
########################################################
#
# Copyright (c) 2003-2012 by University of Queensland
# Earth Systems Science Computational Center (ESSCC)
# http://www.uq.edu.au/esscc
#
# Primary Business: Queensland, Australia
# Licensed under the Open Software License version 3.0
# http://www.opensource.org/licenses/osl-3.0.php
#
########################################################

__copyright__="""Copyright (c) 2003-2012 by University of Queensland
Earth Systems Science Computational Center (ESSCC)
http://www.uq.edu.au/esscc
Primary Business: Queensland, Australia"""
__license__="""Licensed under the Open Software License version 3.0
http://www.opensource.org/licenses/osl-3.0.php"""
__url__="https://launchpad.net/escript-finley"

"""
basic tests for functions in util.py effecting the spatial distribution 

it is assumed that the domain is the usint square/cube 

not all these test will run for all domains. check the doc string for the assumptions of a particular test

:var __author__: name of author
:var __copyright__: copyrights
:var __license__: licence agreement
:var __url__: url entry point on documentation
:var __version__: version
:var __date__: date of the version
"""
__author__="Lutz Gross, l.gross@uq.edu.au"

import unittest
from esys.escript import *
from numpy import array


class Test_Util_SpatialFunctions_noGradOnBoundary_noContact(unittest.TestCase):
   RES_TOL=1.e-8
   def test_x_ofDomain(self):
     """
     test getX() of the domain to be in the [0,1]^dim box
     """
     dim=self.domain.getDim()
     x=self.domain.getX()
     self.assertEqual(x.getShape(),(dim,),"wrong shape of result.")
     self.assertAlmostEqual(inf(x[0]),0.,int(-log10(self.RES_TOL)),"min x0 wrong")
     self.assertAlmostEqual(sup(x[0]),1.,int(-log10(self.RES_TOL)),"max x0 wrong")
     self.assertAlmostEqual(inf(x[1]),0.,int(-log10(self.RES_TOL)),"min x1 wrong")
     self.assertAlmostEqual(sup(x[1]),1.,int(-log10(self.RES_TOL)),"max x1 wrong")
     if dim>2:
       self.assertAlmostEqual(inf(x[2]),0.,int(-log10(self.RES_TOL)),"min x2 wrong")
       self.assertAlmostEqual(sup(x[2]),1.,int(-log10(self.RES_TOL)),"max x2 wrong")

   def test_SolutionOrder(self):
      """
      test the approximation order 
      """
      self.assertTrue(self.order == Solution(self.domain).getApproximationOrder(), "wrong order (Solution)")
      self.assertTrue(self.order == ContinuousFunction(self.domain).getApproximationOrder(), "wrong order (continuous function)")
      self.assertTrue(1 == ReducedSolution(self.domain).getApproximationOrder(), "wrong order (ReducedSolution)")
      self.assertTrue(1 == ReducedContinuousFunction(self.domain).getApproximationOrder(), "wrong order (Reduced continuous function)")
      for i in range(self.domain.getDim()):
         for k in range(Function(self.domain).getApproximationOrder()+1):
             self.assertAlmostEqual(integrate(Function(self.domain).getX()[i]**k),1./(k+1),8,"wrong integral (i=%s, order = %s)"%(i,k))
         for k in range(ReducedFunction(self.domain).getApproximationOrder()+1):
             self.assertAlmostEqual(integrate(ReducedFunction(self.domain).getX()[i]**k),1./(k+1),8,"wrong integral (i=%s, order = %s (reduced))"%(i,k))


   def test_normal_FunctionOnBoundary(self):
     """
     test getNormal() on boundary

     assumptions: FunctionOnBoundary(self.domain) exists
     """
     dim=self.domain.getDim()
     f=FunctionOnBoundary(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref.setTaggedValue(200,[0,0,1])
         ref.setTaggedValue(100,[0,0,-1])
         ref.setTaggedValue(20,[0,1,0])
         ref.setTaggedValue(10,[0,-1,0])
         ref.setTaggedValue(2,[1,0,0])
         ref.setTaggedValue(1,[-1,0,0])
     else:
         ref.setTaggedValue(2,[1,0])
         ref.setTaggedValue(1,[-1,0])
         ref.setTaggedValue(20, [0,1])
         ref.setTaggedValue(10, [0,-1])

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertTrue(Lsup(ref-res)<=self.RES_TOL,"wrong result")

   def test_normal_ReducedFunctionOnBoundary(self):
     """
     test getNormal() on boundary

     assumptions: FunctionOnBoundary(self.domain) exists
     """
     dim=self.domain.getDim()
     f=ReducedFunctionOnBoundary(self.domain)
     x=f.getX()
     ref=Vector(0.,what=f)
     if dim==3:
         ref.setTaggedValue(200,[0,0,1])
         ref.setTaggedValue(100,[0,0,-1])
         ref.setTaggedValue(20,[0,1,0])
         ref.setTaggedValue(10,[0,-1,0])
         ref.setTaggedValue(2,[1,0,0])
         ref.setTaggedValue(1,[-1,0,0])
     else:
         ref.setTaggedValue(2,[1,0])
         ref.setTaggedValue(1,[-1,0])
         ref.setTaggedValue(20, [0,1])
         ref.setTaggedValue(10, [0,-1])

     res=f.getNormal()
     self.assertEqual(res.getShape(),(dim,),"wrong shape of result.")
     self.assertEqual(res.getFunctionSpace(),f,"wrong functionspace of result.")
     self.assertTrue(Lsup(ref-res)<=self.RES_TOL,"wrong result")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.465976690369)*x[0]**o+(-0.25136070196)*x[0]+(0.409576841566)*x[1]**o+(0.726707772754)*x[1]
        ref=(0.875553531935)/(o+1.)+(0.237673535397)
      else:
        arg=(0.567785895928)*x[0]**o+(0.928477128396)*x[0]+(-0.625720943088)*x[1]**o+(-0.681460841298)*x[1]+(-0.866801010069)*x[2]**o+(0.159594300637)*x[2]
        ref=(-0.924736057229)/(o+1.)+(0.203305293867)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.839020088668)*x[0]**o+(-0.880288035177)*x[0]+(0.260951734364)*x[1]**o+(-0.624208000982)*x[1]
        arg[1]=(0.985421831314)*x[0]**o+(0.660399229451)*x[0]+(-0.514003848134)*x[1]**o+(0.0312162082593)*x[1]
        ref[0]=(-0.578068354304)/(o+1.)+(-0.75224801808)
        ref[1]=(0.47141798318)/(o+1.)+(0.345807718855)
      else:
        arg[0]=(-0.627307316929)*x[0]**o+(0.500432411035)*x[0]+(0.233895908313)*x[1]**o+(-0.125048135432)*x[1]+(0.0695114009632)*x[2]**o+(0.995170977074)*x[2]
        arg[1]=(0.760033715785)*x[0]**o+(0.48895905139)*x[0]+(0.240372023176)*x[1]**o+(0.33946689896)*x[1]+(-0.774638489419)*x[2]**o+(-0.0777240871074)*x[2]
        ref[0]=(-0.323900007653)/(o+1.)+(0.685277626338)
        ref[1]=(0.225767249542)/(o+1.)+(0.375350931622)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.844492048883)*x[0]**o+(0.727132679659)*x[0]+(-0.862576284722)*x[1]**o+(0.502937561417)*x[1]
        arg[0,1]=(-0.587157419309)*x[0]**o+(0.00327068097107)*x[0]+(0.422592349683)*x[1]**o+(0.167479615369)*x[1]
        arg[0,2]=(0.771156665672)*x[0]**o+(-0.371091138833)*x[0]+(-0.867750378658)*x[1]**o+(-0.744394290185)*x[1]
        arg[0,3]=(0.669718598855)*x[0]**o+(-0.0825909409121)*x[0]+(0.868872948791)*x[1]**o+(0.141891982204)*x[1]
        arg[0,4]=(-0.441072764653)*x[0]**o+(-0.929654649265)*x[0]+(-0.35883498816)*x[1]**o+(-0.84045480381)*x[1]
        arg[1,0]=(0.555668196622)*x[0]**o+(0.94306681043)*x[0]+(0.586512446245)*x[1]**o+(0.702007276327)*x[1]
        arg[1,1]=(0.792083771005)*x[0]**o+(-0.865830139617)*x[0]+(0.0929971631615)*x[1]**o+(0.620446289121)*x[1]
        arg[1,2]=(-0.201978375472)*x[0]**o+(-0.0777206885918)*x[0]+(0.798755459687)*x[1]**o+(-0.278086468846)*x[1]
        arg[1,3]=(0.063028419572)*x[0]**o+(-0.454786232294)*x[0]+(-0.0481393324293)*x[1]**o+(-0.612657132834)*x[1]
        arg[1,4]=(-0.743283787964)*x[0]**o+(0.854827303846)*x[0]+(0.94241635825)*x[1]**o+(0.436937526521)*x[1]
        arg[2,0]=(-0.814975695669)*x[0]**o+(0.579477783769)*x[0]+(0.00998264005965)*x[1]**o+(-0.913665878657)*x[1]
        arg[2,1]=(-0.526833976369)*x[0]**o+(-0.800386086232)*x[0]+(-0.487600053086)*x[1]**o+(0.00242562519003)*x[1]
        arg[2,2]=(-0.964119591874)*x[0]**o+(0.203212394214)*x[0]+(-0.96850837954)*x[1]**o+(-0.214886132512)*x[1]
        arg[2,3]=(-0.356249104564)*x[0]**o+(0.329769293221)*x[0]+(0.936664928754)*x[1]**o+(0.520603503283)*x[1]
        arg[2,4]=(0.255464219424)*x[0]**o+(-0.157783294516)*x[0]+(0.519451787865)*x[1]**o+(0.00960521397376)*x[1]
        arg[3,0]=(-0.117250486034)*x[0]**o+(0.192945431273)*x[0]+(-0.7580154297)*x[1]**o+(-0.308801650294)*x[1]
        arg[3,1]=(0.243931002845)*x[0]**o+(-0.626714191745)*x[0]+(-0.365870571265)*x[1]**o+(0.652257582021)*x[1]
        arg[3,2]=(-0.769756656122)*x[0]**o+(0.711106390017)*x[0]+(0.150749630653)*x[1]**o+(-0.0441309732448)*x[1]
        arg[3,3]=(0.92677837493)*x[0]**o+(0.507199859313)*x[0]+(-0.542275490828)*x[1]**o+(-0.0907270087228)*x[1]
        arg[3,4]=(-0.0306175885885)*x[0]**o+(-0.434666505018)*x[0]+(-0.815633427562)*x[1]**o+(-0.916424797903)*x[1]
        ref[0,0]=(-1.70706833361)/(o+1.)+(0.615035120538)
        ref[0,1]=(-0.164565069626)/(o+1.)+(0.08537514817)
        ref[0,2]=(-0.0965937129851)/(o+1.)+(-0.557742714509)
        ref[0,3]=(1.53859154765)/(o+1.)+(0.0296505206459)
        ref[0,4]=(-0.799907752814)/(o+1.)+(-0.885054726537)
        ref[1,0]=(1.14218064287)/(o+1.)+(0.822537043379)
        ref[1,1]=(0.885080934167)/(o+1.)+(-0.122691925248)
        ref[1,2]=(0.596777084215)/(o+1.)+(-0.177903578719)
        ref[1,3]=(0.0148890871427)/(o+1.)+(-0.533721682564)
        ref[1,4]=(0.199132570285)/(o+1.)+(0.645882415184)
        ref[2,0]=(-0.80499305561)/(o+1.)+(-0.167094047444)
        ref[2,1]=(-1.01443402945)/(o+1.)+(-0.398980230521)
        ref[2,2]=(-1.93262797141)/(o+1.)+(-0.00583686914904)
        ref[2,3]=(0.580415824189)/(o+1.)+(0.425186398252)
        ref[2,4]=(0.774916007289)/(o+1.)+(-0.074089040271)
        ref[3,0]=(-0.875265915734)/(o+1.)+(-0.0579281095106)
        ref[3,1]=(-0.12193956842)/(o+1.)+(0.0127716951382)
        ref[3,2]=(-0.619007025469)/(o+1.)+(0.333487708386)
        ref[3,3]=(0.384502884102)/(o+1.)+(0.208236425295)
        ref[3,4]=(-0.846251016151)/(o+1.)+(-0.675545651461)
      else:
        arg[0,0]=(0.596538443259)*x[0]**o+(0.479462980373)*x[0]+(-0.334469126403)*x[1]**o+(0.784201085556)*x[1]+(-0.231474140531)*x[2]**o+(-0.368096345589)*x[2]
        arg[0,1]=(0.188945383338)*x[0]**o+(-0.95033024122)*x[0]+(0.567904901047)*x[1]**o+(-0.76703296351)*x[1]+(-0.0116083690376)*x[2]**o+(-0.211176589764)*x[2]
        arg[0,2]=(0.225412552449)*x[0]**o+(-0.731615155924)*x[0]+(-0.362727501637)*x[1]**o+(-0.195442119344)*x[1]+(-0.215444456285)*x[2]**o+(0.412715796392)*x[2]
        arg[0,3]=(-0.00941287481408)*x[0]**o+(-0.207271678463)*x[0]+(0.66582433583)*x[1]**o+(-0.297871484308)*x[1]+(-0.819460465536)*x[2]**o+(-0.3316307033)*x[2]
        arg[0,4]=(-0.172265683581)*x[0]**o+(0.881658164973)*x[0]+(0.0267816673708)*x[1]**o+(-0.526757427521)*x[1]+(-0.666953449252)*x[2]**o+(0.366533690002)*x[2]
        arg[1,0]=(-0.575272180021)*x[0]**o+(-0.707236339163)*x[0]+(-0.734723292219)*x[1]**o+(-0.872019568084)*x[1]+(-0.513269190461)*x[2]**o+(-0.95077853535)*x[2]
        arg[1,1]=(0.687652562435)*x[0]**o+(0.500668143461)*x[0]+(0.127678854367)*x[1]**o+(0.135196530391)*x[1]+(0.219757431874)*x[2]**o+(0.43373842168)*x[2]
        arg[1,2]=(-0.674120744855)*x[0]**o+(-0.792769889708)*x[0]+(0.0414602844391)*x[1]**o+(-0.159887164563)*x[1]+(0.406441555189)*x[2]**o+(-0.662210077269)*x[2]
        arg[1,3]=(-0.623120612358)*x[0]**o+(0.895951757448)*x[0]+(0.957407224559)*x[1]**o+(-0.472225118848)*x[1]+(-0.294551998847)*x[2]**o+(0.503584320886)*x[2]
        arg[1,4]=(-0.304220645545)*x[0]**o+(0.812321371589)*x[0]+(0.932257196386)*x[1]**o+(0.736014913915)*x[1]+(0.119140391025)*x[2]**o+(0.910696940879)*x[2]
        arg[2,0]=(0.275470013797)*x[0]**o+(-0.903754203548)*x[0]+(-0.443211232771)*x[1]**o+(0.78815938507)*x[1]+(-0.0802934106722)*x[2]**o+(0.096532189886)*x[2]
        arg[2,1]=(-0.4505392488)*x[0]**o+(-0.668542586506)*x[0]+(0.487967999984)*x[1]**o+(-0.642310745404)*x[1]+(-0.956530771836)*x[2]**o+(0.517544404873)*x[2]
        arg[2,2]=(-0.984032885692)*x[0]**o+(-0.606688527142)*x[0]+(0.0565483776156)*x[1]**o+(-0.945228520749)*x[1]+(0.433070187921)*x[2]**o+(-0.599983370996)*x[2]
        arg[2,3]=(-0.673469008741)*x[0]**o+(0.612144012035)*x[0]+(-0.169497339443)*x[1]**o+(0.267990057152)*x[1]+(0.431716413548)*x[2]**o+(0.188904436012)*x[2]
        arg[2,4]=(0.320782309844)*x[0]**o+(-0.830068082984)*x[0]+(0.900183436056)*x[1]**o+(-0.343951186189)*x[1]+(-0.00137135137854)*x[2]**o+(-0.655764522249)*x[2]
        arg[3,0]=(-0.936193981767)*x[0]**o+(0.885260284217)*x[0]+(0.00397725971098)*x[1]**o+(-0.0864751234501)*x[1]+(0.25001166345)*x[2]**o+(-0.127859505667)*x[2]
        arg[3,1]=(-0.0841986843231)*x[0]**o+(-0.95678469673)*x[0]+(0.867049114883)*x[1]**o+(0.154499266839)*x[1]+(0.0217146474622)*x[2]**o+(-0.945774319527)*x[2]
        arg[3,2]=(-0.668161122306)*x[0]**o+(0.779140555525)*x[0]+(0.217613826157)*x[1]**o+(0.900077039607)*x[1]+(-0.251881965789)*x[2]**o+(-0.331281723354)*x[2]
        arg[3,3]=(-0.95624079226)*x[0]**o+(-0.403226291225)*x[0]+(0.303340462877)*x[1]**o+(-0.269819286355)*x[1]+(0.605804508829)*x[2]**o+(-0.461404715818)*x[2]
        arg[3,4]=(-0.933059329973)*x[0]**o+(-0.00851461368762)*x[0]+(0.0588544443918)*x[1]**o+(-0.682939702389)*x[1]+(-0.213965742261)*x[2]**o+(-0.102965977964)*x[2]
        ref[0,0]=(0.0305951763245)/(o+1.)+(0.44778386017)
        ref[0,1]=(0.745241915348)/(o+1.)+(-0.964269897247)
        ref[0,2]=(-0.352759405473)/(o+1.)+(-0.257170739438)
        ref[0,3]=(-0.163049004521)/(o+1.)+(-0.418386933036)
        ref[0,4]=(-0.812437465462)/(o+1.)+(0.360717213727)
        ref[1,0]=(-1.8232646627)/(o+1.)+(-1.2650172213)
        ref[1,1]=(1.03508884868)/(o+1.)+(0.534801547766)
        ref[1,2]=(-0.226218905227)/(o+1.)+(-0.80743356577)
        ref[1,3]=(0.0397346133545)/(o+1.)+(0.463655479743)
        ref[1,4]=(0.747176941866)/(o+1.)+(1.22951661319)
        ref[2,0]=(-0.248034629647)/(o+1.)+(-0.00953131429598)
        ref[2,1]=(-0.919102020652)/(o+1.)+(-0.396654463518)
        ref[2,2]=(-0.494414320156)/(o+1.)+(-1.07595020944)
        ref[2,3]=(-0.411249934637)/(o+1.)+(0.534519252599)
        ref[2,4]=(1.21959439452)/(o+1.)+(-0.914891895711)
        ref[3,0]=(-0.682205058606)/(o+1.)+(0.33546282755)
        ref[3,1]=(0.804565078022)/(o+1.)+(-0.874029874709)
        ref[3,2]=(-0.702429261938)/(o+1.)+(0.673967935889)
        ref[3,3]=(-0.0470958205546)/(o+1.)+(-0.567225146699)
        ref[3,4]=(-1.08817062784)/(o+1.)+(-0.39721014702)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.401537399178)*x[0]**o+(-0.868937065568)*x[0]+(0.356734884346)*x[1]**o+(0.0137485089422)*x[1]
        arg[0,0,1]=(0.705383698155)*x[0]**o+(-0.884988282352)*x[0]+(-0.979185994641)*x[1]**o+(0.929576942519)*x[1]
        arg[0,1,0]=(-0.292540086272)*x[0]**o+(-0.223230840915)*x[0]+(0.717291109855)*x[1]**o+(-0.826233591275)*x[1]
        arg[0,1,1]=(0.55739919643)*x[0]**o+(0.2571407205)*x[0]+(0.00434404142287)*x[1]**o+(0.200305144451)*x[1]
        arg[1,0,0]=(0.451841947443)*x[0]**o+(-0.576279050146)*x[0]+(-0.128142118055)*x[1]**o+(0.316406960557)*x[1]
        arg[1,0,1]=(-0.07174792925)*x[0]**o+(-0.392935435009)*x[0]+(-0.343535485172)*x[1]**o+(0.619131193274)*x[1]
        arg[1,1,0]=(-0.667031669611)*x[0]**o+(-0.184806094874)*x[0]+(-0.199375296384)*x[1]**o+(0.449351889815)*x[1]
        arg[1,1,1]=(-0.51924508404)*x[0]**o+(0.870508122509)*x[0]+(-0.274149433148)*x[1]**o+(-0.762367215962)*x[1]
        arg[2,0,0]=(-0.501546909709)*x[0]**o+(0.164122129343)*x[0]+(-0.601370828214)*x[1]**o+(0.888046948888)*x[1]
        arg[2,0,1]=(0.272472486899)*x[0]**o+(-0.958776296738)*x[0]+(-0.334676045163)*x[1]**o+(0.410943596779)*x[1]
        arg[2,1,0]=(0.363478201368)*x[0]**o+(0.753492598803)*x[0]+(-0.663726768274)*x[1]**o+(0.875443816963)*x[1]
        arg[2,1,1]=(0.995788536211)*x[0]**o+(-0.123932097555)*x[0]+(0.143025171293)*x[1]**o+(0.724918059214)*x[1]
        arg[3,0,0]=(-0.578717829755)*x[0]**o+(-0.899523186929)*x[0]+(0.58523203854)*x[1]**o+(0.751620714771)*x[1]
        arg[3,0,1]=(0.179841033744)*x[0]**o+(0.268015839688)*x[0]+(0.860292991989)*x[1]**o+(0.0117969420991)*x[1]
        arg[3,1,0]=(-0.575777000777)*x[0]**o+(0.170047012345)*x[0]+(-0.418743243278)*x[1]**o+(-0.278415849566)*x[1]
        arg[3,1,1]=(-0.913076795648)*x[0]**o+(0.486330450107)*x[0]+(-0.423570363453)*x[1]**o+(0.388575474946)*x[1]
        arg[4,0,0]=(0.16695368621)*x[0]**o+(0.133497460585)*x[0]+(-0.444434659151)*x[1]**o+(-0.895229636275)*x[1]
        arg[4,0,1]=(0.432018868106)*x[0]**o+(-0.281976075625)*x[0]+(-0.0455135209769)*x[1]**o+(0.76870037059)*x[1]
        arg[4,1,0]=(0.238542836112)*x[0]**o+(-0.806931884409)*x[0]+(-0.0365533871189)*x[1]**o+(0.731336634477)*x[1]
        arg[4,1,1]=(0.512035255059)*x[0]**o+(-0.683560978902)*x[0]+(-0.953782277737)*x[1]**o+(-0.311664210765)*x[1]
        arg[5,0,0]=(-0.842399299624)*x[0]**o+(0.460676816656)*x[0]+(-0.251781603841)*x[1]**o+(0.321792359074)*x[1]
        arg[5,0,1]=(-0.900576257062)*x[0]**o+(0.842393710528)*x[0]+(0.630445446843)*x[1]**o+(-0.647745144739)*x[1]
        arg[5,1,0]=(0.136687908652)*x[0]**o+(0.0252907890641)*x[0]+(0.813054611844)*x[1]**o+(-0.459930787056)*x[1]
        arg[5,1,1]=(-0.00242518550941)*x[0]**o+(-0.558785348137)*x[0]+(0.0568659980522)*x[1]**o+(0.689293614051)*x[1]
        ref[0,0,0]=(0.758272283524)/(o+1.)+(-0.427594278313)
        ref[0,0,1]=(-0.273802296487)/(o+1.)+(0.0222943300833)
        ref[0,1,0]=(0.424751023583)/(o+1.)+(-0.524732216095)
        ref[0,1,1]=(0.561743237852)/(o+1.)+(0.228722932476)
        ref[1,0,0]=(0.323699829388)/(o+1.)+(-0.129936044795)
        ref[1,0,1]=(-0.415283414421)/(o+1.)+(0.113097879133)
        ref[1,1,0]=(-0.866406965995)/(o+1.)+(0.132272897471)
        ref[1,1,1]=(-0.793394517189)/(o+1.)+(0.0540704532731)
        ref[2,0,0]=(-1.10291773792)/(o+1.)+(0.526084539115)
        ref[2,0,1]=(-0.0622035582637)/(o+1.)+(-0.273916349979)
        ref[2,1,0]=(-0.300248566906)/(o+1.)+(0.814468207883)
        ref[2,1,1]=(1.1388137075)/(o+1.)+(0.300492980829)
        ref[3,0,0]=(0.00651420878498)/(o+1.)+(-0.073951236079)
        ref[3,0,1]=(1.04013402573)/(o+1.)+(0.139906390894)
        ref[3,1,0]=(-0.994520244055)/(o+1.)+(-0.0541844186105)
        ref[3,1,1]=(-1.3366471591)/(o+1.)+(0.437452962527)
        ref[4,0,0]=(-0.27748097294)/(o+1.)+(-0.380866087845)
        ref[4,0,1]=(0.386505347129)/(o+1.)+(0.243362147483)
        ref[4,1,0]=(0.201989448993)/(o+1.)+(-0.0377976249661)
        ref[4,1,1]=(-0.441747022678)/(o+1.)+(-0.497612594833)
        ref[5,0,0]=(-1.09418090346)/(o+1.)+(0.391234587865)
        ref[5,0,1]=(-0.270130810219)/(o+1.)+(0.0973242828949)
        ref[5,1,0]=(0.949742520497)/(o+1.)+(-0.217319998996)
        ref[5,1,1]=(0.0544408125428)/(o+1.)+(0.0652541329568)
      else:
        arg[0,0,0]=(0.411788818613)*x[0]**o+(0.324151698669)*x[0]+(0.656551349616)*x[1]**o+(0.801577401197)*x[1]+(0.690440431022)*x[2]**o+(0.776409870151)*x[2]
        arg[0,0,1]=(-0.224985947659)*x[0]**o+(-0.564703086188)*x[0]+(-0.37196336116)*x[1]**o+(-0.114979853172)*x[1]+(0.274462080499)*x[2]**o+(-0.988489021983)*x[2]
        arg[0,1,0]=(0.512331530009)*x[0]**o+(0.519617843594)*x[0]+(-0.515174861158)*x[1]**o+(-0.675953902994)*x[1]+(0.796111490716)*x[2]**o+(0.229569489858)*x[2]
        arg[0,1,1]=(0.347765144376)*x[0]**o+(-0.0110194730031)*x[0]+(-0.593936645599)*x[1]**o+(-0.372786558089)*x[1]+(-0.604691321362)*x[2]**o+(-0.888002521077)*x[2]
        arg[1,0,0]=(-0.575954903433)*x[0]**o+(0.0160235655616)*x[0]+(0.305335632724)*x[1]**o+(-0.259402553796)*x[1]+(-0.221276854741)*x[2]**o+(-0.519932341069)*x[2]
        arg[1,0,1]=(-0.674987647238)*x[0]**o+(-0.45852667934)*x[0]+(-0.758937440978)*x[1]**o+(0.80566898197)*x[1]+(-0.961619154322)*x[2]**o+(0.861564430243)*x[2]
        arg[1,1,0]=(-0.297475412437)*x[0]**o+(0.881828705604)*x[0]+(-0.823267615874)*x[1]**o+(-0.0232728101313)*x[1]+(-0.676220001453)*x[2]**o+(-0.382414036981)*x[2]
        arg[1,1,1]=(-0.31008204616)*x[0]**o+(0.805195900401)*x[0]+(-0.758209772143)*x[1]**o+(0.861379709073)*x[1]+(0.0631063866518)*x[2]**o+(0.356607080553)*x[2]
        arg[2,0,0]=(0.305214964591)*x[0]**o+(0.815106526139)*x[0]+(0.138797156005)*x[1]**o+(-0.241921896332)*x[1]+(-0.103783035058)*x[2]**o+(-0.161952044984)*x[2]
        arg[2,0,1]=(-0.925375677635)*x[0]**o+(0.0341347568201)*x[0]+(-0.706905924964)*x[1]**o+(-0.193304368451)*x[1]+(-0.449866757104)*x[2]**o+(-0.894963680869)*x[2]
        arg[2,1,0]=(0.822676224279)*x[0]**o+(-0.28758006682)*x[0]+(-0.869025918917)*x[1]**o+(0.0107602044744)*x[1]+(-0.832322765163)*x[2]**o+(0.748440319389)*x[2]
        arg[2,1,1]=(-0.415326508997)*x[0]**o+(-0.879521424817)*x[0]+(-0.971781441062)*x[1]**o+(0.527586634108)*x[1]+(0.336682669856)*x[2]**o+(-0.948414041149)*x[2]
        arg[3,0,0]=(-0.345788531715)*x[0]**o+(0.626583498807)*x[0]+(-0.186418683649)*x[1]**o+(0.3141628424)*x[1]+(-0.781119787783)*x[2]**o+(0.972708430266)*x[2]
        arg[3,0,1]=(0.175536279095)*x[0]**o+(-0.150027774295)*x[0]+(-0.506653680683)*x[1]**o+(-0.482544397029)*x[1]+(-0.59090996152)*x[2]**o+(0.618123625026)*x[2]
        arg[3,1,0]=(0.737984493124)*x[0]**o+(0.650901680297)*x[0]+(-0.361389200009)*x[1]**o+(-0.784304665396)*x[1]+(-0.954132232142)*x[2]**o+(0.684341934348)*x[2]
        arg[3,1,1]=(-0.385757579645)*x[0]**o+(0.251136919845)*x[0]+(0.95143939933)*x[1]**o+(0.501860043225)*x[1]+(-0.675673011184)*x[2]**o+(0.337315450579)*x[2]
        arg[4,0,0]=(-0.740183402139)*x[0]**o+(-0.828278760743)*x[0]+(0.88000889117)*x[1]**o+(-0.0226801022184)*x[1]+(0.6811397662)*x[2]**o+(-0.254627446498)*x[2]
        arg[4,0,1]=(-0.194128952361)*x[0]**o+(-0.524854896388)*x[0]+(0.341100326577)*x[1]**o+(-0.26072443499)*x[1]+(-0.197396331776)*x[2]**o+(-0.589037632112)*x[2]
        arg[4,1,0]=(0.261995024865)*x[0]**o+(0.214193793699)*x[0]+(0.572024600244)*x[1]**o+(0.230659327198)*x[1]+(0.420971705453)*x[2]**o+(0.333445049827)*x[2]
        arg[4,1,1]=(-0.0679525054395)*x[0]**o+(0.0756646296473)*x[0]+(-0.299865836239)*x[1]**o+(-0.828294764914)*x[1]+(0.606170745631)*x[2]**o+(0.624267827393)*x[2]
        arg[5,0,0]=(0.312565542211)*x[0]**o+(-0.368815496171)*x[0]+(-0.0615973473105)*x[1]**o+(0.205012835971)*x[1]+(0.310818553076)*x[2]**o+(-0.287393908304)*x[2]
        arg[5,0,1]=(-0.409626309069)*x[0]**o+(0.977953573562)*x[0]+(0.583013544138)*x[1]**o+(0.542907128677)*x[1]+(0.488538336366)*x[2]**o+(-0.85823611123)*x[2]
        arg[5,1,0]=(-0.899607109395)*x[0]**o+(-0.0459053040522)*x[0]+(0.938423293041)*x[1]**o+(-0.112431665317)*x[1]+(0.620609328687)*x[2]**o+(-0.663082429319)*x[2]
        arg[5,1,1]=(0.00370253144626)*x[0]**o+(0.503733497956)*x[0]+(0.362457785043)*x[1]**o+(0.417891957303)*x[1]+(0.404543677829)*x[2]**o+(0.117828669122)*x[2]
        ref[0,0,0]=(1.75878059925)/(o+1.)+(0.951069485009)
        ref[0,0,1]=(-0.32248722832)/(o+1.)+(-0.834085980672)
        ref[0,1,0]=(0.793268159567)/(o+1.)+(0.036616715229)
        ref[0,1,1]=(-0.850862822585)/(o+1.)+(-0.635904276085)
        ref[1,0,0]=(-0.49189612545)/(o+1.)+(-0.381655664652)
        ref[1,0,1]=(-2.39554424254)/(o+1.)+(0.604353366437)
        ref[1,1,0]=(-1.79696302976)/(o+1.)+(0.238070929246)
        ref[1,1,1]=(-1.00518543165)/(o+1.)+(1.01159134501)
        ref[2,0,0]=(0.340229085537)/(o+1.)+(0.205616292412)
        ref[2,0,1]=(-2.0821483597)/(o+1.)+(-0.52706664625)
        ref[2,1,0]=(-0.878672459802)/(o+1.)+(0.235810228522)
        ref[2,1,1]=(-1.0504252802)/(o+1.)+(-0.650174415929)
        ref[3,0,0]=(-1.31332700315)/(o+1.)+(0.956727385736)
        ref[3,0,1]=(-0.922027363109)/(o+1.)+(-0.00722427314892)
        ref[3,1,0]=(-0.577536939027)/(o+1.)+(0.275469474624)
        ref[3,1,1]=(-0.109991191499)/(o+1.)+(0.545156206825)
        ref[4,0,0]=(0.820965255231)/(o+1.)+(-0.55279315473)
        ref[4,0,1]=(-0.0504249575602)/(o+1.)+(-0.687308481745)
        ref[4,1,0]=(1.25499133056)/(o+1.)+(0.389149085362)
        ref[4,1,1]=(0.238352403952)/(o+1.)+(-0.0641811539369)
        ref[5,0,0]=(0.561786747976)/(o+1.)+(-0.225598284252)
        ref[5,0,1]=(0.661925571434)/(o+1.)+(0.331312295504)
        ref[5,1,0]=(0.659425512333)/(o+1.)+(-0.410709699344)
        ref[5,1,1]=(0.770703994319)/(o+1.)+(0.519727062191)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.164405890316)*x[0]**o+(-0.323303939764)*x[0]+(-0.0516829984856)*x[1]**o+(-0.220765052533)*x[1]
        arg[0,0,0,1]=(0.152172612108)*x[0]**o+(-0.166242164365)*x[0]+(0.807177808571)*x[1]**o+(0.0404817109573)*x[1]
        arg[0,0,1,0]=(0.016912808567)*x[0]**o+(0.623923316774)*x[0]+(-0.542886934905)*x[1]**o+(0.0110784560205)*x[1]
        arg[0,0,1,1]=(-0.663885786909)*x[0]**o+(-0.55563355485)*x[0]+(0.191993284382)*x[1]**o+(0.734973556043)*x[1]
        arg[0,0,2,0]=(-0.110316754183)*x[0]**o+(-0.584022783672)*x[0]+(0.56471376011)*x[1]**o+(0.846317747143)*x[1]
        arg[0,0,2,1]=(-0.767988009278)*x[0]**o+(0.440997912352)*x[0]+(0.45366348093)*x[1]**o+(0.376620990318)*x[1]
        arg[0,1,0,0]=(-0.138006508051)*x[0]**o+(0.8450564766)*x[0]+(-0.273559897633)*x[1]**o+(0.683678615645)*x[1]
        arg[0,1,0,1]=(-0.467521625474)*x[0]**o+(0.32289306962)*x[0]+(0.340974562152)*x[1]**o+(-0.58109391053)*x[1]
        arg[0,1,1,0]=(0.76867991208)*x[0]**o+(0.683527735131)*x[0]+(0.762986872389)*x[1]**o+(-0.189342734286)*x[1]
        arg[0,1,1,1]=(0.125150175351)*x[0]**o+(-0.175190872876)*x[0]+(-0.569766488076)*x[1]**o+(0.787503755206)*x[1]
        arg[0,1,2,0]=(-0.339695860559)*x[0]**o+(0.119452596206)*x[0]+(0.474378864192)*x[1]**o+(0.845982555265)*x[1]
        arg[0,1,2,1]=(0.820551334836)*x[0]**o+(-0.200024721285)*x[0]+(-0.525871855871)*x[1]**o+(-0.714639297791)*x[1]
        arg[0,2,0,0]=(-0.145788788262)*x[0]**o+(-0.695717636982)*x[0]+(0.736564332801)*x[1]**o+(0.634307016587)*x[1]
        arg[0,2,0,1]=(-0.25142006947)*x[0]**o+(0.027028767127)*x[0]+(-0.614777438028)*x[1]**o+(-0.691417785617)*x[1]
        arg[0,2,1,0]=(0.366081140839)*x[0]**o+(0.477490533426)*x[0]+(-0.116611335116)*x[1]**o+(-0.863552165712)*x[1]
        arg[0,2,1,1]=(-0.467659558308)*x[0]**o+(0.0836299545334)*x[0]+(0.903093252833)*x[1]**o+(-0.956858182533)*x[1]
        arg[0,2,2,0]=(0.870921031579)*x[0]**o+(-0.801681014465)*x[0]+(-0.951819608329)*x[1]**o+(-0.626222917084)*x[1]
        arg[0,2,2,1]=(0.414973183041)*x[0]**o+(-0.444580205703)*x[0]+(-0.346458764208)*x[1]**o+(-0.679140100745)*x[1]
        arg[0,3,0,0]=(-0.725326935158)*x[0]**o+(-0.914034759513)*x[0]+(0.702614377806)*x[1]**o+(0.819004276317)*x[1]
        arg[0,3,0,1]=(0.617099932352)*x[0]**o+(-0.155182850011)*x[0]+(-0.066428715129)*x[1]**o+(-0.286480585435)*x[1]
        arg[0,3,1,0]=(0.710942545488)*x[0]**o+(-0.734509277268)*x[0]+(-0.807474546329)*x[1]**o+(0.307631760449)*x[1]
        arg[0,3,1,1]=(-0.709966924732)*x[0]**o+(-0.462297022374)*x[0]+(0.773787774038)*x[1]**o+(0.762137020991)*x[1]
        arg[0,3,2,0]=(-0.425660932209)*x[0]**o+(-0.562130300181)*x[0]+(-0.516719014647)*x[1]**o+(-0.547832887369)*x[1]
        arg[0,3,2,1]=(0.149573176879)*x[0]**o+(-0.208999629436)*x[0]+(-0.191426033553)*x[1]**o+(-0.119046538873)*x[1]
        arg[0,4,0,0]=(-0.99932317151)*x[0]**o+(-0.832461558068)*x[0]+(-0.847073271676)*x[1]**o+(0.882046887729)*x[1]
        arg[0,4,0,1]=(0.14235557233)*x[0]**o+(-0.352572053276)*x[0]+(-0.231989519255)*x[1]**o+(0.146855135398)*x[1]
        arg[0,4,1,0]=(0.550561449012)*x[0]**o+(0.530745681425)*x[0]+(0.198889689472)*x[1]**o+(0.165503108658)*x[1]
        arg[0,4,1,1]=(0.059663401309)*x[0]**o+(0.168377411751)*x[0]+(-0.224682723925)*x[1]**o+(-0.255406907757)*x[1]
        arg[0,4,2,0]=(-0.747776894683)*x[0]**o+(-0.281763358533)*x[0]+(0.660369838923)*x[1]**o+(0.348541234687)*x[1]
        arg[0,4,2,1]=(-0.886248366716)*x[0]**o+(0.665256341312)*x[0]+(-0.118041122298)*x[1]**o+(0.192284814559)*x[1]
        arg[1,0,0,0]=(-0.212291422817)*x[0]**o+(-0.16123490382)*x[0]+(-0.884220059537)*x[1]**o+(-0.40104775897)*x[1]
        arg[1,0,0,1]=(0.389745305389)*x[0]**o+(0.544302066739)*x[0]+(-0.504770065056)*x[1]**o+(0.903371132654)*x[1]
        arg[1,0,1,0]=(0.983764499357)*x[0]**o+(0.0707766278425)*x[0]+(-0.119851167133)*x[1]**o+(0.668490934432)*x[1]
        arg[1,0,1,1]=(-0.61333738947)*x[0]**o+(-0.515837139916)*x[0]+(-0.825015725806)*x[1]**o+(-0.858420009384)*x[1]
        arg[1,0,2,0]=(0.838488008532)*x[0]**o+(0.320020900158)*x[0]+(0.265140913892)*x[1]**o+(0.493460973766)*x[1]
        arg[1,0,2,1]=(-0.957171379585)*x[0]**o+(0.911667647687)*x[0]+(0.0634495769152)*x[1]**o+(-0.244783653159)*x[1]
        arg[1,1,0,0]=(0.854312398231)*x[0]**o+(0.0177604750376)*x[0]+(-0.263135825468)*x[1]**o+(0.196452389095)*x[1]
        arg[1,1,0,1]=(0.496952473267)*x[0]**o+(-0.505581680291)*x[0]+(0.227653536528)*x[1]**o+(-0.138354150837)*x[1]
        arg[1,1,1,0]=(0.569682536816)*x[0]**o+(-0.523174667456)*x[0]+(-0.434073415962)*x[1]**o+(-0.853667257479)*x[1]
        arg[1,1,1,1]=(0.636882167235)*x[0]**o+(-0.122206427025)*x[0]+(0.934361025063)*x[1]**o+(-0.43037160177)*x[1]
        arg[1,1,2,0]=(0.0798318147504)*x[0]**o+(0.825231320185)*x[0]+(0.810701650808)*x[1]**o+(-0.544356371381)*x[1]
        arg[1,1,2,1]=(-0.426710170704)*x[0]**o+(0.200697079677)*x[0]+(0.242027752153)*x[1]**o+(0.746762119874)*x[1]
        arg[1,2,0,0]=(-0.497091856216)*x[0]**o+(0.0953816987137)*x[0]+(-0.388615292775)*x[1]**o+(0.794962452203)*x[1]
        arg[1,2,0,1]=(0.484769626964)*x[0]**o+(0.700484219762)*x[0]+(-0.230377697673)*x[1]**o+(-0.110327221848)*x[1]
        arg[1,2,1,0]=(-0.791806541129)*x[0]**o+(0.0199911517557)*x[0]+(-0.441657020836)*x[1]**o+(-0.972999930466)*x[1]
        arg[1,2,1,1]=(-0.566750983493)*x[0]**o+(0.193689535753)*x[0]+(0.659310468709)*x[1]**o+(0.209045229813)*x[1]
        arg[1,2,2,0]=(0.852553900341)*x[0]**o+(-0.869275488172)*x[0]+(-0.308429715369)*x[1]**o+(0.096335995297)*x[1]
        arg[1,2,2,1]=(0.169351018917)*x[0]**o+(0.269577471686)*x[0]+(-0.226840840616)*x[1]**o+(-0.94229540867)*x[1]
        arg[1,3,0,0]=(0.481711489234)*x[0]**o+(0.293576608201)*x[0]+(0.101998881767)*x[1]**o+(-0.10730442531)*x[1]
        arg[1,3,0,1]=(-0.824144820539)*x[0]**o+(0.174754759038)*x[0]+(-0.251803196848)*x[1]**o+(-0.389656255459)*x[1]
        arg[1,3,1,0]=(0.438310633462)*x[0]**o+(-0.268051020546)*x[0]+(0.670456978527)*x[1]**o+(-0.547393352784)*x[1]
        arg[1,3,1,1]=(-0.95082246973)*x[0]**o+(0.739061218443)*x[0]+(0.209495253219)*x[1]**o+(-0.604801908581)*x[1]
        arg[1,3,2,0]=(0.696900724687)*x[0]**o+(0.501191311274)*x[0]+(0.982952568975)*x[1]**o+(0.1345704134)*x[1]
        arg[1,3,2,1]=(0.823865642457)*x[0]**o+(0.344980106206)*x[0]+(0.0740071873515)*x[1]**o+(-0.950460516779)*x[1]
        arg[1,4,0,0]=(-0.428525532814)*x[0]**o+(0.496012936411)*x[0]+(-0.0756953876324)*x[1]**o+(0.8382068924)*x[1]
        arg[1,4,0,1]=(0.950388644852)*x[0]**o+(-0.225116488186)*x[0]+(0.515802878592)*x[1]**o+(-0.51004230039)*x[1]
        arg[1,4,1,0]=(0.868733254296)*x[0]**o+(-0.506544709529)*x[0]+(0.28935579533)*x[1]**o+(0.100277736612)*x[1]
        arg[1,4,1,1]=(-0.104533384722)*x[0]**o+(0.598995288066)*x[0]+(-0.738946772839)*x[1]**o+(-0.23245555216)*x[1]
        arg[1,4,2,0]=(-0.286232179213)*x[0]**o+(-0.31267596034)*x[0]+(0.588681250431)*x[1]**o+(-0.579857978487)*x[1]
        arg[1,4,2,1]=(0.405666376198)*x[0]**o+(0.380029842735)*x[0]+(-0.646672003588)*x[1]**o+(0.419606979519)*x[1]
        arg[2,0,0,0]=(0.183075434304)*x[0]**o+(0.0929532345426)*x[0]+(-0.742007638693)*x[1]**o+(-0.521914593386)*x[1]
        arg[2,0,0,1]=(0.292878118412)*x[0]**o+(-0.427165220789)*x[0]+(0.540033289295)*x[1]**o+(0.997945616043)*x[1]
        arg[2,0,1,0]=(0.872801856197)*x[0]**o+(0.0427577769706)*x[0]+(0.595354626662)*x[1]**o+(-0.878459740609)*x[1]
        arg[2,0,1,1]=(-0.259988866684)*x[0]**o+(0.154043636253)*x[0]+(0.0580047878991)*x[1]**o+(0.40696744056)*x[1]
        arg[2,0,2,0]=(-0.602648518466)*x[0]**o+(0.30877596654)*x[0]+(-0.268563459489)*x[1]**o+(-0.526379663585)*x[1]
        arg[2,0,2,1]=(-0.375777994931)*x[0]**o+(-0.175440011709)*x[0]+(-0.578650003477)*x[1]**o+(0.822157683321)*x[1]
        arg[2,1,0,0]=(-0.44389226502)*x[0]**o+(0.683580346916)*x[0]+(0.471857559282)*x[1]**o+(-0.869624665398)*x[1]
        arg[2,1,0,1]=(0.768272315036)*x[0]**o+(-0.709688321973)*x[0]+(0.200727881948)*x[1]**o+(0.617326752785)*x[1]
        arg[2,1,1,0]=(-0.371169228982)*x[0]**o+(-0.0248248405104)*x[0]+(-0.899904749204)*x[1]**o+(0.796779883083)*x[1]
        arg[2,1,1,1]=(0.220088670807)*x[0]**o+(0.691718163435)*x[0]+(-0.0834793048429)*x[1]**o+(-0.748110928988)*x[1]
        arg[2,1,2,0]=(0.246045293748)*x[0]**o+(-0.669660154233)*x[0]+(0.601040129345)*x[1]**o+(-0.0893792369983)*x[1]
        arg[2,1,2,1]=(0.318708374425)*x[0]**o+(-0.973048583674)*x[0]+(-0.575891882633)*x[1]**o+(0.229050058038)*x[1]
        arg[2,2,0,0]=(0.358008410645)*x[0]**o+(-0.834388697702)*x[0]+(0.0931465346237)*x[1]**o+(0.503934981947)*x[1]
        arg[2,2,0,1]=(-0.257597986495)*x[0]**o+(-0.58070568528)*x[0]+(0.735541935653)*x[1]**o+(-0.704666216263)*x[1]
        arg[2,2,1,0]=(0.0207421250729)*x[0]**o+(-0.724644198871)*x[0]+(0.637124335031)*x[1]**o+(-0.547300138808)*x[1]
        arg[2,2,1,1]=(0.280179217797)*x[0]**o+(-0.387279903944)*x[0]+(0.516275341085)*x[1]**o+(0.781250973385)*x[1]
        arg[2,2,2,0]=(-0.97771111517)*x[0]**o+(-0.140997969015)*x[0]+(-0.495112130955)*x[1]**o+(0.151523726543)*x[1]
        arg[2,2,2,1]=(-0.48638258963)*x[0]**o+(-0.251363723104)*x[0]+(-0.484911659879)*x[1]**o+(-0.841285872322)*x[1]
        arg[2,3,0,0]=(-0.70159761565)*x[0]**o+(-0.661658396735)*x[0]+(-0.380920250292)*x[1]**o+(0.682598882025)*x[1]
        arg[2,3,0,1]=(0.477056088757)*x[0]**o+(0.697397619448)*x[0]+(0.28617323153)*x[1]**o+(-0.0443690469658)*x[1]
        arg[2,3,1,0]=(-0.823038323615)*x[0]**o+(-0.742146866184)*x[0]+(0.258604766787)*x[1]**o+(-0.740368164141)*x[1]
        arg[2,3,1,1]=(0.958751077013)*x[0]**o+(0.353976218746)*x[0]+(-0.954344840728)*x[1]**o+(0.436915389196)*x[1]
        arg[2,3,2,0]=(0.432342075923)*x[0]**o+(-0.0699761039653)*x[0]+(0.701490046186)*x[1]**o+(0.0888793443125)*x[1]
        arg[2,3,2,1]=(0.612536667525)*x[0]**o+(0.107433723026)*x[0]+(-0.611874456965)*x[1]**o+(-0.253338899867)*x[1]
        arg[2,4,0,0]=(-0.0234819568317)*x[0]**o+(-0.0642964170457)*x[0]+(0.556679188908)*x[1]**o+(0.865922901373)*x[1]
        arg[2,4,0,1]=(0.669235019483)*x[0]**o+(0.744361910609)*x[0]+(0.383397302403)*x[1]**o+(0.669587709042)*x[1]
        arg[2,4,1,0]=(-0.143461051192)*x[0]**o+(-0.0272388772724)*x[0]+(0.723868767782)*x[1]**o+(0.303617835707)*x[1]
        arg[2,4,1,1]=(-0.829765013546)*x[0]**o+(-0.784193619495)*x[0]+(0.384376849236)*x[1]**o+(0.280904709362)*x[1]
        arg[2,4,2,0]=(0.38737050935)*x[0]**o+(0.749614613361)*x[0]+(-0.340848293488)*x[1]**o+(0.96291850214)*x[1]
        arg[2,4,2,1]=(0.619896872679)*x[0]**o+(0.970694589398)*x[0]+(-0.300698305892)*x[1]**o+(0.794408153506)*x[1]
        arg[3,0,0,0]=(-0.808442463698)*x[0]**o+(-0.527555255968)*x[0]+(0.246776596713)*x[1]**o+(-0.719809811547)*x[1]
        arg[3,0,0,1]=(-0.390016002326)*x[0]**o+(-0.864168521951)*x[0]+(-0.154649478359)*x[1]**o+(0.210766031283)*x[1]
        arg[3,0,1,0]=(-0.941916382966)*x[0]**o+(-0.0352551994705)*x[0]+(-0.660026082442)*x[1]**o+(-0.869241000235)*x[1]
        arg[3,0,1,1]=(-0.287992443961)*x[0]**o+(0.292551506636)*x[0]+(-0.619953648778)*x[1]**o+(0.746095811695)*x[1]
        arg[3,0,2,0]=(-0.233659077726)*x[0]**o+(-0.568647415596)*x[0]+(-0.22048065705)*x[1]**o+(0.658105381132)*x[1]
        arg[3,0,2,1]=(0.871784854897)*x[0]**o+(-0.118389479971)*x[0]+(0.143290959871)*x[1]**o+(0.753198625961)*x[1]
        arg[3,1,0,0]=(0.993988451174)*x[0]**o+(0.451849763406)*x[0]+(-0.972805640824)*x[1]**o+(0.796893857099)*x[1]
        arg[3,1,0,1]=(-0.424353439197)*x[0]**o+(-0.575799061446)*x[0]+(-0.233936106648)*x[1]**o+(-0.451864349448)*x[1]
        arg[3,1,1,0]=(0.639821887614)*x[0]**o+(0.771041944334)*x[0]+(0.420297869214)*x[1]**o+(0.671734395857)*x[1]
        arg[3,1,1,1]=(0.398121897075)*x[0]**o+(0.647679737081)*x[0]+(-0.544457308885)*x[1]**o+(0.58955287395)*x[1]
        arg[3,1,2,0]=(-0.118222351008)*x[0]**o+(-0.943807502172)*x[0]+(0.866866987138)*x[1]**o+(-0.0793340319339)*x[1]
        arg[3,1,2,1]=(-0.157286204244)*x[0]**o+(-0.789214982692)*x[0]+(0.646008960971)*x[1]**o+(0.866687256647)*x[1]
        arg[3,2,0,0]=(-0.620625598995)*x[0]**o+(-0.12874444072)*x[0]+(0.4402928966)*x[1]**o+(-0.0344340173898)*x[1]
        arg[3,2,0,1]=(-0.714597102332)*x[0]**o+(0.730206152989)*x[0]+(0.795717819655)*x[1]**o+(0.0543647593833)*x[1]
        arg[3,2,1,0]=(0.698721346048)*x[0]**o+(0.599204472096)*x[0]+(0.323529910402)*x[1]**o+(0.362678869119)*x[1]
        arg[3,2,1,1]=(0.976107824554)*x[0]**o+(0.559889392243)*x[0]+(0.240016231718)*x[1]**o+(0.343322878078)*x[1]
        arg[3,2,2,0]=(0.609695396838)*x[0]**o+(0.786331471916)*x[0]+(-0.965489152976)*x[1]**o+(0.553653504493)*x[1]
        arg[3,2,2,1]=(0.897952533418)*x[0]**o+(0.227242627929)*x[0]+(-0.121752169524)*x[1]**o+(0.844617689997)*x[1]
        arg[3,3,0,0]=(-0.249698211819)*x[0]**o+(-0.400095108874)*x[0]+(-0.44300562221)*x[1]**o+(0.387777895056)*x[1]
        arg[3,3,0,1]=(0.626469203641)*x[0]**o+(0.741893669108)*x[0]+(0.261882509899)*x[1]**o+(-0.726628233604)*x[1]
        arg[3,3,1,0]=(-0.525223500134)*x[0]**o+(0.918314547123)*x[0]+(0.426441176651)*x[1]**o+(0.338058005974)*x[1]
        arg[3,3,1,1]=(-0.31358142709)*x[0]**o+(-0.411070225036)*x[0]+(0.0902340021221)*x[1]**o+(0.404135075929)*x[1]
        arg[3,3,2,0]=(0.528842663849)*x[0]**o+(0.474768688828)*x[0]+(0.697643730083)*x[1]**o+(-0.170335125341)*x[1]
        arg[3,3,2,1]=(0.875221054674)*x[0]**o+(-0.441830492507)*x[0]+(-0.60312069438)*x[1]**o+(0.664895429753)*x[1]
        arg[3,4,0,0]=(0.563422972203)*x[0]**o+(-0.595004250424)*x[0]+(0.888927367905)*x[1]**o+(0.553640735136)*x[1]
        arg[3,4,0,1]=(0.79563891979)*x[0]**o+(-0.461626171695)*x[0]+(0.700194671524)*x[1]**o+(-0.687505010217)*x[1]
        arg[3,4,1,0]=(0.655248139354)*x[0]**o+(0.588175755435)*x[0]+(-0.29236593461)*x[1]**o+(0.575186295035)*x[1]
        arg[3,4,1,1]=(0.405730566113)*x[0]**o+(-0.161117488254)*x[0]+(0.912224633472)*x[1]**o+(0.0292423764587)*x[1]
        arg[3,4,2,0]=(-0.405819525278)*x[0]**o+(-0.362840757195)*x[0]+(0.961950188511)*x[1]**o+(0.205213542149)*x[1]
        arg[3,4,2,1]=(0.703066280254)*x[0]**o+(-0.936055321446)*x[0]+(0.861070629413)*x[1]**o+(0.0856428368256)*x[1]
        ref[0,0,0,0]=(-0.216088888802)/(o+1.)+(-0.272034496148)
        ref[0,0,0,1]=(0.959350420679)/(o+1.)+(-0.062880226704)
        ref[0,0,1,0]=(-0.525974126338)/(o+1.)+(0.317500886397)
        ref[0,0,1,1]=(-0.471892502528)/(o+1.)+(0.0896700005967)
        ref[0,0,2,0]=(0.454397005927)/(o+1.)+(0.131147481736)
        ref[0,0,2,1]=(-0.314324528348)/(o+1.)+(0.408809451335)
        ref[0,1,0,0]=(-0.411566405684)/(o+1.)+(0.764367546122)
        ref[0,1,0,1]=(-0.126547063321)/(o+1.)+(-0.129100420455)
        ref[0,1,1,0]=(1.53166678447)/(o+1.)+(0.247092500423)
        ref[0,1,1,1]=(-0.444616312725)/(o+1.)+(0.306156441165)
        ref[0,1,2,0]=(0.134683003633)/(o+1.)+(0.482717575736)
        ref[0,1,2,1]=(0.294679478965)/(o+1.)+(-0.457332009538)
        ref[0,2,0,0]=(0.590775544538)/(o+1.)+(-0.0307053101973)
        ref[0,2,0,1]=(-0.866197507498)/(o+1.)+(-0.332194509245)
        ref[0,2,1,0]=(0.249469805723)/(o+1.)+(-0.193030816143)
        ref[0,2,1,1]=(0.435433694525)/(o+1.)+(-0.436614114)
        ref[0,2,2,0]=(-0.0808985767498)/(o+1.)+(-0.713951965774)
        ref[0,2,2,1]=(0.0685144188336)/(o+1.)+(-0.561860153224)
        ref[0,3,0,0]=(-0.0227125573528)/(o+1.)+(-0.0475152415978)
        ref[0,3,0,1]=(0.550671217223)/(o+1.)+(-0.220831717723)
        ref[0,3,1,0]=(-0.0965320008409)/(o+1.)+(-0.213438758409)
        ref[0,3,1,1]=(0.0638208493064)/(o+1.)+(0.149919999309)
        ref[0,3,2,0]=(-0.942379946856)/(o+1.)+(-0.554981593775)
        ref[0,3,2,1]=(-0.0418528566735)/(o+1.)+(-0.164023084154)
        ref[0,4,0,0]=(-1.84639644319)/(o+1.)+(0.0247926648306)
        ref[0,4,0,1]=(-0.0896339469252)/(o+1.)+(-0.102858458939)
        ref[0,4,1,0]=(0.749451138484)/(o+1.)+(0.348124395042)
        ref[0,4,1,1]=(-0.165019322616)/(o+1.)+(-0.0435147480027)
        ref[0,4,2,0]=(-0.0874070557597)/(o+1.)+(0.033388938077)
        ref[0,4,2,1]=(-1.00428948901)/(o+1.)+(0.428770577935)
        ref[1,0,0,0]=(-1.09651148235)/(o+1.)+(-0.281141331395)
        ref[1,0,0,1]=(-0.115024759667)/(o+1.)+(0.723836599697)
        ref[1,0,1,0]=(0.863913332224)/(o+1.)+(0.369633781137)
        ref[1,0,1,1]=(-1.43835311528)/(o+1.)+(-0.68712857465)
        ref[1,0,2,0]=(1.10362892242)/(o+1.)+(0.406740936962)
        ref[1,0,2,1]=(-0.893721802669)/(o+1.)+(0.333441997264)
        ref[1,1,0,0]=(0.591176572763)/(o+1.)+(0.107106432066)
        ref[1,1,0,1]=(0.724606009795)/(o+1.)+(-0.321967915564)
        ref[1,1,1,0]=(0.135609120854)/(o+1.)+(-0.688420962467)
        ref[1,1,1,1]=(1.5712431923)/(o+1.)+(-0.276289014397)
        ref[1,1,2,0]=(0.890533465558)/(o+1.)+(0.140437474402)
        ref[1,1,2,1]=(-0.184682418551)/(o+1.)+(0.473729599776)
        ref[1,2,0,0]=(-0.88570714899)/(o+1.)+(0.445172075458)
        ref[1,2,0,1]=(0.254391929291)/(o+1.)+(0.295078498957)
        ref[1,2,1,0]=(-1.23346356196)/(o+1.)+(-0.476504389355)
        ref[1,2,1,1]=(0.0925594852161)/(o+1.)+(0.201367382783)
        ref[1,2,2,0]=(0.544124184972)/(o+1.)+(-0.386469746437)
        ref[1,2,2,1]=(-0.057489821699)/(o+1.)+(-0.336358968492)
        ref[1,3,0,0]=(0.583710371001)/(o+1.)+(0.0931360914458)
        ref[1,3,0,1]=(-1.07594801739)/(o+1.)+(-0.107450748211)
        ref[1,3,1,0]=(1.10876761199)/(o+1.)+(-0.407722186665)
        ref[1,3,1,1]=(-0.741327216511)/(o+1.)+(0.0671296549307)
        ref[1,3,2,0]=(1.67985329366)/(o+1.)+(0.317880862337)
        ref[1,3,2,1]=(0.897872829808)/(o+1.)+(-0.302740205287)
        ref[1,4,0,0]=(-0.504220920447)/(o+1.)+(0.667109914405)
        ref[1,4,0,1]=(1.46619152344)/(o+1.)+(-0.367579394288)
        ref[1,4,1,0]=(1.15808904963)/(o+1.)+(-0.203133486458)
        ref[1,4,1,1]=(-0.843480157561)/(o+1.)+(0.183269867953)
        ref[1,4,2,0]=(0.302449071217)/(o+1.)+(-0.446266969414)
        ref[1,4,2,1]=(-0.24100562739)/(o+1.)+(0.399818411127)
        ref[2,0,0,0]=(-0.558932204389)/(o+1.)+(-0.214480679421)
        ref[2,0,0,1]=(0.832911407707)/(o+1.)+(0.285390197627)
        ref[2,0,1,0]=(1.46815648286)/(o+1.)+(-0.417850981819)
        ref[2,0,1,1]=(-0.201984078785)/(o+1.)+(0.280505538406)
        ref[2,0,2,0]=(-0.871211977955)/(o+1.)+(-0.108801848522)
        ref[2,0,2,1]=(-0.954427998409)/(o+1.)+(0.323358835806)
        ref[2,1,0,0]=(0.0279652942615)/(o+1.)+(-0.0930221592407)
        ref[2,1,0,1]=(0.969000196983)/(o+1.)+(-0.0461807845941)
        ref[2,1,1,0]=(-1.27107397819)/(o+1.)+(0.385977521287)
        ref[2,1,1,1]=(0.136609365964)/(o+1.)+(-0.0281963827764)
        ref[2,1,2,0]=(0.847085423093)/(o+1.)+(-0.379519695616)
        ref[2,1,2,1]=(-0.257183508208)/(o+1.)+(-0.371999262818)
        ref[2,2,0,0]=(0.451154945268)/(o+1.)+(-0.165226857878)
        ref[2,2,0,1]=(0.477943949158)/(o+1.)+(-0.642685950772)
        ref[2,2,1,0]=(0.657866460103)/(o+1.)+(-0.635972168839)
        ref[2,2,1,1]=(0.796454558882)/(o+1.)+(0.19698553472)
        ref[2,2,2,0]=(-1.47282324613)/(o+1.)+(0.00526287876389)
        ref[2,2,2,1]=(-0.971294249509)/(o+1.)+(-0.546324797713)
        ref[2,3,0,0]=(-1.08251786594)/(o+1.)+(0.0104702426452)
        ref[2,3,0,1]=(0.763229320287)/(o+1.)+(0.326514286241)
        ref[2,3,1,0]=(-0.564433556829)/(o+1.)+(-0.741257515163)
        ref[2,3,1,1]=(0.00440623628526)/(o+1.)+(0.395445803971)
        ref[2,3,2,0]=(1.13383212211)/(o+1.)+(0.0094516201736)
        ref[2,3,2,1]=(0.000662210559809)/(o+1.)+(-0.0729525884203)
        ref[2,4,0,0]=(0.533197232076)/(o+1.)+(0.400813242164)
        ref[2,4,0,1]=(1.05263232189)/(o+1.)+(0.706974809826)
        ref[2,4,1,0]=(0.580407716589)/(o+1.)+(0.138189479217)
        ref[2,4,1,1]=(-0.44538816431)/(o+1.)+(-0.251644455066)
        ref[2,4,2,0]=(0.0465222158616)/(o+1.)+(0.85626655775)
        ref[2,4,2,1]=(0.319198566788)/(o+1.)+(0.882551371452)
        ref[3,0,0,0]=(-0.561665866984)/(o+1.)+(-0.623682533758)
        ref[3,0,0,1]=(-0.544665480685)/(o+1.)+(-0.326701245334)
        ref[3,0,1,0]=(-1.60194246541)/(o+1.)+(-0.452248099853)
        ref[3,0,1,1]=(-0.907946092739)/(o+1.)+(0.519323659165)
        ref[3,0,2,0]=(-0.454139734776)/(o+1.)+(0.0447289827682)
        ref[3,0,2,1]=(1.01507581477)/(o+1.)+(0.317404572995)
        ref[3,1,0,0]=(0.0211828103496)/(o+1.)+(0.624371810253)
        ref[3,1,0,1]=(-0.658289545845)/(o+1.)+(-0.513831705447)
        ref[3,1,1,0]=(1.06011975683)/(o+1.)+(0.721388170095)
        ref[3,1,1,1]=(-0.14633541181)/(o+1.)+(0.618616305516)
        ref[3,1,2,0]=(0.74864463613)/(o+1.)+(-0.511570767053)
        ref[3,1,2,1]=(0.488722756727)/(o+1.)+(0.0387361369774)
        ref[3,2,0,0]=(-0.180332702395)/(o+1.)+(-0.0815892290549)
        ref[3,2,0,1]=(0.0811207173227)/(o+1.)+(0.392285456186)
        ref[3,2,1,0]=(1.02225125645)/(o+1.)+(0.480941670608)
        ref[3,2,1,1]=(1.21612405627)/(o+1.)+(0.45160613516)
        ref[3,2,2,0]=(-0.355793756138)/(o+1.)+(0.669992488205)
        ref[3,2,2,1]=(0.776200363894)/(o+1.)+(0.535930158963)
        ref[3,3,0,0]=(-0.692703834029)/(o+1.)+(-0.00615860690903)
        ref[3,3,0,1]=(0.88835171354)/(o+1.)+(0.00763271775166)
        ref[3,3,1,0]=(-0.0987823234827)/(o+1.)+(0.628186276549)
        ref[3,3,1,1]=(-0.223347424968)/(o+1.)+(-0.00346757455392)
        ref[3,3,2,0]=(1.22648639393)/(o+1.)+(0.152216781743)
        ref[3,3,2,1]=(0.272100360294)/(o+1.)+(0.111532468623)
        ref[3,4,0,0]=(1.45235034011)/(o+1.)+(-0.0206817576441)
        ref[3,4,0,1]=(1.49583359131)/(o+1.)+(-0.574565590956)
        ref[3,4,1,0]=(0.362882204744)/(o+1.)+(0.581681025235)
        ref[3,4,1,1]=(1.31795519959)/(o+1.)+(-0.0659375558974)
        ref[3,4,2,0]=(0.556130663233)/(o+1.)+(-0.078813607523)
        ref[3,4,2,1]=(1.56413690967)/(o+1.)+(-0.42520624231)
      else:
        arg[0,0,0,0]=(0.956854727968)*x[0]**o+(-0.43263548815)*x[0]+(-0.1485543705)*x[1]**o+(-0.411060684134)*x[1]+(-0.306829513954)*x[2]**o+(0.89077924839)*x[2]
        arg[0,0,0,1]=(0.125905642347)*x[0]**o+(-0.645491335766)*x[0]+(-0.237571989123)*x[1]**o+(-0.239301942511)*x[1]+(-0.384215893518)*x[2]**o+(-0.371218179274)*x[2]
        arg[0,0,1,0]=(0.0204273419969)*x[0]**o+(0.354631489501)*x[0]+(0.235086936835)*x[1]**o+(0.647551574475)*x[1]+(-0.729066874572)*x[2]**o+(0.177733791046)*x[2]
        arg[0,0,1,1]=(-0.0938492166558)*x[0]**o+(-0.484339074939)*x[0]+(0.619333193491)*x[1]**o+(0.809157671781)*x[1]+(-0.28367956249)*x[2]**o+(-0.1758249267)*x[2]
        arg[0,0,2,0]=(-0.334944259519)*x[0]**o+(-0.126145567301)*x[0]+(-0.646972616146)*x[1]**o+(0.20287553181)*x[1]+(-0.383816067326)*x[2]**o+(-0.0265591031498)*x[2]
        arg[0,0,2,1]=(0.103308564583)*x[0]**o+(0.016108181951)*x[0]+(-0.367450237136)*x[1]**o+(0.657618788015)*x[1]+(-0.8327532511)*x[2]**o+(-0.267286876849)*x[2]
        arg[0,1,0,0]=(0.688995722152)*x[0]**o+(-0.982012541087)*x[0]+(-0.232339851099)*x[1]**o+(-0.097863461266)*x[1]+(0.25916639806)*x[2]**o+(0.0538637018789)*x[2]
        arg[0,1,0,1]=(0.446698337789)*x[0]**o+(-0.309434573629)*x[0]+(0.617004214016)*x[1]**o+(0.747666594435)*x[1]+(-0.869684246717)*x[2]**o+(0.75016186442)*x[2]
        arg[0,1,1,0]=(-0.559974353252)*x[0]**o+(0.966642794905)*x[0]+(-0.228473000256)*x[1]**o+(-0.435093965632)*x[1]+(0.426986383284)*x[2]**o+(0.131300817321)*x[2]
        arg[0,1,1,1]=(-0.696027598385)*x[0]**o+(-0.804778535506)*x[0]+(0.930171049019)*x[1]**o+(-0.900826020873)*x[1]+(0.844689166523)*x[2]**o+(0.0349718139676)*x[2]
        arg[0,1,2,0]=(0.359170897328)*x[0]**o+(0.992944582308)*x[0]+(0.579162388074)*x[1]**o+(0.743410851741)*x[1]+(-0.204325306226)*x[2]**o+(-0.992005484878)*x[2]
        arg[0,1,2,1]=(-0.883554988936)*x[0]**o+(0.486947073334)*x[0]+(-0.78078860326)*x[1]**o+(-0.134196299284)*x[1]+(-0.980852067031)*x[2]**o+(0.453023187355)*x[2]
        arg[0,2,0,0]=(0.804524524371)*x[0]**o+(0.397924403155)*x[0]+(0.712297039229)*x[1]**o+(0.713485622558)*x[1]+(-0.966783610543)*x[2]**o+(0.497869404576)*x[2]
        arg[0,2,0,1]=(-0.564706930001)*x[0]**o+(-0.482026983244)*x[0]+(0.284926154564)*x[1]**o+(0.844423322149)*x[1]+(0.444203847061)*x[2]**o+(0.333244188639)*x[2]
        arg[0,2,1,0]=(-0.630410083438)*x[0]**o+(-0.831218126629)*x[0]+(0.400237927956)*x[1]**o+(-0.70518378137)*x[1]+(-0.803754805917)*x[2]**o+(0.60059455602)*x[2]
        arg[0,2,1,1]=(0.198272547853)*x[0]**o+(0.0840356091496)*x[0]+(0.114385146155)*x[1]**o+(0.731057038288)*x[1]+(-0.976056408523)*x[2]**o+(-0.132565954304)*x[2]
        arg[0,2,2,0]=(0.983300141657)*x[0]**o+(-0.733168195162)*x[0]+(-0.948802450335)*x[1]**o+(-0.756616078178)*x[1]+(0.450302374372)*x[2]**o+(0.802115535717)*x[2]
        arg[0,2,2,1]=(-0.680853503774)*x[0]**o+(-0.910402862928)*x[0]+(0.583775613868)*x[1]**o+(0.741516645611)*x[1]+(-0.550498467533)*x[2]**o+(-0.610648481959)*x[2]
        arg[0,3,0,0]=(0.981909313325)*x[0]**o+(0.148145112359)*x[0]+(-0.989721117518)*x[1]**o+(0.0246558730659)*x[1]+(0.721439979982)*x[2]**o+(0.896052837889)*x[2]
        arg[0,3,0,1]=(0.109831691542)*x[0]**o+(-0.349555366958)*x[0]+(-0.54822642636)*x[1]**o+(0.121919022172)*x[1]+(-0.580228247481)*x[2]**o+(0.564276455713)*x[2]
        arg[0,3,1,0]=(0.106091093353)*x[0]**o+(0.410111742327)*x[0]+(-0.0669026478824)*x[1]**o+(-0.662684907574)*x[1]+(-0.351815520918)*x[2]**o+(-0.324802958038)*x[2]
        arg[0,3,1,1]=(0.857685803258)*x[0]**o+(-0.782627954043)*x[0]+(-0.555036389439)*x[1]**o+(-0.0126716685608)*x[1]+(0.260461379777)*x[2]**o+(0.117590260591)*x[2]
        arg[0,3,2,0]=(-0.987873674604)*x[0]**o+(0.0779662295512)*x[0]+(-0.928229202803)*x[1]**o+(0.508015712707)*x[1]+(0.248638760509)*x[2]**o+(-0.712158091156)*x[2]
        arg[0,3,2,1]=(-0.977973356138)*x[0]**o+(0.270015563339)*x[0]+(0.811157094127)*x[1]**o+(-0.756460384936)*x[1]+(-0.735557557504)*x[2]**o+(0.780333584929)*x[2]
        arg[0,4,0,0]=(-0.762444411646)*x[0]**o+(0.386894014691)*x[0]+(0.763898738855)*x[1]**o+(-0.422896581278)*x[1]+(0.428591445563)*x[2]**o+(-0.617626926396)*x[2]
        arg[0,4,0,1]=(-0.269524167286)*x[0]**o+(0.996895823788)*x[0]+(0.771396605337)*x[1]**o+(-0.993794940262)*x[1]+(-0.534803587481)*x[2]**o+(-0.928627946122)*x[2]
        arg[0,4,1,0]=(0.978497459584)*x[0]**o+(-0.343496507186)*x[0]+(-0.250581087884)*x[1]**o+(-0.855743445852)*x[1]+(-0.127512246124)*x[2]**o+(-0.133936486447)*x[2]
        arg[0,4,1,1]=(-0.844450818471)*x[0]**o+(0.263329730105)*x[0]+(0.411212783735)*x[1]**o+(-0.816379748952)*x[1]+(-0.443806901857)*x[2]**o+(0.1262969097)*x[2]
        arg[0,4,2,0]=(-0.326938381124)*x[0]**o+(-0.707054728551)*x[0]+(-0.725513874889)*x[1]**o+(0.655254977879)*x[1]+(0.772734170665)*x[2]**o+(-0.963856465024)*x[2]
        arg[0,4,2,1]=(0.896415185133)*x[0]**o+(-0.722431478737)*x[0]+(0.942058336843)*x[1]**o+(0.838230971246)*x[1]+(-0.279821311439)*x[2]**o+(-0.330130961206)*x[2]
        arg[1,0,0,0]=(-0.466537465363)*x[0]**o+(0.86493624537)*x[0]+(-0.943049537777)*x[1]**o+(0.658727518981)*x[1]+(-0.130355508544)*x[2]**o+(-0.158245981876)*x[2]
        arg[1,0,0,1]=(0.247046604605)*x[0]**o+(-0.287683750248)*x[0]+(0.266630203142)*x[1]**o+(0.372409648579)*x[1]+(-0.239411725166)*x[2]**o+(0.537339451105)*x[2]
        arg[1,0,1,0]=(-0.232312387253)*x[0]**o+(-0.0493954012747)*x[0]+(-0.461085479525)*x[1]**o+(-0.0744729096417)*x[1]+(-0.795163064273)*x[2]**o+(0.613075576806)*x[2]
        arg[1,0,1,1]=(-0.68570821748)*x[0]**o+(-0.492507082229)*x[0]+(-0.0673636097828)*x[1]**o+(0.569119992806)*x[1]+(-0.227653595613)*x[2]**o+(0.0337433636837)*x[2]
        arg[1,0,2,0]=(-0.124371489405)*x[0]**o+(-0.726635218442)*x[0]+(-0.835418167414)*x[1]**o+(-0.00414597707753)*x[1]+(0.8365917104)*x[2]**o+(-0.507314639615)*x[2]
        arg[1,0,2,1]=(-0.973228057052)*x[0]**o+(0.123228228923)*x[0]+(0.545712842666)*x[1]**o+(-0.451331566047)*x[1]+(0.299501907067)*x[2]**o+(-0.24898389883)*x[2]
        arg[1,1,0,0]=(-0.940690906836)*x[0]**o+(-0.357571197214)*x[0]+(0.98643983012)*x[1]**o+(-0.82113140872)*x[1]+(0.460009285424)*x[2]**o+(-0.0298714172583)*x[2]
        arg[1,1,0,1]=(-0.881543245921)*x[0]**o+(-0.640741814727)*x[0]+(-0.749244677096)*x[1]**o+(0.47161101909)*x[1]+(0.712893375828)*x[2]**o+(0.600151133959)*x[2]
        arg[1,1,1,0]=(-0.217214692405)*x[0]**o+(-0.504211076658)*x[0]+(0.857127736087)*x[1]**o+(0.0439761116172)*x[1]+(0.956380848189)*x[2]**o+(0.698482155013)*x[2]
        arg[1,1,1,1]=(-0.288969728734)*x[0]**o+(-0.63960886193)*x[0]+(-0.741293494258)*x[1]**o+(-0.125034323955)*x[1]+(-0.608855286755)*x[2]**o+(0.462427737944)*x[2]
        arg[1,1,2,0]=(0.126242043984)*x[0]**o+(0.811846484451)*x[0]+(-0.522920862127)*x[1]**o+(0.678142081935)*x[1]+(0.0392586436898)*x[2]**o+(0.403908304728)*x[2]
        arg[1,1,2,1]=(-0.290103861843)*x[0]**o+(0.336130241928)*x[0]+(-0.872641307678)*x[1]**o+(-0.372020796526)*x[1]+(0.172572077906)*x[2]**o+(-0.789431556966)*x[2]
        arg[1,2,0,0]=(-0.158114079913)*x[0]**o+(0.954919898537)*x[0]+(-0.711834968234)*x[1]**o+(-0.390487270434)*x[1]+(-0.086581514295)*x[2]**o+(0.442047084988)*x[2]
        arg[1,2,0,1]=(-0.914034594356)*x[0]**o+(0.521137372279)*x[0]+(0.12454416569)*x[1]**o+(0.627815188423)*x[1]+(0.770244577882)*x[2]**o+(-0.846691748091)*x[2]
        arg[1,2,1,0]=(0.769631824465)*x[0]**o+(0.988840796717)*x[0]+(-0.401836868027)*x[1]**o+(0.401846957832)*x[1]+(-0.877826997285)*x[2]**o+(0.857725824229)*x[2]
        arg[1,2,1,1]=(0.173539365276)*x[0]**o+(0.373206670706)*x[0]+(-0.597288730366)*x[1]**o+(-0.770220794555)*x[1]+(-0.586510454467)*x[2]**o+(-0.569370842441)*x[2]
        arg[1,2,2,0]=(0.673765155818)*x[0]**o+(-0.642914516709)*x[0]+(0.794438851788)*x[1]**o+(-0.17557754186)*x[1]+(0.997565143863)*x[2]**o+(0.722460144483)*x[2]
        arg[1,2,2,1]=(-0.878452482718)*x[0]**o+(-0.846862156166)*x[0]+(0.462351027504)*x[1]**o+(0.267435606154)*x[1]+(-0.452135086983)*x[2]**o+(0.701577407252)*x[2]
        arg[1,3,0,0]=(-0.0557389135964)*x[0]**o+(-0.781831460089)*x[0]+(0.214846092931)*x[1]**o+(0.597436260353)*x[1]+(0.713968428327)*x[2]**o+(-0.456335676572)*x[2]
        arg[1,3,0,1]=(0.288266531585)*x[0]**o+(-0.259529556521)*x[0]+(0.0286107975184)*x[1]**o+(0.465095455484)*x[1]+(-0.747648909025)*x[2]**o+(0.682866470382)*x[2]
        arg[1,3,1,0]=(-0.720339104668)*x[0]**o+(0.961100526257)*x[0]+(0.0814885381756)*x[1]**o+(-0.448775223649)*x[1]+(0.250972143198)*x[2]**o+(-0.240231439947)*x[2]
        arg[1,3,1,1]=(-0.459305786748)*x[0]**o+(0.272222143865)*x[0]+(-0.499164099401)*x[1]**o+(-0.225162547271)*x[1]+(-0.92261124167)*x[2]**o+(-0.436659790026)*x[2]
        arg[1,3,2,0]=(0.890752966329)*x[0]**o+(-0.775207922011)*x[0]+(0.715960284721)*x[1]**o+(0.843422841789)*x[1]+(0.513788850466)*x[2]**o+(-0.325733710281)*x[2]
        arg[1,3,2,1]=(0.958726172433)*x[0]**o+(-0.743672393249)*x[0]+(-0.523783133945)*x[1]**o+(0.133861976051)*x[1]+(-0.109230940116)*x[2]**o+(0.819716327751)*x[2]
        arg[1,4,0,0]=(0.54439729551)*x[0]**o+(0.0763503792087)*x[0]+(0.0634218657498)*x[1]**o+(-0.405471845044)*x[1]+(0.316327091756)*x[2]**o+(-0.539493863939)*x[2]
        arg[1,4,0,1]=(0.527330373182)*x[0]**o+(0.77531447123)*x[0]+(-0.481869315442)*x[1]**o+(0.556927692362)*x[1]+(-0.267215098309)*x[2]**o+(0.958919406465)*x[2]
        arg[1,4,1,0]=(0.540950372012)*x[0]**o+(-0.95002626611)*x[0]+(-0.523544509893)*x[1]**o+(-0.583056635057)*x[1]+(-0.00288409928193)*x[2]**o+(0.671385098427)*x[2]
        arg[1,4,1,1]=(-0.39802076092)*x[0]**o+(-0.583264022146)*x[0]+(-0.656261495961)*x[1]**o+(-0.527677078418)*x[1]+(-0.2632498932)*x[2]**o+(-0.537632951507)*x[2]
        arg[1,4,2,0]=(-0.77176416566)*x[0]**o+(0.812696440735)*x[0]+(0.0925789242005)*x[1]**o+(0.299698967601)*x[1]+(-0.0185377066587)*x[2]**o+(-0.883247318682)*x[2]
        arg[1,4,2,1]=(0.360949407452)*x[0]**o+(-0.463659754812)*x[0]+(0.799066387132)*x[1]**o+(0.0318151346916)*x[1]+(-0.380675976437)*x[2]**o+(-0.299341342112)*x[2]
        arg[2,0,0,0]=(-0.985332287141)*x[0]**o+(0.33398206098)*x[0]+(0.39204502373)*x[1]**o+(-0.586826360517)*x[1]+(0.868240161369)*x[2]**o+(-0.257897525907)*x[2]
        arg[2,0,0,1]=(-0.801369266967)*x[0]**o+(0.333496542809)*x[0]+(0.459444994717)*x[1]**o+(-0.950297568688)*x[1]+(-0.760722281765)*x[2]**o+(0.0933850221377)*x[2]
        arg[2,0,1,0]=(0.53372811812)*x[0]**o+(0.999551098511)*x[0]+(-0.761650890163)*x[1]**o+(-0.80456940786)*x[1]+(-0.452089806623)*x[2]**o+(-0.40841169158)*x[2]
        arg[2,0,1,1]=(-0.782294374468)*x[0]**o+(0.614736817827)*x[0]+(0.262995910218)*x[1]**o+(0.067349886804)*x[1]+(-0.167494656909)*x[2]**o+(-0.730462378769)*x[2]
        arg[2,0,2,0]=(0.261001857862)*x[0]**o+(0.959318765248)*x[0]+(-0.177224016298)*x[1]**o+(-0.979637664014)*x[1]+(0.858937930292)*x[2]**o+(-0.627116377798)*x[2]
        arg[2,0,2,1]=(-0.238083578675)*x[0]**o+(0.586367991705)*x[0]+(0.747378561289)*x[1]**o+(-0.540178926415)*x[1]+(-0.519781240185)*x[2]**o+(-0.42715014981)*x[2]
        arg[2,1,0,0]=(-0.21022574437)*x[0]**o+(-0.292359099348)*x[0]+(0.449614855785)*x[1]**o+(-0.416904619153)*x[1]+(0.507537288404)*x[2]**o+(0.943584716996)*x[2]
        arg[2,1,0,1]=(-0.410591525695)*x[0]**o+(-0.258796394127)*x[0]+(-0.474714731399)*x[1]**o+(-0.518288451462)*x[1]+(-0.15640903664)*x[2]**o+(0.97242574307)*x[2]
        arg[2,1,1,0]=(-0.930865834892)*x[0]**o+(0.961887630616)*x[0]+(0.688706862295)*x[1]**o+(-0.375163634404)*x[1]+(0.378954001649)*x[2]**o+(-0.747140890466)*x[2]
        arg[2,1,1,1]=(0.0454701384411)*x[0]**o+(-0.37651801332)*x[0]+(0.0808750051583)*x[1]**o+(-0.530355850075)*x[1]+(0.769675417234)*x[2]**o+(-0.767046792016)*x[2]
        arg[2,1,2,0]=(0.179573418696)*x[0]**o+(-0.0547918709025)*x[0]+(0.739792606043)*x[1]**o+(0.849619562235)*x[1]+(-0.92840796838)*x[2]**o+(-0.302941296337)*x[2]
        arg[2,1,2,1]=(0.160657807137)*x[0]**o+(0.551943403654)*x[0]+(0.649164334204)*x[1]**o+(0.298882266215)*x[1]+(0.33872360276)*x[2]**o+(-0.558910868754)*x[2]
        arg[2,2,0,0]=(-0.704330511528)*x[0]**o+(-0.326943167116)*x[0]+(0.626863819473)*x[1]**o+(-0.107147718189)*x[1]+(0.00668670393016)*x[2]**o+(0.572223662188)*x[2]
        arg[2,2,0,1]=(0.879695137799)*x[0]**o+(-0.313488700497)*x[0]+(-0.113134766565)*x[1]**o+(-0.131007759381)*x[1]+(0.55460969593)*x[2]**o+(0.898706374575)*x[2]
        arg[2,2,1,0]=(0.329932124981)*x[0]**o+(-0.624965470976)*x[0]+(0.82906774364)*x[1]**o+(-0.681403786977)*x[1]+(-0.447119491747)*x[2]**o+(0.143763601614)*x[2]
        arg[2,2,1,1]=(0.0367746207528)*x[0]**o+(0.47553259605)*x[0]+(-0.517723045973)*x[1]**o+(0.263932783195)*x[1]+(-0.131032589284)*x[2]**o+(-0.132729291841)*x[2]
        arg[2,2,2,0]=(-0.108349266564)*x[0]**o+(-0.684748849483)*x[0]+(0.203271582231)*x[1]**o+(-0.860842085896)*x[1]+(0.469684577785)*x[2]**o+(0.195736308182)*x[2]
        arg[2,2,2,1]=(0.915666799349)*x[0]**o+(0.383338953756)*x[0]+(-0.554011249579)*x[1]**o+(0.232285573367)*x[1]+(0.39288151674)*x[2]**o+(0.692592233016)*x[2]
        arg[2,3,0,0]=(0.81179207663)*x[0]**o+(0.786809236572)*x[0]+(-0.814758445611)*x[1]**o+(-0.78461653396)*x[1]+(0.171435802778)*x[2]**o+(0.660186909193)*x[2]
        arg[2,3,0,1]=(0.752694523669)*x[0]**o+(0.301473785296)*x[0]+(0.610379955695)*x[1]**o+(0.573729399495)*x[1]+(0.225983751859)*x[2]**o+(-0.349530143446)*x[2]
        arg[2,3,1,0]=(-0.647825490424)*x[0]**o+(-0.232573858642)*x[0]+(0.773519085771)*x[1]**o+(-0.224497979868)*x[1]+(-0.681726885952)*x[2]**o+(-0.217968759388)*x[2]
        arg[2,3,1,1]=(0.391888417642)*x[0]**o+(-0.240556139087)*x[0]+(-0.852397797723)*x[1]**o+(0.318342758913)*x[1]+(-0.773804983598)*x[2]**o+(-0.957590491908)*x[2]
        arg[2,3,2,0]=(0.7126582007)*x[0]**o+(0.806022164546)*x[0]+(0.232348951782)*x[1]**o+(-0.985774366344)*x[1]+(-0.365161907541)*x[2]**o+(0.0218977623299)*x[2]
        arg[2,3,2,1]=(-0.775068843872)*x[0]**o+(-0.595291330879)*x[0]+(0.681122481841)*x[1]**o+(0.152514252732)*x[1]+(0.510976191808)*x[2]**o+(-0.69539067488)*x[2]
        arg[2,4,0,0]=(-0.362761749405)*x[0]**o+(0.494347958562)*x[0]+(0.215244694522)*x[1]**o+(-0.527767969217)*x[1]+(-0.115976071059)*x[2]**o+(-0.998233415841)*x[2]
        arg[2,4,0,1]=(0.620291015098)*x[0]**o+(-0.222644358028)*x[0]+(0.0377714785847)*x[1]**o+(0.663978623913)*x[1]+(0.996481523389)*x[2]**o+(-0.689519418095)*x[2]
        arg[2,4,1,0]=(-0.0374746407394)*x[0]**o+(0.3186355021)*x[0]+(-0.704305640533)*x[1]**o+(-0.632264150766)*x[1]+(-0.166691199609)*x[2]**o+(-0.794128441334)*x[2]
        arg[2,4,1,1]=(0.117806334919)*x[0]**o+(-0.717334572919)*x[0]+(0.665341054571)*x[1]**o+(0.354208273372)*x[1]+(0.349883631586)*x[2]**o+(-0.260626035314)*x[2]
        arg[2,4,2,0]=(0.160337827685)*x[0]**o+(-0.92734430881)*x[0]+(0.872112181363)*x[1]**o+(-0.796372970466)*x[1]+(-0.776882702685)*x[2]**o+(0.246816968474)*x[2]
        arg[2,4,2,1]=(-0.317926812571)*x[0]**o+(-0.795779593996)*x[0]+(0.785287724793)*x[1]**o+(0.839033515634)*x[1]+(0.433187943714)*x[2]**o+(0.308175885606)*x[2]
        arg[3,0,0,0]=(-0.555193292291)*x[0]**o+(-0.0280452948283)*x[0]+(0.988455127227)*x[1]**o+(-0.637707060271)*x[1]+(-0.95062571916)*x[2]**o+(-0.563121211553)*x[2]
        arg[3,0,0,1]=(-0.169558534946)*x[0]**o+(0.886096637835)*x[0]+(-0.666307491933)*x[1]**o+(0.650321801985)*x[1]+(-0.268934932123)*x[2]**o+(-0.406617199906)*x[2]
        arg[3,0,1,0]=(0.497351781449)*x[0]**o+(0.768617651685)*x[0]+(0.200962956226)*x[1]**o+(0.694236078954)*x[1]+(0.0323702864321)*x[2]**o+(-0.666046265832)*x[2]
        arg[3,0,1,1]=(0.425151807387)*x[0]**o+(-0.886432438604)*x[0]+(-0.963257203569)*x[1]**o+(-0.698524427437)*x[1]+(0.0165810713355)*x[2]**o+(-0.142193988951)*x[2]
        arg[3,0,2,0]=(0.965746112788)*x[0]**o+(-0.333068915886)*x[0]+(0.331031372002)*x[1]**o+(-0.384004796757)*x[1]+(0.0144121092048)*x[2]**o+(-0.352607264387)*x[2]
        arg[3,0,2,1]=(-0.391557775789)*x[0]**o+(0.127020330586)*x[0]+(0.758832450081)*x[1]**o+(0.49669414469)*x[1]+(0.358403498634)*x[2]**o+(0.184963946351)*x[2]
        arg[3,1,0,0]=(0.978053897464)*x[0]**o+(-0.211787313494)*x[0]+(0.927706482976)*x[1]**o+(-0.451292956408)*x[1]+(0.592824645174)*x[2]**o+(0.896240907546)*x[2]
        arg[3,1,0,1]=(0.0780568875295)*x[0]**o+(0.130098698523)*x[0]+(-0.281871571918)*x[1]**o+(-0.596362849045)*x[1]+(-0.436455789199)*x[2]**o+(-0.657145053021)*x[2]
        arg[3,1,1,0]=(-0.278111817329)*x[0]**o+(-0.627384158565)*x[0]+(-0.995802044048)*x[1]**o+(0.42485189159)*x[1]+(0.483473275475)*x[2]**o+(-0.177928974458)*x[2]
        arg[3,1,1,1]=(0.127721047557)*x[0]**o+(0.93147711978)*x[0]+(-0.350992052283)*x[1]**o+(0.0993773637672)*x[1]+(0.29111744653)*x[2]**o+(-0.821710278883)*x[2]
        arg[3,1,2,0]=(-0.697941918276)*x[0]**o+(-0.585872232852)*x[0]+(0.697967779925)*x[1]**o+(-0.811186381728)*x[1]+(0.708726108998)*x[2]**o+(0.289602027081)*x[2]
        arg[3,1,2,1]=(-0.461424409147)*x[0]**o+(0.335089386385)*x[0]+(-0.964838280406)*x[1]**o+(0.828627561583)*x[1]+(-0.861416281013)*x[2]**o+(0.898354675269)*x[2]
        arg[3,2,0,0]=(0.752543705358)*x[0]**o+(0.164487088496)*x[0]+(0.756164002406)*x[1]**o+(-0.740019514032)*x[1]+(0.107245489678)*x[2]**o+(0.677707700079)*x[2]
        arg[3,2,0,1]=(0.763792557134)*x[0]**o+(0.725507550672)*x[0]+(0.171323949629)*x[1]**o+(-0.287172075258)*x[1]+(-0.94865226715)*x[2]**o+(-0.163093465336)*x[2]
        arg[3,2,1,0]=(-0.327892721048)*x[0]**o+(0.334145897353)*x[0]+(-0.936827293436)*x[1]**o+(0.0822687872593)*x[1]+(0.259794829604)*x[2]**o+(0.376908280759)*x[2]
        arg[3,2,1,1]=(0.595012545955)*x[0]**o+(0.617490014693)*x[0]+(-0.658036388462)*x[1]**o+(-0.921055393527)*x[1]+(-0.257878085344)*x[2]**o+(-0.623113089161)*x[2]
        arg[3,2,2,0]=(0.961175705937)*x[0]**o+(0.656031660842)*x[0]+(0.964101580317)*x[1]**o+(0.824033157042)*x[1]+(0.510244886658)*x[2]**o+(-0.236501162729)*x[2]
        arg[3,2,2,1]=(-0.128405588923)*x[0]**o+(-0.0327141502894)*x[0]+(0.909941485131)*x[1]**o+(0.79947258121)*x[1]+(0.635883050534)*x[2]**o+(0.841997575796)*x[2]
        arg[3,3,0,0]=(-0.906234907643)*x[0]**o+(-0.316076976009)*x[0]+(0.863208664533)*x[1]**o+(-0.968640170523)*x[1]+(0.791637381686)*x[2]**o+(-0.543353462098)*x[2]
        arg[3,3,0,1]=(-0.586709036832)*x[0]**o+(-0.299493216592)*x[0]+(-0.173966725899)*x[1]**o+(0.0971838711677)*x[1]+(0.385662922244)*x[2]**o+(0.288932625368)*x[2]
        arg[3,3,1,0]=(-0.33865322795)*x[0]**o+(0.442176445777)*x[0]+(0.706183231282)*x[1]**o+(0.62528018966)*x[1]+(-0.377746006979)*x[2]**o+(-0.261455124139)*x[2]
        arg[3,3,1,1]=(0.834855553776)*x[0]**o+(0.755005824764)*x[0]+(0.446589332183)*x[1]**o+(0.41883521757)*x[1]+(0.820177207209)*x[2]**o+(0.603811552083)*x[2]
        arg[3,3,2,0]=(-0.624282266193)*x[0]**o+(0.28109086861)*x[0]+(0.44286011655)*x[1]**o+(-0.117876187343)*x[1]+(-0.267660131821)*x[2]**o+(0.880359406159)*x[2]
        arg[3,3,2,1]=(-0.0722687033317)*x[0]**o+(-0.7013147214)*x[0]+(0.00614303249005)*x[1]**o+(-0.658138946692)*x[1]+(0.0557277088337)*x[2]**o+(0.614362766103)*x[2]
        arg[3,4,0,0]=(0.935244440586)*x[0]**o+(-0.0699305155896)*x[0]+(-0.849303362918)*x[1]**o+(0.728062537708)*x[1]+(0.404621498028)*x[2]**o+(0.611486631437)*x[2]
        arg[3,4,0,1]=(0.168117782017)*x[0]**o+(-0.0454396423181)*x[0]+(0.46021270423)*x[1]**o+(0.597343569492)*x[1]+(0.561215321486)*x[2]**o+(0.715771227777)*x[2]
        arg[3,4,1,0]=(-0.502675682461)*x[0]**o+(0.543910335185)*x[0]+(0.721732643536)*x[1]**o+(-0.862126899566)*x[1]+(0.794704104653)*x[2]**o+(-0.284248276797)*x[2]
        arg[3,4,1,1]=(-0.580751817984)*x[0]**o+(0.482838505281)*x[0]+(-0.426066109367)*x[1]**o+(-0.456724382007)*x[1]+(-0.770655426938)*x[2]**o+(-0.226232224569)*x[2]
        arg[3,4,2,0]=(0.199832861603)*x[0]**o+(-0.586239366436)*x[0]+(0.730019631541)*x[1]**o+(0.185265476259)*x[1]+(0.514243351617)*x[2]**o+(-0.757742857839)*x[2]
        arg[3,4,2,1]=(-0.941843068195)*x[0]**o+(0.211732407161)*x[0]+(-0.99705602485)*x[1]**o+(-0.198037742645)*x[1]+(0.154058794178)*x[2]**o+(0.0675962671791)*x[2]
        ref[0,0,0,0]=(0.501470843513)/(o+1.)+(0.0235415380528)
        ref[0,0,0,1]=(-0.495882240293)/(o+1.)+(-0.628005728776)
        ref[0,0,1,0]=(-0.47355259574)/(o+1.)+(0.589958427511)
        ref[0,0,1,1]=(0.241804414345)/(o+1.)+(0.0744968350706)
        ref[0,0,2,0]=(-1.36573294299)/(o+1.)+(0.0250854306796)
        ref[0,0,2,1]=(-1.09689492365)/(o+1.)+(0.203220046558)
        ref[0,1,0,0]=(0.715822269113)/(o+1.)+(-0.513006150237)
        ref[0,1,0,1]=(0.194018305088)/(o+1.)+(0.594196942613)
        ref[0,1,1,0]=(-0.361460970224)/(o+1.)+(0.331424823297)
        ref[0,1,1,1]=(1.07883261716)/(o+1.)+(-0.835316371206)
        ref[0,1,2,0]=(0.734007979175)/(o+1.)+(0.372174974585)
        ref[0,1,2,1]=(-2.64519565923)/(o+1.)+(0.402886980703)
        ref[0,2,0,0]=(0.550037953057)/(o+1.)+(0.804639715144)
        ref[0,2,0,1]=(0.164423071623)/(o+1.)+(0.347820263772)
        ref[0,2,1,0]=(-1.0339269614)/(o+1.)+(-0.46790367599)
        ref[0,2,1,1]=(-0.663398714515)/(o+1.)+(0.341263346566)
        ref[0,2,2,0]=(0.484800065694)/(o+1.)+(-0.343834368812)
        ref[0,2,2,1]=(-0.647576357439)/(o+1.)+(-0.389767349638)
        ref[0,3,0,0]=(0.713628175789)/(o+1.)+(0.534426911657)
        ref[0,3,0,1]=(-1.0186229823)/(o+1.)+(0.168320055464)
        ref[0,3,1,0]=(-0.312627075447)/(o+1.)+(-0.288688061643)
        ref[0,3,1,1]=(0.563110793595)/(o+1.)+(-0.338854681006)
        ref[0,3,2,0]=(-1.6674641169)/(o+1.)+(-0.0630880744491)
        ref[0,3,2,1]=(-0.902373819515)/(o+1.)+(0.146944381666)
        ref[0,4,0,0]=(0.430045772773)/(o+1.)+(-0.326814746492)
        ref[0,4,0,1]=(-0.0329311494305)/(o+1.)+(-0.462763531298)
        ref[0,4,1,0]=(0.600404125575)/(o+1.)+(-0.666588219743)
        ref[0,4,1,1]=(-0.877044936593)/(o+1.)+(-0.213376554573)
        ref[0,4,2,0]=(-0.279718085349)/(o+1.)+(-0.507828107848)
        ref[0,4,2,1]=(1.55865221054)/(o+1.)+(-0.107165734349)
        ref[1,0,0,0]=(-1.53994251168)/(o+1.)+(0.682708891237)
        ref[1,0,0,1]=(0.27426508258)/(o+1.)+(0.311032674718)
        ref[1,0,1,0]=(-1.48856093105)/(o+1.)+(0.244603632945)
        ref[1,0,1,1]=(-0.980725422876)/(o+1.)+(0.0551781371305)
        ref[1,0,2,0]=(-0.123197946419)/(o+1.)+(-0.619047917567)
        ref[1,0,2,1]=(-0.128013307318)/(o+1.)+(-0.288543617977)
        ref[1,1,0,0]=(0.505758208708)/(o+1.)+(-0.604287011596)
        ref[1,1,0,1]=(-0.917894547188)/(o+1.)+(0.215510169161)
        ref[1,1,1,0]=(1.59629389187)/(o+1.)+(0.119123594986)
        ref[1,1,1,1]=(-1.63911850975)/(o+1.)+(-0.15110772397)
        ref[1,1,2,0]=(-0.357420174452)/(o+1.)+(0.946948435557)
        ref[1,1,2,1]=(-0.990173091615)/(o+1.)+(-0.412661055782)
        ref[1,2,0,0]=(-0.956530562442)/(o+1.)+(0.503239856545)
        ref[1,2,0,1]=(-0.0192458507848)/(o+1.)+(0.151130406306)
        ref[1,2,1,0]=(-0.510032040848)/(o+1.)+(1.12420678939)
        ref[1,2,1,1]=(-1.01025981956)/(o+1.)+(-0.483192483145)
        ref[1,2,2,0]=(2.46576915147)/(o+1.)+(-0.0480159570431)
        ref[1,2,2,1]=(-0.868236542197)/(o+1.)+(0.0610754286198)
        ref[1,3,0,0]=(0.873075607662)/(o+1.)+(-0.320365438154)
        ref[1,3,0,1]=(-0.430771579922)/(o+1.)+(0.444216184672)
        ref[1,3,1,0]=(-0.387878423294)/(o+1.)+(0.13604693133)
        ref[1,3,1,1]=(-1.88108112782)/(o+1.)+(-0.194800096716)
        ref[1,3,2,0]=(2.12050210152)/(o+1.)+(-0.128759395251)
        ref[1,3,2,1]=(0.325712098372)/(o+1.)+(0.104952955276)
        ref[1,4,0,0]=(0.924146253016)/(o+1.)+(-0.434307664887)
        ref[1,4,0,1]=(-0.221754040568)/(o+1.)+(1.14558078503)
        ref[1,4,1,0]=(0.0145217628367)/(o+1.)+(-0.43084890137)
        ref[1,4,1,1]=(-1.31753215008)/(o+1.)+(-0.824287026036)
        ref[1,4,2,0]=(-0.697722948118)/(o+1.)+(0.114574044827)
        ref[1,4,2,1]=(0.779339818147)/(o+1.)+(-0.365592981116)
        ref[2,0,0,0]=(0.274952897958)/(o+1.)+(-0.255370912722)
        ref[2,0,0,1]=(-1.10264655401)/(o+1.)+(-0.261708001871)
        ref[2,0,1,0]=(-0.680012578665)/(o+1.)+(-0.106715000464)
        ref[2,0,1,1]=(-0.686793121159)/(o+1.)+(-0.0241878370688)
        ref[2,0,2,0]=(0.942715771856)/(o+1.)+(-0.323717638282)
        ref[2,0,2,1]=(-0.0104862575713)/(o+1.)+(-0.19048054226)
        ref[2,1,0,0]=(0.746926399818)/(o+1.)+(0.117160499248)
        ref[2,1,0,1]=(-1.04171529373)/(o+1.)+(0.0976704487401)
        ref[2,1,1,0]=(0.136795029052)/(o+1.)+(-0.0802084471266)
        ref[2,1,1,1]=(0.896020560833)/(o+1.)+(-0.836960327706)
        ref[2,1,2,0]=(-0.00904194364113)/(o+1.)+(0.245943197498)
        ref[2,1,2,1]=(1.1485457441)/(o+1.)+(0.145957400558)
        ref[2,2,0,0]=(-0.0707799881243)/(o+1.)+(0.0690663884416)
        ref[2,2,0,1]=(1.32117006716)/(o+1.)+(0.227104957348)
        ref[2,2,1,0]=(0.711880376874)/(o+1.)+(-0.58130282817)
        ref[2,2,1,1]=(-0.611981014504)/(o+1.)+(0.303368043702)
        ref[2,2,2,0]=(0.564606893452)/(o+1.)+(-0.674927313599)
        ref[2,2,2,1]=(0.754537066511)/(o+1.)+(0.654108380069)
        ref[2,3,0,0]=(0.168469433796)/(o+1.)+(0.331189805903)
        ref[2,3,0,1]=(1.58905823122)/(o+1.)+(0.262836520673)
        ref[2,3,1,0]=(-0.556033290604)/(o+1.)+(-0.337520298949)
        ref[2,3,1,1]=(-1.23431436368)/(o+1.)+(-0.439901936041)
        ref[2,3,2,0]=(0.579845244941)/(o+1.)+(-0.0789272197338)
        ref[2,3,2,1]=(0.417029829778)/(o+1.)+(-0.569083876513)
        ref[2,4,0,0]=(-0.263493125943)/(o+1.)+(-0.515826713248)
        ref[2,4,0,1]=(1.65454401707)/(o+1.)+(-0.124092576105)
        ref[2,4,1,0]=(-0.908471480882)/(o+1.)+(-0.553878545)
        ref[2,4,1,1]=(1.13303102108)/(o+1.)+(-0.31187616743)
        ref[2,4,2,0]=(0.255567306363)/(o+1.)+(-0.738450155401)
        ref[2,4,2,1]=(0.900548855936)/(o+1.)+(0.175714903622)
        ref[3,0,0,0]=(-0.517363884224)/(o+1.)+(-0.614436783326)
        ref[3,0,0,1]=(-1.104800959)/(o+1.)+(0.564900619957)
        ref[3,0,1,0]=(0.730685024108)/(o+1.)+(0.398403732404)
        ref[3,0,1,1]=(-0.521524324846)/(o+1.)+(-0.863575427496)
        ref[3,0,2,0]=(1.31118959399)/(o+1.)+(-0.534840488515)
        ref[3,0,2,1]=(0.725678172925)/(o+1.)+(0.404339210814)
        ref[3,1,0,0]=(2.49858502561)/(o+1.)+(0.116580318822)
        ref[3,1,0,1]=(-0.640270473588)/(o+1.)+(-0.561704601771)
        ref[3,1,1,0]=(-0.790440585902)/(o+1.)+(-0.190230620716)
        ref[3,1,1,1]=(0.0678464418036)/(o+1.)+(0.104572102332)
        ref[3,1,2,0]=(0.708751970646)/(o+1.)+(-0.553728293749)
        ref[3,1,2,1]=(-2.28767897057)/(o+1.)+(1.03103581162)
        ref[3,2,0,0]=(1.61595319744)/(o+1.)+(0.0510876372718)
        ref[3,2,0,1]=(-0.0135357603866)/(o+1.)+(0.137621005039)
        ref[3,2,1,0]=(-1.00492518488)/(o+1.)+(0.396661482686)
        ref[3,2,1,1]=(-0.320901927851)/(o+1.)+(-0.463339233997)
        ref[3,2,2,0]=(2.43552217291)/(o+1.)+(0.621781827577)
        ref[3,2,2,1]=(1.41741894674)/(o+1.)+(0.804378003359)
        ref[3,3,0,0]=(0.748611138576)/(o+1.)+(-0.914035304315)
        ref[3,3,0,1]=(-0.375012840487)/(o+1.)+(0.0433116399723)
        ref[3,3,1,0]=(-0.0102160036467)/(o+1.)+(0.403000755649)
        ref[3,3,1,1]=(2.10162209317)/(o+1.)+(0.888826297208)
        ref[3,3,2,0]=(-0.449082281464)/(o+1.)+(0.521787043713)
        ref[3,3,2,1]=(-0.0103979620079)/(o+1.)+(-0.372545450994)
        ref[3,4,0,0]=(0.490562575696)/(o+1.)+(0.634809326777)
        ref[3,4,0,1]=(1.18954580773)/(o+1.)+(0.633837577475)
        ref[3,4,1,0]=(1.01376106573)/(o+1.)+(-0.301232420589)
        ref[3,4,1,1]=(-1.77747335429)/(o+1.)+(-0.100059050648)
        ref[3,4,2,0]=(1.44409584476)/(o+1.)+(-0.579358374008)
        ref[3,4,2,1]=(-1.78484029887)/(o+1.)+(0.0406454658476)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.788336609161)*x[0]**o+(-0.46539601764)*x[0]+(0.0970136798824)*x[1]**o+(-0.120665298671)*x[1]
        ref=(-0.691322929278)/(o+1.)+(-0.293030658156)
      else:
        arg=(0.922255261003)*x[0]**o+(0.0862232364707)*x[0]+(-0.94152635039)*x[1]**o+(0.771078486789)*x[1]+(-0.487642191181)*x[2]**o+(0.971754960362)*x[2]
        ref=(-0.506913280568)/(o+1.)+(0.914528341811)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.193071015033)*x[0]**o+(-0.281957941612)*x[0]+(0.269334552863)*x[1]**o+(0.0332262192804)*x[1]
        arg[1]=(-0.982634724425)*x[0]**o+(0.872638820737)*x[0]+(0.828028964603)*x[1]**o+(0.70464418136)*x[1]
        ref[0]=(0.0762635378296)/(o+1.)+(-0.124365861166)
        ref[1]=(-0.154605759822)/(o+1.)+(0.788641501048)
      else:
        arg[0]=(0.0746983082972)*x[0]**o+(0.281818436343)*x[0]+(-0.835321029206)*x[1]**o+(-0.349524270524)*x[1]+(-0.471905088601)*x[2]**o+(0.156327167402)*x[2]
        arg[1]=(-0.971935771255)*x[0]**o+(0.943151542057)*x[0]+(-0.60084258169)*x[1]**o+(0.458392855527)*x[1]+(-0.0173384187065)*x[2]**o+(0.538410338758)*x[2]
        ref[0]=(-1.23252780951)/(o+1.)+(0.0443106666103)
        ref[1]=(-1.59011677165)/(o+1.)+(0.969977368171)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.568783264759)*x[0]**o+(0.690377280868)*x[0]+(-0.689750922806)*x[1]**o+(-0.262828324126)*x[1]
        arg[0,1]=(-0.981657284673)*x[0]**o+(0.37118874852)*x[0]+(0.733992104146)*x[1]**o+(0.350268735699)*x[1]
        arg[0,2]=(0.713336100765)*x[0]**o+(0.737674447814)*x[0]+(0.239606043833)*x[1]**o+(0.814911366356)*x[1]
        arg[0,3]=(0.171706130967)*x[0]**o+(0.56324398491)*x[0]+(0.328884588679)*x[1]**o+(0.13307521808)*x[1]
        arg[0,4]=(-0.309130885049)*x[0]**o+(0.353596646674)*x[0]+(-0.282764686556)*x[1]**o+(-0.56977508596)*x[1]
        arg[1,0]=(0.974136161109)*x[0]**o+(-0.432477120938)*x[0]+(-0.841212164751)*x[1]**o+(-0.429211937256)*x[1]
        arg[1,1]=(0.0303890146585)*x[0]**o+(-0.712865964696)*x[0]+(0.572471431129)*x[1]**o+(-0.194324510922)*x[1]
        arg[1,2]=(-0.757134870981)*x[0]**o+(-0.018314520594)*x[0]+(0.250423290169)*x[1]**o+(-0.876678823729)*x[1]
        arg[1,3]=(-0.351885929068)*x[0]**o+(-0.561258605321)*x[0]+(0.208995575986)*x[1]**o+(0.428437856056)*x[1]
        arg[1,4]=(-0.313383912526)*x[0]**o+(-0.713484404328)*x[0]+(-0.279619274966)*x[1]**o+(0.72420661494)*x[1]
        arg[2,0]=(-0.686382641827)*x[0]**o+(0.678578408675)*x[0]+(-0.696532353786)*x[1]**o+(-0.154644995634)*x[1]
        arg[2,1]=(0.368548759951)*x[0]**o+(-0.674856247769)*x[0]+(0.908734263899)*x[1]**o+(0.331173572353)*x[1]
        arg[2,2]=(-0.1279136556)*x[0]**o+(-0.161474840876)*x[0]+(-0.775009015881)*x[1]**o+(-0.689552100815)*x[1]
        arg[2,3]=(-0.638450048485)*x[0]**o+(0.755485009392)*x[0]+(-0.793363159431)*x[1]**o+(0.255883452369)*x[1]
        arg[2,4]=(-0.831891512113)*x[0]**o+(-0.811678751836)*x[0]+(-0.487866621558)*x[1]**o+(0.240951890349)*x[1]
        arg[3,0]=(0.448249735424)*x[0]**o+(0.0809467993259)*x[0]+(0.232269153692)*x[1]**o+(0.477512628661)*x[1]
        arg[3,1]=(0.494093407104)*x[0]**o+(0.196761586998)*x[0]+(-0.702155296644)*x[1]**o+(-0.748641693731)*x[1]
        arg[3,2]=(0.432370067164)*x[0]**o+(-0.365671005874)*x[0]+(0.452009309356)*x[1]**o+(-0.951328172558)*x[1]
        arg[3,3]=(-0.993290240028)*x[0]**o+(0.172534714933)*x[0]+(0.00917693958413)*x[1]**o+(0.417515657393)*x[1]
        arg[3,4]=(0.541299019375)*x[0]**o+(0.479118172057)*x[0]+(-0.502475287154)*x[1]**o+(-0.04397899716)*x[1]
        ref[0,0]=(-0.120967658047)/(o+1.)+(0.213774478371)
        ref[0,1]=(-0.247665180527)/(o+1.)+(0.360728742109)
        ref[0,2]=(0.952942144598)/(o+1.)+(0.776292907085)
        ref[0,3]=(0.500590719646)/(o+1.)+(0.348159601495)
        ref[0,4]=(-0.591895571605)/(o+1.)+(-0.108089219643)
        ref[1,0]=(0.132923996358)/(o+1.)+(-0.430844529097)
        ref[1,1]=(0.602860445788)/(o+1.)+(-0.453595237809)
        ref[1,2]=(-0.506711580812)/(o+1.)+(-0.447496672161)
        ref[1,3]=(-0.142890353081)/(o+1.)+(-0.066410374632)
        ref[1,4]=(-0.593003187492)/(o+1.)+(0.00536110530621)
        ref[2,0]=(-1.38291499561)/(o+1.)+(0.261966706521)
        ref[2,1]=(1.27728302385)/(o+1.)+(-0.171841337708)
        ref[2,2]=(-0.902922671481)/(o+1.)+(-0.425513470846)
        ref[2,3]=(-1.43181320792)/(o+1.)+(0.505684230881)
        ref[2,4]=(-1.31975813367)/(o+1.)+(-0.285363430743)
        ref[3,0]=(0.680518889117)/(o+1.)+(0.279229713993)
        ref[3,1]=(-0.20806188954)/(o+1.)+(-0.275940053366)
        ref[3,2]=(0.88437937652)/(o+1.)+(-0.658499589216)
        ref[3,3]=(-0.984113300443)/(o+1.)+(0.295025186163)
        ref[3,4]=(0.0388237322207)/(o+1.)+(0.217569587448)
      else:
        arg[0,0]=(-0.866239193113)*x[0]**o+(0.479559390215)*x[0]+(0.512920648125)*x[1]**o+(-0.504775214511)*x[1]+(0.462210441279)*x[2]**o+(0.000249131516168)*x[2]
        arg[0,1]=(-0.11397206495)*x[0]**o+(0.0762933475692)*x[0]+(0.347716642793)*x[1]**o+(0.373262578958)*x[1]+(0.965468442129)*x[2]**o+(-0.743205128452)*x[2]
        arg[0,2]=(-0.214519408107)*x[0]**o+(-0.663675446512)*x[0]+(-0.424102625804)*x[1]**o+(-0.426644021856)*x[1]+(-0.0731662128704)*x[2]**o+(-0.375185307096)*x[2]
        arg[0,3]=(0.599595484989)*x[0]**o+(-0.303187098474)*x[0]+(0.278524134467)*x[1]**o+(-0.579432425528)*x[1]+(0.006698672684)*x[2]**o+(-0.820721120517)*x[2]
        arg[0,4]=(0.538847126305)*x[0]**o+(0.58093792853)*x[0]+(-0.21189097773)*x[1]**o+(0.637101424777)*x[1]+(0.0155994636183)*x[2]**o+(0.488675370945)*x[2]
        arg[1,0]=(-0.775581481064)*x[0]**o+(-0.968381856562)*x[0]+(-0.660755920498)*x[1]**o+(0.746481830633)*x[1]+(-0.300061615221)*x[2]**o+(-0.720031101537)*x[2]
        arg[1,1]=(0.990001351802)*x[0]**o+(-0.168997300944)*x[0]+(0.0134555819367)*x[1]**o+(0.969355591009)*x[1]+(-0.347553438064)*x[2]**o+(0.648537935984)*x[2]
        arg[1,2]=(-0.327760535516)*x[0]**o+(0.517489199537)*x[0]+(-0.0589046196438)*x[1]**o+(-0.893401439507)*x[1]+(-0.0821634552571)*x[2]**o+(0.861266440343)*x[2]
        arg[1,3]=(-0.808782893446)*x[0]**o+(-0.196363777506)*x[0]+(0.623837845847)*x[1]**o+(-0.05392424847)*x[1]+(0.616011657554)*x[2]**o+(0.800682023434)*x[2]
        arg[1,4]=(-0.425073222968)*x[0]**o+(-0.635248695204)*x[0]+(0.947222621311)*x[1]**o+(0.253198813179)*x[1]+(-0.907568459712)*x[2]**o+(-0.513049744062)*x[2]
        arg[2,0]=(-0.103867107528)*x[0]**o+(0.893031214224)*x[0]+(0.151292631908)*x[1]**o+(-0.407189878657)*x[1]+(0.618120612304)*x[2]**o+(0.983992964931)*x[2]
        arg[2,1]=(0.310432646627)*x[0]**o+(0.235143842057)*x[0]+(0.729683627093)*x[1]**o+(0.891778928523)*x[1]+(-0.148927367367)*x[2]**o+(0.913403960844)*x[2]
        arg[2,2]=(0.610293692912)*x[0]**o+(-0.28497731539)*x[0]+(-0.926462472232)*x[1]**o+(-0.427657211191)*x[1]+(0.467905160751)*x[2]**o+(0.195578740767)*x[2]
        arg[2,3]=(0.854869452114)*x[0]**o+(-0.752530632831)*x[0]+(0.0128220565485)*x[1]**o+(-0.726013651836)*x[1]+(-0.0979888078519)*x[2]**o+(-0.658955703807)*x[2]
        arg[2,4]=(0.536447848289)*x[0]**o+(-0.658205406579)*x[0]+(0.66427484293)*x[1]**o+(-0.250014589496)*x[1]+(-0.0654257159073)*x[2]**o+(0.538196496913)*x[2]
        arg[3,0]=(-0.514665034974)*x[0]**o+(-0.278189747423)*x[0]+(0.402996329453)*x[1]**o+(-0.0362969716175)*x[1]+(-0.707890815004)*x[2]**o+(-0.298762865096)*x[2]
        arg[3,1]=(0.7342439327)*x[0]**o+(-0.387336714116)*x[0]+(-0.67888634151)*x[1]**o+(0.647322640846)*x[1]+(-0.0346274738413)*x[2]**o+(0.985950301913)*x[2]
        arg[3,2]=(0.687555064124)*x[0]**o+(0.345506248144)*x[0]+(0.658692258064)*x[1]**o+(0.810796031447)*x[1]+(0.102608006398)*x[2]**o+(0.349877767777)*x[2]
        arg[3,3]=(-0.456241806837)*x[0]**o+(-0.181224535865)*x[0]+(0.401842456169)*x[1]**o+(-0.413936620519)*x[1]+(0.436794604974)*x[2]**o+(0.95051042082)*x[2]
        arg[3,4]=(-0.330739393408)*x[0]**o+(-0.261876658969)*x[0]+(0.287474017623)*x[1]**o+(0.463691723118)*x[1]+(0.757597823206)*x[2]**o+(0.49336246207)*x[2]
        ref[0,0]=(0.108891896292)/(o+1.)+(-0.01248334639)
        ref[0,1]=(1.19921301997)/(o+1.)+(-0.146824600962)
        ref[0,2]=(-0.711788246781)/(o+1.)+(-0.732752387732)
        ref[0,3]=(0.88481829214)/(o+1.)+(-0.85167032226)
        ref[0,4]=(0.342555612194)/(o+1.)+(0.853357362126)
        ref[1,0]=(-1.73639901678)/(o+1.)+(-0.470965563733)
        ref[1,1]=(0.655903495675)/(o+1.)+(0.724448113024)
        ref[1,2]=(-0.468828610417)/(o+1.)+(0.242677100186)
        ref[1,3]=(0.431066609955)/(o+1.)+(0.275196998729)
        ref[1,4]=(-0.385419061368)/(o+1.)+(-0.447549813044)
        ref[2,0]=(0.665546136685)/(o+1.)+(0.734917150249)
        ref[2,1]=(0.891188906354)/(o+1.)+(1.02016336571)
        ref[2,2]=(0.151736381432)/(o+1.)+(-0.258527892907)
        ref[2,3]=(0.76970270081)/(o+1.)+(-1.06874999424)
        ref[2,4]=(1.13529697531)/(o+1.)+(-0.185011749581)
        ref[3,0]=(-0.819559520525)/(o+1.)+(-0.306624792068)
        ref[3,1]=(0.020730117348)/(o+1.)+(0.622968114321)
        ref[3,2]=(1.44885532859)/(o+1.)+(0.753090023684)
        ref[3,3]=(0.382395254306)/(o+1.)+(0.177674632218)
        ref[3,4]=(0.714332447421)/(o+1.)+(0.34758876311)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.779261125448)*x[0]**o+(-0.640503838419)*x[0]+(0.589085035572)*x[1]**o+(-0.0426687544445)*x[1]
        arg[0,0,1]=(-0.419361292473)*x[0]**o+(0.0850684591011)*x[0]+(-0.440758336688)*x[1]**o+(0.151015611594)*x[1]
        arg[0,1,0]=(-0.727043966709)*x[0]**o+(0.711730063365)*x[0]+(0.657759819512)*x[1]**o+(-0.210981156609)*x[1]
        arg[0,1,1]=(-0.276491768644)*x[0]**o+(-0.324865684188)*x[0]+(-0.921856877627)*x[1]**o+(-0.246833902926)*x[1]
        arg[1,0,0]=(0.720541615152)*x[0]**o+(-0.0406851820986)*x[0]+(0.33976001267)*x[1]**o+(0.937989079378)*x[1]
        arg[1,0,1]=(0.239272962875)*x[0]**o+(0.482026352856)*x[0]+(0.159148474482)*x[1]**o+(0.680831564145)*x[1]
        arg[1,1,0]=(0.55170300456)*x[0]**o+(-0.261357073238)*x[0]+(0.298185649596)*x[1]**o+(-0.540598468483)*x[1]
        arg[1,1,1]=(0.508262701362)*x[0]**o+(-0.531692643817)*x[0]+(0.151424571021)*x[1]**o+(-0.77488755481)*x[1]
        arg[2,0,0]=(0.265201389846)*x[0]**o+(0.377300848247)*x[0]+(0.0402148688087)*x[1]**o+(0.0263806695524)*x[1]
        arg[2,0,1]=(0.321397565411)*x[0]**o+(-0.277650913698)*x[0]+(0.909430556637)*x[1]**o+(-0.43244155765)*x[1]
        arg[2,1,0]=(-0.671417887851)*x[0]**o+(-0.772328065901)*x[0]+(0.0896878386442)*x[1]**o+(0.584773963567)*x[1]
        arg[2,1,1]=(0.318995094412)*x[0]**o+(-0.396578444213)*x[0]+(-0.627767362562)*x[1]**o+(-0.588717342984)*x[1]
        arg[3,0,0]=(-0.154999830633)*x[0]**o+(-0.263434767944)*x[0]+(0.0704862479101)*x[1]**o+(0.216237505495)*x[1]
        arg[3,0,1]=(0.956868643095)*x[0]**o+(-0.0331644426873)*x[0]+(-0.236340025853)*x[1]**o+(-0.814705425837)*x[1]
        arg[3,1,0]=(-0.477985064084)*x[0]**o+(-0.368896833289)*x[0]+(0.208122873207)*x[1]**o+(-0.566525593432)*x[1]
        arg[3,1,1]=(-0.299257275248)*x[0]**o+(0.937022356746)*x[0]+(0.671724793932)*x[1]**o+(0.80705820341)*x[1]
        arg[4,0,0]=(0.857855249735)*x[0]**o+(0.537756541143)*x[0]+(0.611784225987)*x[1]**o+(0.944061991186)*x[1]
        arg[4,0,1]=(0.825779290308)*x[0]**o+(-0.636461636739)*x[0]+(-0.382007450888)*x[1]**o+(0.170178262303)*x[1]
        arg[4,1,0]=(0.275666857657)*x[0]**o+(-0.404983003194)*x[0]+(0.476932065223)*x[1]**o+(0.931645726493)*x[1]
        arg[4,1,1]=(0.956182179652)*x[0]**o+(0.543482655321)*x[0]+(0.60492494312)*x[1]**o+(0.384497838341)*x[1]
        arg[5,0,0]=(0.476937368598)*x[0]**o+(-0.588434801718)*x[0]+(0.865080125217)*x[1]**o+(-0.954493375041)*x[1]
        arg[5,0,1]=(0.151850214007)*x[0]**o+(0.252651124764)*x[0]+(-0.570132867968)*x[1]**o+(-0.0885902498736)*x[1]
        arg[5,1,0]=(-0.69330471578)*x[0]**o+(0.165874518143)*x[0]+(-0.610503496045)*x[1]**o+(-0.871951964138)*x[1]
        arg[5,1,1]=(-0.539191264945)*x[0]**o+(0.558474736927)*x[0]+(0.752045569996)*x[1]**o+(-0.5305308467)*x[1]
        ref[0,0,0]=(1.36834616102)/(o+1.)+(-0.341586296432)
        ref[0,0,1]=(-0.860119629161)/(o+1.)+(0.118042035348)
        ref[0,1,0]=(-0.0692841471964)/(o+1.)+(0.250374453378)
        ref[0,1,1]=(-1.19834864627)/(o+1.)+(-0.285849793557)
        ref[1,0,0]=(1.06030162782)/(o+1.)+(0.448651948639)
        ref[1,0,1]=(0.398421437357)/(o+1.)+(0.5814289585)
        ref[1,1,0]=(0.849888654155)/(o+1.)+(-0.400977770861)
        ref[1,1,1]=(0.659687272383)/(o+1.)+(-0.653290099314)
        ref[2,0,0]=(0.305416258655)/(o+1.)+(0.2018407589)
        ref[2,0,1]=(1.23082812205)/(o+1.)+(-0.355046235674)
        ref[2,1,0]=(-0.581730049206)/(o+1.)+(-0.0937770511668)
        ref[2,1,1]=(-0.30877226815)/(o+1.)+(-0.492647893598)
        ref[3,0,0]=(-0.0845135827233)/(o+1.)+(-0.0235986312242)
        ref[3,0,1]=(0.720528617241)/(o+1.)+(-0.423934934262)
        ref[3,1,0]=(-0.269862190877)/(o+1.)+(-0.46771121336)
        ref[3,1,1]=(0.372467518684)/(o+1.)+(0.872040280078)
        ref[4,0,0]=(1.46963947572)/(o+1.)+(0.740909266165)
        ref[4,0,1]=(0.44377183942)/(o+1.)+(-0.233141687218)
        ref[4,1,0]=(0.75259892288)/(o+1.)+(0.26333136165)
        ref[4,1,1]=(1.56110712277)/(o+1.)+(0.463990246831)
        ref[5,0,0]=(1.34201749382)/(o+1.)+(-0.771464088379)
        ref[5,0,1]=(-0.418282653961)/(o+1.)+(0.0820304374451)
        ref[5,1,0]=(-1.30380821182)/(o+1.)+(-0.353038722997)
        ref[5,1,1]=(0.212854305051)/(o+1.)+(0.0139719451134)
      else:
        arg[0,0,0]=(-0.825926001774)*x[0]**o+(0.418049316138)*x[0]+(0.668695279849)*x[1]**o+(-0.7323605258)*x[1]+(-0.637717742224)*x[2]**o+(0.257065974143)*x[2]
        arg[0,0,1]=(-0.169779448443)*x[0]**o+(-0.648612931029)*x[0]+(-0.643814352995)*x[1]**o+(0.396704595574)*x[1]+(-0.248766540261)*x[2]**o+(-0.736410673282)*x[2]
        arg[0,1,0]=(-0.371929441192)*x[0]**o+(0.438661796459)*x[0]+(-0.893450426264)*x[1]**o+(0.643494077636)*x[1]+(0.252099102122)*x[2]**o+(-0.0857798075076)*x[2]
        arg[0,1,1]=(0.336251711034)*x[0]**o+(-0.317426111079)*x[0]+(-0.801221986791)*x[1]**o+(0.281123033042)*x[1]+(0.921153543163)*x[2]**o+(0.497003835923)*x[2]
        arg[1,0,0]=(0.40897311872)*x[0]**o+(0.326894801733)*x[0]+(0.216268807497)*x[1]**o+(0.946235364974)*x[1]+(-0.480995485096)*x[2]**o+(0.869772937982)*x[2]
        arg[1,0,1]=(0.108715554449)*x[0]**o+(0.0847529231253)*x[0]+(0.866120965844)*x[1]**o+(0.43989772616)*x[1]+(0.67962061792)*x[2]**o+(0.90311988632)*x[2]
        arg[1,1,0]=(0.28571484173)*x[0]**o+(0.593050819853)*x[0]+(-0.384864010146)*x[1]**o+(-0.57920566259)*x[1]+(-0.304024223769)*x[2]**o+(-0.0279518764989)*x[2]
        arg[1,1,1]=(0.249126658892)*x[0]**o+(0.243355410215)*x[0]+(0.154816978135)*x[1]**o+(-0.260343533153)*x[1]+(-0.973133787852)*x[2]**o+(0.752545508531)*x[2]
        arg[2,0,0]=(0.103612497259)*x[0]**o+(0.0919234144108)*x[0]+(0.416394965635)*x[1]**o+(-0.0132359939268)*x[1]+(-0.522775855178)*x[2]**o+(-0.985811066906)*x[2]
        arg[2,0,1]=(-0.592641488144)*x[0]**o+(0.703703144485)*x[0]+(-0.0149046721584)*x[1]**o+(0.220368283581)*x[1]+(-0.656858164281)*x[2]**o+(0.165951102113)*x[2]
        arg[2,1,0]=(0.309492848012)*x[0]**o+(-0.301037623847)*x[0]+(-0.224814259669)*x[1]**o+(-0.950794406688)*x[1]+(-0.296088293363)*x[2]**o+(-0.712930293168)*x[2]
        arg[2,1,1]=(0.0592633594166)*x[0]**o+(-0.179007619512)*x[0]+(-0.922169690831)*x[1]**o+(-0.778228075089)*x[1]+(-0.135749453777)*x[2]**o+(-0.146265051618)*x[2]
        arg[3,0,0]=(-0.71773633584)*x[0]**o+(0.471551385738)*x[0]+(-0.041415831943)*x[1]**o+(0.343380640343)*x[1]+(-0.211821446187)*x[2]**o+(-0.661201893183)*x[2]
        arg[3,0,1]=(-0.159071045836)*x[0]**o+(0.216715127143)*x[0]+(-0.846847485213)*x[1]**o+(0.662226888693)*x[1]+(0.050806337433)*x[2]**o+(0.744071347051)*x[2]
        arg[3,1,0]=(-0.363431877105)*x[0]**o+(-0.461513470653)*x[0]+(-0.130947882739)*x[1]**o+(-0.624729599981)*x[1]+(0.778572567699)*x[2]**o+(0.324584316744)*x[2]
        arg[3,1,1]=(0.395432262346)*x[0]**o+(0.826450301981)*x[0]+(0.176721657975)*x[1]**o+(-0.370954952405)*x[1]+(-0.708437370609)*x[2]**o+(-0.150075114354)*x[2]
        arg[4,0,0]=(-0.18662583335)*x[0]**o+(-0.0120831234457)*x[0]+(0.805686139762)*x[1]**o+(0.30195413136)*x[1]+(0.648067854031)*x[2]**o+(0.28073240632)*x[2]
        arg[4,0,1]=(-0.913625644854)*x[0]**o+(-0.264637827098)*x[0]+(-0.6531590441)*x[1]**o+(0.396422722623)*x[1]+(0.280591271146)*x[2]**o+(0.498346191707)*x[2]
        arg[4,1,0]=(0.208029597907)*x[0]**o+(-0.295747909396)*x[0]+(-0.541509895186)*x[1]**o+(-0.511394562575)*x[1]+(-0.459197562982)*x[2]**o+(-0.749433614666)*x[2]
        arg[4,1,1]=(-0.29700028905)*x[0]**o+(-0.590932598278)*x[0]+(0.36219700213)*x[1]**o+(0.482759242709)*x[1]+(-0.135696240971)*x[2]**o+(0.49658807849)*x[2]
        arg[5,0,0]=(-0.800345624219)*x[0]**o+(0.259709746207)*x[0]+(0.480073493816)*x[1]**o+(0.499332550127)*x[1]+(0.691794186846)*x[2]**o+(-0.319369042249)*x[2]
        arg[5,0,1]=(0.300879052118)*x[0]**o+(-0.985396817441)*x[0]+(0.971136879002)*x[1]**o+(0.806560091745)*x[1]+(0.361308859099)*x[2]**o+(0.264277089732)*x[2]
        arg[5,1,0]=(0.271857906548)*x[0]**o+(0.338260411675)*x[0]+(-0.170292665803)*x[1]**o+(0.675491566142)*x[1]+(0.897921755421)*x[2]**o+(-0.829233864443)*x[2]
        arg[5,1,1]=(0.673366845241)*x[0]**o+(0.121940550422)*x[0]+(0.116054155545)*x[1]**o+(0.0456556938689)*x[1]+(-0.978127590673)*x[2]**o+(0.180985351995)*x[2]
        ref[0,0,0]=(-0.794948464148)/(o+1.)+(-0.0286226177598)
        ref[0,0,1]=(-1.0623603417)/(o+1.)+(-0.494159504368)
        ref[0,1,0]=(-1.01328076533)/(o+1.)+(0.498188033294)
        ref[0,1,1]=(0.456183267406)/(o+1.)+(0.230350378943)
        ref[1,0,0]=(0.144246441122)/(o+1.)+(1.07145155234)
        ref[1,0,1]=(1.65445713821)/(o+1.)+(0.713885267803)
        ref[1,1,0]=(-0.403173392185)/(o+1.)+(-0.0070533596178)
        ref[1,1,1]=(-0.569190150825)/(o+1.)+(0.367778692797)
        ref[2,0,0]=(-0.00276839228366)/(o+1.)+(-0.453561823211)
        ref[2,0,1]=(-1.26440432458)/(o+1.)+(0.545011265089)
        ref[2,1,0]=(-0.21140970502)/(o+1.)+(-0.982381161851)
        ref[2,1,1]=(-0.998655785191)/(o+1.)+(-0.55175037311)
        ref[3,0,0]=(-0.97097361397)/(o+1.)+(0.0768650664488)
        ref[3,0,1]=(-0.955112193616)/(o+1.)+(0.811506681444)
        ref[3,1,0]=(0.284192807855)/(o+1.)+(-0.380829376946)
        ref[3,1,1]=(-0.136283450288)/(o+1.)+(0.152710117611)
        ref[4,0,0]=(1.26712816044)/(o+1.)+(0.285301707117)
        ref[4,0,1]=(-1.28619341781)/(o+1.)+(0.315065543616)
        ref[4,1,0]=(-0.792677860261)/(o+1.)+(-0.778288043318)
        ref[4,1,1]=(-0.0704995278906)/(o+1.)+(0.194207361461)
        ref[5,0,0]=(0.371522056443)/(o+1.)+(0.219836627043)
        ref[5,0,1]=(1.63332479022)/(o+1.)+(0.0427201820182)
        ref[5,1,0]=(0.999486996166)/(o+1.)+(0.0922590566873)
        ref[5,1,1]=(-0.188706589887)/(o+1.)+(0.174290798143)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.647932589381)*x[0]**o+(-0.092686685918)*x[0]+(0.167888464541)*x[1]**o+(0.698445840627)*x[1]
        arg[0,0,0,1]=(0.413204959658)*x[0]**o+(0.772926029512)*x[0]+(-0.63702948026)*x[1]**o+(0.608392823861)*x[1]
        arg[0,0,1,0]=(0.303506932052)*x[0]**o+(0.650309664063)*x[0]+(-0.805238344485)*x[1]**o+(-0.62971680995)*x[1]
        arg[0,0,1,1]=(-0.205701190736)*x[0]**o+(-0.0834042840814)*x[0]+(-0.525839398431)*x[1]**o+(0.779973223005)*x[1]
        arg[0,0,2,0]=(-0.97954633631)*x[0]**o+(-0.298973871114)*x[0]+(0.858022652298)*x[1]**o+(0.667671796632)*x[1]
        arg[0,0,2,1]=(-0.790133712345)*x[0]**o+(-0.733743414808)*x[0]+(0.337185148145)*x[1]**o+(-0.927853051244)*x[1]
        arg[0,1,0,0]=(0.617198112458)*x[0]**o+(0.70438849416)*x[0]+(-0.861661596364)*x[1]**o+(-0.251796116219)*x[1]
        arg[0,1,0,1]=(0.37107453636)*x[0]**o+(-0.490440661153)*x[0]+(0.849252675242)*x[1]**o+(-0.896547775077)*x[1]
        arg[0,1,1,0]=(0.375080744007)*x[0]**o+(-0.609557663503)*x[0]+(-0.930846980977)*x[1]**o+(0.0623096843381)*x[1]
        arg[0,1,1,1]=(-0.809863615805)*x[0]**o+(-0.999724521803)*x[0]+(0.493027030633)*x[1]**o+(-0.672546253752)*x[1]
        arg[0,1,2,0]=(0.219294077412)*x[0]**o+(-0.7854819847)*x[0]+(-0.353026321798)*x[1]**o+(-0.965348418509)*x[1]
        arg[0,1,2,1]=(0.177236321722)*x[0]**o+(-0.860491592895)*x[0]+(-0.458724268899)*x[1]**o+(-0.476508229653)*x[1]
        arg[0,2,0,0]=(0.0760611767074)*x[0]**o+(-0.00296398964042)*x[0]+(-0.972843039436)*x[1]**o+(-0.0422057743881)*x[1]
        arg[0,2,0,1]=(-0.926737777701)*x[0]**o+(-0.895791593378)*x[0]+(-0.776529150964)*x[1]**o+(0.557115706978)*x[1]
        arg[0,2,1,0]=(0.353264702858)*x[0]**o+(0.632583933138)*x[0]+(-0.354112684469)*x[1]**o+(0.0541346174117)*x[1]
        arg[0,2,1,1]=(-0.124181989217)*x[0]**o+(-0.803262099858)*x[0]+(0.733246361984)*x[1]**o+(0.0639821744038)*x[1]
        arg[0,2,2,0]=(-0.446733845136)*x[0]**o+(-0.187373171381)*x[0]+(-0.312309043572)*x[1]**o+(0.830789380477)*x[1]
        arg[0,2,2,1]=(0.403889932995)*x[0]**o+(-0.243474625233)*x[0]+(0.35707779102)*x[1]**o+(0.29725421988)*x[1]
        arg[0,3,0,0]=(-0.856376546696)*x[0]**o+(0.448431504377)*x[0]+(-0.535187100108)*x[1]**o+(-0.90994242894)*x[1]
        arg[0,3,0,1]=(-0.321475754775)*x[0]**o+(0.0389358177737)*x[0]+(-0.0949733279735)*x[1]**o+(0.921937240884)*x[1]
        arg[0,3,1,0]=(0.591276336387)*x[0]**o+(-0.598236438303)*x[0]+(-0.829829995185)*x[1]**o+(0.539669346291)*x[1]
        arg[0,3,1,1]=(-0.944726366074)*x[0]**o+(-0.796928273271)*x[0]+(-0.424606656681)*x[1]**o+(0.35341306063)*x[1]
        arg[0,3,2,0]=(-0.850914599406)*x[0]**o+(0.791179202404)*x[0]+(-0.0220845565851)*x[1]**o+(0.594756136957)*x[1]
        arg[0,3,2,1]=(0.355862680641)*x[0]**o+(0.134261680432)*x[0]+(0.477798827075)*x[1]**o+(0.508452338499)*x[1]
        arg[0,4,0,0]=(0.481974944781)*x[0]**o+(0.259271175551)*x[0]+(0.953827692515)*x[1]**o+(0.364160709544)*x[1]
        arg[0,4,0,1]=(-0.316627315143)*x[0]**o+(-0.856616735533)*x[0]+(0.308508314695)*x[1]**o+(-0.677602758367)*x[1]
        arg[0,4,1,0]=(-0.76069824745)*x[0]**o+(0.282351651183)*x[0]+(-0.15964276552)*x[1]**o+(-0.344276428838)*x[1]
        arg[0,4,1,1]=(9.00669303738e-05)*x[0]**o+(0.948498888115)*x[0]+(0.690370267333)*x[1]**o+(0.619119188327)*x[1]
        arg[0,4,2,0]=(0.957692561039)*x[0]**o+(-0.147005436631)*x[0]+(-0.177840574215)*x[1]**o+(-0.28259452614)*x[1]
        arg[0,4,2,1]=(-0.431771943451)*x[0]**o+(-0.158132166651)*x[0]+(-0.134385316147)*x[1]**o+(0.751243590835)*x[1]
        arg[1,0,0,0]=(-0.816841420645)*x[0]**o+(0.55608373851)*x[0]+(-0.885838014201)*x[1]**o+(-0.467159163867)*x[1]
        arg[1,0,0,1]=(-0.543361117401)*x[0]**o+(0.329490019675)*x[0]+(0.942703992489)*x[1]**o+(-0.261273838803)*x[1]
        arg[1,0,1,0]=(0.865108106801)*x[0]**o+(-0.376972427024)*x[0]+(-0.0257304693919)*x[1]**o+(0.694210909726)*x[1]
        arg[1,0,1,1]=(0.428354498121)*x[0]**o+(0.962704601892)*x[0]+(-0.442636806139)*x[1]**o+(0.424724924149)*x[1]
        arg[1,0,2,0]=(-0.157277242299)*x[0]**o+(0.487588443251)*x[0]+(-0.813089096341)*x[1]**o+(0.0406316718829)*x[1]
        arg[1,0,2,1]=(0.345322702588)*x[0]**o+(0.224429042659)*x[0]+(-0.24109926586)*x[1]**o+(0.433141241931)*x[1]
        arg[1,1,0,0]=(0.956901263519)*x[0]**o+(0.48418062914)*x[0]+(0.14089030013)*x[1]**o+(-0.492352201717)*x[1]
        arg[1,1,0,1]=(-0.679781695411)*x[0]**o+(0.29857196636)*x[0]+(-0.234968984185)*x[1]**o+(0.114575919411)*x[1]
        arg[1,1,1,0]=(0.488378272934)*x[0]**o+(-0.358508327343)*x[0]+(-0.337958846849)*x[1]**o+(0.189303452912)*x[1]
        arg[1,1,1,1]=(-0.506993859574)*x[0]**o+(-0.682478021258)*x[0]+(-0.540344338899)*x[1]**o+(0.366965535976)*x[1]
        arg[1,1,2,0]=(-0.681151638572)*x[0]**o+(0.790837550009)*x[0]+(-0.810785716023)*x[1]**o+(-0.00739097962416)*x[1]
        arg[1,1,2,1]=(-0.355290573677)*x[0]**o+(-0.263294580108)*x[0]+(0.10469018402)*x[1]**o+(-0.650388283116)*x[1]
        arg[1,2,0,0]=(0.906212689905)*x[0]**o+(-0.673956103538)*x[0]+(0.305335278934)*x[1]**o+(-0.0805949628719)*x[1]
        arg[1,2,0,1]=(-0.98311373852)*x[0]**o+(-0.165316704542)*x[0]+(0.197637770603)*x[1]**o+(0.647989184342)*x[1]
        arg[1,2,1,0]=(0.415990736251)*x[0]**o+(-0.710476520822)*x[0]+(0.791512786296)*x[1]**o+(-0.10853188742)*x[1]
        arg[1,2,1,1]=(0.04337234954)*x[0]**o+(0.595650094449)*x[0]+(0.228098665648)*x[1]**o+(-0.868930899722)*x[1]
        arg[1,2,2,0]=(-0.791842482688)*x[0]**o+(-0.611246387621)*x[0]+(0.0504978789412)*x[1]**o+(0.624626300694)*x[1]
        arg[1,2,2,1]=(0.0207375595263)*x[0]**o+(0.560638072838)*x[0]+(0.44126312508)*x[1]**o+(0.169151353084)*x[1]
        arg[1,3,0,0]=(-0.0512069089279)*x[0]**o+(0.51110271913)*x[0]+(0.0785633458626)*x[1]**o+(0.836821453174)*x[1]
        arg[1,3,0,1]=(0.185737468851)*x[0]**o+(0.516071031227)*x[0]+(0.0995388802342)*x[1]**o+(0.25527909092)*x[1]
        arg[1,3,1,0]=(0.318240681435)*x[0]**o+(-0.442725217398)*x[0]+(0.533066616909)*x[1]**o+(0.679570707339)*x[1]
        arg[1,3,1,1]=(0.353367397644)*x[0]**o+(-0.396487189861)*x[0]+(-0.117773711001)*x[1]**o+(0.885897675293)*x[1]
        arg[1,3,2,0]=(-0.647645988694)*x[0]**o+(-0.755258662164)*x[0]+(0.288300257472)*x[1]**o+(0.505342385389)*x[1]
        arg[1,3,2,1]=(0.000781386924262)*x[0]**o+(0.190244787656)*x[0]+(0.589388834071)*x[1]**o+(0.669705413859)*x[1]
        arg[1,4,0,0]=(0.320107896841)*x[0]**o+(0.673508197912)*x[0]+(0.667026895874)*x[1]**o+(-0.818578971562)*x[1]
        arg[1,4,0,1]=(-0.226755915796)*x[0]**o+(0.71800207991)*x[0]+(0.850490212043)*x[1]**o+(0.261708964138)*x[1]
        arg[1,4,1,0]=(-0.0147350282361)*x[0]**o+(0.743889536542)*x[0]+(-0.898802067607)*x[1]**o+(0.711589170571)*x[1]
        arg[1,4,1,1]=(-0.816145493818)*x[0]**o+(0.405931568806)*x[0]+(0.159156576579)*x[1]**o+(0.143576116984)*x[1]
        arg[1,4,2,0]=(0.994791165013)*x[0]**o+(0.65054901122)*x[0]+(0.039218059733)*x[1]**o+(-0.662978657934)*x[1]
        arg[1,4,2,1]=(-0.792277825616)*x[0]**o+(-0.257224872111)*x[0]+(-0.244502392787)*x[1]**o+(-0.276042189495)*x[1]
        arg[2,0,0,0]=(0.133211578016)*x[0]**o+(0.10029141733)*x[0]+(-0.52480547182)*x[1]**o+(0.814652707336)*x[1]
        arg[2,0,0,1]=(0.836429340685)*x[0]**o+(-0.837709322969)*x[0]+(-0.442953015462)*x[1]**o+(-0.984757344044)*x[1]
        arg[2,0,1,0]=(0.893738853578)*x[0]**o+(-0.238050980408)*x[0]+(-0.913668527302)*x[1]**o+(0.895136777323)*x[1]
        arg[2,0,1,1]=(-0.0457361766611)*x[0]**o+(0.414814798208)*x[0]+(-0.57261759439)*x[1]**o+(-0.358969048076)*x[1]
        arg[2,0,2,0]=(-0.320394419991)*x[0]**o+(-0.990129952721)*x[0]+(0.890888895016)*x[1]**o+(0.241063301452)*x[1]
        arg[2,0,2,1]=(-0.604164679932)*x[0]**o+(-0.598939404114)*x[0]+(-0.606633388454)*x[1]**o+(-0.666879580873)*x[1]
        arg[2,1,0,0]=(0.68515547034)*x[0]**o+(0.892247437859)*x[0]+(0.642003383991)*x[1]**o+(0.569240970392)*x[1]
        arg[2,1,0,1]=(0.918713108587)*x[0]**o+(0.865547962284)*x[0]+(0.517952260633)*x[1]**o+(-0.307334010857)*x[1]
        arg[2,1,1,0]=(0.353382468404)*x[0]**o+(-0.542718692644)*x[0]+(-0.957182863451)*x[1]**o+(-0.889352632225)*x[1]
        arg[2,1,1,1]=(0.505185847817)*x[0]**o+(-0.71137143775)*x[0]+(0.853785404375)*x[1]**o+(-0.731028307335)*x[1]
        arg[2,1,2,0]=(-0.0805774093876)*x[0]**o+(0.949831911897)*x[0]+(0.511009248375)*x[1]**o+(-0.0192710206538)*x[1]
        arg[2,1,2,1]=(-0.987709862593)*x[0]**o+(-0.536077508064)*x[0]+(-0.0726882784026)*x[1]**o+(-0.27275457372)*x[1]
        arg[2,2,0,0]=(0.271001953238)*x[0]**o+(0.0446246860776)*x[0]+(0.103657462724)*x[1]**o+(0.359900072694)*x[1]
        arg[2,2,0,1]=(0.613460277623)*x[0]**o+(-0.23324730519)*x[0]+(0.188458167579)*x[1]**o+(-0.574087648909)*x[1]
        arg[2,2,1,0]=(-0.787444207567)*x[0]**o+(0.143772612473)*x[0]+(-0.720451072166)*x[1]**o+(0.936819171338)*x[1]
        arg[2,2,1,1]=(-0.219701379296)*x[0]**o+(-0.343100221149)*x[0]+(-0.972737482064)*x[1]**o+(-0.614208415137)*x[1]
        arg[2,2,2,0]=(0.0292070295438)*x[0]**o+(-0.694263454424)*x[0]+(0.188064440036)*x[1]**o+(0.0901763471697)*x[1]
        arg[2,2,2,1]=(0.407629585245)*x[0]**o+(-0.745619322197)*x[0]+(0.949044983116)*x[1]**o+(-0.316057288128)*x[1]
        arg[2,3,0,0]=(0.124212182643)*x[0]**o+(0.621703244759)*x[0]+(-0.0592136772296)*x[1]**o+(0.571981450786)*x[1]
        arg[2,3,0,1]=(-0.683989210442)*x[0]**o+(-0.318774594157)*x[0]+(0.850235115277)*x[1]**o+(-0.871075179838)*x[1]
        arg[2,3,1,0]=(0.478657634044)*x[0]**o+(-0.43702074673)*x[0]+(0.535572237843)*x[1]**o+(0.22354698199)*x[1]
        arg[2,3,1,1]=(0.0391087869133)*x[0]**o+(0.765140505702)*x[0]+(-0.66279875624)*x[1]**o+(-0.41964006149)*x[1]
        arg[2,3,2,0]=(-0.11123799186)*x[0]**o+(0.533398540082)*x[0]+(-0.931488199925)*x[1]**o+(-0.83080969281)*x[1]
        arg[2,3,2,1]=(-0.138259313941)*x[0]**o+(-0.508937684332)*x[0]+(0.583603445711)*x[1]**o+(-0.0782724872225)*x[1]
        arg[2,4,0,0]=(-0.316223936951)*x[0]**o+(0.329776289164)*x[0]+(-0.692491559115)*x[1]**o+(0.814572730294)*x[1]
        arg[2,4,0,1]=(0.74890960681)*x[0]**o+(-0.79167945549)*x[0]+(0.175328119487)*x[1]**o+(-0.253003240064)*x[1]
        arg[2,4,1,0]=(0.657827625141)*x[0]**o+(-0.268878597055)*x[0]+(0.247374286921)*x[1]**o+(0.414500041415)*x[1]
        arg[2,4,1,1]=(-0.187587046202)*x[0]**o+(-0.594983066039)*x[0]+(-0.203784406879)*x[1]**o+(0.122237566739)*x[1]
        arg[2,4,2,0]=(0.671210794607)*x[0]**o+(0.886706498415)*x[0]+(-0.368997114212)*x[1]**o+(0.0418311781187)*x[1]
        arg[2,4,2,1]=(-0.272424416171)*x[0]**o+(-0.537477247783)*x[0]+(0.0405655834868)*x[1]**o+(-0.179684833703)*x[1]
        arg[3,0,0,0]=(0.457602964193)*x[0]**o+(-0.186625111516)*x[0]+(-0.75244105343)*x[1]**o+(-0.278893445377)*x[1]
        arg[3,0,0,1]=(-0.690378721784)*x[0]**o+(0.0157608350188)*x[0]+(-0.203765434813)*x[1]**o+(0.69771096957)*x[1]
        arg[3,0,1,0]=(-0.884824688493)*x[0]**o+(0.26613451569)*x[0]+(0.226081491639)*x[1]**o+(0.597724298389)*x[1]
        arg[3,0,1,1]=(0.0372124540969)*x[0]**o+(-0.420242551812)*x[0]+(-0.625683613012)*x[1]**o+(0.279682305563)*x[1]
        arg[3,0,2,0]=(-0.570873872458)*x[0]**o+(-0.82361574923)*x[0]+(0.970204188337)*x[1]**o+(-0.242725003569)*x[1]
        arg[3,0,2,1]=(-0.714826095124)*x[0]**o+(-0.334939427182)*x[0]+(-0.730058530908)*x[1]**o+(0.287148154312)*x[1]
        arg[3,1,0,0]=(0.164234307274)*x[0]**o+(-0.388849394356)*x[0]+(0.391518896617)*x[1]**o+(0.00142750126236)*x[1]
        arg[3,1,0,1]=(-0.327700781196)*x[0]**o+(0.0442604682565)*x[0]+(-0.683834466536)*x[1]**o+(-0.665567774985)*x[1]
        arg[3,1,1,0]=(0.135979934041)*x[0]**o+(-0.822614011173)*x[0]+(-0.946061768113)*x[1]**o+(-0.585442102671)*x[1]
        arg[3,1,1,1]=(0.235568769871)*x[0]**o+(0.738233870703)*x[0]+(0.437750339594)*x[1]**o+(-0.44521724321)*x[1]
        arg[3,1,2,0]=(0.449256898185)*x[0]**o+(-0.464211755196)*x[0]+(0.616886411002)*x[1]**o+(0.885500269963)*x[1]
        arg[3,1,2,1]=(0.0851105760559)*x[0]**o+(0.888976499957)*x[0]+(0.0842441111314)*x[1]**o+(-0.366994083394)*x[1]
        arg[3,2,0,0]=(0.77120050858)*x[0]**o+(0.354835584949)*x[0]+(0.406813878866)*x[1]**o+(-0.600042478179)*x[1]
        arg[3,2,0,1]=(-0.747876879184)*x[0]**o+(0.527204926634)*x[0]+(0.134965492713)*x[1]**o+(0.497662739459)*x[1]
        arg[3,2,1,0]=(0.367550591163)*x[0]**o+(-0.700608509353)*x[0]+(0.0737696587931)*x[1]**o+(-0.164466899816)*x[1]
        arg[3,2,1,1]=(0.0908791429433)*x[0]**o+(-0.168468767791)*x[0]+(0.85978873292)*x[1]**o+(-0.785241412884)*x[1]
        arg[3,2,2,0]=(-0.296455818292)*x[0]**o+(-0.0277004420512)*x[0]+(0.409183414933)*x[1]**o+(-0.879831945543)*x[1]
        arg[3,2,2,1]=(0.971512010223)*x[0]**o+(-0.811249982262)*x[0]+(-0.846549204737)*x[1]**o+(0.689986186691)*x[1]
        arg[3,3,0,0]=(-0.800235545702)*x[0]**o+(-0.796021196346)*x[0]+(-0.0660744375539)*x[1]**o+(-0.826632192842)*x[1]
        arg[3,3,0,1]=(0.305678237361)*x[0]**o+(0.203936000462)*x[0]+(-0.583171879217)*x[1]**o+(0.924112859119)*x[1]
        arg[3,3,1,0]=(-0.937148913762)*x[0]**o+(-0.172547991203)*x[0]+(0.928491154238)*x[1]**o+(-0.502876301351)*x[1]
        arg[3,3,1,1]=(0.731413856616)*x[0]**o+(0.837213097155)*x[0]+(-0.650032764358)*x[1]**o+(0.96973961589)*x[1]
        arg[3,3,2,0]=(-0.501994458598)*x[0]**o+(-0.705938508685)*x[0]+(0.664394117366)*x[1]**o+(0.692950029135)*x[1]
        arg[3,3,2,1]=(-0.0948399323409)*x[0]**o+(-0.686744554674)*x[0]+(0.577765681583)*x[1]**o+(0.969310750204)*x[1]
        arg[3,4,0,0]=(-0.484011788432)*x[0]**o+(-0.560744139846)*x[0]+(0.072097608773)*x[1]**o+(-0.103052228904)*x[1]
        arg[3,4,0,1]=(-0.201158934695)*x[0]**o+(0.463105335811)*x[0]+(0.733365706682)*x[1]**o+(-0.330035241794)*x[1]
        arg[3,4,1,0]=(0.252379277102)*x[0]**o+(-0.110596657969)*x[0]+(0.763687257878)*x[1]**o+(-0.345598371263)*x[1]
        arg[3,4,1,1]=(0.0868412855617)*x[0]**o+(0.385098679406)*x[0]+(0.291487288978)*x[1]**o+(-0.112145846209)*x[1]
        arg[3,4,2,0]=(-0.913724685255)*x[0]**o+(-0.0348735884313)*x[0]+(0.550688528861)*x[1]**o+(0.798394152329)*x[1]
        arg[3,4,2,1]=(-0.92870523219)*x[0]**o+(0.177354002469)*x[0]+(0.350393765313)*x[1]**o+(-0.974997198741)*x[1]
        ref[0,0,0,0]=(0.815821053922)/(o+1.)+(0.302879577355)
        ref[0,0,0,1]=(-0.223824520603)/(o+1.)+(0.690659426687)
        ref[0,0,1,0]=(-0.501731412433)/(o+1.)+(0.0102964270565)
        ref[0,0,1,1]=(-0.731540589167)/(o+1.)+(0.348284469462)
        ref[0,0,2,0]=(-0.121523684012)/(o+1.)+(0.184348962759)
        ref[0,0,2,1]=(-0.4529485642)/(o+1.)+(-0.830798233026)
        ref[0,1,0,0]=(-0.244463483906)/(o+1.)+(0.226296188971)
        ref[0,1,0,1]=(1.2203272116)/(o+1.)+(-0.693494218115)
        ref[0,1,1,0]=(-0.55576623697)/(o+1.)+(-0.273623989583)
        ref[0,1,1,1]=(-0.316836585171)/(o+1.)+(-0.836135387777)
        ref[0,1,2,0]=(-0.133732244387)/(o+1.)+(-0.875415201605)
        ref[0,1,2,1]=(-0.281487947177)/(o+1.)+(-0.668499911274)
        ref[0,2,0,0]=(-0.896781862729)/(o+1.)+(-0.0225848820142)
        ref[0,2,0,1]=(-1.70326692867)/(o+1.)+(-0.1693379432)
        ref[0,2,1,0]=(-0.00084798161079)/(o+1.)+(0.343359275275)
        ref[0,2,1,1]=(0.609064372767)/(o+1.)+(-0.369639962727)
        ref[0,2,2,0]=(-0.759042888708)/(o+1.)+(0.321708104548)
        ref[0,2,2,1]=(0.760967724015)/(o+1.)+(0.0268897973234)
        ref[0,3,0,0]=(-1.3915636468)/(o+1.)+(-0.230755462281)
        ref[0,3,0,1]=(-0.416449082748)/(o+1.)+(0.480436529329)
        ref[0,3,1,0]=(-0.238553658798)/(o+1.)+(-0.029283546006)
        ref[0,3,1,1]=(-1.36933302275)/(o+1.)+(-0.221757606321)
        ref[0,3,2,0]=(-0.872999155991)/(o+1.)+(0.69296766968)
        ref[0,3,2,1]=(0.833661507716)/(o+1.)+(0.321357009466)
        ref[0,4,0,0]=(1.4358026373)/(o+1.)+(0.311715942547)
        ref[0,4,0,1]=(-0.00811900044778)/(o+1.)+(-0.76710974695)
        ref[0,4,1,0]=(-0.92034101297)/(o+1.)+(-0.0309623888275)
        ref[0,4,1,1]=(0.690460334263)/(o+1.)+(0.783809038221)
        ref[0,4,2,0]=(0.779851986824)/(o+1.)+(-0.214799981386)
        ref[0,4,2,1]=(-0.566157259598)/(o+1.)+(0.296555712092)
        ref[1,0,0,0]=(-1.70267943485)/(o+1.)+(0.0444622873215)
        ref[1,0,0,1]=(0.399342875087)/(o+1.)+(0.0341080904363)
        ref[1,0,1,0]=(0.839377637409)/(o+1.)+(0.158619241351)
        ref[1,0,1,1]=(-0.0142823080176)/(o+1.)+(0.693714763021)
        ref[1,0,2,0]=(-0.97036633864)/(o+1.)+(0.264110057567)
        ref[1,0,2,1]=(0.104223436728)/(o+1.)+(0.328785142295)
        ref[1,1,0,0]=(1.09779156365)/(o+1.)+(-0.00408578628855)
        ref[1,1,0,1]=(-0.914750679597)/(o+1.)+(0.206573942886)
        ref[1,1,1,0]=(0.150419426085)/(o+1.)+(-0.0846024372156)
        ref[1,1,1,1]=(-1.04733819847)/(o+1.)+(-0.157756242641)
        ref[1,1,2,0]=(-1.49193735459)/(o+1.)+(0.391723285192)
        ref[1,1,2,1]=(-0.250600389657)/(o+1.)+(-0.456841431612)
        ref[1,2,0,0]=(1.21154796884)/(o+1.)+(-0.377275533205)
        ref[1,2,0,1]=(-0.785475967917)/(o+1.)+(0.2413362399)
        ref[1,2,1,0]=(1.20750352255)/(o+1.)+(-0.409504204121)
        ref[1,2,1,1]=(0.271471015188)/(o+1.)+(-0.136640402636)
        ref[1,2,2,0]=(-0.741344603747)/(o+1.)+(0.00668995653618)
        ref[1,2,2,1]=(0.462000684606)/(o+1.)+(0.364894712961)
        ref[1,3,0,0]=(0.0273564369347)/(o+1.)+(0.673962086152)
        ref[1,3,0,1]=(0.285276349085)/(o+1.)+(0.385675061073)
        ref[1,3,1,0]=(0.851307298344)/(o+1.)+(0.118422744971)
        ref[1,3,1,1]=(0.235593686643)/(o+1.)+(0.244705242716)
        ref[1,3,2,0]=(-0.359345731222)/(o+1.)+(-0.124958138388)
        ref[1,3,2,1]=(0.590170220995)/(o+1.)+(0.429975100757)
        ref[1,4,0,0]=(0.987134792715)/(o+1.)+(-0.0725353868253)
        ref[1,4,0,1]=(0.623734296247)/(o+1.)+(0.489855522024)
        ref[1,4,1,0]=(-0.913537095843)/(o+1.)+(0.727739353556)
        ref[1,4,1,1]=(-0.656988917238)/(o+1.)+(0.274753842895)
        ref[1,4,2,0]=(1.03400922475)/(o+1.)+(-0.00621482335664)
        ref[1,4,2,1]=(-1.0367802184)/(o+1.)+(-0.266633530803)
        ref[2,0,0,0]=(-0.391593893803)/(o+1.)+(0.457472062333)
        ref[2,0,0,1]=(0.393476325223)/(o+1.)+(-0.911233333506)
        ref[2,0,1,0]=(-0.0199296737245)/(o+1.)+(0.328542898457)
        ref[2,0,1,1]=(-0.618353771051)/(o+1.)+(0.0279228750664)
        ref[2,0,2,0]=(0.570494475025)/(o+1.)+(-0.374533325635)
        ref[2,0,2,1]=(-1.21079806839)/(o+1.)+(-0.632909492494)
        ref[2,1,0,0]=(1.32715885433)/(o+1.)+(0.730744204126)
        ref[2,1,0,1]=(1.43666536922)/(o+1.)+(0.279106975714)
        ref[2,1,1,0]=(-0.603800395047)/(o+1.)+(-0.716035662434)
        ref[2,1,1,1]=(1.35897125219)/(o+1.)+(-0.721199872542)
        ref[2,1,2,0]=(0.430431838988)/(o+1.)+(0.465280445622)
        ref[2,1,2,1]=(-1.060398141)/(o+1.)+(-0.404416040892)
        ref[2,2,0,0]=(0.374659415961)/(o+1.)+(0.202262379386)
        ref[2,2,0,1]=(0.801918445202)/(o+1.)+(-0.403667477049)
        ref[2,2,1,0]=(-1.50789527973)/(o+1.)+(0.540295891905)
        ref[2,2,1,1]=(-1.19243886136)/(o+1.)+(-0.478654318143)
        ref[2,2,2,0]=(0.21727146958)/(o+1.)+(-0.302043553627)
        ref[2,2,2,1]=(1.35667456836)/(o+1.)+(-0.530838305162)
        ref[2,3,0,0]=(0.0649985054133)/(o+1.)+(0.596842347772)
        ref[2,3,0,1]=(0.166245904835)/(o+1.)+(-0.594924886997)
        ref[2,3,1,0]=(1.01422987189)/(o+1.)+(-0.10673688237)
        ref[2,3,1,1]=(-0.623689969327)/(o+1.)+(0.172750222106)
        ref[2,3,2,0]=(-1.04272619178)/(o+1.)+(-0.148705576364)
        ref[2,3,2,1]=(0.445344131771)/(o+1.)+(-0.293605085777)
        ref[2,4,0,0]=(-1.00871549607)/(o+1.)+(0.572174509729)
        ref[2,4,0,1]=(0.924237726297)/(o+1.)+(-0.522341347777)
        ref[2,4,1,0]=(0.905201912063)/(o+1.)+(0.0728107221801)
        ref[2,4,1,1]=(-0.391371453081)/(o+1.)+(-0.23637274965)
        ref[2,4,2,0]=(0.302213680395)/(o+1.)+(0.464268838267)
        ref[2,4,2,1]=(-0.231858832685)/(o+1.)+(-0.358581040743)
        ref[3,0,0,0]=(-0.294838089237)/(o+1.)+(-0.232759278447)
        ref[3,0,0,1]=(-0.894144156597)/(o+1.)+(0.356735902294)
        ref[3,0,1,0]=(-0.658743196854)/(o+1.)+(0.43192940704)
        ref[3,0,1,1]=(-0.588471158915)/(o+1.)+(-0.0702801231243)
        ref[3,0,2,0]=(0.399330315879)/(o+1.)+(-0.5331703764)
        ref[3,0,2,1]=(-1.44488462603)/(o+1.)+(-0.0238956364351)
        ref[3,1,0,0]=(0.555753203891)/(o+1.)+(-0.193710946547)
        ref[3,1,0,1]=(-1.01153524773)/(o+1.)+(-0.310653653364)
        ref[3,1,1,0]=(-0.810081834072)/(o+1.)+(-0.704028056922)
        ref[3,1,1,1]=(0.673319109465)/(o+1.)+(0.146508313746)
        ref[3,1,2,0]=(1.06614330919)/(o+1.)+(0.210644257384)
        ref[3,1,2,1]=(0.169354687187)/(o+1.)+(0.260991208281)
        ref[3,2,0,0]=(1.17801438745)/(o+1.)+(-0.122603446615)
        ref[3,2,0,1]=(-0.612911386471)/(o+1.)+(0.512433833047)
        ref[3,2,1,0]=(0.441320249956)/(o+1.)+(-0.432537704584)
        ref[3,2,1,1]=(0.950667875863)/(o+1.)+(-0.476855090338)
        ref[3,2,2,0]=(0.112727596641)/(o+1.)+(-0.453766193797)
        ref[3,2,2,1]=(0.124962805486)/(o+1.)+(-0.0606318977858)
        ref[3,3,0,0]=(-0.866309983256)/(o+1.)+(-0.811326694594)
        ref[3,3,0,1]=(-0.277493641857)/(o+1.)+(0.564024429791)
        ref[3,3,1,0]=(-0.00865775952435)/(o+1.)+(-0.337712146277)
        ref[3,3,1,1]=(0.0813810922576)/(o+1.)+(0.903476356523)
        ref[3,3,2,0]=(0.162399658768)/(o+1.)+(-0.00649423977499)
        ref[3,3,2,1]=(0.482925749243)/(o+1.)+(0.141283097765)
        ref[3,4,0,0]=(-0.411914179659)/(o+1.)+(-0.331898184375)
        ref[3,4,0,1]=(0.532206771987)/(o+1.)+(0.0665350470083)
        ref[3,4,1,0]=(1.01606653498)/(o+1.)+(-0.228097514616)
        ref[3,4,1,1]=(0.37832857454)/(o+1.)+(0.136476416599)
        ref[3,4,2,0]=(-0.363036156393)/(o+1.)+(0.381760281949)
        ref[3,4,2,1]=(-0.578311466877)/(o+1.)+(-0.398821598136)
      else:
        arg[0,0,0,0]=(0.860676142569)*x[0]**o+(-0.156046991162)*x[0]+(-0.0462355579656)*x[1]**o+(0.70842844961)*x[1]+(0.617299123548)*x[2]**o+(-0.865174329565)*x[2]
        arg[0,0,0,1]=(-0.284308060865)*x[0]**o+(0.788848579389)*x[0]+(0.629321669042)*x[1]**o+(0.066637849809)*x[1]+(0.114152381009)*x[2]**o+(-0.466687745846)*x[2]
        arg[0,0,1,0]=(-0.690823198517)*x[0]**o+(-0.141510804573)*x[0]+(0.969978831967)*x[1]**o+(0.307926382658)*x[1]+(0.710373009094)*x[2]**o+(-0.494828315907)*x[2]
        arg[0,0,1,1]=(0.953361413913)*x[0]**o+(-0.103742611936)*x[0]+(0.63934955554)*x[1]**o+(-0.0778114476564)*x[1]+(-0.221185122698)*x[2]**o+(-0.287432047404)*x[2]
        arg[0,0,2,0]=(0.378306380558)*x[0]**o+(-0.117743922672)*x[0]+(-0.0606376630208)*x[1]**o+(-0.59439189505)*x[1]+(0.10507320657)*x[2]**o+(0.856411423067)*x[2]
        arg[0,0,2,1]=(0.777734672779)*x[0]**o+(0.286956002007)*x[0]+(-0.931811626583)*x[1]**o+(-0.806626659661)*x[1]+(-0.420390949282)*x[2]**o+(0.854080006578)*x[2]
        arg[0,1,0,0]=(0.403426948964)*x[0]**o+(0.707099419296)*x[0]+(0.593137130456)*x[1]**o+(-0.757560771598)*x[1]+(0.369632217934)*x[2]**o+(-0.348480285397)*x[2]
        arg[0,1,0,1]=(-0.491454189782)*x[0]**o+(0.547276640565)*x[0]+(-0.955045542708)*x[1]**o+(0.529989872531)*x[1]+(-0.161086874799)*x[2]**o+(-0.181810040724)*x[2]
        arg[0,1,1,0]=(0.46282177104)*x[0]**o+(0.254346937251)*x[0]+(0.824797105546)*x[1]**o+(-0.449979162246)*x[1]+(0.22307703071)*x[2]**o+(-0.592072135668)*x[2]
        arg[0,1,1,1]=(0.329230675141)*x[0]**o+(-0.733034697145)*x[0]+(0.762195609347)*x[1]**o+(-0.639097591362)*x[1]+(0.765354927905)*x[2]**o+(0.313571493405)*x[2]
        arg[0,1,2,0]=(0.3874127072)*x[0]**o+(-0.0569429727569)*x[0]+(-0.203121198708)*x[1]**o+(-0.736379499164)*x[1]+(-0.980629375436)*x[2]**o+(0.955793660048)*x[2]
        arg[0,1,2,1]=(-0.656255469455)*x[0]**o+(0.711904808699)*x[0]+(-0.363253678399)*x[1]**o+(0.921875000807)*x[1]+(0.686394816431)*x[2]**o+(0.818500716704)*x[2]
        arg[0,2,0,0]=(-0.931555347846)*x[0]**o+(-0.249573400742)*x[0]+(0.617166716404)*x[1]**o+(-0.929587051108)*x[1]+(0.830928715064)*x[2]**o+(0.461637431613)*x[2]
        arg[0,2,0,1]=(-0.8677719945)*x[0]**o+(-0.176892180445)*x[0]+(-0.881887456464)*x[1]**o+(0.923355677513)*x[1]+(-0.978544668453)*x[2]**o+(-0.427009384287)*x[2]
        arg[0,2,1,0]=(-0.0707937433084)*x[0]**o+(-0.872892141393)*x[0]+(0.606234798381)*x[1]**o+(0.935845076847)*x[1]+(-0.97559052201)*x[2]**o+(-0.31138518209)*x[2]
        arg[0,2,1,1]=(0.450210597865)*x[0]**o+(0.842939440378)*x[0]+(-0.333422856577)*x[1]**o+(-0.492809679443)*x[1]+(-0.254728278043)*x[2]**o+(-0.0702767077952)*x[2]
        arg[0,2,2,0]=(-0.250463216903)*x[0]**o+(0.756576293763)*x[0]+(0.596902629953)*x[1]**o+(-0.887080794947)*x[1]+(0.027319948008)*x[2]**o+(0.552496782776)*x[2]
        arg[0,2,2,1]=(-0.0883102140591)*x[0]**o+(0.988759610101)*x[0]+(-0.208152392508)*x[1]**o+(0.938999108564)*x[1]+(-0.773883611417)*x[2]**o+(0.653413306338)*x[2]
        arg[0,3,0,0]=(-0.399644211606)*x[0]**o+(-0.978595776407)*x[0]+(0.99454650986)*x[1]**o+(0.143313668715)*x[1]+(0.328592750842)*x[2]**o+(-0.958804581101)*x[2]
        arg[0,3,0,1]=(-0.0667966885959)*x[0]**o+(0.332653574704)*x[0]+(-0.762025150916)*x[1]**o+(0.895044013374)*x[1]+(0.627013843362)*x[2]**o+(-0.682392672232)*x[2]
        arg[0,3,1,0]=(-0.834479453773)*x[0]**o+(-0.375930859033)*x[0]+(0.129507966861)*x[1]**o+(0.995364280408)*x[1]+(-0.126081191572)*x[2]**o+(0.0344262779491)*x[2]
        arg[0,3,1,1]=(0.796952126596)*x[0]**o+(0.310562243091)*x[0]+(0.854548623788)*x[1]**o+(-0.719100537284)*x[1]+(-0.236266783438)*x[2]**o+(-0.646647129209)*x[2]
        arg[0,3,2,0]=(0.844073289115)*x[0]**o+(0.423300345347)*x[0]+(0.126354796852)*x[1]**o+(0.180372257606)*x[1]+(-0.564712722146)*x[2]**o+(0.466988430183)*x[2]
        arg[0,3,2,1]=(0.228775824367)*x[0]**o+(0.40701094831)*x[0]+(0.223101054593)*x[1]**o+(-0.894937905383)*x[1]+(0.425392942773)*x[2]**o+(-0.844238667169)*x[2]
        arg[0,4,0,0]=(0.3363848738)*x[0]**o+(-0.985327066062)*x[0]+(-0.107847971678)*x[1]**o+(0.596121801488)*x[1]+(0.655589801663)*x[2]**o+(-0.729650731949)*x[2]
        arg[0,4,0,1]=(-0.176435679219)*x[0]**o+(0.382558711725)*x[0]+(-0.658212492337)*x[1]**o+(0.584904267612)*x[1]+(-0.534684876705)*x[2]**o+(-0.441262614915)*x[2]
        arg[0,4,1,0]=(0.281301737167)*x[0]**o+(-0.642486274791)*x[0]+(0.420753884869)*x[1]**o+(0.482835955591)*x[1]+(0.169431333788)*x[2]**o+(0.922841193825)*x[2]
        arg[0,4,1,1]=(0.235812590604)*x[0]**o+(-0.446233855396)*x[0]+(0.465015868566)*x[1]**o+(-0.780441276888)*x[1]+(0.513632761351)*x[2]**o+(-0.0576482811678)*x[2]
        arg[0,4,2,0]=(-0.617245874462)*x[0]**o+(0.183661473466)*x[0]+(-0.711148911239)*x[1]**o+(0.125228991306)*x[1]+(-0.500218603959)*x[2]**o+(0.70597577229)*x[2]
        arg[0,4,2,1]=(0.56330304117)*x[0]**o+(0.631284986628)*x[0]+(-0.222485175776)*x[1]**o+(0.0639235737228)*x[1]+(-0.276613868747)*x[2]**o+(0.407488564857)*x[2]
        arg[1,0,0,0]=(0.159032146915)*x[0]**o+(-0.716040989693)*x[0]+(0.410709172613)*x[1]**o+(-0.527667963652)*x[1]+(-0.160515156497)*x[2]**o+(0.294448742382)*x[2]
        arg[1,0,0,1]=(0.941139035835)*x[0]**o+(0.707410046094)*x[0]+(-0.172095410803)*x[1]**o+(-0.967035204304)*x[1]+(0.870395272646)*x[2]**o+(0.830870986802)*x[2]
        arg[1,0,1,0]=(0.731015285309)*x[0]**o+(0.450519048697)*x[0]+(-0.484568714675)*x[1]**o+(0.454201220108)*x[1]+(-0.161850459099)*x[2]**o+(-0.617463263396)*x[2]
        arg[1,0,1,1]=(-0.958841628847)*x[0]**o+(-0.0667813400021)*x[0]+(0.659551780082)*x[1]**o+(0.514424015304)*x[1]+(0.920583608125)*x[2]**o+(0.89564899572)*x[2]
        arg[1,0,2,0]=(0.0482630479817)*x[0]**o+(0.911326064322)*x[0]+(-0.275134422092)*x[1]**o+(0.216495232459)*x[1]+(-0.973586171195)*x[2]**o+(0.275694688611)*x[2]
        arg[1,0,2,1]=(0.769733688293)*x[0]**o+(0.771829018421)*x[0]+(-0.22434549079)*x[1]**o+(0.131966885817)*x[1]+(-0.109424705679)*x[2]**o+(-0.369585136379)*x[2]
        arg[1,1,0,0]=(-0.0160180905877)*x[0]**o+(-0.268260755267)*x[0]+(-0.213940412022)*x[1]**o+(0.665987805279)*x[1]+(0.968879828017)*x[2]**o+(-0.449435538913)*x[2]
        arg[1,1,0,1]=(0.361055351825)*x[0]**o+(0.463294478488)*x[0]+(0.692655668486)*x[1]**o+(0.151415573206)*x[1]+(0.902280461515)*x[2]**o+(0.690926934073)*x[2]
        arg[1,1,1,0]=(0.591998602299)*x[0]**o+(0.516482177709)*x[0]+(-0.298511909182)*x[1]**o+(0.281818071894)*x[1]+(0.207089953979)*x[2]**o+(0.917004722388)*x[2]
        arg[1,1,1,1]=(0.479832408236)*x[0]**o+(-0.643374914165)*x[0]+(0.370844835249)*x[1]**o+(0.560458499239)*x[1]+(-0.865852319767)*x[2]**o+(0.560572130123)*x[2]
        arg[1,1,2,0]=(0.709293526635)*x[0]**o+(-0.388347223155)*x[0]+(0.60797330273)*x[1]**o+(-0.847624783099)*x[1]+(-0.0476367756382)*x[2]**o+(0.637841988478)*x[2]
        arg[1,1,2,1]=(0.343458045422)*x[0]**o+(-0.764451449587)*x[0]+(-0.444256801761)*x[1]**o+(0.385385997921)*x[1]+(-0.722095136671)*x[2]**o+(0.677852473207)*x[2]
        arg[1,2,0,0]=(0.710372887392)*x[0]**o+(-0.628494679119)*x[0]+(-0.59513507905)*x[1]**o+(-0.527824874295)*x[1]+(-0.611189170789)*x[2]**o+(0.090651976806)*x[2]
        arg[1,2,0,1]=(-0.980160150289)*x[0]**o+(-0.0286097581598)*x[0]+(-0.407906975953)*x[1]**o+(-0.319422835951)*x[1]+(-0.489787310147)*x[2]**o+(-0.694386341354)*x[2]
        arg[1,2,1,0]=(-0.416867224376)*x[0]**o+(0.631406877489)*x[0]+(-0.93939195942)*x[1]**o+(0.266858600313)*x[1]+(0.982173221612)*x[2]**o+(0.112901364617)*x[2]
        arg[1,2,1,1]=(0.412001480348)*x[0]**o+(0.624281489887)*x[0]+(0.306558457751)*x[1]**o+(0.95864454419)*x[1]+(0.299175627954)*x[2]**o+(-0.0531991017347)*x[2]
        arg[1,2,2,0]=(-0.0342261621741)*x[0]**o+(-0.587245704486)*x[0]+(-0.976407464107)*x[1]**o+(-0.0157215030322)*x[1]+(-0.529857472095)*x[2]**o+(-0.419626697673)*x[2]
        arg[1,2,2,1]=(-0.681737301037)*x[0]**o+(-0.980349460702)*x[0]+(-0.839469631393)*x[1]**o+(-0.412417814907)*x[1]+(0.451717398947)*x[2]**o+(0.962167685987)*x[2]
        arg[1,3,0,0]=(0.283453965913)*x[0]**o+(0.204445403787)*x[0]+(-0.806150376236)*x[1]**o+(0.122187094069)*x[1]+(-0.679291367964)*x[2]**o+(-0.875353966489)*x[2]
        arg[1,3,0,1]=(0.115974475175)*x[0]**o+(0.701473497654)*x[0]+(0.298012120304)*x[1]**o+(0.567597901709)*x[1]+(0.721211831476)*x[2]**o+(-0.255389572775)*x[2]
        arg[1,3,1,0]=(0.393891431092)*x[0]**o+(-0.563382537925)*x[0]+(0.974194829106)*x[1]**o+(0.659688559688)*x[1]+(-0.394266143828)*x[2]**o+(0.875266200191)*x[2]
        arg[1,3,1,1]=(-0.456930975697)*x[0]**o+(0.334590626231)*x[0]+(-0.227176883362)*x[1]**o+(0.76775699173)*x[1]+(0.839371167887)*x[2]**o+(-0.155473916326)*x[2]
        arg[1,3,2,0]=(0.346761981787)*x[0]**o+(-0.625121706322)*x[0]+(0.340655154638)*x[1]**o+(-0.0646323546999)*x[1]+(-0.211347402677)*x[2]**o+(0.233661266919)*x[2]
        arg[1,3,2,1]=(-0.086096644956)*x[0]**o+(-0.845623653209)*x[0]+(-0.0866093549957)*x[1]**o+(-0.467286349011)*x[1]+(-0.0621455515357)*x[2]**o+(0.164779979085)*x[2]
        arg[1,4,0,0]=(0.42022138208)*x[0]**o+(0.136984451355)*x[0]+(0.664770979086)*x[1]**o+(-0.537368896193)*x[1]+(0.874256726691)*x[2]**o+(-0.58614100824)*x[2]
        arg[1,4,0,1]=(-0.805860964792)*x[0]**o+(0.758511304044)*x[0]+(-0.574686863541)*x[1]**o+(-0.410873101247)*x[1]+(0.932015171551)*x[2]**o+(0.439987195351)*x[2]
        arg[1,4,1,0]=(-0.515187049123)*x[0]**o+(0.308078404334)*x[0]+(0.940959771774)*x[1]**o+(0.466269579691)*x[1]+(-0.190617144843)*x[2]**o+(0.738172533582)*x[2]
        arg[1,4,1,1]=(0.283326036586)*x[0]**o+(-0.693631747857)*x[0]+(0.478965080319)*x[1]**o+(-0.0465389396775)*x[1]+(-0.72886286959)*x[2]**o+(-0.9346424149)*x[2]
        arg[1,4,2,0]=(-0.71539766285)*x[0]**o+(0.361875987071)*x[0]+(0.905270678876)*x[1]**o+(-0.675024725702)*x[1]+(0.378259473658)*x[2]**o+(0.0520394226711)*x[2]
        arg[1,4,2,1]=(-0.319600525923)*x[0]**o+(0.0449956327134)*x[0]+(0.707418658521)*x[1]**o+(0.210137972776)*x[1]+(-0.767860728688)*x[2]**o+(0.232340825758)*x[2]
        arg[2,0,0,0]=(0.92844243651)*x[0]**o+(0.956810319504)*x[0]+(0.762198571194)*x[1]**o+(-0.363107634629)*x[1]+(0.0750393505254)*x[2]**o+(-0.43573581041)*x[2]
        arg[2,0,0,1]=(-0.193613121742)*x[0]**o+(0.440027871839)*x[0]+(-0.607905506346)*x[1]**o+(0.894840354958)*x[1]+(0.601279436178)*x[2]**o+(0.46806560762)*x[2]
        arg[2,0,1,0]=(0.370901517567)*x[0]**o+(-0.738769905647)*x[0]+(-0.698692979664)*x[1]**o+(-0.286318840335)*x[1]+(-0.879367315688)*x[2]**o+(-0.959912132991)*x[2]
        arg[2,0,1,1]=(-0.278939533774)*x[0]**o+(0.183795552426)*x[0]+(-0.341943551466)*x[1]**o+(0.491847260291)*x[1]+(-0.141030429384)*x[2]**o+(0.256287855539)*x[2]
        arg[2,0,2,0]=(-0.135930357616)*x[0]**o+(-0.382515749949)*x[0]+(-0.43030873581)*x[1]**o+(-0.0211891159981)*x[1]+(-0.377865120002)*x[2]**o+(0.109942325029)*x[2]
        arg[2,0,2,1]=(0.288827267201)*x[0]**o+(-0.907970544468)*x[0]+(-0.622258309213)*x[1]**o+(-0.763273879894)*x[1]+(0.438715013671)*x[2]**o+(-0.93873140988)*x[2]
        arg[2,1,0,0]=(-0.963487369906)*x[0]**o+(0.899051290288)*x[0]+(-0.268257956548)*x[1]**o+(-0.229092726584)*x[1]+(0.427024112314)*x[2]**o+(0.0435138095138)*x[2]
        arg[2,1,0,1]=(0.89996926597)*x[0]**o+(-0.206500706229)*x[0]+(-0.191769123095)*x[1]**o+(-0.348924155298)*x[1]+(-0.74405653776)*x[2]**o+(0.0724946691864)*x[2]
        arg[2,1,1,0]=(0.338170397901)*x[0]**o+(-0.725761081242)*x[0]+(-0.45928272676)*x[1]**o+(0.795535326783)*x[1]+(-0.316443844644)*x[2]**o+(-0.844205401892)*x[2]
        arg[2,1,1,1]=(0.509131757681)*x[0]**o+(0.319175801535)*x[0]+(-0.807058926279)*x[1]**o+(0.63928269996)*x[1]+(0.219324073432)*x[2]**o+(0.351217517434)*x[2]
        arg[2,1,2,0]=(-0.534090721611)*x[0]**o+(0.27146424294)*x[0]+(0.250778065716)*x[1]**o+(-0.901322994167)*x[1]+(-0.57588768699)*x[2]**o+(0.405043003456)*x[2]
        arg[2,1,2,1]=(0.101699906269)*x[0]**o+(0.361960318738)*x[0]+(0.278974923593)*x[1]**o+(-0.336289352037)*x[1]+(-0.677353936215)*x[2]**o+(0.0726793347753)*x[2]
        arg[2,2,0,0]=(-0.734961283823)*x[0]**o+(-0.593268557578)*x[0]+(0.310896858565)*x[1]**o+(-0.482326137951)*x[1]+(0.886472957099)*x[2]**o+(-0.0403563476581)*x[2]
        arg[2,2,0,1]=(-0.166071725106)*x[0]**o+(-0.454245686818)*x[0]+(-0.892674508991)*x[1]**o+(0.581554190251)*x[1]+(0.995323654908)*x[2]**o+(-0.0336714433349)*x[2]
        arg[2,2,1,0]=(-0.437491848099)*x[0]**o+(-0.195039508213)*x[0]+(0.822011086215)*x[1]**o+(0.771044412863)*x[1]+(0.697534261252)*x[2]**o+(0.666510217725)*x[2]
        arg[2,2,1,1]=(-0.529148276275)*x[0]**o+(-0.093016460024)*x[0]+(0.663641010649)*x[1]**o+(0.940439359843)*x[1]+(0.754151159716)*x[2]**o+(0.852349310591)*x[2]
        arg[2,2,2,0]=(0.859523244682)*x[0]**o+(0.704825125128)*x[0]+(0.221832129598)*x[1]**o+(0.573940020456)*x[1]+(0.935673751591)*x[2]**o+(0.23235170475)*x[2]
        arg[2,2,2,1]=(0.0359047473512)*x[0]**o+(0.836816935443)*x[0]+(0.6411608273)*x[1]**o+(0.291526953908)*x[1]+(0.735564515362)*x[2]**o+(-0.465781621769)*x[2]
        arg[2,3,0,0]=(0.352740040216)*x[0]**o+(0.54224020581)*x[0]+(0.681356174115)*x[1]**o+(-0.97589117869)*x[1]+(0.880755719884)*x[2]**o+(0.562259403362)*x[2]
        arg[2,3,0,1]=(0.563708471054)*x[0]**o+(0.525523219723)*x[0]+(0.423068360374)*x[1]**o+(0.29239160821)*x[1]+(0.874360368885)*x[2]**o+(0.922265851359)*x[2]
        arg[2,3,1,0]=(0.694309936305)*x[0]**o+(-0.409224769672)*x[0]+(-0.748835864803)*x[1]**o+(-0.21841534641)*x[1]+(0.497059735624)*x[2]**o+(-0.753234694148)*x[2]
        arg[2,3,1,1]=(0.109706510352)*x[0]**o+(-0.0160720403678)*x[0]+(0.888778080744)*x[1]**o+(-0.938692716268)*x[1]+(0.529262951478)*x[2]**o+(0.481430680214)*x[2]
        arg[2,3,2,0]=(0.65368659132)*x[0]**o+(-0.988176396392)*x[0]+(0.989546739333)*x[1]**o+(0.23178715846)*x[1]+(-0.557645438328)*x[2]**o+(0.063228513147)*x[2]
        arg[2,3,2,1]=(0.286316141831)*x[0]**o+(0.668138463424)*x[0]+(-0.950430671533)*x[1]**o+(-0.978012107187)*x[1]+(-0.107618382464)*x[2]**o+(-0.684084421276)*x[2]
        arg[2,4,0,0]=(-0.63510659432)*x[0]**o+(0.264242935995)*x[0]+(0.960647241624)*x[1]**o+(0.58483379289)*x[1]+(0.7255270271)*x[2]**o+(-0.476151653985)*x[2]
        arg[2,4,0,1]=(-0.221953908705)*x[0]**o+(-0.994143786509)*x[0]+(-0.359525421944)*x[1]**o+(0.533966197668)*x[1]+(-0.29739666047)*x[2]**o+(-0.0824204124466)*x[2]
        arg[2,4,1,0]=(-0.447043625871)*x[0]**o+(-0.904485300564)*x[0]+(0.26179071803)*x[1]**o+(0.437320719437)*x[1]+(-0.56501466909)*x[2]**o+(0.129662379166)*x[2]
        arg[2,4,1,1]=(0.259809756366)*x[0]**o+(0.446788808701)*x[0]+(-0.143702396023)*x[1]**o+(0.879531090832)*x[1]+(0.410047703156)*x[2]**o+(0.755750006468)*x[2]
        arg[2,4,2,0]=(-0.730881615951)*x[0]**o+(-0.357110792705)*x[0]+(0.552722654621)*x[1]**o+(0.965623848242)*x[1]+(0.555951256049)*x[2]**o+(0.259021370764)*x[2]
        arg[2,4,2,1]=(-0.465880467377)*x[0]**o+(-0.840984822217)*x[0]+(0.665817382039)*x[1]**o+(-0.390932961529)*x[1]+(-0.0857939888315)*x[2]**o+(-0.712185917672)*x[2]
        arg[3,0,0,0]=(0.710281429223)*x[0]**o+(-0.723903639057)*x[0]+(-0.267455509065)*x[1]**o+(-0.731684188612)*x[1]+(-0.287442751768)*x[2]**o+(-0.266871749947)*x[2]
        arg[3,0,0,1]=(0.460723870663)*x[0]**o+(0.0502154310124)*x[0]+(-0.260241363193)*x[1]**o+(-0.516473088648)*x[1]+(-0.925984113039)*x[2]**o+(-0.0297616649018)*x[2]
        arg[3,0,1,0]=(0.523186648278)*x[0]**o+(0.937042293367)*x[0]+(0.811740916657)*x[1]**o+(0.0544805175015)*x[1]+(-0.543484279955)*x[2]**o+(0.244714198306)*x[2]
        arg[3,0,1,1]=(-0.69212630167)*x[0]**o+(0.868708276385)*x[0]+(0.473290025811)*x[1]**o+(0.137043602294)*x[1]+(-0.19291314113)*x[2]**o+(-0.755147036188)*x[2]
        arg[3,0,2,0]=(0.0472638421122)*x[0]**o+(0.403135551064)*x[0]+(0.290083188289)*x[1]**o+(0.96419060449)*x[1]+(-0.240544184077)*x[2]**o+(-0.183028108529)*x[2]
        arg[3,0,2,1]=(0.354253383908)*x[0]**o+(0.628816528774)*x[0]+(0.588194429756)*x[1]**o+(-0.852128734457)*x[1]+(-0.84774778925)*x[2]**o+(0.796310123562)*x[2]
        arg[3,1,0,0]=(0.528880495806)*x[0]**o+(-0.74176589205)*x[0]+(-0.447816938371)*x[1]**o+(0.611297564802)*x[1]+(0.742134198482)*x[2]**o+(-0.625840664202)*x[2]
        arg[3,1,0,1]=(0.419633746259)*x[0]**o+(-0.247742274578)*x[0]+(0.302328885303)*x[1]**o+(-0.629854480502)*x[1]+(0.493919301288)*x[2]**o+(0.902910119017)*x[2]
        arg[3,1,1,0]=(0.509160481656)*x[0]**o+(-0.388129364133)*x[0]+(-0.595185230598)*x[1]**o+(-0.289592836865)*x[1]+(0.615663908134)*x[2]**o+(-0.560687711511)*x[2]
        arg[3,1,1,1]=(0.647782473568)*x[0]**o+(0.935840881725)*x[0]+(0.440348212287)*x[1]**o+(-0.489797988572)*x[1]+(0.766592631787)*x[2]**o+(0.217399649246)*x[2]
        arg[3,1,2,0]=(0.64944026906)*x[0]**o+(0.487409918475)*x[0]+(-0.0322210593784)*x[1]**o+(0.281311337861)*x[1]+(-0.694252407571)*x[2]**o+(-0.659678386121)*x[2]
        arg[3,1,2,1]=(0.806357278285)*x[0]**o+(0.419430374567)*x[0]+(-0.0525796600116)*x[1]**o+(-0.556886341914)*x[1]+(-0.288209814683)*x[2]**o+(-0.309696771495)*x[2]
        arg[3,2,0,0]=(-0.300806388527)*x[0]**o+(-0.39334416908)*x[0]+(0.000711617540648)*x[1]**o+(-0.618539958736)*x[1]+(0.0574098223152)*x[2]**o+(-0.351595384168)*x[2]
        arg[3,2,0,1]=(-0.898675793892)*x[0]**o+(0.904479958277)*x[0]+(0.0781945046018)*x[1]**o+(-0.0461053392586)*x[1]+(0.844798127031)*x[2]**o+(-0.82306838709)*x[2]
        arg[3,2,1,0]=(-0.268742335093)*x[0]**o+(-0.532639944208)*x[0]+(0.0101227109551)*x[1]**o+(-0.438312865518)*x[1]+(0.104288648246)*x[2]**o+(-0.115669027311)*x[2]
        arg[3,2,1,1]=(0.639871308792)*x[0]**o+(0.260847628514)*x[0]+(-0.455670719167)*x[1]**o+(0.210619035618)*x[1]+(-0.401809068119)*x[2]**o+(-0.0625194517026)*x[2]
        arg[3,2,2,0]=(-0.713684223829)*x[0]**o+(0.214281604766)*x[0]+(-0.60152592324)*x[1]**o+(0.249807228316)*x[1]+(-0.939395078261)*x[2]**o+(-0.519744396388)*x[2]
        arg[3,2,2,1]=(-0.340588993443)*x[0]**o+(-0.0165543854997)*x[0]+(-0.366479789338)*x[1]**o+(0.316347630987)*x[1]+(-0.457055798679)*x[2]**o+(-0.132368055553)*x[2]
        arg[3,3,0,0]=(-0.745211736715)*x[0]**o+(0.0106787556884)*x[0]+(-0.448180943446)*x[1]**o+(0.694982281891)*x[1]+(0.924431258221)*x[2]**o+(-0.841633806168)*x[2]
        arg[3,3,0,1]=(0.368054364362)*x[0]**o+(0.19288251992)*x[0]+(0.565854068035)*x[1]**o+(0.747534544754)*x[1]+(0.473165789569)*x[2]**o+(-0.59399986706)*x[2]
        arg[3,3,1,0]=(-0.142707671693)*x[0]**o+(0.177095043535)*x[0]+(-0.348946846847)*x[1]**o+(-0.545430013244)*x[1]+(-0.671405653136)*x[2]**o+(-0.531699969702)*x[2]
        arg[3,3,1,1]=(0.783201055631)*x[0]**o+(0.899495042323)*x[0]+(0.117802924059)*x[1]**o+(-0.277400900172)*x[1]+(-0.351268362305)*x[2]**o+(0.52756480897)*x[2]
        arg[3,3,2,0]=(-0.866758006698)*x[0]**o+(0.293139875238)*x[0]+(-0.318394231317)*x[1]**o+(-0.783701622124)*x[1]+(-0.95137399655)*x[2]**o+(-0.474198736639)*x[2]
        arg[3,3,2,1]=(0.0428398560619)*x[0]**o+(-0.970180601102)*x[0]+(-0.753897573393)*x[1]**o+(-0.0873352699251)*x[1]+(-0.131440532574)*x[2]**o+(0.875241605077)*x[2]
        arg[3,4,0,0]=(0.32399493643)*x[0]**o+(0.67295092524)*x[0]+(-0.29221110671)*x[1]**o+(0.118066079866)*x[1]+(-0.566541154667)*x[2]**o+(-0.806367499079)*x[2]
        arg[3,4,0,1]=(0.278776399755)*x[0]**o+(0.472327826919)*x[0]+(-0.956871861949)*x[1]**o+(0.109236874792)*x[1]+(-0.494095116883)*x[2]**o+(-0.194313775518)*x[2]
        arg[3,4,1,0]=(0.650920037415)*x[0]**o+(0.977698516929)*x[0]+(-0.0730777837998)*x[1]**o+(-0.576543152032)*x[1]+(0.781481934657)*x[2]**o+(-0.389328840554)*x[2]
        arg[3,4,1,1]=(-0.798341529522)*x[0]**o+(-0.630245234712)*x[0]+(-0.444550323489)*x[1]**o+(-0.539189248295)*x[1]+(-0.904964423673)*x[2]**o+(-0.611560462796)*x[2]
        arg[3,4,2,0]=(-0.392991417685)*x[0]**o+(0.313599061567)*x[0]+(-0.215747337398)*x[1]**o+(-0.310316872476)*x[1]+(-0.269452617785)*x[2]**o+(0.684353794645)*x[2]
        arg[3,4,2,1]=(-0.586668781403)*x[0]**o+(-0.0711284937177)*x[0]+(0.408273850302)*x[1]**o+(0.761860299288)*x[1]+(0.902079242449)*x[2]**o+(0.70161017722)*x[2]
        ref[0,0,0,0]=(1.43173970815)/(o+1.)+(-0.156396435558)
        ref[0,0,0,1]=(0.459165989186)/(o+1.)+(0.194399341676)
        ref[0,0,1,0]=(0.989528642545)/(o+1.)+(-0.164206368911)
        ref[0,0,1,1]=(1.37152584676)/(o+1.)+(-0.234493053498)
        ref[0,0,2,0]=(0.422741924108)/(o+1.)+(0.0721378026728)
        ref[0,0,2,1]=(-0.574467903086)/(o+1.)+(0.167204674462)
        ref[0,1,0,0]=(1.36619629735)/(o+1.)+(-0.19947081885)
        ref[0,1,0,1]=(-1.60758660729)/(o+1.)+(0.447728236186)
        ref[0,1,1,0]=(1.5106959073)/(o+1.)+(-0.393852180331)
        ref[0,1,1,1]=(1.85678121239)/(o+1.)+(-0.529280397551)
        ref[0,1,2,0]=(-0.796337866944)/(o+1.)+(0.0812355940636)
        ref[0,1,2,1]=(-0.333114331422)/(o+1.)+(1.22614026311)
        ref[0,2,0,0]=(0.516540083621)/(o+1.)+(-0.358761510118)
        ref[0,2,0,1]=(-2.72820411942)/(o+1.)+(0.15972705639)
        ref[0,2,1,0]=(-0.440149466937)/(o+1.)+(-0.124216123318)
        ref[0,2,1,1]=(-0.137940536755)/(o+1.)+(0.13992652657)
        ref[0,2,2,0]=(0.373759361058)/(o+1.)+(0.210996140796)
        ref[0,2,2,1]=(-1.07034621798)/(o+1.)+(1.2905860125)
        ref[0,3,0,0]=(0.923495049095)/(o+1.)+(-0.897043344396)
        ref[0,3,0,1]=(-0.20180799615)/(o+1.)+(0.272652457923)
        ref[0,3,1,0]=(-0.831052678484)/(o+1.)+(0.326929849662)
        ref[0,3,1,1]=(1.41523396695)/(o+1.)+(-0.5275927117)
        ref[0,3,2,0]=(0.405715363821)/(o+1.)+(0.535330516568)
        ref[0,3,2,1]=(0.877269821733)/(o+1.)+(-0.666082812121)
        ref[0,4,0,0]=(0.884126703786)/(o+1.)+(-0.559427998262)
        ref[0,4,0,1]=(-1.36933304826)/(o+1.)+(0.263100182211)
        ref[0,4,1,0]=(0.871486955825)/(o+1.)+(0.381595437312)
        ref[0,4,1,1]=(1.21446122052)/(o+1.)+(-0.642161706726)
        ref[0,4,2,0]=(-1.82861338966)/(o+1.)+(0.507433118531)
        ref[0,4,2,1]=(0.0642039966469)/(o+1.)+(0.551348562604)
        ref[1,0,0,0]=(0.40922616303)/(o+1.)+(-0.474630105482)
        ref[1,0,0,1]=(1.63943889768)/(o+1.)+(0.285622914296)
        ref[1,0,1,0]=(0.0845961115353)/(o+1.)+(0.143628502705)
        ref[1,0,1,1]=(0.62129375936)/(o+1.)+(0.671645835511)
        ref[1,0,2,0]=(-1.2004575453)/(o+1.)+(0.701757992696)
        ref[1,0,2,1]=(0.435963491825)/(o+1.)+(0.26710538393)
        ref[1,1,0,0]=(0.738921325407)/(o+1.)+(-0.0258542444502)
        ref[1,1,0,1]=(1.95599148183)/(o+1.)+(0.652818492883)
        ref[1,1,1,0]=(0.500576647095)/(o+1.)+(0.857652485995)
        ref[1,1,1,1]=(-0.0151750762821)/(o+1.)+(0.238827857598)
        ref[1,1,2,0]=(1.26963005373)/(o+1.)+(-0.299065008888)
        ref[1,1,2,1]=(-0.82289389301)/(o+1.)+(0.14939351077)
        ref[1,2,0,0]=(-0.495951362447)/(o+1.)+(-0.532833788304)
        ref[1,2,0,1]=(-1.87785443639)/(o+1.)+(-0.521209467732)
        ref[1,2,1,0]=(-0.374085962184)/(o+1.)+(0.505583421209)
        ref[1,2,1,1]=(1.01773556605)/(o+1.)+(0.764863466171)
        ref[1,2,2,0]=(-1.54049109838)/(o+1.)+(-0.511296952595)
        ref[1,2,2,1]=(-1.06948953348)/(o+1.)+(-0.215299794811)
        ref[1,3,0,0]=(-1.20198777829)/(o+1.)+(-0.274360734317)
        ref[1,3,0,1]=(1.13519842695)/(o+1.)+(0.506840913294)
        ref[1,3,1,0]=(0.973820116369)/(o+1.)+(0.485786110977)
        ref[1,3,1,1]=(0.155263308828)/(o+1.)+(0.473436850817)
        ref[1,3,2,0]=(0.476069733748)/(o+1.)+(-0.228046397052)
        ref[1,3,2,1]=(-0.234851551487)/(o+1.)+(-0.574065011567)
        ref[1,4,0,0]=(1.95924908786)/(o+1.)+(-0.493262726539)
        ref[1,4,0,1]=(-0.448532656782)/(o+1.)+(0.393812699074)
        ref[1,4,1,0]=(0.235155577808)/(o+1.)+(0.756260258803)
        ref[1,4,1,1]=(0.0334282473144)/(o+1.)+(-0.837406551217)
        ref[1,4,2,0]=(0.568132489684)/(o+1.)+(-0.13055465798)
        ref[1,4,2,1]=(-0.380042596089)/(o+1.)+(0.243737215624)
        ref[2,0,0,0]=(1.76568035823)/(o+1.)+(0.0789834372326)
        ref[2,0,0,1]=(-0.20023919191)/(o+1.)+(0.901466917208)
        ref[2,0,1,0]=(-1.20715877778)/(o+1.)+(-0.992500439486)
        ref[2,0,1,1]=(-0.761913514624)/(o+1.)+(0.465965334128)
        ref[2,0,2,0]=(-0.944104213428)/(o+1.)+(-0.146881270459)
        ref[2,0,2,1]=(0.105283971659)/(o+1.)+(-1.30498791712)
        ref[2,1,0,0]=(-0.804721214141)/(o+1.)+(0.356736186609)
        ref[2,1,0,1]=(-0.0358563948854)/(o+1.)+(-0.24146509617)
        ref[2,1,1,0]=(-0.437556173502)/(o+1.)+(-0.387215578175)
        ref[2,1,1,1]=(-0.0786030951662)/(o+1.)+(0.654838009465)
        ref[2,1,2,0]=(-0.859200342885)/(o+1.)+(-0.112407873885)
        ref[2,1,2,1]=(-0.296679106354)/(o+1.)+(0.0491751507381)
        ref[2,2,0,0]=(0.462408531841)/(o+1.)+(-0.557975521593)
        ref[2,2,0,1]=(-0.0634225791893)/(o+1.)+(0.0468185300489)
        ref[2,2,1,0]=(1.08205349937)/(o+1.)+(0.621257561187)
        ref[2,2,1,1]=(0.88864389409)/(o+1.)+(0.849886105205)
        ref[2,2,2,0]=(2.01702912587)/(o+1.)+(0.755558425167)
        ref[2,2,2,1]=(1.41263009001)/(o+1.)+(0.331281133791)
        ref[2,3,0,0]=(1.91485193421)/(o+1.)+(0.0643042152414)
        ref[2,3,0,1]=(1.86113720031)/(o+1.)+(0.870090339646)
        ref[2,3,1,0]=(0.442533807126)/(o+1.)+(-0.690437405116)
        ref[2,3,1,1]=(1.52774754257)/(o+1.)+(-0.236667038211)
        ref[2,3,2,0]=(1.08558789233)/(o+1.)+(-0.346580362392)
        ref[2,3,2,1]=(-0.771732912166)/(o+1.)+(-0.49697903252)
        ref[2,4,0,0]=(1.0510676744)/(o+1.)+(0.18646253745)
        ref[2,4,0,1]=(-0.878875991119)/(o+1.)+(-0.271299000644)
        ref[2,4,1,0]=(-0.75026757693)/(o+1.)+(-0.168751100981)
        ref[2,4,1,1]=(0.526155063499)/(o+1.)+(1.041034953)
        ref[2,4,2,0]=(0.377792294719)/(o+1.)+(0.433767213151)
        ref[2,4,2,1]=(0.114142925831)/(o+1.)+(-0.972051850709)
        ref[3,0,0,0]=(0.15538316839)/(o+1.)+(-0.861229788808)
        ref[3,0,0,1]=(-0.725501605569)/(o+1.)+(-0.248009661269)
        ref[3,0,1,0]=(0.791443284981)/(o+1.)+(0.618118504587)
        ref[3,0,1,1]=(-0.411749416988)/(o+1.)+(0.125302421245)
        ref[3,0,2,0]=(0.0968028463245)/(o+1.)+(0.592149023513)
        ref[3,0,2,1]=(0.0947000244134)/(o+1.)+(0.286498958939)
        ref[3,1,0,0]=(0.823197755916)/(o+1.)+(-0.378154495725)
        ref[3,1,0,1]=(1.21588193285)/(o+1.)+(0.0126566819682)
        ref[3,1,1,0]=(0.529639159192)/(o+1.)+(-0.619204956255)
        ref[3,1,1,1]=(1.85472331764)/(o+1.)+(0.3317212712)
        ref[3,1,2,0]=(-0.0770331978898)/(o+1.)+(0.0545214351076)
        ref[3,1,2,1]=(0.46556780359)/(o+1.)+(-0.223576369421)
        ref[3,2,0,0]=(-0.242684948671)/(o+1.)+(-0.681739755992)
        ref[3,2,0,1]=(0.0243168377401)/(o+1.)+(0.0176531159642)
        ref[3,2,1,0]=(-0.154330975892)/(o+1.)+(-0.543310918519)
        ref[3,2,1,1]=(-0.217608478494)/(o+1.)+(0.204473606215)
        ref[3,2,2,0]=(-2.25460522533)/(o+1.)+(-0.0278277816531)
        ref[3,2,2,1]=(-1.16412458146)/(o+1.)+(0.0837125949668)
        ref[3,3,0,0]=(-0.26896142194)/(o+1.)+(-0.0679863842941)
        ref[3,3,0,1]=(1.40707422197)/(o+1.)+(0.173208598807)
        ref[3,3,1,0]=(-1.16306017168)/(o+1.)+(-0.450017469706)
        ref[3,3,1,1]=(0.549735617385)/(o+1.)+(0.57482947556)
        ref[3,3,2,0]=(-2.13652623456)/(o+1.)+(-0.482380241762)
        ref[3,3,2,1]=(-0.842498249905)/(o+1.)+(-0.0911371329748)
        ref[3,4,0,0]=(-0.534757324947)/(o+1.)+(-0.00767524698684)
        ref[3,4,0,1]=(-1.17219057908)/(o+1.)+(0.193625463097)
        ref[3,4,1,0]=(1.35932418827)/(o+1.)+(0.00591326217154)
        ref[3,4,1,1]=(-2.14785627668)/(o+1.)+(-0.890497472901)
        ref[3,4,2,0]=(-0.878191372868)/(o+1.)+(0.343817991868)
        ref[3,4,2,1]=(0.723684311348)/(o+1.)+(0.696170991395)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank0(self):
      """
      tests integral of rank 0 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.217470460227)*x[0]**o+(0.664132905811)*x[0]+(0.493878617743)*x[1]**o+(0.111327715714)*x[1]
        ref=(0.276408157516)/(o+1.)+(0.387730310763)
      else:
        arg=(0.371937838886)*x[0]**o+(0.653846807778)*x[0]+(0.172358058188)*x[1]**o+(-0.928017916614)*x[1]+(0.116156616174)*x[2]**o+(0.371332501889)*x[2]
        ref=(0.660452513249)/(o+1.)+(0.0485806965265)
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank1(self):
      """
      tests integral of rank 1 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.329952046701)*x[0]**o+(0.12757022314)*x[0]+(0.198280362833)*x[1]**o+(0.129179854334)*x[1]
        arg[1]=(0.185895161332)*x[0]**o+(0.270740935936)*x[0]+(0.630122833594)*x[1]**o+(-0.0696403698787)*x[1]
        ref[0]=(-0.131671683869)/(o+1.)+(0.128375038737)
        ref[1]=(0.816017994925)/(o+1.)+(0.100550283029)
      else:
        arg[0]=(0.526257699936)*x[0]**o+(0.0731442796396)*x[0]+(-0.266121642119)*x[1]**o+(-0.929478242136)*x[1]+(0.743092861949)*x[2]**o+(0.22690419858)*x[2]
        arg[1]=(-0.121387039147)*x[0]**o+(-0.0580608263784)*x[0]+(0.761164059949)*x[1]**o+(0.821938972431)*x[1]+(-0.756423175962)*x[2]**o+(0.213207132353)*x[2]
        ref[0]=(1.00322891977)/(o+1.)+(-0.314714881958)
        ref[1]=(-0.11664615516)/(o+1.)+(0.488542639203)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank2(self):
      """
      tests integral of rank 2 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.963318847499)*x[0]**o+(-0.184977238308)*x[0]+(0.153633231879)*x[1]**o+(-0.640028315216)*x[1]
        arg[0,1]=(0.0879103656102)*x[0]**o+(-0.928854425726)*x[0]+(-0.138566176095)*x[1]**o+(-0.670118398936)*x[1]
        arg[0,2]=(0.920856326204)*x[0]**o+(0.461047594595)*x[0]+(-0.914778320363)*x[1]**o+(-0.561644217568)*x[1]
        arg[0,3]=(0.506148289443)*x[0]**o+(-0.474044343536)*x[0]+(0.742127575836)*x[1]**o+(-0.52349314159)*x[1]
        arg[0,4]=(0.0810224543145)*x[0]**o+(-0.307992483498)*x[0]+(0.749673245291)*x[1]**o+(-0.554099873044)*x[1]
        arg[1,0]=(-0.270897099553)*x[0]**o+(-0.222178461886)*x[0]+(0.755729049844)*x[1]**o+(0.740743777894)*x[1]
        arg[1,1]=(0.0220934176819)*x[0]**o+(-0.309641658323)*x[0]+(-0.951833592348)*x[1]**o+(0.430554526918)*x[1]
        arg[1,2]=(0.176033311218)*x[0]**o+(0.728188001962)*x[0]+(-0.428357954006)*x[1]**o+(0.955059500699)*x[1]
        arg[1,3]=(0.946680422474)*x[0]**o+(-0.251968148743)*x[0]+(-0.3704600461)*x[1]**o+(0.692080922297)*x[1]
        arg[1,4]=(0.122655375591)*x[0]**o+(0.617285886533)*x[0]+(0.729184247986)*x[1]**o+(0.30369620264)*x[1]
        arg[2,0]=(-0.957780501271)*x[0]**o+(0.438290429468)*x[0]+(-0.504841431279)*x[1]**o+(0.935350834608)*x[1]
        arg[2,1]=(-0.49410765407)*x[0]**o+(-0.635142349002)*x[0]+(0.700131616572)*x[1]**o+(0.99523209861)*x[1]
        arg[2,2]=(0.0967686936303)*x[0]**o+(0.987754073381)*x[0]+(-0.161220173386)*x[1]**o+(-0.343926097784)*x[1]
        arg[2,3]=(0.560040886628)*x[0]**o+(-0.843778720323)*x[0]+(0.345354034369)*x[1]**o+(0.440584543858)*x[1]
        arg[2,4]=(0.988977815245)*x[0]**o+(0.661922598357)*x[0]+(-0.425421002549)*x[1]**o+(0.624279872655)*x[1]
        arg[3,0]=(-0.997217068098)*x[0]**o+(-0.973497091595)*x[0]+(-0.0980683851598)*x[1]**o+(0.408081557041)*x[1]
        arg[3,1]=(0.0827616522414)*x[0]**o+(0.931022772362)*x[0]+(-0.588688206358)*x[1]**o+(0.900467348992)*x[1]
        arg[3,2]=(0.0933642874383)*x[0]**o+(0.398732664158)*x[0]+(-0.437102212783)*x[1]**o+(-0.126538302466)*x[1]
        arg[3,3]=(0.251989644486)*x[0]**o+(0.765809815972)*x[0]+(-0.325455283605)*x[1]**o+(0.171819372538)*x[1]
        arg[3,4]=(0.649407682145)*x[0]**o+(0.963776897105)*x[0]+(0.671915752131)*x[1]**o+(-0.468729629153)*x[1]
        ref[0,0]=(-0.80968561562)/(o+1.)+(-0.412502776762)
        ref[0,1]=(-0.0506558104848)/(o+1.)+(-0.799486412331)
        ref[0,2]=(0.00607800584081)/(o+1.)+(-0.0502983114867)
        ref[0,3]=(1.24827586528)/(o+1.)+(-0.498768742563)
        ref[0,4]=(0.830695699605)/(o+1.)+(-0.431046178271)
        ref[1,0]=(0.484831950292)/(o+1.)+(0.259282658004)
        ref[1,1]=(-0.929740174666)/(o+1.)+(0.0604564342974)
        ref[1,2]=(-0.252324642788)/(o+1.)+(0.84162375133)
        ref[1,3]=(0.576220376373)/(o+1.)+(0.220056386777)
        ref[1,4]=(0.851839623577)/(o+1.)+(0.460491044586)
        ref[2,0]=(-1.46262193255)/(o+1.)+(0.686820632038)
        ref[2,1]=(0.206023962501)/(o+1.)+(0.180044874804)
        ref[2,2]=(-0.0644514797559)/(o+1.)+(0.321913987798)
        ref[2,3]=(0.905394920997)/(o+1.)+(-0.201597088233)
        ref[2,4]=(0.563556812695)/(o+1.)+(0.643101235506)
        ref[3,0]=(-1.09528545326)/(o+1.)+(-0.282707767277)
        ref[3,1]=(-0.505926554117)/(o+1.)+(0.915745060677)
        ref[3,2]=(-0.343737925345)/(o+1.)+(0.136097180846)
        ref[3,3]=(-0.0734656391194)/(o+1.)+(0.468814594255)
        ref[3,4]=(1.32132343428)/(o+1.)+(0.247523633976)
      else:
        arg[0,0]=(-0.512698606017)*x[0]**o+(-0.156853163498)*x[0]+(0.247997355633)*x[1]**o+(-0.363003068904)*x[1]+(-0.940097989108)*x[2]**o+(-0.661063850288)*x[2]
        arg[0,1]=(-0.886390181577)*x[0]**o+(0.724049511142)*x[0]+(0.260684592163)*x[1]**o+(0.265648885034)*x[1]+(0.123084281693)*x[2]**o+(0.181492979151)*x[2]
        arg[0,2]=(-0.570543998851)*x[0]**o+(0.0541041165388)*x[0]+(0.630994990544)*x[1]**o+(-0.89311079559)*x[1]+(0.750714073617)*x[2]**o+(-0.0684796380527)*x[2]
        arg[0,3]=(0.605486031268)*x[0]**o+(0.0387617487672)*x[0]+(-0.365027101406)*x[1]**o+(0.323275180514)*x[1]+(-0.763142660003)*x[2]**o+(-0.331069480921)*x[2]
        arg[0,4]=(-0.0779035848472)*x[0]**o+(-0.0234904624931)*x[0]+(0.591635786678)*x[1]**o+(-0.793680126332)*x[1]+(-0.874742296757)*x[2]**o+(-0.184528885157)*x[2]
        arg[1,0]=(-0.62480972544)*x[0]**o+(-0.99242215547)*x[0]+(0.634250679117)*x[1]**o+(-0.785448642692)*x[1]+(-0.013335654402)*x[2]**o+(0.52112750417)*x[2]
        arg[1,1]=(0.611541580179)*x[0]**o+(-0.203144044802)*x[0]+(-0.411277184995)*x[1]**o+(0.373487145939)*x[1]+(-0.679339345026)*x[2]**o+(-0.632413536218)*x[2]
        arg[1,2]=(-0.258927016205)*x[0]**o+(0.0242628959926)*x[0]+(-0.596909760383)*x[1]**o+(-0.295787354787)*x[1]+(-0.337930022823)*x[2]**o+(-0.825085100446)*x[2]
        arg[1,3]=(-0.0976663781507)*x[0]**o+(0.716600132838)*x[0]+(-0.682155341898)*x[1]**o+(-0.154484864167)*x[1]+(0.754686874908)*x[2]**o+(-0.102453470268)*x[2]
        arg[1,4]=(0.268714339964)*x[0]**o+(0.974234398844)*x[0]+(-0.370113158384)*x[1]**o+(0.0745228956861)*x[1]+(-0.0674370676047)*x[2]**o+(0.132605448278)*x[2]
        arg[2,0]=(-0.00750893262562)*x[0]**o+(0.507641887134)*x[0]+(0.638483932117)*x[1]**o+(-0.940295540912)*x[1]+(-0.201492519698)*x[2]**o+(0.504878595818)*x[2]
        arg[2,1]=(-0.0283522980593)*x[0]**o+(-0.76577346605)*x[0]+(0.0863472830598)*x[1]**o+(-0.904447859187)*x[1]+(-0.203390005492)*x[2]**o+(-0.741629136645)*x[2]
        arg[2,2]=(-0.0196123831667)*x[0]**o+(-0.320721372799)*x[0]+(0.0670421287749)*x[1]**o+(-0.415029682416)*x[1]+(0.450269143475)*x[2]**o+(-0.735201210502)*x[2]
        arg[2,3]=(-0.621537330758)*x[0]**o+(0.21573881897)*x[0]+(0.0364275955835)*x[1]**o+(0.873908952751)*x[1]+(-0.197046790303)*x[2]**o+(0.655654722165)*x[2]
        arg[2,4]=(0.432531007791)*x[0]**o+(-0.65605251166)*x[0]+(0.728245204833)*x[1]**o+(-0.769998936989)*x[1]+(-0.797879623133)*x[2]**o+(0.36575554332)*x[2]
        arg[3,0]=(0.512219734043)*x[0]**o+(-0.177942735603)*x[0]+(-0.307293350237)*x[1]**o+(-0.780181541911)*x[1]+(-0.510201727643)*x[2]**o+(-0.567980715525)*x[2]
        arg[3,1]=(-0.405123989366)*x[0]**o+(-0.133316566583)*x[0]+(0.552868689995)*x[1]**o+(-0.921149434884)*x[1]+(0.510851078054)*x[2]**o+(0.7207533866)*x[2]
        arg[3,2]=(0.150011788901)*x[0]**o+(-0.55947555552)*x[0]+(-0.353366762442)*x[1]**o+(0.204954197301)*x[1]+(-0.48254632377)*x[2]**o+(-0.272738900127)*x[2]
        arg[3,3]=(-0.690981207461)*x[0]**o+(0.38683432128)*x[0]+(0.655731808738)*x[1]**o+(0.162269870301)*x[1]+(-0.373468659959)*x[2]**o+(-0.841431521009)*x[2]
        arg[3,4]=(0.156455675319)*x[0]**o+(0.501690084272)*x[0]+(-0.339255791314)*x[1]**o+(0.0832260784264)*x[1]+(0.394137401365)*x[2]**o+(-0.92403906326)*x[2]
        ref[0,0]=(-1.20479923949)/(o+1.)+(-0.590460041345)
        ref[0,1]=(-0.502621307721)/(o+1.)+(0.585595687663)
        ref[0,2]=(0.811165065311)/(o+1.)+(-0.453743158552)
        ref[0,3]=(-0.522683730141)/(o+1.)+(0.0154837241797)
        ref[0,4]=(-0.361010094926)/(o+1.)+(-0.500849736991)
        ref[1,0]=(-0.00389470072479)/(o+1.)+(-0.628371646996)
        ref[1,1]=(-0.479074949842)/(o+1.)+(-0.23103521754)
        ref[1,2]=(-1.19376679941)/(o+1.)+(-0.54830477962)
        ref[1,3]=(-0.0251348451402)/(o+1.)+(0.229830899202)
        ref[1,4]=(-0.168835886025)/(o+1.)+(0.590681371404)
        ref[2,0]=(0.429482479793)/(o+1.)+(0.0361124710203)
        ref[2,1]=(-0.145395020492)/(o+1.)+(-1.20592523094)
        ref[2,2]=(0.497698889084)/(o+1.)+(-0.735476132859)
        ref[2,3]=(-0.782156525477)/(o+1.)+(0.872651246943)
        ref[2,4]=(0.362896589491)/(o+1.)+(-0.530147952664)
        ref[3,0]=(-0.305275343837)/(o+1.)+(-0.763052496519)
        ref[3,1]=(0.658595778683)/(o+1.)+(-0.166856307434)
        ref[3,2]=(-0.685901297311)/(o+1.)+(-0.313630129173)
        ref[3,3]=(-0.408718058683)/(o+1.)+(-0.146163664714)
        ref[3,4]=(0.21133728537)/(o+1.)+(-0.169561450281)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank3(self):
      """
      tests integral of rank 3 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.165423911829)*x[0]**o+(0.912542062064)*x[0]+(0.515840721225)*x[1]**o+(-0.855922021017)*x[1]
        arg[0,0,1]=(0.591873802546)*x[0]**o+(-0.039385167187)*x[0]+(0.0261484884274)*x[1]**o+(0.227950240434)*x[1]
        arg[0,1,0]=(-0.825833712046)*x[0]**o+(0.989606797765)*x[0]+(0.975260558791)*x[1]**o+(0.297963603984)*x[1]
        arg[0,1,1]=(-0.658054502482)*x[0]**o+(-0.865505090963)*x[0]+(0.817184084453)*x[1]**o+(-0.409588098899)*x[1]
        arg[1,0,0]=(0.538456987198)*x[0]**o+(-0.0768787509391)*x[0]+(0.424620431835)*x[1]**o+(0.799917948831)*x[1]
        arg[1,0,1]=(0.125945558199)*x[0]**o+(-0.48320049996)*x[0]+(0.105474962327)*x[1]**o+(-0.090713555582)*x[1]
        arg[1,1,0]=(0.835213375418)*x[0]**o+(-0.513407038769)*x[0]+(-0.342890811735)*x[1]**o+(-0.285057736244)*x[1]
        arg[1,1,1]=(0.762782085749)*x[0]**o+(-0.308457087352)*x[0]+(0.23756726753)*x[1]**o+(-0.909365956504)*x[1]
        arg[2,0,0]=(-0.76467980678)*x[0]**o+(-0.265552922006)*x[0]+(-0.261862355539)*x[1]**o+(-0.399407404618)*x[1]
        arg[2,0,1]=(0.79401193617)*x[0]**o+(-0.0300461495216)*x[0]+(-0.978834133508)*x[1]**o+(-0.984845382599)*x[1]
        arg[2,1,0]=(0.898033018534)*x[0]**o+(-0.184513430159)*x[0]+(0.0430634098526)*x[1]**o+(0.62213313954)*x[1]
        arg[2,1,1]=(0.126926439246)*x[0]**o+(0.671259428917)*x[0]+(-0.396062855417)*x[1]**o+(0.0461825691916)*x[1]
        arg[3,0,0]=(0.185756594279)*x[0]**o+(0.570946291973)*x[0]+(0.0673731944956)*x[1]**o+(0.367766610292)*x[1]
        arg[3,0,1]=(0.021269224946)*x[0]**o+(-0.463663279063)*x[0]+(-0.687930988068)*x[1]**o+(0.184714204849)*x[1]
        arg[3,1,0]=(0.754288639303)*x[0]**o+(-0.94612827517)*x[0]+(0.802463196709)*x[1]**o+(0.202355582077)*x[1]
        arg[3,1,1]=(0.840002857786)*x[0]**o+(-0.692207739447)*x[0]+(0.486569391181)*x[1]**o+(-0.713960793192)*x[1]
        arg[4,0,0]=(-0.205794539655)*x[0]**o+(-0.930458992533)*x[0]+(-0.62580205172)*x[1]**o+(-0.884738110673)*x[1]
        arg[4,0,1]=(-0.91433612349)*x[0]**o+(0.694961527804)*x[0]+(-0.510941334746)*x[1]**o+(-0.510725271706)*x[1]
        arg[4,1,0]=(0.852785499366)*x[0]**o+(0.870826628851)*x[0]+(-0.577529754198)*x[1]**o+(-0.583793684857)*x[1]
        arg[4,1,1]=(-0.604869226881)*x[0]**o+(0.482463572402)*x[0]+(0.829654316885)*x[1]**o+(0.597543822455)*x[1]
        arg[5,0,0]=(0.413459469749)*x[0]**o+(0.865143264083)*x[0]+(0.986197644249)*x[1]**o+(-0.2511021346)*x[1]
        arg[5,0,1]=(0.811572168213)*x[0]**o+(-0.110121334437)*x[0]+(-0.985492232242)*x[1]**o+(-0.913436084156)*x[1]
        arg[5,1,0]=(-0.660869051422)*x[0]**o+(-0.743235620427)*x[0]+(0.205393778238)*x[1]**o+(0.296069404687)*x[1]
        arg[5,1,1]=(-0.13285011267)*x[0]**o+(0.659400366834)*x[0]+(-0.889984368942)*x[1]**o+(0.584893563559)*x[1]
        ref[0,0,0]=(0.350416809396)/(o+1.)+(0.0283100205234)
        ref[0,0,1]=(0.618022290973)/(o+1.)+(0.0942825366236)
        ref[0,1,0]=(0.149426846745)/(o+1.)+(0.643785200874)
        ref[0,1,1]=(0.159129581971)/(o+1.)+(-0.637546594931)
        ref[1,0,0]=(0.963077419032)/(o+1.)+(0.361519598946)
        ref[1,0,1]=(0.231420520526)/(o+1.)+(-0.286957027771)
        ref[1,1,0]=(0.492322563683)/(o+1.)+(-0.399232387506)
        ref[1,1,1]=(1.00034935328)/(o+1.)+(-0.608911521928)
        ref[2,0,0]=(-1.02654216232)/(o+1.)+(-0.332480163312)
        ref[2,0,1]=(-0.184822197338)/(o+1.)+(-0.507445766061)
        ref[2,1,0]=(0.941096428387)/(o+1.)+(0.21880985469)
        ref[2,1,1]=(-0.269136416172)/(o+1.)+(0.358720999054)
        ref[3,0,0]=(0.253129788774)/(o+1.)+(0.469356451133)
        ref[3,0,1]=(-0.666661763122)/(o+1.)+(-0.139474537107)
        ref[3,1,0]=(1.55675183601)/(o+1.)+(-0.371886346546)
        ref[3,1,1]=(1.32657224897)/(o+1.)+(-0.703084266319)
        ref[4,0,0]=(-0.831596591375)/(o+1.)+(-0.907598551603)
        ref[4,0,1]=(-1.42527745824)/(o+1.)+(0.0921181280487)
        ref[4,1,0]=(0.275255745168)/(o+1.)+(0.143516471997)
        ref[4,1,1]=(0.224785090004)/(o+1.)+(0.540003697429)
        ref[5,0,0]=(1.399657114)/(o+1.)+(0.307020564742)
        ref[5,0,1]=(-0.173920064028)/(o+1.)+(-0.511778709296)
        ref[5,1,0]=(-0.455475273184)/(o+1.)+(-0.22358310787)
        ref[5,1,1]=(-1.02283448161)/(o+1.)+(0.622146965196)
      else:
        arg[0,0,0]=(-0.564478749422)*x[0]**o+(-0.0357223504179)*x[0]+(0.396603298865)*x[1]**o+(-0.461941327703)*x[1]+(-0.667580657928)*x[2]**o+(-0.589957067263)*x[2]
        arg[0,0,1]=(0.930616008951)*x[0]**o+(-0.764067565635)*x[0]+(-0.71350617205)*x[1]**o+(-0.530942701898)*x[1]+(-0.103797356434)*x[2]**o+(0.410704036673)*x[2]
        arg[0,1,0]=(-0.59127392299)*x[0]**o+(0.452490122131)*x[0]+(0.714793625631)*x[1]**o+(-0.933114051133)*x[1]+(-0.693437840325)*x[2]**o+(0.979267231764)*x[2]
        arg[0,1,1]=(0.179682332267)*x[0]**o+(-0.419960476425)*x[0]+(-0.10986598341)*x[1]**o+(0.912986428497)*x[1]+(0.0886946782573)*x[2]**o+(0.516065958873)*x[2]
        arg[1,0,0]=(0.524446863481)*x[0]**o+(-0.439668193423)*x[0]+(0.681259811488)*x[1]**o+(0.439275262258)*x[1]+(0.84573881706)*x[2]**o+(0.377640553175)*x[2]
        arg[1,0,1]=(-0.007464905203)*x[0]**o+(0.234727832471)*x[0]+(0.0428956356618)*x[1]**o+(0.195772930172)*x[1]+(0.457790934614)*x[2]**o+(0.445309166495)*x[2]
        arg[1,1,0]=(0.821283375827)*x[0]**o+(0.212480005798)*x[0]+(0.148817990961)*x[1]**o+(0.119716150877)*x[1]+(0.658959126865)*x[2]**o+(-0.346216916131)*x[2]
        arg[1,1,1]=(-0.0587222422894)*x[0]**o+(-0.325321661973)*x[0]+(0.149293818689)*x[1]**o+(-0.632073115037)*x[1]+(0.018604271874)*x[2]**o+(-0.412288240166)*x[2]
        arg[2,0,0]=(-0.22588082169)*x[0]**o+(0.197780461217)*x[0]+(0.387336989213)*x[1]**o+(-0.680021661016)*x[1]+(-0.903782254346)*x[2]**o+(-0.273268109853)*x[2]
        arg[2,0,1]=(0.415876394297)*x[0]**o+(0.199596795904)*x[0]+(-0.445817134682)*x[1]**o+(0.672539590684)*x[1]+(0.907825473964)*x[2]**o+(0.459560469691)*x[2]
        arg[2,1,0]=(-0.052146887134)*x[0]**o+(-0.229450574738)*x[0]+(0.0929720151604)*x[1]**o+(-0.893861034491)*x[1]+(-0.294030328982)*x[2]**o+(-0.193256779109)*x[2]
        arg[2,1,1]=(0.361171422066)*x[0]**o+(-0.212022056495)*x[0]+(0.0672221746963)*x[1]**o+(0.57506871515)*x[1]+(0.654894469509)*x[2]**o+(0.0749777260533)*x[2]
        arg[3,0,0]=(0.505355253404)*x[0]**o+(-0.0753563586359)*x[0]+(0.544299279727)*x[1]**o+(-0.781954131328)*x[1]+(0.534390713891)*x[2]**o+(-0.719289344378)*x[2]
        arg[3,0,1]=(-0.589309652827)*x[0]**o+(-0.0557421151178)*x[0]+(-0.275311076837)*x[1]**o+(0.998180518149)*x[1]+(-0.844823856433)*x[2]**o+(-0.988843805162)*x[2]
        arg[3,1,0]=(-0.411075104301)*x[0]**o+(-0.405896828894)*x[0]+(0.242301856034)*x[1]**o+(-0.760648969984)*x[1]+(0.0940367373673)*x[2]**o+(-0.121420230853)*x[2]
        arg[3,1,1]=(-0.201426146967)*x[0]**o+(0.722847134095)*x[0]+(-0.00743580369433)*x[1]**o+(-0.593158423904)*x[1]+(0.630575332064)*x[2]**o+(-0.309839791343)*x[2]
        arg[4,0,0]=(0.370057366738)*x[0]**o+(0.35683022394)*x[0]+(0.886304520049)*x[1]**o+(0.834488899465)*x[1]+(0.27408661832)*x[2]**o+(0.325007409933)*x[2]
        arg[4,0,1]=(0.386153377977)*x[0]**o+(-0.889700309398)*x[0]+(-0.60695970137)*x[1]**o+(0.701208035732)*x[1]+(0.435164510485)*x[2]**o+(0.40223148148)*x[2]
        arg[4,1,0]=(-0.227256041906)*x[0]**o+(-0.0406115125763)*x[0]+(-0.0298974314183)*x[1]**o+(0.887167464033)*x[1]+(0.1004751563)*x[2]**o+(0.826924545719)*x[2]
        arg[4,1,1]=(0.733575640137)*x[0]**o+(-0.376269474248)*x[0]+(-0.841290711112)*x[1]**o+(-0.0633013917179)*x[1]+(0.335730636406)*x[2]**o+(0.913607658221)*x[2]
        arg[5,0,0]=(-0.896019254334)*x[0]**o+(0.465470987762)*x[0]+(0.338543664838)*x[1]**o+(0.182102591196)*x[1]+(0.476317455739)*x[2]**o+(0.191367623163)*x[2]
        arg[5,0,1]=(0.828458939676)*x[0]**o+(-0.37853033744)*x[0]+(-0.540809170906)*x[1]**o+(0.367869598038)*x[1]+(0.321196361443)*x[2]**o+(0.548981819769)*x[2]
        arg[5,1,0]=(-0.0805871335579)*x[0]**o+(-0.554679375424)*x[0]+(-0.259438414137)*x[1]**o+(0.72362534417)*x[1]+(-0.838855028889)*x[2]**o+(0.929802758377)*x[2]
        arg[5,1,1]=(-0.947910259125)*x[0]**o+(0.654012824998)*x[0]+(0.425232602078)*x[1]**o+(-0.649345176159)*x[1]+(-0.946280438348)*x[2]**o+(-0.429547284912)*x[2]
        ref[0,0,0]=(-0.835456108484)/(o+1.)+(-0.543810372692)
        ref[0,0,1]=(0.113312480467)/(o+1.)+(-0.44215311543)
        ref[0,1,0]=(-0.569918137684)/(o+1.)+(0.249321651381)
        ref[0,1,1]=(0.158511027114)/(o+1.)+(0.504545955472)
        ref[1,0,0]=(2.05144549203)/(o+1.)+(0.188623811006)
        ref[1,0,1]=(0.493221665073)/(o+1.)+(0.437904964569)
        ref[1,1,0]=(1.62906049365)/(o+1.)+(-0.00701037972825)
        ref[1,1,1]=(0.109175848273)/(o+1.)+(-0.684841508588)
        ref[2,0,0]=(-0.742326086822)/(o+1.)+(-0.377754654826)
        ref[2,0,1]=(0.877884733578)/(o+1.)+(0.665848428139)
        ref[2,1,0]=(-0.253205200956)/(o+1.)+(-0.658284194169)
        ref[2,1,1]=(1.08328806627)/(o+1.)+(0.219012192354)
        ref[3,0,0]=(1.58404524702)/(o+1.)+(-0.788299917171)
        ref[3,0,1]=(-1.7094445861)/(o+1.)+(-0.0232027010654)
        ref[3,1,0]=(-0.0747365109002)/(o+1.)+(-0.643983014866)
        ref[3,1,1]=(0.421713381403)/(o+1.)+(-0.0900755405762)
        ref[4,0,0]=(1.53044850511)/(o+1.)+(0.758163266669)
        ref[4,0,1]=(0.214358187092)/(o+1.)+(0.106869603907)
        ref[4,1,0]=(-0.156678317024)/(o+1.)+(0.836740248588)
        ref[4,1,1]=(0.228015565432)/(o+1.)+(0.237018396127)
        ref[5,0,0]=(-0.0811581337568)/(o+1.)+(0.419470601061)
        ref[5,0,1]=(0.608846130213)/(o+1.)+(0.269160540183)
        ref[5,1,0]=(-1.17888057658)/(o+1.)+(0.549374363561)
        ref[5,1,1]=(-1.4689580954)/(o+1.)+(-0.212439818036)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromData_Function_rank4(self):
      """
      tests integral of rank 4 Data on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.216014106349)*x[0]**o+(0.98826230471)*x[0]+(0.522566177319)*x[1]**o+(-0.672520997452)*x[1]
        arg[0,0,0,1]=(0.368024408084)*x[0]**o+(0.047972971332)*x[0]+(-0.307374327321)*x[1]**o+(0.0745169735457)*x[1]
        arg[0,0,1,0]=(0.438735045452)*x[0]**o+(0.76355958752)*x[0]+(0.200829025437)*x[1]**o+(0.56912407759)*x[1]
        arg[0,0,1,1]=(-0.0328769393712)*x[0]**o+(-0.39701363611)*x[0]+(0.704125532519)*x[1]**o+(-0.6602435084)*x[1]
        arg[0,0,2,0]=(-0.547717992717)*x[0]**o+(-0.844256763776)*x[0]+(0.621300470147)*x[1]**o+(-0.105216181226)*x[1]
        arg[0,0,2,1]=(-0.723474063234)*x[0]**o+(-0.980738972901)*x[0]+(-0.183578659566)*x[1]**o+(0.586038450047)*x[1]
        arg[0,1,0,0]=(0.505379907001)*x[0]**o+(-0.470514711277)*x[0]+(0.799116811747)*x[1]**o+(0.682760846386)*x[1]
        arg[0,1,0,1]=(0.353992609198)*x[0]**o+(-0.848107655585)*x[0]+(-0.774065459855)*x[1]**o+(0.0323137462292)*x[1]
        arg[0,1,1,0]=(0.799068805694)*x[0]**o+(-0.979761821209)*x[0]+(0.642674987867)*x[1]**o+(-0.639029952044)*x[1]
        arg[0,1,1,1]=(0.542556683327)*x[0]**o+(0.876753086356)*x[0]+(-0.387121297643)*x[1]**o+(0.377543179628)*x[1]
        arg[0,1,2,0]=(0.424672899596)*x[0]**o+(-0.134184751569)*x[0]+(-0.561877304114)*x[1]**o+(0.549933699347)*x[1]
        arg[0,1,2,1]=(-0.126173294501)*x[0]**o+(-0.338964866844)*x[0]+(-0.163536115759)*x[1]**o+(0.747551558186)*x[1]
        arg[0,2,0,0]=(0.304334183452)*x[0]**o+(0.927809971187)*x[0]+(0.961504942988)*x[1]**o+(0.0698014252396)*x[1]
        arg[0,2,0,1]=(0.488820088455)*x[0]**o+(0.890735247208)*x[0]+(0.686782901398)*x[1]**o+(0.972954994564)*x[1]
        arg[0,2,1,0]=(-0.737577792766)*x[0]**o+(-0.351698913646)*x[0]+(0.392591489865)*x[1]**o+(0.79238240143)*x[1]
        arg[0,2,1,1]=(-0.997254027544)*x[0]**o+(0.579800166465)*x[0]+(-0.95000690643)*x[1]**o+(-0.166463083476)*x[1]
        arg[0,2,2,0]=(-0.999703749644)*x[0]**o+(0.0415211466468)*x[0]+(0.0816160245139)*x[1]**o+(-0.0921114587195)*x[1]
        arg[0,2,2,1]=(-0.172818721386)*x[0]**o+(0.10475692446)*x[0]+(-0.855815339933)*x[1]**o+(-0.50842580034)*x[1]
        arg[0,3,0,0]=(-0.154563121513)*x[0]**o+(-0.0956305513369)*x[0]+(-0.893783521038)*x[1]**o+(0.10772474638)*x[1]
        arg[0,3,0,1]=(0.365331678349)*x[0]**o+(0.403124820832)*x[0]+(-0.0750500521493)*x[1]**o+(0.62559762808)*x[1]
        arg[0,3,1,0]=(-0.42235875521)*x[0]**o+(0.831133238533)*x[0]+(-0.893322944982)*x[1]**o+(0.936074096319)*x[1]
        arg[0,3,1,1]=(0.587382334689)*x[0]**o+(0.996190710631)*x[0]+(-0.796032791654)*x[1]**o+(0.783491453801)*x[1]
        arg[0,3,2,0]=(-0.369527027508)*x[0]**o+(-0.501417552089)*x[0]+(-0.372105439045)*x[1]**o+(0.653283000152)*x[1]
        arg[0,3,2,1]=(0.914893750314)*x[0]**o+(-0.769183323793)*x[0]+(0.571638486717)*x[1]**o+(-0.939784652259)*x[1]
        arg[0,4,0,0]=(0.134859060251)*x[0]**o+(-0.297363165275)*x[0]+(0.198957028974)*x[1]**o+(-0.828531725127)*x[1]
        arg[0,4,0,1]=(-0.371375370217)*x[0]**o+(-0.0422058718402)*x[0]+(0.502010302278)*x[1]**o+(0.442051802597)*x[1]
        arg[0,4,1,0]=(0.513346571173)*x[0]**o+(0.182640867356)*x[0]+(0.76200283664)*x[1]**o+(0.134640432361)*x[1]
        arg[0,4,1,1]=(-0.815543712756)*x[0]**o+(0.0306226823724)*x[0]+(-0.0211231926242)*x[1]**o+(-0.718800228294)*x[1]
        arg[0,4,2,0]=(-0.980596107555)*x[0]**o+(-0.714237107759)*x[0]+(0.547986215851)*x[1]**o+(0.180131474534)*x[1]
        arg[0,4,2,1]=(-0.460780402239)*x[0]**o+(-0.483771113997)*x[0]+(-0.017140263255)*x[1]**o+(-0.127947184627)*x[1]
        arg[1,0,0,0]=(-0.0187443155707)*x[0]**o+(-0.490914415873)*x[0]+(-0.720022496093)*x[1]**o+(0.0265633770733)*x[1]
        arg[1,0,0,1]=(-0.634855765904)*x[0]**o+(0.345031954366)*x[0]+(-0.0532392363316)*x[1]**o+(0.983295524564)*x[1]
        arg[1,0,1,0]=(0.262182040476)*x[0]**o+(-0.883857758733)*x[0]+(0.880670543535)*x[1]**o+(0.193477695378)*x[1]
        arg[1,0,1,1]=(-0.0825976168371)*x[0]**o+(0.161838720337)*x[0]+(-0.0110587499089)*x[1]**o+(-0.634455402489)*x[1]
        arg[1,0,2,0]=(0.284221958342)*x[0]**o+(-0.948548729029)*x[0]+(0.430914344458)*x[1]**o+(-0.18563638241)*x[1]
        arg[1,0,2,1]=(-0.459558022898)*x[0]**o+(0.431907572866)*x[0]+(-0.0597669070218)*x[1]**o+(0.0694271372225)*x[1]
        arg[1,1,0,0]=(0.54071078447)*x[0]**o+(-0.244570994356)*x[0]+(0.00506219762373)*x[1]**o+(-0.0816378019186)*x[1]
        arg[1,1,0,1]=(-0.100752221379)*x[0]**o+(0.535738345566)*x[0]+(0.730325284553)*x[1]**o+(-0.885044748717)*x[1]
        arg[1,1,1,0]=(-0.0422460353138)*x[0]**o+(-0.986416505995)*x[0]+(-0.111766776997)*x[1]**o+(-0.842179226197)*x[1]
        arg[1,1,1,1]=(0.0690281593169)*x[0]**o+(-0.885054188663)*x[0]+(-0.680349314907)*x[1]**o+(0.00277190934345)*x[1]
        arg[1,1,2,0]=(0.502632146841)*x[0]**o+(-0.197332151163)*x[0]+(-0.0942257361149)*x[1]**o+(0.045171692749)*x[1]
        arg[1,1,2,1]=(0.427910466818)*x[0]**o+(-0.61355947216)*x[0]+(0.465128040517)*x[1]**o+(0.0874536555354)*x[1]
        arg[1,2,0,0]=(0.943819415417)*x[0]**o+(0.884020001869)*x[0]+(-0.769166027477)*x[1]**o+(-0.10032307068)*x[1]
        arg[1,2,0,1]=(0.0670224828564)*x[0]**o+(-0.645424481071)*x[0]+(-0.756361652678)*x[1]**o+(-0.274687916196)*x[1]
        arg[1,2,1,0]=(0.397836742229)*x[0]**o+(-0.708791058745)*x[0]+(0.920869587663)*x[1]**o+(-0.250037069918)*x[1]
        arg[1,2,1,1]=(0.722570486719)*x[0]**o+(-0.358082774936)*x[0]+(0.786344673969)*x[1]**o+(0.310591893637)*x[1]
        arg[1,2,2,0]=(-0.379404573553)*x[0]**o+(-0.969753501645)*x[0]+(-0.0745714915337)*x[1]**o+(-0.262943796911)*x[1]
        arg[1,2,2,1]=(0.817999459632)*x[0]**o+(0.731516351792)*x[0]+(-0.531411613325)*x[1]**o+(-0.824858061154)*x[1]
        arg[1,3,0,0]=(-0.206295444021)*x[0]**o+(-0.781973944918)*x[0]+(-0.173845545382)*x[1]**o+(-0.360122364573)*x[1]
        arg[1,3,0,1]=(-0.0604068005719)*x[0]**o+(0.238691416507)*x[0]+(-0.82262982031)*x[1]**o+(-0.893751755968)*x[1]
        arg[1,3,1,0]=(-0.686409669386)*x[0]**o+(0.711830599399)*x[0]+(0.0839688142166)*x[1]**o+(-0.99463449684)*x[1]
        arg[1,3,1,1]=(-0.0864745975077)*x[0]**o+(0.00068907895313)*x[0]+(-0.766810513534)*x[1]**o+(0.685236202954)*x[1]
        arg[1,3,2,0]=(0.0909338812695)*x[0]**o+(-0.709595654169)*x[0]+(-0.483428504499)*x[1]**o+(-0.0707341191006)*x[1]
        arg[1,3,2,1]=(-0.38342983855)*x[0]**o+(-0.0407824575865)*x[0]+(-0.201329040891)*x[1]**o+(0.0240007937141)*x[1]
        arg[1,4,0,0]=(0.512465803173)*x[0]**o+(-0.371898603608)*x[0]+(-0.79453595861)*x[1]**o+(0.340546986375)*x[1]
        arg[1,4,0,1]=(0.742383198442)*x[0]**o+(-0.603548189912)*x[0]+(-0.270277825546)*x[1]**o+(-0.525100913894)*x[1]
        arg[1,4,1,0]=(-0.693074384029)*x[0]**o+(-0.825037404844)*x[0]+(-0.626116177978)*x[1]**o+(0.628159342479)*x[1]
        arg[1,4,1,1]=(-0.595844943255)*x[0]**o+(0.250758194807)*x[0]+(0.504390052142)*x[1]**o+(-0.124588859702)*x[1]
        arg[1,4,2,0]=(-0.651727490997)*x[0]**o+(0.411138504151)*x[0]+(0.323769160331)*x[1]**o+(0.904379126671)*x[1]
        arg[1,4,2,1]=(-0.71307850178)*x[0]**o+(0.734964608119)*x[0]+(0.348693698506)*x[1]**o+(0.272270639772)*x[1]
        arg[2,0,0,0]=(0.558236567771)*x[0]**o+(0.516985445776)*x[0]+(0.0278915137861)*x[1]**o+(-0.268619977833)*x[1]
        arg[2,0,0,1]=(-0.871181709658)*x[0]**o+(-0.132251987744)*x[0]+(-0.215497073461)*x[1]**o+(0.508008002082)*x[1]
        arg[2,0,1,0]=(0.880795999572)*x[0]**o+(-0.231151166496)*x[0]+(-0.698846681617)*x[1]**o+(0.914483413825)*x[1]
        arg[2,0,1,1]=(0.684792911211)*x[0]**o+(-0.801408782989)*x[0]+(0.689713232168)*x[1]**o+(0.142847173797)*x[1]
        arg[2,0,2,0]=(-0.912566319945)*x[0]**o+(0.811138574657)*x[0]+(-0.884250879071)*x[1]**o+(-0.177206225477)*x[1]
        arg[2,0,2,1]=(-0.898648046678)*x[0]**o+(0.581554373688)*x[0]+(0.567642878443)*x[1]**o+(0.636312832751)*x[1]
        arg[2,1,0,0]=(-0.788853365648)*x[0]**o+(-0.135329327941)*x[0]+(0.180850147594)*x[1]**o+(0.742619029335)*x[1]
        arg[2,1,0,1]=(-0.824539980645)*x[0]**o+(0.857173393615)*x[0]+(0.850455264916)*x[1]**o+(0.975422322464)*x[1]
        arg[2,1,1,0]=(0.462293517317)*x[0]**o+(-0.173835273794)*x[0]+(-0.800072514276)*x[1]**o+(0.140214114627)*x[1]
        arg[2,1,1,1]=(-0.187890814293)*x[0]**o+(0.534472364495)*x[0]+(-0.799640665571)*x[1]**o+(0.301287955302)*x[1]
        arg[2,1,2,0]=(0.278901492946)*x[0]**o+(0.620544326886)*x[0]+(-0.705702443752)*x[1]**o+(-0.489690636328)*x[1]
        arg[2,1,2,1]=(-0.926772900235)*x[0]**o+(0.134654742845)*x[0]+(0.356561975337)*x[1]**o+(-0.919999359857)*x[1]
        arg[2,2,0,0]=(-0.675956301416)*x[0]**o+(0.149582589205)*x[0]+(0.186440827385)*x[1]**o+(0.88381610446)*x[1]
        arg[2,2,0,1]=(-0.697267962915)*x[0]**o+(0.991366596424)*x[0]+(0.310290094618)*x[1]**o+(-0.881771505066)*x[1]
        arg[2,2,1,0]=(0.506027358474)*x[0]**o+(0.294136137354)*x[0]+(0.732003937836)*x[1]**o+(0.633469131504)*x[1]
        arg[2,2,1,1]=(-0.427146039728)*x[0]**o+(-0.975737063903)*x[0]+(-0.934913653772)*x[1]**o+(-0.662568325005)*x[1]
        arg[2,2,2,0]=(0.477849916621)*x[0]**o+(-0.798883883367)*x[0]+(-0.154217684441)*x[1]**o+(-0.175817547912)*x[1]
        arg[2,2,2,1]=(0.935192620387)*x[0]**o+(-0.193802588249)*x[0]+(0.406181949213)*x[1]**o+(0.20422475396)*x[1]
        arg[2,3,0,0]=(0.965944278856)*x[0]**o+(-0.736175411725)*x[0]+(0.92404347386)*x[1]**o+(0.273152788392)*x[1]
        arg[2,3,0,1]=(-0.606944412413)*x[0]**o+(0.692555413476)*x[0]+(0.812085676275)*x[1]**o+(-0.0809920218886)*x[1]
        arg[2,3,1,0]=(-0.314634777406)*x[0]**o+(0.750181618592)*x[0]+(0.0554739779642)*x[1]**o+(-0.333788326726)*x[1]
        arg[2,3,1,1]=(0.85610815703)*x[0]**o+(0.910096005524)*x[0]+(-0.227092814843)*x[1]**o+(-0.0849532598914)*x[1]
        arg[2,3,2,0]=(0.735064128504)*x[0]**o+(-0.253919528943)*x[0]+(0.8457463307)*x[1]**o+(0.865727345925)*x[1]
        arg[2,3,2,1]=(0.520607989389)*x[0]**o+(0.569210317105)*x[0]+(0.500780081677)*x[1]**o+(-0.787630914351)*x[1]
        arg[2,4,0,0]=(0.179774694296)*x[0]**o+(-0.890872128607)*x[0]+(-0.185758177247)*x[1]**o+(0.888231734606)*x[1]
        arg[2,4,0,1]=(0.059863462227)*x[0]**o+(-0.0796463123948)*x[0]+(-0.104564520504)*x[1]**o+(-0.336182528617)*x[1]
        arg[2,4,1,0]=(-0.636828275967)*x[0]**o+(0.156894572218)*x[0]+(-0.884260096115)*x[1]**o+(-0.967431438795)*x[1]
        arg[2,4,1,1]=(-0.958168110389)*x[0]**o+(-0.0424772631477)*x[0]+(0.512363473316)*x[1]**o+(-0.716729419668)*x[1]
        arg[2,4,2,0]=(0.19736654467)*x[0]**o+(-0.288723518002)*x[0]+(0.243674007033)*x[1]**o+(0.911210601142)*x[1]
        arg[2,4,2,1]=(-0.847493343055)*x[0]**o+(0.115003265593)*x[0]+(0.360071936714)*x[1]**o+(-0.0265385061141)*x[1]
        arg[3,0,0,0]=(-0.510550590836)*x[0]**o+(-0.266174606631)*x[0]+(0.771580708159)*x[1]**o+(-0.770554374259)*x[1]
        arg[3,0,0,1]=(-0.65795024738)*x[0]**o+(-0.83394628525)*x[0]+(0.812461325452)*x[1]**o+(0.191593733482)*x[1]
        arg[3,0,1,0]=(0.136878718548)*x[0]**o+(-0.000345061692795)*x[0]+(-0.0780280205008)*x[1]**o+(0.173925537373)*x[1]
        arg[3,0,1,1]=(0.584079470149)*x[0]**o+(0.138674493948)*x[0]+(-0.563553909861)*x[1]**o+(-0.254304838008)*x[1]
        arg[3,0,2,0]=(0.326192887554)*x[0]**o+(0.993615980303)*x[0]+(-0.26995279693)*x[1]**o+(0.216112615287)*x[1]
        arg[3,0,2,1]=(0.684786021819)*x[0]**o+(0.158934017292)*x[0]+(-0.457448199191)*x[1]**o+(0.644722228163)*x[1]
        arg[3,1,0,0]=(-0.198252162262)*x[0]**o+(-0.821832378445)*x[0]+(0.831807441916)*x[1]**o+(-0.70643670471)*x[1]
        arg[3,1,0,1]=(0.361698759016)*x[0]**o+(0.556511107101)*x[0]+(0.892219700601)*x[1]**o+(-0.861598520361)*x[1]
        arg[3,1,1,0]=(-0.107946706634)*x[0]**o+(-0.910191934756)*x[0]+(0.502507607871)*x[1]**o+(0.649089810516)*x[1]
        arg[3,1,1,1]=(-0.92959760672)*x[0]**o+(0.950682520961)*x[0]+(0.103321004187)*x[1]**o+(-0.631562667815)*x[1]
        arg[3,1,2,0]=(0.858565242025)*x[0]**o+(-0.106810606075)*x[0]+(0.236377425048)*x[1]**o+(-0.296320858145)*x[1]
        arg[3,1,2,1]=(0.148558873625)*x[0]**o+(0.458488334082)*x[0]+(-0.0172824905523)*x[1]**o+(-0.321212766581)*x[1]
        arg[3,2,0,0]=(-0.239733890336)*x[0]**o+(0.558819103697)*x[0]+(-0.151683129885)*x[1]**o+(0.153995780645)*x[1]
        arg[3,2,0,1]=(-0.949100270937)*x[0]**o+(0.0687171499139)*x[0]+(0.968028021831)*x[1]**o+(-0.935600499345)*x[1]
        arg[3,2,1,0]=(-0.739823767237)*x[0]**o+(0.417825150389)*x[0]+(0.318612170707)*x[1]**o+(0.627557337421)*x[1]
        arg[3,2,1,1]=(-0.965142167324)*x[0]**o+(0.981446854383)*x[0]+(0.266191185951)*x[1]**o+(-0.581303142841)*x[1]
        arg[3,2,2,0]=(-0.00312606993817)*x[0]**o+(-0.191885030828)*x[0]+(-0.342525068561)*x[1]**o+(-0.953710538626)*x[1]
        arg[3,2,2,1]=(-0.689127187476)*x[0]**o+(-0.107190366063)*x[0]+(-0.0556716250594)*x[1]**o+(-0.609462166754)*x[1]
        arg[3,3,0,0]=(-0.765237197853)*x[0]**o+(-0.0707283944117)*x[0]+(0.891789872287)*x[1]**o+(-0.255939791858)*x[1]
        arg[3,3,0,1]=(-0.367658362103)*x[0]**o+(0.27406309422)*x[0]+(0.866288683947)*x[1]**o+(0.403194030287)*x[1]
        arg[3,3,1,0]=(-0.362795000993)*x[0]**o+(-0.498774415158)*x[0]+(0.0726808331583)*x[1]**o+(-0.871066791136)*x[1]
        arg[3,3,1,1]=(-0.177149059814)*x[0]**o+(0.287463753612)*x[0]+(0.101726559941)*x[1]**o+(0.561797204517)*x[1]
        arg[3,3,2,0]=(0.770615121039)*x[0]**o+(-0.503402148251)*x[0]+(0.319753007634)*x[1]**o+(0.586868894261)*x[1]
        arg[3,3,2,1]=(0.924125904724)*x[0]**o+(-0.508988920423)*x[0]+(-0.866752504824)*x[1]**o+(0.518544700652)*x[1]
        arg[3,4,0,0]=(-0.973660291124)*x[0]**o+(0.746869369283)*x[0]+(-0.634097171844)*x[1]**o+(0.723638944736)*x[1]
        arg[3,4,0,1]=(-0.599366869066)*x[0]**o+(-0.59893829688)*x[0]+(0.124483401866)*x[1]**o+(-0.924820315792)*x[1]
        arg[3,4,1,0]=(-0.707374787729)*x[0]**o+(-0.494748755022)*x[0]+(0.391752209698)*x[1]**o+(0.149710640116)*x[1]
        arg[3,4,1,1]=(0.329587748119)*x[0]**o+(-0.788289803764)*x[0]+(-0.145100163626)*x[1]**o+(0.141493810237)*x[1]
        arg[3,4,2,0]=(0.17496644337)*x[0]**o+(-0.270242104589)*x[0]+(0.996684379179)*x[1]**o+(-0.751697440046)*x[1]
        arg[3,4,2,1]=(0.79046837395)*x[0]**o+(0.568394473062)*x[0]+(0.359419854016)*x[1]**o+(0.630686948244)*x[1]
        ref[0,0,0,0]=(0.306552070969)/(o+1.)+(0.157870653629)
        ref[0,0,0,1]=(0.0606500807631)/(o+1.)+(0.0612449724389)
        ref[0,0,1,0]=(0.639564070889)/(o+1.)+(0.666341832555)
        ref[0,0,1,1]=(0.671248593147)/(o+1.)+(-0.528628572255)
        ref[0,0,2,0]=(0.0735824774293)/(o+1.)+(-0.474736472501)
        ref[0,0,2,1]=(-0.9070527228)/(o+1.)+(-0.197350261427)
        ref[0,1,0,0]=(1.30449671875)/(o+1.)+(0.106123067554)
        ref[0,1,0,1]=(-0.420072850657)/(o+1.)+(-0.407896954678)
        ref[0,1,1,0]=(1.44174379356)/(o+1.)+(-0.809395886627)
        ref[0,1,1,1]=(0.155435385684)/(o+1.)+(0.627148132992)
        ref[0,1,2,0]=(-0.137204404518)/(o+1.)+(0.207874473889)
        ref[0,1,2,1]=(-0.289709410261)/(o+1.)+(0.204293345671)
        ref[0,2,0,0]=(1.26583912644)/(o+1.)+(0.498805698213)
        ref[0,2,0,1]=(1.17560298985)/(o+1.)+(0.931845120886)
        ref[0,2,1,0]=(-0.344986302901)/(o+1.)+(0.220341743892)
        ref[0,2,1,1]=(-1.94726093397)/(o+1.)+(0.206668541494)
        ref[0,2,2,0]=(-0.91808772513)/(o+1.)+(-0.0252951560363)
        ref[0,2,2,1]=(-1.02863406132)/(o+1.)+(-0.20183443794)
        ref[0,3,0,0]=(-1.04834664255)/(o+1.)+(0.00604709752155)
        ref[0,3,0,1]=(0.2902816262)/(o+1.)+(0.514361224456)
        ref[0,3,1,0]=(-1.31568170019)/(o+1.)+(0.883603667426)
        ref[0,3,1,1]=(-0.208650456965)/(o+1.)+(0.889841082216)
        ref[0,3,2,0]=(-0.741632466554)/(o+1.)+(0.0759327240318)
        ref[0,3,2,1]=(1.48653223703)/(o+1.)+(-0.854483988026)
        ref[0,4,0,0]=(0.333816089225)/(o+1.)+(-0.562947445201)
        ref[0,4,0,1]=(0.130634932061)/(o+1.)+(0.199922965378)
        ref[0,4,1,0]=(1.27534940781)/(o+1.)+(0.158640649858)
        ref[0,4,1,1]=(-0.83666690538)/(o+1.)+(-0.344088772961)
        ref[0,4,2,0]=(-0.432609891704)/(o+1.)+(-0.267052816612)
        ref[0,4,2,1]=(-0.477920665494)/(o+1.)+(-0.305859149312)
        ref[1,0,0,0]=(-0.738766811663)/(o+1.)+(-0.2321755194)
        ref[1,0,0,1]=(-0.688095002235)/(o+1.)+(0.664163739465)
        ref[1,0,1,0]=(1.14285258401)/(o+1.)+(-0.345190031677)
        ref[1,0,1,1]=(-0.093656366746)/(o+1.)+(-0.236308341076)
        ref[1,0,2,0]=(0.7151363028)/(o+1.)+(-0.56709255572)
        ref[1,0,2,1]=(-0.51932492992)/(o+1.)+(0.250667355044)
        ref[1,1,0,0]=(0.545772982093)/(o+1.)+(-0.163104398138)
        ref[1,1,0,1]=(0.629573063174)/(o+1.)+(-0.174653201576)
        ref[1,1,1,0]=(-0.154012812311)/(o+1.)+(-0.914297866096)
        ref[1,1,1,1]=(-0.61132115559)/(o+1.)+(-0.44114113966)
        ref[1,1,2,0]=(0.408406410726)/(o+1.)+(-0.076080229207)
        ref[1,1,2,1]=(0.893038507335)/(o+1.)+(-0.263052908312)
        ref[1,2,0,0]=(0.17465338794)/(o+1.)+(0.391848465594)
        ref[1,2,0,1]=(-0.689339169821)/(o+1.)+(-0.460056198634)
        ref[1,2,1,0]=(1.31870632989)/(o+1.)+(-0.479414064332)
        ref[1,2,1,1]=(1.50891516069)/(o+1.)+(-0.0237454406495)
        ref[1,2,2,0]=(-0.453976065087)/(o+1.)+(-0.616348649278)
        ref[1,2,2,1]=(0.286587846307)/(o+1.)+(-0.0466708546809)
        ref[1,3,0,0]=(-0.380140989403)/(o+1.)+(-0.571048154745)
        ref[1,3,0,1]=(-0.883036620882)/(o+1.)+(-0.32753016973)
        ref[1,3,1,0]=(-0.60244085517)/(o+1.)+(-0.141401948721)
        ref[1,3,1,1]=(-0.853285111042)/(o+1.)+(0.342962640954)
        ref[1,3,2,0]=(-0.39249462323)/(o+1.)+(-0.390164886635)
        ref[1,3,2,1]=(-0.584758879441)/(o+1.)+(-0.00839083193621)
        ref[1,4,0,0]=(-0.282070155438)/(o+1.)+(-0.0156758086169)
        ref[1,4,0,1]=(0.472105372896)/(o+1.)+(-0.564324551903)
        ref[1,4,1,0]=(-1.31919056201)/(o+1.)+(-0.0984390311826)
        ref[1,4,1,1]=(-0.0914548911124)/(o+1.)+(0.0630846675525)
        ref[1,4,2,0]=(-0.327958330666)/(o+1.)+(0.657758815411)
        ref[1,4,2,1]=(-0.364384803274)/(o+1.)+(0.503617623946)
        ref[2,0,0,0]=(0.586128081557)/(o+1.)+(0.124182733972)
        ref[2,0,0,1]=(-1.08667878312)/(o+1.)+(0.187878007169)
        ref[2,0,1,0]=(0.181949317955)/(o+1.)+(0.341666123665)
        ref[2,0,1,1]=(1.37450614338)/(o+1.)+(-0.329280804596)
        ref[2,0,2,0]=(-1.79681719902)/(o+1.)+(0.31696617459)
        ref[2,0,2,1]=(-0.331005168235)/(o+1.)+(0.608933603219)
        ref[2,1,0,0]=(-0.608003218053)/(o+1.)+(0.303644850697)
        ref[2,1,0,1]=(0.0259152842714)/(o+1.)+(0.916297858039)
        ref[2,1,1,0]=(-0.33777899696)/(o+1.)+(-0.0168105795836)
        ref[2,1,1,1]=(-0.987531479864)/(o+1.)+(0.417880159899)
        ref[2,1,2,0]=(-0.426800950806)/(o+1.)+(0.0654268452787)
        ref[2,1,2,1]=(-0.570210924898)/(o+1.)+(-0.392672308506)
        ref[2,2,0,0]=(-0.489515474032)/(o+1.)+(0.516699346832)
        ref[2,2,0,1]=(-0.386977868298)/(o+1.)+(0.054797545679)
        ref[2,2,1,0]=(1.23803129631)/(o+1.)+(0.463802634429)
        ref[2,2,1,1]=(-1.3620596935)/(o+1.)+(-0.819152694454)
        ref[2,2,2,0]=(0.32363223218)/(o+1.)+(-0.48735071564)
        ref[2,2,2,1]=(1.3413745696)/(o+1.)+(0.00521108285542)
        ref[2,3,0,0]=(1.88998775272)/(o+1.)+(-0.231511311666)
        ref[2,3,0,1]=(0.205141263861)/(o+1.)+(0.305781695794)
        ref[2,3,1,0]=(-0.259160799442)/(o+1.)+(0.208196645933)
        ref[2,3,1,1]=(0.629015342187)/(o+1.)+(0.412571372816)
        ref[2,3,2,0]=(1.5808104592)/(o+1.)+(0.305903908491)
        ref[2,3,2,1]=(1.02138807107)/(o+1.)+(-0.109210298623)
        ref[2,4,0,0]=(-0.00598348295078)/(o+1.)+(-0.00132019700044)
        ref[2,4,0,1]=(-0.0447010582767)/(o+1.)+(-0.207914420506)
        ref[2,4,1,0]=(-1.52108837208)/(o+1.)+(-0.405268433289)
        ref[2,4,1,1]=(-0.445804637073)/(o+1.)+(-0.379603341408)
        ref[2,4,2,0]=(0.441040551703)/(o+1.)+(0.31124354157)
        ref[2,4,2,1]=(-0.487421406342)/(o+1.)+(0.0442323797396)
        ref[3,0,0,0]=(0.261030117323)/(o+1.)+(-0.518364490445)
        ref[3,0,0,1]=(0.154511078072)/(o+1.)+(-0.321176275884)
        ref[3,0,1,0]=(0.0588506980476)/(o+1.)+(0.0867902378403)
        ref[3,0,1,1]=(0.0205255602871)/(o+1.)+(-0.0578151720304)
        ref[3,0,2,0]=(0.0562400906237)/(o+1.)+(0.604864297795)
        ref[3,0,2,1]=(0.227337822629)/(o+1.)+(0.401828122727)
        ref[3,1,0,0]=(0.633555279654)/(o+1.)+(-0.764134541578)
        ref[3,1,0,1]=(1.25391845962)/(o+1.)+(-0.15254370663)
        ref[3,1,1,0]=(0.394560901238)/(o+1.)+(-0.13055106212)
        ref[3,1,1,1]=(-0.826276602534)/(o+1.)+(0.159559926573)
        ref[3,1,2,0]=(1.09494266707)/(o+1.)+(-0.20156573211)
        ref[3,1,2,1]=(0.131276383073)/(o+1.)+(0.0686377837506)
        ref[3,2,0,0]=(-0.391417020221)/(o+1.)+(0.356407442171)
        ref[3,2,0,1]=(0.0189277508933)/(o+1.)+(-0.433441674716)
        ref[3,2,1,0]=(-0.42121159653)/(o+1.)+(0.522691243905)
        ref[3,2,1,1]=(-0.698950981373)/(o+1.)+(0.200071855771)
        ref[3,2,2,0]=(-0.345651138499)/(o+1.)+(-0.572797784727)
        ref[3,2,2,1]=(-0.744798812536)/(o+1.)+(-0.358326266408)
        ref[3,3,0,0]=(0.126552674434)/(o+1.)+(-0.163334093135)
        ref[3,3,0,1]=(0.498630321844)/(o+1.)+(0.338628562253)
        ref[3,3,1,0]=(-0.290114167835)/(o+1.)+(-0.684920603147)
        ref[3,3,1,1]=(-0.0754224998727)/(o+1.)+(0.424630479064)
        ref[3,3,2,0]=(1.09036812867)/(o+1.)+(0.0417333730049)
        ref[3,3,2,1]=(0.0573733998999)/(o+1.)+(0.00477789011464)
        ref[3,4,0,0]=(-1.60775746297)/(o+1.)+(0.73525415701)
        ref[3,4,0,1]=(-0.4748834672)/(o+1.)+(-0.761879306336)
        ref[3,4,1,0]=(-0.31562257803)/(o+1.)+(-0.172519057453)
        ref[3,4,1,1]=(0.184487584493)/(o+1.)+(-0.323397996764)
        ref[3,4,2,0]=(1.17165082255)/(o+1.)+(-0.510969772317)
        ref[3,4,2,1]=(1.14988822797)/(o+1.)+(0.599540710653)
      else:
        arg[0,0,0,0]=(-0.517412938774)*x[0]**o+(0.824804313111)*x[0]+(-0.681706563374)*x[1]**o+(-0.259540440272)*x[1]+(0.876735283937)*x[2]**o+(0.307072859807)*x[2]
        arg[0,0,0,1]=(-0.545981202251)*x[0]**o+(-0.530080933212)*x[0]+(-0.339289111599)*x[1]**o+(-0.726311605088)*x[1]+(0.334837705444)*x[2]**o+(-0.422667311064)*x[2]
        arg[0,0,1,0]=(-0.94261387328)*x[0]**o+(-0.315825326734)*x[0]+(0.542502801277)*x[1]**o+(-0.927525943646)*x[1]+(-0.802118238213)*x[2]**o+(-0.106512314706)*x[2]
        arg[0,0,1,1]=(-0.877940099286)*x[0]**o+(-0.801256557869)*x[0]+(-0.155983792242)*x[1]**o+(0.200547191496)*x[1]+(-0.960553133432)*x[2]**o+(0.968862792367)*x[2]
        arg[0,0,2,0]=(-0.870342547946)*x[0]**o+(-0.956447402661)*x[0]+(-0.506797803004)*x[1]**o+(0.342841675107)*x[1]+(0.528354159344)*x[2]**o+(-0.824546729564)*x[2]
        arg[0,0,2,1]=(-0.411899137296)*x[0]**o+(0.133736458559)*x[0]+(0.852726131619)*x[1]**o+(0.692795229813)*x[1]+(0.047372576403)*x[2]**o+(0.810485765945)*x[2]
        arg[0,1,0,0]=(0.961758233254)*x[0]**o+(-0.606307924964)*x[0]+(0.0612696412665)*x[1]**o+(-0.947823541802)*x[1]+(-0.113021104365)*x[2]**o+(-0.00802440322155)*x[2]
        arg[0,1,0,1]=(-0.684097052467)*x[0]**o+(0.467140835537)*x[0]+(-0.0742537091205)*x[1]**o+(0.541641916143)*x[1]+(0.987491548505)*x[2]**o+(-0.558380262255)*x[2]
        arg[0,1,1,0]=(-0.901821988855)*x[0]**o+(0.34443714654)*x[0]+(-0.682552654793)*x[1]**o+(0.398276086461)*x[1]+(0.334644264965)*x[2]**o+(0.21454845251)*x[2]
        arg[0,1,1,1]=(0.929614060295)*x[0]**o+(-0.322609307947)*x[0]+(-0.426962200724)*x[1]**o+(0.464659992564)*x[1]+(-0.13983863674)*x[2]**o+(-0.681648817325)*x[2]
        arg[0,1,2,0]=(0.504166508914)*x[0]**o+(-0.978562859041)*x[0]+(-0.769650404806)*x[1]**o+(0.443500137153)*x[1]+(0.969669965936)*x[2]**o+(0.401742207403)*x[2]
        arg[0,1,2,1]=(-0.478561324822)*x[0]**o+(-0.772019838834)*x[0]+(-0.779194844829)*x[1]**o+(-0.747980400065)*x[1]+(0.827885608192)*x[2]**o+(0.135571151528)*x[2]
        arg[0,2,0,0]=(-0.210426221052)*x[0]**o+(-0.464938141903)*x[0]+(-0.776844066203)*x[1]**o+(0.267244555705)*x[1]+(-0.773135929429)*x[2]**o+(-0.00846060888954)*x[2]
        arg[0,2,0,1]=(0.0945858334531)*x[0]**o+(-0.492400895713)*x[0]+(-0.128564669727)*x[1]**o+(-0.0951458189983)*x[1]+(-0.54426715378)*x[2]**o+(0.106545346072)*x[2]
        arg[0,2,1,0]=(-0.482424571392)*x[0]**o+(0.0431208155652)*x[0]+(0.557138601636)*x[1]**o+(-0.725564282551)*x[1]+(0.9081107419)*x[2]**o+(0.0179201585352)*x[2]
        arg[0,2,1,1]=(-0.555731853052)*x[0]**o+(0.447137254671)*x[0]+(0.820958717703)*x[1]**o+(-0.580251434504)*x[1]+(0.227884622965)*x[2]**o+(0.34460189453)*x[2]
        arg[0,2,2,0]=(-0.25868983898)*x[0]**o+(0.759165881486)*x[0]+(-0.499903058341)*x[1]**o+(0.896471093639)*x[1]+(0.351955520633)*x[2]**o+(0.16566756566)*x[2]
        arg[0,2,2,1]=(-0.37859354794)*x[0]**o+(0.925387188779)*x[0]+(0.0111055293719)*x[1]**o+(-0.311998642154)*x[1]+(0.441009182046)*x[2]**o+(0.0968247918271)*x[2]
        arg[0,3,0,0]=(-0.461665900641)*x[0]**o+(0.812002051396)*x[0]+(-0.767456852027)*x[1]**o+(0.809069781575)*x[1]+(0.690416953885)*x[2]**o+(0.171662369216)*x[2]
        arg[0,3,0,1]=(-0.804811324639)*x[0]**o+(-0.0962711523954)*x[0]+(0.980130059825)*x[1]**o+(0.98279180603)*x[1]+(-0.0614353559054)*x[2]**o+(0.645637153538)*x[2]
        arg[0,3,1,0]=(0.97878415369)*x[0]**o+(-0.676321251698)*x[0]+(0.916075195717)*x[1]**o+(0.244509683371)*x[1]+(-0.939563193145)*x[2]**o+(-0.866506108913)*x[2]
        arg[0,3,1,1]=(-0.676965895162)*x[0]**o+(-0.124355338519)*x[0]+(0.680151101195)*x[1]**o+(-0.839412433156)*x[1]+(0.434088587223)*x[2]**o+(-0.275286048166)*x[2]
        arg[0,3,2,0]=(0.414822536626)*x[0]**o+(0.493836498194)*x[0]+(0.967077097745)*x[1]**o+(-0.0115301157869)*x[1]+(-0.307994657929)*x[2]**o+(-0.967766608312)*x[2]
        arg[0,3,2,1]=(-0.34064670348)*x[0]**o+(-0.975121872905)*x[0]+(0.0839457874766)*x[1]**o+(0.0995106978636)*x[1]+(-0.533837237516)*x[2]**o+(0.771268044302)*x[2]
        arg[0,4,0,0]=(0.223197544637)*x[0]**o+(0.712738871576)*x[0]+(-0.386972075747)*x[1]**o+(0.289829507744)*x[1]+(0.153418793839)*x[2]**o+(0.876087048492)*x[2]
        arg[0,4,0,1]=(0.696750948653)*x[0]**o+(0.466905335953)*x[0]+(-0.611476041907)*x[1]**o+(-0.939902457901)*x[1]+(-0.752675432265)*x[2]**o+(-0.869502717044)*x[2]
        arg[0,4,1,0]=(-0.465562201723)*x[0]**o+(-0.388716352051)*x[0]+(0.260044093503)*x[1]**o+(0.919379163294)*x[1]+(-0.902453394797)*x[2]**o+(0.378318744505)*x[2]
        arg[0,4,1,1]=(0.121990842946)*x[0]**o+(-0.691267894153)*x[0]+(-0.703960568755)*x[1]**o+(-0.582404896428)*x[1]+(0.57560180036)*x[2]**o+(0.537099448589)*x[2]
        arg[0,4,2,0]=(-0.495452196513)*x[0]**o+(-0.939389935669)*x[0]+(-0.237205717598)*x[1]**o+(0.61376516207)*x[1]+(-0.506082738728)*x[2]**o+(0.407704148542)*x[2]
        arg[0,4,2,1]=(0.450976180771)*x[0]**o+(-0.00739258898437)*x[0]+(0.210130848093)*x[1]**o+(0.624639007009)*x[1]+(0.173795671727)*x[2]**o+(0.508269170853)*x[2]
        arg[1,0,0,0]=(0.609685942177)*x[0]**o+(0.609299494609)*x[0]+(0.321046388121)*x[1]**o+(0.598197189498)*x[1]+(0.536092177618)*x[2]**o+(0.576427431157)*x[2]
        arg[1,0,0,1]=(0.412500706234)*x[0]**o+(0.905942239017)*x[0]+(-0.974980343169)*x[1]**o+(-0.131245997607)*x[1]+(0.745809068018)*x[2]**o+(0.543470258153)*x[2]
        arg[1,0,1,0]=(-0.903032545947)*x[0]**o+(-0.778693321051)*x[0]+(0.399342164105)*x[1]**o+(0.478212896767)*x[1]+(0.254596269802)*x[2]**o+(0.70308382001)*x[2]
        arg[1,0,1,1]=(-0.281529506793)*x[0]**o+(0.74166036027)*x[0]+(0.629713747725)*x[1]**o+(-0.275164746597)*x[1]+(0.37523909979)*x[2]**o+(-0.878026427978)*x[2]
        arg[1,0,2,0]=(0.236027764061)*x[0]**o+(-0.816910805083)*x[0]+(0.0782011624185)*x[1]**o+(-0.644004245666)*x[1]+(0.293586462602)*x[2]**o+(0.557045894695)*x[2]
        arg[1,0,2,1]=(0.868061727269)*x[0]**o+(0.53329318943)*x[0]+(0.367603915235)*x[1]**o+(0.89280277925)*x[1]+(-0.316595424948)*x[2]**o+(0.593830565147)*x[2]
        arg[1,1,0,0]=(0.886950595996)*x[0]**o+(-0.290323112286)*x[0]+(-0.863347030089)*x[1]**o+(0.545096994541)*x[1]+(0.338200944281)*x[2]**o+(-0.79898415048)*x[2]
        arg[1,1,0,1]=(0.103761462006)*x[0]**o+(-0.046916161864)*x[0]+(0.0312549239467)*x[1]**o+(0.214045118984)*x[1]+(-0.523161766402)*x[2]**o+(0.714207028317)*x[2]
        arg[1,1,1,0]=(0.494943183987)*x[0]**o+(-0.0543096386006)*x[0]+(-0.263875120588)*x[1]**o+(0.776031489925)*x[1]+(0.881224752145)*x[2]**o+(-0.650038283228)*x[2]
        arg[1,1,1,1]=(0.0871887540513)*x[0]**o+(-0.953848841349)*x[0]+(-0.223751459589)*x[1]**o+(0.227400823993)*x[1]+(-0.224200553009)*x[2]**o+(0.594012970899)*x[2]
        arg[1,1,2,0]=(-0.443116951346)*x[0]**o+(0.891806149626)*x[0]+(0.94731530917)*x[1]**o+(0.739499894809)*x[1]+(-0.92613053703)*x[2]**o+(0.464883087528)*x[2]
        arg[1,1,2,1]=(-0.149831145559)*x[0]**o+(0.913999704597)*x[0]+(0.76082978152)*x[1]**o+(-0.441535347622)*x[1]+(0.902841586936)*x[2]**o+(-0.509022004948)*x[2]
        arg[1,2,0,0]=(0.508230188489)*x[0]**o+(0.241669921231)*x[0]+(-0.802811365056)*x[1]**o+(0.298424305948)*x[1]+(-0.949407454532)*x[2]**o+(-0.131426880214)*x[2]
        arg[1,2,0,1]=(0.65372313274)*x[0]**o+(0.195891642771)*x[0]+(-0.160845429219)*x[1]**o+(0.500815823724)*x[1]+(0.875785249339)*x[2]**o+(0.325024339211)*x[2]
        arg[1,2,1,0]=(0.908405682536)*x[0]**o+(-0.146675166477)*x[0]+(0.24033815873)*x[1]**o+(-0.62814219771)*x[1]+(-0.512365788351)*x[2]**o+(-0.103817191415)*x[2]
        arg[1,2,1,1]=(0.824305237501)*x[0]**o+(0.175753992729)*x[0]+(0.992502114126)*x[1]**o+(0.973110627555)*x[1]+(0.355352002177)*x[2]**o+(0.112252263955)*x[2]
        arg[1,2,2,0]=(-0.485996285003)*x[0]**o+(-0.532551123887)*x[0]+(0.708205121405)*x[1]**o+(-0.803269038441)*x[1]+(-0.232049124534)*x[2]**o+(0.55437652866)*x[2]
        arg[1,2,2,1]=(0.551934776224)*x[0]**o+(0.727629664538)*x[0]+(0.197393511049)*x[1]**o+(-0.146619812174)*x[1]+(0.316592306936)*x[2]**o+(0.0263881493809)*x[2]
        arg[1,3,0,0]=(-0.918554166254)*x[0]**o+(-0.839150758884)*x[0]+(0.436713305329)*x[1]**o+(-0.410131731387)*x[1]+(0.321246028791)*x[2]**o+(0.706389384376)*x[2]
        arg[1,3,0,1]=(-0.870168753295)*x[0]**o+(-0.00473364206025)*x[0]+(0.594648540734)*x[1]**o+(-0.924887214906)*x[1]+(0.58858894515)*x[2]**o+(0.0152152159566)*x[2]
        arg[1,3,1,0]=(0.64005604234)*x[0]**o+(0.308265334836)*x[0]+(0.321645661286)*x[1]**o+(0.670017614875)*x[1]+(0.712764136703)*x[2]**o+(0.458986964342)*x[2]
        arg[1,3,1,1]=(-0.201376578944)*x[0]**o+(-0.271793865118)*x[0]+(-0.798181331582)*x[1]**o+(0.690100877945)*x[1]+(0.600350295622)*x[2]**o+(0.341066858436)*x[2]
        arg[1,3,2,0]=(0.459832113137)*x[0]**o+(0.0566780273877)*x[0]+(0.221723079112)*x[1]**o+(-0.681001728734)*x[1]+(0.265505135729)*x[2]**o+(-0.389365671367)*x[2]
        arg[1,3,2,1]=(0.715444384493)*x[0]**o+(-0.705697520786)*x[0]+(0.945208081248)*x[1]**o+(0.0233171867808)*x[1]+(0.740156166802)*x[2]**o+(-0.144314337012)*x[2]
        arg[1,4,0,0]=(-0.490162545806)*x[0]**o+(0.542981268396)*x[0]+(-0.841471465564)*x[1]**o+(0.79091798393)*x[1]+(0.245598142884)*x[2]**o+(0.4740296042)*x[2]
        arg[1,4,0,1]=(-0.504504615322)*x[0]**o+(0.45494067251)*x[0]+(-0.612427248728)*x[1]**o+(0.0880340051743)*x[1]+(0.645906227711)*x[2]**o+(0.772038416622)*x[2]
        arg[1,4,1,0]=(0.945538632492)*x[0]**o+(0.98519603491)*x[0]+(-0.245495451859)*x[1]**o+(-0.70732864026)*x[1]+(0.378733662069)*x[2]**o+(0.579554415018)*x[2]
        arg[1,4,1,1]=(0.886659064398)*x[0]**o+(-0.491870874438)*x[0]+(0.751814742427)*x[1]**o+(0.430589443767)*x[1]+(-0.729193959143)*x[2]**o+(-0.699344758538)*x[2]
        arg[1,4,2,0]=(-0.4190140481)*x[0]**o+(0.84509612452)*x[0]+(-0.263608911857)*x[1]**o+(0.693648282371)*x[1]+(-0.144463932166)*x[2]**o+(0.384293689266)*x[2]
        arg[1,4,2,1]=(-0.266973780787)*x[0]**o+(0.613988367707)*x[0]+(0.686955199789)*x[1]**o+(-0.301585818785)*x[1]+(-0.574601202558)*x[2]**o+(-0.430718616071)*x[2]
        arg[2,0,0,0]=(-0.117390727803)*x[0]**o+(-0.25850139145)*x[0]+(0.324909086031)*x[1]**o+(-0.593112241553)*x[1]+(-0.51016768177)*x[2]**o+(0.392299415334)*x[2]
        arg[2,0,0,1]=(0.290478531696)*x[0]**o+(0.758153023626)*x[0]+(0.531003993361)*x[1]**o+(-0.112481510374)*x[1]+(0.950222158203)*x[2]**o+(-0.742480526204)*x[2]
        arg[2,0,1,0]=(-0.754820921225)*x[0]**o+(-0.408482044077)*x[0]+(-0.229697426486)*x[1]**o+(-0.983010424713)*x[1]+(-0.286618604257)*x[2]**o+(-0.489034108451)*x[2]
        arg[2,0,1,1]=(0.96775757822)*x[0]**o+(0.797765725721)*x[0]+(-0.90936176648)*x[1]**o+(-0.372755885193)*x[1]+(-0.144727874143)*x[2]**o+(0.341624944551)*x[2]
        arg[2,0,2,0]=(0.672779422278)*x[0]**o+(-0.936102224964)*x[0]+(-0.763620352857)*x[1]**o+(0.265802509744)*x[1]+(0.275798875047)*x[2]**o+(0.235397520942)*x[2]
        arg[2,0,2,1]=(0.192866248041)*x[0]**o+(0.656863606616)*x[0]+(-0.902796948663)*x[1]**o+(0.692123089256)*x[1]+(0.311287987388)*x[2]**o+(0.555427507546)*x[2]
        arg[2,1,0,0]=(0.464585537593)*x[0]**o+(0.161744615208)*x[0]+(-0.325569920662)*x[1]**o+(-0.47142644441)*x[1]+(0.885151021784)*x[2]**o+(-0.840595219086)*x[2]
        arg[2,1,0,1]=(-0.659253397488)*x[0]**o+(0.54391314862)*x[0]+(-0.595612529952)*x[1]**o+(-0.0472758114379)*x[1]+(0.876391930188)*x[2]**o+(-0.440313199946)*x[2]
        arg[2,1,1,0]=(-0.445591423364)*x[0]**o+(0.179144841652)*x[0]+(-0.294166502086)*x[1]**o+(-0.373820778739)*x[1]+(0.464609314863)*x[2]**o+(0.407893781978)*x[2]
        arg[2,1,1,1]=(-0.767512700255)*x[0]**o+(0.63198312168)*x[0]+(0.652318598902)*x[1]**o+(0.817874971823)*x[1]+(0.397211574885)*x[2]**o+(0.33174354886)*x[2]
        arg[2,1,2,0]=(0.473801282131)*x[0]**o+(-0.778895701394)*x[0]+(-0.801488721847)*x[1]**o+(-0.585840161723)*x[1]+(-0.805496721223)*x[2]**o+(0.585971840421)*x[2]
        arg[2,1,2,1]=(0.472329620891)*x[0]**o+(-0.0473805663117)*x[0]+(0.0894117727522)*x[1]**o+(-0.308473625733)*x[1]+(-0.658376953956)*x[2]**o+(0.32027592346)*x[2]
        arg[2,2,0,0]=(0.843241796369)*x[0]**o+(0.346857959798)*x[0]+(0.0111223078269)*x[1]**o+(-0.949623227267)*x[1]+(-0.152362393376)*x[2]**o+(-0.886340001141)*x[2]
        arg[2,2,0,1]=(-0.555393018762)*x[0]**o+(0.666364344488)*x[0]+(0.181237902671)*x[1]**o+(0.670959040345)*x[1]+(0.57383285317)*x[2]**o+(-0.451027832775)*x[2]
        arg[2,2,1,0]=(0.249709446431)*x[0]**o+(0.568753063379)*x[0]+(-0.812386750097)*x[1]**o+(0.506805124672)*x[1]+(-0.454844217662)*x[2]**o+(0.843435269288)*x[2]
        arg[2,2,1,1]=(0.753144930025)*x[0]**o+(0.0181642998269)*x[0]+(0.622635241593)*x[1]**o+(-0.372883842874)*x[1]+(0.366920965577)*x[2]**o+(-0.092391524816)*x[2]
        arg[2,2,2,0]=(0.309601906611)*x[0]**o+(-0.502573192965)*x[0]+(-0.420795137333)*x[1]**o+(-0.807109356307)*x[1]+(-0.802887583918)*x[2]**o+(0.552609684817)*x[2]
        arg[2,2,2,1]=(0.320933223342)*x[0]**o+(0.541931544704)*x[0]+(0.988715316931)*x[1]**o+(-0.4197325237)*x[1]+(-0.438705512435)*x[2]**o+(-0.540175093599)*x[2]
        arg[2,3,0,0]=(0.018870015518)*x[0]**o+(-0.615120691773)*x[0]+(0.596295280542)*x[1]**o+(0.210874310884)*x[1]+(-0.652194502321)*x[2]**o+(0.74019575754)*x[2]
        arg[2,3,0,1]=(0.820285141622)*x[0]**o+(0.130985741886)*x[0]+(-0.397387767472)*x[1]**o+(0.975103051154)*x[1]+(0.0887369214611)*x[2]**o+(0.88498731668)*x[2]
        arg[2,3,1,0]=(0.160503336808)*x[0]**o+(0.641180113144)*x[0]+(-0.408109675737)*x[1]**o+(-0.355443951133)*x[1]+(-0.429622016495)*x[2]**o+(0.998729833779)*x[2]
        arg[2,3,1,1]=(-0.646524352112)*x[0]**o+(-0.289894880059)*x[0]+(0.909761879864)*x[1]**o+(0.30453651619)*x[1]+(0.884596709091)*x[2]**o+(0.329408737003)*x[2]
        arg[2,3,2,0]=(0.459385919432)*x[0]**o+(0.416447271916)*x[0]+(-0.400798659819)*x[1]**o+(-0.769043818187)*x[1]+(-0.631682400803)*x[2]**o+(-0.80150361084)*x[2]
        arg[2,3,2,1]=(-0.544824298969)*x[0]**o+(0.191301965562)*x[0]+(0.41061102918)*x[1]**o+(-0.999195759064)*x[1]+(-0.0226105975802)*x[2]**o+(-0.137078691725)*x[2]
        arg[2,4,0,0]=(-0.624033952375)*x[0]**o+(0.0792222332007)*x[0]+(0.766756718455)*x[1]**o+(0.581487379511)*x[1]+(-0.437940391245)*x[2]**o+(0.115617341492)*x[2]
        arg[2,4,0,1]=(-0.300526068838)*x[0]**o+(-0.822844066866)*x[0]+(0.660949310367)*x[1]**o+(-0.0993949179102)*x[1]+(0.599884615655)*x[2]**o+(-0.903177033523)*x[2]
        arg[2,4,1,0]=(-0.654642517091)*x[0]**o+(0.428391621765)*x[0]+(0.932377441464)*x[1]**o+(-0.459525189684)*x[1]+(0.667425199675)*x[2]**o+(-0.937213410686)*x[2]
        arg[2,4,1,1]=(0.258067525957)*x[0]**o+(0.619943768383)*x[0]+(0.458217567511)*x[1]**o+(0.0812191081554)*x[1]+(0.743260637332)*x[2]**o+(-0.796150472293)*x[2]
        arg[2,4,2,0]=(0.782473636556)*x[0]**o+(-0.441405613305)*x[0]+(0.3951595327)*x[1]**o+(-0.449855781187)*x[1]+(-0.964130179212)*x[2]**o+(-0.299199410525)*x[2]
        arg[2,4,2,1]=(-0.878340584397)*x[0]**o+(-0.674575785429)*x[0]+(0.838137461257)*x[1]**o+(-0.498710475715)*x[1]+(-0.117205127994)*x[2]**o+(0.0934669704542)*x[2]
        arg[3,0,0,0]=(0.139631366137)*x[0]**o+(-0.893001605238)*x[0]+(-0.420780773096)*x[1]**o+(0.439055863393)*x[1]+(0.173113363237)*x[2]**o+(0.893867443017)*x[2]
        arg[3,0,0,1]=(-0.804823042124)*x[0]**o+(0.250231181298)*x[0]+(0.358464048439)*x[1]**o+(-0.847391231172)*x[1]+(-0.515391975323)*x[2]**o+(-0.999640302937)*x[2]
        arg[3,0,1,0]=(0.328676968943)*x[0]**o+(-0.0521242139559)*x[0]+(0.861548411537)*x[1]**o+(-0.288979961428)*x[1]+(-0.927901849987)*x[2]**o+(0.857869958971)*x[2]
        arg[3,0,1,1]=(-0.654866876537)*x[0]**o+(-0.0515462321916)*x[0]+(-0.224164685097)*x[1]**o+(-0.84207399757)*x[1]+(-0.442392040219)*x[2]**o+(-0.915734076984)*x[2]
        arg[3,0,2,0]=(-0.322517406382)*x[0]**o+(-0.757146499161)*x[0]+(-0.541690197465)*x[1]**o+(0.931090695413)*x[1]+(-0.355505263744)*x[2]**o+(-0.773185404514)*x[2]
        arg[3,0,2,1]=(0.421030649833)*x[0]**o+(-0.83768939688)*x[0]+(-0.863791071447)*x[1]**o+(0.78393205545)*x[1]+(-0.489893911681)*x[2]**o+(-0.291989245709)*x[2]
        arg[3,1,0,0]=(0.550839051007)*x[0]**o+(-0.883853789938)*x[0]+(-0.339529971923)*x[1]**o+(-0.756936847344)*x[1]+(0.724888858656)*x[2]**o+(-0.981057762224)*x[2]
        arg[3,1,0,1]=(0.711792743156)*x[0]**o+(0.233248429378)*x[0]+(0.526076154408)*x[1]**o+(0.349364804166)*x[1]+(0.271475992486)*x[2]**o+(0.782429402713)*x[2]
        arg[3,1,1,0]=(-0.499072848146)*x[0]**o+(-0.0604888692308)*x[0]+(0.545442893272)*x[1]**o+(-0.688797522515)*x[1]+(0.305341295222)*x[2]**o+(0.130010601331)*x[2]
        arg[3,1,1,1]=(0.940695359452)*x[0]**o+(0.113015472294)*x[0]+(-0.506254395321)*x[1]**o+(-0.252044277039)*x[1]+(0.0172784972094)*x[2]**o+(-0.184256978193)*x[2]
        arg[3,1,2,0]=(0.550472143981)*x[0]**o+(-0.729809590776)*x[0]+(-0.293765999495)*x[1]**o+(-0.378042976721)*x[1]+(0.973242167993)*x[2]**o+(-0.75350391134)*x[2]
        arg[3,1,2,1]=(0.369266178598)*x[0]**o+(-0.764034492366)*x[0]+(-0.713325992328)*x[1]**o+(0.948932375717)*x[1]+(-0.335074977323)*x[2]**o+(0.117829591745)*x[2]
        arg[3,2,0,0]=(0.426458076699)*x[0]**o+(-0.520551329627)*x[0]+(-0.773464312884)*x[1]**o+(0.26332468989)*x[1]+(0.143497828773)*x[2]**o+(0.0676278275986)*x[2]
        arg[3,2,0,1]=(0.85806016531)*x[0]**o+(0.755361032731)*x[0]+(0.403942188112)*x[1]**o+(0.0113665816202)*x[1]+(-0.439835633198)*x[2]**o+(0.620930145183)*x[2]
        arg[3,2,1,0]=(-0.852838624821)*x[0]**o+(-0.918594624263)*x[0]+(-0.572898555139)*x[1]**o+(-0.14353390918)*x[1]+(-0.716010071307)*x[2]**o+(0.134687635305)*x[2]
        arg[3,2,1,1]=(-0.14282590746)*x[0]**o+(-0.169314199229)*x[0]+(0.880745511071)*x[1]**o+(0.540733710441)*x[1]+(-0.131424658303)*x[2]**o+(-0.93209431119)*x[2]
        arg[3,2,2,0]=(-0.864016075688)*x[0]**o+(-0.265794760131)*x[0]+(0.444625188371)*x[1]**o+(0.133693703972)*x[1]+(-0.749799597936)*x[2]**o+(0.886355820615)*x[2]
        arg[3,2,2,1]=(0.676281241167)*x[0]**o+(-0.351800298325)*x[0]+(-0.751498380099)*x[1]**o+(-0.561737786104)*x[1]+(-0.984872672853)*x[2]**o+(0.476840947707)*x[2]
        arg[3,3,0,0]=(0.457874366022)*x[0]**o+(-0.126929649966)*x[0]+(0.666077559139)*x[1]**o+(-0.966437669224)*x[1]+(-0.849000502816)*x[2]**o+(0.990790747926)*x[2]
        arg[3,3,0,1]=(0.45136859326)*x[0]**o+(0.771355456911)*x[0]+(-0.678558472676)*x[1]**o+(0.0366869703264)*x[1]+(-0.257247358952)*x[2]**o+(-0.290852285538)*x[2]
        arg[3,3,1,0]=(-0.623007069616)*x[0]**o+(-0.365255080044)*x[0]+(0.441927420573)*x[1]**o+(0.331640625664)*x[1]+(-0.21153951456)*x[2]**o+(0.297967669123)*x[2]
        arg[3,3,1,1]=(-0.836057742118)*x[0]**o+(0.959873293328)*x[0]+(-0.52034405502)*x[1]**o+(-0.178838686591)*x[1]+(0.317610563306)*x[2]**o+(-0.555134650833)*x[2]
        arg[3,3,2,0]=(0.0819108022024)*x[0]**o+(0.69432296821)*x[0]+(-0.23685434367)*x[1]**o+(-0.0442137744391)*x[1]+(-0.879175086057)*x[2]**o+(-0.28031950012)*x[2]
        arg[3,3,2,1]=(0.340083286763)*x[0]**o+(-0.381942838645)*x[0]+(0.570579443852)*x[1]**o+(0.772192708863)*x[1]+(0.42813556157)*x[2]**o+(0.735683065499)*x[2]
        arg[3,4,0,0]=(-0.866813692418)*x[0]**o+(-0.53715159706)*x[0]+(0.0725900211922)*x[1]**o+(0.0891167848614)*x[1]+(-0.381482922347)*x[2]**o+(-0.521504123034)*x[2]
        arg[3,4,0,1]=(0.487750922723)*x[0]**o+(0.967820473849)*x[0]+(0.305600886352)*x[1]**o+(0.350777657643)*x[1]+(-0.383731485788)*x[2]**o+(0.814016126476)*x[2]
        arg[3,4,1,0]=(-0.186663244344)*x[0]**o+(0.144872424379)*x[0]+(0.844454339304)*x[1]**o+(0.252133708483)*x[1]+(-0.748235508222)*x[2]**o+(0.702952185343)*x[2]
        arg[3,4,1,1]=(-0.962296047927)*x[0]**o+(0.874277032826)*x[0]+(-0.404689253219)*x[1]**o+(-0.174075799004)*x[1]+(0.65086350106)*x[2]**o+(-0.109850650677)*x[2]
        arg[3,4,2,0]=(-0.59515836027)*x[0]**o+(-0.0917296352911)*x[0]+(0.834779238353)*x[1]**o+(0.951252871777)*x[1]+(0.983008684497)*x[2]**o+(-0.784806954891)*x[2]
        arg[3,4,2,1]=(0.603110971708)*x[0]**o+(0.291507042984)*x[0]+(-0.401523286981)*x[1]**o+(-0.994760476862)*x[1]+(0.410516007475)*x[2]**o+(0.368982230001)*x[2]
        ref[0,0,0,0]=(-0.322384218211)/(o+1.)+(0.436168366323)
        ref[0,0,0,1]=(-0.550432608405)/(o+1.)+(-0.839529924682)
        ref[0,0,1,0]=(-1.20222931022)/(o+1.)+(-0.674931792543)
        ref[0,0,1,1]=(-1.99447702496)/(o+1.)+(0.184076712997)
        ref[0,0,2,0]=(-0.848786191606)/(o+1.)+(-0.719076228559)
        ref[0,0,2,1]=(0.488199570726)/(o+1.)+(0.818508727159)
        ref[0,1,0,0]=(0.910006770156)/(o+1.)+(-0.781077934994)
        ref[0,1,0,1]=(0.229140786917)/(o+1.)+(0.225201244712)
        ref[0,1,1,0]=(-1.24973037868)/(o+1.)+(0.478630842756)
        ref[0,1,1,1]=(0.362813222831)/(o+1.)+(-0.269799066354)
        ref[0,1,2,0]=(0.704186070044)/(o+1.)+(-0.0666602572428)
        ref[0,1,2,1]=(-0.429870561459)/(o+1.)+(-0.692214543685)
        ref[0,2,0,0]=(-1.76040621668)/(o+1.)+(-0.103077097544)
        ref[0,2,0,1]=(-0.578245990054)/(o+1.)+(-0.24050068432)
        ref[0,2,1,0]=(0.982824772144)/(o+1.)+(-0.332261654225)
        ref[0,2,1,1]=(0.493111487615)/(o+1.)+(0.105743857349)
        ref[0,2,2,0]=(-0.406637376689)/(o+1.)+(0.910652270393)
        ref[0,2,2,1]=(0.0735211634779)/(o+1.)+(0.355106669226)
        ref[0,3,0,0]=(-0.538705798782)/(o+1.)+(0.896367101094)
        ref[0,3,0,1]=(0.11388337928)/(o+1.)+(0.766078903586)
        ref[0,3,1,0]=(0.955296156262)/(o+1.)+(-0.64915883862)
        ref[0,3,1,1]=(0.437273793256)/(o+1.)+(-0.61952690992)
        ref[0,3,2,0]=(1.07390497644)/(o+1.)+(-0.242730112953)
        ref[0,3,2,1]=(-0.79053815352)/(o+1.)+(-0.0521715653697)
        ref[0,4,0,0]=(-0.0103557372704)/(o+1.)+(0.939327713906)
        ref[0,4,0,1]=(-0.66740052552)/(o+1.)+(-0.671249919496)
        ref[0,4,1,0]=(-1.10797150302)/(o+1.)+(0.454490777873)
        ref[0,4,1,1]=(-0.00636792544909)/(o+1.)+(-0.368286670996)
        ref[0,4,2,0]=(-1.23874065284)/(o+1.)+(0.0410396874718)
        ref[0,4,2,1]=(0.834902700591)/(o+1.)+(0.562757794439)
        ref[1,0,0,0]=(1.46682450791)/(o+1.)+(0.891962057632)
        ref[1,0,0,1]=(0.183329431084)/(o+1.)+(0.659083249781)
        ref[1,0,1,0]=(-0.24909411204)/(o+1.)+(0.201301697863)
        ref[1,0,1,1]=(0.723423340723)/(o+1.)+(-0.205765407152)
        ref[1,0,2,0]=(0.607815389082)/(o+1.)+(-0.451934578027)
        ref[1,0,2,1]=(0.919070217555)/(o+1.)+(1.00996326691)
        ref[1,1,0,0]=(0.361804510188)/(o+1.)+(-0.272105134113)
        ref[1,1,0,1]=(-0.388145380449)/(o+1.)+(0.440667992719)
        ref[1,1,1,0]=(1.11229281554)/(o+1.)+(0.0358417840478)
        ref[1,1,1,1]=(-0.360763258547)/(o+1.)+(-0.0662175232282)
        ref[1,1,2,0]=(-0.421932179206)/(o+1.)+(1.04809456598)
        ref[1,1,2,1]=(1.5138402229)/(o+1.)+(-0.0182788239862)
        ref[1,2,0,0]=(-1.2439886311)/(o+1.)+(0.204333673483)
        ref[1,2,0,1]=(1.36866295286)/(o+1.)+(0.510865902853)
        ref[1,2,1,0]=(0.636378052915)/(o+1.)+(-0.439317277801)
        ref[1,2,1,1]=(2.1721593538)/(o+1.)+(0.63055844212)
        ref[1,2,2,0]=(-0.00984028813171)/(o+1.)+(-0.390721816834)
        ref[1,2,2,1]=(1.06592059421)/(o+1.)+(0.303699000873)
        ref[1,3,0,0]=(-0.160594832134)/(o+1.)+(-0.271446552947)
        ref[1,3,0,1]=(0.313068732589)/(o+1.)+(-0.457202820505)
        ref[1,3,1,0]=(1.67446584033)/(o+1.)+(0.718634957026)
        ref[1,3,1,1]=(-0.399207614903)/(o+1.)+(0.379686935632)
        ref[1,3,2,0]=(0.947060327978)/(o+1.)+(-0.506844686357)
        ref[1,3,2,1]=(2.40080863254)/(o+1.)+(-0.413347335508)
        ref[1,4,0,0]=(-1.08603586849)/(o+1.)+(0.903964428263)
        ref[1,4,0,1]=(-0.471025636339)/(o+1.)+(0.657506547153)
        ref[1,4,1,0]=(1.0787768427)/(o+1.)+(0.428710904834)
        ref[1,4,1,1]=(0.909279847682)/(o+1.)+(-0.380313094604)
        ref[1,4,2,0]=(-0.827086892123)/(o+1.)+(0.961519048079)
        ref[1,4,2,1]=(-0.154619783556)/(o+1.)+(-0.0591580335741)
        ref[2,0,0,0]=(-0.302649323542)/(o+1.)+(-0.229657108834)
        ref[2,0,0,1]=(1.77170468326)/(o+1.)+(-0.0484045064764)
        ref[2,0,1,0]=(-1.27113695197)/(o+1.)+(-0.94026328862)
        ref[2,0,1,1]=(-0.0863320624027)/(o+1.)+(0.383317392539)
        ref[2,0,2,0]=(0.184957944467)/(o+1.)+(-0.217451097139)
        ref[2,0,2,1]=(-0.398642713235)/(o+1.)+(0.952207101709)
        ref[2,1,0,0]=(1.02416663872)/(o+1.)+(-0.575138524144)
        ref[2,1,0,1]=(-0.378473997252)/(o+1.)+(0.0281620686178)
        ref[2,1,1,0]=(-0.275148610588)/(o+1.)+(0.106608922446)
        ref[2,1,1,1]=(0.282017473532)/(o+1.)+(0.890800821181)
        ref[2,1,2,0]=(-1.13318416094)/(o+1.)+(-0.389382011348)
        ref[2,1,2,1]=(-0.0966355603128)/(o+1.)+(-0.0177891342927)
        ref[2,2,0,0]=(0.70200171082)/(o+1.)+(-0.744552634305)
        ref[2,2,0,1]=(0.199677737079)/(o+1.)+(0.443147776029)
        ref[2,2,1,0]=(-1.01752152133)/(o+1.)+(0.959496728669)
        ref[2,2,1,1]=(1.74270113719)/(o+1.)+(-0.223555533932)
        ref[2,2,2,0]=(-0.91408081464)/(o+1.)+(-0.378536432227)
        ref[2,2,2,1]=(0.870943027839)/(o+1.)+(-0.208988036298)
        ref[2,3,0,0]=(-0.0370292062615)/(o+1.)+(0.167974688326)
        ref[2,3,0,1]=(0.51163429561)/(o+1.)+(0.99553805486)
        ref[2,3,1,0]=(-0.677228355424)/(o+1.)+(0.642232997895)
        ref[2,3,1,1]=(1.14783423684)/(o+1.)+(0.172025186567)
        ref[2,3,2,0]=(-0.57309514119)/(o+1.)+(-0.577050078555)
        ref[2,3,2,1]=(-0.156823867369)/(o+1.)+(-0.472486242613)
        ref[2,4,0,0]=(-0.295217625165)/(o+1.)+(0.388163477102)
        ref[2,4,0,1]=(0.960307857184)/(o+1.)+(-0.91270800915)
        ref[2,4,1,0]=(0.945160124047)/(o+1.)+(-0.484173489303)
        ref[2,4,1,1]=(1.4595457308)/(o+1.)+(-0.0474937978771)
        ref[2,4,2,0]=(0.213502990044)/(o+1.)+(-0.595230402509)
        ref[2,4,2,1]=(-0.157408251134)/(o+1.)+(-0.539909645345)
        ref[3,0,0,0]=(-0.108036043722)/(o+1.)+(0.219960850586)
        ref[3,0,0,1]=(-0.961750969008)/(o+1.)+(-0.798400176406)
        ref[3,0,1,0]=(0.262323530493)/(o+1.)+(0.258382891794)
        ref[3,0,1,1]=(-1.32142360185)/(o+1.)+(-0.904677153373)
        ref[3,0,2,0]=(-1.21971286759)/(o+1.)+(-0.299620604131)
        ref[3,0,2,1]=(-0.932654333294)/(o+1.)+(-0.17287329357)
        ref[3,1,0,0]=(0.93619793774)/(o+1.)+(-1.31092419975)
        ref[3,1,0,1]=(1.50934489005)/(o+1.)+(0.682521318129)
        ref[3,1,1,0]=(0.351711340348)/(o+1.)+(-0.309637895207)
        ref[3,1,1,1]=(0.451719461341)/(o+1.)+(-0.161642891469)
        ref[3,1,2,0]=(1.22994831248)/(o+1.)+(-0.930678239419)
        ref[3,1,2,1]=(-0.679134791053)/(o+1.)+(0.151363737548)
        ref[3,2,0,0]=(-0.203508407413)/(o+1.)+(-0.0947994060691)
        ref[3,2,0,1]=(0.822166720224)/(o+1.)+(0.693828879767)
        ref[3,2,1,0]=(-2.14174725127)/(o+1.)+(-0.463720449069)
        ref[3,2,1,1]=(0.606494945308)/(o+1.)+(-0.280337399989)
        ref[3,2,2,0]=(-1.16919048525)/(o+1.)+(0.377127382228)
        ref[3,2,2,1]=(-1.06008981178)/(o+1.)+(-0.218348568361)
        ref[3,3,0,0]=(0.274951422345)/(o+1.)+(-0.0512882856318)
        ref[3,3,0,1]=(-0.484437238367)/(o+1.)+(0.25859507085)
        ref[3,3,1,0]=(-0.392619163603)/(o+1.)+(0.132176607371)
        ref[3,3,1,1]=(-1.03879123383)/(o+1.)+(0.112949977952)
        ref[3,3,2,0]=(-1.03411862752)/(o+1.)+(0.184894846825)
        ref[3,3,2,1]=(1.33879829218)/(o+1.)+(0.562966467859)
        ref[3,4,0,0]=(-1.17570659357)/(o+1.)+(-0.484769467616)
        ref[3,4,0,1]=(0.409620323287)/(o+1.)+(1.06630712898)
        ref[3,4,1,0]=(-0.0904444132627)/(o+1.)+(0.549979159103)
        ref[3,4,1,1]=(-0.716121800087)/(o+1.)+(0.295175291572)
        ref[3,4,2,0]=(1.22262956258)/(o+1.)+(0.0373581407971)
        ref[3,4,2,1]=(0.612103692201)/(o+1.)+(-0.167135601938)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.640162884104)*x[0]**o+(-0.686170347303)*x[0]+(0.0531418393824)*x[1]**o+(-0.0288840863965)*x[1]
        ref=(-0.587021044721)/(o+1.)+(-0.35752721685)
      else:
        s=(0.835059122976)*x[0]**o+(-0.876192510448)*x[0]+(0.229125107737)*x[1]**o+(-0.378923833918)*x[1]+(0.734383439043)*x[2]**o+(0.987756783011)*x[2]
        ref=(1.79856766975)/(o+1.)+(-0.133679780678)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.292137623974)*x[0]**o+(0.75950865395)*x[0]+(-0.205837757637)*x[1]**o+(0.90443054103)*x[1]
        s[1]=(0.407997882771)*x[0]**o+(0.335693471916)*x[0]+(0.755336837856)*x[1]**o+(0.218679559306)*x[1]
        ref[0]=(-0.497975381612)/(o+1.)+(0.83196959749)
        ref[1]=(1.16333472063)/(o+1.)+(0.277186515611)
      else:
        s[0]=(-0.393835090841)*x[0]**o+(-0.474828597107)*x[0]+(0.465782088981)*x[1]**o+(-0.882630637032)*x[1]+(0.849599914303)*x[2]**o+(0.737421732808)*x[2]
        s[1]=(-0.589146617652)*x[0]**o+(-0.412139119114)*x[0]+(0.970406630484)*x[1]**o+(0.363942463909)*x[1]+(-0.372574721521)*x[2]**o+(-0.798081163974)*x[2]
        ref[0]=(0.921546912444)/(o+1.)+(-0.310018750665)
        ref[1]=(0.00868529131054)/(o+1.)+(-0.42313890959)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.542564643665)*x[0]**o+(0.869139399504)*x[0]+(-0.918464318956)*x[1]**o+(0.258557879364)*x[1]
        s[0,1]=(0.797108788422)*x[0]**o+(-0.200368785173)*x[0]+(-0.518597299854)*x[1]**o+(-0.647852667874)*x[1]
        s[0,2]=(0.692223658744)*x[0]**o+(-0.103586429107)*x[0]+(0.282165531986)*x[1]**o+(-0.191815912482)*x[1]
        s[0,3]=(-0.370167730862)*x[0]**o+(-0.822713005318)*x[0]+(-0.00317012401339)*x[1]**o+(-0.848121245241)*x[1]
        s[0,4]=(-0.52733376174)*x[0]**o+(-0.314966701977)*x[0]+(0.755974759737)*x[1]**o+(-0.483207203205)*x[1]
        s[1,0]=(-0.709714610043)*x[0]**o+(0.647231907198)*x[0]+(-0.577474021122)*x[1]**o+(-0.124504757353)*x[1]
        s[1,1]=(-0.723690611716)*x[0]**o+(-0.617424586153)*x[0]+(-0.536447933353)*x[1]**o+(0.255892253884)*x[1]
        s[1,2]=(-0.201052994401)*x[0]**o+(-0.519971533366)*x[0]+(0.515314592901)*x[1]**o+(0.831783913039)*x[1]
        s[1,3]=(0.505769313912)*x[0]**o+(-0.460235423669)*x[0]+(0.982458353968)*x[1]**o+(0.479022475503)*x[1]
        s[1,4]=(0.624910824579)*x[0]**o+(0.563956018517)*x[0]+(0.0129379080194)*x[1]**o+(-0.634601977633)*x[1]
        s[2,0]=(-0.765921383834)*x[0]**o+(0.0199045688425)*x[0]+(-0.0666443330576)*x[1]**o+(-0.929035862007)*x[1]
        s[2,1]=(0.698423948822)*x[0]**o+(-0.529311289681)*x[0]+(-0.563339072497)*x[1]**o+(-0.494985449663)*x[1]
        s[2,2]=(0.101812328764)*x[0]**o+(-0.922547259485)*x[0]+(-0.45388897835)*x[1]**o+(0.357458826452)*x[1]
        s[2,3]=(0.643691149734)*x[0]**o+(-0.791395473264)*x[0]+(-0.919285047298)*x[1]**o+(-0.732314921151)*x[1]
        s[2,4]=(0.770509137995)*x[0]**o+(-0.399023743738)*x[0]+(-0.0887239237676)*x[1]**o+(0.846766840231)*x[1]
        s[3,0]=(-0.258188916476)*x[0]**o+(-0.192762172857)*x[0]+(-0.291360534186)*x[1]**o+(-0.828483492861)*x[1]
        s[3,1]=(-0.180563503871)*x[0]**o+(0.411356882323)*x[0]+(0.144166964346)*x[1]**o+(-0.817750491386)*x[1]
        s[3,2]=(-0.56008111033)*x[0]**o+(-0.903908126304)*x[0]+(-0.795710525329)*x[1]**o+(0.42957467817)*x[1]
        s[3,3]=(-0.435823321725)*x[0]**o+(-0.687506280092)*x[0]+(0.512172892719)*x[1]**o+(0.624268656549)*x[1]
        s[3,4]=(0.0995998270974)*x[0]**o+(0.54337587883)*x[0]+(-0.589282344971)*x[1]**o+(0.551709444954)*x[1]
        ref[0,0]=(-0.375899675291)/(o+1.)+(0.563848639434)
        ref[0,1]=(0.278511488568)/(o+1.)+(-0.424110726524)
        ref[0,2]=(0.97438919073)/(o+1.)+(-0.147701170794)
        ref[0,3]=(-0.373337854876)/(o+1.)+(-0.835417125279)
        ref[0,4]=(0.228640997997)/(o+1.)+(-0.399086952591)
        ref[1,0]=(-1.28718863116)/(o+1.)+(0.261363574923)
        ref[1,1]=(-1.26013854507)/(o+1.)+(-0.180766166134)
        ref[1,2]=(0.3142615985)/(o+1.)+(0.155906189837)
        ref[1,3]=(1.48822766788)/(o+1.)+(0.00939352591692)
        ref[1,4]=(0.637848732599)/(o+1.)+(-0.0353229795581)
        ref[2,0]=(-0.832565716891)/(o+1.)+(-0.454565646582)
        ref[2,1]=(0.135084876326)/(o+1.)+(-0.512148369672)
        ref[2,2]=(-0.352076649587)/(o+1.)+(-0.282544216517)
        ref[2,3]=(-0.275593897564)/(o+1.)+(-0.761855197208)
        ref[2,4]=(0.681785214228)/(o+1.)+(0.223871548247)
        ref[3,0]=(-0.549549450662)/(o+1.)+(-0.510622832859)
        ref[3,1]=(-0.0363965395244)/(o+1.)+(-0.203196804532)
        ref[3,2]=(-1.35579163566)/(o+1.)+(-0.237166724067)
        ref[3,3]=(0.0763495709944)/(o+1.)+(-0.0316188117717)
        ref[3,4]=(-0.489682517873)/(o+1.)+(0.547542661892)
      else:
        s[0,0]=(-0.547777970569)*x[0]**o+(-0.644569779184)*x[0]+(-0.247447574507)*x[1]**o+(-0.165674972238)*x[1]+(0.492011982518)*x[2]**o+(0.824373100652)*x[2]
        s[0,1]=(0.714120627471)*x[0]**o+(-0.343344240726)*x[0]+(0.846874965081)*x[1]**o+(0.955417725709)*x[1]+(-0.462347625853)*x[2]**o+(0.236518016678)*x[2]
        s[0,2]=(0.649286582806)*x[0]**o+(-0.712764590386)*x[0]+(-0.364435855331)*x[1]**o+(0.529624579838)*x[1]+(-0.478005452391)*x[2]**o+(0.79676822654)*x[2]
        s[0,3]=(0.186538430487)*x[0]**o+(0.873009863028)*x[0]+(-0.617246224023)*x[1]**o+(0.212670710234)*x[1]+(-0.568924203263)*x[2]**o+(-0.591868989872)*x[2]
        s[0,4]=(0.0764403834746)*x[0]**o+(0.580116472215)*x[0]+(-0.964832163438)*x[1]**o+(0.146621202039)*x[1]+(0.303667547632)*x[2]**o+(-0.572109268772)*x[2]
        s[1,0]=(0.0466497527173)*x[0]**o+(-0.660397185235)*x[0]+(0.314235627776)*x[1]**o+(-0.0219053335847)*x[1]+(0.0341017375209)*x[2]**o+(-0.138388505919)*x[2]
        s[1,1]=(-0.306253609605)*x[0]**o+(-0.274894845016)*x[0]+(0.590915613652)*x[1]**o+(0.466423547973)*x[1]+(-0.65224024787)*x[2]**o+(0.011478276889)*x[2]
        s[1,2]=(-0.850811667442)*x[0]**o+(-0.362611692783)*x[0]+(-0.132198189858)*x[1]**o+(0.613225855964)*x[1]+(-0.995844863385)*x[2]**o+(-0.00228774595319)*x[2]
        s[1,3]=(-0.331021120468)*x[0]**o+(0.262316324101)*x[0]+(0.693647465307)*x[1]**o+(-0.758663328061)*x[1]+(0.563291863323)*x[2]**o+(-0.464786381036)*x[2]
        s[1,4]=(-0.297063926313)*x[0]**o+(0.258694451069)*x[0]+(0.453986223203)*x[1]**o+(0.516652862875)*x[1]+(-0.119453447306)*x[2]**o+(0.13369866553)*x[2]
        s[2,0]=(0.895304992051)*x[0]**o+(-0.581521556328)*x[0]+(-0.102534354406)*x[1]**o+(-0.312191328622)*x[1]+(0.00625028874943)*x[2]**o+(0.455967398752)*x[2]
        s[2,1]=(0.0129094351309)*x[0]**o+(-0.284360799298)*x[0]+(0.254610787505)*x[1]**o+(0.232923077121)*x[1]+(0.884382417671)*x[2]**o+(0.879710026011)*x[2]
        s[2,2]=(0.978354808904)*x[0]**o+(0.45417061753)*x[0]+(-0.617831296441)*x[1]**o+(-0.540161947193)*x[1]+(0.749817874967)*x[2]**o+(0.85276098284)*x[2]
        s[2,3]=(0.790036783517)*x[0]**o+(-0.461944315129)*x[0]+(0.0177930616139)*x[1]**o+(0.827267186021)*x[1]+(-0.5681970366)*x[2]**o+(0.0800529373343)*x[2]
        s[2,4]=(0.320372577784)*x[0]**o+(-0.988843069534)*x[0]+(0.366830399498)*x[1]**o+(0.0749365557309)*x[1]+(-0.964416326732)*x[2]**o+(0.0589434543979)*x[2]
        s[3,0]=(-0.143156890467)*x[0]**o+(0.665143613007)*x[0]+(0.324524292861)*x[1]**o+(0.573292071297)*x[1]+(0.55975986068)*x[2]**o+(-0.703161304253)*x[2]
        s[3,1]=(0.827768123756)*x[0]**o+(-0.565184841493)*x[0]+(0.724124619351)*x[1]**o+(-0.375933144184)*x[1]+(-0.186573364344)*x[2]**o+(0.174303241336)*x[2]
        s[3,2]=(0.19505934899)*x[0]**o+(-0.378308488179)*x[0]+(-0.542115440091)*x[1]**o+(-0.864956237152)*x[1]+(0.0489116293809)*x[2]**o+(0.149786871599)*x[2]
        s[3,3]=(-0.151225736973)*x[0]**o+(-0.491229793833)*x[0]+(-0.704529608083)*x[1]**o+(-0.790051884866)*x[1]+(-0.202633581578)*x[2]**o+(0.0769118984781)*x[2]
        s[3,4]=(0.523527337562)*x[0]**o+(0.266094832079)*x[0]+(-0.147028009699)*x[1]**o+(-0.683770701764)*x[1]+(0.794443408406)*x[2]**o+(-0.0536824425549)*x[2]
        ref[0,0]=(-0.303213562559)/(o+1.)+(0.00706417461518)
        ref[0,1]=(1.0986479667)/(o+1.)+(0.42429575083)
        ref[0,2]=(-0.193154724916)/(o+1.)+(0.306814107996)
        ref[0,3]=(-0.999631996799)/(o+1.)+(0.246905791695)
        ref[0,4]=(-0.584724232331)/(o+1.)+(0.0773142027411)
        ref[1,0]=(0.394987118014)/(o+1.)+(-0.410345512369)
        ref[1,1]=(-0.367578243823)/(o+1.)+(0.101503489923)
        ref[1,2]=(-1.97885472069)/(o+1.)+(0.124163208614)
        ref[1,3]=(0.925918208162)/(o+1.)+(-0.480566692498)
        ref[1,4]=(0.0374688495844)/(o+1.)+(0.454522989737)
        ref[2,0]=(0.799020926395)/(o+1.)+(-0.218872743099)
        ref[2,1]=(1.15190264031)/(o+1.)+(0.414136151917)
        ref[2,2]=(1.11034138743)/(o+1.)+(0.383384826588)
        ref[2,3]=(0.239632808531)/(o+1.)+(0.222687904113)
        ref[2,4]=(-0.27721334945)/(o+1.)+(-0.427481529703)
        ref[3,0]=(0.741127263073)/(o+1.)+(0.267637190026)
        ref[3,1]=(1.36531937876)/(o+1.)+(-0.383407372171)
        ref[3,2]=(-0.29814446172)/(o+1.)+(-0.546738926866)
        ref[3,3]=(-1.05838892663)/(o+1.)+(-0.602184890111)
        ref[3,4]=(1.17094273627)/(o+1.)+(-0.23567915612)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.630010583318)*x[0]**o+(0.0494796640212)*x[0]+(0.471665362166)*x[1]**o+(0.392459033784)*x[1]
        s[0,0,1]=(-0.423295675063)*x[0]**o+(0.24240367131)*x[0]+(-0.724581249901)*x[1]**o+(-0.131295119605)*x[1]
        s[0,1,0]=(-0.410928342692)*x[0]**o+(-0.541418037237)*x[0]+(-0.232546576693)*x[1]**o+(-0.0784119675747)*x[1]
        s[0,1,1]=(0.789988066595)*x[0]**o+(-0.243498836345)*x[0]+(0.368561675765)*x[1]**o+(-0.747610528524)*x[1]
        s[1,0,0]=(-0.326293541804)*x[0]**o+(0.101066389163)*x[0]+(-0.184367851)*x[1]**o+(0.568769737008)*x[1]
        s[1,0,1]=(-0.015503180143)*x[0]**o+(0.819464848547)*x[0]+(0.979976262017)*x[1]**o+(-0.0389216169058)*x[1]
        s[1,1,0]=(-0.542766421816)*x[0]**o+(-0.373053336375)*x[0]+(-0.938010294052)*x[1]**o+(-0.399667401896)*x[1]
        s[1,1,1]=(0.0200936228416)*x[0]**o+(0.770519603961)*x[0]+(0.620955090369)*x[1]**o+(-0.330894822252)*x[1]
        s[2,0,0]=(-0.171145765903)*x[0]**o+(0.612313985433)*x[0]+(0.0479528062441)*x[1]**o+(-0.281058939492)*x[1]
        s[2,0,1]=(-0.332826833153)*x[0]**o+(0.944629304493)*x[0]+(-0.487576018254)*x[1]**o+(-0.904875326977)*x[1]
        s[2,1,0]=(-0.00808555466364)*x[0]**o+(0.245367446684)*x[0]+(-0.164397692869)*x[1]**o+(-0.198574049753)*x[1]
        s[2,1,1]=(-0.0469077421177)*x[0]**o+(0.89449436367)*x[0]+(-0.802161009378)*x[1]**o+(-0.0767107312174)*x[1]
        s[3,0,0]=(0.530561159497)*x[0]**o+(-0.0909174483134)*x[0]+(0.159443779335)*x[1]**o+(-0.178020813545)*x[1]
        s[3,0,1]=(-0.216531776351)*x[0]**o+(-0.531806095078)*x[0]+(0.653531437331)*x[1]**o+(0.21339193674)*x[1]
        s[3,1,0]=(0.254391857637)*x[0]**o+(0.619393713071)*x[0]+(-0.260306661561)*x[1]**o+(-0.473743696717)*x[1]
        s[3,1,1]=(-0.921367458187)*x[0]**o+(-0.69212037605)*x[0]+(0.6216551239)*x[1]**o+(0.7963578436)*x[1]
        s[4,0,0]=(-0.506208138693)*x[0]**o+(-0.103718686465)*x[0]+(0.771724985573)*x[1]**o+(0.815749872065)*x[1]
        s[4,0,1]=(-0.21355375924)*x[0]**o+(0.635638481649)*x[0]+(0.274581141313)*x[1]**o+(-0.947357183419)*x[1]
        s[4,1,0]=(0.35658463006)*x[0]**o+(-0.351876832623)*x[0]+(-0.0999414279107)*x[1]**o+(0.212979499129)*x[1]
        s[4,1,1]=(0.494190940488)*x[0]**o+(0.0851596543145)*x[0]+(-0.433440340709)*x[1]**o+(-0.0173667523032)*x[1]
        s[5,0,0]=(0.53169691054)*x[0]**o+(0.952395156793)*x[0]+(-0.127903378991)*x[1]**o+(-0.605192547481)*x[1]
        s[5,0,1]=(0.795535841954)*x[0]**o+(-0.615332450079)*x[0]+(0.764586784938)*x[1]**o+(0.420251114523)*x[1]
        s[5,1,0]=(-0.699722370082)*x[0]**o+(0.608404015259)*x[0]+(-0.166384830092)*x[1]**o+(-0.0611871373695)*x[1]
        s[5,1,1]=(0.545721411662)*x[0]**o+(-0.426115151805)*x[0]+(-0.438291880099)*x[1]**o+(0.0686028790393)*x[1]
        ref[0,0,0]=(1.10167594548)/(o+1.)+(0.220969348903)
        ref[0,0,1]=(-1.14787692496)/(o+1.)+(0.055554275852)
        ref[0,1,0]=(-0.643474919385)/(o+1.)+(-0.309915002406)
        ref[0,1,1]=(1.15854974236)/(o+1.)+(-0.495554682434)
        ref[1,0,0]=(-0.510661392804)/(o+1.)+(0.334918063085)
        ref[1,0,1]=(0.964473081874)/(o+1.)+(0.390271615821)
        ref[1,1,0]=(-1.48077671587)/(o+1.)+(-0.386360369136)
        ref[1,1,1]=(0.64104871321)/(o+1.)+(0.219812390854)
        ref[2,0,0]=(-0.123192959659)/(o+1.)+(0.16562752297)
        ref[2,0,1]=(-0.820402851406)/(o+1.)+(0.0198769887578)
        ref[2,1,0]=(-0.172483247533)/(o+1.)+(0.0233966984659)
        ref[2,1,1]=(-0.849068751496)/(o+1.)+(0.408891816226)
        ref[3,0,0]=(0.690004938832)/(o+1.)+(-0.134469130929)
        ref[3,0,1]=(0.43699966098)/(o+1.)+(-0.159207079169)
        ref[3,1,0]=(-0.00591480392415)/(o+1.)+(0.0728250081768)
        ref[3,1,1]=(-0.299712334287)/(o+1.)+(0.0521187337749)
        ref[4,0,0]=(0.265516846879)/(o+1.)+(0.3560155928)
        ref[4,0,1]=(0.0610273820724)/(o+1.)+(-0.155859350885)
        ref[4,1,0]=(0.256643202149)/(o+1.)+(-0.0694486667468)
        ref[4,1,1]=(0.0607505997785)/(o+1.)+(0.0338964510057)
        ref[5,0,0]=(0.403793531549)/(o+1.)+(0.173601304656)
        ref[5,0,1]=(1.56012262689)/(o+1.)+(-0.0975406677777)
        ref[5,1,0]=(-0.866107200175)/(o+1.)+(0.273608438945)
        ref[5,1,1]=(0.107429531563)/(o+1.)+(-0.178756136383)
      else:
        s[0,0,0]=(0.584550224653)*x[0]**o+(-0.106569322848)*x[0]+(0.728893354651)*x[1]**o+(0.237029650377)*x[1]+(-0.546491966553)*x[2]**o+(0.0390698430932)*x[2]
        s[0,0,1]=(0.0636227774465)*x[0]**o+(0.934881844205)*x[0]+(-0.545166942655)*x[1]**o+(0.08148757249)*x[1]+(-0.179461185461)*x[2]**o+(0.641018091939)*x[2]
        s[0,1,0]=(0.325463516488)*x[0]**o+(-0.620730155784)*x[0]+(0.154991661017)*x[1]**o+(0.245046216313)*x[1]+(0.645726837044)*x[2]**o+(-0.230093713841)*x[2]
        s[0,1,1]=(-0.321016164146)*x[0]**o+(-0.0509943423674)*x[0]+(-0.479999557029)*x[1]**o+(0.569709724552)*x[1]+(0.190221146753)*x[2]**o+(-0.919477787005)*x[2]
        s[1,0,0]=(0.929113150774)*x[0]**o+(-0.733802947668)*x[0]+(0.8108269457)*x[1]**o+(-0.488163027335)*x[1]+(-0.745575620351)*x[2]**o+(-0.486955144862)*x[2]
        s[1,0,1]=(-0.017329684129)*x[0]**o+(0.147982077718)*x[0]+(-0.144958156543)*x[1]**o+(-0.300572864309)*x[1]+(0.00787918928353)*x[2]**o+(-0.532824168573)*x[2]
        s[1,1,0]=(0.685828903913)*x[0]**o+(-0.567114040704)*x[0]+(-0.0471407827804)*x[1]**o+(0.511415127351)*x[1]+(-0.554086444893)*x[2]**o+(-0.226435378121)*x[2]
        s[1,1,1]=(-0.415793464117)*x[0]**o+(-0.777707665965)*x[0]+(-0.846630968003)*x[1]**o+(0.389596880295)*x[1]+(-0.213682650766)*x[2]**o+(-0.0851017843465)*x[2]
        s[2,0,0]=(0.932561239744)*x[0]**o+(0.0615709255597)*x[0]+(0.634288279118)*x[1]**o+(0.293277319848)*x[1]+(-0.429481015254)*x[2]**o+(0.778577347672)*x[2]
        s[2,0,1]=(0.857123773566)*x[0]**o+(-0.47446916241)*x[0]+(0.951211790713)*x[1]**o+(0.443028899356)*x[1]+(0.0861361311835)*x[2]**o+(0.597277083437)*x[2]
        s[2,1,0]=(0.874693704519)*x[0]**o+(-0.920294692148)*x[0]+(-0.03228863342)*x[1]**o+(0.248843294667)*x[1]+(-0.926343823806)*x[2]**o+(0.15109837142)*x[2]
        s[2,1,1]=(0.840368685645)*x[0]**o+(-0.200563324888)*x[0]+(0.88039674632)*x[1]**o+(-0.349512401298)*x[1]+(0.25915053789)*x[2]**o+(-0.48356523029)*x[2]
        s[3,0,0]=(0.758393510867)*x[0]**o+(0.666093057943)*x[0]+(-0.453617931299)*x[1]**o+(-0.134152843992)*x[1]+(-0.0117598390465)*x[2]**o+(-0.730981155782)*x[2]
        s[3,0,1]=(0.205572845432)*x[0]**o+(0.46935055342)*x[0]+(-0.415086164167)*x[1]**o+(0.276709685737)*x[1]+(0.341889576347)*x[2]**o+(0.170610214566)*x[2]
        s[3,1,0]=(0.926107056617)*x[0]**o+(-0.928101461122)*x[0]+(-0.639162904468)*x[1]**o+(-0.787501002441)*x[1]+(0.615156638997)*x[2]**o+(0.444030323472)*x[2]
        s[3,1,1]=(0.458304153272)*x[0]**o+(0.119383772473)*x[0]+(-0.816833779036)*x[1]**o+(0.428980054603)*x[1]+(0.176196071727)*x[2]**o+(0.294498083663)*x[2]
        s[4,0,0]=(-0.184477855098)*x[0]**o+(-0.281312358692)*x[0]+(-0.259869292769)*x[1]**o+(0.670877642172)*x[1]+(0.710140723846)*x[2]**o+(-0.738455024569)*x[2]
        s[4,0,1]=(0.790994543706)*x[0]**o+(0.594490423343)*x[0]+(0.737641957747)*x[1]**o+(-0.974154697219)*x[1]+(-0.228328212417)*x[2]**o+(-0.870422565759)*x[2]
        s[4,1,0]=(0.538277719545)*x[0]**o+(0.0405414212643)*x[0]+(-0.340695322783)*x[1]**o+(-0.925052511534)*x[1]+(0.77026668428)*x[2]**o+(0.872129470246)*x[2]
        s[4,1,1]=(0.35315629997)*x[0]**o+(-0.581808852999)*x[0]+(-0.506082324541)*x[1]**o+(-0.979298372486)*x[1]+(0.269130557602)*x[2]**o+(0.0297188290624)*x[2]
        s[5,0,0]=(0.13627346547)*x[0]**o+(-0.634005099347)*x[0]+(0.413163361533)*x[1]**o+(-0.445924057754)*x[1]+(0.317763103856)*x[2]**o+(-0.0934833794122)*x[2]
        s[5,0,1]=(0.775864446863)*x[0]**o+(0.647005517482)*x[0]+(-0.922488575786)*x[1]**o+(0.592973645448)*x[1]+(0.752015212966)*x[2]**o+(-0.700335290015)*x[2]
        s[5,1,0]=(0.376707292205)*x[0]**o+(-0.185352369466)*x[0]+(0.701616000772)*x[1]**o+(-0.496738161848)*x[1]+(-0.722103675479)*x[2]**o+(-0.124090359996)*x[2]
        s[5,1,1]=(-0.070302198698)*x[0]**o+(-0.701418359643)*x[0]+(0.131136807633)*x[1]**o+(-0.0581346703562)*x[1]+(0.378664686889)*x[2]**o+(0.785268912587)*x[2]
        ref[0,0,0]=(0.766951612751)/(o+1.)+(0.0847650853109)
        ref[0,0,1]=(-0.66100535067)/(o+1.)+(0.828693754317)
        ref[0,1,0]=(1.12618201455)/(o+1.)+(-0.302888826656)
        ref[0,1,1]=(-0.610794574422)/(o+1.)+(-0.20038120241)
        ref[1,0,0]=(0.994364476124)/(o+1.)+(-0.854460559932)
        ref[1,0,1]=(-0.154408651388)/(o+1.)+(-0.342707477582)
        ref[1,1,0]=(0.0846016762395)/(o+1.)+(-0.141067145737)
        ref[1,1,1]=(-1.47610708289)/(o+1.)+(-0.236606285008)
        ref[2,0,0]=(1.13736850361)/(o+1.)+(0.56671279654)
        ref[2,0,1]=(1.89447169546)/(o+1.)+(0.282918410191)
        ref[2,1,0]=(-0.0839387527076)/(o+1.)+(-0.260176513031)
        ref[2,1,1]=(1.97991596985)/(o+1.)+(-0.516820478238)
        ref[3,0,0]=(0.293015740521)/(o+1.)+(-0.0995204709152)
        ref[3,0,1]=(0.132376257612)/(o+1.)+(0.458335226861)
        ref[3,1,0]=(0.902100791146)/(o+1.)+(-0.635786070046)
        ref[3,1,1]=(-0.182333554037)/(o+1.)+(0.421430955369)
        ref[4,0,0]=(0.265793575979)/(o+1.)+(-0.174444870545)
        ref[4,0,1]=(1.30030828904)/(o+1.)+(-0.625043419818)
        ref[4,1,0]=(0.967849081042)/(o+1.)+(-0.00619081001188)
        ref[4,1,1]=(0.116204533031)/(o+1.)+(-0.765694198211)
        ref[5,0,0]=(0.867199930859)/(o+1.)+(-0.586706268257)
        ref[5,0,1]=(0.605391084042)/(o+1.)+(0.269821936457)
        ref[5,1,0]=(0.356219617498)/(o+1.)+(-0.403090445655)
        ref[5,1,1]=(0.439499295823)/(o+1.)+(0.0128579412935)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.904602515522)*x[0]**o+(-0.371521279208)*x[0]+(-0.452933934201)*x[1]**o+(-0.202117873586)*x[1]
        s[0,0,0,1]=(0.046504512522)*x[0]**o+(-0.23169139456)*x[0]+(0.260745907068)*x[1]**o+(-0.280939854842)*x[1]
        s[0,0,1,0]=(0.150287081226)*x[0]**o+(-0.62438094501)*x[0]+(-0.617182852213)*x[1]**o+(0.0366613552193)*x[1]
        s[0,0,1,1]=(0.0710288489579)*x[0]**o+(-0.111273065295)*x[0]+(0.41817519756)*x[1]**o+(0.792759228517)*x[1]
        s[0,0,2,0]=(-0.00835504723814)*x[0]**o+(-0.928910786463)*x[0]+(-0.830115466368)*x[1]**o+(0.117896100163)*x[1]
        s[0,0,2,1]=(0.178088696257)*x[0]**o+(0.378404605517)*x[0]+(-0.0609283403081)*x[1]**o+(-0.0543695218661)*x[1]
        s[0,1,0,0]=(-0.0689997578822)*x[0]**o+(-0.632975314677)*x[0]+(0.817036198517)*x[1]**o+(0.83235818794)*x[1]
        s[0,1,0,1]=(0.307747809333)*x[0]**o+(-0.119316653583)*x[0]+(-0.962502450634)*x[1]**o+(0.730005736334)*x[1]
        s[0,1,1,0]=(0.474932601365)*x[0]**o+(-0.672778327766)*x[0]+(0.0931366097943)*x[1]**o+(0.441430893263)*x[1]
        s[0,1,1,1]=(-0.688077109622)*x[0]**o+(-0.0904781446959)*x[0]+(0.75334586689)*x[1]**o+(0.64166855195)*x[1]
        s[0,1,2,0]=(-0.445313569413)*x[0]**o+(0.0860176511221)*x[0]+(0.816569466868)*x[1]**o+(-0.716517326506)*x[1]
        s[0,1,2,1]=(-0.282441352341)*x[0]**o+(-0.63704220861)*x[0]+(-0.0493662512733)*x[1]**o+(0.568004633808)*x[1]
        s[0,2,0,0]=(-0.39629286578)*x[0]**o+(-0.930416708861)*x[0]+(0.426475002251)*x[1]**o+(-0.156685168112)*x[1]
        s[0,2,0,1]=(-0.932671691096)*x[0]**o+(-0.918584855074)*x[0]+(0.820320229167)*x[1]**o+(0.39108460143)*x[1]
        s[0,2,1,0]=(0.904161654788)*x[0]**o+(0.580424354995)*x[0]+(0.901999297026)*x[1]**o+(0.16227344306)*x[1]
        s[0,2,1,1]=(0.80438774629)*x[0]**o+(0.132549332211)*x[0]+(0.799348114478)*x[1]**o+(-0.22361807585)*x[1]
        s[0,2,2,0]=(-0.561135188306)*x[0]**o+(0.584405140552)*x[0]+(0.655074058246)*x[1]**o+(0.145893780708)*x[1]
        s[0,2,2,1]=(-0.252147320901)*x[0]**o+(0.798853343056)*x[0]+(0.615093018404)*x[1]**o+(0.394704774949)*x[1]
        s[0,3,0,0]=(0.36142643768)*x[0]**o+(0.307386971305)*x[0]+(-0.294243769289)*x[1]**o+(-0.516116390793)*x[1]
        s[0,3,0,1]=(-0.258524392445)*x[0]**o+(-0.955667526515)*x[0]+(-0.268164834645)*x[1]**o+(-0.0982143901681)*x[1]
        s[0,3,1,0]=(-0.363719831483)*x[0]**o+(0.542839794876)*x[0]+(-0.591176577173)*x[1]**o+(-0.885630814734)*x[1]
        s[0,3,1,1]=(-0.857066901153)*x[0]**o+(0.146459811163)*x[0]+(0.167525505888)*x[1]**o+(-0.823988951028)*x[1]
        s[0,3,2,0]=(-0.67987029357)*x[0]**o+(0.484426195638)*x[0]+(0.357896203022)*x[1]**o+(-0.225752357797)*x[1]
        s[0,3,2,1]=(-0.580488963435)*x[0]**o+(-0.561249356735)*x[0]+(0.383669718576)*x[1]**o+(0.657077607148)*x[1]
        s[0,4,0,0]=(0.69915173204)*x[0]**o+(-0.98282597178)*x[0]+(-0.107525636764)*x[1]**o+(0.860091695416)*x[1]
        s[0,4,0,1]=(-0.52652663231)*x[0]**o+(0.0466888351687)*x[0]+(0.490026321752)*x[1]**o+(0.694751470792)*x[1]
        s[0,4,1,0]=(0.824302118883)*x[0]**o+(0.00719006856334)*x[0]+(-0.539797289173)*x[1]**o+(0.690123831564)*x[1]
        s[0,4,1,1]=(-0.472128229772)*x[0]**o+(-0.648689330166)*x[0]+(0.391117917633)*x[1]**o+(-0.162627834297)*x[1]
        s[0,4,2,0]=(0.225959403713)*x[0]**o+(-0.955982321286)*x[0]+(0.98648257289)*x[1]**o+(0.685606762198)*x[1]
        s[0,4,2,1]=(0.344295953653)*x[0]**o+(0.0937724197016)*x[0]+(0.334346155325)*x[1]**o+(0.240553977121)*x[1]
        s[1,0,0,0]=(0.195329436142)*x[0]**o+(-0.391787107203)*x[0]+(0.282744654065)*x[1]**o+(0.525383075525)*x[1]
        s[1,0,0,1]=(0.0500438294543)*x[0]**o+(0.91622484244)*x[0]+(-0.678531170299)*x[1]**o+(-0.992341400238)*x[1]
        s[1,0,1,0]=(0.297013531565)*x[0]**o+(-0.11033494708)*x[0]+(-0.0204768747597)*x[1]**o+(0.911044821444)*x[1]
        s[1,0,1,1]=(-0.318409579239)*x[0]**o+(-0.0397282629607)*x[0]+(0.126590560999)*x[1]**o+(-0.583058191524)*x[1]
        s[1,0,2,0]=(-0.928353567554)*x[0]**o+(-0.290772724182)*x[0]+(0.114112299004)*x[1]**o+(0.320854358475)*x[1]
        s[1,0,2,1]=(-0.570768568717)*x[0]**o+(-0.824979336359)*x[0]+(-0.521612560762)*x[1]**o+(0.892935684254)*x[1]
        s[1,1,0,0]=(0.409271921712)*x[0]**o+(0.656688417546)*x[0]+(-0.552252115532)*x[1]**o+(-0.252684939643)*x[1]
        s[1,1,0,1]=(0.650926797031)*x[0]**o+(0.937738666477)*x[0]+(-0.280790296797)*x[1]**o+(0.214208558313)*x[1]
        s[1,1,1,0]=(0.1256718371)*x[0]**o+(0.261580183407)*x[0]+(-0.57943553377)*x[1]**o+(-0.746871082838)*x[1]
        s[1,1,1,1]=(0.935864480093)*x[0]**o+(-0.438674963211)*x[0]+(0.18368021533)*x[1]**o+(-0.694240108689)*x[1]
        s[1,1,2,0]=(-0.0397072007891)*x[0]**o+(-0.745343758112)*x[0]+(0.641693619013)*x[1]**o+(-0.658786784991)*x[1]
        s[1,1,2,1]=(-0.695558285084)*x[0]**o+(-0.650222199462)*x[0]+(-0.386725498078)*x[1]**o+(-0.674446830484)*x[1]
        s[1,2,0,0]=(0.905187721767)*x[0]**o+(-0.415012814051)*x[0]+(0.797167315255)*x[1]**o+(-0.80569654981)*x[1]
        s[1,2,0,1]=(-0.425831083968)*x[0]**o+(-0.306155499216)*x[0]+(0.878423386932)*x[1]**o+(-0.204033374375)*x[1]
        s[1,2,1,0]=(0.0315352399527)*x[0]**o+(-0.0864621608168)*x[0]+(0.953279645196)*x[1]**o+(-0.350121337825)*x[1]
        s[1,2,1,1]=(0.30461258451)*x[0]**o+(-0.825063058949)*x[0]+(0.512058260572)*x[1]**o+(0.748809814194)*x[1]
        s[1,2,2,0]=(-0.739360626393)*x[0]**o+(-0.163606333648)*x[0]+(0.0928076120668)*x[1]**o+(0.720798765678)*x[1]
        s[1,2,2,1]=(0.199986177011)*x[0]**o+(-0.355098022147)*x[0]+(-0.429976374764)*x[1]**o+(-0.333973984813)*x[1]
        s[1,3,0,0]=(0.245099889527)*x[0]**o+(-0.272006414876)*x[0]+(0.19038825631)*x[1]**o+(0.141537376325)*x[1]
        s[1,3,0,1]=(0.964892470064)*x[0]**o+(0.266000699533)*x[0]+(0.281992883406)*x[1]**o+(0.0470036363258)*x[1]
        s[1,3,1,0]=(-0.406024010771)*x[0]**o+(0.211515087571)*x[0]+(-0.139150083168)*x[1]**o+(0.694931678647)*x[1]
        s[1,3,1,1]=(0.719958915132)*x[0]**o+(0.425474993616)*x[0]+(0.381998339685)*x[1]**o+(-0.587210328527)*x[1]
        s[1,3,2,0]=(-0.446628131313)*x[0]**o+(0.147944239049)*x[0]+(0.858318344908)*x[1]**o+(0.141912002794)*x[1]
        s[1,3,2,1]=(0.123237888525)*x[0]**o+(0.609628760233)*x[0]+(-0.0997977880881)*x[1]**o+(0.776602457692)*x[1]
        s[1,4,0,0]=(-0.205992404955)*x[0]**o+(0.95994896532)*x[0]+(-0.668831643097)*x[1]**o+(0.400688887056)*x[1]
        s[1,4,0,1]=(-0.551761785809)*x[0]**o+(-0.188933577465)*x[0]+(-0.445368366446)*x[1]**o+(-0.487934192641)*x[1]
        s[1,4,1,0]=(0.806637485565)*x[0]**o+(-0.912751633205)*x[0]+(-0.0317735780285)*x[1]**o+(0.397954044901)*x[1]
        s[1,4,1,1]=(0.321015788412)*x[0]**o+(0.0387288594742)*x[0]+(0.924286425546)*x[1]**o+(0.636709951581)*x[1]
        s[1,4,2,0]=(0.822453145379)*x[0]**o+(-0.763204841293)*x[0]+(-0.821998150153)*x[1]**o+(-0.0455190284992)*x[1]
        s[1,4,2,1]=(-0.769855780302)*x[0]**o+(0.175243832232)*x[0]+(0.643564391513)*x[1]**o+(0.167745614405)*x[1]
        s[2,0,0,0]=(-0.703387009485)*x[0]**o+(-0.165672894741)*x[0]+(-0.593523456406)*x[1]**o+(-0.953916267625)*x[1]
        s[2,0,0,1]=(0.189269691722)*x[0]**o+(0.0650904594513)*x[0]+(-0.972603715891)*x[1]**o+(0.632748799808)*x[1]
        s[2,0,1,0]=(-0.355043086131)*x[0]**o+(0.223886836405)*x[0]+(0.0812465154386)*x[1]**o+(-0.123342881585)*x[1]
        s[2,0,1,1]=(0.111949886338)*x[0]**o+(0.333936233092)*x[0]+(-0.726855801653)*x[1]**o+(-0.410644679001)*x[1]
        s[2,0,2,0]=(-0.514134171061)*x[0]**o+(-0.920122884271)*x[0]+(-0.257750657117)*x[1]**o+(-0.920243692547)*x[1]
        s[2,0,2,1]=(-0.845077655219)*x[0]**o+(-0.208901924124)*x[0]+(0.337378697599)*x[1]**o+(0.801979415473)*x[1]
        s[2,1,0,0]=(0.193043677084)*x[0]**o+(-0.0599530199511)*x[0]+(0.43567411879)*x[1]**o+(0.500546331575)*x[1]
        s[2,1,0,1]=(0.599088933494)*x[0]**o+(-0.468604163828)*x[0]+(0.429967954899)*x[1]**o+(-0.492384378174)*x[1]
        s[2,1,1,0]=(-0.720574232874)*x[0]**o+(0.821661207615)*x[0]+(0.724662820892)*x[1]**o+(0.753198897961)*x[1]
        s[2,1,1,1]=(-0.448359465447)*x[0]**o+(0.693373974155)*x[0]+(-0.484872729016)*x[1]**o+(-0.192659481547)*x[1]
        s[2,1,2,0]=(0.703509253973)*x[0]**o+(-0.0365462091464)*x[0]+(0.240885950313)*x[1]**o+(-0.321055059013)*x[1]
        s[2,1,2,1]=(0.247817476916)*x[0]**o+(-0.138605619124)*x[0]+(0.697567286153)*x[1]**o+(0.727793395093)*x[1]
        s[2,2,0,0]=(0.0419479009665)*x[0]**o+(0.66706255286)*x[0]+(-0.330849950899)*x[1]**o+(-0.600398002711)*x[1]
        s[2,2,0,1]=(0.0302303643583)*x[0]**o+(0.851488267489)*x[0]+(-0.568849620793)*x[1]**o+(-0.363234175722)*x[1]
        s[2,2,1,0]=(-0.0230217508795)*x[0]**o+(0.118232576333)*x[0]+(0.0810817815326)*x[1]**o+(-0.58292650684)*x[1]
        s[2,2,1,1]=(-0.372568177431)*x[0]**o+(-0.613455954736)*x[0]+(0.223650823501)*x[1]**o+(-0.310111084313)*x[1]
        s[2,2,2,0]=(-0.297770107968)*x[0]**o+(0.334407378326)*x[0]+(0.0820495048815)*x[1]**o+(0.670675755602)*x[1]
        s[2,2,2,1]=(-0.884675664948)*x[0]**o+(-0.141329306491)*x[0]+(0.125135636826)*x[1]**o+(-0.911348264716)*x[1]
        s[2,3,0,0]=(0.538935527955)*x[0]**o+(0.513300520882)*x[0]+(0.850744598079)*x[1]**o+(0.751590880335)*x[1]
        s[2,3,0,1]=(-0.821182874204)*x[0]**o+(-0.186188144846)*x[0]+(-0.107871578183)*x[1]**o+(0.462281422576)*x[1]
        s[2,3,1,0]=(0.699998387809)*x[0]**o+(0.497080130045)*x[0]+(-0.876146393883)*x[1]**o+(-0.966792776701)*x[1]
        s[2,3,1,1]=(-0.755486256523)*x[0]**o+(0.482429742628)*x[0]+(0.64207235708)*x[1]**o+(0.792668637521)*x[1]
        s[2,3,2,0]=(0.480810051431)*x[0]**o+(0.526844732935)*x[0]+(0.358083926997)*x[1]**o+(0.975251790749)*x[1]
        s[2,3,2,1]=(-0.71380930491)*x[0]**o+(0.321062714102)*x[0]+(-0.393765851515)*x[1]**o+(0.758037335781)*x[1]
        s[2,4,0,0]=(0.799658695022)*x[0]**o+(-0.970884788767)*x[0]+(-0.909912887472)*x[1]**o+(-0.118569854978)*x[1]
        s[2,4,0,1]=(0.22665071178)*x[0]**o+(-0.866764191923)*x[0]+(0.760310143236)*x[1]**o+(0.979622869213)*x[1]
        s[2,4,1,0]=(-0.639860894248)*x[0]**o+(-0.254761334846)*x[0]+(0.844133251872)*x[1]**o+(-0.0823600617007)*x[1]
        s[2,4,1,1]=(0.699864723463)*x[0]**o+(-0.941417107514)*x[0]+(-0.939935969783)*x[1]**o+(0.307212220608)*x[1]
        s[2,4,2,0]=(-0.597501168242)*x[0]**o+(-0.855810189779)*x[0]+(-0.616246893869)*x[1]**o+(0.862508316402)*x[1]
        s[2,4,2,1]=(0.279133829297)*x[0]**o+(-0.963434838906)*x[0]+(-0.927999672088)*x[1]**o+(-0.633147829562)*x[1]
        s[3,0,0,0]=(-0.228495811891)*x[0]**o+(-0.253261426268)*x[0]+(-0.487641848136)*x[1]**o+(0.321239316844)*x[1]
        s[3,0,0,1]=(-0.226385127812)*x[0]**o+(-0.903861474811)*x[0]+(0.0285492108309)*x[1]**o+(0.199168585275)*x[1]
        s[3,0,1,0]=(-0.0670214317839)*x[0]**o+(-0.169290931005)*x[0]+(0.859675236893)*x[1]**o+(0.673436325782)*x[1]
        s[3,0,1,1]=(0.589201260546)*x[0]**o+(-0.931566266675)*x[0]+(0.127884139041)*x[1]**o+(-0.0831271642138)*x[1]
        s[3,0,2,0]=(-0.228265145653)*x[0]**o+(-0.449360641451)*x[0]+(0.0895266621221)*x[1]**o+(0.516248722513)*x[1]
        s[3,0,2,1]=(0.726271772612)*x[0]**o+(-0.834727339534)*x[0]+(0.779809207203)*x[1]**o+(-0.2745165742)*x[1]
        s[3,1,0,0]=(-0.528300011184)*x[0]**o+(0.0728636912947)*x[0]+(-0.953454233615)*x[1]**o+(0.582219621875)*x[1]
        s[3,1,0,1]=(0.196609308635)*x[0]**o+(0.162717112564)*x[0]+(0.449469968326)*x[1]**o+(-0.174119135554)*x[1]
        s[3,1,1,0]=(-0.261855816805)*x[0]**o+(0.607391359963)*x[0]+(-0.793775568608)*x[1]**o+(-0.950278043961)*x[1]
        s[3,1,1,1]=(0.240905287034)*x[0]**o+(-0.204299485241)*x[0]+(-0.825848364133)*x[1]**o+(-0.00971732871232)*x[1]
        s[3,1,2,0]=(0.798049323787)*x[0]**o+(0.0710454406309)*x[0]+(0.990006532214)*x[1]**o+(0.711055965038)*x[1]
        s[3,1,2,1]=(-0.406803353709)*x[0]**o+(0.120946506707)*x[0]+(-0.410263830047)*x[1]**o+(0.988997635486)*x[1]
        s[3,2,0,0]=(0.181718966759)*x[0]**o+(-0.0262352524424)*x[0]+(-0.1917125307)*x[1]**o+(-0.169130797999)*x[1]
        s[3,2,0,1]=(-0.00378400436872)*x[0]**o+(-0.970892218185)*x[0]+(-0.563995443607)*x[1]**o+(0.906996441633)*x[1]
        s[3,2,1,0]=(-0.858412521643)*x[0]**o+(-0.133539980742)*x[0]+(-0.692474245554)*x[1]**o+(0.0128241106616)*x[1]
        s[3,2,1,1]=(-0.272651295849)*x[0]**o+(-0.185053306751)*x[0]+(-0.034278709119)*x[1]**o+(-0.201998293805)*x[1]
        s[3,2,2,0]=(-0.0376999069998)*x[0]**o+(-0.895215920335)*x[0]+(0.944947071273)*x[1]**o+(-0.732964572204)*x[1]
        s[3,2,2,1]=(0.0771318167573)*x[0]**o+(0.051862360377)*x[0]+(-0.837260137737)*x[1]**o+(-0.770673420806)*x[1]
        s[3,3,0,0]=(-0.258989060244)*x[0]**o+(-0.699818576225)*x[0]+(-0.311518509076)*x[1]**o+(0.517397173211)*x[1]
        s[3,3,0,1]=(-0.443559295683)*x[0]**o+(-0.0544611458142)*x[0]+(0.983503123196)*x[1]**o+(0.248879114721)*x[1]
        s[3,3,1,0]=(0.626729673965)*x[0]**o+(0.347363767931)*x[0]+(0.956080389131)*x[1]**o+(0.429076008118)*x[1]
        s[3,3,1,1]=(-0.508758891471)*x[0]**o+(0.616615567996)*x[0]+(0.221644986884)*x[1]**o+(-0.904935728116)*x[1]
        s[3,3,2,0]=(0.254341631925)*x[0]**o+(-0.908688314004)*x[0]+(-0.0449719498093)*x[1]**o+(0.871189405684)*x[1]
        s[3,3,2,1]=(0.506459636421)*x[0]**o+(0.978066502797)*x[0]+(-0.212934438396)*x[1]**o+(-0.355665670198)*x[1]
        s[3,4,0,0]=(0.0152292845756)*x[0]**o+(-0.400615592376)*x[0]+(0.264352702586)*x[1]**o+(-0.0677780689342)*x[1]
        s[3,4,0,1]=(0.734077033298)*x[0]**o+(-0.171218476666)*x[0]+(-0.399285387071)*x[1]**o+(-0.589596788961)*x[1]
        s[3,4,1,0]=(-0.903602643334)*x[0]**o+(0.0527455440694)*x[0]+(-0.973273256059)*x[1]**o+(0.76054623312)*x[1]
        s[3,4,1,1]=(0.368219867735)*x[0]**o+(0.857361275285)*x[0]+(0.297273583978)*x[1]**o+(0.243120582563)*x[1]
        s[3,4,2,0]=(0.518637118464)*x[0]**o+(-0.961371552746)*x[0]+(0.283809772558)*x[1]**o+(-0.947184379737)*x[1]
        s[3,4,2,1]=(-0.514398206056)*x[0]**o+(-0.499373736713)*x[0]+(-0.756154352548)*x[1]**o+(0.814078333554)*x[1]
        ref[0,0,0,0]=(0.451668581322)/(o+1.)+(-0.286819576397)
        ref[0,0,0,1]=(0.30725041959)/(o+1.)+(-0.256315624701)
        ref[0,0,1,0]=(-0.466895770987)/(o+1.)+(-0.293859794895)
        ref[0,0,1,1]=(0.489204046518)/(o+1.)+(0.340743081611)
        ref[0,0,2,0]=(-0.838470513606)/(o+1.)+(-0.40550734315)
        ref[0,0,2,1]=(0.117160355949)/(o+1.)+(0.162017541826)
        ref[0,1,0,0]=(0.748036440634)/(o+1.)+(0.0996914366319)
        ref[0,1,0,1]=(-0.654754641301)/(o+1.)+(0.305344541376)
        ref[0,1,1,0]=(0.568069211159)/(o+1.)+(-0.115673717252)
        ref[0,1,1,1]=(0.0652687572678)/(o+1.)+(0.275595203627)
        ref[0,1,2,0]=(0.371255897455)/(o+1.)+(-0.315249837692)
        ref[0,1,2,1]=(-0.331807603614)/(o+1.)+(-0.0345187874006)
        ref[0,2,0,0]=(0.0301821364712)/(o+1.)+(-0.543550938487)
        ref[0,2,0,1]=(-0.112351461929)/(o+1.)+(-0.263750126822)
        ref[0,2,1,0]=(1.80616095181)/(o+1.)+(0.371348899028)
        ref[0,2,1,1]=(1.60373586077)/(o+1.)+(-0.0455343718196)
        ref[0,2,2,0]=(0.0939388699401)/(o+1.)+(0.36514946063)
        ref[0,2,2,1]=(0.362945697503)/(o+1.)+(0.596779059003)
        ref[0,3,0,0]=(0.0671826683913)/(o+1.)+(-0.104364709744)
        ref[0,3,0,1]=(-0.52668922709)/(o+1.)+(-0.526940958341)
        ref[0,3,1,0]=(-0.954896408656)/(o+1.)+(-0.171395509929)
        ref[0,3,1,1]=(-0.689541395265)/(o+1.)+(-0.338764569932)
        ref[0,3,2,0]=(-0.321974090548)/(o+1.)+(0.12933691892)
        ref[0,3,2,1]=(-0.196819244859)/(o+1.)+(0.0479141252065)
        ref[0,4,0,0]=(0.591626095276)/(o+1.)+(-0.061367138182)
        ref[0,4,0,1]=(-0.0365003105586)/(o+1.)+(0.37072015298)
        ref[0,4,1,0]=(0.28450482971)/(o+1.)+(0.348656950064)
        ref[0,4,1,1]=(-0.0810103121392)/(o+1.)+(-0.405658582232)
        ref[0,4,2,0]=(1.2124419766)/(o+1.)+(-0.135187779544)
        ref[0,4,2,1]=(0.678642108977)/(o+1.)+(0.167163198411)
        ref[1,0,0,0]=(0.478074090207)/(o+1.)+(0.066797984161)
        ref[1,0,0,1]=(-0.628487340845)/(o+1.)+(-0.0380582788988)
        ref[1,0,1,0]=(0.276536656806)/(o+1.)+(0.400354937182)
        ref[1,0,1,1]=(-0.191819018239)/(o+1.)+(-0.311393227243)
        ref[1,0,2,0]=(-0.81424126855)/(o+1.)+(0.0150408171462)
        ref[1,0,2,1]=(-1.09238112948)/(o+1.)+(0.0339781739475)
        ref[1,1,0,0]=(-0.142980193819)/(o+1.)+(0.202001738951)
        ref[1,1,0,1]=(0.370136500233)/(o+1.)+(0.575973612395)
        ref[1,1,1,0]=(-0.45376369667)/(o+1.)+(-0.242645449715)
        ref[1,1,1,1]=(1.11954469542)/(o+1.)+(-0.56645753595)
        ref[1,1,2,0]=(0.601986418224)/(o+1.)+(-0.702065271552)
        ref[1,1,2,1]=(-1.08228378316)/(o+1.)+(-0.662334514973)
        ref[1,2,0,0]=(1.70235503702)/(o+1.)+(-0.61035468193)
        ref[1,2,0,1]=(0.452592302963)/(o+1.)+(-0.255094436796)
        ref[1,2,1,0]=(0.984814885149)/(o+1.)+(-0.218291749321)
        ref[1,2,1,1]=(0.816670845082)/(o+1.)+(-0.0381266223775)
        ref[1,2,2,0]=(-0.646553014327)/(o+1.)+(0.278596216015)
        ref[1,2,2,1]=(-0.229990197753)/(o+1.)+(-0.34453600348)
        ref[1,3,0,0]=(0.435488145837)/(o+1.)+(-0.0652345192758)
        ref[1,3,0,1]=(1.24688535347)/(o+1.)+(0.15650216793)
        ref[1,3,1,0]=(-0.545174093939)/(o+1.)+(0.453223383109)
        ref[1,3,1,1]=(1.10195725482)/(o+1.)+(-0.0808676674551)
        ref[1,3,2,0]=(0.411690213595)/(o+1.)+(0.144928120921)
        ref[1,3,2,1]=(0.0234401004369)/(o+1.)+(0.693115608963)
        ref[1,4,0,0]=(-0.874824048052)/(o+1.)+(0.680318926188)
        ref[1,4,0,1]=(-0.997130152255)/(o+1.)+(-0.338433885053)
        ref[1,4,1,0]=(0.774863907537)/(o+1.)+(-0.257398794152)
        ref[1,4,1,1]=(1.24530221396)/(o+1.)+(0.337719405527)
        ref[1,4,2,0]=(0.000454995225935)/(o+1.)+(-0.404361934896)
        ref[1,4,2,1]=(-0.126291388789)/(o+1.)+(0.171494723319)
        ref[2,0,0,0]=(-1.29691046589)/(o+1.)+(-0.559794581183)
        ref[2,0,0,1]=(-0.783334024169)/(o+1.)+(0.348919629629)
        ref[2,0,1,0]=(-0.273796570692)/(o+1.)+(0.0502719774097)
        ref[2,0,1,1]=(-0.614905915315)/(o+1.)+(-0.0383542229546)
        ref[2,0,2,0]=(-0.771884828178)/(o+1.)+(-0.920183288409)
        ref[2,0,2,1]=(-0.50769895762)/(o+1.)+(0.296538745675)
        ref[2,1,0,0]=(0.628717795874)/(o+1.)+(0.220296655812)
        ref[2,1,0,1]=(1.02905688839)/(o+1.)+(-0.480494271001)
        ref[2,1,1,0]=(0.00408858801768)/(o+1.)+(0.787430052788)
        ref[2,1,1,1]=(-0.933232194463)/(o+1.)+(0.250357246304)
        ref[2,1,2,0]=(0.944395204286)/(o+1.)+(-0.17880063408)
        ref[2,1,2,1]=(0.945384763069)/(o+1.)+(0.294593887985)
        ref[2,2,0,0]=(-0.288902049932)/(o+1.)+(0.0333322750744)
        ref[2,2,0,1]=(-0.538619256435)/(o+1.)+(0.244127045883)
        ref[2,2,1,0]=(0.0580600306531)/(o+1.)+(-0.232346965253)
        ref[2,2,1,1]=(-0.148917353929)/(o+1.)+(-0.461783519524)
        ref[2,2,2,0]=(-0.215720603086)/(o+1.)+(0.502541566964)
        ref[2,2,2,1]=(-0.759540028122)/(o+1.)+(-0.526338785603)
        ref[2,3,0,0]=(1.38968012603)/(o+1.)+(0.632445700608)
        ref[2,3,0,1]=(-0.929054452386)/(o+1.)+(0.138046638865)
        ref[2,3,1,0]=(-0.176148006073)/(o+1.)+(-0.234856323328)
        ref[2,3,1,1]=(-0.113413899443)/(o+1.)+(0.637549190074)
        ref[2,3,2,0]=(0.838893978428)/(o+1.)+(0.751048261842)
        ref[2,3,2,1]=(-1.10757515642)/(o+1.)+(0.539550024941)
        ref[2,4,0,0]=(-0.11025419245)/(o+1.)+(-0.544727321873)
        ref[2,4,0,1]=(0.986960855016)/(o+1.)+(0.056429338645)
        ref[2,4,1,0]=(0.204272357624)/(o+1.)+(-0.168560698273)
        ref[2,4,1,1]=(-0.24007124632)/(o+1.)+(-0.317102443453)
        ref[2,4,2,0]=(-1.21374806211)/(o+1.)+(0.00334906331113)
        ref[2,4,2,1]=(-0.648865842791)/(o+1.)+(-0.798291334234)
        ref[3,0,0,0]=(-0.716137660027)/(o+1.)+(0.033988945288)
        ref[3,0,0,1]=(-0.197835916981)/(o+1.)+(-0.352346444768)
        ref[3,0,1,0]=(0.792653805109)/(o+1.)+(0.252072697388)
        ref[3,0,1,1]=(0.717085399588)/(o+1.)+(-0.507346715445)
        ref[3,0,2,0]=(-0.138738483531)/(o+1.)+(0.0334440405313)
        ref[3,0,2,1]=(1.50608097982)/(o+1.)+(-0.554621956867)
        ref[3,1,0,0]=(-1.4817542448)/(o+1.)+(0.327541656585)
        ref[3,1,0,1]=(0.64607927696)/(o+1.)+(-0.00570101149523)
        ref[3,1,1,0]=(-1.05563138541)/(o+1.)+(-0.171443341999)
        ref[3,1,1,1]=(-0.5849430771)/(o+1.)+(-0.107008406977)
        ref[3,1,2,0]=(1.788055856)/(o+1.)+(0.391050702835)
        ref[3,1,2,1]=(-0.817067183756)/(o+1.)+(0.554972071097)
        ref[3,2,0,0]=(-0.00999356394068)/(o+1.)+(-0.0976830252209)
        ref[3,2,0,1]=(-0.567779447976)/(o+1.)+(-0.031947888276)
        ref[3,2,1,0]=(-1.5508867672)/(o+1.)+(-0.0603579350402)
        ref[3,2,1,1]=(-0.306930004968)/(o+1.)+(-0.193525800278)
        ref[3,2,2,0]=(0.907247164273)/(o+1.)+(-0.81409024627)
        ref[3,2,2,1]=(-0.76012832098)/(o+1.)+(-0.359405530215)
        ref[3,3,0,0]=(-0.57050756932)/(o+1.)+(-0.0912107015068)
        ref[3,3,0,1]=(0.539943827514)/(o+1.)+(0.0972089844535)
        ref[3,3,1,0]=(1.5828100631)/(o+1.)+(0.388219888025)
        ref[3,3,1,1]=(-0.287113904587)/(o+1.)+(-0.14416008006)
        ref[3,3,2,0]=(0.209369682116)/(o+1.)+(-0.01874945416)
        ref[3,3,2,1]=(0.293525198025)/(o+1.)+(0.311200416299)
        ref[3,4,0,0]=(0.279581987162)/(o+1.)+(-0.234196830655)
        ref[3,4,0,1]=(0.334791646227)/(o+1.)+(-0.380407632814)
        ref[3,4,1,0]=(-1.87687589939)/(o+1.)+(0.406645888595)
        ref[3,4,1,1]=(0.665493451713)/(o+1.)+(0.550240928924)
        ref[3,4,2,0]=(0.802446891023)/(o+1.)+(-0.954277966241)
        ref[3,4,2,1]=(-1.2705525586)/(o+1.)+(0.15735229842)
      else:
        s[0,0,0,0]=(0.641376947921)*x[0]**o+(-0.761647983943)*x[0]+(0.1869803104)*x[1]**o+(-0.325986059158)*x[1]+(-0.0520537723186)*x[2]**o+(-0.242075762859)*x[2]
        s[0,0,0,1]=(0.774456605418)*x[0]**o+(0.321485563379)*x[0]+(0.0248450395096)*x[1]**o+(-0.425638835963)*x[1]+(-0.0622024423983)*x[2]**o+(0.0431057597271)*x[2]
        s[0,0,1,0]=(0.661432460103)*x[0]**o+(-0.653382716149)*x[0]+(0.596912471264)*x[1]**o+(-0.375947479542)*x[1]+(0.370999295823)*x[2]**o+(0.791872078833)*x[2]
        s[0,0,1,1]=(0.258860589295)*x[0]**o+(-0.802340135742)*x[0]+(0.569149399765)*x[1]**o+(-0.505649920225)*x[1]+(0.384267048627)*x[2]**o+(-0.791015249057)*x[2]
        s[0,0,2,0]=(0.219942214071)*x[0]**o+(0.360851774828)*x[0]+(-0.456714182675)*x[1]**o+(0.0965079164275)*x[1]+(-0.752467360132)*x[2]**o+(-0.285016148963)*x[2]
        s[0,0,2,1]=(0.227249920303)*x[0]**o+(-0.63371992192)*x[0]+(-0.751996139921)*x[1]**o+(-0.804517033328)*x[1]+(-0.0202988256583)*x[2]**o+(0.531173510334)*x[2]
        s[0,1,0,0]=(0.530838912723)*x[0]**o+(0.230970767172)*x[0]+(-0.424131392941)*x[1]**o+(-0.606417012546)*x[1]+(-0.318937522783)*x[2]**o+(0.179390519765)*x[2]
        s[0,1,0,1]=(-0.776569569832)*x[0]**o+(0.918265838431)*x[0]+(-0.091878191762)*x[1]**o+(0.698951377419)*x[1]+(0.244940000061)*x[2]**o+(0.0452610075428)*x[2]
        s[0,1,1,0]=(0.435082117831)*x[0]**o+(0.889881605799)*x[0]+(-0.219619308431)*x[1]**o+(-0.672303092957)*x[1]+(0.987619326906)*x[2]**o+(0.220516905778)*x[2]
        s[0,1,1,1]=(-0.118683589446)*x[0]**o+(0.965421464735)*x[0]+(-0.863562140452)*x[1]**o+(-0.231242073932)*x[1]+(0.568240493507)*x[2]**o+(0.734578820791)*x[2]
        s[0,1,2,0]=(0.511447709837)*x[0]**o+(-0.274507694143)*x[0]+(0.576288098696)*x[1]**o+(0.655476727813)*x[1]+(0.602315737594)*x[2]**o+(0.817938429306)*x[2]
        s[0,1,2,1]=(-0.319953701981)*x[0]**o+(-0.0816890923671)*x[0]+(0.84471771093)*x[1]**o+(0.216072534675)*x[1]+(0.782233227772)*x[2]**o+(0.761080396806)*x[2]
        s[0,2,0,0]=(0.371971964019)*x[0]**o+(-0.914300658178)*x[0]+(-0.421914056897)*x[1]**o+(-0.366900498221)*x[1]+(-0.224742744541)*x[2]**o+(0.718117565395)*x[2]
        s[0,2,0,1]=(0.390948765614)*x[0]**o+(0.523332138936)*x[0]+(0.207458000485)*x[1]**o+(0.704266764721)*x[1]+(0.194539635633)*x[2]**o+(-0.0310558120915)*x[2]
        s[0,2,1,0]=(-0.00308409831643)*x[0]**o+(0.0874338943679)*x[0]+(0.792479180633)*x[1]**o+(-0.840935676827)*x[1]+(0.183535198853)*x[2]**o+(-0.632484318094)*x[2]
        s[0,2,1,1]=(0.93055010195)*x[0]**o+(-0.592883255994)*x[0]+(-0.293724124968)*x[1]**o+(-0.532867438247)*x[1]+(-0.731629824662)*x[2]**o+(-0.926457959344)*x[2]
        s[0,2,2,0]=(-0.410766259996)*x[0]**o+(0.153463001676)*x[0]+(-0.22434101683)*x[1]**o+(0.792184946048)*x[1]+(0.217569318007)*x[2]**o+(-0.746503700782)*x[2]
        s[0,2,2,1]=(-0.48097027987)*x[0]**o+(-0.855259653713)*x[0]+(0.297296031945)*x[1]**o+(0.486702064445)*x[1]+(-0.15982283079)*x[2]**o+(0.943473459324)*x[2]
        s[0,3,0,0]=(0.440444162838)*x[0]**o+(-0.411163880392)*x[0]+(0.434891712314)*x[1]**o+(0.421032339491)*x[1]+(0.185417751167)*x[2]**o+(0.774341796864)*x[2]
        s[0,3,0,1]=(0.418605026678)*x[0]**o+(0.697823127475)*x[0]+(-0.490471272795)*x[1]**o+(0.889959972523)*x[1]+(-0.935571921221)*x[2]**o+(-0.477163919777)*x[2]
        s[0,3,1,0]=(0.240863375001)*x[0]**o+(0.171535577067)*x[0]+(-0.798318413682)*x[1]**o+(0.165523160187)*x[1]+(-0.793244034855)*x[2]**o+(-0.639277575041)*x[2]
        s[0,3,1,1]=(0.202212256877)*x[0]**o+(0.387471670954)*x[0]+(-0.198376856629)*x[1]**o+(-0.679099770813)*x[1]+(-0.213566318573)*x[2]**o+(-0.591165980661)*x[2]
        s[0,3,2,0]=(0.764050063724)*x[0]**o+(-0.921472809292)*x[0]+(-0.483029501792)*x[1]**o+(0.198096560699)*x[1]+(-0.161449532882)*x[2]**o+(-0.854460510477)*x[2]
        s[0,3,2,1]=(0.491500019407)*x[0]**o+(0.830069678076)*x[0]+(-0.597452015974)*x[1]**o+(-0.893928655959)*x[1]+(0.940149982314)*x[2]**o+(0.17300719331)*x[2]
        s[0,4,0,0]=(0.0798620352831)*x[0]**o+(0.0237937127416)*x[0]+(-0.086351109139)*x[1]**o+(0.729882388424)*x[1]+(-0.7065299588)*x[2]**o+(-0.632673206138)*x[2]
        s[0,4,0,1]=(-0.809989577091)*x[0]**o+(0.0981392253023)*x[0]+(0.831656529801)*x[1]**o+(0.0337424060561)*x[1]+(-0.210067786784)*x[2]**o+(-0.0549477247313)*x[2]
        s[0,4,1,0]=(-0.0632498403207)*x[0]**o+(-0.992515766051)*x[0]+(0.579065868983)*x[1]**o+(0.201496477221)*x[1]+(0.147678629343)*x[2]**o+(0.396978535826)*x[2]
        s[0,4,1,1]=(0.538470621207)*x[0]**o+(0.125447714672)*x[0]+(0.345561739505)*x[1]**o+(-0.898311293341)*x[1]+(0.695965393051)*x[2]**o+(0.456779053964)*x[2]
        s[0,4,2,0]=(-0.461938934992)*x[0]**o+(0.815268894146)*x[0]+(0.529702893063)*x[1]**o+(0.106752164506)*x[1]+(-0.892819156049)*x[2]**o+(-0.540251931665)*x[2]
        s[0,4,2,1]=(-0.890961526208)*x[0]**o+(-0.0522965595586)*x[0]+(-0.166811931177)*x[1]**o+(-0.32795191314)*x[1]+(0.827693520674)*x[2]**o+(-0.858862162004)*x[2]
        s[1,0,0,0]=(0.0341612512071)*x[0]**o+(0.704407020254)*x[0]+(-0.0212728069872)*x[1]**o+(-0.360120370601)*x[1]+(0.0355417904312)*x[2]**o+(-0.899016835477)*x[2]
        s[1,0,0,1]=(-0.254336879728)*x[0]**o+(-0.749583807992)*x[0]+(-0.822540014653)*x[1]**o+(0.321686636614)*x[1]+(-0.110121756691)*x[2]**o+(0.995172970249)*x[2]
        s[1,0,1,0]=(0.538783330989)*x[0]**o+(0.425931208203)*x[0]+(0.808850719257)*x[1]**o+(-0.616824777466)*x[1]+(-0.999643426203)*x[2]**o+(0.206956920828)*x[2]
        s[1,0,1,1]=(0.92122724205)*x[0]**o+(-0.0077578775328)*x[0]+(-0.455369944999)*x[1]**o+(0.069992464289)*x[1]+(-0.625922232223)*x[2]**o+(0.308582889054)*x[2]
        s[1,0,2,0]=(0.812275176702)*x[0]**o+(-0.0737918971904)*x[0]+(-0.352834721968)*x[1]**o+(-0.485715856695)*x[1]+(0.893012104541)*x[2]**o+(-0.50277227482)*x[2]
        s[1,0,2,1]=(0.327001325412)*x[0]**o+(0.248420422978)*x[0]+(0.934489184596)*x[1]**o+(0.645665157141)*x[1]+(-0.125952174434)*x[2]**o+(-0.589610796444)*x[2]
        s[1,1,0,0]=(-0.19439618112)*x[0]**o+(-0.473153238554)*x[0]+(0.441593313035)*x[1]**o+(0.560223203285)*x[1]+(0.744701337265)*x[2]**o+(0.588942879247)*x[2]
        s[1,1,0,1]=(0.956622537359)*x[0]**o+(0.388369191252)*x[0]+(-0.912880779816)*x[1]**o+(-0.930624449932)*x[1]+(-0.0693131490103)*x[2]**o+(0.79864169138)*x[2]
        s[1,1,1,0]=(0.0802315160845)*x[0]**o+(0.360021747567)*x[0]+(-0.838148172909)*x[1]**o+(0.974574791036)*x[1]+(0.134930445828)*x[2]**o+(-0.224819866647)*x[2]
        s[1,1,1,1]=(0.0408467896812)*x[0]**o+(-0.527931565684)*x[0]+(-0.294779051686)*x[1]**o+(0.655648246879)*x[1]+(0.206641334088)*x[2]**o+(0.244207224731)*x[2]
        s[1,1,2,0]=(-0.695324333815)*x[0]**o+(-0.76302643793)*x[0]+(-0.445448086412)*x[1]**o+(0.387620295728)*x[1]+(0.410986742302)*x[2]**o+(-0.903330791078)*x[2]
        s[1,1,2,1]=(0.725983511781)*x[0]**o+(-0.55616923731)*x[0]+(0.694922396401)*x[1]**o+(0.969679789365)*x[1]+(0.390140728655)*x[2]**o+(0.616506146743)*x[2]
        s[1,2,0,0]=(-0.847983007272)*x[0]**o+(-0.21379616326)*x[0]+(-0.221990648875)*x[1]**o+(0.542144543115)*x[1]+(-0.476016238668)*x[2]**o+(0.239835257111)*x[2]
        s[1,2,0,1]=(0.382592141805)*x[0]**o+(-0.547322693057)*x[0]+(0.207135932694)*x[1]**o+(0.479160574257)*x[1]+(-0.406772807762)*x[2]**o+(0.0827537744696)*x[2]
        s[1,2,1,0]=(0.385177578855)*x[0]**o+(-0.96749949076)*x[0]+(-0.980528334919)*x[1]**o+(0.846145425556)*x[1]+(0.663229596459)*x[2]**o+(0.234589730098)*x[2]
        s[1,2,1,1]=(0.326538279652)*x[0]**o+(0.864444241537)*x[0]+(0.925715710223)*x[1]**o+(0.901973119241)*x[1]+(0.489827899451)*x[2]**o+(-0.446654474045)*x[2]
        s[1,2,2,0]=(0.623521048768)*x[0]**o+(0.00694630100059)*x[0]+(-0.316266037281)*x[1]**o+(-0.296669528851)*x[1]+(-0.345499066232)*x[2]**o+(-0.781251040298)*x[2]
        s[1,2,2,1]=(-0.564332000062)*x[0]**o+(-0.838410587038)*x[0]+(0.105923226211)*x[1]**o+(0.027080971794)*x[1]+(0.656502442527)*x[2]**o+(-0.968700980875)*x[2]
        s[1,3,0,0]=(0.28297404936)*x[0]**o+(0.906254282744)*x[0]+(0.426459225459)*x[1]**o+(0.474780047643)*x[1]+(0.915469451254)*x[2]**o+(0.150014279679)*x[2]
        s[1,3,0,1]=(0.417919974897)*x[0]**o+(0.118528892402)*x[0]+(0.456086489559)*x[1]**o+(0.407436737785)*x[1]+(-0.594040737087)*x[2]**o+(-0.0968244395915)*x[2]
        s[1,3,1,0]=(-0.00608149061984)*x[0]**o+(0.279861847458)*x[0]+(0.622153448276)*x[1]**o+(0.307580940395)*x[1]+(-0.311958042371)*x[2]**o+(0.951637572927)*x[2]
        s[1,3,1,1]=(0.0620010964861)*x[0]**o+(0.694489586135)*x[0]+(-0.397214573709)*x[1]**o+(0.816641484229)*x[1]+(-0.55010098877)*x[2]**o+(-0.252677099684)*x[2]
        s[1,3,2,0]=(0.347469613465)*x[0]**o+(-0.793025539854)*x[0]+(0.340939082154)*x[1]**o+(0.00135134646417)*x[1]+(-0.253753620216)*x[2]**o+(-0.795217820163)*x[2]
        s[1,3,2,1]=(-0.77663362989)*x[0]**o+(-0.0454794482729)*x[0]+(0.0346722039087)*x[1]**o+(-0.137241685281)*x[1]+(0.984186419894)*x[2]**o+(0.884865835233)*x[2]
        s[1,4,0,0]=(-0.344544201299)*x[0]**o+(0.118604431816)*x[0]+(0.735491059587)*x[1]**o+(-0.474594418844)*x[1]+(-0.160183641095)*x[2]**o+(0.0442675403903)*x[2]
        s[1,4,0,1]=(0.159437065336)*x[0]**o+(0.357454800453)*x[0]+(0.0623836301056)*x[1]**o+(0.255296236156)*x[1]+(0.586460110084)*x[2]**o+(-0.557870816384)*x[2]
        s[1,4,1,0]=(0.272212988717)*x[0]**o+(-0.00218257591936)*x[0]+(0.637653920938)*x[1]**o+(0.474154612654)*x[1]+(0.0927963495169)*x[2]**o+(0.688169553045)*x[2]
        s[1,4,1,1]=(-0.296752052135)*x[0]**o+(0.222825403373)*x[0]+(-0.61297870931)*x[1]**o+(-0.772169949259)*x[1]+(0.267924738357)*x[2]**o+(-0.98640521973)*x[2]
        s[1,4,2,0]=(-0.528327447355)*x[0]**o+(-0.13640459815)*x[0]+(-0.839608161898)*x[1]**o+(0.79996626812)*x[1]+(0.0889785028146)*x[2]**o+(0.483698551365)*x[2]
        s[1,4,2,1]=(-0.888180576683)*x[0]**o+(-0.243274686308)*x[0]+(-0.459488747018)*x[1]**o+(0.770881162569)*x[1]+(-0.502642334958)*x[2]**o+(-0.0596414308521)*x[2]
        s[2,0,0,0]=(-0.765317614505)*x[0]**o+(-0.220818348989)*x[0]+(-0.185228616416)*x[1]**o+(-0.0787003843488)*x[1]+(-0.412520273332)*x[2]**o+(0.192760450932)*x[2]
        s[2,0,0,1]=(0.913449879489)*x[0]**o+(-0.219345049089)*x[0]+(0.54504962001)*x[1]**o+(0.623236284643)*x[1]+(-0.350717162156)*x[2]**o+(-0.623681076852)*x[2]
        s[2,0,1,0]=(-0.940664057127)*x[0]**o+(0.251319839352)*x[0]+(0.877556299495)*x[1]**o+(0.016160084659)*x[1]+(-0.544235646954)*x[2]**o+(-0.343760171279)*x[2]
        s[2,0,1,1]=(0.324227271021)*x[0]**o+(-0.380692454159)*x[0]+(-0.727849414028)*x[1]**o+(0.955281983149)*x[1]+(-0.898073831095)*x[2]**o+(0.858127409364)*x[2]
        s[2,0,2,0]=(-0.111449036516)*x[0]**o+(0.928827923826)*x[0]+(-0.992899512514)*x[1]**o+(0.0150334039442)*x[1]+(-0.552135666421)*x[2]**o+(0.406949591012)*x[2]
        s[2,0,2,1]=(0.241832900797)*x[0]**o+(0.684688614786)*x[0]+(0.842064214069)*x[1]**o+(-0.718696633586)*x[1]+(0.352196331118)*x[2]**o+(-0.260343576356)*x[2]
        s[2,1,0,0]=(0.426302002061)*x[0]**o+(0.573786857786)*x[0]+(0.909827742357)*x[1]**o+(0.34656984965)*x[1]+(0.379863389758)*x[2]**o+(0.0322921482901)*x[2]
        s[2,1,0,1]=(0.34423098891)*x[0]**o+(0.959825965873)*x[0]+(-0.458147326954)*x[1]**o+(-0.190830182513)*x[1]+(-0.0836184656488)*x[2]**o+(0.711921183181)*x[2]
        s[2,1,1,0]=(0.813184534496)*x[0]**o+(-0.551568400877)*x[0]+(0.0282287303962)*x[1]**o+(0.421838849382)*x[1]+(0.312118130223)*x[2]**o+(-0.276104599646)*x[2]
        s[2,1,1,1]=(0.39842349776)*x[0]**o+(0.0505250495963)*x[0]+(-0.398145839899)*x[1]**o+(0.712110112191)*x[1]+(-0.646997386195)*x[2]**o+(-0.584123612783)*x[2]
        s[2,1,2,0]=(-0.722790769831)*x[0]**o+(0.850227076858)*x[0]+(0.204013681627)*x[1]**o+(0.924250450519)*x[1]+(-0.0736995922058)*x[2]**o+(-0.597799712798)*x[2]
        s[2,1,2,1]=(0.314644875138)*x[0]**o+(-0.636988948936)*x[0]+(-0.641641656145)*x[1]**o+(0.788809853074)*x[1]+(0.993531553111)*x[2]**o+(-0.495755039677)*x[2]
        s[2,2,0,0]=(0.633851639171)*x[0]**o+(0.523452525521)*x[0]+(-0.382111999068)*x[1]**o+(-0.63124196783)*x[1]+(-0.147565214833)*x[2]**o+(0.0331370104049)*x[2]
        s[2,2,0,1]=(0.526384634874)*x[0]**o+(0.133282991706)*x[0]+(0.805017307077)*x[1]**o+(-0.266487317047)*x[1]+(0.33606063535)*x[2]**o+(0.805493533241)*x[2]
        s[2,2,1,0]=(0.346357232135)*x[0]**o+(-0.787017955245)*x[0]+(0.30877740144)*x[1]**o+(0.451911959523)*x[1]+(0.747842891938)*x[2]**o+(-0.44945658191)*x[2]
        s[2,2,1,1]=(-0.546211248475)*x[0]**o+(0.247263733668)*x[0]+(-0.594514097121)*x[1]**o+(-0.264637607913)*x[1]+(0.418219952147)*x[2]**o+(-0.816923592759)*x[2]
        s[2,2,2,0]=(0.0563761896319)*x[0]**o+(-0.736507329679)*x[0]+(0.470349998404)*x[1]**o+(-0.932042349224)*x[1]+(0.462401759004)*x[2]**o+(-0.714392002134)*x[2]
        s[2,2,2,1]=(0.240181845755)*x[0]**o+(-0.55763353473)*x[0]+(0.285496776429)*x[1]**o+(0.842557595355)*x[1]+(-0.672368632752)*x[2]**o+(0.298762100435)*x[2]
        s[2,3,0,0]=(0.316677723523)*x[0]**o+(-0.637036430356)*x[0]+(0.00972363132121)*x[1]**o+(-0.30937691623)*x[1]+(-0.757958054703)*x[2]**o+(0.261972786754)*x[2]
        s[2,3,0,1]=(-0.0111918506129)*x[0]**o+(-0.79939173162)*x[0]+(0.237847735286)*x[1]**o+(-0.78668226084)*x[1]+(0.707378088836)*x[2]**o+(-0.934201311773)*x[2]
        s[2,3,1,0]=(-0.0276842306357)*x[0]**o+(0.353182001738)*x[0]+(0.458929280756)*x[1]**o+(0.845593054688)*x[1]+(-0.0437171694787)*x[2]**o+(-0.75271342809)*x[2]
        s[2,3,1,1]=(0.752394290258)*x[0]**o+(-0.179837111909)*x[0]+(0.939093046468)*x[1]**o+(0.460733186456)*x[1]+(0.747517309414)*x[2]**o+(-0.58266983679)*x[2]
        s[2,3,2,0]=(-0.951300235705)*x[0]**o+(-0.350444110401)*x[0]+(-0.904564312825)*x[1]**o+(0.356471509479)*x[1]+(0.0971202512406)*x[2]**o+(0.227621817712)*x[2]
        s[2,3,2,1]=(0.404143843413)*x[0]**o+(-0.911041017819)*x[0]+(-0.567863808091)*x[1]**o+(0.228161895607)*x[1]+(-0.160584718896)*x[2]**o+(-0.951937728172)*x[2]
        s[2,4,0,0]=(-0.284708280204)*x[0]**o+(-0.225655800044)*x[0]+(0.0189816391307)*x[1]**o+(-0.847695463473)*x[1]+(-0.603035310159)*x[2]**o+(0.878440760667)*x[2]
        s[2,4,0,1]=(0.249580969821)*x[0]**o+(0.00196364400716)*x[0]+(-0.580222966198)*x[1]**o+(0.266039411545)*x[1]+(-0.0857426202045)*x[2]**o+(0.297517844516)*x[2]
        s[2,4,1,0]=(-0.847202130689)*x[0]**o+(-0.409645962819)*x[0]+(-0.914867914136)*x[1]**o+(0.464829575727)*x[1]+(0.2358554853)*x[2]**o+(-0.951825555951)*x[2]
        s[2,4,1,1]=(-0.925109552513)*x[0]**o+(0.57253113036)*x[0]+(0.531590407505)*x[1]**o+(-0.812964837422)*x[1]+(-0.354555503225)*x[2]**o+(-0.453054364223)*x[2]
        s[2,4,2,0]=(0.669533379208)*x[0]**o+(0.147512160447)*x[0]+(0.730871788446)*x[1]**o+(-0.2646655671)*x[1]+(-0.890950482274)*x[2]**o+(0.28244962845)*x[2]
        s[2,4,2,1]=(-0.11059763341)*x[0]**o+(0.276920644999)*x[0]+(-0.869145286116)*x[1]**o+(-0.403772968269)*x[1]+(0.319872723054)*x[2]**o+(-0.172168932529)*x[2]
        s[3,0,0,0]=(-0.309323319641)*x[0]**o+(-0.520023896018)*x[0]+(-0.91435931387)*x[1]**o+(-0.994973476595)*x[1]+(-0.372209395986)*x[2]**o+(0.980356617984)*x[2]
        s[3,0,0,1]=(0.741365113034)*x[0]**o+(-0.42466697031)*x[0]+(-0.800148478866)*x[1]**o+(0.891726804851)*x[1]+(0.173578031033)*x[2]**o+(-0.830047254257)*x[2]
        s[3,0,1,0]=(0.292501009071)*x[0]**o+(-0.643570264019)*x[0]+(0.549854764281)*x[1]**o+(-0.359686767247)*x[1]+(-0.242285567291)*x[2]**o+(-0.905609679312)*x[2]
        s[3,0,1,1]=(-0.340043108142)*x[0]**o+(0.654957588422)*x[0]+(-0.709296972181)*x[1]**o+(0.997889126174)*x[1]+(-0.0231005104624)*x[2]**o+(0.736312548493)*x[2]
        s[3,0,2,0]=(-0.724460864229)*x[0]**o+(-0.367115506582)*x[0]+(0.932338820643)*x[1]**o+(0.386420866744)*x[1]+(0.0937739144019)*x[2]**o+(0.980197176095)*x[2]
        s[3,0,2,1]=(-0.123268193799)*x[0]**o+(0.775389790643)*x[0]+(0.374761099968)*x[1]**o+(-0.410566663234)*x[1]+(0.61390455683)*x[2]**o+(-0.424980785321)*x[2]
        s[3,1,0,0]=(-0.931705529761)*x[0]**o+(-0.883420554924)*x[0]+(0.364171802711)*x[1]**o+(-0.120720945763)*x[1]+(-0.264662446682)*x[2]**o+(0.694245672188)*x[2]
        s[3,1,0,1]=(0.507167558745)*x[0]**o+(0.41310141465)*x[0]+(0.951808421277)*x[1]**o+(0.194810425697)*x[1]+(0.656365011417)*x[2]**o+(-0.869365458629)*x[2]
        s[3,1,1,0]=(-0.00813347718406)*x[0]**o+(0.410334738969)*x[0]+(0.749864796873)*x[1]**o+(-0.592730297251)*x[1]+(-0.739489287628)*x[2]**o+(0.620937075039)*x[2]
        s[3,1,1,1]=(-0.598495133204)*x[0]**o+(0.933561890579)*x[0]+(0.711594544245)*x[1]**o+(-0.307136681784)*x[1]+(-0.391014626741)*x[2]**o+(0.237572771088)*x[2]
        s[3,1,2,0]=(0.0815836502329)*x[0]**o+(0.265301269129)*x[0]+(0.151163193573)*x[1]**o+(0.734126505497)*x[1]+(-0.607886397818)*x[2]**o+(-0.96735041904)*x[2]
        s[3,1,2,1]=(0.264092716207)*x[0]**o+(-0.0674826292731)*x[0]+(0.163729761978)*x[1]**o+(-0.76481590065)*x[1]+(0.23022017141)*x[2]**o+(0.0749358120678)*x[2]
        s[3,2,0,0]=(-0.0551562544206)*x[0]**o+(0.529626011841)*x[0]+(0.394111309195)*x[1]**o+(-0.315588836981)*x[1]+(-0.821692537082)*x[2]**o+(0.687214393749)*x[2]
        s[3,2,0,1]=(-0.400905852693)*x[0]**o+(-0.588175641647)*x[0]+(-0.759129469971)*x[1]**o+(-0.234537321811)*x[1]+(-0.506929882395)*x[2]**o+(-0.354376736058)*x[2]
        s[3,2,1,0]=(0.517928718456)*x[0]**o+(0.340801973663)*x[0]+(0.648735816398)*x[1]**o+(0.566053973455)*x[1]+(0.42463576352)*x[2]**o+(0.0873509981339)*x[2]
        s[3,2,1,1]=(0.452898115077)*x[0]**o+(0.122517234268)*x[0]+(-0.98577530096)*x[1]**o+(-0.353319422026)*x[1]+(0.926541753948)*x[2]**o+(-0.393031058333)*x[2]
        s[3,2,2,0]=(0.667750444088)*x[0]**o+(0.00375536977016)*x[0]+(0.268195379568)*x[1]**o+(0.50950653377)*x[1]+(-0.262580606419)*x[2]**o+(-0.417690499605)*x[2]
        s[3,2,2,1]=(-0.16803158003)*x[0]**o+(0.186259658616)*x[0]+(-0.200850683296)*x[1]**o+(-0.443680999529)*x[1]+(0.0208407252003)*x[2]**o+(-0.724292481324)*x[2]
        s[3,3,0,0]=(-0.35345741052)*x[0]**o+(0.217412970886)*x[0]+(0.27307889565)*x[1]**o+(0.267600255726)*x[1]+(-0.973212265009)*x[2]**o+(-0.129667297553)*x[2]
        s[3,3,0,1]=(-0.891104426101)*x[0]**o+(-0.613334258669)*x[0]+(-0.373828493642)*x[1]**o+(-0.0702748602911)*x[1]+(0.236710052698)*x[2]**o+(-0.358872144334)*x[2]
        s[3,3,1,0]=(-0.0928125660128)*x[0]**o+(0.413580667461)*x[0]+(0.8369960454)*x[1]**o+(0.77176338641)*x[1]+(-0.0577873035722)*x[2]**o+(-0.702932886813)*x[2]
        s[3,3,1,1]=(-0.241426495256)*x[0]**o+(0.525912190909)*x[0]+(-0.428192935989)*x[1]**o+(0.892165668837)*x[1]+(0.0334032549752)*x[2]**o+(0.882084846304)*x[2]
        s[3,3,2,0]=(-0.891362453397)*x[0]**o+(-0.339599362928)*x[0]+(0.473329812864)*x[1]**o+(-0.371616732026)*x[1]+(0.660699523287)*x[2]**o+(-0.119309478457)*x[2]
        s[3,3,2,1]=(0.243522118981)*x[0]**o+(0.155411321418)*x[0]+(-0.230275201921)*x[1]**o+(0.63515968566)*x[1]+(0.794898971569)*x[2]**o+(0.409779242619)*x[2]
        s[3,4,0,0]=(-0.854632433212)*x[0]**o+(-0.195836100347)*x[0]+(-0.437665256019)*x[1]**o+(0.576362236865)*x[1]+(-0.0965898541616)*x[2]**o+(0.477972926354)*x[2]
        s[3,4,0,1]=(0.570617061345)*x[0]**o+(-0.0667297487904)*x[0]+(0.81823573301)*x[1]**o+(-0.721189008734)*x[1]+(0.314807619532)*x[2]**o+(-0.376322628683)*x[2]
        s[3,4,1,0]=(-0.580668389702)*x[0]**o+(-0.60760839523)*x[0]+(0.686535934072)*x[1]**o+(0.0779142152446)*x[1]+(0.373571653006)*x[2]**o+(-0.88636603136)*x[2]
        s[3,4,1,1]=(0.987949055842)*x[0]**o+(-0.370309316213)*x[0]+(0.491077591851)*x[1]**o+(0.544978214302)*x[1]+(0.0979248628028)*x[2]**o+(0.0838359159228)*x[2]
        s[3,4,2,0]=(-0.0541483763088)*x[0]**o+(0.491394550602)*x[0]+(0.492110322334)*x[1]**o+(-0.113182373476)*x[1]+(0.807657591761)*x[2]**o+(0.387085950646)*x[2]
        s[3,4,2,1]=(-0.420727803772)*x[0]**o+(-0.105139523222)*x[0]+(-0.51755674286)*x[1]**o+(-0.219056797552)*x[1]+(-0.0414122967561)*x[2]**o+(-0.161491064048)*x[2]
        ref[0,0,0,0]=(0.776303486002)/(o+1.)+(-0.66485490298)
        ref[0,0,0,1]=(0.73709920253)/(o+1.)+(-0.0305237564282)
        ref[0,0,1,0]=(1.62934422719)/(o+1.)+(-0.118729058429)
        ref[0,0,1,1]=(1.21227703769)/(o+1.)+(-1.04950265251)
        ref[0,0,2,0]=(-0.989239328736)/(o+1.)+(0.0861717711464)
        ref[0,0,2,1]=(-0.545045045276)/(o+1.)+(-0.453531722457)
        ref[0,1,0,0]=(-0.212230003001)/(o+1.)+(-0.0980278628049)
        ref[0,1,0,1]=(-0.623507761533)/(o+1.)+(0.831239111696)
        ref[0,1,1,0]=(1.20308213631)/(o+1.)+(0.21904770931)
        ref[0,1,1,1]=(-0.414005236391)/(o+1.)+(0.734379105797)
        ref[0,1,2,0]=(1.69005154613)/(o+1.)+(0.599453731488)
        ref[0,1,2,1]=(1.30699723672)/(o+1.)+(0.447731919557)
        ref[0,2,0,0]=(-0.274684837419)/(o+1.)+(-0.281541795502)
        ref[0,2,0,1]=(0.792946401732)/(o+1.)+(0.598271545783)
        ref[0,2,1,0]=(0.972930281169)/(o+1.)+(-0.692993050277)
        ref[0,2,1,1]=(-0.0948038476807)/(o+1.)+(-1.02610432679)
        ref[0,2,2,0]=(-0.417537958819)/(o+1.)+(0.0995721234709)
        ref[0,2,2,1]=(-0.343497078715)/(o+1.)+(0.287457935028)
        ref[0,3,0,0]=(1.06075362632)/(o+1.)+(0.392105127981)
        ref[0,3,0,1]=(-1.00743816734)/(o+1.)+(0.55530959011)
        ref[0,3,1,0]=(-1.35069907354)/(o+1.)+(-0.151109418893)
        ref[0,3,1,1]=(-0.209730918326)/(o+1.)+(-0.44139704026)
        ref[0,3,2,0]=(0.119571029049)/(o+1.)+(-0.788918379535)
        ref[0,3,2,1]=(0.834197985747)/(o+1.)+(0.0545741077131)
        ref[0,4,0,0]=(-0.713019032656)/(o+1.)+(0.0605014475136)
        ref[0,4,0,1]=(-0.188400834074)/(o+1.)+(0.0384669533136)
        ref[0,4,1,0]=(0.663494658006)/(o+1.)+(-0.197020376502)
        ref[0,4,1,1]=(1.57999775376)/(o+1.)+(-0.158042262352)
        ref[0,4,2,0]=(-0.825055197978)/(o+1.)+(0.190884563494)
        ref[0,4,2,1]=(-0.230079936711)/(o+1.)+(-0.619555317351)
        ref[1,0,0,0]=(0.0484302346511)/(o+1.)+(-0.277365092912)
        ref[1,0,0,1]=(-1.18699865107)/(o+1.)+(0.283637899436)
        ref[1,0,1,0]=(0.347990624042)/(o+1.)+(0.00803167578244)
        ref[1,0,1,1]=(-0.160064935171)/(o+1.)+(0.185408737905)
        ref[1,0,2,0]=(1.35245255927)/(o+1.)+(-0.531140014353)
        ref[1,0,2,1]=(1.13553833557)/(o+1.)+(0.152237391837)
        ref[1,1,0,0]=(0.99189846918)/(o+1.)+(0.338006421989)
        ref[1,1,0,1]=(-0.0255713914667)/(o+1.)+(0.12819321635)
        ref[1,1,1,0]=(-0.622986210996)/(o+1.)+(0.554888335978)
        ref[1,1,1,1]=(-0.0472909279168)/(o+1.)+(0.185961952963)
        ref[1,1,2,0]=(-0.729785677925)/(o+1.)+(-0.63936846664)
        ref[1,1,2,1]=(1.81104663684)/(o+1.)+(0.515008349399)
        ref[1,2,0,0]=(-1.54598989482)/(o+1.)+(0.284091818483)
        ref[1,2,0,1]=(0.182955266737)/(o+1.)+(0.00729582783438)
        ref[1,2,1,0]=(0.0678788403948)/(o+1.)+(0.0566178324469)
        ref[1,2,1,1]=(1.74208188933)/(o+1.)+(0.659881443367)
        ref[1,2,2,0]=(-0.0382440547455)/(o+1.)+(-0.535487134074)
        ref[1,2,2,1]=(0.198093668675)/(o+1.)+(-0.89001529806)
        ref[1,3,0,0]=(1.62490272607)/(o+1.)+(0.765524305033)
        ref[1,3,0,1]=(0.279965727369)/(o+1.)+(0.214570595298)
        ref[1,3,1,0]=(0.304113915286)/(o+1.)+(0.76954018039)
        ref[1,3,1,1]=(-0.885314465993)/(o+1.)+(0.62922698534)
        ref[1,3,2,0]=(0.434655075403)/(o+1.)+(-0.793446006776)
        ref[1,3,2,1]=(0.242224993913)/(o+1.)+(0.35107235084)
        ref[1,4,0,0]=(0.230763217193)/(o+1.)+(-0.155861223318)
        ref[1,4,0,1]=(0.808280805525)/(o+1.)+(0.0274401101125)
        ref[1,4,1,0]=(1.00266325917)/(o+1.)+(0.58007079489)
        ref[1,4,1,1]=(-0.641806023088)/(o+1.)+(-0.767874882808)
        ref[1,4,2,0]=(-1.27895710644)/(o+1.)+(0.573630110668)
        ref[1,4,2,1]=(-1.85031165866)/(o+1.)+(0.233982522704)
        ref[2,0,0,0]=(-1.36306650425)/(o+1.)+(-0.0533791412027)
        ref[2,0,0,1]=(1.10778233734)/(o+1.)+(-0.109894920649)
        ref[2,0,1,0]=(-0.607343404586)/(o+1.)+(-0.0381401236341)
        ref[2,0,1,1]=(-1.3016959741)/(o+1.)+(0.716358469177)
        ref[2,0,2,0]=(-1.65648421545)/(o+1.)+(0.675405459391)
        ref[2,0,2,1]=(1.43609344598)/(o+1.)+(-0.147175797578)
        ref[2,1,0,0]=(1.71599313418)/(o+1.)+(0.476324427863)
        ref[2,1,0,1]=(-0.197534803692)/(o+1.)+(0.74045848327)
        ref[2,1,1,0]=(1.15353139511)/(o+1.)+(-0.202917075571)
        ref[2,1,1,1]=(-0.646719728334)/(o+1.)+(0.089255774502)
        ref[2,1,2,0]=(-0.59247668041)/(o+1.)+(0.588338907289)
        ref[2,1,2,1]=(0.666534772105)/(o+1.)+(-0.171967067769)
        ref[2,2,0,0]=(0.10417442527)/(o+1.)+(-0.0373262159521)
        ref[2,2,0,1]=(1.6674625773)/(o+1.)+(0.33614460395)
        ref[2,2,1,0]=(1.40297752551)/(o+1.)+(-0.392281288816)
        ref[2,2,1,1]=(-0.72250539345)/(o+1.)+(-0.417148733502)
        ref[2,2,2,0]=(0.98912794704)/(o+1.)+(-1.19147084052)
        ref[2,2,2,1]=(-0.146690010568)/(o+1.)+(0.29184308053)
        ref[2,3,0,0]=(-0.431556699859)/(o+1.)+(-0.342220279916)
        ref[2,3,0,1]=(0.934033973509)/(o+1.)+(-1.26013765212)
        ref[2,3,1,0]=(0.387527880641)/(o+1.)+(0.223030814168)
        ref[2,3,1,1]=(2.43900464614)/(o+1.)+(-0.150886881122)
        ref[2,3,2,0]=(-1.75874429729)/(o+1.)+(0.116824608395)
        ref[2,3,2,1]=(-0.324304683574)/(o+1.)+(-0.817408425192)
        ref[2,4,0,0]=(-0.868761951233)/(o+1.)+(-0.0974552514252)
        ref[2,4,0,1]=(-0.416384616581)/(o+1.)+(0.282760450034)
        ref[2,4,1,0]=(-1.52621455953)/(o+1.)+(-0.448320971521)
        ref[2,4,1,1]=(-0.748074648234)/(o+1.)+(-0.346744035642)
        ref[2,4,2,0]=(0.50945468538)/(o+1.)+(0.0826481108988)
        ref[2,4,2,1]=(-0.659870196472)/(o+1.)+(-0.1495106279)
        ref[3,0,0,0]=(-1.5958920295)/(o+1.)+(-0.267320377314)
        ref[3,0,0,1]=(0.114794665201)/(o+1.)+(-0.181493709858)
        ref[3,0,1,0]=(0.600070206061)/(o+1.)+(-0.954433355289)
        ref[3,0,1,1]=(-1.07244059079)/(o+1.)+(1.19457963154)
        ref[3,0,2,0]=(0.301651870816)/(o+1.)+(0.499751268128)
        ref[3,0,2,1]=(0.865397463)/(o+1.)+(-0.0300788289559)
        ref[3,1,0,0]=(-0.832196173732)/(o+1.)+(-0.15494791425)
        ref[3,1,0,1]=(2.11534099144)/(o+1.)+(-0.130726809141)
        ref[3,1,1,0]=(0.00224203206025)/(o+1.)+(0.219270758379)
        ref[3,1,1,1]=(-0.277915215701)/(o+1.)+(0.431998989942)
        ref[3,1,2,0]=(-0.375139554013)/(o+1.)+(0.0160386777929)
        ref[3,1,2,1]=(0.658042649595)/(o+1.)+(-0.378681358928)
        ref[3,2,0,0]=(-0.482737482307)/(o+1.)+(0.450625784304)
        ref[3,2,0,1]=(-1.66696520506)/(o+1.)+(-0.588544849758)
        ref[3,2,1,0]=(1.59130029837)/(o+1.)+(0.497103472626)
        ref[3,2,1,1]=(0.393664568064)/(o+1.)+(-0.311916623046)
        ref[3,2,2,0]=(0.673365217238)/(o+1.)+(0.0477857019673)
        ref[3,2,2,1]=(-0.348041538126)/(o+1.)+(-0.490856911118)
        ref[3,3,0,0]=(-1.05359077988)/(o+1.)+(0.17767296453)
        ref[3,3,0,1]=(-1.02822286704)/(o+1.)+(-0.521240631647)
        ref[3,3,1,0]=(0.686396175815)/(o+1.)+(0.241205583529)
        ref[3,3,1,1]=(-0.63621617627)/(o+1.)+(1.15008135302)
        ref[3,3,2,0]=(0.242666882753)/(o+1.)+(-0.415262786705)
        ref[3,3,2,1]=(0.808145888629)/(o+1.)+(0.600175124848)
        ref[3,4,0,0]=(-1.38888754339)/(o+1.)+(0.429249531436)
        ref[3,4,0,1]=(1.70366041389)/(o+1.)+(-0.582120693104)
        ref[3,4,1,0]=(0.479439197375)/(o+1.)+(-0.708030105673)
        ref[3,4,1,1]=(1.5769515105)/(o+1.)+(0.129252407006)
        ref[3,4,2,0]=(1.24561953779)/(o+1.)+(0.382649063886)
        ref[3,4,2,1]=(-0.979696843387)/(o+1.)+(-0.242843692411)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Solution_rank0(self):
      """
      tests integral of rank 0 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.496833468021)*x[0]**o+(-0.119656931817)*x[0]+(-0.456435785007)*x[1]**o+(0.0378732106767)*x[1]
        ref=(-0.953269253027)/(o+1.)+(-0.04089186057)
      else:
        s=(0.960756800558)*x[0]**o+(-0.297563111204)*x[0]+(0.657965711967)*x[1]**o+(-0.942706539964)*x[1]+(-0.460398488427)*x[2]**o+(0.539495746387)*x[2]
        ref=(1.1583240241)/(o+1.)+(-0.350386952391)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Solution_rank1(self):
      """
      tests integral of rank 1 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.521294857582)*x[0]**o+(0.813966078452)*x[0]+(-0.460972018313)*x[1]**o+(-0.274643646978)*x[1]
        s[1]=(0.64160433254)*x[0]**o+(0.100182020329)*x[0]+(-0.877560875536)*x[1]**o+(0.0967945574365)*x[1]
        ref[0]=(-0.982266875895)/(o+1.)+(0.269661215737)
        ref[1]=(-0.235956542996)/(o+1.)+(0.0984882888828)
      else:
        s[0]=(-0.935408759921)*x[0]**o+(-0.636823488942)*x[0]+(-0.937973746046)*x[1]**o+(-0.548377694248)*x[1]+(-0.325664041787)*x[2]**o+(-0.859644239487)*x[2]
        s[1]=(-0.618325226925)*x[0]**o+(-0.464240689062)*x[0]+(-0.333395219266)*x[1]**o+(0.75384521826)*x[1]+(0.812529741058)*x[2]**o+(0.293845075804)*x[2]
        ref[0]=(-2.19904654775)/(o+1.)+(-1.02242271134)
        ref[1]=(-0.139190705133)/(o+1.)+(0.291724802501)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Solution_rank2(self):
      """
      tests integral of rank 2 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.745131837074)*x[0]**o+(0.954368069641)*x[0]+(0.582989014357)*x[1]**o+(0.649474790852)*x[1]
        s[0,1]=(0.0682420943207)*x[0]**o+(0.890255947523)*x[0]+(-0.501961353506)*x[1]**o+(-0.468049947986)*x[1]
        s[0,2]=(-0.376326981417)*x[0]**o+(0.318377164087)*x[0]+(0.518313700818)*x[1]**o+(-0.624749027625)*x[1]
        s[0,3]=(-0.0127578587596)*x[0]**o+(0.510088777647)*x[0]+(-0.00731027766723)*x[1]**o+(0.404116290004)*x[1]
        s[0,4]=(-0.566065672786)*x[0]**o+(-0.797903214842)*x[0]+(0.305896347185)*x[1]**o+(-0.44763489831)*x[1]
        s[1,0]=(0.136108309012)*x[0]**o+(0.8858491692)*x[0]+(0.63174549565)*x[1]**o+(0.12884968267)*x[1]
        s[1,1]=(-0.420219415055)*x[0]**o+(-0.474087579489)*x[0]+(-0.639616730493)*x[1]**o+(-0.339713160471)*x[1]
        s[1,2]=(0.730001695284)*x[0]**o+(-0.281914379046)*x[0]+(-0.0897917239674)*x[1]**o+(-0.521366256858)*x[1]
        s[1,3]=(0.507225459863)*x[0]**o+(-0.279054149386)*x[0]+(-0.310997653203)*x[1]**o+(0.798594910647)*x[1]
        s[1,4]=(-0.473780124237)*x[0]**o+(-0.521982314393)*x[0]+(-0.66517104956)*x[1]**o+(0.906212634536)*x[1]
        s[2,0]=(0.981595846105)*x[0]**o+(0.306232557856)*x[0]+(0.144004451791)*x[1]**o+(0.590440862044)*x[1]
        s[2,1]=(-0.465282108724)*x[0]**o+(-0.911349734596)*x[0]+(0.724954062739)*x[1]**o+(0.0617692321941)*x[1]
        s[2,2]=(-0.698461944366)*x[0]**o+(0.69038728172)*x[0]+(-0.0999661256163)*x[1]**o+(0.81993889447)*x[1]
        s[2,3]=(-0.850065489825)*x[0]**o+(-0.137124990846)*x[0]+(0.390609467506)*x[1]**o+(0.0928979112394)*x[1]
        s[2,4]=(-0.831110879728)*x[0]**o+(0.85125177454)*x[0]+(0.322669554869)*x[1]**o+(0.199419883126)*x[1]
        s[3,0]=(0.560516174698)*x[0]**o+(-0.952963851912)*x[0]+(0.0840374448937)*x[1]**o+(0.852644015588)*x[1]
        s[3,1]=(-0.738776742753)*x[0]**o+(-0.718650015551)*x[0]+(-0.48025826111)*x[1]**o+(0.163919964403)*x[1]
        s[3,2]=(-0.498985570012)*x[0]**o+(0.796130560575)*x[0]+(0.78812252561)*x[1]**o+(0.427402467946)*x[1]
        s[3,3]=(0.900695900592)*x[0]**o+(-0.68276543281)*x[0]+(-0.448699598266)*x[1]**o+(0.338876907438)*x[1]
        s[3,4]=(-0.101161698815)*x[0]**o+(0.24782725682)*x[0]+(0.194568299968)*x[1]**o+(-0.232062305313)*x[1]
        ref[0,0]=(1.32812085143)/(o+1.)+(0.801921430247)
        ref[0,1]=(-0.433719259185)/(o+1.)+(0.211102999768)
        ref[0,2]=(0.141986719401)/(o+1.)+(-0.153185931769)
        ref[0,3]=(-0.0200681364268)/(o+1.)+(0.457102533825)
        ref[0,4]=(-0.260169325601)/(o+1.)+(-0.622769056576)
        ref[1,0]=(0.767853804662)/(o+1.)+(0.507349425935)
        ref[1,1]=(-1.05983614555)/(o+1.)+(-0.40690036998)
        ref[1,2]=(0.640209971317)/(o+1.)+(-0.401640317952)
        ref[1,3]=(0.19622780666)/(o+1.)+(0.259770380631)
        ref[1,4]=(-1.1389511738)/(o+1.)+(0.192115160071)
        ref[2,0]=(1.1256002979)/(o+1.)+(0.44833670995)
        ref[2,1]=(0.259671954016)/(o+1.)+(-0.424790251201)
        ref[2,2]=(-0.798428069982)/(o+1.)+(0.755163088095)
        ref[2,3]=(-0.459456022319)/(o+1.)+(-0.0221135398034)
        ref[2,4]=(-0.508441324858)/(o+1.)+(0.525335828833)
        ref[3,0]=(0.644553619592)/(o+1.)+(-0.0501599181621)
        ref[3,1]=(-1.21903500386)/(o+1.)+(-0.277365025574)
        ref[3,2]=(0.289136955597)/(o+1.)+(0.611766514261)
        ref[3,3]=(0.451996302326)/(o+1.)+(-0.171944262686)
        ref[3,4]=(0.0934066011531)/(o+1.)+(0.00788247575383)
      else:
        s[0,0]=(0.581126436866)*x[0]**o+(0.307736169306)*x[0]+(0.46264603817)*x[1]**o+(-0.658226966972)*x[1]+(0.560314472945)*x[2]**o+(-0.71392138112)*x[2]
        s[0,1]=(0.41763244695)*x[0]**o+(-0.249710720358)*x[0]+(0.318827586572)*x[1]**o+(-0.766338584888)*x[1]+(-0.235847336194)*x[2]**o+(0.720822035745)*x[2]
        s[0,2]=(0.753177939244)*x[0]**o+(0.575601283201)*x[0]+(0.833261599709)*x[1]**o+(0.285673583448)*x[1]+(-0.850617074406)*x[2]**o+(0.976207021957)*x[2]
        s[0,3]=(-0.246323991583)*x[0]**o+(-0.425601468065)*x[0]+(-0.631382577618)*x[1]**o+(0.303177559003)*x[1]+(-0.772499313037)*x[2]**o+(0.0976456268892)*x[2]
        s[0,4]=(0.459782891083)*x[0]**o+(0.695607546341)*x[0]+(-0.758437555397)*x[1]**o+(-0.0467038938803)*x[1]+(0.077637691262)*x[2]**o+(0.632686471042)*x[2]
        s[1,0]=(0.62677312896)*x[0]**o+(-0.538520494446)*x[0]+(0.322526032671)*x[1]**o+(0.155570348616)*x[1]+(-0.645711565238)*x[2]**o+(-0.505639404827)*x[2]
        s[1,1]=(0.760035088253)*x[0]**o+(-0.653433840013)*x[0]+(0.541102747261)*x[1]**o+(0.102121826295)*x[1]+(0.203147157734)*x[2]**o+(0.825692237055)*x[2]
        s[1,2]=(0.83286976398)*x[0]**o+(-0.483950138218)*x[0]+(-0.988388711773)*x[1]**o+(0.276946264031)*x[1]+(-0.769100710972)*x[2]**o+(-0.857758652165)*x[2]
        s[1,3]=(0.965261287995)*x[0]**o+(-0.582299103263)*x[0]+(0.650118954781)*x[1]**o+(-0.714080635544)*x[1]+(-0.984672595257)*x[2]**o+(-0.890248882426)*x[2]
        s[1,4]=(0.780505914972)*x[0]**o+(-0.469283145875)*x[0]+(0.65642477278)*x[1]**o+(-0.0385309033271)*x[1]+(-0.838684124311)*x[2]**o+(0.715186480156)*x[2]
        s[2,0]=(-0.957976655406)*x[0]**o+(-0.326902969464)*x[0]+(0.519819227981)*x[1]**o+(0.526859170123)*x[1]+(-0.83643924332)*x[2]**o+(-0.833906618158)*x[2]
        s[2,1]=(0.511380988321)*x[0]**o+(0.205939303999)*x[0]+(-0.560011037377)*x[1]**o+(-0.74412420954)*x[1]+(-0.413858375939)*x[2]**o+(-0.818302325723)*x[2]
        s[2,2]=(0.669120009156)*x[0]**o+(-0.680617749061)*x[0]+(-0.855851359703)*x[1]**o+(0.681103383833)*x[1]+(0.358345655307)*x[2]**o+(-0.834176731331)*x[2]
        s[2,3]=(0.113991673968)*x[0]**o+(-0.167621358311)*x[0]+(-0.824506843066)*x[1]**o+(0.576032397062)*x[1]+(0.98813748839)*x[2]**o+(-0.954277123787)*x[2]
        s[2,4]=(0.758065768891)*x[0]**o+(0.550472010004)*x[0]+(0.541968389568)*x[1]**o+(-0.893196211859)*x[1]+(0.55846204484)*x[2]**o+(-0.154935285741)*x[2]
        s[3,0]=(0.319452905704)*x[0]**o+(0.416924332754)*x[0]+(0.695996329307)*x[1]**o+(-0.512723809281)*x[1]+(-0.0438582487583)*x[2]**o+(-0.202757150975)*x[2]
        s[3,1]=(0.300936121076)*x[0]**o+(0.0199168771219)*x[0]+(-0.79311689061)*x[1]**o+(0.120821678434)*x[1]+(0.832598541766)*x[2]**o+(-0.13378517127)*x[2]
        s[3,2]=(0.816251625164)*x[0]**o+(0.180324577125)*x[0]+(0.869902904875)*x[1]**o+(-0.300483761133)*x[1]+(0.405728532808)*x[2]**o+(-0.772682148538)*x[2]
        s[3,3]=(-0.0882536430174)*x[0]**o+(0.67244001887)*x[0]+(-0.295701327277)*x[1]**o+(-0.235065967505)*x[1]+(0.905094390215)*x[2]**o+(-0.132120939051)*x[2]
        s[3,4]=(0.976955651594)*x[0]**o+(0.36283338584)*x[0]+(0.21945555266)*x[1]**o+(-0.305668951381)*x[1]+(0.581731431193)*x[2]**o+(-0.308954336316)*x[2]
        ref[0,0]=(1.60408694798)/(o+1.)+(-0.532206089393)
        ref[0,1]=(0.500612697329)/(o+1.)+(-0.147613634751)
        ref[0,2]=(0.735822464547)/(o+1.)+(0.918740944303)
        ref[0,3]=(-1.65020588224)/(o+1.)+(-0.0123891410865)
        ref[0,4]=(-0.221016973052)/(o+1.)+(0.640795061752)
        ref[1,0]=(0.303587596392)/(o+1.)+(-0.444294775329)
        ref[1,1]=(1.50428499325)/(o+1.)+(0.137190111668)
        ref[1,2]=(-0.924619658766)/(o+1.)+(-0.532381263176)
        ref[1,3]=(0.630707647519)/(o+1.)+(-1.09331431062)
        ref[1,4]=(0.598246563441)/(o+1.)+(0.103686215477)
        ref[2,0]=(-1.27459667074)/(o+1.)+(-0.31697520875)
        ref[2,1]=(-0.462488424995)/(o+1.)+(-0.678243615632)
        ref[2,2]=(0.17161430476)/(o+1.)+(-0.416845548279)
        ref[2,3]=(0.277622319292)/(o+1.)+(-0.272933042518)
        ref[2,4]=(1.8584962033)/(o+1.)+(-0.248829743798)
        ref[3,0]=(0.971590986252)/(o+1.)+(-0.149278313751)
        ref[3,1]=(0.340417772233)/(o+1.)+(0.00347669214316)
        ref[3,2]=(2.09188306285)/(o+1.)+(-0.446420666273)
        ref[3,3]=(0.521139419921)/(o+1.)+(0.152626556157)
        ref[3,4]=(1.77814263545)/(o+1.)+(-0.125894950929)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Solution_rank3(self):
      """
      tests integral of rank 3 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.327469743415)*x[0]**o+(0.986515360087)*x[0]+(-0.264743586601)*x[1]**o+(-0.417911841646)*x[1]
        s[0,0,1]=(-0.367335531482)*x[0]**o+(-0.294564095391)*x[0]+(0.948849162205)*x[1]**o+(-0.672727814952)*x[1]
        s[0,1,0]=(0.201971500659)*x[0]**o+(0.715648302042)*x[0]+(0.200753457036)*x[1]**o+(-0.447431022755)*x[1]
        s[0,1,1]=(0.391503527712)*x[0]**o+(0.561563485529)*x[0]+(0.0756343714486)*x[1]**o+(-0.232567658257)*x[1]
        s[1,0,0]=(0.0416189159889)*x[0]**o+(0.933172580543)*x[0]+(0.906370262092)*x[1]**o+(0.843408035954)*x[1]
        s[1,0,1]=(-0.0489328569668)*x[0]**o+(0.305933018225)*x[0]+(0.459563500395)*x[1]**o+(-0.305696910459)*x[1]
        s[1,1,0]=(-0.166432892993)*x[0]**o+(0.323257103336)*x[0]+(-0.876673724336)*x[1]**o+(-0.246533332987)*x[1]
        s[1,1,1]=(0.331551737096)*x[0]**o+(-0.473411829001)*x[0]+(0.658917172856)*x[1]**o+(-0.645603256937)*x[1]
        s[2,0,0]=(-0.779837619461)*x[0]**o+(0.726847293406)*x[0]+(0.756283586425)*x[1]**o+(0.820695230411)*x[1]
        s[2,0,1]=(-0.0446700642358)*x[0]**o+(-0.00612226458731)*x[0]+(-0.255350731215)*x[1]**o+(0.279072151991)*x[1]
        s[2,1,0]=(0.952754783643)*x[0]**o+(0.575497372286)*x[0]+(-0.996409669059)*x[1]**o+(-0.993504128277)*x[1]
        s[2,1,1]=(-0.964146199975)*x[0]**o+(-0.89830816406)*x[0]+(0.27874102283)*x[1]**o+(-0.193698734442)*x[1]
        s[3,0,0]=(-0.22434829265)*x[0]**o+(-0.451123824419)*x[0]+(0.0996011155458)*x[1]**o+(-0.0196796989303)*x[1]
        s[3,0,1]=(0.479762785194)*x[0]**o+(-0.0875868570981)*x[0]+(-0.604020314973)*x[1]**o+(-0.948950136203)*x[1]
        s[3,1,0]=(0.380546378709)*x[0]**o+(0.968895520825)*x[0]+(0.0717757394195)*x[1]**o+(0.00273829931123)*x[1]
        s[3,1,1]=(0.999877803305)*x[0]**o+(0.726497905193)*x[0]+(-0.156097974361)*x[1]**o+(0.875069921439)*x[1]
        s[4,0,0]=(0.298060895282)*x[0]**o+(-0.25017046132)*x[0]+(-0.6409928584)*x[1]**o+(0.237325951363)*x[1]
        s[4,0,1]=(0.176241833972)*x[0]**o+(0.362302520221)*x[0]+(-0.632667110076)*x[1]**o+(0.206684577439)*x[1]
        s[4,1,0]=(-0.0658280231422)*x[0]**o+(-0.830173560366)*x[0]+(-0.408668580492)*x[1]**o+(0.131219154432)*x[1]
        s[4,1,1]=(-0.739357616315)*x[0]**o+(0.331825079796)*x[0]+(-0.950532071867)*x[1]**o+(0.430637319576)*x[1]
        s[5,0,0]=(-0.374056178585)*x[0]**o+(-0.503558441322)*x[0]+(0.253751100374)*x[1]**o+(-0.169874527745)*x[1]
        s[5,0,1]=(-0.540168118335)*x[0]**o+(-0.599437156114)*x[0]+(-0.545843659509)*x[1]**o+(-0.144502584314)*x[1]
        s[5,1,0]=(0.734871232877)*x[0]**o+(0.294937992469)*x[0]+(0.318246966782)*x[1]**o+(0.301215884904)*x[1]
        s[5,1,1]=(-0.48224128272)*x[0]**o+(-0.971913873914)*x[0]+(-0.853997429288)*x[1]**o+(0.911950472127)*x[1]
        ref[0,0,0]=(0.062726156814)/(o+1.)+(0.284301759221)
        ref[0,0,1]=(0.581513630723)/(o+1.)+(-0.483645955172)
        ref[0,1,0]=(0.402724957695)/(o+1.)+(0.134108639643)
        ref[0,1,1]=(0.467137899161)/(o+1.)+(0.164497913636)
        ref[1,0,0]=(0.947989178081)/(o+1.)+(0.888290308248)
        ref[1,0,1]=(0.410630643428)/(o+1.)+(0.000118053882864)
        ref[1,1,0]=(-1.04310661733)/(o+1.)+(0.0383618851745)
        ref[1,1,1]=(0.990468909952)/(o+1.)+(-0.559507542969)
        ref[2,0,0]=(-0.0235540330353)/(o+1.)+(0.773771261908)
        ref[2,0,1]=(-0.300020795451)/(o+1.)+(0.136474943702)
        ref[2,1,0]=(-0.043654885416)/(o+1.)+(-0.209003377996)
        ref[2,1,1]=(-0.685405177144)/(o+1.)+(-0.546003449251)
        ref[3,0,0]=(-0.124747177104)/(o+1.)+(-0.235401761675)
        ref[3,0,1]=(-0.124257529779)/(o+1.)+(-0.51826849665)
        ref[3,1,0]=(0.452322118128)/(o+1.)+(0.485816910068)
        ref[3,1,1]=(0.843779828944)/(o+1.)+(0.800783913316)
        ref[4,0,0]=(-0.342931963118)/(o+1.)+(-0.00642225497804)
        ref[4,0,1]=(-0.456425276104)/(o+1.)+(0.28449354883)
        ref[4,1,0]=(-0.474496603634)/(o+1.)+(-0.349477202967)
        ref[4,1,1]=(-1.68988968818)/(o+1.)+(0.381231199686)
        ref[5,0,0]=(-0.120305078211)/(o+1.)+(-0.336716484533)
        ref[5,0,1]=(-1.08601177784)/(o+1.)+(-0.371969870214)
        ref[5,1,0]=(1.05311819966)/(o+1.)+(0.298076938687)
        ref[5,1,1]=(-1.33623871201)/(o+1.)+(-0.0299817008935)
      else:
        s[0,0,0]=(0.907642728831)*x[0]**o+(0.24755436099)*x[0]+(0.329429496561)*x[1]**o+(0.195804615806)*x[1]+(0.635928214783)*x[2]**o+(0.286575554416)*x[2]
        s[0,0,1]=(-0.596463751156)*x[0]**o+(-0.573557158304)*x[0]+(-0.799372215818)*x[1]**o+(0.917917612224)*x[1]+(0.743915017372)*x[2]**o+(0.697340711065)*x[2]
        s[0,1,0]=(-0.226377288829)*x[0]**o+(0.11441928128)*x[0]+(0.915343934109)*x[1]**o+(0.0235994828916)*x[1]+(0.837206417747)*x[2]**o+(-0.809536949964)*x[2]
        s[0,1,1]=(-0.807835571238)*x[0]**o+(-0.313842967415)*x[0]+(0.143974822075)*x[1]**o+(0.772050096219)*x[1]+(0.0698457301843)*x[2]**o+(-0.878711129876)*x[2]
        s[1,0,0]=(0.806964016183)*x[0]**o+(0.906862034833)*x[0]+(0.656990650598)*x[1]**o+(0.858146180457)*x[1]+(-0.455748564953)*x[2]**o+(0.0415537957765)*x[2]
        s[1,0,1]=(0.930192572382)*x[0]**o+(0.406175274412)*x[0]+(0.522683472096)*x[1]**o+(-0.40154832923)*x[1]+(0.62436447915)*x[2]**o+(0.942684294891)*x[2]
        s[1,1,0]=(-0.190480904061)*x[0]**o+(-0.884913647118)*x[0]+(0.894504380487)*x[1]**o+(0.472867062235)*x[1]+(-0.226561577541)*x[2]**o+(0.0817340770676)*x[2]
        s[1,1,1]=(-0.785577817856)*x[0]**o+(0.286061214076)*x[0]+(-0.29241495776)*x[1]**o+(-0.885469334696)*x[1]+(-0.728917274605)*x[2]**o+(-0.400170372532)*x[2]
        s[2,0,0]=(-0.0443801095191)*x[0]**o+(-0.580757683753)*x[0]+(-0.680012816252)*x[1]**o+(-0.24174589493)*x[1]+(-0.32345463822)*x[2]**o+(0.488362285647)*x[2]
        s[2,0,1]=(0.427683981593)*x[0]**o+(0.709521979359)*x[0]+(-0.485171186139)*x[1]**o+(-0.466170281294)*x[1]+(0.668169758775)*x[2]**o+(-0.32376435425)*x[2]
        s[2,1,0]=(-0.322582923837)*x[0]**o+(-0.243805747954)*x[0]+(-0.282254505116)*x[1]**o+(0.276894252913)*x[1]+(0.425316275665)*x[2]**o+(-0.843958018195)*x[2]
        s[2,1,1]=(-0.172148420657)*x[0]**o+(0.288740472914)*x[0]+(0.3130139801)*x[1]**o+(0.0909556111434)*x[1]+(-0.134361659944)*x[2]**o+(-0.958259889515)*x[2]
        s[3,0,0]=(0.509175759598)*x[0]**o+(0.901681022023)*x[0]+(-0.04878880083)*x[1]**o+(-0.292935057546)*x[1]+(-0.73460973687)*x[2]**o+(0.143266558286)*x[2]
        s[3,0,1]=(-0.525256676545)*x[0]**o+(0.330501687555)*x[0]+(-0.533311903214)*x[1]**o+(-0.558985165298)*x[1]+(-0.726910027746)*x[2]**o+(-0.430022743518)*x[2]
        s[3,1,0]=(-0.801052113128)*x[0]**o+(0.383850418649)*x[0]+(-0.74692163952)*x[1]**o+(0.119907629434)*x[1]+(-0.688440032866)*x[2]**o+(0.258856702215)*x[2]
        s[3,1,1]=(0.662483893926)*x[0]**o+(0.321392974926)*x[0]+(0.797623010994)*x[1]**o+(-0.555546458227)*x[1]+(0.140050792242)*x[2]**o+(0.819042937008)*x[2]
        s[4,0,0]=(0.208697187872)*x[0]**o+(0.347715032509)*x[0]+(0.900642395285)*x[1]**o+(-0.121096976816)*x[1]+(0.473919668027)*x[2]**o+(0.693182424772)*x[2]
        s[4,0,1]=(0.185689163302)*x[0]**o+(0.923973904464)*x[0]+(-0.091048447659)*x[1]**o+(-0.331567067076)*x[1]+(0.0639420800968)*x[2]**o+(0.0645745629094)*x[2]
        s[4,1,0]=(-0.097237311089)*x[0]**o+(0.0234720818008)*x[0]+(-0.476652797084)*x[1]**o+(-0.258741500235)*x[1]+(-0.644183703276)*x[2]**o+(-0.953514564143)*x[2]
        s[4,1,1]=(-0.647701499983)*x[0]**o+(0.514337030449)*x[0]+(-0.779360494982)*x[1]**o+(-0.467345311588)*x[1]+(0.166149183082)*x[2]**o+(0.12480608055)*x[2]
        s[5,0,0]=(-0.349937644226)*x[0]**o+(0.673132604681)*x[0]+(0.793703484606)*x[1]**o+(0.958967045691)*x[1]+(-0.361047380062)*x[2]**o+(0.758206969142)*x[2]
        s[5,0,1]=(0.415890706186)*x[0]**o+(0.748811321123)*x[0]+(-0.629959299026)*x[1]**o+(0.706391001435)*x[1]+(-0.831407461643)*x[2]**o+(0.698632094335)*x[2]
        s[5,1,0]=(-0.914414551259)*x[0]**o+(0.85324200236)*x[0]+(0.273543170188)*x[1]**o+(0.311955664492)*x[1]+(0.726975838083)*x[2]**o+(0.671917428159)*x[2]
        s[5,1,1]=(0.450170734221)*x[0]**o+(-0.928357299044)*x[0]+(-0.645229625203)*x[1]**o+(0.756723626659)*x[1]+(-0.85050032636)*x[2]**o+(0.70593481655)*x[2]
        ref[0,0,0]=(1.87300044017)/(o+1.)+(0.364967265606)
        ref[0,0,1]=(-0.651920949601)/(o+1.)+(0.520850582493)
        ref[0,1,0]=(1.52617306303)/(o+1.)+(-0.335759092896)
        ref[0,1,1]=(-0.594015018979)/(o+1.)+(-0.210252000536)
        ref[1,0,0]=(1.00820610183)/(o+1.)+(0.903281005533)
        ref[1,0,1]=(2.07724052363)/(o+1.)+(0.473655620037)
        ref[1,1,0]=(0.477461898885)/(o+1.)+(-0.165156253908)
        ref[1,1,1]=(-1.80691005022)/(o+1.)+(-0.499789246576)
        ref[2,0,0]=(-1.04784756399)/(o+1.)+(-0.167070646517)
        ref[2,0,1]=(0.610682554229)/(o+1.)+(-0.0402063280924)
        ref[2,1,0]=(-0.179521153288)/(o+1.)+(-0.405434756618)
        ref[2,1,1]=(0.00650389949838)/(o+1.)+(-0.289281902729)
        ref[3,0,0]=(-0.274222778102)/(o+1.)+(0.376006261382)
        ref[3,0,1]=(-1.7854786075)/(o+1.)+(-0.329253110631)
        ref[3,1,0]=(-2.23641378551)/(o+1.)+(0.381307375149)
        ref[3,1,1]=(1.60015769716)/(o+1.)+(0.292444726853)
        ref[4,0,0]=(1.58325925118)/(o+1.)+(0.459900240233)
        ref[4,0,1]=(0.15858279574)/(o+1.)+(0.328490700148)
        ref[4,1,0]=(-1.21807381145)/(o+1.)+(-0.594391991289)
        ref[4,1,1]=(-1.26091281188)/(o+1.)+(0.085898899706)
        ref[5,0,0]=(0.0827184603185)/(o+1.)+(1.19515330976)
        ref[5,0,1]=(-1.04547605448)/(o+1.)+(1.07691720845)
        ref[5,1,0]=(0.086104457012)/(o+1.)+(0.918557547505)
        ref[5,1,1]=(-1.04555921734)/(o+1.)+(0.267150572083)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Solution_rank4(self):
      """
      tests integral of rank 4 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.501583472757)*x[0]**o+(0.174276674809)*x[0]+(0.474559132041)*x[1]**o+(0.944197178797)*x[1]
        s[0,0,0,1]=(-0.850492891287)*x[0]**o+(-0.778714553562)*x[0]+(-0.864025500909)*x[1]**o+(0.905856111362)*x[1]
        s[0,0,1,0]=(0.837983487193)*x[0]**o+(0.789080933347)*x[0]+(0.0618430819498)*x[1]**o+(0.791928371037)*x[1]
        s[0,0,1,1]=(-0.678469228871)*x[0]**o+(0.702133334239)*x[0]+(-0.937443836374)*x[1]**o+(-0.130947584181)*x[1]
        s[0,0,2,0]=(0.84013488851)*x[0]**o+(0.683811818455)*x[0]+(0.981450595677)*x[1]**o+(-0.813078675111)*x[1]
        s[0,0,2,1]=(0.00148903051979)*x[0]**o+(-0.740776021571)*x[0]+(0.917048043634)*x[1]**o+(0.362159065586)*x[1]
        s[0,1,0,0]=(0.285313019304)*x[0]**o+(0.0108652086531)*x[0]+(0.323356296135)*x[1]**o+(0.541189488955)*x[1]
        s[0,1,0,1]=(0.545540725184)*x[0]**o+(-0.803623471109)*x[0]+(0.861711269888)*x[1]**o+(0.614843604935)*x[1]
        s[0,1,1,0]=(0.44131453032)*x[0]**o+(-0.560146628455)*x[0]+(-0.390081444316)*x[1]**o+(-0.814215730743)*x[1]
        s[0,1,1,1]=(-0.120309068092)*x[0]**o+(-0.992223638563)*x[0]+(-0.757944249628)*x[1]**o+(-0.906900780015)*x[1]
        s[0,1,2,0]=(0.278503974882)*x[0]**o+(0.471665153392)*x[0]+(0.745287973293)*x[1]**o+(-0.572972893657)*x[1]
        s[0,1,2,1]=(0.840866880159)*x[0]**o+(-0.969714982065)*x[0]+(-0.181097771175)*x[1]**o+(-0.645812375407)*x[1]
        s[0,2,0,0]=(-0.791010044792)*x[0]**o+(-0.895489878363)*x[0]+(-0.32385199033)*x[1]**o+(0.892979821152)*x[1]
        s[0,2,0,1]=(0.966330767971)*x[0]**o+(0.757433262697)*x[0]+(0.859264987879)*x[1]**o+(0.836273715628)*x[1]
        s[0,2,1,0]=(0.764347587379)*x[0]**o+(0.498460349969)*x[0]+(0.896024738679)*x[1]**o+(0.341308865054)*x[1]
        s[0,2,1,1]=(-0.276964987386)*x[0]**o+(-0.294300496732)*x[0]+(-0.755886597374)*x[1]**o+(0.0964446535133)*x[1]
        s[0,2,2,0]=(0.933876243978)*x[0]**o+(0.694067023434)*x[0]+(-0.626635561624)*x[1]**o+(0.107630740485)*x[1]
        s[0,2,2,1]=(0.905601520891)*x[0]**o+(0.246861090683)*x[0]+(0.364620749293)*x[1]**o+(-0.34329084894)*x[1]
        s[0,3,0,0]=(-0.484050880674)*x[0]**o+(-0.344211446029)*x[0]+(-0.304888212647)*x[1]**o+(-0.918515747538)*x[1]
        s[0,3,0,1]=(0.352307753352)*x[0]**o+(-0.497295643302)*x[0]+(-0.0859041918813)*x[1]**o+(-0.566829131134)*x[1]
        s[0,3,1,0]=(0.458949103752)*x[0]**o+(-0.940473606591)*x[0]+(0.500249302466)*x[1]**o+(-0.192176323799)*x[1]
        s[0,3,1,1]=(0.714312790267)*x[0]**o+(0.219740572431)*x[0]+(-0.621517496057)*x[1]**o+(-0.79963217554)*x[1]
        s[0,3,2,0]=(-0.041379012308)*x[0]**o+(-0.674987043823)*x[0]+(-0.0165846458396)*x[1]**o+(0.643074958959)*x[1]
        s[0,3,2,1]=(-0.258839387368)*x[0]**o+(0.63938656748)*x[0]+(0.0042420382287)*x[1]**o+(-0.711660235823)*x[1]
        s[0,4,0,0]=(-0.905262277861)*x[0]**o+(0.995066153797)*x[0]+(0.895316712432)*x[1]**o+(0.794484306559)*x[1]
        s[0,4,0,1]=(0.858709170162)*x[0]**o+(-0.896059882819)*x[0]+(-0.322447413352)*x[1]**o+(0.356724683447)*x[1]
        s[0,4,1,0]=(-0.496930600469)*x[0]**o+(-0.715930772072)*x[0]+(0.514137034309)*x[1]**o+(0.738095094325)*x[1]
        s[0,4,1,1]=(-0.475961686127)*x[0]**o+(0.406077221295)*x[0]+(-0.934472851356)*x[1]**o+(0.184564622216)*x[1]
        s[0,4,2,0]=(0.50929205392)*x[0]**o+(-0.361652258356)*x[0]+(-0.828564581552)*x[1]**o+(0.864597967503)*x[1]
        s[0,4,2,1]=(-0.670572914815)*x[0]**o+(-0.917836333665)*x[0]+(0.329473913157)*x[1]**o+(0.649536686851)*x[1]
        s[1,0,0,0]=(0.507327212873)*x[0]**o+(-0.784119587219)*x[0]+(0.861200611677)*x[1]**o+(-0.537142113189)*x[1]
        s[1,0,0,1]=(0.307279452436)*x[0]**o+(-0.331535898217)*x[0]+(0.864752548746)*x[1]**o+(-0.41254012361)*x[1]
        s[1,0,1,0]=(0.723666220176)*x[0]**o+(-0.528755526062)*x[0]+(0.182515067109)*x[1]**o+(-0.33319767285)*x[1]
        s[1,0,1,1]=(-0.652532069809)*x[0]**o+(-0.497647481159)*x[0]+(-0.726676961172)*x[1]**o+(0.901803102587)*x[1]
        s[1,0,2,0]=(0.966840239828)*x[0]**o+(-0.559625286504)*x[0]+(0.856786274248)*x[1]**o+(-0.231280485522)*x[1]
        s[1,0,2,1]=(0.898750450503)*x[0]**o+(-0.241387492118)*x[0]+(-0.242302411171)*x[1]**o+(-0.931590118111)*x[1]
        s[1,1,0,0]=(0.298117967411)*x[0]**o+(-0.887837423565)*x[0]+(-0.804393274952)*x[1]**o+(-0.339940646012)*x[1]
        s[1,1,0,1]=(-0.3303942703)*x[0]**o+(0.22588379719)*x[0]+(0.427447092807)*x[1]**o+(-0.940368147505)*x[1]
        s[1,1,1,0]=(-0.227054817206)*x[0]**o+(0.76367061207)*x[0]+(-0.382609156261)*x[1]**o+(-0.869111355614)*x[1]
        s[1,1,1,1]=(0.538281073824)*x[0]**o+(-0.454205347349)*x[0]+(-0.121007985262)*x[1]**o+(0.0290163849694)*x[1]
        s[1,1,2,0]=(-0.888958335111)*x[0]**o+(0.779682676236)*x[0]+(-0.875785352701)*x[1]**o+(-0.891960671849)*x[1]
        s[1,1,2,1]=(-0.173332519418)*x[0]**o+(-0.104421077299)*x[0]+(0.906729441747)*x[1]**o+(-0.616671337017)*x[1]
        s[1,2,0,0]=(-0.305885548664)*x[0]**o+(-0.86724411868)*x[0]+(-0.787155086634)*x[1]**o+(-0.898599442906)*x[1]
        s[1,2,0,1]=(0.848368426968)*x[0]**o+(0.0802812152769)*x[0]+(-0.378057863356)*x[1]**o+(-0.428488725116)*x[1]
        s[1,2,1,0]=(-0.290675868214)*x[0]**o+(0.342045945019)*x[0]+(-0.421298816185)*x[1]**o+(-0.553807359319)*x[1]
        s[1,2,1,1]=(0.0289012545961)*x[0]**o+(0.423276922447)*x[0]+(-0.911749389958)*x[1]**o+(-0.39409305776)*x[1]
        s[1,2,2,0]=(0.904480195539)*x[0]**o+(-0.729896491926)*x[0]+(-0.21305322735)*x[1]**o+(-0.347444381896)*x[1]
        s[1,2,2,1]=(0.34039428449)*x[0]**o+(-0.246758392869)*x[0]+(-0.159391459268)*x[1]**o+(-0.630134063159)*x[1]
        s[1,3,0,0]=(0.202017077356)*x[0]**o+(-0.195969669611)*x[0]+(0.267580338059)*x[1]**o+(-0.274282886887)*x[1]
        s[1,3,0,1]=(-0.183473257308)*x[0]**o+(-0.646999501714)*x[0]+(-0.746037354076)*x[1]**o+(-0.180977233442)*x[1]
        s[1,3,1,0]=(0.895075853694)*x[0]**o+(-0.820090467371)*x[0]+(-0.040561177436)*x[1]**o+(0.749395910572)*x[1]
        s[1,3,1,1]=(-0.232039710082)*x[0]**o+(-0.633561589743)*x[0]+(0.862514378491)*x[1]**o+(0.782438844898)*x[1]
        s[1,3,2,0]=(-0.0228789327036)*x[0]**o+(0.604368007016)*x[0]+(0.413803296189)*x[1]**o+(0.651011054925)*x[1]
        s[1,3,2,1]=(-0.682393812631)*x[0]**o+(-0.508380759547)*x[0]+(-0.958515947271)*x[1]**o+(-0.396382186265)*x[1]
        s[1,4,0,0]=(0.772945258804)*x[0]**o+(0.981773233361)*x[0]+(-0.415876208997)*x[1]**o+(0.391268836611)*x[1]
        s[1,4,0,1]=(-0.111414446087)*x[0]**o+(-0.381712931273)*x[0]+(0.389894843435)*x[1]**o+(-0.276144369314)*x[1]
        s[1,4,1,0]=(0.544309017751)*x[0]**o+(0.920841251767)*x[0]+(-0.524322161349)*x[1]**o+(-0.745967887424)*x[1]
        s[1,4,1,1]=(0.0993088568573)*x[0]**o+(0.172455640245)*x[0]+(-0.310227024922)*x[1]**o+(0.503773577532)*x[1]
        s[1,4,2,0]=(-0.282979849337)*x[0]**o+(0.573540669768)*x[0]+(0.974343187515)*x[1]**o+(0.643215559873)*x[1]
        s[1,4,2,1]=(0.616949303196)*x[0]**o+(-0.206142689323)*x[0]+(0.594875479161)*x[1]**o+(0.390460380914)*x[1]
        s[2,0,0,0]=(-0.164379488326)*x[0]**o+(-0.613380471473)*x[0]+(-0.309735177264)*x[1]**o+(0.722815882927)*x[1]
        s[2,0,0,1]=(-0.443733554915)*x[0]**o+(-0.226908044072)*x[0]+(0.22281947804)*x[1]**o+(0.321557448698)*x[1]
        s[2,0,1,0]=(0.979229428359)*x[0]**o+(-0.410075647119)*x[0]+(-0.0869658073133)*x[1]**o+(0.472354461211)*x[1]
        s[2,0,1,1]=(0.279554115394)*x[0]**o+(0.982131537982)*x[0]+(0.865627129266)*x[1]**o+(0.0595046994689)*x[1]
        s[2,0,2,0]=(-0.928629057762)*x[0]**o+(-0.0888000293033)*x[0]+(0.925425619622)*x[1]**o+(0.286877459131)*x[1]
        s[2,0,2,1]=(-0.128076644671)*x[0]**o+(0.769985562586)*x[0]+(-0.326281257692)*x[1]**o+(-0.904253231113)*x[1]
        s[2,1,0,0]=(-7.49817068653e-05)*x[0]**o+(0.944960174862)*x[0]+(0.394130491438)*x[1]**o+(0.904576980195)*x[1]
        s[2,1,0,1]=(0.51260445176)*x[0]**o+(0.871495843591)*x[0]+(-0.0117378185696)*x[1]**o+(-0.447028599408)*x[1]
        s[2,1,1,0]=(0.385302842134)*x[0]**o+(0.68530194308)*x[0]+(0.908717705423)*x[1]**o+(0.0657269588494)*x[1]
        s[2,1,1,1]=(-0.175101480962)*x[0]**o+(-0.0126863839895)*x[0]+(0.544228922923)*x[1]**o+(-0.856957800534)*x[1]
        s[2,1,2,0]=(-0.74757544736)*x[0]**o+(-0.705933683532)*x[0]+(0.496113134499)*x[1]**o+(0.134139074285)*x[1]
        s[2,1,2,1]=(0.661170765505)*x[0]**o+(-0.0954233403413)*x[0]+(0.796712312861)*x[1]**o+(0.941668952301)*x[1]
        s[2,2,0,0]=(0.396185569444)*x[0]**o+(0.352712187886)*x[0]+(0.132847269784)*x[1]**o+(0.00561831278948)*x[1]
        s[2,2,0,1]=(-0.284087190739)*x[0]**o+(-0.307024290069)*x[0]+(-0.958481172381)*x[1]**o+(0.0918318779969)*x[1]
        s[2,2,1,0]=(0.0633474988001)*x[0]**o+(-0.650310080822)*x[0]+(0.576579254607)*x[1]**o+(0.674521364804)*x[1]
        s[2,2,1,1]=(0.579536883642)*x[0]**o+(0.160337559302)*x[0]+(-0.806829552191)*x[1]**o+(-0.45313287787)*x[1]
        s[2,2,2,0]=(-0.922597521221)*x[0]**o+(-0.250217857798)*x[0]+(-0.0762271301664)*x[1]**o+(0.610046304117)*x[1]
        s[2,2,2,1]=(0.259914145502)*x[0]**o+(0.00944250446491)*x[0]+(-0.217644702814)*x[1]**o+(-0.396061779281)*x[1]
        s[2,3,0,0]=(-0.639026927928)*x[0]**o+(-0.0136042147736)*x[0]+(0.147051301049)*x[1]**o+(-0.116721318722)*x[1]
        s[2,3,0,1]=(-0.921716795774)*x[0]**o+(-0.573953102898)*x[0]+(-0.292599705087)*x[1]**o+(0.425695557831)*x[1]
        s[2,3,1,0]=(-0.137488915425)*x[0]**o+(-0.49947594329)*x[0]+(-0.0290388465996)*x[1]**o+(-0.246424924833)*x[1]
        s[2,3,1,1]=(0.231049448089)*x[0]**o+(-0.268957451079)*x[0]+(0.900504385567)*x[1]**o+(-0.331066572068)*x[1]
        s[2,3,2,0]=(-0.442867405047)*x[0]**o+(-0.862895940039)*x[0]+(0.130486555031)*x[1]**o+(-0.325005280104)*x[1]
        s[2,3,2,1]=(0.0505322685584)*x[0]**o+(0.588209043808)*x[0]+(-0.85729410482)*x[1]**o+(0.20026415404)*x[1]
        s[2,4,0,0]=(0.111265750114)*x[0]**o+(-0.368157597131)*x[0]+(0.264359624294)*x[1]**o+(0.880890103316)*x[1]
        s[2,4,0,1]=(0.7435570617)*x[0]**o+(-0.794211194611)*x[0]+(0.506134754195)*x[1]**o+(0.363318494554)*x[1]
        s[2,4,1,0]=(0.557218657611)*x[0]**o+(-0.584592285851)*x[0]+(-0.551098106406)*x[1]**o+(-0.253662162858)*x[1]
        s[2,4,1,1]=(0.508828804121)*x[0]**o+(0.4399952689)*x[0]+(-0.997629557499)*x[1]**o+(0.651550324191)*x[1]
        s[2,4,2,0]=(-0.292367705037)*x[0]**o+(0.325506979353)*x[0]+(0.345654565875)*x[1]**o+(-0.0615253936178)*x[1]
        s[2,4,2,1]=(-0.408469289071)*x[0]**o+(-0.242826710026)*x[0]+(0.968562541409)*x[1]**o+(-0.691883337576)*x[1]
        s[3,0,0,0]=(-0.166166381364)*x[0]**o+(0.67558604652)*x[0]+(0.681136750192)*x[1]**o+(-0.59276757042)*x[1]
        s[3,0,0,1]=(0.139326266787)*x[0]**o+(-0.855321186969)*x[0]+(0.87031372876)*x[1]**o+(0.167702944821)*x[1]
        s[3,0,1,0]=(0.0382861750276)*x[0]**o+(-0.15200218165)*x[0]+(0.877979117505)*x[1]**o+(-0.0121347649803)*x[1]
        s[3,0,1,1]=(0.498616833176)*x[0]**o+(-0.663932730791)*x[0]+(-0.351696535062)*x[1]**o+(0.635672218271)*x[1]
        s[3,0,2,0]=(0.570569612815)*x[0]**o+(0.302777992546)*x[0]+(0.234111661233)*x[1]**o+(0.23736500566)*x[1]
        s[3,0,2,1]=(0.103264640389)*x[0]**o+(-0.665245672415)*x[0]+(-0.507585301735)*x[1]**o+(-0.624579920066)*x[1]
        s[3,1,0,0]=(-0.424026522629)*x[0]**o+(-0.289542833446)*x[0]+(-0.208440343753)*x[1]**o+(0.670301272477)*x[1]
        s[3,1,0,1]=(0.884688677823)*x[0]**o+(-0.547024055336)*x[0]+(-0.760788413661)*x[1]**o+(0.892759772046)*x[1]
        s[3,1,1,0]=(-0.193616699826)*x[0]**o+(-0.817184462119)*x[0]+(-0.405430250489)*x[1]**o+(0.697242407135)*x[1]
        s[3,1,1,1]=(-0.224055211394)*x[0]**o+(0.1446788843)*x[0]+(0.242791365909)*x[1]**o+(-0.00969799364721)*x[1]
        s[3,1,2,0]=(-0.86134785548)*x[0]**o+(0.808223751392)*x[0]+(-0.59203015099)*x[1]**o+(-0.774006816622)*x[1]
        s[3,1,2,1]=(0.60450330353)*x[0]**o+(0.47388266469)*x[0]+(0.160423399019)*x[1]**o+(0.177522838509)*x[1]
        s[3,2,0,0]=(-0.0704547054622)*x[0]**o+(-0.28005227128)*x[0]+(0.842667852448)*x[1]**o+(-0.660858857912)*x[1]
        s[3,2,0,1]=(0.950935693626)*x[0]**o+(-0.707135492206)*x[0]+(0.383664604588)*x[1]**o+(0.301522608591)*x[1]
        s[3,2,1,0]=(-0.843657734623)*x[0]**o+(0.0313079420866)*x[0]+(0.842733149317)*x[1]**o+(0.8363224382)*x[1]
        s[3,2,1,1]=(0.590344608942)*x[0]**o+(-0.688010374961)*x[0]+(0.419847151895)*x[1]**o+(0.703183045156)*x[1]
        s[3,2,2,0]=(0.726161847781)*x[0]**o+(-0.644844589861)*x[0]+(0.615915050884)*x[1]**o+(0.928915762636)*x[1]
        s[3,2,2,1]=(-0.272507884368)*x[0]**o+(0.915695022479)*x[0]+(-0.461159868664)*x[1]**o+(-0.959034792496)*x[1]
        s[3,3,0,0]=(0.454712361388)*x[0]**o+(-0.212357162336)*x[0]+(0.153326792468)*x[1]**o+(-0.891670991055)*x[1]
        s[3,3,0,1]=(0.524401133703)*x[0]**o+(-0.649106057963)*x[0]+(0.879052771991)*x[1]**o+(-0.710445481523)*x[1]
        s[3,3,1,0]=(0.614330071709)*x[0]**o+(-0.632228392701)*x[0]+(0.272026382411)*x[1]**o+(-0.261384418355)*x[1]
        s[3,3,1,1]=(0.552126023508)*x[0]**o+(0.266617783195)*x[0]+(-0.183251756988)*x[1]**o+(0.905027547625)*x[1]
        s[3,3,2,0]=(-0.671825095536)*x[0]**o+(-0.0942290736973)*x[0]+(0.354473207574)*x[1]**o+(-0.61776928413)*x[1]
        s[3,3,2,1]=(0.8582228544)*x[0]**o+(-0.68683359365)*x[0]+(0.856354836335)*x[1]**o+(0.588764055551)*x[1]
        s[3,4,0,0]=(-0.535790362987)*x[0]**o+(0.835465266424)*x[0]+(-0.422274723942)*x[1]**o+(-0.979702488165)*x[1]
        s[3,4,0,1]=(-0.846846124079)*x[0]**o+(-0.52263128994)*x[0]+(0.0349528062482)*x[1]**o+(-0.498515393094)*x[1]
        s[3,4,1,0]=(-0.927241665975)*x[0]**o+(0.0353321330647)*x[0]+(-0.225562802107)*x[1]**o+(0.8207319568)*x[1]
        s[3,4,1,1]=(-0.174397417012)*x[0]**o+(-0.0398036522533)*x[0]+(0.0755396127937)*x[1]**o+(-0.102836810613)*x[1]
        s[3,4,2,0]=(-0.968779473612)*x[0]**o+(0.339954457114)*x[0]+(-0.613643260944)*x[1]**o+(0.301848346204)*x[1]
        s[3,4,2,1]=(0.994429790023)*x[0]**o+(0.159760881455)*x[0]+(0.252053920576)*x[1]**o+(-0.562820021013)*x[1]
        ref[0,0,0,0]=(0.976142604799)/(o+1.)+(0.559236926803)
        ref[0,0,0,1]=(-1.7145183922)/(o+1.)+(0.0635707789)
        ref[0,0,1,0]=(0.899826569143)/(o+1.)+(0.790504652192)
        ref[0,0,1,1]=(-1.61591306524)/(o+1.)+(0.285592875029)
        ref[0,0,2,0]=(1.82158548419)/(o+1.)+(-0.0646334283283)
        ref[0,0,2,1]=(0.918537074153)/(o+1.)+(-0.189308477993)
        ref[0,1,0,0]=(0.608669315438)/(o+1.)+(0.276027348804)
        ref[0,1,0,1]=(1.40725199507)/(o+1.)+(-0.0943899330873)
        ref[0,1,1,0]=(0.0512330860046)/(o+1.)+(-0.687181179599)
        ref[0,1,1,1]=(-0.878253317721)/(o+1.)+(-0.949562209289)
        ref[0,1,2,0]=(1.02379194818)/(o+1.)+(-0.0506538701325)
        ref[0,1,2,1]=(0.659769108985)/(o+1.)+(-0.807763678736)
        ref[0,2,0,0]=(-1.11486203512)/(o+1.)+(-0.00125502860562)
        ref[0,2,0,1]=(1.82559575585)/(o+1.)+(0.796853489163)
        ref[0,2,1,0]=(1.66037232606)/(o+1.)+(0.419884607511)
        ref[0,2,1,1]=(-1.03285158476)/(o+1.)+(-0.0989279216092)
        ref[0,2,2,0]=(0.307240682354)/(o+1.)+(0.400848881959)
        ref[0,2,2,1]=(1.27022227018)/(o+1.)+(-0.0482148791284)
        ref[0,3,0,0]=(-0.788939093322)/(o+1.)+(-0.631363596784)
        ref[0,3,0,1]=(0.26640356147)/(o+1.)+(-0.532062387218)
        ref[0,3,1,0]=(0.959198406218)/(o+1.)+(-0.566324965195)
        ref[0,3,1,1]=(0.0927952942109)/(o+1.)+(-0.289945801554)
        ref[0,3,2,0]=(-0.0579636581476)/(o+1.)+(-0.0159560424323)
        ref[0,3,2,1]=(-0.254597349139)/(o+1.)+(-0.0361368341719)
        ref[0,4,0,0]=(-0.00994556542921)/(o+1.)+(0.894775230178)
        ref[0,4,0,1]=(0.53626175681)/(o+1.)+(-0.269667599686)
        ref[0,4,1,0]=(0.0172064338404)/(o+1.)+(0.0110821611268)
        ref[0,4,1,1]=(-1.41043453748)/(o+1.)+(0.295320921755)
        ref[0,4,2,0]=(-0.319272527632)/(o+1.)+(0.251472854574)
        ref[0,4,2,1]=(-0.341099001658)/(o+1.)+(-0.134149823407)
        ref[1,0,0,0]=(1.36852782455)/(o+1.)+(-0.660630850204)
        ref[1,0,0,1]=(1.17203200118)/(o+1.)+(-0.372038010914)
        ref[1,0,1,0]=(0.906181287285)/(o+1.)+(-0.430976599456)
        ref[1,0,1,1]=(-1.37920903098)/(o+1.)+(0.202077810714)
        ref[1,0,2,0]=(1.82362651408)/(o+1.)+(-0.395452886013)
        ref[1,0,2,1]=(0.656448039332)/(o+1.)+(-0.586488805114)
        ref[1,1,0,0]=(-0.506275307541)/(o+1.)+(-0.613889034788)
        ref[1,1,0,1]=(0.0970528225071)/(o+1.)+(-0.357242175157)
        ref[1,1,1,0]=(-0.609663973467)/(o+1.)+(-0.0527203717721)
        ref[1,1,1,1]=(0.417273088561)/(o+1.)+(-0.21259448119)
        ref[1,1,2,0]=(-1.76474368781)/(o+1.)+(-0.0561389978065)
        ref[1,1,2,1]=(0.733396922329)/(o+1.)+(-0.360546207158)
        ref[1,2,0,0]=(-1.0930406353)/(o+1.)+(-0.882921780793)
        ref[1,2,0,1]=(0.470310563612)/(o+1.)+(-0.17410375492)
        ref[1,2,1,0]=(-0.711974684398)/(o+1.)+(-0.10588070715)
        ref[1,2,1,1]=(-0.882848135361)/(o+1.)+(0.0145919323434)
        ref[1,2,2,0]=(0.691426968189)/(o+1.)+(-0.538670436911)
        ref[1,2,2,1]=(0.181002825221)/(o+1.)+(-0.438446228014)
        ref[1,3,0,0]=(0.469597415414)/(o+1.)+(-0.235126278249)
        ref[1,3,0,1]=(-0.929510611384)/(o+1.)+(-0.413988367578)
        ref[1,3,1,0]=(0.854514676258)/(o+1.)+(-0.0353472783996)
        ref[1,3,1,1]=(0.630474668409)/(o+1.)+(0.0744386275773)
        ref[1,3,2,0]=(0.390924363485)/(o+1.)+(0.627689530971)
        ref[1,3,2,1]=(-1.6409097599)/(o+1.)+(-0.452381472906)
        ref[1,4,0,0]=(0.357069049807)/(o+1.)+(0.686521034986)
        ref[1,4,0,1]=(0.278480397348)/(o+1.)+(-0.328928650294)
        ref[1,4,1,0]=(0.0199868564018)/(o+1.)+(0.0874366821718)
        ref[1,4,1,1]=(-0.210918168065)/(o+1.)+(0.338114608888)
        ref[1,4,2,0]=(0.691363338179)/(o+1.)+(0.608378114821)
        ref[1,4,2,1]=(1.21182478236)/(o+1.)+(0.0921588457958)
        ref[2,0,0,0]=(-0.47411466559)/(o+1.)+(0.0547177057268)
        ref[2,0,0,1]=(-0.220914076875)/(o+1.)+(0.0473247023129)
        ref[2,0,1,0]=(0.892263621045)/(o+1.)+(0.0311394070465)
        ref[2,0,1,1]=(1.14518124466)/(o+1.)+(0.520818118726)
        ref[2,0,2,0]=(-0.00320343813922)/(o+1.)+(0.0990387149137)
        ref[2,0,2,1]=(-0.454357902363)/(o+1.)+(-0.0671338342638)
        ref[2,1,0,0]=(0.394055509731)/(o+1.)+(0.924768577529)
        ref[2,1,0,1]=(0.500866633191)/(o+1.)+(0.212233622092)
        ref[2,1,1,0]=(1.29402054756)/(o+1.)+(0.375514450965)
        ref[2,1,1,1]=(0.369127441961)/(o+1.)+(-0.434822092262)
        ref[2,1,2,0]=(-0.251462312861)/(o+1.)+(-0.285897304624)
        ref[2,1,2,1]=(1.45788307837)/(o+1.)+(0.42312280598)
        ref[2,2,0,0]=(0.529032839228)/(o+1.)+(0.179165250338)
        ref[2,2,0,1]=(-1.24256836312)/(o+1.)+(-0.107596206036)
        ref[2,2,1,0]=(0.639926753407)/(o+1.)+(0.0121056419913)
        ref[2,2,1,1]=(-0.227292668549)/(o+1.)+(-0.146397659284)
        ref[2,2,2,0]=(-0.998824651388)/(o+1.)+(0.179914223159)
        ref[2,2,2,1]=(0.0422694426877)/(o+1.)+(-0.193309637408)
        ref[2,3,0,0]=(-0.491975626879)/(o+1.)+(-0.0651627667478)
        ref[2,3,0,1]=(-1.21431650086)/(o+1.)+(-0.0741287725332)
        ref[2,3,1,0]=(-0.166527762025)/(o+1.)+(-0.372950434061)
        ref[2,3,1,1]=(1.13155383366)/(o+1.)+(-0.300012011573)
        ref[2,3,2,0]=(-0.312380850015)/(o+1.)+(-0.593950610072)
        ref[2,3,2,1]=(-0.806761836262)/(o+1.)+(0.394236598924)
        ref[2,4,0,0]=(0.375625374408)/(o+1.)+(0.256366253093)
        ref[2,4,0,1]=(1.2496918159)/(o+1.)+(-0.215446350028)
        ref[2,4,1,0]=(0.00612055120519)/(o+1.)+(-0.419127224354)
        ref[2,4,1,1]=(-0.488800753378)/(o+1.)+(0.545772796545)
        ref[2,4,2,0]=(0.0532868608376)/(o+1.)+(0.131990792868)
        ref[2,4,2,1]=(0.560093252338)/(o+1.)+(-0.467355023801)
        ref[3,0,0,0]=(0.514970368828)/(o+1.)+(0.0414092380501)
        ref[3,0,0,1]=(1.00963999555)/(o+1.)+(-0.343809121074)
        ref[3,0,1,0]=(0.916265292533)/(o+1.)+(-0.0820684733153)
        ref[3,0,1,1]=(0.146920298114)/(o+1.)+(-0.0141302562603)
        ref[3,0,2,0]=(0.804681274048)/(o+1.)+(0.270071499103)
        ref[3,0,2,1]=(-0.404320661346)/(o+1.)+(-0.644912796241)
        ref[3,1,0,0]=(-0.632466866383)/(o+1.)+(0.190379219516)
        ref[3,1,0,1]=(0.123900264162)/(o+1.)+(0.172867858355)
        ref[3,1,1,0]=(-0.599046950315)/(o+1.)+(-0.0599710274916)
        ref[3,1,1,1]=(0.0187361545152)/(o+1.)+(0.0674904453263)
        ref[3,1,2,0]=(-1.45337800647)/(o+1.)+(0.0171084673849)
        ref[3,1,2,1]=(0.764926702549)/(o+1.)+(0.325702751599)
        ref[3,2,0,0]=(0.772213146985)/(o+1.)+(-0.470455564596)
        ref[3,2,0,1]=(1.33460029821)/(o+1.)+(-0.202806441808)
        ref[3,2,1,0]=(-0.00092458530583)/(o+1.)+(0.433815190143)
        ref[3,2,1,1]=(1.01019176084)/(o+1.)+(0.00758633509776)
        ref[3,2,2,0]=(1.34207689867)/(o+1.)+(0.142035586387)
        ref[3,2,2,1]=(-0.733667753032)/(o+1.)+(-0.0216698850086)
        ref[3,3,0,0]=(0.608039153856)/(o+1.)+(-0.552014076696)
        ref[3,3,0,1]=(1.40345390569)/(o+1.)+(-0.679775769743)
        ref[3,3,1,0]=(0.88635645412)/(o+1.)+(-0.446806405528)
        ref[3,3,1,1]=(0.368874266519)/(o+1.)+(0.58582266541)
        ref[3,3,2,0]=(-0.317351887963)/(o+1.)+(-0.355999178914)
        ref[3,3,2,1]=(1.71457769073)/(o+1.)+(-0.0490347690493)
        ref[3,4,0,0]=(-0.958065086929)/(o+1.)+(-0.0721186108707)
        ref[3,4,0,1]=(-0.811893317831)/(o+1.)+(-0.510573341517)
        ref[3,4,1,0]=(-1.15280446808)/(o+1.)+(0.428032044932)
        ref[3,4,1,1]=(-0.0988578042187)/(o+1.)+(-0.0713202314331)
        ref[3,4,2,0]=(-1.58242273456)/(o+1.)+(0.320901401659)
        ref[3,4,2,1]=(1.2464837106)/(o+1.)+(-0.201529569779)
      else:
        s[0,0,0,0]=(0.541504168376)*x[0]**o+(0.0573239599909)*x[0]+(0.0647250584452)*x[1]**o+(-0.703570101211)*x[1]+(0.328542610633)*x[2]**o+(-0.223400684365)*x[2]
        s[0,0,0,1]=(0.815893107692)*x[0]**o+(0.865481882711)*x[0]+(-0.46283176119)*x[1]**o+(-0.180772911655)*x[1]+(0.93497108116)*x[2]**o+(-0.742753139638)*x[2]
        s[0,0,1,0]=(-0.974209081736)*x[0]**o+(-0.672767028246)*x[0]+(-0.741240928071)*x[1]**o+(0.33719490832)*x[1]+(-0.686997621399)*x[2]**o+(-0.276966804283)*x[2]
        s[0,0,1,1]=(-0.857143471623)*x[0]**o+(-0.808079211925)*x[0]+(-0.953814060073)*x[1]**o+(0.961661670434)*x[1]+(0.189072084458)*x[2]**o+(-0.103208629753)*x[2]
        s[0,0,2,0]=(-0.20991399963)*x[0]**o+(-0.442037964882)*x[0]+(0.322578941372)*x[1]**o+(-0.161870387325)*x[1]+(0.585691936782)*x[2]**o+(-0.267825636094)*x[2]
        s[0,0,2,1]=(0.567642417211)*x[0]**o+(0.824356280251)*x[0]+(0.414108305262)*x[1]**o+(0.79515753838)*x[1]+(-0.894154134914)*x[2]**o+(-0.307671452735)*x[2]
        s[0,1,0,0]=(0.536896074085)*x[0]**o+(0.361562637795)*x[0]+(-0.158827271642)*x[1]**o+(0.0705880551767)*x[1]+(-0.332232895988)*x[2]**o+(-0.0305180398365)*x[2]
        s[0,1,0,1]=(0.633176778138)*x[0]**o+(-0.572376784007)*x[0]+(-0.294505207444)*x[1]**o+(-0.378967371632)*x[1]+(-0.89662018673)*x[2]**o+(0.287814422967)*x[2]
        s[0,1,1,0]=(-0.61097365344)*x[0]**o+(-0.363660280252)*x[0]+(0.190746604006)*x[1]**o+(-0.510926077271)*x[1]+(-0.957696939615)*x[2]**o+(-0.789299643561)*x[2]
        s[0,1,1,1]=(-0.835396537806)*x[0]**o+(-0.836746663696)*x[0]+(0.702723030555)*x[1]**o+(0.548005714327)*x[1]+(-0.924774335596)*x[2]**o+(-0.439240327617)*x[2]
        s[0,1,2,0]=(0.461024657935)*x[0]**o+(-0.373409962194)*x[0]+(0.743442026381)*x[1]**o+(0.657544000829)*x[1]+(0.674057315581)*x[2]**o+(-0.0183044871772)*x[2]
        s[0,1,2,1]=(0.164119453764)*x[0]**o+(-0.31204382122)*x[0]+(-0.424802851747)*x[1]**o+(-0.341469957442)*x[1]+(-0.231054265963)*x[2]**o+(0.128985361513)*x[2]
        s[0,2,0,0]=(0.963584200519)*x[0]**o+(-0.465275784807)*x[0]+(0.421929871546)*x[1]**o+(-0.211860414792)*x[1]+(-0.403745447737)*x[2]**o+(-0.379173494789)*x[2]
        s[0,2,0,1]=(-0.849597493426)*x[0]**o+(-0.174912181728)*x[0]+(-0.0534228854624)*x[1]**o+(0.826630462481)*x[1]+(-0.0210209049463)*x[2]**o+(-0.675409496795)*x[2]
        s[0,2,1,0]=(-0.565229688917)*x[0]**o+(0.485718751072)*x[0]+(0.700861807655)*x[1]**o+(0.901169377907)*x[1]+(0.890394572646)*x[2]**o+(0.596255846932)*x[2]
        s[0,2,1,1]=(0.876434872654)*x[0]**o+(-0.83625911596)*x[0]+(-0.539508005116)*x[1]**o+(-0.790017898115)*x[1]+(-0.043576285617)*x[2]**o+(-0.90281792431)*x[2]
        s[0,2,2,0]=(0.688686103492)*x[0]**o+(-0.479018765365)*x[0]+(-0.0041922877839)*x[1]**o+(0.360321450098)*x[1]+(0.520980145515)*x[2]**o+(0.781262738907)*x[2]
        s[0,2,2,1]=(-0.416899995117)*x[0]**o+(-0.018267124332)*x[0]+(-0.052482515113)*x[1]**o+(0.6928630854)*x[1]+(-0.163882932309)*x[2]**o+(0.478296669995)*x[2]
        s[0,3,0,0]=(-0.80317047721)*x[0]**o+(-0.475554188062)*x[0]+(0.194711479429)*x[1]**o+(0.866725258981)*x[1]+(0.500781673148)*x[2]**o+(-0.360414664576)*x[2]
        s[0,3,0,1]=(0.243316356193)*x[0]**o+(-0.109598635489)*x[0]+(0.579141755165)*x[1]**o+(0.876481264859)*x[1]+(-0.880903923571)*x[2]**o+(0.142946599744)*x[2]
        s[0,3,1,0]=(-0.507217017458)*x[0]**o+(0.641911252335)*x[0]+(-0.764452998266)*x[1]**o+(-0.505598974991)*x[1]+(0.641213955525)*x[2]**o+(-0.44758893679)*x[2]
        s[0,3,1,1]=(0.473545973781)*x[0]**o+(-0.00833748934844)*x[0]+(-0.00699017656361)*x[1]**o+(-0.643164722085)*x[1]+(0.726844277071)*x[2]**o+(-0.559279257114)*x[2]
        s[0,3,2,0]=(0.558099790948)*x[0]**o+(0.808990368539)*x[0]+(0.0431682029749)*x[1]**o+(0.0173216046894)*x[1]+(-0.394219668262)*x[2]**o+(-0.0336030329937)*x[2]
        s[0,3,2,1]=(-0.0222232210481)*x[0]**o+(-0.510190960052)*x[0]+(0.569389223293)*x[1]**o+(0.117046338884)*x[1]+(-0.671864592568)*x[2]**o+(-0.674239142474)*x[2]
        s[0,4,0,0]=(-0.644798822323)*x[0]**o+(-0.342116629892)*x[0]+(-0.168244417587)*x[1]**o+(-0.108919120953)*x[1]+(0.804101464972)*x[2]**o+(-0.0927979618696)*x[2]
        s[0,4,0,1]=(0.195838618741)*x[0]**o+(-0.111914225664)*x[0]+(-0.754003728212)*x[1]**o+(-0.096873749243)*x[1]+(-0.770744141788)*x[2]**o+(0.0291059262795)*x[2]
        s[0,4,1,0]=(0.321738393515)*x[0]**o+(-0.262140344086)*x[0]+(-0.130813570922)*x[1]**o+(0.749305862624)*x[1]+(-0.0933499641013)*x[2]**o+(0.583868661476)*x[2]
        s[0,4,1,1]=(-0.948399949958)*x[0]**o+(-0.514552416108)*x[0]+(0.403700494592)*x[1]**o+(0.319753801921)*x[1]+(0.068451176553)*x[2]**o+(0.236776599218)*x[2]
        s[0,4,2,0]=(-0.206993609035)*x[0]**o+(-0.693436356612)*x[0]+(0.376741069465)*x[1]**o+(-0.377934787271)*x[1]+(0.600888347529)*x[2]**o+(-0.806497559367)*x[2]
        s[0,4,2,1]=(0.519911073798)*x[0]**o+(0.209533160142)*x[0]+(-0.828631380624)*x[1]**o+(-0.758712343322)*x[1]+(0.469217244822)*x[2]**o+(0.777639952133)*x[2]
        s[1,0,0,0]=(-0.00938493494863)*x[0]**o+(0.035357976182)*x[0]+(0.530712339191)*x[1]**o+(0.163841019664)*x[1]+(-0.481052524963)*x[2]**o+(0.165894455618)*x[2]
        s[1,0,0,1]=(0.405172592393)*x[0]**o+(0.748490170154)*x[0]+(-0.579541033967)*x[1]**o+(-0.612353189754)*x[1]+(-0.0878018425036)*x[2]**o+(0.909881661391)*x[2]
        s[1,0,1,0]=(-0.131136361267)*x[0]**o+(0.0576430483263)*x[0]+(0.21369263102)*x[1]**o+(0.00660590703342)*x[1]+(0.079311350103)*x[2]**o+(0.878257991876)*x[2]
        s[1,0,1,1]=(-0.585221206397)*x[0]**o+(-0.791686741073)*x[0]+(0.631923355264)*x[1]**o+(-0.0277720531745)*x[1]+(-0.0777198468194)*x[2]**o+(-0.407162573619)*x[2]
        s[1,0,2,0]=(0.544255664465)*x[0]**o+(-0.51929274141)*x[0]+(-0.614260874796)*x[1]**o+(-0.951824251618)*x[1]+(0.48400676445)*x[2]**o+(-0.210092738748)*x[2]
        s[1,0,2,1]=(-0.901122575086)*x[0]**o+(-0.855042125739)*x[0]+(-0.660979212714)*x[1]**o+(0.874395337943)*x[1]+(-0.286202842553)*x[2]**o+(0.981270297086)*x[2]
        s[1,1,0,0]=(-0.752469757419)*x[0]**o+(0.0583950789171)*x[0]+(-0.699699086971)*x[1]**o+(-0.461650511393)*x[1]+(-0.272665950769)*x[2]**o+(-0.873526828849)*x[2]
        s[1,1,0,1]=(0.219706522447)*x[0]**o+(-0.73250655794)*x[0]+(0.0679566752122)*x[1]**o+(0.731592986993)*x[1]+(0.557341326754)*x[2]**o+(0.256780037036)*x[2]
        s[1,1,1,0]=(-0.820468252627)*x[0]**o+(0.163409736336)*x[0]+(-0.593882214431)*x[1]**o+(0.778760445642)*x[1]+(-0.822051914962)*x[2]**o+(0.920172785813)*x[2]
        s[1,1,1,1]=(-0.355904745872)*x[0]**o+(-0.192775924552)*x[0]+(0.598345031205)*x[1]**o+(-0.108244476269)*x[1]+(-0.895456581435)*x[2]**o+(0.207881298634)*x[2]
        s[1,1,2,0]=(0.781059740818)*x[0]**o+(0.186409854519)*x[0]+(0.92458071805)*x[1]**o+(0.197380801637)*x[1]+(-0.402971455411)*x[2]**o+(0.42922557636)*x[2]
        s[1,1,2,1]=(0.0064822572955)*x[0]**o+(0.846443443074)*x[0]+(0.500688303847)*x[1]**o+(-0.884507156164)*x[1]+(0.433224067545)*x[2]**o+(-0.0383476145137)*x[2]
        s[1,2,0,0]=(-0.945907369747)*x[0]**o+(-0.149404885342)*x[0]+(-0.911317716302)*x[1]**o+(0.718244019047)*x[1]+(0.097223049045)*x[2]**o+(0.95630310286)*x[2]
        s[1,2,0,1]=(-0.942956848602)*x[0]**o+(0.24519384765)*x[0]+(0.450231851653)*x[1]**o+(0.875902538898)*x[1]+(0.552657508084)*x[2]**o+(0.644477914693)*x[2]
        s[1,2,1,0]=(0.600235706495)*x[0]**o+(-0.142524845444)*x[0]+(0.216130207621)*x[1]**o+(-0.396701569264)*x[1]+(0.316884903524)*x[2]**o+(-0.0586626235465)*x[2]
        s[1,2,1,1]=(0.329111957957)*x[0]**o+(-0.508751506135)*x[0]+(0.911087373052)*x[1]**o+(-0.0983342912554)*x[1]+(-0.842070723886)*x[2]**o+(0.0746739548918)*x[2]
        s[1,2,2,0]=(0.387988330934)*x[0]**o+(-0.8408548375)*x[0]+(0.838833668833)*x[1]**o+(0.965224043099)*x[1]+(-0.860139996894)*x[2]**o+(-0.925778998652)*x[2]
        s[1,2,2,1]=(0.238958121236)*x[0]**o+(0.10607561646)*x[0]+(0.302265720665)*x[1]**o+(0.145320472959)*x[1]+(-0.677724287444)*x[2]**o+(-0.666145512045)*x[2]
        s[1,3,0,0]=(0.360431863997)*x[0]**o+(0.509815440438)*x[0]+(-0.178896172032)*x[1]**o+(0.316774928052)*x[1]+(-0.44743491823)*x[2]**o+(-0.651310611493)*x[2]
        s[1,3,0,1]=(-0.65053757716)*x[0]**o+(-0.375438056145)*x[0]+(0.360834001552)*x[1]**o+(-0.250086720496)*x[1]+(-0.71501469699)*x[2]**o+(0.0150228087513)*x[2]
        s[1,3,1,0]=(0.0628062775266)*x[0]**o+(-0.61619322157)*x[0]+(0.45067341985)*x[1]**o+(0.355471837743)*x[1]+(-0.174333549061)*x[2]**o+(-0.942897180184)*x[2]
        s[1,3,1,1]=(0.629724537179)*x[0]**o+(-0.0558311952601)*x[0]+(-0.956862453932)*x[1]**o+(-0.552443418291)*x[1]+(0.146303452118)*x[2]**o+(-0.770969067419)*x[2]
        s[1,3,2,0]=(-0.178686519235)*x[0]**o+(-0.171143660316)*x[0]+(-0.162449341623)*x[1]**o+(-0.1105712214)*x[1]+(0.119821653614)*x[2]**o+(0.209872978827)*x[2]
        s[1,3,2,1]=(0.312724690007)*x[0]**o+(-0.692554198329)*x[0]+(0.502590520541)*x[1]**o+(-0.729216407269)*x[1]+(0.302595866989)*x[2]**o+(-0.815710825541)*x[2]
        s[1,4,0,0]=(0.112006695888)*x[0]**o+(0.829399837819)*x[0]+(-0.419232858393)*x[1]**o+(0.146007406445)*x[1]+(0.304188744047)*x[2]**o+(0.46242131026)*x[2]
        s[1,4,0,1]=(0.964310426412)*x[0]**o+(0.196629594947)*x[0]+(0.124667636162)*x[1]**o+(-0.304792580501)*x[1]+(0.180285405611)*x[2]**o+(-0.543512746454)*x[2]
        s[1,4,1,0]=(-0.708536876552)*x[0]**o+(-0.680163366548)*x[0]+(0.0584969206489)*x[1]**o+(0.847180583253)*x[1]+(0.157570405709)*x[2]**o+(-0.746490964212)*x[2]
        s[1,4,1,1]=(0.0845668429821)*x[0]**o+(-0.889880680014)*x[0]+(-0.0771948903819)*x[1]**o+(-0.533952084212)*x[1]+(0.346150298712)*x[2]**o+(-0.78816093534)*x[2]
        s[1,4,2,0]=(0.375822387867)*x[0]**o+(-0.357919731843)*x[0]+(-0.373519792105)*x[1]**o+(0.641319045383)*x[1]+(0.351324745466)*x[2]**o+(-0.186703738887)*x[2]
        s[1,4,2,1]=(-0.60888359369)*x[0]**o+(0.694522376656)*x[0]+(-0.169823125514)*x[1]**o+(-0.856247555905)*x[1]+(0.00519595740385)*x[2]**o+(-0.875798589653)*x[2]
        s[2,0,0,0]=(0.469380548227)*x[0]**o+(-0.868913385859)*x[0]+(0.441755799104)*x[1]**o+(-0.493212921066)*x[1]+(-0.11743848736)*x[2]**o+(0.21526923935)*x[2]
        s[2,0,0,1]=(-0.151212323369)*x[0]**o+(-0.999958063251)*x[0]+(0.449187517422)*x[1]**o+(-0.583657187784)*x[1]+(-0.905753638698)*x[2]**o+(-0.945179993617)*x[2]
        s[2,0,1,0]=(-0.40460984789)*x[0]**o+(0.510786603887)*x[0]+(-0.953825153871)*x[1]**o+(0.226098717119)*x[1]+(-0.799990467305)*x[2]**o+(-0.429211164318)*x[2]
        s[2,0,1,1]=(0.867644883712)*x[0]**o+(0.714940923764)*x[0]+(-0.122362538438)*x[1]**o+(-0.44142264498)*x[1]+(0.437303901854)*x[2]**o+(0.841319415724)*x[2]
        s[2,0,2,0]=(0.13789233553)*x[0]**o+(0.472191869459)*x[0]+(0.468664678229)*x[1]**o+(-0.584558488838)*x[1]+(0.183146176041)*x[2]**o+(-0.893499132079)*x[2]
        s[2,0,2,1]=(0.875084187567)*x[0]**o+(0.550980479395)*x[0]+(0.312182122865)*x[1]**o+(-0.764545704288)*x[1]+(0.421014399576)*x[2]**o+(-0.652582305594)*x[2]
        s[2,1,0,0]=(-0.397160997842)*x[0]**o+(-0.401717267945)*x[0]+(0.612062897035)*x[1]**o+(0.147514233621)*x[1]+(0.994896686982)*x[2]**o+(-0.650669842592)*x[2]
        s[2,1,0,1]=(0.0859312342607)*x[0]**o+(-0.425173641674)*x[0]+(0.748616439433)*x[1]**o+(0.764026759608)*x[1]+(-0.990504424444)*x[2]**o+(-0.399709104793)*x[2]
        s[2,1,1,0]=(0.732840180154)*x[0]**o+(0.970502200871)*x[0]+(-0.413467445613)*x[1]**o+(-0.246800235561)*x[1]+(-0.881207449946)*x[2]**o+(0.882360789659)*x[2]
        s[2,1,1,1]=(0.689150562511)*x[0]**o+(-0.0933156400969)*x[0]+(-0.0951449636495)*x[1]**o+(0.143121640121)*x[1]+(-0.554471425755)*x[2]**o+(0.683081191337)*x[2]
        s[2,1,2,0]=(0.829863223837)*x[0]**o+(0.216508858271)*x[0]+(0.99016446302)*x[1]**o+(0.319444245585)*x[1]+(0.668648650551)*x[2]**o+(-0.896787700447)*x[2]
        s[2,1,2,1]=(-0.152966345983)*x[0]**o+(-0.775534953656)*x[0]+(0.0774085937181)*x[1]**o+(-0.59686756092)*x[1]+(0.224888875286)*x[2]**o+(-0.770576865174)*x[2]
        s[2,2,0,0]=(-0.831308688707)*x[0]**o+(0.0286055654454)*x[0]+(0.555782586294)*x[1]**o+(-0.863596073174)*x[1]+(0.0660913483948)*x[2]**o+(-0.691221916393)*x[2]
        s[2,2,0,1]=(0.97986500415)*x[0]**o+(-0.687991179162)*x[0]+(0.835947454347)*x[1]**o+(-0.425024918711)*x[1]+(0.632327959013)*x[2]**o+(-0.929603792153)*x[2]
        s[2,2,1,0]=(-0.785177923154)*x[0]**o+(0.498524809242)*x[0]+(0.657015457864)*x[1]**o+(-0.14042512075)*x[1]+(-0.523655529492)*x[2]**o+(0.0676392379718)*x[2]
        s[2,2,1,1]=(-0.282279979706)*x[0]**o+(0.402558983226)*x[0]+(0.349297772011)*x[1]**o+(-0.274102896576)*x[1]+(-0.976904085905)*x[2]**o+(-0.862615408506)*x[2]
        s[2,2,2,0]=(-0.442415692301)*x[0]**o+(-0.490020432017)*x[0]+(0.423330456256)*x[1]**o+(0.0308403124141)*x[1]+(-0.59889279992)*x[2]**o+(-0.784624469809)*x[2]
        s[2,2,2,1]=(-0.0571094173044)*x[0]**o+(-0.685673967597)*x[0]+(0.884019972654)*x[1]**o+(0.076633908669)*x[1]+(-0.436603596414)*x[2]**o+(0.643765166823)*x[2]
        s[2,3,0,0]=(-0.834122691134)*x[0]**o+(-0.316705170482)*x[0]+(0.0232732484874)*x[1]**o+(0.254076347383)*x[1]+(0.510661944881)*x[2]**o+(0.638231760578)*x[2]
        s[2,3,0,1]=(0.130809328027)*x[0]**o+(0.0475045505152)*x[0]+(0.253435146439)*x[1]**o+(0.15750814431)*x[1]+(0.930835277675)*x[2]**o+(0.215956216657)*x[2]
        s[2,3,1,0]=(-0.931893183482)*x[0]**o+(-0.165642362248)*x[0]+(-0.7975167088)*x[1]**o+(0.943743900193)*x[1]+(-0.834729159735)*x[2]**o+(0.310938564524)*x[2]
        s[2,3,1,1]=(-0.952550061226)*x[0]**o+(0.824399352189)*x[0]+(0.931907966548)*x[1]**o+(-0.46156960063)*x[1]+(0.6225788167)*x[2]**o+(-0.989192673067)*x[2]
        s[2,3,2,0]=(0.756326608673)*x[0]**o+(0.771528948716)*x[0]+(0.239160366604)*x[1]**o+(-0.452631969831)*x[1]+(-0.237358134546)*x[2]**o+(-0.839427184027)*x[2]
        s[2,3,2,1]=(0.0238504992639)*x[0]**o+(-0.103939311831)*x[0]+(0.291372616066)*x[1]**o+(0.672484596501)*x[1]+(0.494387122319)*x[2]**o+(0.486078041674)*x[2]
        s[2,4,0,0]=(-0.856552973776)*x[0]**o+(0.220667747539)*x[0]+(-0.901887042412)*x[1]**o+(0.162974138385)*x[1]+(0.884973432866)*x[2]**o+(-0.843375672742)*x[2]
        s[2,4,0,1]=(-0.581001018196)*x[0]**o+(0.766670058764)*x[0]+(0.195299626813)*x[1]**o+(0.491689692116)*x[1]+(-0.0851700136621)*x[2]**o+(0.556688214561)*x[2]
        s[2,4,1,0]=(0.341242000274)*x[0]**o+(0.254872207602)*x[0]+(0.721651164807)*x[1]**o+(0.817267858784)*x[1]+(0.525309787334)*x[2]**o+(-0.189345432506)*x[2]
        s[2,4,1,1]=(-0.461247890207)*x[0]**o+(-0.295916833174)*x[0]+(-0.795114931826)*x[1]**o+(0.760634909709)*x[1]+(0.628604868176)*x[2]**o+(-0.91582350733)*x[2]
        s[2,4,2,0]=(0.388037825656)*x[0]**o+(0.667391082754)*x[0]+(-0.76856572381)*x[1]**o+(-0.684287348308)*x[1]+(0.0434449534518)*x[2]**o+(-0.627587380339)*x[2]
        s[2,4,2,1]=(0.493746914184)*x[0]**o+(-0.47088277759)*x[0]+(-0.64457015181)*x[1]**o+(0.710566646648)*x[1]+(-0.603534864394)*x[2]**o+(0.952591999625)*x[2]
        s[3,0,0,0]=(-0.964941424537)*x[0]**o+(-0.701861794912)*x[0]+(0.631389613785)*x[1]**o+(0.216075041942)*x[1]+(-0.325257371696)*x[2]**o+(-0.291703451654)*x[2]
        s[3,0,0,1]=(-0.96299134603)*x[0]**o+(0.625796224057)*x[0]+(-0.418463355374)*x[1]**o+(-0.256336497204)*x[1]+(0.712728188959)*x[2]**o+(-0.770370932567)*x[2]
        s[3,0,1,0]=(-0.292515423665)*x[0]**o+(0.686334822809)*x[0]+(0.0305376943649)*x[1]**o+(0.872668037223)*x[1]+(-0.0541424770809)*x[2]**o+(-0.572416266221)*x[2]
        s[3,0,1,1]=(-0.690323374741)*x[0]**o+(-0.671336462076)*x[0]+(-0.166196941189)*x[1]**o+(-0.384967652901)*x[1]+(-0.61068991009)*x[2]**o+(0.776653422289)*x[2]
        s[3,0,2,0]=(-0.944428014582)*x[0]**o+(-0.301748262202)*x[0]+(-0.686840009057)*x[1]**o+(0.712060295627)*x[1]+(-0.254326623988)*x[2]**o+(0.400337107846)*x[2]
        s[3,0,2,1]=(-0.290963532129)*x[0]**o+(0.0269234770835)*x[0]+(-0.950364000347)*x[1]**o+(-0.460706646264)*x[1]+(-0.198333139868)*x[2]**o+(0.629854716945)*x[2]
        s[3,1,0,0]=(-0.555184083664)*x[0]**o+(0.131468013243)*x[0]+(0.282129000779)*x[1]**o+(-0.198577931709)*x[1]+(-0.175106810125)*x[2]**o+(0.00542537950498)*x[2]
        s[3,1,0,1]=(-0.284255526343)*x[0]**o+(-0.841249154497)*x[0]+(0.0213305454566)*x[1]**o+(-0.357636696086)*x[1]+(0.20888768221)*x[2]**o+(-0.503785042238)*x[2]
        s[3,1,1,0]=(-0.330999455854)*x[0]**o+(-0.159160032544)*x[0]+(0.52907127808)*x[1]**o+(0.0550513576484)*x[1]+(0.511074173459)*x[2]**o+(0.0159456796202)*x[2]
        s[3,1,1,1]=(-0.79143588969)*x[0]**o+(0.971248239785)*x[0]+(-0.0345752805688)*x[1]**o+(-0.0153770141636)*x[1]+(-0.716027682667)*x[2]**o+(-0.524455141515)*x[2]
        s[3,1,2,0]=(-0.582525406979)*x[0]**o+(0.397353443964)*x[0]+(-0.519936987381)*x[1]**o+(-0.21693645368)*x[1]+(0.869686334938)*x[2]**o+(-0.151814599828)*x[2]
        s[3,1,2,1]=(0.868456250139)*x[0]**o+(-0.568324799548)*x[0]+(-0.00346073709208)*x[1]**o+(-0.943912775939)*x[1]+(0.561497263651)*x[2]**o+(-0.923788209036)*x[2]
        s[3,2,0,0]=(-0.797982975817)*x[0]**o+(0.947494036065)*x[0]+(-0.250051189868)*x[1]**o+(0.472713798841)*x[1]+(0.902863101354)*x[2]**o+(0.613145652065)*x[2]
        s[3,2,0,1]=(0.252252158112)*x[0]**o+(0.301818562132)*x[0]+(0.664012773607)*x[1]**o+(0.0296698446724)*x[1]+(-0.00729763735658)*x[2]**o+(0.557599604788)*x[2]
        s[3,2,1,0]=(0.0570129540722)*x[0]**o+(-0.453241439283)*x[0]+(-0.71043548568)*x[1]**o+(-0.138422568238)*x[1]+(-0.611853026036)*x[2]**o+(0.253347515085)*x[2]
        s[3,2,1,1]=(-0.66828746218)*x[0]**o+(-0.751232642462)*x[0]+(-0.390775113865)*x[1]**o+(0.995150799921)*x[1]+(-0.292932619723)*x[2]**o+(-0.600225027769)*x[2]
        s[3,2,2,0]=(0.065423903522)*x[0]**o+(0.384335040067)*x[0]+(0.19683258538)*x[1]**o+(0.267437363607)*x[1]+(0.0294176448501)*x[2]**o+(0.155211007088)*x[2]
        s[3,2,2,1]=(-0.851636568209)*x[0]**o+(-0.761490675173)*x[0]+(-0.729643742868)*x[1]**o+(0.478780948031)*x[1]+(-0.036088351558)*x[2]**o+(0.823507165674)*x[2]
        s[3,3,0,0]=(0.482377341472)*x[0]**o+(0.139556800754)*x[0]+(-0.0112720934043)*x[1]**o+(0.298852046645)*x[1]+(0.145554009271)*x[2]**o+(-0.859974982251)*x[2]
        s[3,3,0,1]=(-0.137104991135)*x[0]**o+(-0.857876623449)*x[0]+(0.258417639346)*x[1]**o+(-0.389256714236)*x[1]+(-0.936199013774)*x[2]**o+(-0.312698752628)*x[2]
        s[3,3,1,0]=(-0.488668823485)*x[0]**o+(-0.771705852614)*x[0]+(-0.450034458754)*x[1]**o+(0.632396692772)*x[1]+(-0.105848719123)*x[2]**o+(0.358481590951)*x[2]
        s[3,3,1,1]=(-0.559261115256)*x[0]**o+(0.770050098012)*x[0]+(-0.125673261546)*x[1]**o+(0.863587697472)*x[1]+(0.94178472327)*x[2]**o+(0.691350347846)*x[2]
        s[3,3,2,0]=(0.348276322002)*x[0]**o+(-0.367507421115)*x[0]+(0.171562046637)*x[1]**o+(-0.98530139967)*x[1]+(-0.374019522945)*x[2]**o+(-0.0772490996658)*x[2]
        s[3,3,2,1]=(-0.91862240772)*x[0]**o+(-0.771336713978)*x[0]+(-0.510735333043)*x[1]**o+(-0.793667481925)*x[1]+(0.793023981359)*x[2]**o+(-0.211127108731)*x[2]
        s[3,4,0,0]=(-0.439892942063)*x[0]**o+(-0.452356550969)*x[0]+(0.255344783422)*x[1]**o+(0.966971078667)*x[1]+(0.770960867316)*x[2]**o+(-0.877647551728)*x[2]
        s[3,4,0,1]=(0.926860716799)*x[0]**o+(-0.325430377076)*x[0]+(0.593914577372)*x[1]**o+(-0.237069180519)*x[1]+(0.0779887304243)*x[2]**o+(0.0757684397401)*x[2]
        s[3,4,1,0]=(-0.865144446511)*x[0]**o+(0.220573510075)*x[0]+(0.574182016443)*x[1]**o+(0.585433407304)*x[1]+(-0.304138910485)*x[2]**o+(-0.26316229504)*x[2]
        s[3,4,1,1]=(-0.0941252376404)*x[0]**o+(-0.441659917216)*x[0]+(0.833947587523)*x[1]**o+(0.226992443402)*x[1]+(0.902425846073)*x[2]**o+(-0.525442142595)*x[2]
        s[3,4,2,0]=(0.782184984647)*x[0]**o+(0.0260219673241)*x[0]+(0.664501465465)*x[1]**o+(-0.571404130587)*x[1]+(0.419283611654)*x[2]**o+(-0.419135271863)*x[2]
        s[3,4,2,1]=(-0.944783897524)*x[0]**o+(-0.173931506017)*x[0]+(-0.0465484376595)*x[1]**o+(0.718604875734)*x[1]+(-0.388606992705)*x[2]**o+(-0.450843297215)*x[2]
        ref[0,0,0,0]=(0.934771837454)/(o+1.)+(-0.434823412793)
        ref[0,0,0,1]=(1.28803242766)/(o+1.)+(-0.0290220842909)
        ref[0,0,1,0]=(-2.40244763121)/(o+1.)+(-0.306269462104)
        ref[0,0,1,1]=(-1.62188544724)/(o+1.)+(0.0251869143786)
        ref[0,0,2,0]=(0.698356878524)/(o+1.)+(-0.435866994151)
        ref[0,0,2,1]=(0.0875965875576)/(o+1.)+(0.655921182948)
        ref[0,1,0,0]=(0.0458359064556)/(o+1.)+(0.200816326568)
        ref[0,1,0,1]=(-0.557948616036)/(o+1.)+(-0.331764866336)
        ref[0,1,1,0]=(-1.37792398905)/(o+1.)+(-0.831943000542)
        ref[0,1,1,1]=(-1.05744784285)/(o+1.)+(-0.363990638493)
        ref[0,1,2,0]=(1.8785239999)/(o+1.)+(0.132914775729)
        ref[0,1,2,1]=(-0.491737663947)/(o+1.)+(-0.262264208574)
        ref[0,2,0,0]=(0.981768624328)/(o+1.)+(-0.528154847194)
        ref[0,2,0,1]=(-0.924041283835)/(o+1.)+(-0.0118456080214)
        ref[0,2,1,0]=(1.02602669138)/(o+1.)+(0.991571987955)
        ref[0,2,1,1]=(0.293350581922)/(o+1.)+(-1.26454746919)
        ref[0,2,2,0]=(1.20547396122)/(o+1.)+(0.33128271182)
        ref[0,2,2,1]=(-0.633265442539)/(o+1.)+(0.576446315531)
        ref[0,3,0,0]=(-0.107677324633)/(o+1.)+(0.0153782031713)
        ref[0,3,0,1]=(-0.058445812213)/(o+1.)+(0.454914614557)
        ref[0,3,1,0]=(-0.630456060199)/(o+1.)+(-0.155638329723)
        ref[0,3,1,1]=(1.19340007429)/(o+1.)+(-0.605390734274)
        ref[0,3,2,0]=(0.207048325661)/(o+1.)+(0.396354470118)
        ref[0,3,2,1]=(-0.124698590323)/(o+1.)+(-0.533691881821)
        ref[0,4,0,0]=(-0.00894177493875)/(o+1.)+(-0.271916856357)
        ref[0,4,0,1]=(-1.32890925126)/(o+1.)+(-0.0898410243139)
        ref[0,4,1,0]=(0.0975748584924)/(o+1.)+(0.535517090007)
        ref[0,4,1,1]=(-0.476248278814)/(o+1.)+(0.0209889925155)
        ref[0,4,2,0]=(0.770635807959)/(o+1.)+(-0.938934351625)
        ref[0,4,2,1]=(0.160496937995)/(o+1.)+(0.114230384477)
        ref[1,0,0,0]=(0.0402748792799)/(o+1.)+(0.182546725732)
        ref[1,0,0,1]=(-0.262170284078)/(o+1.)+(0.523009320895)
        ref[1,0,1,0]=(0.161867619856)/(o+1.)+(0.471253473618)
        ref[1,0,1,1]=(-0.0310176979524)/(o+1.)+(-0.613310683933)
        ref[1,0,2,0]=(0.414001554119)/(o+1.)+(-0.840604865888)
        ref[1,0,2,1]=(-1.84830463035)/(o+1.)+(0.500311754645)
        ref[1,1,0,0]=(-1.72483479516)/(o+1.)+(-0.638391130662)
        ref[1,1,0,1]=(0.845004524414)/(o+1.)+(0.127933233045)
        ref[1,1,1,0]=(-2.23640238202)/(o+1.)+(0.931171483895)
        ref[1,1,1,1]=(-0.653016296101)/(o+1.)+(-0.0465695510937)
        ref[1,1,2,0]=(1.30266900346)/(o+1.)+(0.406508116258)
        ref[1,1,2,1]=(0.940394628687)/(o+1.)+(-0.0382056638017)
        ref[1,2,0,0]=(-1.760002037)/(o+1.)+(0.762571118282)
        ref[1,2,0,1]=(0.0599325111348)/(o+1.)+(0.88278715062)
        ref[1,2,1,0]=(1.13325081764)/(o+1.)+(-0.298944519128)
        ref[1,2,1,1]=(0.398128607122)/(o+1.)+(-0.266205921249)
        ref[1,2,2,0]=(0.366682002873)/(o+1.)+(-0.400704896527)
        ref[1,2,2,1]=(-0.136500445542)/(o+1.)+(-0.207374711313)
        ref[1,3,0,0]=(-0.265899226265)/(o+1.)+(0.0876398784989)
        ref[1,3,0,1]=(-1.0047182726)/(o+1.)+(-0.305250983945)
        ref[1,3,1,0]=(0.339146148316)/(o+1.)+(-0.601809282005)
        ref[1,3,1,1]=(-0.180834464635)/(o+1.)+(-0.689621840485)
        ref[1,3,2,0]=(-0.221314207244)/(o+1.)+(-0.0359209514441)
        ref[1,3,2,1]=(1.11791107754)/(o+1.)+(-1.11874071557)
        ref[1,4,0,0]=(-0.00303741845779)/(o+1.)+(0.718914277262)
        ref[1,4,0,1]=(1.26926346819)/(o+1.)+(-0.325837866004)
        ref[1,4,1,0]=(-0.492469550194)/(o+1.)+(-0.289736873753)
        ref[1,4,1,1]=(0.353522251312)/(o+1.)+(-1.10599684978)
        ref[1,4,2,0]=(0.353627341229)/(o+1.)+(0.0483477873267)
        ref[1,4,2,1]=(-0.7735107618)/(o+1.)+(-0.518761884451)
        ref[2,0,0,0]=(0.793697859971)/(o+1.)+(-0.573428533787)
        ref[2,0,0,1]=(-0.607778444644)/(o+1.)+(-1.26439762233)
        ref[2,0,1,0]=(-2.15842546907)/(o+1.)+(0.153837078344)
        ref[2,0,1,1]=(1.18258624713)/(o+1.)+(0.557418847254)
        ref[2,0,2,0]=(0.7897031898)/(o+1.)+(-0.502932875729)
        ref[2,0,2,1]=(1.60828071001)/(o+1.)+(-0.433073765244)
        ref[2,1,0,0]=(1.20979858617)/(o+1.)+(-0.452436438458)
        ref[2,1,0,1]=(-0.15595675075)/(o+1.)+(-0.0304279934298)
        ref[2,1,1,0]=(-0.561834715406)/(o+1.)+(0.803031377485)
        ref[2,1,1,1]=(0.0395341731061)/(o+1.)+(0.366443595681)
        ref[2,1,2,0]=(2.48867633741)/(o+1.)+(-0.180417298296)
        ref[2,1,2,1]=(0.149331123022)/(o+1.)+(-1.07148968987)
        ref[2,2,0,0]=(-0.209434754019)/(o+1.)+(-0.763106212061)
        ref[2,2,0,1]=(2.44814041751)/(o+1.)+(-1.02130994501)
        ref[2,2,1,0]=(-0.651817994783)/(o+1.)+(0.212869463232)
        ref[2,2,1,1]=(-0.9098862936)/(o+1.)+(-0.367079660928)
        ref[2,2,2,0]=(-0.617978035965)/(o+1.)+(-0.621902294706)
        ref[2,2,2,1]=(0.390306958936)/(o+1.)+(0.0173625539477)
        ref[2,3,0,0]=(-0.300187497765)/(o+1.)+(0.28780146874)
        ref[2,3,0,1]=(1.31507975214)/(o+1.)+(0.210484455741)
        ref[2,3,1,0]=(-2.56413905202)/(o+1.)+(0.544520051234)
        ref[2,3,1,1]=(0.601936722022)/(o+1.)+(-0.313181460754)
        ref[2,3,2,0]=(0.758128840732)/(o+1.)+(-0.260265102571)
        ref[2,3,2,1]=(0.809610237649)/(o+1.)+(0.527311663172)
        ref[2,4,0,0]=(-0.873466583322)/(o+1.)+(-0.229866893409)
        ref[2,4,0,1]=(-0.470871405045)/(o+1.)+(0.907523982721)
        ref[2,4,1,0]=(1.58820295242)/(o+1.)+(0.44139731694)
        ref[2,4,1,1]=(-0.627757953856)/(o+1.)+(-0.225552715397)
        ref[2,4,2,0]=(-0.337082944702)/(o+1.)+(-0.322241822947)
        ref[2,4,2,1]=(-0.75435810202)/(o+1.)+(0.596137934342)
        ref[3,0,0,0]=(-0.658809182448)/(o+1.)+(-0.388745102312)
        ref[3,0,0,1]=(-0.668726512445)/(o+1.)+(-0.200455602857)
        ref[3,0,1,0]=(-0.316120206381)/(o+1.)+(0.493293296905)
        ref[3,0,1,1]=(-1.46721022602)/(o+1.)+(-0.139825346344)
        ref[3,0,2,0]=(-1.88559464763)/(o+1.)+(0.405324570636)
        ref[3,0,2,1]=(-1.43966067234)/(o+1.)+(0.0980357738822)
        ref[3,1,0,0]=(-0.448161893009)/(o+1.)+(-0.0308422694803)
        ref[3,1,0,1]=(-0.0540372986773)/(o+1.)+(-0.851335446411)
        ref[3,1,1,0]=(0.709145995686)/(o+1.)+(-0.0440814976379)
        ref[3,1,1,1]=(-1.54203885293)/(o+1.)+(0.215708042053)
        ref[3,1,2,0]=(-0.232776059422)/(o+1.)+(0.0143011952278)
        ref[3,1,2,1]=(1.4264927767)/(o+1.)+(-1.21801289226)
        ref[3,2,0,0]=(-0.145171064331)/(o+1.)+(1.01667674349)
        ref[3,2,0,1]=(0.908967294363)/(o+1.)+(0.444544005797)
        ref[3,2,1,0]=(-1.26527555764)/(o+1.)+(-0.169158246218)
        ref[3,2,1,1]=(-1.35199519577)/(o+1.)+(-0.178153435155)
        ref[3,2,2,0]=(0.291674133752)/(o+1.)+(0.403491705381)
        ref[3,2,2,1]=(-1.61736866264)/(o+1.)+(0.270398719266)
        ref[3,3,0,0]=(0.616659257339)/(o+1.)+(-0.210783067426)
        ref[3,3,0,1]=(-0.814886365562)/(o+1.)+(-0.779916045157)
        ref[3,3,1,0]=(-1.04455200136)/(o+1.)+(0.109586215554)
        ref[3,3,1,1]=(0.256850346468)/(o+1.)+(1.16249407166)
        ref[3,3,2,0]=(0.145818845694)/(o+1.)+(-0.715028960225)
        ref[3,3,2,1]=(-0.636333759404)/(o+1.)+(-0.888065652317)
        ref[3,4,0,0]=(0.586412708675)/(o+1.)+(-0.181516512015)
        ref[3,4,0,1]=(1.5987640246)/(o+1.)+(-0.243365558927)
        ref[3,4,1,0]=(-0.595101340552)/(o+1.)+(0.271422311169)
        ref[3,4,1,1]=(1.64224819596)/(o+1.)+(-0.370054808205)
        ref[3,4,2,0]=(1.86597006177)/(o+1.)+(-0.482258717563)
        ref[3,4,2,1]=(-1.37993932789)/(o+1.)+(0.0469150362512)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.518496555195)*x[0]+(-1.7081835316)*x[1]
        ref=(-0.587946015314)/(o+1.)+(-0.300870480545)
      else:
        s=(0.271043713323)*x[0]+(-1.57101724485)*x[1]+(-0.650553623396)*x[2]
        ref=(-1.53180577827)/(o+1.)+(-0.209360688325)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.228569380378)*x[0]+(0.808739010281)*x[1]
        s[1]=(0.658420467565)*x[0]+(0.755018594164)*x[1]
        ref[0]=(0.635537444802)/(o+1.)+(-0.0276839074492)
        ref[1]=(0.431019006365)/(o+1.)+(0.491210027682)
      else:
        s[0]=(0.376210963228)*x[0]+(-1.32850082986)*x[1]+(0.536274160408)*x[2]
        s[1]=(-0.771050457843)*x[0]+(-0.16954629666)*x[1]+(-0.579642261351)*x[2]
        ref[0]=(-0.923045104748)/(o+1.)+(0.253514699262)
        ref[1]=(0.10262370672)/(o+1.)+(-0.811431361287)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(-0.405824989536)*x[0]+(-1.36105849061)*x[1]
        s[0,1]=(0.759504921251)*x[0]+(0.455515132277)*x[1]
        s[0,2]=(-1.37469026888)*x[0]+(1.35869839203)*x[1]
        s[0,3]=(-0.0438580623198)*x[0]+(0.34881551198)*x[1]
        s[0,4]=(-0.0097850380672)*x[0]+(0.61199301527)*x[1]
        s[1,0]=(-1.24950855151)*x[0]+(0.560604227152)*x[1]
        s[1,1]=(0.897125563557)*x[0]+(-0.362799426566)*x[1]
        s[1,2]=(-0.475265733496)*x[0]+(-0.906032283495)*x[1]
        s[1,3]=(0.184749910708)*x[0]+(0.961807058286)*x[1]
        s[1,4]=(-1.8130820241)*x[0]+(-0.0716801569248)*x[1]
        s[2,0]=(0.401163196364)*x[0]+(-0.122822662117)*x[1]
        s[2,1]=(-0.183141841931)*x[0]+(0.159646202564)*x[1]
        s[2,2]=(0.494070531254)*x[0]+(0.0283267498222)*x[1]
        s[2,3]=(-0.46326385119)*x[0]+(0.416856391877)*x[1]
        s[2,4]=(0.669972632028)*x[0]+(1.01207764348)*x[1]
        s[3,0]=(0.579509485187)*x[0]+(0.578664253542)*x[1]
        s[3,1]=(1.1904242989)*x[0]+(-0.617672917808)*x[1]
        s[3,2]=(1.01069090914)*x[0]+(-0.132861271259)*x[1]
        s[3,3]=(-1.02014113453)*x[0]+(-1.10411835631)*x[1]
        s[3,4]=(-0.178208550651)*x[0]+(-0.590646644191)*x[1]
        ref[0,0]=(-1.81750521429)/(o+1.)+(0.0253108670731)
        ref[0,1]=(1.13215741955)/(o+1.)+(0.0414313169864)
        ref[0,2]=(-0.346109803394)/(o+1.)+(0.165058963273)
        ref[0,3]=(1.16722378378)/(o+1.)+(-0.431133167061)
        ref[0,4]=(1.39150379152)/(o+1.)+(-0.39464790716)
        ref[1,0]=(-0.249819369301)/(o+1.)+(-0.219542477529)
        ref[1,1]=(-0.253832127476)/(o+1.)+(0.394079132234)
        ref[1,2]=(-0.514583575657)/(o+1.)+(-0.433357220667)
        ref[1,3]=(1.16430206364)/(o+1.)+(-0.00887254732364)
        ref[1,4]=(-1.40353822359)/(o+1.)+(-0.240611978717)
        ref[2,0]=(0.0129701699159)/(o+1.)+(0.132685182166)
        ref[2,1]=(-0.000259682696025)/(o+1.)+(-0.0116179783357)
        ref[2,2]=(-0.736785877312)/(o+1.)+(0.629591579194)
        ref[2,3]=(-0.319819594662)/(o+1.)+(0.136706067675)
        ref[2,4]=(1.34735651213)/(o+1.)+(0.167346881689)
        ref[3,0]=(1.02891998838)/(o+1.)+(0.0646268751734)
        ref[3,1]=(0.0226437097602)/(o+1.)+(0.275053835665)
        ref[3,2]=(-0.367947454589)/(o+1.)+(0.622888546236)
        ref[3,3]=(-1.07416112237)/(o+1.)+(-0.525049184236)
        ref[3,4]=(-0.423224656448)/(o+1.)+(-0.172815269197)
      else:
        s[0,0]=(0.391957465543)*x[0]+(-0.836212796005)*x[1]+(0.541170293661)*x[2]
        s[0,1]=(-0.324966494738)*x[0]+(0.62718596172)*x[1]+(-0.00110999366491)*x[2]
        s[0,2]=(0.377719216476)*x[0]+(1.20630293683)*x[1]+(-0.936497207115)*x[2]
        s[0,3]=(-1.52853135673)*x[0]+(-0.634683048191)*x[1]+(-0.533312108123)*x[2]
        s[0,4]=(-0.659359670617)*x[0]+(-1.23995360412)*x[1]+(0.669989279254)*x[2]
        s[1,0]=(-0.45120395504)*x[0]+(1.25119902426)*x[1]+(0.11139465291)*x[2]
        s[1,1]=(0.431699115955)*x[0]+(1.165662618)*x[1]+(-1.12543997002)*x[2]
        s[1,2]=(0.16640215316)*x[0]+(-1.81838137531)*x[1]+(-0.111705546449)*x[2]
        s[1,3]=(-1.56135929502)*x[0]+(0.425946749357)*x[1]+(0.255013889109)*x[2]
        s[1,4]=(0.320981318781)*x[0]+(-0.536492696361)*x[1]+(1.21768285382)*x[2]
        s[2,0]=(-0.600249234632)*x[0]+(-0.497626140165)*x[1]+(-0.0606420495108)*x[2]
        s[2,1]=(-0.144412727094)*x[0]+(-1.09139116844)*x[1]+(-0.39344520166)*x[2]
        s[2,2]=(-0.433006957898)*x[0]+(-0.287282800253)*x[1]+(8.69987670264e-05)*x[2]
        s[2,3]=(0.805491532167)*x[0]+(-1.17746983966)*x[1]+(0.720861745039)*x[2]
        s[2,4]=(-0.735630820744)*x[0]+(-1.01122803474)*x[1]+(1.72681276559)*x[2]
        s[3,0]=(-0.739861204924)*x[0]+(0.739077496919)*x[1]+(-0.0950784436747)*x[2]
        s[3,1]=(0.428369405071)*x[0]+(-0.0613035842349)*x[1]+(1.22804321951)*x[2]
        s[3,2]=(0.617885252444)*x[0]+(-1.63428007618)*x[1]+(-0.543704110813)*x[2]
        s[3,3]=(-1.52397266949)*x[0]+(-1.33821038473)*x[1]+(0.578214428096)*x[2]
        s[3,4]=(-0.904835367377)*x[0]+(0.156039380736)*x[1]+(-0.833632978592)*x[2]
        ref[0,0]=(-0.865179337718)/(o+1.)+(0.481047150458)
        ref[0,1]=(-0.427981609749)/(o+1.)+(0.364545541533)
        ref[0,2]=(0.738756225478)/(o+1.)+(-0.0456156396413)
        ref[0,3]=(-2.19753999815)/(o+1.)+(-0.249493257444)
        ref[0,4]=(-1.44493155356)/(o+1.)+(0.107803779038)
        ref[1,0]=(1.72841073711)/(o+1.)+(-0.408510507493)
        ref[1,1]=(0.688403407203)/(o+1.)+(-0.108240821633)
        ref[1,2]=(-0.0562832360974)/(o+1.)+(-0.853700766251)
        ref[1,3]=(-0.866037983384)/(o+1.)+(-0.00718033658349)
        ref[1,4]=(1.27849436197)/(o+1.)+(-0.138161442865)
        ref[2,0]=(-0.450814686687)/(o+1.)+(-0.353851368811)
        ref[2,1]=(-0.608210959793)/(o+1.)+(-0.510519068702)
        ref[2,2]=(-1.55008808776)/(o+1.)+(0.414942664188)
        ref[2,3]=(0.242744047944)/(o+1.)+(0.0530696948023)
        ref[2,4]=(-0.793195276992)/(o+1.)+(0.386574593552)
        ref[3,0]=(-0.33646344984)/(o+1.)+(0.12030064908)
        ref[3,1]=(1.37599194013)/(o+1.)+(0.109558550106)
        ref[3,2]=(0.0175612882637)/(o+1.)+(-0.788830111409)
        ref[3,3]=(-0.747828767237)/(o+1.)+(-0.768069929445)
        ref[3,4]=(-0.876520699909)/(o+1.)+(-0.352954132662)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.64806859919)*x[0]+(0.669871100749)*x[1]
        s[0,0,1]=(0.351912095781)*x[0]+(-0.795221209702)*x[1]
        s[0,1,0]=(1.67848694486)*x[0]+(0.294895162463)*x[1]
        s[0,1,1]=(0.360904647807)*x[0]+(1.08975889449)*x[1]
        s[1,0,0]=(-0.0924096394032)*x[0]+(-1.45324202762)*x[1]
        s[1,0,1]=(-0.160994756309)*x[0]+(0.551455091958)*x[1]
        s[1,1,0]=(1.32478072001)*x[0]+(-1.43998718605)*x[1]
        s[1,1,1]=(0.481907947963)*x[0]+(-1.62005808161)*x[1]
        s[2,0,0]=(1.28445085089)*x[0]+(0.139212942532)*x[1]
        s[2,0,1]=(0.13496510335)*x[0]+(-0.3090345869)*x[1]
        s[2,1,0]=(1.34702478342)*x[0]+(0.77762332551)*x[1]
        s[2,1,1]=(0.2270682866)*x[0]+(0.585652373115)*x[1]
        s[3,0,0]=(0.974462828497)*x[0]+(-0.0761666893839)*x[1]
        s[3,0,1]=(0.787654033315)*x[0]+(-0.665952609368)*x[1]
        s[3,1,0]=(-0.535337193665)*x[0]+(-0.10003612917)*x[1]
        s[3,1,1]=(0.90669722873)*x[0]+(0.0637346383222)*x[1]
        s[4,0,0]=(0.224906594767)*x[0]+(0.255067891624)*x[1]
        s[4,0,1]=(1.45652552317)*x[0]+(-0.896697311539)*x[1]
        s[4,1,0]=(0.0111655859631)*x[0]+(1.35361034898)*x[1]
        s[4,1,1]=(-1.49696987574)*x[0]+(-0.0630120006374)*x[1]
        s[5,0,0]=(-1.37528487034)*x[0]+(-1.03082532988)*x[1]
        s[5,0,1]=(0.298794708742)*x[0]+(-0.338420113242)*x[1]
        s[5,1,0]=(0.101365713196)*x[0]+(0.46239070902)*x[1]
        s[5,1,1]=(-0.329260078889)*x[0]+(0.0114285623122)*x[1]
        ref[0,0,0]=(1.04954961047)/(o+1.)+(0.134195044736)
        ref[0,0,1]=(-0.941346926582)/(o+1.)+(0.24901890633)
        ref[0,1,0]=(0.33540890863)/(o+1.)+(0.818986599345)
        ref[0,1,1]=(-0.0452512610385)/(o+1.)+(0.747957401666)
        ref[1,0,0]=(-0.0571552186929)/(o+1.)+(-0.744248224165)
        ref[1,0,1]=(-0.256660001756)/(o+1.)+(0.323560168702)
        ref[1,1,0]=(0.173031418514)/(o+1.)+(-0.144118942277)
        ref[1,1,1]=(-1.27331535296)/(o+1.)+(0.0675826096564)
        ref[2,0,0]=(1.27442931542)/(o+1.)+(0.0746172389969)
        ref[2,0,1]=(1.23134060492)/(o+1.)+(-0.702705044233)
        ref[2,1,0]=(1.44361222924)/(o+1.)+(0.340517939842)
        ref[2,1,1]=(0.858404770157)/(o+1.)+(-0.0228420552212)
        ref[3,0,0]=(0.53268376719)/(o+1.)+(0.182806185962)
        ref[3,0,1]=(0.962024080252)/(o+1.)+(-0.420161328153)
        ref[3,1,0]=(-1.06376900656)/(o+1.)+(0.214197841863)
        ref[3,1,1]=(0.437728927482)/(o+1.)+(0.266351469785)
        ref[4,0,0]=(-0.247848566495)/(o+1.)+(0.363911526443)
        ref[4,0,1]=(0.275565249528)/(o+1.)+(0.142131481049)
        ref[4,1,0]=(0.8446292547)/(o+1.)+(0.260073340122)
        ref[4,1,1]=(-1.92408551747)/(o+1.)+(0.182051820549)
        ref[5,0,0]=(-1.70867649172)/(o+1.)+(-0.348716854247)
        ref[5,0,1]=(-0.216851661682)/(o+1.)+(0.088613128591)
        ref[5,1,0]=(1.89734425595)/(o+1.)+(-0.666793916865)
        ref[5,1,1]=(-0.281093101446)/(o+1.)+(-0.0183692075654)
      else:
        s[0,0,0]=(0.142896594259)*x[0]+(-0.458264763854)*x[1]+(0.521893125785)*x[2]
        s[0,0,1]=(-0.141846975275)*x[0]+(-0.976845409768)*x[1]+(-1.14147520898)*x[2]
        s[0,1,0]=(0.141747271548)*x[0]+(-1.21811158828)*x[1]+(0.0640715930615)*x[2]
        s[0,1,1]=(0.462551385436)*x[0]+(-0.87929813895)*x[1]+(0.944848279649)*x[2]
        s[1,0,0]=(0.246197516828)*x[0]+(-0.0275751890238)*x[1]+(0.278318853806)*x[2]
        s[1,0,1]=(0.525663042785)*x[0]+(0.191011208259)*x[1]+(-1.27931005652)*x[2]
        s[1,1,0]=(-1.12276137904)*x[0]+(0.474659469962)*x[1]+(0.15450537857)*x[2]
        s[1,1,1]=(0.321959890457)*x[0]+(-0.0325903104321)*x[1]+(-0.735505871467)*x[2]
        s[2,0,0]=(-0.190194685931)*x[0]+(-0.277506154989)*x[1]+(0.0984734800625)*x[2]
        s[2,0,1]=(0.286493694398)*x[0]+(0.841559451892)*x[1]+(0.20572294365)*x[2]
        s[2,1,0]=(-0.0132570386632)*x[0]+(-0.160857995452)*x[1]+(-0.813758770212)*x[2]
        s[2,1,1]=(0.00519333657271)*x[0]+(0.0324836424574)*x[1]+(0.493299621992)*x[2]
        s[3,0,0]=(0.300624746821)*x[0]+(0.211207146953)*x[1]+(1.0860831317)*x[2]
        s[3,0,1]=(-0.696336553938)*x[0]+(-1.28551896118)*x[1]+(0.738769481443)*x[2]
        s[3,1,0]=(0.15510282832)*x[0]+(-1.15795290548)*x[1]+(0.543235378485)*x[2]
        s[3,1,1]=(-0.714308237763)*x[0]+(0.239072001232)*x[1]+(-0.187168055861)*x[2]
        s[4,0,0]=(0.624611061079)*x[0]+(1.11568209096)*x[1]+(-1.46683771155)*x[2]
        s[4,0,1]=(1.45301860552)*x[0]+(0.74921646591)*x[1]+(-0.796630526412)*x[2]
        s[4,1,0]=(0.560555339207)*x[0]+(-0.421267308186)*x[1]+(-0.222312671328)*x[2]
        s[4,1,1]=(0.807845684562)*x[0]+(0.323266951017)*x[1]+(1.51940950039)*x[2]
        s[5,0,0]=(-0.558067921385)*x[0]+(-0.214928793546)*x[1]+(1.01402046886)*x[2]
        s[5,0,1]=(1.20688436616)*x[0]+(0.0247589671447)*x[1]+(-0.368536410815)*x[2]
        s[5,1,0]=(-0.324541665933)*x[0]+(-0.205515389215)*x[1]+(0.758278431203)*x[2]
        s[5,1,1]=(-0.620405057772)*x[0]+(-0.67647528448)*x[1]+(1.32328947052)*x[2]
        ref[0,0,0]=(0.260018146786)/(o+1.)+(-0.0267465952979)
        ref[0,0,1]=(-0.435502524101)/(o+1.)+(-0.912332534962)
        ref[0,1,0]=(-0.0330362836983)/(o+1.)+(-0.489628219987)
        ref[0,1,1]=(1.07484163998)/(o+1.)+(-0.273370056924)
        ref[1,0,0]=(-0.523779853488)/(o+1.)+(0.510360517549)
        ref[1,0,1]=(0.448972814496)/(o+1.)+(-0.505804309986)
        ref[1,1,0]=(-1.09840793106)/(o+1.)+(0.302405700276)
        ref[1,1,1]=(0.130511241273)/(o+1.)+(-0.288323766357)
        ref[2,0,0]=(-0.805299870557)/(o+1.)+(0.21803625485)
        ref[2,0,1]=(0.271181617798)/(o+1.)+(0.531297236071)
        ref[2,1,0]=(-1.29575884246)/(o+1.)+(0.153942519064)
        ref[2,1,1]=(-0.600379662508)/(o+1.)+(0.565678131765)
        ref[3,0,0]=(1.87814767227)/(o+1.)+(-0.140116323394)
        ref[3,0,1]=(-0.195825320078)/(o+1.)+(-0.523630356799)
        ref[3,1,0]=(-0.959261224384)/(o+1.)+(0.249823262853)
        ref[3,1,1]=(1.05531553015)/(o+1.)+(-0.858859911271)
        ref[4,0,0]=(0.65497734235)/(o+1.)+(-0.190760950929)
        ref[4,0,1]=(1.37092574409)/(o+1.)+(0.0173394004651)
        ref[4,1,0]=(1.27754767629)/(o+1.)+(-0.6802861583)
        ref[4,1,1]=(1.23997013471)/(o+1.)+(0.70527600063)
        ref[5,0,0]=(-0.383142890439)/(o+1.)+(0.312083322186)
        ref[5,0,1]=(1.64380369652)/(o+1.)+(-0.390348387013)
        ref[5,1,0]=(-0.812713549757)/(o+1.)+(0.520467462905)
        ref[5,1,1]=(0.600505313178)/(o+1.)+(-0.287048092454)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.235888975732)*x[0]+(-0.734372259615)*x[1]
        s[0,0,0,1]=(0.0665724430605)*x[0]+(-0.0890002398629)*x[1]
        s[0,0,1,0]=(-1.19774759676)*x[0]+(1.00321581155)*x[1]
        s[0,0,1,1]=(0.93370004223)*x[0]+(-0.953835581453)*x[1]
        s[0,0,2,0]=(-0.0557204731966)*x[0]+(0.94945217411)*x[1]
        s[0,0,2,1]=(-0.772628465097)*x[0]+(1.77100579145)*x[1]
        s[0,1,0,0]=(-0.405653863463)*x[0]+(0.494482129094)*x[1]
        s[0,1,0,1]=(0.636236763209)*x[0]+(-1.63254203234)*x[1]
        s[0,1,1,0]=(0.941764211044)*x[0]+(0.699520209508)*x[1]
        s[0,1,1,1]=(1.05045643599)*x[0]+(0.0596636815695)*x[1]
        s[0,1,2,0]=(0.578668042684)*x[0]+(0.285442340394)*x[1]
        s[0,1,2,1]=(0.469149909671)*x[0]+(1.26526906754)*x[1]
        s[0,2,0,0]=(1.05435496407)*x[0]+(-0.592645628655)*x[1]
        s[0,2,0,1]=(0.066181979241)*x[0]+(-1.42368311585)*x[1]
        s[0,2,1,0]=(-0.896588999422)*x[0]+(1.44881386304)*x[1]
        s[0,2,1,1]=(-0.862544854709)*x[0]+(-0.398360134784)*x[1]
        s[0,2,2,0]=(1.02969811319)*x[0]+(0.318993808627)*x[1]
        s[0,2,2,1]=(0.0143773859487)*x[0]+(-0.327118788648)*x[1]
        s[0,3,0,0]=(0.021196438108)*x[0]+(1.8515159226)*x[1]
        s[0,3,0,1]=(0.627726779803)*x[0]+(-0.302716383303)*x[1]
        s[0,3,1,0]=(0.165624370858)*x[0]+(0.274919760437)*x[1]
        s[0,3,1,1]=(0.283214725106)*x[0]+(-0.872378951063)*x[1]
        s[0,3,2,0]=(-0.83391985409)*x[0]+(0.731217556829)*x[1]
        s[0,3,2,1]=(0.918833558754)*x[0]+(0.714715370412)*x[1]
        s[0,4,0,0]=(-0.503211677055)*x[0]+(0.0289310095577)*x[1]
        s[0,4,0,1]=(0.597129238132)*x[0]+(0.540475624893)*x[1]
        s[0,4,1,0]=(0.687211570349)*x[0]+(0.594018658674)*x[1]
        s[0,4,1,1]=(1.5970484642)*x[0]+(-0.495319600999)*x[1]
        s[0,4,2,0]=(1.20627016067)*x[0]+(0.461261823726)*x[1]
        s[0,4,2,1]=(0.165927055028)*x[0]+(1.09007328428)*x[1]
        s[1,0,0,0]=(1.16105917598)*x[0]+(1.82590164801)*x[1]
        s[1,0,0,1]=(-1.13479961391)*x[0]+(0.0351464050353)*x[1]
        s[1,0,1,0]=(-0.318954647255)*x[0]+(0.210946813363)*x[1]
        s[1,0,1,1]=(1.31393538581)*x[0]+(0.198463680367)*x[1]
        s[1,0,2,0]=(1.17799558613)*x[0]+(0.514357347746)*x[1]
        s[1,0,2,1]=(0.571191341129)*x[0]+(0.694971250457)*x[1]
        s[1,1,0,0]=(0.132901410189)*x[0]+(-0.262070778236)*x[1]
        s[1,1,0,1]=(-0.871431837436)*x[0]+(-0.514707012778)*x[1]
        s[1,1,1,0]=(-0.319003186732)*x[0]+(-0.014973803902)*x[1]
        s[1,1,1,1]=(0.260213741998)*x[0]+(1.32908581921)*x[1]
        s[1,1,2,0]=(0.811339404893)*x[0]+(-0.304564391555)*x[1]
        s[1,1,2,1]=(-0.062193223101)*x[0]+(-0.911704705741)*x[1]
        s[1,2,0,0]=(0.0932623640149)*x[0]+(1.12570360092)*x[1]
        s[1,2,0,1]=(1.16504942277)*x[0]+(-0.385840092559)*x[1]
        s[1,2,1,0]=(1.3440680469)*x[0]+(0.532696669818)*x[1]
        s[1,2,1,1]=(1.31050128725)*x[0]+(0.00906728815983)*x[1]
        s[1,2,2,0]=(-1.36931832474)*x[0]+(-1.70562158659)*x[1]
        s[1,2,2,1]=(-1.8375028036)*x[0]+(1.52366054242)*x[1]
        s[1,3,0,0]=(0.474023303611)*x[0]+(0.437518737767)*x[1]
        s[1,3,0,1]=(-0.42921544272)*x[0]+(0.387647202618)*x[1]
        s[1,3,1,0]=(0.863045413096)*x[0]+(-0.36260979035)*x[1]
        s[1,3,1,1]=(-0.654365401799)*x[0]+(0.533926246461)*x[1]
        s[1,3,2,0]=(0.418152564814)*x[0]+(1.11544647745)*x[1]
        s[1,3,2,1]=(-1.60947873228)*x[0]+(0.591727770738)*x[1]
        s[1,4,0,0]=(-1.51807668572)*x[0]+(-1.01613349614)*x[1]
        s[1,4,0,1]=(-1.50163863727)*x[0]+(1.50376272164)*x[1]
        s[1,4,1,0]=(0.320776223369)*x[0]+(-0.700391339598)*x[1]
        s[1,4,1,1]=(0.745224476612)*x[0]+(-0.850870606209)*x[1]
        s[1,4,2,0]=(-0.179598404379)*x[0]+(1.75799776411)*x[1]
        s[1,4,2,1]=(0.00909330900106)*x[0]+(-0.077399205942)*x[1]
        s[2,0,0,0]=(0.505719447396)*x[0]+(-0.275603298829)*x[1]
        s[2,0,0,1]=(1.46787325811)*x[0]+(0.534793377012)*x[1]
        s[2,0,1,0]=(0.77645891983)*x[0]+(0.540540443218)*x[1]
        s[2,0,1,1]=(0.429661881583)*x[0]+(1.43421311479)*x[1]
        s[2,0,2,0]=(1.02806543045)*x[0]+(0.931834883879)*x[1]
        s[2,0,2,1]=(0.763031666632)*x[0]+(0.873653388204)*x[1]
        s[2,1,0,0]=(0.261772733655)*x[0]+(-1.18679909814)*x[1]
        s[2,1,0,1]=(-1.61662908583)*x[0]+(1.10646142209)*x[1]
        s[2,1,1,0]=(0.188527773811)*x[0]+(-0.772410233645)*x[1]
        s[2,1,1,1]=(-0.869931105194)*x[0]+(1.51014899284)*x[1]
        s[2,1,2,0]=(0.252478837689)*x[0]+(1.22480134325)*x[1]
        s[2,1,2,1]=(1.49578867187)*x[0]+(-0.877973056373)*x[1]
        s[2,2,0,0]=(-0.209568005598)*x[0]+(-0.0462992389671)*x[1]
        s[2,2,0,1]=(-0.298130269216)*x[0]+(-0.908352713549)*x[1]
        s[2,2,1,0]=(-0.586370615835)*x[0]+(-1.72704626602)*x[1]
        s[2,2,1,1]=(1.26217059553)*x[0]+(0.676548549908)*x[1]
        s[2,2,2,0]=(0.961200845015)*x[0]+(0.612151157303)*x[1]
        s[2,2,2,1]=(0.700873758828)*x[0]+(0.800662765762)*x[1]
        s[2,3,0,0]=(-0.319747641322)*x[0]+(0.689929330298)*x[1]
        s[2,3,0,1]=(-1.04393759081)*x[0]+(-1.80449230281)*x[1]
        s[2,3,1,0]=(-0.0410072962318)*x[0]+(-0.678000289402)*x[1]
        s[2,3,1,1]=(-0.329381648117)*x[0]+(-0.935487069857)*x[1]
        s[2,3,2,0]=(-0.907460976718)*x[0]+(0.0314212805943)*x[1]
        s[2,3,2,1]=(0.593604907619)*x[0]+(-0.479146057787)*x[1]
        s[2,4,0,0]=(-0.0337048332079)*x[0]+(0.239721443625)*x[1]
        s[2,4,0,1]=(0.721609580339)*x[0]+(-1.06464423044)*x[1]
        s[2,4,1,0]=(0.670568080353)*x[0]+(1.28774286264)*x[1]
        s[2,4,1,1]=(0.603608338555)*x[0]+(0.323996221169)*x[1]
        s[2,4,2,0]=(0.0371268677698)*x[0]+(0.401814556385)*x[1]
        s[2,4,2,1]=(0.436926475796)*x[0]+(-0.819433312575)*x[1]
        s[3,0,0,0]=(-0.0190186791033)*x[0]+(0.501672647061)*x[1]
        s[3,0,0,1]=(0.00060856018325)*x[0]+(-0.414668792931)*x[1]
        s[3,0,1,0]=(0.699428985955)*x[0]+(0.612631528371)*x[1]
        s[3,0,1,1]=(-0.659371576463)*x[0]+(-0.0982545515259)*x[1]
        s[3,0,2,0]=(0.444351889183)*x[0]+(0.53998411713)*x[1]
        s[3,0,2,1]=(-0.281563709018)*x[0]+(-0.486501437256)*x[1]
        s[3,1,0,0]=(0.407521411466)*x[0]+(-0.749870044816)*x[1]
        s[3,1,0,1]=(-0.638780264813)*x[0]+(0.230397511262)*x[1]
        s[3,1,1,0]=(1.16000440204)*x[0]+(0.13074916095)*x[1]
        s[3,1,1,1]=(-1.30550439046)*x[0]+(-1.33788704039)*x[1]
        s[3,1,2,0]=(-0.540767742404)*x[0]+(-0.148316212185)*x[1]
        s[3,1,2,1]=(1.70452499495)*x[0]+(0.935421717463)*x[1]
        s[3,2,0,0]=(0.320990904519)*x[0]+(0.467953609756)*x[1]
        s[3,2,0,1]=(-0.26002228203)*x[0]+(-0.925765706976)*x[1]
        s[3,2,1,0]=(0.925754935382)*x[0]+(0.240432086422)*x[1]
        s[3,2,1,1]=(0.0214680080146)*x[0]+(-1.11507977191)*x[1]
        s[3,2,2,0]=(-0.187946153679)*x[0]+(0.024781322847)*x[1]
        s[3,2,2,1]=(-1.41866510495)*x[0]+(0.596173034069)*x[1]
        s[3,3,0,0]=(0.77087649562)*x[0]+(0.848168811103)*x[1]
        s[3,3,0,1]=(0.218810390517)*x[0]+(0.116909485563)*x[1]
        s[3,3,1,0]=(-0.889981738645)*x[0]+(-0.506355969844)*x[1]
        s[3,3,1,1]=(1.15870903185)*x[0]+(-0.8158119166)*x[1]
        s[3,3,2,0]=(-0.345230989856)*x[0]+(-0.138496974465)*x[1]
        s[3,3,2,1]=(-0.819830787817)*x[0]+(-1.66747998417)*x[1]
        s[3,4,0,0]=(-1.74647564112)*x[0]+(0.22309219172)*x[1]
        s[3,4,0,1]=(0.146804717412)*x[0]+(0.493084518827)*x[1]
        s[3,4,1,0]=(-0.536943383893)*x[0]+(-0.686288132381)*x[1]
        s[3,4,1,1]=(-0.142983666789)*x[0]+(-0.183927597857)*x[1]
        s[3,4,2,0]=(-0.173071632973)*x[0]+(-0.696870700802)*x[1]
        s[3,4,2,1]=(-0.359163238166)*x[0]+(-0.172033035327)*x[1]
        ref[0,0,0,0]=(-0.603598770198)/(o+1.)+(0.0525577431575)
        ref[0,0,0,1]=(1.04505857398)/(o+1.)+(-0.533743185393)
        ref[0,0,1,0]=(0.00772071573504)/(o+1.)+(-0.101126250473)
        ref[0,0,1,1]=(-0.290587702404)/(o+1.)+(0.135226081591)
        ref[0,0,2,0]=(0.614426628234)/(o+1.)+(0.139652536339)
        ref[0,0,2,1]=(-0.0228094425795)/(o+1.)+(0.510593384465)
        ref[0,1,0,0]=(-0.436885078328)/(o+1.)+(0.262856671979)
        ref[0,1,0,1]=(0.0510259667548)/(o+1.)+(-0.523665617945)
        ref[0,1,1,0]=(0.744462079536)/(o+1.)+(0.448411170508)
        ref[0,1,1,1]=(1.52720962375)/(o+1.)+(-0.208544753096)
        ref[0,1,2,0]=(0.73405306307)/(o+1.)+(0.0650286600044)
        ref[0,1,2,1]=(0.550924984933)/(o+1.)+(0.591746996137)
        ref[0,2,0,0]=(-0.347797813257)/(o+1.)+(0.404753574338)
        ref[0,2,0,1]=(0.187043042708)/(o+1.)+(-0.772272089659)
        ref[0,2,1,0]=(-0.316098850787)/(o+1.)+(0.434161857202)
        ref[0,2,1,1]=(-0.536124734534)/(o+1.)+(-0.36239012748)
        ref[0,2,2,0]=(0.585045523113)/(o+1.)+(0.381823199352)
        ref[0,2,2,1]=(0.422243751365)/(o+1.)+(-0.367492577032)
        ref[0,3,0,0]=(0.91381158808)/(o+1.)+(0.479450386313)
        ref[0,3,0,1]=(0.209840671177)/(o+1.)+(0.0575848626616)
        ref[0,3,1,0]=(0.158639778589)/(o+1.)+(0.140952176353)
        ref[0,3,1,1]=(-0.364808895601)/(o+1.)+(-0.112177665178)
        ref[0,3,2,0]=(-0.198937574741)/(o+1.)+(0.0481176387404)
        ref[0,3,2,1]=(1.05420599501)/(o+1.)+(0.289671467078)
        ref[0,4,0,0]=(0.811701728881)/(o+1.)+(-0.642991198189)
        ref[0,4,0,1]=(0.693754647803)/(o+1.)+(0.221925107611)
        ref[0,4,1,0]=(1.05453609011)/(o+1.)+(0.113347069457)
        ref[0,4,1,1]=(0.117274271978)/(o+1.)+(0.49222729561)
        ref[0,4,2,0]=(0.496795563283)/(o+1.)+(0.585368210556)
        ref[0,4,2,1]=(0.83272690983)/(o+1.)+(0.211636714741)
        ref[1,0,0,0]=(1.46878592273)/(o+1.)+(0.75908745063)
        ref[1,0,0,1]=(-0.250785630529)/(o+1.)+(-0.424433789171)
        ref[1,0,1,0]=(0.161786720779)/(o+1.)+(-0.134897277336)
        ref[1,0,1,1]=(1.12719401256)/(o+1.)+(0.192602526813)
        ref[1,0,2,0]=(0.7197235394)/(o+1.)+(0.486314697239)
        ref[1,0,2,1]=(1.02804458464)/(o+1.)+(0.119059003471)
        ref[1,1,0,0]=(0.0361584520541)/(o+1.)+(-0.0826639100507)
        ref[1,1,0,1]=(-1.06091812324)/(o+1.)+(-0.162610363489)
        ref[1,1,1,0]=(1.13776585837)/(o+1.)+(-0.735871424501)
        ref[1,1,1,1]=(1.4740891177)/(o+1.)+(0.0576052217538)
        ref[1,1,2,0]=(0.280616623834)/(o+1.)+(0.113079194752)
        ref[1,1,2,1]=(-0.491890699653)/(o+1.)+(-0.241003614595)
        ref[1,2,0,0]=(-0.0342838844534)/(o+1.)+(0.626624924695)
        ref[1,2,0,1]=(0.770911868081)/(o+1.)+(0.00414873106652)
        ref[1,2,1,0]=(0.30749084682)/(o+1.)+(0.784636934951)
        ref[1,2,1,1]=(1.8638645006)/(o+1.)+(-0.272147962595)
        ref[1,2,2,0]=(-1.46841213372)/(o+1.)+(-0.803263888803)
        ref[1,2,2,1]=(-0.302454198157)/(o+1.)+(-0.00569403150815)
        ref[1,3,0,0]=(0.493917522063)/(o+1.)+(0.208812259657)
        ref[1,3,0,1]=(0.476274657976)/(o+1.)+(-0.258921449039)
        ref[1,3,1,0]=(-0.736248174893)/(o+1.)+(0.618341898819)
        ref[1,3,1,1]=(-0.697129670691)/(o+1.)+(0.288345257677)
        ref[1,3,2,0]=(0.303153912627)/(o+1.)+(0.615222564816)
        ref[1,3,2,1]=(-0.58122630211)/(o+1.)+(-0.218262329715)
        ref[1,4,0,0]=(-1.15442442968)/(o+1.)+(-0.689892876091)
        ref[1,4,0,1]=(0.0585826445649)/(o+1.)+(-0.0282292800982)
        ref[1,4,1,0]=(0.3252337262)/(o+1.)+(-0.352424421214)
        ref[1,4,1,1]=(0.512701244088)/(o+1.)+(-0.309173686843)
        ref[1,4,2,0]=(1.5506882457)/(o+1.)+(0.0138555570174)
        ref[1,4,2,1]=(-0.940773035102)/(o+1.)+(0.436233569081)
        ref[2,0,0,0]=(1.54247754874)/(o+1.)+(-0.656180700088)
        ref[2,0,0,1]=(0.988306961756)/(o+1.)+(0.507179836684)
        ref[2,0,1,0]=(-0.416628824296)/(o+1.)+(0.866814093672)
        ref[2,0,1,1]=(1.41006066195)/(o+1.)+(0.22690716721)
        ref[2,0,2,0]=(0.353011705537)/(o+1.)+(0.803444304396)
        ref[2,0,2,1]=(0.687642260874)/(o+1.)+(0.474521396981)
        ref[2,1,0,0]=(-0.218198987148)/(o+1.)+(-0.353413688668)
        ref[2,1,0,1]=(-0.635306433562)/(o+1.)+(0.0625693849126)
        ref[2,1,1,0]=(0.53416867099)/(o+1.)+(-0.559025565412)
        ref[2,1,1,1]=(0.00543686305333)/(o+1.)+(0.317390512297)
        ref[2,1,2,0]=(0.916573120847)/(o+1.)+(0.280353530048)
        ref[2,1,2,1]=(-0.0537588152546)/(o+1.)+(0.335787215373)
        ref[2,2,0,0]=(1.0055660817)/(o+1.)+(-0.630716663133)
        ref[2,2,0,1]=(0.420474306577)/(o+1.)+(-0.813478644671)
        ref[2,2,1,0]=(-0.723008657238)/(o+1.)+(-0.795204112308)
        ref[2,2,1,1]=(0.306188174034)/(o+1.)+(0.816265485704)
        ref[2,2,2,0]=(0.806158569368)/(o+1.)+(0.383596716475)
        ref[2,2,2,1]=(1.6353377865)/(o+1.)+(-0.066900630957)
        ref[2,3,0,0]=(-1.24813850815)/(o+1.)+(0.809160098565)
        ref[2,3,0,1]=(-1.56942041127)/(o+1.)+(-0.639504741178)
        ref[2,3,1,0]=(-1.49605878837)/(o+1.)+(0.388525601366)
        ref[2,3,1,1]=(-0.448403645878)/(o+1.)+(-0.408232536048)
        ref[2,3,2,0]=(-1.13568794342)/(o+1.)+(0.129824123646)
        ref[2,3,2,1]=(0.596733666259)/(o+1.)+(-0.241137408214)
        ref[2,4,0,0]=(0.0191094707617)/(o+1.)+(0.0934535698276)
        ref[2,4,0,1]=(0.185204097773)/(o+1.)+(-0.264119373935)
        ref[2,4,1,0]=(1.55855563639)/(o+1.)+(0.199877653305)
        ref[2,4,1,1]=(0.11469142184)/(o+1.)+(0.406456568942)
        ref[2,4,2,0]=(-0.954321542363)/(o+1.)+(0.696631483259)
        ref[2,4,2,1]=(-0.810227566168)/(o+1.)+(0.213860364695)
        ref[3,0,0,0]=(0.453436736881)/(o+1.)+(0.0146086155385)
        ref[3,0,0,1]=(-0.210824760654)/(o+1.)+(-0.101617736047)
        ref[3,0,1,0]=(0.414458928012)/(o+1.)+(0.448800793157)
        ref[3,0,1,1]=(0.346129385596)/(o+1.)+(-0.551877756793)
        ref[3,0,2,0]=(1.85496275992)/(o+1.)+(-0.435313376805)
        ref[3,0,2,1]=(-0.110500317363)/(o+1.)+(-0.328782414455)
        ref[3,1,0,0]=(0.0399449935267)/(o+1.)+(-0.191146813438)
        ref[3,1,0,1]=(0.133080949177)/(o+1.)+(-0.270731851364)
        ref[3,1,1,0]=(0.139648356416)/(o+1.)+(0.575552603289)
        ref[3,1,1,1]=(-1.45813597924)/(o+1.)+(-0.592627725804)
        ref[3,1,2,0]=(-0.773854684679)/(o+1.)+(0.0423853650452)
        ref[3,1,2,1]=(1.2726274277)/(o+1.)+(0.683659642354)
        ref[3,2,0,0]=(0.555324507349)/(o+1.)+(0.116810003463)
        ref[3,2,0,1]=(-0.310977730161)/(o+1.)+(-0.437405129422)
        ref[3,2,1,0]=(1.05800939782)/(o+1.)+(0.0540888119933)
        ref[3,2,1,1]=(0.386222714751)/(o+1.)+(-0.739917239322)
        ref[3,2,2,0]=(-0.95705464593)/(o+1.)+(0.396944907549)
        ref[3,2,2,1]=(-0.491612682819)/(o+1.)+(-0.165439694033)
        ref[3,3,0,0]=(0.666972772584)/(o+1.)+(0.47603626707)
        ref[3,3,0,1]=(0.268506355312)/(o+1.)+(0.0336067603842)
        ref[3,3,1,0]=(-0.57724956178)/(o+1.)+(-0.409544073355)
        ref[3,3,1,1]=(-0.518808765862)/(o+1.)+(0.430852940555)
        ref[3,3,2,0]=(-1.95248352029)/(o+1.)+(0.734377777986)
        ref[3,3,2,1]=(-1.14565232021)/(o+1.)+(-0.67082922589)
        ref[3,4,0,0]=(-1.13455573259)/(o+1.)+(-0.1944138584)
        ref[3,4,0,1]=(-0.508914286468)/(o+1.)+(0.574401761354)
        ref[3,4,1,0]=(-0.319104038837)/(o+1.)+(-0.452063738718)
        ref[3,4,1,1]=(-0.100356733327)/(o+1.)+(-0.11327726566)
        ref[3,4,2,0]=(-0.39970881352)/(o+1.)+(-0.235116760127)
        ref[3,4,2,1]=(-0.339631070901)/(o+1.)+(-0.0957826012957)
      else:
        s[0,0,0,0]=(0.0794776522341)*x[0]+(1.69294456255)*x[1]+(0.824115844979)*x[2]
        s[0,0,0,1]=(0.0720523378919)*x[0]+(-1.39724154561)*x[1]+(-0.010988810956)*x[2]
        s[0,0,1,0]=(0.985556321125)*x[0]+(1.12133844031)*x[1]+(-0.262491219522)*x[2]
        s[0,0,1,1]=(0.789643681767)*x[0]+(1.02860624013)*x[1]+(-0.450770685703)*x[2]
        s[0,0,2,0]=(-0.850427026537)*x[0]+(-0.393993518297)*x[1]+(0.12816296306)*x[2]
        s[0,0,2,1]=(-0.47508130541)*x[0]+(-0.393857510976)*x[1]+(0.66003641476)*x[2]
        s[0,1,0,0]=(-1.87725252884)*x[0]+(-0.668075530728)*x[1]+(0.938485679482)*x[2]
        s[0,1,0,1]=(0.820537136967)*x[0]+(-0.453620529157)*x[1]+(0.811944595883)*x[2]
        s[0,1,1,0]=(0.914995161971)*x[0]+(-0.196153821444)*x[1]+(0.938555859255)*x[2]
        s[0,1,1,1]=(-0.866633718493)*x[0]+(-0.534596859321)*x[1]+(0.77514383239)*x[2]
        s[0,1,2,0]=(0.183094085342)*x[0]+(0.727499336912)*x[1]+(0.170986687969)*x[2]
        s[0,1,2,1]=(-1.36220673904)*x[0]+(0.0822122492483)*x[1]+(0.124227107716)*x[2]
        s[0,2,0,0]=(1.55269479425)*x[0]+(-0.14574857122)*x[1]+(1.24120289667)*x[2]
        s[0,2,0,1]=(-0.464038186768)*x[0]+(0.00554383028356)*x[1]+(-1.14838250985)*x[2]
        s[0,2,1,0]=(0.904608031363)*x[0]+(-0.313004420619)*x[1]+(-0.10529298073)*x[2]
        s[0,2,1,1]=(0.156623578469)*x[0]+(0.55127494399)*x[1]+(-0.385447926692)*x[2]
        s[0,2,2,0]=(-0.766804018933)*x[0]+(0.18519619356)*x[1]+(1.00915540081)*x[2]
        s[0,2,2,1]=(0.00231157548456)*x[0]+(0.338768642838)*x[1]+(-0.684228602438)*x[2]
        s[0,3,0,0]=(-0.834810486599)*x[0]+(-0.096360306458)*x[1]+(1.28924889761)*x[2]
        s[0,3,0,1]=(-0.658443843543)*x[0]+(0.889474267886)*x[1]+(0.566763124725)*x[2]
        s[0,3,1,0]=(-0.109777093963)*x[0]+(0.228053887631)*x[1]+(1.22952539906)*x[2]
        s[0,3,1,1]=(-0.749479260801)*x[0]+(-1.01102432725)*x[1]+(-0.196861989291)*x[2]
        s[0,3,2,0]=(0.650389523509)*x[0]+(0.151701374527)*x[1]+(0.506246388483)*x[2]
        s[0,3,2,1]=(-0.648510958338)*x[0]+(-0.263590202304)*x[1]+(-1.35992044414)*x[2]
        s[0,4,0,0]=(0.196976454523)*x[0]+(-0.193947226599)*x[1]+(1.67641586045)*x[2]
        s[0,4,0,1]=(-0.455868626908)*x[0]+(0.506302321882)*x[1]+(1.1666778393)*x[2]
        s[0,4,1,0]=(-0.788545609857)*x[0]+(-0.487332926053)*x[1]+(-0.0645141715605)*x[2]
        s[0,4,1,1]=(-0.320424709093)*x[0]+(-0.683301403769)*x[1]+(-0.222245108268)*x[2]
        s[0,4,2,0]=(0.876149348517)*x[0]+(0.12449602802)*x[1]+(-0.220531205966)*x[2]
        s[0,4,2,1]=(-0.2764172159)*x[0]+(-0.894571003479)*x[1]+(-0.595830070177)*x[2]
        s[1,0,0,0]=(-0.79676665323)*x[0]+(-1.25965739347)*x[1]+(-0.319474113915)*x[2]
        s[1,0,0,1]=(1.601348448)*x[0]+(0.41027863439)*x[1]+(-0.346758807915)*x[2]
        s[1,0,1,0]=(0.181292913253)*x[0]+(1.10494312205)*x[1]+(0.196973772813)*x[2]
        s[1,0,1,1]=(0.562861818795)*x[0]+(-0.997990211102)*x[1]+(0.0332706615838)*x[2]
        s[1,0,2,0]=(-0.989984164479)*x[0]+(-0.886186583204)*x[1]+(-0.228620900973)*x[2]
        s[1,0,2,1]=(0.841365917044)*x[0]+(-0.0519066039156)*x[1]+(-0.81126199301)*x[2]
        s[1,1,0,0]=(-0.613125126303)*x[0]+(0.482770874125)*x[1]+(0.038374960382)*x[2]
        s[1,1,0,1]=(-0.203719663632)*x[0]+(-0.281873286682)*x[1]+(-0.0303453795347)*x[2]
        s[1,1,1,0]=(1.11287699536)*x[0]+(-1.04759296757)*x[1]+(1.64257695374)*x[2]
        s[1,1,1,1]=(-0.0861307607908)*x[0]+(0.261314888791)*x[1]+(-0.204955973137)*x[2]
        s[1,1,2,0]=(0.821662110998)*x[0]+(-1.05649310988)*x[1]+(0.609822823353)*x[2]
        s[1,1,2,1]=(-0.730526005063)*x[0]+(1.62018363097)*x[1]+(0.99111776492)*x[2]
        s[1,2,0,0]=(-1.25039777751)*x[0]+(0.811070034681)*x[1]+(1.32505446058)*x[2]
        s[1,2,0,1]=(0.753110041895)*x[0]+(0.186756235207)*x[1]+(-1.30430176967)*x[2]
        s[1,2,1,0]=(0.146182105242)*x[0]+(-0.863831927779)*x[1]+(-1.1030247274)*x[2]
        s[1,2,1,1]=(0.674634709918)*x[0]+(-0.560866548824)*x[1]+(0.576408322189)*x[2]
        s[1,2,2,0]=(-0.573489753302)*x[0]+(1.57147013427)*x[1]+(0.711691292064)*x[2]
        s[1,2,2,1]=(0.549483527739)*x[0]+(-1.14179814973)*x[1]+(1.60524619335)*x[2]
        s[1,3,0,0]=(0.197453494754)*x[0]+(-0.456559966113)*x[1]+(-1.21423667747)*x[2]
        s[1,3,0,1]=(-1.76312274691)*x[0]+(-1.04778192886)*x[1]+(-0.185488490674)*x[2]
        s[1,3,1,0]=(0.0162460011376)*x[0]+(1.16407944993)*x[1]+(1.16424268292)*x[2]
        s[1,3,1,1]=(-0.513913288204)*x[0]+(-1.46536589962)*x[1]+(-0.906205233261)*x[2]
        s[1,3,2,0]=(-0.0157400074256)*x[0]+(-0.293812243081)*x[1]+(0.815280978304)*x[2]
        s[1,3,2,1]=(1.57630585875)*x[0]+(0.290739280065)*x[1]+(-0.771398356417)*x[2]
        s[1,4,0,0]=(-0.323882779703)*x[0]+(-0.763929395312)*x[1]+(0.738403272406)*x[2]
        s[1,4,0,1]=(-0.476278532043)*x[0]+(-0.11039317349)*x[1]+(0.346684885005)*x[2]
        s[1,4,1,0]=(-1.36304063164)*x[0]+(-0.660619165296)*x[1]+(-0.903566907206)*x[2]
        s[1,4,1,1]=(0.0863780283437)*x[0]+(0.466792514558)*x[1]+(0.448776109878)*x[2]
        s[1,4,2,0]=(0.74478619068)*x[0]+(0.446314847678)*x[1]+(-0.0481741032015)*x[2]
        s[1,4,2,1]=(0.746506715398)*x[0]+(-0.432201202785)*x[1]+(0.203440105689)*x[2]
        s[2,0,0,0]=(-0.785805880678)*x[0]+(-0.862407720473)*x[1]+(-1.21640925829)*x[2]
        s[2,0,0,1]=(0.647190817984)*x[0]+(0.519851314653)*x[1]+(0.390946186931)*x[2]
        s[2,0,1,0]=(-1.24831725337)*x[0]+(1.59413938467)*x[1]+(-1.33358250136)*x[2]
        s[2,0,1,1]=(-0.152213072763)*x[0]+(-1.09357954369)*x[1]+(-1.25586196425)*x[2]
        s[2,0,2,0]=(-0.0314861108304)*x[0]+(0.293907529511)*x[1]+(-0.21039874768)*x[2]
        s[2,0,2,1]=(0.0467194832302)*x[0]+(0.527286460462)*x[1]+(-0.415618293894)*x[2]
        s[2,1,0,0]=(-0.189575051656)*x[0]+(-0.844885847783)*x[1]+(-0.334235875002)*x[2]
        s[2,1,0,1]=(-0.00822711174783)*x[0]+(-1.5897272285)*x[1]+(1.50002681163)*x[2]
        s[2,1,1,0]=(-0.841981498104)*x[0]+(-1.10376539971)*x[1]+(0.311385102139)*x[2]
        s[2,1,1,1]=(1.12897934543)*x[0]+(-0.966261308872)*x[1]+(-1.16359871767)*x[2]
        s[2,1,2,0]=(-0.957189463062)*x[0]+(0.0471569985184)*x[1]+(0.12392515517)*x[2]
        s[2,1,2,1]=(-0.752369087888)*x[0]+(1.18176351896)*x[1]+(0.334788137458)*x[2]
        s[2,2,0,0]=(1.22014205343)*x[0]+(0.784413576281)*x[1]+(0.0962245348039)*x[2]
        s[2,2,0,1]=(-0.407023035962)*x[0]+(1.11597107325)*x[1]+(0.458612291763)*x[2]
        s[2,2,1,0]=(-0.0496175828387)*x[0]+(-0.0536216340718)*x[1]+(0.771683970464)*x[2]
        s[2,2,1,1]=(-0.356958793645)*x[0]+(-1.38580705404)*x[1]+(-0.670167616875)*x[2]
        s[2,2,2,0]=(0.0804001816041)*x[0]+(-1.16114869406)*x[1]+(0.475472363874)*x[2]
        s[2,2,2,1]=(0.98238551642)*x[0]+(-1.12578358118)*x[1]+(-0.00970359156434)*x[2]
        s[2,3,0,0]=(1.0110556058)*x[0]+(-0.488112488916)*x[1]+(0.243005554591)*x[2]
        s[2,3,0,1]=(-0.16668051279)*x[0]+(0.178067528024)*x[1]+(0.969038052731)*x[2]
        s[2,3,1,0]=(-0.00345650882131)*x[0]+(-0.941671547756)*x[1]+(-0.49908897105)*x[2]
        s[2,3,1,1]=(0.716683430797)*x[0]+(-0.585708024289)*x[1]+(-1.19073441986)*x[2]
        s[2,3,2,0]=(0.400515369791)*x[0]+(-0.454556618814)*x[1]+(0.0782324352019)*x[2]
        s[2,3,2,1]=(-0.164628093414)*x[0]+(-1.26464027792)*x[1]+(0.179511210919)*x[2]
        s[2,4,0,0]=(1.33277550072)*x[0]+(1.18570575409)*x[1]+(0.381346013717)*x[2]
        s[2,4,0,1]=(0.588290284687)*x[0]+(-1.08302445047)*x[1]+(0.166898904485)*x[2]
        s[2,4,1,0]=(0.131178111015)*x[0]+(-0.749855470189)*x[1]+(-0.1247474719)*x[2]
        s[2,4,1,1]=(-0.97245336122)*x[0]+(0.873932906896)*x[1]+(0.120264062027)*x[2]
        s[2,4,2,0]=(-0.722531128456)*x[0]+(0.19636289816)*x[1]+(-0.907945347107)*x[2]
        s[2,4,2,1]=(-0.711111750197)*x[0]+(0.31947094036)*x[1]+(-1.79333697163)*x[2]
        s[3,0,0,0]=(-0.184519023765)*x[0]+(0.680221159423)*x[1]+(-0.0244123658903)*x[2]
        s[3,0,0,1]=(-1.14573637546)*x[0]+(0.0616977700293)*x[1]+(-0.654808111844)*x[2]
        s[3,0,1,0]=(-1.09297543018)*x[0]+(-1.04184353717)*x[1]+(0.384120100538)*x[2]
        s[3,0,1,1]=(0.105247846819)*x[0]+(-1.31600150124)*x[1]+(1.43496073616)*x[2]
        s[3,0,2,0]=(0.881148598911)*x[0]+(-0.478723216871)*x[1]+(-0.0746344543132)*x[2]
        s[3,0,2,1]=(-1.20828681587)*x[0]+(-1.65808839437)*x[1]+(-0.636678098535)*x[2]
        s[3,1,0,0]=(-0.213394309668)*x[0]+(-0.240354748844)*x[1]+(0.117282970374)*x[2]
        s[3,1,0,1]=(1.37462845901)*x[0]+(-0.41583894581)*x[1]+(-0.625626047701)*x[2]
        s[3,1,1,0]=(1.83991491983)*x[0]+(1.2778674571)*x[1]+(1.35465906139)*x[2]
        s[3,1,1,1]=(1.58168546715)*x[0]+(0.829809996131)*x[1]+(0.681232260732)*x[2]
        s[3,1,2,0]=(-0.0212907664121)*x[0]+(1.00109823691)*x[1]+(-1.30845432825)*x[2]
        s[3,1,2,1]=(0.169920360674)*x[0]+(-0.864995850094)*x[1]+(0.302979947072)*x[2]
        s[3,2,0,0]=(-0.154084636739)*x[0]+(0.319647474298)*x[1]+(-1.23506947259)*x[2]
        s[3,2,0,1]=(-0.199216213818)*x[0]+(1.28921202953)*x[1]+(1.65805122541)*x[2]
        s[3,2,1,0]=(-1.86774194364)*x[0]+(-0.101723263166)*x[1]+(-0.526029332636)*x[2]
        s[3,2,1,1]=(-0.177960964857)*x[0]+(-0.899073564225)*x[1]+(0.551378002191)*x[2]
        s[3,2,2,0]=(0.18491319197)*x[0]+(-0.286906366286)*x[1]+(0.583859098776)*x[2]
        s[3,2,2,1]=(0.38623837272)*x[0]+(-0.653272502002)*x[1]+(-1.46718517275)*x[2]
        s[3,3,0,0]=(0.0494572779834)*x[0]+(-0.0566099666381)*x[1]+(-0.113874384717)*x[2]
        s[3,3,0,1]=(0.0241152098269)*x[0]+(-0.317061594612)*x[1]+(1.23769551101)*x[2]
        s[3,3,1,0]=(-0.107294616786)*x[0]+(-0.651410283508)*x[1]+(0.833104181804)*x[2]
        s[3,3,1,1]=(0.16676828757)*x[0]+(-0.670496617303)*x[1]+(1.15686719798)*x[2]
        s[3,3,2,0]=(0.159950143242)*x[0]+(0.0436823700228)*x[1]+(-0.645027680285)*x[2]
        s[3,3,2,1]=(-1.2697347822)*x[0]+(1.22938250997)*x[1]+(0.208587928792)*x[2]
        s[3,4,0,0]=(0.575105413412)*x[0]+(-1.44998456404)*x[1]+(-0.516125694177)*x[2]
        s[3,4,0,1]=(0.0387618783003)*x[0]+(0.859152861676)*x[1]+(1.29847202073)*x[2]
        s[3,4,1,0]=(-0.842510438531)*x[0]+(0.191878858518)*x[1]+(-0.73453059399)*x[2]
        s[3,4,1,1]=(1.41435810392)*x[0]+(-0.855734345577)*x[1]+(-0.425917153945)*x[2]
        s[3,4,2,0]=(-0.58228358663)*x[0]+(0.00936548588525)*x[1]+(-1.38536320116)*x[2]
        s[3,4,2,1]=(0.0419678913201)*x[0]+(1.02999559903)*x[1]+(-0.464745184871)*x[2]
        ref[0,0,0,0]=(1.30172410662)/(o+1.)+(0.647406976573)
        ref[0,0,0,1]=(-1.09442521373)/(o+1.)+(-0.120876402476)
        ref[0,0,1,0]=(2.01757642043)/(o+1.)+(-0.0865864392547)
        ref[0,0,1,1]=(1.80034002198)/(o+1.)+(-0.216430392892)
        ref[0,0,2,0]=(1.13142112916)/(o+1.)+(-1.12383935547)
        ref[0,0,2,1]=(-0.905515986966)/(o+1.)+(0.34830679267)
        ref[0,1,0,0]=(-0.210441134504)/(o+1.)+(-0.698200622791)
        ref[0,1,0,1]=(0.360807961621)/(o+1.)+(0.409026621036)
        ref[0,1,1,0]=(1.5147969883)/(o+1.)+(0.0713001057429)
        ref[0,1,1,1]=(0.396123567843)/(o+1.)+(-0.511105156634)
        ref[0,1,2,0]=(0.184468760768)/(o+1.)+(0.448555674728)
        ref[0,1,2,1]=(-0.133892978353)/(o+1.)+(-0.510937201862)
        ref[0,2,0,0]=(1.66137009462)/(o+1.)+(0.493389512541)
        ref[0,2,0,1]=(0.358452769174)/(o+1.)+(-0.982664817753)
        ref[0,2,1,0]=(0.966434135345)/(o+1.)+(-0.240061752665)
        ref[0,2,1,1]=(-0.401753883503)/(o+1.)+(0.362102239635)
        ref[0,2,2,0]=(-0.0153619747046)/(o+1.)+(0.221454775071)
        ref[0,2,2,1]=(-0.104138117282)/(o+1.)+(-0.119505133417)
        ref[0,3,0,0]=(-0.966204569194)/(o+1.)+(0.662141336872)
        ref[0,3,0,1]=(0.587661220327)/(o+1.)+(0.10506616437)
        ref[0,3,1,0]=(2.02288199931)/(o+1.)+(-0.337539903289)
        ref[0,3,1,1]=(-1.78842825396)/(o+1.)+(-0.0844686616883)
        ref[0,3,2,0]=(0.927282266945)/(o+1.)+(0.190527509786)
        ref[0,3,2,1]=(-0.679518039292)/(o+1.)+(-0.796251782745)
        ref[0,4,0,0]=(0.555388124049)/(o+1.)+(0.562028482163)
        ref[0,4,0,1]=(0.0870397234082)/(o+1.)+(0.565035905435)
        ref[0,4,1,0]=(-1.1463502212)/(o+1.)+(-0.0970212431374)
        ref[0,4,1,1]=(0.233233715042)/(o+1.)+(-0.729602468086)
        ref[0,4,2,0]=(-0.782889963002)/(o+1.)+(0.781502066786)
        ref[0,4,2,1]=(0.763433249649)/(o+1.)+(-1.2651257696)
        ref[1,0,0,0]=(-1.00441831898)/(o+1.)+(-0.685739920821)
        ref[1,0,0,1]=(0.751247546824)/(o+1.)+(0.456810363824)
        ref[1,0,1,0]=(2.22764882334)/(o+1.)+(-0.372219507613)
        ref[1,0,1,1]=(-1.19623631804)/(o+1.)+(0.39718929366)
        ref[1,0,2,0]=(-1.60343741738)/(o+1.)+(-0.250677115636)
        ref[1,0,2,1]=(-1.25680789752)/(o+1.)+(0.61750260882)
        ref[1,1,0,0]=(-1.45519189991)/(o+1.)+(0.681606304054)
        ref[1,1,0,1]=(0.840628495308)/(o+1.)+(-0.678283412578)
        ref[1,1,1,0]=(1.41826768667)/(o+1.)+(0.144796647431)
        ref[1,1,1,1]=(-0.933723251106)/(o+1.)+(0.451975702985)
        ref[1,1,2,0]=(-0.408324994487)/(o+1.)+(0.39165840948)
        ref[1,1,2,1]=(1.23290517672)/(o+1.)+(0.323935107053)
        ref[1,2,0,0]=(0.825483086179)/(o+1.)+(0.0301218157877)
        ref[1,2,0,1]=(-0.0793934768662)/(o+1.)+(-0.142521007854)
        ref[1,2,1,0]=(-0.0709650957035)/(o+1.)+(-0.874854727116)
        ref[1,2,1,1]=(1.10714952407)/(o+1.)+(-0.208486520392)
        ref[1,2,2,0]=(0.99013461187)/(o+1.)+(0.359768530582)
        ref[1,2,2,1]=(0.0356627876162)/(o+1.)+(0.488634391872)
        ref[1,3,0,0]=(-0.474442326803)/(o+1.)+(-0.499450411015)
        ref[1,3,0,1]=(-0.673232992552)/(o+1.)+(-1.16158008694)
        ref[1,3,1,0]=(0.411652221393)/(o+1.)+(0.966457956295)
        ref[1,3,1,1]=(-1.76698398732)/(o+1.)+(-0.559250216883)
        ref[1,3,2,0]=(1.9874048718)/(o+1.)+(-0.740838072)
        ref[1,3,2,1]=(0.661784629796)/(o+1.)+(0.216931076303)
        ref[1,4,0,0]=(-0.513029515019)/(o+1.)+(0.0818103062045)
        ref[1,4,0,1]=(0.282755142694)/(o+1.)+(-0.261370981611)
        ref[1,4,1,0]=(-1.96973437805)/(o+1.)+(-0.478746163042)
        ref[1,4,1,1]=(0.908774952181)/(o+1.)+(0.0465858502998)
        ref[1,4,2,0]=(1.3422757017)/(o+1.)+(-0.0996743832723)
        ref[1,4,2,1]=(1.21609351443)/(o+1.)+(-0.349173948065)
        ref[2,0,0,0]=(-1.01744592956)/(o+1.)+(-0.923588464944)
        ref[2,0,0,1]=(1.10931736222)/(o+1.)+(0.224335478673)
        ref[2,0,1,0]=(-0.416479828321)/(o+1.)+(-0.285640270869)
        ref[2,0,1,1]=(-1.5671174369)/(o+1.)+(-0.467268571897)
        ref[2,0,2,0]=(-0.274214960025)/(o+1.)+(0.163118815513)
        ref[2,0,2,1]=(0.000680528469817)/(o+1.)+(0.0788535606642)
        ref[2,1,0,0]=(-0.844192475139)/(o+1.)+(-0.262252149651)
        ref[2,1,0,1]=(0.69399848864)/(o+1.)+(-0.395963008628)
        ref[2,1,1,0]=(-1.37742489692)/(o+1.)+(-0.128468449379)
        ref[2,1,1,1]=(0.282095133276)/(o+1.)+(-0.641487907195)
        ref[2,1,2,0]=(-1.4727375549)/(o+1.)+(0.343315122764)
        ref[2,1,2,1]=(0.67637649169)/(o+1.)+(0.0439030384195)
        ref[2,2,0,0]=(1.58598306034)/(o+1.)+(0.257398552088)
        ref[2,2,0,1]=(-0.187654980952)/(o+1.)+(0.677607655001)
        ref[2,2,1,0]=(0.225457027395)/(o+1.)+(0.221493863079)
        ref[2,2,1,1]=(-1.66885547416)/(o+1.)+(-0.372038995202)
        ref[2,2,2,0]=(-1.42708510165)/(o+1.)+(0.410904476533)
        ref[2,2,2,1]=(0.352696956576)/(o+1.)+(-0.252899306452)
        ref[2,3,0,0]=(0.266365999523)/(o+1.)+(0.249791335977)
        ref[2,3,0,1]=(0.926084735451)/(o+1.)+(0.0271701662571)
        ref[2,3,1,0]=(-0.138787407751)/(o+1.)+(-0.652714809939)
        ref[2,3,1,1]=(-0.834750293031)/(o+1.)+(-0.112504360161)
        ref[2,3,2,0]=(-0.806210182586)/(o+1.)+(0.415200684382)
        ref[2,3,2,1]=(0.577899253552)/(o+1.)+(-0.913828206984)
        ref[2,4,0,0]=(1.38802720612)/(o+1.)+(0.755900031203)
        ref[2,4,0,1]=(-1.33634216788)/(o+1.)+(0.504253453293)
        ref[2,4,1,0]=(-1.13393506986)/(o+1.)+(0.195255119392)
        ref[2,4,1,1]=(-0.384732222354)/(o+1.)+(0.203237915028)
        ref[2,4,2,0]=(-1.18866097183)/(o+1.)+(-0.122726302784)
        ref[2,4,2,1]=(-1.41914128138)/(o+1.)+(-0.382918250045)
        ref[3,0,0,0]=(-1.3617379499)/(o+1.)+(0.916513859836)
        ref[3,0,0,1]=(-0.577580679368)/(o+1.)+(-0.580633018955)
        ref[3,0,1,0]=(-0.730926522369)/(o+1.)+(-0.50988617222)
        ref[3,0,1,1]=(0.852352706539)/(o+1.)+(-0.314072812402)
        ref[3,0,2,0]=(0.63914695576)/(o+1.)+(-0.155678014016)
        ref[3,0,2,1]=(-2.31555597527)/(o+1.)+(-0.593748666755)
        ref[3,1,0,0]=(-0.31308710102)/(o+1.)+(-0.011689493559)
        ref[3,1,0,1]=(0.565049631081)/(o+1.)+(-0.115943082793)
        ref[3,1,1,0]=(1.76264026827)/(o+1.)+(1.35490058502)
        ref[3,1,1,1]=(1.34557111025)/(o+1.)+(0.873578306883)
        ref[3,1,2,0]=(0.924587947935)/(o+1.)+(-0.626617402844)
        ref[3,1,2,1]=(-0.679695689354)/(o+1.)+(0.143800073503)
        ref[3,2,0,0]=(-0.408824291494)/(o+1.)+(-0.330341171766)
        ref[3,2,0,1]=(1.02728720579)/(o+1.)+(0.860379917667)
        ref[3,2,1,0]=(-1.99227170213)/(o+1.)+(-0.251611418656)
        ref[3,2,1,1]=(-0.844161339595)/(o+1.)+(0.159252406352)
        ref[3,2,2,0]=(1.52009831693)/(o+1.)+(-0.519116196235)
        ref[3,2,2,1]=(0.323553722724)/(o+1.)+(-1.02888651238)
        ref[3,3,0,0]=(-1.39647666034)/(o+1.)+(0.637724793483)
        ref[3,3,0,1]=(0.960901654854)/(o+1.)+(-0.00807626431577)
        ref[3,3,1,0]=(0.383998350003)/(o+1.)+(-0.154799534246)
        ref[3,3,1,1]=(-0.851062083214)/(o+1.)+(0.752100475729)
        ref[3,3,2,0]=(0.788418151505)/(o+1.)+(-0.614906659263)
        ref[3,3,2,1]=(-0.1028922667)/(o+1.)+(0.135563961629)
        ref[3,4,0,0]=(-0.540240088251)/(o+1.)+(-0.425382378277)
        ref[3,4,0,1]=(0.236730341461)/(o+1.)+(0.979828209623)
        ref[3,4,1,0]=(-0.9880898216)/(o+1.)+(-0.198536176202)
        ref[3,4,1,1]=(1.06221574228)/(o+1.)+(-0.46475456894)
        ref[3,4,2,0]=(-1.89471538439)/(o+1.)+(-0.0317829587581)
        ref[3,4,2,1]=(-1.20699948566)/(o+1.)+(0.907108895566)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.518496555195)*x[0]+(-1.7081835316)*x[1]
        ref=(-0.587946015314)/(o+1.)+(-0.300870480545)
      else:
        s=(0.271043713323)*x[0]+(-1.57101724485)*x[1]+(-0.650553623396)*x[2]
        ref=(-1.53180577827)/(o+1.)+(-0.209360688325)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.228569380378)*x[0]+(0.808739010281)*x[1]
        s[1]=(0.658420467565)*x[0]+(0.755018594164)*x[1]
        ref[0]=(0.635537444802)/(o+1.)+(-0.0276839074492)
        ref[1]=(0.431019006365)/(o+1.)+(0.491210027682)
      else:
        s[0]=(0.376210963228)*x[0]+(-1.32850082986)*x[1]+(0.536274160408)*x[2]
        s[1]=(-0.771050457843)*x[0]+(-0.16954629666)*x[1]+(-0.579642261351)*x[2]
        ref[0]=(-0.923045104748)/(o+1.)+(0.253514699262)
        ref[1]=(0.10262370672)/(o+1.)+(-0.811431361287)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(-0.405824989536)*x[0]+(-1.36105849061)*x[1]
        s[0,1]=(0.759504921251)*x[0]+(0.455515132277)*x[1]
        s[0,2]=(-1.37469026888)*x[0]+(1.35869839203)*x[1]
        s[0,3]=(-0.0438580623198)*x[0]+(0.34881551198)*x[1]
        s[0,4]=(-0.0097850380672)*x[0]+(0.61199301527)*x[1]
        s[1,0]=(-1.24950855151)*x[0]+(0.560604227152)*x[1]
        s[1,1]=(0.897125563557)*x[0]+(-0.362799426566)*x[1]
        s[1,2]=(-0.475265733496)*x[0]+(-0.906032283495)*x[1]
        s[1,3]=(0.184749910708)*x[0]+(0.961807058286)*x[1]
        s[1,4]=(-1.8130820241)*x[0]+(-0.0716801569248)*x[1]
        s[2,0]=(0.401163196364)*x[0]+(-0.122822662117)*x[1]
        s[2,1]=(-0.183141841931)*x[0]+(0.159646202564)*x[1]
        s[2,2]=(0.494070531254)*x[0]+(0.0283267498222)*x[1]
        s[2,3]=(-0.46326385119)*x[0]+(0.416856391877)*x[1]
        s[2,4]=(0.669972632028)*x[0]+(1.01207764348)*x[1]
        s[3,0]=(0.579509485187)*x[0]+(0.578664253542)*x[1]
        s[3,1]=(1.1904242989)*x[0]+(-0.617672917808)*x[1]
        s[3,2]=(1.01069090914)*x[0]+(-0.132861271259)*x[1]
        s[3,3]=(-1.02014113453)*x[0]+(-1.10411835631)*x[1]
        s[3,4]=(-0.178208550651)*x[0]+(-0.590646644191)*x[1]
        ref[0,0]=(-1.81750521429)/(o+1.)+(0.0253108670731)
        ref[0,1]=(1.13215741955)/(o+1.)+(0.0414313169864)
        ref[0,2]=(-0.346109803394)/(o+1.)+(0.165058963273)
        ref[0,3]=(1.16722378378)/(o+1.)+(-0.431133167061)
        ref[0,4]=(1.39150379152)/(o+1.)+(-0.39464790716)
        ref[1,0]=(-0.249819369301)/(o+1.)+(-0.219542477529)
        ref[1,1]=(-0.253832127476)/(o+1.)+(0.394079132234)
        ref[1,2]=(-0.514583575657)/(o+1.)+(-0.433357220667)
        ref[1,3]=(1.16430206364)/(o+1.)+(-0.00887254732364)
        ref[1,4]=(-1.40353822359)/(o+1.)+(-0.240611978717)
        ref[2,0]=(0.0129701699159)/(o+1.)+(0.132685182166)
        ref[2,1]=(-0.000259682696025)/(o+1.)+(-0.0116179783357)
        ref[2,2]=(-0.736785877312)/(o+1.)+(0.629591579194)
        ref[2,3]=(-0.319819594662)/(o+1.)+(0.136706067675)
        ref[2,4]=(1.34735651213)/(o+1.)+(0.167346881689)
        ref[3,0]=(1.02891998838)/(o+1.)+(0.0646268751734)
        ref[3,1]=(0.0226437097602)/(o+1.)+(0.275053835665)
        ref[3,2]=(-0.367947454589)/(o+1.)+(0.622888546236)
        ref[3,3]=(-1.07416112237)/(o+1.)+(-0.525049184236)
        ref[3,4]=(-0.423224656448)/(o+1.)+(-0.172815269197)
      else:
        s[0,0]=(0.391957465543)*x[0]+(-0.836212796005)*x[1]+(0.541170293661)*x[2]
        s[0,1]=(-0.324966494738)*x[0]+(0.62718596172)*x[1]+(-0.00110999366491)*x[2]
        s[0,2]=(0.377719216476)*x[0]+(1.20630293683)*x[1]+(-0.936497207115)*x[2]
        s[0,3]=(-1.52853135673)*x[0]+(-0.634683048191)*x[1]+(-0.533312108123)*x[2]
        s[0,4]=(-0.659359670617)*x[0]+(-1.23995360412)*x[1]+(0.669989279254)*x[2]
        s[1,0]=(-0.45120395504)*x[0]+(1.25119902426)*x[1]+(0.11139465291)*x[2]
        s[1,1]=(0.431699115955)*x[0]+(1.165662618)*x[1]+(-1.12543997002)*x[2]
        s[1,2]=(0.16640215316)*x[0]+(-1.81838137531)*x[1]+(-0.111705546449)*x[2]
        s[1,3]=(-1.56135929502)*x[0]+(0.425946749357)*x[1]+(0.255013889109)*x[2]
        s[1,4]=(0.320981318781)*x[0]+(-0.536492696361)*x[1]+(1.21768285382)*x[2]
        s[2,0]=(-0.600249234632)*x[0]+(-0.497626140165)*x[1]+(-0.0606420495108)*x[2]
        s[2,1]=(-0.144412727094)*x[0]+(-1.09139116844)*x[1]+(-0.39344520166)*x[2]
        s[2,2]=(-0.433006957898)*x[0]+(-0.287282800253)*x[1]+(8.69987670264e-05)*x[2]
        s[2,3]=(0.805491532167)*x[0]+(-1.17746983966)*x[1]+(0.720861745039)*x[2]
        s[2,4]=(-0.735630820744)*x[0]+(-1.01122803474)*x[1]+(1.72681276559)*x[2]
        s[3,0]=(-0.739861204924)*x[0]+(0.739077496919)*x[1]+(-0.0950784436747)*x[2]
        s[3,1]=(0.428369405071)*x[0]+(-0.0613035842349)*x[1]+(1.22804321951)*x[2]
        s[3,2]=(0.617885252444)*x[0]+(-1.63428007618)*x[1]+(-0.543704110813)*x[2]
        s[3,3]=(-1.52397266949)*x[0]+(-1.33821038473)*x[1]+(0.578214428096)*x[2]
        s[3,4]=(-0.904835367377)*x[0]+(0.156039380736)*x[1]+(-0.833632978592)*x[2]
        ref[0,0]=(-0.865179337718)/(o+1.)+(0.481047150458)
        ref[0,1]=(-0.427981609749)/(o+1.)+(0.364545541533)
        ref[0,2]=(0.738756225478)/(o+1.)+(-0.0456156396413)
        ref[0,3]=(-2.19753999815)/(o+1.)+(-0.249493257444)
        ref[0,4]=(-1.44493155356)/(o+1.)+(0.107803779038)
        ref[1,0]=(1.72841073711)/(o+1.)+(-0.408510507493)
        ref[1,1]=(0.688403407203)/(o+1.)+(-0.108240821633)
        ref[1,2]=(-0.0562832360974)/(o+1.)+(-0.853700766251)
        ref[1,3]=(-0.866037983384)/(o+1.)+(-0.00718033658349)
        ref[1,4]=(1.27849436197)/(o+1.)+(-0.138161442865)
        ref[2,0]=(-0.450814686687)/(o+1.)+(-0.353851368811)
        ref[2,1]=(-0.608210959793)/(o+1.)+(-0.510519068702)
        ref[2,2]=(-1.55008808776)/(o+1.)+(0.414942664188)
        ref[2,3]=(0.242744047944)/(o+1.)+(0.0530696948023)
        ref[2,4]=(-0.793195276992)/(o+1.)+(0.386574593552)
        ref[3,0]=(-0.33646344984)/(o+1.)+(0.12030064908)
        ref[3,1]=(1.37599194013)/(o+1.)+(0.109558550106)
        ref[3,2]=(0.0175612882637)/(o+1.)+(-0.788830111409)
        ref[3,3]=(-0.747828767237)/(o+1.)+(-0.768069929445)
        ref[3,4]=(-0.876520699909)/(o+1.)+(-0.352954132662)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.64806859919)*x[0]+(0.669871100749)*x[1]
        s[0,0,1]=(0.351912095781)*x[0]+(-0.795221209702)*x[1]
        s[0,1,0]=(1.67848694486)*x[0]+(0.294895162463)*x[1]
        s[0,1,1]=(0.360904647807)*x[0]+(1.08975889449)*x[1]
        s[1,0,0]=(-0.0924096394032)*x[0]+(-1.45324202762)*x[1]
        s[1,0,1]=(-0.160994756309)*x[0]+(0.551455091958)*x[1]
        s[1,1,0]=(1.32478072001)*x[0]+(-1.43998718605)*x[1]
        s[1,1,1]=(0.481907947963)*x[0]+(-1.62005808161)*x[1]
        s[2,0,0]=(1.28445085089)*x[0]+(0.139212942532)*x[1]
        s[2,0,1]=(0.13496510335)*x[0]+(-0.3090345869)*x[1]
        s[2,1,0]=(1.34702478342)*x[0]+(0.77762332551)*x[1]
        s[2,1,1]=(0.2270682866)*x[0]+(0.585652373115)*x[1]
        s[3,0,0]=(0.974462828497)*x[0]+(-0.0761666893839)*x[1]
        s[3,0,1]=(0.787654033315)*x[0]+(-0.665952609368)*x[1]
        s[3,1,0]=(-0.535337193665)*x[0]+(-0.10003612917)*x[1]
        s[3,1,1]=(0.90669722873)*x[0]+(0.0637346383222)*x[1]
        s[4,0,0]=(0.224906594767)*x[0]+(0.255067891624)*x[1]
        s[4,0,1]=(1.45652552317)*x[0]+(-0.896697311539)*x[1]
        s[4,1,0]=(0.0111655859631)*x[0]+(1.35361034898)*x[1]
        s[4,1,1]=(-1.49696987574)*x[0]+(-0.0630120006374)*x[1]
        s[5,0,0]=(-1.37528487034)*x[0]+(-1.03082532988)*x[1]
        s[5,0,1]=(0.298794708742)*x[0]+(-0.338420113242)*x[1]
        s[5,1,0]=(0.101365713196)*x[0]+(0.46239070902)*x[1]
        s[5,1,1]=(-0.329260078889)*x[0]+(0.0114285623122)*x[1]
        ref[0,0,0]=(1.04954961047)/(o+1.)+(0.134195044736)
        ref[0,0,1]=(-0.941346926582)/(o+1.)+(0.24901890633)
        ref[0,1,0]=(0.33540890863)/(o+1.)+(0.818986599345)
        ref[0,1,1]=(-0.0452512610385)/(o+1.)+(0.747957401666)
        ref[1,0,0]=(-0.0571552186929)/(o+1.)+(-0.744248224165)
        ref[1,0,1]=(-0.256660001756)/(o+1.)+(0.323560168702)
        ref[1,1,0]=(0.173031418514)/(o+1.)+(-0.144118942277)
        ref[1,1,1]=(-1.27331535296)/(o+1.)+(0.0675826096564)
        ref[2,0,0]=(1.27442931542)/(o+1.)+(0.0746172389969)
        ref[2,0,1]=(1.23134060492)/(o+1.)+(-0.702705044233)
        ref[2,1,0]=(1.44361222924)/(o+1.)+(0.340517939842)
        ref[2,1,1]=(0.858404770157)/(o+1.)+(-0.0228420552212)
        ref[3,0,0]=(0.53268376719)/(o+1.)+(0.182806185962)
        ref[3,0,1]=(0.962024080252)/(o+1.)+(-0.420161328153)
        ref[3,1,0]=(-1.06376900656)/(o+1.)+(0.214197841863)
        ref[3,1,1]=(0.437728927482)/(o+1.)+(0.266351469785)
        ref[4,0,0]=(-0.247848566495)/(o+1.)+(0.363911526443)
        ref[4,0,1]=(0.275565249528)/(o+1.)+(0.142131481049)
        ref[4,1,0]=(0.8446292547)/(o+1.)+(0.260073340122)
        ref[4,1,1]=(-1.92408551747)/(o+1.)+(0.182051820549)
        ref[5,0,0]=(-1.70867649172)/(o+1.)+(-0.348716854247)
        ref[5,0,1]=(-0.216851661682)/(o+1.)+(0.088613128591)
        ref[5,1,0]=(1.89734425595)/(o+1.)+(-0.666793916865)
        ref[5,1,1]=(-0.281093101446)/(o+1.)+(-0.0183692075654)
      else:
        s[0,0,0]=(0.142896594259)*x[0]+(-0.458264763854)*x[1]+(0.521893125785)*x[2]
        s[0,0,1]=(-0.141846975275)*x[0]+(-0.976845409768)*x[1]+(-1.14147520898)*x[2]
        s[0,1,0]=(0.141747271548)*x[0]+(-1.21811158828)*x[1]+(0.0640715930615)*x[2]
        s[0,1,1]=(0.462551385436)*x[0]+(-0.87929813895)*x[1]+(0.944848279649)*x[2]
        s[1,0,0]=(0.246197516828)*x[0]+(-0.0275751890238)*x[1]+(0.278318853806)*x[2]
        s[1,0,1]=(0.525663042785)*x[0]+(0.191011208259)*x[1]+(-1.27931005652)*x[2]
        s[1,1,0]=(-1.12276137904)*x[0]+(0.474659469962)*x[1]+(0.15450537857)*x[2]
        s[1,1,1]=(0.321959890457)*x[0]+(-0.0325903104321)*x[1]+(-0.735505871467)*x[2]
        s[2,0,0]=(-0.190194685931)*x[0]+(-0.277506154989)*x[1]+(0.0984734800625)*x[2]
        s[2,0,1]=(0.286493694398)*x[0]+(0.841559451892)*x[1]+(0.20572294365)*x[2]
        s[2,1,0]=(-0.0132570386632)*x[0]+(-0.160857995452)*x[1]+(-0.813758770212)*x[2]
        s[2,1,1]=(0.00519333657271)*x[0]+(0.0324836424574)*x[1]+(0.493299621992)*x[2]
        s[3,0,0]=(0.300624746821)*x[0]+(0.211207146953)*x[1]+(1.0860831317)*x[2]
        s[3,0,1]=(-0.696336553938)*x[0]+(-1.28551896118)*x[1]+(0.738769481443)*x[2]
        s[3,1,0]=(0.15510282832)*x[0]+(-1.15795290548)*x[1]+(0.543235378485)*x[2]
        s[3,1,1]=(-0.714308237763)*x[0]+(0.239072001232)*x[1]+(-0.187168055861)*x[2]
        s[4,0,0]=(0.624611061079)*x[0]+(1.11568209096)*x[1]+(-1.46683771155)*x[2]
        s[4,0,1]=(1.45301860552)*x[0]+(0.74921646591)*x[1]+(-0.796630526412)*x[2]
        s[4,1,0]=(0.560555339207)*x[0]+(-0.421267308186)*x[1]+(-0.222312671328)*x[2]
        s[4,1,1]=(0.807845684562)*x[0]+(0.323266951017)*x[1]+(1.51940950039)*x[2]
        s[5,0,0]=(-0.558067921385)*x[0]+(-0.214928793546)*x[1]+(1.01402046886)*x[2]
        s[5,0,1]=(1.20688436616)*x[0]+(0.0247589671447)*x[1]+(-0.368536410815)*x[2]
        s[5,1,0]=(-0.324541665933)*x[0]+(-0.205515389215)*x[1]+(0.758278431203)*x[2]
        s[5,1,1]=(-0.620405057772)*x[0]+(-0.67647528448)*x[1]+(1.32328947052)*x[2]
        ref[0,0,0]=(0.260018146786)/(o+1.)+(-0.0267465952979)
        ref[0,0,1]=(-0.435502524101)/(o+1.)+(-0.912332534962)
        ref[0,1,0]=(-0.0330362836983)/(o+1.)+(-0.489628219987)
        ref[0,1,1]=(1.07484163998)/(o+1.)+(-0.273370056924)
        ref[1,0,0]=(-0.523779853488)/(o+1.)+(0.510360517549)
        ref[1,0,1]=(0.448972814496)/(o+1.)+(-0.505804309986)
        ref[1,1,0]=(-1.09840793106)/(o+1.)+(0.302405700276)
        ref[1,1,1]=(0.130511241273)/(o+1.)+(-0.288323766357)
        ref[2,0,0]=(-0.805299870557)/(o+1.)+(0.21803625485)
        ref[2,0,1]=(0.271181617798)/(o+1.)+(0.531297236071)
        ref[2,1,0]=(-1.29575884246)/(o+1.)+(0.153942519064)
        ref[2,1,1]=(-0.600379662508)/(o+1.)+(0.565678131765)
        ref[3,0,0]=(1.87814767227)/(o+1.)+(-0.140116323394)
        ref[3,0,1]=(-0.195825320078)/(o+1.)+(-0.523630356799)
        ref[3,1,0]=(-0.959261224384)/(o+1.)+(0.249823262853)
        ref[3,1,1]=(1.05531553015)/(o+1.)+(-0.858859911271)
        ref[4,0,0]=(0.65497734235)/(o+1.)+(-0.190760950929)
        ref[4,0,1]=(1.37092574409)/(o+1.)+(0.0173394004651)
        ref[4,1,0]=(1.27754767629)/(o+1.)+(-0.6802861583)
        ref[4,1,1]=(1.23997013471)/(o+1.)+(0.70527600063)
        ref[5,0,0]=(-0.383142890439)/(o+1.)+(0.312083322186)
        ref[5,0,1]=(1.64380369652)/(o+1.)+(-0.390348387013)
        ref[5,1,0]=(-0.812713549757)/(o+1.)+(0.520467462905)
        ref[5,1,1]=(0.600505313178)/(o+1.)+(-0.287048092454)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.235888975732)*x[0]+(-0.734372259615)*x[1]
        s[0,0,0,1]=(0.0665724430605)*x[0]+(-0.0890002398629)*x[1]
        s[0,0,1,0]=(-1.19774759676)*x[0]+(1.00321581155)*x[1]
        s[0,0,1,1]=(0.93370004223)*x[0]+(-0.953835581453)*x[1]
        s[0,0,2,0]=(-0.0557204731966)*x[0]+(0.94945217411)*x[1]
        s[0,0,2,1]=(-0.772628465097)*x[0]+(1.77100579145)*x[1]
        s[0,1,0,0]=(-0.405653863463)*x[0]+(0.494482129094)*x[1]
        s[0,1,0,1]=(0.636236763209)*x[0]+(-1.63254203234)*x[1]
        s[0,1,1,0]=(0.941764211044)*x[0]+(0.699520209508)*x[1]
        s[0,1,1,1]=(1.05045643599)*x[0]+(0.0596636815695)*x[1]
        s[0,1,2,0]=(0.578668042684)*x[0]+(0.285442340394)*x[1]
        s[0,1,2,1]=(0.469149909671)*x[0]+(1.26526906754)*x[1]
        s[0,2,0,0]=(1.05435496407)*x[0]+(-0.592645628655)*x[1]
        s[0,2,0,1]=(0.066181979241)*x[0]+(-1.42368311585)*x[1]
        s[0,2,1,0]=(-0.896588999422)*x[0]+(1.44881386304)*x[1]
        s[0,2,1,1]=(-0.862544854709)*x[0]+(-0.398360134784)*x[1]
        s[0,2,2,0]=(1.02969811319)*x[0]+(0.318993808627)*x[1]
        s[0,2,2,1]=(0.0143773859487)*x[0]+(-0.327118788648)*x[1]
        s[0,3,0,0]=(0.021196438108)*x[0]+(1.8515159226)*x[1]
        s[0,3,0,1]=(0.627726779803)*x[0]+(-0.302716383303)*x[1]
        s[0,3,1,0]=(0.165624370858)*x[0]+(0.274919760437)*x[1]
        s[0,3,1,1]=(0.283214725106)*x[0]+(-0.872378951063)*x[1]
        s[0,3,2,0]=(-0.83391985409)*x[0]+(0.731217556829)*x[1]
        s[0,3,2,1]=(0.918833558754)*x[0]+(0.714715370412)*x[1]
        s[0,4,0,0]=(-0.503211677055)*x[0]+(0.0289310095577)*x[1]
        s[0,4,0,1]=(0.597129238132)*x[0]+(0.540475624893)*x[1]
        s[0,4,1,0]=(0.687211570349)*x[0]+(0.594018658674)*x[1]
        s[0,4,1,1]=(1.5970484642)*x[0]+(-0.495319600999)*x[1]
        s[0,4,2,0]=(1.20627016067)*x[0]+(0.461261823726)*x[1]
        s[0,4,2,1]=(0.165927055028)*x[0]+(1.09007328428)*x[1]
        s[1,0,0,0]=(1.16105917598)*x[0]+(1.82590164801)*x[1]
        s[1,0,0,1]=(-1.13479961391)*x[0]+(0.0351464050353)*x[1]
        s[1,0,1,0]=(-0.318954647255)*x[0]+(0.210946813363)*x[1]
        s[1,0,1,1]=(1.31393538581)*x[0]+(0.198463680367)*x[1]
        s[1,0,2,0]=(1.17799558613)*x[0]+(0.514357347746)*x[1]
        s[1,0,2,1]=(0.571191341129)*x[0]+(0.694971250457)*x[1]
        s[1,1,0,0]=(0.132901410189)*x[0]+(-0.262070778236)*x[1]
        s[1,1,0,1]=(-0.871431837436)*x[0]+(-0.514707012778)*x[1]
        s[1,1,1,0]=(-0.319003186732)*x[0]+(-0.014973803902)*x[1]
        s[1,1,1,1]=(0.260213741998)*x[0]+(1.32908581921)*x[1]
        s[1,1,2,0]=(0.811339404893)*x[0]+(-0.304564391555)*x[1]
        s[1,1,2,1]=(-0.062193223101)*x[0]+(-0.911704705741)*x[1]
        s[1,2,0,0]=(0.0932623640149)*x[0]+(1.12570360092)*x[1]
        s[1,2,0,1]=(1.16504942277)*x[0]+(-0.385840092559)*x[1]
        s[1,2,1,0]=(1.3440680469)*x[0]+(0.532696669818)*x[1]
        s[1,2,1,1]=(1.31050128725)*x[0]+(0.00906728815983)*x[1]
        s[1,2,2,0]=(-1.36931832474)*x[0]+(-1.70562158659)*x[1]
        s[1,2,2,1]=(-1.8375028036)*x[0]+(1.52366054242)*x[1]
        s[1,3,0,0]=(0.474023303611)*x[0]+(0.437518737767)*x[1]
        s[1,3,0,1]=(-0.42921544272)*x[0]+(0.387647202618)*x[1]
        s[1,3,1,0]=(0.863045413096)*x[0]+(-0.36260979035)*x[1]
        s[1,3,1,1]=(-0.654365401799)*x[0]+(0.533926246461)*x[1]
        s[1,3,2,0]=(0.418152564814)*x[0]+(1.11544647745)*x[1]
        s[1,3,2,1]=(-1.60947873228)*x[0]+(0.591727770738)*x[1]
        s[1,4,0,0]=(-1.51807668572)*x[0]+(-1.01613349614)*x[1]
        s[1,4,0,1]=(-1.50163863727)*x[0]+(1.50376272164)*x[1]
        s[1,4,1,0]=(0.320776223369)*x[0]+(-0.700391339598)*x[1]
        s[1,4,1,1]=(0.745224476612)*x[0]+(-0.850870606209)*x[1]
        s[1,4,2,0]=(-0.179598404379)*x[0]+(1.75799776411)*x[1]
        s[1,4,2,1]=(0.00909330900106)*x[0]+(-0.077399205942)*x[1]
        s[2,0,0,0]=(0.505719447396)*x[0]+(-0.275603298829)*x[1]
        s[2,0,0,1]=(1.46787325811)*x[0]+(0.534793377012)*x[1]
        s[2,0,1,0]=(0.77645891983)*x[0]+(0.540540443218)*x[1]
        s[2,0,1,1]=(0.429661881583)*x[0]+(1.43421311479)*x[1]
        s[2,0,2,0]=(1.02806543045)*x[0]+(0.931834883879)*x[1]
        s[2,0,2,1]=(0.763031666632)*x[0]+(0.873653388204)*x[1]
        s[2,1,0,0]=(0.261772733655)*x[0]+(-1.18679909814)*x[1]
        s[2,1,0,1]=(-1.61662908583)*x[0]+(1.10646142209)*x[1]
        s[2,1,1,0]=(0.188527773811)*x[0]+(-0.772410233645)*x[1]
        s[2,1,1,1]=(-0.869931105194)*x[0]+(1.51014899284)*x[1]
        s[2,1,2,0]=(0.252478837689)*x[0]+(1.22480134325)*x[1]
        s[2,1,2,1]=(1.49578867187)*x[0]+(-0.877973056373)*x[1]
        s[2,2,0,0]=(-0.209568005598)*x[0]+(-0.0462992389671)*x[1]
        s[2,2,0,1]=(-0.298130269216)*x[0]+(-0.908352713549)*x[1]
        s[2,2,1,0]=(-0.586370615835)*x[0]+(-1.72704626602)*x[1]
        s[2,2,1,1]=(1.26217059553)*x[0]+(0.676548549908)*x[1]
        s[2,2,2,0]=(0.961200845015)*x[0]+(0.612151157303)*x[1]
        s[2,2,2,1]=(0.700873758828)*x[0]+(0.800662765762)*x[1]
        s[2,3,0,0]=(-0.319747641322)*x[0]+(0.689929330298)*x[1]
        s[2,3,0,1]=(-1.04393759081)*x[0]+(-1.80449230281)*x[1]
        s[2,3,1,0]=(-0.0410072962318)*x[0]+(-0.678000289402)*x[1]
        s[2,3,1,1]=(-0.329381648117)*x[0]+(-0.935487069857)*x[1]
        s[2,3,2,0]=(-0.907460976718)*x[0]+(0.0314212805943)*x[1]
        s[2,3,2,1]=(0.593604907619)*x[0]+(-0.479146057787)*x[1]
        s[2,4,0,0]=(-0.0337048332079)*x[0]+(0.239721443625)*x[1]
        s[2,4,0,1]=(0.721609580339)*x[0]+(-1.06464423044)*x[1]
        s[2,4,1,0]=(0.670568080353)*x[0]+(1.28774286264)*x[1]
        s[2,4,1,1]=(0.603608338555)*x[0]+(0.323996221169)*x[1]
        s[2,4,2,0]=(0.0371268677698)*x[0]+(0.401814556385)*x[1]
        s[2,4,2,1]=(0.436926475796)*x[0]+(-0.819433312575)*x[1]
        s[3,0,0,0]=(-0.0190186791033)*x[0]+(0.501672647061)*x[1]
        s[3,0,0,1]=(0.00060856018325)*x[0]+(-0.414668792931)*x[1]
        s[3,0,1,0]=(0.699428985955)*x[0]+(0.612631528371)*x[1]
        s[3,0,1,1]=(-0.659371576463)*x[0]+(-0.0982545515259)*x[1]
        s[3,0,2,0]=(0.444351889183)*x[0]+(0.53998411713)*x[1]
        s[3,0,2,1]=(-0.281563709018)*x[0]+(-0.486501437256)*x[1]
        s[3,1,0,0]=(0.407521411466)*x[0]+(-0.749870044816)*x[1]
        s[3,1,0,1]=(-0.638780264813)*x[0]+(0.230397511262)*x[1]
        s[3,1,1,0]=(1.16000440204)*x[0]+(0.13074916095)*x[1]
        s[3,1,1,1]=(-1.30550439046)*x[0]+(-1.33788704039)*x[1]
        s[3,1,2,0]=(-0.540767742404)*x[0]+(-0.148316212185)*x[1]
        s[3,1,2,1]=(1.70452499495)*x[0]+(0.935421717463)*x[1]
        s[3,2,0,0]=(0.320990904519)*x[0]+(0.467953609756)*x[1]
        s[3,2,0,1]=(-0.26002228203)*x[0]+(-0.925765706976)*x[1]
        s[3,2,1,0]=(0.925754935382)*x[0]+(0.240432086422)*x[1]
        s[3,2,1,1]=(0.0214680080146)*x[0]+(-1.11507977191)*x[1]
        s[3,2,2,0]=(-0.187946153679)*x[0]+(0.024781322847)*x[1]
        s[3,2,2,1]=(-1.41866510495)*x[0]+(0.596173034069)*x[1]
        s[3,3,0,0]=(0.77087649562)*x[0]+(0.848168811103)*x[1]
        s[3,3,0,1]=(0.218810390517)*x[0]+(0.116909485563)*x[1]
        s[3,3,1,0]=(-0.889981738645)*x[0]+(-0.506355969844)*x[1]
        s[3,3,1,1]=(1.15870903185)*x[0]+(-0.8158119166)*x[1]
        s[3,3,2,0]=(-0.345230989856)*x[0]+(-0.138496974465)*x[1]
        s[3,3,2,1]=(-0.819830787817)*x[0]+(-1.66747998417)*x[1]
        s[3,4,0,0]=(-1.74647564112)*x[0]+(0.22309219172)*x[1]
        s[3,4,0,1]=(0.146804717412)*x[0]+(0.493084518827)*x[1]
        s[3,4,1,0]=(-0.536943383893)*x[0]+(-0.686288132381)*x[1]
        s[3,4,1,1]=(-0.142983666789)*x[0]+(-0.183927597857)*x[1]
        s[3,4,2,0]=(-0.173071632973)*x[0]+(-0.696870700802)*x[1]
        s[3,4,2,1]=(-0.359163238166)*x[0]+(-0.172033035327)*x[1]
        ref[0,0,0,0]=(-0.603598770198)/(o+1.)+(0.0525577431575)
        ref[0,0,0,1]=(1.04505857398)/(o+1.)+(-0.533743185393)
        ref[0,0,1,0]=(0.00772071573504)/(o+1.)+(-0.101126250473)
        ref[0,0,1,1]=(-0.290587702404)/(o+1.)+(0.135226081591)
        ref[0,0,2,0]=(0.614426628234)/(o+1.)+(0.139652536339)
        ref[0,0,2,1]=(-0.0228094425795)/(o+1.)+(0.510593384465)
        ref[0,1,0,0]=(-0.436885078328)/(o+1.)+(0.262856671979)
        ref[0,1,0,1]=(0.0510259667548)/(o+1.)+(-0.523665617945)
        ref[0,1,1,0]=(0.744462079536)/(o+1.)+(0.448411170508)
        ref[0,1,1,1]=(1.52720962375)/(o+1.)+(-0.208544753096)
        ref[0,1,2,0]=(0.73405306307)/(o+1.)+(0.0650286600044)
        ref[0,1,2,1]=(0.550924984933)/(o+1.)+(0.591746996137)
        ref[0,2,0,0]=(-0.347797813257)/(o+1.)+(0.404753574338)
        ref[0,2,0,1]=(0.187043042708)/(o+1.)+(-0.772272089659)
        ref[0,2,1,0]=(-0.316098850787)/(o+1.)+(0.434161857202)
        ref[0,2,1,1]=(-0.536124734534)/(o+1.)+(-0.36239012748)
        ref[0,2,2,0]=(0.585045523113)/(o+1.)+(0.381823199352)
        ref[0,2,2,1]=(0.422243751365)/(o+1.)+(-0.367492577032)
        ref[0,3,0,0]=(0.91381158808)/(o+1.)+(0.479450386313)
        ref[0,3,0,1]=(0.209840671177)/(o+1.)+(0.0575848626616)
        ref[0,3,1,0]=(0.158639778589)/(o+1.)+(0.140952176353)
        ref[0,3,1,1]=(-0.364808895601)/(o+1.)+(-0.112177665178)
        ref[0,3,2,0]=(-0.198937574741)/(o+1.)+(0.0481176387404)
        ref[0,3,2,1]=(1.05420599501)/(o+1.)+(0.289671467078)
        ref[0,4,0,0]=(0.811701728881)/(o+1.)+(-0.642991198189)
        ref[0,4,0,1]=(0.693754647803)/(o+1.)+(0.221925107611)
        ref[0,4,1,0]=(1.05453609011)/(o+1.)+(0.113347069457)
        ref[0,4,1,1]=(0.117274271978)/(o+1.)+(0.49222729561)
        ref[0,4,2,0]=(0.496795563283)/(o+1.)+(0.585368210556)
        ref[0,4,2,1]=(0.83272690983)/(o+1.)+(0.211636714741)
        ref[1,0,0,0]=(1.46878592273)/(o+1.)+(0.75908745063)
        ref[1,0,0,1]=(-0.250785630529)/(o+1.)+(-0.424433789171)
        ref[1,0,1,0]=(0.161786720779)/(o+1.)+(-0.134897277336)
        ref[1,0,1,1]=(1.12719401256)/(o+1.)+(0.192602526813)
        ref[1,0,2,0]=(0.7197235394)/(o+1.)+(0.486314697239)
        ref[1,0,2,1]=(1.02804458464)/(o+1.)+(0.119059003471)
        ref[1,1,0,0]=(0.0361584520541)/(o+1.)+(-0.0826639100507)
        ref[1,1,0,1]=(-1.06091812324)/(o+1.)+(-0.162610363489)
        ref[1,1,1,0]=(1.13776585837)/(o+1.)+(-0.735871424501)
        ref[1,1,1,1]=(1.4740891177)/(o+1.)+(0.0576052217538)
        ref[1,1,2,0]=(0.280616623834)/(o+1.)+(0.113079194752)
        ref[1,1,2,1]=(-0.491890699653)/(o+1.)+(-0.241003614595)
        ref[1,2,0,0]=(-0.0342838844534)/(o+1.)+(0.626624924695)
        ref[1,2,0,1]=(0.770911868081)/(o+1.)+(0.00414873106652)
        ref[1,2,1,0]=(0.30749084682)/(o+1.)+(0.784636934951)
        ref[1,2,1,1]=(1.8638645006)/(o+1.)+(-0.272147962595)
        ref[1,2,2,0]=(-1.46841213372)/(o+1.)+(-0.803263888803)
        ref[1,2,2,1]=(-0.302454198157)/(o+1.)+(-0.00569403150815)
        ref[1,3,0,0]=(0.493917522063)/(o+1.)+(0.208812259657)
        ref[1,3,0,1]=(0.476274657976)/(o+1.)+(-0.258921449039)
        ref[1,3,1,0]=(-0.736248174893)/(o+1.)+(0.618341898819)
        ref[1,3,1,1]=(-0.697129670691)/(o+1.)+(0.288345257677)
        ref[1,3,2,0]=(0.303153912627)/(o+1.)+(0.615222564816)
        ref[1,3,2,1]=(-0.58122630211)/(o+1.)+(-0.218262329715)
        ref[1,4,0,0]=(-1.15442442968)/(o+1.)+(-0.689892876091)
        ref[1,4,0,1]=(0.0585826445649)/(o+1.)+(-0.0282292800982)
        ref[1,4,1,0]=(0.3252337262)/(o+1.)+(-0.352424421214)
        ref[1,4,1,1]=(0.512701244088)/(o+1.)+(-0.309173686843)
        ref[1,4,2,0]=(1.5506882457)/(o+1.)+(0.0138555570174)
        ref[1,4,2,1]=(-0.940773035102)/(o+1.)+(0.436233569081)
        ref[2,0,0,0]=(1.54247754874)/(o+1.)+(-0.656180700088)
        ref[2,0,0,1]=(0.988306961756)/(o+1.)+(0.507179836684)
        ref[2,0,1,0]=(-0.416628824296)/(o+1.)+(0.866814093672)
        ref[2,0,1,1]=(1.41006066195)/(o+1.)+(0.22690716721)
        ref[2,0,2,0]=(0.353011705537)/(o+1.)+(0.803444304396)
        ref[2,0,2,1]=(0.687642260874)/(o+1.)+(0.474521396981)
        ref[2,1,0,0]=(-0.218198987148)/(o+1.)+(-0.353413688668)
        ref[2,1,0,1]=(-0.635306433562)/(o+1.)+(0.0625693849126)
        ref[2,1,1,0]=(0.53416867099)/(o+1.)+(-0.559025565412)
        ref[2,1,1,1]=(0.00543686305333)/(o+1.)+(0.317390512297)
        ref[2,1,2,0]=(0.916573120847)/(o+1.)+(0.280353530048)
        ref[2,1,2,1]=(-0.0537588152546)/(o+1.)+(0.335787215373)
        ref[2,2,0,0]=(1.0055660817)/(o+1.)+(-0.630716663133)
        ref[2,2,0,1]=(0.420474306577)/(o+1.)+(-0.813478644671)
        ref[2,2,1,0]=(-0.723008657238)/(o+1.)+(-0.795204112308)
        ref[2,2,1,1]=(0.306188174034)/(o+1.)+(0.816265485704)
        ref[2,2,2,0]=(0.806158569368)/(o+1.)+(0.383596716475)
        ref[2,2,2,1]=(1.6353377865)/(o+1.)+(-0.066900630957)
        ref[2,3,0,0]=(-1.24813850815)/(o+1.)+(0.809160098565)
        ref[2,3,0,1]=(-1.56942041127)/(o+1.)+(-0.639504741178)
        ref[2,3,1,0]=(-1.49605878837)/(o+1.)+(0.388525601366)
        ref[2,3,1,1]=(-0.448403645878)/(o+1.)+(-0.408232536048)
        ref[2,3,2,0]=(-1.13568794342)/(o+1.)+(0.129824123646)
        ref[2,3,2,1]=(0.596733666259)/(o+1.)+(-0.241137408214)
        ref[2,4,0,0]=(0.0191094707617)/(o+1.)+(0.0934535698276)
        ref[2,4,0,1]=(0.185204097773)/(o+1.)+(-0.264119373935)
        ref[2,4,1,0]=(1.55855563639)/(o+1.)+(0.199877653305)
        ref[2,4,1,1]=(0.11469142184)/(o+1.)+(0.406456568942)
        ref[2,4,2,0]=(-0.954321542363)/(o+1.)+(0.696631483259)
        ref[2,4,2,1]=(-0.810227566168)/(o+1.)+(0.213860364695)
        ref[3,0,0,0]=(0.453436736881)/(o+1.)+(0.0146086155385)
        ref[3,0,0,1]=(-0.210824760654)/(o+1.)+(-0.101617736047)
        ref[3,0,1,0]=(0.414458928012)/(o+1.)+(0.448800793157)
        ref[3,0,1,1]=(0.346129385596)/(o+1.)+(-0.551877756793)
        ref[3,0,2,0]=(1.85496275992)/(o+1.)+(-0.435313376805)
        ref[3,0,2,1]=(-0.110500317363)/(o+1.)+(-0.328782414455)
        ref[3,1,0,0]=(0.0399449935267)/(o+1.)+(-0.191146813438)
        ref[3,1,0,1]=(0.133080949177)/(o+1.)+(-0.270731851364)
        ref[3,1,1,0]=(0.139648356416)/(o+1.)+(0.575552603289)
        ref[3,1,1,1]=(-1.45813597924)/(o+1.)+(-0.592627725804)
        ref[3,1,2,0]=(-0.773854684679)/(o+1.)+(0.0423853650452)
        ref[3,1,2,1]=(1.2726274277)/(o+1.)+(0.683659642354)
        ref[3,2,0,0]=(0.555324507349)/(o+1.)+(0.116810003463)
        ref[3,2,0,1]=(-0.310977730161)/(o+1.)+(-0.437405129422)
        ref[3,2,1,0]=(1.05800939782)/(o+1.)+(0.0540888119933)
        ref[3,2,1,1]=(0.386222714751)/(o+1.)+(-0.739917239322)
        ref[3,2,2,0]=(-0.95705464593)/(o+1.)+(0.396944907549)
        ref[3,2,2,1]=(-0.491612682819)/(o+1.)+(-0.165439694033)
        ref[3,3,0,0]=(0.666972772584)/(o+1.)+(0.47603626707)
        ref[3,3,0,1]=(0.268506355312)/(o+1.)+(0.0336067603842)
        ref[3,3,1,0]=(-0.57724956178)/(o+1.)+(-0.409544073355)
        ref[3,3,1,1]=(-0.518808765862)/(o+1.)+(0.430852940555)
        ref[3,3,2,0]=(-1.95248352029)/(o+1.)+(0.734377777986)
        ref[3,3,2,1]=(-1.14565232021)/(o+1.)+(-0.67082922589)
        ref[3,4,0,0]=(-1.13455573259)/(o+1.)+(-0.1944138584)
        ref[3,4,0,1]=(-0.508914286468)/(o+1.)+(0.574401761354)
        ref[3,4,1,0]=(-0.319104038837)/(o+1.)+(-0.452063738718)
        ref[3,4,1,1]=(-0.100356733327)/(o+1.)+(-0.11327726566)
        ref[3,4,2,0]=(-0.39970881352)/(o+1.)+(-0.235116760127)
        ref[3,4,2,1]=(-0.339631070901)/(o+1.)+(-0.0957826012957)
      else:
        s[0,0,0,0]=(0.0794776522341)*x[0]+(1.69294456255)*x[1]+(0.824115844979)*x[2]
        s[0,0,0,1]=(0.0720523378919)*x[0]+(-1.39724154561)*x[1]+(-0.010988810956)*x[2]
        s[0,0,1,0]=(0.985556321125)*x[0]+(1.12133844031)*x[1]+(-0.262491219522)*x[2]
        s[0,0,1,1]=(0.789643681767)*x[0]+(1.02860624013)*x[1]+(-0.450770685703)*x[2]
        s[0,0,2,0]=(-0.850427026537)*x[0]+(-0.393993518297)*x[1]+(0.12816296306)*x[2]
        s[0,0,2,1]=(-0.47508130541)*x[0]+(-0.393857510976)*x[1]+(0.66003641476)*x[2]
        s[0,1,0,0]=(-1.87725252884)*x[0]+(-0.668075530728)*x[1]+(0.938485679482)*x[2]
        s[0,1,0,1]=(0.820537136967)*x[0]+(-0.453620529157)*x[1]+(0.811944595883)*x[2]
        s[0,1,1,0]=(0.914995161971)*x[0]+(-0.196153821444)*x[1]+(0.938555859255)*x[2]
        s[0,1,1,1]=(-0.866633718493)*x[0]+(-0.534596859321)*x[1]+(0.77514383239)*x[2]
        s[0,1,2,0]=(0.183094085342)*x[0]+(0.727499336912)*x[1]+(0.170986687969)*x[2]
        s[0,1,2,1]=(-1.36220673904)*x[0]+(0.0822122492483)*x[1]+(0.124227107716)*x[2]
        s[0,2,0,0]=(1.55269479425)*x[0]+(-0.14574857122)*x[1]+(1.24120289667)*x[2]
        s[0,2,0,1]=(-0.464038186768)*x[0]+(0.00554383028356)*x[1]+(-1.14838250985)*x[2]
        s[0,2,1,0]=(0.904608031363)*x[0]+(-0.313004420619)*x[1]+(-0.10529298073)*x[2]
        s[0,2,1,1]=(0.156623578469)*x[0]+(0.55127494399)*x[1]+(-0.385447926692)*x[2]
        s[0,2,2,0]=(-0.766804018933)*x[0]+(0.18519619356)*x[1]+(1.00915540081)*x[2]
        s[0,2,2,1]=(0.00231157548456)*x[0]+(0.338768642838)*x[1]+(-0.684228602438)*x[2]
        s[0,3,0,0]=(-0.834810486599)*x[0]+(-0.096360306458)*x[1]+(1.28924889761)*x[2]
        s[0,3,0,1]=(-0.658443843543)*x[0]+(0.889474267886)*x[1]+(0.566763124725)*x[2]
        s[0,3,1,0]=(-0.109777093963)*x[0]+(0.228053887631)*x[1]+(1.22952539906)*x[2]
        s[0,3,1,1]=(-0.749479260801)*x[0]+(-1.01102432725)*x[1]+(-0.196861989291)*x[2]
        s[0,3,2,0]=(0.650389523509)*x[0]+(0.151701374527)*x[1]+(0.506246388483)*x[2]
        s[0,3,2,1]=(-0.648510958338)*x[0]+(-0.263590202304)*x[1]+(-1.35992044414)*x[2]
        s[0,4,0,0]=(0.196976454523)*x[0]+(-0.193947226599)*x[1]+(1.67641586045)*x[2]
        s[0,4,0,1]=(-0.455868626908)*x[0]+(0.506302321882)*x[1]+(1.1666778393)*x[2]
        s[0,4,1,0]=(-0.788545609857)*x[0]+(-0.487332926053)*x[1]+(-0.0645141715605)*x[2]
        s[0,4,1,1]=(-0.320424709093)*x[0]+(-0.683301403769)*x[1]+(-0.222245108268)*x[2]
        s[0,4,2,0]=(0.876149348517)*x[0]+(0.12449602802)*x[1]+(-0.220531205966)*x[2]
        s[0,4,2,1]=(-0.2764172159)*x[0]+(-0.894571003479)*x[1]+(-0.595830070177)*x[2]
        s[1,0,0,0]=(-0.79676665323)*x[0]+(-1.25965739347)*x[1]+(-0.319474113915)*x[2]
        s[1,0,0,1]=(1.601348448)*x[0]+(0.41027863439)*x[1]+(-0.346758807915)*x[2]
        s[1,0,1,0]=(0.181292913253)*x[0]+(1.10494312205)*x[1]+(0.196973772813)*x[2]
        s[1,0,1,1]=(0.562861818795)*x[0]+(-0.997990211102)*x[1]+(0.0332706615838)*x[2]
        s[1,0,2,0]=(-0.989984164479)*x[0]+(-0.886186583204)*x[1]+(-0.228620900973)*x[2]
        s[1,0,2,1]=(0.841365917044)*x[0]+(-0.0519066039156)*x[1]+(-0.81126199301)*x[2]
        s[1,1,0,0]=(-0.613125126303)*x[0]+(0.482770874125)*x[1]+(0.038374960382)*x[2]
        s[1,1,0,1]=(-0.203719663632)*x[0]+(-0.281873286682)*x[1]+(-0.0303453795347)*x[2]
        s[1,1,1,0]=(1.11287699536)*x[0]+(-1.04759296757)*x[1]+(1.64257695374)*x[2]
        s[1,1,1,1]=(-0.0861307607908)*x[0]+(0.261314888791)*x[1]+(-0.204955973137)*x[2]
        s[1,1,2,0]=(0.821662110998)*x[0]+(-1.05649310988)*x[1]+(0.609822823353)*x[2]
        s[1,1,2,1]=(-0.730526005063)*x[0]+(1.62018363097)*x[1]+(0.99111776492)*x[2]
        s[1,2,0,0]=(-1.25039777751)*x[0]+(0.811070034681)*x[1]+(1.32505446058)*x[2]
        s[1,2,0,1]=(0.753110041895)*x[0]+(0.186756235207)*x[1]+(-1.30430176967)*x[2]
        s[1,2,1,0]=(0.146182105242)*x[0]+(-0.863831927779)*x[1]+(-1.1030247274)*x[2]
        s[1,2,1,1]=(0.674634709918)*x[0]+(-0.560866548824)*x[1]+(0.576408322189)*x[2]
        s[1,2,2,0]=(-0.573489753302)*x[0]+(1.57147013427)*x[1]+(0.711691292064)*x[2]
        s[1,2,2,1]=(0.549483527739)*x[0]+(-1.14179814973)*x[1]+(1.60524619335)*x[2]
        s[1,3,0,0]=(0.197453494754)*x[0]+(-0.456559966113)*x[1]+(-1.21423667747)*x[2]
        s[1,3,0,1]=(-1.76312274691)*x[0]+(-1.04778192886)*x[1]+(-0.185488490674)*x[2]
        s[1,3,1,0]=(0.0162460011376)*x[0]+(1.16407944993)*x[1]+(1.16424268292)*x[2]
        s[1,3,1,1]=(-0.513913288204)*x[0]+(-1.46536589962)*x[1]+(-0.906205233261)*x[2]
        s[1,3,2,0]=(-0.0157400074256)*x[0]+(-0.293812243081)*x[1]+(0.815280978304)*x[2]
        s[1,3,2,1]=(1.57630585875)*x[0]+(0.290739280065)*x[1]+(-0.771398356417)*x[2]
        s[1,4,0,0]=(-0.323882779703)*x[0]+(-0.763929395312)*x[1]+(0.738403272406)*x[2]
        s[1,4,0,1]=(-0.476278532043)*x[0]+(-0.11039317349)*x[1]+(0.346684885005)*x[2]
        s[1,4,1,0]=(-1.36304063164)*x[0]+(-0.660619165296)*x[1]+(-0.903566907206)*x[2]
        s[1,4,1,1]=(0.0863780283437)*x[0]+(0.466792514558)*x[1]+(0.448776109878)*x[2]
        s[1,4,2,0]=(0.74478619068)*x[0]+(0.446314847678)*x[1]+(-0.0481741032015)*x[2]
        s[1,4,2,1]=(0.746506715398)*x[0]+(-0.432201202785)*x[1]+(0.203440105689)*x[2]
        s[2,0,0,0]=(-0.785805880678)*x[0]+(-0.862407720473)*x[1]+(-1.21640925829)*x[2]
        s[2,0,0,1]=(0.647190817984)*x[0]+(0.519851314653)*x[1]+(0.390946186931)*x[2]
        s[2,0,1,0]=(-1.24831725337)*x[0]+(1.59413938467)*x[1]+(-1.33358250136)*x[2]
        s[2,0,1,1]=(-0.152213072763)*x[0]+(-1.09357954369)*x[1]+(-1.25586196425)*x[2]
        s[2,0,2,0]=(-0.0314861108304)*x[0]+(0.293907529511)*x[1]+(-0.21039874768)*x[2]
        s[2,0,2,1]=(0.0467194832302)*x[0]+(0.527286460462)*x[1]+(-0.415618293894)*x[2]
        s[2,1,0,0]=(-0.189575051656)*x[0]+(-0.844885847783)*x[1]+(-0.334235875002)*x[2]
        s[2,1,0,1]=(-0.00822711174783)*x[0]+(-1.5897272285)*x[1]+(1.50002681163)*x[2]
        s[2,1,1,0]=(-0.841981498104)*x[0]+(-1.10376539971)*x[1]+(0.311385102139)*x[2]
        s[2,1,1,1]=(1.12897934543)*x[0]+(-0.966261308872)*x[1]+(-1.16359871767)*x[2]
        s[2,1,2,0]=(-0.957189463062)*x[0]+(0.0471569985184)*x[1]+(0.12392515517)*x[2]
        s[2,1,2,1]=(-0.752369087888)*x[0]+(1.18176351896)*x[1]+(0.334788137458)*x[2]
        s[2,2,0,0]=(1.22014205343)*x[0]+(0.784413576281)*x[1]+(0.0962245348039)*x[2]
        s[2,2,0,1]=(-0.407023035962)*x[0]+(1.11597107325)*x[1]+(0.458612291763)*x[2]
        s[2,2,1,0]=(-0.0496175828387)*x[0]+(-0.0536216340718)*x[1]+(0.771683970464)*x[2]
        s[2,2,1,1]=(-0.356958793645)*x[0]+(-1.38580705404)*x[1]+(-0.670167616875)*x[2]
        s[2,2,2,0]=(0.0804001816041)*x[0]+(-1.16114869406)*x[1]+(0.475472363874)*x[2]
        s[2,2,2,1]=(0.98238551642)*x[0]+(-1.12578358118)*x[1]+(-0.00970359156434)*x[2]
        s[2,3,0,0]=(1.0110556058)*x[0]+(-0.488112488916)*x[1]+(0.243005554591)*x[2]
        s[2,3,0,1]=(-0.16668051279)*x[0]+(0.178067528024)*x[1]+(0.969038052731)*x[2]
        s[2,3,1,0]=(-0.00345650882131)*x[0]+(-0.941671547756)*x[1]+(-0.49908897105)*x[2]
        s[2,3,1,1]=(0.716683430797)*x[0]+(-0.585708024289)*x[1]+(-1.19073441986)*x[2]
        s[2,3,2,0]=(0.400515369791)*x[0]+(-0.454556618814)*x[1]+(0.0782324352019)*x[2]
        s[2,3,2,1]=(-0.164628093414)*x[0]+(-1.26464027792)*x[1]+(0.179511210919)*x[2]
        s[2,4,0,0]=(1.33277550072)*x[0]+(1.18570575409)*x[1]+(0.381346013717)*x[2]
        s[2,4,0,1]=(0.588290284687)*x[0]+(-1.08302445047)*x[1]+(0.166898904485)*x[2]
        s[2,4,1,0]=(0.131178111015)*x[0]+(-0.749855470189)*x[1]+(-0.1247474719)*x[2]
        s[2,4,1,1]=(-0.97245336122)*x[0]+(0.873932906896)*x[1]+(0.120264062027)*x[2]
        s[2,4,2,0]=(-0.722531128456)*x[0]+(0.19636289816)*x[1]+(-0.907945347107)*x[2]
        s[2,4,2,1]=(-0.711111750197)*x[0]+(0.31947094036)*x[1]+(-1.79333697163)*x[2]
        s[3,0,0,0]=(-0.184519023765)*x[0]+(0.680221159423)*x[1]+(-0.0244123658903)*x[2]
        s[3,0,0,1]=(-1.14573637546)*x[0]+(0.0616977700293)*x[1]+(-0.654808111844)*x[2]
        s[3,0,1,0]=(-1.09297543018)*x[0]+(-1.04184353717)*x[1]+(0.384120100538)*x[2]
        s[3,0,1,1]=(0.105247846819)*x[0]+(-1.31600150124)*x[1]+(1.43496073616)*x[2]
        s[3,0,2,0]=(0.881148598911)*x[0]+(-0.478723216871)*x[1]+(-0.0746344543132)*x[2]
        s[3,0,2,1]=(-1.20828681587)*x[0]+(-1.65808839437)*x[1]+(-0.636678098535)*x[2]
        s[3,1,0,0]=(-0.213394309668)*x[0]+(-0.240354748844)*x[1]+(0.117282970374)*x[2]
        s[3,1,0,1]=(1.37462845901)*x[0]+(-0.41583894581)*x[1]+(-0.625626047701)*x[2]
        s[3,1,1,0]=(1.83991491983)*x[0]+(1.2778674571)*x[1]+(1.35465906139)*x[2]
        s[3,1,1,1]=(1.58168546715)*x[0]+(0.829809996131)*x[1]+(0.681232260732)*x[2]
        s[3,1,2,0]=(-0.0212907664121)*x[0]+(1.00109823691)*x[1]+(-1.30845432825)*x[2]
        s[3,1,2,1]=(0.169920360674)*x[0]+(-0.864995850094)*x[1]+(0.302979947072)*x[2]
        s[3,2,0,0]=(-0.154084636739)*x[0]+(0.319647474298)*x[1]+(-1.23506947259)*x[2]
        s[3,2,0,1]=(-0.199216213818)*x[0]+(1.28921202953)*x[1]+(1.65805122541)*x[2]
        s[3,2,1,0]=(-1.86774194364)*x[0]+(-0.101723263166)*x[1]+(-0.526029332636)*x[2]
        s[3,2,1,1]=(-0.177960964857)*x[0]+(-0.899073564225)*x[1]+(0.551378002191)*x[2]
        s[3,2,2,0]=(0.18491319197)*x[0]+(-0.286906366286)*x[1]+(0.583859098776)*x[2]
        s[3,2,2,1]=(0.38623837272)*x[0]+(-0.653272502002)*x[1]+(-1.46718517275)*x[2]
        s[3,3,0,0]=(0.0494572779834)*x[0]+(-0.0566099666381)*x[1]+(-0.113874384717)*x[2]
        s[3,3,0,1]=(0.0241152098269)*x[0]+(-0.317061594612)*x[1]+(1.23769551101)*x[2]
        s[3,3,1,0]=(-0.107294616786)*x[0]+(-0.651410283508)*x[1]+(0.833104181804)*x[2]
        s[3,3,1,1]=(0.16676828757)*x[0]+(-0.670496617303)*x[1]+(1.15686719798)*x[2]
        s[3,3,2,0]=(0.159950143242)*x[0]+(0.0436823700228)*x[1]+(-0.645027680285)*x[2]
        s[3,3,2,1]=(-1.2697347822)*x[0]+(1.22938250997)*x[1]+(0.208587928792)*x[2]
        s[3,4,0,0]=(0.575105413412)*x[0]+(-1.44998456404)*x[1]+(-0.516125694177)*x[2]
        s[3,4,0,1]=(0.0387618783003)*x[0]+(0.859152861676)*x[1]+(1.29847202073)*x[2]
        s[3,4,1,0]=(-0.842510438531)*x[0]+(0.191878858518)*x[1]+(-0.73453059399)*x[2]
        s[3,4,1,1]=(1.41435810392)*x[0]+(-0.855734345577)*x[1]+(-0.425917153945)*x[2]
        s[3,4,2,0]=(-0.58228358663)*x[0]+(0.00936548588525)*x[1]+(-1.38536320116)*x[2]
        s[3,4,2,1]=(0.0419678913201)*x[0]+(1.02999559903)*x[1]+(-0.464745184871)*x[2]
        ref[0,0,0,0]=(1.30172410662)/(o+1.)+(0.647406976573)
        ref[0,0,0,1]=(-1.09442521373)/(o+1.)+(-0.120876402476)
        ref[0,0,1,0]=(2.01757642043)/(o+1.)+(-0.0865864392547)
        ref[0,0,1,1]=(1.80034002198)/(o+1.)+(-0.216430392892)
        ref[0,0,2,0]=(1.13142112916)/(o+1.)+(-1.12383935547)
        ref[0,0,2,1]=(-0.905515986966)/(o+1.)+(0.34830679267)
        ref[0,1,0,0]=(-0.210441134504)/(o+1.)+(-0.698200622791)
        ref[0,1,0,1]=(0.360807961621)/(o+1.)+(0.409026621036)
        ref[0,1,1,0]=(1.5147969883)/(o+1.)+(0.0713001057429)
        ref[0,1,1,1]=(0.396123567843)/(o+1.)+(-0.511105156634)
        ref[0,1,2,0]=(0.184468760768)/(o+1.)+(0.448555674728)
        ref[0,1,2,1]=(-0.133892978353)/(o+1.)+(-0.510937201862)
        ref[0,2,0,0]=(1.66137009462)/(o+1.)+(0.493389512541)
        ref[0,2,0,1]=(0.358452769174)/(o+1.)+(-0.982664817753)
        ref[0,2,1,0]=(0.966434135345)/(o+1.)+(-0.240061752665)
        ref[0,2,1,1]=(-0.401753883503)/(o+1.)+(0.362102239635)
        ref[0,2,2,0]=(-0.0153619747046)/(o+1.)+(0.221454775071)
        ref[0,2,2,1]=(-0.104138117282)/(o+1.)+(-0.119505133417)
        ref[0,3,0,0]=(-0.966204569194)/(o+1.)+(0.662141336872)
        ref[0,3,0,1]=(0.587661220327)/(o+1.)+(0.10506616437)
        ref[0,3,1,0]=(2.02288199931)/(o+1.)+(-0.337539903289)
        ref[0,3,1,1]=(-1.78842825396)/(o+1.)+(-0.0844686616883)
        ref[0,3,2,0]=(0.927282266945)/(o+1.)+(0.190527509786)
        ref[0,3,2,1]=(-0.679518039292)/(o+1.)+(-0.796251782745)
        ref[0,4,0,0]=(0.555388124049)/(o+1.)+(0.562028482163)
        ref[0,4,0,1]=(0.0870397234082)/(o+1.)+(0.565035905435)
        ref[0,4,1,0]=(-1.1463502212)/(o+1.)+(-0.0970212431374)
        ref[0,4,1,1]=(0.233233715042)/(o+1.)+(-0.729602468086)
        ref[0,4,2,0]=(-0.782889963002)/(o+1.)+(0.781502066786)
        ref[0,4,2,1]=(0.763433249649)/(o+1.)+(-1.2651257696)
        ref[1,0,0,0]=(-1.00441831898)/(o+1.)+(-0.685739920821)
        ref[1,0,0,1]=(0.751247546824)/(o+1.)+(0.456810363824)
        ref[1,0,1,0]=(2.22764882334)/(o+1.)+(-0.372219507613)
        ref[1,0,1,1]=(-1.19623631804)/(o+1.)+(0.39718929366)
        ref[1,0,2,0]=(-1.60343741738)/(o+1.)+(-0.250677115636)
        ref[1,0,2,1]=(-1.25680789752)/(o+1.)+(0.61750260882)
        ref[1,1,0,0]=(-1.45519189991)/(o+1.)+(0.681606304054)
        ref[1,1,0,1]=(0.840628495308)/(o+1.)+(-0.678283412578)
        ref[1,1,1,0]=(1.41826768667)/(o+1.)+(0.144796647431)
        ref[1,1,1,1]=(-0.933723251106)/(o+1.)+(0.451975702985)
        ref[1,1,2,0]=(-0.408324994487)/(o+1.)+(0.39165840948)
        ref[1,1,2,1]=(1.23290517672)/(o+1.)+(0.323935107053)
        ref[1,2,0,0]=(0.825483086179)/(o+1.)+(0.0301218157877)
        ref[1,2,0,1]=(-0.0793934768662)/(o+1.)+(-0.142521007854)
        ref[1,2,1,0]=(-0.0709650957035)/(o+1.)+(-0.874854727116)
        ref[1,2,1,1]=(1.10714952407)/(o+1.)+(-0.208486520392)
        ref[1,2,2,0]=(0.99013461187)/(o+1.)+(0.359768530582)
        ref[1,2,2,1]=(0.0356627876162)/(o+1.)+(0.488634391872)
        ref[1,3,0,0]=(-0.474442326803)/(o+1.)+(-0.499450411015)
        ref[1,3,0,1]=(-0.673232992552)/(o+1.)+(-1.16158008694)
        ref[1,3,1,0]=(0.411652221393)/(o+1.)+(0.966457956295)
        ref[1,3,1,1]=(-1.76698398732)/(o+1.)+(-0.559250216883)
        ref[1,3,2,0]=(1.9874048718)/(o+1.)+(-0.740838072)
        ref[1,3,2,1]=(0.661784629796)/(o+1.)+(0.216931076303)
        ref[1,4,0,0]=(-0.513029515019)/(o+1.)+(0.0818103062045)
        ref[1,4,0,1]=(0.282755142694)/(o+1.)+(-0.261370981611)
        ref[1,4,1,0]=(-1.96973437805)/(o+1.)+(-0.478746163042)
        ref[1,4,1,1]=(0.908774952181)/(o+1.)+(0.0465858502998)
        ref[1,4,2,0]=(1.3422757017)/(o+1.)+(-0.0996743832723)
        ref[1,4,2,1]=(1.21609351443)/(o+1.)+(-0.349173948065)
        ref[2,0,0,0]=(-1.01744592956)/(o+1.)+(-0.923588464944)
        ref[2,0,0,1]=(1.10931736222)/(o+1.)+(0.224335478673)
        ref[2,0,1,0]=(-0.416479828321)/(o+1.)+(-0.285640270869)
        ref[2,0,1,1]=(-1.5671174369)/(o+1.)+(-0.467268571897)
        ref[2,0,2,0]=(-0.274214960025)/(o+1.)+(0.163118815513)
        ref[2,0,2,1]=(0.000680528469817)/(o+1.)+(0.0788535606642)
        ref[2,1,0,0]=(-0.844192475139)/(o+1.)+(-0.262252149651)
        ref[2,1,0,1]=(0.69399848864)/(o+1.)+(-0.395963008628)
        ref[2,1,1,0]=(-1.37742489692)/(o+1.)+(-0.128468449379)
        ref[2,1,1,1]=(0.282095133276)/(o+1.)+(-0.641487907195)
        ref[2,1,2,0]=(-1.4727375549)/(o+1.)+(0.343315122764)
        ref[2,1,2,1]=(0.67637649169)/(o+1.)+(0.0439030384195)
        ref[2,2,0,0]=(1.58598306034)/(o+1.)+(0.257398552088)
        ref[2,2,0,1]=(-0.187654980952)/(o+1.)+(0.677607655001)
        ref[2,2,1,0]=(0.225457027395)/(o+1.)+(0.221493863079)
        ref[2,2,1,1]=(-1.66885547416)/(o+1.)+(-0.372038995202)
        ref[2,2,2,0]=(-1.42708510165)/(o+1.)+(0.410904476533)
        ref[2,2,2,1]=(0.352696956576)/(o+1.)+(-0.252899306452)
        ref[2,3,0,0]=(0.266365999523)/(o+1.)+(0.249791335977)
        ref[2,3,0,1]=(0.926084735451)/(o+1.)+(0.0271701662571)
        ref[2,3,1,0]=(-0.138787407751)/(o+1.)+(-0.652714809939)
        ref[2,3,1,1]=(-0.834750293031)/(o+1.)+(-0.112504360161)
        ref[2,3,2,0]=(-0.806210182586)/(o+1.)+(0.415200684382)
        ref[2,3,2,1]=(0.577899253552)/(o+1.)+(-0.913828206984)
        ref[2,4,0,0]=(1.38802720612)/(o+1.)+(0.755900031203)
        ref[2,4,0,1]=(-1.33634216788)/(o+1.)+(0.504253453293)
        ref[2,4,1,0]=(-1.13393506986)/(o+1.)+(0.195255119392)
        ref[2,4,1,1]=(-0.384732222354)/(o+1.)+(0.203237915028)
        ref[2,4,2,0]=(-1.18866097183)/(o+1.)+(-0.122726302784)
        ref[2,4,2,1]=(-1.41914128138)/(o+1.)+(-0.382918250045)
        ref[3,0,0,0]=(-1.3617379499)/(o+1.)+(0.916513859836)
        ref[3,0,0,1]=(-0.577580679368)/(o+1.)+(-0.580633018955)
        ref[3,0,1,0]=(-0.730926522369)/(o+1.)+(-0.50988617222)
        ref[3,0,1,1]=(0.852352706539)/(o+1.)+(-0.314072812402)
        ref[3,0,2,0]=(0.63914695576)/(o+1.)+(-0.155678014016)
        ref[3,0,2,1]=(-2.31555597527)/(o+1.)+(-0.593748666755)
        ref[3,1,0,0]=(-0.31308710102)/(o+1.)+(-0.011689493559)
        ref[3,1,0,1]=(0.565049631081)/(o+1.)+(-0.115943082793)
        ref[3,1,1,0]=(1.76264026827)/(o+1.)+(1.35490058502)
        ref[3,1,1,1]=(1.34557111025)/(o+1.)+(0.873578306883)
        ref[3,1,2,0]=(0.924587947935)/(o+1.)+(-0.626617402844)
        ref[3,1,2,1]=(-0.679695689354)/(o+1.)+(0.143800073503)
        ref[3,2,0,0]=(-0.408824291494)/(o+1.)+(-0.330341171766)
        ref[3,2,0,1]=(1.02728720579)/(o+1.)+(0.860379917667)
        ref[3,2,1,0]=(-1.99227170213)/(o+1.)+(-0.251611418656)
        ref[3,2,1,1]=(-0.844161339595)/(o+1.)+(0.159252406352)
        ref[3,2,2,0]=(1.52009831693)/(o+1.)+(-0.519116196235)
        ref[3,2,2,1]=(0.323553722724)/(o+1.)+(-1.02888651238)
        ref[3,3,0,0]=(-1.39647666034)/(o+1.)+(0.637724793483)
        ref[3,3,0,1]=(0.960901654854)/(o+1.)+(-0.00807626431577)
        ref[3,3,1,0]=(0.383998350003)/(o+1.)+(-0.154799534246)
        ref[3,3,1,1]=(-0.851062083214)/(o+1.)+(0.752100475729)
        ref[3,3,2,0]=(0.788418151505)/(o+1.)+(-0.614906659263)
        ref[3,3,2,1]=(-0.1028922667)/(o+1.)+(0.135563961629)
        ref[3,4,0,0]=(-0.540240088251)/(o+1.)+(-0.425382378277)
        ref[3,4,0,1]=(0.236730341461)/(o+1.)+(0.979828209623)
        ref[3,4,1,0]=(-0.9880898216)/(o+1.)+(-0.198536176202)
        ref[3,4,1,1]=(1.06221574228)/(o+1.)+(-0.46475456894)
        ref[3,4,2,0]=(-1.89471538439)/(o+1.)+(-0.0317829587581)
        ref[3,4,2,1]=(-1.20699948566)/(o+1.)+(0.907108895566)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Function_rank0(self):
      """
      tests integral of rank 0 Symbol on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.107044842683)*x[0]**o+(-0.198407800146)*x[0]+(0.877472712926)*x[1]**o+(-0.647405831691)*x[1]
        ref=(0.770427870243)/(o+1.)+(-0.422906815918)
      else:
        s=(-0.559912240782)*x[0]**o+(0.766296134438)*x[0]+(-0.110980192077)*x[1]**o+(-0.608865520099)*x[1]+(0.672667086497)*x[2]**o+(0.334463167521)*x[2]
        ref=(0.00177465363795)/(o+1.)+(0.24594689093)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Function_rank1(self):
      """
      tests integral of rank 1 Symbol on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.416213605972)*x[0]**o+(0.77813592218)*x[0]+(-0.222437793488)*x[1]**o+(-0.933738289486)*x[1]
        s[1]=(0.455257144475)*x[0]**o+(-0.622417334686)*x[0]+(0.806205896623)*x[1]**o+(0.675691766701)*x[1]
        ref[0]=(-0.638651399461)/(o+1.)+(-0.0778011836529)
        ref[1]=(1.2614630411)/(o+1.)+(0.0266372160075)
      else:
        s[0]=(0.753437228142)*x[0]**o+(0.148952810054)*x[0]+(0.179818865036)*x[1]**o+(0.490547213201)*x[1]+(0.977606414203)*x[2]**o+(0.0928058817324)*x[2]
        s[1]=(0.964988411956)*x[0]**o+(0.446937048611)*x[0]+(-0.110739634254)*x[1]**o+(0.906002899971)*x[1]+(0.666451027663)*x[2]**o+(0.270230684952)*x[2]
        ref[0]=(1.91086250738)/(o+1.)+(0.366152952493)
        ref[1]=(1.52069980537)/(o+1.)+(0.811585316767)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Function_rank2(self):
      """
      tests integral of rank 2 Symbol on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.08214321783)*x[0]**o+(-0.496010136907)*x[0]+(-0.661161523248)*x[1]**o+(-0.151256978215)*x[1]
        s[0,1]=(0.00779835697637)*x[0]**o+(0.307130297185)*x[0]+(-0.111598029557)*x[1]**o+(-0.00186165192551)*x[1]
        s[0,2]=(0.0978324862172)*x[0]**o+(-0.746313762674)*x[0]+(0.519229413642)*x[1]**o+(0.892797139942)*x[1]
        s[0,3]=(-0.598407415713)*x[0]**o+(0.801307410865)*x[0]+(0.894609711877)*x[1]**o+(0.0520830645255)*x[1]
        s[0,4]=(0.580674315696)*x[0]**o+(-0.000690290928564)*x[0]+(0.909550857813)*x[1]**o+(-0.766950054968)*x[1]
        s[1,0]=(0.293745692233)*x[0]**o+(0.581206623889)*x[0]+(0.154848883067)*x[1]**o+(0.260519542891)*x[1]
        s[1,1]=(0.81822409326)*x[0]**o+(0.0527752269091)*x[0]+(0.960312253084)*x[1]**o+(-0.844575939421)*x[1]
        s[1,2]=(-0.186016521867)*x[0]**o+(0.316565783375)*x[0]+(-0.206256264607)*x[1]**o+(0.951167349262)*x[1]
        s[1,3]=(-0.355700637442)*x[0]**o+(-0.636455488757)*x[0]+(0.623605194008)*x[1]**o+(-0.0426220883099)*x[1]
        s[1,4]=(0.72832697443)*x[0]**o+(-0.349553513535)*x[0]+(-0.128283294734)*x[1]**o+(-0.280029190018)*x[1]
        s[2,0]=(0.634737331487)*x[0]**o+(-0.0784762897091)*x[0]+(0.49107451702)*x[1]**o+(-0.927830147754)*x[1]
        s[2,1]=(-0.561516427218)*x[0]**o+(0.414180445476)*x[0]+(-0.32846350728)*x[1]**o+(0.653302722231)*x[1]
        s[2,2]=(-0.285665357177)*x[0]**o+(-0.495353890909)*x[0]+(-0.300417232522)*x[1]**o+(0.828720486127)*x[1]
        s[2,3]=(0.582282364197)*x[0]**o+(-0.491365235361)*x[0]+(-0.481532775048)*x[1]**o+(0.843340014315)*x[1]
        s[2,4]=(0.467528444795)*x[0]**o+(-0.617430509987)*x[0]+(0.0370576618542)*x[1]**o+(0.124933927604)*x[1]
        s[3,0]=(0.76865308901)*x[0]**o+(0.270205155941)*x[0]+(0.239796830473)*x[1]**o+(-0.732593342574)*x[1]
        s[3,1]=(0.276346458614)*x[0]**o+(-0.637344640378)*x[0]+(-0.697338900523)*x[1]**o+(0.136084835992)*x[1]
        s[3,2]=(-0.932808156816)*x[0]**o+(0.296855273797)*x[0]+(-0.417934077535)*x[1]**o+(0.897857223334)*x[1]
        s[3,3]=(0.358813012298)*x[0]**o+(-0.421306401815)*x[0]+(0.682226516534)*x[1]**o+(-0.0533913429448)*x[1]
        s[3,4]=(0.27862707257)*x[0]**o+(-0.227499516484)*x[0]+(0.502409049994)*x[1]**o+(0.348218429657)*x[1]
        ref[0,0]=(-0.579018305418)/(o+1.)+(-0.323633557561)
        ref[0,1]=(-0.103799672581)/(o+1.)+(0.15263432263)
        ref[0,2]=(0.617061899859)/(o+1.)+(0.0732416886339)
        ref[0,3]=(0.296202296164)/(o+1.)+(0.426695237695)
        ref[0,4]=(1.49022517351)/(o+1.)+(-0.383820172948)
        ref[1,0]=(0.4485945753)/(o+1.)+(0.42086308339)
        ref[1,1]=(1.77853634634)/(o+1.)+(-0.395900356256)
        ref[1,2]=(-0.392272786474)/(o+1.)+(0.633866566319)
        ref[1,3]=(0.267904556566)/(o+1.)+(-0.339538788533)
        ref[1,4]=(0.600043679696)/(o+1.)+(-0.314791351776)
        ref[2,0]=(1.12581184851)/(o+1.)+(-0.503153218732)
        ref[2,1]=(-0.889979934498)/(o+1.)+(0.533741583854)
        ref[2,2]=(-0.586082589699)/(o+1.)+(0.166683297609)
        ref[2,3]=(0.100749589149)/(o+1.)+(0.175987389477)
        ref[2,4]=(0.504586106649)/(o+1.)+(-0.246248291192)
        ref[3,0]=(1.00844991948)/(o+1.)+(-0.231194093316)
        ref[3,1]=(-0.420992441909)/(o+1.)+(-0.250629902193)
        ref[3,2]=(-1.35074223435)/(o+1.)+(0.597356248566)
        ref[3,3]=(1.04103952883)/(o+1.)+(-0.23734887238)
        ref[3,4]=(0.781036122564)/(o+1.)+(0.0603594565866)
      else:
        s[0,0]=(-0.941295734032)*x[0]**o+(0.396841358423)*x[0]+(-0.490021992176)*x[1]**o+(-0.346405280728)*x[1]+(-0.201846160921)*x[2]**o+(0.20749764579)*x[2]
        s[0,1]=(-0.88981448497)*x[0]**o+(-0.884955504659)*x[0]+(0.182259548742)*x[1]**o+(-0.235489618403)*x[1]+(-0.977189329039)*x[2]**o+(-0.514950614838)*x[2]
        s[0,2]=(0.328187150969)*x[0]**o+(-0.448497961254)*x[0]+(0.472389791923)*x[1]**o+(-0.624930304822)*x[1]+(0.265474988322)*x[2]**o+(-0.0214969683756)*x[2]
        s[0,3]=(-0.735757404485)*x[0]**o+(-0.92658492028)*x[0]+(0.684533158897)*x[1]**o+(-0.865066520203)*x[1]+(-0.190890639018)*x[2]**o+(-0.529642573907)*x[2]
        s[0,4]=(0.955420842735)*x[0]**o+(0.773074765628)*x[0]+(-0.382656468864)*x[1]**o+(0.47181739927)*x[1]+(0.873841536673)*x[2]**o+(-0.526372620067)*x[2]
        s[1,0]=(0.770067221954)*x[0]**o+(-0.406438793272)*x[0]+(0.51351865241)*x[1]**o+(-0.856834133418)*x[1]+(0.589461089026)*x[2]**o+(-0.141812408947)*x[2]
        s[1,1]=(-0.812514809415)*x[0]**o+(-0.456460678755)*x[0]+(0.430628764669)*x[1]**o+(0.262068556418)*x[1]+(0.597424763719)*x[2]**o+(-0.157908736594)*x[2]
        s[1,2]=(-0.46727482012)*x[0]**o+(-0.744512472884)*x[0]+(-0.365664650476)*x[1]**o+(-0.361046498637)*x[1]+(-0.166284258648)*x[2]**o+(0.528262791587)*x[2]
        s[1,3]=(-0.273755634801)*x[0]**o+(-0.15873969342)*x[0]+(-0.368357879665)*x[1]**o+(0.283273650468)*x[1]+(-0.0175031061849)*x[2]**o+(-0.0417868805286)*x[2]
        s[1,4]=(0.349815667052)*x[0]**o+(-0.175629838767)*x[0]+(0.272305468439)*x[1]**o+(-0.750728619331)*x[1]+(0.0723037064016)*x[2]**o+(0.911924754086)*x[2]
        s[2,0]=(0.834136592552)*x[0]**o+(0.227306622837)*x[0]+(-0.468357820454)*x[1]**o+(0.0429839240965)*x[1]+(0.397233264579)*x[2]**o+(0.645297775174)*x[2]
        s[2,1]=(-0.420265217156)*x[0]**o+(-0.762906499224)*x[0]+(0.16960694356)*x[1]**o+(0.381251517238)*x[1]+(-0.740637261618)*x[2]**o+(-0.437283161876)*x[2]
        s[2,2]=(-0.957535003492)*x[0]**o+(0.546534458865)*x[0]+(0.10634355782)*x[1]**o+(0.0575530053655)*x[1]+(0.395304821296)*x[2]**o+(0.0507237509151)*x[2]
        s[2,3]=(0.130545158576)*x[0]**o+(-0.297688721316)*x[0]+(-0.193136907135)*x[1]**o+(-0.530687107912)*x[1]+(0.891112852449)*x[2]**o+(0.89025257856)*x[2]
        s[2,4]=(-0.452227707094)*x[0]**o+(0.518745990389)*x[0]+(-0.848908958687)*x[1]**o+(-0.984222753382)*x[1]+(-0.725703746452)*x[2]**o+(0.29370062408)*x[2]
        s[3,0]=(0.451818718553)*x[0]**o+(0.15228149344)*x[0]+(0.309329014013)*x[1]**o+(-0.629074550954)*x[1]+(0.925898897131)*x[2]**o+(-0.530392928354)*x[2]
        s[3,1]=(0.48661045622)*x[0]**o+(0.362115939707)*x[0]+(0.461789600557)*x[1]**o+(-0.0265394801756)*x[1]+(-0.52060582127)*x[2]**o+(0.847046142749)*x[2]
        s[3,2]=(0.356328974486)*x[0]**o+(0.735492369055)*x[0]+(0.833634465835)*x[1]**o+(-0.603533125221)*x[1]+(-0.799865132292)*x[2]**o+(0.255351535539)*x[2]
        s[3,3]=(0.821371453867)*x[0]**o+(0.360956876043)*x[0]+(0.0191507582216)*x[1]**o+(-0.995259844736)*x[1]+(-0.00388518722366)*x[2]**o+(-0.0360171649464)*x[2]
        s[3,4]=(-0.248806545065)*x[0]**o+(-0.598745625081)*x[0]+(0.722047538053)*x[1]**o+(-0.840877899002)*x[1]+(0.920023026949)*x[2]**o+(0.503521861458)*x[2]
        ref[0,0]=(-1.63316388713)/(o+1.)+(0.128966861742)
        ref[0,1]=(-1.68474426527)/(o+1.)+(-0.81769786895)
        ref[0,2]=(1.06605193121)/(o+1.)+(-0.547462617226)
        ref[0,3]=(-0.242114884606)/(o+1.)+(-1.16064700719)
        ref[0,4]=(1.44660591054)/(o+1.)+(0.359259772415)
        ref[1,0]=(1.87304696339)/(o+1.)+(-0.702542667819)
        ref[1,1]=(0.215538718973)/(o+1.)+(-0.176150429465)
        ref[1,2]=(-0.999223729244)/(o+1.)+(-0.288648089967)
        ref[1,3]=(-0.659616620651)/(o+1.)+(0.04137353826)
        ref[1,4]=(0.694424841892)/(o+1.)+(-0.00721685200611)
        ref[2,0]=(0.763012036677)/(o+1.)+(0.457794161054)
        ref[2,1]=(-0.991295535214)/(o+1.)+(-0.409469071931)
        ref[2,2]=(-0.455886624375)/(o+1.)+(0.327405607573)
        ref[2,3]=(0.82852110389)/(o+1.)+(0.030938374666)
        ref[2,4]=(-2.02684041223)/(o+1.)+(-0.0858880694563)
        ref[3,0]=(1.6870466297)/(o+1.)+(-0.503592992934)
        ref[3,1]=(0.427794235507)/(o+1.)+(0.59131130114)
        ref[3,2]=(0.39009830803)/(o+1.)+(0.193655389686)
        ref[3,3]=(0.836637024865)/(o+1.)+(-0.33516006682)
        ref[3,4]=(1.39326401994)/(o+1.)+(-0.468050831312)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Function_rank3(self):
      """
      tests integral of rank 3 Symbol on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(-0.944773955793)*x[0]**o+(-0.662996681032)*x[0]+(-0.442867713759)*x[1]**o+(-0.229023301726)*x[1]
        s[0,0,1]=(0.0407144183428)*x[0]**o+(-0.669891322439)*x[0]+(0.920328407938)*x[1]**o+(-0.320681600926)*x[1]
        s[0,1,0]=(0.179693170721)*x[0]**o+(-0.671664515028)*x[0]+(0.525108213743)*x[1]**o+(0.310882938487)*x[1]
        s[0,1,1]=(-0.0540671676637)*x[0]**o+(0.000853108264254)*x[0]+(-0.370081882537)*x[1]**o+(-0.138484250063)*x[1]
        s[1,0,0]=(0.631472269828)*x[0]**o+(-0.294848092138)*x[0]+(0.0730087097944)*x[1]**o+(0.0292051228565)*x[1]
        s[1,0,1]=(-0.576009492093)*x[0]**o+(-0.228105833611)*x[0]+(-0.0563734559287)*x[1]**o+(-0.242583876037)*x[1]
        s[1,1,0]=(-0.597734903158)*x[0]**o+(0.960078413053)*x[0]+(0.881951380528)*x[1]**o+(-0.186121672274)*x[1]
        s[1,1,1]=(-0.117005971726)*x[0]**o+(-0.809886380446)*x[0]+(-0.981092676193)*x[1]**o+(0.0636030759105)*x[1]
        s[2,0,0]=(-0.197553649616)*x[0]**o+(-0.820287165629)*x[0]+(0.359670409851)*x[1]**o+(-0.861222386581)*x[1]
        s[2,0,1]=(0.968998274457)*x[0]**o+(0.376230063695)*x[0]+(0.283097885679)*x[1]**o+(-0.128262765246)*x[1]
        s[2,1,0]=(0.0310410423851)*x[0]**o+(0.611929570715)*x[0]+(-0.972311035606)*x[1]**o+(-0.568277782449)*x[1]
        s[2,1,1]=(0.95074765381)*x[0]**o+(0.338991372672)*x[0]+(0.264103717561)*x[1]**o+(0.370834284282)*x[1]
        s[3,0,0]=(-0.664867197259)*x[0]**o+(0.917684307761)*x[0]+(-0.319774527149)*x[1]**o+(0.432915185753)*x[1]
        s[3,0,1]=(-0.162944024583)*x[0]**o+(-0.123907346782)*x[0]+(0.325356901629)*x[1]**o+(0.90249028492)*x[1]
        s[3,1,0]=(-0.135439910119)*x[0]**o+(0.920226918882)*x[0]+(0.112667508952)*x[1]**o+(0.710942378416)*x[1]
        s[3,1,1]=(-0.370025988485)*x[0]**o+(0.146841174972)*x[0]+(-0.209033207026)*x[1]**o+(0.270439136519)*x[1]
        s[4,0,0]=(0.724119312565)*x[0]**o+(0.726636721509)*x[0]+(-0.643509558883)*x[1]**o+(-0.544447102654)*x[1]
        s[4,0,1]=(-0.563641108028)*x[0]**o+(-0.320916186803)*x[0]+(0.935527371187)*x[1]**o+(0.535917271795)*x[1]
        s[4,1,0]=(-0.482028335236)*x[0]**o+(-0.510759210235)*x[0]+(-0.105477148218)*x[1]**o+(0.54069501931)*x[1]
        s[4,1,1]=(0.305620947599)*x[0]**o+(-0.724738374933)*x[0]+(0.470881233144)*x[1]**o+(-0.705410563798)*x[1]
        s[5,0,0]=(0.168173718235)*x[0]**o+(0.578831386242)*x[0]+(-0.869711732793)*x[1]**o+(-0.176143888818)*x[1]
        s[5,0,1]=(-0.228160489841)*x[0]**o+(-0.0297036971933)*x[0]+(0.391702396499)*x[1]**o+(0.875713079041)*x[1]
        s[5,1,0]=(-0.910154283353)*x[0]**o+(-0.0651085103088)*x[0]+(0.0884987382756)*x[1]**o+(-0.937409873275)*x[1]
        s[5,1,1]=(0.41106472762)*x[0]**o+(-0.801661249269)*x[0]+(0.201137556319)*x[1]**o+(-0.665423718838)*x[1]
        ref[0,0,0]=(-1.38764166955)/(o+1.)+(-0.446009991379)
        ref[0,0,1]=(0.961042826281)/(o+1.)+(-0.495286461682)
        ref[0,1,0]=(0.704801384464)/(o+1.)+(-0.180390788271)
        ref[0,1,1]=(-0.424149050201)/(o+1.)+(-0.0688155708996)
        ref[1,0,0]=(0.704480979622)/(o+1.)+(-0.132821484641)
        ref[1,0,1]=(-0.632382948021)/(o+1.)+(-0.235344854824)
        ref[1,1,0]=(0.28421647737)/(o+1.)+(0.386978370389)
        ref[1,1,1]=(-1.09809864792)/(o+1.)+(-0.373141652268)
        ref[2,0,0]=(0.162116760235)/(o+1.)+(-0.840754776105)
        ref[2,0,1]=(1.25209616014)/(o+1.)+(0.123983649224)
        ref[2,1,0]=(-0.941269993221)/(o+1.)+(0.0218258941331)
        ref[2,1,1]=(1.21485137137)/(o+1.)+(0.354912828477)
        ref[3,0,0]=(-0.984641724408)/(o+1.)+(0.675299746757)
        ref[3,0,1]=(0.162412877046)/(o+1.)+(0.389291469069)
        ref[3,1,0]=(-0.0227724011669)/(o+1.)+(0.815584648649)
        ref[3,1,1]=(-0.579059195511)/(o+1.)+(0.208640155746)
        ref[4,0,0]=(0.0806097536818)/(o+1.)+(0.0910948094271)
        ref[4,0,1]=(0.371886263159)/(o+1.)+(0.107500542496)
        ref[4,1,0]=(-0.587505483453)/(o+1.)+(0.0149679045376)
        ref[4,1,1]=(0.776502180743)/(o+1.)+(-0.715074469366)
        ref[5,0,0]=(-0.701538014558)/(o+1.)+(0.201343748712)
        ref[5,0,1]=(0.163541906658)/(o+1.)+(0.423004690924)
        ref[5,1,0]=(-0.821655545078)/(o+1.)+(-0.501259191792)
        ref[5,1,1]=(0.612202283939)/(o+1.)+(-0.733542484054)
      else:
        s[0,0,0]=(-0.0829784527698)*x[0]**o+(-0.332549632616)*x[0]+(0.255962043751)*x[1]**o+(-0.176352635053)*x[1]+(-0.954055133743)*x[2]**o+(-0.898668627164)*x[2]
        s[0,0,1]=(-0.0566639977041)*x[0]**o+(0.64899823634)*x[0]+(0.339703741475)*x[1]**o+(-0.754613713569)*x[1]+(0.379536135428)*x[2]**o+(0.118057489576)*x[2]
        s[0,1,0]=(-0.718453260796)*x[0]**o+(0.997156187689)*x[0]+(0.914514642799)*x[1]**o+(-0.633873909376)*x[1]+(-0.485691739694)*x[2]**o+(0.415370908143)*x[2]
        s[0,1,1]=(0.553256681553)*x[0]**o+(-0.191413319548)*x[0]+(0.800894762679)*x[1]**o+(0.919882584486)*x[1]+(0.127165880997)*x[2]**o+(-0.690488492514)*x[2]
        s[1,0,0]=(-0.142765560936)*x[0]**o+(-0.303407652797)*x[0]+(0.989394601479)*x[1]**o+(-0.484741465949)*x[1]+(0.732648062756)*x[2]**o+(-0.678857629867)*x[2]
        s[1,0,1]=(0.288139139594)*x[0]**o+(-0.534820162236)*x[0]+(0.554980676789)*x[1]**o+(-0.917630291921)*x[1]+(0.170042991043)*x[2]**o+(0.826809263768)*x[2]
        s[1,1,0]=(-0.901850450639)*x[0]**o+(-0.0365789634899)*x[0]+(0.0622336448434)*x[1]**o+(0.22005965603)*x[1]+(0.857183927574)*x[2]**o+(0.254453636804)*x[2]
        s[1,1,1]=(-0.888020161965)*x[0]**o+(-0.0741224453822)*x[0]+(0.887224038781)*x[1]**o+(0.742137909151)*x[1]+(-0.587100789782)*x[2]**o+(-0.928415179888)*x[2]
        s[2,0,0]=(0.00216356197373)*x[0]**o+(0.248793770338)*x[0]+(0.518030493147)*x[1]**o+(-0.508299580849)*x[1]+(-0.0290711721801)*x[2]**o+(0.662305993163)*x[2]
        s[2,0,1]=(0.924293736923)*x[0]**o+(0.218836928692)*x[0]+(-0.270314297479)*x[1]**o+(0.20290032209)*x[1]+(-0.87536943379)*x[2]**o+(-0.864192540105)*x[2]
        s[2,1,0]=(0.416159016379)*x[0]**o+(-0.30257442476)*x[0]+(-0.257298173268)*x[1]**o+(-0.398564079155)*x[1]+(-0.0699018670952)*x[2]**o+(-0.540145396091)*x[2]
        s[2,1,1]=(0.82076420765)*x[0]**o+(-0.265319227281)*x[0]+(-0.156971036756)*x[1]**o+(-0.468527020278)*x[1]+(0.2455572117)*x[2]**o+(0.237918192218)*x[2]
        s[3,0,0]=(-0.645207425303)*x[0]**o+(0.992013339436)*x[0]+(-0.814115520055)*x[1]**o+(-0.0863699557395)*x[1]+(0.356684159665)*x[2]**o+(0.0215161017879)*x[2]
        s[3,0,1]=(0.429124334865)*x[0]**o+(-0.516632695494)*x[0]+(0.276076705271)*x[1]**o+(0.73233754297)*x[1]+(-0.499960919138)*x[2]**o+(0.927448433933)*x[2]
        s[3,1,0]=(0.105016655639)*x[0]**o+(0.627563943644)*x[0]+(0.0134384788636)*x[1]**o+(-0.491675197932)*x[1]+(0.850898525327)*x[2]**o+(-0.153767354797)*x[2]
        s[3,1,1]=(0.0927473537694)*x[0]**o+(-0.164202910181)*x[0]+(-0.246744412752)*x[1]**o+(0.812782310771)*x[1]+(0.452025251957)*x[2]**o+(-0.46155534584)*x[2]
        s[4,0,0]=(0.466686843553)*x[0]**o+(-0.454360335871)*x[0]+(0.600983402349)*x[1]**o+(-0.248407743104)*x[1]+(-0.114458429866)*x[2]**o+(0.363120759629)*x[2]
        s[4,0,1]=(0.108248019353)*x[0]**o+(-0.386567024916)*x[0]+(0.589911190112)*x[1]**o+(0.974355188979)*x[1]+(0.956434046018)*x[2]**o+(-0.750308948178)*x[2]
        s[4,1,0]=(-0.0575495701591)*x[0]**o+(-0.297079422559)*x[0]+(0.173709916352)*x[1]**o+(-0.703203763304)*x[1]+(-0.206920064792)*x[2]**o+(-0.566180018312)*x[2]
        s[4,1,1]=(0.752722443401)*x[0]**o+(0.287487922131)*x[0]+(-0.253690778509)*x[1]**o+(0.158399655258)*x[1]+(0.365133322462)*x[2]**o+(-0.765340889542)*x[2]
        s[5,0,0]=(-0.249224479124)*x[0]**o+(-0.81760956491)*x[0]+(0.664454285698)*x[1]**o+(-0.111438607721)*x[1]+(-0.34058205464)*x[2]**o+(-0.604036230847)*x[2]
        s[5,0,1]=(-0.0489505541715)*x[0]**o+(-0.216719586033)*x[0]+(0.862292178064)*x[1]**o+(-0.978934839395)*x[1]+(-0.229118015068)*x[2]**o+(-0.220547177874)*x[2]
        s[5,1,0]=(-0.200117801874)*x[0]**o+(0.826116119716)*x[0]+(0.180058706489)*x[1]**o+(0.938844014751)*x[1]+(0.232853505918)*x[2]**o+(-0.0718472890928)*x[2]
        s[5,1,1]=(0.335820727851)*x[0]**o+(-0.30183416972)*x[0]+(-0.69421482762)*x[1]**o+(0.105894103212)*x[1]+(0.620400009549)*x[2]**o+(0.970029758381)*x[2]
        ref[0,0,0]=(-0.781071542761)/(o+1.)+(-0.703785447417)
        ref[0,0,1]=(0.662575879198)/(o+1.)+(0.00622100617352)
        ref[0,1,0]=(-0.289630357691)/(o+1.)+(0.389326593228)
        ref[0,1,1]=(1.48131732523)/(o+1.)+(0.0189903862116)
        ref[1,0,0]=(1.5792771033)/(o+1.)+(-0.733503374307)
        ref[1,0,1]=(1.01316280743)/(o+1.)+(-0.312820595194)
        ref[1,1,0]=(0.0175671217784)/(o+1.)+(0.218967164672)
        ref[1,1,1]=(-0.587896912967)/(o+1.)+(-0.13019985806)
        ref[2,0,0]=(0.49112288294)/(o+1.)+(0.201400091326)
        ref[2,0,1]=(-0.221389994346)/(o+1.)+(-0.221227644661)
        ref[2,1,0]=(0.0889589760167)/(o+1.)+(-0.620641950003)
        ref[2,1,1]=(0.909350382595)/(o+1.)+(-0.24796402767)
        ref[3,0,0]=(-1.10263878569)/(o+1.)+(0.463579742742)
        ref[3,0,1]=(0.205240120999)/(o+1.)+(0.571576640705)
        ref[3,1,0]=(0.96935365983)/(o+1.)+(-0.00893930454231)
        ref[3,1,1]=(0.298028192975)/(o+1.)+(0.0935120273748)
        ref[4,0,0]=(0.953211816036)/(o+1.)+(-0.169823659673)
        ref[4,0,1]=(1.65459325548)/(o+1.)+(-0.0812603920577)
        ref[4,1,0]=(-0.0907597185993)/(o+1.)+(-0.783231602088)
        ref[4,1,1]=(0.864164987355)/(o+1.)+(-0.159726656077)
        ref[5,0,0]=(0.0746477519342)/(o+1.)+(-0.766542201739)
        ref[5,0,1]=(0.584223608825)/(o+1.)+(-0.708100801651)
        ref[5,1,0]=(0.212794410532)/(o+1.)+(0.846556422687)
        ref[5,1,1]=(0.262005909781)/(o+1.)+(0.387044845936)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunction_fromSymbol_Function_rank4(self):
      """
      tests integral of rank 4 Symbol on the Function

      assumptions: Function(self.domain) exists
                   self.domain supports integral on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.64587181131)*x[0]**o+(-0.142518178492)*x[0]+(-0.399846110498)*x[1]**o+(0.00236392286301)*x[1]
        s[0,0,0,1]=(0.298109366098)*x[0]**o+(-0.0843450895271)*x[0]+(0.574794599006)*x[1]**o+(-0.107008590623)*x[1]
        s[0,0,1,0]=(-0.424486704098)*x[0]**o+(-0.382352426027)*x[0]+(0.275387919625)*x[1]**o+(-0.161015391255)*x[1]
        s[0,0,1,1]=(0.643084919048)*x[0]**o+(0.44303782474)*x[0]+(0.17010706601)*x[1]**o+(-0.160095051974)*x[1]
        s[0,0,2,0]=(0.255268276945)*x[0]**o+(-0.867088236584)*x[0]+(-0.809452452666)*x[1]**o+(-0.376237703808)*x[1]
        s[0,0,2,1]=(-0.773368404022)*x[0]**o+(-0.655491257176)*x[0]+(-0.256178554483)*x[1]**o+(0.582354235979)*x[1]
        s[0,1,0,0]=(0.312387735876)*x[0]**o+(0.36653787868)*x[0]+(-0.160583709945)*x[1]**o+(-0.791133941857)*x[1]
        s[0,1,0,1]=(-0.57693493814)*x[0]**o+(-0.862135922856)*x[0]+(0.0439301419063)*x[1]**o+(0.959289798443)*x[1]
        s[0,1,1,0]=(0.121311324776)*x[0]**o+(0.907585047473)*x[0]+(-0.923686957239)*x[1]**o+(-0.652706978591)*x[1]
        s[0,1,1,1]=(0.682901593615)*x[0]**o+(-0.306481093391)*x[0]+(-0.35714675383)*x[1]**o+(0.500943467097)*x[1]
        s[0,1,2,0]=(0.647130385973)*x[0]**o+(0.873804224811)*x[0]+(0.78779813495)*x[1]**o+(0.430789257821)*x[1]
        s[0,1,2,1]=(0.135833683749)*x[0]**o+(0.32862510689)*x[0]+(-0.48350127573)*x[1]**o+(-0.649354230648)*x[1]
        s[0,2,0,0]=(-0.641045806017)*x[0]**o+(-0.654315036977)*x[0]+(-0.729951508044)*x[1]**o+(0.766934507787)*x[1]
        s[0,2,0,1]=(0.481734616023)*x[0]**o+(-0.687904467365)*x[0]+(0.630779635862)*x[1]**o+(-0.982134985324)*x[1]
        s[0,2,1,0]=(-0.0488969750979)*x[0]**o+(-0.804550194421)*x[0]+(-0.721206215475)*x[1]**o+(0.184754632486)*x[1]
        s[0,2,1,1]=(-0.0554773461175)*x[0]**o+(0.501447670577)*x[0]+(0.214749897505)*x[1]**o+(0.912271898356)*x[1]
        s[0,2,2,0]=(0.0902819474814)*x[0]**o+(0.182092397318)*x[0]+(-0.48175183483)*x[1]**o+(-0.598040724808)*x[1]
        s[0,2,2,1]=(0.720795079983)*x[0]**o+(-0.339928403247)*x[0]+(-0.762363891559)*x[1]**o+(0.460915902605)*x[1]
        s[0,3,0,0]=(0.508573641682)*x[0]**o+(-0.015037165373)*x[0]+(0.448986097944)*x[1]**o+(0.471102721761)*x[1]
        s[0,3,0,1]=(-0.341905079065)*x[0]**o+(-0.95312476241)*x[0]+(-0.749402955548)*x[1]**o+(-0.372596238803)*x[1]
        s[0,3,1,0]=(-0.0690403122223)*x[0]**o+(-0.08921421318)*x[0]+(-0.380821740415)*x[1]**o+(-0.599974433304)*x[1]
        s[0,3,1,1]=(-0.50179318754)*x[0]**o+(0.355355851326)*x[0]+(0.721120108213)*x[1]**o+(0.454667741643)*x[1]
        s[0,3,2,0]=(0.213191653894)*x[0]**o+(-0.637815219816)*x[0]+(0.0628021542224)*x[1]**o+(-0.284841867085)*x[1]
        s[0,3,2,1]=(0.226011558647)*x[0]**o+(0.926633205695)*x[0]+(-0.801045292364)*x[1]**o+(0.84320696071)*x[1]
        s[0,4,0,0]=(-0.925639600998)*x[0]**o+(-0.743975200606)*x[0]+(0.950987915324)*x[1]**o+(-0.919549720538)*x[1]
        s[0,4,0,1]=(0.295152910157)*x[0]**o+(-0.155400545856)*x[0]+(-0.888657028463)*x[1]**o+(-0.0245586756163)*x[1]
        s[0,4,1,0]=(0.659416651507)*x[0]**o+(-0.0687605790427)*x[0]+(0.287026029448)*x[1]**o+(0.749704220615)*x[1]
        s[0,4,1,1]=(0.195399326633)*x[0]**o+(-0.690030508171)*x[0]+(0.812925291301)*x[1]**o+(-0.288219851462)*x[1]
        s[0,4,2,0]=(0.92210509534)*x[0]**o+(-0.746377740152)*x[0]+(-0.263846201131)*x[1]**o+(0.241927075102)*x[1]
        s[0,4,2,1]=(-0.342776236669)*x[0]**o+(-0.678319456068)*x[0]+(0.533446356713)*x[1]**o+(0.893285310119)*x[1]
        s[1,0,0,0]=(-0.35370685183)*x[0]**o+(-0.649199740924)*x[0]+(-0.284557848052)*x[1]**o+(-0.108918514336)*x[1]
        s[1,0,0,1]=(0.435799676535)*x[0]**o+(-0.560606926167)*x[0]+(0.422473364618)*x[1]**o+(-0.644113034355)*x[1]
        s[1,0,1,0]=(0.371384490035)*x[0]**o+(0.468745463643)*x[0]+(0.824650091185)*x[1]**o+(0.779212641032)*x[1]
        s[1,0,1,1]=(0.566680148615)*x[0]**o+(0.70499851212)*x[0]+(-0.622667334132)*x[1]**o+(0.0793283623173)*x[1]
        s[1,0,2,0]=(-0.137422675839)*x[0]**o+(-0.0828536384824)*x[0]+(0.34029178349)*x[1]**o+(-0.572205817396)*x[1]
        s[1,0,2,1]=(0.849275927844)*x[0]**o+(0.588676957111)*x[0]+(-0.351435931217)*x[1]**o+(-0.985192646644)*x[1]
        s[1,1,0,0]=(-0.832427042757)*x[0]**o+(0.511670148283)*x[0]+(-0.225009791535)*x[1]**o+(-0.726486541188)*x[1]
        s[1,1,0,1]=(-0.370148483489)*x[0]**o+(0.0364804235508)*x[0]+(-0.740077305856)*x[1]**o+(0.123450524426)*x[1]
        s[1,1,1,0]=(0.989855829356)*x[0]**o+(-0.183176166338)*x[0]+(0.0388003238684)*x[1]**o+(-0.0395036587674)*x[1]
        s[1,1,1,1]=(-0.775089474977)*x[0]**o+(-0.976089661518)*x[0]+(0.387965308645)*x[1]**o+(0.712155002639)*x[1]
        s[1,1,2,0]=(0.86011220616)*x[0]**o+(-0.95431213868)*x[0]+(-0.495540857845)*x[1]**o+(-0.701111767974)*x[1]
        s[1,1,2,1]=(0.827371223575)*x[0]**o+(-0.650031455862)*x[0]+(-0.0269816386714)*x[1]**o+(-0.490304255652)*x[1]
        s[1,2,0,0]=(-0.718249056807)*x[0]**o+(-0.812216925478)*x[0]+(0.358363619384)*x[1]**o+(-0.555431408236)*x[1]
        s[1,2,0,1]=(0.502681744448)*x[0]**o+(0.568656095603)*x[0]+(-0.642171078627)*x[1]**o+(-0.678715038904)*x[1]
        s[1,2,1,0]=(-0.321002503528)*x[0]**o+(0.399915822381)*x[0]+(0.378527801988)*x[1]**o+(0.261422806614)*x[1]
        s[1,2,1,1]=(-0.296716183029)*x[0]**o+(0.497650651416)*x[0]+(-0.386922791311)*x[1]**o+(-0.91901564657)*x[1]
        s[1,2,2,0]=(-0.828712534186)*x[0]**o+(-0.0114777410069)*x[0]+(-0.142857241422)*x[1]**o+(0.527573020984)*x[1]
        s[1,2,2,1]=(0.341639667906)*x[0]**o+(0.773780839708)*x[0]+(0.625113762311)*x[1]**o+(0.700528961096)*x[1]
        s[1,3,0,0]=(0.763866858532)*x[0]**o+(0.934384862727)*x[0]+(-0.192635958925)*x[1]**o+(-0.531023677701)*x[1]
        s[1,3,0,1]=(-0.559232654796)*x[0]**o+(-0.381248359127)*x[0]+(0.042953808137)*x[1]**o+(-0.0258385267166)*x[1]
        s[1,3,1,0]=(-0.596069058223)*x[0]**o+(0.611041390676)*x[0]+(0.320050026897)*x[1]**o+(0.31950524284)*x[1]
        s[1,3,1,1]=(0.874401923132)*x[0]**o+(0.995518258803)*x[0]+(-0.535120262939)*x[1]**o+(0.160955202392)*x[1]
        s[1,3,2,0]=(0.352942439272)*x[0]**o+(-0.198121930351)*x[0]+(0.458180780997)*x[1]**o+(-0.415159831726)*x[1]
        s[1,3,2,1]=(-0.511099937481)*x[0]**o+(0.297820678198)*x[0]+(0.908702936015)*x[1]**o+(0.965700845817)*x[1]
        s[1,4,0,0]=(0.484158989021)*x[0]**o+(-0.339968620521)*x[0]+(-0.213631360058)*x[1]**o+(0.0279496812648)*x[1]
        s[1,4,0,1]=(-0.217367899821)*x[0]**o+(0.433458086428)*x[0]+(0.244095485621)*x[1]**o+(-0.210532522829)*x[1]
        s[1,4,1,0]=(-0.254889249994)*x[0]**o+(-0.61248326904)*x[0]+(0.372253508249)*x[1]**o+(-0.642709327921)*x[1]
        s[1,4,1,1]=(0.880474654934)*x[0]**o+(0.68839232824)*x[0]+(0.189353506343)*x[1]**o+(-0.880653838654)*x[1]
        s[1,4,2,0]=(0.894790715592)*x[0]**o+(-0.0486173810442)*x[0]+(-0.246550765358)*x[1]**o+(-0.343955068452)*x[1]
        s[1,4,2,1]=(-0.801075573104)*x[0]**o+(-0.749933700542)*x[0]+(0.527976119218)*x[1]**o+(-0.671158980138)*x[1]
        s[2,0,0,0]=(0.114632100765)*x[0]**o+(0.61717357379)*x[0]+(-0.0549019501391)*x[1]**o+(-0.749508575346)*x[1]
        s[2,0,0,1]=(-0.130970404194)*x[0]**o+(0.707199821591)*x[0]+(-0.903645272692)*x[1]**o+(-0.0334328361989)*x[1]
        s[2,0,1,0]=(-0.715712763894)*x[0]**o+(-0.675427980461)*x[0]+(-0.366134531475)*x[1]**o+(0.201815564551)*x[1]
        s[2,0,1,1]=(-0.269870188604)*x[0]**o+(-0.147598844928)*x[0]+(0.677047514973)*x[1]**o+(-0.675438319339)*x[1]
        s[2,0,2,0]=(0.212062724301)*x[0]**o+(-0.0250868236126)*x[0]+(-0.100922604124)*x[1]**o+(0.427937448863)*x[1]
        s[2,0,2,1]=(-0.666239860886)*x[0]**o+(0.438278520576)*x[0]+(0.544120816825)*x[1]**o+(-0.473903266225)*x[1]
        s[2,1,0,0]=(-0.361166717507)*x[0]**o+(0.901751988241)*x[0]+(0.638709648231)*x[1]**o+(-0.235793145944)*x[1]
        s[2,1,0,1]=(-0.11628481466)*x[0]**o+(-0.0851943604844)*x[0]+(0.832299993845)*x[1]**o+(-0.619236617136)*x[1]
        s[2,1,1,0]=(-0.960053497178)*x[0]**o+(-0.455025261633)*x[0]+(-0.701715648299)*x[1]**o+(0.728639782857)*x[1]
        s[2,1,1,1]=(-0.533289133278)*x[0]**o+(-0.370102308202)*x[0]+(0.99106398096)*x[1]**o+(-0.0538096992941)*x[1]
        s[2,1,2,0]=(0.224553214823)*x[0]**o+(-0.500053815427)*x[0]+(-0.0704009106671)*x[1]**o+(0.0488366166495)*x[1]
        s[2,1,2,1]=(0.337180739675)*x[0]**o+(-0.981874806959)*x[0]+(-0.731568712654)*x[1]**o+(0.0847072795393)*x[1]
        s[2,2,0,0]=(-0.222325355327)*x[0]**o+(-0.276984511347)*x[0]+(-0.948381537292)*x[1]**o+(0.766624524385)*x[1]
        s[2,2,0,1]=(0.938320523187)*x[0]**o+(0.808908551771)*x[0]+(0.765330540739)*x[1]**o+(0.41183946621)*x[1]
        s[2,2,1,0]=(-0.356223597296)*x[0]**o+(0.898624256746)*x[0]+(0.491456887622)*x[1]**o+(0.454642872239)*x[1]
        s[2,2,1,1]=(0.70662625865)*x[0]**o+(0.0268161060988)*x[0]+(-0.17883477863)*x[1]**o+(-0.379062708108)*x[1]
        s[2,2,2,0]=(0.405231213751)*x[0]**o+(-0.339541044657)*x[0]+(0.92369283669)*x[1]**o+(-0.521296842329)*x[1]
        s[2,2,2,1]=(0.371621085625)*x[0]**o+(-0.277933239024)*x[0]+(0.99545384432)*x[1]**o+(-0.458904147226)*x[1]
        s[2,3,0,0]=(-0.978371890906)*x[0]**o+(-0.220612890504)*x[0]+(0.485724985276)*x[1]**o+(0.566941748451)*x[1]
        s[2,3,0,1]=(-0.784455960641)*x[0]**o+(0.488256734495)*x[0]+(-0.85203669445)*x[1]**o+(-0.573678796739)*x[1]
        s[2,3,1,0]=(-0.083067317174)*x[0]**o+(0.284336362391)*x[0]+(-0.671352852822)*x[1]**o+(0.585391001713)*x[1]
        s[2,3,1,1]=(0.453596080554)*x[0]**o+(0.382646103643)*x[0]+(0.976202291904)*x[1]**o+(0.269885074758)*x[1]
        s[2,3,2,0]=(0.58278560743)*x[0]**o+(-0.641861358694)*x[0]+(0.65546441735)*x[1]**o+(-0.510301705975)*x[1]
        s[2,3,2,1]=(-0.462458062923)*x[0]**o+(-0.860624656355)*x[0]+(-0.795012311004)*x[1]**o+(-0.945015232064)*x[1]
        s[2,4,0,0]=(0.373515573367)*x[0]**o+(0.429646026706)*x[0]+(0.107731590254)*x[1]**o+(0.404788302157)*x[1]
        s[2,4,0,1]=(0.922575542165)*x[0]**o+(0.990525976075)*x[0]+(-0.920047278057)*x[1]**o+(-0.606366502396)*x[1]
        s[2,4,1,0]=(-0.395749593133)*x[0]**o+(-0.303444216059)*x[0]+(0.279159896939)*x[1]**o+(-0.998124949675)*x[1]
        s[2,4,1,1]=(0.608308919638)*x[0]**o+(0.547633707859)*x[0]+(-0.335546391995)*x[1]**o+(-0.49712570743)*x[1]
        s[2,4,2,0]=(-0.764049221842)*x[0]**o+(-0.691060860074)*x[0]+(0.576144131381)*x[1]**o+(-0.0692145325011)*x[1]
        s[2,4,2,1]=(-0.445193960624)*x[0]**o+(-0.299673373394)*x[0]+(-0.17618349631)*x[1]**o+(0.6882025213)*x[1]
        s[3,0,0,0]=(-0.363639361841)*x[0]**o+(-0.637031005499)*x[0]+(0.750565431989)*x[1]**o+(-0.713309519647)*x[1]
        s[3,0,0,1]=(0.86686475536)*x[0]**o+(0.857916166663)*x[0]+(-0.233119477582)*x[1]**o+(-0.275718321599)*x[1]
        s[3,0,1,0]=(0.676379750639)*x[0]**o+(-0.221464786801)*x[0]+(-0.307372968314)*x[1]**o+(-0.974878791542)*x[1]
        s[3,0,1,1]=(0.963332442265)*x[0]**o+(0.873237489049)*x[0]+(0.347621440028)*x[1]**o+(-0.447236045492)*x[1]
        s[3,0,2,0]=(0.574231383242)*x[0]**o+(0.406006911504)*x[0]+(0.998282029235)*x[1]**o+(-0.713526415934)*x[1]
        s[3,0,2,1]=(0.130477890591)*x[0]**o+(-0.994273713498)*x[0]+(-0.836645831178)*x[1]**o+(0.23964737976)*x[1]
        s[3,1,0,0]=(0.681027983965)*x[0]**o+(0.277762028836)*x[0]+(-0.476680449158)*x[1]**o+(0.863218206262)*x[1]
        s[3,1,0,1]=(-0.551857608855)*x[0]**o+(0.714701927775)*x[0]+(0.0805460030018)*x[1]**o+(-0.75844120919)*x[1]
        s[3,1,1,0]=(-0.699952165487)*x[0]**o+(-0.829498201164)*x[0]+(0.518643791102)*x[1]**o+(0.748348341191)*x[1]
        s[3,1,1,1]=(0.628926227064)*x[0]**o+(0.756291174101)*x[0]+(0.847249240418)*x[1]**o+(-0.694296012132)*x[1]
        s[3,1,2,0]=(0.300066092357)*x[0]**o+(0.0979478221525)*x[0]+(-0.652853644798)*x[1]**o+(0.87892206219)*x[1]
        s[3,1,2,1]=(0.820578435316)*x[0]**o+(-0.151267836854)*x[0]+(0.288662297283)*x[1]**o+(-0.530205788227)*x[1]
        s[3,2,0,0]=(-0.262081376064)*x[0]**o+(0.564488927009)*x[0]+(-0.574271798456)*x[1]**o+(-0.103341985058)*x[1]
        s[3,2,0,1]=(0.264373345575)*x[0]**o+(0.505975223545)*x[0]+(0.760869432805)*x[1]**o+(-0.348245510796)*x[1]
        s[3,2,1,0]=(0.382085638525)*x[0]**o+(0.271178748741)*x[0]+(-0.344266263181)*x[1]**o+(0.949904496331)*x[1]
        s[3,2,1,1]=(0.570382182257)*x[0]**o+(-0.507665638153)*x[0]+(0.675674419969)*x[1]**o+(-0.741319149319)*x[1]
        s[3,2,2,0]=(-0.0159674521901)*x[0]**o+(-0.201449258763)*x[0]+(-0.774168601079)*x[1]**o+(-0.333025684689)*x[1]
        s[3,2,2,1]=(0.21116026219)*x[0]**o+(0.174264855243)*x[0]+(-0.693006543432)*x[1]**o+(-0.117768474854)*x[1]
        s[3,3,0,0]=(-0.0965568139991)*x[0]**o+(-0.305779611724)*x[0]+(-0.905575174356)*x[1]**o+(-0.0438869174233)*x[1]
        s[3,3,0,1]=(-0.517028280471)*x[0]**o+(0.725783549398)*x[0]+(0.346469604255)*x[1]**o+(-0.958724910661)*x[1]
        s[3,3,1,0]=(0.673217439387)*x[0]**o+(0.171407887035)*x[0]+(-0.931242988664)*x[1]**o+(0.905785986466)*x[1]
        s[3,3,1,1]=(-0.806309377174)*x[0]**o+(-0.39154927098)*x[0]+(-0.353245837689)*x[1]**o+(-0.0891421463801)*x[1]
        s[3,3,2,0]=(0.54365433687)*x[0]**o+(0.328598607479)*x[0]+(-0.134078804517)*x[1]**o+(0.368297327809)*x[1]
        s[3,3,2,1]=(0.309908919248)*x[0]**o+(-0.366081335409)*x[0]+(0.287094150861)*x[1]**o+(-0.311725137682)*x[1]
        s[3,4,0,0]=(0.0350561429151)*x[0]**o+(-0.587757161762)*x[0]+(0.105785456873)*x[1]**o+(0.00732306049631)*x[1]
        s[3,4,0,1]=(0.110213440973)*x[0]**o+(-0.429550096051)*x[0]+(0.471367364287)*x[1]**o+(0.580191251114)*x[1]
        s[3,4,1,0]=(0.349823338156)*x[0]**o+(0.474522056666)*x[0]+(-0.190656365622)*x[1]**o+(0.872370974782)*x[1]
        s[3,4,1,1]=(-0.694637797093)*x[0]**o+(0.648769674687)*x[0]+(-0.0962824527574)*x[1]**o+(0.669702141463)*x[1]
        s[3,4,2,0]=(-0.954369212248)*x[0]**o+(0.558821994909)*x[0]+(-0.196151296515)*x[1]**o+(0.475540095693)*x[1]
        s[3,4,2,1]=(0.946611204124)*x[0]**o+(-0.452828842027)*x[0]+(-0.167471954345)*x[1]**o+(0.0326624440493)*x[1]
        ref[0,0,0,0]=(0.246025700812)/(o+1.)+(-0.0700771278147)
        ref[0,0,0,1]=(0.872903965104)/(o+1.)+(-0.0956768400749)
        ref[0,0,1,0]=(-0.149098784473)/(o+1.)+(-0.271683908641)
        ref[0,0,1,1]=(0.813191985058)/(o+1.)+(0.141471386383)
        ref[0,0,2,0]=(-0.554184175721)/(o+1.)+(-0.621662970196)
        ref[0,0,2,1]=(-1.0295469585)/(o+1.)+(-0.0365685105983)
        ref[0,1,0,0]=(0.151804025931)/(o+1.)+(-0.212298031588)
        ref[0,1,0,1]=(-0.533004796234)/(o+1.)+(0.0485769377937)
        ref[0,1,1,0]=(-0.802375632463)/(o+1.)+(0.127439034441)
        ref[0,1,1,1]=(0.325754839785)/(o+1.)+(0.0972311868531)
        ref[0,1,2,0]=(1.43492852092)/(o+1.)+(0.652296741316)
        ref[0,1,2,1]=(-0.347667591981)/(o+1.)+(-0.160364561879)
        ref[0,2,0,0]=(-1.37099731406)/(o+1.)+(0.0563097354051)
        ref[0,2,0,1]=(1.11251425188)/(o+1.)+(-0.835019726344)
        ref[0,2,1,0]=(-0.770103190572)/(o+1.)+(-0.309897780967)
        ref[0,2,1,1]=(0.159272551388)/(o+1.)+(0.706859784466)
        ref[0,2,2,0]=(-0.391469887349)/(o+1.)+(-0.207974163745)
        ref[0,2,2,1]=(-0.0415688115757)/(o+1.)+(0.0604937496788)
        ref[0,3,0,0]=(0.957559739626)/(o+1.)+(0.228032778194)
        ref[0,3,0,1]=(-1.09130803461)/(o+1.)+(-0.662860500606)
        ref[0,3,1,0]=(-0.449862052638)/(o+1.)+(-0.344594323242)
        ref[0,3,1,1]=(0.219326920673)/(o+1.)+(0.405011796484)
        ref[0,3,2,0]=(0.275993808116)/(o+1.)+(-0.461328543451)
        ref[0,3,2,1]=(-0.575033733717)/(o+1.)+(0.884920083202)
        ref[0,4,0,0]=(0.0253483143264)/(o+1.)+(-0.831762460572)
        ref[0,4,0,1]=(-0.593504118306)/(o+1.)+(-0.0899796107363)
        ref[0,4,1,0]=(0.946442680954)/(o+1.)+(0.340471820786)
        ref[0,4,1,1]=(1.00832461793)/(o+1.)+(-0.489125179816)
        ref[0,4,2,0]=(0.658258894209)/(o+1.)+(-0.252225332525)
        ref[0,4,2,1]=(0.190670120044)/(o+1.)+(0.107482927026)
        ref[1,0,0,0]=(-0.638264699882)/(o+1.)+(-0.37905912763)
        ref[1,0,0,1]=(0.858273041153)/(o+1.)+(-0.602359980261)
        ref[1,0,1,0]=(1.19603458122)/(o+1.)+(0.623979052337)
        ref[1,0,1,1]=(-0.0559871855173)/(o+1.)+(0.392163437219)
        ref[1,0,2,0]=(0.202869107651)/(o+1.)+(-0.327529727939)
        ref[1,0,2,1]=(0.497839996627)/(o+1.)+(-0.198257844767)
        ref[1,1,0,0]=(-1.05743683429)/(o+1.)+(-0.107408196452)
        ref[1,1,0,1]=(-1.11022578935)/(o+1.)+(0.0799654739886)
        ref[1,1,1,0]=(1.02865615322)/(o+1.)+(-0.111339912553)
        ref[1,1,1,1]=(-0.387124166332)/(o+1.)+(-0.13196732944)
        ref[1,1,2,0]=(0.364571348315)/(o+1.)+(-0.827711953327)
        ref[1,1,2,1]=(0.800389584904)/(o+1.)+(-0.570167855757)
        ref[1,2,0,0]=(-0.359885437423)/(o+1.)+(-0.683824166857)
        ref[1,2,0,1]=(-0.139489334179)/(o+1.)+(-0.0550294716506)
        ref[1,2,1,0]=(0.05752529846)/(o+1.)+(0.330669314498)
        ref[1,2,1,1]=(-0.68363897434)/(o+1.)+(-0.210682497577)
        ref[1,2,2,0]=(-0.971569775609)/(o+1.)+(0.258047639988)
        ref[1,2,2,1]=(0.966753430217)/(o+1.)+(0.737154900402)
        ref[1,3,0,0]=(0.571230899607)/(o+1.)+(0.201680592513)
        ref[1,3,0,1]=(-0.516278846659)/(o+1.)+(-0.203543442922)
        ref[1,3,1,0]=(-0.276019031326)/(o+1.)+(0.465273316758)
        ref[1,3,1,1]=(0.339281660193)/(o+1.)+(0.578236730598)
        ref[1,3,2,0]=(0.811123220269)/(o+1.)+(-0.306640881038)
        ref[1,3,2,1]=(0.397602998535)/(o+1.)+(0.631760762007)
        ref[1,4,0,0]=(0.270527628963)/(o+1.)+(-0.156009469628)
        ref[1,4,0,1]=(0.0267275858003)/(o+1.)+(0.1114627818)
        ref[1,4,1,0]=(0.117364258255)/(o+1.)+(-0.627596298481)
        ref[1,4,1,1]=(1.06982816128)/(o+1.)+(-0.0961307552069)
        ref[1,4,2,0]=(0.648239950234)/(o+1.)+(-0.196286224748)
        ref[1,4,2,1]=(-0.273099453886)/(o+1.)+(-0.71054634034)
        ref[2,0,0,0]=(0.0597301506259)/(o+1.)+(-0.0661675007782)
        ref[2,0,0,1]=(-1.03461567689)/(o+1.)+(0.336883492696)
        ref[2,0,1,0]=(-1.08184729537)/(o+1.)+(-0.236806207955)
        ref[2,0,1,1]=(0.407177326369)/(o+1.)+(-0.411518582134)
        ref[2,0,2,0]=(0.111140120177)/(o+1.)+(0.201425312625)
        ref[2,0,2,1]=(-0.122119044061)/(o+1.)+(-0.0178123728248)
        ref[2,1,0,0]=(0.277542930723)/(o+1.)+(0.332979421149)
        ref[2,1,0,1]=(0.716015179185)/(o+1.)+(-0.35221548881)
        ref[2,1,1,0]=(-1.66176914548)/(o+1.)+(0.136807260612)
        ref[2,1,1,1]=(0.457774847682)/(o+1.)+(-0.211956003748)
        ref[2,1,2,0]=(0.154152304156)/(o+1.)+(-0.225608599389)
        ref[2,1,2,1]=(-0.394387972979)/(o+1.)+(-0.44858376371)
        ref[2,2,0,0]=(-1.17070689262)/(o+1.)+(0.244820006519)
        ref[2,2,0,1]=(1.70365106393)/(o+1.)+(0.61037400899)
        ref[2,2,1,0]=(0.135233290326)/(o+1.)+(0.676633564493)
        ref[2,2,1,1]=(0.527791480019)/(o+1.)+(-0.176123301004)
        ref[2,2,2,0]=(1.32892405044)/(o+1.)+(-0.430418943493)
        ref[2,2,2,1]=(1.36707492995)/(o+1.)+(-0.368418693125)
        ref[2,3,0,0]=(-0.492646905629)/(o+1.)+(0.173164428973)
        ref[2,3,0,1]=(-1.63649265509)/(o+1.)+(-0.0427110311218)
        ref[2,3,1,0]=(-0.754420169996)/(o+1.)+(0.434863682052)
        ref[2,3,1,1]=(1.42979837246)/(o+1.)+(0.3262655892)
        ref[2,3,2,0]=(1.23825002478)/(o+1.)+(-0.576081532335)
        ref[2,3,2,1]=(-1.25747037393)/(o+1.)+(-0.902819944209)
        ref[2,4,0,0]=(0.48124716362)/(o+1.)+(0.417217164431)
        ref[2,4,0,1]=(0.00252826410759)/(o+1.)+(0.192079736839)
        ref[2,4,1,0]=(-0.116589696194)/(o+1.)+(-0.650784582867)
        ref[2,4,1,1]=(0.272762527643)/(o+1.)+(0.0252540002143)
        ref[2,4,2,0]=(-0.187905090461)/(o+1.)+(-0.380137696288)
        ref[2,4,2,1]=(-0.621377456934)/(o+1.)+(0.194264573953)
        ref[3,0,0,0]=(0.386926070148)/(o+1.)+(-0.675170262573)
        ref[3,0,0,1]=(0.633745277778)/(o+1.)+(0.291098922532)
        ref[3,0,1,0]=(0.369006782325)/(o+1.)+(-0.598171789171)
        ref[3,0,1,1]=(1.31095388229)/(o+1.)+(0.213000721778)
        ref[3,0,2,0]=(1.57251341248)/(o+1.)+(-0.153759752215)
        ref[3,0,2,1]=(-0.706167940587)/(o+1.)+(-0.377313166869)
        ref[3,1,0,0]=(0.204347534807)/(o+1.)+(0.570490117549)
        ref[3,1,0,1]=(-0.471311605853)/(o+1.)+(-0.0218696407078)
        ref[3,1,1,0]=(-0.181308374385)/(o+1.)+(-0.0405749299865)
        ref[3,1,1,1]=(1.47617546748)/(o+1.)+(0.0309975809845)
        ref[3,1,2,0]=(-0.352787552441)/(o+1.)+(0.488434942171)
        ref[3,1,2,1]=(1.1092407326)/(o+1.)+(-0.34073681254)
        ref[3,2,0,0]=(-0.83635317452)/(o+1.)+(0.230573470976)
        ref[3,2,0,1]=(1.02524277838)/(o+1.)+(0.0788648563748)
        ref[3,2,1,0]=(0.0378193753437)/(o+1.)+(0.610541622536)
        ref[3,2,1,1]=(1.24605660223)/(o+1.)+(-0.624492393736)
        ref[3,2,2,0]=(-0.79013605327)/(o+1.)+(-0.267237471726)
        ref[3,2,2,1]=(-0.481846281241)/(o+1.)+(0.0282481901946)
        ref[3,3,0,0]=(-1.00213198835)/(o+1.)+(-0.174833264573)
        ref[3,3,0,1]=(-0.170558676217)/(o+1.)+(-0.116470680631)
        ref[3,3,1,0]=(-0.258025549277)/(o+1.)+(0.538596936751)
        ref[3,3,1,1]=(-1.15955521486)/(o+1.)+(-0.24034570868)
        ref[3,3,2,0]=(0.409575532352)/(o+1.)+(0.348447967644)
        ref[3,3,2,1]=(0.597003070109)/(o+1.)+(-0.338903236546)
        ref[3,4,0,0]=(0.140841599788)/(o+1.)+(-0.290217050633)
        ref[3,4,0,1]=(0.581580805261)/(o+1.)+(0.0753205775318)
        ref[3,4,1,0]=(0.159166972534)/(o+1.)+(0.673446515724)
        ref[3,4,1,1]=(-0.79092024985)/(o+1.)+(0.659235908075)
        ref[3,4,2,0]=(-1.15052050876)/(o+1.)+(0.517181045301)
        ref[3,4,2,1]=(0.779139249778)/(o+1.)+(-0.210083198989)
      else:
        s[0,0,0,0]=(0.748319783349)*x[0]**o+(-0.176566570347)*x[0]+(-0.302910837408)*x[1]**o+(0.492350205415)*x[1]+(-0.160456161783)*x[2]**o+(0.868280836814)*x[2]
        s[0,0,0,1]=(-0.51136577317)*x[0]**o+(-0.242417336787)*x[0]+(-0.785996766849)*x[1]**o+(0.606869170308)*x[1]+(0.49391109913)*x[2]**o+(-0.319568202304)*x[2]
        s[0,0,1,0]=(0.640062862527)*x[0]**o+(0.566706264858)*x[0]+(0.907582097741)*x[1]**o+(0.350910171053)*x[1]+(-0.909242439832)*x[2]**o+(-0.85540876524)*x[2]
        s[0,0,1,1]=(-0.377548918631)*x[0]**o+(-0.630218128454)*x[0]+(0.45141559373)*x[1]**o+(0.0832708083581)*x[1]+(0.233405447853)*x[2]**o+(0.567903425032)*x[2]
        s[0,0,2,0]=(-0.497312249883)*x[0]**o+(-0.651244043641)*x[0]+(-0.575450843765)*x[1]**o+(0.0219939930053)*x[1]+(0.570655374403)*x[2]**o+(-0.402040210721)*x[2]
        s[0,0,2,1]=(0.383694368889)*x[0]**o+(-0.280765736983)*x[0]+(0.644560587428)*x[1]**o+(-0.263879759253)*x[1]+(0.78943504044)*x[2]**o+(0.838120969604)*x[2]
        s[0,1,0,0]=(0.331871753344)*x[0]**o+(0.731388799379)*x[0]+(0.394440515222)*x[1]**o+(-0.337910220731)*x[1]+(-0.078492234799)*x[2]**o+(0.388988268348)*x[2]
        s[0,1,0,1]=(0.197262971559)*x[0]**o+(0.482585248548)*x[0]+(0.0890475484335)*x[1]**o+(-0.154881471591)*x[1]+(0.14978160538)*x[2]**o+(-0.115243504627)*x[2]
        s[0,1,1,0]=(0.338072090662)*x[0]**o+(0.238967458874)*x[0]+(0.4078504965)*x[1]**o+(-0.941497950412)*x[1]+(0.539654579256)*x[2]**o+(0.777502770306)*x[2]
        s[0,1,1,1]=(-0.876894224774)*x[0]**o+(-0.413741121497)*x[0]+(0.841988144013)*x[1]**o+(-0.987067492493)*x[1]+(-0.0142647738806)*x[2]**o+(0.809756634084)*x[2]
        s[0,1,2,0]=(0.771962423748)*x[0]**o+(-0.274151695697)*x[0]+(-0.0583329632543)*x[1]**o+(0.290755268126)*x[1]+(0.986842038664)*x[2]**o+(-0.445513638841)*x[2]
        s[0,1,2,1]=(0.235289294125)*x[0]**o+(-0.0230216115543)*x[0]+(-0.39449823454)*x[1]**o+(0.758538375985)*x[1]+(-0.990485958396)*x[2]**o+(0.441787118653)*x[2]
        s[0,2,0,0]=(-0.860580472907)*x[0]**o+(-0.480513252563)*x[0]+(0.388567023408)*x[1]**o+(-0.435900869228)*x[1]+(-0.483812008458)*x[2]**o+(0.87821418943)*x[2]
        s[0,2,0,1]=(-0.391914341084)*x[0]**o+(-0.813798019851)*x[0]+(0.681793443132)*x[1]**o+(-0.422727355552)*x[1]+(-0.0599492576813)*x[2]**o+(0.630302059584)*x[2]
        s[0,2,1,0]=(-0.857349416782)*x[0]**o+(0.0937767571422)*x[0]+(-0.935044609271)*x[1]**o+(-0.825241166723)*x[1]+(-0.78223171556)*x[2]**o+(0.33550368528)*x[2]
        s[0,2,1,1]=(-0.366924211693)*x[0]**o+(0.572797118425)*x[0]+(-0.289194982186)*x[1]**o+(-0.942701172085)*x[1]+(0.614219031682)*x[2]**o+(-0.389753632964)*x[2]
        s[0,2,2,0]=(-0.404732972636)*x[0]**o+(0.631781218361)*x[0]+(-0.263486337549)*x[1]**o+(0.631746356812)*x[1]+(-0.515291196501)*x[2]**o+(-0.363991020547)*x[2]
        s[0,2,2,1]=(0.892685628814)*x[0]**o+(0.963218649844)*x[0]+(0.483947727789)*x[1]**o+(0.211028649587)*x[1]+(0.50159178018)*x[2]**o+(0.0711392272561)*x[2]
        s[0,3,0,0]=(-0.33216085676)*x[0]**o+(0.906269165577)*x[0]+(0.874123825651)*x[1]**o+(-0.570808752482)*x[1]+(-0.561182318614)*x[2]**o+(-0.884125641249)*x[2]
        s[0,3,0,1]=(-0.0703055529684)*x[0]**o+(-0.101044608869)*x[0]+(0.829269084242)*x[1]**o+(-0.497613621381)*x[1]+(0.297140846283)*x[2]**o+(0.566992550301)*x[2]
        s[0,3,1,0]=(0.00230797481965)*x[0]**o+(-0.561978070323)*x[0]+(-0.874987267344)*x[1]**o+(0.404602330516)*x[1]+(0.541390115931)*x[2]**o+(0.109832470538)*x[2]
        s[0,3,1,1]=(-0.738073409216)*x[0]**o+(0.016899800363)*x[0]+(0.606773214528)*x[1]**o+(0.607363930641)*x[1]+(-0.755469399679)*x[2]**o+(-0.585951709489)*x[2]
        s[0,3,2,0]=(0.754566844048)*x[0]**o+(0.353623703809)*x[0]+(0.300465743201)*x[1]**o+(0.4804155825)*x[1]+(-0.500709775706)*x[2]**o+(-0.184866087487)*x[2]
        s[0,3,2,1]=(0.656642762895)*x[0]**o+(-0.522249569275)*x[0]+(-0.689897112218)*x[1]**o+(0.197660165382)*x[1]+(0.790618474391)*x[2]**o+(0.34534294776)*x[2]
        s[0,4,0,0]=(0.951780164354)*x[0]**o+(0.227261667276)*x[0]+(-0.388326921756)*x[1]**o+(-0.143812077493)*x[1]+(0.869741228576)*x[2]**o+(0.896296431361)*x[2]
        s[0,4,0,1]=(0.0418495997166)*x[0]**o+(0.378737546698)*x[0]+(-0.893460535382)*x[1]**o+(-0.0409222699203)*x[1]+(-0.6955627414)*x[2]**o+(-0.993500126323)*x[2]
        s[0,4,1,0]=(-0.47303219937)*x[0]**o+(0.488878345444)*x[0]+(0.110295068747)*x[1]**o+(-0.775897017507)*x[1]+(-0.664013808104)*x[2]**o+(-0.125758726632)*x[2]
        s[0,4,1,1]=(0.326490818022)*x[0]**o+(-0.785954928184)*x[0]+(0.85336153105)*x[1]**o+(-0.573522941199)*x[1]+(-0.42185352182)*x[2]**o+(-0.425049668912)*x[2]
        s[0,4,2,0]=(0.290783142101)*x[0]**o+(-0.502056588268)*x[0]+(0.433355340131)*x[1]**o+(-0.717832338133)*x[1]+(0.254866069131)*x[2]**o+(-0.383849864093)*x[2]
        s[0,4,2,1]=(-0.796340985968)*x[0]**o+(-0.732115333195)*x[0]+(-0.829732009003)*x[1]**o+(-0.790621874069)*x[1]+(-0.60818338505)*x[2]**o+(-0.892012123169)*x[2]
        s[1,0,0,0]=(0.150096606609)*x[0]**o+(0.713142060809)*x[0]+(0.0898795205212)*x[1]**o+(-0.682356428859)*x[1]+(0.275326716654)*x[2]**o+(-0.356136748426)*x[2]
        s[1,0,0,1]=(0.104574120129)*x[0]**o+(-0.895395804614)*x[0]+(-0.00895409609851)*x[1]**o+(0.852200905159)*x[1]+(0.25757145031)*x[2]**o+(-0.578299505429)*x[2]
        s[1,0,1,0]=(-0.83104126362)*x[0]**o+(-0.587286782794)*x[0]+(0.043755441563)*x[1]**o+(-0.357299738488)*x[1]+(0.900592455437)*x[2]**o+(0.885645422129)*x[2]
        s[1,0,1,1]=(-0.144444995355)*x[0]**o+(-0.667350132103)*x[0]+(0.949183394884)*x[1]**o+(-0.519086948324)*x[1]+(-0.74215320159)*x[2]**o+(-0.202102401774)*x[2]
        s[1,0,2,0]=(0.303496934218)*x[0]**o+(-0.02461446864)*x[0]+(-0.0867456086666)*x[1]**o+(0.730962877622)*x[1]+(-0.851607522381)*x[2]**o+(-0.239310033319)*x[2]
        s[1,0,2,1]=(-0.512139821912)*x[0]**o+(-0.69403755459)*x[0]+(0.462840604615)*x[1]**o+(-0.192161565941)*x[1]+(0.807307771096)*x[2]**o+(0.440024157111)*x[2]
        s[1,1,0,0]=(-0.693987625939)*x[0]**o+(0.410652593337)*x[0]+(0.665893670901)*x[1]**o+(0.525886124327)*x[1]+(0.200291951309)*x[2]**o+(0.793353011987)*x[2]
        s[1,1,0,1]=(-0.667911914141)*x[0]**o+(0.693960372567)*x[0]+(-0.0892935574405)*x[1]**o+(0.669184962503)*x[1]+(0.664202374165)*x[2]**o+(-0.474657421474)*x[2]
        s[1,1,1,0]=(0.4341035606)*x[0]**o+(0.707815942288)*x[0]+(-0.204720100594)*x[1]**o+(-0.816120724182)*x[1]+(-0.259566839542)*x[2]**o+(0.918010553929)*x[2]
        s[1,1,1,1]=(0.280925026809)*x[0]**o+(0.895509529721)*x[0]+(0.244175246656)*x[1]**o+(0.692059608476)*x[1]+(-0.424963623478)*x[2]**o+(0.985618462184)*x[2]
        s[1,1,2,0]=(0.172285590575)*x[0]**o+(0.0614208586474)*x[0]+(-0.751024157153)*x[1]**o+(0.182364162229)*x[1]+(-0.400270263752)*x[2]**o+(0.6410957443)*x[2]
        s[1,1,2,1]=(0.621541338202)*x[0]**o+(0.716601026013)*x[0]+(0.822288477475)*x[1]**o+(0.200859656154)*x[1]+(-0.651437159651)*x[2]**o+(-0.554048539377)*x[2]
        s[1,2,0,0]=(-0.835149004101)*x[0]**o+(-0.748380867211)*x[0]+(-0.445526738596)*x[1]**o+(0.782304294095)*x[1]+(0.827204947885)*x[2]**o+(0.188779249069)*x[2]
        s[1,2,0,1]=(0.354866953137)*x[0]**o+(-0.981132219204)*x[0]+(0.488043961513)*x[1]**o+(0.257052078929)*x[1]+(0.719111294043)*x[2]**o+(-0.298289862834)*x[2]
        s[1,2,1,0]=(0.552605101038)*x[0]**o+(0.145016605474)*x[0]+(0.119161182879)*x[1]**o+(0.723703065293)*x[1]+(-0.011955307053)*x[2]**o+(0.374814299212)*x[2]
        s[1,2,1,1]=(0.448055014229)*x[0]**o+(-0.320498330109)*x[0]+(0.812977543935)*x[1]**o+(0.374118976348)*x[1]+(0.656782601024)*x[2]**o+(0.0845883407242)*x[2]
        s[1,2,2,0]=(0.418470977027)*x[0]**o+(-0.324185586615)*x[0]+(0.576184154734)*x[1]**o+(-0.244791711079)*x[1]+(-0.383448002434)*x[2]**o+(-0.425391146295)*x[2]
        s[1,2,2,1]=(0.885670287748)*x[0]**o+(0.804245948464)*x[0]+(-0.122365449147)*x[1]**o+(0.0526689576387)*x[1]+(0.886983215847)*x[2]**o+(-0.9130795011)*x[2]
        s[1,3,0,0]=(-0.698202951872)*x[0]**o+(-0.753867598728)*x[0]+(-0.729720876147)*x[1]**o+(-0.535739350698)*x[1]+(0.140570380613)*x[2]**o+(-0.0716035886696)*x[2]
        s[1,3,0,1]=(-0.898459992298)*x[0]**o+(0.372462814699)*x[0]+(0.605465414867)*x[1]**o+(-0.684605485569)*x[1]+(0.549437793056)*x[2]**o+(-0.010737341414)*x[2]
        s[1,3,1,0]=(0.792329973599)*x[0]**o+(-0.52685018301)*x[0]+(0.200314760956)*x[1]**o+(0.632873255355)*x[1]+(-0.43152569425)*x[2]**o+(0.348904687244)*x[2]
        s[1,3,1,1]=(0.209464885105)*x[0]**o+(-0.0146787858502)*x[0]+(0.320235817099)*x[1]**o+(0.86370910616)*x[1]+(-0.202145495919)*x[2]**o+(-0.25657518419)*x[2]
        s[1,3,2,0]=(0.489051201235)*x[0]**o+(0.349594264623)*x[0]+(-0.574002000861)*x[1]**o+(0.390761344969)*x[1]+(0.1514845518)*x[2]**o+(0.802240236468)*x[2]
        s[1,3,2,1]=(-0.0182530256393)*x[0]**o+(0.303828930846)*x[0]+(0.618612491362)*x[1]**o+(0.293473693133)*x[1]+(0.763201722939)*x[2]**o+(-0.652537522965)*x[2]
        s[1,4,0,0]=(0.349015278137)*x[0]**o+(-0.0165197033193)*x[0]+(0.63228324617)*x[1]**o+(-0.934745316284)*x[1]+(-0.187390369535)*x[2]**o+(-0.456599706963)*x[2]
        s[1,4,0,1]=(0.310212294648)*x[0]**o+(-0.898896972668)*x[0]+(0.229475812326)*x[1]**o+(0.787159228942)*x[1]+(0.133488840506)*x[2]**o+(0.949893906424)*x[2]
        s[1,4,1,0]=(-0.869667859962)*x[0]**o+(0.624210860706)*x[0]+(0.122085293209)*x[1]**o+(-0.624604470463)*x[1]+(-0.878410325648)*x[2]**o+(-0.0841801197816)*x[2]
        s[1,4,1,1]=(0.706521970977)*x[0]**o+(0.23837086782)*x[0]+(-0.960682169998)*x[1]**o+(-0.707493453388)*x[1]+(0.828048621994)*x[2]**o+(0.327258769743)*x[2]
        s[1,4,2,0]=(-0.291203807685)*x[0]**o+(0.707463915698)*x[0]+(-0.860720761064)*x[1]**o+(0.802477835676)*x[1]+(-0.16905212173)*x[2]**o+(0.453497949714)*x[2]
        s[1,4,2,1]=(0.789863893897)*x[0]**o+(-0.945472074403)*x[0]+(-0.047780484871)*x[1]**o+(0.199167500985)*x[1]+(0.575959160298)*x[2]**o+(-0.181712597239)*x[2]
        s[2,0,0,0]=(0.61569741708)*x[0]**o+(0.609998488119)*x[0]+(-0.943339902899)*x[1]**o+(-0.962332686128)*x[1]+(0.458927713248)*x[2]**o+(-0.191444752971)*x[2]
        s[2,0,0,1]=(0.0306432921503)*x[0]**o+(0.326065705299)*x[0]+(0.29054908043)*x[1]**o+(-0.498110577924)*x[1]+(-0.855604413911)*x[2]**o+(-0.832549081863)*x[2]
        s[2,0,1,0]=(0.539469235745)*x[0]**o+(0.549053450527)*x[0]+(0.379811452156)*x[1]**o+(0.383008863276)*x[1]+(-0.835777000108)*x[2]**o+(0.418879791451)*x[2]
        s[2,0,1,1]=(0.0773392682115)*x[0]**o+(0.594680641339)*x[0]+(-0.514207317809)*x[1]**o+(0.516677819046)*x[1]+(-0.616548060204)*x[2]**o+(-0.487308340419)*x[2]
        s[2,0,2,0]=(0.308538286588)*x[0]**o+(0.553188024822)*x[0]+(0.640606076984)*x[1]**o+(0.851001392437)*x[1]+(-0.175136140727)*x[2]**o+(-0.0677219547255)*x[2]
        s[2,0,2,1]=(0.689683007295)*x[0]**o+(0.0205486604859)*x[0]+(-0.513197632413)*x[1]**o+(-0.589507997499)*x[1]+(0.0967263356508)*x[2]**o+(-0.743224844971)*x[2]
        s[2,1,0,0]=(0.175040132619)*x[0]**o+(0.0564766932006)*x[0]+(-0.641666937619)*x[1]**o+(0.820974466919)*x[1]+(0.494310429863)*x[2]**o+(-0.704304819607)*x[2]
        s[2,1,0,1]=(-0.890775824863)*x[0]**o+(-0.500686412296)*x[0]+(0.153470124596)*x[1]**o+(-0.64297395847)*x[1]+(-0.716566099394)*x[2]**o+(0.74578557063)*x[2]
        s[2,1,1,0]=(-0.999358965815)*x[0]**o+(-0.657733542154)*x[0]+(0.197035042134)*x[1]**o+(-0.887786199345)*x[1]+(-0.0685059299499)*x[2]**o+(0.498972129916)*x[2]
        s[2,1,1,1]=(0.627376530136)*x[0]**o+(-0.681228319483)*x[0]+(0.552935442983)*x[1]**o+(-0.813921943144)*x[1]+(-0.192761641367)*x[2]**o+(-0.248022829303)*x[2]
        s[2,1,2,0]=(0.845230425625)*x[0]**o+(-0.503049369073)*x[0]+(-0.705939778787)*x[1]**o+(-0.862373899236)*x[1]+(-0.327695135868)*x[2]**o+(-0.469012024699)*x[2]
        s[2,1,2,1]=(-0.509720328843)*x[0]**o+(-0.595784133936)*x[0]+(0.703072807212)*x[1]**o+(0.413326203128)*x[1]+(0.819831655306)*x[2]**o+(-0.494002066589)*x[2]
        s[2,2,0,0]=(-0.349642870638)*x[0]**o+(0.530770622335)*x[0]+(0.132675549915)*x[1]**o+(-0.234998432688)*x[1]+(-0.928123366402)*x[2]**o+(-0.578031336794)*x[2]
        s[2,2,0,1]=(-0.38335511074)*x[0]**o+(-0.993906933916)*x[0]+(-0.82535132107)*x[1]**o+(0.0112642210987)*x[1]+(0.904749217207)*x[2]**o+(-0.951098241145)*x[2]
        s[2,2,1,0]=(-0.387786853958)*x[0]**o+(0.455299305562)*x[0]+(0.690526696082)*x[1]**o+(0.192809198912)*x[1]+(-0.0217815689027)*x[2]**o+(0.853966231049)*x[2]
        s[2,2,1,1]=(0.399193164297)*x[0]**o+(0.19349639518)*x[0]+(0.317501609824)*x[1]**o+(0.240510267197)*x[1]+(0.0450712212053)*x[2]**o+(0.565853605939)*x[2]
        s[2,2,2,0]=(0.0853562286915)*x[0]**o+(0.122662799465)*x[0]+(0.26271658282)*x[1]**o+(0.90975618664)*x[1]+(-0.4111835609)*x[2]**o+(0.813644048231)*x[2]
        s[2,2,2,1]=(-0.852594000157)*x[0]**o+(-0.38507970023)*x[0]+(-0.232827312227)*x[1]**o+(0.00790525919862)*x[1]+(-0.3900612508)*x[2]**o+(-0.655554313153)*x[2]
        s[2,3,0,0]=(0.636184557317)*x[0]**o+(-0.549119485575)*x[0]+(-0.948821359516)*x[1]**o+(0.131451687899)*x[1]+(0.670399101185)*x[2]**o+(0.489217491983)*x[2]
        s[2,3,0,1]=(-0.856150893763)*x[0]**o+(-0.280815300266)*x[0]+(-0.523050266221)*x[1]**o+(0.520191604865)*x[1]+(0.0521766680699)*x[2]**o+(0.900107417471)*x[2]
        s[2,3,1,0]=(0.826705547671)*x[0]**o+(0.840439089501)*x[0]+(0.170261390802)*x[1]**o+(0.0811180372799)*x[1]+(-0.150910532868)*x[2]**o+(-0.312396418734)*x[2]
        s[2,3,1,1]=(-0.761542055765)*x[0]**o+(-0.972348306591)*x[0]+(0.872426053365)*x[1]**o+(-0.395809987519)*x[1]+(0.731896957461)*x[2]**o+(0.743098137394)*x[2]
        s[2,3,2,0]=(0.702021470374)*x[0]**o+(-0.378108346769)*x[0]+(-0.931728063859)*x[1]**o+(0.387496129463)*x[1]+(-0.0627914582829)*x[2]**o+(-0.648454760891)*x[2]
        s[2,3,2,1]=(0.694922709107)*x[0]**o+(0.348731283066)*x[0]+(-0.772113301946)*x[1]**o+(0.958191340446)*x[1]+(0.227272159891)*x[2]**o+(-0.553319421819)*x[2]
        s[2,4,0,0]=(0.201119725976)*x[0]**o+(0.593100698188)*x[0]+(-0.140057809068)*x[1]**o+(-0.0952566649925)*x[1]+(-0.851826547872)*x[2]**o+(0.505822022941)*x[2]
        s[2,4,0,1]=(0.0212756051792)*x[0]**o+(0.612323788569)*x[0]+(-0.582655513686)*x[1]**o+(0.461920977901)*x[1]+(0.45615084988)*x[2]**o+(-0.831485043604)*x[2]
        s[2,4,1,0]=(-0.896264627494)*x[0]**o+(0.130655495842)*x[0]+(-0.51905734937)*x[1]**o+(-0.850916009006)*x[1]+(-0.927481150342)*x[2]**o+(-0.394829448395)*x[2]
        s[2,4,1,1]=(0.589011113198)*x[0]**o+(0.779837449297)*x[0]+(-0.404231103183)*x[1]**o+(-0.818485472133)*x[1]+(-0.975586808573)*x[2]**o+(-0.140406688026)*x[2]
        s[2,4,2,0]=(0.621779862491)*x[0]**o+(-0.761880142681)*x[0]+(0.373695582445)*x[1]**o+(0.822746723299)*x[1]+(-0.0828625534579)*x[2]**o+(0.0737978389046)*x[2]
        s[2,4,2,1]=(-0.565625067907)*x[0]**o+(-0.494605807934)*x[0]+(-0.16016442955)*x[1]**o+(0.154859091053)*x[1]+(0.274690562477)*x[2]**o+(-0.0880681582431)*x[2]
        s[3,0,0,0]=(-0.876284142284)*x[0]**o+(0.833701643665)*x[0]+(-0.343394927755)*x[1]**o+(0.750989322932)*x[1]+(-0.0929682612771)*x[2]**o+(-0.098548374762)*x[2]
        s[3,0,0,1]=(-0.894241201097)*x[0]**o+(0.960924642371)*x[0]+(-0.385879848199)*x[1]**o+(-0.0740790803068)*x[1]+(-0.202098139954)*x[2]**o+(0.881367694352)*x[2]
        s[3,0,1,0]=(-0.462333015258)*x[0]**o+(-0.209675407943)*x[0]+(-0.162580359293)*x[1]**o+(0.699769354405)*x[1]+(0.505842739219)*x[2]**o+(-0.0511954656814)*x[2]
        s[3,0,1,1]=(-0.560563273256)*x[0]**o+(-0.798740244759)*x[0]+(0.661483757231)*x[1]**o+(0.160669606575)*x[1]+(-0.433798248002)*x[2]**o+(-0.573448279579)*x[2]
        s[3,0,2,0]=(-0.05963441333)*x[0]**o+(0.123422468842)*x[0]+(-0.0794480113474)*x[1]**o+(-0.980566064632)*x[1]+(0.172807620144)*x[2]**o+(0.290739484628)*x[2]
        s[3,0,2,1]=(-0.824869737132)*x[0]**o+(-0.236911913997)*x[0]+(-0.168741303105)*x[1]**o+(0.115993599766)*x[1]+(0.345866352676)*x[2]**o+(-0.810213944202)*x[2]
        s[3,1,0,0]=(-0.413325465485)*x[0]**o+(0.622383829933)*x[0]+(-0.517445935411)*x[1]**o+(0.968177050325)*x[1]+(-0.257376473971)*x[2]**o+(0.614956161076)*x[2]
        s[3,1,0,1]=(0.998189367064)*x[0]**o+(0.222064685116)*x[0]+(0.0082228342872)*x[1]**o+(-0.779543154084)*x[1]+(-0.566689004021)*x[2]**o+(0.580783912818)*x[2]
        s[3,1,1,0]=(0.245064518006)*x[0]**o+(0.252652697672)*x[0]+(-0.0750782925877)*x[1]**o+(0.216217715236)*x[1]+(-0.0766336852144)*x[2]**o+(0.819380420323)*x[2]
        s[3,1,1,1]=(-0.262865184606)*x[0]**o+(0.200232864872)*x[0]+(0.378217877682)*x[1]**o+(-0.560693320189)*x[1]+(0.235131793324)*x[2]**o+(0.0374878565262)*x[2]
        s[3,1,2,0]=(0.475538018977)*x[0]**o+(-0.731782099691)*x[0]+(-0.983895939904)*x[1]**o+(-0.505746472336)*x[1]+(-0.393214081403)*x[2]**o+(0.48342413719)*x[2]
        s[3,1,2,1]=(0.298214379)*x[0]**o+(-0.232901801182)*x[0]+(-0.47025072398)*x[1]**o+(-0.0935590248775)*x[1]+(-0.270755608412)*x[2]**o+(0.232963599748)*x[2]
        s[3,2,0,0]=(0.544547455045)*x[0]**o+(-0.238443336164)*x[0]+(0.104607358043)*x[1]**o+(0.700551882808)*x[1]+(-0.154865547453)*x[2]**o+(0.897044366696)*x[2]
        s[3,2,0,1]=(-0.226438512542)*x[0]**o+(-0.889161505097)*x[0]+(0.844759329847)*x[1]**o+(0.67393984486)*x[1]+(0.600583328853)*x[2]**o+(0.14085768238)*x[2]
        s[3,2,1,0]=(-0.303519143875)*x[0]**o+(-0.480627582343)*x[0]+(0.711647844961)*x[1]**o+(-0.827593833605)*x[1]+(-0.551933240305)*x[2]**o+(-0.358612992123)*x[2]
        s[3,2,1,1]=(-0.284375845165)*x[0]**o+(-0.144660681111)*x[0]+(-0.520160492072)*x[1]**o+(0.175169216201)*x[1]+(-0.500839582917)*x[2]**o+(-0.816514398637)*x[2]
        s[3,2,2,0]=(-0.563952264023)*x[0]**o+(0.104321734788)*x[0]+(-0.469018385204)*x[1]**o+(0.0093577376129)*x[1]+(-0.732696251175)*x[2]**o+(-0.178803413304)*x[2]
        s[3,2,2,1]=(0.774737106755)*x[0]**o+(0.860675883412)*x[0]+(0.79787325945)*x[1]**o+(0.664362658951)*x[1]+(-0.178147456548)*x[2]**o+(-0.253548140518)*x[2]
        s[3,3,0,0]=(-0.218464111878)*x[0]**o+(-0.255988638178)*x[0]+(0.278107535041)*x[1]**o+(0.40326101595)*x[1]+(-0.2331761004)*x[2]**o+(-0.317553406707)*x[2]
        s[3,3,0,1]=(-0.537282071415)*x[0]**o+(0.601952446965)*x[0]+(0.381109335541)*x[1]**o+(0.669261809466)*x[1]+(0.945072492166)*x[2]**o+(0.853596132501)*x[2]
        s[3,3,1,0]=(0.847664465983)*x[0]**o+(0.8066212639)*x[0]+(-0.857726357776)*x[1]**o+(0.474266813247)*x[1]+(-0.201605613201)*x[2]**o+(-0.793421334897)*x[2]
        s[3,3,1,1]=(-0.420235472332)*x[0]**o+(0.691721844986)*x[0]+(-0.265495062345)*x[1]**o+(0.275291637819)*x[1]+(0.935918852363)*x[2]**o+(0.682857746339)*x[2]
        s[3,3,2,0]=(-0.507370968068)*x[0]**o+(0.73174079427)*x[0]+(0.260834780217)*x[1]**o+(0.0234116684785)*x[1]+(-0.259405028241)*x[2]**o+(-0.95866122285)*x[2]
        s[3,3,2,1]=(-0.739777982614)*x[0]**o+(-0.634784565277)*x[0]+(-0.459700756224)*x[1]**o+(-0.668028439149)*x[1]+(0.543643060627)*x[2]**o+(0.481120453087)*x[2]
        s[3,4,0,0]=(-0.564521203332)*x[0]**o+(0.875264891992)*x[0]+(-0.711596964515)*x[1]**o+(0.516599187629)*x[1]+(0.803187045453)*x[2]**o+(0.130416944939)*x[2]
        s[3,4,0,1]=(-0.233194158029)*x[0]**o+(0.124405220571)*x[0]+(-0.0969251448946)*x[1]**o+(-0.13351738113)*x[1]+(-0.426939325117)*x[2]**o+(-0.0127989581875)*x[2]
        s[3,4,1,0]=(-0.654825260357)*x[0]**o+(-0.073851613222)*x[0]+(-0.866970362806)*x[1]**o+(-0.417768390991)*x[1]+(0.876189107492)*x[2]**o+(0.57905826773)*x[2]
        s[3,4,1,1]=(0.62991933581)*x[0]**o+(0.0333499904703)*x[0]+(0.296962053291)*x[1]**o+(-0.203151609982)*x[1]+(-0.416709189249)*x[2]**o+(-0.815746328506)*x[2]
        s[3,4,2,0]=(-0.720775085901)*x[0]**o+(0.96283858374)*x[0]+(0.23423515114)*x[1]**o+(0.640634860355)*x[1]+(-0.819346396686)*x[2]**o+(-0.807002446047)*x[2]
        s[3,4,2,1]=(-0.0608202717996)*x[0]**o+(0.339885794306)*x[0]+(-0.317314034496)*x[1]**o+(-0.672508392809)*x[1]+(0.63517267097)*x[2]**o+(-0.00885735344423)*x[2]
        ref[0,0,0,0]=(0.284952784158)/(o+1.)+(0.592032235941)
        ref[0,0,0,1]=(-0.803451440889)/(o+1.)+(0.0224418156084)
        ref[0,0,1,0]=(0.638402520436)/(o+1.)+(0.0311038353355)
        ref[0,0,1,1]=(0.307272122952)/(o+1.)+(0.0104780524684)
        ref[0,0,2,0]=(-0.502107719245)/(o+1.)+(-0.515645130678)
        ref[0,0,2,1]=(1.81768999676)/(o+1.)+(0.146737736684)
        ref[0,1,0,0]=(0.647820033767)/(o+1.)+(0.391233423499)
        ref[0,1,0,1]=(0.436092125372)/(o+1.)+(0.106230136165)
        ref[0,1,1,0]=(1.28557716642)/(o+1.)+(0.0374861393842)
        ref[0,1,1,1]=(-0.0491708546408)/(o+1.)+(-0.295525989953)
        ref[0,1,2,0]=(1.70047149916)/(o+1.)+(-0.214455033205)
        ref[0,1,2,1]=(-1.14969489881)/(o+1.)+(0.588651941542)
        ref[0,2,0,0]=(-0.955825457958)/(o+1.)+(-0.0190999661809)
        ref[0,2,0,1]=(0.229929844366)/(o+1.)+(-0.30311165791)
        ref[0,2,1,0]=(-2.57462574161)/(o+1.)+(-0.197980362151)
        ref[0,2,1,1]=(-0.0419001621974)/(o+1.)+(-0.379828843312)
        ref[0,2,2,0]=(-1.18351050669)/(o+1.)+(0.449768277313)
        ref[0,2,2,1]=(1.87822513678)/(o+1.)+(0.622693263344)
        ref[0,3,0,0]=(-0.0192193497223)/(o+1.)+(-0.274332614077)
        ref[0,3,0,1]=(1.05610437756)/(o+1.)+(-0.0158328399745)
        ref[0,3,1,0]=(-0.331289176593)/(o+1.)+(-0.023771634635)
        ref[0,3,1,1]=(-0.886769594367)/(o+1.)+(0.0191560107572)
        ref[0,3,2,0]=(0.554322811543)/(o+1.)+(0.324586599411)
        ref[0,3,2,1]=(0.757364125068)/(o+1.)+(0.0103767719331)
        ref[0,4,0,0]=(1.43319447117)/(o+1.)+(0.489873010572)
        ref[0,4,0,1]=(-1.54717367706)/(o+1.)+(-0.327842424773)
        ref[0,4,1,0]=(-1.02675093873)/(o+1.)+(-0.206388699347)
        ref[0,4,1,1]=(0.757998827253)/(o+1.)+(-0.892263769148)
        ref[0,4,2,0]=(0.979004551362)/(o+1.)+(-0.801869395247)
        ref[0,4,2,1]=(-2.23425638002)/(o+1.)+(-1.20737466522)
        ref[1,0,0,0]=(0.515302843784)/(o+1.)+(-0.162675558238)
        ref[1,0,0,1]=(0.353191474341)/(o+1.)+(-0.310747202442)
        ref[1,0,1,0]=(0.11330663338)/(o+1.)+(-0.0294705495767)
        ref[1,0,1,1]=(0.0625851979387)/(o+1.)+(-0.6942697411)
        ref[1,0,2,0]=(-0.634856196829)/(o+1.)+(0.233519187832)
        ref[1,0,2,1]=(0.758008553799)/(o+1.)+(-0.22308748171)
        ref[1,1,0,0]=(0.172197996272)/(o+1.)+(0.864945864826)
        ref[1,1,0,1]=(-0.0930030974171)/(o+1.)+(0.444243956798)
        ref[1,1,1,0]=(-0.0301833795362)/(o+1.)+(0.404852886018)
        ref[1,1,1,1]=(0.100136649986)/(o+1.)+(1.28659380019)
        ref[1,1,2,0]=(-0.97900883033)/(o+1.)+(0.442440382588)
        ref[1,1,2,1]=(0.792392656026)/(o+1.)+(0.181706071395)
        ref[1,2,0,0]=(-0.453470794812)/(o+1.)+(0.111351337976)
        ref[1,2,0,1]=(1.56202220869)/(o+1.)+(-0.511185001555)
        ref[1,2,1,0]=(0.659810976864)/(o+1.)+(0.621766984989)
        ref[1,2,1,1]=(1.91781515919)/(o+1.)+(0.0691044934815)
        ref[1,2,2,0]=(0.611207129327)/(o+1.)+(-0.497184221995)
        ref[1,2,2,1]=(1.65028805445)/(o+1.)+(-0.0280822974985)
        ref[1,3,0,0]=(-1.28735344741)/(o+1.)+(-0.680605269048)
        ref[1,3,0,1]=(0.256443215625)/(o+1.)+(-0.161440006142)
        ref[1,3,1,0]=(0.561119040305)/(o+1.)+(0.227463879795)
        ref[1,3,1,1]=(0.327555206284)/(o+1.)+(0.29622756806)
        ref[1,3,2,0]=(0.0665337521742)/(o+1.)+(0.77129792303)
        ref[1,3,2,1]=(1.36356118866)/(o+1.)+(-0.0276174494925)
        ref[1,4,0,0]=(0.793908154771)/(o+1.)+(-0.703932363283)
        ref[1,4,0,1]=(0.67317694748)/(o+1.)+(0.419078081349)
        ref[1,4,1,0]=(-1.6259928924)/(o+1.)+(-0.0422868647694)
        ref[1,4,1,1]=(0.573888422973)/(o+1.)+(-0.0709319079127)
        ref[1,4,2,0]=(-1.32097669048)/(o+1.)+(0.981719850544)
        ref[1,4,2,1]=(1.31804256932)/(o+1.)+(-0.464008585329)
        ref[2,0,0,0]=(0.131285227429)/(o+1.)+(-0.27188947549)
        ref[2,0,0,1]=(-0.534412041331)/(o+1.)+(-0.502296977244)
        ref[2,0,1,0]=(0.0835036877928)/(o+1.)+(0.675471052627)
        ref[2,0,1,1]=(-1.0534161098)/(o+1.)+(0.312025059983)
        ref[2,0,2,0]=(0.774008222845)/(o+1.)+(0.668233731266)
        ref[2,0,2,1]=(0.273211710533)/(o+1.)+(-0.656092090992)
        ref[2,1,0,0]=(0.0276836248637)/(o+1.)+(0.0865731702563)
        ref[2,1,0,1]=(-1.45387179966)/(o+1.)+(-0.198937400068)
        ref[2,1,1,0]=(-0.870829853631)/(o+1.)+(-0.523273805792)
        ref[2,1,1,1]=(0.987550331752)/(o+1.)+(-0.871586545966)
        ref[2,1,2,0]=(-0.188404489031)/(o+1.)+(-0.917217646504)
        ref[2,1,2,1]=(1.01318413367)/(o+1.)+(-0.338229998699)
        ref[2,2,0,0]=(-1.14509068713)/(o+1.)+(-0.141129573573)
        ref[2,2,0,1]=(-0.303957214603)/(o+1.)+(-0.966870476981)
        ref[2,2,1,0]=(0.280958273221)/(o+1.)+(0.751037367762)
        ref[2,2,1,1]=(0.761765995326)/(o+1.)+(0.499930134158)
        ref[2,2,2,0]=(-0.0631107493882)/(o+1.)+(0.923031517168)
        ref[2,2,2,1]=(-1.47548256318)/(o+1.)+(-0.516364377092)
        ref[2,3,0,0]=(0.357762298985)/(o+1.)+(0.0357748471536)
        ref[2,3,0,1]=(-1.32702449191)/(o+1.)+(0.569741861035)
        ref[2,3,1,0]=(0.846056405606)/(o+1.)+(0.304580354023)
        ref[2,3,1,1]=(0.842780955061)/(o+1.)+(-0.312530078358)
        ref[2,3,2,0]=(-0.292498051768)/(o+1.)+(-0.319533489098)
        ref[2,3,2,1]=(0.150081567052)/(o+1.)+(0.376801600846)
        ref[2,4,0,0]=(-0.790764630964)/(o+1.)+(0.501833028068)
        ref[2,4,0,1]=(-0.105229058628)/(o+1.)+(0.121379861433)
        ref[2,4,1,0]=(-2.34280312721)/(o+1.)+(-0.557544980779)
        ref[2,4,1,1]=(-0.790806798558)/(o+1.)+(-0.0895273554309)
        ref[2,4,2,0]=(0.912612891478)/(o+1.)+(0.0673322097612)
        ref[2,4,2,1]=(-0.45109893498)/(o+1.)+(-0.213907437562)
        ref[3,0,0,0]=(-1.31264733132)/(o+1.)+(0.743071295917)
        ref[3,0,0,1]=(-1.48221918925)/(o+1.)+(0.884106628208)
        ref[3,0,1,0]=(-0.119070635332)/(o+1.)+(0.21944924039)
        ref[3,0,1,1]=(-0.332877764028)/(o+1.)+(-0.605759458881)
        ref[3,0,2,0]=(0.0337251954668)/(o+1.)+(-0.283202055581)
        ref[3,0,2,1]=(-0.64774468756)/(o+1.)+(-0.465566129216)
        ref[3,1,0,0]=(-1.18814787487)/(o+1.)+(1.10275852067)
        ref[3,1,0,1]=(0.43972319733)/(o+1.)+(0.0116527219252)
        ref[3,1,1,0]=(0.0933525402037)/(o+1.)+(0.644125416616)
        ref[3,1,1,1]=(0.350484486401)/(o+1.)+(-0.161486299395)
        ref[3,1,2,0]=(-0.90157200233)/(o+1.)+(-0.377052217418)
        ref[3,1,2,1]=(-0.442791953392)/(o+1.)+(-0.0467486131556)
        ref[3,2,0,0]=(0.494289265635)/(o+1.)+(0.67957645667)
        ref[3,2,0,1]=(1.21890414616)/(o+1.)+(-0.0371819889285)
        ref[3,2,1,0]=(-0.143804539219)/(o+1.)+(-0.833417204036)
        ref[3,2,1,1]=(-1.30537592015)/(o+1.)+(-0.393002931773)
        ref[3,2,2,0]=(-1.7656669004)/(o+1.)+(-0.0325619704518)
        ref[3,2,2,1]=(1.39446290966)/(o+1.)+(0.635745200922)
        ref[3,3,0,0]=(-0.173532677236)/(o+1.)+(-0.0851405144673)
        ref[3,3,0,1]=(0.788899756292)/(o+1.)+(1.06240519447)
        ref[3,3,1,0]=(-0.211667504994)/(o+1.)+(0.243733371125)
        ref[3,3,1,1]=(0.250188317686)/(o+1.)+(0.824935614572)
        ref[3,3,2,0]=(-0.505941216092)/(o+1.)+(-0.101754380051)
        ref[3,3,2,1]=(-0.655835678211)/(o+1.)+(-0.410846275669)
        ref[3,4,0,0]=(-0.472931122393)/(o+1.)+(0.76114051228)
        ref[3,4,0,1]=(-0.757058628041)/(o+1.)+(-0.0109555593731)
        ref[3,4,1,0]=(-0.645606515671)/(o+1.)+(0.0437191317586)
        ref[3,4,1,1]=(0.510172199852)/(o+1.)+(-0.492773974009)
        ref[3,4,2,0]=(-1.30588633145)/(o+1.)+(0.398235499024)
        ref[3,4,2,1]=(0.257038364674)/(o+1.)+(-0.170739975974)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.465976690369)*x[0]**o+(-0.25136070196)*x[0]+(0.409576841566)*x[1]**o+(0.726707772754)*x[1]
        ref=(0.875553531935)/(o+1.)+(0.237673535397)
      else:
        arg=(0.567785895928)*x[0]**o+(0.928477128396)*x[0]+(-0.625720943088)*x[1]**o+(-0.681460841298)*x[1]+(-0.866801010069)*x[2]**o+(0.159594300637)*x[2]
        ref=(-0.924736057229)/(o+1.)+(0.203305293867)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.839020088668)*x[0]**o+(-0.880288035177)*x[0]+(0.260951734364)*x[1]**o+(-0.624208000982)*x[1]
        arg[1]=(0.985421831314)*x[0]**o+(0.660399229451)*x[0]+(-0.514003848134)*x[1]**o+(0.0312162082593)*x[1]
        ref[0]=(-0.578068354304)/(o+1.)+(-0.75224801808)
        ref[1]=(0.47141798318)/(o+1.)+(0.345807718855)
      else:
        arg[0]=(-0.627307316929)*x[0]**o+(0.500432411035)*x[0]+(0.233895908313)*x[1]**o+(-0.125048135432)*x[1]+(0.0695114009632)*x[2]**o+(0.995170977074)*x[2]
        arg[1]=(0.760033715785)*x[0]**o+(0.48895905139)*x[0]+(0.240372023176)*x[1]**o+(0.33946689896)*x[1]+(-0.774638489419)*x[2]**o+(-0.0777240871074)*x[2]
        ref[0]=(-0.323900007653)/(o+1.)+(0.685277626338)
        ref[1]=(0.225767249542)/(o+1.)+(0.375350931622)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.844492048883)*x[0]**o+(0.727132679659)*x[0]+(-0.862576284722)*x[1]**o+(0.502937561417)*x[1]
        arg[0,1]=(-0.587157419309)*x[0]**o+(0.00327068097107)*x[0]+(0.422592349683)*x[1]**o+(0.167479615369)*x[1]
        arg[0,2]=(0.771156665672)*x[0]**o+(-0.371091138833)*x[0]+(-0.867750378658)*x[1]**o+(-0.744394290185)*x[1]
        arg[0,3]=(0.669718598855)*x[0]**o+(-0.0825909409121)*x[0]+(0.868872948791)*x[1]**o+(0.141891982204)*x[1]
        arg[0,4]=(-0.441072764653)*x[0]**o+(-0.929654649265)*x[0]+(-0.35883498816)*x[1]**o+(-0.84045480381)*x[1]
        arg[1,0]=(0.555668196622)*x[0]**o+(0.94306681043)*x[0]+(0.586512446245)*x[1]**o+(0.702007276327)*x[1]
        arg[1,1]=(0.792083771005)*x[0]**o+(-0.865830139617)*x[0]+(0.0929971631615)*x[1]**o+(0.620446289121)*x[1]
        arg[1,2]=(-0.201978375472)*x[0]**o+(-0.0777206885918)*x[0]+(0.798755459687)*x[1]**o+(-0.278086468846)*x[1]
        arg[1,3]=(0.063028419572)*x[0]**o+(-0.454786232294)*x[0]+(-0.0481393324293)*x[1]**o+(-0.612657132834)*x[1]
        arg[1,4]=(-0.743283787964)*x[0]**o+(0.854827303846)*x[0]+(0.94241635825)*x[1]**o+(0.436937526521)*x[1]
        arg[2,0]=(-0.814975695669)*x[0]**o+(0.579477783769)*x[0]+(0.00998264005965)*x[1]**o+(-0.913665878657)*x[1]
        arg[2,1]=(-0.526833976369)*x[0]**o+(-0.800386086232)*x[0]+(-0.487600053086)*x[1]**o+(0.00242562519003)*x[1]
        arg[2,2]=(-0.964119591874)*x[0]**o+(0.203212394214)*x[0]+(-0.96850837954)*x[1]**o+(-0.214886132512)*x[1]
        arg[2,3]=(-0.356249104564)*x[0]**o+(0.329769293221)*x[0]+(0.936664928754)*x[1]**o+(0.520603503283)*x[1]
        arg[2,4]=(0.255464219424)*x[0]**o+(-0.157783294516)*x[0]+(0.519451787865)*x[1]**o+(0.00960521397376)*x[1]
        arg[3,0]=(-0.117250486034)*x[0]**o+(0.192945431273)*x[0]+(-0.7580154297)*x[1]**o+(-0.308801650294)*x[1]
        arg[3,1]=(0.243931002845)*x[0]**o+(-0.626714191745)*x[0]+(-0.365870571265)*x[1]**o+(0.652257582021)*x[1]
        arg[3,2]=(-0.769756656122)*x[0]**o+(0.711106390017)*x[0]+(0.150749630653)*x[1]**o+(-0.0441309732448)*x[1]
        arg[3,3]=(0.92677837493)*x[0]**o+(0.507199859313)*x[0]+(-0.542275490828)*x[1]**o+(-0.0907270087228)*x[1]
        arg[3,4]=(-0.0306175885885)*x[0]**o+(-0.434666505018)*x[0]+(-0.815633427562)*x[1]**o+(-0.916424797903)*x[1]
        ref[0,0]=(-1.70706833361)/(o+1.)+(0.615035120538)
        ref[0,1]=(-0.164565069626)/(o+1.)+(0.08537514817)
        ref[0,2]=(-0.0965937129851)/(o+1.)+(-0.557742714509)
        ref[0,3]=(1.53859154765)/(o+1.)+(0.0296505206459)
        ref[0,4]=(-0.799907752814)/(o+1.)+(-0.885054726537)
        ref[1,0]=(1.14218064287)/(o+1.)+(0.822537043379)
        ref[1,1]=(0.885080934167)/(o+1.)+(-0.122691925248)
        ref[1,2]=(0.596777084215)/(o+1.)+(-0.177903578719)
        ref[1,3]=(0.0148890871427)/(o+1.)+(-0.533721682564)
        ref[1,4]=(0.199132570285)/(o+1.)+(0.645882415184)
        ref[2,0]=(-0.80499305561)/(o+1.)+(-0.167094047444)
        ref[2,1]=(-1.01443402945)/(o+1.)+(-0.398980230521)
        ref[2,2]=(-1.93262797141)/(o+1.)+(-0.00583686914904)
        ref[2,3]=(0.580415824189)/(o+1.)+(0.425186398252)
        ref[2,4]=(0.774916007289)/(o+1.)+(-0.074089040271)
        ref[3,0]=(-0.875265915734)/(o+1.)+(-0.0579281095106)
        ref[3,1]=(-0.12193956842)/(o+1.)+(0.0127716951382)
        ref[3,2]=(-0.619007025469)/(o+1.)+(0.333487708386)
        ref[3,3]=(0.384502884102)/(o+1.)+(0.208236425295)
        ref[3,4]=(-0.846251016151)/(o+1.)+(-0.675545651461)
      else:
        arg[0,0]=(0.596538443259)*x[0]**o+(0.479462980373)*x[0]+(-0.334469126403)*x[1]**o+(0.784201085556)*x[1]+(-0.231474140531)*x[2]**o+(-0.368096345589)*x[2]
        arg[0,1]=(0.188945383338)*x[0]**o+(-0.95033024122)*x[0]+(0.567904901047)*x[1]**o+(-0.76703296351)*x[1]+(-0.0116083690376)*x[2]**o+(-0.211176589764)*x[2]
        arg[0,2]=(0.225412552449)*x[0]**o+(-0.731615155924)*x[0]+(-0.362727501637)*x[1]**o+(-0.195442119344)*x[1]+(-0.215444456285)*x[2]**o+(0.412715796392)*x[2]
        arg[0,3]=(-0.00941287481408)*x[0]**o+(-0.207271678463)*x[0]+(0.66582433583)*x[1]**o+(-0.297871484308)*x[1]+(-0.819460465536)*x[2]**o+(-0.3316307033)*x[2]
        arg[0,4]=(-0.172265683581)*x[0]**o+(0.881658164973)*x[0]+(0.0267816673708)*x[1]**o+(-0.526757427521)*x[1]+(-0.666953449252)*x[2]**o+(0.366533690002)*x[2]
        arg[1,0]=(-0.575272180021)*x[0]**o+(-0.707236339163)*x[0]+(-0.734723292219)*x[1]**o+(-0.872019568084)*x[1]+(-0.513269190461)*x[2]**o+(-0.95077853535)*x[2]
        arg[1,1]=(0.687652562435)*x[0]**o+(0.500668143461)*x[0]+(0.127678854367)*x[1]**o+(0.135196530391)*x[1]+(0.219757431874)*x[2]**o+(0.43373842168)*x[2]
        arg[1,2]=(-0.674120744855)*x[0]**o+(-0.792769889708)*x[0]+(0.0414602844391)*x[1]**o+(-0.159887164563)*x[1]+(0.406441555189)*x[2]**o+(-0.662210077269)*x[2]
        arg[1,3]=(-0.623120612358)*x[0]**o+(0.895951757448)*x[0]+(0.957407224559)*x[1]**o+(-0.472225118848)*x[1]+(-0.294551998847)*x[2]**o+(0.503584320886)*x[2]
        arg[1,4]=(-0.304220645545)*x[0]**o+(0.812321371589)*x[0]+(0.932257196386)*x[1]**o+(0.736014913915)*x[1]+(0.119140391025)*x[2]**o+(0.910696940879)*x[2]
        arg[2,0]=(0.275470013797)*x[0]**o+(-0.903754203548)*x[0]+(-0.443211232771)*x[1]**o+(0.78815938507)*x[1]+(-0.0802934106722)*x[2]**o+(0.096532189886)*x[2]
        arg[2,1]=(-0.4505392488)*x[0]**o+(-0.668542586506)*x[0]+(0.487967999984)*x[1]**o+(-0.642310745404)*x[1]+(-0.956530771836)*x[2]**o+(0.517544404873)*x[2]
        arg[2,2]=(-0.984032885692)*x[0]**o+(-0.606688527142)*x[0]+(0.0565483776156)*x[1]**o+(-0.945228520749)*x[1]+(0.433070187921)*x[2]**o+(-0.599983370996)*x[2]
        arg[2,3]=(-0.673469008741)*x[0]**o+(0.612144012035)*x[0]+(-0.169497339443)*x[1]**o+(0.267990057152)*x[1]+(0.431716413548)*x[2]**o+(0.188904436012)*x[2]
        arg[2,4]=(0.320782309844)*x[0]**o+(-0.830068082984)*x[0]+(0.900183436056)*x[1]**o+(-0.343951186189)*x[1]+(-0.00137135137854)*x[2]**o+(-0.655764522249)*x[2]
        arg[3,0]=(-0.936193981767)*x[0]**o+(0.885260284217)*x[0]+(0.00397725971098)*x[1]**o+(-0.0864751234501)*x[1]+(0.25001166345)*x[2]**o+(-0.127859505667)*x[2]
        arg[3,1]=(-0.0841986843231)*x[0]**o+(-0.95678469673)*x[0]+(0.867049114883)*x[1]**o+(0.154499266839)*x[1]+(0.0217146474622)*x[2]**o+(-0.945774319527)*x[2]
        arg[3,2]=(-0.668161122306)*x[0]**o+(0.779140555525)*x[0]+(0.217613826157)*x[1]**o+(0.900077039607)*x[1]+(-0.251881965789)*x[2]**o+(-0.331281723354)*x[2]
        arg[3,3]=(-0.95624079226)*x[0]**o+(-0.403226291225)*x[0]+(0.303340462877)*x[1]**o+(-0.269819286355)*x[1]+(0.605804508829)*x[2]**o+(-0.461404715818)*x[2]
        arg[3,4]=(-0.933059329973)*x[0]**o+(-0.00851461368762)*x[0]+(0.0588544443918)*x[1]**o+(-0.682939702389)*x[1]+(-0.213965742261)*x[2]**o+(-0.102965977964)*x[2]
        ref[0,0]=(0.0305951763245)/(o+1.)+(0.44778386017)
        ref[0,1]=(0.745241915348)/(o+1.)+(-0.964269897247)
        ref[0,2]=(-0.352759405473)/(o+1.)+(-0.257170739438)
        ref[0,3]=(-0.163049004521)/(o+1.)+(-0.418386933036)
        ref[0,4]=(-0.812437465462)/(o+1.)+(0.360717213727)
        ref[1,0]=(-1.8232646627)/(o+1.)+(-1.2650172213)
        ref[1,1]=(1.03508884868)/(o+1.)+(0.534801547766)
        ref[1,2]=(-0.226218905227)/(o+1.)+(-0.80743356577)
        ref[1,3]=(0.0397346133545)/(o+1.)+(0.463655479743)
        ref[1,4]=(0.747176941866)/(o+1.)+(1.22951661319)
        ref[2,0]=(-0.248034629647)/(o+1.)+(-0.00953131429598)
        ref[2,1]=(-0.919102020652)/(o+1.)+(-0.396654463518)
        ref[2,2]=(-0.494414320156)/(o+1.)+(-1.07595020944)
        ref[2,3]=(-0.411249934637)/(o+1.)+(0.534519252599)
        ref[2,4]=(1.21959439452)/(o+1.)+(-0.914891895711)
        ref[3,0]=(-0.682205058606)/(o+1.)+(0.33546282755)
        ref[3,1]=(0.804565078022)/(o+1.)+(-0.874029874709)
        ref[3,2]=(-0.702429261938)/(o+1.)+(0.673967935889)
        ref[3,3]=(-0.0470958205546)/(o+1.)+(-0.567225146699)
        ref[3,4]=(-1.08817062784)/(o+1.)+(-0.39721014702)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.401537399178)*x[0]**o+(-0.868937065568)*x[0]+(0.356734884346)*x[1]**o+(0.0137485089422)*x[1]
        arg[0,0,1]=(0.705383698155)*x[0]**o+(-0.884988282352)*x[0]+(-0.979185994641)*x[1]**o+(0.929576942519)*x[1]
        arg[0,1,0]=(-0.292540086272)*x[0]**o+(-0.223230840915)*x[0]+(0.717291109855)*x[1]**o+(-0.826233591275)*x[1]
        arg[0,1,1]=(0.55739919643)*x[0]**o+(0.2571407205)*x[0]+(0.00434404142287)*x[1]**o+(0.200305144451)*x[1]
        arg[1,0,0]=(0.451841947443)*x[0]**o+(-0.576279050146)*x[0]+(-0.128142118055)*x[1]**o+(0.316406960557)*x[1]
        arg[1,0,1]=(-0.07174792925)*x[0]**o+(-0.392935435009)*x[0]+(-0.343535485172)*x[1]**o+(0.619131193274)*x[1]
        arg[1,1,0]=(-0.667031669611)*x[0]**o+(-0.184806094874)*x[0]+(-0.199375296384)*x[1]**o+(0.449351889815)*x[1]
        arg[1,1,1]=(-0.51924508404)*x[0]**o+(0.870508122509)*x[0]+(-0.274149433148)*x[1]**o+(-0.762367215962)*x[1]
        arg[2,0,0]=(-0.501546909709)*x[0]**o+(0.164122129343)*x[0]+(-0.601370828214)*x[1]**o+(0.888046948888)*x[1]
        arg[2,0,1]=(0.272472486899)*x[0]**o+(-0.958776296738)*x[0]+(-0.334676045163)*x[1]**o+(0.410943596779)*x[1]
        arg[2,1,0]=(0.363478201368)*x[0]**o+(0.753492598803)*x[0]+(-0.663726768274)*x[1]**o+(0.875443816963)*x[1]
        arg[2,1,1]=(0.995788536211)*x[0]**o+(-0.123932097555)*x[0]+(0.143025171293)*x[1]**o+(0.724918059214)*x[1]
        arg[3,0,0]=(-0.578717829755)*x[0]**o+(-0.899523186929)*x[0]+(0.58523203854)*x[1]**o+(0.751620714771)*x[1]
        arg[3,0,1]=(0.179841033744)*x[0]**o+(0.268015839688)*x[0]+(0.860292991989)*x[1]**o+(0.0117969420991)*x[1]
        arg[3,1,0]=(-0.575777000777)*x[0]**o+(0.170047012345)*x[0]+(-0.418743243278)*x[1]**o+(-0.278415849566)*x[1]
        arg[3,1,1]=(-0.913076795648)*x[0]**o+(0.486330450107)*x[0]+(-0.423570363453)*x[1]**o+(0.388575474946)*x[1]
        arg[4,0,0]=(0.16695368621)*x[0]**o+(0.133497460585)*x[0]+(-0.444434659151)*x[1]**o+(-0.895229636275)*x[1]
        arg[4,0,1]=(0.432018868106)*x[0]**o+(-0.281976075625)*x[0]+(-0.0455135209769)*x[1]**o+(0.76870037059)*x[1]
        arg[4,1,0]=(0.238542836112)*x[0]**o+(-0.806931884409)*x[0]+(-0.0365533871189)*x[1]**o+(0.731336634477)*x[1]
        arg[4,1,1]=(0.512035255059)*x[0]**o+(-0.683560978902)*x[0]+(-0.953782277737)*x[1]**o+(-0.311664210765)*x[1]
        arg[5,0,0]=(-0.842399299624)*x[0]**o+(0.460676816656)*x[0]+(-0.251781603841)*x[1]**o+(0.321792359074)*x[1]
        arg[5,0,1]=(-0.900576257062)*x[0]**o+(0.842393710528)*x[0]+(0.630445446843)*x[1]**o+(-0.647745144739)*x[1]
        arg[5,1,0]=(0.136687908652)*x[0]**o+(0.0252907890641)*x[0]+(0.813054611844)*x[1]**o+(-0.459930787056)*x[1]
        arg[5,1,1]=(-0.00242518550941)*x[0]**o+(-0.558785348137)*x[0]+(0.0568659980522)*x[1]**o+(0.689293614051)*x[1]
        ref[0,0,0]=(0.758272283524)/(o+1.)+(-0.427594278313)
        ref[0,0,1]=(-0.273802296487)/(o+1.)+(0.0222943300833)
        ref[0,1,0]=(0.424751023583)/(o+1.)+(-0.524732216095)
        ref[0,1,1]=(0.561743237852)/(o+1.)+(0.228722932476)
        ref[1,0,0]=(0.323699829388)/(o+1.)+(-0.129936044795)
        ref[1,0,1]=(-0.415283414421)/(o+1.)+(0.113097879133)
        ref[1,1,0]=(-0.866406965995)/(o+1.)+(0.132272897471)
        ref[1,1,1]=(-0.793394517189)/(o+1.)+(0.0540704532731)
        ref[2,0,0]=(-1.10291773792)/(o+1.)+(0.526084539115)
        ref[2,0,1]=(-0.0622035582637)/(o+1.)+(-0.273916349979)
        ref[2,1,0]=(-0.300248566906)/(o+1.)+(0.814468207883)
        ref[2,1,1]=(1.1388137075)/(o+1.)+(0.300492980829)
        ref[3,0,0]=(0.00651420878498)/(o+1.)+(-0.073951236079)
        ref[3,0,1]=(1.04013402573)/(o+1.)+(0.139906390894)
        ref[3,1,0]=(-0.994520244055)/(o+1.)+(-0.0541844186105)
        ref[3,1,1]=(-1.3366471591)/(o+1.)+(0.437452962527)
        ref[4,0,0]=(-0.27748097294)/(o+1.)+(-0.380866087845)
        ref[4,0,1]=(0.386505347129)/(o+1.)+(0.243362147483)
        ref[4,1,0]=(0.201989448993)/(o+1.)+(-0.0377976249661)
        ref[4,1,1]=(-0.441747022678)/(o+1.)+(-0.497612594833)
        ref[5,0,0]=(-1.09418090346)/(o+1.)+(0.391234587865)
        ref[5,0,1]=(-0.270130810219)/(o+1.)+(0.0973242828949)
        ref[5,1,0]=(0.949742520497)/(o+1.)+(-0.217319998996)
        ref[5,1,1]=(0.0544408125428)/(o+1.)+(0.0652541329568)
      else:
        arg[0,0,0]=(0.411788818613)*x[0]**o+(0.324151698669)*x[0]+(0.656551349616)*x[1]**o+(0.801577401197)*x[1]+(0.690440431022)*x[2]**o+(0.776409870151)*x[2]
        arg[0,0,1]=(-0.224985947659)*x[0]**o+(-0.564703086188)*x[0]+(-0.37196336116)*x[1]**o+(-0.114979853172)*x[1]+(0.274462080499)*x[2]**o+(-0.988489021983)*x[2]
        arg[0,1,0]=(0.512331530009)*x[0]**o+(0.519617843594)*x[0]+(-0.515174861158)*x[1]**o+(-0.675953902994)*x[1]+(0.796111490716)*x[2]**o+(0.229569489858)*x[2]
        arg[0,1,1]=(0.347765144376)*x[0]**o+(-0.0110194730031)*x[0]+(-0.593936645599)*x[1]**o+(-0.372786558089)*x[1]+(-0.604691321362)*x[2]**o+(-0.888002521077)*x[2]
        arg[1,0,0]=(-0.575954903433)*x[0]**o+(0.0160235655616)*x[0]+(0.305335632724)*x[1]**o+(-0.259402553796)*x[1]+(-0.221276854741)*x[2]**o+(-0.519932341069)*x[2]
        arg[1,0,1]=(-0.674987647238)*x[0]**o+(-0.45852667934)*x[0]+(-0.758937440978)*x[1]**o+(0.80566898197)*x[1]+(-0.961619154322)*x[2]**o+(0.861564430243)*x[2]
        arg[1,1,0]=(-0.297475412437)*x[0]**o+(0.881828705604)*x[0]+(-0.823267615874)*x[1]**o+(-0.0232728101313)*x[1]+(-0.676220001453)*x[2]**o+(-0.382414036981)*x[2]
        arg[1,1,1]=(-0.31008204616)*x[0]**o+(0.805195900401)*x[0]+(-0.758209772143)*x[1]**o+(0.861379709073)*x[1]+(0.0631063866518)*x[2]**o+(0.356607080553)*x[2]
        arg[2,0,0]=(0.305214964591)*x[0]**o+(0.815106526139)*x[0]+(0.138797156005)*x[1]**o+(-0.241921896332)*x[1]+(-0.103783035058)*x[2]**o+(-0.161952044984)*x[2]
        arg[2,0,1]=(-0.925375677635)*x[0]**o+(0.0341347568201)*x[0]+(-0.706905924964)*x[1]**o+(-0.193304368451)*x[1]+(-0.449866757104)*x[2]**o+(-0.894963680869)*x[2]
        arg[2,1,0]=(0.822676224279)*x[0]**o+(-0.28758006682)*x[0]+(-0.869025918917)*x[1]**o+(0.0107602044744)*x[1]+(-0.832322765163)*x[2]**o+(0.748440319389)*x[2]
        arg[2,1,1]=(-0.415326508997)*x[0]**o+(-0.879521424817)*x[0]+(-0.971781441062)*x[1]**o+(0.527586634108)*x[1]+(0.336682669856)*x[2]**o+(-0.948414041149)*x[2]
        arg[3,0,0]=(-0.345788531715)*x[0]**o+(0.626583498807)*x[0]+(-0.186418683649)*x[1]**o+(0.3141628424)*x[1]+(-0.781119787783)*x[2]**o+(0.972708430266)*x[2]
        arg[3,0,1]=(0.175536279095)*x[0]**o+(-0.150027774295)*x[0]+(-0.506653680683)*x[1]**o+(-0.482544397029)*x[1]+(-0.59090996152)*x[2]**o+(0.618123625026)*x[2]
        arg[3,1,0]=(0.737984493124)*x[0]**o+(0.650901680297)*x[0]+(-0.361389200009)*x[1]**o+(-0.784304665396)*x[1]+(-0.954132232142)*x[2]**o+(0.684341934348)*x[2]
        arg[3,1,1]=(-0.385757579645)*x[0]**o+(0.251136919845)*x[0]+(0.95143939933)*x[1]**o+(0.501860043225)*x[1]+(-0.675673011184)*x[2]**o+(0.337315450579)*x[2]
        arg[4,0,0]=(-0.740183402139)*x[0]**o+(-0.828278760743)*x[0]+(0.88000889117)*x[1]**o+(-0.0226801022184)*x[1]+(0.6811397662)*x[2]**o+(-0.254627446498)*x[2]
        arg[4,0,1]=(-0.194128952361)*x[0]**o+(-0.524854896388)*x[0]+(0.341100326577)*x[1]**o+(-0.26072443499)*x[1]+(-0.197396331776)*x[2]**o+(-0.589037632112)*x[2]
        arg[4,1,0]=(0.261995024865)*x[0]**o+(0.214193793699)*x[0]+(0.572024600244)*x[1]**o+(0.230659327198)*x[1]+(0.420971705453)*x[2]**o+(0.333445049827)*x[2]
        arg[4,1,1]=(-0.0679525054395)*x[0]**o+(0.0756646296473)*x[0]+(-0.299865836239)*x[1]**o+(-0.828294764914)*x[1]+(0.606170745631)*x[2]**o+(0.624267827393)*x[2]
        arg[5,0,0]=(0.312565542211)*x[0]**o+(-0.368815496171)*x[0]+(-0.0615973473105)*x[1]**o+(0.205012835971)*x[1]+(0.310818553076)*x[2]**o+(-0.287393908304)*x[2]
        arg[5,0,1]=(-0.409626309069)*x[0]**o+(0.977953573562)*x[0]+(0.583013544138)*x[1]**o+(0.542907128677)*x[1]+(0.488538336366)*x[2]**o+(-0.85823611123)*x[2]
        arg[5,1,0]=(-0.899607109395)*x[0]**o+(-0.0459053040522)*x[0]+(0.938423293041)*x[1]**o+(-0.112431665317)*x[1]+(0.620609328687)*x[2]**o+(-0.663082429319)*x[2]
        arg[5,1,1]=(0.00370253144626)*x[0]**o+(0.503733497956)*x[0]+(0.362457785043)*x[1]**o+(0.417891957303)*x[1]+(0.404543677829)*x[2]**o+(0.117828669122)*x[2]
        ref[0,0,0]=(1.75878059925)/(o+1.)+(0.951069485009)
        ref[0,0,1]=(-0.32248722832)/(o+1.)+(-0.834085980672)
        ref[0,1,0]=(0.793268159567)/(o+1.)+(0.036616715229)
        ref[0,1,1]=(-0.850862822585)/(o+1.)+(-0.635904276085)
        ref[1,0,0]=(-0.49189612545)/(o+1.)+(-0.381655664652)
        ref[1,0,1]=(-2.39554424254)/(o+1.)+(0.604353366437)
        ref[1,1,0]=(-1.79696302976)/(o+1.)+(0.238070929246)
        ref[1,1,1]=(-1.00518543165)/(o+1.)+(1.01159134501)
        ref[2,0,0]=(0.340229085537)/(o+1.)+(0.205616292412)
        ref[2,0,1]=(-2.0821483597)/(o+1.)+(-0.52706664625)
        ref[2,1,0]=(-0.878672459802)/(o+1.)+(0.235810228522)
        ref[2,1,1]=(-1.0504252802)/(o+1.)+(-0.650174415929)
        ref[3,0,0]=(-1.31332700315)/(o+1.)+(0.956727385736)
        ref[3,0,1]=(-0.922027363109)/(o+1.)+(-0.00722427314892)
        ref[3,1,0]=(-0.577536939027)/(o+1.)+(0.275469474624)
        ref[3,1,1]=(-0.109991191499)/(o+1.)+(0.545156206825)
        ref[4,0,0]=(0.820965255231)/(o+1.)+(-0.55279315473)
        ref[4,0,1]=(-0.0504249575602)/(o+1.)+(-0.687308481745)
        ref[4,1,0]=(1.25499133056)/(o+1.)+(0.389149085362)
        ref[4,1,1]=(0.238352403952)/(o+1.)+(-0.0641811539369)
        ref[5,0,0]=(0.561786747976)/(o+1.)+(-0.225598284252)
        ref[5,0,1]=(0.661925571434)/(o+1.)+(0.331312295504)
        ref[5,1,0]=(0.659425512333)/(o+1.)+(-0.410709699344)
        ref[5,1,1]=(0.770703994319)/(o+1.)+(0.519727062191)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.164405890316)*x[0]**o+(-0.323303939764)*x[0]+(-0.0516829984856)*x[1]**o+(-0.220765052533)*x[1]
        arg[0,0,0,1]=(0.152172612108)*x[0]**o+(-0.166242164365)*x[0]+(0.807177808571)*x[1]**o+(0.0404817109573)*x[1]
        arg[0,0,1,0]=(0.016912808567)*x[0]**o+(0.623923316774)*x[0]+(-0.542886934905)*x[1]**o+(0.0110784560205)*x[1]
        arg[0,0,1,1]=(-0.663885786909)*x[0]**o+(-0.55563355485)*x[0]+(0.191993284382)*x[1]**o+(0.734973556043)*x[1]
        arg[0,0,2,0]=(-0.110316754183)*x[0]**o+(-0.584022783672)*x[0]+(0.56471376011)*x[1]**o+(0.846317747143)*x[1]
        arg[0,0,2,1]=(-0.767988009278)*x[0]**o+(0.440997912352)*x[0]+(0.45366348093)*x[1]**o+(0.376620990318)*x[1]
        arg[0,1,0,0]=(-0.138006508051)*x[0]**o+(0.8450564766)*x[0]+(-0.273559897633)*x[1]**o+(0.683678615645)*x[1]
        arg[0,1,0,1]=(-0.467521625474)*x[0]**o+(0.32289306962)*x[0]+(0.340974562152)*x[1]**o+(-0.58109391053)*x[1]
        arg[0,1,1,0]=(0.76867991208)*x[0]**o+(0.683527735131)*x[0]+(0.762986872389)*x[1]**o+(-0.189342734286)*x[1]
        arg[0,1,1,1]=(0.125150175351)*x[0]**o+(-0.175190872876)*x[0]+(-0.569766488076)*x[1]**o+(0.787503755206)*x[1]
        arg[0,1,2,0]=(-0.339695860559)*x[0]**o+(0.119452596206)*x[0]+(0.474378864192)*x[1]**o+(0.845982555265)*x[1]
        arg[0,1,2,1]=(0.820551334836)*x[0]**o+(-0.200024721285)*x[0]+(-0.525871855871)*x[1]**o+(-0.714639297791)*x[1]
        arg[0,2,0,0]=(-0.145788788262)*x[0]**o+(-0.695717636982)*x[0]+(0.736564332801)*x[1]**o+(0.634307016587)*x[1]
        arg[0,2,0,1]=(-0.25142006947)*x[0]**o+(0.027028767127)*x[0]+(-0.614777438028)*x[1]**o+(-0.691417785617)*x[1]
        arg[0,2,1,0]=(0.366081140839)*x[0]**o+(0.477490533426)*x[0]+(-0.116611335116)*x[1]**o+(-0.863552165712)*x[1]
        arg[0,2,1,1]=(-0.467659558308)*x[0]**o+(0.0836299545334)*x[0]+(0.903093252833)*x[1]**o+(-0.956858182533)*x[1]
        arg[0,2,2,0]=(0.870921031579)*x[0]**o+(-0.801681014465)*x[0]+(-0.951819608329)*x[1]**o+(-0.626222917084)*x[1]
        arg[0,2,2,1]=(0.414973183041)*x[0]**o+(-0.444580205703)*x[0]+(-0.346458764208)*x[1]**o+(-0.679140100745)*x[1]
        arg[0,3,0,0]=(-0.725326935158)*x[0]**o+(-0.914034759513)*x[0]+(0.702614377806)*x[1]**o+(0.819004276317)*x[1]
        arg[0,3,0,1]=(0.617099932352)*x[0]**o+(-0.155182850011)*x[0]+(-0.066428715129)*x[1]**o+(-0.286480585435)*x[1]
        arg[0,3,1,0]=(0.710942545488)*x[0]**o+(-0.734509277268)*x[0]+(-0.807474546329)*x[1]**o+(0.307631760449)*x[1]
        arg[0,3,1,1]=(-0.709966924732)*x[0]**o+(-0.462297022374)*x[0]+(0.773787774038)*x[1]**o+(0.762137020991)*x[1]
        arg[0,3,2,0]=(-0.425660932209)*x[0]**o+(-0.562130300181)*x[0]+(-0.516719014647)*x[1]**o+(-0.547832887369)*x[1]
        arg[0,3,2,1]=(0.149573176879)*x[0]**o+(-0.208999629436)*x[0]+(-0.191426033553)*x[1]**o+(-0.119046538873)*x[1]
        arg[0,4,0,0]=(-0.99932317151)*x[0]**o+(-0.832461558068)*x[0]+(-0.847073271676)*x[1]**o+(0.882046887729)*x[1]
        arg[0,4,0,1]=(0.14235557233)*x[0]**o+(-0.352572053276)*x[0]+(-0.231989519255)*x[1]**o+(0.146855135398)*x[1]
        arg[0,4,1,0]=(0.550561449012)*x[0]**o+(0.530745681425)*x[0]+(0.198889689472)*x[1]**o+(0.165503108658)*x[1]
        arg[0,4,1,1]=(0.059663401309)*x[0]**o+(0.168377411751)*x[0]+(-0.224682723925)*x[1]**o+(-0.255406907757)*x[1]
        arg[0,4,2,0]=(-0.747776894683)*x[0]**o+(-0.281763358533)*x[0]+(0.660369838923)*x[1]**o+(0.348541234687)*x[1]
        arg[0,4,2,1]=(-0.886248366716)*x[0]**o+(0.665256341312)*x[0]+(-0.118041122298)*x[1]**o+(0.192284814559)*x[1]
        arg[1,0,0,0]=(-0.212291422817)*x[0]**o+(-0.16123490382)*x[0]+(-0.884220059537)*x[1]**o+(-0.40104775897)*x[1]
        arg[1,0,0,1]=(0.389745305389)*x[0]**o+(0.544302066739)*x[0]+(-0.504770065056)*x[1]**o+(0.903371132654)*x[1]
        arg[1,0,1,0]=(0.983764499357)*x[0]**o+(0.0707766278425)*x[0]+(-0.119851167133)*x[1]**o+(0.668490934432)*x[1]
        arg[1,0,1,1]=(-0.61333738947)*x[0]**o+(-0.515837139916)*x[0]+(-0.825015725806)*x[1]**o+(-0.858420009384)*x[1]
        arg[1,0,2,0]=(0.838488008532)*x[0]**o+(0.320020900158)*x[0]+(0.265140913892)*x[1]**o+(0.493460973766)*x[1]
        arg[1,0,2,1]=(-0.957171379585)*x[0]**o+(0.911667647687)*x[0]+(0.0634495769152)*x[1]**o+(-0.244783653159)*x[1]
        arg[1,1,0,0]=(0.854312398231)*x[0]**o+(0.0177604750376)*x[0]+(-0.263135825468)*x[1]**o+(0.196452389095)*x[1]
        arg[1,1,0,1]=(0.496952473267)*x[0]**o+(-0.505581680291)*x[0]+(0.227653536528)*x[1]**o+(-0.138354150837)*x[1]
        arg[1,1,1,0]=(0.569682536816)*x[0]**o+(-0.523174667456)*x[0]+(-0.434073415962)*x[1]**o+(-0.853667257479)*x[1]
        arg[1,1,1,1]=(0.636882167235)*x[0]**o+(-0.122206427025)*x[0]+(0.934361025063)*x[1]**o+(-0.43037160177)*x[1]
        arg[1,1,2,0]=(0.0798318147504)*x[0]**o+(0.825231320185)*x[0]+(0.810701650808)*x[1]**o+(-0.544356371381)*x[1]
        arg[1,1,2,1]=(-0.426710170704)*x[0]**o+(0.200697079677)*x[0]+(0.242027752153)*x[1]**o+(0.746762119874)*x[1]
        arg[1,2,0,0]=(-0.497091856216)*x[0]**o+(0.0953816987137)*x[0]+(-0.388615292775)*x[1]**o+(0.794962452203)*x[1]
        arg[1,2,0,1]=(0.484769626964)*x[0]**o+(0.700484219762)*x[0]+(-0.230377697673)*x[1]**o+(-0.110327221848)*x[1]
        arg[1,2,1,0]=(-0.791806541129)*x[0]**o+(0.0199911517557)*x[0]+(-0.441657020836)*x[1]**o+(-0.972999930466)*x[1]
        arg[1,2,1,1]=(-0.566750983493)*x[0]**o+(0.193689535753)*x[0]+(0.659310468709)*x[1]**o+(0.209045229813)*x[1]
        arg[1,2,2,0]=(0.852553900341)*x[0]**o+(-0.869275488172)*x[0]+(-0.308429715369)*x[1]**o+(0.096335995297)*x[1]
        arg[1,2,2,1]=(0.169351018917)*x[0]**o+(0.269577471686)*x[0]+(-0.226840840616)*x[1]**o+(-0.94229540867)*x[1]
        arg[1,3,0,0]=(0.481711489234)*x[0]**o+(0.293576608201)*x[0]+(0.101998881767)*x[1]**o+(-0.10730442531)*x[1]
        arg[1,3,0,1]=(-0.824144820539)*x[0]**o+(0.174754759038)*x[0]+(-0.251803196848)*x[1]**o+(-0.389656255459)*x[1]
        arg[1,3,1,0]=(0.438310633462)*x[0]**o+(-0.268051020546)*x[0]+(0.670456978527)*x[1]**o+(-0.547393352784)*x[1]
        arg[1,3,1,1]=(-0.95082246973)*x[0]**o+(0.739061218443)*x[0]+(0.209495253219)*x[1]**o+(-0.604801908581)*x[1]
        arg[1,3,2,0]=(0.696900724687)*x[0]**o+(0.501191311274)*x[0]+(0.982952568975)*x[1]**o+(0.1345704134)*x[1]
        arg[1,3,2,1]=(0.823865642457)*x[0]**o+(0.344980106206)*x[0]+(0.0740071873515)*x[1]**o+(-0.950460516779)*x[1]
        arg[1,4,0,0]=(-0.428525532814)*x[0]**o+(0.496012936411)*x[0]+(-0.0756953876324)*x[1]**o+(0.8382068924)*x[1]
        arg[1,4,0,1]=(0.950388644852)*x[0]**o+(-0.225116488186)*x[0]+(0.515802878592)*x[1]**o+(-0.51004230039)*x[1]
        arg[1,4,1,0]=(0.868733254296)*x[0]**o+(-0.506544709529)*x[0]+(0.28935579533)*x[1]**o+(0.100277736612)*x[1]
        arg[1,4,1,1]=(-0.104533384722)*x[0]**o+(0.598995288066)*x[0]+(-0.738946772839)*x[1]**o+(-0.23245555216)*x[1]
        arg[1,4,2,0]=(-0.286232179213)*x[0]**o+(-0.31267596034)*x[0]+(0.588681250431)*x[1]**o+(-0.579857978487)*x[1]
        arg[1,4,2,1]=(0.405666376198)*x[0]**o+(0.380029842735)*x[0]+(-0.646672003588)*x[1]**o+(0.419606979519)*x[1]
        arg[2,0,0,0]=(0.183075434304)*x[0]**o+(0.0929532345426)*x[0]+(-0.742007638693)*x[1]**o+(-0.521914593386)*x[1]
        arg[2,0,0,1]=(0.292878118412)*x[0]**o+(-0.427165220789)*x[0]+(0.540033289295)*x[1]**o+(0.997945616043)*x[1]
        arg[2,0,1,0]=(0.872801856197)*x[0]**o+(0.0427577769706)*x[0]+(0.595354626662)*x[1]**o+(-0.878459740609)*x[1]
        arg[2,0,1,1]=(-0.259988866684)*x[0]**o+(0.154043636253)*x[0]+(0.0580047878991)*x[1]**o+(0.40696744056)*x[1]
        arg[2,0,2,0]=(-0.602648518466)*x[0]**o+(0.30877596654)*x[0]+(-0.268563459489)*x[1]**o+(-0.526379663585)*x[1]
        arg[2,0,2,1]=(-0.375777994931)*x[0]**o+(-0.175440011709)*x[0]+(-0.578650003477)*x[1]**o+(0.822157683321)*x[1]
        arg[2,1,0,0]=(-0.44389226502)*x[0]**o+(0.683580346916)*x[0]+(0.471857559282)*x[1]**o+(-0.869624665398)*x[1]
        arg[2,1,0,1]=(0.768272315036)*x[0]**o+(-0.709688321973)*x[0]+(0.200727881948)*x[1]**o+(0.617326752785)*x[1]
        arg[2,1,1,0]=(-0.371169228982)*x[0]**o+(-0.0248248405104)*x[0]+(-0.899904749204)*x[1]**o+(0.796779883083)*x[1]
        arg[2,1,1,1]=(0.220088670807)*x[0]**o+(0.691718163435)*x[0]+(-0.0834793048429)*x[1]**o+(-0.748110928988)*x[1]
        arg[2,1,2,0]=(0.246045293748)*x[0]**o+(-0.669660154233)*x[0]+(0.601040129345)*x[1]**o+(-0.0893792369983)*x[1]
        arg[2,1,2,1]=(0.318708374425)*x[0]**o+(-0.973048583674)*x[0]+(-0.575891882633)*x[1]**o+(0.229050058038)*x[1]
        arg[2,2,0,0]=(0.358008410645)*x[0]**o+(-0.834388697702)*x[0]+(0.0931465346237)*x[1]**o+(0.503934981947)*x[1]
        arg[2,2,0,1]=(-0.257597986495)*x[0]**o+(-0.58070568528)*x[0]+(0.735541935653)*x[1]**o+(-0.704666216263)*x[1]
        arg[2,2,1,0]=(0.0207421250729)*x[0]**o+(-0.724644198871)*x[0]+(0.637124335031)*x[1]**o+(-0.547300138808)*x[1]
        arg[2,2,1,1]=(0.280179217797)*x[0]**o+(-0.387279903944)*x[0]+(0.516275341085)*x[1]**o+(0.781250973385)*x[1]
        arg[2,2,2,0]=(-0.97771111517)*x[0]**o+(-0.140997969015)*x[0]+(-0.495112130955)*x[1]**o+(0.151523726543)*x[1]
        arg[2,2,2,1]=(-0.48638258963)*x[0]**o+(-0.251363723104)*x[0]+(-0.484911659879)*x[1]**o+(-0.841285872322)*x[1]
        arg[2,3,0,0]=(-0.70159761565)*x[0]**o+(-0.661658396735)*x[0]+(-0.380920250292)*x[1]**o+(0.682598882025)*x[1]
        arg[2,3,0,1]=(0.477056088757)*x[0]**o+(0.697397619448)*x[0]+(0.28617323153)*x[1]**o+(-0.0443690469658)*x[1]
        arg[2,3,1,0]=(-0.823038323615)*x[0]**o+(-0.742146866184)*x[0]+(0.258604766787)*x[1]**o+(-0.740368164141)*x[1]
        arg[2,3,1,1]=(0.958751077013)*x[0]**o+(0.353976218746)*x[0]+(-0.954344840728)*x[1]**o+(0.436915389196)*x[1]
        arg[2,3,2,0]=(0.432342075923)*x[0]**o+(-0.0699761039653)*x[0]+(0.701490046186)*x[1]**o+(0.0888793443125)*x[1]
        arg[2,3,2,1]=(0.612536667525)*x[0]**o+(0.107433723026)*x[0]+(-0.611874456965)*x[1]**o+(-0.253338899867)*x[1]
        arg[2,4,0,0]=(-0.0234819568317)*x[0]**o+(-0.0642964170457)*x[0]+(0.556679188908)*x[1]**o+(0.865922901373)*x[1]
        arg[2,4,0,1]=(0.669235019483)*x[0]**o+(0.744361910609)*x[0]+(0.383397302403)*x[1]**o+(0.669587709042)*x[1]
        arg[2,4,1,0]=(-0.143461051192)*x[0]**o+(-0.0272388772724)*x[0]+(0.723868767782)*x[1]**o+(0.303617835707)*x[1]
        arg[2,4,1,1]=(-0.829765013546)*x[0]**o+(-0.784193619495)*x[0]+(0.384376849236)*x[1]**o+(0.280904709362)*x[1]
        arg[2,4,2,0]=(0.38737050935)*x[0]**o+(0.749614613361)*x[0]+(-0.340848293488)*x[1]**o+(0.96291850214)*x[1]
        arg[2,4,2,1]=(0.619896872679)*x[0]**o+(0.970694589398)*x[0]+(-0.300698305892)*x[1]**o+(0.794408153506)*x[1]
        arg[3,0,0,0]=(-0.808442463698)*x[0]**o+(-0.527555255968)*x[0]+(0.246776596713)*x[1]**o+(-0.719809811547)*x[1]
        arg[3,0,0,1]=(-0.390016002326)*x[0]**o+(-0.864168521951)*x[0]+(-0.154649478359)*x[1]**o+(0.210766031283)*x[1]
        arg[3,0,1,0]=(-0.941916382966)*x[0]**o+(-0.0352551994705)*x[0]+(-0.660026082442)*x[1]**o+(-0.869241000235)*x[1]
        arg[3,0,1,1]=(-0.287992443961)*x[0]**o+(0.292551506636)*x[0]+(-0.619953648778)*x[1]**o+(0.746095811695)*x[1]
        arg[3,0,2,0]=(-0.233659077726)*x[0]**o+(-0.568647415596)*x[0]+(-0.22048065705)*x[1]**o+(0.658105381132)*x[1]
        arg[3,0,2,1]=(0.871784854897)*x[0]**o+(-0.118389479971)*x[0]+(0.143290959871)*x[1]**o+(0.753198625961)*x[1]
        arg[3,1,0,0]=(0.993988451174)*x[0]**o+(0.451849763406)*x[0]+(-0.972805640824)*x[1]**o+(0.796893857099)*x[1]
        arg[3,1,0,1]=(-0.424353439197)*x[0]**o+(-0.575799061446)*x[0]+(-0.233936106648)*x[1]**o+(-0.451864349448)*x[1]
        arg[3,1,1,0]=(0.639821887614)*x[0]**o+(0.771041944334)*x[0]+(0.420297869214)*x[1]**o+(0.671734395857)*x[1]
        arg[3,1,1,1]=(0.398121897075)*x[0]**o+(0.647679737081)*x[0]+(-0.544457308885)*x[1]**o+(0.58955287395)*x[1]
        arg[3,1,2,0]=(-0.118222351008)*x[0]**o+(-0.943807502172)*x[0]+(0.866866987138)*x[1]**o+(-0.0793340319339)*x[1]
        arg[3,1,2,1]=(-0.157286204244)*x[0]**o+(-0.789214982692)*x[0]+(0.646008960971)*x[1]**o+(0.866687256647)*x[1]
        arg[3,2,0,0]=(-0.620625598995)*x[0]**o+(-0.12874444072)*x[0]+(0.4402928966)*x[1]**o+(-0.0344340173898)*x[1]
        arg[3,2,0,1]=(-0.714597102332)*x[0]**o+(0.730206152989)*x[0]+(0.795717819655)*x[1]**o+(0.0543647593833)*x[1]
        arg[3,2,1,0]=(0.698721346048)*x[0]**o+(0.599204472096)*x[0]+(0.323529910402)*x[1]**o+(0.362678869119)*x[1]
        arg[3,2,1,1]=(0.976107824554)*x[0]**o+(0.559889392243)*x[0]+(0.240016231718)*x[1]**o+(0.343322878078)*x[1]
        arg[3,2,2,0]=(0.609695396838)*x[0]**o+(0.786331471916)*x[0]+(-0.965489152976)*x[1]**o+(0.553653504493)*x[1]
        arg[3,2,2,1]=(0.897952533418)*x[0]**o+(0.227242627929)*x[0]+(-0.121752169524)*x[1]**o+(0.844617689997)*x[1]
        arg[3,3,0,0]=(-0.249698211819)*x[0]**o+(-0.400095108874)*x[0]+(-0.44300562221)*x[1]**o+(0.387777895056)*x[1]
        arg[3,3,0,1]=(0.626469203641)*x[0]**o+(0.741893669108)*x[0]+(0.261882509899)*x[1]**o+(-0.726628233604)*x[1]
        arg[3,3,1,0]=(-0.525223500134)*x[0]**o+(0.918314547123)*x[0]+(0.426441176651)*x[1]**o+(0.338058005974)*x[1]
        arg[3,3,1,1]=(-0.31358142709)*x[0]**o+(-0.411070225036)*x[0]+(0.0902340021221)*x[1]**o+(0.404135075929)*x[1]
        arg[3,3,2,0]=(0.528842663849)*x[0]**o+(0.474768688828)*x[0]+(0.697643730083)*x[1]**o+(-0.170335125341)*x[1]
        arg[3,3,2,1]=(0.875221054674)*x[0]**o+(-0.441830492507)*x[0]+(-0.60312069438)*x[1]**o+(0.664895429753)*x[1]
        arg[3,4,0,0]=(0.563422972203)*x[0]**o+(-0.595004250424)*x[0]+(0.888927367905)*x[1]**o+(0.553640735136)*x[1]
        arg[3,4,0,1]=(0.79563891979)*x[0]**o+(-0.461626171695)*x[0]+(0.700194671524)*x[1]**o+(-0.687505010217)*x[1]
        arg[3,4,1,0]=(0.655248139354)*x[0]**o+(0.588175755435)*x[0]+(-0.29236593461)*x[1]**o+(0.575186295035)*x[1]
        arg[3,4,1,1]=(0.405730566113)*x[0]**o+(-0.161117488254)*x[0]+(0.912224633472)*x[1]**o+(0.0292423764587)*x[1]
        arg[3,4,2,0]=(-0.405819525278)*x[0]**o+(-0.362840757195)*x[0]+(0.961950188511)*x[1]**o+(0.205213542149)*x[1]
        arg[3,4,2,1]=(0.703066280254)*x[0]**o+(-0.936055321446)*x[0]+(0.861070629413)*x[1]**o+(0.0856428368256)*x[1]
        ref[0,0,0,0]=(-0.216088888802)/(o+1.)+(-0.272034496148)
        ref[0,0,0,1]=(0.959350420679)/(o+1.)+(-0.062880226704)
        ref[0,0,1,0]=(-0.525974126338)/(o+1.)+(0.317500886397)
        ref[0,0,1,1]=(-0.471892502528)/(o+1.)+(0.0896700005967)
        ref[0,0,2,0]=(0.454397005927)/(o+1.)+(0.131147481736)
        ref[0,0,2,1]=(-0.314324528348)/(o+1.)+(0.408809451335)
        ref[0,1,0,0]=(-0.411566405684)/(o+1.)+(0.764367546122)
        ref[0,1,0,1]=(-0.126547063321)/(o+1.)+(-0.129100420455)
        ref[0,1,1,0]=(1.53166678447)/(o+1.)+(0.247092500423)
        ref[0,1,1,1]=(-0.444616312725)/(o+1.)+(0.306156441165)
        ref[0,1,2,0]=(0.134683003633)/(o+1.)+(0.482717575736)
        ref[0,1,2,1]=(0.294679478965)/(o+1.)+(-0.457332009538)
        ref[0,2,0,0]=(0.590775544538)/(o+1.)+(-0.0307053101973)
        ref[0,2,0,1]=(-0.866197507498)/(o+1.)+(-0.332194509245)
        ref[0,2,1,0]=(0.249469805723)/(o+1.)+(-0.193030816143)
        ref[0,2,1,1]=(0.435433694525)/(o+1.)+(-0.436614114)
        ref[0,2,2,0]=(-0.0808985767498)/(o+1.)+(-0.713951965774)
        ref[0,2,2,1]=(0.0685144188336)/(o+1.)+(-0.561860153224)
        ref[0,3,0,0]=(-0.0227125573528)/(o+1.)+(-0.0475152415978)
        ref[0,3,0,1]=(0.550671217223)/(o+1.)+(-0.220831717723)
        ref[0,3,1,0]=(-0.0965320008409)/(o+1.)+(-0.213438758409)
        ref[0,3,1,1]=(0.0638208493064)/(o+1.)+(0.149919999309)
        ref[0,3,2,0]=(-0.942379946856)/(o+1.)+(-0.554981593775)
        ref[0,3,2,1]=(-0.0418528566735)/(o+1.)+(-0.164023084154)
        ref[0,4,0,0]=(-1.84639644319)/(o+1.)+(0.0247926648306)
        ref[0,4,0,1]=(-0.0896339469252)/(o+1.)+(-0.102858458939)
        ref[0,4,1,0]=(0.749451138484)/(o+1.)+(0.348124395042)
        ref[0,4,1,1]=(-0.165019322616)/(o+1.)+(-0.0435147480027)
        ref[0,4,2,0]=(-0.0874070557597)/(o+1.)+(0.033388938077)
        ref[0,4,2,1]=(-1.00428948901)/(o+1.)+(0.428770577935)
        ref[1,0,0,0]=(-1.09651148235)/(o+1.)+(-0.281141331395)
        ref[1,0,0,1]=(-0.115024759667)/(o+1.)+(0.723836599697)
        ref[1,0,1,0]=(0.863913332224)/(o+1.)+(0.369633781137)
        ref[1,0,1,1]=(-1.43835311528)/(o+1.)+(-0.68712857465)
        ref[1,0,2,0]=(1.10362892242)/(o+1.)+(0.406740936962)
        ref[1,0,2,1]=(-0.893721802669)/(o+1.)+(0.333441997264)
        ref[1,1,0,0]=(0.591176572763)/(o+1.)+(0.107106432066)
        ref[1,1,0,1]=(0.724606009795)/(o+1.)+(-0.321967915564)
        ref[1,1,1,0]=(0.135609120854)/(o+1.)+(-0.688420962467)
        ref[1,1,1,1]=(1.5712431923)/(o+1.)+(-0.276289014397)
        ref[1,1,2,0]=(0.890533465558)/(o+1.)+(0.140437474402)
        ref[1,1,2,1]=(-0.184682418551)/(o+1.)+(0.473729599776)
        ref[1,2,0,0]=(-0.88570714899)/(o+1.)+(0.445172075458)
        ref[1,2,0,1]=(0.254391929291)/(o+1.)+(0.295078498957)
        ref[1,2,1,0]=(-1.23346356196)/(o+1.)+(-0.476504389355)
        ref[1,2,1,1]=(0.0925594852161)/(o+1.)+(0.201367382783)
        ref[1,2,2,0]=(0.544124184972)/(o+1.)+(-0.386469746437)
        ref[1,2,2,1]=(-0.057489821699)/(o+1.)+(-0.336358968492)
        ref[1,3,0,0]=(0.583710371001)/(o+1.)+(0.0931360914458)
        ref[1,3,0,1]=(-1.07594801739)/(o+1.)+(-0.107450748211)
        ref[1,3,1,0]=(1.10876761199)/(o+1.)+(-0.407722186665)
        ref[1,3,1,1]=(-0.741327216511)/(o+1.)+(0.0671296549307)
        ref[1,3,2,0]=(1.67985329366)/(o+1.)+(0.317880862337)
        ref[1,3,2,1]=(0.897872829808)/(o+1.)+(-0.302740205287)
        ref[1,4,0,0]=(-0.504220920447)/(o+1.)+(0.667109914405)
        ref[1,4,0,1]=(1.46619152344)/(o+1.)+(-0.367579394288)
        ref[1,4,1,0]=(1.15808904963)/(o+1.)+(-0.203133486458)
        ref[1,4,1,1]=(-0.843480157561)/(o+1.)+(0.183269867953)
        ref[1,4,2,0]=(0.302449071217)/(o+1.)+(-0.446266969414)
        ref[1,4,2,1]=(-0.24100562739)/(o+1.)+(0.399818411127)
        ref[2,0,0,0]=(-0.558932204389)/(o+1.)+(-0.214480679421)
        ref[2,0,0,1]=(0.832911407707)/(o+1.)+(0.285390197627)
        ref[2,0,1,0]=(1.46815648286)/(o+1.)+(-0.417850981819)
        ref[2,0,1,1]=(-0.201984078785)/(o+1.)+(0.280505538406)
        ref[2,0,2,0]=(-0.871211977955)/(o+1.)+(-0.108801848522)
        ref[2,0,2,1]=(-0.954427998409)/(o+1.)+(0.323358835806)
        ref[2,1,0,0]=(0.0279652942615)/(o+1.)+(-0.0930221592407)
        ref[2,1,0,1]=(0.969000196983)/(o+1.)+(-0.0461807845941)
        ref[2,1,1,0]=(-1.27107397819)/(o+1.)+(0.385977521287)
        ref[2,1,1,1]=(0.136609365964)/(o+1.)+(-0.0281963827764)
        ref[2,1,2,0]=(0.847085423093)/(o+1.)+(-0.379519695616)
        ref[2,1,2,1]=(-0.257183508208)/(o+1.)+(-0.371999262818)
        ref[2,2,0,0]=(0.451154945268)/(o+1.)+(-0.165226857878)
        ref[2,2,0,1]=(0.477943949158)/(o+1.)+(-0.642685950772)
        ref[2,2,1,0]=(0.657866460103)/(o+1.)+(-0.635972168839)
        ref[2,2,1,1]=(0.796454558882)/(o+1.)+(0.19698553472)
        ref[2,2,2,0]=(-1.47282324613)/(o+1.)+(0.00526287876389)
        ref[2,2,2,1]=(-0.971294249509)/(o+1.)+(-0.546324797713)
        ref[2,3,0,0]=(-1.08251786594)/(o+1.)+(0.0104702426452)
        ref[2,3,0,1]=(0.763229320287)/(o+1.)+(0.326514286241)
        ref[2,3,1,0]=(-0.564433556829)/(o+1.)+(-0.741257515163)
        ref[2,3,1,1]=(0.00440623628526)/(o+1.)+(0.395445803971)
        ref[2,3,2,0]=(1.13383212211)/(o+1.)+(0.0094516201736)
        ref[2,3,2,1]=(0.000662210559809)/(o+1.)+(-0.0729525884203)
        ref[2,4,0,0]=(0.533197232076)/(o+1.)+(0.400813242164)
        ref[2,4,0,1]=(1.05263232189)/(o+1.)+(0.706974809826)
        ref[2,4,1,0]=(0.580407716589)/(o+1.)+(0.138189479217)
        ref[2,4,1,1]=(-0.44538816431)/(o+1.)+(-0.251644455066)
        ref[2,4,2,0]=(0.0465222158616)/(o+1.)+(0.85626655775)
        ref[2,4,2,1]=(0.319198566788)/(o+1.)+(0.882551371452)
        ref[3,0,0,0]=(-0.561665866984)/(o+1.)+(-0.623682533758)
        ref[3,0,0,1]=(-0.544665480685)/(o+1.)+(-0.326701245334)
        ref[3,0,1,0]=(-1.60194246541)/(o+1.)+(-0.452248099853)
        ref[3,0,1,1]=(-0.907946092739)/(o+1.)+(0.519323659165)
        ref[3,0,2,0]=(-0.454139734776)/(o+1.)+(0.0447289827682)
        ref[3,0,2,1]=(1.01507581477)/(o+1.)+(0.317404572995)
        ref[3,1,0,0]=(0.0211828103496)/(o+1.)+(0.624371810253)
        ref[3,1,0,1]=(-0.658289545845)/(o+1.)+(-0.513831705447)
        ref[3,1,1,0]=(1.06011975683)/(o+1.)+(0.721388170095)
        ref[3,1,1,1]=(-0.14633541181)/(o+1.)+(0.618616305516)
        ref[3,1,2,0]=(0.74864463613)/(o+1.)+(-0.511570767053)
        ref[3,1,2,1]=(0.488722756727)/(o+1.)+(0.0387361369774)
        ref[3,2,0,0]=(-0.180332702395)/(o+1.)+(-0.0815892290549)
        ref[3,2,0,1]=(0.0811207173227)/(o+1.)+(0.392285456186)
        ref[3,2,1,0]=(1.02225125645)/(o+1.)+(0.480941670608)
        ref[3,2,1,1]=(1.21612405627)/(o+1.)+(0.45160613516)
        ref[3,2,2,0]=(-0.355793756138)/(o+1.)+(0.669992488205)
        ref[3,2,2,1]=(0.776200363894)/(o+1.)+(0.535930158963)
        ref[3,3,0,0]=(-0.692703834029)/(o+1.)+(-0.00615860690903)
        ref[3,3,0,1]=(0.88835171354)/(o+1.)+(0.00763271775166)
        ref[3,3,1,0]=(-0.0987823234827)/(o+1.)+(0.628186276549)
        ref[3,3,1,1]=(-0.223347424968)/(o+1.)+(-0.00346757455392)
        ref[3,3,2,0]=(1.22648639393)/(o+1.)+(0.152216781743)
        ref[3,3,2,1]=(0.272100360294)/(o+1.)+(0.111532468623)
        ref[3,4,0,0]=(1.45235034011)/(o+1.)+(-0.0206817576441)
        ref[3,4,0,1]=(1.49583359131)/(o+1.)+(-0.574565590956)
        ref[3,4,1,0]=(0.362882204744)/(o+1.)+(0.581681025235)
        ref[3,4,1,1]=(1.31795519959)/(o+1.)+(-0.0659375558974)
        ref[3,4,2,0]=(0.556130663233)/(o+1.)+(-0.078813607523)
        ref[3,4,2,1]=(1.56413690967)/(o+1.)+(-0.42520624231)
      else:
        arg[0,0,0,0]=(0.956854727968)*x[0]**o+(-0.43263548815)*x[0]+(-0.1485543705)*x[1]**o+(-0.411060684134)*x[1]+(-0.306829513954)*x[2]**o+(0.89077924839)*x[2]
        arg[0,0,0,1]=(0.125905642347)*x[0]**o+(-0.645491335766)*x[0]+(-0.237571989123)*x[1]**o+(-0.239301942511)*x[1]+(-0.384215893518)*x[2]**o+(-0.371218179274)*x[2]
        arg[0,0,1,0]=(0.0204273419969)*x[0]**o+(0.354631489501)*x[0]+(0.235086936835)*x[1]**o+(0.647551574475)*x[1]+(-0.729066874572)*x[2]**o+(0.177733791046)*x[2]
        arg[0,0,1,1]=(-0.0938492166558)*x[0]**o+(-0.484339074939)*x[0]+(0.619333193491)*x[1]**o+(0.809157671781)*x[1]+(-0.28367956249)*x[2]**o+(-0.1758249267)*x[2]
        arg[0,0,2,0]=(-0.334944259519)*x[0]**o+(-0.126145567301)*x[0]+(-0.646972616146)*x[1]**o+(0.20287553181)*x[1]+(-0.383816067326)*x[2]**o+(-0.0265591031498)*x[2]
        arg[0,0,2,1]=(0.103308564583)*x[0]**o+(0.016108181951)*x[0]+(-0.367450237136)*x[1]**o+(0.657618788015)*x[1]+(-0.8327532511)*x[2]**o+(-0.267286876849)*x[2]
        arg[0,1,0,0]=(0.688995722152)*x[0]**o+(-0.982012541087)*x[0]+(-0.232339851099)*x[1]**o+(-0.097863461266)*x[1]+(0.25916639806)*x[2]**o+(0.0538637018789)*x[2]
        arg[0,1,0,1]=(0.446698337789)*x[0]**o+(-0.309434573629)*x[0]+(0.617004214016)*x[1]**o+(0.747666594435)*x[1]+(-0.869684246717)*x[2]**o+(0.75016186442)*x[2]
        arg[0,1,1,0]=(-0.559974353252)*x[0]**o+(0.966642794905)*x[0]+(-0.228473000256)*x[1]**o+(-0.435093965632)*x[1]+(0.426986383284)*x[2]**o+(0.131300817321)*x[2]
        arg[0,1,1,1]=(-0.696027598385)*x[0]**o+(-0.804778535506)*x[0]+(0.930171049019)*x[1]**o+(-0.900826020873)*x[1]+(0.844689166523)*x[2]**o+(0.0349718139676)*x[2]
        arg[0,1,2,0]=(0.359170897328)*x[0]**o+(0.992944582308)*x[0]+(0.579162388074)*x[1]**o+(0.743410851741)*x[1]+(-0.204325306226)*x[2]**o+(-0.992005484878)*x[2]
        arg[0,1,2,1]=(-0.883554988936)*x[0]**o+(0.486947073334)*x[0]+(-0.78078860326)*x[1]**o+(-0.134196299284)*x[1]+(-0.980852067031)*x[2]**o+(0.453023187355)*x[2]
        arg[0,2,0,0]=(0.804524524371)*x[0]**o+(0.397924403155)*x[0]+(0.712297039229)*x[1]**o+(0.713485622558)*x[1]+(-0.966783610543)*x[2]**o+(0.497869404576)*x[2]
        arg[0,2,0,1]=(-0.564706930001)*x[0]**o+(-0.482026983244)*x[0]+(0.284926154564)*x[1]**o+(0.844423322149)*x[1]+(0.444203847061)*x[2]**o+(0.333244188639)*x[2]
        arg[0,2,1,0]=(-0.630410083438)*x[0]**o+(-0.831218126629)*x[0]+(0.400237927956)*x[1]**o+(-0.70518378137)*x[1]+(-0.803754805917)*x[2]**o+(0.60059455602)*x[2]
        arg[0,2,1,1]=(0.198272547853)*x[0]**o+(0.0840356091496)*x[0]+(0.114385146155)*x[1]**o+(0.731057038288)*x[1]+(-0.976056408523)*x[2]**o+(-0.132565954304)*x[2]
        arg[0,2,2,0]=(0.983300141657)*x[0]**o+(-0.733168195162)*x[0]+(-0.948802450335)*x[1]**o+(-0.756616078178)*x[1]+(0.450302374372)*x[2]**o+(0.802115535717)*x[2]
        arg[0,2,2,1]=(-0.680853503774)*x[0]**o+(-0.910402862928)*x[0]+(0.583775613868)*x[1]**o+(0.741516645611)*x[1]+(-0.550498467533)*x[2]**o+(-0.610648481959)*x[2]
        arg[0,3,0,0]=(0.981909313325)*x[0]**o+(0.148145112359)*x[0]+(-0.989721117518)*x[1]**o+(0.0246558730659)*x[1]+(0.721439979982)*x[2]**o+(0.896052837889)*x[2]
        arg[0,3,0,1]=(0.109831691542)*x[0]**o+(-0.349555366958)*x[0]+(-0.54822642636)*x[1]**o+(0.121919022172)*x[1]+(-0.580228247481)*x[2]**o+(0.564276455713)*x[2]
        arg[0,3,1,0]=(0.106091093353)*x[0]**o+(0.410111742327)*x[0]+(-0.0669026478824)*x[1]**o+(-0.662684907574)*x[1]+(-0.351815520918)*x[2]**o+(-0.324802958038)*x[2]
        arg[0,3,1,1]=(0.857685803258)*x[0]**o+(-0.782627954043)*x[0]+(-0.555036389439)*x[1]**o+(-0.0126716685608)*x[1]+(0.260461379777)*x[2]**o+(0.117590260591)*x[2]
        arg[0,3,2,0]=(-0.987873674604)*x[0]**o+(0.0779662295512)*x[0]+(-0.928229202803)*x[1]**o+(0.508015712707)*x[1]+(0.248638760509)*x[2]**o+(-0.712158091156)*x[2]
        arg[0,3,2,1]=(-0.977973356138)*x[0]**o+(0.270015563339)*x[0]+(0.811157094127)*x[1]**o+(-0.756460384936)*x[1]+(-0.735557557504)*x[2]**o+(0.780333584929)*x[2]
        arg[0,4,0,0]=(-0.762444411646)*x[0]**o+(0.386894014691)*x[0]+(0.763898738855)*x[1]**o+(-0.422896581278)*x[1]+(0.428591445563)*x[2]**o+(-0.617626926396)*x[2]
        arg[0,4,0,1]=(-0.269524167286)*x[0]**o+(0.996895823788)*x[0]+(0.771396605337)*x[1]**o+(-0.993794940262)*x[1]+(-0.534803587481)*x[2]**o+(-0.928627946122)*x[2]
        arg[0,4,1,0]=(0.978497459584)*x[0]**o+(-0.343496507186)*x[0]+(-0.250581087884)*x[1]**o+(-0.855743445852)*x[1]+(-0.127512246124)*x[2]**o+(-0.133936486447)*x[2]
        arg[0,4,1,1]=(-0.844450818471)*x[0]**o+(0.263329730105)*x[0]+(0.411212783735)*x[1]**o+(-0.816379748952)*x[1]+(-0.443806901857)*x[2]**o+(0.1262969097)*x[2]
        arg[0,4,2,0]=(-0.326938381124)*x[0]**o+(-0.707054728551)*x[0]+(-0.725513874889)*x[1]**o+(0.655254977879)*x[1]+(0.772734170665)*x[2]**o+(-0.963856465024)*x[2]
        arg[0,4,2,1]=(0.896415185133)*x[0]**o+(-0.722431478737)*x[0]+(0.942058336843)*x[1]**o+(0.838230971246)*x[1]+(-0.279821311439)*x[2]**o+(-0.330130961206)*x[2]
        arg[1,0,0,0]=(-0.466537465363)*x[0]**o+(0.86493624537)*x[0]+(-0.943049537777)*x[1]**o+(0.658727518981)*x[1]+(-0.130355508544)*x[2]**o+(-0.158245981876)*x[2]
        arg[1,0,0,1]=(0.247046604605)*x[0]**o+(-0.287683750248)*x[0]+(0.266630203142)*x[1]**o+(0.372409648579)*x[1]+(-0.239411725166)*x[2]**o+(0.537339451105)*x[2]
        arg[1,0,1,0]=(-0.232312387253)*x[0]**o+(-0.0493954012747)*x[0]+(-0.461085479525)*x[1]**o+(-0.0744729096417)*x[1]+(-0.795163064273)*x[2]**o+(0.613075576806)*x[2]
        arg[1,0,1,1]=(-0.68570821748)*x[0]**o+(-0.492507082229)*x[0]+(-0.0673636097828)*x[1]**o+(0.569119992806)*x[1]+(-0.227653595613)*x[2]**o+(0.0337433636837)*x[2]
        arg[1,0,2,0]=(-0.124371489405)*x[0]**o+(-0.726635218442)*x[0]+(-0.835418167414)*x[1]**o+(-0.00414597707753)*x[1]+(0.8365917104)*x[2]**o+(-0.507314639615)*x[2]
        arg[1,0,2,1]=(-0.973228057052)*x[0]**o+(0.123228228923)*x[0]+(0.545712842666)*x[1]**o+(-0.451331566047)*x[1]+(0.299501907067)*x[2]**o+(-0.24898389883)*x[2]
        arg[1,1,0,0]=(-0.940690906836)*x[0]**o+(-0.357571197214)*x[0]+(0.98643983012)*x[1]**o+(-0.82113140872)*x[1]+(0.460009285424)*x[2]**o+(-0.0298714172583)*x[2]
        arg[1,1,0,1]=(-0.881543245921)*x[0]**o+(-0.640741814727)*x[0]+(-0.749244677096)*x[1]**o+(0.47161101909)*x[1]+(0.712893375828)*x[2]**o+(0.600151133959)*x[2]
        arg[1,1,1,0]=(-0.217214692405)*x[0]**o+(-0.504211076658)*x[0]+(0.857127736087)*x[1]**o+(0.0439761116172)*x[1]+(0.956380848189)*x[2]**o+(0.698482155013)*x[2]
        arg[1,1,1,1]=(-0.288969728734)*x[0]**o+(-0.63960886193)*x[0]+(-0.741293494258)*x[1]**o+(-0.125034323955)*x[1]+(-0.608855286755)*x[2]**o+(0.462427737944)*x[2]
        arg[1,1,2,0]=(0.126242043984)*x[0]**o+(0.811846484451)*x[0]+(-0.522920862127)*x[1]**o+(0.678142081935)*x[1]+(0.0392586436898)*x[2]**o+(0.403908304728)*x[2]
        arg[1,1,2,1]=(-0.290103861843)*x[0]**o+(0.336130241928)*x[0]+(-0.872641307678)*x[1]**o+(-0.372020796526)*x[1]+(0.172572077906)*x[2]**o+(-0.789431556966)*x[2]
        arg[1,2,0,0]=(-0.158114079913)*x[0]**o+(0.954919898537)*x[0]+(-0.711834968234)*x[1]**o+(-0.390487270434)*x[1]+(-0.086581514295)*x[2]**o+(0.442047084988)*x[2]
        arg[1,2,0,1]=(-0.914034594356)*x[0]**o+(0.521137372279)*x[0]+(0.12454416569)*x[1]**o+(0.627815188423)*x[1]+(0.770244577882)*x[2]**o+(-0.846691748091)*x[2]
        arg[1,2,1,0]=(0.769631824465)*x[0]**o+(0.988840796717)*x[0]+(-0.401836868027)*x[1]**o+(0.401846957832)*x[1]+(-0.877826997285)*x[2]**o+(0.857725824229)*x[2]
        arg[1,2,1,1]=(0.173539365276)*x[0]**o+(0.373206670706)*x[0]+(-0.597288730366)*x[1]**o+(-0.770220794555)*x[1]+(-0.586510454467)*x[2]**o+(-0.569370842441)*x[2]
        arg[1,2,2,0]=(0.673765155818)*x[0]**o+(-0.642914516709)*x[0]+(0.794438851788)*x[1]**o+(-0.17557754186)*x[1]+(0.997565143863)*x[2]**o+(0.722460144483)*x[2]
        arg[1,2,2,1]=(-0.878452482718)*x[0]**o+(-0.846862156166)*x[0]+(0.462351027504)*x[1]**o+(0.267435606154)*x[1]+(-0.452135086983)*x[2]**o+(0.701577407252)*x[2]
        arg[1,3,0,0]=(-0.0557389135964)*x[0]**o+(-0.781831460089)*x[0]+(0.214846092931)*x[1]**o+(0.597436260353)*x[1]+(0.713968428327)*x[2]**o+(-0.456335676572)*x[2]
        arg[1,3,0,1]=(0.288266531585)*x[0]**o+(-0.259529556521)*x[0]+(0.0286107975184)*x[1]**o+(0.465095455484)*x[1]+(-0.747648909025)*x[2]**o+(0.682866470382)*x[2]
        arg[1,3,1,0]=(-0.720339104668)*x[0]**o+(0.961100526257)*x[0]+(0.0814885381756)*x[1]**o+(-0.448775223649)*x[1]+(0.250972143198)*x[2]**o+(-0.240231439947)*x[2]
        arg[1,3,1,1]=(-0.459305786748)*x[0]**o+(0.272222143865)*x[0]+(-0.499164099401)*x[1]**o+(-0.225162547271)*x[1]+(-0.92261124167)*x[2]**o+(-0.436659790026)*x[2]
        arg[1,3,2,0]=(0.890752966329)*x[0]**o+(-0.775207922011)*x[0]+(0.715960284721)*x[1]**o+(0.843422841789)*x[1]+(0.513788850466)*x[2]**o+(-0.325733710281)*x[2]
        arg[1,3,2,1]=(0.958726172433)*x[0]**o+(-0.743672393249)*x[0]+(-0.523783133945)*x[1]**o+(0.133861976051)*x[1]+(-0.109230940116)*x[2]**o+(0.819716327751)*x[2]
        arg[1,4,0,0]=(0.54439729551)*x[0]**o+(0.0763503792087)*x[0]+(0.0634218657498)*x[1]**o+(-0.405471845044)*x[1]+(0.316327091756)*x[2]**o+(-0.539493863939)*x[2]
        arg[1,4,0,1]=(0.527330373182)*x[0]**o+(0.77531447123)*x[0]+(-0.481869315442)*x[1]**o+(0.556927692362)*x[1]+(-0.267215098309)*x[2]**o+(0.958919406465)*x[2]
        arg[1,4,1,0]=(0.540950372012)*x[0]**o+(-0.95002626611)*x[0]+(-0.523544509893)*x[1]**o+(-0.583056635057)*x[1]+(-0.00288409928193)*x[2]**o+(0.671385098427)*x[2]
        arg[1,4,1,1]=(-0.39802076092)*x[0]**o+(-0.583264022146)*x[0]+(-0.656261495961)*x[1]**o+(-0.527677078418)*x[1]+(-0.2632498932)*x[2]**o+(-0.537632951507)*x[2]
        arg[1,4,2,0]=(-0.77176416566)*x[0]**o+(0.812696440735)*x[0]+(0.0925789242005)*x[1]**o+(0.299698967601)*x[1]+(-0.0185377066587)*x[2]**o+(-0.883247318682)*x[2]
        arg[1,4,2,1]=(0.360949407452)*x[0]**o+(-0.463659754812)*x[0]+(0.799066387132)*x[1]**o+(0.0318151346916)*x[1]+(-0.380675976437)*x[2]**o+(-0.299341342112)*x[2]
        arg[2,0,0,0]=(-0.985332287141)*x[0]**o+(0.33398206098)*x[0]+(0.39204502373)*x[1]**o+(-0.586826360517)*x[1]+(0.868240161369)*x[2]**o+(-0.257897525907)*x[2]
        arg[2,0,0,1]=(-0.801369266967)*x[0]**o+(0.333496542809)*x[0]+(0.459444994717)*x[1]**o+(-0.950297568688)*x[1]+(-0.760722281765)*x[2]**o+(0.0933850221377)*x[2]
        arg[2,0,1,0]=(0.53372811812)*x[0]**o+(0.999551098511)*x[0]+(-0.761650890163)*x[1]**o+(-0.80456940786)*x[1]+(-0.452089806623)*x[2]**o+(-0.40841169158)*x[2]
        arg[2,0,1,1]=(-0.782294374468)*x[0]**o+(0.614736817827)*x[0]+(0.262995910218)*x[1]**o+(0.067349886804)*x[1]+(-0.167494656909)*x[2]**o+(-0.730462378769)*x[2]
        arg[2,0,2,0]=(0.261001857862)*x[0]**o+(0.959318765248)*x[0]+(-0.177224016298)*x[1]**o+(-0.979637664014)*x[1]+(0.858937930292)*x[2]**o+(-0.627116377798)*x[2]
        arg[2,0,2,1]=(-0.238083578675)*x[0]**o+(0.586367991705)*x[0]+(0.747378561289)*x[1]**o+(-0.540178926415)*x[1]+(-0.519781240185)*x[2]**o+(-0.42715014981)*x[2]
        arg[2,1,0,0]=(-0.21022574437)*x[0]**o+(-0.292359099348)*x[0]+(0.449614855785)*x[1]**o+(-0.416904619153)*x[1]+(0.507537288404)*x[2]**o+(0.943584716996)*x[2]
        arg[2,1,0,1]=(-0.410591525695)*x[0]**o+(-0.258796394127)*x[0]+(-0.474714731399)*x[1]**o+(-0.518288451462)*x[1]+(-0.15640903664)*x[2]**o+(0.97242574307)*x[2]
        arg[2,1,1,0]=(-0.930865834892)*x[0]**o+(0.961887630616)*x[0]+(0.688706862295)*x[1]**o+(-0.375163634404)*x[1]+(0.378954001649)*x[2]**o+(-0.747140890466)*x[2]
        arg[2,1,1,1]=(0.0454701384411)*x[0]**o+(-0.37651801332)*x[0]+(0.0808750051583)*x[1]**o+(-0.530355850075)*x[1]+(0.769675417234)*x[2]**o+(-0.767046792016)*x[2]
        arg[2,1,2,0]=(0.179573418696)*x[0]**o+(-0.0547918709025)*x[0]+(0.739792606043)*x[1]**o+(0.849619562235)*x[1]+(-0.92840796838)*x[2]**o+(-0.302941296337)*x[2]
        arg[2,1,2,1]=(0.160657807137)*x[0]**o+(0.551943403654)*x[0]+(0.649164334204)*x[1]**o+(0.298882266215)*x[1]+(0.33872360276)*x[2]**o+(-0.558910868754)*x[2]
        arg[2,2,0,0]=(-0.704330511528)*x[0]**o+(-0.326943167116)*x[0]+(0.626863819473)*x[1]**o+(-0.107147718189)*x[1]+(0.00668670393016)*x[2]**o+(0.572223662188)*x[2]
        arg[2,2,0,1]=(0.879695137799)*x[0]**o+(-0.313488700497)*x[0]+(-0.113134766565)*x[1]**o+(-0.131007759381)*x[1]+(0.55460969593)*x[2]**o+(0.898706374575)*x[2]
        arg[2,2,1,0]=(0.329932124981)*x[0]**o+(-0.624965470976)*x[0]+(0.82906774364)*x[1]**o+(-0.681403786977)*x[1]+(-0.447119491747)*x[2]**o+(0.143763601614)*x[2]
        arg[2,2,1,1]=(0.0367746207528)*x[0]**o+(0.47553259605)*x[0]+(-0.517723045973)*x[1]**o+(0.263932783195)*x[1]+(-0.131032589284)*x[2]**o+(-0.132729291841)*x[2]
        arg[2,2,2,0]=(-0.108349266564)*x[0]**o+(-0.684748849483)*x[0]+(0.203271582231)*x[1]**o+(-0.860842085896)*x[1]+(0.469684577785)*x[2]**o+(0.195736308182)*x[2]
        arg[2,2,2,1]=(0.915666799349)*x[0]**o+(0.383338953756)*x[0]+(-0.554011249579)*x[1]**o+(0.232285573367)*x[1]+(0.39288151674)*x[2]**o+(0.692592233016)*x[2]
        arg[2,3,0,0]=(0.81179207663)*x[0]**o+(0.786809236572)*x[0]+(-0.814758445611)*x[1]**o+(-0.78461653396)*x[1]+(0.171435802778)*x[2]**o+(0.660186909193)*x[2]
        arg[2,3,0,1]=(0.752694523669)*x[0]**o+(0.301473785296)*x[0]+(0.610379955695)*x[1]**o+(0.573729399495)*x[1]+(0.225983751859)*x[2]**o+(-0.349530143446)*x[2]
        arg[2,3,1,0]=(-0.647825490424)*x[0]**o+(-0.232573858642)*x[0]+(0.773519085771)*x[1]**o+(-0.224497979868)*x[1]+(-0.681726885952)*x[2]**o+(-0.217968759388)*x[2]
        arg[2,3,1,1]=(0.391888417642)*x[0]**o+(-0.240556139087)*x[0]+(-0.852397797723)*x[1]**o+(0.318342758913)*x[1]+(-0.773804983598)*x[2]**o+(-0.957590491908)*x[2]
        arg[2,3,2,0]=(0.7126582007)*x[0]**o+(0.806022164546)*x[0]+(0.232348951782)*x[1]**o+(-0.985774366344)*x[1]+(-0.365161907541)*x[2]**o+(0.0218977623299)*x[2]
        arg[2,3,2,1]=(-0.775068843872)*x[0]**o+(-0.595291330879)*x[0]+(0.681122481841)*x[1]**o+(0.152514252732)*x[1]+(0.510976191808)*x[2]**o+(-0.69539067488)*x[2]
        arg[2,4,0,0]=(-0.362761749405)*x[0]**o+(0.494347958562)*x[0]+(0.215244694522)*x[1]**o+(-0.527767969217)*x[1]+(-0.115976071059)*x[2]**o+(-0.998233415841)*x[2]
        arg[2,4,0,1]=(0.620291015098)*x[0]**o+(-0.222644358028)*x[0]+(0.0377714785847)*x[1]**o+(0.663978623913)*x[1]+(0.996481523389)*x[2]**o+(-0.689519418095)*x[2]
        arg[2,4,1,0]=(-0.0374746407394)*x[0]**o+(0.3186355021)*x[0]+(-0.704305640533)*x[1]**o+(-0.632264150766)*x[1]+(-0.166691199609)*x[2]**o+(-0.794128441334)*x[2]
        arg[2,4,1,1]=(0.117806334919)*x[0]**o+(-0.717334572919)*x[0]+(0.665341054571)*x[1]**o+(0.354208273372)*x[1]+(0.349883631586)*x[2]**o+(-0.260626035314)*x[2]
        arg[2,4,2,0]=(0.160337827685)*x[0]**o+(-0.92734430881)*x[0]+(0.872112181363)*x[1]**o+(-0.796372970466)*x[1]+(-0.776882702685)*x[2]**o+(0.246816968474)*x[2]
        arg[2,4,2,1]=(-0.317926812571)*x[0]**o+(-0.795779593996)*x[0]+(0.785287724793)*x[1]**o+(0.839033515634)*x[1]+(0.433187943714)*x[2]**o+(0.308175885606)*x[2]
        arg[3,0,0,0]=(-0.555193292291)*x[0]**o+(-0.0280452948283)*x[0]+(0.988455127227)*x[1]**o+(-0.637707060271)*x[1]+(-0.95062571916)*x[2]**o+(-0.563121211553)*x[2]
        arg[3,0,0,1]=(-0.169558534946)*x[0]**o+(0.886096637835)*x[0]+(-0.666307491933)*x[1]**o+(0.650321801985)*x[1]+(-0.268934932123)*x[2]**o+(-0.406617199906)*x[2]
        arg[3,0,1,0]=(0.497351781449)*x[0]**o+(0.768617651685)*x[0]+(0.200962956226)*x[1]**o+(0.694236078954)*x[1]+(0.0323702864321)*x[2]**o+(-0.666046265832)*x[2]
        arg[3,0,1,1]=(0.425151807387)*x[0]**o+(-0.886432438604)*x[0]+(-0.963257203569)*x[1]**o+(-0.698524427437)*x[1]+(0.0165810713355)*x[2]**o+(-0.142193988951)*x[2]
        arg[3,0,2,0]=(0.965746112788)*x[0]**o+(-0.333068915886)*x[0]+(0.331031372002)*x[1]**o+(-0.384004796757)*x[1]+(0.0144121092048)*x[2]**o+(-0.352607264387)*x[2]
        arg[3,0,2,1]=(-0.391557775789)*x[0]**o+(0.127020330586)*x[0]+(0.758832450081)*x[1]**o+(0.49669414469)*x[1]+(0.358403498634)*x[2]**o+(0.184963946351)*x[2]
        arg[3,1,0,0]=(0.978053897464)*x[0]**o+(-0.211787313494)*x[0]+(0.927706482976)*x[1]**o+(-0.451292956408)*x[1]+(0.592824645174)*x[2]**o+(0.896240907546)*x[2]
        arg[3,1,0,1]=(0.0780568875295)*x[0]**o+(0.130098698523)*x[0]+(-0.281871571918)*x[1]**o+(-0.596362849045)*x[1]+(-0.436455789199)*x[2]**o+(-0.657145053021)*x[2]
        arg[3,1,1,0]=(-0.278111817329)*x[0]**o+(-0.627384158565)*x[0]+(-0.995802044048)*x[1]**o+(0.42485189159)*x[1]+(0.483473275475)*x[2]**o+(-0.177928974458)*x[2]
        arg[3,1,1,1]=(0.127721047557)*x[0]**o+(0.93147711978)*x[0]+(-0.350992052283)*x[1]**o+(0.0993773637672)*x[1]+(0.29111744653)*x[2]**o+(-0.821710278883)*x[2]
        arg[3,1,2,0]=(-0.697941918276)*x[0]**o+(-0.585872232852)*x[0]+(0.697967779925)*x[1]**o+(-0.811186381728)*x[1]+(0.708726108998)*x[2]**o+(0.289602027081)*x[2]
        arg[3,1,2,1]=(-0.461424409147)*x[0]**o+(0.335089386385)*x[0]+(-0.964838280406)*x[1]**o+(0.828627561583)*x[1]+(-0.861416281013)*x[2]**o+(0.898354675269)*x[2]
        arg[3,2,0,0]=(0.752543705358)*x[0]**o+(0.164487088496)*x[0]+(0.756164002406)*x[1]**o+(-0.740019514032)*x[1]+(0.107245489678)*x[2]**o+(0.677707700079)*x[2]
        arg[3,2,0,1]=(0.763792557134)*x[0]**o+(0.725507550672)*x[0]+(0.171323949629)*x[1]**o+(-0.287172075258)*x[1]+(-0.94865226715)*x[2]**o+(-0.163093465336)*x[2]
        arg[3,2,1,0]=(-0.327892721048)*x[0]**o+(0.334145897353)*x[0]+(-0.936827293436)*x[1]**o+(0.0822687872593)*x[1]+(0.259794829604)*x[2]**o+(0.376908280759)*x[2]
        arg[3,2,1,1]=(0.595012545955)*x[0]**o+(0.617490014693)*x[0]+(-0.658036388462)*x[1]**o+(-0.921055393527)*x[1]+(-0.257878085344)*x[2]**o+(-0.623113089161)*x[2]
        arg[3,2,2,0]=(0.961175705937)*x[0]**o+(0.656031660842)*x[0]+(0.964101580317)*x[1]**o+(0.824033157042)*x[1]+(0.510244886658)*x[2]**o+(-0.236501162729)*x[2]
        arg[3,2,2,1]=(-0.128405588923)*x[0]**o+(-0.0327141502894)*x[0]+(0.909941485131)*x[1]**o+(0.79947258121)*x[1]+(0.635883050534)*x[2]**o+(0.841997575796)*x[2]
        arg[3,3,0,0]=(-0.906234907643)*x[0]**o+(-0.316076976009)*x[0]+(0.863208664533)*x[1]**o+(-0.968640170523)*x[1]+(0.791637381686)*x[2]**o+(-0.543353462098)*x[2]
        arg[3,3,0,1]=(-0.586709036832)*x[0]**o+(-0.299493216592)*x[0]+(-0.173966725899)*x[1]**o+(0.0971838711677)*x[1]+(0.385662922244)*x[2]**o+(0.288932625368)*x[2]
        arg[3,3,1,0]=(-0.33865322795)*x[0]**o+(0.442176445777)*x[0]+(0.706183231282)*x[1]**o+(0.62528018966)*x[1]+(-0.377746006979)*x[2]**o+(-0.261455124139)*x[2]
        arg[3,3,1,1]=(0.834855553776)*x[0]**o+(0.755005824764)*x[0]+(0.446589332183)*x[1]**o+(0.41883521757)*x[1]+(0.820177207209)*x[2]**o+(0.603811552083)*x[2]
        arg[3,3,2,0]=(-0.624282266193)*x[0]**o+(0.28109086861)*x[0]+(0.44286011655)*x[1]**o+(-0.117876187343)*x[1]+(-0.267660131821)*x[2]**o+(0.880359406159)*x[2]
        arg[3,3,2,1]=(-0.0722687033317)*x[0]**o+(-0.7013147214)*x[0]+(0.00614303249005)*x[1]**o+(-0.658138946692)*x[1]+(0.0557277088337)*x[2]**o+(0.614362766103)*x[2]
        arg[3,4,0,0]=(0.935244440586)*x[0]**o+(-0.0699305155896)*x[0]+(-0.849303362918)*x[1]**o+(0.728062537708)*x[1]+(0.404621498028)*x[2]**o+(0.611486631437)*x[2]
        arg[3,4,0,1]=(0.168117782017)*x[0]**o+(-0.0454396423181)*x[0]+(0.46021270423)*x[1]**o+(0.597343569492)*x[1]+(0.561215321486)*x[2]**o+(0.715771227777)*x[2]
        arg[3,4,1,0]=(-0.502675682461)*x[0]**o+(0.543910335185)*x[0]+(0.721732643536)*x[1]**o+(-0.862126899566)*x[1]+(0.794704104653)*x[2]**o+(-0.284248276797)*x[2]
        arg[3,4,1,1]=(-0.580751817984)*x[0]**o+(0.482838505281)*x[0]+(-0.426066109367)*x[1]**o+(-0.456724382007)*x[1]+(-0.770655426938)*x[2]**o+(-0.226232224569)*x[2]
        arg[3,4,2,0]=(0.199832861603)*x[0]**o+(-0.586239366436)*x[0]+(0.730019631541)*x[1]**o+(0.185265476259)*x[1]+(0.514243351617)*x[2]**o+(-0.757742857839)*x[2]
        arg[3,4,2,1]=(-0.941843068195)*x[0]**o+(0.211732407161)*x[0]+(-0.99705602485)*x[1]**o+(-0.198037742645)*x[1]+(0.154058794178)*x[2]**o+(0.0675962671791)*x[2]
        ref[0,0,0,0]=(0.501470843513)/(o+1.)+(0.0235415380528)
        ref[0,0,0,1]=(-0.495882240293)/(o+1.)+(-0.628005728776)
        ref[0,0,1,0]=(-0.47355259574)/(o+1.)+(0.589958427511)
        ref[0,0,1,1]=(0.241804414345)/(o+1.)+(0.0744968350706)
        ref[0,0,2,0]=(-1.36573294299)/(o+1.)+(0.0250854306796)
        ref[0,0,2,1]=(-1.09689492365)/(o+1.)+(0.203220046558)
        ref[0,1,0,0]=(0.715822269113)/(o+1.)+(-0.513006150237)
        ref[0,1,0,1]=(0.194018305088)/(o+1.)+(0.594196942613)
        ref[0,1,1,0]=(-0.361460970224)/(o+1.)+(0.331424823297)
        ref[0,1,1,1]=(1.07883261716)/(o+1.)+(-0.835316371206)
        ref[0,1,2,0]=(0.734007979175)/(o+1.)+(0.372174974585)
        ref[0,1,2,1]=(-2.64519565923)/(o+1.)+(0.402886980703)
        ref[0,2,0,0]=(0.550037953057)/(o+1.)+(0.804639715144)
        ref[0,2,0,1]=(0.164423071623)/(o+1.)+(0.347820263772)
        ref[0,2,1,0]=(-1.0339269614)/(o+1.)+(-0.46790367599)
        ref[0,2,1,1]=(-0.663398714515)/(o+1.)+(0.341263346566)
        ref[0,2,2,0]=(0.484800065694)/(o+1.)+(-0.343834368812)
        ref[0,2,2,1]=(-0.647576357439)/(o+1.)+(-0.389767349638)
        ref[0,3,0,0]=(0.713628175789)/(o+1.)+(0.534426911657)
        ref[0,3,0,1]=(-1.0186229823)/(o+1.)+(0.168320055464)
        ref[0,3,1,0]=(-0.312627075447)/(o+1.)+(-0.288688061643)
        ref[0,3,1,1]=(0.563110793595)/(o+1.)+(-0.338854681006)
        ref[0,3,2,0]=(-1.6674641169)/(o+1.)+(-0.0630880744491)
        ref[0,3,2,1]=(-0.902373819515)/(o+1.)+(0.146944381666)
        ref[0,4,0,0]=(0.430045772773)/(o+1.)+(-0.326814746492)
        ref[0,4,0,1]=(-0.0329311494305)/(o+1.)+(-0.462763531298)
        ref[0,4,1,0]=(0.600404125575)/(o+1.)+(-0.666588219743)
        ref[0,4,1,1]=(-0.877044936593)/(o+1.)+(-0.213376554573)
        ref[0,4,2,0]=(-0.279718085349)/(o+1.)+(-0.507828107848)
        ref[0,4,2,1]=(1.55865221054)/(o+1.)+(-0.107165734349)
        ref[1,0,0,0]=(-1.53994251168)/(o+1.)+(0.682708891237)
        ref[1,0,0,1]=(0.27426508258)/(o+1.)+(0.311032674718)
        ref[1,0,1,0]=(-1.48856093105)/(o+1.)+(0.244603632945)
        ref[1,0,1,1]=(-0.980725422876)/(o+1.)+(0.0551781371305)
        ref[1,0,2,0]=(-0.123197946419)/(o+1.)+(-0.619047917567)
        ref[1,0,2,1]=(-0.128013307318)/(o+1.)+(-0.288543617977)
        ref[1,1,0,0]=(0.505758208708)/(o+1.)+(-0.604287011596)
        ref[1,1,0,1]=(-0.917894547188)/(o+1.)+(0.215510169161)
        ref[1,1,1,0]=(1.59629389187)/(o+1.)+(0.119123594986)
        ref[1,1,1,1]=(-1.63911850975)/(o+1.)+(-0.15110772397)
        ref[1,1,2,0]=(-0.357420174452)/(o+1.)+(0.946948435557)
        ref[1,1,2,1]=(-0.990173091615)/(o+1.)+(-0.412661055782)
        ref[1,2,0,0]=(-0.956530562442)/(o+1.)+(0.503239856545)
        ref[1,2,0,1]=(-0.0192458507848)/(o+1.)+(0.151130406306)
        ref[1,2,1,0]=(-0.510032040848)/(o+1.)+(1.12420678939)
        ref[1,2,1,1]=(-1.01025981956)/(o+1.)+(-0.483192483145)
        ref[1,2,2,0]=(2.46576915147)/(o+1.)+(-0.0480159570431)
        ref[1,2,2,1]=(-0.868236542197)/(o+1.)+(0.0610754286198)
        ref[1,3,0,0]=(0.873075607662)/(o+1.)+(-0.320365438154)
        ref[1,3,0,1]=(-0.430771579922)/(o+1.)+(0.444216184672)
        ref[1,3,1,0]=(-0.387878423294)/(o+1.)+(0.13604693133)
        ref[1,3,1,1]=(-1.88108112782)/(o+1.)+(-0.194800096716)
        ref[1,3,2,0]=(2.12050210152)/(o+1.)+(-0.128759395251)
        ref[1,3,2,1]=(0.325712098372)/(o+1.)+(0.104952955276)
        ref[1,4,0,0]=(0.924146253016)/(o+1.)+(-0.434307664887)
        ref[1,4,0,1]=(-0.221754040568)/(o+1.)+(1.14558078503)
        ref[1,4,1,0]=(0.0145217628367)/(o+1.)+(-0.43084890137)
        ref[1,4,1,1]=(-1.31753215008)/(o+1.)+(-0.824287026036)
        ref[1,4,2,0]=(-0.697722948118)/(o+1.)+(0.114574044827)
        ref[1,4,2,1]=(0.779339818147)/(o+1.)+(-0.365592981116)
        ref[2,0,0,0]=(0.274952897958)/(o+1.)+(-0.255370912722)
        ref[2,0,0,1]=(-1.10264655401)/(o+1.)+(-0.261708001871)
        ref[2,0,1,0]=(-0.680012578665)/(o+1.)+(-0.106715000464)
        ref[2,0,1,1]=(-0.686793121159)/(o+1.)+(-0.0241878370688)
        ref[2,0,2,0]=(0.942715771856)/(o+1.)+(-0.323717638282)
        ref[2,0,2,1]=(-0.0104862575713)/(o+1.)+(-0.19048054226)
        ref[2,1,0,0]=(0.746926399818)/(o+1.)+(0.117160499248)
        ref[2,1,0,1]=(-1.04171529373)/(o+1.)+(0.0976704487401)
        ref[2,1,1,0]=(0.136795029052)/(o+1.)+(-0.0802084471266)
        ref[2,1,1,1]=(0.896020560833)/(o+1.)+(-0.836960327706)
        ref[2,1,2,0]=(-0.00904194364113)/(o+1.)+(0.245943197498)
        ref[2,1,2,1]=(1.1485457441)/(o+1.)+(0.145957400558)
        ref[2,2,0,0]=(-0.0707799881243)/(o+1.)+(0.0690663884416)
        ref[2,2,0,1]=(1.32117006716)/(o+1.)+(0.227104957348)
        ref[2,2,1,0]=(0.711880376874)/(o+1.)+(-0.58130282817)
        ref[2,2,1,1]=(-0.611981014504)/(o+1.)+(0.303368043702)
        ref[2,2,2,0]=(0.564606893452)/(o+1.)+(-0.674927313599)
        ref[2,2,2,1]=(0.754537066511)/(o+1.)+(0.654108380069)
        ref[2,3,0,0]=(0.168469433796)/(o+1.)+(0.331189805903)
        ref[2,3,0,1]=(1.58905823122)/(o+1.)+(0.262836520673)
        ref[2,3,1,0]=(-0.556033290604)/(o+1.)+(-0.337520298949)
        ref[2,3,1,1]=(-1.23431436368)/(o+1.)+(-0.439901936041)
        ref[2,3,2,0]=(0.579845244941)/(o+1.)+(-0.0789272197338)
        ref[2,3,2,1]=(0.417029829778)/(o+1.)+(-0.569083876513)
        ref[2,4,0,0]=(-0.263493125943)/(o+1.)+(-0.515826713248)
        ref[2,4,0,1]=(1.65454401707)/(o+1.)+(-0.124092576105)
        ref[2,4,1,0]=(-0.908471480882)/(o+1.)+(-0.553878545)
        ref[2,4,1,1]=(1.13303102108)/(o+1.)+(-0.31187616743)
        ref[2,4,2,0]=(0.255567306363)/(o+1.)+(-0.738450155401)
        ref[2,4,2,1]=(0.900548855936)/(o+1.)+(0.175714903622)
        ref[3,0,0,0]=(-0.517363884224)/(o+1.)+(-0.614436783326)
        ref[3,0,0,1]=(-1.104800959)/(o+1.)+(0.564900619957)
        ref[3,0,1,0]=(0.730685024108)/(o+1.)+(0.398403732404)
        ref[3,0,1,1]=(-0.521524324846)/(o+1.)+(-0.863575427496)
        ref[3,0,2,0]=(1.31118959399)/(o+1.)+(-0.534840488515)
        ref[3,0,2,1]=(0.725678172925)/(o+1.)+(0.404339210814)
        ref[3,1,0,0]=(2.49858502561)/(o+1.)+(0.116580318822)
        ref[3,1,0,1]=(-0.640270473588)/(o+1.)+(-0.561704601771)
        ref[3,1,1,0]=(-0.790440585902)/(o+1.)+(-0.190230620716)
        ref[3,1,1,1]=(0.0678464418036)/(o+1.)+(0.104572102332)
        ref[3,1,2,0]=(0.708751970646)/(o+1.)+(-0.553728293749)
        ref[3,1,2,1]=(-2.28767897057)/(o+1.)+(1.03103581162)
        ref[3,2,0,0]=(1.61595319744)/(o+1.)+(0.0510876372718)
        ref[3,2,0,1]=(-0.0135357603866)/(o+1.)+(0.137621005039)
        ref[3,2,1,0]=(-1.00492518488)/(o+1.)+(0.396661482686)
        ref[3,2,1,1]=(-0.320901927851)/(o+1.)+(-0.463339233997)
        ref[3,2,2,0]=(2.43552217291)/(o+1.)+(0.621781827577)
        ref[3,2,2,1]=(1.41741894674)/(o+1.)+(0.804378003359)
        ref[3,3,0,0]=(0.748611138576)/(o+1.)+(-0.914035304315)
        ref[3,3,0,1]=(-0.375012840487)/(o+1.)+(0.0433116399723)
        ref[3,3,1,0]=(-0.0102160036467)/(o+1.)+(0.403000755649)
        ref[3,3,1,1]=(2.10162209317)/(o+1.)+(0.888826297208)
        ref[3,3,2,0]=(-0.449082281464)/(o+1.)+(0.521787043713)
        ref[3,3,2,1]=(-0.0103979620079)/(o+1.)+(-0.372545450994)
        ref[3,4,0,0]=(0.490562575696)/(o+1.)+(0.634809326777)
        ref[3,4,0,1]=(1.18954580773)/(o+1.)+(0.633837577475)
        ref[3,4,1,0]=(1.01376106573)/(o+1.)+(-0.301232420589)
        ref[3,4,1,1]=(-1.77747335429)/(o+1.)+(-0.100059050648)
        ref[3,4,2,0]=(1.44409584476)/(o+1.)+(-0.579358374008)
        ref[3,4,2,1]=(-1.78484029887)/(o+1.)+(0.0406454658476)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.788336609161)*x[0]**o+(-0.46539601764)*x[0]+(0.0970136798824)*x[1]**o+(-0.120665298671)*x[1]
        ref=(-0.691322929278)/(o+1.)+(-0.293030658156)
      else:
        arg=(0.922255261003)*x[0]**o+(0.0862232364707)*x[0]+(-0.94152635039)*x[1]**o+(0.771078486789)*x[1]+(-0.487642191181)*x[2]**o+(0.971754960362)*x[2]
        ref=(-0.506913280568)/(o+1.)+(0.914528341811)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.193071015033)*x[0]**o+(-0.281957941612)*x[0]+(0.269334552863)*x[1]**o+(0.0332262192804)*x[1]
        arg[1]=(-0.982634724425)*x[0]**o+(0.872638820737)*x[0]+(0.828028964603)*x[1]**o+(0.70464418136)*x[1]
        ref[0]=(0.0762635378296)/(o+1.)+(-0.124365861166)
        ref[1]=(-0.154605759822)/(o+1.)+(0.788641501048)
      else:
        arg[0]=(0.0746983082972)*x[0]**o+(0.281818436343)*x[0]+(-0.835321029206)*x[1]**o+(-0.349524270524)*x[1]+(-0.471905088601)*x[2]**o+(0.156327167402)*x[2]
        arg[1]=(-0.971935771255)*x[0]**o+(0.943151542057)*x[0]+(-0.60084258169)*x[1]**o+(0.458392855527)*x[1]+(-0.0173384187065)*x[2]**o+(0.538410338758)*x[2]
        ref[0]=(-1.23252780951)/(o+1.)+(0.0443106666103)
        ref[1]=(-1.59011677165)/(o+1.)+(0.969977368171)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.568783264759)*x[0]**o+(0.690377280868)*x[0]+(-0.689750922806)*x[1]**o+(-0.262828324126)*x[1]
        arg[0,1]=(-0.981657284673)*x[0]**o+(0.37118874852)*x[0]+(0.733992104146)*x[1]**o+(0.350268735699)*x[1]
        arg[0,2]=(0.713336100765)*x[0]**o+(0.737674447814)*x[0]+(0.239606043833)*x[1]**o+(0.814911366356)*x[1]
        arg[0,3]=(0.171706130967)*x[0]**o+(0.56324398491)*x[0]+(0.328884588679)*x[1]**o+(0.13307521808)*x[1]
        arg[0,4]=(-0.309130885049)*x[0]**o+(0.353596646674)*x[0]+(-0.282764686556)*x[1]**o+(-0.56977508596)*x[1]
        arg[1,0]=(0.974136161109)*x[0]**o+(-0.432477120938)*x[0]+(-0.841212164751)*x[1]**o+(-0.429211937256)*x[1]
        arg[1,1]=(0.0303890146585)*x[0]**o+(-0.712865964696)*x[0]+(0.572471431129)*x[1]**o+(-0.194324510922)*x[1]
        arg[1,2]=(-0.757134870981)*x[0]**o+(-0.018314520594)*x[0]+(0.250423290169)*x[1]**o+(-0.876678823729)*x[1]
        arg[1,3]=(-0.351885929068)*x[0]**o+(-0.561258605321)*x[0]+(0.208995575986)*x[1]**o+(0.428437856056)*x[1]
        arg[1,4]=(-0.313383912526)*x[0]**o+(-0.713484404328)*x[0]+(-0.279619274966)*x[1]**o+(0.72420661494)*x[1]
        arg[2,0]=(-0.686382641827)*x[0]**o+(0.678578408675)*x[0]+(-0.696532353786)*x[1]**o+(-0.154644995634)*x[1]
        arg[2,1]=(0.368548759951)*x[0]**o+(-0.674856247769)*x[0]+(0.908734263899)*x[1]**o+(0.331173572353)*x[1]
        arg[2,2]=(-0.1279136556)*x[0]**o+(-0.161474840876)*x[0]+(-0.775009015881)*x[1]**o+(-0.689552100815)*x[1]
        arg[2,3]=(-0.638450048485)*x[0]**o+(0.755485009392)*x[0]+(-0.793363159431)*x[1]**o+(0.255883452369)*x[1]
        arg[2,4]=(-0.831891512113)*x[0]**o+(-0.811678751836)*x[0]+(-0.487866621558)*x[1]**o+(0.240951890349)*x[1]
        arg[3,0]=(0.448249735424)*x[0]**o+(0.0809467993259)*x[0]+(0.232269153692)*x[1]**o+(0.477512628661)*x[1]
        arg[3,1]=(0.494093407104)*x[0]**o+(0.196761586998)*x[0]+(-0.702155296644)*x[1]**o+(-0.748641693731)*x[1]
        arg[3,2]=(0.432370067164)*x[0]**o+(-0.365671005874)*x[0]+(0.452009309356)*x[1]**o+(-0.951328172558)*x[1]
        arg[3,3]=(-0.993290240028)*x[0]**o+(0.172534714933)*x[0]+(0.00917693958413)*x[1]**o+(0.417515657393)*x[1]
        arg[3,4]=(0.541299019375)*x[0]**o+(0.479118172057)*x[0]+(-0.502475287154)*x[1]**o+(-0.04397899716)*x[1]
        ref[0,0]=(-0.120967658047)/(o+1.)+(0.213774478371)
        ref[0,1]=(-0.247665180527)/(o+1.)+(0.360728742109)
        ref[0,2]=(0.952942144598)/(o+1.)+(0.776292907085)
        ref[0,3]=(0.500590719646)/(o+1.)+(0.348159601495)
        ref[0,4]=(-0.591895571605)/(o+1.)+(-0.108089219643)
        ref[1,0]=(0.132923996358)/(o+1.)+(-0.430844529097)
        ref[1,1]=(0.602860445788)/(o+1.)+(-0.453595237809)
        ref[1,2]=(-0.506711580812)/(o+1.)+(-0.447496672161)
        ref[1,3]=(-0.142890353081)/(o+1.)+(-0.066410374632)
        ref[1,4]=(-0.593003187492)/(o+1.)+(0.00536110530621)
        ref[2,0]=(-1.38291499561)/(o+1.)+(0.261966706521)
        ref[2,1]=(1.27728302385)/(o+1.)+(-0.171841337708)
        ref[2,2]=(-0.902922671481)/(o+1.)+(-0.425513470846)
        ref[2,3]=(-1.43181320792)/(o+1.)+(0.505684230881)
        ref[2,4]=(-1.31975813367)/(o+1.)+(-0.285363430743)
        ref[3,0]=(0.680518889117)/(o+1.)+(0.279229713993)
        ref[3,1]=(-0.20806188954)/(o+1.)+(-0.275940053366)
        ref[3,2]=(0.88437937652)/(o+1.)+(-0.658499589216)
        ref[3,3]=(-0.984113300443)/(o+1.)+(0.295025186163)
        ref[3,4]=(0.0388237322207)/(o+1.)+(0.217569587448)
      else:
        arg[0,0]=(-0.866239193113)*x[0]**o+(0.479559390215)*x[0]+(0.512920648125)*x[1]**o+(-0.504775214511)*x[1]+(0.462210441279)*x[2]**o+(0.000249131516168)*x[2]
        arg[0,1]=(-0.11397206495)*x[0]**o+(0.0762933475692)*x[0]+(0.347716642793)*x[1]**o+(0.373262578958)*x[1]+(0.965468442129)*x[2]**o+(-0.743205128452)*x[2]
        arg[0,2]=(-0.214519408107)*x[0]**o+(-0.663675446512)*x[0]+(-0.424102625804)*x[1]**o+(-0.426644021856)*x[1]+(-0.0731662128704)*x[2]**o+(-0.375185307096)*x[2]
        arg[0,3]=(0.599595484989)*x[0]**o+(-0.303187098474)*x[0]+(0.278524134467)*x[1]**o+(-0.579432425528)*x[1]+(0.006698672684)*x[2]**o+(-0.820721120517)*x[2]
        arg[0,4]=(0.538847126305)*x[0]**o+(0.58093792853)*x[0]+(-0.21189097773)*x[1]**o+(0.637101424777)*x[1]+(0.0155994636183)*x[2]**o+(0.488675370945)*x[2]
        arg[1,0]=(-0.775581481064)*x[0]**o+(-0.968381856562)*x[0]+(-0.660755920498)*x[1]**o+(0.746481830633)*x[1]+(-0.300061615221)*x[2]**o+(-0.720031101537)*x[2]
        arg[1,1]=(0.990001351802)*x[0]**o+(-0.168997300944)*x[0]+(0.0134555819367)*x[1]**o+(0.969355591009)*x[1]+(-0.347553438064)*x[2]**o+(0.648537935984)*x[2]
        arg[1,2]=(-0.327760535516)*x[0]**o+(0.517489199537)*x[0]+(-0.0589046196438)*x[1]**o+(-0.893401439507)*x[1]+(-0.0821634552571)*x[2]**o+(0.861266440343)*x[2]
        arg[1,3]=(-0.808782893446)*x[0]**o+(-0.196363777506)*x[0]+(0.623837845847)*x[1]**o+(-0.05392424847)*x[1]+(0.616011657554)*x[2]**o+(0.800682023434)*x[2]
        arg[1,4]=(-0.425073222968)*x[0]**o+(-0.635248695204)*x[0]+(0.947222621311)*x[1]**o+(0.253198813179)*x[1]+(-0.907568459712)*x[2]**o+(-0.513049744062)*x[2]
        arg[2,0]=(-0.103867107528)*x[0]**o+(0.893031214224)*x[0]+(0.151292631908)*x[1]**o+(-0.407189878657)*x[1]+(0.618120612304)*x[2]**o+(0.983992964931)*x[2]
        arg[2,1]=(0.310432646627)*x[0]**o+(0.235143842057)*x[0]+(0.729683627093)*x[1]**o+(0.891778928523)*x[1]+(-0.148927367367)*x[2]**o+(0.913403960844)*x[2]
        arg[2,2]=(0.610293692912)*x[0]**o+(-0.28497731539)*x[0]+(-0.926462472232)*x[1]**o+(-0.427657211191)*x[1]+(0.467905160751)*x[2]**o+(0.195578740767)*x[2]
        arg[2,3]=(0.854869452114)*x[0]**o+(-0.752530632831)*x[0]+(0.0128220565485)*x[1]**o+(-0.726013651836)*x[1]+(-0.0979888078519)*x[2]**o+(-0.658955703807)*x[2]
        arg[2,4]=(0.536447848289)*x[0]**o+(-0.658205406579)*x[0]+(0.66427484293)*x[1]**o+(-0.250014589496)*x[1]+(-0.0654257159073)*x[2]**o+(0.538196496913)*x[2]
        arg[3,0]=(-0.514665034974)*x[0]**o+(-0.278189747423)*x[0]+(0.402996329453)*x[1]**o+(-0.0362969716175)*x[1]+(-0.707890815004)*x[2]**o+(-0.298762865096)*x[2]
        arg[3,1]=(0.7342439327)*x[0]**o+(-0.387336714116)*x[0]+(-0.67888634151)*x[1]**o+(0.647322640846)*x[1]+(-0.0346274738413)*x[2]**o+(0.985950301913)*x[2]
        arg[3,2]=(0.687555064124)*x[0]**o+(0.345506248144)*x[0]+(0.658692258064)*x[1]**o+(0.810796031447)*x[1]+(0.102608006398)*x[2]**o+(0.349877767777)*x[2]
        arg[3,3]=(-0.456241806837)*x[0]**o+(-0.181224535865)*x[0]+(0.401842456169)*x[1]**o+(-0.413936620519)*x[1]+(0.436794604974)*x[2]**o+(0.95051042082)*x[2]
        arg[3,4]=(-0.330739393408)*x[0]**o+(-0.261876658969)*x[0]+(0.287474017623)*x[1]**o+(0.463691723118)*x[1]+(0.757597823206)*x[2]**o+(0.49336246207)*x[2]
        ref[0,0]=(0.108891896292)/(o+1.)+(-0.01248334639)
        ref[0,1]=(1.19921301997)/(o+1.)+(-0.146824600962)
        ref[0,2]=(-0.711788246781)/(o+1.)+(-0.732752387732)
        ref[0,3]=(0.88481829214)/(o+1.)+(-0.85167032226)
        ref[0,4]=(0.342555612194)/(o+1.)+(0.853357362126)
        ref[1,0]=(-1.73639901678)/(o+1.)+(-0.470965563733)
        ref[1,1]=(0.655903495675)/(o+1.)+(0.724448113024)
        ref[1,2]=(-0.468828610417)/(o+1.)+(0.242677100186)
        ref[1,3]=(0.431066609955)/(o+1.)+(0.275196998729)
        ref[1,4]=(-0.385419061368)/(o+1.)+(-0.447549813044)
        ref[2,0]=(0.665546136685)/(o+1.)+(0.734917150249)
        ref[2,1]=(0.891188906354)/(o+1.)+(1.02016336571)
        ref[2,2]=(0.151736381432)/(o+1.)+(-0.258527892907)
        ref[2,3]=(0.76970270081)/(o+1.)+(-1.06874999424)
        ref[2,4]=(1.13529697531)/(o+1.)+(-0.185011749581)
        ref[3,0]=(-0.819559520525)/(o+1.)+(-0.306624792068)
        ref[3,1]=(0.020730117348)/(o+1.)+(0.622968114321)
        ref[3,2]=(1.44885532859)/(o+1.)+(0.753090023684)
        ref[3,3]=(0.382395254306)/(o+1.)+(0.177674632218)
        ref[3,4]=(0.714332447421)/(o+1.)+(0.34758876311)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.779261125448)*x[0]**o+(-0.640503838419)*x[0]+(0.589085035572)*x[1]**o+(-0.0426687544445)*x[1]
        arg[0,0,1]=(-0.419361292473)*x[0]**o+(0.0850684591011)*x[0]+(-0.440758336688)*x[1]**o+(0.151015611594)*x[1]
        arg[0,1,0]=(-0.727043966709)*x[0]**o+(0.711730063365)*x[0]+(0.657759819512)*x[1]**o+(-0.210981156609)*x[1]
        arg[0,1,1]=(-0.276491768644)*x[0]**o+(-0.324865684188)*x[0]+(-0.921856877627)*x[1]**o+(-0.246833902926)*x[1]
        arg[1,0,0]=(0.720541615152)*x[0]**o+(-0.0406851820986)*x[0]+(0.33976001267)*x[1]**o+(0.937989079378)*x[1]
        arg[1,0,1]=(0.239272962875)*x[0]**o+(0.482026352856)*x[0]+(0.159148474482)*x[1]**o+(0.680831564145)*x[1]
        arg[1,1,0]=(0.55170300456)*x[0]**o+(-0.261357073238)*x[0]+(0.298185649596)*x[1]**o+(-0.540598468483)*x[1]
        arg[1,1,1]=(0.508262701362)*x[0]**o+(-0.531692643817)*x[0]+(0.151424571021)*x[1]**o+(-0.77488755481)*x[1]
        arg[2,0,0]=(0.265201389846)*x[0]**o+(0.377300848247)*x[0]+(0.0402148688087)*x[1]**o+(0.0263806695524)*x[1]
        arg[2,0,1]=(0.321397565411)*x[0]**o+(-0.277650913698)*x[0]+(0.909430556637)*x[1]**o+(-0.43244155765)*x[1]
        arg[2,1,0]=(-0.671417887851)*x[0]**o+(-0.772328065901)*x[0]+(0.0896878386442)*x[1]**o+(0.584773963567)*x[1]
        arg[2,1,1]=(0.318995094412)*x[0]**o+(-0.396578444213)*x[0]+(-0.627767362562)*x[1]**o+(-0.588717342984)*x[1]
        arg[3,0,0]=(-0.154999830633)*x[0]**o+(-0.263434767944)*x[0]+(0.0704862479101)*x[1]**o+(0.216237505495)*x[1]
        arg[3,0,1]=(0.956868643095)*x[0]**o+(-0.0331644426873)*x[0]+(-0.236340025853)*x[1]**o+(-0.814705425837)*x[1]
        arg[3,1,0]=(-0.477985064084)*x[0]**o+(-0.368896833289)*x[0]+(0.208122873207)*x[1]**o+(-0.566525593432)*x[1]
        arg[3,1,1]=(-0.299257275248)*x[0]**o+(0.937022356746)*x[0]+(0.671724793932)*x[1]**o+(0.80705820341)*x[1]
        arg[4,0,0]=(0.857855249735)*x[0]**o+(0.537756541143)*x[0]+(0.611784225987)*x[1]**o+(0.944061991186)*x[1]
        arg[4,0,1]=(0.825779290308)*x[0]**o+(-0.636461636739)*x[0]+(-0.382007450888)*x[1]**o+(0.170178262303)*x[1]
        arg[4,1,0]=(0.275666857657)*x[0]**o+(-0.404983003194)*x[0]+(0.476932065223)*x[1]**o+(0.931645726493)*x[1]
        arg[4,1,1]=(0.956182179652)*x[0]**o+(0.543482655321)*x[0]+(0.60492494312)*x[1]**o+(0.384497838341)*x[1]
        arg[5,0,0]=(0.476937368598)*x[0]**o+(-0.588434801718)*x[0]+(0.865080125217)*x[1]**o+(-0.954493375041)*x[1]
        arg[5,0,1]=(0.151850214007)*x[0]**o+(0.252651124764)*x[0]+(-0.570132867968)*x[1]**o+(-0.0885902498736)*x[1]
        arg[5,1,0]=(-0.69330471578)*x[0]**o+(0.165874518143)*x[0]+(-0.610503496045)*x[1]**o+(-0.871951964138)*x[1]
        arg[5,1,1]=(-0.539191264945)*x[0]**o+(0.558474736927)*x[0]+(0.752045569996)*x[1]**o+(-0.5305308467)*x[1]
        ref[0,0,0]=(1.36834616102)/(o+1.)+(-0.341586296432)
        ref[0,0,1]=(-0.860119629161)/(o+1.)+(0.118042035348)
        ref[0,1,0]=(-0.0692841471964)/(o+1.)+(0.250374453378)
        ref[0,1,1]=(-1.19834864627)/(o+1.)+(-0.285849793557)
        ref[1,0,0]=(1.06030162782)/(o+1.)+(0.448651948639)
        ref[1,0,1]=(0.398421437357)/(o+1.)+(0.5814289585)
        ref[1,1,0]=(0.849888654155)/(o+1.)+(-0.400977770861)
        ref[1,1,1]=(0.659687272383)/(o+1.)+(-0.653290099314)
        ref[2,0,0]=(0.305416258655)/(o+1.)+(0.2018407589)
        ref[2,0,1]=(1.23082812205)/(o+1.)+(-0.355046235674)
        ref[2,1,0]=(-0.581730049206)/(o+1.)+(-0.0937770511668)
        ref[2,1,1]=(-0.30877226815)/(o+1.)+(-0.492647893598)
        ref[3,0,0]=(-0.0845135827233)/(o+1.)+(-0.0235986312242)
        ref[3,0,1]=(0.720528617241)/(o+1.)+(-0.423934934262)
        ref[3,1,0]=(-0.269862190877)/(o+1.)+(-0.46771121336)
        ref[3,1,1]=(0.372467518684)/(o+1.)+(0.872040280078)
        ref[4,0,0]=(1.46963947572)/(o+1.)+(0.740909266165)
        ref[4,0,1]=(0.44377183942)/(o+1.)+(-0.233141687218)
        ref[4,1,0]=(0.75259892288)/(o+1.)+(0.26333136165)
        ref[4,1,1]=(1.56110712277)/(o+1.)+(0.463990246831)
        ref[5,0,0]=(1.34201749382)/(o+1.)+(-0.771464088379)
        ref[5,0,1]=(-0.418282653961)/(o+1.)+(0.0820304374451)
        ref[5,1,0]=(-1.30380821182)/(o+1.)+(-0.353038722997)
        ref[5,1,1]=(0.212854305051)/(o+1.)+(0.0139719451134)
      else:
        arg[0,0,0]=(-0.825926001774)*x[0]**o+(0.418049316138)*x[0]+(0.668695279849)*x[1]**o+(-0.7323605258)*x[1]+(-0.637717742224)*x[2]**o+(0.257065974143)*x[2]
        arg[0,0,1]=(-0.169779448443)*x[0]**o+(-0.648612931029)*x[0]+(-0.643814352995)*x[1]**o+(0.396704595574)*x[1]+(-0.248766540261)*x[2]**o+(-0.736410673282)*x[2]
        arg[0,1,0]=(-0.371929441192)*x[0]**o+(0.438661796459)*x[0]+(-0.893450426264)*x[1]**o+(0.643494077636)*x[1]+(0.252099102122)*x[2]**o+(-0.0857798075076)*x[2]
        arg[0,1,1]=(0.336251711034)*x[0]**o+(-0.317426111079)*x[0]+(-0.801221986791)*x[1]**o+(0.281123033042)*x[1]+(0.921153543163)*x[2]**o+(0.497003835923)*x[2]
        arg[1,0,0]=(0.40897311872)*x[0]**o+(0.326894801733)*x[0]+(0.216268807497)*x[1]**o+(0.946235364974)*x[1]+(-0.480995485096)*x[2]**o+(0.869772937982)*x[2]
        arg[1,0,1]=(0.108715554449)*x[0]**o+(0.0847529231253)*x[0]+(0.866120965844)*x[1]**o+(0.43989772616)*x[1]+(0.67962061792)*x[2]**o+(0.90311988632)*x[2]
        arg[1,1,0]=(0.28571484173)*x[0]**o+(0.593050819853)*x[0]+(-0.384864010146)*x[1]**o+(-0.57920566259)*x[1]+(-0.304024223769)*x[2]**o+(-0.0279518764989)*x[2]
        arg[1,1,1]=(0.249126658892)*x[0]**o+(0.243355410215)*x[0]+(0.154816978135)*x[1]**o+(-0.260343533153)*x[1]+(-0.973133787852)*x[2]**o+(0.752545508531)*x[2]
        arg[2,0,0]=(0.103612497259)*x[0]**o+(0.0919234144108)*x[0]+(0.416394965635)*x[1]**o+(-0.0132359939268)*x[1]+(-0.522775855178)*x[2]**o+(-0.985811066906)*x[2]
        arg[2,0,1]=(-0.592641488144)*x[0]**o+(0.703703144485)*x[0]+(-0.0149046721584)*x[1]**o+(0.220368283581)*x[1]+(-0.656858164281)*x[2]**o+(0.165951102113)*x[2]
        arg[2,1,0]=(0.309492848012)*x[0]**o+(-0.301037623847)*x[0]+(-0.224814259669)*x[1]**o+(-0.950794406688)*x[1]+(-0.296088293363)*x[2]**o+(-0.712930293168)*x[2]
        arg[2,1,1]=(0.0592633594166)*x[0]**o+(-0.179007619512)*x[0]+(-0.922169690831)*x[1]**o+(-0.778228075089)*x[1]+(-0.135749453777)*x[2]**o+(-0.146265051618)*x[2]
        arg[3,0,0]=(-0.71773633584)*x[0]**o+(0.471551385738)*x[0]+(-0.041415831943)*x[1]**o+(0.343380640343)*x[1]+(-0.211821446187)*x[2]**o+(-0.661201893183)*x[2]
        arg[3,0,1]=(-0.159071045836)*x[0]**o+(0.216715127143)*x[0]+(-0.846847485213)*x[1]**o+(0.662226888693)*x[1]+(0.050806337433)*x[2]**o+(0.744071347051)*x[2]
        arg[3,1,0]=(-0.363431877105)*x[0]**o+(-0.461513470653)*x[0]+(-0.130947882739)*x[1]**o+(-0.624729599981)*x[1]+(0.778572567699)*x[2]**o+(0.324584316744)*x[2]
        arg[3,1,1]=(0.395432262346)*x[0]**o+(0.826450301981)*x[0]+(0.176721657975)*x[1]**o+(-0.370954952405)*x[1]+(-0.708437370609)*x[2]**o+(-0.150075114354)*x[2]
        arg[4,0,0]=(-0.18662583335)*x[0]**o+(-0.0120831234457)*x[0]+(0.805686139762)*x[1]**o+(0.30195413136)*x[1]+(0.648067854031)*x[2]**o+(0.28073240632)*x[2]
        arg[4,0,1]=(-0.913625644854)*x[0]**o+(-0.264637827098)*x[0]+(-0.6531590441)*x[1]**o+(0.396422722623)*x[1]+(0.280591271146)*x[2]**o+(0.498346191707)*x[2]
        arg[4,1,0]=(0.208029597907)*x[0]**o+(-0.295747909396)*x[0]+(-0.541509895186)*x[1]**o+(-0.511394562575)*x[1]+(-0.459197562982)*x[2]**o+(-0.749433614666)*x[2]
        arg[4,1,1]=(-0.29700028905)*x[0]**o+(-0.590932598278)*x[0]+(0.36219700213)*x[1]**o+(0.482759242709)*x[1]+(-0.135696240971)*x[2]**o+(0.49658807849)*x[2]
        arg[5,0,0]=(-0.800345624219)*x[0]**o+(0.259709746207)*x[0]+(0.480073493816)*x[1]**o+(0.499332550127)*x[1]+(0.691794186846)*x[2]**o+(-0.319369042249)*x[2]
        arg[5,0,1]=(0.300879052118)*x[0]**o+(-0.985396817441)*x[0]+(0.971136879002)*x[1]**o+(0.806560091745)*x[1]+(0.361308859099)*x[2]**o+(0.264277089732)*x[2]
        arg[5,1,0]=(0.271857906548)*x[0]**o+(0.338260411675)*x[0]+(-0.170292665803)*x[1]**o+(0.675491566142)*x[1]+(0.897921755421)*x[2]**o+(-0.829233864443)*x[2]
        arg[5,1,1]=(0.673366845241)*x[0]**o+(0.121940550422)*x[0]+(0.116054155545)*x[1]**o+(0.0456556938689)*x[1]+(-0.978127590673)*x[2]**o+(0.180985351995)*x[2]
        ref[0,0,0]=(-0.794948464148)/(o+1.)+(-0.0286226177598)
        ref[0,0,1]=(-1.0623603417)/(o+1.)+(-0.494159504368)
        ref[0,1,0]=(-1.01328076533)/(o+1.)+(0.498188033294)
        ref[0,1,1]=(0.456183267406)/(o+1.)+(0.230350378943)
        ref[1,0,0]=(0.144246441122)/(o+1.)+(1.07145155234)
        ref[1,0,1]=(1.65445713821)/(o+1.)+(0.713885267803)
        ref[1,1,0]=(-0.403173392185)/(o+1.)+(-0.0070533596178)
        ref[1,1,1]=(-0.569190150825)/(o+1.)+(0.367778692797)
        ref[2,0,0]=(-0.00276839228366)/(o+1.)+(-0.453561823211)
        ref[2,0,1]=(-1.26440432458)/(o+1.)+(0.545011265089)
        ref[2,1,0]=(-0.21140970502)/(o+1.)+(-0.982381161851)
        ref[2,1,1]=(-0.998655785191)/(o+1.)+(-0.55175037311)
        ref[3,0,0]=(-0.97097361397)/(o+1.)+(0.0768650664488)
        ref[3,0,1]=(-0.955112193616)/(o+1.)+(0.811506681444)
        ref[3,1,0]=(0.284192807855)/(o+1.)+(-0.380829376946)
        ref[3,1,1]=(-0.136283450288)/(o+1.)+(0.152710117611)
        ref[4,0,0]=(1.26712816044)/(o+1.)+(0.285301707117)
        ref[4,0,1]=(-1.28619341781)/(o+1.)+(0.315065543616)
        ref[4,1,0]=(-0.792677860261)/(o+1.)+(-0.778288043318)
        ref[4,1,1]=(-0.0704995278906)/(o+1.)+(0.194207361461)
        ref[5,0,0]=(0.371522056443)/(o+1.)+(0.219836627043)
        ref[5,0,1]=(1.63332479022)/(o+1.)+(0.0427201820182)
        ref[5,1,0]=(0.999486996166)/(o+1.)+(0.0922590566873)
        ref[5,1,1]=(-0.188706589887)/(o+1.)+(0.174290798143)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.647932589381)*x[0]**o+(-0.092686685918)*x[0]+(0.167888464541)*x[1]**o+(0.698445840627)*x[1]
        arg[0,0,0,1]=(0.413204959658)*x[0]**o+(0.772926029512)*x[0]+(-0.63702948026)*x[1]**o+(0.608392823861)*x[1]
        arg[0,0,1,0]=(0.303506932052)*x[0]**o+(0.650309664063)*x[0]+(-0.805238344485)*x[1]**o+(-0.62971680995)*x[1]
        arg[0,0,1,1]=(-0.205701190736)*x[0]**o+(-0.0834042840814)*x[0]+(-0.525839398431)*x[1]**o+(0.779973223005)*x[1]
        arg[0,0,2,0]=(-0.97954633631)*x[0]**o+(-0.298973871114)*x[0]+(0.858022652298)*x[1]**o+(0.667671796632)*x[1]
        arg[0,0,2,1]=(-0.790133712345)*x[0]**o+(-0.733743414808)*x[0]+(0.337185148145)*x[1]**o+(-0.927853051244)*x[1]
        arg[0,1,0,0]=(0.617198112458)*x[0]**o+(0.70438849416)*x[0]+(-0.861661596364)*x[1]**o+(-0.251796116219)*x[1]
        arg[0,1,0,1]=(0.37107453636)*x[0]**o+(-0.490440661153)*x[0]+(0.849252675242)*x[1]**o+(-0.896547775077)*x[1]
        arg[0,1,1,0]=(0.375080744007)*x[0]**o+(-0.609557663503)*x[0]+(-0.930846980977)*x[1]**o+(0.0623096843381)*x[1]
        arg[0,1,1,1]=(-0.809863615805)*x[0]**o+(-0.999724521803)*x[0]+(0.493027030633)*x[1]**o+(-0.672546253752)*x[1]
        arg[0,1,2,0]=(0.219294077412)*x[0]**o+(-0.7854819847)*x[0]+(-0.353026321798)*x[1]**o+(-0.965348418509)*x[1]
        arg[0,1,2,1]=(0.177236321722)*x[0]**o+(-0.860491592895)*x[0]+(-0.458724268899)*x[1]**o+(-0.476508229653)*x[1]
        arg[0,2,0,0]=(0.0760611767074)*x[0]**o+(-0.00296398964042)*x[0]+(-0.972843039436)*x[1]**o+(-0.0422057743881)*x[1]
        arg[0,2,0,1]=(-0.926737777701)*x[0]**o+(-0.895791593378)*x[0]+(-0.776529150964)*x[1]**o+(0.557115706978)*x[1]
        arg[0,2,1,0]=(0.353264702858)*x[0]**o+(0.632583933138)*x[0]+(-0.354112684469)*x[1]**o+(0.0541346174117)*x[1]
        arg[0,2,1,1]=(-0.124181989217)*x[0]**o+(-0.803262099858)*x[0]+(0.733246361984)*x[1]**o+(0.0639821744038)*x[1]
        arg[0,2,2,0]=(-0.446733845136)*x[0]**o+(-0.187373171381)*x[0]+(-0.312309043572)*x[1]**o+(0.830789380477)*x[1]
        arg[0,2,2,1]=(0.403889932995)*x[0]**o+(-0.243474625233)*x[0]+(0.35707779102)*x[1]**o+(0.29725421988)*x[1]
        arg[0,3,0,0]=(-0.856376546696)*x[0]**o+(0.448431504377)*x[0]+(-0.535187100108)*x[1]**o+(-0.90994242894)*x[1]
        arg[0,3,0,1]=(-0.321475754775)*x[0]**o+(0.0389358177737)*x[0]+(-0.0949733279735)*x[1]**o+(0.921937240884)*x[1]
        arg[0,3,1,0]=(0.591276336387)*x[0]**o+(-0.598236438303)*x[0]+(-0.829829995185)*x[1]**o+(0.539669346291)*x[1]
        arg[0,3,1,1]=(-0.944726366074)*x[0]**o+(-0.796928273271)*x[0]+(-0.424606656681)*x[1]**o+(0.35341306063)*x[1]
        arg[0,3,2,0]=(-0.850914599406)*x[0]**o+(0.791179202404)*x[0]+(-0.0220845565851)*x[1]**o+(0.594756136957)*x[1]
        arg[0,3,2,1]=(0.355862680641)*x[0]**o+(0.134261680432)*x[0]+(0.477798827075)*x[1]**o+(0.508452338499)*x[1]
        arg[0,4,0,0]=(0.481974944781)*x[0]**o+(0.259271175551)*x[0]+(0.953827692515)*x[1]**o+(0.364160709544)*x[1]
        arg[0,4,0,1]=(-0.316627315143)*x[0]**o+(-0.856616735533)*x[0]+(0.308508314695)*x[1]**o+(-0.677602758367)*x[1]
        arg[0,4,1,0]=(-0.76069824745)*x[0]**o+(0.282351651183)*x[0]+(-0.15964276552)*x[1]**o+(-0.344276428838)*x[1]
        arg[0,4,1,1]=(9.00669303738e-05)*x[0]**o+(0.948498888115)*x[0]+(0.690370267333)*x[1]**o+(0.619119188327)*x[1]
        arg[0,4,2,0]=(0.957692561039)*x[0]**o+(-0.147005436631)*x[0]+(-0.177840574215)*x[1]**o+(-0.28259452614)*x[1]
        arg[0,4,2,1]=(-0.431771943451)*x[0]**o+(-0.158132166651)*x[0]+(-0.134385316147)*x[1]**o+(0.751243590835)*x[1]
        arg[1,0,0,0]=(-0.816841420645)*x[0]**o+(0.55608373851)*x[0]+(-0.885838014201)*x[1]**o+(-0.467159163867)*x[1]
        arg[1,0,0,1]=(-0.543361117401)*x[0]**o+(0.329490019675)*x[0]+(0.942703992489)*x[1]**o+(-0.261273838803)*x[1]
        arg[1,0,1,0]=(0.865108106801)*x[0]**o+(-0.376972427024)*x[0]+(-0.0257304693919)*x[1]**o+(0.694210909726)*x[1]
        arg[1,0,1,1]=(0.428354498121)*x[0]**o+(0.962704601892)*x[0]+(-0.442636806139)*x[1]**o+(0.424724924149)*x[1]
        arg[1,0,2,0]=(-0.157277242299)*x[0]**o+(0.487588443251)*x[0]+(-0.813089096341)*x[1]**o+(0.0406316718829)*x[1]
        arg[1,0,2,1]=(0.345322702588)*x[0]**o+(0.224429042659)*x[0]+(-0.24109926586)*x[1]**o+(0.433141241931)*x[1]
        arg[1,1,0,0]=(0.956901263519)*x[0]**o+(0.48418062914)*x[0]+(0.14089030013)*x[1]**o+(-0.492352201717)*x[1]
        arg[1,1,0,1]=(-0.679781695411)*x[0]**o+(0.29857196636)*x[0]+(-0.234968984185)*x[1]**o+(0.114575919411)*x[1]
        arg[1,1,1,0]=(0.488378272934)*x[0]**o+(-0.358508327343)*x[0]+(-0.337958846849)*x[1]**o+(0.189303452912)*x[1]
        arg[1,1,1,1]=(-0.506993859574)*x[0]**o+(-0.682478021258)*x[0]+(-0.540344338899)*x[1]**o+(0.366965535976)*x[1]
        arg[1,1,2,0]=(-0.681151638572)*x[0]**o+(0.790837550009)*x[0]+(-0.810785716023)*x[1]**o+(-0.00739097962416)*x[1]
        arg[1,1,2,1]=(-0.355290573677)*x[0]**o+(-0.263294580108)*x[0]+(0.10469018402)*x[1]**o+(-0.650388283116)*x[1]
        arg[1,2,0,0]=(0.906212689905)*x[0]**o+(-0.673956103538)*x[0]+(0.305335278934)*x[1]**o+(-0.0805949628719)*x[1]
        arg[1,2,0,1]=(-0.98311373852)*x[0]**o+(-0.165316704542)*x[0]+(0.197637770603)*x[1]**o+(0.647989184342)*x[1]
        arg[1,2,1,0]=(0.415990736251)*x[0]**o+(-0.710476520822)*x[0]+(0.791512786296)*x[1]**o+(-0.10853188742)*x[1]
        arg[1,2,1,1]=(0.04337234954)*x[0]**o+(0.595650094449)*x[0]+(0.228098665648)*x[1]**o+(-0.868930899722)*x[1]
        arg[1,2,2,0]=(-0.791842482688)*x[0]**o+(-0.611246387621)*x[0]+(0.0504978789412)*x[1]**o+(0.624626300694)*x[1]
        arg[1,2,2,1]=(0.0207375595263)*x[0]**o+(0.560638072838)*x[0]+(0.44126312508)*x[1]**o+(0.169151353084)*x[1]
        arg[1,3,0,0]=(-0.0512069089279)*x[0]**o+(0.51110271913)*x[0]+(0.0785633458626)*x[1]**o+(0.836821453174)*x[1]
        arg[1,3,0,1]=(0.185737468851)*x[0]**o+(0.516071031227)*x[0]+(0.0995388802342)*x[1]**o+(0.25527909092)*x[1]
        arg[1,3,1,0]=(0.318240681435)*x[0]**o+(-0.442725217398)*x[0]+(0.533066616909)*x[1]**o+(0.679570707339)*x[1]
        arg[1,3,1,1]=(0.353367397644)*x[0]**o+(-0.396487189861)*x[0]+(-0.117773711001)*x[1]**o+(0.885897675293)*x[1]
        arg[1,3,2,0]=(-0.647645988694)*x[0]**o+(-0.755258662164)*x[0]+(0.288300257472)*x[1]**o+(0.505342385389)*x[1]
        arg[1,3,2,1]=(0.000781386924262)*x[0]**o+(0.190244787656)*x[0]+(0.589388834071)*x[1]**o+(0.669705413859)*x[1]
        arg[1,4,0,0]=(0.320107896841)*x[0]**o+(0.673508197912)*x[0]+(0.667026895874)*x[1]**o+(-0.818578971562)*x[1]
        arg[1,4,0,1]=(-0.226755915796)*x[0]**o+(0.71800207991)*x[0]+(0.850490212043)*x[1]**o+(0.261708964138)*x[1]
        arg[1,4,1,0]=(-0.0147350282361)*x[0]**o+(0.743889536542)*x[0]+(-0.898802067607)*x[1]**o+(0.711589170571)*x[1]
        arg[1,4,1,1]=(-0.816145493818)*x[0]**o+(0.405931568806)*x[0]+(0.159156576579)*x[1]**o+(0.143576116984)*x[1]
        arg[1,4,2,0]=(0.994791165013)*x[0]**o+(0.65054901122)*x[0]+(0.039218059733)*x[1]**o+(-0.662978657934)*x[1]
        arg[1,4,2,1]=(-0.792277825616)*x[0]**o+(-0.257224872111)*x[0]+(-0.244502392787)*x[1]**o+(-0.276042189495)*x[1]
        arg[2,0,0,0]=(0.133211578016)*x[0]**o+(0.10029141733)*x[0]+(-0.52480547182)*x[1]**o+(0.814652707336)*x[1]
        arg[2,0,0,1]=(0.836429340685)*x[0]**o+(-0.837709322969)*x[0]+(-0.442953015462)*x[1]**o+(-0.984757344044)*x[1]
        arg[2,0,1,0]=(0.893738853578)*x[0]**o+(-0.238050980408)*x[0]+(-0.913668527302)*x[1]**o+(0.895136777323)*x[1]
        arg[2,0,1,1]=(-0.0457361766611)*x[0]**o+(0.414814798208)*x[0]+(-0.57261759439)*x[1]**o+(-0.358969048076)*x[1]
        arg[2,0,2,0]=(-0.320394419991)*x[0]**o+(-0.990129952721)*x[0]+(0.890888895016)*x[1]**o+(0.241063301452)*x[1]
        arg[2,0,2,1]=(-0.604164679932)*x[0]**o+(-0.598939404114)*x[0]+(-0.606633388454)*x[1]**o+(-0.666879580873)*x[1]
        arg[2,1,0,0]=(0.68515547034)*x[0]**o+(0.892247437859)*x[0]+(0.642003383991)*x[1]**o+(0.569240970392)*x[1]
        arg[2,1,0,1]=(0.918713108587)*x[0]**o+(0.865547962284)*x[0]+(0.517952260633)*x[1]**o+(-0.307334010857)*x[1]
        arg[2,1,1,0]=(0.353382468404)*x[0]**o+(-0.542718692644)*x[0]+(-0.957182863451)*x[1]**o+(-0.889352632225)*x[1]
        arg[2,1,1,1]=(0.505185847817)*x[0]**o+(-0.71137143775)*x[0]+(0.853785404375)*x[1]**o+(-0.731028307335)*x[1]
        arg[2,1,2,0]=(-0.0805774093876)*x[0]**o+(0.949831911897)*x[0]+(0.511009248375)*x[1]**o+(-0.0192710206538)*x[1]
        arg[2,1,2,1]=(-0.987709862593)*x[0]**o+(-0.536077508064)*x[0]+(-0.0726882784026)*x[1]**o+(-0.27275457372)*x[1]
        arg[2,2,0,0]=(0.271001953238)*x[0]**o+(0.0446246860776)*x[0]+(0.103657462724)*x[1]**o+(0.359900072694)*x[1]
        arg[2,2,0,1]=(0.613460277623)*x[0]**o+(-0.23324730519)*x[0]+(0.188458167579)*x[1]**o+(-0.574087648909)*x[1]
        arg[2,2,1,0]=(-0.787444207567)*x[0]**o+(0.143772612473)*x[0]+(-0.720451072166)*x[1]**o+(0.936819171338)*x[1]
        arg[2,2,1,1]=(-0.219701379296)*x[0]**o+(-0.343100221149)*x[0]+(-0.972737482064)*x[1]**o+(-0.614208415137)*x[1]
        arg[2,2,2,0]=(0.0292070295438)*x[0]**o+(-0.694263454424)*x[0]+(0.188064440036)*x[1]**o+(0.0901763471697)*x[1]
        arg[2,2,2,1]=(0.407629585245)*x[0]**o+(-0.745619322197)*x[0]+(0.949044983116)*x[1]**o+(-0.316057288128)*x[1]
        arg[2,3,0,0]=(0.124212182643)*x[0]**o+(0.621703244759)*x[0]+(-0.0592136772296)*x[1]**o+(0.571981450786)*x[1]
        arg[2,3,0,1]=(-0.683989210442)*x[0]**o+(-0.318774594157)*x[0]+(0.850235115277)*x[1]**o+(-0.871075179838)*x[1]
        arg[2,3,1,0]=(0.478657634044)*x[0]**o+(-0.43702074673)*x[0]+(0.535572237843)*x[1]**o+(0.22354698199)*x[1]
        arg[2,3,1,1]=(0.0391087869133)*x[0]**o+(0.765140505702)*x[0]+(-0.66279875624)*x[1]**o+(-0.41964006149)*x[1]
        arg[2,3,2,0]=(-0.11123799186)*x[0]**o+(0.533398540082)*x[0]+(-0.931488199925)*x[1]**o+(-0.83080969281)*x[1]
        arg[2,3,2,1]=(-0.138259313941)*x[0]**o+(-0.508937684332)*x[0]+(0.583603445711)*x[1]**o+(-0.0782724872225)*x[1]
        arg[2,4,0,0]=(-0.316223936951)*x[0]**o+(0.329776289164)*x[0]+(-0.692491559115)*x[1]**o+(0.814572730294)*x[1]
        arg[2,4,0,1]=(0.74890960681)*x[0]**o+(-0.79167945549)*x[0]+(0.175328119487)*x[1]**o+(-0.253003240064)*x[1]
        arg[2,4,1,0]=(0.657827625141)*x[0]**o+(-0.268878597055)*x[0]+(0.247374286921)*x[1]**o+(0.414500041415)*x[1]
        arg[2,4,1,1]=(-0.187587046202)*x[0]**o+(-0.594983066039)*x[0]+(-0.203784406879)*x[1]**o+(0.122237566739)*x[1]
        arg[2,4,2,0]=(0.671210794607)*x[0]**o+(0.886706498415)*x[0]+(-0.368997114212)*x[1]**o+(0.0418311781187)*x[1]
        arg[2,4,2,1]=(-0.272424416171)*x[0]**o+(-0.537477247783)*x[0]+(0.0405655834868)*x[1]**o+(-0.179684833703)*x[1]
        arg[3,0,0,0]=(0.457602964193)*x[0]**o+(-0.186625111516)*x[0]+(-0.75244105343)*x[1]**o+(-0.278893445377)*x[1]
        arg[3,0,0,1]=(-0.690378721784)*x[0]**o+(0.0157608350188)*x[0]+(-0.203765434813)*x[1]**o+(0.69771096957)*x[1]
        arg[3,0,1,0]=(-0.884824688493)*x[0]**o+(0.26613451569)*x[0]+(0.226081491639)*x[1]**o+(0.597724298389)*x[1]
        arg[3,0,1,1]=(0.0372124540969)*x[0]**o+(-0.420242551812)*x[0]+(-0.625683613012)*x[1]**o+(0.279682305563)*x[1]
        arg[3,0,2,0]=(-0.570873872458)*x[0]**o+(-0.82361574923)*x[0]+(0.970204188337)*x[1]**o+(-0.242725003569)*x[1]
        arg[3,0,2,1]=(-0.714826095124)*x[0]**o+(-0.334939427182)*x[0]+(-0.730058530908)*x[1]**o+(0.287148154312)*x[1]
        arg[3,1,0,0]=(0.164234307274)*x[0]**o+(-0.388849394356)*x[0]+(0.391518896617)*x[1]**o+(0.00142750126236)*x[1]
        arg[3,1,0,1]=(-0.327700781196)*x[0]**o+(0.0442604682565)*x[0]+(-0.683834466536)*x[1]**o+(-0.665567774985)*x[1]
        arg[3,1,1,0]=(0.135979934041)*x[0]**o+(-0.822614011173)*x[0]+(-0.946061768113)*x[1]**o+(-0.585442102671)*x[1]
        arg[3,1,1,1]=(0.235568769871)*x[0]**o+(0.738233870703)*x[0]+(0.437750339594)*x[1]**o+(-0.44521724321)*x[1]
        arg[3,1,2,0]=(0.449256898185)*x[0]**o+(-0.464211755196)*x[0]+(0.616886411002)*x[1]**o+(0.885500269963)*x[1]
        arg[3,1,2,1]=(0.0851105760559)*x[0]**o+(0.888976499957)*x[0]+(0.0842441111314)*x[1]**o+(-0.366994083394)*x[1]
        arg[3,2,0,0]=(0.77120050858)*x[0]**o+(0.354835584949)*x[0]+(0.406813878866)*x[1]**o+(-0.600042478179)*x[1]
        arg[3,2,0,1]=(-0.747876879184)*x[0]**o+(0.527204926634)*x[0]+(0.134965492713)*x[1]**o+(0.497662739459)*x[1]
        arg[3,2,1,0]=(0.367550591163)*x[0]**o+(-0.700608509353)*x[0]+(0.0737696587931)*x[1]**o+(-0.164466899816)*x[1]
        arg[3,2,1,1]=(0.0908791429433)*x[0]**o+(-0.168468767791)*x[0]+(0.85978873292)*x[1]**o+(-0.785241412884)*x[1]
        arg[3,2,2,0]=(-0.296455818292)*x[0]**o+(-0.0277004420512)*x[0]+(0.409183414933)*x[1]**o+(-0.879831945543)*x[1]
        arg[3,2,2,1]=(0.971512010223)*x[0]**o+(-0.811249982262)*x[0]+(-0.846549204737)*x[1]**o+(0.689986186691)*x[1]
        arg[3,3,0,0]=(-0.800235545702)*x[0]**o+(-0.796021196346)*x[0]+(-0.0660744375539)*x[1]**o+(-0.826632192842)*x[1]
        arg[3,3,0,1]=(0.305678237361)*x[0]**o+(0.203936000462)*x[0]+(-0.583171879217)*x[1]**o+(0.924112859119)*x[1]
        arg[3,3,1,0]=(-0.937148913762)*x[0]**o+(-0.172547991203)*x[0]+(0.928491154238)*x[1]**o+(-0.502876301351)*x[1]
        arg[3,3,1,1]=(0.731413856616)*x[0]**o+(0.837213097155)*x[0]+(-0.650032764358)*x[1]**o+(0.96973961589)*x[1]
        arg[3,3,2,0]=(-0.501994458598)*x[0]**o+(-0.705938508685)*x[0]+(0.664394117366)*x[1]**o+(0.692950029135)*x[1]
        arg[3,3,2,1]=(-0.0948399323409)*x[0]**o+(-0.686744554674)*x[0]+(0.577765681583)*x[1]**o+(0.969310750204)*x[1]
        arg[3,4,0,0]=(-0.484011788432)*x[0]**o+(-0.560744139846)*x[0]+(0.072097608773)*x[1]**o+(-0.103052228904)*x[1]
        arg[3,4,0,1]=(-0.201158934695)*x[0]**o+(0.463105335811)*x[0]+(0.733365706682)*x[1]**o+(-0.330035241794)*x[1]
        arg[3,4,1,0]=(0.252379277102)*x[0]**o+(-0.110596657969)*x[0]+(0.763687257878)*x[1]**o+(-0.345598371263)*x[1]
        arg[3,4,1,1]=(0.0868412855617)*x[0]**o+(0.385098679406)*x[0]+(0.291487288978)*x[1]**o+(-0.112145846209)*x[1]
        arg[3,4,2,0]=(-0.913724685255)*x[0]**o+(-0.0348735884313)*x[0]+(0.550688528861)*x[1]**o+(0.798394152329)*x[1]
        arg[3,4,2,1]=(-0.92870523219)*x[0]**o+(0.177354002469)*x[0]+(0.350393765313)*x[1]**o+(-0.974997198741)*x[1]
        ref[0,0,0,0]=(0.815821053922)/(o+1.)+(0.302879577355)
        ref[0,0,0,1]=(-0.223824520603)/(o+1.)+(0.690659426687)
        ref[0,0,1,0]=(-0.501731412433)/(o+1.)+(0.0102964270565)
        ref[0,0,1,1]=(-0.731540589167)/(o+1.)+(0.348284469462)
        ref[0,0,2,0]=(-0.121523684012)/(o+1.)+(0.184348962759)
        ref[0,0,2,1]=(-0.4529485642)/(o+1.)+(-0.830798233026)
        ref[0,1,0,0]=(-0.244463483906)/(o+1.)+(0.226296188971)
        ref[0,1,0,1]=(1.2203272116)/(o+1.)+(-0.693494218115)
        ref[0,1,1,0]=(-0.55576623697)/(o+1.)+(-0.273623989583)
        ref[0,1,1,1]=(-0.316836585171)/(o+1.)+(-0.836135387777)
        ref[0,1,2,0]=(-0.133732244387)/(o+1.)+(-0.875415201605)
        ref[0,1,2,1]=(-0.281487947177)/(o+1.)+(-0.668499911274)
        ref[0,2,0,0]=(-0.896781862729)/(o+1.)+(-0.0225848820142)
        ref[0,2,0,1]=(-1.70326692867)/(o+1.)+(-0.1693379432)
        ref[0,2,1,0]=(-0.00084798161079)/(o+1.)+(0.343359275275)
        ref[0,2,1,1]=(0.609064372767)/(o+1.)+(-0.369639962727)
        ref[0,2,2,0]=(-0.759042888708)/(o+1.)+(0.321708104548)
        ref[0,2,2,1]=(0.760967724015)/(o+1.)+(0.0268897973234)
        ref[0,3,0,0]=(-1.3915636468)/(o+1.)+(-0.230755462281)
        ref[0,3,0,1]=(-0.416449082748)/(o+1.)+(0.480436529329)
        ref[0,3,1,0]=(-0.238553658798)/(o+1.)+(-0.029283546006)
        ref[0,3,1,1]=(-1.36933302275)/(o+1.)+(-0.221757606321)
        ref[0,3,2,0]=(-0.872999155991)/(o+1.)+(0.69296766968)
        ref[0,3,2,1]=(0.833661507716)/(o+1.)+(0.321357009466)
        ref[0,4,0,0]=(1.4358026373)/(o+1.)+(0.311715942547)
        ref[0,4,0,1]=(-0.00811900044778)/(o+1.)+(-0.76710974695)
        ref[0,4,1,0]=(-0.92034101297)/(o+1.)+(-0.0309623888275)
        ref[0,4,1,1]=(0.690460334263)/(o+1.)+(0.783809038221)
        ref[0,4,2,0]=(0.779851986824)/(o+1.)+(-0.214799981386)
        ref[0,4,2,1]=(-0.566157259598)/(o+1.)+(0.296555712092)
        ref[1,0,0,0]=(-1.70267943485)/(o+1.)+(0.0444622873215)
        ref[1,0,0,1]=(0.399342875087)/(o+1.)+(0.0341080904363)
        ref[1,0,1,0]=(0.839377637409)/(o+1.)+(0.158619241351)
        ref[1,0,1,1]=(-0.0142823080176)/(o+1.)+(0.693714763021)
        ref[1,0,2,0]=(-0.97036633864)/(o+1.)+(0.264110057567)
        ref[1,0,2,1]=(0.104223436728)/(o+1.)+(0.328785142295)
        ref[1,1,0,0]=(1.09779156365)/(o+1.)+(-0.00408578628855)
        ref[1,1,0,1]=(-0.914750679597)/(o+1.)+(0.206573942886)
        ref[1,1,1,0]=(0.150419426085)/(o+1.)+(-0.0846024372156)
        ref[1,1,1,1]=(-1.04733819847)/(o+1.)+(-0.157756242641)
        ref[1,1,2,0]=(-1.49193735459)/(o+1.)+(0.391723285192)
        ref[1,1,2,1]=(-0.250600389657)/(o+1.)+(-0.456841431612)
        ref[1,2,0,0]=(1.21154796884)/(o+1.)+(-0.377275533205)
        ref[1,2,0,1]=(-0.785475967917)/(o+1.)+(0.2413362399)
        ref[1,2,1,0]=(1.20750352255)/(o+1.)+(-0.409504204121)
        ref[1,2,1,1]=(0.271471015188)/(o+1.)+(-0.136640402636)
        ref[1,2,2,0]=(-0.741344603747)/(o+1.)+(0.00668995653618)
        ref[1,2,2,1]=(0.462000684606)/(o+1.)+(0.364894712961)
        ref[1,3,0,0]=(0.0273564369347)/(o+1.)+(0.673962086152)
        ref[1,3,0,1]=(0.285276349085)/(o+1.)+(0.385675061073)
        ref[1,3,1,0]=(0.851307298344)/(o+1.)+(0.118422744971)
        ref[1,3,1,1]=(0.235593686643)/(o+1.)+(0.244705242716)
        ref[1,3,2,0]=(-0.359345731222)/(o+1.)+(-0.124958138388)
        ref[1,3,2,1]=(0.590170220995)/(o+1.)+(0.429975100757)
        ref[1,4,0,0]=(0.987134792715)/(o+1.)+(-0.0725353868253)
        ref[1,4,0,1]=(0.623734296247)/(o+1.)+(0.489855522024)
        ref[1,4,1,0]=(-0.913537095843)/(o+1.)+(0.727739353556)
        ref[1,4,1,1]=(-0.656988917238)/(o+1.)+(0.274753842895)
        ref[1,4,2,0]=(1.03400922475)/(o+1.)+(-0.00621482335664)
        ref[1,4,2,1]=(-1.0367802184)/(o+1.)+(-0.266633530803)
        ref[2,0,0,0]=(-0.391593893803)/(o+1.)+(0.457472062333)
        ref[2,0,0,1]=(0.393476325223)/(o+1.)+(-0.911233333506)
        ref[2,0,1,0]=(-0.0199296737245)/(o+1.)+(0.328542898457)
        ref[2,0,1,1]=(-0.618353771051)/(o+1.)+(0.0279228750664)
        ref[2,0,2,0]=(0.570494475025)/(o+1.)+(-0.374533325635)
        ref[2,0,2,1]=(-1.21079806839)/(o+1.)+(-0.632909492494)
        ref[2,1,0,0]=(1.32715885433)/(o+1.)+(0.730744204126)
        ref[2,1,0,1]=(1.43666536922)/(o+1.)+(0.279106975714)
        ref[2,1,1,0]=(-0.603800395047)/(o+1.)+(-0.716035662434)
        ref[2,1,1,1]=(1.35897125219)/(o+1.)+(-0.721199872542)
        ref[2,1,2,0]=(0.430431838988)/(o+1.)+(0.465280445622)
        ref[2,1,2,1]=(-1.060398141)/(o+1.)+(-0.404416040892)
        ref[2,2,0,0]=(0.374659415961)/(o+1.)+(0.202262379386)
        ref[2,2,0,1]=(0.801918445202)/(o+1.)+(-0.403667477049)
        ref[2,2,1,0]=(-1.50789527973)/(o+1.)+(0.540295891905)
        ref[2,2,1,1]=(-1.19243886136)/(o+1.)+(-0.478654318143)
        ref[2,2,2,0]=(0.21727146958)/(o+1.)+(-0.302043553627)
        ref[2,2,2,1]=(1.35667456836)/(o+1.)+(-0.530838305162)
        ref[2,3,0,0]=(0.0649985054133)/(o+1.)+(0.596842347772)
        ref[2,3,0,1]=(0.166245904835)/(o+1.)+(-0.594924886997)
        ref[2,3,1,0]=(1.01422987189)/(o+1.)+(-0.10673688237)
        ref[2,3,1,1]=(-0.623689969327)/(o+1.)+(0.172750222106)
        ref[2,3,2,0]=(-1.04272619178)/(o+1.)+(-0.148705576364)
        ref[2,3,2,1]=(0.445344131771)/(o+1.)+(-0.293605085777)
        ref[2,4,0,0]=(-1.00871549607)/(o+1.)+(0.572174509729)
        ref[2,4,0,1]=(0.924237726297)/(o+1.)+(-0.522341347777)
        ref[2,4,1,0]=(0.905201912063)/(o+1.)+(0.0728107221801)
        ref[2,4,1,1]=(-0.391371453081)/(o+1.)+(-0.23637274965)
        ref[2,4,2,0]=(0.302213680395)/(o+1.)+(0.464268838267)
        ref[2,4,2,1]=(-0.231858832685)/(o+1.)+(-0.358581040743)
        ref[3,0,0,0]=(-0.294838089237)/(o+1.)+(-0.232759278447)
        ref[3,0,0,1]=(-0.894144156597)/(o+1.)+(0.356735902294)
        ref[3,0,1,0]=(-0.658743196854)/(o+1.)+(0.43192940704)
        ref[3,0,1,1]=(-0.588471158915)/(o+1.)+(-0.0702801231243)
        ref[3,0,2,0]=(0.399330315879)/(o+1.)+(-0.5331703764)
        ref[3,0,2,1]=(-1.44488462603)/(o+1.)+(-0.0238956364351)
        ref[3,1,0,0]=(0.555753203891)/(o+1.)+(-0.193710946547)
        ref[3,1,0,1]=(-1.01153524773)/(o+1.)+(-0.310653653364)
        ref[3,1,1,0]=(-0.810081834072)/(o+1.)+(-0.704028056922)
        ref[3,1,1,1]=(0.673319109465)/(o+1.)+(0.146508313746)
        ref[3,1,2,0]=(1.06614330919)/(o+1.)+(0.210644257384)
        ref[3,1,2,1]=(0.169354687187)/(o+1.)+(0.260991208281)
        ref[3,2,0,0]=(1.17801438745)/(o+1.)+(-0.122603446615)
        ref[3,2,0,1]=(-0.612911386471)/(o+1.)+(0.512433833047)
        ref[3,2,1,0]=(0.441320249956)/(o+1.)+(-0.432537704584)
        ref[3,2,1,1]=(0.950667875863)/(o+1.)+(-0.476855090338)
        ref[3,2,2,0]=(0.112727596641)/(o+1.)+(-0.453766193797)
        ref[3,2,2,1]=(0.124962805486)/(o+1.)+(-0.0606318977858)
        ref[3,3,0,0]=(-0.866309983256)/(o+1.)+(-0.811326694594)
        ref[3,3,0,1]=(-0.277493641857)/(o+1.)+(0.564024429791)
        ref[3,3,1,0]=(-0.00865775952435)/(o+1.)+(-0.337712146277)
        ref[3,3,1,1]=(0.0813810922576)/(o+1.)+(0.903476356523)
        ref[3,3,2,0]=(0.162399658768)/(o+1.)+(-0.00649423977499)
        ref[3,3,2,1]=(0.482925749243)/(o+1.)+(0.141283097765)
        ref[3,4,0,0]=(-0.411914179659)/(o+1.)+(-0.331898184375)
        ref[3,4,0,1]=(0.532206771987)/(o+1.)+(0.0665350470083)
        ref[3,4,1,0]=(1.01606653498)/(o+1.)+(-0.228097514616)
        ref[3,4,1,1]=(0.37832857454)/(o+1.)+(0.136476416599)
        ref[3,4,2,0]=(-0.363036156393)/(o+1.)+(0.381760281949)
        ref[3,4,2,1]=(-0.578311466877)/(o+1.)+(-0.398821598136)
      else:
        arg[0,0,0,0]=(0.860676142569)*x[0]**o+(-0.156046991162)*x[0]+(-0.0462355579656)*x[1]**o+(0.70842844961)*x[1]+(0.617299123548)*x[2]**o+(-0.865174329565)*x[2]
        arg[0,0,0,1]=(-0.284308060865)*x[0]**o+(0.788848579389)*x[0]+(0.629321669042)*x[1]**o+(0.066637849809)*x[1]+(0.114152381009)*x[2]**o+(-0.466687745846)*x[2]
        arg[0,0,1,0]=(-0.690823198517)*x[0]**o+(-0.141510804573)*x[0]+(0.969978831967)*x[1]**o+(0.307926382658)*x[1]+(0.710373009094)*x[2]**o+(-0.494828315907)*x[2]
        arg[0,0,1,1]=(0.953361413913)*x[0]**o+(-0.103742611936)*x[0]+(0.63934955554)*x[1]**o+(-0.0778114476564)*x[1]+(-0.221185122698)*x[2]**o+(-0.287432047404)*x[2]
        arg[0,0,2,0]=(0.378306380558)*x[0]**o+(-0.117743922672)*x[0]+(-0.0606376630208)*x[1]**o+(-0.59439189505)*x[1]+(0.10507320657)*x[2]**o+(0.856411423067)*x[2]
        arg[0,0,2,1]=(0.777734672779)*x[0]**o+(0.286956002007)*x[0]+(-0.931811626583)*x[1]**o+(-0.806626659661)*x[1]+(-0.420390949282)*x[2]**o+(0.854080006578)*x[2]
        arg[0,1,0,0]=(0.403426948964)*x[0]**o+(0.707099419296)*x[0]+(0.593137130456)*x[1]**o+(-0.757560771598)*x[1]+(0.369632217934)*x[2]**o+(-0.348480285397)*x[2]
        arg[0,1,0,1]=(-0.491454189782)*x[0]**o+(0.547276640565)*x[0]+(-0.955045542708)*x[1]**o+(0.529989872531)*x[1]+(-0.161086874799)*x[2]**o+(-0.181810040724)*x[2]
        arg[0,1,1,0]=(0.46282177104)*x[0]**o+(0.254346937251)*x[0]+(0.824797105546)*x[1]**o+(-0.449979162246)*x[1]+(0.22307703071)*x[2]**o+(-0.592072135668)*x[2]
        arg[0,1,1,1]=(0.329230675141)*x[0]**o+(-0.733034697145)*x[0]+(0.762195609347)*x[1]**o+(-0.639097591362)*x[1]+(0.765354927905)*x[2]**o+(0.313571493405)*x[2]
        arg[0,1,2,0]=(0.3874127072)*x[0]**o+(-0.0569429727569)*x[0]+(-0.203121198708)*x[1]**o+(-0.736379499164)*x[1]+(-0.980629375436)*x[2]**o+(0.955793660048)*x[2]
        arg[0,1,2,1]=(-0.656255469455)*x[0]**o+(0.711904808699)*x[0]+(-0.363253678399)*x[1]**o+(0.921875000807)*x[1]+(0.686394816431)*x[2]**o+(0.818500716704)*x[2]
        arg[0,2,0,0]=(-0.931555347846)*x[0]**o+(-0.249573400742)*x[0]+(0.617166716404)*x[1]**o+(-0.929587051108)*x[1]+(0.830928715064)*x[2]**o+(0.461637431613)*x[2]
        arg[0,2,0,1]=(-0.8677719945)*x[0]**o+(-0.176892180445)*x[0]+(-0.881887456464)*x[1]**o+(0.923355677513)*x[1]+(-0.978544668453)*x[2]**o+(-0.427009384287)*x[2]
        arg[0,2,1,0]=(-0.0707937433084)*x[0]**o+(-0.872892141393)*x[0]+(0.606234798381)*x[1]**o+(0.935845076847)*x[1]+(-0.97559052201)*x[2]**o+(-0.31138518209)*x[2]
        arg[0,2,1,1]=(0.450210597865)*x[0]**o+(0.842939440378)*x[0]+(-0.333422856577)*x[1]**o+(-0.492809679443)*x[1]+(-0.254728278043)*x[2]**o+(-0.0702767077952)*x[2]
        arg[0,2,2,0]=(-0.250463216903)*x[0]**o+(0.756576293763)*x[0]+(0.596902629953)*x[1]**o+(-0.887080794947)*x[1]+(0.027319948008)*x[2]**o+(0.552496782776)*x[2]
        arg[0,2,2,1]=(-0.0883102140591)*x[0]**o+(0.988759610101)*x[0]+(-0.208152392508)*x[1]**o+(0.938999108564)*x[1]+(-0.773883611417)*x[2]**o+(0.653413306338)*x[2]
        arg[0,3,0,0]=(-0.399644211606)*x[0]**o+(-0.978595776407)*x[0]+(0.99454650986)*x[1]**o+(0.143313668715)*x[1]+(0.328592750842)*x[2]**o+(-0.958804581101)*x[2]
        arg[0,3,0,1]=(-0.0667966885959)*x[0]**o+(0.332653574704)*x[0]+(-0.762025150916)*x[1]**o+(0.895044013374)*x[1]+(0.627013843362)*x[2]**o+(-0.682392672232)*x[2]
        arg[0,3,1,0]=(-0.834479453773)*x[0]**o+(-0.375930859033)*x[0]+(0.129507966861)*x[1]**o+(0.995364280408)*x[1]+(-0.126081191572)*x[2]**o+(0.0344262779491)*x[2]
        arg[0,3,1,1]=(0.796952126596)*x[0]**o+(0.310562243091)*x[0]+(0.854548623788)*x[1]**o+(-0.719100537284)*x[1]+(-0.236266783438)*x[2]**o+(-0.646647129209)*x[2]
        arg[0,3,2,0]=(0.844073289115)*x[0]**o+(0.423300345347)*x[0]+(0.126354796852)*x[1]**o+(0.180372257606)*x[1]+(-0.564712722146)*x[2]**o+(0.466988430183)*x[2]
        arg[0,3,2,1]=(0.228775824367)*x[0]**o+(0.40701094831)*x[0]+(0.223101054593)*x[1]**o+(-0.894937905383)*x[1]+(0.425392942773)*x[2]**o+(-0.844238667169)*x[2]
        arg[0,4,0,0]=(0.3363848738)*x[0]**o+(-0.985327066062)*x[0]+(-0.107847971678)*x[1]**o+(0.596121801488)*x[1]+(0.655589801663)*x[2]**o+(-0.729650731949)*x[2]
        arg[0,4,0,1]=(-0.176435679219)*x[0]**o+(0.382558711725)*x[0]+(-0.658212492337)*x[1]**o+(0.584904267612)*x[1]+(-0.534684876705)*x[2]**o+(-0.441262614915)*x[2]
        arg[0,4,1,0]=(0.281301737167)*x[0]**o+(-0.642486274791)*x[0]+(0.420753884869)*x[1]**o+(0.482835955591)*x[1]+(0.169431333788)*x[2]**o+(0.922841193825)*x[2]
        arg[0,4,1,1]=(0.235812590604)*x[0]**o+(-0.446233855396)*x[0]+(0.465015868566)*x[1]**o+(-0.780441276888)*x[1]+(0.513632761351)*x[2]**o+(-0.0576482811678)*x[2]
        arg[0,4,2,0]=(-0.617245874462)*x[0]**o+(0.183661473466)*x[0]+(-0.711148911239)*x[1]**o+(0.125228991306)*x[1]+(-0.500218603959)*x[2]**o+(0.70597577229)*x[2]
        arg[0,4,2,1]=(0.56330304117)*x[0]**o+(0.631284986628)*x[0]+(-0.222485175776)*x[1]**o+(0.0639235737228)*x[1]+(-0.276613868747)*x[2]**o+(0.407488564857)*x[2]
        arg[1,0,0,0]=(0.159032146915)*x[0]**o+(-0.716040989693)*x[0]+(0.410709172613)*x[1]**o+(-0.527667963652)*x[1]+(-0.160515156497)*x[2]**o+(0.294448742382)*x[2]
        arg[1,0,0,1]=(0.941139035835)*x[0]**o+(0.707410046094)*x[0]+(-0.172095410803)*x[1]**o+(-0.967035204304)*x[1]+(0.870395272646)*x[2]**o+(0.830870986802)*x[2]
        arg[1,0,1,0]=(0.731015285309)*x[0]**o+(0.450519048697)*x[0]+(-0.484568714675)*x[1]**o+(0.454201220108)*x[1]+(-0.161850459099)*x[2]**o+(-0.617463263396)*x[2]
        arg[1,0,1,1]=(-0.958841628847)*x[0]**o+(-0.0667813400021)*x[0]+(0.659551780082)*x[1]**o+(0.514424015304)*x[1]+(0.920583608125)*x[2]**o+(0.89564899572)*x[2]
        arg[1,0,2,0]=(0.0482630479817)*x[0]**o+(0.911326064322)*x[0]+(-0.275134422092)*x[1]**o+(0.216495232459)*x[1]+(-0.973586171195)*x[2]**o+(0.275694688611)*x[2]
        arg[1,0,2,1]=(0.769733688293)*x[0]**o+(0.771829018421)*x[0]+(-0.22434549079)*x[1]**o+(0.131966885817)*x[1]+(-0.109424705679)*x[2]**o+(-0.369585136379)*x[2]
        arg[1,1,0,0]=(-0.0160180905877)*x[0]**o+(-0.268260755267)*x[0]+(-0.213940412022)*x[1]**o+(0.665987805279)*x[1]+(0.968879828017)*x[2]**o+(-0.449435538913)*x[2]
        arg[1,1,0,1]=(0.361055351825)*x[0]**o+(0.463294478488)*x[0]+(0.692655668486)*x[1]**o+(0.151415573206)*x[1]+(0.902280461515)*x[2]**o+(0.690926934073)*x[2]
        arg[1,1,1,0]=(0.591998602299)*x[0]**o+(0.516482177709)*x[0]+(-0.298511909182)*x[1]**o+(0.281818071894)*x[1]+(0.207089953979)*x[2]**o+(0.917004722388)*x[2]
        arg[1,1,1,1]=(0.479832408236)*x[0]**o+(-0.643374914165)*x[0]+(0.370844835249)*x[1]**o+(0.560458499239)*x[1]+(-0.865852319767)*x[2]**o+(0.560572130123)*x[2]
        arg[1,1,2,0]=(0.709293526635)*x[0]**o+(-0.388347223155)*x[0]+(0.60797330273)*x[1]**o+(-0.847624783099)*x[1]+(-0.0476367756382)*x[2]**o+(0.637841988478)*x[2]
        arg[1,1,2,1]=(0.343458045422)*x[0]**o+(-0.764451449587)*x[0]+(-0.444256801761)*x[1]**o+(0.385385997921)*x[1]+(-0.722095136671)*x[2]**o+(0.677852473207)*x[2]
        arg[1,2,0,0]=(0.710372887392)*x[0]**o+(-0.628494679119)*x[0]+(-0.59513507905)*x[1]**o+(-0.527824874295)*x[1]+(-0.611189170789)*x[2]**o+(0.090651976806)*x[2]
        arg[1,2,0,1]=(-0.980160150289)*x[0]**o+(-0.0286097581598)*x[0]+(-0.407906975953)*x[1]**o+(-0.319422835951)*x[1]+(-0.489787310147)*x[2]**o+(-0.694386341354)*x[2]
        arg[1,2,1,0]=(-0.416867224376)*x[0]**o+(0.631406877489)*x[0]+(-0.93939195942)*x[1]**o+(0.266858600313)*x[1]+(0.982173221612)*x[2]**o+(0.112901364617)*x[2]
        arg[1,2,1,1]=(0.412001480348)*x[0]**o+(0.624281489887)*x[0]+(0.306558457751)*x[1]**o+(0.95864454419)*x[1]+(0.299175627954)*x[2]**o+(-0.0531991017347)*x[2]
        arg[1,2,2,0]=(-0.0342261621741)*x[0]**o+(-0.587245704486)*x[0]+(-0.976407464107)*x[1]**o+(-0.0157215030322)*x[1]+(-0.529857472095)*x[2]**o+(-0.419626697673)*x[2]
        arg[1,2,2,1]=(-0.681737301037)*x[0]**o+(-0.980349460702)*x[0]+(-0.839469631393)*x[1]**o+(-0.412417814907)*x[1]+(0.451717398947)*x[2]**o+(0.962167685987)*x[2]
        arg[1,3,0,0]=(0.283453965913)*x[0]**o+(0.204445403787)*x[0]+(-0.806150376236)*x[1]**o+(0.122187094069)*x[1]+(-0.679291367964)*x[2]**o+(-0.875353966489)*x[2]
        arg[1,3,0,1]=(0.115974475175)*x[0]**o+(0.701473497654)*x[0]+(0.298012120304)*x[1]**o+(0.567597901709)*x[1]+(0.721211831476)*x[2]**o+(-0.255389572775)*x[2]
        arg[1,3,1,0]=(0.393891431092)*x[0]**o+(-0.563382537925)*x[0]+(0.974194829106)*x[1]**o+(0.659688559688)*x[1]+(-0.394266143828)*x[2]**o+(0.875266200191)*x[2]
        arg[1,3,1,1]=(-0.456930975697)*x[0]**o+(0.334590626231)*x[0]+(-0.227176883362)*x[1]**o+(0.76775699173)*x[1]+(0.839371167887)*x[2]**o+(-0.155473916326)*x[2]
        arg[1,3,2,0]=(0.346761981787)*x[0]**o+(-0.625121706322)*x[0]+(0.340655154638)*x[1]**o+(-0.0646323546999)*x[1]+(-0.211347402677)*x[2]**o+(0.233661266919)*x[2]
        arg[1,3,2,1]=(-0.086096644956)*x[0]**o+(-0.845623653209)*x[0]+(-0.0866093549957)*x[1]**o+(-0.467286349011)*x[1]+(-0.0621455515357)*x[2]**o+(0.164779979085)*x[2]
        arg[1,4,0,0]=(0.42022138208)*x[0]**o+(0.136984451355)*x[0]+(0.664770979086)*x[1]**o+(-0.537368896193)*x[1]+(0.874256726691)*x[2]**o+(-0.58614100824)*x[2]
        arg[1,4,0,1]=(-0.805860964792)*x[0]**o+(0.758511304044)*x[0]+(-0.574686863541)*x[1]**o+(-0.410873101247)*x[1]+(0.932015171551)*x[2]**o+(0.439987195351)*x[2]
        arg[1,4,1,0]=(-0.515187049123)*x[0]**o+(0.308078404334)*x[0]+(0.940959771774)*x[1]**o+(0.466269579691)*x[1]+(-0.190617144843)*x[2]**o+(0.738172533582)*x[2]
        arg[1,4,1,1]=(0.283326036586)*x[0]**o+(-0.693631747857)*x[0]+(0.478965080319)*x[1]**o+(-0.0465389396775)*x[1]+(-0.72886286959)*x[2]**o+(-0.9346424149)*x[2]
        arg[1,4,2,0]=(-0.71539766285)*x[0]**o+(0.361875987071)*x[0]+(0.905270678876)*x[1]**o+(-0.675024725702)*x[1]+(0.378259473658)*x[2]**o+(0.0520394226711)*x[2]
        arg[1,4,2,1]=(-0.319600525923)*x[0]**o+(0.0449956327134)*x[0]+(0.707418658521)*x[1]**o+(0.210137972776)*x[1]+(-0.767860728688)*x[2]**o+(0.232340825758)*x[2]
        arg[2,0,0,0]=(0.92844243651)*x[0]**o+(0.956810319504)*x[0]+(0.762198571194)*x[1]**o+(-0.363107634629)*x[1]+(0.0750393505254)*x[2]**o+(-0.43573581041)*x[2]
        arg[2,0,0,1]=(-0.193613121742)*x[0]**o+(0.440027871839)*x[0]+(-0.607905506346)*x[1]**o+(0.894840354958)*x[1]+(0.601279436178)*x[2]**o+(0.46806560762)*x[2]
        arg[2,0,1,0]=(0.370901517567)*x[0]**o+(-0.738769905647)*x[0]+(-0.698692979664)*x[1]**o+(-0.286318840335)*x[1]+(-0.879367315688)*x[2]**o+(-0.959912132991)*x[2]
        arg[2,0,1,1]=(-0.278939533774)*x[0]**o+(0.183795552426)*x[0]+(-0.341943551466)*x[1]**o+(0.491847260291)*x[1]+(-0.141030429384)*x[2]**o+(0.256287855539)*x[2]
        arg[2,0,2,0]=(-0.135930357616)*x[0]**o+(-0.382515749949)*x[0]+(-0.43030873581)*x[1]**o+(-0.0211891159981)*x[1]+(-0.377865120002)*x[2]**o+(0.109942325029)*x[2]
        arg[2,0,2,1]=(0.288827267201)*x[0]**o+(-0.907970544468)*x[0]+(-0.622258309213)*x[1]**o+(-0.763273879894)*x[1]+(0.438715013671)*x[2]**o+(-0.93873140988)*x[2]
        arg[2,1,0,0]=(-0.963487369906)*x[0]**o+(0.899051290288)*x[0]+(-0.268257956548)*x[1]**o+(-0.229092726584)*x[1]+(0.427024112314)*x[2]**o+(0.0435138095138)*x[2]
        arg[2,1,0,1]=(0.89996926597)*x[0]**o+(-0.206500706229)*x[0]+(-0.191769123095)*x[1]**o+(-0.348924155298)*x[1]+(-0.74405653776)*x[2]**o+(0.0724946691864)*x[2]
        arg[2,1,1,0]=(0.338170397901)*x[0]**o+(-0.725761081242)*x[0]+(-0.45928272676)*x[1]**o+(0.795535326783)*x[1]+(-0.316443844644)*x[2]**o+(-0.844205401892)*x[2]
        arg[2,1,1,1]=(0.509131757681)*x[0]**o+(0.319175801535)*x[0]+(-0.807058926279)*x[1]**o+(0.63928269996)*x[1]+(0.219324073432)*x[2]**o+(0.351217517434)*x[2]
        arg[2,1,2,0]=(-0.534090721611)*x[0]**o+(0.27146424294)*x[0]+(0.250778065716)*x[1]**o+(-0.901322994167)*x[1]+(-0.57588768699)*x[2]**o+(0.405043003456)*x[2]
        arg[2,1,2,1]=(0.101699906269)*x[0]**o+(0.361960318738)*x[0]+(0.278974923593)*x[1]**o+(-0.336289352037)*x[1]+(-0.677353936215)*x[2]**o+(0.0726793347753)*x[2]
        arg[2,2,0,0]=(-0.734961283823)*x[0]**o+(-0.593268557578)*x[0]+(0.310896858565)*x[1]**o+(-0.482326137951)*x[1]+(0.886472957099)*x[2]**o+(-0.0403563476581)*x[2]
        arg[2,2,0,1]=(-0.166071725106)*x[0]**o+(-0.454245686818)*x[0]+(-0.892674508991)*x[1]**o+(0.581554190251)*x[1]+(0.995323654908)*x[2]**o+(-0.0336714433349)*x[2]
        arg[2,2,1,0]=(-0.437491848099)*x[0]**o+(-0.195039508213)*x[0]+(0.822011086215)*x[1]**o+(0.771044412863)*x[1]+(0.697534261252)*x[2]**o+(0.666510217725)*x[2]
        arg[2,2,1,1]=(-0.529148276275)*x[0]**o+(-0.093016460024)*x[0]+(0.663641010649)*x[1]**o+(0.940439359843)*x[1]+(0.754151159716)*x[2]**o+(0.852349310591)*x[2]
        arg[2,2,2,0]=(0.859523244682)*x[0]**o+(0.704825125128)*x[0]+(0.221832129598)*x[1]**o+(0.573940020456)*x[1]+(0.935673751591)*x[2]**o+(0.23235170475)*x[2]
        arg[2,2,2,1]=(0.0359047473512)*x[0]**o+(0.836816935443)*x[0]+(0.6411608273)*x[1]**o+(0.291526953908)*x[1]+(0.735564515362)*x[2]**o+(-0.465781621769)*x[2]
        arg[2,3,0,0]=(0.352740040216)*x[0]**o+(0.54224020581)*x[0]+(0.681356174115)*x[1]**o+(-0.97589117869)*x[1]+(0.880755719884)*x[2]**o+(0.562259403362)*x[2]
        arg[2,3,0,1]=(0.563708471054)*x[0]**o+(0.525523219723)*x[0]+(0.423068360374)*x[1]**o+(0.29239160821)*x[1]+(0.874360368885)*x[2]**o+(0.922265851359)*x[2]
        arg[2,3,1,0]=(0.694309936305)*x[0]**o+(-0.409224769672)*x[0]+(-0.748835864803)*x[1]**o+(-0.21841534641)*x[1]+(0.497059735624)*x[2]**o+(-0.753234694148)*x[2]
        arg[2,3,1,1]=(0.109706510352)*x[0]**o+(-0.0160720403678)*x[0]+(0.888778080744)*x[1]**o+(-0.938692716268)*x[1]+(0.529262951478)*x[2]**o+(0.481430680214)*x[2]
        arg[2,3,2,0]=(0.65368659132)*x[0]**o+(-0.988176396392)*x[0]+(0.989546739333)*x[1]**o+(0.23178715846)*x[1]+(-0.557645438328)*x[2]**o+(0.063228513147)*x[2]
        arg[2,3,2,1]=(0.286316141831)*x[0]**o+(0.668138463424)*x[0]+(-0.950430671533)*x[1]**o+(-0.978012107187)*x[1]+(-0.107618382464)*x[2]**o+(-0.684084421276)*x[2]
        arg[2,4,0,0]=(-0.63510659432)*x[0]**o+(0.264242935995)*x[0]+(0.960647241624)*x[1]**o+(0.58483379289)*x[1]+(0.7255270271)*x[2]**o+(-0.476151653985)*x[2]
        arg[2,4,0,1]=(-0.221953908705)*x[0]**o+(-0.994143786509)*x[0]+(-0.359525421944)*x[1]**o+(0.533966197668)*x[1]+(-0.29739666047)*x[2]**o+(-0.0824204124466)*x[2]
        arg[2,4,1,0]=(-0.447043625871)*x[0]**o+(-0.904485300564)*x[0]+(0.26179071803)*x[1]**o+(0.437320719437)*x[1]+(-0.56501466909)*x[2]**o+(0.129662379166)*x[2]
        arg[2,4,1,1]=(0.259809756366)*x[0]**o+(0.446788808701)*x[0]+(-0.143702396023)*x[1]**o+(0.879531090832)*x[1]+(0.410047703156)*x[2]**o+(0.755750006468)*x[2]
        arg[2,4,2,0]=(-0.730881615951)*x[0]**o+(-0.357110792705)*x[0]+(0.552722654621)*x[1]**o+(0.965623848242)*x[1]+(0.555951256049)*x[2]**o+(0.259021370764)*x[2]
        arg[2,4,2,1]=(-0.465880467377)*x[0]**o+(-0.840984822217)*x[0]+(0.665817382039)*x[1]**o+(-0.390932961529)*x[1]+(-0.0857939888315)*x[2]**o+(-0.712185917672)*x[2]
        arg[3,0,0,0]=(0.710281429223)*x[0]**o+(-0.723903639057)*x[0]+(-0.267455509065)*x[1]**o+(-0.731684188612)*x[1]+(-0.287442751768)*x[2]**o+(-0.266871749947)*x[2]
        arg[3,0,0,1]=(0.460723870663)*x[0]**o+(0.0502154310124)*x[0]+(-0.260241363193)*x[1]**o+(-0.516473088648)*x[1]+(-0.925984113039)*x[2]**o+(-0.0297616649018)*x[2]
        arg[3,0,1,0]=(0.523186648278)*x[0]**o+(0.937042293367)*x[0]+(0.811740916657)*x[1]**o+(0.0544805175015)*x[1]+(-0.543484279955)*x[2]**o+(0.244714198306)*x[2]
        arg[3,0,1,1]=(-0.69212630167)*x[0]**o+(0.868708276385)*x[0]+(0.473290025811)*x[1]**o+(0.137043602294)*x[1]+(-0.19291314113)*x[2]**o+(-0.755147036188)*x[2]
        arg[3,0,2,0]=(0.0472638421122)*x[0]**o+(0.403135551064)*x[0]+(0.290083188289)*x[1]**o+(0.96419060449)*x[1]+(-0.240544184077)*x[2]**o+(-0.183028108529)*x[2]
        arg[3,0,2,1]=(0.354253383908)*x[0]**o+(0.628816528774)*x[0]+(0.588194429756)*x[1]**o+(-0.852128734457)*x[1]+(-0.84774778925)*x[2]**o+(0.796310123562)*x[2]
        arg[3,1,0,0]=(0.528880495806)*x[0]**o+(-0.74176589205)*x[0]+(-0.447816938371)*x[1]**o+(0.611297564802)*x[1]+(0.742134198482)*x[2]**o+(-0.625840664202)*x[2]
        arg[3,1,0,1]=(0.419633746259)*x[0]**o+(-0.247742274578)*x[0]+(0.302328885303)*x[1]**o+(-0.629854480502)*x[1]+(0.493919301288)*x[2]**o+(0.902910119017)*x[2]
        arg[3,1,1,0]=(0.509160481656)*x[0]**o+(-0.388129364133)*x[0]+(-0.595185230598)*x[1]**o+(-0.289592836865)*x[1]+(0.615663908134)*x[2]**o+(-0.560687711511)*x[2]
        arg[3,1,1,1]=(0.647782473568)*x[0]**o+(0.935840881725)*x[0]+(0.440348212287)*x[1]**o+(-0.489797988572)*x[1]+(0.766592631787)*x[2]**o+(0.217399649246)*x[2]
        arg[3,1,2,0]=(0.64944026906)*x[0]**o+(0.487409918475)*x[0]+(-0.0322210593784)*x[1]**o+(0.281311337861)*x[1]+(-0.694252407571)*x[2]**o+(-0.659678386121)*x[2]
        arg[3,1,2,1]=(0.806357278285)*x[0]**o+(0.419430374567)*x[0]+(-0.0525796600116)*x[1]**o+(-0.556886341914)*x[1]+(-0.288209814683)*x[2]**o+(-0.309696771495)*x[2]
        arg[3,2,0,0]=(-0.300806388527)*x[0]**o+(-0.39334416908)*x[0]+(0.000711617540648)*x[1]**o+(-0.618539958736)*x[1]+(0.0574098223152)*x[2]**o+(-0.351595384168)*x[2]
        arg[3,2,0,1]=(-0.898675793892)*x[0]**o+(0.904479958277)*x[0]+(0.0781945046018)*x[1]**o+(-0.0461053392586)*x[1]+(0.844798127031)*x[2]**o+(-0.82306838709)*x[2]
        arg[3,2,1,0]=(-0.268742335093)*x[0]**o+(-0.532639944208)*x[0]+(0.0101227109551)*x[1]**o+(-0.438312865518)*x[1]+(0.104288648246)*x[2]**o+(-0.115669027311)*x[2]
        arg[3,2,1,1]=(0.639871308792)*x[0]**o+(0.260847628514)*x[0]+(-0.455670719167)*x[1]**o+(0.210619035618)*x[1]+(-0.401809068119)*x[2]**o+(-0.0625194517026)*x[2]
        arg[3,2,2,0]=(-0.713684223829)*x[0]**o+(0.214281604766)*x[0]+(-0.60152592324)*x[1]**o+(0.249807228316)*x[1]+(-0.939395078261)*x[2]**o+(-0.519744396388)*x[2]
        arg[3,2,2,1]=(-0.340588993443)*x[0]**o+(-0.0165543854997)*x[0]+(-0.366479789338)*x[1]**o+(0.316347630987)*x[1]+(-0.457055798679)*x[2]**o+(-0.132368055553)*x[2]
        arg[3,3,0,0]=(-0.745211736715)*x[0]**o+(0.0106787556884)*x[0]+(-0.448180943446)*x[1]**o+(0.694982281891)*x[1]+(0.924431258221)*x[2]**o+(-0.841633806168)*x[2]
        arg[3,3,0,1]=(0.368054364362)*x[0]**o+(0.19288251992)*x[0]+(0.565854068035)*x[1]**o+(0.747534544754)*x[1]+(0.473165789569)*x[2]**o+(-0.59399986706)*x[2]
        arg[3,3,1,0]=(-0.142707671693)*x[0]**o+(0.177095043535)*x[0]+(-0.348946846847)*x[1]**o+(-0.545430013244)*x[1]+(-0.671405653136)*x[2]**o+(-0.531699969702)*x[2]
        arg[3,3,1,1]=(0.783201055631)*x[0]**o+(0.899495042323)*x[0]+(0.117802924059)*x[1]**o+(-0.277400900172)*x[1]+(-0.351268362305)*x[2]**o+(0.52756480897)*x[2]
        arg[3,3,2,0]=(-0.866758006698)*x[0]**o+(0.293139875238)*x[0]+(-0.318394231317)*x[1]**o+(-0.783701622124)*x[1]+(-0.95137399655)*x[2]**o+(-0.474198736639)*x[2]
        arg[3,3,2,1]=(0.0428398560619)*x[0]**o+(-0.970180601102)*x[0]+(-0.753897573393)*x[1]**o+(-0.0873352699251)*x[1]+(-0.131440532574)*x[2]**o+(0.875241605077)*x[2]
        arg[3,4,0,0]=(0.32399493643)*x[0]**o+(0.67295092524)*x[0]+(-0.29221110671)*x[1]**o+(0.118066079866)*x[1]+(-0.566541154667)*x[2]**o+(-0.806367499079)*x[2]
        arg[3,4,0,1]=(0.278776399755)*x[0]**o+(0.472327826919)*x[0]+(-0.956871861949)*x[1]**o+(0.109236874792)*x[1]+(-0.494095116883)*x[2]**o+(-0.194313775518)*x[2]
        arg[3,4,1,0]=(0.650920037415)*x[0]**o+(0.977698516929)*x[0]+(-0.0730777837998)*x[1]**o+(-0.576543152032)*x[1]+(0.781481934657)*x[2]**o+(-0.389328840554)*x[2]
        arg[3,4,1,1]=(-0.798341529522)*x[0]**o+(-0.630245234712)*x[0]+(-0.444550323489)*x[1]**o+(-0.539189248295)*x[1]+(-0.904964423673)*x[2]**o+(-0.611560462796)*x[2]
        arg[3,4,2,0]=(-0.392991417685)*x[0]**o+(0.313599061567)*x[0]+(-0.215747337398)*x[1]**o+(-0.310316872476)*x[1]+(-0.269452617785)*x[2]**o+(0.684353794645)*x[2]
        arg[3,4,2,1]=(-0.586668781403)*x[0]**o+(-0.0711284937177)*x[0]+(0.408273850302)*x[1]**o+(0.761860299288)*x[1]+(0.902079242449)*x[2]**o+(0.70161017722)*x[2]
        ref[0,0,0,0]=(1.43173970815)/(o+1.)+(-0.156396435558)
        ref[0,0,0,1]=(0.459165989186)/(o+1.)+(0.194399341676)
        ref[0,0,1,0]=(0.989528642545)/(o+1.)+(-0.164206368911)
        ref[0,0,1,1]=(1.37152584676)/(o+1.)+(-0.234493053498)
        ref[0,0,2,0]=(0.422741924108)/(o+1.)+(0.0721378026728)
        ref[0,0,2,1]=(-0.574467903086)/(o+1.)+(0.167204674462)
        ref[0,1,0,0]=(1.36619629735)/(o+1.)+(-0.19947081885)
        ref[0,1,0,1]=(-1.60758660729)/(o+1.)+(0.447728236186)
        ref[0,1,1,0]=(1.5106959073)/(o+1.)+(-0.393852180331)
        ref[0,1,1,1]=(1.85678121239)/(o+1.)+(-0.529280397551)
        ref[0,1,2,0]=(-0.796337866944)/(o+1.)+(0.0812355940636)
        ref[0,1,2,1]=(-0.333114331422)/(o+1.)+(1.22614026311)
        ref[0,2,0,0]=(0.516540083621)/(o+1.)+(-0.358761510118)
        ref[0,2,0,1]=(-2.72820411942)/(o+1.)+(0.15972705639)
        ref[0,2,1,0]=(-0.440149466937)/(o+1.)+(-0.124216123318)
        ref[0,2,1,1]=(-0.137940536755)/(o+1.)+(0.13992652657)
        ref[0,2,2,0]=(0.373759361058)/(o+1.)+(0.210996140796)
        ref[0,2,2,1]=(-1.07034621798)/(o+1.)+(1.2905860125)
        ref[0,3,0,0]=(0.923495049095)/(o+1.)+(-0.897043344396)
        ref[0,3,0,1]=(-0.20180799615)/(o+1.)+(0.272652457923)
        ref[0,3,1,0]=(-0.831052678484)/(o+1.)+(0.326929849662)
        ref[0,3,1,1]=(1.41523396695)/(o+1.)+(-0.5275927117)
        ref[0,3,2,0]=(0.405715363821)/(o+1.)+(0.535330516568)
        ref[0,3,2,1]=(0.877269821733)/(o+1.)+(-0.666082812121)
        ref[0,4,0,0]=(0.884126703786)/(o+1.)+(-0.559427998262)
        ref[0,4,0,1]=(-1.36933304826)/(o+1.)+(0.263100182211)
        ref[0,4,1,0]=(0.871486955825)/(o+1.)+(0.381595437312)
        ref[0,4,1,1]=(1.21446122052)/(o+1.)+(-0.642161706726)
        ref[0,4,2,0]=(-1.82861338966)/(o+1.)+(0.507433118531)
        ref[0,4,2,1]=(0.0642039966469)/(o+1.)+(0.551348562604)
        ref[1,0,0,0]=(0.40922616303)/(o+1.)+(-0.474630105482)
        ref[1,0,0,1]=(1.63943889768)/(o+1.)+(0.285622914296)
        ref[1,0,1,0]=(0.0845961115353)/(o+1.)+(0.143628502705)
        ref[1,0,1,1]=(0.62129375936)/(o+1.)+(0.671645835511)
        ref[1,0,2,0]=(-1.2004575453)/(o+1.)+(0.701757992696)
        ref[1,0,2,1]=(0.435963491825)/(o+1.)+(0.26710538393)
        ref[1,1,0,0]=(0.738921325407)/(o+1.)+(-0.0258542444502)
        ref[1,1,0,1]=(1.95599148183)/(o+1.)+(0.652818492883)
        ref[1,1,1,0]=(0.500576647095)/(o+1.)+(0.857652485995)
        ref[1,1,1,1]=(-0.0151750762821)/(o+1.)+(0.238827857598)
        ref[1,1,2,0]=(1.26963005373)/(o+1.)+(-0.299065008888)
        ref[1,1,2,1]=(-0.82289389301)/(o+1.)+(0.14939351077)
        ref[1,2,0,0]=(-0.495951362447)/(o+1.)+(-0.532833788304)
        ref[1,2,0,1]=(-1.87785443639)/(o+1.)+(-0.521209467732)
        ref[1,2,1,0]=(-0.374085962184)/(o+1.)+(0.505583421209)
        ref[1,2,1,1]=(1.01773556605)/(o+1.)+(0.764863466171)
        ref[1,2,2,0]=(-1.54049109838)/(o+1.)+(-0.511296952595)
        ref[1,2,2,1]=(-1.06948953348)/(o+1.)+(-0.215299794811)
        ref[1,3,0,0]=(-1.20198777829)/(o+1.)+(-0.274360734317)
        ref[1,3,0,1]=(1.13519842695)/(o+1.)+(0.506840913294)
        ref[1,3,1,0]=(0.973820116369)/(o+1.)+(0.485786110977)
        ref[1,3,1,1]=(0.155263308828)/(o+1.)+(0.473436850817)
        ref[1,3,2,0]=(0.476069733748)/(o+1.)+(-0.228046397052)
        ref[1,3,2,1]=(-0.234851551487)/(o+1.)+(-0.574065011567)
        ref[1,4,0,0]=(1.95924908786)/(o+1.)+(-0.493262726539)
        ref[1,4,0,1]=(-0.448532656782)/(o+1.)+(0.393812699074)
        ref[1,4,1,0]=(0.235155577808)/(o+1.)+(0.756260258803)
        ref[1,4,1,1]=(0.0334282473144)/(o+1.)+(-0.837406551217)
        ref[1,4,2,0]=(0.568132489684)/(o+1.)+(-0.13055465798)
        ref[1,4,2,1]=(-0.380042596089)/(o+1.)+(0.243737215624)
        ref[2,0,0,0]=(1.76568035823)/(o+1.)+(0.0789834372326)
        ref[2,0,0,1]=(-0.20023919191)/(o+1.)+(0.901466917208)
        ref[2,0,1,0]=(-1.20715877778)/(o+1.)+(-0.992500439486)
        ref[2,0,1,1]=(-0.761913514624)/(o+1.)+(0.465965334128)
        ref[2,0,2,0]=(-0.944104213428)/(o+1.)+(-0.146881270459)
        ref[2,0,2,1]=(0.105283971659)/(o+1.)+(-1.30498791712)
        ref[2,1,0,0]=(-0.804721214141)/(o+1.)+(0.356736186609)
        ref[2,1,0,1]=(-0.0358563948854)/(o+1.)+(-0.24146509617)
        ref[2,1,1,0]=(-0.437556173502)/(o+1.)+(-0.387215578175)
        ref[2,1,1,1]=(-0.0786030951662)/(o+1.)+(0.654838009465)
        ref[2,1,2,0]=(-0.859200342885)/(o+1.)+(-0.112407873885)
        ref[2,1,2,1]=(-0.296679106354)/(o+1.)+(0.0491751507381)
        ref[2,2,0,0]=(0.462408531841)/(o+1.)+(-0.557975521593)
        ref[2,2,0,1]=(-0.0634225791893)/(o+1.)+(0.0468185300489)
        ref[2,2,1,0]=(1.08205349937)/(o+1.)+(0.621257561187)
        ref[2,2,1,1]=(0.88864389409)/(o+1.)+(0.849886105205)
        ref[2,2,2,0]=(2.01702912587)/(o+1.)+(0.755558425167)
        ref[2,2,2,1]=(1.41263009001)/(o+1.)+(0.331281133791)
        ref[2,3,0,0]=(1.91485193421)/(o+1.)+(0.0643042152414)
        ref[2,3,0,1]=(1.86113720031)/(o+1.)+(0.870090339646)
        ref[2,3,1,0]=(0.442533807126)/(o+1.)+(-0.690437405116)
        ref[2,3,1,1]=(1.52774754257)/(o+1.)+(-0.236667038211)
        ref[2,3,2,0]=(1.08558789233)/(o+1.)+(-0.346580362392)
        ref[2,3,2,1]=(-0.771732912166)/(o+1.)+(-0.49697903252)
        ref[2,4,0,0]=(1.0510676744)/(o+1.)+(0.18646253745)
        ref[2,4,0,1]=(-0.878875991119)/(o+1.)+(-0.271299000644)
        ref[2,4,1,0]=(-0.75026757693)/(o+1.)+(-0.168751100981)
        ref[2,4,1,1]=(0.526155063499)/(o+1.)+(1.041034953)
        ref[2,4,2,0]=(0.377792294719)/(o+1.)+(0.433767213151)
        ref[2,4,2,1]=(0.114142925831)/(o+1.)+(-0.972051850709)
        ref[3,0,0,0]=(0.15538316839)/(o+1.)+(-0.861229788808)
        ref[3,0,0,1]=(-0.725501605569)/(o+1.)+(-0.248009661269)
        ref[3,0,1,0]=(0.791443284981)/(o+1.)+(0.618118504587)
        ref[3,0,1,1]=(-0.411749416988)/(o+1.)+(0.125302421245)
        ref[3,0,2,0]=(0.0968028463245)/(o+1.)+(0.592149023513)
        ref[3,0,2,1]=(0.0947000244134)/(o+1.)+(0.286498958939)
        ref[3,1,0,0]=(0.823197755916)/(o+1.)+(-0.378154495725)
        ref[3,1,0,1]=(1.21588193285)/(o+1.)+(0.0126566819682)
        ref[3,1,1,0]=(0.529639159192)/(o+1.)+(-0.619204956255)
        ref[3,1,1,1]=(1.85472331764)/(o+1.)+(0.3317212712)
        ref[3,1,2,0]=(-0.0770331978898)/(o+1.)+(0.0545214351076)
        ref[3,1,2,1]=(0.46556780359)/(o+1.)+(-0.223576369421)
        ref[3,2,0,0]=(-0.242684948671)/(o+1.)+(-0.681739755992)
        ref[3,2,0,1]=(0.0243168377401)/(o+1.)+(0.0176531159642)
        ref[3,2,1,0]=(-0.154330975892)/(o+1.)+(-0.543310918519)
        ref[3,2,1,1]=(-0.217608478494)/(o+1.)+(0.204473606215)
        ref[3,2,2,0]=(-2.25460522533)/(o+1.)+(-0.0278277816531)
        ref[3,2,2,1]=(-1.16412458146)/(o+1.)+(0.0837125949668)
        ref[3,3,0,0]=(-0.26896142194)/(o+1.)+(-0.0679863842941)
        ref[3,3,0,1]=(1.40707422197)/(o+1.)+(0.173208598807)
        ref[3,3,1,0]=(-1.16306017168)/(o+1.)+(-0.450017469706)
        ref[3,3,1,1]=(0.549735617385)/(o+1.)+(0.57482947556)
        ref[3,3,2,0]=(-2.13652623456)/(o+1.)+(-0.482380241762)
        ref[3,3,2,1]=(-0.842498249905)/(o+1.)+(-0.0911371329748)
        ref[3,4,0,0]=(-0.534757324947)/(o+1.)+(-0.00767524698684)
        ref[3,4,0,1]=(-1.17219057908)/(o+1.)+(0.193625463097)
        ref[3,4,1,0]=(1.35932418827)/(o+1.)+(0.00591326217154)
        ref[3,4,1,1]=(-2.14785627668)/(o+1.)+(-0.890497472901)
        ref[3,4,2,0]=(-0.878191372868)/(o+1.)+(0.343817991868)
        ref[3,4,2,1]=(0.723684311348)/(o+1.)+(0.696170991395)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.729700679932)*x[0]+(-1.38329769179)*x[1]
        ref=(0.329728333821)/(o+1.)+(-0.491662672839)
      else:
        arg=(-1.75076012766)*x[0]+(-0.0862990889349)*x[1]+(0.637429232833)*x[2]
        ref=(-0.577517867896)/(o+1.)+(-0.311056057931)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(1.0296233191)*x[0]+(-0.347985432388)*x[1]
        arg[1]=(-0.182082548052)*x[0]+(0.114822539064)*x[1]
        ref[0]=(0.876900301573)/(o+1.)+(-0.0976312074308)
        ref[1]=(-1.09276229827)/(o+1.)+(0.512751144643)
      else:
        arg[0]=(0.390938934986)*x[0]+(-0.17813109236)*x[1]+(0.92611177061)*x[2]
        arg[1]=(0.169562781554)*x[0]+(-0.149625135256)*x[1]+(0.357445984273)*x[2]
        ref[0]=(1.07308269316)/(o+1.)+(0.0329184600363)
        ref[1]=(-1.07527269336)/(o+1.)+(0.726328161966)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.798668686601)*x[0]+(-0.502017718729)*x[1]
        arg[0,1]=(-0.682368031728)*x[0]+(-0.915048032379)*x[1]
        arg[0,2]=(-0.0237771508029)*x[0]+(0.177441262563)*x[1]
        arg[0,3]=(0.135600942216)*x[0]+(0.496697614771)*x[1]
        arg[0,4]=(0.495272681631)*x[0]+(-0.457599050745)*x[1]
        arg[1,0]=(-0.252415627225)*x[0]+(-0.820872546406)*x[1]
        arg[1,1]=(-0.183362696688)*x[0]+(-0.721277471492)*x[1]
        arg[1,2]=(0.401101418876)*x[0]+(0.0483185737314)*x[1]
        arg[1,3]=(0.251246583055)*x[0]+(-0.837901564153)*x[1]
        arg[1,4]=(0.2984760032)*x[0]+(0.0504569221307)*x[1]
        arg[2,0]=(1.18468310151)*x[0]+(-0.545612792222)*x[1]
        arg[2,1]=(-1.08451540544)*x[0]+(0.820927137753)*x[1]
        arg[2,2]=(0.85435600905)*x[0]+(-0.309819803422)*x[1]
        arg[2,3]=(-0.184148030844)*x[0]+(0.978315441698)*x[1]
        arg[2,4]=(1.36943188922)*x[0]+(-1.3176984733)*x[1]
        arg[3,0]=(0.42189213134)*x[0]+(1.23097633986)*x[1]
        arg[3,1]=(0.0409353425792)*x[0]+(-1.15555830367)*x[1]
        arg[3,2]=(0.557942070395)*x[0]+(-0.591659744114)*x[1]
        arg[3,3]=(-1.07432449167)*x[0]+(-0.665475086161)*x[1]
        arg[3,4]=(-1.00421972544)*x[0]+(1.16737316771)*x[1]
        ref[0,0]=(-0.400434079596)/(o+1.)+(-0.450126162867)
        ref[0,1]=(-1.35836659237)/(o+1.)+(-0.11952473587)
        ref[0,2]=(0.154775934077)/(o+1.)+(-0.000555911158727)
        ref[0,3]=(-0.177525502969)/(o+1.)+(0.404912029978)
        ref[0,4]=(1.00122501632)/(o+1.)+(-0.481775692719)
        ref[1,0]=(-1.37507083806)/(o+1.)+(0.150891332214)
        ref[1,1]=(-1.37429063854)/(o+1.)+(0.234825235182)
        ref[1,2]=(1.30069236547)/(o+1.)+(-0.425636186429)
        ref[1,3]=(-0.414658324515)/(o+1.)+(-0.0859983282917)
        ref[1,4]=(0.757097138678)/(o+1.)+(-0.204082106674)
        ref[2,0]=(1.09244584366)/(o+1.)+(-0.226687767185)
        ref[2,1]=(0.109481169277)/(o+1.)+(-0.186534718481)
        ref[2,2]=(0.587125473152)/(o+1.)+(-0.0212946337622)
        ref[2,3]=(0.00878695699842)/(o+1.)+(0.392690226928)
        ref[2,4]=(0.306810290785)/(o+1.)+(-0.127538437433)
        ref[3,0]=(0.317826050636)/(o+1.)+(0.667521210281)
        ref[3,1]=(-0.276254182573)/(o+1.)+(-0.419184389261)
        ref[3,2]=(-0.753831729065)/(o+1.)+(0.360057027673)
        ref[3,3]=(-1.40505001195)/(o+1.)+(-0.167374782943)
        ref[3,4]=(0.53112066724)/(o+1.)+(-0.183983612488)
      else:
        arg[0,0]=(1.54447818842)*x[0]+(-0.560568590939)*x[1]+(-1.33749549019)*x[2]
        arg[0,1]=(-0.535544800192)*x[0]+(-0.732513980801)*x[1]+(0.309855030569)*x[2]
        arg[0,2]=(-0.45251604551)*x[0]+(-0.0933932995351)*x[1]+(-0.465724167794)*x[2]
        arg[0,3]=(0.10344455234)*x[0]+(-0.0451254971766)*x[1]+(-1.39912042607)*x[2]
        arg[0,4]=(0.617585449939)*x[0]+(1.27357810776)*x[1]+(-0.840161322302)*x[2]
        arg[1,0]=(0.650274355535)*x[0]+(0.547805408891)*x[1]+(0.314350326062)*x[2]
        arg[1,1]=(0.697200474806)*x[0]+(0.558681734309)*x[1]+(-0.0277936766536)*x[2]
        arg[1,2]=(-0.508578328229)*x[0]+(-0.513894226666)*x[1]+(-0.528417418023)*x[2]
        arg[1,3]=(-1.50391414876)*x[0]+(0.174080260823)*x[1]+(-1.48528641514)*x[2]
        arg[1,4]=(0.859470776434)*x[0]+(-0.121348493404)*x[1]+(1.69796346547)*x[2]
        arg[2,0]=(1.00593912049)*x[0]+(0.794254020467)*x[1]+(-1.02770358461)*x[2]
        arg[2,1]=(-1.64586076188)*x[0]+(-0.83927167526)*x[1]+(-1.13088449805)*x[2]
        arg[2,2]=(0.0628777004357)*x[0]+(-1.27316091149)*x[1]+(0.417655047748)*x[2]
        arg[2,3]=(-1.43661526118)*x[0]+(0.0676042964527)*x[1]+(0.174947572442)*x[2]
        arg[2,4]=(0.504347051098)*x[0]+(-1.62976096707)*x[1]+(0.61579932957)*x[2]
        arg[3,0]=(1.05212645875)*x[0]+(-1.10517003049)*x[1]+(0.902242097595)*x[2]
        arg[3,1]=(-0.997611854104)*x[0]+(-1.53611378293)*x[1]+(0.322835276897)*x[2]
        arg[3,2]=(-0.815135194015)*x[0]+(0.357734619962)*x[1]+(0.684966069635)*x[2]
        arg[3,3]=(0.4545417685)*x[0]+(0.386262464083)*x[1]+(0.131414279708)*x[2]
        arg[3,4]=(-0.222694646001)*x[0]+(-0.470499833621)*x[1]+(-0.593290862391)*x[2]
        ref[0,0]=(-0.593803934901)/(o+1.)+(0.120109021094)
        ref[0,1]=(0.0569115377867)/(o+1.)+(-0.507557644105)
        ref[0,2]=(-0.202741590274)/(o+1.)+(-0.404445961283)
        ref[0,3]=(-0.516888777015)/(o+1.)+(-0.411956296948)
        ref[0,4]=(0.52292787417)/(o+1.)+(0.264037180616)
        ref[1,0]=(-0.102685279742)/(o+1.)+(0.807557685115)
        ref[1,1]=(1.44527264276)/(o+1.)+(-0.108592055148)
        ref[1,2]=(-0.448767718864)/(o+1.)+(-0.551061127027)
        ref[1,3]=(-1.19936004685)/(o+1.)+(-0.807880128113)
        ref[1,4]=(1.09918755605)/(o+1.)+(0.668449096223)
        ref[2,0]=(-0.873708007453)/(o+1.)+(0.823098781902)
        ref[2,1]=(-1.9697776018)/(o+1.)+(-0.823119666698)
        ref[2,2]=(0.598023292545)/(o+1.)+(-0.695325727928)
        ref[2,3]=(-0.486077421455)/(o+1.)+(-0.353992985415)
        ref[2,4]=(0.65037669171)/(o+1.)+(-0.579995639058)
        ref[3,0]=(0.115401170876)/(o+1.)+(0.366898677493)
        ref[3,1]=(-0.836209920174)/(o+1.)+(-0.68734021998)
        ref[3,2]=(0.753819223079)/(o+1.)+(-0.263126863749)
        ref[3,3]=(-0.75294945942)/(o+1.)+(0.862583985856)
        ref[3,4]=(-1.44356156537)/(o+1.)+(0.0785381116785)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.409425595197)*x[0]+(-0.992427840513)*x[1]
        arg[0,0,1]=(0.645693185569)*x[0]+(-1.38338827813)*x[1]
        arg[0,1,0]=(0.838847764041)*x[0]+(-1.01358491173)*x[1]
        arg[0,1,1]=(0.381938057912)*x[0]+(0.61979806953)*x[1]
        arg[1,0,0]=(-1.36783302742)*x[0]+(-0.132422986617)*x[1]
        arg[1,0,1]=(-0.137283745381)*x[0]+(0.633534018338)*x[1]
        arg[1,1,0]=(0.0698189930006)*x[0]+(-1.45980415637)*x[1]
        arg[1,1,1]=(-0.391725001578)*x[0]+(0.355258883756)*x[1]
        arg[2,0,0]=(-0.596091494216)*x[0]+(-0.911870026464)*x[1]
        arg[2,0,1]=(0.138007894626)*x[0]+(1.76859278751)*x[1]
        arg[2,1,0]=(-0.791803431843)*x[0]+(-0.813170959331)*x[1]
        arg[2,1,1]=(-1.05466714677)*x[0]+(-1.12822729059)*x[1]
        arg[3,0,0]=(-0.417845531677)*x[0]+(1.13689234148)*x[1]
        arg[3,0,1]=(-0.0323347813877)*x[0]+(1.14381520037)*x[1]
        arg[3,1,0]=(0.639524393892)*x[0]+(0.864696525716)*x[1]
        arg[3,1,1]=(0.593104139176)*x[0]+(0.812642260561)*x[1]
        arg[4,0,0]=(-0.0580034066457)*x[0]+(0.802938738147)*x[1]
        arg[4,0,1]=(0.722143376522)*x[0]+(0.843814429252)*x[1]
        arg[4,1,0]=(-0.574057925551)*x[0]+(-0.298232158953)*x[1]
        arg[4,1,1]=(0.422529038525)*x[0]+(0.841942812244)*x[1]
        arg[5,0,0]=(1.90272013575)*x[0]+(-1.84811498267)*x[1]
        arg[5,0,1]=(-1.52156314991)*x[0]+(-0.797833417606)*x[1]
        arg[5,1,0]=(0.723748480291)*x[0]+(0.8667659268)*x[1]
        arg[5,1,1]=(-0.70289499551)*x[0]+(-0.726595802952)*x[1]
        ref[0,0,0]=(-1.1719312491)/(o+1.)+(-0.114961093303)
        ref[0,0,1]=(0.229993657297)/(o+1.)+(-0.483844374927)
        ref[0,1,0]=(-0.630442924796)/(o+1.)+(0.227852888553)
        ref[0,1,1]=(-0.806392286115)/(o+1.)+(0.904064206778)
        ref[1,0,0]=(-1.38312106397)/(o+1.)+(-0.0585674750338)
        ref[1,0,1]=(0.595552955095)/(o+1.)+(-0.049651341069)
        ref[1,1,0]=(-1.63606241234)/(o+1.)+(0.123038624483)
        ref[1,1,1]=(-0.340463713286)/(o+1.)+(0.151998797732)
        ref[2,0,0]=(-1.31925158304)/(o+1.)+(-0.094354968822)
        ref[2,0,1]=(0.699277735512)/(o+1.)+(0.60366147331)
        ref[2,1,0]=(-0.219878898763)/(o+1.)+(-0.692547746205)
        ref[2,1,1]=(-1.46750249607)/(o+1.)+(-0.357695970647)
        ref[3,0,0]=(0.71941698906)/(o+1.)+(-0.000185089626866)
        ref[3,0,1]=(-0.526993313502)/(o+1.)+(0.819236866243)
        ref[3,1,0]=(1.51671320209)/(o+1.)+(-0.00624614123899)
        ref[3,1,1]=(1.05185683613)/(o+1.)+(0.176944781803)
        ref[4,0,0]=(1.20432711487)/(o+1.)+(-0.229695891683)
        ref[4,0,1]=(1.26863112516)/(o+1.)+(0.14866334031)
        ref[4,1,0]=(0.34377834331)/(o+1.)+(-0.608034213907)
        ref[4,1,1]=(-0.189568721058)/(o+1.)+(0.727020285914)
        ref[5,0,0]=(0.116981035747)/(o+1.)+(-0.0311879413343)
        ref[5,0,1]=(-0.638728138162)/(o+1.)+(-0.840334214679)
        ref[5,1,0]=(0.891634589919)/(o+1.)+(0.349439908586)
        ref[5,1,1]=(-0.552480077628)/(o+1.)+(-0.438505360418)
      else:
        arg[0,0,0]=(-0.662295604848)*x[0]+(1.04606412328)*x[1]+(-0.376008345205)*x[2]
        arg[0,0,1]=(1.25030158303)*x[0]+(-0.0735167236581)*x[1]+(1.31623868156)*x[2]
        arg[0,1,0]=(0.717953620717)*x[0]+(-0.700464709991)*x[1]+(0.0345381771501)*x[2]
        arg[0,1,1]=(0.225944321569)*x[0]+(-1.09510173632)*x[1]+(-0.371235420166)*x[2]
        arg[1,0,0]=(-0.0755865552996)*x[0]+(-0.168214695387)*x[1]+(-0.437427629408)*x[2]
        arg[1,0,1]=(-1.52674685128)*x[0]+(-0.408158359188)*x[1]+(0.750091649824)*x[2]
        arg[1,1,0]=(-0.572400948498)*x[0]+(-0.802826119419)*x[1]+(-1.0233358643)*x[2]
        arg[1,1,1]=(0.623156259061)*x[0]+(0.378774331134)*x[1]+(0.0419766309042)*x[2]
        arg[2,0,0]=(0.443271002813)*x[0]+(-0.930092172267)*x[1]+(-0.20199150178)*x[2]
        arg[2,0,1]=(0.408614848881)*x[0]+(-1.3007040171)*x[1]+(-0.872615408174)*x[2]
        arg[2,1,0]=(-0.174910742406)*x[0]+(1.64872582781)*x[1]+(0.701172036891)*x[2]
        arg[2,1,1]=(0.417145974895)*x[0]+(0.428983721984)*x[1]+(-0.000787616769604)*x[2]
        arg[3,0,0]=(0.655644872855)*x[0]+(-0.268612479186)*x[1]+(0.11779790898)*x[2]
        arg[3,0,1]=(-0.244670332245)*x[0]+(-0.857502890419)*x[1]+(0.111792341131)*x[2]
        arg[3,1,0]=(0.131303542183)*x[0]+(0.231230724482)*x[1]+(0.116997126513)*x[2]
        arg[3,1,1]=(0.725434968911)*x[0]+(1.00206695553)*x[1]+(-0.319715936784)*x[2]
        arg[4,0,0]=(1.30949067662)*x[0]+(-1.14809917605)*x[1]+(0.00529778192088)*x[2]
        arg[4,0,1]=(0.969626585498)*x[0]+(-0.517908035476)*x[1]+(1.0467168328)*x[2]
        arg[4,1,0]=(0.362824899886)*x[0]+(0.890586602899)*x[1]+(-0.135163953653)*x[2]
        arg[4,1,1]=(-0.10353577028)*x[0]+(-0.0997924233294)*x[1]+(0.308973949876)*x[2]
        arg[5,0,0]=(-1.37242244095)*x[0]+(-0.746928231432)*x[1]+(-1.37385481504)*x[2]
        arg[5,0,1]=(0.231730881472)*x[0]+(-0.250265629505)*x[1]+(-0.903599685792)*x[2]
        arg[5,1,0]=(-0.813783753893)*x[0]+(0.41081474894)*x[1]+(-0.541868665288)*x[2]
        arg[5,1,1]=(0.19296641571)*x[0]+(1.8056250646)*x[1]+(0.283675443031)*x[2]
        ref[0,0,0]=(0.859538687531)/(o+1.)+(-0.425889257151)
        ref[0,0,1]=(-0.0674840316504)/(o+1.)+(1.28025378629)
        ref[0,1,0]=(-0.76922088579)/(o+1.)+(0.410623986833)
        ref[0,1,1]=(-0.582981712778)/(o+1.)+(-0.328705561068)
        ref[1,0,0]=(1.25605113432)/(o+1.)+(-0.968640007206)
        ref[1,0,1]=(-1.59448994908)/(o+1.)+(0.204838194218)
        ref[1,1,0]=(-2.07497671334)/(o+1.)+(-0.161793109438)
        ref[1,1,1]=(0.199659680372)/(o+1.)+(0.422123770364)
        ref[2,0,0]=(-1.86970457224)/(o+1.)+(0.590445950504)
        ref[2,0,1]=(-1.09505237157)/(o+1.)+(-0.334826102411)
        ref[2,1,0]=(0.395146564722)/(o+1.)+(0.889920278785)
        ref[2,1,1]=(1.13315474193)/(o+1.)+(-0.143906330909)
        ref[3,0,0]=(-0.804398097072)/(o+1.)+(0.65461419986)
        ref[3,0,1]=(-0.829074327243)/(o+1.)+(-0.0806532771448)
        ref[3,1,0]=(0.126278821088)/(o+1.)+(0.176626286045)
        ref[3,1,1]=(-0.498753315826)/(o+1.)+(0.953269651743)
        ref[4,0,0]=(0.33849802718)/(o+1.)+(-0.0859043723417)
        ref[4,0,1]=(0.153172891455)/(o+1.)+(0.672631245684)
        ref[4,1,0]=(1.42803542207)/(o+1.)+(-0.154893936468)
        ref[4,1,1]=(0.605555698814)/(o+1.)+(-0.249954971274)
        ref[5,0,0]=(-1.21823919194)/(o+1.)+(-1.13748314774)
        ref[5,0,1]=(-1.13684484437)/(o+1.)+(0.107355205274)
        ref[5,1,0]=(-1.21991007491)/(o+1.)+(0.137536202333)
        ref[5,1,1]=(2.05825898341)/(o+1.)+(0.112003969962)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.2631296794)*x[0]+(-0.130117190301)*x[1]
        arg[0,0,0,1]=(0.635732952218)*x[0]+(-0.647555274619)*x[1]
        arg[0,0,1,0]=(1.06301179339)*x[0]+(-0.227813591181)*x[1]
        arg[0,0,1,1]=(-0.519979711376)*x[0]+(1.04747480856)*x[1]
        arg[0,0,2,0]=(0.0232818794686)*x[0]+(-1.14799687633)*x[1]
        arg[0,0,2,1]=(0.359605391616)*x[0]+(-1.32942998109)*x[1]
        arg[0,1,0,0]=(-0.555284020205)*x[0]+(0.11676202509)*x[1]
        arg[0,1,0,1]=(-1.83479976366)*x[0]+(0.413168061643)*x[1]
        arg[0,1,1,0]=(0.239736306037)*x[0]+(0.680002261596)*x[1]
        arg[0,1,1,1]=(0.0555978436642)*x[0]+(1.28910533087)*x[1]
        arg[0,1,2,0]=(0.732696470486)*x[0]+(0.538837927425)*x[1]
        arg[0,1,2,1]=(0.528018376979)*x[0]+(1.20428218411)*x[1]
        arg[0,2,0,0]=(0.863803666388)*x[0]+(0.187082373754)*x[1]
        arg[0,2,0,1]=(-1.41490739709)*x[0]+(-0.741935114995)*x[1]
        arg[0,2,1,0]=(0.0501211708116)*x[0]+(-0.370371645307)*x[1]
        arg[0,2,1,1]=(-0.0353423795999)*x[0]+(1.00911245551)*x[1]
        arg[0,2,2,0]=(1.79997397117)*x[0]+(0.240103369892)*x[1]
        arg[0,2,2,1]=(0.926052414873)*x[0]+(0.00914357056458)*x[1]
        arg[0,3,0,0]=(-1.38911305622)*x[0]+(-1.29285413766)*x[1]
        arg[0,3,0,1]=(0.464908270265)*x[0]+(-0.354824878801)*x[1]
        arg[0,3,1,0]=(0.17477491298)*x[0]+(0.882807721541)*x[1]
        arg[0,3,1,1]=(0.633177386897)*x[0]+(-0.312628707154)*x[1]
        arg[0,3,2,0]=(0.0841289229919)*x[0]+(-0.244085514309)*x[1]
        arg[0,3,2,1]=(0.198631529554)*x[0]+(-0.324387433152)*x[1]
        arg[0,4,0,0]=(-1.41503119851)*x[0]+(-0.760818812719)*x[1]
        arg[0,4,0,1]=(-0.0326236565753)*x[0]+(0.835716629913)*x[1]
        arg[0,4,1,0]=(1.84482629546)*x[0]+(-0.341557269651)*x[1]
        arg[0,4,1,1]=(-0.380600663728)*x[0]+(-0.295663906776)*x[1]
        arg[0,4,2,0]=(0.411308057157)*x[0]+(1.0109767907)*x[1]
        arg[0,4,2,1]=(-1.10494637617)*x[0]+(0.401374247774)*x[1]
        arg[1,0,0,0]=(0.437019391027)*x[0]+(0.0507256248679)*x[1]
        arg[1,0,0,1]=(-0.399457976892)*x[0]+(1.60118981524)*x[1]
        arg[1,0,1,0]=(-1.51289007675)*x[0]+(0.821406477166)*x[1]
        arg[1,0,1,1]=(0.405770274494)*x[0]+(0.637591733337)*x[1]
        arg[1,0,2,0]=(-1.4945593715)*x[0]+(1.0919176668)*x[1]
        arg[1,0,2,1]=(-0.482868528044)*x[0]+(0.578004764959)*x[1]
        arg[1,1,0,0]=(-1.44298476434)*x[0]+(0.197151370973)*x[1]
        arg[1,1,0,1]=(1.38517677391)*x[0]+(0.915352009234)*x[1]
        arg[1,1,1,0]=(1.69025511761)*x[0]+(0.0203268727501)*x[1]
        arg[1,1,1,1]=(0.615392847773)*x[0]+(-0.301873832451)*x[1]
        arg[1,1,2,0]=(0.3081860115)*x[0]+(-0.495230532585)*x[1]
        arg[1,1,2,1]=(-0.0258752746808)*x[0]+(0.441629273958)*x[1]
        arg[1,2,0,0]=(-0.947577742489)*x[0]+(0.218632343222)*x[1]
        arg[1,2,0,1]=(0.520029690788)*x[0]+(0.403806873912)*x[1]
        arg[1,2,1,0]=(1.32820278922)*x[0]+(-1.68314905373)*x[1]
        arg[1,2,1,1]=(1.30409876916)*x[0]+(-0.721819171819)*x[1]
        arg[1,2,2,0]=(1.37872038109)*x[0]+(0.0725139769178)*x[1]
        arg[1,2,2,1]=(0.35593830057)*x[0]+(0.28751298465)*x[1]
        arg[1,3,0,0]=(-1.17950311406)*x[0]+(-0.232819397686)*x[1]
        arg[1,3,0,1]=(0.17340184781)*x[0]+(-0.43329852484)*x[1]
        arg[1,3,1,0]=(-1.71125828064)*x[0]+(0.957284609667)*x[1]
        arg[1,3,1,1]=(-0.429043922461)*x[0]+(0.15749154348)*x[1]
        arg[1,3,2,0]=(-1.42626208855)*x[0]+(-1.12812746347)*x[1]
        arg[1,3,2,1]=(0.490516645689)*x[0]+(0.559039786338)*x[1]
        arg[1,4,0,0]=(0.436395831147)*x[0]+(0.382691724897)*x[1]
        arg[1,4,0,1]=(1.27903773839)*x[0]+(-0.0642376846324)*x[1]
        arg[1,4,1,0]=(0.503159792753)*x[0]+(0.392973700403)*x[1]
        arg[1,4,1,1]=(0.0881500114326)*x[0]+(0.834121935361)*x[1]
        arg[1,4,2,0]=(-0.620745962116)*x[0]+(-0.0770814791601)*x[1]
        arg[1,4,2,1]=(0.0231684675117)*x[0]+(-1.33994827808)*x[1]
        arg[2,0,0,0]=(-0.972311570714)*x[0]+(-0.235293375636)*x[1]
        arg[2,0,0,1]=(0.21633355466)*x[0]+(0.11325401613)*x[1]
        arg[2,0,1,0]=(0.922986727965)*x[0]+(-1.00516246282)*x[1]
        arg[2,0,1,1]=(-0.740778981838)*x[0]+(0.616728971608)*x[1]
        arg[2,0,2,0]=(1.73201175195)*x[0]+(-1.1634325209)*x[1]
        arg[2,0,2,1]=(1.15829744421)*x[0]+(-1.87219685708)*x[1]
        arg[2,1,0,0]=(-0.458901790255)*x[0]+(1.13812715713)*x[1]
        arg[2,1,0,1]=(1.03315725208)*x[0]+(-0.112487270341)*x[1]
        arg[2,1,1,0]=(-1.7070220527)*x[0]+(0.126058081856)*x[1]
        arg[2,1,1,1]=(0.863301459388)*x[0]+(0.226670498194)*x[1]
        arg[2,1,2,0]=(-1.0012992544)*x[0]+(-0.195604893926)*x[1]
        arg[2,1,2,1]=(-0.128048233768)*x[0]+(0.827383267773)*x[1]
        arg[2,2,0,0]=(-0.0876478859699)*x[0]+(1.69600082596)*x[1]
        arg[2,2,0,1]=(-0.616278041482)*x[0]+(-0.995548774612)*x[1]
        arg[2,2,1,0]=(0.780699602657)*x[0]+(-0.721498796198)*x[1]
        arg[2,2,1,1]=(-1.17591990575)*x[0]+(-0.66080847955)*x[1]
        arg[2,2,2,0]=(1.71259481962)*x[0]+(-1.12456334234)*x[1]
        arg[2,2,2,1]=(-0.721345747555)*x[0]+(-1.28222132706)*x[1]
        arg[2,3,0,0]=(-1.25971324146)*x[0]+(-0.00559590465243)*x[1]
        arg[2,3,0,1]=(-0.49234245798)*x[0]+(0.30900436328)*x[1]
        arg[2,3,1,0]=(1.3068919966)*x[0]+(-0.891484126193)*x[1]
        arg[2,3,1,1]=(0.319277203381)*x[0]+(-0.535113401495)*x[1]
        arg[2,3,2,0]=(0.529480518887)*x[0]+(0.00452686996008)*x[1]
        arg[2,3,2,1]=(-1.0978905592)*x[0]+(-0.421513714335)*x[1]
        arg[2,4,0,0]=(-1.27711509307)*x[0]+(-1.42208296178)*x[1]
        arg[2,4,0,1]=(-0.621292314868)*x[0]+(-1.10962518414)*x[1]
        arg[2,4,1,0]=(-0.183478331213)*x[0]+(1.0775900659)*x[1]
        arg[2,4,1,1]=(1.07354953663)*x[0]+(0.869765287054)*x[1]
        arg[2,4,2,0]=(-0.696050139647)*x[0]+(-0.120769338549)*x[1]
        arg[2,4,2,1]=(-0.895597167617)*x[0]+(1.41025466606)*x[1]
        arg[3,0,0,0]=(1.15262876611)*x[0]+(-0.505428168285)*x[1]
        arg[3,0,0,1]=(-0.277585299138)*x[0]+(0.352683243558)*x[1]
        arg[3,0,1,0]=(0.218694567952)*x[0]+(-0.170325565499)*x[1]
        arg[3,0,1,1]=(0.297325594036)*x[0]+(0.459308129811)*x[1]
        arg[3,0,2,0]=(-0.572509624935)*x[0]+(0.756188981433)*x[1]
        arg[3,0,2,1]=(0.226462444009)*x[0]+(-1.42100784266)*x[1]
        arg[3,1,0,0]=(-0.28388824767)*x[0]+(-0.306064847514)*x[1]
        arg[3,1,0,1]=(-0.805831316478)*x[0]+(0.756933495912)*x[1]
        arg[3,1,1,0]=(1.37344361233)*x[0]+(-0.377600924857)*x[1]
        arg[3,1,1,1]=(-0.0725302533884)*x[0]+(-0.425074450201)*x[1]
        arg[3,1,2,0]=(-0.066659431931)*x[0]+(1.39252394891)*x[1]
        arg[3,1,2,1]=(0.593440019224)*x[0]+(-0.945629564385)*x[1]
        arg[3,2,0,0]=(0.101850471598)*x[0]+(0.0552356415787)*x[1]
        arg[3,2,0,1]=(-0.407603005372)*x[0]+(-0.849371670514)*x[1]
        arg[3,2,1,0]=(0.435403192155)*x[0]+(0.213673097377)*x[1]
        arg[3,2,1,1]=(0.424081872219)*x[0]+(-0.0800529063085)*x[1]
        arg[3,2,2,0]=(0.189428134309)*x[0]+(-0.205813522374)*x[1]
        arg[3,2,2,1]=(-0.51449771505)*x[0]+(0.0570738169919)*x[1]
        arg[3,3,0,0]=(1.05431625119)*x[0]+(-0.0583139615539)*x[1]
        arg[3,3,0,1]=(0.0292261008496)*x[0]+(-0.569482715214)*x[1]
        arg[3,3,1,0]=(-0.155861158232)*x[0]+(0.628422800294)*x[1]
        arg[3,3,1,1]=(-1.24050157391)*x[0]+(-1.83949885089)*x[1]
        arg[3,3,2,0]=(-0.140709404689)*x[0]+(0.975529832713)*x[1]
        arg[3,3,2,1]=(-0.290824210426)*x[0]+(-0.198373932879)*x[1]
        arg[3,4,0,0]=(0.627236720719)*x[0]+(-0.306676521516)*x[1]
        arg[3,4,0,1]=(-0.964303098626)*x[0]+(-0.701698363587)*x[1]
        arg[3,4,1,0]=(-0.0447860906704)*x[0]+(0.324659175162)*x[1]
        arg[3,4,1,1]=(1.28273443124)*x[0]+(-0.764329261143)*x[1]
        arg[3,4,2,0]=(-0.023232709502)*x[0]+(-0.124231212653)*x[1]
        arg[3,4,2,1]=(0.510754136991)*x[0]+(0.570492207442)*x[1]
        ref[0,0,0,0]=(0.648764087616)/(o+1.)+(-0.521005478658)
        ref[0,0,0,1]=(-0.0127476670646)/(o+1.)+(0.000462672331599)
        ref[0,0,1,0]=(0.258601710827)/(o+1.)+(0.288298245692)
        ref[0,0,1,1]=(0.0577415806545)/(o+1.)+(0.234876758266)
        ref[0,0,2,0]=(-0.91129780757)/(o+1.)+(-0.106708594644)
        ref[0,0,2,1]=(0.0841946322525)/(o+1.)+(-0.527009610865)
        ref[0,1,0,0]=(0.111101563297)/(o+1.)+(-0.274811779206)
        ref[0,1,0,1]=(-0.504062576651)/(o+1.)+(-0.458784562683)
        ref[0,1,1,0]=(0.65560969277)/(o+1.)+(0.132064437431)
        ref[0,1,1,1]=(0.0866382135955)/(o+1.)+(0.629032480469)
        ref[0,1,2,0]=(1.36074036394)/(o+1.)+(-0.0446029830125)
        ref[0,1,2,1]=(1.16071753175)/(o+1.)+(0.285791514672)
        ref[0,2,0,0]=(0.693065154772)/(o+1.)+(0.178910442685)
        ref[0,2,0,1]=(-1.41209835364)/(o+1.)+(-0.372372079222)
        ref[0,2,1,0]=(-0.542625642859)/(o+1.)+(0.111187584182)
        ref[0,2,1,1]=(0.628714449405)/(o+1.)+(0.172527813253)
        ref[0,2,2,0]=(1.62887921583)/(o+1.)+(0.205599062617)
        ref[0,2,2,1]=(1.0640301541)/(o+1.)+(-0.0644170843301)
        ref[0,3,0,0]=(-1.09870552313)/(o+1.)+(-0.791630835375)
        ref[0,3,0,1]=(0.260706656546)/(o+1.)+(-0.0753116325413)
        ref[0,3,1,0]=(-0.0543842358149)/(o+1.)+(0.555983435168)
        ref[0,3,1,1]=(0.0369075211917)/(o+1.)+(0.141820579276)
        ref[0,3,2,0]=(-0.784250881508)/(o+1.)+(0.312147145095)
        ref[0,3,2,1]=(-0.178450092422)/(o+1.)+(0.0263470944113)
        ref[0,4,0,0]=(-0.993852999635)/(o+1.)+(-0.590998505799)
        ref[0,4,0,1]=(-0.0449039600312)/(o+1.)+(0.423998466684)
        ref[0,4,1,0]=(0.583535069477)/(o+1.)+(0.459866978166)
        ref[0,4,1,1]=(-0.251020886878)/(o+1.)+(-0.212621841813)
        ref[0,4,2,0]=(0.408688561613)/(o+1.)+(0.506798143121)
        ref[0,4,2,1]=(0.537090447747)/(o+1.)+(-0.620331288072)
        ref[1,0,0,0]=(0.37428496839)/(o+1.)+(0.0567300237522)
        ref[1,0,0,1]=(0.777672833448)/(o+1.)+(0.212029502451)
        ref[1,0,1,0]=(-0.745422608709)/(o+1.)+(0.026969504563)
        ref[1,0,1,1]=(0.466434983875)/(o+1.)+(0.288463511978)
        ref[1,0,2,0]=(0.237457788829)/(o+1.)+(-0.320049746767)
        ref[1,0,2,1]=(0.346974181328)/(o+1.)+(-0.125918972207)
        ref[1,1,0,0]=(-0.823737519034)/(o+1.)+(-0.211047937165)
        ref[1,1,0,1]=(1.00324369541)/(o+1.)+(0.648642543871)
        ref[1,1,1,0]=(0.674013158596)/(o+1.)+(0.518284415882)
        ref[1,1,1,1]=(0.403142463029)/(o+1.)+(-0.0448117238535)
        ref[1,1,2,0]=(0.0399328921703)/(o+1.)+(-0.113488706628)
        ref[1,1,2,1]=(0.111446633645)/(o+1.)+(0.152153682816)
        ref[1,2,0,0]=(-0.481646721692)/(o+1.)+(-0.123649338787)
        ref[1,2,0,1]=(0.857170000026)/(o+1.)+(0.033333282337)
        ref[1,2,1,0]=(-0.445362398126)/(o+1.)+(0.0452080668082)
        ref[1,2,1,1]=(-0.256201827386)/(o+1.)+(0.419240712362)
        ref[1,2,2,0]=(1.02991996311)/(o+1.)+(0.210657197448)
        ref[1,2,2,1]=(0.355299316895)/(o+1.)+(0.144075984162)
        ref[1,3,0,0]=(-0.679726322154)/(o+1.)+(-0.366298094796)
        ref[1,3,0,1]=(-0.67885027329)/(o+1.)+(0.20947679813)
        ref[1,3,1,0]=(-0.163081485233)/(o+1.)+(-0.295446092868)
        ref[1,3,1,1]=(-0.035768623673)/(o+1.)+(-0.117891877654)
        ref[1,3,2,0]=(-0.639368943038)/(o+1.)+(-0.957510304493)
        ref[1,3,2,1]=(0.160615066001)/(o+1.)+(0.444470683013)
        ref[1,4,0,0]=(1.0649487216)/(o+1.)+(-0.122930582777)
        ref[1,4,0,1]=(0.22766574122)/(o+1.)+(0.493567156271)
        ref[1,4,1,0]=(-0.481296537412)/(o+1.)+(0.688715015285)
        ref[1,4,1,1]=(-0.195268237132)/(o+1.)+(0.558770091963)
        ref[1,4,2,0]=(0.180618258117)/(o+1.)+(-0.439222849696)
        ref[1,4,2,1]=(0.0261012984826)/(o+1.)+(-0.671440554524)
        ref[2,0,0,0]=(-0.301101739758)/(o+1.)+(-0.453251603296)
        ref[2,0,0,1]=(0.00878006761908)/(o+1.)+(0.160403751586)
        ref[2,0,1,0]=(-0.80011300584)/(o+1.)+(0.35896863549)
        ref[2,0,1,1]=(0.00248831604776)/(o+1.)+(-0.0632691631389)
        ref[2,0,2,0]=(-0.0569220251136)/(o+1.)+(0.312750628078)
        ref[2,0,2,1]=(-0.440109525096)/(o+1.)+(-0.136894943888)
        ref[2,1,0,0]=(-0.465652661854)/(o+1.)+(0.572439014362)
        ref[2,1,0,1]=(0.0522958647723)/(o+1.)+(0.434187058485)
        ref[2,1,1,0]=(0.00194260180185)/(o+1.)+(-0.791453286324)
        ref[2,1,1,1]=(0.981996325082)/(o+1.)+(0.0539878162501)
        ref[2,1,2,0]=(-0.243334369882)/(o+1.)+(-0.476784889223)
        ref[2,1,2,1]=(1.53208640175)/(o+1.)+(-0.416375683873)
        ref[2,2,0,0]=(1.27849139532)/(o+1.)+(0.164930772337)
        ref[2,2,0,1]=(-1.4813889023)/(o+1.)+(-0.0652189568992)
        ref[2,2,1,0]=(-0.728282325831)/(o+1.)+(0.393741566145)
        ref[2,2,1,1]=(-0.537765426582)/(o+1.)+(-0.649481479359)
        ref[2,2,2,0]=(-0.266224153785)/(o+1.)+(0.427127815533)
        ref[2,2,2,1]=(-0.862349098277)/(o+1.)+(-0.570608988167)
        ref[2,3,0,0]=(-1.09340368619)/(o+1.)+(-0.0859527299593)
        ref[2,3,0,1]=(1.02842400947)/(o+1.)+(-0.605881052086)
        ref[2,3,1,0]=(0.0131810342609)/(o+1.)+(0.20111341807)
        ref[2,3,1,1]=(-1.11244447248)/(o+1.)+(0.448304137186)
        ref[2,3,2,0]=(-0.978348257673)/(o+1.)+(0.75617782326)
        ref[2,3,2,1]=(-0.140801959181)/(o+1.)+(-0.689301157175)
        ref[2,4,0,0]=(-0.733659962936)/(o+1.)+(-0.982769045955)
        ref[2,4,0,1]=(-1.6278906672)/(o+1.)+(-0.0515134159034)
        ref[2,4,1,0]=(1.52117858199)/(o+1.)+(-0.313533423654)
        ref[2,4,1,1]=(0.621221455007)/(o+1.)+(0.661046684337)
        ref[2,4,2,0]=(-0.157580028393)/(o+1.)+(-0.329619724901)
        ref[2,4,2,1]=(0.52303412622)/(o+1.)+(-0.00418831388809)
        ref[3,0,0,0]=(0.337890126112)/(o+1.)+(0.154655235856)
        ref[3,0,0,1]=(-0.100490423466)/(o+1.)+(0.0877941839427)
        ref[3,0,1,0]=(0.37549944857)/(o+1.)+(-0.163565223059)
        ref[3,0,1,1]=(-0.3212149071)/(o+1.)+(0.538924315474)
        ref[3,0,2,0]=(-0.359815373231)/(o+1.)+(0.271747364864)
        ref[3,0,2,1]=(-0.890575856923)/(o+1.)+(-0.151984770865)
        ref[3,1,0,0]=(-1.10384271628)/(o+1.)+(0.256944810546)
        ref[3,1,0,1]=(0.647425296535)/(o+1.)+(-0.348161558551)
        ref[3,1,1,0]=(1.41806770884)/(o+1.)+(-0.211112510685)
        ref[3,1,1,1]=(0.106528838317)/(o+1.)+(-0.302066770953)
        ref[3,1,2,0]=(0.917213295746)/(o+1.)+(0.204325610614)
        ref[3,1,2,1]=(-0.24738513406)/(o+1.)+(-0.0524022055505)
        ref[3,2,0,0]=(-0.0537683888373)/(o+1.)+(0.105427251007)
        ref[3,2,0,1]=(-1.24647336894)/(o+1.)+(-0.00525065347416)
        ref[3,2,1,0]=(0.187132797735)/(o+1.)+(0.230971745898)
        ref[3,2,1,1]=(0.492750417331)/(o+1.)+(-0.0743607257106)
        ref[3,2,2,0]=(0.062835017839)/(o+1.)+(-0.0396102029518)
        ref[3,2,2,1]=(-0.967272854749)/(o+1.)+(0.254924478345)
        ref[3,3,0,0]=(1.41291903854)/(o+1.)+(-0.208458374448)
        ref[3,3,0,1]=(-0.740922121718)/(o+1.)+(0.100332753677)
        ref[3,3,1,0]=(1.2839228132)/(o+1.)+(-0.405680585568)
        ref[3,3,1,1]=(-1.69516258968)/(o+1.)+(-0.69241891756)
        ref[3,3,2,0]=(0.216594952745)/(o+1.)+(0.309112737639)
        ref[3,3,2,1]=(-0.634806027047)/(o+1.)+(0.0728039418712)
        ref[3,4,0,0]=(-0.310491357088)/(o+1.)+(0.315525778146)
        ref[3,4,0,1]=(-0.751915044154)/(o+1.)+(-0.457043209029)
        ref[3,4,1,0]=(-0.339963591836)/(o+1.)+(0.309918338163)
        ref[3,4,1,1]=(0.0720250143477)/(o+1.)+(0.223190077872)
        ref[3,4,2,0]=(-0.0508554115866)/(o+1.)+(-0.0483042552841)
        ref[3,4,2,1]=(0.754037720807)/(o+1.)+(0.163604311813)
      else:
        arg[0,0,0,0]=(-1.24350959032)*x[0]+(-0.580181370761)*x[1]+(-0.174543112173)*x[2]
        arg[0,0,0,1]=(1.43450421585)*x[0]+(0.717062915298)*x[1]+(-0.91866562143)*x[2]
        arg[0,0,1,0]=(0.0188823859409)*x[0]+(1.24568141807)*x[1]+(0.0842604494481)*x[2]
        arg[0,0,1,1]=(0.871591276766)*x[0]+(0.653031319605)*x[1]+(0.037714556611)*x[2]
        arg[0,0,2,0]=(1.60473786555)*x[0]+(1.47509195705)*x[1]+(-0.103485610136)*x[2]
        arg[0,0,2,1]=(1.68030531336)*x[0]+(0.551201281257)*x[1]+(1.26951824468)*x[2]
        arg[0,1,0,0]=(0.491638014782)*x[0]+(1.26264219762)*x[1]+(-0.921297094204)*x[2]
        arg[0,1,0,1]=(0.20108233628)*x[0]+(-1.36096031252)*x[1]+(-0.878077046379)*x[2]
        arg[0,1,1,0]=(0.900775475455)*x[0]+(0.193423726606)*x[1]+(-0.139955126775)*x[2]
        arg[0,1,1,1]=(-0.0354326925578)*x[0]+(-0.464646718718)*x[1]+(1.49240476583)*x[2]
        arg[0,1,2,0]=(-1.40991971664)*x[0]+(-1.7747715584)*x[1]+(-0.96387650203)*x[2]
        arg[0,1,2,1]=(-0.476443682267)*x[0]+(0.55076135194)*x[1]+(0.190289559512)*x[2]
        arg[0,2,0,0]=(0.535490315669)*x[0]+(0.566915366497)*x[1]+(-0.41513643529)*x[2]
        arg[0,2,0,1]=(-0.18986562478)*x[0]+(-0.62347050233)*x[1]+(1.18004907973)*x[2]
        arg[0,2,1,0]=(-0.288440432828)*x[0]+(0.0375642125411)*x[1]+(0.574653643212)*x[2]
        arg[0,2,1,1]=(1.03327739612)*x[0]+(-0.0443874572943)*x[1]+(0.972142765323)*x[2]
        arg[0,2,2,0]=(-1.27131901306)*x[0]+(0.619179457229)*x[1]+(-0.436029340417)*x[2]
        arg[0,2,2,1]=(0.0573099137019)*x[0]+(0.211373076629)*x[1]+(0.859991510967)*x[2]
        arg[0,3,0,0]=(0.650974277613)*x[0]+(-0.926352792387)*x[1]+(-1.84479828618)*x[2]
        arg[0,3,0,1]=(-0.293748564815)*x[0]+(0.53683587198)*x[1]+(1.51147453744)*x[2]
        arg[0,3,1,0]=(0.837889627485)*x[0]+(0.62568225238)*x[1]+(0.365321068424)*x[2]
        arg[0,3,1,1]=(-0.671841278343)*x[0]+(-0.00822894090357)*x[1]+(-1.35180265459)*x[2]
        arg[0,3,2,0]=(0.823443588183)*x[0]+(-0.366903894987)*x[1]+(-0.878402795844)*x[2]
        arg[0,3,2,1]=(-0.856292830088)*x[0]+(-0.134584368493)*x[1]+(-0.299126849425)*x[2]
        arg[0,4,0,0]=(0.62792423481)*x[0]+(0.781414088068)*x[1]+(-1.12299220031)*x[2]
        arg[0,4,0,1]=(-0.00743646774917)*x[0]+(0.52784004028)*x[1]+(-1.44284109205)*x[2]
        arg[0,4,1,0]=(-0.855519018884)*x[0]+(-0.222306698876)*x[1]+(0.0502517275952)*x[2]
        arg[0,4,1,1]=(-0.311060622512)*x[0]+(0.321283094479)*x[1]+(-0.484147152851)*x[2]
        arg[0,4,2,0]=(-1.43220161062)*x[0]+(-0.27822988291)*x[1]+(-0.692367106541)*x[2]
        arg[0,4,2,1]=(1.3956686934)*x[0]+(-1.26426333906)*x[1]+(-1.09343602989)*x[2]
        arg[1,0,0,0]=(0.122525927098)*x[0]+(-0.484009414389)*x[1]+(-0.822036495419)*x[2]
        arg[1,0,0,1]=(-0.276748173334)*x[0]+(0.35307031133)*x[1]+(-1.23466402387)*x[2]
        arg[1,0,1,0]=(-0.479355493223)*x[0]+(-0.56368304348)*x[1]+(0.54679107219)*x[2]
        arg[1,0,1,1]=(-0.483618214285)*x[0]+(1.60438722757)*x[1]+(0.542389963596)*x[2]
        arg[1,0,2,0]=(0.130988624202)*x[0]+(-0.449468495509)*x[1]+(0.254640520023)*x[2]
        arg[1,0,2,1]=(1.13790232796)*x[0]+(-0.716218913004)*x[1]+(-0.193031264955)*x[2]
        arg[1,1,0,0]=(-0.0653152151059)*x[0]+(1.07890922508)*x[1]+(0.324995864884)*x[2]
        arg[1,1,0,1]=(0.769151345051)*x[0]+(-1.35911748672)*x[1]+(0.299326803785)*x[2]
        arg[1,1,1,0]=(-0.431128032697)*x[0]+(1.09261915758)*x[1]+(-0.029871998484)*x[2]
        arg[1,1,1,1]=(-0.922618915502)*x[0]+(1.30318141394)*x[1]+(-0.324039981399)*x[2]
        arg[1,1,2,0]=(-0.668178375077)*x[0]+(0.132203602988)*x[1]+(0.236035070297)*x[2]
        arg[1,1,2,1]=(-0.518594419598)*x[0]+(-0.0429695477263)*x[1]+(1.11877941395)*x[2]
        arg[1,2,0,0]=(0.610776644247)*x[0]+(-0.023900598958)*x[1]+(0.0480141870788)*x[2]
        arg[1,2,0,1]=(0.562736426016)*x[0]+(-0.294969624732)*x[1]+(-0.23898469143)*x[2]
        arg[1,2,1,0]=(0.645173726197)*x[0]+(-1.18365074993)*x[1]+(0.102581636468)*x[2]
        arg[1,2,1,1]=(-0.518976597355)*x[0]+(0.372817596528)*x[1]+(0.49035249397)*x[2]
        arg[1,2,2,0]=(0.0192623731103)*x[0]+(-0.241279406644)*x[1]+(-0.420015571981)*x[2]
        arg[1,2,2,1]=(0.318641827599)*x[0]+(-1.5357263392)*x[1]+(0.350150621892)*x[2]
        arg[1,3,0,0]=(-0.719401628191)*x[0]+(-1.77806447133)*x[1]+(1.19750443209)*x[2]
        arg[1,3,0,1]=(-1.00122204304)*x[0]+(0.675162216383)*x[1]+(1.1180105058)*x[2]
        arg[1,3,1,0]=(0.331149947519)*x[0]+(0.310832741708)*x[1]+(0.087866556795)*x[2]
        arg[1,3,1,1]=(-0.156343054316)*x[0]+(1.00796776488)*x[1]+(-1.53460740228)*x[2]
        arg[1,3,2,0]=(0.714240747972)*x[0]+(0.968170992322)*x[1]+(-0.962819191598)*x[2]
        arg[1,3,2,1]=(-0.200412114379)*x[0]+(-0.173781234544)*x[1]+(0.899584749309)*x[2]
        arg[1,4,0,0]=(-1.38011177642)*x[0]+(-0.316506470647)*x[1]+(0.266952180469)*x[2]
        arg[1,4,0,1]=(0.343359684584)*x[0]+(-1.13244466712)*x[1]+(-0.81840155125)*x[2]
        arg[1,4,1,0]=(0.295299129957)*x[0]+(0.199810907018)*x[1]+(0.631717407771)*x[2]
        arg[1,4,1,1]=(-0.46840784478)*x[0]+(-0.402683932206)*x[1]+(-0.782209293177)*x[2]
        arg[1,4,2,0]=(-1.8006389736)*x[0]+(0.119727360087)*x[1]+(0.482076626168)*x[2]
        arg[1,4,2,1]=(0.96553885281)*x[0]+(0.425097333238)*x[1]+(0.54533551085)*x[2]
        arg[2,0,0,0]=(0.932725602425)*x[0]+(-0.513780652569)*x[1]+(0.275742915648)*x[2]
        arg[2,0,0,1]=(1.10053578072)*x[0]+(0.787895323287)*x[1]+(-1.46544221328)*x[2]
        arg[2,0,1,0]=(0.83657269666)*x[0]+(-1.15717482853)*x[1]+(1.1090271297)*x[2]
        arg[2,0,1,1]=(1.59370135994)*x[0]+(0.452015708845)*x[1]+(0.0862230059082)*x[2]
        arg[2,0,2,0]=(-0.954832509866)*x[0]+(0.51192257473)*x[1]+(-1.48499771643)*x[2]
        arg[2,0,2,1]=(-0.556993361475)*x[0]+(0.438110051421)*x[1]+(0.237254018143)*x[2]
        arg[2,1,0,0]=(-0.318980966383)*x[0]+(-0.263071719097)*x[1]+(-0.348122442404)*x[2]
        arg[2,1,0,1]=(0.100191287326)*x[0]+(-0.422996999974)*x[1]+(-0.850702087112)*x[2]
        arg[2,1,1,0]=(-0.198942237141)*x[0]+(0.277459568173)*x[1]+(-1.66195774421)*x[2]
        arg[2,1,1,1]=(-1.62211381863)*x[0]+(-0.13417844308)*x[1]+(1.07441481356)*x[2]
        arg[2,1,2,0]=(-0.183120762554)*x[0]+(-0.158546687568)*x[1]+(1.05938565108)*x[2]
        arg[2,1,2,1]=(-0.530596736933)*x[0]+(-0.507872303289)*x[1]+(0.0437745116021)*x[2]
        arg[2,2,0,0]=(-0.430338834548)*x[0]+(-1.01989755268)*x[1]+(0.0364152729771)*x[2]
        arg[2,2,0,1]=(-0.55129175645)*x[0]+(0.283907226465)*x[1]+(0.281403899376)*x[2]
        arg[2,2,1,0]=(0.119036949181)*x[0]+(0.187265158741)*x[1]+(-0.0845814658663)*x[2]
        arg[2,2,1,1]=(1.39900884968)*x[0]+(1.42806588781)*x[1]+(0.211444691268)*x[2]
        arg[2,2,2,0]=(-0.609664666581)*x[0]+(-0.29988332429)*x[1]+(-0.28125800646)*x[2]
        arg[2,2,2,1]=(-1.77545247157)*x[0]+(0.337828315779)*x[1]+(-0.239141292616)*x[2]
        arg[2,3,0,0]=(-1.23141604656)*x[0]+(0.322998749903)*x[1]+(-1.70685281816)*x[2]
        arg[2,3,0,1]=(-1.4416737434)*x[0]+(1.49740114642)*x[1]+(0.732057753882)*x[2]
        arg[2,3,1,0]=(-1.59112622077)*x[0]+(0.665317566947)*x[1]+(0.826417253499)*x[2]
        arg[2,3,1,1]=(-1.22000830454)*x[0]+(0.618713896065)*x[1]+(1.27354984534)*x[2]
        arg[2,3,2,0]=(0.125639723221)*x[0]+(0.663241510656)*x[1]+(-0.426459288281)*x[2]
        arg[2,3,2,1]=(1.39120212155)*x[0]+(0.9115748844)*x[1]+(1.77889030223)*x[2]
        arg[2,4,0,0]=(-0.217927656387)*x[0]+(1.20391185439)*x[1]+(-0.188197734908)*x[2]
        arg[2,4,0,1]=(-0.442009082947)*x[0]+(0.399477101959)*x[1]+(-1.33974368956)*x[2]
        arg[2,4,1,0]=(0.636534809382)*x[0]+(0.872923470503)*x[1]+(-0.922286702006)*x[2]
        arg[2,4,1,1]=(-0.831840624431)*x[0]+(0.946223248622)*x[1]+(0.380300039133)*x[2]
        arg[2,4,2,0]=(0.0803088903901)*x[0]+(-0.857685204574)*x[1]+(-0.479475079431)*x[2]
        arg[2,4,2,1]=(-1.7233942516)*x[0]+(-1.38771345941)*x[1]+(0.761958201633)*x[2]
        arg[3,0,0,0]=(-0.738032075796)*x[0]+(1.33264583681)*x[1]+(-1.21101988385)*x[2]
        arg[3,0,0,1]=(1.40336327022)*x[0]+(-0.77078885213)*x[1]+(1.70960627204)*x[2]
        arg[3,0,1,0]=(0.452477629041)*x[0]+(0.235432009434)*x[1]+(1.69314009822)*x[2]
        arg[3,0,1,1]=(-0.816213143646)*x[0]+(-1.42050350309)*x[1]+(-0.0806105456758)*x[2]
        arg[3,0,2,0]=(0.0400708086023)*x[0]+(-0.336948609628)*x[1]+(-0.113687293995)*x[2]
        arg[3,0,2,1]=(0.414077492095)*x[0]+(-1.06937716207)*x[1]+(1.25972041173)*x[2]
        arg[3,1,0,0]=(-1.02975555201)*x[0]+(-0.123306171085)*x[1]+(-1.74750909033)*x[2]
        arg[3,1,0,1]=(-1.01988698009)*x[0]+(-0.142188487205)*x[1]+(0.274514515492)*x[2]
        arg[3,1,1,0]=(0.440626586141)*x[0]+(-0.0569580581917)*x[1]+(-0.77065631367)*x[2]
        arg[3,1,1,1]=(0.987903067972)*x[0]+(-1.47543187274)*x[1]+(-0.0888962732893)*x[2]
        arg[3,1,2,0]=(0.150506204416)*x[0]+(0.035476396146)*x[1]+(-0.852061690418)*x[2]
        arg[3,1,2,1]=(-1.15179800132)*x[0]+(-0.425570974979)*x[1]+(1.10092214699)*x[2]
        arg[3,2,0,0]=(1.01255188057)*x[0]+(1.24279465527)*x[1]+(1.25190246102)*x[2]
        arg[3,2,0,1]=(-1.16275538385)*x[0]+(1.13672085108)*x[1]+(0.131953731231)*x[2]
        arg[3,2,1,0]=(0.774290747173)*x[0]+(0.275410883089)*x[1]+(-0.825692266439)*x[2]
        arg[3,2,1,1]=(-1.07241838074)*x[0]+(-0.0974675737726)*x[1]+(1.1856849697)*x[2]
        arg[3,2,2,0]=(0.234209248938)*x[0]+(0.756893234577)*x[1]+(-0.365141055079)*x[2]
        arg[3,2,2,1]=(-1.6668858309)*x[0]+(-0.916769120288)*x[1]+(-0.716365681217)*x[2]
        arg[3,3,0,0]=(-0.36890215249)*x[0]+(-0.581062815398)*x[1]+(1.6280164507)*x[2]
        arg[3,3,0,1]=(-0.933634743559)*x[0]+(-0.0657657779145)*x[1]+(1.35011362804)*x[2]
        arg[3,3,1,0]=(1.13739289322)*x[0]+(0.0103208127738)*x[1]+(1.01628193014)*x[2]
        arg[3,3,1,1]=(0.037979696459)*x[0]+(-1.76926650495)*x[1]+(-1.28602572757)*x[2]
        arg[3,3,2,0]=(0.797830431467)*x[0]+(0.732108360923)*x[1]+(-0.436390867333)*x[2]
        arg[3,3,2,1]=(-0.571865447888)*x[0]+(0.895109047357)*x[1]+(-0.673234648776)*x[2]
        arg[3,4,0,0]=(-1.59667195158)*x[0]+(-1.06146105331)*x[1]+(-1.81255923783)*x[2]
        arg[3,4,0,1]=(-0.0721327214647)*x[0]+(1.23982376787)*x[1]+(1.43101011)*x[2]
        arg[3,4,1,0]=(0.117288395573)*x[0]+(-0.37678966212)*x[1]+(-1.14107285089)*x[2]
        arg[3,4,1,1]=(-0.303270532321)*x[0]+(1.33784402324)*x[1]+(-0.651624796413)*x[2]
        arg[3,4,2,0]=(0.215849013256)*x[0]+(0.247315697644)*x[1]+(0.404644392153)*x[2]
        arg[3,4,2,1]=(0.975781190478)*x[0]+(-0.252285063767)*x[1]+(-0.145694910854)*x[2]
        ref[0,0,0,0]=(-0.718934738033)/(o+1.)+(-0.639649667608)
        ref[0,0,0,1]=(1.01007662915)/(o+1.)+(0.111412440284)
        ref[0,0,1,0]=(0.827868953278)/(o+1.)+(0.26047765009)
        ref[0,0,1,1]=(1.48122352358)/(o+1.)+(0.0405568147034)
        ref[0,0,2,0]=(1.33282099436)/(o+1.)+(0.821761609054)
        ref[0,0,2,1]=(1.65298089862)/(o+1.)+(0.924021970342)
        ref[0,1,0,0]=(0.930162571453)/(o+1.)+(-0.0485897266263)
        ref[0,1,0,1]=(-0.347599158269)/(o+1.)+(-0.845177932176)
        ref[0,1,1,0]=(0.623152121663)/(o+1.)+(0.165545976811)
        ref[0,1,1,1]=(0.0506217690841)/(o+1.)+(0.470851792737)
        ref[0,1,2,0]=(-1.99181491207)/(o+1.)+(-1.0783764325)
        ref[0,1,2,1]=(-0.731065007801)/(o+1.)+(0.497836118493)
        ref[0,2,0,0]=(-0.0812249009151)/(o+1.)+(0.384247073896)
        ref[0,2,0,1]=(-0.115768505253)/(o+1.)+(0.241240728938)
        ref[0,2,1,0]=(0.152728009607)/(o+1.)+(0.0855247066589)
        ref[0,2,1,1]=(1.43086526503)/(o+1.)+(0.265083719563)
        ref[0,2,2,0]=(-0.967906369418)/(o+1.)+(-0.060131263415)
        ref[0,2,2,1]=(-0.269269238419)/(o+1.)+(0.698971869859)
        ref[0,3,0,0]=(-1.71659801402)/(o+1.)+(-0.201789393467)
        ref[0,3,0,1]=(0.265343635273)/(o+1.)+(0.744609104665)
        ref[0,3,1,0]=(1.83021875237)/(o+1.)+(-0.000662902039879)
        ref[0,3,1,1]=(-0.317529833375)/(o+1.)+(-0.857171520229)
        ref[0,3,2,0]=(-0.705757693524)/(o+1.)+(0.141947295438)
        ref[0,3,2,1]=(-1.27000495392)/(o+1.)+(-0.00999954704202)
        ref[0,4,0,0]=(0.061892522425)/(o+1.)+(0.112226800069)
        ref[0,4,0,1]=(-0.747062267148)/(o+1.)+(-0.0876876261864)
        ref[0,4,1,0]=(-1.19488448842)/(o+1.)+(0.0836552491261)
        ref[0,4,1,1]=(1.36573976292)/(o+1.)+(-0.9198322219)
        ref[0,4,2,0]=(-1.67982141697)/(o+1.)+(-0.361488591554)
        ref[0,4,2,1]=(-0.598050084547)/(o+1.)+(-0.181990295501)
        ref[1,0,0,0]=(-2.45927674426)/(o+1.)+(0.637878380777)
        ref[1,0,0,1]=(-0.575223026885)/(o+1.)+(-0.291559429493)
        ref[1,0,1,0]=(0.0871765959469)/(o+1.)+(-0.29171203023)
        ref[1,0,1,1]=(0.94050575309)/(o+1.)+(0.361326611893)
        ref[1,0,2,0]=(-0.306500880914)/(o+1.)+(0.121330764815)
        ref[1,0,2,1]=(1.60748296414)/(o+1.)+(-0.689415407072)
        ref[1,1,0,0]=(0.640401899755)/(o+1.)+(0.34909398755)
        ref[1,1,0,1]=(0.134022909144)/(o+1.)+(-0.212331123516)
        ref[1,1,1,0]=(0.248008519593)/(o+1.)+(0.191805303402)
        ref[1,1,1,1]=(0.638911732328)/(o+1.)+(-0.291194607647)
        ref[1,1,2,0]=(0.109264462855)/(o+1.)+(-0.204602082323)
        ref[1,1,2,1]=(-0.185963112656)/(o+1.)+(0.371589279643)
        ref[1,2,0,0]=(-0.590669733193)/(o+1.)+(0.61277998278)
        ref[1,2,0,1]=(-1.15856908266)/(o+1.)+(0.593675596257)
        ref[1,2,1,0]=(-0.451642187813)/(o+1.)+(0.00787340027166)
        ref[1,2,1,1]=(-0.428757501259)/(o+1.)+(0.386475497201)
        ref[1,2,2,0]=(-0.414532914271)/(o+1.)+(-0.113749845622)
        ref[1,2,2,1]=(-0.513613710163)/(o+1.)+(-0.176660089771)
        ref[1,3,0,0]=(-0.890487934256)/(o+1.)+(-0.20473686659)
        ref[1,3,0,1]=(1.64512779208)/(o+1.)+(-0.426588556466)
        ref[1,3,1,0]=(0.831527595966)/(o+1.)+(-0.0508391749725)
        ref[1,3,1,1]=(-0.718921003194)/(o+1.)+(0.0179691557385)
        ref[1,3,2,0]=(0.405959213301)/(o+1.)+(0.156816667697)
        ref[1,3,2,1]=(0.426319486322)/(o+1.)+(0.0495359570318)
        ref[1,4,0,0]=(-1.11364330874)/(o+1.)+(-0.158011378929)
        ref[1,4,0,1]=(-0.0993967101457)/(o+1.)+(-0.754044911817)
        ref[1,4,1,0]=(0.979493241598)/(o+1.)+(0.0736671015741)
        ref[1,4,1,1]=(-1.78692478393)/(o+1.)+(0.0668118568857)
        ref[1,4,2,0]=(-0.691473647622)/(o+1.)+(-0.253680669863)
        ref[1,4,2,1]=(1.16914997228)/(o+1.)+(0.383410862311)
        ref[2,0,0,0]=(0.790187638399)/(o+1.)+(-0.047749886448)
        ref[2,0,0,1]=(0.429625240932)/(o+1.)+(-0.00331817510394)
        ref[2,0,1,0]=(-0.101344997877)/(o+1.)+(0.444884997853)
        ref[2,0,1,1]=(2.15519953039)/(o+1.)+(-0.0116297278494)
        ref[2,0,2,0]=(-1.40159594866)/(o+1.)+(-0.263155851453)
        ref[2,0,2,1]=(-0.658391287862)/(o+1.)+(0.388380997976)
        ref[2,1,0,0]=(-0.773933630328)/(o+1.)+(-0.0781207487777)
        ref[2,1,0,1]=(0.0015551432788)/(o+1.)+(-0.587531471519)
        ref[2,1,1,0]=(0.122693324758)/(o+1.)+(-0.853066868967)
        ref[2,1,1,1]=(-0.219255630453)/(o+1.)+(-0.231310908847)
        ref[2,1,2,0]=(1.33176471286)/(o+1.)+(-0.307023255951)
        ref[2,1,2,1]=(-1.67471265259)/(o+1.)+(0.340009061986)
        ref[2,2,0,0]=(-0.727222063897)/(o+1.)+(-0.343299525178)
        ref[2,2,0,1]=(-1.87667590405)/(o+1.)+(0.945347636719)
        ref[2,2,1,0]=(0.776003927859)/(o+1.)+(-0.277141642902)
        ref[2,2,1,1]=(1.62850322537)/(o+1.)+(0.705008101698)
        ref[2,2,2,0]=(1.31869548331)/(o+1.)+(-1.25475074032)
        ref[2,2,2,1]=(-1.71691241115)/(o+1.)+(0.0200734813728)
        ref[2,3,0,0]=(-0.37741220744)/(o+1.)+(-1.11892895368)
        ref[2,3,0,1]=(0.684763737686)/(o+1.)+(0.0515107096073)
        ref[2,3,1,0]=(-0.0572023243047)/(o+1.)+(-0.0210945380092)
        ref[2,3,1,1]=(0.031457290932)/(o+1.)+(0.320399072966)
        ref[2,3,2,0]=(-0.71613334286)/(o+1.)+(0.539277644228)
        ref[2,3,2,1]=(2.10068005021)/(o+1.)+(0.990493628981)
        ref[2,4,0,0]=(0.830379294407)/(o+1.)+(-0.016296415655)
        ref[2,4,0,1]=(-1.05197706105)/(o+1.)+(-0.165149304749)
        ref[2,4,1,0]=(-0.0854959336547)/(o+1.)+(0.336333755767)
        ref[2,4,1,1]=(1.15845594915)/(o+1.)+(-0.331886642912)
        ref[2,4,2,0]=(-1.25248721318)/(o+1.)+(-0.00218209021539)
        ref[2,4,2,1]=(-0.991752052473)/(o+1.)+(-0.678698728452)
        ref[3,0,0,0]=(-0.262071815816)/(o+1.)+(-0.177167153508)
        ref[3,0,0,1]=(1.20598261029)/(o+1.)+(0.568099039922)
        ref[3,0,1,0]=(1.21237099375)/(o+1.)+(0.584339371474)
        ref[3,0,1,1]=(-1.04628498282)/(o+1.)+(-0.635521104799)
        ref[3,0,2,0]=(0.630843868383)/(o+1.)+(-0.520704481702)
        ref[3,0,2,1]=(1.56136208039)/(o+1.)+(-0.478470669318)
        ref[3,1,0,0]=(-0.372752943047)/(o+1.)+(-1.26390893519)
        ref[3,1,0,1]=(0.192139112377)/(o+1.)+(-0.53985003209)
        ref[3,1,1,0]=(-1.18480584023)/(o+1.)+(0.398909027255)
        ref[3,1,1,1]=(0.402069277892)/(o+1.)+(-0.489247177975)
        ref[3,1,2,0]=(0.27608763117)/(o+1.)+(-0.471083360513)
        ref[3,1,2,1]=(-0.273636427696)/(o+1.)+(-0.101405200809)
        ref[3,2,0,0]=(1.92572771773)/(o+1.)+(0.790760639563)
        ref[3,2,0,1]=(0.16264122626)/(o+1.)+(-0.0283610138982)
        ref[3,2,1,0]=(0.377561018039)/(o+1.)+(-0.076775827108)
        ref[3,2,1,1]=(-0.652466873272)/(o+1.)+(0.334132944227)
        ref[3,2,2,0]=(-0.0304721826842)/(o+1.)+(0.32821680556)
        ref[3,2,2,1]=(-1.52932569088)/(o+1.)+(-0.885347470761)
        ref[3,3,0,0]=(-0.451938162279)/(o+1.)+(0.564994822546)
        ref[3,3,0,1]=(-0.477583137662)/(o+1.)+(0.414148122113)
        ref[3,3,1,0]=(1.38037325689)/(o+1.)+(0.391811189625)
        ref[3,3,1,1]=(-1.01333159097)/(o+1.)+(-1.00199047255)
        ref[3,3,2,0]=(1.3222644334)/(o+1.)+(-0.114358254173)
        ref[3,3,2,1]=(-0.248989333346)/(o+1.)+(-0.0505008579799)
        ref[3,4,0,0]=(-1.92196761736)/(o+1.)+(-1.27436231268)
        ref[3,4,0,1]=(0.894404962457)/(o+1.)+(0.852148096971)
        ref[3,4,1,0]=(0.868763436571)/(o+1.)+(-1.134668777)
        ref[3,4,1,1]=(0.438747123022)/(o+1.)+(-0.0278992142568)
        ref[3,4,2,0]=(0.532575431402)/(o+1.)+(0.167616835826)
        ref[3,4,2,1]=(0.796744496799)/(o+1.)+(-0.109471640471)
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.217470460227)*x[0]**o+(0.664132905811)*x[0]+(0.493878617743)*x[1]**o+(0.111327715714)*x[1]
        ref=(0.276408157516)/(o+1.)+(0.387730310763)
      else:
        arg=(0.371937838886)*x[0]**o+(0.653846807778)*x[0]+(0.172358058188)*x[1]**o+(-0.928017916614)*x[1]+(0.116156616174)*x[2]**o+(0.371332501889)*x[2]
        ref=(0.660452513249)/(o+1.)+(0.0485806965265)
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.329952046701)*x[0]**o+(0.12757022314)*x[0]+(0.198280362833)*x[1]**o+(0.129179854334)*x[1]
        arg[1]=(0.185895161332)*x[0]**o+(0.270740935936)*x[0]+(0.630122833594)*x[1]**o+(-0.0696403698787)*x[1]
        ref[0]=(-0.131671683869)/(o+1.)+(0.128375038737)
        ref[1]=(0.816017994925)/(o+1.)+(0.100550283029)
      else:
        arg[0]=(0.526257699936)*x[0]**o+(0.0731442796396)*x[0]+(-0.266121642119)*x[1]**o+(-0.929478242136)*x[1]+(0.743092861949)*x[2]**o+(0.22690419858)*x[2]
        arg[1]=(-0.121387039147)*x[0]**o+(-0.0580608263784)*x[0]+(0.761164059949)*x[1]**o+(0.821938972431)*x[1]+(-0.756423175962)*x[2]**o+(0.213207132353)*x[2]
        ref[0]=(1.00322891977)/(o+1.)+(-0.314714881958)
        ref[1]=(-0.11664615516)/(o+1.)+(0.488542639203)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.963318847499)*x[0]**o+(-0.184977238308)*x[0]+(0.153633231879)*x[1]**o+(-0.640028315216)*x[1]
        arg[0,1]=(0.0879103656102)*x[0]**o+(-0.928854425726)*x[0]+(-0.138566176095)*x[1]**o+(-0.670118398936)*x[1]
        arg[0,2]=(0.920856326204)*x[0]**o+(0.461047594595)*x[0]+(-0.914778320363)*x[1]**o+(-0.561644217568)*x[1]
        arg[0,3]=(0.506148289443)*x[0]**o+(-0.474044343536)*x[0]+(0.742127575836)*x[1]**o+(-0.52349314159)*x[1]
        arg[0,4]=(0.0810224543145)*x[0]**o+(-0.307992483498)*x[0]+(0.749673245291)*x[1]**o+(-0.554099873044)*x[1]
        arg[1,0]=(-0.270897099553)*x[0]**o+(-0.222178461886)*x[0]+(0.755729049844)*x[1]**o+(0.740743777894)*x[1]
        arg[1,1]=(0.0220934176819)*x[0]**o+(-0.309641658323)*x[0]+(-0.951833592348)*x[1]**o+(0.430554526918)*x[1]
        arg[1,2]=(0.176033311218)*x[0]**o+(0.728188001962)*x[0]+(-0.428357954006)*x[1]**o+(0.955059500699)*x[1]
        arg[1,3]=(0.946680422474)*x[0]**o+(-0.251968148743)*x[0]+(-0.3704600461)*x[1]**o+(0.692080922297)*x[1]
        arg[1,4]=(0.122655375591)*x[0]**o+(0.617285886533)*x[0]+(0.729184247986)*x[1]**o+(0.30369620264)*x[1]
        arg[2,0]=(-0.957780501271)*x[0]**o+(0.438290429468)*x[0]+(-0.504841431279)*x[1]**o+(0.935350834608)*x[1]
        arg[2,1]=(-0.49410765407)*x[0]**o+(-0.635142349002)*x[0]+(0.700131616572)*x[1]**o+(0.99523209861)*x[1]
        arg[2,2]=(0.0967686936303)*x[0]**o+(0.987754073381)*x[0]+(-0.161220173386)*x[1]**o+(-0.343926097784)*x[1]
        arg[2,3]=(0.560040886628)*x[0]**o+(-0.843778720323)*x[0]+(0.345354034369)*x[1]**o+(0.440584543858)*x[1]
        arg[2,4]=(0.988977815245)*x[0]**o+(0.661922598357)*x[0]+(-0.425421002549)*x[1]**o+(0.624279872655)*x[1]
        arg[3,0]=(-0.997217068098)*x[0]**o+(-0.973497091595)*x[0]+(-0.0980683851598)*x[1]**o+(0.408081557041)*x[1]
        arg[3,1]=(0.0827616522414)*x[0]**o+(0.931022772362)*x[0]+(-0.588688206358)*x[1]**o+(0.900467348992)*x[1]
        arg[3,2]=(0.0933642874383)*x[0]**o+(0.398732664158)*x[0]+(-0.437102212783)*x[1]**o+(-0.126538302466)*x[1]
        arg[3,3]=(0.251989644486)*x[0]**o+(0.765809815972)*x[0]+(-0.325455283605)*x[1]**o+(0.171819372538)*x[1]
        arg[3,4]=(0.649407682145)*x[0]**o+(0.963776897105)*x[0]+(0.671915752131)*x[1]**o+(-0.468729629153)*x[1]
        ref[0,0]=(-0.80968561562)/(o+1.)+(-0.412502776762)
        ref[0,1]=(-0.0506558104848)/(o+1.)+(-0.799486412331)
        ref[0,2]=(0.00607800584081)/(o+1.)+(-0.0502983114867)
        ref[0,3]=(1.24827586528)/(o+1.)+(-0.498768742563)
        ref[0,4]=(0.830695699605)/(o+1.)+(-0.431046178271)
        ref[1,0]=(0.484831950292)/(o+1.)+(0.259282658004)
        ref[1,1]=(-0.929740174666)/(o+1.)+(0.0604564342974)
        ref[1,2]=(-0.252324642788)/(o+1.)+(0.84162375133)
        ref[1,3]=(0.576220376373)/(o+1.)+(0.220056386777)
        ref[1,4]=(0.851839623577)/(o+1.)+(0.460491044586)
        ref[2,0]=(-1.46262193255)/(o+1.)+(0.686820632038)
        ref[2,1]=(0.206023962501)/(o+1.)+(0.180044874804)
        ref[2,2]=(-0.0644514797559)/(o+1.)+(0.321913987798)
        ref[2,3]=(0.905394920997)/(o+1.)+(-0.201597088233)
        ref[2,4]=(0.563556812695)/(o+1.)+(0.643101235506)
        ref[3,0]=(-1.09528545326)/(o+1.)+(-0.282707767277)
        ref[3,1]=(-0.505926554117)/(o+1.)+(0.915745060677)
        ref[3,2]=(-0.343737925345)/(o+1.)+(0.136097180846)
        ref[3,3]=(-0.0734656391194)/(o+1.)+(0.468814594255)
        ref[3,4]=(1.32132343428)/(o+1.)+(0.247523633976)
      else:
        arg[0,0]=(-0.512698606017)*x[0]**o+(-0.156853163498)*x[0]+(0.247997355633)*x[1]**o+(-0.363003068904)*x[1]+(-0.940097989108)*x[2]**o+(-0.661063850288)*x[2]
        arg[0,1]=(-0.886390181577)*x[0]**o+(0.724049511142)*x[0]+(0.260684592163)*x[1]**o+(0.265648885034)*x[1]+(0.123084281693)*x[2]**o+(0.181492979151)*x[2]
        arg[0,2]=(-0.570543998851)*x[0]**o+(0.0541041165388)*x[0]+(0.630994990544)*x[1]**o+(-0.89311079559)*x[1]+(0.750714073617)*x[2]**o+(-0.0684796380527)*x[2]
        arg[0,3]=(0.605486031268)*x[0]**o+(0.0387617487672)*x[0]+(-0.365027101406)*x[1]**o+(0.323275180514)*x[1]+(-0.763142660003)*x[2]**o+(-0.331069480921)*x[2]
        arg[0,4]=(-0.0779035848472)*x[0]**o+(-0.0234904624931)*x[0]+(0.591635786678)*x[1]**o+(-0.793680126332)*x[1]+(-0.874742296757)*x[2]**o+(-0.184528885157)*x[2]
        arg[1,0]=(-0.62480972544)*x[0]**o+(-0.99242215547)*x[0]+(0.634250679117)*x[1]**o+(-0.785448642692)*x[1]+(-0.013335654402)*x[2]**o+(0.52112750417)*x[2]
        arg[1,1]=(0.611541580179)*x[0]**o+(-0.203144044802)*x[0]+(-0.411277184995)*x[1]**o+(0.373487145939)*x[1]+(-0.679339345026)*x[2]**o+(-0.632413536218)*x[2]
        arg[1,2]=(-0.258927016205)*x[0]**o+(0.0242628959926)*x[0]+(-0.596909760383)*x[1]**o+(-0.295787354787)*x[1]+(-0.337930022823)*x[2]**o+(-0.825085100446)*x[2]
        arg[1,3]=(-0.0976663781507)*x[0]**o+(0.716600132838)*x[0]+(-0.682155341898)*x[1]**o+(-0.154484864167)*x[1]+(0.754686874908)*x[2]**o+(-0.102453470268)*x[2]
        arg[1,4]=(0.268714339964)*x[0]**o+(0.974234398844)*x[0]+(-0.370113158384)*x[1]**o+(0.0745228956861)*x[1]+(-0.0674370676047)*x[2]**o+(0.132605448278)*x[2]
        arg[2,0]=(-0.00750893262562)*x[0]**o+(0.507641887134)*x[0]+(0.638483932117)*x[1]**o+(-0.940295540912)*x[1]+(-0.201492519698)*x[2]**o+(0.504878595818)*x[2]
        arg[2,1]=(-0.0283522980593)*x[0]**o+(-0.76577346605)*x[0]+(0.0863472830598)*x[1]**o+(-0.904447859187)*x[1]+(-0.203390005492)*x[2]**o+(-0.741629136645)*x[2]
        arg[2,2]=(-0.0196123831667)*x[0]**o+(-0.320721372799)*x[0]+(0.0670421287749)*x[1]**o+(-0.415029682416)*x[1]+(0.450269143475)*x[2]**o+(-0.735201210502)*x[2]
        arg[2,3]=(-0.621537330758)*x[0]**o+(0.21573881897)*x[0]+(0.0364275955835)*x[1]**o+(0.873908952751)*x[1]+(-0.197046790303)*x[2]**o+(0.655654722165)*x[2]
        arg[2,4]=(0.432531007791)*x[0]**o+(-0.65605251166)*x[0]+(0.728245204833)*x[1]**o+(-0.769998936989)*x[1]+(-0.797879623133)*x[2]**o+(0.36575554332)*x[2]
        arg[3,0]=(0.512219734043)*x[0]**o+(-0.177942735603)*x[0]+(-0.307293350237)*x[1]**o+(-0.780181541911)*x[1]+(-0.510201727643)*x[2]**o+(-0.567980715525)*x[2]
        arg[3,1]=(-0.405123989366)*x[0]**o+(-0.133316566583)*x[0]+(0.552868689995)*x[1]**o+(-0.921149434884)*x[1]+(0.510851078054)*x[2]**o+(0.7207533866)*x[2]
        arg[3,2]=(0.150011788901)*x[0]**o+(-0.55947555552)*x[0]+(-0.353366762442)*x[1]**o+(0.204954197301)*x[1]+(-0.48254632377)*x[2]**o+(-0.272738900127)*x[2]
        arg[3,3]=(-0.690981207461)*x[0]**o+(0.38683432128)*x[0]+(0.655731808738)*x[1]**o+(0.162269870301)*x[1]+(-0.373468659959)*x[2]**o+(-0.841431521009)*x[2]
        arg[3,4]=(0.156455675319)*x[0]**o+(0.501690084272)*x[0]+(-0.339255791314)*x[1]**o+(0.0832260784264)*x[1]+(0.394137401365)*x[2]**o+(-0.92403906326)*x[2]
        ref[0,0]=(-1.20479923949)/(o+1.)+(-0.590460041345)
        ref[0,1]=(-0.502621307721)/(o+1.)+(0.585595687663)
        ref[0,2]=(0.811165065311)/(o+1.)+(-0.453743158552)
        ref[0,3]=(-0.522683730141)/(o+1.)+(0.0154837241797)
        ref[0,4]=(-0.361010094926)/(o+1.)+(-0.500849736991)
        ref[1,0]=(-0.00389470072479)/(o+1.)+(-0.628371646996)
        ref[1,1]=(-0.479074949842)/(o+1.)+(-0.23103521754)
        ref[1,2]=(-1.19376679941)/(o+1.)+(-0.54830477962)
        ref[1,3]=(-0.0251348451402)/(o+1.)+(0.229830899202)
        ref[1,4]=(-0.168835886025)/(o+1.)+(0.590681371404)
        ref[2,0]=(0.429482479793)/(o+1.)+(0.0361124710203)
        ref[2,1]=(-0.145395020492)/(o+1.)+(-1.20592523094)
        ref[2,2]=(0.497698889084)/(o+1.)+(-0.735476132859)
        ref[2,3]=(-0.782156525477)/(o+1.)+(0.872651246943)
        ref[2,4]=(0.362896589491)/(o+1.)+(-0.530147952664)
        ref[3,0]=(-0.305275343837)/(o+1.)+(-0.763052496519)
        ref[3,1]=(0.658595778683)/(o+1.)+(-0.166856307434)
        ref[3,2]=(-0.685901297311)/(o+1.)+(-0.313630129173)
        ref[3,3]=(-0.408718058683)/(o+1.)+(-0.146163664714)
        ref[3,4]=(0.21133728537)/(o+1.)+(-0.169561450281)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.165423911829)*x[0]**o+(0.912542062064)*x[0]+(0.515840721225)*x[1]**o+(-0.855922021017)*x[1]
        arg[0,0,1]=(0.591873802546)*x[0]**o+(-0.039385167187)*x[0]+(0.0261484884274)*x[1]**o+(0.227950240434)*x[1]
        arg[0,1,0]=(-0.825833712046)*x[0]**o+(0.989606797765)*x[0]+(0.975260558791)*x[1]**o+(0.297963603984)*x[1]
        arg[0,1,1]=(-0.658054502482)*x[0]**o+(-0.865505090963)*x[0]+(0.817184084453)*x[1]**o+(-0.409588098899)*x[1]
        arg[1,0,0]=(0.538456987198)*x[0]**o+(-0.0768787509391)*x[0]+(0.424620431835)*x[1]**o+(0.799917948831)*x[1]
        arg[1,0,1]=(0.125945558199)*x[0]**o+(-0.48320049996)*x[0]+(0.105474962327)*x[1]**o+(-0.090713555582)*x[1]
        arg[1,1,0]=(0.835213375418)*x[0]**o+(-0.513407038769)*x[0]+(-0.342890811735)*x[1]**o+(-0.285057736244)*x[1]
        arg[1,1,1]=(0.762782085749)*x[0]**o+(-0.308457087352)*x[0]+(0.23756726753)*x[1]**o+(-0.909365956504)*x[1]
        arg[2,0,0]=(-0.76467980678)*x[0]**o+(-0.265552922006)*x[0]+(-0.261862355539)*x[1]**o+(-0.399407404618)*x[1]
        arg[2,0,1]=(0.79401193617)*x[0]**o+(-0.0300461495216)*x[0]+(-0.978834133508)*x[1]**o+(-0.984845382599)*x[1]
        arg[2,1,0]=(0.898033018534)*x[0]**o+(-0.184513430159)*x[0]+(0.0430634098526)*x[1]**o+(0.62213313954)*x[1]
        arg[2,1,1]=(0.126926439246)*x[0]**o+(0.671259428917)*x[0]+(-0.396062855417)*x[1]**o+(0.0461825691916)*x[1]
        arg[3,0,0]=(0.185756594279)*x[0]**o+(0.570946291973)*x[0]+(0.0673731944956)*x[1]**o+(0.367766610292)*x[1]
        arg[3,0,1]=(0.021269224946)*x[0]**o+(-0.463663279063)*x[0]+(-0.687930988068)*x[1]**o+(0.184714204849)*x[1]
        arg[3,1,0]=(0.754288639303)*x[0]**o+(-0.94612827517)*x[0]+(0.802463196709)*x[1]**o+(0.202355582077)*x[1]
        arg[3,1,1]=(0.840002857786)*x[0]**o+(-0.692207739447)*x[0]+(0.486569391181)*x[1]**o+(-0.713960793192)*x[1]
        arg[4,0,0]=(-0.205794539655)*x[0]**o+(-0.930458992533)*x[0]+(-0.62580205172)*x[1]**o+(-0.884738110673)*x[1]
        arg[4,0,1]=(-0.91433612349)*x[0]**o+(0.694961527804)*x[0]+(-0.510941334746)*x[1]**o+(-0.510725271706)*x[1]
        arg[4,1,0]=(0.852785499366)*x[0]**o+(0.870826628851)*x[0]+(-0.577529754198)*x[1]**o+(-0.583793684857)*x[1]
        arg[4,1,1]=(-0.604869226881)*x[0]**o+(0.482463572402)*x[0]+(0.829654316885)*x[1]**o+(0.597543822455)*x[1]
        arg[5,0,0]=(0.413459469749)*x[0]**o+(0.865143264083)*x[0]+(0.986197644249)*x[1]**o+(-0.2511021346)*x[1]
        arg[5,0,1]=(0.811572168213)*x[0]**o+(-0.110121334437)*x[0]+(-0.985492232242)*x[1]**o+(-0.913436084156)*x[1]
        arg[5,1,0]=(-0.660869051422)*x[0]**o+(-0.743235620427)*x[0]+(0.205393778238)*x[1]**o+(0.296069404687)*x[1]
        arg[5,1,1]=(-0.13285011267)*x[0]**o+(0.659400366834)*x[0]+(-0.889984368942)*x[1]**o+(0.584893563559)*x[1]
        ref[0,0,0]=(0.350416809396)/(o+1.)+(0.0283100205234)
        ref[0,0,1]=(0.618022290973)/(o+1.)+(0.0942825366236)
        ref[0,1,0]=(0.149426846745)/(o+1.)+(0.643785200874)
        ref[0,1,1]=(0.159129581971)/(o+1.)+(-0.637546594931)
        ref[1,0,0]=(0.963077419032)/(o+1.)+(0.361519598946)
        ref[1,0,1]=(0.231420520526)/(o+1.)+(-0.286957027771)
        ref[1,1,0]=(0.492322563683)/(o+1.)+(-0.399232387506)
        ref[1,1,1]=(1.00034935328)/(o+1.)+(-0.608911521928)
        ref[2,0,0]=(-1.02654216232)/(o+1.)+(-0.332480163312)
        ref[2,0,1]=(-0.184822197338)/(o+1.)+(-0.507445766061)
        ref[2,1,0]=(0.941096428387)/(o+1.)+(0.21880985469)
        ref[2,1,1]=(-0.269136416172)/(o+1.)+(0.358720999054)
        ref[3,0,0]=(0.253129788774)/(o+1.)+(0.469356451133)
        ref[3,0,1]=(-0.666661763122)/(o+1.)+(-0.139474537107)
        ref[3,1,0]=(1.55675183601)/(o+1.)+(-0.371886346546)
        ref[3,1,1]=(1.32657224897)/(o+1.)+(-0.703084266319)
        ref[4,0,0]=(-0.831596591375)/(o+1.)+(-0.907598551603)
        ref[4,0,1]=(-1.42527745824)/(o+1.)+(0.0921181280487)
        ref[4,1,0]=(0.275255745168)/(o+1.)+(0.143516471997)
        ref[4,1,1]=(0.224785090004)/(o+1.)+(0.540003697429)
        ref[5,0,0]=(1.399657114)/(o+1.)+(0.307020564742)
        ref[5,0,1]=(-0.173920064028)/(o+1.)+(-0.511778709296)
        ref[5,1,0]=(-0.455475273184)/(o+1.)+(-0.22358310787)
        ref[5,1,1]=(-1.02283448161)/(o+1.)+(0.622146965196)
      else:
        arg[0,0,0]=(-0.564478749422)*x[0]**o+(-0.0357223504179)*x[0]+(0.396603298865)*x[1]**o+(-0.461941327703)*x[1]+(-0.667580657928)*x[2]**o+(-0.589957067263)*x[2]
        arg[0,0,1]=(0.930616008951)*x[0]**o+(-0.764067565635)*x[0]+(-0.71350617205)*x[1]**o+(-0.530942701898)*x[1]+(-0.103797356434)*x[2]**o+(0.410704036673)*x[2]
        arg[0,1,0]=(-0.59127392299)*x[0]**o+(0.452490122131)*x[0]+(0.714793625631)*x[1]**o+(-0.933114051133)*x[1]+(-0.693437840325)*x[2]**o+(0.979267231764)*x[2]
        arg[0,1,1]=(0.179682332267)*x[0]**o+(-0.419960476425)*x[0]+(-0.10986598341)*x[1]**o+(0.912986428497)*x[1]+(0.0886946782573)*x[2]**o+(0.516065958873)*x[2]
        arg[1,0,0]=(0.524446863481)*x[0]**o+(-0.439668193423)*x[0]+(0.681259811488)*x[1]**o+(0.439275262258)*x[1]+(0.84573881706)*x[2]**o+(0.377640553175)*x[2]
        arg[1,0,1]=(-0.007464905203)*x[0]**o+(0.234727832471)*x[0]+(0.0428956356618)*x[1]**o+(0.195772930172)*x[1]+(0.457790934614)*x[2]**o+(0.445309166495)*x[2]
        arg[1,1,0]=(0.821283375827)*x[0]**o+(0.212480005798)*x[0]+(0.148817990961)*x[1]**o+(0.119716150877)*x[1]+(0.658959126865)*x[2]**o+(-0.346216916131)*x[2]
        arg[1,1,1]=(-0.0587222422894)*x[0]**o+(-0.325321661973)*x[0]+(0.149293818689)*x[1]**o+(-0.632073115037)*x[1]+(0.018604271874)*x[2]**o+(-0.412288240166)*x[2]
        arg[2,0,0]=(-0.22588082169)*x[0]**o+(0.197780461217)*x[0]+(0.387336989213)*x[1]**o+(-0.680021661016)*x[1]+(-0.903782254346)*x[2]**o+(-0.273268109853)*x[2]
        arg[2,0,1]=(0.415876394297)*x[0]**o+(0.199596795904)*x[0]+(-0.445817134682)*x[1]**o+(0.672539590684)*x[1]+(0.907825473964)*x[2]**o+(0.459560469691)*x[2]
        arg[2,1,0]=(-0.052146887134)*x[0]**o+(-0.229450574738)*x[0]+(0.0929720151604)*x[1]**o+(-0.893861034491)*x[1]+(-0.294030328982)*x[2]**o+(-0.193256779109)*x[2]
        arg[2,1,1]=(0.361171422066)*x[0]**o+(-0.212022056495)*x[0]+(0.0672221746963)*x[1]**o+(0.57506871515)*x[1]+(0.654894469509)*x[2]**o+(0.0749777260533)*x[2]
        arg[3,0,0]=(0.505355253404)*x[0]**o+(-0.0753563586359)*x[0]+(0.544299279727)*x[1]**o+(-0.781954131328)*x[1]+(0.534390713891)*x[2]**o+(-0.719289344378)*x[2]
        arg[3,0,1]=(-0.589309652827)*x[0]**o+(-0.0557421151178)*x[0]+(-0.275311076837)*x[1]**o+(0.998180518149)*x[1]+(-0.844823856433)*x[2]**o+(-0.988843805162)*x[2]
        arg[3,1,0]=(-0.411075104301)*x[0]**o+(-0.405896828894)*x[0]+(0.242301856034)*x[1]**o+(-0.760648969984)*x[1]+(0.0940367373673)*x[2]**o+(-0.121420230853)*x[2]
        arg[3,1,1]=(-0.201426146967)*x[0]**o+(0.722847134095)*x[0]+(-0.00743580369433)*x[1]**o+(-0.593158423904)*x[1]+(0.630575332064)*x[2]**o+(-0.309839791343)*x[2]
        arg[4,0,0]=(0.370057366738)*x[0]**o+(0.35683022394)*x[0]+(0.886304520049)*x[1]**o+(0.834488899465)*x[1]+(0.27408661832)*x[2]**o+(0.325007409933)*x[2]
        arg[4,0,1]=(0.386153377977)*x[0]**o+(-0.889700309398)*x[0]+(-0.60695970137)*x[1]**o+(0.701208035732)*x[1]+(0.435164510485)*x[2]**o+(0.40223148148)*x[2]
        arg[4,1,0]=(-0.227256041906)*x[0]**o+(-0.0406115125763)*x[0]+(-0.0298974314183)*x[1]**o+(0.887167464033)*x[1]+(0.1004751563)*x[2]**o+(0.826924545719)*x[2]
        arg[4,1,1]=(0.733575640137)*x[0]**o+(-0.376269474248)*x[0]+(-0.841290711112)*x[1]**o+(-0.0633013917179)*x[1]+(0.335730636406)*x[2]**o+(0.913607658221)*x[2]
        arg[5,0,0]=(-0.896019254334)*x[0]**o+(0.465470987762)*x[0]+(0.338543664838)*x[1]**o+(0.182102591196)*x[1]+(0.476317455739)*x[2]**o+(0.191367623163)*x[2]
        arg[5,0,1]=(0.828458939676)*x[0]**o+(-0.37853033744)*x[0]+(-0.540809170906)*x[1]**o+(0.367869598038)*x[1]+(0.321196361443)*x[2]**o+(0.548981819769)*x[2]
        arg[5,1,0]=(-0.0805871335579)*x[0]**o+(-0.554679375424)*x[0]+(-0.259438414137)*x[1]**o+(0.72362534417)*x[1]+(-0.838855028889)*x[2]**o+(0.929802758377)*x[2]
        arg[5,1,1]=(-0.947910259125)*x[0]**o+(0.654012824998)*x[0]+(0.425232602078)*x[1]**o+(-0.649345176159)*x[1]+(-0.946280438348)*x[2]**o+(-0.429547284912)*x[2]
        ref[0,0,0]=(-0.835456108484)/(o+1.)+(-0.543810372692)
        ref[0,0,1]=(0.113312480467)/(o+1.)+(-0.44215311543)
        ref[0,1,0]=(-0.569918137684)/(o+1.)+(0.249321651381)
        ref[0,1,1]=(0.158511027114)/(o+1.)+(0.504545955472)
        ref[1,0,0]=(2.05144549203)/(o+1.)+(0.188623811006)
        ref[1,0,1]=(0.493221665073)/(o+1.)+(0.437904964569)
        ref[1,1,0]=(1.62906049365)/(o+1.)+(-0.00701037972825)
        ref[1,1,1]=(0.109175848273)/(o+1.)+(-0.684841508588)
        ref[2,0,0]=(-0.742326086822)/(o+1.)+(-0.377754654826)
        ref[2,0,1]=(0.877884733578)/(o+1.)+(0.665848428139)
        ref[2,1,0]=(-0.253205200956)/(o+1.)+(-0.658284194169)
        ref[2,1,1]=(1.08328806627)/(o+1.)+(0.219012192354)
        ref[3,0,0]=(1.58404524702)/(o+1.)+(-0.788299917171)
        ref[3,0,1]=(-1.7094445861)/(o+1.)+(-0.0232027010654)
        ref[3,1,0]=(-0.0747365109002)/(o+1.)+(-0.643983014866)
        ref[3,1,1]=(0.421713381403)/(o+1.)+(-0.0900755405762)
        ref[4,0,0]=(1.53044850511)/(o+1.)+(0.758163266669)
        ref[4,0,1]=(0.214358187092)/(o+1.)+(0.106869603907)
        ref[4,1,0]=(-0.156678317024)/(o+1.)+(0.836740248588)
        ref[4,1,1]=(0.228015565432)/(o+1.)+(0.237018396127)
        ref[5,0,0]=(-0.0811581337568)/(o+1.)+(0.419470601061)
        ref[5,0,1]=(0.608846130213)/(o+1.)+(0.269160540183)
        ref[5,1,0]=(-1.17888057658)/(o+1.)+(0.549374363561)
        ref[5,1,1]=(-1.4689580954)/(o+1.)+(-0.212439818036)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromData_ReducedFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.216014106349)*x[0]**o+(0.98826230471)*x[0]+(0.522566177319)*x[1]**o+(-0.672520997452)*x[1]
        arg[0,0,0,1]=(0.368024408084)*x[0]**o+(0.047972971332)*x[0]+(-0.307374327321)*x[1]**o+(0.0745169735457)*x[1]
        arg[0,0,1,0]=(0.438735045452)*x[0]**o+(0.76355958752)*x[0]+(0.200829025437)*x[1]**o+(0.56912407759)*x[1]
        arg[0,0,1,1]=(-0.0328769393712)*x[0]**o+(-0.39701363611)*x[0]+(0.704125532519)*x[1]**o+(-0.6602435084)*x[1]
        arg[0,0,2,0]=(-0.547717992717)*x[0]**o+(-0.844256763776)*x[0]+(0.621300470147)*x[1]**o+(-0.105216181226)*x[1]
        arg[0,0,2,1]=(-0.723474063234)*x[0]**o+(-0.980738972901)*x[0]+(-0.183578659566)*x[1]**o+(0.586038450047)*x[1]
        arg[0,1,0,0]=(0.505379907001)*x[0]**o+(-0.470514711277)*x[0]+(0.799116811747)*x[1]**o+(0.682760846386)*x[1]
        arg[0,1,0,1]=(0.353992609198)*x[0]**o+(-0.848107655585)*x[0]+(-0.774065459855)*x[1]**o+(0.0323137462292)*x[1]
        arg[0,1,1,0]=(0.799068805694)*x[0]**o+(-0.979761821209)*x[0]+(0.642674987867)*x[1]**o+(-0.639029952044)*x[1]
        arg[0,1,1,1]=(0.542556683327)*x[0]**o+(0.876753086356)*x[0]+(-0.387121297643)*x[1]**o+(0.377543179628)*x[1]
        arg[0,1,2,0]=(0.424672899596)*x[0]**o+(-0.134184751569)*x[0]+(-0.561877304114)*x[1]**o+(0.549933699347)*x[1]
        arg[0,1,2,1]=(-0.126173294501)*x[0]**o+(-0.338964866844)*x[0]+(-0.163536115759)*x[1]**o+(0.747551558186)*x[1]
        arg[0,2,0,0]=(0.304334183452)*x[0]**o+(0.927809971187)*x[0]+(0.961504942988)*x[1]**o+(0.0698014252396)*x[1]
        arg[0,2,0,1]=(0.488820088455)*x[0]**o+(0.890735247208)*x[0]+(0.686782901398)*x[1]**o+(0.972954994564)*x[1]
        arg[0,2,1,0]=(-0.737577792766)*x[0]**o+(-0.351698913646)*x[0]+(0.392591489865)*x[1]**o+(0.79238240143)*x[1]
        arg[0,2,1,1]=(-0.997254027544)*x[0]**o+(0.579800166465)*x[0]+(-0.95000690643)*x[1]**o+(-0.166463083476)*x[1]
        arg[0,2,2,0]=(-0.999703749644)*x[0]**o+(0.0415211466468)*x[0]+(0.0816160245139)*x[1]**o+(-0.0921114587195)*x[1]
        arg[0,2,2,1]=(-0.172818721386)*x[0]**o+(0.10475692446)*x[0]+(-0.855815339933)*x[1]**o+(-0.50842580034)*x[1]
        arg[0,3,0,0]=(-0.154563121513)*x[0]**o+(-0.0956305513369)*x[0]+(-0.893783521038)*x[1]**o+(0.10772474638)*x[1]
        arg[0,3,0,1]=(0.365331678349)*x[0]**o+(0.403124820832)*x[0]+(-0.0750500521493)*x[1]**o+(0.62559762808)*x[1]
        arg[0,3,1,0]=(-0.42235875521)*x[0]**o+(0.831133238533)*x[0]+(-0.893322944982)*x[1]**o+(0.936074096319)*x[1]
        arg[0,3,1,1]=(0.587382334689)*x[0]**o+(0.996190710631)*x[0]+(-0.796032791654)*x[1]**o+(0.783491453801)*x[1]
        arg[0,3,2,0]=(-0.369527027508)*x[0]**o+(-0.501417552089)*x[0]+(-0.372105439045)*x[1]**o+(0.653283000152)*x[1]
        arg[0,3,2,1]=(0.914893750314)*x[0]**o+(-0.769183323793)*x[0]+(0.571638486717)*x[1]**o+(-0.939784652259)*x[1]
        arg[0,4,0,0]=(0.134859060251)*x[0]**o+(-0.297363165275)*x[0]+(0.198957028974)*x[1]**o+(-0.828531725127)*x[1]
        arg[0,4,0,1]=(-0.371375370217)*x[0]**o+(-0.0422058718402)*x[0]+(0.502010302278)*x[1]**o+(0.442051802597)*x[1]
        arg[0,4,1,0]=(0.513346571173)*x[0]**o+(0.182640867356)*x[0]+(0.76200283664)*x[1]**o+(0.134640432361)*x[1]
        arg[0,4,1,1]=(-0.815543712756)*x[0]**o+(0.0306226823724)*x[0]+(-0.0211231926242)*x[1]**o+(-0.718800228294)*x[1]
        arg[0,4,2,0]=(-0.980596107555)*x[0]**o+(-0.714237107759)*x[0]+(0.547986215851)*x[1]**o+(0.180131474534)*x[1]
        arg[0,4,2,1]=(-0.460780402239)*x[0]**o+(-0.483771113997)*x[0]+(-0.017140263255)*x[1]**o+(-0.127947184627)*x[1]
        arg[1,0,0,0]=(-0.0187443155707)*x[0]**o+(-0.490914415873)*x[0]+(-0.720022496093)*x[1]**o+(0.0265633770733)*x[1]
        arg[1,0,0,1]=(-0.634855765904)*x[0]**o+(0.345031954366)*x[0]+(-0.0532392363316)*x[1]**o+(0.983295524564)*x[1]
        arg[1,0,1,0]=(0.262182040476)*x[0]**o+(-0.883857758733)*x[0]+(0.880670543535)*x[1]**o+(0.193477695378)*x[1]
        arg[1,0,1,1]=(-0.0825976168371)*x[0]**o+(0.161838720337)*x[0]+(-0.0110587499089)*x[1]**o+(-0.634455402489)*x[1]
        arg[1,0,2,0]=(0.284221958342)*x[0]**o+(-0.948548729029)*x[0]+(0.430914344458)*x[1]**o+(-0.18563638241)*x[1]
        arg[1,0,2,1]=(-0.459558022898)*x[0]**o+(0.431907572866)*x[0]+(-0.0597669070218)*x[1]**o+(0.0694271372225)*x[1]
        arg[1,1,0,0]=(0.54071078447)*x[0]**o+(-0.244570994356)*x[0]+(0.00506219762373)*x[1]**o+(-0.0816378019186)*x[1]
        arg[1,1,0,1]=(-0.100752221379)*x[0]**o+(0.535738345566)*x[0]+(0.730325284553)*x[1]**o+(-0.885044748717)*x[1]
        arg[1,1,1,0]=(-0.0422460353138)*x[0]**o+(-0.986416505995)*x[0]+(-0.111766776997)*x[1]**o+(-0.842179226197)*x[1]
        arg[1,1,1,1]=(0.0690281593169)*x[0]**o+(-0.885054188663)*x[0]+(-0.680349314907)*x[1]**o+(0.00277190934345)*x[1]
        arg[1,1,2,0]=(0.502632146841)*x[0]**o+(-0.197332151163)*x[0]+(-0.0942257361149)*x[1]**o+(0.045171692749)*x[1]
        arg[1,1,2,1]=(0.427910466818)*x[0]**o+(-0.61355947216)*x[0]+(0.465128040517)*x[1]**o+(0.0874536555354)*x[1]
        arg[1,2,0,0]=(0.943819415417)*x[0]**o+(0.884020001869)*x[0]+(-0.769166027477)*x[1]**o+(-0.10032307068)*x[1]
        arg[1,2,0,1]=(0.0670224828564)*x[0]**o+(-0.645424481071)*x[0]+(-0.756361652678)*x[1]**o+(-0.274687916196)*x[1]
        arg[1,2,1,0]=(0.397836742229)*x[0]**o+(-0.708791058745)*x[0]+(0.920869587663)*x[1]**o+(-0.250037069918)*x[1]
        arg[1,2,1,1]=(0.722570486719)*x[0]**o+(-0.358082774936)*x[0]+(0.786344673969)*x[1]**o+(0.310591893637)*x[1]
        arg[1,2,2,0]=(-0.379404573553)*x[0]**o+(-0.969753501645)*x[0]+(-0.0745714915337)*x[1]**o+(-0.262943796911)*x[1]
        arg[1,2,2,1]=(0.817999459632)*x[0]**o+(0.731516351792)*x[0]+(-0.531411613325)*x[1]**o+(-0.824858061154)*x[1]
        arg[1,3,0,0]=(-0.206295444021)*x[0]**o+(-0.781973944918)*x[0]+(-0.173845545382)*x[1]**o+(-0.360122364573)*x[1]
        arg[1,3,0,1]=(-0.0604068005719)*x[0]**o+(0.238691416507)*x[0]+(-0.82262982031)*x[1]**o+(-0.893751755968)*x[1]
        arg[1,3,1,0]=(-0.686409669386)*x[0]**o+(0.711830599399)*x[0]+(0.0839688142166)*x[1]**o+(-0.99463449684)*x[1]
        arg[1,3,1,1]=(-0.0864745975077)*x[0]**o+(0.00068907895313)*x[0]+(-0.766810513534)*x[1]**o+(0.685236202954)*x[1]
        arg[1,3,2,0]=(0.0909338812695)*x[0]**o+(-0.709595654169)*x[0]+(-0.483428504499)*x[1]**o+(-0.0707341191006)*x[1]
        arg[1,3,2,1]=(-0.38342983855)*x[0]**o+(-0.0407824575865)*x[0]+(-0.201329040891)*x[1]**o+(0.0240007937141)*x[1]
        arg[1,4,0,0]=(0.512465803173)*x[0]**o+(-0.371898603608)*x[0]+(-0.79453595861)*x[1]**o+(0.340546986375)*x[1]
        arg[1,4,0,1]=(0.742383198442)*x[0]**o+(-0.603548189912)*x[0]+(-0.270277825546)*x[1]**o+(-0.525100913894)*x[1]
        arg[1,4,1,0]=(-0.693074384029)*x[0]**o+(-0.825037404844)*x[0]+(-0.626116177978)*x[1]**o+(0.628159342479)*x[1]
        arg[1,4,1,1]=(-0.595844943255)*x[0]**o+(0.250758194807)*x[0]+(0.504390052142)*x[1]**o+(-0.124588859702)*x[1]
        arg[1,4,2,0]=(-0.651727490997)*x[0]**o+(0.411138504151)*x[0]+(0.323769160331)*x[1]**o+(0.904379126671)*x[1]
        arg[1,4,2,1]=(-0.71307850178)*x[0]**o+(0.734964608119)*x[0]+(0.348693698506)*x[1]**o+(0.272270639772)*x[1]
        arg[2,0,0,0]=(0.558236567771)*x[0]**o+(0.516985445776)*x[0]+(0.0278915137861)*x[1]**o+(-0.268619977833)*x[1]
        arg[2,0,0,1]=(-0.871181709658)*x[0]**o+(-0.132251987744)*x[0]+(-0.215497073461)*x[1]**o+(0.508008002082)*x[1]
        arg[2,0,1,0]=(0.880795999572)*x[0]**o+(-0.231151166496)*x[0]+(-0.698846681617)*x[1]**o+(0.914483413825)*x[1]
        arg[2,0,1,1]=(0.684792911211)*x[0]**o+(-0.801408782989)*x[0]+(0.689713232168)*x[1]**o+(0.142847173797)*x[1]
        arg[2,0,2,0]=(-0.912566319945)*x[0]**o+(0.811138574657)*x[0]+(-0.884250879071)*x[1]**o+(-0.177206225477)*x[1]
        arg[2,0,2,1]=(-0.898648046678)*x[0]**o+(0.581554373688)*x[0]+(0.567642878443)*x[1]**o+(0.636312832751)*x[1]
        arg[2,1,0,0]=(-0.788853365648)*x[0]**o+(-0.135329327941)*x[0]+(0.180850147594)*x[1]**o+(0.742619029335)*x[1]
        arg[2,1,0,1]=(-0.824539980645)*x[0]**o+(0.857173393615)*x[0]+(0.850455264916)*x[1]**o+(0.975422322464)*x[1]
        arg[2,1,1,0]=(0.462293517317)*x[0]**o+(-0.173835273794)*x[0]+(-0.800072514276)*x[1]**o+(0.140214114627)*x[1]
        arg[2,1,1,1]=(-0.187890814293)*x[0]**o+(0.534472364495)*x[0]+(-0.799640665571)*x[1]**o+(0.301287955302)*x[1]
        arg[2,1,2,0]=(0.278901492946)*x[0]**o+(0.620544326886)*x[0]+(-0.705702443752)*x[1]**o+(-0.489690636328)*x[1]
        arg[2,1,2,1]=(-0.926772900235)*x[0]**o+(0.134654742845)*x[0]+(0.356561975337)*x[1]**o+(-0.919999359857)*x[1]
        arg[2,2,0,0]=(-0.675956301416)*x[0]**o+(0.149582589205)*x[0]+(0.186440827385)*x[1]**o+(0.88381610446)*x[1]
        arg[2,2,0,1]=(-0.697267962915)*x[0]**o+(0.991366596424)*x[0]+(0.310290094618)*x[1]**o+(-0.881771505066)*x[1]
        arg[2,2,1,0]=(0.506027358474)*x[0]**o+(0.294136137354)*x[0]+(0.732003937836)*x[1]**o+(0.633469131504)*x[1]
        arg[2,2,1,1]=(-0.427146039728)*x[0]**o+(-0.975737063903)*x[0]+(-0.934913653772)*x[1]**o+(-0.662568325005)*x[1]
        arg[2,2,2,0]=(0.477849916621)*x[0]**o+(-0.798883883367)*x[0]+(-0.154217684441)*x[1]**o+(-0.175817547912)*x[1]
        arg[2,2,2,1]=(0.935192620387)*x[0]**o+(-0.193802588249)*x[0]+(0.406181949213)*x[1]**o+(0.20422475396)*x[1]
        arg[2,3,0,0]=(0.965944278856)*x[0]**o+(-0.736175411725)*x[0]+(0.92404347386)*x[1]**o+(0.273152788392)*x[1]
        arg[2,3,0,1]=(-0.606944412413)*x[0]**o+(0.692555413476)*x[0]+(0.812085676275)*x[1]**o+(-0.0809920218886)*x[1]
        arg[2,3,1,0]=(-0.314634777406)*x[0]**o+(0.750181618592)*x[0]+(0.0554739779642)*x[1]**o+(-0.333788326726)*x[1]
        arg[2,3,1,1]=(0.85610815703)*x[0]**o+(0.910096005524)*x[0]+(-0.227092814843)*x[1]**o+(-0.0849532598914)*x[1]
        arg[2,3,2,0]=(0.735064128504)*x[0]**o+(-0.253919528943)*x[0]+(0.8457463307)*x[1]**o+(0.865727345925)*x[1]
        arg[2,3,2,1]=(0.520607989389)*x[0]**o+(0.569210317105)*x[0]+(0.500780081677)*x[1]**o+(-0.787630914351)*x[1]
        arg[2,4,0,0]=(0.179774694296)*x[0]**o+(-0.890872128607)*x[0]+(-0.185758177247)*x[1]**o+(0.888231734606)*x[1]
        arg[2,4,0,1]=(0.059863462227)*x[0]**o+(-0.0796463123948)*x[0]+(-0.104564520504)*x[1]**o+(-0.336182528617)*x[1]
        arg[2,4,1,0]=(-0.636828275967)*x[0]**o+(0.156894572218)*x[0]+(-0.884260096115)*x[1]**o+(-0.967431438795)*x[1]
        arg[2,4,1,1]=(-0.958168110389)*x[0]**o+(-0.0424772631477)*x[0]+(0.512363473316)*x[1]**o+(-0.716729419668)*x[1]
        arg[2,4,2,0]=(0.19736654467)*x[0]**o+(-0.288723518002)*x[0]+(0.243674007033)*x[1]**o+(0.911210601142)*x[1]
        arg[2,4,2,1]=(-0.847493343055)*x[0]**o+(0.115003265593)*x[0]+(0.360071936714)*x[1]**o+(-0.0265385061141)*x[1]
        arg[3,0,0,0]=(-0.510550590836)*x[0]**o+(-0.266174606631)*x[0]+(0.771580708159)*x[1]**o+(-0.770554374259)*x[1]
        arg[3,0,0,1]=(-0.65795024738)*x[0]**o+(-0.83394628525)*x[0]+(0.812461325452)*x[1]**o+(0.191593733482)*x[1]
        arg[3,0,1,0]=(0.136878718548)*x[0]**o+(-0.000345061692795)*x[0]+(-0.0780280205008)*x[1]**o+(0.173925537373)*x[1]
        arg[3,0,1,1]=(0.584079470149)*x[0]**o+(0.138674493948)*x[0]+(-0.563553909861)*x[1]**o+(-0.254304838008)*x[1]
        arg[3,0,2,0]=(0.326192887554)*x[0]**o+(0.993615980303)*x[0]+(-0.26995279693)*x[1]**o+(0.216112615287)*x[1]
        arg[3,0,2,1]=(0.684786021819)*x[0]**o+(0.158934017292)*x[0]+(-0.457448199191)*x[1]**o+(0.644722228163)*x[1]
        arg[3,1,0,0]=(-0.198252162262)*x[0]**o+(-0.821832378445)*x[0]+(0.831807441916)*x[1]**o+(-0.70643670471)*x[1]
        arg[3,1,0,1]=(0.361698759016)*x[0]**o+(0.556511107101)*x[0]+(0.892219700601)*x[1]**o+(-0.861598520361)*x[1]
        arg[3,1,1,0]=(-0.107946706634)*x[0]**o+(-0.910191934756)*x[0]+(0.502507607871)*x[1]**o+(0.649089810516)*x[1]
        arg[3,1,1,1]=(-0.92959760672)*x[0]**o+(0.950682520961)*x[0]+(0.103321004187)*x[1]**o+(-0.631562667815)*x[1]
        arg[3,1,2,0]=(0.858565242025)*x[0]**o+(-0.106810606075)*x[0]+(0.236377425048)*x[1]**o+(-0.296320858145)*x[1]
        arg[3,1,2,1]=(0.148558873625)*x[0]**o+(0.458488334082)*x[0]+(-0.0172824905523)*x[1]**o+(-0.321212766581)*x[1]
        arg[3,2,0,0]=(-0.239733890336)*x[0]**o+(0.558819103697)*x[0]+(-0.151683129885)*x[1]**o+(0.153995780645)*x[1]
        arg[3,2,0,1]=(-0.949100270937)*x[0]**o+(0.0687171499139)*x[0]+(0.968028021831)*x[1]**o+(-0.935600499345)*x[1]
        arg[3,2,1,0]=(-0.739823767237)*x[0]**o+(0.417825150389)*x[0]+(0.318612170707)*x[1]**o+(0.627557337421)*x[1]
        arg[3,2,1,1]=(-0.965142167324)*x[0]**o+(0.981446854383)*x[0]+(0.266191185951)*x[1]**o+(-0.581303142841)*x[1]
        arg[3,2,2,0]=(-0.00312606993817)*x[0]**o+(-0.191885030828)*x[0]+(-0.342525068561)*x[1]**o+(-0.953710538626)*x[1]
        arg[3,2,2,1]=(-0.689127187476)*x[0]**o+(-0.107190366063)*x[0]+(-0.0556716250594)*x[1]**o+(-0.609462166754)*x[1]
        arg[3,3,0,0]=(-0.765237197853)*x[0]**o+(-0.0707283944117)*x[0]+(0.891789872287)*x[1]**o+(-0.255939791858)*x[1]
        arg[3,3,0,1]=(-0.367658362103)*x[0]**o+(0.27406309422)*x[0]+(0.866288683947)*x[1]**o+(0.403194030287)*x[1]
        arg[3,3,1,0]=(-0.362795000993)*x[0]**o+(-0.498774415158)*x[0]+(0.0726808331583)*x[1]**o+(-0.871066791136)*x[1]
        arg[3,3,1,1]=(-0.177149059814)*x[0]**o+(0.287463753612)*x[0]+(0.101726559941)*x[1]**o+(0.561797204517)*x[1]
        arg[3,3,2,0]=(0.770615121039)*x[0]**o+(-0.503402148251)*x[0]+(0.319753007634)*x[1]**o+(0.586868894261)*x[1]
        arg[3,3,2,1]=(0.924125904724)*x[0]**o+(-0.508988920423)*x[0]+(-0.866752504824)*x[1]**o+(0.518544700652)*x[1]
        arg[3,4,0,0]=(-0.973660291124)*x[0]**o+(0.746869369283)*x[0]+(-0.634097171844)*x[1]**o+(0.723638944736)*x[1]
        arg[3,4,0,1]=(-0.599366869066)*x[0]**o+(-0.59893829688)*x[0]+(0.124483401866)*x[1]**o+(-0.924820315792)*x[1]
        arg[3,4,1,0]=(-0.707374787729)*x[0]**o+(-0.494748755022)*x[0]+(0.391752209698)*x[1]**o+(0.149710640116)*x[1]
        arg[3,4,1,1]=(0.329587748119)*x[0]**o+(-0.788289803764)*x[0]+(-0.145100163626)*x[1]**o+(0.141493810237)*x[1]
        arg[3,4,2,0]=(0.17496644337)*x[0]**o+(-0.270242104589)*x[0]+(0.996684379179)*x[1]**o+(-0.751697440046)*x[1]
        arg[3,4,2,1]=(0.79046837395)*x[0]**o+(0.568394473062)*x[0]+(0.359419854016)*x[1]**o+(0.630686948244)*x[1]
        ref[0,0,0,0]=(0.306552070969)/(o+1.)+(0.157870653629)
        ref[0,0,0,1]=(0.0606500807631)/(o+1.)+(0.0612449724389)
        ref[0,0,1,0]=(0.639564070889)/(o+1.)+(0.666341832555)
        ref[0,0,1,1]=(0.671248593147)/(o+1.)+(-0.528628572255)
        ref[0,0,2,0]=(0.0735824774293)/(o+1.)+(-0.474736472501)
        ref[0,0,2,1]=(-0.9070527228)/(o+1.)+(-0.197350261427)
        ref[0,1,0,0]=(1.30449671875)/(o+1.)+(0.106123067554)
        ref[0,1,0,1]=(-0.420072850657)/(o+1.)+(-0.407896954678)
        ref[0,1,1,0]=(1.44174379356)/(o+1.)+(-0.809395886627)
        ref[0,1,1,1]=(0.155435385684)/(o+1.)+(0.627148132992)
        ref[0,1,2,0]=(-0.137204404518)/(o+1.)+(0.207874473889)
        ref[0,1,2,1]=(-0.289709410261)/(o+1.)+(0.204293345671)
        ref[0,2,0,0]=(1.26583912644)/(o+1.)+(0.498805698213)
        ref[0,2,0,1]=(1.17560298985)/(o+1.)+(0.931845120886)
        ref[0,2,1,0]=(-0.344986302901)/(o+1.)+(0.220341743892)
        ref[0,2,1,1]=(-1.94726093397)/(o+1.)+(0.206668541494)
        ref[0,2,2,0]=(-0.91808772513)/(o+1.)+(-0.0252951560363)
        ref[0,2,2,1]=(-1.02863406132)/(o+1.)+(-0.20183443794)
        ref[0,3,0,0]=(-1.04834664255)/(o+1.)+(0.00604709752155)
        ref[0,3,0,1]=(0.2902816262)/(o+1.)+(0.514361224456)
        ref[0,3,1,0]=(-1.31568170019)/(o+1.)+(0.883603667426)
        ref[0,3,1,1]=(-0.208650456965)/(o+1.)+(0.889841082216)
        ref[0,3,2,0]=(-0.741632466554)/(o+1.)+(0.0759327240318)
        ref[0,3,2,1]=(1.48653223703)/(o+1.)+(-0.854483988026)
        ref[0,4,0,0]=(0.333816089225)/(o+1.)+(-0.562947445201)
        ref[0,4,0,1]=(0.130634932061)/(o+1.)+(0.199922965378)
        ref[0,4,1,0]=(1.27534940781)/(o+1.)+(0.158640649858)
        ref[0,4,1,1]=(-0.83666690538)/(o+1.)+(-0.344088772961)
        ref[0,4,2,0]=(-0.432609891704)/(o+1.)+(-0.267052816612)
        ref[0,4,2,1]=(-0.477920665494)/(o+1.)+(-0.305859149312)
        ref[1,0,0,0]=(-0.738766811663)/(o+1.)+(-0.2321755194)
        ref[1,0,0,1]=(-0.688095002235)/(o+1.)+(0.664163739465)
        ref[1,0,1,0]=(1.14285258401)/(o+1.)+(-0.345190031677)
        ref[1,0,1,1]=(-0.093656366746)/(o+1.)+(-0.236308341076)
        ref[1,0,2,0]=(0.7151363028)/(o+1.)+(-0.56709255572)
        ref[1,0,2,1]=(-0.51932492992)/(o+1.)+(0.250667355044)
        ref[1,1,0,0]=(0.545772982093)/(o+1.)+(-0.163104398138)
        ref[1,1,0,1]=(0.629573063174)/(o+1.)+(-0.174653201576)
        ref[1,1,1,0]=(-0.154012812311)/(o+1.)+(-0.914297866096)
        ref[1,1,1,1]=(-0.61132115559)/(o+1.)+(-0.44114113966)
        ref[1,1,2,0]=(0.408406410726)/(o+1.)+(-0.076080229207)
        ref[1,1,2,1]=(0.893038507335)/(o+1.)+(-0.263052908312)
        ref[1,2,0,0]=(0.17465338794)/(o+1.)+(0.391848465594)
        ref[1,2,0,1]=(-0.689339169821)/(o+1.)+(-0.460056198634)
        ref[1,2,1,0]=(1.31870632989)/(o+1.)+(-0.479414064332)
        ref[1,2,1,1]=(1.50891516069)/(o+1.)+(-0.0237454406495)
        ref[1,2,2,0]=(-0.453976065087)/(o+1.)+(-0.616348649278)
        ref[1,2,2,1]=(0.286587846307)/(o+1.)+(-0.0466708546809)
        ref[1,3,0,0]=(-0.380140989403)/(o+1.)+(-0.571048154745)
        ref[1,3,0,1]=(-0.883036620882)/(o+1.)+(-0.32753016973)
        ref[1,3,1,0]=(-0.60244085517)/(o+1.)+(-0.141401948721)
        ref[1,3,1,1]=(-0.853285111042)/(o+1.)+(0.342962640954)
        ref[1,3,2,0]=(-0.39249462323)/(o+1.)+(-0.390164886635)
        ref[1,3,2,1]=(-0.584758879441)/(o+1.)+(-0.00839083193621)
        ref[1,4,0,0]=(-0.282070155438)/(o+1.)+(-0.0156758086169)
        ref[1,4,0,1]=(0.472105372896)/(o+1.)+(-0.564324551903)
        ref[1,4,1,0]=(-1.31919056201)/(o+1.)+(-0.0984390311826)
        ref[1,4,1,1]=(-0.0914548911124)/(o+1.)+(0.0630846675525)
        ref[1,4,2,0]=(-0.327958330666)/(o+1.)+(0.657758815411)
        ref[1,4,2,1]=(-0.364384803274)/(o+1.)+(0.503617623946)
        ref[2,0,0,0]=(0.586128081557)/(o+1.)+(0.124182733972)
        ref[2,0,0,1]=(-1.08667878312)/(o+1.)+(0.187878007169)
        ref[2,0,1,0]=(0.181949317955)/(o+1.)+(0.341666123665)
        ref[2,0,1,1]=(1.37450614338)/(o+1.)+(-0.329280804596)
        ref[2,0,2,0]=(-1.79681719902)/(o+1.)+(0.31696617459)
        ref[2,0,2,1]=(-0.331005168235)/(o+1.)+(0.608933603219)
        ref[2,1,0,0]=(-0.608003218053)/(o+1.)+(0.303644850697)
        ref[2,1,0,1]=(0.0259152842714)/(o+1.)+(0.916297858039)
        ref[2,1,1,0]=(-0.33777899696)/(o+1.)+(-0.0168105795836)
        ref[2,1,1,1]=(-0.987531479864)/(o+1.)+(0.417880159899)
        ref[2,1,2,0]=(-0.426800950806)/(o+1.)+(0.0654268452787)
        ref[2,1,2,1]=(-0.570210924898)/(o+1.)+(-0.392672308506)
        ref[2,2,0,0]=(-0.489515474032)/(o+1.)+(0.516699346832)
        ref[2,2,0,1]=(-0.386977868298)/(o+1.)+(0.054797545679)
        ref[2,2,1,0]=(1.23803129631)/(o+1.)+(0.463802634429)
        ref[2,2,1,1]=(-1.3620596935)/(o+1.)+(-0.819152694454)
        ref[2,2,2,0]=(0.32363223218)/(o+1.)+(-0.48735071564)
        ref[2,2,2,1]=(1.3413745696)/(o+1.)+(0.00521108285542)
        ref[2,3,0,0]=(1.88998775272)/(o+1.)+(-0.231511311666)
        ref[2,3,0,1]=(0.205141263861)/(o+1.)+(0.305781695794)
        ref[2,3,1,0]=(-0.259160799442)/(o+1.)+(0.208196645933)
        ref[2,3,1,1]=(0.629015342187)/(o+1.)+(0.412571372816)
        ref[2,3,2,0]=(1.5808104592)/(o+1.)+(0.305903908491)
        ref[2,3,2,1]=(1.02138807107)/(o+1.)+(-0.109210298623)
        ref[2,4,0,0]=(-0.00598348295078)/(o+1.)+(-0.00132019700044)
        ref[2,4,0,1]=(-0.0447010582767)/(o+1.)+(-0.207914420506)
        ref[2,4,1,0]=(-1.52108837208)/(o+1.)+(-0.405268433289)
        ref[2,4,1,1]=(-0.445804637073)/(o+1.)+(-0.379603341408)
        ref[2,4,2,0]=(0.441040551703)/(o+1.)+(0.31124354157)
        ref[2,4,2,1]=(-0.487421406342)/(o+1.)+(0.0442323797396)
        ref[3,0,0,0]=(0.261030117323)/(o+1.)+(-0.518364490445)
        ref[3,0,0,1]=(0.154511078072)/(o+1.)+(-0.321176275884)
        ref[3,0,1,0]=(0.0588506980476)/(o+1.)+(0.0867902378403)
        ref[3,0,1,1]=(0.0205255602871)/(o+1.)+(-0.0578151720304)
        ref[3,0,2,0]=(0.0562400906237)/(o+1.)+(0.604864297795)
        ref[3,0,2,1]=(0.227337822629)/(o+1.)+(0.401828122727)
        ref[3,1,0,0]=(0.633555279654)/(o+1.)+(-0.764134541578)
        ref[3,1,0,1]=(1.25391845962)/(o+1.)+(-0.15254370663)
        ref[3,1,1,0]=(0.394560901238)/(o+1.)+(-0.13055106212)
        ref[3,1,1,1]=(-0.826276602534)/(o+1.)+(0.159559926573)
        ref[3,1,2,0]=(1.09494266707)/(o+1.)+(-0.20156573211)
        ref[3,1,2,1]=(0.131276383073)/(o+1.)+(0.0686377837506)
        ref[3,2,0,0]=(-0.391417020221)/(o+1.)+(0.356407442171)
        ref[3,2,0,1]=(0.0189277508933)/(o+1.)+(-0.433441674716)
        ref[3,2,1,0]=(-0.42121159653)/(o+1.)+(0.522691243905)
        ref[3,2,1,1]=(-0.698950981373)/(o+1.)+(0.200071855771)
        ref[3,2,2,0]=(-0.345651138499)/(o+1.)+(-0.572797784727)
        ref[3,2,2,1]=(-0.744798812536)/(o+1.)+(-0.358326266408)
        ref[3,3,0,0]=(0.126552674434)/(o+1.)+(-0.163334093135)
        ref[3,3,0,1]=(0.498630321844)/(o+1.)+(0.338628562253)
        ref[3,3,1,0]=(-0.290114167835)/(o+1.)+(-0.684920603147)
        ref[3,3,1,1]=(-0.0754224998727)/(o+1.)+(0.424630479064)
        ref[3,3,2,0]=(1.09036812867)/(o+1.)+(0.0417333730049)
        ref[3,3,2,1]=(0.0573733998999)/(o+1.)+(0.00477789011464)
        ref[3,4,0,0]=(-1.60775746297)/(o+1.)+(0.73525415701)
        ref[3,4,0,1]=(-0.4748834672)/(o+1.)+(-0.761879306336)
        ref[3,4,1,0]=(-0.31562257803)/(o+1.)+(-0.172519057453)
        ref[3,4,1,1]=(0.184487584493)/(o+1.)+(-0.323397996764)
        ref[3,4,2,0]=(1.17165082255)/(o+1.)+(-0.510969772317)
        ref[3,4,2,1]=(1.14988822797)/(o+1.)+(0.599540710653)
      else:
        arg[0,0,0,0]=(-0.517412938774)*x[0]**o+(0.824804313111)*x[0]+(-0.681706563374)*x[1]**o+(-0.259540440272)*x[1]+(0.876735283937)*x[2]**o+(0.307072859807)*x[2]
        arg[0,0,0,1]=(-0.545981202251)*x[0]**o+(-0.530080933212)*x[0]+(-0.339289111599)*x[1]**o+(-0.726311605088)*x[1]+(0.334837705444)*x[2]**o+(-0.422667311064)*x[2]
        arg[0,0,1,0]=(-0.94261387328)*x[0]**o+(-0.315825326734)*x[0]+(0.542502801277)*x[1]**o+(-0.927525943646)*x[1]+(-0.802118238213)*x[2]**o+(-0.106512314706)*x[2]
        arg[0,0,1,1]=(-0.877940099286)*x[0]**o+(-0.801256557869)*x[0]+(-0.155983792242)*x[1]**o+(0.200547191496)*x[1]+(-0.960553133432)*x[2]**o+(0.968862792367)*x[2]
        arg[0,0,2,0]=(-0.870342547946)*x[0]**o+(-0.956447402661)*x[0]+(-0.506797803004)*x[1]**o+(0.342841675107)*x[1]+(0.528354159344)*x[2]**o+(-0.824546729564)*x[2]
        arg[0,0,2,1]=(-0.411899137296)*x[0]**o+(0.133736458559)*x[0]+(0.852726131619)*x[1]**o+(0.692795229813)*x[1]+(0.047372576403)*x[2]**o+(0.810485765945)*x[2]
        arg[0,1,0,0]=(0.961758233254)*x[0]**o+(-0.606307924964)*x[0]+(0.0612696412665)*x[1]**o+(-0.947823541802)*x[1]+(-0.113021104365)*x[2]**o+(-0.00802440322155)*x[2]
        arg[0,1,0,1]=(-0.684097052467)*x[0]**o+(0.467140835537)*x[0]+(-0.0742537091205)*x[1]**o+(0.541641916143)*x[1]+(0.987491548505)*x[2]**o+(-0.558380262255)*x[2]
        arg[0,1,1,0]=(-0.901821988855)*x[0]**o+(0.34443714654)*x[0]+(-0.682552654793)*x[1]**o+(0.398276086461)*x[1]+(0.334644264965)*x[2]**o+(0.21454845251)*x[2]
        arg[0,1,1,1]=(0.929614060295)*x[0]**o+(-0.322609307947)*x[0]+(-0.426962200724)*x[1]**o+(0.464659992564)*x[1]+(-0.13983863674)*x[2]**o+(-0.681648817325)*x[2]
        arg[0,1,2,0]=(0.504166508914)*x[0]**o+(-0.978562859041)*x[0]+(-0.769650404806)*x[1]**o+(0.443500137153)*x[1]+(0.969669965936)*x[2]**o+(0.401742207403)*x[2]
        arg[0,1,2,1]=(-0.478561324822)*x[0]**o+(-0.772019838834)*x[0]+(-0.779194844829)*x[1]**o+(-0.747980400065)*x[1]+(0.827885608192)*x[2]**o+(0.135571151528)*x[2]
        arg[0,2,0,0]=(-0.210426221052)*x[0]**o+(-0.464938141903)*x[0]+(-0.776844066203)*x[1]**o+(0.267244555705)*x[1]+(-0.773135929429)*x[2]**o+(-0.00846060888954)*x[2]
        arg[0,2,0,1]=(0.0945858334531)*x[0]**o+(-0.492400895713)*x[0]+(-0.128564669727)*x[1]**o+(-0.0951458189983)*x[1]+(-0.54426715378)*x[2]**o+(0.106545346072)*x[2]
        arg[0,2,1,0]=(-0.482424571392)*x[0]**o+(0.0431208155652)*x[0]+(0.557138601636)*x[1]**o+(-0.725564282551)*x[1]+(0.9081107419)*x[2]**o+(0.0179201585352)*x[2]
        arg[0,2,1,1]=(-0.555731853052)*x[0]**o+(0.447137254671)*x[0]+(0.820958717703)*x[1]**o+(-0.580251434504)*x[1]+(0.227884622965)*x[2]**o+(0.34460189453)*x[2]
        arg[0,2,2,0]=(-0.25868983898)*x[0]**o+(0.759165881486)*x[0]+(-0.499903058341)*x[1]**o+(0.896471093639)*x[1]+(0.351955520633)*x[2]**o+(0.16566756566)*x[2]
        arg[0,2,2,1]=(-0.37859354794)*x[0]**o+(0.925387188779)*x[0]+(0.0111055293719)*x[1]**o+(-0.311998642154)*x[1]+(0.441009182046)*x[2]**o+(0.0968247918271)*x[2]
        arg[0,3,0,0]=(-0.461665900641)*x[0]**o+(0.812002051396)*x[0]+(-0.767456852027)*x[1]**o+(0.809069781575)*x[1]+(0.690416953885)*x[2]**o+(0.171662369216)*x[2]
        arg[0,3,0,1]=(-0.804811324639)*x[0]**o+(-0.0962711523954)*x[0]+(0.980130059825)*x[1]**o+(0.98279180603)*x[1]+(-0.0614353559054)*x[2]**o+(0.645637153538)*x[2]
        arg[0,3,1,0]=(0.97878415369)*x[0]**o+(-0.676321251698)*x[0]+(0.916075195717)*x[1]**o+(0.244509683371)*x[1]+(-0.939563193145)*x[2]**o+(-0.866506108913)*x[2]
        arg[0,3,1,1]=(-0.676965895162)*x[0]**o+(-0.124355338519)*x[0]+(0.680151101195)*x[1]**o+(-0.839412433156)*x[1]+(0.434088587223)*x[2]**o+(-0.275286048166)*x[2]
        arg[0,3,2,0]=(0.414822536626)*x[0]**o+(0.493836498194)*x[0]+(0.967077097745)*x[1]**o+(-0.0115301157869)*x[1]+(-0.307994657929)*x[2]**o+(-0.967766608312)*x[2]
        arg[0,3,2,1]=(-0.34064670348)*x[0]**o+(-0.975121872905)*x[0]+(0.0839457874766)*x[1]**o+(0.0995106978636)*x[1]+(-0.533837237516)*x[2]**o+(0.771268044302)*x[2]
        arg[0,4,0,0]=(0.223197544637)*x[0]**o+(0.712738871576)*x[0]+(-0.386972075747)*x[1]**o+(0.289829507744)*x[1]+(0.153418793839)*x[2]**o+(0.876087048492)*x[2]
        arg[0,4,0,1]=(0.696750948653)*x[0]**o+(0.466905335953)*x[0]+(-0.611476041907)*x[1]**o+(-0.939902457901)*x[1]+(-0.752675432265)*x[2]**o+(-0.869502717044)*x[2]
        arg[0,4,1,0]=(-0.465562201723)*x[0]**o+(-0.388716352051)*x[0]+(0.260044093503)*x[1]**o+(0.919379163294)*x[1]+(-0.902453394797)*x[2]**o+(0.378318744505)*x[2]
        arg[0,4,1,1]=(0.121990842946)*x[0]**o+(-0.691267894153)*x[0]+(-0.703960568755)*x[1]**o+(-0.582404896428)*x[1]+(0.57560180036)*x[2]**o+(0.537099448589)*x[2]
        arg[0,4,2,0]=(-0.495452196513)*x[0]**o+(-0.939389935669)*x[0]+(-0.237205717598)*x[1]**o+(0.61376516207)*x[1]+(-0.506082738728)*x[2]**o+(0.407704148542)*x[2]
        arg[0,4,2,1]=(0.450976180771)*x[0]**o+(-0.00739258898437)*x[0]+(0.210130848093)*x[1]**o+(0.624639007009)*x[1]+(0.173795671727)*x[2]**o+(0.508269170853)*x[2]
        arg[1,0,0,0]=(0.609685942177)*x[0]**o+(0.609299494609)*x[0]+(0.321046388121)*x[1]**o+(0.598197189498)*x[1]+(0.536092177618)*x[2]**o+(0.576427431157)*x[2]
        arg[1,0,0,1]=(0.412500706234)*x[0]**o+(0.905942239017)*x[0]+(-0.974980343169)*x[1]**o+(-0.131245997607)*x[1]+(0.745809068018)*x[2]**o+(0.543470258153)*x[2]
        arg[1,0,1,0]=(-0.903032545947)*x[0]**o+(-0.778693321051)*x[0]+(0.399342164105)*x[1]**o+(0.478212896767)*x[1]+(0.254596269802)*x[2]**o+(0.70308382001)*x[2]
        arg[1,0,1,1]=(-0.281529506793)*x[0]**o+(0.74166036027)*x[0]+(0.629713747725)*x[1]**o+(-0.275164746597)*x[1]+(0.37523909979)*x[2]**o+(-0.878026427978)*x[2]
        arg[1,0,2,0]=(0.236027764061)*x[0]**o+(-0.816910805083)*x[0]+(0.0782011624185)*x[1]**o+(-0.644004245666)*x[1]+(0.293586462602)*x[2]**o+(0.557045894695)*x[2]
        arg[1,0,2,1]=(0.868061727269)*x[0]**o+(0.53329318943)*x[0]+(0.367603915235)*x[1]**o+(0.89280277925)*x[1]+(-0.316595424948)*x[2]**o+(0.593830565147)*x[2]
        arg[1,1,0,0]=(0.886950595996)*x[0]**o+(-0.290323112286)*x[0]+(-0.863347030089)*x[1]**o+(0.545096994541)*x[1]+(0.338200944281)*x[2]**o+(-0.79898415048)*x[2]
        arg[1,1,0,1]=(0.103761462006)*x[0]**o+(-0.046916161864)*x[0]+(0.0312549239467)*x[1]**o+(0.214045118984)*x[1]+(-0.523161766402)*x[2]**o+(0.714207028317)*x[2]
        arg[1,1,1,0]=(0.494943183987)*x[0]**o+(-0.0543096386006)*x[0]+(-0.263875120588)*x[1]**o+(0.776031489925)*x[1]+(0.881224752145)*x[2]**o+(-0.650038283228)*x[2]
        arg[1,1,1,1]=(0.0871887540513)*x[0]**o+(-0.953848841349)*x[0]+(-0.223751459589)*x[1]**o+(0.227400823993)*x[1]+(-0.224200553009)*x[2]**o+(0.594012970899)*x[2]
        arg[1,1,2,0]=(-0.443116951346)*x[0]**o+(0.891806149626)*x[0]+(0.94731530917)*x[1]**o+(0.739499894809)*x[1]+(-0.92613053703)*x[2]**o+(0.464883087528)*x[2]
        arg[1,1,2,1]=(-0.149831145559)*x[0]**o+(0.913999704597)*x[0]+(0.76082978152)*x[1]**o+(-0.441535347622)*x[1]+(0.902841586936)*x[2]**o+(-0.509022004948)*x[2]
        arg[1,2,0,0]=(0.508230188489)*x[0]**o+(0.241669921231)*x[0]+(-0.802811365056)*x[1]**o+(0.298424305948)*x[1]+(-0.949407454532)*x[2]**o+(-0.131426880214)*x[2]
        arg[1,2,0,1]=(0.65372313274)*x[0]**o+(0.195891642771)*x[0]+(-0.160845429219)*x[1]**o+(0.500815823724)*x[1]+(0.875785249339)*x[2]**o+(0.325024339211)*x[2]
        arg[1,2,1,0]=(0.908405682536)*x[0]**o+(-0.146675166477)*x[0]+(0.24033815873)*x[1]**o+(-0.62814219771)*x[1]+(-0.512365788351)*x[2]**o+(-0.103817191415)*x[2]
        arg[1,2,1,1]=(0.824305237501)*x[0]**o+(0.175753992729)*x[0]+(0.992502114126)*x[1]**o+(0.973110627555)*x[1]+(0.355352002177)*x[2]**o+(0.112252263955)*x[2]
        arg[1,2,2,0]=(-0.485996285003)*x[0]**o+(-0.532551123887)*x[0]+(0.708205121405)*x[1]**o+(-0.803269038441)*x[1]+(-0.232049124534)*x[2]**o+(0.55437652866)*x[2]
        arg[1,2,2,1]=(0.551934776224)*x[0]**o+(0.727629664538)*x[0]+(0.197393511049)*x[1]**o+(-0.146619812174)*x[1]+(0.316592306936)*x[2]**o+(0.0263881493809)*x[2]
        arg[1,3,0,0]=(-0.918554166254)*x[0]**o+(-0.839150758884)*x[0]+(0.436713305329)*x[1]**o+(-0.410131731387)*x[1]+(0.321246028791)*x[2]**o+(0.706389384376)*x[2]
        arg[1,3,0,1]=(-0.870168753295)*x[0]**o+(-0.00473364206025)*x[0]+(0.594648540734)*x[1]**o+(-0.924887214906)*x[1]+(0.58858894515)*x[2]**o+(0.0152152159566)*x[2]
        arg[1,3,1,0]=(0.64005604234)*x[0]**o+(0.308265334836)*x[0]+(0.321645661286)*x[1]**o+(0.670017614875)*x[1]+(0.712764136703)*x[2]**o+(0.458986964342)*x[2]
        arg[1,3,1,1]=(-0.201376578944)*x[0]**o+(-0.271793865118)*x[0]+(-0.798181331582)*x[1]**o+(0.690100877945)*x[1]+(0.600350295622)*x[2]**o+(0.341066858436)*x[2]
        arg[1,3,2,0]=(0.459832113137)*x[0]**o+(0.0566780273877)*x[0]+(0.221723079112)*x[1]**o+(-0.681001728734)*x[1]+(0.265505135729)*x[2]**o+(-0.389365671367)*x[2]
        arg[1,3,2,1]=(0.715444384493)*x[0]**o+(-0.705697520786)*x[0]+(0.945208081248)*x[1]**o+(0.0233171867808)*x[1]+(0.740156166802)*x[2]**o+(-0.144314337012)*x[2]
        arg[1,4,0,0]=(-0.490162545806)*x[0]**o+(0.542981268396)*x[0]+(-0.841471465564)*x[1]**o+(0.79091798393)*x[1]+(0.245598142884)*x[2]**o+(0.4740296042)*x[2]
        arg[1,4,0,1]=(-0.504504615322)*x[0]**o+(0.45494067251)*x[0]+(-0.612427248728)*x[1]**o+(0.0880340051743)*x[1]+(0.645906227711)*x[2]**o+(0.772038416622)*x[2]
        arg[1,4,1,0]=(0.945538632492)*x[0]**o+(0.98519603491)*x[0]+(-0.245495451859)*x[1]**o+(-0.70732864026)*x[1]+(0.378733662069)*x[2]**o+(0.579554415018)*x[2]
        arg[1,4,1,1]=(0.886659064398)*x[0]**o+(-0.491870874438)*x[0]+(0.751814742427)*x[1]**o+(0.430589443767)*x[1]+(-0.729193959143)*x[2]**o+(-0.699344758538)*x[2]
        arg[1,4,2,0]=(-0.4190140481)*x[0]**o+(0.84509612452)*x[0]+(-0.263608911857)*x[1]**o+(0.693648282371)*x[1]+(-0.144463932166)*x[2]**o+(0.384293689266)*x[2]
        arg[1,4,2,1]=(-0.266973780787)*x[0]**o+(0.613988367707)*x[0]+(0.686955199789)*x[1]**o+(-0.301585818785)*x[1]+(-0.574601202558)*x[2]**o+(-0.430718616071)*x[2]
        arg[2,0,0,0]=(-0.117390727803)*x[0]**o+(-0.25850139145)*x[0]+(0.324909086031)*x[1]**o+(-0.593112241553)*x[1]+(-0.51016768177)*x[2]**o+(0.392299415334)*x[2]
        arg[2,0,0,1]=(0.290478531696)*x[0]**o+(0.758153023626)*x[0]+(0.531003993361)*x[1]**o+(-0.112481510374)*x[1]+(0.950222158203)*x[2]**o+(-0.742480526204)*x[2]
        arg[2,0,1,0]=(-0.754820921225)*x[0]**o+(-0.408482044077)*x[0]+(-0.229697426486)*x[1]**o+(-0.983010424713)*x[1]+(-0.286618604257)*x[2]**o+(-0.489034108451)*x[2]
        arg[2,0,1,1]=(0.96775757822)*x[0]**o+(0.797765725721)*x[0]+(-0.90936176648)*x[1]**o+(-0.372755885193)*x[1]+(-0.144727874143)*x[2]**o+(0.341624944551)*x[2]
        arg[2,0,2,0]=(0.672779422278)*x[0]**o+(-0.936102224964)*x[0]+(-0.763620352857)*x[1]**o+(0.265802509744)*x[1]+(0.275798875047)*x[2]**o+(0.235397520942)*x[2]
        arg[2,0,2,1]=(0.192866248041)*x[0]**o+(0.656863606616)*x[0]+(-0.902796948663)*x[1]**o+(0.692123089256)*x[1]+(0.311287987388)*x[2]**o+(0.555427507546)*x[2]
        arg[2,1,0,0]=(0.464585537593)*x[0]**o+(0.161744615208)*x[0]+(-0.325569920662)*x[1]**o+(-0.47142644441)*x[1]+(0.885151021784)*x[2]**o+(-0.840595219086)*x[2]
        arg[2,1,0,1]=(-0.659253397488)*x[0]**o+(0.54391314862)*x[0]+(-0.595612529952)*x[1]**o+(-0.0472758114379)*x[1]+(0.876391930188)*x[2]**o+(-0.440313199946)*x[2]
        arg[2,1,1,0]=(-0.445591423364)*x[0]**o+(0.179144841652)*x[0]+(-0.294166502086)*x[1]**o+(-0.373820778739)*x[1]+(0.464609314863)*x[2]**o+(0.407893781978)*x[2]
        arg[2,1,1,1]=(-0.767512700255)*x[0]**o+(0.63198312168)*x[0]+(0.652318598902)*x[1]**o+(0.817874971823)*x[1]+(0.397211574885)*x[2]**o+(0.33174354886)*x[2]
        arg[2,1,2,0]=(0.473801282131)*x[0]**o+(-0.778895701394)*x[0]+(-0.801488721847)*x[1]**o+(-0.585840161723)*x[1]+(-0.805496721223)*x[2]**o+(0.585971840421)*x[2]
        arg[2,1,2,1]=(0.472329620891)*x[0]**o+(-0.0473805663117)*x[0]+(0.0894117727522)*x[1]**o+(-0.308473625733)*x[1]+(-0.658376953956)*x[2]**o+(0.32027592346)*x[2]
        arg[2,2,0,0]=(0.843241796369)*x[0]**o+(0.346857959798)*x[0]+(0.0111223078269)*x[1]**o+(-0.949623227267)*x[1]+(-0.152362393376)*x[2]**o+(-0.886340001141)*x[2]
        arg[2,2,0,1]=(-0.555393018762)*x[0]**o+(0.666364344488)*x[0]+(0.181237902671)*x[1]**o+(0.670959040345)*x[1]+(0.57383285317)*x[2]**o+(-0.451027832775)*x[2]
        arg[2,2,1,0]=(0.249709446431)*x[0]**o+(0.568753063379)*x[0]+(-0.812386750097)*x[1]**o+(0.506805124672)*x[1]+(-0.454844217662)*x[2]**o+(0.843435269288)*x[2]
        arg[2,2,1,1]=(0.753144930025)*x[0]**o+(0.0181642998269)*x[0]+(0.622635241593)*x[1]**o+(-0.372883842874)*x[1]+(0.366920965577)*x[2]**o+(-0.092391524816)*x[2]
        arg[2,2,2,0]=(0.309601906611)*x[0]**o+(-0.502573192965)*x[0]+(-0.420795137333)*x[1]**o+(-0.807109356307)*x[1]+(-0.802887583918)*x[2]**o+(0.552609684817)*x[2]
        arg[2,2,2,1]=(0.320933223342)*x[0]**o+(0.541931544704)*x[0]+(0.988715316931)*x[1]**o+(-0.4197325237)*x[1]+(-0.438705512435)*x[2]**o+(-0.540175093599)*x[2]
        arg[2,3,0,0]=(0.018870015518)*x[0]**o+(-0.615120691773)*x[0]+(0.596295280542)*x[1]**o+(0.210874310884)*x[1]+(-0.652194502321)*x[2]**o+(0.74019575754)*x[2]
        arg[2,3,0,1]=(0.820285141622)*x[0]**o+(0.130985741886)*x[0]+(-0.397387767472)*x[1]**o+(0.975103051154)*x[1]+(0.0887369214611)*x[2]**o+(0.88498731668)*x[2]
        arg[2,3,1,0]=(0.160503336808)*x[0]**o+(0.641180113144)*x[0]+(-0.408109675737)*x[1]**o+(-0.355443951133)*x[1]+(-0.429622016495)*x[2]**o+(0.998729833779)*x[2]
        arg[2,3,1,1]=(-0.646524352112)*x[0]**o+(-0.289894880059)*x[0]+(0.909761879864)*x[1]**o+(0.30453651619)*x[1]+(0.884596709091)*x[2]**o+(0.329408737003)*x[2]
        arg[2,3,2,0]=(0.459385919432)*x[0]**o+(0.416447271916)*x[0]+(-0.400798659819)*x[1]**o+(-0.769043818187)*x[1]+(-0.631682400803)*x[2]**o+(-0.80150361084)*x[2]
        arg[2,3,2,1]=(-0.544824298969)*x[0]**o+(0.191301965562)*x[0]+(0.41061102918)*x[1]**o+(-0.999195759064)*x[1]+(-0.0226105975802)*x[2]**o+(-0.137078691725)*x[2]
        arg[2,4,0,0]=(-0.624033952375)*x[0]**o+(0.0792222332007)*x[0]+(0.766756718455)*x[1]**o+(0.581487379511)*x[1]+(-0.437940391245)*x[2]**o+(0.115617341492)*x[2]
        arg[2,4,0,1]=(-0.300526068838)*x[0]**o+(-0.822844066866)*x[0]+(0.660949310367)*x[1]**o+(-0.0993949179102)*x[1]+(0.599884615655)*x[2]**o+(-0.903177033523)*x[2]
        arg[2,4,1,0]=(-0.654642517091)*x[0]**o+(0.428391621765)*x[0]+(0.932377441464)*x[1]**o+(-0.459525189684)*x[1]+(0.667425199675)*x[2]**o+(-0.937213410686)*x[2]
        arg[2,4,1,1]=(0.258067525957)*x[0]**o+(0.619943768383)*x[0]+(0.458217567511)*x[1]**o+(0.0812191081554)*x[1]+(0.743260637332)*x[2]**o+(-0.796150472293)*x[2]
        arg[2,4,2,0]=(0.782473636556)*x[0]**o+(-0.441405613305)*x[0]+(0.3951595327)*x[1]**o+(-0.449855781187)*x[1]+(-0.964130179212)*x[2]**o+(-0.299199410525)*x[2]
        arg[2,4,2,1]=(-0.878340584397)*x[0]**o+(-0.674575785429)*x[0]+(0.838137461257)*x[1]**o+(-0.498710475715)*x[1]+(-0.117205127994)*x[2]**o+(0.0934669704542)*x[2]
        arg[3,0,0,0]=(0.139631366137)*x[0]**o+(-0.893001605238)*x[0]+(-0.420780773096)*x[1]**o+(0.439055863393)*x[1]+(0.173113363237)*x[2]**o+(0.893867443017)*x[2]
        arg[3,0,0,1]=(-0.804823042124)*x[0]**o+(0.250231181298)*x[0]+(0.358464048439)*x[1]**o+(-0.847391231172)*x[1]+(-0.515391975323)*x[2]**o+(-0.999640302937)*x[2]
        arg[3,0,1,0]=(0.328676968943)*x[0]**o+(-0.0521242139559)*x[0]+(0.861548411537)*x[1]**o+(-0.288979961428)*x[1]+(-0.927901849987)*x[2]**o+(0.857869958971)*x[2]
        arg[3,0,1,1]=(-0.654866876537)*x[0]**o+(-0.0515462321916)*x[0]+(-0.224164685097)*x[1]**o+(-0.84207399757)*x[1]+(-0.442392040219)*x[2]**o+(-0.915734076984)*x[2]
        arg[3,0,2,0]=(-0.322517406382)*x[0]**o+(-0.757146499161)*x[0]+(-0.541690197465)*x[1]**o+(0.931090695413)*x[1]+(-0.355505263744)*x[2]**o+(-0.773185404514)*x[2]
        arg[3,0,2,1]=(0.421030649833)*x[0]**o+(-0.83768939688)*x[0]+(-0.863791071447)*x[1]**o+(0.78393205545)*x[1]+(-0.489893911681)*x[2]**o+(-0.291989245709)*x[2]
        arg[3,1,0,0]=(0.550839051007)*x[0]**o+(-0.883853789938)*x[0]+(-0.339529971923)*x[1]**o+(-0.756936847344)*x[1]+(0.724888858656)*x[2]**o+(-0.981057762224)*x[2]
        arg[3,1,0,1]=(0.711792743156)*x[0]**o+(0.233248429378)*x[0]+(0.526076154408)*x[1]**o+(0.349364804166)*x[1]+(0.271475992486)*x[2]**o+(0.782429402713)*x[2]
        arg[3,1,1,0]=(-0.499072848146)*x[0]**o+(-0.0604888692308)*x[0]+(0.545442893272)*x[1]**o+(-0.688797522515)*x[1]+(0.305341295222)*x[2]**o+(0.130010601331)*x[2]
        arg[3,1,1,1]=(0.940695359452)*x[0]**o+(0.113015472294)*x[0]+(-0.506254395321)*x[1]**o+(-0.252044277039)*x[1]+(0.0172784972094)*x[2]**o+(-0.184256978193)*x[2]
        arg[3,1,2,0]=(0.550472143981)*x[0]**o+(-0.729809590776)*x[0]+(-0.293765999495)*x[1]**o+(-0.378042976721)*x[1]+(0.973242167993)*x[2]**o+(-0.75350391134)*x[2]
        arg[3,1,2,1]=(0.369266178598)*x[0]**o+(-0.764034492366)*x[0]+(-0.713325992328)*x[1]**o+(0.948932375717)*x[1]+(-0.335074977323)*x[2]**o+(0.117829591745)*x[2]
        arg[3,2,0,0]=(0.426458076699)*x[0]**o+(-0.520551329627)*x[0]+(-0.773464312884)*x[1]**o+(0.26332468989)*x[1]+(0.143497828773)*x[2]**o+(0.0676278275986)*x[2]
        arg[3,2,0,1]=(0.85806016531)*x[0]**o+(0.755361032731)*x[0]+(0.403942188112)*x[1]**o+(0.0113665816202)*x[1]+(-0.439835633198)*x[2]**o+(0.620930145183)*x[2]
        arg[3,2,1,0]=(-0.852838624821)*x[0]**o+(-0.918594624263)*x[0]+(-0.572898555139)*x[1]**o+(-0.14353390918)*x[1]+(-0.716010071307)*x[2]**o+(0.134687635305)*x[2]
        arg[3,2,1,1]=(-0.14282590746)*x[0]**o+(-0.169314199229)*x[0]+(0.880745511071)*x[1]**o+(0.540733710441)*x[1]+(-0.131424658303)*x[2]**o+(-0.93209431119)*x[2]
        arg[3,2,2,0]=(-0.864016075688)*x[0]**o+(-0.265794760131)*x[0]+(0.444625188371)*x[1]**o+(0.133693703972)*x[1]+(-0.749799597936)*x[2]**o+(0.886355820615)*x[2]
        arg[3,2,2,1]=(0.676281241167)*x[0]**o+(-0.351800298325)*x[0]+(-0.751498380099)*x[1]**o+(-0.561737786104)*x[1]+(-0.984872672853)*x[2]**o+(0.476840947707)*x[2]
        arg[3,3,0,0]=(0.457874366022)*x[0]**o+(-0.126929649966)*x[0]+(0.666077559139)*x[1]**o+(-0.966437669224)*x[1]+(-0.849000502816)*x[2]**o+(0.990790747926)*x[2]
        arg[3,3,0,1]=(0.45136859326)*x[0]**o+(0.771355456911)*x[0]+(-0.678558472676)*x[1]**o+(0.0366869703264)*x[1]+(-0.257247358952)*x[2]**o+(-0.290852285538)*x[2]
        arg[3,3,1,0]=(-0.623007069616)*x[0]**o+(-0.365255080044)*x[0]+(0.441927420573)*x[1]**o+(0.331640625664)*x[1]+(-0.21153951456)*x[2]**o+(0.297967669123)*x[2]
        arg[3,3,1,1]=(-0.836057742118)*x[0]**o+(0.959873293328)*x[0]+(-0.52034405502)*x[1]**o+(-0.178838686591)*x[1]+(0.317610563306)*x[2]**o+(-0.555134650833)*x[2]
        arg[3,3,2,0]=(0.0819108022024)*x[0]**o+(0.69432296821)*x[0]+(-0.23685434367)*x[1]**o+(-0.0442137744391)*x[1]+(-0.879175086057)*x[2]**o+(-0.28031950012)*x[2]
        arg[3,3,2,1]=(0.340083286763)*x[0]**o+(-0.381942838645)*x[0]+(0.570579443852)*x[1]**o+(0.772192708863)*x[1]+(0.42813556157)*x[2]**o+(0.735683065499)*x[2]
        arg[3,4,0,0]=(-0.866813692418)*x[0]**o+(-0.53715159706)*x[0]+(0.0725900211922)*x[1]**o+(0.0891167848614)*x[1]+(-0.381482922347)*x[2]**o+(-0.521504123034)*x[2]
        arg[3,4,0,1]=(0.487750922723)*x[0]**o+(0.967820473849)*x[0]+(0.305600886352)*x[1]**o+(0.350777657643)*x[1]+(-0.383731485788)*x[2]**o+(0.814016126476)*x[2]
        arg[3,4,1,0]=(-0.186663244344)*x[0]**o+(0.144872424379)*x[0]+(0.844454339304)*x[1]**o+(0.252133708483)*x[1]+(-0.748235508222)*x[2]**o+(0.702952185343)*x[2]
        arg[3,4,1,1]=(-0.962296047927)*x[0]**o+(0.874277032826)*x[0]+(-0.404689253219)*x[1]**o+(-0.174075799004)*x[1]+(0.65086350106)*x[2]**o+(-0.109850650677)*x[2]
        arg[3,4,2,0]=(-0.59515836027)*x[0]**o+(-0.0917296352911)*x[0]+(0.834779238353)*x[1]**o+(0.951252871777)*x[1]+(0.983008684497)*x[2]**o+(-0.784806954891)*x[2]
        arg[3,4,2,1]=(0.603110971708)*x[0]**o+(0.291507042984)*x[0]+(-0.401523286981)*x[1]**o+(-0.994760476862)*x[1]+(0.410516007475)*x[2]**o+(0.368982230001)*x[2]
        ref[0,0,0,0]=(-0.322384218211)/(o+1.)+(0.436168366323)
        ref[0,0,0,1]=(-0.550432608405)/(o+1.)+(-0.839529924682)
        ref[0,0,1,0]=(-1.20222931022)/(o+1.)+(-0.674931792543)
        ref[0,0,1,1]=(-1.99447702496)/(o+1.)+(0.184076712997)
        ref[0,0,2,0]=(-0.848786191606)/(o+1.)+(-0.719076228559)
        ref[0,0,2,1]=(0.488199570726)/(o+1.)+(0.818508727159)
        ref[0,1,0,0]=(0.910006770156)/(o+1.)+(-0.781077934994)
        ref[0,1,0,1]=(0.229140786917)/(o+1.)+(0.225201244712)
        ref[0,1,1,0]=(-1.24973037868)/(o+1.)+(0.478630842756)
        ref[0,1,1,1]=(0.362813222831)/(o+1.)+(-0.269799066354)
        ref[0,1,2,0]=(0.704186070044)/(o+1.)+(-0.0666602572428)
        ref[0,1,2,1]=(-0.429870561459)/(o+1.)+(-0.692214543685)
        ref[0,2,0,0]=(-1.76040621668)/(o+1.)+(-0.103077097544)
        ref[0,2,0,1]=(-0.578245990054)/(o+1.)+(-0.24050068432)
        ref[0,2,1,0]=(0.982824772144)/(o+1.)+(-0.332261654225)
        ref[0,2,1,1]=(0.493111487615)/(o+1.)+(0.105743857349)
        ref[0,2,2,0]=(-0.406637376689)/(o+1.)+(0.910652270393)
        ref[0,2,2,1]=(0.0735211634779)/(o+1.)+(0.355106669226)
        ref[0,3,0,0]=(-0.538705798782)/(o+1.)+(0.896367101094)
        ref[0,3,0,1]=(0.11388337928)/(o+1.)+(0.766078903586)
        ref[0,3,1,0]=(0.955296156262)/(o+1.)+(-0.64915883862)
        ref[0,3,1,1]=(0.437273793256)/(o+1.)+(-0.61952690992)
        ref[0,3,2,0]=(1.07390497644)/(o+1.)+(-0.242730112953)
        ref[0,3,2,1]=(-0.79053815352)/(o+1.)+(-0.0521715653697)
        ref[0,4,0,0]=(-0.0103557372704)/(o+1.)+(0.939327713906)
        ref[0,4,0,1]=(-0.66740052552)/(o+1.)+(-0.671249919496)
        ref[0,4,1,0]=(-1.10797150302)/(o+1.)+(0.454490777873)
        ref[0,4,1,1]=(-0.00636792544909)/(o+1.)+(-0.368286670996)
        ref[0,4,2,0]=(-1.23874065284)/(o+1.)+(0.0410396874718)
        ref[0,4,2,1]=(0.834902700591)/(o+1.)+(0.562757794439)
        ref[1,0,0,0]=(1.46682450791)/(o+1.)+(0.891962057632)
        ref[1,0,0,1]=(0.183329431084)/(o+1.)+(0.659083249781)
        ref[1,0,1,0]=(-0.24909411204)/(o+1.)+(0.201301697863)
        ref[1,0,1,1]=(0.723423340723)/(o+1.)+(-0.205765407152)
        ref[1,0,2,0]=(0.607815389082)/(o+1.)+(-0.451934578027)
        ref[1,0,2,1]=(0.919070217555)/(o+1.)+(1.00996326691)
        ref[1,1,0,0]=(0.361804510188)/(o+1.)+(-0.272105134113)
        ref[1,1,0,1]=(-0.388145380449)/(o+1.)+(0.440667992719)
        ref[1,1,1,0]=(1.11229281554)/(o+1.)+(0.0358417840478)
        ref[1,1,1,1]=(-0.360763258547)/(o+1.)+(-0.0662175232282)
        ref[1,1,2,0]=(-0.421932179206)/(o+1.)+(1.04809456598)
        ref[1,1,2,1]=(1.5138402229)/(o+1.)+(-0.0182788239862)
        ref[1,2,0,0]=(-1.2439886311)/(o+1.)+(0.204333673483)
        ref[1,2,0,1]=(1.36866295286)/(o+1.)+(0.510865902853)
        ref[1,2,1,0]=(0.636378052915)/(o+1.)+(-0.439317277801)
        ref[1,2,1,1]=(2.1721593538)/(o+1.)+(0.63055844212)
        ref[1,2,2,0]=(-0.00984028813171)/(o+1.)+(-0.390721816834)
        ref[1,2,2,1]=(1.06592059421)/(o+1.)+(0.303699000873)
        ref[1,3,0,0]=(-0.160594832134)/(o+1.)+(-0.271446552947)
        ref[1,3,0,1]=(0.313068732589)/(o+1.)+(-0.457202820505)
        ref[1,3,1,0]=(1.67446584033)/(o+1.)+(0.718634957026)
        ref[1,3,1,1]=(-0.399207614903)/(o+1.)+(0.379686935632)
        ref[1,3,2,0]=(0.947060327978)/(o+1.)+(-0.506844686357)
        ref[1,3,2,1]=(2.40080863254)/(o+1.)+(-0.413347335508)
        ref[1,4,0,0]=(-1.08603586849)/(o+1.)+(0.903964428263)
        ref[1,4,0,1]=(-0.471025636339)/(o+1.)+(0.657506547153)
        ref[1,4,1,0]=(1.0787768427)/(o+1.)+(0.428710904834)
        ref[1,4,1,1]=(0.909279847682)/(o+1.)+(-0.380313094604)
        ref[1,4,2,0]=(-0.827086892123)/(o+1.)+(0.961519048079)
        ref[1,4,2,1]=(-0.154619783556)/(o+1.)+(-0.0591580335741)
        ref[2,0,0,0]=(-0.302649323542)/(o+1.)+(-0.229657108834)
        ref[2,0,0,1]=(1.77170468326)/(o+1.)+(-0.0484045064764)
        ref[2,0,1,0]=(-1.27113695197)/(o+1.)+(-0.94026328862)
        ref[2,0,1,1]=(-0.0863320624027)/(o+1.)+(0.383317392539)
        ref[2,0,2,0]=(0.184957944467)/(o+1.)+(-0.217451097139)
        ref[2,0,2,1]=(-0.398642713235)/(o+1.)+(0.952207101709)
        ref[2,1,0,0]=(1.02416663872)/(o+1.)+(-0.575138524144)
        ref[2,1,0,1]=(-0.378473997252)/(o+1.)+(0.0281620686178)
        ref[2,1,1,0]=(-0.275148610588)/(o+1.)+(0.106608922446)
        ref[2,1,1,1]=(0.282017473532)/(o+1.)+(0.890800821181)
        ref[2,1,2,0]=(-1.13318416094)/(o+1.)+(-0.389382011348)
        ref[2,1,2,1]=(-0.0966355603128)/(o+1.)+(-0.0177891342927)
        ref[2,2,0,0]=(0.70200171082)/(o+1.)+(-0.744552634305)
        ref[2,2,0,1]=(0.199677737079)/(o+1.)+(0.443147776029)
        ref[2,2,1,0]=(-1.01752152133)/(o+1.)+(0.959496728669)
        ref[2,2,1,1]=(1.74270113719)/(o+1.)+(-0.223555533932)
        ref[2,2,2,0]=(-0.91408081464)/(o+1.)+(-0.378536432227)
        ref[2,2,2,1]=(0.870943027839)/(o+1.)+(-0.208988036298)
        ref[2,3,0,0]=(-0.0370292062615)/(o+1.)+(0.167974688326)
        ref[2,3,0,1]=(0.51163429561)/(o+1.)+(0.99553805486)
        ref[2,3,1,0]=(-0.677228355424)/(o+1.)+(0.642232997895)
        ref[2,3,1,1]=(1.14783423684)/(o+1.)+(0.172025186567)
        ref[2,3,2,0]=(-0.57309514119)/(o+1.)+(-0.577050078555)
        ref[2,3,2,1]=(-0.156823867369)/(o+1.)+(-0.472486242613)
        ref[2,4,0,0]=(-0.295217625165)/(o+1.)+(0.388163477102)
        ref[2,4,0,1]=(0.960307857184)/(o+1.)+(-0.91270800915)
        ref[2,4,1,0]=(0.945160124047)/(o+1.)+(-0.484173489303)
        ref[2,4,1,1]=(1.4595457308)/(o+1.)+(-0.0474937978771)
        ref[2,4,2,0]=(0.213502990044)/(o+1.)+(-0.595230402509)
        ref[2,4,2,1]=(-0.157408251134)/(o+1.)+(-0.539909645345)
        ref[3,0,0,0]=(-0.108036043722)/(o+1.)+(0.219960850586)
        ref[3,0,0,1]=(-0.961750969008)/(o+1.)+(-0.798400176406)
        ref[3,0,1,0]=(0.262323530493)/(o+1.)+(0.258382891794)
        ref[3,0,1,1]=(-1.32142360185)/(o+1.)+(-0.904677153373)
        ref[3,0,2,0]=(-1.21971286759)/(o+1.)+(-0.299620604131)
        ref[3,0,2,1]=(-0.932654333294)/(o+1.)+(-0.17287329357)
        ref[3,1,0,0]=(0.93619793774)/(o+1.)+(-1.31092419975)
        ref[3,1,0,1]=(1.50934489005)/(o+1.)+(0.682521318129)
        ref[3,1,1,0]=(0.351711340348)/(o+1.)+(-0.309637895207)
        ref[3,1,1,1]=(0.451719461341)/(o+1.)+(-0.161642891469)
        ref[3,1,2,0]=(1.22994831248)/(o+1.)+(-0.930678239419)
        ref[3,1,2,1]=(-0.679134791053)/(o+1.)+(0.151363737548)
        ref[3,2,0,0]=(-0.203508407413)/(o+1.)+(-0.0947994060691)
        ref[3,2,0,1]=(0.822166720224)/(o+1.)+(0.693828879767)
        ref[3,2,1,0]=(-2.14174725127)/(o+1.)+(-0.463720449069)
        ref[3,2,1,1]=(0.606494945308)/(o+1.)+(-0.280337399989)
        ref[3,2,2,0]=(-1.16919048525)/(o+1.)+(0.377127382228)
        ref[3,2,2,1]=(-1.06008981178)/(o+1.)+(-0.218348568361)
        ref[3,3,0,0]=(0.274951422345)/(o+1.)+(-0.0512882856318)
        ref[3,3,0,1]=(-0.484437238367)/(o+1.)+(0.25859507085)
        ref[3,3,1,0]=(-0.392619163603)/(o+1.)+(0.132176607371)
        ref[3,3,1,1]=(-1.03879123383)/(o+1.)+(0.112949977952)
        ref[3,3,2,0]=(-1.03411862752)/(o+1.)+(0.184894846825)
        ref[3,3,2,1]=(1.33879829218)/(o+1.)+(0.562966467859)
        ref[3,4,0,0]=(-1.17570659357)/(o+1.)+(-0.484769467616)
        ref[3,4,0,1]=(0.409620323287)/(o+1.)+(1.06630712898)
        ref[3,4,1,0]=(-0.0904444132627)/(o+1.)+(0.549979159103)
        ref[3,4,1,1]=(-0.716121800087)/(o+1.)+(0.295175291572)
        ref[3,4,2,0]=(1.22262956258)/(o+1.)+(0.0373581407971)
        ref[3,4,2,1]=(0.612103692201)/(o+1.)+(-0.167135601938)
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.640162884104)*x[0]**o+(-0.686170347303)*x[0]+(0.0531418393824)*x[1]**o+(-0.0288840863965)*x[1]
        ref=(-0.587021044721)/(o+1.)+(-0.35752721685)
      else:
        s=(0.835059122976)*x[0]**o+(-0.876192510448)*x[0]+(0.229125107737)*x[1]**o+(-0.378923833918)*x[1]+(0.734383439043)*x[2]**o+(0.987756783011)*x[2]
        ref=(1.79856766975)/(o+1.)+(-0.133679780678)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.292137623974)*x[0]**o+(0.75950865395)*x[0]+(-0.205837757637)*x[1]**o+(0.90443054103)*x[1]
        s[1]=(0.407997882771)*x[0]**o+(0.335693471916)*x[0]+(0.755336837856)*x[1]**o+(0.218679559306)*x[1]
        ref[0]=(-0.497975381612)/(o+1.)+(0.83196959749)
        ref[1]=(1.16333472063)/(o+1.)+(0.277186515611)
      else:
        s[0]=(-0.393835090841)*x[0]**o+(-0.474828597107)*x[0]+(0.465782088981)*x[1]**o+(-0.882630637032)*x[1]+(0.849599914303)*x[2]**o+(0.737421732808)*x[2]
        s[1]=(-0.589146617652)*x[0]**o+(-0.412139119114)*x[0]+(0.970406630484)*x[1]**o+(0.363942463909)*x[1]+(-0.372574721521)*x[2]**o+(-0.798081163974)*x[2]
        ref[0]=(0.921546912444)/(o+1.)+(-0.310018750665)
        ref[1]=(0.00868529131054)/(o+1.)+(-0.42313890959)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.542564643665)*x[0]**o+(0.869139399504)*x[0]+(-0.918464318956)*x[1]**o+(0.258557879364)*x[1]
        s[0,1]=(0.797108788422)*x[0]**o+(-0.200368785173)*x[0]+(-0.518597299854)*x[1]**o+(-0.647852667874)*x[1]
        s[0,2]=(0.692223658744)*x[0]**o+(-0.103586429107)*x[0]+(0.282165531986)*x[1]**o+(-0.191815912482)*x[1]
        s[0,3]=(-0.370167730862)*x[0]**o+(-0.822713005318)*x[0]+(-0.00317012401339)*x[1]**o+(-0.848121245241)*x[1]
        s[0,4]=(-0.52733376174)*x[0]**o+(-0.314966701977)*x[0]+(0.755974759737)*x[1]**o+(-0.483207203205)*x[1]
        s[1,0]=(-0.709714610043)*x[0]**o+(0.647231907198)*x[0]+(-0.577474021122)*x[1]**o+(-0.124504757353)*x[1]
        s[1,1]=(-0.723690611716)*x[0]**o+(-0.617424586153)*x[0]+(-0.536447933353)*x[1]**o+(0.255892253884)*x[1]
        s[1,2]=(-0.201052994401)*x[0]**o+(-0.519971533366)*x[0]+(0.515314592901)*x[1]**o+(0.831783913039)*x[1]
        s[1,3]=(0.505769313912)*x[0]**o+(-0.460235423669)*x[0]+(0.982458353968)*x[1]**o+(0.479022475503)*x[1]
        s[1,4]=(0.624910824579)*x[0]**o+(0.563956018517)*x[0]+(0.0129379080194)*x[1]**o+(-0.634601977633)*x[1]
        s[2,0]=(-0.765921383834)*x[0]**o+(0.0199045688425)*x[0]+(-0.0666443330576)*x[1]**o+(-0.929035862007)*x[1]
        s[2,1]=(0.698423948822)*x[0]**o+(-0.529311289681)*x[0]+(-0.563339072497)*x[1]**o+(-0.494985449663)*x[1]
        s[2,2]=(0.101812328764)*x[0]**o+(-0.922547259485)*x[0]+(-0.45388897835)*x[1]**o+(0.357458826452)*x[1]
        s[2,3]=(0.643691149734)*x[0]**o+(-0.791395473264)*x[0]+(-0.919285047298)*x[1]**o+(-0.732314921151)*x[1]
        s[2,4]=(0.770509137995)*x[0]**o+(-0.399023743738)*x[0]+(-0.0887239237676)*x[1]**o+(0.846766840231)*x[1]
        s[3,0]=(-0.258188916476)*x[0]**o+(-0.192762172857)*x[0]+(-0.291360534186)*x[1]**o+(-0.828483492861)*x[1]
        s[3,1]=(-0.180563503871)*x[0]**o+(0.411356882323)*x[0]+(0.144166964346)*x[1]**o+(-0.817750491386)*x[1]
        s[3,2]=(-0.56008111033)*x[0]**o+(-0.903908126304)*x[0]+(-0.795710525329)*x[1]**o+(0.42957467817)*x[1]
        s[3,3]=(-0.435823321725)*x[0]**o+(-0.687506280092)*x[0]+(0.512172892719)*x[1]**o+(0.624268656549)*x[1]
        s[3,4]=(0.0995998270974)*x[0]**o+(0.54337587883)*x[0]+(-0.589282344971)*x[1]**o+(0.551709444954)*x[1]
        ref[0,0]=(-0.375899675291)/(o+1.)+(0.563848639434)
        ref[0,1]=(0.278511488568)/(o+1.)+(-0.424110726524)
        ref[0,2]=(0.97438919073)/(o+1.)+(-0.147701170794)
        ref[0,3]=(-0.373337854876)/(o+1.)+(-0.835417125279)
        ref[0,4]=(0.228640997997)/(o+1.)+(-0.399086952591)
        ref[1,0]=(-1.28718863116)/(o+1.)+(0.261363574923)
        ref[1,1]=(-1.26013854507)/(o+1.)+(-0.180766166134)
        ref[1,2]=(0.3142615985)/(o+1.)+(0.155906189837)
        ref[1,3]=(1.48822766788)/(o+1.)+(0.00939352591692)
        ref[1,4]=(0.637848732599)/(o+1.)+(-0.0353229795581)
        ref[2,0]=(-0.832565716891)/(o+1.)+(-0.454565646582)
        ref[2,1]=(0.135084876326)/(o+1.)+(-0.512148369672)
        ref[2,2]=(-0.352076649587)/(o+1.)+(-0.282544216517)
        ref[2,3]=(-0.275593897564)/(o+1.)+(-0.761855197208)
        ref[2,4]=(0.681785214228)/(o+1.)+(0.223871548247)
        ref[3,0]=(-0.549549450662)/(o+1.)+(-0.510622832859)
        ref[3,1]=(-0.0363965395244)/(o+1.)+(-0.203196804532)
        ref[3,2]=(-1.35579163566)/(o+1.)+(-0.237166724067)
        ref[3,3]=(0.0763495709944)/(o+1.)+(-0.0316188117717)
        ref[3,4]=(-0.489682517873)/(o+1.)+(0.547542661892)
      else:
        s[0,0]=(-0.547777970569)*x[0]**o+(-0.644569779184)*x[0]+(-0.247447574507)*x[1]**o+(-0.165674972238)*x[1]+(0.492011982518)*x[2]**o+(0.824373100652)*x[2]
        s[0,1]=(0.714120627471)*x[0]**o+(-0.343344240726)*x[0]+(0.846874965081)*x[1]**o+(0.955417725709)*x[1]+(-0.462347625853)*x[2]**o+(0.236518016678)*x[2]
        s[0,2]=(0.649286582806)*x[0]**o+(-0.712764590386)*x[0]+(-0.364435855331)*x[1]**o+(0.529624579838)*x[1]+(-0.478005452391)*x[2]**o+(0.79676822654)*x[2]
        s[0,3]=(0.186538430487)*x[0]**o+(0.873009863028)*x[0]+(-0.617246224023)*x[1]**o+(0.212670710234)*x[1]+(-0.568924203263)*x[2]**o+(-0.591868989872)*x[2]
        s[0,4]=(0.0764403834746)*x[0]**o+(0.580116472215)*x[0]+(-0.964832163438)*x[1]**o+(0.146621202039)*x[1]+(0.303667547632)*x[2]**o+(-0.572109268772)*x[2]
        s[1,0]=(0.0466497527173)*x[0]**o+(-0.660397185235)*x[0]+(0.314235627776)*x[1]**o+(-0.0219053335847)*x[1]+(0.0341017375209)*x[2]**o+(-0.138388505919)*x[2]
        s[1,1]=(-0.306253609605)*x[0]**o+(-0.274894845016)*x[0]+(0.590915613652)*x[1]**o+(0.466423547973)*x[1]+(-0.65224024787)*x[2]**o+(0.011478276889)*x[2]
        s[1,2]=(-0.850811667442)*x[0]**o+(-0.362611692783)*x[0]+(-0.132198189858)*x[1]**o+(0.613225855964)*x[1]+(-0.995844863385)*x[2]**o+(-0.00228774595319)*x[2]
        s[1,3]=(-0.331021120468)*x[0]**o+(0.262316324101)*x[0]+(0.693647465307)*x[1]**o+(-0.758663328061)*x[1]+(0.563291863323)*x[2]**o+(-0.464786381036)*x[2]
        s[1,4]=(-0.297063926313)*x[0]**o+(0.258694451069)*x[0]+(0.453986223203)*x[1]**o+(0.516652862875)*x[1]+(-0.119453447306)*x[2]**o+(0.13369866553)*x[2]
        s[2,0]=(0.895304992051)*x[0]**o+(-0.581521556328)*x[0]+(-0.102534354406)*x[1]**o+(-0.312191328622)*x[1]+(0.00625028874943)*x[2]**o+(0.455967398752)*x[2]
        s[2,1]=(0.0129094351309)*x[0]**o+(-0.284360799298)*x[0]+(0.254610787505)*x[1]**o+(0.232923077121)*x[1]+(0.884382417671)*x[2]**o+(0.879710026011)*x[2]
        s[2,2]=(0.978354808904)*x[0]**o+(0.45417061753)*x[0]+(-0.617831296441)*x[1]**o+(-0.540161947193)*x[1]+(0.749817874967)*x[2]**o+(0.85276098284)*x[2]
        s[2,3]=(0.790036783517)*x[0]**o+(-0.461944315129)*x[0]+(0.0177930616139)*x[1]**o+(0.827267186021)*x[1]+(-0.5681970366)*x[2]**o+(0.0800529373343)*x[2]
        s[2,4]=(0.320372577784)*x[0]**o+(-0.988843069534)*x[0]+(0.366830399498)*x[1]**o+(0.0749365557309)*x[1]+(-0.964416326732)*x[2]**o+(0.0589434543979)*x[2]
        s[3,0]=(-0.143156890467)*x[0]**o+(0.665143613007)*x[0]+(0.324524292861)*x[1]**o+(0.573292071297)*x[1]+(0.55975986068)*x[2]**o+(-0.703161304253)*x[2]
        s[3,1]=(0.827768123756)*x[0]**o+(-0.565184841493)*x[0]+(0.724124619351)*x[1]**o+(-0.375933144184)*x[1]+(-0.186573364344)*x[2]**o+(0.174303241336)*x[2]
        s[3,2]=(0.19505934899)*x[0]**o+(-0.378308488179)*x[0]+(-0.542115440091)*x[1]**o+(-0.864956237152)*x[1]+(0.0489116293809)*x[2]**o+(0.149786871599)*x[2]
        s[3,3]=(-0.151225736973)*x[0]**o+(-0.491229793833)*x[0]+(-0.704529608083)*x[1]**o+(-0.790051884866)*x[1]+(-0.202633581578)*x[2]**o+(0.0769118984781)*x[2]
        s[3,4]=(0.523527337562)*x[0]**o+(0.266094832079)*x[0]+(-0.147028009699)*x[1]**o+(-0.683770701764)*x[1]+(0.794443408406)*x[2]**o+(-0.0536824425549)*x[2]
        ref[0,0]=(-0.303213562559)/(o+1.)+(0.00706417461518)
        ref[0,1]=(1.0986479667)/(o+1.)+(0.42429575083)
        ref[0,2]=(-0.193154724916)/(o+1.)+(0.306814107996)
        ref[0,3]=(-0.999631996799)/(o+1.)+(0.246905791695)
        ref[0,4]=(-0.584724232331)/(o+1.)+(0.0773142027411)
        ref[1,0]=(0.394987118014)/(o+1.)+(-0.410345512369)
        ref[1,1]=(-0.367578243823)/(o+1.)+(0.101503489923)
        ref[1,2]=(-1.97885472069)/(o+1.)+(0.124163208614)
        ref[1,3]=(0.925918208162)/(o+1.)+(-0.480566692498)
        ref[1,4]=(0.0374688495844)/(o+1.)+(0.454522989737)
        ref[2,0]=(0.799020926395)/(o+1.)+(-0.218872743099)
        ref[2,1]=(1.15190264031)/(o+1.)+(0.414136151917)
        ref[2,2]=(1.11034138743)/(o+1.)+(0.383384826588)
        ref[2,3]=(0.239632808531)/(o+1.)+(0.222687904113)
        ref[2,4]=(-0.27721334945)/(o+1.)+(-0.427481529703)
        ref[3,0]=(0.741127263073)/(o+1.)+(0.267637190026)
        ref[3,1]=(1.36531937876)/(o+1.)+(-0.383407372171)
        ref[3,2]=(-0.29814446172)/(o+1.)+(-0.546738926866)
        ref[3,3]=(-1.05838892663)/(o+1.)+(-0.602184890111)
        ref[3,4]=(1.17094273627)/(o+1.)+(-0.23567915612)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.630010583318)*x[0]**o+(0.0494796640212)*x[0]+(0.471665362166)*x[1]**o+(0.392459033784)*x[1]
        s[0,0,1]=(-0.423295675063)*x[0]**o+(0.24240367131)*x[0]+(-0.724581249901)*x[1]**o+(-0.131295119605)*x[1]
        s[0,1,0]=(-0.410928342692)*x[0]**o+(-0.541418037237)*x[0]+(-0.232546576693)*x[1]**o+(-0.0784119675747)*x[1]
        s[0,1,1]=(0.789988066595)*x[0]**o+(-0.243498836345)*x[0]+(0.368561675765)*x[1]**o+(-0.747610528524)*x[1]
        s[1,0,0]=(-0.326293541804)*x[0]**o+(0.101066389163)*x[0]+(-0.184367851)*x[1]**o+(0.568769737008)*x[1]
        s[1,0,1]=(-0.015503180143)*x[0]**o+(0.819464848547)*x[0]+(0.979976262017)*x[1]**o+(-0.0389216169058)*x[1]
        s[1,1,0]=(-0.542766421816)*x[0]**o+(-0.373053336375)*x[0]+(-0.938010294052)*x[1]**o+(-0.399667401896)*x[1]
        s[1,1,1]=(0.0200936228416)*x[0]**o+(0.770519603961)*x[0]+(0.620955090369)*x[1]**o+(-0.330894822252)*x[1]
        s[2,0,0]=(-0.171145765903)*x[0]**o+(0.612313985433)*x[0]+(0.0479528062441)*x[1]**o+(-0.281058939492)*x[1]
        s[2,0,1]=(-0.332826833153)*x[0]**o+(0.944629304493)*x[0]+(-0.487576018254)*x[1]**o+(-0.904875326977)*x[1]
        s[2,1,0]=(-0.00808555466364)*x[0]**o+(0.245367446684)*x[0]+(-0.164397692869)*x[1]**o+(-0.198574049753)*x[1]
        s[2,1,1]=(-0.0469077421177)*x[0]**o+(0.89449436367)*x[0]+(-0.802161009378)*x[1]**o+(-0.0767107312174)*x[1]
        s[3,0,0]=(0.530561159497)*x[0]**o+(-0.0909174483134)*x[0]+(0.159443779335)*x[1]**o+(-0.178020813545)*x[1]
        s[3,0,1]=(-0.216531776351)*x[0]**o+(-0.531806095078)*x[0]+(0.653531437331)*x[1]**o+(0.21339193674)*x[1]
        s[3,1,0]=(0.254391857637)*x[0]**o+(0.619393713071)*x[0]+(-0.260306661561)*x[1]**o+(-0.473743696717)*x[1]
        s[3,1,1]=(-0.921367458187)*x[0]**o+(-0.69212037605)*x[0]+(0.6216551239)*x[1]**o+(0.7963578436)*x[1]
        s[4,0,0]=(-0.506208138693)*x[0]**o+(-0.103718686465)*x[0]+(0.771724985573)*x[1]**o+(0.815749872065)*x[1]
        s[4,0,1]=(-0.21355375924)*x[0]**o+(0.635638481649)*x[0]+(0.274581141313)*x[1]**o+(-0.947357183419)*x[1]
        s[4,1,0]=(0.35658463006)*x[0]**o+(-0.351876832623)*x[0]+(-0.0999414279107)*x[1]**o+(0.212979499129)*x[1]
        s[4,1,1]=(0.494190940488)*x[0]**o+(0.0851596543145)*x[0]+(-0.433440340709)*x[1]**o+(-0.0173667523032)*x[1]
        s[5,0,0]=(0.53169691054)*x[0]**o+(0.952395156793)*x[0]+(-0.127903378991)*x[1]**o+(-0.605192547481)*x[1]
        s[5,0,1]=(0.795535841954)*x[0]**o+(-0.615332450079)*x[0]+(0.764586784938)*x[1]**o+(0.420251114523)*x[1]
        s[5,1,0]=(-0.699722370082)*x[0]**o+(0.608404015259)*x[0]+(-0.166384830092)*x[1]**o+(-0.0611871373695)*x[1]
        s[5,1,1]=(0.545721411662)*x[0]**o+(-0.426115151805)*x[0]+(-0.438291880099)*x[1]**o+(0.0686028790393)*x[1]
        ref[0,0,0]=(1.10167594548)/(o+1.)+(0.220969348903)
        ref[0,0,1]=(-1.14787692496)/(o+1.)+(0.055554275852)
        ref[0,1,0]=(-0.643474919385)/(o+1.)+(-0.309915002406)
        ref[0,1,1]=(1.15854974236)/(o+1.)+(-0.495554682434)
        ref[1,0,0]=(-0.510661392804)/(o+1.)+(0.334918063085)
        ref[1,0,1]=(0.964473081874)/(o+1.)+(0.390271615821)
        ref[1,1,0]=(-1.48077671587)/(o+1.)+(-0.386360369136)
        ref[1,1,1]=(0.64104871321)/(o+1.)+(0.219812390854)
        ref[2,0,0]=(-0.123192959659)/(o+1.)+(0.16562752297)
        ref[2,0,1]=(-0.820402851406)/(o+1.)+(0.0198769887578)
        ref[2,1,0]=(-0.172483247533)/(o+1.)+(0.0233966984659)
        ref[2,1,1]=(-0.849068751496)/(o+1.)+(0.408891816226)
        ref[3,0,0]=(0.690004938832)/(o+1.)+(-0.134469130929)
        ref[3,0,1]=(0.43699966098)/(o+1.)+(-0.159207079169)
        ref[3,1,0]=(-0.00591480392415)/(o+1.)+(0.0728250081768)
        ref[3,1,1]=(-0.299712334287)/(o+1.)+(0.0521187337749)
        ref[4,0,0]=(0.265516846879)/(o+1.)+(0.3560155928)
        ref[4,0,1]=(0.0610273820724)/(o+1.)+(-0.155859350885)
        ref[4,1,0]=(0.256643202149)/(o+1.)+(-0.0694486667468)
        ref[4,1,1]=(0.0607505997785)/(o+1.)+(0.0338964510057)
        ref[5,0,0]=(0.403793531549)/(o+1.)+(0.173601304656)
        ref[5,0,1]=(1.56012262689)/(o+1.)+(-0.0975406677777)
        ref[5,1,0]=(-0.866107200175)/(o+1.)+(0.273608438945)
        ref[5,1,1]=(0.107429531563)/(o+1.)+(-0.178756136383)
      else:
        s[0,0,0]=(0.584550224653)*x[0]**o+(-0.106569322848)*x[0]+(0.728893354651)*x[1]**o+(0.237029650377)*x[1]+(-0.546491966553)*x[2]**o+(0.0390698430932)*x[2]
        s[0,0,1]=(0.0636227774465)*x[0]**o+(0.934881844205)*x[0]+(-0.545166942655)*x[1]**o+(0.08148757249)*x[1]+(-0.179461185461)*x[2]**o+(0.641018091939)*x[2]
        s[0,1,0]=(0.325463516488)*x[0]**o+(-0.620730155784)*x[0]+(0.154991661017)*x[1]**o+(0.245046216313)*x[1]+(0.645726837044)*x[2]**o+(-0.230093713841)*x[2]
        s[0,1,1]=(-0.321016164146)*x[0]**o+(-0.0509943423674)*x[0]+(-0.479999557029)*x[1]**o+(0.569709724552)*x[1]+(0.190221146753)*x[2]**o+(-0.919477787005)*x[2]
        s[1,0,0]=(0.929113150774)*x[0]**o+(-0.733802947668)*x[0]+(0.8108269457)*x[1]**o+(-0.488163027335)*x[1]+(-0.745575620351)*x[2]**o+(-0.486955144862)*x[2]
        s[1,0,1]=(-0.017329684129)*x[0]**o+(0.147982077718)*x[0]+(-0.144958156543)*x[1]**o+(-0.300572864309)*x[1]+(0.00787918928353)*x[2]**o+(-0.532824168573)*x[2]
        s[1,1,0]=(0.685828903913)*x[0]**o+(-0.567114040704)*x[0]+(-0.0471407827804)*x[1]**o+(0.511415127351)*x[1]+(-0.554086444893)*x[2]**o+(-0.226435378121)*x[2]
        s[1,1,1]=(-0.415793464117)*x[0]**o+(-0.777707665965)*x[0]+(-0.846630968003)*x[1]**o+(0.389596880295)*x[1]+(-0.213682650766)*x[2]**o+(-0.0851017843465)*x[2]
        s[2,0,0]=(0.932561239744)*x[0]**o+(0.0615709255597)*x[0]+(0.634288279118)*x[1]**o+(0.293277319848)*x[1]+(-0.429481015254)*x[2]**o+(0.778577347672)*x[2]
        s[2,0,1]=(0.857123773566)*x[0]**o+(-0.47446916241)*x[0]+(0.951211790713)*x[1]**o+(0.443028899356)*x[1]+(0.0861361311835)*x[2]**o+(0.597277083437)*x[2]
        s[2,1,0]=(0.874693704519)*x[0]**o+(-0.920294692148)*x[0]+(-0.03228863342)*x[1]**o+(0.248843294667)*x[1]+(-0.926343823806)*x[2]**o+(0.15109837142)*x[2]
        s[2,1,1]=(0.840368685645)*x[0]**o+(-0.200563324888)*x[0]+(0.88039674632)*x[1]**o+(-0.349512401298)*x[1]+(0.25915053789)*x[2]**o+(-0.48356523029)*x[2]
        s[3,0,0]=(0.758393510867)*x[0]**o+(0.666093057943)*x[0]+(-0.453617931299)*x[1]**o+(-0.134152843992)*x[1]+(-0.0117598390465)*x[2]**o+(-0.730981155782)*x[2]
        s[3,0,1]=(0.205572845432)*x[0]**o+(0.46935055342)*x[0]+(-0.415086164167)*x[1]**o+(0.276709685737)*x[1]+(0.341889576347)*x[2]**o+(0.170610214566)*x[2]
        s[3,1,0]=(0.926107056617)*x[0]**o+(-0.928101461122)*x[0]+(-0.639162904468)*x[1]**o+(-0.787501002441)*x[1]+(0.615156638997)*x[2]**o+(0.444030323472)*x[2]
        s[3,1,1]=(0.458304153272)*x[0]**o+(0.119383772473)*x[0]+(-0.816833779036)*x[1]**o+(0.428980054603)*x[1]+(0.176196071727)*x[2]**o+(0.294498083663)*x[2]
        s[4,0,0]=(-0.184477855098)*x[0]**o+(-0.281312358692)*x[0]+(-0.259869292769)*x[1]**o+(0.670877642172)*x[1]+(0.710140723846)*x[2]**o+(-0.738455024569)*x[2]
        s[4,0,1]=(0.790994543706)*x[0]**o+(0.594490423343)*x[0]+(0.737641957747)*x[1]**o+(-0.974154697219)*x[1]+(-0.228328212417)*x[2]**o+(-0.870422565759)*x[2]
        s[4,1,0]=(0.538277719545)*x[0]**o+(0.0405414212643)*x[0]+(-0.340695322783)*x[1]**o+(-0.925052511534)*x[1]+(0.77026668428)*x[2]**o+(0.872129470246)*x[2]
        s[4,1,1]=(0.35315629997)*x[0]**o+(-0.581808852999)*x[0]+(-0.506082324541)*x[1]**o+(-0.979298372486)*x[1]+(0.269130557602)*x[2]**o+(0.0297188290624)*x[2]
        s[5,0,0]=(0.13627346547)*x[0]**o+(-0.634005099347)*x[0]+(0.413163361533)*x[1]**o+(-0.445924057754)*x[1]+(0.317763103856)*x[2]**o+(-0.0934833794122)*x[2]
        s[5,0,1]=(0.775864446863)*x[0]**o+(0.647005517482)*x[0]+(-0.922488575786)*x[1]**o+(0.592973645448)*x[1]+(0.752015212966)*x[2]**o+(-0.700335290015)*x[2]
        s[5,1,0]=(0.376707292205)*x[0]**o+(-0.185352369466)*x[0]+(0.701616000772)*x[1]**o+(-0.496738161848)*x[1]+(-0.722103675479)*x[2]**o+(-0.124090359996)*x[2]
        s[5,1,1]=(-0.070302198698)*x[0]**o+(-0.701418359643)*x[0]+(0.131136807633)*x[1]**o+(-0.0581346703562)*x[1]+(0.378664686889)*x[2]**o+(0.785268912587)*x[2]
        ref[0,0,0]=(0.766951612751)/(o+1.)+(0.0847650853109)
        ref[0,0,1]=(-0.66100535067)/(o+1.)+(0.828693754317)
        ref[0,1,0]=(1.12618201455)/(o+1.)+(-0.302888826656)
        ref[0,1,1]=(-0.610794574422)/(o+1.)+(-0.20038120241)
        ref[1,0,0]=(0.994364476124)/(o+1.)+(-0.854460559932)
        ref[1,0,1]=(-0.154408651388)/(o+1.)+(-0.342707477582)
        ref[1,1,0]=(0.0846016762395)/(o+1.)+(-0.141067145737)
        ref[1,1,1]=(-1.47610708289)/(o+1.)+(-0.236606285008)
        ref[2,0,0]=(1.13736850361)/(o+1.)+(0.56671279654)
        ref[2,0,1]=(1.89447169546)/(o+1.)+(0.282918410191)
        ref[2,1,0]=(-0.0839387527076)/(o+1.)+(-0.260176513031)
        ref[2,1,1]=(1.97991596985)/(o+1.)+(-0.516820478238)
        ref[3,0,0]=(0.293015740521)/(o+1.)+(-0.0995204709152)
        ref[3,0,1]=(0.132376257612)/(o+1.)+(0.458335226861)
        ref[3,1,0]=(0.902100791146)/(o+1.)+(-0.635786070046)
        ref[3,1,1]=(-0.182333554037)/(o+1.)+(0.421430955369)
        ref[4,0,0]=(0.265793575979)/(o+1.)+(-0.174444870545)
        ref[4,0,1]=(1.30030828904)/(o+1.)+(-0.625043419818)
        ref[4,1,0]=(0.967849081042)/(o+1.)+(-0.00619081001188)
        ref[4,1,1]=(0.116204533031)/(o+1.)+(-0.765694198211)
        ref[5,0,0]=(0.867199930859)/(o+1.)+(-0.586706268257)
        ref[5,0,1]=(0.605391084042)/(o+1.)+(0.269821936457)
        ref[5,1,0]=(0.356219617498)/(o+1.)+(-0.403090445655)
        ref[5,1,1]=(0.439499295823)/(o+1.)+(0.0128579412935)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.904602515522)*x[0]**o+(-0.371521279208)*x[0]+(-0.452933934201)*x[1]**o+(-0.202117873586)*x[1]
        s[0,0,0,1]=(0.046504512522)*x[0]**o+(-0.23169139456)*x[0]+(0.260745907068)*x[1]**o+(-0.280939854842)*x[1]
        s[0,0,1,0]=(0.150287081226)*x[0]**o+(-0.62438094501)*x[0]+(-0.617182852213)*x[1]**o+(0.0366613552193)*x[1]
        s[0,0,1,1]=(0.0710288489579)*x[0]**o+(-0.111273065295)*x[0]+(0.41817519756)*x[1]**o+(0.792759228517)*x[1]
        s[0,0,2,0]=(-0.00835504723814)*x[0]**o+(-0.928910786463)*x[0]+(-0.830115466368)*x[1]**o+(0.117896100163)*x[1]
        s[0,0,2,1]=(0.178088696257)*x[0]**o+(0.378404605517)*x[0]+(-0.0609283403081)*x[1]**o+(-0.0543695218661)*x[1]
        s[0,1,0,0]=(-0.0689997578822)*x[0]**o+(-0.632975314677)*x[0]+(0.817036198517)*x[1]**o+(0.83235818794)*x[1]
        s[0,1,0,1]=(0.307747809333)*x[0]**o+(-0.119316653583)*x[0]+(-0.962502450634)*x[1]**o+(0.730005736334)*x[1]
        s[0,1,1,0]=(0.474932601365)*x[0]**o+(-0.672778327766)*x[0]+(0.0931366097943)*x[1]**o+(0.441430893263)*x[1]
        s[0,1,1,1]=(-0.688077109622)*x[0]**o+(-0.0904781446959)*x[0]+(0.75334586689)*x[1]**o+(0.64166855195)*x[1]
        s[0,1,2,0]=(-0.445313569413)*x[0]**o+(0.0860176511221)*x[0]+(0.816569466868)*x[1]**o+(-0.716517326506)*x[1]
        s[0,1,2,1]=(-0.282441352341)*x[0]**o+(-0.63704220861)*x[0]+(-0.0493662512733)*x[1]**o+(0.568004633808)*x[1]
        s[0,2,0,0]=(-0.39629286578)*x[0]**o+(-0.930416708861)*x[0]+(0.426475002251)*x[1]**o+(-0.156685168112)*x[1]
        s[0,2,0,1]=(-0.932671691096)*x[0]**o+(-0.918584855074)*x[0]+(0.820320229167)*x[1]**o+(0.39108460143)*x[1]
        s[0,2,1,0]=(0.904161654788)*x[0]**o+(0.580424354995)*x[0]+(0.901999297026)*x[1]**o+(0.16227344306)*x[1]
        s[0,2,1,1]=(0.80438774629)*x[0]**o+(0.132549332211)*x[0]+(0.799348114478)*x[1]**o+(-0.22361807585)*x[1]
        s[0,2,2,0]=(-0.561135188306)*x[0]**o+(0.584405140552)*x[0]+(0.655074058246)*x[1]**o+(0.145893780708)*x[1]
        s[0,2,2,1]=(-0.252147320901)*x[0]**o+(0.798853343056)*x[0]+(0.615093018404)*x[1]**o+(0.394704774949)*x[1]
        s[0,3,0,0]=(0.36142643768)*x[0]**o+(0.307386971305)*x[0]+(-0.294243769289)*x[1]**o+(-0.516116390793)*x[1]
        s[0,3,0,1]=(-0.258524392445)*x[0]**o+(-0.955667526515)*x[0]+(-0.268164834645)*x[1]**o+(-0.0982143901681)*x[1]
        s[0,3,1,0]=(-0.363719831483)*x[0]**o+(0.542839794876)*x[0]+(-0.591176577173)*x[1]**o+(-0.885630814734)*x[1]
        s[0,3,1,1]=(-0.857066901153)*x[0]**o+(0.146459811163)*x[0]+(0.167525505888)*x[1]**o+(-0.823988951028)*x[1]
        s[0,3,2,0]=(-0.67987029357)*x[0]**o+(0.484426195638)*x[0]+(0.357896203022)*x[1]**o+(-0.225752357797)*x[1]
        s[0,3,2,1]=(-0.580488963435)*x[0]**o+(-0.561249356735)*x[0]+(0.383669718576)*x[1]**o+(0.657077607148)*x[1]
        s[0,4,0,0]=(0.69915173204)*x[0]**o+(-0.98282597178)*x[0]+(-0.107525636764)*x[1]**o+(0.860091695416)*x[1]
        s[0,4,0,1]=(-0.52652663231)*x[0]**o+(0.0466888351687)*x[0]+(0.490026321752)*x[1]**o+(0.694751470792)*x[1]
        s[0,4,1,0]=(0.824302118883)*x[0]**o+(0.00719006856334)*x[0]+(-0.539797289173)*x[1]**o+(0.690123831564)*x[1]
        s[0,4,1,1]=(-0.472128229772)*x[0]**o+(-0.648689330166)*x[0]+(0.391117917633)*x[1]**o+(-0.162627834297)*x[1]
        s[0,4,2,0]=(0.225959403713)*x[0]**o+(-0.955982321286)*x[0]+(0.98648257289)*x[1]**o+(0.685606762198)*x[1]
        s[0,4,2,1]=(0.344295953653)*x[0]**o+(0.0937724197016)*x[0]+(0.334346155325)*x[1]**o+(0.240553977121)*x[1]
        s[1,0,0,0]=(0.195329436142)*x[0]**o+(-0.391787107203)*x[0]+(0.282744654065)*x[1]**o+(0.525383075525)*x[1]
        s[1,0,0,1]=(0.0500438294543)*x[0]**o+(0.91622484244)*x[0]+(-0.678531170299)*x[1]**o+(-0.992341400238)*x[1]
        s[1,0,1,0]=(0.297013531565)*x[0]**o+(-0.11033494708)*x[0]+(-0.0204768747597)*x[1]**o+(0.911044821444)*x[1]
        s[1,0,1,1]=(-0.318409579239)*x[0]**o+(-0.0397282629607)*x[0]+(0.126590560999)*x[1]**o+(-0.583058191524)*x[1]
        s[1,0,2,0]=(-0.928353567554)*x[0]**o+(-0.290772724182)*x[0]+(0.114112299004)*x[1]**o+(0.320854358475)*x[1]
        s[1,0,2,1]=(-0.570768568717)*x[0]**o+(-0.824979336359)*x[0]+(-0.521612560762)*x[1]**o+(0.892935684254)*x[1]
        s[1,1,0,0]=(0.409271921712)*x[0]**o+(0.656688417546)*x[0]+(-0.552252115532)*x[1]**o+(-0.252684939643)*x[1]
        s[1,1,0,1]=(0.650926797031)*x[0]**o+(0.937738666477)*x[0]+(-0.280790296797)*x[1]**o+(0.214208558313)*x[1]
        s[1,1,1,0]=(0.1256718371)*x[0]**o+(0.261580183407)*x[0]+(-0.57943553377)*x[1]**o+(-0.746871082838)*x[1]
        s[1,1,1,1]=(0.935864480093)*x[0]**o+(-0.438674963211)*x[0]+(0.18368021533)*x[1]**o+(-0.694240108689)*x[1]
        s[1,1,2,0]=(-0.0397072007891)*x[0]**o+(-0.745343758112)*x[0]+(0.641693619013)*x[1]**o+(-0.658786784991)*x[1]
        s[1,1,2,1]=(-0.695558285084)*x[0]**o+(-0.650222199462)*x[0]+(-0.386725498078)*x[1]**o+(-0.674446830484)*x[1]
        s[1,2,0,0]=(0.905187721767)*x[0]**o+(-0.415012814051)*x[0]+(0.797167315255)*x[1]**o+(-0.80569654981)*x[1]
        s[1,2,0,1]=(-0.425831083968)*x[0]**o+(-0.306155499216)*x[0]+(0.878423386932)*x[1]**o+(-0.204033374375)*x[1]
        s[1,2,1,0]=(0.0315352399527)*x[0]**o+(-0.0864621608168)*x[0]+(0.953279645196)*x[1]**o+(-0.350121337825)*x[1]
        s[1,2,1,1]=(0.30461258451)*x[0]**o+(-0.825063058949)*x[0]+(0.512058260572)*x[1]**o+(0.748809814194)*x[1]
        s[1,2,2,0]=(-0.739360626393)*x[0]**o+(-0.163606333648)*x[0]+(0.0928076120668)*x[1]**o+(0.720798765678)*x[1]
        s[1,2,2,1]=(0.199986177011)*x[0]**o+(-0.355098022147)*x[0]+(-0.429976374764)*x[1]**o+(-0.333973984813)*x[1]
        s[1,3,0,0]=(0.245099889527)*x[0]**o+(-0.272006414876)*x[0]+(0.19038825631)*x[1]**o+(0.141537376325)*x[1]
        s[1,3,0,1]=(0.964892470064)*x[0]**o+(0.266000699533)*x[0]+(0.281992883406)*x[1]**o+(0.0470036363258)*x[1]
        s[1,3,1,0]=(-0.406024010771)*x[0]**o+(0.211515087571)*x[0]+(-0.139150083168)*x[1]**o+(0.694931678647)*x[1]
        s[1,3,1,1]=(0.719958915132)*x[0]**o+(0.425474993616)*x[0]+(0.381998339685)*x[1]**o+(-0.587210328527)*x[1]
        s[1,3,2,0]=(-0.446628131313)*x[0]**o+(0.147944239049)*x[0]+(0.858318344908)*x[1]**o+(0.141912002794)*x[1]
        s[1,3,2,1]=(0.123237888525)*x[0]**o+(0.609628760233)*x[0]+(-0.0997977880881)*x[1]**o+(0.776602457692)*x[1]
        s[1,4,0,0]=(-0.205992404955)*x[0]**o+(0.95994896532)*x[0]+(-0.668831643097)*x[1]**o+(0.400688887056)*x[1]
        s[1,4,0,1]=(-0.551761785809)*x[0]**o+(-0.188933577465)*x[0]+(-0.445368366446)*x[1]**o+(-0.487934192641)*x[1]
        s[1,4,1,0]=(0.806637485565)*x[0]**o+(-0.912751633205)*x[0]+(-0.0317735780285)*x[1]**o+(0.397954044901)*x[1]
        s[1,4,1,1]=(0.321015788412)*x[0]**o+(0.0387288594742)*x[0]+(0.924286425546)*x[1]**o+(0.636709951581)*x[1]
        s[1,4,2,0]=(0.822453145379)*x[0]**o+(-0.763204841293)*x[0]+(-0.821998150153)*x[1]**o+(-0.0455190284992)*x[1]
        s[1,4,2,1]=(-0.769855780302)*x[0]**o+(0.175243832232)*x[0]+(0.643564391513)*x[1]**o+(0.167745614405)*x[1]
        s[2,0,0,0]=(-0.703387009485)*x[0]**o+(-0.165672894741)*x[0]+(-0.593523456406)*x[1]**o+(-0.953916267625)*x[1]
        s[2,0,0,1]=(0.189269691722)*x[0]**o+(0.0650904594513)*x[0]+(-0.972603715891)*x[1]**o+(0.632748799808)*x[1]
        s[2,0,1,0]=(-0.355043086131)*x[0]**o+(0.223886836405)*x[0]+(0.0812465154386)*x[1]**o+(-0.123342881585)*x[1]
        s[2,0,1,1]=(0.111949886338)*x[0]**o+(0.333936233092)*x[0]+(-0.726855801653)*x[1]**o+(-0.410644679001)*x[1]
        s[2,0,2,0]=(-0.514134171061)*x[0]**o+(-0.920122884271)*x[0]+(-0.257750657117)*x[1]**o+(-0.920243692547)*x[1]
        s[2,0,2,1]=(-0.845077655219)*x[0]**o+(-0.208901924124)*x[0]+(0.337378697599)*x[1]**o+(0.801979415473)*x[1]
        s[2,1,0,0]=(0.193043677084)*x[0]**o+(-0.0599530199511)*x[0]+(0.43567411879)*x[1]**o+(0.500546331575)*x[1]
        s[2,1,0,1]=(0.599088933494)*x[0]**o+(-0.468604163828)*x[0]+(0.429967954899)*x[1]**o+(-0.492384378174)*x[1]
        s[2,1,1,0]=(-0.720574232874)*x[0]**o+(0.821661207615)*x[0]+(0.724662820892)*x[1]**o+(0.753198897961)*x[1]
        s[2,1,1,1]=(-0.448359465447)*x[0]**o+(0.693373974155)*x[0]+(-0.484872729016)*x[1]**o+(-0.192659481547)*x[1]
        s[2,1,2,0]=(0.703509253973)*x[0]**o+(-0.0365462091464)*x[0]+(0.240885950313)*x[1]**o+(-0.321055059013)*x[1]
        s[2,1,2,1]=(0.247817476916)*x[0]**o+(-0.138605619124)*x[0]+(0.697567286153)*x[1]**o+(0.727793395093)*x[1]
        s[2,2,0,0]=(0.0419479009665)*x[0]**o+(0.66706255286)*x[0]+(-0.330849950899)*x[1]**o+(-0.600398002711)*x[1]
        s[2,2,0,1]=(0.0302303643583)*x[0]**o+(0.851488267489)*x[0]+(-0.568849620793)*x[1]**o+(-0.363234175722)*x[1]
        s[2,2,1,0]=(-0.0230217508795)*x[0]**o+(0.118232576333)*x[0]+(0.0810817815326)*x[1]**o+(-0.58292650684)*x[1]
        s[2,2,1,1]=(-0.372568177431)*x[0]**o+(-0.613455954736)*x[0]+(0.223650823501)*x[1]**o+(-0.310111084313)*x[1]
        s[2,2,2,0]=(-0.297770107968)*x[0]**o+(0.334407378326)*x[0]+(0.0820495048815)*x[1]**o+(0.670675755602)*x[1]
        s[2,2,2,1]=(-0.884675664948)*x[0]**o+(-0.141329306491)*x[0]+(0.125135636826)*x[1]**o+(-0.911348264716)*x[1]
        s[2,3,0,0]=(0.538935527955)*x[0]**o+(0.513300520882)*x[0]+(0.850744598079)*x[1]**o+(0.751590880335)*x[1]
        s[2,3,0,1]=(-0.821182874204)*x[0]**o+(-0.186188144846)*x[0]+(-0.107871578183)*x[1]**o+(0.462281422576)*x[1]
        s[2,3,1,0]=(0.699998387809)*x[0]**o+(0.497080130045)*x[0]+(-0.876146393883)*x[1]**o+(-0.966792776701)*x[1]
        s[2,3,1,1]=(-0.755486256523)*x[0]**o+(0.482429742628)*x[0]+(0.64207235708)*x[1]**o+(0.792668637521)*x[1]
        s[2,3,2,0]=(0.480810051431)*x[0]**o+(0.526844732935)*x[0]+(0.358083926997)*x[1]**o+(0.975251790749)*x[1]
        s[2,3,2,1]=(-0.71380930491)*x[0]**o+(0.321062714102)*x[0]+(-0.393765851515)*x[1]**o+(0.758037335781)*x[1]
        s[2,4,0,0]=(0.799658695022)*x[0]**o+(-0.970884788767)*x[0]+(-0.909912887472)*x[1]**o+(-0.118569854978)*x[1]
        s[2,4,0,1]=(0.22665071178)*x[0]**o+(-0.866764191923)*x[0]+(0.760310143236)*x[1]**o+(0.979622869213)*x[1]
        s[2,4,1,0]=(-0.639860894248)*x[0]**o+(-0.254761334846)*x[0]+(0.844133251872)*x[1]**o+(-0.0823600617007)*x[1]
        s[2,4,1,1]=(0.699864723463)*x[0]**o+(-0.941417107514)*x[0]+(-0.939935969783)*x[1]**o+(0.307212220608)*x[1]
        s[2,4,2,0]=(-0.597501168242)*x[0]**o+(-0.855810189779)*x[0]+(-0.616246893869)*x[1]**o+(0.862508316402)*x[1]
        s[2,4,2,1]=(0.279133829297)*x[0]**o+(-0.963434838906)*x[0]+(-0.927999672088)*x[1]**o+(-0.633147829562)*x[1]
        s[3,0,0,0]=(-0.228495811891)*x[0]**o+(-0.253261426268)*x[0]+(-0.487641848136)*x[1]**o+(0.321239316844)*x[1]
        s[3,0,0,1]=(-0.226385127812)*x[0]**o+(-0.903861474811)*x[0]+(0.0285492108309)*x[1]**o+(0.199168585275)*x[1]
        s[3,0,1,0]=(-0.0670214317839)*x[0]**o+(-0.169290931005)*x[0]+(0.859675236893)*x[1]**o+(0.673436325782)*x[1]
        s[3,0,1,1]=(0.589201260546)*x[0]**o+(-0.931566266675)*x[0]+(0.127884139041)*x[1]**o+(-0.0831271642138)*x[1]
        s[3,0,2,0]=(-0.228265145653)*x[0]**o+(-0.449360641451)*x[0]+(0.0895266621221)*x[1]**o+(0.516248722513)*x[1]
        s[3,0,2,1]=(0.726271772612)*x[0]**o+(-0.834727339534)*x[0]+(0.779809207203)*x[1]**o+(-0.2745165742)*x[1]
        s[3,1,0,0]=(-0.528300011184)*x[0]**o+(0.0728636912947)*x[0]+(-0.953454233615)*x[1]**o+(0.582219621875)*x[1]
        s[3,1,0,1]=(0.196609308635)*x[0]**o+(0.162717112564)*x[0]+(0.449469968326)*x[1]**o+(-0.174119135554)*x[1]
        s[3,1,1,0]=(-0.261855816805)*x[0]**o+(0.607391359963)*x[0]+(-0.793775568608)*x[1]**o+(-0.950278043961)*x[1]
        s[3,1,1,1]=(0.240905287034)*x[0]**o+(-0.204299485241)*x[0]+(-0.825848364133)*x[1]**o+(-0.00971732871232)*x[1]
        s[3,1,2,0]=(0.798049323787)*x[0]**o+(0.0710454406309)*x[0]+(0.990006532214)*x[1]**o+(0.711055965038)*x[1]
        s[3,1,2,1]=(-0.406803353709)*x[0]**o+(0.120946506707)*x[0]+(-0.410263830047)*x[1]**o+(0.988997635486)*x[1]
        s[3,2,0,0]=(0.181718966759)*x[0]**o+(-0.0262352524424)*x[0]+(-0.1917125307)*x[1]**o+(-0.169130797999)*x[1]
        s[3,2,0,1]=(-0.00378400436872)*x[0]**o+(-0.970892218185)*x[0]+(-0.563995443607)*x[1]**o+(0.906996441633)*x[1]
        s[3,2,1,0]=(-0.858412521643)*x[0]**o+(-0.133539980742)*x[0]+(-0.692474245554)*x[1]**o+(0.0128241106616)*x[1]
        s[3,2,1,1]=(-0.272651295849)*x[0]**o+(-0.185053306751)*x[0]+(-0.034278709119)*x[1]**o+(-0.201998293805)*x[1]
        s[3,2,2,0]=(-0.0376999069998)*x[0]**o+(-0.895215920335)*x[0]+(0.944947071273)*x[1]**o+(-0.732964572204)*x[1]
        s[3,2,2,1]=(0.0771318167573)*x[0]**o+(0.051862360377)*x[0]+(-0.837260137737)*x[1]**o+(-0.770673420806)*x[1]
        s[3,3,0,0]=(-0.258989060244)*x[0]**o+(-0.699818576225)*x[0]+(-0.311518509076)*x[1]**o+(0.517397173211)*x[1]
        s[3,3,0,1]=(-0.443559295683)*x[0]**o+(-0.0544611458142)*x[0]+(0.983503123196)*x[1]**o+(0.248879114721)*x[1]
        s[3,3,1,0]=(0.626729673965)*x[0]**o+(0.347363767931)*x[0]+(0.956080389131)*x[1]**o+(0.429076008118)*x[1]
        s[3,3,1,1]=(-0.508758891471)*x[0]**o+(0.616615567996)*x[0]+(0.221644986884)*x[1]**o+(-0.904935728116)*x[1]
        s[3,3,2,0]=(0.254341631925)*x[0]**o+(-0.908688314004)*x[0]+(-0.0449719498093)*x[1]**o+(0.871189405684)*x[1]
        s[3,3,2,1]=(0.506459636421)*x[0]**o+(0.978066502797)*x[0]+(-0.212934438396)*x[1]**o+(-0.355665670198)*x[1]
        s[3,4,0,0]=(0.0152292845756)*x[0]**o+(-0.400615592376)*x[0]+(0.264352702586)*x[1]**o+(-0.0677780689342)*x[1]
        s[3,4,0,1]=(0.734077033298)*x[0]**o+(-0.171218476666)*x[0]+(-0.399285387071)*x[1]**o+(-0.589596788961)*x[1]
        s[3,4,1,0]=(-0.903602643334)*x[0]**o+(0.0527455440694)*x[0]+(-0.973273256059)*x[1]**o+(0.76054623312)*x[1]
        s[3,4,1,1]=(0.368219867735)*x[0]**o+(0.857361275285)*x[0]+(0.297273583978)*x[1]**o+(0.243120582563)*x[1]
        s[3,4,2,0]=(0.518637118464)*x[0]**o+(-0.961371552746)*x[0]+(0.283809772558)*x[1]**o+(-0.947184379737)*x[1]
        s[3,4,2,1]=(-0.514398206056)*x[0]**o+(-0.499373736713)*x[0]+(-0.756154352548)*x[1]**o+(0.814078333554)*x[1]
        ref[0,0,0,0]=(0.451668581322)/(o+1.)+(-0.286819576397)
        ref[0,0,0,1]=(0.30725041959)/(o+1.)+(-0.256315624701)
        ref[0,0,1,0]=(-0.466895770987)/(o+1.)+(-0.293859794895)
        ref[0,0,1,1]=(0.489204046518)/(o+1.)+(0.340743081611)
        ref[0,0,2,0]=(-0.838470513606)/(o+1.)+(-0.40550734315)
        ref[0,0,2,1]=(0.117160355949)/(o+1.)+(0.162017541826)
        ref[0,1,0,0]=(0.748036440634)/(o+1.)+(0.0996914366319)
        ref[0,1,0,1]=(-0.654754641301)/(o+1.)+(0.305344541376)
        ref[0,1,1,0]=(0.568069211159)/(o+1.)+(-0.115673717252)
        ref[0,1,1,1]=(0.0652687572678)/(o+1.)+(0.275595203627)
        ref[0,1,2,0]=(0.371255897455)/(o+1.)+(-0.315249837692)
        ref[0,1,2,1]=(-0.331807603614)/(o+1.)+(-0.0345187874006)
        ref[0,2,0,0]=(0.0301821364712)/(o+1.)+(-0.543550938487)
        ref[0,2,0,1]=(-0.112351461929)/(o+1.)+(-0.263750126822)
        ref[0,2,1,0]=(1.80616095181)/(o+1.)+(0.371348899028)
        ref[0,2,1,1]=(1.60373586077)/(o+1.)+(-0.0455343718196)
        ref[0,2,2,0]=(0.0939388699401)/(o+1.)+(0.36514946063)
        ref[0,2,2,1]=(0.362945697503)/(o+1.)+(0.596779059003)
        ref[0,3,0,0]=(0.0671826683913)/(o+1.)+(-0.104364709744)
        ref[0,3,0,1]=(-0.52668922709)/(o+1.)+(-0.526940958341)
        ref[0,3,1,0]=(-0.954896408656)/(o+1.)+(-0.171395509929)
        ref[0,3,1,1]=(-0.689541395265)/(o+1.)+(-0.338764569932)
        ref[0,3,2,0]=(-0.321974090548)/(o+1.)+(0.12933691892)
        ref[0,3,2,1]=(-0.196819244859)/(o+1.)+(0.0479141252065)
        ref[0,4,0,0]=(0.591626095276)/(o+1.)+(-0.061367138182)
        ref[0,4,0,1]=(-0.0365003105586)/(o+1.)+(0.37072015298)
        ref[0,4,1,0]=(0.28450482971)/(o+1.)+(0.348656950064)
        ref[0,4,1,1]=(-0.0810103121392)/(o+1.)+(-0.405658582232)
        ref[0,4,2,0]=(1.2124419766)/(o+1.)+(-0.135187779544)
        ref[0,4,2,1]=(0.678642108977)/(o+1.)+(0.167163198411)
        ref[1,0,0,0]=(0.478074090207)/(o+1.)+(0.066797984161)
        ref[1,0,0,1]=(-0.628487340845)/(o+1.)+(-0.0380582788988)
        ref[1,0,1,0]=(0.276536656806)/(o+1.)+(0.400354937182)
        ref[1,0,1,1]=(-0.191819018239)/(o+1.)+(-0.311393227243)
        ref[1,0,2,0]=(-0.81424126855)/(o+1.)+(0.0150408171462)
        ref[1,0,2,1]=(-1.09238112948)/(o+1.)+(0.0339781739475)
        ref[1,1,0,0]=(-0.142980193819)/(o+1.)+(0.202001738951)
        ref[1,1,0,1]=(0.370136500233)/(o+1.)+(0.575973612395)
        ref[1,1,1,0]=(-0.45376369667)/(o+1.)+(-0.242645449715)
        ref[1,1,1,1]=(1.11954469542)/(o+1.)+(-0.56645753595)
        ref[1,1,2,0]=(0.601986418224)/(o+1.)+(-0.702065271552)
        ref[1,1,2,1]=(-1.08228378316)/(o+1.)+(-0.662334514973)
        ref[1,2,0,0]=(1.70235503702)/(o+1.)+(-0.61035468193)
        ref[1,2,0,1]=(0.452592302963)/(o+1.)+(-0.255094436796)
        ref[1,2,1,0]=(0.984814885149)/(o+1.)+(-0.218291749321)
        ref[1,2,1,1]=(0.816670845082)/(o+1.)+(-0.0381266223775)
        ref[1,2,2,0]=(-0.646553014327)/(o+1.)+(0.278596216015)
        ref[1,2,2,1]=(-0.229990197753)/(o+1.)+(-0.34453600348)
        ref[1,3,0,0]=(0.435488145837)/(o+1.)+(-0.0652345192758)
        ref[1,3,0,1]=(1.24688535347)/(o+1.)+(0.15650216793)
        ref[1,3,1,0]=(-0.545174093939)/(o+1.)+(0.453223383109)
        ref[1,3,1,1]=(1.10195725482)/(o+1.)+(-0.0808676674551)
        ref[1,3,2,0]=(0.411690213595)/(o+1.)+(0.144928120921)
        ref[1,3,2,1]=(0.0234401004369)/(o+1.)+(0.693115608963)
        ref[1,4,0,0]=(-0.874824048052)/(o+1.)+(0.680318926188)
        ref[1,4,0,1]=(-0.997130152255)/(o+1.)+(-0.338433885053)
        ref[1,4,1,0]=(0.774863907537)/(o+1.)+(-0.257398794152)
        ref[1,4,1,1]=(1.24530221396)/(o+1.)+(0.337719405527)
        ref[1,4,2,0]=(0.000454995225935)/(o+1.)+(-0.404361934896)
        ref[1,4,2,1]=(-0.126291388789)/(o+1.)+(0.171494723319)
        ref[2,0,0,0]=(-1.29691046589)/(o+1.)+(-0.559794581183)
        ref[2,0,0,1]=(-0.783334024169)/(o+1.)+(0.348919629629)
        ref[2,0,1,0]=(-0.273796570692)/(o+1.)+(0.0502719774097)
        ref[2,0,1,1]=(-0.614905915315)/(o+1.)+(-0.0383542229546)
        ref[2,0,2,0]=(-0.771884828178)/(o+1.)+(-0.920183288409)
        ref[2,0,2,1]=(-0.50769895762)/(o+1.)+(0.296538745675)
        ref[2,1,0,0]=(0.628717795874)/(o+1.)+(0.220296655812)
        ref[2,1,0,1]=(1.02905688839)/(o+1.)+(-0.480494271001)
        ref[2,1,1,0]=(0.00408858801768)/(o+1.)+(0.787430052788)
        ref[2,1,1,1]=(-0.933232194463)/(o+1.)+(0.250357246304)
        ref[2,1,2,0]=(0.944395204286)/(o+1.)+(-0.17880063408)
        ref[2,1,2,1]=(0.945384763069)/(o+1.)+(0.294593887985)
        ref[2,2,0,0]=(-0.288902049932)/(o+1.)+(0.0333322750744)
        ref[2,2,0,1]=(-0.538619256435)/(o+1.)+(0.244127045883)
        ref[2,2,1,0]=(0.0580600306531)/(o+1.)+(-0.232346965253)
        ref[2,2,1,1]=(-0.148917353929)/(o+1.)+(-0.461783519524)
        ref[2,2,2,0]=(-0.215720603086)/(o+1.)+(0.502541566964)
        ref[2,2,2,1]=(-0.759540028122)/(o+1.)+(-0.526338785603)
        ref[2,3,0,0]=(1.38968012603)/(o+1.)+(0.632445700608)
        ref[2,3,0,1]=(-0.929054452386)/(o+1.)+(0.138046638865)
        ref[2,3,1,0]=(-0.176148006073)/(o+1.)+(-0.234856323328)
        ref[2,3,1,1]=(-0.113413899443)/(o+1.)+(0.637549190074)
        ref[2,3,2,0]=(0.838893978428)/(o+1.)+(0.751048261842)
        ref[2,3,2,1]=(-1.10757515642)/(o+1.)+(0.539550024941)
        ref[2,4,0,0]=(-0.11025419245)/(o+1.)+(-0.544727321873)
        ref[2,4,0,1]=(0.986960855016)/(o+1.)+(0.056429338645)
        ref[2,4,1,0]=(0.204272357624)/(o+1.)+(-0.168560698273)
        ref[2,4,1,1]=(-0.24007124632)/(o+1.)+(-0.317102443453)
        ref[2,4,2,0]=(-1.21374806211)/(o+1.)+(0.00334906331113)
        ref[2,4,2,1]=(-0.648865842791)/(o+1.)+(-0.798291334234)
        ref[3,0,0,0]=(-0.716137660027)/(o+1.)+(0.033988945288)
        ref[3,0,0,1]=(-0.197835916981)/(o+1.)+(-0.352346444768)
        ref[3,0,1,0]=(0.792653805109)/(o+1.)+(0.252072697388)
        ref[3,0,1,1]=(0.717085399588)/(o+1.)+(-0.507346715445)
        ref[3,0,2,0]=(-0.138738483531)/(o+1.)+(0.0334440405313)
        ref[3,0,2,1]=(1.50608097982)/(o+1.)+(-0.554621956867)
        ref[3,1,0,0]=(-1.4817542448)/(o+1.)+(0.327541656585)
        ref[3,1,0,1]=(0.64607927696)/(o+1.)+(-0.00570101149523)
        ref[3,1,1,0]=(-1.05563138541)/(o+1.)+(-0.171443341999)
        ref[3,1,1,1]=(-0.5849430771)/(o+1.)+(-0.107008406977)
        ref[3,1,2,0]=(1.788055856)/(o+1.)+(0.391050702835)
        ref[3,1,2,1]=(-0.817067183756)/(o+1.)+(0.554972071097)
        ref[3,2,0,0]=(-0.00999356394068)/(o+1.)+(-0.0976830252209)
        ref[3,2,0,1]=(-0.567779447976)/(o+1.)+(-0.031947888276)
        ref[3,2,1,0]=(-1.5508867672)/(o+1.)+(-0.0603579350402)
        ref[3,2,1,1]=(-0.306930004968)/(o+1.)+(-0.193525800278)
        ref[3,2,2,0]=(0.907247164273)/(o+1.)+(-0.81409024627)
        ref[3,2,2,1]=(-0.76012832098)/(o+1.)+(-0.359405530215)
        ref[3,3,0,0]=(-0.57050756932)/(o+1.)+(-0.0912107015068)
        ref[3,3,0,1]=(0.539943827514)/(o+1.)+(0.0972089844535)
        ref[3,3,1,0]=(1.5828100631)/(o+1.)+(0.388219888025)
        ref[3,3,1,1]=(-0.287113904587)/(o+1.)+(-0.14416008006)
        ref[3,3,2,0]=(0.209369682116)/(o+1.)+(-0.01874945416)
        ref[3,3,2,1]=(0.293525198025)/(o+1.)+(0.311200416299)
        ref[3,4,0,0]=(0.279581987162)/(o+1.)+(-0.234196830655)
        ref[3,4,0,1]=(0.334791646227)/(o+1.)+(-0.380407632814)
        ref[3,4,1,0]=(-1.87687589939)/(o+1.)+(0.406645888595)
        ref[3,4,1,1]=(0.665493451713)/(o+1.)+(0.550240928924)
        ref[3,4,2,0]=(0.802446891023)/(o+1.)+(-0.954277966241)
        ref[3,4,2,1]=(-1.2705525586)/(o+1.)+(0.15735229842)
      else:
        s[0,0,0,0]=(0.641376947921)*x[0]**o+(-0.761647983943)*x[0]+(0.1869803104)*x[1]**o+(-0.325986059158)*x[1]+(-0.0520537723186)*x[2]**o+(-0.242075762859)*x[2]
        s[0,0,0,1]=(0.774456605418)*x[0]**o+(0.321485563379)*x[0]+(0.0248450395096)*x[1]**o+(-0.425638835963)*x[1]+(-0.0622024423983)*x[2]**o+(0.0431057597271)*x[2]
        s[0,0,1,0]=(0.661432460103)*x[0]**o+(-0.653382716149)*x[0]+(0.596912471264)*x[1]**o+(-0.375947479542)*x[1]+(0.370999295823)*x[2]**o+(0.791872078833)*x[2]
        s[0,0,1,1]=(0.258860589295)*x[0]**o+(-0.802340135742)*x[0]+(0.569149399765)*x[1]**o+(-0.505649920225)*x[1]+(0.384267048627)*x[2]**o+(-0.791015249057)*x[2]
        s[0,0,2,0]=(0.219942214071)*x[0]**o+(0.360851774828)*x[0]+(-0.456714182675)*x[1]**o+(0.0965079164275)*x[1]+(-0.752467360132)*x[2]**o+(-0.285016148963)*x[2]
        s[0,0,2,1]=(0.227249920303)*x[0]**o+(-0.63371992192)*x[0]+(-0.751996139921)*x[1]**o+(-0.804517033328)*x[1]+(-0.0202988256583)*x[2]**o+(0.531173510334)*x[2]
        s[0,1,0,0]=(0.530838912723)*x[0]**o+(0.230970767172)*x[0]+(-0.424131392941)*x[1]**o+(-0.606417012546)*x[1]+(-0.318937522783)*x[2]**o+(0.179390519765)*x[2]
        s[0,1,0,1]=(-0.776569569832)*x[0]**o+(0.918265838431)*x[0]+(-0.091878191762)*x[1]**o+(0.698951377419)*x[1]+(0.244940000061)*x[2]**o+(0.0452610075428)*x[2]
        s[0,1,1,0]=(0.435082117831)*x[0]**o+(0.889881605799)*x[0]+(-0.219619308431)*x[1]**o+(-0.672303092957)*x[1]+(0.987619326906)*x[2]**o+(0.220516905778)*x[2]
        s[0,1,1,1]=(-0.118683589446)*x[0]**o+(0.965421464735)*x[0]+(-0.863562140452)*x[1]**o+(-0.231242073932)*x[1]+(0.568240493507)*x[2]**o+(0.734578820791)*x[2]
        s[0,1,2,0]=(0.511447709837)*x[0]**o+(-0.274507694143)*x[0]+(0.576288098696)*x[1]**o+(0.655476727813)*x[1]+(0.602315737594)*x[2]**o+(0.817938429306)*x[2]
        s[0,1,2,1]=(-0.319953701981)*x[0]**o+(-0.0816890923671)*x[0]+(0.84471771093)*x[1]**o+(0.216072534675)*x[1]+(0.782233227772)*x[2]**o+(0.761080396806)*x[2]
        s[0,2,0,0]=(0.371971964019)*x[0]**o+(-0.914300658178)*x[0]+(-0.421914056897)*x[1]**o+(-0.366900498221)*x[1]+(-0.224742744541)*x[2]**o+(0.718117565395)*x[2]
        s[0,2,0,1]=(0.390948765614)*x[0]**o+(0.523332138936)*x[0]+(0.207458000485)*x[1]**o+(0.704266764721)*x[1]+(0.194539635633)*x[2]**o+(-0.0310558120915)*x[2]
        s[0,2,1,0]=(-0.00308409831643)*x[0]**o+(0.0874338943679)*x[0]+(0.792479180633)*x[1]**o+(-0.840935676827)*x[1]+(0.183535198853)*x[2]**o+(-0.632484318094)*x[2]
        s[0,2,1,1]=(0.93055010195)*x[0]**o+(-0.592883255994)*x[0]+(-0.293724124968)*x[1]**o+(-0.532867438247)*x[1]+(-0.731629824662)*x[2]**o+(-0.926457959344)*x[2]
        s[0,2,2,0]=(-0.410766259996)*x[0]**o+(0.153463001676)*x[0]+(-0.22434101683)*x[1]**o+(0.792184946048)*x[1]+(0.217569318007)*x[2]**o+(-0.746503700782)*x[2]
        s[0,2,2,1]=(-0.48097027987)*x[0]**o+(-0.855259653713)*x[0]+(0.297296031945)*x[1]**o+(0.486702064445)*x[1]+(-0.15982283079)*x[2]**o+(0.943473459324)*x[2]
        s[0,3,0,0]=(0.440444162838)*x[0]**o+(-0.411163880392)*x[0]+(0.434891712314)*x[1]**o+(0.421032339491)*x[1]+(0.185417751167)*x[2]**o+(0.774341796864)*x[2]
        s[0,3,0,1]=(0.418605026678)*x[0]**o+(0.697823127475)*x[0]+(-0.490471272795)*x[1]**o+(0.889959972523)*x[1]+(-0.935571921221)*x[2]**o+(-0.477163919777)*x[2]
        s[0,3,1,0]=(0.240863375001)*x[0]**o+(0.171535577067)*x[0]+(-0.798318413682)*x[1]**o+(0.165523160187)*x[1]+(-0.793244034855)*x[2]**o+(-0.639277575041)*x[2]
        s[0,3,1,1]=(0.202212256877)*x[0]**o+(0.387471670954)*x[0]+(-0.198376856629)*x[1]**o+(-0.679099770813)*x[1]+(-0.213566318573)*x[2]**o+(-0.591165980661)*x[2]
        s[0,3,2,0]=(0.764050063724)*x[0]**o+(-0.921472809292)*x[0]+(-0.483029501792)*x[1]**o+(0.198096560699)*x[1]+(-0.161449532882)*x[2]**o+(-0.854460510477)*x[2]
        s[0,3,2,1]=(0.491500019407)*x[0]**o+(0.830069678076)*x[0]+(-0.597452015974)*x[1]**o+(-0.893928655959)*x[1]+(0.940149982314)*x[2]**o+(0.17300719331)*x[2]
        s[0,4,0,0]=(0.0798620352831)*x[0]**o+(0.0237937127416)*x[0]+(-0.086351109139)*x[1]**o+(0.729882388424)*x[1]+(-0.7065299588)*x[2]**o+(-0.632673206138)*x[2]
        s[0,4,0,1]=(-0.809989577091)*x[0]**o+(0.0981392253023)*x[0]+(0.831656529801)*x[1]**o+(0.0337424060561)*x[1]+(-0.210067786784)*x[2]**o+(-0.0549477247313)*x[2]
        s[0,4,1,0]=(-0.0632498403207)*x[0]**o+(-0.992515766051)*x[0]+(0.579065868983)*x[1]**o+(0.201496477221)*x[1]+(0.147678629343)*x[2]**o+(0.396978535826)*x[2]
        s[0,4,1,1]=(0.538470621207)*x[0]**o+(0.125447714672)*x[0]+(0.345561739505)*x[1]**o+(-0.898311293341)*x[1]+(0.695965393051)*x[2]**o+(0.456779053964)*x[2]
        s[0,4,2,0]=(-0.461938934992)*x[0]**o+(0.815268894146)*x[0]+(0.529702893063)*x[1]**o+(0.106752164506)*x[1]+(-0.892819156049)*x[2]**o+(-0.540251931665)*x[2]
        s[0,4,2,1]=(-0.890961526208)*x[0]**o+(-0.0522965595586)*x[0]+(-0.166811931177)*x[1]**o+(-0.32795191314)*x[1]+(0.827693520674)*x[2]**o+(-0.858862162004)*x[2]
        s[1,0,0,0]=(0.0341612512071)*x[0]**o+(0.704407020254)*x[0]+(-0.0212728069872)*x[1]**o+(-0.360120370601)*x[1]+(0.0355417904312)*x[2]**o+(-0.899016835477)*x[2]
        s[1,0,0,1]=(-0.254336879728)*x[0]**o+(-0.749583807992)*x[0]+(-0.822540014653)*x[1]**o+(0.321686636614)*x[1]+(-0.110121756691)*x[2]**o+(0.995172970249)*x[2]
        s[1,0,1,0]=(0.538783330989)*x[0]**o+(0.425931208203)*x[0]+(0.808850719257)*x[1]**o+(-0.616824777466)*x[1]+(-0.999643426203)*x[2]**o+(0.206956920828)*x[2]
        s[1,0,1,1]=(0.92122724205)*x[0]**o+(-0.0077578775328)*x[0]+(-0.455369944999)*x[1]**o+(0.069992464289)*x[1]+(-0.625922232223)*x[2]**o+(0.308582889054)*x[2]
        s[1,0,2,0]=(0.812275176702)*x[0]**o+(-0.0737918971904)*x[0]+(-0.352834721968)*x[1]**o+(-0.485715856695)*x[1]+(0.893012104541)*x[2]**o+(-0.50277227482)*x[2]
        s[1,0,2,1]=(0.327001325412)*x[0]**o+(0.248420422978)*x[0]+(0.934489184596)*x[1]**o+(0.645665157141)*x[1]+(-0.125952174434)*x[2]**o+(-0.589610796444)*x[2]
        s[1,1,0,0]=(-0.19439618112)*x[0]**o+(-0.473153238554)*x[0]+(0.441593313035)*x[1]**o+(0.560223203285)*x[1]+(0.744701337265)*x[2]**o+(0.588942879247)*x[2]
        s[1,1,0,1]=(0.956622537359)*x[0]**o+(0.388369191252)*x[0]+(-0.912880779816)*x[1]**o+(-0.930624449932)*x[1]+(-0.0693131490103)*x[2]**o+(0.79864169138)*x[2]
        s[1,1,1,0]=(0.0802315160845)*x[0]**o+(0.360021747567)*x[0]+(-0.838148172909)*x[1]**o+(0.974574791036)*x[1]+(0.134930445828)*x[2]**o+(-0.224819866647)*x[2]
        s[1,1,1,1]=(0.0408467896812)*x[0]**o+(-0.527931565684)*x[0]+(-0.294779051686)*x[1]**o+(0.655648246879)*x[1]+(0.206641334088)*x[2]**o+(0.244207224731)*x[2]
        s[1,1,2,0]=(-0.695324333815)*x[0]**o+(-0.76302643793)*x[0]+(-0.445448086412)*x[1]**o+(0.387620295728)*x[1]+(0.410986742302)*x[2]**o+(-0.903330791078)*x[2]
        s[1,1,2,1]=(0.725983511781)*x[0]**o+(-0.55616923731)*x[0]+(0.694922396401)*x[1]**o+(0.969679789365)*x[1]+(0.390140728655)*x[2]**o+(0.616506146743)*x[2]
        s[1,2,0,0]=(-0.847983007272)*x[0]**o+(-0.21379616326)*x[0]+(-0.221990648875)*x[1]**o+(0.542144543115)*x[1]+(-0.476016238668)*x[2]**o+(0.239835257111)*x[2]
        s[1,2,0,1]=(0.382592141805)*x[0]**o+(-0.547322693057)*x[0]+(0.207135932694)*x[1]**o+(0.479160574257)*x[1]+(-0.406772807762)*x[2]**o+(0.0827537744696)*x[2]
        s[1,2,1,0]=(0.385177578855)*x[0]**o+(-0.96749949076)*x[0]+(-0.980528334919)*x[1]**o+(0.846145425556)*x[1]+(0.663229596459)*x[2]**o+(0.234589730098)*x[2]
        s[1,2,1,1]=(0.326538279652)*x[0]**o+(0.864444241537)*x[0]+(0.925715710223)*x[1]**o+(0.901973119241)*x[1]+(0.489827899451)*x[2]**o+(-0.446654474045)*x[2]
        s[1,2,2,0]=(0.623521048768)*x[0]**o+(0.00694630100059)*x[0]+(-0.316266037281)*x[1]**o+(-0.296669528851)*x[1]+(-0.345499066232)*x[2]**o+(-0.781251040298)*x[2]
        s[1,2,2,1]=(-0.564332000062)*x[0]**o+(-0.838410587038)*x[0]+(0.105923226211)*x[1]**o+(0.027080971794)*x[1]+(0.656502442527)*x[2]**o+(-0.968700980875)*x[2]
        s[1,3,0,0]=(0.28297404936)*x[0]**o+(0.906254282744)*x[0]+(0.426459225459)*x[1]**o+(0.474780047643)*x[1]+(0.915469451254)*x[2]**o+(0.150014279679)*x[2]
        s[1,3,0,1]=(0.417919974897)*x[0]**o+(0.118528892402)*x[0]+(0.456086489559)*x[1]**o+(0.407436737785)*x[1]+(-0.594040737087)*x[2]**o+(-0.0968244395915)*x[2]
        s[1,3,1,0]=(-0.00608149061984)*x[0]**o+(0.279861847458)*x[0]+(0.622153448276)*x[1]**o+(0.307580940395)*x[1]+(-0.311958042371)*x[2]**o+(0.951637572927)*x[2]
        s[1,3,1,1]=(0.0620010964861)*x[0]**o+(0.694489586135)*x[0]+(-0.397214573709)*x[1]**o+(0.816641484229)*x[1]+(-0.55010098877)*x[2]**o+(-0.252677099684)*x[2]
        s[1,3,2,0]=(0.347469613465)*x[0]**o+(-0.793025539854)*x[0]+(0.340939082154)*x[1]**o+(0.00135134646417)*x[1]+(-0.253753620216)*x[2]**o+(-0.795217820163)*x[2]
        s[1,3,2,1]=(-0.77663362989)*x[0]**o+(-0.0454794482729)*x[0]+(0.0346722039087)*x[1]**o+(-0.137241685281)*x[1]+(0.984186419894)*x[2]**o+(0.884865835233)*x[2]
        s[1,4,0,0]=(-0.344544201299)*x[0]**o+(0.118604431816)*x[0]+(0.735491059587)*x[1]**o+(-0.474594418844)*x[1]+(-0.160183641095)*x[2]**o+(0.0442675403903)*x[2]
        s[1,4,0,1]=(0.159437065336)*x[0]**o+(0.357454800453)*x[0]+(0.0623836301056)*x[1]**o+(0.255296236156)*x[1]+(0.586460110084)*x[2]**o+(-0.557870816384)*x[2]
        s[1,4,1,0]=(0.272212988717)*x[0]**o+(-0.00218257591936)*x[0]+(0.637653920938)*x[1]**o+(0.474154612654)*x[1]+(0.0927963495169)*x[2]**o+(0.688169553045)*x[2]
        s[1,4,1,1]=(-0.296752052135)*x[0]**o+(0.222825403373)*x[0]+(-0.61297870931)*x[1]**o+(-0.772169949259)*x[1]+(0.267924738357)*x[2]**o+(-0.98640521973)*x[2]
        s[1,4,2,0]=(-0.528327447355)*x[0]**o+(-0.13640459815)*x[0]+(-0.839608161898)*x[1]**o+(0.79996626812)*x[1]+(0.0889785028146)*x[2]**o+(0.483698551365)*x[2]
        s[1,4,2,1]=(-0.888180576683)*x[0]**o+(-0.243274686308)*x[0]+(-0.459488747018)*x[1]**o+(0.770881162569)*x[1]+(-0.502642334958)*x[2]**o+(-0.0596414308521)*x[2]
        s[2,0,0,0]=(-0.765317614505)*x[0]**o+(-0.220818348989)*x[0]+(-0.185228616416)*x[1]**o+(-0.0787003843488)*x[1]+(-0.412520273332)*x[2]**o+(0.192760450932)*x[2]
        s[2,0,0,1]=(0.913449879489)*x[0]**o+(-0.219345049089)*x[0]+(0.54504962001)*x[1]**o+(0.623236284643)*x[1]+(-0.350717162156)*x[2]**o+(-0.623681076852)*x[2]
        s[2,0,1,0]=(-0.940664057127)*x[0]**o+(0.251319839352)*x[0]+(0.877556299495)*x[1]**o+(0.016160084659)*x[1]+(-0.544235646954)*x[2]**o+(-0.343760171279)*x[2]
        s[2,0,1,1]=(0.324227271021)*x[0]**o+(-0.380692454159)*x[0]+(-0.727849414028)*x[1]**o+(0.955281983149)*x[1]+(-0.898073831095)*x[2]**o+(0.858127409364)*x[2]
        s[2,0,2,0]=(-0.111449036516)*x[0]**o+(0.928827923826)*x[0]+(-0.992899512514)*x[1]**o+(0.0150334039442)*x[1]+(-0.552135666421)*x[2]**o+(0.406949591012)*x[2]
        s[2,0,2,1]=(0.241832900797)*x[0]**o+(0.684688614786)*x[0]+(0.842064214069)*x[1]**o+(-0.718696633586)*x[1]+(0.352196331118)*x[2]**o+(-0.260343576356)*x[2]
        s[2,1,0,0]=(0.426302002061)*x[0]**o+(0.573786857786)*x[0]+(0.909827742357)*x[1]**o+(0.34656984965)*x[1]+(0.379863389758)*x[2]**o+(0.0322921482901)*x[2]
        s[2,1,0,1]=(0.34423098891)*x[0]**o+(0.959825965873)*x[0]+(-0.458147326954)*x[1]**o+(-0.190830182513)*x[1]+(-0.0836184656488)*x[2]**o+(0.711921183181)*x[2]
        s[2,1,1,0]=(0.813184534496)*x[0]**o+(-0.551568400877)*x[0]+(0.0282287303962)*x[1]**o+(0.421838849382)*x[1]+(0.312118130223)*x[2]**o+(-0.276104599646)*x[2]
        s[2,1,1,1]=(0.39842349776)*x[0]**o+(0.0505250495963)*x[0]+(-0.398145839899)*x[1]**o+(0.712110112191)*x[1]+(-0.646997386195)*x[2]**o+(-0.584123612783)*x[2]
        s[2,1,2,0]=(-0.722790769831)*x[0]**o+(0.850227076858)*x[0]+(0.204013681627)*x[1]**o+(0.924250450519)*x[1]+(-0.0736995922058)*x[2]**o+(-0.597799712798)*x[2]
        s[2,1,2,1]=(0.314644875138)*x[0]**o+(-0.636988948936)*x[0]+(-0.641641656145)*x[1]**o+(0.788809853074)*x[1]+(0.993531553111)*x[2]**o+(-0.495755039677)*x[2]
        s[2,2,0,0]=(0.633851639171)*x[0]**o+(0.523452525521)*x[0]+(-0.382111999068)*x[1]**o+(-0.63124196783)*x[1]+(-0.147565214833)*x[2]**o+(0.0331370104049)*x[2]
        s[2,2,0,1]=(0.526384634874)*x[0]**o+(0.133282991706)*x[0]+(0.805017307077)*x[1]**o+(-0.266487317047)*x[1]+(0.33606063535)*x[2]**o+(0.805493533241)*x[2]
        s[2,2,1,0]=(0.346357232135)*x[0]**o+(-0.787017955245)*x[0]+(0.30877740144)*x[1]**o+(0.451911959523)*x[1]+(0.747842891938)*x[2]**o+(-0.44945658191)*x[2]
        s[2,2,1,1]=(-0.546211248475)*x[0]**o+(0.247263733668)*x[0]+(-0.594514097121)*x[1]**o+(-0.264637607913)*x[1]+(0.418219952147)*x[2]**o+(-0.816923592759)*x[2]
        s[2,2,2,0]=(0.0563761896319)*x[0]**o+(-0.736507329679)*x[0]+(0.470349998404)*x[1]**o+(-0.932042349224)*x[1]+(0.462401759004)*x[2]**o+(-0.714392002134)*x[2]
        s[2,2,2,1]=(0.240181845755)*x[0]**o+(-0.55763353473)*x[0]+(0.285496776429)*x[1]**o+(0.842557595355)*x[1]+(-0.672368632752)*x[2]**o+(0.298762100435)*x[2]
        s[2,3,0,0]=(0.316677723523)*x[0]**o+(-0.637036430356)*x[0]+(0.00972363132121)*x[1]**o+(-0.30937691623)*x[1]+(-0.757958054703)*x[2]**o+(0.261972786754)*x[2]
        s[2,3,0,1]=(-0.0111918506129)*x[0]**o+(-0.79939173162)*x[0]+(0.237847735286)*x[1]**o+(-0.78668226084)*x[1]+(0.707378088836)*x[2]**o+(-0.934201311773)*x[2]
        s[2,3,1,0]=(-0.0276842306357)*x[0]**o+(0.353182001738)*x[0]+(0.458929280756)*x[1]**o+(0.845593054688)*x[1]+(-0.0437171694787)*x[2]**o+(-0.75271342809)*x[2]
        s[2,3,1,1]=(0.752394290258)*x[0]**o+(-0.179837111909)*x[0]+(0.939093046468)*x[1]**o+(0.460733186456)*x[1]+(0.747517309414)*x[2]**o+(-0.58266983679)*x[2]
        s[2,3,2,0]=(-0.951300235705)*x[0]**o+(-0.350444110401)*x[0]+(-0.904564312825)*x[1]**o+(0.356471509479)*x[1]+(0.0971202512406)*x[2]**o+(0.227621817712)*x[2]
        s[2,3,2,1]=(0.404143843413)*x[0]**o+(-0.911041017819)*x[0]+(-0.567863808091)*x[1]**o+(0.228161895607)*x[1]+(-0.160584718896)*x[2]**o+(-0.951937728172)*x[2]
        s[2,4,0,0]=(-0.284708280204)*x[0]**o+(-0.225655800044)*x[0]+(0.0189816391307)*x[1]**o+(-0.847695463473)*x[1]+(-0.603035310159)*x[2]**o+(0.878440760667)*x[2]
        s[2,4,0,1]=(0.249580969821)*x[0]**o+(0.00196364400716)*x[0]+(-0.580222966198)*x[1]**o+(0.266039411545)*x[1]+(-0.0857426202045)*x[2]**o+(0.297517844516)*x[2]
        s[2,4,1,0]=(-0.847202130689)*x[0]**o+(-0.409645962819)*x[0]+(-0.914867914136)*x[1]**o+(0.464829575727)*x[1]+(0.2358554853)*x[2]**o+(-0.951825555951)*x[2]
        s[2,4,1,1]=(-0.925109552513)*x[0]**o+(0.57253113036)*x[0]+(0.531590407505)*x[1]**o+(-0.812964837422)*x[1]+(-0.354555503225)*x[2]**o+(-0.453054364223)*x[2]
        s[2,4,2,0]=(0.669533379208)*x[0]**o+(0.147512160447)*x[0]+(0.730871788446)*x[1]**o+(-0.2646655671)*x[1]+(-0.890950482274)*x[2]**o+(0.28244962845)*x[2]
        s[2,4,2,1]=(-0.11059763341)*x[0]**o+(0.276920644999)*x[0]+(-0.869145286116)*x[1]**o+(-0.403772968269)*x[1]+(0.319872723054)*x[2]**o+(-0.172168932529)*x[2]
        s[3,0,0,0]=(-0.309323319641)*x[0]**o+(-0.520023896018)*x[0]+(-0.91435931387)*x[1]**o+(-0.994973476595)*x[1]+(-0.372209395986)*x[2]**o+(0.980356617984)*x[2]
        s[3,0,0,1]=(0.741365113034)*x[0]**o+(-0.42466697031)*x[0]+(-0.800148478866)*x[1]**o+(0.891726804851)*x[1]+(0.173578031033)*x[2]**o+(-0.830047254257)*x[2]
        s[3,0,1,0]=(0.292501009071)*x[0]**o+(-0.643570264019)*x[0]+(0.549854764281)*x[1]**o+(-0.359686767247)*x[1]+(-0.242285567291)*x[2]**o+(-0.905609679312)*x[2]
        s[3,0,1,1]=(-0.340043108142)*x[0]**o+(0.654957588422)*x[0]+(-0.709296972181)*x[1]**o+(0.997889126174)*x[1]+(-0.0231005104624)*x[2]**o+(0.736312548493)*x[2]
        s[3,0,2,0]=(-0.724460864229)*x[0]**o+(-0.367115506582)*x[0]+(0.932338820643)*x[1]**o+(0.386420866744)*x[1]+(0.0937739144019)*x[2]**o+(0.980197176095)*x[2]
        s[3,0,2,1]=(-0.123268193799)*x[0]**o+(0.775389790643)*x[0]+(0.374761099968)*x[1]**o+(-0.410566663234)*x[1]+(0.61390455683)*x[2]**o+(-0.424980785321)*x[2]
        s[3,1,0,0]=(-0.931705529761)*x[0]**o+(-0.883420554924)*x[0]+(0.364171802711)*x[1]**o+(-0.120720945763)*x[1]+(-0.264662446682)*x[2]**o+(0.694245672188)*x[2]
        s[3,1,0,1]=(0.507167558745)*x[0]**o+(0.41310141465)*x[0]+(0.951808421277)*x[1]**o+(0.194810425697)*x[1]+(0.656365011417)*x[2]**o+(-0.869365458629)*x[2]
        s[3,1,1,0]=(-0.00813347718406)*x[0]**o+(0.410334738969)*x[0]+(0.749864796873)*x[1]**o+(-0.592730297251)*x[1]+(-0.739489287628)*x[2]**o+(0.620937075039)*x[2]
        s[3,1,1,1]=(-0.598495133204)*x[0]**o+(0.933561890579)*x[0]+(0.711594544245)*x[1]**o+(-0.307136681784)*x[1]+(-0.391014626741)*x[2]**o+(0.237572771088)*x[2]
        s[3,1,2,0]=(0.0815836502329)*x[0]**o+(0.265301269129)*x[0]+(0.151163193573)*x[1]**o+(0.734126505497)*x[1]+(-0.607886397818)*x[2]**o+(-0.96735041904)*x[2]
        s[3,1,2,1]=(0.264092716207)*x[0]**o+(-0.0674826292731)*x[0]+(0.163729761978)*x[1]**o+(-0.76481590065)*x[1]+(0.23022017141)*x[2]**o+(0.0749358120678)*x[2]
        s[3,2,0,0]=(-0.0551562544206)*x[0]**o+(0.529626011841)*x[0]+(0.394111309195)*x[1]**o+(-0.315588836981)*x[1]+(-0.821692537082)*x[2]**o+(0.687214393749)*x[2]
        s[3,2,0,1]=(-0.400905852693)*x[0]**o+(-0.588175641647)*x[0]+(-0.759129469971)*x[1]**o+(-0.234537321811)*x[1]+(-0.506929882395)*x[2]**o+(-0.354376736058)*x[2]
        s[3,2,1,0]=(0.517928718456)*x[0]**o+(0.340801973663)*x[0]+(0.648735816398)*x[1]**o+(0.566053973455)*x[1]+(0.42463576352)*x[2]**o+(0.0873509981339)*x[2]
        s[3,2,1,1]=(0.452898115077)*x[0]**o+(0.122517234268)*x[0]+(-0.98577530096)*x[1]**o+(-0.353319422026)*x[1]+(0.926541753948)*x[2]**o+(-0.393031058333)*x[2]
        s[3,2,2,0]=(0.667750444088)*x[0]**o+(0.00375536977016)*x[0]+(0.268195379568)*x[1]**o+(0.50950653377)*x[1]+(-0.262580606419)*x[2]**o+(-0.417690499605)*x[2]
        s[3,2,2,1]=(-0.16803158003)*x[0]**o+(0.186259658616)*x[0]+(-0.200850683296)*x[1]**o+(-0.443680999529)*x[1]+(0.0208407252003)*x[2]**o+(-0.724292481324)*x[2]
        s[3,3,0,0]=(-0.35345741052)*x[0]**o+(0.217412970886)*x[0]+(0.27307889565)*x[1]**o+(0.267600255726)*x[1]+(-0.973212265009)*x[2]**o+(-0.129667297553)*x[2]
        s[3,3,0,1]=(-0.891104426101)*x[0]**o+(-0.613334258669)*x[0]+(-0.373828493642)*x[1]**o+(-0.0702748602911)*x[1]+(0.236710052698)*x[2]**o+(-0.358872144334)*x[2]
        s[3,3,1,0]=(-0.0928125660128)*x[0]**o+(0.413580667461)*x[0]+(0.8369960454)*x[1]**o+(0.77176338641)*x[1]+(-0.0577873035722)*x[2]**o+(-0.702932886813)*x[2]
        s[3,3,1,1]=(-0.241426495256)*x[0]**o+(0.525912190909)*x[0]+(-0.428192935989)*x[1]**o+(0.892165668837)*x[1]+(0.0334032549752)*x[2]**o+(0.882084846304)*x[2]
        s[3,3,2,0]=(-0.891362453397)*x[0]**o+(-0.339599362928)*x[0]+(0.473329812864)*x[1]**o+(-0.371616732026)*x[1]+(0.660699523287)*x[2]**o+(-0.119309478457)*x[2]
        s[3,3,2,1]=(0.243522118981)*x[0]**o+(0.155411321418)*x[0]+(-0.230275201921)*x[1]**o+(0.63515968566)*x[1]+(0.794898971569)*x[2]**o+(0.409779242619)*x[2]
        s[3,4,0,0]=(-0.854632433212)*x[0]**o+(-0.195836100347)*x[0]+(-0.437665256019)*x[1]**o+(0.576362236865)*x[1]+(-0.0965898541616)*x[2]**o+(0.477972926354)*x[2]
        s[3,4,0,1]=(0.570617061345)*x[0]**o+(-0.0667297487904)*x[0]+(0.81823573301)*x[1]**o+(-0.721189008734)*x[1]+(0.314807619532)*x[2]**o+(-0.376322628683)*x[2]
        s[3,4,1,0]=(-0.580668389702)*x[0]**o+(-0.60760839523)*x[0]+(0.686535934072)*x[1]**o+(0.0779142152446)*x[1]+(0.373571653006)*x[2]**o+(-0.88636603136)*x[2]
        s[3,4,1,1]=(0.987949055842)*x[0]**o+(-0.370309316213)*x[0]+(0.491077591851)*x[1]**o+(0.544978214302)*x[1]+(0.0979248628028)*x[2]**o+(0.0838359159228)*x[2]
        s[3,4,2,0]=(-0.0541483763088)*x[0]**o+(0.491394550602)*x[0]+(0.492110322334)*x[1]**o+(-0.113182373476)*x[1]+(0.807657591761)*x[2]**o+(0.387085950646)*x[2]
        s[3,4,2,1]=(-0.420727803772)*x[0]**o+(-0.105139523222)*x[0]+(-0.51755674286)*x[1]**o+(-0.219056797552)*x[1]+(-0.0414122967561)*x[2]**o+(-0.161491064048)*x[2]
        ref[0,0,0,0]=(0.776303486002)/(o+1.)+(-0.66485490298)
        ref[0,0,0,1]=(0.73709920253)/(o+1.)+(-0.0305237564282)
        ref[0,0,1,0]=(1.62934422719)/(o+1.)+(-0.118729058429)
        ref[0,0,1,1]=(1.21227703769)/(o+1.)+(-1.04950265251)
        ref[0,0,2,0]=(-0.989239328736)/(o+1.)+(0.0861717711464)
        ref[0,0,2,1]=(-0.545045045276)/(o+1.)+(-0.453531722457)
        ref[0,1,0,0]=(-0.212230003001)/(o+1.)+(-0.0980278628049)
        ref[0,1,0,1]=(-0.623507761533)/(o+1.)+(0.831239111696)
        ref[0,1,1,0]=(1.20308213631)/(o+1.)+(0.21904770931)
        ref[0,1,1,1]=(-0.414005236391)/(o+1.)+(0.734379105797)
        ref[0,1,2,0]=(1.69005154613)/(o+1.)+(0.599453731488)
        ref[0,1,2,1]=(1.30699723672)/(o+1.)+(0.447731919557)
        ref[0,2,0,0]=(-0.274684837419)/(o+1.)+(-0.281541795502)
        ref[0,2,0,1]=(0.792946401732)/(o+1.)+(0.598271545783)
        ref[0,2,1,0]=(0.972930281169)/(o+1.)+(-0.692993050277)
        ref[0,2,1,1]=(-0.0948038476807)/(o+1.)+(-1.02610432679)
        ref[0,2,2,0]=(-0.417537958819)/(o+1.)+(0.0995721234709)
        ref[0,2,2,1]=(-0.343497078715)/(o+1.)+(0.287457935028)
        ref[0,3,0,0]=(1.06075362632)/(o+1.)+(0.392105127981)
        ref[0,3,0,1]=(-1.00743816734)/(o+1.)+(0.55530959011)
        ref[0,3,1,0]=(-1.35069907354)/(o+1.)+(-0.151109418893)
        ref[0,3,1,1]=(-0.209730918326)/(o+1.)+(-0.44139704026)
        ref[0,3,2,0]=(0.119571029049)/(o+1.)+(-0.788918379535)
        ref[0,3,2,1]=(0.834197985747)/(o+1.)+(0.0545741077131)
        ref[0,4,0,0]=(-0.713019032656)/(o+1.)+(0.0605014475136)
        ref[0,4,0,1]=(-0.188400834074)/(o+1.)+(0.0384669533136)
        ref[0,4,1,0]=(0.663494658006)/(o+1.)+(-0.197020376502)
        ref[0,4,1,1]=(1.57999775376)/(o+1.)+(-0.158042262352)
        ref[0,4,2,0]=(-0.825055197978)/(o+1.)+(0.190884563494)
        ref[0,4,2,1]=(-0.230079936711)/(o+1.)+(-0.619555317351)
        ref[1,0,0,0]=(0.0484302346511)/(o+1.)+(-0.277365092912)
        ref[1,0,0,1]=(-1.18699865107)/(o+1.)+(0.283637899436)
        ref[1,0,1,0]=(0.347990624042)/(o+1.)+(0.00803167578244)
        ref[1,0,1,1]=(-0.160064935171)/(o+1.)+(0.185408737905)
        ref[1,0,2,0]=(1.35245255927)/(o+1.)+(-0.531140014353)
        ref[1,0,2,1]=(1.13553833557)/(o+1.)+(0.152237391837)
        ref[1,1,0,0]=(0.99189846918)/(o+1.)+(0.338006421989)
        ref[1,1,0,1]=(-0.0255713914667)/(o+1.)+(0.12819321635)
        ref[1,1,1,0]=(-0.622986210996)/(o+1.)+(0.554888335978)
        ref[1,1,1,1]=(-0.0472909279168)/(o+1.)+(0.185961952963)
        ref[1,1,2,0]=(-0.729785677925)/(o+1.)+(-0.63936846664)
        ref[1,1,2,1]=(1.81104663684)/(o+1.)+(0.515008349399)
        ref[1,2,0,0]=(-1.54598989482)/(o+1.)+(0.284091818483)
        ref[1,2,0,1]=(0.182955266737)/(o+1.)+(0.00729582783438)
        ref[1,2,1,0]=(0.0678788403948)/(o+1.)+(0.0566178324469)
        ref[1,2,1,1]=(1.74208188933)/(o+1.)+(0.659881443367)
        ref[1,2,2,0]=(-0.0382440547455)/(o+1.)+(-0.535487134074)
        ref[1,2,2,1]=(0.198093668675)/(o+1.)+(-0.89001529806)
        ref[1,3,0,0]=(1.62490272607)/(o+1.)+(0.765524305033)
        ref[1,3,0,1]=(0.279965727369)/(o+1.)+(0.214570595298)
        ref[1,3,1,0]=(0.304113915286)/(o+1.)+(0.76954018039)
        ref[1,3,1,1]=(-0.885314465993)/(o+1.)+(0.62922698534)
        ref[1,3,2,0]=(0.434655075403)/(o+1.)+(-0.793446006776)
        ref[1,3,2,1]=(0.242224993913)/(o+1.)+(0.35107235084)
        ref[1,4,0,0]=(0.230763217193)/(o+1.)+(-0.155861223318)
        ref[1,4,0,1]=(0.808280805525)/(o+1.)+(0.0274401101125)
        ref[1,4,1,0]=(1.00266325917)/(o+1.)+(0.58007079489)
        ref[1,4,1,1]=(-0.641806023088)/(o+1.)+(-0.767874882808)
        ref[1,4,2,0]=(-1.27895710644)/(o+1.)+(0.573630110668)
        ref[1,4,2,1]=(-1.85031165866)/(o+1.)+(0.233982522704)
        ref[2,0,0,0]=(-1.36306650425)/(o+1.)+(-0.0533791412027)
        ref[2,0,0,1]=(1.10778233734)/(o+1.)+(-0.109894920649)
        ref[2,0,1,0]=(-0.607343404586)/(o+1.)+(-0.0381401236341)
        ref[2,0,1,1]=(-1.3016959741)/(o+1.)+(0.716358469177)
        ref[2,0,2,0]=(-1.65648421545)/(o+1.)+(0.675405459391)
        ref[2,0,2,1]=(1.43609344598)/(o+1.)+(-0.147175797578)
        ref[2,1,0,0]=(1.71599313418)/(o+1.)+(0.476324427863)
        ref[2,1,0,1]=(-0.197534803692)/(o+1.)+(0.74045848327)
        ref[2,1,1,0]=(1.15353139511)/(o+1.)+(-0.202917075571)
        ref[2,1,1,1]=(-0.646719728334)/(o+1.)+(0.089255774502)
        ref[2,1,2,0]=(-0.59247668041)/(o+1.)+(0.588338907289)
        ref[2,1,2,1]=(0.666534772105)/(o+1.)+(-0.171967067769)
        ref[2,2,0,0]=(0.10417442527)/(o+1.)+(-0.0373262159521)
        ref[2,2,0,1]=(1.6674625773)/(o+1.)+(0.33614460395)
        ref[2,2,1,0]=(1.40297752551)/(o+1.)+(-0.392281288816)
        ref[2,2,1,1]=(-0.72250539345)/(o+1.)+(-0.417148733502)
        ref[2,2,2,0]=(0.98912794704)/(o+1.)+(-1.19147084052)
        ref[2,2,2,1]=(-0.146690010568)/(o+1.)+(0.29184308053)
        ref[2,3,0,0]=(-0.431556699859)/(o+1.)+(-0.342220279916)
        ref[2,3,0,1]=(0.934033973509)/(o+1.)+(-1.26013765212)
        ref[2,3,1,0]=(0.387527880641)/(o+1.)+(0.223030814168)
        ref[2,3,1,1]=(2.43900464614)/(o+1.)+(-0.150886881122)
        ref[2,3,2,0]=(-1.75874429729)/(o+1.)+(0.116824608395)
        ref[2,3,2,1]=(-0.324304683574)/(o+1.)+(-0.817408425192)
        ref[2,4,0,0]=(-0.868761951233)/(o+1.)+(-0.0974552514252)
        ref[2,4,0,1]=(-0.416384616581)/(o+1.)+(0.282760450034)
        ref[2,4,1,0]=(-1.52621455953)/(o+1.)+(-0.448320971521)
        ref[2,4,1,1]=(-0.748074648234)/(o+1.)+(-0.346744035642)
        ref[2,4,2,0]=(0.50945468538)/(o+1.)+(0.0826481108988)
        ref[2,4,2,1]=(-0.659870196472)/(o+1.)+(-0.1495106279)
        ref[3,0,0,0]=(-1.5958920295)/(o+1.)+(-0.267320377314)
        ref[3,0,0,1]=(0.114794665201)/(o+1.)+(-0.181493709858)
        ref[3,0,1,0]=(0.600070206061)/(o+1.)+(-0.954433355289)
        ref[3,0,1,1]=(-1.07244059079)/(o+1.)+(1.19457963154)
        ref[3,0,2,0]=(0.301651870816)/(o+1.)+(0.499751268128)
        ref[3,0,2,1]=(0.865397463)/(o+1.)+(-0.0300788289559)
        ref[3,1,0,0]=(-0.832196173732)/(o+1.)+(-0.15494791425)
        ref[3,1,0,1]=(2.11534099144)/(o+1.)+(-0.130726809141)
        ref[3,1,1,0]=(0.00224203206025)/(o+1.)+(0.219270758379)
        ref[3,1,1,1]=(-0.277915215701)/(o+1.)+(0.431998989942)
        ref[3,1,2,0]=(-0.375139554013)/(o+1.)+(0.0160386777929)
        ref[3,1,2,1]=(0.658042649595)/(o+1.)+(-0.378681358928)
        ref[3,2,0,0]=(-0.482737482307)/(o+1.)+(0.450625784304)
        ref[3,2,0,1]=(-1.66696520506)/(o+1.)+(-0.588544849758)
        ref[3,2,1,0]=(1.59130029837)/(o+1.)+(0.497103472626)
        ref[3,2,1,1]=(0.393664568064)/(o+1.)+(-0.311916623046)
        ref[3,2,2,0]=(0.673365217238)/(o+1.)+(0.0477857019673)
        ref[3,2,2,1]=(-0.348041538126)/(o+1.)+(-0.490856911118)
        ref[3,3,0,0]=(-1.05359077988)/(o+1.)+(0.17767296453)
        ref[3,3,0,1]=(-1.02822286704)/(o+1.)+(-0.521240631647)
        ref[3,3,1,0]=(0.686396175815)/(o+1.)+(0.241205583529)
        ref[3,3,1,1]=(-0.63621617627)/(o+1.)+(1.15008135302)
        ref[3,3,2,0]=(0.242666882753)/(o+1.)+(-0.415262786705)
        ref[3,3,2,1]=(0.808145888629)/(o+1.)+(0.600175124848)
        ref[3,4,0,0]=(-1.38888754339)/(o+1.)+(0.429249531436)
        ref[3,4,0,1]=(1.70366041389)/(o+1.)+(-0.582120693104)
        ref[3,4,1,0]=(0.479439197375)/(o+1.)+(-0.708030105673)
        ref[3,4,1,1]=(1.5769515105)/(o+1.)+(0.129252407006)
        ref[3,4,2,0]=(1.24561953779)/(o+1.)+(0.382649063886)
        ref[3,4,2,1]=(-0.979696843387)/(o+1.)+(-0.242843692411)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_Solution_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.496833468021)*x[0]**o+(-0.119656931817)*x[0]+(-0.456435785007)*x[1]**o+(0.0378732106767)*x[1]
        ref=(-0.953269253027)/(o+1.)+(-0.04089186057)
      else:
        s=(0.960756800558)*x[0]**o+(-0.297563111204)*x[0]+(0.657965711967)*x[1]**o+(-0.942706539964)*x[1]+(-0.460398488427)*x[2]**o+(0.539495746387)*x[2]
        ref=(1.1583240241)/(o+1.)+(-0.350386952391)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_Solution_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.521294857582)*x[0]**o+(0.813966078452)*x[0]+(-0.460972018313)*x[1]**o+(-0.274643646978)*x[1]
        s[1]=(0.64160433254)*x[0]**o+(0.100182020329)*x[0]+(-0.877560875536)*x[1]**o+(0.0967945574365)*x[1]
        ref[0]=(-0.982266875895)/(o+1.)+(0.269661215737)
        ref[1]=(-0.235956542996)/(o+1.)+(0.0984882888828)
      else:
        s[0]=(-0.935408759921)*x[0]**o+(-0.636823488942)*x[0]+(-0.937973746046)*x[1]**o+(-0.548377694248)*x[1]+(-0.325664041787)*x[2]**o+(-0.859644239487)*x[2]
        s[1]=(-0.618325226925)*x[0]**o+(-0.464240689062)*x[0]+(-0.333395219266)*x[1]**o+(0.75384521826)*x[1]+(0.812529741058)*x[2]**o+(0.293845075804)*x[2]
        ref[0]=(-2.19904654775)/(o+1.)+(-1.02242271134)
        ref[1]=(-0.139190705133)/(o+1.)+(0.291724802501)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_Solution_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.745131837074)*x[0]**o+(0.954368069641)*x[0]+(0.582989014357)*x[1]**o+(0.649474790852)*x[1]
        s[0,1]=(0.0682420943207)*x[0]**o+(0.890255947523)*x[0]+(-0.501961353506)*x[1]**o+(-0.468049947986)*x[1]
        s[0,2]=(-0.376326981417)*x[0]**o+(0.318377164087)*x[0]+(0.518313700818)*x[1]**o+(-0.624749027625)*x[1]
        s[0,3]=(-0.0127578587596)*x[0]**o+(0.510088777647)*x[0]+(-0.00731027766723)*x[1]**o+(0.404116290004)*x[1]
        s[0,4]=(-0.566065672786)*x[0]**o+(-0.797903214842)*x[0]+(0.305896347185)*x[1]**o+(-0.44763489831)*x[1]
        s[1,0]=(0.136108309012)*x[0]**o+(0.8858491692)*x[0]+(0.63174549565)*x[1]**o+(0.12884968267)*x[1]
        s[1,1]=(-0.420219415055)*x[0]**o+(-0.474087579489)*x[0]+(-0.639616730493)*x[1]**o+(-0.339713160471)*x[1]
        s[1,2]=(0.730001695284)*x[0]**o+(-0.281914379046)*x[0]+(-0.0897917239674)*x[1]**o+(-0.521366256858)*x[1]
        s[1,3]=(0.507225459863)*x[0]**o+(-0.279054149386)*x[0]+(-0.310997653203)*x[1]**o+(0.798594910647)*x[1]
        s[1,4]=(-0.473780124237)*x[0]**o+(-0.521982314393)*x[0]+(-0.66517104956)*x[1]**o+(0.906212634536)*x[1]
        s[2,0]=(0.981595846105)*x[0]**o+(0.306232557856)*x[0]+(0.144004451791)*x[1]**o+(0.590440862044)*x[1]
        s[2,1]=(-0.465282108724)*x[0]**o+(-0.911349734596)*x[0]+(0.724954062739)*x[1]**o+(0.0617692321941)*x[1]
        s[2,2]=(-0.698461944366)*x[0]**o+(0.69038728172)*x[0]+(-0.0999661256163)*x[1]**o+(0.81993889447)*x[1]
        s[2,3]=(-0.850065489825)*x[0]**o+(-0.137124990846)*x[0]+(0.390609467506)*x[1]**o+(0.0928979112394)*x[1]
        s[2,4]=(-0.831110879728)*x[0]**o+(0.85125177454)*x[0]+(0.322669554869)*x[1]**o+(0.199419883126)*x[1]
        s[3,0]=(0.560516174698)*x[0]**o+(-0.952963851912)*x[0]+(0.0840374448937)*x[1]**o+(0.852644015588)*x[1]
        s[3,1]=(-0.738776742753)*x[0]**o+(-0.718650015551)*x[0]+(-0.48025826111)*x[1]**o+(0.163919964403)*x[1]
        s[3,2]=(-0.498985570012)*x[0]**o+(0.796130560575)*x[0]+(0.78812252561)*x[1]**o+(0.427402467946)*x[1]
        s[3,3]=(0.900695900592)*x[0]**o+(-0.68276543281)*x[0]+(-0.448699598266)*x[1]**o+(0.338876907438)*x[1]
        s[3,4]=(-0.101161698815)*x[0]**o+(0.24782725682)*x[0]+(0.194568299968)*x[1]**o+(-0.232062305313)*x[1]
        ref[0,0]=(1.32812085143)/(o+1.)+(0.801921430247)
        ref[0,1]=(-0.433719259185)/(o+1.)+(0.211102999768)
        ref[0,2]=(0.141986719401)/(o+1.)+(-0.153185931769)
        ref[0,3]=(-0.0200681364268)/(o+1.)+(0.457102533825)
        ref[0,4]=(-0.260169325601)/(o+1.)+(-0.622769056576)
        ref[1,0]=(0.767853804662)/(o+1.)+(0.507349425935)
        ref[1,1]=(-1.05983614555)/(o+1.)+(-0.40690036998)
        ref[1,2]=(0.640209971317)/(o+1.)+(-0.401640317952)
        ref[1,3]=(0.19622780666)/(o+1.)+(0.259770380631)
        ref[1,4]=(-1.1389511738)/(o+1.)+(0.192115160071)
        ref[2,0]=(1.1256002979)/(o+1.)+(0.44833670995)
        ref[2,1]=(0.259671954016)/(o+1.)+(-0.424790251201)
        ref[2,2]=(-0.798428069982)/(o+1.)+(0.755163088095)
        ref[2,3]=(-0.459456022319)/(o+1.)+(-0.0221135398034)
        ref[2,4]=(-0.508441324858)/(o+1.)+(0.525335828833)
        ref[3,0]=(0.644553619592)/(o+1.)+(-0.0501599181621)
        ref[3,1]=(-1.21903500386)/(o+1.)+(-0.277365025574)
        ref[3,2]=(0.289136955597)/(o+1.)+(0.611766514261)
        ref[3,3]=(0.451996302326)/(o+1.)+(-0.171944262686)
        ref[3,4]=(0.0934066011531)/(o+1.)+(0.00788247575383)
      else:
        s[0,0]=(0.581126436866)*x[0]**o+(0.307736169306)*x[0]+(0.46264603817)*x[1]**o+(-0.658226966972)*x[1]+(0.560314472945)*x[2]**o+(-0.71392138112)*x[2]
        s[0,1]=(0.41763244695)*x[0]**o+(-0.249710720358)*x[0]+(0.318827586572)*x[1]**o+(-0.766338584888)*x[1]+(-0.235847336194)*x[2]**o+(0.720822035745)*x[2]
        s[0,2]=(0.753177939244)*x[0]**o+(0.575601283201)*x[0]+(0.833261599709)*x[1]**o+(0.285673583448)*x[1]+(-0.850617074406)*x[2]**o+(0.976207021957)*x[2]
        s[0,3]=(-0.246323991583)*x[0]**o+(-0.425601468065)*x[0]+(-0.631382577618)*x[1]**o+(0.303177559003)*x[1]+(-0.772499313037)*x[2]**o+(0.0976456268892)*x[2]
        s[0,4]=(0.459782891083)*x[0]**o+(0.695607546341)*x[0]+(-0.758437555397)*x[1]**o+(-0.0467038938803)*x[1]+(0.077637691262)*x[2]**o+(0.632686471042)*x[2]
        s[1,0]=(0.62677312896)*x[0]**o+(-0.538520494446)*x[0]+(0.322526032671)*x[1]**o+(0.155570348616)*x[1]+(-0.645711565238)*x[2]**o+(-0.505639404827)*x[2]
        s[1,1]=(0.760035088253)*x[0]**o+(-0.653433840013)*x[0]+(0.541102747261)*x[1]**o+(0.102121826295)*x[1]+(0.203147157734)*x[2]**o+(0.825692237055)*x[2]
        s[1,2]=(0.83286976398)*x[0]**o+(-0.483950138218)*x[0]+(-0.988388711773)*x[1]**o+(0.276946264031)*x[1]+(-0.769100710972)*x[2]**o+(-0.857758652165)*x[2]
        s[1,3]=(0.965261287995)*x[0]**o+(-0.582299103263)*x[0]+(0.650118954781)*x[1]**o+(-0.714080635544)*x[1]+(-0.984672595257)*x[2]**o+(-0.890248882426)*x[2]
        s[1,4]=(0.780505914972)*x[0]**o+(-0.469283145875)*x[0]+(0.65642477278)*x[1]**o+(-0.0385309033271)*x[1]+(-0.838684124311)*x[2]**o+(0.715186480156)*x[2]
        s[2,0]=(-0.957976655406)*x[0]**o+(-0.326902969464)*x[0]+(0.519819227981)*x[1]**o+(0.526859170123)*x[1]+(-0.83643924332)*x[2]**o+(-0.833906618158)*x[2]
        s[2,1]=(0.511380988321)*x[0]**o+(0.205939303999)*x[0]+(-0.560011037377)*x[1]**o+(-0.74412420954)*x[1]+(-0.413858375939)*x[2]**o+(-0.818302325723)*x[2]
        s[2,2]=(0.669120009156)*x[0]**o+(-0.680617749061)*x[0]+(-0.855851359703)*x[1]**o+(0.681103383833)*x[1]+(0.358345655307)*x[2]**o+(-0.834176731331)*x[2]
        s[2,3]=(0.113991673968)*x[0]**o+(-0.167621358311)*x[0]+(-0.824506843066)*x[1]**o+(0.576032397062)*x[1]+(0.98813748839)*x[2]**o+(-0.954277123787)*x[2]
        s[2,4]=(0.758065768891)*x[0]**o+(0.550472010004)*x[0]+(0.541968389568)*x[1]**o+(-0.893196211859)*x[1]+(0.55846204484)*x[2]**o+(-0.154935285741)*x[2]
        s[3,0]=(0.319452905704)*x[0]**o+(0.416924332754)*x[0]+(0.695996329307)*x[1]**o+(-0.512723809281)*x[1]+(-0.0438582487583)*x[2]**o+(-0.202757150975)*x[2]
        s[3,1]=(0.300936121076)*x[0]**o+(0.0199168771219)*x[0]+(-0.79311689061)*x[1]**o+(0.120821678434)*x[1]+(0.832598541766)*x[2]**o+(-0.13378517127)*x[2]
        s[3,2]=(0.816251625164)*x[0]**o+(0.180324577125)*x[0]+(0.869902904875)*x[1]**o+(-0.300483761133)*x[1]+(0.405728532808)*x[2]**o+(-0.772682148538)*x[2]
        s[3,3]=(-0.0882536430174)*x[0]**o+(0.67244001887)*x[0]+(-0.295701327277)*x[1]**o+(-0.235065967505)*x[1]+(0.905094390215)*x[2]**o+(-0.132120939051)*x[2]
        s[3,4]=(0.976955651594)*x[0]**o+(0.36283338584)*x[0]+(0.21945555266)*x[1]**o+(-0.305668951381)*x[1]+(0.581731431193)*x[2]**o+(-0.308954336316)*x[2]
        ref[0,0]=(1.60408694798)/(o+1.)+(-0.532206089393)
        ref[0,1]=(0.500612697329)/(o+1.)+(-0.147613634751)
        ref[0,2]=(0.735822464547)/(o+1.)+(0.918740944303)
        ref[0,3]=(-1.65020588224)/(o+1.)+(-0.0123891410865)
        ref[0,4]=(-0.221016973052)/(o+1.)+(0.640795061752)
        ref[1,0]=(0.303587596392)/(o+1.)+(-0.444294775329)
        ref[1,1]=(1.50428499325)/(o+1.)+(0.137190111668)
        ref[1,2]=(-0.924619658766)/(o+1.)+(-0.532381263176)
        ref[1,3]=(0.630707647519)/(o+1.)+(-1.09331431062)
        ref[1,4]=(0.598246563441)/(o+1.)+(0.103686215477)
        ref[2,0]=(-1.27459667074)/(o+1.)+(-0.31697520875)
        ref[2,1]=(-0.462488424995)/(o+1.)+(-0.678243615632)
        ref[2,2]=(0.17161430476)/(o+1.)+(-0.416845548279)
        ref[2,3]=(0.277622319292)/(o+1.)+(-0.272933042518)
        ref[2,4]=(1.8584962033)/(o+1.)+(-0.248829743798)
        ref[3,0]=(0.971590986252)/(o+1.)+(-0.149278313751)
        ref[3,1]=(0.340417772233)/(o+1.)+(0.00347669214316)
        ref[3,2]=(2.09188306285)/(o+1.)+(-0.446420666273)
        ref[3,3]=(0.521139419921)/(o+1.)+(0.152626556157)
        ref[3,4]=(1.77814263545)/(o+1.)+(-0.125894950929)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_Solution_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.327469743415)*x[0]**o+(0.986515360087)*x[0]+(-0.264743586601)*x[1]**o+(-0.417911841646)*x[1]
        s[0,0,1]=(-0.367335531482)*x[0]**o+(-0.294564095391)*x[0]+(0.948849162205)*x[1]**o+(-0.672727814952)*x[1]
        s[0,1,0]=(0.201971500659)*x[0]**o+(0.715648302042)*x[0]+(0.200753457036)*x[1]**o+(-0.447431022755)*x[1]
        s[0,1,1]=(0.391503527712)*x[0]**o+(0.561563485529)*x[0]+(0.0756343714486)*x[1]**o+(-0.232567658257)*x[1]
        s[1,0,0]=(0.0416189159889)*x[0]**o+(0.933172580543)*x[0]+(0.906370262092)*x[1]**o+(0.843408035954)*x[1]
        s[1,0,1]=(-0.0489328569668)*x[0]**o+(0.305933018225)*x[0]+(0.459563500395)*x[1]**o+(-0.305696910459)*x[1]
        s[1,1,0]=(-0.166432892993)*x[0]**o+(0.323257103336)*x[0]+(-0.876673724336)*x[1]**o+(-0.246533332987)*x[1]
        s[1,1,1]=(0.331551737096)*x[0]**o+(-0.473411829001)*x[0]+(0.658917172856)*x[1]**o+(-0.645603256937)*x[1]
        s[2,0,0]=(-0.779837619461)*x[0]**o+(0.726847293406)*x[0]+(0.756283586425)*x[1]**o+(0.820695230411)*x[1]
        s[2,0,1]=(-0.0446700642358)*x[0]**o+(-0.00612226458731)*x[0]+(-0.255350731215)*x[1]**o+(0.279072151991)*x[1]
        s[2,1,0]=(0.952754783643)*x[0]**o+(0.575497372286)*x[0]+(-0.996409669059)*x[1]**o+(-0.993504128277)*x[1]
        s[2,1,1]=(-0.964146199975)*x[0]**o+(-0.89830816406)*x[0]+(0.27874102283)*x[1]**o+(-0.193698734442)*x[1]
        s[3,0,0]=(-0.22434829265)*x[0]**o+(-0.451123824419)*x[0]+(0.0996011155458)*x[1]**o+(-0.0196796989303)*x[1]
        s[3,0,1]=(0.479762785194)*x[0]**o+(-0.0875868570981)*x[0]+(-0.604020314973)*x[1]**o+(-0.948950136203)*x[1]
        s[3,1,0]=(0.380546378709)*x[0]**o+(0.968895520825)*x[0]+(0.0717757394195)*x[1]**o+(0.00273829931123)*x[1]
        s[3,1,1]=(0.999877803305)*x[0]**o+(0.726497905193)*x[0]+(-0.156097974361)*x[1]**o+(0.875069921439)*x[1]
        s[4,0,0]=(0.298060895282)*x[0]**o+(-0.25017046132)*x[0]+(-0.6409928584)*x[1]**o+(0.237325951363)*x[1]
        s[4,0,1]=(0.176241833972)*x[0]**o+(0.362302520221)*x[0]+(-0.632667110076)*x[1]**o+(0.206684577439)*x[1]
        s[4,1,0]=(-0.0658280231422)*x[0]**o+(-0.830173560366)*x[0]+(-0.408668580492)*x[1]**o+(0.131219154432)*x[1]
        s[4,1,1]=(-0.739357616315)*x[0]**o+(0.331825079796)*x[0]+(-0.950532071867)*x[1]**o+(0.430637319576)*x[1]
        s[5,0,0]=(-0.374056178585)*x[0]**o+(-0.503558441322)*x[0]+(0.253751100374)*x[1]**o+(-0.169874527745)*x[1]
        s[5,0,1]=(-0.540168118335)*x[0]**o+(-0.599437156114)*x[0]+(-0.545843659509)*x[1]**o+(-0.144502584314)*x[1]
        s[5,1,0]=(0.734871232877)*x[0]**o+(0.294937992469)*x[0]+(0.318246966782)*x[1]**o+(0.301215884904)*x[1]
        s[5,1,1]=(-0.48224128272)*x[0]**o+(-0.971913873914)*x[0]+(-0.853997429288)*x[1]**o+(0.911950472127)*x[1]
        ref[0,0,0]=(0.062726156814)/(o+1.)+(0.284301759221)
        ref[0,0,1]=(0.581513630723)/(o+1.)+(-0.483645955172)
        ref[0,1,0]=(0.402724957695)/(o+1.)+(0.134108639643)
        ref[0,1,1]=(0.467137899161)/(o+1.)+(0.164497913636)
        ref[1,0,0]=(0.947989178081)/(o+1.)+(0.888290308248)
        ref[1,0,1]=(0.410630643428)/(o+1.)+(0.000118053882864)
        ref[1,1,0]=(-1.04310661733)/(o+1.)+(0.0383618851745)
        ref[1,1,1]=(0.990468909952)/(o+1.)+(-0.559507542969)
        ref[2,0,0]=(-0.0235540330353)/(o+1.)+(0.773771261908)
        ref[2,0,1]=(-0.300020795451)/(o+1.)+(0.136474943702)
        ref[2,1,0]=(-0.043654885416)/(o+1.)+(-0.209003377996)
        ref[2,1,1]=(-0.685405177144)/(o+1.)+(-0.546003449251)
        ref[3,0,0]=(-0.124747177104)/(o+1.)+(-0.235401761675)
        ref[3,0,1]=(-0.124257529779)/(o+1.)+(-0.51826849665)
        ref[3,1,0]=(0.452322118128)/(o+1.)+(0.485816910068)
        ref[3,1,1]=(0.843779828944)/(o+1.)+(0.800783913316)
        ref[4,0,0]=(-0.342931963118)/(o+1.)+(-0.00642225497804)
        ref[4,0,1]=(-0.456425276104)/(o+1.)+(0.28449354883)
        ref[4,1,0]=(-0.474496603634)/(o+1.)+(-0.349477202967)
        ref[4,1,1]=(-1.68988968818)/(o+1.)+(0.381231199686)
        ref[5,0,0]=(-0.120305078211)/(o+1.)+(-0.336716484533)
        ref[5,0,1]=(-1.08601177784)/(o+1.)+(-0.371969870214)
        ref[5,1,0]=(1.05311819966)/(o+1.)+(0.298076938687)
        ref[5,1,1]=(-1.33623871201)/(o+1.)+(-0.0299817008935)
      else:
        s[0,0,0]=(0.907642728831)*x[0]**o+(0.24755436099)*x[0]+(0.329429496561)*x[1]**o+(0.195804615806)*x[1]+(0.635928214783)*x[2]**o+(0.286575554416)*x[2]
        s[0,0,1]=(-0.596463751156)*x[0]**o+(-0.573557158304)*x[0]+(-0.799372215818)*x[1]**o+(0.917917612224)*x[1]+(0.743915017372)*x[2]**o+(0.697340711065)*x[2]
        s[0,1,0]=(-0.226377288829)*x[0]**o+(0.11441928128)*x[0]+(0.915343934109)*x[1]**o+(0.0235994828916)*x[1]+(0.837206417747)*x[2]**o+(-0.809536949964)*x[2]
        s[0,1,1]=(-0.807835571238)*x[0]**o+(-0.313842967415)*x[0]+(0.143974822075)*x[1]**o+(0.772050096219)*x[1]+(0.0698457301843)*x[2]**o+(-0.878711129876)*x[2]
        s[1,0,0]=(0.806964016183)*x[0]**o+(0.906862034833)*x[0]+(0.656990650598)*x[1]**o+(0.858146180457)*x[1]+(-0.455748564953)*x[2]**o+(0.0415537957765)*x[2]
        s[1,0,1]=(0.930192572382)*x[0]**o+(0.406175274412)*x[0]+(0.522683472096)*x[1]**o+(-0.40154832923)*x[1]+(0.62436447915)*x[2]**o+(0.942684294891)*x[2]
        s[1,1,0]=(-0.190480904061)*x[0]**o+(-0.884913647118)*x[0]+(0.894504380487)*x[1]**o+(0.472867062235)*x[1]+(-0.226561577541)*x[2]**o+(0.0817340770676)*x[2]
        s[1,1,1]=(-0.785577817856)*x[0]**o+(0.286061214076)*x[0]+(-0.29241495776)*x[1]**o+(-0.885469334696)*x[1]+(-0.728917274605)*x[2]**o+(-0.400170372532)*x[2]
        s[2,0,0]=(-0.0443801095191)*x[0]**o+(-0.580757683753)*x[0]+(-0.680012816252)*x[1]**o+(-0.24174589493)*x[1]+(-0.32345463822)*x[2]**o+(0.488362285647)*x[2]
        s[2,0,1]=(0.427683981593)*x[0]**o+(0.709521979359)*x[0]+(-0.485171186139)*x[1]**o+(-0.466170281294)*x[1]+(0.668169758775)*x[2]**o+(-0.32376435425)*x[2]
        s[2,1,0]=(-0.322582923837)*x[0]**o+(-0.243805747954)*x[0]+(-0.282254505116)*x[1]**o+(0.276894252913)*x[1]+(0.425316275665)*x[2]**o+(-0.843958018195)*x[2]
        s[2,1,1]=(-0.172148420657)*x[0]**o+(0.288740472914)*x[0]+(0.3130139801)*x[1]**o+(0.0909556111434)*x[1]+(-0.134361659944)*x[2]**o+(-0.958259889515)*x[2]
        s[3,0,0]=(0.509175759598)*x[0]**o+(0.901681022023)*x[0]+(-0.04878880083)*x[1]**o+(-0.292935057546)*x[1]+(-0.73460973687)*x[2]**o+(0.143266558286)*x[2]
        s[3,0,1]=(-0.525256676545)*x[0]**o+(0.330501687555)*x[0]+(-0.533311903214)*x[1]**o+(-0.558985165298)*x[1]+(-0.726910027746)*x[2]**o+(-0.430022743518)*x[2]
        s[3,1,0]=(-0.801052113128)*x[0]**o+(0.383850418649)*x[0]+(-0.74692163952)*x[1]**o+(0.119907629434)*x[1]+(-0.688440032866)*x[2]**o+(0.258856702215)*x[2]
        s[3,1,1]=(0.662483893926)*x[0]**o+(0.321392974926)*x[0]+(0.797623010994)*x[1]**o+(-0.555546458227)*x[1]+(0.140050792242)*x[2]**o+(0.819042937008)*x[2]
        s[4,0,0]=(0.208697187872)*x[0]**o+(0.347715032509)*x[0]+(0.900642395285)*x[1]**o+(-0.121096976816)*x[1]+(0.473919668027)*x[2]**o+(0.693182424772)*x[2]
        s[4,0,1]=(0.185689163302)*x[0]**o+(0.923973904464)*x[0]+(-0.091048447659)*x[1]**o+(-0.331567067076)*x[1]+(0.0639420800968)*x[2]**o+(0.0645745629094)*x[2]
        s[4,1,0]=(-0.097237311089)*x[0]**o+(0.0234720818008)*x[0]+(-0.476652797084)*x[1]**o+(-0.258741500235)*x[1]+(-0.644183703276)*x[2]**o+(-0.953514564143)*x[2]
        s[4,1,1]=(-0.647701499983)*x[0]**o+(0.514337030449)*x[0]+(-0.779360494982)*x[1]**o+(-0.467345311588)*x[1]+(0.166149183082)*x[2]**o+(0.12480608055)*x[2]
        s[5,0,0]=(-0.349937644226)*x[0]**o+(0.673132604681)*x[0]+(0.793703484606)*x[1]**o+(0.958967045691)*x[1]+(-0.361047380062)*x[2]**o+(0.758206969142)*x[2]
        s[5,0,1]=(0.415890706186)*x[0]**o+(0.748811321123)*x[0]+(-0.629959299026)*x[1]**o+(0.706391001435)*x[1]+(-0.831407461643)*x[2]**o+(0.698632094335)*x[2]
        s[5,1,0]=(-0.914414551259)*x[0]**o+(0.85324200236)*x[0]+(0.273543170188)*x[1]**o+(0.311955664492)*x[1]+(0.726975838083)*x[2]**o+(0.671917428159)*x[2]
        s[5,1,1]=(0.450170734221)*x[0]**o+(-0.928357299044)*x[0]+(-0.645229625203)*x[1]**o+(0.756723626659)*x[1]+(-0.85050032636)*x[2]**o+(0.70593481655)*x[2]
        ref[0,0,0]=(1.87300044017)/(o+1.)+(0.364967265606)
        ref[0,0,1]=(-0.651920949601)/(o+1.)+(0.520850582493)
        ref[0,1,0]=(1.52617306303)/(o+1.)+(-0.335759092896)
        ref[0,1,1]=(-0.594015018979)/(o+1.)+(-0.210252000536)
        ref[1,0,0]=(1.00820610183)/(o+1.)+(0.903281005533)
        ref[1,0,1]=(2.07724052363)/(o+1.)+(0.473655620037)
        ref[1,1,0]=(0.477461898885)/(o+1.)+(-0.165156253908)
        ref[1,1,1]=(-1.80691005022)/(o+1.)+(-0.499789246576)
        ref[2,0,0]=(-1.04784756399)/(o+1.)+(-0.167070646517)
        ref[2,0,1]=(0.610682554229)/(o+1.)+(-0.0402063280924)
        ref[2,1,0]=(-0.179521153288)/(o+1.)+(-0.405434756618)
        ref[2,1,1]=(0.00650389949838)/(o+1.)+(-0.289281902729)
        ref[3,0,0]=(-0.274222778102)/(o+1.)+(0.376006261382)
        ref[3,0,1]=(-1.7854786075)/(o+1.)+(-0.329253110631)
        ref[3,1,0]=(-2.23641378551)/(o+1.)+(0.381307375149)
        ref[3,1,1]=(1.60015769716)/(o+1.)+(0.292444726853)
        ref[4,0,0]=(1.58325925118)/(o+1.)+(0.459900240233)
        ref[4,0,1]=(0.15858279574)/(o+1.)+(0.328490700148)
        ref[4,1,0]=(-1.21807381145)/(o+1.)+(-0.594391991289)
        ref[4,1,1]=(-1.26091281188)/(o+1.)+(0.085898899706)
        ref[5,0,0]=(0.0827184603185)/(o+1.)+(1.19515330976)
        ref[5,0,1]=(-1.04547605448)/(o+1.)+(1.07691720845)
        ref[5,1,0]=(0.086104457012)/(o+1.)+(0.918557547505)
        ref[5,1,1]=(-1.04555921734)/(o+1.)+(0.267150572083)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_Solution_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.501583472757)*x[0]**o+(0.174276674809)*x[0]+(0.474559132041)*x[1]**o+(0.944197178797)*x[1]
        s[0,0,0,1]=(-0.850492891287)*x[0]**o+(-0.778714553562)*x[0]+(-0.864025500909)*x[1]**o+(0.905856111362)*x[1]
        s[0,0,1,0]=(0.837983487193)*x[0]**o+(0.789080933347)*x[0]+(0.0618430819498)*x[1]**o+(0.791928371037)*x[1]
        s[0,0,1,1]=(-0.678469228871)*x[0]**o+(0.702133334239)*x[0]+(-0.937443836374)*x[1]**o+(-0.130947584181)*x[1]
        s[0,0,2,0]=(0.84013488851)*x[0]**o+(0.683811818455)*x[0]+(0.981450595677)*x[1]**o+(-0.813078675111)*x[1]
        s[0,0,2,1]=(0.00148903051979)*x[0]**o+(-0.740776021571)*x[0]+(0.917048043634)*x[1]**o+(0.362159065586)*x[1]
        s[0,1,0,0]=(0.285313019304)*x[0]**o+(0.0108652086531)*x[0]+(0.323356296135)*x[1]**o+(0.541189488955)*x[1]
        s[0,1,0,1]=(0.545540725184)*x[0]**o+(-0.803623471109)*x[0]+(0.861711269888)*x[1]**o+(0.614843604935)*x[1]
        s[0,1,1,0]=(0.44131453032)*x[0]**o+(-0.560146628455)*x[0]+(-0.390081444316)*x[1]**o+(-0.814215730743)*x[1]
        s[0,1,1,1]=(-0.120309068092)*x[0]**o+(-0.992223638563)*x[0]+(-0.757944249628)*x[1]**o+(-0.906900780015)*x[1]
        s[0,1,2,0]=(0.278503974882)*x[0]**o+(0.471665153392)*x[0]+(0.745287973293)*x[1]**o+(-0.572972893657)*x[1]
        s[0,1,2,1]=(0.840866880159)*x[0]**o+(-0.969714982065)*x[0]+(-0.181097771175)*x[1]**o+(-0.645812375407)*x[1]
        s[0,2,0,0]=(-0.791010044792)*x[0]**o+(-0.895489878363)*x[0]+(-0.32385199033)*x[1]**o+(0.892979821152)*x[1]
        s[0,2,0,1]=(0.966330767971)*x[0]**o+(0.757433262697)*x[0]+(0.859264987879)*x[1]**o+(0.836273715628)*x[1]
        s[0,2,1,0]=(0.764347587379)*x[0]**o+(0.498460349969)*x[0]+(0.896024738679)*x[1]**o+(0.341308865054)*x[1]
        s[0,2,1,1]=(-0.276964987386)*x[0]**o+(-0.294300496732)*x[0]+(-0.755886597374)*x[1]**o+(0.0964446535133)*x[1]
        s[0,2,2,0]=(0.933876243978)*x[0]**o+(0.694067023434)*x[0]+(-0.626635561624)*x[1]**o+(0.107630740485)*x[1]
        s[0,2,2,1]=(0.905601520891)*x[0]**o+(0.246861090683)*x[0]+(0.364620749293)*x[1]**o+(-0.34329084894)*x[1]
        s[0,3,0,0]=(-0.484050880674)*x[0]**o+(-0.344211446029)*x[0]+(-0.304888212647)*x[1]**o+(-0.918515747538)*x[1]
        s[0,3,0,1]=(0.352307753352)*x[0]**o+(-0.497295643302)*x[0]+(-0.0859041918813)*x[1]**o+(-0.566829131134)*x[1]
        s[0,3,1,0]=(0.458949103752)*x[0]**o+(-0.940473606591)*x[0]+(0.500249302466)*x[1]**o+(-0.192176323799)*x[1]
        s[0,3,1,1]=(0.714312790267)*x[0]**o+(0.219740572431)*x[0]+(-0.621517496057)*x[1]**o+(-0.79963217554)*x[1]
        s[0,3,2,0]=(-0.041379012308)*x[0]**o+(-0.674987043823)*x[0]+(-0.0165846458396)*x[1]**o+(0.643074958959)*x[1]
        s[0,3,2,1]=(-0.258839387368)*x[0]**o+(0.63938656748)*x[0]+(0.0042420382287)*x[1]**o+(-0.711660235823)*x[1]
        s[0,4,0,0]=(-0.905262277861)*x[0]**o+(0.995066153797)*x[0]+(0.895316712432)*x[1]**o+(0.794484306559)*x[1]
        s[0,4,0,1]=(0.858709170162)*x[0]**o+(-0.896059882819)*x[0]+(-0.322447413352)*x[1]**o+(0.356724683447)*x[1]
        s[0,4,1,0]=(-0.496930600469)*x[0]**o+(-0.715930772072)*x[0]+(0.514137034309)*x[1]**o+(0.738095094325)*x[1]
        s[0,4,1,1]=(-0.475961686127)*x[0]**o+(0.406077221295)*x[0]+(-0.934472851356)*x[1]**o+(0.184564622216)*x[1]
        s[0,4,2,0]=(0.50929205392)*x[0]**o+(-0.361652258356)*x[0]+(-0.828564581552)*x[1]**o+(0.864597967503)*x[1]
        s[0,4,2,1]=(-0.670572914815)*x[0]**o+(-0.917836333665)*x[0]+(0.329473913157)*x[1]**o+(0.649536686851)*x[1]
        s[1,0,0,0]=(0.507327212873)*x[0]**o+(-0.784119587219)*x[0]+(0.861200611677)*x[1]**o+(-0.537142113189)*x[1]
        s[1,0,0,1]=(0.307279452436)*x[0]**o+(-0.331535898217)*x[0]+(0.864752548746)*x[1]**o+(-0.41254012361)*x[1]
        s[1,0,1,0]=(0.723666220176)*x[0]**o+(-0.528755526062)*x[0]+(0.182515067109)*x[1]**o+(-0.33319767285)*x[1]
        s[1,0,1,1]=(-0.652532069809)*x[0]**o+(-0.497647481159)*x[0]+(-0.726676961172)*x[1]**o+(0.901803102587)*x[1]
        s[1,0,2,0]=(0.966840239828)*x[0]**o+(-0.559625286504)*x[0]+(0.856786274248)*x[1]**o+(-0.231280485522)*x[1]
        s[1,0,2,1]=(0.898750450503)*x[0]**o+(-0.241387492118)*x[0]+(-0.242302411171)*x[1]**o+(-0.931590118111)*x[1]
        s[1,1,0,0]=(0.298117967411)*x[0]**o+(-0.887837423565)*x[0]+(-0.804393274952)*x[1]**o+(-0.339940646012)*x[1]
        s[1,1,0,1]=(-0.3303942703)*x[0]**o+(0.22588379719)*x[0]+(0.427447092807)*x[1]**o+(-0.940368147505)*x[1]
        s[1,1,1,0]=(-0.227054817206)*x[0]**o+(0.76367061207)*x[0]+(-0.382609156261)*x[1]**o+(-0.869111355614)*x[1]
        s[1,1,1,1]=(0.538281073824)*x[0]**o+(-0.454205347349)*x[0]+(-0.121007985262)*x[1]**o+(0.0290163849694)*x[1]
        s[1,1,2,0]=(-0.888958335111)*x[0]**o+(0.779682676236)*x[0]+(-0.875785352701)*x[1]**o+(-0.891960671849)*x[1]
        s[1,1,2,1]=(-0.173332519418)*x[0]**o+(-0.104421077299)*x[0]+(0.906729441747)*x[1]**o+(-0.616671337017)*x[1]
        s[1,2,0,0]=(-0.305885548664)*x[0]**o+(-0.86724411868)*x[0]+(-0.787155086634)*x[1]**o+(-0.898599442906)*x[1]
        s[1,2,0,1]=(0.848368426968)*x[0]**o+(0.0802812152769)*x[0]+(-0.378057863356)*x[1]**o+(-0.428488725116)*x[1]
        s[1,2,1,0]=(-0.290675868214)*x[0]**o+(0.342045945019)*x[0]+(-0.421298816185)*x[1]**o+(-0.553807359319)*x[1]
        s[1,2,1,1]=(0.0289012545961)*x[0]**o+(0.423276922447)*x[0]+(-0.911749389958)*x[1]**o+(-0.39409305776)*x[1]
        s[1,2,2,0]=(0.904480195539)*x[0]**o+(-0.729896491926)*x[0]+(-0.21305322735)*x[1]**o+(-0.347444381896)*x[1]
        s[1,2,2,1]=(0.34039428449)*x[0]**o+(-0.246758392869)*x[0]+(-0.159391459268)*x[1]**o+(-0.630134063159)*x[1]
        s[1,3,0,0]=(0.202017077356)*x[0]**o+(-0.195969669611)*x[0]+(0.267580338059)*x[1]**o+(-0.274282886887)*x[1]
        s[1,3,0,1]=(-0.183473257308)*x[0]**o+(-0.646999501714)*x[0]+(-0.746037354076)*x[1]**o+(-0.180977233442)*x[1]
        s[1,3,1,0]=(0.895075853694)*x[0]**o+(-0.820090467371)*x[0]+(-0.040561177436)*x[1]**o+(0.749395910572)*x[1]
        s[1,3,1,1]=(-0.232039710082)*x[0]**o+(-0.633561589743)*x[0]+(0.862514378491)*x[1]**o+(0.782438844898)*x[1]
        s[1,3,2,0]=(-0.0228789327036)*x[0]**o+(0.604368007016)*x[0]+(0.413803296189)*x[1]**o+(0.651011054925)*x[1]
        s[1,3,2,1]=(-0.682393812631)*x[0]**o+(-0.508380759547)*x[0]+(-0.958515947271)*x[1]**o+(-0.396382186265)*x[1]
        s[1,4,0,0]=(0.772945258804)*x[0]**o+(0.981773233361)*x[0]+(-0.415876208997)*x[1]**o+(0.391268836611)*x[1]
        s[1,4,0,1]=(-0.111414446087)*x[0]**o+(-0.381712931273)*x[0]+(0.389894843435)*x[1]**o+(-0.276144369314)*x[1]
        s[1,4,1,0]=(0.544309017751)*x[0]**o+(0.920841251767)*x[0]+(-0.524322161349)*x[1]**o+(-0.745967887424)*x[1]
        s[1,4,1,1]=(0.0993088568573)*x[0]**o+(0.172455640245)*x[0]+(-0.310227024922)*x[1]**o+(0.503773577532)*x[1]
        s[1,4,2,0]=(-0.282979849337)*x[0]**o+(0.573540669768)*x[0]+(0.974343187515)*x[1]**o+(0.643215559873)*x[1]
        s[1,4,2,1]=(0.616949303196)*x[0]**o+(-0.206142689323)*x[0]+(0.594875479161)*x[1]**o+(0.390460380914)*x[1]
        s[2,0,0,0]=(-0.164379488326)*x[0]**o+(-0.613380471473)*x[0]+(-0.309735177264)*x[1]**o+(0.722815882927)*x[1]
        s[2,0,0,1]=(-0.443733554915)*x[0]**o+(-0.226908044072)*x[0]+(0.22281947804)*x[1]**o+(0.321557448698)*x[1]
        s[2,0,1,0]=(0.979229428359)*x[0]**o+(-0.410075647119)*x[0]+(-0.0869658073133)*x[1]**o+(0.472354461211)*x[1]
        s[2,0,1,1]=(0.279554115394)*x[0]**o+(0.982131537982)*x[0]+(0.865627129266)*x[1]**o+(0.0595046994689)*x[1]
        s[2,0,2,0]=(-0.928629057762)*x[0]**o+(-0.0888000293033)*x[0]+(0.925425619622)*x[1]**o+(0.286877459131)*x[1]
        s[2,0,2,1]=(-0.128076644671)*x[0]**o+(0.769985562586)*x[0]+(-0.326281257692)*x[1]**o+(-0.904253231113)*x[1]
        s[2,1,0,0]=(-7.49817068653e-05)*x[0]**o+(0.944960174862)*x[0]+(0.394130491438)*x[1]**o+(0.904576980195)*x[1]
        s[2,1,0,1]=(0.51260445176)*x[0]**o+(0.871495843591)*x[0]+(-0.0117378185696)*x[1]**o+(-0.447028599408)*x[1]
        s[2,1,1,0]=(0.385302842134)*x[0]**o+(0.68530194308)*x[0]+(0.908717705423)*x[1]**o+(0.0657269588494)*x[1]
        s[2,1,1,1]=(-0.175101480962)*x[0]**o+(-0.0126863839895)*x[0]+(0.544228922923)*x[1]**o+(-0.856957800534)*x[1]
        s[2,1,2,0]=(-0.74757544736)*x[0]**o+(-0.705933683532)*x[0]+(0.496113134499)*x[1]**o+(0.134139074285)*x[1]
        s[2,1,2,1]=(0.661170765505)*x[0]**o+(-0.0954233403413)*x[0]+(0.796712312861)*x[1]**o+(0.941668952301)*x[1]
        s[2,2,0,0]=(0.396185569444)*x[0]**o+(0.352712187886)*x[0]+(0.132847269784)*x[1]**o+(0.00561831278948)*x[1]
        s[2,2,0,1]=(-0.284087190739)*x[0]**o+(-0.307024290069)*x[0]+(-0.958481172381)*x[1]**o+(0.0918318779969)*x[1]
        s[2,2,1,0]=(0.0633474988001)*x[0]**o+(-0.650310080822)*x[0]+(0.576579254607)*x[1]**o+(0.674521364804)*x[1]
        s[2,2,1,1]=(0.579536883642)*x[0]**o+(0.160337559302)*x[0]+(-0.806829552191)*x[1]**o+(-0.45313287787)*x[1]
        s[2,2,2,0]=(-0.922597521221)*x[0]**o+(-0.250217857798)*x[0]+(-0.0762271301664)*x[1]**o+(0.610046304117)*x[1]
        s[2,2,2,1]=(0.259914145502)*x[0]**o+(0.00944250446491)*x[0]+(-0.217644702814)*x[1]**o+(-0.396061779281)*x[1]
        s[2,3,0,0]=(-0.639026927928)*x[0]**o+(-0.0136042147736)*x[0]+(0.147051301049)*x[1]**o+(-0.116721318722)*x[1]
        s[2,3,0,1]=(-0.921716795774)*x[0]**o+(-0.573953102898)*x[0]+(-0.292599705087)*x[1]**o+(0.425695557831)*x[1]
        s[2,3,1,0]=(-0.137488915425)*x[0]**o+(-0.49947594329)*x[0]+(-0.0290388465996)*x[1]**o+(-0.246424924833)*x[1]
        s[2,3,1,1]=(0.231049448089)*x[0]**o+(-0.268957451079)*x[0]+(0.900504385567)*x[1]**o+(-0.331066572068)*x[1]
        s[2,3,2,0]=(-0.442867405047)*x[0]**o+(-0.862895940039)*x[0]+(0.130486555031)*x[1]**o+(-0.325005280104)*x[1]
        s[2,3,2,1]=(0.0505322685584)*x[0]**o+(0.588209043808)*x[0]+(-0.85729410482)*x[1]**o+(0.20026415404)*x[1]
        s[2,4,0,0]=(0.111265750114)*x[0]**o+(-0.368157597131)*x[0]+(0.264359624294)*x[1]**o+(0.880890103316)*x[1]
        s[2,4,0,1]=(0.7435570617)*x[0]**o+(-0.794211194611)*x[0]+(0.506134754195)*x[1]**o+(0.363318494554)*x[1]
        s[2,4,1,0]=(0.557218657611)*x[0]**o+(-0.584592285851)*x[0]+(-0.551098106406)*x[1]**o+(-0.253662162858)*x[1]
        s[2,4,1,1]=(0.508828804121)*x[0]**o+(0.4399952689)*x[0]+(-0.997629557499)*x[1]**o+(0.651550324191)*x[1]
        s[2,4,2,0]=(-0.292367705037)*x[0]**o+(0.325506979353)*x[0]+(0.345654565875)*x[1]**o+(-0.0615253936178)*x[1]
        s[2,4,2,1]=(-0.408469289071)*x[0]**o+(-0.242826710026)*x[0]+(0.968562541409)*x[1]**o+(-0.691883337576)*x[1]
        s[3,0,0,0]=(-0.166166381364)*x[0]**o+(0.67558604652)*x[0]+(0.681136750192)*x[1]**o+(-0.59276757042)*x[1]
        s[3,0,0,1]=(0.139326266787)*x[0]**o+(-0.855321186969)*x[0]+(0.87031372876)*x[1]**o+(0.167702944821)*x[1]
        s[3,0,1,0]=(0.0382861750276)*x[0]**o+(-0.15200218165)*x[0]+(0.877979117505)*x[1]**o+(-0.0121347649803)*x[1]
        s[3,0,1,1]=(0.498616833176)*x[0]**o+(-0.663932730791)*x[0]+(-0.351696535062)*x[1]**o+(0.635672218271)*x[1]
        s[3,0,2,0]=(0.570569612815)*x[0]**o+(0.302777992546)*x[0]+(0.234111661233)*x[1]**o+(0.23736500566)*x[1]
        s[3,0,2,1]=(0.103264640389)*x[0]**o+(-0.665245672415)*x[0]+(-0.507585301735)*x[1]**o+(-0.624579920066)*x[1]
        s[3,1,0,0]=(-0.424026522629)*x[0]**o+(-0.289542833446)*x[0]+(-0.208440343753)*x[1]**o+(0.670301272477)*x[1]
        s[3,1,0,1]=(0.884688677823)*x[0]**o+(-0.547024055336)*x[0]+(-0.760788413661)*x[1]**o+(0.892759772046)*x[1]
        s[3,1,1,0]=(-0.193616699826)*x[0]**o+(-0.817184462119)*x[0]+(-0.405430250489)*x[1]**o+(0.697242407135)*x[1]
        s[3,1,1,1]=(-0.224055211394)*x[0]**o+(0.1446788843)*x[0]+(0.242791365909)*x[1]**o+(-0.00969799364721)*x[1]
        s[3,1,2,0]=(-0.86134785548)*x[0]**o+(0.808223751392)*x[0]+(-0.59203015099)*x[1]**o+(-0.774006816622)*x[1]
        s[3,1,2,1]=(0.60450330353)*x[0]**o+(0.47388266469)*x[0]+(0.160423399019)*x[1]**o+(0.177522838509)*x[1]
        s[3,2,0,0]=(-0.0704547054622)*x[0]**o+(-0.28005227128)*x[0]+(0.842667852448)*x[1]**o+(-0.660858857912)*x[1]
        s[3,2,0,1]=(0.950935693626)*x[0]**o+(-0.707135492206)*x[0]+(0.383664604588)*x[1]**o+(0.301522608591)*x[1]
        s[3,2,1,0]=(-0.843657734623)*x[0]**o+(0.0313079420866)*x[0]+(0.842733149317)*x[1]**o+(0.8363224382)*x[1]
        s[3,2,1,1]=(0.590344608942)*x[0]**o+(-0.688010374961)*x[0]+(0.419847151895)*x[1]**o+(0.703183045156)*x[1]
        s[3,2,2,0]=(0.726161847781)*x[0]**o+(-0.644844589861)*x[0]+(0.615915050884)*x[1]**o+(0.928915762636)*x[1]
        s[3,2,2,1]=(-0.272507884368)*x[0]**o+(0.915695022479)*x[0]+(-0.461159868664)*x[1]**o+(-0.959034792496)*x[1]
        s[3,3,0,0]=(0.454712361388)*x[0]**o+(-0.212357162336)*x[0]+(0.153326792468)*x[1]**o+(-0.891670991055)*x[1]
        s[3,3,0,1]=(0.524401133703)*x[0]**o+(-0.649106057963)*x[0]+(0.879052771991)*x[1]**o+(-0.710445481523)*x[1]
        s[3,3,1,0]=(0.614330071709)*x[0]**o+(-0.632228392701)*x[0]+(0.272026382411)*x[1]**o+(-0.261384418355)*x[1]
        s[3,3,1,1]=(0.552126023508)*x[0]**o+(0.266617783195)*x[0]+(-0.183251756988)*x[1]**o+(0.905027547625)*x[1]
        s[3,3,2,0]=(-0.671825095536)*x[0]**o+(-0.0942290736973)*x[0]+(0.354473207574)*x[1]**o+(-0.61776928413)*x[1]
        s[3,3,2,1]=(0.8582228544)*x[0]**o+(-0.68683359365)*x[0]+(0.856354836335)*x[1]**o+(0.588764055551)*x[1]
        s[3,4,0,0]=(-0.535790362987)*x[0]**o+(0.835465266424)*x[0]+(-0.422274723942)*x[1]**o+(-0.979702488165)*x[1]
        s[3,4,0,1]=(-0.846846124079)*x[0]**o+(-0.52263128994)*x[0]+(0.0349528062482)*x[1]**o+(-0.498515393094)*x[1]
        s[3,4,1,0]=(-0.927241665975)*x[0]**o+(0.0353321330647)*x[0]+(-0.225562802107)*x[1]**o+(0.8207319568)*x[1]
        s[3,4,1,1]=(-0.174397417012)*x[0]**o+(-0.0398036522533)*x[0]+(0.0755396127937)*x[1]**o+(-0.102836810613)*x[1]
        s[3,4,2,0]=(-0.968779473612)*x[0]**o+(0.339954457114)*x[0]+(-0.613643260944)*x[1]**o+(0.301848346204)*x[1]
        s[3,4,2,1]=(0.994429790023)*x[0]**o+(0.159760881455)*x[0]+(0.252053920576)*x[1]**o+(-0.562820021013)*x[1]
        ref[0,0,0,0]=(0.976142604799)/(o+1.)+(0.559236926803)
        ref[0,0,0,1]=(-1.7145183922)/(o+1.)+(0.0635707789)
        ref[0,0,1,0]=(0.899826569143)/(o+1.)+(0.790504652192)
        ref[0,0,1,1]=(-1.61591306524)/(o+1.)+(0.285592875029)
        ref[0,0,2,0]=(1.82158548419)/(o+1.)+(-0.0646334283283)
        ref[0,0,2,1]=(0.918537074153)/(o+1.)+(-0.189308477993)
        ref[0,1,0,0]=(0.608669315438)/(o+1.)+(0.276027348804)
        ref[0,1,0,1]=(1.40725199507)/(o+1.)+(-0.0943899330873)
        ref[0,1,1,0]=(0.0512330860046)/(o+1.)+(-0.687181179599)
        ref[0,1,1,1]=(-0.878253317721)/(o+1.)+(-0.949562209289)
        ref[0,1,2,0]=(1.02379194818)/(o+1.)+(-0.0506538701325)
        ref[0,1,2,1]=(0.659769108985)/(o+1.)+(-0.807763678736)
        ref[0,2,0,0]=(-1.11486203512)/(o+1.)+(-0.00125502860562)
        ref[0,2,0,1]=(1.82559575585)/(o+1.)+(0.796853489163)
        ref[0,2,1,0]=(1.66037232606)/(o+1.)+(0.419884607511)
        ref[0,2,1,1]=(-1.03285158476)/(o+1.)+(-0.0989279216092)
        ref[0,2,2,0]=(0.307240682354)/(o+1.)+(0.400848881959)
        ref[0,2,2,1]=(1.27022227018)/(o+1.)+(-0.0482148791284)
        ref[0,3,0,0]=(-0.788939093322)/(o+1.)+(-0.631363596784)
        ref[0,3,0,1]=(0.26640356147)/(o+1.)+(-0.532062387218)
        ref[0,3,1,0]=(0.959198406218)/(o+1.)+(-0.566324965195)
        ref[0,3,1,1]=(0.0927952942109)/(o+1.)+(-0.289945801554)
        ref[0,3,2,0]=(-0.0579636581476)/(o+1.)+(-0.0159560424323)
        ref[0,3,2,1]=(-0.254597349139)/(o+1.)+(-0.0361368341719)
        ref[0,4,0,0]=(-0.00994556542921)/(o+1.)+(0.894775230178)
        ref[0,4,0,1]=(0.53626175681)/(o+1.)+(-0.269667599686)
        ref[0,4,1,0]=(0.0172064338404)/(o+1.)+(0.0110821611268)
        ref[0,4,1,1]=(-1.41043453748)/(o+1.)+(0.295320921755)
        ref[0,4,2,0]=(-0.319272527632)/(o+1.)+(0.251472854574)
        ref[0,4,2,1]=(-0.341099001658)/(o+1.)+(-0.134149823407)
        ref[1,0,0,0]=(1.36852782455)/(o+1.)+(-0.660630850204)
        ref[1,0,0,1]=(1.17203200118)/(o+1.)+(-0.372038010914)
        ref[1,0,1,0]=(0.906181287285)/(o+1.)+(-0.430976599456)
        ref[1,0,1,1]=(-1.37920903098)/(o+1.)+(0.202077810714)
        ref[1,0,2,0]=(1.82362651408)/(o+1.)+(-0.395452886013)
        ref[1,0,2,1]=(0.656448039332)/(o+1.)+(-0.586488805114)
        ref[1,1,0,0]=(-0.506275307541)/(o+1.)+(-0.613889034788)
        ref[1,1,0,1]=(0.0970528225071)/(o+1.)+(-0.357242175157)
        ref[1,1,1,0]=(-0.609663973467)/(o+1.)+(-0.0527203717721)
        ref[1,1,1,1]=(0.417273088561)/(o+1.)+(-0.21259448119)
        ref[1,1,2,0]=(-1.76474368781)/(o+1.)+(-0.0561389978065)
        ref[1,1,2,1]=(0.733396922329)/(o+1.)+(-0.360546207158)
        ref[1,2,0,0]=(-1.0930406353)/(o+1.)+(-0.882921780793)
        ref[1,2,0,1]=(0.470310563612)/(o+1.)+(-0.17410375492)
        ref[1,2,1,0]=(-0.711974684398)/(o+1.)+(-0.10588070715)
        ref[1,2,1,1]=(-0.882848135361)/(o+1.)+(0.0145919323434)
        ref[1,2,2,0]=(0.691426968189)/(o+1.)+(-0.538670436911)
        ref[1,2,2,1]=(0.181002825221)/(o+1.)+(-0.438446228014)
        ref[1,3,0,0]=(0.469597415414)/(o+1.)+(-0.235126278249)
        ref[1,3,0,1]=(-0.929510611384)/(o+1.)+(-0.413988367578)
        ref[1,3,1,0]=(0.854514676258)/(o+1.)+(-0.0353472783996)
        ref[1,3,1,1]=(0.630474668409)/(o+1.)+(0.0744386275773)
        ref[1,3,2,0]=(0.390924363485)/(o+1.)+(0.627689530971)
        ref[1,3,2,1]=(-1.6409097599)/(o+1.)+(-0.452381472906)
        ref[1,4,0,0]=(0.357069049807)/(o+1.)+(0.686521034986)
        ref[1,4,0,1]=(0.278480397348)/(o+1.)+(-0.328928650294)
        ref[1,4,1,0]=(0.0199868564018)/(o+1.)+(0.0874366821718)
        ref[1,4,1,1]=(-0.210918168065)/(o+1.)+(0.338114608888)
        ref[1,4,2,0]=(0.691363338179)/(o+1.)+(0.608378114821)
        ref[1,4,2,1]=(1.21182478236)/(o+1.)+(0.0921588457958)
        ref[2,0,0,0]=(-0.47411466559)/(o+1.)+(0.0547177057268)
        ref[2,0,0,1]=(-0.220914076875)/(o+1.)+(0.0473247023129)
        ref[2,0,1,0]=(0.892263621045)/(o+1.)+(0.0311394070465)
        ref[2,0,1,1]=(1.14518124466)/(o+1.)+(0.520818118726)
        ref[2,0,2,0]=(-0.00320343813922)/(o+1.)+(0.0990387149137)
        ref[2,0,2,1]=(-0.454357902363)/(o+1.)+(-0.0671338342638)
        ref[2,1,0,0]=(0.394055509731)/(o+1.)+(0.924768577529)
        ref[2,1,0,1]=(0.500866633191)/(o+1.)+(0.212233622092)
        ref[2,1,1,0]=(1.29402054756)/(o+1.)+(0.375514450965)
        ref[2,1,1,1]=(0.369127441961)/(o+1.)+(-0.434822092262)
        ref[2,1,2,0]=(-0.251462312861)/(o+1.)+(-0.285897304624)
        ref[2,1,2,1]=(1.45788307837)/(o+1.)+(0.42312280598)
        ref[2,2,0,0]=(0.529032839228)/(o+1.)+(0.179165250338)
        ref[2,2,0,1]=(-1.24256836312)/(o+1.)+(-0.107596206036)
        ref[2,2,1,0]=(0.639926753407)/(o+1.)+(0.0121056419913)
        ref[2,2,1,1]=(-0.227292668549)/(o+1.)+(-0.146397659284)
        ref[2,2,2,0]=(-0.998824651388)/(o+1.)+(0.179914223159)
        ref[2,2,2,1]=(0.0422694426877)/(o+1.)+(-0.193309637408)
        ref[2,3,0,0]=(-0.491975626879)/(o+1.)+(-0.0651627667478)
        ref[2,3,0,1]=(-1.21431650086)/(o+1.)+(-0.0741287725332)
        ref[2,3,1,0]=(-0.166527762025)/(o+1.)+(-0.372950434061)
        ref[2,3,1,1]=(1.13155383366)/(o+1.)+(-0.300012011573)
        ref[2,3,2,0]=(-0.312380850015)/(o+1.)+(-0.593950610072)
        ref[2,3,2,1]=(-0.806761836262)/(o+1.)+(0.394236598924)
        ref[2,4,0,0]=(0.375625374408)/(o+1.)+(0.256366253093)
        ref[2,4,0,1]=(1.2496918159)/(o+1.)+(-0.215446350028)
        ref[2,4,1,0]=(0.00612055120519)/(o+1.)+(-0.419127224354)
        ref[2,4,1,1]=(-0.488800753378)/(o+1.)+(0.545772796545)
        ref[2,4,2,0]=(0.0532868608376)/(o+1.)+(0.131990792868)
        ref[2,4,2,1]=(0.560093252338)/(o+1.)+(-0.467355023801)
        ref[3,0,0,0]=(0.514970368828)/(o+1.)+(0.0414092380501)
        ref[3,0,0,1]=(1.00963999555)/(o+1.)+(-0.343809121074)
        ref[3,0,1,0]=(0.916265292533)/(o+1.)+(-0.0820684733153)
        ref[3,0,1,1]=(0.146920298114)/(o+1.)+(-0.0141302562603)
        ref[3,0,2,0]=(0.804681274048)/(o+1.)+(0.270071499103)
        ref[3,0,2,1]=(-0.404320661346)/(o+1.)+(-0.644912796241)
        ref[3,1,0,0]=(-0.632466866383)/(o+1.)+(0.190379219516)
        ref[3,1,0,1]=(0.123900264162)/(o+1.)+(0.172867858355)
        ref[3,1,1,0]=(-0.599046950315)/(o+1.)+(-0.0599710274916)
        ref[3,1,1,1]=(0.0187361545152)/(o+1.)+(0.0674904453263)
        ref[3,1,2,0]=(-1.45337800647)/(o+1.)+(0.0171084673849)
        ref[3,1,2,1]=(0.764926702549)/(o+1.)+(0.325702751599)
        ref[3,2,0,0]=(0.772213146985)/(o+1.)+(-0.470455564596)
        ref[3,2,0,1]=(1.33460029821)/(o+1.)+(-0.202806441808)
        ref[3,2,1,0]=(-0.00092458530583)/(o+1.)+(0.433815190143)
        ref[3,2,1,1]=(1.01019176084)/(o+1.)+(0.00758633509776)
        ref[3,2,2,0]=(1.34207689867)/(o+1.)+(0.142035586387)
        ref[3,2,2,1]=(-0.733667753032)/(o+1.)+(-0.0216698850086)
        ref[3,3,0,0]=(0.608039153856)/(o+1.)+(-0.552014076696)
        ref[3,3,0,1]=(1.40345390569)/(o+1.)+(-0.679775769743)
        ref[3,3,1,0]=(0.88635645412)/(o+1.)+(-0.446806405528)
        ref[3,3,1,1]=(0.368874266519)/(o+1.)+(0.58582266541)
        ref[3,3,2,0]=(-0.317351887963)/(o+1.)+(-0.355999178914)
        ref[3,3,2,1]=(1.71457769073)/(o+1.)+(-0.0490347690493)
        ref[3,4,0,0]=(-0.958065086929)/(o+1.)+(-0.0721186108707)
        ref[3,4,0,1]=(-0.811893317831)/(o+1.)+(-0.510573341517)
        ref[3,4,1,0]=(-1.15280446808)/(o+1.)+(0.428032044932)
        ref[3,4,1,1]=(-0.0988578042187)/(o+1.)+(-0.0713202314331)
        ref[3,4,2,0]=(-1.58242273456)/(o+1.)+(0.320901401659)
        ref[3,4,2,1]=(1.2464837106)/(o+1.)+(-0.201529569779)
      else:
        s[0,0,0,0]=(0.541504168376)*x[0]**o+(0.0573239599909)*x[0]+(0.0647250584452)*x[1]**o+(-0.703570101211)*x[1]+(0.328542610633)*x[2]**o+(-0.223400684365)*x[2]
        s[0,0,0,1]=(0.815893107692)*x[0]**o+(0.865481882711)*x[0]+(-0.46283176119)*x[1]**o+(-0.180772911655)*x[1]+(0.93497108116)*x[2]**o+(-0.742753139638)*x[2]
        s[0,0,1,0]=(-0.974209081736)*x[0]**o+(-0.672767028246)*x[0]+(-0.741240928071)*x[1]**o+(0.33719490832)*x[1]+(-0.686997621399)*x[2]**o+(-0.276966804283)*x[2]
        s[0,0,1,1]=(-0.857143471623)*x[0]**o+(-0.808079211925)*x[0]+(-0.953814060073)*x[1]**o+(0.961661670434)*x[1]+(0.189072084458)*x[2]**o+(-0.103208629753)*x[2]
        s[0,0,2,0]=(-0.20991399963)*x[0]**o+(-0.442037964882)*x[0]+(0.322578941372)*x[1]**o+(-0.161870387325)*x[1]+(0.585691936782)*x[2]**o+(-0.267825636094)*x[2]
        s[0,0,2,1]=(0.567642417211)*x[0]**o+(0.824356280251)*x[0]+(0.414108305262)*x[1]**o+(0.79515753838)*x[1]+(-0.894154134914)*x[2]**o+(-0.307671452735)*x[2]
        s[0,1,0,0]=(0.536896074085)*x[0]**o+(0.361562637795)*x[0]+(-0.158827271642)*x[1]**o+(0.0705880551767)*x[1]+(-0.332232895988)*x[2]**o+(-0.0305180398365)*x[2]
        s[0,1,0,1]=(0.633176778138)*x[0]**o+(-0.572376784007)*x[0]+(-0.294505207444)*x[1]**o+(-0.378967371632)*x[1]+(-0.89662018673)*x[2]**o+(0.287814422967)*x[2]
        s[0,1,1,0]=(-0.61097365344)*x[0]**o+(-0.363660280252)*x[0]+(0.190746604006)*x[1]**o+(-0.510926077271)*x[1]+(-0.957696939615)*x[2]**o+(-0.789299643561)*x[2]
        s[0,1,1,1]=(-0.835396537806)*x[0]**o+(-0.836746663696)*x[0]+(0.702723030555)*x[1]**o+(0.548005714327)*x[1]+(-0.924774335596)*x[2]**o+(-0.439240327617)*x[2]
        s[0,1,2,0]=(0.461024657935)*x[0]**o+(-0.373409962194)*x[0]+(0.743442026381)*x[1]**o+(0.657544000829)*x[1]+(0.674057315581)*x[2]**o+(-0.0183044871772)*x[2]
        s[0,1,2,1]=(0.164119453764)*x[0]**o+(-0.31204382122)*x[0]+(-0.424802851747)*x[1]**o+(-0.341469957442)*x[1]+(-0.231054265963)*x[2]**o+(0.128985361513)*x[2]
        s[0,2,0,0]=(0.963584200519)*x[0]**o+(-0.465275784807)*x[0]+(0.421929871546)*x[1]**o+(-0.211860414792)*x[1]+(-0.403745447737)*x[2]**o+(-0.379173494789)*x[2]
        s[0,2,0,1]=(-0.849597493426)*x[0]**o+(-0.174912181728)*x[0]+(-0.0534228854624)*x[1]**o+(0.826630462481)*x[1]+(-0.0210209049463)*x[2]**o+(-0.675409496795)*x[2]
        s[0,2,1,0]=(-0.565229688917)*x[0]**o+(0.485718751072)*x[0]+(0.700861807655)*x[1]**o+(0.901169377907)*x[1]+(0.890394572646)*x[2]**o+(0.596255846932)*x[2]
        s[0,2,1,1]=(0.876434872654)*x[0]**o+(-0.83625911596)*x[0]+(-0.539508005116)*x[1]**o+(-0.790017898115)*x[1]+(-0.043576285617)*x[2]**o+(-0.90281792431)*x[2]
        s[0,2,2,0]=(0.688686103492)*x[0]**o+(-0.479018765365)*x[0]+(-0.0041922877839)*x[1]**o+(0.360321450098)*x[1]+(0.520980145515)*x[2]**o+(0.781262738907)*x[2]
        s[0,2,2,1]=(-0.416899995117)*x[0]**o+(-0.018267124332)*x[0]+(-0.052482515113)*x[1]**o+(0.6928630854)*x[1]+(-0.163882932309)*x[2]**o+(0.478296669995)*x[2]
        s[0,3,0,0]=(-0.80317047721)*x[0]**o+(-0.475554188062)*x[0]+(0.194711479429)*x[1]**o+(0.866725258981)*x[1]+(0.500781673148)*x[2]**o+(-0.360414664576)*x[2]
        s[0,3,0,1]=(0.243316356193)*x[0]**o+(-0.109598635489)*x[0]+(0.579141755165)*x[1]**o+(0.876481264859)*x[1]+(-0.880903923571)*x[2]**o+(0.142946599744)*x[2]
        s[0,3,1,0]=(-0.507217017458)*x[0]**o+(0.641911252335)*x[0]+(-0.764452998266)*x[1]**o+(-0.505598974991)*x[1]+(0.641213955525)*x[2]**o+(-0.44758893679)*x[2]
        s[0,3,1,1]=(0.473545973781)*x[0]**o+(-0.00833748934844)*x[0]+(-0.00699017656361)*x[1]**o+(-0.643164722085)*x[1]+(0.726844277071)*x[2]**o+(-0.559279257114)*x[2]
        s[0,3,2,0]=(0.558099790948)*x[0]**o+(0.808990368539)*x[0]+(0.0431682029749)*x[1]**o+(0.0173216046894)*x[1]+(-0.394219668262)*x[2]**o+(-0.0336030329937)*x[2]
        s[0,3,2,1]=(-0.0222232210481)*x[0]**o+(-0.510190960052)*x[0]+(0.569389223293)*x[1]**o+(0.117046338884)*x[1]+(-0.671864592568)*x[2]**o+(-0.674239142474)*x[2]
        s[0,4,0,0]=(-0.644798822323)*x[0]**o+(-0.342116629892)*x[0]+(-0.168244417587)*x[1]**o+(-0.108919120953)*x[1]+(0.804101464972)*x[2]**o+(-0.0927979618696)*x[2]
        s[0,4,0,1]=(0.195838618741)*x[0]**o+(-0.111914225664)*x[0]+(-0.754003728212)*x[1]**o+(-0.096873749243)*x[1]+(-0.770744141788)*x[2]**o+(0.0291059262795)*x[2]
        s[0,4,1,0]=(0.321738393515)*x[0]**o+(-0.262140344086)*x[0]+(-0.130813570922)*x[1]**o+(0.749305862624)*x[1]+(-0.0933499641013)*x[2]**o+(0.583868661476)*x[2]
        s[0,4,1,1]=(-0.948399949958)*x[0]**o+(-0.514552416108)*x[0]+(0.403700494592)*x[1]**o+(0.319753801921)*x[1]+(0.068451176553)*x[2]**o+(0.236776599218)*x[2]
        s[0,4,2,0]=(-0.206993609035)*x[0]**o+(-0.693436356612)*x[0]+(0.376741069465)*x[1]**o+(-0.377934787271)*x[1]+(0.600888347529)*x[2]**o+(-0.806497559367)*x[2]
        s[0,4,2,1]=(0.519911073798)*x[0]**o+(0.209533160142)*x[0]+(-0.828631380624)*x[1]**o+(-0.758712343322)*x[1]+(0.469217244822)*x[2]**o+(0.777639952133)*x[2]
        s[1,0,0,0]=(-0.00938493494863)*x[0]**o+(0.035357976182)*x[0]+(0.530712339191)*x[1]**o+(0.163841019664)*x[1]+(-0.481052524963)*x[2]**o+(0.165894455618)*x[2]
        s[1,0,0,1]=(0.405172592393)*x[0]**o+(0.748490170154)*x[0]+(-0.579541033967)*x[1]**o+(-0.612353189754)*x[1]+(-0.0878018425036)*x[2]**o+(0.909881661391)*x[2]
        s[1,0,1,0]=(-0.131136361267)*x[0]**o+(0.0576430483263)*x[0]+(0.21369263102)*x[1]**o+(0.00660590703342)*x[1]+(0.079311350103)*x[2]**o+(0.878257991876)*x[2]
        s[1,0,1,1]=(-0.585221206397)*x[0]**o+(-0.791686741073)*x[0]+(0.631923355264)*x[1]**o+(-0.0277720531745)*x[1]+(-0.0777198468194)*x[2]**o+(-0.407162573619)*x[2]
        s[1,0,2,0]=(0.544255664465)*x[0]**o+(-0.51929274141)*x[0]+(-0.614260874796)*x[1]**o+(-0.951824251618)*x[1]+(0.48400676445)*x[2]**o+(-0.210092738748)*x[2]
        s[1,0,2,1]=(-0.901122575086)*x[0]**o+(-0.855042125739)*x[0]+(-0.660979212714)*x[1]**o+(0.874395337943)*x[1]+(-0.286202842553)*x[2]**o+(0.981270297086)*x[2]
        s[1,1,0,0]=(-0.752469757419)*x[0]**o+(0.0583950789171)*x[0]+(-0.699699086971)*x[1]**o+(-0.461650511393)*x[1]+(-0.272665950769)*x[2]**o+(-0.873526828849)*x[2]
        s[1,1,0,1]=(0.219706522447)*x[0]**o+(-0.73250655794)*x[0]+(0.0679566752122)*x[1]**o+(0.731592986993)*x[1]+(0.557341326754)*x[2]**o+(0.256780037036)*x[2]
        s[1,1,1,0]=(-0.820468252627)*x[0]**o+(0.163409736336)*x[0]+(-0.593882214431)*x[1]**o+(0.778760445642)*x[1]+(-0.822051914962)*x[2]**o+(0.920172785813)*x[2]
        s[1,1,1,1]=(-0.355904745872)*x[0]**o+(-0.192775924552)*x[0]+(0.598345031205)*x[1]**o+(-0.108244476269)*x[1]+(-0.895456581435)*x[2]**o+(0.207881298634)*x[2]
        s[1,1,2,0]=(0.781059740818)*x[0]**o+(0.186409854519)*x[0]+(0.92458071805)*x[1]**o+(0.197380801637)*x[1]+(-0.402971455411)*x[2]**o+(0.42922557636)*x[2]
        s[1,1,2,1]=(0.0064822572955)*x[0]**o+(0.846443443074)*x[0]+(0.500688303847)*x[1]**o+(-0.884507156164)*x[1]+(0.433224067545)*x[2]**o+(-0.0383476145137)*x[2]
        s[1,2,0,0]=(-0.945907369747)*x[0]**o+(-0.149404885342)*x[0]+(-0.911317716302)*x[1]**o+(0.718244019047)*x[1]+(0.097223049045)*x[2]**o+(0.95630310286)*x[2]
        s[1,2,0,1]=(-0.942956848602)*x[0]**o+(0.24519384765)*x[0]+(0.450231851653)*x[1]**o+(0.875902538898)*x[1]+(0.552657508084)*x[2]**o+(0.644477914693)*x[2]
        s[1,2,1,0]=(0.600235706495)*x[0]**o+(-0.142524845444)*x[0]+(0.216130207621)*x[1]**o+(-0.396701569264)*x[1]+(0.316884903524)*x[2]**o+(-0.0586626235465)*x[2]
        s[1,2,1,1]=(0.329111957957)*x[0]**o+(-0.508751506135)*x[0]+(0.911087373052)*x[1]**o+(-0.0983342912554)*x[1]+(-0.842070723886)*x[2]**o+(0.0746739548918)*x[2]
        s[1,2,2,0]=(0.387988330934)*x[0]**o+(-0.8408548375)*x[0]+(0.838833668833)*x[1]**o+(0.965224043099)*x[1]+(-0.860139996894)*x[2]**o+(-0.925778998652)*x[2]
        s[1,2,2,1]=(0.238958121236)*x[0]**o+(0.10607561646)*x[0]+(0.302265720665)*x[1]**o+(0.145320472959)*x[1]+(-0.677724287444)*x[2]**o+(-0.666145512045)*x[2]
        s[1,3,0,0]=(0.360431863997)*x[0]**o+(0.509815440438)*x[0]+(-0.178896172032)*x[1]**o+(0.316774928052)*x[1]+(-0.44743491823)*x[2]**o+(-0.651310611493)*x[2]
        s[1,3,0,1]=(-0.65053757716)*x[0]**o+(-0.375438056145)*x[0]+(0.360834001552)*x[1]**o+(-0.250086720496)*x[1]+(-0.71501469699)*x[2]**o+(0.0150228087513)*x[2]
        s[1,3,1,0]=(0.0628062775266)*x[0]**o+(-0.61619322157)*x[0]+(0.45067341985)*x[1]**o+(0.355471837743)*x[1]+(-0.174333549061)*x[2]**o+(-0.942897180184)*x[2]
        s[1,3,1,1]=(0.629724537179)*x[0]**o+(-0.0558311952601)*x[0]+(-0.956862453932)*x[1]**o+(-0.552443418291)*x[1]+(0.146303452118)*x[2]**o+(-0.770969067419)*x[2]
        s[1,3,2,0]=(-0.178686519235)*x[0]**o+(-0.171143660316)*x[0]+(-0.162449341623)*x[1]**o+(-0.1105712214)*x[1]+(0.119821653614)*x[2]**o+(0.209872978827)*x[2]
        s[1,3,2,1]=(0.312724690007)*x[0]**o+(-0.692554198329)*x[0]+(0.502590520541)*x[1]**o+(-0.729216407269)*x[1]+(0.302595866989)*x[2]**o+(-0.815710825541)*x[2]
        s[1,4,0,0]=(0.112006695888)*x[0]**o+(0.829399837819)*x[0]+(-0.419232858393)*x[1]**o+(0.146007406445)*x[1]+(0.304188744047)*x[2]**o+(0.46242131026)*x[2]
        s[1,4,0,1]=(0.964310426412)*x[0]**o+(0.196629594947)*x[0]+(0.124667636162)*x[1]**o+(-0.304792580501)*x[1]+(0.180285405611)*x[2]**o+(-0.543512746454)*x[2]
        s[1,4,1,0]=(-0.708536876552)*x[0]**o+(-0.680163366548)*x[0]+(0.0584969206489)*x[1]**o+(0.847180583253)*x[1]+(0.157570405709)*x[2]**o+(-0.746490964212)*x[2]
        s[1,4,1,1]=(0.0845668429821)*x[0]**o+(-0.889880680014)*x[0]+(-0.0771948903819)*x[1]**o+(-0.533952084212)*x[1]+(0.346150298712)*x[2]**o+(-0.78816093534)*x[2]
        s[1,4,2,0]=(0.375822387867)*x[0]**o+(-0.357919731843)*x[0]+(-0.373519792105)*x[1]**o+(0.641319045383)*x[1]+(0.351324745466)*x[2]**o+(-0.186703738887)*x[2]
        s[1,4,2,1]=(-0.60888359369)*x[0]**o+(0.694522376656)*x[0]+(-0.169823125514)*x[1]**o+(-0.856247555905)*x[1]+(0.00519595740385)*x[2]**o+(-0.875798589653)*x[2]
        s[2,0,0,0]=(0.469380548227)*x[0]**o+(-0.868913385859)*x[0]+(0.441755799104)*x[1]**o+(-0.493212921066)*x[1]+(-0.11743848736)*x[2]**o+(0.21526923935)*x[2]
        s[2,0,0,1]=(-0.151212323369)*x[0]**o+(-0.999958063251)*x[0]+(0.449187517422)*x[1]**o+(-0.583657187784)*x[1]+(-0.905753638698)*x[2]**o+(-0.945179993617)*x[2]
        s[2,0,1,0]=(-0.40460984789)*x[0]**o+(0.510786603887)*x[0]+(-0.953825153871)*x[1]**o+(0.226098717119)*x[1]+(-0.799990467305)*x[2]**o+(-0.429211164318)*x[2]
        s[2,0,1,1]=(0.867644883712)*x[0]**o+(0.714940923764)*x[0]+(-0.122362538438)*x[1]**o+(-0.44142264498)*x[1]+(0.437303901854)*x[2]**o+(0.841319415724)*x[2]
        s[2,0,2,0]=(0.13789233553)*x[0]**o+(0.472191869459)*x[0]+(0.468664678229)*x[1]**o+(-0.584558488838)*x[1]+(0.183146176041)*x[2]**o+(-0.893499132079)*x[2]
        s[2,0,2,1]=(0.875084187567)*x[0]**o+(0.550980479395)*x[0]+(0.312182122865)*x[1]**o+(-0.764545704288)*x[1]+(0.421014399576)*x[2]**o+(-0.652582305594)*x[2]
        s[2,1,0,0]=(-0.397160997842)*x[0]**o+(-0.401717267945)*x[0]+(0.612062897035)*x[1]**o+(0.147514233621)*x[1]+(0.994896686982)*x[2]**o+(-0.650669842592)*x[2]
        s[2,1,0,1]=(0.0859312342607)*x[0]**o+(-0.425173641674)*x[0]+(0.748616439433)*x[1]**o+(0.764026759608)*x[1]+(-0.990504424444)*x[2]**o+(-0.399709104793)*x[2]
        s[2,1,1,0]=(0.732840180154)*x[0]**o+(0.970502200871)*x[0]+(-0.413467445613)*x[1]**o+(-0.246800235561)*x[1]+(-0.881207449946)*x[2]**o+(0.882360789659)*x[2]
        s[2,1,1,1]=(0.689150562511)*x[0]**o+(-0.0933156400969)*x[0]+(-0.0951449636495)*x[1]**o+(0.143121640121)*x[1]+(-0.554471425755)*x[2]**o+(0.683081191337)*x[2]
        s[2,1,2,0]=(0.829863223837)*x[0]**o+(0.216508858271)*x[0]+(0.99016446302)*x[1]**o+(0.319444245585)*x[1]+(0.668648650551)*x[2]**o+(-0.896787700447)*x[2]
        s[2,1,2,1]=(-0.152966345983)*x[0]**o+(-0.775534953656)*x[0]+(0.0774085937181)*x[1]**o+(-0.59686756092)*x[1]+(0.224888875286)*x[2]**o+(-0.770576865174)*x[2]
        s[2,2,0,0]=(-0.831308688707)*x[0]**o+(0.0286055654454)*x[0]+(0.555782586294)*x[1]**o+(-0.863596073174)*x[1]+(0.0660913483948)*x[2]**o+(-0.691221916393)*x[2]
        s[2,2,0,1]=(0.97986500415)*x[0]**o+(-0.687991179162)*x[0]+(0.835947454347)*x[1]**o+(-0.425024918711)*x[1]+(0.632327959013)*x[2]**o+(-0.929603792153)*x[2]
        s[2,2,1,0]=(-0.785177923154)*x[0]**o+(0.498524809242)*x[0]+(0.657015457864)*x[1]**o+(-0.14042512075)*x[1]+(-0.523655529492)*x[2]**o+(0.0676392379718)*x[2]
        s[2,2,1,1]=(-0.282279979706)*x[0]**o+(0.402558983226)*x[0]+(0.349297772011)*x[1]**o+(-0.274102896576)*x[1]+(-0.976904085905)*x[2]**o+(-0.862615408506)*x[2]
        s[2,2,2,0]=(-0.442415692301)*x[0]**o+(-0.490020432017)*x[0]+(0.423330456256)*x[1]**o+(0.0308403124141)*x[1]+(-0.59889279992)*x[2]**o+(-0.784624469809)*x[2]
        s[2,2,2,1]=(-0.0571094173044)*x[0]**o+(-0.685673967597)*x[0]+(0.884019972654)*x[1]**o+(0.076633908669)*x[1]+(-0.436603596414)*x[2]**o+(0.643765166823)*x[2]
        s[2,3,0,0]=(-0.834122691134)*x[0]**o+(-0.316705170482)*x[0]+(0.0232732484874)*x[1]**o+(0.254076347383)*x[1]+(0.510661944881)*x[2]**o+(0.638231760578)*x[2]
        s[2,3,0,1]=(0.130809328027)*x[0]**o+(0.0475045505152)*x[0]+(0.253435146439)*x[1]**o+(0.15750814431)*x[1]+(0.930835277675)*x[2]**o+(0.215956216657)*x[2]
        s[2,3,1,0]=(-0.931893183482)*x[0]**o+(-0.165642362248)*x[0]+(-0.7975167088)*x[1]**o+(0.943743900193)*x[1]+(-0.834729159735)*x[2]**o+(0.310938564524)*x[2]
        s[2,3,1,1]=(-0.952550061226)*x[0]**o+(0.824399352189)*x[0]+(0.931907966548)*x[1]**o+(-0.46156960063)*x[1]+(0.6225788167)*x[2]**o+(-0.989192673067)*x[2]
        s[2,3,2,0]=(0.756326608673)*x[0]**o+(0.771528948716)*x[0]+(0.239160366604)*x[1]**o+(-0.452631969831)*x[1]+(-0.237358134546)*x[2]**o+(-0.839427184027)*x[2]
        s[2,3,2,1]=(0.0238504992639)*x[0]**o+(-0.103939311831)*x[0]+(0.291372616066)*x[1]**o+(0.672484596501)*x[1]+(0.494387122319)*x[2]**o+(0.486078041674)*x[2]
        s[2,4,0,0]=(-0.856552973776)*x[0]**o+(0.220667747539)*x[0]+(-0.901887042412)*x[1]**o+(0.162974138385)*x[1]+(0.884973432866)*x[2]**o+(-0.843375672742)*x[2]
        s[2,4,0,1]=(-0.581001018196)*x[0]**o+(0.766670058764)*x[0]+(0.195299626813)*x[1]**o+(0.491689692116)*x[1]+(-0.0851700136621)*x[2]**o+(0.556688214561)*x[2]
        s[2,4,1,0]=(0.341242000274)*x[0]**o+(0.254872207602)*x[0]+(0.721651164807)*x[1]**o+(0.817267858784)*x[1]+(0.525309787334)*x[2]**o+(-0.189345432506)*x[2]
        s[2,4,1,1]=(-0.461247890207)*x[0]**o+(-0.295916833174)*x[0]+(-0.795114931826)*x[1]**o+(0.760634909709)*x[1]+(0.628604868176)*x[2]**o+(-0.91582350733)*x[2]
        s[2,4,2,0]=(0.388037825656)*x[0]**o+(0.667391082754)*x[0]+(-0.76856572381)*x[1]**o+(-0.684287348308)*x[1]+(0.0434449534518)*x[2]**o+(-0.627587380339)*x[2]
        s[2,4,2,1]=(0.493746914184)*x[0]**o+(-0.47088277759)*x[0]+(-0.64457015181)*x[1]**o+(0.710566646648)*x[1]+(-0.603534864394)*x[2]**o+(0.952591999625)*x[2]
        s[3,0,0,0]=(-0.964941424537)*x[0]**o+(-0.701861794912)*x[0]+(0.631389613785)*x[1]**o+(0.216075041942)*x[1]+(-0.325257371696)*x[2]**o+(-0.291703451654)*x[2]
        s[3,0,0,1]=(-0.96299134603)*x[0]**o+(0.625796224057)*x[0]+(-0.418463355374)*x[1]**o+(-0.256336497204)*x[1]+(0.712728188959)*x[2]**o+(-0.770370932567)*x[2]
        s[3,0,1,0]=(-0.292515423665)*x[0]**o+(0.686334822809)*x[0]+(0.0305376943649)*x[1]**o+(0.872668037223)*x[1]+(-0.0541424770809)*x[2]**o+(-0.572416266221)*x[2]
        s[3,0,1,1]=(-0.690323374741)*x[0]**o+(-0.671336462076)*x[0]+(-0.166196941189)*x[1]**o+(-0.384967652901)*x[1]+(-0.61068991009)*x[2]**o+(0.776653422289)*x[2]
        s[3,0,2,0]=(-0.944428014582)*x[0]**o+(-0.301748262202)*x[0]+(-0.686840009057)*x[1]**o+(0.712060295627)*x[1]+(-0.254326623988)*x[2]**o+(0.400337107846)*x[2]
        s[3,0,2,1]=(-0.290963532129)*x[0]**o+(0.0269234770835)*x[0]+(-0.950364000347)*x[1]**o+(-0.460706646264)*x[1]+(-0.198333139868)*x[2]**o+(0.629854716945)*x[2]
        s[3,1,0,0]=(-0.555184083664)*x[0]**o+(0.131468013243)*x[0]+(0.282129000779)*x[1]**o+(-0.198577931709)*x[1]+(-0.175106810125)*x[2]**o+(0.00542537950498)*x[2]
        s[3,1,0,1]=(-0.284255526343)*x[0]**o+(-0.841249154497)*x[0]+(0.0213305454566)*x[1]**o+(-0.357636696086)*x[1]+(0.20888768221)*x[2]**o+(-0.503785042238)*x[2]
        s[3,1,1,0]=(-0.330999455854)*x[0]**o+(-0.159160032544)*x[0]+(0.52907127808)*x[1]**o+(0.0550513576484)*x[1]+(0.511074173459)*x[2]**o+(0.0159456796202)*x[2]
        s[3,1,1,1]=(-0.79143588969)*x[0]**o+(0.971248239785)*x[0]+(-0.0345752805688)*x[1]**o+(-0.0153770141636)*x[1]+(-0.716027682667)*x[2]**o+(-0.524455141515)*x[2]
        s[3,1,2,0]=(-0.582525406979)*x[0]**o+(0.397353443964)*x[0]+(-0.519936987381)*x[1]**o+(-0.21693645368)*x[1]+(0.869686334938)*x[2]**o+(-0.151814599828)*x[2]
        s[3,1,2,1]=(0.868456250139)*x[0]**o+(-0.568324799548)*x[0]+(-0.00346073709208)*x[1]**o+(-0.943912775939)*x[1]+(0.561497263651)*x[2]**o+(-0.923788209036)*x[2]
        s[3,2,0,0]=(-0.797982975817)*x[0]**o+(0.947494036065)*x[0]+(-0.250051189868)*x[1]**o+(0.472713798841)*x[1]+(0.902863101354)*x[2]**o+(0.613145652065)*x[2]
        s[3,2,0,1]=(0.252252158112)*x[0]**o+(0.301818562132)*x[0]+(0.664012773607)*x[1]**o+(0.0296698446724)*x[1]+(-0.00729763735658)*x[2]**o+(0.557599604788)*x[2]
        s[3,2,1,0]=(0.0570129540722)*x[0]**o+(-0.453241439283)*x[0]+(-0.71043548568)*x[1]**o+(-0.138422568238)*x[1]+(-0.611853026036)*x[2]**o+(0.253347515085)*x[2]
        s[3,2,1,1]=(-0.66828746218)*x[0]**o+(-0.751232642462)*x[0]+(-0.390775113865)*x[1]**o+(0.995150799921)*x[1]+(-0.292932619723)*x[2]**o+(-0.600225027769)*x[2]
        s[3,2,2,0]=(0.065423903522)*x[0]**o+(0.384335040067)*x[0]+(0.19683258538)*x[1]**o+(0.267437363607)*x[1]+(0.0294176448501)*x[2]**o+(0.155211007088)*x[2]
        s[3,2,2,1]=(-0.851636568209)*x[0]**o+(-0.761490675173)*x[0]+(-0.729643742868)*x[1]**o+(0.478780948031)*x[1]+(-0.036088351558)*x[2]**o+(0.823507165674)*x[2]
        s[3,3,0,0]=(0.482377341472)*x[0]**o+(0.139556800754)*x[0]+(-0.0112720934043)*x[1]**o+(0.298852046645)*x[1]+(0.145554009271)*x[2]**o+(-0.859974982251)*x[2]
        s[3,3,0,1]=(-0.137104991135)*x[0]**o+(-0.857876623449)*x[0]+(0.258417639346)*x[1]**o+(-0.389256714236)*x[1]+(-0.936199013774)*x[2]**o+(-0.312698752628)*x[2]
        s[3,3,1,0]=(-0.488668823485)*x[0]**o+(-0.771705852614)*x[0]+(-0.450034458754)*x[1]**o+(0.632396692772)*x[1]+(-0.105848719123)*x[2]**o+(0.358481590951)*x[2]
        s[3,3,1,1]=(-0.559261115256)*x[0]**o+(0.770050098012)*x[0]+(-0.125673261546)*x[1]**o+(0.863587697472)*x[1]+(0.94178472327)*x[2]**o+(0.691350347846)*x[2]
        s[3,3,2,0]=(0.348276322002)*x[0]**o+(-0.367507421115)*x[0]+(0.171562046637)*x[1]**o+(-0.98530139967)*x[1]+(-0.374019522945)*x[2]**o+(-0.0772490996658)*x[2]
        s[3,3,2,1]=(-0.91862240772)*x[0]**o+(-0.771336713978)*x[0]+(-0.510735333043)*x[1]**o+(-0.793667481925)*x[1]+(0.793023981359)*x[2]**o+(-0.211127108731)*x[2]
        s[3,4,0,0]=(-0.439892942063)*x[0]**o+(-0.452356550969)*x[0]+(0.255344783422)*x[1]**o+(0.966971078667)*x[1]+(0.770960867316)*x[2]**o+(-0.877647551728)*x[2]
        s[3,4,0,1]=(0.926860716799)*x[0]**o+(-0.325430377076)*x[0]+(0.593914577372)*x[1]**o+(-0.237069180519)*x[1]+(0.0779887304243)*x[2]**o+(0.0757684397401)*x[2]
        s[3,4,1,0]=(-0.865144446511)*x[0]**o+(0.220573510075)*x[0]+(0.574182016443)*x[1]**o+(0.585433407304)*x[1]+(-0.304138910485)*x[2]**o+(-0.26316229504)*x[2]
        s[3,4,1,1]=(-0.0941252376404)*x[0]**o+(-0.441659917216)*x[0]+(0.833947587523)*x[1]**o+(0.226992443402)*x[1]+(0.902425846073)*x[2]**o+(-0.525442142595)*x[2]
        s[3,4,2,0]=(0.782184984647)*x[0]**o+(0.0260219673241)*x[0]+(0.664501465465)*x[1]**o+(-0.571404130587)*x[1]+(0.419283611654)*x[2]**o+(-0.419135271863)*x[2]
        s[3,4,2,1]=(-0.944783897524)*x[0]**o+(-0.173931506017)*x[0]+(-0.0465484376595)*x[1]**o+(0.718604875734)*x[1]+(-0.388606992705)*x[2]**o+(-0.450843297215)*x[2]
        ref[0,0,0,0]=(0.934771837454)/(o+1.)+(-0.434823412793)
        ref[0,0,0,1]=(1.28803242766)/(o+1.)+(-0.0290220842909)
        ref[0,0,1,0]=(-2.40244763121)/(o+1.)+(-0.306269462104)
        ref[0,0,1,1]=(-1.62188544724)/(o+1.)+(0.0251869143786)
        ref[0,0,2,0]=(0.698356878524)/(o+1.)+(-0.435866994151)
        ref[0,0,2,1]=(0.0875965875576)/(o+1.)+(0.655921182948)
        ref[0,1,0,0]=(0.0458359064556)/(o+1.)+(0.200816326568)
        ref[0,1,0,1]=(-0.557948616036)/(o+1.)+(-0.331764866336)
        ref[0,1,1,0]=(-1.37792398905)/(o+1.)+(-0.831943000542)
        ref[0,1,1,1]=(-1.05744784285)/(o+1.)+(-0.363990638493)
        ref[0,1,2,0]=(1.8785239999)/(o+1.)+(0.132914775729)
        ref[0,1,2,1]=(-0.491737663947)/(o+1.)+(-0.262264208574)
        ref[0,2,0,0]=(0.981768624328)/(o+1.)+(-0.528154847194)
        ref[0,2,0,1]=(-0.924041283835)/(o+1.)+(-0.0118456080214)
        ref[0,2,1,0]=(1.02602669138)/(o+1.)+(0.991571987955)
        ref[0,2,1,1]=(0.293350581922)/(o+1.)+(-1.26454746919)
        ref[0,2,2,0]=(1.20547396122)/(o+1.)+(0.33128271182)
        ref[0,2,2,1]=(-0.633265442539)/(o+1.)+(0.576446315531)
        ref[0,3,0,0]=(-0.107677324633)/(o+1.)+(0.0153782031713)
        ref[0,3,0,1]=(-0.058445812213)/(o+1.)+(0.454914614557)
        ref[0,3,1,0]=(-0.630456060199)/(o+1.)+(-0.155638329723)
        ref[0,3,1,1]=(1.19340007429)/(o+1.)+(-0.605390734274)
        ref[0,3,2,0]=(0.207048325661)/(o+1.)+(0.396354470118)
        ref[0,3,2,1]=(-0.124698590323)/(o+1.)+(-0.533691881821)
        ref[0,4,0,0]=(-0.00894177493875)/(o+1.)+(-0.271916856357)
        ref[0,4,0,1]=(-1.32890925126)/(o+1.)+(-0.0898410243139)
        ref[0,4,1,0]=(0.0975748584924)/(o+1.)+(0.535517090007)
        ref[0,4,1,1]=(-0.476248278814)/(o+1.)+(0.0209889925155)
        ref[0,4,2,0]=(0.770635807959)/(o+1.)+(-0.938934351625)
        ref[0,4,2,1]=(0.160496937995)/(o+1.)+(0.114230384477)
        ref[1,0,0,0]=(0.0402748792799)/(o+1.)+(0.182546725732)
        ref[1,0,0,1]=(-0.262170284078)/(o+1.)+(0.523009320895)
        ref[1,0,1,0]=(0.161867619856)/(o+1.)+(0.471253473618)
        ref[1,0,1,1]=(-0.0310176979524)/(o+1.)+(-0.613310683933)
        ref[1,0,2,0]=(0.414001554119)/(o+1.)+(-0.840604865888)
        ref[1,0,2,1]=(-1.84830463035)/(o+1.)+(0.500311754645)
        ref[1,1,0,0]=(-1.72483479516)/(o+1.)+(-0.638391130662)
        ref[1,1,0,1]=(0.845004524414)/(o+1.)+(0.127933233045)
        ref[1,1,1,0]=(-2.23640238202)/(o+1.)+(0.931171483895)
        ref[1,1,1,1]=(-0.653016296101)/(o+1.)+(-0.0465695510937)
        ref[1,1,2,0]=(1.30266900346)/(o+1.)+(0.406508116258)
        ref[1,1,2,1]=(0.940394628687)/(o+1.)+(-0.0382056638017)
        ref[1,2,0,0]=(-1.760002037)/(o+1.)+(0.762571118282)
        ref[1,2,0,1]=(0.0599325111348)/(o+1.)+(0.88278715062)
        ref[1,2,1,0]=(1.13325081764)/(o+1.)+(-0.298944519128)
        ref[1,2,1,1]=(0.398128607122)/(o+1.)+(-0.266205921249)
        ref[1,2,2,0]=(0.366682002873)/(o+1.)+(-0.400704896527)
        ref[1,2,2,1]=(-0.136500445542)/(o+1.)+(-0.207374711313)
        ref[1,3,0,0]=(-0.265899226265)/(o+1.)+(0.0876398784989)
        ref[1,3,0,1]=(-1.0047182726)/(o+1.)+(-0.305250983945)
        ref[1,3,1,0]=(0.339146148316)/(o+1.)+(-0.601809282005)
        ref[1,3,1,1]=(-0.180834464635)/(o+1.)+(-0.689621840485)
        ref[1,3,2,0]=(-0.221314207244)/(o+1.)+(-0.0359209514441)
        ref[1,3,2,1]=(1.11791107754)/(o+1.)+(-1.11874071557)
        ref[1,4,0,0]=(-0.00303741845779)/(o+1.)+(0.718914277262)
        ref[1,4,0,1]=(1.26926346819)/(o+1.)+(-0.325837866004)
        ref[1,4,1,0]=(-0.492469550194)/(o+1.)+(-0.289736873753)
        ref[1,4,1,1]=(0.353522251312)/(o+1.)+(-1.10599684978)
        ref[1,4,2,0]=(0.353627341229)/(o+1.)+(0.0483477873267)
        ref[1,4,2,1]=(-0.7735107618)/(o+1.)+(-0.518761884451)
        ref[2,0,0,0]=(0.793697859971)/(o+1.)+(-0.573428533787)
        ref[2,0,0,1]=(-0.607778444644)/(o+1.)+(-1.26439762233)
        ref[2,0,1,0]=(-2.15842546907)/(o+1.)+(0.153837078344)
        ref[2,0,1,1]=(1.18258624713)/(o+1.)+(0.557418847254)
        ref[2,0,2,0]=(0.7897031898)/(o+1.)+(-0.502932875729)
        ref[2,0,2,1]=(1.60828071001)/(o+1.)+(-0.433073765244)
        ref[2,1,0,0]=(1.20979858617)/(o+1.)+(-0.452436438458)
        ref[2,1,0,1]=(-0.15595675075)/(o+1.)+(-0.0304279934298)
        ref[2,1,1,0]=(-0.561834715406)/(o+1.)+(0.803031377485)
        ref[2,1,1,1]=(0.0395341731061)/(o+1.)+(0.366443595681)
        ref[2,1,2,0]=(2.48867633741)/(o+1.)+(-0.180417298296)
        ref[2,1,2,1]=(0.149331123022)/(o+1.)+(-1.07148968987)
        ref[2,2,0,0]=(-0.209434754019)/(o+1.)+(-0.763106212061)
        ref[2,2,0,1]=(2.44814041751)/(o+1.)+(-1.02130994501)
        ref[2,2,1,0]=(-0.651817994783)/(o+1.)+(0.212869463232)
        ref[2,2,1,1]=(-0.9098862936)/(o+1.)+(-0.367079660928)
        ref[2,2,2,0]=(-0.617978035965)/(o+1.)+(-0.621902294706)
        ref[2,2,2,1]=(0.390306958936)/(o+1.)+(0.0173625539477)
        ref[2,3,0,0]=(-0.300187497765)/(o+1.)+(0.28780146874)
        ref[2,3,0,1]=(1.31507975214)/(o+1.)+(0.210484455741)
        ref[2,3,1,0]=(-2.56413905202)/(o+1.)+(0.544520051234)
        ref[2,3,1,1]=(0.601936722022)/(o+1.)+(-0.313181460754)
        ref[2,3,2,0]=(0.758128840732)/(o+1.)+(-0.260265102571)
        ref[2,3,2,1]=(0.809610237649)/(o+1.)+(0.527311663172)
        ref[2,4,0,0]=(-0.873466583322)/(o+1.)+(-0.229866893409)
        ref[2,4,0,1]=(-0.470871405045)/(o+1.)+(0.907523982721)
        ref[2,4,1,0]=(1.58820295242)/(o+1.)+(0.44139731694)
        ref[2,4,1,1]=(-0.627757953856)/(o+1.)+(-0.225552715397)
        ref[2,4,2,0]=(-0.337082944702)/(o+1.)+(-0.322241822947)
        ref[2,4,2,1]=(-0.75435810202)/(o+1.)+(0.596137934342)
        ref[3,0,0,0]=(-0.658809182448)/(o+1.)+(-0.388745102312)
        ref[3,0,0,1]=(-0.668726512445)/(o+1.)+(-0.200455602857)
        ref[3,0,1,0]=(-0.316120206381)/(o+1.)+(0.493293296905)
        ref[3,0,1,1]=(-1.46721022602)/(o+1.)+(-0.139825346344)
        ref[3,0,2,0]=(-1.88559464763)/(o+1.)+(0.405324570636)
        ref[3,0,2,1]=(-1.43966067234)/(o+1.)+(0.0980357738822)
        ref[3,1,0,0]=(-0.448161893009)/(o+1.)+(-0.0308422694803)
        ref[3,1,0,1]=(-0.0540372986773)/(o+1.)+(-0.851335446411)
        ref[3,1,1,0]=(0.709145995686)/(o+1.)+(-0.0440814976379)
        ref[3,1,1,1]=(-1.54203885293)/(o+1.)+(0.215708042053)
        ref[3,1,2,0]=(-0.232776059422)/(o+1.)+(0.0143011952278)
        ref[3,1,2,1]=(1.4264927767)/(o+1.)+(-1.21801289226)
        ref[3,2,0,0]=(-0.145171064331)/(o+1.)+(1.01667674349)
        ref[3,2,0,1]=(0.908967294363)/(o+1.)+(0.444544005797)
        ref[3,2,1,0]=(-1.26527555764)/(o+1.)+(-0.169158246218)
        ref[3,2,1,1]=(-1.35199519577)/(o+1.)+(-0.178153435155)
        ref[3,2,2,0]=(0.291674133752)/(o+1.)+(0.403491705381)
        ref[3,2,2,1]=(-1.61736866264)/(o+1.)+(0.270398719266)
        ref[3,3,0,0]=(0.616659257339)/(o+1.)+(-0.210783067426)
        ref[3,3,0,1]=(-0.814886365562)/(o+1.)+(-0.779916045157)
        ref[3,3,1,0]=(-1.04455200136)/(o+1.)+(0.109586215554)
        ref[3,3,1,1]=(0.256850346468)/(o+1.)+(1.16249407166)
        ref[3,3,2,0]=(0.145818845694)/(o+1.)+(-0.715028960225)
        ref[3,3,2,1]=(-0.636333759404)/(o+1.)+(-0.888065652317)
        ref[3,4,0,0]=(0.586412708675)/(o+1.)+(-0.181516512015)
        ref[3,4,0,1]=(1.5987640246)/(o+1.)+(-0.243365558927)
        ref[3,4,1,0]=(-0.595101340552)/(o+1.)+(0.271422311169)
        ref[3,4,1,1]=(1.64224819596)/(o+1.)+(-0.370054808205)
        ref[3,4,2,0]=(1.86597006177)/(o+1.)+(-0.482258717563)
        ref[3,4,2,1]=(-1.37993932789)/(o+1.)+(0.0469150362512)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.518496555195)*x[0]+(-1.7081835316)*x[1]
        ref=(-0.587946015314)/(o+1.)+(-0.300870480545)
      else:
        s=(0.271043713323)*x[0]+(-1.57101724485)*x[1]+(-0.650553623396)*x[2]
        ref=(-1.53180577827)/(o+1.)+(-0.209360688325)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.228569380378)*x[0]+(0.808739010281)*x[1]
        s[1]=(0.658420467565)*x[0]+(0.755018594164)*x[1]
        ref[0]=(0.635537444802)/(o+1.)+(-0.0276839074492)
        ref[1]=(0.431019006365)/(o+1.)+(0.491210027682)
      else:
        s[0]=(0.376210963228)*x[0]+(-1.32850082986)*x[1]+(0.536274160408)*x[2]
        s[1]=(-0.771050457843)*x[0]+(-0.16954629666)*x[1]+(-0.579642261351)*x[2]
        ref[0]=(-0.923045104748)/(o+1.)+(0.253514699262)
        ref[1]=(0.10262370672)/(o+1.)+(-0.811431361287)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(-0.405824989536)*x[0]+(-1.36105849061)*x[1]
        s[0,1]=(0.759504921251)*x[0]+(0.455515132277)*x[1]
        s[0,2]=(-1.37469026888)*x[0]+(1.35869839203)*x[1]
        s[0,3]=(-0.0438580623198)*x[0]+(0.34881551198)*x[1]
        s[0,4]=(-0.0097850380672)*x[0]+(0.61199301527)*x[1]
        s[1,0]=(-1.24950855151)*x[0]+(0.560604227152)*x[1]
        s[1,1]=(0.897125563557)*x[0]+(-0.362799426566)*x[1]
        s[1,2]=(-0.475265733496)*x[0]+(-0.906032283495)*x[1]
        s[1,3]=(0.184749910708)*x[0]+(0.961807058286)*x[1]
        s[1,4]=(-1.8130820241)*x[0]+(-0.0716801569248)*x[1]
        s[2,0]=(0.401163196364)*x[0]+(-0.122822662117)*x[1]
        s[2,1]=(-0.183141841931)*x[0]+(0.159646202564)*x[1]
        s[2,2]=(0.494070531254)*x[0]+(0.0283267498222)*x[1]
        s[2,3]=(-0.46326385119)*x[0]+(0.416856391877)*x[1]
        s[2,4]=(0.669972632028)*x[0]+(1.01207764348)*x[1]
        s[3,0]=(0.579509485187)*x[0]+(0.578664253542)*x[1]
        s[3,1]=(1.1904242989)*x[0]+(-0.617672917808)*x[1]
        s[3,2]=(1.01069090914)*x[0]+(-0.132861271259)*x[1]
        s[3,3]=(-1.02014113453)*x[0]+(-1.10411835631)*x[1]
        s[3,4]=(-0.178208550651)*x[0]+(-0.590646644191)*x[1]
        ref[0,0]=(-1.81750521429)/(o+1.)+(0.0253108670731)
        ref[0,1]=(1.13215741955)/(o+1.)+(0.0414313169864)
        ref[0,2]=(-0.346109803394)/(o+1.)+(0.165058963273)
        ref[0,3]=(1.16722378378)/(o+1.)+(-0.431133167061)
        ref[0,4]=(1.39150379152)/(o+1.)+(-0.39464790716)
        ref[1,0]=(-0.249819369301)/(o+1.)+(-0.219542477529)
        ref[1,1]=(-0.253832127476)/(o+1.)+(0.394079132234)
        ref[1,2]=(-0.514583575657)/(o+1.)+(-0.433357220667)
        ref[1,3]=(1.16430206364)/(o+1.)+(-0.00887254732364)
        ref[1,4]=(-1.40353822359)/(o+1.)+(-0.240611978717)
        ref[2,0]=(0.0129701699159)/(o+1.)+(0.132685182166)
        ref[2,1]=(-0.000259682696025)/(o+1.)+(-0.0116179783357)
        ref[2,2]=(-0.736785877312)/(o+1.)+(0.629591579194)
        ref[2,3]=(-0.319819594662)/(o+1.)+(0.136706067675)
        ref[2,4]=(1.34735651213)/(o+1.)+(0.167346881689)
        ref[3,0]=(1.02891998838)/(o+1.)+(0.0646268751734)
        ref[3,1]=(0.0226437097602)/(o+1.)+(0.275053835665)
        ref[3,2]=(-0.367947454589)/(o+1.)+(0.622888546236)
        ref[3,3]=(-1.07416112237)/(o+1.)+(-0.525049184236)
        ref[3,4]=(-0.423224656448)/(o+1.)+(-0.172815269197)
      else:
        s[0,0]=(0.391957465543)*x[0]+(-0.836212796005)*x[1]+(0.541170293661)*x[2]
        s[0,1]=(-0.324966494738)*x[0]+(0.62718596172)*x[1]+(-0.00110999366491)*x[2]
        s[0,2]=(0.377719216476)*x[0]+(1.20630293683)*x[1]+(-0.936497207115)*x[2]
        s[0,3]=(-1.52853135673)*x[0]+(-0.634683048191)*x[1]+(-0.533312108123)*x[2]
        s[0,4]=(-0.659359670617)*x[0]+(-1.23995360412)*x[1]+(0.669989279254)*x[2]
        s[1,0]=(-0.45120395504)*x[0]+(1.25119902426)*x[1]+(0.11139465291)*x[2]
        s[1,1]=(0.431699115955)*x[0]+(1.165662618)*x[1]+(-1.12543997002)*x[2]
        s[1,2]=(0.16640215316)*x[0]+(-1.81838137531)*x[1]+(-0.111705546449)*x[2]
        s[1,3]=(-1.56135929502)*x[0]+(0.425946749357)*x[1]+(0.255013889109)*x[2]
        s[1,4]=(0.320981318781)*x[0]+(-0.536492696361)*x[1]+(1.21768285382)*x[2]
        s[2,0]=(-0.600249234632)*x[0]+(-0.497626140165)*x[1]+(-0.0606420495108)*x[2]
        s[2,1]=(-0.144412727094)*x[0]+(-1.09139116844)*x[1]+(-0.39344520166)*x[2]
        s[2,2]=(-0.433006957898)*x[0]+(-0.287282800253)*x[1]+(8.69987670264e-05)*x[2]
        s[2,3]=(0.805491532167)*x[0]+(-1.17746983966)*x[1]+(0.720861745039)*x[2]
        s[2,4]=(-0.735630820744)*x[0]+(-1.01122803474)*x[1]+(1.72681276559)*x[2]
        s[3,0]=(-0.739861204924)*x[0]+(0.739077496919)*x[1]+(-0.0950784436747)*x[2]
        s[3,1]=(0.428369405071)*x[0]+(-0.0613035842349)*x[1]+(1.22804321951)*x[2]
        s[3,2]=(0.617885252444)*x[0]+(-1.63428007618)*x[1]+(-0.543704110813)*x[2]
        s[3,3]=(-1.52397266949)*x[0]+(-1.33821038473)*x[1]+(0.578214428096)*x[2]
        s[3,4]=(-0.904835367377)*x[0]+(0.156039380736)*x[1]+(-0.833632978592)*x[2]
        ref[0,0]=(-0.865179337718)/(o+1.)+(0.481047150458)
        ref[0,1]=(-0.427981609749)/(o+1.)+(0.364545541533)
        ref[0,2]=(0.738756225478)/(o+1.)+(-0.0456156396413)
        ref[0,3]=(-2.19753999815)/(o+1.)+(-0.249493257444)
        ref[0,4]=(-1.44493155356)/(o+1.)+(0.107803779038)
        ref[1,0]=(1.72841073711)/(o+1.)+(-0.408510507493)
        ref[1,1]=(0.688403407203)/(o+1.)+(-0.108240821633)
        ref[1,2]=(-0.0562832360974)/(o+1.)+(-0.853700766251)
        ref[1,3]=(-0.866037983384)/(o+1.)+(-0.00718033658349)
        ref[1,4]=(1.27849436197)/(o+1.)+(-0.138161442865)
        ref[2,0]=(-0.450814686687)/(o+1.)+(-0.353851368811)
        ref[2,1]=(-0.608210959793)/(o+1.)+(-0.510519068702)
        ref[2,2]=(-1.55008808776)/(o+1.)+(0.414942664188)
        ref[2,3]=(0.242744047944)/(o+1.)+(0.0530696948023)
        ref[2,4]=(-0.793195276992)/(o+1.)+(0.386574593552)
        ref[3,0]=(-0.33646344984)/(o+1.)+(0.12030064908)
        ref[3,1]=(1.37599194013)/(o+1.)+(0.109558550106)
        ref[3,2]=(0.0175612882637)/(o+1.)+(-0.788830111409)
        ref[3,3]=(-0.747828767237)/(o+1.)+(-0.768069929445)
        ref[3,4]=(-0.876520699909)/(o+1.)+(-0.352954132662)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.64806859919)*x[0]+(0.669871100749)*x[1]
        s[0,0,1]=(0.351912095781)*x[0]+(-0.795221209702)*x[1]
        s[0,1,0]=(1.67848694486)*x[0]+(0.294895162463)*x[1]
        s[0,1,1]=(0.360904647807)*x[0]+(1.08975889449)*x[1]
        s[1,0,0]=(-0.0924096394032)*x[0]+(-1.45324202762)*x[1]
        s[1,0,1]=(-0.160994756309)*x[0]+(0.551455091958)*x[1]
        s[1,1,0]=(1.32478072001)*x[0]+(-1.43998718605)*x[1]
        s[1,1,1]=(0.481907947963)*x[0]+(-1.62005808161)*x[1]
        s[2,0,0]=(1.28445085089)*x[0]+(0.139212942532)*x[1]
        s[2,0,1]=(0.13496510335)*x[0]+(-0.3090345869)*x[1]
        s[2,1,0]=(1.34702478342)*x[0]+(0.77762332551)*x[1]
        s[2,1,1]=(0.2270682866)*x[0]+(0.585652373115)*x[1]
        s[3,0,0]=(0.974462828497)*x[0]+(-0.0761666893839)*x[1]
        s[3,0,1]=(0.787654033315)*x[0]+(-0.665952609368)*x[1]
        s[3,1,0]=(-0.535337193665)*x[0]+(-0.10003612917)*x[1]
        s[3,1,1]=(0.90669722873)*x[0]+(0.0637346383222)*x[1]
        s[4,0,0]=(0.224906594767)*x[0]+(0.255067891624)*x[1]
        s[4,0,1]=(1.45652552317)*x[0]+(-0.896697311539)*x[1]
        s[4,1,0]=(0.0111655859631)*x[0]+(1.35361034898)*x[1]
        s[4,1,1]=(-1.49696987574)*x[0]+(-0.0630120006374)*x[1]
        s[5,0,0]=(-1.37528487034)*x[0]+(-1.03082532988)*x[1]
        s[5,0,1]=(0.298794708742)*x[0]+(-0.338420113242)*x[1]
        s[5,1,0]=(0.101365713196)*x[0]+(0.46239070902)*x[1]
        s[5,1,1]=(-0.329260078889)*x[0]+(0.0114285623122)*x[1]
        ref[0,0,0]=(1.04954961047)/(o+1.)+(0.134195044736)
        ref[0,0,1]=(-0.941346926582)/(o+1.)+(0.24901890633)
        ref[0,1,0]=(0.33540890863)/(o+1.)+(0.818986599345)
        ref[0,1,1]=(-0.0452512610385)/(o+1.)+(0.747957401666)
        ref[1,0,0]=(-0.0571552186929)/(o+1.)+(-0.744248224165)
        ref[1,0,1]=(-0.256660001756)/(o+1.)+(0.323560168702)
        ref[1,1,0]=(0.173031418514)/(o+1.)+(-0.144118942277)
        ref[1,1,1]=(-1.27331535296)/(o+1.)+(0.0675826096564)
        ref[2,0,0]=(1.27442931542)/(o+1.)+(0.0746172389969)
        ref[2,0,1]=(1.23134060492)/(o+1.)+(-0.702705044233)
        ref[2,1,0]=(1.44361222924)/(o+1.)+(0.340517939842)
        ref[2,1,1]=(0.858404770157)/(o+1.)+(-0.0228420552212)
        ref[3,0,0]=(0.53268376719)/(o+1.)+(0.182806185962)
        ref[3,0,1]=(0.962024080252)/(o+1.)+(-0.420161328153)
        ref[3,1,0]=(-1.06376900656)/(o+1.)+(0.214197841863)
        ref[3,1,1]=(0.437728927482)/(o+1.)+(0.266351469785)
        ref[4,0,0]=(-0.247848566495)/(o+1.)+(0.363911526443)
        ref[4,0,1]=(0.275565249528)/(o+1.)+(0.142131481049)
        ref[4,1,0]=(0.8446292547)/(o+1.)+(0.260073340122)
        ref[4,1,1]=(-1.92408551747)/(o+1.)+(0.182051820549)
        ref[5,0,0]=(-1.70867649172)/(o+1.)+(-0.348716854247)
        ref[5,0,1]=(-0.216851661682)/(o+1.)+(0.088613128591)
        ref[5,1,0]=(1.89734425595)/(o+1.)+(-0.666793916865)
        ref[5,1,1]=(-0.281093101446)/(o+1.)+(-0.0183692075654)
      else:
        s[0,0,0]=(0.142896594259)*x[0]+(-0.458264763854)*x[1]+(0.521893125785)*x[2]
        s[0,0,1]=(-0.141846975275)*x[0]+(-0.976845409768)*x[1]+(-1.14147520898)*x[2]
        s[0,1,0]=(0.141747271548)*x[0]+(-1.21811158828)*x[1]+(0.0640715930615)*x[2]
        s[0,1,1]=(0.462551385436)*x[0]+(-0.87929813895)*x[1]+(0.944848279649)*x[2]
        s[1,0,0]=(0.246197516828)*x[0]+(-0.0275751890238)*x[1]+(0.278318853806)*x[2]
        s[1,0,1]=(0.525663042785)*x[0]+(0.191011208259)*x[1]+(-1.27931005652)*x[2]
        s[1,1,0]=(-1.12276137904)*x[0]+(0.474659469962)*x[1]+(0.15450537857)*x[2]
        s[1,1,1]=(0.321959890457)*x[0]+(-0.0325903104321)*x[1]+(-0.735505871467)*x[2]
        s[2,0,0]=(-0.190194685931)*x[0]+(-0.277506154989)*x[1]+(0.0984734800625)*x[2]
        s[2,0,1]=(0.286493694398)*x[0]+(0.841559451892)*x[1]+(0.20572294365)*x[2]
        s[2,1,0]=(-0.0132570386632)*x[0]+(-0.160857995452)*x[1]+(-0.813758770212)*x[2]
        s[2,1,1]=(0.00519333657271)*x[0]+(0.0324836424574)*x[1]+(0.493299621992)*x[2]
        s[3,0,0]=(0.300624746821)*x[0]+(0.211207146953)*x[1]+(1.0860831317)*x[2]
        s[3,0,1]=(-0.696336553938)*x[0]+(-1.28551896118)*x[1]+(0.738769481443)*x[2]
        s[3,1,0]=(0.15510282832)*x[0]+(-1.15795290548)*x[1]+(0.543235378485)*x[2]
        s[3,1,1]=(-0.714308237763)*x[0]+(0.239072001232)*x[1]+(-0.187168055861)*x[2]
        s[4,0,0]=(0.624611061079)*x[0]+(1.11568209096)*x[1]+(-1.46683771155)*x[2]
        s[4,0,1]=(1.45301860552)*x[0]+(0.74921646591)*x[1]+(-0.796630526412)*x[2]
        s[4,1,0]=(0.560555339207)*x[0]+(-0.421267308186)*x[1]+(-0.222312671328)*x[2]
        s[4,1,1]=(0.807845684562)*x[0]+(0.323266951017)*x[1]+(1.51940950039)*x[2]
        s[5,0,0]=(-0.558067921385)*x[0]+(-0.214928793546)*x[1]+(1.01402046886)*x[2]
        s[5,0,1]=(1.20688436616)*x[0]+(0.0247589671447)*x[1]+(-0.368536410815)*x[2]
        s[5,1,0]=(-0.324541665933)*x[0]+(-0.205515389215)*x[1]+(0.758278431203)*x[2]
        s[5,1,1]=(-0.620405057772)*x[0]+(-0.67647528448)*x[1]+(1.32328947052)*x[2]
        ref[0,0,0]=(0.260018146786)/(o+1.)+(-0.0267465952979)
        ref[0,0,1]=(-0.435502524101)/(o+1.)+(-0.912332534962)
        ref[0,1,0]=(-0.0330362836983)/(o+1.)+(-0.489628219987)
        ref[0,1,1]=(1.07484163998)/(o+1.)+(-0.273370056924)
        ref[1,0,0]=(-0.523779853488)/(o+1.)+(0.510360517549)
        ref[1,0,1]=(0.448972814496)/(o+1.)+(-0.505804309986)
        ref[1,1,0]=(-1.09840793106)/(o+1.)+(0.302405700276)
        ref[1,1,1]=(0.130511241273)/(o+1.)+(-0.288323766357)
        ref[2,0,0]=(-0.805299870557)/(o+1.)+(0.21803625485)
        ref[2,0,1]=(0.271181617798)/(o+1.)+(0.531297236071)
        ref[2,1,0]=(-1.29575884246)/(o+1.)+(0.153942519064)
        ref[2,1,1]=(-0.600379662508)/(o+1.)+(0.565678131765)
        ref[3,0,0]=(1.87814767227)/(o+1.)+(-0.140116323394)
        ref[3,0,1]=(-0.195825320078)/(o+1.)+(-0.523630356799)
        ref[3,1,0]=(-0.959261224384)/(o+1.)+(0.249823262853)
        ref[3,1,1]=(1.05531553015)/(o+1.)+(-0.858859911271)
        ref[4,0,0]=(0.65497734235)/(o+1.)+(-0.190760950929)
        ref[4,0,1]=(1.37092574409)/(o+1.)+(0.0173394004651)
        ref[4,1,0]=(1.27754767629)/(o+1.)+(-0.6802861583)
        ref[4,1,1]=(1.23997013471)/(o+1.)+(0.70527600063)
        ref[5,0,0]=(-0.383142890439)/(o+1.)+(0.312083322186)
        ref[5,0,1]=(1.64380369652)/(o+1.)+(-0.390348387013)
        ref[5,1,0]=(-0.812713549757)/(o+1.)+(0.520467462905)
        ref[5,1,1]=(0.600505313178)/(o+1.)+(-0.287048092454)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.235888975732)*x[0]+(-0.734372259615)*x[1]
        s[0,0,0,1]=(0.0665724430605)*x[0]+(-0.0890002398629)*x[1]
        s[0,0,1,0]=(-1.19774759676)*x[0]+(1.00321581155)*x[1]
        s[0,0,1,1]=(0.93370004223)*x[0]+(-0.953835581453)*x[1]
        s[0,0,2,0]=(-0.0557204731966)*x[0]+(0.94945217411)*x[1]
        s[0,0,2,1]=(-0.772628465097)*x[0]+(1.77100579145)*x[1]
        s[0,1,0,0]=(-0.405653863463)*x[0]+(0.494482129094)*x[1]
        s[0,1,0,1]=(0.636236763209)*x[0]+(-1.63254203234)*x[1]
        s[0,1,1,0]=(0.941764211044)*x[0]+(0.699520209508)*x[1]
        s[0,1,1,1]=(1.05045643599)*x[0]+(0.0596636815695)*x[1]
        s[0,1,2,0]=(0.578668042684)*x[0]+(0.285442340394)*x[1]
        s[0,1,2,1]=(0.469149909671)*x[0]+(1.26526906754)*x[1]
        s[0,2,0,0]=(1.05435496407)*x[0]+(-0.592645628655)*x[1]
        s[0,2,0,1]=(0.066181979241)*x[0]+(-1.42368311585)*x[1]
        s[0,2,1,0]=(-0.896588999422)*x[0]+(1.44881386304)*x[1]
        s[0,2,1,1]=(-0.862544854709)*x[0]+(-0.398360134784)*x[1]
        s[0,2,2,0]=(1.02969811319)*x[0]+(0.318993808627)*x[1]
        s[0,2,2,1]=(0.0143773859487)*x[0]+(-0.327118788648)*x[1]
        s[0,3,0,0]=(0.021196438108)*x[0]+(1.8515159226)*x[1]
        s[0,3,0,1]=(0.627726779803)*x[0]+(-0.302716383303)*x[1]
        s[0,3,1,0]=(0.165624370858)*x[0]+(0.274919760437)*x[1]
        s[0,3,1,1]=(0.283214725106)*x[0]+(-0.872378951063)*x[1]
        s[0,3,2,0]=(-0.83391985409)*x[0]+(0.731217556829)*x[1]
        s[0,3,2,1]=(0.918833558754)*x[0]+(0.714715370412)*x[1]
        s[0,4,0,0]=(-0.503211677055)*x[0]+(0.0289310095577)*x[1]
        s[0,4,0,1]=(0.597129238132)*x[0]+(0.540475624893)*x[1]
        s[0,4,1,0]=(0.687211570349)*x[0]+(0.594018658674)*x[1]
        s[0,4,1,1]=(1.5970484642)*x[0]+(-0.495319600999)*x[1]
        s[0,4,2,0]=(1.20627016067)*x[0]+(0.461261823726)*x[1]
        s[0,4,2,1]=(0.165927055028)*x[0]+(1.09007328428)*x[1]
        s[1,0,0,0]=(1.16105917598)*x[0]+(1.82590164801)*x[1]
        s[1,0,0,1]=(-1.13479961391)*x[0]+(0.0351464050353)*x[1]
        s[1,0,1,0]=(-0.318954647255)*x[0]+(0.210946813363)*x[1]
        s[1,0,1,1]=(1.31393538581)*x[0]+(0.198463680367)*x[1]
        s[1,0,2,0]=(1.17799558613)*x[0]+(0.514357347746)*x[1]
        s[1,0,2,1]=(0.571191341129)*x[0]+(0.694971250457)*x[1]
        s[1,1,0,0]=(0.132901410189)*x[0]+(-0.262070778236)*x[1]
        s[1,1,0,1]=(-0.871431837436)*x[0]+(-0.514707012778)*x[1]
        s[1,1,1,0]=(-0.319003186732)*x[0]+(-0.014973803902)*x[1]
        s[1,1,1,1]=(0.260213741998)*x[0]+(1.32908581921)*x[1]
        s[1,1,2,0]=(0.811339404893)*x[0]+(-0.304564391555)*x[1]
        s[1,1,2,1]=(-0.062193223101)*x[0]+(-0.911704705741)*x[1]
        s[1,2,0,0]=(0.0932623640149)*x[0]+(1.12570360092)*x[1]
        s[1,2,0,1]=(1.16504942277)*x[0]+(-0.385840092559)*x[1]
        s[1,2,1,0]=(1.3440680469)*x[0]+(0.532696669818)*x[1]
        s[1,2,1,1]=(1.31050128725)*x[0]+(0.00906728815983)*x[1]
        s[1,2,2,0]=(-1.36931832474)*x[0]+(-1.70562158659)*x[1]
        s[1,2,2,1]=(-1.8375028036)*x[0]+(1.52366054242)*x[1]
        s[1,3,0,0]=(0.474023303611)*x[0]+(0.437518737767)*x[1]
        s[1,3,0,1]=(-0.42921544272)*x[0]+(0.387647202618)*x[1]
        s[1,3,1,0]=(0.863045413096)*x[0]+(-0.36260979035)*x[1]
        s[1,3,1,1]=(-0.654365401799)*x[0]+(0.533926246461)*x[1]
        s[1,3,2,0]=(0.418152564814)*x[0]+(1.11544647745)*x[1]
        s[1,3,2,1]=(-1.60947873228)*x[0]+(0.591727770738)*x[1]
        s[1,4,0,0]=(-1.51807668572)*x[0]+(-1.01613349614)*x[1]
        s[1,4,0,1]=(-1.50163863727)*x[0]+(1.50376272164)*x[1]
        s[1,4,1,0]=(0.320776223369)*x[0]+(-0.700391339598)*x[1]
        s[1,4,1,1]=(0.745224476612)*x[0]+(-0.850870606209)*x[1]
        s[1,4,2,0]=(-0.179598404379)*x[0]+(1.75799776411)*x[1]
        s[1,4,2,1]=(0.00909330900106)*x[0]+(-0.077399205942)*x[1]
        s[2,0,0,0]=(0.505719447396)*x[0]+(-0.275603298829)*x[1]
        s[2,0,0,1]=(1.46787325811)*x[0]+(0.534793377012)*x[1]
        s[2,0,1,0]=(0.77645891983)*x[0]+(0.540540443218)*x[1]
        s[2,0,1,1]=(0.429661881583)*x[0]+(1.43421311479)*x[1]
        s[2,0,2,0]=(1.02806543045)*x[0]+(0.931834883879)*x[1]
        s[2,0,2,1]=(0.763031666632)*x[0]+(0.873653388204)*x[1]
        s[2,1,0,0]=(0.261772733655)*x[0]+(-1.18679909814)*x[1]
        s[2,1,0,1]=(-1.61662908583)*x[0]+(1.10646142209)*x[1]
        s[2,1,1,0]=(0.188527773811)*x[0]+(-0.772410233645)*x[1]
        s[2,1,1,1]=(-0.869931105194)*x[0]+(1.51014899284)*x[1]
        s[2,1,2,0]=(0.252478837689)*x[0]+(1.22480134325)*x[1]
        s[2,1,2,1]=(1.49578867187)*x[0]+(-0.877973056373)*x[1]
        s[2,2,0,0]=(-0.209568005598)*x[0]+(-0.0462992389671)*x[1]
        s[2,2,0,1]=(-0.298130269216)*x[0]+(-0.908352713549)*x[1]
        s[2,2,1,0]=(-0.586370615835)*x[0]+(-1.72704626602)*x[1]
        s[2,2,1,1]=(1.26217059553)*x[0]+(0.676548549908)*x[1]
        s[2,2,2,0]=(0.961200845015)*x[0]+(0.612151157303)*x[1]
        s[2,2,2,1]=(0.700873758828)*x[0]+(0.800662765762)*x[1]
        s[2,3,0,0]=(-0.319747641322)*x[0]+(0.689929330298)*x[1]
        s[2,3,0,1]=(-1.04393759081)*x[0]+(-1.80449230281)*x[1]
        s[2,3,1,0]=(-0.0410072962318)*x[0]+(-0.678000289402)*x[1]
        s[2,3,1,1]=(-0.329381648117)*x[0]+(-0.935487069857)*x[1]
        s[2,3,2,0]=(-0.907460976718)*x[0]+(0.0314212805943)*x[1]
        s[2,3,2,1]=(0.593604907619)*x[0]+(-0.479146057787)*x[1]
        s[2,4,0,0]=(-0.0337048332079)*x[0]+(0.239721443625)*x[1]
        s[2,4,0,1]=(0.721609580339)*x[0]+(-1.06464423044)*x[1]
        s[2,4,1,0]=(0.670568080353)*x[0]+(1.28774286264)*x[1]
        s[2,4,1,1]=(0.603608338555)*x[0]+(0.323996221169)*x[1]
        s[2,4,2,0]=(0.0371268677698)*x[0]+(0.401814556385)*x[1]
        s[2,4,2,1]=(0.436926475796)*x[0]+(-0.819433312575)*x[1]
        s[3,0,0,0]=(-0.0190186791033)*x[0]+(0.501672647061)*x[1]
        s[3,0,0,1]=(0.00060856018325)*x[0]+(-0.414668792931)*x[1]
        s[3,0,1,0]=(0.699428985955)*x[0]+(0.612631528371)*x[1]
        s[3,0,1,1]=(-0.659371576463)*x[0]+(-0.0982545515259)*x[1]
        s[3,0,2,0]=(0.444351889183)*x[0]+(0.53998411713)*x[1]
        s[3,0,2,1]=(-0.281563709018)*x[0]+(-0.486501437256)*x[1]
        s[3,1,0,0]=(0.407521411466)*x[0]+(-0.749870044816)*x[1]
        s[3,1,0,1]=(-0.638780264813)*x[0]+(0.230397511262)*x[1]
        s[3,1,1,0]=(1.16000440204)*x[0]+(0.13074916095)*x[1]
        s[3,1,1,1]=(-1.30550439046)*x[0]+(-1.33788704039)*x[1]
        s[3,1,2,0]=(-0.540767742404)*x[0]+(-0.148316212185)*x[1]
        s[3,1,2,1]=(1.70452499495)*x[0]+(0.935421717463)*x[1]
        s[3,2,0,0]=(0.320990904519)*x[0]+(0.467953609756)*x[1]
        s[3,2,0,1]=(-0.26002228203)*x[0]+(-0.925765706976)*x[1]
        s[3,2,1,0]=(0.925754935382)*x[0]+(0.240432086422)*x[1]
        s[3,2,1,1]=(0.0214680080146)*x[0]+(-1.11507977191)*x[1]
        s[3,2,2,0]=(-0.187946153679)*x[0]+(0.024781322847)*x[1]
        s[3,2,2,1]=(-1.41866510495)*x[0]+(0.596173034069)*x[1]
        s[3,3,0,0]=(0.77087649562)*x[0]+(0.848168811103)*x[1]
        s[3,3,0,1]=(0.218810390517)*x[0]+(0.116909485563)*x[1]
        s[3,3,1,0]=(-0.889981738645)*x[0]+(-0.506355969844)*x[1]
        s[3,3,1,1]=(1.15870903185)*x[0]+(-0.8158119166)*x[1]
        s[3,3,2,0]=(-0.345230989856)*x[0]+(-0.138496974465)*x[1]
        s[3,3,2,1]=(-0.819830787817)*x[0]+(-1.66747998417)*x[1]
        s[3,4,0,0]=(-1.74647564112)*x[0]+(0.22309219172)*x[1]
        s[3,4,0,1]=(0.146804717412)*x[0]+(0.493084518827)*x[1]
        s[3,4,1,0]=(-0.536943383893)*x[0]+(-0.686288132381)*x[1]
        s[3,4,1,1]=(-0.142983666789)*x[0]+(-0.183927597857)*x[1]
        s[3,4,2,0]=(-0.173071632973)*x[0]+(-0.696870700802)*x[1]
        s[3,4,2,1]=(-0.359163238166)*x[0]+(-0.172033035327)*x[1]
        ref[0,0,0,0]=(-0.603598770198)/(o+1.)+(0.0525577431575)
        ref[0,0,0,1]=(1.04505857398)/(o+1.)+(-0.533743185393)
        ref[0,0,1,0]=(0.00772071573504)/(o+1.)+(-0.101126250473)
        ref[0,0,1,1]=(-0.290587702404)/(o+1.)+(0.135226081591)
        ref[0,0,2,0]=(0.614426628234)/(o+1.)+(0.139652536339)
        ref[0,0,2,1]=(-0.0228094425795)/(o+1.)+(0.510593384465)
        ref[0,1,0,0]=(-0.436885078328)/(o+1.)+(0.262856671979)
        ref[0,1,0,1]=(0.0510259667548)/(o+1.)+(-0.523665617945)
        ref[0,1,1,0]=(0.744462079536)/(o+1.)+(0.448411170508)
        ref[0,1,1,1]=(1.52720962375)/(o+1.)+(-0.208544753096)
        ref[0,1,2,0]=(0.73405306307)/(o+1.)+(0.0650286600044)
        ref[0,1,2,1]=(0.550924984933)/(o+1.)+(0.591746996137)
        ref[0,2,0,0]=(-0.347797813257)/(o+1.)+(0.404753574338)
        ref[0,2,0,1]=(0.187043042708)/(o+1.)+(-0.772272089659)
        ref[0,2,1,0]=(-0.316098850787)/(o+1.)+(0.434161857202)
        ref[0,2,1,1]=(-0.536124734534)/(o+1.)+(-0.36239012748)
        ref[0,2,2,0]=(0.585045523113)/(o+1.)+(0.381823199352)
        ref[0,2,2,1]=(0.422243751365)/(o+1.)+(-0.367492577032)
        ref[0,3,0,0]=(0.91381158808)/(o+1.)+(0.479450386313)
        ref[0,3,0,1]=(0.209840671177)/(o+1.)+(0.0575848626616)
        ref[0,3,1,0]=(0.158639778589)/(o+1.)+(0.140952176353)
        ref[0,3,1,1]=(-0.364808895601)/(o+1.)+(-0.112177665178)
        ref[0,3,2,0]=(-0.198937574741)/(o+1.)+(0.0481176387404)
        ref[0,3,2,1]=(1.05420599501)/(o+1.)+(0.289671467078)
        ref[0,4,0,0]=(0.811701728881)/(o+1.)+(-0.642991198189)
        ref[0,4,0,1]=(0.693754647803)/(o+1.)+(0.221925107611)
        ref[0,4,1,0]=(1.05453609011)/(o+1.)+(0.113347069457)
        ref[0,4,1,1]=(0.117274271978)/(o+1.)+(0.49222729561)
        ref[0,4,2,0]=(0.496795563283)/(o+1.)+(0.585368210556)
        ref[0,4,2,1]=(0.83272690983)/(o+1.)+(0.211636714741)
        ref[1,0,0,0]=(1.46878592273)/(o+1.)+(0.75908745063)
        ref[1,0,0,1]=(-0.250785630529)/(o+1.)+(-0.424433789171)
        ref[1,0,1,0]=(0.161786720779)/(o+1.)+(-0.134897277336)
        ref[1,0,1,1]=(1.12719401256)/(o+1.)+(0.192602526813)
        ref[1,0,2,0]=(0.7197235394)/(o+1.)+(0.486314697239)
        ref[1,0,2,1]=(1.02804458464)/(o+1.)+(0.119059003471)
        ref[1,1,0,0]=(0.0361584520541)/(o+1.)+(-0.0826639100507)
        ref[1,1,0,1]=(-1.06091812324)/(o+1.)+(-0.162610363489)
        ref[1,1,1,0]=(1.13776585837)/(o+1.)+(-0.735871424501)
        ref[1,1,1,1]=(1.4740891177)/(o+1.)+(0.0576052217538)
        ref[1,1,2,0]=(0.280616623834)/(o+1.)+(0.113079194752)
        ref[1,1,2,1]=(-0.491890699653)/(o+1.)+(-0.241003614595)
        ref[1,2,0,0]=(-0.0342838844534)/(o+1.)+(0.626624924695)
        ref[1,2,0,1]=(0.770911868081)/(o+1.)+(0.00414873106652)
        ref[1,2,1,0]=(0.30749084682)/(o+1.)+(0.784636934951)
        ref[1,2,1,1]=(1.8638645006)/(o+1.)+(-0.272147962595)
        ref[1,2,2,0]=(-1.46841213372)/(o+1.)+(-0.803263888803)
        ref[1,2,2,1]=(-0.302454198157)/(o+1.)+(-0.00569403150815)
        ref[1,3,0,0]=(0.493917522063)/(o+1.)+(0.208812259657)
        ref[1,3,0,1]=(0.476274657976)/(o+1.)+(-0.258921449039)
        ref[1,3,1,0]=(-0.736248174893)/(o+1.)+(0.618341898819)
        ref[1,3,1,1]=(-0.697129670691)/(o+1.)+(0.288345257677)
        ref[1,3,2,0]=(0.303153912627)/(o+1.)+(0.615222564816)
        ref[1,3,2,1]=(-0.58122630211)/(o+1.)+(-0.218262329715)
        ref[1,4,0,0]=(-1.15442442968)/(o+1.)+(-0.689892876091)
        ref[1,4,0,1]=(0.0585826445649)/(o+1.)+(-0.0282292800982)
        ref[1,4,1,0]=(0.3252337262)/(o+1.)+(-0.352424421214)
        ref[1,4,1,1]=(0.512701244088)/(o+1.)+(-0.309173686843)
        ref[1,4,2,0]=(1.5506882457)/(o+1.)+(0.0138555570174)
        ref[1,4,2,1]=(-0.940773035102)/(o+1.)+(0.436233569081)
        ref[2,0,0,0]=(1.54247754874)/(o+1.)+(-0.656180700088)
        ref[2,0,0,1]=(0.988306961756)/(o+1.)+(0.507179836684)
        ref[2,0,1,0]=(-0.416628824296)/(o+1.)+(0.866814093672)
        ref[2,0,1,1]=(1.41006066195)/(o+1.)+(0.22690716721)
        ref[2,0,2,0]=(0.353011705537)/(o+1.)+(0.803444304396)
        ref[2,0,2,1]=(0.687642260874)/(o+1.)+(0.474521396981)
        ref[2,1,0,0]=(-0.218198987148)/(o+1.)+(-0.353413688668)
        ref[2,1,0,1]=(-0.635306433562)/(o+1.)+(0.0625693849126)
        ref[2,1,1,0]=(0.53416867099)/(o+1.)+(-0.559025565412)
        ref[2,1,1,1]=(0.00543686305333)/(o+1.)+(0.317390512297)
        ref[2,1,2,0]=(0.916573120847)/(o+1.)+(0.280353530048)
        ref[2,1,2,1]=(-0.0537588152546)/(o+1.)+(0.335787215373)
        ref[2,2,0,0]=(1.0055660817)/(o+1.)+(-0.630716663133)
        ref[2,2,0,1]=(0.420474306577)/(o+1.)+(-0.813478644671)
        ref[2,2,1,0]=(-0.723008657238)/(o+1.)+(-0.795204112308)
        ref[2,2,1,1]=(0.306188174034)/(o+1.)+(0.816265485704)
        ref[2,2,2,0]=(0.806158569368)/(o+1.)+(0.383596716475)
        ref[2,2,2,1]=(1.6353377865)/(o+1.)+(-0.066900630957)
        ref[2,3,0,0]=(-1.24813850815)/(o+1.)+(0.809160098565)
        ref[2,3,0,1]=(-1.56942041127)/(o+1.)+(-0.639504741178)
        ref[2,3,1,0]=(-1.49605878837)/(o+1.)+(0.388525601366)
        ref[2,3,1,1]=(-0.448403645878)/(o+1.)+(-0.408232536048)
        ref[2,3,2,0]=(-1.13568794342)/(o+1.)+(0.129824123646)
        ref[2,3,2,1]=(0.596733666259)/(o+1.)+(-0.241137408214)
        ref[2,4,0,0]=(0.0191094707617)/(o+1.)+(0.0934535698276)
        ref[2,4,0,1]=(0.185204097773)/(o+1.)+(-0.264119373935)
        ref[2,4,1,0]=(1.55855563639)/(o+1.)+(0.199877653305)
        ref[2,4,1,1]=(0.11469142184)/(o+1.)+(0.406456568942)
        ref[2,4,2,0]=(-0.954321542363)/(o+1.)+(0.696631483259)
        ref[2,4,2,1]=(-0.810227566168)/(o+1.)+(0.213860364695)
        ref[3,0,0,0]=(0.453436736881)/(o+1.)+(0.0146086155385)
        ref[3,0,0,1]=(-0.210824760654)/(o+1.)+(-0.101617736047)
        ref[3,0,1,0]=(0.414458928012)/(o+1.)+(0.448800793157)
        ref[3,0,1,1]=(0.346129385596)/(o+1.)+(-0.551877756793)
        ref[3,0,2,0]=(1.85496275992)/(o+1.)+(-0.435313376805)
        ref[3,0,2,1]=(-0.110500317363)/(o+1.)+(-0.328782414455)
        ref[3,1,0,0]=(0.0399449935267)/(o+1.)+(-0.191146813438)
        ref[3,1,0,1]=(0.133080949177)/(o+1.)+(-0.270731851364)
        ref[3,1,1,0]=(0.139648356416)/(o+1.)+(0.575552603289)
        ref[3,1,1,1]=(-1.45813597924)/(o+1.)+(-0.592627725804)
        ref[3,1,2,0]=(-0.773854684679)/(o+1.)+(0.0423853650452)
        ref[3,1,2,1]=(1.2726274277)/(o+1.)+(0.683659642354)
        ref[3,2,0,0]=(0.555324507349)/(o+1.)+(0.116810003463)
        ref[3,2,0,1]=(-0.310977730161)/(o+1.)+(-0.437405129422)
        ref[3,2,1,0]=(1.05800939782)/(o+1.)+(0.0540888119933)
        ref[3,2,1,1]=(0.386222714751)/(o+1.)+(-0.739917239322)
        ref[3,2,2,0]=(-0.95705464593)/(o+1.)+(0.396944907549)
        ref[3,2,2,1]=(-0.491612682819)/(o+1.)+(-0.165439694033)
        ref[3,3,0,0]=(0.666972772584)/(o+1.)+(0.47603626707)
        ref[3,3,0,1]=(0.268506355312)/(o+1.)+(0.0336067603842)
        ref[3,3,1,0]=(-0.57724956178)/(o+1.)+(-0.409544073355)
        ref[3,3,1,1]=(-0.518808765862)/(o+1.)+(0.430852940555)
        ref[3,3,2,0]=(-1.95248352029)/(o+1.)+(0.734377777986)
        ref[3,3,2,1]=(-1.14565232021)/(o+1.)+(-0.67082922589)
        ref[3,4,0,0]=(-1.13455573259)/(o+1.)+(-0.1944138584)
        ref[3,4,0,1]=(-0.508914286468)/(o+1.)+(0.574401761354)
        ref[3,4,1,0]=(-0.319104038837)/(o+1.)+(-0.452063738718)
        ref[3,4,1,1]=(-0.100356733327)/(o+1.)+(-0.11327726566)
        ref[3,4,2,0]=(-0.39970881352)/(o+1.)+(-0.235116760127)
        ref[3,4,2,1]=(-0.339631070901)/(o+1.)+(-0.0957826012957)
      else:
        s[0,0,0,0]=(0.0794776522341)*x[0]+(1.69294456255)*x[1]+(0.824115844979)*x[2]
        s[0,0,0,1]=(0.0720523378919)*x[0]+(-1.39724154561)*x[1]+(-0.010988810956)*x[2]
        s[0,0,1,0]=(0.985556321125)*x[0]+(1.12133844031)*x[1]+(-0.262491219522)*x[2]
        s[0,0,1,1]=(0.789643681767)*x[0]+(1.02860624013)*x[1]+(-0.450770685703)*x[2]
        s[0,0,2,0]=(-0.850427026537)*x[0]+(-0.393993518297)*x[1]+(0.12816296306)*x[2]
        s[0,0,2,1]=(-0.47508130541)*x[0]+(-0.393857510976)*x[1]+(0.66003641476)*x[2]
        s[0,1,0,0]=(-1.87725252884)*x[0]+(-0.668075530728)*x[1]+(0.938485679482)*x[2]
        s[0,1,0,1]=(0.820537136967)*x[0]+(-0.453620529157)*x[1]+(0.811944595883)*x[2]
        s[0,1,1,0]=(0.914995161971)*x[0]+(-0.196153821444)*x[1]+(0.938555859255)*x[2]
        s[0,1,1,1]=(-0.866633718493)*x[0]+(-0.534596859321)*x[1]+(0.77514383239)*x[2]
        s[0,1,2,0]=(0.183094085342)*x[0]+(0.727499336912)*x[1]+(0.170986687969)*x[2]
        s[0,1,2,1]=(-1.36220673904)*x[0]+(0.0822122492483)*x[1]+(0.124227107716)*x[2]
        s[0,2,0,0]=(1.55269479425)*x[0]+(-0.14574857122)*x[1]+(1.24120289667)*x[2]
        s[0,2,0,1]=(-0.464038186768)*x[0]+(0.00554383028356)*x[1]+(-1.14838250985)*x[2]
        s[0,2,1,0]=(0.904608031363)*x[0]+(-0.313004420619)*x[1]+(-0.10529298073)*x[2]
        s[0,2,1,1]=(0.156623578469)*x[0]+(0.55127494399)*x[1]+(-0.385447926692)*x[2]
        s[0,2,2,0]=(-0.766804018933)*x[0]+(0.18519619356)*x[1]+(1.00915540081)*x[2]
        s[0,2,2,1]=(0.00231157548456)*x[0]+(0.338768642838)*x[1]+(-0.684228602438)*x[2]
        s[0,3,0,0]=(-0.834810486599)*x[0]+(-0.096360306458)*x[1]+(1.28924889761)*x[2]
        s[0,3,0,1]=(-0.658443843543)*x[0]+(0.889474267886)*x[1]+(0.566763124725)*x[2]
        s[0,3,1,0]=(-0.109777093963)*x[0]+(0.228053887631)*x[1]+(1.22952539906)*x[2]
        s[0,3,1,1]=(-0.749479260801)*x[0]+(-1.01102432725)*x[1]+(-0.196861989291)*x[2]
        s[0,3,2,0]=(0.650389523509)*x[0]+(0.151701374527)*x[1]+(0.506246388483)*x[2]
        s[0,3,2,1]=(-0.648510958338)*x[0]+(-0.263590202304)*x[1]+(-1.35992044414)*x[2]
        s[0,4,0,0]=(0.196976454523)*x[0]+(-0.193947226599)*x[1]+(1.67641586045)*x[2]
        s[0,4,0,1]=(-0.455868626908)*x[0]+(0.506302321882)*x[1]+(1.1666778393)*x[2]
        s[0,4,1,0]=(-0.788545609857)*x[0]+(-0.487332926053)*x[1]+(-0.0645141715605)*x[2]
        s[0,4,1,1]=(-0.320424709093)*x[0]+(-0.683301403769)*x[1]+(-0.222245108268)*x[2]
        s[0,4,2,0]=(0.876149348517)*x[0]+(0.12449602802)*x[1]+(-0.220531205966)*x[2]
        s[0,4,2,1]=(-0.2764172159)*x[0]+(-0.894571003479)*x[1]+(-0.595830070177)*x[2]
        s[1,0,0,0]=(-0.79676665323)*x[0]+(-1.25965739347)*x[1]+(-0.319474113915)*x[2]
        s[1,0,0,1]=(1.601348448)*x[0]+(0.41027863439)*x[1]+(-0.346758807915)*x[2]
        s[1,0,1,0]=(0.181292913253)*x[0]+(1.10494312205)*x[1]+(0.196973772813)*x[2]
        s[1,0,1,1]=(0.562861818795)*x[0]+(-0.997990211102)*x[1]+(0.0332706615838)*x[2]
        s[1,0,2,0]=(-0.989984164479)*x[0]+(-0.886186583204)*x[1]+(-0.228620900973)*x[2]
        s[1,0,2,1]=(0.841365917044)*x[0]+(-0.0519066039156)*x[1]+(-0.81126199301)*x[2]
        s[1,1,0,0]=(-0.613125126303)*x[0]+(0.482770874125)*x[1]+(0.038374960382)*x[2]
        s[1,1,0,1]=(-0.203719663632)*x[0]+(-0.281873286682)*x[1]+(-0.0303453795347)*x[2]
        s[1,1,1,0]=(1.11287699536)*x[0]+(-1.04759296757)*x[1]+(1.64257695374)*x[2]
        s[1,1,1,1]=(-0.0861307607908)*x[0]+(0.261314888791)*x[1]+(-0.204955973137)*x[2]
        s[1,1,2,0]=(0.821662110998)*x[0]+(-1.05649310988)*x[1]+(0.609822823353)*x[2]
        s[1,1,2,1]=(-0.730526005063)*x[0]+(1.62018363097)*x[1]+(0.99111776492)*x[2]
        s[1,2,0,0]=(-1.25039777751)*x[0]+(0.811070034681)*x[1]+(1.32505446058)*x[2]
        s[1,2,0,1]=(0.753110041895)*x[0]+(0.186756235207)*x[1]+(-1.30430176967)*x[2]
        s[1,2,1,0]=(0.146182105242)*x[0]+(-0.863831927779)*x[1]+(-1.1030247274)*x[2]
        s[1,2,1,1]=(0.674634709918)*x[0]+(-0.560866548824)*x[1]+(0.576408322189)*x[2]
        s[1,2,2,0]=(-0.573489753302)*x[0]+(1.57147013427)*x[1]+(0.711691292064)*x[2]
        s[1,2,2,1]=(0.549483527739)*x[0]+(-1.14179814973)*x[1]+(1.60524619335)*x[2]
        s[1,3,0,0]=(0.197453494754)*x[0]+(-0.456559966113)*x[1]+(-1.21423667747)*x[2]
        s[1,3,0,1]=(-1.76312274691)*x[0]+(-1.04778192886)*x[1]+(-0.185488490674)*x[2]
        s[1,3,1,0]=(0.0162460011376)*x[0]+(1.16407944993)*x[1]+(1.16424268292)*x[2]
        s[1,3,1,1]=(-0.513913288204)*x[0]+(-1.46536589962)*x[1]+(-0.906205233261)*x[2]
        s[1,3,2,0]=(-0.0157400074256)*x[0]+(-0.293812243081)*x[1]+(0.815280978304)*x[2]
        s[1,3,2,1]=(1.57630585875)*x[0]+(0.290739280065)*x[1]+(-0.771398356417)*x[2]
        s[1,4,0,0]=(-0.323882779703)*x[0]+(-0.763929395312)*x[1]+(0.738403272406)*x[2]
        s[1,4,0,1]=(-0.476278532043)*x[0]+(-0.11039317349)*x[1]+(0.346684885005)*x[2]
        s[1,4,1,0]=(-1.36304063164)*x[0]+(-0.660619165296)*x[1]+(-0.903566907206)*x[2]
        s[1,4,1,1]=(0.0863780283437)*x[0]+(0.466792514558)*x[1]+(0.448776109878)*x[2]
        s[1,4,2,0]=(0.74478619068)*x[0]+(0.446314847678)*x[1]+(-0.0481741032015)*x[2]
        s[1,4,2,1]=(0.746506715398)*x[0]+(-0.432201202785)*x[1]+(0.203440105689)*x[2]
        s[2,0,0,0]=(-0.785805880678)*x[0]+(-0.862407720473)*x[1]+(-1.21640925829)*x[2]
        s[2,0,0,1]=(0.647190817984)*x[0]+(0.519851314653)*x[1]+(0.390946186931)*x[2]
        s[2,0,1,0]=(-1.24831725337)*x[0]+(1.59413938467)*x[1]+(-1.33358250136)*x[2]
        s[2,0,1,1]=(-0.152213072763)*x[0]+(-1.09357954369)*x[1]+(-1.25586196425)*x[2]
        s[2,0,2,0]=(-0.0314861108304)*x[0]+(0.293907529511)*x[1]+(-0.21039874768)*x[2]
        s[2,0,2,1]=(0.0467194832302)*x[0]+(0.527286460462)*x[1]+(-0.415618293894)*x[2]
        s[2,1,0,0]=(-0.189575051656)*x[0]+(-0.844885847783)*x[1]+(-0.334235875002)*x[2]
        s[2,1,0,1]=(-0.00822711174783)*x[0]+(-1.5897272285)*x[1]+(1.50002681163)*x[2]
        s[2,1,1,0]=(-0.841981498104)*x[0]+(-1.10376539971)*x[1]+(0.311385102139)*x[2]
        s[2,1,1,1]=(1.12897934543)*x[0]+(-0.966261308872)*x[1]+(-1.16359871767)*x[2]
        s[2,1,2,0]=(-0.957189463062)*x[0]+(0.0471569985184)*x[1]+(0.12392515517)*x[2]
        s[2,1,2,1]=(-0.752369087888)*x[0]+(1.18176351896)*x[1]+(0.334788137458)*x[2]
        s[2,2,0,0]=(1.22014205343)*x[0]+(0.784413576281)*x[1]+(0.0962245348039)*x[2]
        s[2,2,0,1]=(-0.407023035962)*x[0]+(1.11597107325)*x[1]+(0.458612291763)*x[2]
        s[2,2,1,0]=(-0.0496175828387)*x[0]+(-0.0536216340718)*x[1]+(0.771683970464)*x[2]
        s[2,2,1,1]=(-0.356958793645)*x[0]+(-1.38580705404)*x[1]+(-0.670167616875)*x[2]
        s[2,2,2,0]=(0.0804001816041)*x[0]+(-1.16114869406)*x[1]+(0.475472363874)*x[2]
        s[2,2,2,1]=(0.98238551642)*x[0]+(-1.12578358118)*x[1]+(-0.00970359156434)*x[2]
        s[2,3,0,0]=(1.0110556058)*x[0]+(-0.488112488916)*x[1]+(0.243005554591)*x[2]
        s[2,3,0,1]=(-0.16668051279)*x[0]+(0.178067528024)*x[1]+(0.969038052731)*x[2]
        s[2,3,1,0]=(-0.00345650882131)*x[0]+(-0.941671547756)*x[1]+(-0.49908897105)*x[2]
        s[2,3,1,1]=(0.716683430797)*x[0]+(-0.585708024289)*x[1]+(-1.19073441986)*x[2]
        s[2,3,2,0]=(0.400515369791)*x[0]+(-0.454556618814)*x[1]+(0.0782324352019)*x[2]
        s[2,3,2,1]=(-0.164628093414)*x[0]+(-1.26464027792)*x[1]+(0.179511210919)*x[2]
        s[2,4,0,0]=(1.33277550072)*x[0]+(1.18570575409)*x[1]+(0.381346013717)*x[2]
        s[2,4,0,1]=(0.588290284687)*x[0]+(-1.08302445047)*x[1]+(0.166898904485)*x[2]
        s[2,4,1,0]=(0.131178111015)*x[0]+(-0.749855470189)*x[1]+(-0.1247474719)*x[2]
        s[2,4,1,1]=(-0.97245336122)*x[0]+(0.873932906896)*x[1]+(0.120264062027)*x[2]
        s[2,4,2,0]=(-0.722531128456)*x[0]+(0.19636289816)*x[1]+(-0.907945347107)*x[2]
        s[2,4,2,1]=(-0.711111750197)*x[0]+(0.31947094036)*x[1]+(-1.79333697163)*x[2]
        s[3,0,0,0]=(-0.184519023765)*x[0]+(0.680221159423)*x[1]+(-0.0244123658903)*x[2]
        s[3,0,0,1]=(-1.14573637546)*x[0]+(0.0616977700293)*x[1]+(-0.654808111844)*x[2]
        s[3,0,1,0]=(-1.09297543018)*x[0]+(-1.04184353717)*x[1]+(0.384120100538)*x[2]
        s[3,0,1,1]=(0.105247846819)*x[0]+(-1.31600150124)*x[1]+(1.43496073616)*x[2]
        s[3,0,2,0]=(0.881148598911)*x[0]+(-0.478723216871)*x[1]+(-0.0746344543132)*x[2]
        s[3,0,2,1]=(-1.20828681587)*x[0]+(-1.65808839437)*x[1]+(-0.636678098535)*x[2]
        s[3,1,0,0]=(-0.213394309668)*x[0]+(-0.240354748844)*x[1]+(0.117282970374)*x[2]
        s[3,1,0,1]=(1.37462845901)*x[0]+(-0.41583894581)*x[1]+(-0.625626047701)*x[2]
        s[3,1,1,0]=(1.83991491983)*x[0]+(1.2778674571)*x[1]+(1.35465906139)*x[2]
        s[3,1,1,1]=(1.58168546715)*x[0]+(0.829809996131)*x[1]+(0.681232260732)*x[2]
        s[3,1,2,0]=(-0.0212907664121)*x[0]+(1.00109823691)*x[1]+(-1.30845432825)*x[2]
        s[3,1,2,1]=(0.169920360674)*x[0]+(-0.864995850094)*x[1]+(0.302979947072)*x[2]
        s[3,2,0,0]=(-0.154084636739)*x[0]+(0.319647474298)*x[1]+(-1.23506947259)*x[2]
        s[3,2,0,1]=(-0.199216213818)*x[0]+(1.28921202953)*x[1]+(1.65805122541)*x[2]
        s[3,2,1,0]=(-1.86774194364)*x[0]+(-0.101723263166)*x[1]+(-0.526029332636)*x[2]
        s[3,2,1,1]=(-0.177960964857)*x[0]+(-0.899073564225)*x[1]+(0.551378002191)*x[2]
        s[3,2,2,0]=(0.18491319197)*x[0]+(-0.286906366286)*x[1]+(0.583859098776)*x[2]
        s[3,2,2,1]=(0.38623837272)*x[0]+(-0.653272502002)*x[1]+(-1.46718517275)*x[2]
        s[3,3,0,0]=(0.0494572779834)*x[0]+(-0.0566099666381)*x[1]+(-0.113874384717)*x[2]
        s[3,3,0,1]=(0.0241152098269)*x[0]+(-0.317061594612)*x[1]+(1.23769551101)*x[2]
        s[3,3,1,0]=(-0.107294616786)*x[0]+(-0.651410283508)*x[1]+(0.833104181804)*x[2]
        s[3,3,1,1]=(0.16676828757)*x[0]+(-0.670496617303)*x[1]+(1.15686719798)*x[2]
        s[3,3,2,0]=(0.159950143242)*x[0]+(0.0436823700228)*x[1]+(-0.645027680285)*x[2]
        s[3,3,2,1]=(-1.2697347822)*x[0]+(1.22938250997)*x[1]+(0.208587928792)*x[2]
        s[3,4,0,0]=(0.575105413412)*x[0]+(-1.44998456404)*x[1]+(-0.516125694177)*x[2]
        s[3,4,0,1]=(0.0387618783003)*x[0]+(0.859152861676)*x[1]+(1.29847202073)*x[2]
        s[3,4,1,0]=(-0.842510438531)*x[0]+(0.191878858518)*x[1]+(-0.73453059399)*x[2]
        s[3,4,1,1]=(1.41435810392)*x[0]+(-0.855734345577)*x[1]+(-0.425917153945)*x[2]
        s[3,4,2,0]=(-0.58228358663)*x[0]+(0.00936548588525)*x[1]+(-1.38536320116)*x[2]
        s[3,4,2,1]=(0.0419678913201)*x[0]+(1.02999559903)*x[1]+(-0.464745184871)*x[2]
        ref[0,0,0,0]=(1.30172410662)/(o+1.)+(0.647406976573)
        ref[0,0,0,1]=(-1.09442521373)/(o+1.)+(-0.120876402476)
        ref[0,0,1,0]=(2.01757642043)/(o+1.)+(-0.0865864392547)
        ref[0,0,1,1]=(1.80034002198)/(o+1.)+(-0.216430392892)
        ref[0,0,2,0]=(1.13142112916)/(o+1.)+(-1.12383935547)
        ref[0,0,2,1]=(-0.905515986966)/(o+1.)+(0.34830679267)
        ref[0,1,0,0]=(-0.210441134504)/(o+1.)+(-0.698200622791)
        ref[0,1,0,1]=(0.360807961621)/(o+1.)+(0.409026621036)
        ref[0,1,1,0]=(1.5147969883)/(o+1.)+(0.0713001057429)
        ref[0,1,1,1]=(0.396123567843)/(o+1.)+(-0.511105156634)
        ref[0,1,2,0]=(0.184468760768)/(o+1.)+(0.448555674728)
        ref[0,1,2,1]=(-0.133892978353)/(o+1.)+(-0.510937201862)
        ref[0,2,0,0]=(1.66137009462)/(o+1.)+(0.493389512541)
        ref[0,2,0,1]=(0.358452769174)/(o+1.)+(-0.982664817753)
        ref[0,2,1,0]=(0.966434135345)/(o+1.)+(-0.240061752665)
        ref[0,2,1,1]=(-0.401753883503)/(o+1.)+(0.362102239635)
        ref[0,2,2,0]=(-0.0153619747046)/(o+1.)+(0.221454775071)
        ref[0,2,2,1]=(-0.104138117282)/(o+1.)+(-0.119505133417)
        ref[0,3,0,0]=(-0.966204569194)/(o+1.)+(0.662141336872)
        ref[0,3,0,1]=(0.587661220327)/(o+1.)+(0.10506616437)
        ref[0,3,1,0]=(2.02288199931)/(o+1.)+(-0.337539903289)
        ref[0,3,1,1]=(-1.78842825396)/(o+1.)+(-0.0844686616883)
        ref[0,3,2,0]=(0.927282266945)/(o+1.)+(0.190527509786)
        ref[0,3,2,1]=(-0.679518039292)/(o+1.)+(-0.796251782745)
        ref[0,4,0,0]=(0.555388124049)/(o+1.)+(0.562028482163)
        ref[0,4,0,1]=(0.0870397234082)/(o+1.)+(0.565035905435)
        ref[0,4,1,0]=(-1.1463502212)/(o+1.)+(-0.0970212431374)
        ref[0,4,1,1]=(0.233233715042)/(o+1.)+(-0.729602468086)
        ref[0,4,2,0]=(-0.782889963002)/(o+1.)+(0.781502066786)
        ref[0,4,2,1]=(0.763433249649)/(o+1.)+(-1.2651257696)
        ref[1,0,0,0]=(-1.00441831898)/(o+1.)+(-0.685739920821)
        ref[1,0,0,1]=(0.751247546824)/(o+1.)+(0.456810363824)
        ref[1,0,1,0]=(2.22764882334)/(o+1.)+(-0.372219507613)
        ref[1,0,1,1]=(-1.19623631804)/(o+1.)+(0.39718929366)
        ref[1,0,2,0]=(-1.60343741738)/(o+1.)+(-0.250677115636)
        ref[1,0,2,1]=(-1.25680789752)/(o+1.)+(0.61750260882)
        ref[1,1,0,0]=(-1.45519189991)/(o+1.)+(0.681606304054)
        ref[1,1,0,1]=(0.840628495308)/(o+1.)+(-0.678283412578)
        ref[1,1,1,0]=(1.41826768667)/(o+1.)+(0.144796647431)
        ref[1,1,1,1]=(-0.933723251106)/(o+1.)+(0.451975702985)
        ref[1,1,2,0]=(-0.408324994487)/(o+1.)+(0.39165840948)
        ref[1,1,2,1]=(1.23290517672)/(o+1.)+(0.323935107053)
        ref[1,2,0,0]=(0.825483086179)/(o+1.)+(0.0301218157877)
        ref[1,2,0,1]=(-0.0793934768662)/(o+1.)+(-0.142521007854)
        ref[1,2,1,0]=(-0.0709650957035)/(o+1.)+(-0.874854727116)
        ref[1,2,1,1]=(1.10714952407)/(o+1.)+(-0.208486520392)
        ref[1,2,2,0]=(0.99013461187)/(o+1.)+(0.359768530582)
        ref[1,2,2,1]=(0.0356627876162)/(o+1.)+(0.488634391872)
        ref[1,3,0,0]=(-0.474442326803)/(o+1.)+(-0.499450411015)
        ref[1,3,0,1]=(-0.673232992552)/(o+1.)+(-1.16158008694)
        ref[1,3,1,0]=(0.411652221393)/(o+1.)+(0.966457956295)
        ref[1,3,1,1]=(-1.76698398732)/(o+1.)+(-0.559250216883)
        ref[1,3,2,0]=(1.9874048718)/(o+1.)+(-0.740838072)
        ref[1,3,2,1]=(0.661784629796)/(o+1.)+(0.216931076303)
        ref[1,4,0,0]=(-0.513029515019)/(o+1.)+(0.0818103062045)
        ref[1,4,0,1]=(0.282755142694)/(o+1.)+(-0.261370981611)
        ref[1,4,1,0]=(-1.96973437805)/(o+1.)+(-0.478746163042)
        ref[1,4,1,1]=(0.908774952181)/(o+1.)+(0.0465858502998)
        ref[1,4,2,0]=(1.3422757017)/(o+1.)+(-0.0996743832723)
        ref[1,4,2,1]=(1.21609351443)/(o+1.)+(-0.349173948065)
        ref[2,0,0,0]=(-1.01744592956)/(o+1.)+(-0.923588464944)
        ref[2,0,0,1]=(1.10931736222)/(o+1.)+(0.224335478673)
        ref[2,0,1,0]=(-0.416479828321)/(o+1.)+(-0.285640270869)
        ref[2,0,1,1]=(-1.5671174369)/(o+1.)+(-0.467268571897)
        ref[2,0,2,0]=(-0.274214960025)/(o+1.)+(0.163118815513)
        ref[2,0,2,1]=(0.000680528469817)/(o+1.)+(0.0788535606642)
        ref[2,1,0,0]=(-0.844192475139)/(o+1.)+(-0.262252149651)
        ref[2,1,0,1]=(0.69399848864)/(o+1.)+(-0.395963008628)
        ref[2,1,1,0]=(-1.37742489692)/(o+1.)+(-0.128468449379)
        ref[2,1,1,1]=(0.282095133276)/(o+1.)+(-0.641487907195)
        ref[2,1,2,0]=(-1.4727375549)/(o+1.)+(0.343315122764)
        ref[2,1,2,1]=(0.67637649169)/(o+1.)+(0.0439030384195)
        ref[2,2,0,0]=(1.58598306034)/(o+1.)+(0.257398552088)
        ref[2,2,0,1]=(-0.187654980952)/(o+1.)+(0.677607655001)
        ref[2,2,1,0]=(0.225457027395)/(o+1.)+(0.221493863079)
        ref[2,2,1,1]=(-1.66885547416)/(o+1.)+(-0.372038995202)
        ref[2,2,2,0]=(-1.42708510165)/(o+1.)+(0.410904476533)
        ref[2,2,2,1]=(0.352696956576)/(o+1.)+(-0.252899306452)
        ref[2,3,0,0]=(0.266365999523)/(o+1.)+(0.249791335977)
        ref[2,3,0,1]=(0.926084735451)/(o+1.)+(0.0271701662571)
        ref[2,3,1,0]=(-0.138787407751)/(o+1.)+(-0.652714809939)
        ref[2,3,1,1]=(-0.834750293031)/(o+1.)+(-0.112504360161)
        ref[2,3,2,0]=(-0.806210182586)/(o+1.)+(0.415200684382)
        ref[2,3,2,1]=(0.577899253552)/(o+1.)+(-0.913828206984)
        ref[2,4,0,0]=(1.38802720612)/(o+1.)+(0.755900031203)
        ref[2,4,0,1]=(-1.33634216788)/(o+1.)+(0.504253453293)
        ref[2,4,1,0]=(-1.13393506986)/(o+1.)+(0.195255119392)
        ref[2,4,1,1]=(-0.384732222354)/(o+1.)+(0.203237915028)
        ref[2,4,2,0]=(-1.18866097183)/(o+1.)+(-0.122726302784)
        ref[2,4,2,1]=(-1.41914128138)/(o+1.)+(-0.382918250045)
        ref[3,0,0,0]=(-1.3617379499)/(o+1.)+(0.916513859836)
        ref[3,0,0,1]=(-0.577580679368)/(o+1.)+(-0.580633018955)
        ref[3,0,1,0]=(-0.730926522369)/(o+1.)+(-0.50988617222)
        ref[3,0,1,1]=(0.852352706539)/(o+1.)+(-0.314072812402)
        ref[3,0,2,0]=(0.63914695576)/(o+1.)+(-0.155678014016)
        ref[3,0,2,1]=(-2.31555597527)/(o+1.)+(-0.593748666755)
        ref[3,1,0,0]=(-0.31308710102)/(o+1.)+(-0.011689493559)
        ref[3,1,0,1]=(0.565049631081)/(o+1.)+(-0.115943082793)
        ref[3,1,1,0]=(1.76264026827)/(o+1.)+(1.35490058502)
        ref[3,1,1,1]=(1.34557111025)/(o+1.)+(0.873578306883)
        ref[3,1,2,0]=(0.924587947935)/(o+1.)+(-0.626617402844)
        ref[3,1,2,1]=(-0.679695689354)/(o+1.)+(0.143800073503)
        ref[3,2,0,0]=(-0.408824291494)/(o+1.)+(-0.330341171766)
        ref[3,2,0,1]=(1.02728720579)/(o+1.)+(0.860379917667)
        ref[3,2,1,0]=(-1.99227170213)/(o+1.)+(-0.251611418656)
        ref[3,2,1,1]=(-0.844161339595)/(o+1.)+(0.159252406352)
        ref[3,2,2,0]=(1.52009831693)/(o+1.)+(-0.519116196235)
        ref[3,2,2,1]=(0.323553722724)/(o+1.)+(-1.02888651238)
        ref[3,3,0,0]=(-1.39647666034)/(o+1.)+(0.637724793483)
        ref[3,3,0,1]=(0.960901654854)/(o+1.)+(-0.00807626431577)
        ref[3,3,1,0]=(0.383998350003)/(o+1.)+(-0.154799534246)
        ref[3,3,1,1]=(-0.851062083214)/(o+1.)+(0.752100475729)
        ref[3,3,2,0]=(0.788418151505)/(o+1.)+(-0.614906659263)
        ref[3,3,2,1]=(-0.1028922667)/(o+1.)+(0.135563961629)
        ref[3,4,0,0]=(-0.540240088251)/(o+1.)+(-0.425382378277)
        ref[3,4,0,1]=(0.236730341461)/(o+1.)+(0.979828209623)
        ref[3,4,1,0]=(-0.9880898216)/(o+1.)+(-0.198536176202)
        ref[3,4,1,1]=(1.06221574228)/(o+1.)+(-0.46475456894)
        ref[3,4,2,0]=(-1.89471538439)/(o+1.)+(-0.0317829587581)
        ref[3,4,2,1]=(-1.20699948566)/(o+1.)+(0.907108895566)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.518496555195)*x[0]+(-1.7081835316)*x[1]
        ref=(-0.587946015314)/(o+1.)+(-0.300870480545)
      else:
        s=(0.271043713323)*x[0]+(-1.57101724485)*x[1]+(-0.650553623396)*x[2]
        ref=(-1.53180577827)/(o+1.)+(-0.209360688325)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.228569380378)*x[0]+(0.808739010281)*x[1]
        s[1]=(0.658420467565)*x[0]+(0.755018594164)*x[1]
        ref[0]=(0.635537444802)/(o+1.)+(-0.0276839074492)
        ref[1]=(0.431019006365)/(o+1.)+(0.491210027682)
      else:
        s[0]=(0.376210963228)*x[0]+(-1.32850082986)*x[1]+(0.536274160408)*x[2]
        s[1]=(-0.771050457843)*x[0]+(-0.16954629666)*x[1]+(-0.579642261351)*x[2]
        ref[0]=(-0.923045104748)/(o+1.)+(0.253514699262)
        ref[1]=(0.10262370672)/(o+1.)+(-0.811431361287)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(-0.405824989536)*x[0]+(-1.36105849061)*x[1]
        s[0,1]=(0.759504921251)*x[0]+(0.455515132277)*x[1]
        s[0,2]=(-1.37469026888)*x[0]+(1.35869839203)*x[1]
        s[0,3]=(-0.0438580623198)*x[0]+(0.34881551198)*x[1]
        s[0,4]=(-0.0097850380672)*x[0]+(0.61199301527)*x[1]
        s[1,0]=(-1.24950855151)*x[0]+(0.560604227152)*x[1]
        s[1,1]=(0.897125563557)*x[0]+(-0.362799426566)*x[1]
        s[1,2]=(-0.475265733496)*x[0]+(-0.906032283495)*x[1]
        s[1,3]=(0.184749910708)*x[0]+(0.961807058286)*x[1]
        s[1,4]=(-1.8130820241)*x[0]+(-0.0716801569248)*x[1]
        s[2,0]=(0.401163196364)*x[0]+(-0.122822662117)*x[1]
        s[2,1]=(-0.183141841931)*x[0]+(0.159646202564)*x[1]
        s[2,2]=(0.494070531254)*x[0]+(0.0283267498222)*x[1]
        s[2,3]=(-0.46326385119)*x[0]+(0.416856391877)*x[1]
        s[2,4]=(0.669972632028)*x[0]+(1.01207764348)*x[1]
        s[3,0]=(0.579509485187)*x[0]+(0.578664253542)*x[1]
        s[3,1]=(1.1904242989)*x[0]+(-0.617672917808)*x[1]
        s[3,2]=(1.01069090914)*x[0]+(-0.132861271259)*x[1]
        s[3,3]=(-1.02014113453)*x[0]+(-1.10411835631)*x[1]
        s[3,4]=(-0.178208550651)*x[0]+(-0.590646644191)*x[1]
        ref[0,0]=(-1.81750521429)/(o+1.)+(0.0253108670731)
        ref[0,1]=(1.13215741955)/(o+1.)+(0.0414313169864)
        ref[0,2]=(-0.346109803394)/(o+1.)+(0.165058963273)
        ref[0,3]=(1.16722378378)/(o+1.)+(-0.431133167061)
        ref[0,4]=(1.39150379152)/(o+1.)+(-0.39464790716)
        ref[1,0]=(-0.249819369301)/(o+1.)+(-0.219542477529)
        ref[1,1]=(-0.253832127476)/(o+1.)+(0.394079132234)
        ref[1,2]=(-0.514583575657)/(o+1.)+(-0.433357220667)
        ref[1,3]=(1.16430206364)/(o+1.)+(-0.00887254732364)
        ref[1,4]=(-1.40353822359)/(o+1.)+(-0.240611978717)
        ref[2,0]=(0.0129701699159)/(o+1.)+(0.132685182166)
        ref[2,1]=(-0.000259682696025)/(o+1.)+(-0.0116179783357)
        ref[2,2]=(-0.736785877312)/(o+1.)+(0.629591579194)
        ref[2,3]=(-0.319819594662)/(o+1.)+(0.136706067675)
        ref[2,4]=(1.34735651213)/(o+1.)+(0.167346881689)
        ref[3,0]=(1.02891998838)/(o+1.)+(0.0646268751734)
        ref[3,1]=(0.0226437097602)/(o+1.)+(0.275053835665)
        ref[3,2]=(-0.367947454589)/(o+1.)+(0.622888546236)
        ref[3,3]=(-1.07416112237)/(o+1.)+(-0.525049184236)
        ref[3,4]=(-0.423224656448)/(o+1.)+(-0.172815269197)
      else:
        s[0,0]=(0.391957465543)*x[0]+(-0.836212796005)*x[1]+(0.541170293661)*x[2]
        s[0,1]=(-0.324966494738)*x[0]+(0.62718596172)*x[1]+(-0.00110999366491)*x[2]
        s[0,2]=(0.377719216476)*x[0]+(1.20630293683)*x[1]+(-0.936497207115)*x[2]
        s[0,3]=(-1.52853135673)*x[0]+(-0.634683048191)*x[1]+(-0.533312108123)*x[2]
        s[0,4]=(-0.659359670617)*x[0]+(-1.23995360412)*x[1]+(0.669989279254)*x[2]
        s[1,0]=(-0.45120395504)*x[0]+(1.25119902426)*x[1]+(0.11139465291)*x[2]
        s[1,1]=(0.431699115955)*x[0]+(1.165662618)*x[1]+(-1.12543997002)*x[2]
        s[1,2]=(0.16640215316)*x[0]+(-1.81838137531)*x[1]+(-0.111705546449)*x[2]
        s[1,3]=(-1.56135929502)*x[0]+(0.425946749357)*x[1]+(0.255013889109)*x[2]
        s[1,4]=(0.320981318781)*x[0]+(-0.536492696361)*x[1]+(1.21768285382)*x[2]
        s[2,0]=(-0.600249234632)*x[0]+(-0.497626140165)*x[1]+(-0.0606420495108)*x[2]
        s[2,1]=(-0.144412727094)*x[0]+(-1.09139116844)*x[1]+(-0.39344520166)*x[2]
        s[2,2]=(-0.433006957898)*x[0]+(-0.287282800253)*x[1]+(8.69987670264e-05)*x[2]
        s[2,3]=(0.805491532167)*x[0]+(-1.17746983966)*x[1]+(0.720861745039)*x[2]
        s[2,4]=(-0.735630820744)*x[0]+(-1.01122803474)*x[1]+(1.72681276559)*x[2]
        s[3,0]=(-0.739861204924)*x[0]+(0.739077496919)*x[1]+(-0.0950784436747)*x[2]
        s[3,1]=(0.428369405071)*x[0]+(-0.0613035842349)*x[1]+(1.22804321951)*x[2]
        s[3,2]=(0.617885252444)*x[0]+(-1.63428007618)*x[1]+(-0.543704110813)*x[2]
        s[3,3]=(-1.52397266949)*x[0]+(-1.33821038473)*x[1]+(0.578214428096)*x[2]
        s[3,4]=(-0.904835367377)*x[0]+(0.156039380736)*x[1]+(-0.833632978592)*x[2]
        ref[0,0]=(-0.865179337718)/(o+1.)+(0.481047150458)
        ref[0,1]=(-0.427981609749)/(o+1.)+(0.364545541533)
        ref[0,2]=(0.738756225478)/(o+1.)+(-0.0456156396413)
        ref[0,3]=(-2.19753999815)/(o+1.)+(-0.249493257444)
        ref[0,4]=(-1.44493155356)/(o+1.)+(0.107803779038)
        ref[1,0]=(1.72841073711)/(o+1.)+(-0.408510507493)
        ref[1,1]=(0.688403407203)/(o+1.)+(-0.108240821633)
        ref[1,2]=(-0.0562832360974)/(o+1.)+(-0.853700766251)
        ref[1,3]=(-0.866037983384)/(o+1.)+(-0.00718033658349)
        ref[1,4]=(1.27849436197)/(o+1.)+(-0.138161442865)
        ref[2,0]=(-0.450814686687)/(o+1.)+(-0.353851368811)
        ref[2,1]=(-0.608210959793)/(o+1.)+(-0.510519068702)
        ref[2,2]=(-1.55008808776)/(o+1.)+(0.414942664188)
        ref[2,3]=(0.242744047944)/(o+1.)+(0.0530696948023)
        ref[2,4]=(-0.793195276992)/(o+1.)+(0.386574593552)
        ref[3,0]=(-0.33646344984)/(o+1.)+(0.12030064908)
        ref[3,1]=(1.37599194013)/(o+1.)+(0.109558550106)
        ref[3,2]=(0.0175612882637)/(o+1.)+(-0.788830111409)
        ref[3,3]=(-0.747828767237)/(o+1.)+(-0.768069929445)
        ref[3,4]=(-0.876520699909)/(o+1.)+(-0.352954132662)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.64806859919)*x[0]+(0.669871100749)*x[1]
        s[0,0,1]=(0.351912095781)*x[0]+(-0.795221209702)*x[1]
        s[0,1,0]=(1.67848694486)*x[0]+(0.294895162463)*x[1]
        s[0,1,1]=(0.360904647807)*x[0]+(1.08975889449)*x[1]
        s[1,0,0]=(-0.0924096394032)*x[0]+(-1.45324202762)*x[1]
        s[1,0,1]=(-0.160994756309)*x[0]+(0.551455091958)*x[1]
        s[1,1,0]=(1.32478072001)*x[0]+(-1.43998718605)*x[1]
        s[1,1,1]=(0.481907947963)*x[0]+(-1.62005808161)*x[1]
        s[2,0,0]=(1.28445085089)*x[0]+(0.139212942532)*x[1]
        s[2,0,1]=(0.13496510335)*x[0]+(-0.3090345869)*x[1]
        s[2,1,0]=(1.34702478342)*x[0]+(0.77762332551)*x[1]
        s[2,1,1]=(0.2270682866)*x[0]+(0.585652373115)*x[1]
        s[3,0,0]=(0.974462828497)*x[0]+(-0.0761666893839)*x[1]
        s[3,0,1]=(0.787654033315)*x[0]+(-0.665952609368)*x[1]
        s[3,1,0]=(-0.535337193665)*x[0]+(-0.10003612917)*x[1]
        s[3,1,1]=(0.90669722873)*x[0]+(0.0637346383222)*x[1]
        s[4,0,0]=(0.224906594767)*x[0]+(0.255067891624)*x[1]
        s[4,0,1]=(1.45652552317)*x[0]+(-0.896697311539)*x[1]
        s[4,1,0]=(0.0111655859631)*x[0]+(1.35361034898)*x[1]
        s[4,1,1]=(-1.49696987574)*x[0]+(-0.0630120006374)*x[1]
        s[5,0,0]=(-1.37528487034)*x[0]+(-1.03082532988)*x[1]
        s[5,0,1]=(0.298794708742)*x[0]+(-0.338420113242)*x[1]
        s[5,1,0]=(0.101365713196)*x[0]+(0.46239070902)*x[1]
        s[5,1,1]=(-0.329260078889)*x[0]+(0.0114285623122)*x[1]
        ref[0,0,0]=(1.04954961047)/(o+1.)+(0.134195044736)
        ref[0,0,1]=(-0.941346926582)/(o+1.)+(0.24901890633)
        ref[0,1,0]=(0.33540890863)/(o+1.)+(0.818986599345)
        ref[0,1,1]=(-0.0452512610385)/(o+1.)+(0.747957401666)
        ref[1,0,0]=(-0.0571552186929)/(o+1.)+(-0.744248224165)
        ref[1,0,1]=(-0.256660001756)/(o+1.)+(0.323560168702)
        ref[1,1,0]=(0.173031418514)/(o+1.)+(-0.144118942277)
        ref[1,1,1]=(-1.27331535296)/(o+1.)+(0.0675826096564)
        ref[2,0,0]=(1.27442931542)/(o+1.)+(0.0746172389969)
        ref[2,0,1]=(1.23134060492)/(o+1.)+(-0.702705044233)
        ref[2,1,0]=(1.44361222924)/(o+1.)+(0.340517939842)
        ref[2,1,1]=(0.858404770157)/(o+1.)+(-0.0228420552212)
        ref[3,0,0]=(0.53268376719)/(o+1.)+(0.182806185962)
        ref[3,0,1]=(0.962024080252)/(o+1.)+(-0.420161328153)
        ref[3,1,0]=(-1.06376900656)/(o+1.)+(0.214197841863)
        ref[3,1,1]=(0.437728927482)/(o+1.)+(0.266351469785)
        ref[4,0,0]=(-0.247848566495)/(o+1.)+(0.363911526443)
        ref[4,0,1]=(0.275565249528)/(o+1.)+(0.142131481049)
        ref[4,1,0]=(0.8446292547)/(o+1.)+(0.260073340122)
        ref[4,1,1]=(-1.92408551747)/(o+1.)+(0.182051820549)
        ref[5,0,0]=(-1.70867649172)/(o+1.)+(-0.348716854247)
        ref[5,0,1]=(-0.216851661682)/(o+1.)+(0.088613128591)
        ref[5,1,0]=(1.89734425595)/(o+1.)+(-0.666793916865)
        ref[5,1,1]=(-0.281093101446)/(o+1.)+(-0.0183692075654)
      else:
        s[0,0,0]=(0.142896594259)*x[0]+(-0.458264763854)*x[1]+(0.521893125785)*x[2]
        s[0,0,1]=(-0.141846975275)*x[0]+(-0.976845409768)*x[1]+(-1.14147520898)*x[2]
        s[0,1,0]=(0.141747271548)*x[0]+(-1.21811158828)*x[1]+(0.0640715930615)*x[2]
        s[0,1,1]=(0.462551385436)*x[0]+(-0.87929813895)*x[1]+(0.944848279649)*x[2]
        s[1,0,0]=(0.246197516828)*x[0]+(-0.0275751890238)*x[1]+(0.278318853806)*x[2]
        s[1,0,1]=(0.525663042785)*x[0]+(0.191011208259)*x[1]+(-1.27931005652)*x[2]
        s[1,1,0]=(-1.12276137904)*x[0]+(0.474659469962)*x[1]+(0.15450537857)*x[2]
        s[1,1,1]=(0.321959890457)*x[0]+(-0.0325903104321)*x[1]+(-0.735505871467)*x[2]
        s[2,0,0]=(-0.190194685931)*x[0]+(-0.277506154989)*x[1]+(0.0984734800625)*x[2]
        s[2,0,1]=(0.286493694398)*x[0]+(0.841559451892)*x[1]+(0.20572294365)*x[2]
        s[2,1,0]=(-0.0132570386632)*x[0]+(-0.160857995452)*x[1]+(-0.813758770212)*x[2]
        s[2,1,1]=(0.00519333657271)*x[0]+(0.0324836424574)*x[1]+(0.493299621992)*x[2]
        s[3,0,0]=(0.300624746821)*x[0]+(0.211207146953)*x[1]+(1.0860831317)*x[2]
        s[3,0,1]=(-0.696336553938)*x[0]+(-1.28551896118)*x[1]+(0.738769481443)*x[2]
        s[3,1,0]=(0.15510282832)*x[0]+(-1.15795290548)*x[1]+(0.543235378485)*x[2]
        s[3,1,1]=(-0.714308237763)*x[0]+(0.239072001232)*x[1]+(-0.187168055861)*x[2]
        s[4,0,0]=(0.624611061079)*x[0]+(1.11568209096)*x[1]+(-1.46683771155)*x[2]
        s[4,0,1]=(1.45301860552)*x[0]+(0.74921646591)*x[1]+(-0.796630526412)*x[2]
        s[4,1,0]=(0.560555339207)*x[0]+(-0.421267308186)*x[1]+(-0.222312671328)*x[2]
        s[4,1,1]=(0.807845684562)*x[0]+(0.323266951017)*x[1]+(1.51940950039)*x[2]
        s[5,0,0]=(-0.558067921385)*x[0]+(-0.214928793546)*x[1]+(1.01402046886)*x[2]
        s[5,0,1]=(1.20688436616)*x[0]+(0.0247589671447)*x[1]+(-0.368536410815)*x[2]
        s[5,1,0]=(-0.324541665933)*x[0]+(-0.205515389215)*x[1]+(0.758278431203)*x[2]
        s[5,1,1]=(-0.620405057772)*x[0]+(-0.67647528448)*x[1]+(1.32328947052)*x[2]
        ref[0,0,0]=(0.260018146786)/(o+1.)+(-0.0267465952979)
        ref[0,0,1]=(-0.435502524101)/(o+1.)+(-0.912332534962)
        ref[0,1,0]=(-0.0330362836983)/(o+1.)+(-0.489628219987)
        ref[0,1,1]=(1.07484163998)/(o+1.)+(-0.273370056924)
        ref[1,0,0]=(-0.523779853488)/(o+1.)+(0.510360517549)
        ref[1,0,1]=(0.448972814496)/(o+1.)+(-0.505804309986)
        ref[1,1,0]=(-1.09840793106)/(o+1.)+(0.302405700276)
        ref[1,1,1]=(0.130511241273)/(o+1.)+(-0.288323766357)
        ref[2,0,0]=(-0.805299870557)/(o+1.)+(0.21803625485)
        ref[2,0,1]=(0.271181617798)/(o+1.)+(0.531297236071)
        ref[2,1,0]=(-1.29575884246)/(o+1.)+(0.153942519064)
        ref[2,1,1]=(-0.600379662508)/(o+1.)+(0.565678131765)
        ref[3,0,0]=(1.87814767227)/(o+1.)+(-0.140116323394)
        ref[3,0,1]=(-0.195825320078)/(o+1.)+(-0.523630356799)
        ref[3,1,0]=(-0.959261224384)/(o+1.)+(0.249823262853)
        ref[3,1,1]=(1.05531553015)/(o+1.)+(-0.858859911271)
        ref[4,0,0]=(0.65497734235)/(o+1.)+(-0.190760950929)
        ref[4,0,1]=(1.37092574409)/(o+1.)+(0.0173394004651)
        ref[4,1,0]=(1.27754767629)/(o+1.)+(-0.6802861583)
        ref[4,1,1]=(1.23997013471)/(o+1.)+(0.70527600063)
        ref[5,0,0]=(-0.383142890439)/(o+1.)+(0.312083322186)
        ref[5,0,1]=(1.64380369652)/(o+1.)+(-0.390348387013)
        ref[5,1,0]=(-0.812713549757)/(o+1.)+(0.520467462905)
        ref[5,1,1]=(0.600505313178)/(o+1.)+(-0.287048092454)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.235888975732)*x[0]+(-0.734372259615)*x[1]
        s[0,0,0,1]=(0.0665724430605)*x[0]+(-0.0890002398629)*x[1]
        s[0,0,1,0]=(-1.19774759676)*x[0]+(1.00321581155)*x[1]
        s[0,0,1,1]=(0.93370004223)*x[0]+(-0.953835581453)*x[1]
        s[0,0,2,0]=(-0.0557204731966)*x[0]+(0.94945217411)*x[1]
        s[0,0,2,1]=(-0.772628465097)*x[0]+(1.77100579145)*x[1]
        s[0,1,0,0]=(-0.405653863463)*x[0]+(0.494482129094)*x[1]
        s[0,1,0,1]=(0.636236763209)*x[0]+(-1.63254203234)*x[1]
        s[0,1,1,0]=(0.941764211044)*x[0]+(0.699520209508)*x[1]
        s[0,1,1,1]=(1.05045643599)*x[0]+(0.0596636815695)*x[1]
        s[0,1,2,0]=(0.578668042684)*x[0]+(0.285442340394)*x[1]
        s[0,1,2,1]=(0.469149909671)*x[0]+(1.26526906754)*x[1]
        s[0,2,0,0]=(1.05435496407)*x[0]+(-0.592645628655)*x[1]
        s[0,2,0,1]=(0.066181979241)*x[0]+(-1.42368311585)*x[1]
        s[0,2,1,0]=(-0.896588999422)*x[0]+(1.44881386304)*x[1]
        s[0,2,1,1]=(-0.862544854709)*x[0]+(-0.398360134784)*x[1]
        s[0,2,2,0]=(1.02969811319)*x[0]+(0.318993808627)*x[1]
        s[0,2,2,1]=(0.0143773859487)*x[0]+(-0.327118788648)*x[1]
        s[0,3,0,0]=(0.021196438108)*x[0]+(1.8515159226)*x[1]
        s[0,3,0,1]=(0.627726779803)*x[0]+(-0.302716383303)*x[1]
        s[0,3,1,0]=(0.165624370858)*x[0]+(0.274919760437)*x[1]
        s[0,3,1,1]=(0.283214725106)*x[0]+(-0.872378951063)*x[1]
        s[0,3,2,0]=(-0.83391985409)*x[0]+(0.731217556829)*x[1]
        s[0,3,2,1]=(0.918833558754)*x[0]+(0.714715370412)*x[1]
        s[0,4,0,0]=(-0.503211677055)*x[0]+(0.0289310095577)*x[1]
        s[0,4,0,1]=(0.597129238132)*x[0]+(0.540475624893)*x[1]
        s[0,4,1,0]=(0.687211570349)*x[0]+(0.594018658674)*x[1]
        s[0,4,1,1]=(1.5970484642)*x[0]+(-0.495319600999)*x[1]
        s[0,4,2,0]=(1.20627016067)*x[0]+(0.461261823726)*x[1]
        s[0,4,2,1]=(0.165927055028)*x[0]+(1.09007328428)*x[1]
        s[1,0,0,0]=(1.16105917598)*x[0]+(1.82590164801)*x[1]
        s[1,0,0,1]=(-1.13479961391)*x[0]+(0.0351464050353)*x[1]
        s[1,0,1,0]=(-0.318954647255)*x[0]+(0.210946813363)*x[1]
        s[1,0,1,1]=(1.31393538581)*x[0]+(0.198463680367)*x[1]
        s[1,0,2,0]=(1.17799558613)*x[0]+(0.514357347746)*x[1]
        s[1,0,2,1]=(0.571191341129)*x[0]+(0.694971250457)*x[1]
        s[1,1,0,0]=(0.132901410189)*x[0]+(-0.262070778236)*x[1]
        s[1,1,0,1]=(-0.871431837436)*x[0]+(-0.514707012778)*x[1]
        s[1,1,1,0]=(-0.319003186732)*x[0]+(-0.014973803902)*x[1]
        s[1,1,1,1]=(0.260213741998)*x[0]+(1.32908581921)*x[1]
        s[1,1,2,0]=(0.811339404893)*x[0]+(-0.304564391555)*x[1]
        s[1,1,2,1]=(-0.062193223101)*x[0]+(-0.911704705741)*x[1]
        s[1,2,0,0]=(0.0932623640149)*x[0]+(1.12570360092)*x[1]
        s[1,2,0,1]=(1.16504942277)*x[0]+(-0.385840092559)*x[1]
        s[1,2,1,0]=(1.3440680469)*x[0]+(0.532696669818)*x[1]
        s[1,2,1,1]=(1.31050128725)*x[0]+(0.00906728815983)*x[1]
        s[1,2,2,0]=(-1.36931832474)*x[0]+(-1.70562158659)*x[1]
        s[1,2,2,1]=(-1.8375028036)*x[0]+(1.52366054242)*x[1]
        s[1,3,0,0]=(0.474023303611)*x[0]+(0.437518737767)*x[1]
        s[1,3,0,1]=(-0.42921544272)*x[0]+(0.387647202618)*x[1]
        s[1,3,1,0]=(0.863045413096)*x[0]+(-0.36260979035)*x[1]
        s[1,3,1,1]=(-0.654365401799)*x[0]+(0.533926246461)*x[1]
        s[1,3,2,0]=(0.418152564814)*x[0]+(1.11544647745)*x[1]
        s[1,3,2,1]=(-1.60947873228)*x[0]+(0.591727770738)*x[1]
        s[1,4,0,0]=(-1.51807668572)*x[0]+(-1.01613349614)*x[1]
        s[1,4,0,1]=(-1.50163863727)*x[0]+(1.50376272164)*x[1]
        s[1,4,1,0]=(0.320776223369)*x[0]+(-0.700391339598)*x[1]
        s[1,4,1,1]=(0.745224476612)*x[0]+(-0.850870606209)*x[1]
        s[1,4,2,0]=(-0.179598404379)*x[0]+(1.75799776411)*x[1]
        s[1,4,2,1]=(0.00909330900106)*x[0]+(-0.077399205942)*x[1]
        s[2,0,0,0]=(0.505719447396)*x[0]+(-0.275603298829)*x[1]
        s[2,0,0,1]=(1.46787325811)*x[0]+(0.534793377012)*x[1]
        s[2,0,1,0]=(0.77645891983)*x[0]+(0.540540443218)*x[1]
        s[2,0,1,1]=(0.429661881583)*x[0]+(1.43421311479)*x[1]
        s[2,0,2,0]=(1.02806543045)*x[0]+(0.931834883879)*x[1]
        s[2,0,2,1]=(0.763031666632)*x[0]+(0.873653388204)*x[1]
        s[2,1,0,0]=(0.261772733655)*x[0]+(-1.18679909814)*x[1]
        s[2,1,0,1]=(-1.61662908583)*x[0]+(1.10646142209)*x[1]
        s[2,1,1,0]=(0.188527773811)*x[0]+(-0.772410233645)*x[1]
        s[2,1,1,1]=(-0.869931105194)*x[0]+(1.51014899284)*x[1]
        s[2,1,2,0]=(0.252478837689)*x[0]+(1.22480134325)*x[1]
        s[2,1,2,1]=(1.49578867187)*x[0]+(-0.877973056373)*x[1]
        s[2,2,0,0]=(-0.209568005598)*x[0]+(-0.0462992389671)*x[1]
        s[2,2,0,1]=(-0.298130269216)*x[0]+(-0.908352713549)*x[1]
        s[2,2,1,0]=(-0.586370615835)*x[0]+(-1.72704626602)*x[1]
        s[2,2,1,1]=(1.26217059553)*x[0]+(0.676548549908)*x[1]
        s[2,2,2,0]=(0.961200845015)*x[0]+(0.612151157303)*x[1]
        s[2,2,2,1]=(0.700873758828)*x[0]+(0.800662765762)*x[1]
        s[2,3,0,0]=(-0.319747641322)*x[0]+(0.689929330298)*x[1]
        s[2,3,0,1]=(-1.04393759081)*x[0]+(-1.80449230281)*x[1]
        s[2,3,1,0]=(-0.0410072962318)*x[0]+(-0.678000289402)*x[1]
        s[2,3,1,1]=(-0.329381648117)*x[0]+(-0.935487069857)*x[1]
        s[2,3,2,0]=(-0.907460976718)*x[0]+(0.0314212805943)*x[1]
        s[2,3,2,1]=(0.593604907619)*x[0]+(-0.479146057787)*x[1]
        s[2,4,0,0]=(-0.0337048332079)*x[0]+(0.239721443625)*x[1]
        s[2,4,0,1]=(0.721609580339)*x[0]+(-1.06464423044)*x[1]
        s[2,4,1,0]=(0.670568080353)*x[0]+(1.28774286264)*x[1]
        s[2,4,1,1]=(0.603608338555)*x[0]+(0.323996221169)*x[1]
        s[2,4,2,0]=(0.0371268677698)*x[0]+(0.401814556385)*x[1]
        s[2,4,2,1]=(0.436926475796)*x[0]+(-0.819433312575)*x[1]
        s[3,0,0,0]=(-0.0190186791033)*x[0]+(0.501672647061)*x[1]
        s[3,0,0,1]=(0.00060856018325)*x[0]+(-0.414668792931)*x[1]
        s[3,0,1,0]=(0.699428985955)*x[0]+(0.612631528371)*x[1]
        s[3,0,1,1]=(-0.659371576463)*x[0]+(-0.0982545515259)*x[1]
        s[3,0,2,0]=(0.444351889183)*x[0]+(0.53998411713)*x[1]
        s[3,0,2,1]=(-0.281563709018)*x[0]+(-0.486501437256)*x[1]
        s[3,1,0,0]=(0.407521411466)*x[0]+(-0.749870044816)*x[1]
        s[3,1,0,1]=(-0.638780264813)*x[0]+(0.230397511262)*x[1]
        s[3,1,1,0]=(1.16000440204)*x[0]+(0.13074916095)*x[1]
        s[3,1,1,1]=(-1.30550439046)*x[0]+(-1.33788704039)*x[1]
        s[3,1,2,0]=(-0.540767742404)*x[0]+(-0.148316212185)*x[1]
        s[3,1,2,1]=(1.70452499495)*x[0]+(0.935421717463)*x[1]
        s[3,2,0,0]=(0.320990904519)*x[0]+(0.467953609756)*x[1]
        s[3,2,0,1]=(-0.26002228203)*x[0]+(-0.925765706976)*x[1]
        s[3,2,1,0]=(0.925754935382)*x[0]+(0.240432086422)*x[1]
        s[3,2,1,1]=(0.0214680080146)*x[0]+(-1.11507977191)*x[1]
        s[3,2,2,0]=(-0.187946153679)*x[0]+(0.024781322847)*x[1]
        s[3,2,2,1]=(-1.41866510495)*x[0]+(0.596173034069)*x[1]
        s[3,3,0,0]=(0.77087649562)*x[0]+(0.848168811103)*x[1]
        s[3,3,0,1]=(0.218810390517)*x[0]+(0.116909485563)*x[1]
        s[3,3,1,0]=(-0.889981738645)*x[0]+(-0.506355969844)*x[1]
        s[3,3,1,1]=(1.15870903185)*x[0]+(-0.8158119166)*x[1]
        s[3,3,2,0]=(-0.345230989856)*x[0]+(-0.138496974465)*x[1]
        s[3,3,2,1]=(-0.819830787817)*x[0]+(-1.66747998417)*x[1]
        s[3,4,0,0]=(-1.74647564112)*x[0]+(0.22309219172)*x[1]
        s[3,4,0,1]=(0.146804717412)*x[0]+(0.493084518827)*x[1]
        s[3,4,1,0]=(-0.536943383893)*x[0]+(-0.686288132381)*x[1]
        s[3,4,1,1]=(-0.142983666789)*x[0]+(-0.183927597857)*x[1]
        s[3,4,2,0]=(-0.173071632973)*x[0]+(-0.696870700802)*x[1]
        s[3,4,2,1]=(-0.359163238166)*x[0]+(-0.172033035327)*x[1]
        ref[0,0,0,0]=(-0.603598770198)/(o+1.)+(0.0525577431575)
        ref[0,0,0,1]=(1.04505857398)/(o+1.)+(-0.533743185393)
        ref[0,0,1,0]=(0.00772071573504)/(o+1.)+(-0.101126250473)
        ref[0,0,1,1]=(-0.290587702404)/(o+1.)+(0.135226081591)
        ref[0,0,2,0]=(0.614426628234)/(o+1.)+(0.139652536339)
        ref[0,0,2,1]=(-0.0228094425795)/(o+1.)+(0.510593384465)
        ref[0,1,0,0]=(-0.436885078328)/(o+1.)+(0.262856671979)
        ref[0,1,0,1]=(0.0510259667548)/(o+1.)+(-0.523665617945)
        ref[0,1,1,0]=(0.744462079536)/(o+1.)+(0.448411170508)
        ref[0,1,1,1]=(1.52720962375)/(o+1.)+(-0.208544753096)
        ref[0,1,2,0]=(0.73405306307)/(o+1.)+(0.0650286600044)
        ref[0,1,2,1]=(0.550924984933)/(o+1.)+(0.591746996137)
        ref[0,2,0,0]=(-0.347797813257)/(o+1.)+(0.404753574338)
        ref[0,2,0,1]=(0.187043042708)/(o+1.)+(-0.772272089659)
        ref[0,2,1,0]=(-0.316098850787)/(o+1.)+(0.434161857202)
        ref[0,2,1,1]=(-0.536124734534)/(o+1.)+(-0.36239012748)
        ref[0,2,2,0]=(0.585045523113)/(o+1.)+(0.381823199352)
        ref[0,2,2,1]=(0.422243751365)/(o+1.)+(-0.367492577032)
        ref[0,3,0,0]=(0.91381158808)/(o+1.)+(0.479450386313)
        ref[0,3,0,1]=(0.209840671177)/(o+1.)+(0.0575848626616)
        ref[0,3,1,0]=(0.158639778589)/(o+1.)+(0.140952176353)
        ref[0,3,1,1]=(-0.364808895601)/(o+1.)+(-0.112177665178)
        ref[0,3,2,0]=(-0.198937574741)/(o+1.)+(0.0481176387404)
        ref[0,3,2,1]=(1.05420599501)/(o+1.)+(0.289671467078)
        ref[0,4,0,0]=(0.811701728881)/(o+1.)+(-0.642991198189)
        ref[0,4,0,1]=(0.693754647803)/(o+1.)+(0.221925107611)
        ref[0,4,1,0]=(1.05453609011)/(o+1.)+(0.113347069457)
        ref[0,4,1,1]=(0.117274271978)/(o+1.)+(0.49222729561)
        ref[0,4,2,0]=(0.496795563283)/(o+1.)+(0.585368210556)
        ref[0,4,2,1]=(0.83272690983)/(o+1.)+(0.211636714741)
        ref[1,0,0,0]=(1.46878592273)/(o+1.)+(0.75908745063)
        ref[1,0,0,1]=(-0.250785630529)/(o+1.)+(-0.424433789171)
        ref[1,0,1,0]=(0.161786720779)/(o+1.)+(-0.134897277336)
        ref[1,0,1,1]=(1.12719401256)/(o+1.)+(0.192602526813)
        ref[1,0,2,0]=(0.7197235394)/(o+1.)+(0.486314697239)
        ref[1,0,2,1]=(1.02804458464)/(o+1.)+(0.119059003471)
        ref[1,1,0,0]=(0.0361584520541)/(o+1.)+(-0.0826639100507)
        ref[1,1,0,1]=(-1.06091812324)/(o+1.)+(-0.162610363489)
        ref[1,1,1,0]=(1.13776585837)/(o+1.)+(-0.735871424501)
        ref[1,1,1,1]=(1.4740891177)/(o+1.)+(0.0576052217538)
        ref[1,1,2,0]=(0.280616623834)/(o+1.)+(0.113079194752)
        ref[1,1,2,1]=(-0.491890699653)/(o+1.)+(-0.241003614595)
        ref[1,2,0,0]=(-0.0342838844534)/(o+1.)+(0.626624924695)
        ref[1,2,0,1]=(0.770911868081)/(o+1.)+(0.00414873106652)
        ref[1,2,1,0]=(0.30749084682)/(o+1.)+(0.784636934951)
        ref[1,2,1,1]=(1.8638645006)/(o+1.)+(-0.272147962595)
        ref[1,2,2,0]=(-1.46841213372)/(o+1.)+(-0.803263888803)
        ref[1,2,2,1]=(-0.302454198157)/(o+1.)+(-0.00569403150815)
        ref[1,3,0,0]=(0.493917522063)/(o+1.)+(0.208812259657)
        ref[1,3,0,1]=(0.476274657976)/(o+1.)+(-0.258921449039)
        ref[1,3,1,0]=(-0.736248174893)/(o+1.)+(0.618341898819)
        ref[1,3,1,1]=(-0.697129670691)/(o+1.)+(0.288345257677)
        ref[1,3,2,0]=(0.303153912627)/(o+1.)+(0.615222564816)
        ref[1,3,2,1]=(-0.58122630211)/(o+1.)+(-0.218262329715)
        ref[1,4,0,0]=(-1.15442442968)/(o+1.)+(-0.689892876091)
        ref[1,4,0,1]=(0.0585826445649)/(o+1.)+(-0.0282292800982)
        ref[1,4,1,0]=(0.3252337262)/(o+1.)+(-0.352424421214)
        ref[1,4,1,1]=(0.512701244088)/(o+1.)+(-0.309173686843)
        ref[1,4,2,0]=(1.5506882457)/(o+1.)+(0.0138555570174)
        ref[1,4,2,1]=(-0.940773035102)/(o+1.)+(0.436233569081)
        ref[2,0,0,0]=(1.54247754874)/(o+1.)+(-0.656180700088)
        ref[2,0,0,1]=(0.988306961756)/(o+1.)+(0.507179836684)
        ref[2,0,1,0]=(-0.416628824296)/(o+1.)+(0.866814093672)
        ref[2,0,1,1]=(1.41006066195)/(o+1.)+(0.22690716721)
        ref[2,0,2,0]=(0.353011705537)/(o+1.)+(0.803444304396)
        ref[2,0,2,1]=(0.687642260874)/(o+1.)+(0.474521396981)
        ref[2,1,0,0]=(-0.218198987148)/(o+1.)+(-0.353413688668)
        ref[2,1,0,1]=(-0.635306433562)/(o+1.)+(0.0625693849126)
        ref[2,1,1,0]=(0.53416867099)/(o+1.)+(-0.559025565412)
        ref[2,1,1,1]=(0.00543686305333)/(o+1.)+(0.317390512297)
        ref[2,1,2,0]=(0.916573120847)/(o+1.)+(0.280353530048)
        ref[2,1,2,1]=(-0.0537588152546)/(o+1.)+(0.335787215373)
        ref[2,2,0,0]=(1.0055660817)/(o+1.)+(-0.630716663133)
        ref[2,2,0,1]=(0.420474306577)/(o+1.)+(-0.813478644671)
        ref[2,2,1,0]=(-0.723008657238)/(o+1.)+(-0.795204112308)
        ref[2,2,1,1]=(0.306188174034)/(o+1.)+(0.816265485704)
        ref[2,2,2,0]=(0.806158569368)/(o+1.)+(0.383596716475)
        ref[2,2,2,1]=(1.6353377865)/(o+1.)+(-0.066900630957)
        ref[2,3,0,0]=(-1.24813850815)/(o+1.)+(0.809160098565)
        ref[2,3,0,1]=(-1.56942041127)/(o+1.)+(-0.639504741178)
        ref[2,3,1,0]=(-1.49605878837)/(o+1.)+(0.388525601366)
        ref[2,3,1,1]=(-0.448403645878)/(o+1.)+(-0.408232536048)
        ref[2,3,2,0]=(-1.13568794342)/(o+1.)+(0.129824123646)
        ref[2,3,2,1]=(0.596733666259)/(o+1.)+(-0.241137408214)
        ref[2,4,0,0]=(0.0191094707617)/(o+1.)+(0.0934535698276)
        ref[2,4,0,1]=(0.185204097773)/(o+1.)+(-0.264119373935)
        ref[2,4,1,0]=(1.55855563639)/(o+1.)+(0.199877653305)
        ref[2,4,1,1]=(0.11469142184)/(o+1.)+(0.406456568942)
        ref[2,4,2,0]=(-0.954321542363)/(o+1.)+(0.696631483259)
        ref[2,4,2,1]=(-0.810227566168)/(o+1.)+(0.213860364695)
        ref[3,0,0,0]=(0.453436736881)/(o+1.)+(0.0146086155385)
        ref[3,0,0,1]=(-0.210824760654)/(o+1.)+(-0.101617736047)
        ref[3,0,1,0]=(0.414458928012)/(o+1.)+(0.448800793157)
        ref[3,0,1,1]=(0.346129385596)/(o+1.)+(-0.551877756793)
        ref[3,0,2,0]=(1.85496275992)/(o+1.)+(-0.435313376805)
        ref[3,0,2,1]=(-0.110500317363)/(o+1.)+(-0.328782414455)
        ref[3,1,0,0]=(0.0399449935267)/(o+1.)+(-0.191146813438)
        ref[3,1,0,1]=(0.133080949177)/(o+1.)+(-0.270731851364)
        ref[3,1,1,0]=(0.139648356416)/(o+1.)+(0.575552603289)
        ref[3,1,1,1]=(-1.45813597924)/(o+1.)+(-0.592627725804)
        ref[3,1,2,0]=(-0.773854684679)/(o+1.)+(0.0423853650452)
        ref[3,1,2,1]=(1.2726274277)/(o+1.)+(0.683659642354)
        ref[3,2,0,0]=(0.555324507349)/(o+1.)+(0.116810003463)
        ref[3,2,0,1]=(-0.310977730161)/(o+1.)+(-0.437405129422)
        ref[3,2,1,0]=(1.05800939782)/(o+1.)+(0.0540888119933)
        ref[3,2,1,1]=(0.386222714751)/(o+1.)+(-0.739917239322)
        ref[3,2,2,0]=(-0.95705464593)/(o+1.)+(0.396944907549)
        ref[3,2,2,1]=(-0.491612682819)/(o+1.)+(-0.165439694033)
        ref[3,3,0,0]=(0.666972772584)/(o+1.)+(0.47603626707)
        ref[3,3,0,1]=(0.268506355312)/(o+1.)+(0.0336067603842)
        ref[3,3,1,0]=(-0.57724956178)/(o+1.)+(-0.409544073355)
        ref[3,3,1,1]=(-0.518808765862)/(o+1.)+(0.430852940555)
        ref[3,3,2,0]=(-1.95248352029)/(o+1.)+(0.734377777986)
        ref[3,3,2,1]=(-1.14565232021)/(o+1.)+(-0.67082922589)
        ref[3,4,0,0]=(-1.13455573259)/(o+1.)+(-0.1944138584)
        ref[3,4,0,1]=(-0.508914286468)/(o+1.)+(0.574401761354)
        ref[3,4,1,0]=(-0.319104038837)/(o+1.)+(-0.452063738718)
        ref[3,4,1,1]=(-0.100356733327)/(o+1.)+(-0.11327726566)
        ref[3,4,2,0]=(-0.39970881352)/(o+1.)+(-0.235116760127)
        ref[3,4,2,1]=(-0.339631070901)/(o+1.)+(-0.0957826012957)
      else:
        s[0,0,0,0]=(0.0794776522341)*x[0]+(1.69294456255)*x[1]+(0.824115844979)*x[2]
        s[0,0,0,1]=(0.0720523378919)*x[0]+(-1.39724154561)*x[1]+(-0.010988810956)*x[2]
        s[0,0,1,0]=(0.985556321125)*x[0]+(1.12133844031)*x[1]+(-0.262491219522)*x[2]
        s[0,0,1,1]=(0.789643681767)*x[0]+(1.02860624013)*x[1]+(-0.450770685703)*x[2]
        s[0,0,2,0]=(-0.850427026537)*x[0]+(-0.393993518297)*x[1]+(0.12816296306)*x[2]
        s[0,0,2,1]=(-0.47508130541)*x[0]+(-0.393857510976)*x[1]+(0.66003641476)*x[2]
        s[0,1,0,0]=(-1.87725252884)*x[0]+(-0.668075530728)*x[1]+(0.938485679482)*x[2]
        s[0,1,0,1]=(0.820537136967)*x[0]+(-0.453620529157)*x[1]+(0.811944595883)*x[2]
        s[0,1,1,0]=(0.914995161971)*x[0]+(-0.196153821444)*x[1]+(0.938555859255)*x[2]
        s[0,1,1,1]=(-0.866633718493)*x[0]+(-0.534596859321)*x[1]+(0.77514383239)*x[2]
        s[0,1,2,0]=(0.183094085342)*x[0]+(0.727499336912)*x[1]+(0.170986687969)*x[2]
        s[0,1,2,1]=(-1.36220673904)*x[0]+(0.0822122492483)*x[1]+(0.124227107716)*x[2]
        s[0,2,0,0]=(1.55269479425)*x[0]+(-0.14574857122)*x[1]+(1.24120289667)*x[2]
        s[0,2,0,1]=(-0.464038186768)*x[0]+(0.00554383028356)*x[1]+(-1.14838250985)*x[2]
        s[0,2,1,0]=(0.904608031363)*x[0]+(-0.313004420619)*x[1]+(-0.10529298073)*x[2]
        s[0,2,1,1]=(0.156623578469)*x[0]+(0.55127494399)*x[1]+(-0.385447926692)*x[2]
        s[0,2,2,0]=(-0.766804018933)*x[0]+(0.18519619356)*x[1]+(1.00915540081)*x[2]
        s[0,2,2,1]=(0.00231157548456)*x[0]+(0.338768642838)*x[1]+(-0.684228602438)*x[2]
        s[0,3,0,0]=(-0.834810486599)*x[0]+(-0.096360306458)*x[1]+(1.28924889761)*x[2]
        s[0,3,0,1]=(-0.658443843543)*x[0]+(0.889474267886)*x[1]+(0.566763124725)*x[2]
        s[0,3,1,0]=(-0.109777093963)*x[0]+(0.228053887631)*x[1]+(1.22952539906)*x[2]
        s[0,3,1,1]=(-0.749479260801)*x[0]+(-1.01102432725)*x[1]+(-0.196861989291)*x[2]
        s[0,3,2,0]=(0.650389523509)*x[0]+(0.151701374527)*x[1]+(0.506246388483)*x[2]
        s[0,3,2,1]=(-0.648510958338)*x[0]+(-0.263590202304)*x[1]+(-1.35992044414)*x[2]
        s[0,4,0,0]=(0.196976454523)*x[0]+(-0.193947226599)*x[1]+(1.67641586045)*x[2]
        s[0,4,0,1]=(-0.455868626908)*x[0]+(0.506302321882)*x[1]+(1.1666778393)*x[2]
        s[0,4,1,0]=(-0.788545609857)*x[0]+(-0.487332926053)*x[1]+(-0.0645141715605)*x[2]
        s[0,4,1,1]=(-0.320424709093)*x[0]+(-0.683301403769)*x[1]+(-0.222245108268)*x[2]
        s[0,4,2,0]=(0.876149348517)*x[0]+(0.12449602802)*x[1]+(-0.220531205966)*x[2]
        s[0,4,2,1]=(-0.2764172159)*x[0]+(-0.894571003479)*x[1]+(-0.595830070177)*x[2]
        s[1,0,0,0]=(-0.79676665323)*x[0]+(-1.25965739347)*x[1]+(-0.319474113915)*x[2]
        s[1,0,0,1]=(1.601348448)*x[0]+(0.41027863439)*x[1]+(-0.346758807915)*x[2]
        s[1,0,1,0]=(0.181292913253)*x[0]+(1.10494312205)*x[1]+(0.196973772813)*x[2]
        s[1,0,1,1]=(0.562861818795)*x[0]+(-0.997990211102)*x[1]+(0.0332706615838)*x[2]
        s[1,0,2,0]=(-0.989984164479)*x[0]+(-0.886186583204)*x[1]+(-0.228620900973)*x[2]
        s[1,0,2,1]=(0.841365917044)*x[0]+(-0.0519066039156)*x[1]+(-0.81126199301)*x[2]
        s[1,1,0,0]=(-0.613125126303)*x[0]+(0.482770874125)*x[1]+(0.038374960382)*x[2]
        s[1,1,0,1]=(-0.203719663632)*x[0]+(-0.281873286682)*x[1]+(-0.0303453795347)*x[2]
        s[1,1,1,0]=(1.11287699536)*x[0]+(-1.04759296757)*x[1]+(1.64257695374)*x[2]
        s[1,1,1,1]=(-0.0861307607908)*x[0]+(0.261314888791)*x[1]+(-0.204955973137)*x[2]
        s[1,1,2,0]=(0.821662110998)*x[0]+(-1.05649310988)*x[1]+(0.609822823353)*x[2]
        s[1,1,2,1]=(-0.730526005063)*x[0]+(1.62018363097)*x[1]+(0.99111776492)*x[2]
        s[1,2,0,0]=(-1.25039777751)*x[0]+(0.811070034681)*x[1]+(1.32505446058)*x[2]
        s[1,2,0,1]=(0.753110041895)*x[0]+(0.186756235207)*x[1]+(-1.30430176967)*x[2]
        s[1,2,1,0]=(0.146182105242)*x[0]+(-0.863831927779)*x[1]+(-1.1030247274)*x[2]
        s[1,2,1,1]=(0.674634709918)*x[0]+(-0.560866548824)*x[1]+(0.576408322189)*x[2]
        s[1,2,2,0]=(-0.573489753302)*x[0]+(1.57147013427)*x[1]+(0.711691292064)*x[2]
        s[1,2,2,1]=(0.549483527739)*x[0]+(-1.14179814973)*x[1]+(1.60524619335)*x[2]
        s[1,3,0,0]=(0.197453494754)*x[0]+(-0.456559966113)*x[1]+(-1.21423667747)*x[2]
        s[1,3,0,1]=(-1.76312274691)*x[0]+(-1.04778192886)*x[1]+(-0.185488490674)*x[2]
        s[1,3,1,0]=(0.0162460011376)*x[0]+(1.16407944993)*x[1]+(1.16424268292)*x[2]
        s[1,3,1,1]=(-0.513913288204)*x[0]+(-1.46536589962)*x[1]+(-0.906205233261)*x[2]
        s[1,3,2,0]=(-0.0157400074256)*x[0]+(-0.293812243081)*x[1]+(0.815280978304)*x[2]
        s[1,3,2,1]=(1.57630585875)*x[0]+(0.290739280065)*x[1]+(-0.771398356417)*x[2]
        s[1,4,0,0]=(-0.323882779703)*x[0]+(-0.763929395312)*x[1]+(0.738403272406)*x[2]
        s[1,4,0,1]=(-0.476278532043)*x[0]+(-0.11039317349)*x[1]+(0.346684885005)*x[2]
        s[1,4,1,0]=(-1.36304063164)*x[0]+(-0.660619165296)*x[1]+(-0.903566907206)*x[2]
        s[1,4,1,1]=(0.0863780283437)*x[0]+(0.466792514558)*x[1]+(0.448776109878)*x[2]
        s[1,4,2,0]=(0.74478619068)*x[0]+(0.446314847678)*x[1]+(-0.0481741032015)*x[2]
        s[1,4,2,1]=(0.746506715398)*x[0]+(-0.432201202785)*x[1]+(0.203440105689)*x[2]
        s[2,0,0,0]=(-0.785805880678)*x[0]+(-0.862407720473)*x[1]+(-1.21640925829)*x[2]
        s[2,0,0,1]=(0.647190817984)*x[0]+(0.519851314653)*x[1]+(0.390946186931)*x[2]
        s[2,0,1,0]=(-1.24831725337)*x[0]+(1.59413938467)*x[1]+(-1.33358250136)*x[2]
        s[2,0,1,1]=(-0.152213072763)*x[0]+(-1.09357954369)*x[1]+(-1.25586196425)*x[2]
        s[2,0,2,0]=(-0.0314861108304)*x[0]+(0.293907529511)*x[1]+(-0.21039874768)*x[2]
        s[2,0,2,1]=(0.0467194832302)*x[0]+(0.527286460462)*x[1]+(-0.415618293894)*x[2]
        s[2,1,0,0]=(-0.189575051656)*x[0]+(-0.844885847783)*x[1]+(-0.334235875002)*x[2]
        s[2,1,0,1]=(-0.00822711174783)*x[0]+(-1.5897272285)*x[1]+(1.50002681163)*x[2]
        s[2,1,1,0]=(-0.841981498104)*x[0]+(-1.10376539971)*x[1]+(0.311385102139)*x[2]
        s[2,1,1,1]=(1.12897934543)*x[0]+(-0.966261308872)*x[1]+(-1.16359871767)*x[2]
        s[2,1,2,0]=(-0.957189463062)*x[0]+(0.0471569985184)*x[1]+(0.12392515517)*x[2]
        s[2,1,2,1]=(-0.752369087888)*x[0]+(1.18176351896)*x[1]+(0.334788137458)*x[2]
        s[2,2,0,0]=(1.22014205343)*x[0]+(0.784413576281)*x[1]+(0.0962245348039)*x[2]
        s[2,2,0,1]=(-0.407023035962)*x[0]+(1.11597107325)*x[1]+(0.458612291763)*x[2]
        s[2,2,1,0]=(-0.0496175828387)*x[0]+(-0.0536216340718)*x[1]+(0.771683970464)*x[2]
        s[2,2,1,1]=(-0.356958793645)*x[0]+(-1.38580705404)*x[1]+(-0.670167616875)*x[2]
        s[2,2,2,0]=(0.0804001816041)*x[0]+(-1.16114869406)*x[1]+(0.475472363874)*x[2]
        s[2,2,2,1]=(0.98238551642)*x[0]+(-1.12578358118)*x[1]+(-0.00970359156434)*x[2]
        s[2,3,0,0]=(1.0110556058)*x[0]+(-0.488112488916)*x[1]+(0.243005554591)*x[2]
        s[2,3,0,1]=(-0.16668051279)*x[0]+(0.178067528024)*x[1]+(0.969038052731)*x[2]
        s[2,3,1,0]=(-0.00345650882131)*x[0]+(-0.941671547756)*x[1]+(-0.49908897105)*x[2]
        s[2,3,1,1]=(0.716683430797)*x[0]+(-0.585708024289)*x[1]+(-1.19073441986)*x[2]
        s[2,3,2,0]=(0.400515369791)*x[0]+(-0.454556618814)*x[1]+(0.0782324352019)*x[2]
        s[2,3,2,1]=(-0.164628093414)*x[0]+(-1.26464027792)*x[1]+(0.179511210919)*x[2]
        s[2,4,0,0]=(1.33277550072)*x[0]+(1.18570575409)*x[1]+(0.381346013717)*x[2]
        s[2,4,0,1]=(0.588290284687)*x[0]+(-1.08302445047)*x[1]+(0.166898904485)*x[2]
        s[2,4,1,0]=(0.131178111015)*x[0]+(-0.749855470189)*x[1]+(-0.1247474719)*x[2]
        s[2,4,1,1]=(-0.97245336122)*x[0]+(0.873932906896)*x[1]+(0.120264062027)*x[2]
        s[2,4,2,0]=(-0.722531128456)*x[0]+(0.19636289816)*x[1]+(-0.907945347107)*x[2]
        s[2,4,2,1]=(-0.711111750197)*x[0]+(0.31947094036)*x[1]+(-1.79333697163)*x[2]
        s[3,0,0,0]=(-0.184519023765)*x[0]+(0.680221159423)*x[1]+(-0.0244123658903)*x[2]
        s[3,0,0,1]=(-1.14573637546)*x[0]+(0.0616977700293)*x[1]+(-0.654808111844)*x[2]
        s[3,0,1,0]=(-1.09297543018)*x[0]+(-1.04184353717)*x[1]+(0.384120100538)*x[2]
        s[3,0,1,1]=(0.105247846819)*x[0]+(-1.31600150124)*x[1]+(1.43496073616)*x[2]
        s[3,0,2,0]=(0.881148598911)*x[0]+(-0.478723216871)*x[1]+(-0.0746344543132)*x[2]
        s[3,0,2,1]=(-1.20828681587)*x[0]+(-1.65808839437)*x[1]+(-0.636678098535)*x[2]
        s[3,1,0,0]=(-0.213394309668)*x[0]+(-0.240354748844)*x[1]+(0.117282970374)*x[2]
        s[3,1,0,1]=(1.37462845901)*x[0]+(-0.41583894581)*x[1]+(-0.625626047701)*x[2]
        s[3,1,1,0]=(1.83991491983)*x[0]+(1.2778674571)*x[1]+(1.35465906139)*x[2]
        s[3,1,1,1]=(1.58168546715)*x[0]+(0.829809996131)*x[1]+(0.681232260732)*x[2]
        s[3,1,2,0]=(-0.0212907664121)*x[0]+(1.00109823691)*x[1]+(-1.30845432825)*x[2]
        s[3,1,2,1]=(0.169920360674)*x[0]+(-0.864995850094)*x[1]+(0.302979947072)*x[2]
        s[3,2,0,0]=(-0.154084636739)*x[0]+(0.319647474298)*x[1]+(-1.23506947259)*x[2]
        s[3,2,0,1]=(-0.199216213818)*x[0]+(1.28921202953)*x[1]+(1.65805122541)*x[2]
        s[3,2,1,0]=(-1.86774194364)*x[0]+(-0.101723263166)*x[1]+(-0.526029332636)*x[2]
        s[3,2,1,1]=(-0.177960964857)*x[0]+(-0.899073564225)*x[1]+(0.551378002191)*x[2]
        s[3,2,2,0]=(0.18491319197)*x[0]+(-0.286906366286)*x[1]+(0.583859098776)*x[2]
        s[3,2,2,1]=(0.38623837272)*x[0]+(-0.653272502002)*x[1]+(-1.46718517275)*x[2]
        s[3,3,0,0]=(0.0494572779834)*x[0]+(-0.0566099666381)*x[1]+(-0.113874384717)*x[2]
        s[3,3,0,1]=(0.0241152098269)*x[0]+(-0.317061594612)*x[1]+(1.23769551101)*x[2]
        s[3,3,1,0]=(-0.107294616786)*x[0]+(-0.651410283508)*x[1]+(0.833104181804)*x[2]
        s[3,3,1,1]=(0.16676828757)*x[0]+(-0.670496617303)*x[1]+(1.15686719798)*x[2]
        s[3,3,2,0]=(0.159950143242)*x[0]+(0.0436823700228)*x[1]+(-0.645027680285)*x[2]
        s[3,3,2,1]=(-1.2697347822)*x[0]+(1.22938250997)*x[1]+(0.208587928792)*x[2]
        s[3,4,0,0]=(0.575105413412)*x[0]+(-1.44998456404)*x[1]+(-0.516125694177)*x[2]
        s[3,4,0,1]=(0.0387618783003)*x[0]+(0.859152861676)*x[1]+(1.29847202073)*x[2]
        s[3,4,1,0]=(-0.842510438531)*x[0]+(0.191878858518)*x[1]+(-0.73453059399)*x[2]
        s[3,4,1,1]=(1.41435810392)*x[0]+(-0.855734345577)*x[1]+(-0.425917153945)*x[2]
        s[3,4,2,0]=(-0.58228358663)*x[0]+(0.00936548588525)*x[1]+(-1.38536320116)*x[2]
        s[3,4,2,1]=(0.0419678913201)*x[0]+(1.02999559903)*x[1]+(-0.464745184871)*x[2]
        ref[0,0,0,0]=(1.30172410662)/(o+1.)+(0.647406976573)
        ref[0,0,0,1]=(-1.09442521373)/(o+1.)+(-0.120876402476)
        ref[0,0,1,0]=(2.01757642043)/(o+1.)+(-0.0865864392547)
        ref[0,0,1,1]=(1.80034002198)/(o+1.)+(-0.216430392892)
        ref[0,0,2,0]=(1.13142112916)/(o+1.)+(-1.12383935547)
        ref[0,0,2,1]=(-0.905515986966)/(o+1.)+(0.34830679267)
        ref[0,1,0,0]=(-0.210441134504)/(o+1.)+(-0.698200622791)
        ref[0,1,0,1]=(0.360807961621)/(o+1.)+(0.409026621036)
        ref[0,1,1,0]=(1.5147969883)/(o+1.)+(0.0713001057429)
        ref[0,1,1,1]=(0.396123567843)/(o+1.)+(-0.511105156634)
        ref[0,1,2,0]=(0.184468760768)/(o+1.)+(0.448555674728)
        ref[0,1,2,1]=(-0.133892978353)/(o+1.)+(-0.510937201862)
        ref[0,2,0,0]=(1.66137009462)/(o+1.)+(0.493389512541)
        ref[0,2,0,1]=(0.358452769174)/(o+1.)+(-0.982664817753)
        ref[0,2,1,0]=(0.966434135345)/(o+1.)+(-0.240061752665)
        ref[0,2,1,1]=(-0.401753883503)/(o+1.)+(0.362102239635)
        ref[0,2,2,0]=(-0.0153619747046)/(o+1.)+(0.221454775071)
        ref[0,2,2,1]=(-0.104138117282)/(o+1.)+(-0.119505133417)
        ref[0,3,0,0]=(-0.966204569194)/(o+1.)+(0.662141336872)
        ref[0,3,0,1]=(0.587661220327)/(o+1.)+(0.10506616437)
        ref[0,3,1,0]=(2.02288199931)/(o+1.)+(-0.337539903289)
        ref[0,3,1,1]=(-1.78842825396)/(o+1.)+(-0.0844686616883)
        ref[0,3,2,0]=(0.927282266945)/(o+1.)+(0.190527509786)
        ref[0,3,2,1]=(-0.679518039292)/(o+1.)+(-0.796251782745)
        ref[0,4,0,0]=(0.555388124049)/(o+1.)+(0.562028482163)
        ref[0,4,0,1]=(0.0870397234082)/(o+1.)+(0.565035905435)
        ref[0,4,1,0]=(-1.1463502212)/(o+1.)+(-0.0970212431374)
        ref[0,4,1,1]=(0.233233715042)/(o+1.)+(-0.729602468086)
        ref[0,4,2,0]=(-0.782889963002)/(o+1.)+(0.781502066786)
        ref[0,4,2,1]=(0.763433249649)/(o+1.)+(-1.2651257696)
        ref[1,0,0,0]=(-1.00441831898)/(o+1.)+(-0.685739920821)
        ref[1,0,0,1]=(0.751247546824)/(o+1.)+(0.456810363824)
        ref[1,0,1,0]=(2.22764882334)/(o+1.)+(-0.372219507613)
        ref[1,0,1,1]=(-1.19623631804)/(o+1.)+(0.39718929366)
        ref[1,0,2,0]=(-1.60343741738)/(o+1.)+(-0.250677115636)
        ref[1,0,2,1]=(-1.25680789752)/(o+1.)+(0.61750260882)
        ref[1,1,0,0]=(-1.45519189991)/(o+1.)+(0.681606304054)
        ref[1,1,0,1]=(0.840628495308)/(o+1.)+(-0.678283412578)
        ref[1,1,1,0]=(1.41826768667)/(o+1.)+(0.144796647431)
        ref[1,1,1,1]=(-0.933723251106)/(o+1.)+(0.451975702985)
        ref[1,1,2,0]=(-0.408324994487)/(o+1.)+(0.39165840948)
        ref[1,1,2,1]=(1.23290517672)/(o+1.)+(0.323935107053)
        ref[1,2,0,0]=(0.825483086179)/(o+1.)+(0.0301218157877)
        ref[1,2,0,1]=(-0.0793934768662)/(o+1.)+(-0.142521007854)
        ref[1,2,1,0]=(-0.0709650957035)/(o+1.)+(-0.874854727116)
        ref[1,2,1,1]=(1.10714952407)/(o+1.)+(-0.208486520392)
        ref[1,2,2,0]=(0.99013461187)/(o+1.)+(0.359768530582)
        ref[1,2,2,1]=(0.0356627876162)/(o+1.)+(0.488634391872)
        ref[1,3,0,0]=(-0.474442326803)/(o+1.)+(-0.499450411015)
        ref[1,3,0,1]=(-0.673232992552)/(o+1.)+(-1.16158008694)
        ref[1,3,1,0]=(0.411652221393)/(o+1.)+(0.966457956295)
        ref[1,3,1,1]=(-1.76698398732)/(o+1.)+(-0.559250216883)
        ref[1,3,2,0]=(1.9874048718)/(o+1.)+(-0.740838072)
        ref[1,3,2,1]=(0.661784629796)/(o+1.)+(0.216931076303)
        ref[1,4,0,0]=(-0.513029515019)/(o+1.)+(0.0818103062045)
        ref[1,4,0,1]=(0.282755142694)/(o+1.)+(-0.261370981611)
        ref[1,4,1,0]=(-1.96973437805)/(o+1.)+(-0.478746163042)
        ref[1,4,1,1]=(0.908774952181)/(o+1.)+(0.0465858502998)
        ref[1,4,2,0]=(1.3422757017)/(o+1.)+(-0.0996743832723)
        ref[1,4,2,1]=(1.21609351443)/(o+1.)+(-0.349173948065)
        ref[2,0,0,0]=(-1.01744592956)/(o+1.)+(-0.923588464944)
        ref[2,0,0,1]=(1.10931736222)/(o+1.)+(0.224335478673)
        ref[2,0,1,0]=(-0.416479828321)/(o+1.)+(-0.285640270869)
        ref[2,0,1,1]=(-1.5671174369)/(o+1.)+(-0.467268571897)
        ref[2,0,2,0]=(-0.274214960025)/(o+1.)+(0.163118815513)
        ref[2,0,2,1]=(0.000680528469817)/(o+1.)+(0.0788535606642)
        ref[2,1,0,0]=(-0.844192475139)/(o+1.)+(-0.262252149651)
        ref[2,1,0,1]=(0.69399848864)/(o+1.)+(-0.395963008628)
        ref[2,1,1,0]=(-1.37742489692)/(o+1.)+(-0.128468449379)
        ref[2,1,1,1]=(0.282095133276)/(o+1.)+(-0.641487907195)
        ref[2,1,2,0]=(-1.4727375549)/(o+1.)+(0.343315122764)
        ref[2,1,2,1]=(0.67637649169)/(o+1.)+(0.0439030384195)
        ref[2,2,0,0]=(1.58598306034)/(o+1.)+(0.257398552088)
        ref[2,2,0,1]=(-0.187654980952)/(o+1.)+(0.677607655001)
        ref[2,2,1,0]=(0.225457027395)/(o+1.)+(0.221493863079)
        ref[2,2,1,1]=(-1.66885547416)/(o+1.)+(-0.372038995202)
        ref[2,2,2,0]=(-1.42708510165)/(o+1.)+(0.410904476533)
        ref[2,2,2,1]=(0.352696956576)/(o+1.)+(-0.252899306452)
        ref[2,3,0,0]=(0.266365999523)/(o+1.)+(0.249791335977)
        ref[2,3,0,1]=(0.926084735451)/(o+1.)+(0.0271701662571)
        ref[2,3,1,0]=(-0.138787407751)/(o+1.)+(-0.652714809939)
        ref[2,3,1,1]=(-0.834750293031)/(o+1.)+(-0.112504360161)
        ref[2,3,2,0]=(-0.806210182586)/(o+1.)+(0.415200684382)
        ref[2,3,2,1]=(0.577899253552)/(o+1.)+(-0.913828206984)
        ref[2,4,0,0]=(1.38802720612)/(o+1.)+(0.755900031203)
        ref[2,4,0,1]=(-1.33634216788)/(o+1.)+(0.504253453293)
        ref[2,4,1,0]=(-1.13393506986)/(o+1.)+(0.195255119392)
        ref[2,4,1,1]=(-0.384732222354)/(o+1.)+(0.203237915028)
        ref[2,4,2,0]=(-1.18866097183)/(o+1.)+(-0.122726302784)
        ref[2,4,2,1]=(-1.41914128138)/(o+1.)+(-0.382918250045)
        ref[3,0,0,0]=(-1.3617379499)/(o+1.)+(0.916513859836)
        ref[3,0,0,1]=(-0.577580679368)/(o+1.)+(-0.580633018955)
        ref[3,0,1,0]=(-0.730926522369)/(o+1.)+(-0.50988617222)
        ref[3,0,1,1]=(0.852352706539)/(o+1.)+(-0.314072812402)
        ref[3,0,2,0]=(0.63914695576)/(o+1.)+(-0.155678014016)
        ref[3,0,2,1]=(-2.31555597527)/(o+1.)+(-0.593748666755)
        ref[3,1,0,0]=(-0.31308710102)/(o+1.)+(-0.011689493559)
        ref[3,1,0,1]=(0.565049631081)/(o+1.)+(-0.115943082793)
        ref[3,1,1,0]=(1.76264026827)/(o+1.)+(1.35490058502)
        ref[3,1,1,1]=(1.34557111025)/(o+1.)+(0.873578306883)
        ref[3,1,2,0]=(0.924587947935)/(o+1.)+(-0.626617402844)
        ref[3,1,2,1]=(-0.679695689354)/(o+1.)+(0.143800073503)
        ref[3,2,0,0]=(-0.408824291494)/(o+1.)+(-0.330341171766)
        ref[3,2,0,1]=(1.02728720579)/(o+1.)+(0.860379917667)
        ref[3,2,1,0]=(-1.99227170213)/(o+1.)+(-0.251611418656)
        ref[3,2,1,1]=(-0.844161339595)/(o+1.)+(0.159252406352)
        ref[3,2,2,0]=(1.52009831693)/(o+1.)+(-0.519116196235)
        ref[3,2,2,1]=(0.323553722724)/(o+1.)+(-1.02888651238)
        ref[3,3,0,0]=(-1.39647666034)/(o+1.)+(0.637724793483)
        ref[3,3,0,1]=(0.960901654854)/(o+1.)+(-0.00807626431577)
        ref[3,3,1,0]=(0.383998350003)/(o+1.)+(-0.154799534246)
        ref[3,3,1,1]=(-0.851062083214)/(o+1.)+(0.752100475729)
        ref[3,3,2,0]=(0.788418151505)/(o+1.)+(-0.614906659263)
        ref[3,3,2,1]=(-0.1028922667)/(o+1.)+(0.135563961629)
        ref[3,4,0,0]=(-0.540240088251)/(o+1.)+(-0.425382378277)
        ref[3,4,0,1]=(0.236730341461)/(o+1.)+(0.979828209623)
        ref[3,4,1,0]=(-0.9880898216)/(o+1.)+(-0.198536176202)
        ref[3,4,1,1]=(1.06221574228)/(o+1.)+(-0.46475456894)
        ref[3,4,2,0]=(-1.89471538439)/(o+1.)+(-0.0317829587581)
        ref[3,4,2,1]=(-1.20699948566)/(o+1.)+(0.907108895566)
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.107044842683)*x[0]**o+(-0.198407800146)*x[0]+(0.877472712926)*x[1]**o+(-0.647405831691)*x[1]
        ref=(0.770427870243)/(o+1.)+(-0.422906815918)
      else:
        s=(-0.559912240782)*x[0]**o+(0.766296134438)*x[0]+(-0.110980192077)*x[1]**o+(-0.608865520099)*x[1]+(0.672667086497)*x[2]**o+(0.334463167521)*x[2]
        ref=(0.00177465363795)/(o+1.)+(0.24594689093)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.416213605972)*x[0]**o+(0.77813592218)*x[0]+(-0.222437793488)*x[1]**o+(-0.933738289486)*x[1]
        s[1]=(0.455257144475)*x[0]**o+(-0.622417334686)*x[0]+(0.806205896623)*x[1]**o+(0.675691766701)*x[1]
        ref[0]=(-0.638651399461)/(o+1.)+(-0.0778011836529)
        ref[1]=(1.2614630411)/(o+1.)+(0.0266372160075)
      else:
        s[0]=(0.753437228142)*x[0]**o+(0.148952810054)*x[0]+(0.179818865036)*x[1]**o+(0.490547213201)*x[1]+(0.977606414203)*x[2]**o+(0.0928058817324)*x[2]
        s[1]=(0.964988411956)*x[0]**o+(0.446937048611)*x[0]+(-0.110739634254)*x[1]**o+(0.906002899971)*x[1]+(0.666451027663)*x[2]**o+(0.270230684952)*x[2]
        ref[0]=(1.91086250738)/(o+1.)+(0.366152952493)
        ref[1]=(1.52069980537)/(o+1.)+(0.811585316767)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.08214321783)*x[0]**o+(-0.496010136907)*x[0]+(-0.661161523248)*x[1]**o+(-0.151256978215)*x[1]
        s[0,1]=(0.00779835697637)*x[0]**o+(0.307130297185)*x[0]+(-0.111598029557)*x[1]**o+(-0.00186165192551)*x[1]
        s[0,2]=(0.0978324862172)*x[0]**o+(-0.746313762674)*x[0]+(0.519229413642)*x[1]**o+(0.892797139942)*x[1]
        s[0,3]=(-0.598407415713)*x[0]**o+(0.801307410865)*x[0]+(0.894609711877)*x[1]**o+(0.0520830645255)*x[1]
        s[0,4]=(0.580674315696)*x[0]**o+(-0.000690290928564)*x[0]+(0.909550857813)*x[1]**o+(-0.766950054968)*x[1]
        s[1,0]=(0.293745692233)*x[0]**o+(0.581206623889)*x[0]+(0.154848883067)*x[1]**o+(0.260519542891)*x[1]
        s[1,1]=(0.81822409326)*x[0]**o+(0.0527752269091)*x[0]+(0.960312253084)*x[1]**o+(-0.844575939421)*x[1]
        s[1,2]=(-0.186016521867)*x[0]**o+(0.316565783375)*x[0]+(-0.206256264607)*x[1]**o+(0.951167349262)*x[1]
        s[1,3]=(-0.355700637442)*x[0]**o+(-0.636455488757)*x[0]+(0.623605194008)*x[1]**o+(-0.0426220883099)*x[1]
        s[1,4]=(0.72832697443)*x[0]**o+(-0.349553513535)*x[0]+(-0.128283294734)*x[1]**o+(-0.280029190018)*x[1]
        s[2,0]=(0.634737331487)*x[0]**o+(-0.0784762897091)*x[0]+(0.49107451702)*x[1]**o+(-0.927830147754)*x[1]
        s[2,1]=(-0.561516427218)*x[0]**o+(0.414180445476)*x[0]+(-0.32846350728)*x[1]**o+(0.653302722231)*x[1]
        s[2,2]=(-0.285665357177)*x[0]**o+(-0.495353890909)*x[0]+(-0.300417232522)*x[1]**o+(0.828720486127)*x[1]
        s[2,3]=(0.582282364197)*x[0]**o+(-0.491365235361)*x[0]+(-0.481532775048)*x[1]**o+(0.843340014315)*x[1]
        s[2,4]=(0.467528444795)*x[0]**o+(-0.617430509987)*x[0]+(0.0370576618542)*x[1]**o+(0.124933927604)*x[1]
        s[3,0]=(0.76865308901)*x[0]**o+(0.270205155941)*x[0]+(0.239796830473)*x[1]**o+(-0.732593342574)*x[1]
        s[3,1]=(0.276346458614)*x[0]**o+(-0.637344640378)*x[0]+(-0.697338900523)*x[1]**o+(0.136084835992)*x[1]
        s[3,2]=(-0.932808156816)*x[0]**o+(0.296855273797)*x[0]+(-0.417934077535)*x[1]**o+(0.897857223334)*x[1]
        s[3,3]=(0.358813012298)*x[0]**o+(-0.421306401815)*x[0]+(0.682226516534)*x[1]**o+(-0.0533913429448)*x[1]
        s[3,4]=(0.27862707257)*x[0]**o+(-0.227499516484)*x[0]+(0.502409049994)*x[1]**o+(0.348218429657)*x[1]
        ref[0,0]=(-0.579018305418)/(o+1.)+(-0.323633557561)
        ref[0,1]=(-0.103799672581)/(o+1.)+(0.15263432263)
        ref[0,2]=(0.617061899859)/(o+1.)+(0.0732416886339)
        ref[0,3]=(0.296202296164)/(o+1.)+(0.426695237695)
        ref[0,4]=(1.49022517351)/(o+1.)+(-0.383820172948)
        ref[1,0]=(0.4485945753)/(o+1.)+(0.42086308339)
        ref[1,1]=(1.77853634634)/(o+1.)+(-0.395900356256)
        ref[1,2]=(-0.392272786474)/(o+1.)+(0.633866566319)
        ref[1,3]=(0.267904556566)/(o+1.)+(-0.339538788533)
        ref[1,4]=(0.600043679696)/(o+1.)+(-0.314791351776)
        ref[2,0]=(1.12581184851)/(o+1.)+(-0.503153218732)
        ref[2,1]=(-0.889979934498)/(o+1.)+(0.533741583854)
        ref[2,2]=(-0.586082589699)/(o+1.)+(0.166683297609)
        ref[2,3]=(0.100749589149)/(o+1.)+(0.175987389477)
        ref[2,4]=(0.504586106649)/(o+1.)+(-0.246248291192)
        ref[3,0]=(1.00844991948)/(o+1.)+(-0.231194093316)
        ref[3,1]=(-0.420992441909)/(o+1.)+(-0.250629902193)
        ref[3,2]=(-1.35074223435)/(o+1.)+(0.597356248566)
        ref[3,3]=(1.04103952883)/(o+1.)+(-0.23734887238)
        ref[3,4]=(0.781036122564)/(o+1.)+(0.0603594565866)
      else:
        s[0,0]=(-0.941295734032)*x[0]**o+(0.396841358423)*x[0]+(-0.490021992176)*x[1]**o+(-0.346405280728)*x[1]+(-0.201846160921)*x[2]**o+(0.20749764579)*x[2]
        s[0,1]=(-0.88981448497)*x[0]**o+(-0.884955504659)*x[0]+(0.182259548742)*x[1]**o+(-0.235489618403)*x[1]+(-0.977189329039)*x[2]**o+(-0.514950614838)*x[2]
        s[0,2]=(0.328187150969)*x[0]**o+(-0.448497961254)*x[0]+(0.472389791923)*x[1]**o+(-0.624930304822)*x[1]+(0.265474988322)*x[2]**o+(-0.0214969683756)*x[2]
        s[0,3]=(-0.735757404485)*x[0]**o+(-0.92658492028)*x[0]+(0.684533158897)*x[1]**o+(-0.865066520203)*x[1]+(-0.190890639018)*x[2]**o+(-0.529642573907)*x[2]
        s[0,4]=(0.955420842735)*x[0]**o+(0.773074765628)*x[0]+(-0.382656468864)*x[1]**o+(0.47181739927)*x[1]+(0.873841536673)*x[2]**o+(-0.526372620067)*x[2]
        s[1,0]=(0.770067221954)*x[0]**o+(-0.406438793272)*x[0]+(0.51351865241)*x[1]**o+(-0.856834133418)*x[1]+(0.589461089026)*x[2]**o+(-0.141812408947)*x[2]
        s[1,1]=(-0.812514809415)*x[0]**o+(-0.456460678755)*x[0]+(0.430628764669)*x[1]**o+(0.262068556418)*x[1]+(0.597424763719)*x[2]**o+(-0.157908736594)*x[2]
        s[1,2]=(-0.46727482012)*x[0]**o+(-0.744512472884)*x[0]+(-0.365664650476)*x[1]**o+(-0.361046498637)*x[1]+(-0.166284258648)*x[2]**o+(0.528262791587)*x[2]
        s[1,3]=(-0.273755634801)*x[0]**o+(-0.15873969342)*x[0]+(-0.368357879665)*x[1]**o+(0.283273650468)*x[1]+(-0.0175031061849)*x[2]**o+(-0.0417868805286)*x[2]
        s[1,4]=(0.349815667052)*x[0]**o+(-0.175629838767)*x[0]+(0.272305468439)*x[1]**o+(-0.750728619331)*x[1]+(0.0723037064016)*x[2]**o+(0.911924754086)*x[2]
        s[2,0]=(0.834136592552)*x[0]**o+(0.227306622837)*x[0]+(-0.468357820454)*x[1]**o+(0.0429839240965)*x[1]+(0.397233264579)*x[2]**o+(0.645297775174)*x[2]
        s[2,1]=(-0.420265217156)*x[0]**o+(-0.762906499224)*x[0]+(0.16960694356)*x[1]**o+(0.381251517238)*x[1]+(-0.740637261618)*x[2]**o+(-0.437283161876)*x[2]
        s[2,2]=(-0.957535003492)*x[0]**o+(0.546534458865)*x[0]+(0.10634355782)*x[1]**o+(0.0575530053655)*x[1]+(0.395304821296)*x[2]**o+(0.0507237509151)*x[2]
        s[2,3]=(0.130545158576)*x[0]**o+(-0.297688721316)*x[0]+(-0.193136907135)*x[1]**o+(-0.530687107912)*x[1]+(0.891112852449)*x[2]**o+(0.89025257856)*x[2]
        s[2,4]=(-0.452227707094)*x[0]**o+(0.518745990389)*x[0]+(-0.848908958687)*x[1]**o+(-0.984222753382)*x[1]+(-0.725703746452)*x[2]**o+(0.29370062408)*x[2]
        s[3,0]=(0.451818718553)*x[0]**o+(0.15228149344)*x[0]+(0.309329014013)*x[1]**o+(-0.629074550954)*x[1]+(0.925898897131)*x[2]**o+(-0.530392928354)*x[2]
        s[3,1]=(0.48661045622)*x[0]**o+(0.362115939707)*x[0]+(0.461789600557)*x[1]**o+(-0.0265394801756)*x[1]+(-0.52060582127)*x[2]**o+(0.847046142749)*x[2]
        s[3,2]=(0.356328974486)*x[0]**o+(0.735492369055)*x[0]+(0.833634465835)*x[1]**o+(-0.603533125221)*x[1]+(-0.799865132292)*x[2]**o+(0.255351535539)*x[2]
        s[3,3]=(0.821371453867)*x[0]**o+(0.360956876043)*x[0]+(0.0191507582216)*x[1]**o+(-0.995259844736)*x[1]+(-0.00388518722366)*x[2]**o+(-0.0360171649464)*x[2]
        s[3,4]=(-0.248806545065)*x[0]**o+(-0.598745625081)*x[0]+(0.722047538053)*x[1]**o+(-0.840877899002)*x[1]+(0.920023026949)*x[2]**o+(0.503521861458)*x[2]
        ref[0,0]=(-1.63316388713)/(o+1.)+(0.128966861742)
        ref[0,1]=(-1.68474426527)/(o+1.)+(-0.81769786895)
        ref[0,2]=(1.06605193121)/(o+1.)+(-0.547462617226)
        ref[0,3]=(-0.242114884606)/(o+1.)+(-1.16064700719)
        ref[0,4]=(1.44660591054)/(o+1.)+(0.359259772415)
        ref[1,0]=(1.87304696339)/(o+1.)+(-0.702542667819)
        ref[1,1]=(0.215538718973)/(o+1.)+(-0.176150429465)
        ref[1,2]=(-0.999223729244)/(o+1.)+(-0.288648089967)
        ref[1,3]=(-0.659616620651)/(o+1.)+(0.04137353826)
        ref[1,4]=(0.694424841892)/(o+1.)+(-0.00721685200611)
        ref[2,0]=(0.763012036677)/(o+1.)+(0.457794161054)
        ref[2,1]=(-0.991295535214)/(o+1.)+(-0.409469071931)
        ref[2,2]=(-0.455886624375)/(o+1.)+(0.327405607573)
        ref[2,3]=(0.82852110389)/(o+1.)+(0.030938374666)
        ref[2,4]=(-2.02684041223)/(o+1.)+(-0.0858880694563)
        ref[3,0]=(1.6870466297)/(o+1.)+(-0.503592992934)
        ref[3,1]=(0.427794235507)/(o+1.)+(0.59131130114)
        ref[3,2]=(0.39009830803)/(o+1.)+(0.193655389686)
        ref[3,3]=(0.836637024865)/(o+1.)+(-0.33516006682)
        ref[3,4]=(1.39326401994)/(o+1.)+(-0.468050831312)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(-0.944773955793)*x[0]**o+(-0.662996681032)*x[0]+(-0.442867713759)*x[1]**o+(-0.229023301726)*x[1]
        s[0,0,1]=(0.0407144183428)*x[0]**o+(-0.669891322439)*x[0]+(0.920328407938)*x[1]**o+(-0.320681600926)*x[1]
        s[0,1,0]=(0.179693170721)*x[0]**o+(-0.671664515028)*x[0]+(0.525108213743)*x[1]**o+(0.310882938487)*x[1]
        s[0,1,1]=(-0.0540671676637)*x[0]**o+(0.000853108264254)*x[0]+(-0.370081882537)*x[1]**o+(-0.138484250063)*x[1]
        s[1,0,0]=(0.631472269828)*x[0]**o+(-0.294848092138)*x[0]+(0.0730087097944)*x[1]**o+(0.0292051228565)*x[1]
        s[1,0,1]=(-0.576009492093)*x[0]**o+(-0.228105833611)*x[0]+(-0.0563734559287)*x[1]**o+(-0.242583876037)*x[1]
        s[1,1,0]=(-0.597734903158)*x[0]**o+(0.960078413053)*x[0]+(0.881951380528)*x[1]**o+(-0.186121672274)*x[1]
        s[1,1,1]=(-0.117005971726)*x[0]**o+(-0.809886380446)*x[0]+(-0.981092676193)*x[1]**o+(0.0636030759105)*x[1]
        s[2,0,0]=(-0.197553649616)*x[0]**o+(-0.820287165629)*x[0]+(0.359670409851)*x[1]**o+(-0.861222386581)*x[1]
        s[2,0,1]=(0.968998274457)*x[0]**o+(0.376230063695)*x[0]+(0.283097885679)*x[1]**o+(-0.128262765246)*x[1]
        s[2,1,0]=(0.0310410423851)*x[0]**o+(0.611929570715)*x[0]+(-0.972311035606)*x[1]**o+(-0.568277782449)*x[1]
        s[2,1,1]=(0.95074765381)*x[0]**o+(0.338991372672)*x[0]+(0.264103717561)*x[1]**o+(0.370834284282)*x[1]
        s[3,0,0]=(-0.664867197259)*x[0]**o+(0.917684307761)*x[0]+(-0.319774527149)*x[1]**o+(0.432915185753)*x[1]
        s[3,0,1]=(-0.162944024583)*x[0]**o+(-0.123907346782)*x[0]+(0.325356901629)*x[1]**o+(0.90249028492)*x[1]
        s[3,1,0]=(-0.135439910119)*x[0]**o+(0.920226918882)*x[0]+(0.112667508952)*x[1]**o+(0.710942378416)*x[1]
        s[3,1,1]=(-0.370025988485)*x[0]**o+(0.146841174972)*x[0]+(-0.209033207026)*x[1]**o+(0.270439136519)*x[1]
        s[4,0,0]=(0.724119312565)*x[0]**o+(0.726636721509)*x[0]+(-0.643509558883)*x[1]**o+(-0.544447102654)*x[1]
        s[4,0,1]=(-0.563641108028)*x[0]**o+(-0.320916186803)*x[0]+(0.935527371187)*x[1]**o+(0.535917271795)*x[1]
        s[4,1,0]=(-0.482028335236)*x[0]**o+(-0.510759210235)*x[0]+(-0.105477148218)*x[1]**o+(0.54069501931)*x[1]
        s[4,1,1]=(0.305620947599)*x[0]**o+(-0.724738374933)*x[0]+(0.470881233144)*x[1]**o+(-0.705410563798)*x[1]
        s[5,0,0]=(0.168173718235)*x[0]**o+(0.578831386242)*x[0]+(-0.869711732793)*x[1]**o+(-0.176143888818)*x[1]
        s[5,0,1]=(-0.228160489841)*x[0]**o+(-0.0297036971933)*x[0]+(0.391702396499)*x[1]**o+(0.875713079041)*x[1]
        s[5,1,0]=(-0.910154283353)*x[0]**o+(-0.0651085103088)*x[0]+(0.0884987382756)*x[1]**o+(-0.937409873275)*x[1]
        s[5,1,1]=(0.41106472762)*x[0]**o+(-0.801661249269)*x[0]+(0.201137556319)*x[1]**o+(-0.665423718838)*x[1]
        ref[0,0,0]=(-1.38764166955)/(o+1.)+(-0.446009991379)
        ref[0,0,1]=(0.961042826281)/(o+1.)+(-0.495286461682)
        ref[0,1,0]=(0.704801384464)/(o+1.)+(-0.180390788271)
        ref[0,1,1]=(-0.424149050201)/(o+1.)+(-0.0688155708996)
        ref[1,0,0]=(0.704480979622)/(o+1.)+(-0.132821484641)
        ref[1,0,1]=(-0.632382948021)/(o+1.)+(-0.235344854824)
        ref[1,1,0]=(0.28421647737)/(o+1.)+(0.386978370389)
        ref[1,1,1]=(-1.09809864792)/(o+1.)+(-0.373141652268)
        ref[2,0,0]=(0.162116760235)/(o+1.)+(-0.840754776105)
        ref[2,0,1]=(1.25209616014)/(o+1.)+(0.123983649224)
        ref[2,1,0]=(-0.941269993221)/(o+1.)+(0.0218258941331)
        ref[2,1,1]=(1.21485137137)/(o+1.)+(0.354912828477)
        ref[3,0,0]=(-0.984641724408)/(o+1.)+(0.675299746757)
        ref[3,0,1]=(0.162412877046)/(o+1.)+(0.389291469069)
        ref[3,1,0]=(-0.0227724011669)/(o+1.)+(0.815584648649)
        ref[3,1,1]=(-0.579059195511)/(o+1.)+(0.208640155746)
        ref[4,0,0]=(0.0806097536818)/(o+1.)+(0.0910948094271)
        ref[4,0,1]=(0.371886263159)/(o+1.)+(0.107500542496)
        ref[4,1,0]=(-0.587505483453)/(o+1.)+(0.0149679045376)
        ref[4,1,1]=(0.776502180743)/(o+1.)+(-0.715074469366)
        ref[5,0,0]=(-0.701538014558)/(o+1.)+(0.201343748712)
        ref[5,0,1]=(0.163541906658)/(o+1.)+(0.423004690924)
        ref[5,1,0]=(-0.821655545078)/(o+1.)+(-0.501259191792)
        ref[5,1,1]=(0.612202283939)/(o+1.)+(-0.733542484054)
      else:
        s[0,0,0]=(-0.0829784527698)*x[0]**o+(-0.332549632616)*x[0]+(0.255962043751)*x[1]**o+(-0.176352635053)*x[1]+(-0.954055133743)*x[2]**o+(-0.898668627164)*x[2]
        s[0,0,1]=(-0.0566639977041)*x[0]**o+(0.64899823634)*x[0]+(0.339703741475)*x[1]**o+(-0.754613713569)*x[1]+(0.379536135428)*x[2]**o+(0.118057489576)*x[2]
        s[0,1,0]=(-0.718453260796)*x[0]**o+(0.997156187689)*x[0]+(0.914514642799)*x[1]**o+(-0.633873909376)*x[1]+(-0.485691739694)*x[2]**o+(0.415370908143)*x[2]
        s[0,1,1]=(0.553256681553)*x[0]**o+(-0.191413319548)*x[0]+(0.800894762679)*x[1]**o+(0.919882584486)*x[1]+(0.127165880997)*x[2]**o+(-0.690488492514)*x[2]
        s[1,0,0]=(-0.142765560936)*x[0]**o+(-0.303407652797)*x[0]+(0.989394601479)*x[1]**o+(-0.484741465949)*x[1]+(0.732648062756)*x[2]**o+(-0.678857629867)*x[2]
        s[1,0,1]=(0.288139139594)*x[0]**o+(-0.534820162236)*x[0]+(0.554980676789)*x[1]**o+(-0.917630291921)*x[1]+(0.170042991043)*x[2]**o+(0.826809263768)*x[2]
        s[1,1,0]=(-0.901850450639)*x[0]**o+(-0.0365789634899)*x[0]+(0.0622336448434)*x[1]**o+(0.22005965603)*x[1]+(0.857183927574)*x[2]**o+(0.254453636804)*x[2]
        s[1,1,1]=(-0.888020161965)*x[0]**o+(-0.0741224453822)*x[0]+(0.887224038781)*x[1]**o+(0.742137909151)*x[1]+(-0.587100789782)*x[2]**o+(-0.928415179888)*x[2]
        s[2,0,0]=(0.00216356197373)*x[0]**o+(0.248793770338)*x[0]+(0.518030493147)*x[1]**o+(-0.508299580849)*x[1]+(-0.0290711721801)*x[2]**o+(0.662305993163)*x[2]
        s[2,0,1]=(0.924293736923)*x[0]**o+(0.218836928692)*x[0]+(-0.270314297479)*x[1]**o+(0.20290032209)*x[1]+(-0.87536943379)*x[2]**o+(-0.864192540105)*x[2]
        s[2,1,0]=(0.416159016379)*x[0]**o+(-0.30257442476)*x[0]+(-0.257298173268)*x[1]**o+(-0.398564079155)*x[1]+(-0.0699018670952)*x[2]**o+(-0.540145396091)*x[2]
        s[2,1,1]=(0.82076420765)*x[0]**o+(-0.265319227281)*x[0]+(-0.156971036756)*x[1]**o+(-0.468527020278)*x[1]+(0.2455572117)*x[2]**o+(0.237918192218)*x[2]
        s[3,0,0]=(-0.645207425303)*x[0]**o+(0.992013339436)*x[0]+(-0.814115520055)*x[1]**o+(-0.0863699557395)*x[1]+(0.356684159665)*x[2]**o+(0.0215161017879)*x[2]
        s[3,0,1]=(0.429124334865)*x[0]**o+(-0.516632695494)*x[0]+(0.276076705271)*x[1]**o+(0.73233754297)*x[1]+(-0.499960919138)*x[2]**o+(0.927448433933)*x[2]
        s[3,1,0]=(0.105016655639)*x[0]**o+(0.627563943644)*x[0]+(0.0134384788636)*x[1]**o+(-0.491675197932)*x[1]+(0.850898525327)*x[2]**o+(-0.153767354797)*x[2]
        s[3,1,1]=(0.0927473537694)*x[0]**o+(-0.164202910181)*x[0]+(-0.246744412752)*x[1]**o+(0.812782310771)*x[1]+(0.452025251957)*x[2]**o+(-0.46155534584)*x[2]
        s[4,0,0]=(0.466686843553)*x[0]**o+(-0.454360335871)*x[0]+(0.600983402349)*x[1]**o+(-0.248407743104)*x[1]+(-0.114458429866)*x[2]**o+(0.363120759629)*x[2]
        s[4,0,1]=(0.108248019353)*x[0]**o+(-0.386567024916)*x[0]+(0.589911190112)*x[1]**o+(0.974355188979)*x[1]+(0.956434046018)*x[2]**o+(-0.750308948178)*x[2]
        s[4,1,0]=(-0.0575495701591)*x[0]**o+(-0.297079422559)*x[0]+(0.173709916352)*x[1]**o+(-0.703203763304)*x[1]+(-0.206920064792)*x[2]**o+(-0.566180018312)*x[2]
        s[4,1,1]=(0.752722443401)*x[0]**o+(0.287487922131)*x[0]+(-0.253690778509)*x[1]**o+(0.158399655258)*x[1]+(0.365133322462)*x[2]**o+(-0.765340889542)*x[2]
        s[5,0,0]=(-0.249224479124)*x[0]**o+(-0.81760956491)*x[0]+(0.664454285698)*x[1]**o+(-0.111438607721)*x[1]+(-0.34058205464)*x[2]**o+(-0.604036230847)*x[2]
        s[5,0,1]=(-0.0489505541715)*x[0]**o+(-0.216719586033)*x[0]+(0.862292178064)*x[1]**o+(-0.978934839395)*x[1]+(-0.229118015068)*x[2]**o+(-0.220547177874)*x[2]
        s[5,1,0]=(-0.200117801874)*x[0]**o+(0.826116119716)*x[0]+(0.180058706489)*x[1]**o+(0.938844014751)*x[1]+(0.232853505918)*x[2]**o+(-0.0718472890928)*x[2]
        s[5,1,1]=(0.335820727851)*x[0]**o+(-0.30183416972)*x[0]+(-0.69421482762)*x[1]**o+(0.105894103212)*x[1]+(0.620400009549)*x[2]**o+(0.970029758381)*x[2]
        ref[0,0,0]=(-0.781071542761)/(o+1.)+(-0.703785447417)
        ref[0,0,1]=(0.662575879198)/(o+1.)+(0.00622100617352)
        ref[0,1,0]=(-0.289630357691)/(o+1.)+(0.389326593228)
        ref[0,1,1]=(1.48131732523)/(o+1.)+(0.0189903862116)
        ref[1,0,0]=(1.5792771033)/(o+1.)+(-0.733503374307)
        ref[1,0,1]=(1.01316280743)/(o+1.)+(-0.312820595194)
        ref[1,1,0]=(0.0175671217784)/(o+1.)+(0.218967164672)
        ref[1,1,1]=(-0.587896912967)/(o+1.)+(-0.13019985806)
        ref[2,0,0]=(0.49112288294)/(o+1.)+(0.201400091326)
        ref[2,0,1]=(-0.221389994346)/(o+1.)+(-0.221227644661)
        ref[2,1,0]=(0.0889589760167)/(o+1.)+(-0.620641950003)
        ref[2,1,1]=(0.909350382595)/(o+1.)+(-0.24796402767)
        ref[3,0,0]=(-1.10263878569)/(o+1.)+(0.463579742742)
        ref[3,0,1]=(0.205240120999)/(o+1.)+(0.571576640705)
        ref[3,1,0]=(0.96935365983)/(o+1.)+(-0.00893930454231)
        ref[3,1,1]=(0.298028192975)/(o+1.)+(0.0935120273748)
        ref[4,0,0]=(0.953211816036)/(o+1.)+(-0.169823659673)
        ref[4,0,1]=(1.65459325548)/(o+1.)+(-0.0812603920577)
        ref[4,1,0]=(-0.0907597185993)/(o+1.)+(-0.783231602088)
        ref[4,1,1]=(0.864164987355)/(o+1.)+(-0.159726656077)
        ref[5,0,0]=(0.0746477519342)/(o+1.)+(-0.766542201739)
        ref[5,0,1]=(0.584223608825)/(o+1.)+(-0.708100801651)
        ref[5,1,0]=(0.212794410532)/(o+1.)+(0.846556422687)
        ref[5,1,1]=(0.262005909781)/(o+1.)+(0.387044845936)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunction_fromSymbol_ReducedFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunction

      assumptions: ReducedFunction(self.domain) exists
                   self.domain supports integral on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(0.64587181131)*x[0]**o+(-0.142518178492)*x[0]+(-0.399846110498)*x[1]**o+(0.00236392286301)*x[1]
        s[0,0,0,1]=(0.298109366098)*x[0]**o+(-0.0843450895271)*x[0]+(0.574794599006)*x[1]**o+(-0.107008590623)*x[1]
        s[0,0,1,0]=(-0.424486704098)*x[0]**o+(-0.382352426027)*x[0]+(0.275387919625)*x[1]**o+(-0.161015391255)*x[1]
        s[0,0,1,1]=(0.643084919048)*x[0]**o+(0.44303782474)*x[0]+(0.17010706601)*x[1]**o+(-0.160095051974)*x[1]
        s[0,0,2,0]=(0.255268276945)*x[0]**o+(-0.867088236584)*x[0]+(-0.809452452666)*x[1]**o+(-0.376237703808)*x[1]
        s[0,0,2,1]=(-0.773368404022)*x[0]**o+(-0.655491257176)*x[0]+(-0.256178554483)*x[1]**o+(0.582354235979)*x[1]
        s[0,1,0,0]=(0.312387735876)*x[0]**o+(0.36653787868)*x[0]+(-0.160583709945)*x[1]**o+(-0.791133941857)*x[1]
        s[0,1,0,1]=(-0.57693493814)*x[0]**o+(-0.862135922856)*x[0]+(0.0439301419063)*x[1]**o+(0.959289798443)*x[1]
        s[0,1,1,0]=(0.121311324776)*x[0]**o+(0.907585047473)*x[0]+(-0.923686957239)*x[1]**o+(-0.652706978591)*x[1]
        s[0,1,1,1]=(0.682901593615)*x[0]**o+(-0.306481093391)*x[0]+(-0.35714675383)*x[1]**o+(0.500943467097)*x[1]
        s[0,1,2,0]=(0.647130385973)*x[0]**o+(0.873804224811)*x[0]+(0.78779813495)*x[1]**o+(0.430789257821)*x[1]
        s[0,1,2,1]=(0.135833683749)*x[0]**o+(0.32862510689)*x[0]+(-0.48350127573)*x[1]**o+(-0.649354230648)*x[1]
        s[0,2,0,0]=(-0.641045806017)*x[0]**o+(-0.654315036977)*x[0]+(-0.729951508044)*x[1]**o+(0.766934507787)*x[1]
        s[0,2,0,1]=(0.481734616023)*x[0]**o+(-0.687904467365)*x[0]+(0.630779635862)*x[1]**o+(-0.982134985324)*x[1]
        s[0,2,1,0]=(-0.0488969750979)*x[0]**o+(-0.804550194421)*x[0]+(-0.721206215475)*x[1]**o+(0.184754632486)*x[1]
        s[0,2,1,1]=(-0.0554773461175)*x[0]**o+(0.501447670577)*x[0]+(0.214749897505)*x[1]**o+(0.912271898356)*x[1]
        s[0,2,2,0]=(0.0902819474814)*x[0]**o+(0.182092397318)*x[0]+(-0.48175183483)*x[1]**o+(-0.598040724808)*x[1]
        s[0,2,2,1]=(0.720795079983)*x[0]**o+(-0.339928403247)*x[0]+(-0.762363891559)*x[1]**o+(0.460915902605)*x[1]
        s[0,3,0,0]=(0.508573641682)*x[0]**o+(-0.015037165373)*x[0]+(0.448986097944)*x[1]**o+(0.471102721761)*x[1]
        s[0,3,0,1]=(-0.341905079065)*x[0]**o+(-0.95312476241)*x[0]+(-0.749402955548)*x[1]**o+(-0.372596238803)*x[1]
        s[0,3,1,0]=(-0.0690403122223)*x[0]**o+(-0.08921421318)*x[0]+(-0.380821740415)*x[1]**o+(-0.599974433304)*x[1]
        s[0,3,1,1]=(-0.50179318754)*x[0]**o+(0.355355851326)*x[0]+(0.721120108213)*x[1]**o+(0.454667741643)*x[1]
        s[0,3,2,0]=(0.213191653894)*x[0]**o+(-0.637815219816)*x[0]+(0.0628021542224)*x[1]**o+(-0.284841867085)*x[1]
        s[0,3,2,1]=(0.226011558647)*x[0]**o+(0.926633205695)*x[0]+(-0.801045292364)*x[1]**o+(0.84320696071)*x[1]
        s[0,4,0,0]=(-0.925639600998)*x[0]**o+(-0.743975200606)*x[0]+(0.950987915324)*x[1]**o+(-0.919549720538)*x[1]
        s[0,4,0,1]=(0.295152910157)*x[0]**o+(-0.155400545856)*x[0]+(-0.888657028463)*x[1]**o+(-0.0245586756163)*x[1]
        s[0,4,1,0]=(0.659416651507)*x[0]**o+(-0.0687605790427)*x[0]+(0.287026029448)*x[1]**o+(0.749704220615)*x[1]
        s[0,4,1,1]=(0.195399326633)*x[0]**o+(-0.690030508171)*x[0]+(0.812925291301)*x[1]**o+(-0.288219851462)*x[1]
        s[0,4,2,0]=(0.92210509534)*x[0]**o+(-0.746377740152)*x[0]+(-0.263846201131)*x[1]**o+(0.241927075102)*x[1]
        s[0,4,2,1]=(-0.342776236669)*x[0]**o+(-0.678319456068)*x[0]+(0.533446356713)*x[1]**o+(0.893285310119)*x[1]
        s[1,0,0,0]=(-0.35370685183)*x[0]**o+(-0.649199740924)*x[0]+(-0.284557848052)*x[1]**o+(-0.108918514336)*x[1]
        s[1,0,0,1]=(0.435799676535)*x[0]**o+(-0.560606926167)*x[0]+(0.422473364618)*x[1]**o+(-0.644113034355)*x[1]
        s[1,0,1,0]=(0.371384490035)*x[0]**o+(0.468745463643)*x[0]+(0.824650091185)*x[1]**o+(0.779212641032)*x[1]
        s[1,0,1,1]=(0.566680148615)*x[0]**o+(0.70499851212)*x[0]+(-0.622667334132)*x[1]**o+(0.0793283623173)*x[1]
        s[1,0,2,0]=(-0.137422675839)*x[0]**o+(-0.0828536384824)*x[0]+(0.34029178349)*x[1]**o+(-0.572205817396)*x[1]
        s[1,0,2,1]=(0.849275927844)*x[0]**o+(0.588676957111)*x[0]+(-0.351435931217)*x[1]**o+(-0.985192646644)*x[1]
        s[1,1,0,0]=(-0.832427042757)*x[0]**o+(0.511670148283)*x[0]+(-0.225009791535)*x[1]**o+(-0.726486541188)*x[1]
        s[1,1,0,1]=(-0.370148483489)*x[0]**o+(0.0364804235508)*x[0]+(-0.740077305856)*x[1]**o+(0.123450524426)*x[1]
        s[1,1,1,0]=(0.989855829356)*x[0]**o+(-0.183176166338)*x[0]+(0.0388003238684)*x[1]**o+(-0.0395036587674)*x[1]
        s[1,1,1,1]=(-0.775089474977)*x[0]**o+(-0.976089661518)*x[0]+(0.387965308645)*x[1]**o+(0.712155002639)*x[1]
        s[1,1,2,0]=(0.86011220616)*x[0]**o+(-0.95431213868)*x[0]+(-0.495540857845)*x[1]**o+(-0.701111767974)*x[1]
        s[1,1,2,1]=(0.827371223575)*x[0]**o+(-0.650031455862)*x[0]+(-0.0269816386714)*x[1]**o+(-0.490304255652)*x[1]
        s[1,2,0,0]=(-0.718249056807)*x[0]**o+(-0.812216925478)*x[0]+(0.358363619384)*x[1]**o+(-0.555431408236)*x[1]
        s[1,2,0,1]=(0.502681744448)*x[0]**o+(0.568656095603)*x[0]+(-0.642171078627)*x[1]**o+(-0.678715038904)*x[1]
        s[1,2,1,0]=(-0.321002503528)*x[0]**o+(0.399915822381)*x[0]+(0.378527801988)*x[1]**o+(0.261422806614)*x[1]
        s[1,2,1,1]=(-0.296716183029)*x[0]**o+(0.497650651416)*x[0]+(-0.386922791311)*x[1]**o+(-0.91901564657)*x[1]
        s[1,2,2,0]=(-0.828712534186)*x[0]**o+(-0.0114777410069)*x[0]+(-0.142857241422)*x[1]**o+(0.527573020984)*x[1]
        s[1,2,2,1]=(0.341639667906)*x[0]**o+(0.773780839708)*x[0]+(0.625113762311)*x[1]**o+(0.700528961096)*x[1]
        s[1,3,0,0]=(0.763866858532)*x[0]**o+(0.934384862727)*x[0]+(-0.192635958925)*x[1]**o+(-0.531023677701)*x[1]
        s[1,3,0,1]=(-0.559232654796)*x[0]**o+(-0.381248359127)*x[0]+(0.042953808137)*x[1]**o+(-0.0258385267166)*x[1]
        s[1,3,1,0]=(-0.596069058223)*x[0]**o+(0.611041390676)*x[0]+(0.320050026897)*x[1]**o+(0.31950524284)*x[1]
        s[1,3,1,1]=(0.874401923132)*x[0]**o+(0.995518258803)*x[0]+(-0.535120262939)*x[1]**o+(0.160955202392)*x[1]
        s[1,3,2,0]=(0.352942439272)*x[0]**o+(-0.198121930351)*x[0]+(0.458180780997)*x[1]**o+(-0.415159831726)*x[1]
        s[1,3,2,1]=(-0.511099937481)*x[0]**o+(0.297820678198)*x[0]+(0.908702936015)*x[1]**o+(0.965700845817)*x[1]
        s[1,4,0,0]=(0.484158989021)*x[0]**o+(-0.339968620521)*x[0]+(-0.213631360058)*x[1]**o+(0.0279496812648)*x[1]
        s[1,4,0,1]=(-0.217367899821)*x[0]**o+(0.433458086428)*x[0]+(0.244095485621)*x[1]**o+(-0.210532522829)*x[1]
        s[1,4,1,0]=(-0.254889249994)*x[0]**o+(-0.61248326904)*x[0]+(0.372253508249)*x[1]**o+(-0.642709327921)*x[1]
        s[1,4,1,1]=(0.880474654934)*x[0]**o+(0.68839232824)*x[0]+(0.189353506343)*x[1]**o+(-0.880653838654)*x[1]
        s[1,4,2,0]=(0.894790715592)*x[0]**o+(-0.0486173810442)*x[0]+(-0.246550765358)*x[1]**o+(-0.343955068452)*x[1]
        s[1,4,2,1]=(-0.801075573104)*x[0]**o+(-0.749933700542)*x[0]+(0.527976119218)*x[1]**o+(-0.671158980138)*x[1]
        s[2,0,0,0]=(0.114632100765)*x[0]**o+(0.61717357379)*x[0]+(-0.0549019501391)*x[1]**o+(-0.749508575346)*x[1]
        s[2,0,0,1]=(-0.130970404194)*x[0]**o+(0.707199821591)*x[0]+(-0.903645272692)*x[1]**o+(-0.0334328361989)*x[1]
        s[2,0,1,0]=(-0.715712763894)*x[0]**o+(-0.675427980461)*x[0]+(-0.366134531475)*x[1]**o+(0.201815564551)*x[1]
        s[2,0,1,1]=(-0.269870188604)*x[0]**o+(-0.147598844928)*x[0]+(0.677047514973)*x[1]**o+(-0.675438319339)*x[1]
        s[2,0,2,0]=(0.212062724301)*x[0]**o+(-0.0250868236126)*x[0]+(-0.100922604124)*x[1]**o+(0.427937448863)*x[1]
        s[2,0,2,1]=(-0.666239860886)*x[0]**o+(0.438278520576)*x[0]+(0.544120816825)*x[1]**o+(-0.473903266225)*x[1]
        s[2,1,0,0]=(-0.361166717507)*x[0]**o+(0.901751988241)*x[0]+(0.638709648231)*x[1]**o+(-0.235793145944)*x[1]
        s[2,1,0,1]=(-0.11628481466)*x[0]**o+(-0.0851943604844)*x[0]+(0.832299993845)*x[1]**o+(-0.619236617136)*x[1]
        s[2,1,1,0]=(-0.960053497178)*x[0]**o+(-0.455025261633)*x[0]+(-0.701715648299)*x[1]**o+(0.728639782857)*x[1]
        s[2,1,1,1]=(-0.533289133278)*x[0]**o+(-0.370102308202)*x[0]+(0.99106398096)*x[1]**o+(-0.0538096992941)*x[1]
        s[2,1,2,0]=(0.224553214823)*x[0]**o+(-0.500053815427)*x[0]+(-0.0704009106671)*x[1]**o+(0.0488366166495)*x[1]
        s[2,1,2,1]=(0.337180739675)*x[0]**o+(-0.981874806959)*x[0]+(-0.731568712654)*x[1]**o+(0.0847072795393)*x[1]
        s[2,2,0,0]=(-0.222325355327)*x[0]**o+(-0.276984511347)*x[0]+(-0.948381537292)*x[1]**o+(0.766624524385)*x[1]
        s[2,2,0,1]=(0.938320523187)*x[0]**o+(0.808908551771)*x[0]+(0.765330540739)*x[1]**o+(0.41183946621)*x[1]
        s[2,2,1,0]=(-0.356223597296)*x[0]**o+(0.898624256746)*x[0]+(0.491456887622)*x[1]**o+(0.454642872239)*x[1]
        s[2,2,1,1]=(0.70662625865)*x[0]**o+(0.0268161060988)*x[0]+(-0.17883477863)*x[1]**o+(-0.379062708108)*x[1]
        s[2,2,2,0]=(0.405231213751)*x[0]**o+(-0.339541044657)*x[0]+(0.92369283669)*x[1]**o+(-0.521296842329)*x[1]
        s[2,2,2,1]=(0.371621085625)*x[0]**o+(-0.277933239024)*x[0]+(0.99545384432)*x[1]**o+(-0.458904147226)*x[1]
        s[2,3,0,0]=(-0.978371890906)*x[0]**o+(-0.220612890504)*x[0]+(0.485724985276)*x[1]**o+(0.566941748451)*x[1]
        s[2,3,0,1]=(-0.784455960641)*x[0]**o+(0.488256734495)*x[0]+(-0.85203669445)*x[1]**o+(-0.573678796739)*x[1]
        s[2,3,1,0]=(-0.083067317174)*x[0]**o+(0.284336362391)*x[0]+(-0.671352852822)*x[1]**o+(0.585391001713)*x[1]
        s[2,3,1,1]=(0.453596080554)*x[0]**o+(0.382646103643)*x[0]+(0.976202291904)*x[1]**o+(0.269885074758)*x[1]
        s[2,3,2,0]=(0.58278560743)*x[0]**o+(-0.641861358694)*x[0]+(0.65546441735)*x[1]**o+(-0.510301705975)*x[1]
        s[2,3,2,1]=(-0.462458062923)*x[0]**o+(-0.860624656355)*x[0]+(-0.795012311004)*x[1]**o+(-0.945015232064)*x[1]
        s[2,4,0,0]=(0.373515573367)*x[0]**o+(0.429646026706)*x[0]+(0.107731590254)*x[1]**o+(0.404788302157)*x[1]
        s[2,4,0,1]=(0.922575542165)*x[0]**o+(0.990525976075)*x[0]+(-0.920047278057)*x[1]**o+(-0.606366502396)*x[1]
        s[2,4,1,0]=(-0.395749593133)*x[0]**o+(-0.303444216059)*x[0]+(0.279159896939)*x[1]**o+(-0.998124949675)*x[1]
        s[2,4,1,1]=(0.608308919638)*x[0]**o+(0.547633707859)*x[0]+(-0.335546391995)*x[1]**o+(-0.49712570743)*x[1]
        s[2,4,2,0]=(-0.764049221842)*x[0]**o+(-0.691060860074)*x[0]+(0.576144131381)*x[1]**o+(-0.0692145325011)*x[1]
        s[2,4,2,1]=(-0.445193960624)*x[0]**o+(-0.299673373394)*x[0]+(-0.17618349631)*x[1]**o+(0.6882025213)*x[1]
        s[3,0,0,0]=(-0.363639361841)*x[0]**o+(-0.637031005499)*x[0]+(0.750565431989)*x[1]**o+(-0.713309519647)*x[1]
        s[3,0,0,1]=(0.86686475536)*x[0]**o+(0.857916166663)*x[0]+(-0.233119477582)*x[1]**o+(-0.275718321599)*x[1]
        s[3,0,1,0]=(0.676379750639)*x[0]**o+(-0.221464786801)*x[0]+(-0.307372968314)*x[1]**o+(-0.974878791542)*x[1]
        s[3,0,1,1]=(0.963332442265)*x[0]**o+(0.873237489049)*x[0]+(0.347621440028)*x[1]**o+(-0.447236045492)*x[1]
        s[3,0,2,0]=(0.574231383242)*x[0]**o+(0.406006911504)*x[0]+(0.998282029235)*x[1]**o+(-0.713526415934)*x[1]
        s[3,0,2,1]=(0.130477890591)*x[0]**o+(-0.994273713498)*x[0]+(-0.836645831178)*x[1]**o+(0.23964737976)*x[1]
        s[3,1,0,0]=(0.681027983965)*x[0]**o+(0.277762028836)*x[0]+(-0.476680449158)*x[1]**o+(0.863218206262)*x[1]
        s[3,1,0,1]=(-0.551857608855)*x[0]**o+(0.714701927775)*x[0]+(0.0805460030018)*x[1]**o+(-0.75844120919)*x[1]
        s[3,1,1,0]=(-0.699952165487)*x[0]**o+(-0.829498201164)*x[0]+(0.518643791102)*x[1]**o+(0.748348341191)*x[1]
        s[3,1,1,1]=(0.628926227064)*x[0]**o+(0.756291174101)*x[0]+(0.847249240418)*x[1]**o+(-0.694296012132)*x[1]
        s[3,1,2,0]=(0.300066092357)*x[0]**o+(0.0979478221525)*x[0]+(-0.652853644798)*x[1]**o+(0.87892206219)*x[1]
        s[3,1,2,1]=(0.820578435316)*x[0]**o+(-0.151267836854)*x[0]+(0.288662297283)*x[1]**o+(-0.530205788227)*x[1]
        s[3,2,0,0]=(-0.262081376064)*x[0]**o+(0.564488927009)*x[0]+(-0.574271798456)*x[1]**o+(-0.103341985058)*x[1]
        s[3,2,0,1]=(0.264373345575)*x[0]**o+(0.505975223545)*x[0]+(0.760869432805)*x[1]**o+(-0.348245510796)*x[1]
        s[3,2,1,0]=(0.382085638525)*x[0]**o+(0.271178748741)*x[0]+(-0.344266263181)*x[1]**o+(0.949904496331)*x[1]
        s[3,2,1,1]=(0.570382182257)*x[0]**o+(-0.507665638153)*x[0]+(0.675674419969)*x[1]**o+(-0.741319149319)*x[1]
        s[3,2,2,0]=(-0.0159674521901)*x[0]**o+(-0.201449258763)*x[0]+(-0.774168601079)*x[1]**o+(-0.333025684689)*x[1]
        s[3,2,2,1]=(0.21116026219)*x[0]**o+(0.174264855243)*x[0]+(-0.693006543432)*x[1]**o+(-0.117768474854)*x[1]
        s[3,3,0,0]=(-0.0965568139991)*x[0]**o+(-0.305779611724)*x[0]+(-0.905575174356)*x[1]**o+(-0.0438869174233)*x[1]
        s[3,3,0,1]=(-0.517028280471)*x[0]**o+(0.725783549398)*x[0]+(0.346469604255)*x[1]**o+(-0.958724910661)*x[1]
        s[3,3,1,0]=(0.673217439387)*x[0]**o+(0.171407887035)*x[0]+(-0.931242988664)*x[1]**o+(0.905785986466)*x[1]
        s[3,3,1,1]=(-0.806309377174)*x[0]**o+(-0.39154927098)*x[0]+(-0.353245837689)*x[1]**o+(-0.0891421463801)*x[1]
        s[3,3,2,0]=(0.54365433687)*x[0]**o+(0.328598607479)*x[0]+(-0.134078804517)*x[1]**o+(0.368297327809)*x[1]
        s[3,3,2,1]=(0.309908919248)*x[0]**o+(-0.366081335409)*x[0]+(0.287094150861)*x[1]**o+(-0.311725137682)*x[1]
        s[3,4,0,0]=(0.0350561429151)*x[0]**o+(-0.587757161762)*x[0]+(0.105785456873)*x[1]**o+(0.00732306049631)*x[1]
        s[3,4,0,1]=(0.110213440973)*x[0]**o+(-0.429550096051)*x[0]+(0.471367364287)*x[1]**o+(0.580191251114)*x[1]
        s[3,4,1,0]=(0.349823338156)*x[0]**o+(0.474522056666)*x[0]+(-0.190656365622)*x[1]**o+(0.872370974782)*x[1]
        s[3,4,1,1]=(-0.694637797093)*x[0]**o+(0.648769674687)*x[0]+(-0.0962824527574)*x[1]**o+(0.669702141463)*x[1]
        s[3,4,2,0]=(-0.954369212248)*x[0]**o+(0.558821994909)*x[0]+(-0.196151296515)*x[1]**o+(0.475540095693)*x[1]
        s[3,4,2,1]=(0.946611204124)*x[0]**o+(-0.452828842027)*x[0]+(-0.167471954345)*x[1]**o+(0.0326624440493)*x[1]
        ref[0,0,0,0]=(0.246025700812)/(o+1.)+(-0.0700771278147)
        ref[0,0,0,1]=(0.872903965104)/(o+1.)+(-0.0956768400749)
        ref[0,0,1,0]=(-0.149098784473)/(o+1.)+(-0.271683908641)
        ref[0,0,1,1]=(0.813191985058)/(o+1.)+(0.141471386383)
        ref[0,0,2,0]=(-0.554184175721)/(o+1.)+(-0.621662970196)
        ref[0,0,2,1]=(-1.0295469585)/(o+1.)+(-0.0365685105983)
        ref[0,1,0,0]=(0.151804025931)/(o+1.)+(-0.212298031588)
        ref[0,1,0,1]=(-0.533004796234)/(o+1.)+(0.0485769377937)
        ref[0,1,1,0]=(-0.802375632463)/(o+1.)+(0.127439034441)
        ref[0,1,1,1]=(0.325754839785)/(o+1.)+(0.0972311868531)
        ref[0,1,2,0]=(1.43492852092)/(o+1.)+(0.652296741316)
        ref[0,1,2,1]=(-0.347667591981)/(o+1.)+(-0.160364561879)
        ref[0,2,0,0]=(-1.37099731406)/(o+1.)+(0.0563097354051)
        ref[0,2,0,1]=(1.11251425188)/(o+1.)+(-0.835019726344)
        ref[0,2,1,0]=(-0.770103190572)/(o+1.)+(-0.309897780967)
        ref[0,2,1,1]=(0.159272551388)/(o+1.)+(0.706859784466)
        ref[0,2,2,0]=(-0.391469887349)/(o+1.)+(-0.207974163745)
        ref[0,2,2,1]=(-0.0415688115757)/(o+1.)+(0.0604937496788)
        ref[0,3,0,0]=(0.957559739626)/(o+1.)+(0.228032778194)
        ref[0,3,0,1]=(-1.09130803461)/(o+1.)+(-0.662860500606)
        ref[0,3,1,0]=(-0.449862052638)/(o+1.)+(-0.344594323242)
        ref[0,3,1,1]=(0.219326920673)/(o+1.)+(0.405011796484)
        ref[0,3,2,0]=(0.275993808116)/(o+1.)+(-0.461328543451)
        ref[0,3,2,1]=(-0.575033733717)/(o+1.)+(0.884920083202)
        ref[0,4,0,0]=(0.0253483143264)/(o+1.)+(-0.831762460572)
        ref[0,4,0,1]=(-0.593504118306)/(o+1.)+(-0.0899796107363)
        ref[0,4,1,0]=(0.946442680954)/(o+1.)+(0.340471820786)
        ref[0,4,1,1]=(1.00832461793)/(o+1.)+(-0.489125179816)
        ref[0,4,2,0]=(0.658258894209)/(o+1.)+(-0.252225332525)
        ref[0,4,2,1]=(0.190670120044)/(o+1.)+(0.107482927026)
        ref[1,0,0,0]=(-0.638264699882)/(o+1.)+(-0.37905912763)
        ref[1,0,0,1]=(0.858273041153)/(o+1.)+(-0.602359980261)
        ref[1,0,1,0]=(1.19603458122)/(o+1.)+(0.623979052337)
        ref[1,0,1,1]=(-0.0559871855173)/(o+1.)+(0.392163437219)
        ref[1,0,2,0]=(0.202869107651)/(o+1.)+(-0.327529727939)
        ref[1,0,2,1]=(0.497839996627)/(o+1.)+(-0.198257844767)
        ref[1,1,0,0]=(-1.05743683429)/(o+1.)+(-0.107408196452)
        ref[1,1,0,1]=(-1.11022578935)/(o+1.)+(0.0799654739886)
        ref[1,1,1,0]=(1.02865615322)/(o+1.)+(-0.111339912553)
        ref[1,1,1,1]=(-0.387124166332)/(o+1.)+(-0.13196732944)
        ref[1,1,2,0]=(0.364571348315)/(o+1.)+(-0.827711953327)
        ref[1,1,2,1]=(0.800389584904)/(o+1.)+(-0.570167855757)
        ref[1,2,0,0]=(-0.359885437423)/(o+1.)+(-0.683824166857)
        ref[1,2,0,1]=(-0.139489334179)/(o+1.)+(-0.0550294716506)
        ref[1,2,1,0]=(0.05752529846)/(o+1.)+(0.330669314498)
        ref[1,2,1,1]=(-0.68363897434)/(o+1.)+(-0.210682497577)
        ref[1,2,2,0]=(-0.971569775609)/(o+1.)+(0.258047639988)
        ref[1,2,2,1]=(0.966753430217)/(o+1.)+(0.737154900402)
        ref[1,3,0,0]=(0.571230899607)/(o+1.)+(0.201680592513)
        ref[1,3,0,1]=(-0.516278846659)/(o+1.)+(-0.203543442922)
        ref[1,3,1,0]=(-0.276019031326)/(o+1.)+(0.465273316758)
        ref[1,3,1,1]=(0.339281660193)/(o+1.)+(0.578236730598)
        ref[1,3,2,0]=(0.811123220269)/(o+1.)+(-0.306640881038)
        ref[1,3,2,1]=(0.397602998535)/(o+1.)+(0.631760762007)
        ref[1,4,0,0]=(0.270527628963)/(o+1.)+(-0.156009469628)
        ref[1,4,0,1]=(0.0267275858003)/(o+1.)+(0.1114627818)
        ref[1,4,1,0]=(0.117364258255)/(o+1.)+(-0.627596298481)
        ref[1,4,1,1]=(1.06982816128)/(o+1.)+(-0.0961307552069)
        ref[1,4,2,0]=(0.648239950234)/(o+1.)+(-0.196286224748)
        ref[1,4,2,1]=(-0.273099453886)/(o+1.)+(-0.71054634034)
        ref[2,0,0,0]=(0.0597301506259)/(o+1.)+(-0.0661675007782)
        ref[2,0,0,1]=(-1.03461567689)/(o+1.)+(0.336883492696)
        ref[2,0,1,0]=(-1.08184729537)/(o+1.)+(-0.236806207955)
        ref[2,0,1,1]=(0.407177326369)/(o+1.)+(-0.411518582134)
        ref[2,0,2,0]=(0.111140120177)/(o+1.)+(0.201425312625)
        ref[2,0,2,1]=(-0.122119044061)/(o+1.)+(-0.0178123728248)
        ref[2,1,0,0]=(0.277542930723)/(o+1.)+(0.332979421149)
        ref[2,1,0,1]=(0.716015179185)/(o+1.)+(-0.35221548881)
        ref[2,1,1,0]=(-1.66176914548)/(o+1.)+(0.136807260612)
        ref[2,1,1,1]=(0.457774847682)/(o+1.)+(-0.211956003748)
        ref[2,1,2,0]=(0.154152304156)/(o+1.)+(-0.225608599389)
        ref[2,1,2,1]=(-0.394387972979)/(o+1.)+(-0.44858376371)
        ref[2,2,0,0]=(-1.17070689262)/(o+1.)+(0.244820006519)
        ref[2,2,0,1]=(1.70365106393)/(o+1.)+(0.61037400899)
        ref[2,2,1,0]=(0.135233290326)/(o+1.)+(0.676633564493)
        ref[2,2,1,1]=(0.527791480019)/(o+1.)+(-0.176123301004)
        ref[2,2,2,0]=(1.32892405044)/(o+1.)+(-0.430418943493)
        ref[2,2,2,1]=(1.36707492995)/(o+1.)+(-0.368418693125)
        ref[2,3,0,0]=(-0.492646905629)/(o+1.)+(0.173164428973)
        ref[2,3,0,1]=(-1.63649265509)/(o+1.)+(-0.0427110311218)
        ref[2,3,1,0]=(-0.754420169996)/(o+1.)+(0.434863682052)
        ref[2,3,1,1]=(1.42979837246)/(o+1.)+(0.3262655892)
        ref[2,3,2,0]=(1.23825002478)/(o+1.)+(-0.576081532335)
        ref[2,3,2,1]=(-1.25747037393)/(o+1.)+(-0.902819944209)
        ref[2,4,0,0]=(0.48124716362)/(o+1.)+(0.417217164431)
        ref[2,4,0,1]=(0.00252826410759)/(o+1.)+(0.192079736839)
        ref[2,4,1,0]=(-0.116589696194)/(o+1.)+(-0.650784582867)
        ref[2,4,1,1]=(0.272762527643)/(o+1.)+(0.0252540002143)
        ref[2,4,2,0]=(-0.187905090461)/(o+1.)+(-0.380137696288)
        ref[2,4,2,1]=(-0.621377456934)/(o+1.)+(0.194264573953)
        ref[3,0,0,0]=(0.386926070148)/(o+1.)+(-0.675170262573)
        ref[3,0,0,1]=(0.633745277778)/(o+1.)+(0.291098922532)
        ref[3,0,1,0]=(0.369006782325)/(o+1.)+(-0.598171789171)
        ref[3,0,1,1]=(1.31095388229)/(o+1.)+(0.213000721778)
        ref[3,0,2,0]=(1.57251341248)/(o+1.)+(-0.153759752215)
        ref[3,0,2,1]=(-0.706167940587)/(o+1.)+(-0.377313166869)
        ref[3,1,0,0]=(0.204347534807)/(o+1.)+(0.570490117549)
        ref[3,1,0,1]=(-0.471311605853)/(o+1.)+(-0.0218696407078)
        ref[3,1,1,0]=(-0.181308374385)/(o+1.)+(-0.0405749299865)
        ref[3,1,1,1]=(1.47617546748)/(o+1.)+(0.0309975809845)
        ref[3,1,2,0]=(-0.352787552441)/(o+1.)+(0.488434942171)
        ref[3,1,2,1]=(1.1092407326)/(o+1.)+(-0.34073681254)
        ref[3,2,0,0]=(-0.83635317452)/(o+1.)+(0.230573470976)
        ref[3,2,0,1]=(1.02524277838)/(o+1.)+(0.0788648563748)
        ref[3,2,1,0]=(0.0378193753437)/(o+1.)+(0.610541622536)
        ref[3,2,1,1]=(1.24605660223)/(o+1.)+(-0.624492393736)
        ref[3,2,2,0]=(-0.79013605327)/(o+1.)+(-0.267237471726)
        ref[3,2,2,1]=(-0.481846281241)/(o+1.)+(0.0282481901946)
        ref[3,3,0,0]=(-1.00213198835)/(o+1.)+(-0.174833264573)
        ref[3,3,0,1]=(-0.170558676217)/(o+1.)+(-0.116470680631)
        ref[3,3,1,0]=(-0.258025549277)/(o+1.)+(0.538596936751)
        ref[3,3,1,1]=(-1.15955521486)/(o+1.)+(-0.24034570868)
        ref[3,3,2,0]=(0.409575532352)/(o+1.)+(0.348447967644)
        ref[3,3,2,1]=(0.597003070109)/(o+1.)+(-0.338903236546)
        ref[3,4,0,0]=(0.140841599788)/(o+1.)+(-0.290217050633)
        ref[3,4,0,1]=(0.581580805261)/(o+1.)+(0.0753205775318)
        ref[3,4,1,0]=(0.159166972534)/(o+1.)+(0.673446515724)
        ref[3,4,1,1]=(-0.79092024985)/(o+1.)+(0.659235908075)
        ref[3,4,2,0]=(-1.15052050876)/(o+1.)+(0.517181045301)
        ref[3,4,2,1]=(0.779139249778)/(o+1.)+(-0.210083198989)
      else:
        s[0,0,0,0]=(0.748319783349)*x[0]**o+(-0.176566570347)*x[0]+(-0.302910837408)*x[1]**o+(0.492350205415)*x[1]+(-0.160456161783)*x[2]**o+(0.868280836814)*x[2]
        s[0,0,0,1]=(-0.51136577317)*x[0]**o+(-0.242417336787)*x[0]+(-0.785996766849)*x[1]**o+(0.606869170308)*x[1]+(0.49391109913)*x[2]**o+(-0.319568202304)*x[2]
        s[0,0,1,0]=(0.640062862527)*x[0]**o+(0.566706264858)*x[0]+(0.907582097741)*x[1]**o+(0.350910171053)*x[1]+(-0.909242439832)*x[2]**o+(-0.85540876524)*x[2]
        s[0,0,1,1]=(-0.377548918631)*x[0]**o+(-0.630218128454)*x[0]+(0.45141559373)*x[1]**o+(0.0832708083581)*x[1]+(0.233405447853)*x[2]**o+(0.567903425032)*x[2]
        s[0,0,2,0]=(-0.497312249883)*x[0]**o+(-0.651244043641)*x[0]+(-0.575450843765)*x[1]**o+(0.0219939930053)*x[1]+(0.570655374403)*x[2]**o+(-0.402040210721)*x[2]
        s[0,0,2,1]=(0.383694368889)*x[0]**o+(-0.280765736983)*x[0]+(0.644560587428)*x[1]**o+(-0.263879759253)*x[1]+(0.78943504044)*x[2]**o+(0.838120969604)*x[2]
        s[0,1,0,0]=(0.331871753344)*x[0]**o+(0.731388799379)*x[0]+(0.394440515222)*x[1]**o+(-0.337910220731)*x[1]+(-0.078492234799)*x[2]**o+(0.388988268348)*x[2]
        s[0,1,0,1]=(0.197262971559)*x[0]**o+(0.482585248548)*x[0]+(0.0890475484335)*x[1]**o+(-0.154881471591)*x[1]+(0.14978160538)*x[2]**o+(-0.115243504627)*x[2]
        s[0,1,1,0]=(0.338072090662)*x[0]**o+(0.238967458874)*x[0]+(0.4078504965)*x[1]**o+(-0.941497950412)*x[1]+(0.539654579256)*x[2]**o+(0.777502770306)*x[2]
        s[0,1,1,1]=(-0.876894224774)*x[0]**o+(-0.413741121497)*x[0]+(0.841988144013)*x[1]**o+(-0.987067492493)*x[1]+(-0.0142647738806)*x[2]**o+(0.809756634084)*x[2]
        s[0,1,2,0]=(0.771962423748)*x[0]**o+(-0.274151695697)*x[0]+(-0.0583329632543)*x[1]**o+(0.290755268126)*x[1]+(0.986842038664)*x[2]**o+(-0.445513638841)*x[2]
        s[0,1,2,1]=(0.235289294125)*x[0]**o+(-0.0230216115543)*x[0]+(-0.39449823454)*x[1]**o+(0.758538375985)*x[1]+(-0.990485958396)*x[2]**o+(0.441787118653)*x[2]
        s[0,2,0,0]=(-0.860580472907)*x[0]**o+(-0.480513252563)*x[0]+(0.388567023408)*x[1]**o+(-0.435900869228)*x[1]+(-0.483812008458)*x[2]**o+(0.87821418943)*x[2]
        s[0,2,0,1]=(-0.391914341084)*x[0]**o+(-0.813798019851)*x[0]+(0.681793443132)*x[1]**o+(-0.422727355552)*x[1]+(-0.0599492576813)*x[2]**o+(0.630302059584)*x[2]
        s[0,2,1,0]=(-0.857349416782)*x[0]**o+(0.0937767571422)*x[0]+(-0.935044609271)*x[1]**o+(-0.825241166723)*x[1]+(-0.78223171556)*x[2]**o+(0.33550368528)*x[2]
        s[0,2,1,1]=(-0.366924211693)*x[0]**o+(0.572797118425)*x[0]+(-0.289194982186)*x[1]**o+(-0.942701172085)*x[1]+(0.614219031682)*x[2]**o+(-0.389753632964)*x[2]
        s[0,2,2,0]=(-0.404732972636)*x[0]**o+(0.631781218361)*x[0]+(-0.263486337549)*x[1]**o+(0.631746356812)*x[1]+(-0.515291196501)*x[2]**o+(-0.363991020547)*x[2]
        s[0,2,2,1]=(0.892685628814)*x[0]**o+(0.963218649844)*x[0]+(0.483947727789)*x[1]**o+(0.211028649587)*x[1]+(0.50159178018)*x[2]**o+(0.0711392272561)*x[2]
        s[0,3,0,0]=(-0.33216085676)*x[0]**o+(0.906269165577)*x[0]+(0.874123825651)*x[1]**o+(-0.570808752482)*x[1]+(-0.561182318614)*x[2]**o+(-0.884125641249)*x[2]
        s[0,3,0,1]=(-0.0703055529684)*x[0]**o+(-0.101044608869)*x[0]+(0.829269084242)*x[1]**o+(-0.497613621381)*x[1]+(0.297140846283)*x[2]**o+(0.566992550301)*x[2]
        s[0,3,1,0]=(0.00230797481965)*x[0]**o+(-0.561978070323)*x[0]+(-0.874987267344)*x[1]**o+(0.404602330516)*x[1]+(0.541390115931)*x[2]**o+(0.109832470538)*x[2]
        s[0,3,1,1]=(-0.738073409216)*x[0]**o+(0.016899800363)*x[0]+(0.606773214528)*x[1]**o+(0.607363930641)*x[1]+(-0.755469399679)*x[2]**o+(-0.585951709489)*x[2]
        s[0,3,2,0]=(0.754566844048)*x[0]**o+(0.353623703809)*x[0]+(0.300465743201)*x[1]**o+(0.4804155825)*x[1]+(-0.500709775706)*x[2]**o+(-0.184866087487)*x[2]
        s[0,3,2,1]=(0.656642762895)*x[0]**o+(-0.522249569275)*x[0]+(-0.689897112218)*x[1]**o+(0.197660165382)*x[1]+(0.790618474391)*x[2]**o+(0.34534294776)*x[2]
        s[0,4,0,0]=(0.951780164354)*x[0]**o+(0.227261667276)*x[0]+(-0.388326921756)*x[1]**o+(-0.143812077493)*x[1]+(0.869741228576)*x[2]**o+(0.896296431361)*x[2]
        s[0,4,0,1]=(0.0418495997166)*x[0]**o+(0.378737546698)*x[0]+(-0.893460535382)*x[1]**o+(-0.0409222699203)*x[1]+(-0.6955627414)*x[2]**o+(-0.993500126323)*x[2]
        s[0,4,1,0]=(-0.47303219937)*x[0]**o+(0.488878345444)*x[0]+(0.110295068747)*x[1]**o+(-0.775897017507)*x[1]+(-0.664013808104)*x[2]**o+(-0.125758726632)*x[2]
        s[0,4,1,1]=(0.326490818022)*x[0]**o+(-0.785954928184)*x[0]+(0.85336153105)*x[1]**o+(-0.573522941199)*x[1]+(-0.42185352182)*x[2]**o+(-0.425049668912)*x[2]
        s[0,4,2,0]=(0.290783142101)*x[0]**o+(-0.502056588268)*x[0]+(0.433355340131)*x[1]**o+(-0.717832338133)*x[1]+(0.254866069131)*x[2]**o+(-0.383849864093)*x[2]
        s[0,4,2,1]=(-0.796340985968)*x[0]**o+(-0.732115333195)*x[0]+(-0.829732009003)*x[1]**o+(-0.790621874069)*x[1]+(-0.60818338505)*x[2]**o+(-0.892012123169)*x[2]
        s[1,0,0,0]=(0.150096606609)*x[0]**o+(0.713142060809)*x[0]+(0.0898795205212)*x[1]**o+(-0.682356428859)*x[1]+(0.275326716654)*x[2]**o+(-0.356136748426)*x[2]
        s[1,0,0,1]=(0.104574120129)*x[0]**o+(-0.895395804614)*x[0]+(-0.00895409609851)*x[1]**o+(0.852200905159)*x[1]+(0.25757145031)*x[2]**o+(-0.578299505429)*x[2]
        s[1,0,1,0]=(-0.83104126362)*x[0]**o+(-0.587286782794)*x[0]+(0.043755441563)*x[1]**o+(-0.357299738488)*x[1]+(0.900592455437)*x[2]**o+(0.885645422129)*x[2]
        s[1,0,1,1]=(-0.144444995355)*x[0]**o+(-0.667350132103)*x[0]+(0.949183394884)*x[1]**o+(-0.519086948324)*x[1]+(-0.74215320159)*x[2]**o+(-0.202102401774)*x[2]
        s[1,0,2,0]=(0.303496934218)*x[0]**o+(-0.02461446864)*x[0]+(-0.0867456086666)*x[1]**o+(0.730962877622)*x[1]+(-0.851607522381)*x[2]**o+(-0.239310033319)*x[2]
        s[1,0,2,1]=(-0.512139821912)*x[0]**o+(-0.69403755459)*x[0]+(0.462840604615)*x[1]**o+(-0.192161565941)*x[1]+(0.807307771096)*x[2]**o+(0.440024157111)*x[2]
        s[1,1,0,0]=(-0.693987625939)*x[0]**o+(0.410652593337)*x[0]+(0.665893670901)*x[1]**o+(0.525886124327)*x[1]+(0.200291951309)*x[2]**o+(0.793353011987)*x[2]
        s[1,1,0,1]=(-0.667911914141)*x[0]**o+(0.693960372567)*x[0]+(-0.0892935574405)*x[1]**o+(0.669184962503)*x[1]+(0.664202374165)*x[2]**o+(-0.474657421474)*x[2]
        s[1,1,1,0]=(0.4341035606)*x[0]**o+(0.707815942288)*x[0]+(-0.204720100594)*x[1]**o+(-0.816120724182)*x[1]+(-0.259566839542)*x[2]**o+(0.918010553929)*x[2]
        s[1,1,1,1]=(0.280925026809)*x[0]**o+(0.895509529721)*x[0]+(0.244175246656)*x[1]**o+(0.692059608476)*x[1]+(-0.424963623478)*x[2]**o+(0.985618462184)*x[2]
        s[1,1,2,0]=(0.172285590575)*x[0]**o+(0.0614208586474)*x[0]+(-0.751024157153)*x[1]**o+(0.182364162229)*x[1]+(-0.400270263752)*x[2]**o+(0.6410957443)*x[2]
        s[1,1,2,1]=(0.621541338202)*x[0]**o+(0.716601026013)*x[0]+(0.822288477475)*x[1]**o+(0.200859656154)*x[1]+(-0.651437159651)*x[2]**o+(-0.554048539377)*x[2]
        s[1,2,0,0]=(-0.835149004101)*x[0]**o+(-0.748380867211)*x[0]+(-0.445526738596)*x[1]**o+(0.782304294095)*x[1]+(0.827204947885)*x[2]**o+(0.188779249069)*x[2]
        s[1,2,0,1]=(0.354866953137)*x[0]**o+(-0.981132219204)*x[0]+(0.488043961513)*x[1]**o+(0.257052078929)*x[1]+(0.719111294043)*x[2]**o+(-0.298289862834)*x[2]
        s[1,2,1,0]=(0.552605101038)*x[0]**o+(0.145016605474)*x[0]+(0.119161182879)*x[1]**o+(0.723703065293)*x[1]+(-0.011955307053)*x[2]**o+(0.374814299212)*x[2]
        s[1,2,1,1]=(0.448055014229)*x[0]**o+(-0.320498330109)*x[0]+(0.812977543935)*x[1]**o+(0.374118976348)*x[1]+(0.656782601024)*x[2]**o+(0.0845883407242)*x[2]
        s[1,2,2,0]=(0.418470977027)*x[0]**o+(-0.324185586615)*x[0]+(0.576184154734)*x[1]**o+(-0.244791711079)*x[1]+(-0.383448002434)*x[2]**o+(-0.425391146295)*x[2]
        s[1,2,2,1]=(0.885670287748)*x[0]**o+(0.804245948464)*x[0]+(-0.122365449147)*x[1]**o+(0.0526689576387)*x[1]+(0.886983215847)*x[2]**o+(-0.9130795011)*x[2]
        s[1,3,0,0]=(-0.698202951872)*x[0]**o+(-0.753867598728)*x[0]+(-0.729720876147)*x[1]**o+(-0.535739350698)*x[1]+(0.140570380613)*x[2]**o+(-0.0716035886696)*x[2]
        s[1,3,0,1]=(-0.898459992298)*x[0]**o+(0.372462814699)*x[0]+(0.605465414867)*x[1]**o+(-0.684605485569)*x[1]+(0.549437793056)*x[2]**o+(-0.010737341414)*x[2]
        s[1,3,1,0]=(0.792329973599)*x[0]**o+(-0.52685018301)*x[0]+(0.200314760956)*x[1]**o+(0.632873255355)*x[1]+(-0.43152569425)*x[2]**o+(0.348904687244)*x[2]
        s[1,3,1,1]=(0.209464885105)*x[0]**o+(-0.0146787858502)*x[0]+(0.320235817099)*x[1]**o+(0.86370910616)*x[1]+(-0.202145495919)*x[2]**o+(-0.25657518419)*x[2]
        s[1,3,2,0]=(0.489051201235)*x[0]**o+(0.349594264623)*x[0]+(-0.574002000861)*x[1]**o+(0.390761344969)*x[1]+(0.1514845518)*x[2]**o+(0.802240236468)*x[2]
        s[1,3,2,1]=(-0.0182530256393)*x[0]**o+(0.303828930846)*x[0]+(0.618612491362)*x[1]**o+(0.293473693133)*x[1]+(0.763201722939)*x[2]**o+(-0.652537522965)*x[2]
        s[1,4,0,0]=(0.349015278137)*x[0]**o+(-0.0165197033193)*x[0]+(0.63228324617)*x[1]**o+(-0.934745316284)*x[1]+(-0.187390369535)*x[2]**o+(-0.456599706963)*x[2]
        s[1,4,0,1]=(0.310212294648)*x[0]**o+(-0.898896972668)*x[0]+(0.229475812326)*x[1]**o+(0.787159228942)*x[1]+(0.133488840506)*x[2]**o+(0.949893906424)*x[2]
        s[1,4,1,0]=(-0.869667859962)*x[0]**o+(0.624210860706)*x[0]+(0.122085293209)*x[1]**o+(-0.624604470463)*x[1]+(-0.878410325648)*x[2]**o+(-0.0841801197816)*x[2]
        s[1,4,1,1]=(0.706521970977)*x[0]**o+(0.23837086782)*x[0]+(-0.960682169998)*x[1]**o+(-0.707493453388)*x[1]+(0.828048621994)*x[2]**o+(0.327258769743)*x[2]
        s[1,4,2,0]=(-0.291203807685)*x[0]**o+(0.707463915698)*x[0]+(-0.860720761064)*x[1]**o+(0.802477835676)*x[1]+(-0.16905212173)*x[2]**o+(0.453497949714)*x[2]
        s[1,4,2,1]=(0.789863893897)*x[0]**o+(-0.945472074403)*x[0]+(-0.047780484871)*x[1]**o+(0.199167500985)*x[1]+(0.575959160298)*x[2]**o+(-0.181712597239)*x[2]
        s[2,0,0,0]=(0.61569741708)*x[0]**o+(0.609998488119)*x[0]+(-0.943339902899)*x[1]**o+(-0.962332686128)*x[1]+(0.458927713248)*x[2]**o+(-0.191444752971)*x[2]
        s[2,0,0,1]=(0.0306432921503)*x[0]**o+(0.326065705299)*x[0]+(0.29054908043)*x[1]**o+(-0.498110577924)*x[1]+(-0.855604413911)*x[2]**o+(-0.832549081863)*x[2]
        s[2,0,1,0]=(0.539469235745)*x[0]**o+(0.549053450527)*x[0]+(0.379811452156)*x[1]**o+(0.383008863276)*x[1]+(-0.835777000108)*x[2]**o+(0.418879791451)*x[2]
        s[2,0,1,1]=(0.0773392682115)*x[0]**o+(0.594680641339)*x[0]+(-0.514207317809)*x[1]**o+(0.516677819046)*x[1]+(-0.616548060204)*x[2]**o+(-0.487308340419)*x[2]
        s[2,0,2,0]=(0.308538286588)*x[0]**o+(0.553188024822)*x[0]+(0.640606076984)*x[1]**o+(0.851001392437)*x[1]+(-0.175136140727)*x[2]**o+(-0.0677219547255)*x[2]
        s[2,0,2,1]=(0.689683007295)*x[0]**o+(0.0205486604859)*x[0]+(-0.513197632413)*x[1]**o+(-0.589507997499)*x[1]+(0.0967263356508)*x[2]**o+(-0.743224844971)*x[2]
        s[2,1,0,0]=(0.175040132619)*x[0]**o+(0.0564766932006)*x[0]+(-0.641666937619)*x[1]**o+(0.820974466919)*x[1]+(0.494310429863)*x[2]**o+(-0.704304819607)*x[2]
        s[2,1,0,1]=(-0.890775824863)*x[0]**o+(-0.500686412296)*x[0]+(0.153470124596)*x[1]**o+(-0.64297395847)*x[1]+(-0.716566099394)*x[2]**o+(0.74578557063)*x[2]
        s[2,1,1,0]=(-0.999358965815)*x[0]**o+(-0.657733542154)*x[0]+(0.197035042134)*x[1]**o+(-0.887786199345)*x[1]+(-0.0685059299499)*x[2]**o+(0.498972129916)*x[2]
        s[2,1,1,1]=(0.627376530136)*x[0]**o+(-0.681228319483)*x[0]+(0.552935442983)*x[1]**o+(-0.813921943144)*x[1]+(-0.192761641367)*x[2]**o+(-0.248022829303)*x[2]
        s[2,1,2,0]=(0.845230425625)*x[0]**o+(-0.503049369073)*x[0]+(-0.705939778787)*x[1]**o+(-0.862373899236)*x[1]+(-0.327695135868)*x[2]**o+(-0.469012024699)*x[2]
        s[2,1,2,1]=(-0.509720328843)*x[0]**o+(-0.595784133936)*x[0]+(0.703072807212)*x[1]**o+(0.413326203128)*x[1]+(0.819831655306)*x[2]**o+(-0.494002066589)*x[2]
        s[2,2,0,0]=(-0.349642870638)*x[0]**o+(0.530770622335)*x[0]+(0.132675549915)*x[1]**o+(-0.234998432688)*x[1]+(-0.928123366402)*x[2]**o+(-0.578031336794)*x[2]
        s[2,2,0,1]=(-0.38335511074)*x[0]**o+(-0.993906933916)*x[0]+(-0.82535132107)*x[1]**o+(0.0112642210987)*x[1]+(0.904749217207)*x[2]**o+(-0.951098241145)*x[2]
        s[2,2,1,0]=(-0.387786853958)*x[0]**o+(0.455299305562)*x[0]+(0.690526696082)*x[1]**o+(0.192809198912)*x[1]+(-0.0217815689027)*x[2]**o+(0.853966231049)*x[2]
        s[2,2,1,1]=(0.399193164297)*x[0]**o+(0.19349639518)*x[0]+(0.317501609824)*x[1]**o+(0.240510267197)*x[1]+(0.0450712212053)*x[2]**o+(0.565853605939)*x[2]
        s[2,2,2,0]=(0.0853562286915)*x[0]**o+(0.122662799465)*x[0]+(0.26271658282)*x[1]**o+(0.90975618664)*x[1]+(-0.4111835609)*x[2]**o+(0.813644048231)*x[2]
        s[2,2,2,1]=(-0.852594000157)*x[0]**o+(-0.38507970023)*x[0]+(-0.232827312227)*x[1]**o+(0.00790525919862)*x[1]+(-0.3900612508)*x[2]**o+(-0.655554313153)*x[2]
        s[2,3,0,0]=(0.636184557317)*x[0]**o+(-0.549119485575)*x[0]+(-0.948821359516)*x[1]**o+(0.131451687899)*x[1]+(0.670399101185)*x[2]**o+(0.489217491983)*x[2]
        s[2,3,0,1]=(-0.856150893763)*x[0]**o+(-0.280815300266)*x[0]+(-0.523050266221)*x[1]**o+(0.520191604865)*x[1]+(0.0521766680699)*x[2]**o+(0.900107417471)*x[2]
        s[2,3,1,0]=(0.826705547671)*x[0]**o+(0.840439089501)*x[0]+(0.170261390802)*x[1]**o+(0.0811180372799)*x[1]+(-0.150910532868)*x[2]**o+(-0.312396418734)*x[2]
        s[2,3,1,1]=(-0.761542055765)*x[0]**o+(-0.972348306591)*x[0]+(0.872426053365)*x[1]**o+(-0.395809987519)*x[1]+(0.731896957461)*x[2]**o+(0.743098137394)*x[2]
        s[2,3,2,0]=(0.702021470374)*x[0]**o+(-0.378108346769)*x[0]+(-0.931728063859)*x[1]**o+(0.387496129463)*x[1]+(-0.0627914582829)*x[2]**o+(-0.648454760891)*x[2]
        s[2,3,2,1]=(0.694922709107)*x[0]**o+(0.348731283066)*x[0]+(-0.772113301946)*x[1]**o+(0.958191340446)*x[1]+(0.227272159891)*x[2]**o+(-0.553319421819)*x[2]
        s[2,4,0,0]=(0.201119725976)*x[0]**o+(0.593100698188)*x[0]+(-0.140057809068)*x[1]**o+(-0.0952566649925)*x[1]+(-0.851826547872)*x[2]**o+(0.505822022941)*x[2]
        s[2,4,0,1]=(0.0212756051792)*x[0]**o+(0.612323788569)*x[0]+(-0.582655513686)*x[1]**o+(0.461920977901)*x[1]+(0.45615084988)*x[2]**o+(-0.831485043604)*x[2]
        s[2,4,1,0]=(-0.896264627494)*x[0]**o+(0.130655495842)*x[0]+(-0.51905734937)*x[1]**o+(-0.850916009006)*x[1]+(-0.927481150342)*x[2]**o+(-0.394829448395)*x[2]
        s[2,4,1,1]=(0.589011113198)*x[0]**o+(0.779837449297)*x[0]+(-0.404231103183)*x[1]**o+(-0.818485472133)*x[1]+(-0.975586808573)*x[2]**o+(-0.140406688026)*x[2]
        s[2,4,2,0]=(0.621779862491)*x[0]**o+(-0.761880142681)*x[0]+(0.373695582445)*x[1]**o+(0.822746723299)*x[1]+(-0.0828625534579)*x[2]**o+(0.0737978389046)*x[2]
        s[2,4,2,1]=(-0.565625067907)*x[0]**o+(-0.494605807934)*x[0]+(-0.16016442955)*x[1]**o+(0.154859091053)*x[1]+(0.274690562477)*x[2]**o+(-0.0880681582431)*x[2]
        s[3,0,0,0]=(-0.876284142284)*x[0]**o+(0.833701643665)*x[0]+(-0.343394927755)*x[1]**o+(0.750989322932)*x[1]+(-0.0929682612771)*x[2]**o+(-0.098548374762)*x[2]
        s[3,0,0,1]=(-0.894241201097)*x[0]**o+(0.960924642371)*x[0]+(-0.385879848199)*x[1]**o+(-0.0740790803068)*x[1]+(-0.202098139954)*x[2]**o+(0.881367694352)*x[2]
        s[3,0,1,0]=(-0.462333015258)*x[0]**o+(-0.209675407943)*x[0]+(-0.162580359293)*x[1]**o+(0.699769354405)*x[1]+(0.505842739219)*x[2]**o+(-0.0511954656814)*x[2]
        s[3,0,1,1]=(-0.560563273256)*x[0]**o+(-0.798740244759)*x[0]+(0.661483757231)*x[1]**o+(0.160669606575)*x[1]+(-0.433798248002)*x[2]**o+(-0.573448279579)*x[2]
        s[3,0,2,0]=(-0.05963441333)*x[0]**o+(0.123422468842)*x[0]+(-0.0794480113474)*x[1]**o+(-0.980566064632)*x[1]+(0.172807620144)*x[2]**o+(0.290739484628)*x[2]
        s[3,0,2,1]=(-0.824869737132)*x[0]**o+(-0.236911913997)*x[0]+(-0.168741303105)*x[1]**o+(0.115993599766)*x[1]+(0.345866352676)*x[2]**o+(-0.810213944202)*x[2]
        s[3,1,0,0]=(-0.413325465485)*x[0]**o+(0.622383829933)*x[0]+(-0.517445935411)*x[1]**o+(0.968177050325)*x[1]+(-0.257376473971)*x[2]**o+(0.614956161076)*x[2]
        s[3,1,0,1]=(0.998189367064)*x[0]**o+(0.222064685116)*x[0]+(0.0082228342872)*x[1]**o+(-0.779543154084)*x[1]+(-0.566689004021)*x[2]**o+(0.580783912818)*x[2]
        s[3,1,1,0]=(0.245064518006)*x[0]**o+(0.252652697672)*x[0]+(-0.0750782925877)*x[1]**o+(0.216217715236)*x[1]+(-0.0766336852144)*x[2]**o+(0.819380420323)*x[2]
        s[3,1,1,1]=(-0.262865184606)*x[0]**o+(0.200232864872)*x[0]+(0.378217877682)*x[1]**o+(-0.560693320189)*x[1]+(0.235131793324)*x[2]**o+(0.0374878565262)*x[2]
        s[3,1,2,0]=(0.475538018977)*x[0]**o+(-0.731782099691)*x[0]+(-0.983895939904)*x[1]**o+(-0.505746472336)*x[1]+(-0.393214081403)*x[2]**o+(0.48342413719)*x[2]
        s[3,1,2,1]=(0.298214379)*x[0]**o+(-0.232901801182)*x[0]+(-0.47025072398)*x[1]**o+(-0.0935590248775)*x[1]+(-0.270755608412)*x[2]**o+(0.232963599748)*x[2]
        s[3,2,0,0]=(0.544547455045)*x[0]**o+(-0.238443336164)*x[0]+(0.104607358043)*x[1]**o+(0.700551882808)*x[1]+(-0.154865547453)*x[2]**o+(0.897044366696)*x[2]
        s[3,2,0,1]=(-0.226438512542)*x[0]**o+(-0.889161505097)*x[0]+(0.844759329847)*x[1]**o+(0.67393984486)*x[1]+(0.600583328853)*x[2]**o+(0.14085768238)*x[2]
        s[3,2,1,0]=(-0.303519143875)*x[0]**o+(-0.480627582343)*x[0]+(0.711647844961)*x[1]**o+(-0.827593833605)*x[1]+(-0.551933240305)*x[2]**o+(-0.358612992123)*x[2]
        s[3,2,1,1]=(-0.284375845165)*x[0]**o+(-0.144660681111)*x[0]+(-0.520160492072)*x[1]**o+(0.175169216201)*x[1]+(-0.500839582917)*x[2]**o+(-0.816514398637)*x[2]
        s[3,2,2,0]=(-0.563952264023)*x[0]**o+(0.104321734788)*x[0]+(-0.469018385204)*x[1]**o+(0.0093577376129)*x[1]+(-0.732696251175)*x[2]**o+(-0.178803413304)*x[2]
        s[3,2,2,1]=(0.774737106755)*x[0]**o+(0.860675883412)*x[0]+(0.79787325945)*x[1]**o+(0.664362658951)*x[1]+(-0.178147456548)*x[2]**o+(-0.253548140518)*x[2]
        s[3,3,0,0]=(-0.218464111878)*x[0]**o+(-0.255988638178)*x[0]+(0.278107535041)*x[1]**o+(0.40326101595)*x[1]+(-0.2331761004)*x[2]**o+(-0.317553406707)*x[2]
        s[3,3,0,1]=(-0.537282071415)*x[0]**o+(0.601952446965)*x[0]+(0.381109335541)*x[1]**o+(0.669261809466)*x[1]+(0.945072492166)*x[2]**o+(0.853596132501)*x[2]
        s[3,3,1,0]=(0.847664465983)*x[0]**o+(0.8066212639)*x[0]+(-0.857726357776)*x[1]**o+(0.474266813247)*x[1]+(-0.201605613201)*x[2]**o+(-0.793421334897)*x[2]
        s[3,3,1,1]=(-0.420235472332)*x[0]**o+(0.691721844986)*x[0]+(-0.265495062345)*x[1]**o+(0.275291637819)*x[1]+(0.935918852363)*x[2]**o+(0.682857746339)*x[2]
        s[3,3,2,0]=(-0.507370968068)*x[0]**o+(0.73174079427)*x[0]+(0.260834780217)*x[1]**o+(0.0234116684785)*x[1]+(-0.259405028241)*x[2]**o+(-0.95866122285)*x[2]
        s[3,3,2,1]=(-0.739777982614)*x[0]**o+(-0.634784565277)*x[0]+(-0.459700756224)*x[1]**o+(-0.668028439149)*x[1]+(0.543643060627)*x[2]**o+(0.481120453087)*x[2]
        s[3,4,0,0]=(-0.564521203332)*x[0]**o+(0.875264891992)*x[0]+(-0.711596964515)*x[1]**o+(0.516599187629)*x[1]+(0.803187045453)*x[2]**o+(0.130416944939)*x[2]
        s[3,4,0,1]=(-0.233194158029)*x[0]**o+(0.124405220571)*x[0]+(-0.0969251448946)*x[1]**o+(-0.13351738113)*x[1]+(-0.426939325117)*x[2]**o+(-0.0127989581875)*x[2]
        s[3,4,1,0]=(-0.654825260357)*x[0]**o+(-0.073851613222)*x[0]+(-0.866970362806)*x[1]**o+(-0.417768390991)*x[1]+(0.876189107492)*x[2]**o+(0.57905826773)*x[2]
        s[3,4,1,1]=(0.62991933581)*x[0]**o+(0.0333499904703)*x[0]+(0.296962053291)*x[1]**o+(-0.203151609982)*x[1]+(-0.416709189249)*x[2]**o+(-0.815746328506)*x[2]
        s[3,4,2,0]=(-0.720775085901)*x[0]**o+(0.96283858374)*x[0]+(0.23423515114)*x[1]**o+(0.640634860355)*x[1]+(-0.819346396686)*x[2]**o+(-0.807002446047)*x[2]
        s[3,4,2,1]=(-0.0608202717996)*x[0]**o+(0.339885794306)*x[0]+(-0.317314034496)*x[1]**o+(-0.672508392809)*x[1]+(0.63517267097)*x[2]**o+(-0.00885735344423)*x[2]
        ref[0,0,0,0]=(0.284952784158)/(o+1.)+(0.592032235941)
        ref[0,0,0,1]=(-0.803451440889)/(o+1.)+(0.0224418156084)
        ref[0,0,1,0]=(0.638402520436)/(o+1.)+(0.0311038353355)
        ref[0,0,1,1]=(0.307272122952)/(o+1.)+(0.0104780524684)
        ref[0,0,2,0]=(-0.502107719245)/(o+1.)+(-0.515645130678)
        ref[0,0,2,1]=(1.81768999676)/(o+1.)+(0.146737736684)
        ref[0,1,0,0]=(0.647820033767)/(o+1.)+(0.391233423499)
        ref[0,1,0,1]=(0.436092125372)/(o+1.)+(0.106230136165)
        ref[0,1,1,0]=(1.28557716642)/(o+1.)+(0.0374861393842)
        ref[0,1,1,1]=(-0.0491708546408)/(o+1.)+(-0.295525989953)
        ref[0,1,2,0]=(1.70047149916)/(o+1.)+(-0.214455033205)
        ref[0,1,2,1]=(-1.14969489881)/(o+1.)+(0.588651941542)
        ref[0,2,0,0]=(-0.955825457958)/(o+1.)+(-0.0190999661809)
        ref[0,2,0,1]=(0.229929844366)/(o+1.)+(-0.30311165791)
        ref[0,2,1,0]=(-2.57462574161)/(o+1.)+(-0.197980362151)
        ref[0,2,1,1]=(-0.0419001621974)/(o+1.)+(-0.379828843312)
        ref[0,2,2,0]=(-1.18351050669)/(o+1.)+(0.449768277313)
        ref[0,2,2,1]=(1.87822513678)/(o+1.)+(0.622693263344)
        ref[0,3,0,0]=(-0.0192193497223)/(o+1.)+(-0.274332614077)
        ref[0,3,0,1]=(1.05610437756)/(o+1.)+(-0.0158328399745)
        ref[0,3,1,0]=(-0.331289176593)/(o+1.)+(-0.023771634635)
        ref[0,3,1,1]=(-0.886769594367)/(o+1.)+(0.0191560107572)
        ref[0,3,2,0]=(0.554322811543)/(o+1.)+(0.324586599411)
        ref[0,3,2,1]=(0.757364125068)/(o+1.)+(0.0103767719331)
        ref[0,4,0,0]=(1.43319447117)/(o+1.)+(0.489873010572)
        ref[0,4,0,1]=(-1.54717367706)/(o+1.)+(-0.327842424773)
        ref[0,4,1,0]=(-1.02675093873)/(o+1.)+(-0.206388699347)
        ref[0,4,1,1]=(0.757998827253)/(o+1.)+(-0.892263769148)
        ref[0,4,2,0]=(0.979004551362)/(o+1.)+(-0.801869395247)
        ref[0,4,2,1]=(-2.23425638002)/(o+1.)+(-1.20737466522)
        ref[1,0,0,0]=(0.515302843784)/(o+1.)+(-0.162675558238)
        ref[1,0,0,1]=(0.353191474341)/(o+1.)+(-0.310747202442)
        ref[1,0,1,0]=(0.11330663338)/(o+1.)+(-0.0294705495767)
        ref[1,0,1,1]=(0.0625851979387)/(o+1.)+(-0.6942697411)
        ref[1,0,2,0]=(-0.634856196829)/(o+1.)+(0.233519187832)
        ref[1,0,2,1]=(0.758008553799)/(o+1.)+(-0.22308748171)
        ref[1,1,0,0]=(0.172197996272)/(o+1.)+(0.864945864826)
        ref[1,1,0,1]=(-0.0930030974171)/(o+1.)+(0.444243956798)
        ref[1,1,1,0]=(-0.0301833795362)/(o+1.)+(0.404852886018)
        ref[1,1,1,1]=(0.100136649986)/(o+1.)+(1.28659380019)
        ref[1,1,2,0]=(-0.97900883033)/(o+1.)+(0.442440382588)
        ref[1,1,2,1]=(0.792392656026)/(o+1.)+(0.181706071395)
        ref[1,2,0,0]=(-0.453470794812)/(o+1.)+(0.111351337976)
        ref[1,2,0,1]=(1.56202220869)/(o+1.)+(-0.511185001555)
        ref[1,2,1,0]=(0.659810976864)/(o+1.)+(0.621766984989)
        ref[1,2,1,1]=(1.91781515919)/(o+1.)+(0.0691044934815)
        ref[1,2,2,0]=(0.611207129327)/(o+1.)+(-0.497184221995)
        ref[1,2,2,1]=(1.65028805445)/(o+1.)+(-0.0280822974985)
        ref[1,3,0,0]=(-1.28735344741)/(o+1.)+(-0.680605269048)
        ref[1,3,0,1]=(0.256443215625)/(o+1.)+(-0.161440006142)
        ref[1,3,1,0]=(0.561119040305)/(o+1.)+(0.227463879795)
        ref[1,3,1,1]=(0.327555206284)/(o+1.)+(0.29622756806)
        ref[1,3,2,0]=(0.0665337521742)/(o+1.)+(0.77129792303)
        ref[1,3,2,1]=(1.36356118866)/(o+1.)+(-0.0276174494925)
        ref[1,4,0,0]=(0.793908154771)/(o+1.)+(-0.703932363283)
        ref[1,4,0,1]=(0.67317694748)/(o+1.)+(0.419078081349)
        ref[1,4,1,0]=(-1.6259928924)/(o+1.)+(-0.0422868647694)
        ref[1,4,1,1]=(0.573888422973)/(o+1.)+(-0.0709319079127)
        ref[1,4,2,0]=(-1.32097669048)/(o+1.)+(0.981719850544)
        ref[1,4,2,1]=(1.31804256932)/(o+1.)+(-0.464008585329)
        ref[2,0,0,0]=(0.131285227429)/(o+1.)+(-0.27188947549)
        ref[2,0,0,1]=(-0.534412041331)/(o+1.)+(-0.502296977244)
        ref[2,0,1,0]=(0.0835036877928)/(o+1.)+(0.675471052627)
        ref[2,0,1,1]=(-1.0534161098)/(o+1.)+(0.312025059983)
        ref[2,0,2,0]=(0.774008222845)/(o+1.)+(0.668233731266)
        ref[2,0,2,1]=(0.273211710533)/(o+1.)+(-0.656092090992)
        ref[2,1,0,0]=(0.0276836248637)/(o+1.)+(0.0865731702563)
        ref[2,1,0,1]=(-1.45387179966)/(o+1.)+(-0.198937400068)
        ref[2,1,1,0]=(-0.870829853631)/(o+1.)+(-0.523273805792)
        ref[2,1,1,1]=(0.987550331752)/(o+1.)+(-0.871586545966)
        ref[2,1,2,0]=(-0.188404489031)/(o+1.)+(-0.917217646504)
        ref[2,1,2,1]=(1.01318413367)/(o+1.)+(-0.338229998699)
        ref[2,2,0,0]=(-1.14509068713)/(o+1.)+(-0.141129573573)
        ref[2,2,0,1]=(-0.303957214603)/(o+1.)+(-0.966870476981)
        ref[2,2,1,0]=(0.280958273221)/(o+1.)+(0.751037367762)
        ref[2,2,1,1]=(0.761765995326)/(o+1.)+(0.499930134158)
        ref[2,2,2,0]=(-0.0631107493882)/(o+1.)+(0.923031517168)
        ref[2,2,2,1]=(-1.47548256318)/(o+1.)+(-0.516364377092)
        ref[2,3,0,0]=(0.357762298985)/(o+1.)+(0.0357748471536)
        ref[2,3,0,1]=(-1.32702449191)/(o+1.)+(0.569741861035)
        ref[2,3,1,0]=(0.846056405606)/(o+1.)+(0.304580354023)
        ref[2,3,1,1]=(0.842780955061)/(o+1.)+(-0.312530078358)
        ref[2,3,2,0]=(-0.292498051768)/(o+1.)+(-0.319533489098)
        ref[2,3,2,1]=(0.150081567052)/(o+1.)+(0.376801600846)
        ref[2,4,0,0]=(-0.790764630964)/(o+1.)+(0.501833028068)
        ref[2,4,0,1]=(-0.105229058628)/(o+1.)+(0.121379861433)
        ref[2,4,1,0]=(-2.34280312721)/(o+1.)+(-0.557544980779)
        ref[2,4,1,1]=(-0.790806798558)/(o+1.)+(-0.0895273554309)
        ref[2,4,2,0]=(0.912612891478)/(o+1.)+(0.0673322097612)
        ref[2,4,2,1]=(-0.45109893498)/(o+1.)+(-0.213907437562)
        ref[3,0,0,0]=(-1.31264733132)/(o+1.)+(0.743071295917)
        ref[3,0,0,1]=(-1.48221918925)/(o+1.)+(0.884106628208)
        ref[3,0,1,0]=(-0.119070635332)/(o+1.)+(0.21944924039)
        ref[3,0,1,1]=(-0.332877764028)/(o+1.)+(-0.605759458881)
        ref[3,0,2,0]=(0.0337251954668)/(o+1.)+(-0.283202055581)
        ref[3,0,2,1]=(-0.64774468756)/(o+1.)+(-0.465566129216)
        ref[3,1,0,0]=(-1.18814787487)/(o+1.)+(1.10275852067)
        ref[3,1,0,1]=(0.43972319733)/(o+1.)+(0.0116527219252)
        ref[3,1,1,0]=(0.0933525402037)/(o+1.)+(0.644125416616)
        ref[3,1,1,1]=(0.350484486401)/(o+1.)+(-0.161486299395)
        ref[3,1,2,0]=(-0.90157200233)/(o+1.)+(-0.377052217418)
        ref[3,1,2,1]=(-0.442791953392)/(o+1.)+(-0.0467486131556)
        ref[3,2,0,0]=(0.494289265635)/(o+1.)+(0.67957645667)
        ref[3,2,0,1]=(1.21890414616)/(o+1.)+(-0.0371819889285)
        ref[3,2,1,0]=(-0.143804539219)/(o+1.)+(-0.833417204036)
        ref[3,2,1,1]=(-1.30537592015)/(o+1.)+(-0.393002931773)
        ref[3,2,2,0]=(-1.7656669004)/(o+1.)+(-0.0325619704518)
        ref[3,2,2,1]=(1.39446290966)/(o+1.)+(0.635745200922)
        ref[3,3,0,0]=(-0.173532677236)/(o+1.)+(-0.0851405144673)
        ref[3,3,0,1]=(0.788899756292)/(o+1.)+(1.06240519447)
        ref[3,3,1,0]=(-0.211667504994)/(o+1.)+(0.243733371125)
        ref[3,3,1,1]=(0.250188317686)/(o+1.)+(0.824935614572)
        ref[3,3,2,0]=(-0.505941216092)/(o+1.)+(-0.101754380051)
        ref[3,3,2,1]=(-0.655835678211)/(o+1.)+(-0.410846275669)
        ref[3,4,0,0]=(-0.472931122393)/(o+1.)+(0.76114051228)
        ref[3,4,0,1]=(-0.757058628041)/(o+1.)+(-0.0109555593731)
        ref[3,4,1,0]=(-0.645606515671)/(o+1.)+(0.0437191317586)
        ref[3,4,1,1]=(0.510172199852)/(o+1.)+(-0.492773974009)
        ref[3,4,2,0]=(-1.30588633145)/(o+1.)+(0.398235499024)
        ref[3,4,2,1]=(0.257038364674)/(o+1.)+(-0.170739975974)
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.506621877215)*x[0]**o+(-0.45560383685)*x[0]+(-0.598137301806)*x[1]**o+(0.784241595615)*x[1]
        ref=(-0.0915154245913)*(1+2.*(dim-1.)/(o+1.))+(0.328637758765)*dim
      else:
        arg=(-0.505489131484)*x[0]**o+(0.58139159259)*x[0]+(0.458086828888)*x[1]**o+(0.549090746672)*x[1]+(-0.0882325936167)*x[2]**o+(0.0170295954808)*x[2]
        ref=(-0.135634896213)*(1+2.*(dim-1.)/(o+1.))+(1.14751193474)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.88779433873)*x[0]**o+(0.507955012301)*x[0]+(0.652739274924)*x[1]**o+(-0.971975755989)*x[1]
        arg[1]=(-0.0812690931139)*x[0]**o+(-0.283880962354)*x[0]+(0.094904517439)*x[1]**o+(-0.940528164353)*x[1]
        ref[0]=(-0.235055063806)*(1+2.*(dim-1.)/(o+1.))+(-0.464020743688)*dim
        ref[1]=(0.0136354243251)*(1+2.*(dim-1.)/(o+1.))+(-1.22440912671)*dim
      else:
        arg[0]=(-0.506412384389)*x[0]**o+(-0.465794405394)*x[0]+(0.346977499982)*x[1]**o+(0.386181176987)*x[1]+(-0.476729572397)*x[2]**o+(-0.956945933508)*x[2]
        arg[1]=(0.267092807416)*x[0]**o+(-0.54211155802)*x[0]+(-0.27815411727)*x[1]**o+(-0.984542450102)*x[1]+(0.656389296076)*x[2]**o+(0.0395863971401)*x[2]
        ref[0]=(-0.636164456804)*(1+2.*(dim-1.)/(o+1.))+(-1.03655916191)*dim
        ref[1]=(0.645327986222)*(1+2.*(dim-1.)/(o+1.))+(-1.48706761098)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.0607636892262)*x[0]**o+(0.740866012668)*x[0]+(0.145223136928)*x[1]**o+(-0.738387773526)*x[1]
        arg[0,1]=(0.0292598484511)*x[0]**o+(-0.189208290889)*x[0]+(0.902021405165)*x[1]**o+(0.423106545559)*x[1]
        arg[0,2]=(0.519658217226)*x[0]**o+(-0.768406018136)*x[0]+(-0.189532527097)*x[1]**o+(0.256417429833)*x[1]
        arg[0,3]=(-0.565327663481)*x[0]**o+(0.876762003673)*x[0]+(-0.243804758063)*x[1]**o+(0.399476036454)*x[1]
        arg[0,4]=(0.0632252409818)*x[0]**o+(0.0951130662538)*x[0]+(0.185662425763)*x[1]**o+(0.907446482506)*x[1]
        arg[1,0]=(-0.430504584794)*x[0]**o+(0.864814314023)*x[0]+(0.47460095501)*x[1]**o+(-0.599679142188)*x[1]
        arg[1,1]=(0.699618175298)*x[0]**o+(-0.561782605517)*x[0]+(-0.304250865015)*x[1]**o+(-0.341836776795)*x[1]
        arg[1,2]=(0.856956767316)*x[0]**o+(0.610026903213)*x[0]+(-0.666749954599)*x[1]**o+(0.885388885234)*x[1]
        arg[1,3]=(0.699377007631)*x[0]**o+(0.699906750287)*x[0]+(0.486690669414)*x[1]**o+(0.167425901549)*x[1]
        arg[1,4]=(0.466701415896)*x[0]**o+(-0.0719752759126)*x[0]+(0.795662047062)*x[1]**o+(0.979862503926)*x[1]
        arg[2,0]=(-0.912218732612)*x[0]**o+(0.950990472894)*x[0]+(-0.527999708902)*x[1]**o+(0.695735593043)*x[1]
        arg[2,1]=(-0.223468270048)*x[0]**o+(-0.882565147673)*x[0]+(0.750516681645)*x[1]**o+(-0.652794590093)*x[1]
        arg[2,2]=(0.897646682412)*x[0]**o+(-0.815401214058)*x[0]+(0.629680741327)*x[1]**o+(-0.9716988156)*x[1]
        arg[2,3]=(0.279739517524)*x[0]**o+(-0.0859797206016)*x[0]+(-0.0454325622757)*x[1]**o+(0.301214753868)*x[1]
        arg[2,4]=(-0.358056827092)*x[0]**o+(0.254207762658)*x[0]+(0.831499219864)*x[1]**o+(0.0930388028811)*x[1]
        arg[3,0]=(0.0761520020047)*x[0]**o+(0.650696394486)*x[0]+(-0.132369660678)*x[1]**o+(-0.321220385971)*x[1]
        arg[3,1]=(0.238715231585)*x[0]**o+(-0.297219391083)*x[0]+(-0.15446552614)*x[1]**o+(0.860560615056)*x[1]
        arg[3,2]=(0.708864754538)*x[0]**o+(-0.221186702958)*x[0]+(-0.478127729891)*x[1]**o+(0.0309604412212)*x[1]
        arg[3,3]=(0.362909366479)*x[0]**o+(0.382279106813)*x[0]+(0.827905062086)*x[1]**o+(-0.303048355949)*x[1]
        arg[3,4]=(0.573736946844)*x[0]**o+(0.645059764766)*x[0]+(0.301945383678)*x[1]**o+(-0.103083649994)*x[1]
        ref[0,0]=(0.205986826154)*(1+2.*(dim-1.)/(o+1.))+(0.0024782391428)*dim
        ref[0,1]=(0.931281253617)*(1+2.*(dim-1.)/(o+1.))+(0.23389825467)*dim
        ref[0,2]=(0.330125690129)*(1+2.*(dim-1.)/(o+1.))+(-0.511988588303)*dim
        ref[0,3]=(-0.809132421544)*(1+2.*(dim-1.)/(o+1.))+(1.27623804013)*dim
        ref[0,4]=(0.248887666745)*(1+2.*(dim-1.)/(o+1.))+(1.00255954876)*dim
        ref[1,0]=(0.044096370216)*(1+2.*(dim-1.)/(o+1.))+(0.265135171835)*dim
        ref[1,1]=(0.395367310283)*(1+2.*(dim-1.)/(o+1.))+(-0.903619382312)*dim
        ref[1,2]=(0.190206812717)*(1+2.*(dim-1.)/(o+1.))+(1.49541578845)*dim
        ref[1,3]=(1.18606767705)*(1+2.*(dim-1.)/(o+1.))+(0.867332651836)*dim
        ref[1,4]=(1.26236346296)*(1+2.*(dim-1.)/(o+1.))+(0.907887228013)*dim
        ref[2,0]=(-1.44021844151)*(1+2.*(dim-1.)/(o+1.))+(1.64672606594)*dim
        ref[2,1]=(0.527048411596)*(1+2.*(dim-1.)/(o+1.))+(-1.53535973777)*dim
        ref[2,2]=(1.52732742374)*(1+2.*(dim-1.)/(o+1.))+(-1.78710002966)*dim
        ref[2,3]=(0.234306955249)*(1+2.*(dim-1.)/(o+1.))+(0.215235033266)*dim
        ref[2,4]=(0.473442392772)*(1+2.*(dim-1.)/(o+1.))+(0.34724656554)*dim
        ref[3,0]=(-0.0562176586729)*(1+2.*(dim-1.)/(o+1.))+(0.329476008516)*dim
        ref[3,1]=(0.0842497054449)*(1+2.*(dim-1.)/(o+1.))+(0.563341223973)*dim
        ref[3,2]=(0.230737024647)*(1+2.*(dim-1.)/(o+1.))+(-0.190226261737)*dim
        ref[3,3]=(1.19081442856)*(1+2.*(dim-1.)/(o+1.))+(0.079230750864)*dim
        ref[3,4]=(0.875682330522)*(1+2.*(dim-1.)/(o+1.))+(0.541976114772)*dim
      else:
        arg[0,0]=(0.0640108068236)*x[0]**o+(0.0322830166043)*x[0]+(-0.703691335255)*x[1]**o+(-0.185228770711)*x[1]+(-0.738714166313)*x[2]**o+(-0.423576647077)*x[2]
        arg[0,1]=(-0.738341351945)*x[0]**o+(0.396306395436)*x[0]+(-0.369974848568)*x[1]**o+(0.750933208783)*x[1]+(0.554069869127)*x[2]**o+(0.634275493051)*x[2]
        arg[0,2]=(-0.259193248708)*x[0]**o+(-0.900007449583)*x[0]+(0.894379436915)*x[1]**o+(0.997814320752)*x[1]+(0.233192116622)*x[2]**o+(0.986616259649)*x[2]
        arg[0,3]=(0.381480250594)*x[0]**o+(0.734823219057)*x[0]+(0.569912440169)*x[1]**o+(-0.441084504695)*x[1]+(0.312539701584)*x[2]**o+(-0.3733004974)*x[2]
        arg[0,4]=(-0.779059807622)*x[0]**o+(0.538777922322)*x[0]+(-0.0695879211612)*x[1]**o+(-0.204862216499)*x[1]+(-0.0109698525672)*x[2]**o+(0.981875117617)*x[2]
        arg[1,0]=(-0.415857630044)*x[0]**o+(-0.356607866913)*x[0]+(-0.628854326903)*x[1]**o+(0.0401233333505)*x[1]+(0.687470169579)*x[2]**o+(-0.565349753114)*x[2]
        arg[1,1]=(-0.438364910402)*x[0]**o+(0.498078194367)*x[0]+(-0.120729597948)*x[1]**o+(-0.622076117713)*x[1]+(0.995130029172)*x[2]**o+(-0.7234601064)*x[2]
        arg[1,2]=(-0.957869585558)*x[0]**o+(0.374693219824)*x[0]+(0.0355192120972)*x[1]**o+(-0.516770140651)*x[1]+(0.521887953791)*x[2]**o+(0.00771564236502)*x[2]
        arg[1,3]=(-0.0495580194062)*x[0]**o+(-0.183895326318)*x[0]+(-0.55304086303)*x[1]**o+(0.920204449879)*x[1]+(0.143165855157)*x[2]**o+(-0.773684620462)*x[2]
        arg[1,4]=(0.795913131087)*x[0]**o+(0.653826129199)*x[0]+(-0.585562760813)*x[1]**o+(0.274083108918)*x[1]+(-0.165275654116)*x[2]**o+(0.358678758212)*x[2]
        arg[2,0]=(-0.0699685857843)*x[0]**o+(-0.42922435931)*x[0]+(-0.0494308981938)*x[1]**o+(-0.956597829176)*x[1]+(-0.912941305307)*x[2]**o+(0.214777332562)*x[2]
        arg[2,1]=(-0.900983338173)*x[0]**o+(-0.779729884378)*x[0]+(-0.860100306508)*x[1]**o+(0.14211537281)*x[1]+(-0.984439938285)*x[2]**o+(-0.549987779842)*x[2]
        arg[2,2]=(0.801489512389)*x[0]**o+(0.896468923826)*x[0]+(-0.155737972887)*x[1]**o+(0.639864330477)*x[1]+(0.798598705038)*x[2]**o+(-0.863019041525)*x[2]
        arg[2,3]=(-0.453612136504)*x[0]**o+(0.522662491688)*x[0]+(-0.463058076486)*x[1]**o+(0.398018809499)*x[1]+(-0.612044961611)*x[2]**o+(-0.481951883217)*x[2]
        arg[2,4]=(-0.0739336519425)*x[0]**o+(-0.241753740154)*x[0]+(0.0176292058898)*x[1]**o+(0.435641132809)*x[1]+(-0.0592828584976)*x[2]**o+(0.430115620801)*x[2]
        arg[3,0]=(0.604784358152)*x[0]**o+(-0.817214179503)*x[0]+(-0.423104272008)*x[1]**o+(0.287787881773)*x[1]+(0.678036469241)*x[2]**o+(0.578347129563)*x[2]
        arg[3,1]=(0.362390778598)*x[0]**o+(-0.389066580119)*x[0]+(-0.934087580396)*x[1]**o+(0.907587485226)*x[1]+(-0.902083043406)*x[2]**o+(0.904707538314)*x[2]
        arg[3,2]=(0.862463207465)*x[0]**o+(0.904923219336)*x[0]+(0.421361696042)*x[1]**o+(-0.779329196171)*x[1]+(0.059521522394)*x[2]**o+(-0.744510067387)*x[2]
        arg[3,3]=(0.412029226851)*x[0]**o+(-0.649730143511)*x[0]+(0.579628017001)*x[1]**o+(-0.949180147129)*x[1]+(-0.073016492309)*x[2]**o+(0.666554170757)*x[2]
        arg[3,4]=(-0.700401396504)*x[0]**o+(-0.89194120812)*x[0]+(0.592387042213)*x[1]**o+(-0.906632787891)*x[1]+(-0.486072860812)*x[2]**o+(-0.505386176523)*x[2]
        ref[0,0]=(-1.37839469474)*(1+2.*(dim-1.)/(o+1.))+(-0.576522401184)*dim
        ref[0,1]=(-0.554246331386)*(1+2.*(dim-1.)/(o+1.))+(1.78151509727)*dim
        ref[0,2]=(0.868378304829)*(1+2.*(dim-1.)/(o+1.))+(1.08442313082)*dim
        ref[0,3]=(1.26393239235)*(1+2.*(dim-1.)/(o+1.))+(-0.0795617830385)*dim
        ref[0,4]=(-0.85961758135)*(1+2.*(dim-1.)/(o+1.))+(1.31579082344)*dim
        ref[1,0]=(-0.357241787368)*(1+2.*(dim-1.)/(o+1.))+(-0.881834286676)*dim
        ref[1,1]=(0.436035520821)*(1+2.*(dim-1.)/(o+1.))+(-0.847458029746)*dim
        ref[1,2]=(-0.40046241967)*(1+2.*(dim-1.)/(o+1.))+(-0.134361278462)*dim
        ref[1,3]=(-0.459433027279)*(1+2.*(dim-1.)/(o+1.))+(-0.0373754969011)*dim
        ref[1,4]=(0.0450747161574)*(1+2.*(dim-1.)/(o+1.))+(1.28658799633)*dim
        ref[2,0]=(-1.03234078929)*(1+2.*(dim-1.)/(o+1.))+(-1.17104485592)*dim
        ref[2,1]=(-2.74552358297)*(1+2.*(dim-1.)/(o+1.))+(-1.18760229141)*dim
        ref[2,2]=(1.44435024454)*(1+2.*(dim-1.)/(o+1.))+(0.673314212777)*dim
        ref[2,3]=(-1.5287151746)*(1+2.*(dim-1.)/(o+1.))+(0.43872941797)*dim
        ref[2,4]=(-0.11558730455)*(1+2.*(dim-1.)/(o+1.))+(0.624003013456)*dim
        ref[3,0]=(0.859716555385)*(1+2.*(dim-1.)/(o+1.))+(0.0489208318321)*dim
        ref[3,1]=(-1.4737798452)*(1+2.*(dim-1.)/(o+1.))+(1.42322844342)*dim
        ref[3,2]=(1.3433464259)*(1+2.*(dim-1.)/(o+1.))+(-0.618916044222)*dim
        ref[3,3]=(0.918640751542)*(1+2.*(dim-1.)/(o+1.))+(-0.932356119883)*dim
        ref[3,4]=(-0.594087215103)*(1+2.*(dim-1.)/(o+1.))+(-2.30396017253)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.662707930283)*x[0]**o+(0.625512077969)*x[0]+(-0.222208840922)*x[1]**o+(0.374614317804)*x[1]
        arg[0,0,1]=(0.685952252808)*x[0]**o+(-0.88219802797)*x[0]+(-0.948918831634)*x[1]**o+(-0.00615202896776)*x[1]
        arg[0,1,0]=(-0.555231206657)*x[0]**o+(0.997112425959)*x[0]+(-0.186247870511)*x[1]**o+(-0.0658892046229)*x[1]
        arg[0,1,1]=(-0.735288285178)*x[0]**o+(0.941234848614)*x[0]+(-0.307761737256)*x[1]**o+(0.933293069433)*x[1]
        arg[1,0,0]=(-0.618191337825)*x[0]**o+(-0.12930895166)*x[0]+(0.76213325083)*x[1]**o+(-0.744049403156)*x[1]
        arg[1,0,1]=(-0.793412570981)*x[0]**o+(0.533175323401)*x[0]+(-0.87398581932)*x[1]**o+(-0.279651754725)*x[1]
        arg[1,1,0]=(-0.0654077907694)*x[0]**o+(0.239040458488)*x[0]+(0.730639927837)*x[1]**o+(0.805020533329)*x[1]
        arg[1,1,1]=(-0.560195748542)*x[0]**o+(-0.265389430135)*x[0]+(-0.295142324741)*x[1]**o+(-0.929699771024)*x[1]
        arg[2,0,0]=(0.935760865577)*x[0]**o+(-0.332458843131)*x[0]+(0.257302695187)*x[1]**o+(-0.347547927033)*x[1]
        arg[2,0,1]=(-0.245770653276)*x[0]**o+(0.126945988409)*x[0]+(0.302793434995)*x[1]**o+(-0.997483200172)*x[1]
        arg[2,1,0]=(0.523494364951)*x[0]**o+(-0.323782740449)*x[0]+(-0.563752593454)*x[1]**o+(-0.358325752838)*x[1]
        arg[2,1,1]=(-0.506330114515)*x[0]**o+(-0.733234050567)*x[0]+(0.734774667778)*x[1]**o+(0.795264566776)*x[1]
        arg[3,0,0]=(-0.21258295235)*x[0]**o+(-0.929331330718)*x[0]+(-0.546061847742)*x[1]**o+(0.0613931491065)*x[1]
        arg[3,0,1]=(-0.357847946656)*x[0]**o+(-0.133795684508)*x[0]+(-0.796646028688)*x[1]**o+(0.875795801429)*x[1]
        arg[3,1,0]=(0.923717804039)*x[0]**o+(-0.176266633621)*x[0]+(-0.605892867534)*x[1]**o+(0.633377612365)*x[1]
        arg[3,1,1]=(0.729025508129)*x[0]**o+(0.412034955257)*x[0]+(0.0933054864485)*x[1]**o+(0.386019085532)*x[1]
        arg[4,0,0]=(-0.0705524918008)*x[0]**o+(-0.869605321304)*x[0]+(0.244351267755)*x[1]**o+(0.898883388102)*x[1]
        arg[4,0,1]=(-0.453297333327)*x[0]**o+(-0.67824185123)*x[0]+(0.0497322840245)*x[1]**o+(0.581677879069)*x[1]
        arg[4,1,0]=(0.343975006358)*x[0]**o+(0.724084357103)*x[0]+(-0.186292335362)*x[1]**o+(-0.257997075018)*x[1]
        arg[4,1,1]=(-0.389731923798)*x[0]**o+(0.0949155578616)*x[0]+(-0.298428138752)*x[1]**o+(0.819510069666)*x[1]
        arg[5,0,0]=(0.827670169513)*x[0]**o+(-0.584607978668)*x[0]+(0.669056948658)*x[1]**o+(-0.393158541985)*x[1]
        arg[5,0,1]=(-0.471551968206)*x[0]**o+(-0.339964831396)*x[0]+(0.354554393554)*x[1]**o+(-0.334773591957)*x[1]
        arg[5,1,0]=(0.113562904864)*x[0]**o+(-0.775045505123)*x[0]+(0.783044874931)*x[1]**o+(0.52265843873)*x[1]
        arg[5,1,1]=(-0.115476036205)*x[0]**o+(-0.290822012728)*x[0]+(-0.398510477854)*x[1]**o+(0.398737205542)*x[1]
        ref[0,0,0]=(0.440499089362)*(1+2.*(dim-1.)/(o+1.))+(1.00012639577)*dim
        ref[0,0,1]=(-0.262966578826)*(1+2.*(dim-1.)/(o+1.))+(-0.888350056938)*dim
        ref[0,1,0]=(-0.741479077167)*(1+2.*(dim-1.)/(o+1.))+(0.931223221336)*dim
        ref[0,1,1]=(-1.04305002243)*(1+2.*(dim-1.)/(o+1.))+(1.87452791805)*dim
        ref[1,0,0]=(0.143941913005)*(1+2.*(dim-1.)/(o+1.))+(-0.873358354816)*dim
        ref[1,0,1]=(-1.6673983903)*(1+2.*(dim-1.)/(o+1.))+(0.253523568676)*dim
        ref[1,1,0]=(0.665232137068)*(1+2.*(dim-1.)/(o+1.))+(1.04406099182)*dim
        ref[1,1,1]=(-0.855338073283)*(1+2.*(dim-1.)/(o+1.))+(-1.19508920116)*dim
        ref[2,0,0]=(1.19306356076)*(1+2.*(dim-1.)/(o+1.))+(-0.680006770164)*dim
        ref[2,0,1]=(0.0570227817187)*(1+2.*(dim-1.)/(o+1.))+(-0.870537211764)*dim
        ref[2,1,0]=(-0.0402582285028)*(1+2.*(dim-1.)/(o+1.))+(-0.682108493287)*dim
        ref[2,1,1]=(0.228444553263)*(1+2.*(dim-1.)/(o+1.))+(0.0620305162089)*dim
        ref[3,0,0]=(-0.758644800093)*(1+2.*(dim-1.)/(o+1.))+(-0.867938181611)*dim
        ref[3,0,1]=(-1.15449397534)*(1+2.*(dim-1.)/(o+1.))+(0.742000116921)*dim
        ref[3,1,0]=(0.317824936505)*(1+2.*(dim-1.)/(o+1.))+(0.457110978744)*dim
        ref[3,1,1]=(0.822330994578)*(1+2.*(dim-1.)/(o+1.))+(0.798054040789)*dim
        ref[4,0,0]=(0.173798775954)*(1+2.*(dim-1.)/(o+1.))+(0.0292780667983)*dim
        ref[4,0,1]=(-0.403565049303)*(1+2.*(dim-1.)/(o+1.))+(-0.0965639721601)*dim
        ref[4,1,0]=(0.157682670995)*(1+2.*(dim-1.)/(o+1.))+(0.466087282085)*dim
        ref[4,1,1]=(-0.68816006255)*(1+2.*(dim-1.)/(o+1.))+(0.914425627528)*dim
        ref[5,0,0]=(1.49672711817)*(1+2.*(dim-1.)/(o+1.))+(-0.977766520654)*dim
        ref[5,0,1]=(-0.116997574651)*(1+2.*(dim-1.)/(o+1.))+(-0.674738423353)*dim
        ref[5,1,0]=(0.896607779795)*(1+2.*(dim-1.)/(o+1.))+(-0.252387066394)*dim
        ref[5,1,1]=(-0.513986514058)*(1+2.*(dim-1.)/(o+1.))+(0.107915192814)*dim
      else:
        arg[0,0,0]=(-0.418671467289)*x[0]**o+(0.498387164484)*x[0]+(0.208896299314)*x[1]**o+(-0.643038678771)*x[1]+(-0.85475466439)*x[2]**o+(0.59586368009)*x[2]
        arg[0,0,1]=(0.149229891253)*x[0]**o+(0.229836799955)*x[0]+(0.557946253001)*x[1]**o+(-0.434223003739)*x[1]+(-0.250979831671)*x[2]**o+(0.762582320269)*x[2]
        arg[0,1,0]=(-0.950902166094)*x[0]**o+(-0.312913424444)*x[0]+(-0.573887973428)*x[1]**o+(-0.372485255069)*x[1]+(-0.18228472562)*x[2]**o+(-0.566679093436)*x[2]
        arg[0,1,1]=(-0.191769651257)*x[0]**o+(-0.344504306057)*x[0]+(0.355926399715)*x[1]**o+(-0.721341598375)*x[1]+(0.362401271289)*x[2]**o+(0.601351093654)*x[2]
        arg[1,0,0]=(-0.375572370375)*x[0]**o+(-0.532723139333)*x[0]+(-0.738166755424)*x[1]**o+(0.785506757134)*x[1]+(0.192667976423)*x[2]**o+(-0.06534893872)*x[2]
        arg[1,0,1]=(-0.353750712453)*x[0]**o+(0.401872866285)*x[0]+(-0.747783222746)*x[1]**o+(0.663116752574)*x[1]+(-0.162954382466)*x[2]**o+(0.318112656687)*x[2]
        arg[1,1,0]=(-0.711345828912)*x[0]**o+(0.764701438614)*x[0]+(-0.232576947703)*x[1]**o+(-0.50541791533)*x[1]+(0.604603682696)*x[2]**o+(0.498828792648)*x[2]
        arg[1,1,1]=(-0.72474879447)*x[0]**o+(0.888266999605)*x[0]+(0.639903282531)*x[1]**o+(-0.777435844156)*x[1]+(-0.0814738680964)*x[2]**o+(-0.0135286853237)*x[2]
        arg[2,0,0]=(0.839793308998)*x[0]**o+(0.374051014226)*x[0]+(0.594653957466)*x[1]**o+(-0.962352782345)*x[1]+(0.888632514336)*x[2]**o+(0.491195067464)*x[2]
        arg[2,0,1]=(-0.827615661598)*x[0]**o+(0.675227575488)*x[0]+(-0.93068580658)*x[1]**o+(0.694434337954)*x[1]+(0.3583179263)*x[2]**o+(0.775147111187)*x[2]
        arg[2,1,0]=(-0.772656999491)*x[0]**o+(-0.464476706435)*x[0]+(0.569832694977)*x[1]**o+(0.500517676542)*x[1]+(-0.272153062279)*x[2]**o+(-0.398436498506)*x[2]
        arg[2,1,1]=(-0.3065198839)*x[0]**o+(-0.454353346315)*x[0]+(-0.539448433869)*x[1]**o+(0.776950547078)*x[1]+(-0.987178147402)*x[2]**o+(0.368935549126)*x[2]
        arg[3,0,0]=(-0.516131622594)*x[0]**o+(-0.680922252714)*x[0]+(0.304951059501)*x[1]**o+(-0.236361364379)*x[1]+(-0.37983306233)*x[2]**o+(0.693426329316)*x[2]
        arg[3,0,1]=(0.85094983056)*x[0]**o+(-0.356782375867)*x[0]+(0.362116478047)*x[1]**o+(0.269126381543)*x[1]+(0.950254707197)*x[2]**o+(0.592028802301)*x[2]
        arg[3,1,0]=(-0.17721437307)*x[0]**o+(-0.127432965766)*x[0]+(-0.754975923735)*x[1]**o+(0.755938625221)*x[1]+(0.776393089777)*x[2]**o+(0.913028629874)*x[2]
        arg[3,1,1]=(-0.779840479711)*x[0]**o+(0.432392796542)*x[0]+(0.0457670482312)*x[1]**o+(0.493363120993)*x[1]+(-0.290308698356)*x[2]**o+(0.263565329052)*x[2]
        arg[4,0,0]=(-0.594371032991)*x[0]**o+(0.673573089348)*x[0]+(0.0301238273919)*x[1]**o+(-0.900766544194)*x[1]+(0.937125579182)*x[2]**o+(-0.360512287398)*x[2]
        arg[4,0,1]=(-0.187561420369)*x[0]**o+(-0.727058667406)*x[0]+(0.643904893485)*x[1]**o+(0.547358609515)*x[1]+(-0.810276689361)*x[2]**o+(0.347967967049)*x[2]
        arg[4,1,0]=(0.0261452776327)*x[0]**o+(0.630125453041)*x[0]+(0.884509651558)*x[1]**o+(-0.580251885597)*x[1]+(0.141608918047)*x[2]**o+(0.417140237181)*x[2]
        arg[4,1,1]=(0.943149763428)*x[0]**o+(0.0414421588123)*x[0]+(-0.453944232191)*x[1]**o+(0.817963360891)*x[1]+(-0.0132464046434)*x[2]**o+(-0.247024507218)*x[2]
        arg[5,0,0]=(0.993356103007)*x[0]**o+(-0.604991443085)*x[0]+(-0.0141942193536)*x[1]**o+(-0.467982818195)*x[1]+(0.907608332852)*x[2]**o+(-0.918494340721)*x[2]
        arg[5,0,1]=(0.782079295393)*x[0]**o+(0.0125642236587)*x[0]+(-0.512764292821)*x[1]**o+(-0.0699705327379)*x[1]+(0.141341885735)*x[2]**o+(0.420508183565)*x[2]
        arg[5,1,0]=(-0.835665486427)*x[0]**o+(-0.536850925449)*x[0]+(-0.0958499285885)*x[1]**o+(-0.541119926637)*x[1]+(0.341886033336)*x[2]**o+(-0.772823771771)*x[2]
        arg[5,1,1]=(-0.391737530391)*x[0]**o+(-0.160045087084)*x[0]+(-0.984350699381)*x[1]**o+(0.303782477449)*x[1]+(-0.279906615513)*x[2]**o+(-0.690872289955)*x[2]
        ref[0,0,0]=(-1.06452983237)*(1+2.*(dim-1.)/(o+1.))+(0.451212165803)*dim
        ref[0,0,1]=(0.456196312583)*(1+2.*(dim-1.)/(o+1.))+(0.558196116486)*dim
        ref[0,1,0]=(-1.70707486514)*(1+2.*(dim-1.)/(o+1.))+(-1.25207777295)*dim
        ref[0,1,1]=(0.526558019747)*(1+2.*(dim-1.)/(o+1.))+(-0.464494810779)*dim
        ref[1,0,0]=(-0.921071149376)*(1+2.*(dim-1.)/(o+1.))+(0.187434679081)*dim
        ref[1,0,1]=(-1.26448831766)*(1+2.*(dim-1.)/(o+1.))+(1.38310227555)*dim
        ref[1,1,0]=(-0.339319093919)*(1+2.*(dim-1.)/(o+1.))+(0.758112315932)*dim
        ref[1,1,1]=(-0.166319380035)*(1+2.*(dim-1.)/(o+1.))+(0.0973024701251)*dim
        ref[2,0,0]=(2.3230797808)*(1+2.*(dim-1.)/(o+1.))+(-0.0971067006546)*dim
        ref[2,0,1]=(-1.39998354188)*(1+2.*(dim-1.)/(o+1.))+(2.14480902463)*dim
        ref[2,1,0]=(-0.474977366793)*(1+2.*(dim-1.)/(o+1.))+(-0.362395528399)*dim
        ref[2,1,1]=(-1.83314646517)*(1+2.*(dim-1.)/(o+1.))+(0.69153274989)*dim
        ref[3,0,0]=(-0.591013625423)*(1+2.*(dim-1.)/(o+1.))+(-0.223857287777)*dim
        ref[3,0,1]=(2.1633210158)*(1+2.*(dim-1.)/(o+1.))+(0.504372807977)*dim
        ref[3,1,0]=(-0.155797207029)*(1+2.*(dim-1.)/(o+1.))+(1.54153428933)*dim
        ref[3,1,1]=(-1.02438212984)*(1+2.*(dim-1.)/(o+1.))+(1.18932124659)*dim
        ref[4,0,0]=(0.372878373583)*(1+2.*(dim-1.)/(o+1.))+(-0.587705742243)*dim
        ref[4,0,1]=(-0.353933216245)*(1+2.*(dim-1.)/(o+1.))+(0.168267909158)*dim
        ref[4,1,0]=(1.05226384724)*(1+2.*(dim-1.)/(o+1.))+(0.467013804625)*dim
        ref[4,1,1]=(0.475959126593)*(1+2.*(dim-1.)/(o+1.))+(0.612381012486)*dim
        ref[5,0,0]=(1.88677021651)*(1+2.*(dim-1.)/(o+1.))+(-1.991468602)*dim
        ref[5,0,1]=(0.410656888307)*(1+2.*(dim-1.)/(o+1.))+(0.363101874486)*dim
        ref[5,1,0]=(-0.58962938168)*(1+2.*(dim-1.)/(o+1.))+(-1.85079462386)*dim
        ref[5,1,1]=(-1.65599484529)*(1+2.*(dim-1.)/(o+1.))+(-0.54713489959)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.606358724075)*x[0]**o+(-0.6326664687)*x[0]+(-0.521101256804)*x[1]**o+(0.702113497865)*x[1]
        arg[0,0,0,1]=(0.528315511137)*x[0]**o+(-0.116566989978)*x[0]+(-0.350847384045)*x[1]**o+(-0.581988878345)*x[1]
        arg[0,0,1,0]=(0.565076347032)*x[0]**o+(0.752532782601)*x[0]+(0.00706163321089)*x[1]**o+(-0.89265244135)*x[1]
        arg[0,0,1,1]=(-0.342136497951)*x[0]**o+(0.753153993866)*x[0]+(-0.199643944066)*x[1]**o+(0.995411752312)*x[1]
        arg[0,0,2,0]=(0.621283193015)*x[0]**o+(0.591186318333)*x[0]+(0.880169584588)*x[1]**o+(-0.11839698224)*x[1]
        arg[0,0,2,1]=(-0.831745199475)*x[0]**o+(0.293656166596)*x[0]+(0.0717745661823)*x[1]**o+(0.827680695965)*x[1]
        arg[0,1,0,0]=(0.686864676707)*x[0]**o+(-0.6552976723)*x[0]+(-0.65589888943)*x[1]**o+(-0.878054891784)*x[1]
        arg[0,1,0,1]=(-0.43359086203)*x[0]**o+(0.723819586663)*x[0]+(0.0741799035834)*x[1]**o+(0.13005305188)*x[1]
        arg[0,1,1,0]=(-0.0605346127445)*x[0]**o+(-0.717729061173)*x[0]+(0.825539660518)*x[1]**o+(0.873033938391)*x[1]
        arg[0,1,1,1]=(-0.286483607332)*x[0]**o+(-0.577255191498)*x[0]+(-0.328362280683)*x[1]**o+(0.499078653745)*x[1]
        arg[0,1,2,0]=(-0.334132932545)*x[0]**o+(0.462461110225)*x[0]+(-0.932311565873)*x[1]**o+(0.934683608666)*x[1]
        arg[0,1,2,1]=(-0.383203086272)*x[0]**o+(0.261035781624)*x[0]+(-0.203936125231)*x[1]**o+(0.657584310755)*x[1]
        arg[0,2,0,0]=(-0.637642146695)*x[0]**o+(0.80524493224)*x[0]+(-0.139137135459)*x[1]**o+(0.947434162736)*x[1]
        arg[0,2,0,1]=(-0.0738996477961)*x[0]**o+(0.0663360260869)*x[0]+(-0.599113718918)*x[1]**o+(-0.349514990363)*x[1]
        arg[0,2,1,0]=(-0.0539211250603)*x[0]**o+(0.262567947534)*x[0]+(0.115386681137)*x[1]**o+(-0.68179682822)*x[1]
        arg[0,2,1,1]=(-0.770114869627)*x[0]**o+(-0.388068404619)*x[0]+(-0.712914716647)*x[1]**o+(-0.810536809871)*x[1]
        arg[0,2,2,0]=(0.274425895455)*x[0]**o+(0.338467327995)*x[0]+(-0.78223636363)*x[1]**o+(0.406149090351)*x[1]
        arg[0,2,2,1]=(-0.748672361572)*x[0]**o+(0.508311939723)*x[0]+(0.875797107098)*x[1]**o+(0.0992795348726)*x[1]
        arg[0,3,0,0]=(-0.547350688209)*x[0]**o+(0.671683747412)*x[0]+(-0.0954212965969)*x[1]**o+(0.983487887304)*x[1]
        arg[0,3,0,1]=(0.626104338883)*x[0]**o+(-0.825311449834)*x[0]+(-0.422521614246)*x[1]**o+(0.304208397592)*x[1]
        arg[0,3,1,0]=(0.698777992012)*x[0]**o+(0.00291670095874)*x[0]+(0.327527809646)*x[1]**o+(0.283289333886)*x[1]
        arg[0,3,1,1]=(-0.0197648150597)*x[0]**o+(-0.902860814776)*x[0]+(0.663961228673)*x[1]**o+(0.675952386365)*x[1]
        arg[0,3,2,0]=(0.286747814345)*x[0]**o+(-0.0641946364041)*x[0]+(0.812019228115)*x[1]**o+(-0.225698609001)*x[1]
        arg[0,3,2,1]=(0.879776322226)*x[0]**o+(-0.251856143047)*x[0]+(-0.252403834757)*x[1]**o+(0.650790406323)*x[1]
        arg[0,4,0,0]=(-0.58782572388)*x[0]**o+(-0.144122228965)*x[0]+(0.521707348991)*x[1]**o+(-0.0460840614459)*x[1]
        arg[0,4,0,1]=(-0.777497837497)*x[0]**o+(-0.229011563089)*x[0]+(0.754657080443)*x[1]**o+(-0.690675521089)*x[1]
        arg[0,4,1,0]=(0.351528557097)*x[0]**o+(0.00192986530026)*x[0]+(-0.414849484269)*x[1]**o+(-0.250205659505)*x[1]
        arg[0,4,1,1]=(-0.814037796843)*x[0]**o+(0.00188512986627)*x[0]+(-0.810799710998)*x[1]**o+(-0.114406174711)*x[1]
        arg[0,4,2,0]=(-0.755511212878)*x[0]**o+(-0.502206951212)*x[0]+(0.555738333365)*x[1]**o+(-0.220359547926)*x[1]
        arg[0,4,2,1]=(0.247108894345)*x[0]**o+(0.479239235148)*x[0]+(0.663958827816)*x[1]**o+(0.220938477625)*x[1]
        arg[1,0,0,0]=(0.818774052595)*x[0]**o+(-0.651584728881)*x[0]+(-0.780674092348)*x[1]**o+(-0.397277502644)*x[1]
        arg[1,0,0,1]=(0.482101508249)*x[0]**o+(-0.523618202277)*x[0]+(0.806406836488)*x[1]**o+(-0.883738297664)*x[1]
        arg[1,0,1,0]=(0.272517667346)*x[0]**o+(-0.700461889638)*x[0]+(0.280736354342)*x[1]**o+(0.182983247089)*x[1]
        arg[1,0,1,1]=(0.566589111451)*x[0]**o+(-0.876744547538)*x[0]+(-0.934741703566)*x[1]**o+(-0.562770667492)*x[1]
        arg[1,0,2,0]=(-0.954128714623)*x[0]**o+(0.115699770127)*x[0]+(-0.580663262426)*x[1]**o+(-0.451619694161)*x[1]
        arg[1,0,2,1]=(0.268605150185)*x[0]**o+(-0.233415174179)*x[0]+(0.477143801066)*x[1]**o+(0.443696994244)*x[1]
        arg[1,1,0,0]=(-0.00240846686485)*x[0]**o+(0.776336851643)*x[0]+(0.488314964094)*x[1]**o+(-0.840196476788)*x[1]
        arg[1,1,0,1]=(-0.511732155638)*x[0]**o+(0.528782039485)*x[0]+(0.51274502262)*x[1]**o+(-0.700213610705)*x[1]
        arg[1,1,1,0]=(0.300359000499)*x[0]**o+(0.405101876401)*x[0]+(-0.47022583299)*x[1]**o+(-0.977634916267)*x[1]
        arg[1,1,1,1]=(-0.494482318948)*x[0]**o+(0.254112728167)*x[0]+(-0.541672595343)*x[1]**o+(-0.270970064522)*x[1]
        arg[1,1,2,0]=(-0.586624294233)*x[0]**o+(0.629601920904)*x[0]+(-0.525092252936)*x[1]**o+(-0.44611635978)*x[1]
        arg[1,1,2,1]=(0.716927085139)*x[0]**o+(0.700193925267)*x[0]+(0.0624921446647)*x[1]**o+(0.109485129877)*x[1]
        arg[1,2,0,0]=(0.325729307898)*x[0]**o+(0.773432634097)*x[0]+(-0.675612597356)*x[1]**o+(-0.0554913733692)*x[1]
        arg[1,2,0,1]=(-0.381606382262)*x[0]**o+(-0.689040382854)*x[0]+(-0.580191973398)*x[1]**o+(0.951689147387)*x[1]
        arg[1,2,1,0]=(-0.527642403744)*x[0]**o+(-0.936915203927)*x[0]+(0.631733997289)*x[1]**o+(-0.715975680015)*x[1]
        arg[1,2,1,1]=(0.0886898373323)*x[0]**o+(-0.0943320873915)*x[0]+(0.470668750443)*x[1]**o+(0.773832960093)*x[1]
        arg[1,2,2,0]=(-0.727720290197)*x[0]**o+(0.422119803173)*x[0]+(0.705882951779)*x[1]**o+(-0.350801865244)*x[1]
        arg[1,2,2,1]=(0.879335475407)*x[0]**o+(-0.0307168471577)*x[0]+(0.458578151544)*x[1]**o+(0.788796576339)*x[1]
        arg[1,3,0,0]=(0.182143508146)*x[0]**o+(-0.680787176288)*x[0]+(0.626099489145)*x[1]**o+(-0.771540378408)*x[1]
        arg[1,3,0,1]=(-0.578341681282)*x[0]**o+(0.856407179963)*x[0]+(0.525319425356)*x[1]**o+(0.345651851179)*x[1]
        arg[1,3,1,0]=(-0.316652838058)*x[0]**o+(0.652409898929)*x[0]+(-0.82123442746)*x[1]**o+(-0.136868558081)*x[1]
        arg[1,3,1,1]=(-0.755530700082)*x[0]**o+(0.354429731374)*x[0]+(-0.14330437792)*x[1]**o+(-0.221454495186)*x[1]
        arg[1,3,2,0]=(-0.520261483769)*x[0]**o+(0.0873707179205)*x[0]+(-0.882995356682)*x[1]**o+(-0.773921687804)*x[1]
        arg[1,3,2,1]=(0.404278952718)*x[0]**o+(0.651166017764)*x[0]+(-0.496695253747)*x[1]**o+(0.625671754182)*x[1]
        arg[1,4,0,0]=(-0.730831539043)*x[0]**o+(0.219467989041)*x[0]+(-0.853784008431)*x[1]**o+(0.274799624897)*x[1]
        arg[1,4,0,1]=(0.893313801855)*x[0]**o+(-0.539687651773)*x[0]+(-0.411582029239)*x[1]**o+(-0.293521483805)*x[1]
        arg[1,4,1,0]=(0.484878049594)*x[0]**o+(0.218930293134)*x[0]+(-0.138362733141)*x[1]**o+(-0.663241354177)*x[1]
        arg[1,4,1,1]=(-0.608410207857)*x[0]**o+(-0.0921006917331)*x[0]+(-0.885705130565)*x[1]**o+(0.103724764404)*x[1]
        arg[1,4,2,0]=(0.150946260009)*x[0]**o+(-0.601834829203)*x[0]+(0.0141091402169)*x[1]**o+(-0.98316187826)*x[1]
        arg[1,4,2,1]=(0.98764918249)*x[0]**o+(-0.841155365687)*x[0]+(0.185525084432)*x[1]**o+(-0.613421841538)*x[1]
        arg[2,0,0,0]=(0.485337970037)*x[0]**o+(-0.898602108837)*x[0]+(0.847812629133)*x[1]**o+(-0.23522051021)*x[1]
        arg[2,0,0,1]=(0.835641029366)*x[0]**o+(-0.0524661277338)*x[0]+(0.41697147238)*x[1]**o+(-0.622813671618)*x[1]
        arg[2,0,1,0]=(0.505181088966)*x[0]**o+(-0.870141755006)*x[0]+(0.451139050304)*x[1]**o+(0.957790881903)*x[1]
        arg[2,0,1,1]=(0.822955028051)*x[0]**o+(-0.103028201548)*x[0]+(-0.604614127349)*x[1]**o+(-0.655334670762)*x[1]
        arg[2,0,2,0]=(-0.166647802545)*x[0]**o+(-0.466885186725)*x[0]+(0.0782333081578)*x[1]**o+(0.814414769115)*x[1]
        arg[2,0,2,1]=(-0.980310280968)*x[0]**o+(-0.884306329761)*x[0]+(-0.895550319566)*x[1]**o+(0.160657671233)*x[1]
        arg[2,1,0,0]=(-0.846405435483)*x[0]**o+(-0.625465093616)*x[0]+(-0.0789995975268)*x[1]**o+(-0.440180761086)*x[1]
        arg[2,1,0,1]=(-0.949855662419)*x[0]**o+(-0.013707742372)*x[0]+(0.0397729185783)*x[1]**o+(-0.836054083952)*x[1]
        arg[2,1,1,0]=(0.427246928002)*x[0]**o+(0.835369889909)*x[0]+(0.925300799558)*x[1]**o+(-0.328266770631)*x[1]
        arg[2,1,1,1]=(-0.36143365614)*x[0]**o+(-0.744107540819)*x[0]+(-0.358554131639)*x[1]**o+(-0.562674543946)*x[1]
        arg[2,1,2,0]=(0.60580594275)*x[0]**o+(0.644796993923)*x[0]+(-0.682407021927)*x[1]**o+(0.938016908061)*x[1]
        arg[2,1,2,1]=(-0.934393274317)*x[0]**o+(-0.718420272664)*x[0]+(0.0357039032065)*x[1]**o+(0.446387943344)*x[1]
        arg[2,2,0,0]=(-0.329105411737)*x[0]**o+(0.588935500072)*x[0]+(-0.833362483432)*x[1]**o+(0.629996961074)*x[1]
        arg[2,2,0,1]=(0.723977530226)*x[0]**o+(0.0698430385616)*x[0]+(-0.503761809156)*x[1]**o+(0.907472279839)*x[1]
        arg[2,2,1,0]=(-0.767545510793)*x[0]**o+(0.615094783682)*x[0]+(0.894594933703)*x[1]**o+(0.922357219049)*x[1]
        arg[2,2,1,1]=(-0.155739055576)*x[0]**o+(0.646142299486)*x[0]+(0.865932681534)*x[1]**o+(-0.126887380871)*x[1]
        arg[2,2,2,0]=(-0.629534320112)*x[0]**o+(-0.0904366301836)*x[0]+(0.193019806731)*x[1]**o+(-0.866080160466)*x[1]
        arg[2,2,2,1]=(-0.714527761053)*x[0]**o+(-0.813712879405)*x[0]+(-0.291734238438)*x[1]**o+(0.232675536415)*x[1]
        arg[2,3,0,0]=(-0.447474292879)*x[0]**o+(-0.639046887788)*x[0]+(0.159616683647)*x[1]**o+(0.564127141623)*x[1]
        arg[2,3,0,1]=(-0.608953936619)*x[0]**o+(0.273179484103)*x[0]+(0.42693156927)*x[1]**o+(-0.983331075647)*x[1]
        arg[2,3,1,0]=(0.855859483301)*x[0]**o+(-0.432495503925)*x[0]+(-0.67928203769)*x[1]**o+(-0.856541006702)*x[1]
        arg[2,3,1,1]=(-0.581185594204)*x[0]**o+(0.947720323693)*x[0]+(-0.520212852628)*x[1]**o+(-0.37563221594)*x[1]
        arg[2,3,2,0]=(0.351358492535)*x[0]**o+(0.367210525846)*x[0]+(0.97870179383)*x[1]**o+(-0.199559052968)*x[1]
        arg[2,3,2,1]=(0.797294518541)*x[0]**o+(0.546397721552)*x[0]+(0.374121851646)*x[1]**o+(0.704436689798)*x[1]
        arg[2,4,0,0]=(-0.756047775149)*x[0]**o+(-0.705403684825)*x[0]+(-0.662421213376)*x[1]**o+(0.390069314234)*x[1]
        arg[2,4,0,1]=(-0.658842889869)*x[0]**o+(0.766273769822)*x[0]+(-0.779201775962)*x[1]**o+(-0.968670270742)*x[1]
        arg[2,4,1,0]=(-0.336220564628)*x[0]**o+(-0.134745604319)*x[0]+(0.391939888428)*x[1]**o+(-0.113113999792)*x[1]
        arg[2,4,1,1]=(-0.8345363851)*x[0]**o+(0.339223929857)*x[0]+(0.628403353717)*x[1]**o+(0.799364167182)*x[1]
        arg[2,4,2,0]=(-0.898050446701)*x[0]**o+(-0.666373288962)*x[0]+(-0.620778021659)*x[1]**o+(-0.990359850049)*x[1]
        arg[2,4,2,1]=(-0.933931027675)*x[0]**o+(-0.119077281033)*x[0]+(0.557313533556)*x[1]**o+(0.160336158304)*x[1]
        arg[3,0,0,0]=(0.972462636314)*x[0]**o+(0.607913740671)*x[0]+(0.490168539257)*x[1]**o+(-0.12515921391)*x[1]
        arg[3,0,0,1]=(0.856370886333)*x[0]**o+(-0.143225009869)*x[0]+(-0.330381845603)*x[1]**o+(-0.271958059018)*x[1]
        arg[3,0,1,0]=(0.86742129112)*x[0]**o+(-0.0259756341823)*x[0]+(-0.0137047849942)*x[1]**o+(-0.0165444161044)*x[1]
        arg[3,0,1,1]=(-0.445996036941)*x[0]**o+(0.247936136111)*x[0]+(-0.381484740124)*x[1]**o+(0.836678417006)*x[1]
        arg[3,0,2,0]=(0.161387456771)*x[0]**o+(0.687533753311)*x[0]+(-0.463845417603)*x[1]**o+(0.665023823118)*x[1]
        arg[3,0,2,1]=(0.243410612742)*x[0]**o+(-0.52431675734)*x[0]+(0.75630192488)*x[1]**o+(-0.703601557317)*x[1]
        arg[3,1,0,0]=(0.793530689056)*x[0]**o+(0.814806726176)*x[0]+(0.582051068232)*x[1]**o+(0.692432923027)*x[1]
        arg[3,1,0,1]=(-0.283506909465)*x[0]**o+(0.21438128538)*x[0]+(-0.327476122861)*x[1]**o+(0.990178870269)*x[1]
        arg[3,1,1,0]=(-0.117386491827)*x[0]**o+(0.747771272481)*x[0]+(-0.439767754004)*x[1]**o+(-0.905218338639)*x[1]
        arg[3,1,1,1]=(-0.544805647706)*x[0]**o+(-0.0484033735345)*x[0]+(0.861443439156)*x[1]**o+(-0.992086899129)*x[1]
        arg[3,1,2,0]=(0.596208599112)*x[0]**o+(-0.276136075021)*x[0]+(0.079604120535)*x[1]**o+(0.124232758054)*x[1]
        arg[3,1,2,1]=(0.809847641841)*x[0]**o+(0.956041370087)*x[0]+(0.65224420189)*x[1]**o+(0.565327743771)*x[1]
        arg[3,2,0,0]=(-0.264599479107)*x[0]**o+(-0.276726556903)*x[0]+(0.577911819561)*x[1]**o+(-0.656157381756)*x[1]
        arg[3,2,0,1]=(-0.0636193103458)*x[0]**o+(0.941482878452)*x[0]+(0.877581030494)*x[1]**o+(-0.254911385455)*x[1]
        arg[3,2,1,0]=(-0.28083673633)*x[0]**o+(0.849095032935)*x[0]+(0.6825284068)*x[1]**o+(0.445605171749)*x[1]
        arg[3,2,1,1]=(-0.729331034619)*x[0]**o+(0.036400895394)*x[0]+(0.365733359214)*x[1]**o+(0.0305324156609)*x[1]
        arg[3,2,2,0]=(-0.70932226428)*x[0]**o+(0.310455571562)*x[0]+(-0.101305643894)*x[1]**o+(0.346873071627)*x[1]
        arg[3,2,2,1]=(0.712053329488)*x[0]**o+(-0.845042122758)*x[0]+(0.909517101069)*x[1]**o+(-0.163109058477)*x[1]
        arg[3,3,0,0]=(0.451744730688)*x[0]**o+(0.308756227526)*x[0]+(0.13963307493)*x[1]**o+(-0.945789640023)*x[1]
        arg[3,3,0,1]=(-0.852609785834)*x[0]**o+(-0.821513824231)*x[0]+(-0.544690113726)*x[1]**o+(0.335547843106)*x[1]
        arg[3,3,1,0]=(0.337600834432)*x[0]**o+(-0.170181468776)*x[0]+(-0.302415145845)*x[1]**o+(0.985425450212)*x[1]
        arg[3,3,1,1]=(-0.638242591048)*x[0]**o+(-0.0221442303187)*x[0]+(0.421311946694)*x[1]**o+(-0.43495081316)*x[1]
        arg[3,3,2,0]=(-0.734168622382)*x[0]**o+(-0.889300572069)*x[0]+(0.808161580285)*x[1]**o+(-0.526348332439)*x[1]
        arg[3,3,2,1]=(0.998918324168)*x[0]**o+(-0.953729666152)*x[0]+(0.665101882861)*x[1]**o+(-0.581118145725)*x[1]
        arg[3,4,0,0]=(-0.251226888098)*x[0]**o+(-0.711615999514)*x[0]+(0.291022154479)*x[1]**o+(0.0772744442604)*x[1]
        arg[3,4,0,1]=(0.162379537091)*x[0]**o+(0.363750962136)*x[0]+(0.635943839083)*x[1]**o+(0.513856953389)*x[1]
        arg[3,4,1,0]=(-0.99052354199)*x[0]**o+(0.289202668813)*x[0]+(-0.759078503012)*x[1]**o+(-0.588801682903)*x[1]
        arg[3,4,1,1]=(-0.942278699306)*x[0]**o+(0.476003512559)*x[0]+(0.873920293292)*x[1]**o+(0.867413991424)*x[1]
        arg[3,4,2,0]=(-0.602386702281)*x[0]**o+(0.694432164577)*x[0]+(0.367008950869)*x[1]**o+(-0.803331549449)*x[1]
        arg[3,4,2,1]=(-0.00204465287104)*x[0]**o+(-0.697572466685)*x[0]+(-0.84755978799)*x[1]**o+(-0.758946055726)*x[1]
        ref[0,0,0,0]=(0.0852574672711)*(1+2.*(dim-1.)/(o+1.))+(0.069447029165)*dim
        ref[0,0,0,1]=(0.177468127092)*(1+2.*(dim-1.)/(o+1.))+(-0.698555868323)*dim
        ref[0,0,1,0]=(0.572137980243)*(1+2.*(dim-1.)/(o+1.))+(-0.140119658749)*dim
        ref[0,0,1,1]=(-0.541780442017)*(1+2.*(dim-1.)/(o+1.))+(1.74856574618)*dim
        ref[0,0,2,0]=(1.5014527776)*(1+2.*(dim-1.)/(o+1.))+(0.472789336092)*dim
        ref[0,0,2,1]=(-0.759970633293)*(1+2.*(dim-1.)/(o+1.))+(1.12133686256)*dim
        ref[0,1,0,0]=(0.030965787278)*(1+2.*(dim-1.)/(o+1.))+(-1.53335256408)*dim
        ref[0,1,0,1]=(-0.359410958446)*(1+2.*(dim-1.)/(o+1.))+(0.853872638544)*dim
        ref[0,1,1,0]=(0.765005047774)*(1+2.*(dim-1.)/(o+1.))+(0.155304877218)*dim
        ref[0,1,1,1]=(-0.614845888015)*(1+2.*(dim-1.)/(o+1.))+(-0.0781765377528)*dim
        ref[0,1,2,0]=(-1.26644449842)*(1+2.*(dim-1.)/(o+1.))+(1.39714471889)*dim
        ref[0,1,2,1]=(-0.587139211502)*(1+2.*(dim-1.)/(o+1.))+(0.918620092379)*dim
        ref[0,2,0,0]=(-0.776779282155)*(1+2.*(dim-1.)/(o+1.))+(1.75267909498)*dim
        ref[0,2,0,1]=(-0.673013366714)*(1+2.*(dim-1.)/(o+1.))+(-0.283178964276)*dim
        ref[0,2,1,0]=(0.0614655560763)*(1+2.*(dim-1.)/(o+1.))+(-0.419228880686)*dim
        ref[0,2,1,1]=(-1.48302958627)*(1+2.*(dim-1.)/(o+1.))+(-1.19860521449)*dim
        ref[0,2,2,0]=(-0.507810468175)*(1+2.*(dim-1.)/(o+1.))+(0.744616418346)*dim
        ref[0,2,2,1]=(0.127124745526)*(1+2.*(dim-1.)/(o+1.))+(0.607591474596)*dim
        ref[0,3,0,0]=(-0.642771984806)*(1+2.*(dim-1.)/(o+1.))+(1.65517163472)*dim
        ref[0,3,0,1]=(0.203582724637)*(1+2.*(dim-1.)/(o+1.))+(-0.521103052242)*dim
        ref[0,3,1,0]=(1.02630580166)*(1+2.*(dim-1.)/(o+1.))+(0.286206034845)*dim
        ref[0,3,1,1]=(0.644196413613)*(1+2.*(dim-1.)/(o+1.))+(-0.226908428411)*dim
        ref[0,3,2,0]=(1.09876704246)*(1+2.*(dim-1.)/(o+1.))+(-0.289893245405)*dim
        ref[0,3,2,1]=(0.627372487469)*(1+2.*(dim-1.)/(o+1.))+(0.398934263276)*dim
        ref[0,4,0,0]=(-0.0661183748894)*(1+2.*(dim-1.)/(o+1.))+(-0.190206290411)*dim
        ref[0,4,0,1]=(-0.022840757054)*(1+2.*(dim-1.)/(o+1.))+(-0.919687084178)*dim
        ref[0,4,1,0]=(-0.063320927172)*(1+2.*(dim-1.)/(o+1.))+(-0.248275794205)*dim
        ref[0,4,1,1]=(-1.62483750784)*(1+2.*(dim-1.)/(o+1.))+(-0.112521044845)*dim
        ref[0,4,2,0]=(-0.199772879513)*(1+2.*(dim-1.)/(o+1.))+(-0.722566499138)*dim
        ref[0,4,2,1]=(0.911067722161)*(1+2.*(dim-1.)/(o+1.))+(0.700177712774)*dim
        ref[1,0,0,0]=(0.0380999602468)*(1+2.*(dim-1.)/(o+1.))+(-1.04886223152)*dim
        ref[1,0,0,1]=(1.28850834474)*(1+2.*(dim-1.)/(o+1.))+(-1.40735649994)*dim
        ref[1,0,1,0]=(0.553254021688)*(1+2.*(dim-1.)/(o+1.))+(-0.517478642549)*dim
        ref[1,0,1,1]=(-0.368152592115)*(1+2.*(dim-1.)/(o+1.))+(-1.43951521503)*dim
        ref[1,0,2,0]=(-1.53479197705)*(1+2.*(dim-1.)/(o+1.))+(-0.335919924034)*dim
        ref[1,0,2,1]=(0.745748951251)*(1+2.*(dim-1.)/(o+1.))+(0.210281820065)*dim
        ref[1,1,0,0]=(0.485906497229)*(1+2.*(dim-1.)/(o+1.))+(-0.0638596251453)*dim
        ref[1,1,0,1]=(0.00101286698233)*(1+2.*(dim-1.)/(o+1.))+(-0.17143157122)*dim
        ref[1,1,1,0]=(-0.169866832491)*(1+2.*(dim-1.)/(o+1.))+(-0.572533039866)*dim
        ref[1,1,1,1]=(-1.03615491429)*(1+2.*(dim-1.)/(o+1.))+(-0.016857336355)*dim
        ref[1,1,2,0]=(-1.11171654717)*(1+2.*(dim-1.)/(o+1.))+(0.183485561125)*dim
        ref[1,1,2,1]=(0.779419229804)*(1+2.*(dim-1.)/(o+1.))+(0.809679055143)*dim
        ref[1,2,0,0]=(-0.349883289458)*(1+2.*(dim-1.)/(o+1.))+(0.717941260728)*dim
        ref[1,2,0,1]=(-0.96179835566)*(1+2.*(dim-1.)/(o+1.))+(0.262648764533)*dim
        ref[1,2,1,0]=(0.104091593546)*(1+2.*(dim-1.)/(o+1.))+(-1.65289088394)*dim
        ref[1,2,1,1]=(0.559358587775)*(1+2.*(dim-1.)/(o+1.))+(0.679500872701)*dim
        ref[1,2,2,0]=(-0.0218373384179)*(1+2.*(dim-1.)/(o+1.))+(0.0713179379289)*dim
        ref[1,2,2,1]=(1.33791362695)*(1+2.*(dim-1.)/(o+1.))+(0.758079729181)*dim
        ref[1,3,0,0]=(0.808242997292)*(1+2.*(dim-1.)/(o+1.))+(-1.4523275547)*dim
        ref[1,3,0,1]=(-0.0530222559262)*(1+2.*(dim-1.)/(o+1.))+(1.20205903114)*dim
        ref[1,3,1,0]=(-1.13788726552)*(1+2.*(dim-1.)/(o+1.))+(0.515541340848)*dim
        ref[1,3,1,1]=(-0.898835078002)*(1+2.*(dim-1.)/(o+1.))+(0.132975236188)*dim
        ref[1,3,2,0]=(-1.40325684045)*(1+2.*(dim-1.)/(o+1.))+(-0.686550969884)*dim
        ref[1,3,2,1]=(-0.0924163010291)*(1+2.*(dim-1.)/(o+1.))+(1.27683777195)*dim
        ref[1,4,0,0]=(-1.58461554747)*(1+2.*(dim-1.)/(o+1.))+(0.494267613937)*dim
        ref[1,4,0,1]=(0.481731772616)*(1+2.*(dim-1.)/(o+1.))+(-0.833209135578)*dim
        ref[1,4,1,0]=(0.346515316453)*(1+2.*(dim-1.)/(o+1.))+(-0.444311061043)*dim
        ref[1,4,1,1]=(-1.49411533842)*(1+2.*(dim-1.)/(o+1.))+(0.0116240726713)*dim
        ref[1,4,2,0]=(0.165055400226)*(1+2.*(dim-1.)/(o+1.))+(-1.58499670746)*dim
        ref[1,4,2,1]=(1.17317426692)*(1+2.*(dim-1.)/(o+1.))+(-1.45457720722)*dim
        ref[2,0,0,0]=(1.33315059917)*(1+2.*(dim-1.)/(o+1.))+(-1.13382261905)*dim
        ref[2,0,0,1]=(1.25261250175)*(1+2.*(dim-1.)/(o+1.))+(-0.675279799352)*dim
        ref[2,0,1,0]=(0.95632013927)*(1+2.*(dim-1.)/(o+1.))+(0.0876491268976)*dim
        ref[2,0,1,1]=(0.218340900702)*(1+2.*(dim-1.)/(o+1.))+(-0.75836287231)*dim
        ref[2,0,2,0]=(-0.0884144943875)*(1+2.*(dim-1.)/(o+1.))+(0.34752958239)*dim
        ref[2,0,2,1]=(-1.87586060053)*(1+2.*(dim-1.)/(o+1.))+(-0.723648658528)*dim
        ref[2,1,0,0]=(-0.92540503301)*(1+2.*(dim-1.)/(o+1.))+(-1.0656458547)*dim
        ref[2,1,0,1]=(-0.910082743841)*(1+2.*(dim-1.)/(o+1.))+(-0.849761826324)*dim
        ref[2,1,1,0]=(1.35254772756)*(1+2.*(dim-1.)/(o+1.))+(0.507103119278)*dim
        ref[2,1,1,1]=(-0.719987787779)*(1+2.*(dim-1.)/(o+1.))+(-1.30678208477)*dim
        ref[2,1,2,0]=(-0.0766010791773)*(1+2.*(dim-1.)/(o+1.))+(1.58281390198)*dim
        ref[2,1,2,1]=(-0.89868937111)*(1+2.*(dim-1.)/(o+1.))+(-0.27203232932)*dim
        ref[2,2,0,0]=(-1.16246789517)*(1+2.*(dim-1.)/(o+1.))+(1.21893246115)*dim
        ref[2,2,0,1]=(0.22021572107)*(1+2.*(dim-1.)/(o+1.))+(0.977315318401)*dim
        ref[2,2,1,0]=(0.12704942291)*(1+2.*(dim-1.)/(o+1.))+(1.53745200273)*dim
        ref[2,2,1,1]=(0.710193625958)*(1+2.*(dim-1.)/(o+1.))+(0.519254918615)*dim
        ref[2,2,2,0]=(-0.43651451338)*(1+2.*(dim-1.)/(o+1.))+(-0.956516790649)*dim
        ref[2,2,2,1]=(-1.00626199949)*(1+2.*(dim-1.)/(o+1.))+(-0.581037342991)*dim
        ref[2,3,0,0]=(-0.287857609232)*(1+2.*(dim-1.)/(o+1.))+(-0.0749197461649)*dim
        ref[2,3,0,1]=(-0.182022367349)*(1+2.*(dim-1.)/(o+1.))+(-0.710151591543)*dim
        ref[2,3,1,0]=(0.176577445611)*(1+2.*(dim-1.)/(o+1.))+(-1.28903651063)*dim
        ref[2,3,1,1]=(-1.10139844683)*(1+2.*(dim-1.)/(o+1.))+(0.572088107754)*dim
        ref[2,3,2,0]=(1.33006028637)*(1+2.*(dim-1.)/(o+1.))+(0.167651472878)*dim
        ref[2,3,2,1]=(1.17141637019)*(1+2.*(dim-1.)/(o+1.))+(1.25083441135)*dim
        ref[2,4,0,0]=(-1.41846898852)*(1+2.*(dim-1.)/(o+1.))+(-0.315334370591)*dim
        ref[2,4,0,1]=(-1.43804466583)*(1+2.*(dim-1.)/(o+1.))+(-0.20239650092)*dim
        ref[2,4,1,0]=(0.0557193237997)*(1+2.*(dim-1.)/(o+1.))+(-0.247859604111)*dim
        ref[2,4,1,1]=(-0.206133031383)*(1+2.*(dim-1.)/(o+1.))+(1.13858809704)*dim
        ref[2,4,2,0]=(-1.51882846836)*(1+2.*(dim-1.)/(o+1.))+(-1.65673313901)*dim
        ref[2,4,2,1]=(-0.376617494119)*(1+2.*(dim-1.)/(o+1.))+(0.041258877271)*dim
        ref[3,0,0,0]=(1.46263117557)*(1+2.*(dim-1.)/(o+1.))+(0.482754526761)*dim
        ref[3,0,0,1]=(0.52598904073)*(1+2.*(dim-1.)/(o+1.))+(-0.415183068887)*dim
        ref[3,0,1,0]=(0.853716506126)*(1+2.*(dim-1.)/(o+1.))+(-0.0425200502867)*dim
        ref[3,0,1,1]=(-0.827480777065)*(1+2.*(dim-1.)/(o+1.))+(1.08461455312)*dim
        ref[3,0,2,0]=(-0.302457960832)*(1+2.*(dim-1.)/(o+1.))+(1.35255757643)*dim
        ref[3,0,2,1]=(0.999712537623)*(1+2.*(dim-1.)/(o+1.))+(-1.22791831466)*dim
        ref[3,1,0,0]=(1.37558175729)*(1+2.*(dim-1.)/(o+1.))+(1.5072396492)*dim
        ref[3,1,0,1]=(-0.610983032326)*(1+2.*(dim-1.)/(o+1.))+(1.20456015565)*dim
        ref[3,1,1,0]=(-0.557154245831)*(1+2.*(dim-1.)/(o+1.))+(-0.157447066158)*dim
        ref[3,1,1,1]=(0.316637791451)*(1+2.*(dim-1.)/(o+1.))+(-1.04049027266)*dim
        ref[3,1,2,0]=(0.675812719647)*(1+2.*(dim-1.)/(o+1.))+(-0.151903316967)*dim
        ref[3,1,2,1]=(1.46209184373)*(1+2.*(dim-1.)/(o+1.))+(1.52136911386)*dim
        ref[3,2,0,0]=(0.313312340454)*(1+2.*(dim-1.)/(o+1.))+(-0.93288393866)*dim
        ref[3,2,0,1]=(0.813961720148)*(1+2.*(dim-1.)/(o+1.))+(0.686571492997)*dim
        ref[3,2,1,0]=(0.40169167047)*(1+2.*(dim-1.)/(o+1.))+(1.29470020468)*dim
        ref[3,2,1,1]=(-0.363597675405)*(1+2.*(dim-1.)/(o+1.))+(0.066933311055)*dim
        ref[3,2,2,0]=(-0.810627908174)*(1+2.*(dim-1.)/(o+1.))+(0.657328643189)*dim
        ref[3,2,2,1]=(1.62157043056)*(1+2.*(dim-1.)/(o+1.))+(-1.00815118124)*dim
        ref[3,3,0,0]=(0.591377805617)*(1+2.*(dim-1.)/(o+1.))+(-0.637033412497)*dim
        ref[3,3,0,1]=(-1.39729989956)*(1+2.*(dim-1.)/(o+1.))+(-0.485965981125)*dim
        ref[3,3,1,0]=(0.0351856885862)*(1+2.*(dim-1.)/(o+1.))+(0.815243981437)*dim
        ref[3,3,1,1]=(-0.216930644354)*(1+2.*(dim-1.)/(o+1.))+(-0.457095043478)*dim
        ref[3,3,2,0]=(0.0739929579029)*(1+2.*(dim-1.)/(o+1.))+(-1.41564890451)*dim
        ref[3,3,2,1]=(1.66402020703)*(1+2.*(dim-1.)/(o+1.))+(-1.53484781188)*dim
        ref[3,4,0,0]=(0.0397952663815)*(1+2.*(dim-1.)/(o+1.))+(-0.634341555254)*dim
        ref[3,4,0,1]=(0.798323376174)*(1+2.*(dim-1.)/(o+1.))+(0.877607915525)*dim
        ref[3,4,1,0]=(-1.749602045)*(1+2.*(dim-1.)/(o+1.))+(-0.29959901409)*dim
        ref[3,4,1,1]=(-0.0683584060144)*(1+2.*(dim-1.)/(o+1.))+(1.34341750398)*dim
        ref[3,4,2,0]=(-0.235377751413)*(1+2.*(dim-1.)/(o+1.))+(-0.108899384872)*dim
        ref[3,4,2,1]=(-0.849604440861)*(1+2.*(dim-1.)/(o+1.))+(-1.45651852241)*dim
      else:
        arg[0,0,0,0]=(-0.229219068739)*x[0]**o+(0.293241016924)*x[0]+(-0.2730739542)*x[1]**o+(0.980438707356)*x[1]+(-0.712463216466)*x[2]**o+(0.201036317883)*x[2]
        arg[0,0,0,1]=(0.626271947422)*x[0]**o+(0.379293769575)*x[0]+(-0.476378592342)*x[1]**o+(-0.0089906260934)*x[1]+(-0.433154385811)*x[2]**o+(0.126766298748)*x[2]
        arg[0,0,1,0]=(-0.517950283035)*x[0]**o+(-0.219780986066)*x[0]+(0.345048580545)*x[1]**o+(0.116833851128)*x[1]+(-0.790641523314)*x[2]**o+(0.40506306417)*x[2]
        arg[0,0,1,1]=(0.910924420538)*x[0]**o+(0.281102231596)*x[0]+(0.115454208619)*x[1]**o+(0.164254056171)*x[1]+(0.44960060103)*x[2]**o+(-0.14238069165)*x[2]
        arg[0,0,2,0]=(0.86195808217)*x[0]**o+(0.775425956811)*x[0]+(0.214559021523)*x[1]**o+(0.400663063038)*x[1]+(0.995663608954)*x[2]**o+(0.228658543268)*x[2]
        arg[0,0,2,1]=(0.9025917492)*x[0]**o+(0.643205662573)*x[0]+(-0.304382459664)*x[1]**o+(-0.653256441532)*x[1]+(0.591142459807)*x[2]**o+(-0.888899030043)*x[2]
        arg[0,1,0,0]=(-0.874720784838)*x[0]**o+(-0.961456282274)*x[0]+(-0.954972993457)*x[1]**o+(0.00963939708746)*x[1]+(-0.884848381277)*x[2]**o+(-0.859924591622)*x[2]
        arg[0,1,0,1]=(0.133592941787)*x[0]**o+(0.12505415793)*x[0]+(-0.872197297559)*x[1]**o+(-0.666936976734)*x[1]+(-0.540166108036)*x[2]**o+(0.739145002101)*x[2]
        arg[0,1,1,0]=(0.543825266564)*x[0]**o+(0.167854064952)*x[0]+(-0.876426395953)*x[1]**o+(0.317538118917)*x[1]+(0.0454531201303)*x[2]**o+(0.822731343286)*x[2]
        arg[0,1,1,1]=(0.294557050513)*x[0]**o+(0.811698886899)*x[0]+(0.906213580147)*x[1]**o+(0.47895749797)*x[1]+(-0.937617554883)*x[2]**o+(0.171907458148)*x[2]
        arg[0,1,2,0]=(-0.843310949654)*x[0]**o+(0.766452648675)*x[0]+(-0.801151193917)*x[1]**o+(-0.373068887299)*x[1]+(-0.908043738284)*x[2]**o+(0.742278324563)*x[2]
        arg[0,1,2,1]=(-0.83259110525)*x[0]**o+(-0.303796015231)*x[0]+(0.153727560059)*x[1]**o+(-0.848393854019)*x[1]+(0.887795931916)*x[2]**o+(0.713486413074)*x[2]
        arg[0,2,0,0]=(-0.928603918182)*x[0]**o+(0.631213545526)*x[0]+(0.0149216054247)*x[1]**o+(0.136454313226)*x[1]+(-0.182080277504)*x[2]**o+(-0.898437757742)*x[2]
        arg[0,2,0,1]=(0.553844129645)*x[0]**o+(0.91439079798)*x[0]+(0.260424446678)*x[1]**o+(0.786629777779)*x[1]+(-0.412353876109)*x[2]**o+(0.496953000487)*x[2]
        arg[0,2,1,0]=(-0.48365422571)*x[0]**o+(-0.772252454664)*x[0]+(0.692592845143)*x[1]**o+(0.861992232175)*x[1]+(0.65169271692)*x[2]**o+(0.889175716524)*x[2]
        arg[0,2,1,1]=(-0.761086015475)*x[0]**o+(-0.616528617148)*x[0]+(-0.949399556926)*x[1]**o+(-0.783355074724)*x[1]+(-0.260470660608)*x[2]**o+(-0.854537542803)*x[2]
        arg[0,2,2,0]=(-0.812649439778)*x[0]**o+(-0.844862957855)*x[0]+(0.719928253134)*x[1]**o+(-0.892803903061)*x[1]+(-0.434753054503)*x[2]**o+(0.814188548451)*x[2]
        arg[0,2,2,1]=(-0.831106881173)*x[0]**o+(0.893305297647)*x[0]+(-0.351357795644)*x[1]**o+(0.0318574503478)*x[1]+(0.721896657873)*x[2]**o+(-0.75215838035)*x[2]
        arg[0,3,0,0]=(-0.868084492214)*x[0]**o+(-0.108398365838)*x[0]+(0.994384983885)*x[1]**o+(0.199445529335)*x[1]+(-0.570120483013)*x[2]**o+(0.992752456282)*x[2]
        arg[0,3,0,1]=(-0.304553428923)*x[0]**o+(-0.589461304318)*x[0]+(-0.570142565817)*x[1]**o+(-0.858530013475)*x[1]+(0.0690574539796)*x[2]**o+(-0.0321587525658)*x[2]
        arg[0,3,1,0]=(-0.729413026954)*x[0]**o+(-0.728248951404)*x[0]+(-0.568598599095)*x[1]**o+(0.0823012989572)*x[1]+(0.248614424573)*x[2]**o+(-0.116584878805)*x[2]
        arg[0,3,1,1]=(0.586846538084)*x[0]**o+(-0.406933529762)*x[0]+(0.839125682244)*x[1]**o+(0.237338453386)*x[1]+(-0.953101889056)*x[2]**o+(0.164147861616)*x[2]
        arg[0,3,2,0]=(-0.603895026116)*x[0]**o+(-0.401045478885)*x[0]+(-0.524739550832)*x[1]**o+(-0.077907095917)*x[1]+(0.535404409567)*x[2]**o+(0.745721528391)*x[2]
        arg[0,3,2,1]=(-0.189151084147)*x[0]**o+(-0.865927541126)*x[0]+(0.55265091657)*x[1]**o+(-0.592847311104)*x[1]+(-0.467509383222)*x[2]**o+(-0.779649276678)*x[2]
        arg[0,4,0,0]=(-0.564283606569)*x[0]**o+(-0.588029571592)*x[0]+(0.544948665159)*x[1]**o+(0.140885576219)*x[1]+(-0.333531842397)*x[2]**o+(-0.579095398292)*x[2]
        arg[0,4,0,1]=(-0.189127238455)*x[0]**o+(-0.693907127162)*x[0]+(0.40947793861)*x[1]**o+(-0.371098494308)*x[1]+(-0.189273002147)*x[2]**o+(-0.863169616331)*x[2]
        arg[0,4,1,0]=(-0.929838141145)*x[0]**o+(-0.0237583231579)*x[0]+(0.721069982341)*x[1]**o+(0.0378559913864)*x[1]+(-0.354729434476)*x[2]**o+(0.486256042969)*x[2]
        arg[0,4,1,1]=(0.349085173473)*x[0]**o+(0.398939355666)*x[0]+(-0.165772805366)*x[1]**o+(-0.532771334584)*x[1]+(-0.971868054413)*x[2]**o+(-0.38430185761)*x[2]
        arg[0,4,2,0]=(0.0330282399287)*x[0]**o+(0.157336131691)*x[0]+(-0.232739262685)*x[1]**o+(-0.280149712596)*x[1]+(0.374058430291)*x[2]**o+(0.594113015973)*x[2]
        arg[0,4,2,1]=(0.841740974112)*x[0]**o+(-0.454359680369)*x[0]+(-0.29970574726)*x[1]**o+(-0.234747232007)*x[1]+(0.590942390095)*x[2]**o+(-0.241807214234)*x[2]
        arg[1,0,0,0]=(0.815585848775)*x[0]**o+(-0.915020392674)*x[0]+(-0.0246530790781)*x[1]**o+(-0.783552074981)*x[1]+(0.539847778607)*x[2]**o+(-0.239345161181)*x[2]
        arg[1,0,0,1]=(-0.34212551354)*x[0]**o+(-0.231769108271)*x[0]+(-0.495709071836)*x[1]**o+(-0.301124823696)*x[1]+(-0.567450536966)*x[2]**o+(0.553640896859)*x[2]
        arg[1,0,1,0]=(0.507091065517)*x[0]**o+(-0.844317985101)*x[0]+(0.488126307787)*x[1]**o+(-0.624165922596)*x[1]+(0.00960189690918)*x[2]**o+(-0.35617062688)*x[2]
        arg[1,0,1,1]=(-0.219877421252)*x[0]**o+(-0.574843930095)*x[0]+(-0.846995760292)*x[1]**o+(-0.778496435041)*x[1]+(-0.817593097977)*x[2]**o+(0.387406829939)*x[2]
        arg[1,0,2,0]=(-0.774922277048)*x[0]**o+(-0.238152671014)*x[0]+(-0.913770877808)*x[1]**o+(-0.737305937434)*x[1]+(-0.13977056904)*x[2]**o+(-0.0156668131016)*x[2]
        arg[1,0,2,1]=(0.565993265911)*x[0]**o+(0.0690768496992)*x[0]+(-0.019845214223)*x[1]**o+(0.16253166701)*x[1]+(0.848117215885)*x[2]**o+(0.838075165243)*x[2]
        arg[1,1,0,0]=(0.871499694148)*x[0]**o+(-0.640851595157)*x[0]+(-0.170827284611)*x[1]**o+(-0.855550549988)*x[1]+(0.796086141249)*x[2]**o+(-0.214410672377)*x[2]
        arg[1,1,0,1]=(-0.903644604832)*x[0]**o+(0.0604780584681)*x[0]+(0.12800839523)*x[1]**o+(-0.594186752891)*x[1]+(0.415537334962)*x[2]**o+(0.122944611099)*x[2]
        arg[1,1,1,0]=(-0.611432775957)*x[0]**o+(0.391078622669)*x[0]+(-0.943091949273)*x[1]**o+(0.952739057317)*x[1]+(0.205104161722)*x[2]**o+(-0.946040001739)*x[2]
        arg[1,1,1,1]=(0.122941344579)*x[0]**o+(0.836981501711)*x[0]+(0.488970739079)*x[1]**o+(0.210062341907)*x[1]+(0.0264620393683)*x[2]**o+(-0.396386059797)*x[2]
        arg[1,1,2,0]=(-0.782658774038)*x[0]**o+(0.00611865020069)*x[0]+(0.806802875462)*x[1]**o+(-0.504323859583)*x[1]+(0.796276800331)*x[2]**o+(-0.792865116401)*x[2]
        arg[1,1,2,1]=(-0.249170081188)*x[0]**o+(-0.0234780571796)*x[0]+(-0.314017467872)*x[1]**o+(-0.0819148249592)*x[1]+(-0.71553368486)*x[2]**o+(0.763463207899)*x[2]
        arg[1,2,0,0]=(-0.722263632571)*x[0]**o+(-0.922870211147)*x[0]+(0.380794419995)*x[1]**o+(-0.76088532378)*x[1]+(0.417253071012)*x[2]**o+(0.152980049966)*x[2]
        arg[1,2,0,1]=(-0.667661997967)*x[0]**o+(-0.520570751147)*x[0]+(0.526368189149)*x[1]**o+(-0.0729376045288)*x[1]+(-0.434764684251)*x[2]**o+(0.124284939267)*x[2]
        arg[1,2,1,0]=(0.610868035553)*x[0]**o+(0.958644088455)*x[0]+(0.876029854699)*x[1]**o+(-0.278348787699)*x[1]+(-0.576695356283)*x[2]**o+(-0.302696556744)*x[2]
        arg[1,2,1,1]=(-0.224897853162)*x[0]**o+(0.89405397732)*x[0]+(-0.865281799513)*x[1]**o+(0.778866760053)*x[1]+(-0.939295291326)*x[2]**o+(0.256725768393)*x[2]
        arg[1,2,2,0]=(0.763785330202)*x[0]**o+(-0.92861265304)*x[0]+(-0.73488309202)*x[1]**o+(0.912711310857)*x[1]+(-0.359622626692)*x[2]**o+(-0.901261216611)*x[2]
        arg[1,2,2,1]=(-0.70687794303)*x[0]**o+(0.186497640523)*x[0]+(-0.587602932017)*x[1]**o+(0.439204174904)*x[1]+(0.0527573129765)*x[2]**o+(-0.121921819047)*x[2]
        arg[1,3,0,0]=(-0.576513727689)*x[0]**o+(0.228971758182)*x[0]+(0.282735129059)*x[1]**o+(-0.375270097915)*x[1]+(0.360087142563)*x[2]**o+(-0.0629946372113)*x[2]
        arg[1,3,0,1]=(-0.934375471561)*x[0]**o+(-0.710481522462)*x[0]+(-0.506438555546)*x[1]**o+(0.515278994703)*x[1]+(-0.789552111338)*x[2]**o+(0.556694887963)*x[2]
        arg[1,3,1,0]=(0.418081142778)*x[0]**o+(-0.990334755424)*x[0]+(0.465998147831)*x[1]**o+(0.85211792131)*x[1]+(0.64293315455)*x[2]**o+(0.418636667223)*x[2]
        arg[1,3,1,1]=(-0.192682640907)*x[0]**o+(-0.367176463246)*x[0]+(-0.512565915338)*x[1]**o+(0.431102518715)*x[1]+(-0.0213613616827)*x[2]**o+(-0.329868944656)*x[2]
        arg[1,3,2,0]=(-0.63277285658)*x[0]**o+(0.679819079734)*x[0]+(0.986520110022)*x[1]**o+(0.988046257154)*x[1]+(0.973419900492)*x[2]**o+(-0.177540534942)*x[2]
        arg[1,3,2,1]=(-0.660456888389)*x[0]**o+(0.546277948387)*x[0]+(-0.437065207396)*x[1]**o+(-0.310754727958)*x[1]+(-0.786950057282)*x[2]**o+(0.959526365177)*x[2]
        arg[1,4,0,0]=(0.764190976082)*x[0]**o+(0.970535029727)*x[0]+(0.389051006993)*x[1]**o+(-0.289346925704)*x[1]+(-0.386263606576)*x[2]**o+(0.624234063589)*x[2]
        arg[1,4,0,1]=(-0.982030683687)*x[0]**o+(-0.310500292581)*x[0]+(-0.559096285575)*x[1]**o+(-0.622541272039)*x[1]+(0.49924233081)*x[2]**o+(-0.219258681197)*x[2]
        arg[1,4,1,0]=(-0.124268932891)*x[0]**o+(0.102696501388)*x[0]+(-0.135123912862)*x[1]**o+(-0.960215132796)*x[1]+(-0.63171550556)*x[2]**o+(-0.493725285999)*x[2]
        arg[1,4,1,1]=(-0.397915436399)*x[0]**o+(-0.629822230732)*x[0]+(-0.65097233947)*x[1]**o+(0.915396047247)*x[1]+(0.835684562564)*x[2]**o+(-0.538096550078)*x[2]
        arg[1,4,2,0]=(-0.946429448216)*x[0]**o+(-0.310412085571)*x[0]+(-0.0984587257278)*x[1]**o+(-0.662090989911)*x[1]+(-0.424711481307)*x[2]**o+(0.00916879493127)*x[2]
        arg[1,4,2,1]=(0.919399549311)*x[0]**o+(-0.890110018315)*x[0]+(-0.685928475328)*x[1]**o+(-0.4731186395)*x[1]+(0.961887038718)*x[2]**o+(0.850957316131)*x[2]
        arg[2,0,0,0]=(0.834851801283)*x[0]**o+(-0.401572429371)*x[0]+(0.313309302528)*x[1]**o+(0.0368590314435)*x[1]+(-0.299807168778)*x[2]**o+(0.250599904818)*x[2]
        arg[2,0,0,1]=(-0.717562866174)*x[0]**o+(0.374797888506)*x[0]+(0.254329847579)*x[1]**o+(-0.987429692718)*x[1]+(-0.886758521324)*x[2]**o+(0.309486900444)*x[2]
        arg[2,0,1,0]=(0.139962143021)*x[0]**o+(-0.134735774213)*x[0]+(-0.231254041503)*x[1]**o+(0.0694785371931)*x[1]+(0.769929630132)*x[2]**o+(0.486646094292)*x[2]
        arg[2,0,1,1]=(0.349691701871)*x[0]**o+(-0.0308285150496)*x[0]+(0.00977143515921)*x[1]**o+(0.0970015800122)*x[1]+(-0.725991176667)*x[2]**o+(0.433663449433)*x[2]
        arg[2,0,2,0]=(0.564474039592)*x[0]**o+(0.950620065312)*x[0]+(0.256658075134)*x[1]**o+(0.774458453189)*x[1]+(0.00181734756534)*x[2]**o+(-0.270638852561)*x[2]
        arg[2,0,2,1]=(-0.175813811707)*x[0]**o+(0.821341326493)*x[0]+(-0.464966990893)*x[1]**o+(-0.310747066257)*x[1]+(-0.413557861251)*x[2]**o+(-0.799160067701)*x[2]
        arg[2,1,0,0]=(-0.449817381621)*x[0]**o+(-0.0573780378383)*x[0]+(-0.22675349133)*x[1]**o+(-0.150314340624)*x[1]+(-0.116693721705)*x[2]**o+(-0.20577625675)*x[2]
        arg[2,1,0,1]=(0.955318430674)*x[0]**o+(-0.148100168239)*x[0]+(0.741705440574)*x[1]**o+(-0.0768022054634)*x[1]+(-0.721807872287)*x[2]**o+(-0.898235041116)*x[2]
        arg[2,1,1,0]=(-0.183120810472)*x[0]**o+(0.307305712226)*x[0]+(0.699835684443)*x[1]**o+(0.964082351856)*x[1]+(-0.0348051023458)*x[2]**o+(0.544435382916)*x[2]
        arg[2,1,1,1]=(0.786862912642)*x[0]**o+(0.37388939814)*x[0]+(0.608937302857)*x[1]**o+(0.122305286217)*x[1]+(-0.892536849678)*x[2]**o+(0.119257478145)*x[2]
        arg[2,1,2,0]=(0.758154908623)*x[0]**o+(0.695188856067)*x[0]+(0.458313343957)*x[1]**o+(0.979157471455)*x[1]+(0.194237063725)*x[2]**o+(0.773001101454)*x[2]
        arg[2,1,2,1]=(0.0754678814151)*x[0]**o+(-0.639556181007)*x[0]+(0.934627102078)*x[1]**o+(0.0832082572611)*x[1]+(0.668822756858)*x[2]**o+(0.627532210297)*x[2]
        arg[2,2,0,0]=(-0.192921889174)*x[0]**o+(0.989703287553)*x[0]+(0.00517568784693)*x[1]**o+(-0.360076937731)*x[1]+(0.947056110784)*x[2]**o+(0.469841550381)*x[2]
        arg[2,2,0,1]=(0.201501669595)*x[0]**o+(-0.959957961834)*x[0]+(0.0229159090152)*x[1]**o+(0.396188263084)*x[1]+(0.186211667989)*x[2]**o+(0.304503474778)*x[2]
        arg[2,2,1,0]=(0.333974995608)*x[0]**o+(0.210202867626)*x[0]+(-0.410565897643)*x[1]**o+(0.52659985644)*x[1]+(0.983960564934)*x[2]**o+(-0.439323691866)*x[2]
        arg[2,2,1,1]=(0.706119074329)*x[0]**o+(-0.00220875175563)*x[0]+(0.143693169123)*x[1]**o+(0.217254593235)*x[1]+(-0.412057975252)*x[2]**o+(-0.967360361422)*x[2]
        arg[2,2,2,0]=(0.217307356597)*x[0]**o+(0.861728452163)*x[0]+(-0.193239462241)*x[1]**o+(0.872273176526)*x[1]+(-0.436414721507)*x[2]**o+(-0.62398425635)*x[2]
        arg[2,2,2,1]=(0.467099668794)*x[0]**o+(0.109742400918)*x[0]+(-0.987522191801)*x[1]**o+(-0.250251964275)*x[1]+(0.442152166523)*x[2]**o+(-0.709157679576)*x[2]
        arg[2,3,0,0]=(0.769666242031)*x[0]**o+(0.2352745864)*x[0]+(-0.302893533611)*x[1]**o+(0.0806677655098)*x[1]+(0.886434893279)*x[2]**o+(0.366016417047)*x[2]
        arg[2,3,0,1]=(-0.612204706187)*x[0]**o+(-0.242961993133)*x[0]+(0.946225816441)*x[1]**o+(0.528642650581)*x[1]+(-0.0166837450337)*x[2]**o+(-0.0889397945206)*x[2]
        arg[2,3,1,0]=(0.218589745288)*x[0]**o+(-0.296125811204)*x[0]+(0.358743045538)*x[1]**o+(-0.945778818656)*x[1]+(-0.33155158646)*x[2]**o+(-0.293038455557)*x[2]
        arg[2,3,1,1]=(0.753737723433)*x[0]**o+(-0.772926141856)*x[0]+(-0.0107183417902)*x[1]**o+(0.125538612735)*x[1]+(0.225277420156)*x[2]**o+(-0.54876651451)*x[2]
        arg[2,3,2,0]=(0.947419453319)*x[0]**o+(-0.101935160067)*x[0]+(0.481841505589)*x[1]**o+(-0.873842146032)*x[1]+(-0.377764402402)*x[2]**o+(-0.361560947647)*x[2]
        arg[2,3,2,1]=(0.745703609187)*x[0]**o+(-0.488401078192)*x[0]+(0.190261500318)*x[1]**o+(-0.22176180212)*x[1]+(0.80182740764)*x[2]**o+(-0.504953098351)*x[2]
        arg[2,4,0,0]=(0.476355625206)*x[0]**o+(0.452225231178)*x[0]+(0.669315429911)*x[1]**o+(0.159511010158)*x[1]+(0.652056592943)*x[2]**o+(0.44515033609)*x[2]
        arg[2,4,0,1]=(0.512194809916)*x[0]**o+(0.654452558318)*x[0]+(-0.526832882528)*x[1]**o+(0.590859897981)*x[1]+(-0.155989159043)*x[2]**o+(-0.689178072559)*x[2]
        arg[2,4,1,0]=(0.0450244624553)*x[0]**o+(0.0790299000335)*x[0]+(-0.690643908994)*x[1]**o+(-0.262426911912)*x[1]+(-0.426935471303)*x[2]**o+(-0.434570506153)*x[2]
        arg[2,4,1,1]=(-0.04445890079)*x[0]**o+(-0.541574770252)*x[0]+(-0.879290270122)*x[1]**o+(0.545302743626)*x[1]+(-0.836472565261)*x[2]**o+(0.66750272736)*x[2]
        arg[2,4,2,0]=(-0.477788509677)*x[0]**o+(0.706195950613)*x[0]+(0.628430029293)*x[1]**o+(-0.297608932813)*x[1]+(0.38364501438)*x[2]**o+(0.855444938831)*x[2]
        arg[2,4,2,1]=(0.123183000451)*x[0]**o+(0.41176132265)*x[0]+(-0.379498883356)*x[1]**o+(-0.53442961658)*x[1]+(-0.101215056317)*x[2]**o+(0.247971839461)*x[2]
        arg[3,0,0,0]=(-0.239170701904)*x[0]**o+(0.369420936359)*x[0]+(0.972913513036)*x[1]**o+(0.775248415403)*x[1]+(-0.850310640223)*x[2]**o+(-0.815345534491)*x[2]
        arg[3,0,0,1]=(0.591124498004)*x[0]**o+(0.529522171858)*x[0]+(-0.974005221641)*x[1]**o+(0.430102201571)*x[1]+(0.570519406976)*x[2]**o+(-0.0915083364671)*x[2]
        arg[3,0,1,0]=(-0.27678065557)*x[0]**o+(-0.53225308602)*x[0]+(0.902356190541)*x[1]**o+(-0.564808732046)*x[1]+(0.0981283775668)*x[2]**o+(-0.64957631107)*x[2]
        arg[3,0,1,1]=(0.230523873408)*x[0]**o+(-0.382745197262)*x[0]+(0.41507101812)*x[1]**o+(0.48492708335)*x[1]+(-0.11672353881)*x[2]**o+(-0.56902283397)*x[2]
        arg[3,0,2,0]=(-0.400023140104)*x[0]**o+(-0.406234012546)*x[0]+(-0.17810623045)*x[1]**o+(-0.833748300281)*x[1]+(-0.463689953863)*x[2]**o+(-0.11095482465)*x[2]
        arg[3,0,2,1]=(0.448885918108)*x[0]**o+(0.0447046792888)*x[0]+(-0.781652457281)*x[1]**o+(-0.893726375109)*x[1]+(0.918099233406)*x[2]**o+(0.560717614604)*x[2]
        arg[3,1,0,0]=(0.046700262853)*x[0]**o+(0.729867044213)*x[0]+(-0.870868205823)*x[1]**o+(-0.711267787629)*x[1]+(0.760550112312)*x[2]**o+(-0.0188334901698)*x[2]
        arg[3,1,0,1]=(0.304305551549)*x[0]**o+(-0.73527291962)*x[0]+(0.722134673974)*x[1]**o+(-0.40822784698)*x[1]+(0.912464122456)*x[2]**o+(-0.857656020349)*x[2]
        arg[3,1,1,0]=(-0.857864508282)*x[0]**o+(-0.22160002809)*x[0]+(-0.0892530574233)*x[1]**o+(-0.547259900671)*x[1]+(0.641858628113)*x[2]**o+(-0.203700292205)*x[2]
        arg[3,1,1,1]=(-0.47064403668)*x[0]**o+(-0.105885697358)*x[0]+(-0.899570240486)*x[1]**o+(0.660942689756)*x[1]+(-0.365836826289)*x[2]**o+(0.9593955249)*x[2]
        arg[3,1,2,0]=(-0.51407433496)*x[0]**o+(0.369092182634)*x[0]+(-0.740281092517)*x[1]**o+(0.383443939366)*x[1]+(-0.140318759419)*x[2]**o+(-0.481274651104)*x[2]
        arg[3,1,2,1]=(-0.938924125636)*x[0]**o+(0.41994781084)*x[0]+(-0.946320511542)*x[1]**o+(-0.86702606561)*x[1]+(0.63547569926)*x[2]**o+(0.923988068696)*x[2]
        arg[3,2,0,0]=(0.212833889165)*x[0]**o+(-0.0823077975238)*x[0]+(0.038483237489)*x[1]**o+(0.393784376623)*x[1]+(0.573190557103)*x[2]**o+(0.403015743632)*x[2]
        arg[3,2,0,1]=(-0.169294531943)*x[0]**o+(-0.769228919251)*x[0]+(-0.907462814057)*x[1]**o+(-0.188574782291)*x[1]+(-0.859793027531)*x[2]**o+(-0.673334626964)*x[2]
        arg[3,2,1,0]=(0.646744388517)*x[0]**o+(0.346749855021)*x[0]+(-0.314382791156)*x[1]**o+(-0.0396545336605)*x[1]+(0.844678566939)*x[2]**o+(0.724721336727)*x[2]
        arg[3,2,1,1]=(-0.471761417345)*x[0]**o+(-0.801631597222)*x[0]+(-0.376444012893)*x[1]**o+(0.920939382926)*x[1]+(-0.728019331398)*x[2]**o+(0.905808341144)*x[2]
        arg[3,2,2,0]=(-0.713759272803)*x[0]**o+(-0.930638013429)*x[0]+(0.864451879958)*x[1]**o+(0.0566794358824)*x[1]+(-0.292092226905)*x[2]**o+(-0.195695530268)*x[2]
        arg[3,2,2,1]=(0.843437618194)*x[0]**o+(0.214911962718)*x[0]+(0.475690162826)*x[1]**o+(0.0744570063618)*x[1]+(0.351711310616)*x[2]**o+(-0.187223450548)*x[2]
        arg[3,3,0,0]=(0.108651500878)*x[0]**o+(-0.402624590847)*x[0]+(0.120700159277)*x[1]**o+(-0.363139807635)*x[1]+(-0.62413441951)*x[2]**o+(-0.474236131596)*x[2]
        arg[3,3,0,1]=(-0.0836636547536)*x[0]**o+(-0.187959781777)*x[0]+(0.756983584308)*x[1]**o+(0.756573839599)*x[1]+(0.195640878564)*x[2]**o+(-0.566596474032)*x[2]
        arg[3,3,1,0]=(-0.171427844449)*x[0]**o+(-0.266807792484)*x[0]+(0.380046037965)*x[1]**o+(-0.129297371793)*x[1]+(0.79139100315)*x[2]**o+(0.223804751067)*x[2]
        arg[3,3,1,1]=(-0.054757303264)*x[0]**o+(-0.823878124196)*x[0]+(0.80695256107)*x[1]**o+(0.583250192211)*x[1]+(0.894937939133)*x[2]**o+(0.878757173892)*x[2]
        arg[3,3,2,0]=(0.13923802434)*x[0]**o+(-0.600732296834)*x[0]+(0.111226716676)*x[1]**o+(-0.474524749496)*x[1]+(0.674290476359)*x[2]**o+(-0.06714767759)*x[2]
        arg[3,3,2,1]=(0.246187769682)*x[0]**o+(-0.871969439162)*x[0]+(0.56748693564)*x[1]**o+(-0.0831774328079)*x[1]+(-0.27380467532)*x[2]**o+(-0.623711953244)*x[2]
        arg[3,4,0,0]=(-0.345947366312)*x[0]**o+(-0.304842609526)*x[0]+(-0.720722450633)*x[1]**o+(0.889712922756)*x[1]+(-0.0451362945237)*x[2]**o+(-0.52344522887)*x[2]
        arg[3,4,0,1]=(-0.830119647904)*x[0]**o+(0.577888363835)*x[0]+(0.180691945154)*x[1]**o+(-0.489554894073)*x[1]+(0.848489044545)*x[2]**o+(-0.243139985297)*x[2]
        arg[3,4,1,0]=(0.846597693999)*x[0]**o+(0.757477455396)*x[0]+(-0.827604908657)*x[1]**o+(-0.108392312717)*x[1]+(0.597454707613)*x[2]**o+(0.35846243312)*x[2]
        arg[3,4,1,1]=(0.107973753109)*x[0]**o+(0.578649460223)*x[0]+(0.609012033524)*x[1]**o+(-0.8428861922)*x[1]+(0.42144759968)*x[2]**o+(-0.938952208084)*x[2]
        arg[3,4,2,0]=(0.286586330618)*x[0]**o+(0.7324646289)*x[0]+(-0.448461377611)*x[1]**o+(0.62562408085)*x[1]+(0.352429283116)*x[2]**o+(0.112047034296)*x[2]
        arg[3,4,2,1]=(-0.431663387082)*x[0]**o+(-0.612934322552)*x[0]+(-0.426644797074)*x[1]**o+(0.768645027094)*x[1]+(-0.453242885757)*x[2]**o+(0.893748345809)*x[2]
        ref[0,0,0,0]=(-1.2147562394)*(1+2.*(dim-1.)/(o+1.))+(1.47471604216)*dim
        ref[0,0,0,1]=(-0.283261030731)*(1+2.*(dim-1.)/(o+1.))+(0.49706944223)*dim
        ref[0,0,1,0]=(-0.963543225803)*(1+2.*(dim-1.)/(o+1.))+(0.302115929232)*dim
        ref[0,0,1,1]=(1.47597923019)*(1+2.*(dim-1.)/(o+1.))+(0.302975596116)*dim
        ref[0,0,2,0]=(2.07218071265)*(1+2.*(dim-1.)/(o+1.))+(1.40474756312)*dim
        ref[0,0,2,1]=(1.18935174934)*(1+2.*(dim-1.)/(o+1.))+(-0.898949809002)*dim
        ref[0,1,0,0]=(-2.71454215957)*(1+2.*(dim-1.)/(o+1.))+(-1.81174147681)*dim
        ref[0,1,0,1]=(-1.27877046381)*(1+2.*(dim-1.)/(o+1.))+(0.197262183297)*dim
        ref[0,1,1,0]=(-0.287148009258)*(1+2.*(dim-1.)/(o+1.))+(1.30812352716)*dim
        ref[0,1,1,1]=(0.263153075777)*(1+2.*(dim-1.)/(o+1.))+(1.46256384302)*dim
        ref[0,1,2,0]=(-2.55250588186)*(1+2.*(dim-1.)/(o+1.))+(1.13566208594)*dim
        ref[0,1,2,1]=(0.208932386724)*(1+2.*(dim-1.)/(o+1.))+(-0.438703456176)*dim
        ref[0,2,0,0]=(-1.09576259026)*(1+2.*(dim-1.)/(o+1.))+(-0.13076989899)*dim
        ref[0,2,0,1]=(0.401914700214)*(1+2.*(dim-1.)/(o+1.))+(2.19797357625)*dim
        ref[0,2,1,0]=(0.860631336354)*(1+2.*(dim-1.)/(o+1.))+(0.978915494035)*dim
        ref[0,2,1,1]=(-1.97095623301)*(1+2.*(dim-1.)/(o+1.))+(-2.25442123467)*dim
        ref[0,2,2,0]=(-0.527474241147)*(1+2.*(dim-1.)/(o+1.))+(-0.923478312465)*dim
        ref[0,2,2,1]=(-0.460568018944)*(1+2.*(dim-1.)/(o+1.))+(0.173004367644)*dim
        ref[0,3,0,0]=(-0.443819991343)*(1+2.*(dim-1.)/(o+1.))+(1.08379961978)*dim
        ref[0,3,0,1]=(-0.80563854076)*(1+2.*(dim-1.)/(o+1.))+(-1.48015007036)*dim
        ref[0,3,1,0]=(-1.04939720148)*(1+2.*(dim-1.)/(o+1.))+(-0.762532531252)*dim
        ref[0,3,1,1]=(0.472870331273)*(1+2.*(dim-1.)/(o+1.))+(-0.00544721475928)*dim
        ref[0,3,2,0]=(-0.593230167381)*(1+2.*(dim-1.)/(o+1.))+(0.26676895359)*dim
        ref[0,3,2,1]=(-0.104009550799)*(1+2.*(dim-1.)/(o+1.))+(-2.23842412891)*dim
        ref[0,4,0,0]=(-0.352866783806)*(1+2.*(dim-1.)/(o+1.))+(-1.02623939367)*dim
        ref[0,4,0,1]=(0.0310776980081)*(1+2.*(dim-1.)/(o+1.))+(-1.9281752378)*dim
        ref[0,4,1,0]=(-0.56349759328)*(1+2.*(dim-1.)/(o+1.))+(0.500353711197)*dim
        ref[0,4,1,1]=(-0.788555686306)*(1+2.*(dim-1.)/(o+1.))+(-0.518133836527)*dim
        ref[0,4,2,0]=(0.174347407535)*(1+2.*(dim-1.)/(o+1.))+(0.471299435068)*dim
        ref[0,4,2,1]=(1.13297761695)*(1+2.*(dim-1.)/(o+1.))+(-0.93091412661)*dim
        ref[1,0,0,0]=(1.3307805483)*(1+2.*(dim-1.)/(o+1.))+(-1.93791762884)*dim
        ref[1,0,0,1]=(-1.40528512234)*(1+2.*(dim-1.)/(o+1.))+(0.0207469648915)*dim
        ref[1,0,1,0]=(1.00481927021)*(1+2.*(dim-1.)/(o+1.))+(-1.82465453458)*dim
        ref[1,0,1,1]=(-1.88446627952)*(1+2.*(dim-1.)/(o+1.))+(-0.965933535196)*dim
        ref[1,0,2,0]=(-1.8284637239)*(1+2.*(dim-1.)/(o+1.))+(-0.991125421549)*dim
        ref[1,0,2,1]=(1.39426526757)*(1+2.*(dim-1.)/(o+1.))+(1.06968368195)*dim
        ref[1,1,0,0]=(1.49675855079)*(1+2.*(dim-1.)/(o+1.))+(-1.71081281752)*dim
        ref[1,1,0,1]=(-0.36009887464)*(1+2.*(dim-1.)/(o+1.))+(-0.410764083324)*dim
        ref[1,1,1,0]=(-1.34942056351)*(1+2.*(dim-1.)/(o+1.))+(0.397777678247)*dim
        ref[1,1,1,1]=(0.638374123027)*(1+2.*(dim-1.)/(o+1.))+(0.650657783821)*dim
        ref[1,1,2,0]=(0.820420901755)*(1+2.*(dim-1.)/(o+1.))+(-1.29107032578)*dim
        ref[1,1,2,1]=(-1.27872123392)*(1+2.*(dim-1.)/(o+1.))+(0.65807032576)*dim
        ref[1,2,0,0]=(0.075783858436)*(1+2.*(dim-1.)/(o+1.))+(-1.53077548496)*dim
        ref[1,2,0,1]=(-0.576058493068)*(1+2.*(dim-1.)/(o+1.))+(-0.469223416409)*dim
        ref[1,2,1,0]=(0.91020253397)*(1+2.*(dim-1.)/(o+1.))+(0.377598744012)*dim
        ref[1,2,1,1]=(-2.029474944)*(1+2.*(dim-1.)/(o+1.))+(1.92964650577)*dim
        ref[1,2,2,0]=(-0.330720388509)*(1+2.*(dim-1.)/(o+1.))+(-0.917162558794)*dim
        ref[1,2,2,1]=(-1.24172356207)*(1+2.*(dim-1.)/(o+1.))+(0.503779996381)*dim
        ref[1,3,0,0]=(0.0663085439332)*(1+2.*(dim-1.)/(o+1.))+(-0.209292976945)*dim
        ref[1,3,0,1]=(-2.23036613844)*(1+2.*(dim-1.)/(o+1.))+(0.361492360204)*dim
        ref[1,3,1,0]=(1.52701244516)*(1+2.*(dim-1.)/(o+1.))+(0.280419833109)*dim
        ref[1,3,1,1]=(-0.726609917928)*(1+2.*(dim-1.)/(o+1.))+(-0.265942889187)*dim
        ref[1,3,2,0]=(1.32716715393)*(1+2.*(dim-1.)/(o+1.))+(1.49032480195)*dim
        ref[1,3,2,1]=(-1.88447215307)*(1+2.*(dim-1.)/(o+1.))+(1.19504958561)*dim
        ref[1,4,0,0]=(0.766978376499)*(1+2.*(dim-1.)/(o+1.))+(1.30542216761)*dim
        ref[1,4,0,1]=(-1.04188463845)*(1+2.*(dim-1.)/(o+1.))+(-1.15230024582)*dim
        ref[1,4,1,0]=(-0.891108351313)*(1+2.*(dim-1.)/(o+1.))+(-1.35124391741)*dim
        ref[1,4,1,1]=(-0.213203213304)*(1+2.*(dim-1.)/(o+1.))+(-0.252522733564)*dim
        ref[1,4,2,0]=(-1.46959965525)*(1+2.*(dim-1.)/(o+1.))+(-0.963334280551)*dim
        ref[1,4,2,1]=(1.1953581127)*(1+2.*(dim-1.)/(o+1.))+(-0.512271341684)*dim
        ref[2,0,0,0]=(0.848353935033)*(1+2.*(dim-1.)/(o+1.))+(-0.11411349311)*dim
        ref[2,0,0,1]=(-1.34999153992)*(1+2.*(dim-1.)/(o+1.))+(-0.303144903768)*dim
        ref[2,0,1,0]=(0.67863773165)*(1+2.*(dim-1.)/(o+1.))+(0.421388857272)*dim
        ref[2,0,1,1]=(-0.366528039637)*(1+2.*(dim-1.)/(o+1.))+(0.499836514396)*dim
        ref[2,0,2,0]=(0.822949462292)*(1+2.*(dim-1.)/(o+1.))+(1.45443966594)*dim
        ref[2,0,2,1]=(-1.05433866385)*(1+2.*(dim-1.)/(o+1.))+(-0.288565807465)*dim
        ref[2,1,0,0]=(-0.793264594657)*(1+2.*(dim-1.)/(o+1.))+(-0.413468635212)*dim
        ref[2,1,0,1]=(0.975215998961)*(1+2.*(dim-1.)/(o+1.))+(-1.12313741482)*dim
        ref[2,1,1,0]=(0.481909771625)*(1+2.*(dim-1.)/(o+1.))+(1.815823447)*dim
        ref[2,1,1,1]=(0.503263365821)*(1+2.*(dim-1.)/(o+1.))+(0.615452162503)*dim
        ref[2,1,2,0]=(1.41070531631)*(1+2.*(dim-1.)/(o+1.))+(2.44734742898)*dim
        ref[2,1,2,1]=(1.67891774035)*(1+2.*(dim-1.)/(o+1.))+(0.0711842865505)*dim
        ref[2,2,0,0]=(0.759309909457)*(1+2.*(dim-1.)/(o+1.))+(1.0994679002)*dim
        ref[2,2,0,1]=(0.410629246599)*(1+2.*(dim-1.)/(o+1.))+(-0.259266223972)*dim
        ref[2,2,1,0]=(0.907369662899)*(1+2.*(dim-1.)/(o+1.))+(0.297479032201)*dim
        ref[2,2,1,1]=(0.437754268199)*(1+2.*(dim-1.)/(o+1.))+(-0.752314519943)*dim
        ref[2,2,2,0]=(-0.412346827151)*(1+2.*(dim-1.)/(o+1.))+(1.11001737234)*dim
        ref[2,2,2,1]=(-0.0782703564833)*(1+2.*(dim-1.)/(o+1.))+(-0.849667242932)*dim
        ref[2,3,0,0]=(1.3532076017)*(1+2.*(dim-1.)/(o+1.))+(0.681958768956)*dim
        ref[2,3,0,1]=(0.317337365221)*(1+2.*(dim-1.)/(o+1.))+(0.196740862928)*dim
        ref[2,3,1,0]=(0.245781204366)*(1+2.*(dim-1.)/(o+1.))+(-1.53494308542)*dim
        ref[2,3,1,1]=(0.968296801799)*(1+2.*(dim-1.)/(o+1.))+(-1.19615404363)*dim
        ref[2,3,2,0]=(1.05149655651)*(1+2.*(dim-1.)/(o+1.))+(-1.33733825375)*dim
        ref[2,3,2,1]=(1.73779251715)*(1+2.*(dim-1.)/(o+1.))+(-1.21511597866)*dim
        ref[2,4,0,0]=(1.79772764806)*(1+2.*(dim-1.)/(o+1.))+(1.05688657743)*dim
        ref[2,4,0,1]=(-0.170627231654)*(1+2.*(dim-1.)/(o+1.))+(0.55613438374)*dim
        ref[2,4,1,0]=(-1.07255491784)*(1+2.*(dim-1.)/(o+1.))+(-0.617967518031)*dim
        ref[2,4,1,1]=(-1.76022173617)*(1+2.*(dim-1.)/(o+1.))+(0.671230700733)*dim
        ref[2,4,2,0]=(0.534286533995)*(1+2.*(dim-1.)/(o+1.))+(1.26403195663)*dim
        ref[2,4,2,1]=(-0.357530939222)*(1+2.*(dim-1.)/(o+1.))+(0.12530354553)*dim
        ref[3,0,0,0]=(-0.11656782909)*(1+2.*(dim-1.)/(o+1.))+(0.32932381727)*dim
        ref[3,0,0,1]=(0.187638683339)*(1+2.*(dim-1.)/(o+1.))+(0.868116036962)*dim
        ref[3,0,1,0]=(0.723703912538)*(1+2.*(dim-1.)/(o+1.))+(-1.74663812914)*dim
        ref[3,0,1,1]=(0.528871352717)*(1+2.*(dim-1.)/(o+1.))+(-0.466840947882)*dim
        ref[3,0,2,0]=(-1.04181932442)*(1+2.*(dim-1.)/(o+1.))+(-1.35093713748)*dim
        ref[3,0,2,1]=(0.585332694234)*(1+2.*(dim-1.)/(o+1.))+(-0.288304081216)*dim
        ref[3,1,0,0]=(-0.0636178306576)*(1+2.*(dim-1.)/(o+1.))+(-0.000234233586678)*dim
        ref[3,1,0,1]=(1.93890434798)*(1+2.*(dim-1.)/(o+1.))+(-2.00115678695)*dim
        ref[3,1,1,0]=(-0.305258937592)*(1+2.*(dim-1.)/(o+1.))+(-0.972560220966)*dim
        ref[3,1,1,1]=(-1.73605110346)*(1+2.*(dim-1.)/(o+1.))+(1.5144525173)*dim
        ref[3,1,2,0]=(-1.3946741869)*(1+2.*(dim-1.)/(o+1.))+(0.271261470896)*dim
        ref[3,1,2,1]=(-1.24976893792)*(1+2.*(dim-1.)/(o+1.))+(0.476909813927)*dim
        ref[3,2,0,0]=(0.824507683756)*(1+2.*(dim-1.)/(o+1.))+(0.71449232273)*dim
        ref[3,2,0,1]=(-1.93655037353)*(1+2.*(dim-1.)/(o+1.))+(-1.63113832851)*dim
        ref[3,2,1,0]=(1.1770401643)*(1+2.*(dim-1.)/(o+1.))+(1.03181665809)*dim
        ref[3,2,1,1]=(-1.57622476164)*(1+2.*(dim-1.)/(o+1.))+(1.02511612685)*dim
        ref[3,2,2,0]=(-0.141399619749)*(1+2.*(dim-1.)/(o+1.))+(-1.06965410781)*dim
        ref[3,2,2,1]=(1.67083909164)*(1+2.*(dim-1.)/(o+1.))+(0.102145518531)*dim
        ref[3,3,0,0]=(-0.394782759355)*(1+2.*(dim-1.)/(o+1.))+(-1.24000053008)*dim
        ref[3,3,0,1]=(0.868960808118)*(1+2.*(dim-1.)/(o+1.))+(0.00201758379073)*dim
        ref[3,3,1,0]=(1.00000919667)*(1+2.*(dim-1.)/(o+1.))+(-0.17230041321)*dim
        ref[3,3,1,1]=(1.64713319694)*(1+2.*(dim-1.)/(o+1.))+(0.638129241907)*dim
        ref[3,3,2,0]=(0.924755217374)*(1+2.*(dim-1.)/(o+1.))+(-1.14240472392)*dim
        ref[3,3,2,1]=(0.539870030002)*(1+2.*(dim-1.)/(o+1.))+(-1.57885882521)*dim
        ref[3,4,0,0]=(-1.11180611147)*(1+2.*(dim-1.)/(o+1.))+(0.0614250843596)*dim
        ref[3,4,0,1]=(0.199061341795)*(1+2.*(dim-1.)/(o+1.))+(-0.154806515534)*dim
        ref[3,4,1,0]=(0.616447492955)*(1+2.*(dim-1.)/(o+1.))+(1.0075475758)*dim
        ref[3,4,1,1]=(1.13843338631)*(1+2.*(dim-1.)/(o+1.))+(-1.20318894006)*dim
        ref[3,4,2,0]=(0.190554236123)*(1+2.*(dim-1.)/(o+1.))+(1.47013574405)*dim
        ref[3,4,2,1]=(-1.31155106991)*(1+2.*(dim-1.)/(o+1.))+(1.04945905035)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.416878137535)*x[0]**o+(0.0809721296208)*x[0]+(-0.275084069478)*x[1]**o+(-0.704226656249)*x[1]
        ref=(0.141794068057)*(1+2.*(dim-1.)/(o+1.))+(-0.623254526628)*dim
      else:
        arg=(0.587142789605)*x[0]**o+(0.894018391929)*x[0]+(-0.417709558814)*x[1]**o+(0.0970658484968)*x[1]+(-0.840982437679)*x[2]**o+(-0.400242562091)*x[2]
        ref=(-0.671549206889)*(1+2.*(dim-1.)/(o+1.))+(0.590841678335)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.051335729494)*x[0]**o+(0.257057962305)*x[0]+(0.00794214446212)*x[1]**o+(0.52307365061)*x[1]
        arg[1]=(0.419915493803)*x[0]**o+(-0.0105826268931)*x[0]+(0.369274332127)*x[1]**o+(0.340796701053)*x[1]
        ref[0]=(0.0592778739561)*(1+2.*(dim-1.)/(o+1.))+(0.780131612915)*dim
        ref[1]=(0.789189825931)*(1+2.*(dim-1.)/(o+1.))+(0.33021407416)*dim
      else:
        arg[0]=(-0.483302491019)*x[0]**o+(0.223892398776)*x[0]+(-0.555612541317)*x[1]**o+(0.75535374219)*x[1]+(-0.208573228788)*x[2]**o+(0.479879212564)*x[2]
        arg[1]=(-0.012341185885)*x[0]**o+(-0.959474230066)*x[0]+(0.397671123434)*x[1]**o+(-0.120340696903)*x[1]+(-0.645757653205)*x[2]**o+(-0.216057959957)*x[2]
        ref[0]=(-1.24748826112)*(1+2.*(dim-1.)/(o+1.))+(1.45912535353)*dim
        ref[1]=(-0.260427715657)*(1+2.*(dim-1.)/(o+1.))+(-1.29587288693)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.493237934301)*x[0]**o+(-0.977176505302)*x[0]+(-0.837740403623)*x[1]**o+(0.204709065496)*x[1]
        arg[0,1]=(0.794064508577)*x[0]**o+(-0.917060945623)*x[0]+(0.0114185601474)*x[1]**o+(0.547108694374)*x[1]
        arg[0,2]=(-0.989778984923)*x[0]**o+(-0.499651095311)*x[0]+(-0.126817505183)*x[1]**o+(0.889077385524)*x[1]
        arg[0,3]=(-0.223453594493)*x[0]**o+(-0.17983229649)*x[0]+(0.352602269148)*x[1]**o+(0.00617724258273)*x[1]
        arg[0,4]=(0.401784356952)*x[0]**o+(0.839308128877)*x[0]+(0.322943402455)*x[1]**o+(0.770206521944)*x[1]
        arg[1,0]=(0.491015078355)*x[0]**o+(0.194909202321)*x[0]+(0.38986016491)*x[1]**o+(0.671106836372)*x[1]
        arg[1,1]=(0.200697856784)*x[0]**o+(0.389746053244)*x[0]+(0.850237988574)*x[1]**o+(-0.677750684271)*x[1]
        arg[1,2]=(0.300444129016)*x[0]**o+(-0.502811932395)*x[0]+(0.390751425646)*x[1]**o+(-0.0234058348564)*x[1]
        arg[1,3]=(-0.451410318068)*x[0]**o+(0.0959092789621)*x[0]+(0.364602756931)*x[1]**o+(-0.637883149677)*x[1]
        arg[1,4]=(0.84148898142)*x[0]**o+(0.447701396385)*x[0]+(-0.925960934959)*x[1]**o+(-0.216915201171)*x[1]
        arg[2,0]=(0.497685540053)*x[0]**o+(0.706445401439)*x[0]+(0.278688226804)*x[1]**o+(0.263560807934)*x[1]
        arg[2,1]=(0.846700886302)*x[0]**o+(0.794459987626)*x[0]+(0.206707548962)*x[1]**o+(0.536938178952)*x[1]
        arg[2,2]=(-0.825060225713)*x[0]**o+(0.00650901742391)*x[0]+(0.500591036904)*x[1]**o+(0.86251606869)*x[1]
        arg[2,3]=(-0.490203731083)*x[0]**o+(-0.758885551614)*x[0]+(0.44212368096)*x[1]**o+(0.422183463853)*x[1]
        arg[2,4]=(0.326180837967)*x[0]**o+(0.0511693925274)*x[0]+(0.450713817213)*x[1]**o+(-0.788372992203)*x[1]
        arg[3,0]=(0.427328335694)*x[0]**o+(-0.549570782632)*x[0]+(0.0484849428866)*x[1]**o+(-0.488921618371)*x[1]
        arg[3,1]=(-0.000395237589224)*x[0]**o+(-0.902889768871)*x[0]+(-0.991821787559)*x[1]**o+(-0.907842202962)*x[1]
        arg[3,2]=(0.737156280391)*x[0]**o+(0.104714673958)*x[0]+(0.102528971835)*x[1]**o+(-0.0982626427735)*x[1]
        arg[3,3]=(0.971460723075)*x[0]**o+(-0.759821212391)*x[0]+(0.385963754923)*x[1]**o+(-0.533924725996)*x[1]
        arg[3,4]=(0.25440424638)*x[0]**o+(-0.82432588605)*x[0]+(0.759285911224)*x[1]**o+(0.328467600746)*x[1]
        ref[0,0]=(-1.33097833792)*(1+2.*(dim-1.)/(o+1.))+(-0.772467439806)*dim
        ref[0,1]=(0.805483068724)*(1+2.*(dim-1.)/(o+1.))+(-0.36995225125)*dim
        ref[0,2]=(-1.11659649011)*(1+2.*(dim-1.)/(o+1.))+(0.389426290213)*dim
        ref[0,3]=(0.129148674655)*(1+2.*(dim-1.)/(o+1.))+(-0.173655053907)*dim
        ref[0,4]=(0.724727759407)*(1+2.*(dim-1.)/(o+1.))+(1.60951465082)*dim
        ref[1,0]=(0.880875243265)*(1+2.*(dim-1.)/(o+1.))+(0.866016038693)*dim
        ref[1,1]=(1.05093584536)*(1+2.*(dim-1.)/(o+1.))+(-0.288004631027)*dim
        ref[1,2]=(0.691195554662)*(1+2.*(dim-1.)/(o+1.))+(-0.526217767251)*dim
        ref[1,3]=(-0.0868075611369)*(1+2.*(dim-1.)/(o+1.))+(-0.541973870715)*dim
        ref[1,4]=(-0.0844719535392)*(1+2.*(dim-1.)/(o+1.))+(0.230786195214)*dim
        ref[2,0]=(0.776373766857)*(1+2.*(dim-1.)/(o+1.))+(0.970006209374)*dim
        ref[2,1]=(1.05340843526)*(1+2.*(dim-1.)/(o+1.))+(1.33139816658)*dim
        ref[2,2]=(-0.324469188808)*(1+2.*(dim-1.)/(o+1.))+(0.869025086114)*dim
        ref[2,3]=(-0.0480800501232)*(1+2.*(dim-1.)/(o+1.))+(-0.336702087761)*dim
        ref[2,4]=(0.776894655179)*(1+2.*(dim-1.)/(o+1.))+(-0.737203599675)*dim
        ref[3,0]=(0.47581327858)*(1+2.*(dim-1.)/(o+1.))+(-1.038492401)*dim
        ref[3,1]=(-0.992217025148)*(1+2.*(dim-1.)/(o+1.))+(-1.81073197183)*dim
        ref[3,2]=(0.839685252226)*(1+2.*(dim-1.)/(o+1.))+(0.00645203118434)*dim
        ref[3,3]=(1.357424478)*(1+2.*(dim-1.)/(o+1.))+(-1.29374593839)*dim
        ref[3,4]=(1.0136901576)*(1+2.*(dim-1.)/(o+1.))+(-0.495858285305)*dim
      else:
        arg[0,0]=(-0.543785449843)*x[0]**o+(-0.593471310422)*x[0]+(0.223732084513)*x[1]**o+(-0.126390494354)*x[1]+(0.602561108649)*x[2]**o+(-0.879650196727)*x[2]
        arg[0,1]=(-0.373708985284)*x[0]**o+(0.274723212925)*x[0]+(0.0901068561926)*x[1]**o+(0.0705878267418)*x[1]+(-0.87124855534)*x[2]**o+(-0.342156987834)*x[2]
        arg[0,2]=(0.197662792758)*x[0]**o+(0.993389348528)*x[0]+(0.300080063659)*x[1]**o+(-0.87329423474)*x[1]+(-0.708861668092)*x[2]**o+(-0.145567510817)*x[2]
        arg[0,3]=(-0.641680604707)*x[0]**o+(0.502870155472)*x[0]+(0.726460257571)*x[1]**o+(-0.796246941548)*x[1]+(-0.034138694789)*x[2]**o+(-0.756773445728)*x[2]
        arg[0,4]=(-0.910301138852)*x[0]**o+(0.370603440687)*x[0]+(0.30908632679)*x[1]**o+(0.399307258883)*x[1]+(0.734329887794)*x[2]**o+(-0.105355323212)*x[2]
        arg[1,0]=(-0.998162269861)*x[0]**o+(0.744436497607)*x[0]+(0.474991680035)*x[1]**o+(-0.523297766774)*x[1]+(-0.147411103004)*x[2]**o+(-0.366260797475)*x[2]
        arg[1,1]=(-0.895068058812)*x[0]**o+(-0.543026567202)*x[0]+(0.681910579939)*x[1]**o+(0.839238385122)*x[1]+(-0.943454697555)*x[2]**o+(-0.942882687085)*x[2]
        arg[1,2]=(-0.0327934757673)*x[0]**o+(-0.551060110327)*x[0]+(0.0246419122114)*x[1]**o+(0.290492943846)*x[1]+(-0.81629260256)*x[2]**o+(-0.0703654573398)*x[2]
        arg[1,3]=(0.0101368428324)*x[0]**o+(0.169228100407)*x[0]+(-0.0720675504637)*x[1]**o+(0.854569517064)*x[1]+(-0.185495614033)*x[2]**o+(0.159402892807)*x[2]
        arg[1,4]=(-0.879056526056)*x[0]**o+(-0.499865495878)*x[0]+(0.389281340553)*x[1]**o+(0.208323326648)*x[1]+(0.201495185869)*x[2]**o+(0.287693342543)*x[2]
        arg[2,0]=(-0.0980367959573)*x[0]**o+(-0.743908923973)*x[0]+(0.612332291133)*x[1]**o+(-0.833162883242)*x[1]+(-0.149234146487)*x[2]**o+(-0.265352382043)*x[2]
        arg[2,1]=(-0.731053054407)*x[0]**o+(-0.328824749845)*x[0]+(-0.270771833154)*x[1]**o+(-0.924718425874)*x[1]+(-0.333737248999)*x[2]**o+(-0.437707530432)*x[2]
        arg[2,2]=(0.0568240255423)*x[0]**o+(0.889929679474)*x[0]+(-0.550346007257)*x[1]**o+(-0.276114383355)*x[1]+(0.818856335741)*x[2]**o+(0.949830759195)*x[2]
        arg[2,3]=(-0.487278099196)*x[0]**o+(0.621833922878)*x[0]+(-0.641187646288)*x[1]**o+(-0.62187991518)*x[1]+(0.193101177186)*x[2]**o+(0.574039256278)*x[2]
        arg[2,4]=(0.411046922073)*x[0]**o+(-0.10196277248)*x[0]+(-0.880065306034)*x[1]**o+(0.629582927604)*x[1]+(-0.364618567951)*x[2]**o+(0.130293365676)*x[2]
        arg[3,0]=(-0.625244224555)*x[0]**o+(0.50345243521)*x[0]+(0.701616120451)*x[1]**o+(0.44401843794)*x[1]+(-0.881068054195)*x[2]**o+(0.337405641389)*x[2]
        arg[3,1]=(0.244113319139)*x[0]**o+(0.739760310044)*x[0]+(0.439528183979)*x[1]**o+(0.808496622595)*x[1]+(0.432069187682)*x[2]**o+(-0.106394550529)*x[2]
        arg[3,2]=(-0.200973139796)*x[0]**o+(-0.373932231384)*x[0]+(-0.438946407503)*x[1]**o+(0.543079830762)*x[1]+(-0.499599010695)*x[2]**o+(0.501967674296)*x[2]
        arg[3,3]=(-0.128263000609)*x[0]**o+(-0.408181851435)*x[0]+(0.403315054177)*x[1]**o+(0.75067593555)*x[1]+(0.361607987217)*x[2]**o+(0.269483801463)*x[2]
        arg[3,4]=(0.821969994275)*x[0]**o+(0.350771424951)*x[0]+(-0.761235507511)*x[1]**o+(0.148329827876)*x[1]+(-0.918493845231)*x[2]**o+(0.592094636272)*x[2]
        ref[0,0]=(0.282507743319)*(1+2.*(dim-1.)/(o+1.))+(-1.5995120015)*dim
        ref[0,1]=(-1.15485068443)*(1+2.*(dim-1.)/(o+1.))+(0.00315405183214)*dim
        ref[0,2]=(-0.211118811675)*(1+2.*(dim-1.)/(o+1.))+(-0.0254723970291)*dim
        ref[0,3]=(0.0506409580755)*(1+2.*(dim-1.)/(o+1.))+(-1.0501502318)*dim
        ref[0,4]=(0.133115075732)*(1+2.*(dim-1.)/(o+1.))+(0.664555376358)*dim
        ref[1,0]=(-0.67058169283)*(1+2.*(dim-1.)/(o+1.))+(-0.145122066642)*dim
        ref[1,1]=(-1.15661217643)*(1+2.*(dim-1.)/(o+1.))+(-0.646670869164)*dim
        ref[1,2]=(-0.824444166116)*(1+2.*(dim-1.)/(o+1.))+(-0.330932623821)*dim
        ref[1,3]=(-0.247426321664)*(1+2.*(dim-1.)/(o+1.))+(1.18320051028)*dim
        ref[1,4]=(-0.288279999634)*(1+2.*(dim-1.)/(o+1.))+(-0.00384882668765)*dim
        ref[2,0]=(0.365061348688)*(1+2.*(dim-1.)/(o+1.))+(-1.84242418926)*dim
        ref[2,1]=(-1.33556213656)*(1+2.*(dim-1.)/(o+1.))+(-1.69125070615)*dim
        ref[2,2]=(0.325334354026)*(1+2.*(dim-1.)/(o+1.))+(1.56364605531)*dim
        ref[2,3]=(-0.935364568298)*(1+2.*(dim-1.)/(o+1.))+(0.573993263976)*dim
        ref[2,4]=(-0.833636951913)*(1+2.*(dim-1.)/(o+1.))+(0.6579135208)*dim
        ref[3,0]=(-0.804696158299)*(1+2.*(dim-1.)/(o+1.))+(1.28487651454)*dim
        ref[3,1]=(1.1157106908)*(1+2.*(dim-1.)/(o+1.))+(1.44186238211)*dim
        ref[3,2]=(-1.13951855799)*(1+2.*(dim-1.)/(o+1.))+(0.671115273674)*dim
        ref[3,3]=(0.636660040784)*(1+2.*(dim-1.)/(o+1.))+(0.611977885578)*dim
        ref[3,4]=(-0.857759358467)*(1+2.*(dim-1.)/(o+1.))+(1.0911958891)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.331954160625)*x[0]**o+(0.713423346586)*x[0]+(0.922333771405)*x[1]**o+(-0.306799323245)*x[1]
        arg[0,0,1]=(-0.837278273696)*x[0]**o+(-0.322068356249)*x[0]+(-0.952056283446)*x[1]**o+(0.78147215355)*x[1]
        arg[0,1,0]=(-0.71217197633)*x[0]**o+(-0.222121067189)*x[0]+(-0.740851323426)*x[1]**o+(-0.664415003166)*x[1]
        arg[0,1,1]=(0.304797844524)*x[0]**o+(-0.644642249615)*x[0]+(-0.942312793933)*x[1]**o+(0.758857258357)*x[1]
        arg[1,0,0]=(-0.978734634026)*x[0]**o+(0.697829375729)*x[0]+(-0.161023313194)*x[1]**o+(0.61960280308)*x[1]
        arg[1,0,1]=(0.65091302745)*x[0]**o+(-0.411592147008)*x[0]+(0.748017319704)*x[1]**o+(0.751661257345)*x[1]
        arg[1,1,0]=(-0.752749816617)*x[0]**o+(-0.889417028079)*x[0]+(0.187075769961)*x[1]**o+(0.96605423194)*x[1]
        arg[1,1,1]=(-0.791276410919)*x[0]**o+(-0.66910236639)*x[0]+(-0.107283654349)*x[1]**o+(0.556315348977)*x[1]
        arg[2,0,0]=(0.0717781211643)*x[0]**o+(-0.895974517834)*x[0]+(-0.388350799289)*x[1]**o+(0.55541171641)*x[1]
        arg[2,0,1]=(0.886475701949)*x[0]**o+(0.0794480579219)*x[0]+(-0.77671944551)*x[1]**o+(-0.29611930088)*x[1]
        arg[2,1,0]=(0.891530897911)*x[0]**o+(-0.767892439984)*x[0]+(-0.692871902002)*x[1]**o+(0.296346783431)*x[1]
        arg[2,1,1]=(0.904417221616)*x[0]**o+(-0.379208335246)*x[0]+(0.491675297059)*x[1]**o+(0.994131453216)*x[1]
        arg[3,0,0]=(-0.278750282129)*x[0]**o+(-0.889235749208)*x[0]+(-0.970025256676)*x[1]**o+(0.459265023633)*x[1]
        arg[3,0,1]=(-0.718548177793)*x[0]**o+(-0.801618785436)*x[0]+(-0.0489351007933)*x[1]**o+(-0.0331281021404)*x[1]
        arg[3,1,0]=(0.838645233036)*x[0]**o+(0.174830460073)*x[0]+(0.343887140274)*x[1]**o+(0.475643625956)*x[1]
        arg[3,1,1]=(-0.376629360434)*x[0]**o+(0.965850026388)*x[0]+(-0.62284011199)*x[1]**o+(-0.548261522438)*x[1]
        arg[4,0,0]=(-0.747109622101)*x[0]**o+(-0.793286186527)*x[0]+(0.700502366132)*x[1]**o+(0.708459489695)*x[1]
        arg[4,0,1]=(-0.372920704339)*x[0]**o+(-0.856662712404)*x[0]+(0.735191558102)*x[1]**o+(0.451959287084)*x[1]
        arg[4,1,0]=(-0.825227447648)*x[0]**o+(-0.978701927456)*x[0]+(0.856562109765)*x[1]**o+(0.226820057686)*x[1]
        arg[4,1,1]=(-0.883951567752)*x[0]**o+(0.346114896878)*x[0]+(0.22103721575)*x[1]**o+(0.766907412924)*x[1]
        arg[5,0,0]=(0.440140408264)*x[0]**o+(0.576442002932)*x[0]+(0.181136518153)*x[1]**o+(0.467290325887)*x[1]
        arg[5,0,1]=(0.323653297993)*x[0]**o+(-0.0169080219964)*x[0]+(-0.790889359324)*x[1]**o+(-0.752609210154)*x[1]
        arg[5,1,0]=(-0.771859095701)*x[0]**o+(0.392074466446)*x[0]+(0.672832544535)*x[1]**o+(-0.518196376899)*x[1]
        arg[5,1,1]=(-0.714106835785)*x[0]**o+(-0.392974811152)*x[0]+(0.477349412757)*x[1]**o+(-0.978738343032)*x[1]
        ref[0,0,0]=(0.59037961078)*(1+2.*(dim-1.)/(o+1.))+(0.406624023341)*dim
        ref[0,0,1]=(-1.78933455714)*(1+2.*(dim-1.)/(o+1.))+(0.459403797301)*dim
        ref[0,1,0]=(-1.45302329976)*(1+2.*(dim-1.)/(o+1.))+(-0.886536070355)*dim
        ref[0,1,1]=(-0.637514949409)*(1+2.*(dim-1.)/(o+1.))+(0.114215008741)*dim
        ref[1,0,0]=(-1.13975794722)*(1+2.*(dim-1.)/(o+1.))+(1.31743217881)*dim
        ref[1,0,1]=(1.39893034715)*(1+2.*(dim-1.)/(o+1.))+(0.340069110337)*dim
        ref[1,1,0]=(-0.565674046656)*(1+2.*(dim-1.)/(o+1.))+(0.076637203861)*dim
        ref[1,1,1]=(-0.898560065268)*(1+2.*(dim-1.)/(o+1.))+(-0.112787017413)*dim
        ref[2,0,0]=(-0.316572678125)*(1+2.*(dim-1.)/(o+1.))+(-0.340562801424)*dim
        ref[2,0,1]=(0.10975625644)*(1+2.*(dim-1.)/(o+1.))+(-0.216671242958)*dim
        ref[2,1,0]=(0.198658995909)*(1+2.*(dim-1.)/(o+1.))+(-0.471545656553)*dim
        ref[2,1,1]=(1.39609251868)*(1+2.*(dim-1.)/(o+1.))+(0.614923117971)*dim
        ref[3,0,0]=(-1.2487755388)*(1+2.*(dim-1.)/(o+1.))+(-0.429970725575)*dim
        ref[3,0,1]=(-0.767483278586)*(1+2.*(dim-1.)/(o+1.))+(-0.834746887577)*dim
        ref[3,1,0]=(1.18253237331)*(1+2.*(dim-1.)/(o+1.))+(0.650474086029)*dim
        ref[3,1,1]=(-0.999469472424)*(1+2.*(dim-1.)/(o+1.))+(0.41758850395)*dim
        ref[4,0,0]=(-0.0466072559697)*(1+2.*(dim-1.)/(o+1.))+(-0.0848266968319)*dim
        ref[4,0,1]=(0.362270853763)*(1+2.*(dim-1.)/(o+1.))+(-0.404703425319)*dim
        ref[4,1,0]=(0.0313346621166)*(1+2.*(dim-1.)/(o+1.))+(-0.75188186977)*dim
        ref[4,1,1]=(-0.662914352001)*(1+2.*(dim-1.)/(o+1.))+(1.1130223098)*dim
        ref[5,0,0]=(0.621276926417)*(1+2.*(dim-1.)/(o+1.))+(1.04373232882)*dim
        ref[5,0,1]=(-0.467236061331)*(1+2.*(dim-1.)/(o+1.))+(-0.76951723215)*dim
        ref[5,1,0]=(-0.0990265511653)*(1+2.*(dim-1.)/(o+1.))+(-0.126121910453)*dim
        ref[5,1,1]=(-0.236757423028)*(1+2.*(dim-1.)/(o+1.))+(-1.37171315418)*dim
      else:
        arg[0,0,0]=(-0.676850371385)*x[0]**o+(0.343194249293)*x[0]+(-0.289569238124)*x[1]**o+(0.546143025789)*x[1]+(-0.292429216951)*x[2]**o+(0.100638805907)*x[2]
        arg[0,0,1]=(-0.558403199885)*x[0]**o+(0.369265961569)*x[0]+(0.098539502761)*x[1]**o+(-0.555747072012)*x[1]+(0.0142505043744)*x[2]**o+(0.110677590143)*x[2]
        arg[0,1,0]=(0.480265324249)*x[0]**o+(-0.0801799981418)*x[0]+(-0.659040110213)*x[1]**o+(0.315685251032)*x[1]+(0.146222780181)*x[2]**o+(-0.307845817615)*x[2]
        arg[0,1,1]=(-0.469643195547)*x[0]**o+(0.631180947887)*x[0]+(0.0873749097691)*x[1]**o+(-0.907805635326)*x[1]+(-0.445238378297)*x[2]**o+(-0.583750385876)*x[2]
        arg[1,0,0]=(-0.519648347947)*x[0]**o+(-0.580046561107)*x[0]+(-0.474702629402)*x[1]**o+(-0.76698196707)*x[1]+(0.631290211156)*x[2]**o+(0.283150656338)*x[2]
        arg[1,0,1]=(0.828049802595)*x[0]**o+(-0.0996271417739)*x[0]+(0.13247948046)*x[1]**o+(0.64657332105)*x[1]+(0.901448888477)*x[2]**o+(0.895379597076)*x[2]
        arg[1,1,0]=(-0.197367099356)*x[0]**o+(-0.933210661759)*x[0]+(-0.0134349749548)*x[1]**o+(-0.651111981269)*x[1]+(0.256945439519)*x[2]**o+(0.528412689094)*x[2]
        arg[1,1,1]=(-0.922943172982)*x[0]**o+(-0.883102479542)*x[0]+(-0.0037345936911)*x[1]**o+(-0.598256159261)*x[1]+(-0.8562089409)*x[2]**o+(-0.298502617143)*x[2]
        arg[2,0,0]=(0.499285053237)*x[0]**o+(-0.347877602065)*x[0]+(0.783117300662)*x[1]**o+(0.0413016895448)*x[1]+(0.22555777414)*x[2]**o+(0.089246166469)*x[2]
        arg[2,0,1]=(-0.180602931202)*x[0]**o+(0.883958054663)*x[0]+(0.904974519387)*x[1]**o+(-0.169618307434)*x[1]+(0.361963890467)*x[2]**o+(-0.364236450353)*x[2]
        arg[2,1,0]=(0.332239090882)*x[0]**o+(-0.392239851855)*x[0]+(0.905142609029)*x[1]**o+(-0.471493540964)*x[1]+(-0.305327788141)*x[2]**o+(0.581213078834)*x[2]
        arg[2,1,1]=(0.481276189253)*x[0]**o+(0.645400809127)*x[0]+(-0.840844470697)*x[1]**o+(-0.47587396865)*x[1]+(0.134145050668)*x[2]**o+(-0.315081369951)*x[2]
        arg[3,0,0]=(-0.480497217552)*x[0]**o+(0.595345393671)*x[0]+(0.776315451392)*x[1]**o+(-0.396110510671)*x[1]+(0.0437422791996)*x[2]**o+(0.348675326203)*x[2]
        arg[3,0,1]=(0.867117243092)*x[0]**o+(-0.222483151231)*x[0]+(0.0856439550139)*x[1]**o+(0.355855801395)*x[1]+(-0.779380984653)*x[2]**o+(0.799440639884)*x[2]
        arg[3,1,0]=(-0.869234399353)*x[0]**o+(-0.747500724089)*x[0]+(-0.83154206968)*x[1]**o+(-0.771906212766)*x[1]+(-0.188734869619)*x[2]**o+(0.7312382704)*x[2]
        arg[3,1,1]=(-0.729677636148)*x[0]**o+(-0.358066533942)*x[0]+(-0.342651052704)*x[1]**o+(-0.531453801913)*x[1]+(-0.898300547827)*x[2]**o+(0.0387220517653)*x[2]
        arg[4,0,0]=(-0.380489041999)*x[0]**o+(0.517823660835)*x[0]+(0.722149782685)*x[1]**o+(0.510169845258)*x[1]+(-0.979340093133)*x[2]**o+(0.846050053391)*x[2]
        arg[4,0,1]=(0.390895580804)*x[0]**o+(0.122461788897)*x[0]+(-0.809243912452)*x[1]**o+(0.871689879803)*x[1]+(-0.991689466458)*x[2]**o+(-0.0373953689911)*x[2]
        arg[4,1,0]=(0.00670611033858)*x[0]**o+(-0.700084486793)*x[0]+(0.365358002826)*x[1]**o+(-0.928347441251)*x[1]+(-0.675081280461)*x[2]**o+(-0.262871553902)*x[2]
        arg[4,1,1]=(-0.566521655794)*x[0]**o+(-0.23325381374)*x[0]+(0.744538505444)*x[1]**o+(0.263970013855)*x[1]+(0.360984907343)*x[2]**o+(0.704696800059)*x[2]
        arg[5,0,0]=(0.754229123633)*x[0]**o+(-0.523353302691)*x[0]+(0.404181763418)*x[1]**o+(0.175414644029)*x[1]+(0.627455124887)*x[2]**o+(-0.450163514418)*x[2]
        arg[5,0,1]=(0.804547370453)*x[0]**o+(-0.748034210243)*x[0]+(-0.943261499953)*x[1]**o+(0.360715335834)*x[1]+(-0.952785406204)*x[2]**o+(0.188838771539)*x[2]
        arg[5,1,0]=(0.653730390007)*x[0]**o+(0.663746229033)*x[0]+(0.454911944273)*x[1]**o+(-0.548110740107)*x[1]+(0.799171265206)*x[2]**o+(-0.646356900812)*x[2]
        arg[5,1,1]=(0.533464266683)*x[0]**o+(-0.0650985333228)*x[0]+(-0.599880569232)*x[1]**o+(0.645120695827)*x[1]+(0.985316456657)*x[2]**o+(0.527848503917)*x[2]
        ref[0,0,0]=(-1.25884882646)*(1+2.*(dim-1.)/(o+1.))+(0.989976080988)*dim
        ref[0,0,1]=(-0.445613192749)*(1+2.*(dim-1.)/(o+1.))+(-0.0758035202994)*dim
        ref[0,1,0]=(-0.0325520057832)*(1+2.*(dim-1.)/(o+1.))+(-0.0723405647253)*dim
        ref[0,1,1]=(-0.827506664075)*(1+2.*(dim-1.)/(o+1.))+(-0.860375073315)*dim
        ref[1,0,0]=(-0.363060766193)*(1+2.*(dim-1.)/(o+1.))+(-1.06387787184)*dim
        ref[1,0,1]=(1.86197817153)*(1+2.*(dim-1.)/(o+1.))+(1.44232577635)*dim
        ref[1,1,0]=(0.0461433652081)*(1+2.*(dim-1.)/(o+1.))+(-1.05590995393)*dim
        ref[1,1,1]=(-1.78288670757)*(1+2.*(dim-1.)/(o+1.))+(-1.77986125595)*dim
        ref[2,0,0]=(1.50796012804)*(1+2.*(dim-1.)/(o+1.))+(-0.217329746051)*dim
        ref[2,0,1]=(1.08633547865)*(1+2.*(dim-1.)/(o+1.))+(0.350103296876)*dim
        ref[2,1,0]=(0.932053911771)*(1+2.*(dim-1.)/(o+1.))+(-0.282520313985)*dim
        ref[2,1,1]=(-0.225423230776)*(1+2.*(dim-1.)/(o+1.))+(-0.145554529474)*dim
        ref[3,0,0]=(0.33956051304)*(1+2.*(dim-1.)/(o+1.))+(0.547910209204)*dim
        ref[3,0,1]=(0.173380213454)*(1+2.*(dim-1.)/(o+1.))+(0.932813290049)*dim
        ref[3,1,0]=(-1.88951133865)*(1+2.*(dim-1.)/(o+1.))+(-0.788168666455)*dim
        ref[3,1,1]=(-1.97062923668)*(1+2.*(dim-1.)/(o+1.))+(-0.850798284089)*dim
        ref[4,0,0]=(-0.637679352447)*(1+2.*(dim-1.)/(o+1.))+(1.87404355948)*dim
        ref[4,0,1]=(-1.41003779811)*(1+2.*(dim-1.)/(o+1.))+(0.956756299709)*dim
        ref[4,1,0]=(-0.303017167297)*(1+2.*(dim-1.)/(o+1.))+(-1.89130348195)*dim
        ref[4,1,1]=(0.539001756993)*(1+2.*(dim-1.)/(o+1.))+(0.735413000173)*dim
        ref[5,0,0]=(1.78586601194)*(1+2.*(dim-1.)/(o+1.))+(-0.79810217308)*dim
        ref[5,0,1]=(-1.0914995357)*(1+2.*(dim-1.)/(o+1.))+(-0.198480102871)*dim
        ref[5,1,0]=(1.90781359949)*(1+2.*(dim-1.)/(o+1.))+(-0.530721411887)*dim
        ref[5,1,1]=(0.918900154108)*(1+2.*(dim-1.)/(o+1.))+(1.10787066642)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.181966612335)*x[0]**o+(-0.461883712153)*x[0]+(-0.420217253531)*x[1]**o+(0.52197109374)*x[1]
        arg[0,0,0,1]=(-0.122704378373)*x[0]**o+(0.549708706786)*x[0]+(-0.728108908074)*x[1]**o+(-0.839262748545)*x[1]
        arg[0,0,1,0]=(0.58557741427)*x[0]**o+(-0.970633305521)*x[0]+(0.215513381389)*x[1]**o+(-0.178877868196)*x[1]
        arg[0,0,1,1]=(0.646118894595)*x[0]**o+(0.839917605008)*x[0]+(-0.608843960266)*x[1]**o+(-0.840842208189)*x[1]
        arg[0,0,2,0]=(-0.126004594384)*x[0]**o+(0.663487355648)*x[0]+(0.801751176464)*x[1]**o+(-0.76620842256)*x[1]
        arg[0,0,2,1]=(-0.244425023768)*x[0]**o+(0.736809513316)*x[0]+(0.736776514098)*x[1]**o+(0.382648973627)*x[1]
        arg[0,1,0,0]=(-0.264460763052)*x[0]**o+(-0.325868872518)*x[0]+(-0.0636203008538)*x[1]**o+(-0.398059740187)*x[1]
        arg[0,1,0,1]=(0.443639751435)*x[0]**o+(0.523903519493)*x[0]+(-0.222575248757)*x[1]**o+(-0.168203019401)*x[1]
        arg[0,1,1,0]=(-0.329523829258)*x[0]**o+(-0.908408437436)*x[0]+(0.695441587034)*x[1]**o+(-0.236957609006)*x[1]
        arg[0,1,1,1]=(-0.57363301023)*x[0]**o+(-0.168677045141)*x[0]+(0.455235475098)*x[1]**o+(-0.498803217861)*x[1]
        arg[0,1,2,0]=(0.09906404243)*x[0]**o+(0.48682122702)*x[0]+(-0.250940180366)*x[1]**o+(-0.641013418008)*x[1]
        arg[0,1,2,1]=(-0.963193555161)*x[0]**o+(0.129868691037)*x[0]+(0.821192174481)*x[1]**o+(-0.245576703223)*x[1]
        arg[0,2,0,0]=(-0.867410251576)*x[0]**o+(0.170194043006)*x[0]+(0.837669366536)*x[1]**o+(0.285772890787)*x[1]
        arg[0,2,0,1]=(0.767089114154)*x[0]**o+(-0.425174482112)*x[0]+(-0.460840645159)*x[1]**o+(-0.314295694392)*x[1]
        arg[0,2,1,0]=(-0.932934920269)*x[0]**o+(0.399552332498)*x[0]+(-0.137764102275)*x[1]**o+(-0.0375128952657)*x[1]
        arg[0,2,1,1]=(0.507999826854)*x[0]**o+(-0.60826518686)*x[0]+(0.766838061511)*x[1]**o+(0.800723296773)*x[1]
        arg[0,2,2,0]=(0.372064864992)*x[0]**o+(0.29029476011)*x[0]+(0.203299629846)*x[1]**o+(-0.386357855884)*x[1]
        arg[0,2,2,1]=(0.661268311956)*x[0]**o+(0.60257115577)*x[0]+(0.88881711536)*x[1]**o+(0.335431084294)*x[1]
        arg[0,3,0,0]=(-0.756295024079)*x[0]**o+(-0.982472476783)*x[0]+(-0.264060825724)*x[1]**o+(0.082684535426)*x[1]
        arg[0,3,0,1]=(0.514580120205)*x[0]**o+(0.524816925461)*x[0]+(-0.745096223094)*x[1]**o+(-0.0796881815577)*x[1]
        arg[0,3,1,0]=(0.597556449974)*x[0]**o+(0.527680495979)*x[0]+(0.517638167762)*x[1]**o+(-0.357836534374)*x[1]
        arg[0,3,1,1]=(0.354366560101)*x[0]**o+(-0.617697325436)*x[0]+(0.339071431286)*x[1]**o+(0.299934132319)*x[1]
        arg[0,3,2,0]=(0.26563455123)*x[0]**o+(0.821495614519)*x[0]+(0.177496963332)*x[1]**o+(-0.154521746257)*x[1]
        arg[0,3,2,1]=(0.983477388632)*x[0]**o+(0.524806318175)*x[0]+(-0.604096007297)*x[1]**o+(0.350360164328)*x[1]
        arg[0,4,0,0]=(0.311369778768)*x[0]**o+(-0.648416457513)*x[0]+(0.463469290711)*x[1]**o+(0.0719628828709)*x[1]
        arg[0,4,0,1]=(-0.743081056195)*x[0]**o+(-0.803573973461)*x[0]+(-0.766982068691)*x[1]**o+(0.387240020398)*x[1]
        arg[0,4,1,0]=(-0.0715117411345)*x[0]**o+(-0.341871715212)*x[0]+(0.0593378446291)*x[1]**o+(-0.10183864979)*x[1]
        arg[0,4,1,1]=(0.538850329978)*x[0]**o+(0.0280201497914)*x[0]+(0.350206859589)*x[1]**o+(0.316831165676)*x[1]
        arg[0,4,2,0]=(-0.327852385434)*x[0]**o+(0.408866306659)*x[0]+(0.940378517318)*x[1]**o+(0.578309623908)*x[1]
        arg[0,4,2,1]=(0.378380112483)*x[0]**o+(0.696828716067)*x[0]+(-0.908233000582)*x[1]**o+(-0.328611483675)*x[1]
        arg[1,0,0,0]=(0.753991558436)*x[0]**o+(0.0542891140783)*x[0]+(-0.428934638249)*x[1]**o+(0.318155063446)*x[1]
        arg[1,0,0,1]=(-0.387609085343)*x[0]**o+(0.711023478158)*x[0]+(-0.969026331169)*x[1]**o+(0.456175434422)*x[1]
        arg[1,0,1,0]=(-0.468504469917)*x[0]**o+(-0.314756281356)*x[0]+(0.0145647044588)*x[1]**o+(0.074035761055)*x[1]
        arg[1,0,1,1]=(0.240749620446)*x[0]**o+(0.906736060537)*x[0]+(0.391320123368)*x[1]**o+(-0.261639387239)*x[1]
        arg[1,0,2,0]=(-0.461436333094)*x[0]**o+(-0.809258428817)*x[0]+(0.0670777162266)*x[1]**o+(-0.445304287721)*x[1]
        arg[1,0,2,1]=(-0.569794448172)*x[0]**o+(-0.0624052432724)*x[0]+(-0.759840810858)*x[1]**o+(-0.609106375016)*x[1]
        arg[1,1,0,0]=(0.751570143287)*x[0]**o+(0.78189919599)*x[0]+(-0.103246458909)*x[1]**o+(0.732156635312)*x[1]
        arg[1,1,0,1]=(-0.0126214545951)*x[0]**o+(-0.612853966263)*x[0]+(-0.238555698245)*x[1]**o+(-0.668009933698)*x[1]
        arg[1,1,1,0]=(0.816325838743)*x[0]**o+(0.545518357738)*x[0]+(0.893296287925)*x[1]**o+(0.694579919864)*x[1]
        arg[1,1,1,1]=(0.927638834795)*x[0]**o+(0.750347688466)*x[0]+(-0.15924462002)*x[1]**o+(-0.952995938892)*x[1]
        arg[1,1,2,0]=(0.970761253678)*x[0]**o+(0.205308852313)*x[0]+(0.909209906751)*x[1]**o+(-0.178635180174)*x[1]
        arg[1,1,2,1]=(-0.632065128479)*x[0]**o+(-0.300905398347)*x[0]+(0.865317638002)*x[1]**o+(0.886398250669)*x[1]
        arg[1,2,0,0]=(-0.0529293620539)*x[0]**o+(-0.909022003179)*x[0]+(0.235231916345)*x[1]**o+(-0.623380752415)*x[1]
        arg[1,2,0,1]=(0.873344987616)*x[0]**o+(0.144231546529)*x[0]+(0.307375566381)*x[1]**o+(-0.201617507084)*x[1]
        arg[1,2,1,0]=(-0.957945047658)*x[0]**o+(-0.432955474381)*x[0]+(0.0742288164869)*x[1]**o+(-0.106080430366)*x[1]
        arg[1,2,1,1]=(-0.664227561512)*x[0]**o+(0.0821007781774)*x[0]+(-0.541506076231)*x[1]**o+(-0.88307957937)*x[1]
        arg[1,2,2,0]=(0.754702675232)*x[0]**o+(0.175176916913)*x[0]+(0.519607893979)*x[1]**o+(-0.24129526204)*x[1]
        arg[1,2,2,1]=(-0.415660998528)*x[0]**o+(-0.964071393069)*x[0]+(0.458656303526)*x[1]**o+(0.752153826846)*x[1]
        arg[1,3,0,0]=(-0.15638393187)*x[0]**o+(0.0828824719)*x[0]+(-0.0846205011509)*x[1]**o+(-0.438939463736)*x[1]
        arg[1,3,0,1]=(0.222994482074)*x[0]**o+(-0.488426101986)*x[0]+(-0.799262651825)*x[1]**o+(0.0638880385288)*x[1]
        arg[1,3,1,0]=(0.690529298963)*x[0]**o+(-0.845248154386)*x[0]+(0.748635110096)*x[1]**o+(0.820882250869)*x[1]
        arg[1,3,1,1]=(-0.0989162396328)*x[0]**o+(-0.171252734498)*x[0]+(-0.363447812361)*x[1]**o+(-0.06937701403)*x[1]
        arg[1,3,2,0]=(-0.739647298124)*x[0]**o+(-0.83384135572)*x[0]+(0.546124768813)*x[1]**o+(0.415761104064)*x[1]
        arg[1,3,2,1]=(0.859487140106)*x[0]**o+(0.481214854724)*x[0]+(0.788685241166)*x[1]**o+(-0.776410333263)*x[1]
        arg[1,4,0,0]=(-0.312428038681)*x[0]**o+(-0.178596293803)*x[0]+(-0.229770546807)*x[1]**o+(-0.43765633326)*x[1]
        arg[1,4,0,1]=(-0.974305286188)*x[0]**o+(-0.978154435248)*x[0]+(-0.671799350916)*x[1]**o+(0.645287291694)*x[1]
        arg[1,4,1,0]=(-0.219711371151)*x[0]**o+(-0.432784911532)*x[0]+(0.552940388883)*x[1]**o+(-0.622742205614)*x[1]
        arg[1,4,1,1]=(0.677289749287)*x[0]**o+(-0.92825132284)*x[0]+(0.818767358259)*x[1]**o+(-0.426000012305)*x[1]
        arg[1,4,2,0]=(-0.507474696573)*x[0]**o+(0.0882151625837)*x[0]+(0.826611187078)*x[1]**o+(0.157884229299)*x[1]
        arg[1,4,2,1]=(0.902398378671)*x[0]**o+(-0.372679984773)*x[0]+(-0.998272135687)*x[1]**o+(0.292294917465)*x[1]
        arg[2,0,0,0]=(0.699035433674)*x[0]**o+(0.380640017665)*x[0]+(-0.650735771214)*x[1]**o+(-0.55187467604)*x[1]
        arg[2,0,0,1]=(0.703438074924)*x[0]**o+(0.795508285134)*x[0]+(-0.0650847718675)*x[1]**o+(-0.0250766383001)*x[1]
        arg[2,0,1,0]=(0.493174434488)*x[0]**o+(-0.775523541313)*x[0]+(0.23184619496)*x[1]**o+(0.343368328234)*x[1]
        arg[2,0,1,1]=(0.404974280918)*x[0]**o+(0.0673627086244)*x[0]+(-0.843957699955)*x[1]**o+(0.645895713789)*x[1]
        arg[2,0,2,0]=(0.985047143685)*x[0]**o+(0.969582300743)*x[0]+(0.929427858098)*x[1]**o+(-0.951787413112)*x[1]
        arg[2,0,2,1]=(0.618505056215)*x[0]**o+(0.685760720523)*x[0]+(0.14594727158)*x[1]**o+(0.816579780815)*x[1]
        arg[2,1,0,0]=(0.0824953418089)*x[0]**o+(-0.896753351387)*x[0]+(0.415841596602)*x[1]**o+(0.349436600222)*x[1]
        arg[2,1,0,1]=(-0.904246301273)*x[0]**o+(0.51403543319)*x[0]+(0.0950147832662)*x[1]**o+(0.0482220256459)*x[1]
        arg[2,1,1,0]=(0.653418547332)*x[0]**o+(0.703094950575)*x[0]+(-0.294924381355)*x[1]**o+(0.889408306649)*x[1]
        arg[2,1,1,1]=(-0.329361050221)*x[0]**o+(-0.93072021487)*x[0]+(0.0845194032793)*x[1]**o+(-0.310617126183)*x[1]
        arg[2,1,2,0]=(-0.282371473471)*x[0]**o+(0.387458491178)*x[0]+(0.643908016757)*x[1]**o+(0.303741991451)*x[1]
        arg[2,1,2,1]=(-0.416872502153)*x[0]**o+(0.373465764073)*x[0]+(-0.90689140487)*x[1]**o+(-0.351052228378)*x[1]
        arg[2,2,0,0]=(0.494198770026)*x[0]**o+(0.744917026846)*x[0]+(-0.806591346974)*x[1]**o+(-0.812396479226)*x[1]
        arg[2,2,0,1]=(0.230573331783)*x[0]**o+(-0.449815323831)*x[0]+(-0.662178492084)*x[1]**o+(0.0281786827584)*x[1]
        arg[2,2,1,0]=(-0.546296599764)*x[0]**o+(-0.457939754236)*x[0]+(-0.0432225723667)*x[1]**o+(0.0431706529852)*x[1]
        arg[2,2,1,1]=(-0.808174460895)*x[0]**o+(-0.473584203382)*x[0]+(0.567514752294)*x[1]**o+(-0.115780229631)*x[1]
        arg[2,2,2,0]=(-0.134425008633)*x[0]**o+(-0.103550646345)*x[0]+(-0.588484951305)*x[1]**o+(0.955148281194)*x[1]
        arg[2,2,2,1]=(0.550095417437)*x[0]**o+(0.943564321046)*x[0]+(-0.990366145789)*x[1]**o+(0.872652302953)*x[1]
        arg[2,3,0,0]=(0.209341033615)*x[0]**o+(-0.285197252362)*x[0]+(0.961111991429)*x[1]**o+(0.810022268034)*x[1]
        arg[2,3,0,1]=(0.238319839582)*x[0]**o+(-0.714990515235)*x[0]+(0.0925371590198)*x[1]**o+(0.938145796375)*x[1]
        arg[2,3,1,0]=(0.889068347007)*x[0]**o+(-0.301904392078)*x[0]+(0.41386941601)*x[1]**o+(-0.973115118701)*x[1]
        arg[2,3,1,1]=(-0.522083695846)*x[0]**o+(-0.582455208417)*x[0]+(0.725431346859)*x[1]**o+(-0.950157996033)*x[1]
        arg[2,3,2,0]=(-0.897838142917)*x[0]**o+(0.745488698358)*x[0]+(-0.38910727894)*x[1]**o+(0.614539961316)*x[1]
        arg[2,3,2,1]=(0.924500645496)*x[0]**o+(-0.68988446876)*x[0]+(-0.466238481805)*x[1]**o+(-0.957703006413)*x[1]
        arg[2,4,0,0]=(0.280692018949)*x[0]**o+(0.600320489083)*x[0]+(-0.402390231845)*x[1]**o+(0.906721159456)*x[1]
        arg[2,4,0,1]=(0.50594574787)*x[0]**o+(0.130986910768)*x[0]+(0.272643351394)*x[1]**o+(0.53790545383)*x[1]
        arg[2,4,1,0]=(-0.627308461808)*x[0]**o+(0.690560028115)*x[0]+(-0.455768647207)*x[1]**o+(0.134926199015)*x[1]
        arg[2,4,1,1]=(-0.501163279392)*x[0]**o+(0.980144824993)*x[0]+(-0.0358930723173)*x[1]**o+(0.720157452357)*x[1]
        arg[2,4,2,0]=(-0.609861117539)*x[0]**o+(0.965699461849)*x[0]+(0.305674136837)*x[1]**o+(-0.578333683729)*x[1]
        arg[2,4,2,1]=(-0.921051416627)*x[0]**o+(-0.703479746028)*x[0]+(-0.0189027931731)*x[1]**o+(0.403415497212)*x[1]
        arg[3,0,0,0]=(-0.846266714886)*x[0]**o+(-0.103621657972)*x[0]+(0.217340457606)*x[1]**o+(-0.537654290693)*x[1]
        arg[3,0,0,1]=(0.186251745403)*x[0]**o+(0.694458841658)*x[0]+(0.126837362883)*x[1]**o+(0.0997502401792)*x[1]
        arg[3,0,1,0]=(0.153737843143)*x[0]**o+(0.811590040899)*x[0]+(0.189363446699)*x[1]**o+(0.00688399857093)*x[1]
        arg[3,0,1,1]=(-0.912444421339)*x[0]**o+(-0.68504171198)*x[0]+(-0.820717153919)*x[1]**o+(-0.00475951426309)*x[1]
        arg[3,0,2,0]=(0.908522439455)*x[0]**o+(-0.911126936858)*x[0]+(0.0424198145962)*x[1]**o+(-0.352097158128)*x[1]
        arg[3,0,2,1]=(0.935127271967)*x[0]**o+(0.154700792222)*x[0]+(-0.283148208173)*x[1]**o+(-0.607541482501)*x[1]
        arg[3,1,0,0]=(-0.131329160543)*x[0]**o+(0.24170180662)*x[0]+(-0.00893591013587)*x[1]**o+(0.262635048958)*x[1]
        arg[3,1,0,1]=(-0.559200924561)*x[0]**o+(0.509310116425)*x[0]+(0.484486720282)*x[1]**o+(-0.89314080386)*x[1]
        arg[3,1,1,0]=(0.355094442114)*x[0]**o+(-0.748398915688)*x[0]+(-0.379187476624)*x[1]**o+(0.447409858108)*x[1]
        arg[3,1,1,1]=(0.170855109103)*x[0]**o+(-0.48925054508)*x[0]+(0.840559727559)*x[1]**o+(0.0359562471889)*x[1]
        arg[3,1,2,0]=(0.827586094521)*x[0]**o+(0.0231820503547)*x[0]+(0.457192172855)*x[1]**o+(0.826960230377)*x[1]
        arg[3,1,2,1]=(-0.962890676345)*x[0]**o+(-0.112439391098)*x[0]+(-0.865510686025)*x[1]**o+(-0.539059426684)*x[1]
        arg[3,2,0,0]=(-0.165030011664)*x[0]**o+(-0.285694090419)*x[0]+(-0.522579583443)*x[1]**o+(0.73946640768)*x[1]
        arg[3,2,0,1]=(0.809351962503)*x[0]**o+(-0.872047628102)*x[0]+(0.61415433255)*x[1]**o+(0.132268088986)*x[1]
        arg[3,2,1,0]=(-0.587823544835)*x[0]**o+(-0.416092975475)*x[0]+(0.829027070431)*x[1]**o+(0.805271620727)*x[1]
        arg[3,2,1,1]=(-0.0107935093527)*x[0]**o+(0.988434287979)*x[0]+(0.941743055822)*x[1]**o+(0.137807879334)*x[1]
        arg[3,2,2,0]=(-0.855931527408)*x[0]**o+(0.575075608849)*x[0]+(0.11916967111)*x[1]**o+(-0.472844336533)*x[1]
        arg[3,2,2,1]=(-0.542112514875)*x[0]**o+(0.081308081083)*x[0]+(-0.934575648869)*x[1]**o+(-0.218208369535)*x[1]
        arg[3,3,0,0]=(0.486635391195)*x[0]**o+(0.995728500002)*x[0]+(-0.686767180456)*x[1]**o+(0.246849483855)*x[1]
        arg[3,3,0,1]=(-0.331571848212)*x[0]**o+(0.801870431206)*x[0]+(-0.91142654055)*x[1]**o+(0.089854528457)*x[1]
        arg[3,3,1,0]=(-0.657455652936)*x[0]**o+(-0.721466128651)*x[0]+(0.0420989631458)*x[1]**o+(-0.982316453325)*x[1]
        arg[3,3,1,1]=(-0.709257565086)*x[0]**o+(0.689028730506)*x[0]+(-0.196516960475)*x[1]**o+(-0.216704841188)*x[1]
        arg[3,3,2,0]=(-0.910547650394)*x[0]**o+(-0.977471949987)*x[0]+(0.37319115353)*x[1]**o+(-0.248978955796)*x[1]
        arg[3,3,2,1]=(-0.71025804576)*x[0]**o+(0.373559860529)*x[0]+(-0.9064074733)*x[1]**o+(-0.542012058073)*x[1]
        arg[3,4,0,0]=(-0.649273580176)*x[0]**o+(0.525784948768)*x[0]+(-0.213929862516)*x[1]**o+(-0.162295172275)*x[1]
        arg[3,4,0,1]=(0.209784328978)*x[0]**o+(-0.972601183002)*x[0]+(-0.730886568265)*x[1]**o+(0.966415518759)*x[1]
        arg[3,4,1,0]=(-0.334060094293)*x[0]**o+(0.29371522739)*x[0]+(0.936156150108)*x[1]**o+(-0.257026769476)*x[1]
        arg[3,4,1,1]=(-0.0465451767064)*x[0]**o+(0.115525862669)*x[0]+(0.891280073228)*x[1]**o+(-0.0436233989979)*x[1]
        arg[3,4,2,0]=(-0.748697989123)*x[0]**o+(0.246672415445)*x[0]+(0.62642302161)*x[1]**o+(-0.411644161038)*x[1]
        arg[3,4,2,1]=(0.850132316298)*x[0]**o+(0.939463534845)*x[0]+(0.0437756618735)*x[1]**o+(-0.955198146074)*x[1]
        ref[0,0,0,0]=(-0.238250641196)*(1+2.*(dim-1.)/(o+1.))+(0.0600873815876)*dim
        ref[0,0,0,1]=(-0.850813286447)*(1+2.*(dim-1.)/(o+1.))+(-0.28955404176)*dim
        ref[0,0,1,0]=(0.801090795659)*(1+2.*(dim-1.)/(o+1.))+(-1.14951117372)*dim
        ref[0,0,1,1]=(0.0372749343294)*(1+2.*(dim-1.)/(o+1.))+(-0.000924603180129)*dim
        ref[0,0,2,0]=(0.675746582079)*(1+2.*(dim-1.)/(o+1.))+(-0.102721066912)*dim
        ref[0,0,2,1]=(0.49235149033)*(1+2.*(dim-1.)/(o+1.))+(1.11945848694)*dim
        ref[0,1,0,0]=(-0.328081063906)*(1+2.*(dim-1.)/(o+1.))+(-0.723928612706)*dim
        ref[0,1,0,1]=(0.221064502678)*(1+2.*(dim-1.)/(o+1.))+(0.355700500092)*dim
        ref[0,1,1,0]=(0.365917757777)*(1+2.*(dim-1.)/(o+1.))+(-1.14536604644)*dim
        ref[0,1,1,1]=(-0.118397535132)*(1+2.*(dim-1.)/(o+1.))+(-0.667480263002)*dim
        ref[0,1,2,0]=(-0.151876137936)*(1+2.*(dim-1.)/(o+1.))+(-0.154192190988)*dim
        ref[0,1,2,1]=(-0.14200138068)*(1+2.*(dim-1.)/(o+1.))+(-0.115708012186)*dim
        ref[0,2,0,0]=(-0.0297408850402)*(1+2.*(dim-1.)/(o+1.))+(0.455966933793)*dim
        ref[0,2,0,1]=(0.306248468995)*(1+2.*(dim-1.)/(o+1.))+(-0.739470176504)*dim
        ref[0,2,1,0]=(-1.07069902254)*(1+2.*(dim-1.)/(o+1.))+(0.362039437233)*dim
        ref[0,2,1,1]=(1.27483788836)*(1+2.*(dim-1.)/(o+1.))+(0.192458109913)*dim
        ref[0,2,2,0]=(0.575364494839)*(1+2.*(dim-1.)/(o+1.))+(-0.0960630957749)*dim
        ref[0,2,2,1]=(1.55008542732)*(1+2.*(dim-1.)/(o+1.))+(0.938002240064)*dim
        ref[0,3,0,0]=(-1.0203558498)*(1+2.*(dim-1.)/(o+1.))+(-0.899787941357)*dim
        ref[0,3,0,1]=(-0.230516102889)*(1+2.*(dim-1.)/(o+1.))+(0.445128743903)*dim
        ref[0,3,1,0]=(1.11519461774)*(1+2.*(dim-1.)/(o+1.))+(0.169843961604)*dim
        ref[0,3,1,1]=(0.693437991388)*(1+2.*(dim-1.)/(o+1.))+(-0.317763193117)*dim
        ref[0,3,2,0]=(0.443131514562)*(1+2.*(dim-1.)/(o+1.))+(0.666973868262)*dim
        ref[0,3,2,1]=(0.379381381335)*(1+2.*(dim-1.)/(o+1.))+(0.875166482503)*dim
        ref[0,4,0,0]=(0.774839069479)*(1+2.*(dim-1.)/(o+1.))+(-0.576453574642)*dim
        ref[0,4,0,1]=(-1.51006312489)*(1+2.*(dim-1.)/(o+1.))+(-0.416333953063)*dim
        ref[0,4,1,0]=(-0.0121738965054)*(1+2.*(dim-1.)/(o+1.))+(-0.443710365002)*dim
        ref[0,4,1,1]=(0.889057189568)*(1+2.*(dim-1.)/(o+1.))+(0.344851315468)*dim
        ref[0,4,2,0]=(0.612526131884)*(1+2.*(dim-1.)/(o+1.))+(0.987175930567)*dim
        ref[0,4,2,1]=(-0.529852888099)*(1+2.*(dim-1.)/(o+1.))+(0.368217232392)*dim
        ref[1,0,0,0]=(0.325056920187)*(1+2.*(dim-1.)/(o+1.))+(0.372444177524)*dim
        ref[1,0,0,1]=(-1.35663541651)*(1+2.*(dim-1.)/(o+1.))+(1.16719891258)*dim
        ref[1,0,1,0]=(-0.453939765458)*(1+2.*(dim-1.)/(o+1.))+(-0.240720520301)*dim
        ref[1,0,1,1]=(0.632069743814)*(1+2.*(dim-1.)/(o+1.))+(0.645096673298)*dim
        ref[1,0,2,0]=(-0.394358616867)*(1+2.*(dim-1.)/(o+1.))+(-1.25456271654)*dim
        ref[1,0,2,1]=(-1.32963525903)*(1+2.*(dim-1.)/(o+1.))+(-0.671511618288)*dim
        ref[1,1,0,0]=(0.648323684378)*(1+2.*(dim-1.)/(o+1.))+(1.5140558313)*dim
        ref[1,1,0,1]=(-0.251177152841)*(1+2.*(dim-1.)/(o+1.))+(-1.28086389996)*dim
        ref[1,1,1,0]=(1.70962212667)*(1+2.*(dim-1.)/(o+1.))+(1.2400982776)*dim
        ref[1,1,1,1]=(0.768394214775)*(1+2.*(dim-1.)/(o+1.))+(-0.202648250425)*dim
        ref[1,1,2,0]=(1.87997116043)*(1+2.*(dim-1.)/(o+1.))+(0.0266736721392)*dim
        ref[1,1,2,1]=(0.233252509524)*(1+2.*(dim-1.)/(o+1.))+(0.585492852322)*dim
        ref[1,2,0,0]=(0.182302554291)*(1+2.*(dim-1.)/(o+1.))+(-1.53240275559)*dim
        ref[1,2,0,1]=(1.180720554)*(1+2.*(dim-1.)/(o+1.))+(-0.0573859605549)*dim
        ref[1,2,1,0]=(-0.883716231171)*(1+2.*(dim-1.)/(o+1.))+(-0.539035904746)*dim
        ref[1,2,1,1]=(-1.20573363774)*(1+2.*(dim-1.)/(o+1.))+(-0.800978801193)*dim
        ref[1,2,2,0]=(1.27431056921)*(1+2.*(dim-1.)/(o+1.))+(-0.066118345127)*dim
        ref[1,2,2,1]=(0.0429953049979)*(1+2.*(dim-1.)/(o+1.))+(-0.211917566223)*dim
        ref[1,3,0,0]=(-0.241004433021)*(1+2.*(dim-1.)/(o+1.))+(-0.356056991836)*dim
        ref[1,3,0,1]=(-0.576268169751)*(1+2.*(dim-1.)/(o+1.))+(-0.424538063457)*dim
        ref[1,3,1,0]=(1.43916440906)*(1+2.*(dim-1.)/(o+1.))+(-0.0243659035174)*dim
        ref[1,3,1,1]=(-0.462364051993)*(1+2.*(dim-1.)/(o+1.))+(-0.240629748528)*dim
        ref[1,3,2,0]=(-0.193522529311)*(1+2.*(dim-1.)/(o+1.))+(-0.418080251656)*dim
        ref[1,3,2,1]=(1.64817238127)*(1+2.*(dim-1.)/(o+1.))+(-0.295195478539)*dim
        ref[1,4,0,0]=(-0.542198585489)*(1+2.*(dim-1.)/(o+1.))+(-0.616252627062)*dim
        ref[1,4,0,1]=(-1.6461046371)*(1+2.*(dim-1.)/(o+1.))+(-0.332867143554)*dim
        ref[1,4,1,0]=(0.333229017732)*(1+2.*(dim-1.)/(o+1.))+(-1.05552711715)*dim
        ref[1,4,1,1]=(1.49605710755)*(1+2.*(dim-1.)/(o+1.))+(-1.35425133515)*dim
        ref[1,4,2,0]=(0.319136490505)*(1+2.*(dim-1.)/(o+1.))+(0.246099391883)*dim
        ref[1,4,2,1]=(-0.0958737570155)*(1+2.*(dim-1.)/(o+1.))+(-0.0803850673082)*dim
        ref[2,0,0,0]=(0.0482996624604)*(1+2.*(dim-1.)/(o+1.))+(-0.171234658375)*dim
        ref[2,0,0,1]=(0.638353303057)*(1+2.*(dim-1.)/(o+1.))+(0.770431646834)*dim
        ref[2,0,1,0]=(0.725020629449)*(1+2.*(dim-1.)/(o+1.))+(-0.432155213079)*dim
        ref[2,0,1,1]=(-0.438983419037)*(1+2.*(dim-1.)/(o+1.))+(0.713258422414)*dim
        ref[2,0,2,0]=(1.91447500178)*(1+2.*(dim-1.)/(o+1.))+(0.0177948876308)*dim
        ref[2,0,2,1]=(0.764452327795)*(1+2.*(dim-1.)/(o+1.))+(1.50234050134)*dim
        ref[2,1,0,0]=(0.498336938411)*(1+2.*(dim-1.)/(o+1.))+(-0.547316751164)*dim
        ref[2,1,0,1]=(-0.809231518007)*(1+2.*(dim-1.)/(o+1.))+(0.562257458836)*dim
        ref[2,1,1,0]=(0.358494165977)*(1+2.*(dim-1.)/(o+1.))+(1.59250325722)*dim
        ref[2,1,1,1]=(-0.244841646941)*(1+2.*(dim-1.)/(o+1.))+(-1.24133734105)*dim
        ref[2,1,2,0]=(0.361536543287)*(1+2.*(dim-1.)/(o+1.))+(0.691200482629)*dim
        ref[2,1,2,1]=(-1.32376390702)*(1+2.*(dim-1.)/(o+1.))+(0.0224135356952)*dim
        ref[2,2,0,0]=(-0.312392576948)*(1+2.*(dim-1.)/(o+1.))+(-0.06747945238)*dim
        ref[2,2,0,1]=(-0.431605160301)*(1+2.*(dim-1.)/(o+1.))+(-0.421636641073)*dim
        ref[2,2,1,0]=(-0.589519172131)*(1+2.*(dim-1.)/(o+1.))+(-0.414769101251)*dim
        ref[2,2,1,1]=(-0.240659708601)*(1+2.*(dim-1.)/(o+1.))+(-0.589364433013)*dim
        ref[2,2,2,0]=(-0.722909959939)*(1+2.*(dim-1.)/(o+1.))+(0.851597634849)*dim
        ref[2,2,2,1]=(-0.440270728352)*(1+2.*(dim-1.)/(o+1.))+(1.816216624)*dim
        ref[2,3,0,0]=(1.17045302504)*(1+2.*(dim-1.)/(o+1.))+(0.524825015672)*dim
        ref[2,3,0,1]=(0.330856998602)*(1+2.*(dim-1.)/(o+1.))+(0.22315528114)*dim
        ref[2,3,1,0]=(1.30293776302)*(1+2.*(dim-1.)/(o+1.))+(-1.27501951078)*dim
        ref[2,3,1,1]=(0.203347651013)*(1+2.*(dim-1.)/(o+1.))+(-1.53261320445)*dim
        ref[2,3,2,0]=(-1.28694542186)*(1+2.*(dim-1.)/(o+1.))+(1.36002865967)*dim
        ref[2,3,2,1]=(0.458262163691)*(1+2.*(dim-1.)/(o+1.))+(-1.64758747517)*dim
        ref[2,4,0,0]=(-0.121698212896)*(1+2.*(dim-1.)/(o+1.))+(1.50704164854)*dim
        ref[2,4,0,1]=(0.778589099264)*(1+2.*(dim-1.)/(o+1.))+(0.668892364598)*dim
        ref[2,4,1,0]=(-1.08307710902)*(1+2.*(dim-1.)/(o+1.))+(0.82548622713)*dim
        ref[2,4,1,1]=(-0.537056351709)*(1+2.*(dim-1.)/(o+1.))+(1.70030227735)*dim
        ref[2,4,2,0]=(-0.304186980702)*(1+2.*(dim-1.)/(o+1.))+(0.38736577812)*dim
        ref[2,4,2,1]=(-0.9399542098)*(1+2.*(dim-1.)/(o+1.))+(-0.300064248815)*dim
        ref[3,0,0,0]=(-0.62892625728)*(1+2.*(dim-1.)/(o+1.))+(-0.641275948665)*dim
        ref[3,0,0,1]=(0.313089108286)*(1+2.*(dim-1.)/(o+1.))+(0.794209081837)*dim
        ref[3,0,1,0]=(0.343101289842)*(1+2.*(dim-1.)/(o+1.))+(0.81847403947)*dim
        ref[3,0,1,1]=(-1.73316157526)*(1+2.*(dim-1.)/(o+1.))+(-0.689801226243)*dim
        ref[3,0,2,0]=(0.950942254051)*(1+2.*(dim-1.)/(o+1.))+(-1.26322409499)*dim
        ref[3,0,2,1]=(0.651979063795)*(1+2.*(dim-1.)/(o+1.))+(-0.452840690279)*dim
        ref[3,1,0,0]=(-0.140265070679)*(1+2.*(dim-1.)/(o+1.))+(0.504336855579)*dim
        ref[3,1,0,1]=(-0.074714204279)*(1+2.*(dim-1.)/(o+1.))+(-0.383830687435)*dim
        ref[3,1,1,0]=(-0.0240930345095)*(1+2.*(dim-1.)/(o+1.))+(-0.30098905758)*dim
        ref[3,1,1,1]=(1.01141483666)*(1+2.*(dim-1.)/(o+1.))+(-0.453294297891)*dim
        ref[3,1,2,0]=(1.28477826738)*(1+2.*(dim-1.)/(o+1.))+(0.850142280731)*dim
        ref[3,1,2,1]=(-1.82840136237)*(1+2.*(dim-1.)/(o+1.))+(-0.651498817783)*dim
        ref[3,2,0,0]=(-0.687609595107)*(1+2.*(dim-1.)/(o+1.))+(0.453772317262)*dim
        ref[3,2,0,1]=(1.42350629505)*(1+2.*(dim-1.)/(o+1.))+(-0.739779539116)*dim
        ref[3,2,1,0]=(0.241203525596)*(1+2.*(dim-1.)/(o+1.))+(0.389178645253)*dim
        ref[3,2,1,1]=(0.930949546469)*(1+2.*(dim-1.)/(o+1.))+(1.12624216731)*dim
        ref[3,2,2,0]=(-0.736761856298)*(1+2.*(dim-1.)/(o+1.))+(0.102231272315)*dim
        ref[3,2,2,1]=(-1.47668816374)*(1+2.*(dim-1.)/(o+1.))+(-0.136900288452)*dim
        ref[3,3,0,0]=(-0.200131789262)*(1+2.*(dim-1.)/(o+1.))+(1.24257798386)*dim
        ref[3,3,0,1]=(-1.24299838876)*(1+2.*(dim-1.)/(o+1.))+(0.891724959663)*dim
        ref[3,3,1,0]=(-0.61535668979)*(1+2.*(dim-1.)/(o+1.))+(-1.70378258198)*dim
        ref[3,3,1,1]=(-0.905774525561)*(1+2.*(dim-1.)/(o+1.))+(0.472323889318)*dim
        ref[3,3,2,0]=(-0.537356496864)*(1+2.*(dim-1.)/(o+1.))+(-1.22645090578)*dim
        ref[3,3,2,1]=(-1.61666551906)*(1+2.*(dim-1.)/(o+1.))+(-0.168452197544)*dim
        ref[3,4,0,0]=(-0.863203442693)*(1+2.*(dim-1.)/(o+1.))+(0.363489776493)*dim
        ref[3,4,0,1]=(-0.521102239287)*(1+2.*(dim-1.)/(o+1.))+(-0.00618566424301)*dim
        ref[3,4,1,0]=(0.602096055815)*(1+2.*(dim-1.)/(o+1.))+(0.0366884579134)*dim
        ref[3,4,1,1]=(0.844734896522)*(1+2.*(dim-1.)/(o+1.))+(0.0719024636709)*dim
        ref[3,4,2,0]=(-0.122274967513)*(1+2.*(dim-1.)/(o+1.))+(-0.164971745593)*dim
        ref[3,4,2,1]=(0.893907978172)*(1+2.*(dim-1.)/(o+1.))+(-0.0157346112292)*dim
      else:
        arg[0,0,0,0]=(-0.0445963085788)*x[0]**o+(0.649401127647)*x[0]+(-0.949035518207)*x[1]**o+(0.620675629012)*x[1]+(0.379727527503)*x[2]**o+(0.0321033403935)*x[2]
        arg[0,0,0,1]=(-0.334483142061)*x[0]**o+(-0.408749824143)*x[0]+(-0.427042441291)*x[1]**o+(-0.263818177378)*x[1]+(-0.0490001897328)*x[2]**o+(0.925792628022)*x[2]
        arg[0,0,1,0]=(0.632498844224)*x[0]**o+(-0.756487058254)*x[0]+(-0.0277391041902)*x[1]**o+(0.615017171892)*x[1]+(-0.0956053270216)*x[2]**o+(-0.0383148218356)*x[2]
        arg[0,0,1,1]=(0.560189295143)*x[0]**o+(-0.565632451768)*x[0]+(0.131609132282)*x[1]**o+(0.689012872903)*x[1]+(-0.0542418297927)*x[2]**o+(-0.816694105058)*x[2]
        arg[0,0,2,0]=(-0.606523401606)*x[0]**o+(-0.129782638438)*x[0]+(0.286554604329)*x[1]**o+(-0.264164004519)*x[1]+(-0.145168356054)*x[2]**o+(0.346147680755)*x[2]
        arg[0,0,2,1]=(-0.847528471559)*x[0]**o+(-0.50641170288)*x[0]+(-0.773325001399)*x[1]**o+(0.517649361249)*x[1]+(-0.531000881434)*x[2]**o+(0.220222077563)*x[2]
        arg[0,1,0,0]=(-0.0946495378384)*x[0]**o+(-0.312650449416)*x[0]+(0.183321918505)*x[1]**o+(-0.359448777446)*x[1]+(0.713219915733)*x[2]**o+(-0.436742264028)*x[2]
        arg[0,1,0,1]=(0.856947337921)*x[0]**o+(0.497228192699)*x[0]+(0.351718755501)*x[1]**o+(-0.928210477058)*x[1]+(0.586853747677)*x[2]**o+(-0.915642724205)*x[2]
        arg[0,1,1,0]=(-0.745330469339)*x[0]**o+(-0.724610807566)*x[0]+(0.821804346007)*x[1]**o+(0.241991559492)*x[1]+(0.508380905493)*x[2]**o+(0.756706999359)*x[2]
        arg[0,1,1,1]=(-0.647952004012)*x[0]**o+(0.940964624139)*x[0]+(-0.620895246845)*x[1]**o+(0.577736885116)*x[1]+(0.228716730257)*x[2]**o+(0.336914427419)*x[2]
        arg[0,1,2,0]=(-0.671541115529)*x[0]**o+(0.153595444626)*x[0]+(0.420649971716)*x[1]**o+(-0.459078302559)*x[1]+(-0.583263239511)*x[2]**o+(-0.927250932344)*x[2]
        arg[0,1,2,1]=(-0.893954406097)*x[0]**o+(0.266297374506)*x[0]+(-0.777463374391)*x[1]**o+(0.0905062188555)*x[1]+(-0.668756964585)*x[2]**o+(-0.580980380472)*x[2]
        arg[0,2,0,0]=(0.424449538697)*x[0]**o+(0.0787502268502)*x[0]+(-0.645996834851)*x[1]**o+(-0.0409730346273)*x[1]+(0.835463212243)*x[2]**o+(0.716052012249)*x[2]
        arg[0,2,0,1]=(0.0241441154932)*x[0]**o+(-0.580242261617)*x[0]+(0.693484316968)*x[1]**o+(0.198598746328)*x[1]+(-0.586283866696)*x[2]**o+(-0.595008846645)*x[2]
        arg[0,2,1,0]=(0.074643270363)*x[0]**o+(0.454053362138)*x[0]+(0.861855819289)*x[1]**o+(-0.0310194739742)*x[1]+(-0.35738290172)*x[2]**o+(-0.400226698584)*x[2]
        arg[0,2,1,1]=(0.61780464269)*x[0]**o+(-0.702818770068)*x[0]+(0.0840653474306)*x[1]**o+(0.786842849442)*x[1]+(0.0704342755849)*x[2]**o+(0.386459053939)*x[2]
        arg[0,2,2,0]=(-0.815363379053)*x[0]**o+(-0.621823615812)*x[0]+(-0.13535005543)*x[1]**o+(-0.127080917834)*x[1]+(-0.178866000441)*x[2]**o+(-0.0753933476244)*x[2]
        arg[0,2,2,1]=(0.400194366678)*x[0]**o+(0.075003395019)*x[0]+(-0.35708995046)*x[1]**o+(-0.370028795924)*x[1]+(-0.674944523932)*x[2]**o+(-0.437373855225)*x[2]
        arg[0,3,0,0]=(-0.736317751792)*x[0]**o+(0.671627391036)*x[0]+(0.0768737429532)*x[1]**o+(-0.90036869218)*x[1]+(0.796767364925)*x[2]**o+(-0.185034722102)*x[2]
        arg[0,3,0,1]=(0.530921403378)*x[0]**o+(0.970385957718)*x[0]+(-0.75659968724)*x[1]**o+(-0.437200812366)*x[1]+(0.0519374612092)*x[2]**o+(0.657323406602)*x[2]
        arg[0,3,1,0]=(-0.109751102222)*x[0]**o+(-0.154990305064)*x[0]+(-0.502004367351)*x[1]**o+(0.535266424739)*x[1]+(-0.968926773395)*x[2]**o+(-0.138032126456)*x[2]
        arg[0,3,1,1]=(0.509506900641)*x[0]**o+(-0.872357145765)*x[0]+(-0.154129018225)*x[1]**o+(0.0190914956768)*x[1]+(0.119094726078)*x[2]**o+(0.29593674915)*x[2]
        arg[0,3,2,0]=(-0.133398979819)*x[0]**o+(-0.780852347888)*x[0]+(-0.346576066687)*x[1]**o+(0.187923380934)*x[1]+(0.939817544138)*x[2]**o+(-0.497081081638)*x[2]
        arg[0,3,2,1]=(-0.908783140711)*x[0]**o+(0.541754659297)*x[0]+(-0.297350066105)*x[1]**o+(0.118676031731)*x[1]+(0.985854358996)*x[2]**o+(-0.424715359194)*x[2]
        arg[0,4,0,0]=(0.341191520307)*x[0]**o+(-0.252381555804)*x[0]+(-0.307289805415)*x[1]**o+(0.762228901652)*x[1]+(0.390901852611)*x[2]**o+(0.147349139229)*x[2]
        arg[0,4,0,1]=(0.236625833503)*x[0]**o+(0.856322905978)*x[0]+(-0.160244298088)*x[1]**o+(-0.971940590481)*x[1]+(-0.0829366716357)*x[2]**o+(-0.687610358703)*x[2]
        arg[0,4,1,0]=(0.619591455739)*x[0]**o+(-0.62589432832)*x[0]+(-0.0275687591477)*x[1]**o+(0.367590198294)*x[1]+(0.745367285277)*x[2]**o+(0.232571513762)*x[2]
        arg[0,4,1,1]=(0.317669904199)*x[0]**o+(0.741734360559)*x[0]+(-0.923644886773)*x[1]**o+(-0.725994528938)*x[1]+(-0.949385908603)*x[2]**o+(-0.0515088731457)*x[2]
        arg[0,4,2,0]=(0.82547889841)*x[0]**o+(-0.926970312436)*x[0]+(0.946941744691)*x[1]**o+(-0.171582097545)*x[1]+(-0.185628570945)*x[2]**o+(0.40495393086)*x[2]
        arg[0,4,2,1]=(0.51879613184)*x[0]**o+(-0.517447781109)*x[0]+(-0.288826585096)*x[1]**o+(-0.736890998485)*x[1]+(-0.167119326999)*x[2]**o+(-0.653082372893)*x[2]
        arg[1,0,0,0]=(-0.879008868811)*x[0]**o+(-0.0613987509179)*x[0]+(-0.671288168832)*x[1]**o+(-0.959517297906)*x[1]+(0.292554039162)*x[2]**o+(0.942246368678)*x[2]
        arg[1,0,0,1]=(0.599956815218)*x[0]**o+(0.731601305574)*x[0]+(-0.61593947163)*x[1]**o+(-0.597924796707)*x[1]+(-0.791620695721)*x[2]**o+(-0.562978718552)*x[2]
        arg[1,0,1,0]=(0.448381207242)*x[0]**o+(-0.178004176549)*x[0]+(0.631841924913)*x[1]**o+(0.311836637044)*x[1]+(0.0619475841251)*x[2]**o+(0.606564490624)*x[2]
        arg[1,0,1,1]=(0.159907398582)*x[0]**o+(-0.278605461684)*x[0]+(0.822914081538)*x[1]**o+(0.75112795022)*x[1]+(0.91389895436)*x[2]**o+(0.929177149059)*x[2]
        arg[1,0,2,0]=(-0.472286972241)*x[0]**o+(0.00950883684535)*x[0]+(0.41110884679)*x[1]**o+(0.109422159345)*x[1]+(-0.707436024011)*x[2]**o+(0.795057014561)*x[2]
        arg[1,0,2,1]=(-0.0156784802605)*x[0]**o+(-0.657283760492)*x[0]+(-0.0422880440635)*x[1]**o+(-0.326436491982)*x[1]+(-0.543917764251)*x[2]**o+(-0.691698129266)*x[2]
        arg[1,1,0,0]=(-0.814424723604)*x[0]**o+(0.746801606056)*x[0]+(0.22025107532)*x[1]**o+(-0.785254149057)*x[1]+(-0.845637273475)*x[2]**o+(-0.468272754144)*x[2]
        arg[1,1,0,1]=(-0.0815177460988)*x[0]**o+(-0.47210691717)*x[0]+(-0.285964111717)*x[1]**o+(0.527387496664)*x[1]+(0.87059583429)*x[2]**o+(-0.543899395468)*x[2]
        arg[1,1,1,0]=(0.424163426295)*x[0]**o+(0.649692204305)*x[0]+(0.0570010318344)*x[1]**o+(-0.426047169951)*x[1]+(0.334750596342)*x[2]**o+(-0.746549570927)*x[2]
        arg[1,1,1,1]=(-0.120493849358)*x[0]**o+(0.567677110538)*x[0]+(-0.355891513526)*x[1]**o+(0.68069794787)*x[1]+(0.786560156246)*x[2]**o+(-0.422804642829)*x[2]
        arg[1,1,2,0]=(-0.80298202932)*x[0]**o+(-0.674766348751)*x[0]+(-0.29945072468)*x[1]**o+(0.385941328636)*x[1]+(-0.756272624651)*x[2]**o+(-0.607541099269)*x[2]
        arg[1,1,2,1]=(0.915020713521)*x[0]**o+(-0.875957199283)*x[0]+(0.0217597029367)*x[1]**o+(-0.0649461013997)*x[1]+(-0.302450259541)*x[2]**o+(0.738498325526)*x[2]
        arg[1,2,0,0]=(0.239905462913)*x[0]**o+(0.197348893262)*x[0]+(-0.151173414577)*x[1]**o+(0.368405558905)*x[1]+(0.891448032058)*x[2]**o+(0.874975769684)*x[2]
        arg[1,2,0,1]=(-0.304971614823)*x[0]**o+(-0.0166065674003)*x[0]+(0.883397157318)*x[1]**o+(-0.918803270424)*x[1]+(0.590875856583)*x[2]**o+(-0.360659400522)*x[2]
        arg[1,2,1,0]=(0.185368540041)*x[0]**o+(0.498866146124)*x[0]+(0.591601152982)*x[1]**o+(-0.986932631453)*x[1]+(-0.525706872664)*x[2]**o+(-0.0462993899877)*x[2]
        arg[1,2,1,1]=(0.260677894495)*x[0]**o+(-0.216847258131)*x[0]+(0.932876363707)*x[1]**o+(0.761630676808)*x[1]+(0.260889221066)*x[2]**o+(0.546259908582)*x[2]
        arg[1,2,2,0]=(-0.316803600202)*x[0]**o+(-0.898044300642)*x[0]+(-0.0865001657852)*x[1]**o+(0.760424275199)*x[1]+(0.771564532427)*x[2]**o+(-0.0852237819025)*x[2]
        arg[1,2,2,1]=(-0.54274371504)*x[0]**o+(0.0822524750896)*x[0]+(0.872655885939)*x[1]**o+(0.594054041006)*x[1]+(0.322150325252)*x[2]**o+(0.774020612668)*x[2]
        arg[1,3,0,0]=(0.034593806358)*x[0]**o+(0.541638857337)*x[0]+(0.291359214696)*x[1]**o+(-0.203872730757)*x[1]+(0.0312260615928)*x[2]**o+(0.62568815669)*x[2]
        arg[1,3,0,1]=(-0.31929050432)*x[0]**o+(-0.931756481553)*x[0]+(0.755537511071)*x[1]**o+(-0.182578278538)*x[1]+(0.00958902784528)*x[2]**o+(0.345520390047)*x[2]
        arg[1,3,1,0]=(-0.435083251076)*x[0]**o+(0.248332963857)*x[0]+(-0.122701495966)*x[1]**o+(0.669350594899)*x[1]+(-0.0758596665333)*x[2]**o+(0.0784997907526)*x[2]
        arg[1,3,1,1]=(-0.725479421981)*x[0]**o+(0.657242746138)*x[0]+(-0.199958263735)*x[1]**o+(-0.784165440236)*x[1]+(-0.550936676647)*x[2]**o+(-0.31907813038)*x[2]
        arg[1,3,2,0]=(0.657177346512)*x[0]**o+(-0.168053866293)*x[0]+(0.782142168258)*x[1]**o+(0.660514922866)*x[1]+(0.788106295165)*x[2]**o+(-0.254290645125)*x[2]
        arg[1,3,2,1]=(-0.215166688682)*x[0]**o+(0.743577328376)*x[0]+(-0.374827404054)*x[1]**o+(0.267885661143)*x[1]+(0.728818301014)*x[2]**o+(0.6828999278)*x[2]
        arg[1,4,0,0]=(0.371646403413)*x[0]**o+(0.00998777673275)*x[0]+(-0.313378345871)*x[1]**o+(-0.952580545501)*x[1]+(-0.680596974717)*x[2]**o+(-0.181426263871)*x[2]
        arg[1,4,0,1]=(-0.828437965469)*x[0]**o+(0.302443878606)*x[0]+(0.581310358652)*x[1]**o+(0.714092786618)*x[1]+(0.650179420347)*x[2]**o+(0.663482256421)*x[2]
        arg[1,4,1,0]=(0.220318191456)*x[0]**o+(0.443108440795)*x[0]+(-0.538413779146)*x[1]**o+(0.660052783252)*x[1]+(-0.846698606343)*x[2]**o+(-0.0447806256913)*x[2]
        arg[1,4,1,1]=(0.866921895372)*x[0]**o+(0.776291161128)*x[0]+(-0.385318933957)*x[1]**o+(0.841514060614)*x[1]+(-0.401465329431)*x[2]**o+(0.474475442638)*x[2]
        arg[1,4,2,0]=(-0.222535436657)*x[0]**o+(0.921341450452)*x[0]+(-0.765043798058)*x[1]**o+(-0.492946667792)*x[1]+(0.758817008648)*x[2]**o+(0.0950595321909)*x[2]
        arg[1,4,2,1]=(0.333884542089)*x[0]**o+(0.671265678974)*x[0]+(-0.719262407869)*x[1]**o+(0.498176251457)*x[1]+(-0.867726438404)*x[2]**o+(-0.872614195608)*x[2]
        arg[2,0,0,0]=(-0.303331399994)*x[0]**o+(-0.563887060818)*x[0]+(0.64537626236)*x[1]**o+(-0.657611314342)*x[1]+(0.800882746828)*x[2]**o+(0.700671892157)*x[2]
        arg[2,0,0,1]=(0.417691239931)*x[0]**o+(-0.822789801933)*x[0]+(0.712214951539)*x[1]**o+(0.160960575173)*x[1]+(0.0400205244239)*x[2]**o+(0.49892939112)*x[2]
        arg[2,0,1,0]=(-0.828677767124)*x[0]**o+(-0.347411290435)*x[0]+(-0.700177402223)*x[1]**o+(-0.42498634463)*x[1]+(0.984172081328)*x[2]**o+(-0.677168643292)*x[2]
        arg[2,0,1,1]=(-0.423091398412)*x[0]**o+(-0.786913239913)*x[0]+(-0.888102258019)*x[1]**o+(-0.119525894432)*x[1]+(-0.158513498081)*x[2]**o+(0.503403674219)*x[2]
        arg[2,0,2,0]=(0.547217868614)*x[0]**o+(-0.737401808325)*x[0]+(-0.865530737659)*x[1]**o+(-0.750671412749)*x[1]+(-0.06731420082)*x[2]**o+(0.658572191108)*x[2]
        arg[2,0,2,1]=(0.501785022047)*x[0]**o+(-0.0542441989122)*x[0]+(-0.379614025724)*x[1]**o+(-0.785618793316)*x[1]+(0.994476365391)*x[2]**o+(-0.436996817494)*x[2]
        arg[2,1,0,0]=(0.0329439903445)*x[0]**o+(0.688097592632)*x[0]+(0.554030860489)*x[1]**o+(0.60949736297)*x[1]+(0.499716171437)*x[2]**o+(0.501464282299)*x[2]
        arg[2,1,0,1]=(0.459629414019)*x[0]**o+(0.810658146454)*x[0]+(-0.902023596001)*x[1]**o+(-0.982998009297)*x[1]+(0.487736529196)*x[2]**o+(-0.509164692477)*x[2]
        arg[2,1,1,0]=(0.0818151896334)*x[0]**o+(-0.931023788966)*x[0]+(-0.217349745187)*x[1]**o+(-0.785586110912)*x[1]+(0.16783052001)*x[2]**o+(-0.515867681053)*x[2]
        arg[2,1,1,1]=(-0.758874975703)*x[0]**o+(0.219408816573)*x[0]+(0.508187203896)*x[1]**o+(0.306700464008)*x[1]+(-0.641891226105)*x[2]**o+(-0.137752888266)*x[2]
        arg[2,1,2,0]=(-0.952430879482)*x[0]**o+(0.179895758684)*x[0]+(0.502285174801)*x[1]**o+(-0.82300826605)*x[1]+(-0.267848001419)*x[2]**o+(-0.882529775244)*x[2]
        arg[2,1,2,1]=(0.371372868601)*x[0]**o+(-0.521801538326)*x[0]+(0.894906517432)*x[1]**o+(0.205799057307)*x[1]+(-0.315840560279)*x[2]**o+(0.65823096555)*x[2]
        arg[2,2,0,0]=(0.585088762354)*x[0]**o+(0.527662713737)*x[0]+(0.83511043292)*x[1]**o+(-0.263597549489)*x[1]+(-0.342210974079)*x[2]**o+(0.532149528802)*x[2]
        arg[2,2,0,1]=(0.985868978475)*x[0]**o+(0.162798954854)*x[0]+(-0.887525429296)*x[1]**o+(0.618905848201)*x[1]+(-0.213643353241)*x[2]**o+(0.0170513098359)*x[2]
        arg[2,2,1,0]=(-0.216402064475)*x[0]**o+(0.350853951704)*x[0]+(0.592219675443)*x[1]**o+(-0.915293916561)*x[1]+(0.160588856435)*x[2]**o+(-0.574686420146)*x[2]
        arg[2,2,1,1]=(0.233118497878)*x[0]**o+(0.482035958353)*x[0]+(-0.581971661478)*x[1]**o+(0.832450520208)*x[1]+(-0.430914444794)*x[2]**o+(-0.791658025494)*x[2]
        arg[2,2,2,0]=(-0.710516844)*x[0]**o+(0.0266391730256)*x[0]+(-0.154541997151)*x[1]**o+(-0.232843778379)*x[1]+(0.453522207498)*x[2]**o+(-0.688939516711)*x[2]
        arg[2,2,2,1]=(0.26860058468)*x[0]**o+(-0.969638661751)*x[0]+(0.623784614857)*x[1]**o+(0.686359051774)*x[1]+(0.677056704385)*x[2]**o+(0.156688176847)*x[2]
        arg[2,3,0,0]=(-0.0620909841534)*x[0]**o+(-0.364946660535)*x[0]+(-0.388275430806)*x[1]**o+(-0.947908160286)*x[1]+(0.0774302410344)*x[2]**o+(0.374469033649)*x[2]
        arg[2,3,0,1]=(0.164845147121)*x[0]**o+(0.480213031852)*x[0]+(0.58157030274)*x[1]**o+(0.882820521347)*x[1]+(0.584714825748)*x[2]**o+(0.0285154925376)*x[2]
        arg[2,3,1,0]=(0.664711234489)*x[0]**o+(0.247415435607)*x[0]+(0.741592287061)*x[1]**o+(0.963505986041)*x[1]+(-0.987844687003)*x[2]**o+(0.998294459659)*x[2]
        arg[2,3,1,1]=(-0.734529316764)*x[0]**o+(-0.177480004713)*x[0]+(0.923555424779)*x[1]**o+(0.727331498256)*x[1]+(0.452312367674)*x[2]**o+(0.00652243396085)*x[2]
        arg[2,3,2,0]=(-0.367377897953)*x[0]**o+(-0.157529442804)*x[0]+(-0.121881528621)*x[1]**o+(0.865116726884)*x[1]+(-0.200087079925)*x[2]**o+(0.930889629788)*x[2]
        arg[2,3,2,1]=(-0.900394562978)*x[0]**o+(-0.762334071565)*x[0]+(-0.723776245294)*x[1]**o+(0.332728994729)*x[1]+(0.390453779523)*x[2]**o+(0.0447350797932)*x[2]
        arg[2,4,0,0]=(0.178230443819)*x[0]**o+(0.0898732089623)*x[0]+(0.808448952186)*x[1]**o+(0.164579824494)*x[1]+(-0.553032263217)*x[2]**o+(-0.782141879825)*x[2]
        arg[2,4,0,1]=(-0.105582488753)*x[0]**o+(0.995415696949)*x[0]+(0.978971670365)*x[1]**o+(-0.977866735017)*x[1]+(0.156938966924)*x[2]**o+(-0.120871488936)*x[2]
        arg[2,4,1,0]=(0.652354601889)*x[0]**o+(0.381186764672)*x[0]+(-0.783479778972)*x[1]**o+(-0.158898829615)*x[1]+(0.794693790129)*x[2]**o+(0.914720093741)*x[2]
        arg[2,4,1,1]=(0.571506419376)*x[0]**o+(0.906086581152)*x[0]+(-0.798519792125)*x[1]**o+(-0.22840897962)*x[1]+(-0.750891721145)*x[2]**o+(-0.591014038221)*x[2]
        arg[2,4,2,0]=(0.515982716306)*x[0]**o+(0.299569326845)*x[0]+(-0.407423180268)*x[1]**o+(-0.252103363573)*x[1]+(-0.0228238994531)*x[2]**o+(-0.282150542844)*x[2]
        arg[2,4,2,1]=(0.5335415942)*x[0]**o+(-0.885728570489)*x[0]+(-0.750810131327)*x[1]**o+(0.604053940679)*x[1]+(-0.0946405866686)*x[2]**o+(-0.423845465234)*x[2]
        arg[3,0,0,0]=(-0.498025341136)*x[0]**o+(-0.296354416633)*x[0]+(-0.600874889655)*x[1]**o+(0.544042300882)*x[1]+(0.318784851038)*x[2]**o+(-0.801109111437)*x[2]
        arg[3,0,0,1]=(-0.492361949745)*x[0]**o+(-0.708188092599)*x[0]+(-0.863737808435)*x[1]**o+(0.674889485553)*x[1]+(-0.982254652233)*x[2]**o+(0.682172862189)*x[2]
        arg[3,0,1,0]=(-0.872692397532)*x[0]**o+(0.563852836655)*x[0]+(-0.0219331486599)*x[1]**o+(-0.858175679559)*x[1]+(0.802432631538)*x[2]**o+(-0.655545287387)*x[2]
        arg[3,0,1,1]=(-0.538117166854)*x[0]**o+(-0.62507518352)*x[0]+(0.101524298816)*x[1]**o+(0.590432448393)*x[1]+(0.668057302006)*x[2]**o+(0.405382330056)*x[2]
        arg[3,0,2,0]=(-0.322916455064)*x[0]**o+(0.950624857669)*x[0]+(0.485623846818)*x[1]**o+(-0.12611522301)*x[1]+(0.503586393973)*x[2]**o+(0.109451719709)*x[2]
        arg[3,0,2,1]=(0.0289965995645)*x[0]**o+(0.0101440661359)*x[0]+(-0.303827119958)*x[1]**o+(0.355269935141)*x[1]+(0.29889427526)*x[2]**o+(-0.707999988781)*x[2]
        arg[3,1,0,0]=(-0.541382521848)*x[0]**o+(0.534108535206)*x[0]+(0.875610773119)*x[1]**o+(-0.575849439451)*x[1]+(-0.0836939606164)*x[2]**o+(0.433678856889)*x[2]
        arg[3,1,0,1]=(0.0322576144027)*x[0]**o+(0.794219359514)*x[0]+(0.629442621265)*x[1]**o+(0.0517726053499)*x[1]+(-0.776050101686)*x[2]**o+(0.697910196708)*x[2]
        arg[3,1,1,0]=(-0.198211682395)*x[0]**o+(-0.926268101941)*x[0]+(-0.904760383993)*x[1]**o+(-0.820075667018)*x[1]+(-0.101262044428)*x[2]**o+(-0.0621221772941)*x[2]
        arg[3,1,1,1]=(-0.531324749521)*x[0]**o+(0.736017207718)*x[0]+(0.487299696259)*x[1]**o+(-0.00604423021253)*x[1]+(-0.635722590088)*x[2]**o+(-0.143501730927)*x[2]
        arg[3,1,2,0]=(-0.731421637848)*x[0]**o+(-0.0371648783851)*x[0]+(-0.290362466891)*x[1]**o+(-0.375048584106)*x[1]+(-0.854622532092)*x[2]**o+(-0.794295905388)*x[2]
        arg[3,1,2,1]=(-0.766383422105)*x[0]**o+(0.149248159984)*x[0]+(0.970103674649)*x[1]**o+(-0.117794368107)*x[1]+(0.0178775633047)*x[2]**o+(-0.221126829278)*x[2]
        arg[3,2,0,0]=(0.706586887601)*x[0]**o+(0.162520023447)*x[0]+(-0.731539880336)*x[1]**o+(-0.992558519713)*x[1]+(0.426569303014)*x[2]**o+(0.0451783819487)*x[2]
        arg[3,2,0,1]=(0.0392295508272)*x[0]**o+(-0.157929005382)*x[0]+(0.0636418899878)*x[1]**o+(0.892501759578)*x[1]+(-0.368161001401)*x[2]**o+(-0.715520794543)*x[2]
        arg[3,2,1,0]=(0.889667707133)*x[0]**o+(0.308604168817)*x[0]+(-0.275558663983)*x[1]**o+(-0.408210095215)*x[1]+(-0.0918000869138)*x[2]**o+(0.142754064645)*x[2]
        arg[3,2,1,1]=(0.585831954407)*x[0]**o+(-0.705004375516)*x[0]+(0.657277099061)*x[1]**o+(-0.972401794257)*x[1]+(-0.658049310396)*x[2]**o+(-0.553961555706)*x[2]
        arg[3,2,2,0]=(0.88916769079)*x[0]**o+(-0.1401162684)*x[0]+(-0.396496633317)*x[1]**o+(0.215029522669)*x[1]+(-0.619023294208)*x[2]**o+(-0.187670706968)*x[2]
        arg[3,2,2,1]=(-0.262983340384)*x[0]**o+(-0.763236528477)*x[0]+(-0.339720289405)*x[1]**o+(-0.283077066751)*x[1]+(0.625600522278)*x[2]**o+(-0.132697432298)*x[2]
        arg[3,3,0,0]=(-0.346408105223)*x[0]**o+(0.299760534041)*x[0]+(0.567951405594)*x[1]**o+(-0.880683463459)*x[1]+(0.784127947661)*x[2]**o+(0.268895744072)*x[2]
        arg[3,3,0,1]=(0.064159345936)*x[0]**o+(-0.685098798117)*x[0]+(-0.182310885124)*x[1]**o+(-0.700081075257)*x[1]+(0.543959312437)*x[2]**o+(0.125104170434)*x[2]
        arg[3,3,1,0]=(0.460375837083)*x[0]**o+(0.82355800923)*x[0]+(-0.186000244328)*x[1]**o+(0.8418355439)*x[1]+(-0.375269292115)*x[2]**o+(0.712431840155)*x[2]
        arg[3,3,1,1]=(0.57260867805)*x[0]**o+(-0.146405851776)*x[0]+(-0.50640779246)*x[1]**o+(0.741760309299)*x[1]+(0.069752390409)*x[2]**o+(-0.376724191471)*x[2]
        arg[3,3,2,0]=(0.795493408403)*x[0]**o+(0.396861648344)*x[0]+(-0.728716068325)*x[1]**o+(0.739330072416)*x[1]+(0.202019645585)*x[2]**o+(0.841240509566)*x[2]
        arg[3,3,2,1]=(-0.967372744621)*x[0]**o+(-0.334044549662)*x[0]+(-0.0686982670779)*x[1]**o+(0.180083293629)*x[1]+(-0.208697793793)*x[2]**o+(0.0305446253751)*x[2]
        arg[3,4,0,0]=(0.94168350996)*x[0]**o+(0.0545532586137)*x[0]+(-0.300430532207)*x[1]**o+(-0.960362884765)*x[1]+(0.738261342527)*x[2]**o+(-0.351477574571)*x[2]
        arg[3,4,0,1]=(-0.525893857873)*x[0]**o+(0.227252001588)*x[0]+(0.179831911446)*x[1]**o+(0.350952712138)*x[1]+(-0.671550635312)*x[2]**o+(-0.338834217909)*x[2]
        arg[3,4,1,0]=(-0.60610056656)*x[0]**o+(0.219192673102)*x[0]+(-0.801595717702)*x[1]**o+(-0.696853420225)*x[1]+(-0.495201990047)*x[2]**o+(-0.500683649414)*x[2]
        arg[3,4,1,1]=(-0.37343020332)*x[0]**o+(-0.585536591527)*x[0]+(0.81794997494)*x[1]**o+(-0.240232527585)*x[1]+(-0.538970681636)*x[2]**o+(-0.146105210473)*x[2]
        arg[3,4,2,0]=(0.857322149358)*x[0]**o+(-0.552847286493)*x[0]+(0.530970378294)*x[1]**o+(0.0746811386922)*x[1]+(0.946170976628)*x[2]**o+(-0.329916962713)*x[2]
        arg[3,4,2,1]=(-0.353323100685)*x[0]**o+(0.808796594618)*x[0]+(0.481450588496)*x[1]**o+(0.1202874944)*x[1]+(0.671293736572)*x[2]**o+(-0.905074286629)*x[2]
        ref[0,0,0,0]=(-0.613904299283)*(1+2.*(dim-1.)/(o+1.))+(1.30218009705)*dim
        ref[0,0,0,1]=(-0.810525773084)*(1+2.*(dim-1.)/(o+1.))+(0.253224626501)*dim
        ref[0,0,1,0]=(0.509154413012)*(1+2.*(dim-1.)/(o+1.))+(-0.179784708198)*dim
        ref[0,0,1,1]=(0.637556597632)*(1+2.*(dim-1.)/(o+1.))+(-0.693313683923)*dim
        ref[0,0,2,0]=(-0.465137153331)*(1+2.*(dim-1.)/(o+1.))+(-0.0477989622018)*dim
        ref[0,0,2,1]=(-2.15185435439)*(1+2.*(dim-1.)/(o+1.))+(0.231459735932)*dim
        ref[0,1,0,0]=(0.8018922964)*(1+2.*(dim-1.)/(o+1.))+(-1.10884149089)*dim
        ref[0,1,0,1]=(1.7955198411)*(1+2.*(dim-1.)/(o+1.))+(-1.34662500856)*dim
        ref[0,1,1,0]=(0.584854782162)*(1+2.*(dim-1.)/(o+1.))+(0.274087751285)*dim
        ref[0,1,1,1]=(-1.0401305206)*(1+2.*(dim-1.)/(o+1.))+(1.85561593667)*dim
        ref[0,1,2,0]=(-0.834154383324)*(1+2.*(dim-1.)/(o+1.))+(-1.23273379028)*dim
        ref[0,1,2,1]=(-2.34017474507)*(1+2.*(dim-1.)/(o+1.))+(-0.224176787111)*dim
        ref[0,2,0,0]=(0.613915916089)*(1+2.*(dim-1.)/(o+1.))+(0.753829204471)*dim
        ref[0,2,0,1]=(0.131344565765)*(1+2.*(dim-1.)/(o+1.))+(-0.976652361934)*dim
        ref[0,2,1,0]=(0.579116187931)*(1+2.*(dim-1.)/(o+1.))+(0.0228071895801)*dim
        ref[0,2,1,1]=(0.772304265705)*(1+2.*(dim-1.)/(o+1.))+(0.470483133313)*dim
        ref[0,2,2,0]=(-1.12957943492)*(1+2.*(dim-1.)/(o+1.))+(-0.82429788127)*dim
        ref[0,2,2,1]=(-0.631840107715)*(1+2.*(dim-1.)/(o+1.))+(-0.732399256131)*dim
        ref[0,3,0,0]=(0.137323356087)*(1+2.*(dim-1.)/(o+1.))+(-0.413776023245)*dim
        ref[0,3,0,1]=(-0.173740822652)*(1+2.*(dim-1.)/(o+1.))+(1.19050855195)*dim
        ref[0,3,1,0]=(-1.58068224297)*(1+2.*(dim-1.)/(o+1.))+(0.24224399322)*dim
        ref[0,3,1,1]=(0.474472608494)*(1+2.*(dim-1.)/(o+1.))+(-0.557328900938)*dim
        ref[0,3,2,0]=(0.459842497633)*(1+2.*(dim-1.)/(o+1.))+(-1.09001004859)*dim
        ref[0,3,2,1]=(-0.22027884782)*(1+2.*(dim-1.)/(o+1.))+(0.235715331835)*dim
        ref[0,4,0,0]=(0.424803567504)*(1+2.*(dim-1.)/(o+1.))+(0.657196485077)*dim
        ref[0,4,0,1]=(-0.00655513622033)*(1+2.*(dim-1.)/(o+1.))+(-0.803228043207)*dim
        ref[0,4,1,0]=(1.33738998187)*(1+2.*(dim-1.)/(o+1.))+(-0.0257326162636)*dim
        ref[0,4,1,1]=(-1.55536089118)*(1+2.*(dim-1.)/(o+1.))+(-0.0357690415245)*dim
        ref[0,4,2,0]=(1.58679207216)*(1+2.*(dim-1.)/(o+1.))+(-0.693598479122)*dim
        ref[0,4,2,1]=(0.0628502197447)*(1+2.*(dim-1.)/(o+1.))+(-1.90742115249)*dim
        ref[1,0,0,0]=(-1.25774299848)*(1+2.*(dim-1.)/(o+1.))+(-0.0786696801461)*dim
        ref[1,0,0,1]=(-0.807603352133)*(1+2.*(dim-1.)/(o+1.))+(-0.429302209684)*dim
        ref[1,0,1,0]=(1.14217071628)*(1+2.*(dim-1.)/(o+1.))+(0.740396951119)*dim
        ref[1,0,1,1]=(1.89672043448)*(1+2.*(dim-1.)/(o+1.))+(1.4016996376)*dim
        ref[1,0,2,0]=(-0.768614149462)*(1+2.*(dim-1.)/(o+1.))+(0.913988010751)*dim
        ref[1,0,2,1]=(-0.601884288575)*(1+2.*(dim-1.)/(o+1.))+(-1.67541838174)*dim
        ref[1,1,0,0]=(-1.43981092176)*(1+2.*(dim-1.)/(o+1.))+(-0.506725297145)*dim
        ref[1,1,0,1]=(0.503113976474)*(1+2.*(dim-1.)/(o+1.))+(-0.488618815974)*dim
        ref[1,1,1,0]=(0.815915054472)*(1+2.*(dim-1.)/(o+1.))+(-0.522904536573)*dim
        ref[1,1,1,1]=(0.310174793362)*(1+2.*(dim-1.)/(o+1.))+(0.825570415579)*dim
        ref[1,1,2,0]=(-1.85870537865)*(1+2.*(dim-1.)/(o+1.))+(-0.896366119383)*dim
        ref[1,1,2,1]=(0.634330156916)*(1+2.*(dim-1.)/(o+1.))+(-0.202404975156)*dim
        ref[1,2,0,0]=(0.980180080393)*(1+2.*(dim-1.)/(o+1.))+(1.44073022185)*dim
        ref[1,2,0,1]=(1.16930139908)*(1+2.*(dim-1.)/(o+1.))+(-1.29606923835)*dim
        ref[1,2,1,0]=(0.251262820359)*(1+2.*(dim-1.)/(o+1.))+(-0.534365875316)*dim
        ref[1,2,1,1]=(1.45444347927)*(1+2.*(dim-1.)/(o+1.))+(1.09104332726)*dim
        ref[1,2,2,0]=(0.36826076644)*(1+2.*(dim-1.)/(o+1.))+(-0.222843807345)*dim
        ref[1,2,2,1]=(0.652062496151)*(1+2.*(dim-1.)/(o+1.))+(1.45032712876)*dim
        ref[1,3,0,0]=(0.357179082647)*(1+2.*(dim-1.)/(o+1.))+(0.96345428327)*dim
        ref[1,3,0,1]=(0.445836034597)*(1+2.*(dim-1.)/(o+1.))+(-0.768814370045)*dim
        ref[1,3,1,0]=(-0.633644413575)*(1+2.*(dim-1.)/(o+1.))+(0.996183349509)*dim
        ref[1,3,1,1]=(-1.47637436236)*(1+2.*(dim-1.)/(o+1.))+(-0.446000824477)*dim
        ref[1,3,2,0]=(2.22742580994)*(1+2.*(dim-1.)/(o+1.))+(0.238170411449)*dim
        ref[1,3,2,1]=(0.138824208277)*(1+2.*(dim-1.)/(o+1.))+(1.69436291732)*dim
        ref[1,4,0,0]=(-0.622328917175)*(1+2.*(dim-1.)/(o+1.))+(-1.12401903264)*dim
        ref[1,4,0,1]=(0.40305181353)*(1+2.*(dim-1.)/(o+1.))+(1.68001892165)*dim
        ref[1,4,1,0]=(-1.16479419403)*(1+2.*(dim-1.)/(o+1.))+(1.05838059836)*dim
        ref[1,4,1,1]=(0.0801376319834)*(1+2.*(dim-1.)/(o+1.))+(2.09228066438)*dim
        ref[1,4,2,0]=(-0.228762226067)*(1+2.*(dim-1.)/(o+1.))+(0.523454314851)*dim
        ref[1,4,2,1]=(-1.25310430418)*(1+2.*(dim-1.)/(o+1.))+(0.296827734822)*dim
        ref[2,0,0,0]=(1.14292760919)*(1+2.*(dim-1.)/(o+1.))+(-0.520826483003)*dim
        ref[2,0,0,1]=(1.16992671589)*(1+2.*(dim-1.)/(o+1.))+(-0.162899835641)*dim
        ref[2,0,1,0]=(-0.544683088018)*(1+2.*(dim-1.)/(o+1.))+(-1.44956627836)*dim
        ref[2,0,1,1]=(-1.46970715451)*(1+2.*(dim-1.)/(o+1.))+(-0.403035460127)*dim
        ref[2,0,2,0]=(-0.385627069865)*(1+2.*(dim-1.)/(o+1.))+(-0.829501029966)*dim
        ref[2,0,2,1]=(1.11664736171)*(1+2.*(dim-1.)/(o+1.))+(-1.27685980972)*dim
        ref[2,1,0,0]=(1.08669102227)*(1+2.*(dim-1.)/(o+1.))+(1.7990592379)*dim
        ref[2,1,0,1]=(0.0453423472143)*(1+2.*(dim-1.)/(o+1.))+(-0.68150455532)*dim
        ref[2,1,1,0]=(0.0322959644566)*(1+2.*(dim-1.)/(o+1.))+(-2.23247758093)*dim
        ref[2,1,1,1]=(-0.892578997912)*(1+2.*(dim-1.)/(o+1.))+(0.388356392315)*dim
        ref[2,1,2,0]=(-0.7179937061)*(1+2.*(dim-1.)/(o+1.))+(-1.52564228261)*dim
        ref[2,1,2,1]=(0.950438825754)*(1+2.*(dim-1.)/(o+1.))+(0.342228484532)*dim
        ref[2,2,0,0]=(1.0779882212)*(1+2.*(dim-1.)/(o+1.))+(0.796214693049)*dim
        ref[2,2,0,1]=(-0.115299804062)*(1+2.*(dim-1.)/(o+1.))+(0.798756112891)*dim
        ref[2,2,1,0]=(0.536406467403)*(1+2.*(dim-1.)/(o+1.))+(-1.139126385)*dim
        ref[2,2,1,1]=(-0.779767608394)*(1+2.*(dim-1.)/(o+1.))+(0.522828453067)*dim
        ref[2,2,2,0]=(-0.411536633652)*(1+2.*(dim-1.)/(o+1.))+(-0.895144122064)*dim
        ref[2,2,2,1]=(1.56944190392)*(1+2.*(dim-1.)/(o+1.))+(-0.12659143313)*dim
        ref[2,3,0,0]=(-0.372936173925)*(1+2.*(dim-1.)/(o+1.))+(-0.938385787171)*dim
        ref[2,3,0,1]=(1.33113027561)*(1+2.*(dim-1.)/(o+1.))+(1.39154904574)*dim
        ref[2,3,1,0]=(0.418458834548)*(1+2.*(dim-1.)/(o+1.))+(2.20921588131)*dim
        ref[2,3,1,1]=(0.64133847569)*(1+2.*(dim-1.)/(o+1.))+(0.556373927504)*dim
        ref[2,3,2,0]=(-0.689346506499)*(1+2.*(dim-1.)/(o+1.))+(1.63847691387)*dim
        ref[2,3,2,1]=(-1.23371702875)*(1+2.*(dim-1.)/(o+1.))+(-0.384869997042)*dim
        ref[2,4,0,0]=(0.433647132788)*(1+2.*(dim-1.)/(o+1.))+(-0.527688846368)*dim
        ref[2,4,0,1]=(1.03032814854)*(1+2.*(dim-1.)/(o+1.))+(-0.103322527004)*dim
        ref[2,4,1,0]=(0.663568613046)*(1+2.*(dim-1.)/(o+1.))+(1.1370080288)*dim
        ref[2,4,1,1]=(-0.977905093893)*(1+2.*(dim-1.)/(o+1.))+(0.0866635633103)*dim
        ref[2,4,2,0]=(0.085735636585)*(1+2.*(dim-1.)/(o+1.))+(-0.234684579572)*dim
        ref[2,4,2,1]=(-0.311909123795)*(1+2.*(dim-1.)/(o+1.))+(-0.705520095044)*dim
        ref[3,0,0,0]=(-0.780115379753)*(1+2.*(dim-1.)/(o+1.))+(-0.553421227188)*dim
        ref[3,0,0,1]=(-2.33835441041)*(1+2.*(dim-1.)/(o+1.))+(0.648874255143)*dim
        ref[3,0,1,0]=(-0.0921929146543)*(1+2.*(dim-1.)/(o+1.))+(-0.949868130291)*dim
        ref[3,0,1,1]=(0.231464433969)*(1+2.*(dim-1.)/(o+1.))+(0.370739594929)*dim
        ref[3,0,2,0]=(0.666293785727)*(1+2.*(dim-1.)/(o+1.))+(0.933961354368)*dim
        ref[3,0,2,1]=(0.0240637548661)*(1+2.*(dim-1.)/(o+1.))+(-0.342585987504)*dim
        ref[3,1,0,0]=(0.250534290654)*(1+2.*(dim-1.)/(o+1.))+(0.391937952644)*dim
        ref[3,1,0,1]=(-0.114349866018)*(1+2.*(dim-1.)/(o+1.))+(1.54390216157)*dim
        ref[3,1,1,0]=(-1.20423411082)*(1+2.*(dim-1.)/(o+1.))+(-1.80846594625)*dim
        ref[3,1,1,1]=(-0.679747643349)*(1+2.*(dim-1.)/(o+1.))+(0.586471246578)*dim
        ref[3,1,2,0]=(-1.87640663683)*(1+2.*(dim-1.)/(o+1.))+(-1.20650936788)*dim
        ref[3,1,2,1]=(0.221597815849)*(1+2.*(dim-1.)/(o+1.))+(-0.189673037401)*dim
        ref[3,2,0,0]=(0.40161631028)*(1+2.*(dim-1.)/(o+1.))+(-0.784860114318)*dim
        ref[3,2,0,1]=(-0.265289560586)*(1+2.*(dim-1.)/(o+1.))+(0.0190519596518)*dim
        ref[3,2,1,0]=(0.522308956237)*(1+2.*(dim-1.)/(o+1.))+(0.0431481382473)*dim
        ref[3,2,1,1]=(0.585059743072)*(1+2.*(dim-1.)/(o+1.))+(-2.23136772548)*dim
        ref[3,2,2,0]=(-0.126352236735)*(1+2.*(dim-1.)/(o+1.))+(-0.112757452698)*dim
        ref[3,2,2,1]=(0.0228968924883)*(1+2.*(dim-1.)/(o+1.))+(-1.17901102753)*dim
        ref[3,3,0,0]=(1.00567124803)*(1+2.*(dim-1.)/(o+1.))+(-0.312027185345)*dim
        ref[3,3,0,1]=(0.425807773249)*(1+2.*(dim-1.)/(o+1.))+(-1.26007570294)*dim
        ref[3,3,1,0]=(-0.10089369936)*(1+2.*(dim-1.)/(o+1.))+(2.37782539328)*dim
        ref[3,3,1,1]=(0.135953275999)*(1+2.*(dim-1.)/(o+1.))+(0.218630266051)*dim
        ref[3,3,2,0]=(0.268796985663)*(1+2.*(dim-1.)/(o+1.))+(1.97743223033)*dim
        ref[3,3,2,1]=(-1.24476880549)*(1+2.*(dim-1.)/(o+1.))+(-0.123416630658)*dim
        ref[3,4,0,0]=(1.37951432028)*(1+2.*(dim-1.)/(o+1.))+(-1.25728720072)*dim
        ref[3,4,0,1]=(-1.01761258174)*(1+2.*(dim-1.)/(o+1.))+(0.239370495817)*dim
        ref[3,4,1,0]=(-1.90289827431)*(1+2.*(dim-1.)/(o+1.))+(-0.978344396537)*dim
        ref[3,4,1,1]=(-0.0944509100161)*(1+2.*(dim-1.)/(o+1.))+(-0.971874329585)*dim
        ref[3,4,2,0]=(2.33446350428)*(1+2.*(dim-1.)/(o+1.))+(-0.808083110513)*dim
        ref[3,4,2,1]=(0.799421224383)*(1+2.*(dim-1.)/(o+1.))+(0.0240098023897)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank0(self):
      """
      tests integral of rank 0 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.626062912046)*x[0]**o+(0.600712856149)*x[0]+(0.0478861882442)*x[1]**o+(0.77008552734)*x[1]
        ref=(0.67394910029)*(1+2.*(dim-1.)/(o+1.))+(1.37079838349)*dim
      else:
        arg=(0.85588880126)*x[0]**o+(-0.898804533462)*x[0]+(-0.270816921088)*x[1]**o+(-0.842823912373)*x[1]+(-0.47559690762)*x[2]**o+(0.751770125129)*x[2]
        ref=(0.109474972552)*(1+2.*(dim-1.)/(o+1.))+(-0.989858320706)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank1(self):
      """
      tests integral of rank 1 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.298913063662)*x[0]**o+(-0.250822875451)*x[0]+(0.707311253225)*x[1]**o+(0.0408091445557)*x[1]
        arg[1]=(-0.70143850042)*x[0]**o+(-0.80121227022)*x[0]+(0.0731514590408)*x[1]**o+(0.621969876113)*x[1]
        ref[0]=(1.00622431689)*(1+2.*(dim-1.)/(o+1.))+(-0.210013730895)*dim
        ref[1]=(-0.628287041379)*(1+2.*(dim-1.)/(o+1.))+(-0.179242394108)*dim
      else:
        arg[0]=(0.676976938336)*x[0]**o+(-0.928207054947)*x[0]+(0.864749399199)*x[1]**o+(0.864358390848)*x[1]+(-0.0255569869984)*x[2]**o+(0.0268536872793)*x[2]
        arg[1]=(0.270542056965)*x[0]**o+(0.0696194435081)*x[0]+(0.401360001043)*x[1]**o+(0.917064528874)*x[1]+(-0.755510144559)*x[2]**o+(0.881244515908)*x[2]
        ref[0]=(1.51616935054)*(1+2.*(dim-1.)/(o+1.))+(-0.0369949768193)*dim
        ref[1]=(-0.083608086551)*(1+2.*(dim-1.)/(o+1.))+(1.86792848829)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank2(self):
      """
      tests integral of rank 2 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.381843423534)*x[0]**o+(-0.563210012057)*x[0]+(0.730631055205)*x[1]**o+(0.13636238698)*x[1]
        arg[0,1]=(0.306357653338)*x[0]**o+(-0.797201232101)*x[0]+(-0.66333793372)*x[1]**o+(-0.943650360464)*x[1]
        arg[0,2]=(-0.34058649734)*x[0]**o+(-0.27047792232)*x[0]+(0.57156687267)*x[1]**o+(-0.659114355468)*x[1]
        arg[0,3]=(0.73319004289)*x[0]**o+(0.229139540294)*x[0]+(-0.51558765225)*x[1]**o+(0.45052521066)*x[1]
        arg[0,4]=(-0.264117198359)*x[0]**o+(0.807950109095)*x[0]+(-0.290558748289)*x[1]**o+(0.161864708045)*x[1]
        arg[1,0]=(0.988732415297)*x[0]**o+(-0.725047414136)*x[0]+(-0.722337599885)*x[1]**o+(0.739130135256)*x[1]
        arg[1,1]=(-0.315086233906)*x[0]**o+(-0.783177846939)*x[0]+(-0.271931454399)*x[1]**o+(-0.935744613016)*x[1]
        arg[1,2]=(-0.787047622811)*x[0]**o+(0.00746242211892)*x[0]+(-0.539716273578)*x[1]**o+(-0.340449115622)*x[1]
        arg[1,3]=(-0.782830405823)*x[0]**o+(0.842318742811)*x[0]+(-0.118894493096)*x[1]**o+(0.645943277412)*x[1]
        arg[1,4]=(0.0594068367647)*x[0]**o+(-0.184711682975)*x[0]+(-0.0643050731295)*x[1]**o+(-0.981593287261)*x[1]
        arg[2,0]=(0.17170754931)*x[0]**o+(0.582651363494)*x[0]+(0.334422749985)*x[1]**o+(0.852747429144)*x[1]
        arg[2,1]=(0.683967516208)*x[0]**o+(-0.612989314204)*x[0]+(0.845175135723)*x[1]**o+(0.374223395314)*x[1]
        arg[2,2]=(0.588917536035)*x[0]**o+(0.0130713873939)*x[0]+(0.881942872071)*x[1]**o+(0.683941046151)*x[1]
        arg[2,3]=(-0.63861264887)*x[0]**o+(0.453919752415)*x[0]+(0.448731342326)*x[1]**o+(-0.259183783496)*x[1]
        arg[2,4]=(-0.132175638549)*x[0]**o+(0.710071566415)*x[0]+(-0.870119332524)*x[1]**o+(0.523394538403)*x[1]
        arg[3,0]=(0.768036036634)*x[0]**o+(-0.218321673662)*x[0]+(0.50224512768)*x[1]**o+(0.694191572559)*x[1]
        arg[3,1]=(-0.298369075336)*x[0]**o+(-0.276776010079)*x[0]+(-0.659309162004)*x[1]**o+(0.436841182692)*x[1]
        arg[3,2]=(-0.00370049871499)*x[0]**o+(-0.41700259949)*x[0]+(0.726806931339)*x[1]**o+(-0.113165181924)*x[1]
        arg[3,3]=(0.288175924931)*x[0]**o+(-0.489667003521)*x[0]+(-0.531842953651)*x[1]**o+(-0.443517285929)*x[1]
        arg[3,4]=(-0.645574931589)*x[0]**o+(0.464738749544)*x[0]+(-0.538213524368)*x[1]**o+(-0.94443292146)*x[1]
        ref[0,0]=(0.348787631671)*(1+2.*(dim-1.)/(o+1.))+(-0.426847625077)*dim
        ref[0,1]=(-0.356980280382)*(1+2.*(dim-1.)/(o+1.))+(-1.74085159257)*dim
        ref[0,2]=(0.23098037533)*(1+2.*(dim-1.)/(o+1.))+(-0.929592277787)*dim
        ref[0,3]=(0.21760239064)*(1+2.*(dim-1.)/(o+1.))+(0.679664750954)*dim
        ref[0,4]=(-0.554675946648)*(1+2.*(dim-1.)/(o+1.))+(0.96981481714)*dim
        ref[1,0]=(0.266394815412)*(1+2.*(dim-1.)/(o+1.))+(0.0140827211195)*dim
        ref[1,1]=(-0.587017688304)*(1+2.*(dim-1.)/(o+1.))+(-1.71892245996)*dim
        ref[1,2]=(-1.32676389639)*(1+2.*(dim-1.)/(o+1.))+(-0.332986693503)*dim
        ref[1,3]=(-0.901724898919)*(1+2.*(dim-1.)/(o+1.))+(1.48826202022)*dim
        ref[1,4]=(-0.00489823636486)*(1+2.*(dim-1.)/(o+1.))+(-1.16630497024)*dim
        ref[2,0]=(0.506130299295)*(1+2.*(dim-1.)/(o+1.))+(1.43539879264)*dim
        ref[2,1]=(1.52914265193)*(1+2.*(dim-1.)/(o+1.))+(-0.238765918889)*dim
        ref[2,2]=(1.47086040811)*(1+2.*(dim-1.)/(o+1.))+(0.697012433545)*dim
        ref[2,3]=(-0.189881306544)*(1+2.*(dim-1.)/(o+1.))+(0.194735968919)*dim
        ref[2,4]=(-1.00229497107)*(1+2.*(dim-1.)/(o+1.))+(1.23346610482)*dim
        ref[3,0]=(1.27028116431)*(1+2.*(dim-1.)/(o+1.))+(0.475869898897)*dim
        ref[3,1]=(-0.957678237339)*(1+2.*(dim-1.)/(o+1.))+(0.160065172613)*dim
        ref[3,2]=(0.723106432624)*(1+2.*(dim-1.)/(o+1.))+(-0.530167781414)*dim
        ref[3,3]=(-0.243667028719)*(1+2.*(dim-1.)/(o+1.))+(-0.93318428945)*dim
        ref[3,4]=(-1.18378845596)*(1+2.*(dim-1.)/(o+1.))+(-0.479694171917)*dim
      else:
        arg[0,0]=(0.751222317885)*x[0]**o+(0.953666267123)*x[0]+(-0.707417048158)*x[1]**o+(0.310686543483)*x[1]+(0.326267797641)*x[2]**o+(-0.722772500437)*x[2]
        arg[0,1]=(-0.038709086264)*x[0]**o+(-0.483307819428)*x[0]+(0.611120835303)*x[1]**o+(0.885935954604)*x[1]+(0.630199910046)*x[2]**o+(-0.101764126615)*x[2]
        arg[0,2]=(0.890606731439)*x[0]**o+(-0.0555491679283)*x[0]+(0.334835177686)*x[1]**o+(0.999030888839)*x[1]+(-0.891653525154)*x[2]**o+(-0.14056280616)*x[2]
        arg[0,3]=(-0.458305223634)*x[0]**o+(0.679075077126)*x[0]+(0.1176620176)*x[1]**o+(0.685223087684)*x[1]+(-0.952262623464)*x[2]**o+(-0.905568306628)*x[2]
        arg[0,4]=(0.623179050934)*x[0]**o+(-0.0150847487218)*x[0]+(0.0823495403769)*x[1]**o+(0.48303169009)*x[1]+(0.610276253905)*x[2]**o+(0.353104308072)*x[2]
        arg[1,0]=(-0.709185342895)*x[0]**o+(0.789980598884)*x[0]+(0.655538102878)*x[1]**o+(0.713913568814)*x[1]+(0.222131732773)*x[2]**o+(0.712644729381)*x[2]
        arg[1,1]=(-0.683646421822)*x[0]**o+(-0.943691516528)*x[0]+(-0.134261002557)*x[1]**o+(0.28225941018)*x[1]+(0.906235993779)*x[2]**o+(-0.819214603609)*x[2]
        arg[1,2]=(0.713781993938)*x[0]**o+(-0.189551367619)*x[0]+(-0.021310841804)*x[1]**o+(-0.873466428012)*x[1]+(0.248337664546)*x[2]**o+(0.911824874935)*x[2]
        arg[1,3]=(-0.393142772637)*x[0]**o+(0.773501237865)*x[0]+(0.759039429922)*x[1]**o+(0.67036876499)*x[1]+(-0.921411979023)*x[2]**o+(-0.649371534276)*x[2]
        arg[1,4]=(0.526708082723)*x[0]**o+(-0.522449586718)*x[0]+(0.951176548558)*x[1]**o+(0.604602039898)*x[1]+(0.345888192325)*x[2]**o+(-0.973729954334)*x[2]
        arg[2,0]=(-0.139931215293)*x[0]**o+(-0.996072915638)*x[0]+(-0.0451731431264)*x[1]**o+(0.433694224874)*x[1]+(-0.955189760302)*x[2]**o+(-0.68361667242)*x[2]
        arg[2,1]=(0.139626552661)*x[0]**o+(-0.91217468751)*x[0]+(-0.854252204911)*x[1]**o+(0.417539913464)*x[1]+(0.557096050124)*x[2]**o+(-0.195272359868)*x[2]
        arg[2,2]=(0.960352725049)*x[0]**o+(0.98274500341)*x[0]+(0.967782565025)*x[1]**o+(0.566011353381)*x[1]+(0.583885094583)*x[2]**o+(-0.943173847936)*x[2]
        arg[2,3]=(0.959388333494)*x[0]**o+(0.531908797155)*x[0]+(-0.62531126203)*x[1]**o+(-0.288763293646)*x[1]+(-0.823398780526)*x[2]**o+(-0.164348871248)*x[2]
        arg[2,4]=(0.73397702179)*x[0]**o+(0.407844592399)*x[0]+(-0.835769355727)*x[1]**o+(0.742360626394)*x[1]+(0.31130186529)*x[2]**o+(0.114221168872)*x[2]
        arg[3,0]=(-0.42278940444)*x[0]**o+(-0.513158064442)*x[0]+(-0.133233894233)*x[1]**o+(0.701381913887)*x[1]+(0.283776333553)*x[2]**o+(-0.805096661642)*x[2]
        arg[3,1]=(-0.417338508269)*x[0]**o+(0.535961900783)*x[0]+(0.0836386907734)*x[1]**o+(0.449680928704)*x[1]+(0.998998322868)*x[2]**o+(0.983057825698)*x[2]
        arg[3,2]=(0.24102382276)*x[0]**o+(0.0291470157398)*x[0]+(-0.539542451698)*x[1]**o+(0.281017043631)*x[1]+(-0.464652133499)*x[2]**o+(0.652828720185)*x[2]
        arg[3,3]=(-0.482577920013)*x[0]**o+(0.931756756394)*x[0]+(0.96648206338)*x[1]**o+(0.73941459549)*x[1]+(0.136743015429)*x[2]**o+(0.41477914722)*x[2]
        arg[3,4]=(0.879881885972)*x[0]**o+(0.208970079585)*x[0]+(-0.200079958414)*x[1]**o+(-0.0291238772241)*x[1]+(-0.333569505857)*x[2]**o+(0.846835729875)*x[2]
        ref[0,0]=(0.370073067368)*(1+2.*(dim-1.)/(o+1.))+(0.541580310169)*dim
        ref[0,1]=(1.20261165909)*(1+2.*(dim-1.)/(o+1.))+(0.300864008561)*dim
        ref[0,2]=(0.333788383971)*(1+2.*(dim-1.)/(o+1.))+(0.80291891475)*dim
        ref[0,3]=(-1.2929058295)*(1+2.*(dim-1.)/(o+1.))+(0.458729858183)*dim
        ref[0,4]=(1.31580484522)*(1+2.*(dim-1.)/(o+1.))+(0.82105124944)*dim
        ref[1,0]=(0.168484492756)*(1+2.*(dim-1.)/(o+1.))+(2.21653889708)*dim
        ref[1,1]=(0.0883285694004)*(1+2.*(dim-1.)/(o+1.))+(-1.48064670996)*dim
        ref[1,2]=(0.940808816681)*(1+2.*(dim-1.)/(o+1.))+(-0.151192920696)*dim
        ref[1,3]=(-0.555515321739)*(1+2.*(dim-1.)/(o+1.))+(0.79449846858)*dim
        ref[1,4]=(1.82377282361)*(1+2.*(dim-1.)/(o+1.))+(-0.891577501154)*dim
        ref[2,0]=(-1.14029411872)*(1+2.*(dim-1.)/(o+1.))+(-1.24599536318)*dim
        ref[2,1]=(-0.157529602127)*(1+2.*(dim-1.)/(o+1.))+(-0.689907133913)*dim
        ref[2,2]=(2.51202038466)*(1+2.*(dim-1.)/(o+1.))+(0.605582508855)*dim
        ref[2,3]=(-0.489321709063)*(1+2.*(dim-1.)/(o+1.))+(0.0787966322607)*dim
        ref[2,4]=(0.209509531353)*(1+2.*(dim-1.)/(o+1.))+(1.26442638766)*dim
        ref[3,0]=(-0.27224696512)*(1+2.*(dim-1.)/(o+1.))+(-0.616872812198)*dim
        ref[3,1]=(0.665298505372)*(1+2.*(dim-1.)/(o+1.))+(1.96870065519)*dim
        ref[3,2]=(-0.763170762437)*(1+2.*(dim-1.)/(o+1.))+(0.962992779555)*dim
        ref[3,3]=(0.620647158797)*(1+2.*(dim-1.)/(o+1.))+(2.0859504991)*dim
        ref[3,4]=(0.346232421701)*(1+2.*(dim-1.)/(o+1.))+(1.02668193224)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank3(self):
      """
      tests integral of rank 3 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.201115186772)*x[0]**o+(0.324758103748)*x[0]+(0.285065583242)*x[1]**o+(0.0264985832782)*x[1]
        arg[0,0,1]=(-0.443023003973)*x[0]**o+(-0.194732927607)*x[0]+(0.64414908519)*x[1]**o+(-0.248261857779)*x[1]
        arg[0,1,0]=(0.587534172195)*x[0]**o+(0.021088951369)*x[0]+(-0.986195557459)*x[1]**o+(0.595659005312)*x[1]
        arg[0,1,1]=(-0.916961145134)*x[0]**o+(-0.766484469605)*x[0]+(-0.281345964683)*x[1]**o+(-0.399113856826)*x[1]
        arg[1,0,0]=(0.175943424949)*x[0]**o+(0.344853386001)*x[0]+(0.175474681848)*x[1]**o+(-0.0775102874697)*x[1]
        arg[1,0,1]=(-0.874074467007)*x[0]**o+(0.903506914387)*x[0]+(0.177699372538)*x[1]**o+(0.969818154968)*x[1]
        arg[1,1,0]=(0.765751472387)*x[0]**o+(-0.0288724203687)*x[0]+(0.655676828008)*x[1]**o+(0.22281307247)*x[1]
        arg[1,1,1]=(0.871009077152)*x[0]**o+(-0.20052353101)*x[0]+(-0.889496767621)*x[1]**o+(0.8458762887)*x[1]
        arg[2,0,0]=(-0.15377513542)*x[0]**o+(0.963714892768)*x[0]+(0.942514815983)*x[1]**o+(0.846111018838)*x[1]
        arg[2,0,1]=(0.000727086938663)*x[0]**o+(0.243136987492)*x[0]+(-0.308674899886)*x[1]**o+(-0.963435664436)*x[1]
        arg[2,1,0]=(0.780912245051)*x[0]**o+(0.227139041778)*x[0]+(-0.336203258384)*x[1]**o+(0.092568284679)*x[1]
        arg[2,1,1]=(0.0464862681168)*x[0]**o+(0.284493375176)*x[0]+(-0.811073509077)*x[1]**o+(0.54280235419)*x[1]
        arg[3,0,0]=(0.605642304612)*x[0]**o+(-0.103442921567)*x[0]+(0.106253324829)*x[1]**o+(0.562480007904)*x[1]
        arg[3,0,1]=(0.497668538008)*x[0]**o+(-0.859679982844)*x[0]+(-0.255499355194)*x[1]**o+(0.0725919239243)*x[1]
        arg[3,1,0]=(0.641826893201)*x[0]**o+(0.889514581887)*x[0]+(-0.211781857836)*x[1]**o+(0.45900873193)*x[1]
        arg[3,1,1]=(-0.29349600565)*x[0]**o+(-0.627542961817)*x[0]+(0.313009642686)*x[1]**o+(0.338028134809)*x[1]
        arg[4,0,0]=(-0.641915241397)*x[0]**o+(-0.682547311604)*x[0]+(-0.106472077451)*x[1]**o+(0.103402239052)*x[1]
        arg[4,0,1]=(-0.822562520978)*x[0]**o+(-0.430539530156)*x[0]+(-0.0225528898923)*x[1]**o+(-0.0908986665674)*x[1]
        arg[4,1,0]=(-0.456160645978)*x[0]**o+(-0.264592336242)*x[0]+(-0.672702978868)*x[1]**o+(-0.893197955639)*x[1]
        arg[4,1,1]=(-0.714689169676)*x[0]**o+(0.513612105215)*x[0]+(-0.3093340377)*x[1]**o+(-0.34467035156)*x[1]
        arg[5,0,0]=(-0.993761664154)*x[0]**o+(0.375953817044)*x[0]+(0.0463240503072)*x[1]**o+(-0.595646829466)*x[1]
        arg[5,0,1]=(0.453145186545)*x[0]**o+(0.331353840672)*x[0]+(-0.908533078632)*x[1]**o+(0.755175839674)*x[1]
        arg[5,1,0]=(-0.736874540101)*x[0]**o+(-0.867437170846)*x[0]+(-0.472395301622)*x[1]**o+(0.78289019661)*x[1]
        arg[5,1,1]=(-0.900966156016)*x[0]**o+(-0.160363876891)*x[0]+(-0.0969222200475)*x[1]**o+(0.0527983470624)*x[1]
        ref[0,0,0]=(0.486180770014)*(1+2.*(dim-1.)/(o+1.))+(0.351256687026)*dim
        ref[0,0,1]=(0.201126081217)*(1+2.*(dim-1.)/(o+1.))+(-0.442994785386)*dim
        ref[0,1,0]=(-0.398661385264)*(1+2.*(dim-1.)/(o+1.))+(0.616747956681)*dim
        ref[0,1,1]=(-1.19830710982)*(1+2.*(dim-1.)/(o+1.))+(-1.16559832643)*dim
        ref[1,0,0]=(0.351418106797)*(1+2.*(dim-1.)/(o+1.))+(0.267343098531)*dim
        ref[1,0,1]=(-0.696375094469)*(1+2.*(dim-1.)/(o+1.))+(1.87332506936)*dim
        ref[1,1,0]=(1.42142830039)*(1+2.*(dim-1.)/(o+1.))+(0.193940652101)*dim
        ref[1,1,1]=(-0.0184876904693)*(1+2.*(dim-1.)/(o+1.))+(0.64535275769)*dim
        ref[2,0,0]=(0.788739680562)*(1+2.*(dim-1.)/(o+1.))+(1.80982591161)*dim
        ref[2,0,1]=(-0.307947812947)*(1+2.*(dim-1.)/(o+1.))+(-0.720298676944)*dim
        ref[2,1,0]=(0.444708986667)*(1+2.*(dim-1.)/(o+1.))+(0.319707326457)*dim
        ref[2,1,1]=(-0.76458724096)*(1+2.*(dim-1.)/(o+1.))+(0.827295729366)*dim
        ref[3,0,0]=(0.711895629441)*(1+2.*(dim-1.)/(o+1.))+(0.459037086337)*dim
        ref[3,0,1]=(0.242169182814)*(1+2.*(dim-1.)/(o+1.))+(-0.78708805892)*dim
        ref[3,1,0]=(0.430045035365)*(1+2.*(dim-1.)/(o+1.))+(1.34852331382)*dim
        ref[3,1,1]=(0.019513637036)*(1+2.*(dim-1.)/(o+1.))+(-0.289514827008)*dim
        ref[4,0,0]=(-0.748387318848)*(1+2.*(dim-1.)/(o+1.))+(-0.579145072551)*dim
        ref[4,0,1]=(-0.845115410871)*(1+2.*(dim-1.)/(o+1.))+(-0.521438196723)*dim
        ref[4,1,0]=(-1.12886362485)*(1+2.*(dim-1.)/(o+1.))+(-1.15779029188)*dim
        ref[4,1,1]=(-1.02402320738)*(1+2.*(dim-1.)/(o+1.))+(0.168941753655)*dim
        ref[5,0,0]=(-0.947437613847)*(1+2.*(dim-1.)/(o+1.))+(-0.219693012422)*dim
        ref[5,0,1]=(-0.455387892087)*(1+2.*(dim-1.)/(o+1.))+(1.08652968035)*dim
        ref[5,1,0]=(-1.20926984172)*(1+2.*(dim-1.)/(o+1.))+(-0.0845469742352)*dim
        ref[5,1,1]=(-0.997888376064)*(1+2.*(dim-1.)/(o+1.))+(-0.107565529828)*dim
      else:
        arg[0,0,0]=(-0.760764230601)*x[0]**o+(-0.955215307904)*x[0]+(0.376251560231)*x[1]**o+(0.147849416477)*x[1]+(0.627037741745)*x[2]**o+(-0.93996244415)*x[2]
        arg[0,0,1]=(0.904449671196)*x[0]**o+(-0.297804849063)*x[0]+(0.773059038176)*x[1]**o+(-0.836104534497)*x[1]+(-0.150040994258)*x[2]**o+(0.837988726131)*x[2]
        arg[0,1,0]=(0.354255190058)*x[0]**o+(0.463977999833)*x[0]+(0.218855388406)*x[1]**o+(-0.163086423536)*x[1]+(0.653790071045)*x[2]**o+(-0.457721486689)*x[2]
        arg[0,1,1]=(0.357317871424)*x[0]**o+(-0.752862019242)*x[0]+(0.674187130836)*x[1]**o+(-0.983503125785)*x[1]+(-0.776153051078)*x[2]**o+(0.390700377058)*x[2]
        arg[1,0,0]=(-0.465423805042)*x[0]**o+(0.629855541232)*x[0]+(-0.405300083887)*x[1]**o+(-0.211254598072)*x[1]+(0.502581881406)*x[2]**o+(-0.853779249469)*x[2]
        arg[1,0,1]=(0.716025450344)*x[0]**o+(0.876021396077)*x[0]+(-0.530324190586)*x[1]**o+(0.700207449985)*x[1]+(-0.788565669024)*x[2]**o+(0.963118484198)*x[2]
        arg[1,1,0]=(0.813943574582)*x[0]**o+(-0.938778607938)*x[0]+(-0.451619613547)*x[1]**o+(-0.916990884105)*x[1]+(-0.671730738367)*x[2]**o+(0.109664560689)*x[2]
        arg[1,1,1]=(-0.557598711518)*x[0]**o+(-0.541691706474)*x[0]+(0.500874328683)*x[1]**o+(-0.959653364589)*x[1]+(0.572057415183)*x[2]**o+(-0.842306660222)*x[2]
        arg[2,0,0]=(0.0470537168687)*x[0]**o+(-0.151521222488)*x[0]+(0.168279496456)*x[1]**o+(0.869948894317)*x[1]+(-0.554196665086)*x[2]**o+(0.444189090676)*x[2]
        arg[2,0,1]=(-0.662480394662)*x[0]**o+(0.506778773041)*x[0]+(0.664368833417)*x[1]**o+(-0.76728005759)*x[1]+(0.729668063743)*x[2]**o+(-0.476499535761)*x[2]
        arg[2,1,0]=(0.40405640406)*x[0]**o+(0.537691275434)*x[0]+(0.675845932799)*x[1]**o+(-0.663037107454)*x[1]+(-0.947246204559)*x[2]**o+(-0.210977364198)*x[2]
        arg[2,1,1]=(-0.460999541334)*x[0]**o+(-0.403551870206)*x[0]+(-0.59751233801)*x[1]**o+(0.389018763902)*x[1]+(-0.916883519483)*x[2]**o+(0.712714107738)*x[2]
        arg[3,0,0]=(-0.547139864054)*x[0]**o+(0.91774354334)*x[0]+(-0.953248927323)*x[1]**o+(0.693673610785)*x[1]+(-0.474082903895)*x[2]**o+(0.187174904886)*x[2]
        arg[3,0,1]=(0.538597533339)*x[0]**o+(0.903368885629)*x[0]+(-0.548058780044)*x[1]**o+(-0.380406303455)*x[1]+(-0.389244091021)*x[2]**o+(0.137402766164)*x[2]
        arg[3,1,0]=(-0.11691167078)*x[0]**o+(-0.812294829383)*x[0]+(0.592192015269)*x[1]**o+(-0.3278056345)*x[1]+(-0.60462640896)*x[2]**o+(-0.701541632465)*x[2]
        arg[3,1,1]=(-0.196382907034)*x[0]**o+(-0.496234735923)*x[0]+(-0.646502033751)*x[1]**o+(-0.18981379172)*x[1]+(0.627608199527)*x[2]**o+(-0.468659552408)*x[2]
        arg[4,0,0]=(0.56789650708)*x[0]**o+(0.854984079971)*x[0]+(-0.641473205573)*x[1]**o+(-0.791240666887)*x[1]+(0.453562162212)*x[2]**o+(0.849111318036)*x[2]
        arg[4,0,1]=(0.90836228457)*x[0]**o+(-0.666684673376)*x[0]+(0.245928632693)*x[1]**o+(0.313580017685)*x[1]+(0.353126203728)*x[2]**o+(0.467530918033)*x[2]
        arg[4,1,0]=(0.670926585272)*x[0]**o+(0.768555336283)*x[0]+(-0.170656824353)*x[1]**o+(-0.346425800834)*x[1]+(-0.969641886683)*x[2]**o+(0.775335418066)*x[2]
        arg[4,1,1]=(-0.152500967035)*x[0]**o+(0.575400101502)*x[0]+(0.901331752685)*x[1]**o+(-0.672576642993)*x[1]+(0.803176375377)*x[2]**o+(0.202785337179)*x[2]
        arg[5,0,0]=(-0.798851736802)*x[0]**o+(-0.221276048656)*x[0]+(0.773295190339)*x[1]**o+(0.0903112854292)*x[1]+(0.884961364642)*x[2]**o+(0.326429585537)*x[2]
        arg[5,0,1]=(0.482124059007)*x[0]**o+(-0.39227677557)*x[0]+(-0.23074922677)*x[1]**o+(0.986783507045)*x[1]+(0.591759402442)*x[2]**o+(-0.680278617308)*x[2]
        arg[5,1,0]=(-0.857456972644)*x[0]**o+(-0.200661025977)*x[0]+(-0.850031395341)*x[1]**o+(0.830863037447)*x[1]+(-0.323765625026)*x[2]**o+(-0.969463605839)*x[2]
        arg[5,1,1]=(-0.582538216697)*x[0]**o+(-0.495303456972)*x[0]+(0.0134139524699)*x[1]**o+(-0.750144603651)*x[1]+(-0.632380446311)*x[2]**o+(0.38782755017)*x[2]
        ref[0,0,0]=(0.242525071375)*(1+2.*(dim-1.)/(o+1.))+(-1.74732833558)*dim
        ref[0,0,1]=(1.52746771511)*(1+2.*(dim-1.)/(o+1.))+(-0.295920657428)*dim
        ref[0,1,0]=(1.22690064951)*(1+2.*(dim-1.)/(o+1.))+(-0.156829910392)*dim
        ref[0,1,1]=(0.255351951182)*(1+2.*(dim-1.)/(o+1.))+(-1.34566476797)*dim
        ref[1,0,0]=(-0.368142007523)*(1+2.*(dim-1.)/(o+1.))+(-0.43517830631)*dim
        ref[1,0,1]=(-0.602864409266)*(1+2.*(dim-1.)/(o+1.))+(2.53934733026)*dim
        ref[1,1,0]=(-0.309406777332)*(1+2.*(dim-1.)/(o+1.))+(-1.74610493135)*dim
        ref[1,1,1]=(0.515333032348)*(1+2.*(dim-1.)/(o+1.))+(-2.34365173129)*dim
        ref[2,0,0]=(-0.338863451761)*(1+2.*(dim-1.)/(o+1.))+(1.16261676251)*dim
        ref[2,0,1]=(0.731556502498)*(1+2.*(dim-1.)/(o+1.))+(-0.73700082031)*dim
        ref[2,1,0]=(0.132656132299)*(1+2.*(dim-1.)/(o+1.))+(-0.336323196217)*dim
        ref[2,1,1]=(-1.97539539883)*(1+2.*(dim-1.)/(o+1.))+(0.698181001434)*dim
        ref[3,0,0]=(-1.97447169527)*(1+2.*(dim-1.)/(o+1.))+(1.79859205901)*dim
        ref[3,0,1]=(-0.398705337727)*(1+2.*(dim-1.)/(o+1.))+(0.660365348338)*dim
        ref[3,1,0]=(-0.129346064471)*(1+2.*(dim-1.)/(o+1.))+(-1.84164209635)*dim
        ref[3,1,1]=(-0.215276741258)*(1+2.*(dim-1.)/(o+1.))+(-1.15470808005)*dim
        ref[4,0,0]=(0.379985463719)*(1+2.*(dim-1.)/(o+1.))+(0.91285473112)*dim
        ref[4,0,1]=(1.50741712099)*(1+2.*(dim-1.)/(o+1.))+(0.114426262342)*dim
        ref[4,1,0]=(-0.469372125764)*(1+2.*(dim-1.)/(o+1.))+(1.19746495352)*dim
        ref[4,1,1]=(1.55200716103)*(1+2.*(dim-1.)/(o+1.))+(0.105608795688)*dim
        ref[5,0,0]=(0.859404818179)*(1+2.*(dim-1.)/(o+1.))+(0.19546482231)*dim
        ref[5,0,1]=(0.843134234679)*(1+2.*(dim-1.)/(o+1.))+(-0.085771885833)*dim
        ref[5,1,0]=(-2.03125399301)*(1+2.*(dim-1.)/(o+1.))+(-0.339261594369)*dim
        ref[5,1,1]=(-1.20150471054)*(1+2.*(dim-1.)/(o+1.))+(-0.857620510454)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank4(self):
      """
      tests integral of rank 4 Data on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.161068279134)*x[0]**o+(-0.20016693624)*x[0]+(0.360967926263)*x[1]**o+(0.519774730324)*x[1]
        arg[0,0,0,1]=(-0.470257945181)*x[0]**o+(0.0652334206708)*x[0]+(0.652139991357)*x[1]**o+(-0.592482756578)*x[1]
        arg[0,0,1,0]=(0.143126570256)*x[0]**o+(0.683474249212)*x[0]+(0.839611756722)*x[1]**o+(-0.436612603526)*x[1]
        arg[0,0,1,1]=(-0.82328367976)*x[0]**o+(-0.848145278113)*x[0]+(-0.649046694126)*x[1]**o+(0.7686615034)*x[1]
        arg[0,0,2,0]=(-0.89676527056)*x[0]**o+(-0.466316806084)*x[0]+(-0.454908744166)*x[1]**o+(0.425524454897)*x[1]
        arg[0,0,2,1]=(-0.720978656424)*x[0]**o+(-0.655449460498)*x[0]+(0.466383340682)*x[1]**o+(-0.609615749634)*x[1]
        arg[0,1,0,0]=(-0.358670380794)*x[0]**o+(-0.764939003508)*x[0]+(-0.550898156906)*x[1]**o+(0.803150103728)*x[1]
        arg[0,1,0,1]=(-0.652241076457)*x[0]**o+(-0.614033274928)*x[0]+(-0.308770011731)*x[1]**o+(0.925847136965)*x[1]
        arg[0,1,1,0]=(0.710539026622)*x[0]**o+(-0.345587804078)*x[0]+(-0.170026226995)*x[1]**o+(0.340557574804)*x[1]
        arg[0,1,1,1]=(0.348069554895)*x[0]**o+(-0.633629515001)*x[0]+(-0.199233786202)*x[1]**o+(0.500291488845)*x[1]
        arg[0,1,2,0]=(0.262179751197)*x[0]**o+(-0.55460029497)*x[0]+(-0.354335283745)*x[1]**o+(0.301130103531)*x[1]
        arg[0,1,2,1]=(0.212709964525)*x[0]**o+(0.205907518129)*x[0]+(0.167040570615)*x[1]**o+(-0.0684155270612)*x[1]
        arg[0,2,0,0]=(-0.272141616042)*x[0]**o+(0.187045318075)*x[0]+(-0.364974531949)*x[1]**o+(-0.969702579161)*x[1]
        arg[0,2,0,1]=(-0.997155483171)*x[0]**o+(0.864439328971)*x[0]+(0.887907754657)*x[1]**o+(-0.323021835147)*x[1]
        arg[0,2,1,0]=(-0.544228085563)*x[0]**o+(-0.56752289714)*x[0]+(-0.389234395916)*x[1]**o+(-0.969287386144)*x[1]
        arg[0,2,1,1]=(-0.786136450866)*x[0]**o+(0.589465339989)*x[0]+(-0.411595778776)*x[1]**o+(0.726975706536)*x[1]
        arg[0,2,2,0]=(-0.788299350517)*x[0]**o+(0.697037609691)*x[0]+(-0.391342918209)*x[1]**o+(0.432368662365)*x[1]
        arg[0,2,2,1]=(-0.390536554506)*x[0]**o+(0.547528359725)*x[0]+(0.58957966188)*x[1]**o+(0.146918361066)*x[1]
        arg[0,3,0,0]=(0.176059579799)*x[0]**o+(-0.774041865433)*x[0]+(0.52369869825)*x[1]**o+(-0.812474996262)*x[1]
        arg[0,3,0,1]=(-0.938727595513)*x[0]**o+(0.734277246158)*x[0]+(-0.500452658276)*x[1]**o+(0.591562924701)*x[1]
        arg[0,3,1,0]=(-0.109970490351)*x[0]**o+(-0.463382333915)*x[0]+(0.430588389365)*x[1]**o+(-0.174978181104)*x[1]
        arg[0,3,1,1]=(-0.564486046162)*x[0]**o+(0.325599920584)*x[0]+(-0.177334934858)*x[1]**o+(-0.789210879368)*x[1]
        arg[0,3,2,0]=(-0.409769653157)*x[0]**o+(-0.674668709075)*x[0]+(0.977465383205)*x[1]**o+(-0.178249094926)*x[1]
        arg[0,3,2,1]=(0.112394584736)*x[0]**o+(-0.229767891006)*x[0]+(0.952422132907)*x[1]**o+(-0.133952947363)*x[1]
        arg[0,4,0,0]=(0.927747625444)*x[0]**o+(-0.147390883234)*x[0]+(-0.0523168758823)*x[1]**o+(0.698643952882)*x[1]
        arg[0,4,0,1]=(0.986627759385)*x[0]**o+(-0.0486375625834)*x[0]+(0.27598075052)*x[1]**o+(0.970272004318)*x[1]
        arg[0,4,1,0]=(-0.461643441984)*x[0]**o+(-0.815261492927)*x[0]+(-0.0806771439228)*x[1]**o+(0.877302861263)*x[1]
        arg[0,4,1,1]=(-0.108403920958)*x[0]**o+(0.651566419087)*x[0]+(-0.629972296369)*x[1]**o+(-0.235413582155)*x[1]
        arg[0,4,2,0]=(-0.177527213798)*x[0]**o+(0.987202381273)*x[0]+(-0.15522298966)*x[1]**o+(0.234268561378)*x[1]
        arg[0,4,2,1]=(0.521954293269)*x[0]**o+(-0.776898814469)*x[0]+(0.783085515988)*x[1]**o+(-0.469921197982)*x[1]
        arg[1,0,0,0]=(0.858157489751)*x[0]**o+(0.444181035326)*x[0]+(0.3426806552)*x[1]**o+(-0.219249318925)*x[1]
        arg[1,0,0,1]=(0.405857394564)*x[0]**o+(0.178361294933)*x[0]+(0.0879917927852)*x[1]**o+(-0.0689581267347)*x[1]
        arg[1,0,1,0]=(0.63175565694)*x[0]**o+(0.0810462363705)*x[0]+(0.0350399072162)*x[1]**o+(-0.269906961338)*x[1]
        arg[1,0,1,1]=(-0.273973794035)*x[0]**o+(0.824519533391)*x[0]+(0.596785621044)*x[1]**o+(-0.243043370261)*x[1]
        arg[1,0,2,0]=(-0.959246118913)*x[0]**o+(0.3570024829)*x[0]+(-0.260098162669)*x[1]**o+(0.341044029457)*x[1]
        arg[1,0,2,1]=(0.966671029776)*x[0]**o+(0.598898116395)*x[0]+(0.392360541053)*x[1]**o+(-0.322723646751)*x[1]
        arg[1,1,0,0]=(-0.658365963559)*x[0]**o+(-0.674107595463)*x[0]+(-0.0103911551403)*x[1]**o+(0.858879803218)*x[1]
        arg[1,1,0,1]=(0.340459206102)*x[0]**o+(-0.681434505015)*x[0]+(0.917412924129)*x[1]**o+(0.703699154381)*x[1]
        arg[1,1,1,0]=(0.0473840209756)*x[0]**o+(-0.787070458106)*x[0]+(0.11979244907)*x[1]**o+(0.541238063489)*x[1]
        arg[1,1,1,1]=(0.184943768213)*x[0]**o+(-0.942961758488)*x[0]+(-0.844785429901)*x[1]**o+(-0.0109779269179)*x[1]
        arg[1,1,2,0]=(0.921469748172)*x[0]**o+(-0.122684636542)*x[0]+(0.149552532716)*x[1]**o+(-0.0102623511735)*x[1]
        arg[1,1,2,1]=(-0.720466710917)*x[0]**o+(-0.755841224189)*x[0]+(0.372635950892)*x[1]**o+(0.271604227274)*x[1]
        arg[1,2,0,0]=(-0.548557725319)*x[0]**o+(0.745340396995)*x[0]+(0.461556064961)*x[1]**o+(0.696460891668)*x[1]
        arg[1,2,0,1]=(-0.738939238644)*x[0]**o+(0.344829756443)*x[0]+(0.272341172248)*x[1]**o+(0.768063359753)*x[1]
        arg[1,2,1,0]=(-0.213161389496)*x[0]**o+(0.458730188555)*x[0]+(-0.783049463147)*x[1]**o+(0.785273989074)*x[1]
        arg[1,2,1,1]=(0.189092239774)*x[0]**o+(-0.0649917239139)*x[0]+(0.0171281685468)*x[1]**o+(0.397065812267)*x[1]
        arg[1,2,2,0]=(-0.809663598067)*x[0]**o+(0.587214021528)*x[0]+(0.912814601228)*x[1]**o+(0.860792008172)*x[1]
        arg[1,2,2,1]=(0.487927123118)*x[0]**o+(-0.760831928598)*x[0]+(-0.770486346056)*x[1]**o+(-0.715077841171)*x[1]
        arg[1,3,0,0]=(0.0202747853647)*x[0]**o+(0.0666642857036)*x[0]+(-0.717756666598)*x[1]**o+(-0.481009806467)*x[1]
        arg[1,3,0,1]=(-0.251164550146)*x[0]**o+(0.0519935051437)*x[0]+(-0.703400138413)*x[1]**o+(-0.990329903234)*x[1]
        arg[1,3,1,0]=(0.970793375169)*x[0]**o+(-0.365675269998)*x[0]+(0.77706784549)*x[1]**o+(-0.237785293481)*x[1]
        arg[1,3,1,1]=(-0.162786953006)*x[0]**o+(0.985402077919)*x[0]+(0.739483663725)*x[1]**o+(0.559554321882)*x[1]
        arg[1,3,2,0]=(0.184701463663)*x[0]**o+(0.946173117295)*x[0]+(-0.451420616712)*x[1]**o+(0.244506578089)*x[1]
        arg[1,3,2,1]=(0.894307549357)*x[0]**o+(0.474983970689)*x[0]+(0.626273940678)*x[1]**o+(0.107985157225)*x[1]
        arg[1,4,0,0]=(-0.180994231135)*x[0]**o+(-0.899808624585)*x[0]+(0.395935138342)*x[1]**o+(-0.387788497342)*x[1]
        arg[1,4,0,1]=(0.0713358452056)*x[0]**o+(0.505276043139)*x[0]+(-0.635565006093)*x[1]**o+(0.956951408169)*x[1]
        arg[1,4,1,0]=(0.355608577722)*x[0]**o+(0.58482326687)*x[0]+(0.654137690443)*x[1]**o+(-0.849016362888)*x[1]
        arg[1,4,1,1]=(-0.275966957458)*x[0]**o+(0.856095068377)*x[0]+(-0.760077050664)*x[1]**o+(-0.904159432699)*x[1]
        arg[1,4,2,0]=(-0.0637454805096)*x[0]**o+(-0.833057387633)*x[0]+(0.279008913569)*x[1]**o+(-0.093356790193)*x[1]
        arg[1,4,2,1]=(0.165869666472)*x[0]**o+(-0.4554345089)*x[0]+(0.697149094381)*x[1]**o+(0.244460087482)*x[1]
        arg[2,0,0,0]=(0.388188511348)*x[0]**o+(-0.00552398040882)*x[0]+(0.0679336322536)*x[1]**o+(-0.281132460216)*x[1]
        arg[2,0,0,1]=(0.968019690109)*x[0]**o+(0.525335591339)*x[0]+(0.660069087546)*x[1]**o+(-0.795388487336)*x[1]
        arg[2,0,1,0]=(-0.444010450901)*x[0]**o+(-0.498771184141)*x[0]+(-0.394999176962)*x[1]**o+(-0.579230388911)*x[1]
        arg[2,0,1,1]=(-0.96211072138)*x[0]**o+(-0.0977519108256)*x[0]+(0.206179333903)*x[1]**o+(-0.206155312014)*x[1]
        arg[2,0,2,0]=(0.131307090845)*x[0]**o+(0.179396626286)*x[0]+(-0.436118623166)*x[1]**o+(0.829727875456)*x[1]
        arg[2,0,2,1]=(-0.206336435516)*x[0]**o+(0.175744210813)*x[0]+(0.818204199205)*x[1]**o+(0.433292546177)*x[1]
        arg[2,1,0,0]=(0.920189410059)*x[0]**o+(-0.629286762249)*x[0]+(-0.14444305096)*x[1]**o+(0.564022686795)*x[1]
        arg[2,1,0,1]=(0.43550707444)*x[0]**o+(-0.0203691253164)*x[0]+(0.632472807002)*x[1]**o+(0.722390213081)*x[1]
        arg[2,1,1,0]=(0.827084071535)*x[0]**o+(0.864139638834)*x[0]+(0.795893591186)*x[1]**o+(0.0492667773115)*x[1]
        arg[2,1,1,1]=(0.863813787478)*x[0]**o+(0.268800194137)*x[0]+(-0.567637801205)*x[1]**o+(0.931221493563)*x[1]
        arg[2,1,2,0]=(-0.613463525024)*x[0]**o+(-0.0688030533166)*x[0]+(-0.786395692457)*x[1]**o+(0.862189706883)*x[1]
        arg[2,1,2,1]=(0.315306814851)*x[0]**o+(0.76879509223)*x[0]+(0.451451342779)*x[1]**o+(-0.478873421268)*x[1]
        arg[2,2,0,0]=(-0.785116047841)*x[0]**o+(0.347614087057)*x[0]+(-0.0106716042015)*x[1]**o+(-0.350898768187)*x[1]
        arg[2,2,0,1]=(-0.987778931187)*x[0]**o+(-0.380379654757)*x[0]+(-0.383604999655)*x[1]**o+(0.022203803886)*x[1]
        arg[2,2,1,0]=(0.508327115069)*x[0]**o+(0.443678958632)*x[0]+(-0.320069211602)*x[1]**o+(-0.778140184171)*x[1]
        arg[2,2,1,1]=(-0.532025577107)*x[0]**o+(-0.0475799386434)*x[0]+(0.650417175266)*x[1]**o+(0.33739661369)*x[1]
        arg[2,2,2,0]=(-0.406412097333)*x[0]**o+(0.0510500610605)*x[0]+(-0.25755890225)*x[1]**o+(-0.350209779392)*x[1]
        arg[2,2,2,1]=(0.401361048956)*x[0]**o+(-0.594793585281)*x[0]+(0.155257431052)*x[1]**o+(-0.674885216524)*x[1]
        arg[2,3,0,0]=(0.352606222236)*x[0]**o+(0.64529574473)*x[0]+(-0.838258091233)*x[1]**o+(0.527911574536)*x[1]
        arg[2,3,0,1]=(-0.62652610251)*x[0]**o+(0.879664434218)*x[0]+(0.730845316192)*x[1]**o+(0.0318659384791)*x[1]
        arg[2,3,1,0]=(-0.311152313179)*x[0]**o+(-0.333154399786)*x[0]+(0.743640958069)*x[1]**o+(0.313394944493)*x[1]
        arg[2,3,1,1]=(-0.728968478378)*x[0]**o+(0.500229380574)*x[0]+(0.756628133572)*x[1]**o+(0.270533146994)*x[1]
        arg[2,3,2,0]=(-0.219775843272)*x[0]**o+(0.882943220401)*x[0]+(0.79341547188)*x[1]**o+(0.503486054551)*x[1]
        arg[2,3,2,1]=(-0.850285149353)*x[0]**o+(0.78820884302)*x[0]+(0.348024963962)*x[1]**o+(0.83743616537)*x[1]
        arg[2,4,0,0]=(0.868649421127)*x[0]**o+(-0.622361223999)*x[0]+(0.836321867516)*x[1]**o+(0.423582994131)*x[1]
        arg[2,4,0,1]=(0.140963355402)*x[0]**o+(0.190496469385)*x[0]+(0.765311231744)*x[1]**o+(0.368696267095)*x[1]
        arg[2,4,1,0]=(-0.680854080076)*x[0]**o+(-0.729641193897)*x[0]+(-0.843347646585)*x[1]**o+(-0.0993036180842)*x[1]
        arg[2,4,1,1]=(-0.130039648356)*x[0]**o+(0.534197300899)*x[0]+(-0.142588861186)*x[1]**o+(-0.308376644249)*x[1]
        arg[2,4,2,0]=(0.331212825202)*x[0]**o+(0.466247840288)*x[0]+(-0.201714210846)*x[1]**o+(0.80521345364)*x[1]
        arg[2,4,2,1]=(0.454299813247)*x[0]**o+(0.646694941459)*x[0]+(-0.929854585458)*x[1]**o+(-0.365252296746)*x[1]
        arg[3,0,0,0]=(-0.623915395721)*x[0]**o+(-0.789827882448)*x[0]+(-0.862568457189)*x[1]**o+(-0.141799958679)*x[1]
        arg[3,0,0,1]=(0.197040281353)*x[0]**o+(0.155785418081)*x[0]+(-0.587898788817)*x[1]**o+(0.700612189317)*x[1]
        arg[3,0,1,0]=(-0.899654954269)*x[0]**o+(-0.801426785588)*x[0]+(-0.528678401571)*x[1]**o+(-0.224185097206)*x[1]
        arg[3,0,1,1]=(0.291298188198)*x[0]**o+(0.786698690825)*x[0]+(0.925634940697)*x[1]**o+(0.617661715197)*x[1]
        arg[3,0,2,0]=(0.788272473274)*x[0]**o+(-0.0400044705274)*x[0]+(-0.669092974469)*x[1]**o+(-0.063697728491)*x[1]
        arg[3,0,2,1]=(-0.0612398565521)*x[0]**o+(-0.884311783009)*x[0]+(0.444689078628)*x[1]**o+(0.302741981611)*x[1]
        arg[3,1,0,0]=(0.483756249722)*x[0]**o+(-0.214185139167)*x[0]+(0.526387014144)*x[1]**o+(-0.172619975799)*x[1]
        arg[3,1,0,1]=(0.0685836489777)*x[0]**o+(-0.425861440836)*x[0]+(0.94669899123)*x[1]**o+(-0.649068988929)*x[1]
        arg[3,1,1,0]=(0.935225203148)*x[0]**o+(-0.649378812349)*x[0]+(-0.791553458732)*x[1]**o+(0.776307740809)*x[1]
        arg[3,1,1,1]=(-0.45319945403)*x[0]**o+(-0.473310763902)*x[0]+(-0.268480872039)*x[1]**o+(0.847382076174)*x[1]
        arg[3,1,2,0]=(-0.573022434421)*x[0]**o+(0.584143976397)*x[0]+(0.131688457808)*x[1]**o+(-0.628894863435)*x[1]
        arg[3,1,2,1]=(0.628566928363)*x[0]**o+(-0.868564569338)*x[0]+(-0.482273553287)*x[1]**o+(0.679652796519)*x[1]
        arg[3,2,0,0]=(-0.64046509323)*x[0]**o+(-0.108928537952)*x[0]+(-0.512288515479)*x[1]**o+(0.683072290041)*x[1]
        arg[3,2,0,1]=(0.472589328326)*x[0]**o+(0.219741638671)*x[0]+(0.136368535311)*x[1]**o+(-0.525113003647)*x[1]
        arg[3,2,1,0]=(-0.45731256203)*x[0]**o+(0.0854410475382)*x[0]+(0.257725845542)*x[1]**o+(0.430430185351)*x[1]
        arg[3,2,1,1]=(0.948699678725)*x[0]**o+(-0.00538658505415)*x[0]+(0.921898600708)*x[1]**o+(-0.448522286404)*x[1]
        arg[3,2,2,0]=(-0.0863353766204)*x[0]**o+(-0.652417730681)*x[0]+(0.755871413533)*x[1]**o+(0.206065623655)*x[1]
        arg[3,2,2,1]=(0.263075920527)*x[0]**o+(0.94162407402)*x[0]+(0.750015836722)*x[1]**o+(-0.815154267718)*x[1]
        arg[3,3,0,0]=(-0.606984299829)*x[0]**o+(0.120704522147)*x[0]+(-0.324211323778)*x[1]**o+(-0.123647081263)*x[1]
        arg[3,3,0,1]=(0.838995936232)*x[0]**o+(0.854147779134)*x[0]+(-0.192271552952)*x[1]**o+(0.0281108740635)*x[1]
        arg[3,3,1,0]=(-0.0343159887291)*x[0]**o+(0.242432659911)*x[0]+(-0.37299830191)*x[1]**o+(-0.456792158625)*x[1]
        arg[3,3,1,1]=(-0.305277044867)*x[0]**o+(-0.930386582491)*x[0]+(-0.979921197617)*x[1]**o+(-0.507736755766)*x[1]
        arg[3,3,2,0]=(0.548675744195)*x[0]**o+(0.755596482286)*x[0]+(0.0722267054576)*x[1]**o+(-0.393768047035)*x[1]
        arg[3,3,2,1]=(0.74781993508)*x[0]**o+(-0.655033306209)*x[0]+(0.480653847364)*x[1]**o+(-0.169400714085)*x[1]
        arg[3,4,0,0]=(0.269186769422)*x[0]**o+(-0.0786214595922)*x[0]+(-0.975508552024)*x[1]**o+(-0.798249509899)*x[1]
        arg[3,4,0,1]=(-0.599201275441)*x[0]**o+(0.473490319504)*x[0]+(0.0277044125859)*x[1]**o+(0.818107843614)*x[1]
        arg[3,4,1,0]=(-0.46272642862)*x[0]**o+(0.740282148493)*x[0]+(0.35998379499)*x[1]**o+(-0.938681489717)*x[1]
        arg[3,4,1,1]=(0.187067286301)*x[0]**o+(-0.284171344521)*x[0]+(0.612229918327)*x[1]**o+(0.3917984881)*x[1]
        arg[3,4,2,0]=(-0.535270910341)*x[0]**o+(-0.00723888213927)*x[0]+(-0.776984534618)*x[1]**o+(-0.0899536033273)*x[1]
        arg[3,4,2,1]=(-0.382290376155)*x[0]**o+(-0.0593490177599)*x[0]+(-0.19716515678)*x[1]**o+(-0.268991693613)*x[1]
        ref[0,0,0,0]=(0.19989964713)*(1+2.*(dim-1.)/(o+1.))+(0.319607794084)*dim
        ref[0,0,0,1]=(0.181882046176)*(1+2.*(dim-1.)/(o+1.))+(-0.527249335907)*dim
        ref[0,0,1,0]=(0.982738326978)*(1+2.*(dim-1.)/(o+1.))+(0.246861645686)*dim
        ref[0,0,1,1]=(-1.47233037389)*(1+2.*(dim-1.)/(o+1.))+(-0.0794837747136)*dim
        ref[0,0,2,0]=(-1.35167401473)*(1+2.*(dim-1.)/(o+1.))+(-0.0407923511861)*dim
        ref[0,0,2,1]=(-0.254595315742)*(1+2.*(dim-1.)/(o+1.))+(-1.26506521013)*dim
        ref[0,1,0,0]=(-0.909568537701)*(1+2.*(dim-1.)/(o+1.))+(0.0382111002203)*dim
        ref[0,1,0,1]=(-0.961011088187)*(1+2.*(dim-1.)/(o+1.))+(0.311813862038)*dim
        ref[0,1,1,0]=(0.540512799627)*(1+2.*(dim-1.)/(o+1.))+(-0.00503022927381)*dim
        ref[0,1,1,1]=(0.148835768693)*(1+2.*(dim-1.)/(o+1.))+(-0.133338026157)*dim
        ref[0,1,2,0]=(-0.0921555325482)*(1+2.*(dim-1.)/(o+1.))+(-0.253470191439)*dim
        ref[0,1,2,1]=(0.37975053514)*(1+2.*(dim-1.)/(o+1.))+(0.137491991068)*dim
        ref[0,2,0,0]=(-0.637116147991)*(1+2.*(dim-1.)/(o+1.))+(-0.782657261086)*dim
        ref[0,2,0,1]=(-0.109247728514)*(1+2.*(dim-1.)/(o+1.))+(0.541417493824)*dim
        ref[0,2,1,0]=(-0.933462481479)*(1+2.*(dim-1.)/(o+1.))+(-1.53681028328)*dim
        ref[0,2,1,1]=(-1.19773222964)*(1+2.*(dim-1.)/(o+1.))+(1.31644104653)*dim
        ref[0,2,2,0]=(-1.17964226873)*(1+2.*(dim-1.)/(o+1.))+(1.12940627206)*dim
        ref[0,2,2,1]=(0.199043107375)*(1+2.*(dim-1.)/(o+1.))+(0.694446720792)*dim
        ref[0,3,0,0]=(0.699758278049)*(1+2.*(dim-1.)/(o+1.))+(-1.5865168617)*dim
        ref[0,3,0,1]=(-1.43918025379)*(1+2.*(dim-1.)/(o+1.))+(1.32584017086)*dim
        ref[0,3,1,0]=(0.320617899014)*(1+2.*(dim-1.)/(o+1.))+(-0.63836051502)*dim
        ref[0,3,1,1]=(-0.741820981021)*(1+2.*(dim-1.)/(o+1.))+(-0.463610958784)*dim
        ref[0,3,2,0]=(0.567695730048)*(1+2.*(dim-1.)/(o+1.))+(-0.852917804)*dim
        ref[0,3,2,1]=(1.06481671764)*(1+2.*(dim-1.)/(o+1.))+(-0.363720838369)*dim
        ref[0,4,0,0]=(0.875430749562)*(1+2.*(dim-1.)/(o+1.))+(0.551253069648)*dim
        ref[0,4,0,1]=(1.26260850991)*(1+2.*(dim-1.)/(o+1.))+(0.921634441735)*dim
        ref[0,4,1,0]=(-0.542320585907)*(1+2.*(dim-1.)/(o+1.))+(0.0620413683361)*dim
        ref[0,4,1,1]=(-0.738376217327)*(1+2.*(dim-1.)/(o+1.))+(0.416152836932)*dim
        ref[0,4,2,0]=(-0.332750203458)*(1+2.*(dim-1.)/(o+1.))+(1.22147094265)*dim
        ref[0,4,2,1]=(1.30503980926)*(1+2.*(dim-1.)/(o+1.))+(-1.24682001245)*dim
        ref[1,0,0,0]=(1.20083814495)*(1+2.*(dim-1.)/(o+1.))+(0.224931716401)*dim
        ref[1,0,0,1]=(0.493849187349)*(1+2.*(dim-1.)/(o+1.))+(0.109403168198)*dim
        ref[1,0,1,0]=(0.666795564157)*(1+2.*(dim-1.)/(o+1.))+(-0.188860724968)*dim
        ref[1,0,1,1]=(0.322811827009)*(1+2.*(dim-1.)/(o+1.))+(0.58147616313)*dim
        ref[1,0,2,0]=(-1.21934428158)*(1+2.*(dim-1.)/(o+1.))+(0.698046512357)*dim
        ref[1,0,2,1]=(1.35903157083)*(1+2.*(dim-1.)/(o+1.))+(0.276174469644)*dim
        ref[1,1,0,0]=(-0.668757118699)*(1+2.*(dim-1.)/(o+1.))+(0.184772207756)*dim
        ref[1,1,0,1]=(1.25787213023)*(1+2.*(dim-1.)/(o+1.))+(0.0222646493659)*dim
        ref[1,1,1,0]=(0.167176470046)*(1+2.*(dim-1.)/(o+1.))+(-0.245832394617)*dim
        ref[1,1,1,1]=(-0.659841661688)*(1+2.*(dim-1.)/(o+1.))+(-0.953939685406)*dim
        ref[1,1,2,0]=(1.07102228089)*(1+2.*(dim-1.)/(o+1.))+(-0.132946987715)*dim
        ref[1,1,2,1]=(-0.347830760025)*(1+2.*(dim-1.)/(o+1.))+(-0.484236996915)*dim
        ref[1,2,0,0]=(-0.087001660358)*(1+2.*(dim-1.)/(o+1.))+(1.44180128866)*dim
        ref[1,2,0,1]=(-0.466598066397)*(1+2.*(dim-1.)/(o+1.))+(1.1128931162)*dim
        ref[1,2,1,0]=(-0.996210852643)*(1+2.*(dim-1.)/(o+1.))+(1.24400417763)*dim
        ref[1,2,1,1]=(0.206220408321)*(1+2.*(dim-1.)/(o+1.))+(0.332074088353)*dim
        ref[1,2,2,0]=(0.103151003161)*(1+2.*(dim-1.)/(o+1.))+(1.4480060297)*dim
        ref[1,2,2,1]=(-0.282559222939)*(1+2.*(dim-1.)/(o+1.))+(-1.47590976977)*dim
        ref[1,3,0,0]=(-0.697481881234)*(1+2.*(dim-1.)/(o+1.))+(-0.414345520764)*dim
        ref[1,3,0,1]=(-0.954564688559)*(1+2.*(dim-1.)/(o+1.))+(-0.93833639809)*dim
        ref[1,3,1,0]=(1.74786122066)*(1+2.*(dim-1.)/(o+1.))+(-0.603460563479)*dim
        ref[1,3,1,1]=(0.576696710719)*(1+2.*(dim-1.)/(o+1.))+(1.5449563998)*dim
        ref[1,3,2,0]=(-0.26671915305)*(1+2.*(dim-1.)/(o+1.))+(1.19067969538)*dim
        ref[1,3,2,1]=(1.52058149003)*(1+2.*(dim-1.)/(o+1.))+(0.582969127914)*dim
        ref[1,4,0,0]=(0.214940907208)*(1+2.*(dim-1.)/(o+1.))+(-1.28759712193)*dim
        ref[1,4,0,1]=(-0.564229160887)*(1+2.*(dim-1.)/(o+1.))+(1.46222745131)*dim
        ref[1,4,1,0]=(1.00974626816)*(1+2.*(dim-1.)/(o+1.))+(-0.264193096019)*dim
        ref[1,4,1,1]=(-1.03604400812)*(1+2.*(dim-1.)/(o+1.))+(-0.048064364322)*dim
        ref[1,4,2,0]=(0.215263433059)*(1+2.*(dim-1.)/(o+1.))+(-0.926414177826)*dim
        ref[1,4,2,1]=(0.863018760853)*(1+2.*(dim-1.)/(o+1.))+(-0.210974421418)*dim
        ref[2,0,0,0]=(0.456122143601)*(1+2.*(dim-1.)/(o+1.))+(-0.286656440625)*dim
        ref[2,0,0,1]=(1.62808877765)*(1+2.*(dim-1.)/(o+1.))+(-0.270052895997)*dim
        ref[2,0,1,0]=(-0.839009627862)*(1+2.*(dim-1.)/(o+1.))+(-1.07800157305)*dim
        ref[2,0,1,1]=(-0.755931387477)*(1+2.*(dim-1.)/(o+1.))+(-0.30390722284)*dim
        ref[2,0,2,0]=(-0.304811532322)*(1+2.*(dim-1.)/(o+1.))+(1.00912450174)*dim
        ref[2,0,2,1]=(0.61186776369)*(1+2.*(dim-1.)/(o+1.))+(0.60903675699)*dim
        ref[2,1,0,0]=(0.775746359099)*(1+2.*(dim-1.)/(o+1.))+(-0.0652640754538)*dim
        ref[2,1,0,1]=(1.06797988144)*(1+2.*(dim-1.)/(o+1.))+(0.702021087765)*dim
        ref[2,1,1,0]=(1.62297766272)*(1+2.*(dim-1.)/(o+1.))+(0.913406416145)*dim
        ref[2,1,1,1]=(0.296175986273)*(1+2.*(dim-1.)/(o+1.))+(1.2000216877)*dim
        ref[2,1,2,0]=(-1.39985921748)*(1+2.*(dim-1.)/(o+1.))+(0.793386653567)*dim
        ref[2,1,2,1]=(0.76675815763)*(1+2.*(dim-1.)/(o+1.))+(0.289921670962)*dim
        ref[2,2,0,0]=(-0.795787652043)*(1+2.*(dim-1.)/(o+1.))+(-0.0032846811301)*dim
        ref[2,2,0,1]=(-1.37138393084)*(1+2.*(dim-1.)/(o+1.))+(-0.358175850871)*dim
        ref[2,2,1,0]=(0.188257903467)*(1+2.*(dim-1.)/(o+1.))+(-0.334461225539)*dim
        ref[2,2,1,1]=(0.118391598158)*(1+2.*(dim-1.)/(o+1.))+(0.289816675047)*dim
        ref[2,2,2,0]=(-0.663970999582)*(1+2.*(dim-1.)/(o+1.))+(-0.299159718332)*dim
        ref[2,2,2,1]=(0.556618480008)*(1+2.*(dim-1.)/(o+1.))+(-1.2696788018)*dim
        ref[2,3,0,0]=(-0.485651868997)*(1+2.*(dim-1.)/(o+1.))+(1.17320731927)*dim
        ref[2,3,0,1]=(0.104319213682)*(1+2.*(dim-1.)/(o+1.))+(0.911530372697)*dim
        ref[2,3,1,0]=(0.43248864489)*(1+2.*(dim-1.)/(o+1.))+(-0.0197594552935)*dim
        ref[2,3,1,1]=(0.0276596551936)*(1+2.*(dim-1.)/(o+1.))+(0.770762527568)*dim
        ref[2,3,2,0]=(0.573639628608)*(1+2.*(dim-1.)/(o+1.))+(1.38642927495)*dim
        ref[2,3,2,1]=(-0.502260185391)*(1+2.*(dim-1.)/(o+1.))+(1.62564500839)*dim
        ref[2,4,0,0]=(1.70497128864)*(1+2.*(dim-1.)/(o+1.))+(-0.198778229868)*dim
        ref[2,4,0,1]=(0.906274587146)*(1+2.*(dim-1.)/(o+1.))+(0.55919273648)*dim
        ref[2,4,1,0]=(-1.52420172666)*(1+2.*(dim-1.)/(o+1.))+(-0.828944811982)*dim
        ref[2,4,1,1]=(-0.272628509541)*(1+2.*(dim-1.)/(o+1.))+(0.225820656649)*dim
        ref[2,4,2,0]=(0.129498614356)*(1+2.*(dim-1.)/(o+1.))+(1.27146129393)*dim
        ref[2,4,2,1]=(-0.475554772211)*(1+2.*(dim-1.)/(o+1.))+(0.281442644713)*dim
        ref[3,0,0,0]=(-1.48648385291)*(1+2.*(dim-1.)/(o+1.))+(-0.931627841127)*dim
        ref[3,0,0,1]=(-0.390858507465)*(1+2.*(dim-1.)/(o+1.))+(0.856397607398)*dim
        ref[3,0,1,0]=(-1.42833335584)*(1+2.*(dim-1.)/(o+1.))+(-1.02561188279)*dim
        ref[3,0,1,1]=(1.2169331289)*(1+2.*(dim-1.)/(o+1.))+(1.40436040602)*dim
        ref[3,0,2,0]=(0.119179498805)*(1+2.*(dim-1.)/(o+1.))+(-0.103702199018)*dim
        ref[3,0,2,1]=(0.383449222076)*(1+2.*(dim-1.)/(o+1.))+(-0.581569801398)*dim
        ref[3,1,0,0]=(1.01014326387)*(1+2.*(dim-1.)/(o+1.))+(-0.386805114966)*dim
        ref[3,1,0,1]=(1.01528264021)*(1+2.*(dim-1.)/(o+1.))+(-1.07493042977)*dim
        ref[3,1,1,0]=(0.143671744416)*(1+2.*(dim-1.)/(o+1.))+(0.12692892846)*dim
        ref[3,1,1,1]=(-0.721680326068)*(1+2.*(dim-1.)/(o+1.))+(0.374071312272)*dim
        ref[3,1,2,0]=(-0.441333976613)*(1+2.*(dim-1.)/(o+1.))+(-0.0447508870373)*dim
        ref[3,1,2,1]=(0.146293375075)*(1+2.*(dim-1.)/(o+1.))+(-0.188911772819)*dim
        ref[3,2,0,0]=(-1.15275360871)*(1+2.*(dim-1.)/(o+1.))+(0.57414375209)*dim
        ref[3,2,0,1]=(0.608957863637)*(1+2.*(dim-1.)/(o+1.))+(-0.305371364976)*dim
        ref[3,2,1,0]=(-0.199586716488)*(1+2.*(dim-1.)/(o+1.))+(0.515871232889)*dim
        ref[3,2,1,1]=(1.87059827943)*(1+2.*(dim-1.)/(o+1.))+(-0.453908871458)*dim
        ref[3,2,2,0]=(0.669536036913)*(1+2.*(dim-1.)/(o+1.))+(-0.446352107026)*dim
        ref[3,2,2,1]=(1.01309175725)*(1+2.*(dim-1.)/(o+1.))+(0.126469806303)*dim
        ref[3,3,0,0]=(-0.931195623607)*(1+2.*(dim-1.)/(o+1.))+(-0.00294255911617)*dim
        ref[3,3,0,1]=(0.64672438328)*(1+2.*(dim-1.)/(o+1.))+(0.882258653197)*dim
        ref[3,3,1,0]=(-0.407314290639)*(1+2.*(dim-1.)/(o+1.))+(-0.214359498715)*dim
        ref[3,3,1,1]=(-1.28519824248)*(1+2.*(dim-1.)/(o+1.))+(-1.43812333826)*dim
        ref[3,3,2,0]=(0.620902449652)*(1+2.*(dim-1.)/(o+1.))+(0.361828435252)*dim
        ref[3,3,2,1]=(1.22847378244)*(1+2.*(dim-1.)/(o+1.))+(-0.824434020294)*dim
        ref[3,4,0,0]=(-0.706321782603)*(1+2.*(dim-1.)/(o+1.))+(-0.876870969492)*dim
        ref[3,4,0,1]=(-0.571496862855)*(1+2.*(dim-1.)/(o+1.))+(1.29159816312)*dim
        ref[3,4,1,0]=(-0.102742633631)*(1+2.*(dim-1.)/(o+1.))+(-0.198399341224)*dim
        ref[3,4,1,1]=(0.799297204627)*(1+2.*(dim-1.)/(o+1.))+(0.107627143578)*dim
        ref[3,4,2,0]=(-1.31225544496)*(1+2.*(dim-1.)/(o+1.))+(-0.0971924854666)*dim
        ref[3,4,2,1]=(-0.579455532935)*(1+2.*(dim-1.)/(o+1.))+(-0.328340711373)*dim
      else:
        arg[0,0,0,0]=(-0.727570048979)*x[0]**o+(-0.769770353069)*x[0]+(-0.532381054852)*x[1]**o+(0.590589994369)*x[1]+(0.775348726284)*x[2]**o+(-0.840239939796)*x[2]
        arg[0,0,0,1]=(-0.441856577751)*x[0]**o+(-0.39234308235)*x[0]+(-0.834378647798)*x[1]**o+(0.84269918161)*x[1]+(0.441102591728)*x[2]**o+(0.166331058665)*x[2]
        arg[0,0,1,0]=(0.391773069521)*x[0]**o+(0.496759920942)*x[0]+(0.555386226902)*x[1]**o+(0.980197783628)*x[1]+(0.549622021652)*x[2]**o+(-0.596179893504)*x[2]
        arg[0,0,1,1]=(-0.104197387892)*x[0]**o+(-0.64636505375)*x[0]+(0.377612339028)*x[1]**o+(-0.215111619231)*x[1]+(0.637249692625)*x[2]**o+(-0.0998227822351)*x[2]
        arg[0,0,2,0]=(0.780693043561)*x[0]**o+(0.571405344342)*x[0]+(-0.793680154472)*x[1]**o+(-0.0481469136436)*x[1]+(0.959561051293)*x[2]**o+(0.0756152497582)*x[2]
        arg[0,0,2,1]=(0.88678762282)*x[0]**o+(0.254753272214)*x[0]+(0.344376798516)*x[1]**o+(-0.908135902735)*x[1]+(-0.3584387726)*x[2]**o+(-0.890225037617)*x[2]
        arg[0,1,0,0]=(0.680262917473)*x[0]**o+(0.65860357305)*x[0]+(0.859832322473)*x[1]**o+(0.326105580592)*x[1]+(0.132170297876)*x[2]**o+(-0.928606849775)*x[2]
        arg[0,1,0,1]=(-0.500503822293)*x[0]**o+(-0.456331012174)*x[0]+(0.900478846187)*x[1]**o+(0.350542503874)*x[1]+(0.755849038617)*x[2]**o+(-0.514240007142)*x[2]
        arg[0,1,1,0]=(0.124404217291)*x[0]**o+(-0.499651000805)*x[0]+(-0.968178694708)*x[1]**o+(0.306780500818)*x[1]+(-0.946470794819)*x[2]**o+(0.919424575523)*x[2]
        arg[0,1,1,1]=(0.722520017474)*x[0]**o+(-0.718216593718)*x[0]+(0.801196051455)*x[1]**o+(-0.791951278546)*x[1]+(-0.675975417529)*x[2]**o+(0.355069676047)*x[2]
        arg[0,1,2,0]=(-0.137017428857)*x[0]**o+(-0.760267879487)*x[0]+(0.625869259413)*x[1]**o+(-0.697626713352)*x[1]+(-0.467876751423)*x[2]**o+(0.456475454123)*x[2]
        arg[0,1,2,1]=(0.646536442797)*x[0]**o+(0.0981671744166)*x[0]+(0.477147756414)*x[1]**o+(-0.23211823966)*x[1]+(-0.553808167147)*x[2]**o+(0.405402676302)*x[2]
        arg[0,2,0,0]=(-0.0604502966371)*x[0]**o+(0.297995946813)*x[0]+(0.837045433858)*x[1]**o+(-0.270340963575)*x[1]+(-0.58462190484)*x[2]**o+(0.292953993287)*x[2]
        arg[0,2,0,1]=(-0.463109088268)*x[0]**o+(0.346260157368)*x[0]+(0.0228615151352)*x[1]**o+(0.905507471559)*x[1]+(-0.423777553665)*x[2]**o+(-0.45964535654)*x[2]
        arg[0,2,1,0]=(-0.50761294542)*x[0]**o+(-0.465260578419)*x[0]+(-0.729953382373)*x[1]**o+(0.903666219263)*x[1]+(-0.756794344711)*x[2]**o+(0.817263008431)*x[2]
        arg[0,2,1,1]=(-0.551253781393)*x[0]**o+(-0.911481117802)*x[0]+(0.180349330394)*x[1]**o+(0.419118491074)*x[1]+(-0.691917875703)*x[2]**o+(-0.835365504347)*x[2]
        arg[0,2,2,0]=(0.32567932803)*x[0]**o+(0.413167839883)*x[0]+(0.781320691089)*x[1]**o+(-0.777223093057)*x[1]+(0.297556831029)*x[2]**o+(-0.952676550814)*x[2]
        arg[0,2,2,1]=(0.226250474873)*x[0]**o+(0.279287242249)*x[0]+(0.421824124143)*x[1]**o+(-0.97664927412)*x[1]+(0.703245736838)*x[2]**o+(-0.629660622476)*x[2]
        arg[0,3,0,0]=(-0.138403445102)*x[0]**o+(0.0979152923187)*x[0]+(-0.177116218507)*x[1]**o+(0.0100523961708)*x[1]+(-0.634159492232)*x[2]**o+(-0.847618220907)*x[2]
        arg[0,3,0,1]=(0.612268738957)*x[0]**o+(-0.462682392204)*x[0]+(-0.146623835506)*x[1]**o+(0.699258469487)*x[1]+(0.0139983038691)*x[2]**o+(0.236461011719)*x[2]
        arg[0,3,1,0]=(-0.0518123102809)*x[0]**o+(0.386822059696)*x[0]+(-0.208858744655)*x[1]**o+(0.971393362011)*x[1]+(-0.171649562077)*x[2]**o+(-0.507479926411)*x[2]
        arg[0,3,1,1]=(0.0140396270391)*x[0]**o+(-0.837711757872)*x[0]+(-0.00108924459289)*x[1]**o+(-0.724087080496)*x[1]+(-0.793071741694)*x[2]**o+(0.38654936043)*x[2]
        arg[0,3,2,0]=(-0.897711437057)*x[0]**o+(0.562533782571)*x[0]+(-0.354013039442)*x[1]**o+(-0.473034213198)*x[1]+(0.4598333464)*x[2]**o+(0.467503992754)*x[2]
        arg[0,3,2,1]=(-0.198441232877)*x[0]**o+(-0.411568849672)*x[0]+(0.0850211273328)*x[1]**o+(-0.333702582633)*x[1]+(0.727888461746)*x[2]**o+(-0.174835541819)*x[2]
        arg[0,4,0,0]=(-0.114679585752)*x[0]**o+(0.0446596462171)*x[0]+(-0.950604674991)*x[1]**o+(-0.307622186566)*x[1]+(0.956712754366)*x[2]**o+(0.334630480874)*x[2]
        arg[0,4,0,1]=(0.345475911118)*x[0]**o+(0.834620576123)*x[0]+(0.903176273684)*x[1]**o+(-0.364211856466)*x[1]+(0.910903954488)*x[2]**o+(0.392248545744)*x[2]
        arg[0,4,1,0]=(-0.593360797686)*x[0]**o+(-0.0199892857836)*x[0]+(0.293839091284)*x[1]**o+(0.580208854742)*x[1]+(0.060499817751)*x[2]**o+(0.907498841261)*x[2]
        arg[0,4,1,1]=(-0.446192713548)*x[0]**o+(0.29870251614)*x[0]+(0.754231657585)*x[1]**o+(0.321710920549)*x[1]+(0.287978396665)*x[2]**o+(-0.963473165492)*x[2]
        arg[0,4,2,0]=(0.780174386478)*x[0]**o+(0.603377914483)*x[0]+(-0.841703120145)*x[1]**o+(0.316522795462)*x[1]+(-0.192916583378)*x[2]**o+(-0.734091913995)*x[2]
        arg[0,4,2,1]=(-0.130567656486)*x[0]**o+(-0.342928842892)*x[0]+(-0.0827376128775)*x[1]**o+(-0.771104057041)*x[1]+(0.733571283734)*x[2]**o+(-0.880531838119)*x[2]
        arg[1,0,0,0]=(-0.634428121043)*x[0]**o+(0.294135783753)*x[0]+(-0.566305561339)*x[1]**o+(0.0332700472211)*x[1]+(0.128130119463)*x[2]**o+(0.0468678470468)*x[2]
        arg[1,0,0,1]=(0.82168858107)*x[0]**o+(-0.75544177649)*x[0]+(0.260115562827)*x[1]**o+(-0.149575433725)*x[1]+(-0.780945910032)*x[2]**o+(-0.0539291857071)*x[2]
        arg[1,0,1,0]=(-0.97885291605)*x[0]**o+(0.777955971883)*x[0]+(0.115355866293)*x[1]**o+(-0.456449594589)*x[1]+(-0.832816352387)*x[2]**o+(-0.418070257158)*x[2]
        arg[1,0,1,1]=(-0.908396506245)*x[0]**o+(-0.982147653872)*x[0]+(0.138657052347)*x[1]**o+(0.633837624386)*x[1]+(-0.529748303776)*x[2]**o+(-0.950225664343)*x[2]
        arg[1,0,2,0]=(0.733974435377)*x[0]**o+(-0.758228777635)*x[0]+(0.841368499977)*x[1]**o+(0.871715889444)*x[1]+(0.541055759071)*x[2]**o+(0.669798500731)*x[2]
        arg[1,0,2,1]=(-0.572994548563)*x[0]**o+(0.806241853906)*x[0]+(0.141217952929)*x[1]**o+(0.832291361477)*x[1]+(0.207737014762)*x[2]**o+(-0.279769695917)*x[2]
        arg[1,1,0,0]=(-0.394247339923)*x[0]**o+(0.548338039327)*x[0]+(-0.845822483858)*x[1]**o+(-0.941161341738)*x[1]+(-0.244513027249)*x[2]**o+(-0.524790843131)*x[2]
        arg[1,1,0,1]=(0.509994957918)*x[0]**o+(0.691470630092)*x[0]+(0.249241759795)*x[1]**o+(-0.995293157716)*x[1]+(0.162652088127)*x[2]**o+(0.873578806104)*x[2]
        arg[1,1,1,0]=(-0.22367137056)*x[0]**o+(-0.896306406631)*x[0]+(0.613214186844)*x[1]**o+(-0.460055562163)*x[1]+(-0.704736634406)*x[2]**o+(-0.517741014186)*x[2]
        arg[1,1,1,1]=(-0.426072795689)*x[0]**o+(-0.13452184053)*x[0]+(0.908778671123)*x[1]**o+(0.194472828672)*x[1]+(-0.658909495204)*x[2]**o+(-0.234260758459)*x[2]
        arg[1,1,2,0]=(0.236086562474)*x[0]**o+(0.33312461561)*x[0]+(0.867089643801)*x[1]**o+(-0.720157305369)*x[1]+(0.812845198077)*x[2]**o+(-0.375694713863)*x[2]
        arg[1,1,2,1]=(-0.561145817489)*x[0]**o+(0.746630137163)*x[0]+(-0.809280840008)*x[1]**o+(-0.690912076356)*x[1]+(0.88933641231)*x[2]**o+(0.246647817564)*x[2]
        arg[1,2,0,0]=(0.701319152729)*x[0]**o+(-0.309074024603)*x[0]+(0.567988628717)*x[1]**o+(-0.0596939135728)*x[1]+(0.242135294918)*x[2]**o+(-0.101136005445)*x[2]
        arg[1,2,0,1]=(0.880665674235)*x[0]**o+(-0.833066777852)*x[0]+(-0.0556340786728)*x[1]**o+(0.620095583249)*x[1]+(-0.390556829051)*x[2]**o+(0.793481461864)*x[2]
        arg[1,2,1,0]=(0.832046852431)*x[0]**o+(0.4276990324)*x[0]+(-0.486292331356)*x[1]**o+(0.990296730852)*x[1]+(-0.381732943927)*x[2]**o+(0.225838720655)*x[2]
        arg[1,2,1,1]=(-0.971477034333)*x[0]**o+(0.819409035213)*x[0]+(-0.610341797903)*x[1]**o+(0.842186877805)*x[1]+(0.571187503044)*x[2]**o+(-0.150565006374)*x[2]
        arg[1,2,2,0]=(0.266080624706)*x[0]**o+(0.893377400845)*x[0]+(0.718525426441)*x[1]**o+(-0.847615622286)*x[1]+(0.614914866)*x[2]**o+(0.826503774974)*x[2]
        arg[1,2,2,1]=(0.998060871997)*x[0]**o+(-0.572523112154)*x[0]+(-0.786785614822)*x[1]**o+(0.556952019953)*x[1]+(0.804849394796)*x[2]**o+(-0.201157281118)*x[2]
        arg[1,3,0,0]=(-0.322048105751)*x[0]**o+(0.215441634353)*x[0]+(0.203298400258)*x[1]**o+(0.452071048161)*x[1]+(0.0134315902731)*x[2]**o+(0.240098266028)*x[2]
        arg[1,3,0,1]=(-0.109845136841)*x[0]**o+(0.0479663457074)*x[0]+(0.899394604931)*x[1]**o+(0.853608214151)*x[1]+(-0.27487885579)*x[2]**o+(-0.0236702866567)*x[2]
        arg[1,3,1,0]=(0.948550474293)*x[0]**o+(0.359077475995)*x[0]+(0.655387467125)*x[1]**o+(-0.0501114084308)*x[1]+(-0.244470769181)*x[2]**o+(0.670276065978)*x[2]
        arg[1,3,1,1]=(-0.860769147822)*x[0]**o+(-0.738449680804)*x[0]+(0.840959189643)*x[1]**o+(0.229930283198)*x[1]+(-0.521034843253)*x[2]**o+(-0.50213233896)*x[2]
        arg[1,3,2,0]=(0.29910216992)*x[0]**o+(0.411953562132)*x[0]+(0.777584507308)*x[1]**o+(0.41482951697)*x[1]+(-0.948751119313)*x[2]**o+(-0.526573791564)*x[2]
        arg[1,3,2,1]=(-0.768167653244)*x[0]**o+(-0.318762199437)*x[0]+(0.7566706966)*x[1]**o+(-0.151347269999)*x[1]+(-0.626690637485)*x[2]**o+(-0.731518248609)*x[2]
        arg[1,4,0,0]=(0.0147733854143)*x[0]**o+(-0.812988077571)*x[0]+(0.815385483312)*x[1]**o+(0.285326283687)*x[1]+(0.872502797762)*x[2]**o+(-0.748392032843)*x[2]
        arg[1,4,0,1]=(0.172931930795)*x[0]**o+(0.425088340505)*x[0]+(0.0122969950852)*x[1]**o+(0.556865378995)*x[1]+(0.614686173604)*x[2]**o+(0.966581266651)*x[2]
        arg[1,4,1,0]=(0.725270959871)*x[0]**o+(-0.672683754253)*x[0]+(0.486696038678)*x[1]**o+(-0.53856418144)*x[1]+(-0.295197744155)*x[2]**o+(0.0610876962829)*x[2]
        arg[1,4,1,1]=(0.942300770794)*x[0]**o+(0.876440453364)*x[0]+(0.432652207093)*x[1]**o+(-0.562629961162)*x[1]+(0.497526473424)*x[2]**o+(-0.165824324383)*x[2]
        arg[1,4,2,0]=(-0.668927175078)*x[0]**o+(0.587283304457)*x[0]+(0.588305038304)*x[1]**o+(0.087906366831)*x[1]+(-0.907721198172)*x[2]**o+(-0.506067369924)*x[2]
        arg[1,4,2,1]=(0.899071894553)*x[0]**o+(0.920678485704)*x[0]+(-0.244107214117)*x[1]**o+(0.196327093214)*x[1]+(-0.538075261013)*x[2]**o+(-0.925646646321)*x[2]
        arg[2,0,0,0]=(-0.113452114076)*x[0]**o+(0.541178389983)*x[0]+(0.501083424074)*x[1]**o+(-0.234654083639)*x[1]+(-0.435569149263)*x[2]**o+(0.671517947867)*x[2]
        arg[2,0,0,1]=(-0.00139653535956)*x[0]**o+(-0.93731587459)*x[0]+(-0.387786776826)*x[1]**o+(0.361445822788)*x[1]+(-0.0958490804452)*x[2]**o+(0.22475163005)*x[2]
        arg[2,0,1,0]=(0.100832420019)*x[0]**o+(-0.957577198313)*x[0]+(0.244479124631)*x[1]**o+(-0.589295650233)*x[1]+(-0.11020792463)*x[2]**o+(0.969023424875)*x[2]
        arg[2,0,1,1]=(-0.82757398282)*x[0]**o+(0.665274113482)*x[0]+(0.219624146627)*x[1]**o+(0.604687589384)*x[1]+(0.539774513376)*x[2]**o+(0.676136495778)*x[2]
        arg[2,0,2,0]=(-0.869539421193)*x[0]**o+(0.449645390105)*x[0]+(-0.216356671763)*x[1]**o+(-0.191616800811)*x[1]+(-0.324126591392)*x[2]**o+(0.726527544226)*x[2]
        arg[2,0,2,1]=(-0.522340761977)*x[0]**o+(-0.253136332372)*x[0]+(-0.993693358789)*x[1]**o+(0.689338979208)*x[1]+(-0.73030598892)*x[2]**o+(0.604868622965)*x[2]
        arg[2,1,0,0]=(-0.87285436276)*x[0]**o+(-0.665917385941)*x[0]+(0.105570880359)*x[1]**o+(0.773780432278)*x[1]+(-0.0932677950401)*x[2]**o+(0.55563257373)*x[2]
        arg[2,1,0,1]=(-0.717823895492)*x[0]**o+(-0.843992380543)*x[0]+(0.348730031558)*x[1]**o+(-0.0641727895105)*x[1]+(-0.575332519937)*x[2]**o+(0.566731137573)*x[2]
        arg[2,1,1,0]=(-0.749216309273)*x[0]**o+(0.945232679736)*x[0]+(-0.984694983452)*x[1]**o+(-0.697422850579)*x[1]+(0.573407778839)*x[2]**o+(0.654548962131)*x[2]
        arg[2,1,1,1]=(-0.951637762959)*x[0]**o+(-0.0299724748286)*x[0]+(-0.784527116834)*x[1]**o+(0.490138786409)*x[1]+(0.945743565011)*x[2]**o+(0.158024494836)*x[2]
        arg[2,1,2,0]=(0.80439224967)*x[0]**o+(0.889509225531)*x[0]+(-0.178335319647)*x[1]**o+(-0.365669987413)*x[1]+(-0.170055755127)*x[2]**o+(0.258843711555)*x[2]
        arg[2,1,2,1]=(0.520706901588)*x[0]**o+(0.897178241272)*x[0]+(-0.128495163664)*x[1]**o+(-0.404742110163)*x[1]+(0.140333496135)*x[2]**o+(0.31429703244)*x[2]
        arg[2,2,0,0]=(-0.933201442136)*x[0]**o+(0.0083650170184)*x[0]+(-0.281807642508)*x[1]**o+(-0.09193002721)*x[1]+(-0.422582781227)*x[2]**o+(0.358915935551)*x[2]
        arg[2,2,0,1]=(-0.09712053324)*x[0]**o+(-0.0752448353447)*x[0]+(0.41291697894)*x[1]**o+(0.593823615119)*x[1]+(0.954183095989)*x[2]**o+(0.479576723309)*x[2]
        arg[2,2,1,0]=(-0.48592273499)*x[0]**o+(-0.555807193072)*x[0]+(0.842926147107)*x[1]**o+(-0.0760474608316)*x[1]+(0.864590424311)*x[2]**o+(-0.854721691042)*x[2]
        arg[2,2,1,1]=(0.238261310086)*x[0]**o+(-0.506114025767)*x[0]+(-0.534154656267)*x[1]**o+(-0.282262066272)*x[1]+(-0.241950678835)*x[2]**o+(-0.596913007205)*x[2]
        arg[2,2,2,0]=(0.35002425052)*x[0]**o+(-0.724235915518)*x[0]+(-0.443311896206)*x[1]**o+(0.0183169190559)*x[1]+(-0.180995873436)*x[2]**o+(0.949105760518)*x[2]
        arg[2,2,2,1]=(-0.524596480672)*x[0]**o+(-0.72511087285)*x[0]+(-0.793159588655)*x[1]**o+(0.197579071177)*x[1]+(0.214510478855)*x[2]**o+(0.592011219443)*x[2]
        arg[2,3,0,0]=(-0.777144207151)*x[0]**o+(0.821815068413)*x[0]+(0.828183715461)*x[1]**o+(-0.77742744879)*x[1]+(0.520088528639)*x[2]**o+(0.564796500907)*x[2]
        arg[2,3,0,1]=(-0.444252706506)*x[0]**o+(-0.999658246414)*x[0]+(0.458490846376)*x[1]**o+(0.206371738627)*x[1]+(-0.5943830165)*x[2]**o+(0.125954475804)*x[2]
        arg[2,3,1,0]=(-0.499555379078)*x[0]**o+(0.667767779032)*x[0]+(-0.169053599596)*x[1]**o+(0.587608309525)*x[1]+(-0.99317016838)*x[2]**o+(-0.580232194372)*x[2]
        arg[2,3,1,1]=(0.881736170108)*x[0]**o+(-0.0221128885544)*x[0]+(-0.239202356219)*x[1]**o+(-0.825250840397)*x[1]+(-0.34138233902)*x[2]**o+(-0.900437209752)*x[2]
        arg[2,3,2,0]=(-0.810415345283)*x[0]**o+(0.671639285402)*x[0]+(0.0577696593549)*x[1]**o+(0.12466039802)*x[1]+(-0.164952472389)*x[2]**o+(-0.881356759483)*x[2]
        arg[2,3,2,1]=(0.641411687742)*x[0]**o+(0.939465110042)*x[0]+(0.182541049996)*x[1]**o+(-0.502132165421)*x[1]+(-0.914212670088)*x[2]**o+(0.361356243622)*x[2]
        arg[2,4,0,0]=(0.696033328872)*x[0]**o+(0.854329989895)*x[0]+(0.66092830442)*x[1]**o+(-0.165623528752)*x[1]+(0.587322879532)*x[2]**o+(-0.441145527759)*x[2]
        arg[2,4,0,1]=(0.885589185979)*x[0]**o+(0.170856006811)*x[0]+(-0.500114387063)*x[1]**o+(-0.772574748562)*x[1]+(0.599300325869)*x[2]**o+(0.0950082175327)*x[2]
        arg[2,4,1,0]=(0.877739402691)*x[0]**o+(0.494715865857)*x[0]+(-0.4548916845)*x[1]**o+(-0.108381552398)*x[1]+(-0.620522177203)*x[2]**o+(-0.375546939131)*x[2]
        arg[2,4,1,1]=(-0.108044365878)*x[0]**o+(-0.592058059304)*x[0]+(-0.64481990676)*x[1]**o+(-0.903674017194)*x[1]+(0.0964052945913)*x[2]**o+(0.357819426568)*x[2]
        arg[2,4,2,0]=(-0.686157836559)*x[0]**o+(0.585514351597)*x[0]+(-0.5623443108)*x[1]**o+(-0.636059380122)*x[1]+(0.469863235081)*x[2]**o+(0.374567303622)*x[2]
        arg[2,4,2,1]=(0.214800008136)*x[0]**o+(0.0680448664819)*x[0]+(0.722930302776)*x[1]**o+(0.107573165145)*x[1]+(0.0027804595356)*x[2]**o+(0.725566605544)*x[2]
        arg[3,0,0,0]=(-0.273256791086)*x[0]**o+(0.801129507153)*x[0]+(-0.840227644476)*x[1]**o+(-0.31594258736)*x[1]+(-0.140806177065)*x[2]**o+(-0.0958103115672)*x[2]
        arg[3,0,0,1]=(0.588534583503)*x[0]**o+(0.242235627811)*x[0]+(-0.541539488722)*x[1]**o+(0.624385574868)*x[1]+(-0.0641371754789)*x[2]**o+(-0.635676751281)*x[2]
        arg[3,0,1,0]=(0.81967100218)*x[0]**o+(0.671597165436)*x[0]+(0.160179001916)*x[1]**o+(-0.226714136475)*x[1]+(-0.523709112534)*x[2]**o+(0.932797007296)*x[2]
        arg[3,0,1,1]=(0.707793115256)*x[0]**o+(-0.0995173271712)*x[0]+(-0.594597614398)*x[1]**o+(-0.72096550461)*x[1]+(-0.373631371949)*x[2]**o+(-0.282151921803)*x[2]
        arg[3,0,2,0]=(0.873307297971)*x[0]**o+(0.0950091191819)*x[0]+(-0.534108539131)*x[1]**o+(0.567381193719)*x[1]+(-0.276691140405)*x[2]**o+(0.84882104254)*x[2]
        arg[3,0,2,1]=(-0.895148878411)*x[0]**o+(-0.880973451462)*x[0]+(0.105602819158)*x[1]**o+(-0.836475687487)*x[1]+(-0.18380164164)*x[2]**o+(-0.391643939001)*x[2]
        arg[3,1,0,0]=(-0.144070349397)*x[0]**o+(-0.247578520564)*x[0]+(0.512469248899)*x[1]**o+(0.0369808259097)*x[1]+(0.67517925106)*x[2]**o+(-0.306122403522)*x[2]
        arg[3,1,0,1]=(0.389023591419)*x[0]**o+(0.0335324297663)*x[0]+(0.166380432563)*x[1]**o+(0.415133512101)*x[1]+(-0.956055398028)*x[2]**o+(-0.069254560597)*x[2]
        arg[3,1,1,0]=(-0.253194755795)*x[0]**o+(-0.432240637655)*x[0]+(0.886730830871)*x[1]**o+(0.94685399231)*x[1]+(-0.0193581886643)*x[2]**o+(-0.828124732153)*x[2]
        arg[3,1,1,1]=(0.948739050057)*x[0]**o+(-0.200402018913)*x[0]+(0.062573402994)*x[1]**o+(-0.408727698463)*x[1]+(-0.0334411087309)*x[2]**o+(-0.676147119957)*x[2]
        arg[3,1,2,0]=(0.249237167322)*x[0]**o+(-0.621387879739)*x[0]+(-0.745335388751)*x[1]**o+(-0.183173004074)*x[1]+(0.56632553768)*x[2]**o+(-0.519909176525)*x[2]
        arg[3,1,2,1]=(-0.660822495815)*x[0]**o+(0.0459275869317)*x[0]+(-0.0853936912232)*x[1]**o+(-0.895502076207)*x[1]+(0.150013238845)*x[2]**o+(0.201527905115)*x[2]
        arg[3,2,0,0]=(0.0846880899978)*x[0]**o+(-0.450721252185)*x[0]+(0.0100224498309)*x[1]**o+(-0.729399043167)*x[1]+(0.249073157161)*x[2]**o+(-0.313056982897)*x[2]
        arg[3,2,0,1]=(0.349123663394)*x[0]**o+(-0.0439266646004)*x[0]+(0.201749457716)*x[1]**o+(-0.29961557135)*x[1]+(0.247657591065)*x[2]**o+(0.702448030656)*x[2]
        arg[3,2,1,0]=(-0.67551758529)*x[0]**o+(-0.856271486186)*x[0]+(-0.337698270178)*x[1]**o+(-0.855526501932)*x[1]+(0.204628923051)*x[2]**o+(0.0425229547861)*x[2]
        arg[3,2,1,1]=(0.913895958603)*x[0]**o+(0.33488814727)*x[0]+(0.725768429772)*x[1]**o+(0.31487179035)*x[1]+(-0.926895768372)*x[2]**o+(-0.303623128744)*x[2]
        arg[3,2,2,0]=(0.695497510732)*x[0]**o+(-0.71819800795)*x[0]+(0.0949892877746)*x[1]**o+(-0.371952684199)*x[1]+(-0.712679288901)*x[2]**o+(-0.839032166941)*x[2]
        arg[3,2,2,1]=(0.847915632309)*x[0]**o+(0.489320767225)*x[0]+(0.884694888545)*x[1]**o+(-0.836734066966)*x[1]+(-0.782643394677)*x[2]**o+(0.998012000062)*x[2]
        arg[3,3,0,0]=(0.400343531732)*x[0]**o+(-0.977399566427)*x[0]+(-0.0491815761683)*x[1]**o+(0.820343030844)*x[1]+(0.145276965371)*x[2]**o+(0.106181410329)*x[2]
        arg[3,3,0,1]=(0.39509119571)*x[0]**o+(-0.532152199068)*x[0]+(0.0961163031465)*x[1]**o+(-0.845734786516)*x[1]+(0.0685207262665)*x[2]**o+(0.10725771579)*x[2]
        arg[3,3,1,0]=(-0.326475146957)*x[0]**o+(0.0491193078417)*x[0]+(-0.193405294452)*x[1]**o+(0.283579426128)*x[1]+(0.885442239373)*x[2]**o+(-0.838544460442)*x[2]
        arg[3,3,1,1]=(-0.502234549601)*x[0]**o+(-0.764617462736)*x[0]+(0.52701849705)*x[1]**o+(-0.423245302614)*x[1]+(0.728096320995)*x[2]**o+(0.143356081954)*x[2]
        arg[3,3,2,0]=(0.924562662058)*x[0]**o+(-0.0120799887424)*x[0]+(-0.342886608544)*x[1]**o+(0.609435784717)*x[1]+(0.392296911228)*x[2]**o+(0.572866548262)*x[2]
        arg[3,3,2,1]=(0.661267105524)*x[0]**o+(-0.71847134372)*x[0]+(-0.913866381825)*x[1]**o+(-0.502556458508)*x[1]+(0.201214639904)*x[2]**o+(-0.431939125167)*x[2]
        arg[3,4,0,0]=(0.391785002874)*x[0]**o+(0.424885787931)*x[0]+(-0.377558346944)*x[1]**o+(-0.556924666311)*x[1]+(0.511852194353)*x[2]**o+(0.514261734074)*x[2]
        arg[3,4,0,1]=(-0.269148156354)*x[0]**o+(0.692033978804)*x[0]+(-0.390604590454)*x[1]**o+(-0.551681907742)*x[1]+(0.0613720420346)*x[2]**o+(-0.508841982991)*x[2]
        arg[3,4,1,0]=(0.593170387694)*x[0]**o+(-0.00676980821753)*x[0]+(0.788564472913)*x[1]**o+(0.823094059865)*x[1]+(0.52895485971)*x[2]**o+(0.88849618397)*x[2]
        arg[3,4,1,1]=(0.772185283881)*x[0]**o+(0.759014894154)*x[0]+(-0.142739578724)*x[1]**o+(-0.141899255898)*x[1]+(0.479086662114)*x[2]**o+(0.68993908918)*x[2]
        arg[3,4,2,0]=(0.998097253401)*x[0]**o+(0.621254339576)*x[0]+(-0.692540865624)*x[1]**o+(0.716545091382)*x[1]+(0.478209703594)*x[2]**o+(0.259063088226)*x[2]
        arg[3,4,2,1]=(-0.321962318028)*x[0]**o+(0.0377531939042)*x[0]+(-0.362041758315)*x[1]**o+(0.632474956936)*x[1]+(0.133972776903)*x[2]**o+(-0.341803258375)*x[2]
        ref[0,0,0,0]=(-0.484602377547)*(1+2.*(dim-1.)/(o+1.))+(-1.0194202985)*dim
        ref[0,0,0,1]=(-0.835132633821)*(1+2.*(dim-1.)/(o+1.))+(0.616687157924)*dim
        ref[0,0,1,0]=(1.49678131808)*(1+2.*(dim-1.)/(o+1.))+(0.880777811066)*dim
        ref[0,0,1,1]=(0.910664643761)*(1+2.*(dim-1.)/(o+1.))+(-0.961299455217)*dim
        ref[0,0,2,0]=(0.946573940383)*(1+2.*(dim-1.)/(o+1.))+(0.598873680457)*dim
        ref[0,0,2,1]=(0.872725648736)*(1+2.*(dim-1.)/(o+1.))+(-1.54360766814)*dim
        ref[0,1,0,0]=(1.67226553782)*(1+2.*(dim-1.)/(o+1.))+(0.0561023038674)*dim
        ref[0,1,0,1]=(1.15582406251)*(1+2.*(dim-1.)/(o+1.))+(-0.620028515442)*dim
        ref[0,1,1,0]=(-1.79024527224)*(1+2.*(dim-1.)/(o+1.))+(0.726554075535)*dim
        ref[0,1,1,1]=(0.8477406514)*(1+2.*(dim-1.)/(o+1.))+(-1.15509819622)*dim
        ref[0,1,2,0]=(0.0209750791326)*(1+2.*(dim-1.)/(o+1.))+(-1.00141913872)*dim
        ref[0,1,2,1]=(0.569876032064)*(1+2.*(dim-1.)/(o+1.))+(0.271451611059)*dim
        ref[0,2,0,0]=(0.19197323238)*(1+2.*(dim-1.)/(o+1.))+(0.320608976525)*dim
        ref[0,2,0,1]=(-0.864025126798)*(1+2.*(dim-1.)/(o+1.))+(0.792122272387)*dim
        ref[0,2,1,0]=(-1.9943606725)*(1+2.*(dim-1.)/(o+1.))+(1.25566864927)*dim
        ref[0,2,1,1]=(-1.0628223267)*(1+2.*(dim-1.)/(o+1.))+(-1.32772813107)*dim
        ref[0,2,2,0]=(1.40455685015)*(1+2.*(dim-1.)/(o+1.))+(-1.31673180399)*dim
        ref[0,2,2,1]=(1.35132033585)*(1+2.*(dim-1.)/(o+1.))+(-1.32702265435)*dim
        ref[0,3,0,0]=(-0.949679155841)*(1+2.*(dim-1.)/(o+1.))+(-0.739650532418)*dim
        ref[0,3,0,1]=(0.47964320732)*(1+2.*(dim-1.)/(o+1.))+(0.473037089003)*dim
        ref[0,3,1,0]=(-0.432320617013)*(1+2.*(dim-1.)/(o+1.))+(0.850735495297)*dim
        ref[0,3,1,1]=(-0.780121359247)*(1+2.*(dim-1.)/(o+1.))+(-1.17524947794)*dim
        ref[0,3,2,0]=(-0.791891130098)*(1+2.*(dim-1.)/(o+1.))+(0.557003562128)*dim
        ref[0,3,2,1]=(0.614468356202)*(1+2.*(dim-1.)/(o+1.))+(-0.920106974125)*dim
        ref[0,4,0,0]=(-0.108571506377)*(1+2.*(dim-1.)/(o+1.))+(0.0716679405256)*dim
        ref[0,4,0,1]=(2.15955613929)*(1+2.*(dim-1.)/(o+1.))+(0.862657265402)*dim
        ref[0,4,1,0]=(-0.239021888652)*(1+2.*(dim-1.)/(o+1.))+(1.46771841022)*dim
        ref[0,4,1,1]=(0.596017340702)*(1+2.*(dim-1.)/(o+1.))+(-0.343059728803)*dim
        ref[0,4,2,0]=(-0.254445317045)*(1+2.*(dim-1.)/(o+1.))+(0.185808795951)*dim
        ref[0,4,2,1]=(0.520266014371)*(1+2.*(dim-1.)/(o+1.))+(-1.99456473805)*dim
        ref[1,0,0,0]=(-1.07260356292)*(1+2.*(dim-1.)/(o+1.))+(0.374273678021)*dim
        ref[1,0,0,1]=(0.300858233864)*(1+2.*(dim-1.)/(o+1.))+(-0.958946395922)*dim
        ref[1,0,1,0]=(-1.69631340214)*(1+2.*(dim-1.)/(o+1.))+(-0.096563879864)*dim
        ref[1,0,1,1]=(-1.29948775767)*(1+2.*(dim-1.)/(o+1.))+(-1.29853569383)*dim
        ref[1,0,2,0]=(2.11639869443)*(1+2.*(dim-1.)/(o+1.))+(0.783285612541)*dim
        ref[1,0,2,1]=(-0.224039580872)*(1+2.*(dim-1.)/(o+1.))+(1.35876351947)*dim
        ref[1,1,0,0]=(-1.48458285103)*(1+2.*(dim-1.)/(o+1.))+(-0.917614145542)*dim
        ref[1,1,0,1]=(0.92188880584)*(1+2.*(dim-1.)/(o+1.))+(0.569756278479)*dim
        ref[1,1,1,0]=(-0.315193818122)*(1+2.*(dim-1.)/(o+1.))+(-1.87410298298)*dim
        ref[1,1,1,1]=(-0.17620361977)*(1+2.*(dim-1.)/(o+1.))+(-0.174309770317)*dim
        ref[1,1,2,0]=(1.91602140435)*(1+2.*(dim-1.)/(o+1.))+(-0.762727403622)*dim
        ref[1,1,2,1]=(-0.481090245186)*(1+2.*(dim-1.)/(o+1.))+(0.302365878371)*dim
        ref[1,2,0,0]=(1.51144307636)*(1+2.*(dim-1.)/(o+1.))+(-0.469903943621)*dim
        ref[1,2,0,1]=(0.434474766511)*(1+2.*(dim-1.)/(o+1.))+(0.580510267262)*dim
        ref[1,2,1,0]=(-0.0359784228521)*(1+2.*(dim-1.)/(o+1.))+(1.64383448391)*dim
        ref[1,2,1,1]=(-1.01063132919)*(1+2.*(dim-1.)/(o+1.))+(1.51103090664)*dim
        ref[1,2,2,0]=(1.59952091715)*(1+2.*(dim-1.)/(o+1.))+(0.872265553532)*dim
        ref[1,2,2,1]=(1.01612465197)*(1+2.*(dim-1.)/(o+1.))+(-0.216728373319)*dim
        ref[1,3,0,0]=(-0.10531811522)*(1+2.*(dim-1.)/(o+1.))+(0.907610948542)*dim
        ref[1,3,0,1]=(0.5146706123)*(1+2.*(dim-1.)/(o+1.))+(0.877904273202)*dim
        ref[1,3,1,0]=(1.35946717224)*(1+2.*(dim-1.)/(o+1.))+(0.979242133542)*dim
        ref[1,3,1,1]=(-0.540844801432)*(1+2.*(dim-1.)/(o+1.))+(-1.01065173657)*dim
        ref[1,3,2,0]=(0.127935557915)*(1+2.*(dim-1.)/(o+1.))+(0.300209287538)*dim
        ref[1,3,2,1]=(-0.63818759413)*(1+2.*(dim-1.)/(o+1.))+(-1.20162771805)*dim
        ref[1,4,0,0]=(1.70266166649)*(1+2.*(dim-1.)/(o+1.))+(-1.27605382673)*dim
        ref[1,4,0,1]=(0.799915099484)*(1+2.*(dim-1.)/(o+1.))+(1.94853498615)*dim
        ref[1,4,1,0]=(0.916769254394)*(1+2.*(dim-1.)/(o+1.))+(-1.15016023941)*dim
        ref[1,4,1,1]=(1.87247945131)*(1+2.*(dim-1.)/(o+1.))+(0.147986167818)*dim
        ref[1,4,2,0]=(-0.988343334946)*(1+2.*(dim-1.)/(o+1.))+(0.169122301365)*dim
        ref[1,4,2,1]=(0.116889419423)*(1+2.*(dim-1.)/(o+1.))+(0.191358932597)*dim
        ref[2,0,0,0]=(-0.0479378392653)*(1+2.*(dim-1.)/(o+1.))+(0.978042254211)*dim
        ref[2,0,0,1]=(-0.485032392631)*(1+2.*(dim-1.)/(o+1.))+(-0.351118421752)*dim
        ref[2,0,1,0]=(0.23510362002)*(1+2.*(dim-1.)/(o+1.))+(-0.577849423671)*dim
        ref[2,0,1,1]=(-0.068175322817)*(1+2.*(dim-1.)/(o+1.))+(1.94609819864)*dim
        ref[2,0,2,0]=(-1.41002268435)*(1+2.*(dim-1.)/(o+1.))+(0.98455613352)*dim
        ref[2,0,2,1]=(-2.24634010969)*(1+2.*(dim-1.)/(o+1.))+(1.0410712698)*dim
        ref[2,1,0,0]=(-0.860551277441)*(1+2.*(dim-1.)/(o+1.))+(0.663495620068)*dim
        ref[2,1,0,1]=(-0.944426383871)*(1+2.*(dim-1.)/(o+1.))+(-0.341434032481)*dim
        ref[2,1,1,0]=(-1.16050351389)*(1+2.*(dim-1.)/(o+1.))+(0.902358791288)*dim
        ref[2,1,1,1]=(-0.790421314781)*(1+2.*(dim-1.)/(o+1.))+(0.618190806416)*dim
        ref[2,1,2,0]=(0.456001174897)*(1+2.*(dim-1.)/(o+1.))+(0.782682949672)*dim
        ref[2,1,2,1]=(0.532545234059)*(1+2.*(dim-1.)/(o+1.))+(0.806733163549)*dim
        ref[2,2,0,0]=(-1.63759186587)*(1+2.*(dim-1.)/(o+1.))+(0.275350925359)*dim
        ref[2,2,0,1]=(1.26997954169)*(1+2.*(dim-1.)/(o+1.))+(0.998155503083)*dim
        ref[2,2,1,0]=(1.22159383643)*(1+2.*(dim-1.)/(o+1.))+(-1.48657634495)*dim
        ref[2,2,1,1]=(-0.537844025016)*(1+2.*(dim-1.)/(o+1.))+(-1.38528909924)*dim
        ref[2,2,2,0]=(-0.274283519122)*(1+2.*(dim-1.)/(o+1.))+(0.243186764056)*dim
        ref[2,2,2,1]=(-1.10324559047)*(1+2.*(dim-1.)/(o+1.))+(0.0644794177698)*dim
        ref[2,3,0,0]=(0.571128036949)*(1+2.*(dim-1.)/(o+1.))+(0.60918412053)*dim
        ref[2,3,0,1]=(-0.580144876629)*(1+2.*(dim-1.)/(o+1.))+(-0.667332031982)*dim
        ref[2,3,1,0]=(-1.66177914705)*(1+2.*(dim-1.)/(o+1.))+(0.675143894185)*dim
        ref[2,3,1,1]=(0.301151474869)*(1+2.*(dim-1.)/(o+1.))+(-1.7478009387)*dim
        ref[2,3,2,0]=(-0.917598158317)*(1+2.*(dim-1.)/(o+1.))+(-0.0850570760611)*dim
        ref[2,3,2,1]=(-0.09025993235)*(1+2.*(dim-1.)/(o+1.))+(0.798689188243)*dim
        ref[2,4,0,0]=(1.94428451282)*(1+2.*(dim-1.)/(o+1.))+(0.247560933384)*dim
        ref[2,4,0,1]=(0.984775124785)*(1+2.*(dim-1.)/(o+1.))+(-0.506710524219)*dim
        ref[2,4,1,0]=(-0.197674459012)*(1+2.*(dim-1.)/(o+1.))+(0.0107873743289)*dim
        ref[2,4,1,1]=(-0.656458978046)*(1+2.*(dim-1.)/(o+1.))+(-1.13791264993)*dim
        ref[2,4,2,0]=(-0.778638912279)*(1+2.*(dim-1.)/(o+1.))+(0.324022275097)*dim
        ref[2,4,2,1]=(0.940510770447)*(1+2.*(dim-1.)/(o+1.))+(0.901184637171)*dim
        ref[3,0,0,0]=(-1.25429061263)*(1+2.*(dim-1.)/(o+1.))+(0.389376608226)*dim
        ref[3,0,0,1]=(-0.0171420806973)*(1+2.*(dim-1.)/(o+1.))+(0.230944451397)*dim
        ref[3,0,1,0]=(0.456140891562)*(1+2.*(dim-1.)/(o+1.))+(1.37768003626)*dim
        ref[3,0,1,1]=(-0.260435871091)*(1+2.*(dim-1.)/(o+1.))+(-1.10263475358)*dim
        ref[3,0,2,0]=(0.0625076184349)*(1+2.*(dim-1.)/(o+1.))+(1.51121135544)*dim
        ref[3,0,2,1]=(-0.973347700893)*(1+2.*(dim-1.)/(o+1.))+(-2.10909307795)*dim
        ref[3,1,0,0]=(1.04357815056)*(1+2.*(dim-1.)/(o+1.))+(-0.516720098176)*dim
        ref[3,1,0,1]=(-0.400651374046)*(1+2.*(dim-1.)/(o+1.))+(0.37941138127)*dim
        ref[3,1,1,0]=(0.614177886412)*(1+2.*(dim-1.)/(o+1.))+(-0.313511377498)*dim
        ref[3,1,1,1]=(0.97787134432)*(1+2.*(dim-1.)/(o+1.))+(-1.28527683733)*dim
        ref[3,1,2,0]=(0.0702273162506)*(1+2.*(dim-1.)/(o+1.))+(-1.32447006034)*dim
        ref[3,1,2,1]=(-0.596202948193)*(1+2.*(dim-1.)/(o+1.))+(-0.648046584161)*dim
        ref[3,2,0,0]=(0.34378369699)*(1+2.*(dim-1.)/(o+1.))+(-1.49317727825)*dim
        ref[3,2,0,1]=(0.798530712175)*(1+2.*(dim-1.)/(o+1.))+(0.358905794706)*dim
        ref[3,2,1,0]=(-0.808586932417)*(1+2.*(dim-1.)/(o+1.))+(-1.66927503333)*dim
        ref[3,2,1,1]=(0.712768620002)*(1+2.*(dim-1.)/(o+1.))+(0.346136808876)*dim
        ref[3,2,2,0]=(0.0778075096053)*(1+2.*(dim-1.)/(o+1.))+(-1.92918285909)*dim
        ref[3,2,2,1]=(0.949967126178)*(1+2.*(dim-1.)/(o+1.))+(0.650598700321)*dim
        ref[3,3,0,0]=(0.496438920934)*(1+2.*(dim-1.)/(o+1.))+(-0.0508751252546)*dim
        ref[3,3,0,1]=(0.559728225123)*(1+2.*(dim-1.)/(o+1.))+(-1.27062926979)*dim
        ref[3,3,1,0]=(0.365561797964)*(1+2.*(dim-1.)/(o+1.))+(-0.505845726473)*dim
        ref[3,3,1,1]=(0.752880268444)*(1+2.*(dim-1.)/(o+1.))+(-1.0445066834)*dim
        ref[3,3,2,0]=(0.973972964742)*(1+2.*(dim-1.)/(o+1.))+(1.17022234424)*dim
        ref[3,3,2,1]=(-0.0513846363967)*(1+2.*(dim-1.)/(o+1.))+(-1.65296692739)*dim
        ref[3,4,0,0]=(0.526078850284)*(1+2.*(dim-1.)/(o+1.))+(0.382222855695)*dim
        ref[3,4,0,1]=(-0.598380704773)*(1+2.*(dim-1.)/(o+1.))+(-0.368489911929)*dim
        ref[3,4,1,0]=(1.91068972032)*(1+2.*(dim-1.)/(o+1.))+(1.70482043562)*dim
        ref[3,4,1,1]=(1.10853236727)*(1+2.*(dim-1.)/(o+1.))+(1.30705472744)*dim
        ref[3,4,2,0]=(0.78376609137)*(1+2.*(dim-1.)/(o+1.))+(1.59686251918)*dim
        ref[3,4,2,1]=(-0.550031299439)*(1+2.*(dim-1.)/(o+1.))+(0.328424892465)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.276000764164)*x[0]**o+(-0.468805844816)*x[0]+(0.216181374155)*x[1]**o+(0.670874609882)*x[1]
        ref=(0.492182138319)*(1+2.*(dim-1.)/(o+1.))+(0.202068765065)*dim
      else:
        s=(0.959211859452)*x[0]**o+(0.856777496752)*x[0]+(-0.689682159726)*x[1]**o+(-0.0511538529405)*x[1]+(0.909142526772)*x[2]**o+(-0.851436719686)*x[2]
        ref=(1.1786722265)*(1+2.*(dim-1.)/(o+1.))+(-0.0458130758744)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(0.959628680059)*x[0]**o+(0.316358798368)*x[0]+(-0.11695939796)*x[1]**o+(0.467189695616)*x[1]
        s[1]=(-0.0914212792301)*x[0]**o+(-0.876964392495)*x[0]+(-0.241429076296)*x[1]**o+(-0.136283508929)*x[1]
        ref[0]=(0.842669282099)*(1+2.*(dim-1.)/(o+1.))+(0.783548493984)*dim
        ref[1]=(-0.332850355526)*(1+2.*(dim-1.)/(o+1.))+(-1.01324790142)*dim
      else:
        s[0]=(0.438221600861)*x[0]**o+(-0.675281168341)*x[0]+(-0.977876997798)*x[1]**o+(0.263292240228)*x[1]+(0.298700889488)*x[2]**o+(0.218464753318)*x[2]
        s[1]=(0.129309327387)*x[0]**o+(0.718818780853)*x[0]+(-0.829433536453)*x[1]**o+(0.189387074977)*x[1]+(-0.370123421171)*x[2]**o+(-0.944375947073)*x[2]
        ref[0]=(-0.240954507449)*(1+2.*(dim-1.)/(o+1.))+(-0.193524174795)*dim
        ref[1]=(-1.07024763024)*(1+2.*(dim-1.)/(o+1.))+(-0.0361700912426)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.179044774852)*x[0]**o+(-0.771154031526)*x[0]+(0.283792506671)*x[1]**o+(-0.402392048)*x[1]
        s[0,1]=(-0.232541435218)*x[0]**o+(-0.0245965827061)*x[0]+(0.645399819776)*x[1]**o+(-0.561735270705)*x[1]
        s[0,2]=(0.145947743942)*x[0]**o+(-0.575168618436)*x[0]+(-0.323186337184)*x[1]**o+(0.286768456651)*x[1]
        s[0,3]=(-0.850294957994)*x[0]**o+(0.338532201346)*x[0]+(-0.0793135342296)*x[1]**o+(-0.863030213073)*x[1]
        s[0,4]=(-0.250033615938)*x[0]**o+(0.0646802670443)*x[0]+(0.757218484623)*x[1]**o+(-0.136938974237)*x[1]
        s[1,0]=(-0.388186073304)*x[0]**o+(0.816982222815)*x[0]+(0.966648677362)*x[1]**o+(0.433441368441)*x[1]
        s[1,1]=(-0.26320466546)*x[0]**o+(0.814680348027)*x[0]+(0.121083870568)*x[1]**o+(0.546344037777)*x[1]
        s[1,2]=(0.999583479755)*x[0]**o+(-0.613738351596)*x[0]+(-0.15105692369)*x[1]**o+(-0.826110754043)*x[1]
        s[1,3]=(0.823160030826)*x[0]**o+(-0.986858175002)*x[0]+(0.979035971406)*x[1]**o+(0.495551722082)*x[1]
        s[1,4]=(0.322276907312)*x[0]**o+(-0.768815547878)*x[0]+(0.944581045957)*x[1]**o+(0.638853629633)*x[1]
        s[2,0]=(-0.715790872031)*x[0]**o+(-0.983342426781)*x[0]+(-0.584719832165)*x[1]**o+(0.372036497473)*x[1]
        s[2,1]=(0.00365756286977)*x[0]**o+(-0.715355493926)*x[0]+(0.812526897221)*x[1]**o+(-0.676997560296)*x[1]
        s[2,2]=(0.853542765125)*x[0]**o+(-0.468938820981)*x[0]+(0.958637393127)*x[1]**o+(-0.934834551087)*x[1]
        s[2,3]=(0.0411833783296)*x[0]**o+(0.668096714355)*x[0]+(-0.68394500964)*x[1]**o+(-0.915192447045)*x[1]
        s[2,4]=(0.477937998635)*x[0]**o+(-0.655023923727)*x[0]+(0.31178650811)*x[1]**o+(0.573276677618)*x[1]
        s[3,0]=(0.0827434421975)*x[0]**o+(0.116951368308)*x[0]+(-0.53041218008)*x[1]**o+(0.923217710706)*x[1]
        s[3,1]=(0.0352752409951)*x[0]**o+(0.186756758251)*x[0]+(-0.650196963391)*x[1]**o+(0.758872769044)*x[1]
        s[3,2]=(0.627293036559)*x[0]**o+(-0.101693009282)*x[0]+(-0.651126401234)*x[1]**o+(0.0946956663513)*x[1]
        s[3,3]=(-0.0645383041667)*x[0]**o+(0.411155076439)*x[0]+(0.936010412212)*x[1]**o+(-0.662112489076)*x[1]
        s[3,4]=(-0.749505083645)*x[0]**o+(0.767738728633)*x[0]+(-0.0521697451237)*x[1]**o+(0.743320460752)*x[1]
        ref[0,0]=(0.462837281523)*(1+2.*(dim-1.)/(o+1.))+(-1.17354607953)*dim
        ref[0,1]=(0.412858384558)*(1+2.*(dim-1.)/(o+1.))+(-0.586331853411)*dim
        ref[0,2]=(-0.177238593242)*(1+2.*(dim-1.)/(o+1.))+(-0.288400161786)*dim
        ref[0,3]=(-0.929608492223)*(1+2.*(dim-1.)/(o+1.))+(-0.524498011727)*dim
        ref[0,4]=(0.507184868685)*(1+2.*(dim-1.)/(o+1.))+(-0.0722587071931)*dim
        ref[1,0]=(0.578462604058)*(1+2.*(dim-1.)/(o+1.))+(1.25042359126)*dim
        ref[1,1]=(-0.142120794892)*(1+2.*(dim-1.)/(o+1.))+(1.3610243858)*dim
        ref[1,2]=(0.848526556066)*(1+2.*(dim-1.)/(o+1.))+(-1.43984910564)*dim
        ref[1,3]=(1.80219600223)*(1+2.*(dim-1.)/(o+1.))+(-0.49130645292)*dim
        ref[1,4]=(1.26685795327)*(1+2.*(dim-1.)/(o+1.))+(-0.129961918245)*dim
        ref[2,0]=(-1.3005107042)*(1+2.*(dim-1.)/(o+1.))+(-0.611305929308)*dim
        ref[2,1]=(0.816184460091)*(1+2.*(dim-1.)/(o+1.))+(-1.39235305422)*dim
        ref[2,2]=(1.81218015825)*(1+2.*(dim-1.)/(o+1.))+(-1.40377337207)*dim
        ref[2,3]=(-0.64276163131)*(1+2.*(dim-1.)/(o+1.))+(-0.247095732689)*dim
        ref[2,4]=(0.789724506745)*(1+2.*(dim-1.)/(o+1.))+(-0.0817472461083)*dim
        ref[3,0]=(-0.447668737882)*(1+2.*(dim-1.)/(o+1.))+(1.04016907901)*dim
        ref[3,1]=(-0.614921722396)*(1+2.*(dim-1.)/(o+1.))+(0.945629527295)*dim
        ref[3,2]=(-0.0238333646741)*(1+2.*(dim-1.)/(o+1.))+(-0.00699734293025)*dim
        ref[3,3]=(0.871472108046)*(1+2.*(dim-1.)/(o+1.))+(-0.250957412637)*dim
        ref[3,4]=(-0.801674828768)*(1+2.*(dim-1.)/(o+1.))+(1.51105918939)*dim
      else:
        s[0,0]=(-0.125646626393)*x[0]**o+(-0.248909495972)*x[0]+(0.694342489041)*x[1]**o+(-0.944409107438)*x[1]+(0.319473721262)*x[2]**o+(0.601806182335)*x[2]
        s[0,1]=(-0.505114968317)*x[0]**o+(-0.328058153553)*x[0]+(0.231927714111)*x[1]**o+(-0.20605100827)*x[1]+(0.441122232963)*x[2]**o+(-0.800716234123)*x[2]
        s[0,2]=(0.995136206899)*x[0]**o+(-0.853681974117)*x[0]+(-0.0627243072319)*x[1]**o+(0.974181482578)*x[1]+(0.719967223861)*x[2]**o+(0.714716693035)*x[2]
        s[0,3]=(-0.441934133426)*x[0]**o+(0.853576641618)*x[0]+(-0.981667008539)*x[1]**o+(0.628894290368)*x[1]+(-0.860902752279)*x[2]**o+(-0.265897164492)*x[2]
        s[0,4]=(0.793316938319)*x[0]**o+(-0.210698953461)*x[0]+(-0.642203222971)*x[1]**o+(-0.426236159989)*x[1]+(-0.425899483366)*x[2]**o+(0.160038345546)*x[2]
        s[1,0]=(0.708795168562)*x[0]**o+(-0.196948028271)*x[0]+(-0.935375576786)*x[1]**o+(0.0861198689411)*x[1]+(-0.921983001302)*x[2]**o+(0.460387290735)*x[2]
        s[1,1]=(0.929230715741)*x[0]**o+(-0.469953215705)*x[0]+(-0.327574657931)*x[1]**o+(0.508027765515)*x[1]+(-0.599239898886)*x[2]**o+(0.63360774997)*x[2]
        s[1,2]=(0.521095649571)*x[0]**o+(-0.306540618845)*x[0]+(0.989778780969)*x[1]**o+(0.942448043532)*x[1]+(0.696619526146)*x[2]**o+(0.316003759)*x[2]
        s[1,3]=(0.862728599037)*x[0]**o+(0.710693247125)*x[0]+(-0.319771560638)*x[1]**o+(0.00195165068796)*x[1]+(0.196231338915)*x[2]**o+(-0.540258673852)*x[2]
        s[1,4]=(0.608993531839)*x[0]**o+(-0.538278966186)*x[0]+(0.805409107065)*x[1]**o+(-0.306251894849)*x[1]+(0.137370330819)*x[2]**o+(0.723689377138)*x[2]
        s[2,0]=(0.725845088985)*x[0]**o+(-0.146963254784)*x[0]+(-0.33277537985)*x[1]**o+(0.905316206858)*x[1]+(0.833942228903)*x[2]**o+(-0.828915480014)*x[2]
        s[2,1]=(0.405007463559)*x[0]**o+(-0.58741177463)*x[0]+(-0.732389736864)*x[1]**o+(0.810006779688)*x[1]+(0.0934482848718)*x[2]**o+(-0.780478896444)*x[2]
        s[2,2]=(0.330277184347)*x[0]**o+(0.542618623736)*x[0]+(0.0984021435812)*x[1]**o+(0.528357675511)*x[1]+(-0.822522098782)*x[2]**o+(0.327999189729)*x[2]
        s[2,3]=(-0.65382619524)*x[0]**o+(0.683040803962)*x[0]+(-0.951007838791)*x[1]**o+(-0.907245567669)*x[1]+(-0.645828264634)*x[2]**o+(0.471775924817)*x[2]
        s[2,4]=(-0.881174731542)*x[0]**o+(0.902183751937)*x[0]+(0.0550862007787)*x[1]**o+(-0.251644685884)*x[1]+(0.220290011733)*x[2]**o+(-0.433798608361)*x[2]
        s[3,0]=(-0.822853797457)*x[0]**o+(0.92597725252)*x[0]+(0.151102447707)*x[1]**o+(-0.643773009686)*x[1]+(-0.82012950381)*x[2]**o+(-0.735602052424)*x[2]
        s[3,1]=(-0.0841053011733)*x[0]**o+(0.0930350084683)*x[0]+(0.276873742716)*x[1]**o+(-0.624783698693)*x[1]+(0.861829520969)*x[2]**o+(0.523311788689)*x[2]
        s[3,2]=(0.570824564972)*x[0]**o+(-0.258483951766)*x[0]+(-0.550248707418)*x[1]**o+(-0.762093783646)*x[1]+(-0.580531558601)*x[2]**o+(0.529122851261)*x[2]
        s[3,3]=(-0.945884266871)*x[0]**o+(0.947833330446)*x[0]+(0.94128603099)*x[1]**o+(-0.353478856114)*x[1]+(0.125480323773)*x[2]**o+(0.119572435249)*x[2]
        s[3,4]=(-0.0190004624928)*x[0]**o+(0.935780695738)*x[0]+(0.525428336584)*x[1]**o+(-0.94896065125)*x[1]+(-0.277950627306)*x[2]**o+(0.0152064531175)*x[2]
        ref[0,0]=(0.88816958391)*(1+2.*(dim-1.)/(o+1.))+(-0.591512421075)*dim
        ref[0,1]=(0.167934978757)*(1+2.*(dim-1.)/(o+1.))+(-1.33482539595)*dim
        ref[0,2]=(1.65237912353)*(1+2.*(dim-1.)/(o+1.))+(0.835216201496)*dim
        ref[0,3]=(-2.28450389424)*(1+2.*(dim-1.)/(o+1.))+(1.21657376749)*dim
        ref[0,4]=(-0.274785768018)*(1+2.*(dim-1.)/(o+1.))+(-0.476896767904)*dim
        ref[1,0]=(-1.14856340953)*(1+2.*(dim-1.)/(o+1.))+(0.349559131405)*dim
        ref[1,1]=(0.0024161589232)*(1+2.*(dim-1.)/(o+1.))+(0.67168229978)*dim
        ref[1,2]=(2.20749395669)*(1+2.*(dim-1.)/(o+1.))+(0.951911183687)*dim
        ref[1,3]=(0.739188377314)*(1+2.*(dim-1.)/(o+1.))+(0.172386223961)*dim
        ref[1,4]=(1.55177296972)*(1+2.*(dim-1.)/(o+1.))+(-0.120841483898)*dim
        ref[2,0]=(1.22701193804)*(1+2.*(dim-1.)/(o+1.))+(-0.0705625279405)*dim
        ref[2,1]=(-0.233933988434)*(1+2.*(dim-1.)/(o+1.))+(-0.557883891386)*dim
        ref[2,2]=(-0.393842770855)*(1+2.*(dim-1.)/(o+1.))+(1.39897548898)*dim
        ref[2,3]=(-2.25066229866)*(1+2.*(dim-1.)/(o+1.))+(0.247571161111)*dim
        ref[2,4]=(-0.60579851903)*(1+2.*(dim-1.)/(o+1.))+(0.216740457692)*dim
        ref[3,0]=(-1.49188085356)*(1+2.*(dim-1.)/(o+1.))+(-0.453397809589)*dim
        ref[3,1]=(1.05459796251)*(1+2.*(dim-1.)/(o+1.))+(-0.00843690153556)*dim
        ref[3,2]=(-0.559955701048)*(1+2.*(dim-1.)/(o+1.))+(-0.491454884151)*dim
        ref[3,3]=(0.120882087893)*(1+2.*(dim-1.)/(o+1.))+(0.713926909581)*dim
        ref[3,4]=(0.228477246785)*(1+2.*(dim-1.)/(o+1.))+(0.00202649760578)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(-0.926772326685)*x[0]**o+(0.387486521987)*x[0]+(0.434356695921)*x[1]**o+(0.612313168985)*x[1]
        s[0,0,1]=(0.0416123602896)*x[0]**o+(0.688827142008)*x[0]+(-0.983728357354)*x[1]**o+(0.360807139866)*x[1]
        s[0,1,0]=(0.695266078006)*x[0]**o+(0.415695465239)*x[0]+(-0.679624573491)*x[1]**o+(0.428762900621)*x[1]
        s[0,1,1]=(0.313944654416)*x[0]**o+(0.0809925793325)*x[0]+(0.621639152049)*x[1]**o+(0.611203880481)*x[1]
        s[1,0,0]=(0.779313803934)*x[0]**o+(0.417564759976)*x[0]+(0.309499891258)*x[1]**o+(0.309580948656)*x[1]
        s[1,0,1]=(-0.165921233657)*x[0]**o+(0.754873050152)*x[0]+(-0.086251970981)*x[1]**o+(0.326647616211)*x[1]
        s[1,1,0]=(-0.803682583733)*x[0]**o+(-0.532584958606)*x[0]+(0.161316696631)*x[1]**o+(-0.956754947772)*x[1]
        s[1,1,1]=(0.229536665796)*x[0]**o+(0.436690509304)*x[0]+(0.111716074991)*x[1]**o+(0.0152188972478)*x[1]
        s[2,0,0]=(0.431277528457)*x[0]**o+(-0.549627071548)*x[0]+(0.896434242812)*x[1]**o+(0.310788829239)*x[1]
        s[2,0,1]=(0.337481979979)*x[0]**o+(-0.115120163813)*x[0]+(0.261492211787)*x[1]**o+(-0.78544941385)*x[1]
        s[2,1,0]=(-0.930198652575)*x[0]**o+(0.378441956425)*x[0]+(0.10805648813)*x[1]**o+(-0.436097313908)*x[1]
        s[2,1,1]=(0.249959291812)*x[0]**o+(0.836638711074)*x[0]+(0.0728794590137)*x[1]**o+(0.66862732862)*x[1]
        s[3,0,0]=(-0.738201854061)*x[0]**o+(0.772813906884)*x[0]+(-0.181342905715)*x[1]**o+(0.804464512448)*x[1]
        s[3,0,1]=(-0.196081244053)*x[0]**o+(0.994865977727)*x[0]+(0.842797040161)*x[1]**o+(0.821509016638)*x[1]
        s[3,1,0]=(0.833815509702)*x[0]**o+(0.577383594944)*x[0]+(-0.570130606584)*x[1]**o+(-0.0805462111487)*x[1]
        s[3,1,1]=(-0.8554995951)*x[0]**o+(0.720808128606)*x[0]+(0.457881428951)*x[1]**o+(-0.0471201858303)*x[1]
        s[4,0,0]=(0.48208350846)*x[0]**o+(0.47231745613)*x[0]+(-0.889701258367)*x[1]**o+(-0.630128265742)*x[1]
        s[4,0,1]=(-0.252911141671)*x[0]**o+(0.49998513292)*x[0]+(0.394792632318)*x[1]**o+(-0.622944802336)*x[1]
        s[4,1,0]=(-0.677909127453)*x[0]**o+(0.396645756207)*x[0]+(-0.723263540777)*x[1]**o+(-0.339508447056)*x[1]
        s[4,1,1]=(-0.807950411362)*x[0]**o+(-0.149082460629)*x[0]+(-0.841669340631)*x[1]**o+(0.888900711982)*x[1]
        s[5,0,0]=(-0.560462354311)*x[0]**o+(0.00127986811316)*x[0]+(-0.21694183911)*x[1]**o+(-0.815224807309)*x[1]
        s[5,0,1]=(0.767918970281)*x[0]**o+(-0.0975926175116)*x[0]+(0.281160215987)*x[1]**o+(0.38258099137)*x[1]
        s[5,1,0]=(-0.0755717349023)*x[0]**o+(-0.321990446321)*x[0]+(-0.203426543697)*x[1]**o+(0.782003123159)*x[1]
        s[5,1,1]=(-0.947191134408)*x[0]**o+(-0.710118492404)*x[0]+(-0.0923673220139)*x[1]**o+(-0.103352087028)*x[1]
        ref[0,0,0]=(-0.492415630764)*(1+2.*(dim-1.)/(o+1.))+(0.999799690972)*dim
        ref[0,0,1]=(-0.942115997065)*(1+2.*(dim-1.)/(o+1.))+(1.04963428187)*dim
        ref[0,1,0]=(0.0156415045147)*(1+2.*(dim-1.)/(o+1.))+(0.84445836586)*dim
        ref[0,1,1]=(0.935583806465)*(1+2.*(dim-1.)/(o+1.))+(0.692196459814)*dim
        ref[1,0,0]=(1.08881369519)*(1+2.*(dim-1.)/(o+1.))+(0.727145708632)*dim
        ref[1,0,1]=(-0.252173204638)*(1+2.*(dim-1.)/(o+1.))+(1.08152066636)*dim
        ref[1,1,0]=(-0.642365887101)*(1+2.*(dim-1.)/(o+1.))+(-1.48933990638)*dim
        ref[1,1,1]=(0.341252740787)*(1+2.*(dim-1.)/(o+1.))+(0.451909406552)*dim
        ref[2,0,0]=(1.32771177127)*(1+2.*(dim-1.)/(o+1.))+(-0.238838242309)*dim
        ref[2,0,1]=(0.598974191766)*(1+2.*(dim-1.)/(o+1.))+(-0.900569577663)*dim
        ref[2,1,0]=(-0.822142164446)*(1+2.*(dim-1.)/(o+1.))+(-0.0576553574825)*dim
        ref[2,1,1]=(0.322838750825)*(1+2.*(dim-1.)/(o+1.))+(1.50526603969)*dim
        ref[3,0,0]=(-0.919544759776)*(1+2.*(dim-1.)/(o+1.))+(1.57727841933)*dim
        ref[3,0,1]=(0.646715796109)*(1+2.*(dim-1.)/(o+1.))+(1.81637499436)*dim
        ref[3,1,0]=(0.263684903118)*(1+2.*(dim-1.)/(o+1.))+(0.496837383795)*dim
        ref[3,1,1]=(-0.39761816615)*(1+2.*(dim-1.)/(o+1.))+(0.673687942776)*dim
        ref[4,0,0]=(-0.407617749907)*(1+2.*(dim-1.)/(o+1.))+(-0.157810809613)*dim
        ref[4,0,1]=(0.141881490647)*(1+2.*(dim-1.)/(o+1.))+(-0.122959669416)*dim
        ref[4,1,0]=(-1.40117266823)*(1+2.*(dim-1.)/(o+1.))+(0.0571373091516)*dim
        ref[4,1,1]=(-1.64961975199)*(1+2.*(dim-1.)/(o+1.))+(0.739818251353)*dim
        ref[5,0,0]=(-0.777404193421)*(1+2.*(dim-1.)/(o+1.))+(-0.813944939196)*dim
        ref[5,0,1]=(1.04907918627)*(1+2.*(dim-1.)/(o+1.))+(0.284988373859)*dim
        ref[5,1,0]=(-0.278998278599)*(1+2.*(dim-1.)/(o+1.))+(0.460012676837)*dim
        ref[5,1,1]=(-1.03955845642)*(1+2.*(dim-1.)/(o+1.))+(-0.813470579432)*dim
      else:
        s[0,0,0]=(0.887591619739)*x[0]**o+(0.407256482081)*x[0]+(0.382561775603)*x[1]**o+(0.170135617346)*x[1]+(-0.745510574059)*x[2]**o+(-0.455506612867)*x[2]
        s[0,0,1]=(-0.906429753951)*x[0]**o+(0.650007305971)*x[0]+(0.679810544399)*x[1]**o+(-0.327318089388)*x[1]+(-0.682816086768)*x[2]**o+(0.213571445023)*x[2]
        s[0,1,0]=(-0.101321617665)*x[0]**o+(-0.830248450523)*x[0]+(0.968297844076)*x[1]**o+(-0.461609225074)*x[1]+(-0.803382287303)*x[2]**o+(-0.264990872553)*x[2]
        s[0,1,1]=(0.762025047301)*x[0]**o+(0.352751067716)*x[0]+(0.085082244422)*x[1]**o+(0.260281207839)*x[1]+(-0.838657429282)*x[2]**o+(-0.449142572742)*x[2]
        s[1,0,0]=(0.60404314316)*x[0]**o+(0.433627969764)*x[0]+(-0.187903979441)*x[1]**o+(-0.490170615337)*x[1]+(0.854829687117)*x[2]**o+(-0.0562457996778)*x[2]
        s[1,0,1]=(0.379185262336)*x[0]**o+(-0.127987724933)*x[0]+(-0.199149895558)*x[1]**o+(0.177834623561)*x[1]+(-0.528371897238)*x[2]**o+(-0.752877478181)*x[2]
        s[1,1,0]=(-0.639722907848)*x[0]**o+(0.595935571051)*x[0]+(0.660417028114)*x[1]**o+(-0.70337252144)*x[1]+(0.0532141342539)*x[2]**o+(0.335609331703)*x[2]
        s[1,1,1]=(0.918579942175)*x[0]**o+(-0.419345002272)*x[0]+(-0.733731200432)*x[1]**o+(-0.926576582075)*x[1]+(-0.596748064397)*x[2]**o+(-0.994146900668)*x[2]
        s[2,0,0]=(0.465936668326)*x[0]**o+(0.146304452558)*x[0]+(0.434641974102)*x[1]**o+(0.7475528675)*x[1]+(-0.9968692715)*x[2]**o+(0.761573606267)*x[2]
        s[2,0,1]=(-0.19499782219)*x[0]**o+(-0.0487068806171)*x[0]+(-0.466751862198)*x[1]**o+(0.748775488587)*x[1]+(-0.159041931091)*x[2]**o+(-0.786654864976)*x[2]
        s[2,1,0]=(0.497506698992)*x[0]**o+(0.00446974112015)*x[0]+(-0.57139775079)*x[1]**o+(0.794232026784)*x[1]+(0.788382405769)*x[2]**o+(-0.611606691539)*x[2]
        s[2,1,1]=(-0.744517152085)*x[0]**o+(0.826326809885)*x[0]+(-0.489647614501)*x[1]**o+(-0.798398764067)*x[1]+(0.622474900672)*x[2]**o+(-0.958167077873)*x[2]
        s[3,0,0]=(-0.54850657455)*x[0]**o+(0.907733944445)*x[0]+(0.150662161005)*x[1]**o+(0.329278515834)*x[1]+(0.612014869233)*x[2]**o+(-0.332197388537)*x[2]
        s[3,0,1]=(-0.977591157702)*x[0]**o+(0.0314123979019)*x[0]+(0.357325837849)*x[1]**o+(-0.121203337609)*x[1]+(-0.417779485304)*x[2]**o+(0.56339482028)*x[2]
        s[3,1,0]=(-0.0882818794492)*x[0]**o+(0.373727429506)*x[0]+(0.251705556572)*x[1]**o+(0.75208764669)*x[1]+(-0.243231080216)*x[2]**o+(0.408665593633)*x[2]
        s[3,1,1]=(0.902672970131)*x[0]**o+(0.0816189076234)*x[0]+(-0.0188617954975)*x[1]**o+(0.248727356198)*x[1]+(0.659217806257)*x[2]**o+(-0.521141634196)*x[2]
        s[4,0,0]=(-0.884907344323)*x[0]**o+(0.284225021738)*x[0]+(0.810073434041)*x[1]**o+(0.267855075455)*x[1]+(0.0897833985076)*x[2]**o+(0.622374196423)*x[2]
        s[4,0,1]=(0.512127530167)*x[0]**o+(-0.0208132068954)*x[0]+(0.258260560413)*x[1]**o+(0.206670909439)*x[1]+(-0.359616018773)*x[2]**o+(0.538953196654)*x[2]
        s[4,1,0]=(-0.775312439486)*x[0]**o+(0.781309943125)*x[0]+(0.277745830089)*x[1]**o+(-0.166968517573)*x[1]+(0.428260154734)*x[2]**o+(0.753276413709)*x[2]
        s[4,1,1]=(-0.575086193799)*x[0]**o+(0.488745961307)*x[0]+(0.828550042541)*x[1]**o+(-0.82220696841)*x[1]+(0.48247983233)*x[2]**o+(0.348574290989)*x[2]
        s[5,0,0]=(0.904922447578)*x[0]**o+(0.94372036658)*x[0]+(0.554305596085)*x[1]**o+(-0.633630196279)*x[1]+(0.0304645794125)*x[2]**o+(-0.338076751695)*x[2]
        s[5,0,1]=(0.157663141664)*x[0]**o+(0.27698264503)*x[0]+(-0.244811019178)*x[1]**o+(-0.86016833799)*x[1]+(0.0037249377228)*x[2]**o+(0.415013895956)*x[2]
        s[5,1,0]=(-0.0651731085463)*x[0]**o+(0.656433177612)*x[0]+(-0.693226351108)*x[1]**o+(-0.195335968106)*x[1]+(0.735806703428)*x[2]**o+(-0.284744282673)*x[2]
        s[5,1,1]=(0.14124610976)*x[0]**o+(0.114249449125)*x[0]+(-0.89675797948)*x[1]**o+(-0.109343809939)*x[1]+(-0.171059459418)*x[2]**o+(-0.125911673497)*x[2]
        ref[0,0,0]=(0.524642821284)*(1+2.*(dim-1.)/(o+1.))+(0.12188548656)*dim
        ref[0,0,1]=(-0.90943529632)*(1+2.*(dim-1.)/(o+1.))+(0.536260661607)*dim
        ref[0,1,0]=(0.0635939391079)*(1+2.*(dim-1.)/(o+1.))+(-1.55684854815)*dim
        ref[0,1,1]=(0.0084498624415)*(1+2.*(dim-1.)/(o+1.))+(0.163889702813)*dim
        ref[1,0,0]=(1.27096885084)*(1+2.*(dim-1.)/(o+1.))+(-0.112788445251)*dim
        ref[1,0,1]=(-0.34833653046)*(1+2.*(dim-1.)/(o+1.))+(-0.703030579553)*dim
        ref[1,1,0]=(0.0739082545202)*(1+2.*(dim-1.)/(o+1.))+(0.228172381314)*dim
        ref[1,1,1]=(-0.411899322654)*(1+2.*(dim-1.)/(o+1.))+(-2.34006848501)*dim
        ref[2,0,0]=(-0.0962906290716)*(1+2.*(dim-1.)/(o+1.))+(1.65543092632)*dim
        ref[2,0,1]=(-0.820791615479)*(1+2.*(dim-1.)/(o+1.))+(-0.0865862570057)*dim
        ref[2,1,0]=(0.714491353971)*(1+2.*(dim-1.)/(o+1.))+(0.187095076365)*dim
        ref[2,1,1]=(-0.611689865914)*(1+2.*(dim-1.)/(o+1.))+(-0.930239032054)*dim
        ref[3,0,0]=(0.214170455688)*(1+2.*(dim-1.)/(o+1.))+(0.904815071742)*dim
        ref[3,0,1]=(-1.03804480516)*(1+2.*(dim-1.)/(o+1.))+(0.473603880573)*dim
        ref[3,1,0]=(-0.0798074030926)*(1+2.*(dim-1.)/(o+1.))+(1.53448066983)*dim
        ref[3,1,1]=(1.54302898089)*(1+2.*(dim-1.)/(o+1.))+(-0.190795370375)*dim
        ref[4,0,0]=(0.0149494882254)*(1+2.*(dim-1.)/(o+1.))+(1.17445429362)*dim
        ref[4,0,1]=(0.410772071806)*(1+2.*(dim-1.)/(o+1.))+(0.724810899198)*dim
        ref[4,1,0]=(-0.0693064546636)*(1+2.*(dim-1.)/(o+1.))+(1.36761783926)*dim
        ref[4,1,1]=(0.735943681072)*(1+2.*(dim-1.)/(o+1.))+(0.0151132838872)*dim
        ref[5,0,0]=(1.48969262308)*(1+2.*(dim-1.)/(o+1.))+(-0.0279865813938)*dim
        ref[5,0,1]=(-0.0834229397911)*(1+2.*(dim-1.)/(o+1.))+(-0.168171797005)*dim
        ref[5,1,0]=(-0.0225927562259)*(1+2.*(dim-1.)/(o+1.))+(0.176352926834)*dim
        ref[5,1,1]=(-0.926571329138)*(1+2.*(dim-1.)/(o+1.))+(-0.121006034312)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the FunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-0.0564495520087)*x[0]**o+(0.322467947613)*x[0]+(-0.426040737283)*x[1]**o+(0.743030997838)*x[1]
        s[0,0,0,1]=(-0.705100478682)*x[0]**o+(0.0428011755277)*x[0]+(-0.667969683035)*x[1]**o+(-0.05268016216)*x[1]
        s[0,0,1,0]=(-0.646229122275)*x[0]**o+(-0.117856575975)*x[0]+(-0.928260620991)*x[1]**o+(-0.0295789972002)*x[1]
        s[0,0,1,1]=(0.684057019666)*x[0]**o+(0.217883017426)*x[0]+(-0.8003965891)*x[1]**o+(0.805481050427)*x[1]
        s[0,0,2,0]=(-0.517102518951)*x[0]**o+(0.703108669274)*x[0]+(-0.0818548881221)*x[1]**o+(0.967661353324)*x[1]
        s[0,0,2,1]=(-0.311422038099)*x[0]**o+(0.647051468428)*x[0]+(-0.0134135628666)*x[1]**o+(0.401772150809)*x[1]
        s[0,1,0,0]=(-0.372031418539)*x[0]**o+(-0.348570224054)*x[0]+(0.971394255739)*x[1]**o+(0.220452856894)*x[1]
        s[0,1,0,1]=(0.726171816717)*x[0]**o+(0.0115054389412)*x[0]+(-0.663899850543)*x[1]**o+(-0.00862058334168)*x[1]
        s[0,1,1,0]=(0.0546574475678)*x[0]**o+(-0.110568027517)*x[0]+(-0.490412123307)*x[1]**o+(0.70968579481)*x[1]
        s[0,1,1,1]=(0.525121644008)*x[0]**o+(-0.460449092422)*x[0]+(-0.105141027916)*x[1]**o+(0.70599530088)*x[1]
        s[0,1,2,0]=(-0.333393404511)*x[0]**o+(0.956105748878)*x[0]+(-0.542294839284)*x[1]**o+(0.856767614724)*x[1]
        s[0,1,2,1]=(0.499018678584)*x[0]**o+(0.709512775393)*x[0]+(-0.0278608905417)*x[1]**o+(-0.408880317793)*x[1]
        s[0,2,0,0]=(0.196709165539)*x[0]**o+(-0.824269904377)*x[0]+(-0.776756196058)*x[1]**o+(-0.21242798078)*x[1]
        s[0,2,0,1]=(0.101174745892)*x[0]**o+(0.103180763582)*x[0]+(-0.25979513904)*x[1]**o+(-0.595455963398)*x[1]
        s[0,2,1,0]=(-0.490524427754)*x[0]**o+(0.22473843241)*x[0]+(-0.559043751054)*x[1]**o+(-0.396861202642)*x[1]
        s[0,2,1,1]=(0.366671902637)*x[0]**o+(0.00444476264911)*x[0]+(0.84769491016)*x[1]**o+(0.340088796185)*x[1]
        s[0,2,2,0]=(0.0843962714222)*x[0]**o+(-0.678807745661)*x[0]+(0.0710532661834)*x[1]**o+(-0.131143081822)*x[1]
        s[0,2,2,1]=(-0.505967901175)*x[0]**o+(0.389944048429)*x[0]+(-0.607195013377)*x[1]**o+(0.615596624068)*x[1]
        s[0,3,0,0]=(0.480464466311)*x[0]**o+(-0.200660236866)*x[0]+(-0.366683832098)*x[1]**o+(0.473029292775)*x[1]
        s[0,3,0,1]=(0.503865126139)*x[0]**o+(0.620947845674)*x[0]+(0.482553725938)*x[1]**o+(0.903952597709)*x[1]
        s[0,3,1,0]=(-0.935816186601)*x[0]**o+(-0.694323019844)*x[0]+(-0.222643250757)*x[1]**o+(-0.61395534471)*x[1]
        s[0,3,1,1]=(-0.366248207885)*x[0]**o+(0.21981177916)*x[0]+(0.829964990663)*x[1]**o+(-0.458138466604)*x[1]
        s[0,3,2,0]=(-0.133505096099)*x[0]**o+(0.889073061727)*x[0]+(0.748032746123)*x[1]**o+(0.457372622346)*x[1]
        s[0,3,2,1]=(-0.940388978849)*x[0]**o+(0.194505986498)*x[0]+(0.277476781766)*x[1]**o+(0.0315390566538)*x[1]
        s[0,4,0,0]=(-0.246674730207)*x[0]**o+(-0.145685680956)*x[0]+(-0.855397050636)*x[1]**o+(0.332401911997)*x[1]
        s[0,4,0,1]=(-0.786787596226)*x[0]**o+(0.000178838426883)*x[0]+(0.0315355043243)*x[1]**o+(0.502956944109)*x[1]
        s[0,4,1,0]=(0.456353731831)*x[0]**o+(0.224571019711)*x[0]+(0.787432530547)*x[1]**o+(0.900715775247)*x[1]
        s[0,4,1,1]=(-0.87764927624)*x[0]**o+(0.547980575611)*x[0]+(-0.287094420401)*x[1]**o+(0.664160174901)*x[1]
        s[0,4,2,0]=(-0.195894642929)*x[0]**o+(-0.595129209807)*x[0]+(0.865849102635)*x[1]**o+(-0.677159397082)*x[1]
        s[0,4,2,1]=(-0.841735734151)*x[0]**o+(0.625862078643)*x[0]+(-0.545516568256)*x[1]**o+(0.114109991331)*x[1]
        s[1,0,0,0]=(0.209458674327)*x[0]**o+(-0.20896563634)*x[0]+(-0.579063332782)*x[1]**o+(-0.0433216462031)*x[1]
        s[1,0,0,1]=(0.0646951342633)*x[0]**o+(-0.999495478916)*x[0]+(-0.679168301408)*x[1]**o+(-0.357289498732)*x[1]
        s[1,0,1,0]=(0.19078385999)*x[0]**o+(0.14391750283)*x[0]+(-0.571814090672)*x[1]**o+(0.19532813646)*x[1]
        s[1,0,1,1]=(0.3385444312)*x[0]**o+(-0.961513536858)*x[0]+(-0.729134108689)*x[1]**o+(-0.477737542344)*x[1]
        s[1,0,2,0]=(0.545433953869)*x[0]**o+(-0.746630455132)*x[0]+(-0.49008038028)*x[1]**o+(-0.156535798648)*x[1]
        s[1,0,2,1]=(-0.718991145693)*x[0]**o+(-0.264354651201)*x[0]+(0.0812886298976)*x[1]**o+(-0.331692687243)*x[1]
        s[1,1,0,0]=(-0.646431085196)*x[0]**o+(-0.150073929112)*x[0]+(-0.919422164553)*x[1]**o+(-0.754093106669)*x[1]
        s[1,1,0,1]=(-0.424167875207)*x[0]**o+(0.00436241598561)*x[0]+(0.423024130972)*x[1]**o+(0.0414727269723)*x[1]
        s[1,1,1,0]=(0.972554573539)*x[0]**o+(-0.00871610987967)*x[0]+(-0.968130421895)*x[1]**o+(0.323276868881)*x[1]
        s[1,1,1,1]=(-0.223846361715)*x[0]**o+(0.0984359694216)*x[0]+(0.367109095942)*x[1]**o+(0.39421742616)*x[1]
        s[1,1,2,0]=(-0.532854459841)*x[0]**o+(0.0932005176653)*x[0]+(0.300915725885)*x[1]**o+(-0.407779066395)*x[1]
        s[1,1,2,1]=(-0.977204985435)*x[0]**o+(0.632461453595)*x[0]+(0.222769752858)*x[1]**o+(0.97524903541)*x[1]
        s[1,2,0,0]=(0.798817773565)*x[0]**o+(-0.687809695378)*x[0]+(-0.77661557915)*x[1]**o+(0.242848730169)*x[1]
        s[1,2,0,1]=(0.852738374774)*x[0]**o+(-0.353097828509)*x[0]+(-0.947070627494)*x[1]**o+(0.686527427094)*x[1]
        s[1,2,1,0]=(0.315473111735)*x[0]**o+(0.671113387299)*x[0]+(0.328881731518)*x[1]**o+(-0.269227909853)*x[1]
        s[1,2,1,1]=(-0.877500814779)*x[0]**o+(-0.508518061024)*x[0]+(-0.486367952918)*x[1]**o+(0.465181865784)*x[1]
        s[1,2,2,0]=(0.321396065376)*x[0]**o+(-0.977515734724)*x[0]+(-0.745047664798)*x[1]**o+(0.00770829984163)*x[1]
        s[1,2,2,1]=(0.217194865535)*x[0]**o+(0.84002706514)*x[0]+(0.4186527058)*x[1]**o+(-0.0197865947851)*x[1]
        s[1,3,0,0]=(-0.434267890176)*x[0]**o+(0.677387101394)*x[0]+(-0.433143616343)*x[1]**o+(0.76079607685)*x[1]
        s[1,3,0,1]=(-0.568019213241)*x[0]**o+(0.940292544942)*x[0]+(0.351503513272)*x[1]**o+(-0.8227358798)*x[1]
        s[1,3,1,0]=(0.225598112956)*x[0]**o+(-0.158411926824)*x[0]+(-0.0770214720299)*x[1]**o+(0.667242375231)*x[1]
        s[1,3,1,1]=(0.497365374413)*x[0]**o+(-0.402552536909)*x[0]+(-0.760975595637)*x[1]**o+(0.423931229263)*x[1]
        s[1,3,2,0]=(0.260548515487)*x[0]**o+(-0.723043083794)*x[0]+(0.40423813784)*x[1]**o+(0.591112253369)*x[1]
        s[1,3,2,1]=(0.0657703077553)*x[0]**o+(0.610016697083)*x[0]+(-0.515291622078)*x[1]**o+(-0.0821391959295)*x[1]
        s[1,4,0,0]=(-0.0842960021063)*x[0]**o+(0.830851714471)*x[0]+(0.571649443759)*x[1]**o+(-0.884015388056)*x[1]
        s[1,4,0,1]=(0.681962300417)*x[0]**o+(0.549746462866)*x[0]+(0.0330915607536)*x[1]**o+(0.757109551626)*x[1]
        s[1,4,1,0]=(0.413804087081)*x[0]**o+(-0.0713221635854)*x[0]+(0.153990572806)*x[1]**o+(-0.448202273823)*x[1]
        s[1,4,1,1]=(-0.109978839431)*x[0]**o+(-0.720641165805)*x[0]+(-0.465701345584)*x[1]**o+(-0.423310622549)*x[1]
        s[1,4,2,0]=(0.0884455154661)*x[0]**o+(0.341657397119)*x[0]+(-0.282148805378)*x[1]**o+(0.567581544922)*x[1]
        s[1,4,2,1]=(0.384270691665)*x[0]**o+(0.986879879066)*x[0]+(0.725843240631)*x[1]**o+(-0.888932402046)*x[1]
        s[2,0,0,0]=(0.915028230934)*x[0]**o+(0.0308158620736)*x[0]+(-0.534330519368)*x[1]**o+(-0.668490082908)*x[1]
        s[2,0,0,1]=(0.0447246773914)*x[0]**o+(-0.764286683802)*x[0]+(0.985272626016)*x[1]**o+(-0.582955807402)*x[1]
        s[2,0,1,0]=(-0.0982089493146)*x[0]**o+(-0.119886638989)*x[0]+(0.0918316786338)*x[1]**o+(-0.275013043117)*x[1]
        s[2,0,1,1]=(-0.187529321909)*x[0]**o+(0.715634182891)*x[0]+(0.554464679108)*x[1]**o+(0.75060799895)*x[1]
        s[2,0,2,0]=(0.402781810737)*x[0]**o+(-0.885441671822)*x[0]+(0.647449131552)*x[1]**o+(-0.82134489597)*x[1]
        s[2,0,2,1]=(0.903766197812)*x[0]**o+(0.397983290493)*x[0]+(-0.161556012403)*x[1]**o+(0.441970897)*x[1]
        s[2,1,0,0]=(-0.0286858171901)*x[0]**o+(0.152444209719)*x[0]+(0.713144598152)*x[1]**o+(-0.330803066479)*x[1]
        s[2,1,0,1]=(0.221312544677)*x[0]**o+(0.76681848391)*x[0]+(0.4237403202)*x[1]**o+(0.219978904876)*x[1]
        s[2,1,1,0]=(0.350534590086)*x[0]**o+(0.933385654173)*x[0]+(0.349813301929)*x[1]**o+(0.0427424949923)*x[1]
        s[2,1,1,1]=(0.660478429879)*x[0]**o+(-0.487740867789)*x[0]+(-0.270542163333)*x[1]**o+(-0.405080985084)*x[1]
        s[2,1,2,0]=(0.0114890380351)*x[0]**o+(-0.156620602559)*x[0]+(-0.494024071381)*x[1]**o+(0.164138650655)*x[1]
        s[2,1,2,1]=(0.688285293331)*x[0]**o+(-0.723548244407)*x[0]+(-0.979169222862)*x[1]**o+(-0.815167412147)*x[1]
        s[2,2,0,0]=(0.743910713659)*x[0]**o+(-0.319702207578)*x[0]+(0.464318179749)*x[1]**o+(-0.911256589955)*x[1]
        s[2,2,0,1]=(0.178943711133)*x[0]**o+(0.878989967814)*x[0]+(-0.584920241418)*x[1]**o+(-0.232803309459)*x[1]
        s[2,2,1,0]=(0.289708048016)*x[0]**o+(0.0250010477203)*x[0]+(0.231695416136)*x[1]**o+(0.00458067191189)*x[1]
        s[2,2,1,1]=(0.252996966945)*x[0]**o+(0.718948618791)*x[0]+(0.238406858981)*x[1]**o+(-0.222687536947)*x[1]
        s[2,2,2,0]=(0.707394678426)*x[0]**o+(-0.341044895187)*x[0]+(-0.924768233451)*x[1]**o+(-0.717376693436)*x[1]
        s[2,2,2,1]=(0.826870828657)*x[0]**o+(0.596221683232)*x[0]+(0.13184086469)*x[1]**o+(-0.122283119637)*x[1]
        s[2,3,0,0]=(0.082622085842)*x[0]**o+(0.802466780483)*x[0]+(-0.374536755651)*x[1]**o+(0.299695470983)*x[1]
        s[2,3,0,1]=(0.806722430379)*x[0]**o+(-0.307373050709)*x[0]+(-0.504241501795)*x[1]**o+(0.0365801039206)*x[1]
        s[2,3,1,0]=(-0.189680944199)*x[0]**o+(0.86059525782)*x[0]+(0.612911991629)*x[1]**o+(-0.868979910208)*x[1]
        s[2,3,1,1]=(-0.0606711884277)*x[0]**o+(-0.543604644984)*x[0]+(0.292808478791)*x[1]**o+(0.0024260857339)*x[1]
        s[2,3,2,0]=(0.258222812918)*x[0]**o+(0.304770046568)*x[0]+(-0.426946619435)*x[1]**o+(-0.0521912646177)*x[1]
        s[2,3,2,1]=(-0.324355225171)*x[0]**o+(-0.784048371347)*x[0]+(0.169611573731)*x[1]**o+(-0.548169504296)*x[1]
        s[2,4,0,0]=(0.0155438234769)*x[0]**o+(-0.620846826317)*x[0]+(0.409826337485)*x[1]**o+(0.129190074563)*x[1]
        s[2,4,0,1]=(0.923545504209)*x[0]**o+(-0.155101945409)*x[0]+(-0.717093202578)*x[1]**o+(-0.480920400256)*x[1]
        s[2,4,1,0]=(0.313709415121)*x[0]**o+(0.736484420744)*x[0]+(-0.0293655314952)*x[1]**o+(-0.599267629638)*x[1]
        s[2,4,1,1]=(0.463358526622)*x[0]**o+(-0.592469734442)*x[0]+(0.613104805703)*x[1]**o+(-0.178209835993)*x[1]
        s[2,4,2,0]=(-0.713644697339)*x[0]**o+(0.799721953045)*x[0]+(0.0589328481367)*x[1]**o+(0.547328786304)*x[1]
        s[2,4,2,1]=(0.993775334243)*x[0]**o+(0.553815998267)*x[0]+(0.1992862938)*x[1]**o+(0.81025353823)*x[1]
        s[3,0,0,0]=(0.160105719249)*x[0]**o+(-0.555881043013)*x[0]+(0.595178125526)*x[1]**o+(0.581499038384)*x[1]
        s[3,0,0,1]=(-0.073553285882)*x[0]**o+(0.302577114557)*x[0]+(0.170358740039)*x[1]**o+(0.75537952732)*x[1]
        s[3,0,1,0]=(-0.64563114781)*x[0]**o+(0.196969117846)*x[0]+(-0.972196629648)*x[1]**o+(-0.996586995934)*x[1]
        s[3,0,1,1]=(0.833966809791)*x[0]**o+(0.95981521797)*x[0]+(0.816649660191)*x[1]**o+(-0.220324291634)*x[1]
        s[3,0,2,0]=(0.102957823139)*x[0]**o+(-0.442375614091)*x[0]+(0.686402985053)*x[1]**o+(0.0731364854586)*x[1]
        s[3,0,2,1]=(-0.168212668775)*x[0]**o+(0.314439484314)*x[0]+(0.619151838083)*x[1]**o+(-0.360098253956)*x[1]
        s[3,1,0,0]=(-0.416264395501)*x[0]**o+(0.214984128512)*x[0]+(0.358642811015)*x[1]**o+(-0.869373823201)*x[1]
        s[3,1,0,1]=(0.181499327771)*x[0]**o+(-0.331648488714)*x[0]+(-0.191432513473)*x[1]**o+(0.0930773945136)*x[1]
        s[3,1,1,0]=(-0.976247847012)*x[0]**o+(-0.575393818806)*x[0]+(0.0929450161482)*x[1]**o+(-0.276689604101)*x[1]
        s[3,1,1,1]=(0.185352455216)*x[0]**o+(-0.931509714205)*x[0]+(-0.6870722095)*x[1]**o+(0.0624138101885)*x[1]
        s[3,1,2,0]=(-0.281042580465)*x[0]**o+(0.966028964753)*x[0]+(-0.419450359271)*x[1]**o+(0.873992760636)*x[1]
        s[3,1,2,1]=(0.993820925583)*x[0]**o+(0.690209908944)*x[0]+(0.148117153189)*x[1]**o+(0.71818108405)*x[1]
        s[3,2,0,0]=(-0.549354119752)*x[0]**o+(-0.282097817235)*x[0]+(0.896456051498)*x[1]**o+(-0.452485791074)*x[1]
        s[3,2,0,1]=(-0.0830679821316)*x[0]**o+(0.40212443752)*x[0]+(0.380555934515)*x[1]**o+(0.987505234574)*x[1]
        s[3,2,1,0]=(0.997713740817)*x[0]**o+(-0.692039069659)*x[0]+(-0.565960747782)*x[1]**o+(0.445414151858)*x[1]
        s[3,2,1,1]=(0.9675137078)*x[0]**o+(0.84769422726)*x[0]+(0.0938577014253)*x[1]**o+(-0.431299521474)*x[1]
        s[3,2,2,0]=(-0.868876045235)*x[0]**o+(-0.955493151846)*x[0]+(-0.9143773492)*x[1]**o+(0.899030565498)*x[1]
        s[3,2,2,1]=(0.629496450112)*x[0]**o+(-0.0218037231989)*x[0]+(-0.933196760824)*x[1]**o+(-0.549980925362)*x[1]
        s[3,3,0,0]=(-0.73789282568)*x[0]**o+(0.612764312125)*x[0]+(-0.80912954116)*x[1]**o+(0.608837621314)*x[1]
        s[3,3,0,1]=(0.465678249887)*x[0]**o+(-0.780539933825)*x[0]+(0.149236887279)*x[1]**o+(-0.842449293644)*x[1]
        s[3,3,1,0]=(-0.853285341888)*x[0]**o+(0.347485246788)*x[0]+(0.219773434631)*x[1]**o+(0.702046500252)*x[1]
        s[3,3,1,1]=(-0.0500921104166)*x[0]**o+(0.620181272972)*x[0]+(-0.514576066373)*x[1]**o+(0.0487351718037)*x[1]
        s[3,3,2,0]=(0.0904463554055)*x[0]**o+(0.268365306318)*x[0]+(-0.164684341011)*x[1]**o+(0.115207872226)*x[1]
        s[3,3,2,1]=(0.359085658137)*x[0]**o+(0.144112368075)*x[0]+(0.0650644733865)*x[1]**o+(0.176937328733)*x[1]
        s[3,4,0,0]=(-0.602904552775)*x[0]**o+(-0.648691919828)*x[0]+(-0.41237635973)*x[1]**o+(-0.282240953949)*x[1]
        s[3,4,0,1]=(-0.938255401084)*x[0]**o+(0.263050201557)*x[0]+(0.175643545166)*x[1]**o+(-0.995029122689)*x[1]
        s[3,4,1,0]=(0.728511178435)*x[0]**o+(-0.0673148421745)*x[0]+(-0.814903409759)*x[1]**o+(0.0229206126733)*x[1]
        s[3,4,1,1]=(-0.790381209272)*x[0]**o+(-0.326044592264)*x[0]+(-0.616797433118)*x[1]**o+(-0.0963030542755)*x[1]
        s[3,4,2,0]=(-0.286555245913)*x[0]**o+(-0.725972210674)*x[0]+(-0.809797611215)*x[1]**o+(0.477382881829)*x[1]
        s[3,4,2,1]=(0.947314594641)*x[0]**o+(0.731567714373)*x[0]+(0.104324647674)*x[1]**o+(0.0771726257761)*x[1]
        ref[0,0,0,0]=(-0.482490289291)*(1+2.*(dim-1.)/(o+1.))+(1.06549894545)*dim
        ref[0,0,0,1]=(-1.37307016172)*(1+2.*(dim-1.)/(o+1.))+(-0.00987898663227)*dim
        ref[0,0,1,0]=(-1.57448974327)*(1+2.*(dim-1.)/(o+1.))+(-0.147435573175)*dim
        ref[0,0,1,1]=(-0.116339569434)*(1+2.*(dim-1.)/(o+1.))+(1.02336406785)*dim
        ref[0,0,2,0]=(-0.598957407073)*(1+2.*(dim-1.)/(o+1.))+(1.6707700226)*dim
        ref[0,0,2,1]=(-0.324835600966)*(1+2.*(dim-1.)/(o+1.))+(1.04882361924)*dim
        ref[0,1,0,0]=(0.5993628372)*(1+2.*(dim-1.)/(o+1.))+(-0.128117367161)*dim
        ref[0,1,0,1]=(0.062271966174)*(1+2.*(dim-1.)/(o+1.))+(0.00288485559949)*dim
        ref[0,1,1,0]=(-0.43575467574)*(1+2.*(dim-1.)/(o+1.))+(0.599117767293)*dim
        ref[0,1,1,1]=(0.419980616092)*(1+2.*(dim-1.)/(o+1.))+(0.245546208457)*dim
        ref[0,1,2,0]=(-0.875688243795)*(1+2.*(dim-1.)/(o+1.))+(1.8128733636)*dim
        ref[0,1,2,1]=(0.471157788042)*(1+2.*(dim-1.)/(o+1.))+(0.3006324576)*dim
        ref[0,2,0,0]=(-0.580047030519)*(1+2.*(dim-1.)/(o+1.))+(-1.03669788516)*dim
        ref[0,2,0,1]=(-0.158620393148)*(1+2.*(dim-1.)/(o+1.))+(-0.492275199816)*dim
        ref[0,2,1,0]=(-1.04956817881)*(1+2.*(dim-1.)/(o+1.))+(-0.172122770232)*dim
        ref[0,2,1,1]=(1.2143668128)*(1+2.*(dim-1.)/(o+1.))+(0.344533558834)*dim
        ref[0,2,2,0]=(0.155449537606)*(1+2.*(dim-1.)/(o+1.))+(-0.809950827483)*dim
        ref[0,2,2,1]=(-1.11316291455)*(1+2.*(dim-1.)/(o+1.))+(1.0055406725)*dim
        ref[0,3,0,0]=(0.113780634213)*(1+2.*(dim-1.)/(o+1.))+(0.272369055909)*dim
        ref[0,3,0,1]=(0.986418852077)*(1+2.*(dim-1.)/(o+1.))+(1.52490044338)*dim
        ref[0,3,1,0]=(-1.15845943736)*(1+2.*(dim-1.)/(o+1.))+(-1.30827836455)*dim
        ref[0,3,1,1]=(0.463716782779)*(1+2.*(dim-1.)/(o+1.))+(-0.238326687444)*dim
        ref[0,3,2,0]=(0.614527650024)*(1+2.*(dim-1.)/(o+1.))+(1.34644568407)*dim
        ref[0,3,2,1]=(-0.662912197084)*(1+2.*(dim-1.)/(o+1.))+(0.226045043152)*dim
        ref[0,4,0,0]=(-1.10207178084)*(1+2.*(dim-1.)/(o+1.))+(0.186716231041)*dim
        ref[0,4,0,1]=(-0.755252091901)*(1+2.*(dim-1.)/(o+1.))+(0.503135782536)*dim
        ref[0,4,1,0]=(1.24378626238)*(1+2.*(dim-1.)/(o+1.))+(1.12528679496)*dim
        ref[0,4,1,1]=(-1.16474369664)*(1+2.*(dim-1.)/(o+1.))+(1.21214075051)*dim
        ref[0,4,2,0]=(0.669954459707)*(1+2.*(dim-1.)/(o+1.))+(-1.27228860689)*dim
        ref[0,4,2,1]=(-1.38725230241)*(1+2.*(dim-1.)/(o+1.))+(0.739972069974)*dim
        ref[1,0,0,0]=(-0.369604658454)*(1+2.*(dim-1.)/(o+1.))+(-0.252287282543)*dim
        ref[1,0,0,1]=(-0.614473167144)*(1+2.*(dim-1.)/(o+1.))+(-1.35678497765)*dim
        ref[1,0,1,0]=(-0.381030230682)*(1+2.*(dim-1.)/(o+1.))+(0.339245639291)*dim
        ref[1,0,1,1]=(-0.39058967749)*(1+2.*(dim-1.)/(o+1.))+(-1.4392510792)*dim
        ref[1,0,2,0]=(0.0553535735892)*(1+2.*(dim-1.)/(o+1.))+(-0.903166253779)*dim
        ref[1,0,2,1]=(-0.637702515795)*(1+2.*(dim-1.)/(o+1.))+(-0.596047338443)*dim
        ref[1,1,0,0]=(-1.56585324975)*(1+2.*(dim-1.)/(o+1.))+(-0.904167035781)*dim
        ref[1,1,0,1]=(-0.00114374423549)*(1+2.*(dim-1.)/(o+1.))+(0.0458351429579)*dim
        ref[1,1,1,0]=(0.00442415164442)*(1+2.*(dim-1.)/(o+1.))+(0.314560759001)*dim
        ref[1,1,1,1]=(0.143262734228)*(1+2.*(dim-1.)/(o+1.))+(0.492653395582)*dim
        ref[1,1,2,0]=(-0.231938733956)*(1+2.*(dim-1.)/(o+1.))+(-0.314578548729)*dim
        ref[1,1,2,1]=(-0.754435232577)*(1+2.*(dim-1.)/(o+1.))+(1.60771048901)*dim
        ref[1,2,0,0]=(0.0222021944152)*(1+2.*(dim-1.)/(o+1.))+(-0.444960965209)*dim
        ref[1,2,0,1]=(-0.0943322527204)*(1+2.*(dim-1.)/(o+1.))+(0.333429598586)*dim
        ref[1,2,1,0]=(0.644354843253)*(1+2.*(dim-1.)/(o+1.))+(0.401885477446)*dim
        ref[1,2,1,1]=(-1.3638687677)*(1+2.*(dim-1.)/(o+1.))+(-0.0433361952409)*dim
        ref[1,2,2,0]=(-0.423651599421)*(1+2.*(dim-1.)/(o+1.))+(-0.969807434883)*dim
        ref[1,2,2,1]=(0.635847571335)*(1+2.*(dim-1.)/(o+1.))+(0.820240470355)*dim
        ref[1,3,0,0]=(-0.867411506519)*(1+2.*(dim-1.)/(o+1.))+(1.43818317824)*dim
        ref[1,3,0,1]=(-0.21651569997)*(1+2.*(dim-1.)/(o+1.))+(0.117556665142)*dim
        ref[1,3,1,0]=(0.148576640926)*(1+2.*(dim-1.)/(o+1.))+(0.508830448407)*dim
        ref[1,3,1,1]=(-0.263610221224)*(1+2.*(dim-1.)/(o+1.))+(0.0213786923536)*dim
        ref[1,3,2,0]=(0.664786653328)*(1+2.*(dim-1.)/(o+1.))+(-0.131930830425)*dim
        ref[1,3,2,1]=(-0.449521314323)*(1+2.*(dim-1.)/(o+1.))+(0.527877501154)*dim
        ref[1,4,0,0]=(0.487353441653)*(1+2.*(dim-1.)/(o+1.))+(-0.0531636735847)*dim
        ref[1,4,0,1]=(0.715053861171)*(1+2.*(dim-1.)/(o+1.))+(1.30685601449)*dim
        ref[1,4,1,0]=(0.567794659886)*(1+2.*(dim-1.)/(o+1.))+(-0.519524437409)*dim
        ref[1,4,1,1]=(-0.575680185015)*(1+2.*(dim-1.)/(o+1.))+(-1.14395178835)*dim
        ref[1,4,2,0]=(-0.193703289912)*(1+2.*(dim-1.)/(o+1.))+(0.909238942041)*dim
        ref[1,4,2,1]=(1.1101139323)*(1+2.*(dim-1.)/(o+1.))+(0.0979474770197)*dim
        ref[2,0,0,0]=(0.380697711566)*(1+2.*(dim-1.)/(o+1.))+(-0.637674220835)*dim
        ref[2,0,0,1]=(1.02999730341)*(1+2.*(dim-1.)/(o+1.))+(-1.3472424912)*dim
        ref[2,0,1,0]=(-0.00637727068081)*(1+2.*(dim-1.)/(o+1.))+(-0.394899682106)*dim
        ref[2,0,1,1]=(0.366935357199)*(1+2.*(dim-1.)/(o+1.))+(1.46624218184)*dim
        ref[2,0,2,0]=(1.05023094229)*(1+2.*(dim-1.)/(o+1.))+(-1.70678656779)*dim
        ref[2,0,2,1]=(0.742210185409)*(1+2.*(dim-1.)/(o+1.))+(0.839954187494)*dim
        ref[2,1,0,0]=(0.684458780962)*(1+2.*(dim-1.)/(o+1.))+(-0.17835885676)*dim
        ref[2,1,0,1]=(0.645052864877)*(1+2.*(dim-1.)/(o+1.))+(0.986797388786)*dim
        ref[2,1,1,0]=(0.700347892015)*(1+2.*(dim-1.)/(o+1.))+(0.976128149165)*dim
        ref[2,1,1,1]=(0.389936266545)*(1+2.*(dim-1.)/(o+1.))+(-0.892821852874)*dim
        ref[2,1,2,0]=(-0.482535033346)*(1+2.*(dim-1.)/(o+1.))+(0.00751804809504)*dim
        ref[2,1,2,1]=(-0.29088392953)*(1+2.*(dim-1.)/(o+1.))+(-1.53871565655)*dim
        ref[2,2,0,0]=(1.20822889341)*(1+2.*(dim-1.)/(o+1.))+(-1.23095879753)*dim
        ref[2,2,0,1]=(-0.405976530285)*(1+2.*(dim-1.)/(o+1.))+(0.646186658355)*dim
        ref[2,2,1,0]=(0.521403464152)*(1+2.*(dim-1.)/(o+1.))+(0.0295817196322)*dim
        ref[2,2,1,1]=(0.491403825926)*(1+2.*(dim-1.)/(o+1.))+(0.496261081844)*dim
        ref[2,2,2,0]=(-0.217373555025)*(1+2.*(dim-1.)/(o+1.))+(-1.05842158862)*dim
        ref[2,2,2,1]=(0.958711693347)*(1+2.*(dim-1.)/(o+1.))+(0.473938563595)*dim
        ref[2,3,0,0]=(-0.291914669809)*(1+2.*(dim-1.)/(o+1.))+(1.10216225147)*dim
        ref[2,3,0,1]=(0.302480928584)*(1+2.*(dim-1.)/(o+1.))+(-0.270792946789)*dim
        ref[2,3,1,0]=(0.423231047429)*(1+2.*(dim-1.)/(o+1.))+(-0.00838465238812)*dim
        ref[2,3,1,1]=(0.232137290364)*(1+2.*(dim-1.)/(o+1.))+(-0.54117855925)*dim
        ref[2,3,2,0]=(-0.168723806518)*(1+2.*(dim-1.)/(o+1.))+(0.25257878195)*dim
        ref[2,3,2,1]=(-0.154743651439)*(1+2.*(dim-1.)/(o+1.))+(-1.33221787564)*dim
        ref[2,4,0,0]=(0.425370160961)*(1+2.*(dim-1.)/(o+1.))+(-0.491656751754)*dim
        ref[2,4,0,1]=(0.206452301631)*(1+2.*(dim-1.)/(o+1.))+(-0.636022345665)*dim
        ref[2,4,1,0]=(0.284343883626)*(1+2.*(dim-1.)/(o+1.))+(0.137216791106)*dim
        ref[2,4,1,1]=(1.07646333233)*(1+2.*(dim-1.)/(o+1.))+(-0.770679570435)*dim
        ref[2,4,2,0]=(-0.654711849202)*(1+2.*(dim-1.)/(o+1.))+(1.34705073935)*dim
        ref[2,4,2,1]=(1.19306162804)*(1+2.*(dim-1.)/(o+1.))+(1.3640695365)*dim
        ref[3,0,0,0]=(0.755283844775)*(1+2.*(dim-1.)/(o+1.))+(0.0256179953709)*dim
        ref[3,0,0,1]=(0.0968054541574)*(1+2.*(dim-1.)/(o+1.))+(1.05795664188)*dim
        ref[3,0,1,0]=(-1.61782777746)*(1+2.*(dim-1.)/(o+1.))+(-0.799617878088)*dim
        ref[3,0,1,1]=(1.65061646998)*(1+2.*(dim-1.)/(o+1.))+(0.739490926336)*dim
        ref[3,0,2,0]=(0.789360808191)*(1+2.*(dim-1.)/(o+1.))+(-0.369239128632)*dim
        ref[3,0,2,1]=(0.450939169308)*(1+2.*(dim-1.)/(o+1.))+(-0.0456587696413)*dim
        ref[3,1,0,0]=(-0.0576215844858)*(1+2.*(dim-1.)/(o+1.))+(-0.654389694689)*dim
        ref[3,1,0,1]=(-0.00993318570204)*(1+2.*(dim-1.)/(o+1.))+(-0.2385710942)*dim
        ref[3,1,1,0]=(-0.883302830864)*(1+2.*(dim-1.)/(o+1.))+(-0.852083422907)*dim
        ref[3,1,1,1]=(-0.501719754284)*(1+2.*(dim-1.)/(o+1.))+(-0.869095904016)*dim
        ref[3,1,2,0]=(-0.700492939735)*(1+2.*(dim-1.)/(o+1.))+(1.84002172539)*dim
        ref[3,1,2,1]=(1.14193807877)*(1+2.*(dim-1.)/(o+1.))+(1.40839099299)*dim
        ref[3,2,0,0]=(0.347101931746)*(1+2.*(dim-1.)/(o+1.))+(-0.73458360831)*dim
        ref[3,2,0,1]=(0.297487952383)*(1+2.*(dim-1.)/(o+1.))+(1.38962967209)*dim
        ref[3,2,1,0]=(0.431752993035)*(1+2.*(dim-1.)/(o+1.))+(-0.246624917801)*dim
        ref[3,2,1,1]=(1.06137140922)*(1+2.*(dim-1.)/(o+1.))+(0.416394705787)*dim
        ref[3,2,2,0]=(-1.78325339443)*(1+2.*(dim-1.)/(o+1.))+(-0.0564625863481)*dim
        ref[3,2,2,1]=(-0.303700310713)*(1+2.*(dim-1.)/(o+1.))+(-0.571784648561)*dim
        ref[3,3,0,0]=(-1.54702236684)*(1+2.*(dim-1.)/(o+1.))+(1.22160193344)*dim
        ref[3,3,0,1]=(0.614915137167)*(1+2.*(dim-1.)/(o+1.))+(-1.62298922747)*dim
        ref[3,3,1,0]=(-0.633511907257)*(1+2.*(dim-1.)/(o+1.))+(1.04953174704)*dim
        ref[3,3,1,1]=(-0.56466817679)*(1+2.*(dim-1.)/(o+1.))+(0.668916444776)*dim
        ref[3,3,2,0]=(-0.074237985605)*(1+2.*(dim-1.)/(o+1.))+(0.383573178544)*dim
        ref[3,3,2,1]=(0.424150131524)*(1+2.*(dim-1.)/(o+1.))+(0.321049696808)*dim
        ref[3,4,0,0]=(-1.01528091251)*(1+2.*(dim-1.)/(o+1.))+(-0.930932873777)*dim
        ref[3,4,0,1]=(-0.762611855917)*(1+2.*(dim-1.)/(o+1.))+(-0.731978921132)*dim
        ref[3,4,1,0]=(-0.0863922313234)*(1+2.*(dim-1.)/(o+1.))+(-0.0443942295012)*dim
        ref[3,4,1,1]=(-1.40717864239)*(1+2.*(dim-1.)/(o+1.))+(-0.42234764654)*dim
        ref[3,4,2,0]=(-1.09635285713)*(1+2.*(dim-1.)/(o+1.))+(-0.248589328845)*dim
        ref[3,4,2,1]=(1.05163924231)*(1+2.*(dim-1.)/(o+1.))+(0.808740340149)*dim
      else:
        s[0,0,0,0]=(0.315885178942)*x[0]**o+(-0.259227433663)*x[0]+(-0.985139490966)*x[1]**o+(-0.365399875853)*x[1]+(-0.378614709474)*x[2]**o+(-0.749166773032)*x[2]
        s[0,0,0,1]=(-0.980745316199)*x[0]**o+(0.917227239621)*x[0]+(-0.00961787312321)*x[1]**o+(-0.803272961034)*x[1]+(-0.569841488886)*x[2]**o+(-0.470914343673)*x[2]
        s[0,0,1,0]=(0.474857040946)*x[0]**o+(0.737872920694)*x[0]+(0.0428239150717)*x[1]**o+(0.805671300306)*x[1]+(-0.372628413406)*x[2]**o+(0.719721248997)*x[2]
        s[0,0,1,1]=(-0.711880589909)*x[0]**o+(0.321391465679)*x[0]+(0.519295140087)*x[1]**o+(0.038319529181)*x[1]+(0.38347458279)*x[2]**o+(0.331699888768)*x[2]
        s[0,0,2,0]=(-0.749496365536)*x[0]**o+(0.587604152085)*x[0]+(-0.713982951135)*x[1]**o+(-0.846872978696)*x[1]+(0.187878469529)*x[2]**o+(-0.605222544928)*x[2]
        s[0,0,2,1]=(-0.410414088285)*x[0]**o+(-0.767933775709)*x[0]+(-0.60156292058)*x[1]**o+(-0.489314475101)*x[1]+(0.888771153163)*x[2]**o+(-0.600341282488)*x[2]
        s[0,1,0,0]=(0.57725226271)*x[0]**o+(0.634144506828)*x[0]+(0.341370019875)*x[1]**o+(0.586961840417)*x[1]+(0.459018061293)*x[2]**o+(0.0746845088082)*x[2]
        s[0,1,0,1]=(0.0407919221935)*x[0]**o+(-0.360125630891)*x[0]+(-0.148792597333)*x[1]**o+(0.396869345515)*x[1]+(-0.745920425203)*x[2]**o+(0.0343202170743)*x[2]
        s[0,1,1,0]=(-0.924932256421)*x[0]**o+(0.516566988169)*x[0]+(-0.152377696008)*x[1]**o+(0.873885948456)*x[1]+(-0.360869839536)*x[2]**o+(-0.388854183331)*x[2]
        s[0,1,1,1]=(-0.323969456625)*x[0]**o+(0.304506990935)*x[0]+(-0.990292733104)*x[1]**o+(-0.245365945948)*x[1]+(0.181900978883)*x[2]**o+(-0.538788421192)*x[2]
        s[0,1,2,0]=(-0.374550735743)*x[0]**o+(0.564340899121)*x[0]+(0.801278948399)*x[1]**o+(-0.462054604952)*x[1]+(-0.785576728179)*x[2]**o+(0.0926519203136)*x[2]
        s[0,1,2,1]=(0.310136431429)*x[0]**o+(-0.00821101099015)*x[0]+(-0.974448609702)*x[1]**o+(0.348959120777)*x[1]+(-0.757748781922)*x[2]**o+(0.297755375582)*x[2]
        s[0,2,0,0]=(-0.812242021105)*x[0]**o+(0.898358963888)*x[0]+(0.77445189337)*x[1]**o+(0.786716425501)*x[1]+(0.544028378828)*x[2]**o+(0.549375386728)*x[2]
        s[0,2,0,1]=(-0.0378209686309)*x[0]**o+(-0.802035255239)*x[0]+(0.750004779554)*x[1]**o+(-0.671471110194)*x[1]+(0.993314295529)*x[2]**o+(-0.162020573179)*x[2]
        s[0,2,1,0]=(-0.492426591625)*x[0]**o+(0.902435821013)*x[0]+(0.557810230975)*x[1]**o+(-0.283019951662)*x[1]+(-0.987590772512)*x[2]**o+(-0.892849711933)*x[2]
        s[0,2,1,1]=(0.714593519237)*x[0]**o+(0.964018148555)*x[0]+(0.510176911254)*x[1]**o+(0.630953794541)*x[1]+(0.0297056372684)*x[2]**o+(0.554337702949)*x[2]
        s[0,2,2,0]=(-0.0754028072226)*x[0]**o+(0.0696670414466)*x[0]+(0.303440801227)*x[1]**o+(-0.876096813532)*x[1]+(-0.772838653879)*x[2]**o+(0.891062956492)*x[2]
        s[0,2,2,1]=(-0.517283509506)*x[0]**o+(-0.890513911735)*x[0]+(-0.850002026469)*x[1]**o+(0.0485442983018)*x[1]+(-0.690340415981)*x[2]**o+(0.356372135225)*x[2]
        s[0,3,0,0]=(0.00589321043146)*x[0]**o+(0.833040936257)*x[0]+(-0.779061736965)*x[1]**o+(0.596999628455)*x[1]+(0.185497386523)*x[2]**o+(-0.580282832451)*x[2]
        s[0,3,0,1]=(0.536520364057)*x[0]**o+(0.119573894206)*x[0]+(-0.481591297936)*x[1]**o+(-0.131066608017)*x[1]+(-0.714448088992)*x[2]**o+(-0.15245786606)*x[2]
        s[0,3,1,0]=(-0.649688821569)*x[0]**o+(0.872646765296)*x[0]+(0.996914263841)*x[1]**o+(-0.771999134903)*x[1]+(0.388158636968)*x[2]**o+(0.997694871854)*x[2]
        s[0,3,1,1]=(0.60805862552)*x[0]**o+(0.496342905951)*x[0]+(0.0678538981395)*x[1]**o+(-0.723458192199)*x[1]+(-0.227951366099)*x[2]**o+(-0.0454845476315)*x[2]
        s[0,3,2,0]=(0.927213775724)*x[0]**o+(-0.447976073985)*x[0]+(-0.176332048298)*x[1]**o+(0.177420385987)*x[1]+(0.752794515408)*x[2]**o+(-0.273183979432)*x[2]
        s[0,3,2,1]=(-0.648274135241)*x[0]**o+(0.16199869622)*x[0]+(-0.902866229405)*x[1]**o+(0.469114855264)*x[1]+(0.810779308988)*x[2]**o+(0.946561654853)*x[2]
        s[0,4,0,0]=(0.20080337911)*x[0]**o+(0.492510182922)*x[0]+(-0.39283470481)*x[1]**o+(-0.47885185637)*x[1]+(-0.925823446617)*x[2]**o+(0.222218483415)*x[2]
        s[0,4,0,1]=(-0.0972069635212)*x[0]**o+(0.183203800054)*x[0]+(-0.654862802864)*x[1]**o+(0.262618285977)*x[1]+(0.854563567237)*x[2]**o+(0.0964536047741)*x[2]
        s[0,4,1,0]=(0.45227435438)*x[0]**o+(0.538656962105)*x[0]+(0.365921783628)*x[1]**o+(0.215899953977)*x[1]+(0.592413111565)*x[2]**o+(-0.062048122597)*x[2]
        s[0,4,1,1]=(-0.818221902033)*x[0]**o+(-0.160355040635)*x[0]+(0.167475874143)*x[1]**o+(0.918522475482)*x[1]+(0.223565847733)*x[2]**o+(-0.552706943026)*x[2]
        s[0,4,2,0]=(0.917266921581)*x[0]**o+(-0.142334389128)*x[0]+(0.387287963005)*x[1]**o+(0.330170937971)*x[1]+(-0.822125826009)*x[2]**o+(-0.18853564744)*x[2]
        s[0,4,2,1]=(-0.720218394391)*x[0]**o+(0.844297728137)*x[0]+(0.428917747313)*x[1]**o+(0.0234117516707)*x[1]+(0.877188952533)*x[2]**o+(-0.580181524673)*x[2]
        s[1,0,0,0]=(-0.525360603901)*x[0]**o+(-0.444250370569)*x[0]+(-0.669034103414)*x[1]**o+(-0.463421879769)*x[1]+(-0.759457132782)*x[2]**o+(0.604035926001)*x[2]
        s[1,0,0,1]=(-0.0897354786393)*x[0]**o+(0.705360710589)*x[0]+(0.696483863344)*x[1]**o+(-0.241174862697)*x[1]+(-0.561951314355)*x[2]**o+(-0.912030054302)*x[2]
        s[1,0,1,0]=(0.52054104485)*x[0]**o+(0.0430575354511)*x[0]+(0.158290317145)*x[1]**o+(0.101019650181)*x[1]+(0.524338380356)*x[2]**o+(0.090701232108)*x[2]
        s[1,0,1,1]=(0.825208335954)*x[0]**o+(-0.473377827796)*x[0]+(0.68825627399)*x[1]**o+(0.494858806038)*x[1]+(0.224456734493)*x[2]**o+(0.406206820762)*x[2]
        s[1,0,2,0]=(-0.960366948345)*x[0]**o+(-0.234157118981)*x[0]+(-0.113359629154)*x[1]**o+(0.301361613247)*x[1]+(-0.594339813983)*x[2]**o+(-0.263211884586)*x[2]
        s[1,0,2,1]=(-0.94506913355)*x[0]**o+(0.389157389807)*x[0]+(-0.0470471678881)*x[1]**o+(-0.545905083414)*x[1]+(0.189132629724)*x[2]**o+(0.873560571105)*x[2]
        s[1,1,0,0]=(0.196235297476)*x[0]**o+(0.122584939735)*x[0]+(0.508485269882)*x[1]**o+(0.525003719219)*x[1]+(0.961756294197)*x[2]**o+(-0.766404136898)*x[2]
        s[1,1,0,1]=(0.158187473618)*x[0]**o+(0.570375022293)*x[0]+(0.815862093528)*x[1]**o+(-0.967297717974)*x[1]+(0.290548253702)*x[2]**o+(0.603934169794)*x[2]
        s[1,1,1,0]=(0.37168083831)*x[0]**o+(0.378783317761)*x[0]+(-0.62105516013)*x[1]**o+(-0.243670994716)*x[1]+(0.0493547361834)*x[2]**o+(-0.463410303505)*x[2]
        s[1,1,1,1]=(0.935366553895)*x[0]**o+(-0.62914115892)*x[0]+(-0.251619616412)*x[1]**o+(-0.00509465672734)*x[1]+(0.801300357872)*x[2]**o+(-0.225391368884)*x[2]
        s[1,1,2,0]=(-0.182365850951)*x[0]**o+(-0.510497578546)*x[0]+(-0.976490140605)*x[1]**o+(-0.810521612611)*x[1]+(0.568861407451)*x[2]**o+(0.984242045677)*x[2]
        s[1,1,2,1]=(0.438879837985)*x[0]**o+(0.123241828005)*x[0]+(-0.0920198739704)*x[1]**o+(0.0163670845773)*x[1]+(-0.170074389128)*x[2]**o+(0.9808945977)*x[2]
        s[1,2,0,0]=(0.844376099112)*x[0]**o+(0.281113346834)*x[0]+(0.678586875156)*x[1]**o+(-0.339719035096)*x[1]+(0.149112798051)*x[2]**o+(0.345737934766)*x[2]
        s[1,2,0,1]=(-0.823808399101)*x[0]**o+(0.741400052542)*x[0]+(0.631490960369)*x[1]**o+(0.884534048234)*x[1]+(0.986087890092)*x[2]**o+(0.0633984741173)*x[2]
        s[1,2,1,0]=(0.577630620249)*x[0]**o+(-0.875400887572)*x[0]+(0.282570000665)*x[1]**o+(-0.720651390173)*x[1]+(0.740020216206)*x[2]**o+(0.116758611432)*x[2]
        s[1,2,1,1]=(0.827573407856)*x[0]**o+(0.788455460143)*x[0]+(-0.935960102047)*x[1]**o+(0.128092793671)*x[1]+(-0.637271207159)*x[2]**o+(-0.0949607133369)*x[2]
        s[1,2,2,0]=(0.653489510769)*x[0]**o+(0.127242563347)*x[0]+(0.441720108695)*x[1]**o+(0.176710497085)*x[1]+(-0.466525521911)*x[2]**o+(-0.106726607773)*x[2]
        s[1,2,2,1]=(0.544895430642)*x[0]**o+(0.617184925179)*x[0]+(-0.590535213263)*x[1]**o+(-0.240993032963)*x[1]+(0.847624557098)*x[2]**o+(0.825419570696)*x[2]
        s[1,3,0,0]=(0.632663921163)*x[0]**o+(0.190826102102)*x[0]+(0.849534763016)*x[1]**o+(-0.0237561687913)*x[1]+(-0.269077077602)*x[2]**o+(-0.549527552663)*x[2]
        s[1,3,0,1]=(-0.573827583713)*x[0]**o+(0.698099772692)*x[0]+(-0.0111629174288)*x[1]**o+(0.779552959539)*x[1]+(0.268204963497)*x[2]**o+(0.598849955092)*x[2]
        s[1,3,1,0]=(-0.83963180468)*x[0]**o+(-0.688432739537)*x[0]+(-0.354553847784)*x[1]**o+(0.818998341858)*x[1]+(-0.214142424648)*x[2]**o+(-0.0349087206605)*x[2]
        s[1,3,1,1]=(0.961416480403)*x[0]**o+(0.214086874997)*x[0]+(-0.00837344828613)*x[1]**o+(0.907896124099)*x[1]+(-0.101452170826)*x[2]**o+(0.532739507876)*x[2]
        s[1,3,2,0]=(0.695510036287)*x[0]**o+(0.365801535675)*x[0]+(0.72964983115)*x[1]**o+(0.598376640377)*x[1]+(-0.786372424004)*x[2]**o+(-0.307873549549)*x[2]
        s[1,3,2,1]=(-0.00267955266574)*x[0]**o+(-0.36091144037)*x[0]+(-0.679691091713)*x[1]**o+(-0.728893213661)*x[1]+(-0.925050172841)*x[2]**o+(-0.218818592747)*x[2]
        s[1,4,0,0]=(-0.10754938599)*x[0]**o+(-0.412916506659)*x[0]+(0.688731006932)*x[1]**o+(0.878299598346)*x[1]+(0.726030761617)*x[2]**o+(0.637038101344)*x[2]
        s[1,4,0,1]=(0.262443563953)*x[0]**o+(-0.768431899484)*x[0]+(0.807995951545)*x[1]**o+(0.729996462998)*x[1]+(-0.128003445373)*x[2]**o+(0.516856350881)*x[2]
        s[1,4,1,0]=(-0.275354824026)*x[0]**o+(0.0620548936694)*x[0]+(-0.708292160104)*x[1]**o+(-0.138724590156)*x[1]+(-0.937760817006)*x[2]**o+(0.295878139279)*x[2]
        s[1,4,1,1]=(-0.194978075848)*x[0]**o+(0.116302091644)*x[0]+(0.668500205751)*x[1]**o+(-0.276843526204)*x[1]+(0.554624275832)*x[2]**o+(-0.0252161509726)*x[2]
        s[1,4,2,0]=(-0.980507466062)*x[0]**o+(0.345696047502)*x[0]+(0.990984322861)*x[1]**o+(-0.29237796986)*x[1]+(-0.113106912241)*x[2]**o+(-0.171569411982)*x[2]
        s[1,4,2,1]=(-0.996430023986)*x[0]**o+(0.0129252732357)*x[0]+(0.432605766267)*x[1]**o+(0.245516428361)*x[1]+(-0.287283002725)*x[2]**o+(0.496535294994)*x[2]
        s[2,0,0,0]=(-0.521251352762)*x[0]**o+(0.814082050334)*x[0]+(-0.342514979075)*x[1]**o+(0.180394009823)*x[1]+(-0.928876765188)*x[2]**o+(-0.177317444177)*x[2]
        s[2,0,0,1]=(-0.752594467046)*x[0]**o+(0.00697110526032)*x[0]+(0.107735313159)*x[1]**o+(0.507279243411)*x[1]+(0.246185410903)*x[2]**o+(-0.705550571172)*x[2]
        s[2,0,1,0]=(0.987074535948)*x[0]**o+(0.120612341024)*x[0]+(-0.00891981771691)*x[1]**o+(0.412034737534)*x[1]+(-0.641732350314)*x[2]**o+(0.702296660315)*x[2]
        s[2,0,1,1]=(0.545360348518)*x[0]**o+(0.246057841211)*x[0]+(0.68096977198)*x[1]**o+(0.16979635608)*x[1]+(-0.786467966954)*x[2]**o+(-0.604979391255)*x[2]
        s[2,0,2,0]=(0.390620126799)*x[0]**o+(-0.297208846965)*x[0]+(-0.1676231355)*x[1]**o+(-0.999842891217)*x[1]+(-0.525384600734)*x[2]**o+(-0.616431037869)*x[2]
        s[2,0,2,1]=(-0.114488632507)*x[0]**o+(-0.837696558005)*x[0]+(0.0856093277648)*x[1]**o+(-0.813930500792)*x[1]+(0.0242825112259)*x[2]**o+(-0.553431782224)*x[2]
        s[2,1,0,0]=(-0.158829154482)*x[0]**o+(0.944427245379)*x[0]+(0.288168271085)*x[1]**o+(0.294610923624)*x[1]+(-0.93996240515)*x[2]**o+(-0.141444934145)*x[2]
        s[2,1,0,1]=(0.600407255802)*x[0]**o+(0.466366546805)*x[0]+(0.293813875924)*x[1]**o+(0.824442365447)*x[1]+(-0.95914204615)*x[2]**o+(0.268073473809)*x[2]
        s[2,1,1,0]=(-0.00912285930883)*x[0]**o+(0.541241315284)*x[0]+(-0.941279012281)*x[1]**o+(0.192591379725)*x[1]+(0.84906414331)*x[2]**o+(0.739941920728)*x[2]
        s[2,1,1,1]=(-0.515663862321)*x[0]**o+(-0.995173844425)*x[0]+(-0.0545298263549)*x[1]**o+(0.0547616321893)*x[1]+(0.906580176018)*x[2]**o+(-0.724193715365)*x[2]
        s[2,1,2,0]=(-0.961884774945)*x[0]**o+(0.228071784467)*x[0]+(0.109594642294)*x[1]**o+(0.390512248069)*x[1]+(0.584516492005)*x[2]**o+(0.732112742614)*x[2]
        s[2,1,2,1]=(-0.447795495117)*x[0]**o+(-0.436915184002)*x[0]+(0.0097966293808)*x[1]**o+(0.00984049811243)*x[1]+(-0.534980632513)*x[2]**o+(0.847070567413)*x[2]
        s[2,2,0,0]=(-0.72527144583)*x[0]**o+(-0.328380984831)*x[0]+(0.891931683176)*x[1]**o+(0.731425008677)*x[1]+(0.451269066706)*x[2]**o+(-0.876119382129)*x[2]
        s[2,2,0,1]=(-0.645145440462)*x[0]**o+(0.549953455814)*x[0]+(-0.714731040969)*x[1]**o+(-0.367442134998)*x[1]+(-0.784974893863)*x[2]**o+(0.993294593495)*x[2]
        s[2,2,1,0]=(0.795416456648)*x[0]**o+(-0.249185994779)*x[0]+(0.0887737619644)*x[1]**o+(-0.8527851189)*x[1]+(-0.977545483463)*x[2]**o+(-0.458811422497)*x[2]
        s[2,2,1,1]=(0.448509047555)*x[0]**o+(0.251160841663)*x[0]+(-0.958180262344)*x[1]**o+(0.00687385479851)*x[1]+(0.415237807094)*x[2]**o+(0.454049069214)*x[2]
        s[2,2,2,0]=(-0.875182150725)*x[0]**o+(-0.164810616491)*x[0]+(-0.854984081654)*x[1]**o+(-0.487839136852)*x[1]+(-0.991611140703)*x[2]**o+(0.131746964462)*x[2]
        s[2,2,2,1]=(0.190113983979)*x[0]**o+(-0.0580500408307)*x[0]+(-0.987736788697)*x[1]**o+(0.29105974832)*x[1]+(-0.841830938414)*x[2]**o+(-0.399303249251)*x[2]
        s[2,3,0,0]=(0.375874939141)*x[0]**o+(0.0179896833971)*x[0]+(0.935171068662)*x[1]**o+(-0.605279673785)*x[1]+(-0.315907419394)*x[2]**o+(-0.653351592585)*x[2]
        s[2,3,0,1]=(-0.307367451597)*x[0]**o+(0.433464337349)*x[0]+(0.120694815275)*x[1]**o+(0.882573418509)*x[1]+(0.720880126974)*x[2]**o+(0.413515788634)*x[2]
        s[2,3,1,0]=(-0.360768726229)*x[0]**o+(-0.345798386124)*x[0]+(-0.158591700207)*x[1]**o+(0.777571415125)*x[1]+(-0.515278006685)*x[2]**o+(-0.60141050103)*x[2]
        s[2,3,1,1]=(0.19968785285)*x[0]**o+(-0.678957085866)*x[0]+(-0.504487760025)*x[1]**o+(-0.938195450799)*x[1]+(-0.869784726008)*x[2]**o+(0.611936589332)*x[2]
        s[2,3,2,0]=(-0.94908432374)*x[0]**o+(-0.764309365026)*x[0]+(0.299970983215)*x[1]**o+(0.76972605976)*x[1]+(0.808781373563)*x[2]**o+(-0.744690590895)*x[2]
        s[2,3,2,1]=(-0.827801524944)*x[0]**o+(-0.639131625181)*x[0]+(0.115482861386)*x[1]**o+(0.749852345379)*x[1]+(-0.876800704567)*x[2]**o+(-0.335471532349)*x[2]
        s[2,4,0,0]=(0.957721855837)*x[0]**o+(-0.37091118004)*x[0]+(-0.457918571765)*x[1]**o+(0.574281982261)*x[1]+(-0.0381561925569)*x[2]**o+(0.583211845309)*x[2]
        s[2,4,0,1]=(0.922272060507)*x[0]**o+(0.934043162659)*x[0]+(0.665312385413)*x[1]**o+(0.82839206727)*x[1]+(0.192535528872)*x[2]**o+(-0.170352598091)*x[2]
        s[2,4,1,0]=(0.696400806982)*x[0]**o+(-0.864993988074)*x[0]+(0.175067366294)*x[1]**o+(-0.117341125055)*x[1]+(-0.0862802982578)*x[2]**o+(0.360686011537)*x[2]
        s[2,4,1,1]=(0.285042052581)*x[0]**o+(0.819442194445)*x[0]+(0.985647806774)*x[1]**o+(-0.706227171911)*x[1]+(-0.854322012297)*x[2]**o+(-0.854051496942)*x[2]
        s[2,4,2,0]=(0.493375212662)*x[0]**o+(0.0135080286945)*x[0]+(-0.620056636744)*x[1]**o+(-0.160959399742)*x[1]+(-0.524788079589)*x[2]**o+(-0.421933761234)*x[2]
        s[2,4,2,1]=(0.657882929009)*x[0]**o+(0.164321066198)*x[0]+(0.0849684071734)*x[1]**o+(0.128314490298)*x[1]+(0.0313579868936)*x[2]**o+(0.608750041614)*x[2]
        s[3,0,0,0]=(0.401854741306)*x[0]**o+(0.134373785831)*x[0]+(-0.518732727663)*x[1]**o+(0.962631263931)*x[1]+(0.756955386629)*x[2]**o+(0.10153666864)*x[2]
        s[3,0,0,1]=(0.407206244876)*x[0]**o+(-0.432733726651)*x[0]+(-0.940640930336)*x[1]**o+(0.745788564728)*x[1]+(-0.342370630964)*x[2]**o+(-0.535375080975)*x[2]
        s[3,0,1,0]=(0.360769684176)*x[0]**o+(-0.931274190319)*x[0]+(0.86097262982)*x[1]**o+(0.767924495942)*x[1]+(0.751955166412)*x[2]**o+(-0.243315246145)*x[2]
        s[3,0,1,1]=(0.729403674327)*x[0]**o+(0.400842302945)*x[0]+(0.275966885949)*x[1]**o+(-0.546578452065)*x[1]+(-0.844280661655)*x[2]**o+(0.0715593226033)*x[2]
        s[3,0,2,0]=(-0.26926147241)*x[0]**o+(-0.313445919689)*x[0]+(0.0745755455487)*x[1]**o+(0.561797432558)*x[1]+(0.746592037865)*x[2]**o+(0.691841700089)*x[2]
        s[3,0,2,1]=(0.260642362402)*x[0]**o+(0.585261059676)*x[0]+(-0.226552249351)*x[1]**o+(-0.22018440207)*x[1]+(-0.913385081787)*x[2]**o+(0.641093619341)*x[2]
        s[3,1,0,0]=(0.371513393673)*x[0]**o+(0.0170798820983)*x[0]+(0.564278651267)*x[1]**o+(0.680930162447)*x[1]+(-0.706423458637)*x[2]**o+(0.0764206259648)*x[2]
        s[3,1,0,1]=(-0.948003266655)*x[0]**o+(0.491205780333)*x[0]+(-0.758784222468)*x[1]**o+(0.963242289935)*x[1]+(0.598059292168)*x[2]**o+(0.864032824015)*x[2]
        s[3,1,1,0]=(-0.238580509808)*x[0]**o+(-0.604062371822)*x[0]+(0.575641671747)*x[1]**o+(0.36375183411)*x[1]+(-0.972736699609)*x[2]**o+(0.880286281096)*x[2]
        s[3,1,1,1]=(0.0699909126019)*x[0]**o+(0.134320946509)*x[0]+(0.610933092241)*x[1]**o+(0.371352827221)*x[1]+(0.0323703390883)*x[2]**o+(-0.274126481723)*x[2]
        s[3,1,2,0]=(0.453740440957)*x[0]**o+(-0.943147597599)*x[0]+(0.881661129342)*x[1]**o+(-0.877404630304)*x[1]+(-0.565813043124)*x[2]**o+(0.730370433019)*x[2]
        s[3,1,2,1]=(0.55962110239)*x[0]**o+(-0.304753265168)*x[0]+(-0.446631806503)*x[1]**o+(-0.247296550227)*x[1]+(0.704691317199)*x[2]**o+(0.885258048422)*x[2]
        s[3,2,0,0]=(0.165892354646)*x[0]**o+(-0.910149277809)*x[0]+(0.512472386138)*x[1]**o+(0.783845430116)*x[1]+(0.748568950882)*x[2]**o+(0.279356093192)*x[2]
        s[3,2,0,1]=(-0.152647132155)*x[0]**o+(0.59980439016)*x[0]+(0.75599611367)*x[1]**o+(-0.492517161389)*x[1]+(-0.510828418476)*x[2]**o+(0.813006872834)*x[2]
        s[3,2,1,0]=(-0.732207745061)*x[0]**o+(-0.304109063819)*x[0]+(0.330171325627)*x[1]**o+(-0.934563542072)*x[1]+(0.529201115138)*x[2]**o+(0.837219324241)*x[2]
        s[3,2,1,1]=(0.697323448629)*x[0]**o+(0.34999597431)*x[0]+(0.67453985446)*x[1]**o+(-0.607043070971)*x[1]+(0.774743224308)*x[2]**o+(0.560978199188)*x[2]
        s[3,2,2,0]=(-0.826819922732)*x[0]**o+(-0.0808978789297)*x[0]+(0.391865159047)*x[1]**o+(-0.34035597887)*x[1]+(0.129988822983)*x[2]**o+(0.0782066826859)*x[2]
        s[3,2,2,1]=(-0.503921164807)*x[0]**o+(0.229433648691)*x[0]+(-0.224244626319)*x[1]**o+(0.307825062312)*x[1]+(0.778558288408)*x[2]**o+(0.236537878219)*x[2]
        s[3,3,0,0]=(-0.465572503617)*x[0]**o+(-0.505092598655)*x[0]+(-0.285232464978)*x[1]**o+(0.295822892399)*x[1]+(-0.678626607589)*x[2]**o+(-0.762910104233)*x[2]
        s[3,3,0,1]=(-0.313839797733)*x[0]**o+(-0.866968909594)*x[0]+(0.193089207451)*x[1]**o+(0.350572793209)*x[1]+(0.0985650768561)*x[2]**o+(0.61459734708)*x[2]
        s[3,3,1,0]=(-0.686947068473)*x[0]**o+(0.803092292826)*x[0]+(0.229346725739)*x[1]**o+(-0.523883629849)*x[1]+(-0.825851523578)*x[2]**o+(0.911991149266)*x[2]
        s[3,3,1,1]=(-0.0229534891291)*x[0]**o+(0.459587997571)*x[0]+(-0.0323375614527)*x[1]**o+(-0.960355649174)*x[1]+(-0.157463174284)*x[2]**o+(0.721625063029)*x[2]
        s[3,3,2,0]=(-0.244675134158)*x[0]**o+(-0.750815470458)*x[0]+(0.350391343156)*x[1]**o+(0.246114995378)*x[1]+(-0.299954768936)*x[2]**o+(0.621800581218)*x[2]
        s[3,3,2,1]=(0.517173606541)*x[0]**o+(0.175764075574)*x[0]+(0.423442626185)*x[1]**o+(0.803990330201)*x[1]+(0.286217667564)*x[2]**o+(-0.471348114366)*x[2]
        s[3,4,0,0]=(-0.351936186325)*x[0]**o+(-0.0279135493684)*x[0]+(0.143368480037)*x[1]**o+(-0.0914938899855)*x[1]+(-0.160075587837)*x[2]**o+(-0.504374107434)*x[2]
        s[3,4,0,1]=(0.585914411245)*x[0]**o+(0.274811379635)*x[0]+(-0.177534785865)*x[1]**o+(0.567196593954)*x[1]+(-0.510174432853)*x[2]**o+(0.930747960349)*x[2]
        s[3,4,1,0]=(0.992675673771)*x[0]**o+(-0.709559401209)*x[0]+(0.996108059004)*x[1]**o+(-0.264682085878)*x[1]+(0.336432533695)*x[2]**o+(0.439349601512)*x[2]
        s[3,4,1,1]=(-0.383904129095)*x[0]**o+(-0.640989586457)*x[0]+(0.876802160463)*x[1]**o+(-0.941096425806)*x[1]+(0.636317329751)*x[2]**o+(-0.922925940931)*x[2]
        s[3,4,2,0]=(0.215934904956)*x[0]**o+(-0.0763907072691)*x[0]+(0.864686802474)*x[1]**o+(-0.100958345332)*x[1]+(-0.545187503302)*x[2]**o+(0.896671191263)*x[2]
        s[3,4,2,1]=(-0.53798548609)*x[0]**o+(0.189564902905)*x[0]+(-0.352161561502)*x[1]**o+(0.726662085086)*x[1]+(0.514655541645)*x[2]**o+(-0.36993932162)*x[2]
        ref[0,0,0,0]=(-1.0478690215)*(1+2.*(dim-1.)/(o+1.))+(-1.37379408255)*dim
        ref[0,0,0,1]=(-1.56020467821)*(1+2.*(dim-1.)/(o+1.))+(-0.356960065085)*dim
        ref[0,0,1,0]=(0.145052542612)*(1+2.*(dim-1.)/(o+1.))+(2.26326547)*dim
        ref[0,0,1,1]=(0.190889132968)*(1+2.*(dim-1.)/(o+1.))+(0.691410883628)*dim
        ref[0,0,2,0]=(-1.27560084714)*(1+2.*(dim-1.)/(o+1.))+(-0.864491371539)*dim
        ref[0,0,2,1]=(-0.123205855702)*(1+2.*(dim-1.)/(o+1.))+(-1.8575895333)*dim
        ref[0,1,0,0]=(1.37764034388)*(1+2.*(dim-1.)/(o+1.))+(1.29579085605)*dim
        ref[0,1,0,1]=(-0.853921100343)*(1+2.*(dim-1.)/(o+1.))+(0.0710639316977)*dim
        ref[0,1,1,0]=(-1.43817979196)*(1+2.*(dim-1.)/(o+1.))+(1.00159875329)*dim
        ref[0,1,1,1]=(-1.13236121085)*(1+2.*(dim-1.)/(o+1.))+(-0.479647376205)*dim
        ref[0,1,2,0]=(-0.358848515523)*(1+2.*(dim-1.)/(o+1.))+(0.194938214483)*dim
        ref[0,1,2,1]=(-1.42206096019)*(1+2.*(dim-1.)/(o+1.))+(0.638503485369)*dim
        ref[0,2,0,0]=(0.506238251094)*(1+2.*(dim-1.)/(o+1.))+(2.23445077612)*dim
        ref[0,2,0,1]=(1.70549810645)*(1+2.*(dim-1.)/(o+1.))+(-1.63552693861)*dim
        ref[0,2,1,0]=(-0.922207133162)*(1+2.*(dim-1.)/(o+1.))+(-0.273433842582)*dim
        ref[0,2,1,1]=(1.25447606776)*(1+2.*(dim-1.)/(o+1.))+(2.14930964604)*dim
        ref[0,2,2,0]=(-0.544800659874)*(1+2.*(dim-1.)/(o+1.))+(0.0846331844061)*dim
        ref[0,2,2,1]=(-2.05762595196)*(1+2.*(dim-1.)/(o+1.))+(-0.485597478209)*dim
        ref[0,3,0,0]=(-0.587671140011)*(1+2.*(dim-1.)/(o+1.))+(0.84975773226)*dim
        ref[0,3,0,1]=(-0.659519022871)*(1+2.*(dim-1.)/(o+1.))+(-0.163950579871)*dim
        ref[0,3,1,0]=(0.73538407924)*(1+2.*(dim-1.)/(o+1.))+(1.09834250225)*dim
        ref[0,3,1,1]=(0.44796115756)*(1+2.*(dim-1.)/(o+1.))+(-0.27259983388)*dim
        ref[0,3,2,0]=(1.50367624283)*(1+2.*(dim-1.)/(o+1.))+(-0.54373966743)*dim
        ref[0,3,2,1]=(-0.740361055659)*(1+2.*(dim-1.)/(o+1.))+(1.57767520634)*dim
        ref[0,4,0,0]=(-1.11785477232)*(1+2.*(dim-1.)/(o+1.))+(0.235876809967)*dim
        ref[0,4,0,1]=(0.102493800852)*(1+2.*(dim-1.)/(o+1.))+(0.542275690805)*dim
        ref[0,4,1,0]=(1.41060924957)*(1+2.*(dim-1.)/(o+1.))+(0.692508793485)*dim
        ref[0,4,1,1]=(-0.427180180157)*(1+2.*(dim-1.)/(o+1.))+(0.205460491821)*dim
        ref[0,4,2,0]=(0.482429058578)*(1+2.*(dim-1.)/(o+1.))+(-0.000699098596959)*dim
        ref[0,4,2,1]=(0.585888305456)*(1+2.*(dim-1.)/(o+1.))+(0.287527955135)*dim
        ref[1,0,0,0]=(-1.9538518401)*(1+2.*(dim-1.)/(o+1.))+(-0.303636324337)*dim
        ref[1,0,0,1]=(0.0447970703505)*(1+2.*(dim-1.)/(o+1.))+(-0.447844206411)*dim
        ref[1,0,1,0]=(1.20316974235)*(1+2.*(dim-1.)/(o+1.))+(0.23477841774)*dim
        ref[1,0,1,1]=(1.73792134444)*(1+2.*(dim-1.)/(o+1.))+(0.427687799004)*dim
        ref[1,0,2,0]=(-1.66806639148)*(1+2.*(dim-1.)/(o+1.))+(-0.19600739032)*dim
        ref[1,0,2,1]=(-0.802983671714)*(1+2.*(dim-1.)/(o+1.))+(0.716812877498)*dim
        ref[1,1,0,0]=(1.66647686155)*(1+2.*(dim-1.)/(o+1.))+(-0.118815477944)*dim
        ref[1,1,0,1]=(1.26459782085)*(1+2.*(dim-1.)/(o+1.))+(0.207011474112)*dim
        ref[1,1,1,0]=(-0.200019585637)*(1+2.*(dim-1.)/(o+1.))+(-0.32829798046)*dim
        ref[1,1,1,1]=(1.48504729536)*(1+2.*(dim-1.)/(o+1.))+(-0.859627184532)*dim
        ref[1,1,2,0]=(-0.589994584106)*(1+2.*(dim-1.)/(o+1.))+(-0.336777145479)*dim
        ref[1,1,2,1]=(0.176785574887)*(1+2.*(dim-1.)/(o+1.))+(1.12050351028)*dim
        ref[1,2,0,0]=(1.67207577232)*(1+2.*(dim-1.)/(o+1.))+(0.287132246505)*dim
        ref[1,2,0,1]=(0.79377045136)*(1+2.*(dim-1.)/(o+1.))+(1.68933257489)*dim
        ref[1,2,1,0]=(1.60022083712)*(1+2.*(dim-1.)/(o+1.))+(-1.47929366631)*dim
        ref[1,2,1,1]=(-0.74565790135)*(1+2.*(dim-1.)/(o+1.))+(0.821587540478)*dim
        ref[1,2,2,0]=(0.628684097552)*(1+2.*(dim-1.)/(o+1.))+(0.19722645266)*dim
        ref[1,2,2,1]=(0.801984774477)*(1+2.*(dim-1.)/(o+1.))+(1.20161146291)*dim
        ref[1,3,0,0]=(1.21312160658)*(1+2.*(dim-1.)/(o+1.))+(-0.382457619353)*dim
        ref[1,3,0,1]=(-0.316785537645)*(1+2.*(dim-1.)/(o+1.))+(2.07650268732)*dim
        ref[1,3,1,0]=(-1.40832807711)*(1+2.*(dim-1.)/(o+1.))+(0.09565688166)*dim
        ref[1,3,1,1]=(0.851590861291)*(1+2.*(dim-1.)/(o+1.))+(1.65472250697)*dim
        ref[1,3,2,0]=(0.638787443433)*(1+2.*(dim-1.)/(o+1.))+(0.656304626504)*dim
        ref[1,3,2,1]=(-1.60742081722)*(1+2.*(dim-1.)/(o+1.))+(-1.30862324678)*dim
        ref[1,4,0,0]=(1.30721238256)*(1+2.*(dim-1.)/(o+1.))+(1.10242119303)*dim
        ref[1,4,0,1]=(0.942436070125)*(1+2.*(dim-1.)/(o+1.))+(0.478420914396)*dim
        ref[1,4,1,0]=(-1.92140780114)*(1+2.*(dim-1.)/(o+1.))+(0.219208442793)*dim
        ref[1,4,1,1]=(1.02814640573)*(1+2.*(dim-1.)/(o+1.))+(-0.185757585533)*dim
        ref[1,4,2,0]=(-0.102630055442)*(1+2.*(dim-1.)/(o+1.))+(-0.118251334341)*dim
        ref[1,4,2,1]=(-0.851107260444)*(1+2.*(dim-1.)/(o+1.))+(0.754976996591)*dim
        ref[2,0,0,0]=(-1.79264309703)*(1+2.*(dim-1.)/(o+1.))+(0.81715861598)*dim
        ref[2,0,0,1]=(-0.398673742984)*(1+2.*(dim-1.)/(o+1.))+(-0.191300222501)*dim
        ref[2,0,1,0]=(0.336422367917)*(1+2.*(dim-1.)/(o+1.))+(1.23494373887)*dim
        ref[2,0,1,1]=(0.439862153544)*(1+2.*(dim-1.)/(o+1.))+(-0.189125193964)*dim
        ref[2,0,2,0]=(-0.302387609436)*(1+2.*(dim-1.)/(o+1.))+(-1.91348277605)*dim
        ref[2,0,2,1]=(-0.00459679351666)*(1+2.*(dim-1.)/(o+1.))+(-2.20505884102)*dim
        ref[2,1,0,0]=(-0.810623288547)*(1+2.*(dim-1.)/(o+1.))+(1.09759323486)*dim
        ref[2,1,0,1]=(-0.0649209144233)*(1+2.*(dim-1.)/(o+1.))+(1.55888238606)*dim
        ref[2,1,1,0]=(-0.10133772828)*(1+2.*(dim-1.)/(o+1.))+(1.47377461574)*dim
        ref[2,1,1,1]=(0.336386487342)*(1+2.*(dim-1.)/(o+1.))+(-1.6646059276)*dim
        ref[2,1,2,0]=(-0.267773640646)*(1+2.*(dim-1.)/(o+1.))+(1.35069677515)*dim
        ref[2,1,2,1]=(-0.972979498249)*(1+2.*(dim-1.)/(o+1.))+(0.419995881524)*dim
        ref[2,2,0,0]=(0.617929304052)*(1+2.*(dim-1.)/(o+1.))+(-0.473075358283)*dim
        ref[2,2,0,1]=(-2.14485137529)*(1+2.*(dim-1.)/(o+1.))+(1.17580591431)*dim
        ref[2,2,1,0]=(-0.0933552648497)*(1+2.*(dim-1.)/(o+1.))+(-1.56078253618)*dim
        ref[2,2,1,1]=(-0.094433407695)*(1+2.*(dim-1.)/(o+1.))+(0.712083765676)*dim
        ref[2,2,2,0]=(-2.72177737308)*(1+2.*(dim-1.)/(o+1.))+(-0.520902788881)*dim
        ref[2,2,2,1]=(-1.63945374313)*(1+2.*(dim-1.)/(o+1.))+(-0.166293541762)*dim
        ref[2,3,0,0]=(0.995138588408)*(1+2.*(dim-1.)/(o+1.))+(-1.24064158297)*dim
        ref[2,3,0,1]=(0.534207490651)*(1+2.*(dim-1.)/(o+1.))+(1.72955354449)*dim
        ref[2,3,1,0]=(-1.03463843312)*(1+2.*(dim-1.)/(o+1.))+(-0.169637472029)*dim
        ref[2,3,1,1]=(-1.17458463318)*(1+2.*(dim-1.)/(o+1.))+(-1.00521594733)*dim
        ref[2,3,2,0]=(0.159668033037)*(1+2.*(dim-1.)/(o+1.))+(-0.739273896161)*dim
        ref[2,3,2,1]=(-1.58911936813)*(1+2.*(dim-1.)/(o+1.))+(-0.22475081215)*dim
        ref[2,4,0,0]=(0.461647091515)*(1+2.*(dim-1.)/(o+1.))+(0.786582647531)*dim
        ref[2,4,0,1]=(1.78011997479)*(1+2.*(dim-1.)/(o+1.))+(1.59208263184)*dim
        ref[2,4,1,0]=(0.785187875018)*(1+2.*(dim-1.)/(o+1.))+(-0.621649101592)*dim
        ref[2,4,1,1]=(0.416367847058)*(1+2.*(dim-1.)/(o+1.))+(-0.740836474409)*dim
        ref[2,4,2,0]=(-0.651469503672)*(1+2.*(dim-1.)/(o+1.))+(-0.569385132282)*dim
        ref[2,4,2,1]=(0.774209323076)*(1+2.*(dim-1.)/(o+1.))+(0.90138559811)*dim
        ref[3,0,0,0]=(0.640077400272)*(1+2.*(dim-1.)/(o+1.))+(1.1985417184)*dim
        ref[3,0,0,1]=(-0.875805316424)*(1+2.*(dim-1.)/(o+1.))+(-0.222320242899)*dim
        ref[3,0,1,0]=(1.97369748041)*(1+2.*(dim-1.)/(o+1.))+(-0.406664940521)*dim
        ref[3,0,1,1]=(0.161089898621)*(1+2.*(dim-1.)/(o+1.))+(-0.0741768265174)*dim
        ref[3,0,2,0]=(0.551906111004)*(1+2.*(dim-1.)/(o+1.))+(0.940193212958)*dim
        ref[3,0,2,1]=(-0.879294968735)*(1+2.*(dim-1.)/(o+1.))+(1.00617027695)*dim
        ref[3,1,0,0]=(0.229368586303)*(1+2.*(dim-1.)/(o+1.))+(0.77443067051)*dim
        ref[3,1,0,1]=(-1.10872819696)*(1+2.*(dim-1.)/(o+1.))+(2.31848089428)*dim
        ref[3,1,1,0]=(-0.63567553767)*(1+2.*(dim-1.)/(o+1.))+(0.639975743384)*dim
        ref[3,1,1,1]=(0.713294343931)*(1+2.*(dim-1.)/(o+1.))+(0.231547292007)*dim
        ref[3,1,2,0]=(0.769588527176)*(1+2.*(dim-1.)/(o+1.))+(-1.09018179488)*dim
        ref[3,1,2,1]=(0.817680613086)*(1+2.*(dim-1.)/(o+1.))+(0.333208233027)*dim
        ref[3,2,0,0]=(1.42693369167)*(1+2.*(dim-1.)/(o+1.))+(0.153052245499)*dim
        ref[3,2,0,1]=(0.0925205630387)*(1+2.*(dim-1.)/(o+1.))+(0.920294101604)*dim
        ref[3,2,1,0]=(0.127164695704)*(1+2.*(dim-1.)/(o+1.))+(-0.40145328165)*dim
        ref[3,2,1,1]=(2.1466065274)*(1+2.*(dim-1.)/(o+1.))+(0.303931102527)*dim
        ref[3,2,2,0]=(-0.304965940702)*(1+2.*(dim-1.)/(o+1.))+(-0.343047175114)*dim
        ref[3,2,2,1]=(0.0503924972819)*(1+2.*(dim-1.)/(o+1.))+(0.773796589222)*dim
        ref[3,3,0,0]=(-1.42943157618)*(1+2.*(dim-1.)/(o+1.))+(-0.972179810489)*dim
        ref[3,3,0,1]=(-0.0221855134254)*(1+2.*(dim-1.)/(o+1.))+(0.0982012306951)*dim
        ref[3,3,1,0]=(-1.28345186631)*(1+2.*(dim-1.)/(o+1.))+(1.19119981224)*dim
        ref[3,3,1,1]=(-0.212754224866)*(1+2.*(dim-1.)/(o+1.))+(0.220857411425)*dim
        ref[3,3,2,0]=(-0.194238559937)*(1+2.*(dim-1.)/(o+1.))+(0.117100106138)*dim
        ref[3,3,2,1]=(1.22683390029)*(1+2.*(dim-1.)/(o+1.))+(0.508406291409)*dim
        ref[3,4,0,0]=(-0.368643294124)*(1+2.*(dim-1.)/(o+1.))+(-0.623781546787)*dim
        ref[3,4,0,1]=(-0.101794807473)*(1+2.*(dim-1.)/(o+1.))+(1.77275593394)*dim
        ref[3,4,1,0]=(2.32521626647)*(1+2.*(dim-1.)/(o+1.))+(-0.534891885575)*dim
        ref[3,4,1,1]=(1.12921536112)*(1+2.*(dim-1.)/(o+1.))+(-2.50501195319)*dim
        ref[3,4,2,0]=(0.535434204129)*(1+2.*(dim-1.)/(o+1.))+(0.719322138661)*dim
        ref[3,4,2,1]=(-0.375491505946)*(1+2.*(dim-1.)/(o+1.))+(0.546287666371)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_Solution_rank0(self):
      """
      tests integral of rank 0 Symbol on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.747700242689)*x[0]**o+(-0.24253900484)*x[0]+(0.337715187384)*x[1]**o+(0.338764998592)*x[1]
        ref=(-0.409985055305)*(1+2.*(dim-1.)/(o+1.))+(0.0962259937514)*dim
      else:
        s=(-0.25178656924)*x[0]**o+(-0.295938166993)*x[0]+(0.0692176296902)*x[1]**o+(-0.514895147995)*x[1]+(0.34638140815)*x[2]**o+(-0.702417755803)*x[2]
        ref=(0.163812468601)*(1+2.*(dim-1.)/(o+1.))+(-1.51325107079)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_Solution_rank1(self):
      """
      tests integral of rank 1 Symbol on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(0.555314122492)*x[0]**o+(-0.808960080698)*x[0]+(0.884426560701)*x[1]**o+(-0.243549906992)*x[1]
        s[1]=(-0.0719219833378)*x[0]**o+(-0.403661400366)*x[0]+(0.666085789176)*x[1]**o+(-0.914772770267)*x[1]
        ref[0]=(1.43974068319)*(1+2.*(dim-1.)/(o+1.))+(-1.05250998769)*dim
        ref[1]=(0.594163805838)*(1+2.*(dim-1.)/(o+1.))+(-1.31843417063)*dim
      else:
        s[0]=(-0.826331014544)*x[0]**o+(-0.321975774952)*x[0]+(0.112378767915)*x[1]**o+(-0.756344985678)*x[1]+(0.719456760908)*x[2]**o+(-0.353602934037)*x[2]
        s[1]=(0.869195210566)*x[0]**o+(-0.315975726841)*x[0]+(0.712421301529)*x[1]**o+(0.215376989329)*x[1]+(0.964920869405)*x[2]**o+(0.922113174792)*x[2]
        ref[0]=(0.00550451427885)*(1+2.*(dim-1.)/(o+1.))+(-1.43192369467)*dim
        ref[1]=(2.5465373815)*(1+2.*(dim-1.)/(o+1.))+(0.82151443728)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_Solution_rank2(self):
      """
      tests integral of rank 2 Symbol on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.271641608649)*x[0]**o+(-0.685616858784)*x[0]+(0.644560357341)*x[1]**o+(0.294292988645)*x[1]
        s[0,1]=(-0.868199164404)*x[0]**o+(-0.352185589942)*x[0]+(-0.939001057807)*x[1]**o+(-0.806858466969)*x[1]
        s[0,2]=(-0.908891077111)*x[0]**o+(0.75605446917)*x[0]+(-0.921669916524)*x[1]**o+(-0.375347220131)*x[1]
        s[0,3]=(0.508605238535)*x[0]**o+(0.478464667724)*x[0]+(-0.632750076182)*x[1]**o+(-0.0150029824707)*x[1]
        s[0,4]=(-0.886187964736)*x[0]**o+(0.397671872055)*x[0]+(-0.216676829324)*x[1]**o+(0.986585859674)*x[1]
        s[1,0]=(0.630864334183)*x[0]**o+(-0.722653479356)*x[0]+(0.211515635986)*x[1]**o+(0.83331541895)*x[1]
        s[1,1]=(-0.354681281315)*x[0]**o+(-0.848733894579)*x[0]+(-0.101868246651)*x[1]**o+(-0.505303431554)*x[1]
        s[1,2]=(0.685694963182)*x[0]**o+(-0.950829894589)*x[0]+(-0.31792598055)*x[1]**o+(0.426249652943)*x[1]
        s[1,3]=(-0.614038433138)*x[0]**o+(0.723563718847)*x[0]+(0.752373220947)*x[1]**o+(0.613330337959)*x[1]
        s[1,4]=(-0.998818105353)*x[0]**o+(0.269274342639)*x[0]+(-0.902068964726)*x[1]**o+(-0.191309514428)*x[1]
        s[2,0]=(0.249323085595)*x[0]**o+(0.411304825013)*x[0]+(0.312377963164)*x[1]**o+(-0.311446803448)*x[1]
        s[2,1]=(-0.290453739481)*x[0]**o+(-0.202745911645)*x[0]+(-0.309021366331)*x[1]**o+(-0.173018322631)*x[1]
        s[2,2]=(0.691837531968)*x[0]**o+(0.672203641337)*x[0]+(0.0217882582663)*x[1]**o+(-0.776821250479)*x[1]
        s[2,3]=(0.00878696630163)*x[0]**o+(-0.565790699611)*x[0]+(-0.748591951135)*x[1]**o+(-0.46249464823)*x[1]
        s[2,4]=(-0.638178053099)*x[0]**o+(0.898061575109)*x[0]+(0.498715053696)*x[1]**o+(-0.236164377825)*x[1]
        s[3,0]=(0.843383855229)*x[0]**o+(0.652320113529)*x[0]+(0.34384880146)*x[1]**o+(-0.539878095901)*x[1]
        s[3,1]=(-0.0571664516277)*x[0]**o+(-0.658773983784)*x[0]+(0.160437286579)*x[1]**o+(-0.805774084186)*x[1]
        s[3,2]=(0.340124131227)*x[0]**o+(-0.464143943808)*x[0]+(-0.215724355021)*x[1]**o+(0.206995713005)*x[1]
        s[3,3]=(0.825533013397)*x[0]**o+(0.695982840481)*x[0]+(0.355542963602)*x[1]**o+(-0.401849802264)*x[1]
        s[3,4]=(-0.367312805197)*x[0]**o+(-0.012173639311)*x[0]+(0.581566155275)*x[1]**o+(-0.650284343373)*x[1]
        ref[0,0]=(0.91620196599)*(1+2.*(dim-1.)/(o+1.))+(-0.39132387014)*dim
        ref[0,1]=(-1.80720022221)*(1+2.*(dim-1.)/(o+1.))+(-1.15904405691)*dim
        ref[0,2]=(-1.83056099363)*(1+2.*(dim-1.)/(o+1.))+(0.380707249038)*dim
        ref[0,3]=(-0.124144837647)*(1+2.*(dim-1.)/(o+1.))+(0.463461685254)*dim
        ref[0,4]=(-1.10286479406)*(1+2.*(dim-1.)/(o+1.))+(1.38425773173)*dim
        ref[1,0]=(0.842379970168)*(1+2.*(dim-1.)/(o+1.))+(0.110661939594)*dim
        ref[1,1]=(-0.456549527966)*(1+2.*(dim-1.)/(o+1.))+(-1.35403732613)*dim
        ref[1,2]=(0.367768982632)*(1+2.*(dim-1.)/(o+1.))+(-0.524580241646)*dim
        ref[1,3]=(0.138334787809)*(1+2.*(dim-1.)/(o+1.))+(1.33689405681)*dim
        ref[1,4]=(-1.90088707008)*(1+2.*(dim-1.)/(o+1.))+(0.0779648282107)*dim
        ref[2,0]=(0.561701048758)*(1+2.*(dim-1.)/(o+1.))+(0.0998580215651)*dim
        ref[2,1]=(-0.599475105813)*(1+2.*(dim-1.)/(o+1.))+(-0.375764234276)*dim
        ref[2,2]=(0.713625790235)*(1+2.*(dim-1.)/(o+1.))+(-0.104617609143)*dim
        ref[2,3]=(-0.739804984833)*(1+2.*(dim-1.)/(o+1.))+(-1.02828534784)*dim
        ref[2,4]=(-0.139462999403)*(1+2.*(dim-1.)/(o+1.))+(0.661897197284)*dim
        ref[3,0]=(1.18723265669)*(1+2.*(dim-1.)/(o+1.))+(0.112442017629)*dim
        ref[3,1]=(0.103270834951)*(1+2.*(dim-1.)/(o+1.))+(-1.46454806797)*dim
        ref[3,2]=(0.124399776206)*(1+2.*(dim-1.)/(o+1.))+(-0.257148230803)*dim
        ref[3,3]=(1.181075977)*(1+2.*(dim-1.)/(o+1.))+(0.294133038216)*dim
        ref[3,4]=(0.214253350078)*(1+2.*(dim-1.)/(o+1.))+(-0.662457982684)*dim
      else:
        s[0,0]=(0.488698481646)*x[0]**o+(-0.0857636443438)*x[0]+(0.879463791392)*x[1]**o+(-0.748151251773)*x[1]+(0.342272219635)*x[2]**o+(0.148033760154)*x[2]
        s[0,1]=(0.716943874304)*x[0]**o+(-0.582838388256)*x[0]+(0.799901107796)*x[1]**o+(-0.814196005238)*x[1]+(0.229983496729)*x[2]**o+(-0.153919124028)*x[2]
        s[0,2]=(-0.898513198581)*x[0]**o+(0.380948693198)*x[0]+(0.357237014893)*x[1]**o+(0.650057181845)*x[1]+(0.192484147287)*x[2]**o+(-0.00114028159222)*x[2]
        s[0,3]=(0.882708886033)*x[0]**o+(0.412063277674)*x[0]+(-0.801825884205)*x[1]**o+(0.421365647832)*x[1]+(0.0732340922267)*x[2]**o+(0.831590828322)*x[2]
        s[0,4]=(-0.893541294976)*x[0]**o+(0.616353784204)*x[0]+(0.43931453774)*x[1]**o+(0.60163593029)*x[1]+(0.171225928123)*x[2]**o+(-0.809624161102)*x[2]
        s[1,0]=(0.3743694715)*x[0]**o+(0.973362616567)*x[0]+(-0.790022350688)*x[1]**o+(0.438346450916)*x[1]+(-0.85644858607)*x[2]**o+(0.861754412538)*x[2]
        s[1,1]=(0.0682853717145)*x[0]**o+(-0.483847902471)*x[0]+(0.830085512097)*x[1]**o+(0.330508826392)*x[1]+(-0.341254677789)*x[2]**o+(0.0388136714901)*x[2]
        s[1,2]=(0.740241060442)*x[0]**o+(0.569699186818)*x[0]+(0.292637941654)*x[1]**o+(-0.126938078274)*x[1]+(-0.400665686855)*x[2]**o+(-0.0981290395971)*x[2]
        s[1,3]=(0.611312886004)*x[0]**o+(0.570096324275)*x[0]+(0.222499932388)*x[1]**o+(-0.276552061893)*x[1]+(-0.370405232883)*x[2]**o+(-0.91757884865)*x[2]
        s[1,4]=(-0.844423209218)*x[0]**o+(-0.357603990072)*x[0]+(-0.291569596715)*x[1]**o+(0.127364564811)*x[1]+(-0.427115349787)*x[2]**o+(-0.689895532918)*x[2]
        s[2,0]=(-0.798050448891)*x[0]**o+(-0.197062766104)*x[0]+(-0.363892466443)*x[1]**o+(0.154013357585)*x[1]+(-0.398682782201)*x[2]**o+(-0.529551264949)*x[2]
        s[2,1]=(-0.0912777859374)*x[0]**o+(-0.675884283313)*x[0]+(0.516144657124)*x[1]**o+(-0.53265938856)*x[1]+(0.95476185698)*x[2]**o+(-0.313549200036)*x[2]
        s[2,2]=(0.868959641933)*x[0]**o+(0.962373087968)*x[0]+(-0.848379282457)*x[1]**o+(0.820817107146)*x[1]+(0.498221865324)*x[2]**o+(-0.885755811184)*x[2]
        s[2,3]=(0.552181728294)*x[0]**o+(-0.943385218578)*x[0]+(-0.650886151517)*x[1]**o+(0.204702323464)*x[1]+(-0.419413465102)*x[2]**o+(0.761759803195)*x[2]
        s[2,4]=(0.0446686935799)*x[0]**o+(-0.850084331283)*x[0]+(0.215197552732)*x[1]**o+(0.0524953736999)*x[1]+(-0.828056084121)*x[2]**o+(-0.960944367985)*x[2]
        s[3,0]=(-0.475014597092)*x[0]**o+(-0.100740967917)*x[0]+(0.76769378138)*x[1]**o+(0.846896186419)*x[1]+(0.794108335473)*x[2]**o+(-0.34345830075)*x[2]
        s[3,1]=(0.10453345316)*x[0]**o+(-0.677797655695)*x[0]+(0.471066711536)*x[1]**o+(0.534601952265)*x[1]+(-0.685589576412)*x[2]**o+(-0.025255709402)*x[2]
        s[3,2]=(0.0410162645138)*x[0]**o+(-0.948466999545)*x[0]+(0.147271170264)*x[1]**o+(0.17422632965)*x[1]+(0.425319252082)*x[2]**o+(-0.395792817384)*x[2]
        s[3,3]=(0.301093567659)*x[0]**o+(-0.48405705123)*x[0]+(-0.681930643655)*x[1]**o+(0.848540291162)*x[1]+(0.815437001935)*x[2]**o+(-0.734620301961)*x[2]
        s[3,4]=(-0.586209686649)*x[0]**o+(-0.747006165956)*x[0]+(0.0368543812482)*x[1]**o+(-0.976301323297)*x[1]+(0.748206708653)*x[2]**o+(0.833845422962)*x[2]
        ref[0,0]=(1.71043449267)*(1+2.*(dim-1.)/(o+1.))+(-0.685881135963)*dim
        ref[0,1]=(1.74682847883)*(1+2.*(dim-1.)/(o+1.))+(-1.55095351752)*dim
        ref[0,2]=(-0.348792036401)*(1+2.*(dim-1.)/(o+1.))+(1.02986559345)*dim
        ref[0,3]=(0.154117094054)*(1+2.*(dim-1.)/(o+1.))+(1.66501975383)*dim
        ref[0,4]=(-0.283000829113)*(1+2.*(dim-1.)/(o+1.))+(0.408365553392)*dim
        ref[1,0]=(-1.27210146526)*(1+2.*(dim-1.)/(o+1.))+(2.27346348002)*dim
        ref[1,1]=(0.557116206022)*(1+2.*(dim-1.)/(o+1.))+(-0.114525404589)*dim
        ref[1,2]=(0.632213315241)*(1+2.*(dim-1.)/(o+1.))+(0.344632068947)*dim
        ref[1,3]=(0.463407585509)*(1+2.*(dim-1.)/(o+1.))+(-0.624034586267)*dim
        ref[1,4]=(-1.56310815572)*(1+2.*(dim-1.)/(o+1.))+(-0.920134958179)*dim
        ref[2,0]=(-1.56062569754)*(1+2.*(dim-1.)/(o+1.))+(-0.572600673468)*dim
        ref[2,1]=(1.37962872817)*(1+2.*(dim-1.)/(o+1.))+(-1.52209287191)*dim
        ref[2,2]=(0.5188022248)*(1+2.*(dim-1.)/(o+1.))+(0.897434383931)*dim
        ref[2,3]=(-0.518117888325)*(1+2.*(dim-1.)/(o+1.))+(0.0230769080811)*dim
        ref[2,4]=(-0.568189837809)*(1+2.*(dim-1.)/(o+1.))+(-1.75853332557)*dim
        ref[3,0]=(1.08678751976)*(1+2.*(dim-1.)/(o+1.))+(0.402696917752)*dim
        ref[3,1]=(-0.109989411716)*(1+2.*(dim-1.)/(o+1.))+(-0.168451412832)*dim
        ref[3,2]=(0.61360668686)*(1+2.*(dim-1.)/(o+1.))+(-1.17003348728)*dim
        ref[3,3]=(0.434599925938)*(1+2.*(dim-1.)/(o+1.))+(-0.370137062028)*dim
        ref[3,4]=(0.198851403252)*(1+2.*(dim-1.)/(o+1.))+(-0.88946206629)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_Solution_rank3(self):
      """
      tests integral of rank 3 Symbol on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.591558717335)*x[0]**o+(-0.671839230736)*x[0]+(0.107743853024)*x[1]**o+(0.595347539232)*x[1]
        s[0,0,1]=(0.25500384698)*x[0]**o+(0.944399712188)*x[0]+(-0.999249484326)*x[1]**o+(-0.495212132457)*x[1]
        s[0,1,0]=(0.236244996704)*x[0]**o+(0.704807264803)*x[0]+(-0.242028416637)*x[1]**o+(-0.553320357705)*x[1]
        s[0,1,1]=(0.552969384309)*x[0]**o+(-0.941333552159)*x[0]+(-0.696785515448)*x[1]**o+(0.851994229748)*x[1]
        s[1,0,0]=(0.0953739386185)*x[0]**o+(-0.50227621093)*x[0]+(-0.994199466154)*x[1]**o+(0.975204139168)*x[1]
        s[1,0,1]=(0.0251612361822)*x[0]**o+(-0.984244259014)*x[0]+(0.928117852579)*x[1]**o+(0.0302634017269)*x[1]
        s[1,1,0]=(-0.800615084919)*x[0]**o+(0.71993242468)*x[0]+(0.714856567768)*x[1]**o+(-0.497753089167)*x[1]
        s[1,1,1]=(-0.916226186485)*x[0]**o+(0.81028396707)*x[0]+(-0.156900817441)*x[1]**o+(-0.093034093703)*x[1]
        s[2,0,0]=(-0.738732605581)*x[0]**o+(-0.77907436774)*x[0]+(0.948664797284)*x[1]**o+(-0.0764721201972)*x[1]
        s[2,0,1]=(-0.11146154208)*x[0]**o+(0.145299573751)*x[0]+(0.970129117378)*x[1]**o+(-0.272338564084)*x[1]
        s[2,1,0]=(-0.197948912143)*x[0]**o+(-0.902880824888)*x[0]+(0.579307317816)*x[1]**o+(-0.852996341605)*x[1]
        s[2,1,1]=(0.563711575633)*x[0]**o+(-0.288761476406)*x[0]+(0.7854621452)*x[1]**o+(-0.784647403092)*x[1]
        s[3,0,0]=(-0.116979739809)*x[0]**o+(0.964821934046)*x[0]+(0.441879236444)*x[1]**o+(0.245324297501)*x[1]
        s[3,0,1]=(-0.63895256779)*x[0]**o+(0.446532064378)*x[0]+(0.703018532945)*x[1]**o+(0.398978822243)*x[1]
        s[3,1,0]=(0.347022657024)*x[0]**o+(0.330266740052)*x[0]+(0.9044236351)*x[1]**o+(0.463728515326)*x[1]
        s[3,1,1]=(0.673344024583)*x[0]**o+(0.218503531592)*x[0]+(-0.981631285231)*x[1]**o+(0.493256781677)*x[1]
        s[4,0,0]=(-0.0337376408638)*x[0]**o+(0.006116830911)*x[0]+(-0.186073327302)*x[1]**o+(0.0809017607404)*x[1]
        s[4,0,1]=(0.624443889423)*x[0]**o+(0.727204963203)*x[0]+(0.724716349734)*x[1]**o+(-0.360798281504)*x[1]
        s[4,1,0]=(-0.316602217463)*x[0]**o+(-0.775985201313)*x[0]+(0.313467720837)*x[1]**o+(0.19768983092)*x[1]
        s[4,1,1]=(0.578665229532)*x[0]**o+(-0.962255191175)*x[0]+(0.341888555373)*x[1]**o+(-0.389440724436)*x[1]
        s[5,0,0]=(0.256912289915)*x[0]**o+(-0.171401072804)*x[0]+(-0.646582287213)*x[1]**o+(0.211366542089)*x[1]
        s[5,0,1]=(0.393310638354)*x[0]**o+(0.319973991926)*x[0]+(0.626363008584)*x[1]**o+(-0.0843826544875)*x[1]
        s[5,1,0]=(0.425476941925)*x[0]**o+(0.913999716353)*x[0]+(-0.534398944156)*x[1]**o+(-0.83386676719)*x[1]
        s[5,1,1]=(0.660737270239)*x[0]**o+(-0.828622795768)*x[0]+(0.488808193457)*x[1]**o+(-0.352537348251)*x[1]
        ref[0,0,0]=(0.699302570359)*(1+2.*(dim-1.)/(o+1.))+(-0.076491691504)*dim
        ref[0,0,1]=(-0.744245637345)*(1+2.*(dim-1.)/(o+1.))+(0.449187579731)*dim
        ref[0,1,0]=(-0.00578341993308)*(1+2.*(dim-1.)/(o+1.))+(0.151486907099)*dim
        ref[0,1,1]=(-0.14381613114)*(1+2.*(dim-1.)/(o+1.))+(-0.0893393224107)*dim
        ref[1,0,0]=(-0.898825527536)*(1+2.*(dim-1.)/(o+1.))+(0.472927928237)*dim
        ref[1,0,1]=(0.953279088761)*(1+2.*(dim-1.)/(o+1.))+(-0.953980857287)*dim
        ref[1,1,0]=(-0.0857585171509)*(1+2.*(dim-1.)/(o+1.))+(0.222179335513)*dim
        ref[1,1,1]=(-1.07312700393)*(1+2.*(dim-1.)/(o+1.))+(0.717249873367)*dim
        ref[2,0,0]=(0.209932191703)*(1+2.*(dim-1.)/(o+1.))+(-0.855546487937)*dim
        ref[2,0,1]=(0.858667575297)*(1+2.*(dim-1.)/(o+1.))+(-0.127038990333)*dim
        ref[2,1,0]=(0.381358405673)*(1+2.*(dim-1.)/(o+1.))+(-1.75587716649)*dim
        ref[2,1,1]=(1.34917372083)*(1+2.*(dim-1.)/(o+1.))+(-1.0734088795)*dim
        ref[3,0,0]=(0.324899496635)*(1+2.*(dim-1.)/(o+1.))+(1.21014623155)*dim
        ref[3,0,1]=(0.0640659651547)*(1+2.*(dim-1.)/(o+1.))+(0.84551088662)*dim
        ref[3,1,0]=(1.25144629212)*(1+2.*(dim-1.)/(o+1.))+(0.793995255378)*dim
        ref[3,1,1]=(-0.308287260648)*(1+2.*(dim-1.)/(o+1.))+(0.711760313269)*dim
        ref[4,0,0]=(-0.219810968166)*(1+2.*(dim-1.)/(o+1.))+(0.0870185916514)*dim
        ref[4,0,1]=(1.34916023916)*(1+2.*(dim-1.)/(o+1.))+(0.366406681699)*dim
        ref[4,1,0]=(-0.00313449662664)*(1+2.*(dim-1.)/(o+1.))+(-0.578295370393)*dim
        ref[4,1,1]=(0.920553784904)*(1+2.*(dim-1.)/(o+1.))+(-1.35169591561)*dim
        ref[5,0,0]=(-0.389669997298)*(1+2.*(dim-1.)/(o+1.))+(0.0399654692848)*dim
        ref[5,0,1]=(1.01967364694)*(1+2.*(dim-1.)/(o+1.))+(0.235591337439)*dim
        ref[5,1,0]=(-0.108922002231)*(1+2.*(dim-1.)/(o+1.))+(0.0801329491631)*dim
        ref[5,1,1]=(1.1495454637)*(1+2.*(dim-1.)/(o+1.))+(-1.18116014402)*dim
      else:
        s[0,0,0]=(0.267370933512)*x[0]**o+(0.155479029604)*x[0]+(0.483226114118)*x[1]**o+(0.0331367362161)*x[1]+(-0.958604071433)*x[2]**o+(-0.938589620173)*x[2]
        s[0,0,1]=(0.558184625201)*x[0]**o+(-0.442315117978)*x[0]+(0.443008244723)*x[1]**o+(0.10094012777)*x[1]+(-0.884286721018)*x[2]**o+(-0.234859150202)*x[2]
        s[0,1,0]=(0.00707521319666)*x[0]**o+(-0.436445369554)*x[0]+(0.0339511386587)*x[1]**o+(-0.4912959825)*x[1]+(0.387274090979)*x[2]**o+(-0.634777694546)*x[2]
        s[0,1,1]=(-0.169879301168)*x[0]**o+(0.745060242424)*x[0]+(-0.518240485868)*x[1]**o+(0.0555475122839)*x[1]+(0.420456483719)*x[2]**o+(-0.805694025108)*x[2]
        s[1,0,0]=(0.626862061657)*x[0]**o+(0.286735499998)*x[0]+(-0.154542230383)*x[1]**o+(-0.659059291598)*x[1]+(0.700399779356)*x[2]**o+(-0.149494578344)*x[2]
        s[1,0,1]=(-0.85312717829)*x[0]**o+(0.827572890814)*x[0]+(0.929068438324)*x[1]**o+(0.783234890681)*x[1]+(0.132385694002)*x[2]**o+(-0.0375332819318)*x[2]
        s[1,1,0]=(0.828428181973)*x[0]**o+(0.489617430458)*x[0]+(0.0915776189363)*x[1]**o+(-0.425176021653)*x[1]+(0.80387559654)*x[2]**o+(-0.681443382816)*x[2]
        s[1,1,1]=(-0.65230596594)*x[0]**o+(0.170131891444)*x[0]+(0.0830524896765)*x[1]**o+(-0.221764900615)*x[1]+(-0.645608220391)*x[2]**o+(0.400795086408)*x[2]
        s[2,0,0]=(0.893199207029)*x[0]**o+(-0.872998450109)*x[0]+(0.270918838822)*x[1]**o+(0.0111491233197)*x[1]+(-0.0264854465472)*x[2]**o+(0.955927944305)*x[2]
        s[2,0,1]=(0.240242096164)*x[0]**o+(-0.161479794451)*x[0]+(-0.66104753617)*x[1]**o+(0.946080550241)*x[1]+(-0.397485102557)*x[2]**o+(0.886966644933)*x[2]
        s[2,1,0]=(0.80637989046)*x[0]**o+(-0.876797025996)*x[0]+(-0.250156824474)*x[1]**o+(0.956885396073)*x[1]+(-0.721161989016)*x[2]**o+(-0.957921265653)*x[2]
        s[2,1,1]=(-0.769495191281)*x[0]**o+(-0.51589948382)*x[0]+(0.472246475849)*x[1]**o+(0.833747696193)*x[1]+(-0.810945745724)*x[2]**o+(0.00314465415233)*x[2]
        s[3,0,0]=(-0.710791737222)*x[0]**o+(0.608387037146)*x[0]+(0.911509216107)*x[1]**o+(-0.292243195309)*x[1]+(-0.957172951675)*x[2]**o+(0.524831114463)*x[2]
        s[3,0,1]=(0.892735233153)*x[0]**o+(0.28183990495)*x[0]+(0.168860806917)*x[1]**o+(0.388164284517)*x[1]+(0.144580027346)*x[2]**o+(0.881037146671)*x[2]
        s[3,1,0]=(0.221930595909)*x[0]**o+(0.903824792034)*x[0]+(0.271360511407)*x[1]**o+(-0.951642792919)*x[1]+(-0.100229570494)*x[2]**o+(-0.355195001599)*x[2]
        s[3,1,1]=(0.875769927418)*x[0]**o+(-0.161604269687)*x[0]+(0.74568148533)*x[1]**o+(-0.553173791916)*x[1]+(-0.458908157129)*x[2]**o+(-0.247242445331)*x[2]
        s[4,0,0]=(-0.515082637068)*x[0]**o+(0.456575233293)*x[0]+(-0.607867524532)*x[1]**o+(-0.155257786478)*x[1]+(-0.845324963524)*x[2]**o+(-0.862815318477)*x[2]
        s[4,0,1]=(0.922203259385)*x[0]**o+(-0.752313452972)*x[0]+(0.10594289591)*x[1]**o+(-0.370275136504)*x[1]+(-0.0979341688932)*x[2]**o+(-0.066305843593)*x[2]
        s[4,1,0]=(-0.849429313411)*x[0]**o+(-0.28348706029)*x[0]+(0.423697075901)*x[1]**o+(0.724791178434)*x[1]+(0.608888753747)*x[2]**o+(-0.975600085008)*x[2]
        s[4,1,1]=(-0.277835181883)*x[0]**o+(0.255125892715)*x[0]+(-0.801810191818)*x[1]**o+(-0.107375403424)*x[1]+(0.746887097735)*x[2]**o+(0.406084178688)*x[2]
        s[5,0,0]=(-0.137289244094)*x[0]**o+(0.76622915286)*x[0]+(-0.354285839406)*x[1]**o+(-0.195958436789)*x[1]+(-0.836211369427)*x[2]**o+(-0.684714854823)*x[2]
        s[5,0,1]=(0.263960517384)*x[0]**o+(0.908002470581)*x[0]+(0.439729777064)*x[1]**o+(-0.274244779759)*x[1]+(0.659934789159)*x[2]**o+(-0.672573099237)*x[2]
        s[5,1,0]=(-0.20882025835)*x[0]**o+(-0.528304766773)*x[0]+(-0.279646805628)*x[1]**o+(-0.23736563075)*x[1]+(-0.439565255672)*x[2]**o+(0.956815955079)*x[2]
        s[5,1,1]=(-0.294480977855)*x[0]**o+(-0.51059423384)*x[0]+(0.371920382084)*x[1]**o+(0.648458734295)*x[1]+(0.0335200465334)*x[2]**o+(-0.923801240024)*x[2]
        ref[0,0,0]=(-0.208007023803)*(1+2.*(dim-1.)/(o+1.))+(-0.749973854353)*dim
        ref[0,0,1]=(0.116906148907)*(1+2.*(dim-1.)/(o+1.))+(-0.57623414041)*dim
        ref[0,1,0]=(0.428300442835)*(1+2.*(dim-1.)/(o+1.))+(-1.5625190466)*dim
        ref[0,1,1]=(-0.267663303317)*(1+2.*(dim-1.)/(o+1.))+(-0.00508627039966)*dim
        ref[1,0,0]=(1.17271961063)*(1+2.*(dim-1.)/(o+1.))+(-0.521818369944)*dim
        ref[1,0,1]=(0.208326954036)*(1+2.*(dim-1.)/(o+1.))+(1.57327449956)*dim
        ref[1,1,0]=(1.72388139745)*(1+2.*(dim-1.)/(o+1.))+(-0.61700197401)*dim
        ref[1,1,1]=(-1.21486169665)*(1+2.*(dim-1.)/(o+1.))+(0.349162077237)*dim
        ref[2,0,0]=(1.1376325993)*(1+2.*(dim-1.)/(o+1.))+(0.0940786175155)*dim
        ref[2,0,1]=(-0.818290542564)*(1+2.*(dim-1.)/(o+1.))+(1.67156740072)*dim
        ref[2,1,0]=(-0.16493892303)*(1+2.*(dim-1.)/(o+1.))+(-0.877832895575)*dim
        ref[2,1,1]=(-1.10819446116)*(1+2.*(dim-1.)/(o+1.))+(0.320992866525)*dim
        ref[3,0,0]=(-0.75645547279)*(1+2.*(dim-1.)/(o+1.))+(0.8409749563)*dim
        ref[3,0,1]=(1.20617606742)*(1+2.*(dim-1.)/(o+1.))+(1.55104133614)*dim
        ref[3,1,0]=(0.393061536822)*(1+2.*(dim-1.)/(o+1.))+(-0.403013002484)*dim
        ref[3,1,1]=(1.16254325562)*(1+2.*(dim-1.)/(o+1.))+(-0.962020506935)*dim
        ref[4,0,0]=(-1.96827512512)*(1+2.*(dim-1.)/(o+1.))+(-0.561497871662)*dim
        ref[4,0,1]=(0.930211986402)*(1+2.*(dim-1.)/(o+1.))+(-1.18889443307)*dim
        ref[4,1,0]=(0.183156516238)*(1+2.*(dim-1.)/(o+1.))+(-0.534295966865)*dim
        ref[4,1,1]=(-0.332758275966)*(1+2.*(dim-1.)/(o+1.))+(0.553834667978)*dim
        ref[5,0,0]=(-1.32778645293)*(1+2.*(dim-1.)/(o+1.))+(-0.114444138751)*dim
        ref[5,0,1]=(1.36362508361)*(1+2.*(dim-1.)/(o+1.))+(-0.038815408415)*dim
        ref[5,1,0]=(-0.928032319651)*(1+2.*(dim-1.)/(o+1.))+(0.191145557556)*dim
        ref[5,1,1]=(0.110959450763)*(1+2.*(dim-1.)/(o+1.))+(-0.785936739569)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_Solution_rank4(self):
      """
      tests integral of rank 4 Symbol on the FunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-0.512388618715)*x[0]**o+(-0.456786510339)*x[0]+(-0.369955618695)*x[1]**o+(0.284589526283)*x[1]
        s[0,0,0,1]=(0.45694429765)*x[0]**o+(-0.951226590686)*x[0]+(0.812195765321)*x[1]**o+(-0.450606422037)*x[1]
        s[0,0,1,0]=(0.917138960838)*x[0]**o+(0.846015644843)*x[0]+(0.311194587689)*x[1]**o+(0.318765403889)*x[1]
        s[0,0,1,1]=(0.912716498884)*x[0]**o+(-0.309788565863)*x[0]+(0.582625846661)*x[1]**o+(-0.786597172188)*x[1]
        s[0,0,2,0]=(-0.335643205116)*x[0]**o+(-0.162356382231)*x[0]+(0.33024736395)*x[1]**o+(0.137133359333)*x[1]
        s[0,0,2,1]=(-0.853219359404)*x[0]**o+(0.20806173242)*x[0]+(0.904614991529)*x[1]**o+(-0.0927322673776)*x[1]
        s[0,1,0,0]=(-0.116418078552)*x[0]**o+(-0.450766009165)*x[0]+(-0.109027531459)*x[1]**o+(-0.393345607974)*x[1]
        s[0,1,0,1]=(-0.425371185843)*x[0]**o+(-0.513194578298)*x[0]+(0.2204981189)*x[1]**o+(0.959807684568)*x[1]
        s[0,1,1,0]=(-0.184006516133)*x[0]**o+(0.724279611778)*x[0]+(0.233453086887)*x[1]**o+(-0.304433218327)*x[1]
        s[0,1,1,1]=(-0.675234346597)*x[0]**o+(0.444869722037)*x[0]+(-0.0503551086423)*x[1]**o+(-0.0766529306108)*x[1]
        s[0,1,2,0]=(0.841378546422)*x[0]**o+(-0.415767745596)*x[0]+(0.203941853234)*x[1]**o+(0.908654560689)*x[1]
        s[0,1,2,1]=(0.920754900511)*x[0]**o+(-0.630439736764)*x[0]+(-0.58028124877)*x[1]**o+(0.118829809363)*x[1]
        s[0,2,0,0]=(-0.307376557058)*x[0]**o+(0.782274756896)*x[0]+(0.996380983481)*x[1]**o+(-0.91041006941)*x[1]
        s[0,2,0,1]=(0.445454838496)*x[0]**o+(-0.278761052714)*x[0]+(-0.211616677518)*x[1]**o+(-0.982593785248)*x[1]
        s[0,2,1,0]=(-0.304509030828)*x[0]**o+(0.671529029626)*x[0]+(-0.883970223268)*x[1]**o+(-0.5778649808)*x[1]
        s[0,2,1,1]=(-0.463199486208)*x[0]**o+(0.346898357742)*x[0]+(-0.838746712872)*x[1]**o+(-0.303045089364)*x[1]
        s[0,2,2,0]=(-0.561239833326)*x[0]**o+(0.0861796115948)*x[0]+(-0.233810358053)*x[1]**o+(0.667590859355)*x[1]
        s[0,2,2,1]=(-0.0566127020774)*x[0]**o+(-0.702171135934)*x[0]+(0.156408646225)*x[1]**o+(0.503705135348)*x[1]
        s[0,3,0,0]=(-0.213903115172)*x[0]**o+(0.442577130625)*x[0]+(0.973915195866)*x[1]**o+(0.666138357006)*x[1]
        s[0,3,0,1]=(0.888546931537)*x[0]**o+(-0.581620837696)*x[0]+(-0.493141457997)*x[1]**o+(0.572137876409)*x[1]
        s[0,3,1,0]=(-0.198788230198)*x[0]**o+(0.399480321258)*x[0]+(0.437641855366)*x[1]**o+(-0.996597321011)*x[1]
        s[0,3,1,1]=(0.220561040916)*x[0]**o+(-0.613655442822)*x[0]+(0.961868845852)*x[1]**o+(-0.0593807869087)*x[1]
        s[0,3,2,0]=(0.590719334754)*x[0]**o+(0.675522468843)*x[0]+(0.420015363742)*x[1]**o+(-0.779408872787)*x[1]
        s[0,3,2,1]=(-0.739675077785)*x[0]**o+(0.868116107031)*x[0]+(-0.986504469737)*x[1]**o+(0.060324260277)*x[1]
        s[0,4,0,0]=(0.0611797596706)*x[0]**o+(0.804524773235)*x[0]+(0.987009736595)*x[1]**o+(0.811062048039)*x[1]
        s[0,4,0,1]=(0.47425001801)*x[0]**o+(-0.921861981358)*x[0]+(0.344921072375)*x[1]**o+(-0.601620009804)*x[1]
        s[0,4,1,0]=(0.0906565717033)*x[0]**o+(0.152797946513)*x[0]+(-0.567184628241)*x[1]**o+(-0.454348146448)*x[1]
        s[0,4,1,1]=(-0.0896053537692)*x[0]**o+(-0.895659709172)*x[0]+(0.0992922782923)*x[1]**o+(0.491225950996)*x[1]
        s[0,4,2,0]=(-0.810399276984)*x[0]**o+(-0.304522341713)*x[0]+(-0.247445852791)*x[1]**o+(-0.043454035062)*x[1]
        s[0,4,2,1]=(0.798398078646)*x[0]**o+(0.919534042298)*x[0]+(-0.558649299444)*x[1]**o+(-0.583442934514)*x[1]
        s[1,0,0,0]=(0.00573882500165)*x[0]**o+(-0.942216312116)*x[0]+(0.220926627969)*x[1]**o+(-0.363895609692)*x[1]
        s[1,0,0,1]=(-0.849679135789)*x[0]**o+(0.527924450097)*x[0]+(0.79947401514)*x[1]**o+(0.110179654418)*x[1]
        s[1,0,1,0]=(-0.990390680757)*x[0]**o+(0.115770686732)*x[0]+(0.0519459377792)*x[1]**o+(-0.405025360743)*x[1]
        s[1,0,1,1]=(-0.217170752126)*x[0]**o+(0.104026029604)*x[0]+(0.935952773688)*x[1]**o+(0.205909597691)*x[1]
        s[1,0,2,0]=(0.355936537495)*x[0]**o+(0.123599975203)*x[0]+(0.992745060396)*x[1]**o+(0.230955110212)*x[1]
        s[1,0,2,1]=(-0.337760143624)*x[0]**o+(-0.120785608957)*x[0]+(0.519621755185)*x[1]**o+(0.648090620962)*x[1]
        s[1,1,0,0]=(0.425227533029)*x[0]**o+(0.431947155274)*x[0]+(0.864871771139)*x[1]**o+(-0.421776717503)*x[1]
        s[1,1,0,1]=(-0.155048210381)*x[0]**o+(0.98349398745)*x[0]+(0.616262478346)*x[1]**o+(0.354130744218)*x[1]
        s[1,1,1,0]=(-0.255422611215)*x[0]**o+(0.610000128329)*x[0]+(0.692199306628)*x[1]**o+(-0.469569470881)*x[1]
        s[1,1,1,1]=(-0.00176886794175)*x[0]**o+(-0.408069848741)*x[0]+(0.708816789956)*x[1]**o+(0.823083702629)*x[1]
        s[1,1,2,0]=(0.229145338062)*x[0]**o+(-0.468402804906)*x[0]+(-0.352476560454)*x[1]**o+(0.968823931554)*x[1]
        s[1,1,2,1]=(-0.414346479022)*x[0]**o+(-0.0181085935776)*x[0]+(-0.170858935466)*x[1]**o+(-0.816432101623)*x[1]
        s[1,2,0,0]=(-0.0948608343933)*x[0]**o+(0.0068199755948)*x[0]+(-0.145376398386)*x[1]**o+(0.755438498846)*x[1]
        s[1,2,0,1]=(-0.665050941689)*x[0]**o+(-0.719231494887)*x[0]+(0.41492847392)*x[1]**o+(0.708962325583)*x[1]
        s[1,2,1,0]=(0.621628314609)*x[0]**o+(-0.997324959187)*x[0]+(0.211421770712)*x[1]**o+(0.272336354113)*x[1]
        s[1,2,1,1]=(-0.744309774218)*x[0]**o+(-0.901777776216)*x[0]+(-0.727378511259)*x[1]**o+(0.276331604178)*x[1]
        s[1,2,2,0]=(0.119470505814)*x[0]**o+(-0.377429184576)*x[0]+(0.634631061717)*x[1]**o+(0.71726575358)*x[1]
        s[1,2,2,1]=(0.196579032787)*x[0]**o+(-0.417495130697)*x[0]+(-0.469739874787)*x[1]**o+(-0.775290208858)*x[1]
        s[1,3,0,0]=(0.406483838142)*x[0]**o+(0.184354975226)*x[0]+(0.762003327026)*x[1]**o+(0.0300461028065)*x[1]
        s[1,3,0,1]=(-0.844790086513)*x[0]**o+(0.663686153475)*x[0]+(0.567312988723)*x[1]**o+(0.508368547154)*x[1]
        s[1,3,1,0]=(-0.6417260084)*x[0]**o+(0.592764591647)*x[0]+(-0.335643553939)*x[1]**o+(0.152782626)*x[1]
        s[1,3,1,1]=(-0.570864881875)*x[0]**o+(-0.610576918843)*x[0]+(-0.0434074971721)*x[1]**o+(0.373353748894)*x[1]
        s[1,3,2,0]=(0.951712868404)*x[0]**o+(0.837001943514)*x[0]+(-0.661919006256)*x[1]**o+(0.136870953867)*x[1]
        s[1,3,2,1]=(0.338496531342)*x[0]**o+(0.646315658139)*x[0]+(0.173960800908)*x[1]**o+(0.730170280099)*x[1]
        s[1,4,0,0]=(-0.00733305213967)*x[0]**o+(0.473349687079)*x[0]+(0.433262136464)*x[1]**o+(-0.122843871456)*x[1]
        s[1,4,0,1]=(-0.609798017036)*x[0]**o+(-0.271729269891)*x[0]+(-0.0473090363303)*x[1]**o+(-0.429215577266)*x[1]
        s[1,4,1,0]=(0.860194281311)*x[0]**o+(0.267299506296)*x[0]+(0.564929961344)*x[1]**o+(0.4738297918)*x[1]
        s[1,4,1,1]=(0.170949641101)*x[0]**o+(0.710820092329)*x[0]+(0.934882329613)*x[1]**o+(0.607064104034)*x[1]
        s[1,4,2,0]=(-0.299596799604)*x[0]**o+(-0.533859449337)*x[0]+(0.751725833171)*x[1]**o+(-0.690767322166)*x[1]
        s[1,4,2,1]=(-0.624424753627)*x[0]**o+(-0.947238508194)*x[0]+(0.456597368873)*x[1]**o+(-0.800064510218)*x[1]
        s[2,0,0,0]=(0.566979423938)*x[0]**o+(0.924856128006)*x[0]+(0.894661507372)*x[1]**o+(0.737960904672)*x[1]
        s[2,0,0,1]=(0.424224691717)*x[0]**o+(-0.290057017691)*x[0]+(0.524889948009)*x[1]**o+(-0.968728724115)*x[1]
        s[2,0,1,0]=(0.116336639818)*x[0]**o+(0.192371600611)*x[0]+(-0.303824973684)*x[1]**o+(-0.32000509975)*x[1]
        s[2,0,1,1]=(-0.327127663029)*x[0]**o+(0.715544836917)*x[0]+(0.347278471442)*x[1]**o+(0.109909231847)*x[1]
        s[2,0,2,0]=(0.793922217119)*x[0]**o+(0.262726473054)*x[0]+(0.172201555322)*x[1]**o+(0.770334659224)*x[1]
        s[2,0,2,1]=(0.988217831671)*x[0]**o+(0.399655753685)*x[0]+(-0.143569502754)*x[1]**o+(0.0617091702351)*x[1]
        s[2,1,0,0]=(0.775224667659)*x[0]**o+(-0.0668878550232)*x[0]+(-0.424553437498)*x[1]**o+(-0.472878522461)*x[1]
        s[2,1,0,1]=(0.214811724455)*x[0]**o+(0.331016899879)*x[0]+(0.137063379737)*x[1]**o+(-0.0540696322974)*x[1]
        s[2,1,1,0]=(-0.463688617296)*x[0]**o+(-0.373502340044)*x[0]+(-0.0840959194547)*x[1]**o+(0.399767480048)*x[1]
        s[2,1,1,1]=(-0.245792577875)*x[0]**o+(-0.996843380732)*x[0]+(-0.249517835294)*x[1]**o+(-0.594721563199)*x[1]
        s[2,1,2,0]=(0.671238831228)*x[0]**o+(0.899659387933)*x[0]+(-0.964232607177)*x[1]**o+(0.637445200931)*x[1]
        s[2,1,2,1]=(0.746879641209)*x[0]**o+(-0.247785412499)*x[0]+(-0.370865590053)*x[1]**o+(-0.302746229959)*x[1]
        s[2,2,0,0]=(0.336257487071)*x[0]**o+(-0.179673308736)*x[0]+(0.912054571446)*x[1]**o+(-0.716018450571)*x[1]
        s[2,2,0,1]=(-0.686601572148)*x[0]**o+(0.120899496999)*x[0]+(-0.419150767247)*x[1]**o+(-0.418707334342)*x[1]
        s[2,2,1,0]=(-0.354036063032)*x[0]**o+(0.363634583771)*x[0]+(0.691849369366)*x[1]**o+(-0.887830896718)*x[1]
        s[2,2,1,1]=(0.711473627816)*x[0]**o+(0.556386580914)*x[0]+(0.888396281963)*x[1]**o+(0.386590176165)*x[1]
        s[2,2,2,0]=(0.419814106708)*x[0]**o+(-0.0352463077763)*x[0]+(0.600680097958)*x[1]**o+(-0.884870139538)*x[1]
        s[2,2,2,1]=(0.203572624179)*x[0]**o+(-0.915173748563)*x[0]+(0.0466882954353)*x[1]**o+(0.567223966741)*x[1]
        s[2,3,0,0]=(0.365889449332)*x[0]**o+(-0.0875793555978)*x[0]+(-0.514531990304)*x[1]**o+(0.641244565067)*x[1]
        s[2,3,0,1]=(0.801255486954)*x[0]**o+(0.324497137754)*x[0]+(-0.750291931827)*x[1]**o+(-0.972736614855)*x[1]
        s[2,3,1,0]=(0.540943455318)*x[0]**o+(-0.507084574203)*x[0]+(0.903787268326)*x[1]**o+(0.192067385499)*x[1]
        s[2,3,1,1]=(0.50229010977)*x[0]**o+(0.327771429044)*x[0]+(-0.871646777584)*x[1]**o+(-0.915312646394)*x[1]
        s[2,3,2,0]=(0.816437071516)*x[0]**o+(-0.324552837786)*x[0]+(0.32191911659)*x[1]**o+(0.448902757154)*x[1]
        s[2,3,2,1]=(-0.688163580944)*x[0]**o+(0.00185099386065)*x[0]+(0.629596314583)*x[1]**o+(0.898968857239)*x[1]
        s[2,4,0,0]=(0.558694638173)*x[0]**o+(0.879817228992)*x[0]+(-0.779489055632)*x[1]**o+(-0.597316482587)*x[1]
        s[2,4,0,1]=(0.405266767625)*x[0]**o+(-0.803650202792)*x[0]+(-0.209683708637)*x[1]**o+(0.843260589899)*x[1]
        s[2,4,1,0]=(0.228094700844)*x[0]**o+(0.28915480893)*x[0]+(0.404160989937)*x[1]**o+(-0.837717657152)*x[1]
        s[2,4,1,1]=(-0.628771431514)*x[0]**o+(0.70484286338)*x[0]+(0.162686726925)*x[1]**o+(0.743576454116)*x[1]
        s[2,4,2,0]=(0.139366390708)*x[0]**o+(0.452232237593)*x[0]+(-0.0536300454799)*x[1]**o+(-0.0220704990737)*x[1]
        s[2,4,2,1]=(0.407548912264)*x[0]**o+(-0.184892320636)*x[0]+(0.945920847222)*x[1]**o+(-0.326239004837)*x[1]
        s[3,0,0,0]=(-0.893632035291)*x[0]**o+(0.605568638642)*x[0]+(-0.605327037586)*x[1]**o+(-0.129824948634)*x[1]
        s[3,0,0,1]=(0.058430352705)*x[0]**o+(0.724766033805)*x[0]+(0.900105388199)*x[1]**o+(-0.248904105045)*x[1]
        s[3,0,1,0]=(-0.151387812319)*x[0]**o+(0.434769353572)*x[0]+(0.525149052261)*x[1]**o+(-0.949942908295)*x[1]
        s[3,0,1,1]=(0.419206109353)*x[0]**o+(0.828163036942)*x[0]+(-0.388930812616)*x[1]**o+(0.247705907297)*x[1]
        s[3,0,2,0]=(0.0520861249252)*x[0]**o+(0.545440655049)*x[0]+(0.123957894743)*x[1]**o+(0.726676820938)*x[1]
        s[3,0,2,1]=(-0.832443800699)*x[0]**o+(-0.873219166703)*x[0]+(-0.09459670039)*x[1]**o+(-0.490598426074)*x[1]
        s[3,1,0,0]=(-0.0396898083178)*x[0]**o+(0.13897357947)*x[0]+(-0.570943434112)*x[1]**o+(0.113736405636)*x[1]
        s[3,1,0,1]=(-0.753753572629)*x[0]**o+(0.0105397073947)*x[0]+(-0.100456113737)*x[1]**o+(0.702912637963)*x[1]
        s[3,1,1,0]=(-0.2396162336)*x[0]**o+(0.0957817653067)*x[0]+(-0.426329991998)*x[1]**o+(-0.806921457576)*x[1]
        s[3,1,1,1]=(-0.00676639677643)*x[0]**o+(0.644421172464)*x[0]+(0.893174135862)*x[1]**o+(0.921265996277)*x[1]
        s[3,1,2,0]=(-0.370431629052)*x[0]**o+(-0.431968328195)*x[0]+(0.322907749791)*x[1]**o+(-0.734928059791)*x[1]
        s[3,1,2,1]=(0.565520417451)*x[0]**o+(0.5972374649)*x[0]+(-0.384953451928)*x[1]**o+(0.708513444822)*x[1]
        s[3,2,0,0]=(0.546973382463)*x[0]**o+(-0.391535907448)*x[0]+(0.743984877534)*x[1]**o+(0.690664640674)*x[1]
        s[3,2,0,1]=(0.286366773252)*x[0]**o+(-0.536931967236)*x[0]+(0.914555659603)*x[1]**o+(-0.978176990473)*x[1]
        s[3,2,1,0]=(-0.272652614089)*x[0]**o+(0.0235157061398)*x[0]+(-0.147489880453)*x[1]**o+(-0.698449540889)*x[1]
        s[3,2,1,1]=(0.856000001512)*x[0]**o+(-0.49804787796)*x[0]+(-0.993780459188)*x[1]**o+(0.851338346493)*x[1]
        s[3,2,2,0]=(-0.57527355268)*x[0]**o+(-0.760820441576)*x[0]+(-0.654096561486)*x[1]**o+(-0.736649206713)*x[1]
        s[3,2,2,1]=(-0.262714050973)*x[0]**o+(-0.182627531212)*x[0]+(0.845508653482)*x[1]**o+(0.335330629033)*x[1]
        s[3,3,0,0]=(0.284468477631)*x[0]**o+(0.778849843484)*x[0]+(-0.210037009712)*x[1]**o+(0.261684901446)*x[1]
        s[3,3,0,1]=(-0.702010952741)*x[0]**o+(-0.147730819043)*x[0]+(-0.251392638434)*x[1]**o+(-0.450438064218)*x[1]
        s[3,3,1,0]=(0.37599915758)*x[0]**o+(-0.663740528508)*x[0]+(0.831132392336)*x[1]**o+(-0.686462967317)*x[1]
        s[3,3,1,1]=(0.104633103538)*x[0]**o+(0.152999779786)*x[0]+(-0.617511701255)*x[1]**o+(-0.6261759882)*x[1]
        s[3,3,2,0]=(0.996985674737)*x[0]**o+(-0.113914220536)*x[0]+(0.815302144446)*x[1]**o+(0.426458911854)*x[1]
        s[3,3,2,1]=(0.79490425189)*x[0]**o+(-0.352566895288)*x[0]+(0.595219934007)*x[1]**o+(0.931962305688)*x[1]
        s[3,4,0,0]=(0.0342109984348)*x[0]**o+(0.756935167027)*x[0]+(-0.910031669765)*x[1]**o+(-0.768302193908)*x[1]
        s[3,4,0,1]=(-0.765106211079)*x[0]**o+(0.971115633922)*x[0]+(-0.682150454882)*x[1]**o+(0.345375833573)*x[1]
        s[3,4,1,0]=(0.392841668036)*x[0]**o+(0.06254598514)*x[0]+(0.66808036785)*x[1]**o+(-0.76724599164)*x[1]
        s[3,4,1,1]=(-0.155018584015)*x[0]**o+(0.100771866989)*x[0]+(-0.800809131287)*x[1]**o+(-0.481511683108)*x[1]
        s[3,4,2,0]=(0.420069461837)*x[0]**o+(0.116833735731)*x[0]+(-0.334781127089)*x[1]**o+(-0.278898259371)*x[1]
        s[3,4,2,1]=(-0.695614335152)*x[0]**o+(-0.116909319428)*x[0]+(-0.136699698838)*x[1]**o+(-0.199293217231)*x[1]
        ref[0,0,0,0]=(-0.882344237411)*(1+2.*(dim-1.)/(o+1.))+(-0.172196984056)*dim
        ref[0,0,0,1]=(1.26914006297)*(1+2.*(dim-1.)/(o+1.))+(-1.40183301272)*dim
        ref[0,0,1,0]=(1.22833354853)*(1+2.*(dim-1.)/(o+1.))+(1.16478104873)*dim
        ref[0,0,1,1]=(1.49534234555)*(1+2.*(dim-1.)/(o+1.))+(-1.09638573805)*dim
        ref[0,0,2,0]=(-0.00539584116507)*(1+2.*(dim-1.)/(o+1.))+(-0.0252230228979)*dim
        ref[0,0,2,1]=(0.0513956321246)*(1+2.*(dim-1.)/(o+1.))+(0.115329465043)*dim
        ref[0,1,0,0]=(-0.225445610011)*(1+2.*(dim-1.)/(o+1.))+(-0.84411161714)*dim
        ref[0,1,0,1]=(-0.204873066942)*(1+2.*(dim-1.)/(o+1.))+(0.44661310627)*dim
        ref[0,1,1,0]=(0.0494465707545)*(1+2.*(dim-1.)/(o+1.))+(0.419846393451)*dim
        ref[0,1,1,1]=(-0.725589455239)*(1+2.*(dim-1.)/(o+1.))+(0.368216791426)*dim
        ref[0,1,2,0]=(1.04532039966)*(1+2.*(dim-1.)/(o+1.))+(0.492886815094)*dim
        ref[0,1,2,1]=(0.340473651741)*(1+2.*(dim-1.)/(o+1.))+(-0.511609927401)*dim
        ref[0,2,0,0]=(0.689004426423)*(1+2.*(dim-1.)/(o+1.))+(-0.128135312514)*dim
        ref[0,2,0,1]=(0.233838160978)*(1+2.*(dim-1.)/(o+1.))+(-1.26135483796)*dim
        ref[0,2,1,0]=(-1.1884792541)*(1+2.*(dim-1.)/(o+1.))+(0.0936640488261)*dim
        ref[0,2,1,1]=(-1.30194619908)*(1+2.*(dim-1.)/(o+1.))+(0.0438532683776)*dim
        ref[0,2,2,0]=(-0.795050191379)*(1+2.*(dim-1.)/(o+1.))+(0.75377047095)*dim
        ref[0,2,2,1]=(0.099795944148)*(1+2.*(dim-1.)/(o+1.))+(-0.198466000586)*dim
        ref[0,3,0,0]=(0.760012080694)*(1+2.*(dim-1.)/(o+1.))+(1.10871548763)*dim
        ref[0,3,0,1]=(0.395405473541)*(1+2.*(dim-1.)/(o+1.))+(-0.00948296128713)*dim
        ref[0,3,1,0]=(0.238853625168)*(1+2.*(dim-1.)/(o+1.))+(-0.597116999753)*dim
        ref[0,3,1,1]=(1.18242988677)*(1+2.*(dim-1.)/(o+1.))+(-0.673036229731)*dim
        ref[0,3,2,0]=(1.0107346985)*(1+2.*(dim-1.)/(o+1.))+(-0.103886403944)*dim
        ref[0,3,2,1]=(-1.72617954752)*(1+2.*(dim-1.)/(o+1.))+(0.928440367308)*dim
        ref[0,4,0,0]=(1.04818949627)*(1+2.*(dim-1.)/(o+1.))+(1.61558682127)*dim
        ref[0,4,0,1]=(0.819171090384)*(1+2.*(dim-1.)/(o+1.))+(-1.52348199116)*dim
        ref[0,4,1,0]=(-0.476528056537)*(1+2.*(dim-1.)/(o+1.))+(-0.301550199935)*dim
        ref[0,4,1,1]=(0.00968692452303)*(1+2.*(dim-1.)/(o+1.))+(-0.404433758176)*dim
        ref[0,4,2,0]=(-1.05784512977)*(1+2.*(dim-1.)/(o+1.))+(-0.347976376775)*dim
        ref[0,4,2,1]=(0.239748779202)*(1+2.*(dim-1.)/(o+1.))+(0.336091107784)*dim
        ref[1,0,0,0]=(0.226665452971)*(1+2.*(dim-1.)/(o+1.))+(-1.30611192181)*dim
        ref[1,0,0,1]=(-0.050205120649)*(1+2.*(dim-1.)/(o+1.))+(0.638104104515)*dim
        ref[1,0,1,0]=(-0.938444742978)*(1+2.*(dim-1.)/(o+1.))+(-0.289254674011)*dim
        ref[1,0,1,1]=(0.718782021562)*(1+2.*(dim-1.)/(o+1.))+(0.309935627295)*dim
        ref[1,0,2,0]=(1.34868159789)*(1+2.*(dim-1.)/(o+1.))+(0.354555085415)*dim
        ref[1,0,2,1]=(0.181861611561)*(1+2.*(dim-1.)/(o+1.))+(0.527305012005)*dim
        ref[1,1,0,0]=(1.29009930417)*(1+2.*(dim-1.)/(o+1.))+(0.0101704377708)*dim
        ref[1,1,0,1]=(0.461214267966)*(1+2.*(dim-1.)/(o+1.))+(1.33762473167)*dim
        ref[1,1,1,0]=(0.436776695414)*(1+2.*(dim-1.)/(o+1.))+(0.140430657448)*dim
        ref[1,1,1,1]=(0.707047922014)*(1+2.*(dim-1.)/(o+1.))+(0.415013853887)*dim
        ref[1,1,2,0]=(-0.123331222393)*(1+2.*(dim-1.)/(o+1.))+(0.500421126648)*dim
        ref[1,1,2,1]=(-0.585205414488)*(1+2.*(dim-1.)/(o+1.))+(-0.8345406952)*dim
        ref[1,2,0,0]=(-0.240237232779)*(1+2.*(dim-1.)/(o+1.))+(0.762258474441)*dim
        ref[1,2,0,1]=(-0.250122467769)*(1+2.*(dim-1.)/(o+1.))+(-0.0102691693039)*dim
        ref[1,2,1,0]=(0.833050085321)*(1+2.*(dim-1.)/(o+1.))+(-0.724988605075)*dim
        ref[1,2,1,1]=(-1.47168828548)*(1+2.*(dim-1.)/(o+1.))+(-0.625446172039)*dim
        ref[1,2,2,0]=(0.75410156753)*(1+2.*(dim-1.)/(o+1.))+(0.339836569004)*dim
        ref[1,2,2,1]=(-0.273160842)*(1+2.*(dim-1.)/(o+1.))+(-1.19278533956)*dim
        ref[1,3,0,0]=(1.16848716517)*(1+2.*(dim-1.)/(o+1.))+(0.214401078033)*dim
        ref[1,3,0,1]=(-0.277477097791)*(1+2.*(dim-1.)/(o+1.))+(1.17205470063)*dim
        ref[1,3,1,0]=(-0.977369562339)*(1+2.*(dim-1.)/(o+1.))+(0.745547217647)*dim
        ref[1,3,1,1]=(-0.614272379047)*(1+2.*(dim-1.)/(o+1.))+(-0.237223169949)*dim
        ref[1,3,2,0]=(0.289793862148)*(1+2.*(dim-1.)/(o+1.))+(0.973872897381)*dim
        ref[1,3,2,1]=(0.512457332251)*(1+2.*(dim-1.)/(o+1.))+(1.37648593824)*dim
        ref[1,4,0,0]=(0.425929084324)*(1+2.*(dim-1.)/(o+1.))+(0.350505815623)*dim
        ref[1,4,0,1]=(-0.657107053366)*(1+2.*(dim-1.)/(o+1.))+(-0.700944847157)*dim
        ref[1,4,1,0]=(1.42512424265)*(1+2.*(dim-1.)/(o+1.))+(0.741129298097)*dim
        ref[1,4,1,1]=(1.10583197071)*(1+2.*(dim-1.)/(o+1.))+(1.31788419636)*dim
        ref[1,4,2,0]=(0.452129033567)*(1+2.*(dim-1.)/(o+1.))+(-1.2246267715)*dim
        ref[1,4,2,1]=(-0.167827384754)*(1+2.*(dim-1.)/(o+1.))+(-1.74730301841)*dim
        ref[2,0,0,0]=(1.46164093131)*(1+2.*(dim-1.)/(o+1.))+(1.66281703268)*dim
        ref[2,0,0,1]=(0.949114639726)*(1+2.*(dim-1.)/(o+1.))+(-1.25878574181)*dim
        ref[2,0,1,0]=(-0.187488333866)*(1+2.*(dim-1.)/(o+1.))+(-0.127633499139)*dim
        ref[2,0,1,1]=(0.020150808413)*(1+2.*(dim-1.)/(o+1.))+(0.825454068764)*dim
        ref[2,0,2,0]=(0.96612377244)*(1+2.*(dim-1.)/(o+1.))+(1.03306113228)*dim
        ref[2,0,2,1]=(0.844648328917)*(1+2.*(dim-1.)/(o+1.))+(0.46136492392)*dim
        ref[2,1,0,0]=(0.350671230162)*(1+2.*(dim-1.)/(o+1.))+(-0.539766377484)*dim
        ref[2,1,0,1]=(0.351875104192)*(1+2.*(dim-1.)/(o+1.))+(0.276947267582)*dim
        ref[2,1,1,0]=(-0.54778453675)*(1+2.*(dim-1.)/(o+1.))+(0.0262651400042)*dim
        ref[2,1,1,1]=(-0.495310413169)*(1+2.*(dim-1.)/(o+1.))+(-1.59156494393)*dim
        ref[2,1,2,0]=(-0.292993775949)*(1+2.*(dim-1.)/(o+1.))+(1.53710458886)*dim
        ref[2,1,2,1]=(0.376014051155)*(1+2.*(dim-1.)/(o+1.))+(-0.550531642457)*dim
        ref[2,2,0,0]=(1.24831205852)*(1+2.*(dim-1.)/(o+1.))+(-0.895691759307)*dim
        ref[2,2,0,1]=(-1.1057523394)*(1+2.*(dim-1.)/(o+1.))+(-0.297807837342)*dim
        ref[2,2,1,0]=(0.337813306334)*(1+2.*(dim-1.)/(o+1.))+(-0.524196312948)*dim
        ref[2,2,1,1]=(1.59986990978)*(1+2.*(dim-1.)/(o+1.))+(0.942976757078)*dim
        ref[2,2,2,0]=(1.02049420467)*(1+2.*(dim-1.)/(o+1.))+(-0.920116447314)*dim
        ref[2,2,2,1]=(0.250260919614)*(1+2.*(dim-1.)/(o+1.))+(-0.347949781822)*dim
        ref[2,3,0,0]=(-0.148642540972)*(1+2.*(dim-1.)/(o+1.))+(0.553665209469)*dim
        ref[2,3,0,1]=(0.0509635551269)*(1+2.*(dim-1.)/(o+1.))+(-0.6482394771)*dim
        ref[2,3,1,0]=(1.44473072364)*(1+2.*(dim-1.)/(o+1.))+(-0.315017188703)*dim
        ref[2,3,1,1]=(-0.369356667814)*(1+2.*(dim-1.)/(o+1.))+(-0.58754121735)*dim
        ref[2,3,2,0]=(1.13835618811)*(1+2.*(dim-1.)/(o+1.))+(0.124349919368)*dim
        ref[2,3,2,1]=(-0.058567266361)*(1+2.*(dim-1.)/(o+1.))+(0.900819851099)*dim
        ref[2,4,0,0]=(-0.220794417459)*(1+2.*(dim-1.)/(o+1.))+(0.282500746405)*dim
        ref[2,4,0,1]=(0.195583058988)*(1+2.*(dim-1.)/(o+1.))+(0.0396103871071)*dim
        ref[2,4,1,0]=(0.63225569078)*(1+2.*(dim-1.)/(o+1.))+(-0.548562848222)*dim
        ref[2,4,1,1]=(-0.466084704588)*(1+2.*(dim-1.)/(o+1.))+(1.4484193175)*dim
        ref[2,4,2,0]=(0.0857363452282)*(1+2.*(dim-1.)/(o+1.))+(0.430161738519)*dim
        ref[2,4,2,1]=(1.35346975949)*(1+2.*(dim-1.)/(o+1.))+(-0.511131325473)*dim
        ref[3,0,0,0]=(-1.49895907288)*(1+2.*(dim-1.)/(o+1.))+(0.475743690008)*dim
        ref[3,0,0,1]=(0.958535740904)*(1+2.*(dim-1.)/(o+1.))+(0.47586192876)*dim
        ref[3,0,1,0]=(0.373761239942)*(1+2.*(dim-1.)/(o+1.))+(-0.515173554723)*dim
        ref[3,0,1,1]=(0.0302752967372)*(1+2.*(dim-1.)/(o+1.))+(1.07586894424)*dim
        ref[3,0,2,0]=(0.176044019669)*(1+2.*(dim-1.)/(o+1.))+(1.27211747599)*dim
        ref[3,0,2,1]=(-0.927040501089)*(1+2.*(dim-1.)/(o+1.))+(-1.36381759278)*dim
        ref[3,1,0,0]=(-0.61063324243)*(1+2.*(dim-1.)/(o+1.))+(0.252709985106)*dim
        ref[3,1,0,1]=(-0.854209686365)*(1+2.*(dim-1.)/(o+1.))+(0.713452345358)*dim
        ref[3,1,1,0]=(-0.665946225598)*(1+2.*(dim-1.)/(o+1.))+(-0.711139692269)*dim
        ref[3,1,1,1]=(0.886407739085)*(1+2.*(dim-1.)/(o+1.))+(1.56568716874)*dim
        ref[3,1,2,0]=(-0.0475238792609)*(1+2.*(dim-1.)/(o+1.))+(-1.16689638799)*dim
        ref[3,1,2,1]=(0.180566965523)*(1+2.*(dim-1.)/(o+1.))+(1.30575090972)*dim
        ref[3,2,0,0]=(1.29095826)*(1+2.*(dim-1.)/(o+1.))+(0.299128733226)*dim
        ref[3,2,0,1]=(1.20092243286)*(1+2.*(dim-1.)/(o+1.))+(-1.51510895771)*dim
        ref[3,2,1,0]=(-0.420142494542)*(1+2.*(dim-1.)/(o+1.))+(-0.674933834749)*dim
        ref[3,2,1,1]=(-0.137780457676)*(1+2.*(dim-1.)/(o+1.))+(0.353290468533)*dim
        ref[3,2,2,0]=(-1.22937011417)*(1+2.*(dim-1.)/(o+1.))+(-1.49746964829)*dim
        ref[3,2,2,1]=(0.582794602509)*(1+2.*(dim-1.)/(o+1.))+(0.152703097821)*dim
        ref[3,3,0,0]=(0.0744314679194)*(1+2.*(dim-1.)/(o+1.))+(1.04053474493)*dim
        ref[3,3,0,1]=(-0.953403591176)*(1+2.*(dim-1.)/(o+1.))+(-0.598168883261)*dim
        ref[3,3,1,0]=(1.20713154992)*(1+2.*(dim-1.)/(o+1.))+(-1.35020349583)*dim
        ref[3,3,1,1]=(-0.512878597717)*(1+2.*(dim-1.)/(o+1.))+(-0.473176208414)*dim
        ref[3,3,2,0]=(1.81228781918)*(1+2.*(dim-1.)/(o+1.))+(0.312544691319)*dim
        ref[3,3,2,1]=(1.3901241859)*(1+2.*(dim-1.)/(o+1.))+(0.5793954104)*dim
        ref[3,4,0,0]=(-0.87582067133)*(1+2.*(dim-1.)/(o+1.))+(-0.0113670268806)*dim
        ref[3,4,0,1]=(-1.44725666596)*(1+2.*(dim-1.)/(o+1.))+(1.31649146749)*dim
        ref[3,4,1,0]=(1.06092203589)*(1+2.*(dim-1.)/(o+1.))+(-0.7047000065)*dim
        ref[3,4,1,1]=(-0.955827715302)*(1+2.*(dim-1.)/(o+1.))+(-0.38073981612)*dim
        ref[3,4,2,0]=(0.0852883347481)*(1+2.*(dim-1.)/(o+1.))+(-0.16206452364)*dim
        ref[3,4,2,1]=(-0.83231403399)*(1+2.*(dim-1.)/(o+1.))+(-0.316202536659)*dim
      else:
        s[0,0,0,0]=(-0.394972833746)*x[0]**o+(-0.555133440221)*x[0]+(-0.268090782116)*x[1]**o+(-0.279086182399)*x[1]+(0.647538832619)*x[2]**o+(0.403928571033)*x[2]
        s[0,0,0,1]=(-0.35392621979)*x[0]**o+(-0.492907886034)*x[0]+(-0.285991202298)*x[1]**o+(0.652256699399)*x[1]+(0.746644949246)*x[2]**o+(-0.81069060154)*x[2]
        s[0,0,1,0]=(-0.373534422869)*x[0]**o+(0.725526776416)*x[0]+(-0.228168552487)*x[1]**o+(0.530464480729)*x[1]+(-0.257805238342)*x[2]**o+(0.482778418004)*x[2]
        s[0,0,1,1]=(-0.19411212431)*x[0]**o+(0.45313120948)*x[0]+(0.159756421366)*x[1]**o+(0.124033159775)*x[1]+(0.395678187632)*x[2]**o+(-0.145495322245)*x[2]
        s[0,0,2,0]=(0.420598692135)*x[0]**o+(0.222648941851)*x[0]+(-0.964380624775)*x[1]**o+(0.419039850926)*x[1]+(-0.908861665427)*x[2]**o+(-0.13977247814)*x[2]
        s[0,0,2,1]=(-0.339655604031)*x[0]**o+(-0.81821640554)*x[0]+(-0.864798652203)*x[1]**o+(-0.695538165934)*x[1]+(0.535554008883)*x[2]**o+(-0.142485273891)*x[2]
        s[0,1,0,0]=(-0.221315603083)*x[0]**o+(0.437569708371)*x[0]+(0.702877377655)*x[1]**o+(0.169385737092)*x[1]+(-0.92490534465)*x[2]**o+(0.264213443224)*x[2]
        s[0,1,0,1]=(0.61945266928)*x[0]**o+(0.699276978747)*x[0]+(0.199776861853)*x[1]**o+(-0.615345507396)*x[1]+(-0.70081403292)*x[2]**o+(0.664056523269)*x[2]
        s[0,1,1,0]=(-0.479722011354)*x[0]**o+(-0.591782822119)*x[0]+(0.919441326242)*x[1]**o+(-0.886693034352)*x[1]+(0.48123492158)*x[2]**o+(0.308135102102)*x[2]
        s[0,1,1,1]=(-0.522342575747)*x[0]**o+(0.662917347867)*x[0]+(0.359887593621)*x[1]**o+(-0.228498407508)*x[1]+(0.705995139546)*x[2]**o+(-0.406581496679)*x[2]
        s[0,1,2,0]=(0.555378475566)*x[0]**o+(-0.680672745096)*x[0]+(-0.269859012441)*x[1]**o+(0.800892438723)*x[1]+(-0.455863363072)*x[2]**o+(-0.991391540158)*x[2]
        s[0,1,2,1]=(0.547403511029)*x[0]**o+(-0.154514192835)*x[0]+(0.177483676876)*x[1]**o+(0.887044119995)*x[1]+(-0.21971493108)*x[2]**o+(-0.262092872636)*x[2]
        s[0,2,0,0]=(0.976262232729)*x[0]**o+(0.548252161786)*x[0]+(-0.743334308912)*x[1]**o+(0.207627841895)*x[1]+(0.853225044125)*x[2]**o+(0.307558901346)*x[2]
        s[0,2,0,1]=(0.503102495469)*x[0]**o+(0.407241098221)*x[0]+(0.52413664456)*x[1]**o+(-0.00773056229007)*x[1]+(-0.292242153733)*x[2]**o+(0.17146852389)*x[2]
        s[0,2,1,0]=(0.13389613662)*x[0]**o+(-0.87587702201)*x[0]+(-0.678875907381)*x[1]**o+(0.543117132397)*x[1]+(0.393096793861)*x[2]**o+(0.959258318284)*x[2]
        s[0,2,1,1]=(-0.0940648527232)*x[0]**o+(-0.598198827088)*x[0]+(0.424517028995)*x[1]**o+(-0.345900941234)*x[1]+(0.758648965957)*x[2]**o+(0.25975303965)*x[2]
        s[0,2,2,0]=(-0.870274070711)*x[0]**o+(-0.78246615425)*x[0]+(0.392295964553)*x[1]**o+(0.844507475338)*x[1]+(0.528851559277)*x[2]**o+(-0.995729427392)*x[2]
        s[0,2,2,1]=(-0.0730197692456)*x[0]**o+(-0.312551702743)*x[0]+(0.623722578367)*x[1]**o+(-0.848292098624)*x[1]+(-0.779643252401)*x[2]**o+(-0.338597223715)*x[2]
        s[0,3,0,0]=(0.872506198169)*x[0]**o+(-0.500297419082)*x[0]+(0.269776051595)*x[1]**o+(0.566518089969)*x[1]+(-0.498369039824)*x[2]**o+(0.607886737439)*x[2]
        s[0,3,0,1]=(0.883053802312)*x[0]**o+(0.833130186655)*x[0]+(-0.921394587681)*x[1]**o+(0.0175686356791)*x[1]+(0.951456010531)*x[2]**o+(-0.571752807102)*x[2]
        s[0,3,1,0]=(-0.712330362129)*x[0]**o+(-0.676885540903)*x[0]+(-0.509633815107)*x[1]**o+(0.999769611236)*x[1]+(-0.186240684121)*x[2]**o+(-0.386806354224)*x[2]
        s[0,3,1,1]=(0.997518452801)*x[0]**o+(-0.269198998158)*x[0]+(-0.775427684973)*x[1]**o+(0.787580393259)*x[1]+(0.195858442466)*x[2]**o+(-0.734837176349)*x[2]
        s[0,3,2,0]=(0.888165993988)*x[0]**o+(-0.611433401469)*x[0]+(0.840713835877)*x[1]**o+(-0.805947825209)*x[1]+(-0.223790913101)*x[2]**o+(0.274074627813)*x[2]
        s[0,3,2,1]=(-0.0586612521858)*x[0]**o+(-0.904292259669)*x[0]+(0.701563259971)*x[1]**o+(-0.919454807365)*x[1]+(-0.432367433099)*x[2]**o+(-0.915055158557)*x[2]
        s[0,4,0,0]=(-0.896479628968)*x[0]**o+(-0.0775794895972)*x[0]+(0.952385683546)*x[1]**o+(0.101370020306)*x[1]+(-0.311916160508)*x[2]**o+(0.327792852845)*x[2]
        s[0,4,0,1]=(-0.368307750793)*x[0]**o+(-0.824637391638)*x[0]+(0.143668778522)*x[1]**o+(0.224471763433)*x[1]+(0.470345995942)*x[2]**o+(-0.238496191133)*x[2]
        s[0,4,1,0]=(0.978800093563)*x[0]**o+(0.356297491637)*x[0]+(-0.846067544583)*x[1]**o+(0.0112192363669)*x[1]+(-0.456308213004)*x[2]**o+(0.328816226899)*x[2]
        s[0,4,1,1]=(0.860778047982)*x[0]**o+(-0.85484855319)*x[0]+(0.431318104432)*x[1]**o+(0.0519725606118)*x[1]+(-0.37452257816)*x[2]**o+(0.387786048315)*x[2]
        s[0,4,2,0]=(-0.809652418528)*x[0]**o+(0.622059542769)*x[0]+(0.0769858780015)*x[1]**o+(-0.1118510671)*x[1]+(0.462693639527)*x[2]**o+(-0.321289200501)*x[2]
        s[0,4,2,1]=(-0.112319716285)*x[0]**o+(-0.97958472439)*x[0]+(0.471152868527)*x[1]**o+(0.979477042696)*x[1]+(0.343437406244)*x[2]**o+(0.330245074789)*x[2]
        s[1,0,0,0]=(-0.664959353312)*x[0]**o+(0.667428688064)*x[0]+(0.379323454697)*x[1]**o+(0.458212123636)*x[1]+(-0.583770985715)*x[2]**o+(0.926174954355)*x[2]
        s[1,0,0,1]=(0.463653553012)*x[0]**o+(-0.998470865927)*x[0]+(-0.0590876546604)*x[1]**o+(0.22211328071)*x[1]+(-0.741158466684)*x[2]**o+(-0.405853581466)*x[2]
        s[1,0,1,0]=(0.261926286639)*x[0]**o+(-0.291967639861)*x[0]+(-0.22306024334)*x[1]**o+(-0.266187302656)*x[1]+(0.0808303795362)*x[2]**o+(-0.0411340553637)*x[2]
        s[1,0,1,1]=(0.823792286605)*x[0]**o+(-0.622764469917)*x[0]+(0.532406278339)*x[1]**o+(-0.799863805036)*x[1]+(0.356182760763)*x[2]**o+(0.633305652052)*x[2]
        s[1,0,2,0]=(0.386490056692)*x[0]**o+(-0.538422996672)*x[0]+(-0.400192004838)*x[1]**o+(0.632203703462)*x[1]+(0.315043759293)*x[2]**o+(-0.10635477353)*x[2]
        s[1,0,2,1]=(0.911736934682)*x[0]**o+(0.874267947717)*x[0]+(0.716599906083)*x[1]**o+(-0.477115250943)*x[1]+(0.296312692731)*x[2]**o+(0.691409352221)*x[2]
        s[1,1,0,0]=(0.193258443163)*x[0]**o+(-0.479351523999)*x[0]+(0.426152720777)*x[1]**o+(0.577292380929)*x[1]+(0.781776817395)*x[2]**o+(-0.0622489459224)*x[2]
        s[1,1,0,1]=(-0.491606711738)*x[0]**o+(-0.138919126056)*x[0]+(0.10073957381)*x[1]**o+(0.560595873872)*x[1]+(0.661805959053)*x[2]**o+(0.999249400576)*x[2]
        s[1,1,1,0]=(-0.197096694248)*x[0]**o+(0.205033675186)*x[0]+(0.265896904632)*x[1]**o+(0.448157190065)*x[1]+(0.284574920599)*x[2]**o+(-0.466032134334)*x[2]
        s[1,1,1,1]=(0.852594063486)*x[0]**o+(-0.139990914704)*x[0]+(0.0792640568063)*x[1]**o+(0.279425947587)*x[1]+(-0.598534354914)*x[2]**o+(-0.423199533073)*x[2]
        s[1,1,2,0]=(0.576687955944)*x[0]**o+(-0.616599888356)*x[0]+(-0.826670910311)*x[1]**o+(0.54433728322)*x[1]+(0.00123761537904)*x[2]**o+(0.716238991968)*x[2]
        s[1,1,2,1]=(0.56855109067)*x[0]**o+(-0.0440632979111)*x[0]+(0.526871640009)*x[1]**o+(0.275141335835)*x[1]+(-0.983229310103)*x[2]**o+(0.700105733507)*x[2]
        s[1,2,0,0]=(-0.184742995569)*x[0]**o+(0.71036642023)*x[0]+(0.981161346217)*x[1]**o+(0.365488167653)*x[1]+(-0.398883902177)*x[2]**o+(-0.482354856056)*x[2]
        s[1,2,0,1]=(0.583421736253)*x[0]**o+(-0.279163572851)*x[0]+(0.503764541147)*x[1]**o+(0.292381662704)*x[1]+(-0.888007033365)*x[2]**o+(-0.448668581094)*x[2]
        s[1,2,1,0]=(0.806086468959)*x[0]**o+(-0.959525959302)*x[0]+(0.734809833547)*x[1]**o+(-0.284558836843)*x[1]+(-0.354048559091)*x[2]**o+(-0.464063301273)*x[2]
        s[1,2,1,1]=(0.20737358797)*x[0]**o+(-0.958445091026)*x[0]+(-0.207486744384)*x[1]**o+(-0.300655801991)*x[1]+(0.534182391535)*x[2]**o+(-0.630595215476)*x[2]
        s[1,2,2,0]=(0.573970416565)*x[0]**o+(0.0868191347586)*x[0]+(0.90996767074)*x[1]**o+(-0.556896647781)*x[1]+(-0.57168584006)*x[2]**o+(-0.000855122180503)*x[2]
        s[1,2,2,1]=(-0.0500416574964)*x[0]**o+(0.134962389253)*x[0]+(0.7037800261)*x[1]**o+(-0.46634099981)*x[1]+(-0.978823012363)*x[2]**o+(0.429717517544)*x[2]
        s[1,3,0,0]=(-0.480934294581)*x[0]**o+(-0.761664422527)*x[0]+(0.563388700943)*x[1]**o+(0.286474912293)*x[1]+(-0.254087945936)*x[2]**o+(0.823788425201)*x[2]
        s[1,3,0,1]=(0.287760411402)*x[0]**o+(-0.667747306058)*x[0]+(-0.801505580774)*x[1]**o+(0.642958552096)*x[1]+(-0.443695399886)*x[2]**o+(-0.840149579234)*x[2]
        s[1,3,1,0]=(-0.789394564521)*x[0]**o+(0.918638482871)*x[0]+(0.584254838389)*x[1]**o+(-0.150259802986)*x[1]+(0.794702325874)*x[2]**o+(0.0246788014415)*x[2]
        s[1,3,1,1]=(-0.1183405979)*x[0]**o+(0.0451549250255)*x[0]+(0.1517987763)*x[1]**o+(-0.200979231459)*x[1]+(-0.0817034681199)*x[2]**o+(0.664693805987)*x[2]
        s[1,3,2,0]=(-0.364619176305)*x[0]**o+(0.989060492595)*x[0]+(-0.452469848404)*x[1]**o+(-0.622993726081)*x[1]+(0.738661997825)*x[2]**o+(0.89325891611)*x[2]
        s[1,3,2,1]=(0.794435451296)*x[0]**o+(0.561632824215)*x[0]+(-0.661460086901)*x[1]**o+(0.110534045458)*x[1]+(-0.11627022197)*x[2]**o+(-0.737456873271)*x[2]
        s[1,4,0,0]=(0.821875192234)*x[0]**o+(0.378653978743)*x[0]+(0.384830574445)*x[1]**o+(0.815762479042)*x[1]+(0.709307414564)*x[2]**o+(0.785945406666)*x[2]
        s[1,4,0,1]=(-0.195473903829)*x[0]**o+(-0.792429456326)*x[0]+(0.475482962939)*x[1]**o+(0.02846998342)*x[1]+(0.433471515314)*x[2]**o+(0.0565560487548)*x[2]
        s[1,4,1,0]=(-0.927049196862)*x[0]**o+(-0.770563924424)*x[0]+(0.722391079777)*x[1]**o+(-0.962242600193)*x[1]+(-0.652853937129)*x[2]**o+(-0.583745627826)*x[2]
        s[1,4,1,1]=(-0.955616614181)*x[0]**o+(0.194945562768)*x[0]+(-0.517066615715)*x[1]**o+(-0.712362265862)*x[1]+(-0.524911207386)*x[2]**o+(-0.882591220196)*x[2]
        s[1,4,2,0]=(0.524532786764)*x[0]**o+(0.284840195202)*x[0]+(0.963855761362)*x[1]**o+(-0.856530057166)*x[1]+(-0.0740070151584)*x[2]**o+(0.226341867159)*x[2]
        s[1,4,2,1]=(0.446800509767)*x[0]**o+(-0.864046465502)*x[0]+(-0.517438631247)*x[1]**o+(0.901374962381)*x[1]+(0.601052625973)*x[2]**o+(-0.863564939303)*x[2]
        s[2,0,0,0]=(-0.701781199314)*x[0]**o+(0.264327906644)*x[0]+(0.971572320137)*x[1]**o+(-0.168607506193)*x[1]+(-0.526717738449)*x[2]**o+(0.226927089015)*x[2]
        s[2,0,0,1]=(-0.847412912499)*x[0]**o+(-0.401413370017)*x[0]+(0.759311464889)*x[1]**o+(0.457894962279)*x[1]+(-0.55434272035)*x[2]**o+(0.120326915639)*x[2]
        s[2,0,1,0]=(-0.517103294855)*x[0]**o+(0.492756717159)*x[0]+(0.973627035185)*x[1]**o+(-0.860482824403)*x[1]+(0.54057457411)*x[2]**o+(-0.874495812872)*x[2]
        s[2,0,1,1]=(-0.932588871815)*x[0]**o+(-0.741303214077)*x[0]+(0.468667081235)*x[1]**o+(0.395164667731)*x[1]+(0.412999483627)*x[2]**o+(0.827200271655)*x[2]
        s[2,0,2,0]=(0.324804049142)*x[0]**o+(0.687197584348)*x[0]+(0.48017530302)*x[1]**o+(0.0594198080606)*x[1]+(-0.243274785097)*x[2]**o+(0.561027778867)*x[2]
        s[2,0,2,1]=(-0.830970533632)*x[0]**o+(-0.399321288145)*x[0]+(-0.715545820929)*x[1]**o+(-0.290047681469)*x[1]+(-0.335483597978)*x[2]**o+(0.666741330684)*x[2]
        s[2,1,0,0]=(-0.734079757702)*x[0]**o+(0.290309222969)*x[0]+(0.764118958044)*x[1]**o+(0.4132335859)*x[1]+(0.257031297672)*x[2]**o+(0.244389627708)*x[2]
        s[2,1,0,1]=(0.687268551061)*x[0]**o+(-0.24563836651)*x[0]+(-0.123765123048)*x[1]**o+(-0.530923035235)*x[1]+(-0.720827493122)*x[2]**o+(0.210975066905)*x[2]
        s[2,1,1,0]=(0.813659604982)*x[0]**o+(0.893957744697)*x[0]+(0.991947458807)*x[1]**o+(0.793408082473)*x[1]+(0.609236290624)*x[2]**o+(-0.0658228268111)*x[2]
        s[2,1,1,1]=(-0.118289813545)*x[0]**o+(0.622749254009)*x[0]+(0.465005967301)*x[1]**o+(-0.878906002448)*x[1]+(0.168374644259)*x[2]**o+(0.149889108873)*x[2]
        s[2,1,2,0]=(-0.255533627884)*x[0]**o+(-0.436918772486)*x[0]+(0.29166400203)*x[1]**o+(0.928052058386)*x[1]+(0.303846927853)*x[2]**o+(0.158922482472)*x[2]
        s[2,1,2,1]=(-0.0504039593843)*x[0]**o+(0.170081032189)*x[0]+(-0.739948814307)*x[1]**o+(-0.52541309551)*x[1]+(0.797681120077)*x[2]**o+(0.247665713856)*x[2]
        s[2,2,0,0]=(-0.400562988004)*x[0]**o+(-0.128489671375)*x[0]+(-0.451867300451)*x[1]**o+(-0.343423869937)*x[1]+(0.663184069698)*x[2]**o+(0.926436292949)*x[2]
        s[2,2,0,1]=(-0.599719778431)*x[0]**o+(0.108577031922)*x[0]+(0.217879834507)*x[1]**o+(0.907899596451)*x[1]+(-0.161562448509)*x[2]**o+(-0.514018490585)*x[2]
        s[2,2,1,0]=(-0.736520696776)*x[0]**o+(0.166243034841)*x[0]+(0.438311085394)*x[1]**o+(0.802937292983)*x[1]+(-0.938817901098)*x[2]**o+(-0.205869039453)*x[2]
        s[2,2,1,1]=(0.555108741686)*x[0]**o+(-0.533145284484)*x[0]+(-0.763891188865)*x[1]**o+(0.515585725441)*x[1]+(0.638581800777)*x[2]**o+(0.629321554502)*x[2]
        s[2,2,2,0]=(-0.790093516383)*x[0]**o+(-0.0154620341689)*x[0]+(0.462011424745)*x[1]**o+(0.598267220586)*x[1]+(0.576054554929)*x[2]**o+(0.482972076666)*x[2]
        s[2,2,2,1]=(0.075201635129)*x[0]**o+(-0.660059195587)*x[0]+(0.00711502517884)*x[1]**o+(-0.391407272114)*x[1]+(-0.213729719317)*x[2]**o+(-0.561573742069)*x[2]
        s[2,3,0,0]=(0.350797999062)*x[0]**o+(0.250500699573)*x[0]+(0.818402537945)*x[1]**o+(-0.899093594018)*x[1]+(-0.162195807738)*x[2]**o+(-0.147586906331)*x[2]
        s[2,3,0,1]=(0.695618803281)*x[0]**o+(0.227239221026)*x[0]+(-0.968505693287)*x[1]**o+(-0.473260919549)*x[1]+(-0.746882331556)*x[2]**o+(0.486703205212)*x[2]
        s[2,3,1,0]=(-0.271378189855)*x[0]**o+(-0.561268063707)*x[0]+(-0.0252286382422)*x[1]**o+(0.229921692249)*x[1]+(0.989187908151)*x[2]**o+(0.643686783302)*x[2]
        s[2,3,1,1]=(0.750682282794)*x[0]**o+(-0.27757575698)*x[0]+(0.604377636669)*x[1]**o+(-0.733927934644)*x[1]+(-0.718229138068)*x[2]**o+(-0.435426875038)*x[2]
        s[2,3,2,0]=(0.0733516934174)*x[0]**o+(-0.174475004056)*x[0]+(0.478942713703)*x[1]**o+(0.662549744231)*x[1]+(-0.565016805898)*x[2]**o+(0.17793235315)*x[2]
        s[2,3,2,1]=(-0.15216608314)*x[0]**o+(0.999325854219)*x[0]+(0.496048192253)*x[1]**o+(-0.674178160594)*x[1]+(-0.927955584861)*x[2]**o+(0.156845563256)*x[2]
        s[2,4,0,0]=(-0.912208708086)*x[0]**o+(0.929579552168)*x[0]+(-0.0445821051156)*x[1]**o+(0.911346880333)*x[1]+(0.309997148008)*x[2]**o+(0.0975168137997)*x[2]
        s[2,4,0,1]=(0.65092333822)*x[0]**o+(0.605218122961)*x[0]+(-0.684013078124)*x[1]**o+(0.088688492152)*x[1]+(-0.0554688873336)*x[2]**o+(-0.292971890481)*x[2]
        s[2,4,1,0]=(-0.437643682456)*x[0]**o+(-0.161540806934)*x[0]+(-0.863085496236)*x[1]**o+(0.714210296119)*x[1]+(0.972219783929)*x[2]**o+(-0.0450752608276)*x[2]
        s[2,4,1,1]=(0.838644689829)*x[0]**o+(0.876692248024)*x[0]+(-0.787815110065)*x[1]**o+(-0.867528859191)*x[1]+(-0.393677502867)*x[2]**o+(-0.522478124784)*x[2]
        s[2,4,2,0]=(-0.606364594289)*x[0]**o+(0.88715735186)*x[0]+(-0.894189269194)*x[1]**o+(0.470967389174)*x[1]+(0.427781385837)*x[2]**o+(0.141054076833)*x[2]
        s[2,4,2,1]=(-0.0329322907424)*x[0]**o+(0.0437971796726)*x[0]+(-0.706648559646)*x[1]**o+(0.132643884341)*x[1]+(0.101209934477)*x[2]**o+(-0.556234167371)*x[2]
        s[3,0,0,0]=(0.25287903621)*x[0]**o+(-0.281221279197)*x[0]+(0.643850500988)*x[1]**o+(0.0269082516545)*x[1]+(0.157104400545)*x[2]**o+(-0.554382511827)*x[2]
        s[3,0,0,1]=(0.130965460225)*x[0]**o+(0.203133851634)*x[0]+(-0.876140347686)*x[1]**o+(0.871803270423)*x[1]+(-0.372911987943)*x[2]**o+(-0.296478871771)*x[2]
        s[3,0,1,0]=(0.395462973708)*x[0]**o+(-0.0463119371121)*x[0]+(-0.810272498493)*x[1]**o+(0.052842476247)*x[1]+(-0.632091659537)*x[2]**o+(0.0477272308506)*x[2]
        s[3,0,1,1]=(-0.15375761581)*x[0]**o+(0.840485309457)*x[0]+(-0.183510151479)*x[1]**o+(0.423252205076)*x[1]+(0.0972125869507)*x[2]**o+(-0.873325211398)*x[2]
        s[3,0,2,0]=(-0.230628763369)*x[0]**o+(-0.446642785223)*x[0]+(0.172403317403)*x[1]**o+(0.6666203719)*x[1]+(0.758132421761)*x[2]**o+(-0.0802337093298)*x[2]
        s[3,0,2,1]=(-0.375928720447)*x[0]**o+(0.390180617554)*x[0]+(0.532916128738)*x[1]**o+(-0.438049708768)*x[1]+(-0.969242415622)*x[2]**o+(-0.979491904432)*x[2]
        s[3,1,0,0]=(0.243017072038)*x[0]**o+(-0.314649526706)*x[0]+(-0.64964134438)*x[1]**o+(-0.319482380304)*x[1]+(0.969793978414)*x[2]**o+(0.235845670735)*x[2]
        s[3,1,0,1]=(0.5372323492)*x[0]**o+(-0.297603522377)*x[0]+(0.718422772931)*x[1]**o+(0.591124970588)*x[1]+(0.148470013869)*x[2]**o+(-0.906959786864)*x[2]
        s[3,1,1,0]=(-0.41004677863)*x[0]**o+(0.519928695266)*x[0]+(-0.41056500013)*x[1]**o+(0.545730013165)*x[1]+(0.0316754455631)*x[2]**o+(0.0735333716599)*x[2]
        s[3,1,1,1]=(-0.413084334125)*x[0]**o+(-0.188907203544)*x[0]+(0.333210666705)*x[1]**o+(0.795876759492)*x[1]+(-0.879028457556)*x[2]**o+(0.597499547791)*x[2]
        s[3,1,2,0]=(0.572164028218)*x[0]**o+(0.561153490096)*x[0]+(-0.277218859151)*x[1]**o+(-0.784951308131)*x[1]+(-0.54076329511)*x[2]**o+(0.232589967523)*x[2]
        s[3,1,2,1]=(0.183000187474)*x[0]**o+(-0.0315492936828)*x[0]+(0.145425750124)*x[1]**o+(0.570531080762)*x[1]+(0.321679325968)*x[2]**o+(0.768693312757)*x[2]
        s[3,2,0,0]=(0.0572073598209)*x[0]**o+(-0.473877874718)*x[0]+(0.19030371495)*x[1]**o+(0.950458284849)*x[1]+(0.520875156632)*x[2]**o+(0.746025413415)*x[2]
        s[3,2,0,1]=(-0.670180903965)*x[0]**o+(0.792714072874)*x[0]+(-0.830626382455)*x[1]**o+(-0.694487379592)*x[1]+(0.708769579296)*x[2]**o+(0.874781581547)*x[2]
        s[3,2,1,0]=(-0.635101696551)*x[0]**o+(0.683413284169)*x[0]+(-0.580556747733)*x[1]**o+(0.228587157705)*x[1]+(0.283498698937)*x[2]**o+(-0.661341563969)*x[2]
        s[3,2,1,1]=(-0.938857602123)*x[0]**o+(-0.855150853757)*x[0]+(0.313962358596)*x[1]**o+(-0.137683481729)*x[1]+(-0.543347522619)*x[2]**o+(0.520234742546)*x[2]
        s[3,2,2,0]=(0.00237595749469)*x[0]**o+(0.151777614372)*x[0]+(-0.207526210221)*x[1]**o+(-0.593427746692)*x[1]+(-0.273081745725)*x[2]**o+(-0.103365939558)*x[2]
        s[3,2,2,1]=(-0.836487016235)*x[0]**o+(-0.499416194189)*x[0]+(0.944785244284)*x[1]**o+(0.316171200523)*x[1]+(0.882197908567)*x[2]**o+(-0.756125841104)*x[2]
        s[3,3,0,0]=(-0.331819442381)*x[0]**o+(0.0868286143286)*x[0]+(-0.0214955303831)*x[1]**o+(-0.802588129882)*x[1]+(-0.79665760605)*x[2]**o+(0.733766062704)*x[2]
        s[3,3,0,1]=(0.0832038861206)*x[0]**o+(-0.229342320163)*x[0]+(0.411514025627)*x[1]**o+(-0.278316368324)*x[1]+(0.158512271865)*x[2]**o+(-0.176346585884)*x[2]
        s[3,3,1,0]=(0.813577309153)*x[0]**o+(0.183181445534)*x[0]+(-0.625433189537)*x[1]**o+(-0.296559044989)*x[1]+(0.735593889495)*x[2]**o+(-0.84760854319)*x[2]
        s[3,3,1,1]=(0.129952300256)*x[0]**o+(-0.900793812246)*x[0]+(0.723956240897)*x[1]**o+(0.906884554981)*x[1]+(-0.828597157505)*x[2]**o+(-0.0453629922512)*x[2]
        s[3,3,2,0]=(0.618506925918)*x[0]**o+(0.537560191905)*x[0]+(0.274041602538)*x[1]**o+(0.128748880801)*x[1]+(-0.0421945967926)*x[2]**o+(-0.402512195673)*x[2]
        s[3,3,2,1]=(-0.882194818723)*x[0]**o+(-0.554059181227)*x[0]+(-0.519424591535)*x[1]**o+(0.0254864032182)*x[1]+(-0.163939472982)*x[2]**o+(-0.0494335206145)*x[2]
        s[3,4,0,0]=(-0.583301433479)*x[0]**o+(0.155758796662)*x[0]+(-0.814066320026)*x[1]**o+(0.152299298845)*x[1]+(0.988818040849)*x[2]**o+(-0.538805205144)*x[2]
        s[3,4,0,1]=(0.83112196296)*x[0]**o+(-0.719967217592)*x[0]+(-0.283369672135)*x[1]**o+(-0.995074593345)*x[1]+(0.129890006169)*x[2]**o+(-0.565650038645)*x[2]
        s[3,4,1,0]=(0.290242298945)*x[0]**o+(0.106236639888)*x[0]+(0.881353634263)*x[1]**o+(0.578003018941)*x[1]+(0.337748119883)*x[2]**o+(0.0495137765079)*x[2]
        s[3,4,1,1]=(-0.392371956935)*x[0]**o+(-0.403685094924)*x[0]+(0.953131101032)*x[1]**o+(0.144162998481)*x[1]+(0.13946212185)*x[2]**o+(-0.951124223008)*x[2]
        s[3,4,2,0]=(0.997965631913)*x[0]**o+(-0.494892241571)*x[0]+(0.887313590406)*x[1]**o+(-0.947323203129)*x[1]+(0.150412298394)*x[2]**o+(-0.0541618066578)*x[2]
        s[3,4,2,1]=(-0.177392402935)*x[0]**o+(0.338956222402)*x[0]+(0.746432610869)*x[1]**o+(-0.4930192355)*x[1]+(-0.210191759258)*x[2]**o+(0.501879127576)*x[2]
        ref[0,0,0,0]=(-0.0155247832423)*(1+2.*(dim-1.)/(o+1.))+(-0.430291051587)*dim
        ref[0,0,0,1]=(0.106727527157)*(1+2.*(dim-1.)/(o+1.))+(-0.651341788175)*dim
        ref[0,0,1,0]=(-0.859508213698)*(1+2.*(dim-1.)/(o+1.))+(1.73876967515)*dim
        ref[0,0,1,1]=(0.361322484688)*(1+2.*(dim-1.)/(o+1.))+(0.43166904701)*dim
        ref[0,0,2,0]=(-1.45264359807)*(1+2.*(dim-1.)/(o+1.))+(0.501916314636)*dim
        ref[0,0,2,1]=(-0.668900247351)*(1+2.*(dim-1.)/(o+1.))+(-1.65623984537)*dim
        ref[0,1,0,0]=(-0.443343570079)*(1+2.*(dim-1.)/(o+1.))+(0.871168888686)*dim
        ref[0,1,0,1]=(0.118415498213)*(1+2.*(dim-1.)/(o+1.))+(0.74798799462)*dim
        ref[0,1,1,0]=(0.920954236468)*(1+2.*(dim-1.)/(o+1.))+(-1.17034075437)*dim
        ref[0,1,1,1]=(0.54354015742)*(1+2.*(dim-1.)/(o+1.))+(0.0278374436795)*dim
        ref[0,1,2,0]=(-0.170343899948)*(1+2.*(dim-1.)/(o+1.))+(-0.871171846531)*dim
        ref[0,1,2,1]=(0.505172256825)*(1+2.*(dim-1.)/(o+1.))+(0.470437054524)*dim
        ref[0,2,0,0]=(1.08615296794)*(1+2.*(dim-1.)/(o+1.))+(1.06343890503)*dim
        ref[0,2,0,1]=(0.734996986295)*(1+2.*(dim-1.)/(o+1.))+(0.570979059821)*dim
        ref[0,2,1,0]=(-0.1518829769)*(1+2.*(dim-1.)/(o+1.))+(0.62649842867)*dim
        ref[0,2,1,1]=(1.08910114223)*(1+2.*(dim-1.)/(o+1.))+(-0.684346728672)*dim
        ref[0,2,2,0]=(0.0508734531183)*(1+2.*(dim-1.)/(o+1.))+(-0.933688106304)*dim
        ref[0,2,2,1]=(-0.228940443279)*(1+2.*(dim-1.)/(o+1.))+(-1.49944102508)*dim
        ref[0,3,0,0]=(0.64391320994)*(1+2.*(dim-1.)/(o+1.))+(0.674107408326)*dim
        ref[0,3,0,1]=(0.913115225162)*(1+2.*(dim-1.)/(o+1.))+(0.278946015231)*dim
        ref[0,3,1,0]=(-1.40820486136)*(1+2.*(dim-1.)/(o+1.))+(-0.0639222838911)*dim
        ref[0,3,1,1]=(0.417949210295)*(1+2.*(dim-1.)/(o+1.))+(-0.216455781248)*dim
        ref[0,3,2,0]=(1.50508891676)*(1+2.*(dim-1.)/(o+1.))+(-1.14330659887)*dim
        ref[0,3,2,1]=(0.210534574687)*(1+2.*(dim-1.)/(o+1.))+(-2.73880222559)*dim
        ref[0,4,0,0]=(-0.25601010593)*(1+2.*(dim-1.)/(o+1.))+(0.351583383554)*dim
        ref[0,4,0,1]=(0.245707023671)*(1+2.*(dim-1.)/(o+1.))+(-0.838661819338)*dim
        ref[0,4,1,0]=(-0.323575664024)*(1+2.*(dim-1.)/(o+1.))+(0.696332954903)*dim
        ref[0,4,1,1]=(0.917573574254)*(1+2.*(dim-1.)/(o+1.))+(-0.415089944263)*dim
        ref[0,4,2,0]=(-0.269972900999)*(1+2.*(dim-1.)/(o+1.))+(0.188919275167)*dim
        ref[0,4,2,1]=(0.702270558486)*(1+2.*(dim-1.)/(o+1.))+(0.330137393094)*dim
        ref[1,0,0,0]=(-0.869406884331)*(1+2.*(dim-1.)/(o+1.))+(2.05181576606)*dim
        ref[1,0,0,1]=(-0.336592568333)*(1+2.*(dim-1.)/(o+1.))+(-1.18221116668)*dim
        ref[1,0,1,0]=(0.119696422835)*(1+2.*(dim-1.)/(o+1.))+(-0.599288997881)*dim
        ref[1,0,1,1]=(1.71238132571)*(1+2.*(dim-1.)/(o+1.))+(-0.7893226229)*dim
        ref[1,0,2,0]=(0.301341811148)*(1+2.*(dim-1.)/(o+1.))+(-0.0125740667408)*dim
        ref[1,0,2,1]=(1.9246495335)*(1+2.*(dim-1.)/(o+1.))+(1.08856204899)*dim
        ref[1,1,0,0]=(1.40118798134)*(1+2.*(dim-1.)/(o+1.))+(0.0356919110082)*dim
        ref[1,1,0,1]=(0.270938821125)*(1+2.*(dim-1.)/(o+1.))+(1.42092614839)*dim
        ref[1,1,1,0]=(0.353375130983)*(1+2.*(dim-1.)/(o+1.))+(0.187158730917)*dim
        ref[1,1,1,1]=(0.333323765378)*(1+2.*(dim-1.)/(o+1.))+(-0.28376450019)*dim
        ref[1,1,2,0]=(-0.248745338987)*(1+2.*(dim-1.)/(o+1.))+(0.643976386833)*dim
        ref[1,1,2,1]=(0.112193420576)*(1+2.*(dim-1.)/(o+1.))+(0.931183771431)*dim
        ref[1,2,0,0]=(0.397534448472)*(1+2.*(dim-1.)/(o+1.))+(0.593499731827)*dim
        ref[1,2,0,1]=(0.199179244034)*(1+2.*(dim-1.)/(o+1.))+(-0.43545049124)*dim
        ref[1,2,1,0]=(1.18684774341)*(1+2.*(dim-1.)/(o+1.))+(-1.70814809742)*dim
        ref[1,2,1,1]=(0.534069235122)*(1+2.*(dim-1.)/(o+1.))+(-1.88969610849)*dim
        ref[1,2,2,0]=(0.912252247246)*(1+2.*(dim-1.)/(o+1.))+(-0.470932635203)*dim
        ref[1,2,2,1]=(-0.32508464376)*(1+2.*(dim-1.)/(o+1.))+(0.0983389069861)*dim
        ref[1,3,0,0]=(-0.171633539573)*(1+2.*(dim-1.)/(o+1.))+(0.348598914967)*dim
        ref[1,3,0,1]=(-0.957440569258)*(1+2.*(dim-1.)/(o+1.))+(-0.864938333196)*dim
        ref[1,3,1,0]=(0.589562599742)*(1+2.*(dim-1.)/(o+1.))+(0.793057481326)*dim
        ref[1,3,1,1]=(-0.0482452897207)*(1+2.*(dim-1.)/(o+1.))+(0.508869499554)*dim
        ref[1,3,2,0]=(-0.0784270268847)*(1+2.*(dim-1.)/(o+1.))+(1.25932568262)*dim
        ref[1,3,2,1]=(0.0167051424243)*(1+2.*(dim-1.)/(o+1.))+(-0.065290003598)*dim
        ref[1,4,0,0]=(1.91601318124)*(1+2.*(dim-1.)/(o+1.))+(1.98036186445)*dim
        ref[1,4,0,1]=(0.713480574424)*(1+2.*(dim-1.)/(o+1.))+(-0.707403424152)*dim
        ref[1,4,1,0]=(-0.857512054213)*(1+2.*(dim-1.)/(o+1.))+(-2.31655215244)*dim
        ref[1,4,1,1]=(-1.99759443728)*(1+2.*(dim-1.)/(o+1.))+(-1.40000792329)*dim
        ref[1,4,2,0]=(1.41438153297)*(1+2.*(dim-1.)/(o+1.))+(-0.345347994805)*dim
        ref[1,4,2,1]=(0.530414504493)*(1+2.*(dim-1.)/(o+1.))+(-0.826236442424)*dim
        ref[2,0,0,0]=(-0.256926617626)*(1+2.*(dim-1.)/(o+1.))+(0.322647489466)*dim
        ref[2,0,0,1]=(-0.64244416796)*(1+2.*(dim-1.)/(o+1.))+(0.176808507901)*dim
        ref[2,0,1,0]=(0.997098314441)*(1+2.*(dim-1.)/(o+1.))+(-1.24222192012)*dim
        ref[2,0,1,1]=(-0.0509223069533)*(1+2.*(dim-1.)/(o+1.))+(0.481061725309)*dim
        ref[2,0,2,0]=(0.561704567066)*(1+2.*(dim-1.)/(o+1.))+(1.30764517127)*dim
        ref[2,0,2,1]=(-1.88199995254)*(1+2.*(dim-1.)/(o+1.))+(-0.0226276389294)*dim
        ref[2,1,0,0]=(0.287070498014)*(1+2.*(dim-1.)/(o+1.))+(0.947932436577)*dim
        ref[2,1,0,1]=(-0.157324065109)*(1+2.*(dim-1.)/(o+1.))+(-0.56558633484)*dim
        ref[2,1,1,0]=(2.41484335441)*(1+2.*(dim-1.)/(o+1.))+(1.62154300036)*dim
        ref[2,1,1,1]=(0.515090798015)*(1+2.*(dim-1.)/(o+1.))+(-0.106267639566)*dim
        ref[2,1,2,0]=(0.339977301998)*(1+2.*(dim-1.)/(o+1.))+(0.650055768373)*dim
        ref[2,1,2,1]=(0.00732834638585)*(1+2.*(dim-1.)/(o+1.))+(-0.107666349465)*dim
        ref[2,2,0,0]=(-0.189246218757)*(1+2.*(dim-1.)/(o+1.))+(0.454522751637)*dim
        ref[2,2,0,1]=(-0.543402392433)*(1+2.*(dim-1.)/(o+1.))+(0.502458137788)*dim
        ref[2,2,1,0]=(-1.23702751248)*(1+2.*(dim-1.)/(o+1.))+(0.76331128837)*dim
        ref[2,2,1,1]=(0.429799353598)*(1+2.*(dim-1.)/(o+1.))+(0.611761995459)*dim
        ref[2,2,2,0]=(0.24797246329)*(1+2.*(dim-1.)/(o+1.))+(1.06577726308)*dim
        ref[2,2,2,1]=(-0.131413059009)*(1+2.*(dim-1.)/(o+1.))+(-1.61304020977)*dim
        ref[2,3,0,0]=(1.00700472927)*(1+2.*(dim-1.)/(o+1.))+(-0.796179800777)*dim
        ref[2,3,0,1]=(-1.01976922156)*(1+2.*(dim-1.)/(o+1.))+(0.240681506689)*dim
        ref[2,3,1,0]=(0.692581080055)*(1+2.*(dim-1.)/(o+1.))+(0.312340411844)*dim
        ref[2,3,1,1]=(0.636830781395)*(1+2.*(dim-1.)/(o+1.))+(-1.44693056666)*dim
        ref[2,3,2,0]=(-0.0127223987779)*(1+2.*(dim-1.)/(o+1.))+(0.666007093324)*dim
        ref[2,3,2,1]=(-0.584073475747)*(1+2.*(dim-1.)/(o+1.))+(0.481993256882)*dim
        ref[2,4,0,0]=(-0.646793665194)*(1+2.*(dim-1.)/(o+1.))+(1.9384432463)*dim
        ref[2,4,0,1]=(-0.0885586272376)*(1+2.*(dim-1.)/(o+1.))+(0.400934724632)*dim
        ref[2,4,1,0]=(-0.328509394762)*(1+2.*(dim-1.)/(o+1.))+(0.507594228357)*dim
        ref[2,4,1,1]=(-0.342847923103)*(1+2.*(dim-1.)/(o+1.))+(-0.513314735952)*dim
        ref[2,4,2,0]=(-1.07277247765)*(1+2.*(dim-1.)/(o+1.))+(1.49917881787)*dim
        ref[2,4,2,1]=(-0.638370915912)*(1+2.*(dim-1.)/(o+1.))+(-0.379793103357)*dim
        ref[3,0,0,0]=(1.05383393774)*(1+2.*(dim-1.)/(o+1.))+(-0.80869553937)*dim
        ref[3,0,0,1]=(-1.1180868754)*(1+2.*(dim-1.)/(o+1.))+(0.778458250286)*dim
        ref[3,0,1,0]=(-1.04690118432)*(1+2.*(dim-1.)/(o+1.))+(0.0542577699855)*dim
        ref[3,0,1,1]=(-0.240055180339)*(1+2.*(dim-1.)/(o+1.))+(0.390412303135)*dim
        ref[3,0,2,0]=(0.699906975796)*(1+2.*(dim-1.)/(o+1.))+(0.139743877347)*dim
        ref[3,0,2,1]=(-0.812255007331)*(1+2.*(dim-1.)/(o+1.))+(-1.02736099565)*dim
        ref[3,1,0,0]=(0.563169706071)*(1+2.*(dim-1.)/(o+1.))+(-0.398286236275)*dim
        ref[3,1,0,1]=(1.404125136)*(1+2.*(dim-1.)/(o+1.))+(-0.613438338653)*dim
        ref[3,1,1,0]=(-0.788936333197)*(1+2.*(dim-1.)/(o+1.))+(1.13919208009)*dim
        ref[3,1,1,1]=(-0.958902124976)*(1+2.*(dim-1.)/(o+1.))+(1.20446910374)*dim
        ref[3,1,2,0]=(-0.245818126043)*(1+2.*(dim-1.)/(o+1.))+(0.00879214948826)*dim
        ref[3,1,2,1]=(0.650105263567)*(1+2.*(dim-1.)/(o+1.))+(1.30767509984)*dim
        ref[3,2,0,0]=(0.768386231403)*(1+2.*(dim-1.)/(o+1.))+(1.22260582355)*dim
        ref[3,2,0,1]=(-0.792037707123)*(1+2.*(dim-1.)/(o+1.))+(0.973008274829)*dim
        ref[3,2,1,0]=(-0.932159745347)*(1+2.*(dim-1.)/(o+1.))+(0.250658877904)*dim
        ref[3,2,1,1]=(-1.16824276615)*(1+2.*(dim-1.)/(o+1.))+(-0.47259959294)*dim
        ref[3,2,2,0]=(-0.478231998451)*(1+2.*(dim-1.)/(o+1.))+(-0.545016071879)*dim
        ref[3,2,2,1]=(0.990496136616)*(1+2.*(dim-1.)/(o+1.))+(-0.939370834771)*dim
        ref[3,3,0,0]=(-1.14997257881)*(1+2.*(dim-1.)/(o+1.))+(0.0180065471512)*dim
        ref[3,3,0,1]=(0.653230183613)*(1+2.*(dim-1.)/(o+1.))+(-0.684005274372)*dim
        ref[3,3,1,0]=(0.923738009112)*(1+2.*(dim-1.)/(o+1.))+(-0.960986142644)*dim
        ref[3,3,1,1]=(0.0253113836472)*(1+2.*(dim-1.)/(o+1.))+(-0.0392722495157)*dim
        ref[3,3,2,0]=(0.850353931663)*(1+2.*(dim-1.)/(o+1.))+(0.263796877033)*dim
        ref[3,3,2,1]=(-1.56555888324)*(1+2.*(dim-1.)/(o+1.))+(-0.578006298623)*dim
        ref[3,4,0,0]=(-0.408549712656)*(1+2.*(dim-1.)/(o+1.))+(-0.230747109637)*dim
        ref[3,4,0,1]=(0.677642296994)*(1+2.*(dim-1.)/(o+1.))+(-2.28069184958)*dim
        ref[3,4,1,0]=(1.50934405309)*(1+2.*(dim-1.)/(o+1.))+(0.733753435337)*dim
        ref[3,4,1,1]=(0.700221265947)*(1+2.*(dim-1.)/(o+1.))+(-1.21064631945)*dim
        ref[3,4,2,0]=(2.03569152071)*(1+2.*(dim-1.)/(o+1.))+(-1.49637725136)*dim
        ref[3,4,2,1]=(0.358848448676)*(1+2.*(dim-1.)/(o+1.))+(0.347816114478)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Symbol on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-1.08871335955)*x[0]+(0.247972631031)*x[1]
        ref=(-0.156403544195)*(1+2.*(dim-1.)/(o+1.))+(-0.684337184327)*dim
      else:
        s=(-0.324052151044)*x[0]+(0.441305482424)*x[1]+(-0.315030176112)*x[2]
        ref=(0.603955649036)*(1+2.*(dim-1.)/(o+1.))+(-0.801732493769)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Symbol on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.260677248137)*x[0]+(0.0211318398374)*x[1]
        s[1]=(-0.815879139039)*x[0]+(-0.918304389572)*x[1]
        ref[0]=(0.843100081089)*(1+2.*(dim-1.)/(o+1.))+(-1.08264548939)*dim
        ref[1]=(-1.3080191481)*(1+2.*(dim-1.)/(o+1.))+(-0.426164380506)*dim
      else:
        s[0]=(1.37908771986)*x[0]+(1.36518233103)*x[1]+(1.013246401)*x[2]
        s[1]=(0.556318683467)*x[0]+(1.45737568352)*x[1]+(0.0165078724907)*x[2]
        ref[0]=(2.05950634594)*(1+2.*(dim-1.)/(o+1.))+(1.69801010595)*dim
        ref[1]=(0.655472308826)*(1+2.*(dim-1.)/(o+1.))+(1.37472993065)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Symbol on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(1.44204352077)*x[0]+(-0.626836867163)*x[1]
        s[0,1]=(-0.295593397487)*x[0]+(-0.309954518607)*x[1]
        s[0,2]=(-0.546882380107)*x[0]+(-0.0925530562931)*x[1]
        s[0,3]=(-0.623335335636)*x[0]+(0.798887138514)*x[1]
        s[0,4]=(0.202234634868)*x[0]+(1.44014401659)*x[1]
        s[1,0]=(-0.772178934874)*x[0]+(-0.962339450381)*x[1]
        s[1,1]=(-0.136928685359)*x[0]+(0.385217139725)*x[1]
        s[1,2]=(-0.00807747505852)*x[0]+(1.08230267787)*x[1]
        s[1,3]=(-1.28422033803)*x[0]+(0.011228260616)*x[1]
        s[1,4]=(0.766746911823)*x[0]+(-0.145254490276)*x[1]
        s[2,0]=(1.65171936301)*x[0]+(1.68030971997)*x[1]
        s[2,1]=(-0.465968023499)*x[0]+(-0.594634053719)*x[1]
        s[2,2]=(-0.140916975546)*x[0]+(0.230405132955)*x[1]
        s[2,3]=(-0.0185338227586)*x[0]+(0.298041051546)*x[1]
        s[2,4]=(-0.024340230778)*x[0]+(-0.791873643457)*x[1]
        s[3,0]=(0.555336313403)*x[0]+(0.606898490822)*x[1]
        s[3,1]=(1.05785941408)*x[0]+(-0.672561947115)*x[1]
        s[3,2]=(1.36138187068)*x[0]+(0.259592704164)*x[1]
        s[3,3]=(0.66211210256)*x[0]+(-1.38753456619)*x[1]
        s[3,4]=(0.0789301309156)*x[0]+(1.24015676345)*x[1]
        ref[0,0]=(0.863657572901)*(1+2.*(dim-1.)/(o+1.))+(-0.0484509192919)*dim
        ref[0,1]=(-0.0608826851336)*(1+2.*(dim-1.)/(o+1.))+(-0.54466523096)*dim
        ref[0,2]=(-1.25630630361)*(1+2.*(dim-1.)/(o+1.))+(0.616870867215)*dim
        ref[0,3]=(-0.618026054387)*(1+2.*(dim-1.)/(o+1.))+(0.793577857265)*dim
        ref[0,4]=(1.55309725329)*(1+2.*(dim-1.)/(o+1.))+(0.0892813981594)*dim
        ref[1,0]=(-0.707427946889)*(1+2.*(dim-1.)/(o+1.))+(-1.02709043837)*dim
        ref[1,1]=(1.0516700664)*(1+2.*(dim-1.)/(o+1.))+(-0.803381612038)*dim
        ref[1,2]=(-0.251691185365)*(1+2.*(dim-1.)/(o+1.))+(1.32591638818)*dim
        ref[1,3]=(0.227444315739)*(1+2.*(dim-1.)/(o+1.))+(-1.50043639316)*dim
        ref[1,4]=(0.636926531504)*(1+2.*(dim-1.)/(o+1.))+(-0.0154341099568)*dim
        ref[2,0]=(1.58333782129)*(1+2.*(dim-1.)/(o+1.))+(1.7486912617)*dim
        ref[2,1]=(0.376537776448)*(1+2.*(dim-1.)/(o+1.))+(-1.43713985367)*dim
        ref[2,2]=(1.33315471883)*(1+2.*(dim-1.)/(o+1.))+(-1.24366656142)*dim
        ref[2,3]=(-0.893476349046)*(1+2.*(dim-1.)/(o+1.))+(1.17298357783)*dim
        ref[2,4]=(-0.556720283569)*(1+2.*(dim-1.)/(o+1.))+(-0.259493590666)*dim
        ref[3,0]=(-0.0344322176845)*(1+2.*(dim-1.)/(o+1.))+(1.19666702191)*dim
        ref[3,1]=(0.356058306775)*(1+2.*(dim-1.)/(o+1.))+(0.0292391601866)*dim
        ref[3,2]=(1.18926653697)*(1+2.*(dim-1.)/(o+1.))+(0.431708037874)*dim
        ref[3,3]=(-0.0855874137907)*(1+2.*(dim-1.)/(o+1.))+(-0.639835049842)*dim
        ref[3,4]=(0.485696084789)*(1+2.*(dim-1.)/(o+1.))+(0.833390809579)*dim
      else:
        s[0,0]=(1.48740802799)*x[0]+(-0.5231718635)*x[1]+(0.621553689326)*x[2]
        s[0,1]=(-0.816432702835)*x[0]+(-1.02472113234)*x[1]+(-0.498250552797)*x[2]
        s[0,2]=(-0.0547881618906)*x[0]+(0.903306651062)*x[1]+(0.471428097681)*x[2]
        s[0,3]=(-0.910642152807)*x[0]+(-0.958224215021)*x[1]+(-1.4403271378)*x[2]
        s[0,4]=(0.116198234381)*x[0]+(-0.636378388418)*x[1]+(0.1270635999)*x[2]
        s[1,0]=(-0.659576291097)*x[0]+(-0.446587669482)*x[1]+(1.38617928144)*x[2]
        s[1,1]=(0.442922554555)*x[0]+(0.2841566742)*x[1]+(-0.384854418521)*x[2]
        s[1,2]=(-0.883255151971)*x[0]+(-1.3015449692)*x[1]+(-1.49785758846)*x[2]
        s[1,3]=(-1.43150392129)*x[0]+(1.13296295364)*x[1]+(0.393410729481)*x[2]
        s[1,4]=(-0.423254486419)*x[0]+(1.58675432746)*x[1]+(1.46179069555)*x[2]
        s[2,0]=(0.467182623429)*x[0]+(-1.23530058455)*x[1]+(0.546258172921)*x[2]
        s[2,1]=(-1.2468563949)*x[0]+(1.05285375014)*x[1]+(1.00605876886)*x[2]
        s[2,2]=(-0.1261124635)*x[0]+(-0.855000687125)*x[1]+(-0.252904511506)*x[2]
        s[2,3]=(1.1950830615)*x[0]+(0.699000137772)*x[1]+(1.33793802349)*x[2]
        s[2,4]=(-1.26581081559)*x[0]+(-0.0366783527385)*x[1]+(-0.736592212733)*x[2]
        s[3,0]=(1.84878865394)*x[0]+(0.504454193662)*x[1]+(-1.15690555199)*x[2]
        s[3,1]=(-0.249497283612)*x[0]+(-0.635940805715)*x[1]+(0.168382197376)*x[2]
        s[3,2]=(-0.167328351962)*x[0]+(0.523012476804)*x[1]+(1.19870547556)*x[2]
        s[3,3]=(-0.499150198998)*x[0]+(0.448461159266)*x[1]+(0.554069089559)*x[2]
        s[3,4]=(-0.520146860421)*x[0]+(1.43515582204)*x[1]+(1.31685474485)*x[2]
        ref[0,0]=(1.56352587364)*(1+2.*(dim-1.)/(o+1.))+(0.0222639801801)*dim
        ref[0,1]=(-1.88354672396)*(1+2.*(dim-1.)/(o+1.))+(-0.455857664013)*dim
        ref[0,2]=(0.631897841337)*(1+2.*(dim-1.)/(o+1.))+(0.688048745515)*dim
        ref[0,3]=(-2.49358987828)*(1+2.*(dim-1.)/(o+1.))+(-0.815603627346)*dim
        ref[0,4]=(-1.43782849721)*(1+2.*(dim-1.)/(o+1.))+(1.04471194307)*dim
        ref[1,0]=(-0.257186244728)*(1+2.*(dim-1.)/(o+1.))+(0.537201565586)*dim
        ref[1,1]=(-1.21753131847)*(1+2.*(dim-1.)/(o+1.))+(1.55975612871)*dim
        ref[1,2]=(-1.83391505519)*(1+2.*(dim-1.)/(o+1.))+(-1.84874265444)*dim
        ref[1,3]=(0.790398824851)*(1+2.*(dim-1.)/(o+1.))+(-0.69552906302)*dim
        ref[1,4]=(1.96838334672)*(1+2.*(dim-1.)/(o+1.))+(0.656907189867)*dim
        ref[2,0]=(-0.486341534759)*(1+2.*(dim-1.)/(o+1.))+(0.264481746559)*dim
        ref[2,1]=(-0.417272431595)*(1+2.*(dim-1.)/(o+1.))+(1.22932855569)*dim
        ref[2,2]=(-0.881184195775)*(1+2.*(dim-1.)/(o+1.))+(-0.352833466355)*dim
        ref[2,3]=(1.01420597285)*(1+2.*(dim-1.)/(o+1.))+(2.21781524991)*dim
        ref[2,4]=(-0.580092347401)*(1+2.*(dim-1.)/(o+1.))+(-1.45898903366)*dim
        ref[3,0]=(0.865857995406)*(1+2.*(dim-1.)/(o+1.))+(0.330479300213)*dim
        ref[3,1]=(0.632704550785)*(1+2.*(dim-1.)/(o+1.))+(-1.34976044274)*dim
        ref[3,2]=(-0.259081469822)*(1+2.*(dim-1.)/(o+1.))+(1.81347107022)*dim
        ref[3,3]=(0.533963525526)*(1+2.*(dim-1.)/(o+1.))+(-0.0305834756989)*dim
        ref[3,4]=(1.19847134659)*(1+2.*(dim-1.)/(o+1.))+(1.03339235988)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Symbol on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.536617774033)*x[0]+(-1.31191068914)*x[1]
        s[0,0,1]=(1.45706261992)*x[0]+(0.33760400949)*x[1]
        s[0,1,0]=(-0.453569968232)*x[0]+(0.798108537393)*x[1]
        s[0,1,1]=(-1.26956398588)*x[0]+(-1.35096551318)*x[1]
        s[1,0,0]=(0.104407588284)*x[0]+(-0.649935519811)*x[1]
        s[1,0,1]=(-1.64047091856)*x[0]+(0.654330987243)*x[1]
        s[1,1,0]=(-0.161010086142)*x[0]+(-0.644341383353)*x[1]
        s[1,1,1]=(1.42345151896)*x[0]+(-0.963960533124)*x[1]
        s[2,0,0]=(0.069463733935)*x[0]+(-0.764246138958)*x[1]
        s[2,0,1]=(0.516672615671)*x[0]+(0.32129769972)*x[1]
        s[2,1,0]=(0.399513997078)*x[0]+(-1.76803609001)*x[1]
        s[2,1,1]=(0.0570653989294)*x[0]+(-0.191869561254)*x[1]
        s[3,0,0]=(-0.261019396242)*x[0]+(0.429544509966)*x[1]
        s[3,0,1]=(-1.40560961029)*x[0]+(-0.739182555238)*x[1]
        s[3,1,0]=(0.314850928569)*x[0]+(1.49030138742)*x[1]
        s[3,1,1]=(-1.27327662424)*x[0]+(-0.0287977839083)*x[1]
        s[4,0,0]=(-0.840541224777)*x[0]+(-0.805224659169)*x[1]
        s[4,0,1]=(0.303904176581)*x[0]+(0.353767617246)*x[1]
        s[4,1,0]=(1.65813306507)*x[0]+(1.12299575058)*x[1]
        s[4,1,1]=(-1.19199290034)*x[0]+(1.6581285511)*x[1]
        s[5,0,0]=(0.120283505072)*x[0]+(1.68044908685)*x[1]
        s[5,0,1]=(0.382489278089)*x[0]+(0.129887092126)*x[1]
        s[5,1,0]=(-1.01890077156)*x[0]+(-0.111140290699)*x[1]
        s[5,1,1]=(-0.0392100253378)*x[0]+(0.579951319719)*x[1]
        ref[0,0,0]=(-0.686284158891)*(1+2.*(dim-1.)/(o+1.))+(-0.089008756216)*dim
        ref[0,0,1]=(1.76417819838)*(1+2.*(dim-1.)/(o+1.))+(0.030488431025)*dim
        ref[0,1,0]=(0.453381498643)*(1+2.*(dim-1.)/(o+1.))+(-0.108842929482)*dim
        ref[0,1,1]=(-1.03607416678)*(1+2.*(dim-1.)/(o+1.))+(-1.58445533227)*dim
        ref[1,0,0]=(0.638386446298)*(1+2.*(dim-1.)/(o+1.))+(-1.18391437783)*dim
        ref[1,0,1]=(0.275023590147)*(1+2.*(dim-1.)/(o+1.))+(-1.26116352146)*dim
        ref[1,1,0]=(-0.869379328882)*(1+2.*(dim-1.)/(o+1.))+(0.0640278593865)*dim
        ref[1,1,1]=(0.597799121533)*(1+2.*(dim-1.)/(o+1.))+(-0.138308135699)*dim
        ref[2,0,0]=(-1.45715684308)*(1+2.*(dim-1.)/(o+1.))+(0.762374438053)*dim
        ref[2,0,1]=(0.929718948925)*(1+2.*(dim-1.)/(o+1.))+(-0.0917486335329)*dim
        ref[2,1,0]=(-0.78092454297)*(1+2.*(dim-1.)/(o+1.))+(-0.587597549963)*dim
        ref[2,1,1]=(-1.21253063514)*(1+2.*(dim-1.)/(o+1.))+(1.07772647282)*dim
        ref[3,0,0]=(1.07007881303)*(1+2.*(dim-1.)/(o+1.))+(-0.901553699308)*dim
        ref[3,0,1]=(-1.77159657093)*(1+2.*(dim-1.)/(o+1.))+(-0.373195594594)*dim
        ref[3,1,0]=(1.10443227539)*(1+2.*(dim-1.)/(o+1.))+(0.700720040595)*dim
        ref[3,1,1]=(0.315983219905)*(1+2.*(dim-1.)/(o+1.))+(-1.61805762805)*dim
        ref[4,0,0]=(-0.602218266612)*(1+2.*(dim-1.)/(o+1.))+(-1.04354761733)*dim
        ref[4,0,1]=(0.297276459441)*(1+2.*(dim-1.)/(o+1.))+(0.360395334386)*dim
        ref[4,1,0]=(1.5584687424)*(1+2.*(dim-1.)/(o+1.))+(1.22266007325)*dim
        ref[4,1,1]=(0.504270780151)*(1+2.*(dim-1.)/(o+1.))+(-0.038135129389)*dim
        ref[5,0,0]=(1.35289487036)*(1+2.*(dim-1.)/(o+1.))+(0.44783772157)*dim
        ref[5,0,1]=(1.57660464714)*(1+2.*(dim-1.)/(o+1.))+(-1.06422827692)*dim
        ref[5,1,0]=(0.1360111664)*(1+2.*(dim-1.)/(o+1.))+(-1.26605222866)*dim
        ref[5,1,1]=(0.024986479373)*(1+2.*(dim-1.)/(o+1.))+(0.515754815008)*dim
      else:
        s[0,0,0]=(-0.491953157793)*x[0]+(0.808487399764)*x[1]+(0.87783526238)*x[2]
        s[0,0,1]=(-0.470158989175)*x[0]+(-0.492379274826)*x[1]+(-0.191290328078)*x[2]
        s[0,1,0]=(-0.356351922052)*x[0]+(0.823887078075)*x[1]+(-0.177422034195)*x[2]
        s[0,1,1]=(-0.378759483085)*x[0]+(0.116067213963)*x[1]+(0.0288265220551)*x[2]
        s[1,0,0]=(-0.325670782382)*x[0]+(-1.01289817775)*x[1]+(0.672608800327)*x[2]
        s[1,0,1]=(-0.296979811719)*x[0]+(0.46777334672)*x[1]+(1.61350930522)*x[2]
        s[1,1,0]=(-0.0574537817174)*x[0]+(1.10617918343)*x[1]+(0.172222519609)*x[2]
        s[1,1,1]=(-0.978413596402)*x[0]+(-1.20699880678)*x[1]+(-0.0600246637942)*x[2]
        s[2,0,0]=(0.637900823637)*x[0]+(1.29751687672)*x[1]+(-0.390955179963)*x[2]
        s[2,0,1]=(-1.15596339731)*x[0]+(-1.04131524409)*x[1]+(0.111161978383)*x[2]
        s[2,1,0]=(-1.33713877493)*x[0]+(0.0677540624633)*x[1]+(0.0662976219802)*x[2]
        s[2,1,1]=(0.0636656461261)*x[0]+(-1.32776475079)*x[1]+(0.365594993394)*x[2]
        s[3,0,0]=(0.454007484649)*x[0]+(-1.56273676656)*x[1]+(-0.783612553822)*x[2]
        s[3,0,1]=(-0.908388718727)*x[0]+(0.496552922533)*x[1]+(1.0802957835)*x[2]
        s[3,1,0]=(0.46784249418)*x[0]+(1.627686011)*x[1]+(1.12464788868)*x[2]
        s[3,1,1]=(1.47169416804)*x[0]+(-0.541043548439)*x[1]+(-0.0203908370923)*x[2]
        s[4,0,0]=(0.153593270409)*x[0]+(0.991271922294)*x[1]+(0.910688732568)*x[2]
        s[4,0,1]=(1.26666528018)*x[0]+(-0.0415946389155)*x[1]+(-0.702890723051)*x[2]
        s[4,1,0]=(0.401000008974)*x[0]+(0.270202736778)*x[1]+(-1.52334516556)*x[2]
        s[4,1,1]=(-0.567247526123)*x[0]+(0.0638484526058)*x[1]+(-0.942447437378)*x[2]
        s[5,0,0]=(1.3141482761)*x[0]+(0.602038495282)*x[1]+(-0.162794144191)*x[2]
        s[5,0,1]=(1.49343941166)*x[0]+(0.340106070028)*x[1]+(0.222963641278)*x[2]
        s[5,1,0]=(1.29779085951)*x[0]+(0.44436851782)*x[1]+(-0.414930735066)*x[2]
        s[5,1,1]=(0.445071462828)*x[0]+(0.0113100435192)*x[1]+(-0.304892738875)*x[2]
        ref[0,0,0]=(0.807154905774)*(1+2.*(dim-1.)/(o+1.))+(0.387214598576)*dim
        ref[0,0,1]=(-0.168668456531)*(1+2.*(dim-1.)/(o+1.))+(-0.985160135549)*dim
        ref[0,1,0]=(-0.405021627391)*(1+2.*(dim-1.)/(o+1.))+(0.69513474922)*dim
        ref[0,1,1]=(0.481047411067)*(1+2.*(dim-1.)/(o+1.))+(-0.714913158134)*dim
        ref[1,0,0]=(-0.241963228819)*(1+2.*(dim-1.)/(o+1.))+(-0.423996930987)*dim
        ref[1,0,1]=(-0.222173232945)*(1+2.*(dim-1.)/(o+1.))+(2.00647607317)*dim
        ref[1,1,0]=(2.08927091112)*(1+2.*(dim-1.)/(o+1.))+(-0.868322989801)*dim
        ref[1,1,1]=(-1.2199872069)*(1+2.*(dim-1.)/(o+1.))+(-1.02544986008)*dim
        ref[2,0,0]=(1.51358492234)*(1+2.*(dim-1.)/(o+1.))+(0.0308775980534)*dim
        ref[2,0,1]=(-1.17419262019)*(1+2.*(dim-1.)/(o+1.))+(-0.911924042825)*dim
        ref[2,1,0]=(-0.808827279389)*(1+2.*(dim-1.)/(o+1.))+(-0.394259811094)*dim
        ref[2,1,1]=(-0.39434462087)*(1+2.*(dim-1.)/(o+1.))+(-0.504159490404)*dim
        ref[3,0,0]=(-0.420509147061)*(1+2.*(dim-1.)/(o+1.))+(-1.47183268867)*dim
        ref[3,0,1]=(-0.509702954626)*(1+2.*(dim-1.)/(o+1.))+(1.17816294193)*dim
        ref[3,1,0]=(2.45580213911)*(1+2.*(dim-1.)/(o+1.))+(0.764374254754)*dim
        ref[3,1,1]=(0.36148972446)*(1+2.*(dim-1.)/(o+1.))+(0.548770058051)*dim
        ref[4,0,0]=(0.723463115757)*(1+2.*(dim-1.)/(o+1.))+(1.33209080951)*dim
        ref[4,0,1]=(-0.584660571656)*(1+2.*(dim-1.)/(o+1.))+(1.10684048987)*dim
        ref[4,1,0]=(0.558877069737)*(1+2.*(dim-1.)/(o+1.))+(-1.41101948955)*dim
        ref[4,1,1]=(-1.46224923696)*(1+2.*(dim-1.)/(o+1.))+(0.016402726067)*dim
        ref[5,0,0]=(0.375824940974)*(1+2.*(dim-1.)/(o+1.))+(1.37756768621)*dim
        ref[5,0,1]=(0.0790629051458)*(1+2.*(dim-1.)/(o+1.))+(1.97744621782)*dim
        ref[5,1,0]=(-0.429416156192)*(1+2.*(dim-1.)/(o+1.))+(1.75664479845)*dim
        ref[5,1,1]=(-0.227051313962)*(1+2.*(dim-1.)/(o+1.))+(0.378540081434)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Symbol on the FunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-1.24466004429)*x[0]+(0.249656583863)*x[1]
        s[0,0,0,1]=(-1.31303014286)*x[0]+(-0.188187818679)*x[1]
        s[0,0,1,0]=(-1.19686759063)*x[0]+(0.202998667098)*x[1]
        s[0,0,1,1]=(-0.240139422625)*x[0]+(1.12145727417)*x[1]
        s[0,0,2,0]=(-0.779228056931)*x[0]+(-0.256394958159)*x[1]
        s[0,0,2,1]=(0.511346899235)*x[0]+(-1.17117401759)*x[1]
        s[0,1,0,0]=(1.20931773874)*x[0]+(0.703848713877)*x[1]
        s[0,1,0,1]=(1.05396707864)*x[0]+(0.561081392967)*x[1]
        s[0,1,1,0]=(-0.940092412899)*x[0]+(0.197879835766)*x[1]
        s[0,1,1,1]=(0.031033640344)*x[0]+(1.47668227837)*x[1]
        s[0,1,2,0]=(-0.973745458182)*x[0]+(-0.699999591381)*x[1]
        s[0,1,2,1]=(0.0637513252131)*x[0]+(0.223369875142)*x[1]
        s[0,2,0,0]=(-0.179817828717)*x[0]+(-1.0618817379)*x[1]
        s[0,2,0,1]=(-1.31848741633)*x[0]+(0.562871230704)*x[1]
        s[0,2,1,0]=(0.20660993199)*x[0]+(-0.260401741099)*x[1]
        s[0,2,1,1]=(-0.119752579601)*x[0]+(1.03659426931)*x[1]
        s[0,2,2,0]=(-1.05257053872)*x[0]+(-0.410999368732)*x[1]
        s[0,2,2,1]=(0.142472752971)*x[0]+(0.687099054848)*x[1]
        s[0,3,0,0]=(-0.735958436285)*x[0]+(0.576103117069)*x[1]
        s[0,3,0,1]=(0.820759477791)*x[0]+(1.40201775771)*x[1]
        s[0,3,1,0]=(-1.43663163172)*x[0]+(-0.928562287497)*x[1]
        s[0,3,1,1]=(-0.640592604343)*x[0]+(-1.37951764035)*x[1]
        s[0,3,2,0]=(0.923170161037)*x[0]+(-1.29865735065)*x[1]
        s[0,3,2,1]=(-1.49185366667)*x[0]+(1.26235954659)*x[1]
        s[0,4,0,0]=(-1.54260726706)*x[0]+(0.386401532954)*x[1]
        s[0,4,0,1]=(0.938275027884)*x[0]+(-0.325978711858)*x[1]
        s[0,4,1,0]=(-0.741422170549)*x[0]+(0.360529257608)*x[1]
        s[0,4,1,1]=(-1.38867328824)*x[0]+(0.368979718783)*x[1]
        s[0,4,2,0]=(-1.19633228454)*x[0]+(0.143128822342)*x[1]
        s[0,4,2,1]=(0.365590220489)*x[0]+(0.702852400998)*x[1]
        s[1,0,0,0]=(0.590708718523)*x[0]+(-1.74588023915)*x[1]
        s[1,0,0,1]=(-0.411269052638)*x[0]+(0.677504850422)*x[1]
        s[1,0,1,0]=(1.54641105105)*x[0]+(-0.227798671895)*x[1]
        s[1,0,1,1]=(0.206986074762)*x[0]+(0.431911634342)*x[1]
        s[1,0,2,0]=(0.180245950747)*x[0]+(-0.766496827464)*x[1]
        s[1,0,2,1]=(0.920490611623)*x[0]+(0.146274601957)*x[1]
        s[1,1,0,0]=(-1.47489897228)*x[0]+(-1.0428187509)*x[1]
        s[1,1,0,1]=(-1.17774302127)*x[0]+(0.44418244557)*x[1]
        s[1,1,1,0]=(-0.340268939926)*x[0]+(-0.362308449805)*x[1]
        s[1,1,1,1]=(1.1122568876)*x[0]+(0.807887213986)*x[1]
        s[1,1,2,0]=(-1.1580670758)*x[0]+(0.61015974696)*x[1]
        s[1,1,2,1]=(-0.21738296838)*x[0]+(-1.15287881926)*x[1]
        s[1,2,0,0]=(-1.63285965164)*x[0]+(0.546571504115)*x[1]
        s[1,2,0,1]=(1.34627844359)*x[0]+(0.0153020013782)*x[1]
        s[1,2,1,0]=(-0.459380401764)*x[0]+(0.653971151372)*x[1]
        s[1,2,1,1]=(0.0918961369392)*x[0]+(0.631274394662)*x[1]
        s[1,2,2,0]=(-0.461548096839)*x[0]+(0.432322561401)*x[1]
        s[1,2,2,1]=(-0.741897062249)*x[0]+(-0.0397725689655)*x[1]
        s[1,3,0,0]=(-0.166826966339)*x[0]+(0.76157630088)*x[1]
        s[1,3,0,1]=(-1.43001526663)*x[0]+(0.0200445726731)*x[1]
        s[1,3,1,0]=(0.776640170933)*x[0]+(1.77522529504)*x[1]
        s[1,3,1,1]=(0.132999332599)*x[0]+(0.91396296202)*x[1]
        s[1,3,2,0]=(-0.435447499973)*x[0]+(-0.714956349315)*x[1]
        s[1,3,2,1]=(1.54444404505)*x[0]+(0.0419015202674)*x[1]
        s[1,4,0,0]=(-1.14197604256)*x[0]+(0.140549307511)*x[1]
        s[1,4,0,1]=(1.11080451267)*x[0]+(0.140105909725)*x[1]
        s[1,4,1,0]=(0.361572095315)*x[0]+(1.73141141151)*x[1]
        s[1,4,1,1]=(1.2968959883)*x[0]+(0.459397526479)*x[1]
        s[1,4,2,0]=(-1.04477229488)*x[0]+(-0.432251933613)*x[1]
        s[1,4,2,1]=(0.166352074158)*x[0]+(0.240538931271)*x[1]
        s[2,0,0,0]=(-0.302792258664)*x[0]+(0.014159412453)*x[1]
        s[2,0,0,1]=(0.161287860507)*x[0]+(-1.66699057701)*x[1]
        s[2,0,1,0]=(0.793079282708)*x[0]+(0.999374748613)*x[1]
        s[2,0,1,1]=(-0.0860624573943)*x[0]+(0.438101549788)*x[1]
        s[2,0,2,0]=(-0.832685279306)*x[0]+(1.20735467656)*x[1]
        s[2,0,2,1]=(-0.0628631179915)*x[0]+(0.720123606355)*x[1]
        s[2,1,0,0]=(0.275852566161)*x[0]+(-0.259595395227)*x[1]
        s[2,1,0,1]=(0.862309966294)*x[0]+(-0.45114296596)*x[1]
        s[2,1,1,0]=(0.14823103051)*x[0]+(-1.16121025546)*x[1]
        s[2,1,1,1]=(-1.34419311122)*x[0]+(1.63903760286)*x[1]
        s[2,1,2,0]=(1.18830622559)*x[0]+(0.985255273279)*x[1]
        s[2,1,2,1]=(-0.0208912272649)*x[0]+(0.543921272375)*x[1]
        s[2,2,0,0]=(0.510821411306)*x[0]+(-1.88592356399)*x[1]
        s[2,2,0,1]=(0.353203575328)*x[0]+(0.328922615441)*x[1]
        s[2,2,1,0]=(1.1536669592)*x[0]+(-1.05621607534)*x[1]
        s[2,2,1,1]=(-0.581456156893)*x[0]+(-0.20275776476)*x[1]
        s[2,2,2,0]=(1.23883996503)*x[0]+(0.577857478009)*x[1]
        s[2,2,2,1]=(0.938162394147)*x[0]+(1.74792706902)*x[1]
        s[2,3,0,0]=(-0.0610369156925)*x[0]+(0.316929103344)*x[1]
        s[2,3,0,1]=(0.115575825651)*x[0]+(0.196603816526)*x[1]
        s[2,3,1,0]=(-0.137564893251)*x[0]+(0.732741603189)*x[1]
        s[2,3,1,1]=(-0.659475613725)*x[0]+(0.126037736276)*x[1]
        s[2,3,2,0]=(1.1911196497)*x[0]+(0.86409088069)*x[1]
        s[2,3,2,1]=(-0.161576876539)*x[0]+(-0.519953247817)*x[1]
        s[2,4,0,0]=(0.653018736743)*x[0]+(-0.141460723253)*x[1]
        s[2,4,0,1]=(0.548072342723)*x[0]+(0.79847118191)*x[1]
        s[2,4,1,0]=(-0.894497169389)*x[0]+(0.358296451187)*x[1]
        s[2,4,1,1]=(0.104304124331)*x[0]+(0.229557340588)*x[1]
        s[2,4,2,0]=(-0.775504740819)*x[0]+(-0.914276983707)*x[1]
        s[2,4,2,1]=(-0.997444443952)*x[0]+(-0.550036262287)*x[1]
        s[3,0,0,0]=(-0.912968382437)*x[0]+(1.34668077004)*x[1]
        s[3,0,0,1]=(0.279466090703)*x[0]+(1.28107040706)*x[1]
        s[3,0,1,0]=(0.282808913079)*x[0]+(0.740167690692)*x[1]
        s[3,0,1,1]=(-1.39563519985)*x[0]+(-0.567913880569)*x[1]
        s[3,0,2,0]=(1.27572544936)*x[0]+(-0.733366778512)*x[1]
        s[3,0,2,1]=(-0.33162427954)*x[0]+(1.35270158006)*x[1]
        s[3,1,0,0]=(0.519174606536)*x[0]+(-0.372514377255)*x[1]
        s[3,1,0,1]=(-1.58036427028)*x[0]+(-0.235983616334)*x[1]
        s[3,1,1,0]=(-0.472835600722)*x[0]+(0.721357411442)*x[1]
        s[3,1,1,1]=(0.448313914421)*x[0]+(-0.138520459243)*x[1]
        s[3,1,2,0]=(-1.51018963552)*x[0]+(1.34139211207)*x[1]
        s[3,1,2,1]=(-0.212789438989)*x[0]+(-1.06575114519)*x[1]
        s[3,2,0,0]=(0.0592065335067)*x[0]+(0.359601543991)*x[1]
        s[3,2,0,1]=(-0.594769346579)*x[0]+(1.5491276729)*x[1]
        s[3,2,1,0]=(0.28455993713)*x[0]+(0.600972748948)*x[1]
        s[3,2,1,1]=(0.398577473206)*x[0]+(0.0479286078109)*x[1]
        s[3,2,2,0]=(-1.76127884396)*x[0]+(-0.538814316335)*x[1]
        s[3,2,2,1]=(-0.390841519342)*x[0]+(0.191259339024)*x[1]
        s[3,3,0,0]=(-0.877110189074)*x[0]+(0.301682318677)*x[1]
        s[3,3,0,1]=(0.112669606669)*x[0]+(1.0395312515)*x[1]
        s[3,3,1,0]=(1.04370247351)*x[0]+(0.581250871082)*x[1]
        s[3,3,1,1]=(-0.143233022608)*x[0]+(0.593143844478)*x[1]
        s[3,3,2,0]=(-0.304532308332)*x[0]+(-0.232727140662)*x[1]
        s[3,3,2,1]=(0.767141825929)*x[0]+(0.0401997820135)*x[1]
        s[3,4,0,0]=(-0.226100649415)*x[0]+(-0.656022574403)*x[1]
        s[3,4,0,1]=(0.679160670663)*x[0]+(0.507720591962)*x[1]
        s[3,4,1,0]=(-1.30886998647)*x[0]+(-1.3539695469)*x[1]
        s[3,4,1,1]=(-1.39638819903)*x[0]+(0.650428829094)*x[1]
        s[3,4,2,0]=(-0.346486286718)*x[0]+(-0.869112300588)*x[1]
        s[3,4,2,1]=(-1.27650086562)*x[0]+(0.660506644039)*x[1]
        ref[0,0,0,0]=(-0.625690911818)*(1+2.*(dim-1.)/(o+1.))+(-0.369312548605)*dim
        ref[0,0,0,1]=(-0.0186378038077)*(1+2.*(dim-1.)/(o+1.))+(-1.48258015773)*dim
        ref[0,0,1,0]=(-1.04194152975)*(1+2.*(dim-1.)/(o+1.))+(0.0480726062254)*dim
        ref[0,0,1,1]=(0.00120841586244)*(1+2.*(dim-1.)/(o+1.))+(0.880109435685)*dim
        ref[0,0,2,0]=(-0.65182866689)*(1+2.*(dim-1.)/(o+1.))+(-0.3837943482)*dim
        ref[0,0,2,1]=(-0.729868010003)*(1+2.*(dim-1.)/(o+1.))+(0.0700408916461)*dim
        ref[0,1,0,0]=(1.26484534964)*(1+2.*(dim-1.)/(o+1.))+(0.648321102977)*dim
        ref[0,1,0,1]=(0.334654674402)*(1+2.*(dim-1.)/(o+1.))+(1.28039379721)*dim
        ref[0,1,1,0]=(0.338629227563)*(1+2.*(dim-1.)/(o+1.))+(-1.0808418047)*dim
        ref[0,1,1,1]=(0.0969508871625)*(1+2.*(dim-1.)/(o+1.))+(1.41076503155)*dim
        ref[0,1,2,0]=(-0.462969154119)*(1+2.*(dim-1.)/(o+1.))+(-1.21077589544)*dim
        ref[0,1,2,1]=(0.46756493995)*(1+2.*(dim-1.)/(o+1.))+(-0.180443739595)*dim
        ref[0,2,0,0]=(-0.0576220389279)*(1+2.*(dim-1.)/(o+1.))+(-1.18407752769)*dim
        ref[0,2,0,1]=(-0.650282644944)*(1+2.*(dim-1.)/(o+1.))+(-0.105333540684)*dim
        ref[0,2,1,0]=(0.136853100987)*(1+2.*(dim-1.)/(o+1.))+(-0.190644910096)*dim
        ref[0,2,1,1]=(-0.637797927534)*(1+2.*(dim-1.)/(o+1.))+(1.55463961724)*dim
        ref[0,2,2,0]=(-0.455136385489)*(1+2.*(dim-1.)/(o+1.))+(-1.00843352196)*dim
        ref[0,2,2,1]=(-0.35303452486)*(1+2.*(dim-1.)/(o+1.))+(1.18260633268)*dim
        ref[0,3,0,0]=(-0.315687014208)*(1+2.*(dim-1.)/(o+1.))+(0.155831694992)*dim
        ref[0,3,0,1]=(1.52158435346)*(1+2.*(dim-1.)/(o+1.))+(0.701192882048)*dim
        ref[0,3,1,0]=(-0.836739848177)*(1+2.*(dim-1.)/(o+1.))+(-1.52845407104)*dim
        ref[0,3,1,1]=(-0.983041539851)*(1+2.*(dim-1.)/(o+1.))+(-1.03706870484)*dim
        ref[0,3,2,0]=(-0.497690471912)*(1+2.*(dim-1.)/(o+1.))+(0.122203282296)*dim
        ref[0,3,2,1]=(0.394951943207)*(1+2.*(dim-1.)/(o+1.))+(-0.624446063287)*dim
        ref[0,4,0,0]=(-0.63974792726)*(1+2.*(dim-1.)/(o+1.))+(-0.516457806843)*dim
        ref[0,4,0,1]=(0.62556865655)*(1+2.*(dim-1.)/(o+1.))+(-0.0132723405234)*dim
        ref[0,4,1,0]=(-1.10166644693)*(1+2.*(dim-1.)/(o+1.))+(0.720773533984)*dim
        ref[0,4,1,1]=(-1.09945957099)*(1+2.*(dim-1.)/(o+1.))+(0.0797660015265)*dim
        ref[0,4,2,0]=(-0.359198665738)*(1+2.*(dim-1.)/(o+1.))+(-0.694004796462)*dim
        ref[0,4,2,1]=(1.60203342245)*(1+2.*(dim-1.)/(o+1.))+(-0.533590800966)*dim
        ref[1,0,0,0]=(0.0630998548187)*(1+2.*(dim-1.)/(o+1.))+(-1.21827137545)*dim
        ref[1,0,0,1]=(-0.362786036469)*(1+2.*(dim-1.)/(o+1.))+(0.629021834252)*dim
        ref[1,0,1,0]=(1.30840533269)*(1+2.*(dim-1.)/(o+1.))+(0.01020704647)*dim
        ref[1,0,1,1]=(0.581668517056)*(1+2.*(dim-1.)/(o+1.))+(0.0572291920482)*dim
        ref[1,0,2,0]=(-1.1596262045)*(1+2.*(dim-1.)/(o+1.))+(0.573375327782)*dim
        ref[1,0,2,1]=(-0.025337463973)*(1+2.*(dim-1.)/(o+1.))+(1.09210267755)*dim
        ref[1,1,0,0]=(-1.46016674214)*(1+2.*(dim-1.)/(o+1.))+(-1.05755098104)*dim
        ref[1,1,0,1]=(-0.410079206814)*(1+2.*(dim-1.)/(o+1.))+(-0.323481368888)*dim
        ref[1,1,1,0]=(0.497609154803)*(1+2.*(dim-1.)/(o+1.))+(-1.20018654453)*dim
        ref[1,1,1,1]=(0.446563754938)*(1+2.*(dim-1.)/(o+1.))+(1.47358034665)*dim
        ref[1,1,2,0]=(0.124142572295)*(1+2.*(dim-1.)/(o+1.))+(-0.672049901132)*dim
        ref[1,1,2,1]=(-1.27946259076)*(1+2.*(dim-1.)/(o+1.))+(-0.09079919688)*dim
        ref[1,2,0,0]=(0.0167916382593)*(1+2.*(dim-1.)/(o+1.))+(-1.10307978578)*dim
        ref[1,2,0,1]=(0.310300272224)*(1+2.*(dim-1.)/(o+1.))+(1.05128017274)*dim
        ref[1,2,1,0]=(-0.384485455128)*(1+2.*(dim-1.)/(o+1.))+(0.579076204736)*dim
        ref[1,2,1,1]=(0.133076827561)*(1+2.*(dim-1.)/(o+1.))+(0.59009370404)*dim
        ref[1,2,2,0]=(1.4126992054)*(1+2.*(dim-1.)/(o+1.))+(-1.44192474084)*dim
        ref[1,2,2,1]=(0.235736468845)*(1+2.*(dim-1.)/(o+1.))+(-1.01740610006)*dim
        ref[1,3,0,0]=(-0.323242689323)*(1+2.*(dim-1.)/(o+1.))+(0.917992023864)*dim
        ref[1,3,0,1]=(-1.56677140485)*(1+2.*(dim-1.)/(o+1.))+(0.156800710897)*dim
        ref[1,3,1,0]=(1.39852204704)*(1+2.*(dim-1.)/(o+1.))+(1.15334341893)*dim
        ref[1,3,1,1]=(1.10144038868)*(1+2.*(dim-1.)/(o+1.))+(-0.0544780940565)*dim
        ref[1,3,2,0]=(-0.619896071163)*(1+2.*(dim-1.)/(o+1.))+(-0.530507778126)*dim
        ref[1,3,2,1]=(0.665968335878)*(1+2.*(dim-1.)/(o+1.))+(0.920377229442)*dim
        ref[1,4,0,0]=(0.24940249757)*(1+2.*(dim-1.)/(o+1.))+(-1.25082923262)*dim
        ref[1,4,0,1]=(1.26616311307)*(1+2.*(dim-1.)/(o+1.))+(-0.0152526906804)*dim
        ref[1,4,1,0]=(1.37246903375)*(1+2.*(dim-1.)/(o+1.))+(0.720514473076)*dim
        ref[1,4,1,1]=(1.20885120083)*(1+2.*(dim-1.)/(o+1.))+(0.547442313947)*dim
        ref[1,4,2,0]=(-0.929784649867)*(1+2.*(dim-1.)/(o+1.))+(-0.547239578625)*dim
        ref[1,4,2,1]=(1.40793646749)*(1+2.*(dim-1.)/(o+1.))+(-1.00104546206)*dim
        ref[2,0,0,0]=(-0.865664592293)*(1+2.*(dim-1.)/(o+1.))+(0.577031746082)*dim
        ref[2,0,0,1]=(-0.383744932135)*(1+2.*(dim-1.)/(o+1.))+(-1.12195778437)*dim
        ref[2,0,1,0]=(0.899513902212)*(1+2.*(dim-1.)/(o+1.))+(0.892940129109)*dim
        ref[2,0,1,1]=(0.422487727431)*(1+2.*(dim-1.)/(o+1.))+(-0.070448635037)*dim
        ref[2,0,2,0]=(-0.0542314325807)*(1+2.*(dim-1.)/(o+1.))+(0.428900829837)*dim
        ref[2,0,2,1]=(-0.333651152721)*(1+2.*(dim-1.)/(o+1.))+(0.990911641084)*dim
        ref[2,1,0,0]=(0.743812676627)*(1+2.*(dim-1.)/(o+1.))+(-0.727555505694)*dim
        ref[2,1,0,1]=(0.487164931518)*(1+2.*(dim-1.)/(o+1.))+(-0.0759979311835)*dim
        ref[2,1,1,0]=(0.146918316537)*(1+2.*(dim-1.)/(o+1.))+(-1.15989754148)*dim
        ref[2,1,1,1]=(-0.239925787022)*(1+2.*(dim-1.)/(o+1.))+(0.534770278668)*dim
        ref[2,1,2,0]=(1.79635305588)*(1+2.*(dim-1.)/(o+1.))+(0.377208442988)*dim
        ref[2,1,2,1]=(-0.275396782095)*(1+2.*(dim-1.)/(o+1.))+(0.798426827206)*dim
        ref[2,2,0,0]=(-0.732792630073)*(1+2.*(dim-1.)/(o+1.))+(-0.642309522614)*dim
        ref[2,2,0,1]=(0.691994610633)*(1+2.*(dim-1.)/(o+1.))+(-0.0098684198643)*dim
        ref[2,2,1,0]=(0.571059981052)*(1+2.*(dim-1.)/(o+1.))+(-0.473609097192)*dim
        ref[2,2,1,1]=(-0.935084534142)*(1+2.*(dim-1.)/(o+1.))+(0.15087061249)*dim
        ref[2,2,2,0]=(0.0826567188857)*(1+2.*(dim-1.)/(o+1.))+(1.73404072415)*dim
        ref[2,2,2,1]=(1.26427004187)*(1+2.*(dim-1.)/(o+1.))+(1.4218194213)*dim
        ref[2,3,0,0]=(-0.398591309648)*(1+2.*(dim-1.)/(o+1.))+(0.6544834973)*dim
        ref[2,3,0,1]=(-0.568822853648)*(1+2.*(dim-1.)/(o+1.))+(0.881002495825)*dim
        ref[2,3,1,0]=(1.18000532793)*(1+2.*(dim-1.)/(o+1.))+(-0.584828617992)*dim
        ref[2,3,1,1]=(-1.60821457294)*(1+2.*(dim-1.)/(o+1.))+(1.07477669549)*dim
        ref[2,3,2,0]=(1.48063830505)*(1+2.*(dim-1.)/(o+1.))+(0.574572225346)*dim
        ref[2,3,2,1]=(0.229132618257)*(1+2.*(dim-1.)/(o+1.))+(-0.910662742613)*dim
        ref[2,4,0,0]=(0.426313638958)*(1+2.*(dim-1.)/(o+1.))+(0.0852443745317)*dim
        ref[2,4,0,1]=(-0.0761298452145)*(1+2.*(dim-1.)/(o+1.))+(1.42267336985)*dim
        ref[2,4,1,0]=(-0.537516881228)*(1+2.*(dim-1.)/(o+1.))+(0.00131616302647)*dim
        ref[2,4,1,1]=(-1.00615029385)*(1+2.*(dim-1.)/(o+1.))+(1.34001175877)*dim
        ref[2,4,2,0]=(-0.250744474716)*(1+2.*(dim-1.)/(o+1.))+(-1.43903724981)*dim
        ref[2,4,2,1]=(0.273616829308)*(1+2.*(dim-1.)/(o+1.))+(-1.82109753555)*dim
        ref[3,0,0,0]=(0.773912240208)*(1+2.*(dim-1.)/(o+1.))+(-0.340199852604)*dim
        ref[3,0,0,1]=(0.00495773386215)*(1+2.*(dim-1.)/(o+1.))+(1.5555787639)*dim
        ref[3,0,1,0]=(-0.586757469166)*(1+2.*(dim-1.)/(o+1.))+(1.60973407294)*dim
        ref[3,0,1,1]=(-0.577120453982)*(1+2.*(dim-1.)/(o+1.))+(-1.38642862644)*dim
        ref[3,0,2,0]=(0.288712056159)*(1+2.*(dim-1.)/(o+1.))+(0.253646614684)*dim
        ref[3,0,2,1]=(-0.248042635081)*(1+2.*(dim-1.)/(o+1.))+(1.2691199356)*dim
        ref[3,1,0,0]=(0.995572341947)*(1+2.*(dim-1.)/(o+1.))+(-0.848912112665)*dim
        ref[3,1,0,1]=(-1.37225234108)*(1+2.*(dim-1.)/(o+1.))+(-0.444095545534)*dim
        ref[3,1,1,0]=(0.441843238326)*(1+2.*(dim-1.)/(o+1.))+(-0.193321427606)*dim
        ref[3,1,1,1]=(0.408346053376)*(1+2.*(dim-1.)/(o+1.))+(-0.098552598199)*dim
        ref[3,1,2,0]=(-0.311970472748)*(1+2.*(dim-1.)/(o+1.))+(0.143172949307)*dim
        ref[3,1,2,1]=(-0.434499269883)*(1+2.*(dim-1.)/(o+1.))+(-0.844041314296)*dim
        ref[3,2,0,0]=(0.480046851902)*(1+2.*(dim-1.)/(o+1.))+(-0.0612387744048)*dim
        ref[3,2,0,1]=(1.12223817367)*(1+2.*(dim-1.)/(o+1.))+(-0.167879847351)*dim
        ref[3,2,1,0]=(-0.345095098378)*(1+2.*(dim-1.)/(o+1.))+(1.23062778446)*dim
        ref[3,2,1,1]=(1.37605147183)*(1+2.*(dim-1.)/(o+1.))+(-0.92954539081)*dim
        ref[3,2,2,0]=(-1.0202068094)*(1+2.*(dim-1.)/(o+1.))+(-1.2798863509)*dim
        ref[3,2,2,1]=(-0.302591270785)*(1+2.*(dim-1.)/(o+1.))+(0.103009090467)*dim
        ref[3,3,0,0]=(0.857177782444)*(1+2.*(dim-1.)/(o+1.))+(-1.43260565284)*dim
        ref[3,3,0,1]=(0.0906238774103)*(1+2.*(dim-1.)/(o+1.))+(1.06157698076)*dim
        ref[3,3,1,0]=(0.42136870665)*(1+2.*(dim-1.)/(o+1.))+(1.20358463794)*dim
        ref[3,3,1,1]=(-0.88620181924)*(1+2.*(dim-1.)/(o+1.))+(1.33611264111)*dim
        ref[3,3,2,0]=(0.444491342755)*(1+2.*(dim-1.)/(o+1.))+(-0.981750791749)*dim
        ref[3,3,2,1]=(-1.06075000108)*(1+2.*(dim-1.)/(o+1.))+(1.86809160902)*dim
        ref[3,4,0,0]=(-0.517172055625)*(1+2.*(dim-1.)/(o+1.))+(-0.364951168193)*dim
        ref[3,4,0,1]=(0.683410864606)*(1+2.*(dim-1.)/(o+1.))+(0.503470398019)*dim
        ref[3,4,1,0]=(-1.49890766435)*(1+2.*(dim-1.)/(o+1.))+(-1.16393186901)*dim
        ref[3,4,1,1]=(0.25999658315)*(1+2.*(dim-1.)/(o+1.))+(-1.00595595308)*dim
        ref[3,4,2,0]=(-0.296406951133)*(1+2.*(dim-1.)/(o+1.))+(-0.919191636173)*dim
        ref[3,4,2,1]=(-0.562797436325)*(1+2.*(dim-1.)/(o+1.))+(-0.0531967852595)*dim
      else:
        s[0,0,0,0]=(0.28520634077)*x[0]+(-0.126741433974)*x[1]+(1.28370871197)*x[2]
        s[0,0,0,1]=(-0.0731693108841)*x[0]+(0.342123624879)*x[1]+(0.302260810729)*x[2]
        s[0,0,1,0]=(-0.933375161617)*x[0]+(0.714257115938)*x[1]+(0.0355044702732)*x[2]
        s[0,0,1,1]=(0.273422423495)*x[0]+(-1.43381666098)*x[1]+(-0.3924531625)*x[2]
        s[0,0,2,0]=(-1.24504380748)*x[0]+(1.36389469156)*x[1]+(1.0146102387)*x[2]
        s[0,0,2,1]=(-0.12931890066)*x[0]+(0.00848180915183)*x[1]+(0.230392813094)*x[2]
        s[0,1,0,0]=(0.390247630566)*x[0]+(0.116544846708)*x[1]+(-0.94552924973)*x[2]
        s[0,1,0,1]=(1.53653205813)*x[0]+(0.984195773879)*x[1]+(0.767182801059)*x[2]
        s[0,1,1,0]=(0.433068815739)*x[0]+(0.196964560742)*x[1]+(0.00150259046749)*x[2]
        s[0,1,1,1]=(0.620083290721)*x[0]+(-0.981816668862)*x[1]+(-0.0602017060991)*x[2]
        s[0,1,2,0]=(0.266920335934)*x[0]+(-1.01077201758)*x[1]+(-1.66555517355)*x[2]
        s[0,1,2,1]=(-0.463794455357)*x[0]+(1.02905222514)*x[1]+(-0.524965351287)*x[2]
        s[0,2,0,0]=(-1.02890742598)*x[0]+(0.10277767711)*x[1]+(-0.709831378433)*x[2]
        s[0,2,0,1]=(0.942130530664)*x[0]+(-0.498478792456)*x[1]+(-0.321634750428)*x[2]
        s[0,2,1,0]=(-1.09979499064)*x[0]+(-0.552242175008)*x[1]+(1.47155591555)*x[2]
        s[0,2,1,1]=(0.038643175545)*x[0]+(-0.367877010362)*x[1]+(-0.422403338025)*x[2]
        s[0,2,2,0]=(0.915798473155)*x[0]+(0.627559693912)*x[1]+(-0.146957763577)*x[2]
        s[0,2,2,1]=(0.35609430038)*x[0]+(0.627155969018)*x[1]+(-0.0538976009369)*x[2]
        s[0,3,0,0]=(0.753367589685)*x[0]+(-1.56174987227)*x[1]+(0.0317386362973)*x[2]
        s[0,3,0,1]=(-1.0814768856)*x[0]+(-0.000317117094536)*x[1]+(0.160394230281)*x[2]
        s[0,3,1,0]=(0.539343002581)*x[0]+(0.600975350875)*x[1]+(-0.579108274756)*x[2]
        s[0,3,1,1]=(1.93822764773)*x[0]+(1.19115168218)*x[1]+(-0.345303177833)*x[2]
        s[0,3,2,0]=(-0.396284929409)*x[0]+(-0.0620671301539)*x[1]+(0.405978876833)*x[2]
        s[0,3,2,1]=(-0.806591193034)*x[0]+(-0.869878106357)*x[1]+(-0.352772882375)*x[2]
        s[0,4,0,0]=(-1.01176424875)*x[0]+(-1.57589661925)*x[1]+(-0.313904557434)*x[2]
        s[0,4,0,1]=(0.277176183617)*x[0]+(0.548456592065)*x[1]+(1.21233269169)*x[2]
        s[0,4,1,0]=(0.916512281025)*x[0]+(-1.53600125673)*x[1]+(-0.486397251593)*x[2]
        s[0,4,1,1]=(1.12553562564)*x[0]+(0.0569780615281)*x[1]+(-0.275491563277)*x[2]
        s[0,4,2,0]=(0.0701240897706)*x[0]+(-0.109284348724)*x[1]+(0.920255378883)*x[2]
        s[0,4,2,1]=(-0.286723671632)*x[0]+(0.521149269283)*x[1]+(0.880566081072)*x[2]
        s[1,0,0,0]=(-0.161092153994)*x[0]+(1.60646814731)*x[1]+(-0.293423133436)*x[2]
        s[1,0,0,1]=(-0.62705410926)*x[0]+(-0.133468132127)*x[1]+(0.435110047414)*x[2]
        s[1,0,1,0]=(0.963863560937)*x[0]+(0.358847992997)*x[1]+(0.847184531597)*x[2]
        s[1,0,1,1]=(0.2445787465)*x[0]+(-0.161812641928)*x[1]+(-0.679720454567)*x[2]
        s[1,0,2,0]=(1.05646386542)*x[0]+(1.0631242639)*x[1]+(-0.762052372247)*x[2]
        s[1,0,2,1]=(-0.950180341823)*x[0]+(1.00890659836)*x[1]+(1.54399946028)*x[2]
        s[1,1,0,0]=(0.290340455371)*x[0]+(-1.13919380812)*x[1]+(-0.659408495899)*x[2]
        s[1,1,0,1]=(1.4258405346)*x[0]+(1.53809682182)*x[1]+(0.0675302762314)*x[2]
        s[1,1,1,0]=(1.21278717612)*x[0]+(0.220266219454)*x[1]+(1.66873955725)*x[2]
        s[1,1,1,1]=(0.771526101684)*x[0]+(-0.067360487854)*x[1]+(1.21982979968)*x[2]
        s[1,1,2,0]=(-0.77604682892)*x[0]+(0.3133384598)*x[1]+(0.645828815985)*x[2]
        s[1,1,2,1]=(-0.107715380402)*x[0]+(-0.0772190829216)*x[1]+(-1.14700080403)*x[2]
        s[1,2,0,0]=(0.324137135934)*x[0]+(1.29562077427)*x[1]+(0.455108897284)*x[2]
        s[1,2,0,1]=(-1.43262429921)*x[0]+(0.773443491708)*x[1]+(0.277297920728)*x[2]
        s[1,2,1,0]=(0.304810473045)*x[0]+(-0.0168967979423)*x[1]+(0.131774697941)*x[2]
        s[1,2,1,1]=(0.387736648788)*x[0]+(0.666220386408)*x[1]+(1.07576988895)*x[2]
        s[1,2,2,0]=(-0.254693787451)*x[0]+(-0.3384585134)*x[1]+(-0.64060329677)*x[2]
        s[1,2,2,1]=(-0.672597816893)*x[0]+(0.145646769601)*x[1]+(-0.994938729226)*x[2]
        s[1,3,0,0]=(-1.1030634788)*x[0]+(-1.80762204727)*x[1]+(1.6309434404)*x[2]
        s[1,3,0,1]=(0.276318339362)*x[0]+(0.385250530079)*x[1]+(-0.758561170758)*x[2]
        s[1,3,1,0]=(0.179910601577)*x[0]+(-0.11818800924)*x[1]+(-0.743123335945)*x[2]
        s[1,3,1,1]=(0.782295683197)*x[0]+(0.749420799157)*x[1]+(-0.656475021802)*x[2]
        s[1,3,2,0]=(0.574829069465)*x[0]+(-0.493394135291)*x[1]+(1.29795947888)*x[2]
        s[1,3,2,1]=(-1.46888017959)*x[0]+(-1.08679474255)*x[1]+(0.156158770501)*x[2]
        s[1,4,0,0]=(-0.860789510784)*x[0]+(0.652130849161)*x[1]+(-0.0894965081242)*x[2]
        s[1,4,0,1]=(0.997352754709)*x[0]+(-0.112776085202)*x[1]+(-0.469141207634)*x[2]
        s[1,4,1,0]=(-0.401561591069)*x[0]+(-0.362165453433)*x[1]+(1.42046539935)*x[2]
        s[1,4,1,1]=(-0.668331880273)*x[0]+(-1.28337116108)*x[1]+(0.368963183759)*x[2]
        s[1,4,2,0]=(-1.15039637411)*x[0]+(0.121579045201)*x[1]+(-0.104055094935)*x[2]
        s[1,4,2,1]=(-0.923080753122)*x[0]+(0.322395936762)*x[1]+(0.233695594483)*x[2]
        s[2,0,0,0]=(-0.206424413737)*x[0]+(0.761779799677)*x[1]+(-0.301048388352)*x[2]
        s[2,0,0,1]=(1.55990080913)*x[0]+(-0.162113046152)*x[1]+(-0.310428209804)*x[2]
        s[2,0,1,0]=(-0.433285639024)*x[0]+(1.36777883595)*x[1]+(-1.17795591805)*x[2]
        s[2,0,1,1]=(1.34284608147)*x[0]+(-0.222861276203)*x[1]+(0.855486771517)*x[2]
        s[2,0,2,0]=(-0.16635189023)*x[0]+(-0.823533654953)*x[1]+(0.797079690281)*x[2]
        s[2,0,2,1]=(0.752310816499)*x[0]+(0.764018784752)*x[1]+(0.056117413692)*x[2]
        s[2,1,0,0]=(0.00166988723156)*x[0]+(0.766304091963)*x[1]+(0.275940662737)*x[2]
        s[2,1,0,1]=(-0.0855732564349)*x[0]+(-1.02664418473)*x[1]+(0.662522947755)*x[2]
        s[2,1,1,0]=(-0.658811588292)*x[0]+(-1.04635899154)*x[1]+(1.05470344529)*x[2]
        s[2,1,1,1]=(-0.119299788842)*x[0]+(-0.782372280168)*x[1]+(-1.79380761901)*x[2]
        s[2,1,2,0]=(1.71887176677)*x[0]+(-0.226946752041)*x[1]+(0.799181606184)*x[2]
        s[2,1,2,1]=(-0.9189461833)*x[0]+(0.916498349691)*x[1]+(0.734617496542)*x[2]
        s[2,2,0,0]=(-0.790364198949)*x[0]+(0.0131553082942)*x[1]+(-0.978651161192)*x[2]
        s[2,2,0,1]=(0.32388442717)*x[0]+(-0.593416422259)*x[1]+(1.10510332946)*x[2]
        s[2,2,1,0]=(0.452042167919)*x[0]+(-0.641462816719)*x[1]+(-0.356784432003)*x[2]
        s[2,2,1,1]=(0.0725380542114)*x[0]+(0.255256127518)*x[1]+(0.979834678019)*x[2]
        s[2,2,2,0]=(1.39372044596)*x[0]+(0.938441851122)*x[1]+(-0.983490819141)*x[2]
        s[2,2,2,1]=(1.2364973033)*x[0]+(-1.2815807633)*x[1]+(-0.360311591593)*x[2]
        s[2,3,0,0]=(-0.0597199849648)*x[0]+(0.376596450311)*x[1]+(-0.494756059286)*x[2]
        s[2,3,0,1]=(-0.1287394847)*x[0]+(0.0455819025025)*x[1]+(1.57146607969)*x[2]
        s[2,3,1,0]=(-1.4907853501)*x[0]+(0.0388186733561)*x[1]+(0.13920185582)*x[2]
        s[2,3,1,1]=(0.0439467610482)*x[0]+(1.43515095249)*x[1]+(-0.205189829549)*x[2]
        s[2,3,2,0]=(0.773581921294)*x[0]+(1.41482523023)*x[1]+(1.02435393168)*x[2]
        s[2,3,2,1]=(0.347805074163)*x[0]+(1.52992663335)*x[1]+(-0.487058346391)*x[2]
        s[2,4,0,0]=(-0.107692405269)*x[0]+(0.147562810855)*x[1]+(1.38077195652)*x[2]
        s[2,4,0,1]=(-0.0410323036269)*x[0]+(0.269432309711)*x[1]+(-0.693128417186)*x[2]
        s[2,4,1,0]=(-0.33861862665)*x[0]+(-0.419772907291)*x[1]+(-1.07238876285)*x[2]
        s[2,4,1,1]=(0.143346805317)*x[0]+(-0.419444960688)*x[1]+(1.59692612607)*x[2]
        s[2,4,2,0]=(0.654582963584)*x[0]+(-0.44163597128)*x[1]+(-0.694330043907)*x[2]
        s[2,4,2,1]=(0.627421193857)*x[0]+(0.597177617998)*x[1]+(-0.451180510793)*x[2]
        s[3,0,0,0]=(-0.101412872135)*x[0]+(-0.396884673414)*x[1]+(0.896841368992)*x[2]
        s[3,0,0,1]=(0.139595535793)*x[0]+(0.624607944421)*x[1]+(0.574170381258)*x[2]
        s[3,0,1,0]=(0.316717544294)*x[0]+(-0.621261709901)*x[1]+(1.31441803912)*x[2]
        s[3,0,1,1]=(0.254418231071)*x[0]+(0.228843906675)*x[1]+(0.25896007767)*x[2]
        s[3,0,2,0]=(-0.247274104613)*x[0]+(1.20694307591)*x[1]+(-0.407808429936)*x[2]
        s[3,0,2,1]=(-1.40605558298)*x[0]+(0.168721449606)*x[1]+(-0.972867940483)*x[2]
        s[3,1,0,0]=(0.245845223266)*x[0]+(-1.51779122633)*x[1]+(-0.779293963961)*x[2]
        s[3,1,0,1]=(0.0325495798223)*x[0]+(-0.12100728517)*x[1]+(1.31848044222)*x[2]
        s[3,1,1,0]=(0.789118948962)*x[0]+(0.589652419254)*x[1]+(-0.08688557206)*x[2]
        s[3,1,1,1]=(-1.83518748403)*x[0]+(-0.500646680661)*x[1]+(0.233750065377)*x[2]
        s[3,1,2,0]=(-0.0771120817018)*x[0]+(0.0904104042813)*x[1]+(0.0933370539969)*x[2]
        s[3,1,2,1]=(-1.19249513127)*x[0]+(0.447348983457)*x[1]+(-0.298916535972)*x[2]
        s[3,2,0,0]=(1.22266670435)*x[0]+(-1.62277644197)*x[1]+(-0.0190709988254)*x[2]
        s[3,2,0,1]=(-0.194952675382)*x[0]+(0.554694000692)*x[1]+(-0.151982113386)*x[2]
        s[3,2,1,0]=(0.511265692083)*x[0]+(-0.802622720144)*x[1]+(-0.242270723374)*x[2]
        s[3,2,1,1]=(0.476021603798)*x[0]+(-0.463494861878)*x[1]+(-0.532143290096)*x[2]
        s[3,2,2,0]=(-0.530708374495)*x[0]+(1.10856138272)*x[1]+(0.409512895954)*x[2]
        s[3,2,2,1]=(0.0214256235243)*x[0]+(0.836046354994)*x[1]+(-0.263352555948)*x[2]
        s[3,3,0,0]=(-0.437181323979)*x[0]+(-0.413716617562)*x[1]+(-0.432122540243)*x[2]
        s[3,3,0,1]=(1.08881737048)*x[0]+(-0.0166529943951)*x[1]+(-0.236235303435)*x[2]
        s[3,3,1,0]=(-0.220528005238)*x[0]+(-0.0637133108381)*x[1]+(-0.722670710646)*x[2]
        s[3,3,1,1]=(0.0267596690177)*x[0]+(0.277583150215)*x[1]+(-1.19820417967)*x[2]
        s[3,3,2,0]=(-0.134456240167)*x[0]+(0.207594364333)*x[1]+(-1.58153388102)*x[2]
        s[3,3,2,1]=(-0.484174332477)*x[0]+(1.24683115653)*x[1]+(0.0960608606087)*x[2]
        s[3,4,0,0]=(-1.06280143452)*x[0]+(0.0909193631754)*x[1]+(0.735785607027)*x[2]
        s[3,4,0,1]=(-0.39199782699)*x[0]+(-0.547457599272)*x[1]+(-0.990008593852)*x[2]
        s[3,4,1,0]=(-0.301784264736)*x[0]+(-0.219227445543)*x[1]+(-0.979494077205)*x[2]
        s[3,4,1,1]=(-0.898860901012)*x[0]+(-0.103378912588)*x[1]+(-0.66619898802)*x[2]
        s[3,4,2,0]=(1.10051932868)*x[0]+(0.407341781092)*x[1]+(-0.680512930996)*x[2]
        s[3,4,2,1]=(-0.0733544592857)*x[0]+(1.18707229675)*x[1]+(-0.777693466919)*x[2]
        ref[0,0,0,0]=(0.438851859071)*(1+2.*(dim-1.)/(o+1.))+(1.0033217597)*dim
        ref[0,0,0,1]=(-0.891607994637)*(1+2.*(dim-1.)/(o+1.))+(1.46282311936)*dim
        ref[0,0,1,0]=(0.817888583209)*(1+2.*(dim-1.)/(o+1.))+(-1.00150215862)*dim
        ref[0,0,1,1]=(-1.639070793)*(1+2.*(dim-1.)/(o+1.))+(0.0862233930162)*dim
        ref[0,0,2,0]=(-0.348875799642)*(1+2.*(dim-1.)/(o+1.))+(1.48233692242)*dim
        ref[0,0,2,1]=(-0.589821443428)*(1+2.*(dim-1.)/(o+1.))+(0.699377165015)*dim
        ref[0,1,0,0]=(-0.832218783182)*(1+2.*(dim-1.)/(o+1.))+(0.393482010727)*dim
        ref[0,1,0,1]=(2.07094909251)*(1+2.*(dim-1.)/(o+1.))+(1.21696154056)*dim
        ref[0,1,1,0]=(0.1125989301)*(1+2.*(dim-1.)/(o+1.))+(0.518937036848)*dim
        ref[0,1,1,1]=(-1.36685685946)*(1+2.*(dim-1.)/(o+1.))+(0.944921775217)*dim
        ref[0,1,2,0]=(-1.47817293944)*(1+2.*(dim-1.)/(o+1.))+(-0.931233915749)*dim
        ref[0,1,2,1]=(-0.213405460956)*(1+2.*(dim-1.)/(o+1.))+(0.253697879455)*dim
        ref[0,2,0,0]=(-0.0805522984701)*(1+2.*(dim-1.)/(o+1.))+(-1.55540882883)*dim
        ref[0,2,0,1]=(0.156888902683)*(1+2.*(dim-1.)/(o+1.))+(-0.034871914902)*dim
        ref[0,2,1,0]=(-0.15684567608)*(1+2.*(dim-1.)/(o+1.))+(-0.0236355740175)*dim
        ref[0,2,1,1]=(-1.39067500442)*(1+2.*(dim-1.)/(o+1.))+(0.639037831581)*dim
        ref[0,2,2,0]=(1.81399897277)*(1+2.*(dim-1.)/(o+1.))+(-0.417598569281)*dim
        ref[0,2,2,1]=(-0.764514507478)*(1+2.*(dim-1.)/(o+1.))+(1.69386717594)*dim
        ref[0,3,0,0]=(-0.392055498198)*(1+2.*(dim-1.)/(o+1.))+(-0.384588148087)*dim
        ref[0,3,0,1]=(-1.62777049183)*(1+2.*(dim-1.)/(o+1.))+(0.706370719421)*dim
        ref[0,3,1,0]=(0.619522807034)*(1+2.*(dim-1.)/(o+1.))+(-0.0583127283348)*dim
        ref[0,3,1,1]=(1.39660007067)*(1+2.*(dim-1.)/(o+1.))+(1.3874760814)*dim
        ref[0,3,2,0]=(0.396663683156)*(1+2.*(dim-1.)/(o+1.))+(-0.449036865886)*dim
        ref[0,3,2,1]=(-1.08765510225)*(1+2.*(dim-1.)/(o+1.))+(-0.941587079516)*dim
        ref[0,4,0,0]=(-1.51265971093)*(1+2.*(dim-1.)/(o+1.))+(-1.3889057145)*dim
        ref[0,4,0,1]=(0.901518176509)*(1+2.*(dim-1.)/(o+1.))+(1.13644729087)*dim
        ref[0,4,1,0]=(-0.58127466259)*(1+2.*(dim-1.)/(o+1.))+(-0.524611564706)*dim
        ref[0,4,1,1]=(1.59425336176)*(1+2.*(dim-1.)/(o+1.))+(-0.687231237871)*dim
        ref[0,4,2,0]=(1.07708364012)*(1+2.*(dim-1.)/(o+1.))+(-0.195988520187)*dim
        ref[0,4,2,1]=(0.108644035622)*(1+2.*(dim-1.)/(o+1.))+(1.0063476431)*dim
        ref[1,0,0,0]=(-0.153352577193)*(1+2.*(dim-1.)/(o+1.))+(1.30530543707)*dim
        ref[1,0,0,1]=(-0.518263005278)*(1+2.*(dim-1.)/(o+1.))+(0.192850811305)*dim
        ref[1,0,1,0]=(0.0968185470945)*(1+2.*(dim-1.)/(o+1.))+(2.07307753844)*dim
        ref[1,0,1,1]=(-0.572562275019)*(1+2.*(dim-1.)/(o+1.))+(-0.0243920749758)*dim
        ref[1,0,2,0]=(1.40149079544)*(1+2.*(dim-1.)/(o+1.))+(-0.043955038372)*dim
        ref[1,0,2,1]=(0.613246507575)*(1+2.*(dim-1.)/(o+1.))+(0.989479209244)*dim
        ref[1,1,0,0]=(-0.444118917423)*(1+2.*(dim-1.)/(o+1.))+(-1.06414293122)*dim
        ref[1,1,0,1]=(2.18067335546)*(1+2.*(dim-1.)/(o+1.))+(0.850794277187)*dim
        ref[1,1,1,0]=(0.976948683131)*(1+2.*(dim-1.)/(o+1.))+(2.1248442697)*dim
        ref[1,1,1,1]=(1.02658988987)*(1+2.*(dim-1.)/(o+1.))+(0.897405523639)*dim
        ref[1,1,2,0]=(-0.442692736232)*(1+2.*(dim-1.)/(o+1.))+(0.625813183096)*dim
        ref[1,1,2,1]=(-0.43212274342)*(1+2.*(dim-1.)/(o+1.))+(-0.899812523933)*dim
        ref[1,2,0,0]=(0.457022725817)*(1+2.*(dim-1.)/(o+1.))+(1.61784408168)*dim
        ref[1,2,0,1]=(0.643173594534)*(1+2.*(dim-1.)/(o+1.))+(-1.02505648131)*dim
        ref[1,2,1,0]=(-0.807339832268)*(1+2.*(dim-1.)/(o+1.))+(1.22702820531)*dim
        ref[1,2,1,1]=(1.42307806941)*(1+2.*(dim-1.)/(o+1.))+(0.706648854736)*dim
        ref[1,2,2,0]=(-1.02472620332)*(1+2.*(dim-1.)/(o+1.))+(-0.209029394299)*dim
        ref[1,2,2,1]=(-0.735490314485)*(1+2.*(dim-1.)/(o+1.))+(-0.786399462032)*dim
        ref[1,3,0,0]=(-0.193471598403)*(1+2.*(dim-1.)/(o+1.))+(-1.08627048726)*dim
        ref[1,3,0,1]=(0.385356410185)*(1+2.*(dim-1.)/(o+1.))+(-0.482348711502)*dim
        ref[1,3,1,0]=(0.387298963898)*(1+2.*(dim-1.)/(o+1.))+(-1.06869970751)*dim
        ref[1,3,1,1]=(0.423963900797)*(1+2.*(dim-1.)/(o+1.))+(0.451277559755)*dim
        ref[1,3,2,0]=(2.23728483025)*(1+2.*(dim-1.)/(o+1.))+(-0.857890417195)*dim
        ref[1,3,2,1]=(-0.537279589749)*(1+2.*(dim-1.)/(o+1.))+(-1.86223656189)*dim
        ref[1,4,0,0]=(0.0728532816595)*(1+2.*(dim-1.)/(o+1.))+(-0.371008451407)*dim
        ref[1,4,0,1]=(-0.159731100663)*(1+2.*(dim-1.)/(o+1.))+(0.575166562536)*dim
        ref[1,4,1,0]=(-1.07464731238)*(1+2.*(dim-1.)/(o+1.))+(1.73138566723)*dim
        ref[1,4,1,1]=(0.0871788096124)*(1+2.*(dim-1.)/(o+1.))+(-1.66991866721)*dim
        ref[1,4,2,0]=(-0.561546011624)*(1+2.*(dim-1.)/(o+1.))+(-0.571326412224)*dim
        ref[1,4,2,1]=(0.00625030418185)*(1+2.*(dim-1.)/(o+1.))+(-0.373239526059)*dim
        ref[2,0,0,0]=(-0.081748991935)*(1+2.*(dim-1.)/(o+1.))+(0.336055989523)*dim
        ref[2,0,0,1]=(-0.324440284217)*(1+2.*(dim-1.)/(o+1.))+(1.41179983739)*dim
        ref[2,0,1,0]=(-1.01122720324)*(1+2.*(dim-1.)/(o+1.))+(0.767764482116)*dim
        ref[2,0,1,1]=(1.90139553236)*(1+2.*(dim-1.)/(o+1.))+(0.0740760444209)*dim
        ref[2,0,2,0]=(1.07285550495)*(1+2.*(dim-1.)/(o+1.))+(-1.26566135985)*dim
        ref[2,0,2,1]=(1.4632067138)*(1+2.*(dim-1.)/(o+1.))+(0.109240301143)*dim
        ref[2,1,0,0]=(0.284420138548)*(1+2.*(dim-1.)/(o+1.))+(0.759494503384)*dim
        ref[2,1,0,1]=(-0.260549938527)*(1+2.*(dim-1.)/(o+1.))+(-0.189144554883)*dim
        ref[2,1,1,0]=(0.442311731675)*(1+2.*(dim-1.)/(o+1.))+(-1.09277886622)*dim
        ref[2,1,1,1]=(-0.10302967331)*(1+2.*(dim-1.)/(o+1.))+(-2.59245001471)*dim
        ref[2,1,2,0]=(0.739203737903)*(1+2.*(dim-1.)/(o+1.))+(1.55190288301)*dim
        ref[2,1,2,1]=(0.079851368615)*(1+2.*(dim-1.)/(o+1.))+(0.652318294318)*dim
        ref[2,2,0,0]=(-1.06662747958)*(1+2.*(dim-1.)/(o+1.))+(-0.689232572268)*dim
        ref[2,2,0,1]=(1.99822376512)*(1+2.*(dim-1.)/(o+1.))+(-1.16265243075)*dim
        ref[2,2,1,0]=(-1.22458303021)*(1+2.*(dim-1.)/(o+1.))+(0.678377949405)*dim
        ref[2,2,1,1]=(1.08782595263)*(1+2.*(dim-1.)/(o+1.))+(0.219802907121)*dim
        ref[2,2,2,0]=(0.905623047084)*(1+2.*(dim-1.)/(o+1.))+(0.44304843086)*dim
        ref[2,2,2,1]=(0.0286572039953)*(1+2.*(dim-1.)/(o+1.))+(-0.434052255584)*dim
        ref[2,3,0,0]=(0.498368738287)*(1+2.*(dim-1.)/(o+1.))+(-0.676248332227)*dim
        ref[2,3,0,1]=(0.0199276642636)*(1+2.*(dim-1.)/(o+1.))+(1.46838083323)*dim
        ref[2,3,1,0]=(-0.948945787905)*(1+2.*(dim-1.)/(o+1.))+(-0.363819033015)*dim
        ref[2,3,1,1]=(1.64627388129)*(1+2.*(dim-1.)/(o+1.))+(-0.372365997305)*dim
        ref[2,3,2,0]=(2.13784190571)*(1+2.*(dim-1.)/(o+1.))+(1.07491917749)*dim
        ref[2,3,2,1]=(-0.233611389691)*(1+2.*(dim-1.)/(o+1.))+(1.62428475081)*dim
        ref[2,4,0,0]=(0.739702054679)*(1+2.*(dim-1.)/(o+1.))+(0.680940307431)*dim
        ref[2,4,0,1]=(0.784529457615)*(1+2.*(dim-1.)/(o+1.))+(-1.24925786872)*dim
        ref[2,4,1,0]=(-0.583682264824)*(1+2.*(dim-1.)/(o+1.))+(-1.24709803197)*dim
        ref[2,4,1,1]=(2.09646981494)*(1+2.*(dim-1.)/(o+1.))+(-0.775641844239)*dim
        ref[2,4,2,0]=(0.259700462448)*(1+2.*(dim-1.)/(o+1.))+(-0.741083514052)*dim
        ref[2,4,2,1]=(0.997073171465)*(1+2.*(dim-1.)/(o+1.))+(-0.223654870403)*dim
        ref[3,0,0,0]=(0.444883466114)*(1+2.*(dim-1.)/(o+1.))+(-0.0463396426708)*dim
        ref[3,0,0,1]=(1.26485847247)*(1+2.*(dim-1.)/(o+1.))+(0.0735153889988)*dim
        ref[3,0,1,0]=(-0.0375651517431)*(1+2.*(dim-1.)/(o+1.))+(1.04743902526)*dim
        ref[3,0,1,1]=(0.5358978799)*(1+2.*(dim-1.)/(o+1.))+(0.206324335516)*dim
        ref[3,0,2,0]=(0.256485361704)*(1+2.*(dim-1.)/(o+1.))+(0.295375179654)*dim
        ref[3,0,2,1]=(-0.592012973207)*(1+2.*(dim-1.)/(o+1.))+(-1.61818910065)*dim
        ref[3,1,0,0]=(-0.526710008267)*(1+2.*(dim-1.)/(o+1.))+(-1.52452995876)*dim
        ref[3,1,0,1]=(1.25161847815)*(1+2.*(dim-1.)/(o+1.))+(-0.0215957412846)*dim
        ref[3,1,1,0]=(0.0535402130696)*(1+2.*(dim-1.)/(o+1.))+(1.23834558309)*dim
        ref[3,1,1,1]=(-0.642702769482)*(1+2.*(dim-1.)/(o+1.))+(-1.45938132983)*dim
        ref[3,1,2,0]=(0.0084187441016)*(1+2.*(dim-1.)/(o+1.))+(0.0982166324748)*dim
        ref[3,1,2,1]=(-0.0835608510497)*(1+2.*(dim-1.)/(o+1.))+(-0.960501832737)*dim
        ref[3,2,0,0]=(-0.233638393296)*(1+2.*(dim-1.)/(o+1.))+(-0.185542343145)*dim
        ref[3,2,0,1]=(1.83727235646)*(1+2.*(dim-1.)/(o+1.))+(-1.62951314453)*dim
        ref[3,2,1,0]=(0.141920587064)*(1+2.*(dim-1.)/(o+1.))+(-0.675548338499)*dim
        ref[3,2,1,1]=(-0.546837220172)*(1+2.*(dim-1.)/(o+1.))+(0.027220671997)*dim
        ref[3,2,2,0]=(1.08956280993)*(1+2.*(dim-1.)/(o+1.))+(-0.102196905755)*dim
        ref[3,2,2,1]=(1.37182911507)*(1+2.*(dim-1.)/(o+1.))+(-0.7777096925)*dim
        ref[3,3,0,0]=(-1.71451575877)*(1+2.*(dim-1.)/(o+1.))+(0.43149527699)*dim
        ref[3,3,0,1]=(0.125482430607)*(1+2.*(dim-1.)/(o+1.))+(0.710446642044)*dim
        ref[3,3,1,0]=(0.731418690422)*(1+2.*(dim-1.)/(o+1.))+(-1.73833071714)*dim
        ref[3,3,1,1]=(0.0925145394024)*(1+2.*(dim-1.)/(o+1.))+(-0.986375899836)*dim
        ref[3,3,2,0]=(-1.07231086612)*(1+2.*(dim-1.)/(o+1.))+(-0.436084890732)*dim
        ref[3,3,2,1]=(0.332375487809)*(1+2.*(dim-1.)/(o+1.))+(0.526342196849)*dim
        ref[3,4,0,0]=(0.730251131978)*(1+2.*(dim-1.)/(o+1.))+(-0.966347596298)*dim
        ref[3,4,0,1]=(0.235975729317)*(1+2.*(dim-1.)/(o+1.))+(-2.16543974943)*dim
        ref[3,4,1,0]=(-2.13209908673)*(1+2.*(dim-1.)/(o+1.))+(0.631593299244)*dim
        ref[3,4,1,1]=(-0.257512397822)*(1+2.*(dim-1.)/(o+1.))+(-1.4109264038)*dim
        ref[3,4,2,0]=(0.201086569751)*(1+2.*(dim-1.)/(o+1.))+(0.626261609021)*dim
        ref[3,4,2,1]=(0.819600430724)*(1+2.*(dim-1.)/(o+1.))+(-0.483576060183)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-1.08871335955)*x[0]+(0.247972631031)*x[1]
        ref=(-0.156403544195)*(1+2.*(dim-1.)/(o+1.))+(-0.684337184327)*dim
      else:
        s=(-0.324052151044)*x[0]+(0.441305482424)*x[1]+(-0.315030176112)*x[2]
        ref=(0.603955649036)*(1+2.*(dim-1.)/(o+1.))+(-0.801732493769)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.260677248137)*x[0]+(0.0211318398374)*x[1]
        s[1]=(-0.815879139039)*x[0]+(-0.918304389572)*x[1]
        ref[0]=(0.843100081089)*(1+2.*(dim-1.)/(o+1.))+(-1.08264548939)*dim
        ref[1]=(-1.3080191481)*(1+2.*(dim-1.)/(o+1.))+(-0.426164380506)*dim
      else:
        s[0]=(1.37908771986)*x[0]+(1.36518233103)*x[1]+(1.013246401)*x[2]
        s[1]=(0.556318683467)*x[0]+(1.45737568352)*x[1]+(0.0165078724907)*x[2]
        ref[0]=(2.05950634594)*(1+2.*(dim-1.)/(o+1.))+(1.69801010595)*dim
        ref[1]=(0.655472308826)*(1+2.*(dim-1.)/(o+1.))+(1.37472993065)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(1.44204352077)*x[0]+(-0.626836867163)*x[1]
        s[0,1]=(-0.295593397487)*x[0]+(-0.309954518607)*x[1]
        s[0,2]=(-0.546882380107)*x[0]+(-0.0925530562931)*x[1]
        s[0,3]=(-0.623335335636)*x[0]+(0.798887138514)*x[1]
        s[0,4]=(0.202234634868)*x[0]+(1.44014401659)*x[1]
        s[1,0]=(-0.772178934874)*x[0]+(-0.962339450381)*x[1]
        s[1,1]=(-0.136928685359)*x[0]+(0.385217139725)*x[1]
        s[1,2]=(-0.00807747505852)*x[0]+(1.08230267787)*x[1]
        s[1,3]=(-1.28422033803)*x[0]+(0.011228260616)*x[1]
        s[1,4]=(0.766746911823)*x[0]+(-0.145254490276)*x[1]
        s[2,0]=(1.65171936301)*x[0]+(1.68030971997)*x[1]
        s[2,1]=(-0.465968023499)*x[0]+(-0.594634053719)*x[1]
        s[2,2]=(-0.140916975546)*x[0]+(0.230405132955)*x[1]
        s[2,3]=(-0.0185338227586)*x[0]+(0.298041051546)*x[1]
        s[2,4]=(-0.024340230778)*x[0]+(-0.791873643457)*x[1]
        s[3,0]=(0.555336313403)*x[0]+(0.606898490822)*x[1]
        s[3,1]=(1.05785941408)*x[0]+(-0.672561947115)*x[1]
        s[3,2]=(1.36138187068)*x[0]+(0.259592704164)*x[1]
        s[3,3]=(0.66211210256)*x[0]+(-1.38753456619)*x[1]
        s[3,4]=(0.0789301309156)*x[0]+(1.24015676345)*x[1]
        ref[0,0]=(0.863657572901)*(1+2.*(dim-1.)/(o+1.))+(-0.0484509192919)*dim
        ref[0,1]=(-0.0608826851336)*(1+2.*(dim-1.)/(o+1.))+(-0.54466523096)*dim
        ref[0,2]=(-1.25630630361)*(1+2.*(dim-1.)/(o+1.))+(0.616870867215)*dim
        ref[0,3]=(-0.618026054387)*(1+2.*(dim-1.)/(o+1.))+(0.793577857265)*dim
        ref[0,4]=(1.55309725329)*(1+2.*(dim-1.)/(o+1.))+(0.0892813981594)*dim
        ref[1,0]=(-0.707427946889)*(1+2.*(dim-1.)/(o+1.))+(-1.02709043837)*dim
        ref[1,1]=(1.0516700664)*(1+2.*(dim-1.)/(o+1.))+(-0.803381612038)*dim
        ref[1,2]=(-0.251691185365)*(1+2.*(dim-1.)/(o+1.))+(1.32591638818)*dim
        ref[1,3]=(0.227444315739)*(1+2.*(dim-1.)/(o+1.))+(-1.50043639316)*dim
        ref[1,4]=(0.636926531504)*(1+2.*(dim-1.)/(o+1.))+(-0.0154341099568)*dim
        ref[2,0]=(1.58333782129)*(1+2.*(dim-1.)/(o+1.))+(1.7486912617)*dim
        ref[2,1]=(0.376537776448)*(1+2.*(dim-1.)/(o+1.))+(-1.43713985367)*dim
        ref[2,2]=(1.33315471883)*(1+2.*(dim-1.)/(o+1.))+(-1.24366656142)*dim
        ref[2,3]=(-0.893476349046)*(1+2.*(dim-1.)/(o+1.))+(1.17298357783)*dim
        ref[2,4]=(-0.556720283569)*(1+2.*(dim-1.)/(o+1.))+(-0.259493590666)*dim
        ref[3,0]=(-0.0344322176845)*(1+2.*(dim-1.)/(o+1.))+(1.19666702191)*dim
        ref[3,1]=(0.356058306775)*(1+2.*(dim-1.)/(o+1.))+(0.0292391601866)*dim
        ref[3,2]=(1.18926653697)*(1+2.*(dim-1.)/(o+1.))+(0.431708037874)*dim
        ref[3,3]=(-0.0855874137907)*(1+2.*(dim-1.)/(o+1.))+(-0.639835049842)*dim
        ref[3,4]=(0.485696084789)*(1+2.*(dim-1.)/(o+1.))+(0.833390809579)*dim
      else:
        s[0,0]=(1.48740802799)*x[0]+(-0.5231718635)*x[1]+(0.621553689326)*x[2]
        s[0,1]=(-0.816432702835)*x[0]+(-1.02472113234)*x[1]+(-0.498250552797)*x[2]
        s[0,2]=(-0.0547881618906)*x[0]+(0.903306651062)*x[1]+(0.471428097681)*x[2]
        s[0,3]=(-0.910642152807)*x[0]+(-0.958224215021)*x[1]+(-1.4403271378)*x[2]
        s[0,4]=(0.116198234381)*x[0]+(-0.636378388418)*x[1]+(0.1270635999)*x[2]
        s[1,0]=(-0.659576291097)*x[0]+(-0.446587669482)*x[1]+(1.38617928144)*x[2]
        s[1,1]=(0.442922554555)*x[0]+(0.2841566742)*x[1]+(-0.384854418521)*x[2]
        s[1,2]=(-0.883255151971)*x[0]+(-1.3015449692)*x[1]+(-1.49785758846)*x[2]
        s[1,3]=(-1.43150392129)*x[0]+(1.13296295364)*x[1]+(0.393410729481)*x[2]
        s[1,4]=(-0.423254486419)*x[0]+(1.58675432746)*x[1]+(1.46179069555)*x[2]
        s[2,0]=(0.467182623429)*x[0]+(-1.23530058455)*x[1]+(0.546258172921)*x[2]
        s[2,1]=(-1.2468563949)*x[0]+(1.05285375014)*x[1]+(1.00605876886)*x[2]
        s[2,2]=(-0.1261124635)*x[0]+(-0.855000687125)*x[1]+(-0.252904511506)*x[2]
        s[2,3]=(1.1950830615)*x[0]+(0.699000137772)*x[1]+(1.33793802349)*x[2]
        s[2,4]=(-1.26581081559)*x[0]+(-0.0366783527385)*x[1]+(-0.736592212733)*x[2]
        s[3,0]=(1.84878865394)*x[0]+(0.504454193662)*x[1]+(-1.15690555199)*x[2]
        s[3,1]=(-0.249497283612)*x[0]+(-0.635940805715)*x[1]+(0.168382197376)*x[2]
        s[3,2]=(-0.167328351962)*x[0]+(0.523012476804)*x[1]+(1.19870547556)*x[2]
        s[3,3]=(-0.499150198998)*x[0]+(0.448461159266)*x[1]+(0.554069089559)*x[2]
        s[3,4]=(-0.520146860421)*x[0]+(1.43515582204)*x[1]+(1.31685474485)*x[2]
        ref[0,0]=(1.56352587364)*(1+2.*(dim-1.)/(o+1.))+(0.0222639801801)*dim
        ref[0,1]=(-1.88354672396)*(1+2.*(dim-1.)/(o+1.))+(-0.455857664013)*dim
        ref[0,2]=(0.631897841337)*(1+2.*(dim-1.)/(o+1.))+(0.688048745515)*dim
        ref[0,3]=(-2.49358987828)*(1+2.*(dim-1.)/(o+1.))+(-0.815603627346)*dim
        ref[0,4]=(-1.43782849721)*(1+2.*(dim-1.)/(o+1.))+(1.04471194307)*dim
        ref[1,0]=(-0.257186244728)*(1+2.*(dim-1.)/(o+1.))+(0.537201565586)*dim
        ref[1,1]=(-1.21753131847)*(1+2.*(dim-1.)/(o+1.))+(1.55975612871)*dim
        ref[1,2]=(-1.83391505519)*(1+2.*(dim-1.)/(o+1.))+(-1.84874265444)*dim
        ref[1,3]=(0.790398824851)*(1+2.*(dim-1.)/(o+1.))+(-0.69552906302)*dim
        ref[1,4]=(1.96838334672)*(1+2.*(dim-1.)/(o+1.))+(0.656907189867)*dim
        ref[2,0]=(-0.486341534759)*(1+2.*(dim-1.)/(o+1.))+(0.264481746559)*dim
        ref[2,1]=(-0.417272431595)*(1+2.*(dim-1.)/(o+1.))+(1.22932855569)*dim
        ref[2,2]=(-0.881184195775)*(1+2.*(dim-1.)/(o+1.))+(-0.352833466355)*dim
        ref[2,3]=(1.01420597285)*(1+2.*(dim-1.)/(o+1.))+(2.21781524991)*dim
        ref[2,4]=(-0.580092347401)*(1+2.*(dim-1.)/(o+1.))+(-1.45898903366)*dim
        ref[3,0]=(0.865857995406)*(1+2.*(dim-1.)/(o+1.))+(0.330479300213)*dim
        ref[3,1]=(0.632704550785)*(1+2.*(dim-1.)/(o+1.))+(-1.34976044274)*dim
        ref[3,2]=(-0.259081469822)*(1+2.*(dim-1.)/(o+1.))+(1.81347107022)*dim
        ref[3,3]=(0.533963525526)*(1+2.*(dim-1.)/(o+1.))+(-0.0305834756989)*dim
        ref[3,4]=(1.19847134659)*(1+2.*(dim-1.)/(o+1.))+(1.03339235988)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.536617774033)*x[0]+(-1.31191068914)*x[1]
        s[0,0,1]=(1.45706261992)*x[0]+(0.33760400949)*x[1]
        s[0,1,0]=(-0.453569968232)*x[0]+(0.798108537393)*x[1]
        s[0,1,1]=(-1.26956398588)*x[0]+(-1.35096551318)*x[1]
        s[1,0,0]=(0.104407588284)*x[0]+(-0.649935519811)*x[1]
        s[1,0,1]=(-1.64047091856)*x[0]+(0.654330987243)*x[1]
        s[1,1,0]=(-0.161010086142)*x[0]+(-0.644341383353)*x[1]
        s[1,1,1]=(1.42345151896)*x[0]+(-0.963960533124)*x[1]
        s[2,0,0]=(0.069463733935)*x[0]+(-0.764246138958)*x[1]
        s[2,0,1]=(0.516672615671)*x[0]+(0.32129769972)*x[1]
        s[2,1,0]=(0.399513997078)*x[0]+(-1.76803609001)*x[1]
        s[2,1,1]=(0.0570653989294)*x[0]+(-0.191869561254)*x[1]
        s[3,0,0]=(-0.261019396242)*x[0]+(0.429544509966)*x[1]
        s[3,0,1]=(-1.40560961029)*x[0]+(-0.739182555238)*x[1]
        s[3,1,0]=(0.314850928569)*x[0]+(1.49030138742)*x[1]
        s[3,1,1]=(-1.27327662424)*x[0]+(-0.0287977839083)*x[1]
        s[4,0,0]=(-0.840541224777)*x[0]+(-0.805224659169)*x[1]
        s[4,0,1]=(0.303904176581)*x[0]+(0.353767617246)*x[1]
        s[4,1,0]=(1.65813306507)*x[0]+(1.12299575058)*x[1]
        s[4,1,1]=(-1.19199290034)*x[0]+(1.6581285511)*x[1]
        s[5,0,0]=(0.120283505072)*x[0]+(1.68044908685)*x[1]
        s[5,0,1]=(0.382489278089)*x[0]+(0.129887092126)*x[1]
        s[5,1,0]=(-1.01890077156)*x[0]+(-0.111140290699)*x[1]
        s[5,1,1]=(-0.0392100253378)*x[0]+(0.579951319719)*x[1]
        ref[0,0,0]=(-0.686284158891)*(1+2.*(dim-1.)/(o+1.))+(-0.089008756216)*dim
        ref[0,0,1]=(1.76417819838)*(1+2.*(dim-1.)/(o+1.))+(0.030488431025)*dim
        ref[0,1,0]=(0.453381498643)*(1+2.*(dim-1.)/(o+1.))+(-0.108842929482)*dim
        ref[0,1,1]=(-1.03607416678)*(1+2.*(dim-1.)/(o+1.))+(-1.58445533227)*dim
        ref[1,0,0]=(0.638386446298)*(1+2.*(dim-1.)/(o+1.))+(-1.18391437783)*dim
        ref[1,0,1]=(0.275023590147)*(1+2.*(dim-1.)/(o+1.))+(-1.26116352146)*dim
        ref[1,1,0]=(-0.869379328882)*(1+2.*(dim-1.)/(o+1.))+(0.0640278593865)*dim
        ref[1,1,1]=(0.597799121533)*(1+2.*(dim-1.)/(o+1.))+(-0.138308135699)*dim
        ref[2,0,0]=(-1.45715684308)*(1+2.*(dim-1.)/(o+1.))+(0.762374438053)*dim
        ref[2,0,1]=(0.929718948925)*(1+2.*(dim-1.)/(o+1.))+(-0.0917486335329)*dim
        ref[2,1,0]=(-0.78092454297)*(1+2.*(dim-1.)/(o+1.))+(-0.587597549963)*dim
        ref[2,1,1]=(-1.21253063514)*(1+2.*(dim-1.)/(o+1.))+(1.07772647282)*dim
        ref[3,0,0]=(1.07007881303)*(1+2.*(dim-1.)/(o+1.))+(-0.901553699308)*dim
        ref[3,0,1]=(-1.77159657093)*(1+2.*(dim-1.)/(o+1.))+(-0.373195594594)*dim
        ref[3,1,0]=(1.10443227539)*(1+2.*(dim-1.)/(o+1.))+(0.700720040595)*dim
        ref[3,1,1]=(0.315983219905)*(1+2.*(dim-1.)/(o+1.))+(-1.61805762805)*dim
        ref[4,0,0]=(-0.602218266612)*(1+2.*(dim-1.)/(o+1.))+(-1.04354761733)*dim
        ref[4,0,1]=(0.297276459441)*(1+2.*(dim-1.)/(o+1.))+(0.360395334386)*dim
        ref[4,1,0]=(1.5584687424)*(1+2.*(dim-1.)/(o+1.))+(1.22266007325)*dim
        ref[4,1,1]=(0.504270780151)*(1+2.*(dim-1.)/(o+1.))+(-0.038135129389)*dim
        ref[5,0,0]=(1.35289487036)*(1+2.*(dim-1.)/(o+1.))+(0.44783772157)*dim
        ref[5,0,1]=(1.57660464714)*(1+2.*(dim-1.)/(o+1.))+(-1.06422827692)*dim
        ref[5,1,0]=(0.1360111664)*(1+2.*(dim-1.)/(o+1.))+(-1.26605222866)*dim
        ref[5,1,1]=(0.024986479373)*(1+2.*(dim-1.)/(o+1.))+(0.515754815008)*dim
      else:
        s[0,0,0]=(-0.491953157793)*x[0]+(0.808487399764)*x[1]+(0.87783526238)*x[2]
        s[0,0,1]=(-0.470158989175)*x[0]+(-0.492379274826)*x[1]+(-0.191290328078)*x[2]
        s[0,1,0]=(-0.356351922052)*x[0]+(0.823887078075)*x[1]+(-0.177422034195)*x[2]
        s[0,1,1]=(-0.378759483085)*x[0]+(0.116067213963)*x[1]+(0.0288265220551)*x[2]
        s[1,0,0]=(-0.325670782382)*x[0]+(-1.01289817775)*x[1]+(0.672608800327)*x[2]
        s[1,0,1]=(-0.296979811719)*x[0]+(0.46777334672)*x[1]+(1.61350930522)*x[2]
        s[1,1,0]=(-0.0574537817174)*x[0]+(1.10617918343)*x[1]+(0.172222519609)*x[2]
        s[1,1,1]=(-0.978413596402)*x[0]+(-1.20699880678)*x[1]+(-0.0600246637942)*x[2]
        s[2,0,0]=(0.637900823637)*x[0]+(1.29751687672)*x[1]+(-0.390955179963)*x[2]
        s[2,0,1]=(-1.15596339731)*x[0]+(-1.04131524409)*x[1]+(0.111161978383)*x[2]
        s[2,1,0]=(-1.33713877493)*x[0]+(0.0677540624633)*x[1]+(0.0662976219802)*x[2]
        s[2,1,1]=(0.0636656461261)*x[0]+(-1.32776475079)*x[1]+(0.365594993394)*x[2]
        s[3,0,0]=(0.454007484649)*x[0]+(-1.56273676656)*x[1]+(-0.783612553822)*x[2]
        s[3,0,1]=(-0.908388718727)*x[0]+(0.496552922533)*x[1]+(1.0802957835)*x[2]
        s[3,1,0]=(0.46784249418)*x[0]+(1.627686011)*x[1]+(1.12464788868)*x[2]
        s[3,1,1]=(1.47169416804)*x[0]+(-0.541043548439)*x[1]+(-0.0203908370923)*x[2]
        s[4,0,0]=(0.153593270409)*x[0]+(0.991271922294)*x[1]+(0.910688732568)*x[2]
        s[4,0,1]=(1.26666528018)*x[0]+(-0.0415946389155)*x[1]+(-0.702890723051)*x[2]
        s[4,1,0]=(0.401000008974)*x[0]+(0.270202736778)*x[1]+(-1.52334516556)*x[2]
        s[4,1,1]=(-0.567247526123)*x[0]+(0.0638484526058)*x[1]+(-0.942447437378)*x[2]
        s[5,0,0]=(1.3141482761)*x[0]+(0.602038495282)*x[1]+(-0.162794144191)*x[2]
        s[5,0,1]=(1.49343941166)*x[0]+(0.340106070028)*x[1]+(0.222963641278)*x[2]
        s[5,1,0]=(1.29779085951)*x[0]+(0.44436851782)*x[1]+(-0.414930735066)*x[2]
        s[5,1,1]=(0.445071462828)*x[0]+(0.0113100435192)*x[1]+(-0.304892738875)*x[2]
        ref[0,0,0]=(0.807154905774)*(1+2.*(dim-1.)/(o+1.))+(0.387214598576)*dim
        ref[0,0,1]=(-0.168668456531)*(1+2.*(dim-1.)/(o+1.))+(-0.985160135549)*dim
        ref[0,1,0]=(-0.405021627391)*(1+2.*(dim-1.)/(o+1.))+(0.69513474922)*dim
        ref[0,1,1]=(0.481047411067)*(1+2.*(dim-1.)/(o+1.))+(-0.714913158134)*dim
        ref[1,0,0]=(-0.241963228819)*(1+2.*(dim-1.)/(o+1.))+(-0.423996930987)*dim
        ref[1,0,1]=(-0.222173232945)*(1+2.*(dim-1.)/(o+1.))+(2.00647607317)*dim
        ref[1,1,0]=(2.08927091112)*(1+2.*(dim-1.)/(o+1.))+(-0.868322989801)*dim
        ref[1,1,1]=(-1.2199872069)*(1+2.*(dim-1.)/(o+1.))+(-1.02544986008)*dim
        ref[2,0,0]=(1.51358492234)*(1+2.*(dim-1.)/(o+1.))+(0.0308775980534)*dim
        ref[2,0,1]=(-1.17419262019)*(1+2.*(dim-1.)/(o+1.))+(-0.911924042825)*dim
        ref[2,1,0]=(-0.808827279389)*(1+2.*(dim-1.)/(o+1.))+(-0.394259811094)*dim
        ref[2,1,1]=(-0.39434462087)*(1+2.*(dim-1.)/(o+1.))+(-0.504159490404)*dim
        ref[3,0,0]=(-0.420509147061)*(1+2.*(dim-1.)/(o+1.))+(-1.47183268867)*dim
        ref[3,0,1]=(-0.509702954626)*(1+2.*(dim-1.)/(o+1.))+(1.17816294193)*dim
        ref[3,1,0]=(2.45580213911)*(1+2.*(dim-1.)/(o+1.))+(0.764374254754)*dim
        ref[3,1,1]=(0.36148972446)*(1+2.*(dim-1.)/(o+1.))+(0.548770058051)*dim
        ref[4,0,0]=(0.723463115757)*(1+2.*(dim-1.)/(o+1.))+(1.33209080951)*dim
        ref[4,0,1]=(-0.584660571656)*(1+2.*(dim-1.)/(o+1.))+(1.10684048987)*dim
        ref[4,1,0]=(0.558877069737)*(1+2.*(dim-1.)/(o+1.))+(-1.41101948955)*dim
        ref[4,1,1]=(-1.46224923696)*(1+2.*(dim-1.)/(o+1.))+(0.016402726067)*dim
        ref[5,0,0]=(0.375824940974)*(1+2.*(dim-1.)/(o+1.))+(1.37756768621)*dim
        ref[5,0,1]=(0.0790629051458)*(1+2.*(dim-1.)/(o+1.))+(1.97744621782)*dim
        ref[5,1,0]=(-0.429416156192)*(1+2.*(dim-1.)/(o+1.))+(1.75664479845)*dim
        ref[5,1,1]=(-0.227051313962)*(1+2.*(dim-1.)/(o+1.))+(0.378540081434)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the FunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-1.24466004429)*x[0]+(0.249656583863)*x[1]
        s[0,0,0,1]=(-1.31303014286)*x[0]+(-0.188187818679)*x[1]
        s[0,0,1,0]=(-1.19686759063)*x[0]+(0.202998667098)*x[1]
        s[0,0,1,1]=(-0.240139422625)*x[0]+(1.12145727417)*x[1]
        s[0,0,2,0]=(-0.779228056931)*x[0]+(-0.256394958159)*x[1]
        s[0,0,2,1]=(0.511346899235)*x[0]+(-1.17117401759)*x[1]
        s[0,1,0,0]=(1.20931773874)*x[0]+(0.703848713877)*x[1]
        s[0,1,0,1]=(1.05396707864)*x[0]+(0.561081392967)*x[1]
        s[0,1,1,0]=(-0.940092412899)*x[0]+(0.197879835766)*x[1]
        s[0,1,1,1]=(0.031033640344)*x[0]+(1.47668227837)*x[1]
        s[0,1,2,0]=(-0.973745458182)*x[0]+(-0.699999591381)*x[1]
        s[0,1,2,1]=(0.0637513252131)*x[0]+(0.223369875142)*x[1]
        s[0,2,0,0]=(-0.179817828717)*x[0]+(-1.0618817379)*x[1]
        s[0,2,0,1]=(-1.31848741633)*x[0]+(0.562871230704)*x[1]
        s[0,2,1,0]=(0.20660993199)*x[0]+(-0.260401741099)*x[1]
        s[0,2,1,1]=(-0.119752579601)*x[0]+(1.03659426931)*x[1]
        s[0,2,2,0]=(-1.05257053872)*x[0]+(-0.410999368732)*x[1]
        s[0,2,2,1]=(0.142472752971)*x[0]+(0.687099054848)*x[1]
        s[0,3,0,0]=(-0.735958436285)*x[0]+(0.576103117069)*x[1]
        s[0,3,0,1]=(0.820759477791)*x[0]+(1.40201775771)*x[1]
        s[0,3,1,0]=(-1.43663163172)*x[0]+(-0.928562287497)*x[1]
        s[0,3,1,1]=(-0.640592604343)*x[0]+(-1.37951764035)*x[1]
        s[0,3,2,0]=(0.923170161037)*x[0]+(-1.29865735065)*x[1]
        s[0,3,2,1]=(-1.49185366667)*x[0]+(1.26235954659)*x[1]
        s[0,4,0,0]=(-1.54260726706)*x[0]+(0.386401532954)*x[1]
        s[0,4,0,1]=(0.938275027884)*x[0]+(-0.325978711858)*x[1]
        s[0,4,1,0]=(-0.741422170549)*x[0]+(0.360529257608)*x[1]
        s[0,4,1,1]=(-1.38867328824)*x[0]+(0.368979718783)*x[1]
        s[0,4,2,0]=(-1.19633228454)*x[0]+(0.143128822342)*x[1]
        s[0,4,2,1]=(0.365590220489)*x[0]+(0.702852400998)*x[1]
        s[1,0,0,0]=(0.590708718523)*x[0]+(-1.74588023915)*x[1]
        s[1,0,0,1]=(-0.411269052638)*x[0]+(0.677504850422)*x[1]
        s[1,0,1,0]=(1.54641105105)*x[0]+(-0.227798671895)*x[1]
        s[1,0,1,1]=(0.206986074762)*x[0]+(0.431911634342)*x[1]
        s[1,0,2,0]=(0.180245950747)*x[0]+(-0.766496827464)*x[1]
        s[1,0,2,1]=(0.920490611623)*x[0]+(0.146274601957)*x[1]
        s[1,1,0,0]=(-1.47489897228)*x[0]+(-1.0428187509)*x[1]
        s[1,1,0,1]=(-1.17774302127)*x[0]+(0.44418244557)*x[1]
        s[1,1,1,0]=(-0.340268939926)*x[0]+(-0.362308449805)*x[1]
        s[1,1,1,1]=(1.1122568876)*x[0]+(0.807887213986)*x[1]
        s[1,1,2,0]=(-1.1580670758)*x[0]+(0.61015974696)*x[1]
        s[1,1,2,1]=(-0.21738296838)*x[0]+(-1.15287881926)*x[1]
        s[1,2,0,0]=(-1.63285965164)*x[0]+(0.546571504115)*x[1]
        s[1,2,0,1]=(1.34627844359)*x[0]+(0.0153020013782)*x[1]
        s[1,2,1,0]=(-0.459380401764)*x[0]+(0.653971151372)*x[1]
        s[1,2,1,1]=(0.0918961369392)*x[0]+(0.631274394662)*x[1]
        s[1,2,2,0]=(-0.461548096839)*x[0]+(0.432322561401)*x[1]
        s[1,2,2,1]=(-0.741897062249)*x[0]+(-0.0397725689655)*x[1]
        s[1,3,0,0]=(-0.166826966339)*x[0]+(0.76157630088)*x[1]
        s[1,3,0,1]=(-1.43001526663)*x[0]+(0.0200445726731)*x[1]
        s[1,3,1,0]=(0.776640170933)*x[0]+(1.77522529504)*x[1]
        s[1,3,1,1]=(0.132999332599)*x[0]+(0.91396296202)*x[1]
        s[1,3,2,0]=(-0.435447499973)*x[0]+(-0.714956349315)*x[1]
        s[1,3,2,1]=(1.54444404505)*x[0]+(0.0419015202674)*x[1]
        s[1,4,0,0]=(-1.14197604256)*x[0]+(0.140549307511)*x[1]
        s[1,4,0,1]=(1.11080451267)*x[0]+(0.140105909725)*x[1]
        s[1,4,1,0]=(0.361572095315)*x[0]+(1.73141141151)*x[1]
        s[1,4,1,1]=(1.2968959883)*x[0]+(0.459397526479)*x[1]
        s[1,4,2,0]=(-1.04477229488)*x[0]+(-0.432251933613)*x[1]
        s[1,4,2,1]=(0.166352074158)*x[0]+(0.240538931271)*x[1]
        s[2,0,0,0]=(-0.302792258664)*x[0]+(0.014159412453)*x[1]
        s[2,0,0,1]=(0.161287860507)*x[0]+(-1.66699057701)*x[1]
        s[2,0,1,0]=(0.793079282708)*x[0]+(0.999374748613)*x[1]
        s[2,0,1,1]=(-0.0860624573943)*x[0]+(0.438101549788)*x[1]
        s[2,0,2,0]=(-0.832685279306)*x[0]+(1.20735467656)*x[1]
        s[2,0,2,1]=(-0.0628631179915)*x[0]+(0.720123606355)*x[1]
        s[2,1,0,0]=(0.275852566161)*x[0]+(-0.259595395227)*x[1]
        s[2,1,0,1]=(0.862309966294)*x[0]+(-0.45114296596)*x[1]
        s[2,1,1,0]=(0.14823103051)*x[0]+(-1.16121025546)*x[1]
        s[2,1,1,1]=(-1.34419311122)*x[0]+(1.63903760286)*x[1]
        s[2,1,2,0]=(1.18830622559)*x[0]+(0.985255273279)*x[1]
        s[2,1,2,1]=(-0.0208912272649)*x[0]+(0.543921272375)*x[1]
        s[2,2,0,0]=(0.510821411306)*x[0]+(-1.88592356399)*x[1]
        s[2,2,0,1]=(0.353203575328)*x[0]+(0.328922615441)*x[1]
        s[2,2,1,0]=(1.1536669592)*x[0]+(-1.05621607534)*x[1]
        s[2,2,1,1]=(-0.581456156893)*x[0]+(-0.20275776476)*x[1]
        s[2,2,2,0]=(1.23883996503)*x[0]+(0.577857478009)*x[1]
        s[2,2,2,1]=(0.938162394147)*x[0]+(1.74792706902)*x[1]
        s[2,3,0,0]=(-0.0610369156925)*x[0]+(0.316929103344)*x[1]
        s[2,3,0,1]=(0.115575825651)*x[0]+(0.196603816526)*x[1]
        s[2,3,1,0]=(-0.137564893251)*x[0]+(0.732741603189)*x[1]
        s[2,3,1,1]=(-0.659475613725)*x[0]+(0.126037736276)*x[1]
        s[2,3,2,0]=(1.1911196497)*x[0]+(0.86409088069)*x[1]
        s[2,3,2,1]=(-0.161576876539)*x[0]+(-0.519953247817)*x[1]
        s[2,4,0,0]=(0.653018736743)*x[0]+(-0.141460723253)*x[1]
        s[2,4,0,1]=(0.548072342723)*x[0]+(0.79847118191)*x[1]
        s[2,4,1,0]=(-0.894497169389)*x[0]+(0.358296451187)*x[1]
        s[2,4,1,1]=(0.104304124331)*x[0]+(0.229557340588)*x[1]
        s[2,4,2,0]=(-0.775504740819)*x[0]+(-0.914276983707)*x[1]
        s[2,4,2,1]=(-0.997444443952)*x[0]+(-0.550036262287)*x[1]
        s[3,0,0,0]=(-0.912968382437)*x[0]+(1.34668077004)*x[1]
        s[3,0,0,1]=(0.279466090703)*x[0]+(1.28107040706)*x[1]
        s[3,0,1,0]=(0.282808913079)*x[0]+(0.740167690692)*x[1]
        s[3,0,1,1]=(-1.39563519985)*x[0]+(-0.567913880569)*x[1]
        s[3,0,2,0]=(1.27572544936)*x[0]+(-0.733366778512)*x[1]
        s[3,0,2,1]=(-0.33162427954)*x[0]+(1.35270158006)*x[1]
        s[3,1,0,0]=(0.519174606536)*x[0]+(-0.372514377255)*x[1]
        s[3,1,0,1]=(-1.58036427028)*x[0]+(-0.235983616334)*x[1]
        s[3,1,1,0]=(-0.472835600722)*x[0]+(0.721357411442)*x[1]
        s[3,1,1,1]=(0.448313914421)*x[0]+(-0.138520459243)*x[1]
        s[3,1,2,0]=(-1.51018963552)*x[0]+(1.34139211207)*x[1]
        s[3,1,2,1]=(-0.212789438989)*x[0]+(-1.06575114519)*x[1]
        s[3,2,0,0]=(0.0592065335067)*x[0]+(0.359601543991)*x[1]
        s[3,2,0,1]=(-0.594769346579)*x[0]+(1.5491276729)*x[1]
        s[3,2,1,0]=(0.28455993713)*x[0]+(0.600972748948)*x[1]
        s[3,2,1,1]=(0.398577473206)*x[0]+(0.0479286078109)*x[1]
        s[3,2,2,0]=(-1.76127884396)*x[0]+(-0.538814316335)*x[1]
        s[3,2,2,1]=(-0.390841519342)*x[0]+(0.191259339024)*x[1]
        s[3,3,0,0]=(-0.877110189074)*x[0]+(0.301682318677)*x[1]
        s[3,3,0,1]=(0.112669606669)*x[0]+(1.0395312515)*x[1]
        s[3,3,1,0]=(1.04370247351)*x[0]+(0.581250871082)*x[1]
        s[3,3,1,1]=(-0.143233022608)*x[0]+(0.593143844478)*x[1]
        s[3,3,2,0]=(-0.304532308332)*x[0]+(-0.232727140662)*x[1]
        s[3,3,2,1]=(0.767141825929)*x[0]+(0.0401997820135)*x[1]
        s[3,4,0,0]=(-0.226100649415)*x[0]+(-0.656022574403)*x[1]
        s[3,4,0,1]=(0.679160670663)*x[0]+(0.507720591962)*x[1]
        s[3,4,1,0]=(-1.30886998647)*x[0]+(-1.3539695469)*x[1]
        s[3,4,1,1]=(-1.39638819903)*x[0]+(0.650428829094)*x[1]
        s[3,4,2,0]=(-0.346486286718)*x[0]+(-0.869112300588)*x[1]
        s[3,4,2,1]=(-1.27650086562)*x[0]+(0.660506644039)*x[1]
        ref[0,0,0,0]=(-0.625690911818)*(1+2.*(dim-1.)/(o+1.))+(-0.369312548605)*dim
        ref[0,0,0,1]=(-0.0186378038077)*(1+2.*(dim-1.)/(o+1.))+(-1.48258015773)*dim
        ref[0,0,1,0]=(-1.04194152975)*(1+2.*(dim-1.)/(o+1.))+(0.0480726062254)*dim
        ref[0,0,1,1]=(0.00120841586244)*(1+2.*(dim-1.)/(o+1.))+(0.880109435685)*dim
        ref[0,0,2,0]=(-0.65182866689)*(1+2.*(dim-1.)/(o+1.))+(-0.3837943482)*dim
        ref[0,0,2,1]=(-0.729868010003)*(1+2.*(dim-1.)/(o+1.))+(0.0700408916461)*dim
        ref[0,1,0,0]=(1.26484534964)*(1+2.*(dim-1.)/(o+1.))+(0.648321102977)*dim
        ref[0,1,0,1]=(0.334654674402)*(1+2.*(dim-1.)/(o+1.))+(1.28039379721)*dim
        ref[0,1,1,0]=(0.338629227563)*(1+2.*(dim-1.)/(o+1.))+(-1.0808418047)*dim
        ref[0,1,1,1]=(0.0969508871625)*(1+2.*(dim-1.)/(o+1.))+(1.41076503155)*dim
        ref[0,1,2,0]=(-0.462969154119)*(1+2.*(dim-1.)/(o+1.))+(-1.21077589544)*dim
        ref[0,1,2,1]=(0.46756493995)*(1+2.*(dim-1.)/(o+1.))+(-0.180443739595)*dim
        ref[0,2,0,0]=(-0.0576220389279)*(1+2.*(dim-1.)/(o+1.))+(-1.18407752769)*dim
        ref[0,2,0,1]=(-0.650282644944)*(1+2.*(dim-1.)/(o+1.))+(-0.105333540684)*dim
        ref[0,2,1,0]=(0.136853100987)*(1+2.*(dim-1.)/(o+1.))+(-0.190644910096)*dim
        ref[0,2,1,1]=(-0.637797927534)*(1+2.*(dim-1.)/(o+1.))+(1.55463961724)*dim
        ref[0,2,2,0]=(-0.455136385489)*(1+2.*(dim-1.)/(o+1.))+(-1.00843352196)*dim
        ref[0,2,2,1]=(-0.35303452486)*(1+2.*(dim-1.)/(o+1.))+(1.18260633268)*dim
        ref[0,3,0,0]=(-0.315687014208)*(1+2.*(dim-1.)/(o+1.))+(0.155831694992)*dim
        ref[0,3,0,1]=(1.52158435346)*(1+2.*(dim-1.)/(o+1.))+(0.701192882048)*dim
        ref[0,3,1,0]=(-0.836739848177)*(1+2.*(dim-1.)/(o+1.))+(-1.52845407104)*dim
        ref[0,3,1,1]=(-0.983041539851)*(1+2.*(dim-1.)/(o+1.))+(-1.03706870484)*dim
        ref[0,3,2,0]=(-0.497690471912)*(1+2.*(dim-1.)/(o+1.))+(0.122203282296)*dim
        ref[0,3,2,1]=(0.394951943207)*(1+2.*(dim-1.)/(o+1.))+(-0.624446063287)*dim
        ref[0,4,0,0]=(-0.63974792726)*(1+2.*(dim-1.)/(o+1.))+(-0.516457806843)*dim
        ref[0,4,0,1]=(0.62556865655)*(1+2.*(dim-1.)/(o+1.))+(-0.0132723405234)*dim
        ref[0,4,1,0]=(-1.10166644693)*(1+2.*(dim-1.)/(o+1.))+(0.720773533984)*dim
        ref[0,4,1,1]=(-1.09945957099)*(1+2.*(dim-1.)/(o+1.))+(0.0797660015265)*dim
        ref[0,4,2,0]=(-0.359198665738)*(1+2.*(dim-1.)/(o+1.))+(-0.694004796462)*dim
        ref[0,4,2,1]=(1.60203342245)*(1+2.*(dim-1.)/(o+1.))+(-0.533590800966)*dim
        ref[1,0,0,0]=(0.0630998548187)*(1+2.*(dim-1.)/(o+1.))+(-1.21827137545)*dim
        ref[1,0,0,1]=(-0.362786036469)*(1+2.*(dim-1.)/(o+1.))+(0.629021834252)*dim
        ref[1,0,1,0]=(1.30840533269)*(1+2.*(dim-1.)/(o+1.))+(0.01020704647)*dim
        ref[1,0,1,1]=(0.581668517056)*(1+2.*(dim-1.)/(o+1.))+(0.0572291920482)*dim
        ref[1,0,2,0]=(-1.1596262045)*(1+2.*(dim-1.)/(o+1.))+(0.573375327782)*dim
        ref[1,0,2,1]=(-0.025337463973)*(1+2.*(dim-1.)/(o+1.))+(1.09210267755)*dim
        ref[1,1,0,0]=(-1.46016674214)*(1+2.*(dim-1.)/(o+1.))+(-1.05755098104)*dim
        ref[1,1,0,1]=(-0.410079206814)*(1+2.*(dim-1.)/(o+1.))+(-0.323481368888)*dim
        ref[1,1,1,0]=(0.497609154803)*(1+2.*(dim-1.)/(o+1.))+(-1.20018654453)*dim
        ref[1,1,1,1]=(0.446563754938)*(1+2.*(dim-1.)/(o+1.))+(1.47358034665)*dim
        ref[1,1,2,0]=(0.124142572295)*(1+2.*(dim-1.)/(o+1.))+(-0.672049901132)*dim
        ref[1,1,2,1]=(-1.27946259076)*(1+2.*(dim-1.)/(o+1.))+(-0.09079919688)*dim
        ref[1,2,0,0]=(0.0167916382593)*(1+2.*(dim-1.)/(o+1.))+(-1.10307978578)*dim
        ref[1,2,0,1]=(0.310300272224)*(1+2.*(dim-1.)/(o+1.))+(1.05128017274)*dim
        ref[1,2,1,0]=(-0.384485455128)*(1+2.*(dim-1.)/(o+1.))+(0.579076204736)*dim
        ref[1,2,1,1]=(0.133076827561)*(1+2.*(dim-1.)/(o+1.))+(0.59009370404)*dim
        ref[1,2,2,0]=(1.4126992054)*(1+2.*(dim-1.)/(o+1.))+(-1.44192474084)*dim
        ref[1,2,2,1]=(0.235736468845)*(1+2.*(dim-1.)/(o+1.))+(-1.01740610006)*dim
        ref[1,3,0,0]=(-0.323242689323)*(1+2.*(dim-1.)/(o+1.))+(0.917992023864)*dim
        ref[1,3,0,1]=(-1.56677140485)*(1+2.*(dim-1.)/(o+1.))+(0.156800710897)*dim
        ref[1,3,1,0]=(1.39852204704)*(1+2.*(dim-1.)/(o+1.))+(1.15334341893)*dim
        ref[1,3,1,1]=(1.10144038868)*(1+2.*(dim-1.)/(o+1.))+(-0.0544780940565)*dim
        ref[1,3,2,0]=(-0.619896071163)*(1+2.*(dim-1.)/(o+1.))+(-0.530507778126)*dim
        ref[1,3,2,1]=(0.665968335878)*(1+2.*(dim-1.)/(o+1.))+(0.920377229442)*dim
        ref[1,4,0,0]=(0.24940249757)*(1+2.*(dim-1.)/(o+1.))+(-1.25082923262)*dim
        ref[1,4,0,1]=(1.26616311307)*(1+2.*(dim-1.)/(o+1.))+(-0.0152526906804)*dim
        ref[1,4,1,0]=(1.37246903375)*(1+2.*(dim-1.)/(o+1.))+(0.720514473076)*dim
        ref[1,4,1,1]=(1.20885120083)*(1+2.*(dim-1.)/(o+1.))+(0.547442313947)*dim
        ref[1,4,2,0]=(-0.929784649867)*(1+2.*(dim-1.)/(o+1.))+(-0.547239578625)*dim
        ref[1,4,2,1]=(1.40793646749)*(1+2.*(dim-1.)/(o+1.))+(-1.00104546206)*dim
        ref[2,0,0,0]=(-0.865664592293)*(1+2.*(dim-1.)/(o+1.))+(0.577031746082)*dim
        ref[2,0,0,1]=(-0.383744932135)*(1+2.*(dim-1.)/(o+1.))+(-1.12195778437)*dim
        ref[2,0,1,0]=(0.899513902212)*(1+2.*(dim-1.)/(o+1.))+(0.892940129109)*dim
        ref[2,0,1,1]=(0.422487727431)*(1+2.*(dim-1.)/(o+1.))+(-0.070448635037)*dim
        ref[2,0,2,0]=(-0.0542314325807)*(1+2.*(dim-1.)/(o+1.))+(0.428900829837)*dim
        ref[2,0,2,1]=(-0.333651152721)*(1+2.*(dim-1.)/(o+1.))+(0.990911641084)*dim
        ref[2,1,0,0]=(0.743812676627)*(1+2.*(dim-1.)/(o+1.))+(-0.727555505694)*dim
        ref[2,1,0,1]=(0.487164931518)*(1+2.*(dim-1.)/(o+1.))+(-0.0759979311835)*dim
        ref[2,1,1,0]=(0.146918316537)*(1+2.*(dim-1.)/(o+1.))+(-1.15989754148)*dim
        ref[2,1,1,1]=(-0.239925787022)*(1+2.*(dim-1.)/(o+1.))+(0.534770278668)*dim
        ref[2,1,2,0]=(1.79635305588)*(1+2.*(dim-1.)/(o+1.))+(0.377208442988)*dim
        ref[2,1,2,1]=(-0.275396782095)*(1+2.*(dim-1.)/(o+1.))+(0.798426827206)*dim
        ref[2,2,0,0]=(-0.732792630073)*(1+2.*(dim-1.)/(o+1.))+(-0.642309522614)*dim
        ref[2,2,0,1]=(0.691994610633)*(1+2.*(dim-1.)/(o+1.))+(-0.0098684198643)*dim
        ref[2,2,1,0]=(0.571059981052)*(1+2.*(dim-1.)/(o+1.))+(-0.473609097192)*dim
        ref[2,2,1,1]=(-0.935084534142)*(1+2.*(dim-1.)/(o+1.))+(0.15087061249)*dim
        ref[2,2,2,0]=(0.0826567188857)*(1+2.*(dim-1.)/(o+1.))+(1.73404072415)*dim
        ref[2,2,2,1]=(1.26427004187)*(1+2.*(dim-1.)/(o+1.))+(1.4218194213)*dim
        ref[2,3,0,0]=(-0.398591309648)*(1+2.*(dim-1.)/(o+1.))+(0.6544834973)*dim
        ref[2,3,0,1]=(-0.568822853648)*(1+2.*(dim-1.)/(o+1.))+(0.881002495825)*dim
        ref[2,3,1,0]=(1.18000532793)*(1+2.*(dim-1.)/(o+1.))+(-0.584828617992)*dim
        ref[2,3,1,1]=(-1.60821457294)*(1+2.*(dim-1.)/(o+1.))+(1.07477669549)*dim
        ref[2,3,2,0]=(1.48063830505)*(1+2.*(dim-1.)/(o+1.))+(0.574572225346)*dim
        ref[2,3,2,1]=(0.229132618257)*(1+2.*(dim-1.)/(o+1.))+(-0.910662742613)*dim
        ref[2,4,0,0]=(0.426313638958)*(1+2.*(dim-1.)/(o+1.))+(0.0852443745317)*dim
        ref[2,4,0,1]=(-0.0761298452145)*(1+2.*(dim-1.)/(o+1.))+(1.42267336985)*dim
        ref[2,4,1,0]=(-0.537516881228)*(1+2.*(dim-1.)/(o+1.))+(0.00131616302647)*dim
        ref[2,4,1,1]=(-1.00615029385)*(1+2.*(dim-1.)/(o+1.))+(1.34001175877)*dim
        ref[2,4,2,0]=(-0.250744474716)*(1+2.*(dim-1.)/(o+1.))+(-1.43903724981)*dim
        ref[2,4,2,1]=(0.273616829308)*(1+2.*(dim-1.)/(o+1.))+(-1.82109753555)*dim
        ref[3,0,0,0]=(0.773912240208)*(1+2.*(dim-1.)/(o+1.))+(-0.340199852604)*dim
        ref[3,0,0,1]=(0.00495773386215)*(1+2.*(dim-1.)/(o+1.))+(1.5555787639)*dim
        ref[3,0,1,0]=(-0.586757469166)*(1+2.*(dim-1.)/(o+1.))+(1.60973407294)*dim
        ref[3,0,1,1]=(-0.577120453982)*(1+2.*(dim-1.)/(o+1.))+(-1.38642862644)*dim
        ref[3,0,2,0]=(0.288712056159)*(1+2.*(dim-1.)/(o+1.))+(0.253646614684)*dim
        ref[3,0,2,1]=(-0.248042635081)*(1+2.*(dim-1.)/(o+1.))+(1.2691199356)*dim
        ref[3,1,0,0]=(0.995572341947)*(1+2.*(dim-1.)/(o+1.))+(-0.848912112665)*dim
        ref[3,1,0,1]=(-1.37225234108)*(1+2.*(dim-1.)/(o+1.))+(-0.444095545534)*dim
        ref[3,1,1,0]=(0.441843238326)*(1+2.*(dim-1.)/(o+1.))+(-0.193321427606)*dim
        ref[3,1,1,1]=(0.408346053376)*(1+2.*(dim-1.)/(o+1.))+(-0.098552598199)*dim
        ref[3,1,2,0]=(-0.311970472748)*(1+2.*(dim-1.)/(o+1.))+(0.143172949307)*dim
        ref[3,1,2,1]=(-0.434499269883)*(1+2.*(dim-1.)/(o+1.))+(-0.844041314296)*dim
        ref[3,2,0,0]=(0.480046851902)*(1+2.*(dim-1.)/(o+1.))+(-0.0612387744048)*dim
        ref[3,2,0,1]=(1.12223817367)*(1+2.*(dim-1.)/(o+1.))+(-0.167879847351)*dim
        ref[3,2,1,0]=(-0.345095098378)*(1+2.*(dim-1.)/(o+1.))+(1.23062778446)*dim
        ref[3,2,1,1]=(1.37605147183)*(1+2.*(dim-1.)/(o+1.))+(-0.92954539081)*dim
        ref[3,2,2,0]=(-1.0202068094)*(1+2.*(dim-1.)/(o+1.))+(-1.2798863509)*dim
        ref[3,2,2,1]=(-0.302591270785)*(1+2.*(dim-1.)/(o+1.))+(0.103009090467)*dim
        ref[3,3,0,0]=(0.857177782444)*(1+2.*(dim-1.)/(o+1.))+(-1.43260565284)*dim
        ref[3,3,0,1]=(0.0906238774103)*(1+2.*(dim-1.)/(o+1.))+(1.06157698076)*dim
        ref[3,3,1,0]=(0.42136870665)*(1+2.*(dim-1.)/(o+1.))+(1.20358463794)*dim
        ref[3,3,1,1]=(-0.88620181924)*(1+2.*(dim-1.)/(o+1.))+(1.33611264111)*dim
        ref[3,3,2,0]=(0.444491342755)*(1+2.*(dim-1.)/(o+1.))+(-0.981750791749)*dim
        ref[3,3,2,1]=(-1.06075000108)*(1+2.*(dim-1.)/(o+1.))+(1.86809160902)*dim
        ref[3,4,0,0]=(-0.517172055625)*(1+2.*(dim-1.)/(o+1.))+(-0.364951168193)*dim
        ref[3,4,0,1]=(0.683410864606)*(1+2.*(dim-1.)/(o+1.))+(0.503470398019)*dim
        ref[3,4,1,0]=(-1.49890766435)*(1+2.*(dim-1.)/(o+1.))+(-1.16393186901)*dim
        ref[3,4,1,1]=(0.25999658315)*(1+2.*(dim-1.)/(o+1.))+(-1.00595595308)*dim
        ref[3,4,2,0]=(-0.296406951133)*(1+2.*(dim-1.)/(o+1.))+(-0.919191636173)*dim
        ref[3,4,2,1]=(-0.562797436325)*(1+2.*(dim-1.)/(o+1.))+(-0.0531967852595)*dim
      else:
        s[0,0,0,0]=(0.28520634077)*x[0]+(-0.126741433974)*x[1]+(1.28370871197)*x[2]
        s[0,0,0,1]=(-0.0731693108841)*x[0]+(0.342123624879)*x[1]+(0.302260810729)*x[2]
        s[0,0,1,0]=(-0.933375161617)*x[0]+(0.714257115938)*x[1]+(0.0355044702732)*x[2]
        s[0,0,1,1]=(0.273422423495)*x[0]+(-1.43381666098)*x[1]+(-0.3924531625)*x[2]
        s[0,0,2,0]=(-1.24504380748)*x[0]+(1.36389469156)*x[1]+(1.0146102387)*x[2]
        s[0,0,2,1]=(-0.12931890066)*x[0]+(0.00848180915183)*x[1]+(0.230392813094)*x[2]
        s[0,1,0,0]=(0.390247630566)*x[0]+(0.116544846708)*x[1]+(-0.94552924973)*x[2]
        s[0,1,0,1]=(1.53653205813)*x[0]+(0.984195773879)*x[1]+(0.767182801059)*x[2]
        s[0,1,1,0]=(0.433068815739)*x[0]+(0.196964560742)*x[1]+(0.00150259046749)*x[2]
        s[0,1,1,1]=(0.620083290721)*x[0]+(-0.981816668862)*x[1]+(-0.0602017060991)*x[2]
        s[0,1,2,0]=(0.266920335934)*x[0]+(-1.01077201758)*x[1]+(-1.66555517355)*x[2]
        s[0,1,2,1]=(-0.463794455357)*x[0]+(1.02905222514)*x[1]+(-0.524965351287)*x[2]
        s[0,2,0,0]=(-1.02890742598)*x[0]+(0.10277767711)*x[1]+(-0.709831378433)*x[2]
        s[0,2,0,1]=(0.942130530664)*x[0]+(-0.498478792456)*x[1]+(-0.321634750428)*x[2]
        s[0,2,1,0]=(-1.09979499064)*x[0]+(-0.552242175008)*x[1]+(1.47155591555)*x[2]
        s[0,2,1,1]=(0.038643175545)*x[0]+(-0.367877010362)*x[1]+(-0.422403338025)*x[2]
        s[0,2,2,0]=(0.915798473155)*x[0]+(0.627559693912)*x[1]+(-0.146957763577)*x[2]
        s[0,2,2,1]=(0.35609430038)*x[0]+(0.627155969018)*x[1]+(-0.0538976009369)*x[2]
        s[0,3,0,0]=(0.753367589685)*x[0]+(-1.56174987227)*x[1]+(0.0317386362973)*x[2]
        s[0,3,0,1]=(-1.0814768856)*x[0]+(-0.000317117094536)*x[1]+(0.160394230281)*x[2]
        s[0,3,1,0]=(0.539343002581)*x[0]+(0.600975350875)*x[1]+(-0.579108274756)*x[2]
        s[0,3,1,1]=(1.93822764773)*x[0]+(1.19115168218)*x[1]+(-0.345303177833)*x[2]
        s[0,3,2,0]=(-0.396284929409)*x[0]+(-0.0620671301539)*x[1]+(0.405978876833)*x[2]
        s[0,3,2,1]=(-0.806591193034)*x[0]+(-0.869878106357)*x[1]+(-0.352772882375)*x[2]
        s[0,4,0,0]=(-1.01176424875)*x[0]+(-1.57589661925)*x[1]+(-0.313904557434)*x[2]
        s[0,4,0,1]=(0.277176183617)*x[0]+(0.548456592065)*x[1]+(1.21233269169)*x[2]
        s[0,4,1,0]=(0.916512281025)*x[0]+(-1.53600125673)*x[1]+(-0.486397251593)*x[2]
        s[0,4,1,1]=(1.12553562564)*x[0]+(0.0569780615281)*x[1]+(-0.275491563277)*x[2]
        s[0,4,2,0]=(0.0701240897706)*x[0]+(-0.109284348724)*x[1]+(0.920255378883)*x[2]
        s[0,4,2,1]=(-0.286723671632)*x[0]+(0.521149269283)*x[1]+(0.880566081072)*x[2]
        s[1,0,0,0]=(-0.161092153994)*x[0]+(1.60646814731)*x[1]+(-0.293423133436)*x[2]
        s[1,0,0,1]=(-0.62705410926)*x[0]+(-0.133468132127)*x[1]+(0.435110047414)*x[2]
        s[1,0,1,0]=(0.963863560937)*x[0]+(0.358847992997)*x[1]+(0.847184531597)*x[2]
        s[1,0,1,1]=(0.2445787465)*x[0]+(-0.161812641928)*x[1]+(-0.679720454567)*x[2]
        s[1,0,2,0]=(1.05646386542)*x[0]+(1.0631242639)*x[1]+(-0.762052372247)*x[2]
        s[1,0,2,1]=(-0.950180341823)*x[0]+(1.00890659836)*x[1]+(1.54399946028)*x[2]
        s[1,1,0,0]=(0.290340455371)*x[0]+(-1.13919380812)*x[1]+(-0.659408495899)*x[2]
        s[1,1,0,1]=(1.4258405346)*x[0]+(1.53809682182)*x[1]+(0.0675302762314)*x[2]
        s[1,1,1,0]=(1.21278717612)*x[0]+(0.220266219454)*x[1]+(1.66873955725)*x[2]
        s[1,1,1,1]=(0.771526101684)*x[0]+(-0.067360487854)*x[1]+(1.21982979968)*x[2]
        s[1,1,2,0]=(-0.77604682892)*x[0]+(0.3133384598)*x[1]+(0.645828815985)*x[2]
        s[1,1,2,1]=(-0.107715380402)*x[0]+(-0.0772190829216)*x[1]+(-1.14700080403)*x[2]
        s[1,2,0,0]=(0.324137135934)*x[0]+(1.29562077427)*x[1]+(0.455108897284)*x[2]
        s[1,2,0,1]=(-1.43262429921)*x[0]+(0.773443491708)*x[1]+(0.277297920728)*x[2]
        s[1,2,1,0]=(0.304810473045)*x[0]+(-0.0168967979423)*x[1]+(0.131774697941)*x[2]
        s[1,2,1,1]=(0.387736648788)*x[0]+(0.666220386408)*x[1]+(1.07576988895)*x[2]
        s[1,2,2,0]=(-0.254693787451)*x[0]+(-0.3384585134)*x[1]+(-0.64060329677)*x[2]
        s[1,2,2,1]=(-0.672597816893)*x[0]+(0.145646769601)*x[1]+(-0.994938729226)*x[2]
        s[1,3,0,0]=(-1.1030634788)*x[0]+(-1.80762204727)*x[1]+(1.6309434404)*x[2]
        s[1,3,0,1]=(0.276318339362)*x[0]+(0.385250530079)*x[1]+(-0.758561170758)*x[2]
        s[1,3,1,0]=(0.179910601577)*x[0]+(-0.11818800924)*x[1]+(-0.743123335945)*x[2]
        s[1,3,1,1]=(0.782295683197)*x[0]+(0.749420799157)*x[1]+(-0.656475021802)*x[2]
        s[1,3,2,0]=(0.574829069465)*x[0]+(-0.493394135291)*x[1]+(1.29795947888)*x[2]
        s[1,3,2,1]=(-1.46888017959)*x[0]+(-1.08679474255)*x[1]+(0.156158770501)*x[2]
        s[1,4,0,0]=(-0.860789510784)*x[0]+(0.652130849161)*x[1]+(-0.0894965081242)*x[2]
        s[1,4,0,1]=(0.997352754709)*x[0]+(-0.112776085202)*x[1]+(-0.469141207634)*x[2]
        s[1,4,1,0]=(-0.401561591069)*x[0]+(-0.362165453433)*x[1]+(1.42046539935)*x[2]
        s[1,4,1,1]=(-0.668331880273)*x[0]+(-1.28337116108)*x[1]+(0.368963183759)*x[2]
        s[1,4,2,0]=(-1.15039637411)*x[0]+(0.121579045201)*x[1]+(-0.104055094935)*x[2]
        s[1,4,2,1]=(-0.923080753122)*x[0]+(0.322395936762)*x[1]+(0.233695594483)*x[2]
        s[2,0,0,0]=(-0.206424413737)*x[0]+(0.761779799677)*x[1]+(-0.301048388352)*x[2]
        s[2,0,0,1]=(1.55990080913)*x[0]+(-0.162113046152)*x[1]+(-0.310428209804)*x[2]
        s[2,0,1,0]=(-0.433285639024)*x[0]+(1.36777883595)*x[1]+(-1.17795591805)*x[2]
        s[2,0,1,1]=(1.34284608147)*x[0]+(-0.222861276203)*x[1]+(0.855486771517)*x[2]
        s[2,0,2,0]=(-0.16635189023)*x[0]+(-0.823533654953)*x[1]+(0.797079690281)*x[2]
        s[2,0,2,1]=(0.752310816499)*x[0]+(0.764018784752)*x[1]+(0.056117413692)*x[2]
        s[2,1,0,0]=(0.00166988723156)*x[0]+(0.766304091963)*x[1]+(0.275940662737)*x[2]
        s[2,1,0,1]=(-0.0855732564349)*x[0]+(-1.02664418473)*x[1]+(0.662522947755)*x[2]
        s[2,1,1,0]=(-0.658811588292)*x[0]+(-1.04635899154)*x[1]+(1.05470344529)*x[2]
        s[2,1,1,1]=(-0.119299788842)*x[0]+(-0.782372280168)*x[1]+(-1.79380761901)*x[2]
        s[2,1,2,0]=(1.71887176677)*x[0]+(-0.226946752041)*x[1]+(0.799181606184)*x[2]
        s[2,1,2,1]=(-0.9189461833)*x[0]+(0.916498349691)*x[1]+(0.734617496542)*x[2]
        s[2,2,0,0]=(-0.790364198949)*x[0]+(0.0131553082942)*x[1]+(-0.978651161192)*x[2]
        s[2,2,0,1]=(0.32388442717)*x[0]+(-0.593416422259)*x[1]+(1.10510332946)*x[2]
        s[2,2,1,0]=(0.452042167919)*x[0]+(-0.641462816719)*x[1]+(-0.356784432003)*x[2]
        s[2,2,1,1]=(0.0725380542114)*x[0]+(0.255256127518)*x[1]+(0.979834678019)*x[2]
        s[2,2,2,0]=(1.39372044596)*x[0]+(0.938441851122)*x[1]+(-0.983490819141)*x[2]
        s[2,2,2,1]=(1.2364973033)*x[0]+(-1.2815807633)*x[1]+(-0.360311591593)*x[2]
        s[2,3,0,0]=(-0.0597199849648)*x[0]+(0.376596450311)*x[1]+(-0.494756059286)*x[2]
        s[2,3,0,1]=(-0.1287394847)*x[0]+(0.0455819025025)*x[1]+(1.57146607969)*x[2]
        s[2,3,1,0]=(-1.4907853501)*x[0]+(0.0388186733561)*x[1]+(0.13920185582)*x[2]
        s[2,3,1,1]=(0.0439467610482)*x[0]+(1.43515095249)*x[1]+(-0.205189829549)*x[2]
        s[2,3,2,0]=(0.773581921294)*x[0]+(1.41482523023)*x[1]+(1.02435393168)*x[2]
        s[2,3,2,1]=(0.347805074163)*x[0]+(1.52992663335)*x[1]+(-0.487058346391)*x[2]
        s[2,4,0,0]=(-0.107692405269)*x[0]+(0.147562810855)*x[1]+(1.38077195652)*x[2]
        s[2,4,0,1]=(-0.0410323036269)*x[0]+(0.269432309711)*x[1]+(-0.693128417186)*x[2]
        s[2,4,1,0]=(-0.33861862665)*x[0]+(-0.419772907291)*x[1]+(-1.07238876285)*x[2]
        s[2,4,1,1]=(0.143346805317)*x[0]+(-0.419444960688)*x[1]+(1.59692612607)*x[2]
        s[2,4,2,0]=(0.654582963584)*x[0]+(-0.44163597128)*x[1]+(-0.694330043907)*x[2]
        s[2,4,2,1]=(0.627421193857)*x[0]+(0.597177617998)*x[1]+(-0.451180510793)*x[2]
        s[3,0,0,0]=(-0.101412872135)*x[0]+(-0.396884673414)*x[1]+(0.896841368992)*x[2]
        s[3,0,0,1]=(0.139595535793)*x[0]+(0.624607944421)*x[1]+(0.574170381258)*x[2]
        s[3,0,1,0]=(0.316717544294)*x[0]+(-0.621261709901)*x[1]+(1.31441803912)*x[2]
        s[3,0,1,1]=(0.254418231071)*x[0]+(0.228843906675)*x[1]+(0.25896007767)*x[2]
        s[3,0,2,0]=(-0.247274104613)*x[0]+(1.20694307591)*x[1]+(-0.407808429936)*x[2]
        s[3,0,2,1]=(-1.40605558298)*x[0]+(0.168721449606)*x[1]+(-0.972867940483)*x[2]
        s[3,1,0,0]=(0.245845223266)*x[0]+(-1.51779122633)*x[1]+(-0.779293963961)*x[2]
        s[3,1,0,1]=(0.0325495798223)*x[0]+(-0.12100728517)*x[1]+(1.31848044222)*x[2]
        s[3,1,1,0]=(0.789118948962)*x[0]+(0.589652419254)*x[1]+(-0.08688557206)*x[2]
        s[3,1,1,1]=(-1.83518748403)*x[0]+(-0.500646680661)*x[1]+(0.233750065377)*x[2]
        s[3,1,2,0]=(-0.0771120817018)*x[0]+(0.0904104042813)*x[1]+(0.0933370539969)*x[2]
        s[3,1,2,1]=(-1.19249513127)*x[0]+(0.447348983457)*x[1]+(-0.298916535972)*x[2]
        s[3,2,0,0]=(1.22266670435)*x[0]+(-1.62277644197)*x[1]+(-0.0190709988254)*x[2]
        s[3,2,0,1]=(-0.194952675382)*x[0]+(0.554694000692)*x[1]+(-0.151982113386)*x[2]
        s[3,2,1,0]=(0.511265692083)*x[0]+(-0.802622720144)*x[1]+(-0.242270723374)*x[2]
        s[3,2,1,1]=(0.476021603798)*x[0]+(-0.463494861878)*x[1]+(-0.532143290096)*x[2]
        s[3,2,2,0]=(-0.530708374495)*x[0]+(1.10856138272)*x[1]+(0.409512895954)*x[2]
        s[3,2,2,1]=(0.0214256235243)*x[0]+(0.836046354994)*x[1]+(-0.263352555948)*x[2]
        s[3,3,0,0]=(-0.437181323979)*x[0]+(-0.413716617562)*x[1]+(-0.432122540243)*x[2]
        s[3,3,0,1]=(1.08881737048)*x[0]+(-0.0166529943951)*x[1]+(-0.236235303435)*x[2]
        s[3,3,1,0]=(-0.220528005238)*x[0]+(-0.0637133108381)*x[1]+(-0.722670710646)*x[2]
        s[3,3,1,1]=(0.0267596690177)*x[0]+(0.277583150215)*x[1]+(-1.19820417967)*x[2]
        s[3,3,2,0]=(-0.134456240167)*x[0]+(0.207594364333)*x[1]+(-1.58153388102)*x[2]
        s[3,3,2,1]=(-0.484174332477)*x[0]+(1.24683115653)*x[1]+(0.0960608606087)*x[2]
        s[3,4,0,0]=(-1.06280143452)*x[0]+(0.0909193631754)*x[1]+(0.735785607027)*x[2]
        s[3,4,0,1]=(-0.39199782699)*x[0]+(-0.547457599272)*x[1]+(-0.990008593852)*x[2]
        s[3,4,1,0]=(-0.301784264736)*x[0]+(-0.219227445543)*x[1]+(-0.979494077205)*x[2]
        s[3,4,1,1]=(-0.898860901012)*x[0]+(-0.103378912588)*x[1]+(-0.66619898802)*x[2]
        s[3,4,2,0]=(1.10051932868)*x[0]+(0.407341781092)*x[1]+(-0.680512930996)*x[2]
        s[3,4,2,1]=(-0.0733544592857)*x[0]+(1.18707229675)*x[1]+(-0.777693466919)*x[2]
        ref[0,0,0,0]=(0.438851859071)*(1+2.*(dim-1.)/(o+1.))+(1.0033217597)*dim
        ref[0,0,0,1]=(-0.891607994637)*(1+2.*(dim-1.)/(o+1.))+(1.46282311936)*dim
        ref[0,0,1,0]=(0.817888583209)*(1+2.*(dim-1.)/(o+1.))+(-1.00150215862)*dim
        ref[0,0,1,1]=(-1.639070793)*(1+2.*(dim-1.)/(o+1.))+(0.0862233930162)*dim
        ref[0,0,2,0]=(-0.348875799642)*(1+2.*(dim-1.)/(o+1.))+(1.48233692242)*dim
        ref[0,0,2,1]=(-0.589821443428)*(1+2.*(dim-1.)/(o+1.))+(0.699377165015)*dim
        ref[0,1,0,0]=(-0.832218783182)*(1+2.*(dim-1.)/(o+1.))+(0.393482010727)*dim
        ref[0,1,0,1]=(2.07094909251)*(1+2.*(dim-1.)/(o+1.))+(1.21696154056)*dim
        ref[0,1,1,0]=(0.1125989301)*(1+2.*(dim-1.)/(o+1.))+(0.518937036848)*dim
        ref[0,1,1,1]=(-1.36685685946)*(1+2.*(dim-1.)/(o+1.))+(0.944921775217)*dim
        ref[0,1,2,0]=(-1.47817293944)*(1+2.*(dim-1.)/(o+1.))+(-0.931233915749)*dim
        ref[0,1,2,1]=(-0.213405460956)*(1+2.*(dim-1.)/(o+1.))+(0.253697879455)*dim
        ref[0,2,0,0]=(-0.0805522984701)*(1+2.*(dim-1.)/(o+1.))+(-1.55540882883)*dim
        ref[0,2,0,1]=(0.156888902683)*(1+2.*(dim-1.)/(o+1.))+(-0.034871914902)*dim
        ref[0,2,1,0]=(-0.15684567608)*(1+2.*(dim-1.)/(o+1.))+(-0.0236355740175)*dim
        ref[0,2,1,1]=(-1.39067500442)*(1+2.*(dim-1.)/(o+1.))+(0.639037831581)*dim
        ref[0,2,2,0]=(1.81399897277)*(1+2.*(dim-1.)/(o+1.))+(-0.417598569281)*dim
        ref[0,2,2,1]=(-0.764514507478)*(1+2.*(dim-1.)/(o+1.))+(1.69386717594)*dim
        ref[0,3,0,0]=(-0.392055498198)*(1+2.*(dim-1.)/(o+1.))+(-0.384588148087)*dim
        ref[0,3,0,1]=(-1.62777049183)*(1+2.*(dim-1.)/(o+1.))+(0.706370719421)*dim
        ref[0,3,1,0]=(0.619522807034)*(1+2.*(dim-1.)/(o+1.))+(-0.0583127283348)*dim
        ref[0,3,1,1]=(1.39660007067)*(1+2.*(dim-1.)/(o+1.))+(1.3874760814)*dim
        ref[0,3,2,0]=(0.396663683156)*(1+2.*(dim-1.)/(o+1.))+(-0.449036865886)*dim
        ref[0,3,2,1]=(-1.08765510225)*(1+2.*(dim-1.)/(o+1.))+(-0.941587079516)*dim
        ref[0,4,0,0]=(-1.51265971093)*(1+2.*(dim-1.)/(o+1.))+(-1.3889057145)*dim
        ref[0,4,0,1]=(0.901518176509)*(1+2.*(dim-1.)/(o+1.))+(1.13644729087)*dim
        ref[0,4,1,0]=(-0.58127466259)*(1+2.*(dim-1.)/(o+1.))+(-0.524611564706)*dim
        ref[0,4,1,1]=(1.59425336176)*(1+2.*(dim-1.)/(o+1.))+(-0.687231237871)*dim
        ref[0,4,2,0]=(1.07708364012)*(1+2.*(dim-1.)/(o+1.))+(-0.195988520187)*dim
        ref[0,4,2,1]=(0.108644035622)*(1+2.*(dim-1.)/(o+1.))+(1.0063476431)*dim
        ref[1,0,0,0]=(-0.153352577193)*(1+2.*(dim-1.)/(o+1.))+(1.30530543707)*dim
        ref[1,0,0,1]=(-0.518263005278)*(1+2.*(dim-1.)/(o+1.))+(0.192850811305)*dim
        ref[1,0,1,0]=(0.0968185470945)*(1+2.*(dim-1.)/(o+1.))+(2.07307753844)*dim
        ref[1,0,1,1]=(-0.572562275019)*(1+2.*(dim-1.)/(o+1.))+(-0.0243920749758)*dim
        ref[1,0,2,0]=(1.40149079544)*(1+2.*(dim-1.)/(o+1.))+(-0.043955038372)*dim
        ref[1,0,2,1]=(0.613246507575)*(1+2.*(dim-1.)/(o+1.))+(0.989479209244)*dim
        ref[1,1,0,0]=(-0.444118917423)*(1+2.*(dim-1.)/(o+1.))+(-1.06414293122)*dim
        ref[1,1,0,1]=(2.18067335546)*(1+2.*(dim-1.)/(o+1.))+(0.850794277187)*dim
        ref[1,1,1,0]=(0.976948683131)*(1+2.*(dim-1.)/(o+1.))+(2.1248442697)*dim
        ref[1,1,1,1]=(1.02658988987)*(1+2.*(dim-1.)/(o+1.))+(0.897405523639)*dim
        ref[1,1,2,0]=(-0.442692736232)*(1+2.*(dim-1.)/(o+1.))+(0.625813183096)*dim
        ref[1,1,2,1]=(-0.43212274342)*(1+2.*(dim-1.)/(o+1.))+(-0.899812523933)*dim
        ref[1,2,0,0]=(0.457022725817)*(1+2.*(dim-1.)/(o+1.))+(1.61784408168)*dim
        ref[1,2,0,1]=(0.643173594534)*(1+2.*(dim-1.)/(o+1.))+(-1.02505648131)*dim
        ref[1,2,1,0]=(-0.807339832268)*(1+2.*(dim-1.)/(o+1.))+(1.22702820531)*dim
        ref[1,2,1,1]=(1.42307806941)*(1+2.*(dim-1.)/(o+1.))+(0.706648854736)*dim
        ref[1,2,2,0]=(-1.02472620332)*(1+2.*(dim-1.)/(o+1.))+(-0.209029394299)*dim
        ref[1,2,2,1]=(-0.735490314485)*(1+2.*(dim-1.)/(o+1.))+(-0.786399462032)*dim
        ref[1,3,0,0]=(-0.193471598403)*(1+2.*(dim-1.)/(o+1.))+(-1.08627048726)*dim
        ref[1,3,0,1]=(0.385356410185)*(1+2.*(dim-1.)/(o+1.))+(-0.482348711502)*dim
        ref[1,3,1,0]=(0.387298963898)*(1+2.*(dim-1.)/(o+1.))+(-1.06869970751)*dim
        ref[1,3,1,1]=(0.423963900797)*(1+2.*(dim-1.)/(o+1.))+(0.451277559755)*dim
        ref[1,3,2,0]=(2.23728483025)*(1+2.*(dim-1.)/(o+1.))+(-0.857890417195)*dim
        ref[1,3,2,1]=(-0.537279589749)*(1+2.*(dim-1.)/(o+1.))+(-1.86223656189)*dim
        ref[1,4,0,0]=(0.0728532816595)*(1+2.*(dim-1.)/(o+1.))+(-0.371008451407)*dim
        ref[1,4,0,1]=(-0.159731100663)*(1+2.*(dim-1.)/(o+1.))+(0.575166562536)*dim
        ref[1,4,1,0]=(-1.07464731238)*(1+2.*(dim-1.)/(o+1.))+(1.73138566723)*dim
        ref[1,4,1,1]=(0.0871788096124)*(1+2.*(dim-1.)/(o+1.))+(-1.66991866721)*dim
        ref[1,4,2,0]=(-0.561546011624)*(1+2.*(dim-1.)/(o+1.))+(-0.571326412224)*dim
        ref[1,4,2,1]=(0.00625030418185)*(1+2.*(dim-1.)/(o+1.))+(-0.373239526059)*dim
        ref[2,0,0,0]=(-0.081748991935)*(1+2.*(dim-1.)/(o+1.))+(0.336055989523)*dim
        ref[2,0,0,1]=(-0.324440284217)*(1+2.*(dim-1.)/(o+1.))+(1.41179983739)*dim
        ref[2,0,1,0]=(-1.01122720324)*(1+2.*(dim-1.)/(o+1.))+(0.767764482116)*dim
        ref[2,0,1,1]=(1.90139553236)*(1+2.*(dim-1.)/(o+1.))+(0.0740760444209)*dim
        ref[2,0,2,0]=(1.07285550495)*(1+2.*(dim-1.)/(o+1.))+(-1.26566135985)*dim
        ref[2,0,2,1]=(1.4632067138)*(1+2.*(dim-1.)/(o+1.))+(0.109240301143)*dim
        ref[2,1,0,0]=(0.284420138548)*(1+2.*(dim-1.)/(o+1.))+(0.759494503384)*dim
        ref[2,1,0,1]=(-0.260549938527)*(1+2.*(dim-1.)/(o+1.))+(-0.189144554883)*dim
        ref[2,1,1,0]=(0.442311731675)*(1+2.*(dim-1.)/(o+1.))+(-1.09277886622)*dim
        ref[2,1,1,1]=(-0.10302967331)*(1+2.*(dim-1.)/(o+1.))+(-2.59245001471)*dim
        ref[2,1,2,0]=(0.739203737903)*(1+2.*(dim-1.)/(o+1.))+(1.55190288301)*dim
        ref[2,1,2,1]=(0.079851368615)*(1+2.*(dim-1.)/(o+1.))+(0.652318294318)*dim
        ref[2,2,0,0]=(-1.06662747958)*(1+2.*(dim-1.)/(o+1.))+(-0.689232572268)*dim
        ref[2,2,0,1]=(1.99822376512)*(1+2.*(dim-1.)/(o+1.))+(-1.16265243075)*dim
        ref[2,2,1,0]=(-1.22458303021)*(1+2.*(dim-1.)/(o+1.))+(0.678377949405)*dim
        ref[2,2,1,1]=(1.08782595263)*(1+2.*(dim-1.)/(o+1.))+(0.219802907121)*dim
        ref[2,2,2,0]=(0.905623047084)*(1+2.*(dim-1.)/(o+1.))+(0.44304843086)*dim
        ref[2,2,2,1]=(0.0286572039953)*(1+2.*(dim-1.)/(o+1.))+(-0.434052255584)*dim
        ref[2,3,0,0]=(0.498368738287)*(1+2.*(dim-1.)/(o+1.))+(-0.676248332227)*dim
        ref[2,3,0,1]=(0.0199276642636)*(1+2.*(dim-1.)/(o+1.))+(1.46838083323)*dim
        ref[2,3,1,0]=(-0.948945787905)*(1+2.*(dim-1.)/(o+1.))+(-0.363819033015)*dim
        ref[2,3,1,1]=(1.64627388129)*(1+2.*(dim-1.)/(o+1.))+(-0.372365997305)*dim
        ref[2,3,2,0]=(2.13784190571)*(1+2.*(dim-1.)/(o+1.))+(1.07491917749)*dim
        ref[2,3,2,1]=(-0.233611389691)*(1+2.*(dim-1.)/(o+1.))+(1.62428475081)*dim
        ref[2,4,0,0]=(0.739702054679)*(1+2.*(dim-1.)/(o+1.))+(0.680940307431)*dim
        ref[2,4,0,1]=(0.784529457615)*(1+2.*(dim-1.)/(o+1.))+(-1.24925786872)*dim
        ref[2,4,1,0]=(-0.583682264824)*(1+2.*(dim-1.)/(o+1.))+(-1.24709803197)*dim
        ref[2,4,1,1]=(2.09646981494)*(1+2.*(dim-1.)/(o+1.))+(-0.775641844239)*dim
        ref[2,4,2,0]=(0.259700462448)*(1+2.*(dim-1.)/(o+1.))+(-0.741083514052)*dim
        ref[2,4,2,1]=(0.997073171465)*(1+2.*(dim-1.)/(o+1.))+(-0.223654870403)*dim
        ref[3,0,0,0]=(0.444883466114)*(1+2.*(dim-1.)/(o+1.))+(-0.0463396426708)*dim
        ref[3,0,0,1]=(1.26485847247)*(1+2.*(dim-1.)/(o+1.))+(0.0735153889988)*dim
        ref[3,0,1,0]=(-0.0375651517431)*(1+2.*(dim-1.)/(o+1.))+(1.04743902526)*dim
        ref[3,0,1,1]=(0.5358978799)*(1+2.*(dim-1.)/(o+1.))+(0.206324335516)*dim
        ref[3,0,2,0]=(0.256485361704)*(1+2.*(dim-1.)/(o+1.))+(0.295375179654)*dim
        ref[3,0,2,1]=(-0.592012973207)*(1+2.*(dim-1.)/(o+1.))+(-1.61818910065)*dim
        ref[3,1,0,0]=(-0.526710008267)*(1+2.*(dim-1.)/(o+1.))+(-1.52452995876)*dim
        ref[3,1,0,1]=(1.25161847815)*(1+2.*(dim-1.)/(o+1.))+(-0.0215957412846)*dim
        ref[3,1,1,0]=(0.0535402130696)*(1+2.*(dim-1.)/(o+1.))+(1.23834558309)*dim
        ref[3,1,1,1]=(-0.642702769482)*(1+2.*(dim-1.)/(o+1.))+(-1.45938132983)*dim
        ref[3,1,2,0]=(0.0084187441016)*(1+2.*(dim-1.)/(o+1.))+(0.0982166324748)*dim
        ref[3,1,2,1]=(-0.0835608510497)*(1+2.*(dim-1.)/(o+1.))+(-0.960501832737)*dim
        ref[3,2,0,0]=(-0.233638393296)*(1+2.*(dim-1.)/(o+1.))+(-0.185542343145)*dim
        ref[3,2,0,1]=(1.83727235646)*(1+2.*(dim-1.)/(o+1.))+(-1.62951314453)*dim
        ref[3,2,1,0]=(0.141920587064)*(1+2.*(dim-1.)/(o+1.))+(-0.675548338499)*dim
        ref[3,2,1,1]=(-0.546837220172)*(1+2.*(dim-1.)/(o+1.))+(0.027220671997)*dim
        ref[3,2,2,0]=(1.08956280993)*(1+2.*(dim-1.)/(o+1.))+(-0.102196905755)*dim
        ref[3,2,2,1]=(1.37182911507)*(1+2.*(dim-1.)/(o+1.))+(-0.7777096925)*dim
        ref[3,3,0,0]=(-1.71451575877)*(1+2.*(dim-1.)/(o+1.))+(0.43149527699)*dim
        ref[3,3,0,1]=(0.125482430607)*(1+2.*(dim-1.)/(o+1.))+(0.710446642044)*dim
        ref[3,3,1,0]=(0.731418690422)*(1+2.*(dim-1.)/(o+1.))+(-1.73833071714)*dim
        ref[3,3,1,1]=(0.0925145394024)*(1+2.*(dim-1.)/(o+1.))+(-0.986375899836)*dim
        ref[3,3,2,0]=(-1.07231086612)*(1+2.*(dim-1.)/(o+1.))+(-0.436084890732)*dim
        ref[3,3,2,1]=(0.332375487809)*(1+2.*(dim-1.)/(o+1.))+(0.526342196849)*dim
        ref[3,4,0,0]=(0.730251131978)*(1+2.*(dim-1.)/(o+1.))+(-0.966347596298)*dim
        ref[3,4,0,1]=(0.235975729317)*(1+2.*(dim-1.)/(o+1.))+(-2.16543974943)*dim
        ref[3,4,1,0]=(-2.13209908673)*(1+2.*(dim-1.)/(o+1.))+(0.631593299244)*dim
        ref[3,4,1,1]=(-0.257512397822)*(1+2.*(dim-1.)/(o+1.))+(-1.4109264038)*dim
        ref[3,4,2,0]=(0.201086569751)*(1+2.*(dim-1.)/(o+1.))+(0.626261609021)*dim
        ref[3,4,2,1]=(0.819600430724)*(1+2.*(dim-1.)/(o+1.))+(-0.483576060183)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank0(self):
      """
      tests integral of rank 0 Symbol on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.652580656299)*x[0]**o+(-0.90619447815)*x[0]+(-0.0960652458634)*x[1]**o+(-0.921815212982)*x[1]
        ref=(0.556515410435)*(1+2.*(dim-1.)/(o+1.))+(-1.82800969113)*dim
      else:
        s=(0.992239317892)*x[0]**o+(0.527946493658)*x[0]+(-0.517852623269)*x[1]**o+(-0.85996207711)*x[1]+(-0.351819025215)*x[2]**o+(0.780874062499)*x[2]
        ref=(0.122567669408)*(1+2.*(dim-1.)/(o+1.))+(0.448858479047)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank1(self):
      """
      tests integral of rank 1 Symbol on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(0.493219509026)*x[0]**o+(-0.987257686105)*x[0]+(0.593292619761)*x[1]**o+(0.404393943308)*x[1]
        s[1]=(-0.00503309013251)*x[0]**o+(0.383977250975)*x[0]+(-0.98427148677)*x[1]**o+(-0.980898646673)*x[1]
        ref[0]=(1.08651212879)*(1+2.*(dim-1.)/(o+1.))+(-0.582863742798)*dim
        ref[1]=(-0.989304576902)*(1+2.*(dim-1.)/(o+1.))+(-0.596921395698)*dim
      else:
        s[0]=(-0.807414180514)*x[0]**o+(-0.887552057443)*x[0]+(0.419062942657)*x[1]**o+(-0.0780307657804)*x[1]+(-0.57375161239)*x[2]**o+(-0.672079552789)*x[2]
        s[1]=(0.0341328802311)*x[0]**o+(0.8952312571)*x[0]+(0.769555352247)*x[1]**o+(-0.974043749526)*x[1]+(0.532943429154)*x[2]**o+(0.265140756994)*x[2]
        ref[0]=(-0.962102850247)*(1+2.*(dim-1.)/(o+1.))+(-1.63766237601)*dim
        ref[1]=(1.33663166163)*(1+2.*(dim-1.)/(o+1.))+(0.186328264568)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank2(self):
      """
      tests integral of rank 2 Symbol on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(-0.874129654878)*x[0]**o+(0.769078079549)*x[0]+(0.169490655493)*x[1]**o+(-0.414105577874)*x[1]
        s[0,1]=(-0.573932038812)*x[0]**o+(0.940575833423)*x[0]+(0.146726598392)*x[1]**o+(-0.709516681116)*x[1]
        s[0,2]=(0.958570693102)*x[0]**o+(0.345205321881)*x[0]+(-0.418462504081)*x[1]**o+(-0.391658885585)*x[1]
        s[0,3]=(0.964572209001)*x[0]**o+(-0.989029471645)*x[0]+(-0.783477665675)*x[1]**o+(-0.730975054597)*x[1]
        s[0,4]=(-0.401820155821)*x[0]**o+(0.937487436813)*x[0]+(-0.436388767817)*x[1]**o+(0.41944569363)*x[1]
        s[1,0]=(-0.218435283434)*x[0]**o+(0.449489023616)*x[0]+(0.835180453816)*x[1]**o+(0.330814802055)*x[1]
        s[1,1]=(0.997378308339)*x[0]**o+(-0.0536851038818)*x[0]+(0.415472188324)*x[1]**o+(-0.541543315513)*x[1]
        s[1,2]=(0.531080153329)*x[0]**o+(0.597962435163)*x[0]+(0.196717837822)*x[1]**o+(-0.847336564598)*x[1]
        s[1,3]=(-0.601657809776)*x[0]**o+(0.143169415905)*x[0]+(0.572624366522)*x[1]**o+(-0.126109854968)*x[1]
        s[1,4]=(-0.0923887797789)*x[0]**o+(0.743510972973)*x[0]+(-0.906380069241)*x[1]**o+(0.857368771305)*x[1]
        s[2,0]=(0.671976577482)*x[0]**o+(0.940982995212)*x[0]+(0.0235322881083)*x[1]**o+(0.551932901955)*x[1]
        s[2,1]=(-0.482943532169)*x[0]**o+(0.4285433846)*x[0]+(-0.118244007829)*x[1]**o+(0.674663307952)*x[1]
        s[2,2]=(-0.191762280668)*x[0]**o+(0.0989844222111)*x[0]+(0.322777346991)*x[1]**o+(0.0282258368072)*x[1]
        s[2,3]=(0.927688375343)*x[0]**o+(-0.616381391004)*x[0]+(0.328443474996)*x[1]**o+(0.0269050191942)*x[1]
        s[2,4]=(-0.441359800454)*x[0]**o+(-0.840582337813)*x[0]+(-0.0889373588313)*x[1]**o+(-0.973572053183)*x[1]
        s[3,0]=(0.346613552119)*x[0]**o+(0.191330676816)*x[0]+(-0.632874511488)*x[1]**o+(0.987141598)*x[1]
        s[3,1]=(-0.871079285785)*x[0]**o+(0.0595183425049)*x[0]+(-0.406377080612)*x[1]**o+(0.584095659381)*x[1]
        s[3,2]=(0.959606634672)*x[0]**o+(-0.00872257186186)*x[0]+(0.123701429421)*x[1]**o+(-0.701522275891)*x[1]
        s[3,3]=(0.791212386918)*x[0]**o+(-0.0420426671732)*x[0]+(-0.149789950006)*x[1]**o+(-0.218645658506)*x[1]
        s[3,4]=(0.784561541275)*x[0]**o+(0.615978682934)*x[0]+(0.918442985828)*x[1]**o+(0.063134842611)*x[1]
        ref[0,0]=(-0.704638999385)*(1+2.*(dim-1.)/(o+1.))+(0.354972501675)*dim
        ref[0,1]=(-0.42720544042)*(1+2.*(dim-1.)/(o+1.))+(0.231059152307)*dim
        ref[0,2]=(0.540108189021)*(1+2.*(dim-1.)/(o+1.))+(-0.046453563704)*dim
        ref[0,3]=(0.181094543326)*(1+2.*(dim-1.)/(o+1.))+(-1.72000452624)*dim
        ref[0,4]=(-0.838208923637)*(1+2.*(dim-1.)/(o+1.))+(1.35693313044)*dim
        ref[1,0]=(0.616745170381)*(1+2.*(dim-1.)/(o+1.))+(0.780303825672)*dim
        ref[1,1]=(1.41285049666)*(1+2.*(dim-1.)/(o+1.))+(-0.595228419395)*dim
        ref[1,2]=(0.727797991151)*(1+2.*(dim-1.)/(o+1.))+(-0.249374129435)*dim
        ref[1,3]=(-0.0290334432542)*(1+2.*(dim-1.)/(o+1.))+(0.017059560937)*dim
        ref[1,4]=(-0.99876884902)*(1+2.*(dim-1.)/(o+1.))+(1.60087974428)*dim
        ref[2,0]=(0.69550886559)*(1+2.*(dim-1.)/(o+1.))+(1.49291589717)*dim
        ref[2,1]=(-0.601187539998)*(1+2.*(dim-1.)/(o+1.))+(1.10320669255)*dim
        ref[2,2]=(0.131015066323)*(1+2.*(dim-1.)/(o+1.))+(0.127210259018)*dim
        ref[2,3]=(1.25613185034)*(1+2.*(dim-1.)/(o+1.))+(-0.58947637181)*dim
        ref[2,4]=(-0.530297159285)*(1+2.*(dim-1.)/(o+1.))+(-1.814154391)*dim
        ref[3,0]=(-0.28626095937)*(1+2.*(dim-1.)/(o+1.))+(1.17847227482)*dim
        ref[3,1]=(-1.2774563664)*(1+2.*(dim-1.)/(o+1.))+(0.643614001886)*dim
        ref[3,2]=(1.08330806409)*(1+2.*(dim-1.)/(o+1.))+(-0.710244847753)*dim
        ref[3,3]=(0.641422436913)*(1+2.*(dim-1.)/(o+1.))+(-0.260688325679)*dim
        ref[3,4]=(1.7030045271)*(1+2.*(dim-1.)/(o+1.))+(0.679113525545)*dim
      else:
        s[0,0]=(-0.890021592515)*x[0]**o+(0.869008439528)*x[0]+(0.0393578211875)*x[1]**o+(-0.898838314128)*x[1]+(0.120962394264)*x[2]**o+(0.820232165098)*x[2]
        s[0,1]=(0.200360087495)*x[0]**o+(-0.510360711019)*x[0]+(0.289719656756)*x[1]**o+(-0.544337956881)*x[1]+(-0.137565893265)*x[2]**o+(0.0780003071465)*x[2]
        s[0,2]=(-0.611390021035)*x[0]**o+(-0.257678585283)*x[0]+(-0.432064087215)*x[1]**o+(0.580503927543)*x[1]+(0.599605038707)*x[2]**o+(-0.116243951128)*x[2]
        s[0,3]=(0.533063923783)*x[0]**o+(0.346854667765)*x[0]+(-0.0316617652896)*x[1]**o+(-0.387870493997)*x[1]+(-0.959091675716)*x[2]**o+(-0.632504268888)*x[2]
        s[0,4]=(0.105421436346)*x[0]**o+(0.0672025049394)*x[0]+(0.295674168162)*x[1]**o+(-0.599191260911)*x[1]+(0.106390236596)*x[2]**o+(0.273159012521)*x[2]
        s[1,0]=(-0.179866026516)*x[0]**o+(-0.474318032618)*x[0]+(-0.0917700558723)*x[1]**o+(-0.957849615453)*x[1]+(0.405593449997)*x[2]**o+(-0.0879542737797)*x[2]
        s[1,1]=(0.263341986971)*x[0]**o+(0.0843810614392)*x[0]+(-0.532163327984)*x[1]**o+(0.0695511301898)*x[1]+(0.425849191612)*x[2]**o+(0.891975316063)*x[2]
        s[1,2]=(0.529944898182)*x[0]**o+(0.3446605434)*x[0]+(0.94881991666)*x[1]**o+(0.60846893826)*x[1]+(0.384682108067)*x[2]**o+(0.484190212271)*x[2]
        s[1,3]=(-0.773280762221)*x[0]**o+(-0.694374711369)*x[0]+(0.595597400746)*x[1]**o+(-0.317729831568)*x[1]+(0.436671059979)*x[2]**o+(0.862685012466)*x[2]
        s[1,4]=(0.901990663611)*x[0]**o+(-0.185951360192)*x[0]+(0.880727230793)*x[1]**o+(-0.696829887245)*x[1]+(-0.871619618845)*x[2]**o+(-0.785915181957)*x[2]
        s[2,0]=(-0.598559485292)*x[0]**o+(0.378821777768)*x[0]+(0.262525874825)*x[1]**o+(-0.633890863472)*x[1]+(0.436882086359)*x[2]**o+(0.152137062224)*x[2]
        s[2,1]=(-0.120094887119)*x[0]**o+(0.774082951002)*x[0]+(0.0910918956038)*x[1]**o+(0.914069932683)*x[1]+(-0.200132987477)*x[2]**o+(0.268683863782)*x[2]
        s[2,2]=(0.769128791381)*x[0]**o+(-0.972503532786)*x[0]+(0.898341628893)*x[1]**o+(-0.804588468967)*x[1]+(0.959444830467)*x[2]**o+(0.729116746427)*x[2]
        s[2,3]=(0.460426389529)*x[0]**o+(-0.631593304662)*x[0]+(-0.197985888231)*x[1]**o+(0.994315324532)*x[1]+(0.143730574254)*x[2]**o+(-0.635611067664)*x[2]
        s[2,4]=(-0.212297893625)*x[0]**o+(0.631051506826)*x[0]+(-0.906149638803)*x[1]**o+(0.95431164216)*x[1]+(0.506425952055)*x[2]**o+(0.476198095725)*x[2]
        s[3,0]=(-0.572583486627)*x[0]**o+(-0.615297124684)*x[0]+(0.344317617297)*x[1]**o+(0.798009040785)*x[1]+(0.646565073237)*x[2]**o+(-0.119267772639)*x[2]
        s[3,1]=(-0.275482328004)*x[0]**o+(-0.225018438903)*x[0]+(0.648843558509)*x[1]**o+(0.971983956502)*x[1]+(0.890972764481)*x[2]**o+(-0.0457638811275)*x[2]
        s[3,2]=(0.625004124231)*x[0]**o+(-0.691087681301)*x[0]+(0.823334046124)*x[1]**o+(0.174931553897)*x[1]+(-0.783435157951)*x[2]**o+(-0.183737298781)*x[2]
        s[3,3]=(0.359619581557)*x[0]**o+(0.360159573822)*x[0]+(0.395267272408)*x[1]**o+(-0.412837344771)*x[1]+(-0.015119491493)*x[2]**o+(-0.0468017493027)*x[2]
        s[3,4]=(0.0939051154145)*x[0]**o+(-0.35421304768)*x[0]+(-0.388406902403)*x[1]**o+(-0.17940478311)*x[1]+(0.247049129844)*x[2]**o+(0.208156547444)*x[2]
        ref[0,0]=(-0.729701377063)*(1+2.*(dim-1.)/(o+1.))+(0.790402290499)*dim
        ref[0,1]=(0.352513850987)*(1+2.*(dim-1.)/(o+1.))+(-0.976698360754)*dim
        ref[0,2]=(-0.443849069543)*(1+2.*(dim-1.)/(o+1.))+(0.206581391132)*dim
        ref[0,3]=(-0.457689517223)*(1+2.*(dim-1.)/(o+1.))+(-0.673520095121)*dim
        ref[0,4]=(0.507485841104)*(1+2.*(dim-1.)/(o+1.))+(-0.258829743451)*dim
        ref[1,0]=(0.13395736761)*(1+2.*(dim-1.)/(o+1.))+(-1.52012192185)*dim
        ref[1,1]=(0.157027850598)*(1+2.*(dim-1.)/(o+1.))+(1.04590750769)*dim
        ref[1,2]=(1.86344692291)*(1+2.*(dim-1.)/(o+1.))+(1.43731969393)*dim
        ref[1,3]=(0.258987698503)*(1+2.*(dim-1.)/(o+1.))+(-0.149419530472)*dim
        ref[1,4]=(0.911098275559)*(1+2.*(dim-1.)/(o+1.))+(-1.66869642939)*dim
        ref[2,0]=(0.100848475893)*(1+2.*(dim-1.)/(o+1.))+(-0.10293202348)*dim
        ref[2,1]=(-0.229135978992)*(1+2.*(dim-1.)/(o+1.))+(1.95683674747)*dim
        ref[2,2]=(2.62691525074)*(1+2.*(dim-1.)/(o+1.))+(-1.04797525533)*dim
        ref[2,3]=(0.406171075551)*(1+2.*(dim-1.)/(o+1.))+(-0.272889047794)*dim
        ref[2,4]=(-0.612021580374)*(1+2.*(dim-1.)/(o+1.))+(2.06156124471)*dim
        ref[3,0]=(0.418299203908)*(1+2.*(dim-1.)/(o+1.))+(0.0634441434621)*dim
        ref[3,1]=(1.26433399499)*(1+2.*(dim-1.)/(o+1.))+(0.701201636471)*dim
        ref[3,2]=(0.664903012404)*(1+2.*(dim-1.)/(o+1.))+(-0.699893426186)*dim
        ref[3,3]=(0.739767362472)*(1+2.*(dim-1.)/(o+1.))+(-0.0994795202511)*dim
        ref[3,4]=(-0.0474526571443)*(1+2.*(dim-1.)/(o+1.))+(-0.325461283346)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank3(self):
      """
      tests integral of rank 3 Symbol on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.589365534484)*x[0]**o+(-0.562703733563)*x[0]+(-0.374217229188)*x[1]**o+(-0.39259467848)*x[1]
        s[0,0,1]=(-0.530861865638)*x[0]**o+(-0.452189463083)*x[0]+(0.130627913931)*x[1]**o+(0.401827715746)*x[1]
        s[0,1,0]=(-0.698034316811)*x[0]**o+(0.95858318685)*x[0]+(0.818020760864)*x[1]**o+(-0.842035209691)*x[1]
        s[0,1,1]=(0.784763132752)*x[0]**o+(-0.345161441944)*x[0]+(-0.0915389865806)*x[1]**o+(0.0413615155914)*x[1]
        s[1,0,0]=(0.844613455277)*x[0]**o+(-0.0227629508737)*x[0]+(0.435622185761)*x[1]**o+(-0.0215311374618)*x[1]
        s[1,0,1]=(-0.0669862120561)*x[0]**o+(-0.846007346441)*x[0]+(-0.413495349659)*x[1]**o+(0.535169065488)*x[1]
        s[1,1,0]=(-0.685697856883)*x[0]**o+(0.958953550286)*x[0]+(0.154980083392)*x[1]**o+(0.0800899796196)*x[1]
        s[1,1,1]=(0.912269483572)*x[0]**o+(0.94047005728)*x[0]+(-0.304144591655)*x[1]**o+(-0.953778202636)*x[1]
        s[2,0,0]=(-0.743960822491)*x[0]**o+(-0.550611814396)*x[0]+(-0.725685162705)*x[1]**o+(-0.339527538083)*x[1]
        s[2,0,1]=(-0.889607185943)*x[0]**o+(0.539329041411)*x[0]+(-0.161095005544)*x[1]**o+(0.0793331043418)*x[1]
        s[2,1,0]=(-0.468621581571)*x[0]**o+(-0.686831337537)*x[0]+(0.097676174696)*x[1]**o+(-0.661357348914)*x[1]
        s[2,1,1]=(0.272669628304)*x[0]**o+(-0.0973814107905)*x[0]+(0.934138436983)*x[1]**o+(0.886285785082)*x[1]
        s[3,0,0]=(0.732970432548)*x[0]**o+(0.867739265413)*x[0]+(-0.868701163584)*x[1]**o+(-0.845146120935)*x[1]
        s[3,0,1]=(0.655147591399)*x[0]**o+(-0.604461769582)*x[0]+(-0.855962338468)*x[1]**o+(-0.256146581175)*x[1]
        s[3,1,0]=(0.818629391238)*x[0]**o+(-0.949077886388)*x[0]+(0.620697264485)*x[1]**o+(0.161138363607)*x[1]
        s[3,1,1]=(0.388279443227)*x[0]**o+(-0.905781758355)*x[0]+(0.224661220296)*x[1]**o+(-0.160247355785)*x[1]
        s[4,0,0]=(0.933160401571)*x[0]**o+(0.238977604536)*x[0]+(-0.819630167281)*x[1]**o+(-0.787572050579)*x[1]
        s[4,0,1]=(-0.998710300262)*x[0]**o+(-0.311299346807)*x[0]+(0.01512022099)*x[1]**o+(0.987659405648)*x[1]
        s[4,1,0]=(-0.394579388382)*x[0]**o+(0.368576283401)*x[0]+(0.523842693559)*x[1]**o+(0.053237259039)*x[1]
        s[4,1,1]=(-0.899554766025)*x[0]**o+(0.0846648733918)*x[0]+(0.0384854956872)*x[1]**o+(0.499251840272)*x[1]
        s[5,0,0]=(0.969036476632)*x[0]**o+(0.555078576958)*x[0]+(0.358154612858)*x[1]**o+(0.988304896709)*x[1]
        s[5,0,1]=(-0.915157232317)*x[0]**o+(0.174863566989)*x[0]+(-0.929161107163)*x[1]**o+(-0.656086746151)*x[1]
        s[5,1,0]=(0.10810079771)*x[0]**o+(-0.929385584784)*x[0]+(-0.390809106715)*x[1]**o+(-0.573678217593)*x[1]
        s[5,1,1]=(0.871461803834)*x[0]**o+(-0.13090668343)*x[0]+(0.479502930134)*x[1]**o+(-0.143336447632)*x[1]
        ref[0,0,0]=(0.215148305297)*(1+2.*(dim-1.)/(o+1.))+(-0.955298412042)*dim
        ref[0,0,1]=(-0.400233951707)*(1+2.*(dim-1.)/(o+1.))+(-0.050361747337)*dim
        ref[0,1,0]=(0.119986444053)*(1+2.*(dim-1.)/(o+1.))+(0.116547977158)*dim
        ref[0,1,1]=(0.693224146172)*(1+2.*(dim-1.)/(o+1.))+(-0.303799926353)*dim
        ref[1,0,0]=(1.28023564104)*(1+2.*(dim-1.)/(o+1.))+(-0.0442940883355)*dim
        ref[1,0,1]=(-0.480481561715)*(1+2.*(dim-1.)/(o+1.))+(-0.310838280953)*dim
        ref[1,1,0]=(-0.530717773491)*(1+2.*(dim-1.)/(o+1.))+(1.03904352991)*dim
        ref[1,1,1]=(0.608124891917)*(1+2.*(dim-1.)/(o+1.))+(-0.0133081453554)*dim
        ref[2,0,0]=(-1.4696459852)*(1+2.*(dim-1.)/(o+1.))+(-0.890139352479)*dim
        ref[2,0,1]=(-1.05070219149)*(1+2.*(dim-1.)/(o+1.))+(0.618662145753)*dim
        ref[2,1,0]=(-0.370945406875)*(1+2.*(dim-1.)/(o+1.))+(-1.34818868645)*dim
        ref[2,1,1]=(1.20680806529)*(1+2.*(dim-1.)/(o+1.))+(0.788904374291)*dim
        ref[3,0,0]=(-0.135730731035)*(1+2.*(dim-1.)/(o+1.))+(0.0225931444774)*dim
        ref[3,0,1]=(-0.200814747068)*(1+2.*(dim-1.)/(o+1.))+(-0.860608350757)*dim
        ref[3,1,0]=(1.43932665572)*(1+2.*(dim-1.)/(o+1.))+(-0.787939522781)*dim
        ref[3,1,1]=(0.612940663523)*(1+2.*(dim-1.)/(o+1.))+(-1.06602911414)*dim
        ref[4,0,0]=(0.113530234291)*(1+2.*(dim-1.)/(o+1.))+(-0.548594446042)*dim
        ref[4,0,1]=(-0.983590079272)*(1+2.*(dim-1.)/(o+1.))+(0.676360058841)*dim
        ref[4,1,0]=(0.129263305177)*(1+2.*(dim-1.)/(o+1.))+(0.42181354244)*dim
        ref[4,1,1]=(-0.861069270338)*(1+2.*(dim-1.)/(o+1.))+(0.583916713663)*dim
        ref[5,0,0]=(1.32719108949)*(1+2.*(dim-1.)/(o+1.))+(1.54338347367)*dim
        ref[5,0,1]=(-1.84431833948)*(1+2.*(dim-1.)/(o+1.))+(-0.481223179161)*dim
        ref[5,1,0]=(-0.282708309005)*(1+2.*(dim-1.)/(o+1.))+(-1.50306380238)*dim
        ref[5,1,1]=(1.35096473397)*(1+2.*(dim-1.)/(o+1.))+(-0.274243131062)*dim
      else:
        s[0,0,0]=(0.107513098422)*x[0]**o+(0.630914018595)*x[0]+(0.223228652263)*x[1]**o+(-0.817313865632)*x[1]+(-0.131761985412)*x[2]**o+(0.765852452518)*x[2]
        s[0,0,1]=(0.677582669653)*x[0]**o+(-0.180457590671)*x[0]+(-0.623782106139)*x[1]**o+(-0.525225679629)*x[1]+(-0.506475447696)*x[2]**o+(0.426953870606)*x[2]
        s[0,1,0]=(0.638676595716)*x[0]**o+(0.401321387041)*x[0]+(0.326953813244)*x[1]**o+(0.376976149503)*x[1]+(0.37300668271)*x[2]**o+(-0.220329878781)*x[2]
        s[0,1,1]=(0.801707476852)*x[0]**o+(0.279766359037)*x[0]+(-0.119071566643)*x[1]**o+(0.842943055619)*x[1]+(-0.186390184701)*x[2]**o+(-0.174806078851)*x[2]
        s[1,0,0]=(0.636253830592)*x[0]**o+(0.0451704175721)*x[0]+(0.160762650006)*x[1]**o+(-0.520550899101)*x[1]+(0.104281196353)*x[2]**o+(-0.158693150263)*x[2]
        s[1,0,1]=(-0.529589537208)*x[0]**o+(0.883520631827)*x[0]+(-0.945287489028)*x[1]**o+(-0.402371494207)*x[1]+(-0.347776652964)*x[2]**o+(-0.801910128228)*x[2]
        s[1,1,0]=(-0.0695656121071)*x[0]**o+(0.318785505529)*x[0]+(-0.755792368009)*x[1]**o+(0.965500773902)*x[1]+(-0.963131216861)*x[2]**o+(-0.470019270105)*x[2]
        s[1,1,1]=(0.0313155860155)*x[0]**o+(-0.745224876276)*x[0]+(0.941969004994)*x[1]**o+(0.0868580544081)*x[1]+(0.814837112058)*x[2]**o+(0.6031495013)*x[2]
        s[2,0,0]=(0.326069961839)*x[0]**o+(-0.226878245433)*x[0]+(0.181649890734)*x[1]**o+(0.124231468404)*x[1]+(-0.607776937521)*x[2]**o+(-0.42430990239)*x[2]
        s[2,0,1]=(-0.0362788731491)*x[0]**o+(0.405545167692)*x[0]+(0.234108391034)*x[1]**o+(0.0613492588911)*x[1]+(0.375991408638)*x[2]**o+(-0.08364867034)*x[2]
        s[2,1,0]=(-0.739151142745)*x[0]**o+(0.895090860344)*x[0]+(-0.892542071223)*x[1]**o+(-0.925594662939)*x[1]+(0.527652868204)*x[2]**o+(-0.291230735869)*x[2]
        s[2,1,1]=(-0.569990909361)*x[0]**o+(0.135614304968)*x[0]+(0.644435688812)*x[1]**o+(0.607772804184)*x[1]+(-0.671154751626)*x[2]**o+(0.908859370499)*x[2]
        s[3,0,0]=(0.678101574906)*x[0]**o+(0.393260837804)*x[0]+(-0.364076766107)*x[1]**o+(0.221403168918)*x[1]+(-0.0881262519405)*x[2]**o+(0.243571413051)*x[2]
        s[3,0,1]=(-0.889370605049)*x[0]**o+(-0.751122823303)*x[0]+(-0.23702556553)*x[1]**o+(0.758207050746)*x[1]+(-0.653730045583)*x[2]**o+(-0.881328128574)*x[2]
        s[3,1,0]=(0.99850080224)*x[0]**o+(-0.13489113732)*x[0]+(-0.593980699882)*x[1]**o+(0.972643960626)*x[1]+(-0.0329293151824)*x[2]**o+(-0.432361021975)*x[2]
        s[3,1,1]=(-0.104833428698)*x[0]**o+(-0.937873960299)*x[0]+(-0.328611446648)*x[1]**o+(0.182782391112)*x[1]+(-0.565312252294)*x[2]**o+(0.0785805823337)*x[2]
        s[4,0,0]=(-0.833754760965)*x[0]**o+(-0.942110442063)*x[0]+(-0.544558564212)*x[1]**o+(-0.496169424902)*x[1]+(-0.805940323103)*x[2]**o+(0.355543043698)*x[2]
        s[4,0,1]=(0.1228603677)*x[0]**o+(0.27927470376)*x[0]+(0.790110425171)*x[1]**o+(0.345056958981)*x[1]+(-0.681665552406)*x[2]**o+(-0.263090945447)*x[2]
        s[4,1,0]=(0.845338153049)*x[0]**o+(0.00877377111324)*x[0]+(0.0538302301089)*x[1]**o+(-0.881754421435)*x[1]+(0.432818307915)*x[2]**o+(0.381735394491)*x[2]
        s[4,1,1]=(-0.679482650754)*x[0]**o+(0.111691609594)*x[0]+(-0.984942957975)*x[1]**o+(0.638763928487)*x[1]+(-0.146178908969)*x[2]**o+(0.752462675681)*x[2]
        s[5,0,0]=(-0.698554091901)*x[0]**o+(0.603862483307)*x[0]+(-0.627044519178)*x[1]**o+(-0.173483224843)*x[1]+(-0.794733618981)*x[2]**o+(0.148563520039)*x[2]
        s[5,0,1]=(-0.0422516673963)*x[0]**o+(0.551902135908)*x[0]+(-0.0620401959854)*x[1]**o+(-0.462654965568)*x[1]+(-0.853501426111)*x[2]**o+(-0.168566246091)*x[2]
        s[5,1,0]=(-0.482405425466)*x[0]**o+(0.215434393331)*x[0]+(0.0597508965493)*x[1]**o+(0.163372916375)*x[1]+(-0.0782540165278)*x[2]**o+(-0.29216681971)*x[2]
        s[5,1,1]=(-0.849963259001)*x[0]**o+(0.0653198724433)*x[0]+(0.673346071067)*x[1]**o+(-0.98400726868)*x[1]+(-0.545894674138)*x[2]**o+(-0.128229571203)*x[2]
        ref[0,0,0]=(0.198979765273)*(1+2.*(dim-1.)/(o+1.))+(0.579452605481)*dim
        ref[0,0,1]=(-0.452674884183)*(1+2.*(dim-1.)/(o+1.))+(-0.278729399693)*dim
        ref[0,1,0]=(1.33863709167)*(1+2.*(dim-1.)/(o+1.))+(0.557967657763)*dim
        ref[0,1,1]=(0.496245725508)*(1+2.*(dim-1.)/(o+1.))+(0.947903335805)*dim
        ref[1,0,0]=(0.901297676951)*(1+2.*(dim-1.)/(o+1.))+(-0.634073631792)*dim
        ref[1,0,1]=(-1.8226536792)*(1+2.*(dim-1.)/(o+1.))+(-0.320760990609)*dim
        ref[1,1,0]=(-1.78848919698)*(1+2.*(dim-1.)/(o+1.))+(0.814267009326)*dim
        ref[1,1,1]=(1.78812170307)*(1+2.*(dim-1.)/(o+1.))+(-0.0552173205681)*dim
        ref[2,0,0]=(-0.100057084948)*(1+2.*(dim-1.)/(o+1.))+(-0.526956679419)*dim
        ref[2,0,1]=(0.573820926522)*(1+2.*(dim-1.)/(o+1.))+(0.383245756243)*dim
        ref[2,1,0]=(-1.10404034576)*(1+2.*(dim-1.)/(o+1.))+(-0.321734538465)*dim
        ref[2,1,1]=(-0.596709972174)*(1+2.*(dim-1.)/(o+1.))+(1.65224647965)*dim
        ref[3,0,0]=(0.225898556859)*(1+2.*(dim-1.)/(o+1.))+(0.858235419772)*dim
        ref[3,0,1]=(-1.78012621616)*(1+2.*(dim-1.)/(o+1.))+(-0.874243901131)*dim
        ref[3,1,0]=(0.371590787175)*(1+2.*(dim-1.)/(o+1.))+(0.405391801331)*dim
        ref[3,1,1]=(-0.99875712764)*(1+2.*(dim-1.)/(o+1.))+(-0.676510986853)*dim
        ref[4,0,0]=(-2.18425364828)*(1+2.*(dim-1.)/(o+1.))+(-1.08273682327)*dim
        ref[4,0,1]=(0.231305240466)*(1+2.*(dim-1.)/(o+1.))+(0.361240717293)*dim
        ref[4,1,0]=(1.33198669107)*(1+2.*(dim-1.)/(o+1.))+(-0.49124525583)*dim
        ref[4,1,1]=(-1.8106045177)*(1+2.*(dim-1.)/(o+1.))+(1.50291821376)*dim
        ref[5,0,0]=(-2.12033223006)*(1+2.*(dim-1.)/(o+1.))+(0.578942778503)*dim
        ref[5,0,1]=(-0.957793289493)*(1+2.*(dim-1.)/(o+1.))+(-0.0793190757515)*dim
        ref[5,1,0]=(-0.500908545444)*(1+2.*(dim-1.)/(o+1.))+(0.0866404899966)*dim
        ref[5,1,1]=(-0.722511862072)*(1+2.*(dim-1.)/(o+1.))+(-1.04691696744)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank4(self):
      """
      tests integral of rank 4 Symbol on the FunctionOnBoundary

      assumptions: FunctionOnBoundary(self.domain) exists
                   self.domain supports integral on FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-0.938527612719)*x[0]**o+(0.18109071498)*x[0]+(0.33583450429)*x[1]**o+(0.875633806815)*x[1]
        s[0,0,0,1]=(0.235264331576)*x[0]**o+(-0.916078537073)*x[0]+(-0.564923840072)*x[1]**o+(0.241815090592)*x[1]
        s[0,0,1,0]=(0.465988219999)*x[0]**o+(-0.234861997786)*x[0]+(-0.826589503087)*x[1]**o+(-0.912695921201)*x[1]
        s[0,0,1,1]=(0.683584251749)*x[0]**o+(-0.445524319089)*x[0]+(-0.794091275716)*x[1]**o+(0.960124513284)*x[1]
        s[0,0,2,0]=(0.864832412476)*x[0]**o+(0.545257658859)*x[0]+(-0.0523175406047)*x[1]**o+(0.637801499066)*x[1]
        s[0,0,2,1]=(-0.345749442294)*x[0]**o+(-0.62275362184)*x[0]+(-0.445774008922)*x[1]**o+(-0.883708969162)*x[1]
        s[0,1,0,0]=(0.495428396131)*x[0]**o+(-0.688537966744)*x[0]+(0.974975943301)*x[1]**o+(0.812712252674)*x[1]
        s[0,1,0,1]=(0.288058971992)*x[0]**o+(0.108248232119)*x[0]+(0.717277962001)*x[1]**o+(-0.27243719684)*x[1]
        s[0,1,1,0]=(0.0359145438363)*x[0]**o+(0.536166090023)*x[0]+(-0.00355608454499)*x[1]**o+(-0.806427949542)*x[1]
        s[0,1,1,1]=(0.166147370504)*x[0]**o+(-0.627148508537)*x[0]+(-0.786840545498)*x[1]**o+(0.281613120233)*x[1]
        s[0,1,2,0]=(-0.133612757005)*x[0]**o+(0.392265237404)*x[0]+(0.453687302779)*x[1]**o+(-0.773645262693)*x[1]
        s[0,1,2,1]=(-0.716222674195)*x[0]**o+(0.670906647893)*x[0]+(-0.462562888282)*x[1]**o+(-0.624961145925)*x[1]
        s[0,2,0,0]=(0.0194565885353)*x[0]**o+(0.182692282753)*x[0]+(-0.638358622688)*x[1]**o+(0.335689393108)*x[1]
        s[0,2,0,1]=(-0.860405914558)*x[0]**o+(0.632285612174)*x[0]+(0.0656792325224)*x[1]**o+(-0.156015955877)*x[1]
        s[0,2,1,0]=(-0.152134223629)*x[0]**o+(-0.862936330258)*x[0]+(0.417270798339)*x[1]**o+(0.963136816902)*x[1]
        s[0,2,1,1]=(-0.258242824318)*x[0]**o+(0.971045381488)*x[0]+(-0.241220971892)*x[1]**o+(0.856378952246)*x[1]
        s[0,2,2,0]=(0.32072997937)*x[0]**o+(0.704100826786)*x[0]+(-0.33756098405)*x[1]**o+(-0.664910217598)*x[1]
        s[0,2,2,1]=(-0.61313065356)*x[0]**o+(0.508087982157)*x[0]+(-0.84053545881)*x[1]**o+(0.160765183388)*x[1]
        s[0,3,0,0]=(-0.75820426921)*x[0]**o+(-0.90659534133)*x[0]+(0.29703074695)*x[1]**o+(-0.0460927434434)*x[1]
        s[0,3,0,1]=(-0.168059238195)*x[0]**o+(0.444168839682)*x[0]+(-0.199109031605)*x[1]**o+(-0.0951683017552)*x[1]
        s[0,3,1,0]=(0.134237505423)*x[0]**o+(-0.793789958021)*x[0]+(-0.806546707234)*x[1]**o+(0.841229990055)*x[1]
        s[0,3,1,1]=(0.0190198296544)*x[0]**o+(-0.535272381244)*x[0]+(0.961010454009)*x[1]**o+(0.00558376325046)*x[1]
        s[0,3,2,0]=(0.363866563396)*x[0]**o+(-0.479776283958)*x[0]+(-0.189244785869)*x[1]**o+(0.807338961992)*x[1]
        s[0,3,2,1]=(0.160239702851)*x[0]**o+(-0.629633641475)*x[0]+(-0.116962193936)*x[1]**o+(-0.0779960566649)*x[1]
        s[0,4,0,0]=(-0.778019803229)*x[0]**o+(0.800046035892)*x[0]+(-0.164619354127)*x[1]**o+(-0.785540965821)*x[1]
        s[0,4,0,1]=(-0.302058711888)*x[0]**o+(0.560865269822)*x[0]+(-0.473079568526)*x[1]**o+(0.995387605735)*x[1]
        s[0,4,1,0]=(-0.411379127233)*x[0]**o+(-0.55185696669)*x[0]+(0.306097815722)*x[1]**o+(0.336660861462)*x[1]
        s[0,4,1,1]=(0.449818597104)*x[0]**o+(0.102791276029)*x[0]+(0.795639210789)*x[1]**o+(0.061474861224)*x[1]
        s[0,4,2,0]=(-0.0490155784051)*x[0]**o+(-0.429224063812)*x[0]+(-0.0368599977199)*x[1]**o+(0.718948163937)*x[1]
        s[0,4,2,1]=(-0.0195621009238)*x[0]**o+(0.162744359327)*x[0]+(0.156319239232)*x[1]**o+(-0.840307263668)*x[1]
        s[1,0,0,0]=(0.818322838345)*x[0]**o+(0.75605127429)*x[0]+(0.0160223113114)*x[1]**o+(0.733116071572)*x[1]
        s[1,0,0,1]=(-0.516143096482)*x[0]**o+(-0.607491330321)*x[0]+(-0.71990703998)*x[1]**o+(0.74041801157)*x[1]
        s[1,0,1,0]=(0.087735317248)*x[0]**o+(0.811033616338)*x[0]+(-0.596466275725)*x[1]**o+(-0.632928705475)*x[1]
        s[1,0,1,1]=(0.574370862223)*x[0]**o+(-0.666343550396)*x[0]+(0.955788950863)*x[1]**o+(-0.901731521626)*x[1]
        s[1,0,2,0]=(-0.0749607390632)*x[0]**o+(-0.431912603732)*x[0]+(-0.292096378559)*x[1]**o+(0.184727807798)*x[1]
        s[1,0,2,1]=(-0.680339079773)*x[0]**o+(-0.751178966013)*x[0]+(0.417442581996)*x[1]**o+(0.173275667315)*x[1]
        s[1,1,0,0]=(0.549976167744)*x[0]**o+(-0.542891673039)*x[0]+(0.780157563665)*x[1]**o+(-0.500193881521)*x[1]
        s[1,1,0,1]=(0.741392053681)*x[0]**o+(0.464245280846)*x[0]+(0.893822624677)*x[1]**o+(-0.726733112592)*x[1]
        s[1,1,1,0]=(0.0669632365681)*x[0]**o+(-0.19242503062)*x[0]+(-0.016747330039)*x[1]**o+(0.565528468045)*x[1]
        s[1,1,1,1]=(0.0478230922838)*x[0]**o+(-0.490882982306)*x[0]+(0.71464197649)*x[1]**o+(0.647758969096)*x[1]
        s[1,1,2,0]=(0.0459694337421)*x[0]**o+(0.22219598918)*x[0]+(0.278878633519)*x[1]**o+(0.0964919048896)*x[1]
        s[1,1,2,1]=(0.951517313537)*x[0]**o+(0.776873672124)*x[0]+(0.276827457342)*x[1]**o+(0.97564662187)*x[1]
        s[1,2,0,0]=(-0.756103809082)*x[0]**o+(-0.776671770692)*x[0]+(0.305546356471)*x[1]**o+(0.61267739505)*x[1]
        s[1,2,0,1]=(-0.00486328528121)*x[0]**o+(0.969855303757)*x[0]+(-0.803341508101)*x[1]**o+(-0.973192859371)*x[1]
        s[1,2,1,0]=(-0.233307547886)*x[0]**o+(0.480312206288)*x[0]+(-0.775060924028)*x[1]**o+(0.371008519793)*x[1]
        s[1,2,1,1]=(-0.548323703436)*x[0]**o+(-0.648506076259)*x[0]+(-0.175775778303)*x[1]**o+(0.860353396366)*x[1]
        s[1,2,2,0]=(-0.631591638879)*x[0]**o+(0.64118025586)*x[0]+(0.791112489411)*x[1]**o+(-0.503438016129)*x[1]
        s[1,2,2,1]=(-0.322064055084)*x[0]**o+(0.513203966068)*x[0]+(-0.123194392107)*x[1]**o+(-0.0530322384215)*x[1]
        s[1,3,0,0]=(0.690957542061)*x[0]**o+(0.582367141789)*x[0]+(0.0548888588807)*x[1]**o+(-0.396676526096)*x[1]
        s[1,3,0,1]=(-0.887786079331)*x[0]**o+(-0.595071924485)*x[0]+(-0.585573810513)*x[1]**o+(0.966304199567)*x[1]
        s[1,3,1,0]=(0.0813018143573)*x[0]**o+(-0.97460530802)*x[0]+(0.157284332761)*x[1]**o+(-0.282617607243)*x[1]
        s[1,3,1,1]=(0.876197676377)*x[0]**o+(-0.151980499847)*x[0]+(-0.605310496113)*x[1]**o+(-0.444245831277)*x[1]
        s[1,3,2,0]=(0.342314316145)*x[0]**o+(-0.448750580332)*x[0]+(0.42604855142)*x[1]**o+(0.646516588045)*x[1]
        s[1,3,2,1]=(0.740620038411)*x[0]**o+(-0.519019257788)*x[0]+(-0.625195631671)*x[1]**o+(0.0692954637811)*x[1]
        s[1,4,0,0]=(-0.444071318016)*x[0]**o+(-0.471544009206)*x[0]+(-0.696279794002)*x[1]**o+(-0.902974292958)*x[1]
        s[1,4,0,1]=(0.542533721359)*x[0]**o+(-0.0736449480756)*x[0]+(-0.408131697388)*x[1]**o+(0.364714013702)*x[1]
        s[1,4,1,0]=(-0.152939586181)*x[0]**o+(0.77896231429)*x[0]+(-0.815480658966)*x[1]**o+(0.154938953512)*x[1]
        s[1,4,1,1]=(-0.29007531171)*x[0]**o+(-0.352728646039)*x[0]+(0.811434496017)*x[1]**o+(0.738596503811)*x[1]
        s[1,4,2,0]=(0.219126056868)*x[0]**o+(-0.888054898232)*x[0]+(0.832367086102)*x[1]**o+(0.423282741614)*x[1]
        s[1,4,2,1]=(-0.0215001265455)*x[0]**o+(-0.267591836227)*x[0]+(0.16855162109)*x[1]**o+(0.450976714016)*x[1]
        s[2,0,0,0]=(0.55451395254)*x[0]**o+(-0.569774836527)*x[0]+(-0.950939837183)*x[1]**o+(0.963606409995)*x[1]
        s[2,0,0,1]=(0.768925257395)*x[0]**o+(-0.271693392598)*x[0]+(0.85042078878)*x[1]**o+(0.854821782113)*x[1]
        s[2,0,1,0]=(-0.0681124804964)*x[0]**o+(0.346262969724)*x[0]+(-0.224031702809)*x[1]**o+(0.184665191293)*x[1]
        s[2,0,1,1]=(0.215579225437)*x[0]**o+(0.650089014616)*x[0]+(-0.632072599272)*x[1]**o+(0.974448958372)*x[1]
        s[2,0,2,0]=(0.943481715059)*x[0]**o+(-0.91197962542)*x[0]+(-0.552378744981)*x[1]**o+(0.958940554729)*x[1]
        s[2,0,2,1]=(0.467266364031)*x[0]**o+(-0.253321061857)*x[0]+(0.935529481271)*x[1]**o+(-0.203578102406)*x[1]
        s[2,1,0,0]=(0.0885012656958)*x[0]**o+(-0.0568114758907)*x[0]+(-0.94244176099)*x[1]**o+(-0.151572915453)*x[1]
        s[2,1,0,1]=(-0.583513164141)*x[0]**o+(0.299189697674)*x[0]+(0.129798849695)*x[1]**o+(0.0486488675811)*x[1]
        s[2,1,1,0]=(-0.230047392884)*x[0]**o+(-0.215141703649)*x[0]+(0.597873677289)*x[1]**o+(-0.104893786277)*x[1]
        s[2,1,1,1]=(-0.193340204287)*x[0]**o+(0.172114988052)*x[0]+(-0.912123490435)*x[1]**o+(0.386237743741)*x[1]
        s[2,1,2,0]=(-0.355870799626)*x[0]**o+(-0.636394104532)*x[0]+(0.14161953364)*x[1]**o+(0.793881407217)*x[1]
        s[2,1,2,1]=(0.623415964738)*x[0]**o+(-0.441080834978)*x[0]+(-0.669760052979)*x[1]**o+(0.504915334197)*x[1]
        s[2,2,0,0]=(0.843403522019)*x[0]**o+(0.497098510545)*x[0]+(0.668852088742)*x[1]**o+(-0.469226675749)*x[1]
        s[2,2,0,1]=(-0.924635841352)*x[0]**o+(0.200500164001)*x[0]+(-0.230616017277)*x[1]**o+(0.758688672394)*x[1]
        s[2,2,1,0]=(-0.180287456852)*x[0]**o+(-0.189123451964)*x[0]+(-0.359118343342)*x[1]**o+(-0.856215073416)*x[1]
        s[2,2,1,1]=(-0.270882735947)*x[0]**o+(0.958913592886)*x[0]+(-0.694010263633)*x[1]**o+(0.860782778563)*x[1]
        s[2,2,2,0]=(-0.808819032861)*x[0]**o+(-0.733697428319)*x[0]+(-0.460941216629)*x[1]**o+(-0.617344596558)*x[1]
        s[2,2,2,1]=(0.451979026099)*x[0]**o+(-0.0858303927258)*x[0]+(0.163184369333)*x[1]**o+(0.801763368843)*x[1]
        s[2,3,0,0]=(-0.367005493485)*x[0]**o+(0.854411289003)*x[0]+(0.733528289773)*x[1]**o+(0.791392876143)*x[1]
        s[2,3,0,1]=(0.911579456326)*x[0]**o+(-0.690061379515)*x[0]+(-0.0831450710122)*x[1]**o+(-0.815038630875)*x[1]
        s[2,3,1,0]=(-0.179108077134)*x[0]**o+(0.609583032751)*x[0]+(-0.16765650892)*x[1]**o+(0.861847300762)*x[1]
        s[2,3,1,1]=(0.384373942853)*x[0]**o+(0.517209389691)*x[0]+(0.841154335433)*x[1]**o+(0.579356331563)*x[1]
        s[2,3,2,0]=(-0.683926277829)*x[0]**o+(0.315128536718)*x[0]+(0.394946403402)*x[1]**o+(0.679738641838)*x[1]
        s[2,3,2,1]=(0.466676304045)*x[0]**o+(-0.0400354010415)*x[0]+(-0.760865349977)*x[1]**o+(0.928180066121)*x[1]
        s[2,4,0,0]=(-0.497221654148)*x[0]**o+(-0.921245715417)*x[0]+(0.908610543692)*x[1]**o+(-0.662743777314)*x[1]
        s[2,4,0,1]=(-0.218025628376)*x[0]**o+(0.981616175697)*x[0]+(-0.300066300268)*x[1]**o+(-0.145065054659)*x[1]
        s[2,4,1,0]=(0.822321927296)*x[0]**o+(0.697029639265)*x[0]+(0.21672820186)*x[1]**o+(-0.136979319447)*x[1]
        s[2,4,1,1]=(-0.33254584204)*x[0]**o+(0.0804948126216)*x[0]+(0.931652185323)*x[1]**o+(-0.206973479206)*x[1]
        s[2,4,2,0]=(-0.20645245153)*x[0]**o+(0.183027858308)*x[0]+(-0.889749226799)*x[1]**o+(0.177254293812)*x[1]
        s[2,4,2,1]=(0.67936145385)*x[0]**o+(0.0269414327289)*x[0]+(-0.892654968646)*x[1]**o+(-0.211494218552)*x[1]
        s[3,0,0,0]=(-0.221318514547)*x[0]**o+(0.282480830569)*x[0]+(0.428045462357)*x[1]**o+(0.6335381978)*x[1]
        s[3,0,0,1]=(-0.052602851764)*x[0]**o+(-0.208342592426)*x[0]+(-0.518314637297)*x[1]**o+(0.0538977682478)*x[1]
        s[3,0,1,0]=(-0.879184433575)*x[0]**o+(-0.413590155319)*x[0]+(-0.973398939671)*x[1]**o+(0.880259070216)*x[1]
        s[3,0,1,1]=(-0.160095692245)*x[0]**o+(-0.215405690187)*x[0]+(0.049417328483)*x[1]**o+(0.755528105808)*x[1]
        s[3,0,2,0]=(0.0775811014995)*x[0]**o+(0.574331743034)*x[0]+(0.656316191524)*x[1]**o+(-0.485245137216)*x[1]
        s[3,0,2,1]=(0.235819746353)*x[0]**o+(-0.0212642690609)*x[0]+(0.519388187993)*x[1]**o+(-0.73686111506)*x[1]
        s[3,1,0,0]=(0.895857350075)*x[0]**o+(0.993441070954)*x[0]+(0.423289404334)*x[1]**o+(0.648503660982)*x[1]
        s[3,1,0,1]=(-0.508341685048)*x[0]**o+(0.132799887315)*x[0]+(-0.615279575964)*x[1]**o+(0.00296513970359)*x[1]
        s[3,1,1,0]=(0.691638662949)*x[0]**o+(-0.432982022284)*x[0]+(-0.724535941079)*x[1]**o+(0.0336147554659)*x[1]
        s[3,1,1,1]=(-0.431062234755)*x[0]**o+(-0.795503344318)*x[0]+(0.302222798071)*x[1]**o+(-0.0813627256685)*x[1]
        s[3,1,2,0]=(-0.695560929102)*x[0]**o+(0.328886265482)*x[0]+(-0.187761431398)*x[1]**o+(-0.997772947746)*x[1]
        s[3,1,2,1]=(0.551492819689)*x[0]**o+(0.00738597993558)*x[0]+(-0.920251030964)*x[1]**o+(-0.209623547453)*x[1]
        s[3,2,0,0]=(-0.945715173023)*x[0]**o+(-0.413860607376)*x[0]+(-0.0065717404946)*x[1]**o+(-0.340070315229)*x[1]
        s[3,2,0,1]=(-0.866814974957)*x[0]**o+(0.190612129778)*x[0]+(-0.323173980343)*x[1]**o+(-0.221048196515)*x[1]
        s[3,2,1,0]=(0.19860994374)*x[0]**o+(-0.521137863979)*x[0]+(-0.347340372612)*x[1]**o+(0.85944792998)*x[1]
        s[3,2,1,1]=(-0.128916019272)*x[0]**o+(-0.850906742746)*x[0]+(0.258631463969)*x[1]**o+(-0.201200252475)*x[1]
        s[3,2,2,0]=(-0.330614479011)*x[0]**o+(-0.489889065789)*x[0]+(-0.244206440351)*x[1]**o+(-0.322097122215)*x[1]
        s[3,2,2,1]=(0.711726054557)*x[0]**o+(0.63375724563)*x[0]+(-0.564782988704)*x[1]**o+(-0.110402977102)*x[1]
        s[3,3,0,0]=(0.994364369344)*x[0]**o+(-0.112814477225)*x[0]+(-0.277853585926)*x[1]**o+(0.952622614986)*x[1]
        s[3,3,0,1]=(-0.4170514524)*x[0]**o+(0.890841357027)*x[0]+(-0.821412916413)*x[1]**o+(0.795074763635)*x[1]
        s[3,3,1,0]=(-0.0247320805051)*x[0]**o+(-0.204605869304)*x[0]+(0.24262118978)*x[1]**o+(0.546548720614)*x[1]
        s[3,3,1,1]=(-0.474740509006)*x[0]**o+(-0.350277551474)*x[0]+(0.596007930161)*x[1]**o+(0.310969198868)*x[1]
        s[3,3,2,0]=(0.132049782738)*x[0]**o+(-0.913784955142)*x[0]+(-0.189796554115)*x[1]**o+(-0.0703212084825)*x[1]
        s[3,3,2,1]=(-0.831318298996)*x[0]**o+(0.759034430202)*x[0]+(-0.642082491468)*x[1]**o+(-0.0998337089515)*x[1]
        s[3,4,0,0]=(-0.947142357256)*x[0]**o+(-0.358864598356)*x[0]+(0.451923869609)*x[1]**o+(0.46347113305)*x[1]
        s[3,4,0,1]=(-0.935820217362)*x[0]**o+(-0.896350370731)*x[0]+(0.939373432056)*x[1]**o+(-0.367350815379)*x[1]
        s[3,4,1,0]=(0.217367733064)*x[0]**o+(-0.348603185172)*x[0]+(-0.846422979973)*x[1]**o+(0.466060007646)*x[1]
        s[3,4,1,1]=(-0.0861518225018)*x[0]**o+(0.129106551242)*x[0]+(-0.563717320923)*x[1]**o+(0.488814871115)*x[1]
        s[3,4,2,0]=(0.864238935579)*x[0]**o+(0.451518233825)*x[0]+(0.515184766232)*x[1]**o+(0.791632366311)*x[1]
        s[3,4,2,1]=(0.226031804126)*x[0]**o+(0.703010170609)*x[0]+(-0.694425471794)*x[1]**o+(-0.398984702926)*x[1]
        ref[0,0,0,0]=(-0.602693108429)*(1+2.*(dim-1.)/(o+1.))+(1.05672452179)*dim
        ref[0,0,0,1]=(-0.329659508496)*(1+2.*(dim-1.)/(o+1.))+(-0.674263446481)*dim
        ref[0,0,1,0]=(-0.360601283088)*(1+2.*(dim-1.)/(o+1.))+(-1.14755791899)*dim
        ref[0,0,1,1]=(-0.110507023967)*(1+2.*(dim-1.)/(o+1.))+(0.514600194196)*dim
        ref[0,0,2,0]=(0.812514871871)*(1+2.*(dim-1.)/(o+1.))+(1.18305915792)*dim
        ref[0,0,2,1]=(-0.791523451216)*(1+2.*(dim-1.)/(o+1.))+(-1.506462591)*dim
        ref[0,1,0,0]=(1.47040433943)*(1+2.*(dim-1.)/(o+1.))+(0.12417428593)*dim
        ref[0,1,0,1]=(1.00533693399)*(1+2.*(dim-1.)/(o+1.))+(-0.164188964721)*dim
        ref[0,1,1,0]=(0.0323584592913)*(1+2.*(dim-1.)/(o+1.))+(-0.270261859519)*dim
        ref[0,1,1,1]=(-0.620693174994)*(1+2.*(dim-1.)/(o+1.))+(-0.345535388304)*dim
        ref[0,1,2,0]=(0.320074545773)*(1+2.*(dim-1.)/(o+1.))+(-0.381380025289)*dim
        ref[0,1,2,1]=(-1.17878556248)*(1+2.*(dim-1.)/(o+1.))+(0.0459455019681)*dim
        ref[0,2,0,0]=(-0.618902034153)*(1+2.*(dim-1.)/(o+1.))+(0.518381675861)*dim
        ref[0,2,0,1]=(-0.794726682036)*(1+2.*(dim-1.)/(o+1.))+(0.476269656297)*dim
        ref[0,2,1,0]=(0.26513657471)*(1+2.*(dim-1.)/(o+1.))+(0.100200486644)*dim
        ref[0,2,1,1]=(-0.49946379621)*(1+2.*(dim-1.)/(o+1.))+(1.82742433373)*dim
        ref[0,2,2,0]=(-0.01683100468)*(1+2.*(dim-1.)/(o+1.))+(0.0391906091876)*dim
        ref[0,2,2,1]=(-1.45366611237)*(1+2.*(dim-1.)/(o+1.))+(0.668853165545)*dim
        ref[0,3,0,0]=(-0.46117352226)*(1+2.*(dim-1.)/(o+1.))+(-0.952688084773)*dim
        ref[0,3,0,1]=(-0.3671682698)*(1+2.*(dim-1.)/(o+1.))+(0.349000537927)*dim
        ref[0,3,1,0]=(-0.672309201811)*(1+2.*(dim-1.)/(o+1.))+(0.0474400320344)*dim
        ref[0,3,1,1]=(0.980030283663)*(1+2.*(dim-1.)/(o+1.))+(-0.529688617993)*dim
        ref[0,3,2,0]=(0.174621777527)*(1+2.*(dim-1.)/(o+1.))+(0.327562678035)*dim
        ref[0,3,2,1]=(0.0432775089146)*(1+2.*(dim-1.)/(o+1.))+(-0.707629698139)*dim
        ref[0,4,0,0]=(-0.942639157356)*(1+2.*(dim-1.)/(o+1.))+(0.0145050700714)*dim
        ref[0,4,0,1]=(-0.775138280415)*(1+2.*(dim-1.)/(o+1.))+(1.55625287556)*dim
        ref[0,4,1,0]=(-0.105281311512)*(1+2.*(dim-1.)/(o+1.))+(-0.215196105228)*dim
        ref[0,4,1,1]=(1.24545780789)*(1+2.*(dim-1.)/(o+1.))+(0.164266137253)*dim
        ref[0,4,2,0]=(-0.0858755761251)*(1+2.*(dim-1.)/(o+1.))+(0.289724100124)*dim
        ref[0,4,2,1]=(0.136757138309)*(1+2.*(dim-1.)/(o+1.))+(-0.677562904341)*dim
        ref[1,0,0,0]=(0.834345149656)*(1+2.*(dim-1.)/(o+1.))+(1.48916734586)*dim
        ref[1,0,0,1]=(-1.23605013646)*(1+2.*(dim-1.)/(o+1.))+(0.132926681249)*dim
        ref[1,0,1,0]=(-0.508730958477)*(1+2.*(dim-1.)/(o+1.))+(0.178104910863)*dim
        ref[1,0,1,1]=(1.53015981309)*(1+2.*(dim-1.)/(o+1.))+(-1.56807507202)*dim
        ref[1,0,2,0]=(-0.367057117622)*(1+2.*(dim-1.)/(o+1.))+(-0.247184795934)*dim
        ref[1,0,2,1]=(-0.262896497777)*(1+2.*(dim-1.)/(o+1.))+(-0.577903298697)*dim
        ref[1,1,0,0]=(1.33013373141)*(1+2.*(dim-1.)/(o+1.))+(-1.04308555456)*dim
        ref[1,1,0,1]=(1.63521467836)*(1+2.*(dim-1.)/(o+1.))+(-0.262487831746)*dim
        ref[1,1,1,0]=(0.0502159065291)*(1+2.*(dim-1.)/(o+1.))+(0.373103437426)*dim
        ref[1,1,1,1]=(0.762465068774)*(1+2.*(dim-1.)/(o+1.))+(0.15687598679)*dim
        ref[1,1,2,0]=(0.324848067261)*(1+2.*(dim-1.)/(o+1.))+(0.318687894069)*dim
        ref[1,1,2,1]=(1.22834477088)*(1+2.*(dim-1.)/(o+1.))+(1.75252029399)*dim
        ref[1,2,0,0]=(-0.450557452611)*(1+2.*(dim-1.)/(o+1.))+(-0.163994375642)*dim
        ref[1,2,0,1]=(-0.808204793382)*(1+2.*(dim-1.)/(o+1.))+(-0.00333755561452)*dim
        ref[1,2,1,0]=(-1.00836847191)*(1+2.*(dim-1.)/(o+1.))+(0.85132072608)*dim
        ref[1,2,1,1]=(-0.724099481739)*(1+2.*(dim-1.)/(o+1.))+(0.211847320106)*dim
        ref[1,2,2,0]=(0.159520850532)*(1+2.*(dim-1.)/(o+1.))+(0.137742239731)*dim
        ref[1,2,2,1]=(-0.44525844719)*(1+2.*(dim-1.)/(o+1.))+(0.460171727646)*dim
        ref[1,3,0,0]=(0.745846400942)*(1+2.*(dim-1.)/(o+1.))+(0.185690615693)*dim
        ref[1,3,0,1]=(-1.47335988984)*(1+2.*(dim-1.)/(o+1.))+(0.371232275082)*dim
        ref[1,3,1,0]=(0.238586147118)*(1+2.*(dim-1.)/(o+1.))+(-1.25722291526)*dim
        ref[1,3,1,1]=(0.270887180264)*(1+2.*(dim-1.)/(o+1.))+(-0.596226331125)*dim
        ref[1,3,2,0]=(0.768362867564)*(1+2.*(dim-1.)/(o+1.))+(0.197766007713)*dim
        ref[1,3,2,1]=(0.11542440674)*(1+2.*(dim-1.)/(o+1.))+(-0.449723794007)*dim
        ref[1,4,0,0]=(-1.14035111202)*(1+2.*(dim-1.)/(o+1.))+(-1.37451830216)*dim
        ref[1,4,0,1]=(0.134402023971)*(1+2.*(dim-1.)/(o+1.))+(0.291069065626)*dim
        ref[1,4,1,0]=(-0.968420245147)*(1+2.*(dim-1.)/(o+1.))+(0.933901267802)*dim
        ref[1,4,1,1]=(0.521359184306)*(1+2.*(dim-1.)/(o+1.))+(0.385867857771)*dim
        ref[1,4,2,0]=(1.05149314297)*(1+2.*(dim-1.)/(o+1.))+(-0.464772156618)*dim
        ref[1,4,2,1]=(0.147051494544)*(1+2.*(dim-1.)/(o+1.))+(0.183384877789)*dim
        ref[2,0,0,0]=(-0.396425884642)*(1+2.*(dim-1.)/(o+1.))+(0.393831573468)*dim
        ref[2,0,0,1]=(1.61934604618)*(1+2.*(dim-1.)/(o+1.))+(0.583128389514)*dim
        ref[2,0,1,0]=(-0.292144183305)*(1+2.*(dim-1.)/(o+1.))+(0.530928161018)*dim
        ref[2,0,1,1]=(-0.416493373835)*(1+2.*(dim-1.)/(o+1.))+(1.62453797299)*dim
        ref[2,0,2,0]=(0.391102970078)*(1+2.*(dim-1.)/(o+1.))+(0.0469609293089)*dim
        ref[2,0,2,1]=(1.4027958453)*(1+2.*(dim-1.)/(o+1.))+(-0.456899164263)*dim
        ref[2,1,0,0]=(-0.853940495294)*(1+2.*(dim-1.)/(o+1.))+(-0.208384391343)*dim
        ref[2,1,0,1]=(-0.453714314446)*(1+2.*(dim-1.)/(o+1.))+(0.347838565255)*dim
        ref[2,1,1,0]=(0.367826284405)*(1+2.*(dim-1.)/(o+1.))+(-0.320035489926)*dim
        ref[2,1,1,1]=(-1.10546369472)*(1+2.*(dim-1.)/(o+1.))+(0.558352731793)*dim
        ref[2,1,2,0]=(-0.214251265986)*(1+2.*(dim-1.)/(o+1.))+(0.157487302684)*dim
        ref[2,1,2,1]=(-0.046344088241)*(1+2.*(dim-1.)/(o+1.))+(0.0638344992189)*dim
        ref[2,2,0,0]=(1.51225561076)*(1+2.*(dim-1.)/(o+1.))+(0.0278718347964)*dim
        ref[2,2,0,1]=(-1.15525185863)*(1+2.*(dim-1.)/(o+1.))+(0.959188836395)*dim
        ref[2,2,1,0]=(-0.539405800194)*(1+2.*(dim-1.)/(o+1.))+(-1.04533852538)*dim
        ref[2,2,1,1]=(-0.96489299958)*(1+2.*(dim-1.)/(o+1.))+(1.81969637145)*dim
        ref[2,2,2,0]=(-1.26976024949)*(1+2.*(dim-1.)/(o+1.))+(-1.35104202488)*dim
        ref[2,2,2,1]=(0.615163395432)*(1+2.*(dim-1.)/(o+1.))+(0.715932976118)*dim
        ref[2,3,0,0]=(0.366522796287)*(1+2.*(dim-1.)/(o+1.))+(1.64580416515)*dim
        ref[2,3,0,1]=(0.828434385314)*(1+2.*(dim-1.)/(o+1.))+(-1.50510001039)*dim
        ref[2,3,1,0]=(-0.346764586054)*(1+2.*(dim-1.)/(o+1.))+(1.47143033351)*dim
        ref[2,3,1,1]=(1.22552827829)*(1+2.*(dim-1.)/(o+1.))+(1.09656572125)*dim
        ref[2,3,2,0]=(-0.288979874427)*(1+2.*(dim-1.)/(o+1.))+(0.994867178556)*dim
        ref[2,3,2,1]=(-0.294189045932)*(1+2.*(dim-1.)/(o+1.))+(0.88814466508)*dim
        ref[2,4,0,0]=(0.411388889544)*(1+2.*(dim-1.)/(o+1.))+(-1.58398949273)*dim
        ref[2,4,0,1]=(-0.518091928645)*(1+2.*(dim-1.)/(o+1.))+(0.836551121038)*dim
        ref[2,4,1,0]=(1.03905012916)*(1+2.*(dim-1.)/(o+1.))+(0.560050319818)*dim
        ref[2,4,1,1]=(0.599106343283)*(1+2.*(dim-1.)/(o+1.))+(-0.126478666584)*dim
        ref[2,4,2,0]=(-1.09620167833)*(1+2.*(dim-1.)/(o+1.))+(0.36028215212)*dim
        ref[2,4,2,1]=(-0.213293514797)*(1+2.*(dim-1.)/(o+1.))+(-0.184552785823)*dim
        ref[3,0,0,0]=(0.20672694781)*(1+2.*(dim-1.)/(o+1.))+(0.91601902837)*dim
        ref[3,0,0,1]=(-0.570917489061)*(1+2.*(dim-1.)/(o+1.))+(-0.154444824179)*dim
        ref[3,0,1,0]=(-1.85258337325)*(1+2.*(dim-1.)/(o+1.))+(0.466668914897)*dim
        ref[3,0,1,1]=(-0.110678363762)*(1+2.*(dim-1.)/(o+1.))+(0.540122415621)*dim
        ref[3,0,2,0]=(0.733897293024)*(1+2.*(dim-1.)/(o+1.))+(0.0890866058179)*dim
        ref[3,0,2,1]=(0.755207934346)*(1+2.*(dim-1.)/(o+1.))+(-0.758125384121)*dim
        ref[3,1,0,0]=(1.31914675441)*(1+2.*(dim-1.)/(o+1.))+(1.64194473194)*dim
        ref[3,1,0,1]=(-1.12362126101)*(1+2.*(dim-1.)/(o+1.))+(0.135765027019)*dim
        ref[3,1,1,0]=(-0.0328972781304)*(1+2.*(dim-1.)/(o+1.))+(-0.399367266818)*dim
        ref[3,1,1,1]=(-0.128839436684)*(1+2.*(dim-1.)/(o+1.))+(-0.876866069986)*dim
        ref[3,1,2,0]=(-0.8833223605)*(1+2.*(dim-1.)/(o+1.))+(-0.668886682264)*dim
        ref[3,1,2,1]=(-0.368758211275)*(1+2.*(dim-1.)/(o+1.))+(-0.202237567517)*dim
        ref[3,2,0,0]=(-0.952286913517)*(1+2.*(dim-1.)/(o+1.))+(-0.753930922605)*dim
        ref[3,2,0,1]=(-1.1899889553)*(1+2.*(dim-1.)/(o+1.))+(-0.030436066737)*dim
        ref[3,2,1,0]=(-0.148730428873)*(1+2.*(dim-1.)/(o+1.))+(0.338310066001)*dim
        ref[3,2,1,1]=(0.129715444697)*(1+2.*(dim-1.)/(o+1.))+(-1.05210699522)*dim
        ref[3,2,2,0]=(-0.574820919362)*(1+2.*(dim-1.)/(o+1.))+(-0.811986188004)*dim
        ref[3,2,2,1]=(0.146943065853)*(1+2.*(dim-1.)/(o+1.))+(0.523354268528)*dim
        ref[3,3,0,0]=(0.716510783419)*(1+2.*(dim-1.)/(o+1.))+(0.839808137761)*dim
        ref[3,3,0,1]=(-1.23846436881)*(1+2.*(dim-1.)/(o+1.))+(1.68591612066)*dim
        ref[3,3,1,0]=(0.217889109275)*(1+2.*(dim-1.)/(o+1.))+(0.34194285131)*dim
        ref[3,3,1,1]=(0.121267421155)*(1+2.*(dim-1.)/(o+1.))+(-0.0393083526063)*dim
        ref[3,3,2,0]=(-0.0577467713773)*(1+2.*(dim-1.)/(o+1.))+(-0.984106163624)*dim
        ref[3,3,2,1]=(-1.47340079046)*(1+2.*(dim-1.)/(o+1.))+(0.65920072125)*dim
        ref[3,4,0,0]=(-0.495218487647)*(1+2.*(dim-1.)/(o+1.))+(0.104606534694)*dim
        ref[3,4,0,1]=(0.00355321469463)*(1+2.*(dim-1.)/(o+1.))+(-1.26370118611)*dim
        ref[3,4,1,0]=(-0.629055246908)*(1+2.*(dim-1.)/(o+1.))+(0.117456822474)*dim
        ref[3,4,1,1]=(-0.649869143425)*(1+2.*(dim-1.)/(o+1.))+(0.617921422357)*dim
        ref[3,4,2,0]=(1.37942370181)*(1+2.*(dim-1.)/(o+1.))+(1.24315060014)*dim
        ref[3,4,2,1]=(-0.468393667668)*(1+2.*(dim-1.)/(o+1.))+(0.304025467684)*dim
      else:
        s[0,0,0,0]=(0.265353968453)*x[0]**o+(-0.988400050542)*x[0]+(0.915581496138)*x[1]**o+(-0.532717689852)*x[1]+(-0.326369232123)*x[2]**o+(-0.751181530498)*x[2]
        s[0,0,0,1]=(-0.700652623793)*x[0]**o+(0.385185933969)*x[0]+(0.213605768081)*x[1]**o+(-0.626286949699)*x[1]+(-0.502285730612)*x[2]**o+(-0.510666455487)*x[2]
        s[0,0,1,0]=(0.662622398823)*x[0]**o+(0.682570980418)*x[0]+(-0.888940566008)*x[1]**o+(0.917775368899)*x[1]+(0.397316182924)*x[2]**o+(0.470957649176)*x[2]
        s[0,0,1,1]=(0.506440820988)*x[0]**o+(0.548696583948)*x[0]+(-0.343755749097)*x[1]**o+(0.635493528274)*x[1]+(0.96169367604)*x[2]**o+(-0.814595321216)*x[2]
        s[0,0,2,0]=(0.0897565668909)*x[0]**o+(-0.707489384492)*x[0]+(0.852406382045)*x[1]**o+(0.264605905284)*x[1]+(-0.76680872637)*x[2]**o+(-0.607486929359)*x[2]
        s[0,0,2,1]=(-0.491708830409)*x[0]**o+(-0.917156620819)*x[0]+(0.138490412136)*x[1]**o+(0.445788359479)*x[1]+(0.660215675311)*x[2]**o+(-0.124958885809)*x[2]
        s[0,1,0,0]=(0.252227873704)*x[0]**o+(-0.268708653921)*x[0]+(-0.735320913237)*x[1]**o+(0.832659354382)*x[1]+(-0.503570989065)*x[2]**o+(-0.416695426775)*x[2]
        s[0,1,0,1]=(0.992266568089)*x[0]**o+(-0.553995410037)*x[0]+(-0.94305015981)*x[1]**o+(-0.277660590986)*x[1]+(-0.106619424885)*x[2]**o+(-0.641996454178)*x[2]
        s[0,1,1,0]=(0.346537921137)*x[0]**o+(-0.752921521456)*x[0]+(0.697296847348)*x[1]**o+(0.127827995978)*x[1]+(-0.024787761696)*x[2]**o+(-0.132279656336)*x[2]
        s[0,1,1,1]=(0.838216254144)*x[0]**o+(-0.0488544208848)*x[0]+(0.618848880659)*x[1]**o+(-0.962908419156)*x[1]+(-0.67378264273)*x[2]**o+(-0.206991847936)*x[2]
        s[0,1,2,0]=(0.41676085558)*x[0]**o+(-0.214807189212)*x[0]+(0.291507151171)*x[1]**o+(0.668437674897)*x[1]+(0.244072891194)*x[2]**o+(0.259834097546)*x[2]
        s[0,1,2,1]=(-0.181271253302)*x[0]**o+(-0.120474808334)*x[0]+(0.996200217415)*x[1]**o+(0.964447373929)*x[1]+(0.642033792194)*x[2]**o+(0.737708747193)*x[2]
        s[0,2,0,0]=(-0.792103621996)*x[0]**o+(-0.688217283375)*x[0]+(0.672498785205)*x[1]**o+(-0.653725451669)*x[1]+(0.184442483292)*x[2]**o+(-0.532250028625)*x[2]
        s[0,2,0,1]=(-0.535900645063)*x[0]**o+(0.910701730932)*x[0]+(-0.942492793063)*x[1]**o+(0.840681049537)*x[1]+(0.717504040832)*x[2]**o+(0.325490269912)*x[2]
        s[0,2,1,0]=(-0.337864688224)*x[0]**o+(0.746375970396)*x[0]+(0.747732580426)*x[1]**o+(-0.282770683276)*x[1]+(-0.790443829627)*x[2]**o+(0.719617282434)*x[2]
        s[0,2,1,1]=(0.288927362375)*x[0]**o+(-0.11015159228)*x[0]+(-0.840599795558)*x[1]**o+(0.400589417352)*x[1]+(0.651562852697)*x[2]**o+(0.472103872054)*x[2]
        s[0,2,2,0]=(0.759453740866)*x[0]**o+(0.764211886443)*x[0]+(-0.596686471329)*x[1]**o+(-0.305500139463)*x[1]+(0.644132204372)*x[2]**o+(0.409169518907)*x[2]
        s[0,2,2,1]=(0.310393680205)*x[0]**o+(-0.55096306264)*x[0]+(-0.0418723645688)*x[1]**o+(0.146737303429)*x[1]+(-0.221068515058)*x[2]**o+(0.347847681074)*x[2]
        s[0,3,0,0]=(-0.121018382943)*x[0]**o+(-0.828966827969)*x[0]+(-0.579574647427)*x[1]**o+(0.505540315546)*x[1]+(0.846008555599)*x[2]**o+(0.722616724538)*x[2]
        s[0,3,0,1]=(0.764966634503)*x[0]**o+(0.0406308216785)*x[0]+(0.627462444709)*x[1]**o+(-0.458074562731)*x[1]+(0.868715029031)*x[2]**o+(-0.326046950875)*x[2]
        s[0,3,1,0]=(0.88733222894)*x[0]**o+(0.199875300475)*x[0]+(0.929525862041)*x[1]**o+(0.280705926018)*x[1]+(-0.85604411697)*x[2]**o+(-0.509002674462)*x[2]
        s[0,3,1,1]=(0.178137426708)*x[0]**o+(0.0040475349026)*x[0]+(0.606739420029)*x[1]**o+(-0.399120462352)*x[1]+(0.688376909613)*x[2]**o+(-0.204172031563)*x[2]
        s[0,3,2,0]=(-0.55097942154)*x[0]**o+(0.518240585643)*x[0]+(0.809100524171)*x[1]**o+(0.309535824824)*x[1]+(0.824892877333)*x[2]**o+(-0.580350157914)*x[2]
        s[0,3,2,1]=(-0.429643456007)*x[0]**o+(0.683690075983)*x[0]+(-0.474187287962)*x[1]**o+(-0.641084362835)*x[1]+(-0.514804055536)*x[2]**o+(0.653494101426)*x[2]
        s[0,4,0,0]=(0.759632202933)*x[0]**o+(-0.953554006806)*x[0]+(-0.564086866677)*x[1]**o+(-0.78087231593)*x[1]+(-0.589063170649)*x[2]**o+(-0.569522356624)*x[2]
        s[0,4,0,1]=(-0.440639409356)*x[0]**o+(-0.887245582242)*x[0]+(-0.731777330438)*x[1]**o+(0.678481115368)*x[1]+(-0.234694289692)*x[2]**o+(-0.481669499353)*x[2]
        s[0,4,1,0]=(0.421338590919)*x[0]**o+(-0.710549762367)*x[0]+(-0.831562216016)*x[1]**o+(0.526145415136)*x[1]+(-0.228391513851)*x[2]**o+(-0.287603858419)*x[2]
        s[0,4,1,1]=(-0.394922177529)*x[0]**o+(-0.749933015384)*x[0]+(0.501851406608)*x[1]**o+(0.414942993798)*x[1]+(0.882492276774)*x[2]**o+(-0.417451715702)*x[2]
        s[0,4,2,0]=(-0.708380417114)*x[0]**o+(0.272117639041)*x[0]+(-0.914251681965)*x[1]**o+(-0.692698068557)*x[1]+(0.740875968835)*x[2]**o+(0.0816142773216)*x[2]
        s[0,4,2,1]=(-0.781489599785)*x[0]**o+(-0.688587387866)*x[0]+(-0.8249818093)*x[1]**o+(0.479876441134)*x[1]+(-0.726425081108)*x[2]**o+(0.467932761023)*x[2]
        s[1,0,0,0]=(0.342121912006)*x[0]**o+(0.136682990021)*x[0]+(0.181997624536)*x[1]**o+(0.168026470492)*x[1]+(0.0115668517645)*x[2]**o+(-0.0377932582246)*x[2]
        s[1,0,0,1]=(-0.377080344591)*x[0]**o+(0.874997839577)*x[0]+(0.998952980553)*x[1]**o+(-0.511330536046)*x[1]+(0.845159267663)*x[2]**o+(0.0349344413329)*x[2]
        s[1,0,1,0]=(0.376824814933)*x[0]**o+(-0.58697348357)*x[0]+(0.764173288619)*x[1]**o+(0.288509758241)*x[1]+(-0.818987335991)*x[2]**o+(0.433785438733)*x[2]
        s[1,0,1,1]=(-0.923417677759)*x[0]**o+(0.703237676923)*x[0]+(-0.0963613139466)*x[1]**o+(0.648863487662)*x[1]+(0.121038690293)*x[2]**o+(-0.428707341383)*x[2]
        s[1,0,2,0]=(-0.192698196902)*x[0]**o+(-0.45684016542)*x[0]+(-0.157529151044)*x[1]**o+(0.665508784857)*x[1]+(-0.0263995814254)*x[2]**o+(-0.711087191185)*x[2]
        s[1,0,2,1]=(-0.262702009507)*x[0]**o+(0.612242386787)*x[0]+(-0.0527177865241)*x[1]**o+(-0.946597487364)*x[1]+(0.683034847912)*x[2]**o+(0.447810652774)*x[2]
        s[1,1,0,0]=(0.0954675694985)*x[0]**o+(-0.692307292788)*x[0]+(0.682808081853)*x[1]**o+(-0.602155379721)*x[1]+(-0.804568326115)*x[2]**o+(0.6383976993)*x[2]
        s[1,1,0,1]=(-0.776359536583)*x[0]**o+(0.593053796057)*x[0]+(-0.560010245091)*x[1]**o+(-0.0167025681078)*x[1]+(0.189236205132)*x[2]**o+(0.625263059129)*x[2]
        s[1,1,1,0]=(-0.294839786045)*x[0]**o+(0.318602896219)*x[0]+(0.790859996391)*x[1]**o+(-0.100239676238)*x[1]+(-0.752119733912)*x[2]**o+(0.909827281572)*x[2]
        s[1,1,1,1]=(0.427920987151)*x[0]**o+(0.0579828435662)*x[0]+(-0.805648184456)*x[1]**o+(0.545638930222)*x[1]+(-0.333821427396)*x[2]**o+(0.489029035079)*x[2]
        s[1,1,2,0]=(-0.102211826161)*x[0]**o+(-0.23589775753)*x[0]+(0.194344132247)*x[1]**o+(0.216623239742)*x[1]+(-0.328848234996)*x[2]**o+(-0.35246467614)*x[2]
        s[1,1,2,1]=(-0.597326323585)*x[0]**o+(-0.727893394493)*x[0]+(-0.219786140301)*x[1]**o+(-0.261271587677)*x[1]+(0.433015479844)*x[2]**o+(0.480572847641)*x[2]
        s[1,2,0,0]=(0.978638976517)*x[0]**o+(0.252195213856)*x[0]+(-0.534544050017)*x[1]**o+(-0.0719497516536)*x[1]+(0.0756026756135)*x[2]**o+(-0.376095264735)*x[2]
        s[1,2,0,1]=(-0.592591339778)*x[0]**o+(-0.87353085535)*x[0]+(0.226804211947)*x[1]**o+(0.257318409091)*x[1]+(0.820507252476)*x[2]**o+(0.490297079988)*x[2]
        s[1,2,1,0]=(0.157573169077)*x[0]**o+(0.128029582112)*x[0]+(-0.954106083921)*x[1]**o+(0.509373015197)*x[1]+(0.788827839704)*x[2]**o+(0.717054840575)*x[2]
        s[1,2,1,1]=(-0.47425212595)*x[0]**o+(-0.608889999781)*x[0]+(-0.334194651773)*x[1]**o+(-0.439282037399)*x[1]+(-0.169103982018)*x[2]**o+(-0.569343903254)*x[2]
        s[1,2,2,0]=(0.686390790562)*x[0]**o+(0.670565132214)*x[0]+(-0.491526583908)*x[1]**o+(-0.887969445154)*x[1]+(0.666707124541)*x[2]**o+(-0.643899609236)*x[2]
        s[1,2,2,1]=(0.158400774071)*x[0]**o+(0.672487819329)*x[0]+(-0.397304779942)*x[1]**o+(0.454998380713)*x[1]+(-0.571811342108)*x[2]**o+(0.294069278512)*x[2]
        s[1,3,0,0]=(-0.652264634829)*x[0]**o+(-0.676541910532)*x[0]+(-0.455722985275)*x[1]**o+(-0.25957175602)*x[1]+(0.619181596074)*x[2]**o+(0.305382546067)*x[2]
        s[1,3,0,1]=(0.958823451851)*x[0]**o+(-0.978084662344)*x[0]+(-0.276737082098)*x[1]**o+(-0.117588783747)*x[1]+(0.61233901794)*x[2]**o+(-0.430645122164)*x[2]
        s[1,3,1,0]=(0.17639490062)*x[0]**o+(-0.072015890916)*x[0]+(-0.0776296985656)*x[1]**o+(0.92506693456)*x[1]+(-0.228890114078)*x[2]**o+(-0.817012691208)*x[2]
        s[1,3,1,1]=(-0.0124492438715)*x[0]**o+(-0.291955568984)*x[0]+(-0.701102657404)*x[1]**o+(0.149273972357)*x[1]+(-0.521101405472)*x[2]**o+(0.183603155794)*x[2]
        s[1,3,2,0]=(0.12397369853)*x[0]**o+(0.436139768082)*x[0]+(0.523613099442)*x[1]**o+(0.0273767825371)*x[1]+(0.4825934187)*x[2]**o+(-0.261047164188)*x[2]
        s[1,3,2,1]=(-0.253163229158)*x[0]**o+(-0.190716207327)*x[0]+(0.104639827693)*x[1]**o+(-0.868381858703)*x[1]+(-0.439763792231)*x[2]**o+(-0.799113479705)*x[2]
        s[1,4,0,0]=(-0.304838319461)*x[0]**o+(0.701590180772)*x[0]+(-0.110558497315)*x[1]**o+(-0.893822894679)*x[1]+(-0.787583691318)*x[2]**o+(-0.89265767325)*x[2]
        s[1,4,0,1]=(0.184781667393)*x[0]**o+(0.00115077722427)*x[0]+(-0.301829747192)*x[1]**o+(0.492639963911)*x[1]+(-0.828978825155)*x[2]**o+(0.528275593254)*x[2]
        s[1,4,1,0]=(0.42616420927)*x[0]**o+(-0.473452048931)*x[0]+(0.756413372908)*x[1]**o+(-0.875430941202)*x[1]+(-0.440394982498)*x[2]**o+(0.387473346762)*x[2]
        s[1,4,1,1]=(-0.974545673815)*x[0]**o+(-0.583230191948)*x[0]+(-0.8604375003)*x[1]**o+(0.792666394984)*x[1]+(0.712418702968)*x[2]**o+(-0.635993401114)*x[2]
        s[1,4,2,0]=(0.898921012186)*x[0]**o+(0.31154532209)*x[0]+(0.100106723409)*x[1]**o+(0.510220124566)*x[1]+(-0.694976306736)*x[2]**o+(0.744462320672)*x[2]
        s[1,4,2,1]=(0.208678468389)*x[0]**o+(-0.029704981412)*x[0]+(-0.709792226721)*x[1]**o+(0.978207622687)*x[1]+(-0.94268835481)*x[2]**o+(0.983998190279)*x[2]
        s[2,0,0,0]=(-0.117454131321)*x[0]**o+(0.325178166996)*x[0]+(-0.955591553442)*x[1]**o+(-0.893511287203)*x[1]+(-0.992448477236)*x[2]**o+(-0.314767960841)*x[2]
        s[2,0,0,1]=(-0.215442066237)*x[0]**o+(0.199095086203)*x[0]+(-0.948101787926)*x[1]**o+(0.833921396167)*x[1]+(-0.945489620081)*x[2]**o+(-0.836168838947)*x[2]
        s[2,0,1,0]=(-0.0147045046721)*x[0]**o+(0.910556812541)*x[0]+(-0.854656661853)*x[1]**o+(0.603960686575)*x[1]+(-0.110464227264)*x[2]**o+(-0.654514120746)*x[2]
        s[2,0,1,1]=(0.12588483106)*x[0]**o+(0.741103813876)*x[0]+(0.273343918839)*x[1]**o+(-0.0496068521746)*x[1]+(0.851852637841)*x[2]**o+(0.513176454441)*x[2]
        s[2,0,2,0]=(-0.949030805361)*x[0]**o+(-0.186923222351)*x[0]+(-0.469739120295)*x[1]**o+(-0.667688306186)*x[1]+(0.561259377808)*x[2]**o+(-0.455307474368)*x[2]
        s[2,0,2,1]=(0.50289454022)*x[0]**o+(0.601455959785)*x[0]+(-0.359319458876)*x[1]**o+(-0.252866723971)*x[1]+(-0.232113717554)*x[2]**o+(-0.923251891682)*x[2]
        s[2,1,0,0]=(0.087617991436)*x[0]**o+(0.825668089054)*x[0]+(-0.246279274065)*x[1]**o+(0.427961138603)*x[1]+(0.134574297189)*x[2]**o+(0.126692670762)*x[2]
        s[2,1,0,1]=(0.618759834233)*x[0]**o+(-0.986840591193)*x[0]+(-0.356764253448)*x[1]**o+(0.861294268808)*x[1]+(0.329156660561)*x[2]**o+(0.454768559063)*x[2]
        s[2,1,1,0]=(0.0592587565219)*x[0]**o+(0.658218480064)*x[0]+(-0.605532183319)*x[1]**o+(0.628709371911)*x[1]+(-0.755778976293)*x[2]**o+(0.200067133647)*x[2]
        s[2,1,1,1]=(0.60715871142)*x[0]**o+(0.119152257078)*x[0]+(-0.456685613073)*x[1]**o+(-0.431830986012)*x[1]+(-0.239978100497)*x[2]**o+(-0.272134038408)*x[2]
        s[2,1,2,0]=(0.313930291609)*x[0]**o+(-0.0335423445806)*x[0]+(0.329631545985)*x[1]**o+(0.578422305734)*x[1]+(0.428275935457)*x[2]**o+(0.0754231472821)*x[2]
        s[2,1,2,1]=(0.95215367321)*x[0]**o+(-0.395110969421)*x[0]+(0.186889413454)*x[1]**o+(-0.211320210609)*x[1]+(-0.950278722072)*x[2]**o+(-0.111767520683)*x[2]
        s[2,2,0,0]=(-0.928704839938)*x[0]**o+(0.856077241772)*x[0]+(-0.0417301431511)*x[1]**o+(-0.76181987359)*x[1]+(-0.992155226136)*x[2]**o+(-0.776267895931)*x[2]
        s[2,2,0,1]=(-0.969248181097)*x[0]**o+(0.245775588093)*x[0]+(-0.505906920175)*x[1]**o+(0.146641331232)*x[1]+(-0.262102422862)*x[2]**o+(-0.561207377079)*x[2]
        s[2,2,1,0]=(-0.718472022429)*x[0]**o+(0.439297055455)*x[0]+(-0.852005557298)*x[1]**o+(0.504211693582)*x[1]+(-0.126256957345)*x[2]**o+(0.20236075339)*x[2]
        s[2,2,1,1]=(-0.547818793062)*x[0]**o+(-0.550864012953)*x[0]+(-0.87720186351)*x[1]**o+(0.962779562428)*x[1]+(-0.663892241535)*x[2]**o+(0.826235643802)*x[2]
        s[2,2,2,0]=(0.619652835608)*x[0]**o+(-0.418942285364)*x[0]+(0.407390229096)*x[1]**o+(0.999750598891)*x[1]+(0.826445107816)*x[2]**o+(-0.827613183804)*x[2]
        s[2,2,2,1]=(0.742447797718)*x[0]**o+(-0.135700593818)*x[0]+(0.691148074885)*x[1]**o+(-0.457826189496)*x[1]+(0.256017777404)*x[2]**o+(-0.165702500648)*x[2]
        s[2,3,0,0]=(-0.43784392872)*x[0]**o+(0.16925005364)*x[0]+(-0.659181589908)*x[1]**o+(0.675362824639)*x[1]+(-0.577126041428)*x[2]**o+(-0.506716255511)*x[2]
        s[2,3,0,1]=(-0.216143074446)*x[0]**o+(0.437507580615)*x[0]+(0.401331581783)*x[1]**o+(0.689412723355)*x[1]+(0.8762344674)*x[2]**o+(-0.582543733655)*x[2]
        s[2,3,1,0]=(0.0788197892383)*x[0]**o+(0.417141756193)*x[0]+(-0.253190931158)*x[1]**o+(0.409575699654)*x[1]+(0.514494619825)*x[2]**o+(0.399143793546)*x[2]
        s[2,3,1,1]=(-0.719766785708)*x[0]**o+(-0.620867182386)*x[0]+(-0.0490130572844)*x[1]**o+(0.5158062442)*x[1]+(0.871371891386)*x[2]**o+(0.568994700919)*x[2]
        s[2,3,2,0]=(0.076389643296)*x[0]**o+(0.508512060622)*x[0]+(-0.78737605411)*x[1]**o+(-0.00342872062952)*x[1]+(-0.659245059343)*x[2]**o+(-0.265456383349)*x[2]
        s[2,3,2,1]=(0.5875893363)*x[0]**o+(-0.718091321577)*x[0]+(-0.801865283024)*x[1]**o+(-0.908030358815)*x[1]+(0.431675901356)*x[2]**o+(-0.819466399747)*x[2]
        s[2,4,0,0]=(0.67994234014)*x[0]**o+(0.0421294866201)*x[0]+(0.914156657838)*x[1]**o+(0.631988857673)*x[1]+(-0.731815343643)*x[2]**o+(-0.226439431169)*x[2]
        s[2,4,0,1]=(0.382719860981)*x[0]**o+(0.0998281702684)*x[0]+(-0.191692152755)*x[1]**o+(-0.632487629935)*x[1]+(-0.47197856624)*x[2]**o+(0.743041964659)*x[2]
        s[2,4,1,0]=(-0.766894309495)*x[0]**o+(0.66365148919)*x[0]+(-0.985125550604)*x[1]**o+(0.977558230989)*x[1]+(-0.786405168576)*x[2]**o+(-0.586888601056)*x[2]
        s[2,4,1,1]=(-0.408491721685)*x[0]**o+(-0.485042299377)*x[0]+(0.0342771405689)*x[1]**o+(-0.0281888804188)*x[1]+(-0.100567975294)*x[2]**o+(0.16781221998)*x[2]
        s[2,4,2,0]=(0.655177009817)*x[0]**o+(0.599751578479)*x[0]+(-0.69519370742)*x[1]**o+(0.765111933279)*x[1]+(-0.600152872818)*x[2]**o+(0.933762224073)*x[2]
        s[2,4,2,1]=(-0.783402080781)*x[0]**o+(0.497892410534)*x[0]+(-0.896301612142)*x[1]**o+(-0.624692866087)*x[1]+(0.384092901066)*x[2]**o+(0.103876065891)*x[2]
        s[3,0,0,0]=(0.727781520125)*x[0]**o+(0.516411031951)*x[0]+(0.520409588798)*x[1]**o+(0.203314941059)*x[1]+(-0.72359147418)*x[2]**o+(0.766852866118)*x[2]
        s[3,0,0,1]=(-0.0843842205054)*x[0]**o+(-0.270711183223)*x[0]+(-0.235601047175)*x[1]**o+(-0.820277497036)*x[1]+(-0.466133585852)*x[2]**o+(0.155997468178)*x[2]
        s[3,0,1,0]=(-0.951270599262)*x[0]**o+(-0.313463217212)*x[0]+(-0.166446810677)*x[1]**o+(0.0311582608036)*x[1]+(-0.143475616634)*x[2]**o+(-0.174089420721)*x[2]
        s[3,0,1,1]=(0.31005632514)*x[0]**o+(-0.473374219519)*x[0]+(0.922718547121)*x[1]**o+(0.354849506312)*x[1]+(0.466885594274)*x[2]**o+(-0.0545966611475)*x[2]
        s[3,0,2,0]=(0.0930631890409)*x[0]**o+(0.238199576739)*x[0]+(0.785659372495)*x[1]**o+(-0.381418540402)*x[1]+(-0.372849688379)*x[2]**o+(-0.550096690451)*x[2]
        s[3,0,2,1]=(-0.0488046933792)*x[0]**o+(-0.85673228641)*x[0]+(-0.888414982245)*x[1]**o+(-0.733480030742)*x[1]+(-0.138791373365)*x[2]**o+(-0.0428634394828)*x[2]
        s[3,1,0,0]=(0.431582682868)*x[0]**o+(0.997952021577)*x[0]+(-0.586523128595)*x[1]**o+(-0.38366122568)*x[1]+(-0.966801100136)*x[2]**o+(-0.814601474666)*x[2]
        s[3,1,0,1]=(0.51469888558)*x[0]**o+(0.852630560538)*x[0]+(-0.431644582882)*x[1]**o+(0.612002981263)*x[1]+(-0.824898907932)*x[2]**o+(0.740125311337)*x[2]
        s[3,1,1,0]=(-0.998816511903)*x[0]**o+(0.501436620365)*x[0]+(-0.876369817412)*x[1]**o+(-0.751901171932)*x[1]+(-0.127524412151)*x[2]**o+(0.556578788174)*x[2]
        s[3,1,1,1]=(0.494572326989)*x[0]**o+(-0.4109962303)*x[0]+(0.332604369701)*x[1]**o+(0.144741772456)*x[1]+(-0.356738712921)*x[2]**o+(-0.930461094693)*x[2]
        s[3,1,2,0]=(0.298299420769)*x[0]**o+(0.465301670992)*x[0]+(-0.625366287818)*x[1]**o+(0.617300960428)*x[1]+(-0.0218635106247)*x[2]**o+(-0.400465814843)*x[2]
        s[3,1,2,1]=(-0.269730676738)*x[0]**o+(0.385672666629)*x[0]+(0.0768574834461)*x[1]**o+(-0.946466342247)*x[1]+(-0.46039524704)*x[2]**o+(-0.092664263096)*x[2]
        s[3,2,0,0]=(0.461319900449)*x[0]**o+(0.894479535569)*x[0]+(-0.331082415608)*x[1]**o+(0.40219970326)*x[1]+(-0.45348751543)*x[2]**o+(0.344838434995)*x[2]
        s[3,2,0,1]=(0.555684742117)*x[0]**o+(0.140415098553)*x[0]+(-0.53477167757)*x[1]**o+(-0.365761904777)*x[1]+(0.96328384437)*x[2]**o+(0.619810575276)*x[2]
        s[3,2,1,0]=(-0.14916965531)*x[0]**o+(0.889806300338)*x[0]+(-0.456091691059)*x[1]**o+(0.532621704874)*x[1]+(0.329988274366)*x[2]**o+(0.800234733169)*x[2]
        s[3,2,1,1]=(0.380606552573)*x[0]**o+(-0.541127952834)*x[0]+(-0.851396276663)*x[1]**o+(0.166128704099)*x[1]+(-0.575246273865)*x[2]**o+(-0.438073609497)*x[2]
        s[3,2,2,0]=(0.845014379056)*x[0]**o+(-0.812352254049)*x[0]+(0.0324309664699)*x[1]**o+(0.558811307761)*x[1]+(-0.865143597202)*x[2]**o+(-0.755672880303)*x[2]
        s[3,2,2,1]=(0.5343293465)*x[0]**o+(-0.840117005289)*x[0]+(-0.30550028433)*x[1]**o+(0.951726967137)*x[1]+(0.469519945822)*x[2]**o+(0.748837114445)*x[2]
        s[3,3,0,0]=(-0.885008843023)*x[0]**o+(-0.170701893945)*x[0]+(-0.374063979126)*x[1]**o+(-0.30389599708)*x[1]+(-0.661322095381)*x[2]**o+(0.838294065739)*x[2]
        s[3,3,0,1]=(0.854428192443)*x[0]**o+(0.2166993324)*x[0]+(0.43944583933)*x[1]**o+(-0.858798284298)*x[1]+(-0.84436470857)*x[2]**o+(-0.567545786739)*x[2]
        s[3,3,1,0]=(0.66702962158)*x[0]**o+(0.720100651321)*x[0]+(0.0956882490639)*x[1]**o+(0.0372232762552)*x[1]+(-0.67009655948)*x[2]**o+(-0.569854055098)*x[2]
        s[3,3,1,1]=(0.146466946465)*x[0]**o+(0.164924499496)*x[0]+(-0.2972769793)*x[1]**o+(0.865363847078)*x[1]+(-0.180431004392)*x[2]**o+(0.265225273207)*x[2]
        s[3,3,2,0]=(-0.830402000378)*x[0]**o+(-0.00161430224161)*x[0]+(-0.863095020073)*x[1]**o+(0.922102273263)*x[1]+(0.385860222623)*x[2]**o+(0.981756782753)*x[2]
        s[3,3,2,1]=(0.567277892172)*x[0]**o+(0.78766223915)*x[0]+(0.412870235103)*x[1]**o+(0.46588554159)*x[1]+(0.716230666858)*x[2]**o+(0.425979074273)*x[2]
        s[3,4,0,0]=(0.679938424982)*x[0]**o+(0.508093418021)*x[0]+(-0.73945776378)*x[1]**o+(-0.178938904098)*x[1]+(0.996547500259)*x[2]**o+(-0.689147793037)*x[2]
        s[3,4,0,1]=(-0.181577713519)*x[0]**o+(-0.413066527133)*x[0]+(0.83941283649)*x[1]**o+(-0.664951450375)*x[1]+(0.993705498629)*x[2]**o+(-0.483417137243)*x[2]
        s[3,4,1,0]=(0.328862164768)*x[0]**o+(0.869116819138)*x[0]+(-0.8079517423)*x[1]**o+(-0.692894104885)*x[1]+(-0.407582178576)*x[2]**o+(0.437242243866)*x[2]
        s[3,4,1,1]=(0.512281196984)*x[0]**o+(-0.155819537445)*x[0]+(0.693970927236)*x[1]**o+(0.785297186585)*x[1]+(-0.514686313027)*x[2]**o+(-0.474084873678)*x[2]
        s[3,4,2,0]=(0.225627329646)*x[0]**o+(0.247722481758)*x[0]+(0.88743622494)*x[1]**o+(-0.953008640988)*x[1]+(-0.428243081608)*x[2]**o+(0.88759395083)*x[2]
        s[3,4,2,1]=(-0.942620080756)*x[0]**o+(-0.794499826777)*x[0]+(-0.191327255697)*x[1]**o+(-0.185051303598)*x[1]+(0.681543419098)*x[2]**o+(-0.142256046241)*x[2]
        ref[0,0,0,0]=(0.854566232468)*(1+2.*(dim-1.)/(o+1.))+(-2.27229927089)*dim
        ref[0,0,0,1]=(-0.989332586323)*(1+2.*(dim-1.)/(o+1.))+(-0.751767471217)*dim
        ref[0,0,1,0]=(0.170998015739)*(1+2.*(dim-1.)/(o+1.))+(2.07130399849)*dim
        ref[0,0,1,1]=(1.12437874793)*(1+2.*(dim-1.)/(o+1.))+(0.369594791005)*dim
        ref[0,0,2,0]=(0.175354222567)*(1+2.*(dim-1.)/(o+1.))+(-1.05037040857)*dim
        ref[0,0,2,1]=(0.306997257038)*(1+2.*(dim-1.)/(o+1.))+(-0.596327147149)*dim
        ref[0,1,0,0]=(-0.986664028597)*(1+2.*(dim-1.)/(o+1.))+(0.147255273686)*dim
        ref[0,1,0,1]=(-0.057403016606)*(1+2.*(dim-1.)/(o+1.))+(-1.4736524552)*dim
        ref[0,1,1,0]=(1.01904700679)*(1+2.*(dim-1.)/(o+1.))+(-0.757373181813)*dim
        ref[0,1,1,1]=(0.783282492073)*(1+2.*(dim-1.)/(o+1.))+(-1.21875468798)*dim
        ref[0,1,2,0]=(0.952340897946)*(1+2.*(dim-1.)/(o+1.))+(0.713464583232)*dim
        ref[0,1,2,1]=(1.45696275631)*(1+2.*(dim-1.)/(o+1.))+(1.58168131279)*dim
        ref[0,2,0,0]=(0.064837646501)*(1+2.*(dim-1.)/(o+1.))+(-1.87419276367)*dim
        ref[0,2,0,1]=(-0.760889397294)*(1+2.*(dim-1.)/(o+1.))+(2.07687305038)*dim
        ref[0,2,1,0]=(-0.380575937425)*(1+2.*(dim-1.)/(o+1.))+(1.18322256955)*dim
        ref[0,2,1,1]=(0.0998904195137)*(1+2.*(dim-1.)/(o+1.))+(0.762541697126)*dim
        ref[0,2,2,0]=(0.806899473909)*(1+2.*(dim-1.)/(o+1.))+(0.867881265887)*dim
        ref[0,2,2,1]=(0.0474528005784)*(1+2.*(dim-1.)/(o+1.))+(-0.0563780781365)*dim
        ref[0,3,0,0]=(0.145415525229)*(1+2.*(dim-1.)/(o+1.))+(0.399190212115)*dim
        ref[0,3,0,1]=(2.26114410824)*(1+2.*(dim-1.)/(o+1.))+(-0.743490691928)*dim
        ref[0,3,1,0]=(0.960813974011)*(1+2.*(dim-1.)/(o+1.))+(-0.0284214479695)*dim
        ref[0,3,1,1]=(1.47325375635)*(1+2.*(dim-1.)/(o+1.))+(-0.599244959013)*dim
        ref[0,3,2,0]=(1.08301397996)*(1+2.*(dim-1.)/(o+1.))+(0.247426252553)*dim
        ref[0,3,2,1]=(-1.4186347995)*(1+2.*(dim-1.)/(o+1.))+(0.696099814573)*dim
        ref[0,4,0,0]=(-0.393517834393)*(1+2.*(dim-1.)/(o+1.))+(-2.30394867936)*dim
        ref[0,4,0,1]=(-1.40711102949)*(1+2.*(dim-1.)/(o+1.))+(-0.690433966228)*dim
        ref[0,4,1,0]=(-0.638615138949)*(1+2.*(dim-1.)/(o+1.))+(-0.47200820565)*dim
        ref[0,4,1,1]=(0.989421505854)*(1+2.*(dim-1.)/(o+1.))+(-0.752441737288)*dim
        ref[0,4,2,0]=(-0.881756130244)*(1+2.*(dim-1.)/(o+1.))+(-0.338966152194)*dim
        ref[0,4,2,1]=(-2.33289649019)*(1+2.*(dim-1.)/(o+1.))+(0.259221814291)*dim
        ref[1,0,0,0]=(0.535686388307)*(1+2.*(dim-1.)/(o+1.))+(0.266916202289)*dim
        ref[1,0,0,1]=(1.46703190363)*(1+2.*(dim-1.)/(o+1.))+(0.398601744865)*dim
        ref[1,0,1,0]=(0.32201076756)*(1+2.*(dim-1.)/(o+1.))+(0.135321713403)*dim
        ref[1,0,1,1]=(-0.898740301412)*(1+2.*(dim-1.)/(o+1.))+(0.923393823202)*dim
        ref[1,0,2,0]=(-0.376626929371)*(1+2.*(dim-1.)/(o+1.))+(-0.502418571748)*dim
        ref[1,0,2,1]=(0.367615051881)*(1+2.*(dim-1.)/(o+1.))+(0.113455552197)*dim
        ref[1,1,0,0]=(-0.0262926747634)*(1+2.*(dim-1.)/(o+1.))+(-0.656064973208)*dim
        ref[1,1,0,1]=(-1.14713357654)*(1+2.*(dim-1.)/(o+1.))+(1.20161428708)*dim
        ref[1,1,1,0]=(-0.256099523566)*(1+2.*(dim-1.)/(o+1.))+(1.12819050155)*dim
        ref[1,1,1,1]=(-0.711548624701)*(1+2.*(dim-1.)/(o+1.))+(1.09265080887)*dim
        ref[1,1,2,0]=(-0.23671592891)*(1+2.*(dim-1.)/(o+1.))+(-0.371739193929)*dim
        ref[1,1,2,1]=(-0.384096984041)*(1+2.*(dim-1.)/(o+1.))+(-0.508592134529)*dim
        ref[1,2,0,0]=(0.519697602114)*(1+2.*(dim-1.)/(o+1.))+(-0.195849802533)*dim
        ref[1,2,0,1]=(0.454720124645)*(1+2.*(dim-1.)/(o+1.))+(-0.125915366271)*dim
        ref[1,2,1,0]=(-0.00770507513982)*(1+2.*(dim-1.)/(o+1.))+(1.35445743788)*dim
        ref[1,2,1,1]=(-0.977550759742)*(1+2.*(dim-1.)/(o+1.))+(-1.61751594043)*dim
        ref[1,2,2,0]=(0.861571331194)*(1+2.*(dim-1.)/(o+1.))+(-0.861303922176)*dim
        ref[1,2,2,1]=(-0.810715347979)*(1+2.*(dim-1.)/(o+1.))+(1.42155547855)*dim
        ref[1,3,0,0]=(-0.48880602403)*(1+2.*(dim-1.)/(o+1.))+(-0.630731120484)*dim
        ref[1,3,0,1]=(1.29442538769)*(1+2.*(dim-1.)/(o+1.))+(-1.52631856825)*dim
        ref[1,3,1,0]=(-0.130124912024)*(1+2.*(dim-1.)/(o+1.))+(0.0360383524367)*dim
        ref[1,3,1,1]=(-1.23465330675)*(1+2.*(dim-1.)/(o+1.))+(0.0409215591672)*dim
        ref[1,3,2,0]=(1.13018021667)*(1+2.*(dim-1.)/(o+1.))+(0.202469386431)*dim
        ref[1,3,2,1]=(-0.588287193696)*(1+2.*(dim-1.)/(o+1.))+(-1.85821154573)*dim
        ref[1,4,0,0]=(-1.20298050809)*(1+2.*(dim-1.)/(o+1.))+(-1.08489038716)*dim
        ref[1,4,0,1]=(-0.946026904954)*(1+2.*(dim-1.)/(o+1.))+(1.02206633439)*dim
        ref[1,4,1,0]=(0.74218259968)*(1+2.*(dim-1.)/(o+1.))+(-0.961409643371)*dim
        ref[1,4,1,1]=(-1.12256447115)*(1+2.*(dim-1.)/(o+1.))+(-0.426557198078)*dim
        ref[1,4,2,0]=(0.304051428859)*(1+2.*(dim-1.)/(o+1.))+(1.56622776733)*dim
        ref[1,4,2,1]=(-1.44380211314)*(1+2.*(dim-1.)/(o+1.))+(1.93250083155)*dim
        ref[2,0,0,0]=(-2.065494162)*(1+2.*(dim-1.)/(o+1.))+(-0.883101081048)*dim
        ref[2,0,0,1]=(-2.10903347424)*(1+2.*(dim-1.)/(o+1.))+(0.196847643424)*dim
        ref[2,0,1,0]=(-0.979825393789)*(1+2.*(dim-1.)/(o+1.))+(0.86000337837)*dim
        ref[2,0,1,1]=(1.25108138774)*(1+2.*(dim-1.)/(o+1.))+(1.20467341614)*dim
        ref[2,0,2,0]=(-0.857510547847)*(1+2.*(dim-1.)/(o+1.))+(-1.30991900291)*dim
        ref[2,0,2,1]=(-0.0885386362105)*(1+2.*(dim-1.)/(o+1.))+(-0.574662655868)*dim
        ref[2,1,0,0]=(-0.0240869854401)*(1+2.*(dim-1.)/(o+1.))+(1.38032189842)*dim
        ref[2,1,0,1]=(0.591152241347)*(1+2.*(dim-1.)/(o+1.))+(0.329222236678)*dim
        ref[2,1,1,0]=(-1.30205240309)*(1+2.*(dim-1.)/(o+1.))+(1.48699498562)*dim
        ref[2,1,1,1]=(-0.0895050021498)*(1+2.*(dim-1.)/(o+1.))+(-0.584812767342)*dim
        ref[2,1,2,0]=(1.07183777305)*(1+2.*(dim-1.)/(o+1.))+(0.620303108435)*dim
        ref[2,1,2,1]=(0.188764364592)*(1+2.*(dim-1.)/(o+1.))+(-0.718198700712)*dim
        ref[2,2,0,0]=(-1.96259020923)*(1+2.*(dim-1.)/(o+1.))+(-0.682010527749)*dim
        ref[2,2,0,1]=(-1.73725752413)*(1+2.*(dim-1.)/(o+1.))+(-0.168790457754)*dim
        ref[2,2,1,0]=(-1.69673453707)*(1+2.*(dim-1.)/(o+1.))+(1.14586950243)*dim
        ref[2,2,1,1]=(-2.08891289811)*(1+2.*(dim-1.)/(o+1.))+(1.23815119328)*dim
        ref[2,2,2,0]=(1.85348817252)*(1+2.*(dim-1.)/(o+1.))+(-0.246804870276)*dim
        ref[2,2,2,1]=(1.68961365001)*(1+2.*(dim-1.)/(o+1.))+(-0.759229283962)*dim
        ref[2,3,0,0]=(-1.67415156006)*(1+2.*(dim-1.)/(o+1.))+(0.337896622768)*dim
        ref[2,3,0,1]=(1.06142297474)*(1+2.*(dim-1.)/(o+1.))+(0.544376570315)*dim
        ref[2,3,1,0]=(0.340123477905)*(1+2.*(dim-1.)/(o+1.))+(1.22586124939)*dim
        ref[2,3,1,1]=(0.102592048393)*(1+2.*(dim-1.)/(o+1.))+(0.463933762733)*dim
        ref[2,3,2,0]=(-1.37023147016)*(1+2.*(dim-1.)/(o+1.))+(0.239626956644)*dim
        ref[2,3,2,1]=(0.217399954632)*(1+2.*(dim-1.)/(o+1.))+(-2.44558808014)*dim
        ref[2,4,0,0]=(0.862283654334)*(1+2.*(dim-1.)/(o+1.))+(0.447678913125)*dim
        ref[2,4,0,1]=(-0.280950858013)*(1+2.*(dim-1.)/(o+1.))+(0.210382504993)*dim
        ref[2,4,1,0]=(-2.53842502868)*(1+2.*(dim-1.)/(o+1.))+(1.05432111912)*dim
        ref[2,4,1,1]=(-0.474782556409)*(1+2.*(dim-1.)/(o+1.))+(-0.345418959816)*dim
        ref[2,4,2,0]=(-0.640169570421)*(1+2.*(dim-1.)/(o+1.))+(2.29862573583)*dim
        ref[2,4,2,1]=(-1.29561079186)*(1+2.*(dim-1.)/(o+1.))+(-0.0229243896616)*dim
        ref[3,0,0,0]=(0.524599634743)*(1+2.*(dim-1.)/(o+1.))+(1.48657883913)*dim
        ref[3,0,0,1]=(-0.786118853533)*(1+2.*(dim-1.)/(o+1.))+(-0.934991212081)*dim
        ref[3,0,1,0]=(-1.26119302657)*(1+2.*(dim-1.)/(o+1.))+(-0.45639437713)*dim
        ref[3,0,1,1]=(1.69966046653)*(1+2.*(dim-1.)/(o+1.))+(-0.173121374354)*dim
        ref[3,0,2,0]=(0.505872873156)*(1+2.*(dim-1.)/(o+1.))+(-0.693315654114)*dim
        ref[3,0,2,1]=(-1.07601104899)*(1+2.*(dim-1.)/(o+1.))+(-1.63307575663)*dim
        ref[3,1,0,0]=(-1.12174154586)*(1+2.*(dim-1.)/(o+1.))+(-0.200310678769)*dim
        ref[3,1,0,1]=(-0.741844605234)*(1+2.*(dim-1.)/(o+1.))+(2.20475885314)*dim
        ref[3,1,1,0]=(-2.00271074147)*(1+2.*(dim-1.)/(o+1.))+(0.306114236606)*dim
        ref[3,1,1,1]=(0.470437983768)*(1+2.*(dim-1.)/(o+1.))+(-1.19671555254)*dim
        ref[3,1,2,0]=(-0.348930377674)*(1+2.*(dim-1.)/(o+1.))+(0.682136816577)*dim
        ref[3,1,2,1]=(-0.653268440332)*(1+2.*(dim-1.)/(o+1.))+(-0.653457938714)*dim
        ref[3,2,0,0]=(-0.323250030589)*(1+2.*(dim-1.)/(o+1.))+(1.64151767382)*dim
        ref[3,2,0,1]=(0.984196908917)*(1+2.*(dim-1.)/(o+1.))+(0.394463769052)*dim
        ref[3,2,1,0]=(-0.275273072003)*(1+2.*(dim-1.)/(o+1.))+(2.22266273838)*dim
        ref[3,2,1,1]=(-1.04603599796)*(1+2.*(dim-1.)/(o+1.))+(-0.813072858233)*dim
        ref[3,2,2,0]=(0.0123017483236)*(1+2.*(dim-1.)/(o+1.))+(-1.00921382659)*dim
        ref[3,2,2,1]=(0.698349007993)*(1+2.*(dim-1.)/(o+1.))+(0.860447076293)*dim
        ref[3,3,0,0]=(-1.92039491753)*(1+2.*(dim-1.)/(o+1.))+(0.363696174714)*dim
        ref[3,3,0,1]=(0.449509323203)*(1+2.*(dim-1.)/(o+1.))+(-1.20964473864)*dim
        ref[3,3,1,0]=(0.0926213111634)*(1+2.*(dim-1.)/(o+1.))+(0.187469872479)*dim
        ref[3,3,1,1]=(-0.331241037228)*(1+2.*(dim-1.)/(o+1.))+(1.29551361978)*dim
        ref[3,3,2,0]=(-1.30763679783)*(1+2.*(dim-1.)/(o+1.))+(1.90224475377)*dim
        ref[3,3,2,1]=(1.69637879413)*(1+2.*(dim-1.)/(o+1.))+(1.67952685501)*dim
        ref[3,4,0,0]=(0.937028161461)*(1+2.*(dim-1.)/(o+1.))+(-0.359993279113)*dim
        ref[3,4,0,1]=(1.6515406216)*(1+2.*(dim-1.)/(o+1.))+(-1.56143511475)*dim
        ref[3,4,1,0]=(-0.886671756108)*(1+2.*(dim-1.)/(o+1.))+(0.613464958119)*dim
        ref[3,4,1,1]=(0.691565811192)*(1+2.*(dim-1.)/(o+1.))+(0.155392775462)*dim
        ref[3,4,2,0]=(0.684820472977)*(1+2.*(dim-1.)/(o+1.))+(0.1823077916)*dim
        ref[3,4,2,1]=(-0.452403917355)*(1+2.*(dim-1.)/(o+1.))+(-1.12180717662)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.506621877215)*x[0]**o+(-0.45560383685)*x[0]+(-0.598137301806)*x[1]**o+(0.784241595615)*x[1]
        ref=(-0.0915154245913)*(1+2.*(dim-1.)/(o+1.))+(0.328637758765)*dim
      else:
        arg=(-0.505489131484)*x[0]**o+(0.58139159259)*x[0]+(0.458086828888)*x[1]**o+(0.549090746672)*x[1]+(-0.0882325936167)*x[2]**o+(0.0170295954808)*x[2]
        ref=(-0.135634896213)*(1+2.*(dim-1.)/(o+1.))+(1.14751193474)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-0.88779433873)*x[0]**o+(0.507955012301)*x[0]+(0.652739274924)*x[1]**o+(-0.971975755989)*x[1]
        arg[1]=(-0.0812690931139)*x[0]**o+(-0.283880962354)*x[0]+(0.094904517439)*x[1]**o+(-0.940528164353)*x[1]
        ref[0]=(-0.235055063806)*(1+2.*(dim-1.)/(o+1.))+(-0.464020743688)*dim
        ref[1]=(0.0136354243251)*(1+2.*(dim-1.)/(o+1.))+(-1.22440912671)*dim
      else:
        arg[0]=(-0.506412384389)*x[0]**o+(-0.465794405394)*x[0]+(0.346977499982)*x[1]**o+(0.386181176987)*x[1]+(-0.476729572397)*x[2]**o+(-0.956945933508)*x[2]
        arg[1]=(0.267092807416)*x[0]**o+(-0.54211155802)*x[0]+(-0.27815411727)*x[1]**o+(-0.984542450102)*x[1]+(0.656389296076)*x[2]**o+(0.0395863971401)*x[2]
        ref[0]=(-0.636164456804)*(1+2.*(dim-1.)/(o+1.))+(-1.03655916191)*dim
        ref[1]=(0.645327986222)*(1+2.*(dim-1.)/(o+1.))+(-1.48706761098)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(0.0607636892262)*x[0]**o+(0.740866012668)*x[0]+(0.145223136928)*x[1]**o+(-0.738387773526)*x[1]
        arg[0,1]=(0.0292598484511)*x[0]**o+(-0.189208290889)*x[0]+(0.902021405165)*x[1]**o+(0.423106545559)*x[1]
        arg[0,2]=(0.519658217226)*x[0]**o+(-0.768406018136)*x[0]+(-0.189532527097)*x[1]**o+(0.256417429833)*x[1]
        arg[0,3]=(-0.565327663481)*x[0]**o+(0.876762003673)*x[0]+(-0.243804758063)*x[1]**o+(0.399476036454)*x[1]
        arg[0,4]=(0.0632252409818)*x[0]**o+(0.0951130662538)*x[0]+(0.185662425763)*x[1]**o+(0.907446482506)*x[1]
        arg[1,0]=(-0.430504584794)*x[0]**o+(0.864814314023)*x[0]+(0.47460095501)*x[1]**o+(-0.599679142188)*x[1]
        arg[1,1]=(0.699618175298)*x[0]**o+(-0.561782605517)*x[0]+(-0.304250865015)*x[1]**o+(-0.341836776795)*x[1]
        arg[1,2]=(0.856956767316)*x[0]**o+(0.610026903213)*x[0]+(-0.666749954599)*x[1]**o+(0.885388885234)*x[1]
        arg[1,3]=(0.699377007631)*x[0]**o+(0.699906750287)*x[0]+(0.486690669414)*x[1]**o+(0.167425901549)*x[1]
        arg[1,4]=(0.466701415896)*x[0]**o+(-0.0719752759126)*x[0]+(0.795662047062)*x[1]**o+(0.979862503926)*x[1]
        arg[2,0]=(-0.912218732612)*x[0]**o+(0.950990472894)*x[0]+(-0.527999708902)*x[1]**o+(0.695735593043)*x[1]
        arg[2,1]=(-0.223468270048)*x[0]**o+(-0.882565147673)*x[0]+(0.750516681645)*x[1]**o+(-0.652794590093)*x[1]
        arg[2,2]=(0.897646682412)*x[0]**o+(-0.815401214058)*x[0]+(0.629680741327)*x[1]**o+(-0.9716988156)*x[1]
        arg[2,3]=(0.279739517524)*x[0]**o+(-0.0859797206016)*x[0]+(-0.0454325622757)*x[1]**o+(0.301214753868)*x[1]
        arg[2,4]=(-0.358056827092)*x[0]**o+(0.254207762658)*x[0]+(0.831499219864)*x[1]**o+(0.0930388028811)*x[1]
        arg[3,0]=(0.0761520020047)*x[0]**o+(0.650696394486)*x[0]+(-0.132369660678)*x[1]**o+(-0.321220385971)*x[1]
        arg[3,1]=(0.238715231585)*x[0]**o+(-0.297219391083)*x[0]+(-0.15446552614)*x[1]**o+(0.860560615056)*x[1]
        arg[3,2]=(0.708864754538)*x[0]**o+(-0.221186702958)*x[0]+(-0.478127729891)*x[1]**o+(0.0309604412212)*x[1]
        arg[3,3]=(0.362909366479)*x[0]**o+(0.382279106813)*x[0]+(0.827905062086)*x[1]**o+(-0.303048355949)*x[1]
        arg[3,4]=(0.573736946844)*x[0]**o+(0.645059764766)*x[0]+(0.301945383678)*x[1]**o+(-0.103083649994)*x[1]
        ref[0,0]=(0.205986826154)*(1+2.*(dim-1.)/(o+1.))+(0.0024782391428)*dim
        ref[0,1]=(0.931281253617)*(1+2.*(dim-1.)/(o+1.))+(0.23389825467)*dim
        ref[0,2]=(0.330125690129)*(1+2.*(dim-1.)/(o+1.))+(-0.511988588303)*dim
        ref[0,3]=(-0.809132421544)*(1+2.*(dim-1.)/(o+1.))+(1.27623804013)*dim
        ref[0,4]=(0.248887666745)*(1+2.*(dim-1.)/(o+1.))+(1.00255954876)*dim
        ref[1,0]=(0.044096370216)*(1+2.*(dim-1.)/(o+1.))+(0.265135171835)*dim
        ref[1,1]=(0.395367310283)*(1+2.*(dim-1.)/(o+1.))+(-0.903619382312)*dim
        ref[1,2]=(0.190206812717)*(1+2.*(dim-1.)/(o+1.))+(1.49541578845)*dim
        ref[1,3]=(1.18606767705)*(1+2.*(dim-1.)/(o+1.))+(0.867332651836)*dim
        ref[1,4]=(1.26236346296)*(1+2.*(dim-1.)/(o+1.))+(0.907887228013)*dim
        ref[2,0]=(-1.44021844151)*(1+2.*(dim-1.)/(o+1.))+(1.64672606594)*dim
        ref[2,1]=(0.527048411596)*(1+2.*(dim-1.)/(o+1.))+(-1.53535973777)*dim
        ref[2,2]=(1.52732742374)*(1+2.*(dim-1.)/(o+1.))+(-1.78710002966)*dim
        ref[2,3]=(0.234306955249)*(1+2.*(dim-1.)/(o+1.))+(0.215235033266)*dim
        ref[2,4]=(0.473442392772)*(1+2.*(dim-1.)/(o+1.))+(0.34724656554)*dim
        ref[3,0]=(-0.0562176586729)*(1+2.*(dim-1.)/(o+1.))+(0.329476008516)*dim
        ref[3,1]=(0.0842497054449)*(1+2.*(dim-1.)/(o+1.))+(0.563341223973)*dim
        ref[3,2]=(0.230737024647)*(1+2.*(dim-1.)/(o+1.))+(-0.190226261737)*dim
        ref[3,3]=(1.19081442856)*(1+2.*(dim-1.)/(o+1.))+(0.079230750864)*dim
        ref[3,4]=(0.875682330522)*(1+2.*(dim-1.)/(o+1.))+(0.541976114772)*dim
      else:
        arg[0,0]=(0.0640108068236)*x[0]**o+(0.0322830166043)*x[0]+(-0.703691335255)*x[1]**o+(-0.185228770711)*x[1]+(-0.738714166313)*x[2]**o+(-0.423576647077)*x[2]
        arg[0,1]=(-0.738341351945)*x[0]**o+(0.396306395436)*x[0]+(-0.369974848568)*x[1]**o+(0.750933208783)*x[1]+(0.554069869127)*x[2]**o+(0.634275493051)*x[2]
        arg[0,2]=(-0.259193248708)*x[0]**o+(-0.900007449583)*x[0]+(0.894379436915)*x[1]**o+(0.997814320752)*x[1]+(0.233192116622)*x[2]**o+(0.986616259649)*x[2]
        arg[0,3]=(0.381480250594)*x[0]**o+(0.734823219057)*x[0]+(0.569912440169)*x[1]**o+(-0.441084504695)*x[1]+(0.312539701584)*x[2]**o+(-0.3733004974)*x[2]
        arg[0,4]=(-0.779059807622)*x[0]**o+(0.538777922322)*x[0]+(-0.0695879211612)*x[1]**o+(-0.204862216499)*x[1]+(-0.0109698525672)*x[2]**o+(0.981875117617)*x[2]
        arg[1,0]=(-0.415857630044)*x[0]**o+(-0.356607866913)*x[0]+(-0.628854326903)*x[1]**o+(0.0401233333505)*x[1]+(0.687470169579)*x[2]**o+(-0.565349753114)*x[2]
        arg[1,1]=(-0.438364910402)*x[0]**o+(0.498078194367)*x[0]+(-0.120729597948)*x[1]**o+(-0.622076117713)*x[1]+(0.995130029172)*x[2]**o+(-0.7234601064)*x[2]
        arg[1,2]=(-0.957869585558)*x[0]**o+(0.374693219824)*x[0]+(0.0355192120972)*x[1]**o+(-0.516770140651)*x[1]+(0.521887953791)*x[2]**o+(0.00771564236502)*x[2]
        arg[1,3]=(-0.0495580194062)*x[0]**o+(-0.183895326318)*x[0]+(-0.55304086303)*x[1]**o+(0.920204449879)*x[1]+(0.143165855157)*x[2]**o+(-0.773684620462)*x[2]
        arg[1,4]=(0.795913131087)*x[0]**o+(0.653826129199)*x[0]+(-0.585562760813)*x[1]**o+(0.274083108918)*x[1]+(-0.165275654116)*x[2]**o+(0.358678758212)*x[2]
        arg[2,0]=(-0.0699685857843)*x[0]**o+(-0.42922435931)*x[0]+(-0.0494308981938)*x[1]**o+(-0.956597829176)*x[1]+(-0.912941305307)*x[2]**o+(0.214777332562)*x[2]
        arg[2,1]=(-0.900983338173)*x[0]**o+(-0.779729884378)*x[0]+(-0.860100306508)*x[1]**o+(0.14211537281)*x[1]+(-0.984439938285)*x[2]**o+(-0.549987779842)*x[2]
        arg[2,2]=(0.801489512389)*x[0]**o+(0.896468923826)*x[0]+(-0.155737972887)*x[1]**o+(0.639864330477)*x[1]+(0.798598705038)*x[2]**o+(-0.863019041525)*x[2]
        arg[2,3]=(-0.453612136504)*x[0]**o+(0.522662491688)*x[0]+(-0.463058076486)*x[1]**o+(0.398018809499)*x[1]+(-0.612044961611)*x[2]**o+(-0.481951883217)*x[2]
        arg[2,4]=(-0.0739336519425)*x[0]**o+(-0.241753740154)*x[0]+(0.0176292058898)*x[1]**o+(0.435641132809)*x[1]+(-0.0592828584976)*x[2]**o+(0.430115620801)*x[2]
        arg[3,0]=(0.604784358152)*x[0]**o+(-0.817214179503)*x[0]+(-0.423104272008)*x[1]**o+(0.287787881773)*x[1]+(0.678036469241)*x[2]**o+(0.578347129563)*x[2]
        arg[3,1]=(0.362390778598)*x[0]**o+(-0.389066580119)*x[0]+(-0.934087580396)*x[1]**o+(0.907587485226)*x[1]+(-0.902083043406)*x[2]**o+(0.904707538314)*x[2]
        arg[3,2]=(0.862463207465)*x[0]**o+(0.904923219336)*x[0]+(0.421361696042)*x[1]**o+(-0.779329196171)*x[1]+(0.059521522394)*x[2]**o+(-0.744510067387)*x[2]
        arg[3,3]=(0.412029226851)*x[0]**o+(-0.649730143511)*x[0]+(0.579628017001)*x[1]**o+(-0.949180147129)*x[1]+(-0.073016492309)*x[2]**o+(0.666554170757)*x[2]
        arg[3,4]=(-0.700401396504)*x[0]**o+(-0.89194120812)*x[0]+(0.592387042213)*x[1]**o+(-0.906632787891)*x[1]+(-0.486072860812)*x[2]**o+(-0.505386176523)*x[2]
        ref[0,0]=(-1.37839469474)*(1+2.*(dim-1.)/(o+1.))+(-0.576522401184)*dim
        ref[0,1]=(-0.554246331386)*(1+2.*(dim-1.)/(o+1.))+(1.78151509727)*dim
        ref[0,2]=(0.868378304829)*(1+2.*(dim-1.)/(o+1.))+(1.08442313082)*dim
        ref[0,3]=(1.26393239235)*(1+2.*(dim-1.)/(o+1.))+(-0.0795617830385)*dim
        ref[0,4]=(-0.85961758135)*(1+2.*(dim-1.)/(o+1.))+(1.31579082344)*dim
        ref[1,0]=(-0.357241787368)*(1+2.*(dim-1.)/(o+1.))+(-0.881834286676)*dim
        ref[1,1]=(0.436035520821)*(1+2.*(dim-1.)/(o+1.))+(-0.847458029746)*dim
        ref[1,2]=(-0.40046241967)*(1+2.*(dim-1.)/(o+1.))+(-0.134361278462)*dim
        ref[1,3]=(-0.459433027279)*(1+2.*(dim-1.)/(o+1.))+(-0.0373754969011)*dim
        ref[1,4]=(0.0450747161574)*(1+2.*(dim-1.)/(o+1.))+(1.28658799633)*dim
        ref[2,0]=(-1.03234078929)*(1+2.*(dim-1.)/(o+1.))+(-1.17104485592)*dim
        ref[2,1]=(-2.74552358297)*(1+2.*(dim-1.)/(o+1.))+(-1.18760229141)*dim
        ref[2,2]=(1.44435024454)*(1+2.*(dim-1.)/(o+1.))+(0.673314212777)*dim
        ref[2,3]=(-1.5287151746)*(1+2.*(dim-1.)/(o+1.))+(0.43872941797)*dim
        ref[2,4]=(-0.11558730455)*(1+2.*(dim-1.)/(o+1.))+(0.624003013456)*dim
        ref[3,0]=(0.859716555385)*(1+2.*(dim-1.)/(o+1.))+(0.0489208318321)*dim
        ref[3,1]=(-1.4737798452)*(1+2.*(dim-1.)/(o+1.))+(1.42322844342)*dim
        ref[3,2]=(1.3433464259)*(1+2.*(dim-1.)/(o+1.))+(-0.618916044222)*dim
        ref[3,3]=(0.918640751542)*(1+2.*(dim-1.)/(o+1.))+(-0.932356119883)*dim
        ref[3,4]=(-0.594087215103)*(1+2.*(dim-1.)/(o+1.))+(-2.30396017253)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.662707930283)*x[0]**o+(0.625512077969)*x[0]+(-0.222208840922)*x[1]**o+(0.374614317804)*x[1]
        arg[0,0,1]=(0.685952252808)*x[0]**o+(-0.88219802797)*x[0]+(-0.948918831634)*x[1]**o+(-0.00615202896776)*x[1]
        arg[0,1,0]=(-0.555231206657)*x[0]**o+(0.997112425959)*x[0]+(-0.186247870511)*x[1]**o+(-0.0658892046229)*x[1]
        arg[0,1,1]=(-0.735288285178)*x[0]**o+(0.941234848614)*x[0]+(-0.307761737256)*x[1]**o+(0.933293069433)*x[1]
        arg[1,0,0]=(-0.618191337825)*x[0]**o+(-0.12930895166)*x[0]+(0.76213325083)*x[1]**o+(-0.744049403156)*x[1]
        arg[1,0,1]=(-0.793412570981)*x[0]**o+(0.533175323401)*x[0]+(-0.87398581932)*x[1]**o+(-0.279651754725)*x[1]
        arg[1,1,0]=(-0.0654077907694)*x[0]**o+(0.239040458488)*x[0]+(0.730639927837)*x[1]**o+(0.805020533329)*x[1]
        arg[1,1,1]=(-0.560195748542)*x[0]**o+(-0.265389430135)*x[0]+(-0.295142324741)*x[1]**o+(-0.929699771024)*x[1]
        arg[2,0,0]=(0.935760865577)*x[0]**o+(-0.332458843131)*x[0]+(0.257302695187)*x[1]**o+(-0.347547927033)*x[1]
        arg[2,0,1]=(-0.245770653276)*x[0]**o+(0.126945988409)*x[0]+(0.302793434995)*x[1]**o+(-0.997483200172)*x[1]
        arg[2,1,0]=(0.523494364951)*x[0]**o+(-0.323782740449)*x[0]+(-0.563752593454)*x[1]**o+(-0.358325752838)*x[1]
        arg[2,1,1]=(-0.506330114515)*x[0]**o+(-0.733234050567)*x[0]+(0.734774667778)*x[1]**o+(0.795264566776)*x[1]
        arg[3,0,0]=(-0.21258295235)*x[0]**o+(-0.929331330718)*x[0]+(-0.546061847742)*x[1]**o+(0.0613931491065)*x[1]
        arg[3,0,1]=(-0.357847946656)*x[0]**o+(-0.133795684508)*x[0]+(-0.796646028688)*x[1]**o+(0.875795801429)*x[1]
        arg[3,1,0]=(0.923717804039)*x[0]**o+(-0.176266633621)*x[0]+(-0.605892867534)*x[1]**o+(0.633377612365)*x[1]
        arg[3,1,1]=(0.729025508129)*x[0]**o+(0.412034955257)*x[0]+(0.0933054864485)*x[1]**o+(0.386019085532)*x[1]
        arg[4,0,0]=(-0.0705524918008)*x[0]**o+(-0.869605321304)*x[0]+(0.244351267755)*x[1]**o+(0.898883388102)*x[1]
        arg[4,0,1]=(-0.453297333327)*x[0]**o+(-0.67824185123)*x[0]+(0.0497322840245)*x[1]**o+(0.581677879069)*x[1]
        arg[4,1,0]=(0.343975006358)*x[0]**o+(0.724084357103)*x[0]+(-0.186292335362)*x[1]**o+(-0.257997075018)*x[1]
        arg[4,1,1]=(-0.389731923798)*x[0]**o+(0.0949155578616)*x[0]+(-0.298428138752)*x[1]**o+(0.819510069666)*x[1]
        arg[5,0,0]=(0.827670169513)*x[0]**o+(-0.584607978668)*x[0]+(0.669056948658)*x[1]**o+(-0.393158541985)*x[1]
        arg[5,0,1]=(-0.471551968206)*x[0]**o+(-0.339964831396)*x[0]+(0.354554393554)*x[1]**o+(-0.334773591957)*x[1]
        arg[5,1,0]=(0.113562904864)*x[0]**o+(-0.775045505123)*x[0]+(0.783044874931)*x[1]**o+(0.52265843873)*x[1]
        arg[5,1,1]=(-0.115476036205)*x[0]**o+(-0.290822012728)*x[0]+(-0.398510477854)*x[1]**o+(0.398737205542)*x[1]
        ref[0,0,0]=(0.440499089362)*(1+2.*(dim-1.)/(o+1.))+(1.00012639577)*dim
        ref[0,0,1]=(-0.262966578826)*(1+2.*(dim-1.)/(o+1.))+(-0.888350056938)*dim
        ref[0,1,0]=(-0.741479077167)*(1+2.*(dim-1.)/(o+1.))+(0.931223221336)*dim
        ref[0,1,1]=(-1.04305002243)*(1+2.*(dim-1.)/(o+1.))+(1.87452791805)*dim
        ref[1,0,0]=(0.143941913005)*(1+2.*(dim-1.)/(o+1.))+(-0.873358354816)*dim
        ref[1,0,1]=(-1.6673983903)*(1+2.*(dim-1.)/(o+1.))+(0.253523568676)*dim
        ref[1,1,0]=(0.665232137068)*(1+2.*(dim-1.)/(o+1.))+(1.04406099182)*dim
        ref[1,1,1]=(-0.855338073283)*(1+2.*(dim-1.)/(o+1.))+(-1.19508920116)*dim
        ref[2,0,0]=(1.19306356076)*(1+2.*(dim-1.)/(o+1.))+(-0.680006770164)*dim
        ref[2,0,1]=(0.0570227817187)*(1+2.*(dim-1.)/(o+1.))+(-0.870537211764)*dim
        ref[2,1,0]=(-0.0402582285028)*(1+2.*(dim-1.)/(o+1.))+(-0.682108493287)*dim
        ref[2,1,1]=(0.228444553263)*(1+2.*(dim-1.)/(o+1.))+(0.0620305162089)*dim
        ref[3,0,0]=(-0.758644800093)*(1+2.*(dim-1.)/(o+1.))+(-0.867938181611)*dim
        ref[3,0,1]=(-1.15449397534)*(1+2.*(dim-1.)/(o+1.))+(0.742000116921)*dim
        ref[3,1,0]=(0.317824936505)*(1+2.*(dim-1.)/(o+1.))+(0.457110978744)*dim
        ref[3,1,1]=(0.822330994578)*(1+2.*(dim-1.)/(o+1.))+(0.798054040789)*dim
        ref[4,0,0]=(0.173798775954)*(1+2.*(dim-1.)/(o+1.))+(0.0292780667983)*dim
        ref[4,0,1]=(-0.403565049303)*(1+2.*(dim-1.)/(o+1.))+(-0.0965639721601)*dim
        ref[4,1,0]=(0.157682670995)*(1+2.*(dim-1.)/(o+1.))+(0.466087282085)*dim
        ref[4,1,1]=(-0.68816006255)*(1+2.*(dim-1.)/(o+1.))+(0.914425627528)*dim
        ref[5,0,0]=(1.49672711817)*(1+2.*(dim-1.)/(o+1.))+(-0.977766520654)*dim
        ref[5,0,1]=(-0.116997574651)*(1+2.*(dim-1.)/(o+1.))+(-0.674738423353)*dim
        ref[5,1,0]=(0.896607779795)*(1+2.*(dim-1.)/(o+1.))+(-0.252387066394)*dim
        ref[5,1,1]=(-0.513986514058)*(1+2.*(dim-1.)/(o+1.))+(0.107915192814)*dim
      else:
        arg[0,0,0]=(-0.418671467289)*x[0]**o+(0.498387164484)*x[0]+(0.208896299314)*x[1]**o+(-0.643038678771)*x[1]+(-0.85475466439)*x[2]**o+(0.59586368009)*x[2]
        arg[0,0,1]=(0.149229891253)*x[0]**o+(0.229836799955)*x[0]+(0.557946253001)*x[1]**o+(-0.434223003739)*x[1]+(-0.250979831671)*x[2]**o+(0.762582320269)*x[2]
        arg[0,1,0]=(-0.950902166094)*x[0]**o+(-0.312913424444)*x[0]+(-0.573887973428)*x[1]**o+(-0.372485255069)*x[1]+(-0.18228472562)*x[2]**o+(-0.566679093436)*x[2]
        arg[0,1,1]=(-0.191769651257)*x[0]**o+(-0.344504306057)*x[0]+(0.355926399715)*x[1]**o+(-0.721341598375)*x[1]+(0.362401271289)*x[2]**o+(0.601351093654)*x[2]
        arg[1,0,0]=(-0.375572370375)*x[0]**o+(-0.532723139333)*x[0]+(-0.738166755424)*x[1]**o+(0.785506757134)*x[1]+(0.192667976423)*x[2]**o+(-0.06534893872)*x[2]
        arg[1,0,1]=(-0.353750712453)*x[0]**o+(0.401872866285)*x[0]+(-0.747783222746)*x[1]**o+(0.663116752574)*x[1]+(-0.162954382466)*x[2]**o+(0.318112656687)*x[2]
        arg[1,1,0]=(-0.711345828912)*x[0]**o+(0.764701438614)*x[0]+(-0.232576947703)*x[1]**o+(-0.50541791533)*x[1]+(0.604603682696)*x[2]**o+(0.498828792648)*x[2]
        arg[1,1,1]=(-0.72474879447)*x[0]**o+(0.888266999605)*x[0]+(0.639903282531)*x[1]**o+(-0.777435844156)*x[1]+(-0.0814738680964)*x[2]**o+(-0.0135286853237)*x[2]
        arg[2,0,0]=(0.839793308998)*x[0]**o+(0.374051014226)*x[0]+(0.594653957466)*x[1]**o+(-0.962352782345)*x[1]+(0.888632514336)*x[2]**o+(0.491195067464)*x[2]
        arg[2,0,1]=(-0.827615661598)*x[0]**o+(0.675227575488)*x[0]+(-0.93068580658)*x[1]**o+(0.694434337954)*x[1]+(0.3583179263)*x[2]**o+(0.775147111187)*x[2]
        arg[2,1,0]=(-0.772656999491)*x[0]**o+(-0.464476706435)*x[0]+(0.569832694977)*x[1]**o+(0.500517676542)*x[1]+(-0.272153062279)*x[2]**o+(-0.398436498506)*x[2]
        arg[2,1,1]=(-0.3065198839)*x[0]**o+(-0.454353346315)*x[0]+(-0.539448433869)*x[1]**o+(0.776950547078)*x[1]+(-0.987178147402)*x[2]**o+(0.368935549126)*x[2]
        arg[3,0,0]=(-0.516131622594)*x[0]**o+(-0.680922252714)*x[0]+(0.304951059501)*x[1]**o+(-0.236361364379)*x[1]+(-0.37983306233)*x[2]**o+(0.693426329316)*x[2]
        arg[3,0,1]=(0.85094983056)*x[0]**o+(-0.356782375867)*x[0]+(0.362116478047)*x[1]**o+(0.269126381543)*x[1]+(0.950254707197)*x[2]**o+(0.592028802301)*x[2]
        arg[3,1,0]=(-0.17721437307)*x[0]**o+(-0.127432965766)*x[0]+(-0.754975923735)*x[1]**o+(0.755938625221)*x[1]+(0.776393089777)*x[2]**o+(0.913028629874)*x[2]
        arg[3,1,1]=(-0.779840479711)*x[0]**o+(0.432392796542)*x[0]+(0.0457670482312)*x[1]**o+(0.493363120993)*x[1]+(-0.290308698356)*x[2]**o+(0.263565329052)*x[2]
        arg[4,0,0]=(-0.594371032991)*x[0]**o+(0.673573089348)*x[0]+(0.0301238273919)*x[1]**o+(-0.900766544194)*x[1]+(0.937125579182)*x[2]**o+(-0.360512287398)*x[2]
        arg[4,0,1]=(-0.187561420369)*x[0]**o+(-0.727058667406)*x[0]+(0.643904893485)*x[1]**o+(0.547358609515)*x[1]+(-0.810276689361)*x[2]**o+(0.347967967049)*x[2]
        arg[4,1,0]=(0.0261452776327)*x[0]**o+(0.630125453041)*x[0]+(0.884509651558)*x[1]**o+(-0.580251885597)*x[1]+(0.141608918047)*x[2]**o+(0.417140237181)*x[2]
        arg[4,1,1]=(0.943149763428)*x[0]**o+(0.0414421588123)*x[0]+(-0.453944232191)*x[1]**o+(0.817963360891)*x[1]+(-0.0132464046434)*x[2]**o+(-0.247024507218)*x[2]
        arg[5,0,0]=(0.993356103007)*x[0]**o+(-0.604991443085)*x[0]+(-0.0141942193536)*x[1]**o+(-0.467982818195)*x[1]+(0.907608332852)*x[2]**o+(-0.918494340721)*x[2]
        arg[5,0,1]=(0.782079295393)*x[0]**o+(0.0125642236587)*x[0]+(-0.512764292821)*x[1]**o+(-0.0699705327379)*x[1]+(0.141341885735)*x[2]**o+(0.420508183565)*x[2]
        arg[5,1,0]=(-0.835665486427)*x[0]**o+(-0.536850925449)*x[0]+(-0.0958499285885)*x[1]**o+(-0.541119926637)*x[1]+(0.341886033336)*x[2]**o+(-0.772823771771)*x[2]
        arg[5,1,1]=(-0.391737530391)*x[0]**o+(-0.160045087084)*x[0]+(-0.984350699381)*x[1]**o+(0.303782477449)*x[1]+(-0.279906615513)*x[2]**o+(-0.690872289955)*x[2]
        ref[0,0,0]=(-1.06452983237)*(1+2.*(dim-1.)/(o+1.))+(0.451212165803)*dim
        ref[0,0,1]=(0.456196312583)*(1+2.*(dim-1.)/(o+1.))+(0.558196116486)*dim
        ref[0,1,0]=(-1.70707486514)*(1+2.*(dim-1.)/(o+1.))+(-1.25207777295)*dim
        ref[0,1,1]=(0.526558019747)*(1+2.*(dim-1.)/(o+1.))+(-0.464494810779)*dim
        ref[1,0,0]=(-0.921071149376)*(1+2.*(dim-1.)/(o+1.))+(0.187434679081)*dim
        ref[1,0,1]=(-1.26448831766)*(1+2.*(dim-1.)/(o+1.))+(1.38310227555)*dim
        ref[1,1,0]=(-0.339319093919)*(1+2.*(dim-1.)/(o+1.))+(0.758112315932)*dim
        ref[1,1,1]=(-0.166319380035)*(1+2.*(dim-1.)/(o+1.))+(0.0973024701251)*dim
        ref[2,0,0]=(2.3230797808)*(1+2.*(dim-1.)/(o+1.))+(-0.0971067006546)*dim
        ref[2,0,1]=(-1.39998354188)*(1+2.*(dim-1.)/(o+1.))+(2.14480902463)*dim
        ref[2,1,0]=(-0.474977366793)*(1+2.*(dim-1.)/(o+1.))+(-0.362395528399)*dim
        ref[2,1,1]=(-1.83314646517)*(1+2.*(dim-1.)/(o+1.))+(0.69153274989)*dim
        ref[3,0,0]=(-0.591013625423)*(1+2.*(dim-1.)/(o+1.))+(-0.223857287777)*dim
        ref[3,0,1]=(2.1633210158)*(1+2.*(dim-1.)/(o+1.))+(0.504372807977)*dim
        ref[3,1,0]=(-0.155797207029)*(1+2.*(dim-1.)/(o+1.))+(1.54153428933)*dim
        ref[3,1,1]=(-1.02438212984)*(1+2.*(dim-1.)/(o+1.))+(1.18932124659)*dim
        ref[4,0,0]=(0.372878373583)*(1+2.*(dim-1.)/(o+1.))+(-0.587705742243)*dim
        ref[4,0,1]=(-0.353933216245)*(1+2.*(dim-1.)/(o+1.))+(0.168267909158)*dim
        ref[4,1,0]=(1.05226384724)*(1+2.*(dim-1.)/(o+1.))+(0.467013804625)*dim
        ref[4,1,1]=(0.475959126593)*(1+2.*(dim-1.)/(o+1.))+(0.612381012486)*dim
        ref[5,0,0]=(1.88677021651)*(1+2.*(dim-1.)/(o+1.))+(-1.991468602)*dim
        ref[5,0,1]=(0.410656888307)*(1+2.*(dim-1.)/(o+1.))+(0.363101874486)*dim
        ref[5,1,0]=(-0.58962938168)*(1+2.*(dim-1.)/(o+1.))+(-1.85079462386)*dim
        ref[5,1,1]=(-1.65599484529)*(1+2.*(dim-1.)/(o+1.))+(-0.54713489959)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.606358724075)*x[0]**o+(-0.6326664687)*x[0]+(-0.521101256804)*x[1]**o+(0.702113497865)*x[1]
        arg[0,0,0,1]=(0.528315511137)*x[0]**o+(-0.116566989978)*x[0]+(-0.350847384045)*x[1]**o+(-0.581988878345)*x[1]
        arg[0,0,1,0]=(0.565076347032)*x[0]**o+(0.752532782601)*x[0]+(0.00706163321089)*x[1]**o+(-0.89265244135)*x[1]
        arg[0,0,1,1]=(-0.342136497951)*x[0]**o+(0.753153993866)*x[0]+(-0.199643944066)*x[1]**o+(0.995411752312)*x[1]
        arg[0,0,2,0]=(0.621283193015)*x[0]**o+(0.591186318333)*x[0]+(0.880169584588)*x[1]**o+(-0.11839698224)*x[1]
        arg[0,0,2,1]=(-0.831745199475)*x[0]**o+(0.293656166596)*x[0]+(0.0717745661823)*x[1]**o+(0.827680695965)*x[1]
        arg[0,1,0,0]=(0.686864676707)*x[0]**o+(-0.6552976723)*x[0]+(-0.65589888943)*x[1]**o+(-0.878054891784)*x[1]
        arg[0,1,0,1]=(-0.43359086203)*x[0]**o+(0.723819586663)*x[0]+(0.0741799035834)*x[1]**o+(0.13005305188)*x[1]
        arg[0,1,1,0]=(-0.0605346127445)*x[0]**o+(-0.717729061173)*x[0]+(0.825539660518)*x[1]**o+(0.873033938391)*x[1]
        arg[0,1,1,1]=(-0.286483607332)*x[0]**o+(-0.577255191498)*x[0]+(-0.328362280683)*x[1]**o+(0.499078653745)*x[1]
        arg[0,1,2,0]=(-0.334132932545)*x[0]**o+(0.462461110225)*x[0]+(-0.932311565873)*x[1]**o+(0.934683608666)*x[1]
        arg[0,1,2,1]=(-0.383203086272)*x[0]**o+(0.261035781624)*x[0]+(-0.203936125231)*x[1]**o+(0.657584310755)*x[1]
        arg[0,2,0,0]=(-0.637642146695)*x[0]**o+(0.80524493224)*x[0]+(-0.139137135459)*x[1]**o+(0.947434162736)*x[1]
        arg[0,2,0,1]=(-0.0738996477961)*x[0]**o+(0.0663360260869)*x[0]+(-0.599113718918)*x[1]**o+(-0.349514990363)*x[1]
        arg[0,2,1,0]=(-0.0539211250603)*x[0]**o+(0.262567947534)*x[0]+(0.115386681137)*x[1]**o+(-0.68179682822)*x[1]
        arg[0,2,1,1]=(-0.770114869627)*x[0]**o+(-0.388068404619)*x[0]+(-0.712914716647)*x[1]**o+(-0.810536809871)*x[1]
        arg[0,2,2,0]=(0.274425895455)*x[0]**o+(0.338467327995)*x[0]+(-0.78223636363)*x[1]**o+(0.406149090351)*x[1]
        arg[0,2,2,1]=(-0.748672361572)*x[0]**o+(0.508311939723)*x[0]+(0.875797107098)*x[1]**o+(0.0992795348726)*x[1]
        arg[0,3,0,0]=(-0.547350688209)*x[0]**o+(0.671683747412)*x[0]+(-0.0954212965969)*x[1]**o+(0.983487887304)*x[1]
        arg[0,3,0,1]=(0.626104338883)*x[0]**o+(-0.825311449834)*x[0]+(-0.422521614246)*x[1]**o+(0.304208397592)*x[1]
        arg[0,3,1,0]=(0.698777992012)*x[0]**o+(0.00291670095874)*x[0]+(0.327527809646)*x[1]**o+(0.283289333886)*x[1]
        arg[0,3,1,1]=(-0.0197648150597)*x[0]**o+(-0.902860814776)*x[0]+(0.663961228673)*x[1]**o+(0.675952386365)*x[1]
        arg[0,3,2,0]=(0.286747814345)*x[0]**o+(-0.0641946364041)*x[0]+(0.812019228115)*x[1]**o+(-0.225698609001)*x[1]
        arg[0,3,2,1]=(0.879776322226)*x[0]**o+(-0.251856143047)*x[0]+(-0.252403834757)*x[1]**o+(0.650790406323)*x[1]
        arg[0,4,0,0]=(-0.58782572388)*x[0]**o+(-0.144122228965)*x[0]+(0.521707348991)*x[1]**o+(-0.0460840614459)*x[1]
        arg[0,4,0,1]=(-0.777497837497)*x[0]**o+(-0.229011563089)*x[0]+(0.754657080443)*x[1]**o+(-0.690675521089)*x[1]
        arg[0,4,1,0]=(0.351528557097)*x[0]**o+(0.00192986530026)*x[0]+(-0.414849484269)*x[1]**o+(-0.250205659505)*x[1]
        arg[0,4,1,1]=(-0.814037796843)*x[0]**o+(0.00188512986627)*x[0]+(-0.810799710998)*x[1]**o+(-0.114406174711)*x[1]
        arg[0,4,2,0]=(-0.755511212878)*x[0]**o+(-0.502206951212)*x[0]+(0.555738333365)*x[1]**o+(-0.220359547926)*x[1]
        arg[0,4,2,1]=(0.247108894345)*x[0]**o+(0.479239235148)*x[0]+(0.663958827816)*x[1]**o+(0.220938477625)*x[1]
        arg[1,0,0,0]=(0.818774052595)*x[0]**o+(-0.651584728881)*x[0]+(-0.780674092348)*x[1]**o+(-0.397277502644)*x[1]
        arg[1,0,0,1]=(0.482101508249)*x[0]**o+(-0.523618202277)*x[0]+(0.806406836488)*x[1]**o+(-0.883738297664)*x[1]
        arg[1,0,1,0]=(0.272517667346)*x[0]**o+(-0.700461889638)*x[0]+(0.280736354342)*x[1]**o+(0.182983247089)*x[1]
        arg[1,0,1,1]=(0.566589111451)*x[0]**o+(-0.876744547538)*x[0]+(-0.934741703566)*x[1]**o+(-0.562770667492)*x[1]
        arg[1,0,2,0]=(-0.954128714623)*x[0]**o+(0.115699770127)*x[0]+(-0.580663262426)*x[1]**o+(-0.451619694161)*x[1]
        arg[1,0,2,1]=(0.268605150185)*x[0]**o+(-0.233415174179)*x[0]+(0.477143801066)*x[1]**o+(0.443696994244)*x[1]
        arg[1,1,0,0]=(-0.00240846686485)*x[0]**o+(0.776336851643)*x[0]+(0.488314964094)*x[1]**o+(-0.840196476788)*x[1]
        arg[1,1,0,1]=(-0.511732155638)*x[0]**o+(0.528782039485)*x[0]+(0.51274502262)*x[1]**o+(-0.700213610705)*x[1]
        arg[1,1,1,0]=(0.300359000499)*x[0]**o+(0.405101876401)*x[0]+(-0.47022583299)*x[1]**o+(-0.977634916267)*x[1]
        arg[1,1,1,1]=(-0.494482318948)*x[0]**o+(0.254112728167)*x[0]+(-0.541672595343)*x[1]**o+(-0.270970064522)*x[1]
        arg[1,1,2,0]=(-0.586624294233)*x[0]**o+(0.629601920904)*x[0]+(-0.525092252936)*x[1]**o+(-0.44611635978)*x[1]
        arg[1,1,2,1]=(0.716927085139)*x[0]**o+(0.700193925267)*x[0]+(0.0624921446647)*x[1]**o+(0.109485129877)*x[1]
        arg[1,2,0,0]=(0.325729307898)*x[0]**o+(0.773432634097)*x[0]+(-0.675612597356)*x[1]**o+(-0.0554913733692)*x[1]
        arg[1,2,0,1]=(-0.381606382262)*x[0]**o+(-0.689040382854)*x[0]+(-0.580191973398)*x[1]**o+(0.951689147387)*x[1]
        arg[1,2,1,0]=(-0.527642403744)*x[0]**o+(-0.936915203927)*x[0]+(0.631733997289)*x[1]**o+(-0.715975680015)*x[1]
        arg[1,2,1,1]=(0.0886898373323)*x[0]**o+(-0.0943320873915)*x[0]+(0.470668750443)*x[1]**o+(0.773832960093)*x[1]
        arg[1,2,2,0]=(-0.727720290197)*x[0]**o+(0.422119803173)*x[0]+(0.705882951779)*x[1]**o+(-0.350801865244)*x[1]
        arg[1,2,2,1]=(0.879335475407)*x[0]**o+(-0.0307168471577)*x[0]+(0.458578151544)*x[1]**o+(0.788796576339)*x[1]
        arg[1,3,0,0]=(0.182143508146)*x[0]**o+(-0.680787176288)*x[0]+(0.626099489145)*x[1]**o+(-0.771540378408)*x[1]
        arg[1,3,0,1]=(-0.578341681282)*x[0]**o+(0.856407179963)*x[0]+(0.525319425356)*x[1]**o+(0.345651851179)*x[1]
        arg[1,3,1,0]=(-0.316652838058)*x[0]**o+(0.652409898929)*x[0]+(-0.82123442746)*x[1]**o+(-0.136868558081)*x[1]
        arg[1,3,1,1]=(-0.755530700082)*x[0]**o+(0.354429731374)*x[0]+(-0.14330437792)*x[1]**o+(-0.221454495186)*x[1]
        arg[1,3,2,0]=(-0.520261483769)*x[0]**o+(0.0873707179205)*x[0]+(-0.882995356682)*x[1]**o+(-0.773921687804)*x[1]
        arg[1,3,2,1]=(0.404278952718)*x[0]**o+(0.651166017764)*x[0]+(-0.496695253747)*x[1]**o+(0.625671754182)*x[1]
        arg[1,4,0,0]=(-0.730831539043)*x[0]**o+(0.219467989041)*x[0]+(-0.853784008431)*x[1]**o+(0.274799624897)*x[1]
        arg[1,4,0,1]=(0.893313801855)*x[0]**o+(-0.539687651773)*x[0]+(-0.411582029239)*x[1]**o+(-0.293521483805)*x[1]
        arg[1,4,1,0]=(0.484878049594)*x[0]**o+(0.218930293134)*x[0]+(-0.138362733141)*x[1]**o+(-0.663241354177)*x[1]
        arg[1,4,1,1]=(-0.608410207857)*x[0]**o+(-0.0921006917331)*x[0]+(-0.885705130565)*x[1]**o+(0.103724764404)*x[1]
        arg[1,4,2,0]=(0.150946260009)*x[0]**o+(-0.601834829203)*x[0]+(0.0141091402169)*x[1]**o+(-0.98316187826)*x[1]
        arg[1,4,2,1]=(0.98764918249)*x[0]**o+(-0.841155365687)*x[0]+(0.185525084432)*x[1]**o+(-0.613421841538)*x[1]
        arg[2,0,0,0]=(0.485337970037)*x[0]**o+(-0.898602108837)*x[0]+(0.847812629133)*x[1]**o+(-0.23522051021)*x[1]
        arg[2,0,0,1]=(0.835641029366)*x[0]**o+(-0.0524661277338)*x[0]+(0.41697147238)*x[1]**o+(-0.622813671618)*x[1]
        arg[2,0,1,0]=(0.505181088966)*x[0]**o+(-0.870141755006)*x[0]+(0.451139050304)*x[1]**o+(0.957790881903)*x[1]
        arg[2,0,1,1]=(0.822955028051)*x[0]**o+(-0.103028201548)*x[0]+(-0.604614127349)*x[1]**o+(-0.655334670762)*x[1]
        arg[2,0,2,0]=(-0.166647802545)*x[0]**o+(-0.466885186725)*x[0]+(0.0782333081578)*x[1]**o+(0.814414769115)*x[1]
        arg[2,0,2,1]=(-0.980310280968)*x[0]**o+(-0.884306329761)*x[0]+(-0.895550319566)*x[1]**o+(0.160657671233)*x[1]
        arg[2,1,0,0]=(-0.846405435483)*x[0]**o+(-0.625465093616)*x[0]+(-0.0789995975268)*x[1]**o+(-0.440180761086)*x[1]
        arg[2,1,0,1]=(-0.949855662419)*x[0]**o+(-0.013707742372)*x[0]+(0.0397729185783)*x[1]**o+(-0.836054083952)*x[1]
        arg[2,1,1,0]=(0.427246928002)*x[0]**o+(0.835369889909)*x[0]+(0.925300799558)*x[1]**o+(-0.328266770631)*x[1]
        arg[2,1,1,1]=(-0.36143365614)*x[0]**o+(-0.744107540819)*x[0]+(-0.358554131639)*x[1]**o+(-0.562674543946)*x[1]
        arg[2,1,2,0]=(0.60580594275)*x[0]**o+(0.644796993923)*x[0]+(-0.682407021927)*x[1]**o+(0.938016908061)*x[1]
        arg[2,1,2,1]=(-0.934393274317)*x[0]**o+(-0.718420272664)*x[0]+(0.0357039032065)*x[1]**o+(0.446387943344)*x[1]
        arg[2,2,0,0]=(-0.329105411737)*x[0]**o+(0.588935500072)*x[0]+(-0.833362483432)*x[1]**o+(0.629996961074)*x[1]
        arg[2,2,0,1]=(0.723977530226)*x[0]**o+(0.0698430385616)*x[0]+(-0.503761809156)*x[1]**o+(0.907472279839)*x[1]
        arg[2,2,1,0]=(-0.767545510793)*x[0]**o+(0.615094783682)*x[0]+(0.894594933703)*x[1]**o+(0.922357219049)*x[1]
        arg[2,2,1,1]=(-0.155739055576)*x[0]**o+(0.646142299486)*x[0]+(0.865932681534)*x[1]**o+(-0.126887380871)*x[1]
        arg[2,2,2,0]=(-0.629534320112)*x[0]**o+(-0.0904366301836)*x[0]+(0.193019806731)*x[1]**o+(-0.866080160466)*x[1]
        arg[2,2,2,1]=(-0.714527761053)*x[0]**o+(-0.813712879405)*x[0]+(-0.291734238438)*x[1]**o+(0.232675536415)*x[1]
        arg[2,3,0,0]=(-0.447474292879)*x[0]**o+(-0.639046887788)*x[0]+(0.159616683647)*x[1]**o+(0.564127141623)*x[1]
        arg[2,3,0,1]=(-0.608953936619)*x[0]**o+(0.273179484103)*x[0]+(0.42693156927)*x[1]**o+(-0.983331075647)*x[1]
        arg[2,3,1,0]=(0.855859483301)*x[0]**o+(-0.432495503925)*x[0]+(-0.67928203769)*x[1]**o+(-0.856541006702)*x[1]
        arg[2,3,1,1]=(-0.581185594204)*x[0]**o+(0.947720323693)*x[0]+(-0.520212852628)*x[1]**o+(-0.37563221594)*x[1]
        arg[2,3,2,0]=(0.351358492535)*x[0]**o+(0.367210525846)*x[0]+(0.97870179383)*x[1]**o+(-0.199559052968)*x[1]
        arg[2,3,2,1]=(0.797294518541)*x[0]**o+(0.546397721552)*x[0]+(0.374121851646)*x[1]**o+(0.704436689798)*x[1]
        arg[2,4,0,0]=(-0.756047775149)*x[0]**o+(-0.705403684825)*x[0]+(-0.662421213376)*x[1]**o+(0.390069314234)*x[1]
        arg[2,4,0,1]=(-0.658842889869)*x[0]**o+(0.766273769822)*x[0]+(-0.779201775962)*x[1]**o+(-0.968670270742)*x[1]
        arg[2,4,1,0]=(-0.336220564628)*x[0]**o+(-0.134745604319)*x[0]+(0.391939888428)*x[1]**o+(-0.113113999792)*x[1]
        arg[2,4,1,1]=(-0.8345363851)*x[0]**o+(0.339223929857)*x[0]+(0.628403353717)*x[1]**o+(0.799364167182)*x[1]
        arg[2,4,2,0]=(-0.898050446701)*x[0]**o+(-0.666373288962)*x[0]+(-0.620778021659)*x[1]**o+(-0.990359850049)*x[1]
        arg[2,4,2,1]=(-0.933931027675)*x[0]**o+(-0.119077281033)*x[0]+(0.557313533556)*x[1]**o+(0.160336158304)*x[1]
        arg[3,0,0,0]=(0.972462636314)*x[0]**o+(0.607913740671)*x[0]+(0.490168539257)*x[1]**o+(-0.12515921391)*x[1]
        arg[3,0,0,1]=(0.856370886333)*x[0]**o+(-0.143225009869)*x[0]+(-0.330381845603)*x[1]**o+(-0.271958059018)*x[1]
        arg[3,0,1,0]=(0.86742129112)*x[0]**o+(-0.0259756341823)*x[0]+(-0.0137047849942)*x[1]**o+(-0.0165444161044)*x[1]
        arg[3,0,1,1]=(-0.445996036941)*x[0]**o+(0.247936136111)*x[0]+(-0.381484740124)*x[1]**o+(0.836678417006)*x[1]
        arg[3,0,2,0]=(0.161387456771)*x[0]**o+(0.687533753311)*x[0]+(-0.463845417603)*x[1]**o+(0.665023823118)*x[1]
        arg[3,0,2,1]=(0.243410612742)*x[0]**o+(-0.52431675734)*x[0]+(0.75630192488)*x[1]**o+(-0.703601557317)*x[1]
        arg[3,1,0,0]=(0.793530689056)*x[0]**o+(0.814806726176)*x[0]+(0.582051068232)*x[1]**o+(0.692432923027)*x[1]
        arg[3,1,0,1]=(-0.283506909465)*x[0]**o+(0.21438128538)*x[0]+(-0.327476122861)*x[1]**o+(0.990178870269)*x[1]
        arg[3,1,1,0]=(-0.117386491827)*x[0]**o+(0.747771272481)*x[0]+(-0.439767754004)*x[1]**o+(-0.905218338639)*x[1]
        arg[3,1,1,1]=(-0.544805647706)*x[0]**o+(-0.0484033735345)*x[0]+(0.861443439156)*x[1]**o+(-0.992086899129)*x[1]
        arg[3,1,2,0]=(0.596208599112)*x[0]**o+(-0.276136075021)*x[0]+(0.079604120535)*x[1]**o+(0.124232758054)*x[1]
        arg[3,1,2,1]=(0.809847641841)*x[0]**o+(0.956041370087)*x[0]+(0.65224420189)*x[1]**o+(0.565327743771)*x[1]
        arg[3,2,0,0]=(-0.264599479107)*x[0]**o+(-0.276726556903)*x[0]+(0.577911819561)*x[1]**o+(-0.656157381756)*x[1]
        arg[3,2,0,1]=(-0.0636193103458)*x[0]**o+(0.941482878452)*x[0]+(0.877581030494)*x[1]**o+(-0.254911385455)*x[1]
        arg[3,2,1,0]=(-0.28083673633)*x[0]**o+(0.849095032935)*x[0]+(0.6825284068)*x[1]**o+(0.445605171749)*x[1]
        arg[3,2,1,1]=(-0.729331034619)*x[0]**o+(0.036400895394)*x[0]+(0.365733359214)*x[1]**o+(0.0305324156609)*x[1]
        arg[3,2,2,0]=(-0.70932226428)*x[0]**o+(0.310455571562)*x[0]+(-0.101305643894)*x[1]**o+(0.346873071627)*x[1]
        arg[3,2,2,1]=(0.712053329488)*x[0]**o+(-0.845042122758)*x[0]+(0.909517101069)*x[1]**o+(-0.163109058477)*x[1]
        arg[3,3,0,0]=(0.451744730688)*x[0]**o+(0.308756227526)*x[0]+(0.13963307493)*x[1]**o+(-0.945789640023)*x[1]
        arg[3,3,0,1]=(-0.852609785834)*x[0]**o+(-0.821513824231)*x[0]+(-0.544690113726)*x[1]**o+(0.335547843106)*x[1]
        arg[3,3,1,0]=(0.337600834432)*x[0]**o+(-0.170181468776)*x[0]+(-0.302415145845)*x[1]**o+(0.985425450212)*x[1]
        arg[3,3,1,1]=(-0.638242591048)*x[0]**o+(-0.0221442303187)*x[0]+(0.421311946694)*x[1]**o+(-0.43495081316)*x[1]
        arg[3,3,2,0]=(-0.734168622382)*x[0]**o+(-0.889300572069)*x[0]+(0.808161580285)*x[1]**o+(-0.526348332439)*x[1]
        arg[3,3,2,1]=(0.998918324168)*x[0]**o+(-0.953729666152)*x[0]+(0.665101882861)*x[1]**o+(-0.581118145725)*x[1]
        arg[3,4,0,0]=(-0.251226888098)*x[0]**o+(-0.711615999514)*x[0]+(0.291022154479)*x[1]**o+(0.0772744442604)*x[1]
        arg[3,4,0,1]=(0.162379537091)*x[0]**o+(0.363750962136)*x[0]+(0.635943839083)*x[1]**o+(0.513856953389)*x[1]
        arg[3,4,1,0]=(-0.99052354199)*x[0]**o+(0.289202668813)*x[0]+(-0.759078503012)*x[1]**o+(-0.588801682903)*x[1]
        arg[3,4,1,1]=(-0.942278699306)*x[0]**o+(0.476003512559)*x[0]+(0.873920293292)*x[1]**o+(0.867413991424)*x[1]
        arg[3,4,2,0]=(-0.602386702281)*x[0]**o+(0.694432164577)*x[0]+(0.367008950869)*x[1]**o+(-0.803331549449)*x[1]
        arg[3,4,2,1]=(-0.00204465287104)*x[0]**o+(-0.697572466685)*x[0]+(-0.84755978799)*x[1]**o+(-0.758946055726)*x[1]
        ref[0,0,0,0]=(0.0852574672711)*(1+2.*(dim-1.)/(o+1.))+(0.069447029165)*dim
        ref[0,0,0,1]=(0.177468127092)*(1+2.*(dim-1.)/(o+1.))+(-0.698555868323)*dim
        ref[0,0,1,0]=(0.572137980243)*(1+2.*(dim-1.)/(o+1.))+(-0.140119658749)*dim
        ref[0,0,1,1]=(-0.541780442017)*(1+2.*(dim-1.)/(o+1.))+(1.74856574618)*dim
        ref[0,0,2,0]=(1.5014527776)*(1+2.*(dim-1.)/(o+1.))+(0.472789336092)*dim
        ref[0,0,2,1]=(-0.759970633293)*(1+2.*(dim-1.)/(o+1.))+(1.12133686256)*dim
        ref[0,1,0,0]=(0.030965787278)*(1+2.*(dim-1.)/(o+1.))+(-1.53335256408)*dim
        ref[0,1,0,1]=(-0.359410958446)*(1+2.*(dim-1.)/(o+1.))+(0.853872638544)*dim
        ref[0,1,1,0]=(0.765005047774)*(1+2.*(dim-1.)/(o+1.))+(0.155304877218)*dim
        ref[0,1,1,1]=(-0.614845888015)*(1+2.*(dim-1.)/(o+1.))+(-0.0781765377528)*dim
        ref[0,1,2,0]=(-1.26644449842)*(1+2.*(dim-1.)/(o+1.))+(1.39714471889)*dim
        ref[0,1,2,1]=(-0.587139211502)*(1+2.*(dim-1.)/(o+1.))+(0.918620092379)*dim
        ref[0,2,0,0]=(-0.776779282155)*(1+2.*(dim-1.)/(o+1.))+(1.75267909498)*dim
        ref[0,2,0,1]=(-0.673013366714)*(1+2.*(dim-1.)/(o+1.))+(-0.283178964276)*dim
        ref[0,2,1,0]=(0.0614655560763)*(1+2.*(dim-1.)/(o+1.))+(-0.419228880686)*dim
        ref[0,2,1,1]=(-1.48302958627)*(1+2.*(dim-1.)/(o+1.))+(-1.19860521449)*dim
        ref[0,2,2,0]=(-0.507810468175)*(1+2.*(dim-1.)/(o+1.))+(0.744616418346)*dim
        ref[0,2,2,1]=(0.127124745526)*(1+2.*(dim-1.)/(o+1.))+(0.607591474596)*dim
        ref[0,3,0,0]=(-0.642771984806)*(1+2.*(dim-1.)/(o+1.))+(1.65517163472)*dim
        ref[0,3,0,1]=(0.203582724637)*(1+2.*(dim-1.)/(o+1.))+(-0.521103052242)*dim
        ref[0,3,1,0]=(1.02630580166)*(1+2.*(dim-1.)/(o+1.))+(0.286206034845)*dim
        ref[0,3,1,1]=(0.644196413613)*(1+2.*(dim-1.)/(o+1.))+(-0.226908428411)*dim
        ref[0,3,2,0]=(1.09876704246)*(1+2.*(dim-1.)/(o+1.))+(-0.289893245405)*dim
        ref[0,3,2,1]=(0.627372487469)*(1+2.*(dim-1.)/(o+1.))+(0.398934263276)*dim
        ref[0,4,0,0]=(-0.0661183748894)*(1+2.*(dim-1.)/(o+1.))+(-0.190206290411)*dim
        ref[0,4,0,1]=(-0.022840757054)*(1+2.*(dim-1.)/(o+1.))+(-0.919687084178)*dim
        ref[0,4,1,0]=(-0.063320927172)*(1+2.*(dim-1.)/(o+1.))+(-0.248275794205)*dim
        ref[0,4,1,1]=(-1.62483750784)*(1+2.*(dim-1.)/(o+1.))+(-0.112521044845)*dim
        ref[0,4,2,0]=(-0.199772879513)*(1+2.*(dim-1.)/(o+1.))+(-0.722566499138)*dim
        ref[0,4,2,1]=(0.911067722161)*(1+2.*(dim-1.)/(o+1.))+(0.700177712774)*dim
        ref[1,0,0,0]=(0.0380999602468)*(1+2.*(dim-1.)/(o+1.))+(-1.04886223152)*dim
        ref[1,0,0,1]=(1.28850834474)*(1+2.*(dim-1.)/(o+1.))+(-1.40735649994)*dim
        ref[1,0,1,0]=(0.553254021688)*(1+2.*(dim-1.)/(o+1.))+(-0.517478642549)*dim
        ref[1,0,1,1]=(-0.368152592115)*(1+2.*(dim-1.)/(o+1.))+(-1.43951521503)*dim
        ref[1,0,2,0]=(-1.53479197705)*(1+2.*(dim-1.)/(o+1.))+(-0.335919924034)*dim
        ref[1,0,2,1]=(0.745748951251)*(1+2.*(dim-1.)/(o+1.))+(0.210281820065)*dim
        ref[1,1,0,0]=(0.485906497229)*(1+2.*(dim-1.)/(o+1.))+(-0.0638596251453)*dim
        ref[1,1,0,1]=(0.00101286698233)*(1+2.*(dim-1.)/(o+1.))+(-0.17143157122)*dim
        ref[1,1,1,0]=(-0.169866832491)*(1+2.*(dim-1.)/(o+1.))+(-0.572533039866)*dim
        ref[1,1,1,1]=(-1.03615491429)*(1+2.*(dim-1.)/(o+1.))+(-0.016857336355)*dim
        ref[1,1,2,0]=(-1.11171654717)*(1+2.*(dim-1.)/(o+1.))+(0.183485561125)*dim
        ref[1,1,2,1]=(0.779419229804)*(1+2.*(dim-1.)/(o+1.))+(0.809679055143)*dim
        ref[1,2,0,0]=(-0.349883289458)*(1+2.*(dim-1.)/(o+1.))+(0.717941260728)*dim
        ref[1,2,0,1]=(-0.96179835566)*(1+2.*(dim-1.)/(o+1.))+(0.262648764533)*dim
        ref[1,2,1,0]=(0.104091593546)*(1+2.*(dim-1.)/(o+1.))+(-1.65289088394)*dim
        ref[1,2,1,1]=(0.559358587775)*(1+2.*(dim-1.)/(o+1.))+(0.679500872701)*dim
        ref[1,2,2,0]=(-0.0218373384179)*(1+2.*(dim-1.)/(o+1.))+(0.0713179379289)*dim
        ref[1,2,2,1]=(1.33791362695)*(1+2.*(dim-1.)/(o+1.))+(0.758079729181)*dim
        ref[1,3,0,0]=(0.808242997292)*(1+2.*(dim-1.)/(o+1.))+(-1.4523275547)*dim
        ref[1,3,0,1]=(-0.0530222559262)*(1+2.*(dim-1.)/(o+1.))+(1.20205903114)*dim
        ref[1,3,1,0]=(-1.13788726552)*(1+2.*(dim-1.)/(o+1.))+(0.515541340848)*dim
        ref[1,3,1,1]=(-0.898835078002)*(1+2.*(dim-1.)/(o+1.))+(0.132975236188)*dim
        ref[1,3,2,0]=(-1.40325684045)*(1+2.*(dim-1.)/(o+1.))+(-0.686550969884)*dim
        ref[1,3,2,1]=(-0.0924163010291)*(1+2.*(dim-1.)/(o+1.))+(1.27683777195)*dim
        ref[1,4,0,0]=(-1.58461554747)*(1+2.*(dim-1.)/(o+1.))+(0.494267613937)*dim
        ref[1,4,0,1]=(0.481731772616)*(1+2.*(dim-1.)/(o+1.))+(-0.833209135578)*dim
        ref[1,4,1,0]=(0.346515316453)*(1+2.*(dim-1.)/(o+1.))+(-0.444311061043)*dim
        ref[1,4,1,1]=(-1.49411533842)*(1+2.*(dim-1.)/(o+1.))+(0.0116240726713)*dim
        ref[1,4,2,0]=(0.165055400226)*(1+2.*(dim-1.)/(o+1.))+(-1.58499670746)*dim
        ref[1,4,2,1]=(1.17317426692)*(1+2.*(dim-1.)/(o+1.))+(-1.45457720722)*dim
        ref[2,0,0,0]=(1.33315059917)*(1+2.*(dim-1.)/(o+1.))+(-1.13382261905)*dim
        ref[2,0,0,1]=(1.25261250175)*(1+2.*(dim-1.)/(o+1.))+(-0.675279799352)*dim
        ref[2,0,1,0]=(0.95632013927)*(1+2.*(dim-1.)/(o+1.))+(0.0876491268976)*dim
        ref[2,0,1,1]=(0.218340900702)*(1+2.*(dim-1.)/(o+1.))+(-0.75836287231)*dim
        ref[2,0,2,0]=(-0.0884144943875)*(1+2.*(dim-1.)/(o+1.))+(0.34752958239)*dim
        ref[2,0,2,1]=(-1.87586060053)*(1+2.*(dim-1.)/(o+1.))+(-0.723648658528)*dim
        ref[2,1,0,0]=(-0.92540503301)*(1+2.*(dim-1.)/(o+1.))+(-1.0656458547)*dim
        ref[2,1,0,1]=(-0.910082743841)*(1+2.*(dim-1.)/(o+1.))+(-0.849761826324)*dim
        ref[2,1,1,0]=(1.35254772756)*(1+2.*(dim-1.)/(o+1.))+(0.507103119278)*dim
        ref[2,1,1,1]=(-0.719987787779)*(1+2.*(dim-1.)/(o+1.))+(-1.30678208477)*dim
        ref[2,1,2,0]=(-0.0766010791773)*(1+2.*(dim-1.)/(o+1.))+(1.58281390198)*dim
        ref[2,1,2,1]=(-0.89868937111)*(1+2.*(dim-1.)/(o+1.))+(-0.27203232932)*dim
        ref[2,2,0,0]=(-1.16246789517)*(1+2.*(dim-1.)/(o+1.))+(1.21893246115)*dim
        ref[2,2,0,1]=(0.22021572107)*(1+2.*(dim-1.)/(o+1.))+(0.977315318401)*dim
        ref[2,2,1,0]=(0.12704942291)*(1+2.*(dim-1.)/(o+1.))+(1.53745200273)*dim
        ref[2,2,1,1]=(0.710193625958)*(1+2.*(dim-1.)/(o+1.))+(0.519254918615)*dim
        ref[2,2,2,0]=(-0.43651451338)*(1+2.*(dim-1.)/(o+1.))+(-0.956516790649)*dim
        ref[2,2,2,1]=(-1.00626199949)*(1+2.*(dim-1.)/(o+1.))+(-0.581037342991)*dim
        ref[2,3,0,0]=(-0.287857609232)*(1+2.*(dim-1.)/(o+1.))+(-0.0749197461649)*dim
        ref[2,3,0,1]=(-0.182022367349)*(1+2.*(dim-1.)/(o+1.))+(-0.710151591543)*dim
        ref[2,3,1,0]=(0.176577445611)*(1+2.*(dim-1.)/(o+1.))+(-1.28903651063)*dim
        ref[2,3,1,1]=(-1.10139844683)*(1+2.*(dim-1.)/(o+1.))+(0.572088107754)*dim
        ref[2,3,2,0]=(1.33006028637)*(1+2.*(dim-1.)/(o+1.))+(0.167651472878)*dim
        ref[2,3,2,1]=(1.17141637019)*(1+2.*(dim-1.)/(o+1.))+(1.25083441135)*dim
        ref[2,4,0,0]=(-1.41846898852)*(1+2.*(dim-1.)/(o+1.))+(-0.315334370591)*dim
        ref[2,4,0,1]=(-1.43804466583)*(1+2.*(dim-1.)/(o+1.))+(-0.20239650092)*dim
        ref[2,4,1,0]=(0.0557193237997)*(1+2.*(dim-1.)/(o+1.))+(-0.247859604111)*dim
        ref[2,4,1,1]=(-0.206133031383)*(1+2.*(dim-1.)/(o+1.))+(1.13858809704)*dim
        ref[2,4,2,0]=(-1.51882846836)*(1+2.*(dim-1.)/(o+1.))+(-1.65673313901)*dim
        ref[2,4,2,1]=(-0.376617494119)*(1+2.*(dim-1.)/(o+1.))+(0.041258877271)*dim
        ref[3,0,0,0]=(1.46263117557)*(1+2.*(dim-1.)/(o+1.))+(0.482754526761)*dim
        ref[3,0,0,1]=(0.52598904073)*(1+2.*(dim-1.)/(o+1.))+(-0.415183068887)*dim
        ref[3,0,1,0]=(0.853716506126)*(1+2.*(dim-1.)/(o+1.))+(-0.0425200502867)*dim
        ref[3,0,1,1]=(-0.827480777065)*(1+2.*(dim-1.)/(o+1.))+(1.08461455312)*dim
        ref[3,0,2,0]=(-0.302457960832)*(1+2.*(dim-1.)/(o+1.))+(1.35255757643)*dim
        ref[3,0,2,1]=(0.999712537623)*(1+2.*(dim-1.)/(o+1.))+(-1.22791831466)*dim
        ref[3,1,0,0]=(1.37558175729)*(1+2.*(dim-1.)/(o+1.))+(1.5072396492)*dim
        ref[3,1,0,1]=(-0.610983032326)*(1+2.*(dim-1.)/(o+1.))+(1.20456015565)*dim
        ref[3,1,1,0]=(-0.557154245831)*(1+2.*(dim-1.)/(o+1.))+(-0.157447066158)*dim
        ref[3,1,1,1]=(0.316637791451)*(1+2.*(dim-1.)/(o+1.))+(-1.04049027266)*dim
        ref[3,1,2,0]=(0.675812719647)*(1+2.*(dim-1.)/(o+1.))+(-0.151903316967)*dim
        ref[3,1,2,1]=(1.46209184373)*(1+2.*(dim-1.)/(o+1.))+(1.52136911386)*dim
        ref[3,2,0,0]=(0.313312340454)*(1+2.*(dim-1.)/(o+1.))+(-0.93288393866)*dim
        ref[3,2,0,1]=(0.813961720148)*(1+2.*(dim-1.)/(o+1.))+(0.686571492997)*dim
        ref[3,2,1,0]=(0.40169167047)*(1+2.*(dim-1.)/(o+1.))+(1.29470020468)*dim
        ref[3,2,1,1]=(-0.363597675405)*(1+2.*(dim-1.)/(o+1.))+(0.066933311055)*dim
        ref[3,2,2,0]=(-0.810627908174)*(1+2.*(dim-1.)/(o+1.))+(0.657328643189)*dim
        ref[3,2,2,1]=(1.62157043056)*(1+2.*(dim-1.)/(o+1.))+(-1.00815118124)*dim
        ref[3,3,0,0]=(0.591377805617)*(1+2.*(dim-1.)/(o+1.))+(-0.637033412497)*dim
        ref[3,3,0,1]=(-1.39729989956)*(1+2.*(dim-1.)/(o+1.))+(-0.485965981125)*dim
        ref[3,3,1,0]=(0.0351856885862)*(1+2.*(dim-1.)/(o+1.))+(0.815243981437)*dim
        ref[3,3,1,1]=(-0.216930644354)*(1+2.*(dim-1.)/(o+1.))+(-0.457095043478)*dim
        ref[3,3,2,0]=(0.0739929579029)*(1+2.*(dim-1.)/(o+1.))+(-1.41564890451)*dim
        ref[3,3,2,1]=(1.66402020703)*(1+2.*(dim-1.)/(o+1.))+(-1.53484781188)*dim
        ref[3,4,0,0]=(0.0397952663815)*(1+2.*(dim-1.)/(o+1.))+(-0.634341555254)*dim
        ref[3,4,0,1]=(0.798323376174)*(1+2.*(dim-1.)/(o+1.))+(0.877607915525)*dim
        ref[3,4,1,0]=(-1.749602045)*(1+2.*(dim-1.)/(o+1.))+(-0.29959901409)*dim
        ref[3,4,1,1]=(-0.0683584060144)*(1+2.*(dim-1.)/(o+1.))+(1.34341750398)*dim
        ref[3,4,2,0]=(-0.235377751413)*(1+2.*(dim-1.)/(o+1.))+(-0.108899384872)*dim
        ref[3,4,2,1]=(-0.849604440861)*(1+2.*(dim-1.)/(o+1.))+(-1.45651852241)*dim
      else:
        arg[0,0,0,0]=(-0.229219068739)*x[0]**o+(0.293241016924)*x[0]+(-0.2730739542)*x[1]**o+(0.980438707356)*x[1]+(-0.712463216466)*x[2]**o+(0.201036317883)*x[2]
        arg[0,0,0,1]=(0.626271947422)*x[0]**o+(0.379293769575)*x[0]+(-0.476378592342)*x[1]**o+(-0.0089906260934)*x[1]+(-0.433154385811)*x[2]**o+(0.126766298748)*x[2]
        arg[0,0,1,0]=(-0.517950283035)*x[0]**o+(-0.219780986066)*x[0]+(0.345048580545)*x[1]**o+(0.116833851128)*x[1]+(-0.790641523314)*x[2]**o+(0.40506306417)*x[2]
        arg[0,0,1,1]=(0.910924420538)*x[0]**o+(0.281102231596)*x[0]+(0.115454208619)*x[1]**o+(0.164254056171)*x[1]+(0.44960060103)*x[2]**o+(-0.14238069165)*x[2]
        arg[0,0,2,0]=(0.86195808217)*x[0]**o+(0.775425956811)*x[0]+(0.214559021523)*x[1]**o+(0.400663063038)*x[1]+(0.995663608954)*x[2]**o+(0.228658543268)*x[2]
        arg[0,0,2,1]=(0.9025917492)*x[0]**o+(0.643205662573)*x[0]+(-0.304382459664)*x[1]**o+(-0.653256441532)*x[1]+(0.591142459807)*x[2]**o+(-0.888899030043)*x[2]
        arg[0,1,0,0]=(-0.874720784838)*x[0]**o+(-0.961456282274)*x[0]+(-0.954972993457)*x[1]**o+(0.00963939708746)*x[1]+(-0.884848381277)*x[2]**o+(-0.859924591622)*x[2]
        arg[0,1,0,1]=(0.133592941787)*x[0]**o+(0.12505415793)*x[0]+(-0.872197297559)*x[1]**o+(-0.666936976734)*x[1]+(-0.540166108036)*x[2]**o+(0.739145002101)*x[2]
        arg[0,1,1,0]=(0.543825266564)*x[0]**o+(0.167854064952)*x[0]+(-0.876426395953)*x[1]**o+(0.317538118917)*x[1]+(0.0454531201303)*x[2]**o+(0.822731343286)*x[2]
        arg[0,1,1,1]=(0.294557050513)*x[0]**o+(0.811698886899)*x[0]+(0.906213580147)*x[1]**o+(0.47895749797)*x[1]+(-0.937617554883)*x[2]**o+(0.171907458148)*x[2]
        arg[0,1,2,0]=(-0.843310949654)*x[0]**o+(0.766452648675)*x[0]+(-0.801151193917)*x[1]**o+(-0.373068887299)*x[1]+(-0.908043738284)*x[2]**o+(0.742278324563)*x[2]
        arg[0,1,2,1]=(-0.83259110525)*x[0]**o+(-0.303796015231)*x[0]+(0.153727560059)*x[1]**o+(-0.848393854019)*x[1]+(0.887795931916)*x[2]**o+(0.713486413074)*x[2]
        arg[0,2,0,0]=(-0.928603918182)*x[0]**o+(0.631213545526)*x[0]+(0.0149216054247)*x[1]**o+(0.136454313226)*x[1]+(-0.182080277504)*x[2]**o+(-0.898437757742)*x[2]
        arg[0,2,0,1]=(0.553844129645)*x[0]**o+(0.91439079798)*x[0]+(0.260424446678)*x[1]**o+(0.786629777779)*x[1]+(-0.412353876109)*x[2]**o+(0.496953000487)*x[2]
        arg[0,2,1,0]=(-0.48365422571)*x[0]**o+(-0.772252454664)*x[0]+(0.692592845143)*x[1]**o+(0.861992232175)*x[1]+(0.65169271692)*x[2]**o+(0.889175716524)*x[2]
        arg[0,2,1,1]=(-0.761086015475)*x[0]**o+(-0.616528617148)*x[0]+(-0.949399556926)*x[1]**o+(-0.783355074724)*x[1]+(-0.260470660608)*x[2]**o+(-0.854537542803)*x[2]
        arg[0,2,2,0]=(-0.812649439778)*x[0]**o+(-0.844862957855)*x[0]+(0.719928253134)*x[1]**o+(-0.892803903061)*x[1]+(-0.434753054503)*x[2]**o+(0.814188548451)*x[2]
        arg[0,2,2,1]=(-0.831106881173)*x[0]**o+(0.893305297647)*x[0]+(-0.351357795644)*x[1]**o+(0.0318574503478)*x[1]+(0.721896657873)*x[2]**o+(-0.75215838035)*x[2]
        arg[0,3,0,0]=(-0.868084492214)*x[0]**o+(-0.108398365838)*x[0]+(0.994384983885)*x[1]**o+(0.199445529335)*x[1]+(-0.570120483013)*x[2]**o+(0.992752456282)*x[2]
        arg[0,3,0,1]=(-0.304553428923)*x[0]**o+(-0.589461304318)*x[0]+(-0.570142565817)*x[1]**o+(-0.858530013475)*x[1]+(0.0690574539796)*x[2]**o+(-0.0321587525658)*x[2]
        arg[0,3,1,0]=(-0.729413026954)*x[0]**o+(-0.728248951404)*x[0]+(-0.568598599095)*x[1]**o+(0.0823012989572)*x[1]+(0.248614424573)*x[2]**o+(-0.116584878805)*x[2]
        arg[0,3,1,1]=(0.586846538084)*x[0]**o+(-0.406933529762)*x[0]+(0.839125682244)*x[1]**o+(0.237338453386)*x[1]+(-0.953101889056)*x[2]**o+(0.164147861616)*x[2]
        arg[0,3,2,0]=(-0.603895026116)*x[0]**o+(-0.401045478885)*x[0]+(-0.524739550832)*x[1]**o+(-0.077907095917)*x[1]+(0.535404409567)*x[2]**o+(0.745721528391)*x[2]
        arg[0,3,2,1]=(-0.189151084147)*x[0]**o+(-0.865927541126)*x[0]+(0.55265091657)*x[1]**o+(-0.592847311104)*x[1]+(-0.467509383222)*x[2]**o+(-0.779649276678)*x[2]
        arg[0,4,0,0]=(-0.564283606569)*x[0]**o+(-0.588029571592)*x[0]+(0.544948665159)*x[1]**o+(0.140885576219)*x[1]+(-0.333531842397)*x[2]**o+(-0.579095398292)*x[2]
        arg[0,4,0,1]=(-0.189127238455)*x[0]**o+(-0.693907127162)*x[0]+(0.40947793861)*x[1]**o+(-0.371098494308)*x[1]+(-0.189273002147)*x[2]**o+(-0.863169616331)*x[2]
        arg[0,4,1,0]=(-0.929838141145)*x[0]**o+(-0.0237583231579)*x[0]+(0.721069982341)*x[1]**o+(0.0378559913864)*x[1]+(-0.354729434476)*x[2]**o+(0.486256042969)*x[2]
        arg[0,4,1,1]=(0.349085173473)*x[0]**o+(0.398939355666)*x[0]+(-0.165772805366)*x[1]**o+(-0.532771334584)*x[1]+(-0.971868054413)*x[2]**o+(-0.38430185761)*x[2]
        arg[0,4,2,0]=(0.0330282399287)*x[0]**o+(0.157336131691)*x[0]+(-0.232739262685)*x[1]**o+(-0.280149712596)*x[1]+(0.374058430291)*x[2]**o+(0.594113015973)*x[2]
        arg[0,4,2,1]=(0.841740974112)*x[0]**o+(-0.454359680369)*x[0]+(-0.29970574726)*x[1]**o+(-0.234747232007)*x[1]+(0.590942390095)*x[2]**o+(-0.241807214234)*x[2]
        arg[1,0,0,0]=(0.815585848775)*x[0]**o+(-0.915020392674)*x[0]+(-0.0246530790781)*x[1]**o+(-0.783552074981)*x[1]+(0.539847778607)*x[2]**o+(-0.239345161181)*x[2]
        arg[1,0,0,1]=(-0.34212551354)*x[0]**o+(-0.231769108271)*x[0]+(-0.495709071836)*x[1]**o+(-0.301124823696)*x[1]+(-0.567450536966)*x[2]**o+(0.553640896859)*x[2]
        arg[1,0,1,0]=(0.507091065517)*x[0]**o+(-0.844317985101)*x[0]+(0.488126307787)*x[1]**o+(-0.624165922596)*x[1]+(0.00960189690918)*x[2]**o+(-0.35617062688)*x[2]
        arg[1,0,1,1]=(-0.219877421252)*x[0]**o+(-0.574843930095)*x[0]+(-0.846995760292)*x[1]**o+(-0.778496435041)*x[1]+(-0.817593097977)*x[2]**o+(0.387406829939)*x[2]
        arg[1,0,2,0]=(-0.774922277048)*x[0]**o+(-0.238152671014)*x[0]+(-0.913770877808)*x[1]**o+(-0.737305937434)*x[1]+(-0.13977056904)*x[2]**o+(-0.0156668131016)*x[2]
        arg[1,0,2,1]=(0.565993265911)*x[0]**o+(0.0690768496992)*x[0]+(-0.019845214223)*x[1]**o+(0.16253166701)*x[1]+(0.848117215885)*x[2]**o+(0.838075165243)*x[2]
        arg[1,1,0,0]=(0.871499694148)*x[0]**o+(-0.640851595157)*x[0]+(-0.170827284611)*x[1]**o+(-0.855550549988)*x[1]+(0.796086141249)*x[2]**o+(-0.214410672377)*x[2]
        arg[1,1,0,1]=(-0.903644604832)*x[0]**o+(0.0604780584681)*x[0]+(0.12800839523)*x[1]**o+(-0.594186752891)*x[1]+(0.415537334962)*x[2]**o+(0.122944611099)*x[2]
        arg[1,1,1,0]=(-0.611432775957)*x[0]**o+(0.391078622669)*x[0]+(-0.943091949273)*x[1]**o+(0.952739057317)*x[1]+(0.205104161722)*x[2]**o+(-0.946040001739)*x[2]
        arg[1,1,1,1]=(0.122941344579)*x[0]**o+(0.836981501711)*x[0]+(0.488970739079)*x[1]**o+(0.210062341907)*x[1]+(0.0264620393683)*x[2]**o+(-0.396386059797)*x[2]
        arg[1,1,2,0]=(-0.782658774038)*x[0]**o+(0.00611865020069)*x[0]+(0.806802875462)*x[1]**o+(-0.504323859583)*x[1]+(0.796276800331)*x[2]**o+(-0.792865116401)*x[2]
        arg[1,1,2,1]=(-0.249170081188)*x[0]**o+(-0.0234780571796)*x[0]+(-0.314017467872)*x[1]**o+(-0.0819148249592)*x[1]+(-0.71553368486)*x[2]**o+(0.763463207899)*x[2]
        arg[1,2,0,0]=(-0.722263632571)*x[0]**o+(-0.922870211147)*x[0]+(0.380794419995)*x[1]**o+(-0.76088532378)*x[1]+(0.417253071012)*x[2]**o+(0.152980049966)*x[2]
        arg[1,2,0,1]=(-0.667661997967)*x[0]**o+(-0.520570751147)*x[0]+(0.526368189149)*x[1]**o+(-0.0729376045288)*x[1]+(-0.434764684251)*x[2]**o+(0.124284939267)*x[2]
        arg[1,2,1,0]=(0.610868035553)*x[0]**o+(0.958644088455)*x[0]+(0.876029854699)*x[1]**o+(-0.278348787699)*x[1]+(-0.576695356283)*x[2]**o+(-0.302696556744)*x[2]
        arg[1,2,1,1]=(-0.224897853162)*x[0]**o+(0.89405397732)*x[0]+(-0.865281799513)*x[1]**o+(0.778866760053)*x[1]+(-0.939295291326)*x[2]**o+(0.256725768393)*x[2]
        arg[1,2,2,0]=(0.763785330202)*x[0]**o+(-0.92861265304)*x[0]+(-0.73488309202)*x[1]**o+(0.912711310857)*x[1]+(-0.359622626692)*x[2]**o+(-0.901261216611)*x[2]
        arg[1,2,2,1]=(-0.70687794303)*x[0]**o+(0.186497640523)*x[0]+(-0.587602932017)*x[1]**o+(0.439204174904)*x[1]+(0.0527573129765)*x[2]**o+(-0.121921819047)*x[2]
        arg[1,3,0,0]=(-0.576513727689)*x[0]**o+(0.228971758182)*x[0]+(0.282735129059)*x[1]**o+(-0.375270097915)*x[1]+(0.360087142563)*x[2]**o+(-0.0629946372113)*x[2]
        arg[1,3,0,1]=(-0.934375471561)*x[0]**o+(-0.710481522462)*x[0]+(-0.506438555546)*x[1]**o+(0.515278994703)*x[1]+(-0.789552111338)*x[2]**o+(0.556694887963)*x[2]
        arg[1,3,1,0]=(0.418081142778)*x[0]**o+(-0.990334755424)*x[0]+(0.465998147831)*x[1]**o+(0.85211792131)*x[1]+(0.64293315455)*x[2]**o+(0.418636667223)*x[2]
        arg[1,3,1,1]=(-0.192682640907)*x[0]**o+(-0.367176463246)*x[0]+(-0.512565915338)*x[1]**o+(0.431102518715)*x[1]+(-0.0213613616827)*x[2]**o+(-0.329868944656)*x[2]
        arg[1,3,2,0]=(-0.63277285658)*x[0]**o+(0.679819079734)*x[0]+(0.986520110022)*x[1]**o+(0.988046257154)*x[1]+(0.973419900492)*x[2]**o+(-0.177540534942)*x[2]
        arg[1,3,2,1]=(-0.660456888389)*x[0]**o+(0.546277948387)*x[0]+(-0.437065207396)*x[1]**o+(-0.310754727958)*x[1]+(-0.786950057282)*x[2]**o+(0.959526365177)*x[2]
        arg[1,4,0,0]=(0.764190976082)*x[0]**o+(0.970535029727)*x[0]+(0.389051006993)*x[1]**o+(-0.289346925704)*x[1]+(-0.386263606576)*x[2]**o+(0.624234063589)*x[2]
        arg[1,4,0,1]=(-0.982030683687)*x[0]**o+(-0.310500292581)*x[0]+(-0.559096285575)*x[1]**o+(-0.622541272039)*x[1]+(0.49924233081)*x[2]**o+(-0.219258681197)*x[2]
        arg[1,4,1,0]=(-0.124268932891)*x[0]**o+(0.102696501388)*x[0]+(-0.135123912862)*x[1]**o+(-0.960215132796)*x[1]+(-0.63171550556)*x[2]**o+(-0.493725285999)*x[2]
        arg[1,4,1,1]=(-0.397915436399)*x[0]**o+(-0.629822230732)*x[0]+(-0.65097233947)*x[1]**o+(0.915396047247)*x[1]+(0.835684562564)*x[2]**o+(-0.538096550078)*x[2]
        arg[1,4,2,0]=(-0.946429448216)*x[0]**o+(-0.310412085571)*x[0]+(-0.0984587257278)*x[1]**o+(-0.662090989911)*x[1]+(-0.424711481307)*x[2]**o+(0.00916879493127)*x[2]
        arg[1,4,2,1]=(0.919399549311)*x[0]**o+(-0.890110018315)*x[0]+(-0.685928475328)*x[1]**o+(-0.4731186395)*x[1]+(0.961887038718)*x[2]**o+(0.850957316131)*x[2]
        arg[2,0,0,0]=(0.834851801283)*x[0]**o+(-0.401572429371)*x[0]+(0.313309302528)*x[1]**o+(0.0368590314435)*x[1]+(-0.299807168778)*x[2]**o+(0.250599904818)*x[2]
        arg[2,0,0,1]=(-0.717562866174)*x[0]**o+(0.374797888506)*x[0]+(0.254329847579)*x[1]**o+(-0.987429692718)*x[1]+(-0.886758521324)*x[2]**o+(0.309486900444)*x[2]
        arg[2,0,1,0]=(0.139962143021)*x[0]**o+(-0.134735774213)*x[0]+(-0.231254041503)*x[1]**o+(0.0694785371931)*x[1]+(0.769929630132)*x[2]**o+(0.486646094292)*x[2]
        arg[2,0,1,1]=(0.349691701871)*x[0]**o+(-0.0308285150496)*x[0]+(0.00977143515921)*x[1]**o+(0.0970015800122)*x[1]+(-0.725991176667)*x[2]**o+(0.433663449433)*x[2]
        arg[2,0,2,0]=(0.564474039592)*x[0]**o+(0.950620065312)*x[0]+(0.256658075134)*x[1]**o+(0.774458453189)*x[1]+(0.00181734756534)*x[2]**o+(-0.270638852561)*x[2]
        arg[2,0,2,1]=(-0.175813811707)*x[0]**o+(0.821341326493)*x[0]+(-0.464966990893)*x[1]**o+(-0.310747066257)*x[1]+(-0.413557861251)*x[2]**o+(-0.799160067701)*x[2]
        arg[2,1,0,0]=(-0.449817381621)*x[0]**o+(-0.0573780378383)*x[0]+(-0.22675349133)*x[1]**o+(-0.150314340624)*x[1]+(-0.116693721705)*x[2]**o+(-0.20577625675)*x[2]
        arg[2,1,0,1]=(0.955318430674)*x[0]**o+(-0.148100168239)*x[0]+(0.741705440574)*x[1]**o+(-0.0768022054634)*x[1]+(-0.721807872287)*x[2]**o+(-0.898235041116)*x[2]
        arg[2,1,1,0]=(-0.183120810472)*x[0]**o+(0.307305712226)*x[0]+(0.699835684443)*x[1]**o+(0.964082351856)*x[1]+(-0.0348051023458)*x[2]**o+(0.544435382916)*x[2]
        arg[2,1,1,1]=(0.786862912642)*x[0]**o+(0.37388939814)*x[0]+(0.608937302857)*x[1]**o+(0.122305286217)*x[1]+(-0.892536849678)*x[2]**o+(0.119257478145)*x[2]
        arg[2,1,2,0]=(0.758154908623)*x[0]**o+(0.695188856067)*x[0]+(0.458313343957)*x[1]**o+(0.979157471455)*x[1]+(0.194237063725)*x[2]**o+(0.773001101454)*x[2]
        arg[2,1,2,1]=(0.0754678814151)*x[0]**o+(-0.639556181007)*x[0]+(0.934627102078)*x[1]**o+(0.0832082572611)*x[1]+(0.668822756858)*x[2]**o+(0.627532210297)*x[2]
        arg[2,2,0,0]=(-0.192921889174)*x[0]**o+(0.989703287553)*x[0]+(0.00517568784693)*x[1]**o+(-0.360076937731)*x[1]+(0.947056110784)*x[2]**o+(0.469841550381)*x[2]
        arg[2,2,0,1]=(0.201501669595)*x[0]**o+(-0.959957961834)*x[0]+(0.0229159090152)*x[1]**o+(0.396188263084)*x[1]+(0.186211667989)*x[2]**o+(0.304503474778)*x[2]
        arg[2,2,1,0]=(0.333974995608)*x[0]**o+(0.210202867626)*x[0]+(-0.410565897643)*x[1]**o+(0.52659985644)*x[1]+(0.983960564934)*x[2]**o+(-0.439323691866)*x[2]
        arg[2,2,1,1]=(0.706119074329)*x[0]**o+(-0.00220875175563)*x[0]+(0.143693169123)*x[1]**o+(0.217254593235)*x[1]+(-0.412057975252)*x[2]**o+(-0.967360361422)*x[2]
        arg[2,2,2,0]=(0.217307356597)*x[0]**o+(0.861728452163)*x[0]+(-0.193239462241)*x[1]**o+(0.872273176526)*x[1]+(-0.436414721507)*x[2]**o+(-0.62398425635)*x[2]
        arg[2,2,2,1]=(0.467099668794)*x[0]**o+(0.109742400918)*x[0]+(-0.987522191801)*x[1]**o+(-0.250251964275)*x[1]+(0.442152166523)*x[2]**o+(-0.709157679576)*x[2]
        arg[2,3,0,0]=(0.769666242031)*x[0]**o+(0.2352745864)*x[0]+(-0.302893533611)*x[1]**o+(0.0806677655098)*x[1]+(0.886434893279)*x[2]**o+(0.366016417047)*x[2]
        arg[2,3,0,1]=(-0.612204706187)*x[0]**o+(-0.242961993133)*x[0]+(0.946225816441)*x[1]**o+(0.528642650581)*x[1]+(-0.0166837450337)*x[2]**o+(-0.0889397945206)*x[2]
        arg[2,3,1,0]=(0.218589745288)*x[0]**o+(-0.296125811204)*x[0]+(0.358743045538)*x[1]**o+(-0.945778818656)*x[1]+(-0.33155158646)*x[2]**o+(-0.293038455557)*x[2]
        arg[2,3,1,1]=(0.753737723433)*x[0]**o+(-0.772926141856)*x[0]+(-0.0107183417902)*x[1]**o+(0.125538612735)*x[1]+(0.225277420156)*x[2]**o+(-0.54876651451)*x[2]
        arg[2,3,2,0]=(0.947419453319)*x[0]**o+(-0.101935160067)*x[0]+(0.481841505589)*x[1]**o+(-0.873842146032)*x[1]+(-0.377764402402)*x[2]**o+(-0.361560947647)*x[2]
        arg[2,3,2,1]=(0.745703609187)*x[0]**o+(-0.488401078192)*x[0]+(0.190261500318)*x[1]**o+(-0.22176180212)*x[1]+(0.80182740764)*x[2]**o+(-0.504953098351)*x[2]
        arg[2,4,0,0]=(0.476355625206)*x[0]**o+(0.452225231178)*x[0]+(0.669315429911)*x[1]**o+(0.159511010158)*x[1]+(0.652056592943)*x[2]**o+(0.44515033609)*x[2]
        arg[2,4,0,1]=(0.512194809916)*x[0]**o+(0.654452558318)*x[0]+(-0.526832882528)*x[1]**o+(0.590859897981)*x[1]+(-0.155989159043)*x[2]**o+(-0.689178072559)*x[2]
        arg[2,4,1,0]=(0.0450244624553)*x[0]**o+(0.0790299000335)*x[0]+(-0.690643908994)*x[1]**o+(-0.262426911912)*x[1]+(-0.426935471303)*x[2]**o+(-0.434570506153)*x[2]
        arg[2,4,1,1]=(-0.04445890079)*x[0]**o+(-0.541574770252)*x[0]+(-0.879290270122)*x[1]**o+(0.545302743626)*x[1]+(-0.836472565261)*x[2]**o+(0.66750272736)*x[2]
        arg[2,4,2,0]=(-0.477788509677)*x[0]**o+(0.706195950613)*x[0]+(0.628430029293)*x[1]**o+(-0.297608932813)*x[1]+(0.38364501438)*x[2]**o+(0.855444938831)*x[2]
        arg[2,4,2,1]=(0.123183000451)*x[0]**o+(0.41176132265)*x[0]+(-0.379498883356)*x[1]**o+(-0.53442961658)*x[1]+(-0.101215056317)*x[2]**o+(0.247971839461)*x[2]
        arg[3,0,0,0]=(-0.239170701904)*x[0]**o+(0.369420936359)*x[0]+(0.972913513036)*x[1]**o+(0.775248415403)*x[1]+(-0.850310640223)*x[2]**o+(-0.815345534491)*x[2]
        arg[3,0,0,1]=(0.591124498004)*x[0]**o+(0.529522171858)*x[0]+(-0.974005221641)*x[1]**o+(0.430102201571)*x[1]+(0.570519406976)*x[2]**o+(-0.0915083364671)*x[2]
        arg[3,0,1,0]=(-0.27678065557)*x[0]**o+(-0.53225308602)*x[0]+(0.902356190541)*x[1]**o+(-0.564808732046)*x[1]+(0.0981283775668)*x[2]**o+(-0.64957631107)*x[2]
        arg[3,0,1,1]=(0.230523873408)*x[0]**o+(-0.382745197262)*x[0]+(0.41507101812)*x[1]**o+(0.48492708335)*x[1]+(-0.11672353881)*x[2]**o+(-0.56902283397)*x[2]
        arg[3,0,2,0]=(-0.400023140104)*x[0]**o+(-0.406234012546)*x[0]+(-0.17810623045)*x[1]**o+(-0.833748300281)*x[1]+(-0.463689953863)*x[2]**o+(-0.11095482465)*x[2]
        arg[3,0,2,1]=(0.448885918108)*x[0]**o+(0.0447046792888)*x[0]+(-0.781652457281)*x[1]**o+(-0.893726375109)*x[1]+(0.918099233406)*x[2]**o+(0.560717614604)*x[2]
        arg[3,1,0,0]=(0.046700262853)*x[0]**o+(0.729867044213)*x[0]+(-0.870868205823)*x[1]**o+(-0.711267787629)*x[1]+(0.760550112312)*x[2]**o+(-0.0188334901698)*x[2]
        arg[3,1,0,1]=(0.304305551549)*x[0]**o+(-0.73527291962)*x[0]+(0.722134673974)*x[1]**o+(-0.40822784698)*x[1]+(0.912464122456)*x[2]**o+(-0.857656020349)*x[2]
        arg[3,1,1,0]=(-0.857864508282)*x[0]**o+(-0.22160002809)*x[0]+(-0.0892530574233)*x[1]**o+(-0.547259900671)*x[1]+(0.641858628113)*x[2]**o+(-0.203700292205)*x[2]
        arg[3,1,1,1]=(-0.47064403668)*x[0]**o+(-0.105885697358)*x[0]+(-0.899570240486)*x[1]**o+(0.660942689756)*x[1]+(-0.365836826289)*x[2]**o+(0.9593955249)*x[2]
        arg[3,1,2,0]=(-0.51407433496)*x[0]**o+(0.369092182634)*x[0]+(-0.740281092517)*x[1]**o+(0.383443939366)*x[1]+(-0.140318759419)*x[2]**o+(-0.481274651104)*x[2]
        arg[3,1,2,1]=(-0.938924125636)*x[0]**o+(0.41994781084)*x[0]+(-0.946320511542)*x[1]**o+(-0.86702606561)*x[1]+(0.63547569926)*x[2]**o+(0.923988068696)*x[2]
        arg[3,2,0,0]=(0.212833889165)*x[0]**o+(-0.0823077975238)*x[0]+(0.038483237489)*x[1]**o+(0.393784376623)*x[1]+(0.573190557103)*x[2]**o+(0.403015743632)*x[2]
        arg[3,2,0,1]=(-0.169294531943)*x[0]**o+(-0.769228919251)*x[0]+(-0.907462814057)*x[1]**o+(-0.188574782291)*x[1]+(-0.859793027531)*x[2]**o+(-0.673334626964)*x[2]
        arg[3,2,1,0]=(0.646744388517)*x[0]**o+(0.346749855021)*x[0]+(-0.314382791156)*x[1]**o+(-0.0396545336605)*x[1]+(0.844678566939)*x[2]**o+(0.724721336727)*x[2]
        arg[3,2,1,1]=(-0.471761417345)*x[0]**o+(-0.801631597222)*x[0]+(-0.376444012893)*x[1]**o+(0.920939382926)*x[1]+(-0.728019331398)*x[2]**o+(0.905808341144)*x[2]
        arg[3,2,2,0]=(-0.713759272803)*x[0]**o+(-0.930638013429)*x[0]+(0.864451879958)*x[1]**o+(0.0566794358824)*x[1]+(-0.292092226905)*x[2]**o+(-0.195695530268)*x[2]
        arg[3,2,2,1]=(0.843437618194)*x[0]**o+(0.214911962718)*x[0]+(0.475690162826)*x[1]**o+(0.0744570063618)*x[1]+(0.351711310616)*x[2]**o+(-0.187223450548)*x[2]
        arg[3,3,0,0]=(0.108651500878)*x[0]**o+(-0.402624590847)*x[0]+(0.120700159277)*x[1]**o+(-0.363139807635)*x[1]+(-0.62413441951)*x[2]**o+(-0.474236131596)*x[2]
        arg[3,3,0,1]=(-0.0836636547536)*x[0]**o+(-0.187959781777)*x[0]+(0.756983584308)*x[1]**o+(0.756573839599)*x[1]+(0.195640878564)*x[2]**o+(-0.566596474032)*x[2]
        arg[3,3,1,0]=(-0.171427844449)*x[0]**o+(-0.266807792484)*x[0]+(0.380046037965)*x[1]**o+(-0.129297371793)*x[1]+(0.79139100315)*x[2]**o+(0.223804751067)*x[2]
        arg[3,3,1,1]=(-0.054757303264)*x[0]**o+(-0.823878124196)*x[0]+(0.80695256107)*x[1]**o+(0.583250192211)*x[1]+(0.894937939133)*x[2]**o+(0.878757173892)*x[2]
        arg[3,3,2,0]=(0.13923802434)*x[0]**o+(-0.600732296834)*x[0]+(0.111226716676)*x[1]**o+(-0.474524749496)*x[1]+(0.674290476359)*x[2]**o+(-0.06714767759)*x[2]
        arg[3,3,2,1]=(0.246187769682)*x[0]**o+(-0.871969439162)*x[0]+(0.56748693564)*x[1]**o+(-0.0831774328079)*x[1]+(-0.27380467532)*x[2]**o+(-0.623711953244)*x[2]
        arg[3,4,0,0]=(-0.345947366312)*x[0]**o+(-0.304842609526)*x[0]+(-0.720722450633)*x[1]**o+(0.889712922756)*x[1]+(-0.0451362945237)*x[2]**o+(-0.52344522887)*x[2]
        arg[3,4,0,1]=(-0.830119647904)*x[0]**o+(0.577888363835)*x[0]+(0.180691945154)*x[1]**o+(-0.489554894073)*x[1]+(0.848489044545)*x[2]**o+(-0.243139985297)*x[2]
        arg[3,4,1,0]=(0.846597693999)*x[0]**o+(0.757477455396)*x[0]+(-0.827604908657)*x[1]**o+(-0.108392312717)*x[1]+(0.597454707613)*x[2]**o+(0.35846243312)*x[2]
        arg[3,4,1,1]=(0.107973753109)*x[0]**o+(0.578649460223)*x[0]+(0.609012033524)*x[1]**o+(-0.8428861922)*x[1]+(0.42144759968)*x[2]**o+(-0.938952208084)*x[2]
        arg[3,4,2,0]=(0.286586330618)*x[0]**o+(0.7324646289)*x[0]+(-0.448461377611)*x[1]**o+(0.62562408085)*x[1]+(0.352429283116)*x[2]**o+(0.112047034296)*x[2]
        arg[3,4,2,1]=(-0.431663387082)*x[0]**o+(-0.612934322552)*x[0]+(-0.426644797074)*x[1]**o+(0.768645027094)*x[1]+(-0.453242885757)*x[2]**o+(0.893748345809)*x[2]
        ref[0,0,0,0]=(-1.2147562394)*(1+2.*(dim-1.)/(o+1.))+(1.47471604216)*dim
        ref[0,0,0,1]=(-0.283261030731)*(1+2.*(dim-1.)/(o+1.))+(0.49706944223)*dim
        ref[0,0,1,0]=(-0.963543225803)*(1+2.*(dim-1.)/(o+1.))+(0.302115929232)*dim
        ref[0,0,1,1]=(1.47597923019)*(1+2.*(dim-1.)/(o+1.))+(0.302975596116)*dim
        ref[0,0,2,0]=(2.07218071265)*(1+2.*(dim-1.)/(o+1.))+(1.40474756312)*dim
        ref[0,0,2,1]=(1.18935174934)*(1+2.*(dim-1.)/(o+1.))+(-0.898949809002)*dim
        ref[0,1,0,0]=(-2.71454215957)*(1+2.*(dim-1.)/(o+1.))+(-1.81174147681)*dim
        ref[0,1,0,1]=(-1.27877046381)*(1+2.*(dim-1.)/(o+1.))+(0.197262183297)*dim
        ref[0,1,1,0]=(-0.287148009258)*(1+2.*(dim-1.)/(o+1.))+(1.30812352716)*dim
        ref[0,1,1,1]=(0.263153075777)*(1+2.*(dim-1.)/(o+1.))+(1.46256384302)*dim
        ref[0,1,2,0]=(-2.55250588186)*(1+2.*(dim-1.)/(o+1.))+(1.13566208594)*dim
        ref[0,1,2,1]=(0.208932386724)*(1+2.*(dim-1.)/(o+1.))+(-0.438703456176)*dim
        ref[0,2,0,0]=(-1.09576259026)*(1+2.*(dim-1.)/(o+1.))+(-0.13076989899)*dim
        ref[0,2,0,1]=(0.401914700214)*(1+2.*(dim-1.)/(o+1.))+(2.19797357625)*dim
        ref[0,2,1,0]=(0.860631336354)*(1+2.*(dim-1.)/(o+1.))+(0.978915494035)*dim
        ref[0,2,1,1]=(-1.97095623301)*(1+2.*(dim-1.)/(o+1.))+(-2.25442123467)*dim
        ref[0,2,2,0]=(-0.527474241147)*(1+2.*(dim-1.)/(o+1.))+(-0.923478312465)*dim
        ref[0,2,2,1]=(-0.460568018944)*(1+2.*(dim-1.)/(o+1.))+(0.173004367644)*dim
        ref[0,3,0,0]=(-0.443819991343)*(1+2.*(dim-1.)/(o+1.))+(1.08379961978)*dim
        ref[0,3,0,1]=(-0.80563854076)*(1+2.*(dim-1.)/(o+1.))+(-1.48015007036)*dim
        ref[0,3,1,0]=(-1.04939720148)*(1+2.*(dim-1.)/(o+1.))+(-0.762532531252)*dim
        ref[0,3,1,1]=(0.472870331273)*(1+2.*(dim-1.)/(o+1.))+(-0.00544721475928)*dim
        ref[0,3,2,0]=(-0.593230167381)*(1+2.*(dim-1.)/(o+1.))+(0.26676895359)*dim
        ref[0,3,2,1]=(-0.104009550799)*(1+2.*(dim-1.)/(o+1.))+(-2.23842412891)*dim
        ref[0,4,0,0]=(-0.352866783806)*(1+2.*(dim-1.)/(o+1.))+(-1.02623939367)*dim
        ref[0,4,0,1]=(0.0310776980081)*(1+2.*(dim-1.)/(o+1.))+(-1.9281752378)*dim
        ref[0,4,1,0]=(-0.56349759328)*(1+2.*(dim-1.)/(o+1.))+(0.500353711197)*dim
        ref[0,4,1,1]=(-0.788555686306)*(1+2.*(dim-1.)/(o+1.))+(-0.518133836527)*dim
        ref[0,4,2,0]=(0.174347407535)*(1+2.*(dim-1.)/(o+1.))+(0.471299435068)*dim
        ref[0,4,2,1]=(1.13297761695)*(1+2.*(dim-1.)/(o+1.))+(-0.93091412661)*dim
        ref[1,0,0,0]=(1.3307805483)*(1+2.*(dim-1.)/(o+1.))+(-1.93791762884)*dim
        ref[1,0,0,1]=(-1.40528512234)*(1+2.*(dim-1.)/(o+1.))+(0.0207469648915)*dim
        ref[1,0,1,0]=(1.00481927021)*(1+2.*(dim-1.)/(o+1.))+(-1.82465453458)*dim
        ref[1,0,1,1]=(-1.88446627952)*(1+2.*(dim-1.)/(o+1.))+(-0.965933535196)*dim
        ref[1,0,2,0]=(-1.8284637239)*(1+2.*(dim-1.)/(o+1.))+(-0.991125421549)*dim
        ref[1,0,2,1]=(1.39426526757)*(1+2.*(dim-1.)/(o+1.))+(1.06968368195)*dim
        ref[1,1,0,0]=(1.49675855079)*(1+2.*(dim-1.)/(o+1.))+(-1.71081281752)*dim
        ref[1,1,0,1]=(-0.36009887464)*(1+2.*(dim-1.)/(o+1.))+(-0.410764083324)*dim
        ref[1,1,1,0]=(-1.34942056351)*(1+2.*(dim-1.)/(o+1.))+(0.397777678247)*dim
        ref[1,1,1,1]=(0.638374123027)*(1+2.*(dim-1.)/(o+1.))+(0.650657783821)*dim
        ref[1,1,2,0]=(0.820420901755)*(1+2.*(dim-1.)/(o+1.))+(-1.29107032578)*dim
        ref[1,1,2,1]=(-1.27872123392)*(1+2.*(dim-1.)/(o+1.))+(0.65807032576)*dim
        ref[1,2,0,0]=(0.075783858436)*(1+2.*(dim-1.)/(o+1.))+(-1.53077548496)*dim
        ref[1,2,0,1]=(-0.576058493068)*(1+2.*(dim-1.)/(o+1.))+(-0.469223416409)*dim
        ref[1,2,1,0]=(0.91020253397)*(1+2.*(dim-1.)/(o+1.))+(0.377598744012)*dim
        ref[1,2,1,1]=(-2.029474944)*(1+2.*(dim-1.)/(o+1.))+(1.92964650577)*dim
        ref[1,2,2,0]=(-0.330720388509)*(1+2.*(dim-1.)/(o+1.))+(-0.917162558794)*dim
        ref[1,2,2,1]=(-1.24172356207)*(1+2.*(dim-1.)/(o+1.))+(0.503779996381)*dim
        ref[1,3,0,0]=(0.0663085439332)*(1+2.*(dim-1.)/(o+1.))+(-0.209292976945)*dim
        ref[1,3,0,1]=(-2.23036613844)*(1+2.*(dim-1.)/(o+1.))+(0.361492360204)*dim
        ref[1,3,1,0]=(1.52701244516)*(1+2.*(dim-1.)/(o+1.))+(0.280419833109)*dim
        ref[1,3,1,1]=(-0.726609917928)*(1+2.*(dim-1.)/(o+1.))+(-0.265942889187)*dim
        ref[1,3,2,0]=(1.32716715393)*(1+2.*(dim-1.)/(o+1.))+(1.49032480195)*dim
        ref[1,3,2,1]=(-1.88447215307)*(1+2.*(dim-1.)/(o+1.))+(1.19504958561)*dim
        ref[1,4,0,0]=(0.766978376499)*(1+2.*(dim-1.)/(o+1.))+(1.30542216761)*dim
        ref[1,4,0,1]=(-1.04188463845)*(1+2.*(dim-1.)/(o+1.))+(-1.15230024582)*dim
        ref[1,4,1,0]=(-0.891108351313)*(1+2.*(dim-1.)/(o+1.))+(-1.35124391741)*dim
        ref[1,4,1,1]=(-0.213203213304)*(1+2.*(dim-1.)/(o+1.))+(-0.252522733564)*dim
        ref[1,4,2,0]=(-1.46959965525)*(1+2.*(dim-1.)/(o+1.))+(-0.963334280551)*dim
        ref[1,4,2,1]=(1.1953581127)*(1+2.*(dim-1.)/(o+1.))+(-0.512271341684)*dim
        ref[2,0,0,0]=(0.848353935033)*(1+2.*(dim-1.)/(o+1.))+(-0.11411349311)*dim
        ref[2,0,0,1]=(-1.34999153992)*(1+2.*(dim-1.)/(o+1.))+(-0.303144903768)*dim
        ref[2,0,1,0]=(0.67863773165)*(1+2.*(dim-1.)/(o+1.))+(0.421388857272)*dim
        ref[2,0,1,1]=(-0.366528039637)*(1+2.*(dim-1.)/(o+1.))+(0.499836514396)*dim
        ref[2,0,2,0]=(0.822949462292)*(1+2.*(dim-1.)/(o+1.))+(1.45443966594)*dim
        ref[2,0,2,1]=(-1.05433866385)*(1+2.*(dim-1.)/(o+1.))+(-0.288565807465)*dim
        ref[2,1,0,0]=(-0.793264594657)*(1+2.*(dim-1.)/(o+1.))+(-0.413468635212)*dim
        ref[2,1,0,1]=(0.975215998961)*(1+2.*(dim-1.)/(o+1.))+(-1.12313741482)*dim
        ref[2,1,1,0]=(0.481909771625)*(1+2.*(dim-1.)/(o+1.))+(1.815823447)*dim
        ref[2,1,1,1]=(0.503263365821)*(1+2.*(dim-1.)/(o+1.))+(0.615452162503)*dim
        ref[2,1,2,0]=(1.41070531631)*(1+2.*(dim-1.)/(o+1.))+(2.44734742898)*dim
        ref[2,1,2,1]=(1.67891774035)*(1+2.*(dim-1.)/(o+1.))+(0.0711842865505)*dim
        ref[2,2,0,0]=(0.759309909457)*(1+2.*(dim-1.)/(o+1.))+(1.0994679002)*dim
        ref[2,2,0,1]=(0.410629246599)*(1+2.*(dim-1.)/(o+1.))+(-0.259266223972)*dim
        ref[2,2,1,0]=(0.907369662899)*(1+2.*(dim-1.)/(o+1.))+(0.297479032201)*dim
        ref[2,2,1,1]=(0.437754268199)*(1+2.*(dim-1.)/(o+1.))+(-0.752314519943)*dim
        ref[2,2,2,0]=(-0.412346827151)*(1+2.*(dim-1.)/(o+1.))+(1.11001737234)*dim
        ref[2,2,2,1]=(-0.0782703564833)*(1+2.*(dim-1.)/(o+1.))+(-0.849667242932)*dim
        ref[2,3,0,0]=(1.3532076017)*(1+2.*(dim-1.)/(o+1.))+(0.681958768956)*dim
        ref[2,3,0,1]=(0.317337365221)*(1+2.*(dim-1.)/(o+1.))+(0.196740862928)*dim
        ref[2,3,1,0]=(0.245781204366)*(1+2.*(dim-1.)/(o+1.))+(-1.53494308542)*dim
        ref[2,3,1,1]=(0.968296801799)*(1+2.*(dim-1.)/(o+1.))+(-1.19615404363)*dim
        ref[2,3,2,0]=(1.05149655651)*(1+2.*(dim-1.)/(o+1.))+(-1.33733825375)*dim
        ref[2,3,2,1]=(1.73779251715)*(1+2.*(dim-1.)/(o+1.))+(-1.21511597866)*dim
        ref[2,4,0,0]=(1.79772764806)*(1+2.*(dim-1.)/(o+1.))+(1.05688657743)*dim
        ref[2,4,0,1]=(-0.170627231654)*(1+2.*(dim-1.)/(o+1.))+(0.55613438374)*dim
        ref[2,4,1,0]=(-1.07255491784)*(1+2.*(dim-1.)/(o+1.))+(-0.617967518031)*dim
        ref[2,4,1,1]=(-1.76022173617)*(1+2.*(dim-1.)/(o+1.))+(0.671230700733)*dim
        ref[2,4,2,0]=(0.534286533995)*(1+2.*(dim-1.)/(o+1.))+(1.26403195663)*dim
        ref[2,4,2,1]=(-0.357530939222)*(1+2.*(dim-1.)/(o+1.))+(0.12530354553)*dim
        ref[3,0,0,0]=(-0.11656782909)*(1+2.*(dim-1.)/(o+1.))+(0.32932381727)*dim
        ref[3,0,0,1]=(0.187638683339)*(1+2.*(dim-1.)/(o+1.))+(0.868116036962)*dim
        ref[3,0,1,0]=(0.723703912538)*(1+2.*(dim-1.)/(o+1.))+(-1.74663812914)*dim
        ref[3,0,1,1]=(0.528871352717)*(1+2.*(dim-1.)/(o+1.))+(-0.466840947882)*dim
        ref[3,0,2,0]=(-1.04181932442)*(1+2.*(dim-1.)/(o+1.))+(-1.35093713748)*dim
        ref[3,0,2,1]=(0.585332694234)*(1+2.*(dim-1.)/(o+1.))+(-0.288304081216)*dim
        ref[3,1,0,0]=(-0.0636178306576)*(1+2.*(dim-1.)/(o+1.))+(-0.000234233586678)*dim
        ref[3,1,0,1]=(1.93890434798)*(1+2.*(dim-1.)/(o+1.))+(-2.00115678695)*dim
        ref[3,1,1,0]=(-0.305258937592)*(1+2.*(dim-1.)/(o+1.))+(-0.972560220966)*dim
        ref[3,1,1,1]=(-1.73605110346)*(1+2.*(dim-1.)/(o+1.))+(1.5144525173)*dim
        ref[3,1,2,0]=(-1.3946741869)*(1+2.*(dim-1.)/(o+1.))+(0.271261470896)*dim
        ref[3,1,2,1]=(-1.24976893792)*(1+2.*(dim-1.)/(o+1.))+(0.476909813927)*dim
        ref[3,2,0,0]=(0.824507683756)*(1+2.*(dim-1.)/(o+1.))+(0.71449232273)*dim
        ref[3,2,0,1]=(-1.93655037353)*(1+2.*(dim-1.)/(o+1.))+(-1.63113832851)*dim
        ref[3,2,1,0]=(1.1770401643)*(1+2.*(dim-1.)/(o+1.))+(1.03181665809)*dim
        ref[3,2,1,1]=(-1.57622476164)*(1+2.*(dim-1.)/(o+1.))+(1.02511612685)*dim
        ref[3,2,2,0]=(-0.141399619749)*(1+2.*(dim-1.)/(o+1.))+(-1.06965410781)*dim
        ref[3,2,2,1]=(1.67083909164)*(1+2.*(dim-1.)/(o+1.))+(0.102145518531)*dim
        ref[3,3,0,0]=(-0.394782759355)*(1+2.*(dim-1.)/(o+1.))+(-1.24000053008)*dim
        ref[3,3,0,1]=(0.868960808118)*(1+2.*(dim-1.)/(o+1.))+(0.00201758379073)*dim
        ref[3,3,1,0]=(1.00000919667)*(1+2.*(dim-1.)/(o+1.))+(-0.17230041321)*dim
        ref[3,3,1,1]=(1.64713319694)*(1+2.*(dim-1.)/(o+1.))+(0.638129241907)*dim
        ref[3,3,2,0]=(0.924755217374)*(1+2.*(dim-1.)/(o+1.))+(-1.14240472392)*dim
        ref[3,3,2,1]=(0.539870030002)*(1+2.*(dim-1.)/(o+1.))+(-1.57885882521)*dim
        ref[3,4,0,0]=(-1.11180611147)*(1+2.*(dim-1.)/(o+1.))+(0.0614250843596)*dim
        ref[3,4,0,1]=(0.199061341795)*(1+2.*(dim-1.)/(o+1.))+(-0.154806515534)*dim
        ref[3,4,1,0]=(0.616447492955)*(1+2.*(dim-1.)/(o+1.))+(1.0075475758)*dim
        ref[3,4,1,1]=(1.13843338631)*(1+2.*(dim-1.)/(o+1.))+(-1.20318894006)*dim
        ref[3,4,2,0]=(0.190554236123)*(1+2.*(dim-1.)/(o+1.))+(1.47013574405)*dim
        ref[3,4,2,1]=(-1.31155106991)*(1+2.*(dim-1.)/(o+1.))+(1.04945905035)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.416878137535)*x[0]**o+(0.0809721296208)*x[0]+(-0.275084069478)*x[1]**o+(-0.704226656249)*x[1]
        ref=(0.141794068057)*(1+2.*(dim-1.)/(o+1.))+(-0.623254526628)*dim
      else:
        arg=(0.587142789605)*x[0]**o+(0.894018391929)*x[0]+(-0.417709558814)*x[1]**o+(0.0970658484968)*x[1]+(-0.840982437679)*x[2]**o+(-0.400242562091)*x[2]
        ref=(-0.671549206889)*(1+2.*(dim-1.)/(o+1.))+(0.590841678335)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.051335729494)*x[0]**o+(0.257057962305)*x[0]+(0.00794214446212)*x[1]**o+(0.52307365061)*x[1]
        arg[1]=(0.419915493803)*x[0]**o+(-0.0105826268931)*x[0]+(0.369274332127)*x[1]**o+(0.340796701053)*x[1]
        ref[0]=(0.0592778739561)*(1+2.*(dim-1.)/(o+1.))+(0.780131612915)*dim
        ref[1]=(0.789189825931)*(1+2.*(dim-1.)/(o+1.))+(0.33021407416)*dim
      else:
        arg[0]=(-0.483302491019)*x[0]**o+(0.223892398776)*x[0]+(-0.555612541317)*x[1]**o+(0.75535374219)*x[1]+(-0.208573228788)*x[2]**o+(0.479879212564)*x[2]
        arg[1]=(-0.012341185885)*x[0]**o+(-0.959474230066)*x[0]+(0.397671123434)*x[1]**o+(-0.120340696903)*x[1]+(-0.645757653205)*x[2]**o+(-0.216057959957)*x[2]
        ref[0]=(-1.24748826112)*(1+2.*(dim-1.)/(o+1.))+(1.45912535353)*dim
        ref[1]=(-0.260427715657)*(1+2.*(dim-1.)/(o+1.))+(-1.29587288693)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.493237934301)*x[0]**o+(-0.977176505302)*x[0]+(-0.837740403623)*x[1]**o+(0.204709065496)*x[1]
        arg[0,1]=(0.794064508577)*x[0]**o+(-0.917060945623)*x[0]+(0.0114185601474)*x[1]**o+(0.547108694374)*x[1]
        arg[0,2]=(-0.989778984923)*x[0]**o+(-0.499651095311)*x[0]+(-0.126817505183)*x[1]**o+(0.889077385524)*x[1]
        arg[0,3]=(-0.223453594493)*x[0]**o+(-0.17983229649)*x[0]+(0.352602269148)*x[1]**o+(0.00617724258273)*x[1]
        arg[0,4]=(0.401784356952)*x[0]**o+(0.839308128877)*x[0]+(0.322943402455)*x[1]**o+(0.770206521944)*x[1]
        arg[1,0]=(0.491015078355)*x[0]**o+(0.194909202321)*x[0]+(0.38986016491)*x[1]**o+(0.671106836372)*x[1]
        arg[1,1]=(0.200697856784)*x[0]**o+(0.389746053244)*x[0]+(0.850237988574)*x[1]**o+(-0.677750684271)*x[1]
        arg[1,2]=(0.300444129016)*x[0]**o+(-0.502811932395)*x[0]+(0.390751425646)*x[1]**o+(-0.0234058348564)*x[1]
        arg[1,3]=(-0.451410318068)*x[0]**o+(0.0959092789621)*x[0]+(0.364602756931)*x[1]**o+(-0.637883149677)*x[1]
        arg[1,4]=(0.84148898142)*x[0]**o+(0.447701396385)*x[0]+(-0.925960934959)*x[1]**o+(-0.216915201171)*x[1]
        arg[2,0]=(0.497685540053)*x[0]**o+(0.706445401439)*x[0]+(0.278688226804)*x[1]**o+(0.263560807934)*x[1]
        arg[2,1]=(0.846700886302)*x[0]**o+(0.794459987626)*x[0]+(0.206707548962)*x[1]**o+(0.536938178952)*x[1]
        arg[2,2]=(-0.825060225713)*x[0]**o+(0.00650901742391)*x[0]+(0.500591036904)*x[1]**o+(0.86251606869)*x[1]
        arg[2,3]=(-0.490203731083)*x[0]**o+(-0.758885551614)*x[0]+(0.44212368096)*x[1]**o+(0.422183463853)*x[1]
        arg[2,4]=(0.326180837967)*x[0]**o+(0.0511693925274)*x[0]+(0.450713817213)*x[1]**o+(-0.788372992203)*x[1]
        arg[3,0]=(0.427328335694)*x[0]**o+(-0.549570782632)*x[0]+(0.0484849428866)*x[1]**o+(-0.488921618371)*x[1]
        arg[3,1]=(-0.000395237589224)*x[0]**o+(-0.902889768871)*x[0]+(-0.991821787559)*x[1]**o+(-0.907842202962)*x[1]
        arg[3,2]=(0.737156280391)*x[0]**o+(0.104714673958)*x[0]+(0.102528971835)*x[1]**o+(-0.0982626427735)*x[1]
        arg[3,3]=(0.971460723075)*x[0]**o+(-0.759821212391)*x[0]+(0.385963754923)*x[1]**o+(-0.533924725996)*x[1]
        arg[3,4]=(0.25440424638)*x[0]**o+(-0.82432588605)*x[0]+(0.759285911224)*x[1]**o+(0.328467600746)*x[1]
        ref[0,0]=(-1.33097833792)*(1+2.*(dim-1.)/(o+1.))+(-0.772467439806)*dim
        ref[0,1]=(0.805483068724)*(1+2.*(dim-1.)/(o+1.))+(-0.36995225125)*dim
        ref[0,2]=(-1.11659649011)*(1+2.*(dim-1.)/(o+1.))+(0.389426290213)*dim
        ref[0,3]=(0.129148674655)*(1+2.*(dim-1.)/(o+1.))+(-0.173655053907)*dim
        ref[0,4]=(0.724727759407)*(1+2.*(dim-1.)/(o+1.))+(1.60951465082)*dim
        ref[1,0]=(0.880875243265)*(1+2.*(dim-1.)/(o+1.))+(0.866016038693)*dim
        ref[1,1]=(1.05093584536)*(1+2.*(dim-1.)/(o+1.))+(-0.288004631027)*dim
        ref[1,2]=(0.691195554662)*(1+2.*(dim-1.)/(o+1.))+(-0.526217767251)*dim
        ref[1,3]=(-0.0868075611369)*(1+2.*(dim-1.)/(o+1.))+(-0.541973870715)*dim
        ref[1,4]=(-0.0844719535392)*(1+2.*(dim-1.)/(o+1.))+(0.230786195214)*dim
        ref[2,0]=(0.776373766857)*(1+2.*(dim-1.)/(o+1.))+(0.970006209374)*dim
        ref[2,1]=(1.05340843526)*(1+2.*(dim-1.)/(o+1.))+(1.33139816658)*dim
        ref[2,2]=(-0.324469188808)*(1+2.*(dim-1.)/(o+1.))+(0.869025086114)*dim
        ref[2,3]=(-0.0480800501232)*(1+2.*(dim-1.)/(o+1.))+(-0.336702087761)*dim
        ref[2,4]=(0.776894655179)*(1+2.*(dim-1.)/(o+1.))+(-0.737203599675)*dim
        ref[3,0]=(0.47581327858)*(1+2.*(dim-1.)/(o+1.))+(-1.038492401)*dim
        ref[3,1]=(-0.992217025148)*(1+2.*(dim-1.)/(o+1.))+(-1.81073197183)*dim
        ref[3,2]=(0.839685252226)*(1+2.*(dim-1.)/(o+1.))+(0.00645203118434)*dim
        ref[3,3]=(1.357424478)*(1+2.*(dim-1.)/(o+1.))+(-1.29374593839)*dim
        ref[3,4]=(1.0136901576)*(1+2.*(dim-1.)/(o+1.))+(-0.495858285305)*dim
      else:
        arg[0,0]=(-0.543785449843)*x[0]**o+(-0.593471310422)*x[0]+(0.223732084513)*x[1]**o+(-0.126390494354)*x[1]+(0.602561108649)*x[2]**o+(-0.879650196727)*x[2]
        arg[0,1]=(-0.373708985284)*x[0]**o+(0.274723212925)*x[0]+(0.0901068561926)*x[1]**o+(0.0705878267418)*x[1]+(-0.87124855534)*x[2]**o+(-0.342156987834)*x[2]
        arg[0,2]=(0.197662792758)*x[0]**o+(0.993389348528)*x[0]+(0.300080063659)*x[1]**o+(-0.87329423474)*x[1]+(-0.708861668092)*x[2]**o+(-0.145567510817)*x[2]
        arg[0,3]=(-0.641680604707)*x[0]**o+(0.502870155472)*x[0]+(0.726460257571)*x[1]**o+(-0.796246941548)*x[1]+(-0.034138694789)*x[2]**o+(-0.756773445728)*x[2]
        arg[0,4]=(-0.910301138852)*x[0]**o+(0.370603440687)*x[0]+(0.30908632679)*x[1]**o+(0.399307258883)*x[1]+(0.734329887794)*x[2]**o+(-0.105355323212)*x[2]
        arg[1,0]=(-0.998162269861)*x[0]**o+(0.744436497607)*x[0]+(0.474991680035)*x[1]**o+(-0.523297766774)*x[1]+(-0.147411103004)*x[2]**o+(-0.366260797475)*x[2]
        arg[1,1]=(-0.895068058812)*x[0]**o+(-0.543026567202)*x[0]+(0.681910579939)*x[1]**o+(0.839238385122)*x[1]+(-0.943454697555)*x[2]**o+(-0.942882687085)*x[2]
        arg[1,2]=(-0.0327934757673)*x[0]**o+(-0.551060110327)*x[0]+(0.0246419122114)*x[1]**o+(0.290492943846)*x[1]+(-0.81629260256)*x[2]**o+(-0.0703654573398)*x[2]
        arg[1,3]=(0.0101368428324)*x[0]**o+(0.169228100407)*x[0]+(-0.0720675504637)*x[1]**o+(0.854569517064)*x[1]+(-0.185495614033)*x[2]**o+(0.159402892807)*x[2]
        arg[1,4]=(-0.879056526056)*x[0]**o+(-0.499865495878)*x[0]+(0.389281340553)*x[1]**o+(0.208323326648)*x[1]+(0.201495185869)*x[2]**o+(0.287693342543)*x[2]
        arg[2,0]=(-0.0980367959573)*x[0]**o+(-0.743908923973)*x[0]+(0.612332291133)*x[1]**o+(-0.833162883242)*x[1]+(-0.149234146487)*x[2]**o+(-0.265352382043)*x[2]
        arg[2,1]=(-0.731053054407)*x[0]**o+(-0.328824749845)*x[0]+(-0.270771833154)*x[1]**o+(-0.924718425874)*x[1]+(-0.333737248999)*x[2]**o+(-0.437707530432)*x[2]
        arg[2,2]=(0.0568240255423)*x[0]**o+(0.889929679474)*x[0]+(-0.550346007257)*x[1]**o+(-0.276114383355)*x[1]+(0.818856335741)*x[2]**o+(0.949830759195)*x[2]
        arg[2,3]=(-0.487278099196)*x[0]**o+(0.621833922878)*x[0]+(-0.641187646288)*x[1]**o+(-0.62187991518)*x[1]+(0.193101177186)*x[2]**o+(0.574039256278)*x[2]
        arg[2,4]=(0.411046922073)*x[0]**o+(-0.10196277248)*x[0]+(-0.880065306034)*x[1]**o+(0.629582927604)*x[1]+(-0.364618567951)*x[2]**o+(0.130293365676)*x[2]
        arg[3,0]=(-0.625244224555)*x[0]**o+(0.50345243521)*x[0]+(0.701616120451)*x[1]**o+(0.44401843794)*x[1]+(-0.881068054195)*x[2]**o+(0.337405641389)*x[2]
        arg[3,1]=(0.244113319139)*x[0]**o+(0.739760310044)*x[0]+(0.439528183979)*x[1]**o+(0.808496622595)*x[1]+(0.432069187682)*x[2]**o+(-0.106394550529)*x[2]
        arg[3,2]=(-0.200973139796)*x[0]**o+(-0.373932231384)*x[0]+(-0.438946407503)*x[1]**o+(0.543079830762)*x[1]+(-0.499599010695)*x[2]**o+(0.501967674296)*x[2]
        arg[3,3]=(-0.128263000609)*x[0]**o+(-0.408181851435)*x[0]+(0.403315054177)*x[1]**o+(0.75067593555)*x[1]+(0.361607987217)*x[2]**o+(0.269483801463)*x[2]
        arg[3,4]=(0.821969994275)*x[0]**o+(0.350771424951)*x[0]+(-0.761235507511)*x[1]**o+(0.148329827876)*x[1]+(-0.918493845231)*x[2]**o+(0.592094636272)*x[2]
        ref[0,0]=(0.282507743319)*(1+2.*(dim-1.)/(o+1.))+(-1.5995120015)*dim
        ref[0,1]=(-1.15485068443)*(1+2.*(dim-1.)/(o+1.))+(0.00315405183214)*dim
        ref[0,2]=(-0.211118811675)*(1+2.*(dim-1.)/(o+1.))+(-0.0254723970291)*dim
        ref[0,3]=(0.0506409580755)*(1+2.*(dim-1.)/(o+1.))+(-1.0501502318)*dim
        ref[0,4]=(0.133115075732)*(1+2.*(dim-1.)/(o+1.))+(0.664555376358)*dim
        ref[1,0]=(-0.67058169283)*(1+2.*(dim-1.)/(o+1.))+(-0.145122066642)*dim
        ref[1,1]=(-1.15661217643)*(1+2.*(dim-1.)/(o+1.))+(-0.646670869164)*dim
        ref[1,2]=(-0.824444166116)*(1+2.*(dim-1.)/(o+1.))+(-0.330932623821)*dim
        ref[1,3]=(-0.247426321664)*(1+2.*(dim-1.)/(o+1.))+(1.18320051028)*dim
        ref[1,4]=(-0.288279999634)*(1+2.*(dim-1.)/(o+1.))+(-0.00384882668765)*dim
        ref[2,0]=(0.365061348688)*(1+2.*(dim-1.)/(o+1.))+(-1.84242418926)*dim
        ref[2,1]=(-1.33556213656)*(1+2.*(dim-1.)/(o+1.))+(-1.69125070615)*dim
        ref[2,2]=(0.325334354026)*(1+2.*(dim-1.)/(o+1.))+(1.56364605531)*dim
        ref[2,3]=(-0.935364568298)*(1+2.*(dim-1.)/(o+1.))+(0.573993263976)*dim
        ref[2,4]=(-0.833636951913)*(1+2.*(dim-1.)/(o+1.))+(0.6579135208)*dim
        ref[3,0]=(-0.804696158299)*(1+2.*(dim-1.)/(o+1.))+(1.28487651454)*dim
        ref[3,1]=(1.1157106908)*(1+2.*(dim-1.)/(o+1.))+(1.44186238211)*dim
        ref[3,2]=(-1.13951855799)*(1+2.*(dim-1.)/(o+1.))+(0.671115273674)*dim
        ref[3,3]=(0.636660040784)*(1+2.*(dim-1.)/(o+1.))+(0.611977885578)*dim
        ref[3,4]=(-0.857759358467)*(1+2.*(dim-1.)/(o+1.))+(1.0911958891)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.331954160625)*x[0]**o+(0.713423346586)*x[0]+(0.922333771405)*x[1]**o+(-0.306799323245)*x[1]
        arg[0,0,1]=(-0.837278273696)*x[0]**o+(-0.322068356249)*x[0]+(-0.952056283446)*x[1]**o+(0.78147215355)*x[1]
        arg[0,1,0]=(-0.71217197633)*x[0]**o+(-0.222121067189)*x[0]+(-0.740851323426)*x[1]**o+(-0.664415003166)*x[1]
        arg[0,1,1]=(0.304797844524)*x[0]**o+(-0.644642249615)*x[0]+(-0.942312793933)*x[1]**o+(0.758857258357)*x[1]
        arg[1,0,0]=(-0.978734634026)*x[0]**o+(0.697829375729)*x[0]+(-0.161023313194)*x[1]**o+(0.61960280308)*x[1]
        arg[1,0,1]=(0.65091302745)*x[0]**o+(-0.411592147008)*x[0]+(0.748017319704)*x[1]**o+(0.751661257345)*x[1]
        arg[1,1,0]=(-0.752749816617)*x[0]**o+(-0.889417028079)*x[0]+(0.187075769961)*x[1]**o+(0.96605423194)*x[1]
        arg[1,1,1]=(-0.791276410919)*x[0]**o+(-0.66910236639)*x[0]+(-0.107283654349)*x[1]**o+(0.556315348977)*x[1]
        arg[2,0,0]=(0.0717781211643)*x[0]**o+(-0.895974517834)*x[0]+(-0.388350799289)*x[1]**o+(0.55541171641)*x[1]
        arg[2,0,1]=(0.886475701949)*x[0]**o+(0.0794480579219)*x[0]+(-0.77671944551)*x[1]**o+(-0.29611930088)*x[1]
        arg[2,1,0]=(0.891530897911)*x[0]**o+(-0.767892439984)*x[0]+(-0.692871902002)*x[1]**o+(0.296346783431)*x[1]
        arg[2,1,1]=(0.904417221616)*x[0]**o+(-0.379208335246)*x[0]+(0.491675297059)*x[1]**o+(0.994131453216)*x[1]
        arg[3,0,0]=(-0.278750282129)*x[0]**o+(-0.889235749208)*x[0]+(-0.970025256676)*x[1]**o+(0.459265023633)*x[1]
        arg[3,0,1]=(-0.718548177793)*x[0]**o+(-0.801618785436)*x[0]+(-0.0489351007933)*x[1]**o+(-0.0331281021404)*x[1]
        arg[3,1,0]=(0.838645233036)*x[0]**o+(0.174830460073)*x[0]+(0.343887140274)*x[1]**o+(0.475643625956)*x[1]
        arg[3,1,1]=(-0.376629360434)*x[0]**o+(0.965850026388)*x[0]+(-0.62284011199)*x[1]**o+(-0.548261522438)*x[1]
        arg[4,0,0]=(-0.747109622101)*x[0]**o+(-0.793286186527)*x[0]+(0.700502366132)*x[1]**o+(0.708459489695)*x[1]
        arg[4,0,1]=(-0.372920704339)*x[0]**o+(-0.856662712404)*x[0]+(0.735191558102)*x[1]**o+(0.451959287084)*x[1]
        arg[4,1,0]=(-0.825227447648)*x[0]**o+(-0.978701927456)*x[0]+(0.856562109765)*x[1]**o+(0.226820057686)*x[1]
        arg[4,1,1]=(-0.883951567752)*x[0]**o+(0.346114896878)*x[0]+(0.22103721575)*x[1]**o+(0.766907412924)*x[1]
        arg[5,0,0]=(0.440140408264)*x[0]**o+(0.576442002932)*x[0]+(0.181136518153)*x[1]**o+(0.467290325887)*x[1]
        arg[5,0,1]=(0.323653297993)*x[0]**o+(-0.0169080219964)*x[0]+(-0.790889359324)*x[1]**o+(-0.752609210154)*x[1]
        arg[5,1,0]=(-0.771859095701)*x[0]**o+(0.392074466446)*x[0]+(0.672832544535)*x[1]**o+(-0.518196376899)*x[1]
        arg[5,1,1]=(-0.714106835785)*x[0]**o+(-0.392974811152)*x[0]+(0.477349412757)*x[1]**o+(-0.978738343032)*x[1]
        ref[0,0,0]=(0.59037961078)*(1+2.*(dim-1.)/(o+1.))+(0.406624023341)*dim
        ref[0,0,1]=(-1.78933455714)*(1+2.*(dim-1.)/(o+1.))+(0.459403797301)*dim
        ref[0,1,0]=(-1.45302329976)*(1+2.*(dim-1.)/(o+1.))+(-0.886536070355)*dim
        ref[0,1,1]=(-0.637514949409)*(1+2.*(dim-1.)/(o+1.))+(0.114215008741)*dim
        ref[1,0,0]=(-1.13975794722)*(1+2.*(dim-1.)/(o+1.))+(1.31743217881)*dim
        ref[1,0,1]=(1.39893034715)*(1+2.*(dim-1.)/(o+1.))+(0.340069110337)*dim
        ref[1,1,0]=(-0.565674046656)*(1+2.*(dim-1.)/(o+1.))+(0.076637203861)*dim
        ref[1,1,1]=(-0.898560065268)*(1+2.*(dim-1.)/(o+1.))+(-0.112787017413)*dim
        ref[2,0,0]=(-0.316572678125)*(1+2.*(dim-1.)/(o+1.))+(-0.340562801424)*dim
        ref[2,0,1]=(0.10975625644)*(1+2.*(dim-1.)/(o+1.))+(-0.216671242958)*dim
        ref[2,1,0]=(0.198658995909)*(1+2.*(dim-1.)/(o+1.))+(-0.471545656553)*dim
        ref[2,1,1]=(1.39609251868)*(1+2.*(dim-1.)/(o+1.))+(0.614923117971)*dim
        ref[3,0,0]=(-1.2487755388)*(1+2.*(dim-1.)/(o+1.))+(-0.429970725575)*dim
        ref[3,0,1]=(-0.767483278586)*(1+2.*(dim-1.)/(o+1.))+(-0.834746887577)*dim
        ref[3,1,0]=(1.18253237331)*(1+2.*(dim-1.)/(o+1.))+(0.650474086029)*dim
        ref[3,1,1]=(-0.999469472424)*(1+2.*(dim-1.)/(o+1.))+(0.41758850395)*dim
        ref[4,0,0]=(-0.0466072559697)*(1+2.*(dim-1.)/(o+1.))+(-0.0848266968319)*dim
        ref[4,0,1]=(0.362270853763)*(1+2.*(dim-1.)/(o+1.))+(-0.404703425319)*dim
        ref[4,1,0]=(0.0313346621166)*(1+2.*(dim-1.)/(o+1.))+(-0.75188186977)*dim
        ref[4,1,1]=(-0.662914352001)*(1+2.*(dim-1.)/(o+1.))+(1.1130223098)*dim
        ref[5,0,0]=(0.621276926417)*(1+2.*(dim-1.)/(o+1.))+(1.04373232882)*dim
        ref[5,0,1]=(-0.467236061331)*(1+2.*(dim-1.)/(o+1.))+(-0.76951723215)*dim
        ref[5,1,0]=(-0.0990265511653)*(1+2.*(dim-1.)/(o+1.))+(-0.126121910453)*dim
        ref[5,1,1]=(-0.236757423028)*(1+2.*(dim-1.)/(o+1.))+(-1.37171315418)*dim
      else:
        arg[0,0,0]=(-0.676850371385)*x[0]**o+(0.343194249293)*x[0]+(-0.289569238124)*x[1]**o+(0.546143025789)*x[1]+(-0.292429216951)*x[2]**o+(0.100638805907)*x[2]
        arg[0,0,1]=(-0.558403199885)*x[0]**o+(0.369265961569)*x[0]+(0.098539502761)*x[1]**o+(-0.555747072012)*x[1]+(0.0142505043744)*x[2]**o+(0.110677590143)*x[2]
        arg[0,1,0]=(0.480265324249)*x[0]**o+(-0.0801799981418)*x[0]+(-0.659040110213)*x[1]**o+(0.315685251032)*x[1]+(0.146222780181)*x[2]**o+(-0.307845817615)*x[2]
        arg[0,1,1]=(-0.469643195547)*x[0]**o+(0.631180947887)*x[0]+(0.0873749097691)*x[1]**o+(-0.907805635326)*x[1]+(-0.445238378297)*x[2]**o+(-0.583750385876)*x[2]
        arg[1,0,0]=(-0.519648347947)*x[0]**o+(-0.580046561107)*x[0]+(-0.474702629402)*x[1]**o+(-0.76698196707)*x[1]+(0.631290211156)*x[2]**o+(0.283150656338)*x[2]
        arg[1,0,1]=(0.828049802595)*x[0]**o+(-0.0996271417739)*x[0]+(0.13247948046)*x[1]**o+(0.64657332105)*x[1]+(0.901448888477)*x[2]**o+(0.895379597076)*x[2]
        arg[1,1,0]=(-0.197367099356)*x[0]**o+(-0.933210661759)*x[0]+(-0.0134349749548)*x[1]**o+(-0.651111981269)*x[1]+(0.256945439519)*x[2]**o+(0.528412689094)*x[2]
        arg[1,1,1]=(-0.922943172982)*x[0]**o+(-0.883102479542)*x[0]+(-0.0037345936911)*x[1]**o+(-0.598256159261)*x[1]+(-0.8562089409)*x[2]**o+(-0.298502617143)*x[2]
        arg[2,0,0]=(0.499285053237)*x[0]**o+(-0.347877602065)*x[0]+(0.783117300662)*x[1]**o+(0.0413016895448)*x[1]+(0.22555777414)*x[2]**o+(0.089246166469)*x[2]
        arg[2,0,1]=(-0.180602931202)*x[0]**o+(0.883958054663)*x[0]+(0.904974519387)*x[1]**o+(-0.169618307434)*x[1]+(0.361963890467)*x[2]**o+(-0.364236450353)*x[2]
        arg[2,1,0]=(0.332239090882)*x[0]**o+(-0.392239851855)*x[0]+(0.905142609029)*x[1]**o+(-0.471493540964)*x[1]+(-0.305327788141)*x[2]**o+(0.581213078834)*x[2]
        arg[2,1,1]=(0.481276189253)*x[0]**o+(0.645400809127)*x[0]+(-0.840844470697)*x[1]**o+(-0.47587396865)*x[1]+(0.134145050668)*x[2]**o+(-0.315081369951)*x[2]
        arg[3,0,0]=(-0.480497217552)*x[0]**o+(0.595345393671)*x[0]+(0.776315451392)*x[1]**o+(-0.396110510671)*x[1]+(0.0437422791996)*x[2]**o+(0.348675326203)*x[2]
        arg[3,0,1]=(0.867117243092)*x[0]**o+(-0.222483151231)*x[0]+(0.0856439550139)*x[1]**o+(0.355855801395)*x[1]+(-0.779380984653)*x[2]**o+(0.799440639884)*x[2]
        arg[3,1,0]=(-0.869234399353)*x[0]**o+(-0.747500724089)*x[0]+(-0.83154206968)*x[1]**o+(-0.771906212766)*x[1]+(-0.188734869619)*x[2]**o+(0.7312382704)*x[2]
        arg[3,1,1]=(-0.729677636148)*x[0]**o+(-0.358066533942)*x[0]+(-0.342651052704)*x[1]**o+(-0.531453801913)*x[1]+(-0.898300547827)*x[2]**o+(0.0387220517653)*x[2]
        arg[4,0,0]=(-0.380489041999)*x[0]**o+(0.517823660835)*x[0]+(0.722149782685)*x[1]**o+(0.510169845258)*x[1]+(-0.979340093133)*x[2]**o+(0.846050053391)*x[2]
        arg[4,0,1]=(0.390895580804)*x[0]**o+(0.122461788897)*x[0]+(-0.809243912452)*x[1]**o+(0.871689879803)*x[1]+(-0.991689466458)*x[2]**o+(-0.0373953689911)*x[2]
        arg[4,1,0]=(0.00670611033858)*x[0]**o+(-0.700084486793)*x[0]+(0.365358002826)*x[1]**o+(-0.928347441251)*x[1]+(-0.675081280461)*x[2]**o+(-0.262871553902)*x[2]
        arg[4,1,1]=(-0.566521655794)*x[0]**o+(-0.23325381374)*x[0]+(0.744538505444)*x[1]**o+(0.263970013855)*x[1]+(0.360984907343)*x[2]**o+(0.704696800059)*x[2]
        arg[5,0,0]=(0.754229123633)*x[0]**o+(-0.523353302691)*x[0]+(0.404181763418)*x[1]**o+(0.175414644029)*x[1]+(0.627455124887)*x[2]**o+(-0.450163514418)*x[2]
        arg[5,0,1]=(0.804547370453)*x[0]**o+(-0.748034210243)*x[0]+(-0.943261499953)*x[1]**o+(0.360715335834)*x[1]+(-0.952785406204)*x[2]**o+(0.188838771539)*x[2]
        arg[5,1,0]=(0.653730390007)*x[0]**o+(0.663746229033)*x[0]+(0.454911944273)*x[1]**o+(-0.548110740107)*x[1]+(0.799171265206)*x[2]**o+(-0.646356900812)*x[2]
        arg[5,1,1]=(0.533464266683)*x[0]**o+(-0.0650985333228)*x[0]+(-0.599880569232)*x[1]**o+(0.645120695827)*x[1]+(0.985316456657)*x[2]**o+(0.527848503917)*x[2]
        ref[0,0,0]=(-1.25884882646)*(1+2.*(dim-1.)/(o+1.))+(0.989976080988)*dim
        ref[0,0,1]=(-0.445613192749)*(1+2.*(dim-1.)/(o+1.))+(-0.0758035202994)*dim
        ref[0,1,0]=(-0.0325520057832)*(1+2.*(dim-1.)/(o+1.))+(-0.0723405647253)*dim
        ref[0,1,1]=(-0.827506664075)*(1+2.*(dim-1.)/(o+1.))+(-0.860375073315)*dim
        ref[1,0,0]=(-0.363060766193)*(1+2.*(dim-1.)/(o+1.))+(-1.06387787184)*dim
        ref[1,0,1]=(1.86197817153)*(1+2.*(dim-1.)/(o+1.))+(1.44232577635)*dim
        ref[1,1,0]=(0.0461433652081)*(1+2.*(dim-1.)/(o+1.))+(-1.05590995393)*dim
        ref[1,1,1]=(-1.78288670757)*(1+2.*(dim-1.)/(o+1.))+(-1.77986125595)*dim
        ref[2,0,0]=(1.50796012804)*(1+2.*(dim-1.)/(o+1.))+(-0.217329746051)*dim
        ref[2,0,1]=(1.08633547865)*(1+2.*(dim-1.)/(o+1.))+(0.350103296876)*dim
        ref[2,1,0]=(0.932053911771)*(1+2.*(dim-1.)/(o+1.))+(-0.282520313985)*dim
        ref[2,1,1]=(-0.225423230776)*(1+2.*(dim-1.)/(o+1.))+(-0.145554529474)*dim
        ref[3,0,0]=(0.33956051304)*(1+2.*(dim-1.)/(o+1.))+(0.547910209204)*dim
        ref[3,0,1]=(0.173380213454)*(1+2.*(dim-1.)/(o+1.))+(0.932813290049)*dim
        ref[3,1,0]=(-1.88951133865)*(1+2.*(dim-1.)/(o+1.))+(-0.788168666455)*dim
        ref[3,1,1]=(-1.97062923668)*(1+2.*(dim-1.)/(o+1.))+(-0.850798284089)*dim
        ref[4,0,0]=(-0.637679352447)*(1+2.*(dim-1.)/(o+1.))+(1.87404355948)*dim
        ref[4,0,1]=(-1.41003779811)*(1+2.*(dim-1.)/(o+1.))+(0.956756299709)*dim
        ref[4,1,0]=(-0.303017167297)*(1+2.*(dim-1.)/(o+1.))+(-1.89130348195)*dim
        ref[4,1,1]=(0.539001756993)*(1+2.*(dim-1.)/(o+1.))+(0.735413000173)*dim
        ref[5,0,0]=(1.78586601194)*(1+2.*(dim-1.)/(o+1.))+(-0.79810217308)*dim
        ref[5,0,1]=(-1.0914995357)*(1+2.*(dim-1.)/(o+1.))+(-0.198480102871)*dim
        ref[5,1,0]=(1.90781359949)*(1+2.*(dim-1.)/(o+1.))+(-0.530721411887)*dim
        ref[5,1,1]=(0.918900154108)*(1+2.*(dim-1.)/(o+1.))+(1.10787066642)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.181966612335)*x[0]**o+(-0.461883712153)*x[0]+(-0.420217253531)*x[1]**o+(0.52197109374)*x[1]
        arg[0,0,0,1]=(-0.122704378373)*x[0]**o+(0.549708706786)*x[0]+(-0.728108908074)*x[1]**o+(-0.839262748545)*x[1]
        arg[0,0,1,0]=(0.58557741427)*x[0]**o+(-0.970633305521)*x[0]+(0.215513381389)*x[1]**o+(-0.178877868196)*x[1]
        arg[0,0,1,1]=(0.646118894595)*x[0]**o+(0.839917605008)*x[0]+(-0.608843960266)*x[1]**o+(-0.840842208189)*x[1]
        arg[0,0,2,0]=(-0.126004594384)*x[0]**o+(0.663487355648)*x[0]+(0.801751176464)*x[1]**o+(-0.76620842256)*x[1]
        arg[0,0,2,1]=(-0.244425023768)*x[0]**o+(0.736809513316)*x[0]+(0.736776514098)*x[1]**o+(0.382648973627)*x[1]
        arg[0,1,0,0]=(-0.264460763052)*x[0]**o+(-0.325868872518)*x[0]+(-0.0636203008538)*x[1]**o+(-0.398059740187)*x[1]
        arg[0,1,0,1]=(0.443639751435)*x[0]**o+(0.523903519493)*x[0]+(-0.222575248757)*x[1]**o+(-0.168203019401)*x[1]
        arg[0,1,1,0]=(-0.329523829258)*x[0]**o+(-0.908408437436)*x[0]+(0.695441587034)*x[1]**o+(-0.236957609006)*x[1]
        arg[0,1,1,1]=(-0.57363301023)*x[0]**o+(-0.168677045141)*x[0]+(0.455235475098)*x[1]**o+(-0.498803217861)*x[1]
        arg[0,1,2,0]=(0.09906404243)*x[0]**o+(0.48682122702)*x[0]+(-0.250940180366)*x[1]**o+(-0.641013418008)*x[1]
        arg[0,1,2,1]=(-0.963193555161)*x[0]**o+(0.129868691037)*x[0]+(0.821192174481)*x[1]**o+(-0.245576703223)*x[1]
        arg[0,2,0,0]=(-0.867410251576)*x[0]**o+(0.170194043006)*x[0]+(0.837669366536)*x[1]**o+(0.285772890787)*x[1]
        arg[0,2,0,1]=(0.767089114154)*x[0]**o+(-0.425174482112)*x[0]+(-0.460840645159)*x[1]**o+(-0.314295694392)*x[1]
        arg[0,2,1,0]=(-0.932934920269)*x[0]**o+(0.399552332498)*x[0]+(-0.137764102275)*x[1]**o+(-0.0375128952657)*x[1]
        arg[0,2,1,1]=(0.507999826854)*x[0]**o+(-0.60826518686)*x[0]+(0.766838061511)*x[1]**o+(0.800723296773)*x[1]
        arg[0,2,2,0]=(0.372064864992)*x[0]**o+(0.29029476011)*x[0]+(0.203299629846)*x[1]**o+(-0.386357855884)*x[1]
        arg[0,2,2,1]=(0.661268311956)*x[0]**o+(0.60257115577)*x[0]+(0.88881711536)*x[1]**o+(0.335431084294)*x[1]
        arg[0,3,0,0]=(-0.756295024079)*x[0]**o+(-0.982472476783)*x[0]+(-0.264060825724)*x[1]**o+(0.082684535426)*x[1]
        arg[0,3,0,1]=(0.514580120205)*x[0]**o+(0.524816925461)*x[0]+(-0.745096223094)*x[1]**o+(-0.0796881815577)*x[1]
        arg[0,3,1,0]=(0.597556449974)*x[0]**o+(0.527680495979)*x[0]+(0.517638167762)*x[1]**o+(-0.357836534374)*x[1]
        arg[0,3,1,1]=(0.354366560101)*x[0]**o+(-0.617697325436)*x[0]+(0.339071431286)*x[1]**o+(0.299934132319)*x[1]
        arg[0,3,2,0]=(0.26563455123)*x[0]**o+(0.821495614519)*x[0]+(0.177496963332)*x[1]**o+(-0.154521746257)*x[1]
        arg[0,3,2,1]=(0.983477388632)*x[0]**o+(0.524806318175)*x[0]+(-0.604096007297)*x[1]**o+(0.350360164328)*x[1]
        arg[0,4,0,0]=(0.311369778768)*x[0]**o+(-0.648416457513)*x[0]+(0.463469290711)*x[1]**o+(0.0719628828709)*x[1]
        arg[0,4,0,1]=(-0.743081056195)*x[0]**o+(-0.803573973461)*x[0]+(-0.766982068691)*x[1]**o+(0.387240020398)*x[1]
        arg[0,4,1,0]=(-0.0715117411345)*x[0]**o+(-0.341871715212)*x[0]+(0.0593378446291)*x[1]**o+(-0.10183864979)*x[1]
        arg[0,4,1,1]=(0.538850329978)*x[0]**o+(0.0280201497914)*x[0]+(0.350206859589)*x[1]**o+(0.316831165676)*x[1]
        arg[0,4,2,0]=(-0.327852385434)*x[0]**o+(0.408866306659)*x[0]+(0.940378517318)*x[1]**o+(0.578309623908)*x[1]
        arg[0,4,2,1]=(0.378380112483)*x[0]**o+(0.696828716067)*x[0]+(-0.908233000582)*x[1]**o+(-0.328611483675)*x[1]
        arg[1,0,0,0]=(0.753991558436)*x[0]**o+(0.0542891140783)*x[0]+(-0.428934638249)*x[1]**o+(0.318155063446)*x[1]
        arg[1,0,0,1]=(-0.387609085343)*x[0]**o+(0.711023478158)*x[0]+(-0.969026331169)*x[1]**o+(0.456175434422)*x[1]
        arg[1,0,1,0]=(-0.468504469917)*x[0]**o+(-0.314756281356)*x[0]+(0.0145647044588)*x[1]**o+(0.074035761055)*x[1]
        arg[1,0,1,1]=(0.240749620446)*x[0]**o+(0.906736060537)*x[0]+(0.391320123368)*x[1]**o+(-0.261639387239)*x[1]
        arg[1,0,2,0]=(-0.461436333094)*x[0]**o+(-0.809258428817)*x[0]+(0.0670777162266)*x[1]**o+(-0.445304287721)*x[1]
        arg[1,0,2,1]=(-0.569794448172)*x[0]**o+(-0.0624052432724)*x[0]+(-0.759840810858)*x[1]**o+(-0.609106375016)*x[1]
        arg[1,1,0,0]=(0.751570143287)*x[0]**o+(0.78189919599)*x[0]+(-0.103246458909)*x[1]**o+(0.732156635312)*x[1]
        arg[1,1,0,1]=(-0.0126214545951)*x[0]**o+(-0.612853966263)*x[0]+(-0.238555698245)*x[1]**o+(-0.668009933698)*x[1]
        arg[1,1,1,0]=(0.816325838743)*x[0]**o+(0.545518357738)*x[0]+(0.893296287925)*x[1]**o+(0.694579919864)*x[1]
        arg[1,1,1,1]=(0.927638834795)*x[0]**o+(0.750347688466)*x[0]+(-0.15924462002)*x[1]**o+(-0.952995938892)*x[1]
        arg[1,1,2,0]=(0.970761253678)*x[0]**o+(0.205308852313)*x[0]+(0.909209906751)*x[1]**o+(-0.178635180174)*x[1]
        arg[1,1,2,1]=(-0.632065128479)*x[0]**o+(-0.300905398347)*x[0]+(0.865317638002)*x[1]**o+(0.886398250669)*x[1]
        arg[1,2,0,0]=(-0.0529293620539)*x[0]**o+(-0.909022003179)*x[0]+(0.235231916345)*x[1]**o+(-0.623380752415)*x[1]
        arg[1,2,0,1]=(0.873344987616)*x[0]**o+(0.144231546529)*x[0]+(0.307375566381)*x[1]**o+(-0.201617507084)*x[1]
        arg[1,2,1,0]=(-0.957945047658)*x[0]**o+(-0.432955474381)*x[0]+(0.0742288164869)*x[1]**o+(-0.106080430366)*x[1]
        arg[1,2,1,1]=(-0.664227561512)*x[0]**o+(0.0821007781774)*x[0]+(-0.541506076231)*x[1]**o+(-0.88307957937)*x[1]
        arg[1,2,2,0]=(0.754702675232)*x[0]**o+(0.175176916913)*x[0]+(0.519607893979)*x[1]**o+(-0.24129526204)*x[1]
        arg[1,2,2,1]=(-0.415660998528)*x[0]**o+(-0.964071393069)*x[0]+(0.458656303526)*x[1]**o+(0.752153826846)*x[1]
        arg[1,3,0,0]=(-0.15638393187)*x[0]**o+(0.0828824719)*x[0]+(-0.0846205011509)*x[1]**o+(-0.438939463736)*x[1]
        arg[1,3,0,1]=(0.222994482074)*x[0]**o+(-0.488426101986)*x[0]+(-0.799262651825)*x[1]**o+(0.0638880385288)*x[1]
        arg[1,3,1,0]=(0.690529298963)*x[0]**o+(-0.845248154386)*x[0]+(0.748635110096)*x[1]**o+(0.820882250869)*x[1]
        arg[1,3,1,1]=(-0.0989162396328)*x[0]**o+(-0.171252734498)*x[0]+(-0.363447812361)*x[1]**o+(-0.06937701403)*x[1]
        arg[1,3,2,0]=(-0.739647298124)*x[0]**o+(-0.83384135572)*x[0]+(0.546124768813)*x[1]**o+(0.415761104064)*x[1]
        arg[1,3,2,1]=(0.859487140106)*x[0]**o+(0.481214854724)*x[0]+(0.788685241166)*x[1]**o+(-0.776410333263)*x[1]
        arg[1,4,0,0]=(-0.312428038681)*x[0]**o+(-0.178596293803)*x[0]+(-0.229770546807)*x[1]**o+(-0.43765633326)*x[1]
        arg[1,4,0,1]=(-0.974305286188)*x[0]**o+(-0.978154435248)*x[0]+(-0.671799350916)*x[1]**o+(0.645287291694)*x[1]
        arg[1,4,1,0]=(-0.219711371151)*x[0]**o+(-0.432784911532)*x[0]+(0.552940388883)*x[1]**o+(-0.622742205614)*x[1]
        arg[1,4,1,1]=(0.677289749287)*x[0]**o+(-0.92825132284)*x[0]+(0.818767358259)*x[1]**o+(-0.426000012305)*x[1]
        arg[1,4,2,0]=(-0.507474696573)*x[0]**o+(0.0882151625837)*x[0]+(0.826611187078)*x[1]**o+(0.157884229299)*x[1]
        arg[1,4,2,1]=(0.902398378671)*x[0]**o+(-0.372679984773)*x[0]+(-0.998272135687)*x[1]**o+(0.292294917465)*x[1]
        arg[2,0,0,0]=(0.699035433674)*x[0]**o+(0.380640017665)*x[0]+(-0.650735771214)*x[1]**o+(-0.55187467604)*x[1]
        arg[2,0,0,1]=(0.703438074924)*x[0]**o+(0.795508285134)*x[0]+(-0.0650847718675)*x[1]**o+(-0.0250766383001)*x[1]
        arg[2,0,1,0]=(0.493174434488)*x[0]**o+(-0.775523541313)*x[0]+(0.23184619496)*x[1]**o+(0.343368328234)*x[1]
        arg[2,0,1,1]=(0.404974280918)*x[0]**o+(0.0673627086244)*x[0]+(-0.843957699955)*x[1]**o+(0.645895713789)*x[1]
        arg[2,0,2,0]=(0.985047143685)*x[0]**o+(0.969582300743)*x[0]+(0.929427858098)*x[1]**o+(-0.951787413112)*x[1]
        arg[2,0,2,1]=(0.618505056215)*x[0]**o+(0.685760720523)*x[0]+(0.14594727158)*x[1]**o+(0.816579780815)*x[1]
        arg[2,1,0,0]=(0.0824953418089)*x[0]**o+(-0.896753351387)*x[0]+(0.415841596602)*x[1]**o+(0.349436600222)*x[1]
        arg[2,1,0,1]=(-0.904246301273)*x[0]**o+(0.51403543319)*x[0]+(0.0950147832662)*x[1]**o+(0.0482220256459)*x[1]
        arg[2,1,1,0]=(0.653418547332)*x[0]**o+(0.703094950575)*x[0]+(-0.294924381355)*x[1]**o+(0.889408306649)*x[1]
        arg[2,1,1,1]=(-0.329361050221)*x[0]**o+(-0.93072021487)*x[0]+(0.0845194032793)*x[1]**o+(-0.310617126183)*x[1]
        arg[2,1,2,0]=(-0.282371473471)*x[0]**o+(0.387458491178)*x[0]+(0.643908016757)*x[1]**o+(0.303741991451)*x[1]
        arg[2,1,2,1]=(-0.416872502153)*x[0]**o+(0.373465764073)*x[0]+(-0.90689140487)*x[1]**o+(-0.351052228378)*x[1]
        arg[2,2,0,0]=(0.494198770026)*x[0]**o+(0.744917026846)*x[0]+(-0.806591346974)*x[1]**o+(-0.812396479226)*x[1]
        arg[2,2,0,1]=(0.230573331783)*x[0]**o+(-0.449815323831)*x[0]+(-0.662178492084)*x[1]**o+(0.0281786827584)*x[1]
        arg[2,2,1,0]=(-0.546296599764)*x[0]**o+(-0.457939754236)*x[0]+(-0.0432225723667)*x[1]**o+(0.0431706529852)*x[1]
        arg[2,2,1,1]=(-0.808174460895)*x[0]**o+(-0.473584203382)*x[0]+(0.567514752294)*x[1]**o+(-0.115780229631)*x[1]
        arg[2,2,2,0]=(-0.134425008633)*x[0]**o+(-0.103550646345)*x[0]+(-0.588484951305)*x[1]**o+(0.955148281194)*x[1]
        arg[2,2,2,1]=(0.550095417437)*x[0]**o+(0.943564321046)*x[0]+(-0.990366145789)*x[1]**o+(0.872652302953)*x[1]
        arg[2,3,0,0]=(0.209341033615)*x[0]**o+(-0.285197252362)*x[0]+(0.961111991429)*x[1]**o+(0.810022268034)*x[1]
        arg[2,3,0,1]=(0.238319839582)*x[0]**o+(-0.714990515235)*x[0]+(0.0925371590198)*x[1]**o+(0.938145796375)*x[1]
        arg[2,3,1,0]=(0.889068347007)*x[0]**o+(-0.301904392078)*x[0]+(0.41386941601)*x[1]**o+(-0.973115118701)*x[1]
        arg[2,3,1,1]=(-0.522083695846)*x[0]**o+(-0.582455208417)*x[0]+(0.725431346859)*x[1]**o+(-0.950157996033)*x[1]
        arg[2,3,2,0]=(-0.897838142917)*x[0]**o+(0.745488698358)*x[0]+(-0.38910727894)*x[1]**o+(0.614539961316)*x[1]
        arg[2,3,2,1]=(0.924500645496)*x[0]**o+(-0.68988446876)*x[0]+(-0.466238481805)*x[1]**o+(-0.957703006413)*x[1]
        arg[2,4,0,0]=(0.280692018949)*x[0]**o+(0.600320489083)*x[0]+(-0.402390231845)*x[1]**o+(0.906721159456)*x[1]
        arg[2,4,0,1]=(0.50594574787)*x[0]**o+(0.130986910768)*x[0]+(0.272643351394)*x[1]**o+(0.53790545383)*x[1]
        arg[2,4,1,0]=(-0.627308461808)*x[0]**o+(0.690560028115)*x[0]+(-0.455768647207)*x[1]**o+(0.134926199015)*x[1]
        arg[2,4,1,1]=(-0.501163279392)*x[0]**o+(0.980144824993)*x[0]+(-0.0358930723173)*x[1]**o+(0.720157452357)*x[1]
        arg[2,4,2,0]=(-0.609861117539)*x[0]**o+(0.965699461849)*x[0]+(0.305674136837)*x[1]**o+(-0.578333683729)*x[1]
        arg[2,4,2,1]=(-0.921051416627)*x[0]**o+(-0.703479746028)*x[0]+(-0.0189027931731)*x[1]**o+(0.403415497212)*x[1]
        arg[3,0,0,0]=(-0.846266714886)*x[0]**o+(-0.103621657972)*x[0]+(0.217340457606)*x[1]**o+(-0.537654290693)*x[1]
        arg[3,0,0,1]=(0.186251745403)*x[0]**o+(0.694458841658)*x[0]+(0.126837362883)*x[1]**o+(0.0997502401792)*x[1]
        arg[3,0,1,0]=(0.153737843143)*x[0]**o+(0.811590040899)*x[0]+(0.189363446699)*x[1]**o+(0.00688399857093)*x[1]
        arg[3,0,1,1]=(-0.912444421339)*x[0]**o+(-0.68504171198)*x[0]+(-0.820717153919)*x[1]**o+(-0.00475951426309)*x[1]
        arg[3,0,2,0]=(0.908522439455)*x[0]**o+(-0.911126936858)*x[0]+(0.0424198145962)*x[1]**o+(-0.352097158128)*x[1]
        arg[3,0,2,1]=(0.935127271967)*x[0]**o+(0.154700792222)*x[0]+(-0.283148208173)*x[1]**o+(-0.607541482501)*x[1]
        arg[3,1,0,0]=(-0.131329160543)*x[0]**o+(0.24170180662)*x[0]+(-0.00893591013587)*x[1]**o+(0.262635048958)*x[1]
        arg[3,1,0,1]=(-0.559200924561)*x[0]**o+(0.509310116425)*x[0]+(0.484486720282)*x[1]**o+(-0.89314080386)*x[1]
        arg[3,1,1,0]=(0.355094442114)*x[0]**o+(-0.748398915688)*x[0]+(-0.379187476624)*x[1]**o+(0.447409858108)*x[1]
        arg[3,1,1,1]=(0.170855109103)*x[0]**o+(-0.48925054508)*x[0]+(0.840559727559)*x[1]**o+(0.0359562471889)*x[1]
        arg[3,1,2,0]=(0.827586094521)*x[0]**o+(0.0231820503547)*x[0]+(0.457192172855)*x[1]**o+(0.826960230377)*x[1]
        arg[3,1,2,1]=(-0.962890676345)*x[0]**o+(-0.112439391098)*x[0]+(-0.865510686025)*x[1]**o+(-0.539059426684)*x[1]
        arg[3,2,0,0]=(-0.165030011664)*x[0]**o+(-0.285694090419)*x[0]+(-0.522579583443)*x[1]**o+(0.73946640768)*x[1]
        arg[3,2,0,1]=(0.809351962503)*x[0]**o+(-0.872047628102)*x[0]+(0.61415433255)*x[1]**o+(0.132268088986)*x[1]
        arg[3,2,1,0]=(-0.587823544835)*x[0]**o+(-0.416092975475)*x[0]+(0.829027070431)*x[1]**o+(0.805271620727)*x[1]
        arg[3,2,1,1]=(-0.0107935093527)*x[0]**o+(0.988434287979)*x[0]+(0.941743055822)*x[1]**o+(0.137807879334)*x[1]
        arg[3,2,2,0]=(-0.855931527408)*x[0]**o+(0.575075608849)*x[0]+(0.11916967111)*x[1]**o+(-0.472844336533)*x[1]
        arg[3,2,2,1]=(-0.542112514875)*x[0]**o+(0.081308081083)*x[0]+(-0.934575648869)*x[1]**o+(-0.218208369535)*x[1]
        arg[3,3,0,0]=(0.486635391195)*x[0]**o+(0.995728500002)*x[0]+(-0.686767180456)*x[1]**o+(0.246849483855)*x[1]
        arg[3,3,0,1]=(-0.331571848212)*x[0]**o+(0.801870431206)*x[0]+(-0.91142654055)*x[1]**o+(0.089854528457)*x[1]
        arg[3,3,1,0]=(-0.657455652936)*x[0]**o+(-0.721466128651)*x[0]+(0.0420989631458)*x[1]**o+(-0.982316453325)*x[1]
        arg[3,3,1,1]=(-0.709257565086)*x[0]**o+(0.689028730506)*x[0]+(-0.196516960475)*x[1]**o+(-0.216704841188)*x[1]
        arg[3,3,2,0]=(-0.910547650394)*x[0]**o+(-0.977471949987)*x[0]+(0.37319115353)*x[1]**o+(-0.248978955796)*x[1]
        arg[3,3,2,1]=(-0.71025804576)*x[0]**o+(0.373559860529)*x[0]+(-0.9064074733)*x[1]**o+(-0.542012058073)*x[1]
        arg[3,4,0,0]=(-0.649273580176)*x[0]**o+(0.525784948768)*x[0]+(-0.213929862516)*x[1]**o+(-0.162295172275)*x[1]
        arg[3,4,0,1]=(0.209784328978)*x[0]**o+(-0.972601183002)*x[0]+(-0.730886568265)*x[1]**o+(0.966415518759)*x[1]
        arg[3,4,1,0]=(-0.334060094293)*x[0]**o+(0.29371522739)*x[0]+(0.936156150108)*x[1]**o+(-0.257026769476)*x[1]
        arg[3,4,1,1]=(-0.0465451767064)*x[0]**o+(0.115525862669)*x[0]+(0.891280073228)*x[1]**o+(-0.0436233989979)*x[1]
        arg[3,4,2,0]=(-0.748697989123)*x[0]**o+(0.246672415445)*x[0]+(0.62642302161)*x[1]**o+(-0.411644161038)*x[1]
        arg[3,4,2,1]=(0.850132316298)*x[0]**o+(0.939463534845)*x[0]+(0.0437756618735)*x[1]**o+(-0.955198146074)*x[1]
        ref[0,0,0,0]=(-0.238250641196)*(1+2.*(dim-1.)/(o+1.))+(0.0600873815876)*dim
        ref[0,0,0,1]=(-0.850813286447)*(1+2.*(dim-1.)/(o+1.))+(-0.28955404176)*dim
        ref[0,0,1,0]=(0.801090795659)*(1+2.*(dim-1.)/(o+1.))+(-1.14951117372)*dim
        ref[0,0,1,1]=(0.0372749343294)*(1+2.*(dim-1.)/(o+1.))+(-0.000924603180129)*dim
        ref[0,0,2,0]=(0.675746582079)*(1+2.*(dim-1.)/(o+1.))+(-0.102721066912)*dim
        ref[0,0,2,1]=(0.49235149033)*(1+2.*(dim-1.)/(o+1.))+(1.11945848694)*dim
        ref[0,1,0,0]=(-0.328081063906)*(1+2.*(dim-1.)/(o+1.))+(-0.723928612706)*dim
        ref[0,1,0,1]=(0.221064502678)*(1+2.*(dim-1.)/(o+1.))+(0.355700500092)*dim
        ref[0,1,1,0]=(0.365917757777)*(1+2.*(dim-1.)/(o+1.))+(-1.14536604644)*dim
        ref[0,1,1,1]=(-0.118397535132)*(1+2.*(dim-1.)/(o+1.))+(-0.667480263002)*dim
        ref[0,1,2,0]=(-0.151876137936)*(1+2.*(dim-1.)/(o+1.))+(-0.154192190988)*dim
        ref[0,1,2,1]=(-0.14200138068)*(1+2.*(dim-1.)/(o+1.))+(-0.115708012186)*dim
        ref[0,2,0,0]=(-0.0297408850402)*(1+2.*(dim-1.)/(o+1.))+(0.455966933793)*dim
        ref[0,2,0,1]=(0.306248468995)*(1+2.*(dim-1.)/(o+1.))+(-0.739470176504)*dim
        ref[0,2,1,0]=(-1.07069902254)*(1+2.*(dim-1.)/(o+1.))+(0.362039437233)*dim
        ref[0,2,1,1]=(1.27483788836)*(1+2.*(dim-1.)/(o+1.))+(0.192458109913)*dim
        ref[0,2,2,0]=(0.575364494839)*(1+2.*(dim-1.)/(o+1.))+(-0.0960630957749)*dim
        ref[0,2,2,1]=(1.55008542732)*(1+2.*(dim-1.)/(o+1.))+(0.938002240064)*dim
        ref[0,3,0,0]=(-1.0203558498)*(1+2.*(dim-1.)/(o+1.))+(-0.899787941357)*dim
        ref[0,3,0,1]=(-0.230516102889)*(1+2.*(dim-1.)/(o+1.))+(0.445128743903)*dim
        ref[0,3,1,0]=(1.11519461774)*(1+2.*(dim-1.)/(o+1.))+(0.169843961604)*dim
        ref[0,3,1,1]=(0.693437991388)*(1+2.*(dim-1.)/(o+1.))+(-0.317763193117)*dim
        ref[0,3,2,0]=(0.443131514562)*(1+2.*(dim-1.)/(o+1.))+(0.666973868262)*dim
        ref[0,3,2,1]=(0.379381381335)*(1+2.*(dim-1.)/(o+1.))+(0.875166482503)*dim
        ref[0,4,0,0]=(0.774839069479)*(1+2.*(dim-1.)/(o+1.))+(-0.576453574642)*dim
        ref[0,4,0,1]=(-1.51006312489)*(1+2.*(dim-1.)/(o+1.))+(-0.416333953063)*dim
        ref[0,4,1,0]=(-0.0121738965054)*(1+2.*(dim-1.)/(o+1.))+(-0.443710365002)*dim
        ref[0,4,1,1]=(0.889057189568)*(1+2.*(dim-1.)/(o+1.))+(0.344851315468)*dim
        ref[0,4,2,0]=(0.612526131884)*(1+2.*(dim-1.)/(o+1.))+(0.987175930567)*dim
        ref[0,4,2,1]=(-0.529852888099)*(1+2.*(dim-1.)/(o+1.))+(0.368217232392)*dim
        ref[1,0,0,0]=(0.325056920187)*(1+2.*(dim-1.)/(o+1.))+(0.372444177524)*dim
        ref[1,0,0,1]=(-1.35663541651)*(1+2.*(dim-1.)/(o+1.))+(1.16719891258)*dim
        ref[1,0,1,0]=(-0.453939765458)*(1+2.*(dim-1.)/(o+1.))+(-0.240720520301)*dim
        ref[1,0,1,1]=(0.632069743814)*(1+2.*(dim-1.)/(o+1.))+(0.645096673298)*dim
        ref[1,0,2,0]=(-0.394358616867)*(1+2.*(dim-1.)/(o+1.))+(-1.25456271654)*dim
        ref[1,0,2,1]=(-1.32963525903)*(1+2.*(dim-1.)/(o+1.))+(-0.671511618288)*dim
        ref[1,1,0,0]=(0.648323684378)*(1+2.*(dim-1.)/(o+1.))+(1.5140558313)*dim
        ref[1,1,0,1]=(-0.251177152841)*(1+2.*(dim-1.)/(o+1.))+(-1.28086389996)*dim
        ref[1,1,1,0]=(1.70962212667)*(1+2.*(dim-1.)/(o+1.))+(1.2400982776)*dim
        ref[1,1,1,1]=(0.768394214775)*(1+2.*(dim-1.)/(o+1.))+(-0.202648250425)*dim
        ref[1,1,2,0]=(1.87997116043)*(1+2.*(dim-1.)/(o+1.))+(0.0266736721392)*dim
        ref[1,1,2,1]=(0.233252509524)*(1+2.*(dim-1.)/(o+1.))+(0.585492852322)*dim
        ref[1,2,0,0]=(0.182302554291)*(1+2.*(dim-1.)/(o+1.))+(-1.53240275559)*dim
        ref[1,2,0,1]=(1.180720554)*(1+2.*(dim-1.)/(o+1.))+(-0.0573859605549)*dim
        ref[1,2,1,0]=(-0.883716231171)*(1+2.*(dim-1.)/(o+1.))+(-0.539035904746)*dim
        ref[1,2,1,1]=(-1.20573363774)*(1+2.*(dim-1.)/(o+1.))+(-0.800978801193)*dim
        ref[1,2,2,0]=(1.27431056921)*(1+2.*(dim-1.)/(o+1.))+(-0.066118345127)*dim
        ref[1,2,2,1]=(0.0429953049979)*(1+2.*(dim-1.)/(o+1.))+(-0.211917566223)*dim
        ref[1,3,0,0]=(-0.241004433021)*(1+2.*(dim-1.)/(o+1.))+(-0.356056991836)*dim
        ref[1,3,0,1]=(-0.576268169751)*(1+2.*(dim-1.)/(o+1.))+(-0.424538063457)*dim
        ref[1,3,1,0]=(1.43916440906)*(1+2.*(dim-1.)/(o+1.))+(-0.0243659035174)*dim
        ref[1,3,1,1]=(-0.462364051993)*(1+2.*(dim-1.)/(o+1.))+(-0.240629748528)*dim
        ref[1,3,2,0]=(-0.193522529311)*(1+2.*(dim-1.)/(o+1.))+(-0.418080251656)*dim
        ref[1,3,2,1]=(1.64817238127)*(1+2.*(dim-1.)/(o+1.))+(-0.295195478539)*dim
        ref[1,4,0,0]=(-0.542198585489)*(1+2.*(dim-1.)/(o+1.))+(-0.616252627062)*dim
        ref[1,4,0,1]=(-1.6461046371)*(1+2.*(dim-1.)/(o+1.))+(-0.332867143554)*dim
        ref[1,4,1,0]=(0.333229017732)*(1+2.*(dim-1.)/(o+1.))+(-1.05552711715)*dim
        ref[1,4,1,1]=(1.49605710755)*(1+2.*(dim-1.)/(o+1.))+(-1.35425133515)*dim
        ref[1,4,2,0]=(0.319136490505)*(1+2.*(dim-1.)/(o+1.))+(0.246099391883)*dim
        ref[1,4,2,1]=(-0.0958737570155)*(1+2.*(dim-1.)/(o+1.))+(-0.0803850673082)*dim
        ref[2,0,0,0]=(0.0482996624604)*(1+2.*(dim-1.)/(o+1.))+(-0.171234658375)*dim
        ref[2,0,0,1]=(0.638353303057)*(1+2.*(dim-1.)/(o+1.))+(0.770431646834)*dim
        ref[2,0,1,0]=(0.725020629449)*(1+2.*(dim-1.)/(o+1.))+(-0.432155213079)*dim
        ref[2,0,1,1]=(-0.438983419037)*(1+2.*(dim-1.)/(o+1.))+(0.713258422414)*dim
        ref[2,0,2,0]=(1.91447500178)*(1+2.*(dim-1.)/(o+1.))+(0.0177948876308)*dim
        ref[2,0,2,1]=(0.764452327795)*(1+2.*(dim-1.)/(o+1.))+(1.50234050134)*dim
        ref[2,1,0,0]=(0.498336938411)*(1+2.*(dim-1.)/(o+1.))+(-0.547316751164)*dim
        ref[2,1,0,1]=(-0.809231518007)*(1+2.*(dim-1.)/(o+1.))+(0.562257458836)*dim
        ref[2,1,1,0]=(0.358494165977)*(1+2.*(dim-1.)/(o+1.))+(1.59250325722)*dim
        ref[2,1,1,1]=(-0.244841646941)*(1+2.*(dim-1.)/(o+1.))+(-1.24133734105)*dim
        ref[2,1,2,0]=(0.361536543287)*(1+2.*(dim-1.)/(o+1.))+(0.691200482629)*dim
        ref[2,1,2,1]=(-1.32376390702)*(1+2.*(dim-1.)/(o+1.))+(0.0224135356952)*dim
        ref[2,2,0,0]=(-0.312392576948)*(1+2.*(dim-1.)/(o+1.))+(-0.06747945238)*dim
        ref[2,2,0,1]=(-0.431605160301)*(1+2.*(dim-1.)/(o+1.))+(-0.421636641073)*dim
        ref[2,2,1,0]=(-0.589519172131)*(1+2.*(dim-1.)/(o+1.))+(-0.414769101251)*dim
        ref[2,2,1,1]=(-0.240659708601)*(1+2.*(dim-1.)/(o+1.))+(-0.589364433013)*dim
        ref[2,2,2,0]=(-0.722909959939)*(1+2.*(dim-1.)/(o+1.))+(0.851597634849)*dim
        ref[2,2,2,1]=(-0.440270728352)*(1+2.*(dim-1.)/(o+1.))+(1.816216624)*dim
        ref[2,3,0,0]=(1.17045302504)*(1+2.*(dim-1.)/(o+1.))+(0.524825015672)*dim
        ref[2,3,0,1]=(0.330856998602)*(1+2.*(dim-1.)/(o+1.))+(0.22315528114)*dim
        ref[2,3,1,0]=(1.30293776302)*(1+2.*(dim-1.)/(o+1.))+(-1.27501951078)*dim
        ref[2,3,1,1]=(0.203347651013)*(1+2.*(dim-1.)/(o+1.))+(-1.53261320445)*dim
        ref[2,3,2,0]=(-1.28694542186)*(1+2.*(dim-1.)/(o+1.))+(1.36002865967)*dim
        ref[2,3,2,1]=(0.458262163691)*(1+2.*(dim-1.)/(o+1.))+(-1.64758747517)*dim
        ref[2,4,0,0]=(-0.121698212896)*(1+2.*(dim-1.)/(o+1.))+(1.50704164854)*dim
        ref[2,4,0,1]=(0.778589099264)*(1+2.*(dim-1.)/(o+1.))+(0.668892364598)*dim
        ref[2,4,1,0]=(-1.08307710902)*(1+2.*(dim-1.)/(o+1.))+(0.82548622713)*dim
        ref[2,4,1,1]=(-0.537056351709)*(1+2.*(dim-1.)/(o+1.))+(1.70030227735)*dim
        ref[2,4,2,0]=(-0.304186980702)*(1+2.*(dim-1.)/(o+1.))+(0.38736577812)*dim
        ref[2,4,2,1]=(-0.9399542098)*(1+2.*(dim-1.)/(o+1.))+(-0.300064248815)*dim
        ref[3,0,0,0]=(-0.62892625728)*(1+2.*(dim-1.)/(o+1.))+(-0.641275948665)*dim
        ref[3,0,0,1]=(0.313089108286)*(1+2.*(dim-1.)/(o+1.))+(0.794209081837)*dim
        ref[3,0,1,0]=(0.343101289842)*(1+2.*(dim-1.)/(o+1.))+(0.81847403947)*dim
        ref[3,0,1,1]=(-1.73316157526)*(1+2.*(dim-1.)/(o+1.))+(-0.689801226243)*dim
        ref[3,0,2,0]=(0.950942254051)*(1+2.*(dim-1.)/(o+1.))+(-1.26322409499)*dim
        ref[3,0,2,1]=(0.651979063795)*(1+2.*(dim-1.)/(o+1.))+(-0.452840690279)*dim
        ref[3,1,0,0]=(-0.140265070679)*(1+2.*(dim-1.)/(o+1.))+(0.504336855579)*dim
        ref[3,1,0,1]=(-0.074714204279)*(1+2.*(dim-1.)/(o+1.))+(-0.383830687435)*dim
        ref[3,1,1,0]=(-0.0240930345095)*(1+2.*(dim-1.)/(o+1.))+(-0.30098905758)*dim
        ref[3,1,1,1]=(1.01141483666)*(1+2.*(dim-1.)/(o+1.))+(-0.453294297891)*dim
        ref[3,1,2,0]=(1.28477826738)*(1+2.*(dim-1.)/(o+1.))+(0.850142280731)*dim
        ref[3,1,2,1]=(-1.82840136237)*(1+2.*(dim-1.)/(o+1.))+(-0.651498817783)*dim
        ref[3,2,0,0]=(-0.687609595107)*(1+2.*(dim-1.)/(o+1.))+(0.453772317262)*dim
        ref[3,2,0,1]=(1.42350629505)*(1+2.*(dim-1.)/(o+1.))+(-0.739779539116)*dim
        ref[3,2,1,0]=(0.241203525596)*(1+2.*(dim-1.)/(o+1.))+(0.389178645253)*dim
        ref[3,2,1,1]=(0.930949546469)*(1+2.*(dim-1.)/(o+1.))+(1.12624216731)*dim
        ref[3,2,2,0]=(-0.736761856298)*(1+2.*(dim-1.)/(o+1.))+(0.102231272315)*dim
        ref[3,2,2,1]=(-1.47668816374)*(1+2.*(dim-1.)/(o+1.))+(-0.136900288452)*dim
        ref[3,3,0,0]=(-0.200131789262)*(1+2.*(dim-1.)/(o+1.))+(1.24257798386)*dim
        ref[3,3,0,1]=(-1.24299838876)*(1+2.*(dim-1.)/(o+1.))+(0.891724959663)*dim
        ref[3,3,1,0]=(-0.61535668979)*(1+2.*(dim-1.)/(o+1.))+(-1.70378258198)*dim
        ref[3,3,1,1]=(-0.905774525561)*(1+2.*(dim-1.)/(o+1.))+(0.472323889318)*dim
        ref[3,3,2,0]=(-0.537356496864)*(1+2.*(dim-1.)/(o+1.))+(-1.22645090578)*dim
        ref[3,3,2,1]=(-1.61666551906)*(1+2.*(dim-1.)/(o+1.))+(-0.168452197544)*dim
        ref[3,4,0,0]=(-0.863203442693)*(1+2.*(dim-1.)/(o+1.))+(0.363489776493)*dim
        ref[3,4,0,1]=(-0.521102239287)*(1+2.*(dim-1.)/(o+1.))+(-0.00618566424301)*dim
        ref[3,4,1,0]=(0.602096055815)*(1+2.*(dim-1.)/(o+1.))+(0.0366884579134)*dim
        ref[3,4,1,1]=(0.844734896522)*(1+2.*(dim-1.)/(o+1.))+(0.0719024636709)*dim
        ref[3,4,2,0]=(-0.122274967513)*(1+2.*(dim-1.)/(o+1.))+(-0.164971745593)*dim
        ref[3,4,2,1]=(0.893907978172)*(1+2.*(dim-1.)/(o+1.))+(-0.0157346112292)*dim
      else:
        arg[0,0,0,0]=(-0.0445963085788)*x[0]**o+(0.649401127647)*x[0]+(-0.949035518207)*x[1]**o+(0.620675629012)*x[1]+(0.379727527503)*x[2]**o+(0.0321033403935)*x[2]
        arg[0,0,0,1]=(-0.334483142061)*x[0]**o+(-0.408749824143)*x[0]+(-0.427042441291)*x[1]**o+(-0.263818177378)*x[1]+(-0.0490001897328)*x[2]**o+(0.925792628022)*x[2]
        arg[0,0,1,0]=(0.632498844224)*x[0]**o+(-0.756487058254)*x[0]+(-0.0277391041902)*x[1]**o+(0.615017171892)*x[1]+(-0.0956053270216)*x[2]**o+(-0.0383148218356)*x[2]
        arg[0,0,1,1]=(0.560189295143)*x[0]**o+(-0.565632451768)*x[0]+(0.131609132282)*x[1]**o+(0.689012872903)*x[1]+(-0.0542418297927)*x[2]**o+(-0.816694105058)*x[2]
        arg[0,0,2,0]=(-0.606523401606)*x[0]**o+(-0.129782638438)*x[0]+(0.286554604329)*x[1]**o+(-0.264164004519)*x[1]+(-0.145168356054)*x[2]**o+(0.346147680755)*x[2]
        arg[0,0,2,1]=(-0.847528471559)*x[0]**o+(-0.50641170288)*x[0]+(-0.773325001399)*x[1]**o+(0.517649361249)*x[1]+(-0.531000881434)*x[2]**o+(0.220222077563)*x[2]
        arg[0,1,0,0]=(-0.0946495378384)*x[0]**o+(-0.312650449416)*x[0]+(0.183321918505)*x[1]**o+(-0.359448777446)*x[1]+(0.713219915733)*x[2]**o+(-0.436742264028)*x[2]
        arg[0,1,0,1]=(0.856947337921)*x[0]**o+(0.497228192699)*x[0]+(0.351718755501)*x[1]**o+(-0.928210477058)*x[1]+(0.586853747677)*x[2]**o+(-0.915642724205)*x[2]
        arg[0,1,1,0]=(-0.745330469339)*x[0]**o+(-0.724610807566)*x[0]+(0.821804346007)*x[1]**o+(0.241991559492)*x[1]+(0.508380905493)*x[2]**o+(0.756706999359)*x[2]
        arg[0,1,1,1]=(-0.647952004012)*x[0]**o+(0.940964624139)*x[0]+(-0.620895246845)*x[1]**o+(0.577736885116)*x[1]+(0.228716730257)*x[2]**o+(0.336914427419)*x[2]
        arg[0,1,2,0]=(-0.671541115529)*x[0]**o+(0.153595444626)*x[0]+(0.420649971716)*x[1]**o+(-0.459078302559)*x[1]+(-0.583263239511)*x[2]**o+(-0.927250932344)*x[2]
        arg[0,1,2,1]=(-0.893954406097)*x[0]**o+(0.266297374506)*x[0]+(-0.777463374391)*x[1]**o+(0.0905062188555)*x[1]+(-0.668756964585)*x[2]**o+(-0.580980380472)*x[2]
        arg[0,2,0,0]=(0.424449538697)*x[0]**o+(0.0787502268502)*x[0]+(-0.645996834851)*x[1]**o+(-0.0409730346273)*x[1]+(0.835463212243)*x[2]**o+(0.716052012249)*x[2]
        arg[0,2,0,1]=(0.0241441154932)*x[0]**o+(-0.580242261617)*x[0]+(0.693484316968)*x[1]**o+(0.198598746328)*x[1]+(-0.586283866696)*x[2]**o+(-0.595008846645)*x[2]
        arg[0,2,1,0]=(0.074643270363)*x[0]**o+(0.454053362138)*x[0]+(0.861855819289)*x[1]**o+(-0.0310194739742)*x[1]+(-0.35738290172)*x[2]**o+(-0.400226698584)*x[2]
        arg[0,2,1,1]=(0.61780464269)*x[0]**o+(-0.702818770068)*x[0]+(0.0840653474306)*x[1]**o+(0.786842849442)*x[1]+(0.0704342755849)*x[2]**o+(0.386459053939)*x[2]
        arg[0,2,2,0]=(-0.815363379053)*x[0]**o+(-0.621823615812)*x[0]+(-0.13535005543)*x[1]**o+(-0.127080917834)*x[1]+(-0.178866000441)*x[2]**o+(-0.0753933476244)*x[2]
        arg[0,2,2,1]=(0.400194366678)*x[0]**o+(0.075003395019)*x[0]+(-0.35708995046)*x[1]**o+(-0.370028795924)*x[1]+(-0.674944523932)*x[2]**o+(-0.437373855225)*x[2]
        arg[0,3,0,0]=(-0.736317751792)*x[0]**o+(0.671627391036)*x[0]+(0.0768737429532)*x[1]**o+(-0.90036869218)*x[1]+(0.796767364925)*x[2]**o+(-0.185034722102)*x[2]
        arg[0,3,0,1]=(0.530921403378)*x[0]**o+(0.970385957718)*x[0]+(-0.75659968724)*x[1]**o+(-0.437200812366)*x[1]+(0.0519374612092)*x[2]**o+(0.657323406602)*x[2]
        arg[0,3,1,0]=(-0.109751102222)*x[0]**o+(-0.154990305064)*x[0]+(-0.502004367351)*x[1]**o+(0.535266424739)*x[1]+(-0.968926773395)*x[2]**o+(-0.138032126456)*x[2]
        arg[0,3,1,1]=(0.509506900641)*x[0]**o+(-0.872357145765)*x[0]+(-0.154129018225)*x[1]**o+(0.0190914956768)*x[1]+(0.119094726078)*x[2]**o+(0.29593674915)*x[2]
        arg[0,3,2,0]=(-0.133398979819)*x[0]**o+(-0.780852347888)*x[0]+(-0.346576066687)*x[1]**o+(0.187923380934)*x[1]+(0.939817544138)*x[2]**o+(-0.497081081638)*x[2]
        arg[0,3,2,1]=(-0.908783140711)*x[0]**o+(0.541754659297)*x[0]+(-0.297350066105)*x[1]**o+(0.118676031731)*x[1]+(0.985854358996)*x[2]**o+(-0.424715359194)*x[2]
        arg[0,4,0,0]=(0.341191520307)*x[0]**o+(-0.252381555804)*x[0]+(-0.307289805415)*x[1]**o+(0.762228901652)*x[1]+(0.390901852611)*x[2]**o+(0.147349139229)*x[2]
        arg[0,4,0,1]=(0.236625833503)*x[0]**o+(0.856322905978)*x[0]+(-0.160244298088)*x[1]**o+(-0.971940590481)*x[1]+(-0.0829366716357)*x[2]**o+(-0.687610358703)*x[2]
        arg[0,4,1,0]=(0.619591455739)*x[0]**o+(-0.62589432832)*x[0]+(-0.0275687591477)*x[1]**o+(0.367590198294)*x[1]+(0.745367285277)*x[2]**o+(0.232571513762)*x[2]
        arg[0,4,1,1]=(0.317669904199)*x[0]**o+(0.741734360559)*x[0]+(-0.923644886773)*x[1]**o+(-0.725994528938)*x[1]+(-0.949385908603)*x[2]**o+(-0.0515088731457)*x[2]
        arg[0,4,2,0]=(0.82547889841)*x[0]**o+(-0.926970312436)*x[0]+(0.946941744691)*x[1]**o+(-0.171582097545)*x[1]+(-0.185628570945)*x[2]**o+(0.40495393086)*x[2]
        arg[0,4,2,1]=(0.51879613184)*x[0]**o+(-0.517447781109)*x[0]+(-0.288826585096)*x[1]**o+(-0.736890998485)*x[1]+(-0.167119326999)*x[2]**o+(-0.653082372893)*x[2]
        arg[1,0,0,0]=(-0.879008868811)*x[0]**o+(-0.0613987509179)*x[0]+(-0.671288168832)*x[1]**o+(-0.959517297906)*x[1]+(0.292554039162)*x[2]**o+(0.942246368678)*x[2]
        arg[1,0,0,1]=(0.599956815218)*x[0]**o+(0.731601305574)*x[0]+(-0.61593947163)*x[1]**o+(-0.597924796707)*x[1]+(-0.791620695721)*x[2]**o+(-0.562978718552)*x[2]
        arg[1,0,1,0]=(0.448381207242)*x[0]**o+(-0.178004176549)*x[0]+(0.631841924913)*x[1]**o+(0.311836637044)*x[1]+(0.0619475841251)*x[2]**o+(0.606564490624)*x[2]
        arg[1,0,1,1]=(0.159907398582)*x[0]**o+(-0.278605461684)*x[0]+(0.822914081538)*x[1]**o+(0.75112795022)*x[1]+(0.91389895436)*x[2]**o+(0.929177149059)*x[2]
        arg[1,0,2,0]=(-0.472286972241)*x[0]**o+(0.00950883684535)*x[0]+(0.41110884679)*x[1]**o+(0.109422159345)*x[1]+(-0.707436024011)*x[2]**o+(0.795057014561)*x[2]
        arg[1,0,2,1]=(-0.0156784802605)*x[0]**o+(-0.657283760492)*x[0]+(-0.0422880440635)*x[1]**o+(-0.326436491982)*x[1]+(-0.543917764251)*x[2]**o+(-0.691698129266)*x[2]
        arg[1,1,0,0]=(-0.814424723604)*x[0]**o+(0.746801606056)*x[0]+(0.22025107532)*x[1]**o+(-0.785254149057)*x[1]+(-0.845637273475)*x[2]**o+(-0.468272754144)*x[2]
        arg[1,1,0,1]=(-0.0815177460988)*x[0]**o+(-0.47210691717)*x[0]+(-0.285964111717)*x[1]**o+(0.527387496664)*x[1]+(0.87059583429)*x[2]**o+(-0.543899395468)*x[2]
        arg[1,1,1,0]=(0.424163426295)*x[0]**o+(0.649692204305)*x[0]+(0.0570010318344)*x[1]**o+(-0.426047169951)*x[1]+(0.334750596342)*x[2]**o+(-0.746549570927)*x[2]
        arg[1,1,1,1]=(-0.120493849358)*x[0]**o+(0.567677110538)*x[0]+(-0.355891513526)*x[1]**o+(0.68069794787)*x[1]+(0.786560156246)*x[2]**o+(-0.422804642829)*x[2]
        arg[1,1,2,0]=(-0.80298202932)*x[0]**o+(-0.674766348751)*x[0]+(-0.29945072468)*x[1]**o+(0.385941328636)*x[1]+(-0.756272624651)*x[2]**o+(-0.607541099269)*x[2]
        arg[1,1,2,1]=(0.915020713521)*x[0]**o+(-0.875957199283)*x[0]+(0.0217597029367)*x[1]**o+(-0.0649461013997)*x[1]+(-0.302450259541)*x[2]**o+(0.738498325526)*x[2]
        arg[1,2,0,0]=(0.239905462913)*x[0]**o+(0.197348893262)*x[0]+(-0.151173414577)*x[1]**o+(0.368405558905)*x[1]+(0.891448032058)*x[2]**o+(0.874975769684)*x[2]
        arg[1,2,0,1]=(-0.304971614823)*x[0]**o+(-0.0166065674003)*x[0]+(0.883397157318)*x[1]**o+(-0.918803270424)*x[1]+(0.590875856583)*x[2]**o+(-0.360659400522)*x[2]
        arg[1,2,1,0]=(0.185368540041)*x[0]**o+(0.498866146124)*x[0]+(0.591601152982)*x[1]**o+(-0.986932631453)*x[1]+(-0.525706872664)*x[2]**o+(-0.0462993899877)*x[2]
        arg[1,2,1,1]=(0.260677894495)*x[0]**o+(-0.216847258131)*x[0]+(0.932876363707)*x[1]**o+(0.761630676808)*x[1]+(0.260889221066)*x[2]**o+(0.546259908582)*x[2]
        arg[1,2,2,0]=(-0.316803600202)*x[0]**o+(-0.898044300642)*x[0]+(-0.0865001657852)*x[1]**o+(0.760424275199)*x[1]+(0.771564532427)*x[2]**o+(-0.0852237819025)*x[2]
        arg[1,2,2,1]=(-0.54274371504)*x[0]**o+(0.0822524750896)*x[0]+(0.872655885939)*x[1]**o+(0.594054041006)*x[1]+(0.322150325252)*x[2]**o+(0.774020612668)*x[2]
        arg[1,3,0,0]=(0.034593806358)*x[0]**o+(0.541638857337)*x[0]+(0.291359214696)*x[1]**o+(-0.203872730757)*x[1]+(0.0312260615928)*x[2]**o+(0.62568815669)*x[2]
        arg[1,3,0,1]=(-0.31929050432)*x[0]**o+(-0.931756481553)*x[0]+(0.755537511071)*x[1]**o+(-0.182578278538)*x[1]+(0.00958902784528)*x[2]**o+(0.345520390047)*x[2]
        arg[1,3,1,0]=(-0.435083251076)*x[0]**o+(0.248332963857)*x[0]+(-0.122701495966)*x[1]**o+(0.669350594899)*x[1]+(-0.0758596665333)*x[2]**o+(0.0784997907526)*x[2]
        arg[1,3,1,1]=(-0.725479421981)*x[0]**o+(0.657242746138)*x[0]+(-0.199958263735)*x[1]**o+(-0.784165440236)*x[1]+(-0.550936676647)*x[2]**o+(-0.31907813038)*x[2]
        arg[1,3,2,0]=(0.657177346512)*x[0]**o+(-0.168053866293)*x[0]+(0.782142168258)*x[1]**o+(0.660514922866)*x[1]+(0.788106295165)*x[2]**o+(-0.254290645125)*x[2]
        arg[1,3,2,1]=(-0.215166688682)*x[0]**o+(0.743577328376)*x[0]+(-0.374827404054)*x[1]**o+(0.267885661143)*x[1]+(0.728818301014)*x[2]**o+(0.6828999278)*x[2]
        arg[1,4,0,0]=(0.371646403413)*x[0]**o+(0.00998777673275)*x[0]+(-0.313378345871)*x[1]**o+(-0.952580545501)*x[1]+(-0.680596974717)*x[2]**o+(-0.181426263871)*x[2]
        arg[1,4,0,1]=(-0.828437965469)*x[0]**o+(0.302443878606)*x[0]+(0.581310358652)*x[1]**o+(0.714092786618)*x[1]+(0.650179420347)*x[2]**o+(0.663482256421)*x[2]
        arg[1,4,1,0]=(0.220318191456)*x[0]**o+(0.443108440795)*x[0]+(-0.538413779146)*x[1]**o+(0.660052783252)*x[1]+(-0.846698606343)*x[2]**o+(-0.0447806256913)*x[2]
        arg[1,4,1,1]=(0.866921895372)*x[0]**o+(0.776291161128)*x[0]+(-0.385318933957)*x[1]**o+(0.841514060614)*x[1]+(-0.401465329431)*x[2]**o+(0.474475442638)*x[2]
        arg[1,4,2,0]=(-0.222535436657)*x[0]**o+(0.921341450452)*x[0]+(-0.765043798058)*x[1]**o+(-0.492946667792)*x[1]+(0.758817008648)*x[2]**o+(0.0950595321909)*x[2]
        arg[1,4,2,1]=(0.333884542089)*x[0]**o+(0.671265678974)*x[0]+(-0.719262407869)*x[1]**o+(0.498176251457)*x[1]+(-0.867726438404)*x[2]**o+(-0.872614195608)*x[2]
        arg[2,0,0,0]=(-0.303331399994)*x[0]**o+(-0.563887060818)*x[0]+(0.64537626236)*x[1]**o+(-0.657611314342)*x[1]+(0.800882746828)*x[2]**o+(0.700671892157)*x[2]
        arg[2,0,0,1]=(0.417691239931)*x[0]**o+(-0.822789801933)*x[0]+(0.712214951539)*x[1]**o+(0.160960575173)*x[1]+(0.0400205244239)*x[2]**o+(0.49892939112)*x[2]
        arg[2,0,1,0]=(-0.828677767124)*x[0]**o+(-0.347411290435)*x[0]+(-0.700177402223)*x[1]**o+(-0.42498634463)*x[1]+(0.984172081328)*x[2]**o+(-0.677168643292)*x[2]
        arg[2,0,1,1]=(-0.423091398412)*x[0]**o+(-0.786913239913)*x[0]+(-0.888102258019)*x[1]**o+(-0.119525894432)*x[1]+(-0.158513498081)*x[2]**o+(0.503403674219)*x[2]
        arg[2,0,2,0]=(0.547217868614)*x[0]**o+(-0.737401808325)*x[0]+(-0.865530737659)*x[1]**o+(-0.750671412749)*x[1]+(-0.06731420082)*x[2]**o+(0.658572191108)*x[2]
        arg[2,0,2,1]=(0.501785022047)*x[0]**o+(-0.0542441989122)*x[0]+(-0.379614025724)*x[1]**o+(-0.785618793316)*x[1]+(0.994476365391)*x[2]**o+(-0.436996817494)*x[2]
        arg[2,1,0,0]=(0.0329439903445)*x[0]**o+(0.688097592632)*x[0]+(0.554030860489)*x[1]**o+(0.60949736297)*x[1]+(0.499716171437)*x[2]**o+(0.501464282299)*x[2]
        arg[2,1,0,1]=(0.459629414019)*x[0]**o+(0.810658146454)*x[0]+(-0.902023596001)*x[1]**o+(-0.982998009297)*x[1]+(0.487736529196)*x[2]**o+(-0.509164692477)*x[2]
        arg[2,1,1,0]=(0.0818151896334)*x[0]**o+(-0.931023788966)*x[0]+(-0.217349745187)*x[1]**o+(-0.785586110912)*x[1]+(0.16783052001)*x[2]**o+(-0.515867681053)*x[2]
        arg[2,1,1,1]=(-0.758874975703)*x[0]**o+(0.219408816573)*x[0]+(0.508187203896)*x[1]**o+(0.306700464008)*x[1]+(-0.641891226105)*x[2]**o+(-0.137752888266)*x[2]
        arg[2,1,2,0]=(-0.952430879482)*x[0]**o+(0.179895758684)*x[0]+(0.502285174801)*x[1]**o+(-0.82300826605)*x[1]+(-0.267848001419)*x[2]**o+(-0.882529775244)*x[2]
        arg[2,1,2,1]=(0.371372868601)*x[0]**o+(-0.521801538326)*x[0]+(0.894906517432)*x[1]**o+(0.205799057307)*x[1]+(-0.315840560279)*x[2]**o+(0.65823096555)*x[2]
        arg[2,2,0,0]=(0.585088762354)*x[0]**o+(0.527662713737)*x[0]+(0.83511043292)*x[1]**o+(-0.263597549489)*x[1]+(-0.342210974079)*x[2]**o+(0.532149528802)*x[2]
        arg[2,2,0,1]=(0.985868978475)*x[0]**o+(0.162798954854)*x[0]+(-0.887525429296)*x[1]**o+(0.618905848201)*x[1]+(-0.213643353241)*x[2]**o+(0.0170513098359)*x[2]
        arg[2,2,1,0]=(-0.216402064475)*x[0]**o+(0.350853951704)*x[0]+(0.592219675443)*x[1]**o+(-0.915293916561)*x[1]+(0.160588856435)*x[2]**o+(-0.574686420146)*x[2]
        arg[2,2,1,1]=(0.233118497878)*x[0]**o+(0.482035958353)*x[0]+(-0.581971661478)*x[1]**o+(0.832450520208)*x[1]+(-0.430914444794)*x[2]**o+(-0.791658025494)*x[2]
        arg[2,2,2,0]=(-0.710516844)*x[0]**o+(0.0266391730256)*x[0]+(-0.154541997151)*x[1]**o+(-0.232843778379)*x[1]+(0.453522207498)*x[2]**o+(-0.688939516711)*x[2]
        arg[2,2,2,1]=(0.26860058468)*x[0]**o+(-0.969638661751)*x[0]+(0.623784614857)*x[1]**o+(0.686359051774)*x[1]+(0.677056704385)*x[2]**o+(0.156688176847)*x[2]
        arg[2,3,0,0]=(-0.0620909841534)*x[0]**o+(-0.364946660535)*x[0]+(-0.388275430806)*x[1]**o+(-0.947908160286)*x[1]+(0.0774302410344)*x[2]**o+(0.374469033649)*x[2]
        arg[2,3,0,1]=(0.164845147121)*x[0]**o+(0.480213031852)*x[0]+(0.58157030274)*x[1]**o+(0.882820521347)*x[1]+(0.584714825748)*x[2]**o+(0.0285154925376)*x[2]
        arg[2,3,1,0]=(0.664711234489)*x[0]**o+(0.247415435607)*x[0]+(0.741592287061)*x[1]**o+(0.963505986041)*x[1]+(-0.987844687003)*x[2]**o+(0.998294459659)*x[2]
        arg[2,3,1,1]=(-0.734529316764)*x[0]**o+(-0.177480004713)*x[0]+(0.923555424779)*x[1]**o+(0.727331498256)*x[1]+(0.452312367674)*x[2]**o+(0.00652243396085)*x[2]
        arg[2,3,2,0]=(-0.367377897953)*x[0]**o+(-0.157529442804)*x[0]+(-0.121881528621)*x[1]**o+(0.865116726884)*x[1]+(-0.200087079925)*x[2]**o+(0.930889629788)*x[2]
        arg[2,3,2,1]=(-0.900394562978)*x[0]**o+(-0.762334071565)*x[0]+(-0.723776245294)*x[1]**o+(0.332728994729)*x[1]+(0.390453779523)*x[2]**o+(0.0447350797932)*x[2]
        arg[2,4,0,0]=(0.178230443819)*x[0]**o+(0.0898732089623)*x[0]+(0.808448952186)*x[1]**o+(0.164579824494)*x[1]+(-0.553032263217)*x[2]**o+(-0.782141879825)*x[2]
        arg[2,4,0,1]=(-0.105582488753)*x[0]**o+(0.995415696949)*x[0]+(0.978971670365)*x[1]**o+(-0.977866735017)*x[1]+(0.156938966924)*x[2]**o+(-0.120871488936)*x[2]
        arg[2,4,1,0]=(0.652354601889)*x[0]**o+(0.381186764672)*x[0]+(-0.783479778972)*x[1]**o+(-0.158898829615)*x[1]+(0.794693790129)*x[2]**o+(0.914720093741)*x[2]
        arg[2,4,1,1]=(0.571506419376)*x[0]**o+(0.906086581152)*x[0]+(-0.798519792125)*x[1]**o+(-0.22840897962)*x[1]+(-0.750891721145)*x[2]**o+(-0.591014038221)*x[2]
        arg[2,4,2,0]=(0.515982716306)*x[0]**o+(0.299569326845)*x[0]+(-0.407423180268)*x[1]**o+(-0.252103363573)*x[1]+(-0.0228238994531)*x[2]**o+(-0.282150542844)*x[2]
        arg[2,4,2,1]=(0.5335415942)*x[0]**o+(-0.885728570489)*x[0]+(-0.750810131327)*x[1]**o+(0.604053940679)*x[1]+(-0.0946405866686)*x[2]**o+(-0.423845465234)*x[2]
        arg[3,0,0,0]=(-0.498025341136)*x[0]**o+(-0.296354416633)*x[0]+(-0.600874889655)*x[1]**o+(0.544042300882)*x[1]+(0.318784851038)*x[2]**o+(-0.801109111437)*x[2]
        arg[3,0,0,1]=(-0.492361949745)*x[0]**o+(-0.708188092599)*x[0]+(-0.863737808435)*x[1]**o+(0.674889485553)*x[1]+(-0.982254652233)*x[2]**o+(0.682172862189)*x[2]
        arg[3,0,1,0]=(-0.872692397532)*x[0]**o+(0.563852836655)*x[0]+(-0.0219331486599)*x[1]**o+(-0.858175679559)*x[1]+(0.802432631538)*x[2]**o+(-0.655545287387)*x[2]
        arg[3,0,1,1]=(-0.538117166854)*x[0]**o+(-0.62507518352)*x[0]+(0.101524298816)*x[1]**o+(0.590432448393)*x[1]+(0.668057302006)*x[2]**o+(0.405382330056)*x[2]
        arg[3,0,2,0]=(-0.322916455064)*x[0]**o+(0.950624857669)*x[0]+(0.485623846818)*x[1]**o+(-0.12611522301)*x[1]+(0.503586393973)*x[2]**o+(0.109451719709)*x[2]
        arg[3,0,2,1]=(0.0289965995645)*x[0]**o+(0.0101440661359)*x[0]+(-0.303827119958)*x[1]**o+(0.355269935141)*x[1]+(0.29889427526)*x[2]**o+(-0.707999988781)*x[2]
        arg[3,1,0,0]=(-0.541382521848)*x[0]**o+(0.534108535206)*x[0]+(0.875610773119)*x[1]**o+(-0.575849439451)*x[1]+(-0.0836939606164)*x[2]**o+(0.433678856889)*x[2]
        arg[3,1,0,1]=(0.0322576144027)*x[0]**o+(0.794219359514)*x[0]+(0.629442621265)*x[1]**o+(0.0517726053499)*x[1]+(-0.776050101686)*x[2]**o+(0.697910196708)*x[2]
        arg[3,1,1,0]=(-0.198211682395)*x[0]**o+(-0.926268101941)*x[0]+(-0.904760383993)*x[1]**o+(-0.820075667018)*x[1]+(-0.101262044428)*x[2]**o+(-0.0621221772941)*x[2]
        arg[3,1,1,1]=(-0.531324749521)*x[0]**o+(0.736017207718)*x[0]+(0.487299696259)*x[1]**o+(-0.00604423021253)*x[1]+(-0.635722590088)*x[2]**o+(-0.143501730927)*x[2]
        arg[3,1,2,0]=(-0.731421637848)*x[0]**o+(-0.0371648783851)*x[0]+(-0.290362466891)*x[1]**o+(-0.375048584106)*x[1]+(-0.854622532092)*x[2]**o+(-0.794295905388)*x[2]
        arg[3,1,2,1]=(-0.766383422105)*x[0]**o+(0.149248159984)*x[0]+(0.970103674649)*x[1]**o+(-0.117794368107)*x[1]+(0.0178775633047)*x[2]**o+(-0.221126829278)*x[2]
        arg[3,2,0,0]=(0.706586887601)*x[0]**o+(0.162520023447)*x[0]+(-0.731539880336)*x[1]**o+(-0.992558519713)*x[1]+(0.426569303014)*x[2]**o+(0.0451783819487)*x[2]
        arg[3,2,0,1]=(0.0392295508272)*x[0]**o+(-0.157929005382)*x[0]+(0.0636418899878)*x[1]**o+(0.892501759578)*x[1]+(-0.368161001401)*x[2]**o+(-0.715520794543)*x[2]
        arg[3,2,1,0]=(0.889667707133)*x[0]**o+(0.308604168817)*x[0]+(-0.275558663983)*x[1]**o+(-0.408210095215)*x[1]+(-0.0918000869138)*x[2]**o+(0.142754064645)*x[2]
        arg[3,2,1,1]=(0.585831954407)*x[0]**o+(-0.705004375516)*x[0]+(0.657277099061)*x[1]**o+(-0.972401794257)*x[1]+(-0.658049310396)*x[2]**o+(-0.553961555706)*x[2]
        arg[3,2,2,0]=(0.88916769079)*x[0]**o+(-0.1401162684)*x[0]+(-0.396496633317)*x[1]**o+(0.215029522669)*x[1]+(-0.619023294208)*x[2]**o+(-0.187670706968)*x[2]
        arg[3,2,2,1]=(-0.262983340384)*x[0]**o+(-0.763236528477)*x[0]+(-0.339720289405)*x[1]**o+(-0.283077066751)*x[1]+(0.625600522278)*x[2]**o+(-0.132697432298)*x[2]
        arg[3,3,0,0]=(-0.346408105223)*x[0]**o+(0.299760534041)*x[0]+(0.567951405594)*x[1]**o+(-0.880683463459)*x[1]+(0.784127947661)*x[2]**o+(0.268895744072)*x[2]
        arg[3,3,0,1]=(0.064159345936)*x[0]**o+(-0.685098798117)*x[0]+(-0.182310885124)*x[1]**o+(-0.700081075257)*x[1]+(0.543959312437)*x[2]**o+(0.125104170434)*x[2]
        arg[3,3,1,0]=(0.460375837083)*x[0]**o+(0.82355800923)*x[0]+(-0.186000244328)*x[1]**o+(0.8418355439)*x[1]+(-0.375269292115)*x[2]**o+(0.712431840155)*x[2]
        arg[3,3,1,1]=(0.57260867805)*x[0]**o+(-0.146405851776)*x[0]+(-0.50640779246)*x[1]**o+(0.741760309299)*x[1]+(0.069752390409)*x[2]**o+(-0.376724191471)*x[2]
        arg[3,3,2,0]=(0.795493408403)*x[0]**o+(0.396861648344)*x[0]+(-0.728716068325)*x[1]**o+(0.739330072416)*x[1]+(0.202019645585)*x[2]**o+(0.841240509566)*x[2]
        arg[3,3,2,1]=(-0.967372744621)*x[0]**o+(-0.334044549662)*x[0]+(-0.0686982670779)*x[1]**o+(0.180083293629)*x[1]+(-0.208697793793)*x[2]**o+(0.0305446253751)*x[2]
        arg[3,4,0,0]=(0.94168350996)*x[0]**o+(0.0545532586137)*x[0]+(-0.300430532207)*x[1]**o+(-0.960362884765)*x[1]+(0.738261342527)*x[2]**o+(-0.351477574571)*x[2]
        arg[3,4,0,1]=(-0.525893857873)*x[0]**o+(0.227252001588)*x[0]+(0.179831911446)*x[1]**o+(0.350952712138)*x[1]+(-0.671550635312)*x[2]**o+(-0.338834217909)*x[2]
        arg[3,4,1,0]=(-0.60610056656)*x[0]**o+(0.219192673102)*x[0]+(-0.801595717702)*x[1]**o+(-0.696853420225)*x[1]+(-0.495201990047)*x[2]**o+(-0.500683649414)*x[2]
        arg[3,4,1,1]=(-0.37343020332)*x[0]**o+(-0.585536591527)*x[0]+(0.81794997494)*x[1]**o+(-0.240232527585)*x[1]+(-0.538970681636)*x[2]**o+(-0.146105210473)*x[2]
        arg[3,4,2,0]=(0.857322149358)*x[0]**o+(-0.552847286493)*x[0]+(0.530970378294)*x[1]**o+(0.0746811386922)*x[1]+(0.946170976628)*x[2]**o+(-0.329916962713)*x[2]
        arg[3,4,2,1]=(-0.353323100685)*x[0]**o+(0.808796594618)*x[0]+(0.481450588496)*x[1]**o+(0.1202874944)*x[1]+(0.671293736572)*x[2]**o+(-0.905074286629)*x[2]
        ref[0,0,0,0]=(-0.613904299283)*(1+2.*(dim-1.)/(o+1.))+(1.30218009705)*dim
        ref[0,0,0,1]=(-0.810525773084)*(1+2.*(dim-1.)/(o+1.))+(0.253224626501)*dim
        ref[0,0,1,0]=(0.509154413012)*(1+2.*(dim-1.)/(o+1.))+(-0.179784708198)*dim
        ref[0,0,1,1]=(0.637556597632)*(1+2.*(dim-1.)/(o+1.))+(-0.693313683923)*dim
        ref[0,0,2,0]=(-0.465137153331)*(1+2.*(dim-1.)/(o+1.))+(-0.0477989622018)*dim
        ref[0,0,2,1]=(-2.15185435439)*(1+2.*(dim-1.)/(o+1.))+(0.231459735932)*dim
        ref[0,1,0,0]=(0.8018922964)*(1+2.*(dim-1.)/(o+1.))+(-1.10884149089)*dim
        ref[0,1,0,1]=(1.7955198411)*(1+2.*(dim-1.)/(o+1.))+(-1.34662500856)*dim
        ref[0,1,1,0]=(0.584854782162)*(1+2.*(dim-1.)/(o+1.))+(0.274087751285)*dim
        ref[0,1,1,1]=(-1.0401305206)*(1+2.*(dim-1.)/(o+1.))+(1.85561593667)*dim
        ref[0,1,2,0]=(-0.834154383324)*(1+2.*(dim-1.)/(o+1.))+(-1.23273379028)*dim
        ref[0,1,2,1]=(-2.34017474507)*(1+2.*(dim-1.)/(o+1.))+(-0.224176787111)*dim
        ref[0,2,0,0]=(0.613915916089)*(1+2.*(dim-1.)/(o+1.))+(0.753829204471)*dim
        ref[0,2,0,1]=(0.131344565765)*(1+2.*(dim-1.)/(o+1.))+(-0.976652361934)*dim
        ref[0,2,1,0]=(0.579116187931)*(1+2.*(dim-1.)/(o+1.))+(0.0228071895801)*dim
        ref[0,2,1,1]=(0.772304265705)*(1+2.*(dim-1.)/(o+1.))+(0.470483133313)*dim
        ref[0,2,2,0]=(-1.12957943492)*(1+2.*(dim-1.)/(o+1.))+(-0.82429788127)*dim
        ref[0,2,2,1]=(-0.631840107715)*(1+2.*(dim-1.)/(o+1.))+(-0.732399256131)*dim
        ref[0,3,0,0]=(0.137323356087)*(1+2.*(dim-1.)/(o+1.))+(-0.413776023245)*dim
        ref[0,3,0,1]=(-0.173740822652)*(1+2.*(dim-1.)/(o+1.))+(1.19050855195)*dim
        ref[0,3,1,0]=(-1.58068224297)*(1+2.*(dim-1.)/(o+1.))+(0.24224399322)*dim
        ref[0,3,1,1]=(0.474472608494)*(1+2.*(dim-1.)/(o+1.))+(-0.557328900938)*dim
        ref[0,3,2,0]=(0.459842497633)*(1+2.*(dim-1.)/(o+1.))+(-1.09001004859)*dim
        ref[0,3,2,1]=(-0.22027884782)*(1+2.*(dim-1.)/(o+1.))+(0.235715331835)*dim
        ref[0,4,0,0]=(0.424803567504)*(1+2.*(dim-1.)/(o+1.))+(0.657196485077)*dim
        ref[0,4,0,1]=(-0.00655513622033)*(1+2.*(dim-1.)/(o+1.))+(-0.803228043207)*dim
        ref[0,4,1,0]=(1.33738998187)*(1+2.*(dim-1.)/(o+1.))+(-0.0257326162636)*dim
        ref[0,4,1,1]=(-1.55536089118)*(1+2.*(dim-1.)/(o+1.))+(-0.0357690415245)*dim
        ref[0,4,2,0]=(1.58679207216)*(1+2.*(dim-1.)/(o+1.))+(-0.693598479122)*dim
        ref[0,4,2,1]=(0.0628502197447)*(1+2.*(dim-1.)/(o+1.))+(-1.90742115249)*dim
        ref[1,0,0,0]=(-1.25774299848)*(1+2.*(dim-1.)/(o+1.))+(-0.0786696801461)*dim
        ref[1,0,0,1]=(-0.807603352133)*(1+2.*(dim-1.)/(o+1.))+(-0.429302209684)*dim
        ref[1,0,1,0]=(1.14217071628)*(1+2.*(dim-1.)/(o+1.))+(0.740396951119)*dim
        ref[1,0,1,1]=(1.89672043448)*(1+2.*(dim-1.)/(o+1.))+(1.4016996376)*dim
        ref[1,0,2,0]=(-0.768614149462)*(1+2.*(dim-1.)/(o+1.))+(0.913988010751)*dim
        ref[1,0,2,1]=(-0.601884288575)*(1+2.*(dim-1.)/(o+1.))+(-1.67541838174)*dim
        ref[1,1,0,0]=(-1.43981092176)*(1+2.*(dim-1.)/(o+1.))+(-0.506725297145)*dim
        ref[1,1,0,1]=(0.503113976474)*(1+2.*(dim-1.)/(o+1.))+(-0.488618815974)*dim
        ref[1,1,1,0]=(0.815915054472)*(1+2.*(dim-1.)/(o+1.))+(-0.522904536573)*dim
        ref[1,1,1,1]=(0.310174793362)*(1+2.*(dim-1.)/(o+1.))+(0.825570415579)*dim
        ref[1,1,2,0]=(-1.85870537865)*(1+2.*(dim-1.)/(o+1.))+(-0.896366119383)*dim
        ref[1,1,2,1]=(0.634330156916)*(1+2.*(dim-1.)/(o+1.))+(-0.202404975156)*dim
        ref[1,2,0,0]=(0.980180080393)*(1+2.*(dim-1.)/(o+1.))+(1.44073022185)*dim
        ref[1,2,0,1]=(1.16930139908)*(1+2.*(dim-1.)/(o+1.))+(-1.29606923835)*dim
        ref[1,2,1,0]=(0.251262820359)*(1+2.*(dim-1.)/(o+1.))+(-0.534365875316)*dim
        ref[1,2,1,1]=(1.45444347927)*(1+2.*(dim-1.)/(o+1.))+(1.09104332726)*dim
        ref[1,2,2,0]=(0.36826076644)*(1+2.*(dim-1.)/(o+1.))+(-0.222843807345)*dim
        ref[1,2,2,1]=(0.652062496151)*(1+2.*(dim-1.)/(o+1.))+(1.45032712876)*dim
        ref[1,3,0,0]=(0.357179082647)*(1+2.*(dim-1.)/(o+1.))+(0.96345428327)*dim
        ref[1,3,0,1]=(0.445836034597)*(1+2.*(dim-1.)/(o+1.))+(-0.768814370045)*dim
        ref[1,3,1,0]=(-0.633644413575)*(1+2.*(dim-1.)/(o+1.))+(0.996183349509)*dim
        ref[1,3,1,1]=(-1.47637436236)*(1+2.*(dim-1.)/(o+1.))+(-0.446000824477)*dim
        ref[1,3,2,0]=(2.22742580994)*(1+2.*(dim-1.)/(o+1.))+(0.238170411449)*dim
        ref[1,3,2,1]=(0.138824208277)*(1+2.*(dim-1.)/(o+1.))+(1.69436291732)*dim
        ref[1,4,0,0]=(-0.622328917175)*(1+2.*(dim-1.)/(o+1.))+(-1.12401903264)*dim
        ref[1,4,0,1]=(0.40305181353)*(1+2.*(dim-1.)/(o+1.))+(1.68001892165)*dim
        ref[1,4,1,0]=(-1.16479419403)*(1+2.*(dim-1.)/(o+1.))+(1.05838059836)*dim
        ref[1,4,1,1]=(0.0801376319834)*(1+2.*(dim-1.)/(o+1.))+(2.09228066438)*dim
        ref[1,4,2,0]=(-0.228762226067)*(1+2.*(dim-1.)/(o+1.))+(0.523454314851)*dim
        ref[1,4,2,1]=(-1.25310430418)*(1+2.*(dim-1.)/(o+1.))+(0.296827734822)*dim
        ref[2,0,0,0]=(1.14292760919)*(1+2.*(dim-1.)/(o+1.))+(-0.520826483003)*dim
        ref[2,0,0,1]=(1.16992671589)*(1+2.*(dim-1.)/(o+1.))+(-0.162899835641)*dim
        ref[2,0,1,0]=(-0.544683088018)*(1+2.*(dim-1.)/(o+1.))+(-1.44956627836)*dim
        ref[2,0,1,1]=(-1.46970715451)*(1+2.*(dim-1.)/(o+1.))+(-0.403035460127)*dim
        ref[2,0,2,0]=(-0.385627069865)*(1+2.*(dim-1.)/(o+1.))+(-0.829501029966)*dim
        ref[2,0,2,1]=(1.11664736171)*(1+2.*(dim-1.)/(o+1.))+(-1.27685980972)*dim
        ref[2,1,0,0]=(1.08669102227)*(1+2.*(dim-1.)/(o+1.))+(1.7990592379)*dim
        ref[2,1,0,1]=(0.0453423472143)*(1+2.*(dim-1.)/(o+1.))+(-0.68150455532)*dim
        ref[2,1,1,0]=(0.0322959644566)*(1+2.*(dim-1.)/(o+1.))+(-2.23247758093)*dim
        ref[2,1,1,1]=(-0.892578997912)*(1+2.*(dim-1.)/(o+1.))+(0.388356392315)*dim
        ref[2,1,2,0]=(-0.7179937061)*(1+2.*(dim-1.)/(o+1.))+(-1.52564228261)*dim
        ref[2,1,2,1]=(0.950438825754)*(1+2.*(dim-1.)/(o+1.))+(0.342228484532)*dim
        ref[2,2,0,0]=(1.0779882212)*(1+2.*(dim-1.)/(o+1.))+(0.796214693049)*dim
        ref[2,2,0,1]=(-0.115299804062)*(1+2.*(dim-1.)/(o+1.))+(0.798756112891)*dim
        ref[2,2,1,0]=(0.536406467403)*(1+2.*(dim-1.)/(o+1.))+(-1.139126385)*dim
        ref[2,2,1,1]=(-0.779767608394)*(1+2.*(dim-1.)/(o+1.))+(0.522828453067)*dim
        ref[2,2,2,0]=(-0.411536633652)*(1+2.*(dim-1.)/(o+1.))+(-0.895144122064)*dim
        ref[2,2,2,1]=(1.56944190392)*(1+2.*(dim-1.)/(o+1.))+(-0.12659143313)*dim
        ref[2,3,0,0]=(-0.372936173925)*(1+2.*(dim-1.)/(o+1.))+(-0.938385787171)*dim
        ref[2,3,0,1]=(1.33113027561)*(1+2.*(dim-1.)/(o+1.))+(1.39154904574)*dim
        ref[2,3,1,0]=(0.418458834548)*(1+2.*(dim-1.)/(o+1.))+(2.20921588131)*dim
        ref[2,3,1,1]=(0.64133847569)*(1+2.*(dim-1.)/(o+1.))+(0.556373927504)*dim
        ref[2,3,2,0]=(-0.689346506499)*(1+2.*(dim-1.)/(o+1.))+(1.63847691387)*dim
        ref[2,3,2,1]=(-1.23371702875)*(1+2.*(dim-1.)/(o+1.))+(-0.384869997042)*dim
        ref[2,4,0,0]=(0.433647132788)*(1+2.*(dim-1.)/(o+1.))+(-0.527688846368)*dim
        ref[2,4,0,1]=(1.03032814854)*(1+2.*(dim-1.)/(o+1.))+(-0.103322527004)*dim
        ref[2,4,1,0]=(0.663568613046)*(1+2.*(dim-1.)/(o+1.))+(1.1370080288)*dim
        ref[2,4,1,1]=(-0.977905093893)*(1+2.*(dim-1.)/(o+1.))+(0.0866635633103)*dim
        ref[2,4,2,0]=(0.085735636585)*(1+2.*(dim-1.)/(o+1.))+(-0.234684579572)*dim
        ref[2,4,2,1]=(-0.311909123795)*(1+2.*(dim-1.)/(o+1.))+(-0.705520095044)*dim
        ref[3,0,0,0]=(-0.780115379753)*(1+2.*(dim-1.)/(o+1.))+(-0.553421227188)*dim
        ref[3,0,0,1]=(-2.33835441041)*(1+2.*(dim-1.)/(o+1.))+(0.648874255143)*dim
        ref[3,0,1,0]=(-0.0921929146543)*(1+2.*(dim-1.)/(o+1.))+(-0.949868130291)*dim
        ref[3,0,1,1]=(0.231464433969)*(1+2.*(dim-1.)/(o+1.))+(0.370739594929)*dim
        ref[3,0,2,0]=(0.666293785727)*(1+2.*(dim-1.)/(o+1.))+(0.933961354368)*dim
        ref[3,0,2,1]=(0.0240637548661)*(1+2.*(dim-1.)/(o+1.))+(-0.342585987504)*dim
        ref[3,1,0,0]=(0.250534290654)*(1+2.*(dim-1.)/(o+1.))+(0.391937952644)*dim
        ref[3,1,0,1]=(-0.114349866018)*(1+2.*(dim-1.)/(o+1.))+(1.54390216157)*dim
        ref[3,1,1,0]=(-1.20423411082)*(1+2.*(dim-1.)/(o+1.))+(-1.80846594625)*dim
        ref[3,1,1,1]=(-0.679747643349)*(1+2.*(dim-1.)/(o+1.))+(0.586471246578)*dim
        ref[3,1,2,0]=(-1.87640663683)*(1+2.*(dim-1.)/(o+1.))+(-1.20650936788)*dim
        ref[3,1,2,1]=(0.221597815849)*(1+2.*(dim-1.)/(o+1.))+(-0.189673037401)*dim
        ref[3,2,0,0]=(0.40161631028)*(1+2.*(dim-1.)/(o+1.))+(-0.784860114318)*dim
        ref[3,2,0,1]=(-0.265289560586)*(1+2.*(dim-1.)/(o+1.))+(0.0190519596518)*dim
        ref[3,2,1,0]=(0.522308956237)*(1+2.*(dim-1.)/(o+1.))+(0.0431481382473)*dim
        ref[3,2,1,1]=(0.585059743072)*(1+2.*(dim-1.)/(o+1.))+(-2.23136772548)*dim
        ref[3,2,2,0]=(-0.126352236735)*(1+2.*(dim-1.)/(o+1.))+(-0.112757452698)*dim
        ref[3,2,2,1]=(0.0228968924883)*(1+2.*(dim-1.)/(o+1.))+(-1.17901102753)*dim
        ref[3,3,0,0]=(1.00567124803)*(1+2.*(dim-1.)/(o+1.))+(-0.312027185345)*dim
        ref[3,3,0,1]=(0.425807773249)*(1+2.*(dim-1.)/(o+1.))+(-1.26007570294)*dim
        ref[3,3,1,0]=(-0.10089369936)*(1+2.*(dim-1.)/(o+1.))+(2.37782539328)*dim
        ref[3,3,1,1]=(0.135953275999)*(1+2.*(dim-1.)/(o+1.))+(0.218630266051)*dim
        ref[3,3,2,0]=(0.268796985663)*(1+2.*(dim-1.)/(o+1.))+(1.97743223033)*dim
        ref[3,3,2,1]=(-1.24476880549)*(1+2.*(dim-1.)/(o+1.))+(-0.123416630658)*dim
        ref[3,4,0,0]=(1.37951432028)*(1+2.*(dim-1.)/(o+1.))+(-1.25728720072)*dim
        ref[3,4,0,1]=(-1.01761258174)*(1+2.*(dim-1.)/(o+1.))+(0.239370495817)*dim
        ref[3,4,1,0]=(-1.90289827431)*(1+2.*(dim-1.)/(o+1.))+(-0.978344396537)*dim
        ref[3,4,1,1]=(-0.0944509100161)*(1+2.*(dim-1.)/(o+1.))+(-0.971874329585)*dim
        ref[3,4,2,0]=(2.33446350428)*(1+2.*(dim-1.)/(o+1.))+(-0.808083110513)*dim
        ref[3,4,2,1]=(0.799421224383)*(1+2.*(dim-1.)/(o+1.))+(0.0240098023897)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(-0.0177156089276)*x[0]+(-1.07750293477)*x[1]
        ref=(0.674554765151)*(1+2.*(dim-1.)/(o+1.))+(-1.76977330884)*dim
      else:
        arg=(0.304688056778)*x[0]+(0.548485298428)*x[1]+(0.672370309114)*x[2]
        ref=(0.0121419382123)*(1+2.*(dim-1.)/(o+1.))+(1.51340172611)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(-1.81332918209)*x[0]+(0.825631690982)*x[1]
        arg[1]=(0.568335098118)*x[0]+(-0.504485431124)*x[1]
        ref[0]=(0.0338855214148)*(1+2.*(dim-1.)/(o+1.))+(-1.02158301252)*dim
        ref[1]=(0.348045520014)*(1+2.*(dim-1.)/(o+1.))+(-0.284195853019)*dim
      else:
        arg[0]=(-0.34247651789)*x[0]+(0.952434108232)*x[1]+(0.956872478767)*x[2]
        arg[1]=(-0.73212870207)*x[0]+(1.25659546958)*x[1]+(-0.690842998605)*x[2]
        ref[0]=(1.22184577749)*(1+2.*(dim-1.)/(o+1.))+(0.344984291617)*dim
        ref[1]=(-0.600249739257)*(1+2.*(dim-1.)/(o+1.))+(0.433873508162)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-1.78262197231)*x[0]+(1.45174573417)*x[1]
        arg[0,1]=(0.213586223199)*x[0]+(-0.917876059198)*x[1]
        arg[0,2]=(-1.12034819339)*x[0]+(0.59993130233)*x[1]
        arg[0,3]=(-0.156732957083)*x[0]+(0.744660873695)*x[1]
        arg[0,4]=(0.217327014577)*x[0]+(1.40157420042)*x[1]
        arg[1,0]=(0.0588042984116)*x[0]+(1.40376148561)*x[1]
        arg[1,1]=(-0.264973303233)*x[0]+(1.77692483003)*x[1]
        arg[1,2]=(-1.17910097102)*x[0]+(-1.20301960774)*x[1]
        arg[1,3]=(-0.511409983468)*x[0]+(-1.02065230228)*x[1]
        arg[1,4]=(1.1254497221)*x[0]+(-1.23163629376)*x[1]
        arg[2,0]=(-0.780207200735)*x[0]+(-1.49983791458)*x[1]
        arg[2,1]=(0.20679132359)*x[0]+(0.640203014983)*x[1]
        arg[2,2]=(0.198135205267)*x[0]+(0.30781411018)*x[1]
        arg[2,3]=(0.465906968942)*x[0]+(-0.812198212)*x[1]
        arg[2,4]=(0.190357350559)*x[0]+(-1.5894498455)*x[1]
        arg[3,0]=(-1.1362863086)*x[0]+(-0.0654142353086)*x[1]
        arg[3,1]=(0.0545924852084)*x[0]+(0.00262152155818)*x[1]
        arg[3,2]=(-0.0497600832907)*x[0]+(-0.609396009303)*x[1]
        arg[3,3]=(0.347834191962)*x[0]+(-0.768159599963)*x[1]
        arg[3,4]=(-0.178383166761)*x[0]+(-0.0690942212822)*x[1]
        ref[0,0]=(0.0436748214249)*(1+2.*(dim-1.)/(o+1.))+(-0.374551059559)*dim
        ref[0,1]=(-1.05854206043)*(1+2.*(dim-1.)/(o+1.))+(0.354252224435)*dim
        ref[0,2]=(-0.71782952995)*(1+2.*(dim-1.)/(o+1.))+(0.197412638885)*dim
        ref[0,3]=(-0.255252590976)*(1+2.*(dim-1.)/(o+1.))+(0.843180507587)*dim
        ref[0,4]=(0.177908820578)*(1+2.*(dim-1.)/(o+1.))+(1.44099239442)*dim
        ref[1,0]=(1.56040501086)*(1+2.*(dim-1.)/(o+1.))+(-0.0978392268304)*dim
        ref[1,1]=(1.33164571828)*(1+2.*(dim-1.)/(o+1.))+(0.180305808513)*dim
        ref[1,2]=(-0.885397553926)*(1+2.*(dim-1.)/(o+1.))+(-1.49672302484)*dim
        ref[1,3]=(0.105451964854)*(1+2.*(dim-1.)/(o+1.))+(-1.63751425061)*dim
        ref[1,4]=(-0.774237262923)*(1+2.*(dim-1.)/(o+1.))+(0.668050691269)*dim
        ref[2,0]=(-0.626493344552)*(1+2.*(dim-1.)/(o+1.))+(-1.65355177076)*dim
        ref[2,1]=(0.383826450148)*(1+2.*(dim-1.)/(o+1.))+(0.463167888425)*dim
        ref[2,2]=(-0.530241676713)*(1+2.*(dim-1.)/(o+1.))+(1.03619099216)*dim
        ref[2,3]=(-1.18245389658)*(1+2.*(dim-1.)/(o+1.))+(0.836162653525)*dim
        ref[2,4]=(-1.32396763993)*(1+2.*(dim-1.)/(o+1.))+(-0.0751248550082)*dim
        ref[3,0]=(0.15610314985)*(1+2.*(dim-1.)/(o+1.))+(-1.35780369376)*dim
        ref[3,1]=(-0.376931239611)*(1+2.*(dim-1.)/(o+1.))+(0.434145246378)*dim
        ref[3,2]=(0.0188741177717)*(1+2.*(dim-1.)/(o+1.))+(-0.678030210365)*dim
        ref[3,3]=(-0.264501990209)*(1+2.*(dim-1.)/(o+1.))+(-0.155823417792)*dim
        ref[3,4]=(-0.394609208758)*(1+2.*(dim-1.)/(o+1.))+(0.147131820715)*dim
      else:
        arg[0,0]=(-1.55600852596)*x[0]+(-0.624393737912)*x[1]+(-0.747032596279)*x[2]
        arg[0,1]=(0.158636782846)*x[0]+(-0.751760245066)*x[1]+(-1.62081735864)*x[2]
        arg[0,2]=(1.18083264152)*x[0]+(-1.16696814878)*x[1]+(-0.567354369023)*x[2]
        arg[0,3]=(0.176344525636)*x[0]+(-0.825319538089)*x[1]+(0.32760822446)*x[2]
        arg[0,4]=(0.570849136808)*x[0]+(-0.86277959325)*x[1]+(-0.253106033513)*x[2]
        arg[1,0]=(0.325475826081)*x[0]+(0.338988467638)*x[1]+(-0.596090654044)*x[2]
        arg[1,1]=(-1.29950586693)*x[0]+(-0.980591079821)*x[1]+(-0.7995507704)*x[2]
        arg[1,2]=(0.108948096835)*x[0]+(0.323001832704)*x[1]+(1.09678982473)*x[2]
        arg[1,3]=(-0.0307191696266)*x[0]+(0.134014984197)*x[1]+(-0.271363693134)*x[2]
        arg[1,4]=(0.0166805959434)*x[0]+(-0.323260440667)*x[1]+(-0.0168023639197)*x[2]
        arg[2,0]=(1.66243206086)*x[0]+(-0.478403313987)*x[1]+(1.45469230407)*x[2]
        arg[2,1]=(0.635927142861)*x[0]+(0.00628277747568)*x[1]+(-0.335502970321)*x[2]
        arg[2,2]=(-0.168440428449)*x[0]+(-0.302470381257)*x[1]+(1.26059253147)*x[2]
        arg[2,3]=(-0.0183286590129)*x[0]+(-0.0950506276918)*x[1]+(-0.721364110928)*x[2]
        arg[2,4]=(-1.04551179644)*x[0]+(0.20728433783)*x[1]+(-0.631521487025)*x[2]
        arg[3,0]=(1.0112442447)*x[0]+(0.0279472213885)*x[1]+(-1.24445628735)*x[2]
        arg[3,1]=(-1.36197290176)*x[0]+(0.145431495907)*x[1]+(0.518562214899)*x[2]
        arg[3,2]=(-0.249567074212)*x[0]+(-1.02991047199)*x[1]+(1.34627546588)*x[2]
        arg[3,3]=(1.72315708213)*x[0]+(0.00515532937983)*x[1]+(-0.479917704073)*x[2]
        arg[3,4]=(1.16297504595)*x[0]+(-0.176518851423)*x[1]+(0.184727222764)*x[2]
        ref[0,0]=(-2.09480681002)*(1+2.*(dim-1.)/(o+1.))+(-0.832628050126)*dim
        ref[0,1]=(-0.759565702286)*(1+2.*(dim-1.)/(o+1.))+(-1.45437511857)*dim
        ref[0,2]=(-1.08318812741)*(1+2.*(dim-1.)/(o+1.))+(0.529698251127)*dim
        ref[0,3]=(1.48988138495)*(1+2.*(dim-1.)/(o+1.))+(-1.81124817294)*dim
        ref[0,4]=(0.116706866039)*(1+2.*(dim-1.)/(o+1.))+(-0.661743355994)*dim
        ref[1,0]=(0.584176790043)*(1+2.*(dim-1.)/(o+1.))+(-0.515803150369)*dim
        ref[1,1]=(-1.25638916576)*(1+2.*(dim-1.)/(o+1.))+(-1.82325855139)*dim
        ref[1,2]=(-0.276318333951)*(1+2.*(dim-1.)/(o+1.))+(1.80505808822)*dim
        ref[1,3]=(-0.732277841606)*(1+2.*(dim-1.)/(o+1.))+(0.564209963043)*dim
        ref[1,4]=(-0.540285295646)*(1+2.*(dim-1.)/(o+1.))+(0.216903087004)*dim
        ref[2,0]=(2.23339929969)*(1+2.*(dim-1.)/(o+1.))+(0.405321751257)*dim
        ref[2,1]=(0.706137929075)*(1+2.*(dim-1.)/(o+1.))+(-0.39943097906)*dim
        ref[2,2]=(-0.474421381251)*(1+2.*(dim-1.)/(o+1.))+(1.26410310301)*dim
        ref[2,3]=(1.43870260444)*(1+2.*(dim-1.)/(o+1.))+(-2.27344600207)*dim
        ref[2,4]=(-1.00154864599)*(1+2.*(dim-1.)/(o+1.))+(-0.468200299646)*dim
        ref[3,0]=(-1.42181650401)*(1+2.*(dim-1.)/(o+1.))+(1.21655168275)*dim
        ref[3,1]=(-0.853191175508)*(1+2.*(dim-1.)/(o+1.))+(0.155211984554)*dim
        ref[3,2]=(-0.823496284217)*(1+2.*(dim-1.)/(o+1.))+(0.890294203901)*dim
        ref[3,3]=(1.54168027082)*(1+2.*(dim-1.)/(o+1.))+(-0.293285563378)*dim
        ref[3,4]=(1.66366656195)*(1+2.*(dim-1.)/(o+1.))+(-0.492483144667)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(-0.775903850503)*x[0]+(0.20657985447)*x[1]
        arg[0,0,1]=(1.42793412139)*x[0]+(1.11703130088)*x[1]
        arg[0,1,0]=(1.89750028353)*x[0]+(1.77903425203)*x[1]
        arg[0,1,1]=(-0.67280619036)*x[0]+(-0.283600511942)*x[1]
        arg[1,0,0]=(0.12590667282)*x[0]+(-0.109778270548)*x[1]
        arg[1,0,1]=(-1.06454156291)*x[0]+(-0.952754071628)*x[1]
        arg[1,1,0]=(1.27939438243)*x[0]+(1.41358757535)*x[1]
        arg[1,1,1]=(-1.68847666929)*x[0]+(-0.13467497005)*x[1]
        arg[2,0,0]=(0.116194220064)*x[0]+(0.457728827844)*x[1]
        arg[2,0,1]=(0.205487821412)*x[0]+(-0.657189976052)*x[1]
        arg[2,1,0]=(0.276192450801)*x[0]+(0.998191496643)*x[1]
        arg[2,1,1]=(1.63929218914)*x[0]+(0.268648283241)*x[1]
        arg[3,0,0]=(-0.419114071163)*x[0]+(-0.365697048391)*x[1]
        arg[3,0,1]=(0.802747458947)*x[0]+(0.137023645119)*x[1]
        arg[3,1,0]=(-0.0712635889142)*x[0]+(0.495489688269)*x[1]
        arg[3,1,1]=(-0.559073395964)*x[0]+(0.248364992354)*x[1]
        arg[4,0,0]=(0.372130696656)*x[0]+(1.22892579337)*x[1]
        arg[4,0,1]=(-1.47944537272)*x[0]+(1.46750125425)*x[1]
        arg[4,1,0]=(-0.0683410762004)*x[0]+(0.0982446782745)*x[1]
        arg[4,1,1]=(0.0698227670903)*x[0]+(0.227875508249)*x[1]
        arg[5,0,0]=(0.237110604157)*x[0]+(0.814767844631)*x[1]
        arg[5,0,1]=(0.172109909542)*x[0]+(-1.56055537357)*x[1]
        arg[5,1,0]=(0.92885449265)*x[0]+(0.820977987322)*x[1]
        arg[5,1,1]=(-1.76483093472)*x[0]+(0.526316351009)*x[1]
        ref[0,0,0]=(0.574633189339)*(1+2.*(dim-1.)/(o+1.))+(-1.14395718537)*dim
        ref[0,0,1]=(1.25832290545)*(1+2.*(dim-1.)/(o+1.))+(1.28664251682)*dim
        ref[0,1,0]=(1.75085914739)*(1+2.*(dim-1.)/(o+1.))+(1.92567538817)*dim
        ref[0,1,1]=(-1.57747201281)*(1+2.*(dim-1.)/(o+1.))+(0.621065310511)*dim
        ref[1,0,0]=(0.710883587246)*(1+2.*(dim-1.)/(o+1.))+(-0.694755184974)*dim
        ref[1,0,1]=(-0.870861363869)*(1+2.*(dim-1.)/(o+1.))+(-1.14643427067)*dim
        ref[1,1,0]=(1.21302125269)*(1+2.*(dim-1.)/(o+1.))+(1.4799607051)*dim
        ref[1,1,1]=(-0.163344093083)*(1+2.*(dim-1.)/(o+1.))+(-1.65980754626)*dim
        ref[2,0,0]=(1.18070929814)*(1+2.*(dim-1.)/(o+1.))+(-0.606786250235)*dim
        ref[2,0,1]=(-1.26011538751)*(1+2.*(dim-1.)/(o+1.))+(0.808413232875)*dim
        ref[2,1,0]=(-0.128253427149)*(1+2.*(dim-1.)/(o+1.))+(1.40263737459)*dim
        ref[2,1,1]=(1.22388877116)*(1+2.*(dim-1.)/(o+1.))+(0.68405170122)*dim
        ref[3,0,0]=(-1.51984063257)*(1+2.*(dim-1.)/(o+1.))+(0.735029513018)*dim
        ref[3,0,1]=(1.63702893126)*(1+2.*(dim-1.)/(o+1.))+(-0.697257827196)*dim
        ref[3,1,0]=(-0.274208683077)*(1+2.*(dim-1.)/(o+1.))+(0.698434782432)*dim
        ref[3,1,1]=(0.649113059071)*(1+2.*(dim-1.)/(o+1.))+(-0.95982146268)*dim
        ref[4,0,0]=(1.11723294318)*(1+2.*(dim-1.)/(o+1.))+(0.483823546848)*dim
        ref[4,0,1]=(-0.0369311601619)*(1+2.*(dim-1.)/(o+1.))+(0.0249870416905)*dim
        ref[4,1,0]=(-0.0969818481854)*(1+2.*(dim-1.)/(o+1.))+(0.126885450259)*dim
        ref[4,1,1]=(0.607834304349)*(1+2.*(dim-1.)/(o+1.))+(-0.31013602901)*dim
        ref[5,0,0]=(0.612356854718)*(1+2.*(dim-1.)/(o+1.))+(0.43952159407)*dim
        ref[5,0,1]=(-0.245849747535)*(1+2.*(dim-1.)/(o+1.))+(-1.1425957165)*dim
        ref[5,1,0]=(0.218025498167)*(1+2.*(dim-1.)/(o+1.))+(1.5318069818)*dim
        ref[5,1,1]=(-0.39419356418)*(1+2.*(dim-1.)/(o+1.))+(-0.844321019532)*dim
      else:
        arg[0,0,0]=(1.52622777835)*x[0]+(-0.627217363559)*x[1]+(-0.221339024656)*x[2]
        arg[0,0,1]=(-0.0735955925339)*x[0]+(-1.44492195287)*x[1]+(-0.0704081180966)*x[2]
        arg[0,1,0]=(-0.330700790074)*x[0]+(-1.04769579979)*x[1]+(0.508799976666)*x[2]
        arg[0,1,1]=(0.349349373439)*x[0]+(0.174054703574)*x[1]+(-0.129296220879)*x[2]
        arg[1,0,0]=(0.0683757878129)*x[0]+(1.33394387082)*x[1]+(0.892302980732)*x[2]
        arg[1,0,1]=(0.61014567223)*x[0]+(1.23846801628)*x[1]+(-0.990674941877)*x[2]
        arg[1,1,0]=(0.964575308424)*x[0]+(0.354822939168)*x[1]+(-0.466180761866)*x[2]
        arg[1,1,1]=(-1.79700594499)*x[0]+(-1.41821199521)*x[1]+(0.366764057361)*x[2]
        arg[2,0,0]=(0.670092254407)*x[0]+(-0.188523225253)*x[1]+(0.363046866484)*x[2]
        arg[2,0,1]=(0.420367065732)*x[0]+(0.798321474665)*x[1]+(-1.11347160928)*x[2]
        arg[2,1,0]=(0.54732858526)*x[0]+(0.249248458735)*x[1]+(1.33905485076)*x[2]
        arg[2,1,1]=(-1.45292600653)*x[0]+(-0.0838212918177)*x[1]+(-0.210961166185)*x[2]
        arg[3,0,0]=(-0.406792355772)*x[0]+(-0.230709788944)*x[1]+(1.12300383125)*x[2]
        arg[3,0,1]=(-1.02748413184)*x[0]+(-1.3840845484)*x[1]+(-0.181262479318)*x[2]
        arg[3,1,0]=(-1.39003872882)*x[0]+(-0.302577309791)*x[1]+(-0.627070288119)*x[2]
        arg[3,1,1]=(0.258496709277)*x[0]+(-1.32012543621)*x[1]+(0.921022534515)*x[2]
        arg[4,0,0]=(-0.639952053735)*x[0]+(0.172303718581)*x[1]+(-1.55894055416)*x[2]
        arg[4,0,1]=(-1.27660129414)*x[0]+(0.142177942574)*x[1]+(0.84156755899)*x[2]
        arg[4,1,0]=(0.49812465572)*x[0]+(1.53755810837)*x[1]+(0.10504146131)*x[2]
        arg[4,1,1]=(0.114534785643)*x[0]+(0.616103931999)*x[1]+(0.515764206037)*x[2]
        arg[5,0,0]=(0.22562941228)*x[0]+(0.70743104453)*x[1]+(0.882965136406)*x[2]
        arg[5,0,1]=(0.674003358504)*x[0]+(0.753008881236)*x[1]+(0.033083269963)*x[2]
        arg[5,1,0]=(-0.495809169314)*x[0]+(-1.09613938465)*x[1]+(-1.03297996938)*x[2]
        arg[5,1,1]=(-0.5222251005)*x[0]+(-0.56129023192)*x[1]+(0.529084741893)*x[2]
        ref[0,0,0]=(0.164333186185)*(1+2.*(dim-1.)/(o+1.))+(0.513338203948)*dim
        ref[0,0,1]=(-0.322072217228)*(1+2.*(dim-1.)/(o+1.))+(-1.26685344627)*dim
        ref[0,1,0]=(1.37061498302)*(1+2.*(dim-1.)/(o+1.))+(-2.24021159622)*dim
        ref[0,1,1]=(0.65996582358)*(1+2.*(dim-1.)/(o+1.))+(-0.265857967445)*dim
        ref[1,0,0]=(1.83616555678)*(1+2.*(dim-1.)/(o+1.))+(0.458457082585)*dim
        ref[1,0,1]=(1.01886896436)*(1+2.*(dim-1.)/(o+1.))+(-0.160930217723)*dim
        ref[1,1,0]=(0.95375222959)*(1+2.*(dim-1.)/(o+1.))+(-0.100534743865)*dim
        ref[1,1,1]=(-1.48059917075)*(1+2.*(dim-1.)/(o+1.))+(-1.3678547121)*dim
        ref[2,0,0]=(-0.0805152837328)*(1+2.*(dim-1.)/(o+1.))+(0.925131179371)*dim
        ref[2,0,1]=(0.155604473013)*(1+2.*(dim-1.)/(o+1.))+(-0.0503875419008)*dim
        ref[2,1,0]=(0.117237902139)*(1+2.*(dim-1.)/(o+1.))+(2.01839399262)*dim
        ref[2,1,1]=(0.196870433027)*(1+2.*(dim-1.)/(o+1.))+(-1.94457889755)*dim
        ref[3,0,0]=(-1.16272511547)*(1+2.*(dim-1.)/(o+1.))+(1.648226802)*dim
        ref[3,0,1]=(-1.54275766077)*(1+2.*(dim-1.)/(o+1.))+(-1.05007349878)*dim
        ref[3,1,0]=(-2.21939430114)*(1+2.*(dim-1.)/(o+1.))+(-0.100292025596)*dim
        ref[3,1,1]=(-0.974866109202)*(1+2.*(dim-1.)/(o+1.))+(0.834259916779)*dim
        ref[4,0,0]=(-1.81015652027)*(1+2.*(dim-1.)/(o+1.))+(-0.216432369042)*dim
        ref[4,0,1]=(-0.0692225921476)*(1+2.*(dim-1.)/(o+1.))+(-0.223633200428)*dim
        ref[4,1,0]=(0.907628019251)*(1+2.*(dim-1.)/(o+1.))+(1.23309620615)*dim
        ref[4,1,1]=(-0.77625708658)*(1+2.*(dim-1.)/(o+1.))+(2.02266001026)*dim
        ref[5,0,0]=(0.040336826806)*(1+2.*(dim-1.)/(o+1.))+(1.77568876641)*dim
        ref[5,0,1]=(-0.741065482842)*(1+2.*(dim-1.)/(o+1.))+(2.20116099255)*dim
        ref[5,1,0]=(-1.93385845474)*(1+2.*(dim-1.)/(o+1.))+(-0.691070068594)*dim
        ref[5,1,1]=(-0.539652361393)*(1+2.*(dim-1.)/(o+1.))+(-0.014778229134)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(0.139765044323)*x[0]+(0.763652540371)*x[1]
        arg[0,0,0,1]=(1.22422998499)*x[0]+(-0.704059494925)*x[1]
        arg[0,0,1,0]=(-0.0738568759745)*x[0]+(0.227297731345)*x[1]
        arg[0,0,1,1]=(-0.438016525512)*x[0]+(-0.00290173552923)*x[1]
        arg[0,0,2,0]=(-0.142466265385)*x[0]+(0.159086109387)*x[1]
        arg[0,0,2,1]=(-1.29207470861)*x[0]+(0.171024552932)*x[1]
        arg[0,1,0,0]=(0.228469040214)*x[0]+(-1.11207386711)*x[1]
        arg[0,1,0,1]=(-0.0716121324998)*x[0]+(-0.293214318059)*x[1]
        arg[0,1,1,0]=(-0.552331240231)*x[0]+(0.483311325934)*x[1]
        arg[0,1,1,1]=(0.759050036013)*x[0]+(-0.192218075397)*x[1]
        arg[0,1,2,0]=(0.736408127399)*x[0]+(0.61962966835)*x[1]
        arg[0,1,2,1]=(-0.170110592043)*x[0]+(1.12024324836)*x[1]
        arg[0,2,0,0]=(1.15358129137)*x[0]+(0.00288801991025)*x[1]
        arg[0,2,0,1]=(0.34101930461)*x[0]+(1.44561406668)*x[1]
        arg[0,2,1,0]=(1.20180064935)*x[0]+(1.53696717349)*x[1]
        arg[0,2,1,1]=(-0.955282659666)*x[0]+(-0.232022391367)*x[1]
        arg[0,2,2,0]=(0.0907361523468)*x[0]+(0.0735957910178)*x[1]
        arg[0,2,2,1]=(0.377295180443)*x[0]+(-0.694318356175)*x[1]
        arg[0,3,0,0]=(1.38936937129)*x[0]+(-1.11460156576)*x[1]
        arg[0,3,0,1]=(-0.477623155591)*x[0]+(-1.38984205639)*x[1]
        arg[0,3,1,0]=(0.0643534425446)*x[0]+(1.38744715898)*x[1]
        arg[0,3,1,1]=(0.0144448337619)*x[0]+(-0.107478295727)*x[1]
        arg[0,3,2,0]=(0.494522779873)*x[0]+(-0.623994602046)*x[1]
        arg[0,3,2,1]=(0.865352526765)*x[0]+(-0.0330471549519)*x[1]
        arg[0,4,0,0]=(0.876802965861)*x[0]+(-0.72991482831)*x[1]
        arg[0,4,0,1]=(-1.59217671369)*x[0]+(-1.17443275047)*x[1]
        arg[0,4,1,0]=(0.980556838858)*x[0]+(-0.134374268551)*x[1]
        arg[0,4,1,1]=(-0.799387621998)*x[0]+(0.290707689722)*x[1]
        arg[0,4,2,0]=(-0.152070506377)*x[0]+(-0.720051351365)*x[1]
        arg[0,4,2,1]=(-0.726386342076)*x[0]+(-0.821126213298)*x[1]
        arg[1,0,0,0]=(0.0108792321621)*x[0]+(1.08482229519)*x[1]
        arg[1,0,0,1]=(0.395066734758)*x[0]+(-0.147576996716)*x[1]
        arg[1,0,1,0]=(1.34674167646)*x[0]+(-1.7762597766)*x[1]
        arg[1,0,1,1]=(0.145955749421)*x[0]+(0.0758020791017)*x[1]
        arg[1,0,2,0]=(-1.3779896638)*x[0]+(0.879047617812)*x[1]
        arg[1,0,2,1]=(-0.034533186491)*x[0]+(-0.961616456719)*x[1]
        arg[1,1,0,0]=(1.93531133223)*x[0]+(0.085806770105)*x[1]
        arg[1,1,0,1]=(-1.40810156335)*x[0]+(0.647954800386)*x[1]
        arg[1,1,1,0]=(-0.898682130534)*x[0]+(0.0293908578939)*x[1]
        arg[1,1,1,1]=(0.906365050374)*x[0]+(0.493604087615)*x[1]
        arg[1,1,2,0]=(0.200697611638)*x[0]+(0.595460978794)*x[1]
        arg[1,1,2,1]=(-0.137201357175)*x[0]+(-0.456327010456)*x[1]
        arg[1,2,0,0]=(-0.435882659968)*x[0]+(-0.0312083604574)*x[1]
        arg[1,2,0,1]=(-0.678923893852)*x[0]+(-0.559827568731)*x[1]
        arg[1,2,1,0]=(-0.721204171338)*x[0]+(-0.054379212442)*x[1]
        arg[1,2,1,1]=(-0.0624007046506)*x[0]+(0.168316149134)*x[1]
        arg[1,2,2,0]=(-1.58839995565)*x[0]+(-0.212071943846)*x[1]
        arg[1,2,2,1]=(-1.12936118963)*x[0]+(1.37132899453)*x[1]
        arg[1,3,0,0]=(-0.382791970195)*x[0]+(-1.08331380956)*x[1]
        arg[1,3,0,1]=(-0.83075188923)*x[0]+(-0.771795988467)*x[1]
        arg[1,3,1,0]=(-0.491393878842)*x[0]+(0.256082770048)*x[1]
        arg[1,3,1,1]=(1.16535762668)*x[0]+(1.15868916612)*x[1]
        arg[1,3,2,0]=(0.930916100823)*x[0]+(0.447810335514)*x[1]
        arg[1,3,2,1]=(0.542008480891)*x[0]+(0.428660301354)*x[1]
        arg[1,4,0,0]=(0.820283847448)*x[0]+(-0.861049765761)*x[1]
        arg[1,4,0,1]=(0.794597952616)*x[0]+(0.484963133102)*x[1]
        arg[1,4,1,0]=(-0.0864183500936)*x[0]+(-0.48652009912)*x[1]
        arg[1,4,1,1]=(-0.729261074705)*x[0]+(0.572520019661)*x[1]
        arg[1,4,2,0]=(0.874563670592)*x[0]+(-1.03978413308)*x[1]
        arg[1,4,2,1]=(-0.119980177921)*x[0]+(-0.416981275935)*x[1]
        arg[2,0,0,0]=(0.00851163981688)*x[0]+(0.0537719250367)*x[1]
        arg[2,0,0,1]=(0.325152071307)*x[0]+(0.852063357275)*x[1]
        arg[2,0,1,0]=(0.3798978673)*x[0]+(0.103107604149)*x[1]
        arg[2,0,1,1]=(0.241778324259)*x[0]+(0.840767359849)*x[1]
        arg[2,0,2,0]=(0.504220642882)*x[0]+(-0.296421218478)*x[1]
        arg[2,0,2,1]=(1.21850888213)*x[0]+(0.361422173343)*x[1]
        arg[2,1,0,0]=(1.23948863042)*x[0]+(-0.454591957846)*x[1]
        arg[2,1,0,1]=(-0.834341417745)*x[0]+(0.836809823188)*x[1]
        arg[2,1,1,0]=(0.305283593888)*x[0]+(-0.443709359754)*x[1]
        arg[2,1,1,1]=(-1.73751494542)*x[0]+(0.28249619071)*x[1]
        arg[2,1,2,0]=(1.64628846378)*x[0]+(0.929422692022)*x[1]
        arg[2,1,2,1]=(1.55964953056)*x[0]+(0.294507226966)*x[1]
        arg[2,2,0,0]=(-0.122178600734)*x[0]+(-0.836106215351)*x[1]
        arg[2,2,0,1]=(0.0560920262009)*x[0]+(-0.959524453552)*x[1]
        arg[2,2,1,0]=(-0.0235093950509)*x[0]+(1.00069942275)*x[1]
        arg[2,2,1,1]=(-1.6685983361)*x[0]+(-0.283415608089)*x[1]
        arg[2,2,2,0]=(1.88757268407)*x[0]+(-0.0970308753337)*x[1]
        arg[2,2,2,1]=(-0.620474974074)*x[0]+(0.24769276713)*x[1]
        arg[2,3,0,0]=(1.01637595389)*x[0]+(0.705160942822)*x[1]
        arg[2,3,0,1]=(1.14707395208)*x[0]+(0.705275592532)*x[1]
        arg[2,3,1,0]=(-0.758160952494)*x[0]+(0.305309048122)*x[1]
        arg[2,3,1,1]=(0.772402918026)*x[0]+(-0.052854508055)*x[1]
        arg[2,3,2,0]=(-0.154978687298)*x[0]+(0.418368826949)*x[1]
        arg[2,3,2,1]=(-0.845976442257)*x[0]+(0.20623978887)*x[1]
        arg[2,4,0,0]=(-0.466185710519)*x[0]+(-0.233766119139)*x[1]
        arg[2,4,0,1]=(-0.922877220469)*x[0]+(-0.688487528304)*x[1]
        arg[2,4,1,0]=(0.95836287786)*x[0]+(0.939609093639)*x[1]
        arg[2,4,1,1]=(-0.11982304155)*x[0]+(-0.257073186684)*x[1]
        arg[2,4,2,0]=(0.345817495959)*x[0]+(0.151979044778)*x[1]
        arg[2,4,2,1]=(-1.05379548205)*x[0]+(-0.436065928744)*x[1]
        arg[3,0,0,0]=(0.744945706398)*x[0]+(-0.649458468468)*x[1]
        arg[3,0,0,1]=(-1.04958965249)*x[0]+(0.639942303499)*x[1]
        arg[3,0,1,0]=(0.337716118074)*x[0]+(1.93458666396)*x[1]
        arg[3,0,1,1]=(0.303244531116)*x[0]+(0.436648998724)*x[1]
        arg[3,0,2,0]=(-1.55484904142)*x[0]+(0.0691158242075)*x[1]
        arg[3,0,2,1]=(-0.184571563941)*x[0]+(-0.617496186026)*x[1]
        arg[3,1,0,0]=(0.54665472418)*x[0]+(-0.685197219861)*x[1]
        arg[3,1,0,1]=(0.270341054659)*x[0]+(0.0123904138595)*x[1]
        arg[3,1,1,0]=(-0.551927232745)*x[0]+(-0.0835992799883)*x[1]
        arg[3,1,1,1]=(-0.731741032186)*x[0]+(0.351597478907)*x[1]
        arg[3,1,2,0]=(-0.678408784965)*x[0]+(-0.0779391332051)*x[1]
        arg[3,1,2,1]=(-0.873182924195)*x[0]+(-0.641772673128)*x[1]
        arg[3,2,0,0]=(-0.826223714873)*x[0]+(0.631844849849)*x[1]
        arg[3,2,0,1]=(-0.899714515249)*x[0]+(-0.010350665009)*x[1]
        arg[3,2,1,0]=(0.290678744119)*x[0]+(-0.5871053663)*x[1]
        arg[3,2,1,1]=(0.208288673139)*x[0]+(0.631917010925)*x[1]
        arg[3,2,2,0]=(-0.262706136542)*x[0]+(0.0853212850096)*x[1]
        arg[3,2,2,1]=(-1.28934754486)*x[0]+(0.686855142119)*x[1]
        arg[3,3,0,0]=(-1.36704032596)*x[0]+(-1.55543480824)*x[1]
        arg[3,3,0,1]=(-1.17014698336)*x[0]+(-0.652446157654)*x[1]
        arg[3,3,1,0]=(-0.547499757476)*x[0]+(-0.156519515585)*x[1]
        arg[3,3,1,1]=(0.228140026291)*x[0]+(-1.41014817371)*x[1]
        arg[3,3,2,0]=(-0.055056583329)*x[0]+(-0.412525603966)*x[1]
        arg[3,3,2,1]=(-0.524790489086)*x[0]+(-0.774411367008)*x[1]
        arg[3,4,0,0]=(-1.33491047191)*x[0]+(1.47855022808)*x[1]
        arg[3,4,0,1]=(-0.301405501225)*x[0]+(0.687178741851)*x[1]
        arg[3,4,1,0]=(0.213050803289)*x[0]+(0.428086322291)*x[1]
        arg[3,4,1,1]=(-1.67469009469)*x[0]+(0.857275314994)*x[1]
        arg[3,4,2,0]=(0.748812148687)*x[0]+(-0.0339722859922)*x[1]
        arg[3,4,2,1]=(-0.829454912563)*x[0]+(-0.637161668052)*x[1]
        ref[0,0,0,0]=(0.538138880568)*(1+2.*(dim-1.)/(o+1.))+(0.365278704126)*dim
        ref[0,0,0,1]=(0.532354921417)*(1+2.*(dim-1.)/(o+1.))+(-0.0121844313566)*dim
        ref[0,0,1,0]=(1.27037934234)*(1+2.*(dim-1.)/(o+1.))+(-1.11693848697)*dim
        ref[0,0,1,1]=(-0.317494874699)*(1+2.*(dim-1.)/(o+1.))+(-0.123423386343)*dim
        ref[0,0,2,0]=(-0.274542501137)*(1+2.*(dim-1.)/(o+1.))+(0.291162345138)*dim
        ref[0,0,2,1]=(-0.669231324784)*(1+2.*(dim-1.)/(o+1.))+(-0.451818830894)*dim
        ref[0,1,0,0]=(-1.55791021679)*(1+2.*(dim-1.)/(o+1.))+(0.674305389899)*dim
        ref[0,1,0,1]=(-1.04225461385)*(1+2.*(dim-1.)/(o+1.))+(0.67742816329)*dim
        ref[0,1,1,0]=(-0.553787932867)*(1+2.*(dim-1.)/(o+1.))+(0.48476801857)*dim
        ref[0,1,1,1]=(0.774284436582)*(1+2.*(dim-1.)/(o+1.))+(-0.207452475966)*dim
        ref[0,1,2,0]=(0.754663298028)*(1+2.*(dim-1.)/(o+1.))+(0.601374497721)*dim
        ref[0,1,2,1]=(0.378489135155)*(1+2.*(dim-1.)/(o+1.))+(0.571643521166)*dim
        ref[0,2,0,0]=(0.84270626402)*(1+2.*(dim-1.)/(o+1.))+(0.313763047258)*dim
        ref[0,2,0,1]=(0.741388749098)*(1+2.*(dim-1.)/(o+1.))+(1.0452446222)*dim
        ref[0,2,1,0]=(1.31801677967)*(1+2.*(dim-1.)/(o+1.))+(1.42075104317)*dim
        ref[0,2,1,1]=(0.342128499103)*(1+2.*(dim-1.)/(o+1.))+(-1.52943355014)*dim
        ref[0,2,2,0]=(-0.2097885639)*(1+2.*(dim-1.)/(o+1.))+(0.374120507264)*dim
        ref[0,2,2,1]=(0.517181369396)*(1+2.*(dim-1.)/(o+1.))+(-0.834204545127)*dim
        ref[0,3,0,0]=(0.546538945104)*(1+2.*(dim-1.)/(o+1.))+(-0.271771139576)*dim
        ref[0,3,0,1]=(-1.11041658618)*(1+2.*(dim-1.)/(o+1.))+(-0.757048625806)*dim
        ref[0,3,1,0]=(1.71248931933)*(1+2.*(dim-1.)/(o+1.))+(-0.260688717804)*dim
        ref[0,3,1,1]=(0.426783834682)*(1+2.*(dim-1.)/(o+1.))+(-0.519817296646)*dim
        ref[0,3,2,0]=(-0.159038051631)*(1+2.*(dim-1.)/(o+1.))+(0.0295662294573)*dim
        ref[0,3,2,1]=(-0.0935533877409)*(1+2.*(dim-1.)/(o+1.))+(0.925858759554)*dim
        ref[0,4,0,0]=(0.254413611162)*(1+2.*(dim-1.)/(o+1.))+(-0.107525473611)*dim
        ref[0,4,0,1]=(-1.80344635667)*(1+2.*(dim-1.)/(o+1.))+(-0.963163107492)*dim
        ref[0,4,1,0]=(0.77936654687)*(1+2.*(dim-1.)/(o+1.))+(0.0668160234379)*dim
        ref[0,4,1,1]=(-0.854656189741)*(1+2.*(dim-1.)/(o+1.))+(0.345976257465)*dim
        ref[0,4,2,0]=(0.236075361826)*(1+2.*(dim-1.)/(o+1.))+(-1.10819721957)*dim
        ref[0,4,2,1]=(-1.41443045549)*(1+2.*(dim-1.)/(o+1.))+(-0.133082099885)*dim
        ref[1,0,0,0]=(0.254601110632)*(1+2.*(dim-1.)/(o+1.))+(0.841100416717)*dim
        ref[1,0,0,1]=(-1.05142234949)*(1+2.*(dim-1.)/(o+1.))+(1.29891208753)*dim
        ref[1,0,1,0]=(-0.476182365596)*(1+2.*(dim-1.)/(o+1.))+(0.0466642654563)*dim
        ref[1,0,1,1]=(-1.2117362156)*(1+2.*(dim-1.)/(o+1.))+(1.43349404413)*dim
        ref[1,0,2,0]=(-0.82694411065)*(1+2.*(dim-1.)/(o+1.))+(0.328002064662)*dim
        ref[1,0,2,1]=(-0.784489634199)*(1+2.*(dim-1.)/(o+1.))+(-0.211660009011)*dim
        ref[1,1,0,0]=(1.67906802629)*(1+2.*(dim-1.)/(o+1.))+(0.342050076043)*dim
        ref[1,1,0,1]=(-0.798764638314)*(1+2.*(dim-1.)/(o+1.))+(0.03861787535)*dim
        ref[1,1,1,0]=(0.348279136631)*(1+2.*(dim-1.)/(o+1.))+(-1.21757040927)*dim
        ref[1,1,1,1]=(0.452104104709)*(1+2.*(dim-1.)/(o+1.))+(0.94786503328)*dim
        ref[1,1,2,0]=(0.855057493837)*(1+2.*(dim-1.)/(o+1.))+(-0.0588989034046)*dim
        ref[1,1,2,1]=(-0.666495642664)*(1+2.*(dim-1.)/(o+1.))+(0.0729672750328)*dim
        ref[1,2,0,0]=(-0.566691483673)*(1+2.*(dim-1.)/(o+1.))+(0.0996004632473)*dim
        ref[1,2,0,1]=(-0.127941886194)*(1+2.*(dim-1.)/(o+1.))+(-1.11080957639)*dim
        ref[1,2,1,0]=(-0.249856548379)*(1+2.*(dim-1.)/(o+1.))+(-0.525726835401)*dim
        ref[1,2,1,1]=(1.00195986166)*(1+2.*(dim-1.)/(o+1.))+(-0.896044417179)*dim
        ref[1,2,2,0]=(-0.96849834562)*(1+2.*(dim-1.)/(o+1.))+(-0.831973553872)*dim
        ref[1,2,2,1]=(0.204504064885)*(1+2.*(dim-1.)/(o+1.))+(0.0374637400154)*dim
        ref[1,3,0,0]=(-0.587139938504)*(1+2.*(dim-1.)/(o+1.))+(-0.878965841253)*dim
        ref[1,3,0,1]=(-0.013960550118)*(1+2.*(dim-1.)/(o+1.))+(-1.58858732758)*dim
        ref[1,3,1,0]=(-0.626445648631)*(1+2.*(dim-1.)/(o+1.))+(0.391134539838)*dim
        ref[1,3,1,1]=(0.793357894316)*(1+2.*(dim-1.)/(o+1.))+(1.53068889848)*dim
        ref[1,3,2,0]=(-0.30795291971)*(1+2.*(dim-1.)/(o+1.))+(1.68667935605)*dim
        ref[1,3,2,1]=(0.865506632636)*(1+2.*(dim-1.)/(o+1.))+(0.105162149608)*dim
        ref[1,4,0,0]=(0.0142921032519)*(1+2.*(dim-1.)/(o+1.))+(-0.0550580215643)*dim
        ref[1,4,0,1]=(1.74635688975)*(1+2.*(dim-1.)/(o+1.))+(-0.466795804028)*dim
        ref[1,4,1,0]=(0.202247079284)*(1+2.*(dim-1.)/(o+1.))+(-0.775185528498)*dim
        ref[1,4,1,1]=(0.518302547238)*(1+2.*(dim-1.)/(o+1.))+(-0.675043602282)*dim
        ref[1,4,2,0]=(-0.0110996487494)*(1+2.*(dim-1.)/(o+1.))+(-0.154120813742)*dim
        ref[1,4,2,1]=(0.445368223328)*(1+2.*(dim-1.)/(o+1.))+(-0.982329677184)*dim
        ref[2,0,0,0]=(1.11858205689)*(1+2.*(dim-1.)/(o+1.))+(-1.05629849203)*dim
        ref[2,0,0,1]=(0.326749663749)*(1+2.*(dim-1.)/(o+1.))+(0.850465764833)*dim
        ref[2,0,1,0]=(1.17260350442)*(1+2.*(dim-1.)/(o+1.))+(-0.689598032967)*dim
        ref[2,0,1,1]=(1.26453211291)*(1+2.*(dim-1.)/(o+1.))+(-0.181986428807)*dim
        ref[2,0,2,0]=(0.28543047736)*(1+2.*(dim-1.)/(o+1.))+(-0.0776310529561)*dim
        ref[2,0,2,1]=(1.67283723918)*(1+2.*(dim-1.)/(o+1.))+(-0.092906183711)*dim
        ref[2,1,0,0]=(0.375606493975)*(1+2.*(dim-1.)/(o+1.))+(0.409290178597)*dim
        ref[2,1,0,1]=(0.391717224425)*(1+2.*(dim-1.)/(o+1.))+(-0.389248818981)*dim
        ref[2,1,1,0]=(0.540376031372)*(1+2.*(dim-1.)/(o+1.))+(-0.678801797238)*dim
        ref[2,1,1,1]=(-0.781880488997)*(1+2.*(dim-1.)/(o+1.))+(-0.673138265714)*dim
        ref[2,1,2,0]=(1.19474704445)*(1+2.*(dim-1.)/(o+1.))+(1.38096411135)*dim
        ref[2,1,2,1]=(0.3333062674)*(1+2.*(dim-1.)/(o+1.))+(1.52085049012)*dim
        ref[2,2,0,0]=(-0.570564746042)*(1+2.*(dim-1.)/(o+1.))+(-0.387720070044)*dim
        ref[2,2,0,1]=(-0.906439974412)*(1+2.*(dim-1.)/(o+1.))+(0.00300754706095)*dim
        ref[2,2,1,0]=(1.35643048678)*(1+2.*(dim-1.)/(o+1.))+(-0.379240459088)*dim
        ref[2,2,1,1]=(-1.68522696847)*(1+2.*(dim-1.)/(o+1.))+(-0.266786975721)*dim
        ref[2,2,2,0]=(1.64773077393)*(1+2.*(dim-1.)/(o+1.))+(0.14281103481)*dim
        ref[2,2,2,1]=(0.161577377439)*(1+2.*(dim-1.)/(o+1.))+(-0.534359584383)*dim
        ref[2,3,0,0]=(0.117577368411)*(1+2.*(dim-1.)/(o+1.))+(1.6039595283)*dim
        ref[2,3,0,1]=(1.07270794583)*(1+2.*(dim-1.)/(o+1.))+(0.779641598778)*dim
        ref[2,3,1,0]=(0.940083930247)*(1+2.*(dim-1.)/(o+1.))+(-1.39293583462)*dim
        ref[2,3,1,1]=(-0.322170646263)*(1+2.*(dim-1.)/(o+1.))+(1.04171905623)*dim
        ref[2,3,2,0]=(-1.00498391043)*(1+2.*(dim-1.)/(o+1.))+(1.26837405008)*dim
        ref[2,3,2,1]=(0.148227404921)*(1+2.*(dim-1.)/(o+1.))+(-0.787964058307)*dim
        ref[2,4,0,0]=(-0.567878073929)*(1+2.*(dim-1.)/(o+1.))+(-0.132073755729)*dim
        ref[2,4,0,1]=(-0.265369521409)*(1+2.*(dim-1.)/(o+1.))+(-1.34599522736)*dim
        ref[2,4,1,0]=(0.87362235129)*(1+2.*(dim-1.)/(o+1.))+(1.02434962021)*dim
        ref[2,4,1,1]=(-1.05435356817)*(1+2.*(dim-1.)/(o+1.))+(0.67745733994)*dim
        ref[2,4,2,0]=(-0.0889174414067)*(1+2.*(dim-1.)/(o+1.))+(0.586713982143)*dim
        ref[2,4,2,1]=(-0.950446212097)*(1+2.*(dim-1.)/(o+1.))+(-0.539415198695)*dim
        ref[3,0,0,0]=(0.585351095733)*(1+2.*(dim-1.)/(o+1.))+(-0.489863857802)*dim
        ref[3,0,0,1]=(0.0818488289135)*(1+2.*(dim-1.)/(o+1.))+(-0.491496177908)*dim
        ref[3,0,1,0]=(1.83790469827)*(1+2.*(dim-1.)/(o+1.))+(0.43439808376)*dim
        ref[3,0,1,1]=(-0.812249042145)*(1+2.*(dim-1.)/(o+1.))+(1.55214257198)*dim
        ref[3,0,2,0]=(-1.62498605972)*(1+2.*(dim-1.)/(o+1.))+(0.139252842511)*dim
        ref[3,0,2,1]=(0.248660359392)*(1+2.*(dim-1.)/(o+1.))+(-1.05072810936)*dim
        ref[3,1,0,0]=(0.701570774477)*(1+2.*(dim-1.)/(o+1.))+(-0.840113270158)*dim
        ref[3,1,0,1]=(-1.01112314775)*(1+2.*(dim-1.)/(o+1.))+(1.29385461627)*dim
        ref[3,1,1,0]=(-0.38604780773)*(1+2.*(dim-1.)/(o+1.))+(-0.249478705003)*dim
        ref[3,1,1,1]=(-0.505728989541)*(1+2.*(dim-1.)/(o+1.))+(0.125585436262)*dim
        ref[3,1,2,0]=(0.170090820399)*(1+2.*(dim-1.)/(o+1.))+(-0.926438738569)*dim
        ref[3,1,2,1]=(-0.929125306455)*(1+2.*(dim-1.)/(o+1.))+(-0.585830290868)*dim
        ref[3,2,0,0]=(-0.155713736332)*(1+2.*(dim-1.)/(o+1.))+(-0.038665128691)*dim
        ref[3,2,0,1]=(-1.52172024486)*(1+2.*(dim-1.)/(o+1.))+(0.6116550646)*dim
        ref[3,2,1,0]=(0.320960641588)*(1+2.*(dim-1.)/(o+1.))+(-0.617387263769)*dim
        ref[3,2,1,1]=(1.35000898226)*(1+2.*(dim-1.)/(o+1.))+(-0.509803298192)*dim
        ref[3,2,2,0]=(-1.08694938974)*(1+2.*(dim-1.)/(o+1.))+(0.909564538204)*dim
        ref[3,2,2,1]=(-0.484173893901)*(1+2.*(dim-1.)/(o+1.))+(-0.11831850884)*dim
        ref[3,3,0,0]=(-1.30137401361)*(1+2.*(dim-1.)/(o+1.))+(-1.62110112059)*dim
        ref[3,3,0,1]=(-1.01955795295)*(1+2.*(dim-1.)/(o+1.))+(-0.803035188058)*dim
        ref[3,3,1,0]=(-0.86522946065)*(1+2.*(dim-1.)/(o+1.))+(0.161210187589)*dim
        ref[3,3,1,1]=(-1.45434545656)*(1+2.*(dim-1.)/(o+1.))+(0.272337309144)*dim
        ref[3,3,2,0]=(-1.09044516001)*(1+2.*(dim-1.)/(o+1.))+(0.622862972717)*dim
        ref[3,3,2,1]=(-1.56164744337)*(1+2.*(dim-1.)/(o+1.))+(0.262445587275)*dim
        ref[3,4,0,0]=(-0.0717688395634)*(1+2.*(dim-1.)/(o+1.))+(0.215408595737)*dim
        ref[3,4,0,1]=(-0.361057910751)*(1+2.*(dim-1.)/(o+1.))+(0.746831151378)*dim
        ref[3,4,1,0]=(1.3065764036)*(1+2.*(dim-1.)/(o+1.))+(-0.665439278018)*dim
        ref[3,4,1,1]=(0.00897444539195)*(1+2.*(dim-1.)/(o+1.))+(-0.826389225091)*dim
        ref[3,4,2,0]=(0.0968985993748)*(1+2.*(dim-1.)/(o+1.))+(0.61794126332)*dim
        ref[3,4,2,1]=(-0.247002319776)*(1+2.*(dim-1.)/(o+1.))+(-1.21961426084)*dim
      else:
        arg[0,0,0,0]=(0.249559417601)*x[0]+(-1.12919415836)*x[1]+(0.535083953587)*x[2]
        arg[0,0,0,1]=(-1.61154525877)*x[0]+(-0.339535503099)*x[1]+(-0.3734414422)*x[2]
        arg[0,0,1,0]=(-1.33426413331)*x[0]+(0.503654661149)*x[1]+(0.900718166029)*x[2]
        arg[0,0,1,1]=(-1.47348135079)*x[0]+(0.294426872223)*x[1]+(-0.092932038156)*x[2]
        arg[0,0,2,0]=(-0.659338105033)*x[0]+(0.276825152205)*x[1]+(-0.825904690993)*x[2]
        arg[0,0,2,1]=(-0.871512897535)*x[0]+(0.16237107696)*x[1]+(-1.37935248231)*x[2]
        arg[0,1,0,0]=(-0.83315792018)*x[0]+(-0.00675809409276)*x[1]+(0.464774846632)*x[2]
        arg[0,1,0,1]=(0.527155362932)*x[0]+(0.856707619808)*x[1]+(-1.68267649456)*x[2]
        arg[0,1,1,0]=(-0.524091384241)*x[0]+(-0.540620074365)*x[1]+(0.992745803077)*x[2]
        arg[0,1,1,1]=(1.61356254869)*x[0]+(1.50422956091)*x[1]+(0.522074610423)*x[2]
        arg[0,1,2,0]=(-0.786383950132)*x[0]+(0.240998133454)*x[1]+(-0.609039834527)*x[2]
        arg[0,1,2,1]=(-0.549608404174)*x[0]+(-0.850827315095)*x[1]+(0.267519458624)*x[2]
        arg[0,2,0,0]=(-0.171255468598)*x[0]+(-0.315808418391)*x[1]+(-1.31408126593)*x[2]
        arg[0,2,0,1]=(-0.0537048497397)*x[0]+(-1.05283989597)*x[1]+(-0.230505226375)*x[2]
        arg[0,2,1,0]=(0.175422426556)*x[0]+(-1.37203245283)*x[1]+(-1.31579047196)*x[2]
        arg[0,2,1,1]=(-0.678719594441)*x[0]+(0.325375648943)*x[1]+(-1.3715383088)*x[2]
        arg[0,2,2,0]=(1.34295310191)*x[0]+(-1.29856762722)*x[1]+(-0.604804069008)*x[2]
        arg[0,2,2,1]=(-0.38925176646)*x[0]+(-0.391078913511)*x[1]+(-0.147119806342)*x[2]
        arg[0,3,0,0]=(0.55897935716)*x[0]+(-0.938161813577)*x[1]+(-1.01170144437)*x[2]
        arg[0,3,0,1]=(-0.214720994691)*x[0]+(-0.936636326408)*x[1]+(1.23434495942)*x[2]
        arg[0,3,1,0]=(0.665136478224)*x[0]+(-0.223132283946)*x[1]+(0.513937660496)*x[2]
        arg[0,3,1,1]=(0.99663942166)*x[0]+(-0.870793744717)*x[1]+(-0.44642261239)*x[2]
        arg[0,3,2,0]=(0.426708530385)*x[0]+(-0.399913590631)*x[1]+(0.621569032554)*x[2]
        arg[0,3,2,1]=(-0.385564378699)*x[0]+(-1.17310910463)*x[1]+(-0.386725711607)*x[2]
        arg[0,4,0,0]=(-0.907605366918)*x[0]+(-1.15620807313)*x[1]+(0.690709991448)*x[2]
        arg[0,4,0,1]=(0.68925487184)*x[0]+(-0.731402302062)*x[1]+(-0.278506722879)*x[2]
        arg[0,4,1,0]=(0.607545316973)*x[0]+(-0.480514687124)*x[1]+(-0.261850296368)*x[2]
        arg[0,4,1,1]=(1.39016358303)*x[0]+(0.571194445557)*x[1]+(1.02359000173)*x[2]
        arg[0,4,2,0]=(1.19042892582)*x[0]+(-0.265599567429)*x[1]+(-0.412933583721)*x[2]
        arg[0,4,2,1]=(0.105768402681)*x[0]+(0.529442827604)*x[1]+(1.70332495695)*x[2]
        arg[1,0,0,0]=(-0.466512450689)*x[0]+(-0.821724116037)*x[1]+(0.315646947734)*x[2]
        arg[1,0,0,1]=(-0.343965024161)*x[0]+(-0.718276908655)*x[1]+(-0.89648206258)*x[2]
        arg[1,0,1,0]=(0.533001165605)*x[0]+(0.0431632707839)*x[1]+(-0.34028160642)*x[2]
        arg[1,0,1,1]=(-0.0905898060243)*x[0]+(-0.634642149101)*x[1]+(-1.13501642684)*x[2]
        arg[1,0,2,0]=(0.660943497868)*x[0]+(-0.354792953445)*x[1]+(1.28780893672)*x[2]
        arg[1,0,2,1]=(-1.2338460337)*x[0]+(-1.07489268654)*x[1]+(-1.32063885434)*x[2]
        arg[1,1,0,0]=(-0.564276365398)*x[0]+(1.03120151704)*x[1]+(-1.15799041025)*x[2]
        arg[1,1,0,1]=(0.715308242213)*x[0]+(1.38372056508)*x[1]+(0.143855804663)*x[2]
        arg[1,1,1,0]=(-0.359352464902)*x[0]+(1.12698364061)*x[1]+(0.592356171876)*x[2]
        arg[1,1,1,1]=(-0.450554686066)*x[0]+(-0.546866217679)*x[1]+(1.0842968414)*x[2]
        arg[1,1,2,0]=(-0.477267172046)*x[0]+(-0.863491255396)*x[1]+(-0.991333852913)*x[2]
        arg[1,1,2,1]=(0.402599196931)*x[0]+(-0.635287203482)*x[1]+(-0.277210261227)*x[2]
        arg[1,2,0,0]=(-1.56932441015)*x[0]+(-1.56442238007)*x[1]+(-0.375354109955)*x[2]
        arg[1,2,0,1]=(0.193969262086)*x[0]+(-0.0913153952113)*x[1]+(-0.764434182025)*x[2]
        arg[1,2,1,0]=(-0.836765580561)*x[0]+(0.84683301026)*x[1]+(-1.05033967172)*x[2]
        arg[1,2,1,1]=(1.68348957076)*x[0]+(-1.14474223991)*x[1]+(-0.0028142026612)*x[2]
        arg[1,2,2,0]=(-0.0921618478149)*x[0]+(-0.555660652821)*x[1]+(-0.398214997034)*x[2]
        arg[1,2,2,1]=(1.03261791276)*x[0]+(-0.650225649301)*x[1]+(1.63375423051)*x[2]
        arg[1,3,0,0]=(-0.348599695393)*x[0]+(-1.20215183267)*x[1]+(-0.1129459677)*x[2]
        arg[1,3,0,1]=(0.877317784105)*x[0]+(-0.587978566735)*x[1]+(1.10042804546)*x[2]
        arg[1,3,1,0]=(-0.547033849227)*x[0]+(-1.28198893788)*x[1]+(-0.424854587074)*x[2]
        arg[1,3,1,1]=(0.739360669292)*x[0]+(-0.66359460621)*x[1]+(1.08588749669)*x[2]
        arg[1,3,2,0]=(0.137130432185)*x[0]+(-0.123436803848)*x[1]+(-1.31161936702)*x[2]
        arg[1,3,2,1]=(-0.0171220220237)*x[0]+(1.44556512776)*x[1]+(0.665008556384)*x[2]
        arg[1,4,0,0]=(-0.304568160155)*x[0]+(-0.453489411096)*x[1]+(0.769736915169)*x[2]
        arg[1,4,0,1]=(-0.389567079496)*x[0]+(1.01465816808)*x[1]+(1.05389792181)*x[2]
        arg[1,4,1,0]=(-0.293019902659)*x[0]+(0.935020375302)*x[1]+(0.465003429909)*x[2]
        arg[1,4,1,1]=(1.26426164192)*x[0]+(0.53660188438)*x[1]+(-1.26997852414)*x[2]
        arg[1,4,2,0]=(0.447013260347)*x[0]+(-0.292646095073)*x[1]+(-0.141236248636)*x[2]
        arg[1,4,2,1]=(0.627246227193)*x[0]+(-1.07310030602)*x[1]+(0.414244174014)*x[2]
        arg[2,0,0,0]=(0.0450404092088)*x[0]+(-0.092484507099)*x[1]+(0.416562031216)*x[2]
        arg[2,0,0,1]=(0.437290559939)*x[0]+(-1.00144259207)*x[1]+(1.26666155425)*x[2]
        arg[2,0,1,0]=(0.0919297766737)*x[0]+(0.925519273623)*x[1]+(-0.582669368425)*x[2]
        arg[2,0,1,1]=(1.40247347323)*x[0]+(-0.896360612277)*x[1]+(-0.235401929947)*x[2]
        arg[2,0,2,0]=(-0.0760543780641)*x[0]+(-1.12543060216)*x[1]+(-1.73005861985)*x[2]
        arg[2,0,2,1]=(-0.0540914682488)*x[0]+(0.686173107995)*x[1]+(-0.499264219762)*x[2]
        arg[2,1,0,0]=(1.29412006754)*x[0]+(1.17564161143)*x[1]+(0.30140132879)*x[2]
        arg[2,1,0,1]=(-0.987572082341)*x[0]+(-1.73401714184)*x[1]+(0.542753625023)*x[2]
        arg[2,1,1,0]=(-0.162426291727)*x[0]+(0.508287986142)*x[1]+(0.42345261935)*x[2]
        arg[2,1,1,1]=(-0.436007145923)*x[0]+(0.344485478615)*x[1]+(0.867081685885)*x[2]
        arg[2,1,2,0]=(1.23340638096)*x[0]+(1.16627983483)*x[1]+(0.291792397233)*x[2]
        arg[2,1,2,1]=(1.81405521045)*x[0]+(-0.358711641776)*x[1]+(1.67495221975)*x[2]
        arg[2,2,0,0]=(-0.998776647505)*x[0]+(1.38487801142)*x[1]+(-0.199886556937)*x[2]
        arg[2,2,0,1]=(-0.151564577745)*x[0]+(0.100245326484)*x[1]+(-0.98457000197)*x[2]
        arg[2,2,1,0]=(-0.767959771993)*x[0]+(-0.641538616024)*x[1]+(0.0206020120898)*x[2]
        arg[2,2,1,1]=(-0.116988718822)*x[0]+(-1.15499563681)*x[1]+(-0.187589599016)*x[2]
        arg[2,2,2,0]=(0.682354528276)*x[0]+(0.606005834005)*x[1]+(0.11468472132)*x[2]
        arg[2,2,2,1]=(-0.547781355521)*x[0]+(1.23047427005)*x[1]+(0.854018236174)*x[2]
        arg[2,3,0,0]=(0.368349078554)*x[0]+(-1.12153473446)*x[1]+(0.114116055257)*x[2]
        arg[2,3,0,1]=(0.838089123891)*x[0]+(0.283469213399)*x[1]+(0.249874271688)*x[2]
        arg[2,3,1,0]=(-1.03882746612)*x[0]+(0.159711106612)*x[1]+(-1.57272656988)*x[2]
        arg[2,3,1,1]=(-1.25439210606)*x[0]+(-0.801588333704)*x[1]+(-0.166928557775)*x[2]
        arg[2,3,2,0]=(0.0650011438717)*x[0]+(0.879974749381)*x[1]+(-0.276534309239)*x[2]
        arg[2,3,2,1]=(-0.774913691581)*x[0]+(-0.360358063631)*x[1]+(-0.368288145309)*x[2]
        arg[2,4,0,0]=(-0.855093207242)*x[0]+(0.82049336972)*x[1]+(-0.496007273696)*x[2]
        arg[2,4,0,1]=(-0.113408635338)*x[0]+(0.116297312211)*x[1]+(-1.05279504872)*x[2]
        arg[2,4,1,0]=(1.59034968962)*x[0]+(-0.993664270946)*x[1]+(-1.44097244774)*x[2]
        arg[2,4,1,1]=(-0.554191144546)*x[0]+(-1.30747375036)*x[1]+(-1.17946836248)*x[2]
        arg[2,4,2,0]=(0.626012137217)*x[0]+(0.518243725142)*x[1]+(0.0933303150622)*x[2]
        arg[2,4,2,1]=(-0.10169564723)*x[0]+(1.45820149276)*x[1]+(-0.0393269028179)*x[2]
        arg[3,0,0,0]=(-0.0404687905735)*x[0]+(0.0715700915639)*x[1]+(1.45121483418)*x[2]
        arg[3,0,0,1]=(1.25470808156)*x[0]+(-0.119317783496)*x[1]+(0.28341020961)*x[2]
        arg[3,0,1,0]=(0.352807293701)*x[0]+(0.230591898707)*x[1]+(-0.580416381826)*x[2]
        arg[3,0,1,1]=(1.6169875055)*x[0]+(0.515147269207)*x[1]+(-0.461088044559)*x[2]
        arg[3,0,2,0]=(0.345374169211)*x[0]+(0.165179946106)*x[1]+(0.0761591935161)*x[2]
        arg[3,0,2,1]=(-0.367306414632)*x[0]+(-0.236156646602)*x[1]+(-0.388955870267)*x[2]
        arg[3,1,0,0]=(-0.27984693051)*x[0]+(-0.0431489812153)*x[1]+(-0.315624571395)*x[2]
        arg[3,1,0,1]=(-0.923109329983)*x[0]+(0.71133451696)*x[1]+(-0.793551470792)*x[2]
        arg[3,1,1,0]=(-0.434858211337)*x[0]+(0.943115368224)*x[1]+(0.195339633021)*x[2]
        arg[3,1,1,1]=(0.517533657447)*x[0]+(-0.975487574798)*x[1]+(0.00389777788827)*x[2]
        arg[3,1,2,0]=(-1.29728995298)*x[0]+(-0.519355341481)*x[1]+(-0.986985327874)*x[2]
        arg[3,1,2,1]=(-0.104824066468)*x[0]+(0.155059832097)*x[1]+(-0.513194354212)*x[2]
        arg[3,2,0,0]=(-0.053676798897)*x[0]+(-0.957854779002)*x[1]+(0.575673251241)*x[2]
        arg[3,2,0,1]=(0.828930053602)*x[0]+(0.122518313839)*x[1]+(-0.711848816688)*x[2]
        arg[3,2,1,0]=(0.468063581449)*x[0]+(0.233244036405)*x[1]+(-1.06333921952)*x[2]
        arg[3,2,1,1]=(-0.178080946586)*x[0]+(0.481804828166)*x[1]+(-0.25332042181)*x[2]
        arg[3,2,2,0]=(-0.2520881536)*x[0]+(-1.03459378989)*x[1]+(-0.284550808049)*x[2]
        arg[3,2,2,1]=(-0.778922518429)*x[0]+(1.0037625091)*x[1]+(-0.252088333444)*x[2]
        arg[3,3,0,0]=(0.467759844055)*x[0]+(0.598630111441)*x[1]+(-0.180631591223)*x[2]
        arg[3,3,0,1]=(-0.704465112342)*x[0]+(0.577992884135)*x[1]+(0.614639798328)*x[2]
        arg[3,3,1,0]=(0.58772281958)*x[0]+(0.267751378071)*x[1]+(0.11546949909)*x[2]
        arg[3,3,1,1]=(-0.220490484007)*x[0]+(-0.346075755455)*x[1]+(0.709249188793)*x[2]
        arg[3,3,2,0]=(-0.03823885045)*x[0]+(1.0070390689)*x[1]+(0.786191739951)*x[2]
        arg[3,3,2,1]=(0.468772686804)*x[0]+(-0.416774817651)*x[1]+(1.65342478719)*x[2]
        arg[3,4,0,0]=(0.69436779527)*x[0]+(-0.644342658009)*x[1]+(-0.6765845785)*x[2]
        arg[3,4,0,1]=(0.210253965392)*x[0]+(0.208735045316)*x[1]+(-0.553442080096)*x[2]
        arg[3,4,1,0]=(-0.0680478107141)*x[0]+(0.770215954791)*x[1]+(-0.876270039559)*x[2]
        arg[3,4,1,1]=(-0.384812708168)*x[0]+(-0.482707860969)*x[1]+(1.00274437781)*x[2]
        arg[3,4,2,0]=(-0.144567417774)*x[0]+(0.277773632892)*x[1]+(1.3123295509)*x[2]
        arg[3,4,2,1]=(0.00870646304914)*x[0]+(-0.254229329332)*x[1]+(-1.66538680273)*x[2]
        ref[0,0,0,0]=(0.841760617241)*(1+2.*(dim-1.)/(o+1.))+(-1.18631140441)*dim
        ref[0,0,0,1]=(-0.610007337864)*(1+2.*(dim-1.)/(o+1.))+(-1.71451486621)*dim
        ref[0,0,1,0]=(-0.765605897612)*(1+2.*(dim-1.)/(o+1.))+(0.83571459148)*dim
        ref[0,0,1,1]=(-0.128014940891)*(1+2.*(dim-1.)/(o+1.))+(-1.14397157584)*dim
        ref[0,0,2,0]=(-0.969876086955)*(1+2.*(dim-1.)/(o+1.))+(-0.238541556867)*dim
        ref[0,0,2,1]=(-0.434074835341)*(1+2.*(dim-1.)/(o+1.))+(-1.65441946755)*dim
        ref[0,1,0,0]=(0.346878720962)*(1+2.*(dim-1.)/(o+1.))+(-0.722019888602)*dim
        ref[0,1,0,1]=(-0.163332213104)*(1+2.*(dim-1.)/(o+1.))+(-0.135481298717)*dim
        ref[0,1,1,0]=(0.64836517897)*(1+2.*(dim-1.)/(o+1.))+(-0.720330834499)*dim
        ref[0,1,1,1]=(1.50578185724)*(1+2.*(dim-1.)/(o+1.))+(2.13408486278)*dim
        ref[0,1,2,0]=(-1.24719412419)*(1+2.*(dim-1.)/(o+1.))+(0.0927684729819)*dim
        ref[0,1,2,1]=(0.798388744364)*(1+2.*(dim-1.)/(o+1.))+(-1.93130500501)*dim
        ref[0,2,0,0]=(-1.9926882083)*(1+2.*(dim-1.)/(o+1.))+(0.191543055387)*dim
        ref[0,2,0,1]=(-2.16203894433)*(1+2.*(dim-1.)/(o+1.))+(0.824988972246)*dim
        ref[0,2,1,0]=(-2.24934415172)*(1+2.*(dim-1.)/(o+1.))+(-0.263056346511)*dim
        ref[0,2,1,1]=(-0.719152132879)*(1+2.*(dim-1.)/(o+1.))+(-1.00573012142)*dim
        ref[0,2,2,0]=(-0.83751635258)*(1+2.*(dim-1.)/(o+1.))+(0.277097758257)*dim
        ref[0,2,2,1]=(-0.979077481565)*(1+2.*(dim-1.)/(o+1.))+(0.0516269952525)*dim
        ref[0,3,0,0]=(-0.623873484234)*(1+2.*(dim-1.)/(o+1.))+(-0.767010416553)*dim
        ref[0,3,0,1]=(-1.0425822452)*(1+2.*(dim-1.)/(o+1.))+(1.12556988353)*dim
        ref[0,3,1,0]=(0.548600253744)*(1+2.*(dim-1.)/(o+1.))+(0.407341601031)*dim
        ref[0,3,1,1]=(-0.531963828356)*(1+2.*(dim-1.)/(o+1.))+(0.211386892908)*dim
        ref[0,3,2,0]=(1.33006849004)*(1+2.*(dim-1.)/(o+1.))+(-0.681704517735)*dim
        ref[0,3,2,1]=(-1.98393315604)*(1+2.*(dim-1.)/(o+1.))+(0.0385339611081)*dim
        ref[0,4,0,0]=(-0.381511396261)*(1+2.*(dim-1.)/(o+1.))+(-0.991592052336)*dim
        ref[0,4,0,1]=(1.00472425157)*(1+2.*(dim-1.)/(o+1.))+(-1.32537840467)*dim
        ref[0,4,1,0]=(1.04630307783)*(1+2.*(dim-1.)/(o+1.))+(-1.18112274434)*dim
        ref[0,4,1,1]=(1.61731240292)*(1+2.*(dim-1.)/(o+1.))+(1.3676356274)*dim
        ref[0,4,2,0]=(1.41653781097)*(1+2.*(dim-1.)/(o+1.))+(-0.904642036296)*dim
        ref[0,4,2,1]=(1.61515833451)*(1+2.*(dim-1.)/(o+1.))+(0.723377852725)*dim
        ref[1,0,0,0]=(-0.879218635)*(1+2.*(dim-1.)/(o+1.))+(-0.0933709839921)*dim
        ref[1,0,0,1]=(-1.26687353853)*(1+2.*(dim-1.)/(o+1.))+(-0.691850456865)*dim
        ref[1,0,1,0]=(1.79928802103)*(1+2.*(dim-1.)/(o+1.))+(-1.56340519106)*dim
        ref[1,0,1,1]=(-1.05495675007)*(1+2.*(dim-1.)/(o+1.))+(-0.805291631901)*dim
        ref[1,0,2,0]=(-0.339423840638)*(1+2.*(dim-1.)/(o+1.))+(1.93338332178)*dim
        ref[1,0,2,1]=(-1.15347644771)*(1+2.*(dim-1.)/(o+1.))+(-2.47590112686)*dim
        ref[1,1,0,0]=(-1.22232845026)*(1+2.*(dim-1.)/(o+1.))+(0.531263191656)*dim
        ref[1,1,0,1]=(0.318005373727)*(1+2.*(dim-1.)/(o+1.))+(1.92487923823)*dim
        ref[1,1,1,0]=(0.157193055461)*(1+2.*(dim-1.)/(o+1.))+(1.20279429213)*dim
        ref[1,1,1,1]=(0.0525914033484)*(1+2.*(dim-1.)/(o+1.))+(0.0342845343071)*dim
        ref[1,1,2,0]=(-0.319573431275)*(1+2.*(dim-1.)/(o+1.))+(-2.01251884908)*dim
        ref[1,1,2,1]=(-1.67524932091)*(1+2.*(dim-1.)/(o+1.))+(1.16535105313)*dim
        ref[1,2,0,0]=(-2.25088459423)*(1+2.*(dim-1.)/(o+1.))+(-1.25821630594)*dim
        ref[1,2,0,1]=(-0.447759463011)*(1+2.*(dim-1.)/(o+1.))+(-0.214020852139)*dim
        ref[1,2,1,0]=(-1.26590071291)*(1+2.*(dim-1.)/(o+1.))+(0.225628470886)*dim
        ref[1,2,1,1]=(-0.372878924951)*(1+2.*(dim-1.)/(o+1.))+(0.90881205314)*dim
        ref[1,2,2,0]=(1.36047043606)*(1+2.*(dim-1.)/(o+1.))+(-2.40650793373)*dim
        ref[1,2,2,1]=(1.13871462186)*(1+2.*(dim-1.)/(o+1.))+(0.877431872103)*dim
        ref[1,3,0,0]=(-1.17967356605)*(1+2.*(dim-1.)/(o+1.))+(-0.48402392971)*dim
        ref[1,3,0,1]=(-0.177974933247)*(1+2.*(dim-1.)/(o+1.))+(1.56774219607)*dim
        ref[1,3,1,0]=(-1.40936193115)*(1+2.*(dim-1.)/(o+1.))+(-0.844515443034)*dim
        ref[1,3,1,1]=(0.790537554795)*(1+2.*(dim-1.)/(o+1.))+(0.371116004973)*dim
        ref[1,3,2,0]=(0.242057644413)*(1+2.*(dim-1.)/(o+1.))+(-1.5399833831)*dim
        ref[1,3,2,1]=(1.70371239948)*(1+2.*(dim-1.)/(o+1.))+(0.389739262639)*dim
        ref[1,4,0,0]=(-0.375645743677)*(1+2.*(dim-1.)/(o+1.))+(0.387325087595)*dim
        ref[1,4,0,1]=(0.0489286471987)*(1+2.*(dim-1.)/(o+1.))+(1.63006036319)*dim
        ref[1,4,1,0]=(1.33203691786)*(1+2.*(dim-1.)/(o+1.))+(-0.225033015313)*dim
        ref[1,4,1,1]=(0.332297845981)*(1+2.*(dim-1.)/(o+1.))+(0.198587156175)*dim
        ref[1,4,2,0]=(-1.22288078692)*(1+2.*(dim-1.)/(o+1.))+(1.23601170355)*dim
        ref[1,4,2,1]=(0.0299493142028)*(1+2.*(dim-1.)/(o+1.))+(-0.0615592190171)*dim
        ref[2,0,0,0]=(0.0963155205346)*(1+2.*(dim-1.)/(o+1.))+(0.272802412791)*dim
        ref[2,0,0,1]=(1.12280541023)*(1+2.*(dim-1.)/(o+1.))+(-0.420295888105)*dim
        ref[2,0,1,0]=(-0.296878488858)*(1+2.*(dim-1.)/(o+1.))+(0.73165817073)*dim
        ref[2,0,1,1]=(-0.407704337172)*(1+2.*(dim-1.)/(o+1.))+(0.678415268175)*dim
        ref[2,0,2,0]=(-2.37402978977)*(1+2.*(dim-1.)/(o+1.))+(-0.557513810308)*dim
        ref[2,0,2,1]=(1.2564150767)*(1+2.*(dim-1.)/(o+1.))+(-1.12359765672)*dim
        ref[2,1,0,0]=(0.824412811392)*(1+2.*(dim-1.)/(o+1.))+(1.94675019637)*dim
        ref[2,1,0,1]=(-1.06505178789)*(1+2.*(dim-1.)/(o+1.))+(-1.11378381126)*dim
        ref[2,1,1,0]=(0.200096305752)*(1+2.*(dim-1.)/(o+1.))+(0.569218008012)*dim
        ref[2,1,1,1]=(1.08680720067)*(1+2.*(dim-1.)/(o+1.))+(-0.311247182096)*dim
        ref[2,1,2,0]=(1.3764914842)*(1+2.*(dim-1.)/(o+1.))+(1.31498712883)*dim
        ref[2,1,2,1]=(2.26537199157)*(1+2.*(dim-1.)/(o+1.))+(0.864923796858)*dim
        ref[2,2,0,0]=(0.750025182887)*(1+2.*(dim-1.)/(o+1.))+(-0.563810375913)*dim
        ref[2,2,0,1]=(-0.806933575047)*(1+2.*(dim-1.)/(o+1.))+(-0.228955678184)*dim
        ref[2,2,1,0]=(-1.53132620021)*(1+2.*(dim-1.)/(o+1.))+(0.14242982428)*dim
        ref[2,2,1,1]=(-0.889411443106)*(1+2.*(dim-1.)/(o+1.))+(-0.570162511538)*dim
        ref[2,2,2,0]=(-0.0309717316948)*(1+2.*(dim-1.)/(o+1.))+(1.4340168153)*dim
        ref[2,2,2,1]=(1.48963237375)*(1+2.*(dim-1.)/(o+1.))+(0.0470787769518)*dim
        ref[2,3,0,0]=(-0.0184174113004)*(1+2.*(dim-1.)/(o+1.))+(-0.620652189349)*dim
        ref[2,3,0,1]=(0.597433991421)*(1+2.*(dim-1.)/(o+1.))+(0.773998617557)*dim
        ref[2,3,1,0]=(-1.10704622688)*(1+2.*(dim-1.)/(o+1.))+(-1.34479670251)*dim
        ref[2,3,1,1]=(-0.369209798891)*(1+2.*(dim-1.)/(o+1.))+(-1.85369919864)*dim
        ref[2,3,2,0]=(-0.286165609161)*(1+2.*(dim-1.)/(o+1.))+(0.954607193175)*dim
        ref[2,3,2,1]=(-0.0948450773617)*(1+2.*(dim-1.)/(o+1.))+(-1.40871482316)*dim
        ref[2,4,0,0]=(-0.350521344827)*(1+2.*(dim-1.)/(o+1.))+(-0.180085766391)*dim
        ref[2,4,0,1]=(-1.10948105015)*(1+2.*(dim-1.)/(o+1.))+(0.0595746783013)*dim
        ref[2,4,1,0]=(-0.251521967553)*(1+2.*(dim-1.)/(o+1.))+(-0.592765061514)*dim
        ref[2,4,1,1]=(-2.22732408506)*(1+2.*(dim-1.)/(o+1.))+(-0.813809172329)*dim
        ref[2,4,2,0]=(-1.4091057051)*(1+2.*(dim-1.)/(o+1.))+(2.64669188252)*dim
        ref[2,4,2,1]=(-0.0262518153279)*(1+2.*(dim-1.)/(o+1.))+(1.34343075804)*dim
        ref[3,0,0,0]=(2.08524442815)*(1+2.*(dim-1.)/(o+1.))+(-0.602928292982)*dim
        ref[3,0,0,1]=(0.227442743255)*(1+2.*(dim-1.)/(o+1.))+(1.19135776442)*dim
        ref[3,0,1,0]=(1.03371796145)*(1+2.*(dim-1.)/(o+1.))+(-1.03073515087)*dim
        ref[3,0,1,1]=(0.205903763449)*(1+2.*(dim-1.)/(o+1.))+(1.4651429667)*dim
        ref[3,0,2,0]=(1.54387435524)*(1+2.*(dim-1.)/(o+1.))+(-0.957161046404)*dim
        ref[3,0,2,1]=(-0.487693556268)*(1+2.*(dim-1.)/(o+1.))+(-0.504725375232)*dim
        ref[3,1,0,0]=(0.319450917291)*(1+2.*(dim-1.)/(o+1.))+(-0.958071400411)*dim
        ref[3,1,0,1]=(-0.0756676477654)*(1+2.*(dim-1.)/(o+1.))+(-0.929658636049)*dim
        ref[3,1,1,0]=(0.0109030665653)*(1+2.*(dim-1.)/(o+1.))+(0.692693723343)*dim
        ref[3,1,1,1]=(0.540904091106)*(1+2.*(dim-1.)/(o+1.))+(-0.994960230569)*dim
        ref[3,1,2,0]=(-1.39229945953)*(1+2.*(dim-1.)/(o+1.))+(-1.4113311628)*dim
        ref[3,1,2,1]=(-0.120906779216)*(1+2.*(dim-1.)/(o+1.))+(-0.342051809366)*dim
        ref[3,2,0,0]=(-0.61477618653)*(1+2.*(dim-1.)/(o+1.))+(0.178917859873)*dim
        ref[3,2,0,1]=(0.286238606318)*(1+2.*(dim-1.)/(o+1.))+(-0.0466390555654)*dim
        ref[3,2,1,0]=(-0.0560560726323)*(1+2.*(dim-1.)/(o+1.))+(-0.305975529033)*dim
        ref[3,2,1,1]=(0.935166077059)*(1+2.*(dim-1.)/(o+1.))+(-0.884762617288)*dim
        ref[3,2,2,0]=(0.57015079501)*(1+2.*(dim-1.)/(o+1.))+(-2.14138354655)*dim
        ref[3,2,2,1]=(-0.699470490706)*(1+2.*(dim-1.)/(o+1.))+(0.672222147932)*dim
        ref[3,3,0,0]=(1.33431543118)*(1+2.*(dim-1.)/(o+1.))+(-0.448557066912)*dim
        ref[3,3,0,1]=(-1.03724498716)*(1+2.*(dim-1.)/(o+1.))+(1.52541255729)*dim
        ref[3,3,1,0]=(-0.00371131395733)*(1+2.*(dim-1.)/(o+1.))+(0.974655010698)*dim
        ref[3,3,1,1]=(0.0918410742793)*(1+2.*(dim-1.)/(o+1.))+(0.0508418750519)*dim
        ref[3,3,2,0]=(1.63372857769)*(1+2.*(dim-1.)/(o+1.))+(0.121263380706)*dim
        ref[3,3,2,1]=(0.761569446324)*(1+2.*(dim-1.)/(o+1.))+(0.943853210019)*dim
        ref[3,4,0,0]=(-0.625751279162)*(1+2.*(dim-1.)/(o+1.))+(-0.000808162076656)*dim
        ref[3,4,0,1]=(-0.113324890356)*(1+2.*(dim-1.)/(o+1.))+(-0.0211281790319)*dim
        ref[3,4,1,0]=(0.296144237506)*(1+2.*(dim-1.)/(o+1.))+(-0.470246132989)*dim
        ref[3,4,1,1]=(0.194264324372)*(1+2.*(dim-1.)/(o+1.))+(-0.0590405157015)*dim
        ref[3,4,2,0]=(1.63851379122)*(1+2.*(dim-1.)/(o+1.))+(-0.192978025203)*dim
        ref[3,4,2,1]=(-0.757374882295)*(1+2.*(dim-1.)/(o+1.))+(-1.15353478672)*dim
      res=integrate(arg,where=w_ref)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank0(self):
      """
      tests integral of rank 0 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      if dim==2:
        arg=(0.626062912046)*x[0]**o+(0.600712856149)*x[0]+(0.0478861882442)*x[1]**o+(0.77008552734)*x[1]
        ref=(0.67394910029)*(1+2.*(dim-1.)/(o+1.))+(1.37079838349)*dim
      else:
        arg=(0.85588880126)*x[0]**o+(-0.898804533462)*x[0]+(-0.270816921088)*x[1]**o+(-0.842823912373)*x[1]+(-0.47559690762)*x[2]**o+(0.751770125129)*x[2]
        ref=(0.109474972552)*(1+2.*(dim-1.)/(o+1.))+(-0.989858320706)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank1(self):
      """
      tests integral of rank 1 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        arg[0]=(0.298913063662)*x[0]**o+(-0.250822875451)*x[0]+(0.707311253225)*x[1]**o+(0.0408091445557)*x[1]
        arg[1]=(-0.70143850042)*x[0]**o+(-0.80121227022)*x[0]+(0.0731514590408)*x[1]**o+(0.621969876113)*x[1]
        ref[0]=(1.00622431689)*(1+2.*(dim-1.)/(o+1.))+(-0.210013730895)*dim
        ref[1]=(-0.628287041379)*(1+2.*(dim-1.)/(o+1.))+(-0.179242394108)*dim
      else:
        arg[0]=(0.676976938336)*x[0]**o+(-0.928207054947)*x[0]+(0.864749399199)*x[1]**o+(0.864358390848)*x[1]+(-0.0255569869984)*x[2]**o+(0.0268536872793)*x[2]
        arg[1]=(0.270542056965)*x[0]**o+(0.0696194435081)*x[0]+(0.401360001043)*x[1]**o+(0.917064528874)*x[1]+(-0.755510144559)*x[2]**o+(0.881244515908)*x[2]
        ref[0]=(1.51616935054)*(1+2.*(dim-1.)/(o+1.))+(-0.0369949768193)*dim
        ref[1]=(-0.083608086551)*(1+2.*(dim-1.)/(o+1.))+(1.86792848829)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank2(self):
      """
      tests integral of rank 2 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        arg[0,0]=(-0.381843423534)*x[0]**o+(-0.563210012057)*x[0]+(0.730631055205)*x[1]**o+(0.13636238698)*x[1]
        arg[0,1]=(0.306357653338)*x[0]**o+(-0.797201232101)*x[0]+(-0.66333793372)*x[1]**o+(-0.943650360464)*x[1]
        arg[0,2]=(-0.34058649734)*x[0]**o+(-0.27047792232)*x[0]+(0.57156687267)*x[1]**o+(-0.659114355468)*x[1]
        arg[0,3]=(0.73319004289)*x[0]**o+(0.229139540294)*x[0]+(-0.51558765225)*x[1]**o+(0.45052521066)*x[1]
        arg[0,4]=(-0.264117198359)*x[0]**o+(0.807950109095)*x[0]+(-0.290558748289)*x[1]**o+(0.161864708045)*x[1]
        arg[1,0]=(0.988732415297)*x[0]**o+(-0.725047414136)*x[0]+(-0.722337599885)*x[1]**o+(0.739130135256)*x[1]
        arg[1,1]=(-0.315086233906)*x[0]**o+(-0.783177846939)*x[0]+(-0.271931454399)*x[1]**o+(-0.935744613016)*x[1]
        arg[1,2]=(-0.787047622811)*x[0]**o+(0.00746242211892)*x[0]+(-0.539716273578)*x[1]**o+(-0.340449115622)*x[1]
        arg[1,3]=(-0.782830405823)*x[0]**o+(0.842318742811)*x[0]+(-0.118894493096)*x[1]**o+(0.645943277412)*x[1]
        arg[1,4]=(0.0594068367647)*x[0]**o+(-0.184711682975)*x[0]+(-0.0643050731295)*x[1]**o+(-0.981593287261)*x[1]
        arg[2,0]=(0.17170754931)*x[0]**o+(0.582651363494)*x[0]+(0.334422749985)*x[1]**o+(0.852747429144)*x[1]
        arg[2,1]=(0.683967516208)*x[0]**o+(-0.612989314204)*x[0]+(0.845175135723)*x[1]**o+(0.374223395314)*x[1]
        arg[2,2]=(0.588917536035)*x[0]**o+(0.0130713873939)*x[0]+(0.881942872071)*x[1]**o+(0.683941046151)*x[1]
        arg[2,3]=(-0.63861264887)*x[0]**o+(0.453919752415)*x[0]+(0.448731342326)*x[1]**o+(-0.259183783496)*x[1]
        arg[2,4]=(-0.132175638549)*x[0]**o+(0.710071566415)*x[0]+(-0.870119332524)*x[1]**o+(0.523394538403)*x[1]
        arg[3,0]=(0.768036036634)*x[0]**o+(-0.218321673662)*x[0]+(0.50224512768)*x[1]**o+(0.694191572559)*x[1]
        arg[3,1]=(-0.298369075336)*x[0]**o+(-0.276776010079)*x[0]+(-0.659309162004)*x[1]**o+(0.436841182692)*x[1]
        arg[3,2]=(-0.00370049871499)*x[0]**o+(-0.41700259949)*x[0]+(0.726806931339)*x[1]**o+(-0.113165181924)*x[1]
        arg[3,3]=(0.288175924931)*x[0]**o+(-0.489667003521)*x[0]+(-0.531842953651)*x[1]**o+(-0.443517285929)*x[1]
        arg[3,4]=(-0.645574931589)*x[0]**o+(0.464738749544)*x[0]+(-0.538213524368)*x[1]**o+(-0.94443292146)*x[1]
        ref[0,0]=(0.348787631671)*(1+2.*(dim-1.)/(o+1.))+(-0.426847625077)*dim
        ref[0,1]=(-0.356980280382)*(1+2.*(dim-1.)/(o+1.))+(-1.74085159257)*dim
        ref[0,2]=(0.23098037533)*(1+2.*(dim-1.)/(o+1.))+(-0.929592277787)*dim
        ref[0,3]=(0.21760239064)*(1+2.*(dim-1.)/(o+1.))+(0.679664750954)*dim
        ref[0,4]=(-0.554675946648)*(1+2.*(dim-1.)/(o+1.))+(0.96981481714)*dim
        ref[1,0]=(0.266394815412)*(1+2.*(dim-1.)/(o+1.))+(0.0140827211195)*dim
        ref[1,1]=(-0.587017688304)*(1+2.*(dim-1.)/(o+1.))+(-1.71892245996)*dim
        ref[1,2]=(-1.32676389639)*(1+2.*(dim-1.)/(o+1.))+(-0.332986693503)*dim
        ref[1,3]=(-0.901724898919)*(1+2.*(dim-1.)/(o+1.))+(1.48826202022)*dim
        ref[1,4]=(-0.00489823636486)*(1+2.*(dim-1.)/(o+1.))+(-1.16630497024)*dim
        ref[2,0]=(0.506130299295)*(1+2.*(dim-1.)/(o+1.))+(1.43539879264)*dim
        ref[2,1]=(1.52914265193)*(1+2.*(dim-1.)/(o+1.))+(-0.238765918889)*dim
        ref[2,2]=(1.47086040811)*(1+2.*(dim-1.)/(o+1.))+(0.697012433545)*dim
        ref[2,3]=(-0.189881306544)*(1+2.*(dim-1.)/(o+1.))+(0.194735968919)*dim
        ref[2,4]=(-1.00229497107)*(1+2.*(dim-1.)/(o+1.))+(1.23346610482)*dim
        ref[3,0]=(1.27028116431)*(1+2.*(dim-1.)/(o+1.))+(0.475869898897)*dim
        ref[3,1]=(-0.957678237339)*(1+2.*(dim-1.)/(o+1.))+(0.160065172613)*dim
        ref[3,2]=(0.723106432624)*(1+2.*(dim-1.)/(o+1.))+(-0.530167781414)*dim
        ref[3,3]=(-0.243667028719)*(1+2.*(dim-1.)/(o+1.))+(-0.93318428945)*dim
        ref[3,4]=(-1.18378845596)*(1+2.*(dim-1.)/(o+1.))+(-0.479694171917)*dim
      else:
        arg[0,0]=(0.751222317885)*x[0]**o+(0.953666267123)*x[0]+(-0.707417048158)*x[1]**o+(0.310686543483)*x[1]+(0.326267797641)*x[2]**o+(-0.722772500437)*x[2]
        arg[0,1]=(-0.038709086264)*x[0]**o+(-0.483307819428)*x[0]+(0.611120835303)*x[1]**o+(0.885935954604)*x[1]+(0.630199910046)*x[2]**o+(-0.101764126615)*x[2]
        arg[0,2]=(0.890606731439)*x[0]**o+(-0.0555491679283)*x[0]+(0.334835177686)*x[1]**o+(0.999030888839)*x[1]+(-0.891653525154)*x[2]**o+(-0.14056280616)*x[2]
        arg[0,3]=(-0.458305223634)*x[0]**o+(0.679075077126)*x[0]+(0.1176620176)*x[1]**o+(0.685223087684)*x[1]+(-0.952262623464)*x[2]**o+(-0.905568306628)*x[2]
        arg[0,4]=(0.623179050934)*x[0]**o+(-0.0150847487218)*x[0]+(0.0823495403769)*x[1]**o+(0.48303169009)*x[1]+(0.610276253905)*x[2]**o+(0.353104308072)*x[2]
        arg[1,0]=(-0.709185342895)*x[0]**o+(0.789980598884)*x[0]+(0.655538102878)*x[1]**o+(0.713913568814)*x[1]+(0.222131732773)*x[2]**o+(0.712644729381)*x[2]
        arg[1,1]=(-0.683646421822)*x[0]**o+(-0.943691516528)*x[0]+(-0.134261002557)*x[1]**o+(0.28225941018)*x[1]+(0.906235993779)*x[2]**o+(-0.819214603609)*x[2]
        arg[1,2]=(0.713781993938)*x[0]**o+(-0.189551367619)*x[0]+(-0.021310841804)*x[1]**o+(-0.873466428012)*x[1]+(0.248337664546)*x[2]**o+(0.911824874935)*x[2]
        arg[1,3]=(-0.393142772637)*x[0]**o+(0.773501237865)*x[0]+(0.759039429922)*x[1]**o+(0.67036876499)*x[1]+(-0.921411979023)*x[2]**o+(-0.649371534276)*x[2]
        arg[1,4]=(0.526708082723)*x[0]**o+(-0.522449586718)*x[0]+(0.951176548558)*x[1]**o+(0.604602039898)*x[1]+(0.345888192325)*x[2]**o+(-0.973729954334)*x[2]
        arg[2,0]=(-0.139931215293)*x[0]**o+(-0.996072915638)*x[0]+(-0.0451731431264)*x[1]**o+(0.433694224874)*x[1]+(-0.955189760302)*x[2]**o+(-0.68361667242)*x[2]
        arg[2,1]=(0.139626552661)*x[0]**o+(-0.91217468751)*x[0]+(-0.854252204911)*x[1]**o+(0.417539913464)*x[1]+(0.557096050124)*x[2]**o+(-0.195272359868)*x[2]
        arg[2,2]=(0.960352725049)*x[0]**o+(0.98274500341)*x[0]+(0.967782565025)*x[1]**o+(0.566011353381)*x[1]+(0.583885094583)*x[2]**o+(-0.943173847936)*x[2]
        arg[2,3]=(0.959388333494)*x[0]**o+(0.531908797155)*x[0]+(-0.62531126203)*x[1]**o+(-0.288763293646)*x[1]+(-0.823398780526)*x[2]**o+(-0.164348871248)*x[2]
        arg[2,4]=(0.73397702179)*x[0]**o+(0.407844592399)*x[0]+(-0.835769355727)*x[1]**o+(0.742360626394)*x[1]+(0.31130186529)*x[2]**o+(0.114221168872)*x[2]
        arg[3,0]=(-0.42278940444)*x[0]**o+(-0.513158064442)*x[0]+(-0.133233894233)*x[1]**o+(0.701381913887)*x[1]+(0.283776333553)*x[2]**o+(-0.805096661642)*x[2]
        arg[3,1]=(-0.417338508269)*x[0]**o+(0.535961900783)*x[0]+(0.0836386907734)*x[1]**o+(0.449680928704)*x[1]+(0.998998322868)*x[2]**o+(0.983057825698)*x[2]
        arg[3,2]=(0.24102382276)*x[0]**o+(0.0291470157398)*x[0]+(-0.539542451698)*x[1]**o+(0.281017043631)*x[1]+(-0.464652133499)*x[2]**o+(0.652828720185)*x[2]
        arg[3,3]=(-0.482577920013)*x[0]**o+(0.931756756394)*x[0]+(0.96648206338)*x[1]**o+(0.73941459549)*x[1]+(0.136743015429)*x[2]**o+(0.41477914722)*x[2]
        arg[3,4]=(0.879881885972)*x[0]**o+(0.208970079585)*x[0]+(-0.200079958414)*x[1]**o+(-0.0291238772241)*x[1]+(-0.333569505857)*x[2]**o+(0.846835729875)*x[2]
        ref[0,0]=(0.370073067368)*(1+2.*(dim-1.)/(o+1.))+(0.541580310169)*dim
        ref[0,1]=(1.20261165909)*(1+2.*(dim-1.)/(o+1.))+(0.300864008561)*dim
        ref[0,2]=(0.333788383971)*(1+2.*(dim-1.)/(o+1.))+(0.80291891475)*dim
        ref[0,3]=(-1.2929058295)*(1+2.*(dim-1.)/(o+1.))+(0.458729858183)*dim
        ref[0,4]=(1.31580484522)*(1+2.*(dim-1.)/(o+1.))+(0.82105124944)*dim
        ref[1,0]=(0.168484492756)*(1+2.*(dim-1.)/(o+1.))+(2.21653889708)*dim
        ref[1,1]=(0.0883285694004)*(1+2.*(dim-1.)/(o+1.))+(-1.48064670996)*dim
        ref[1,2]=(0.940808816681)*(1+2.*(dim-1.)/(o+1.))+(-0.151192920696)*dim
        ref[1,3]=(-0.555515321739)*(1+2.*(dim-1.)/(o+1.))+(0.79449846858)*dim
        ref[1,4]=(1.82377282361)*(1+2.*(dim-1.)/(o+1.))+(-0.891577501154)*dim
        ref[2,0]=(-1.14029411872)*(1+2.*(dim-1.)/(o+1.))+(-1.24599536318)*dim
        ref[2,1]=(-0.157529602127)*(1+2.*(dim-1.)/(o+1.))+(-0.689907133913)*dim
        ref[2,2]=(2.51202038466)*(1+2.*(dim-1.)/(o+1.))+(0.605582508855)*dim
        ref[2,3]=(-0.489321709063)*(1+2.*(dim-1.)/(o+1.))+(0.0787966322607)*dim
        ref[2,4]=(0.209509531353)*(1+2.*(dim-1.)/(o+1.))+(1.26442638766)*dim
        ref[3,0]=(-0.27224696512)*(1+2.*(dim-1.)/(o+1.))+(-0.616872812198)*dim
        ref[3,1]=(0.665298505372)*(1+2.*(dim-1.)/(o+1.))+(1.96870065519)*dim
        ref[3,2]=(-0.763170762437)*(1+2.*(dim-1.)/(o+1.))+(0.962992779555)*dim
        ref[3,3]=(0.620647158797)*(1+2.*(dim-1.)/(o+1.))+(2.0859504991)*dim
        ref[3,4]=(0.346232421701)*(1+2.*(dim-1.)/(o+1.))+(1.02668193224)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank3(self):
      """
      tests integral of rank 3 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        arg[0,0,0]=(0.201115186772)*x[0]**o+(0.324758103748)*x[0]+(0.285065583242)*x[1]**o+(0.0264985832782)*x[1]
        arg[0,0,1]=(-0.443023003973)*x[0]**o+(-0.194732927607)*x[0]+(0.64414908519)*x[1]**o+(-0.248261857779)*x[1]
        arg[0,1,0]=(0.587534172195)*x[0]**o+(0.021088951369)*x[0]+(-0.986195557459)*x[1]**o+(0.595659005312)*x[1]
        arg[0,1,1]=(-0.916961145134)*x[0]**o+(-0.766484469605)*x[0]+(-0.281345964683)*x[1]**o+(-0.399113856826)*x[1]
        arg[1,0,0]=(0.175943424949)*x[0]**o+(0.344853386001)*x[0]+(0.175474681848)*x[1]**o+(-0.0775102874697)*x[1]
        arg[1,0,1]=(-0.874074467007)*x[0]**o+(0.903506914387)*x[0]+(0.177699372538)*x[1]**o+(0.969818154968)*x[1]
        arg[1,1,0]=(0.765751472387)*x[0]**o+(-0.0288724203687)*x[0]+(0.655676828008)*x[1]**o+(0.22281307247)*x[1]
        arg[1,1,1]=(0.871009077152)*x[0]**o+(-0.20052353101)*x[0]+(-0.889496767621)*x[1]**o+(0.8458762887)*x[1]
        arg[2,0,0]=(-0.15377513542)*x[0]**o+(0.963714892768)*x[0]+(0.942514815983)*x[1]**o+(0.846111018838)*x[1]
        arg[2,0,1]=(0.000727086938663)*x[0]**o+(0.243136987492)*x[0]+(-0.308674899886)*x[1]**o+(-0.963435664436)*x[1]
        arg[2,1,0]=(0.780912245051)*x[0]**o+(0.227139041778)*x[0]+(-0.336203258384)*x[1]**o+(0.092568284679)*x[1]
        arg[2,1,1]=(0.0464862681168)*x[0]**o+(0.284493375176)*x[0]+(-0.811073509077)*x[1]**o+(0.54280235419)*x[1]
        arg[3,0,0]=(0.605642304612)*x[0]**o+(-0.103442921567)*x[0]+(0.106253324829)*x[1]**o+(0.562480007904)*x[1]
        arg[3,0,1]=(0.497668538008)*x[0]**o+(-0.859679982844)*x[0]+(-0.255499355194)*x[1]**o+(0.0725919239243)*x[1]
        arg[3,1,0]=(0.641826893201)*x[0]**o+(0.889514581887)*x[0]+(-0.211781857836)*x[1]**o+(0.45900873193)*x[1]
        arg[3,1,1]=(-0.29349600565)*x[0]**o+(-0.627542961817)*x[0]+(0.313009642686)*x[1]**o+(0.338028134809)*x[1]
        arg[4,0,0]=(-0.641915241397)*x[0]**o+(-0.682547311604)*x[0]+(-0.106472077451)*x[1]**o+(0.103402239052)*x[1]
        arg[4,0,1]=(-0.822562520978)*x[0]**o+(-0.430539530156)*x[0]+(-0.0225528898923)*x[1]**o+(-0.0908986665674)*x[1]
        arg[4,1,0]=(-0.456160645978)*x[0]**o+(-0.264592336242)*x[0]+(-0.672702978868)*x[1]**o+(-0.893197955639)*x[1]
        arg[4,1,1]=(-0.714689169676)*x[0]**o+(0.513612105215)*x[0]+(-0.3093340377)*x[1]**o+(-0.34467035156)*x[1]
        arg[5,0,0]=(-0.993761664154)*x[0]**o+(0.375953817044)*x[0]+(0.0463240503072)*x[1]**o+(-0.595646829466)*x[1]
        arg[5,0,1]=(0.453145186545)*x[0]**o+(0.331353840672)*x[0]+(-0.908533078632)*x[1]**o+(0.755175839674)*x[1]
        arg[5,1,0]=(-0.736874540101)*x[0]**o+(-0.867437170846)*x[0]+(-0.472395301622)*x[1]**o+(0.78289019661)*x[1]
        arg[5,1,1]=(-0.900966156016)*x[0]**o+(-0.160363876891)*x[0]+(-0.0969222200475)*x[1]**o+(0.0527983470624)*x[1]
        ref[0,0,0]=(0.486180770014)*(1+2.*(dim-1.)/(o+1.))+(0.351256687026)*dim
        ref[0,0,1]=(0.201126081217)*(1+2.*(dim-1.)/(o+1.))+(-0.442994785386)*dim
        ref[0,1,0]=(-0.398661385264)*(1+2.*(dim-1.)/(o+1.))+(0.616747956681)*dim
        ref[0,1,1]=(-1.19830710982)*(1+2.*(dim-1.)/(o+1.))+(-1.16559832643)*dim
        ref[1,0,0]=(0.351418106797)*(1+2.*(dim-1.)/(o+1.))+(0.267343098531)*dim
        ref[1,0,1]=(-0.696375094469)*(1+2.*(dim-1.)/(o+1.))+(1.87332506936)*dim
        ref[1,1,0]=(1.42142830039)*(1+2.*(dim-1.)/(o+1.))+(0.193940652101)*dim
        ref[1,1,1]=(-0.0184876904693)*(1+2.*(dim-1.)/(o+1.))+(0.64535275769)*dim
        ref[2,0,0]=(0.788739680562)*(1+2.*(dim-1.)/(o+1.))+(1.80982591161)*dim
        ref[2,0,1]=(-0.307947812947)*(1+2.*(dim-1.)/(o+1.))+(-0.720298676944)*dim
        ref[2,1,0]=(0.444708986667)*(1+2.*(dim-1.)/(o+1.))+(0.319707326457)*dim
        ref[2,1,1]=(-0.76458724096)*(1+2.*(dim-1.)/(o+1.))+(0.827295729366)*dim
        ref[3,0,0]=(0.711895629441)*(1+2.*(dim-1.)/(o+1.))+(0.459037086337)*dim
        ref[3,0,1]=(0.242169182814)*(1+2.*(dim-1.)/(o+1.))+(-0.78708805892)*dim
        ref[3,1,0]=(0.430045035365)*(1+2.*(dim-1.)/(o+1.))+(1.34852331382)*dim
        ref[3,1,1]=(0.019513637036)*(1+2.*(dim-1.)/(o+1.))+(-0.289514827008)*dim
        ref[4,0,0]=(-0.748387318848)*(1+2.*(dim-1.)/(o+1.))+(-0.579145072551)*dim
        ref[4,0,1]=(-0.845115410871)*(1+2.*(dim-1.)/(o+1.))+(-0.521438196723)*dim
        ref[4,1,0]=(-1.12886362485)*(1+2.*(dim-1.)/(o+1.))+(-1.15779029188)*dim
        ref[4,1,1]=(-1.02402320738)*(1+2.*(dim-1.)/(o+1.))+(0.168941753655)*dim
        ref[5,0,0]=(-0.947437613847)*(1+2.*(dim-1.)/(o+1.))+(-0.219693012422)*dim
        ref[5,0,1]=(-0.455387892087)*(1+2.*(dim-1.)/(o+1.))+(1.08652968035)*dim
        ref[5,1,0]=(-1.20926984172)*(1+2.*(dim-1.)/(o+1.))+(-0.0845469742352)*dim
        ref[5,1,1]=(-0.997888376064)*(1+2.*(dim-1.)/(o+1.))+(-0.107565529828)*dim
      else:
        arg[0,0,0]=(-0.760764230601)*x[0]**o+(-0.955215307904)*x[0]+(0.376251560231)*x[1]**o+(0.147849416477)*x[1]+(0.627037741745)*x[2]**o+(-0.93996244415)*x[2]
        arg[0,0,1]=(0.904449671196)*x[0]**o+(-0.297804849063)*x[0]+(0.773059038176)*x[1]**o+(-0.836104534497)*x[1]+(-0.150040994258)*x[2]**o+(0.837988726131)*x[2]
        arg[0,1,0]=(0.354255190058)*x[0]**o+(0.463977999833)*x[0]+(0.218855388406)*x[1]**o+(-0.163086423536)*x[1]+(0.653790071045)*x[2]**o+(-0.457721486689)*x[2]
        arg[0,1,1]=(0.357317871424)*x[0]**o+(-0.752862019242)*x[0]+(0.674187130836)*x[1]**o+(-0.983503125785)*x[1]+(-0.776153051078)*x[2]**o+(0.390700377058)*x[2]
        arg[1,0,0]=(-0.465423805042)*x[0]**o+(0.629855541232)*x[0]+(-0.405300083887)*x[1]**o+(-0.211254598072)*x[1]+(0.502581881406)*x[2]**o+(-0.853779249469)*x[2]
        arg[1,0,1]=(0.716025450344)*x[0]**o+(0.876021396077)*x[0]+(-0.530324190586)*x[1]**o+(0.700207449985)*x[1]+(-0.788565669024)*x[2]**o+(0.963118484198)*x[2]
        arg[1,1,0]=(0.813943574582)*x[0]**o+(-0.938778607938)*x[0]+(-0.451619613547)*x[1]**o+(-0.916990884105)*x[1]+(-0.671730738367)*x[2]**o+(0.109664560689)*x[2]
        arg[1,1,1]=(-0.557598711518)*x[0]**o+(-0.541691706474)*x[0]+(0.500874328683)*x[1]**o+(-0.959653364589)*x[1]+(0.572057415183)*x[2]**o+(-0.842306660222)*x[2]
        arg[2,0,0]=(0.0470537168687)*x[0]**o+(-0.151521222488)*x[0]+(0.168279496456)*x[1]**o+(0.869948894317)*x[1]+(-0.554196665086)*x[2]**o+(0.444189090676)*x[2]
        arg[2,0,1]=(-0.662480394662)*x[0]**o+(0.506778773041)*x[0]+(0.664368833417)*x[1]**o+(-0.76728005759)*x[1]+(0.729668063743)*x[2]**o+(-0.476499535761)*x[2]
        arg[2,1,0]=(0.40405640406)*x[0]**o+(0.537691275434)*x[0]+(0.675845932799)*x[1]**o+(-0.663037107454)*x[1]+(-0.947246204559)*x[2]**o+(-0.210977364198)*x[2]
        arg[2,1,1]=(-0.460999541334)*x[0]**o+(-0.403551870206)*x[0]+(-0.59751233801)*x[1]**o+(0.389018763902)*x[1]+(-0.916883519483)*x[2]**o+(0.712714107738)*x[2]
        arg[3,0,0]=(-0.547139864054)*x[0]**o+(0.91774354334)*x[0]+(-0.953248927323)*x[1]**o+(0.693673610785)*x[1]+(-0.474082903895)*x[2]**o+(0.187174904886)*x[2]
        arg[3,0,1]=(0.538597533339)*x[0]**o+(0.903368885629)*x[0]+(-0.548058780044)*x[1]**o+(-0.380406303455)*x[1]+(-0.389244091021)*x[2]**o+(0.137402766164)*x[2]
        arg[3,1,0]=(-0.11691167078)*x[0]**o+(-0.812294829383)*x[0]+(0.592192015269)*x[1]**o+(-0.3278056345)*x[1]+(-0.60462640896)*x[2]**o+(-0.701541632465)*x[2]
        arg[3,1,1]=(-0.196382907034)*x[0]**o+(-0.496234735923)*x[0]+(-0.646502033751)*x[1]**o+(-0.18981379172)*x[1]+(0.627608199527)*x[2]**o+(-0.468659552408)*x[2]
        arg[4,0,0]=(0.56789650708)*x[0]**o+(0.854984079971)*x[0]+(-0.641473205573)*x[1]**o+(-0.791240666887)*x[1]+(0.453562162212)*x[2]**o+(0.849111318036)*x[2]
        arg[4,0,1]=(0.90836228457)*x[0]**o+(-0.666684673376)*x[0]+(0.245928632693)*x[1]**o+(0.313580017685)*x[1]+(0.353126203728)*x[2]**o+(0.467530918033)*x[2]
        arg[4,1,0]=(0.670926585272)*x[0]**o+(0.768555336283)*x[0]+(-0.170656824353)*x[1]**o+(-0.346425800834)*x[1]+(-0.969641886683)*x[2]**o+(0.775335418066)*x[2]
        arg[4,1,1]=(-0.152500967035)*x[0]**o+(0.575400101502)*x[0]+(0.901331752685)*x[1]**o+(-0.672576642993)*x[1]+(0.803176375377)*x[2]**o+(0.202785337179)*x[2]
        arg[5,0,0]=(-0.798851736802)*x[0]**o+(-0.221276048656)*x[0]+(0.773295190339)*x[1]**o+(0.0903112854292)*x[1]+(0.884961364642)*x[2]**o+(0.326429585537)*x[2]
        arg[5,0,1]=(0.482124059007)*x[0]**o+(-0.39227677557)*x[0]+(-0.23074922677)*x[1]**o+(0.986783507045)*x[1]+(0.591759402442)*x[2]**o+(-0.680278617308)*x[2]
        arg[5,1,0]=(-0.857456972644)*x[0]**o+(-0.200661025977)*x[0]+(-0.850031395341)*x[1]**o+(0.830863037447)*x[1]+(-0.323765625026)*x[2]**o+(-0.969463605839)*x[2]
        arg[5,1,1]=(-0.582538216697)*x[0]**o+(-0.495303456972)*x[0]+(0.0134139524699)*x[1]**o+(-0.750144603651)*x[1]+(-0.632380446311)*x[2]**o+(0.38782755017)*x[2]
        ref[0,0,0]=(0.242525071375)*(1+2.*(dim-1.)/(o+1.))+(-1.74732833558)*dim
        ref[0,0,1]=(1.52746771511)*(1+2.*(dim-1.)/(o+1.))+(-0.295920657428)*dim
        ref[0,1,0]=(1.22690064951)*(1+2.*(dim-1.)/(o+1.))+(-0.156829910392)*dim
        ref[0,1,1]=(0.255351951182)*(1+2.*(dim-1.)/(o+1.))+(-1.34566476797)*dim
        ref[1,0,0]=(-0.368142007523)*(1+2.*(dim-1.)/(o+1.))+(-0.43517830631)*dim
        ref[1,0,1]=(-0.602864409266)*(1+2.*(dim-1.)/(o+1.))+(2.53934733026)*dim
        ref[1,1,0]=(-0.309406777332)*(1+2.*(dim-1.)/(o+1.))+(-1.74610493135)*dim
        ref[1,1,1]=(0.515333032348)*(1+2.*(dim-1.)/(o+1.))+(-2.34365173129)*dim
        ref[2,0,0]=(-0.338863451761)*(1+2.*(dim-1.)/(o+1.))+(1.16261676251)*dim
        ref[2,0,1]=(0.731556502498)*(1+2.*(dim-1.)/(o+1.))+(-0.73700082031)*dim
        ref[2,1,0]=(0.132656132299)*(1+2.*(dim-1.)/(o+1.))+(-0.336323196217)*dim
        ref[2,1,1]=(-1.97539539883)*(1+2.*(dim-1.)/(o+1.))+(0.698181001434)*dim
        ref[3,0,0]=(-1.97447169527)*(1+2.*(dim-1.)/(o+1.))+(1.79859205901)*dim
        ref[3,0,1]=(-0.398705337727)*(1+2.*(dim-1.)/(o+1.))+(0.660365348338)*dim
        ref[3,1,0]=(-0.129346064471)*(1+2.*(dim-1.)/(o+1.))+(-1.84164209635)*dim
        ref[3,1,1]=(-0.215276741258)*(1+2.*(dim-1.)/(o+1.))+(-1.15470808005)*dim
        ref[4,0,0]=(0.379985463719)*(1+2.*(dim-1.)/(o+1.))+(0.91285473112)*dim
        ref[4,0,1]=(1.50741712099)*(1+2.*(dim-1.)/(o+1.))+(0.114426262342)*dim
        ref[4,1,0]=(-0.469372125764)*(1+2.*(dim-1.)/(o+1.))+(1.19746495352)*dim
        ref[4,1,1]=(1.55200716103)*(1+2.*(dim-1.)/(o+1.))+(0.105608795688)*dim
        ref[5,0,0]=(0.859404818179)*(1+2.*(dim-1.)/(o+1.))+(0.19546482231)*dim
        ref[5,0,1]=(0.843134234679)*(1+2.*(dim-1.)/(o+1.))+(-0.085771885833)*dim
        ref[5,1,0]=(-2.03125399301)*(1+2.*(dim-1.)/(o+1.))+(-0.339261594369)*dim
        ref[5,1,1]=(-1.20150471054)*(1+2.*(dim-1.)/(o+1.))+(-0.857620510454)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank4(self):
      """
      tests integral of rank 4 Data on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        arg[0,0,0,0]=(-0.161068279134)*x[0]**o+(-0.20016693624)*x[0]+(0.360967926263)*x[1]**o+(0.519774730324)*x[1]
        arg[0,0,0,1]=(-0.470257945181)*x[0]**o+(0.0652334206708)*x[0]+(0.652139991357)*x[1]**o+(-0.592482756578)*x[1]
        arg[0,0,1,0]=(0.143126570256)*x[0]**o+(0.683474249212)*x[0]+(0.839611756722)*x[1]**o+(-0.436612603526)*x[1]
        arg[0,0,1,1]=(-0.82328367976)*x[0]**o+(-0.848145278113)*x[0]+(-0.649046694126)*x[1]**o+(0.7686615034)*x[1]
        arg[0,0,2,0]=(-0.89676527056)*x[0]**o+(-0.466316806084)*x[0]+(-0.454908744166)*x[1]**o+(0.425524454897)*x[1]
        arg[0,0,2,1]=(-0.720978656424)*x[0]**o+(-0.655449460498)*x[0]+(0.466383340682)*x[1]**o+(-0.609615749634)*x[1]
        arg[0,1,0,0]=(-0.358670380794)*x[0]**o+(-0.764939003508)*x[0]+(-0.550898156906)*x[1]**o+(0.803150103728)*x[1]
        arg[0,1,0,1]=(-0.652241076457)*x[0]**o+(-0.614033274928)*x[0]+(-0.308770011731)*x[1]**o+(0.925847136965)*x[1]
        arg[0,1,1,0]=(0.710539026622)*x[0]**o+(-0.345587804078)*x[0]+(-0.170026226995)*x[1]**o+(0.340557574804)*x[1]
        arg[0,1,1,1]=(0.348069554895)*x[0]**o+(-0.633629515001)*x[0]+(-0.199233786202)*x[1]**o+(0.500291488845)*x[1]
        arg[0,1,2,0]=(0.262179751197)*x[0]**o+(-0.55460029497)*x[0]+(-0.354335283745)*x[1]**o+(0.301130103531)*x[1]
        arg[0,1,2,1]=(0.212709964525)*x[0]**o+(0.205907518129)*x[0]+(0.167040570615)*x[1]**o+(-0.0684155270612)*x[1]
        arg[0,2,0,0]=(-0.272141616042)*x[0]**o+(0.187045318075)*x[0]+(-0.364974531949)*x[1]**o+(-0.969702579161)*x[1]
        arg[0,2,0,1]=(-0.997155483171)*x[0]**o+(0.864439328971)*x[0]+(0.887907754657)*x[1]**o+(-0.323021835147)*x[1]
        arg[0,2,1,0]=(-0.544228085563)*x[0]**o+(-0.56752289714)*x[0]+(-0.389234395916)*x[1]**o+(-0.969287386144)*x[1]
        arg[0,2,1,1]=(-0.786136450866)*x[0]**o+(0.589465339989)*x[0]+(-0.411595778776)*x[1]**o+(0.726975706536)*x[1]
        arg[0,2,2,0]=(-0.788299350517)*x[0]**o+(0.697037609691)*x[0]+(-0.391342918209)*x[1]**o+(0.432368662365)*x[1]
        arg[0,2,2,1]=(-0.390536554506)*x[0]**o+(0.547528359725)*x[0]+(0.58957966188)*x[1]**o+(0.146918361066)*x[1]
        arg[0,3,0,0]=(0.176059579799)*x[0]**o+(-0.774041865433)*x[0]+(0.52369869825)*x[1]**o+(-0.812474996262)*x[1]
        arg[0,3,0,1]=(-0.938727595513)*x[0]**o+(0.734277246158)*x[0]+(-0.500452658276)*x[1]**o+(0.591562924701)*x[1]
        arg[0,3,1,0]=(-0.109970490351)*x[0]**o+(-0.463382333915)*x[0]+(0.430588389365)*x[1]**o+(-0.174978181104)*x[1]
        arg[0,3,1,1]=(-0.564486046162)*x[0]**o+(0.325599920584)*x[0]+(-0.177334934858)*x[1]**o+(-0.789210879368)*x[1]
        arg[0,3,2,0]=(-0.409769653157)*x[0]**o+(-0.674668709075)*x[0]+(0.977465383205)*x[1]**o+(-0.178249094926)*x[1]
        arg[0,3,2,1]=(0.112394584736)*x[0]**o+(-0.229767891006)*x[0]+(0.952422132907)*x[1]**o+(-0.133952947363)*x[1]
        arg[0,4,0,0]=(0.927747625444)*x[0]**o+(-0.147390883234)*x[0]+(-0.0523168758823)*x[1]**o+(0.698643952882)*x[1]
        arg[0,4,0,1]=(0.986627759385)*x[0]**o+(-0.0486375625834)*x[0]+(0.27598075052)*x[1]**o+(0.970272004318)*x[1]
        arg[0,4,1,0]=(-0.461643441984)*x[0]**o+(-0.815261492927)*x[0]+(-0.0806771439228)*x[1]**o+(0.877302861263)*x[1]
        arg[0,4,1,1]=(-0.108403920958)*x[0]**o+(0.651566419087)*x[0]+(-0.629972296369)*x[1]**o+(-0.235413582155)*x[1]
        arg[0,4,2,0]=(-0.177527213798)*x[0]**o+(0.987202381273)*x[0]+(-0.15522298966)*x[1]**o+(0.234268561378)*x[1]
        arg[0,4,2,1]=(0.521954293269)*x[0]**o+(-0.776898814469)*x[0]+(0.783085515988)*x[1]**o+(-0.469921197982)*x[1]
        arg[1,0,0,0]=(0.858157489751)*x[0]**o+(0.444181035326)*x[0]+(0.3426806552)*x[1]**o+(-0.219249318925)*x[1]
        arg[1,0,0,1]=(0.405857394564)*x[0]**o+(0.178361294933)*x[0]+(0.0879917927852)*x[1]**o+(-0.0689581267347)*x[1]
        arg[1,0,1,0]=(0.63175565694)*x[0]**o+(0.0810462363705)*x[0]+(0.0350399072162)*x[1]**o+(-0.269906961338)*x[1]
        arg[1,0,1,1]=(-0.273973794035)*x[0]**o+(0.824519533391)*x[0]+(0.596785621044)*x[1]**o+(-0.243043370261)*x[1]
        arg[1,0,2,0]=(-0.959246118913)*x[0]**o+(0.3570024829)*x[0]+(-0.260098162669)*x[1]**o+(0.341044029457)*x[1]
        arg[1,0,2,1]=(0.966671029776)*x[0]**o+(0.598898116395)*x[0]+(0.392360541053)*x[1]**o+(-0.322723646751)*x[1]
        arg[1,1,0,0]=(-0.658365963559)*x[0]**o+(-0.674107595463)*x[0]+(-0.0103911551403)*x[1]**o+(0.858879803218)*x[1]
        arg[1,1,0,1]=(0.340459206102)*x[0]**o+(-0.681434505015)*x[0]+(0.917412924129)*x[1]**o+(0.703699154381)*x[1]
        arg[1,1,1,0]=(0.0473840209756)*x[0]**o+(-0.787070458106)*x[0]+(0.11979244907)*x[1]**o+(0.541238063489)*x[1]
        arg[1,1,1,1]=(0.184943768213)*x[0]**o+(-0.942961758488)*x[0]+(-0.844785429901)*x[1]**o+(-0.0109779269179)*x[1]
        arg[1,1,2,0]=(0.921469748172)*x[0]**o+(-0.122684636542)*x[0]+(0.149552532716)*x[1]**o+(-0.0102623511735)*x[1]
        arg[1,1,2,1]=(-0.720466710917)*x[0]**o+(-0.755841224189)*x[0]+(0.372635950892)*x[1]**o+(0.271604227274)*x[1]
        arg[1,2,0,0]=(-0.548557725319)*x[0]**o+(0.745340396995)*x[0]+(0.461556064961)*x[1]**o+(0.696460891668)*x[1]
        arg[1,2,0,1]=(-0.738939238644)*x[0]**o+(0.344829756443)*x[0]+(0.272341172248)*x[1]**o+(0.768063359753)*x[1]
        arg[1,2,1,0]=(-0.213161389496)*x[0]**o+(0.458730188555)*x[0]+(-0.783049463147)*x[1]**o+(0.785273989074)*x[1]
        arg[1,2,1,1]=(0.189092239774)*x[0]**o+(-0.0649917239139)*x[0]+(0.0171281685468)*x[1]**o+(0.397065812267)*x[1]
        arg[1,2,2,0]=(-0.809663598067)*x[0]**o+(0.587214021528)*x[0]+(0.912814601228)*x[1]**o+(0.860792008172)*x[1]
        arg[1,2,2,1]=(0.487927123118)*x[0]**o+(-0.760831928598)*x[0]+(-0.770486346056)*x[1]**o+(-0.715077841171)*x[1]
        arg[1,3,0,0]=(0.0202747853647)*x[0]**o+(0.0666642857036)*x[0]+(-0.717756666598)*x[1]**o+(-0.481009806467)*x[1]
        arg[1,3,0,1]=(-0.251164550146)*x[0]**o+(0.0519935051437)*x[0]+(-0.703400138413)*x[1]**o+(-0.990329903234)*x[1]
        arg[1,3,1,0]=(0.970793375169)*x[0]**o+(-0.365675269998)*x[0]+(0.77706784549)*x[1]**o+(-0.237785293481)*x[1]
        arg[1,3,1,1]=(-0.162786953006)*x[0]**o+(0.985402077919)*x[0]+(0.739483663725)*x[1]**o+(0.559554321882)*x[1]
        arg[1,3,2,0]=(0.184701463663)*x[0]**o+(0.946173117295)*x[0]+(-0.451420616712)*x[1]**o+(0.244506578089)*x[1]
        arg[1,3,2,1]=(0.894307549357)*x[0]**o+(0.474983970689)*x[0]+(0.626273940678)*x[1]**o+(0.107985157225)*x[1]
        arg[1,4,0,0]=(-0.180994231135)*x[0]**o+(-0.899808624585)*x[0]+(0.395935138342)*x[1]**o+(-0.387788497342)*x[1]
        arg[1,4,0,1]=(0.0713358452056)*x[0]**o+(0.505276043139)*x[0]+(-0.635565006093)*x[1]**o+(0.956951408169)*x[1]
        arg[1,4,1,0]=(0.355608577722)*x[0]**o+(0.58482326687)*x[0]+(0.654137690443)*x[1]**o+(-0.849016362888)*x[1]
        arg[1,4,1,1]=(-0.275966957458)*x[0]**o+(0.856095068377)*x[0]+(-0.760077050664)*x[1]**o+(-0.904159432699)*x[1]
        arg[1,4,2,0]=(-0.0637454805096)*x[0]**o+(-0.833057387633)*x[0]+(0.279008913569)*x[1]**o+(-0.093356790193)*x[1]
        arg[1,4,2,1]=(0.165869666472)*x[0]**o+(-0.4554345089)*x[0]+(0.697149094381)*x[1]**o+(0.244460087482)*x[1]
        arg[2,0,0,0]=(0.388188511348)*x[0]**o+(-0.00552398040882)*x[0]+(0.0679336322536)*x[1]**o+(-0.281132460216)*x[1]
        arg[2,0,0,1]=(0.968019690109)*x[0]**o+(0.525335591339)*x[0]+(0.660069087546)*x[1]**o+(-0.795388487336)*x[1]
        arg[2,0,1,0]=(-0.444010450901)*x[0]**o+(-0.498771184141)*x[0]+(-0.394999176962)*x[1]**o+(-0.579230388911)*x[1]
        arg[2,0,1,1]=(-0.96211072138)*x[0]**o+(-0.0977519108256)*x[0]+(0.206179333903)*x[1]**o+(-0.206155312014)*x[1]
        arg[2,0,2,0]=(0.131307090845)*x[0]**o+(0.179396626286)*x[0]+(-0.436118623166)*x[1]**o+(0.829727875456)*x[1]
        arg[2,0,2,1]=(-0.206336435516)*x[0]**o+(0.175744210813)*x[0]+(0.818204199205)*x[1]**o+(0.433292546177)*x[1]
        arg[2,1,0,0]=(0.920189410059)*x[0]**o+(-0.629286762249)*x[0]+(-0.14444305096)*x[1]**o+(0.564022686795)*x[1]
        arg[2,1,0,1]=(0.43550707444)*x[0]**o+(-0.0203691253164)*x[0]+(0.632472807002)*x[1]**o+(0.722390213081)*x[1]
        arg[2,1,1,0]=(0.827084071535)*x[0]**o+(0.864139638834)*x[0]+(0.795893591186)*x[1]**o+(0.0492667773115)*x[1]
        arg[2,1,1,1]=(0.863813787478)*x[0]**o+(0.268800194137)*x[0]+(-0.567637801205)*x[1]**o+(0.931221493563)*x[1]
        arg[2,1,2,0]=(-0.613463525024)*x[0]**o+(-0.0688030533166)*x[0]+(-0.786395692457)*x[1]**o+(0.862189706883)*x[1]
        arg[2,1,2,1]=(0.315306814851)*x[0]**o+(0.76879509223)*x[0]+(0.451451342779)*x[1]**o+(-0.478873421268)*x[1]
        arg[2,2,0,0]=(-0.785116047841)*x[0]**o+(0.347614087057)*x[0]+(-0.0106716042015)*x[1]**o+(-0.350898768187)*x[1]
        arg[2,2,0,1]=(-0.987778931187)*x[0]**o+(-0.380379654757)*x[0]+(-0.383604999655)*x[1]**o+(0.022203803886)*x[1]
        arg[2,2,1,0]=(0.508327115069)*x[0]**o+(0.443678958632)*x[0]+(-0.320069211602)*x[1]**o+(-0.778140184171)*x[1]
        arg[2,2,1,1]=(-0.532025577107)*x[0]**o+(-0.0475799386434)*x[0]+(0.650417175266)*x[1]**o+(0.33739661369)*x[1]
        arg[2,2,2,0]=(-0.406412097333)*x[0]**o+(0.0510500610605)*x[0]+(-0.25755890225)*x[1]**o+(-0.350209779392)*x[1]
        arg[2,2,2,1]=(0.401361048956)*x[0]**o+(-0.594793585281)*x[0]+(0.155257431052)*x[1]**o+(-0.674885216524)*x[1]
        arg[2,3,0,0]=(0.352606222236)*x[0]**o+(0.64529574473)*x[0]+(-0.838258091233)*x[1]**o+(0.527911574536)*x[1]
        arg[2,3,0,1]=(-0.62652610251)*x[0]**o+(0.879664434218)*x[0]+(0.730845316192)*x[1]**o+(0.0318659384791)*x[1]
        arg[2,3,1,0]=(-0.311152313179)*x[0]**o+(-0.333154399786)*x[0]+(0.743640958069)*x[1]**o+(0.313394944493)*x[1]
        arg[2,3,1,1]=(-0.728968478378)*x[0]**o+(0.500229380574)*x[0]+(0.756628133572)*x[1]**o+(0.270533146994)*x[1]
        arg[2,3,2,0]=(-0.219775843272)*x[0]**o+(0.882943220401)*x[0]+(0.79341547188)*x[1]**o+(0.503486054551)*x[1]
        arg[2,3,2,1]=(-0.850285149353)*x[0]**o+(0.78820884302)*x[0]+(0.348024963962)*x[1]**o+(0.83743616537)*x[1]
        arg[2,4,0,0]=(0.868649421127)*x[0]**o+(-0.622361223999)*x[0]+(0.836321867516)*x[1]**o+(0.423582994131)*x[1]
        arg[2,4,0,1]=(0.140963355402)*x[0]**o+(0.190496469385)*x[0]+(0.765311231744)*x[1]**o+(0.368696267095)*x[1]
        arg[2,4,1,0]=(-0.680854080076)*x[0]**o+(-0.729641193897)*x[0]+(-0.843347646585)*x[1]**o+(-0.0993036180842)*x[1]
        arg[2,4,1,1]=(-0.130039648356)*x[0]**o+(0.534197300899)*x[0]+(-0.142588861186)*x[1]**o+(-0.308376644249)*x[1]
        arg[2,4,2,0]=(0.331212825202)*x[0]**o+(0.466247840288)*x[0]+(-0.201714210846)*x[1]**o+(0.80521345364)*x[1]
        arg[2,4,2,1]=(0.454299813247)*x[0]**o+(0.646694941459)*x[0]+(-0.929854585458)*x[1]**o+(-0.365252296746)*x[1]
        arg[3,0,0,0]=(-0.623915395721)*x[0]**o+(-0.789827882448)*x[0]+(-0.862568457189)*x[1]**o+(-0.141799958679)*x[1]
        arg[3,0,0,1]=(0.197040281353)*x[0]**o+(0.155785418081)*x[0]+(-0.587898788817)*x[1]**o+(0.700612189317)*x[1]
        arg[3,0,1,0]=(-0.899654954269)*x[0]**o+(-0.801426785588)*x[0]+(-0.528678401571)*x[1]**o+(-0.224185097206)*x[1]
        arg[3,0,1,1]=(0.291298188198)*x[0]**o+(0.786698690825)*x[0]+(0.925634940697)*x[1]**o+(0.617661715197)*x[1]
        arg[3,0,2,0]=(0.788272473274)*x[0]**o+(-0.0400044705274)*x[0]+(-0.669092974469)*x[1]**o+(-0.063697728491)*x[1]
        arg[3,0,2,1]=(-0.0612398565521)*x[0]**o+(-0.884311783009)*x[0]+(0.444689078628)*x[1]**o+(0.302741981611)*x[1]
        arg[3,1,0,0]=(0.483756249722)*x[0]**o+(-0.214185139167)*x[0]+(0.526387014144)*x[1]**o+(-0.172619975799)*x[1]
        arg[3,1,0,1]=(0.0685836489777)*x[0]**o+(-0.425861440836)*x[0]+(0.94669899123)*x[1]**o+(-0.649068988929)*x[1]
        arg[3,1,1,0]=(0.935225203148)*x[0]**o+(-0.649378812349)*x[0]+(-0.791553458732)*x[1]**o+(0.776307740809)*x[1]
        arg[3,1,1,1]=(-0.45319945403)*x[0]**o+(-0.473310763902)*x[0]+(-0.268480872039)*x[1]**o+(0.847382076174)*x[1]
        arg[3,1,2,0]=(-0.573022434421)*x[0]**o+(0.584143976397)*x[0]+(0.131688457808)*x[1]**o+(-0.628894863435)*x[1]
        arg[3,1,2,1]=(0.628566928363)*x[0]**o+(-0.868564569338)*x[0]+(-0.482273553287)*x[1]**o+(0.679652796519)*x[1]
        arg[3,2,0,0]=(-0.64046509323)*x[0]**o+(-0.108928537952)*x[0]+(-0.512288515479)*x[1]**o+(0.683072290041)*x[1]
        arg[3,2,0,1]=(0.472589328326)*x[0]**o+(0.219741638671)*x[0]+(0.136368535311)*x[1]**o+(-0.525113003647)*x[1]
        arg[3,2,1,0]=(-0.45731256203)*x[0]**o+(0.0854410475382)*x[0]+(0.257725845542)*x[1]**o+(0.430430185351)*x[1]
        arg[3,2,1,1]=(0.948699678725)*x[0]**o+(-0.00538658505415)*x[0]+(0.921898600708)*x[1]**o+(-0.448522286404)*x[1]
        arg[3,2,2,0]=(-0.0863353766204)*x[0]**o+(-0.652417730681)*x[0]+(0.755871413533)*x[1]**o+(0.206065623655)*x[1]
        arg[3,2,2,1]=(0.263075920527)*x[0]**o+(0.94162407402)*x[0]+(0.750015836722)*x[1]**o+(-0.815154267718)*x[1]
        arg[3,3,0,0]=(-0.606984299829)*x[0]**o+(0.120704522147)*x[0]+(-0.324211323778)*x[1]**o+(-0.123647081263)*x[1]
        arg[3,3,0,1]=(0.838995936232)*x[0]**o+(0.854147779134)*x[0]+(-0.192271552952)*x[1]**o+(0.0281108740635)*x[1]
        arg[3,3,1,0]=(-0.0343159887291)*x[0]**o+(0.242432659911)*x[0]+(-0.37299830191)*x[1]**o+(-0.456792158625)*x[1]
        arg[3,3,1,1]=(-0.305277044867)*x[0]**o+(-0.930386582491)*x[0]+(-0.979921197617)*x[1]**o+(-0.507736755766)*x[1]
        arg[3,3,2,0]=(0.548675744195)*x[0]**o+(0.755596482286)*x[0]+(0.0722267054576)*x[1]**o+(-0.393768047035)*x[1]
        arg[3,3,2,1]=(0.74781993508)*x[0]**o+(-0.655033306209)*x[0]+(0.480653847364)*x[1]**o+(-0.169400714085)*x[1]
        arg[3,4,0,0]=(0.269186769422)*x[0]**o+(-0.0786214595922)*x[0]+(-0.975508552024)*x[1]**o+(-0.798249509899)*x[1]
        arg[3,4,0,1]=(-0.599201275441)*x[0]**o+(0.473490319504)*x[0]+(0.0277044125859)*x[1]**o+(0.818107843614)*x[1]
        arg[3,4,1,0]=(-0.46272642862)*x[0]**o+(0.740282148493)*x[0]+(0.35998379499)*x[1]**o+(-0.938681489717)*x[1]
        arg[3,4,1,1]=(0.187067286301)*x[0]**o+(-0.284171344521)*x[0]+(0.612229918327)*x[1]**o+(0.3917984881)*x[1]
        arg[3,4,2,0]=(-0.535270910341)*x[0]**o+(-0.00723888213927)*x[0]+(-0.776984534618)*x[1]**o+(-0.0899536033273)*x[1]
        arg[3,4,2,1]=(-0.382290376155)*x[0]**o+(-0.0593490177599)*x[0]+(-0.19716515678)*x[1]**o+(-0.268991693613)*x[1]
        ref[0,0,0,0]=(0.19989964713)*(1+2.*(dim-1.)/(o+1.))+(0.319607794084)*dim
        ref[0,0,0,1]=(0.181882046176)*(1+2.*(dim-1.)/(o+1.))+(-0.527249335907)*dim
        ref[0,0,1,0]=(0.982738326978)*(1+2.*(dim-1.)/(o+1.))+(0.246861645686)*dim
        ref[0,0,1,1]=(-1.47233037389)*(1+2.*(dim-1.)/(o+1.))+(-0.0794837747136)*dim
        ref[0,0,2,0]=(-1.35167401473)*(1+2.*(dim-1.)/(o+1.))+(-0.0407923511861)*dim
        ref[0,0,2,1]=(-0.254595315742)*(1+2.*(dim-1.)/(o+1.))+(-1.26506521013)*dim
        ref[0,1,0,0]=(-0.909568537701)*(1+2.*(dim-1.)/(o+1.))+(0.0382111002203)*dim
        ref[0,1,0,1]=(-0.961011088187)*(1+2.*(dim-1.)/(o+1.))+(0.311813862038)*dim
        ref[0,1,1,0]=(0.540512799627)*(1+2.*(dim-1.)/(o+1.))+(-0.00503022927381)*dim
        ref[0,1,1,1]=(0.148835768693)*(1+2.*(dim-1.)/(o+1.))+(-0.133338026157)*dim
        ref[0,1,2,0]=(-0.0921555325482)*(1+2.*(dim-1.)/(o+1.))+(-0.253470191439)*dim
        ref[0,1,2,1]=(0.37975053514)*(1+2.*(dim-1.)/(o+1.))+(0.137491991068)*dim
        ref[0,2,0,0]=(-0.637116147991)*(1+2.*(dim-1.)/(o+1.))+(-0.782657261086)*dim
        ref[0,2,0,1]=(-0.109247728514)*(1+2.*(dim-1.)/(o+1.))+(0.541417493824)*dim
        ref[0,2,1,0]=(-0.933462481479)*(1+2.*(dim-1.)/(o+1.))+(-1.53681028328)*dim
        ref[0,2,1,1]=(-1.19773222964)*(1+2.*(dim-1.)/(o+1.))+(1.31644104653)*dim
        ref[0,2,2,0]=(-1.17964226873)*(1+2.*(dim-1.)/(o+1.))+(1.12940627206)*dim
        ref[0,2,2,1]=(0.199043107375)*(1+2.*(dim-1.)/(o+1.))+(0.694446720792)*dim
        ref[0,3,0,0]=(0.699758278049)*(1+2.*(dim-1.)/(o+1.))+(-1.5865168617)*dim
        ref[0,3,0,1]=(-1.43918025379)*(1+2.*(dim-1.)/(o+1.))+(1.32584017086)*dim
        ref[0,3,1,0]=(0.320617899014)*(1+2.*(dim-1.)/(o+1.))+(-0.63836051502)*dim
        ref[0,3,1,1]=(-0.741820981021)*(1+2.*(dim-1.)/(o+1.))+(-0.463610958784)*dim
        ref[0,3,2,0]=(0.567695730048)*(1+2.*(dim-1.)/(o+1.))+(-0.852917804)*dim
        ref[0,3,2,1]=(1.06481671764)*(1+2.*(dim-1.)/(o+1.))+(-0.363720838369)*dim
        ref[0,4,0,0]=(0.875430749562)*(1+2.*(dim-1.)/(o+1.))+(0.551253069648)*dim
        ref[0,4,0,1]=(1.26260850991)*(1+2.*(dim-1.)/(o+1.))+(0.921634441735)*dim
        ref[0,4,1,0]=(-0.542320585907)*(1+2.*(dim-1.)/(o+1.))+(0.0620413683361)*dim
        ref[0,4,1,1]=(-0.738376217327)*(1+2.*(dim-1.)/(o+1.))+(0.416152836932)*dim
        ref[0,4,2,0]=(-0.332750203458)*(1+2.*(dim-1.)/(o+1.))+(1.22147094265)*dim
        ref[0,4,2,1]=(1.30503980926)*(1+2.*(dim-1.)/(o+1.))+(-1.24682001245)*dim
        ref[1,0,0,0]=(1.20083814495)*(1+2.*(dim-1.)/(o+1.))+(0.224931716401)*dim
        ref[1,0,0,1]=(0.493849187349)*(1+2.*(dim-1.)/(o+1.))+(0.109403168198)*dim
        ref[1,0,1,0]=(0.666795564157)*(1+2.*(dim-1.)/(o+1.))+(-0.188860724968)*dim
        ref[1,0,1,1]=(0.322811827009)*(1+2.*(dim-1.)/(o+1.))+(0.58147616313)*dim
        ref[1,0,2,0]=(-1.21934428158)*(1+2.*(dim-1.)/(o+1.))+(0.698046512357)*dim
        ref[1,0,2,1]=(1.35903157083)*(1+2.*(dim-1.)/(o+1.))+(0.276174469644)*dim
        ref[1,1,0,0]=(-0.668757118699)*(1+2.*(dim-1.)/(o+1.))+(0.184772207756)*dim
        ref[1,1,0,1]=(1.25787213023)*(1+2.*(dim-1.)/(o+1.))+(0.0222646493659)*dim
        ref[1,1,1,0]=(0.167176470046)*(1+2.*(dim-1.)/(o+1.))+(-0.245832394617)*dim
        ref[1,1,1,1]=(-0.659841661688)*(1+2.*(dim-1.)/(o+1.))+(-0.953939685406)*dim
        ref[1,1,2,0]=(1.07102228089)*(1+2.*(dim-1.)/(o+1.))+(-0.132946987715)*dim
        ref[1,1,2,1]=(-0.347830760025)*(1+2.*(dim-1.)/(o+1.))+(-0.484236996915)*dim
        ref[1,2,0,0]=(-0.087001660358)*(1+2.*(dim-1.)/(o+1.))+(1.44180128866)*dim
        ref[1,2,0,1]=(-0.466598066397)*(1+2.*(dim-1.)/(o+1.))+(1.1128931162)*dim
        ref[1,2,1,0]=(-0.996210852643)*(1+2.*(dim-1.)/(o+1.))+(1.24400417763)*dim
        ref[1,2,1,1]=(0.206220408321)*(1+2.*(dim-1.)/(o+1.))+(0.332074088353)*dim
        ref[1,2,2,0]=(0.103151003161)*(1+2.*(dim-1.)/(o+1.))+(1.4480060297)*dim
        ref[1,2,2,1]=(-0.282559222939)*(1+2.*(dim-1.)/(o+1.))+(-1.47590976977)*dim
        ref[1,3,0,0]=(-0.697481881234)*(1+2.*(dim-1.)/(o+1.))+(-0.414345520764)*dim
        ref[1,3,0,1]=(-0.954564688559)*(1+2.*(dim-1.)/(o+1.))+(-0.93833639809)*dim
        ref[1,3,1,0]=(1.74786122066)*(1+2.*(dim-1.)/(o+1.))+(-0.603460563479)*dim
        ref[1,3,1,1]=(0.576696710719)*(1+2.*(dim-1.)/(o+1.))+(1.5449563998)*dim
        ref[1,3,2,0]=(-0.26671915305)*(1+2.*(dim-1.)/(o+1.))+(1.19067969538)*dim
        ref[1,3,2,1]=(1.52058149003)*(1+2.*(dim-1.)/(o+1.))+(0.582969127914)*dim
        ref[1,4,0,0]=(0.214940907208)*(1+2.*(dim-1.)/(o+1.))+(-1.28759712193)*dim
        ref[1,4,0,1]=(-0.564229160887)*(1+2.*(dim-1.)/(o+1.))+(1.46222745131)*dim
        ref[1,4,1,0]=(1.00974626816)*(1+2.*(dim-1.)/(o+1.))+(-0.264193096019)*dim
        ref[1,4,1,1]=(-1.03604400812)*(1+2.*(dim-1.)/(o+1.))+(-0.048064364322)*dim
        ref[1,4,2,0]=(0.215263433059)*(1+2.*(dim-1.)/(o+1.))+(-0.926414177826)*dim
        ref[1,4,2,1]=(0.863018760853)*(1+2.*(dim-1.)/(o+1.))+(-0.210974421418)*dim
        ref[2,0,0,0]=(0.456122143601)*(1+2.*(dim-1.)/(o+1.))+(-0.286656440625)*dim
        ref[2,0,0,1]=(1.62808877765)*(1+2.*(dim-1.)/(o+1.))+(-0.270052895997)*dim
        ref[2,0,1,0]=(-0.839009627862)*(1+2.*(dim-1.)/(o+1.))+(-1.07800157305)*dim
        ref[2,0,1,1]=(-0.755931387477)*(1+2.*(dim-1.)/(o+1.))+(-0.30390722284)*dim
        ref[2,0,2,0]=(-0.304811532322)*(1+2.*(dim-1.)/(o+1.))+(1.00912450174)*dim
        ref[2,0,2,1]=(0.61186776369)*(1+2.*(dim-1.)/(o+1.))+(0.60903675699)*dim
        ref[2,1,0,0]=(0.775746359099)*(1+2.*(dim-1.)/(o+1.))+(-0.0652640754538)*dim
        ref[2,1,0,1]=(1.06797988144)*(1+2.*(dim-1.)/(o+1.))+(0.702021087765)*dim
        ref[2,1,1,0]=(1.62297766272)*(1+2.*(dim-1.)/(o+1.))+(0.913406416145)*dim
        ref[2,1,1,1]=(0.296175986273)*(1+2.*(dim-1.)/(o+1.))+(1.2000216877)*dim
        ref[2,1,2,0]=(-1.39985921748)*(1+2.*(dim-1.)/(o+1.))+(0.793386653567)*dim
        ref[2,1,2,1]=(0.76675815763)*(1+2.*(dim-1.)/(o+1.))+(0.289921670962)*dim
        ref[2,2,0,0]=(-0.795787652043)*(1+2.*(dim-1.)/(o+1.))+(-0.0032846811301)*dim
        ref[2,2,0,1]=(-1.37138393084)*(1+2.*(dim-1.)/(o+1.))+(-0.358175850871)*dim
        ref[2,2,1,0]=(0.188257903467)*(1+2.*(dim-1.)/(o+1.))+(-0.334461225539)*dim
        ref[2,2,1,1]=(0.118391598158)*(1+2.*(dim-1.)/(o+1.))+(0.289816675047)*dim
        ref[2,2,2,0]=(-0.663970999582)*(1+2.*(dim-1.)/(o+1.))+(-0.299159718332)*dim
        ref[2,2,2,1]=(0.556618480008)*(1+2.*(dim-1.)/(o+1.))+(-1.2696788018)*dim
        ref[2,3,0,0]=(-0.485651868997)*(1+2.*(dim-1.)/(o+1.))+(1.17320731927)*dim
        ref[2,3,0,1]=(0.104319213682)*(1+2.*(dim-1.)/(o+1.))+(0.911530372697)*dim
        ref[2,3,1,0]=(0.43248864489)*(1+2.*(dim-1.)/(o+1.))+(-0.0197594552935)*dim
        ref[2,3,1,1]=(0.0276596551936)*(1+2.*(dim-1.)/(o+1.))+(0.770762527568)*dim
        ref[2,3,2,0]=(0.573639628608)*(1+2.*(dim-1.)/(o+1.))+(1.38642927495)*dim
        ref[2,3,2,1]=(-0.502260185391)*(1+2.*(dim-1.)/(o+1.))+(1.62564500839)*dim
        ref[2,4,0,0]=(1.70497128864)*(1+2.*(dim-1.)/(o+1.))+(-0.198778229868)*dim
        ref[2,4,0,1]=(0.906274587146)*(1+2.*(dim-1.)/(o+1.))+(0.55919273648)*dim
        ref[2,4,1,0]=(-1.52420172666)*(1+2.*(dim-1.)/(o+1.))+(-0.828944811982)*dim
        ref[2,4,1,1]=(-0.272628509541)*(1+2.*(dim-1.)/(o+1.))+(0.225820656649)*dim
        ref[2,4,2,0]=(0.129498614356)*(1+2.*(dim-1.)/(o+1.))+(1.27146129393)*dim
        ref[2,4,2,1]=(-0.475554772211)*(1+2.*(dim-1.)/(o+1.))+(0.281442644713)*dim
        ref[3,0,0,0]=(-1.48648385291)*(1+2.*(dim-1.)/(o+1.))+(-0.931627841127)*dim
        ref[3,0,0,1]=(-0.390858507465)*(1+2.*(dim-1.)/(o+1.))+(0.856397607398)*dim
        ref[3,0,1,0]=(-1.42833335584)*(1+2.*(dim-1.)/(o+1.))+(-1.02561188279)*dim
        ref[3,0,1,1]=(1.2169331289)*(1+2.*(dim-1.)/(o+1.))+(1.40436040602)*dim
        ref[3,0,2,0]=(0.119179498805)*(1+2.*(dim-1.)/(o+1.))+(-0.103702199018)*dim
        ref[3,0,2,1]=(0.383449222076)*(1+2.*(dim-1.)/(o+1.))+(-0.581569801398)*dim
        ref[3,1,0,0]=(1.01014326387)*(1+2.*(dim-1.)/(o+1.))+(-0.386805114966)*dim
        ref[3,1,0,1]=(1.01528264021)*(1+2.*(dim-1.)/(o+1.))+(-1.07493042977)*dim
        ref[3,1,1,0]=(0.143671744416)*(1+2.*(dim-1.)/(o+1.))+(0.12692892846)*dim
        ref[3,1,1,1]=(-0.721680326068)*(1+2.*(dim-1.)/(o+1.))+(0.374071312272)*dim
        ref[3,1,2,0]=(-0.441333976613)*(1+2.*(dim-1.)/(o+1.))+(-0.0447508870373)*dim
        ref[3,1,2,1]=(0.146293375075)*(1+2.*(dim-1.)/(o+1.))+(-0.188911772819)*dim
        ref[3,2,0,0]=(-1.15275360871)*(1+2.*(dim-1.)/(o+1.))+(0.57414375209)*dim
        ref[3,2,0,1]=(0.608957863637)*(1+2.*(dim-1.)/(o+1.))+(-0.305371364976)*dim
        ref[3,2,1,0]=(-0.199586716488)*(1+2.*(dim-1.)/(o+1.))+(0.515871232889)*dim
        ref[3,2,1,1]=(1.87059827943)*(1+2.*(dim-1.)/(o+1.))+(-0.453908871458)*dim
        ref[3,2,2,0]=(0.669536036913)*(1+2.*(dim-1.)/(o+1.))+(-0.446352107026)*dim
        ref[3,2,2,1]=(1.01309175725)*(1+2.*(dim-1.)/(o+1.))+(0.126469806303)*dim
        ref[3,3,0,0]=(-0.931195623607)*(1+2.*(dim-1.)/(o+1.))+(-0.00294255911617)*dim
        ref[3,3,0,1]=(0.64672438328)*(1+2.*(dim-1.)/(o+1.))+(0.882258653197)*dim
        ref[3,3,1,0]=(-0.407314290639)*(1+2.*(dim-1.)/(o+1.))+(-0.214359498715)*dim
        ref[3,3,1,1]=(-1.28519824248)*(1+2.*(dim-1.)/(o+1.))+(-1.43812333826)*dim
        ref[3,3,2,0]=(0.620902449652)*(1+2.*(dim-1.)/(o+1.))+(0.361828435252)*dim
        ref[3,3,2,1]=(1.22847378244)*(1+2.*(dim-1.)/(o+1.))+(-0.824434020294)*dim
        ref[3,4,0,0]=(-0.706321782603)*(1+2.*(dim-1.)/(o+1.))+(-0.876870969492)*dim
        ref[3,4,0,1]=(-0.571496862855)*(1+2.*(dim-1.)/(o+1.))+(1.29159816312)*dim
        ref[3,4,1,0]=(-0.102742633631)*(1+2.*(dim-1.)/(o+1.))+(-0.198399341224)*dim
        ref[3,4,1,1]=(0.799297204627)*(1+2.*(dim-1.)/(o+1.))+(0.107627143578)*dim
        ref[3,4,2,0]=(-1.31225544496)*(1+2.*(dim-1.)/(o+1.))+(-0.0971924854666)*dim
        ref[3,4,2,1]=(-0.579455532935)*(1+2.*(dim-1.)/(o+1.))+(-0.328340711373)*dim
      else:
        arg[0,0,0,0]=(-0.727570048979)*x[0]**o+(-0.769770353069)*x[0]+(-0.532381054852)*x[1]**o+(0.590589994369)*x[1]+(0.775348726284)*x[2]**o+(-0.840239939796)*x[2]
        arg[0,0,0,1]=(-0.441856577751)*x[0]**o+(-0.39234308235)*x[0]+(-0.834378647798)*x[1]**o+(0.84269918161)*x[1]+(0.441102591728)*x[2]**o+(0.166331058665)*x[2]
        arg[0,0,1,0]=(0.391773069521)*x[0]**o+(0.496759920942)*x[0]+(0.555386226902)*x[1]**o+(0.980197783628)*x[1]+(0.549622021652)*x[2]**o+(-0.596179893504)*x[2]
        arg[0,0,1,1]=(-0.104197387892)*x[0]**o+(-0.64636505375)*x[0]+(0.377612339028)*x[1]**o+(-0.215111619231)*x[1]+(0.637249692625)*x[2]**o+(-0.0998227822351)*x[2]
        arg[0,0,2,0]=(0.780693043561)*x[0]**o+(0.571405344342)*x[0]+(-0.793680154472)*x[1]**o+(-0.0481469136436)*x[1]+(0.959561051293)*x[2]**o+(0.0756152497582)*x[2]
        arg[0,0,2,1]=(0.88678762282)*x[0]**o+(0.254753272214)*x[0]+(0.344376798516)*x[1]**o+(-0.908135902735)*x[1]+(-0.3584387726)*x[2]**o+(-0.890225037617)*x[2]
        arg[0,1,0,0]=(0.680262917473)*x[0]**o+(0.65860357305)*x[0]+(0.859832322473)*x[1]**o+(0.326105580592)*x[1]+(0.132170297876)*x[2]**o+(-0.928606849775)*x[2]
        arg[0,1,0,1]=(-0.500503822293)*x[0]**o+(-0.456331012174)*x[0]+(0.900478846187)*x[1]**o+(0.350542503874)*x[1]+(0.755849038617)*x[2]**o+(-0.514240007142)*x[2]
        arg[0,1,1,0]=(0.124404217291)*x[0]**o+(-0.499651000805)*x[0]+(-0.968178694708)*x[1]**o+(0.306780500818)*x[1]+(-0.946470794819)*x[2]**o+(0.919424575523)*x[2]
        arg[0,1,1,1]=(0.722520017474)*x[0]**o+(-0.718216593718)*x[0]+(0.801196051455)*x[1]**o+(-0.791951278546)*x[1]+(-0.675975417529)*x[2]**o+(0.355069676047)*x[2]
        arg[0,1,2,0]=(-0.137017428857)*x[0]**o+(-0.760267879487)*x[0]+(0.625869259413)*x[1]**o+(-0.697626713352)*x[1]+(-0.467876751423)*x[2]**o+(0.456475454123)*x[2]
        arg[0,1,2,1]=(0.646536442797)*x[0]**o+(0.0981671744166)*x[0]+(0.477147756414)*x[1]**o+(-0.23211823966)*x[1]+(-0.553808167147)*x[2]**o+(0.405402676302)*x[2]
        arg[0,2,0,0]=(-0.0604502966371)*x[0]**o+(0.297995946813)*x[0]+(0.837045433858)*x[1]**o+(-0.270340963575)*x[1]+(-0.58462190484)*x[2]**o+(0.292953993287)*x[2]
        arg[0,2,0,1]=(-0.463109088268)*x[0]**o+(0.346260157368)*x[0]+(0.0228615151352)*x[1]**o+(0.905507471559)*x[1]+(-0.423777553665)*x[2]**o+(-0.45964535654)*x[2]
        arg[0,2,1,0]=(-0.50761294542)*x[0]**o+(-0.465260578419)*x[0]+(-0.729953382373)*x[1]**o+(0.903666219263)*x[1]+(-0.756794344711)*x[2]**o+(0.817263008431)*x[2]
        arg[0,2,1,1]=(-0.551253781393)*x[0]**o+(-0.911481117802)*x[0]+(0.180349330394)*x[1]**o+(0.419118491074)*x[1]+(-0.691917875703)*x[2]**o+(-0.835365504347)*x[2]
        arg[0,2,2,0]=(0.32567932803)*x[0]**o+(0.413167839883)*x[0]+(0.781320691089)*x[1]**o+(-0.777223093057)*x[1]+(0.297556831029)*x[2]**o+(-0.952676550814)*x[2]
        arg[0,2,2,1]=(0.226250474873)*x[0]**o+(0.279287242249)*x[0]+(0.421824124143)*x[1]**o+(-0.97664927412)*x[1]+(0.703245736838)*x[2]**o+(-0.629660622476)*x[2]
        arg[0,3,0,0]=(-0.138403445102)*x[0]**o+(0.0979152923187)*x[0]+(-0.177116218507)*x[1]**o+(0.0100523961708)*x[1]+(-0.634159492232)*x[2]**o+(-0.847618220907)*x[2]
        arg[0,3,0,1]=(0.612268738957)*x[0]**o+(-0.462682392204)*x[0]+(-0.146623835506)*x[1]**o+(0.699258469487)*x[1]+(0.0139983038691)*x[2]**o+(0.236461011719)*x[2]
        arg[0,3,1,0]=(-0.0518123102809)*x[0]**o+(0.386822059696)*x[0]+(-0.208858744655)*x[1]**o+(0.971393362011)*x[1]+(-0.171649562077)*x[2]**o+(-0.507479926411)*x[2]
        arg[0,3,1,1]=(0.0140396270391)*x[0]**o+(-0.837711757872)*x[0]+(-0.00108924459289)*x[1]**o+(-0.724087080496)*x[1]+(-0.793071741694)*x[2]**o+(0.38654936043)*x[2]
        arg[0,3,2,0]=(-0.897711437057)*x[0]**o+(0.562533782571)*x[0]+(-0.354013039442)*x[1]**o+(-0.473034213198)*x[1]+(0.4598333464)*x[2]**o+(0.467503992754)*x[2]
        arg[0,3,2,1]=(-0.198441232877)*x[0]**o+(-0.411568849672)*x[0]+(0.0850211273328)*x[1]**o+(-0.333702582633)*x[1]+(0.727888461746)*x[2]**o+(-0.174835541819)*x[2]
        arg[0,4,0,0]=(-0.114679585752)*x[0]**o+(0.0446596462171)*x[0]+(-0.950604674991)*x[1]**o+(-0.307622186566)*x[1]+(0.956712754366)*x[2]**o+(0.334630480874)*x[2]
        arg[0,4,0,1]=(0.345475911118)*x[0]**o+(0.834620576123)*x[0]+(0.903176273684)*x[1]**o+(-0.364211856466)*x[1]+(0.910903954488)*x[2]**o+(0.392248545744)*x[2]
        arg[0,4,1,0]=(-0.593360797686)*x[0]**o+(-0.0199892857836)*x[0]+(0.293839091284)*x[1]**o+(0.580208854742)*x[1]+(0.060499817751)*x[2]**o+(0.907498841261)*x[2]
        arg[0,4,1,1]=(-0.446192713548)*x[0]**o+(0.29870251614)*x[0]+(0.754231657585)*x[1]**o+(0.321710920549)*x[1]+(0.287978396665)*x[2]**o+(-0.963473165492)*x[2]
        arg[0,4,2,0]=(0.780174386478)*x[0]**o+(0.603377914483)*x[0]+(-0.841703120145)*x[1]**o+(0.316522795462)*x[1]+(-0.192916583378)*x[2]**o+(-0.734091913995)*x[2]
        arg[0,4,2,1]=(-0.130567656486)*x[0]**o+(-0.342928842892)*x[0]+(-0.0827376128775)*x[1]**o+(-0.771104057041)*x[1]+(0.733571283734)*x[2]**o+(-0.880531838119)*x[2]
        arg[1,0,0,0]=(-0.634428121043)*x[0]**o+(0.294135783753)*x[0]+(-0.566305561339)*x[1]**o+(0.0332700472211)*x[1]+(0.128130119463)*x[2]**o+(0.0468678470468)*x[2]
        arg[1,0,0,1]=(0.82168858107)*x[0]**o+(-0.75544177649)*x[0]+(0.260115562827)*x[1]**o+(-0.149575433725)*x[1]+(-0.780945910032)*x[2]**o+(-0.0539291857071)*x[2]
        arg[1,0,1,0]=(-0.97885291605)*x[0]**o+(0.777955971883)*x[0]+(0.115355866293)*x[1]**o+(-0.456449594589)*x[1]+(-0.832816352387)*x[2]**o+(-0.418070257158)*x[2]
        arg[1,0,1,1]=(-0.908396506245)*x[0]**o+(-0.982147653872)*x[0]+(0.138657052347)*x[1]**o+(0.633837624386)*x[1]+(-0.529748303776)*x[2]**o+(-0.950225664343)*x[2]
        arg[1,0,2,0]=(0.733974435377)*x[0]**o+(-0.758228777635)*x[0]+(0.841368499977)*x[1]**o+(0.871715889444)*x[1]+(0.541055759071)*x[2]**o+(0.669798500731)*x[2]
        arg[1,0,2,1]=(-0.572994548563)*x[0]**o+(0.806241853906)*x[0]+(0.141217952929)*x[1]**o+(0.832291361477)*x[1]+(0.207737014762)*x[2]**o+(-0.279769695917)*x[2]
        arg[1,1,0,0]=(-0.394247339923)*x[0]**o+(0.548338039327)*x[0]+(-0.845822483858)*x[1]**o+(-0.941161341738)*x[1]+(-0.244513027249)*x[2]**o+(-0.524790843131)*x[2]
        arg[1,1,0,1]=(0.509994957918)*x[0]**o+(0.691470630092)*x[0]+(0.249241759795)*x[1]**o+(-0.995293157716)*x[1]+(0.162652088127)*x[2]**o+(0.873578806104)*x[2]
        arg[1,1,1,0]=(-0.22367137056)*x[0]**o+(-0.896306406631)*x[0]+(0.613214186844)*x[1]**o+(-0.460055562163)*x[1]+(-0.704736634406)*x[2]**o+(-0.517741014186)*x[2]
        arg[1,1,1,1]=(-0.426072795689)*x[0]**o+(-0.13452184053)*x[0]+(0.908778671123)*x[1]**o+(0.194472828672)*x[1]+(-0.658909495204)*x[2]**o+(-0.234260758459)*x[2]
        arg[1,1,2,0]=(0.236086562474)*x[0]**o+(0.33312461561)*x[0]+(0.867089643801)*x[1]**o+(-0.720157305369)*x[1]+(0.812845198077)*x[2]**o+(-0.375694713863)*x[2]
        arg[1,1,2,1]=(-0.561145817489)*x[0]**o+(0.746630137163)*x[0]+(-0.809280840008)*x[1]**o+(-0.690912076356)*x[1]+(0.88933641231)*x[2]**o+(0.246647817564)*x[2]
        arg[1,2,0,0]=(0.701319152729)*x[0]**o+(-0.309074024603)*x[0]+(0.567988628717)*x[1]**o+(-0.0596939135728)*x[1]+(0.242135294918)*x[2]**o+(-0.101136005445)*x[2]
        arg[1,2,0,1]=(0.880665674235)*x[0]**o+(-0.833066777852)*x[0]+(-0.0556340786728)*x[1]**o+(0.620095583249)*x[1]+(-0.390556829051)*x[2]**o+(0.793481461864)*x[2]
        arg[1,2,1,0]=(0.832046852431)*x[0]**o+(0.4276990324)*x[0]+(-0.486292331356)*x[1]**o+(0.990296730852)*x[1]+(-0.381732943927)*x[2]**o+(0.225838720655)*x[2]
        arg[1,2,1,1]=(-0.971477034333)*x[0]**o+(0.819409035213)*x[0]+(-0.610341797903)*x[1]**o+(0.842186877805)*x[1]+(0.571187503044)*x[2]**o+(-0.150565006374)*x[2]
        arg[1,2,2,0]=(0.266080624706)*x[0]**o+(0.893377400845)*x[0]+(0.718525426441)*x[1]**o+(-0.847615622286)*x[1]+(0.614914866)*x[2]**o+(0.826503774974)*x[2]
        arg[1,2,2,1]=(0.998060871997)*x[0]**o+(-0.572523112154)*x[0]+(-0.786785614822)*x[1]**o+(0.556952019953)*x[1]+(0.804849394796)*x[2]**o+(-0.201157281118)*x[2]
        arg[1,3,0,0]=(-0.322048105751)*x[0]**o+(0.215441634353)*x[0]+(0.203298400258)*x[1]**o+(0.452071048161)*x[1]+(0.0134315902731)*x[2]**o+(0.240098266028)*x[2]
        arg[1,3,0,1]=(-0.109845136841)*x[0]**o+(0.0479663457074)*x[0]+(0.899394604931)*x[1]**o+(0.853608214151)*x[1]+(-0.27487885579)*x[2]**o+(-0.0236702866567)*x[2]
        arg[1,3,1,0]=(0.948550474293)*x[0]**o+(0.359077475995)*x[0]+(0.655387467125)*x[1]**o+(-0.0501114084308)*x[1]+(-0.244470769181)*x[2]**o+(0.670276065978)*x[2]
        arg[1,3,1,1]=(-0.860769147822)*x[0]**o+(-0.738449680804)*x[0]+(0.840959189643)*x[1]**o+(0.229930283198)*x[1]+(-0.521034843253)*x[2]**o+(-0.50213233896)*x[2]
        arg[1,3,2,0]=(0.29910216992)*x[0]**o+(0.411953562132)*x[0]+(0.777584507308)*x[1]**o+(0.41482951697)*x[1]+(-0.948751119313)*x[2]**o+(-0.526573791564)*x[2]
        arg[1,3,2,1]=(-0.768167653244)*x[0]**o+(-0.318762199437)*x[0]+(0.7566706966)*x[1]**o+(-0.151347269999)*x[1]+(-0.626690637485)*x[2]**o+(-0.731518248609)*x[2]
        arg[1,4,0,0]=(0.0147733854143)*x[0]**o+(-0.812988077571)*x[0]+(0.815385483312)*x[1]**o+(0.285326283687)*x[1]+(0.872502797762)*x[2]**o+(-0.748392032843)*x[2]
        arg[1,4,0,1]=(0.172931930795)*x[0]**o+(0.425088340505)*x[0]+(0.0122969950852)*x[1]**o+(0.556865378995)*x[1]+(0.614686173604)*x[2]**o+(0.966581266651)*x[2]
        arg[1,4,1,0]=(0.725270959871)*x[0]**o+(-0.672683754253)*x[0]+(0.486696038678)*x[1]**o+(-0.53856418144)*x[1]+(-0.295197744155)*x[2]**o+(0.0610876962829)*x[2]
        arg[1,4,1,1]=(0.942300770794)*x[0]**o+(0.876440453364)*x[0]+(0.432652207093)*x[1]**o+(-0.562629961162)*x[1]+(0.497526473424)*x[2]**o+(-0.165824324383)*x[2]
        arg[1,4,2,0]=(-0.668927175078)*x[0]**o+(0.587283304457)*x[0]+(0.588305038304)*x[1]**o+(0.087906366831)*x[1]+(-0.907721198172)*x[2]**o+(-0.506067369924)*x[2]
        arg[1,4,2,1]=(0.899071894553)*x[0]**o+(0.920678485704)*x[0]+(-0.244107214117)*x[1]**o+(0.196327093214)*x[1]+(-0.538075261013)*x[2]**o+(-0.925646646321)*x[2]
        arg[2,0,0,0]=(-0.113452114076)*x[0]**o+(0.541178389983)*x[0]+(0.501083424074)*x[1]**o+(-0.234654083639)*x[1]+(-0.435569149263)*x[2]**o+(0.671517947867)*x[2]
        arg[2,0,0,1]=(-0.00139653535956)*x[0]**o+(-0.93731587459)*x[0]+(-0.387786776826)*x[1]**o+(0.361445822788)*x[1]+(-0.0958490804452)*x[2]**o+(0.22475163005)*x[2]
        arg[2,0,1,0]=(0.100832420019)*x[0]**o+(-0.957577198313)*x[0]+(0.244479124631)*x[1]**o+(-0.589295650233)*x[1]+(-0.11020792463)*x[2]**o+(0.969023424875)*x[2]
        arg[2,0,1,1]=(-0.82757398282)*x[0]**o+(0.665274113482)*x[0]+(0.219624146627)*x[1]**o+(0.604687589384)*x[1]+(0.539774513376)*x[2]**o+(0.676136495778)*x[2]
        arg[2,0,2,0]=(-0.869539421193)*x[0]**o+(0.449645390105)*x[0]+(-0.216356671763)*x[1]**o+(-0.191616800811)*x[1]+(-0.324126591392)*x[2]**o+(0.726527544226)*x[2]
        arg[2,0,2,1]=(-0.522340761977)*x[0]**o+(-0.253136332372)*x[0]+(-0.993693358789)*x[1]**o+(0.689338979208)*x[1]+(-0.73030598892)*x[2]**o+(0.604868622965)*x[2]
        arg[2,1,0,0]=(-0.87285436276)*x[0]**o+(-0.665917385941)*x[0]+(0.105570880359)*x[1]**o+(0.773780432278)*x[1]+(-0.0932677950401)*x[2]**o+(0.55563257373)*x[2]
        arg[2,1,0,1]=(-0.717823895492)*x[0]**o+(-0.843992380543)*x[0]+(0.348730031558)*x[1]**o+(-0.0641727895105)*x[1]+(-0.575332519937)*x[2]**o+(0.566731137573)*x[2]
        arg[2,1,1,0]=(-0.749216309273)*x[0]**o+(0.945232679736)*x[0]+(-0.984694983452)*x[1]**o+(-0.697422850579)*x[1]+(0.573407778839)*x[2]**o+(0.654548962131)*x[2]
        arg[2,1,1,1]=(-0.951637762959)*x[0]**o+(-0.0299724748286)*x[0]+(-0.784527116834)*x[1]**o+(0.490138786409)*x[1]+(0.945743565011)*x[2]**o+(0.158024494836)*x[2]
        arg[2,1,2,0]=(0.80439224967)*x[0]**o+(0.889509225531)*x[0]+(-0.178335319647)*x[1]**o+(-0.365669987413)*x[1]+(-0.170055755127)*x[2]**o+(0.258843711555)*x[2]
        arg[2,1,2,1]=(0.520706901588)*x[0]**o+(0.897178241272)*x[0]+(-0.128495163664)*x[1]**o+(-0.404742110163)*x[1]+(0.140333496135)*x[2]**o+(0.31429703244)*x[2]
        arg[2,2,0,0]=(-0.933201442136)*x[0]**o+(0.0083650170184)*x[0]+(-0.281807642508)*x[1]**o+(-0.09193002721)*x[1]+(-0.422582781227)*x[2]**o+(0.358915935551)*x[2]
        arg[2,2,0,1]=(-0.09712053324)*x[0]**o+(-0.0752448353447)*x[0]+(0.41291697894)*x[1]**o+(0.593823615119)*x[1]+(0.954183095989)*x[2]**o+(0.479576723309)*x[2]
        arg[2,2,1,0]=(-0.48592273499)*x[0]**o+(-0.555807193072)*x[0]+(0.842926147107)*x[1]**o+(-0.0760474608316)*x[1]+(0.864590424311)*x[2]**o+(-0.854721691042)*x[2]
        arg[2,2,1,1]=(0.238261310086)*x[0]**o+(-0.506114025767)*x[0]+(-0.534154656267)*x[1]**o+(-0.282262066272)*x[1]+(-0.241950678835)*x[2]**o+(-0.596913007205)*x[2]
        arg[2,2,2,0]=(0.35002425052)*x[0]**o+(-0.724235915518)*x[0]+(-0.443311896206)*x[1]**o+(0.0183169190559)*x[1]+(-0.180995873436)*x[2]**o+(0.949105760518)*x[2]
        arg[2,2,2,1]=(-0.524596480672)*x[0]**o+(-0.72511087285)*x[0]+(-0.793159588655)*x[1]**o+(0.197579071177)*x[1]+(0.214510478855)*x[2]**o+(0.592011219443)*x[2]
        arg[2,3,0,0]=(-0.777144207151)*x[0]**o+(0.821815068413)*x[0]+(0.828183715461)*x[1]**o+(-0.77742744879)*x[1]+(0.520088528639)*x[2]**o+(0.564796500907)*x[2]
        arg[2,3,0,1]=(-0.444252706506)*x[0]**o+(-0.999658246414)*x[0]+(0.458490846376)*x[1]**o+(0.206371738627)*x[1]+(-0.5943830165)*x[2]**o+(0.125954475804)*x[2]
        arg[2,3,1,0]=(-0.499555379078)*x[0]**o+(0.667767779032)*x[0]+(-0.169053599596)*x[1]**o+(0.587608309525)*x[1]+(-0.99317016838)*x[2]**o+(-0.580232194372)*x[2]
        arg[2,3,1,1]=(0.881736170108)*x[0]**o+(-0.0221128885544)*x[0]+(-0.239202356219)*x[1]**o+(-0.825250840397)*x[1]+(-0.34138233902)*x[2]**o+(-0.900437209752)*x[2]
        arg[2,3,2,0]=(-0.810415345283)*x[0]**o+(0.671639285402)*x[0]+(0.0577696593549)*x[1]**o+(0.12466039802)*x[1]+(-0.164952472389)*x[2]**o+(-0.881356759483)*x[2]
        arg[2,3,2,1]=(0.641411687742)*x[0]**o+(0.939465110042)*x[0]+(0.182541049996)*x[1]**o+(-0.502132165421)*x[1]+(-0.914212670088)*x[2]**o+(0.361356243622)*x[2]
        arg[2,4,0,0]=(0.696033328872)*x[0]**o+(0.854329989895)*x[0]+(0.66092830442)*x[1]**o+(-0.165623528752)*x[1]+(0.587322879532)*x[2]**o+(-0.441145527759)*x[2]
        arg[2,4,0,1]=(0.885589185979)*x[0]**o+(0.170856006811)*x[0]+(-0.500114387063)*x[1]**o+(-0.772574748562)*x[1]+(0.599300325869)*x[2]**o+(0.0950082175327)*x[2]
        arg[2,4,1,0]=(0.877739402691)*x[0]**o+(0.494715865857)*x[0]+(-0.4548916845)*x[1]**o+(-0.108381552398)*x[1]+(-0.620522177203)*x[2]**o+(-0.375546939131)*x[2]
        arg[2,4,1,1]=(-0.108044365878)*x[0]**o+(-0.592058059304)*x[0]+(-0.64481990676)*x[1]**o+(-0.903674017194)*x[1]+(0.0964052945913)*x[2]**o+(0.357819426568)*x[2]
        arg[2,4,2,0]=(-0.686157836559)*x[0]**o+(0.585514351597)*x[0]+(-0.5623443108)*x[1]**o+(-0.636059380122)*x[1]+(0.469863235081)*x[2]**o+(0.374567303622)*x[2]
        arg[2,4,2,1]=(0.214800008136)*x[0]**o+(0.0680448664819)*x[0]+(0.722930302776)*x[1]**o+(0.107573165145)*x[1]+(0.0027804595356)*x[2]**o+(0.725566605544)*x[2]
        arg[3,0,0,0]=(-0.273256791086)*x[0]**o+(0.801129507153)*x[0]+(-0.840227644476)*x[1]**o+(-0.31594258736)*x[1]+(-0.140806177065)*x[2]**o+(-0.0958103115672)*x[2]
        arg[3,0,0,1]=(0.588534583503)*x[0]**o+(0.242235627811)*x[0]+(-0.541539488722)*x[1]**o+(0.624385574868)*x[1]+(-0.0641371754789)*x[2]**o+(-0.635676751281)*x[2]
        arg[3,0,1,0]=(0.81967100218)*x[0]**o+(0.671597165436)*x[0]+(0.160179001916)*x[1]**o+(-0.226714136475)*x[1]+(-0.523709112534)*x[2]**o+(0.932797007296)*x[2]
        arg[3,0,1,1]=(0.707793115256)*x[0]**o+(-0.0995173271712)*x[0]+(-0.594597614398)*x[1]**o+(-0.72096550461)*x[1]+(-0.373631371949)*x[2]**o+(-0.282151921803)*x[2]
        arg[3,0,2,0]=(0.873307297971)*x[0]**o+(0.0950091191819)*x[0]+(-0.534108539131)*x[1]**o+(0.567381193719)*x[1]+(-0.276691140405)*x[2]**o+(0.84882104254)*x[2]
        arg[3,0,2,1]=(-0.895148878411)*x[0]**o+(-0.880973451462)*x[0]+(0.105602819158)*x[1]**o+(-0.836475687487)*x[1]+(-0.18380164164)*x[2]**o+(-0.391643939001)*x[2]
        arg[3,1,0,0]=(-0.144070349397)*x[0]**o+(-0.247578520564)*x[0]+(0.512469248899)*x[1]**o+(0.0369808259097)*x[1]+(0.67517925106)*x[2]**o+(-0.306122403522)*x[2]
        arg[3,1,0,1]=(0.389023591419)*x[0]**o+(0.0335324297663)*x[0]+(0.166380432563)*x[1]**o+(0.415133512101)*x[1]+(-0.956055398028)*x[2]**o+(-0.069254560597)*x[2]
        arg[3,1,1,0]=(-0.253194755795)*x[0]**o+(-0.432240637655)*x[0]+(0.886730830871)*x[1]**o+(0.94685399231)*x[1]+(-0.0193581886643)*x[2]**o+(-0.828124732153)*x[2]
        arg[3,1,1,1]=(0.948739050057)*x[0]**o+(-0.200402018913)*x[0]+(0.062573402994)*x[1]**o+(-0.408727698463)*x[1]+(-0.0334411087309)*x[2]**o+(-0.676147119957)*x[2]
        arg[3,1,2,0]=(0.249237167322)*x[0]**o+(-0.621387879739)*x[0]+(-0.745335388751)*x[1]**o+(-0.183173004074)*x[1]+(0.56632553768)*x[2]**o+(-0.519909176525)*x[2]
        arg[3,1,2,1]=(-0.660822495815)*x[0]**o+(0.0459275869317)*x[0]+(-0.0853936912232)*x[1]**o+(-0.895502076207)*x[1]+(0.150013238845)*x[2]**o+(0.201527905115)*x[2]
        arg[3,2,0,0]=(0.0846880899978)*x[0]**o+(-0.450721252185)*x[0]+(0.0100224498309)*x[1]**o+(-0.729399043167)*x[1]+(0.249073157161)*x[2]**o+(-0.313056982897)*x[2]
        arg[3,2,0,1]=(0.349123663394)*x[0]**o+(-0.0439266646004)*x[0]+(0.201749457716)*x[1]**o+(-0.29961557135)*x[1]+(0.247657591065)*x[2]**o+(0.702448030656)*x[2]
        arg[3,2,1,0]=(-0.67551758529)*x[0]**o+(-0.856271486186)*x[0]+(-0.337698270178)*x[1]**o+(-0.855526501932)*x[1]+(0.204628923051)*x[2]**o+(0.0425229547861)*x[2]
        arg[3,2,1,1]=(0.913895958603)*x[0]**o+(0.33488814727)*x[0]+(0.725768429772)*x[1]**o+(0.31487179035)*x[1]+(-0.926895768372)*x[2]**o+(-0.303623128744)*x[2]
        arg[3,2,2,0]=(0.695497510732)*x[0]**o+(-0.71819800795)*x[0]+(0.0949892877746)*x[1]**o+(-0.371952684199)*x[1]+(-0.712679288901)*x[2]**o+(-0.839032166941)*x[2]
        arg[3,2,2,1]=(0.847915632309)*x[0]**o+(0.489320767225)*x[0]+(0.884694888545)*x[1]**o+(-0.836734066966)*x[1]+(-0.782643394677)*x[2]**o+(0.998012000062)*x[2]
        arg[3,3,0,0]=(0.400343531732)*x[0]**o+(-0.977399566427)*x[0]+(-0.0491815761683)*x[1]**o+(0.820343030844)*x[1]+(0.145276965371)*x[2]**o+(0.106181410329)*x[2]
        arg[3,3,0,1]=(0.39509119571)*x[0]**o+(-0.532152199068)*x[0]+(0.0961163031465)*x[1]**o+(-0.845734786516)*x[1]+(0.0685207262665)*x[2]**o+(0.10725771579)*x[2]
        arg[3,3,1,0]=(-0.326475146957)*x[0]**o+(0.0491193078417)*x[0]+(-0.193405294452)*x[1]**o+(0.283579426128)*x[1]+(0.885442239373)*x[2]**o+(-0.838544460442)*x[2]
        arg[3,3,1,1]=(-0.502234549601)*x[0]**o+(-0.764617462736)*x[0]+(0.52701849705)*x[1]**o+(-0.423245302614)*x[1]+(0.728096320995)*x[2]**o+(0.143356081954)*x[2]
        arg[3,3,2,0]=(0.924562662058)*x[0]**o+(-0.0120799887424)*x[0]+(-0.342886608544)*x[1]**o+(0.609435784717)*x[1]+(0.392296911228)*x[2]**o+(0.572866548262)*x[2]
        arg[3,3,2,1]=(0.661267105524)*x[0]**o+(-0.71847134372)*x[0]+(-0.913866381825)*x[1]**o+(-0.502556458508)*x[1]+(0.201214639904)*x[2]**o+(-0.431939125167)*x[2]
        arg[3,4,0,0]=(0.391785002874)*x[0]**o+(0.424885787931)*x[0]+(-0.377558346944)*x[1]**o+(-0.556924666311)*x[1]+(0.511852194353)*x[2]**o+(0.514261734074)*x[2]
        arg[3,4,0,1]=(-0.269148156354)*x[0]**o+(0.692033978804)*x[0]+(-0.390604590454)*x[1]**o+(-0.551681907742)*x[1]+(0.0613720420346)*x[2]**o+(-0.508841982991)*x[2]
        arg[3,4,1,0]=(0.593170387694)*x[0]**o+(-0.00676980821753)*x[0]+(0.788564472913)*x[1]**o+(0.823094059865)*x[1]+(0.52895485971)*x[2]**o+(0.88849618397)*x[2]
        arg[3,4,1,1]=(0.772185283881)*x[0]**o+(0.759014894154)*x[0]+(-0.142739578724)*x[1]**o+(-0.141899255898)*x[1]+(0.479086662114)*x[2]**o+(0.68993908918)*x[2]
        arg[3,4,2,0]=(0.998097253401)*x[0]**o+(0.621254339576)*x[0]+(-0.692540865624)*x[1]**o+(0.716545091382)*x[1]+(0.478209703594)*x[2]**o+(0.259063088226)*x[2]
        arg[3,4,2,1]=(-0.321962318028)*x[0]**o+(0.0377531939042)*x[0]+(-0.362041758315)*x[1]**o+(0.632474956936)*x[1]+(0.133972776903)*x[2]**o+(-0.341803258375)*x[2]
        ref[0,0,0,0]=(-0.484602377547)*(1+2.*(dim-1.)/(o+1.))+(-1.0194202985)*dim
        ref[0,0,0,1]=(-0.835132633821)*(1+2.*(dim-1.)/(o+1.))+(0.616687157924)*dim
        ref[0,0,1,0]=(1.49678131808)*(1+2.*(dim-1.)/(o+1.))+(0.880777811066)*dim
        ref[0,0,1,1]=(0.910664643761)*(1+2.*(dim-1.)/(o+1.))+(-0.961299455217)*dim
        ref[0,0,2,0]=(0.946573940383)*(1+2.*(dim-1.)/(o+1.))+(0.598873680457)*dim
        ref[0,0,2,1]=(0.872725648736)*(1+2.*(dim-1.)/(o+1.))+(-1.54360766814)*dim
        ref[0,1,0,0]=(1.67226553782)*(1+2.*(dim-1.)/(o+1.))+(0.0561023038674)*dim
        ref[0,1,0,1]=(1.15582406251)*(1+2.*(dim-1.)/(o+1.))+(-0.620028515442)*dim
        ref[0,1,1,0]=(-1.79024527224)*(1+2.*(dim-1.)/(o+1.))+(0.726554075535)*dim
        ref[0,1,1,1]=(0.8477406514)*(1+2.*(dim-1.)/(o+1.))+(-1.15509819622)*dim
        ref[0,1,2,0]=(0.0209750791326)*(1+2.*(dim-1.)/(o+1.))+(-1.00141913872)*dim
        ref[0,1,2,1]=(0.569876032064)*(1+2.*(dim-1.)/(o+1.))+(0.271451611059)*dim
        ref[0,2,0,0]=(0.19197323238)*(1+2.*(dim-1.)/(o+1.))+(0.320608976525)*dim
        ref[0,2,0,1]=(-0.864025126798)*(1+2.*(dim-1.)/(o+1.))+(0.792122272387)*dim
        ref[0,2,1,0]=(-1.9943606725)*(1+2.*(dim-1.)/(o+1.))+(1.25566864927)*dim
        ref[0,2,1,1]=(-1.0628223267)*(1+2.*(dim-1.)/(o+1.))+(-1.32772813107)*dim
        ref[0,2,2,0]=(1.40455685015)*(1+2.*(dim-1.)/(o+1.))+(-1.31673180399)*dim
        ref[0,2,2,1]=(1.35132033585)*(1+2.*(dim-1.)/(o+1.))+(-1.32702265435)*dim
        ref[0,3,0,0]=(-0.949679155841)*(1+2.*(dim-1.)/(o+1.))+(-0.739650532418)*dim
        ref[0,3,0,1]=(0.47964320732)*(1+2.*(dim-1.)/(o+1.))+(0.473037089003)*dim
        ref[0,3,1,0]=(-0.432320617013)*(1+2.*(dim-1.)/(o+1.))+(0.850735495297)*dim
        ref[0,3,1,1]=(-0.780121359247)*(1+2.*(dim-1.)/(o+1.))+(-1.17524947794)*dim
        ref[0,3,2,0]=(-0.791891130098)*(1+2.*(dim-1.)/(o+1.))+(0.557003562128)*dim
        ref[0,3,2,1]=(0.614468356202)*(1+2.*(dim-1.)/(o+1.))+(-0.920106974125)*dim
        ref[0,4,0,0]=(-0.108571506377)*(1+2.*(dim-1.)/(o+1.))+(0.0716679405256)*dim
        ref[0,4,0,1]=(2.15955613929)*(1+2.*(dim-1.)/(o+1.))+(0.862657265402)*dim
        ref[0,4,1,0]=(-0.239021888652)*(1+2.*(dim-1.)/(o+1.))+(1.46771841022)*dim
        ref[0,4,1,1]=(0.596017340702)*(1+2.*(dim-1.)/(o+1.))+(-0.343059728803)*dim
        ref[0,4,2,0]=(-0.254445317045)*(1+2.*(dim-1.)/(o+1.))+(0.185808795951)*dim
        ref[0,4,2,1]=(0.520266014371)*(1+2.*(dim-1.)/(o+1.))+(-1.99456473805)*dim
        ref[1,0,0,0]=(-1.07260356292)*(1+2.*(dim-1.)/(o+1.))+(0.374273678021)*dim
        ref[1,0,0,1]=(0.300858233864)*(1+2.*(dim-1.)/(o+1.))+(-0.958946395922)*dim
        ref[1,0,1,0]=(-1.69631340214)*(1+2.*(dim-1.)/(o+1.))+(-0.096563879864)*dim
        ref[1,0,1,1]=(-1.29948775767)*(1+2.*(dim-1.)/(o+1.))+(-1.29853569383)*dim
        ref[1,0,2,0]=(2.11639869443)*(1+2.*(dim-1.)/(o+1.))+(0.783285612541)*dim
        ref[1,0,2,1]=(-0.224039580872)*(1+2.*(dim-1.)/(o+1.))+(1.35876351947)*dim
        ref[1,1,0,0]=(-1.48458285103)*(1+2.*(dim-1.)/(o+1.))+(-0.917614145542)*dim
        ref[1,1,0,1]=(0.92188880584)*(1+2.*(dim-1.)/(o+1.))+(0.569756278479)*dim
        ref[1,1,1,0]=(-0.315193818122)*(1+2.*(dim-1.)/(o+1.))+(-1.87410298298)*dim
        ref[1,1,1,1]=(-0.17620361977)*(1+2.*(dim-1.)/(o+1.))+(-0.174309770317)*dim
        ref[1,1,2,0]=(1.91602140435)*(1+2.*(dim-1.)/(o+1.))+(-0.762727403622)*dim
        ref[1,1,2,1]=(-0.481090245186)*(1+2.*(dim-1.)/(o+1.))+(0.302365878371)*dim
        ref[1,2,0,0]=(1.51144307636)*(1+2.*(dim-1.)/(o+1.))+(-0.469903943621)*dim
        ref[1,2,0,1]=(0.434474766511)*(1+2.*(dim-1.)/(o+1.))+(0.580510267262)*dim
        ref[1,2,1,0]=(-0.0359784228521)*(1+2.*(dim-1.)/(o+1.))+(1.64383448391)*dim
        ref[1,2,1,1]=(-1.01063132919)*(1+2.*(dim-1.)/(o+1.))+(1.51103090664)*dim
        ref[1,2,2,0]=(1.59952091715)*(1+2.*(dim-1.)/(o+1.))+(0.872265553532)*dim
        ref[1,2,2,1]=(1.01612465197)*(1+2.*(dim-1.)/(o+1.))+(-0.216728373319)*dim
        ref[1,3,0,0]=(-0.10531811522)*(1+2.*(dim-1.)/(o+1.))+(0.907610948542)*dim
        ref[1,3,0,1]=(0.5146706123)*(1+2.*(dim-1.)/(o+1.))+(0.877904273202)*dim
        ref[1,3,1,0]=(1.35946717224)*(1+2.*(dim-1.)/(o+1.))+(0.979242133542)*dim
        ref[1,3,1,1]=(-0.540844801432)*(1+2.*(dim-1.)/(o+1.))+(-1.01065173657)*dim
        ref[1,3,2,0]=(0.127935557915)*(1+2.*(dim-1.)/(o+1.))+(0.300209287538)*dim
        ref[1,3,2,1]=(-0.63818759413)*(1+2.*(dim-1.)/(o+1.))+(-1.20162771805)*dim
        ref[1,4,0,0]=(1.70266166649)*(1+2.*(dim-1.)/(o+1.))+(-1.27605382673)*dim
        ref[1,4,0,1]=(0.799915099484)*(1+2.*(dim-1.)/(o+1.))+(1.94853498615)*dim
        ref[1,4,1,0]=(0.916769254394)*(1+2.*(dim-1.)/(o+1.))+(-1.15016023941)*dim
        ref[1,4,1,1]=(1.87247945131)*(1+2.*(dim-1.)/(o+1.))+(0.147986167818)*dim
        ref[1,4,2,0]=(-0.988343334946)*(1+2.*(dim-1.)/(o+1.))+(0.169122301365)*dim
        ref[1,4,2,1]=(0.116889419423)*(1+2.*(dim-1.)/(o+1.))+(0.191358932597)*dim
        ref[2,0,0,0]=(-0.0479378392653)*(1+2.*(dim-1.)/(o+1.))+(0.978042254211)*dim
        ref[2,0,0,1]=(-0.485032392631)*(1+2.*(dim-1.)/(o+1.))+(-0.351118421752)*dim
        ref[2,0,1,0]=(0.23510362002)*(1+2.*(dim-1.)/(o+1.))+(-0.577849423671)*dim
        ref[2,0,1,1]=(-0.068175322817)*(1+2.*(dim-1.)/(o+1.))+(1.94609819864)*dim
        ref[2,0,2,0]=(-1.41002268435)*(1+2.*(dim-1.)/(o+1.))+(0.98455613352)*dim
        ref[2,0,2,1]=(-2.24634010969)*(1+2.*(dim-1.)/(o+1.))+(1.0410712698)*dim
        ref[2,1,0,0]=(-0.860551277441)*(1+2.*(dim-1.)/(o+1.))+(0.663495620068)*dim
        ref[2,1,0,1]=(-0.944426383871)*(1+2.*(dim-1.)/(o+1.))+(-0.341434032481)*dim
        ref[2,1,1,0]=(-1.16050351389)*(1+2.*(dim-1.)/(o+1.))+(0.902358791288)*dim
        ref[2,1,1,1]=(-0.790421314781)*(1+2.*(dim-1.)/(o+1.))+(0.618190806416)*dim
        ref[2,1,2,0]=(0.456001174897)*(1+2.*(dim-1.)/(o+1.))+(0.782682949672)*dim
        ref[2,1,2,1]=(0.532545234059)*(1+2.*(dim-1.)/(o+1.))+(0.806733163549)*dim
        ref[2,2,0,0]=(-1.63759186587)*(1+2.*(dim-1.)/(o+1.))+(0.275350925359)*dim
        ref[2,2,0,1]=(1.26997954169)*(1+2.*(dim-1.)/(o+1.))+(0.998155503083)*dim
        ref[2,2,1,0]=(1.22159383643)*(1+2.*(dim-1.)/(o+1.))+(-1.48657634495)*dim
        ref[2,2,1,1]=(-0.537844025016)*(1+2.*(dim-1.)/(o+1.))+(-1.38528909924)*dim
        ref[2,2,2,0]=(-0.274283519122)*(1+2.*(dim-1.)/(o+1.))+(0.243186764056)*dim
        ref[2,2,2,1]=(-1.10324559047)*(1+2.*(dim-1.)/(o+1.))+(0.0644794177698)*dim
        ref[2,3,0,0]=(0.571128036949)*(1+2.*(dim-1.)/(o+1.))+(0.60918412053)*dim
        ref[2,3,0,1]=(-0.580144876629)*(1+2.*(dim-1.)/(o+1.))+(-0.667332031982)*dim
        ref[2,3,1,0]=(-1.66177914705)*(1+2.*(dim-1.)/(o+1.))+(0.675143894185)*dim
        ref[2,3,1,1]=(0.301151474869)*(1+2.*(dim-1.)/(o+1.))+(-1.7478009387)*dim
        ref[2,3,2,0]=(-0.917598158317)*(1+2.*(dim-1.)/(o+1.))+(-0.0850570760611)*dim
        ref[2,3,2,1]=(-0.09025993235)*(1+2.*(dim-1.)/(o+1.))+(0.798689188243)*dim
        ref[2,4,0,0]=(1.94428451282)*(1+2.*(dim-1.)/(o+1.))+(0.247560933384)*dim
        ref[2,4,0,1]=(0.984775124785)*(1+2.*(dim-1.)/(o+1.))+(-0.506710524219)*dim
        ref[2,4,1,0]=(-0.197674459012)*(1+2.*(dim-1.)/(o+1.))+(0.0107873743289)*dim
        ref[2,4,1,1]=(-0.656458978046)*(1+2.*(dim-1.)/(o+1.))+(-1.13791264993)*dim
        ref[2,4,2,0]=(-0.778638912279)*(1+2.*(dim-1.)/(o+1.))+(0.324022275097)*dim
        ref[2,4,2,1]=(0.940510770447)*(1+2.*(dim-1.)/(o+1.))+(0.901184637171)*dim
        ref[3,0,0,0]=(-1.25429061263)*(1+2.*(dim-1.)/(o+1.))+(0.389376608226)*dim
        ref[3,0,0,1]=(-0.0171420806973)*(1+2.*(dim-1.)/(o+1.))+(0.230944451397)*dim
        ref[3,0,1,0]=(0.456140891562)*(1+2.*(dim-1.)/(o+1.))+(1.37768003626)*dim
        ref[3,0,1,1]=(-0.260435871091)*(1+2.*(dim-1.)/(o+1.))+(-1.10263475358)*dim
        ref[3,0,2,0]=(0.0625076184349)*(1+2.*(dim-1.)/(o+1.))+(1.51121135544)*dim
        ref[3,0,2,1]=(-0.973347700893)*(1+2.*(dim-1.)/(o+1.))+(-2.10909307795)*dim
        ref[3,1,0,0]=(1.04357815056)*(1+2.*(dim-1.)/(o+1.))+(-0.516720098176)*dim
        ref[3,1,0,1]=(-0.400651374046)*(1+2.*(dim-1.)/(o+1.))+(0.37941138127)*dim
        ref[3,1,1,0]=(0.614177886412)*(1+2.*(dim-1.)/(o+1.))+(-0.313511377498)*dim
        ref[3,1,1,1]=(0.97787134432)*(1+2.*(dim-1.)/(o+1.))+(-1.28527683733)*dim
        ref[3,1,2,0]=(0.0702273162506)*(1+2.*(dim-1.)/(o+1.))+(-1.32447006034)*dim
        ref[3,1,2,1]=(-0.596202948193)*(1+2.*(dim-1.)/(o+1.))+(-0.648046584161)*dim
        ref[3,2,0,0]=(0.34378369699)*(1+2.*(dim-1.)/(o+1.))+(-1.49317727825)*dim
        ref[3,2,0,1]=(0.798530712175)*(1+2.*(dim-1.)/(o+1.))+(0.358905794706)*dim
        ref[3,2,1,0]=(-0.808586932417)*(1+2.*(dim-1.)/(o+1.))+(-1.66927503333)*dim
        ref[3,2,1,1]=(0.712768620002)*(1+2.*(dim-1.)/(o+1.))+(0.346136808876)*dim
        ref[3,2,2,0]=(0.0778075096053)*(1+2.*(dim-1.)/(o+1.))+(-1.92918285909)*dim
        ref[3,2,2,1]=(0.949967126178)*(1+2.*(dim-1.)/(o+1.))+(0.650598700321)*dim
        ref[3,3,0,0]=(0.496438920934)*(1+2.*(dim-1.)/(o+1.))+(-0.0508751252546)*dim
        ref[3,3,0,1]=(0.559728225123)*(1+2.*(dim-1.)/(o+1.))+(-1.27062926979)*dim
        ref[3,3,1,0]=(0.365561797964)*(1+2.*(dim-1.)/(o+1.))+(-0.505845726473)*dim
        ref[3,3,1,1]=(0.752880268444)*(1+2.*(dim-1.)/(o+1.))+(-1.0445066834)*dim
        ref[3,3,2,0]=(0.973972964742)*(1+2.*(dim-1.)/(o+1.))+(1.17022234424)*dim
        ref[3,3,2,1]=(-0.0513846363967)*(1+2.*(dim-1.)/(o+1.))+(-1.65296692739)*dim
        ref[3,4,0,0]=(0.526078850284)*(1+2.*(dim-1.)/(o+1.))+(0.382222855695)*dim
        ref[3,4,0,1]=(-0.598380704773)*(1+2.*(dim-1.)/(o+1.))+(-0.368489911929)*dim
        ref[3,4,1,0]=(1.91068972032)*(1+2.*(dim-1.)/(o+1.))+(1.70482043562)*dim
        ref[3,4,1,1]=(1.10853236727)*(1+2.*(dim-1.)/(o+1.))+(1.30705472744)*dim
        ref[3,4,2,0]=(0.78376609137)*(1+2.*(dim-1.)/(o+1.))+(1.59686251918)*dim
        ref[3,4,2,1]=(-0.550031299439)*(1+2.*(dim-1.)/(o+1.))+(0.328424892465)*dim
      res=integrate(arg)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.276000764164)*x[0]**o+(-0.468805844816)*x[0]+(0.216181374155)*x[1]**o+(0.670874609882)*x[1]
        ref=(0.492182138319)*(1+2.*(dim-1.)/(o+1.))+(0.202068765065)*dim
      else:
        s=(0.959211859452)*x[0]**o+(0.856777496752)*x[0]+(-0.689682159726)*x[1]**o+(-0.0511538529405)*x[1]+(0.909142526772)*x[2]**o+(-0.851436719686)*x[2]
        ref=(1.1786722265)*(1+2.*(dim-1.)/(o+1.))+(-0.0458130758744)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(0.959628680059)*x[0]**o+(0.316358798368)*x[0]+(-0.11695939796)*x[1]**o+(0.467189695616)*x[1]
        s[1]=(-0.0914212792301)*x[0]**o+(-0.876964392495)*x[0]+(-0.241429076296)*x[1]**o+(-0.136283508929)*x[1]
        ref[0]=(0.842669282099)*(1+2.*(dim-1.)/(o+1.))+(0.783548493984)*dim
        ref[1]=(-0.332850355526)*(1+2.*(dim-1.)/(o+1.))+(-1.01324790142)*dim
      else:
        s[0]=(0.438221600861)*x[0]**o+(-0.675281168341)*x[0]+(-0.977876997798)*x[1]**o+(0.263292240228)*x[1]+(0.298700889488)*x[2]**o+(0.218464753318)*x[2]
        s[1]=(0.129309327387)*x[0]**o+(0.718818780853)*x[0]+(-0.829433536453)*x[1]**o+(0.189387074977)*x[1]+(-0.370123421171)*x[2]**o+(-0.944375947073)*x[2]
        ref[0]=(-0.240954507449)*(1+2.*(dim-1.)/(o+1.))+(-0.193524174795)*dim
        ref[1]=(-1.07024763024)*(1+2.*(dim-1.)/(o+1.))+(-0.0361700912426)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.179044774852)*x[0]**o+(-0.771154031526)*x[0]+(0.283792506671)*x[1]**o+(-0.402392048)*x[1]
        s[0,1]=(-0.232541435218)*x[0]**o+(-0.0245965827061)*x[0]+(0.645399819776)*x[1]**o+(-0.561735270705)*x[1]
        s[0,2]=(0.145947743942)*x[0]**o+(-0.575168618436)*x[0]+(-0.323186337184)*x[1]**o+(0.286768456651)*x[1]
        s[0,3]=(-0.850294957994)*x[0]**o+(0.338532201346)*x[0]+(-0.0793135342296)*x[1]**o+(-0.863030213073)*x[1]
        s[0,4]=(-0.250033615938)*x[0]**o+(0.0646802670443)*x[0]+(0.757218484623)*x[1]**o+(-0.136938974237)*x[1]
        s[1,0]=(-0.388186073304)*x[0]**o+(0.816982222815)*x[0]+(0.966648677362)*x[1]**o+(0.433441368441)*x[1]
        s[1,1]=(-0.26320466546)*x[0]**o+(0.814680348027)*x[0]+(0.121083870568)*x[1]**o+(0.546344037777)*x[1]
        s[1,2]=(0.999583479755)*x[0]**o+(-0.613738351596)*x[0]+(-0.15105692369)*x[1]**o+(-0.826110754043)*x[1]
        s[1,3]=(0.823160030826)*x[0]**o+(-0.986858175002)*x[0]+(0.979035971406)*x[1]**o+(0.495551722082)*x[1]
        s[1,4]=(0.322276907312)*x[0]**o+(-0.768815547878)*x[0]+(0.944581045957)*x[1]**o+(0.638853629633)*x[1]
        s[2,0]=(-0.715790872031)*x[0]**o+(-0.983342426781)*x[0]+(-0.584719832165)*x[1]**o+(0.372036497473)*x[1]
        s[2,1]=(0.00365756286977)*x[0]**o+(-0.715355493926)*x[0]+(0.812526897221)*x[1]**o+(-0.676997560296)*x[1]
        s[2,2]=(0.853542765125)*x[0]**o+(-0.468938820981)*x[0]+(0.958637393127)*x[1]**o+(-0.934834551087)*x[1]
        s[2,3]=(0.0411833783296)*x[0]**o+(0.668096714355)*x[0]+(-0.68394500964)*x[1]**o+(-0.915192447045)*x[1]
        s[2,4]=(0.477937998635)*x[0]**o+(-0.655023923727)*x[0]+(0.31178650811)*x[1]**o+(0.573276677618)*x[1]
        s[3,0]=(0.0827434421975)*x[0]**o+(0.116951368308)*x[0]+(-0.53041218008)*x[1]**o+(0.923217710706)*x[1]
        s[3,1]=(0.0352752409951)*x[0]**o+(0.186756758251)*x[0]+(-0.650196963391)*x[1]**o+(0.758872769044)*x[1]
        s[3,2]=(0.627293036559)*x[0]**o+(-0.101693009282)*x[0]+(-0.651126401234)*x[1]**o+(0.0946956663513)*x[1]
        s[3,3]=(-0.0645383041667)*x[0]**o+(0.411155076439)*x[0]+(0.936010412212)*x[1]**o+(-0.662112489076)*x[1]
        s[3,4]=(-0.749505083645)*x[0]**o+(0.767738728633)*x[0]+(-0.0521697451237)*x[1]**o+(0.743320460752)*x[1]
        ref[0,0]=(0.462837281523)*(1+2.*(dim-1.)/(o+1.))+(-1.17354607953)*dim
        ref[0,1]=(0.412858384558)*(1+2.*(dim-1.)/(o+1.))+(-0.586331853411)*dim
        ref[0,2]=(-0.177238593242)*(1+2.*(dim-1.)/(o+1.))+(-0.288400161786)*dim
        ref[0,3]=(-0.929608492223)*(1+2.*(dim-1.)/(o+1.))+(-0.524498011727)*dim
        ref[0,4]=(0.507184868685)*(1+2.*(dim-1.)/(o+1.))+(-0.0722587071931)*dim
        ref[1,0]=(0.578462604058)*(1+2.*(dim-1.)/(o+1.))+(1.25042359126)*dim
        ref[1,1]=(-0.142120794892)*(1+2.*(dim-1.)/(o+1.))+(1.3610243858)*dim
        ref[1,2]=(0.848526556066)*(1+2.*(dim-1.)/(o+1.))+(-1.43984910564)*dim
        ref[1,3]=(1.80219600223)*(1+2.*(dim-1.)/(o+1.))+(-0.49130645292)*dim
        ref[1,4]=(1.26685795327)*(1+2.*(dim-1.)/(o+1.))+(-0.129961918245)*dim
        ref[2,0]=(-1.3005107042)*(1+2.*(dim-1.)/(o+1.))+(-0.611305929308)*dim
        ref[2,1]=(0.816184460091)*(1+2.*(dim-1.)/(o+1.))+(-1.39235305422)*dim
        ref[2,2]=(1.81218015825)*(1+2.*(dim-1.)/(o+1.))+(-1.40377337207)*dim
        ref[2,3]=(-0.64276163131)*(1+2.*(dim-1.)/(o+1.))+(-0.247095732689)*dim
        ref[2,4]=(0.789724506745)*(1+2.*(dim-1.)/(o+1.))+(-0.0817472461083)*dim
        ref[3,0]=(-0.447668737882)*(1+2.*(dim-1.)/(o+1.))+(1.04016907901)*dim
        ref[3,1]=(-0.614921722396)*(1+2.*(dim-1.)/(o+1.))+(0.945629527295)*dim
        ref[3,2]=(-0.0238333646741)*(1+2.*(dim-1.)/(o+1.))+(-0.00699734293025)*dim
        ref[3,3]=(0.871472108046)*(1+2.*(dim-1.)/(o+1.))+(-0.250957412637)*dim
        ref[3,4]=(-0.801674828768)*(1+2.*(dim-1.)/(o+1.))+(1.51105918939)*dim
      else:
        s[0,0]=(-0.125646626393)*x[0]**o+(-0.248909495972)*x[0]+(0.694342489041)*x[1]**o+(-0.944409107438)*x[1]+(0.319473721262)*x[2]**o+(0.601806182335)*x[2]
        s[0,1]=(-0.505114968317)*x[0]**o+(-0.328058153553)*x[0]+(0.231927714111)*x[1]**o+(-0.20605100827)*x[1]+(0.441122232963)*x[2]**o+(-0.800716234123)*x[2]
        s[0,2]=(0.995136206899)*x[0]**o+(-0.853681974117)*x[0]+(-0.0627243072319)*x[1]**o+(0.974181482578)*x[1]+(0.719967223861)*x[2]**o+(0.714716693035)*x[2]
        s[0,3]=(-0.441934133426)*x[0]**o+(0.853576641618)*x[0]+(-0.981667008539)*x[1]**o+(0.628894290368)*x[1]+(-0.860902752279)*x[2]**o+(-0.265897164492)*x[2]
        s[0,4]=(0.793316938319)*x[0]**o+(-0.210698953461)*x[0]+(-0.642203222971)*x[1]**o+(-0.426236159989)*x[1]+(-0.425899483366)*x[2]**o+(0.160038345546)*x[2]
        s[1,0]=(0.708795168562)*x[0]**o+(-0.196948028271)*x[0]+(-0.935375576786)*x[1]**o+(0.0861198689411)*x[1]+(-0.921983001302)*x[2]**o+(0.460387290735)*x[2]
        s[1,1]=(0.929230715741)*x[0]**o+(-0.469953215705)*x[0]+(-0.327574657931)*x[1]**o+(0.508027765515)*x[1]+(-0.599239898886)*x[2]**o+(0.63360774997)*x[2]
        s[1,2]=(0.521095649571)*x[0]**o+(-0.306540618845)*x[0]+(0.989778780969)*x[1]**o+(0.942448043532)*x[1]+(0.696619526146)*x[2]**o+(0.316003759)*x[2]
        s[1,3]=(0.862728599037)*x[0]**o+(0.710693247125)*x[0]+(-0.319771560638)*x[1]**o+(0.00195165068796)*x[1]+(0.196231338915)*x[2]**o+(-0.540258673852)*x[2]
        s[1,4]=(0.608993531839)*x[0]**o+(-0.538278966186)*x[0]+(0.805409107065)*x[1]**o+(-0.306251894849)*x[1]+(0.137370330819)*x[2]**o+(0.723689377138)*x[2]
        s[2,0]=(0.725845088985)*x[0]**o+(-0.146963254784)*x[0]+(-0.33277537985)*x[1]**o+(0.905316206858)*x[1]+(0.833942228903)*x[2]**o+(-0.828915480014)*x[2]
        s[2,1]=(0.405007463559)*x[0]**o+(-0.58741177463)*x[0]+(-0.732389736864)*x[1]**o+(0.810006779688)*x[1]+(0.0934482848718)*x[2]**o+(-0.780478896444)*x[2]
        s[2,2]=(0.330277184347)*x[0]**o+(0.542618623736)*x[0]+(0.0984021435812)*x[1]**o+(0.528357675511)*x[1]+(-0.822522098782)*x[2]**o+(0.327999189729)*x[2]
        s[2,3]=(-0.65382619524)*x[0]**o+(0.683040803962)*x[0]+(-0.951007838791)*x[1]**o+(-0.907245567669)*x[1]+(-0.645828264634)*x[2]**o+(0.471775924817)*x[2]
        s[2,4]=(-0.881174731542)*x[0]**o+(0.902183751937)*x[0]+(0.0550862007787)*x[1]**o+(-0.251644685884)*x[1]+(0.220290011733)*x[2]**o+(-0.433798608361)*x[2]
        s[3,0]=(-0.822853797457)*x[0]**o+(0.92597725252)*x[0]+(0.151102447707)*x[1]**o+(-0.643773009686)*x[1]+(-0.82012950381)*x[2]**o+(-0.735602052424)*x[2]
        s[3,1]=(-0.0841053011733)*x[0]**o+(0.0930350084683)*x[0]+(0.276873742716)*x[1]**o+(-0.624783698693)*x[1]+(0.861829520969)*x[2]**o+(0.523311788689)*x[2]
        s[3,2]=(0.570824564972)*x[0]**o+(-0.258483951766)*x[0]+(-0.550248707418)*x[1]**o+(-0.762093783646)*x[1]+(-0.580531558601)*x[2]**o+(0.529122851261)*x[2]
        s[3,3]=(-0.945884266871)*x[0]**o+(0.947833330446)*x[0]+(0.94128603099)*x[1]**o+(-0.353478856114)*x[1]+(0.125480323773)*x[2]**o+(0.119572435249)*x[2]
        s[3,4]=(-0.0190004624928)*x[0]**o+(0.935780695738)*x[0]+(0.525428336584)*x[1]**o+(-0.94896065125)*x[1]+(-0.277950627306)*x[2]**o+(0.0152064531175)*x[2]
        ref[0,0]=(0.88816958391)*(1+2.*(dim-1.)/(o+1.))+(-0.591512421075)*dim
        ref[0,1]=(0.167934978757)*(1+2.*(dim-1.)/(o+1.))+(-1.33482539595)*dim
        ref[0,2]=(1.65237912353)*(1+2.*(dim-1.)/(o+1.))+(0.835216201496)*dim
        ref[0,3]=(-2.28450389424)*(1+2.*(dim-1.)/(o+1.))+(1.21657376749)*dim
        ref[0,4]=(-0.274785768018)*(1+2.*(dim-1.)/(o+1.))+(-0.476896767904)*dim
        ref[1,0]=(-1.14856340953)*(1+2.*(dim-1.)/(o+1.))+(0.349559131405)*dim
        ref[1,1]=(0.0024161589232)*(1+2.*(dim-1.)/(o+1.))+(0.67168229978)*dim
        ref[1,2]=(2.20749395669)*(1+2.*(dim-1.)/(o+1.))+(0.951911183687)*dim
        ref[1,3]=(0.739188377314)*(1+2.*(dim-1.)/(o+1.))+(0.172386223961)*dim
        ref[1,4]=(1.55177296972)*(1+2.*(dim-1.)/(o+1.))+(-0.120841483898)*dim
        ref[2,0]=(1.22701193804)*(1+2.*(dim-1.)/(o+1.))+(-0.0705625279405)*dim
        ref[2,1]=(-0.233933988434)*(1+2.*(dim-1.)/(o+1.))+(-0.557883891386)*dim
        ref[2,2]=(-0.393842770855)*(1+2.*(dim-1.)/(o+1.))+(1.39897548898)*dim
        ref[2,3]=(-2.25066229866)*(1+2.*(dim-1.)/(o+1.))+(0.247571161111)*dim
        ref[2,4]=(-0.60579851903)*(1+2.*(dim-1.)/(o+1.))+(0.216740457692)*dim
        ref[3,0]=(-1.49188085356)*(1+2.*(dim-1.)/(o+1.))+(-0.453397809589)*dim
        ref[3,1]=(1.05459796251)*(1+2.*(dim-1.)/(o+1.))+(-0.00843690153556)*dim
        ref[3,2]=(-0.559955701048)*(1+2.*(dim-1.)/(o+1.))+(-0.491454884151)*dim
        ref[3,3]=(0.120882087893)*(1+2.*(dim-1.)/(o+1.))+(0.713926909581)*dim
        ref[3,4]=(0.228477246785)*(1+2.*(dim-1.)/(o+1.))+(0.00202649760578)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(-0.926772326685)*x[0]**o+(0.387486521987)*x[0]+(0.434356695921)*x[1]**o+(0.612313168985)*x[1]
        s[0,0,1]=(0.0416123602896)*x[0]**o+(0.688827142008)*x[0]+(-0.983728357354)*x[1]**o+(0.360807139866)*x[1]
        s[0,1,0]=(0.695266078006)*x[0]**o+(0.415695465239)*x[0]+(-0.679624573491)*x[1]**o+(0.428762900621)*x[1]
        s[0,1,1]=(0.313944654416)*x[0]**o+(0.0809925793325)*x[0]+(0.621639152049)*x[1]**o+(0.611203880481)*x[1]
        s[1,0,0]=(0.779313803934)*x[0]**o+(0.417564759976)*x[0]+(0.309499891258)*x[1]**o+(0.309580948656)*x[1]
        s[1,0,1]=(-0.165921233657)*x[0]**o+(0.754873050152)*x[0]+(-0.086251970981)*x[1]**o+(0.326647616211)*x[1]
        s[1,1,0]=(-0.803682583733)*x[0]**o+(-0.532584958606)*x[0]+(0.161316696631)*x[1]**o+(-0.956754947772)*x[1]
        s[1,1,1]=(0.229536665796)*x[0]**o+(0.436690509304)*x[0]+(0.111716074991)*x[1]**o+(0.0152188972478)*x[1]
        s[2,0,0]=(0.431277528457)*x[0]**o+(-0.549627071548)*x[0]+(0.896434242812)*x[1]**o+(0.310788829239)*x[1]
        s[2,0,1]=(0.337481979979)*x[0]**o+(-0.115120163813)*x[0]+(0.261492211787)*x[1]**o+(-0.78544941385)*x[1]
        s[2,1,0]=(-0.930198652575)*x[0]**o+(0.378441956425)*x[0]+(0.10805648813)*x[1]**o+(-0.436097313908)*x[1]
        s[2,1,1]=(0.249959291812)*x[0]**o+(0.836638711074)*x[0]+(0.0728794590137)*x[1]**o+(0.66862732862)*x[1]
        s[3,0,0]=(-0.738201854061)*x[0]**o+(0.772813906884)*x[0]+(-0.181342905715)*x[1]**o+(0.804464512448)*x[1]
        s[3,0,1]=(-0.196081244053)*x[0]**o+(0.994865977727)*x[0]+(0.842797040161)*x[1]**o+(0.821509016638)*x[1]
        s[3,1,0]=(0.833815509702)*x[0]**o+(0.577383594944)*x[0]+(-0.570130606584)*x[1]**o+(-0.0805462111487)*x[1]
        s[3,1,1]=(-0.8554995951)*x[0]**o+(0.720808128606)*x[0]+(0.457881428951)*x[1]**o+(-0.0471201858303)*x[1]
        s[4,0,0]=(0.48208350846)*x[0]**o+(0.47231745613)*x[0]+(-0.889701258367)*x[1]**o+(-0.630128265742)*x[1]
        s[4,0,1]=(-0.252911141671)*x[0]**o+(0.49998513292)*x[0]+(0.394792632318)*x[1]**o+(-0.622944802336)*x[1]
        s[4,1,0]=(-0.677909127453)*x[0]**o+(0.396645756207)*x[0]+(-0.723263540777)*x[1]**o+(-0.339508447056)*x[1]
        s[4,1,1]=(-0.807950411362)*x[0]**o+(-0.149082460629)*x[0]+(-0.841669340631)*x[1]**o+(0.888900711982)*x[1]
        s[5,0,0]=(-0.560462354311)*x[0]**o+(0.00127986811316)*x[0]+(-0.21694183911)*x[1]**o+(-0.815224807309)*x[1]
        s[5,0,1]=(0.767918970281)*x[0]**o+(-0.0975926175116)*x[0]+(0.281160215987)*x[1]**o+(0.38258099137)*x[1]
        s[5,1,0]=(-0.0755717349023)*x[0]**o+(-0.321990446321)*x[0]+(-0.203426543697)*x[1]**o+(0.782003123159)*x[1]
        s[5,1,1]=(-0.947191134408)*x[0]**o+(-0.710118492404)*x[0]+(-0.0923673220139)*x[1]**o+(-0.103352087028)*x[1]
        ref[0,0,0]=(-0.492415630764)*(1+2.*(dim-1.)/(o+1.))+(0.999799690972)*dim
        ref[0,0,1]=(-0.942115997065)*(1+2.*(dim-1.)/(o+1.))+(1.04963428187)*dim
        ref[0,1,0]=(0.0156415045147)*(1+2.*(dim-1.)/(o+1.))+(0.84445836586)*dim
        ref[0,1,1]=(0.935583806465)*(1+2.*(dim-1.)/(o+1.))+(0.692196459814)*dim
        ref[1,0,0]=(1.08881369519)*(1+2.*(dim-1.)/(o+1.))+(0.727145708632)*dim
        ref[1,0,1]=(-0.252173204638)*(1+2.*(dim-1.)/(o+1.))+(1.08152066636)*dim
        ref[1,1,0]=(-0.642365887101)*(1+2.*(dim-1.)/(o+1.))+(-1.48933990638)*dim
        ref[1,1,1]=(0.341252740787)*(1+2.*(dim-1.)/(o+1.))+(0.451909406552)*dim
        ref[2,0,0]=(1.32771177127)*(1+2.*(dim-1.)/(o+1.))+(-0.238838242309)*dim
        ref[2,0,1]=(0.598974191766)*(1+2.*(dim-1.)/(o+1.))+(-0.900569577663)*dim
        ref[2,1,0]=(-0.822142164446)*(1+2.*(dim-1.)/(o+1.))+(-0.0576553574825)*dim
        ref[2,1,1]=(0.322838750825)*(1+2.*(dim-1.)/(o+1.))+(1.50526603969)*dim
        ref[3,0,0]=(-0.919544759776)*(1+2.*(dim-1.)/(o+1.))+(1.57727841933)*dim
        ref[3,0,1]=(0.646715796109)*(1+2.*(dim-1.)/(o+1.))+(1.81637499436)*dim
        ref[3,1,0]=(0.263684903118)*(1+2.*(dim-1.)/(o+1.))+(0.496837383795)*dim
        ref[3,1,1]=(-0.39761816615)*(1+2.*(dim-1.)/(o+1.))+(0.673687942776)*dim
        ref[4,0,0]=(-0.407617749907)*(1+2.*(dim-1.)/(o+1.))+(-0.157810809613)*dim
        ref[4,0,1]=(0.141881490647)*(1+2.*(dim-1.)/(o+1.))+(-0.122959669416)*dim
        ref[4,1,0]=(-1.40117266823)*(1+2.*(dim-1.)/(o+1.))+(0.0571373091516)*dim
        ref[4,1,1]=(-1.64961975199)*(1+2.*(dim-1.)/(o+1.))+(0.739818251353)*dim
        ref[5,0,0]=(-0.777404193421)*(1+2.*(dim-1.)/(o+1.))+(-0.813944939196)*dim
        ref[5,0,1]=(1.04907918627)*(1+2.*(dim-1.)/(o+1.))+(0.284988373859)*dim
        ref[5,1,0]=(-0.278998278599)*(1+2.*(dim-1.)/(o+1.))+(0.460012676837)*dim
        ref[5,1,1]=(-1.03955845642)*(1+2.*(dim-1.)/(o+1.))+(-0.813470579432)*dim
      else:
        s[0,0,0]=(0.887591619739)*x[0]**o+(0.407256482081)*x[0]+(0.382561775603)*x[1]**o+(0.170135617346)*x[1]+(-0.745510574059)*x[2]**o+(-0.455506612867)*x[2]
        s[0,0,1]=(-0.906429753951)*x[0]**o+(0.650007305971)*x[0]+(0.679810544399)*x[1]**o+(-0.327318089388)*x[1]+(-0.682816086768)*x[2]**o+(0.213571445023)*x[2]
        s[0,1,0]=(-0.101321617665)*x[0]**o+(-0.830248450523)*x[0]+(0.968297844076)*x[1]**o+(-0.461609225074)*x[1]+(-0.803382287303)*x[2]**o+(-0.264990872553)*x[2]
        s[0,1,1]=(0.762025047301)*x[0]**o+(0.352751067716)*x[0]+(0.085082244422)*x[1]**o+(0.260281207839)*x[1]+(-0.838657429282)*x[2]**o+(-0.449142572742)*x[2]
        s[1,0,0]=(0.60404314316)*x[0]**o+(0.433627969764)*x[0]+(-0.187903979441)*x[1]**o+(-0.490170615337)*x[1]+(0.854829687117)*x[2]**o+(-0.0562457996778)*x[2]
        s[1,0,1]=(0.379185262336)*x[0]**o+(-0.127987724933)*x[0]+(-0.199149895558)*x[1]**o+(0.177834623561)*x[1]+(-0.528371897238)*x[2]**o+(-0.752877478181)*x[2]
        s[1,1,0]=(-0.639722907848)*x[0]**o+(0.595935571051)*x[0]+(0.660417028114)*x[1]**o+(-0.70337252144)*x[1]+(0.0532141342539)*x[2]**o+(0.335609331703)*x[2]
        s[1,1,1]=(0.918579942175)*x[0]**o+(-0.419345002272)*x[0]+(-0.733731200432)*x[1]**o+(-0.926576582075)*x[1]+(-0.596748064397)*x[2]**o+(-0.994146900668)*x[2]
        s[2,0,0]=(0.465936668326)*x[0]**o+(0.146304452558)*x[0]+(0.434641974102)*x[1]**o+(0.7475528675)*x[1]+(-0.9968692715)*x[2]**o+(0.761573606267)*x[2]
        s[2,0,1]=(-0.19499782219)*x[0]**o+(-0.0487068806171)*x[0]+(-0.466751862198)*x[1]**o+(0.748775488587)*x[1]+(-0.159041931091)*x[2]**o+(-0.786654864976)*x[2]
        s[2,1,0]=(0.497506698992)*x[0]**o+(0.00446974112015)*x[0]+(-0.57139775079)*x[1]**o+(0.794232026784)*x[1]+(0.788382405769)*x[2]**o+(-0.611606691539)*x[2]
        s[2,1,1]=(-0.744517152085)*x[0]**o+(0.826326809885)*x[0]+(-0.489647614501)*x[1]**o+(-0.798398764067)*x[1]+(0.622474900672)*x[2]**o+(-0.958167077873)*x[2]
        s[3,0,0]=(-0.54850657455)*x[0]**o+(0.907733944445)*x[0]+(0.150662161005)*x[1]**o+(0.329278515834)*x[1]+(0.612014869233)*x[2]**o+(-0.332197388537)*x[2]
        s[3,0,1]=(-0.977591157702)*x[0]**o+(0.0314123979019)*x[0]+(0.357325837849)*x[1]**o+(-0.121203337609)*x[1]+(-0.417779485304)*x[2]**o+(0.56339482028)*x[2]
        s[3,1,0]=(-0.0882818794492)*x[0]**o+(0.373727429506)*x[0]+(0.251705556572)*x[1]**o+(0.75208764669)*x[1]+(-0.243231080216)*x[2]**o+(0.408665593633)*x[2]
        s[3,1,1]=(0.902672970131)*x[0]**o+(0.0816189076234)*x[0]+(-0.0188617954975)*x[1]**o+(0.248727356198)*x[1]+(0.659217806257)*x[2]**o+(-0.521141634196)*x[2]
        s[4,0,0]=(-0.884907344323)*x[0]**o+(0.284225021738)*x[0]+(0.810073434041)*x[1]**o+(0.267855075455)*x[1]+(0.0897833985076)*x[2]**o+(0.622374196423)*x[2]
        s[4,0,1]=(0.512127530167)*x[0]**o+(-0.0208132068954)*x[0]+(0.258260560413)*x[1]**o+(0.206670909439)*x[1]+(-0.359616018773)*x[2]**o+(0.538953196654)*x[2]
        s[4,1,0]=(-0.775312439486)*x[0]**o+(0.781309943125)*x[0]+(0.277745830089)*x[1]**o+(-0.166968517573)*x[1]+(0.428260154734)*x[2]**o+(0.753276413709)*x[2]
        s[4,1,1]=(-0.575086193799)*x[0]**o+(0.488745961307)*x[0]+(0.828550042541)*x[1]**o+(-0.82220696841)*x[1]+(0.48247983233)*x[2]**o+(0.348574290989)*x[2]
        s[5,0,0]=(0.904922447578)*x[0]**o+(0.94372036658)*x[0]+(0.554305596085)*x[1]**o+(-0.633630196279)*x[1]+(0.0304645794125)*x[2]**o+(-0.338076751695)*x[2]
        s[5,0,1]=(0.157663141664)*x[0]**o+(0.27698264503)*x[0]+(-0.244811019178)*x[1]**o+(-0.86016833799)*x[1]+(0.0037249377228)*x[2]**o+(0.415013895956)*x[2]
        s[5,1,0]=(-0.0651731085463)*x[0]**o+(0.656433177612)*x[0]+(-0.693226351108)*x[1]**o+(-0.195335968106)*x[1]+(0.735806703428)*x[2]**o+(-0.284744282673)*x[2]
        s[5,1,1]=(0.14124610976)*x[0]**o+(0.114249449125)*x[0]+(-0.89675797948)*x[1]**o+(-0.109343809939)*x[1]+(-0.171059459418)*x[2]**o+(-0.125911673497)*x[2]
        ref[0,0,0]=(0.524642821284)*(1+2.*(dim-1.)/(o+1.))+(0.12188548656)*dim
        ref[0,0,1]=(-0.90943529632)*(1+2.*(dim-1.)/(o+1.))+(0.536260661607)*dim
        ref[0,1,0]=(0.0635939391079)*(1+2.*(dim-1.)/(o+1.))+(-1.55684854815)*dim
        ref[0,1,1]=(0.0084498624415)*(1+2.*(dim-1.)/(o+1.))+(0.163889702813)*dim
        ref[1,0,0]=(1.27096885084)*(1+2.*(dim-1.)/(o+1.))+(-0.112788445251)*dim
        ref[1,0,1]=(-0.34833653046)*(1+2.*(dim-1.)/(o+1.))+(-0.703030579553)*dim
        ref[1,1,0]=(0.0739082545202)*(1+2.*(dim-1.)/(o+1.))+(0.228172381314)*dim
        ref[1,1,1]=(-0.411899322654)*(1+2.*(dim-1.)/(o+1.))+(-2.34006848501)*dim
        ref[2,0,0]=(-0.0962906290716)*(1+2.*(dim-1.)/(o+1.))+(1.65543092632)*dim
        ref[2,0,1]=(-0.820791615479)*(1+2.*(dim-1.)/(o+1.))+(-0.0865862570057)*dim
        ref[2,1,0]=(0.714491353971)*(1+2.*(dim-1.)/(o+1.))+(0.187095076365)*dim
        ref[2,1,1]=(-0.611689865914)*(1+2.*(dim-1.)/(o+1.))+(-0.930239032054)*dim
        ref[3,0,0]=(0.214170455688)*(1+2.*(dim-1.)/(o+1.))+(0.904815071742)*dim
        ref[3,0,1]=(-1.03804480516)*(1+2.*(dim-1.)/(o+1.))+(0.473603880573)*dim
        ref[3,1,0]=(-0.0798074030926)*(1+2.*(dim-1.)/(o+1.))+(1.53448066983)*dim
        ref[3,1,1]=(1.54302898089)*(1+2.*(dim-1.)/(o+1.))+(-0.190795370375)*dim
        ref[4,0,0]=(0.0149494882254)*(1+2.*(dim-1.)/(o+1.))+(1.17445429362)*dim
        ref[4,0,1]=(0.410772071806)*(1+2.*(dim-1.)/(o+1.))+(0.724810899198)*dim
        ref[4,1,0]=(-0.0693064546636)*(1+2.*(dim-1.)/(o+1.))+(1.36761783926)*dim
        ref[4,1,1]=(0.735943681072)*(1+2.*(dim-1.)/(o+1.))+(0.0151132838872)*dim
        ref[5,0,0]=(1.48969262308)*(1+2.*(dim-1.)/(o+1.))+(-0.0279865813938)*dim
        ref[5,0,1]=(-0.0834229397911)*(1+2.*(dim-1.)/(o+1.))+(-0.168171797005)*dim
        ref[5,1,0]=(-0.0225927562259)*(1+2.*(dim-1.)/(o+1.))+(0.176352926834)*dim
        ref[5,1,1]=(-0.926571329138)*(1+2.*(dim-1.)/(o+1.))+(-0.121006034312)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunctionOnBoundary

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-0.0564495520087)*x[0]**o+(0.322467947613)*x[0]+(-0.426040737283)*x[1]**o+(0.743030997838)*x[1]
        s[0,0,0,1]=(-0.705100478682)*x[0]**o+(0.0428011755277)*x[0]+(-0.667969683035)*x[1]**o+(-0.05268016216)*x[1]
        s[0,0,1,0]=(-0.646229122275)*x[0]**o+(-0.117856575975)*x[0]+(-0.928260620991)*x[1]**o+(-0.0295789972002)*x[1]
        s[0,0,1,1]=(0.684057019666)*x[0]**o+(0.217883017426)*x[0]+(-0.8003965891)*x[1]**o+(0.805481050427)*x[1]
        s[0,0,2,0]=(-0.517102518951)*x[0]**o+(0.703108669274)*x[0]+(-0.0818548881221)*x[1]**o+(0.967661353324)*x[1]
        s[0,0,2,1]=(-0.311422038099)*x[0]**o+(0.647051468428)*x[0]+(-0.0134135628666)*x[1]**o+(0.401772150809)*x[1]
        s[0,1,0,0]=(-0.372031418539)*x[0]**o+(-0.348570224054)*x[0]+(0.971394255739)*x[1]**o+(0.220452856894)*x[1]
        s[0,1,0,1]=(0.726171816717)*x[0]**o+(0.0115054389412)*x[0]+(-0.663899850543)*x[1]**o+(-0.00862058334168)*x[1]
        s[0,1,1,0]=(0.0546574475678)*x[0]**o+(-0.110568027517)*x[0]+(-0.490412123307)*x[1]**o+(0.70968579481)*x[1]
        s[0,1,1,1]=(0.525121644008)*x[0]**o+(-0.460449092422)*x[0]+(-0.105141027916)*x[1]**o+(0.70599530088)*x[1]
        s[0,1,2,0]=(-0.333393404511)*x[0]**o+(0.956105748878)*x[0]+(-0.542294839284)*x[1]**o+(0.856767614724)*x[1]
        s[0,1,2,1]=(0.499018678584)*x[0]**o+(0.709512775393)*x[0]+(-0.0278608905417)*x[1]**o+(-0.408880317793)*x[1]
        s[0,2,0,0]=(0.196709165539)*x[0]**o+(-0.824269904377)*x[0]+(-0.776756196058)*x[1]**o+(-0.21242798078)*x[1]
        s[0,2,0,1]=(0.101174745892)*x[0]**o+(0.103180763582)*x[0]+(-0.25979513904)*x[1]**o+(-0.595455963398)*x[1]
        s[0,2,1,0]=(-0.490524427754)*x[0]**o+(0.22473843241)*x[0]+(-0.559043751054)*x[1]**o+(-0.396861202642)*x[1]
        s[0,2,1,1]=(0.366671902637)*x[0]**o+(0.00444476264911)*x[0]+(0.84769491016)*x[1]**o+(0.340088796185)*x[1]
        s[0,2,2,0]=(0.0843962714222)*x[0]**o+(-0.678807745661)*x[0]+(0.0710532661834)*x[1]**o+(-0.131143081822)*x[1]
        s[0,2,2,1]=(-0.505967901175)*x[0]**o+(0.389944048429)*x[0]+(-0.607195013377)*x[1]**o+(0.615596624068)*x[1]
        s[0,3,0,0]=(0.480464466311)*x[0]**o+(-0.200660236866)*x[0]+(-0.366683832098)*x[1]**o+(0.473029292775)*x[1]
        s[0,3,0,1]=(0.503865126139)*x[0]**o+(0.620947845674)*x[0]+(0.482553725938)*x[1]**o+(0.903952597709)*x[1]
        s[0,3,1,0]=(-0.935816186601)*x[0]**o+(-0.694323019844)*x[0]+(-0.222643250757)*x[1]**o+(-0.61395534471)*x[1]
        s[0,3,1,1]=(-0.366248207885)*x[0]**o+(0.21981177916)*x[0]+(0.829964990663)*x[1]**o+(-0.458138466604)*x[1]
        s[0,3,2,0]=(-0.133505096099)*x[0]**o+(0.889073061727)*x[0]+(0.748032746123)*x[1]**o+(0.457372622346)*x[1]
        s[0,3,2,1]=(-0.940388978849)*x[0]**o+(0.194505986498)*x[0]+(0.277476781766)*x[1]**o+(0.0315390566538)*x[1]
        s[0,4,0,0]=(-0.246674730207)*x[0]**o+(-0.145685680956)*x[0]+(-0.855397050636)*x[1]**o+(0.332401911997)*x[1]
        s[0,4,0,1]=(-0.786787596226)*x[0]**o+(0.000178838426883)*x[0]+(0.0315355043243)*x[1]**o+(0.502956944109)*x[1]
        s[0,4,1,0]=(0.456353731831)*x[0]**o+(0.224571019711)*x[0]+(0.787432530547)*x[1]**o+(0.900715775247)*x[1]
        s[0,4,1,1]=(-0.87764927624)*x[0]**o+(0.547980575611)*x[0]+(-0.287094420401)*x[1]**o+(0.664160174901)*x[1]
        s[0,4,2,0]=(-0.195894642929)*x[0]**o+(-0.595129209807)*x[0]+(0.865849102635)*x[1]**o+(-0.677159397082)*x[1]
        s[0,4,2,1]=(-0.841735734151)*x[0]**o+(0.625862078643)*x[0]+(-0.545516568256)*x[1]**o+(0.114109991331)*x[1]
        s[1,0,0,0]=(0.209458674327)*x[0]**o+(-0.20896563634)*x[0]+(-0.579063332782)*x[1]**o+(-0.0433216462031)*x[1]
        s[1,0,0,1]=(0.0646951342633)*x[0]**o+(-0.999495478916)*x[0]+(-0.679168301408)*x[1]**o+(-0.357289498732)*x[1]
        s[1,0,1,0]=(0.19078385999)*x[0]**o+(0.14391750283)*x[0]+(-0.571814090672)*x[1]**o+(0.19532813646)*x[1]
        s[1,0,1,1]=(0.3385444312)*x[0]**o+(-0.961513536858)*x[0]+(-0.729134108689)*x[1]**o+(-0.477737542344)*x[1]
        s[1,0,2,0]=(0.545433953869)*x[0]**o+(-0.746630455132)*x[0]+(-0.49008038028)*x[1]**o+(-0.156535798648)*x[1]
        s[1,0,2,1]=(-0.718991145693)*x[0]**o+(-0.264354651201)*x[0]+(0.0812886298976)*x[1]**o+(-0.331692687243)*x[1]
        s[1,1,0,0]=(-0.646431085196)*x[0]**o+(-0.150073929112)*x[0]+(-0.919422164553)*x[1]**o+(-0.754093106669)*x[1]
        s[1,1,0,1]=(-0.424167875207)*x[0]**o+(0.00436241598561)*x[0]+(0.423024130972)*x[1]**o+(0.0414727269723)*x[1]
        s[1,1,1,0]=(0.972554573539)*x[0]**o+(-0.00871610987967)*x[0]+(-0.968130421895)*x[1]**o+(0.323276868881)*x[1]
        s[1,1,1,1]=(-0.223846361715)*x[0]**o+(0.0984359694216)*x[0]+(0.367109095942)*x[1]**o+(0.39421742616)*x[1]
        s[1,1,2,0]=(-0.532854459841)*x[0]**o+(0.0932005176653)*x[0]+(0.300915725885)*x[1]**o+(-0.407779066395)*x[1]
        s[1,1,2,1]=(-0.977204985435)*x[0]**o+(0.632461453595)*x[0]+(0.222769752858)*x[1]**o+(0.97524903541)*x[1]
        s[1,2,0,0]=(0.798817773565)*x[0]**o+(-0.687809695378)*x[0]+(-0.77661557915)*x[1]**o+(0.242848730169)*x[1]
        s[1,2,0,1]=(0.852738374774)*x[0]**o+(-0.353097828509)*x[0]+(-0.947070627494)*x[1]**o+(0.686527427094)*x[1]
        s[1,2,1,0]=(0.315473111735)*x[0]**o+(0.671113387299)*x[0]+(0.328881731518)*x[1]**o+(-0.269227909853)*x[1]
        s[1,2,1,1]=(-0.877500814779)*x[0]**o+(-0.508518061024)*x[0]+(-0.486367952918)*x[1]**o+(0.465181865784)*x[1]
        s[1,2,2,0]=(0.321396065376)*x[0]**o+(-0.977515734724)*x[0]+(-0.745047664798)*x[1]**o+(0.00770829984163)*x[1]
        s[1,2,2,1]=(0.217194865535)*x[0]**o+(0.84002706514)*x[0]+(0.4186527058)*x[1]**o+(-0.0197865947851)*x[1]
        s[1,3,0,0]=(-0.434267890176)*x[0]**o+(0.677387101394)*x[0]+(-0.433143616343)*x[1]**o+(0.76079607685)*x[1]
        s[1,3,0,1]=(-0.568019213241)*x[0]**o+(0.940292544942)*x[0]+(0.351503513272)*x[1]**o+(-0.8227358798)*x[1]
        s[1,3,1,0]=(0.225598112956)*x[0]**o+(-0.158411926824)*x[0]+(-0.0770214720299)*x[1]**o+(0.667242375231)*x[1]
        s[1,3,1,1]=(0.497365374413)*x[0]**o+(-0.402552536909)*x[0]+(-0.760975595637)*x[1]**o+(0.423931229263)*x[1]
        s[1,3,2,0]=(0.260548515487)*x[0]**o+(-0.723043083794)*x[0]+(0.40423813784)*x[1]**o+(0.591112253369)*x[1]
        s[1,3,2,1]=(0.0657703077553)*x[0]**o+(0.610016697083)*x[0]+(-0.515291622078)*x[1]**o+(-0.0821391959295)*x[1]
        s[1,4,0,0]=(-0.0842960021063)*x[0]**o+(0.830851714471)*x[0]+(0.571649443759)*x[1]**o+(-0.884015388056)*x[1]
        s[1,4,0,1]=(0.681962300417)*x[0]**o+(0.549746462866)*x[0]+(0.0330915607536)*x[1]**o+(0.757109551626)*x[1]
        s[1,4,1,0]=(0.413804087081)*x[0]**o+(-0.0713221635854)*x[0]+(0.153990572806)*x[1]**o+(-0.448202273823)*x[1]
        s[1,4,1,1]=(-0.109978839431)*x[0]**o+(-0.720641165805)*x[0]+(-0.465701345584)*x[1]**o+(-0.423310622549)*x[1]
        s[1,4,2,0]=(0.0884455154661)*x[0]**o+(0.341657397119)*x[0]+(-0.282148805378)*x[1]**o+(0.567581544922)*x[1]
        s[1,4,2,1]=(0.384270691665)*x[0]**o+(0.986879879066)*x[0]+(0.725843240631)*x[1]**o+(-0.888932402046)*x[1]
        s[2,0,0,0]=(0.915028230934)*x[0]**o+(0.0308158620736)*x[0]+(-0.534330519368)*x[1]**o+(-0.668490082908)*x[1]
        s[2,0,0,1]=(0.0447246773914)*x[0]**o+(-0.764286683802)*x[0]+(0.985272626016)*x[1]**o+(-0.582955807402)*x[1]
        s[2,0,1,0]=(-0.0982089493146)*x[0]**o+(-0.119886638989)*x[0]+(0.0918316786338)*x[1]**o+(-0.275013043117)*x[1]
        s[2,0,1,1]=(-0.187529321909)*x[0]**o+(0.715634182891)*x[0]+(0.554464679108)*x[1]**o+(0.75060799895)*x[1]
        s[2,0,2,0]=(0.402781810737)*x[0]**o+(-0.885441671822)*x[0]+(0.647449131552)*x[1]**o+(-0.82134489597)*x[1]
        s[2,0,2,1]=(0.903766197812)*x[0]**o+(0.397983290493)*x[0]+(-0.161556012403)*x[1]**o+(0.441970897)*x[1]
        s[2,1,0,0]=(-0.0286858171901)*x[0]**o+(0.152444209719)*x[0]+(0.713144598152)*x[1]**o+(-0.330803066479)*x[1]
        s[2,1,0,1]=(0.221312544677)*x[0]**o+(0.76681848391)*x[0]+(0.4237403202)*x[1]**o+(0.219978904876)*x[1]
        s[2,1,1,0]=(0.350534590086)*x[0]**o+(0.933385654173)*x[0]+(0.349813301929)*x[1]**o+(0.0427424949923)*x[1]
        s[2,1,1,1]=(0.660478429879)*x[0]**o+(-0.487740867789)*x[0]+(-0.270542163333)*x[1]**o+(-0.405080985084)*x[1]
        s[2,1,2,0]=(0.0114890380351)*x[0]**o+(-0.156620602559)*x[0]+(-0.494024071381)*x[1]**o+(0.164138650655)*x[1]
        s[2,1,2,1]=(0.688285293331)*x[0]**o+(-0.723548244407)*x[0]+(-0.979169222862)*x[1]**o+(-0.815167412147)*x[1]
        s[2,2,0,0]=(0.743910713659)*x[0]**o+(-0.319702207578)*x[0]+(0.464318179749)*x[1]**o+(-0.911256589955)*x[1]
        s[2,2,0,1]=(0.178943711133)*x[0]**o+(0.878989967814)*x[0]+(-0.584920241418)*x[1]**o+(-0.232803309459)*x[1]
        s[2,2,1,0]=(0.289708048016)*x[0]**o+(0.0250010477203)*x[0]+(0.231695416136)*x[1]**o+(0.00458067191189)*x[1]
        s[2,2,1,1]=(0.252996966945)*x[0]**o+(0.718948618791)*x[0]+(0.238406858981)*x[1]**o+(-0.222687536947)*x[1]
        s[2,2,2,0]=(0.707394678426)*x[0]**o+(-0.341044895187)*x[0]+(-0.924768233451)*x[1]**o+(-0.717376693436)*x[1]
        s[2,2,2,1]=(0.826870828657)*x[0]**o+(0.596221683232)*x[0]+(0.13184086469)*x[1]**o+(-0.122283119637)*x[1]
        s[2,3,0,0]=(0.082622085842)*x[0]**o+(0.802466780483)*x[0]+(-0.374536755651)*x[1]**o+(0.299695470983)*x[1]
        s[2,3,0,1]=(0.806722430379)*x[0]**o+(-0.307373050709)*x[0]+(-0.504241501795)*x[1]**o+(0.0365801039206)*x[1]
        s[2,3,1,0]=(-0.189680944199)*x[0]**o+(0.86059525782)*x[0]+(0.612911991629)*x[1]**o+(-0.868979910208)*x[1]
        s[2,3,1,1]=(-0.0606711884277)*x[0]**o+(-0.543604644984)*x[0]+(0.292808478791)*x[1]**o+(0.0024260857339)*x[1]
        s[2,3,2,0]=(0.258222812918)*x[0]**o+(0.304770046568)*x[0]+(-0.426946619435)*x[1]**o+(-0.0521912646177)*x[1]
        s[2,3,2,1]=(-0.324355225171)*x[0]**o+(-0.784048371347)*x[0]+(0.169611573731)*x[1]**o+(-0.548169504296)*x[1]
        s[2,4,0,0]=(0.0155438234769)*x[0]**o+(-0.620846826317)*x[0]+(0.409826337485)*x[1]**o+(0.129190074563)*x[1]
        s[2,4,0,1]=(0.923545504209)*x[0]**o+(-0.155101945409)*x[0]+(-0.717093202578)*x[1]**o+(-0.480920400256)*x[1]
        s[2,4,1,0]=(0.313709415121)*x[0]**o+(0.736484420744)*x[0]+(-0.0293655314952)*x[1]**o+(-0.599267629638)*x[1]
        s[2,4,1,1]=(0.463358526622)*x[0]**o+(-0.592469734442)*x[0]+(0.613104805703)*x[1]**o+(-0.178209835993)*x[1]
        s[2,4,2,0]=(-0.713644697339)*x[0]**o+(0.799721953045)*x[0]+(0.0589328481367)*x[1]**o+(0.547328786304)*x[1]
        s[2,4,2,1]=(0.993775334243)*x[0]**o+(0.553815998267)*x[0]+(0.1992862938)*x[1]**o+(0.81025353823)*x[1]
        s[3,0,0,0]=(0.160105719249)*x[0]**o+(-0.555881043013)*x[0]+(0.595178125526)*x[1]**o+(0.581499038384)*x[1]
        s[3,0,0,1]=(-0.073553285882)*x[0]**o+(0.302577114557)*x[0]+(0.170358740039)*x[1]**o+(0.75537952732)*x[1]
        s[3,0,1,0]=(-0.64563114781)*x[0]**o+(0.196969117846)*x[0]+(-0.972196629648)*x[1]**o+(-0.996586995934)*x[1]
        s[3,0,1,1]=(0.833966809791)*x[0]**o+(0.95981521797)*x[0]+(0.816649660191)*x[1]**o+(-0.220324291634)*x[1]
        s[3,0,2,0]=(0.102957823139)*x[0]**o+(-0.442375614091)*x[0]+(0.686402985053)*x[1]**o+(0.0731364854586)*x[1]
        s[3,0,2,1]=(-0.168212668775)*x[0]**o+(0.314439484314)*x[0]+(0.619151838083)*x[1]**o+(-0.360098253956)*x[1]
        s[3,1,0,0]=(-0.416264395501)*x[0]**o+(0.214984128512)*x[0]+(0.358642811015)*x[1]**o+(-0.869373823201)*x[1]
        s[3,1,0,1]=(0.181499327771)*x[0]**o+(-0.331648488714)*x[0]+(-0.191432513473)*x[1]**o+(0.0930773945136)*x[1]
        s[3,1,1,0]=(-0.976247847012)*x[0]**o+(-0.575393818806)*x[0]+(0.0929450161482)*x[1]**o+(-0.276689604101)*x[1]
        s[3,1,1,1]=(0.185352455216)*x[0]**o+(-0.931509714205)*x[0]+(-0.6870722095)*x[1]**o+(0.0624138101885)*x[1]
        s[3,1,2,0]=(-0.281042580465)*x[0]**o+(0.966028964753)*x[0]+(-0.419450359271)*x[1]**o+(0.873992760636)*x[1]
        s[3,1,2,1]=(0.993820925583)*x[0]**o+(0.690209908944)*x[0]+(0.148117153189)*x[1]**o+(0.71818108405)*x[1]
        s[3,2,0,0]=(-0.549354119752)*x[0]**o+(-0.282097817235)*x[0]+(0.896456051498)*x[1]**o+(-0.452485791074)*x[1]
        s[3,2,0,1]=(-0.0830679821316)*x[0]**o+(0.40212443752)*x[0]+(0.380555934515)*x[1]**o+(0.987505234574)*x[1]
        s[3,2,1,0]=(0.997713740817)*x[0]**o+(-0.692039069659)*x[0]+(-0.565960747782)*x[1]**o+(0.445414151858)*x[1]
        s[3,2,1,1]=(0.9675137078)*x[0]**o+(0.84769422726)*x[0]+(0.0938577014253)*x[1]**o+(-0.431299521474)*x[1]
        s[3,2,2,0]=(-0.868876045235)*x[0]**o+(-0.955493151846)*x[0]+(-0.9143773492)*x[1]**o+(0.899030565498)*x[1]
        s[3,2,2,1]=(0.629496450112)*x[0]**o+(-0.0218037231989)*x[0]+(-0.933196760824)*x[1]**o+(-0.549980925362)*x[1]
        s[3,3,0,0]=(-0.73789282568)*x[0]**o+(0.612764312125)*x[0]+(-0.80912954116)*x[1]**o+(0.608837621314)*x[1]
        s[3,3,0,1]=(0.465678249887)*x[0]**o+(-0.780539933825)*x[0]+(0.149236887279)*x[1]**o+(-0.842449293644)*x[1]
        s[3,3,1,0]=(-0.853285341888)*x[0]**o+(0.347485246788)*x[0]+(0.219773434631)*x[1]**o+(0.702046500252)*x[1]
        s[3,3,1,1]=(-0.0500921104166)*x[0]**o+(0.620181272972)*x[0]+(-0.514576066373)*x[1]**o+(0.0487351718037)*x[1]
        s[3,3,2,0]=(0.0904463554055)*x[0]**o+(0.268365306318)*x[0]+(-0.164684341011)*x[1]**o+(0.115207872226)*x[1]
        s[3,3,2,1]=(0.359085658137)*x[0]**o+(0.144112368075)*x[0]+(0.0650644733865)*x[1]**o+(0.176937328733)*x[1]
        s[3,4,0,0]=(-0.602904552775)*x[0]**o+(-0.648691919828)*x[0]+(-0.41237635973)*x[1]**o+(-0.282240953949)*x[1]
        s[3,4,0,1]=(-0.938255401084)*x[0]**o+(0.263050201557)*x[0]+(0.175643545166)*x[1]**o+(-0.995029122689)*x[1]
        s[3,4,1,0]=(0.728511178435)*x[0]**o+(-0.0673148421745)*x[0]+(-0.814903409759)*x[1]**o+(0.0229206126733)*x[1]
        s[3,4,1,1]=(-0.790381209272)*x[0]**o+(-0.326044592264)*x[0]+(-0.616797433118)*x[1]**o+(-0.0963030542755)*x[1]
        s[3,4,2,0]=(-0.286555245913)*x[0]**o+(-0.725972210674)*x[0]+(-0.809797611215)*x[1]**o+(0.477382881829)*x[1]
        s[3,4,2,1]=(0.947314594641)*x[0]**o+(0.731567714373)*x[0]+(0.104324647674)*x[1]**o+(0.0771726257761)*x[1]
        ref[0,0,0,0]=(-0.482490289291)*(1+2.*(dim-1.)/(o+1.))+(1.06549894545)*dim
        ref[0,0,0,1]=(-1.37307016172)*(1+2.*(dim-1.)/(o+1.))+(-0.00987898663227)*dim
        ref[0,0,1,0]=(-1.57448974327)*(1+2.*(dim-1.)/(o+1.))+(-0.147435573175)*dim
        ref[0,0,1,1]=(-0.116339569434)*(1+2.*(dim-1.)/(o+1.))+(1.02336406785)*dim
        ref[0,0,2,0]=(-0.598957407073)*(1+2.*(dim-1.)/(o+1.))+(1.6707700226)*dim
        ref[0,0,2,1]=(-0.324835600966)*(1+2.*(dim-1.)/(o+1.))+(1.04882361924)*dim
        ref[0,1,0,0]=(0.5993628372)*(1+2.*(dim-1.)/(o+1.))+(-0.128117367161)*dim
        ref[0,1,0,1]=(0.062271966174)*(1+2.*(dim-1.)/(o+1.))+(0.00288485559949)*dim
        ref[0,1,1,0]=(-0.43575467574)*(1+2.*(dim-1.)/(o+1.))+(0.599117767293)*dim
        ref[0,1,1,1]=(0.419980616092)*(1+2.*(dim-1.)/(o+1.))+(0.245546208457)*dim
        ref[0,1,2,0]=(-0.875688243795)*(1+2.*(dim-1.)/(o+1.))+(1.8128733636)*dim
        ref[0,1,2,1]=(0.471157788042)*(1+2.*(dim-1.)/(o+1.))+(0.3006324576)*dim
        ref[0,2,0,0]=(-0.580047030519)*(1+2.*(dim-1.)/(o+1.))+(-1.03669788516)*dim
        ref[0,2,0,1]=(-0.158620393148)*(1+2.*(dim-1.)/(o+1.))+(-0.492275199816)*dim
        ref[0,2,1,0]=(-1.04956817881)*(1+2.*(dim-1.)/(o+1.))+(-0.172122770232)*dim
        ref[0,2,1,1]=(1.2143668128)*(1+2.*(dim-1.)/(o+1.))+(0.344533558834)*dim
        ref[0,2,2,0]=(0.155449537606)*(1+2.*(dim-1.)/(o+1.))+(-0.809950827483)*dim
        ref[0,2,2,1]=(-1.11316291455)*(1+2.*(dim-1.)/(o+1.))+(1.0055406725)*dim
        ref[0,3,0,0]=(0.113780634213)*(1+2.*(dim-1.)/(o+1.))+(0.272369055909)*dim
        ref[0,3,0,1]=(0.986418852077)*(1+2.*(dim-1.)/(o+1.))+(1.52490044338)*dim
        ref[0,3,1,0]=(-1.15845943736)*(1+2.*(dim-1.)/(o+1.))+(-1.30827836455)*dim
        ref[0,3,1,1]=(0.463716782779)*(1+2.*(dim-1.)/(o+1.))+(-0.238326687444)*dim
        ref[0,3,2,0]=(0.614527650024)*(1+2.*(dim-1.)/(o+1.))+(1.34644568407)*dim
        ref[0,3,2,1]=(-0.662912197084)*(1+2.*(dim-1.)/(o+1.))+(0.226045043152)*dim
        ref[0,4,0,0]=(-1.10207178084)*(1+2.*(dim-1.)/(o+1.))+(0.186716231041)*dim
        ref[0,4,0,1]=(-0.755252091901)*(1+2.*(dim-1.)/(o+1.))+(0.503135782536)*dim
        ref[0,4,1,0]=(1.24378626238)*(1+2.*(dim-1.)/(o+1.))+(1.12528679496)*dim
        ref[0,4,1,1]=(-1.16474369664)*(1+2.*(dim-1.)/(o+1.))+(1.21214075051)*dim
        ref[0,4,2,0]=(0.669954459707)*(1+2.*(dim-1.)/(o+1.))+(-1.27228860689)*dim
        ref[0,4,2,1]=(-1.38725230241)*(1+2.*(dim-1.)/(o+1.))+(0.739972069974)*dim
        ref[1,0,0,0]=(-0.369604658454)*(1+2.*(dim-1.)/(o+1.))+(-0.252287282543)*dim
        ref[1,0,0,1]=(-0.614473167144)*(1+2.*(dim-1.)/(o+1.))+(-1.35678497765)*dim
        ref[1,0,1,0]=(-0.381030230682)*(1+2.*(dim-1.)/(o+1.))+(0.339245639291)*dim
        ref[1,0,1,1]=(-0.39058967749)*(1+2.*(dim-1.)/(o+1.))+(-1.4392510792)*dim
        ref[1,0,2,0]=(0.0553535735892)*(1+2.*(dim-1.)/(o+1.))+(-0.903166253779)*dim
        ref[1,0,2,1]=(-0.637702515795)*(1+2.*(dim-1.)/(o+1.))+(-0.596047338443)*dim
        ref[1,1,0,0]=(-1.56585324975)*(1+2.*(dim-1.)/(o+1.))+(-0.904167035781)*dim
        ref[1,1,0,1]=(-0.00114374423549)*(1+2.*(dim-1.)/(o+1.))+(0.0458351429579)*dim
        ref[1,1,1,0]=(0.00442415164442)*(1+2.*(dim-1.)/(o+1.))+(0.314560759001)*dim
        ref[1,1,1,1]=(0.143262734228)*(1+2.*(dim-1.)/(o+1.))+(0.492653395582)*dim
        ref[1,1,2,0]=(-0.231938733956)*(1+2.*(dim-1.)/(o+1.))+(-0.314578548729)*dim
        ref[1,1,2,1]=(-0.754435232577)*(1+2.*(dim-1.)/(o+1.))+(1.60771048901)*dim
        ref[1,2,0,0]=(0.0222021944152)*(1+2.*(dim-1.)/(o+1.))+(-0.444960965209)*dim
        ref[1,2,0,1]=(-0.0943322527204)*(1+2.*(dim-1.)/(o+1.))+(0.333429598586)*dim
        ref[1,2,1,0]=(0.644354843253)*(1+2.*(dim-1.)/(o+1.))+(0.401885477446)*dim
        ref[1,2,1,1]=(-1.3638687677)*(1+2.*(dim-1.)/(o+1.))+(-0.0433361952409)*dim
        ref[1,2,2,0]=(-0.423651599421)*(1+2.*(dim-1.)/(o+1.))+(-0.969807434883)*dim
        ref[1,2,2,1]=(0.635847571335)*(1+2.*(dim-1.)/(o+1.))+(0.820240470355)*dim
        ref[1,3,0,0]=(-0.867411506519)*(1+2.*(dim-1.)/(o+1.))+(1.43818317824)*dim
        ref[1,3,0,1]=(-0.21651569997)*(1+2.*(dim-1.)/(o+1.))+(0.117556665142)*dim
        ref[1,3,1,0]=(0.148576640926)*(1+2.*(dim-1.)/(o+1.))+(0.508830448407)*dim
        ref[1,3,1,1]=(-0.263610221224)*(1+2.*(dim-1.)/(o+1.))+(0.0213786923536)*dim
        ref[1,3,2,0]=(0.664786653328)*(1+2.*(dim-1.)/(o+1.))+(-0.131930830425)*dim
        ref[1,3,2,1]=(-0.449521314323)*(1+2.*(dim-1.)/(o+1.))+(0.527877501154)*dim
        ref[1,4,0,0]=(0.487353441653)*(1+2.*(dim-1.)/(o+1.))+(-0.0531636735847)*dim
        ref[1,4,0,1]=(0.715053861171)*(1+2.*(dim-1.)/(o+1.))+(1.30685601449)*dim
        ref[1,4,1,0]=(0.567794659886)*(1+2.*(dim-1.)/(o+1.))+(-0.519524437409)*dim
        ref[1,4,1,1]=(-0.575680185015)*(1+2.*(dim-1.)/(o+1.))+(-1.14395178835)*dim
        ref[1,4,2,0]=(-0.193703289912)*(1+2.*(dim-1.)/(o+1.))+(0.909238942041)*dim
        ref[1,4,2,1]=(1.1101139323)*(1+2.*(dim-1.)/(o+1.))+(0.0979474770197)*dim
        ref[2,0,0,0]=(0.380697711566)*(1+2.*(dim-1.)/(o+1.))+(-0.637674220835)*dim
        ref[2,0,0,1]=(1.02999730341)*(1+2.*(dim-1.)/(o+1.))+(-1.3472424912)*dim
        ref[2,0,1,0]=(-0.00637727068081)*(1+2.*(dim-1.)/(o+1.))+(-0.394899682106)*dim
        ref[2,0,1,1]=(0.366935357199)*(1+2.*(dim-1.)/(o+1.))+(1.46624218184)*dim
        ref[2,0,2,0]=(1.05023094229)*(1+2.*(dim-1.)/(o+1.))+(-1.70678656779)*dim
        ref[2,0,2,1]=(0.742210185409)*(1+2.*(dim-1.)/(o+1.))+(0.839954187494)*dim
        ref[2,1,0,0]=(0.684458780962)*(1+2.*(dim-1.)/(o+1.))+(-0.17835885676)*dim
        ref[2,1,0,1]=(0.645052864877)*(1+2.*(dim-1.)/(o+1.))+(0.986797388786)*dim
        ref[2,1,1,0]=(0.700347892015)*(1+2.*(dim-1.)/(o+1.))+(0.976128149165)*dim
        ref[2,1,1,1]=(0.389936266545)*(1+2.*(dim-1.)/(o+1.))+(-0.892821852874)*dim
        ref[2,1,2,0]=(-0.482535033346)*(1+2.*(dim-1.)/(o+1.))+(0.00751804809504)*dim
        ref[2,1,2,1]=(-0.29088392953)*(1+2.*(dim-1.)/(o+1.))+(-1.53871565655)*dim
        ref[2,2,0,0]=(1.20822889341)*(1+2.*(dim-1.)/(o+1.))+(-1.23095879753)*dim
        ref[2,2,0,1]=(-0.405976530285)*(1+2.*(dim-1.)/(o+1.))+(0.646186658355)*dim
        ref[2,2,1,0]=(0.521403464152)*(1+2.*(dim-1.)/(o+1.))+(0.0295817196322)*dim
        ref[2,2,1,1]=(0.491403825926)*(1+2.*(dim-1.)/(o+1.))+(0.496261081844)*dim
        ref[2,2,2,0]=(-0.217373555025)*(1+2.*(dim-1.)/(o+1.))+(-1.05842158862)*dim
        ref[2,2,2,1]=(0.958711693347)*(1+2.*(dim-1.)/(o+1.))+(0.473938563595)*dim
        ref[2,3,0,0]=(-0.291914669809)*(1+2.*(dim-1.)/(o+1.))+(1.10216225147)*dim
        ref[2,3,0,1]=(0.302480928584)*(1+2.*(dim-1.)/(o+1.))+(-0.270792946789)*dim
        ref[2,3,1,0]=(0.423231047429)*(1+2.*(dim-1.)/(o+1.))+(-0.00838465238812)*dim
        ref[2,3,1,1]=(0.232137290364)*(1+2.*(dim-1.)/(o+1.))+(-0.54117855925)*dim
        ref[2,3,2,0]=(-0.168723806518)*(1+2.*(dim-1.)/(o+1.))+(0.25257878195)*dim
        ref[2,3,2,1]=(-0.154743651439)*(1+2.*(dim-1.)/(o+1.))+(-1.33221787564)*dim
        ref[2,4,0,0]=(0.425370160961)*(1+2.*(dim-1.)/(o+1.))+(-0.491656751754)*dim
        ref[2,4,0,1]=(0.206452301631)*(1+2.*(dim-1.)/(o+1.))+(-0.636022345665)*dim
        ref[2,4,1,0]=(0.284343883626)*(1+2.*(dim-1.)/(o+1.))+(0.137216791106)*dim
        ref[2,4,1,1]=(1.07646333233)*(1+2.*(dim-1.)/(o+1.))+(-0.770679570435)*dim
        ref[2,4,2,0]=(-0.654711849202)*(1+2.*(dim-1.)/(o+1.))+(1.34705073935)*dim
        ref[2,4,2,1]=(1.19306162804)*(1+2.*(dim-1.)/(o+1.))+(1.3640695365)*dim
        ref[3,0,0,0]=(0.755283844775)*(1+2.*(dim-1.)/(o+1.))+(0.0256179953709)*dim
        ref[3,0,0,1]=(0.0968054541574)*(1+2.*(dim-1.)/(o+1.))+(1.05795664188)*dim
        ref[3,0,1,0]=(-1.61782777746)*(1+2.*(dim-1.)/(o+1.))+(-0.799617878088)*dim
        ref[3,0,1,1]=(1.65061646998)*(1+2.*(dim-1.)/(o+1.))+(0.739490926336)*dim
        ref[3,0,2,0]=(0.789360808191)*(1+2.*(dim-1.)/(o+1.))+(-0.369239128632)*dim
        ref[3,0,2,1]=(0.450939169308)*(1+2.*(dim-1.)/(o+1.))+(-0.0456587696413)*dim
        ref[3,1,0,0]=(-0.0576215844858)*(1+2.*(dim-1.)/(o+1.))+(-0.654389694689)*dim
        ref[3,1,0,1]=(-0.00993318570204)*(1+2.*(dim-1.)/(o+1.))+(-0.2385710942)*dim
        ref[3,1,1,0]=(-0.883302830864)*(1+2.*(dim-1.)/(o+1.))+(-0.852083422907)*dim
        ref[3,1,1,1]=(-0.501719754284)*(1+2.*(dim-1.)/(o+1.))+(-0.869095904016)*dim
        ref[3,1,2,0]=(-0.700492939735)*(1+2.*(dim-1.)/(o+1.))+(1.84002172539)*dim
        ref[3,1,2,1]=(1.14193807877)*(1+2.*(dim-1.)/(o+1.))+(1.40839099299)*dim
        ref[3,2,0,0]=(0.347101931746)*(1+2.*(dim-1.)/(o+1.))+(-0.73458360831)*dim
        ref[3,2,0,1]=(0.297487952383)*(1+2.*(dim-1.)/(o+1.))+(1.38962967209)*dim
        ref[3,2,1,0]=(0.431752993035)*(1+2.*(dim-1.)/(o+1.))+(-0.246624917801)*dim
        ref[3,2,1,1]=(1.06137140922)*(1+2.*(dim-1.)/(o+1.))+(0.416394705787)*dim
        ref[3,2,2,0]=(-1.78325339443)*(1+2.*(dim-1.)/(o+1.))+(-0.0564625863481)*dim
        ref[3,2,2,1]=(-0.303700310713)*(1+2.*(dim-1.)/(o+1.))+(-0.571784648561)*dim
        ref[3,3,0,0]=(-1.54702236684)*(1+2.*(dim-1.)/(o+1.))+(1.22160193344)*dim
        ref[3,3,0,1]=(0.614915137167)*(1+2.*(dim-1.)/(o+1.))+(-1.62298922747)*dim
        ref[3,3,1,0]=(-0.633511907257)*(1+2.*(dim-1.)/(o+1.))+(1.04953174704)*dim
        ref[3,3,1,1]=(-0.56466817679)*(1+2.*(dim-1.)/(o+1.))+(0.668916444776)*dim
        ref[3,3,2,0]=(-0.074237985605)*(1+2.*(dim-1.)/(o+1.))+(0.383573178544)*dim
        ref[3,3,2,1]=(0.424150131524)*(1+2.*(dim-1.)/(o+1.))+(0.321049696808)*dim
        ref[3,4,0,0]=(-1.01528091251)*(1+2.*(dim-1.)/(o+1.))+(-0.930932873777)*dim
        ref[3,4,0,1]=(-0.762611855917)*(1+2.*(dim-1.)/(o+1.))+(-0.731978921132)*dim
        ref[3,4,1,0]=(-0.0863922313234)*(1+2.*(dim-1.)/(o+1.))+(-0.0443942295012)*dim
        ref[3,4,1,1]=(-1.40717864239)*(1+2.*(dim-1.)/(o+1.))+(-0.42234764654)*dim
        ref[3,4,2,0]=(-1.09635285713)*(1+2.*(dim-1.)/(o+1.))+(-0.248589328845)*dim
        ref[3,4,2,1]=(1.05163924231)*(1+2.*(dim-1.)/(o+1.))+(0.808740340149)*dim
      else:
        s[0,0,0,0]=(0.315885178942)*x[0]**o+(-0.259227433663)*x[0]+(-0.985139490966)*x[1]**o+(-0.365399875853)*x[1]+(-0.378614709474)*x[2]**o+(-0.749166773032)*x[2]
        s[0,0,0,1]=(-0.980745316199)*x[0]**o+(0.917227239621)*x[0]+(-0.00961787312321)*x[1]**o+(-0.803272961034)*x[1]+(-0.569841488886)*x[2]**o+(-0.470914343673)*x[2]
        s[0,0,1,0]=(0.474857040946)*x[0]**o+(0.737872920694)*x[0]+(0.0428239150717)*x[1]**o+(0.805671300306)*x[1]+(-0.372628413406)*x[2]**o+(0.719721248997)*x[2]
        s[0,0,1,1]=(-0.711880589909)*x[0]**o+(0.321391465679)*x[0]+(0.519295140087)*x[1]**o+(0.038319529181)*x[1]+(0.38347458279)*x[2]**o+(0.331699888768)*x[2]
        s[0,0,2,0]=(-0.749496365536)*x[0]**o+(0.587604152085)*x[0]+(-0.713982951135)*x[1]**o+(-0.846872978696)*x[1]+(0.187878469529)*x[2]**o+(-0.605222544928)*x[2]
        s[0,0,2,1]=(-0.410414088285)*x[0]**o+(-0.767933775709)*x[0]+(-0.60156292058)*x[1]**o+(-0.489314475101)*x[1]+(0.888771153163)*x[2]**o+(-0.600341282488)*x[2]
        s[0,1,0,0]=(0.57725226271)*x[0]**o+(0.634144506828)*x[0]+(0.341370019875)*x[1]**o+(0.586961840417)*x[1]+(0.459018061293)*x[2]**o+(0.0746845088082)*x[2]
        s[0,1,0,1]=(0.0407919221935)*x[0]**o+(-0.360125630891)*x[0]+(-0.148792597333)*x[1]**o+(0.396869345515)*x[1]+(-0.745920425203)*x[2]**o+(0.0343202170743)*x[2]
        s[0,1,1,0]=(-0.924932256421)*x[0]**o+(0.516566988169)*x[0]+(-0.152377696008)*x[1]**o+(0.873885948456)*x[1]+(-0.360869839536)*x[2]**o+(-0.388854183331)*x[2]
        s[0,1,1,1]=(-0.323969456625)*x[0]**o+(0.304506990935)*x[0]+(-0.990292733104)*x[1]**o+(-0.245365945948)*x[1]+(0.181900978883)*x[2]**o+(-0.538788421192)*x[2]
        s[0,1,2,0]=(-0.374550735743)*x[0]**o+(0.564340899121)*x[0]+(0.801278948399)*x[1]**o+(-0.462054604952)*x[1]+(-0.785576728179)*x[2]**o+(0.0926519203136)*x[2]
        s[0,1,2,1]=(0.310136431429)*x[0]**o+(-0.00821101099015)*x[0]+(-0.974448609702)*x[1]**o+(0.348959120777)*x[1]+(-0.757748781922)*x[2]**o+(0.297755375582)*x[2]
        s[0,2,0,0]=(-0.812242021105)*x[0]**o+(0.898358963888)*x[0]+(0.77445189337)*x[1]**o+(0.786716425501)*x[1]+(0.544028378828)*x[2]**o+(0.549375386728)*x[2]
        s[0,2,0,1]=(-0.0378209686309)*x[0]**o+(-0.802035255239)*x[0]+(0.750004779554)*x[1]**o+(-0.671471110194)*x[1]+(0.993314295529)*x[2]**o+(-0.162020573179)*x[2]
        s[0,2,1,0]=(-0.492426591625)*x[0]**o+(0.902435821013)*x[0]+(0.557810230975)*x[1]**o+(-0.283019951662)*x[1]+(-0.987590772512)*x[2]**o+(-0.892849711933)*x[2]
        s[0,2,1,1]=(0.714593519237)*x[0]**o+(0.964018148555)*x[0]+(0.510176911254)*x[1]**o+(0.630953794541)*x[1]+(0.0297056372684)*x[2]**o+(0.554337702949)*x[2]
        s[0,2,2,0]=(-0.0754028072226)*x[0]**o+(0.0696670414466)*x[0]+(0.303440801227)*x[1]**o+(-0.876096813532)*x[1]+(-0.772838653879)*x[2]**o+(0.891062956492)*x[2]
        s[0,2,2,1]=(-0.517283509506)*x[0]**o+(-0.890513911735)*x[0]+(-0.850002026469)*x[1]**o+(0.0485442983018)*x[1]+(-0.690340415981)*x[2]**o+(0.356372135225)*x[2]
        s[0,3,0,0]=(0.00589321043146)*x[0]**o+(0.833040936257)*x[0]+(-0.779061736965)*x[1]**o+(0.596999628455)*x[1]+(0.185497386523)*x[2]**o+(-0.580282832451)*x[2]
        s[0,3,0,1]=(0.536520364057)*x[0]**o+(0.119573894206)*x[0]+(-0.481591297936)*x[1]**o+(-0.131066608017)*x[1]+(-0.714448088992)*x[2]**o+(-0.15245786606)*x[2]
        s[0,3,1,0]=(-0.649688821569)*x[0]**o+(0.872646765296)*x[0]+(0.996914263841)*x[1]**o+(-0.771999134903)*x[1]+(0.388158636968)*x[2]**o+(0.997694871854)*x[2]
        s[0,3,1,1]=(0.60805862552)*x[0]**o+(0.496342905951)*x[0]+(0.0678538981395)*x[1]**o+(-0.723458192199)*x[1]+(-0.227951366099)*x[2]**o+(-0.0454845476315)*x[2]
        s[0,3,2,0]=(0.927213775724)*x[0]**o+(-0.447976073985)*x[0]+(-0.176332048298)*x[1]**o+(0.177420385987)*x[1]+(0.752794515408)*x[2]**o+(-0.273183979432)*x[2]
        s[0,3,2,1]=(-0.648274135241)*x[0]**o+(0.16199869622)*x[0]+(-0.902866229405)*x[1]**o+(0.469114855264)*x[1]+(0.810779308988)*x[2]**o+(0.946561654853)*x[2]
        s[0,4,0,0]=(0.20080337911)*x[0]**o+(0.492510182922)*x[0]+(-0.39283470481)*x[1]**o+(-0.47885185637)*x[1]+(-0.925823446617)*x[2]**o+(0.222218483415)*x[2]
        s[0,4,0,1]=(-0.0972069635212)*x[0]**o+(0.183203800054)*x[0]+(-0.654862802864)*x[1]**o+(0.262618285977)*x[1]+(0.854563567237)*x[2]**o+(0.0964536047741)*x[2]
        s[0,4,1,0]=(0.45227435438)*x[0]**o+(0.538656962105)*x[0]+(0.365921783628)*x[1]**o+(0.215899953977)*x[1]+(0.592413111565)*x[2]**o+(-0.062048122597)*x[2]
        s[0,4,1,1]=(-0.818221902033)*x[0]**o+(-0.160355040635)*x[0]+(0.167475874143)*x[1]**o+(0.918522475482)*x[1]+(0.223565847733)*x[2]**o+(-0.552706943026)*x[2]
        s[0,4,2,0]=(0.917266921581)*x[0]**o+(-0.142334389128)*x[0]+(0.387287963005)*x[1]**o+(0.330170937971)*x[1]+(-0.822125826009)*x[2]**o+(-0.18853564744)*x[2]
        s[0,4,2,1]=(-0.720218394391)*x[0]**o+(0.844297728137)*x[0]+(0.428917747313)*x[1]**o+(0.0234117516707)*x[1]+(0.877188952533)*x[2]**o+(-0.580181524673)*x[2]
        s[1,0,0,0]=(-0.525360603901)*x[0]**o+(-0.444250370569)*x[0]+(-0.669034103414)*x[1]**o+(-0.463421879769)*x[1]+(-0.759457132782)*x[2]**o+(0.604035926001)*x[2]
        s[1,0,0,1]=(-0.0897354786393)*x[0]**o+(0.705360710589)*x[0]+(0.696483863344)*x[1]**o+(-0.241174862697)*x[1]+(-0.561951314355)*x[2]**o+(-0.912030054302)*x[2]
        s[1,0,1,0]=(0.52054104485)*x[0]**o+(0.0430575354511)*x[0]+(0.158290317145)*x[1]**o+(0.101019650181)*x[1]+(0.524338380356)*x[2]**o+(0.090701232108)*x[2]
        s[1,0,1,1]=(0.825208335954)*x[0]**o+(-0.473377827796)*x[0]+(0.68825627399)*x[1]**o+(0.494858806038)*x[1]+(0.224456734493)*x[2]**o+(0.406206820762)*x[2]
        s[1,0,2,0]=(-0.960366948345)*x[0]**o+(-0.234157118981)*x[0]+(-0.113359629154)*x[1]**o+(0.301361613247)*x[1]+(-0.594339813983)*x[2]**o+(-0.263211884586)*x[2]
        s[1,0,2,1]=(-0.94506913355)*x[0]**o+(0.389157389807)*x[0]+(-0.0470471678881)*x[1]**o+(-0.545905083414)*x[1]+(0.189132629724)*x[2]**o+(0.873560571105)*x[2]
        s[1,1,0,0]=(0.196235297476)*x[0]**o+(0.122584939735)*x[0]+(0.508485269882)*x[1]**o+(0.525003719219)*x[1]+(0.961756294197)*x[2]**o+(-0.766404136898)*x[2]
        s[1,1,0,1]=(0.158187473618)*x[0]**o+(0.570375022293)*x[0]+(0.815862093528)*x[1]**o+(-0.967297717974)*x[1]+(0.290548253702)*x[2]**o+(0.603934169794)*x[2]
        s[1,1,1,0]=(0.37168083831)*x[0]**o+(0.378783317761)*x[0]+(-0.62105516013)*x[1]**o+(-0.243670994716)*x[1]+(0.0493547361834)*x[2]**o+(-0.463410303505)*x[2]
        s[1,1,1,1]=(0.935366553895)*x[0]**o+(-0.62914115892)*x[0]+(-0.251619616412)*x[1]**o+(-0.00509465672734)*x[1]+(0.801300357872)*x[2]**o+(-0.225391368884)*x[2]
        s[1,1,2,0]=(-0.182365850951)*x[0]**o+(-0.510497578546)*x[0]+(-0.976490140605)*x[1]**o+(-0.810521612611)*x[1]+(0.568861407451)*x[2]**o+(0.984242045677)*x[2]
        s[1,1,2,1]=(0.438879837985)*x[0]**o+(0.123241828005)*x[0]+(-0.0920198739704)*x[1]**o+(0.0163670845773)*x[1]+(-0.170074389128)*x[2]**o+(0.9808945977)*x[2]
        s[1,2,0,0]=(0.844376099112)*x[0]**o+(0.281113346834)*x[0]+(0.678586875156)*x[1]**o+(-0.339719035096)*x[1]+(0.149112798051)*x[2]**o+(0.345737934766)*x[2]
        s[1,2,0,1]=(-0.823808399101)*x[0]**o+(0.741400052542)*x[0]+(0.631490960369)*x[1]**o+(0.884534048234)*x[1]+(0.986087890092)*x[2]**o+(0.0633984741173)*x[2]
        s[1,2,1,0]=(0.577630620249)*x[0]**o+(-0.875400887572)*x[0]+(0.282570000665)*x[1]**o+(-0.720651390173)*x[1]+(0.740020216206)*x[2]**o+(0.116758611432)*x[2]
        s[1,2,1,1]=(0.827573407856)*x[0]**o+(0.788455460143)*x[0]+(-0.935960102047)*x[1]**o+(0.128092793671)*x[1]+(-0.637271207159)*x[2]**o+(-0.0949607133369)*x[2]
        s[1,2,2,0]=(0.653489510769)*x[0]**o+(0.127242563347)*x[0]+(0.441720108695)*x[1]**o+(0.176710497085)*x[1]+(-0.466525521911)*x[2]**o+(-0.106726607773)*x[2]
        s[1,2,2,1]=(0.544895430642)*x[0]**o+(0.617184925179)*x[0]+(-0.590535213263)*x[1]**o+(-0.240993032963)*x[1]+(0.847624557098)*x[2]**o+(0.825419570696)*x[2]
        s[1,3,0,0]=(0.632663921163)*x[0]**o+(0.190826102102)*x[0]+(0.849534763016)*x[1]**o+(-0.0237561687913)*x[1]+(-0.269077077602)*x[2]**o+(-0.549527552663)*x[2]
        s[1,3,0,1]=(-0.573827583713)*x[0]**o+(0.698099772692)*x[0]+(-0.0111629174288)*x[1]**o+(0.779552959539)*x[1]+(0.268204963497)*x[2]**o+(0.598849955092)*x[2]
        s[1,3,1,0]=(-0.83963180468)*x[0]**o+(-0.688432739537)*x[0]+(-0.354553847784)*x[1]**o+(0.818998341858)*x[1]+(-0.214142424648)*x[2]**o+(-0.0349087206605)*x[2]
        s[1,3,1,1]=(0.961416480403)*x[0]**o+(0.214086874997)*x[0]+(-0.00837344828613)*x[1]**o+(0.907896124099)*x[1]+(-0.101452170826)*x[2]**o+(0.532739507876)*x[2]
        s[1,3,2,0]=(0.695510036287)*x[0]**o+(0.365801535675)*x[0]+(0.72964983115)*x[1]**o+(0.598376640377)*x[1]+(-0.786372424004)*x[2]**o+(-0.307873549549)*x[2]
        s[1,3,2,1]=(-0.00267955266574)*x[0]**o+(-0.36091144037)*x[0]+(-0.679691091713)*x[1]**o+(-0.728893213661)*x[1]+(-0.925050172841)*x[2]**o+(-0.218818592747)*x[2]
        s[1,4,0,0]=(-0.10754938599)*x[0]**o+(-0.412916506659)*x[0]+(0.688731006932)*x[1]**o+(0.878299598346)*x[1]+(0.726030761617)*x[2]**o+(0.637038101344)*x[2]
        s[1,4,0,1]=(0.262443563953)*x[0]**o+(-0.768431899484)*x[0]+(0.807995951545)*x[1]**o+(0.729996462998)*x[1]+(-0.128003445373)*x[2]**o+(0.516856350881)*x[2]
        s[1,4,1,0]=(-0.275354824026)*x[0]**o+(0.0620548936694)*x[0]+(-0.708292160104)*x[1]**o+(-0.138724590156)*x[1]+(-0.937760817006)*x[2]**o+(0.295878139279)*x[2]
        s[1,4,1,1]=(-0.194978075848)*x[0]**o+(0.116302091644)*x[0]+(0.668500205751)*x[1]**o+(-0.276843526204)*x[1]+(0.554624275832)*x[2]**o+(-0.0252161509726)*x[2]
        s[1,4,2,0]=(-0.980507466062)*x[0]**o+(0.345696047502)*x[0]+(0.990984322861)*x[1]**o+(-0.29237796986)*x[1]+(-0.113106912241)*x[2]**o+(-0.171569411982)*x[2]
        s[1,4,2,1]=(-0.996430023986)*x[0]**o+(0.0129252732357)*x[0]+(0.432605766267)*x[1]**o+(0.245516428361)*x[1]+(-0.287283002725)*x[2]**o+(0.496535294994)*x[2]
        s[2,0,0,0]=(-0.521251352762)*x[0]**o+(0.814082050334)*x[0]+(-0.342514979075)*x[1]**o+(0.180394009823)*x[1]+(-0.928876765188)*x[2]**o+(-0.177317444177)*x[2]
        s[2,0,0,1]=(-0.752594467046)*x[0]**o+(0.00697110526032)*x[0]+(0.107735313159)*x[1]**o+(0.507279243411)*x[1]+(0.246185410903)*x[2]**o+(-0.705550571172)*x[2]
        s[2,0,1,0]=(0.987074535948)*x[0]**o+(0.120612341024)*x[0]+(-0.00891981771691)*x[1]**o+(0.412034737534)*x[1]+(-0.641732350314)*x[2]**o+(0.702296660315)*x[2]
        s[2,0,1,1]=(0.545360348518)*x[0]**o+(0.246057841211)*x[0]+(0.68096977198)*x[1]**o+(0.16979635608)*x[1]+(-0.786467966954)*x[2]**o+(-0.604979391255)*x[2]
        s[2,0,2,0]=(0.390620126799)*x[0]**o+(-0.297208846965)*x[0]+(-0.1676231355)*x[1]**o+(-0.999842891217)*x[1]+(-0.525384600734)*x[2]**o+(-0.616431037869)*x[2]
        s[2,0,2,1]=(-0.114488632507)*x[0]**o+(-0.837696558005)*x[0]+(0.0856093277648)*x[1]**o+(-0.813930500792)*x[1]+(0.0242825112259)*x[2]**o+(-0.553431782224)*x[2]
        s[2,1,0,0]=(-0.158829154482)*x[0]**o+(0.944427245379)*x[0]+(0.288168271085)*x[1]**o+(0.294610923624)*x[1]+(-0.93996240515)*x[2]**o+(-0.141444934145)*x[2]
        s[2,1,0,1]=(0.600407255802)*x[0]**o+(0.466366546805)*x[0]+(0.293813875924)*x[1]**o+(0.824442365447)*x[1]+(-0.95914204615)*x[2]**o+(0.268073473809)*x[2]
        s[2,1,1,0]=(-0.00912285930883)*x[0]**o+(0.541241315284)*x[0]+(-0.941279012281)*x[1]**o+(0.192591379725)*x[1]+(0.84906414331)*x[2]**o+(0.739941920728)*x[2]
        s[2,1,1,1]=(-0.515663862321)*x[0]**o+(-0.995173844425)*x[0]+(-0.0545298263549)*x[1]**o+(0.0547616321893)*x[1]+(0.906580176018)*x[2]**o+(-0.724193715365)*x[2]
        s[2,1,2,0]=(-0.961884774945)*x[0]**o+(0.228071784467)*x[0]+(0.109594642294)*x[1]**o+(0.390512248069)*x[1]+(0.584516492005)*x[2]**o+(0.732112742614)*x[2]
        s[2,1,2,1]=(-0.447795495117)*x[0]**o+(-0.436915184002)*x[0]+(0.0097966293808)*x[1]**o+(0.00984049811243)*x[1]+(-0.534980632513)*x[2]**o+(0.847070567413)*x[2]
        s[2,2,0,0]=(-0.72527144583)*x[0]**o+(-0.328380984831)*x[0]+(0.891931683176)*x[1]**o+(0.731425008677)*x[1]+(0.451269066706)*x[2]**o+(-0.876119382129)*x[2]
        s[2,2,0,1]=(-0.645145440462)*x[0]**o+(0.549953455814)*x[0]+(-0.714731040969)*x[1]**o+(-0.367442134998)*x[1]+(-0.784974893863)*x[2]**o+(0.993294593495)*x[2]
        s[2,2,1,0]=(0.795416456648)*x[0]**o+(-0.249185994779)*x[0]+(0.0887737619644)*x[1]**o+(-0.8527851189)*x[1]+(-0.977545483463)*x[2]**o+(-0.458811422497)*x[2]
        s[2,2,1,1]=(0.448509047555)*x[0]**o+(0.251160841663)*x[0]+(-0.958180262344)*x[1]**o+(0.00687385479851)*x[1]+(0.415237807094)*x[2]**o+(0.454049069214)*x[2]
        s[2,2,2,0]=(-0.875182150725)*x[0]**o+(-0.164810616491)*x[0]+(-0.854984081654)*x[1]**o+(-0.487839136852)*x[1]+(-0.991611140703)*x[2]**o+(0.131746964462)*x[2]
        s[2,2,2,1]=(0.190113983979)*x[0]**o+(-0.0580500408307)*x[0]+(-0.987736788697)*x[1]**o+(0.29105974832)*x[1]+(-0.841830938414)*x[2]**o+(-0.399303249251)*x[2]
        s[2,3,0,0]=(0.375874939141)*x[0]**o+(0.0179896833971)*x[0]+(0.935171068662)*x[1]**o+(-0.605279673785)*x[1]+(-0.315907419394)*x[2]**o+(-0.653351592585)*x[2]
        s[2,3,0,1]=(-0.307367451597)*x[0]**o+(0.433464337349)*x[0]+(0.120694815275)*x[1]**o+(0.882573418509)*x[1]+(0.720880126974)*x[2]**o+(0.413515788634)*x[2]
        s[2,3,1,0]=(-0.360768726229)*x[0]**o+(-0.345798386124)*x[0]+(-0.158591700207)*x[1]**o+(0.777571415125)*x[1]+(-0.515278006685)*x[2]**o+(-0.60141050103)*x[2]
        s[2,3,1,1]=(0.19968785285)*x[0]**o+(-0.678957085866)*x[0]+(-0.504487760025)*x[1]**o+(-0.938195450799)*x[1]+(-0.869784726008)*x[2]**o+(0.611936589332)*x[2]
        s[2,3,2,0]=(-0.94908432374)*x[0]**o+(-0.764309365026)*x[0]+(0.299970983215)*x[1]**o+(0.76972605976)*x[1]+(0.808781373563)*x[2]**o+(-0.744690590895)*x[2]
        s[2,3,2,1]=(-0.827801524944)*x[0]**o+(-0.639131625181)*x[0]+(0.115482861386)*x[1]**o+(0.749852345379)*x[1]+(-0.876800704567)*x[2]**o+(-0.335471532349)*x[2]
        s[2,4,0,0]=(0.957721855837)*x[0]**o+(-0.37091118004)*x[0]+(-0.457918571765)*x[1]**o+(0.574281982261)*x[1]+(-0.0381561925569)*x[2]**o+(0.583211845309)*x[2]
        s[2,4,0,1]=(0.922272060507)*x[0]**o+(0.934043162659)*x[0]+(0.665312385413)*x[1]**o+(0.82839206727)*x[1]+(0.192535528872)*x[2]**o+(-0.170352598091)*x[2]
        s[2,4,1,0]=(0.696400806982)*x[0]**o+(-0.864993988074)*x[0]+(0.175067366294)*x[1]**o+(-0.117341125055)*x[1]+(-0.0862802982578)*x[2]**o+(0.360686011537)*x[2]
        s[2,4,1,1]=(0.285042052581)*x[0]**o+(0.819442194445)*x[0]+(0.985647806774)*x[1]**o+(-0.706227171911)*x[1]+(-0.854322012297)*x[2]**o+(-0.854051496942)*x[2]
        s[2,4,2,0]=(0.493375212662)*x[0]**o+(0.0135080286945)*x[0]+(-0.620056636744)*x[1]**o+(-0.160959399742)*x[1]+(-0.524788079589)*x[2]**o+(-0.421933761234)*x[2]
        s[2,4,2,1]=(0.657882929009)*x[0]**o+(0.164321066198)*x[0]+(0.0849684071734)*x[1]**o+(0.128314490298)*x[1]+(0.0313579868936)*x[2]**o+(0.608750041614)*x[2]
        s[3,0,0,0]=(0.401854741306)*x[0]**o+(0.134373785831)*x[0]+(-0.518732727663)*x[1]**o+(0.962631263931)*x[1]+(0.756955386629)*x[2]**o+(0.10153666864)*x[2]
        s[3,0,0,1]=(0.407206244876)*x[0]**o+(-0.432733726651)*x[0]+(-0.940640930336)*x[1]**o+(0.745788564728)*x[1]+(-0.342370630964)*x[2]**o+(-0.535375080975)*x[2]
        s[3,0,1,0]=(0.360769684176)*x[0]**o+(-0.931274190319)*x[0]+(0.86097262982)*x[1]**o+(0.767924495942)*x[1]+(0.751955166412)*x[2]**o+(-0.243315246145)*x[2]
        s[3,0,1,1]=(0.729403674327)*x[0]**o+(0.400842302945)*x[0]+(0.275966885949)*x[1]**o+(-0.546578452065)*x[1]+(-0.844280661655)*x[2]**o+(0.0715593226033)*x[2]
        s[3,0,2,0]=(-0.26926147241)*x[0]**o+(-0.313445919689)*x[0]+(0.0745755455487)*x[1]**o+(0.561797432558)*x[1]+(0.746592037865)*x[2]**o+(0.691841700089)*x[2]
        s[3,0,2,1]=(0.260642362402)*x[0]**o+(0.585261059676)*x[0]+(-0.226552249351)*x[1]**o+(-0.22018440207)*x[1]+(-0.913385081787)*x[2]**o+(0.641093619341)*x[2]
        s[3,1,0,0]=(0.371513393673)*x[0]**o+(0.0170798820983)*x[0]+(0.564278651267)*x[1]**o+(0.680930162447)*x[1]+(-0.706423458637)*x[2]**o+(0.0764206259648)*x[2]
        s[3,1,0,1]=(-0.948003266655)*x[0]**o+(0.491205780333)*x[0]+(-0.758784222468)*x[1]**o+(0.963242289935)*x[1]+(0.598059292168)*x[2]**o+(0.864032824015)*x[2]
        s[3,1,1,0]=(-0.238580509808)*x[0]**o+(-0.604062371822)*x[0]+(0.575641671747)*x[1]**o+(0.36375183411)*x[1]+(-0.972736699609)*x[2]**o+(0.880286281096)*x[2]
        s[3,1,1,1]=(0.0699909126019)*x[0]**o+(0.134320946509)*x[0]+(0.610933092241)*x[1]**o+(0.371352827221)*x[1]+(0.0323703390883)*x[2]**o+(-0.274126481723)*x[2]
        s[3,1,2,0]=(0.453740440957)*x[0]**o+(-0.943147597599)*x[0]+(0.881661129342)*x[1]**o+(-0.877404630304)*x[1]+(-0.565813043124)*x[2]**o+(0.730370433019)*x[2]
        s[3,1,2,1]=(0.55962110239)*x[0]**o+(-0.304753265168)*x[0]+(-0.446631806503)*x[1]**o+(-0.247296550227)*x[1]+(0.704691317199)*x[2]**o+(0.885258048422)*x[2]
        s[3,2,0,0]=(0.165892354646)*x[0]**o+(-0.910149277809)*x[0]+(0.512472386138)*x[1]**o+(0.783845430116)*x[1]+(0.748568950882)*x[2]**o+(0.279356093192)*x[2]
        s[3,2,0,1]=(-0.152647132155)*x[0]**o+(0.59980439016)*x[0]+(0.75599611367)*x[1]**o+(-0.492517161389)*x[1]+(-0.510828418476)*x[2]**o+(0.813006872834)*x[2]
        s[3,2,1,0]=(-0.732207745061)*x[0]**o+(-0.304109063819)*x[0]+(0.330171325627)*x[1]**o+(-0.934563542072)*x[1]+(0.529201115138)*x[2]**o+(0.837219324241)*x[2]
        s[3,2,1,1]=(0.697323448629)*x[0]**o+(0.34999597431)*x[0]+(0.67453985446)*x[1]**o+(-0.607043070971)*x[1]+(0.774743224308)*x[2]**o+(0.560978199188)*x[2]
        s[3,2,2,0]=(-0.826819922732)*x[0]**o+(-0.0808978789297)*x[0]+(0.391865159047)*x[1]**o+(-0.34035597887)*x[1]+(0.129988822983)*x[2]**o+(0.0782066826859)*x[2]
        s[3,2,2,1]=(-0.503921164807)*x[0]**o+(0.229433648691)*x[0]+(-0.224244626319)*x[1]**o+(0.307825062312)*x[1]+(0.778558288408)*x[2]**o+(0.236537878219)*x[2]
        s[3,3,0,0]=(-0.465572503617)*x[0]**o+(-0.505092598655)*x[0]+(-0.285232464978)*x[1]**o+(0.295822892399)*x[1]+(-0.678626607589)*x[2]**o+(-0.762910104233)*x[2]
        s[3,3,0,1]=(-0.313839797733)*x[0]**o+(-0.866968909594)*x[0]+(0.193089207451)*x[1]**o+(0.350572793209)*x[1]+(0.0985650768561)*x[2]**o+(0.61459734708)*x[2]
        s[3,3,1,0]=(-0.686947068473)*x[0]**o+(0.803092292826)*x[0]+(0.229346725739)*x[1]**o+(-0.523883629849)*x[1]+(-0.825851523578)*x[2]**o+(0.911991149266)*x[2]
        s[3,3,1,1]=(-0.0229534891291)*x[0]**o+(0.459587997571)*x[0]+(-0.0323375614527)*x[1]**o+(-0.960355649174)*x[1]+(-0.157463174284)*x[2]**o+(0.721625063029)*x[2]
        s[3,3,2,0]=(-0.244675134158)*x[0]**o+(-0.750815470458)*x[0]+(0.350391343156)*x[1]**o+(0.246114995378)*x[1]+(-0.299954768936)*x[2]**o+(0.621800581218)*x[2]
        s[3,3,2,1]=(0.517173606541)*x[0]**o+(0.175764075574)*x[0]+(0.423442626185)*x[1]**o+(0.803990330201)*x[1]+(0.286217667564)*x[2]**o+(-0.471348114366)*x[2]
        s[3,4,0,0]=(-0.351936186325)*x[0]**o+(-0.0279135493684)*x[0]+(0.143368480037)*x[1]**o+(-0.0914938899855)*x[1]+(-0.160075587837)*x[2]**o+(-0.504374107434)*x[2]
        s[3,4,0,1]=(0.585914411245)*x[0]**o+(0.274811379635)*x[0]+(-0.177534785865)*x[1]**o+(0.567196593954)*x[1]+(-0.510174432853)*x[2]**o+(0.930747960349)*x[2]
        s[3,4,1,0]=(0.992675673771)*x[0]**o+(-0.709559401209)*x[0]+(0.996108059004)*x[1]**o+(-0.264682085878)*x[1]+(0.336432533695)*x[2]**o+(0.439349601512)*x[2]
        s[3,4,1,1]=(-0.383904129095)*x[0]**o+(-0.640989586457)*x[0]+(0.876802160463)*x[1]**o+(-0.941096425806)*x[1]+(0.636317329751)*x[2]**o+(-0.922925940931)*x[2]
        s[3,4,2,0]=(0.215934904956)*x[0]**o+(-0.0763907072691)*x[0]+(0.864686802474)*x[1]**o+(-0.100958345332)*x[1]+(-0.545187503302)*x[2]**o+(0.896671191263)*x[2]
        s[3,4,2,1]=(-0.53798548609)*x[0]**o+(0.189564902905)*x[0]+(-0.352161561502)*x[1]**o+(0.726662085086)*x[1]+(0.514655541645)*x[2]**o+(-0.36993932162)*x[2]
        ref[0,0,0,0]=(-1.0478690215)*(1+2.*(dim-1.)/(o+1.))+(-1.37379408255)*dim
        ref[0,0,0,1]=(-1.56020467821)*(1+2.*(dim-1.)/(o+1.))+(-0.356960065085)*dim
        ref[0,0,1,0]=(0.145052542612)*(1+2.*(dim-1.)/(o+1.))+(2.26326547)*dim
        ref[0,0,1,1]=(0.190889132968)*(1+2.*(dim-1.)/(o+1.))+(0.691410883628)*dim
        ref[0,0,2,0]=(-1.27560084714)*(1+2.*(dim-1.)/(o+1.))+(-0.864491371539)*dim
        ref[0,0,2,1]=(-0.123205855702)*(1+2.*(dim-1.)/(o+1.))+(-1.8575895333)*dim
        ref[0,1,0,0]=(1.37764034388)*(1+2.*(dim-1.)/(o+1.))+(1.29579085605)*dim
        ref[0,1,0,1]=(-0.853921100343)*(1+2.*(dim-1.)/(o+1.))+(0.0710639316977)*dim
        ref[0,1,1,0]=(-1.43817979196)*(1+2.*(dim-1.)/(o+1.))+(1.00159875329)*dim
        ref[0,1,1,1]=(-1.13236121085)*(1+2.*(dim-1.)/(o+1.))+(-0.479647376205)*dim
        ref[0,1,2,0]=(-0.358848515523)*(1+2.*(dim-1.)/(o+1.))+(0.194938214483)*dim
        ref[0,1,2,1]=(-1.42206096019)*(1+2.*(dim-1.)/(o+1.))+(0.638503485369)*dim
        ref[0,2,0,0]=(0.506238251094)*(1+2.*(dim-1.)/(o+1.))+(2.23445077612)*dim
        ref[0,2,0,1]=(1.70549810645)*(1+2.*(dim-1.)/(o+1.))+(-1.63552693861)*dim
        ref[0,2,1,0]=(-0.922207133162)*(1+2.*(dim-1.)/(o+1.))+(-0.273433842582)*dim
        ref[0,2,1,1]=(1.25447606776)*(1+2.*(dim-1.)/(o+1.))+(2.14930964604)*dim
        ref[0,2,2,0]=(-0.544800659874)*(1+2.*(dim-1.)/(o+1.))+(0.0846331844061)*dim
        ref[0,2,2,1]=(-2.05762595196)*(1+2.*(dim-1.)/(o+1.))+(-0.485597478209)*dim
        ref[0,3,0,0]=(-0.587671140011)*(1+2.*(dim-1.)/(o+1.))+(0.84975773226)*dim
        ref[0,3,0,1]=(-0.659519022871)*(1+2.*(dim-1.)/(o+1.))+(-0.163950579871)*dim
        ref[0,3,1,0]=(0.73538407924)*(1+2.*(dim-1.)/(o+1.))+(1.09834250225)*dim
        ref[0,3,1,1]=(0.44796115756)*(1+2.*(dim-1.)/(o+1.))+(-0.27259983388)*dim
        ref[0,3,2,0]=(1.50367624283)*(1+2.*(dim-1.)/(o+1.))+(-0.54373966743)*dim
        ref[0,3,2,1]=(-0.740361055659)*(1+2.*(dim-1.)/(o+1.))+(1.57767520634)*dim
        ref[0,4,0,0]=(-1.11785477232)*(1+2.*(dim-1.)/(o+1.))+(0.235876809967)*dim
        ref[0,4,0,1]=(0.102493800852)*(1+2.*(dim-1.)/(o+1.))+(0.542275690805)*dim
        ref[0,4,1,0]=(1.41060924957)*(1+2.*(dim-1.)/(o+1.))+(0.692508793485)*dim
        ref[0,4,1,1]=(-0.427180180157)*(1+2.*(dim-1.)/(o+1.))+(0.205460491821)*dim
        ref[0,4,2,0]=(0.482429058578)*(1+2.*(dim-1.)/(o+1.))+(-0.000699098596959)*dim
        ref[0,4,2,1]=(0.585888305456)*(1+2.*(dim-1.)/(o+1.))+(0.287527955135)*dim
        ref[1,0,0,0]=(-1.9538518401)*(1+2.*(dim-1.)/(o+1.))+(-0.303636324337)*dim
        ref[1,0,0,1]=(0.0447970703505)*(1+2.*(dim-1.)/(o+1.))+(-0.447844206411)*dim
        ref[1,0,1,0]=(1.20316974235)*(1+2.*(dim-1.)/(o+1.))+(0.23477841774)*dim
        ref[1,0,1,1]=(1.73792134444)*(1+2.*(dim-1.)/(o+1.))+(0.427687799004)*dim
        ref[1,0,2,0]=(-1.66806639148)*(1+2.*(dim-1.)/(o+1.))+(-0.19600739032)*dim
        ref[1,0,2,1]=(-0.802983671714)*(1+2.*(dim-1.)/(o+1.))+(0.716812877498)*dim
        ref[1,1,0,0]=(1.66647686155)*(1+2.*(dim-1.)/(o+1.))+(-0.118815477944)*dim
        ref[1,1,0,1]=(1.26459782085)*(1+2.*(dim-1.)/(o+1.))+(0.207011474112)*dim
        ref[1,1,1,0]=(-0.200019585637)*(1+2.*(dim-1.)/(o+1.))+(-0.32829798046)*dim
        ref[1,1,1,1]=(1.48504729536)*(1+2.*(dim-1.)/(o+1.))+(-0.859627184532)*dim
        ref[1,1,2,0]=(-0.589994584106)*(1+2.*(dim-1.)/(o+1.))+(-0.336777145479)*dim
        ref[1,1,2,1]=(0.176785574887)*(1+2.*(dim-1.)/(o+1.))+(1.12050351028)*dim
        ref[1,2,0,0]=(1.67207577232)*(1+2.*(dim-1.)/(o+1.))+(0.287132246505)*dim
        ref[1,2,0,1]=(0.79377045136)*(1+2.*(dim-1.)/(o+1.))+(1.68933257489)*dim
        ref[1,2,1,0]=(1.60022083712)*(1+2.*(dim-1.)/(o+1.))+(-1.47929366631)*dim
        ref[1,2,1,1]=(-0.74565790135)*(1+2.*(dim-1.)/(o+1.))+(0.821587540478)*dim
        ref[1,2,2,0]=(0.628684097552)*(1+2.*(dim-1.)/(o+1.))+(0.19722645266)*dim
        ref[1,2,2,1]=(0.801984774477)*(1+2.*(dim-1.)/(o+1.))+(1.20161146291)*dim
        ref[1,3,0,0]=(1.21312160658)*(1+2.*(dim-1.)/(o+1.))+(-0.382457619353)*dim
        ref[1,3,0,1]=(-0.316785537645)*(1+2.*(dim-1.)/(o+1.))+(2.07650268732)*dim
        ref[1,3,1,0]=(-1.40832807711)*(1+2.*(dim-1.)/(o+1.))+(0.09565688166)*dim
        ref[1,3,1,1]=(0.851590861291)*(1+2.*(dim-1.)/(o+1.))+(1.65472250697)*dim
        ref[1,3,2,0]=(0.638787443433)*(1+2.*(dim-1.)/(o+1.))+(0.656304626504)*dim
        ref[1,3,2,1]=(-1.60742081722)*(1+2.*(dim-1.)/(o+1.))+(-1.30862324678)*dim
        ref[1,4,0,0]=(1.30721238256)*(1+2.*(dim-1.)/(o+1.))+(1.10242119303)*dim
        ref[1,4,0,1]=(0.942436070125)*(1+2.*(dim-1.)/(o+1.))+(0.478420914396)*dim
        ref[1,4,1,0]=(-1.92140780114)*(1+2.*(dim-1.)/(o+1.))+(0.219208442793)*dim
        ref[1,4,1,1]=(1.02814640573)*(1+2.*(dim-1.)/(o+1.))+(-0.185757585533)*dim
        ref[1,4,2,0]=(-0.102630055442)*(1+2.*(dim-1.)/(o+1.))+(-0.118251334341)*dim
        ref[1,4,2,1]=(-0.851107260444)*(1+2.*(dim-1.)/(o+1.))+(0.754976996591)*dim
        ref[2,0,0,0]=(-1.79264309703)*(1+2.*(dim-1.)/(o+1.))+(0.81715861598)*dim
        ref[2,0,0,1]=(-0.398673742984)*(1+2.*(dim-1.)/(o+1.))+(-0.191300222501)*dim
        ref[2,0,1,0]=(0.336422367917)*(1+2.*(dim-1.)/(o+1.))+(1.23494373887)*dim
        ref[2,0,1,1]=(0.439862153544)*(1+2.*(dim-1.)/(o+1.))+(-0.189125193964)*dim
        ref[2,0,2,0]=(-0.302387609436)*(1+2.*(dim-1.)/(o+1.))+(-1.91348277605)*dim
        ref[2,0,2,1]=(-0.00459679351666)*(1+2.*(dim-1.)/(o+1.))+(-2.20505884102)*dim
        ref[2,1,0,0]=(-0.810623288547)*(1+2.*(dim-1.)/(o+1.))+(1.09759323486)*dim
        ref[2,1,0,1]=(-0.0649209144233)*(1+2.*(dim-1.)/(o+1.))+(1.55888238606)*dim
        ref[2,1,1,0]=(-0.10133772828)*(1+2.*(dim-1.)/(o+1.))+(1.47377461574)*dim
        ref[2,1,1,1]=(0.336386487342)*(1+2.*(dim-1.)/(o+1.))+(-1.6646059276)*dim
        ref[2,1,2,0]=(-0.267773640646)*(1+2.*(dim-1.)/(o+1.))+(1.35069677515)*dim
        ref[2,1,2,1]=(-0.972979498249)*(1+2.*(dim-1.)/(o+1.))+(0.419995881524)*dim
        ref[2,2,0,0]=(0.617929304052)*(1+2.*(dim-1.)/(o+1.))+(-0.473075358283)*dim
        ref[2,2,0,1]=(-2.14485137529)*(1+2.*(dim-1.)/(o+1.))+(1.17580591431)*dim
        ref[2,2,1,0]=(-0.0933552648497)*(1+2.*(dim-1.)/(o+1.))+(-1.56078253618)*dim
        ref[2,2,1,1]=(-0.094433407695)*(1+2.*(dim-1.)/(o+1.))+(0.712083765676)*dim
        ref[2,2,2,0]=(-2.72177737308)*(1+2.*(dim-1.)/(o+1.))+(-0.520902788881)*dim
        ref[2,2,2,1]=(-1.63945374313)*(1+2.*(dim-1.)/(o+1.))+(-0.166293541762)*dim
        ref[2,3,0,0]=(0.995138588408)*(1+2.*(dim-1.)/(o+1.))+(-1.24064158297)*dim
        ref[2,3,0,1]=(0.534207490651)*(1+2.*(dim-1.)/(o+1.))+(1.72955354449)*dim
        ref[2,3,1,0]=(-1.03463843312)*(1+2.*(dim-1.)/(o+1.))+(-0.169637472029)*dim
        ref[2,3,1,1]=(-1.17458463318)*(1+2.*(dim-1.)/(o+1.))+(-1.00521594733)*dim
        ref[2,3,2,0]=(0.159668033037)*(1+2.*(dim-1.)/(o+1.))+(-0.739273896161)*dim
        ref[2,3,2,1]=(-1.58911936813)*(1+2.*(dim-1.)/(o+1.))+(-0.22475081215)*dim
        ref[2,4,0,0]=(0.461647091515)*(1+2.*(dim-1.)/(o+1.))+(0.786582647531)*dim
        ref[2,4,0,1]=(1.78011997479)*(1+2.*(dim-1.)/(o+1.))+(1.59208263184)*dim
        ref[2,4,1,0]=(0.785187875018)*(1+2.*(dim-1.)/(o+1.))+(-0.621649101592)*dim
        ref[2,4,1,1]=(0.416367847058)*(1+2.*(dim-1.)/(o+1.))+(-0.740836474409)*dim
        ref[2,4,2,0]=(-0.651469503672)*(1+2.*(dim-1.)/(o+1.))+(-0.569385132282)*dim
        ref[2,4,2,1]=(0.774209323076)*(1+2.*(dim-1.)/(o+1.))+(0.90138559811)*dim
        ref[3,0,0,0]=(0.640077400272)*(1+2.*(dim-1.)/(o+1.))+(1.1985417184)*dim
        ref[3,0,0,1]=(-0.875805316424)*(1+2.*(dim-1.)/(o+1.))+(-0.222320242899)*dim
        ref[3,0,1,0]=(1.97369748041)*(1+2.*(dim-1.)/(o+1.))+(-0.406664940521)*dim
        ref[3,0,1,1]=(0.161089898621)*(1+2.*(dim-1.)/(o+1.))+(-0.0741768265174)*dim
        ref[3,0,2,0]=(0.551906111004)*(1+2.*(dim-1.)/(o+1.))+(0.940193212958)*dim
        ref[3,0,2,1]=(-0.879294968735)*(1+2.*(dim-1.)/(o+1.))+(1.00617027695)*dim
        ref[3,1,0,0]=(0.229368586303)*(1+2.*(dim-1.)/(o+1.))+(0.77443067051)*dim
        ref[3,1,0,1]=(-1.10872819696)*(1+2.*(dim-1.)/(o+1.))+(2.31848089428)*dim
        ref[3,1,1,0]=(-0.63567553767)*(1+2.*(dim-1.)/(o+1.))+(0.639975743384)*dim
        ref[3,1,1,1]=(0.713294343931)*(1+2.*(dim-1.)/(o+1.))+(0.231547292007)*dim
        ref[3,1,2,0]=(0.769588527176)*(1+2.*(dim-1.)/(o+1.))+(-1.09018179488)*dim
        ref[3,1,2,1]=(0.817680613086)*(1+2.*(dim-1.)/(o+1.))+(0.333208233027)*dim
        ref[3,2,0,0]=(1.42693369167)*(1+2.*(dim-1.)/(o+1.))+(0.153052245499)*dim
        ref[3,2,0,1]=(0.0925205630387)*(1+2.*(dim-1.)/(o+1.))+(0.920294101604)*dim
        ref[3,2,1,0]=(0.127164695704)*(1+2.*(dim-1.)/(o+1.))+(-0.40145328165)*dim
        ref[3,2,1,1]=(2.1466065274)*(1+2.*(dim-1.)/(o+1.))+(0.303931102527)*dim
        ref[3,2,2,0]=(-0.304965940702)*(1+2.*(dim-1.)/(o+1.))+(-0.343047175114)*dim
        ref[3,2,2,1]=(0.0503924972819)*(1+2.*(dim-1.)/(o+1.))+(0.773796589222)*dim
        ref[3,3,0,0]=(-1.42943157618)*(1+2.*(dim-1.)/(o+1.))+(-0.972179810489)*dim
        ref[3,3,0,1]=(-0.0221855134254)*(1+2.*(dim-1.)/(o+1.))+(0.0982012306951)*dim
        ref[3,3,1,0]=(-1.28345186631)*(1+2.*(dim-1.)/(o+1.))+(1.19119981224)*dim
        ref[3,3,1,1]=(-0.212754224866)*(1+2.*(dim-1.)/(o+1.))+(0.220857411425)*dim
        ref[3,3,2,0]=(-0.194238559937)*(1+2.*(dim-1.)/(o+1.))+(0.117100106138)*dim
        ref[3,3,2,1]=(1.22683390029)*(1+2.*(dim-1.)/(o+1.))+(0.508406291409)*dim
        ref[3,4,0,0]=(-0.368643294124)*(1+2.*(dim-1.)/(o+1.))+(-0.623781546787)*dim
        ref[3,4,0,1]=(-0.101794807473)*(1+2.*(dim-1.)/(o+1.))+(1.77275593394)*dim
        ref[3,4,1,0]=(2.32521626647)*(1+2.*(dim-1.)/(o+1.))+(-0.534891885575)*dim
        ref[3,4,1,1]=(1.12921536112)*(1+2.*(dim-1.)/(o+1.))+(-2.50501195319)*dim
        ref[3,4,2,0]=(0.535434204129)*(1+2.*(dim-1.)/(o+1.))+(0.719322138661)*dim
        ref[3,4,2,1]=(-0.375491505946)*(1+2.*(dim-1.)/(o+1.))+(0.546287666371)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-0.747700242689)*x[0]**o+(-0.24253900484)*x[0]+(0.337715187384)*x[1]**o+(0.338764998592)*x[1]
        ref=(-0.409985055305)*(1+2.*(dim-1.)/(o+1.))+(0.0962259937514)*dim
      else:
        s=(-0.25178656924)*x[0]**o+(-0.295938166993)*x[0]+(0.0692176296902)*x[1]**o+(-0.514895147995)*x[1]+(0.34638140815)*x[2]**o+(-0.702417755803)*x[2]
        ref=(0.163812468601)*(1+2.*(dim-1.)/(o+1.))+(-1.51325107079)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(0.555314122492)*x[0]**o+(-0.808960080698)*x[0]+(0.884426560701)*x[1]**o+(-0.243549906992)*x[1]
        s[1]=(-0.0719219833378)*x[0]**o+(-0.403661400366)*x[0]+(0.666085789176)*x[1]**o+(-0.914772770267)*x[1]
        ref[0]=(1.43974068319)*(1+2.*(dim-1.)/(o+1.))+(-1.05250998769)*dim
        ref[1]=(0.594163805838)*(1+2.*(dim-1.)/(o+1.))+(-1.31843417063)*dim
      else:
        s[0]=(-0.826331014544)*x[0]**o+(-0.321975774952)*x[0]+(0.112378767915)*x[1]**o+(-0.756344985678)*x[1]+(0.719456760908)*x[2]**o+(-0.353602934037)*x[2]
        s[1]=(0.869195210566)*x[0]**o+(-0.315975726841)*x[0]+(0.712421301529)*x[1]**o+(0.215376989329)*x[1]+(0.964920869405)*x[2]**o+(0.922113174792)*x[2]
        ref[0]=(0.00550451427885)*(1+2.*(dim-1.)/(o+1.))+(-1.43192369467)*dim
        ref[1]=(2.5465373815)*(1+2.*(dim-1.)/(o+1.))+(0.82151443728)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(0.271641608649)*x[0]**o+(-0.685616858784)*x[0]+(0.644560357341)*x[1]**o+(0.294292988645)*x[1]
        s[0,1]=(-0.868199164404)*x[0]**o+(-0.352185589942)*x[0]+(-0.939001057807)*x[1]**o+(-0.806858466969)*x[1]
        s[0,2]=(-0.908891077111)*x[0]**o+(0.75605446917)*x[0]+(-0.921669916524)*x[1]**o+(-0.375347220131)*x[1]
        s[0,3]=(0.508605238535)*x[0]**o+(0.478464667724)*x[0]+(-0.632750076182)*x[1]**o+(-0.0150029824707)*x[1]
        s[0,4]=(-0.886187964736)*x[0]**o+(0.397671872055)*x[0]+(-0.216676829324)*x[1]**o+(0.986585859674)*x[1]
        s[1,0]=(0.630864334183)*x[0]**o+(-0.722653479356)*x[0]+(0.211515635986)*x[1]**o+(0.83331541895)*x[1]
        s[1,1]=(-0.354681281315)*x[0]**o+(-0.848733894579)*x[0]+(-0.101868246651)*x[1]**o+(-0.505303431554)*x[1]
        s[1,2]=(0.685694963182)*x[0]**o+(-0.950829894589)*x[0]+(-0.31792598055)*x[1]**o+(0.426249652943)*x[1]
        s[1,3]=(-0.614038433138)*x[0]**o+(0.723563718847)*x[0]+(0.752373220947)*x[1]**o+(0.613330337959)*x[1]
        s[1,4]=(-0.998818105353)*x[0]**o+(0.269274342639)*x[0]+(-0.902068964726)*x[1]**o+(-0.191309514428)*x[1]
        s[2,0]=(0.249323085595)*x[0]**o+(0.411304825013)*x[0]+(0.312377963164)*x[1]**o+(-0.311446803448)*x[1]
        s[2,1]=(-0.290453739481)*x[0]**o+(-0.202745911645)*x[0]+(-0.309021366331)*x[1]**o+(-0.173018322631)*x[1]
        s[2,2]=(0.691837531968)*x[0]**o+(0.672203641337)*x[0]+(0.0217882582663)*x[1]**o+(-0.776821250479)*x[1]
        s[2,3]=(0.00878696630163)*x[0]**o+(-0.565790699611)*x[0]+(-0.748591951135)*x[1]**o+(-0.46249464823)*x[1]
        s[2,4]=(-0.638178053099)*x[0]**o+(0.898061575109)*x[0]+(0.498715053696)*x[1]**o+(-0.236164377825)*x[1]
        s[3,0]=(0.843383855229)*x[0]**o+(0.652320113529)*x[0]+(0.34384880146)*x[1]**o+(-0.539878095901)*x[1]
        s[3,1]=(-0.0571664516277)*x[0]**o+(-0.658773983784)*x[0]+(0.160437286579)*x[1]**o+(-0.805774084186)*x[1]
        s[3,2]=(0.340124131227)*x[0]**o+(-0.464143943808)*x[0]+(-0.215724355021)*x[1]**o+(0.206995713005)*x[1]
        s[3,3]=(0.825533013397)*x[0]**o+(0.695982840481)*x[0]+(0.355542963602)*x[1]**o+(-0.401849802264)*x[1]
        s[3,4]=(-0.367312805197)*x[0]**o+(-0.012173639311)*x[0]+(0.581566155275)*x[1]**o+(-0.650284343373)*x[1]
        ref[0,0]=(0.91620196599)*(1+2.*(dim-1.)/(o+1.))+(-0.39132387014)*dim
        ref[0,1]=(-1.80720022221)*(1+2.*(dim-1.)/(o+1.))+(-1.15904405691)*dim
        ref[0,2]=(-1.83056099363)*(1+2.*(dim-1.)/(o+1.))+(0.380707249038)*dim
        ref[0,3]=(-0.124144837647)*(1+2.*(dim-1.)/(o+1.))+(0.463461685254)*dim
        ref[0,4]=(-1.10286479406)*(1+2.*(dim-1.)/(o+1.))+(1.38425773173)*dim
        ref[1,0]=(0.842379970168)*(1+2.*(dim-1.)/(o+1.))+(0.110661939594)*dim
        ref[1,1]=(-0.456549527966)*(1+2.*(dim-1.)/(o+1.))+(-1.35403732613)*dim
        ref[1,2]=(0.367768982632)*(1+2.*(dim-1.)/(o+1.))+(-0.524580241646)*dim
        ref[1,3]=(0.138334787809)*(1+2.*(dim-1.)/(o+1.))+(1.33689405681)*dim
        ref[1,4]=(-1.90088707008)*(1+2.*(dim-1.)/(o+1.))+(0.0779648282107)*dim
        ref[2,0]=(0.561701048758)*(1+2.*(dim-1.)/(o+1.))+(0.0998580215651)*dim
        ref[2,1]=(-0.599475105813)*(1+2.*(dim-1.)/(o+1.))+(-0.375764234276)*dim
        ref[2,2]=(0.713625790235)*(1+2.*(dim-1.)/(o+1.))+(-0.104617609143)*dim
        ref[2,3]=(-0.739804984833)*(1+2.*(dim-1.)/(o+1.))+(-1.02828534784)*dim
        ref[2,4]=(-0.139462999403)*(1+2.*(dim-1.)/(o+1.))+(0.661897197284)*dim
        ref[3,0]=(1.18723265669)*(1+2.*(dim-1.)/(o+1.))+(0.112442017629)*dim
        ref[3,1]=(0.103270834951)*(1+2.*(dim-1.)/(o+1.))+(-1.46454806797)*dim
        ref[3,2]=(0.124399776206)*(1+2.*(dim-1.)/(o+1.))+(-0.257148230803)*dim
        ref[3,3]=(1.181075977)*(1+2.*(dim-1.)/(o+1.))+(0.294133038216)*dim
        ref[3,4]=(0.214253350078)*(1+2.*(dim-1.)/(o+1.))+(-0.662457982684)*dim
      else:
        s[0,0]=(0.488698481646)*x[0]**o+(-0.0857636443438)*x[0]+(0.879463791392)*x[1]**o+(-0.748151251773)*x[1]+(0.342272219635)*x[2]**o+(0.148033760154)*x[2]
        s[0,1]=(0.716943874304)*x[0]**o+(-0.582838388256)*x[0]+(0.799901107796)*x[1]**o+(-0.814196005238)*x[1]+(0.229983496729)*x[2]**o+(-0.153919124028)*x[2]
        s[0,2]=(-0.898513198581)*x[0]**o+(0.380948693198)*x[0]+(0.357237014893)*x[1]**o+(0.650057181845)*x[1]+(0.192484147287)*x[2]**o+(-0.00114028159222)*x[2]
        s[0,3]=(0.882708886033)*x[0]**o+(0.412063277674)*x[0]+(-0.801825884205)*x[1]**o+(0.421365647832)*x[1]+(0.0732340922267)*x[2]**o+(0.831590828322)*x[2]
        s[0,4]=(-0.893541294976)*x[0]**o+(0.616353784204)*x[0]+(0.43931453774)*x[1]**o+(0.60163593029)*x[1]+(0.171225928123)*x[2]**o+(-0.809624161102)*x[2]
        s[1,0]=(0.3743694715)*x[0]**o+(0.973362616567)*x[0]+(-0.790022350688)*x[1]**o+(0.438346450916)*x[1]+(-0.85644858607)*x[2]**o+(0.861754412538)*x[2]
        s[1,1]=(0.0682853717145)*x[0]**o+(-0.483847902471)*x[0]+(0.830085512097)*x[1]**o+(0.330508826392)*x[1]+(-0.341254677789)*x[2]**o+(0.0388136714901)*x[2]
        s[1,2]=(0.740241060442)*x[0]**o+(0.569699186818)*x[0]+(0.292637941654)*x[1]**o+(-0.126938078274)*x[1]+(-0.400665686855)*x[2]**o+(-0.0981290395971)*x[2]
        s[1,3]=(0.611312886004)*x[0]**o+(0.570096324275)*x[0]+(0.222499932388)*x[1]**o+(-0.276552061893)*x[1]+(-0.370405232883)*x[2]**o+(-0.91757884865)*x[2]
        s[1,4]=(-0.844423209218)*x[0]**o+(-0.357603990072)*x[0]+(-0.291569596715)*x[1]**o+(0.127364564811)*x[1]+(-0.427115349787)*x[2]**o+(-0.689895532918)*x[2]
        s[2,0]=(-0.798050448891)*x[0]**o+(-0.197062766104)*x[0]+(-0.363892466443)*x[1]**o+(0.154013357585)*x[1]+(-0.398682782201)*x[2]**o+(-0.529551264949)*x[2]
        s[2,1]=(-0.0912777859374)*x[0]**o+(-0.675884283313)*x[0]+(0.516144657124)*x[1]**o+(-0.53265938856)*x[1]+(0.95476185698)*x[2]**o+(-0.313549200036)*x[2]
        s[2,2]=(0.868959641933)*x[0]**o+(0.962373087968)*x[0]+(-0.848379282457)*x[1]**o+(0.820817107146)*x[1]+(0.498221865324)*x[2]**o+(-0.885755811184)*x[2]
        s[2,3]=(0.552181728294)*x[0]**o+(-0.943385218578)*x[0]+(-0.650886151517)*x[1]**o+(0.204702323464)*x[1]+(-0.419413465102)*x[2]**o+(0.761759803195)*x[2]
        s[2,4]=(0.0446686935799)*x[0]**o+(-0.850084331283)*x[0]+(0.215197552732)*x[1]**o+(0.0524953736999)*x[1]+(-0.828056084121)*x[2]**o+(-0.960944367985)*x[2]
        s[3,0]=(-0.475014597092)*x[0]**o+(-0.100740967917)*x[0]+(0.76769378138)*x[1]**o+(0.846896186419)*x[1]+(0.794108335473)*x[2]**o+(-0.34345830075)*x[2]
        s[3,1]=(0.10453345316)*x[0]**o+(-0.677797655695)*x[0]+(0.471066711536)*x[1]**o+(0.534601952265)*x[1]+(-0.685589576412)*x[2]**o+(-0.025255709402)*x[2]
        s[3,2]=(0.0410162645138)*x[0]**o+(-0.948466999545)*x[0]+(0.147271170264)*x[1]**o+(0.17422632965)*x[1]+(0.425319252082)*x[2]**o+(-0.395792817384)*x[2]
        s[3,3]=(0.301093567659)*x[0]**o+(-0.48405705123)*x[0]+(-0.681930643655)*x[1]**o+(0.848540291162)*x[1]+(0.815437001935)*x[2]**o+(-0.734620301961)*x[2]
        s[3,4]=(-0.586209686649)*x[0]**o+(-0.747006165956)*x[0]+(0.0368543812482)*x[1]**o+(-0.976301323297)*x[1]+(0.748206708653)*x[2]**o+(0.833845422962)*x[2]
        ref[0,0]=(1.71043449267)*(1+2.*(dim-1.)/(o+1.))+(-0.685881135963)*dim
        ref[0,1]=(1.74682847883)*(1+2.*(dim-1.)/(o+1.))+(-1.55095351752)*dim
        ref[0,2]=(-0.348792036401)*(1+2.*(dim-1.)/(o+1.))+(1.02986559345)*dim
        ref[0,3]=(0.154117094054)*(1+2.*(dim-1.)/(o+1.))+(1.66501975383)*dim
        ref[0,4]=(-0.283000829113)*(1+2.*(dim-1.)/(o+1.))+(0.408365553392)*dim
        ref[1,0]=(-1.27210146526)*(1+2.*(dim-1.)/(o+1.))+(2.27346348002)*dim
        ref[1,1]=(0.557116206022)*(1+2.*(dim-1.)/(o+1.))+(-0.114525404589)*dim
        ref[1,2]=(0.632213315241)*(1+2.*(dim-1.)/(o+1.))+(0.344632068947)*dim
        ref[1,3]=(0.463407585509)*(1+2.*(dim-1.)/(o+1.))+(-0.624034586267)*dim
        ref[1,4]=(-1.56310815572)*(1+2.*(dim-1.)/(o+1.))+(-0.920134958179)*dim
        ref[2,0]=(-1.56062569754)*(1+2.*(dim-1.)/(o+1.))+(-0.572600673468)*dim
        ref[2,1]=(1.37962872817)*(1+2.*(dim-1.)/(o+1.))+(-1.52209287191)*dim
        ref[2,2]=(0.5188022248)*(1+2.*(dim-1.)/(o+1.))+(0.897434383931)*dim
        ref[2,3]=(-0.518117888325)*(1+2.*(dim-1.)/(o+1.))+(0.0230769080811)*dim
        ref[2,4]=(-0.568189837809)*(1+2.*(dim-1.)/(o+1.))+(-1.75853332557)*dim
        ref[3,0]=(1.08678751976)*(1+2.*(dim-1.)/(o+1.))+(0.402696917752)*dim
        ref[3,1]=(-0.109989411716)*(1+2.*(dim-1.)/(o+1.))+(-0.168451412832)*dim
        ref[3,2]=(0.61360668686)*(1+2.*(dim-1.)/(o+1.))+(-1.17003348728)*dim
        ref[3,3]=(0.434599925938)*(1+2.*(dim-1.)/(o+1.))+(-0.370137062028)*dim
        ref[3,4]=(0.198851403252)*(1+2.*(dim-1.)/(o+1.))+(-0.88946206629)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.591558717335)*x[0]**o+(-0.671839230736)*x[0]+(0.107743853024)*x[1]**o+(0.595347539232)*x[1]
        s[0,0,1]=(0.25500384698)*x[0]**o+(0.944399712188)*x[0]+(-0.999249484326)*x[1]**o+(-0.495212132457)*x[1]
        s[0,1,0]=(0.236244996704)*x[0]**o+(0.704807264803)*x[0]+(-0.242028416637)*x[1]**o+(-0.553320357705)*x[1]
        s[0,1,1]=(0.552969384309)*x[0]**o+(-0.941333552159)*x[0]+(-0.696785515448)*x[1]**o+(0.851994229748)*x[1]
        s[1,0,0]=(0.0953739386185)*x[0]**o+(-0.50227621093)*x[0]+(-0.994199466154)*x[1]**o+(0.975204139168)*x[1]
        s[1,0,1]=(0.0251612361822)*x[0]**o+(-0.984244259014)*x[0]+(0.928117852579)*x[1]**o+(0.0302634017269)*x[1]
        s[1,1,0]=(-0.800615084919)*x[0]**o+(0.71993242468)*x[0]+(0.714856567768)*x[1]**o+(-0.497753089167)*x[1]
        s[1,1,1]=(-0.916226186485)*x[0]**o+(0.81028396707)*x[0]+(-0.156900817441)*x[1]**o+(-0.093034093703)*x[1]
        s[2,0,0]=(-0.738732605581)*x[0]**o+(-0.77907436774)*x[0]+(0.948664797284)*x[1]**o+(-0.0764721201972)*x[1]
        s[2,0,1]=(-0.11146154208)*x[0]**o+(0.145299573751)*x[0]+(0.970129117378)*x[1]**o+(-0.272338564084)*x[1]
        s[2,1,0]=(-0.197948912143)*x[0]**o+(-0.902880824888)*x[0]+(0.579307317816)*x[1]**o+(-0.852996341605)*x[1]
        s[2,1,1]=(0.563711575633)*x[0]**o+(-0.288761476406)*x[0]+(0.7854621452)*x[1]**o+(-0.784647403092)*x[1]
        s[3,0,0]=(-0.116979739809)*x[0]**o+(0.964821934046)*x[0]+(0.441879236444)*x[1]**o+(0.245324297501)*x[1]
        s[3,0,1]=(-0.63895256779)*x[0]**o+(0.446532064378)*x[0]+(0.703018532945)*x[1]**o+(0.398978822243)*x[1]
        s[3,1,0]=(0.347022657024)*x[0]**o+(0.330266740052)*x[0]+(0.9044236351)*x[1]**o+(0.463728515326)*x[1]
        s[3,1,1]=(0.673344024583)*x[0]**o+(0.218503531592)*x[0]+(-0.981631285231)*x[1]**o+(0.493256781677)*x[1]
        s[4,0,0]=(-0.0337376408638)*x[0]**o+(0.006116830911)*x[0]+(-0.186073327302)*x[1]**o+(0.0809017607404)*x[1]
        s[4,0,1]=(0.624443889423)*x[0]**o+(0.727204963203)*x[0]+(0.724716349734)*x[1]**o+(-0.360798281504)*x[1]
        s[4,1,0]=(-0.316602217463)*x[0]**o+(-0.775985201313)*x[0]+(0.313467720837)*x[1]**o+(0.19768983092)*x[1]
        s[4,1,1]=(0.578665229532)*x[0]**o+(-0.962255191175)*x[0]+(0.341888555373)*x[1]**o+(-0.389440724436)*x[1]
        s[5,0,0]=(0.256912289915)*x[0]**o+(-0.171401072804)*x[0]+(-0.646582287213)*x[1]**o+(0.211366542089)*x[1]
        s[5,0,1]=(0.393310638354)*x[0]**o+(0.319973991926)*x[0]+(0.626363008584)*x[1]**o+(-0.0843826544875)*x[1]
        s[5,1,0]=(0.425476941925)*x[0]**o+(0.913999716353)*x[0]+(-0.534398944156)*x[1]**o+(-0.83386676719)*x[1]
        s[5,1,1]=(0.660737270239)*x[0]**o+(-0.828622795768)*x[0]+(0.488808193457)*x[1]**o+(-0.352537348251)*x[1]
        ref[0,0,0]=(0.699302570359)*(1+2.*(dim-1.)/(o+1.))+(-0.076491691504)*dim
        ref[0,0,1]=(-0.744245637345)*(1+2.*(dim-1.)/(o+1.))+(0.449187579731)*dim
        ref[0,1,0]=(-0.00578341993308)*(1+2.*(dim-1.)/(o+1.))+(0.151486907099)*dim
        ref[0,1,1]=(-0.14381613114)*(1+2.*(dim-1.)/(o+1.))+(-0.0893393224107)*dim
        ref[1,0,0]=(-0.898825527536)*(1+2.*(dim-1.)/(o+1.))+(0.472927928237)*dim
        ref[1,0,1]=(0.953279088761)*(1+2.*(dim-1.)/(o+1.))+(-0.953980857287)*dim
        ref[1,1,0]=(-0.0857585171509)*(1+2.*(dim-1.)/(o+1.))+(0.222179335513)*dim
        ref[1,1,1]=(-1.07312700393)*(1+2.*(dim-1.)/(o+1.))+(0.717249873367)*dim
        ref[2,0,0]=(0.209932191703)*(1+2.*(dim-1.)/(o+1.))+(-0.855546487937)*dim
        ref[2,0,1]=(0.858667575297)*(1+2.*(dim-1.)/(o+1.))+(-0.127038990333)*dim
        ref[2,1,0]=(0.381358405673)*(1+2.*(dim-1.)/(o+1.))+(-1.75587716649)*dim
        ref[2,1,1]=(1.34917372083)*(1+2.*(dim-1.)/(o+1.))+(-1.0734088795)*dim
        ref[3,0,0]=(0.324899496635)*(1+2.*(dim-1.)/(o+1.))+(1.21014623155)*dim
        ref[3,0,1]=(0.0640659651547)*(1+2.*(dim-1.)/(o+1.))+(0.84551088662)*dim
        ref[3,1,0]=(1.25144629212)*(1+2.*(dim-1.)/(o+1.))+(0.793995255378)*dim
        ref[3,1,1]=(-0.308287260648)*(1+2.*(dim-1.)/(o+1.))+(0.711760313269)*dim
        ref[4,0,0]=(-0.219810968166)*(1+2.*(dim-1.)/(o+1.))+(0.0870185916514)*dim
        ref[4,0,1]=(1.34916023916)*(1+2.*(dim-1.)/(o+1.))+(0.366406681699)*dim
        ref[4,1,0]=(-0.00313449662664)*(1+2.*(dim-1.)/(o+1.))+(-0.578295370393)*dim
        ref[4,1,1]=(0.920553784904)*(1+2.*(dim-1.)/(o+1.))+(-1.35169591561)*dim
        ref[5,0,0]=(-0.389669997298)*(1+2.*(dim-1.)/(o+1.))+(0.0399654692848)*dim
        ref[5,0,1]=(1.01967364694)*(1+2.*(dim-1.)/(o+1.))+(0.235591337439)*dim
        ref[5,1,0]=(-0.108922002231)*(1+2.*(dim-1.)/(o+1.))+(0.0801329491631)*dim
        ref[5,1,1]=(1.1495454637)*(1+2.*(dim-1.)/(o+1.))+(-1.18116014402)*dim
      else:
        s[0,0,0]=(0.267370933512)*x[0]**o+(0.155479029604)*x[0]+(0.483226114118)*x[1]**o+(0.0331367362161)*x[1]+(-0.958604071433)*x[2]**o+(-0.938589620173)*x[2]
        s[0,0,1]=(0.558184625201)*x[0]**o+(-0.442315117978)*x[0]+(0.443008244723)*x[1]**o+(0.10094012777)*x[1]+(-0.884286721018)*x[2]**o+(-0.234859150202)*x[2]
        s[0,1,0]=(0.00707521319666)*x[0]**o+(-0.436445369554)*x[0]+(0.0339511386587)*x[1]**o+(-0.4912959825)*x[1]+(0.387274090979)*x[2]**o+(-0.634777694546)*x[2]
        s[0,1,1]=(-0.169879301168)*x[0]**o+(0.745060242424)*x[0]+(-0.518240485868)*x[1]**o+(0.0555475122839)*x[1]+(0.420456483719)*x[2]**o+(-0.805694025108)*x[2]
        s[1,0,0]=(0.626862061657)*x[0]**o+(0.286735499998)*x[0]+(-0.154542230383)*x[1]**o+(-0.659059291598)*x[1]+(0.700399779356)*x[2]**o+(-0.149494578344)*x[2]
        s[1,0,1]=(-0.85312717829)*x[0]**o+(0.827572890814)*x[0]+(0.929068438324)*x[1]**o+(0.783234890681)*x[1]+(0.132385694002)*x[2]**o+(-0.0375332819318)*x[2]
        s[1,1,0]=(0.828428181973)*x[0]**o+(0.489617430458)*x[0]+(0.0915776189363)*x[1]**o+(-0.425176021653)*x[1]+(0.80387559654)*x[2]**o+(-0.681443382816)*x[2]
        s[1,1,1]=(-0.65230596594)*x[0]**o+(0.170131891444)*x[0]+(0.0830524896765)*x[1]**o+(-0.221764900615)*x[1]+(-0.645608220391)*x[2]**o+(0.400795086408)*x[2]
        s[2,0,0]=(0.893199207029)*x[0]**o+(-0.872998450109)*x[0]+(0.270918838822)*x[1]**o+(0.0111491233197)*x[1]+(-0.0264854465472)*x[2]**o+(0.955927944305)*x[2]
        s[2,0,1]=(0.240242096164)*x[0]**o+(-0.161479794451)*x[0]+(-0.66104753617)*x[1]**o+(0.946080550241)*x[1]+(-0.397485102557)*x[2]**o+(0.886966644933)*x[2]
        s[2,1,0]=(0.80637989046)*x[0]**o+(-0.876797025996)*x[0]+(-0.250156824474)*x[1]**o+(0.956885396073)*x[1]+(-0.721161989016)*x[2]**o+(-0.957921265653)*x[2]
        s[2,1,1]=(-0.769495191281)*x[0]**o+(-0.51589948382)*x[0]+(0.472246475849)*x[1]**o+(0.833747696193)*x[1]+(-0.810945745724)*x[2]**o+(0.00314465415233)*x[2]
        s[3,0,0]=(-0.710791737222)*x[0]**o+(0.608387037146)*x[0]+(0.911509216107)*x[1]**o+(-0.292243195309)*x[1]+(-0.957172951675)*x[2]**o+(0.524831114463)*x[2]
        s[3,0,1]=(0.892735233153)*x[0]**o+(0.28183990495)*x[0]+(0.168860806917)*x[1]**o+(0.388164284517)*x[1]+(0.144580027346)*x[2]**o+(0.881037146671)*x[2]
        s[3,1,0]=(0.221930595909)*x[0]**o+(0.903824792034)*x[0]+(0.271360511407)*x[1]**o+(-0.951642792919)*x[1]+(-0.100229570494)*x[2]**o+(-0.355195001599)*x[2]
        s[3,1,1]=(0.875769927418)*x[0]**o+(-0.161604269687)*x[0]+(0.74568148533)*x[1]**o+(-0.553173791916)*x[1]+(-0.458908157129)*x[2]**o+(-0.247242445331)*x[2]
        s[4,0,0]=(-0.515082637068)*x[0]**o+(0.456575233293)*x[0]+(-0.607867524532)*x[1]**o+(-0.155257786478)*x[1]+(-0.845324963524)*x[2]**o+(-0.862815318477)*x[2]
        s[4,0,1]=(0.922203259385)*x[0]**o+(-0.752313452972)*x[0]+(0.10594289591)*x[1]**o+(-0.370275136504)*x[1]+(-0.0979341688932)*x[2]**o+(-0.066305843593)*x[2]
        s[4,1,0]=(-0.849429313411)*x[0]**o+(-0.28348706029)*x[0]+(0.423697075901)*x[1]**o+(0.724791178434)*x[1]+(0.608888753747)*x[2]**o+(-0.975600085008)*x[2]
        s[4,1,1]=(-0.277835181883)*x[0]**o+(0.255125892715)*x[0]+(-0.801810191818)*x[1]**o+(-0.107375403424)*x[1]+(0.746887097735)*x[2]**o+(0.406084178688)*x[2]
        s[5,0,0]=(-0.137289244094)*x[0]**o+(0.76622915286)*x[0]+(-0.354285839406)*x[1]**o+(-0.195958436789)*x[1]+(-0.836211369427)*x[2]**o+(-0.684714854823)*x[2]
        s[5,0,1]=(0.263960517384)*x[0]**o+(0.908002470581)*x[0]+(0.439729777064)*x[1]**o+(-0.274244779759)*x[1]+(0.659934789159)*x[2]**o+(-0.672573099237)*x[2]
        s[5,1,0]=(-0.20882025835)*x[0]**o+(-0.528304766773)*x[0]+(-0.279646805628)*x[1]**o+(-0.23736563075)*x[1]+(-0.439565255672)*x[2]**o+(0.956815955079)*x[2]
        s[5,1,1]=(-0.294480977855)*x[0]**o+(-0.51059423384)*x[0]+(0.371920382084)*x[1]**o+(0.648458734295)*x[1]+(0.0335200465334)*x[2]**o+(-0.923801240024)*x[2]
        ref[0,0,0]=(-0.208007023803)*(1+2.*(dim-1.)/(o+1.))+(-0.749973854353)*dim
        ref[0,0,1]=(0.116906148907)*(1+2.*(dim-1.)/(o+1.))+(-0.57623414041)*dim
        ref[0,1,0]=(0.428300442835)*(1+2.*(dim-1.)/(o+1.))+(-1.5625190466)*dim
        ref[0,1,1]=(-0.267663303317)*(1+2.*(dim-1.)/(o+1.))+(-0.00508627039966)*dim
        ref[1,0,0]=(1.17271961063)*(1+2.*(dim-1.)/(o+1.))+(-0.521818369944)*dim
        ref[1,0,1]=(0.208326954036)*(1+2.*(dim-1.)/(o+1.))+(1.57327449956)*dim
        ref[1,1,0]=(1.72388139745)*(1+2.*(dim-1.)/(o+1.))+(-0.61700197401)*dim
        ref[1,1,1]=(-1.21486169665)*(1+2.*(dim-1.)/(o+1.))+(0.349162077237)*dim
        ref[2,0,0]=(1.1376325993)*(1+2.*(dim-1.)/(o+1.))+(0.0940786175155)*dim
        ref[2,0,1]=(-0.818290542564)*(1+2.*(dim-1.)/(o+1.))+(1.67156740072)*dim
        ref[2,1,0]=(-0.16493892303)*(1+2.*(dim-1.)/(o+1.))+(-0.877832895575)*dim
        ref[2,1,1]=(-1.10819446116)*(1+2.*(dim-1.)/(o+1.))+(0.320992866525)*dim
        ref[3,0,0]=(-0.75645547279)*(1+2.*(dim-1.)/(o+1.))+(0.8409749563)*dim
        ref[3,0,1]=(1.20617606742)*(1+2.*(dim-1.)/(o+1.))+(1.55104133614)*dim
        ref[3,1,0]=(0.393061536822)*(1+2.*(dim-1.)/(o+1.))+(-0.403013002484)*dim
        ref[3,1,1]=(1.16254325562)*(1+2.*(dim-1.)/(o+1.))+(-0.962020506935)*dim
        ref[4,0,0]=(-1.96827512512)*(1+2.*(dim-1.)/(o+1.))+(-0.561497871662)*dim
        ref[4,0,1]=(0.930211986402)*(1+2.*(dim-1.)/(o+1.))+(-1.18889443307)*dim
        ref[4,1,0]=(0.183156516238)*(1+2.*(dim-1.)/(o+1.))+(-0.534295966865)*dim
        ref[4,1,1]=(-0.332758275966)*(1+2.*(dim-1.)/(o+1.))+(0.553834667978)*dim
        ref[5,0,0]=(-1.32778645293)*(1+2.*(dim-1.)/(o+1.))+(-0.114444138751)*dim
        ref[5,0,1]=(1.36362508361)*(1+2.*(dim-1.)/(o+1.))+(-0.038815408415)*dim
        ref[5,1,0]=(-0.928032319651)*(1+2.*(dim-1.)/(o+1.))+(0.191145557556)*dim
        ref[5,1,1]=(0.110959450763)*(1+2.*(dim-1.)/(o+1.))+(-0.785936739569)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunctionOnBoundary

      assumptions: Solution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-0.512388618715)*x[0]**o+(-0.456786510339)*x[0]+(-0.369955618695)*x[1]**o+(0.284589526283)*x[1]
        s[0,0,0,1]=(0.45694429765)*x[0]**o+(-0.951226590686)*x[0]+(0.812195765321)*x[1]**o+(-0.450606422037)*x[1]
        s[0,0,1,0]=(0.917138960838)*x[0]**o+(0.846015644843)*x[0]+(0.311194587689)*x[1]**o+(0.318765403889)*x[1]
        s[0,0,1,1]=(0.912716498884)*x[0]**o+(-0.309788565863)*x[0]+(0.582625846661)*x[1]**o+(-0.786597172188)*x[1]
        s[0,0,2,0]=(-0.335643205116)*x[0]**o+(-0.162356382231)*x[0]+(0.33024736395)*x[1]**o+(0.137133359333)*x[1]
        s[0,0,2,1]=(-0.853219359404)*x[0]**o+(0.20806173242)*x[0]+(0.904614991529)*x[1]**o+(-0.0927322673776)*x[1]
        s[0,1,0,0]=(-0.116418078552)*x[0]**o+(-0.450766009165)*x[0]+(-0.109027531459)*x[1]**o+(-0.393345607974)*x[1]
        s[0,1,0,1]=(-0.425371185843)*x[0]**o+(-0.513194578298)*x[0]+(0.2204981189)*x[1]**o+(0.959807684568)*x[1]
        s[0,1,1,0]=(-0.184006516133)*x[0]**o+(0.724279611778)*x[0]+(0.233453086887)*x[1]**o+(-0.304433218327)*x[1]
        s[0,1,1,1]=(-0.675234346597)*x[0]**o+(0.444869722037)*x[0]+(-0.0503551086423)*x[1]**o+(-0.0766529306108)*x[1]
        s[0,1,2,0]=(0.841378546422)*x[0]**o+(-0.415767745596)*x[0]+(0.203941853234)*x[1]**o+(0.908654560689)*x[1]
        s[0,1,2,1]=(0.920754900511)*x[0]**o+(-0.630439736764)*x[0]+(-0.58028124877)*x[1]**o+(0.118829809363)*x[1]
        s[0,2,0,0]=(-0.307376557058)*x[0]**o+(0.782274756896)*x[0]+(0.996380983481)*x[1]**o+(-0.91041006941)*x[1]
        s[0,2,0,1]=(0.445454838496)*x[0]**o+(-0.278761052714)*x[0]+(-0.211616677518)*x[1]**o+(-0.982593785248)*x[1]
        s[0,2,1,0]=(-0.304509030828)*x[0]**o+(0.671529029626)*x[0]+(-0.883970223268)*x[1]**o+(-0.5778649808)*x[1]
        s[0,2,1,1]=(-0.463199486208)*x[0]**o+(0.346898357742)*x[0]+(-0.838746712872)*x[1]**o+(-0.303045089364)*x[1]
        s[0,2,2,0]=(-0.561239833326)*x[0]**o+(0.0861796115948)*x[0]+(-0.233810358053)*x[1]**o+(0.667590859355)*x[1]
        s[0,2,2,1]=(-0.0566127020774)*x[0]**o+(-0.702171135934)*x[0]+(0.156408646225)*x[1]**o+(0.503705135348)*x[1]
        s[0,3,0,0]=(-0.213903115172)*x[0]**o+(0.442577130625)*x[0]+(0.973915195866)*x[1]**o+(0.666138357006)*x[1]
        s[0,3,0,1]=(0.888546931537)*x[0]**o+(-0.581620837696)*x[0]+(-0.493141457997)*x[1]**o+(0.572137876409)*x[1]
        s[0,3,1,0]=(-0.198788230198)*x[0]**o+(0.399480321258)*x[0]+(0.437641855366)*x[1]**o+(-0.996597321011)*x[1]
        s[0,3,1,1]=(0.220561040916)*x[0]**o+(-0.613655442822)*x[0]+(0.961868845852)*x[1]**o+(-0.0593807869087)*x[1]
        s[0,3,2,0]=(0.590719334754)*x[0]**o+(0.675522468843)*x[0]+(0.420015363742)*x[1]**o+(-0.779408872787)*x[1]
        s[0,3,2,1]=(-0.739675077785)*x[0]**o+(0.868116107031)*x[0]+(-0.986504469737)*x[1]**o+(0.060324260277)*x[1]
        s[0,4,0,0]=(0.0611797596706)*x[0]**o+(0.804524773235)*x[0]+(0.987009736595)*x[1]**o+(0.811062048039)*x[1]
        s[0,4,0,1]=(0.47425001801)*x[0]**o+(-0.921861981358)*x[0]+(0.344921072375)*x[1]**o+(-0.601620009804)*x[1]
        s[0,4,1,0]=(0.0906565717033)*x[0]**o+(0.152797946513)*x[0]+(-0.567184628241)*x[1]**o+(-0.454348146448)*x[1]
        s[0,4,1,1]=(-0.0896053537692)*x[0]**o+(-0.895659709172)*x[0]+(0.0992922782923)*x[1]**o+(0.491225950996)*x[1]
        s[0,4,2,0]=(-0.810399276984)*x[0]**o+(-0.304522341713)*x[0]+(-0.247445852791)*x[1]**o+(-0.043454035062)*x[1]
        s[0,4,2,1]=(0.798398078646)*x[0]**o+(0.919534042298)*x[0]+(-0.558649299444)*x[1]**o+(-0.583442934514)*x[1]
        s[1,0,0,0]=(0.00573882500165)*x[0]**o+(-0.942216312116)*x[0]+(0.220926627969)*x[1]**o+(-0.363895609692)*x[1]
        s[1,0,0,1]=(-0.849679135789)*x[0]**o+(0.527924450097)*x[0]+(0.79947401514)*x[1]**o+(0.110179654418)*x[1]
        s[1,0,1,0]=(-0.990390680757)*x[0]**o+(0.115770686732)*x[0]+(0.0519459377792)*x[1]**o+(-0.405025360743)*x[1]
        s[1,0,1,1]=(-0.217170752126)*x[0]**o+(0.104026029604)*x[0]+(0.935952773688)*x[1]**o+(0.205909597691)*x[1]
        s[1,0,2,0]=(0.355936537495)*x[0]**o+(0.123599975203)*x[0]+(0.992745060396)*x[1]**o+(0.230955110212)*x[1]
        s[1,0,2,1]=(-0.337760143624)*x[0]**o+(-0.120785608957)*x[0]+(0.519621755185)*x[1]**o+(0.648090620962)*x[1]
        s[1,1,0,0]=(0.425227533029)*x[0]**o+(0.431947155274)*x[0]+(0.864871771139)*x[1]**o+(-0.421776717503)*x[1]
        s[1,1,0,1]=(-0.155048210381)*x[0]**o+(0.98349398745)*x[0]+(0.616262478346)*x[1]**o+(0.354130744218)*x[1]
        s[1,1,1,0]=(-0.255422611215)*x[0]**o+(0.610000128329)*x[0]+(0.692199306628)*x[1]**o+(-0.469569470881)*x[1]
        s[1,1,1,1]=(-0.00176886794175)*x[0]**o+(-0.408069848741)*x[0]+(0.708816789956)*x[1]**o+(0.823083702629)*x[1]
        s[1,1,2,0]=(0.229145338062)*x[0]**o+(-0.468402804906)*x[0]+(-0.352476560454)*x[1]**o+(0.968823931554)*x[1]
        s[1,1,2,1]=(-0.414346479022)*x[0]**o+(-0.0181085935776)*x[0]+(-0.170858935466)*x[1]**o+(-0.816432101623)*x[1]
        s[1,2,0,0]=(-0.0948608343933)*x[0]**o+(0.0068199755948)*x[0]+(-0.145376398386)*x[1]**o+(0.755438498846)*x[1]
        s[1,2,0,1]=(-0.665050941689)*x[0]**o+(-0.719231494887)*x[0]+(0.41492847392)*x[1]**o+(0.708962325583)*x[1]
        s[1,2,1,0]=(0.621628314609)*x[0]**o+(-0.997324959187)*x[0]+(0.211421770712)*x[1]**o+(0.272336354113)*x[1]
        s[1,2,1,1]=(-0.744309774218)*x[0]**o+(-0.901777776216)*x[0]+(-0.727378511259)*x[1]**o+(0.276331604178)*x[1]
        s[1,2,2,0]=(0.119470505814)*x[0]**o+(-0.377429184576)*x[0]+(0.634631061717)*x[1]**o+(0.71726575358)*x[1]
        s[1,2,2,1]=(0.196579032787)*x[0]**o+(-0.417495130697)*x[0]+(-0.469739874787)*x[1]**o+(-0.775290208858)*x[1]
        s[1,3,0,0]=(0.406483838142)*x[0]**o+(0.184354975226)*x[0]+(0.762003327026)*x[1]**o+(0.0300461028065)*x[1]
        s[1,3,0,1]=(-0.844790086513)*x[0]**o+(0.663686153475)*x[0]+(0.567312988723)*x[1]**o+(0.508368547154)*x[1]
        s[1,3,1,0]=(-0.6417260084)*x[0]**o+(0.592764591647)*x[0]+(-0.335643553939)*x[1]**o+(0.152782626)*x[1]
        s[1,3,1,1]=(-0.570864881875)*x[0]**o+(-0.610576918843)*x[0]+(-0.0434074971721)*x[1]**o+(0.373353748894)*x[1]
        s[1,3,2,0]=(0.951712868404)*x[0]**o+(0.837001943514)*x[0]+(-0.661919006256)*x[1]**o+(0.136870953867)*x[1]
        s[1,3,2,1]=(0.338496531342)*x[0]**o+(0.646315658139)*x[0]+(0.173960800908)*x[1]**o+(0.730170280099)*x[1]
        s[1,4,0,0]=(-0.00733305213967)*x[0]**o+(0.473349687079)*x[0]+(0.433262136464)*x[1]**o+(-0.122843871456)*x[1]
        s[1,4,0,1]=(-0.609798017036)*x[0]**o+(-0.271729269891)*x[0]+(-0.0473090363303)*x[1]**o+(-0.429215577266)*x[1]
        s[1,4,1,0]=(0.860194281311)*x[0]**o+(0.267299506296)*x[0]+(0.564929961344)*x[1]**o+(0.4738297918)*x[1]
        s[1,4,1,1]=(0.170949641101)*x[0]**o+(0.710820092329)*x[0]+(0.934882329613)*x[1]**o+(0.607064104034)*x[1]
        s[1,4,2,0]=(-0.299596799604)*x[0]**o+(-0.533859449337)*x[0]+(0.751725833171)*x[1]**o+(-0.690767322166)*x[1]
        s[1,4,2,1]=(-0.624424753627)*x[0]**o+(-0.947238508194)*x[0]+(0.456597368873)*x[1]**o+(-0.800064510218)*x[1]
        s[2,0,0,0]=(0.566979423938)*x[0]**o+(0.924856128006)*x[0]+(0.894661507372)*x[1]**o+(0.737960904672)*x[1]
        s[2,0,0,1]=(0.424224691717)*x[0]**o+(-0.290057017691)*x[0]+(0.524889948009)*x[1]**o+(-0.968728724115)*x[1]
        s[2,0,1,0]=(0.116336639818)*x[0]**o+(0.192371600611)*x[0]+(-0.303824973684)*x[1]**o+(-0.32000509975)*x[1]
        s[2,0,1,1]=(-0.327127663029)*x[0]**o+(0.715544836917)*x[0]+(0.347278471442)*x[1]**o+(0.109909231847)*x[1]
        s[2,0,2,0]=(0.793922217119)*x[0]**o+(0.262726473054)*x[0]+(0.172201555322)*x[1]**o+(0.770334659224)*x[1]
        s[2,0,2,1]=(0.988217831671)*x[0]**o+(0.399655753685)*x[0]+(-0.143569502754)*x[1]**o+(0.0617091702351)*x[1]
        s[2,1,0,0]=(0.775224667659)*x[0]**o+(-0.0668878550232)*x[0]+(-0.424553437498)*x[1]**o+(-0.472878522461)*x[1]
        s[2,1,0,1]=(0.214811724455)*x[0]**o+(0.331016899879)*x[0]+(0.137063379737)*x[1]**o+(-0.0540696322974)*x[1]
        s[2,1,1,0]=(-0.463688617296)*x[0]**o+(-0.373502340044)*x[0]+(-0.0840959194547)*x[1]**o+(0.399767480048)*x[1]
        s[2,1,1,1]=(-0.245792577875)*x[0]**o+(-0.996843380732)*x[0]+(-0.249517835294)*x[1]**o+(-0.594721563199)*x[1]
        s[2,1,2,0]=(0.671238831228)*x[0]**o+(0.899659387933)*x[0]+(-0.964232607177)*x[1]**o+(0.637445200931)*x[1]
        s[2,1,2,1]=(0.746879641209)*x[0]**o+(-0.247785412499)*x[0]+(-0.370865590053)*x[1]**o+(-0.302746229959)*x[1]
        s[2,2,0,0]=(0.336257487071)*x[0]**o+(-0.179673308736)*x[0]+(0.912054571446)*x[1]**o+(-0.716018450571)*x[1]
        s[2,2,0,1]=(-0.686601572148)*x[0]**o+(0.120899496999)*x[0]+(-0.419150767247)*x[1]**o+(-0.418707334342)*x[1]
        s[2,2,1,0]=(-0.354036063032)*x[0]**o+(0.363634583771)*x[0]+(0.691849369366)*x[1]**o+(-0.887830896718)*x[1]
        s[2,2,1,1]=(0.711473627816)*x[0]**o+(0.556386580914)*x[0]+(0.888396281963)*x[1]**o+(0.386590176165)*x[1]
        s[2,2,2,0]=(0.419814106708)*x[0]**o+(-0.0352463077763)*x[0]+(0.600680097958)*x[1]**o+(-0.884870139538)*x[1]
        s[2,2,2,1]=(0.203572624179)*x[0]**o+(-0.915173748563)*x[0]+(0.0466882954353)*x[1]**o+(0.567223966741)*x[1]
        s[2,3,0,0]=(0.365889449332)*x[0]**o+(-0.0875793555978)*x[0]+(-0.514531990304)*x[1]**o+(0.641244565067)*x[1]
        s[2,3,0,1]=(0.801255486954)*x[0]**o+(0.324497137754)*x[0]+(-0.750291931827)*x[1]**o+(-0.972736614855)*x[1]
        s[2,3,1,0]=(0.540943455318)*x[0]**o+(-0.507084574203)*x[0]+(0.903787268326)*x[1]**o+(0.192067385499)*x[1]
        s[2,3,1,1]=(0.50229010977)*x[0]**o+(0.327771429044)*x[0]+(-0.871646777584)*x[1]**o+(-0.915312646394)*x[1]
        s[2,3,2,0]=(0.816437071516)*x[0]**o+(-0.324552837786)*x[0]+(0.32191911659)*x[1]**o+(0.448902757154)*x[1]
        s[2,3,2,1]=(-0.688163580944)*x[0]**o+(0.00185099386065)*x[0]+(0.629596314583)*x[1]**o+(0.898968857239)*x[1]
        s[2,4,0,0]=(0.558694638173)*x[0]**o+(0.879817228992)*x[0]+(-0.779489055632)*x[1]**o+(-0.597316482587)*x[1]
        s[2,4,0,1]=(0.405266767625)*x[0]**o+(-0.803650202792)*x[0]+(-0.209683708637)*x[1]**o+(0.843260589899)*x[1]
        s[2,4,1,0]=(0.228094700844)*x[0]**o+(0.28915480893)*x[0]+(0.404160989937)*x[1]**o+(-0.837717657152)*x[1]
        s[2,4,1,1]=(-0.628771431514)*x[0]**o+(0.70484286338)*x[0]+(0.162686726925)*x[1]**o+(0.743576454116)*x[1]
        s[2,4,2,0]=(0.139366390708)*x[0]**o+(0.452232237593)*x[0]+(-0.0536300454799)*x[1]**o+(-0.0220704990737)*x[1]
        s[2,4,2,1]=(0.407548912264)*x[0]**o+(-0.184892320636)*x[0]+(0.945920847222)*x[1]**o+(-0.326239004837)*x[1]
        s[3,0,0,0]=(-0.893632035291)*x[0]**o+(0.605568638642)*x[0]+(-0.605327037586)*x[1]**o+(-0.129824948634)*x[1]
        s[3,0,0,1]=(0.058430352705)*x[0]**o+(0.724766033805)*x[0]+(0.900105388199)*x[1]**o+(-0.248904105045)*x[1]
        s[3,0,1,0]=(-0.151387812319)*x[0]**o+(0.434769353572)*x[0]+(0.525149052261)*x[1]**o+(-0.949942908295)*x[1]
        s[3,0,1,1]=(0.419206109353)*x[0]**o+(0.828163036942)*x[0]+(-0.388930812616)*x[1]**o+(0.247705907297)*x[1]
        s[3,0,2,0]=(0.0520861249252)*x[0]**o+(0.545440655049)*x[0]+(0.123957894743)*x[1]**o+(0.726676820938)*x[1]
        s[3,0,2,1]=(-0.832443800699)*x[0]**o+(-0.873219166703)*x[0]+(-0.09459670039)*x[1]**o+(-0.490598426074)*x[1]
        s[3,1,0,0]=(-0.0396898083178)*x[0]**o+(0.13897357947)*x[0]+(-0.570943434112)*x[1]**o+(0.113736405636)*x[1]
        s[3,1,0,1]=(-0.753753572629)*x[0]**o+(0.0105397073947)*x[0]+(-0.100456113737)*x[1]**o+(0.702912637963)*x[1]
        s[3,1,1,0]=(-0.2396162336)*x[0]**o+(0.0957817653067)*x[0]+(-0.426329991998)*x[1]**o+(-0.806921457576)*x[1]
        s[3,1,1,1]=(-0.00676639677643)*x[0]**o+(0.644421172464)*x[0]+(0.893174135862)*x[1]**o+(0.921265996277)*x[1]
        s[3,1,2,0]=(-0.370431629052)*x[0]**o+(-0.431968328195)*x[0]+(0.322907749791)*x[1]**o+(-0.734928059791)*x[1]
        s[3,1,2,1]=(0.565520417451)*x[0]**o+(0.5972374649)*x[0]+(-0.384953451928)*x[1]**o+(0.708513444822)*x[1]
        s[3,2,0,0]=(0.546973382463)*x[0]**o+(-0.391535907448)*x[0]+(0.743984877534)*x[1]**o+(0.690664640674)*x[1]
        s[3,2,0,1]=(0.286366773252)*x[0]**o+(-0.536931967236)*x[0]+(0.914555659603)*x[1]**o+(-0.978176990473)*x[1]
        s[3,2,1,0]=(-0.272652614089)*x[0]**o+(0.0235157061398)*x[0]+(-0.147489880453)*x[1]**o+(-0.698449540889)*x[1]
        s[3,2,1,1]=(0.856000001512)*x[0]**o+(-0.49804787796)*x[0]+(-0.993780459188)*x[1]**o+(0.851338346493)*x[1]
        s[3,2,2,0]=(-0.57527355268)*x[0]**o+(-0.760820441576)*x[0]+(-0.654096561486)*x[1]**o+(-0.736649206713)*x[1]
        s[3,2,2,1]=(-0.262714050973)*x[0]**o+(-0.182627531212)*x[0]+(0.845508653482)*x[1]**o+(0.335330629033)*x[1]
        s[3,3,0,0]=(0.284468477631)*x[0]**o+(0.778849843484)*x[0]+(-0.210037009712)*x[1]**o+(0.261684901446)*x[1]
        s[3,3,0,1]=(-0.702010952741)*x[0]**o+(-0.147730819043)*x[0]+(-0.251392638434)*x[1]**o+(-0.450438064218)*x[1]
        s[3,3,1,0]=(0.37599915758)*x[0]**o+(-0.663740528508)*x[0]+(0.831132392336)*x[1]**o+(-0.686462967317)*x[1]
        s[3,3,1,1]=(0.104633103538)*x[0]**o+(0.152999779786)*x[0]+(-0.617511701255)*x[1]**o+(-0.6261759882)*x[1]
        s[3,3,2,0]=(0.996985674737)*x[0]**o+(-0.113914220536)*x[0]+(0.815302144446)*x[1]**o+(0.426458911854)*x[1]
        s[3,3,2,1]=(0.79490425189)*x[0]**o+(-0.352566895288)*x[0]+(0.595219934007)*x[1]**o+(0.931962305688)*x[1]
        s[3,4,0,0]=(0.0342109984348)*x[0]**o+(0.756935167027)*x[0]+(-0.910031669765)*x[1]**o+(-0.768302193908)*x[1]
        s[3,4,0,1]=(-0.765106211079)*x[0]**o+(0.971115633922)*x[0]+(-0.682150454882)*x[1]**o+(0.345375833573)*x[1]
        s[3,4,1,0]=(0.392841668036)*x[0]**o+(0.06254598514)*x[0]+(0.66808036785)*x[1]**o+(-0.76724599164)*x[1]
        s[3,4,1,1]=(-0.155018584015)*x[0]**o+(0.100771866989)*x[0]+(-0.800809131287)*x[1]**o+(-0.481511683108)*x[1]
        s[3,4,2,0]=(0.420069461837)*x[0]**o+(0.116833735731)*x[0]+(-0.334781127089)*x[1]**o+(-0.278898259371)*x[1]
        s[3,4,2,1]=(-0.695614335152)*x[0]**o+(-0.116909319428)*x[0]+(-0.136699698838)*x[1]**o+(-0.199293217231)*x[1]
        ref[0,0,0,0]=(-0.882344237411)*(1+2.*(dim-1.)/(o+1.))+(-0.172196984056)*dim
        ref[0,0,0,1]=(1.26914006297)*(1+2.*(dim-1.)/(o+1.))+(-1.40183301272)*dim
        ref[0,0,1,0]=(1.22833354853)*(1+2.*(dim-1.)/(o+1.))+(1.16478104873)*dim
        ref[0,0,1,1]=(1.49534234555)*(1+2.*(dim-1.)/(o+1.))+(-1.09638573805)*dim
        ref[0,0,2,0]=(-0.00539584116507)*(1+2.*(dim-1.)/(o+1.))+(-0.0252230228979)*dim
        ref[0,0,2,1]=(0.0513956321246)*(1+2.*(dim-1.)/(o+1.))+(0.115329465043)*dim
        ref[0,1,0,0]=(-0.225445610011)*(1+2.*(dim-1.)/(o+1.))+(-0.84411161714)*dim
        ref[0,1,0,1]=(-0.204873066942)*(1+2.*(dim-1.)/(o+1.))+(0.44661310627)*dim
        ref[0,1,1,0]=(0.0494465707545)*(1+2.*(dim-1.)/(o+1.))+(0.419846393451)*dim
        ref[0,1,1,1]=(-0.725589455239)*(1+2.*(dim-1.)/(o+1.))+(0.368216791426)*dim
        ref[0,1,2,0]=(1.04532039966)*(1+2.*(dim-1.)/(o+1.))+(0.492886815094)*dim
        ref[0,1,2,1]=(0.340473651741)*(1+2.*(dim-1.)/(o+1.))+(-0.511609927401)*dim
        ref[0,2,0,0]=(0.689004426423)*(1+2.*(dim-1.)/(o+1.))+(-0.128135312514)*dim
        ref[0,2,0,1]=(0.233838160978)*(1+2.*(dim-1.)/(o+1.))+(-1.26135483796)*dim
        ref[0,2,1,0]=(-1.1884792541)*(1+2.*(dim-1.)/(o+1.))+(0.0936640488261)*dim
        ref[0,2,1,1]=(-1.30194619908)*(1+2.*(dim-1.)/(o+1.))+(0.0438532683776)*dim
        ref[0,2,2,0]=(-0.795050191379)*(1+2.*(dim-1.)/(o+1.))+(0.75377047095)*dim
        ref[0,2,2,1]=(0.099795944148)*(1+2.*(dim-1.)/(o+1.))+(-0.198466000586)*dim
        ref[0,3,0,0]=(0.760012080694)*(1+2.*(dim-1.)/(o+1.))+(1.10871548763)*dim
        ref[0,3,0,1]=(0.395405473541)*(1+2.*(dim-1.)/(o+1.))+(-0.00948296128713)*dim
        ref[0,3,1,0]=(0.238853625168)*(1+2.*(dim-1.)/(o+1.))+(-0.597116999753)*dim
        ref[0,3,1,1]=(1.18242988677)*(1+2.*(dim-1.)/(o+1.))+(-0.673036229731)*dim
        ref[0,3,2,0]=(1.0107346985)*(1+2.*(dim-1.)/(o+1.))+(-0.103886403944)*dim
        ref[0,3,2,1]=(-1.72617954752)*(1+2.*(dim-1.)/(o+1.))+(0.928440367308)*dim
        ref[0,4,0,0]=(1.04818949627)*(1+2.*(dim-1.)/(o+1.))+(1.61558682127)*dim
        ref[0,4,0,1]=(0.819171090384)*(1+2.*(dim-1.)/(o+1.))+(-1.52348199116)*dim
        ref[0,4,1,0]=(-0.476528056537)*(1+2.*(dim-1.)/(o+1.))+(-0.301550199935)*dim
        ref[0,4,1,1]=(0.00968692452303)*(1+2.*(dim-1.)/(o+1.))+(-0.404433758176)*dim
        ref[0,4,2,0]=(-1.05784512977)*(1+2.*(dim-1.)/(o+1.))+(-0.347976376775)*dim
        ref[0,4,2,1]=(0.239748779202)*(1+2.*(dim-1.)/(o+1.))+(0.336091107784)*dim
        ref[1,0,0,0]=(0.226665452971)*(1+2.*(dim-1.)/(o+1.))+(-1.30611192181)*dim
        ref[1,0,0,1]=(-0.050205120649)*(1+2.*(dim-1.)/(o+1.))+(0.638104104515)*dim
        ref[1,0,1,0]=(-0.938444742978)*(1+2.*(dim-1.)/(o+1.))+(-0.289254674011)*dim
        ref[1,0,1,1]=(0.718782021562)*(1+2.*(dim-1.)/(o+1.))+(0.309935627295)*dim
        ref[1,0,2,0]=(1.34868159789)*(1+2.*(dim-1.)/(o+1.))+(0.354555085415)*dim
        ref[1,0,2,1]=(0.181861611561)*(1+2.*(dim-1.)/(o+1.))+(0.527305012005)*dim
        ref[1,1,0,0]=(1.29009930417)*(1+2.*(dim-1.)/(o+1.))+(0.0101704377708)*dim
        ref[1,1,0,1]=(0.461214267966)*(1+2.*(dim-1.)/(o+1.))+(1.33762473167)*dim
        ref[1,1,1,0]=(0.436776695414)*(1+2.*(dim-1.)/(o+1.))+(0.140430657448)*dim
        ref[1,1,1,1]=(0.707047922014)*(1+2.*(dim-1.)/(o+1.))+(0.415013853887)*dim
        ref[1,1,2,0]=(-0.123331222393)*(1+2.*(dim-1.)/(o+1.))+(0.500421126648)*dim
        ref[1,1,2,1]=(-0.585205414488)*(1+2.*(dim-1.)/(o+1.))+(-0.8345406952)*dim
        ref[1,2,0,0]=(-0.240237232779)*(1+2.*(dim-1.)/(o+1.))+(0.762258474441)*dim
        ref[1,2,0,1]=(-0.250122467769)*(1+2.*(dim-1.)/(o+1.))+(-0.0102691693039)*dim
        ref[1,2,1,0]=(0.833050085321)*(1+2.*(dim-1.)/(o+1.))+(-0.724988605075)*dim
        ref[1,2,1,1]=(-1.47168828548)*(1+2.*(dim-1.)/(o+1.))+(-0.625446172039)*dim
        ref[1,2,2,0]=(0.75410156753)*(1+2.*(dim-1.)/(o+1.))+(0.339836569004)*dim
        ref[1,2,2,1]=(-0.273160842)*(1+2.*(dim-1.)/(o+1.))+(-1.19278533956)*dim
        ref[1,3,0,0]=(1.16848716517)*(1+2.*(dim-1.)/(o+1.))+(0.214401078033)*dim
        ref[1,3,0,1]=(-0.277477097791)*(1+2.*(dim-1.)/(o+1.))+(1.17205470063)*dim
        ref[1,3,1,0]=(-0.977369562339)*(1+2.*(dim-1.)/(o+1.))+(0.745547217647)*dim
        ref[1,3,1,1]=(-0.614272379047)*(1+2.*(dim-1.)/(o+1.))+(-0.237223169949)*dim
        ref[1,3,2,0]=(0.289793862148)*(1+2.*(dim-1.)/(o+1.))+(0.973872897381)*dim
        ref[1,3,2,1]=(0.512457332251)*(1+2.*(dim-1.)/(o+1.))+(1.37648593824)*dim
        ref[1,4,0,0]=(0.425929084324)*(1+2.*(dim-1.)/(o+1.))+(0.350505815623)*dim
        ref[1,4,0,1]=(-0.657107053366)*(1+2.*(dim-1.)/(o+1.))+(-0.700944847157)*dim
        ref[1,4,1,0]=(1.42512424265)*(1+2.*(dim-1.)/(o+1.))+(0.741129298097)*dim
        ref[1,4,1,1]=(1.10583197071)*(1+2.*(dim-1.)/(o+1.))+(1.31788419636)*dim
        ref[1,4,2,0]=(0.452129033567)*(1+2.*(dim-1.)/(o+1.))+(-1.2246267715)*dim
        ref[1,4,2,1]=(-0.167827384754)*(1+2.*(dim-1.)/(o+1.))+(-1.74730301841)*dim
        ref[2,0,0,0]=(1.46164093131)*(1+2.*(dim-1.)/(o+1.))+(1.66281703268)*dim
        ref[2,0,0,1]=(0.949114639726)*(1+2.*(dim-1.)/(o+1.))+(-1.25878574181)*dim
        ref[2,0,1,0]=(-0.187488333866)*(1+2.*(dim-1.)/(o+1.))+(-0.127633499139)*dim
        ref[2,0,1,1]=(0.020150808413)*(1+2.*(dim-1.)/(o+1.))+(0.825454068764)*dim
        ref[2,0,2,0]=(0.96612377244)*(1+2.*(dim-1.)/(o+1.))+(1.03306113228)*dim
        ref[2,0,2,1]=(0.844648328917)*(1+2.*(dim-1.)/(o+1.))+(0.46136492392)*dim
        ref[2,1,0,0]=(0.350671230162)*(1+2.*(dim-1.)/(o+1.))+(-0.539766377484)*dim
        ref[2,1,0,1]=(0.351875104192)*(1+2.*(dim-1.)/(o+1.))+(0.276947267582)*dim
        ref[2,1,1,0]=(-0.54778453675)*(1+2.*(dim-1.)/(o+1.))+(0.0262651400042)*dim
        ref[2,1,1,1]=(-0.495310413169)*(1+2.*(dim-1.)/(o+1.))+(-1.59156494393)*dim
        ref[2,1,2,0]=(-0.292993775949)*(1+2.*(dim-1.)/(o+1.))+(1.53710458886)*dim
        ref[2,1,2,1]=(0.376014051155)*(1+2.*(dim-1.)/(o+1.))+(-0.550531642457)*dim
        ref[2,2,0,0]=(1.24831205852)*(1+2.*(dim-1.)/(o+1.))+(-0.895691759307)*dim
        ref[2,2,0,1]=(-1.1057523394)*(1+2.*(dim-1.)/(o+1.))+(-0.297807837342)*dim
        ref[2,2,1,0]=(0.337813306334)*(1+2.*(dim-1.)/(o+1.))+(-0.524196312948)*dim
        ref[2,2,1,1]=(1.59986990978)*(1+2.*(dim-1.)/(o+1.))+(0.942976757078)*dim
        ref[2,2,2,0]=(1.02049420467)*(1+2.*(dim-1.)/(o+1.))+(-0.920116447314)*dim
        ref[2,2,2,1]=(0.250260919614)*(1+2.*(dim-1.)/(o+1.))+(-0.347949781822)*dim
        ref[2,3,0,0]=(-0.148642540972)*(1+2.*(dim-1.)/(o+1.))+(0.553665209469)*dim
        ref[2,3,0,1]=(0.0509635551269)*(1+2.*(dim-1.)/(o+1.))+(-0.6482394771)*dim
        ref[2,3,1,0]=(1.44473072364)*(1+2.*(dim-1.)/(o+1.))+(-0.315017188703)*dim
        ref[2,3,1,1]=(-0.369356667814)*(1+2.*(dim-1.)/(o+1.))+(-0.58754121735)*dim
        ref[2,3,2,0]=(1.13835618811)*(1+2.*(dim-1.)/(o+1.))+(0.124349919368)*dim
        ref[2,3,2,1]=(-0.058567266361)*(1+2.*(dim-1.)/(o+1.))+(0.900819851099)*dim
        ref[2,4,0,0]=(-0.220794417459)*(1+2.*(dim-1.)/(o+1.))+(0.282500746405)*dim
        ref[2,4,0,1]=(0.195583058988)*(1+2.*(dim-1.)/(o+1.))+(0.0396103871071)*dim
        ref[2,4,1,0]=(0.63225569078)*(1+2.*(dim-1.)/(o+1.))+(-0.548562848222)*dim
        ref[2,4,1,1]=(-0.466084704588)*(1+2.*(dim-1.)/(o+1.))+(1.4484193175)*dim
        ref[2,4,2,0]=(0.0857363452282)*(1+2.*(dim-1.)/(o+1.))+(0.430161738519)*dim
        ref[2,4,2,1]=(1.35346975949)*(1+2.*(dim-1.)/(o+1.))+(-0.511131325473)*dim
        ref[3,0,0,0]=(-1.49895907288)*(1+2.*(dim-1.)/(o+1.))+(0.475743690008)*dim
        ref[3,0,0,1]=(0.958535740904)*(1+2.*(dim-1.)/(o+1.))+(0.47586192876)*dim
        ref[3,0,1,0]=(0.373761239942)*(1+2.*(dim-1.)/(o+1.))+(-0.515173554723)*dim
        ref[3,0,1,1]=(0.0302752967372)*(1+2.*(dim-1.)/(o+1.))+(1.07586894424)*dim
        ref[3,0,2,0]=(0.176044019669)*(1+2.*(dim-1.)/(o+1.))+(1.27211747599)*dim
        ref[3,0,2,1]=(-0.927040501089)*(1+2.*(dim-1.)/(o+1.))+(-1.36381759278)*dim
        ref[3,1,0,0]=(-0.61063324243)*(1+2.*(dim-1.)/(o+1.))+(0.252709985106)*dim
        ref[3,1,0,1]=(-0.854209686365)*(1+2.*(dim-1.)/(o+1.))+(0.713452345358)*dim
        ref[3,1,1,0]=(-0.665946225598)*(1+2.*(dim-1.)/(o+1.))+(-0.711139692269)*dim
        ref[3,1,1,1]=(0.886407739085)*(1+2.*(dim-1.)/(o+1.))+(1.56568716874)*dim
        ref[3,1,2,0]=(-0.0475238792609)*(1+2.*(dim-1.)/(o+1.))+(-1.16689638799)*dim
        ref[3,1,2,1]=(0.180566965523)*(1+2.*(dim-1.)/(o+1.))+(1.30575090972)*dim
        ref[3,2,0,0]=(1.29095826)*(1+2.*(dim-1.)/(o+1.))+(0.299128733226)*dim
        ref[3,2,0,1]=(1.20092243286)*(1+2.*(dim-1.)/(o+1.))+(-1.51510895771)*dim
        ref[3,2,1,0]=(-0.420142494542)*(1+2.*(dim-1.)/(o+1.))+(-0.674933834749)*dim
        ref[3,2,1,1]=(-0.137780457676)*(1+2.*(dim-1.)/(o+1.))+(0.353290468533)*dim
        ref[3,2,2,0]=(-1.22937011417)*(1+2.*(dim-1.)/(o+1.))+(-1.49746964829)*dim
        ref[3,2,2,1]=(0.582794602509)*(1+2.*(dim-1.)/(o+1.))+(0.152703097821)*dim
        ref[3,3,0,0]=(0.0744314679194)*(1+2.*(dim-1.)/(o+1.))+(1.04053474493)*dim
        ref[3,3,0,1]=(-0.953403591176)*(1+2.*(dim-1.)/(o+1.))+(-0.598168883261)*dim
        ref[3,3,1,0]=(1.20713154992)*(1+2.*(dim-1.)/(o+1.))+(-1.35020349583)*dim
        ref[3,3,1,1]=(-0.512878597717)*(1+2.*(dim-1.)/(o+1.))+(-0.473176208414)*dim
        ref[3,3,2,0]=(1.81228781918)*(1+2.*(dim-1.)/(o+1.))+(0.312544691319)*dim
        ref[3,3,2,1]=(1.3901241859)*(1+2.*(dim-1.)/(o+1.))+(0.5793954104)*dim
        ref[3,4,0,0]=(-0.87582067133)*(1+2.*(dim-1.)/(o+1.))+(-0.0113670268806)*dim
        ref[3,4,0,1]=(-1.44725666596)*(1+2.*(dim-1.)/(o+1.))+(1.31649146749)*dim
        ref[3,4,1,0]=(1.06092203589)*(1+2.*(dim-1.)/(o+1.))+(-0.7047000065)*dim
        ref[3,4,1,1]=(-0.955827715302)*(1+2.*(dim-1.)/(o+1.))+(-0.38073981612)*dim
        ref[3,4,2,0]=(0.0852883347481)*(1+2.*(dim-1.)/(o+1.))+(-0.16206452364)*dim
        ref[3,4,2,1]=(-0.83231403399)*(1+2.*(dim-1.)/(o+1.))+(-0.316202536659)*dim
      else:
        s[0,0,0,0]=(-0.394972833746)*x[0]**o+(-0.555133440221)*x[0]+(-0.268090782116)*x[1]**o+(-0.279086182399)*x[1]+(0.647538832619)*x[2]**o+(0.403928571033)*x[2]
        s[0,0,0,1]=(-0.35392621979)*x[0]**o+(-0.492907886034)*x[0]+(-0.285991202298)*x[1]**o+(0.652256699399)*x[1]+(0.746644949246)*x[2]**o+(-0.81069060154)*x[2]
        s[0,0,1,0]=(-0.373534422869)*x[0]**o+(0.725526776416)*x[0]+(-0.228168552487)*x[1]**o+(0.530464480729)*x[1]+(-0.257805238342)*x[2]**o+(0.482778418004)*x[2]
        s[0,0,1,1]=(-0.19411212431)*x[0]**o+(0.45313120948)*x[0]+(0.159756421366)*x[1]**o+(0.124033159775)*x[1]+(0.395678187632)*x[2]**o+(-0.145495322245)*x[2]
        s[0,0,2,0]=(0.420598692135)*x[0]**o+(0.222648941851)*x[0]+(-0.964380624775)*x[1]**o+(0.419039850926)*x[1]+(-0.908861665427)*x[2]**o+(-0.13977247814)*x[2]
        s[0,0,2,1]=(-0.339655604031)*x[0]**o+(-0.81821640554)*x[0]+(-0.864798652203)*x[1]**o+(-0.695538165934)*x[1]+(0.535554008883)*x[2]**o+(-0.142485273891)*x[2]
        s[0,1,0,0]=(-0.221315603083)*x[0]**o+(0.437569708371)*x[0]+(0.702877377655)*x[1]**o+(0.169385737092)*x[1]+(-0.92490534465)*x[2]**o+(0.264213443224)*x[2]
        s[0,1,0,1]=(0.61945266928)*x[0]**o+(0.699276978747)*x[0]+(0.199776861853)*x[1]**o+(-0.615345507396)*x[1]+(-0.70081403292)*x[2]**o+(0.664056523269)*x[2]
        s[0,1,1,0]=(-0.479722011354)*x[0]**o+(-0.591782822119)*x[0]+(0.919441326242)*x[1]**o+(-0.886693034352)*x[1]+(0.48123492158)*x[2]**o+(0.308135102102)*x[2]
        s[0,1,1,1]=(-0.522342575747)*x[0]**o+(0.662917347867)*x[0]+(0.359887593621)*x[1]**o+(-0.228498407508)*x[1]+(0.705995139546)*x[2]**o+(-0.406581496679)*x[2]
        s[0,1,2,0]=(0.555378475566)*x[0]**o+(-0.680672745096)*x[0]+(-0.269859012441)*x[1]**o+(0.800892438723)*x[1]+(-0.455863363072)*x[2]**o+(-0.991391540158)*x[2]
        s[0,1,2,1]=(0.547403511029)*x[0]**o+(-0.154514192835)*x[0]+(0.177483676876)*x[1]**o+(0.887044119995)*x[1]+(-0.21971493108)*x[2]**o+(-0.262092872636)*x[2]
        s[0,2,0,0]=(0.976262232729)*x[0]**o+(0.548252161786)*x[0]+(-0.743334308912)*x[1]**o+(0.207627841895)*x[1]+(0.853225044125)*x[2]**o+(0.307558901346)*x[2]
        s[0,2,0,1]=(0.503102495469)*x[0]**o+(0.407241098221)*x[0]+(0.52413664456)*x[1]**o+(-0.00773056229007)*x[1]+(-0.292242153733)*x[2]**o+(0.17146852389)*x[2]
        s[0,2,1,0]=(0.13389613662)*x[0]**o+(-0.87587702201)*x[0]+(-0.678875907381)*x[1]**o+(0.543117132397)*x[1]+(0.393096793861)*x[2]**o+(0.959258318284)*x[2]
        s[0,2,1,1]=(-0.0940648527232)*x[0]**o+(-0.598198827088)*x[0]+(0.424517028995)*x[1]**o+(-0.345900941234)*x[1]+(0.758648965957)*x[2]**o+(0.25975303965)*x[2]
        s[0,2,2,0]=(-0.870274070711)*x[0]**o+(-0.78246615425)*x[0]+(0.392295964553)*x[1]**o+(0.844507475338)*x[1]+(0.528851559277)*x[2]**o+(-0.995729427392)*x[2]
        s[0,2,2,1]=(-0.0730197692456)*x[0]**o+(-0.312551702743)*x[0]+(0.623722578367)*x[1]**o+(-0.848292098624)*x[1]+(-0.779643252401)*x[2]**o+(-0.338597223715)*x[2]
        s[0,3,0,0]=(0.872506198169)*x[0]**o+(-0.500297419082)*x[0]+(0.269776051595)*x[1]**o+(0.566518089969)*x[1]+(-0.498369039824)*x[2]**o+(0.607886737439)*x[2]
        s[0,3,0,1]=(0.883053802312)*x[0]**o+(0.833130186655)*x[0]+(-0.921394587681)*x[1]**o+(0.0175686356791)*x[1]+(0.951456010531)*x[2]**o+(-0.571752807102)*x[2]
        s[0,3,1,0]=(-0.712330362129)*x[0]**o+(-0.676885540903)*x[0]+(-0.509633815107)*x[1]**o+(0.999769611236)*x[1]+(-0.186240684121)*x[2]**o+(-0.386806354224)*x[2]
        s[0,3,1,1]=(0.997518452801)*x[0]**o+(-0.269198998158)*x[0]+(-0.775427684973)*x[1]**o+(0.787580393259)*x[1]+(0.195858442466)*x[2]**o+(-0.734837176349)*x[2]
        s[0,3,2,0]=(0.888165993988)*x[0]**o+(-0.611433401469)*x[0]+(0.840713835877)*x[1]**o+(-0.805947825209)*x[1]+(-0.223790913101)*x[2]**o+(0.274074627813)*x[2]
        s[0,3,2,1]=(-0.0586612521858)*x[0]**o+(-0.904292259669)*x[0]+(0.701563259971)*x[1]**o+(-0.919454807365)*x[1]+(-0.432367433099)*x[2]**o+(-0.915055158557)*x[2]
        s[0,4,0,0]=(-0.896479628968)*x[0]**o+(-0.0775794895972)*x[0]+(0.952385683546)*x[1]**o+(0.101370020306)*x[1]+(-0.311916160508)*x[2]**o+(0.327792852845)*x[2]
        s[0,4,0,1]=(-0.368307750793)*x[0]**o+(-0.824637391638)*x[0]+(0.143668778522)*x[1]**o+(0.224471763433)*x[1]+(0.470345995942)*x[2]**o+(-0.238496191133)*x[2]
        s[0,4,1,0]=(0.978800093563)*x[0]**o+(0.356297491637)*x[0]+(-0.846067544583)*x[1]**o+(0.0112192363669)*x[1]+(-0.456308213004)*x[2]**o+(0.328816226899)*x[2]
        s[0,4,1,1]=(0.860778047982)*x[0]**o+(-0.85484855319)*x[0]+(0.431318104432)*x[1]**o+(0.0519725606118)*x[1]+(-0.37452257816)*x[2]**o+(0.387786048315)*x[2]
        s[0,4,2,0]=(-0.809652418528)*x[0]**o+(0.622059542769)*x[0]+(0.0769858780015)*x[1]**o+(-0.1118510671)*x[1]+(0.462693639527)*x[2]**o+(-0.321289200501)*x[2]
        s[0,4,2,1]=(-0.112319716285)*x[0]**o+(-0.97958472439)*x[0]+(0.471152868527)*x[1]**o+(0.979477042696)*x[1]+(0.343437406244)*x[2]**o+(0.330245074789)*x[2]
        s[1,0,0,0]=(-0.664959353312)*x[0]**o+(0.667428688064)*x[0]+(0.379323454697)*x[1]**o+(0.458212123636)*x[1]+(-0.583770985715)*x[2]**o+(0.926174954355)*x[2]
        s[1,0,0,1]=(0.463653553012)*x[0]**o+(-0.998470865927)*x[0]+(-0.0590876546604)*x[1]**o+(0.22211328071)*x[1]+(-0.741158466684)*x[2]**o+(-0.405853581466)*x[2]
        s[1,0,1,0]=(0.261926286639)*x[0]**o+(-0.291967639861)*x[0]+(-0.22306024334)*x[1]**o+(-0.266187302656)*x[1]+(0.0808303795362)*x[2]**o+(-0.0411340553637)*x[2]
        s[1,0,1,1]=(0.823792286605)*x[0]**o+(-0.622764469917)*x[0]+(0.532406278339)*x[1]**o+(-0.799863805036)*x[1]+(0.356182760763)*x[2]**o+(0.633305652052)*x[2]
        s[1,0,2,0]=(0.386490056692)*x[0]**o+(-0.538422996672)*x[0]+(-0.400192004838)*x[1]**o+(0.632203703462)*x[1]+(0.315043759293)*x[2]**o+(-0.10635477353)*x[2]
        s[1,0,2,1]=(0.911736934682)*x[0]**o+(0.874267947717)*x[0]+(0.716599906083)*x[1]**o+(-0.477115250943)*x[1]+(0.296312692731)*x[2]**o+(0.691409352221)*x[2]
        s[1,1,0,0]=(0.193258443163)*x[0]**o+(-0.479351523999)*x[0]+(0.426152720777)*x[1]**o+(0.577292380929)*x[1]+(0.781776817395)*x[2]**o+(-0.0622489459224)*x[2]
        s[1,1,0,1]=(-0.491606711738)*x[0]**o+(-0.138919126056)*x[0]+(0.10073957381)*x[1]**o+(0.560595873872)*x[1]+(0.661805959053)*x[2]**o+(0.999249400576)*x[2]
        s[1,1,1,0]=(-0.197096694248)*x[0]**o+(0.205033675186)*x[0]+(0.265896904632)*x[1]**o+(0.448157190065)*x[1]+(0.284574920599)*x[2]**o+(-0.466032134334)*x[2]
        s[1,1,1,1]=(0.852594063486)*x[0]**o+(-0.139990914704)*x[0]+(0.0792640568063)*x[1]**o+(0.279425947587)*x[1]+(-0.598534354914)*x[2]**o+(-0.423199533073)*x[2]
        s[1,1,2,0]=(0.576687955944)*x[0]**o+(-0.616599888356)*x[0]+(-0.826670910311)*x[1]**o+(0.54433728322)*x[1]+(0.00123761537904)*x[2]**o+(0.716238991968)*x[2]
        s[1,1,2,1]=(0.56855109067)*x[0]**o+(-0.0440632979111)*x[0]+(0.526871640009)*x[1]**o+(0.275141335835)*x[1]+(-0.983229310103)*x[2]**o+(0.700105733507)*x[2]
        s[1,2,0,0]=(-0.184742995569)*x[0]**o+(0.71036642023)*x[0]+(0.981161346217)*x[1]**o+(0.365488167653)*x[1]+(-0.398883902177)*x[2]**o+(-0.482354856056)*x[2]
        s[1,2,0,1]=(0.583421736253)*x[0]**o+(-0.279163572851)*x[0]+(0.503764541147)*x[1]**o+(0.292381662704)*x[1]+(-0.888007033365)*x[2]**o+(-0.448668581094)*x[2]
        s[1,2,1,0]=(0.806086468959)*x[0]**o+(-0.959525959302)*x[0]+(0.734809833547)*x[1]**o+(-0.284558836843)*x[1]+(-0.354048559091)*x[2]**o+(-0.464063301273)*x[2]
        s[1,2,1,1]=(0.20737358797)*x[0]**o+(-0.958445091026)*x[0]+(-0.207486744384)*x[1]**o+(-0.300655801991)*x[1]+(0.534182391535)*x[2]**o+(-0.630595215476)*x[2]
        s[1,2,2,0]=(0.573970416565)*x[0]**o+(0.0868191347586)*x[0]+(0.90996767074)*x[1]**o+(-0.556896647781)*x[1]+(-0.57168584006)*x[2]**o+(-0.000855122180503)*x[2]
        s[1,2,2,1]=(-0.0500416574964)*x[0]**o+(0.134962389253)*x[0]+(0.7037800261)*x[1]**o+(-0.46634099981)*x[1]+(-0.978823012363)*x[2]**o+(0.429717517544)*x[2]
        s[1,3,0,0]=(-0.480934294581)*x[0]**o+(-0.761664422527)*x[0]+(0.563388700943)*x[1]**o+(0.286474912293)*x[1]+(-0.254087945936)*x[2]**o+(0.823788425201)*x[2]
        s[1,3,0,1]=(0.287760411402)*x[0]**o+(-0.667747306058)*x[0]+(-0.801505580774)*x[1]**o+(0.642958552096)*x[1]+(-0.443695399886)*x[2]**o+(-0.840149579234)*x[2]
        s[1,3,1,0]=(-0.789394564521)*x[0]**o+(0.918638482871)*x[0]+(0.584254838389)*x[1]**o+(-0.150259802986)*x[1]+(0.794702325874)*x[2]**o+(0.0246788014415)*x[2]
        s[1,3,1,1]=(-0.1183405979)*x[0]**o+(0.0451549250255)*x[0]+(0.1517987763)*x[1]**o+(-0.200979231459)*x[1]+(-0.0817034681199)*x[2]**o+(0.664693805987)*x[2]
        s[1,3,2,0]=(-0.364619176305)*x[0]**o+(0.989060492595)*x[0]+(-0.452469848404)*x[1]**o+(-0.622993726081)*x[1]+(0.738661997825)*x[2]**o+(0.89325891611)*x[2]
        s[1,3,2,1]=(0.794435451296)*x[0]**o+(0.561632824215)*x[0]+(-0.661460086901)*x[1]**o+(0.110534045458)*x[1]+(-0.11627022197)*x[2]**o+(-0.737456873271)*x[2]
        s[1,4,0,0]=(0.821875192234)*x[0]**o+(0.378653978743)*x[0]+(0.384830574445)*x[1]**o+(0.815762479042)*x[1]+(0.709307414564)*x[2]**o+(0.785945406666)*x[2]
        s[1,4,0,1]=(-0.195473903829)*x[0]**o+(-0.792429456326)*x[0]+(0.475482962939)*x[1]**o+(0.02846998342)*x[1]+(0.433471515314)*x[2]**o+(0.0565560487548)*x[2]
        s[1,4,1,0]=(-0.927049196862)*x[0]**o+(-0.770563924424)*x[0]+(0.722391079777)*x[1]**o+(-0.962242600193)*x[1]+(-0.652853937129)*x[2]**o+(-0.583745627826)*x[2]
        s[1,4,1,1]=(-0.955616614181)*x[0]**o+(0.194945562768)*x[0]+(-0.517066615715)*x[1]**o+(-0.712362265862)*x[1]+(-0.524911207386)*x[2]**o+(-0.882591220196)*x[2]
        s[1,4,2,0]=(0.524532786764)*x[0]**o+(0.284840195202)*x[0]+(0.963855761362)*x[1]**o+(-0.856530057166)*x[1]+(-0.0740070151584)*x[2]**o+(0.226341867159)*x[2]
        s[1,4,2,1]=(0.446800509767)*x[0]**o+(-0.864046465502)*x[0]+(-0.517438631247)*x[1]**o+(0.901374962381)*x[1]+(0.601052625973)*x[2]**o+(-0.863564939303)*x[2]
        s[2,0,0,0]=(-0.701781199314)*x[0]**o+(0.264327906644)*x[0]+(0.971572320137)*x[1]**o+(-0.168607506193)*x[1]+(-0.526717738449)*x[2]**o+(0.226927089015)*x[2]
        s[2,0,0,1]=(-0.847412912499)*x[0]**o+(-0.401413370017)*x[0]+(0.759311464889)*x[1]**o+(0.457894962279)*x[1]+(-0.55434272035)*x[2]**o+(0.120326915639)*x[2]
        s[2,0,1,0]=(-0.517103294855)*x[0]**o+(0.492756717159)*x[0]+(0.973627035185)*x[1]**o+(-0.860482824403)*x[1]+(0.54057457411)*x[2]**o+(-0.874495812872)*x[2]
        s[2,0,1,1]=(-0.932588871815)*x[0]**o+(-0.741303214077)*x[0]+(0.468667081235)*x[1]**o+(0.395164667731)*x[1]+(0.412999483627)*x[2]**o+(0.827200271655)*x[2]
        s[2,0,2,0]=(0.324804049142)*x[0]**o+(0.687197584348)*x[0]+(0.48017530302)*x[1]**o+(0.0594198080606)*x[1]+(-0.243274785097)*x[2]**o+(0.561027778867)*x[2]
        s[2,0,2,1]=(-0.830970533632)*x[0]**o+(-0.399321288145)*x[0]+(-0.715545820929)*x[1]**o+(-0.290047681469)*x[1]+(-0.335483597978)*x[2]**o+(0.666741330684)*x[2]
        s[2,1,0,0]=(-0.734079757702)*x[0]**o+(0.290309222969)*x[0]+(0.764118958044)*x[1]**o+(0.4132335859)*x[1]+(0.257031297672)*x[2]**o+(0.244389627708)*x[2]
        s[2,1,0,1]=(0.687268551061)*x[0]**o+(-0.24563836651)*x[0]+(-0.123765123048)*x[1]**o+(-0.530923035235)*x[1]+(-0.720827493122)*x[2]**o+(0.210975066905)*x[2]
        s[2,1,1,0]=(0.813659604982)*x[0]**o+(0.893957744697)*x[0]+(0.991947458807)*x[1]**o+(0.793408082473)*x[1]+(0.609236290624)*x[2]**o+(-0.0658228268111)*x[2]
        s[2,1,1,1]=(-0.118289813545)*x[0]**o+(0.622749254009)*x[0]+(0.465005967301)*x[1]**o+(-0.878906002448)*x[1]+(0.168374644259)*x[2]**o+(0.149889108873)*x[2]
        s[2,1,2,0]=(-0.255533627884)*x[0]**o+(-0.436918772486)*x[0]+(0.29166400203)*x[1]**o+(0.928052058386)*x[1]+(0.303846927853)*x[2]**o+(0.158922482472)*x[2]
        s[2,1,2,1]=(-0.0504039593843)*x[0]**o+(0.170081032189)*x[0]+(-0.739948814307)*x[1]**o+(-0.52541309551)*x[1]+(0.797681120077)*x[2]**o+(0.247665713856)*x[2]
        s[2,2,0,0]=(-0.400562988004)*x[0]**o+(-0.128489671375)*x[0]+(-0.451867300451)*x[1]**o+(-0.343423869937)*x[1]+(0.663184069698)*x[2]**o+(0.926436292949)*x[2]
        s[2,2,0,1]=(-0.599719778431)*x[0]**o+(0.108577031922)*x[0]+(0.217879834507)*x[1]**o+(0.907899596451)*x[1]+(-0.161562448509)*x[2]**o+(-0.514018490585)*x[2]
        s[2,2,1,0]=(-0.736520696776)*x[0]**o+(0.166243034841)*x[0]+(0.438311085394)*x[1]**o+(0.802937292983)*x[1]+(-0.938817901098)*x[2]**o+(-0.205869039453)*x[2]
        s[2,2,1,1]=(0.555108741686)*x[0]**o+(-0.533145284484)*x[0]+(-0.763891188865)*x[1]**o+(0.515585725441)*x[1]+(0.638581800777)*x[2]**o+(0.629321554502)*x[2]
        s[2,2,2,0]=(-0.790093516383)*x[0]**o+(-0.0154620341689)*x[0]+(0.462011424745)*x[1]**o+(0.598267220586)*x[1]+(0.576054554929)*x[2]**o+(0.482972076666)*x[2]
        s[2,2,2,1]=(0.075201635129)*x[0]**o+(-0.660059195587)*x[0]+(0.00711502517884)*x[1]**o+(-0.391407272114)*x[1]+(-0.213729719317)*x[2]**o+(-0.561573742069)*x[2]
        s[2,3,0,0]=(0.350797999062)*x[0]**o+(0.250500699573)*x[0]+(0.818402537945)*x[1]**o+(-0.899093594018)*x[1]+(-0.162195807738)*x[2]**o+(-0.147586906331)*x[2]
        s[2,3,0,1]=(0.695618803281)*x[0]**o+(0.227239221026)*x[0]+(-0.968505693287)*x[1]**o+(-0.473260919549)*x[1]+(-0.746882331556)*x[2]**o+(0.486703205212)*x[2]
        s[2,3,1,0]=(-0.271378189855)*x[0]**o+(-0.561268063707)*x[0]+(-0.0252286382422)*x[1]**o+(0.229921692249)*x[1]+(0.989187908151)*x[2]**o+(0.643686783302)*x[2]
        s[2,3,1,1]=(0.750682282794)*x[0]**o+(-0.27757575698)*x[0]+(0.604377636669)*x[1]**o+(-0.733927934644)*x[1]+(-0.718229138068)*x[2]**o+(-0.435426875038)*x[2]
        s[2,3,2,0]=(0.0733516934174)*x[0]**o+(-0.174475004056)*x[0]+(0.478942713703)*x[1]**o+(0.662549744231)*x[1]+(-0.565016805898)*x[2]**o+(0.17793235315)*x[2]
        s[2,3,2,1]=(-0.15216608314)*x[0]**o+(0.999325854219)*x[0]+(0.496048192253)*x[1]**o+(-0.674178160594)*x[1]+(-0.927955584861)*x[2]**o+(0.156845563256)*x[2]
        s[2,4,0,0]=(-0.912208708086)*x[0]**o+(0.929579552168)*x[0]+(-0.0445821051156)*x[1]**o+(0.911346880333)*x[1]+(0.309997148008)*x[2]**o+(0.0975168137997)*x[2]
        s[2,4,0,1]=(0.65092333822)*x[0]**o+(0.605218122961)*x[0]+(-0.684013078124)*x[1]**o+(0.088688492152)*x[1]+(-0.0554688873336)*x[2]**o+(-0.292971890481)*x[2]
        s[2,4,1,0]=(-0.437643682456)*x[0]**o+(-0.161540806934)*x[0]+(-0.863085496236)*x[1]**o+(0.714210296119)*x[1]+(0.972219783929)*x[2]**o+(-0.0450752608276)*x[2]
        s[2,4,1,1]=(0.838644689829)*x[0]**o+(0.876692248024)*x[0]+(-0.787815110065)*x[1]**o+(-0.867528859191)*x[1]+(-0.393677502867)*x[2]**o+(-0.522478124784)*x[2]
        s[2,4,2,0]=(-0.606364594289)*x[0]**o+(0.88715735186)*x[0]+(-0.894189269194)*x[1]**o+(0.470967389174)*x[1]+(0.427781385837)*x[2]**o+(0.141054076833)*x[2]
        s[2,4,2,1]=(-0.0329322907424)*x[0]**o+(0.0437971796726)*x[0]+(-0.706648559646)*x[1]**o+(0.132643884341)*x[1]+(0.101209934477)*x[2]**o+(-0.556234167371)*x[2]
        s[3,0,0,0]=(0.25287903621)*x[0]**o+(-0.281221279197)*x[0]+(0.643850500988)*x[1]**o+(0.0269082516545)*x[1]+(0.157104400545)*x[2]**o+(-0.554382511827)*x[2]
        s[3,0,0,1]=(0.130965460225)*x[0]**o+(0.203133851634)*x[0]+(-0.876140347686)*x[1]**o+(0.871803270423)*x[1]+(-0.372911987943)*x[2]**o+(-0.296478871771)*x[2]
        s[3,0,1,0]=(0.395462973708)*x[0]**o+(-0.0463119371121)*x[0]+(-0.810272498493)*x[1]**o+(0.052842476247)*x[1]+(-0.632091659537)*x[2]**o+(0.0477272308506)*x[2]
        s[3,0,1,1]=(-0.15375761581)*x[0]**o+(0.840485309457)*x[0]+(-0.183510151479)*x[1]**o+(0.423252205076)*x[1]+(0.0972125869507)*x[2]**o+(-0.873325211398)*x[2]
        s[3,0,2,0]=(-0.230628763369)*x[0]**o+(-0.446642785223)*x[0]+(0.172403317403)*x[1]**o+(0.6666203719)*x[1]+(0.758132421761)*x[2]**o+(-0.0802337093298)*x[2]
        s[3,0,2,1]=(-0.375928720447)*x[0]**o+(0.390180617554)*x[0]+(0.532916128738)*x[1]**o+(-0.438049708768)*x[1]+(-0.969242415622)*x[2]**o+(-0.979491904432)*x[2]
        s[3,1,0,0]=(0.243017072038)*x[0]**o+(-0.314649526706)*x[0]+(-0.64964134438)*x[1]**o+(-0.319482380304)*x[1]+(0.969793978414)*x[2]**o+(0.235845670735)*x[2]
        s[3,1,0,1]=(0.5372323492)*x[0]**o+(-0.297603522377)*x[0]+(0.718422772931)*x[1]**o+(0.591124970588)*x[1]+(0.148470013869)*x[2]**o+(-0.906959786864)*x[2]
        s[3,1,1,0]=(-0.41004677863)*x[0]**o+(0.519928695266)*x[0]+(-0.41056500013)*x[1]**o+(0.545730013165)*x[1]+(0.0316754455631)*x[2]**o+(0.0735333716599)*x[2]
        s[3,1,1,1]=(-0.413084334125)*x[0]**o+(-0.188907203544)*x[0]+(0.333210666705)*x[1]**o+(0.795876759492)*x[1]+(-0.879028457556)*x[2]**o+(0.597499547791)*x[2]
        s[3,1,2,0]=(0.572164028218)*x[0]**o+(0.561153490096)*x[0]+(-0.277218859151)*x[1]**o+(-0.784951308131)*x[1]+(-0.54076329511)*x[2]**o+(0.232589967523)*x[2]
        s[3,1,2,1]=(0.183000187474)*x[0]**o+(-0.0315492936828)*x[0]+(0.145425750124)*x[1]**o+(0.570531080762)*x[1]+(0.321679325968)*x[2]**o+(0.768693312757)*x[2]
        s[3,2,0,0]=(0.0572073598209)*x[0]**o+(-0.473877874718)*x[0]+(0.19030371495)*x[1]**o+(0.950458284849)*x[1]+(0.520875156632)*x[2]**o+(0.746025413415)*x[2]
        s[3,2,0,1]=(-0.670180903965)*x[0]**o+(0.792714072874)*x[0]+(-0.830626382455)*x[1]**o+(-0.694487379592)*x[1]+(0.708769579296)*x[2]**o+(0.874781581547)*x[2]
        s[3,2,1,0]=(-0.635101696551)*x[0]**o+(0.683413284169)*x[0]+(-0.580556747733)*x[1]**o+(0.228587157705)*x[1]+(0.283498698937)*x[2]**o+(-0.661341563969)*x[2]
        s[3,2,1,1]=(-0.938857602123)*x[0]**o+(-0.855150853757)*x[0]+(0.313962358596)*x[1]**o+(-0.137683481729)*x[1]+(-0.543347522619)*x[2]**o+(0.520234742546)*x[2]
        s[3,2,2,0]=(0.00237595749469)*x[0]**o+(0.151777614372)*x[0]+(-0.207526210221)*x[1]**o+(-0.593427746692)*x[1]+(-0.273081745725)*x[2]**o+(-0.103365939558)*x[2]
        s[3,2,2,1]=(-0.836487016235)*x[0]**o+(-0.499416194189)*x[0]+(0.944785244284)*x[1]**o+(0.316171200523)*x[1]+(0.882197908567)*x[2]**o+(-0.756125841104)*x[2]
        s[3,3,0,0]=(-0.331819442381)*x[0]**o+(0.0868286143286)*x[0]+(-0.0214955303831)*x[1]**o+(-0.802588129882)*x[1]+(-0.79665760605)*x[2]**o+(0.733766062704)*x[2]
        s[3,3,0,1]=(0.0832038861206)*x[0]**o+(-0.229342320163)*x[0]+(0.411514025627)*x[1]**o+(-0.278316368324)*x[1]+(0.158512271865)*x[2]**o+(-0.176346585884)*x[2]
        s[3,3,1,0]=(0.813577309153)*x[0]**o+(0.183181445534)*x[0]+(-0.625433189537)*x[1]**o+(-0.296559044989)*x[1]+(0.735593889495)*x[2]**o+(-0.84760854319)*x[2]
        s[3,3,1,1]=(0.129952300256)*x[0]**o+(-0.900793812246)*x[0]+(0.723956240897)*x[1]**o+(0.906884554981)*x[1]+(-0.828597157505)*x[2]**o+(-0.0453629922512)*x[2]
        s[3,3,2,0]=(0.618506925918)*x[0]**o+(0.537560191905)*x[0]+(0.274041602538)*x[1]**o+(0.128748880801)*x[1]+(-0.0421945967926)*x[2]**o+(-0.402512195673)*x[2]
        s[3,3,2,1]=(-0.882194818723)*x[0]**o+(-0.554059181227)*x[0]+(-0.519424591535)*x[1]**o+(0.0254864032182)*x[1]+(-0.163939472982)*x[2]**o+(-0.0494335206145)*x[2]
        s[3,4,0,0]=(-0.583301433479)*x[0]**o+(0.155758796662)*x[0]+(-0.814066320026)*x[1]**o+(0.152299298845)*x[1]+(0.988818040849)*x[2]**o+(-0.538805205144)*x[2]
        s[3,4,0,1]=(0.83112196296)*x[0]**o+(-0.719967217592)*x[0]+(-0.283369672135)*x[1]**o+(-0.995074593345)*x[1]+(0.129890006169)*x[2]**o+(-0.565650038645)*x[2]
        s[3,4,1,0]=(0.290242298945)*x[0]**o+(0.106236639888)*x[0]+(0.881353634263)*x[1]**o+(0.578003018941)*x[1]+(0.337748119883)*x[2]**o+(0.0495137765079)*x[2]
        s[3,4,1,1]=(-0.392371956935)*x[0]**o+(-0.403685094924)*x[0]+(0.953131101032)*x[1]**o+(0.144162998481)*x[1]+(0.13946212185)*x[2]**o+(-0.951124223008)*x[2]
        s[3,4,2,0]=(0.997965631913)*x[0]**o+(-0.494892241571)*x[0]+(0.887313590406)*x[1]**o+(-0.947323203129)*x[1]+(0.150412298394)*x[2]**o+(-0.0541618066578)*x[2]
        s[3,4,2,1]=(-0.177392402935)*x[0]**o+(0.338956222402)*x[0]+(0.746432610869)*x[1]**o+(-0.4930192355)*x[1]+(-0.210191759258)*x[2]**o+(0.501879127576)*x[2]
        ref[0,0,0,0]=(-0.0155247832423)*(1+2.*(dim-1.)/(o+1.))+(-0.430291051587)*dim
        ref[0,0,0,1]=(0.106727527157)*(1+2.*(dim-1.)/(o+1.))+(-0.651341788175)*dim
        ref[0,0,1,0]=(-0.859508213698)*(1+2.*(dim-1.)/(o+1.))+(1.73876967515)*dim
        ref[0,0,1,1]=(0.361322484688)*(1+2.*(dim-1.)/(o+1.))+(0.43166904701)*dim
        ref[0,0,2,0]=(-1.45264359807)*(1+2.*(dim-1.)/(o+1.))+(0.501916314636)*dim
        ref[0,0,2,1]=(-0.668900247351)*(1+2.*(dim-1.)/(o+1.))+(-1.65623984537)*dim
        ref[0,1,0,0]=(-0.443343570079)*(1+2.*(dim-1.)/(o+1.))+(0.871168888686)*dim
        ref[0,1,0,1]=(0.118415498213)*(1+2.*(dim-1.)/(o+1.))+(0.74798799462)*dim
        ref[0,1,1,0]=(0.920954236468)*(1+2.*(dim-1.)/(o+1.))+(-1.17034075437)*dim
        ref[0,1,1,1]=(0.54354015742)*(1+2.*(dim-1.)/(o+1.))+(0.0278374436795)*dim
        ref[0,1,2,0]=(-0.170343899948)*(1+2.*(dim-1.)/(o+1.))+(-0.871171846531)*dim
        ref[0,1,2,1]=(0.505172256825)*(1+2.*(dim-1.)/(o+1.))+(0.470437054524)*dim
        ref[0,2,0,0]=(1.08615296794)*(1+2.*(dim-1.)/(o+1.))+(1.06343890503)*dim
        ref[0,2,0,1]=(0.734996986295)*(1+2.*(dim-1.)/(o+1.))+(0.570979059821)*dim
        ref[0,2,1,0]=(-0.1518829769)*(1+2.*(dim-1.)/(o+1.))+(0.62649842867)*dim
        ref[0,2,1,1]=(1.08910114223)*(1+2.*(dim-1.)/(o+1.))+(-0.684346728672)*dim
        ref[0,2,2,0]=(0.0508734531183)*(1+2.*(dim-1.)/(o+1.))+(-0.933688106304)*dim
        ref[0,2,2,1]=(-0.228940443279)*(1+2.*(dim-1.)/(o+1.))+(-1.49944102508)*dim
        ref[0,3,0,0]=(0.64391320994)*(1+2.*(dim-1.)/(o+1.))+(0.674107408326)*dim
        ref[0,3,0,1]=(0.913115225162)*(1+2.*(dim-1.)/(o+1.))+(0.278946015231)*dim
        ref[0,3,1,0]=(-1.40820486136)*(1+2.*(dim-1.)/(o+1.))+(-0.0639222838911)*dim
        ref[0,3,1,1]=(0.417949210295)*(1+2.*(dim-1.)/(o+1.))+(-0.216455781248)*dim
        ref[0,3,2,0]=(1.50508891676)*(1+2.*(dim-1.)/(o+1.))+(-1.14330659887)*dim
        ref[0,3,2,1]=(0.210534574687)*(1+2.*(dim-1.)/(o+1.))+(-2.73880222559)*dim
        ref[0,4,0,0]=(-0.25601010593)*(1+2.*(dim-1.)/(o+1.))+(0.351583383554)*dim
        ref[0,4,0,1]=(0.245707023671)*(1+2.*(dim-1.)/(o+1.))+(-0.838661819338)*dim
        ref[0,4,1,0]=(-0.323575664024)*(1+2.*(dim-1.)/(o+1.))+(0.696332954903)*dim
        ref[0,4,1,1]=(0.917573574254)*(1+2.*(dim-1.)/(o+1.))+(-0.415089944263)*dim
        ref[0,4,2,0]=(-0.269972900999)*(1+2.*(dim-1.)/(o+1.))+(0.188919275167)*dim
        ref[0,4,2,1]=(0.702270558486)*(1+2.*(dim-1.)/(o+1.))+(0.330137393094)*dim
        ref[1,0,0,0]=(-0.869406884331)*(1+2.*(dim-1.)/(o+1.))+(2.05181576606)*dim
        ref[1,0,0,1]=(-0.336592568333)*(1+2.*(dim-1.)/(o+1.))+(-1.18221116668)*dim
        ref[1,0,1,0]=(0.119696422835)*(1+2.*(dim-1.)/(o+1.))+(-0.599288997881)*dim
        ref[1,0,1,1]=(1.71238132571)*(1+2.*(dim-1.)/(o+1.))+(-0.7893226229)*dim
        ref[1,0,2,0]=(0.301341811148)*(1+2.*(dim-1.)/(o+1.))+(-0.0125740667408)*dim
        ref[1,0,2,1]=(1.9246495335)*(1+2.*(dim-1.)/(o+1.))+(1.08856204899)*dim
        ref[1,1,0,0]=(1.40118798134)*(1+2.*(dim-1.)/(o+1.))+(0.0356919110082)*dim
        ref[1,1,0,1]=(0.270938821125)*(1+2.*(dim-1.)/(o+1.))+(1.42092614839)*dim
        ref[1,1,1,0]=(0.353375130983)*(1+2.*(dim-1.)/(o+1.))+(0.187158730917)*dim
        ref[1,1,1,1]=(0.333323765378)*(1+2.*(dim-1.)/(o+1.))+(-0.28376450019)*dim
        ref[1,1,2,0]=(-0.248745338987)*(1+2.*(dim-1.)/(o+1.))+(0.643976386833)*dim
        ref[1,1,2,1]=(0.112193420576)*(1+2.*(dim-1.)/(o+1.))+(0.931183771431)*dim
        ref[1,2,0,0]=(0.397534448472)*(1+2.*(dim-1.)/(o+1.))+(0.593499731827)*dim
        ref[1,2,0,1]=(0.199179244034)*(1+2.*(dim-1.)/(o+1.))+(-0.43545049124)*dim
        ref[1,2,1,0]=(1.18684774341)*(1+2.*(dim-1.)/(o+1.))+(-1.70814809742)*dim
        ref[1,2,1,1]=(0.534069235122)*(1+2.*(dim-1.)/(o+1.))+(-1.88969610849)*dim
        ref[1,2,2,0]=(0.912252247246)*(1+2.*(dim-1.)/(o+1.))+(-0.470932635203)*dim
        ref[1,2,2,1]=(-0.32508464376)*(1+2.*(dim-1.)/(o+1.))+(0.0983389069861)*dim
        ref[1,3,0,0]=(-0.171633539573)*(1+2.*(dim-1.)/(o+1.))+(0.348598914967)*dim
        ref[1,3,0,1]=(-0.957440569258)*(1+2.*(dim-1.)/(o+1.))+(-0.864938333196)*dim
        ref[1,3,1,0]=(0.589562599742)*(1+2.*(dim-1.)/(o+1.))+(0.793057481326)*dim
        ref[1,3,1,1]=(-0.0482452897207)*(1+2.*(dim-1.)/(o+1.))+(0.508869499554)*dim
        ref[1,3,2,0]=(-0.0784270268847)*(1+2.*(dim-1.)/(o+1.))+(1.25932568262)*dim
        ref[1,3,2,1]=(0.0167051424243)*(1+2.*(dim-1.)/(o+1.))+(-0.065290003598)*dim
        ref[1,4,0,0]=(1.91601318124)*(1+2.*(dim-1.)/(o+1.))+(1.98036186445)*dim
        ref[1,4,0,1]=(0.713480574424)*(1+2.*(dim-1.)/(o+1.))+(-0.707403424152)*dim
        ref[1,4,1,0]=(-0.857512054213)*(1+2.*(dim-1.)/(o+1.))+(-2.31655215244)*dim
        ref[1,4,1,1]=(-1.99759443728)*(1+2.*(dim-1.)/(o+1.))+(-1.40000792329)*dim
        ref[1,4,2,0]=(1.41438153297)*(1+2.*(dim-1.)/(o+1.))+(-0.345347994805)*dim
        ref[1,4,2,1]=(0.530414504493)*(1+2.*(dim-1.)/(o+1.))+(-0.826236442424)*dim
        ref[2,0,0,0]=(-0.256926617626)*(1+2.*(dim-1.)/(o+1.))+(0.322647489466)*dim
        ref[2,0,0,1]=(-0.64244416796)*(1+2.*(dim-1.)/(o+1.))+(0.176808507901)*dim
        ref[2,0,1,0]=(0.997098314441)*(1+2.*(dim-1.)/(o+1.))+(-1.24222192012)*dim
        ref[2,0,1,1]=(-0.0509223069533)*(1+2.*(dim-1.)/(o+1.))+(0.481061725309)*dim
        ref[2,0,2,0]=(0.561704567066)*(1+2.*(dim-1.)/(o+1.))+(1.30764517127)*dim
        ref[2,0,2,1]=(-1.88199995254)*(1+2.*(dim-1.)/(o+1.))+(-0.0226276389294)*dim
        ref[2,1,0,0]=(0.287070498014)*(1+2.*(dim-1.)/(o+1.))+(0.947932436577)*dim
        ref[2,1,0,1]=(-0.157324065109)*(1+2.*(dim-1.)/(o+1.))+(-0.56558633484)*dim
        ref[2,1,1,0]=(2.41484335441)*(1+2.*(dim-1.)/(o+1.))+(1.62154300036)*dim
        ref[2,1,1,1]=(0.515090798015)*(1+2.*(dim-1.)/(o+1.))+(-0.106267639566)*dim
        ref[2,1,2,0]=(0.339977301998)*(1+2.*(dim-1.)/(o+1.))+(0.650055768373)*dim
        ref[2,1,2,1]=(0.00732834638585)*(1+2.*(dim-1.)/(o+1.))+(-0.107666349465)*dim
        ref[2,2,0,0]=(-0.189246218757)*(1+2.*(dim-1.)/(o+1.))+(0.454522751637)*dim
        ref[2,2,0,1]=(-0.543402392433)*(1+2.*(dim-1.)/(o+1.))+(0.502458137788)*dim
        ref[2,2,1,0]=(-1.23702751248)*(1+2.*(dim-1.)/(o+1.))+(0.76331128837)*dim
        ref[2,2,1,1]=(0.429799353598)*(1+2.*(dim-1.)/(o+1.))+(0.611761995459)*dim
        ref[2,2,2,0]=(0.24797246329)*(1+2.*(dim-1.)/(o+1.))+(1.06577726308)*dim
        ref[2,2,2,1]=(-0.131413059009)*(1+2.*(dim-1.)/(o+1.))+(-1.61304020977)*dim
        ref[2,3,0,0]=(1.00700472927)*(1+2.*(dim-1.)/(o+1.))+(-0.796179800777)*dim
        ref[2,3,0,1]=(-1.01976922156)*(1+2.*(dim-1.)/(o+1.))+(0.240681506689)*dim
        ref[2,3,1,0]=(0.692581080055)*(1+2.*(dim-1.)/(o+1.))+(0.312340411844)*dim
        ref[2,3,1,1]=(0.636830781395)*(1+2.*(dim-1.)/(o+1.))+(-1.44693056666)*dim
        ref[2,3,2,0]=(-0.0127223987779)*(1+2.*(dim-1.)/(o+1.))+(0.666007093324)*dim
        ref[2,3,2,1]=(-0.584073475747)*(1+2.*(dim-1.)/(o+1.))+(0.481993256882)*dim
        ref[2,4,0,0]=(-0.646793665194)*(1+2.*(dim-1.)/(o+1.))+(1.9384432463)*dim
        ref[2,4,0,1]=(-0.0885586272376)*(1+2.*(dim-1.)/(o+1.))+(0.400934724632)*dim
        ref[2,4,1,0]=(-0.328509394762)*(1+2.*(dim-1.)/(o+1.))+(0.507594228357)*dim
        ref[2,4,1,1]=(-0.342847923103)*(1+2.*(dim-1.)/(o+1.))+(-0.513314735952)*dim
        ref[2,4,2,0]=(-1.07277247765)*(1+2.*(dim-1.)/(o+1.))+(1.49917881787)*dim
        ref[2,4,2,1]=(-0.638370915912)*(1+2.*(dim-1.)/(o+1.))+(-0.379793103357)*dim
        ref[3,0,0,0]=(1.05383393774)*(1+2.*(dim-1.)/(o+1.))+(-0.80869553937)*dim
        ref[3,0,0,1]=(-1.1180868754)*(1+2.*(dim-1.)/(o+1.))+(0.778458250286)*dim
        ref[3,0,1,0]=(-1.04690118432)*(1+2.*(dim-1.)/(o+1.))+(0.0542577699855)*dim
        ref[3,0,1,1]=(-0.240055180339)*(1+2.*(dim-1.)/(o+1.))+(0.390412303135)*dim
        ref[3,0,2,0]=(0.699906975796)*(1+2.*(dim-1.)/(o+1.))+(0.139743877347)*dim
        ref[3,0,2,1]=(-0.812255007331)*(1+2.*(dim-1.)/(o+1.))+(-1.02736099565)*dim
        ref[3,1,0,0]=(0.563169706071)*(1+2.*(dim-1.)/(o+1.))+(-0.398286236275)*dim
        ref[3,1,0,1]=(1.404125136)*(1+2.*(dim-1.)/(o+1.))+(-0.613438338653)*dim
        ref[3,1,1,0]=(-0.788936333197)*(1+2.*(dim-1.)/(o+1.))+(1.13919208009)*dim
        ref[3,1,1,1]=(-0.958902124976)*(1+2.*(dim-1.)/(o+1.))+(1.20446910374)*dim
        ref[3,1,2,0]=(-0.245818126043)*(1+2.*(dim-1.)/(o+1.))+(0.00879214948826)*dim
        ref[3,1,2,1]=(0.650105263567)*(1+2.*(dim-1.)/(o+1.))+(1.30767509984)*dim
        ref[3,2,0,0]=(0.768386231403)*(1+2.*(dim-1.)/(o+1.))+(1.22260582355)*dim
        ref[3,2,0,1]=(-0.792037707123)*(1+2.*(dim-1.)/(o+1.))+(0.973008274829)*dim
        ref[3,2,1,0]=(-0.932159745347)*(1+2.*(dim-1.)/(o+1.))+(0.250658877904)*dim
        ref[3,2,1,1]=(-1.16824276615)*(1+2.*(dim-1.)/(o+1.))+(-0.47259959294)*dim
        ref[3,2,2,0]=(-0.478231998451)*(1+2.*(dim-1.)/(o+1.))+(-0.545016071879)*dim
        ref[3,2,2,1]=(0.990496136616)*(1+2.*(dim-1.)/(o+1.))+(-0.939370834771)*dim
        ref[3,3,0,0]=(-1.14997257881)*(1+2.*(dim-1.)/(o+1.))+(0.0180065471512)*dim
        ref[3,3,0,1]=(0.653230183613)*(1+2.*(dim-1.)/(o+1.))+(-0.684005274372)*dim
        ref[3,3,1,0]=(0.923738009112)*(1+2.*(dim-1.)/(o+1.))+(-0.960986142644)*dim
        ref[3,3,1,1]=(0.0253113836472)*(1+2.*(dim-1.)/(o+1.))+(-0.0392722495157)*dim
        ref[3,3,2,0]=(0.850353931663)*(1+2.*(dim-1.)/(o+1.))+(0.263796877033)*dim
        ref[3,3,2,1]=(-1.56555888324)*(1+2.*(dim-1.)/(o+1.))+(-0.578006298623)*dim
        ref[3,4,0,0]=(-0.408549712656)*(1+2.*(dim-1.)/(o+1.))+(-0.230747109637)*dim
        ref[3,4,0,1]=(0.677642296994)*(1+2.*(dim-1.)/(o+1.))+(-2.28069184958)*dim
        ref[3,4,1,0]=(1.50934405309)*(1+2.*(dim-1.)/(o+1.))+(0.733753435337)*dim
        ref[3,4,1,1]=(0.700221265947)*(1+2.*(dim-1.)/(o+1.))+(-1.21064631945)*dim
        ref[3,4,2,0]=(2.03569152071)*(1+2.*(dim-1.)/(o+1.))+(-1.49637725136)*dim
        ref[3,4,2,1]=(0.358848448676)*(1+2.*(dim-1.)/(o+1.))+(0.347816114478)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-1.08871335955)*x[0]+(0.247972631031)*x[1]
        ref=(-0.156403544195)*(1+2.*(dim-1.)/(o+1.))+(-0.684337184327)*dim
      else:
        s=(-0.324052151044)*x[0]+(0.441305482424)*x[1]+(-0.315030176112)*x[2]
        ref=(0.603955649036)*(1+2.*(dim-1.)/(o+1.))+(-0.801732493769)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.260677248137)*x[0]+(0.0211318398374)*x[1]
        s[1]=(-0.815879139039)*x[0]+(-0.918304389572)*x[1]
        ref[0]=(0.843100081089)*(1+2.*(dim-1.)/(o+1.))+(-1.08264548939)*dim
        ref[1]=(-1.3080191481)*(1+2.*(dim-1.)/(o+1.))+(-0.426164380506)*dim
      else:
        s[0]=(1.37908771986)*x[0]+(1.36518233103)*x[1]+(1.013246401)*x[2]
        s[1]=(0.556318683467)*x[0]+(1.45737568352)*x[1]+(0.0165078724907)*x[2]
        ref[0]=(2.05950634594)*(1+2.*(dim-1.)/(o+1.))+(1.69801010595)*dim
        ref[1]=(0.655472308826)*(1+2.*(dim-1.)/(o+1.))+(1.37472993065)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(1.44204352077)*x[0]+(-0.626836867163)*x[1]
        s[0,1]=(-0.295593397487)*x[0]+(-0.309954518607)*x[1]
        s[0,2]=(-0.546882380107)*x[0]+(-0.0925530562931)*x[1]
        s[0,3]=(-0.623335335636)*x[0]+(0.798887138514)*x[1]
        s[0,4]=(0.202234634868)*x[0]+(1.44014401659)*x[1]
        s[1,0]=(-0.772178934874)*x[0]+(-0.962339450381)*x[1]
        s[1,1]=(-0.136928685359)*x[0]+(0.385217139725)*x[1]
        s[1,2]=(-0.00807747505852)*x[0]+(1.08230267787)*x[1]
        s[1,3]=(-1.28422033803)*x[0]+(0.011228260616)*x[1]
        s[1,4]=(0.766746911823)*x[0]+(-0.145254490276)*x[1]
        s[2,0]=(1.65171936301)*x[0]+(1.68030971997)*x[1]
        s[2,1]=(-0.465968023499)*x[0]+(-0.594634053719)*x[1]
        s[2,2]=(-0.140916975546)*x[0]+(0.230405132955)*x[1]
        s[2,3]=(-0.0185338227586)*x[0]+(0.298041051546)*x[1]
        s[2,4]=(-0.024340230778)*x[0]+(-0.791873643457)*x[1]
        s[3,0]=(0.555336313403)*x[0]+(0.606898490822)*x[1]
        s[3,1]=(1.05785941408)*x[0]+(-0.672561947115)*x[1]
        s[3,2]=(1.36138187068)*x[0]+(0.259592704164)*x[1]
        s[3,3]=(0.66211210256)*x[0]+(-1.38753456619)*x[1]
        s[3,4]=(0.0789301309156)*x[0]+(1.24015676345)*x[1]
        ref[0,0]=(0.863657572901)*(1+2.*(dim-1.)/(o+1.))+(-0.0484509192919)*dim
        ref[0,1]=(-0.0608826851336)*(1+2.*(dim-1.)/(o+1.))+(-0.54466523096)*dim
        ref[0,2]=(-1.25630630361)*(1+2.*(dim-1.)/(o+1.))+(0.616870867215)*dim
        ref[0,3]=(-0.618026054387)*(1+2.*(dim-1.)/(o+1.))+(0.793577857265)*dim
        ref[0,4]=(1.55309725329)*(1+2.*(dim-1.)/(o+1.))+(0.0892813981594)*dim
        ref[1,0]=(-0.707427946889)*(1+2.*(dim-1.)/(o+1.))+(-1.02709043837)*dim
        ref[1,1]=(1.0516700664)*(1+2.*(dim-1.)/(o+1.))+(-0.803381612038)*dim
        ref[1,2]=(-0.251691185365)*(1+2.*(dim-1.)/(o+1.))+(1.32591638818)*dim
        ref[1,3]=(0.227444315739)*(1+2.*(dim-1.)/(o+1.))+(-1.50043639316)*dim
        ref[1,4]=(0.636926531504)*(1+2.*(dim-1.)/(o+1.))+(-0.0154341099568)*dim
        ref[2,0]=(1.58333782129)*(1+2.*(dim-1.)/(o+1.))+(1.7486912617)*dim
        ref[2,1]=(0.376537776448)*(1+2.*(dim-1.)/(o+1.))+(-1.43713985367)*dim
        ref[2,2]=(1.33315471883)*(1+2.*(dim-1.)/(o+1.))+(-1.24366656142)*dim
        ref[2,3]=(-0.893476349046)*(1+2.*(dim-1.)/(o+1.))+(1.17298357783)*dim
        ref[2,4]=(-0.556720283569)*(1+2.*(dim-1.)/(o+1.))+(-0.259493590666)*dim
        ref[3,0]=(-0.0344322176845)*(1+2.*(dim-1.)/(o+1.))+(1.19666702191)*dim
        ref[3,1]=(0.356058306775)*(1+2.*(dim-1.)/(o+1.))+(0.0292391601866)*dim
        ref[3,2]=(1.18926653697)*(1+2.*(dim-1.)/(o+1.))+(0.431708037874)*dim
        ref[3,3]=(-0.0855874137907)*(1+2.*(dim-1.)/(o+1.))+(-0.639835049842)*dim
        ref[3,4]=(0.485696084789)*(1+2.*(dim-1.)/(o+1.))+(0.833390809579)*dim
      else:
        s[0,0]=(1.48740802799)*x[0]+(-0.5231718635)*x[1]+(0.621553689326)*x[2]
        s[0,1]=(-0.816432702835)*x[0]+(-1.02472113234)*x[1]+(-0.498250552797)*x[2]
        s[0,2]=(-0.0547881618906)*x[0]+(0.903306651062)*x[1]+(0.471428097681)*x[2]
        s[0,3]=(-0.910642152807)*x[0]+(-0.958224215021)*x[1]+(-1.4403271378)*x[2]
        s[0,4]=(0.116198234381)*x[0]+(-0.636378388418)*x[1]+(0.1270635999)*x[2]
        s[1,0]=(-0.659576291097)*x[0]+(-0.446587669482)*x[1]+(1.38617928144)*x[2]
        s[1,1]=(0.442922554555)*x[0]+(0.2841566742)*x[1]+(-0.384854418521)*x[2]
        s[1,2]=(-0.883255151971)*x[0]+(-1.3015449692)*x[1]+(-1.49785758846)*x[2]
        s[1,3]=(-1.43150392129)*x[0]+(1.13296295364)*x[1]+(0.393410729481)*x[2]
        s[1,4]=(-0.423254486419)*x[0]+(1.58675432746)*x[1]+(1.46179069555)*x[2]
        s[2,0]=(0.467182623429)*x[0]+(-1.23530058455)*x[1]+(0.546258172921)*x[2]
        s[2,1]=(-1.2468563949)*x[0]+(1.05285375014)*x[1]+(1.00605876886)*x[2]
        s[2,2]=(-0.1261124635)*x[0]+(-0.855000687125)*x[1]+(-0.252904511506)*x[2]
        s[2,3]=(1.1950830615)*x[0]+(0.699000137772)*x[1]+(1.33793802349)*x[2]
        s[2,4]=(-1.26581081559)*x[0]+(-0.0366783527385)*x[1]+(-0.736592212733)*x[2]
        s[3,0]=(1.84878865394)*x[0]+(0.504454193662)*x[1]+(-1.15690555199)*x[2]
        s[3,1]=(-0.249497283612)*x[0]+(-0.635940805715)*x[1]+(0.168382197376)*x[2]
        s[3,2]=(-0.167328351962)*x[0]+(0.523012476804)*x[1]+(1.19870547556)*x[2]
        s[3,3]=(-0.499150198998)*x[0]+(0.448461159266)*x[1]+(0.554069089559)*x[2]
        s[3,4]=(-0.520146860421)*x[0]+(1.43515582204)*x[1]+(1.31685474485)*x[2]
        ref[0,0]=(1.56352587364)*(1+2.*(dim-1.)/(o+1.))+(0.0222639801801)*dim
        ref[0,1]=(-1.88354672396)*(1+2.*(dim-1.)/(o+1.))+(-0.455857664013)*dim
        ref[0,2]=(0.631897841337)*(1+2.*(dim-1.)/(o+1.))+(0.688048745515)*dim
        ref[0,3]=(-2.49358987828)*(1+2.*(dim-1.)/(o+1.))+(-0.815603627346)*dim
        ref[0,4]=(-1.43782849721)*(1+2.*(dim-1.)/(o+1.))+(1.04471194307)*dim
        ref[1,0]=(-0.257186244728)*(1+2.*(dim-1.)/(o+1.))+(0.537201565586)*dim
        ref[1,1]=(-1.21753131847)*(1+2.*(dim-1.)/(o+1.))+(1.55975612871)*dim
        ref[1,2]=(-1.83391505519)*(1+2.*(dim-1.)/(o+1.))+(-1.84874265444)*dim
        ref[1,3]=(0.790398824851)*(1+2.*(dim-1.)/(o+1.))+(-0.69552906302)*dim
        ref[1,4]=(1.96838334672)*(1+2.*(dim-1.)/(o+1.))+(0.656907189867)*dim
        ref[2,0]=(-0.486341534759)*(1+2.*(dim-1.)/(o+1.))+(0.264481746559)*dim
        ref[2,1]=(-0.417272431595)*(1+2.*(dim-1.)/(o+1.))+(1.22932855569)*dim
        ref[2,2]=(-0.881184195775)*(1+2.*(dim-1.)/(o+1.))+(-0.352833466355)*dim
        ref[2,3]=(1.01420597285)*(1+2.*(dim-1.)/(o+1.))+(2.21781524991)*dim
        ref[2,4]=(-0.580092347401)*(1+2.*(dim-1.)/(o+1.))+(-1.45898903366)*dim
        ref[3,0]=(0.865857995406)*(1+2.*(dim-1.)/(o+1.))+(0.330479300213)*dim
        ref[3,1]=(0.632704550785)*(1+2.*(dim-1.)/(o+1.))+(-1.34976044274)*dim
        ref[3,2]=(-0.259081469822)*(1+2.*(dim-1.)/(o+1.))+(1.81347107022)*dim
        ref[3,3]=(0.533963525526)*(1+2.*(dim-1.)/(o+1.))+(-0.0305834756989)*dim
        ref[3,4]=(1.19847134659)*(1+2.*(dim-1.)/(o+1.))+(1.03339235988)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.536617774033)*x[0]+(-1.31191068914)*x[1]
        s[0,0,1]=(1.45706261992)*x[0]+(0.33760400949)*x[1]
        s[0,1,0]=(-0.453569968232)*x[0]+(0.798108537393)*x[1]
        s[0,1,1]=(-1.26956398588)*x[0]+(-1.35096551318)*x[1]
        s[1,0,0]=(0.104407588284)*x[0]+(-0.649935519811)*x[1]
        s[1,0,1]=(-1.64047091856)*x[0]+(0.654330987243)*x[1]
        s[1,1,0]=(-0.161010086142)*x[0]+(-0.644341383353)*x[1]
        s[1,1,1]=(1.42345151896)*x[0]+(-0.963960533124)*x[1]
        s[2,0,0]=(0.069463733935)*x[0]+(-0.764246138958)*x[1]
        s[2,0,1]=(0.516672615671)*x[0]+(0.32129769972)*x[1]
        s[2,1,0]=(0.399513997078)*x[0]+(-1.76803609001)*x[1]
        s[2,1,1]=(0.0570653989294)*x[0]+(-0.191869561254)*x[1]
        s[3,0,0]=(-0.261019396242)*x[0]+(0.429544509966)*x[1]
        s[3,0,1]=(-1.40560961029)*x[0]+(-0.739182555238)*x[1]
        s[3,1,0]=(0.314850928569)*x[0]+(1.49030138742)*x[1]
        s[3,1,1]=(-1.27327662424)*x[0]+(-0.0287977839083)*x[1]
        s[4,0,0]=(-0.840541224777)*x[0]+(-0.805224659169)*x[1]
        s[4,0,1]=(0.303904176581)*x[0]+(0.353767617246)*x[1]
        s[4,1,0]=(1.65813306507)*x[0]+(1.12299575058)*x[1]
        s[4,1,1]=(-1.19199290034)*x[0]+(1.6581285511)*x[1]
        s[5,0,0]=(0.120283505072)*x[0]+(1.68044908685)*x[1]
        s[5,0,1]=(0.382489278089)*x[0]+(0.129887092126)*x[1]
        s[5,1,0]=(-1.01890077156)*x[0]+(-0.111140290699)*x[1]
        s[5,1,1]=(-0.0392100253378)*x[0]+(0.579951319719)*x[1]
        ref[0,0,0]=(-0.686284158891)*(1+2.*(dim-1.)/(o+1.))+(-0.089008756216)*dim
        ref[0,0,1]=(1.76417819838)*(1+2.*(dim-1.)/(o+1.))+(0.030488431025)*dim
        ref[0,1,0]=(0.453381498643)*(1+2.*(dim-1.)/(o+1.))+(-0.108842929482)*dim
        ref[0,1,1]=(-1.03607416678)*(1+2.*(dim-1.)/(o+1.))+(-1.58445533227)*dim
        ref[1,0,0]=(0.638386446298)*(1+2.*(dim-1.)/(o+1.))+(-1.18391437783)*dim
        ref[1,0,1]=(0.275023590147)*(1+2.*(dim-1.)/(o+1.))+(-1.26116352146)*dim
        ref[1,1,0]=(-0.869379328882)*(1+2.*(dim-1.)/(o+1.))+(0.0640278593865)*dim
        ref[1,1,1]=(0.597799121533)*(1+2.*(dim-1.)/(o+1.))+(-0.138308135699)*dim
        ref[2,0,0]=(-1.45715684308)*(1+2.*(dim-1.)/(o+1.))+(0.762374438053)*dim
        ref[2,0,1]=(0.929718948925)*(1+2.*(dim-1.)/(o+1.))+(-0.0917486335329)*dim
        ref[2,1,0]=(-0.78092454297)*(1+2.*(dim-1.)/(o+1.))+(-0.587597549963)*dim
        ref[2,1,1]=(-1.21253063514)*(1+2.*(dim-1.)/(o+1.))+(1.07772647282)*dim
        ref[3,0,0]=(1.07007881303)*(1+2.*(dim-1.)/(o+1.))+(-0.901553699308)*dim
        ref[3,0,1]=(-1.77159657093)*(1+2.*(dim-1.)/(o+1.))+(-0.373195594594)*dim
        ref[3,1,0]=(1.10443227539)*(1+2.*(dim-1.)/(o+1.))+(0.700720040595)*dim
        ref[3,1,1]=(0.315983219905)*(1+2.*(dim-1.)/(o+1.))+(-1.61805762805)*dim
        ref[4,0,0]=(-0.602218266612)*(1+2.*(dim-1.)/(o+1.))+(-1.04354761733)*dim
        ref[4,0,1]=(0.297276459441)*(1+2.*(dim-1.)/(o+1.))+(0.360395334386)*dim
        ref[4,1,0]=(1.5584687424)*(1+2.*(dim-1.)/(o+1.))+(1.22266007325)*dim
        ref[4,1,1]=(0.504270780151)*(1+2.*(dim-1.)/(o+1.))+(-0.038135129389)*dim
        ref[5,0,0]=(1.35289487036)*(1+2.*(dim-1.)/(o+1.))+(0.44783772157)*dim
        ref[5,0,1]=(1.57660464714)*(1+2.*(dim-1.)/(o+1.))+(-1.06422827692)*dim
        ref[5,1,0]=(0.1360111664)*(1+2.*(dim-1.)/(o+1.))+(-1.26605222866)*dim
        ref[5,1,1]=(0.024986479373)*(1+2.*(dim-1.)/(o+1.))+(0.515754815008)*dim
      else:
        s[0,0,0]=(-0.491953157793)*x[0]+(0.808487399764)*x[1]+(0.87783526238)*x[2]
        s[0,0,1]=(-0.470158989175)*x[0]+(-0.492379274826)*x[1]+(-0.191290328078)*x[2]
        s[0,1,0]=(-0.356351922052)*x[0]+(0.823887078075)*x[1]+(-0.177422034195)*x[2]
        s[0,1,1]=(-0.378759483085)*x[0]+(0.116067213963)*x[1]+(0.0288265220551)*x[2]
        s[1,0,0]=(-0.325670782382)*x[0]+(-1.01289817775)*x[1]+(0.672608800327)*x[2]
        s[1,0,1]=(-0.296979811719)*x[0]+(0.46777334672)*x[1]+(1.61350930522)*x[2]
        s[1,1,0]=(-0.0574537817174)*x[0]+(1.10617918343)*x[1]+(0.172222519609)*x[2]
        s[1,1,1]=(-0.978413596402)*x[0]+(-1.20699880678)*x[1]+(-0.0600246637942)*x[2]
        s[2,0,0]=(0.637900823637)*x[0]+(1.29751687672)*x[1]+(-0.390955179963)*x[2]
        s[2,0,1]=(-1.15596339731)*x[0]+(-1.04131524409)*x[1]+(0.111161978383)*x[2]
        s[2,1,0]=(-1.33713877493)*x[0]+(0.0677540624633)*x[1]+(0.0662976219802)*x[2]
        s[2,1,1]=(0.0636656461261)*x[0]+(-1.32776475079)*x[1]+(0.365594993394)*x[2]
        s[3,0,0]=(0.454007484649)*x[0]+(-1.56273676656)*x[1]+(-0.783612553822)*x[2]
        s[3,0,1]=(-0.908388718727)*x[0]+(0.496552922533)*x[1]+(1.0802957835)*x[2]
        s[3,1,0]=(0.46784249418)*x[0]+(1.627686011)*x[1]+(1.12464788868)*x[2]
        s[3,1,1]=(1.47169416804)*x[0]+(-0.541043548439)*x[1]+(-0.0203908370923)*x[2]
        s[4,0,0]=(0.153593270409)*x[0]+(0.991271922294)*x[1]+(0.910688732568)*x[2]
        s[4,0,1]=(1.26666528018)*x[0]+(-0.0415946389155)*x[1]+(-0.702890723051)*x[2]
        s[4,1,0]=(0.401000008974)*x[0]+(0.270202736778)*x[1]+(-1.52334516556)*x[2]
        s[4,1,1]=(-0.567247526123)*x[0]+(0.0638484526058)*x[1]+(-0.942447437378)*x[2]
        s[5,0,0]=(1.3141482761)*x[0]+(0.602038495282)*x[1]+(-0.162794144191)*x[2]
        s[5,0,1]=(1.49343941166)*x[0]+(0.340106070028)*x[1]+(0.222963641278)*x[2]
        s[5,1,0]=(1.29779085951)*x[0]+(0.44436851782)*x[1]+(-0.414930735066)*x[2]
        s[5,1,1]=(0.445071462828)*x[0]+(0.0113100435192)*x[1]+(-0.304892738875)*x[2]
        ref[0,0,0]=(0.807154905774)*(1+2.*(dim-1.)/(o+1.))+(0.387214598576)*dim
        ref[0,0,1]=(-0.168668456531)*(1+2.*(dim-1.)/(o+1.))+(-0.985160135549)*dim
        ref[0,1,0]=(-0.405021627391)*(1+2.*(dim-1.)/(o+1.))+(0.69513474922)*dim
        ref[0,1,1]=(0.481047411067)*(1+2.*(dim-1.)/(o+1.))+(-0.714913158134)*dim
        ref[1,0,0]=(-0.241963228819)*(1+2.*(dim-1.)/(o+1.))+(-0.423996930987)*dim
        ref[1,0,1]=(-0.222173232945)*(1+2.*(dim-1.)/(o+1.))+(2.00647607317)*dim
        ref[1,1,0]=(2.08927091112)*(1+2.*(dim-1.)/(o+1.))+(-0.868322989801)*dim
        ref[1,1,1]=(-1.2199872069)*(1+2.*(dim-1.)/(o+1.))+(-1.02544986008)*dim
        ref[2,0,0]=(1.51358492234)*(1+2.*(dim-1.)/(o+1.))+(0.0308775980534)*dim
        ref[2,0,1]=(-1.17419262019)*(1+2.*(dim-1.)/(o+1.))+(-0.911924042825)*dim
        ref[2,1,0]=(-0.808827279389)*(1+2.*(dim-1.)/(o+1.))+(-0.394259811094)*dim
        ref[2,1,1]=(-0.39434462087)*(1+2.*(dim-1.)/(o+1.))+(-0.504159490404)*dim
        ref[3,0,0]=(-0.420509147061)*(1+2.*(dim-1.)/(o+1.))+(-1.47183268867)*dim
        ref[3,0,1]=(-0.509702954626)*(1+2.*(dim-1.)/(o+1.))+(1.17816294193)*dim
        ref[3,1,0]=(2.45580213911)*(1+2.*(dim-1.)/(o+1.))+(0.764374254754)*dim
        ref[3,1,1]=(0.36148972446)*(1+2.*(dim-1.)/(o+1.))+(0.548770058051)*dim
        ref[4,0,0]=(0.723463115757)*(1+2.*(dim-1.)/(o+1.))+(1.33209080951)*dim
        ref[4,0,1]=(-0.584660571656)*(1+2.*(dim-1.)/(o+1.))+(1.10684048987)*dim
        ref[4,1,0]=(0.558877069737)*(1+2.*(dim-1.)/(o+1.))+(-1.41101948955)*dim
        ref[4,1,1]=(-1.46224923696)*(1+2.*(dim-1.)/(o+1.))+(0.016402726067)*dim
        ref[5,0,0]=(0.375824940974)*(1+2.*(dim-1.)/(o+1.))+(1.37756768621)*dim
        ref[5,0,1]=(0.0790629051458)*(1+2.*(dim-1.)/(o+1.))+(1.97744621782)*dim
        ref[5,1,0]=(-0.429416156192)*(1+2.*(dim-1.)/(o+1.))+(1.75664479845)*dim
        ref[5,1,1]=(-0.227051313962)*(1+2.*(dim-1.)/(o+1.))+(0.378540081434)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-1.24466004429)*x[0]+(0.249656583863)*x[1]
        s[0,0,0,1]=(-1.31303014286)*x[0]+(-0.188187818679)*x[1]
        s[0,0,1,0]=(-1.19686759063)*x[0]+(0.202998667098)*x[1]
        s[0,0,1,1]=(-0.240139422625)*x[0]+(1.12145727417)*x[1]
        s[0,0,2,0]=(-0.779228056931)*x[0]+(-0.256394958159)*x[1]
        s[0,0,2,1]=(0.511346899235)*x[0]+(-1.17117401759)*x[1]
        s[0,1,0,0]=(1.20931773874)*x[0]+(0.703848713877)*x[1]
        s[0,1,0,1]=(1.05396707864)*x[0]+(0.561081392967)*x[1]
        s[0,1,1,0]=(-0.940092412899)*x[0]+(0.197879835766)*x[1]
        s[0,1,1,1]=(0.031033640344)*x[0]+(1.47668227837)*x[1]
        s[0,1,2,0]=(-0.973745458182)*x[0]+(-0.699999591381)*x[1]
        s[0,1,2,1]=(0.0637513252131)*x[0]+(0.223369875142)*x[1]
        s[0,2,0,0]=(-0.179817828717)*x[0]+(-1.0618817379)*x[1]
        s[0,2,0,1]=(-1.31848741633)*x[0]+(0.562871230704)*x[1]
        s[0,2,1,0]=(0.20660993199)*x[0]+(-0.260401741099)*x[1]
        s[0,2,1,1]=(-0.119752579601)*x[0]+(1.03659426931)*x[1]
        s[0,2,2,0]=(-1.05257053872)*x[0]+(-0.410999368732)*x[1]
        s[0,2,2,1]=(0.142472752971)*x[0]+(0.687099054848)*x[1]
        s[0,3,0,0]=(-0.735958436285)*x[0]+(0.576103117069)*x[1]
        s[0,3,0,1]=(0.820759477791)*x[0]+(1.40201775771)*x[1]
        s[0,3,1,0]=(-1.43663163172)*x[0]+(-0.928562287497)*x[1]
        s[0,3,1,1]=(-0.640592604343)*x[0]+(-1.37951764035)*x[1]
        s[0,3,2,0]=(0.923170161037)*x[0]+(-1.29865735065)*x[1]
        s[0,3,2,1]=(-1.49185366667)*x[0]+(1.26235954659)*x[1]
        s[0,4,0,0]=(-1.54260726706)*x[0]+(0.386401532954)*x[1]
        s[0,4,0,1]=(0.938275027884)*x[0]+(-0.325978711858)*x[1]
        s[0,4,1,0]=(-0.741422170549)*x[0]+(0.360529257608)*x[1]
        s[0,4,1,1]=(-1.38867328824)*x[0]+(0.368979718783)*x[1]
        s[0,4,2,0]=(-1.19633228454)*x[0]+(0.143128822342)*x[1]
        s[0,4,2,1]=(0.365590220489)*x[0]+(0.702852400998)*x[1]
        s[1,0,0,0]=(0.590708718523)*x[0]+(-1.74588023915)*x[1]
        s[1,0,0,1]=(-0.411269052638)*x[0]+(0.677504850422)*x[1]
        s[1,0,1,0]=(1.54641105105)*x[0]+(-0.227798671895)*x[1]
        s[1,0,1,1]=(0.206986074762)*x[0]+(0.431911634342)*x[1]
        s[1,0,2,0]=(0.180245950747)*x[0]+(-0.766496827464)*x[1]
        s[1,0,2,1]=(0.920490611623)*x[0]+(0.146274601957)*x[1]
        s[1,1,0,0]=(-1.47489897228)*x[0]+(-1.0428187509)*x[1]
        s[1,1,0,1]=(-1.17774302127)*x[0]+(0.44418244557)*x[1]
        s[1,1,1,0]=(-0.340268939926)*x[0]+(-0.362308449805)*x[1]
        s[1,1,1,1]=(1.1122568876)*x[0]+(0.807887213986)*x[1]
        s[1,1,2,0]=(-1.1580670758)*x[0]+(0.61015974696)*x[1]
        s[1,1,2,1]=(-0.21738296838)*x[0]+(-1.15287881926)*x[1]
        s[1,2,0,0]=(-1.63285965164)*x[0]+(0.546571504115)*x[1]
        s[1,2,0,1]=(1.34627844359)*x[0]+(0.0153020013782)*x[1]
        s[1,2,1,0]=(-0.459380401764)*x[0]+(0.653971151372)*x[1]
        s[1,2,1,1]=(0.0918961369392)*x[0]+(0.631274394662)*x[1]
        s[1,2,2,0]=(-0.461548096839)*x[0]+(0.432322561401)*x[1]
        s[1,2,2,1]=(-0.741897062249)*x[0]+(-0.0397725689655)*x[1]
        s[1,3,0,0]=(-0.166826966339)*x[0]+(0.76157630088)*x[1]
        s[1,3,0,1]=(-1.43001526663)*x[0]+(0.0200445726731)*x[1]
        s[1,3,1,0]=(0.776640170933)*x[0]+(1.77522529504)*x[1]
        s[1,3,1,1]=(0.132999332599)*x[0]+(0.91396296202)*x[1]
        s[1,3,2,0]=(-0.435447499973)*x[0]+(-0.714956349315)*x[1]
        s[1,3,2,1]=(1.54444404505)*x[0]+(0.0419015202674)*x[1]
        s[1,4,0,0]=(-1.14197604256)*x[0]+(0.140549307511)*x[1]
        s[1,4,0,1]=(1.11080451267)*x[0]+(0.140105909725)*x[1]
        s[1,4,1,0]=(0.361572095315)*x[0]+(1.73141141151)*x[1]
        s[1,4,1,1]=(1.2968959883)*x[0]+(0.459397526479)*x[1]
        s[1,4,2,0]=(-1.04477229488)*x[0]+(-0.432251933613)*x[1]
        s[1,4,2,1]=(0.166352074158)*x[0]+(0.240538931271)*x[1]
        s[2,0,0,0]=(-0.302792258664)*x[0]+(0.014159412453)*x[1]
        s[2,0,0,1]=(0.161287860507)*x[0]+(-1.66699057701)*x[1]
        s[2,0,1,0]=(0.793079282708)*x[0]+(0.999374748613)*x[1]
        s[2,0,1,1]=(-0.0860624573943)*x[0]+(0.438101549788)*x[1]
        s[2,0,2,0]=(-0.832685279306)*x[0]+(1.20735467656)*x[1]
        s[2,0,2,1]=(-0.0628631179915)*x[0]+(0.720123606355)*x[1]
        s[2,1,0,0]=(0.275852566161)*x[0]+(-0.259595395227)*x[1]
        s[2,1,0,1]=(0.862309966294)*x[0]+(-0.45114296596)*x[1]
        s[2,1,1,0]=(0.14823103051)*x[0]+(-1.16121025546)*x[1]
        s[2,1,1,1]=(-1.34419311122)*x[0]+(1.63903760286)*x[1]
        s[2,1,2,0]=(1.18830622559)*x[0]+(0.985255273279)*x[1]
        s[2,1,2,1]=(-0.0208912272649)*x[0]+(0.543921272375)*x[1]
        s[2,2,0,0]=(0.510821411306)*x[0]+(-1.88592356399)*x[1]
        s[2,2,0,1]=(0.353203575328)*x[0]+(0.328922615441)*x[1]
        s[2,2,1,0]=(1.1536669592)*x[0]+(-1.05621607534)*x[1]
        s[2,2,1,1]=(-0.581456156893)*x[0]+(-0.20275776476)*x[1]
        s[2,2,2,0]=(1.23883996503)*x[0]+(0.577857478009)*x[1]
        s[2,2,2,1]=(0.938162394147)*x[0]+(1.74792706902)*x[1]
        s[2,3,0,0]=(-0.0610369156925)*x[0]+(0.316929103344)*x[1]
        s[2,3,0,1]=(0.115575825651)*x[0]+(0.196603816526)*x[1]
        s[2,3,1,0]=(-0.137564893251)*x[0]+(0.732741603189)*x[1]
        s[2,3,1,1]=(-0.659475613725)*x[0]+(0.126037736276)*x[1]
        s[2,3,2,0]=(1.1911196497)*x[0]+(0.86409088069)*x[1]
        s[2,3,2,1]=(-0.161576876539)*x[0]+(-0.519953247817)*x[1]
        s[2,4,0,0]=(0.653018736743)*x[0]+(-0.141460723253)*x[1]
        s[2,4,0,1]=(0.548072342723)*x[0]+(0.79847118191)*x[1]
        s[2,4,1,0]=(-0.894497169389)*x[0]+(0.358296451187)*x[1]
        s[2,4,1,1]=(0.104304124331)*x[0]+(0.229557340588)*x[1]
        s[2,4,2,0]=(-0.775504740819)*x[0]+(-0.914276983707)*x[1]
        s[2,4,2,1]=(-0.997444443952)*x[0]+(-0.550036262287)*x[1]
        s[3,0,0,0]=(-0.912968382437)*x[0]+(1.34668077004)*x[1]
        s[3,0,0,1]=(0.279466090703)*x[0]+(1.28107040706)*x[1]
        s[3,0,1,0]=(0.282808913079)*x[0]+(0.740167690692)*x[1]
        s[3,0,1,1]=(-1.39563519985)*x[0]+(-0.567913880569)*x[1]
        s[3,0,2,0]=(1.27572544936)*x[0]+(-0.733366778512)*x[1]
        s[3,0,2,1]=(-0.33162427954)*x[0]+(1.35270158006)*x[1]
        s[3,1,0,0]=(0.519174606536)*x[0]+(-0.372514377255)*x[1]
        s[3,1,0,1]=(-1.58036427028)*x[0]+(-0.235983616334)*x[1]
        s[3,1,1,0]=(-0.472835600722)*x[0]+(0.721357411442)*x[1]
        s[3,1,1,1]=(0.448313914421)*x[0]+(-0.138520459243)*x[1]
        s[3,1,2,0]=(-1.51018963552)*x[0]+(1.34139211207)*x[1]
        s[3,1,2,1]=(-0.212789438989)*x[0]+(-1.06575114519)*x[1]
        s[3,2,0,0]=(0.0592065335067)*x[0]+(0.359601543991)*x[1]
        s[3,2,0,1]=(-0.594769346579)*x[0]+(1.5491276729)*x[1]
        s[3,2,1,0]=(0.28455993713)*x[0]+(0.600972748948)*x[1]
        s[3,2,1,1]=(0.398577473206)*x[0]+(0.0479286078109)*x[1]
        s[3,2,2,0]=(-1.76127884396)*x[0]+(-0.538814316335)*x[1]
        s[3,2,2,1]=(-0.390841519342)*x[0]+(0.191259339024)*x[1]
        s[3,3,0,0]=(-0.877110189074)*x[0]+(0.301682318677)*x[1]
        s[3,3,0,1]=(0.112669606669)*x[0]+(1.0395312515)*x[1]
        s[3,3,1,0]=(1.04370247351)*x[0]+(0.581250871082)*x[1]
        s[3,3,1,1]=(-0.143233022608)*x[0]+(0.593143844478)*x[1]
        s[3,3,2,0]=(-0.304532308332)*x[0]+(-0.232727140662)*x[1]
        s[3,3,2,1]=(0.767141825929)*x[0]+(0.0401997820135)*x[1]
        s[3,4,0,0]=(-0.226100649415)*x[0]+(-0.656022574403)*x[1]
        s[3,4,0,1]=(0.679160670663)*x[0]+(0.507720591962)*x[1]
        s[3,4,1,0]=(-1.30886998647)*x[0]+(-1.3539695469)*x[1]
        s[3,4,1,1]=(-1.39638819903)*x[0]+(0.650428829094)*x[1]
        s[3,4,2,0]=(-0.346486286718)*x[0]+(-0.869112300588)*x[1]
        s[3,4,2,1]=(-1.27650086562)*x[0]+(0.660506644039)*x[1]
        ref[0,0,0,0]=(-0.625690911818)*(1+2.*(dim-1.)/(o+1.))+(-0.369312548605)*dim
        ref[0,0,0,1]=(-0.0186378038077)*(1+2.*(dim-1.)/(o+1.))+(-1.48258015773)*dim
        ref[0,0,1,0]=(-1.04194152975)*(1+2.*(dim-1.)/(o+1.))+(0.0480726062254)*dim
        ref[0,0,1,1]=(0.00120841586244)*(1+2.*(dim-1.)/(o+1.))+(0.880109435685)*dim
        ref[0,0,2,0]=(-0.65182866689)*(1+2.*(dim-1.)/(o+1.))+(-0.3837943482)*dim
        ref[0,0,2,1]=(-0.729868010003)*(1+2.*(dim-1.)/(o+1.))+(0.0700408916461)*dim
        ref[0,1,0,0]=(1.26484534964)*(1+2.*(dim-1.)/(o+1.))+(0.648321102977)*dim
        ref[0,1,0,1]=(0.334654674402)*(1+2.*(dim-1.)/(o+1.))+(1.28039379721)*dim
        ref[0,1,1,0]=(0.338629227563)*(1+2.*(dim-1.)/(o+1.))+(-1.0808418047)*dim
        ref[0,1,1,1]=(0.0969508871625)*(1+2.*(dim-1.)/(o+1.))+(1.41076503155)*dim
        ref[0,1,2,0]=(-0.462969154119)*(1+2.*(dim-1.)/(o+1.))+(-1.21077589544)*dim
        ref[0,1,2,1]=(0.46756493995)*(1+2.*(dim-1.)/(o+1.))+(-0.180443739595)*dim
        ref[0,2,0,0]=(-0.0576220389279)*(1+2.*(dim-1.)/(o+1.))+(-1.18407752769)*dim
        ref[0,2,0,1]=(-0.650282644944)*(1+2.*(dim-1.)/(o+1.))+(-0.105333540684)*dim
        ref[0,2,1,0]=(0.136853100987)*(1+2.*(dim-1.)/(o+1.))+(-0.190644910096)*dim
        ref[0,2,1,1]=(-0.637797927534)*(1+2.*(dim-1.)/(o+1.))+(1.55463961724)*dim
        ref[0,2,2,0]=(-0.455136385489)*(1+2.*(dim-1.)/(o+1.))+(-1.00843352196)*dim
        ref[0,2,2,1]=(-0.35303452486)*(1+2.*(dim-1.)/(o+1.))+(1.18260633268)*dim
        ref[0,3,0,0]=(-0.315687014208)*(1+2.*(dim-1.)/(o+1.))+(0.155831694992)*dim
        ref[0,3,0,1]=(1.52158435346)*(1+2.*(dim-1.)/(o+1.))+(0.701192882048)*dim
        ref[0,3,1,0]=(-0.836739848177)*(1+2.*(dim-1.)/(o+1.))+(-1.52845407104)*dim
        ref[0,3,1,1]=(-0.983041539851)*(1+2.*(dim-1.)/(o+1.))+(-1.03706870484)*dim
        ref[0,3,2,0]=(-0.497690471912)*(1+2.*(dim-1.)/(o+1.))+(0.122203282296)*dim
        ref[0,3,2,1]=(0.394951943207)*(1+2.*(dim-1.)/(o+1.))+(-0.624446063287)*dim
        ref[0,4,0,0]=(-0.63974792726)*(1+2.*(dim-1.)/(o+1.))+(-0.516457806843)*dim
        ref[0,4,0,1]=(0.62556865655)*(1+2.*(dim-1.)/(o+1.))+(-0.0132723405234)*dim
        ref[0,4,1,0]=(-1.10166644693)*(1+2.*(dim-1.)/(o+1.))+(0.720773533984)*dim
        ref[0,4,1,1]=(-1.09945957099)*(1+2.*(dim-1.)/(o+1.))+(0.0797660015265)*dim
        ref[0,4,2,0]=(-0.359198665738)*(1+2.*(dim-1.)/(o+1.))+(-0.694004796462)*dim
        ref[0,4,2,1]=(1.60203342245)*(1+2.*(dim-1.)/(o+1.))+(-0.533590800966)*dim
        ref[1,0,0,0]=(0.0630998548187)*(1+2.*(dim-1.)/(o+1.))+(-1.21827137545)*dim
        ref[1,0,0,1]=(-0.362786036469)*(1+2.*(dim-1.)/(o+1.))+(0.629021834252)*dim
        ref[1,0,1,0]=(1.30840533269)*(1+2.*(dim-1.)/(o+1.))+(0.01020704647)*dim
        ref[1,0,1,1]=(0.581668517056)*(1+2.*(dim-1.)/(o+1.))+(0.0572291920482)*dim
        ref[1,0,2,0]=(-1.1596262045)*(1+2.*(dim-1.)/(o+1.))+(0.573375327782)*dim
        ref[1,0,2,1]=(-0.025337463973)*(1+2.*(dim-1.)/(o+1.))+(1.09210267755)*dim
        ref[1,1,0,0]=(-1.46016674214)*(1+2.*(dim-1.)/(o+1.))+(-1.05755098104)*dim
        ref[1,1,0,1]=(-0.410079206814)*(1+2.*(dim-1.)/(o+1.))+(-0.323481368888)*dim
        ref[1,1,1,0]=(0.497609154803)*(1+2.*(dim-1.)/(o+1.))+(-1.20018654453)*dim
        ref[1,1,1,1]=(0.446563754938)*(1+2.*(dim-1.)/(o+1.))+(1.47358034665)*dim
        ref[1,1,2,0]=(0.124142572295)*(1+2.*(dim-1.)/(o+1.))+(-0.672049901132)*dim
        ref[1,1,2,1]=(-1.27946259076)*(1+2.*(dim-1.)/(o+1.))+(-0.09079919688)*dim
        ref[1,2,0,0]=(0.0167916382593)*(1+2.*(dim-1.)/(o+1.))+(-1.10307978578)*dim
        ref[1,2,0,1]=(0.310300272224)*(1+2.*(dim-1.)/(o+1.))+(1.05128017274)*dim
        ref[1,2,1,0]=(-0.384485455128)*(1+2.*(dim-1.)/(o+1.))+(0.579076204736)*dim
        ref[1,2,1,1]=(0.133076827561)*(1+2.*(dim-1.)/(o+1.))+(0.59009370404)*dim
        ref[1,2,2,0]=(1.4126992054)*(1+2.*(dim-1.)/(o+1.))+(-1.44192474084)*dim
        ref[1,2,2,1]=(0.235736468845)*(1+2.*(dim-1.)/(o+1.))+(-1.01740610006)*dim
        ref[1,3,0,0]=(-0.323242689323)*(1+2.*(dim-1.)/(o+1.))+(0.917992023864)*dim
        ref[1,3,0,1]=(-1.56677140485)*(1+2.*(dim-1.)/(o+1.))+(0.156800710897)*dim
        ref[1,3,1,0]=(1.39852204704)*(1+2.*(dim-1.)/(o+1.))+(1.15334341893)*dim
        ref[1,3,1,1]=(1.10144038868)*(1+2.*(dim-1.)/(o+1.))+(-0.0544780940565)*dim
        ref[1,3,2,0]=(-0.619896071163)*(1+2.*(dim-1.)/(o+1.))+(-0.530507778126)*dim
        ref[1,3,2,1]=(0.665968335878)*(1+2.*(dim-1.)/(o+1.))+(0.920377229442)*dim
        ref[1,4,0,0]=(0.24940249757)*(1+2.*(dim-1.)/(o+1.))+(-1.25082923262)*dim
        ref[1,4,0,1]=(1.26616311307)*(1+2.*(dim-1.)/(o+1.))+(-0.0152526906804)*dim
        ref[1,4,1,0]=(1.37246903375)*(1+2.*(dim-1.)/(o+1.))+(0.720514473076)*dim
        ref[1,4,1,1]=(1.20885120083)*(1+2.*(dim-1.)/(o+1.))+(0.547442313947)*dim
        ref[1,4,2,0]=(-0.929784649867)*(1+2.*(dim-1.)/(o+1.))+(-0.547239578625)*dim
        ref[1,4,2,1]=(1.40793646749)*(1+2.*(dim-1.)/(o+1.))+(-1.00104546206)*dim
        ref[2,0,0,0]=(-0.865664592293)*(1+2.*(dim-1.)/(o+1.))+(0.577031746082)*dim
        ref[2,0,0,1]=(-0.383744932135)*(1+2.*(dim-1.)/(o+1.))+(-1.12195778437)*dim
        ref[2,0,1,0]=(0.899513902212)*(1+2.*(dim-1.)/(o+1.))+(0.892940129109)*dim
        ref[2,0,1,1]=(0.422487727431)*(1+2.*(dim-1.)/(o+1.))+(-0.070448635037)*dim
        ref[2,0,2,0]=(-0.0542314325807)*(1+2.*(dim-1.)/(o+1.))+(0.428900829837)*dim
        ref[2,0,2,1]=(-0.333651152721)*(1+2.*(dim-1.)/(o+1.))+(0.990911641084)*dim
        ref[2,1,0,0]=(0.743812676627)*(1+2.*(dim-1.)/(o+1.))+(-0.727555505694)*dim
        ref[2,1,0,1]=(0.487164931518)*(1+2.*(dim-1.)/(o+1.))+(-0.0759979311835)*dim
        ref[2,1,1,0]=(0.146918316537)*(1+2.*(dim-1.)/(o+1.))+(-1.15989754148)*dim
        ref[2,1,1,1]=(-0.239925787022)*(1+2.*(dim-1.)/(o+1.))+(0.534770278668)*dim
        ref[2,1,2,0]=(1.79635305588)*(1+2.*(dim-1.)/(o+1.))+(0.377208442988)*dim
        ref[2,1,2,1]=(-0.275396782095)*(1+2.*(dim-1.)/(o+1.))+(0.798426827206)*dim
        ref[2,2,0,0]=(-0.732792630073)*(1+2.*(dim-1.)/(o+1.))+(-0.642309522614)*dim
        ref[2,2,0,1]=(0.691994610633)*(1+2.*(dim-1.)/(o+1.))+(-0.0098684198643)*dim
        ref[2,2,1,0]=(0.571059981052)*(1+2.*(dim-1.)/(o+1.))+(-0.473609097192)*dim
        ref[2,2,1,1]=(-0.935084534142)*(1+2.*(dim-1.)/(o+1.))+(0.15087061249)*dim
        ref[2,2,2,0]=(0.0826567188857)*(1+2.*(dim-1.)/(o+1.))+(1.73404072415)*dim
        ref[2,2,2,1]=(1.26427004187)*(1+2.*(dim-1.)/(o+1.))+(1.4218194213)*dim
        ref[2,3,0,0]=(-0.398591309648)*(1+2.*(dim-1.)/(o+1.))+(0.6544834973)*dim
        ref[2,3,0,1]=(-0.568822853648)*(1+2.*(dim-1.)/(o+1.))+(0.881002495825)*dim
        ref[2,3,1,0]=(1.18000532793)*(1+2.*(dim-1.)/(o+1.))+(-0.584828617992)*dim
        ref[2,3,1,1]=(-1.60821457294)*(1+2.*(dim-1.)/(o+1.))+(1.07477669549)*dim
        ref[2,3,2,0]=(1.48063830505)*(1+2.*(dim-1.)/(o+1.))+(0.574572225346)*dim
        ref[2,3,2,1]=(0.229132618257)*(1+2.*(dim-1.)/(o+1.))+(-0.910662742613)*dim
        ref[2,4,0,0]=(0.426313638958)*(1+2.*(dim-1.)/(o+1.))+(0.0852443745317)*dim
        ref[2,4,0,1]=(-0.0761298452145)*(1+2.*(dim-1.)/(o+1.))+(1.42267336985)*dim
        ref[2,4,1,0]=(-0.537516881228)*(1+2.*(dim-1.)/(o+1.))+(0.00131616302647)*dim
        ref[2,4,1,1]=(-1.00615029385)*(1+2.*(dim-1.)/(o+1.))+(1.34001175877)*dim
        ref[2,4,2,0]=(-0.250744474716)*(1+2.*(dim-1.)/(o+1.))+(-1.43903724981)*dim
        ref[2,4,2,1]=(0.273616829308)*(1+2.*(dim-1.)/(o+1.))+(-1.82109753555)*dim
        ref[3,0,0,0]=(0.773912240208)*(1+2.*(dim-1.)/(o+1.))+(-0.340199852604)*dim
        ref[3,0,0,1]=(0.00495773386215)*(1+2.*(dim-1.)/(o+1.))+(1.5555787639)*dim
        ref[3,0,1,0]=(-0.586757469166)*(1+2.*(dim-1.)/(o+1.))+(1.60973407294)*dim
        ref[3,0,1,1]=(-0.577120453982)*(1+2.*(dim-1.)/(o+1.))+(-1.38642862644)*dim
        ref[3,0,2,0]=(0.288712056159)*(1+2.*(dim-1.)/(o+1.))+(0.253646614684)*dim
        ref[3,0,2,1]=(-0.248042635081)*(1+2.*(dim-1.)/(o+1.))+(1.2691199356)*dim
        ref[3,1,0,0]=(0.995572341947)*(1+2.*(dim-1.)/(o+1.))+(-0.848912112665)*dim
        ref[3,1,0,1]=(-1.37225234108)*(1+2.*(dim-1.)/(o+1.))+(-0.444095545534)*dim
        ref[3,1,1,0]=(0.441843238326)*(1+2.*(dim-1.)/(o+1.))+(-0.193321427606)*dim
        ref[3,1,1,1]=(0.408346053376)*(1+2.*(dim-1.)/(o+1.))+(-0.098552598199)*dim
        ref[3,1,2,0]=(-0.311970472748)*(1+2.*(dim-1.)/(o+1.))+(0.143172949307)*dim
        ref[3,1,2,1]=(-0.434499269883)*(1+2.*(dim-1.)/(o+1.))+(-0.844041314296)*dim
        ref[3,2,0,0]=(0.480046851902)*(1+2.*(dim-1.)/(o+1.))+(-0.0612387744048)*dim
        ref[3,2,0,1]=(1.12223817367)*(1+2.*(dim-1.)/(o+1.))+(-0.167879847351)*dim
        ref[3,2,1,0]=(-0.345095098378)*(1+2.*(dim-1.)/(o+1.))+(1.23062778446)*dim
        ref[3,2,1,1]=(1.37605147183)*(1+2.*(dim-1.)/(o+1.))+(-0.92954539081)*dim
        ref[3,2,2,0]=(-1.0202068094)*(1+2.*(dim-1.)/(o+1.))+(-1.2798863509)*dim
        ref[3,2,2,1]=(-0.302591270785)*(1+2.*(dim-1.)/(o+1.))+(0.103009090467)*dim
        ref[3,3,0,0]=(0.857177782444)*(1+2.*(dim-1.)/(o+1.))+(-1.43260565284)*dim
        ref[3,3,0,1]=(0.0906238774103)*(1+2.*(dim-1.)/(o+1.))+(1.06157698076)*dim
        ref[3,3,1,0]=(0.42136870665)*(1+2.*(dim-1.)/(o+1.))+(1.20358463794)*dim
        ref[3,3,1,1]=(-0.88620181924)*(1+2.*(dim-1.)/(o+1.))+(1.33611264111)*dim
        ref[3,3,2,0]=(0.444491342755)*(1+2.*(dim-1.)/(o+1.))+(-0.981750791749)*dim
        ref[3,3,2,1]=(-1.06075000108)*(1+2.*(dim-1.)/(o+1.))+(1.86809160902)*dim
        ref[3,4,0,0]=(-0.517172055625)*(1+2.*(dim-1.)/(o+1.))+(-0.364951168193)*dim
        ref[3,4,0,1]=(0.683410864606)*(1+2.*(dim-1.)/(o+1.))+(0.503470398019)*dim
        ref[3,4,1,0]=(-1.49890766435)*(1+2.*(dim-1.)/(o+1.))+(-1.16393186901)*dim
        ref[3,4,1,1]=(0.25999658315)*(1+2.*(dim-1.)/(o+1.))+(-1.00595595308)*dim
        ref[3,4,2,0]=(-0.296406951133)*(1+2.*(dim-1.)/(o+1.))+(-0.919191636173)*dim
        ref[3,4,2,1]=(-0.562797436325)*(1+2.*(dim-1.)/(o+1.))+(-0.0531967852595)*dim
      else:
        s[0,0,0,0]=(0.28520634077)*x[0]+(-0.126741433974)*x[1]+(1.28370871197)*x[2]
        s[0,0,0,1]=(-0.0731693108841)*x[0]+(0.342123624879)*x[1]+(0.302260810729)*x[2]
        s[0,0,1,0]=(-0.933375161617)*x[0]+(0.714257115938)*x[1]+(0.0355044702732)*x[2]
        s[0,0,1,1]=(0.273422423495)*x[0]+(-1.43381666098)*x[1]+(-0.3924531625)*x[2]
        s[0,0,2,0]=(-1.24504380748)*x[0]+(1.36389469156)*x[1]+(1.0146102387)*x[2]
        s[0,0,2,1]=(-0.12931890066)*x[0]+(0.00848180915183)*x[1]+(0.230392813094)*x[2]
        s[0,1,0,0]=(0.390247630566)*x[0]+(0.116544846708)*x[1]+(-0.94552924973)*x[2]
        s[0,1,0,1]=(1.53653205813)*x[0]+(0.984195773879)*x[1]+(0.767182801059)*x[2]
        s[0,1,1,0]=(0.433068815739)*x[0]+(0.196964560742)*x[1]+(0.00150259046749)*x[2]
        s[0,1,1,1]=(0.620083290721)*x[0]+(-0.981816668862)*x[1]+(-0.0602017060991)*x[2]
        s[0,1,2,0]=(0.266920335934)*x[0]+(-1.01077201758)*x[1]+(-1.66555517355)*x[2]
        s[0,1,2,1]=(-0.463794455357)*x[0]+(1.02905222514)*x[1]+(-0.524965351287)*x[2]
        s[0,2,0,0]=(-1.02890742598)*x[0]+(0.10277767711)*x[1]+(-0.709831378433)*x[2]
        s[0,2,0,1]=(0.942130530664)*x[0]+(-0.498478792456)*x[1]+(-0.321634750428)*x[2]
        s[0,2,1,0]=(-1.09979499064)*x[0]+(-0.552242175008)*x[1]+(1.47155591555)*x[2]
        s[0,2,1,1]=(0.038643175545)*x[0]+(-0.367877010362)*x[1]+(-0.422403338025)*x[2]
        s[0,2,2,0]=(0.915798473155)*x[0]+(0.627559693912)*x[1]+(-0.146957763577)*x[2]
        s[0,2,2,1]=(0.35609430038)*x[0]+(0.627155969018)*x[1]+(-0.0538976009369)*x[2]
        s[0,3,0,0]=(0.753367589685)*x[0]+(-1.56174987227)*x[1]+(0.0317386362973)*x[2]
        s[0,3,0,1]=(-1.0814768856)*x[0]+(-0.000317117094536)*x[1]+(0.160394230281)*x[2]
        s[0,3,1,0]=(0.539343002581)*x[0]+(0.600975350875)*x[1]+(-0.579108274756)*x[2]
        s[0,3,1,1]=(1.93822764773)*x[0]+(1.19115168218)*x[1]+(-0.345303177833)*x[2]
        s[0,3,2,0]=(-0.396284929409)*x[0]+(-0.0620671301539)*x[1]+(0.405978876833)*x[2]
        s[0,3,2,1]=(-0.806591193034)*x[0]+(-0.869878106357)*x[1]+(-0.352772882375)*x[2]
        s[0,4,0,0]=(-1.01176424875)*x[0]+(-1.57589661925)*x[1]+(-0.313904557434)*x[2]
        s[0,4,0,1]=(0.277176183617)*x[0]+(0.548456592065)*x[1]+(1.21233269169)*x[2]
        s[0,4,1,0]=(0.916512281025)*x[0]+(-1.53600125673)*x[1]+(-0.486397251593)*x[2]
        s[0,4,1,1]=(1.12553562564)*x[0]+(0.0569780615281)*x[1]+(-0.275491563277)*x[2]
        s[0,4,2,0]=(0.0701240897706)*x[0]+(-0.109284348724)*x[1]+(0.920255378883)*x[2]
        s[0,4,2,1]=(-0.286723671632)*x[0]+(0.521149269283)*x[1]+(0.880566081072)*x[2]
        s[1,0,0,0]=(-0.161092153994)*x[0]+(1.60646814731)*x[1]+(-0.293423133436)*x[2]
        s[1,0,0,1]=(-0.62705410926)*x[0]+(-0.133468132127)*x[1]+(0.435110047414)*x[2]
        s[1,0,1,0]=(0.963863560937)*x[0]+(0.358847992997)*x[1]+(0.847184531597)*x[2]
        s[1,0,1,1]=(0.2445787465)*x[0]+(-0.161812641928)*x[1]+(-0.679720454567)*x[2]
        s[1,0,2,0]=(1.05646386542)*x[0]+(1.0631242639)*x[1]+(-0.762052372247)*x[2]
        s[1,0,2,1]=(-0.950180341823)*x[0]+(1.00890659836)*x[1]+(1.54399946028)*x[2]
        s[1,1,0,0]=(0.290340455371)*x[0]+(-1.13919380812)*x[1]+(-0.659408495899)*x[2]
        s[1,1,0,1]=(1.4258405346)*x[0]+(1.53809682182)*x[1]+(0.0675302762314)*x[2]
        s[1,1,1,0]=(1.21278717612)*x[0]+(0.220266219454)*x[1]+(1.66873955725)*x[2]
        s[1,1,1,1]=(0.771526101684)*x[0]+(-0.067360487854)*x[1]+(1.21982979968)*x[2]
        s[1,1,2,0]=(-0.77604682892)*x[0]+(0.3133384598)*x[1]+(0.645828815985)*x[2]
        s[1,1,2,1]=(-0.107715380402)*x[0]+(-0.0772190829216)*x[1]+(-1.14700080403)*x[2]
        s[1,2,0,0]=(0.324137135934)*x[0]+(1.29562077427)*x[1]+(0.455108897284)*x[2]
        s[1,2,0,1]=(-1.43262429921)*x[0]+(0.773443491708)*x[1]+(0.277297920728)*x[2]
        s[1,2,1,0]=(0.304810473045)*x[0]+(-0.0168967979423)*x[1]+(0.131774697941)*x[2]
        s[1,2,1,1]=(0.387736648788)*x[0]+(0.666220386408)*x[1]+(1.07576988895)*x[2]
        s[1,2,2,0]=(-0.254693787451)*x[0]+(-0.3384585134)*x[1]+(-0.64060329677)*x[2]
        s[1,2,2,1]=(-0.672597816893)*x[0]+(0.145646769601)*x[1]+(-0.994938729226)*x[2]
        s[1,3,0,0]=(-1.1030634788)*x[0]+(-1.80762204727)*x[1]+(1.6309434404)*x[2]
        s[1,3,0,1]=(0.276318339362)*x[0]+(0.385250530079)*x[1]+(-0.758561170758)*x[2]
        s[1,3,1,0]=(0.179910601577)*x[0]+(-0.11818800924)*x[1]+(-0.743123335945)*x[2]
        s[1,3,1,1]=(0.782295683197)*x[0]+(0.749420799157)*x[1]+(-0.656475021802)*x[2]
        s[1,3,2,0]=(0.574829069465)*x[0]+(-0.493394135291)*x[1]+(1.29795947888)*x[2]
        s[1,3,2,1]=(-1.46888017959)*x[0]+(-1.08679474255)*x[1]+(0.156158770501)*x[2]
        s[1,4,0,0]=(-0.860789510784)*x[0]+(0.652130849161)*x[1]+(-0.0894965081242)*x[2]
        s[1,4,0,1]=(0.997352754709)*x[0]+(-0.112776085202)*x[1]+(-0.469141207634)*x[2]
        s[1,4,1,0]=(-0.401561591069)*x[0]+(-0.362165453433)*x[1]+(1.42046539935)*x[2]
        s[1,4,1,1]=(-0.668331880273)*x[0]+(-1.28337116108)*x[1]+(0.368963183759)*x[2]
        s[1,4,2,0]=(-1.15039637411)*x[0]+(0.121579045201)*x[1]+(-0.104055094935)*x[2]
        s[1,4,2,1]=(-0.923080753122)*x[0]+(0.322395936762)*x[1]+(0.233695594483)*x[2]
        s[2,0,0,0]=(-0.206424413737)*x[0]+(0.761779799677)*x[1]+(-0.301048388352)*x[2]
        s[2,0,0,1]=(1.55990080913)*x[0]+(-0.162113046152)*x[1]+(-0.310428209804)*x[2]
        s[2,0,1,0]=(-0.433285639024)*x[0]+(1.36777883595)*x[1]+(-1.17795591805)*x[2]
        s[2,0,1,1]=(1.34284608147)*x[0]+(-0.222861276203)*x[1]+(0.855486771517)*x[2]
        s[2,0,2,0]=(-0.16635189023)*x[0]+(-0.823533654953)*x[1]+(0.797079690281)*x[2]
        s[2,0,2,1]=(0.752310816499)*x[0]+(0.764018784752)*x[1]+(0.056117413692)*x[2]
        s[2,1,0,0]=(0.00166988723156)*x[0]+(0.766304091963)*x[1]+(0.275940662737)*x[2]
        s[2,1,0,1]=(-0.0855732564349)*x[0]+(-1.02664418473)*x[1]+(0.662522947755)*x[2]
        s[2,1,1,0]=(-0.658811588292)*x[0]+(-1.04635899154)*x[1]+(1.05470344529)*x[2]
        s[2,1,1,1]=(-0.119299788842)*x[0]+(-0.782372280168)*x[1]+(-1.79380761901)*x[2]
        s[2,1,2,0]=(1.71887176677)*x[0]+(-0.226946752041)*x[1]+(0.799181606184)*x[2]
        s[2,1,2,1]=(-0.9189461833)*x[0]+(0.916498349691)*x[1]+(0.734617496542)*x[2]
        s[2,2,0,0]=(-0.790364198949)*x[0]+(0.0131553082942)*x[1]+(-0.978651161192)*x[2]
        s[2,2,0,1]=(0.32388442717)*x[0]+(-0.593416422259)*x[1]+(1.10510332946)*x[2]
        s[2,2,1,0]=(0.452042167919)*x[0]+(-0.641462816719)*x[1]+(-0.356784432003)*x[2]
        s[2,2,1,1]=(0.0725380542114)*x[0]+(0.255256127518)*x[1]+(0.979834678019)*x[2]
        s[2,2,2,0]=(1.39372044596)*x[0]+(0.938441851122)*x[1]+(-0.983490819141)*x[2]
        s[2,2,2,1]=(1.2364973033)*x[0]+(-1.2815807633)*x[1]+(-0.360311591593)*x[2]
        s[2,3,0,0]=(-0.0597199849648)*x[0]+(0.376596450311)*x[1]+(-0.494756059286)*x[2]
        s[2,3,0,1]=(-0.1287394847)*x[0]+(0.0455819025025)*x[1]+(1.57146607969)*x[2]
        s[2,3,1,0]=(-1.4907853501)*x[0]+(0.0388186733561)*x[1]+(0.13920185582)*x[2]
        s[2,3,1,1]=(0.0439467610482)*x[0]+(1.43515095249)*x[1]+(-0.205189829549)*x[2]
        s[2,3,2,0]=(0.773581921294)*x[0]+(1.41482523023)*x[1]+(1.02435393168)*x[2]
        s[2,3,2,1]=(0.347805074163)*x[0]+(1.52992663335)*x[1]+(-0.487058346391)*x[2]
        s[2,4,0,0]=(-0.107692405269)*x[0]+(0.147562810855)*x[1]+(1.38077195652)*x[2]
        s[2,4,0,1]=(-0.0410323036269)*x[0]+(0.269432309711)*x[1]+(-0.693128417186)*x[2]
        s[2,4,1,0]=(-0.33861862665)*x[0]+(-0.419772907291)*x[1]+(-1.07238876285)*x[2]
        s[2,4,1,1]=(0.143346805317)*x[0]+(-0.419444960688)*x[1]+(1.59692612607)*x[2]
        s[2,4,2,0]=(0.654582963584)*x[0]+(-0.44163597128)*x[1]+(-0.694330043907)*x[2]
        s[2,4,2,1]=(0.627421193857)*x[0]+(0.597177617998)*x[1]+(-0.451180510793)*x[2]
        s[3,0,0,0]=(-0.101412872135)*x[0]+(-0.396884673414)*x[1]+(0.896841368992)*x[2]
        s[3,0,0,1]=(0.139595535793)*x[0]+(0.624607944421)*x[1]+(0.574170381258)*x[2]
        s[3,0,1,0]=(0.316717544294)*x[0]+(-0.621261709901)*x[1]+(1.31441803912)*x[2]
        s[3,0,1,1]=(0.254418231071)*x[0]+(0.228843906675)*x[1]+(0.25896007767)*x[2]
        s[3,0,2,0]=(-0.247274104613)*x[0]+(1.20694307591)*x[1]+(-0.407808429936)*x[2]
        s[3,0,2,1]=(-1.40605558298)*x[0]+(0.168721449606)*x[1]+(-0.972867940483)*x[2]
        s[3,1,0,0]=(0.245845223266)*x[0]+(-1.51779122633)*x[1]+(-0.779293963961)*x[2]
        s[3,1,0,1]=(0.0325495798223)*x[0]+(-0.12100728517)*x[1]+(1.31848044222)*x[2]
        s[3,1,1,0]=(0.789118948962)*x[0]+(0.589652419254)*x[1]+(-0.08688557206)*x[2]
        s[3,1,1,1]=(-1.83518748403)*x[0]+(-0.500646680661)*x[1]+(0.233750065377)*x[2]
        s[3,1,2,0]=(-0.0771120817018)*x[0]+(0.0904104042813)*x[1]+(0.0933370539969)*x[2]
        s[3,1,2,1]=(-1.19249513127)*x[0]+(0.447348983457)*x[1]+(-0.298916535972)*x[2]
        s[3,2,0,0]=(1.22266670435)*x[0]+(-1.62277644197)*x[1]+(-0.0190709988254)*x[2]
        s[3,2,0,1]=(-0.194952675382)*x[0]+(0.554694000692)*x[1]+(-0.151982113386)*x[2]
        s[3,2,1,0]=(0.511265692083)*x[0]+(-0.802622720144)*x[1]+(-0.242270723374)*x[2]
        s[3,2,1,1]=(0.476021603798)*x[0]+(-0.463494861878)*x[1]+(-0.532143290096)*x[2]
        s[3,2,2,0]=(-0.530708374495)*x[0]+(1.10856138272)*x[1]+(0.409512895954)*x[2]
        s[3,2,2,1]=(0.0214256235243)*x[0]+(0.836046354994)*x[1]+(-0.263352555948)*x[2]
        s[3,3,0,0]=(-0.437181323979)*x[0]+(-0.413716617562)*x[1]+(-0.432122540243)*x[2]
        s[3,3,0,1]=(1.08881737048)*x[0]+(-0.0166529943951)*x[1]+(-0.236235303435)*x[2]
        s[3,3,1,0]=(-0.220528005238)*x[0]+(-0.0637133108381)*x[1]+(-0.722670710646)*x[2]
        s[3,3,1,1]=(0.0267596690177)*x[0]+(0.277583150215)*x[1]+(-1.19820417967)*x[2]
        s[3,3,2,0]=(-0.134456240167)*x[0]+(0.207594364333)*x[1]+(-1.58153388102)*x[2]
        s[3,3,2,1]=(-0.484174332477)*x[0]+(1.24683115653)*x[1]+(0.0960608606087)*x[2]
        s[3,4,0,0]=(-1.06280143452)*x[0]+(0.0909193631754)*x[1]+(0.735785607027)*x[2]
        s[3,4,0,1]=(-0.39199782699)*x[0]+(-0.547457599272)*x[1]+(-0.990008593852)*x[2]
        s[3,4,1,0]=(-0.301784264736)*x[0]+(-0.219227445543)*x[1]+(-0.979494077205)*x[2]
        s[3,4,1,1]=(-0.898860901012)*x[0]+(-0.103378912588)*x[1]+(-0.66619898802)*x[2]
        s[3,4,2,0]=(1.10051932868)*x[0]+(0.407341781092)*x[1]+(-0.680512930996)*x[2]
        s[3,4,2,1]=(-0.0733544592857)*x[0]+(1.18707229675)*x[1]+(-0.777693466919)*x[2]
        ref[0,0,0,0]=(0.438851859071)*(1+2.*(dim-1.)/(o+1.))+(1.0033217597)*dim
        ref[0,0,0,1]=(-0.891607994637)*(1+2.*(dim-1.)/(o+1.))+(1.46282311936)*dim
        ref[0,0,1,0]=(0.817888583209)*(1+2.*(dim-1.)/(o+1.))+(-1.00150215862)*dim
        ref[0,0,1,1]=(-1.639070793)*(1+2.*(dim-1.)/(o+1.))+(0.0862233930162)*dim
        ref[0,0,2,0]=(-0.348875799642)*(1+2.*(dim-1.)/(o+1.))+(1.48233692242)*dim
        ref[0,0,2,1]=(-0.589821443428)*(1+2.*(dim-1.)/(o+1.))+(0.699377165015)*dim
        ref[0,1,0,0]=(-0.832218783182)*(1+2.*(dim-1.)/(o+1.))+(0.393482010727)*dim
        ref[0,1,0,1]=(2.07094909251)*(1+2.*(dim-1.)/(o+1.))+(1.21696154056)*dim
        ref[0,1,1,0]=(0.1125989301)*(1+2.*(dim-1.)/(o+1.))+(0.518937036848)*dim
        ref[0,1,1,1]=(-1.36685685946)*(1+2.*(dim-1.)/(o+1.))+(0.944921775217)*dim
        ref[0,1,2,0]=(-1.47817293944)*(1+2.*(dim-1.)/(o+1.))+(-0.931233915749)*dim
        ref[0,1,2,1]=(-0.213405460956)*(1+2.*(dim-1.)/(o+1.))+(0.253697879455)*dim
        ref[0,2,0,0]=(-0.0805522984701)*(1+2.*(dim-1.)/(o+1.))+(-1.55540882883)*dim
        ref[0,2,0,1]=(0.156888902683)*(1+2.*(dim-1.)/(o+1.))+(-0.034871914902)*dim
        ref[0,2,1,0]=(-0.15684567608)*(1+2.*(dim-1.)/(o+1.))+(-0.0236355740175)*dim
        ref[0,2,1,1]=(-1.39067500442)*(1+2.*(dim-1.)/(o+1.))+(0.639037831581)*dim
        ref[0,2,2,0]=(1.81399897277)*(1+2.*(dim-1.)/(o+1.))+(-0.417598569281)*dim
        ref[0,2,2,1]=(-0.764514507478)*(1+2.*(dim-1.)/(o+1.))+(1.69386717594)*dim
        ref[0,3,0,0]=(-0.392055498198)*(1+2.*(dim-1.)/(o+1.))+(-0.384588148087)*dim
        ref[0,3,0,1]=(-1.62777049183)*(1+2.*(dim-1.)/(o+1.))+(0.706370719421)*dim
        ref[0,3,1,0]=(0.619522807034)*(1+2.*(dim-1.)/(o+1.))+(-0.0583127283348)*dim
        ref[0,3,1,1]=(1.39660007067)*(1+2.*(dim-1.)/(o+1.))+(1.3874760814)*dim
        ref[0,3,2,0]=(0.396663683156)*(1+2.*(dim-1.)/(o+1.))+(-0.449036865886)*dim
        ref[0,3,2,1]=(-1.08765510225)*(1+2.*(dim-1.)/(o+1.))+(-0.941587079516)*dim
        ref[0,4,0,0]=(-1.51265971093)*(1+2.*(dim-1.)/(o+1.))+(-1.3889057145)*dim
        ref[0,4,0,1]=(0.901518176509)*(1+2.*(dim-1.)/(o+1.))+(1.13644729087)*dim
        ref[0,4,1,0]=(-0.58127466259)*(1+2.*(dim-1.)/(o+1.))+(-0.524611564706)*dim
        ref[0,4,1,1]=(1.59425336176)*(1+2.*(dim-1.)/(o+1.))+(-0.687231237871)*dim
        ref[0,4,2,0]=(1.07708364012)*(1+2.*(dim-1.)/(o+1.))+(-0.195988520187)*dim
        ref[0,4,2,1]=(0.108644035622)*(1+2.*(dim-1.)/(o+1.))+(1.0063476431)*dim
        ref[1,0,0,0]=(-0.153352577193)*(1+2.*(dim-1.)/(o+1.))+(1.30530543707)*dim
        ref[1,0,0,1]=(-0.518263005278)*(1+2.*(dim-1.)/(o+1.))+(0.192850811305)*dim
        ref[1,0,1,0]=(0.0968185470945)*(1+2.*(dim-1.)/(o+1.))+(2.07307753844)*dim
        ref[1,0,1,1]=(-0.572562275019)*(1+2.*(dim-1.)/(o+1.))+(-0.0243920749758)*dim
        ref[1,0,2,0]=(1.40149079544)*(1+2.*(dim-1.)/(o+1.))+(-0.043955038372)*dim
        ref[1,0,2,1]=(0.613246507575)*(1+2.*(dim-1.)/(o+1.))+(0.989479209244)*dim
        ref[1,1,0,0]=(-0.444118917423)*(1+2.*(dim-1.)/(o+1.))+(-1.06414293122)*dim
        ref[1,1,0,1]=(2.18067335546)*(1+2.*(dim-1.)/(o+1.))+(0.850794277187)*dim
        ref[1,1,1,0]=(0.976948683131)*(1+2.*(dim-1.)/(o+1.))+(2.1248442697)*dim
        ref[1,1,1,1]=(1.02658988987)*(1+2.*(dim-1.)/(o+1.))+(0.897405523639)*dim
        ref[1,1,2,0]=(-0.442692736232)*(1+2.*(dim-1.)/(o+1.))+(0.625813183096)*dim
        ref[1,1,2,1]=(-0.43212274342)*(1+2.*(dim-1.)/(o+1.))+(-0.899812523933)*dim
        ref[1,2,0,0]=(0.457022725817)*(1+2.*(dim-1.)/(o+1.))+(1.61784408168)*dim
        ref[1,2,0,1]=(0.643173594534)*(1+2.*(dim-1.)/(o+1.))+(-1.02505648131)*dim
        ref[1,2,1,0]=(-0.807339832268)*(1+2.*(dim-1.)/(o+1.))+(1.22702820531)*dim
        ref[1,2,1,1]=(1.42307806941)*(1+2.*(dim-1.)/(o+1.))+(0.706648854736)*dim
        ref[1,2,2,0]=(-1.02472620332)*(1+2.*(dim-1.)/(o+1.))+(-0.209029394299)*dim
        ref[1,2,2,1]=(-0.735490314485)*(1+2.*(dim-1.)/(o+1.))+(-0.786399462032)*dim
        ref[1,3,0,0]=(-0.193471598403)*(1+2.*(dim-1.)/(o+1.))+(-1.08627048726)*dim
        ref[1,3,0,1]=(0.385356410185)*(1+2.*(dim-1.)/(o+1.))+(-0.482348711502)*dim
        ref[1,3,1,0]=(0.387298963898)*(1+2.*(dim-1.)/(o+1.))+(-1.06869970751)*dim
        ref[1,3,1,1]=(0.423963900797)*(1+2.*(dim-1.)/(o+1.))+(0.451277559755)*dim
        ref[1,3,2,0]=(2.23728483025)*(1+2.*(dim-1.)/(o+1.))+(-0.857890417195)*dim
        ref[1,3,2,1]=(-0.537279589749)*(1+2.*(dim-1.)/(o+1.))+(-1.86223656189)*dim
        ref[1,4,0,0]=(0.0728532816595)*(1+2.*(dim-1.)/(o+1.))+(-0.371008451407)*dim
        ref[1,4,0,1]=(-0.159731100663)*(1+2.*(dim-1.)/(o+1.))+(0.575166562536)*dim
        ref[1,4,1,0]=(-1.07464731238)*(1+2.*(dim-1.)/(o+1.))+(1.73138566723)*dim
        ref[1,4,1,1]=(0.0871788096124)*(1+2.*(dim-1.)/(o+1.))+(-1.66991866721)*dim
        ref[1,4,2,0]=(-0.561546011624)*(1+2.*(dim-1.)/(o+1.))+(-0.571326412224)*dim
        ref[1,4,2,1]=(0.00625030418185)*(1+2.*(dim-1.)/(o+1.))+(-0.373239526059)*dim
        ref[2,0,0,0]=(-0.081748991935)*(1+2.*(dim-1.)/(o+1.))+(0.336055989523)*dim
        ref[2,0,0,1]=(-0.324440284217)*(1+2.*(dim-1.)/(o+1.))+(1.41179983739)*dim
        ref[2,0,1,0]=(-1.01122720324)*(1+2.*(dim-1.)/(o+1.))+(0.767764482116)*dim
        ref[2,0,1,1]=(1.90139553236)*(1+2.*(dim-1.)/(o+1.))+(0.0740760444209)*dim
        ref[2,0,2,0]=(1.07285550495)*(1+2.*(dim-1.)/(o+1.))+(-1.26566135985)*dim
        ref[2,0,2,1]=(1.4632067138)*(1+2.*(dim-1.)/(o+1.))+(0.109240301143)*dim
        ref[2,1,0,0]=(0.284420138548)*(1+2.*(dim-1.)/(o+1.))+(0.759494503384)*dim
        ref[2,1,0,1]=(-0.260549938527)*(1+2.*(dim-1.)/(o+1.))+(-0.189144554883)*dim
        ref[2,1,1,0]=(0.442311731675)*(1+2.*(dim-1.)/(o+1.))+(-1.09277886622)*dim
        ref[2,1,1,1]=(-0.10302967331)*(1+2.*(dim-1.)/(o+1.))+(-2.59245001471)*dim
        ref[2,1,2,0]=(0.739203737903)*(1+2.*(dim-1.)/(o+1.))+(1.55190288301)*dim
        ref[2,1,2,1]=(0.079851368615)*(1+2.*(dim-1.)/(o+1.))+(0.652318294318)*dim
        ref[2,2,0,0]=(-1.06662747958)*(1+2.*(dim-1.)/(o+1.))+(-0.689232572268)*dim
        ref[2,2,0,1]=(1.99822376512)*(1+2.*(dim-1.)/(o+1.))+(-1.16265243075)*dim
        ref[2,2,1,0]=(-1.22458303021)*(1+2.*(dim-1.)/(o+1.))+(0.678377949405)*dim
        ref[2,2,1,1]=(1.08782595263)*(1+2.*(dim-1.)/(o+1.))+(0.219802907121)*dim
        ref[2,2,2,0]=(0.905623047084)*(1+2.*(dim-1.)/(o+1.))+(0.44304843086)*dim
        ref[2,2,2,1]=(0.0286572039953)*(1+2.*(dim-1.)/(o+1.))+(-0.434052255584)*dim
        ref[2,3,0,0]=(0.498368738287)*(1+2.*(dim-1.)/(o+1.))+(-0.676248332227)*dim
        ref[2,3,0,1]=(0.0199276642636)*(1+2.*(dim-1.)/(o+1.))+(1.46838083323)*dim
        ref[2,3,1,0]=(-0.948945787905)*(1+2.*(dim-1.)/(o+1.))+(-0.363819033015)*dim
        ref[2,3,1,1]=(1.64627388129)*(1+2.*(dim-1.)/(o+1.))+(-0.372365997305)*dim
        ref[2,3,2,0]=(2.13784190571)*(1+2.*(dim-1.)/(o+1.))+(1.07491917749)*dim
        ref[2,3,2,1]=(-0.233611389691)*(1+2.*(dim-1.)/(o+1.))+(1.62428475081)*dim
        ref[2,4,0,0]=(0.739702054679)*(1+2.*(dim-1.)/(o+1.))+(0.680940307431)*dim
        ref[2,4,0,1]=(0.784529457615)*(1+2.*(dim-1.)/(o+1.))+(-1.24925786872)*dim
        ref[2,4,1,0]=(-0.583682264824)*(1+2.*(dim-1.)/(o+1.))+(-1.24709803197)*dim
        ref[2,4,1,1]=(2.09646981494)*(1+2.*(dim-1.)/(o+1.))+(-0.775641844239)*dim
        ref[2,4,2,0]=(0.259700462448)*(1+2.*(dim-1.)/(o+1.))+(-0.741083514052)*dim
        ref[2,4,2,1]=(0.997073171465)*(1+2.*(dim-1.)/(o+1.))+(-0.223654870403)*dim
        ref[3,0,0,0]=(0.444883466114)*(1+2.*(dim-1.)/(o+1.))+(-0.0463396426708)*dim
        ref[3,0,0,1]=(1.26485847247)*(1+2.*(dim-1.)/(o+1.))+(0.0735153889988)*dim
        ref[3,0,1,0]=(-0.0375651517431)*(1+2.*(dim-1.)/(o+1.))+(1.04743902526)*dim
        ref[3,0,1,1]=(0.5358978799)*(1+2.*(dim-1.)/(o+1.))+(0.206324335516)*dim
        ref[3,0,2,0]=(0.256485361704)*(1+2.*(dim-1.)/(o+1.))+(0.295375179654)*dim
        ref[3,0,2,1]=(-0.592012973207)*(1+2.*(dim-1.)/(o+1.))+(-1.61818910065)*dim
        ref[3,1,0,0]=(-0.526710008267)*(1+2.*(dim-1.)/(o+1.))+(-1.52452995876)*dim
        ref[3,1,0,1]=(1.25161847815)*(1+2.*(dim-1.)/(o+1.))+(-0.0215957412846)*dim
        ref[3,1,1,0]=(0.0535402130696)*(1+2.*(dim-1.)/(o+1.))+(1.23834558309)*dim
        ref[3,1,1,1]=(-0.642702769482)*(1+2.*(dim-1.)/(o+1.))+(-1.45938132983)*dim
        ref[3,1,2,0]=(0.0084187441016)*(1+2.*(dim-1.)/(o+1.))+(0.0982166324748)*dim
        ref[3,1,2,1]=(-0.0835608510497)*(1+2.*(dim-1.)/(o+1.))+(-0.960501832737)*dim
        ref[3,2,0,0]=(-0.233638393296)*(1+2.*(dim-1.)/(o+1.))+(-0.185542343145)*dim
        ref[3,2,0,1]=(1.83727235646)*(1+2.*(dim-1.)/(o+1.))+(-1.62951314453)*dim
        ref[3,2,1,0]=(0.141920587064)*(1+2.*(dim-1.)/(o+1.))+(-0.675548338499)*dim
        ref[3,2,1,1]=(-0.546837220172)*(1+2.*(dim-1.)/(o+1.))+(0.027220671997)*dim
        ref[3,2,2,0]=(1.08956280993)*(1+2.*(dim-1.)/(o+1.))+(-0.102196905755)*dim
        ref[3,2,2,1]=(1.37182911507)*(1+2.*(dim-1.)/(o+1.))+(-0.7777096925)*dim
        ref[3,3,0,0]=(-1.71451575877)*(1+2.*(dim-1.)/(o+1.))+(0.43149527699)*dim
        ref[3,3,0,1]=(0.125482430607)*(1+2.*(dim-1.)/(o+1.))+(0.710446642044)*dim
        ref[3,3,1,0]=(0.731418690422)*(1+2.*(dim-1.)/(o+1.))+(-1.73833071714)*dim
        ref[3,3,1,1]=(0.0925145394024)*(1+2.*(dim-1.)/(o+1.))+(-0.986375899836)*dim
        ref[3,3,2,0]=(-1.07231086612)*(1+2.*(dim-1.)/(o+1.))+(-0.436084890732)*dim
        ref[3,3,2,1]=(0.332375487809)*(1+2.*(dim-1.)/(o+1.))+(0.526342196849)*dim
        ref[3,4,0,0]=(0.730251131978)*(1+2.*(dim-1.)/(o+1.))+(-0.966347596298)*dim
        ref[3,4,0,1]=(0.235975729317)*(1+2.*(dim-1.)/(o+1.))+(-2.16543974943)*dim
        ref[3,4,1,0]=(-2.13209908673)*(1+2.*(dim-1.)/(o+1.))+(0.631593299244)*dim
        ref[3,4,1,1]=(-0.257512397822)*(1+2.*(dim-1.)/(o+1.))+(-1.4109264038)*dim
        ref[3,4,2,0]=(0.201086569751)*(1+2.*(dim-1.)/(o+1.))+(0.626261609021)*dim
        ref[3,4,2,1]=(0.819600430724)*(1+2.*(dim-1.)/(o+1.))+(-0.483576060183)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(-1.08871335955)*x[0]+(0.247972631031)*x[1]
        ref=(-0.156403544195)*(1+2.*(dim-1.)/(o+1.))+(-0.684337184327)*dim
      else:
        s=(-0.324052151044)*x[0]+(0.441305482424)*x[1]+(-0.315030176112)*x[2]
        ref=(0.603955649036)*(1+2.*(dim-1.)/(o+1.))+(-0.801732493769)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(-0.260677248137)*x[0]+(0.0211318398374)*x[1]
        s[1]=(-0.815879139039)*x[0]+(-0.918304389572)*x[1]
        ref[0]=(0.843100081089)*(1+2.*(dim-1.)/(o+1.))+(-1.08264548939)*dim
        ref[1]=(-1.3080191481)*(1+2.*(dim-1.)/(o+1.))+(-0.426164380506)*dim
      else:
        s[0]=(1.37908771986)*x[0]+(1.36518233103)*x[1]+(1.013246401)*x[2]
        s[1]=(0.556318683467)*x[0]+(1.45737568352)*x[1]+(0.0165078724907)*x[2]
        ref[0]=(2.05950634594)*(1+2.*(dim-1.)/(o+1.))+(1.69801010595)*dim
        ref[1]=(0.655472308826)*(1+2.*(dim-1.)/(o+1.))+(1.37472993065)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(1.44204352077)*x[0]+(-0.626836867163)*x[1]
        s[0,1]=(-0.295593397487)*x[0]+(-0.309954518607)*x[1]
        s[0,2]=(-0.546882380107)*x[0]+(-0.0925530562931)*x[1]
        s[0,3]=(-0.623335335636)*x[0]+(0.798887138514)*x[1]
        s[0,4]=(0.202234634868)*x[0]+(1.44014401659)*x[1]
        s[1,0]=(-0.772178934874)*x[0]+(-0.962339450381)*x[1]
        s[1,1]=(-0.136928685359)*x[0]+(0.385217139725)*x[1]
        s[1,2]=(-0.00807747505852)*x[0]+(1.08230267787)*x[1]
        s[1,3]=(-1.28422033803)*x[0]+(0.011228260616)*x[1]
        s[1,4]=(0.766746911823)*x[0]+(-0.145254490276)*x[1]
        s[2,0]=(1.65171936301)*x[0]+(1.68030971997)*x[1]
        s[2,1]=(-0.465968023499)*x[0]+(-0.594634053719)*x[1]
        s[2,2]=(-0.140916975546)*x[0]+(0.230405132955)*x[1]
        s[2,3]=(-0.0185338227586)*x[0]+(0.298041051546)*x[1]
        s[2,4]=(-0.024340230778)*x[0]+(-0.791873643457)*x[1]
        s[3,0]=(0.555336313403)*x[0]+(0.606898490822)*x[1]
        s[3,1]=(1.05785941408)*x[0]+(-0.672561947115)*x[1]
        s[3,2]=(1.36138187068)*x[0]+(0.259592704164)*x[1]
        s[3,3]=(0.66211210256)*x[0]+(-1.38753456619)*x[1]
        s[3,4]=(0.0789301309156)*x[0]+(1.24015676345)*x[1]
        ref[0,0]=(0.863657572901)*(1+2.*(dim-1.)/(o+1.))+(-0.0484509192919)*dim
        ref[0,1]=(-0.0608826851336)*(1+2.*(dim-1.)/(o+1.))+(-0.54466523096)*dim
        ref[0,2]=(-1.25630630361)*(1+2.*(dim-1.)/(o+1.))+(0.616870867215)*dim
        ref[0,3]=(-0.618026054387)*(1+2.*(dim-1.)/(o+1.))+(0.793577857265)*dim
        ref[0,4]=(1.55309725329)*(1+2.*(dim-1.)/(o+1.))+(0.0892813981594)*dim
        ref[1,0]=(-0.707427946889)*(1+2.*(dim-1.)/(o+1.))+(-1.02709043837)*dim
        ref[1,1]=(1.0516700664)*(1+2.*(dim-1.)/(o+1.))+(-0.803381612038)*dim
        ref[1,2]=(-0.251691185365)*(1+2.*(dim-1.)/(o+1.))+(1.32591638818)*dim
        ref[1,3]=(0.227444315739)*(1+2.*(dim-1.)/(o+1.))+(-1.50043639316)*dim
        ref[1,4]=(0.636926531504)*(1+2.*(dim-1.)/(o+1.))+(-0.0154341099568)*dim
        ref[2,0]=(1.58333782129)*(1+2.*(dim-1.)/(o+1.))+(1.7486912617)*dim
        ref[2,1]=(0.376537776448)*(1+2.*(dim-1.)/(o+1.))+(-1.43713985367)*dim
        ref[2,2]=(1.33315471883)*(1+2.*(dim-1.)/(o+1.))+(-1.24366656142)*dim
        ref[2,3]=(-0.893476349046)*(1+2.*(dim-1.)/(o+1.))+(1.17298357783)*dim
        ref[2,4]=(-0.556720283569)*(1+2.*(dim-1.)/(o+1.))+(-0.259493590666)*dim
        ref[3,0]=(-0.0344322176845)*(1+2.*(dim-1.)/(o+1.))+(1.19666702191)*dim
        ref[3,1]=(0.356058306775)*(1+2.*(dim-1.)/(o+1.))+(0.0292391601866)*dim
        ref[3,2]=(1.18926653697)*(1+2.*(dim-1.)/(o+1.))+(0.431708037874)*dim
        ref[3,3]=(-0.0855874137907)*(1+2.*(dim-1.)/(o+1.))+(-0.639835049842)*dim
        ref[3,4]=(0.485696084789)*(1+2.*(dim-1.)/(o+1.))+(0.833390809579)*dim
      else:
        s[0,0]=(1.48740802799)*x[0]+(-0.5231718635)*x[1]+(0.621553689326)*x[2]
        s[0,1]=(-0.816432702835)*x[0]+(-1.02472113234)*x[1]+(-0.498250552797)*x[2]
        s[0,2]=(-0.0547881618906)*x[0]+(0.903306651062)*x[1]+(0.471428097681)*x[2]
        s[0,3]=(-0.910642152807)*x[0]+(-0.958224215021)*x[1]+(-1.4403271378)*x[2]
        s[0,4]=(0.116198234381)*x[0]+(-0.636378388418)*x[1]+(0.1270635999)*x[2]
        s[1,0]=(-0.659576291097)*x[0]+(-0.446587669482)*x[1]+(1.38617928144)*x[2]
        s[1,1]=(0.442922554555)*x[0]+(0.2841566742)*x[1]+(-0.384854418521)*x[2]
        s[1,2]=(-0.883255151971)*x[0]+(-1.3015449692)*x[1]+(-1.49785758846)*x[2]
        s[1,3]=(-1.43150392129)*x[0]+(1.13296295364)*x[1]+(0.393410729481)*x[2]
        s[1,4]=(-0.423254486419)*x[0]+(1.58675432746)*x[1]+(1.46179069555)*x[2]
        s[2,0]=(0.467182623429)*x[0]+(-1.23530058455)*x[1]+(0.546258172921)*x[2]
        s[2,1]=(-1.2468563949)*x[0]+(1.05285375014)*x[1]+(1.00605876886)*x[2]
        s[2,2]=(-0.1261124635)*x[0]+(-0.855000687125)*x[1]+(-0.252904511506)*x[2]
        s[2,3]=(1.1950830615)*x[0]+(0.699000137772)*x[1]+(1.33793802349)*x[2]
        s[2,4]=(-1.26581081559)*x[0]+(-0.0366783527385)*x[1]+(-0.736592212733)*x[2]
        s[3,0]=(1.84878865394)*x[0]+(0.504454193662)*x[1]+(-1.15690555199)*x[2]
        s[3,1]=(-0.249497283612)*x[0]+(-0.635940805715)*x[1]+(0.168382197376)*x[2]
        s[3,2]=(-0.167328351962)*x[0]+(0.523012476804)*x[1]+(1.19870547556)*x[2]
        s[3,3]=(-0.499150198998)*x[0]+(0.448461159266)*x[1]+(0.554069089559)*x[2]
        s[3,4]=(-0.520146860421)*x[0]+(1.43515582204)*x[1]+(1.31685474485)*x[2]
        ref[0,0]=(1.56352587364)*(1+2.*(dim-1.)/(o+1.))+(0.0222639801801)*dim
        ref[0,1]=(-1.88354672396)*(1+2.*(dim-1.)/(o+1.))+(-0.455857664013)*dim
        ref[0,2]=(0.631897841337)*(1+2.*(dim-1.)/(o+1.))+(0.688048745515)*dim
        ref[0,3]=(-2.49358987828)*(1+2.*(dim-1.)/(o+1.))+(-0.815603627346)*dim
        ref[0,4]=(-1.43782849721)*(1+2.*(dim-1.)/(o+1.))+(1.04471194307)*dim
        ref[1,0]=(-0.257186244728)*(1+2.*(dim-1.)/(o+1.))+(0.537201565586)*dim
        ref[1,1]=(-1.21753131847)*(1+2.*(dim-1.)/(o+1.))+(1.55975612871)*dim
        ref[1,2]=(-1.83391505519)*(1+2.*(dim-1.)/(o+1.))+(-1.84874265444)*dim
        ref[1,3]=(0.790398824851)*(1+2.*(dim-1.)/(o+1.))+(-0.69552906302)*dim
        ref[1,4]=(1.96838334672)*(1+2.*(dim-1.)/(o+1.))+(0.656907189867)*dim
        ref[2,0]=(-0.486341534759)*(1+2.*(dim-1.)/(o+1.))+(0.264481746559)*dim
        ref[2,1]=(-0.417272431595)*(1+2.*(dim-1.)/(o+1.))+(1.22932855569)*dim
        ref[2,2]=(-0.881184195775)*(1+2.*(dim-1.)/(o+1.))+(-0.352833466355)*dim
        ref[2,3]=(1.01420597285)*(1+2.*(dim-1.)/(o+1.))+(2.21781524991)*dim
        ref[2,4]=(-0.580092347401)*(1+2.*(dim-1.)/(o+1.))+(-1.45898903366)*dim
        ref[3,0]=(0.865857995406)*(1+2.*(dim-1.)/(o+1.))+(0.330479300213)*dim
        ref[3,1]=(0.632704550785)*(1+2.*(dim-1.)/(o+1.))+(-1.34976044274)*dim
        ref[3,2]=(-0.259081469822)*(1+2.*(dim-1.)/(o+1.))+(1.81347107022)*dim
        ref[3,3]=(0.533963525526)*(1+2.*(dim-1.)/(o+1.))+(-0.0305834756989)*dim
        ref[3,4]=(1.19847134659)*(1+2.*(dim-1.)/(o+1.))+(1.03339235988)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.536617774033)*x[0]+(-1.31191068914)*x[1]
        s[0,0,1]=(1.45706261992)*x[0]+(0.33760400949)*x[1]
        s[0,1,0]=(-0.453569968232)*x[0]+(0.798108537393)*x[1]
        s[0,1,1]=(-1.26956398588)*x[0]+(-1.35096551318)*x[1]
        s[1,0,0]=(0.104407588284)*x[0]+(-0.649935519811)*x[1]
        s[1,0,1]=(-1.64047091856)*x[0]+(0.654330987243)*x[1]
        s[1,1,0]=(-0.161010086142)*x[0]+(-0.644341383353)*x[1]
        s[1,1,1]=(1.42345151896)*x[0]+(-0.963960533124)*x[1]
        s[2,0,0]=(0.069463733935)*x[0]+(-0.764246138958)*x[1]
        s[2,0,1]=(0.516672615671)*x[0]+(0.32129769972)*x[1]
        s[2,1,0]=(0.399513997078)*x[0]+(-1.76803609001)*x[1]
        s[2,1,1]=(0.0570653989294)*x[0]+(-0.191869561254)*x[1]
        s[3,0,0]=(-0.261019396242)*x[0]+(0.429544509966)*x[1]
        s[3,0,1]=(-1.40560961029)*x[0]+(-0.739182555238)*x[1]
        s[3,1,0]=(0.314850928569)*x[0]+(1.49030138742)*x[1]
        s[3,1,1]=(-1.27327662424)*x[0]+(-0.0287977839083)*x[1]
        s[4,0,0]=(-0.840541224777)*x[0]+(-0.805224659169)*x[1]
        s[4,0,1]=(0.303904176581)*x[0]+(0.353767617246)*x[1]
        s[4,1,0]=(1.65813306507)*x[0]+(1.12299575058)*x[1]
        s[4,1,1]=(-1.19199290034)*x[0]+(1.6581285511)*x[1]
        s[5,0,0]=(0.120283505072)*x[0]+(1.68044908685)*x[1]
        s[5,0,1]=(0.382489278089)*x[0]+(0.129887092126)*x[1]
        s[5,1,0]=(-1.01890077156)*x[0]+(-0.111140290699)*x[1]
        s[5,1,1]=(-0.0392100253378)*x[0]+(0.579951319719)*x[1]
        ref[0,0,0]=(-0.686284158891)*(1+2.*(dim-1.)/(o+1.))+(-0.089008756216)*dim
        ref[0,0,1]=(1.76417819838)*(1+2.*(dim-1.)/(o+1.))+(0.030488431025)*dim
        ref[0,1,0]=(0.453381498643)*(1+2.*(dim-1.)/(o+1.))+(-0.108842929482)*dim
        ref[0,1,1]=(-1.03607416678)*(1+2.*(dim-1.)/(o+1.))+(-1.58445533227)*dim
        ref[1,0,0]=(0.638386446298)*(1+2.*(dim-1.)/(o+1.))+(-1.18391437783)*dim
        ref[1,0,1]=(0.275023590147)*(1+2.*(dim-1.)/(o+1.))+(-1.26116352146)*dim
        ref[1,1,0]=(-0.869379328882)*(1+2.*(dim-1.)/(o+1.))+(0.0640278593865)*dim
        ref[1,1,1]=(0.597799121533)*(1+2.*(dim-1.)/(o+1.))+(-0.138308135699)*dim
        ref[2,0,0]=(-1.45715684308)*(1+2.*(dim-1.)/(o+1.))+(0.762374438053)*dim
        ref[2,0,1]=(0.929718948925)*(1+2.*(dim-1.)/(o+1.))+(-0.0917486335329)*dim
        ref[2,1,0]=(-0.78092454297)*(1+2.*(dim-1.)/(o+1.))+(-0.587597549963)*dim
        ref[2,1,1]=(-1.21253063514)*(1+2.*(dim-1.)/(o+1.))+(1.07772647282)*dim
        ref[3,0,0]=(1.07007881303)*(1+2.*(dim-1.)/(o+1.))+(-0.901553699308)*dim
        ref[3,0,1]=(-1.77159657093)*(1+2.*(dim-1.)/(o+1.))+(-0.373195594594)*dim
        ref[3,1,0]=(1.10443227539)*(1+2.*(dim-1.)/(o+1.))+(0.700720040595)*dim
        ref[3,1,1]=(0.315983219905)*(1+2.*(dim-1.)/(o+1.))+(-1.61805762805)*dim
        ref[4,0,0]=(-0.602218266612)*(1+2.*(dim-1.)/(o+1.))+(-1.04354761733)*dim
        ref[4,0,1]=(0.297276459441)*(1+2.*(dim-1.)/(o+1.))+(0.360395334386)*dim
        ref[4,1,0]=(1.5584687424)*(1+2.*(dim-1.)/(o+1.))+(1.22266007325)*dim
        ref[4,1,1]=(0.504270780151)*(1+2.*(dim-1.)/(o+1.))+(-0.038135129389)*dim
        ref[5,0,0]=(1.35289487036)*(1+2.*(dim-1.)/(o+1.))+(0.44783772157)*dim
        ref[5,0,1]=(1.57660464714)*(1+2.*(dim-1.)/(o+1.))+(-1.06422827692)*dim
        ref[5,1,0]=(0.1360111664)*(1+2.*(dim-1.)/(o+1.))+(-1.26605222866)*dim
        ref[5,1,1]=(0.024986479373)*(1+2.*(dim-1.)/(o+1.))+(0.515754815008)*dim
      else:
        s[0,0,0]=(-0.491953157793)*x[0]+(0.808487399764)*x[1]+(0.87783526238)*x[2]
        s[0,0,1]=(-0.470158989175)*x[0]+(-0.492379274826)*x[1]+(-0.191290328078)*x[2]
        s[0,1,0]=(-0.356351922052)*x[0]+(0.823887078075)*x[1]+(-0.177422034195)*x[2]
        s[0,1,1]=(-0.378759483085)*x[0]+(0.116067213963)*x[1]+(0.0288265220551)*x[2]
        s[1,0,0]=(-0.325670782382)*x[0]+(-1.01289817775)*x[1]+(0.672608800327)*x[2]
        s[1,0,1]=(-0.296979811719)*x[0]+(0.46777334672)*x[1]+(1.61350930522)*x[2]
        s[1,1,0]=(-0.0574537817174)*x[0]+(1.10617918343)*x[1]+(0.172222519609)*x[2]
        s[1,1,1]=(-0.978413596402)*x[0]+(-1.20699880678)*x[1]+(-0.0600246637942)*x[2]
        s[2,0,0]=(0.637900823637)*x[0]+(1.29751687672)*x[1]+(-0.390955179963)*x[2]
        s[2,0,1]=(-1.15596339731)*x[0]+(-1.04131524409)*x[1]+(0.111161978383)*x[2]
        s[2,1,0]=(-1.33713877493)*x[0]+(0.0677540624633)*x[1]+(0.0662976219802)*x[2]
        s[2,1,1]=(0.0636656461261)*x[0]+(-1.32776475079)*x[1]+(0.365594993394)*x[2]
        s[3,0,0]=(0.454007484649)*x[0]+(-1.56273676656)*x[1]+(-0.783612553822)*x[2]
        s[3,0,1]=(-0.908388718727)*x[0]+(0.496552922533)*x[1]+(1.0802957835)*x[2]
        s[3,1,0]=(0.46784249418)*x[0]+(1.627686011)*x[1]+(1.12464788868)*x[2]
        s[3,1,1]=(1.47169416804)*x[0]+(-0.541043548439)*x[1]+(-0.0203908370923)*x[2]
        s[4,0,0]=(0.153593270409)*x[0]+(0.991271922294)*x[1]+(0.910688732568)*x[2]
        s[4,0,1]=(1.26666528018)*x[0]+(-0.0415946389155)*x[1]+(-0.702890723051)*x[2]
        s[4,1,0]=(0.401000008974)*x[0]+(0.270202736778)*x[1]+(-1.52334516556)*x[2]
        s[4,1,1]=(-0.567247526123)*x[0]+(0.0638484526058)*x[1]+(-0.942447437378)*x[2]
        s[5,0,0]=(1.3141482761)*x[0]+(0.602038495282)*x[1]+(-0.162794144191)*x[2]
        s[5,0,1]=(1.49343941166)*x[0]+(0.340106070028)*x[1]+(0.222963641278)*x[2]
        s[5,1,0]=(1.29779085951)*x[0]+(0.44436851782)*x[1]+(-0.414930735066)*x[2]
        s[5,1,1]=(0.445071462828)*x[0]+(0.0113100435192)*x[1]+(-0.304892738875)*x[2]
        ref[0,0,0]=(0.807154905774)*(1+2.*(dim-1.)/(o+1.))+(0.387214598576)*dim
        ref[0,0,1]=(-0.168668456531)*(1+2.*(dim-1.)/(o+1.))+(-0.985160135549)*dim
        ref[0,1,0]=(-0.405021627391)*(1+2.*(dim-1.)/(o+1.))+(0.69513474922)*dim
        ref[0,1,1]=(0.481047411067)*(1+2.*(dim-1.)/(o+1.))+(-0.714913158134)*dim
        ref[1,0,0]=(-0.241963228819)*(1+2.*(dim-1.)/(o+1.))+(-0.423996930987)*dim
        ref[1,0,1]=(-0.222173232945)*(1+2.*(dim-1.)/(o+1.))+(2.00647607317)*dim
        ref[1,1,0]=(2.08927091112)*(1+2.*(dim-1.)/(o+1.))+(-0.868322989801)*dim
        ref[1,1,1]=(-1.2199872069)*(1+2.*(dim-1.)/(o+1.))+(-1.02544986008)*dim
        ref[2,0,0]=(1.51358492234)*(1+2.*(dim-1.)/(o+1.))+(0.0308775980534)*dim
        ref[2,0,1]=(-1.17419262019)*(1+2.*(dim-1.)/(o+1.))+(-0.911924042825)*dim
        ref[2,1,0]=(-0.808827279389)*(1+2.*(dim-1.)/(o+1.))+(-0.394259811094)*dim
        ref[2,1,1]=(-0.39434462087)*(1+2.*(dim-1.)/(o+1.))+(-0.504159490404)*dim
        ref[3,0,0]=(-0.420509147061)*(1+2.*(dim-1.)/(o+1.))+(-1.47183268867)*dim
        ref[3,0,1]=(-0.509702954626)*(1+2.*(dim-1.)/(o+1.))+(1.17816294193)*dim
        ref[3,1,0]=(2.45580213911)*(1+2.*(dim-1.)/(o+1.))+(0.764374254754)*dim
        ref[3,1,1]=(0.36148972446)*(1+2.*(dim-1.)/(o+1.))+(0.548770058051)*dim
        ref[4,0,0]=(0.723463115757)*(1+2.*(dim-1.)/(o+1.))+(1.33209080951)*dim
        ref[4,0,1]=(-0.584660571656)*(1+2.*(dim-1.)/(o+1.))+(1.10684048987)*dim
        ref[4,1,0]=(0.558877069737)*(1+2.*(dim-1.)/(o+1.))+(-1.41101948955)*dim
        ref[4,1,1]=(-1.46224923696)*(1+2.*(dim-1.)/(o+1.))+(0.016402726067)*dim
        ref[5,0,0]=(0.375824940974)*(1+2.*(dim-1.)/(o+1.))+(1.37756768621)*dim
        ref[5,0,1]=(0.0790629051458)*(1+2.*(dim-1.)/(o+1.))+(1.97744621782)*dim
        ref[5,1,0]=(-0.429416156192)*(1+2.*(dim-1.)/(o+1.))+(1.75664479845)*dim
        ref[5,1,1]=(-0.227051313962)*(1+2.*(dim-1.)/(o+1.))+(0.378540081434)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-1.24466004429)*x[0]+(0.249656583863)*x[1]
        s[0,0,0,1]=(-1.31303014286)*x[0]+(-0.188187818679)*x[1]
        s[0,0,1,0]=(-1.19686759063)*x[0]+(0.202998667098)*x[1]
        s[0,0,1,1]=(-0.240139422625)*x[0]+(1.12145727417)*x[1]
        s[0,0,2,0]=(-0.779228056931)*x[0]+(-0.256394958159)*x[1]
        s[0,0,2,1]=(0.511346899235)*x[0]+(-1.17117401759)*x[1]
        s[0,1,0,0]=(1.20931773874)*x[0]+(0.703848713877)*x[1]
        s[0,1,0,1]=(1.05396707864)*x[0]+(0.561081392967)*x[1]
        s[0,1,1,0]=(-0.940092412899)*x[0]+(0.197879835766)*x[1]
        s[0,1,1,1]=(0.031033640344)*x[0]+(1.47668227837)*x[1]
        s[0,1,2,0]=(-0.973745458182)*x[0]+(-0.699999591381)*x[1]
        s[0,1,2,1]=(0.0637513252131)*x[0]+(0.223369875142)*x[1]
        s[0,2,0,0]=(-0.179817828717)*x[0]+(-1.0618817379)*x[1]
        s[0,2,0,1]=(-1.31848741633)*x[0]+(0.562871230704)*x[1]
        s[0,2,1,0]=(0.20660993199)*x[0]+(-0.260401741099)*x[1]
        s[0,2,1,1]=(-0.119752579601)*x[0]+(1.03659426931)*x[1]
        s[0,2,2,0]=(-1.05257053872)*x[0]+(-0.410999368732)*x[1]
        s[0,2,2,1]=(0.142472752971)*x[0]+(0.687099054848)*x[1]
        s[0,3,0,0]=(-0.735958436285)*x[0]+(0.576103117069)*x[1]
        s[0,3,0,1]=(0.820759477791)*x[0]+(1.40201775771)*x[1]
        s[0,3,1,0]=(-1.43663163172)*x[0]+(-0.928562287497)*x[1]
        s[0,3,1,1]=(-0.640592604343)*x[0]+(-1.37951764035)*x[1]
        s[0,3,2,0]=(0.923170161037)*x[0]+(-1.29865735065)*x[1]
        s[0,3,2,1]=(-1.49185366667)*x[0]+(1.26235954659)*x[1]
        s[0,4,0,0]=(-1.54260726706)*x[0]+(0.386401532954)*x[1]
        s[0,4,0,1]=(0.938275027884)*x[0]+(-0.325978711858)*x[1]
        s[0,4,1,0]=(-0.741422170549)*x[0]+(0.360529257608)*x[1]
        s[0,4,1,1]=(-1.38867328824)*x[0]+(0.368979718783)*x[1]
        s[0,4,2,0]=(-1.19633228454)*x[0]+(0.143128822342)*x[1]
        s[0,4,2,1]=(0.365590220489)*x[0]+(0.702852400998)*x[1]
        s[1,0,0,0]=(0.590708718523)*x[0]+(-1.74588023915)*x[1]
        s[1,0,0,1]=(-0.411269052638)*x[0]+(0.677504850422)*x[1]
        s[1,0,1,0]=(1.54641105105)*x[0]+(-0.227798671895)*x[1]
        s[1,0,1,1]=(0.206986074762)*x[0]+(0.431911634342)*x[1]
        s[1,0,2,0]=(0.180245950747)*x[0]+(-0.766496827464)*x[1]
        s[1,0,2,1]=(0.920490611623)*x[0]+(0.146274601957)*x[1]
        s[1,1,0,0]=(-1.47489897228)*x[0]+(-1.0428187509)*x[1]
        s[1,1,0,1]=(-1.17774302127)*x[0]+(0.44418244557)*x[1]
        s[1,1,1,0]=(-0.340268939926)*x[0]+(-0.362308449805)*x[1]
        s[1,1,1,1]=(1.1122568876)*x[0]+(0.807887213986)*x[1]
        s[1,1,2,0]=(-1.1580670758)*x[0]+(0.61015974696)*x[1]
        s[1,1,2,1]=(-0.21738296838)*x[0]+(-1.15287881926)*x[1]
        s[1,2,0,0]=(-1.63285965164)*x[0]+(0.546571504115)*x[1]
        s[1,2,0,1]=(1.34627844359)*x[0]+(0.0153020013782)*x[1]
        s[1,2,1,0]=(-0.459380401764)*x[0]+(0.653971151372)*x[1]
        s[1,2,1,1]=(0.0918961369392)*x[0]+(0.631274394662)*x[1]
        s[1,2,2,0]=(-0.461548096839)*x[0]+(0.432322561401)*x[1]
        s[1,2,2,1]=(-0.741897062249)*x[0]+(-0.0397725689655)*x[1]
        s[1,3,0,0]=(-0.166826966339)*x[0]+(0.76157630088)*x[1]
        s[1,3,0,1]=(-1.43001526663)*x[0]+(0.0200445726731)*x[1]
        s[1,3,1,0]=(0.776640170933)*x[0]+(1.77522529504)*x[1]
        s[1,3,1,1]=(0.132999332599)*x[0]+(0.91396296202)*x[1]
        s[1,3,2,0]=(-0.435447499973)*x[0]+(-0.714956349315)*x[1]
        s[1,3,2,1]=(1.54444404505)*x[0]+(0.0419015202674)*x[1]
        s[1,4,0,0]=(-1.14197604256)*x[0]+(0.140549307511)*x[1]
        s[1,4,0,1]=(1.11080451267)*x[0]+(0.140105909725)*x[1]
        s[1,4,1,0]=(0.361572095315)*x[0]+(1.73141141151)*x[1]
        s[1,4,1,1]=(1.2968959883)*x[0]+(0.459397526479)*x[1]
        s[1,4,2,0]=(-1.04477229488)*x[0]+(-0.432251933613)*x[1]
        s[1,4,2,1]=(0.166352074158)*x[0]+(0.240538931271)*x[1]
        s[2,0,0,0]=(-0.302792258664)*x[0]+(0.014159412453)*x[1]
        s[2,0,0,1]=(0.161287860507)*x[0]+(-1.66699057701)*x[1]
        s[2,0,1,0]=(0.793079282708)*x[0]+(0.999374748613)*x[1]
        s[2,0,1,1]=(-0.0860624573943)*x[0]+(0.438101549788)*x[1]
        s[2,0,2,0]=(-0.832685279306)*x[0]+(1.20735467656)*x[1]
        s[2,0,2,1]=(-0.0628631179915)*x[0]+(0.720123606355)*x[1]
        s[2,1,0,0]=(0.275852566161)*x[0]+(-0.259595395227)*x[1]
        s[2,1,0,1]=(0.862309966294)*x[0]+(-0.45114296596)*x[1]
        s[2,1,1,0]=(0.14823103051)*x[0]+(-1.16121025546)*x[1]
        s[2,1,1,1]=(-1.34419311122)*x[0]+(1.63903760286)*x[1]
        s[2,1,2,0]=(1.18830622559)*x[0]+(0.985255273279)*x[1]
        s[2,1,2,1]=(-0.0208912272649)*x[0]+(0.543921272375)*x[1]
        s[2,2,0,0]=(0.510821411306)*x[0]+(-1.88592356399)*x[1]
        s[2,2,0,1]=(0.353203575328)*x[0]+(0.328922615441)*x[1]
        s[2,2,1,0]=(1.1536669592)*x[0]+(-1.05621607534)*x[1]
        s[2,2,1,1]=(-0.581456156893)*x[0]+(-0.20275776476)*x[1]
        s[2,2,2,0]=(1.23883996503)*x[0]+(0.577857478009)*x[1]
        s[2,2,2,1]=(0.938162394147)*x[0]+(1.74792706902)*x[1]
        s[2,3,0,0]=(-0.0610369156925)*x[0]+(0.316929103344)*x[1]
        s[2,3,0,1]=(0.115575825651)*x[0]+(0.196603816526)*x[1]
        s[2,3,1,0]=(-0.137564893251)*x[0]+(0.732741603189)*x[1]
        s[2,3,1,1]=(-0.659475613725)*x[0]+(0.126037736276)*x[1]
        s[2,3,2,0]=(1.1911196497)*x[0]+(0.86409088069)*x[1]
        s[2,3,2,1]=(-0.161576876539)*x[0]+(-0.519953247817)*x[1]
        s[2,4,0,0]=(0.653018736743)*x[0]+(-0.141460723253)*x[1]
        s[2,4,0,1]=(0.548072342723)*x[0]+(0.79847118191)*x[1]
        s[2,4,1,0]=(-0.894497169389)*x[0]+(0.358296451187)*x[1]
        s[2,4,1,1]=(0.104304124331)*x[0]+(0.229557340588)*x[1]
        s[2,4,2,0]=(-0.775504740819)*x[0]+(-0.914276983707)*x[1]
        s[2,4,2,1]=(-0.997444443952)*x[0]+(-0.550036262287)*x[1]
        s[3,0,0,0]=(-0.912968382437)*x[0]+(1.34668077004)*x[1]
        s[3,0,0,1]=(0.279466090703)*x[0]+(1.28107040706)*x[1]
        s[3,0,1,0]=(0.282808913079)*x[0]+(0.740167690692)*x[1]
        s[3,0,1,1]=(-1.39563519985)*x[0]+(-0.567913880569)*x[1]
        s[3,0,2,0]=(1.27572544936)*x[0]+(-0.733366778512)*x[1]
        s[3,0,2,1]=(-0.33162427954)*x[0]+(1.35270158006)*x[1]
        s[3,1,0,0]=(0.519174606536)*x[0]+(-0.372514377255)*x[1]
        s[3,1,0,1]=(-1.58036427028)*x[0]+(-0.235983616334)*x[1]
        s[3,1,1,0]=(-0.472835600722)*x[0]+(0.721357411442)*x[1]
        s[3,1,1,1]=(0.448313914421)*x[0]+(-0.138520459243)*x[1]
        s[3,1,2,0]=(-1.51018963552)*x[0]+(1.34139211207)*x[1]
        s[3,1,2,1]=(-0.212789438989)*x[0]+(-1.06575114519)*x[1]
        s[3,2,0,0]=(0.0592065335067)*x[0]+(0.359601543991)*x[1]
        s[3,2,0,1]=(-0.594769346579)*x[0]+(1.5491276729)*x[1]
        s[3,2,1,0]=(0.28455993713)*x[0]+(0.600972748948)*x[1]
        s[3,2,1,1]=(0.398577473206)*x[0]+(0.0479286078109)*x[1]
        s[3,2,2,0]=(-1.76127884396)*x[0]+(-0.538814316335)*x[1]
        s[3,2,2,1]=(-0.390841519342)*x[0]+(0.191259339024)*x[1]
        s[3,3,0,0]=(-0.877110189074)*x[0]+(0.301682318677)*x[1]
        s[3,3,0,1]=(0.112669606669)*x[0]+(1.0395312515)*x[1]
        s[3,3,1,0]=(1.04370247351)*x[0]+(0.581250871082)*x[1]
        s[3,3,1,1]=(-0.143233022608)*x[0]+(0.593143844478)*x[1]
        s[3,3,2,0]=(-0.304532308332)*x[0]+(-0.232727140662)*x[1]
        s[3,3,2,1]=(0.767141825929)*x[0]+(0.0401997820135)*x[1]
        s[3,4,0,0]=(-0.226100649415)*x[0]+(-0.656022574403)*x[1]
        s[3,4,0,1]=(0.679160670663)*x[0]+(0.507720591962)*x[1]
        s[3,4,1,0]=(-1.30886998647)*x[0]+(-1.3539695469)*x[1]
        s[3,4,1,1]=(-1.39638819903)*x[0]+(0.650428829094)*x[1]
        s[3,4,2,0]=(-0.346486286718)*x[0]+(-0.869112300588)*x[1]
        s[3,4,2,1]=(-1.27650086562)*x[0]+(0.660506644039)*x[1]
        ref[0,0,0,0]=(-0.625690911818)*(1+2.*(dim-1.)/(o+1.))+(-0.369312548605)*dim
        ref[0,0,0,1]=(-0.0186378038077)*(1+2.*(dim-1.)/(o+1.))+(-1.48258015773)*dim
        ref[0,0,1,0]=(-1.04194152975)*(1+2.*(dim-1.)/(o+1.))+(0.0480726062254)*dim
        ref[0,0,1,1]=(0.00120841586244)*(1+2.*(dim-1.)/(o+1.))+(0.880109435685)*dim
        ref[0,0,2,0]=(-0.65182866689)*(1+2.*(dim-1.)/(o+1.))+(-0.3837943482)*dim
        ref[0,0,2,1]=(-0.729868010003)*(1+2.*(dim-1.)/(o+1.))+(0.0700408916461)*dim
        ref[0,1,0,0]=(1.26484534964)*(1+2.*(dim-1.)/(o+1.))+(0.648321102977)*dim
        ref[0,1,0,1]=(0.334654674402)*(1+2.*(dim-1.)/(o+1.))+(1.28039379721)*dim
        ref[0,1,1,0]=(0.338629227563)*(1+2.*(dim-1.)/(o+1.))+(-1.0808418047)*dim
        ref[0,1,1,1]=(0.0969508871625)*(1+2.*(dim-1.)/(o+1.))+(1.41076503155)*dim
        ref[0,1,2,0]=(-0.462969154119)*(1+2.*(dim-1.)/(o+1.))+(-1.21077589544)*dim
        ref[0,1,2,1]=(0.46756493995)*(1+2.*(dim-1.)/(o+1.))+(-0.180443739595)*dim
        ref[0,2,0,0]=(-0.0576220389279)*(1+2.*(dim-1.)/(o+1.))+(-1.18407752769)*dim
        ref[0,2,0,1]=(-0.650282644944)*(1+2.*(dim-1.)/(o+1.))+(-0.105333540684)*dim
        ref[0,2,1,0]=(0.136853100987)*(1+2.*(dim-1.)/(o+1.))+(-0.190644910096)*dim
        ref[0,2,1,1]=(-0.637797927534)*(1+2.*(dim-1.)/(o+1.))+(1.55463961724)*dim
        ref[0,2,2,0]=(-0.455136385489)*(1+2.*(dim-1.)/(o+1.))+(-1.00843352196)*dim
        ref[0,2,2,1]=(-0.35303452486)*(1+2.*(dim-1.)/(o+1.))+(1.18260633268)*dim
        ref[0,3,0,0]=(-0.315687014208)*(1+2.*(dim-1.)/(o+1.))+(0.155831694992)*dim
        ref[0,3,0,1]=(1.52158435346)*(1+2.*(dim-1.)/(o+1.))+(0.701192882048)*dim
        ref[0,3,1,0]=(-0.836739848177)*(1+2.*(dim-1.)/(o+1.))+(-1.52845407104)*dim
        ref[0,3,1,1]=(-0.983041539851)*(1+2.*(dim-1.)/(o+1.))+(-1.03706870484)*dim
        ref[0,3,2,0]=(-0.497690471912)*(1+2.*(dim-1.)/(o+1.))+(0.122203282296)*dim
        ref[0,3,2,1]=(0.394951943207)*(1+2.*(dim-1.)/(o+1.))+(-0.624446063287)*dim
        ref[0,4,0,0]=(-0.63974792726)*(1+2.*(dim-1.)/(o+1.))+(-0.516457806843)*dim
        ref[0,4,0,1]=(0.62556865655)*(1+2.*(dim-1.)/(o+1.))+(-0.0132723405234)*dim
        ref[0,4,1,0]=(-1.10166644693)*(1+2.*(dim-1.)/(o+1.))+(0.720773533984)*dim
        ref[0,4,1,1]=(-1.09945957099)*(1+2.*(dim-1.)/(o+1.))+(0.0797660015265)*dim
        ref[0,4,2,0]=(-0.359198665738)*(1+2.*(dim-1.)/(o+1.))+(-0.694004796462)*dim
        ref[0,4,2,1]=(1.60203342245)*(1+2.*(dim-1.)/(o+1.))+(-0.533590800966)*dim
        ref[1,0,0,0]=(0.0630998548187)*(1+2.*(dim-1.)/(o+1.))+(-1.21827137545)*dim
        ref[1,0,0,1]=(-0.362786036469)*(1+2.*(dim-1.)/(o+1.))+(0.629021834252)*dim
        ref[1,0,1,0]=(1.30840533269)*(1+2.*(dim-1.)/(o+1.))+(0.01020704647)*dim
        ref[1,0,1,1]=(0.581668517056)*(1+2.*(dim-1.)/(o+1.))+(0.0572291920482)*dim
        ref[1,0,2,0]=(-1.1596262045)*(1+2.*(dim-1.)/(o+1.))+(0.573375327782)*dim
        ref[1,0,2,1]=(-0.025337463973)*(1+2.*(dim-1.)/(o+1.))+(1.09210267755)*dim
        ref[1,1,0,0]=(-1.46016674214)*(1+2.*(dim-1.)/(o+1.))+(-1.05755098104)*dim
        ref[1,1,0,1]=(-0.410079206814)*(1+2.*(dim-1.)/(o+1.))+(-0.323481368888)*dim
        ref[1,1,1,0]=(0.497609154803)*(1+2.*(dim-1.)/(o+1.))+(-1.20018654453)*dim
        ref[1,1,1,1]=(0.446563754938)*(1+2.*(dim-1.)/(o+1.))+(1.47358034665)*dim
        ref[1,1,2,0]=(0.124142572295)*(1+2.*(dim-1.)/(o+1.))+(-0.672049901132)*dim
        ref[1,1,2,1]=(-1.27946259076)*(1+2.*(dim-1.)/(o+1.))+(-0.09079919688)*dim
        ref[1,2,0,0]=(0.0167916382593)*(1+2.*(dim-1.)/(o+1.))+(-1.10307978578)*dim
        ref[1,2,0,1]=(0.310300272224)*(1+2.*(dim-1.)/(o+1.))+(1.05128017274)*dim
        ref[1,2,1,0]=(-0.384485455128)*(1+2.*(dim-1.)/(o+1.))+(0.579076204736)*dim
        ref[1,2,1,1]=(0.133076827561)*(1+2.*(dim-1.)/(o+1.))+(0.59009370404)*dim
        ref[1,2,2,0]=(1.4126992054)*(1+2.*(dim-1.)/(o+1.))+(-1.44192474084)*dim
        ref[1,2,2,1]=(0.235736468845)*(1+2.*(dim-1.)/(o+1.))+(-1.01740610006)*dim
        ref[1,3,0,0]=(-0.323242689323)*(1+2.*(dim-1.)/(o+1.))+(0.917992023864)*dim
        ref[1,3,0,1]=(-1.56677140485)*(1+2.*(dim-1.)/(o+1.))+(0.156800710897)*dim
        ref[1,3,1,0]=(1.39852204704)*(1+2.*(dim-1.)/(o+1.))+(1.15334341893)*dim
        ref[1,3,1,1]=(1.10144038868)*(1+2.*(dim-1.)/(o+1.))+(-0.0544780940565)*dim
        ref[1,3,2,0]=(-0.619896071163)*(1+2.*(dim-1.)/(o+1.))+(-0.530507778126)*dim
        ref[1,3,2,1]=(0.665968335878)*(1+2.*(dim-1.)/(o+1.))+(0.920377229442)*dim
        ref[1,4,0,0]=(0.24940249757)*(1+2.*(dim-1.)/(o+1.))+(-1.25082923262)*dim
        ref[1,4,0,1]=(1.26616311307)*(1+2.*(dim-1.)/(o+1.))+(-0.0152526906804)*dim
        ref[1,4,1,0]=(1.37246903375)*(1+2.*(dim-1.)/(o+1.))+(0.720514473076)*dim
        ref[1,4,1,1]=(1.20885120083)*(1+2.*(dim-1.)/(o+1.))+(0.547442313947)*dim
        ref[1,4,2,0]=(-0.929784649867)*(1+2.*(dim-1.)/(o+1.))+(-0.547239578625)*dim
        ref[1,4,2,1]=(1.40793646749)*(1+2.*(dim-1.)/(o+1.))+(-1.00104546206)*dim
        ref[2,0,0,0]=(-0.865664592293)*(1+2.*(dim-1.)/(o+1.))+(0.577031746082)*dim
        ref[2,0,0,1]=(-0.383744932135)*(1+2.*(dim-1.)/(o+1.))+(-1.12195778437)*dim
        ref[2,0,1,0]=(0.899513902212)*(1+2.*(dim-1.)/(o+1.))+(0.892940129109)*dim
        ref[2,0,1,1]=(0.422487727431)*(1+2.*(dim-1.)/(o+1.))+(-0.070448635037)*dim
        ref[2,0,2,0]=(-0.0542314325807)*(1+2.*(dim-1.)/(o+1.))+(0.428900829837)*dim
        ref[2,0,2,1]=(-0.333651152721)*(1+2.*(dim-1.)/(o+1.))+(0.990911641084)*dim
        ref[2,1,0,0]=(0.743812676627)*(1+2.*(dim-1.)/(o+1.))+(-0.727555505694)*dim
        ref[2,1,0,1]=(0.487164931518)*(1+2.*(dim-1.)/(o+1.))+(-0.0759979311835)*dim
        ref[2,1,1,0]=(0.146918316537)*(1+2.*(dim-1.)/(o+1.))+(-1.15989754148)*dim
        ref[2,1,1,1]=(-0.239925787022)*(1+2.*(dim-1.)/(o+1.))+(0.534770278668)*dim
        ref[2,1,2,0]=(1.79635305588)*(1+2.*(dim-1.)/(o+1.))+(0.377208442988)*dim
        ref[2,1,2,1]=(-0.275396782095)*(1+2.*(dim-1.)/(o+1.))+(0.798426827206)*dim
        ref[2,2,0,0]=(-0.732792630073)*(1+2.*(dim-1.)/(o+1.))+(-0.642309522614)*dim
        ref[2,2,0,1]=(0.691994610633)*(1+2.*(dim-1.)/(o+1.))+(-0.0098684198643)*dim
        ref[2,2,1,0]=(0.571059981052)*(1+2.*(dim-1.)/(o+1.))+(-0.473609097192)*dim
        ref[2,2,1,1]=(-0.935084534142)*(1+2.*(dim-1.)/(o+1.))+(0.15087061249)*dim
        ref[2,2,2,0]=(0.0826567188857)*(1+2.*(dim-1.)/(o+1.))+(1.73404072415)*dim
        ref[2,2,2,1]=(1.26427004187)*(1+2.*(dim-1.)/(o+1.))+(1.4218194213)*dim
        ref[2,3,0,0]=(-0.398591309648)*(1+2.*(dim-1.)/(o+1.))+(0.6544834973)*dim
        ref[2,3,0,1]=(-0.568822853648)*(1+2.*(dim-1.)/(o+1.))+(0.881002495825)*dim
        ref[2,3,1,0]=(1.18000532793)*(1+2.*(dim-1.)/(o+1.))+(-0.584828617992)*dim
        ref[2,3,1,1]=(-1.60821457294)*(1+2.*(dim-1.)/(o+1.))+(1.07477669549)*dim
        ref[2,3,2,0]=(1.48063830505)*(1+2.*(dim-1.)/(o+1.))+(0.574572225346)*dim
        ref[2,3,2,1]=(0.229132618257)*(1+2.*(dim-1.)/(o+1.))+(-0.910662742613)*dim
        ref[2,4,0,0]=(0.426313638958)*(1+2.*(dim-1.)/(o+1.))+(0.0852443745317)*dim
        ref[2,4,0,1]=(-0.0761298452145)*(1+2.*(dim-1.)/(o+1.))+(1.42267336985)*dim
        ref[2,4,1,0]=(-0.537516881228)*(1+2.*(dim-1.)/(o+1.))+(0.00131616302647)*dim
        ref[2,4,1,1]=(-1.00615029385)*(1+2.*(dim-1.)/(o+1.))+(1.34001175877)*dim
        ref[2,4,2,0]=(-0.250744474716)*(1+2.*(dim-1.)/(o+1.))+(-1.43903724981)*dim
        ref[2,4,2,1]=(0.273616829308)*(1+2.*(dim-1.)/(o+1.))+(-1.82109753555)*dim
        ref[3,0,0,0]=(0.773912240208)*(1+2.*(dim-1.)/(o+1.))+(-0.340199852604)*dim
        ref[3,0,0,1]=(0.00495773386215)*(1+2.*(dim-1.)/(o+1.))+(1.5555787639)*dim
        ref[3,0,1,0]=(-0.586757469166)*(1+2.*(dim-1.)/(o+1.))+(1.60973407294)*dim
        ref[3,0,1,1]=(-0.577120453982)*(1+2.*(dim-1.)/(o+1.))+(-1.38642862644)*dim
        ref[3,0,2,0]=(0.288712056159)*(1+2.*(dim-1.)/(o+1.))+(0.253646614684)*dim
        ref[3,0,2,1]=(-0.248042635081)*(1+2.*(dim-1.)/(o+1.))+(1.2691199356)*dim
        ref[3,1,0,0]=(0.995572341947)*(1+2.*(dim-1.)/(o+1.))+(-0.848912112665)*dim
        ref[3,1,0,1]=(-1.37225234108)*(1+2.*(dim-1.)/(o+1.))+(-0.444095545534)*dim
        ref[3,1,1,0]=(0.441843238326)*(1+2.*(dim-1.)/(o+1.))+(-0.193321427606)*dim
        ref[3,1,1,1]=(0.408346053376)*(1+2.*(dim-1.)/(o+1.))+(-0.098552598199)*dim
        ref[3,1,2,0]=(-0.311970472748)*(1+2.*(dim-1.)/(o+1.))+(0.143172949307)*dim
        ref[3,1,2,1]=(-0.434499269883)*(1+2.*(dim-1.)/(o+1.))+(-0.844041314296)*dim
        ref[3,2,0,0]=(0.480046851902)*(1+2.*(dim-1.)/(o+1.))+(-0.0612387744048)*dim
        ref[3,2,0,1]=(1.12223817367)*(1+2.*(dim-1.)/(o+1.))+(-0.167879847351)*dim
        ref[3,2,1,0]=(-0.345095098378)*(1+2.*(dim-1.)/(o+1.))+(1.23062778446)*dim
        ref[3,2,1,1]=(1.37605147183)*(1+2.*(dim-1.)/(o+1.))+(-0.92954539081)*dim
        ref[3,2,2,0]=(-1.0202068094)*(1+2.*(dim-1.)/(o+1.))+(-1.2798863509)*dim
        ref[3,2,2,1]=(-0.302591270785)*(1+2.*(dim-1.)/(o+1.))+(0.103009090467)*dim
        ref[3,3,0,0]=(0.857177782444)*(1+2.*(dim-1.)/(o+1.))+(-1.43260565284)*dim
        ref[3,3,0,1]=(0.0906238774103)*(1+2.*(dim-1.)/(o+1.))+(1.06157698076)*dim
        ref[3,3,1,0]=(0.42136870665)*(1+2.*(dim-1.)/(o+1.))+(1.20358463794)*dim
        ref[3,3,1,1]=(-0.88620181924)*(1+2.*(dim-1.)/(o+1.))+(1.33611264111)*dim
        ref[3,3,2,0]=(0.444491342755)*(1+2.*(dim-1.)/(o+1.))+(-0.981750791749)*dim
        ref[3,3,2,1]=(-1.06075000108)*(1+2.*(dim-1.)/(o+1.))+(1.86809160902)*dim
        ref[3,4,0,0]=(-0.517172055625)*(1+2.*(dim-1.)/(o+1.))+(-0.364951168193)*dim
        ref[3,4,0,1]=(0.683410864606)*(1+2.*(dim-1.)/(o+1.))+(0.503470398019)*dim
        ref[3,4,1,0]=(-1.49890766435)*(1+2.*(dim-1.)/(o+1.))+(-1.16393186901)*dim
        ref[3,4,1,1]=(0.25999658315)*(1+2.*(dim-1.)/(o+1.))+(-1.00595595308)*dim
        ref[3,4,2,0]=(-0.296406951133)*(1+2.*(dim-1.)/(o+1.))+(-0.919191636173)*dim
        ref[3,4,2,1]=(-0.562797436325)*(1+2.*(dim-1.)/(o+1.))+(-0.0531967852595)*dim
      else:
        s[0,0,0,0]=(0.28520634077)*x[0]+(-0.126741433974)*x[1]+(1.28370871197)*x[2]
        s[0,0,0,1]=(-0.0731693108841)*x[0]+(0.342123624879)*x[1]+(0.302260810729)*x[2]
        s[0,0,1,0]=(-0.933375161617)*x[0]+(0.714257115938)*x[1]+(0.0355044702732)*x[2]
        s[0,0,1,1]=(0.273422423495)*x[0]+(-1.43381666098)*x[1]+(-0.3924531625)*x[2]
        s[0,0,2,0]=(-1.24504380748)*x[0]+(1.36389469156)*x[1]+(1.0146102387)*x[2]
        s[0,0,2,1]=(-0.12931890066)*x[0]+(0.00848180915183)*x[1]+(0.230392813094)*x[2]
        s[0,1,0,0]=(0.390247630566)*x[0]+(0.116544846708)*x[1]+(-0.94552924973)*x[2]
        s[0,1,0,1]=(1.53653205813)*x[0]+(0.984195773879)*x[1]+(0.767182801059)*x[2]
        s[0,1,1,0]=(0.433068815739)*x[0]+(0.196964560742)*x[1]+(0.00150259046749)*x[2]
        s[0,1,1,1]=(0.620083290721)*x[0]+(-0.981816668862)*x[1]+(-0.0602017060991)*x[2]
        s[0,1,2,0]=(0.266920335934)*x[0]+(-1.01077201758)*x[1]+(-1.66555517355)*x[2]
        s[0,1,2,1]=(-0.463794455357)*x[0]+(1.02905222514)*x[1]+(-0.524965351287)*x[2]
        s[0,2,0,0]=(-1.02890742598)*x[0]+(0.10277767711)*x[1]+(-0.709831378433)*x[2]
        s[0,2,0,1]=(0.942130530664)*x[0]+(-0.498478792456)*x[1]+(-0.321634750428)*x[2]
        s[0,2,1,0]=(-1.09979499064)*x[0]+(-0.552242175008)*x[1]+(1.47155591555)*x[2]
        s[0,2,1,1]=(0.038643175545)*x[0]+(-0.367877010362)*x[1]+(-0.422403338025)*x[2]
        s[0,2,2,0]=(0.915798473155)*x[0]+(0.627559693912)*x[1]+(-0.146957763577)*x[2]
        s[0,2,2,1]=(0.35609430038)*x[0]+(0.627155969018)*x[1]+(-0.0538976009369)*x[2]
        s[0,3,0,0]=(0.753367589685)*x[0]+(-1.56174987227)*x[1]+(0.0317386362973)*x[2]
        s[0,3,0,1]=(-1.0814768856)*x[0]+(-0.000317117094536)*x[1]+(0.160394230281)*x[2]
        s[0,3,1,0]=(0.539343002581)*x[0]+(0.600975350875)*x[1]+(-0.579108274756)*x[2]
        s[0,3,1,1]=(1.93822764773)*x[0]+(1.19115168218)*x[1]+(-0.345303177833)*x[2]
        s[0,3,2,0]=(-0.396284929409)*x[0]+(-0.0620671301539)*x[1]+(0.405978876833)*x[2]
        s[0,3,2,1]=(-0.806591193034)*x[0]+(-0.869878106357)*x[1]+(-0.352772882375)*x[2]
        s[0,4,0,0]=(-1.01176424875)*x[0]+(-1.57589661925)*x[1]+(-0.313904557434)*x[2]
        s[0,4,0,1]=(0.277176183617)*x[0]+(0.548456592065)*x[1]+(1.21233269169)*x[2]
        s[0,4,1,0]=(0.916512281025)*x[0]+(-1.53600125673)*x[1]+(-0.486397251593)*x[2]
        s[0,4,1,1]=(1.12553562564)*x[0]+(0.0569780615281)*x[1]+(-0.275491563277)*x[2]
        s[0,4,2,0]=(0.0701240897706)*x[0]+(-0.109284348724)*x[1]+(0.920255378883)*x[2]
        s[0,4,2,1]=(-0.286723671632)*x[0]+(0.521149269283)*x[1]+(0.880566081072)*x[2]
        s[1,0,0,0]=(-0.161092153994)*x[0]+(1.60646814731)*x[1]+(-0.293423133436)*x[2]
        s[1,0,0,1]=(-0.62705410926)*x[0]+(-0.133468132127)*x[1]+(0.435110047414)*x[2]
        s[1,0,1,0]=(0.963863560937)*x[0]+(0.358847992997)*x[1]+(0.847184531597)*x[2]
        s[1,0,1,1]=(0.2445787465)*x[0]+(-0.161812641928)*x[1]+(-0.679720454567)*x[2]
        s[1,0,2,0]=(1.05646386542)*x[0]+(1.0631242639)*x[1]+(-0.762052372247)*x[2]
        s[1,0,2,1]=(-0.950180341823)*x[0]+(1.00890659836)*x[1]+(1.54399946028)*x[2]
        s[1,1,0,0]=(0.290340455371)*x[0]+(-1.13919380812)*x[1]+(-0.659408495899)*x[2]
        s[1,1,0,1]=(1.4258405346)*x[0]+(1.53809682182)*x[1]+(0.0675302762314)*x[2]
        s[1,1,1,0]=(1.21278717612)*x[0]+(0.220266219454)*x[1]+(1.66873955725)*x[2]
        s[1,1,1,1]=(0.771526101684)*x[0]+(-0.067360487854)*x[1]+(1.21982979968)*x[2]
        s[1,1,2,0]=(-0.77604682892)*x[0]+(0.3133384598)*x[1]+(0.645828815985)*x[2]
        s[1,1,2,1]=(-0.107715380402)*x[0]+(-0.0772190829216)*x[1]+(-1.14700080403)*x[2]
        s[1,2,0,0]=(0.324137135934)*x[0]+(1.29562077427)*x[1]+(0.455108897284)*x[2]
        s[1,2,0,1]=(-1.43262429921)*x[0]+(0.773443491708)*x[1]+(0.277297920728)*x[2]
        s[1,2,1,0]=(0.304810473045)*x[0]+(-0.0168967979423)*x[1]+(0.131774697941)*x[2]
        s[1,2,1,1]=(0.387736648788)*x[0]+(0.666220386408)*x[1]+(1.07576988895)*x[2]
        s[1,2,2,0]=(-0.254693787451)*x[0]+(-0.3384585134)*x[1]+(-0.64060329677)*x[2]
        s[1,2,2,1]=(-0.672597816893)*x[0]+(0.145646769601)*x[1]+(-0.994938729226)*x[2]
        s[1,3,0,0]=(-1.1030634788)*x[0]+(-1.80762204727)*x[1]+(1.6309434404)*x[2]
        s[1,3,0,1]=(0.276318339362)*x[0]+(0.385250530079)*x[1]+(-0.758561170758)*x[2]
        s[1,3,1,0]=(0.179910601577)*x[0]+(-0.11818800924)*x[1]+(-0.743123335945)*x[2]
        s[1,3,1,1]=(0.782295683197)*x[0]+(0.749420799157)*x[1]+(-0.656475021802)*x[2]
        s[1,3,2,0]=(0.574829069465)*x[0]+(-0.493394135291)*x[1]+(1.29795947888)*x[2]
        s[1,3,2,1]=(-1.46888017959)*x[0]+(-1.08679474255)*x[1]+(0.156158770501)*x[2]
        s[1,4,0,0]=(-0.860789510784)*x[0]+(0.652130849161)*x[1]+(-0.0894965081242)*x[2]
        s[1,4,0,1]=(0.997352754709)*x[0]+(-0.112776085202)*x[1]+(-0.469141207634)*x[2]
        s[1,4,1,0]=(-0.401561591069)*x[0]+(-0.362165453433)*x[1]+(1.42046539935)*x[2]
        s[1,4,1,1]=(-0.668331880273)*x[0]+(-1.28337116108)*x[1]+(0.368963183759)*x[2]
        s[1,4,2,0]=(-1.15039637411)*x[0]+(0.121579045201)*x[1]+(-0.104055094935)*x[2]
        s[1,4,2,1]=(-0.923080753122)*x[0]+(0.322395936762)*x[1]+(0.233695594483)*x[2]
        s[2,0,0,0]=(-0.206424413737)*x[0]+(0.761779799677)*x[1]+(-0.301048388352)*x[2]
        s[2,0,0,1]=(1.55990080913)*x[0]+(-0.162113046152)*x[1]+(-0.310428209804)*x[2]
        s[2,0,1,0]=(-0.433285639024)*x[0]+(1.36777883595)*x[1]+(-1.17795591805)*x[2]
        s[2,0,1,1]=(1.34284608147)*x[0]+(-0.222861276203)*x[1]+(0.855486771517)*x[2]
        s[2,0,2,0]=(-0.16635189023)*x[0]+(-0.823533654953)*x[1]+(0.797079690281)*x[2]
        s[2,0,2,1]=(0.752310816499)*x[0]+(0.764018784752)*x[1]+(0.056117413692)*x[2]
        s[2,1,0,0]=(0.00166988723156)*x[0]+(0.766304091963)*x[1]+(0.275940662737)*x[2]
        s[2,1,0,1]=(-0.0855732564349)*x[0]+(-1.02664418473)*x[1]+(0.662522947755)*x[2]
        s[2,1,1,0]=(-0.658811588292)*x[0]+(-1.04635899154)*x[1]+(1.05470344529)*x[2]
        s[2,1,1,1]=(-0.119299788842)*x[0]+(-0.782372280168)*x[1]+(-1.79380761901)*x[2]
        s[2,1,2,0]=(1.71887176677)*x[0]+(-0.226946752041)*x[1]+(0.799181606184)*x[2]
        s[2,1,2,1]=(-0.9189461833)*x[0]+(0.916498349691)*x[1]+(0.734617496542)*x[2]
        s[2,2,0,0]=(-0.790364198949)*x[0]+(0.0131553082942)*x[1]+(-0.978651161192)*x[2]
        s[2,2,0,1]=(0.32388442717)*x[0]+(-0.593416422259)*x[1]+(1.10510332946)*x[2]
        s[2,2,1,0]=(0.452042167919)*x[0]+(-0.641462816719)*x[1]+(-0.356784432003)*x[2]
        s[2,2,1,1]=(0.0725380542114)*x[0]+(0.255256127518)*x[1]+(0.979834678019)*x[2]
        s[2,2,2,0]=(1.39372044596)*x[0]+(0.938441851122)*x[1]+(-0.983490819141)*x[2]
        s[2,2,2,1]=(1.2364973033)*x[0]+(-1.2815807633)*x[1]+(-0.360311591593)*x[2]
        s[2,3,0,0]=(-0.0597199849648)*x[0]+(0.376596450311)*x[1]+(-0.494756059286)*x[2]
        s[2,3,0,1]=(-0.1287394847)*x[0]+(0.0455819025025)*x[1]+(1.57146607969)*x[2]
        s[2,3,1,0]=(-1.4907853501)*x[0]+(0.0388186733561)*x[1]+(0.13920185582)*x[2]
        s[2,3,1,1]=(0.0439467610482)*x[0]+(1.43515095249)*x[1]+(-0.205189829549)*x[2]
        s[2,3,2,0]=(0.773581921294)*x[0]+(1.41482523023)*x[1]+(1.02435393168)*x[2]
        s[2,3,2,1]=(0.347805074163)*x[0]+(1.52992663335)*x[1]+(-0.487058346391)*x[2]
        s[2,4,0,0]=(-0.107692405269)*x[0]+(0.147562810855)*x[1]+(1.38077195652)*x[2]
        s[2,4,0,1]=(-0.0410323036269)*x[0]+(0.269432309711)*x[1]+(-0.693128417186)*x[2]
        s[2,4,1,0]=(-0.33861862665)*x[0]+(-0.419772907291)*x[1]+(-1.07238876285)*x[2]
        s[2,4,1,1]=(0.143346805317)*x[0]+(-0.419444960688)*x[1]+(1.59692612607)*x[2]
        s[2,4,2,0]=(0.654582963584)*x[0]+(-0.44163597128)*x[1]+(-0.694330043907)*x[2]
        s[2,4,2,1]=(0.627421193857)*x[0]+(0.597177617998)*x[1]+(-0.451180510793)*x[2]
        s[3,0,0,0]=(-0.101412872135)*x[0]+(-0.396884673414)*x[1]+(0.896841368992)*x[2]
        s[3,0,0,1]=(0.139595535793)*x[0]+(0.624607944421)*x[1]+(0.574170381258)*x[2]
        s[3,0,1,0]=(0.316717544294)*x[0]+(-0.621261709901)*x[1]+(1.31441803912)*x[2]
        s[3,0,1,1]=(0.254418231071)*x[0]+(0.228843906675)*x[1]+(0.25896007767)*x[2]
        s[3,0,2,0]=(-0.247274104613)*x[0]+(1.20694307591)*x[1]+(-0.407808429936)*x[2]
        s[3,0,2,1]=(-1.40605558298)*x[0]+(0.168721449606)*x[1]+(-0.972867940483)*x[2]
        s[3,1,0,0]=(0.245845223266)*x[0]+(-1.51779122633)*x[1]+(-0.779293963961)*x[2]
        s[3,1,0,1]=(0.0325495798223)*x[0]+(-0.12100728517)*x[1]+(1.31848044222)*x[2]
        s[3,1,1,0]=(0.789118948962)*x[0]+(0.589652419254)*x[1]+(-0.08688557206)*x[2]
        s[3,1,1,1]=(-1.83518748403)*x[0]+(-0.500646680661)*x[1]+(0.233750065377)*x[2]
        s[3,1,2,0]=(-0.0771120817018)*x[0]+(0.0904104042813)*x[1]+(0.0933370539969)*x[2]
        s[3,1,2,1]=(-1.19249513127)*x[0]+(0.447348983457)*x[1]+(-0.298916535972)*x[2]
        s[3,2,0,0]=(1.22266670435)*x[0]+(-1.62277644197)*x[1]+(-0.0190709988254)*x[2]
        s[3,2,0,1]=(-0.194952675382)*x[0]+(0.554694000692)*x[1]+(-0.151982113386)*x[2]
        s[3,2,1,0]=(0.511265692083)*x[0]+(-0.802622720144)*x[1]+(-0.242270723374)*x[2]
        s[3,2,1,1]=(0.476021603798)*x[0]+(-0.463494861878)*x[1]+(-0.532143290096)*x[2]
        s[3,2,2,0]=(-0.530708374495)*x[0]+(1.10856138272)*x[1]+(0.409512895954)*x[2]
        s[3,2,2,1]=(0.0214256235243)*x[0]+(0.836046354994)*x[1]+(-0.263352555948)*x[2]
        s[3,3,0,0]=(-0.437181323979)*x[0]+(-0.413716617562)*x[1]+(-0.432122540243)*x[2]
        s[3,3,0,1]=(1.08881737048)*x[0]+(-0.0166529943951)*x[1]+(-0.236235303435)*x[2]
        s[3,3,1,0]=(-0.220528005238)*x[0]+(-0.0637133108381)*x[1]+(-0.722670710646)*x[2]
        s[3,3,1,1]=(0.0267596690177)*x[0]+(0.277583150215)*x[1]+(-1.19820417967)*x[2]
        s[3,3,2,0]=(-0.134456240167)*x[0]+(0.207594364333)*x[1]+(-1.58153388102)*x[2]
        s[3,3,2,1]=(-0.484174332477)*x[0]+(1.24683115653)*x[1]+(0.0960608606087)*x[2]
        s[3,4,0,0]=(-1.06280143452)*x[0]+(0.0909193631754)*x[1]+(0.735785607027)*x[2]
        s[3,4,0,1]=(-0.39199782699)*x[0]+(-0.547457599272)*x[1]+(-0.990008593852)*x[2]
        s[3,4,1,0]=(-0.301784264736)*x[0]+(-0.219227445543)*x[1]+(-0.979494077205)*x[2]
        s[3,4,1,1]=(-0.898860901012)*x[0]+(-0.103378912588)*x[1]+(-0.66619898802)*x[2]
        s[3,4,2,0]=(1.10051932868)*x[0]+(0.407341781092)*x[1]+(-0.680512930996)*x[2]
        s[3,4,2,1]=(-0.0733544592857)*x[0]+(1.18707229675)*x[1]+(-0.777693466919)*x[2]
        ref[0,0,0,0]=(0.438851859071)*(1+2.*(dim-1.)/(o+1.))+(1.0033217597)*dim
        ref[0,0,0,1]=(-0.891607994637)*(1+2.*(dim-1.)/(o+1.))+(1.46282311936)*dim
        ref[0,0,1,0]=(0.817888583209)*(1+2.*(dim-1.)/(o+1.))+(-1.00150215862)*dim
        ref[0,0,1,1]=(-1.639070793)*(1+2.*(dim-1.)/(o+1.))+(0.0862233930162)*dim
        ref[0,0,2,0]=(-0.348875799642)*(1+2.*(dim-1.)/(o+1.))+(1.48233692242)*dim
        ref[0,0,2,1]=(-0.589821443428)*(1+2.*(dim-1.)/(o+1.))+(0.699377165015)*dim
        ref[0,1,0,0]=(-0.832218783182)*(1+2.*(dim-1.)/(o+1.))+(0.393482010727)*dim
        ref[0,1,0,1]=(2.07094909251)*(1+2.*(dim-1.)/(o+1.))+(1.21696154056)*dim
        ref[0,1,1,0]=(0.1125989301)*(1+2.*(dim-1.)/(o+1.))+(0.518937036848)*dim
        ref[0,1,1,1]=(-1.36685685946)*(1+2.*(dim-1.)/(o+1.))+(0.944921775217)*dim
        ref[0,1,2,0]=(-1.47817293944)*(1+2.*(dim-1.)/(o+1.))+(-0.931233915749)*dim
        ref[0,1,2,1]=(-0.213405460956)*(1+2.*(dim-1.)/(o+1.))+(0.253697879455)*dim
        ref[0,2,0,0]=(-0.0805522984701)*(1+2.*(dim-1.)/(o+1.))+(-1.55540882883)*dim
        ref[0,2,0,1]=(0.156888902683)*(1+2.*(dim-1.)/(o+1.))+(-0.034871914902)*dim
        ref[0,2,1,0]=(-0.15684567608)*(1+2.*(dim-1.)/(o+1.))+(-0.0236355740175)*dim
        ref[0,2,1,1]=(-1.39067500442)*(1+2.*(dim-1.)/(o+1.))+(0.639037831581)*dim
        ref[0,2,2,0]=(1.81399897277)*(1+2.*(dim-1.)/(o+1.))+(-0.417598569281)*dim
        ref[0,2,2,1]=(-0.764514507478)*(1+2.*(dim-1.)/(o+1.))+(1.69386717594)*dim
        ref[0,3,0,0]=(-0.392055498198)*(1+2.*(dim-1.)/(o+1.))+(-0.384588148087)*dim
        ref[0,3,0,1]=(-1.62777049183)*(1+2.*(dim-1.)/(o+1.))+(0.706370719421)*dim
        ref[0,3,1,0]=(0.619522807034)*(1+2.*(dim-1.)/(o+1.))+(-0.0583127283348)*dim
        ref[0,3,1,1]=(1.39660007067)*(1+2.*(dim-1.)/(o+1.))+(1.3874760814)*dim
        ref[0,3,2,0]=(0.396663683156)*(1+2.*(dim-1.)/(o+1.))+(-0.449036865886)*dim
        ref[0,3,2,1]=(-1.08765510225)*(1+2.*(dim-1.)/(o+1.))+(-0.941587079516)*dim
        ref[0,4,0,0]=(-1.51265971093)*(1+2.*(dim-1.)/(o+1.))+(-1.3889057145)*dim
        ref[0,4,0,1]=(0.901518176509)*(1+2.*(dim-1.)/(o+1.))+(1.13644729087)*dim
        ref[0,4,1,0]=(-0.58127466259)*(1+2.*(dim-1.)/(o+1.))+(-0.524611564706)*dim
        ref[0,4,1,1]=(1.59425336176)*(1+2.*(dim-1.)/(o+1.))+(-0.687231237871)*dim
        ref[0,4,2,0]=(1.07708364012)*(1+2.*(dim-1.)/(o+1.))+(-0.195988520187)*dim
        ref[0,4,2,1]=(0.108644035622)*(1+2.*(dim-1.)/(o+1.))+(1.0063476431)*dim
        ref[1,0,0,0]=(-0.153352577193)*(1+2.*(dim-1.)/(o+1.))+(1.30530543707)*dim
        ref[1,0,0,1]=(-0.518263005278)*(1+2.*(dim-1.)/(o+1.))+(0.192850811305)*dim
        ref[1,0,1,0]=(0.0968185470945)*(1+2.*(dim-1.)/(o+1.))+(2.07307753844)*dim
        ref[1,0,1,1]=(-0.572562275019)*(1+2.*(dim-1.)/(o+1.))+(-0.0243920749758)*dim
        ref[1,0,2,0]=(1.40149079544)*(1+2.*(dim-1.)/(o+1.))+(-0.043955038372)*dim
        ref[1,0,2,1]=(0.613246507575)*(1+2.*(dim-1.)/(o+1.))+(0.989479209244)*dim
        ref[1,1,0,0]=(-0.444118917423)*(1+2.*(dim-1.)/(o+1.))+(-1.06414293122)*dim
        ref[1,1,0,1]=(2.18067335546)*(1+2.*(dim-1.)/(o+1.))+(0.850794277187)*dim
        ref[1,1,1,0]=(0.976948683131)*(1+2.*(dim-1.)/(o+1.))+(2.1248442697)*dim
        ref[1,1,1,1]=(1.02658988987)*(1+2.*(dim-1.)/(o+1.))+(0.897405523639)*dim
        ref[1,1,2,0]=(-0.442692736232)*(1+2.*(dim-1.)/(o+1.))+(0.625813183096)*dim
        ref[1,1,2,1]=(-0.43212274342)*(1+2.*(dim-1.)/(o+1.))+(-0.899812523933)*dim
        ref[1,2,0,0]=(0.457022725817)*(1+2.*(dim-1.)/(o+1.))+(1.61784408168)*dim
        ref[1,2,0,1]=(0.643173594534)*(1+2.*(dim-1.)/(o+1.))+(-1.02505648131)*dim
        ref[1,2,1,0]=(-0.807339832268)*(1+2.*(dim-1.)/(o+1.))+(1.22702820531)*dim
        ref[1,2,1,1]=(1.42307806941)*(1+2.*(dim-1.)/(o+1.))+(0.706648854736)*dim
        ref[1,2,2,0]=(-1.02472620332)*(1+2.*(dim-1.)/(o+1.))+(-0.209029394299)*dim
        ref[1,2,2,1]=(-0.735490314485)*(1+2.*(dim-1.)/(o+1.))+(-0.786399462032)*dim
        ref[1,3,0,0]=(-0.193471598403)*(1+2.*(dim-1.)/(o+1.))+(-1.08627048726)*dim
        ref[1,3,0,1]=(0.385356410185)*(1+2.*(dim-1.)/(o+1.))+(-0.482348711502)*dim
        ref[1,3,1,0]=(0.387298963898)*(1+2.*(dim-1.)/(o+1.))+(-1.06869970751)*dim
        ref[1,3,1,1]=(0.423963900797)*(1+2.*(dim-1.)/(o+1.))+(0.451277559755)*dim
        ref[1,3,2,0]=(2.23728483025)*(1+2.*(dim-1.)/(o+1.))+(-0.857890417195)*dim
        ref[1,3,2,1]=(-0.537279589749)*(1+2.*(dim-1.)/(o+1.))+(-1.86223656189)*dim
        ref[1,4,0,0]=(0.0728532816595)*(1+2.*(dim-1.)/(o+1.))+(-0.371008451407)*dim
        ref[1,4,0,1]=(-0.159731100663)*(1+2.*(dim-1.)/(o+1.))+(0.575166562536)*dim
        ref[1,4,1,0]=(-1.07464731238)*(1+2.*(dim-1.)/(o+1.))+(1.73138566723)*dim
        ref[1,4,1,1]=(0.0871788096124)*(1+2.*(dim-1.)/(o+1.))+(-1.66991866721)*dim
        ref[1,4,2,0]=(-0.561546011624)*(1+2.*(dim-1.)/(o+1.))+(-0.571326412224)*dim
        ref[1,4,2,1]=(0.00625030418185)*(1+2.*(dim-1.)/(o+1.))+(-0.373239526059)*dim
        ref[2,0,0,0]=(-0.081748991935)*(1+2.*(dim-1.)/(o+1.))+(0.336055989523)*dim
        ref[2,0,0,1]=(-0.324440284217)*(1+2.*(dim-1.)/(o+1.))+(1.41179983739)*dim
        ref[2,0,1,0]=(-1.01122720324)*(1+2.*(dim-1.)/(o+1.))+(0.767764482116)*dim
        ref[2,0,1,1]=(1.90139553236)*(1+2.*(dim-1.)/(o+1.))+(0.0740760444209)*dim
        ref[2,0,2,0]=(1.07285550495)*(1+2.*(dim-1.)/(o+1.))+(-1.26566135985)*dim
        ref[2,0,2,1]=(1.4632067138)*(1+2.*(dim-1.)/(o+1.))+(0.109240301143)*dim
        ref[2,1,0,0]=(0.284420138548)*(1+2.*(dim-1.)/(o+1.))+(0.759494503384)*dim
        ref[2,1,0,1]=(-0.260549938527)*(1+2.*(dim-1.)/(o+1.))+(-0.189144554883)*dim
        ref[2,1,1,0]=(0.442311731675)*(1+2.*(dim-1.)/(o+1.))+(-1.09277886622)*dim
        ref[2,1,1,1]=(-0.10302967331)*(1+2.*(dim-1.)/(o+1.))+(-2.59245001471)*dim
        ref[2,1,2,0]=(0.739203737903)*(1+2.*(dim-1.)/(o+1.))+(1.55190288301)*dim
        ref[2,1,2,1]=(0.079851368615)*(1+2.*(dim-1.)/(o+1.))+(0.652318294318)*dim
        ref[2,2,0,0]=(-1.06662747958)*(1+2.*(dim-1.)/(o+1.))+(-0.689232572268)*dim
        ref[2,2,0,1]=(1.99822376512)*(1+2.*(dim-1.)/(o+1.))+(-1.16265243075)*dim
        ref[2,2,1,0]=(-1.22458303021)*(1+2.*(dim-1.)/(o+1.))+(0.678377949405)*dim
        ref[2,2,1,1]=(1.08782595263)*(1+2.*(dim-1.)/(o+1.))+(0.219802907121)*dim
        ref[2,2,2,0]=(0.905623047084)*(1+2.*(dim-1.)/(o+1.))+(0.44304843086)*dim
        ref[2,2,2,1]=(0.0286572039953)*(1+2.*(dim-1.)/(o+1.))+(-0.434052255584)*dim
        ref[2,3,0,0]=(0.498368738287)*(1+2.*(dim-1.)/(o+1.))+(-0.676248332227)*dim
        ref[2,3,0,1]=(0.0199276642636)*(1+2.*(dim-1.)/(o+1.))+(1.46838083323)*dim
        ref[2,3,1,0]=(-0.948945787905)*(1+2.*(dim-1.)/(o+1.))+(-0.363819033015)*dim
        ref[2,3,1,1]=(1.64627388129)*(1+2.*(dim-1.)/(o+1.))+(-0.372365997305)*dim
        ref[2,3,2,0]=(2.13784190571)*(1+2.*(dim-1.)/(o+1.))+(1.07491917749)*dim
        ref[2,3,2,1]=(-0.233611389691)*(1+2.*(dim-1.)/(o+1.))+(1.62428475081)*dim
        ref[2,4,0,0]=(0.739702054679)*(1+2.*(dim-1.)/(o+1.))+(0.680940307431)*dim
        ref[2,4,0,1]=(0.784529457615)*(1+2.*(dim-1.)/(o+1.))+(-1.24925786872)*dim
        ref[2,4,1,0]=(-0.583682264824)*(1+2.*(dim-1.)/(o+1.))+(-1.24709803197)*dim
        ref[2,4,1,1]=(2.09646981494)*(1+2.*(dim-1.)/(o+1.))+(-0.775641844239)*dim
        ref[2,4,2,0]=(0.259700462448)*(1+2.*(dim-1.)/(o+1.))+(-0.741083514052)*dim
        ref[2,4,2,1]=(0.997073171465)*(1+2.*(dim-1.)/(o+1.))+(-0.223654870403)*dim
        ref[3,0,0,0]=(0.444883466114)*(1+2.*(dim-1.)/(o+1.))+(-0.0463396426708)*dim
        ref[3,0,0,1]=(1.26485847247)*(1+2.*(dim-1.)/(o+1.))+(0.0735153889988)*dim
        ref[3,0,1,0]=(-0.0375651517431)*(1+2.*(dim-1.)/(o+1.))+(1.04743902526)*dim
        ref[3,0,1,1]=(0.5358978799)*(1+2.*(dim-1.)/(o+1.))+(0.206324335516)*dim
        ref[3,0,2,0]=(0.256485361704)*(1+2.*(dim-1.)/(o+1.))+(0.295375179654)*dim
        ref[3,0,2,1]=(-0.592012973207)*(1+2.*(dim-1.)/(o+1.))+(-1.61818910065)*dim
        ref[3,1,0,0]=(-0.526710008267)*(1+2.*(dim-1.)/(o+1.))+(-1.52452995876)*dim
        ref[3,1,0,1]=(1.25161847815)*(1+2.*(dim-1.)/(o+1.))+(-0.0215957412846)*dim
        ref[3,1,1,0]=(0.0535402130696)*(1+2.*(dim-1.)/(o+1.))+(1.23834558309)*dim
        ref[3,1,1,1]=(-0.642702769482)*(1+2.*(dim-1.)/(o+1.))+(-1.45938132983)*dim
        ref[3,1,2,0]=(0.0084187441016)*(1+2.*(dim-1.)/(o+1.))+(0.0982166324748)*dim
        ref[3,1,2,1]=(-0.0835608510497)*(1+2.*(dim-1.)/(o+1.))+(-0.960501832737)*dim
        ref[3,2,0,0]=(-0.233638393296)*(1+2.*(dim-1.)/(o+1.))+(-0.185542343145)*dim
        ref[3,2,0,1]=(1.83727235646)*(1+2.*(dim-1.)/(o+1.))+(-1.62951314453)*dim
        ref[3,2,1,0]=(0.141920587064)*(1+2.*(dim-1.)/(o+1.))+(-0.675548338499)*dim
        ref[3,2,1,1]=(-0.546837220172)*(1+2.*(dim-1.)/(o+1.))+(0.027220671997)*dim
        ref[3,2,2,0]=(1.08956280993)*(1+2.*(dim-1.)/(o+1.))+(-0.102196905755)*dim
        ref[3,2,2,1]=(1.37182911507)*(1+2.*(dim-1.)/(o+1.))+(-0.7777096925)*dim
        ref[3,3,0,0]=(-1.71451575877)*(1+2.*(dim-1.)/(o+1.))+(0.43149527699)*dim
        ref[3,3,0,1]=(0.125482430607)*(1+2.*(dim-1.)/(o+1.))+(0.710446642044)*dim
        ref[3,3,1,0]=(0.731418690422)*(1+2.*(dim-1.)/(o+1.))+(-1.73833071714)*dim
        ref[3,3,1,1]=(0.0925145394024)*(1+2.*(dim-1.)/(o+1.))+(-0.986375899836)*dim
        ref[3,3,2,0]=(-1.07231086612)*(1+2.*(dim-1.)/(o+1.))+(-0.436084890732)*dim
        ref[3,3,2,1]=(0.332375487809)*(1+2.*(dim-1.)/(o+1.))+(0.526342196849)*dim
        ref[3,4,0,0]=(0.730251131978)*(1+2.*(dim-1.)/(o+1.))+(-0.966347596298)*dim
        ref[3,4,0,1]=(0.235975729317)*(1+2.*(dim-1.)/(o+1.))+(-2.16543974943)*dim
        ref[3,4,1,0]=(-2.13209908673)*(1+2.*(dim-1.)/(o+1.))+(0.631593299244)*dim
        ref[3,4,1,1]=(-0.257512397822)*(1+2.*(dim-1.)/(o+1.))+(-1.4109264038)*dim
        ref[3,4,2,0]=(0.201086569751)*(1+2.*(dim-1.)/(o+1.))+(0.626261609021)*dim
        ref[3,4,2,1]=(0.819600430724)*(1+2.*(dim-1.)/(o+1.))+(-0.483576060183)*dim
      res=integrate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank0(self):
      """
      tests integral of rank 0 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=())
      s=Data(0,(),w)
      if dim==2:
        s=(0.652580656299)*x[0]**o+(-0.90619447815)*x[0]+(-0.0960652458634)*x[1]**o+(-0.921815212982)*x[1]
        ref=(0.556515410435)*(1+2.*(dim-1.)/(o+1.))+(-1.82800969113)*dim
      else:
        s=(0.992239317892)*x[0]**o+(0.527946493658)*x[0]+(-0.517852623269)*x[1]**o+(-0.85996207711)*x[1]+(-0.351819025215)*x[2]**o+(0.780874062499)*x[2]
        ref=(0.122567669408)*(1+2.*(dim-1.)/(o+1.))+(0.448858479047)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank1(self):
      """
      tests integral of rank 1 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,))
      s=Data(0,(2,),w)
      ref=numpy.zeros((2,),numpy.float_)
      if dim==2:
        s[0]=(0.493219509026)*x[0]**o+(-0.987257686105)*x[0]+(0.593292619761)*x[1]**o+(0.404393943308)*x[1]
        s[1]=(-0.00503309013251)*x[0]**o+(0.383977250975)*x[0]+(-0.98427148677)*x[1]**o+(-0.980898646673)*x[1]
        ref[0]=(1.08651212879)*(1+2.*(dim-1.)/(o+1.))+(-0.582863742798)*dim
        ref[1]=(-0.989304576902)*(1+2.*(dim-1.)/(o+1.))+(-0.596921395698)*dim
      else:
        s[0]=(-0.807414180514)*x[0]**o+(-0.887552057443)*x[0]+(0.419062942657)*x[1]**o+(-0.0780307657804)*x[1]+(-0.57375161239)*x[2]**o+(-0.672079552789)*x[2]
        s[1]=(0.0341328802311)*x[0]**o+(0.8952312571)*x[0]+(0.769555352247)*x[1]**o+(-0.974043749526)*x[1]+(0.532943429154)*x[2]**o+(0.265140756994)*x[2]
        ref[0]=(-0.962102850247)*(1+2.*(dim-1.)/(o+1.))+(-1.63766237601)*dim
        ref[1]=(1.33663166163)*(1+2.*(dim-1.)/(o+1.))+(0.186328264568)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank2(self):
      """
      tests integral of rank 2 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5))
      s=Data(0,(4, 5),w)
      ref=numpy.zeros((4, 5),numpy.float_)
      if dim==2:
        s[0,0]=(-0.874129654878)*x[0]**o+(0.769078079549)*x[0]+(0.169490655493)*x[1]**o+(-0.414105577874)*x[1]
        s[0,1]=(-0.573932038812)*x[0]**o+(0.940575833423)*x[0]+(0.146726598392)*x[1]**o+(-0.709516681116)*x[1]
        s[0,2]=(0.958570693102)*x[0]**o+(0.345205321881)*x[0]+(-0.418462504081)*x[1]**o+(-0.391658885585)*x[1]
        s[0,3]=(0.964572209001)*x[0]**o+(-0.989029471645)*x[0]+(-0.783477665675)*x[1]**o+(-0.730975054597)*x[1]
        s[0,4]=(-0.401820155821)*x[0]**o+(0.937487436813)*x[0]+(-0.436388767817)*x[1]**o+(0.41944569363)*x[1]
        s[1,0]=(-0.218435283434)*x[0]**o+(0.449489023616)*x[0]+(0.835180453816)*x[1]**o+(0.330814802055)*x[1]
        s[1,1]=(0.997378308339)*x[0]**o+(-0.0536851038818)*x[0]+(0.415472188324)*x[1]**o+(-0.541543315513)*x[1]
        s[1,2]=(0.531080153329)*x[0]**o+(0.597962435163)*x[0]+(0.196717837822)*x[1]**o+(-0.847336564598)*x[1]
        s[1,3]=(-0.601657809776)*x[0]**o+(0.143169415905)*x[0]+(0.572624366522)*x[1]**o+(-0.126109854968)*x[1]
        s[1,4]=(-0.0923887797789)*x[0]**o+(0.743510972973)*x[0]+(-0.906380069241)*x[1]**o+(0.857368771305)*x[1]
        s[2,0]=(0.671976577482)*x[0]**o+(0.940982995212)*x[0]+(0.0235322881083)*x[1]**o+(0.551932901955)*x[1]
        s[2,1]=(-0.482943532169)*x[0]**o+(0.4285433846)*x[0]+(-0.118244007829)*x[1]**o+(0.674663307952)*x[1]
        s[2,2]=(-0.191762280668)*x[0]**o+(0.0989844222111)*x[0]+(0.322777346991)*x[1]**o+(0.0282258368072)*x[1]
        s[2,3]=(0.927688375343)*x[0]**o+(-0.616381391004)*x[0]+(0.328443474996)*x[1]**o+(0.0269050191942)*x[1]
        s[2,4]=(-0.441359800454)*x[0]**o+(-0.840582337813)*x[0]+(-0.0889373588313)*x[1]**o+(-0.973572053183)*x[1]
        s[3,0]=(0.346613552119)*x[0]**o+(0.191330676816)*x[0]+(-0.632874511488)*x[1]**o+(0.987141598)*x[1]
        s[3,1]=(-0.871079285785)*x[0]**o+(0.0595183425049)*x[0]+(-0.406377080612)*x[1]**o+(0.584095659381)*x[1]
        s[3,2]=(0.959606634672)*x[0]**o+(-0.00872257186186)*x[0]+(0.123701429421)*x[1]**o+(-0.701522275891)*x[1]
        s[3,3]=(0.791212386918)*x[0]**o+(-0.0420426671732)*x[0]+(-0.149789950006)*x[1]**o+(-0.218645658506)*x[1]
        s[3,4]=(0.784561541275)*x[0]**o+(0.615978682934)*x[0]+(0.918442985828)*x[1]**o+(0.063134842611)*x[1]
        ref[0,0]=(-0.704638999385)*(1+2.*(dim-1.)/(o+1.))+(0.354972501675)*dim
        ref[0,1]=(-0.42720544042)*(1+2.*(dim-1.)/(o+1.))+(0.231059152307)*dim
        ref[0,2]=(0.540108189021)*(1+2.*(dim-1.)/(o+1.))+(-0.046453563704)*dim
        ref[0,3]=(0.181094543326)*(1+2.*(dim-1.)/(o+1.))+(-1.72000452624)*dim
        ref[0,4]=(-0.838208923637)*(1+2.*(dim-1.)/(o+1.))+(1.35693313044)*dim
        ref[1,0]=(0.616745170381)*(1+2.*(dim-1.)/(o+1.))+(0.780303825672)*dim
        ref[1,1]=(1.41285049666)*(1+2.*(dim-1.)/(o+1.))+(-0.595228419395)*dim
        ref[1,2]=(0.727797991151)*(1+2.*(dim-1.)/(o+1.))+(-0.249374129435)*dim
        ref[1,3]=(-0.0290334432542)*(1+2.*(dim-1.)/(o+1.))+(0.017059560937)*dim
        ref[1,4]=(-0.99876884902)*(1+2.*(dim-1.)/(o+1.))+(1.60087974428)*dim
        ref[2,0]=(0.69550886559)*(1+2.*(dim-1.)/(o+1.))+(1.49291589717)*dim
        ref[2,1]=(-0.601187539998)*(1+2.*(dim-1.)/(o+1.))+(1.10320669255)*dim
        ref[2,2]=(0.131015066323)*(1+2.*(dim-1.)/(o+1.))+(0.127210259018)*dim
        ref[2,3]=(1.25613185034)*(1+2.*(dim-1.)/(o+1.))+(-0.58947637181)*dim
        ref[2,4]=(-0.530297159285)*(1+2.*(dim-1.)/(o+1.))+(-1.814154391)*dim
        ref[3,0]=(-0.28626095937)*(1+2.*(dim-1.)/(o+1.))+(1.17847227482)*dim
        ref[3,1]=(-1.2774563664)*(1+2.*(dim-1.)/(o+1.))+(0.643614001886)*dim
        ref[3,2]=(1.08330806409)*(1+2.*(dim-1.)/(o+1.))+(-0.710244847753)*dim
        ref[3,3]=(0.641422436913)*(1+2.*(dim-1.)/(o+1.))+(-0.260688325679)*dim
        ref[3,4]=(1.7030045271)*(1+2.*(dim-1.)/(o+1.))+(0.679113525545)*dim
      else:
        s[0,0]=(-0.890021592515)*x[0]**o+(0.869008439528)*x[0]+(0.0393578211875)*x[1]**o+(-0.898838314128)*x[1]+(0.120962394264)*x[2]**o+(0.820232165098)*x[2]
        s[0,1]=(0.200360087495)*x[0]**o+(-0.510360711019)*x[0]+(0.289719656756)*x[1]**o+(-0.544337956881)*x[1]+(-0.137565893265)*x[2]**o+(0.0780003071465)*x[2]
        s[0,2]=(-0.611390021035)*x[0]**o+(-0.257678585283)*x[0]+(-0.432064087215)*x[1]**o+(0.580503927543)*x[1]+(0.599605038707)*x[2]**o+(-0.116243951128)*x[2]
        s[0,3]=(0.533063923783)*x[0]**o+(0.346854667765)*x[0]+(-0.0316617652896)*x[1]**o+(-0.387870493997)*x[1]+(-0.959091675716)*x[2]**o+(-0.632504268888)*x[2]
        s[0,4]=(0.105421436346)*x[0]**o+(0.0672025049394)*x[0]+(0.295674168162)*x[1]**o+(-0.599191260911)*x[1]+(0.106390236596)*x[2]**o+(0.273159012521)*x[2]
        s[1,0]=(-0.179866026516)*x[0]**o+(-0.474318032618)*x[0]+(-0.0917700558723)*x[1]**o+(-0.957849615453)*x[1]+(0.405593449997)*x[2]**o+(-0.0879542737797)*x[2]
        s[1,1]=(0.263341986971)*x[0]**o+(0.0843810614392)*x[0]+(-0.532163327984)*x[1]**o+(0.0695511301898)*x[1]+(0.425849191612)*x[2]**o+(0.891975316063)*x[2]
        s[1,2]=(0.529944898182)*x[0]**o+(0.3446605434)*x[0]+(0.94881991666)*x[1]**o+(0.60846893826)*x[1]+(0.384682108067)*x[2]**o+(0.484190212271)*x[2]
        s[1,3]=(-0.773280762221)*x[0]**o+(-0.694374711369)*x[0]+(0.595597400746)*x[1]**o+(-0.317729831568)*x[1]+(0.436671059979)*x[2]**o+(0.862685012466)*x[2]
        s[1,4]=(0.901990663611)*x[0]**o+(-0.185951360192)*x[0]+(0.880727230793)*x[1]**o+(-0.696829887245)*x[1]+(-0.871619618845)*x[2]**o+(-0.785915181957)*x[2]
        s[2,0]=(-0.598559485292)*x[0]**o+(0.378821777768)*x[0]+(0.262525874825)*x[1]**o+(-0.633890863472)*x[1]+(0.436882086359)*x[2]**o+(0.152137062224)*x[2]
        s[2,1]=(-0.120094887119)*x[0]**o+(0.774082951002)*x[0]+(0.0910918956038)*x[1]**o+(0.914069932683)*x[1]+(-0.200132987477)*x[2]**o+(0.268683863782)*x[2]
        s[2,2]=(0.769128791381)*x[0]**o+(-0.972503532786)*x[0]+(0.898341628893)*x[1]**o+(-0.804588468967)*x[1]+(0.959444830467)*x[2]**o+(0.729116746427)*x[2]
        s[2,3]=(0.460426389529)*x[0]**o+(-0.631593304662)*x[0]+(-0.197985888231)*x[1]**o+(0.994315324532)*x[1]+(0.143730574254)*x[2]**o+(-0.635611067664)*x[2]
        s[2,4]=(-0.212297893625)*x[0]**o+(0.631051506826)*x[0]+(-0.906149638803)*x[1]**o+(0.95431164216)*x[1]+(0.506425952055)*x[2]**o+(0.476198095725)*x[2]
        s[3,0]=(-0.572583486627)*x[0]**o+(-0.615297124684)*x[0]+(0.344317617297)*x[1]**o+(0.798009040785)*x[1]+(0.646565073237)*x[2]**o+(-0.119267772639)*x[2]
        s[3,1]=(-0.275482328004)*x[0]**o+(-0.225018438903)*x[0]+(0.648843558509)*x[1]**o+(0.971983956502)*x[1]+(0.890972764481)*x[2]**o+(-0.0457638811275)*x[2]
        s[3,2]=(0.625004124231)*x[0]**o+(-0.691087681301)*x[0]+(0.823334046124)*x[1]**o+(0.174931553897)*x[1]+(-0.783435157951)*x[2]**o+(-0.183737298781)*x[2]
        s[3,3]=(0.359619581557)*x[0]**o+(0.360159573822)*x[0]+(0.395267272408)*x[1]**o+(-0.412837344771)*x[1]+(-0.015119491493)*x[2]**o+(-0.0468017493027)*x[2]
        s[3,4]=(0.0939051154145)*x[0]**o+(-0.35421304768)*x[0]+(-0.388406902403)*x[1]**o+(-0.17940478311)*x[1]+(0.247049129844)*x[2]**o+(0.208156547444)*x[2]
        ref[0,0]=(-0.729701377063)*(1+2.*(dim-1.)/(o+1.))+(0.790402290499)*dim
        ref[0,1]=(0.352513850987)*(1+2.*(dim-1.)/(o+1.))+(-0.976698360754)*dim
        ref[0,2]=(-0.443849069543)*(1+2.*(dim-1.)/(o+1.))+(0.206581391132)*dim
        ref[0,3]=(-0.457689517223)*(1+2.*(dim-1.)/(o+1.))+(-0.673520095121)*dim
        ref[0,4]=(0.507485841104)*(1+2.*(dim-1.)/(o+1.))+(-0.258829743451)*dim
        ref[1,0]=(0.13395736761)*(1+2.*(dim-1.)/(o+1.))+(-1.52012192185)*dim
        ref[1,1]=(0.157027850598)*(1+2.*(dim-1.)/(o+1.))+(1.04590750769)*dim
        ref[1,2]=(1.86344692291)*(1+2.*(dim-1.)/(o+1.))+(1.43731969393)*dim
        ref[1,3]=(0.258987698503)*(1+2.*(dim-1.)/(o+1.))+(-0.149419530472)*dim
        ref[1,4]=(0.911098275559)*(1+2.*(dim-1.)/(o+1.))+(-1.66869642939)*dim
        ref[2,0]=(0.100848475893)*(1+2.*(dim-1.)/(o+1.))+(-0.10293202348)*dim
        ref[2,1]=(-0.229135978992)*(1+2.*(dim-1.)/(o+1.))+(1.95683674747)*dim
        ref[2,2]=(2.62691525074)*(1+2.*(dim-1.)/(o+1.))+(-1.04797525533)*dim
        ref[2,3]=(0.406171075551)*(1+2.*(dim-1.)/(o+1.))+(-0.272889047794)*dim
        ref[2,4]=(-0.612021580374)*(1+2.*(dim-1.)/(o+1.))+(2.06156124471)*dim
        ref[3,0]=(0.418299203908)*(1+2.*(dim-1.)/(o+1.))+(0.0634441434621)*dim
        ref[3,1]=(1.26433399499)*(1+2.*(dim-1.)/(o+1.))+(0.701201636471)*dim
        ref[3,2]=(0.664903012404)*(1+2.*(dim-1.)/(o+1.))+(-0.699893426186)*dim
        ref[3,3]=(0.739767362472)*(1+2.*(dim-1.)/(o+1.))+(-0.0994795202511)*dim
        ref[3,4]=(-0.0474526571443)*(1+2.*(dim-1.)/(o+1.))+(-0.325461283346)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank3(self):
      """
      tests integral of rank 3 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2))
      s=Data(0,(6, 2, 2),w)
      ref=numpy.zeros((6, 2, 2),numpy.float_)
      if dim==2:
        s[0,0,0]=(0.589365534484)*x[0]**o+(-0.562703733563)*x[0]+(-0.374217229188)*x[1]**o+(-0.39259467848)*x[1]
        s[0,0,1]=(-0.530861865638)*x[0]**o+(-0.452189463083)*x[0]+(0.130627913931)*x[1]**o+(0.401827715746)*x[1]
        s[0,1,0]=(-0.698034316811)*x[0]**o+(0.95858318685)*x[0]+(0.818020760864)*x[1]**o+(-0.842035209691)*x[1]
        s[0,1,1]=(0.784763132752)*x[0]**o+(-0.345161441944)*x[0]+(-0.0915389865806)*x[1]**o+(0.0413615155914)*x[1]
        s[1,0,0]=(0.844613455277)*x[0]**o+(-0.0227629508737)*x[0]+(0.435622185761)*x[1]**o+(-0.0215311374618)*x[1]
        s[1,0,1]=(-0.0669862120561)*x[0]**o+(-0.846007346441)*x[0]+(-0.413495349659)*x[1]**o+(0.535169065488)*x[1]
        s[1,1,0]=(-0.685697856883)*x[0]**o+(0.958953550286)*x[0]+(0.154980083392)*x[1]**o+(0.0800899796196)*x[1]
        s[1,1,1]=(0.912269483572)*x[0]**o+(0.94047005728)*x[0]+(-0.304144591655)*x[1]**o+(-0.953778202636)*x[1]
        s[2,0,0]=(-0.743960822491)*x[0]**o+(-0.550611814396)*x[0]+(-0.725685162705)*x[1]**o+(-0.339527538083)*x[1]
        s[2,0,1]=(-0.889607185943)*x[0]**o+(0.539329041411)*x[0]+(-0.161095005544)*x[1]**o+(0.0793331043418)*x[1]
        s[2,1,0]=(-0.468621581571)*x[0]**o+(-0.686831337537)*x[0]+(0.097676174696)*x[1]**o+(-0.661357348914)*x[1]
        s[2,1,1]=(0.272669628304)*x[0]**o+(-0.0973814107905)*x[0]+(0.934138436983)*x[1]**o+(0.886285785082)*x[1]
        s[3,0,0]=(0.732970432548)*x[0]**o+(0.867739265413)*x[0]+(-0.868701163584)*x[1]**o+(-0.845146120935)*x[1]
        s[3,0,1]=(0.655147591399)*x[0]**o+(-0.604461769582)*x[0]+(-0.855962338468)*x[1]**o+(-0.256146581175)*x[1]
        s[3,1,0]=(0.818629391238)*x[0]**o+(-0.949077886388)*x[0]+(0.620697264485)*x[1]**o+(0.161138363607)*x[1]
        s[3,1,1]=(0.388279443227)*x[0]**o+(-0.905781758355)*x[0]+(0.224661220296)*x[1]**o+(-0.160247355785)*x[1]
        s[4,0,0]=(0.933160401571)*x[0]**o+(0.238977604536)*x[0]+(-0.819630167281)*x[1]**o+(-0.787572050579)*x[1]
        s[4,0,1]=(-0.998710300262)*x[0]**o+(-0.311299346807)*x[0]+(0.01512022099)*x[1]**o+(0.987659405648)*x[1]
        s[4,1,0]=(-0.394579388382)*x[0]**o+(0.368576283401)*x[0]+(0.523842693559)*x[1]**o+(0.053237259039)*x[1]
        s[4,1,1]=(-0.899554766025)*x[0]**o+(0.0846648733918)*x[0]+(0.0384854956872)*x[1]**o+(0.499251840272)*x[1]
        s[5,0,0]=(0.969036476632)*x[0]**o+(0.555078576958)*x[0]+(0.358154612858)*x[1]**o+(0.988304896709)*x[1]
        s[5,0,1]=(-0.915157232317)*x[0]**o+(0.174863566989)*x[0]+(-0.929161107163)*x[1]**o+(-0.656086746151)*x[1]
        s[5,1,0]=(0.10810079771)*x[0]**o+(-0.929385584784)*x[0]+(-0.390809106715)*x[1]**o+(-0.573678217593)*x[1]
        s[5,1,1]=(0.871461803834)*x[0]**o+(-0.13090668343)*x[0]+(0.479502930134)*x[1]**o+(-0.143336447632)*x[1]
        ref[0,0,0]=(0.215148305297)*(1+2.*(dim-1.)/(o+1.))+(-0.955298412042)*dim
        ref[0,0,1]=(-0.400233951707)*(1+2.*(dim-1.)/(o+1.))+(-0.050361747337)*dim
        ref[0,1,0]=(0.119986444053)*(1+2.*(dim-1.)/(o+1.))+(0.116547977158)*dim
        ref[0,1,1]=(0.693224146172)*(1+2.*(dim-1.)/(o+1.))+(-0.303799926353)*dim
        ref[1,0,0]=(1.28023564104)*(1+2.*(dim-1.)/(o+1.))+(-0.0442940883355)*dim
        ref[1,0,1]=(-0.480481561715)*(1+2.*(dim-1.)/(o+1.))+(-0.310838280953)*dim
        ref[1,1,0]=(-0.530717773491)*(1+2.*(dim-1.)/(o+1.))+(1.03904352991)*dim
        ref[1,1,1]=(0.608124891917)*(1+2.*(dim-1.)/(o+1.))+(-0.0133081453554)*dim
        ref[2,0,0]=(-1.4696459852)*(1+2.*(dim-1.)/(o+1.))+(-0.890139352479)*dim
        ref[2,0,1]=(-1.05070219149)*(1+2.*(dim-1.)/(o+1.))+(0.618662145753)*dim
        ref[2,1,0]=(-0.370945406875)*(1+2.*(dim-1.)/(o+1.))+(-1.34818868645)*dim
        ref[2,1,1]=(1.20680806529)*(1+2.*(dim-1.)/(o+1.))+(0.788904374291)*dim
        ref[3,0,0]=(-0.135730731035)*(1+2.*(dim-1.)/(o+1.))+(0.0225931444774)*dim
        ref[3,0,1]=(-0.200814747068)*(1+2.*(dim-1.)/(o+1.))+(-0.860608350757)*dim
        ref[3,1,0]=(1.43932665572)*(1+2.*(dim-1.)/(o+1.))+(-0.787939522781)*dim
        ref[3,1,1]=(0.612940663523)*(1+2.*(dim-1.)/(o+1.))+(-1.06602911414)*dim
        ref[4,0,0]=(0.113530234291)*(1+2.*(dim-1.)/(o+1.))+(-0.548594446042)*dim
        ref[4,0,1]=(-0.983590079272)*(1+2.*(dim-1.)/(o+1.))+(0.676360058841)*dim
        ref[4,1,0]=(0.129263305177)*(1+2.*(dim-1.)/(o+1.))+(0.42181354244)*dim
        ref[4,1,1]=(-0.861069270338)*(1+2.*(dim-1.)/(o+1.))+(0.583916713663)*dim
        ref[5,0,0]=(1.32719108949)*(1+2.*(dim-1.)/(o+1.))+(1.54338347367)*dim
        ref[5,0,1]=(-1.84431833948)*(1+2.*(dim-1.)/(o+1.))+(-0.481223179161)*dim
        ref[5,1,0]=(-0.282708309005)*(1+2.*(dim-1.)/(o+1.))+(-1.50306380238)*dim
        ref[5,1,1]=(1.35096473397)*(1+2.*(dim-1.)/(o+1.))+(-0.274243131062)*dim
      else:
        s[0,0,0]=(0.107513098422)*x[0]**o+(0.630914018595)*x[0]+(0.223228652263)*x[1]**o+(-0.817313865632)*x[1]+(-0.131761985412)*x[2]**o+(0.765852452518)*x[2]
        s[0,0,1]=(0.677582669653)*x[0]**o+(-0.180457590671)*x[0]+(-0.623782106139)*x[1]**o+(-0.525225679629)*x[1]+(-0.506475447696)*x[2]**o+(0.426953870606)*x[2]
        s[0,1,0]=(0.638676595716)*x[0]**o+(0.401321387041)*x[0]+(0.326953813244)*x[1]**o+(0.376976149503)*x[1]+(0.37300668271)*x[2]**o+(-0.220329878781)*x[2]
        s[0,1,1]=(0.801707476852)*x[0]**o+(0.279766359037)*x[0]+(-0.119071566643)*x[1]**o+(0.842943055619)*x[1]+(-0.186390184701)*x[2]**o+(-0.174806078851)*x[2]
        s[1,0,0]=(0.636253830592)*x[0]**o+(0.0451704175721)*x[0]+(0.160762650006)*x[1]**o+(-0.520550899101)*x[1]+(0.104281196353)*x[2]**o+(-0.158693150263)*x[2]
        s[1,0,1]=(-0.529589537208)*x[0]**o+(0.883520631827)*x[0]+(-0.945287489028)*x[1]**o+(-0.402371494207)*x[1]+(-0.347776652964)*x[2]**o+(-0.801910128228)*x[2]
        s[1,1,0]=(-0.0695656121071)*x[0]**o+(0.318785505529)*x[0]+(-0.755792368009)*x[1]**o+(0.965500773902)*x[1]+(-0.963131216861)*x[2]**o+(-0.470019270105)*x[2]
        s[1,1,1]=(0.0313155860155)*x[0]**o+(-0.745224876276)*x[0]+(0.941969004994)*x[1]**o+(0.0868580544081)*x[1]+(0.814837112058)*x[2]**o+(0.6031495013)*x[2]
        s[2,0,0]=(0.326069961839)*x[0]**o+(-0.226878245433)*x[0]+(0.181649890734)*x[1]**o+(0.124231468404)*x[1]+(-0.607776937521)*x[2]**o+(-0.42430990239)*x[2]
        s[2,0,1]=(-0.0362788731491)*x[0]**o+(0.405545167692)*x[0]+(0.234108391034)*x[1]**o+(0.0613492588911)*x[1]+(0.375991408638)*x[2]**o+(-0.08364867034)*x[2]
        s[2,1,0]=(-0.739151142745)*x[0]**o+(0.895090860344)*x[0]+(-0.892542071223)*x[1]**o+(-0.925594662939)*x[1]+(0.527652868204)*x[2]**o+(-0.291230735869)*x[2]
        s[2,1,1]=(-0.569990909361)*x[0]**o+(0.135614304968)*x[0]+(0.644435688812)*x[1]**o+(0.607772804184)*x[1]+(-0.671154751626)*x[2]**o+(0.908859370499)*x[2]
        s[3,0,0]=(0.678101574906)*x[0]**o+(0.393260837804)*x[0]+(-0.364076766107)*x[1]**o+(0.221403168918)*x[1]+(-0.0881262519405)*x[2]**o+(0.243571413051)*x[2]
        s[3,0,1]=(-0.889370605049)*x[0]**o+(-0.751122823303)*x[0]+(-0.23702556553)*x[1]**o+(0.758207050746)*x[1]+(-0.653730045583)*x[2]**o+(-0.881328128574)*x[2]
        s[3,1,0]=(0.99850080224)*x[0]**o+(-0.13489113732)*x[0]+(-0.593980699882)*x[1]**o+(0.972643960626)*x[1]+(-0.0329293151824)*x[2]**o+(-0.432361021975)*x[2]
        s[3,1,1]=(-0.104833428698)*x[0]**o+(-0.937873960299)*x[0]+(-0.328611446648)*x[1]**o+(0.182782391112)*x[1]+(-0.565312252294)*x[2]**o+(0.0785805823337)*x[2]
        s[4,0,0]=(-0.833754760965)*x[0]**o+(-0.942110442063)*x[0]+(-0.544558564212)*x[1]**o+(-0.496169424902)*x[1]+(-0.805940323103)*x[2]**o+(0.355543043698)*x[2]
        s[4,0,1]=(0.1228603677)*x[0]**o+(0.27927470376)*x[0]+(0.790110425171)*x[1]**o+(0.345056958981)*x[1]+(-0.681665552406)*x[2]**o+(-0.263090945447)*x[2]
        s[4,1,0]=(0.845338153049)*x[0]**o+(0.00877377111324)*x[0]+(0.0538302301089)*x[1]**o+(-0.881754421435)*x[1]+(0.432818307915)*x[2]**o+(0.381735394491)*x[2]
        s[4,1,1]=(-0.679482650754)*x[0]**o+(0.111691609594)*x[0]+(-0.984942957975)*x[1]**o+(0.638763928487)*x[1]+(-0.146178908969)*x[2]**o+(0.752462675681)*x[2]
        s[5,0,0]=(-0.698554091901)*x[0]**o+(0.603862483307)*x[0]+(-0.627044519178)*x[1]**o+(-0.173483224843)*x[1]+(-0.794733618981)*x[2]**o+(0.148563520039)*x[2]
        s[5,0,1]=(-0.0422516673963)*x[0]**o+(0.551902135908)*x[0]+(-0.0620401959854)*x[1]**o+(-0.462654965568)*x[1]+(-0.853501426111)*x[2]**o+(-0.168566246091)*x[2]
        s[5,1,0]=(-0.482405425466)*x[0]**o+(0.215434393331)*x[0]+(0.0597508965493)*x[1]**o+(0.163372916375)*x[1]+(-0.0782540165278)*x[2]**o+(-0.29216681971)*x[2]
        s[5,1,1]=(-0.849963259001)*x[0]**o+(0.0653198724433)*x[0]+(0.673346071067)*x[1]**o+(-0.98400726868)*x[1]+(-0.545894674138)*x[2]**o+(-0.128229571203)*x[2]
        ref[0,0,0]=(0.198979765273)*(1+2.*(dim-1.)/(o+1.))+(0.579452605481)*dim
        ref[0,0,1]=(-0.452674884183)*(1+2.*(dim-1.)/(o+1.))+(-0.278729399693)*dim
        ref[0,1,0]=(1.33863709167)*(1+2.*(dim-1.)/(o+1.))+(0.557967657763)*dim
        ref[0,1,1]=(0.496245725508)*(1+2.*(dim-1.)/(o+1.))+(0.947903335805)*dim
        ref[1,0,0]=(0.901297676951)*(1+2.*(dim-1.)/(o+1.))+(-0.634073631792)*dim
        ref[1,0,1]=(-1.8226536792)*(1+2.*(dim-1.)/(o+1.))+(-0.320760990609)*dim
        ref[1,1,0]=(-1.78848919698)*(1+2.*(dim-1.)/(o+1.))+(0.814267009326)*dim
        ref[1,1,1]=(1.78812170307)*(1+2.*(dim-1.)/(o+1.))+(-0.0552173205681)*dim
        ref[2,0,0]=(-0.100057084948)*(1+2.*(dim-1.)/(o+1.))+(-0.526956679419)*dim
        ref[2,0,1]=(0.573820926522)*(1+2.*(dim-1.)/(o+1.))+(0.383245756243)*dim
        ref[2,1,0]=(-1.10404034576)*(1+2.*(dim-1.)/(o+1.))+(-0.321734538465)*dim
        ref[2,1,1]=(-0.596709972174)*(1+2.*(dim-1.)/(o+1.))+(1.65224647965)*dim
        ref[3,0,0]=(0.225898556859)*(1+2.*(dim-1.)/(o+1.))+(0.858235419772)*dim
        ref[3,0,1]=(-1.78012621616)*(1+2.*(dim-1.)/(o+1.))+(-0.874243901131)*dim
        ref[3,1,0]=(0.371590787175)*(1+2.*(dim-1.)/(o+1.))+(0.405391801331)*dim
        ref[3,1,1]=(-0.99875712764)*(1+2.*(dim-1.)/(o+1.))+(-0.676510986853)*dim
        ref[4,0,0]=(-2.18425364828)*(1+2.*(dim-1.)/(o+1.))+(-1.08273682327)*dim
        ref[4,0,1]=(0.231305240466)*(1+2.*(dim-1.)/(o+1.))+(0.361240717293)*dim
        ref[4,1,0]=(1.33198669107)*(1+2.*(dim-1.)/(o+1.))+(-0.49124525583)*dim
        ref[4,1,1]=(-1.8106045177)*(1+2.*(dim-1.)/(o+1.))+(1.50291821376)*dim
        ref[5,0,0]=(-2.12033223006)*(1+2.*(dim-1.)/(o+1.))+(0.578942778503)*dim
        ref[5,0,1]=(-0.957793289493)*(1+2.*(dim-1.)/(o+1.))+(-0.0793190757515)*dim
        ref[5,1,0]=(-0.500908545444)*(1+2.*(dim-1.)/(o+1.))+(0.0866404899966)*dim
        ref[5,1,1]=(-0.722511862072)*(1+2.*(dim-1.)/(o+1.))+(-1.04691696744)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_integrate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank4(self):
      """
      tests integral of rank 4 Symbol on the ReducedFunctionOnBoundary

      assumptions: ReducedFunctionOnBoundary(self.domain) exists
                   self.domain supports integral on ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2))
      s=Data(0,(4, 5, 3, 2),w)
      ref=numpy.zeros((4, 5, 3, 2),numpy.float_)
      if dim==2:
        s[0,0,0,0]=(-0.938527612719)*x[0]**o+(0.18109071498)*x[0]+(0.33583450429)*x[1]**o+(0.875633806815)*x[1]
        s[0,0,0,1]=(0.235264331576)*x[0]**o+(-0.916078537073)*x[0]+(-0.564923840072)*x[1]**o+(0.241815090592)*x[1]
        s[0,0,1,0]=(0.465988219999)*x[0]**o+(-0.234861997786)*x[0]+(-0.826589503087)*x[1]**o+(-0.912695921201)*x[1]
        s[0,0,1,1]=(0.683584251749)*x[0]**o+(-0.445524319089)*x[0]+(-0.794091275716)*x[1]**o+(0.960124513284)*x[1]
        s[0,0,2,0]=(0.864832412476)*x[0]**o+(0.545257658859)*x[0]+(-0.0523175406047)*x[1]**o+(0.637801499066)*x[1]
        s[0,0,2,1]=(-0.345749442294)*x[0]**o+(-0.62275362184)*x[0]+(-0.445774008922)*x[1]**o+(-0.883708969162)*x[1]
        s[0,1,0,0]=(0.495428396131)*x[0]**o+(-0.688537966744)*x[0]+(0.974975943301)*x[1]**o+(0.812712252674)*x[1]
        s[0,1,0,1]=(0.288058971992)*x[0]**o+(0.108248232119)*x[0]+(0.717277962001)*x[1]**o+(-0.27243719684)*x[1]
        s[0,1,1,0]=(0.0359145438363)*x[0]**o+(0.536166090023)*x[0]+(-0.00355608454499)*x[1]**o+(-0.806427949542)*x[1]
        s[0,1,1,1]=(0.166147370504)*x[0]**o+(-0.627148508537)*x[0]+(-0.786840545498)*x[1]**o+(0.281613120233)*x[1]
        s[0,1,2,0]=(-0.133612757005)*x[0]**o+(0.392265237404)*x[0]+(0.453687302779)*x[1]**o+(-0.773645262693)*x[1]
        s[0,1,2,1]=(-0.716222674195)*x[0]**o+(0.670906647893)*x[0]+(-0.462562888282)*x[1]**o+(-0.624961145925)*x[1]
        s[0,2,0,0]=(0.0194565885353)*x[0]**o+(0.182692282753)*x[0]+(-0.638358622688)*x[1]**o+(0.335689393108)*x[1]
        s[0,2,0,1]=(-0.860405914558)*x[0]**o+(0.632285612174)*x[0]+(0.0656792325224)*x[1]**o+(-0.156015955877)*x[1]
        s[0,2,1,0]=(-0.152134223629)*x[0]**o+(-0.862936330258)*x[0]+(0.417270798339)*x[1]**o+(0.963136816902)*x[1]
        s[0,2,1,1]=(-0.258242824318)*x[0]**o+(0.971045381488)*x[0]+(-0.241220971892)*x[1]**o+(0.856378952246)*x[1]
        s[0,2,2,0]=(0.32072997937)*x[0]**o+(0.704100826786)*x[0]+(-0.33756098405)*x[1]**o+(-0.664910217598)*x[1]
        s[0,2,2,1]=(-0.61313065356)*x[0]**o+(0.508087982157)*x[0]+(-0.84053545881)*x[1]**o+(0.160765183388)*x[1]
        s[0,3,0,0]=(-0.75820426921)*x[0]**o+(-0.90659534133)*x[0]+(0.29703074695)*x[1]**o+(-0.0460927434434)*x[1]
        s[0,3,0,1]=(-0.168059238195)*x[0]**o+(0.444168839682)*x[0]+(-0.199109031605)*x[1]**o+(-0.0951683017552)*x[1]
        s[0,3,1,0]=(0.134237505423)*x[0]**o+(-0.793789958021)*x[0]+(-0.806546707234)*x[1]**o+(0.841229990055)*x[1]
        s[0,3,1,1]=(0.0190198296544)*x[0]**o+(-0.535272381244)*x[0]+(0.961010454009)*x[1]**o+(0.00558376325046)*x[1]
        s[0,3,2,0]=(0.363866563396)*x[0]**o+(-0.479776283958)*x[0]+(-0.189244785869)*x[1]**o+(0.807338961992)*x[1]
        s[0,3,2,1]=(0.160239702851)*x[0]**o+(-0.629633641475)*x[0]+(-0.116962193936)*x[1]**o+(-0.0779960566649)*x[1]
        s[0,4,0,0]=(-0.778019803229)*x[0]**o+(0.800046035892)*x[0]+(-0.164619354127)*x[1]**o+(-0.785540965821)*x[1]
        s[0,4,0,1]=(-0.302058711888)*x[0]**o+(0.560865269822)*x[0]+(-0.473079568526)*x[1]**o+(0.995387605735)*x[1]
        s[0,4,1,0]=(-0.411379127233)*x[0]**o+(-0.55185696669)*x[0]+(0.306097815722)*x[1]**o+(0.336660861462)*x[1]
        s[0,4,1,1]=(0.449818597104)*x[0]**o+(0.102791276029)*x[0]+(0.795639210789)*x[1]**o+(0.061474861224)*x[1]
        s[0,4,2,0]=(-0.0490155784051)*x[0]**o+(-0.429224063812)*x[0]+(-0.0368599977199)*x[1]**o+(0.718948163937)*x[1]
        s[0,4,2,1]=(-0.0195621009238)*x[0]**o+(0.162744359327)*x[0]+(0.156319239232)*x[1]**o+(-0.840307263668)*x[1]
        s[1,0,0,0]=(0.818322838345)*x[0]**o+(0.75605127429)*x[0]+(0.0160223113114)*x[1]**o+(0.733116071572)*x[1]
        s[1,0,0,1]=(-0.516143096482)*x[0]**o+(-0.607491330321)*x[0]+(-0.71990703998)*x[1]**o+(0.74041801157)*x[1]
        s[1,0,1,0]=(0.087735317248)*x[0]**o+(0.811033616338)*x[0]+(-0.596466275725)*x[1]**o+(-0.632928705475)*x[1]
        s[1,0,1,1]=(0.574370862223)*x[0]**o+(-0.666343550396)*x[0]+(0.955788950863)*x[1]**o+(-0.901731521626)*x[1]
        s[1,0,2,0]=(-0.0749607390632)*x[0]**o+(-0.431912603732)*x[0]+(-0.292096378559)*x[1]**o+(0.184727807798)*x[1]
        s[1,0,2,1]=(-0.680339079773)*x[0]**o+(-0.751178966013)*x[0]+(0.417442581996)*x[1]**o+(0.173275667315)*x[1]
        s[1,1,0,0]=(0.549976167744)*x[0]**o+(-0.542891673039)*x[0]+(0.780157563665)*x[1]**o+(-0.500193881521)*x[1]
        s[1,1,0,1]=(0.741392053681)*x[0]**o+(0.464245280846)*x[0]+(0.893822624677)*x[1]**o+(-0.726733112592)*x[1]
        s[1,1,1,0]=(0.0669632365681)*x[0]**o+(-0.19242503062)*x[0]+(-0.016747330039)*x[1]**o+(0.565528468045)*x[1]
        s[1,1,1,1]=(0.0478230922838)*x[0]**o+(-0.490882982306)*x[0]+(0.71464197649)*x[1]**o+(0.647758969096)*x[1]
        s[1,1,2,0]=(0.0459694337421)*x[0]**o+(0.22219598918)*x[0]+(0.278878633519)*x[1]**o+(0.0964919048896)*x[1]
        s[1,1,2,1]=(0.951517313537)*x[0]**o+(0.776873672124)*x[0]+(0.276827457342)*x[1]**o+(0.97564662187)*x[1]
        s[1,2,0,0]=(-0.756103809082)*x[0]**o+(-0.776671770692)*x[0]+(0.305546356471)*x[1]**o+(0.61267739505)*x[1]
        s[1,2,0,1]=(-0.00486328528121)*x[0]**o+(0.969855303757)*x[0]+(-0.803341508101)*x[1]**o+(-0.973192859371)*x[1]
        s[1,2,1,0]=(-0.233307547886)*x[0]**o+(0.480312206288)*x[0]+(-0.775060924028)*x[1]**o+(0.371008519793)*x[1]
        s[1,2,1,1]=(-0.548323703436)*x[0]**o+(-0.648506076259)*x[0]+(-0.175775778303)*x[1]**o+(0.860353396366)*x[1]
        s[1,2,2,0]=(-0.631591638879)*x[0]**o+(0.64118025586)*x[0]+(0.791112489411)*x[1]**o+(-0.503438016129)*x[1]
        s[1,2,2,1]=(-0.322064055084)*x[0]**o+(0.513203966068)*x[0]+(-0.123194392107)*x[1]**o+(-0.0530322384215)*x[1]
        s[1,3,0,0]=(0.690957542061)*x[0]**o+(0.582367141789)*x[0]+(0.0548888588807)*x[1]**o+(-0.396676526096)*x[1]
        s[1,3,0,1]=(-0.887786079331)*x[0]**o+(-0.595071924485)*x[0]+(-0.585573810513)*x[1]**o+(0.966304199567)*x[1]
        s[1,3,1,0]=(0.0813018143573)*x[0]**o+(-0.97460530802)*x[0]+(0.157284332761)*x[1]**o+(-0.282617607243)*x[1]
        s[1,3,1,1]=(0.876197676377)*x[0]**o+(-0.151980499847)*x[0]+(-0.605310496113)*x[1]**o+(-0.444245831277)*x[1]
        s[1,3,2,0]=(0.342314316145)*x[0]**o+(-0.448750580332)*x[0]+(0.42604855142)*x[1]**o+(0.646516588045)*x[1]
        s[1,3,2,1]=(0.740620038411)*x[0]**o+(-0.519019257788)*x[0]+(-0.625195631671)*x[1]**o+(0.0692954637811)*x[1]
        s[1,4,0,0]=(-0.444071318016)*x[0]**o+(-0.471544009206)*x[0]+(-0.696279794002)*x[1]**o+(-0.902974292958)*x[1]
        s[1,4,0,1]=(0.542533721359)*x[0]**o+(-0.0736449480756)*x[0]+(-0.408131697388)*x[1]**o+(0.364714013702)*x[1]
        s[1,4,1,0]=(-0.152939586181)*x[0]**o+(0.77896231429)*x[0]+(-0.815480658966)*x[1]**o+(0.154938953512)*x[1]
        s[1,4,1,1]=(-0.29007531171)*x[0]**o+(-0.352728646039)*x[0]+(0.811434496017)*x[1]**o+(0.738596503811)*x[1]
        s[1,4,2,0]=(0.219126056868)*x[0]**o+(-0.888054898232)*x[0]+(0.832367086102)*x[1]**o+(0.423282741614)*x[1]
        s[1,4,2,1]=(-0.0215001265455)*x[0]**o+(-0.267591836227)*x[0]+(0.16855162109)*x[1]**o+(0.450976714016)*x[1]
        s[2,0,0,0]=(0.55451395254)*x[0]**o+(-0.569774836527)*x[0]+(-0.950939837183)*x[1]**o+(0.963606409995)*x[1]
        s[2,0,0,1]=(0.768925257395)*x[0]**o+(-0.271693392598)*x[0]+(0.85042078878)*x[1]**o+(0.854821782113)*x[1]
        s[2,0,1,0]=(-0.0681124804964)*x[0]**o+(0.346262969724)*x[0]+(-0.224031702809)*x[1]**o+(0.184665191293)*x[1]
        s[2,0,1,1]=(0.215579225437)*x[0]**o+(0.650089014616)*x[0]+(-0.632072599272)*x[1]**o+(0.974448958372)*x[1]
        s[2,0,2,0]=(0.943481715059)*x[0]**o+(-0.91197962542)*x[0]+(-0.552378744981)*x[1]**o+(0.958940554729)*x[1]
        s[2,0,2,1]=(0.467266364031)*x[0]**o+(-0.253321061857)*x[0]+(0.935529481271)*x[1]**o+(-0.203578102406)*x[1]
        s[2,1,0,0]=(0.0885012656958)*x[0]**o+(-0.0568114758907)*x[0]+(-0.94244176099)*x[1]**o+(-0.151572915453)*x[1]
        s[2,1,0,1]=(-0.583513164141)*x[0]**o+(0.299189697674)*x[0]+(0.129798849695)*x[1]**o+(0.0486488675811)*x[1]
        s[2,1,1,0]=(-0.230047392884)*x[0]**o+(-0.215141703649)*x[0]+(0.597873677289)*x[1]**o+(-0.104893786277)*x[1]
        s[2,1,1,1]=(-0.193340204287)*x[0]**o+(0.172114988052)*x[0]+(-0.912123490435)*x[1]**o+(0.386237743741)*x[1]
        s[2,1,2,0]=(-0.355870799626)*x[0]**o+(-0.636394104532)*x[0]+(0.14161953364)*x[1]**o+(0.793881407217)*x[1]
        s[2,1,2,1]=(0.623415964738)*x[0]**o+(-0.441080834978)*x[0]+(-0.669760052979)*x[1]**o+(0.504915334197)*x[1]
        s[2,2,0,0]=(0.843403522019)*x[0]**o+(0.497098510545)*x[0]+(0.668852088742)*x[1]**o+(-0.469226675749)*x[1]
        s[2,2,0,1]=(-0.924635841352)*x[0]**o+(0.200500164001)*x[0]+(-0.230616017277)*x[1]**o+(0.758688672394)*x[1]
        s[2,2,1,0]=(-0.180287456852)*x[0]**o+(-0.189123451964)*x[0]+(-0.359118343342)*x[1]**o+(-0.856215073416)*x[1]
        s[2,2,1,1]=(-0.270882735947)*x[0]**o+(0.958913592886)*x[0]+(-0.694010263633)*x[1]**o+(0.860782778563)*x[1]
        s[2,2,2,0]=(-0.808819032861)*x[0]**o+(-0.733697428319)*x[0]+(-0.460941216629)*x[1]**o+(-0.617344596558)*x[1]
        s[2,2,2,1]=(0.451979026099)*x[0]**o+(-0.0858303927258)*x[0]+(0.163184369333)*x[1]**o+(0.801763368843)*x[1]
        s[2,3,0,0]=(-0.367005493485)*x[0]**o+(0.854411289003)*x[0]+(0.733528289773)*x[1]**o+(0.791392876143)*x[1]
        s[2,3,0,1]=(0.911579456326)*x[0]**o+(-0.690061379515)*x[0]+(-0.0831450710122)*x[1]**o+(-0.815038630875)*x[1]
        s[2,3,1,0]=(-0.179108077134)*x[0]**o+(0.609583032751)*x[0]+(-0.16765650892)*x[1]**o+(0.861847300762)*x[1]
        s[2,3,1,1]=(0.384373942853)*x[0]**o+(0.517209389691)*x[0]+(0.841154335433)*x[1]**o+(0.579356331563)*x[1]
        s[2,3,2,0]=(-0.683926277829)*x[0]**o+(0.315128536718)*x[0]+(0.394946403402)*x[1]**o+(0.679738641838)*x[1]
        s[2,3,2,1]=(0.466676304045)*x[0]**o+(-0.0400354010415)*x[0]+(-0.760865349977)*x[1]**o+(0.928180066121)*x[1]
        s[2,4,0,0]=(-0.497221654148)*x[0]**o+(-0.921245715417)*x[0]+(0.908610543692)*x[1]**o+(-0.662743777314)*x[1]
        s[2,4,0,1]=(-0.218025628376)*x[0]**o+(0.981616175697)*x[0]+(-0.300066300268)*x[1]**o+(-0.145065054659)*x[1]
        s[2,4,1,0]=(0.822321927296)*x[0]**o+(0.697029639265)*x[0]+(0.21672820186)*x[1]**o+(-0.136979319447)*x[1]
        s[2,4,1,1]=(-0.33254584204)*x[0]**o+(0.0804948126216)*x[0]+(0.931652185323)*x[1]**o+(-0.206973479206)*x[1]
        s[2,4,2,0]=(-0.20645245153)*x[0]**o+(0.183027858308)*x[0]+(-0.889749226799)*x[1]**o+(0.177254293812)*x[1]
        s[2,4,2,1]=(0.67936145385)*x[0]**o+(0.0269414327289)*x[0]+(-0.892654968646)*x[1]**o+(-0.211494218552)*x[1]
        s[3,0,0,0]=(-0.221318514547)*x[0]**o+(0.282480830569)*x[0]+(0.428045462357)*x[1]**o+(0.6335381978)*x[1]
        s[3,0,0,1]=(-0.052602851764)*x[0]**o+(-0.208342592426)*x[0]+(-0.518314637297)*x[1]**o+(0.0538977682478)*x[1]
        s[3,0,1,0]=(-0.879184433575)*x[0]**o+(-0.413590155319)*x[0]+(-0.973398939671)*x[1]**o+(0.880259070216)*x[1]
        s[3,0,1,1]=(-0.160095692245)*x[0]**o+(-0.215405690187)*x[0]+(0.049417328483)*x[1]**o+(0.755528105808)*x[1]
        s[3,0,2,0]=(0.0775811014995)*x[0]**o+(0.574331743034)*x[0]+(0.656316191524)*x[1]**o+(-0.485245137216)*x[1]
        s[3,0,2,1]=(0.235819746353)*x[0]**o+(-0.0212642690609)*x[0]+(0.519388187993)*x[1]**o+(-0.73686111506)*x[1]
        s[3,1,0,0]=(0.895857350075)*x[0]**o+(0.993441070954)*x[0]+(0.423289404334)*x[1]**o+(0.648503660982)*x[1]
        s[3,1,0,1]=(-0.508341685048)*x[0]**o+(0.132799887315)*x[0]+(-0.615279575964)*x[1]**o+(0.00296513970359)*x[1]
        s[3,1,1,0]=(0.691638662949)*x[0]**o+(-0.432982022284)*x[0]+(-0.724535941079)*x[1]**o+(0.0336147554659)*x[1]
        s[3,1,1,1]=(-0.431062234755)*x[0]**o+(-0.795503344318)*x[0]+(0.302222798071)*x[1]**o+(-0.0813627256685)*x[1]
        s[3,1,2,0]=(-0.695560929102)*x[0]**o+(0.328886265482)*x[0]+(-0.187761431398)*x[1]**o+(-0.997772947746)*x[1]
        s[3,1,2,1]=(0.551492819689)*x[0]**o+(0.00738597993558)*x[0]+(-0.920251030964)*x[1]**o+(-0.209623547453)*x[1]
        s[3,2,0,0]=(-0.945715173023)*x[0]**o+(-0.413860607376)*x[0]+(-0.0065717404946)*x[1]**o+(-0.340070315229)*x[1]
        s[3,2,0,1]=(-0.866814974957)*x[0]**o+(0.190612129778)*x[0]+(-0.323173980343)*x[1]**o+(-0.221048196515)*x[1]
        s[3,2,1,0]=(0.19860994374)*x[0]**o+(-0.521137863979)*x[0]+(-0.347340372612)*x[1]**o+(0.85944792998)*x[1]
        s[3,2,1,1]=(-0.128916019272)*x[0]**o+(-0.850906742746)*x[0]+(0.258631463969)*x[1]**o+(-0.201200252475)*x[1]
        s[3,2,2,0]=(-0.330614479011)*x[0]**o+(-0.489889065789)*x[0]+(-0.244206440351)*x[1]**o+(-0.322097122215)*x[1]
        s[3,2,2,1]=(0.711726054557)*x[0]**o+(0.63375724563)*x[0]+(-0.564782988704)*x[1]**o+(-0.110402977102)*x[1]
        s[3,3,0,0]=(0.994364369344)*x[0]**o+(-0.112814477225)*x[0]+(-0.277853585926)*x[1]**o+(0.952622614986)*x[1]
        s[3,3,0,1]=(-0.4170514524)*x[0]**o+(0.890841357027)*x[0]+(-0.821412916413)*x[1]**o+(0.795074763635)*x[1]
        s[3,3,1,0]=(-0.0247320805051)*x[0]**o+(-0.204605869304)*x[0]+(0.24262118978)*x[1]**o+(0.546548720614)*x[1]
        s[3,3,1,1]=(-0.474740509006)*x[0]**o+(-0.350277551474)*x[0]+(0.596007930161)*x[1]**o+(0.310969198868)*x[1]
        s[3,3,2,0]=(0.132049782738)*x[0]**o+(-0.913784955142)*x[0]+(-0.189796554115)*x[1]**o+(-0.0703212084825)*x[1]
        s[3,3,2,1]=(-0.831318298996)*x[0]**o+(0.759034430202)*x[0]+(-0.642082491468)*x[1]**o+(-0.0998337089515)*x[1]
        s[3,4,0,0]=(-0.947142357256)*x[0]**o+(-0.358864598356)*x[0]+(0.451923869609)*x[1]**o+(0.46347113305)*x[1]
        s[3,4,0,1]=(-0.935820217362)*x[0]**o+(-0.896350370731)*x[0]+(0.939373432056)*x[1]**o+(-0.367350815379)*x[1]
        s[3,4,1,0]=(0.217367733064)*x[0]**o+(-0.348603185172)*x[0]+(-0.846422979973)*x[1]**o+(0.466060007646)*x[1]
        s[3,4,1,1]=(-0.0861518225018)*x[0]**o+(0.129106551242)*x[0]+(-0.563717320923)*x[1]**o+(0.488814871115)*x[1]
        s[3,4,2,0]=(0.864238935579)*x[0]**o+(0.451518233825)*x[0]+(0.515184766232)*x[1]**o+(0.791632366311)*x[1]
        s[3,4,2,1]=(0.226031804126)*x[0]**o+(0.703010170609)*x[0]+(-0.694425471794)*x[1]**o+(-0.398984702926)*x[1]
        ref[0,0,0,0]=(-0.602693108429)*(1+2.*(dim-1.)/(o+1.))+(1.05672452179)*dim
        ref[0,0,0,1]=(-0.329659508496)*(1+2.*(dim-1.)/(o+1.))+(-0.674263446481)*dim
        ref[0,0,1,0]=(-0.360601283088)*(1+2.*(dim-1.)/(o+1.))+(-1.14755791899)*dim
        ref[0,0,1,1]=(-0.110507023967)*(1+2.*(dim-1.)/(o+1.))+(0.514600194196)*dim
        ref[0,0,2,0]=(0.812514871871)*(1+2.*(dim-1.)/(o+1.))+(1.18305915792)*dim
        ref[0,0,2,1]=(-0.791523451216)*(1+2.*(dim-1.)/(o+1.))+(-1.506462591)*dim
        ref[0,1,0,0]=(1.47040433943)*(1+2.*(dim-1.)/(o+1.))+(0.12417428593)*dim
        ref[0,1,0,1]=(1.00533693399)*(1+2.*(dim-1.)/(o+1.))+(-0.164188964721)*dim
        ref[0,1,1,0]=(0.0323584592913)*(1+2.*(dim-1.)/(o+1.))+(-0.270261859519)*dim
        ref[0,1,1,1]=(-0.620693174994)*(1+2.*(dim-1.)/(o+1.))+(-0.345535388304)*dim
        ref[0,1,2,0]=(0.320074545773)*(1+2.*(dim-1.)/(o+1.))+(-0.381380025289)*dim
        ref[0,1,2,1]=(-1.17878556248)*(1+2.*(dim-1.)/(o+1.))+(0.0459455019681)*dim
        ref[0,2,0,0]=(-0.618902034153)*(1+2.*(dim-1.)/(o+1.))+(0.518381675861)*dim
        ref[0,2,0,1]=(-0.794726682036)*(1+2.*(dim-1.)/(o+1.))+(0.476269656297)*dim
        ref[0,2,1,0]=(0.26513657471)*(1+2.*(dim-1.)/(o+1.))+(0.100200486644)*dim
        ref[0,2,1,1]=(-0.49946379621)*(1+2.*(dim-1.)/(o+1.))+(1.82742433373)*dim
        ref[0,2,2,0]=(-0.01683100468)*(1+2.*(dim-1.)/(o+1.))+(0.0391906091876)*dim
        ref[0,2,2,1]=(-1.45366611237)*(1+2.*(dim-1.)/(o+1.))+(0.668853165545)*dim
        ref[0,3,0,0]=(-0.46117352226)*(1+2.*(dim-1.)/(o+1.))+(-0.952688084773)*dim
        ref[0,3,0,1]=(-0.3671682698)*(1+2.*(dim-1.)/(o+1.))+(0.349000537927)*dim
        ref[0,3,1,0]=(-0.672309201811)*(1+2.*(dim-1.)/(o+1.))+(0.0474400320344)*dim
        ref[0,3,1,1]=(0.980030283663)*(1+2.*(dim-1.)/(o+1.))+(-0.529688617993)*dim
        ref[0,3,2,0]=(0.174621777527)*(1+2.*(dim-1.)/(o+1.))+(0.327562678035)*dim
        ref[0,3,2,1]=(0.0432775089146)*(1+2.*(dim-1.)/(o+1.))+(-0.707629698139)*dim
        ref[0,4,0,0]=(-0.942639157356)*(1+2.*(dim-1.)/(o+1.))+(0.0145050700714)*dim
        ref[0,4,0,1]=(-0.775138280415)*(1+2.*(dim-1.)/(o+1.))+(1.55625287556)*dim
        ref[0,4,1,0]=(-0.105281311512)*(1+2.*(dim-1.)/(o+1.))+(-0.215196105228)*dim
        ref[0,4,1,1]=(1.24545780789)*(1+2.*(dim-1.)/(o+1.))+(0.164266137253)*dim
        ref[0,4,2,0]=(-0.0858755761251)*(1+2.*(dim-1.)/(o+1.))+(0.289724100124)*dim
        ref[0,4,2,1]=(0.136757138309)*(1+2.*(dim-1.)/(o+1.))+(-0.677562904341)*dim
        ref[1,0,0,0]=(0.834345149656)*(1+2.*(dim-1.)/(o+1.))+(1.48916734586)*dim
        ref[1,0,0,1]=(-1.23605013646)*(1+2.*(dim-1.)/(o+1.))+(0.132926681249)*dim
        ref[1,0,1,0]=(-0.508730958477)*(1+2.*(dim-1.)/(o+1.))+(0.178104910863)*dim
        ref[1,0,1,1]=(1.53015981309)*(1+2.*(dim-1.)/(o+1.))+(-1.56807507202)*dim
        ref[1,0,2,0]=(-0.367057117622)*(1+2.*(dim-1.)/(o+1.))+(-0.247184795934)*dim
        ref[1,0,2,1]=(-0.262896497777)*(1+2.*(dim-1.)/(o+1.))+(-0.577903298697)*dim
        ref[1,1,0,0]=(1.33013373141)*(1+2.*(dim-1.)/(o+1.))+(-1.04308555456)*dim
        ref[1,1,0,1]=(1.63521467836)*(1+2.*(dim-1.)/(o+1.))+(-0.262487831746)*dim
        ref[1,1,1,0]=(0.0502159065291)*(1+2.*(dim-1.)/(o+1.))+(0.373103437426)*dim
        ref[1,1,1,1]=(0.762465068774)*(1+2.*(dim-1.)/(o+1.))+(0.15687598679)*dim
        ref[1,1,2,0]=(0.324848067261)*(1+2.*(dim-1.)/(o+1.))+(0.318687894069)*dim
        ref[1,1,2,1]=(1.22834477088)*(1+2.*(dim-1.)/(o+1.))+(1.75252029399)*dim
        ref[1,2,0,0]=(-0.450557452611)*(1+2.*(dim-1.)/(o+1.))+(-0.163994375642)*dim
        ref[1,2,0,1]=(-0.808204793382)*(1+2.*(dim-1.)/(o+1.))+(-0.00333755561452)*dim
        ref[1,2,1,0]=(-1.00836847191)*(1+2.*(dim-1.)/(o+1.))+(0.85132072608)*dim
        ref[1,2,1,1]=(-0.724099481739)*(1+2.*(dim-1.)/(o+1.))+(0.211847320106)*dim
        ref[1,2,2,0]=(0.159520850532)*(1+2.*(dim-1.)/(o+1.))+(0.137742239731)*dim
        ref[1,2,2,1]=(-0.44525844719)*(1+2.*(dim-1.)/(o+1.))+(0.460171727646)*dim
        ref[1,3,0,0]=(0.745846400942)*(1+2.*(dim-1.)/(o+1.))+(0.185690615693)*dim
        ref[1,3,0,1]=(-1.47335988984)*(1+2.*(dim-1.)/(o+1.))+(0.371232275082)*dim
        ref[1,3,1,0]=(0.238586147118)*(1+2.*(dim-1.)/(o+1.))+(-1.25722291526)*dim
        ref[1,3,1,1]=(0.270887180264)*(1+2.*(dim-1.)/(o+1.))+(-0.596226331125)*dim
        ref[1,3,2,0]=(0.768362867564)*(1+2.*(dim-1.)/(o+1.))+(0.197766007713)*dim
        ref[1,3,2,1]=(0.11542440674)*(1+2.*(dim-1.)/(o+1.))+(-0.449723794007)*dim
        ref[1,4,0,0]=(-1.14035111202)*(1+2.*(dim-1.)/(o+1.))+(-1.37451830216)*dim
        ref[1,4,0,1]=(0.134402023971)*(1+2.*(dim-1.)/(o+1.))+(0.291069065626)*dim
        ref[1,4,1,0]=(-0.968420245147)*(1+2.*(dim-1.)/(o+1.))+(0.933901267802)*dim
        ref[1,4,1,1]=(0.521359184306)*(1+2.*(dim-1.)/(o+1.))+(0.385867857771)*dim
        ref[1,4,2,0]=(1.05149314297)*(1+2.*(dim-1.)/(o+1.))+(-0.464772156618)*dim
        ref[1,4,2,1]=(0.147051494544)*(1+2.*(dim-1.)/(o+1.))+(0.183384877789)*dim
        ref[2,0,0,0]=(-0.396425884642)*(1+2.*(dim-1.)/(o+1.))+(0.393831573468)*dim
        ref[2,0,0,1]=(1.61934604618)*(1+2.*(dim-1.)/(o+1.))+(0.583128389514)*dim
        ref[2,0,1,0]=(-0.292144183305)*(1+2.*(dim-1.)/(o+1.))+(0.530928161018)*dim
        ref[2,0,1,1]=(-0.416493373835)*(1+2.*(dim-1.)/(o+1.))+(1.62453797299)*dim
        ref[2,0,2,0]=(0.391102970078)*(1+2.*(dim-1.)/(o+1.))+(0.0469609293089)*dim
        ref[2,0,2,1]=(1.4027958453)*(1+2.*(dim-1.)/(o+1.))+(-0.456899164263)*dim
        ref[2,1,0,0]=(-0.853940495294)*(1+2.*(dim-1.)/(o+1.))+(-0.208384391343)*dim
        ref[2,1,0,1]=(-0.453714314446)*(1+2.*(dim-1.)/(o+1.))+(0.347838565255)*dim
        ref[2,1,1,0]=(0.367826284405)*(1+2.*(dim-1.)/(o+1.))+(-0.320035489926)*dim
        ref[2,1,1,1]=(-1.10546369472)*(1+2.*(dim-1.)/(o+1.))+(0.558352731793)*dim
        ref[2,1,2,0]=(-0.214251265986)*(1+2.*(dim-1.)/(o+1.))+(0.157487302684)*dim
        ref[2,1,2,1]=(-0.046344088241)*(1+2.*(dim-1.)/(o+1.))+(0.0638344992189)*dim
        ref[2,2,0,0]=(1.51225561076)*(1+2.*(dim-1.)/(o+1.))+(0.0278718347964)*dim
        ref[2,2,0,1]=(-1.15525185863)*(1+2.*(dim-1.)/(o+1.))+(0.959188836395)*dim
        ref[2,2,1,0]=(-0.539405800194)*(1+2.*(dim-1.)/(o+1.))+(-1.04533852538)*dim
        ref[2,2,1,1]=(-0.96489299958)*(1+2.*(dim-1.)/(o+1.))+(1.81969637145)*dim
        ref[2,2,2,0]=(-1.26976024949)*(1+2.*(dim-1.)/(o+1.))+(-1.35104202488)*dim
        ref[2,2,2,1]=(0.615163395432)*(1+2.*(dim-1.)/(o+1.))+(0.715932976118)*dim
        ref[2,3,0,0]=(0.366522796287)*(1+2.*(dim-1.)/(o+1.))+(1.64580416515)*dim
        ref[2,3,0,1]=(0.828434385314)*(1+2.*(dim-1.)/(o+1.))+(-1.50510001039)*dim
        ref[2,3,1,0]=(-0.346764586054)*(1+2.*(dim-1.)/(o+1.))+(1.47143033351)*dim
        ref[2,3,1,1]=(1.22552827829)*(1+2.*(dim-1.)/(o+1.))+(1.09656572125)*dim
        ref[2,3,2,0]=(-0.288979874427)*(1+2.*(dim-1.)/(o+1.))+(0.994867178556)*dim
        ref[2,3,2,1]=(-0.294189045932)*(1+2.*(dim-1.)/(o+1.))+(0.88814466508)*dim
        ref[2,4,0,0]=(0.411388889544)*(1+2.*(dim-1.)/(o+1.))+(-1.58398949273)*dim
        ref[2,4,0,1]=(-0.518091928645)*(1+2.*(dim-1.)/(o+1.))+(0.836551121038)*dim
        ref[2,4,1,0]=(1.03905012916)*(1+2.*(dim-1.)/(o+1.))+(0.560050319818)*dim
        ref[2,4,1,1]=(0.599106343283)*(1+2.*(dim-1.)/(o+1.))+(-0.126478666584)*dim
        ref[2,4,2,0]=(-1.09620167833)*(1+2.*(dim-1.)/(o+1.))+(0.36028215212)*dim
        ref[2,4,2,1]=(-0.213293514797)*(1+2.*(dim-1.)/(o+1.))+(-0.184552785823)*dim
        ref[3,0,0,0]=(0.20672694781)*(1+2.*(dim-1.)/(o+1.))+(0.91601902837)*dim
        ref[3,0,0,1]=(-0.570917489061)*(1+2.*(dim-1.)/(o+1.))+(-0.154444824179)*dim
        ref[3,0,1,0]=(-1.85258337325)*(1+2.*(dim-1.)/(o+1.))+(0.466668914897)*dim
        ref[3,0,1,1]=(-0.110678363762)*(1+2.*(dim-1.)/(o+1.))+(0.540122415621)*dim
        ref[3,0,2,0]=(0.733897293024)*(1+2.*(dim-1.)/(o+1.))+(0.0890866058179)*dim
        ref[3,0,2,1]=(0.755207934346)*(1+2.*(dim-1.)/(o+1.))+(-0.758125384121)*dim
        ref[3,1,0,0]=(1.31914675441)*(1+2.*(dim-1.)/(o+1.))+(1.64194473194)*dim
        ref[3,1,0,1]=(-1.12362126101)*(1+2.*(dim-1.)/(o+1.))+(0.135765027019)*dim
        ref[3,1,1,0]=(-0.0328972781304)*(1+2.*(dim-1.)/(o+1.))+(-0.399367266818)*dim
        ref[3,1,1,1]=(-0.128839436684)*(1+2.*(dim-1.)/(o+1.))+(-0.876866069986)*dim
        ref[3,1,2,0]=(-0.8833223605)*(1+2.*(dim-1.)/(o+1.))+(-0.668886682264)*dim
        ref[3,1,2,1]=(-0.368758211275)*(1+2.*(dim-1.)/(o+1.))+(-0.202237567517)*dim
        ref[3,2,0,0]=(-0.952286913517)*(1+2.*(dim-1.)/(o+1.))+(-0.753930922605)*dim
        ref[3,2,0,1]=(-1.1899889553)*(1+2.*(dim-1.)/(o+1.))+(-0.030436066737)*dim
        ref[3,2,1,0]=(-0.148730428873)*(1+2.*(dim-1.)/(o+1.))+(0.338310066001)*dim
        ref[3,2,1,1]=(0.129715444697)*(1+2.*(dim-1.)/(o+1.))+(-1.05210699522)*dim
        ref[3,2,2,0]=(-0.574820919362)*(1+2.*(dim-1.)/(o+1.))+(-0.811986188004)*dim
        ref[3,2,2,1]=(0.146943065853)*(1+2.*(dim-1.)/(o+1.))+(0.523354268528)*dim
        ref[3,3,0,0]=(0.716510783419)*(1+2.*(dim-1.)/(o+1.))+(0.839808137761)*dim
        ref[3,3,0,1]=(-1.23846436881)*(1+2.*(dim-1.)/(o+1.))+(1.68591612066)*dim
        ref[3,3,1,0]=(0.217889109275)*(1+2.*(dim-1.)/(o+1.))+(0.34194285131)*dim
        ref[3,3,1,1]=(0.121267421155)*(1+2.*(dim-1.)/(o+1.))+(-0.0393083526063)*dim
        ref[3,3,2,0]=(-0.0577467713773)*(1+2.*(dim-1.)/(o+1.))+(-0.984106163624)*dim
        ref[3,3,2,1]=(-1.47340079046)*(1+2.*(dim-1.)/(o+1.))+(0.65920072125)*dim
        ref[3,4,0,0]=(-0.495218487647)*(1+2.*(dim-1.)/(o+1.))+(0.104606534694)*dim
        ref[3,4,0,1]=(0.00355321469463)*(1+2.*(dim-1.)/(o+1.))+(-1.26370118611)*dim
        ref[3,4,1,0]=(-0.629055246908)*(1+2.*(dim-1.)/(o+1.))+(0.117456822474)*dim
        ref[3,4,1,1]=(-0.649869143425)*(1+2.*(dim-1.)/(o+1.))+(0.617921422357)*dim
        ref[3,4,2,0]=(1.37942370181)*(1+2.*(dim-1.)/(o+1.))+(1.24315060014)*dim
        ref[3,4,2,1]=(-0.468393667668)*(1+2.*(dim-1.)/(o+1.))+(0.304025467684)*dim
      else:
        s[0,0,0,0]=(0.265353968453)*x[0]**o+(-0.988400050542)*x[0]+(0.915581496138)*x[1]**o+(-0.532717689852)*x[1]+(-0.326369232123)*x[2]**o+(-0.751181530498)*x[2]
        s[0,0,0,1]=(-0.700652623793)*x[0]**o+(0.385185933969)*x[0]+(0.213605768081)*x[1]**o+(-0.626286949699)*x[1]+(-0.502285730612)*x[2]**o+(-0.510666455487)*x[2]
        s[0,0,1,0]=(0.662622398823)*x[0]**o+(0.682570980418)*x[0]+(-0.888940566008)*x[1]**o+(0.917775368899)*x[1]+(0.397316182924)*x[2]**o+(0.470957649176)*x[2]
        s[0,0,1,1]=(0.506440820988)*x[0]**o+(0.548696583948)*x[0]+(-0.343755749097)*x[1]**o+(0.635493528274)*x[1]+(0.96169367604)*x[2]**o+(-0.814595321216)*x[2]
        s[0,0,2,0]=(0.0897565668909)*x[0]**o+(-0.707489384492)*x[0]+(0.852406382045)*x[1]**o+(0.264605905284)*x[1]+(-0.76680872637)*x[2]**o+(-0.607486929359)*x[2]
        s[0,0,2,1]=(-0.491708830409)*x[0]**o+(-0.917156620819)*x[0]+(0.138490412136)*x[1]**o+(0.445788359479)*x[1]+(0.660215675311)*x[2]**o+(-0.124958885809)*x[2]
        s[0,1,0,0]=(0.252227873704)*x[0]**o+(-0.268708653921)*x[0]+(-0.735320913237)*x[1]**o+(0.832659354382)*x[1]+(-0.503570989065)*x[2]**o+(-0.416695426775)*x[2]
        s[0,1,0,1]=(0.992266568089)*x[0]**o+(-0.553995410037)*x[0]+(-0.94305015981)*x[1]**o+(-0.277660590986)*x[1]+(-0.106619424885)*x[2]**o+(-0.641996454178)*x[2]
        s[0,1,1,0]=(0.346537921137)*x[0]**o+(-0.752921521456)*x[0]+(0.697296847348)*x[1]**o+(0.127827995978)*x[1]+(-0.024787761696)*x[2]**o+(-0.132279656336)*x[2]
        s[0,1,1,1]=(0.838216254144)*x[0]**o+(-0.0488544208848)*x[0]+(0.618848880659)*x[1]**o+(-0.962908419156)*x[1]+(-0.67378264273)*x[2]**o+(-0.206991847936)*x[2]
        s[0,1,2,0]=(0.41676085558)*x[0]**o+(-0.214807189212)*x[0]+(0.291507151171)*x[1]**o+(0.668437674897)*x[1]+(0.244072891194)*x[2]**o+(0.259834097546)*x[2]
        s[0,1,2,1]=(-0.181271253302)*x[0]**o+(-0.120474808334)*x[0]+(0.996200217415)*x[1]**o+(0.964447373929)*x[1]+(0.642033792194)*x[2]**o+(0.737708747193)*x[2]
        s[0,2,0,0]=(-0.792103621996)*x[0]**o+(-0.688217283375)*x[0]+(0.672498785205)*x[1]**o+(-0.653725451669)*x[1]+(0.184442483292)*x[2]**o+(-0.532250028625)*x[2]
        s[0,2,0,1]=(-0.535900645063)*x[0]**o+(0.910701730932)*x[0]+(-0.942492793063)*x[1]**o+(0.840681049537)*x[1]+(0.717504040832)*x[2]**o+(0.325490269912)*x[2]
        s[0,2,1,0]=(-0.337864688224)*x[0]**o+(0.746375970396)*x[0]+(0.747732580426)*x[1]**o+(-0.282770683276)*x[1]+(-0.790443829627)*x[2]**o+(0.719617282434)*x[2]
        s[0,2,1,1]=(0.288927362375)*x[0]**o+(-0.11015159228)*x[0]+(-0.840599795558)*x[1]**o+(0.400589417352)*x[1]+(0.651562852697)*x[2]**o+(0.472103872054)*x[2]
        s[0,2,2,0]=(0.759453740866)*x[0]**o+(0.764211886443)*x[0]+(-0.596686471329)*x[1]**o+(-0.305500139463)*x[1]+(0.644132204372)*x[2]**o+(0.409169518907)*x[2]
        s[0,2,2,1]=(0.310393680205)*x[0]**o+(-0.55096306264)*x[0]+(-0.0418723645688)*x[1]**o+(0.146737303429)*x[1]+(-0.221068515058)*x[2]**o+(0.347847681074)*x[2]
        s[0,3,0,0]=(-0.121018382943)*x[0]**o+(-0.828966827969)*x[0]+(-0.579574647427)*x[1]**o+(0.505540315546)*x[1]+(0.846008555599)*x[2]**o+(0.722616724538)*x[2]
        s[0,3,0,1]=(0.764966634503)*x[0]**o+(0.0406308216785)*x[0]+(0.627462444709)*x[1]**o+(-0.458074562731)*x[1]+(0.868715029031)*x[2]**o+(-0.326046950875)*x[2]
        s[0,3,1,0]=(0.88733222894)*x[0]**o+(0.199875300475)*x[0]+(0.929525862041)*x[1]**o+(0.280705926018)*x[1]+(-0.85604411697)*x[2]**o+(-0.509002674462)*x[2]
        s[0,3,1,1]=(0.178137426708)*x[0]**o+(0.0040475349026)*x[0]+(0.606739420029)*x[1]**o+(-0.399120462352)*x[1]+(0.688376909613)*x[2]**o+(-0.204172031563)*x[2]
        s[0,3,2,0]=(-0.55097942154)*x[0]**o+(0.518240585643)*x[0]+(0.809100524171)*x[1]**o+(0.309535824824)*x[1]+(0.824892877333)*x[2]**o+(-0.580350157914)*x[2]
        s[0,3,2,1]=(-0.429643456007)*x[0]**o+(0.683690075983)*x[0]+(-0.474187287962)*x[1]**o+(-0.641084362835)*x[1]+(-0.514804055536)*x[2]**o+(0.653494101426)*x[2]
        s[0,4,0,0]=(0.759632202933)*x[0]**o+(-0.953554006806)*x[0]+(-0.564086866677)*x[1]**o+(-0.78087231593)*x[1]+(-0.589063170649)*x[2]**o+(-0.569522356624)*x[2]
        s[0,4,0,1]=(-0.440639409356)*x[0]**o+(-0.887245582242)*x[0]+(-0.731777330438)*x[1]**o+(0.678481115368)*x[1]+(-0.234694289692)*x[2]**o+(-0.481669499353)*x[2]
        s[0,4,1,0]=(0.421338590919)*x[0]**o+(-0.710549762367)*x[0]+(-0.831562216016)*x[1]**o+(0.526145415136)*x[1]+(-0.228391513851)*x[2]**o+(-0.287603858419)*x[2]
        s[0,4,1,1]=(-0.394922177529)*x[0]**o+(-0.749933015384)*x[0]+(0.501851406608)*x[1]**o+(0.414942993798)*x[1]+(0.882492276774)*x[2]**o+(-0.417451715702)*x[2]
        s[0,4,2,0]=(-0.708380417114)*x[0]**o+(0.272117639041)*x[0]+(-0.914251681965)*x[1]**o+(-0.692698068557)*x[1]+(0.740875968835)*x[2]**o+(0.0816142773216)*x[2]
        s[0,4,2,1]=(-0.781489599785)*x[0]**o+(-0.688587387866)*x[0]+(-0.8249818093)*x[1]**o+(0.479876441134)*x[1]+(-0.726425081108)*x[2]**o+(0.467932761023)*x[2]
        s[1,0,0,0]=(0.342121912006)*x[0]**o+(0.136682990021)*x[0]+(0.181997624536)*x[1]**o+(0.168026470492)*x[1]+(0.0115668517645)*x[2]**o+(-0.0377932582246)*x[2]
        s[1,0,0,1]=(-0.377080344591)*x[0]**o+(0.874997839577)*x[0]+(0.998952980553)*x[1]**o+(-0.511330536046)*x[1]+(0.845159267663)*x[2]**o+(0.0349344413329)*x[2]
        s[1,0,1,0]=(0.376824814933)*x[0]**o+(-0.58697348357)*x[0]+(0.764173288619)*x[1]**o+(0.288509758241)*x[1]+(-0.818987335991)*x[2]**o+(0.433785438733)*x[2]
        s[1,0,1,1]=(-0.923417677759)*x[0]**o+(0.703237676923)*x[0]+(-0.0963613139466)*x[1]**o+(0.648863487662)*x[1]+(0.121038690293)*x[2]**o+(-0.428707341383)*x[2]
        s[1,0,2,0]=(-0.192698196902)*x[0]**o+(-0.45684016542)*x[0]+(-0.157529151044)*x[1]**o+(0.665508784857)*x[1]+(-0.0263995814254)*x[2]**o+(-0.711087191185)*x[2]
        s[1,0,2,1]=(-0.262702009507)*x[0]**o+(0.612242386787)*x[0]+(-0.0527177865241)*x[1]**o+(-0.946597487364)*x[1]+(0.683034847912)*x[2]**o+(0.447810652774)*x[2]
        s[1,1,0,0]=(0.0954675694985)*x[0]**o+(-0.692307292788)*x[0]+(0.682808081853)*x[1]**o+(-0.602155379721)*x[1]+(-0.804568326115)*x[2]**o+(0.6383976993)*x[2]
        s[1,1,0,1]=(-0.776359536583)*x[0]**o+(0.593053796057)*x[0]+(-0.560010245091)*x[1]**o+(-0.0167025681078)*x[1]+(0.189236205132)*x[2]**o+(0.625263059129)*x[2]
        s[1,1,1,0]=(-0.294839786045)*x[0]**o+(0.318602896219)*x[0]+(0.790859996391)*x[1]**o+(-0.100239676238)*x[1]+(-0.752119733912)*x[2]**o+(0.909827281572)*x[2]
        s[1,1,1,1]=(0.427920987151)*x[0]**o+(0.0579828435662)*x[0]+(-0.805648184456)*x[1]**o+(0.545638930222)*x[1]+(-0.333821427396)*x[2]**o+(0.489029035079)*x[2]
        s[1,1,2,0]=(-0.102211826161)*x[0]**o+(-0.23589775753)*x[0]+(0.194344132247)*x[1]**o+(0.216623239742)*x[1]+(-0.328848234996)*x[2]**o+(-0.35246467614)*x[2]
        s[1,1,2,1]=(-0.597326323585)*x[0]**o+(-0.727893394493)*x[0]+(-0.219786140301)*x[1]**o+(-0.261271587677)*x[1]+(0.433015479844)*x[2]**o+(0.480572847641)*x[2]
        s[1,2,0,0]=(0.978638976517)*x[0]**o+(0.252195213856)*x[0]+(-0.534544050017)*x[1]**o+(-0.0719497516536)*x[1]+(0.0756026756135)*x[2]**o+(-0.376095264735)*x[2]
        s[1,2,0,1]=(-0.592591339778)*x[0]**o+(-0.87353085535)*x[0]+(0.226804211947)*x[1]**o+(0.257318409091)*x[1]+(0.820507252476)*x[2]**o+(0.490297079988)*x[2]
        s[1,2,1,0]=(0.157573169077)*x[0]**o+(0.128029582112)*x[0]+(-0.954106083921)*x[1]**o+(0.509373015197)*x[1]+(0.788827839704)*x[2]**o+(0.717054840575)*x[2]
        s[1,2,1,1]=(-0.47425212595)*x[0]**o+(-0.608889999781)*x[0]+(-0.334194651773)*x[1]**o+(-0.439282037399)*x[1]+(-0.169103982018)*x[2]**o+(-0.569343903254)*x[2]
        s[1,2,2,0]=(0.686390790562)*x[0]**o+(0.670565132214)*x[0]+(-0.491526583908)*x[1]**o+(-0.887969445154)*x[1]+(0.666707124541)*x[2]**o+(-0.643899609236)*x[2]
        s[1,2,2,1]=(0.158400774071)*x[0]**o+(0.672487819329)*x[0]+(-0.397304779942)*x[1]**o+(0.454998380713)*x[1]+(-0.571811342108)*x[2]**o+(0.294069278512)*x[2]
        s[1,3,0,0]=(-0.652264634829)*x[0]**o+(-0.676541910532)*x[0]+(-0.455722985275)*x[1]**o+(-0.25957175602)*x[1]+(0.619181596074)*x[2]**o+(0.305382546067)*x[2]
        s[1,3,0,1]=(0.958823451851)*x[0]**o+(-0.978084662344)*x[0]+(-0.276737082098)*x[1]**o+(-0.117588783747)*x[1]+(0.61233901794)*x[2]**o+(-0.430645122164)*x[2]
        s[1,3,1,0]=(0.17639490062)*x[0]**o+(-0.072015890916)*x[0]+(-0.0776296985656)*x[1]**o+(0.92506693456)*x[1]+(-0.228890114078)*x[2]**o+(-0.817012691208)*x[2]
        s[1,3,1,1]=(-0.0124492438715)*x[0]**o+(-0.291955568984)*x[0]+(-0.701102657404)*x[1]**o+(0.149273972357)*x[1]+(-0.521101405472)*x[2]**o+(0.183603155794)*x[2]
        s[1,3,2,0]=(0.12397369853)*x[0]**o+(0.436139768082)*x[0]+(0.523613099442)*x[1]**o+(0.0273767825371)*x[1]+(0.4825934187)*x[2]**o+(-0.261047164188)*x[2]
        s[1,3,2,1]=(-0.253163229158)*x[0]**o+(-0.190716207327)*x[0]+(0.104639827693)*x[1]**o+(-0.868381858703)*x[1]+(-0.439763792231)*x[2]**o+(-0.799113479705)*x[2]
        s[1,4,0,0]=(-0.304838319461)*x[0]**o+(0.701590180772)*x[0]+(-0.110558497315)*x[1]**o+(-0.893822894679)*x[1]+(-0.787583691318)*x[2]**o+(-0.89265767325)*x[2]
        s[1,4,0,1]=(0.184781667393)*x[0]**o+(0.00115077722427)*x[0]+(-0.301829747192)*x[1]**o+(0.492639963911)*x[1]+(-0.828978825155)*x[2]**o+(0.528275593254)*x[2]
        s[1,4,1,0]=(0.42616420927)*x[0]**o+(-0.473452048931)*x[0]+(0.756413372908)*x[1]**o+(-0.875430941202)*x[1]+(-0.440394982498)*x[2]**o+(0.387473346762)*x[2]
        s[1,4,1,1]=(-0.974545673815)*x[0]**o+(-0.583230191948)*x[0]+(-0.8604375003)*x[1]**o+(0.792666394984)*x[1]+(0.712418702968)*x[2]**o+(-0.635993401114)*x[2]
        s[1,4,2,0]=(0.898921012186)*x[0]**o+(0.31154532209)*x[0]+(0.100106723409)*x[1]**o+(0.510220124566)*x[1]+(-0.694976306736)*x[2]**o+(0.744462320672)*x[2]
        s[1,4,2,1]=(0.208678468389)*x[0]**o+(-0.029704981412)*x[0]+(-0.709792226721)*x[1]**o+(0.978207622687)*x[1]+(-0.94268835481)*x[2]**o+(0.983998190279)*x[2]
        s[2,0,0,0]=(-0.117454131321)*x[0]**o+(0.325178166996)*x[0]+(-0.955591553442)*x[1]**o+(-0.893511287203)*x[1]+(-0.992448477236)*x[2]**o+(-0.314767960841)*x[2]
        s[2,0,0,1]=(-0.215442066237)*x[0]**o+(0.199095086203)*x[0]+(-0.948101787926)*x[1]**o+(0.833921396167)*x[1]+(-0.945489620081)*x[2]**o+(-0.836168838947)*x[2]
        s[2,0,1,0]=(-0.0147045046721)*x[0]**o+(0.910556812541)*x[0]+(-0.854656661853)*x[1]**o+(0.603960686575)*x[1]+(-0.110464227264)*x[2]**o+(-0.654514120746)*x[2]
        s[2,0,1,1]=(0.12588483106)*x[0]**o+(0.741103813876)*x[0]+(0.273343918839)*x[1]**o+(-0.0496068521746)*x[1]+(0.851852637841)*x[2]**o+(0.513176454441)*x[2]
        s[2,0,2,0]=(-0.949030805361)*x[0]**o+(-0.186923222351)*x[0]+(-0.469739120295)*x[1]**o+(-0.667688306186)*x[1]+(0.561259377808)*x[2]**o+(-0.455307474368)*x[2]
        s[2,0,2,1]=(0.50289454022)*x[0]**o+(0.601455959785)*x[0]+(-0.359319458876)*x[1]**o+(-0.252866723971)*x[1]+(-0.232113717554)*x[2]**o+(-0.923251891682)*x[2]
        s[2,1,0,0]=(0.087617991436)*x[0]**o+(0.825668089054)*x[0]+(-0.246279274065)*x[1]**o+(0.427961138603)*x[1]+(0.134574297189)*x[2]**o+(0.126692670762)*x[2]
        s[2,1,0,1]=(0.618759834233)*x[0]**o+(-0.986840591193)*x[0]+(-0.356764253448)*x[1]**o+(0.861294268808)*x[1]+(0.329156660561)*x[2]**o+(0.454768559063)*x[2]
        s[2,1,1,0]=(0.0592587565219)*x[0]**o+(0.658218480064)*x[0]+(-0.605532183319)*x[1]**o+(0.628709371911)*x[1]+(-0.755778976293)*x[2]**o+(0.200067133647)*x[2]
        s[2,1,1,1]=(0.60715871142)*x[0]**o+(0.119152257078)*x[0]+(-0.456685613073)*x[1]**o+(-0.431830986012)*x[1]+(-0.239978100497)*x[2]**o+(-0.272134038408)*x[2]
        s[2,1,2,0]=(0.313930291609)*x[0]**o+(-0.0335423445806)*x[0]+(0.329631545985)*x[1]**o+(0.578422305734)*x[1]+(0.428275935457)*x[2]**o+(0.0754231472821)*x[2]
        s[2,1,2,1]=(0.95215367321)*x[0]**o+(-0.395110969421)*x[0]+(0.186889413454)*x[1]**o+(-0.211320210609)*x[1]+(-0.950278722072)*x[2]**o+(-0.111767520683)*x[2]
        s[2,2,0,0]=(-0.928704839938)*x[0]**o+(0.856077241772)*x[0]+(-0.0417301431511)*x[1]**o+(-0.76181987359)*x[1]+(-0.992155226136)*x[2]**o+(-0.776267895931)*x[2]
        s[2,2,0,1]=(-0.969248181097)*x[0]**o+(0.245775588093)*x[0]+(-0.505906920175)*x[1]**o+(0.146641331232)*x[1]+(-0.262102422862)*x[2]**o+(-0.561207377079)*x[2]
        s[2,2,1,0]=(-0.718472022429)*x[0]**o+(0.439297055455)*x[0]+(-0.852005557298)*x[1]**o+(0.504211693582)*x[1]+(-0.126256957345)*x[2]**o+(0.20236075339)*x[2]
        s[2,2,1,1]=(-0.547818793062)*x[0]**o+(-0.550864012953)*x[0]+(-0.87720186351)*x[1]**o+(0.962779562428)*x[1]+(-0.663892241535)*x[2]**o+(0.826235643802)*x[2]
        s[2,2,2,0]=(0.619652835608)*x[0]**o+(-0.418942285364)*x[0]+(0.407390229096)*x[1]**o+(0.999750598891)*x[1]+(0.826445107816)*x[2]**o+(-0.827613183804)*x[2]
        s[2,2,2,1]=(0.742447797718)*x[0]**o+(-0.135700593818)*x[0]+(0.691148074885)*x[1]**o+(-0.457826189496)*x[1]+(0.256017777404)*x[2]**o+(-0.165702500648)*x[2]
        s[2,3,0,0]=(-0.43784392872)*x[0]**o+(0.16925005364)*x[0]+(-0.659181589908)*x[1]**o+(0.675362824639)*x[1]+(-0.577126041428)*x[2]**o+(-0.506716255511)*x[2]
        s[2,3,0,1]=(-0.216143074446)*x[0]**o+(0.437507580615)*x[0]+(0.401331581783)*x[1]**o+(0.689412723355)*x[1]+(0.8762344674)*x[2]**o+(-0.582543733655)*x[2]
        s[2,3,1,0]=(0.0788197892383)*x[0]**o+(0.417141756193)*x[0]+(-0.253190931158)*x[1]**o+(0.409575699654)*x[1]+(0.514494619825)*x[2]**o+(0.399143793546)*x[2]
        s[2,3,1,1]=(-0.719766785708)*x[0]**o+(-0.620867182386)*x[0]+(-0.0490130572844)*x[1]**o+(0.5158062442)*x[1]+(0.871371891386)*x[2]**o+(0.568994700919)*x[2]
        s[2,3,2,0]=(0.076389643296)*x[0]**o+(0.508512060622)*x[0]+(-0.78737605411)*x[1]**o+(-0.00342872062952)*x[1]+(-0.659245059343)*x[2]**o+(-0.265456383349)*x[2]
        s[2,3,2,1]=(0.5875893363)*x[0]**o+(-0.718091321577)*x[0]+(-0.801865283024)*x[1]**o+(-0.908030358815)*x[1]+(0.431675901356)*x[2]**o+(-0.819466399747)*x[2]
        s[2,4,0,0]=(0.67994234014)*x[0]**o+(0.0421294866201)*x[0]+(0.914156657838)*x[1]**o+(0.631988857673)*x[1]+(-0.731815343643)*x[2]**o+(-0.226439431169)*x[2]
        s[2,4,0,1]=(0.382719860981)*x[0]**o+(0.0998281702684)*x[0]+(-0.191692152755)*x[1]**o+(-0.632487629935)*x[1]+(-0.47197856624)*x[2]**o+(0.743041964659)*x[2]
        s[2,4,1,0]=(-0.766894309495)*x[0]**o+(0.66365148919)*x[0]+(-0.985125550604)*x[1]**o+(0.977558230989)*x[1]+(-0.786405168576)*x[2]**o+(-0.586888601056)*x[2]
        s[2,4,1,1]=(-0.408491721685)*x[0]**o+(-0.485042299377)*x[0]+(0.0342771405689)*x[1]**o+(-0.0281888804188)*x[1]+(-0.100567975294)*x[2]**o+(0.16781221998)*x[2]
        s[2,4,2,0]=(0.655177009817)*x[0]**o+(0.599751578479)*x[0]+(-0.69519370742)*x[1]**o+(0.765111933279)*x[1]+(-0.600152872818)*x[2]**o+(0.933762224073)*x[2]
        s[2,4,2,1]=(-0.783402080781)*x[0]**o+(0.497892410534)*x[0]+(-0.896301612142)*x[1]**o+(-0.624692866087)*x[1]+(0.384092901066)*x[2]**o+(0.103876065891)*x[2]
        s[3,0,0,0]=(0.727781520125)*x[0]**o+(0.516411031951)*x[0]+(0.520409588798)*x[1]**o+(0.203314941059)*x[1]+(-0.72359147418)*x[2]**o+(0.766852866118)*x[2]
        s[3,0,0,1]=(-0.0843842205054)*x[0]**o+(-0.270711183223)*x[0]+(-0.235601047175)*x[1]**o+(-0.820277497036)*x[1]+(-0.466133585852)*x[2]**o+(0.155997468178)*x[2]
        s[3,0,1,0]=(-0.951270599262)*x[0]**o+(-0.313463217212)*x[0]+(-0.166446810677)*x[1]**o+(0.0311582608036)*x[1]+(-0.143475616634)*x[2]**o+(-0.174089420721)*x[2]
        s[3,0,1,1]=(0.31005632514)*x[0]**o+(-0.473374219519)*x[0]+(0.922718547121)*x[1]**o+(0.354849506312)*x[1]+(0.466885594274)*x[2]**o+(-0.0545966611475)*x[2]
        s[3,0,2,0]=(0.0930631890409)*x[0]**o+(0.238199576739)*x[0]+(0.785659372495)*x[1]**o+(-0.381418540402)*x[1]+(-0.372849688379)*x[2]**o+(-0.550096690451)*x[2]
        s[3,0,2,1]=(-0.0488046933792)*x[0]**o+(-0.85673228641)*x[0]+(-0.888414982245)*x[1]**o+(-0.733480030742)*x[1]+(-0.138791373365)*x[2]**o+(-0.0428634394828)*x[2]
        s[3,1,0,0]=(0.431582682868)*x[0]**o+(0.997952021577)*x[0]+(-0.586523128595)*x[1]**o+(-0.38366122568)*x[1]+(-0.966801100136)*x[2]**o+(-0.814601474666)*x[2]
        s[3,1,0,1]=(0.51469888558)*x[0]**o+(0.852630560538)*x[0]+(-0.431644582882)*x[1]**o+(0.612002981263)*x[1]+(-0.824898907932)*x[2]**o+(0.740125311337)*x[2]
        s[3,1,1,0]=(-0.998816511903)*x[0]**o+(0.501436620365)*x[0]+(-0.876369817412)*x[1]**o+(-0.751901171932)*x[1]+(-0.127524412151)*x[2]**o+(0.556578788174)*x[2]
        s[3,1,1,1]=(0.494572326989)*x[0]**o+(-0.4109962303)*x[0]+(0.332604369701)*x[1]**o+(0.144741772456)*x[1]+(-0.356738712921)*x[2]**o+(-0.930461094693)*x[2]
        s[3,1,2,0]=(0.298299420769)*x[0]**o+(0.465301670992)*x[0]+(-0.625366287818)*x[1]**o+(0.617300960428)*x[1]+(-0.0218635106247)*x[2]**o+(-0.400465814843)*x[2]
        s[3,1,2,1]=(-0.269730676738)*x[0]**o+(0.385672666629)*x[0]+(0.0768574834461)*x[1]**o+(-0.946466342247)*x[1]+(-0.46039524704)*x[2]**o+(-0.092664263096)*x[2]
        s[3,2,0,0]=(0.461319900449)*x[0]**o+(0.894479535569)*x[0]+(-0.331082415608)*x[1]**o+(0.40219970326)*x[1]+(-0.45348751543)*x[2]**o+(0.344838434995)*x[2]
        s[3,2,0,1]=(0.555684742117)*x[0]**o+(0.140415098553)*x[0]+(-0.53477167757)*x[1]**o+(-0.365761904777)*x[1]+(0.96328384437)*x[2]**o+(0.619810575276)*x[2]
        s[3,2,1,0]=(-0.14916965531)*x[0]**o+(0.889806300338)*x[0]+(-0.456091691059)*x[1]**o+(0.532621704874)*x[1]+(0.329988274366)*x[2]**o+(0.800234733169)*x[2]
        s[3,2,1,1]=(0.380606552573)*x[0]**o+(-0.541127952834)*x[0]+(-0.851396276663)*x[1]**o+(0.166128704099)*x[1]+(-0.575246273865)*x[2]**o+(-0.438073609497)*x[2]
        s[3,2,2,0]=(0.845014379056)*x[0]**o+(-0.812352254049)*x[0]+(0.0324309664699)*x[1]**o+(0.558811307761)*x[1]+(-0.865143597202)*x[2]**o+(-0.755672880303)*x[2]
        s[3,2,2,1]=(0.5343293465)*x[0]**o+(-0.840117005289)*x[0]+(-0.30550028433)*x[1]**o+(0.951726967137)*x[1]+(0.469519945822)*x[2]**o+(0.748837114445)*x[2]
        s[3,3,0,0]=(-0.885008843023)*x[0]**o+(-0.170701893945)*x[0]+(-0.374063979126)*x[1]**o+(-0.30389599708)*x[1]+(-0.661322095381)*x[2]**o+(0.838294065739)*x[2]
        s[3,3,0,1]=(0.854428192443)*x[0]**o+(0.2166993324)*x[0]+(0.43944583933)*x[1]**o+(-0.858798284298)*x[1]+(-0.84436470857)*x[2]**o+(-0.567545786739)*x[2]
        s[3,3,1,0]=(0.66702962158)*x[0]**o+(0.720100651321)*x[0]+(0.0956882490639)*x[1]**o+(0.0372232762552)*x[1]+(-0.67009655948)*x[2]**o+(-0.569854055098)*x[2]
        s[3,3,1,1]=(0.146466946465)*x[0]**o+(0.164924499496)*x[0]+(-0.2972769793)*x[1]**o+(0.865363847078)*x[1]+(-0.180431004392)*x[2]**o+(0.265225273207)*x[2]
        s[3,3,2,0]=(-0.830402000378)*x[0]**o+(-0.00161430224161)*x[0]+(-0.863095020073)*x[1]**o+(0.922102273263)*x[1]+(0.385860222623)*x[2]**o+(0.981756782753)*x[2]
        s[3,3,2,1]=(0.567277892172)*x[0]**o+(0.78766223915)*x[0]+(0.412870235103)*x[1]**o+(0.46588554159)*x[1]+(0.716230666858)*x[2]**o+(0.425979074273)*x[2]
        s[3,4,0,0]=(0.679938424982)*x[0]**o+(0.508093418021)*x[0]+(-0.73945776378)*x[1]**o+(-0.178938904098)*x[1]+(0.996547500259)*x[2]**o+(-0.689147793037)*x[2]
        s[3,4,0,1]=(-0.181577713519)*x[0]**o+(-0.413066527133)*x[0]+(0.83941283649)*x[1]**o+(-0.664951450375)*x[1]+(0.993705498629)*x[2]**o+(-0.483417137243)*x[2]
        s[3,4,1,0]=(0.328862164768)*x[0]**o+(0.869116819138)*x[0]+(-0.8079517423)*x[1]**o+(-0.692894104885)*x[1]+(-0.407582178576)*x[2]**o+(0.437242243866)*x[2]
        s[3,4,1,1]=(0.512281196984)*x[0]**o+(-0.155819537445)*x[0]+(0.693970927236)*x[1]**o+(0.785297186585)*x[1]+(-0.514686313027)*x[2]**o+(-0.474084873678)*x[2]
        s[3,4,2,0]=(0.225627329646)*x[0]**o+(0.247722481758)*x[0]+(0.88743622494)*x[1]**o+(-0.953008640988)*x[1]+(-0.428243081608)*x[2]**o+(0.88759395083)*x[2]
        s[3,4,2,1]=(-0.942620080756)*x[0]**o+(-0.794499826777)*x[0]+(-0.191327255697)*x[1]**o+(-0.185051303598)*x[1]+(0.681543419098)*x[2]**o+(-0.142256046241)*x[2]
        ref[0,0,0,0]=(0.854566232468)*(1+2.*(dim-1.)/(o+1.))+(-2.27229927089)*dim
        ref[0,0,0,1]=(-0.989332586323)*(1+2.*(dim-1.)/(o+1.))+(-0.751767471217)*dim
        ref[0,0,1,0]=(0.170998015739)*(1+2.*(dim-1.)/(o+1.))+(2.07130399849)*dim
        ref[0,0,1,1]=(1.12437874793)*(1+2.*(dim-1.)/(o+1.))+(0.369594791005)*dim
        ref[0,0,2,0]=(0.175354222567)*(1+2.*(dim-1.)/(o+1.))+(-1.05037040857)*dim
        ref[0,0,2,1]=(0.306997257038)*(1+2.*(dim-1.)/(o+1.))+(-0.596327147149)*dim
        ref[0,1,0,0]=(-0.986664028597)*(1+2.*(dim-1.)/(o+1.))+(0.147255273686)*dim
        ref[0,1,0,1]=(-0.057403016606)*(1+2.*(dim-1.)/(o+1.))+(-1.4736524552)*dim
        ref[0,1,1,0]=(1.01904700679)*(1+2.*(dim-1.)/(o+1.))+(-0.757373181813)*dim
        ref[0,1,1,1]=(0.783282492073)*(1+2.*(dim-1.)/(o+1.))+(-1.21875468798)*dim
        ref[0,1,2,0]=(0.952340897946)*(1+2.*(dim-1.)/(o+1.))+(0.713464583232)*dim
        ref[0,1,2,1]=(1.45696275631)*(1+2.*(dim-1.)/(o+1.))+(1.58168131279)*dim
        ref[0,2,0,0]=(0.064837646501)*(1+2.*(dim-1.)/(o+1.))+(-1.87419276367)*dim
        ref[0,2,0,1]=(-0.760889397294)*(1+2.*(dim-1.)/(o+1.))+(2.07687305038)*dim
        ref[0,2,1,0]=(-0.380575937425)*(1+2.*(dim-1.)/(o+1.))+(1.18322256955)*dim
        ref[0,2,1,1]=(0.0998904195137)*(1+2.*(dim-1.)/(o+1.))+(0.762541697126)*dim
        ref[0,2,2,0]=(0.806899473909)*(1+2.*(dim-1.)/(o+1.))+(0.867881265887)*dim
        ref[0,2,2,1]=(0.0474528005784)*(1+2.*(dim-1.)/(o+1.))+(-0.0563780781365)*dim
        ref[0,3,0,0]=(0.145415525229)*(1+2.*(dim-1.)/(o+1.))+(0.399190212115)*dim
        ref[0,3,0,1]=(2.26114410824)*(1+2.*(dim-1.)/(o+1.))+(-0.743490691928)*dim
        ref[0,3,1,0]=(0.960813974011)*(1+2.*(dim-1.)/(o+1.))+(-0.0284214479695)*dim
        ref[0,3,1,1]=(1.47325375635)*(1+2.*(dim-1.)/(o+1.))+(-0.599244959013)*dim
        ref[0,3,2,0]=(1.08301397996)*(1+2.*(dim-1.)/(o+1.))+(0.247426252553)*dim
        ref[0,3,2,1]=(-1.4186347995)*(1+2.*(dim-1.)/(o+1.))+(0.696099814573)*dim
        ref[0,4,0,0]=(-0.393517834393)*(1+2.*(dim-1.)/(o+1.))+(-2.30394867936)*dim
        ref[0,4,0,1]=(-1.40711102949)*(1+2.*(dim-1.)/(o+1.))+(-0.690433966228)*dim
        ref[0,4,1,0]=(-0.638615138949)*(1+2.*(dim-1.)/(o+1.))+(-0.47200820565)*dim
        ref[0,4,1,1]=(0.989421505854)*(1+2.*(dim-1.)/(o+1.))+(-0.752441737288)*dim
        ref[0,4,2,0]=(-0.881756130244)*(1+2.*(dim-1.)/(o+1.))+(-0.338966152194)*dim
        ref[0,4,2,1]=(-2.33289649019)*(1+2.*(dim-1.)/(o+1.))+(0.259221814291)*dim
        ref[1,0,0,0]=(0.535686388307)*(1+2.*(dim-1.)/(o+1.))+(0.266916202289)*dim
        ref[1,0,0,1]=(1.46703190363)*(1+2.*(dim-1.)/(o+1.))+(0.398601744865)*dim
        ref[1,0,1,0]=(0.32201076756)*(1+2.*(dim-1.)/(o+1.))+(0.135321713403)*dim
        ref[1,0,1,1]=(-0.898740301412)*(1+2.*(dim-1.)/(o+1.))+(0.923393823202)*dim
        ref[1,0,2,0]=(-0.376626929371)*(1+2.*(dim-1.)/(o+1.))+(-0.502418571748)*dim
        ref[1,0,2,1]=(0.367615051881)*(1+2.*(dim-1.)/(o+1.))+(0.113455552197)*dim
        ref[1,1,0,0]=(-0.0262926747634)*(1+2.*(dim-1.)/(o+1.))+(-0.656064973208)*dim
        ref[1,1,0,1]=(-1.14713357654)*(1+2.*(dim-1.)/(o+1.))+(1.20161428708)*dim
        ref[1,1,1,0]=(-0.256099523566)*(1+2.*(dim-1.)/(o+1.))+(1.12819050155)*dim
        ref[1,1,1,1]=(-0.711548624701)*(1+2.*(dim-1.)/(o+1.))+(1.09265080887)*dim
        ref[1,1,2,0]=(-0.23671592891)*(1+2.*(dim-1.)/(o+1.))+(-0.371739193929)*dim
        ref[1,1,2,1]=(-0.384096984041)*(1+2.*(dim-1.)/(o+1.))+(-0.508592134529)*dim
        ref[1,2,0,0]=(0.519697602114)*(1+2.*(dim-1.)/(o+1.))+(-0.195849802533)*dim
        ref[1,2,0,1]=(0.454720124645)*(1+2.*(dim-1.)/(o+1.))+(-0.125915366271)*dim
        ref[1,2,1,0]=(-0.00770507513982)*(1+2.*(dim-1.)/(o+1.))+(1.35445743788)*dim
        ref[1,2,1,1]=(-0.977550759742)*(1+2.*(dim-1.)/(o+1.))+(-1.61751594043)*dim
        ref[1,2,2,0]=(0.861571331194)*(1+2.*(dim-1.)/(o+1.))+(-0.861303922176)*dim
        ref[1,2,2,1]=(-0.810715347979)*(1+2.*(dim-1.)/(o+1.))+(1.42155547855)*dim
        ref[1,3,0,0]=(-0.48880602403)*(1+2.*(dim-1.)/(o+1.))+(-0.630731120484)*dim
        ref[1,3,0,1]=(1.29442538769)*(1+2.*(dim-1.)/(o+1.))+(-1.52631856825)*dim
        ref[1,3,1,0]=(-0.130124912024)*(1+2.*(dim-1.)/(o+1.))+(0.0360383524367)*dim
        ref[1,3,1,1]=(-1.23465330675)*(1+2.*(dim-1.)/(o+1.))+(0.0409215591672)*dim
        ref[1,3,2,0]=(1.13018021667)*(1+2.*(dim-1.)/(o+1.))+(0.202469386431)*dim
        ref[1,3,2,1]=(-0.588287193696)*(1+2.*(dim-1.)/(o+1.))+(-1.85821154573)*dim
        ref[1,4,0,0]=(-1.20298050809)*(1+2.*(dim-1.)/(o+1.))+(-1.08489038716)*dim
        ref[1,4,0,1]=(-0.946026904954)*(1+2.*(dim-1.)/(o+1.))+(1.02206633439)*dim
        ref[1,4,1,0]=(0.74218259968)*(1+2.*(dim-1.)/(o+1.))+(-0.961409643371)*dim
        ref[1,4,1,1]=(-1.12256447115)*(1+2.*(dim-1.)/(o+1.))+(-0.426557198078)*dim
        ref[1,4,2,0]=(0.304051428859)*(1+2.*(dim-1.)/(o+1.))+(1.56622776733)*dim
        ref[1,4,2,1]=(-1.44380211314)*(1+2.*(dim-1.)/(o+1.))+(1.93250083155)*dim
        ref[2,0,0,0]=(-2.065494162)*(1+2.*(dim-1.)/(o+1.))+(-0.883101081048)*dim
        ref[2,0,0,1]=(-2.10903347424)*(1+2.*(dim-1.)/(o+1.))+(0.196847643424)*dim
        ref[2,0,1,0]=(-0.979825393789)*(1+2.*(dim-1.)/(o+1.))+(0.86000337837)*dim
        ref[2,0,1,1]=(1.25108138774)*(1+2.*(dim-1.)/(o+1.))+(1.20467341614)*dim
        ref[2,0,2,0]=(-0.857510547847)*(1+2.*(dim-1.)/(o+1.))+(-1.30991900291)*dim
        ref[2,0,2,1]=(-0.0885386362105)*(1+2.*(dim-1.)/(o+1.))+(-0.574662655868)*dim
        ref[2,1,0,0]=(-0.0240869854401)*(1+2.*(dim-1.)/(o+1.))+(1.38032189842)*dim
        ref[2,1,0,1]=(0.591152241347)*(1+2.*(dim-1.)/(o+1.))+(0.329222236678)*dim
        ref[2,1,1,0]=(-1.30205240309)*(1+2.*(dim-1.)/(o+1.))+(1.48699498562)*dim
        ref[2,1,1,1]=(-0.0895050021498)*(1+2.*(dim-1.)/(o+1.))+(-0.584812767342)*dim
        ref[2,1,2,0]=(1.07183777305)*(1+2.*(dim-1.)/(o+1.))+(0.620303108435)*dim
        ref[2,1,2,1]=(0.188764364592)*(1+2.*(dim-1.)/(o+1.))+(-0.718198700712)*dim
        ref[2,2,0,0]=(-1.96259020923)*(1+2.*(dim-1.)/(o+1.))+(-0.682010527749)*dim
        ref[2,2,0,1]=(-1.73725752413)*(1+2.*(dim-1.)/(o+1.))+(-0.168790457754)*dim
        ref[2,2,1,0]=(-1.69673453707)*(1+2.*(dim-1.)/(o+1.))+(1.14586950243)*dim
        ref[2,2,1,1]=(-2.08891289811)*(1+2.*(dim-1.)/(o+1.))+(1.23815119328)*dim
        ref[2,2,2,0]=(1.85348817252)*(1+2.*(dim-1.)/(o+1.))+(-0.246804870276)*dim
        ref[2,2,2,1]=(1.68961365001)*(1+2.*(dim-1.)/(o+1.))+(-0.759229283962)*dim
        ref[2,3,0,0]=(-1.67415156006)*(1+2.*(dim-1.)/(o+1.))+(0.337896622768)*dim
        ref[2,3,0,1]=(1.06142297474)*(1+2.*(dim-1.)/(o+1.))+(0.544376570315)*dim
        ref[2,3,1,0]=(0.340123477905)*(1+2.*(dim-1.)/(o+1.))+(1.22586124939)*dim
        ref[2,3,1,1]=(0.102592048393)*(1+2.*(dim-1.)/(o+1.))+(0.463933762733)*dim
        ref[2,3,2,0]=(-1.37023147016)*(1+2.*(dim-1.)/(o+1.))+(0.239626956644)*dim
        ref[2,3,2,1]=(0.217399954632)*(1+2.*(dim-1.)/(o+1.))+(-2.44558808014)*dim
        ref[2,4,0,0]=(0.862283654334)*(1+2.*(dim-1.)/(o+1.))+(0.447678913125)*dim
        ref[2,4,0,1]=(-0.280950858013)*(1+2.*(dim-1.)/(o+1.))+(0.210382504993)*dim
        ref[2,4,1,0]=(-2.53842502868)*(1+2.*(dim-1.)/(o+1.))+(1.05432111912)*dim
        ref[2,4,1,1]=(-0.474782556409)*(1+2.*(dim-1.)/(o+1.))+(-0.345418959816)*dim
        ref[2,4,2,0]=(-0.640169570421)*(1+2.*(dim-1.)/(o+1.))+(2.29862573583)*dim
        ref[2,4,2,1]=(-1.29561079186)*(1+2.*(dim-1.)/(o+1.))+(-0.0229243896616)*dim
        ref[3,0,0,0]=(0.524599634743)*(1+2.*(dim-1.)/(o+1.))+(1.48657883913)*dim
        ref[3,0,0,1]=(-0.786118853533)*(1+2.*(dim-1.)/(o+1.))+(-0.934991212081)*dim
        ref[3,0,1,0]=(-1.26119302657)*(1+2.*(dim-1.)/(o+1.))+(-0.45639437713)*dim
        ref[3,0,1,1]=(1.69966046653)*(1+2.*(dim-1.)/(o+1.))+(-0.173121374354)*dim
        ref[3,0,2,0]=(0.505872873156)*(1+2.*(dim-1.)/(o+1.))+(-0.693315654114)*dim
        ref[3,0,2,1]=(-1.07601104899)*(1+2.*(dim-1.)/(o+1.))+(-1.63307575663)*dim
        ref[3,1,0,0]=(-1.12174154586)*(1+2.*(dim-1.)/(o+1.))+(-0.200310678769)*dim
        ref[3,1,0,1]=(-0.741844605234)*(1+2.*(dim-1.)/(o+1.))+(2.20475885314)*dim
        ref[3,1,1,0]=(-2.00271074147)*(1+2.*(dim-1.)/(o+1.))+(0.306114236606)*dim
        ref[3,1,1,1]=(0.470437983768)*(1+2.*(dim-1.)/(o+1.))+(-1.19671555254)*dim
        ref[3,1,2,0]=(-0.348930377674)*(1+2.*(dim-1.)/(o+1.))+(0.682136816577)*dim
        ref[3,1,2,1]=(-0.653268440332)*(1+2.*(dim-1.)/(o+1.))+(-0.653457938714)*dim
        ref[3,2,0,0]=(-0.323250030589)*(1+2.*(dim-1.)/(o+1.))+(1.64151767382)*dim
        ref[3,2,0,1]=(0.984196908917)*(1+2.*(dim-1.)/(o+1.))+(0.394463769052)*dim
        ref[3,2,1,0]=(-0.275273072003)*(1+2.*(dim-1.)/(o+1.))+(2.22266273838)*dim
        ref[3,2,1,1]=(-1.04603599796)*(1+2.*(dim-1.)/(o+1.))+(-0.813072858233)*dim
        ref[3,2,2,0]=(0.0123017483236)*(1+2.*(dim-1.)/(o+1.))+(-1.00921382659)*dim
        ref[3,2,2,1]=(0.698349007993)*(1+2.*(dim-1.)/(o+1.))+(0.860447076293)*dim
        ref[3,3,0,0]=(-1.92039491753)*(1+2.*(dim-1.)/(o+1.))+(0.363696174714)*dim
        ref[3,3,0,1]=(0.449509323203)*(1+2.*(dim-1.)/(o+1.))+(-1.20964473864)*dim
        ref[3,3,1,0]=(0.0926213111634)*(1+2.*(dim-1.)/(o+1.))+(0.187469872479)*dim
        ref[3,3,1,1]=(-0.331241037228)*(1+2.*(dim-1.)/(o+1.))+(1.29551361978)*dim
        ref[3,3,2,0]=(-1.30763679783)*(1+2.*(dim-1.)/(o+1.))+(1.90224475377)*dim
        ref[3,3,2,1]=(1.69637879413)*(1+2.*(dim-1.)/(o+1.))+(1.67952685501)*dim
        ref[3,4,0,0]=(0.937028161461)*(1+2.*(dim-1.)/(o+1.))+(-0.359993279113)*dim
        ref[3,4,0,1]=(1.6515406216)*(1+2.*(dim-1.)/(o+1.))+(-1.56143511475)*dim
        ref[3,4,1,0]=(-0.886671756108)*(1+2.*(dim-1.)/(o+1.))+(0.613464958119)*dim
        ref[3,4,1,1]=(0.691565811192)*(1+2.*(dim-1.)/(o+1.))+(0.155392775462)*dim
        ref[3,4,2,0]=(0.684820472977)*(1+2.*(dim-1.)/(o+1.))+(0.1823077916)*dim
        ref[3,4,2,1]=(-0.452403917355)*(1+2.*(dim-1.)/(o+1.))+(-1.12180717662)*dim
      res=integrate(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.260967575068)*x[0]**o+(-0.456559015376)*x[0]+(0.663023090785)*x[1]**o+(-0.000383723402374)*x[1]
        ref[0]=o*(0.260967575068)*x_ref[0]**(o-1)+(-0.456559015376)
        ref[1]=o*(0.663023090785)*x_ref[1]**(o-1)+(-0.000383723402374)
      else:
        arg=(-0.679453642968)*x[0]**o+(-0.836558889087)*x[0]+(0.938200435255)*x[1]**o+(-0.368159735907)*x[1]+(-0.0546850112561)*x[2]**o+(0.931986983035)*x[2]
        ref[0]=o*(-0.679453642968)*x_ref[0]**(o-1)+(-0.836558889087)
        ref[1]=o*(0.938200435255)*x_ref[1]**(o-1)+(-0.368159735907)
        ref[2]=o*(-0.0546850112561)*x_ref[2]**(o-1)+(0.931986983035)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.221278607378)*x[0]**o+(-0.25051444588)*x[0]+(0.752161895582)*x[1]**o+(0.662434618125)*x[1]
        arg[1]=(0.984829669888)*x[0]**o+(0.00500258390598)*x[0]+(-0.292523259369)*x[1]**o+(-0.238277884466)*x[1]
        ref[0,0]=o*(0.221278607378)*x_ref[0]**(o-1)+(-0.25051444588)
        ref[0,1]=o*(0.752161895582)*x_ref[1]**(o-1)+(0.662434618125)
        ref[1,0]=o*(0.984829669888)*x_ref[0]**(o-1)+(0.00500258390598)
        ref[1,1]=o*(-0.292523259369)*x_ref[1]**(o-1)+(-0.238277884466)
      else:
        arg[0]=(0.815886350601)*x[0]**o+(-0.41721938077)*x[0]+(0.478358181839)*x[1]**o+(-0.458860964588)*x[1]+(0.575706872631)*x[2]**o+(-0.965426161177)*x[2]
        arg[1]=(-0.159203003073)*x[0]**o+(-0.985356409774)*x[0]+(0.504125935966)*x[1]**o+(0.703505637664)*x[1]+(-0.480016288837)*x[2]**o+(0.544400209204)*x[2]
        ref[0,0]=o*(0.815886350601)*x_ref[0]**(o-1)+(-0.41721938077)
        ref[0,1]=o*(0.478358181839)*x_ref[1]**(o-1)+(-0.458860964588)
        ref[0,2]=o*(0.575706872631)*x_ref[2]**(o-1)+(-0.965426161177)
        ref[1,0]=o*(-0.159203003073)*x_ref[0]**(o-1)+(-0.985356409774)
        ref[1,1]=o*(0.504125935966)*x_ref[1]**(o-1)+(0.703505637664)
        ref[1,2]=o*(-0.480016288837)*x_ref[2]**(o-1)+(0.544400209204)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(-0.722004066753)*x[0]**o+(-0.911439550186)*x[0]+(0.0793361579222)*x[1]**o+(0.958526722957)*x[1]
        arg[0,1]=(-0.0237328054252)*x[0]**o+(0.832621528642)*x[0]+(0.168484408262)*x[1]**o+(-0.181293214021)*x[1]
        arg[0,2]=(0.0284992392733)*x[0]**o+(0.884411090784)*x[0]+(-0.615596355135)*x[1]**o+(-0.564819619851)*x[1]
        arg[0,3]=(0.37408647134)*x[0]**o+(-0.177901551674)*x[0]+(-0.466165048356)*x[1]**o+(-0.278511961588)*x[1]
        arg[0,4]=(-0.24695435531)*x[0]**o+(-0.624082916034)*x[0]+(-0.343849543401)*x[1]**o+(0.661610450445)*x[1]
        arg[1,0]=(-0.176271330569)*x[0]**o+(-0.661612282712)*x[0]+(-0.937163166613)*x[1]**o+(0.376473179587)*x[1]
        arg[1,1]=(0.964111839461)*x[0]**o+(0.576837069403)*x[0]+(-0.718562476684)*x[1]**o+(0.154883495525)*x[1]
        arg[1,2]=(-0.771556809734)*x[0]**o+(0.720054023732)*x[0]+(-0.390732462938)*x[1]**o+(-0.484359019049)*x[1]
        arg[1,3]=(0.565910084656)*x[0]**o+(0.69105192657)*x[0]+(-0.449408955892)*x[1]**o+(-0.789780518242)*x[1]
        arg[1,4]=(0.602485034177)*x[0]**o+(-0.345139282323)*x[0]+(-0.951334152921)*x[1]**o+(-0.478700081579)*x[1]
        arg[2,0]=(0.00114729937901)*x[0]**o+(0.295356090995)*x[0]+(0.982261544035)*x[1]**o+(0.357076141518)*x[1]
        arg[2,1]=(0.653029330145)*x[0]**o+(-0.0178804068723)*x[0]+(-0.0238303862422)*x[1]**o+(0.983033975535)*x[1]
        arg[2,2]=(0.680878878444)*x[0]**o+(-0.404330541218)*x[0]+(0.856544883866)*x[1]**o+(0.71899071821)*x[1]
        arg[2,3]=(0.0674687202072)*x[0]**o+(-0.748990717637)*x[0]+(0.185597242309)*x[1]**o+(0.623849127358)*x[1]
        arg[2,4]=(0.302869699915)*x[0]**o+(-0.852621552847)*x[0]+(0.115557112172)*x[1]**o+(-0.0827214208715)*x[1]
        arg[3,0]=(0.0929072570643)*x[0]**o+(0.533152382923)*x[0]+(-0.0913754984708)*x[1]**o+(-0.25865969337)*x[1]
        arg[3,1]=(-0.794799886368)*x[0]**o+(-0.865263962776)*x[0]+(0.441343356932)*x[1]**o+(-0.541975483415)*x[1]
        arg[3,2]=(0.261341948686)*x[0]**o+(0.378142729621)*x[0]+(0.0556054089673)*x[1]**o+(-0.854195169318)*x[1]
        arg[3,3]=(0.177222467454)*x[0]**o+(-0.400820197499)*x[0]+(0.88513075927)*x[1]**o+(0.0420082325044)*x[1]
        arg[3,4]=(0.307178422145)*x[0]**o+(-0.203765349758)*x[0]+(-0.546052990937)*x[1]**o+(-0.960312136292)*x[1]
        ref[0,0,0]=o*(-0.722004066753)*x_ref[0]**(o-1)+(-0.911439550186)
        ref[0,0,1]=o*(0.0793361579222)*x_ref[1]**(o-1)+(0.958526722957)
        ref[0,1,0]=o*(-0.0237328054252)*x_ref[0]**(o-1)+(0.832621528642)
        ref[0,1,1]=o*(0.168484408262)*x_ref[1]**(o-1)+(-0.181293214021)
        ref[0,2,0]=o*(0.0284992392733)*x_ref[0]**(o-1)+(0.884411090784)
        ref[0,2,1]=o*(-0.615596355135)*x_ref[1]**(o-1)+(-0.564819619851)
        ref[0,3,0]=o*(0.37408647134)*x_ref[0]**(o-1)+(-0.177901551674)
        ref[0,3,1]=o*(-0.466165048356)*x_ref[1]**(o-1)+(-0.278511961588)
        ref[0,4,0]=o*(-0.24695435531)*x_ref[0]**(o-1)+(-0.624082916034)
        ref[0,4,1]=o*(-0.343849543401)*x_ref[1]**(o-1)+(0.661610450445)
        ref[1,0,0]=o*(-0.176271330569)*x_ref[0]**(o-1)+(-0.661612282712)
        ref[1,0,1]=o*(-0.937163166613)*x_ref[1]**(o-1)+(0.376473179587)
        ref[1,1,0]=o*(0.964111839461)*x_ref[0]**(o-1)+(0.576837069403)
        ref[1,1,1]=o*(-0.718562476684)*x_ref[1]**(o-1)+(0.154883495525)
        ref[1,2,0]=o*(-0.771556809734)*x_ref[0]**(o-1)+(0.720054023732)
        ref[1,2,1]=o*(-0.390732462938)*x_ref[1]**(o-1)+(-0.484359019049)
        ref[1,3,0]=o*(0.565910084656)*x_ref[0]**(o-1)+(0.69105192657)
        ref[1,3,1]=o*(-0.449408955892)*x_ref[1]**(o-1)+(-0.789780518242)
        ref[1,4,0]=o*(0.602485034177)*x_ref[0]**(o-1)+(-0.345139282323)
        ref[1,4,1]=o*(-0.951334152921)*x_ref[1]**(o-1)+(-0.478700081579)
        ref[2,0,0]=o*(0.00114729937901)*x_ref[0]**(o-1)+(0.295356090995)
        ref[2,0,1]=o*(0.982261544035)*x_ref[1]**(o-1)+(0.357076141518)
        ref[2,1,0]=o*(0.653029330145)*x_ref[0]**(o-1)+(-0.0178804068723)
        ref[2,1,1]=o*(-0.0238303862422)*x_ref[1]**(o-1)+(0.983033975535)
        ref[2,2,0]=o*(0.680878878444)*x_ref[0]**(o-1)+(-0.404330541218)
        ref[2,2,1]=o*(0.856544883866)*x_ref[1]**(o-1)+(0.71899071821)
        ref[2,3,0]=o*(0.0674687202072)*x_ref[0]**(o-1)+(-0.748990717637)
        ref[2,3,1]=o*(0.185597242309)*x_ref[1]**(o-1)+(0.623849127358)
        ref[2,4,0]=o*(0.302869699915)*x_ref[0]**(o-1)+(-0.852621552847)
        ref[2,4,1]=o*(0.115557112172)*x_ref[1]**(o-1)+(-0.0827214208715)
        ref[3,0,0]=o*(0.0929072570643)*x_ref[0]**(o-1)+(0.533152382923)
        ref[3,0,1]=o*(-0.0913754984708)*x_ref[1]**(o-1)+(-0.25865969337)
        ref[3,1,0]=o*(-0.794799886368)*x_ref[0]**(o-1)+(-0.865263962776)
        ref[3,1,1]=o*(0.441343356932)*x_ref[1]**(o-1)+(-0.541975483415)
        ref[3,2,0]=o*(0.261341948686)*x_ref[0]**(o-1)+(0.378142729621)
        ref[3,2,1]=o*(0.0556054089673)*x_ref[1]**(o-1)+(-0.854195169318)
        ref[3,3,0]=o*(0.177222467454)*x_ref[0]**(o-1)+(-0.400820197499)
        ref[3,3,1]=o*(0.88513075927)*x_ref[1]**(o-1)+(0.0420082325044)
        ref[3,4,0]=o*(0.307178422145)*x_ref[0]**(o-1)+(-0.203765349758)
        ref[3,4,1]=o*(-0.546052990937)*x_ref[1]**(o-1)+(-0.960312136292)
      else:
        arg[0,0]=(0.767891828291)*x[0]**o+(-0.0840162180616)*x[0]+(-0.942354055114)*x[1]**o+(-0.458324281794)*x[1]+(-0.837589052489)*x[2]**o+(-0.49741065871)*x[2]
        arg[0,1]=(-0.770574732815)*x[0]**o+(0.949362675967)*x[0]+(0.334062389906)*x[1]**o+(0.589134582043)*x[1]+(0.611876443107)*x[2]**o+(-0.476746595186)*x[2]
        arg[0,2]=(-0.113076234857)*x[0]**o+(0.207116555982)*x[0]+(-0.434423505296)*x[1]**o+(0.582451788291)*x[1]+(-0.858451242875)*x[2]**o+(-0.229934112115)*x[2]
        arg[0,3]=(0.55494281628)*x[0]**o+(0.53116989281)*x[0]+(0.150549353151)*x[1]**o+(0.70886785918)*x[1]+(-0.148436740612)*x[2]**o+(0.755584812901)*x[2]
        arg[0,4]=(0.0855991607862)*x[0]**o+(0.509120320151)*x[0]+(-0.303374431746)*x[1]**o+(0.1813545791)*x[1]+(-0.0434667822132)*x[2]**o+(0.395447849194)*x[2]
        arg[1,0]=(-0.779720617706)*x[0]**o+(0.730668258583)*x[0]+(-0.0315884698035)*x[1]**o+(-0.0929845369431)*x[1]+(0.0853332239352)*x[2]**o+(0.303372304789)*x[2]
        arg[1,1]=(-0.687021672626)*x[0]**o+(-0.380139153436)*x[0]+(-0.624774641851)*x[1]**o+(0.331833510141)*x[1]+(-0.747115143343)*x[2]**o+(-0.901384102505)*x[2]
        arg[1,2]=(-0.0528543338921)*x[0]**o+(-0.683764012224)*x[0]+(0.498886779776)*x[1]**o+(-0.396777215862)*x[1]+(0.777051418657)*x[2]**o+(-0.786345142641)*x[2]
        arg[1,3]=(-0.823112017253)*x[0]**o+(0.278818264102)*x[0]+(-0.79372160065)*x[1]**o+(0.254044599609)*x[1]+(-0.372925254827)*x[2]**o+(-0.85054008508)*x[2]
        arg[1,4]=(-0.247156845703)*x[0]**o+(-0.944335229412)*x[0]+(-0.572532409179)*x[1]**o+(0.289091364421)*x[1]+(0.848874739954)*x[2]**o+(0.705098999827)*x[2]
        arg[2,0]=(0.0958493700534)*x[0]**o+(-0.334070579338)*x[0]+(-0.74117408785)*x[1]**o+(0.0609646344894)*x[1]+(-0.328888690751)*x[2]**o+(-0.0430772626841)*x[2]
        arg[2,1]=(0.846871401827)*x[0]**o+(0.379912211333)*x[0]+(-0.197881866966)*x[1]**o+(-0.551498190852)*x[1]+(0.416977164384)*x[2]**o+(0.09398324743)*x[2]
        arg[2,2]=(0.737261473925)*x[0]**o+(0.696028328637)*x[0]+(-0.735110324145)*x[1]**o+(0.949453963809)*x[1]+(-0.894722167055)*x[2]**o+(-0.16124057159)*x[2]
        arg[2,3]=(0.267556059297)*x[0]**o+(-0.633581895464)*x[0]+(-0.176232065927)*x[1]**o+(0.98533106984)*x[1]+(0.404671434387)*x[2]**o+(0.343560595562)*x[2]
        arg[2,4]=(0.502981833324)*x[0]**o+(0.716659442696)*x[0]+(-0.349106283426)*x[1]**o+(0.0726954534053)*x[1]+(0.468516366697)*x[2]**o+(0.328926164383)*x[2]
        arg[3,0]=(0.846509624251)*x[0]**o+(-0.34161660121)*x[0]+(0.25149102353)*x[1]**o+(-0.0130496471853)*x[1]+(-0.556826239055)*x[2]**o+(0.774658667066)*x[2]
        arg[3,1]=(-0.0777319662414)*x[0]**o+(0.992622365536)*x[0]+(0.720504365851)*x[1]**o+(0.0701085839568)*x[1]+(0.475573965193)*x[2]**o+(0.402599619829)*x[2]
        arg[3,2]=(-0.311228599853)*x[0]**o+(-0.732684576916)*x[0]+(0.308213192232)*x[1]**o+(-0.131325871896)*x[1]+(0.964631653839)*x[2]**o+(-0.600646614776)*x[2]
        arg[3,3]=(0.656589254036)*x[0]**o+(-0.275449493207)*x[0]+(-0.611075965309)*x[1]**o+(-0.460879628412)*x[1]+(-0.820861932527)*x[2]**o+(0.424066222721)*x[2]
        arg[3,4]=(0.303269333663)*x[0]**o+(-0.745774722275)*x[0]+(0.82933539351)*x[1]**o+(-0.295464064484)*x[1]+(0.818135297647)*x[2]**o+(-0.989600307642)*x[2]
        ref[0,0,0]=o*(0.767891828291)*x_ref[0]**(o-1)+(-0.0840162180616)
        ref[0,0,1]=o*(-0.942354055114)*x_ref[1]**(o-1)+(-0.458324281794)
        ref[0,0,2]=o*(-0.837589052489)*x_ref[2]**(o-1)+(-0.49741065871)
        ref[0,1,0]=o*(-0.770574732815)*x_ref[0]**(o-1)+(0.949362675967)
        ref[0,1,1]=o*(0.334062389906)*x_ref[1]**(o-1)+(0.589134582043)
        ref[0,1,2]=o*(0.611876443107)*x_ref[2]**(o-1)+(-0.476746595186)
        ref[0,2,0]=o*(-0.113076234857)*x_ref[0]**(o-1)+(0.207116555982)
        ref[0,2,1]=o*(-0.434423505296)*x_ref[1]**(o-1)+(0.582451788291)
        ref[0,2,2]=o*(-0.858451242875)*x_ref[2]**(o-1)+(-0.229934112115)
        ref[0,3,0]=o*(0.55494281628)*x_ref[0]**(o-1)+(0.53116989281)
        ref[0,3,1]=o*(0.150549353151)*x_ref[1]**(o-1)+(0.70886785918)
        ref[0,3,2]=o*(-0.148436740612)*x_ref[2]**(o-1)+(0.755584812901)
        ref[0,4,0]=o*(0.0855991607862)*x_ref[0]**(o-1)+(0.509120320151)
        ref[0,4,1]=o*(-0.303374431746)*x_ref[1]**(o-1)+(0.1813545791)
        ref[0,4,2]=o*(-0.0434667822132)*x_ref[2]**(o-1)+(0.395447849194)
        ref[1,0,0]=o*(-0.779720617706)*x_ref[0]**(o-1)+(0.730668258583)
        ref[1,0,1]=o*(-0.0315884698035)*x_ref[1]**(o-1)+(-0.0929845369431)
        ref[1,0,2]=o*(0.0853332239352)*x_ref[2]**(o-1)+(0.303372304789)
        ref[1,1,0]=o*(-0.687021672626)*x_ref[0]**(o-1)+(-0.380139153436)
        ref[1,1,1]=o*(-0.624774641851)*x_ref[1]**(o-1)+(0.331833510141)
        ref[1,1,2]=o*(-0.747115143343)*x_ref[2]**(o-1)+(-0.901384102505)
        ref[1,2,0]=o*(-0.0528543338921)*x_ref[0]**(o-1)+(-0.683764012224)
        ref[1,2,1]=o*(0.498886779776)*x_ref[1]**(o-1)+(-0.396777215862)
        ref[1,2,2]=o*(0.777051418657)*x_ref[2]**(o-1)+(-0.786345142641)
        ref[1,3,0]=o*(-0.823112017253)*x_ref[0]**(o-1)+(0.278818264102)
        ref[1,3,1]=o*(-0.79372160065)*x_ref[1]**(o-1)+(0.254044599609)
        ref[1,3,2]=o*(-0.372925254827)*x_ref[2]**(o-1)+(-0.85054008508)
        ref[1,4,0]=o*(-0.247156845703)*x_ref[0]**(o-1)+(-0.944335229412)
        ref[1,4,1]=o*(-0.572532409179)*x_ref[1]**(o-1)+(0.289091364421)
        ref[1,4,2]=o*(0.848874739954)*x_ref[2]**(o-1)+(0.705098999827)
        ref[2,0,0]=o*(0.0958493700534)*x_ref[0]**(o-1)+(-0.334070579338)
        ref[2,0,1]=o*(-0.74117408785)*x_ref[1]**(o-1)+(0.0609646344894)
        ref[2,0,2]=o*(-0.328888690751)*x_ref[2]**(o-1)+(-0.0430772626841)
        ref[2,1,0]=o*(0.846871401827)*x_ref[0]**(o-1)+(0.379912211333)
        ref[2,1,1]=o*(-0.197881866966)*x_ref[1]**(o-1)+(-0.551498190852)
        ref[2,1,2]=o*(0.416977164384)*x_ref[2]**(o-1)+(0.09398324743)
        ref[2,2,0]=o*(0.737261473925)*x_ref[0]**(o-1)+(0.696028328637)
        ref[2,2,1]=o*(-0.735110324145)*x_ref[1]**(o-1)+(0.949453963809)
        ref[2,2,2]=o*(-0.894722167055)*x_ref[2]**(o-1)+(-0.16124057159)
        ref[2,3,0]=o*(0.267556059297)*x_ref[0]**(o-1)+(-0.633581895464)
        ref[2,3,1]=o*(-0.176232065927)*x_ref[1]**(o-1)+(0.98533106984)
        ref[2,3,2]=o*(0.404671434387)*x_ref[2]**(o-1)+(0.343560595562)
        ref[2,4,0]=o*(0.502981833324)*x_ref[0]**(o-1)+(0.716659442696)
        ref[2,4,1]=o*(-0.349106283426)*x_ref[1]**(o-1)+(0.0726954534053)
        ref[2,4,2]=o*(0.468516366697)*x_ref[2]**(o-1)+(0.328926164383)
        ref[3,0,0]=o*(0.846509624251)*x_ref[0]**(o-1)+(-0.34161660121)
        ref[3,0,1]=o*(0.25149102353)*x_ref[1]**(o-1)+(-0.0130496471853)
        ref[3,0,2]=o*(-0.556826239055)*x_ref[2]**(o-1)+(0.774658667066)
        ref[3,1,0]=o*(-0.0777319662414)*x_ref[0]**(o-1)+(0.992622365536)
        ref[3,1,1]=o*(0.720504365851)*x_ref[1]**(o-1)+(0.0701085839568)
        ref[3,1,2]=o*(0.475573965193)*x_ref[2]**(o-1)+(0.402599619829)
        ref[3,2,0]=o*(-0.311228599853)*x_ref[0]**(o-1)+(-0.732684576916)
        ref[3,2,1]=o*(0.308213192232)*x_ref[1]**(o-1)+(-0.131325871896)
        ref[3,2,2]=o*(0.964631653839)*x_ref[2]**(o-1)+(-0.600646614776)
        ref[3,3,0]=o*(0.656589254036)*x_ref[0]**(o-1)+(-0.275449493207)
        ref[3,3,1]=o*(-0.611075965309)*x_ref[1]**(o-1)+(-0.460879628412)
        ref[3,3,2]=o*(-0.820861932527)*x_ref[2]**(o-1)+(0.424066222721)
        ref[3,4,0]=o*(0.303269333663)*x_ref[0]**(o-1)+(-0.745774722275)
        ref[3,4,1]=o*(0.82933539351)*x_ref[1]**(o-1)+(-0.295464064484)
        ref[3,4,2]=o*(0.818135297647)*x_ref[2]**(o-1)+(-0.989600307642)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.572693829631)*x[0]**o+(0.583556749477)*x[0]+(0.22594627003)*x[1]**o+(0.0199162626119)*x[1]
        arg[0,0,1]=(0.196266833125)*x[0]**o+(0.544824498113)*x[0]+(0.150736772157)*x[1]**o+(0.362174902716)*x[1]
        arg[0,1,0]=(0.169481539273)*x[0]**o+(0.0355504217105)*x[0]+(-0.156746013218)*x[1]**o+(0.879476257954)*x[1]
        arg[0,1,1]=(0.824826637612)*x[0]**o+(-0.732645177858)*x[0]+(0.732873648361)*x[1]**o+(-0.40090515696)*x[1]
        arg[1,0,0]=(0.77348720999)*x[0]**o+(0.229475702878)*x[0]+(0.686324947328)*x[1]**o+(0.313197971717)*x[1]
        arg[1,0,1]=(-0.966626441904)*x[0]**o+(0.989778282376)*x[0]+(0.0447214020611)*x[1]**o+(0.439337525343)*x[1]
        arg[1,1,0]=(-0.774806567769)*x[0]**o+(0.634591112025)*x[0]+(0.797720133135)*x[1]**o+(0.579137230562)*x[1]
        arg[1,1,1]=(-0.118703583742)*x[0]**o+(0.482822282284)*x[0]+(0.141839968056)*x[1]**o+(0.916615496002)*x[1]
        arg[2,0,0]=(0.7419378407)*x[0]**o+(0.754978933397)*x[0]+(-0.839586800033)*x[1]**o+(-0.525830008753)*x[1]
        arg[2,0,1]=(-0.855040380309)*x[0]**o+(0.423661210447)*x[0]+(0.538277902172)*x[1]**o+(0.127617006384)*x[1]
        arg[2,1,0]=(0.107349578938)*x[0]**o+(0.697738123598)*x[0]+(0.862106802027)*x[1]**o+(-0.829827651082)*x[1]
        arg[2,1,1]=(-0.970208311818)*x[0]**o+(0.806698181512)*x[0]+(-0.374007524144)*x[1]**o+(-0.0600501198736)*x[1]
        arg[3,0,0]=(0.223312188411)*x[0]**o+(-0.130776319489)*x[0]+(0.616021313401)*x[1]**o+(-0.552848902138)*x[1]
        arg[3,0,1]=(-0.461837450406)*x[0]**o+(-0.895912538242)*x[0]+(0.477606324718)*x[1]**o+(0.220148409572)*x[1]
        arg[3,1,0]=(0.0783005249695)*x[0]**o+(-0.762350707337)*x[0]+(0.616470467018)*x[1]**o+(-0.154358859634)*x[1]
        arg[3,1,1]=(-0.604945998091)*x[0]**o+(0.23801834237)*x[0]+(0.500666300529)*x[1]**o+(0.709482270038)*x[1]
        arg[4,0,0]=(0.746627054515)*x[0]**o+(0.550282853502)*x[0]+(-0.223527434147)*x[1]**o+(-0.291559486198)*x[1]
        arg[4,0,1]=(0.512462457431)*x[0]**o+(0.0908594735188)*x[0]+(-0.888862229845)*x[1]**o+(-0.969985508695)*x[1]
        arg[4,1,0]=(0.606213701819)*x[0]**o+(0.147049558356)*x[0]+(-0.0365908849884)*x[1]**o+(-0.584510320466)*x[1]
        arg[4,1,1]=(-0.110180223906)*x[0]**o+(-0.83442554862)*x[0]+(-0.0516097061459)*x[1]**o+(-0.155615857347)*x[1]
        arg[5,0,0]=(0.603753207781)*x[0]**o+(0.302032475639)*x[0]+(-0.68719165993)*x[1]**o+(-0.375829665239)*x[1]
        arg[5,0,1]=(-0.599357977805)*x[0]**o+(0.562486062075)*x[0]+(0.102714751356)*x[1]**o+(-0.114179799185)*x[1]
        arg[5,1,0]=(0.874756501256)*x[0]**o+(0.550416007969)*x[0]+(0.669721605367)*x[1]**o+(0.24372600222)*x[1]
        arg[5,1,1]=(0.651356889701)*x[0]**o+(0.178366196323)*x[0]+(0.762268544189)*x[1]**o+(0.515014711815)*x[1]
        ref[0,0,0,0]=o*(-0.572693829631)*x_ref[0]**(o-1)+(0.583556749477)
        ref[0,0,0,1]=o*(0.22594627003)*x_ref[1]**(o-1)+(0.0199162626119)
        ref[0,0,1,0]=o*(0.196266833125)*x_ref[0]**(o-1)+(0.544824498113)
        ref[0,0,1,1]=o*(0.150736772157)*x_ref[1]**(o-1)+(0.362174902716)
        ref[0,1,0,0]=o*(0.169481539273)*x_ref[0]**(o-1)+(0.0355504217105)
        ref[0,1,0,1]=o*(-0.156746013218)*x_ref[1]**(o-1)+(0.879476257954)
        ref[0,1,1,0]=o*(0.824826637612)*x_ref[0]**(o-1)+(-0.732645177858)
        ref[0,1,1,1]=o*(0.732873648361)*x_ref[1]**(o-1)+(-0.40090515696)
        ref[1,0,0,0]=o*(0.77348720999)*x_ref[0]**(o-1)+(0.229475702878)
        ref[1,0,0,1]=o*(0.686324947328)*x_ref[1]**(o-1)+(0.313197971717)
        ref[1,0,1,0]=o*(-0.966626441904)*x_ref[0]**(o-1)+(0.989778282376)
        ref[1,0,1,1]=o*(0.0447214020611)*x_ref[1]**(o-1)+(0.439337525343)
        ref[1,1,0,0]=o*(-0.774806567769)*x_ref[0]**(o-1)+(0.634591112025)
        ref[1,1,0,1]=o*(0.797720133135)*x_ref[1]**(o-1)+(0.579137230562)
        ref[1,1,1,0]=o*(-0.118703583742)*x_ref[0]**(o-1)+(0.482822282284)
        ref[1,1,1,1]=o*(0.141839968056)*x_ref[1]**(o-1)+(0.916615496002)
        ref[2,0,0,0]=o*(0.7419378407)*x_ref[0]**(o-1)+(0.754978933397)
        ref[2,0,0,1]=o*(-0.839586800033)*x_ref[1]**(o-1)+(-0.525830008753)
        ref[2,0,1,0]=o*(-0.855040380309)*x_ref[0]**(o-1)+(0.423661210447)
        ref[2,0,1,1]=o*(0.538277902172)*x_ref[1]**(o-1)+(0.127617006384)
        ref[2,1,0,0]=o*(0.107349578938)*x_ref[0]**(o-1)+(0.697738123598)
        ref[2,1,0,1]=o*(0.862106802027)*x_ref[1]**(o-1)+(-0.829827651082)
        ref[2,1,1,0]=o*(-0.970208311818)*x_ref[0]**(o-1)+(0.806698181512)
        ref[2,1,1,1]=o*(-0.374007524144)*x_ref[1]**(o-1)+(-0.0600501198736)
        ref[3,0,0,0]=o*(0.223312188411)*x_ref[0]**(o-1)+(-0.130776319489)
        ref[3,0,0,1]=o*(0.616021313401)*x_ref[1]**(o-1)+(-0.552848902138)
        ref[3,0,1,0]=o*(-0.461837450406)*x_ref[0]**(o-1)+(-0.895912538242)
        ref[3,0,1,1]=o*(0.477606324718)*x_ref[1]**(o-1)+(0.220148409572)
        ref[3,1,0,0]=o*(0.0783005249695)*x_ref[0]**(o-1)+(-0.762350707337)
        ref[3,1,0,1]=o*(0.616470467018)*x_ref[1]**(o-1)+(-0.154358859634)
        ref[3,1,1,0]=o*(-0.604945998091)*x_ref[0]**(o-1)+(0.23801834237)
        ref[3,1,1,1]=o*(0.500666300529)*x_ref[1]**(o-1)+(0.709482270038)
        ref[4,0,0,0]=o*(0.746627054515)*x_ref[0]**(o-1)+(0.550282853502)
        ref[4,0,0,1]=o*(-0.223527434147)*x_ref[1]**(o-1)+(-0.291559486198)
        ref[4,0,1,0]=o*(0.512462457431)*x_ref[0]**(o-1)+(0.0908594735188)
        ref[4,0,1,1]=o*(-0.888862229845)*x_ref[1]**(o-1)+(-0.969985508695)
        ref[4,1,0,0]=o*(0.606213701819)*x_ref[0]**(o-1)+(0.147049558356)
        ref[4,1,0,1]=o*(-0.0365908849884)*x_ref[1]**(o-1)+(-0.584510320466)
        ref[4,1,1,0]=o*(-0.110180223906)*x_ref[0]**(o-1)+(-0.83442554862)
        ref[4,1,1,1]=o*(-0.0516097061459)*x_ref[1]**(o-1)+(-0.155615857347)
        ref[5,0,0,0]=o*(0.603753207781)*x_ref[0]**(o-1)+(0.302032475639)
        ref[5,0,0,1]=o*(-0.68719165993)*x_ref[1]**(o-1)+(-0.375829665239)
        ref[5,0,1,0]=o*(-0.599357977805)*x_ref[0]**(o-1)+(0.562486062075)
        ref[5,0,1,1]=o*(0.102714751356)*x_ref[1]**(o-1)+(-0.114179799185)
        ref[5,1,0,0]=o*(0.874756501256)*x_ref[0]**(o-1)+(0.550416007969)
        ref[5,1,0,1]=o*(0.669721605367)*x_ref[1]**(o-1)+(0.24372600222)
        ref[5,1,1,0]=o*(0.651356889701)*x_ref[0]**(o-1)+(0.178366196323)
        ref[5,1,1,1]=o*(0.762268544189)*x_ref[1]**(o-1)+(0.515014711815)
      else:
        arg[0,0,0]=(-0.634739027866)*x[0]**o+(-0.696370463772)*x[0]+(0.503815931458)*x[1]**o+(-0.457824888032)*x[1]+(0.655074800235)*x[2]**o+(-0.456111009206)*x[2]
        arg[0,0,1]=(-0.533710147141)*x[0]**o+(-0.0814085290298)*x[0]+(-0.0894565935545)*x[1]**o+(-0.3775655102)*x[1]+(-0.134051544262)*x[2]**o+(0.000713749886013)*x[2]
        arg[0,1,0]=(-0.64394535332)*x[0]**o+(0.430741797171)*x[0]+(0.985157009863)*x[1]**o+(-0.655331266432)*x[1]+(-0.436629135988)*x[2]**o+(-0.404480211702)*x[2]
        arg[0,1,1]=(-0.0858143551662)*x[0]**o+(0.0452769288327)*x[0]+(0.930452744895)*x[1]**o+(-0.877149214078)*x[1]+(0.675472222413)*x[2]**o+(0.917497654616)*x[2]
        arg[1,0,0]=(0.88346309442)*x[0]**o+(0.892116796676)*x[0]+(0.22765643526)*x[1]**o+(-0.934442833938)*x[1]+(0.0781755340006)*x[2]**o+(0.106009847768)*x[2]
        arg[1,0,1]=(0.20824563298)*x[0]**o+(0.613584602599)*x[0]+(-0.482386438919)*x[1]**o+(-0.155840561301)*x[1]+(0.817890842096)*x[2]**o+(0.798519843541)*x[2]
        arg[1,1,0]=(-0.915853569389)*x[0]**o+(0.573041720657)*x[0]+(0.236967108991)*x[1]**o+(0.421805849528)*x[1]+(-0.576407564364)*x[2]**o+(0.0709741076428)*x[2]
        arg[1,1,1]=(0.530417188741)*x[0]**o+(0.872073272277)*x[0]+(-0.21047376763)*x[1]**o+(0.88490064655)*x[1]+(0.523085602451)*x[2]**o+(-0.926133382905)*x[2]
        arg[2,0,0]=(-0.974186019136)*x[0]**o+(-0.332433414417)*x[0]+(-0.500938330608)*x[1]**o+(-0.113188574497)*x[1]+(0.228656301416)*x[2]**o+(-0.375756425313)*x[2]
        arg[2,0,1]=(0.246057421673)*x[0]**o+(0.231587929687)*x[0]+(0.398413883486)*x[1]**o+(-0.390998026525)*x[1]+(0.273401149062)*x[2]**o+(-0.296142647578)*x[2]
        arg[2,1,0]=(0.283717204372)*x[0]**o+(-0.545852417255)*x[0]+(-0.616345461169)*x[1]**o+(-0.705051061202)*x[1]+(0.664377876346)*x[2]**o+(0.338382503272)*x[2]
        arg[2,1,1]=(-0.433297970157)*x[0]**o+(-0.716600154712)*x[0]+(0.836194932687)*x[1]**o+(0.0903372116119)*x[1]+(0.113374499353)*x[2]**o+(-0.588810627425)*x[2]
        arg[3,0,0]=(0.404388255474)*x[0]**o+(0.521816615546)*x[0]+(-0.0976581170689)*x[1]**o+(-0.678563989037)*x[1]+(0.550953471125)*x[2]**o+(-0.668793844235)*x[2]
        arg[3,0,1]=(-0.66023047868)*x[0]**o+(0.822794930168)*x[0]+(-0.284215677966)*x[1]**o+(-0.779237447318)*x[1]+(-0.191727752719)*x[2]**o+(-0.748596484262)*x[2]
        arg[3,1,0]=(-0.0610807571905)*x[0]**o+(0.044623781518)*x[0]+(0.851226664753)*x[1]**o+(0.921090244094)*x[1]+(0.494414571241)*x[2]**o+(-0.473831333684)*x[2]
        arg[3,1,1]=(0.54651807228)*x[0]**o+(-0.580525990778)*x[0]+(-0.822936606301)*x[1]**o+(-0.0169562878937)*x[1]+(-0.732219249555)*x[2]**o+(-0.435156155008)*x[2]
        arg[4,0,0]=(0.830021032139)*x[0]**o+(0.976836986062)*x[0]+(0.534162405111)*x[1]**o+(-0.286292103539)*x[1]+(0.820276894761)*x[2]**o+(0.245897387706)*x[2]
        arg[4,0,1]=(0.362201493311)*x[0]**o+(-0.186328147108)*x[0]+(-0.91611458572)*x[1]**o+(-0.965258904175)*x[1]+(0.61850292372)*x[2]**o+(0.410229835478)*x[2]
        arg[4,1,0]=(0.0726313013515)*x[0]**o+(-0.974714544914)*x[0]+(0.46313543268)*x[1]**o+(0.100486704072)*x[1]+(-0.149124857233)*x[2]**o+(0.455733236337)*x[2]
        arg[4,1,1]=(-0.875220220201)*x[0]**o+(-0.416152301789)*x[0]+(-0.0145627466423)*x[1]**o+(0.267856009119)*x[1]+(-0.604260174049)*x[2]**o+(0.689107997051)*x[2]
        arg[5,0,0]=(0.196182127971)*x[0]**o+(0.337209503833)*x[0]+(0.595342239158)*x[1]**o+(0.507423300857)*x[1]+(0.934760469389)*x[2]**o+(-0.198686443328)*x[2]
        arg[5,0,1]=(0.77210909967)*x[0]**o+(-0.310778136545)*x[0]+(0.512326103142)*x[1]**o+(-0.501056206243)*x[1]+(0.98755961257)*x[2]**o+(-0.658414340038)*x[2]
        arg[5,1,0]=(0.375277702198)*x[0]**o+(-0.603850047717)*x[0]+(0.677227868069)*x[1]**o+(-0.784640615159)*x[1]+(0.177350721639)*x[2]**o+(0.241053657333)*x[2]
        arg[5,1,1]=(-0.368625522736)*x[0]**o+(0.196977476909)*x[0]+(-0.582451887206)*x[1]**o+(0.598240904158)*x[1]+(0.102240230796)*x[2]**o+(0.487322529158)*x[2]
        ref[0,0,0,0]=o*(-0.634739027866)*x_ref[0]**(o-1)+(-0.696370463772)
        ref[0,0,0,1]=o*(0.503815931458)*x_ref[1]**(o-1)+(-0.457824888032)
        ref[0,0,0,2]=o*(0.655074800235)*x_ref[2]**(o-1)+(-0.456111009206)
        ref[0,0,1,0]=o*(-0.533710147141)*x_ref[0]**(o-1)+(-0.0814085290298)
        ref[0,0,1,1]=o*(-0.0894565935545)*x_ref[1]**(o-1)+(-0.3775655102)
        ref[0,0,1,2]=o*(-0.134051544262)*x_ref[2]**(o-1)+(0.000713749886013)
        ref[0,1,0,0]=o*(-0.64394535332)*x_ref[0]**(o-1)+(0.430741797171)
        ref[0,1,0,1]=o*(0.985157009863)*x_ref[1]**(o-1)+(-0.655331266432)
        ref[0,1,0,2]=o*(-0.436629135988)*x_ref[2]**(o-1)+(-0.404480211702)
        ref[0,1,1,0]=o*(-0.0858143551662)*x_ref[0]**(o-1)+(0.0452769288327)
        ref[0,1,1,1]=o*(0.930452744895)*x_ref[1]**(o-1)+(-0.877149214078)
        ref[0,1,1,2]=o*(0.675472222413)*x_ref[2]**(o-1)+(0.917497654616)
        ref[1,0,0,0]=o*(0.88346309442)*x_ref[0]**(o-1)+(0.892116796676)
        ref[1,0,0,1]=o*(0.22765643526)*x_ref[1]**(o-1)+(-0.934442833938)
        ref[1,0,0,2]=o*(0.0781755340006)*x_ref[2]**(o-1)+(0.106009847768)
        ref[1,0,1,0]=o*(0.20824563298)*x_ref[0]**(o-1)+(0.613584602599)
        ref[1,0,1,1]=o*(-0.482386438919)*x_ref[1]**(o-1)+(-0.155840561301)
        ref[1,0,1,2]=o*(0.817890842096)*x_ref[2]**(o-1)+(0.798519843541)
        ref[1,1,0,0]=o*(-0.915853569389)*x_ref[0]**(o-1)+(0.573041720657)
        ref[1,1,0,1]=o*(0.236967108991)*x_ref[1]**(o-1)+(0.421805849528)
        ref[1,1,0,2]=o*(-0.576407564364)*x_ref[2]**(o-1)+(0.0709741076428)
        ref[1,1,1,0]=o*(0.530417188741)*x_ref[0]**(o-1)+(0.872073272277)
        ref[1,1,1,1]=o*(-0.21047376763)*x_ref[1]**(o-1)+(0.88490064655)
        ref[1,1,1,2]=o*(0.523085602451)*x_ref[2]**(o-1)+(-0.926133382905)
        ref[2,0,0,0]=o*(-0.974186019136)*x_ref[0]**(o-1)+(-0.332433414417)
        ref[2,0,0,1]=o*(-0.500938330608)*x_ref[1]**(o-1)+(-0.113188574497)
        ref[2,0,0,2]=o*(0.228656301416)*x_ref[2]**(o-1)+(-0.375756425313)
        ref[2,0,1,0]=o*(0.246057421673)*x_ref[0]**(o-1)+(0.231587929687)
        ref[2,0,1,1]=o*(0.398413883486)*x_ref[1]**(o-1)+(-0.390998026525)
        ref[2,0,1,2]=o*(0.273401149062)*x_ref[2]**(o-1)+(-0.296142647578)
        ref[2,1,0,0]=o*(0.283717204372)*x_ref[0]**(o-1)+(-0.545852417255)
        ref[2,1,0,1]=o*(-0.616345461169)*x_ref[1]**(o-1)+(-0.705051061202)
        ref[2,1,0,2]=o*(0.664377876346)*x_ref[2]**(o-1)+(0.338382503272)
        ref[2,1,1,0]=o*(-0.433297970157)*x_ref[0]**(o-1)+(-0.716600154712)
        ref[2,1,1,1]=o*(0.836194932687)*x_ref[1]**(o-1)+(0.0903372116119)
        ref[2,1,1,2]=o*(0.113374499353)*x_ref[2]**(o-1)+(-0.588810627425)
        ref[3,0,0,0]=o*(0.404388255474)*x_ref[0]**(o-1)+(0.521816615546)
        ref[3,0,0,1]=o*(-0.0976581170689)*x_ref[1]**(o-1)+(-0.678563989037)
        ref[3,0,0,2]=o*(0.550953471125)*x_ref[2]**(o-1)+(-0.668793844235)
        ref[3,0,1,0]=o*(-0.66023047868)*x_ref[0]**(o-1)+(0.822794930168)
        ref[3,0,1,1]=o*(-0.284215677966)*x_ref[1]**(o-1)+(-0.779237447318)
        ref[3,0,1,2]=o*(-0.191727752719)*x_ref[2]**(o-1)+(-0.748596484262)
        ref[3,1,0,0]=o*(-0.0610807571905)*x_ref[0]**(o-1)+(0.044623781518)
        ref[3,1,0,1]=o*(0.851226664753)*x_ref[1]**(o-1)+(0.921090244094)
        ref[3,1,0,2]=o*(0.494414571241)*x_ref[2]**(o-1)+(-0.473831333684)
        ref[3,1,1,0]=o*(0.54651807228)*x_ref[0]**(o-1)+(-0.580525990778)
        ref[3,1,1,1]=o*(-0.822936606301)*x_ref[1]**(o-1)+(-0.0169562878937)
        ref[3,1,1,2]=o*(-0.732219249555)*x_ref[2]**(o-1)+(-0.435156155008)
        ref[4,0,0,0]=o*(0.830021032139)*x_ref[0]**(o-1)+(0.976836986062)
        ref[4,0,0,1]=o*(0.534162405111)*x_ref[1]**(o-1)+(-0.286292103539)
        ref[4,0,0,2]=o*(0.820276894761)*x_ref[2]**(o-1)+(0.245897387706)
        ref[4,0,1,0]=o*(0.362201493311)*x_ref[0]**(o-1)+(-0.186328147108)
        ref[4,0,1,1]=o*(-0.91611458572)*x_ref[1]**(o-1)+(-0.965258904175)
        ref[4,0,1,2]=o*(0.61850292372)*x_ref[2]**(o-1)+(0.410229835478)
        ref[4,1,0,0]=o*(0.0726313013515)*x_ref[0]**(o-1)+(-0.974714544914)
        ref[4,1,0,1]=o*(0.46313543268)*x_ref[1]**(o-1)+(0.100486704072)
        ref[4,1,0,2]=o*(-0.149124857233)*x_ref[2]**(o-1)+(0.455733236337)
        ref[4,1,1,0]=o*(-0.875220220201)*x_ref[0]**(o-1)+(-0.416152301789)
        ref[4,1,1,1]=o*(-0.0145627466423)*x_ref[1]**(o-1)+(0.267856009119)
        ref[4,1,1,2]=o*(-0.604260174049)*x_ref[2]**(o-1)+(0.689107997051)
        ref[5,0,0,0]=o*(0.196182127971)*x_ref[0]**(o-1)+(0.337209503833)
        ref[5,0,0,1]=o*(0.595342239158)*x_ref[1]**(o-1)+(0.507423300857)
        ref[5,0,0,2]=o*(0.934760469389)*x_ref[2]**(o-1)+(-0.198686443328)
        ref[5,0,1,0]=o*(0.77210909967)*x_ref[0]**(o-1)+(-0.310778136545)
        ref[5,0,1,1]=o*(0.512326103142)*x_ref[1]**(o-1)+(-0.501056206243)
        ref[5,0,1,2]=o*(0.98755961257)*x_ref[2]**(o-1)+(-0.658414340038)
        ref[5,1,0,0]=o*(0.375277702198)*x_ref[0]**(o-1)+(-0.603850047717)
        ref[5,1,0,1]=o*(0.677227868069)*x_ref[1]**(o-1)+(-0.784640615159)
        ref[5,1,0,2]=o*(0.177350721639)*x_ref[2]**(o-1)+(0.241053657333)
        ref[5,1,1,0]=o*(-0.368625522736)*x_ref[0]**(o-1)+(0.196977476909)
        ref[5,1,1,1]=o*(-0.582451887206)*x_ref[1]**(o-1)+(0.598240904158)
        ref[5,1,1,2]=o*(0.102240230796)*x_ref[2]**(o-1)+(0.487322529158)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.851643614423)*x[0]**o+(-0.77638754265)*x[0]+(-0.216746360457)*x[1]**o+(-0.117807364973)*x[1]
        ref[0]=o*(0.851643614423)*x_ref[0]**(o-1)+(-0.77638754265)
        ref[1]=o*(-0.216746360457)*x_ref[1]**(o-1)+(-0.117807364973)
      else:
        arg=(0.846271449213)*x[0]**o+(0.895578416904)*x[0]+(-0.518297490388)*x[1]**o+(0.272255265951)*x[1]+(0.537822992528)*x[2]**o+(0.74226977002)*x[2]
        ref[0]=o*(0.846271449213)*x_ref[0]**(o-1)+(0.895578416904)
        ref[1]=o*(-0.518297490388)*x_ref[1]**(o-1)+(0.272255265951)
        ref[2]=o*(0.537822992528)*x_ref[2]**(o-1)+(0.74226977002)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.365601318395)*x[0]**o+(0.295624555802)*x[0]+(0.540618182217)*x[1]**o+(0.0458416078386)*x[1]
        arg[1]=(0.332276039998)*x[0]**o+(0.422606936253)*x[0]+(-0.221603594854)*x[1]**o+(0.879863420258)*x[1]
        ref[0,0]=o*(0.365601318395)*x_ref[0]**(o-1)+(0.295624555802)
        ref[0,1]=o*(0.540618182217)*x_ref[1]**(o-1)+(0.0458416078386)
        ref[1,0]=o*(0.332276039998)*x_ref[0]**(o-1)+(0.422606936253)
        ref[1,1]=o*(-0.221603594854)*x_ref[1]**(o-1)+(0.879863420258)
      else:
        arg[0]=(-0.258072277129)*x[0]**o+(-0.833085462518)*x[0]+(-0.124240965858)*x[1]**o+(-0.121853073047)*x[1]+(-0.877974471569)*x[2]**o+(0.929660451331)*x[2]
        arg[1]=(0.306520203535)*x[0]**o+(0.635444320723)*x[0]+(-0.766579509915)*x[1]**o+(-0.381189007175)*x[1]+(-0.76608952351)*x[2]**o+(0.00718583364881)*x[2]
        ref[0,0]=o*(-0.258072277129)*x_ref[0]**(o-1)+(-0.833085462518)
        ref[0,1]=o*(-0.124240965858)*x_ref[1]**(o-1)+(-0.121853073047)
        ref[0,2]=o*(-0.877974471569)*x_ref[2]**(o-1)+(0.929660451331)
        ref[1,0]=o*(0.306520203535)*x_ref[0]**(o-1)+(0.635444320723)
        ref[1,1]=o*(-0.766579509915)*x_ref[1]**(o-1)+(-0.381189007175)
        ref[1,2]=o*(-0.76608952351)*x_ref[2]**(o-1)+(0.00718583364881)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.403088506806)*x[0]**o+(-0.436942773988)*x[0]+(-0.0594011537346)*x[1]**o+(0.855063195595)*x[1]
        arg[0,1]=(-0.516992303892)*x[0]**o+(-0.132586737299)*x[0]+(-0.854639239618)*x[1]**o+(0.241200454455)*x[1]
        arg[0,2]=(-0.373941155498)*x[0]**o+(0.933328225549)*x[0]+(-0.986870460316)*x[1]**o+(-0.051265908773)*x[1]
        arg[0,3]=(-0.175701653328)*x[0]**o+(-0.726542005356)*x[0]+(-0.553505635409)*x[1]**o+(-0.199806969718)*x[1]
        arg[0,4]=(-0.750832127973)*x[0]**o+(0.404511387161)*x[0]+(0.874423404897)*x[1]**o+(0.35360129734)*x[1]
        arg[1,0]=(-0.000191271001285)*x[0]**o+(-0.319365399066)*x[0]+(0.398444976089)*x[1]**o+(-0.966391658006)*x[1]
        arg[1,1]=(-0.58562890965)*x[0]**o+(-0.623992315262)*x[0]+(-0.913704774126)*x[1]**o+(0.910305409081)*x[1]
        arg[1,2]=(-0.480491011401)*x[0]**o+(-0.387015926056)*x[0]+(0.845826744282)*x[1]**o+(0.380626980034)*x[1]
        arg[1,3]=(-0.284167706438)*x[0]**o+(0.730574128045)*x[0]+(0.823389953051)*x[1]**o+(0.59483032515)*x[1]
        arg[1,4]=(-0.545147810354)*x[0]**o+(-0.632923743497)*x[0]+(0.619596244096)*x[1]**o+(-0.816596365827)*x[1]
        arg[2,0]=(0.161289895106)*x[0]**o+(0.435472177094)*x[0]+(0.125591455076)*x[1]**o+(0.0207320895937)*x[1]
        arg[2,1]=(0.639457164801)*x[0]**o+(0.2631408826)*x[0]+(0.00255095677899)*x[1]**o+(0.732164720004)*x[1]
        arg[2,2]=(-0.152014092979)*x[0]**o+(0.347024615036)*x[0]+(0.94212294017)*x[1]**o+(0.173939724203)*x[1]
        arg[2,3]=(-0.0324872553975)*x[0]**o+(-0.969492827039)*x[0]+(-0.830876162378)*x[1]**o+(0.601308513636)*x[1]
        arg[2,4]=(0.932999309515)*x[0]**o+(0.0320937041249)*x[0]+(-0.354492894455)*x[1]**o+(-0.752837146132)*x[1]
        arg[3,0]=(0.995147840382)*x[0]**o+(-0.276974278934)*x[0]+(0.905272259272)*x[1]**o+(-0.433178159448)*x[1]
        arg[3,1]=(-0.291708633096)*x[0]**o+(-0.542000555721)*x[0]+(0.372404748364)*x[1]**o+(0.23140215774)*x[1]
        arg[3,2]=(0.88604771071)*x[0]**o+(-0.753078356803)*x[0]+(-0.265302521749)*x[1]**o+(-0.948898850214)*x[1]
        arg[3,3]=(-0.429496844082)*x[0]**o+(0.0731695917195)*x[0]+(0.291387878783)*x[1]**o+(-0.078863825039)*x[1]
        arg[3,4]=(-0.0067999635984)*x[0]**o+(0.488018496019)*x[0]+(-0.875646054486)*x[1]**o+(0.629803284064)*x[1]
        ref[0,0,0]=o*(0.403088506806)*x_ref[0]**(o-1)+(-0.436942773988)
        ref[0,0,1]=o*(-0.0594011537346)*x_ref[1]**(o-1)+(0.855063195595)
        ref[0,1,0]=o*(-0.516992303892)*x_ref[0]**(o-1)+(-0.132586737299)
        ref[0,1,1]=o*(-0.854639239618)*x_ref[1]**(o-1)+(0.241200454455)
        ref[0,2,0]=o*(-0.373941155498)*x_ref[0]**(o-1)+(0.933328225549)
        ref[0,2,1]=o*(-0.986870460316)*x_ref[1]**(o-1)+(-0.051265908773)
        ref[0,3,0]=o*(-0.175701653328)*x_ref[0]**(o-1)+(-0.726542005356)
        ref[0,3,1]=o*(-0.553505635409)*x_ref[1]**(o-1)+(-0.199806969718)
        ref[0,4,0]=o*(-0.750832127973)*x_ref[0]**(o-1)+(0.404511387161)
        ref[0,4,1]=o*(0.874423404897)*x_ref[1]**(o-1)+(0.35360129734)
        ref[1,0,0]=o*(-0.000191271001285)*x_ref[0]**(o-1)+(-0.319365399066)
        ref[1,0,1]=o*(0.398444976089)*x_ref[1]**(o-1)+(-0.966391658006)
        ref[1,1,0]=o*(-0.58562890965)*x_ref[0]**(o-1)+(-0.623992315262)
        ref[1,1,1]=o*(-0.913704774126)*x_ref[1]**(o-1)+(0.910305409081)
        ref[1,2,0]=o*(-0.480491011401)*x_ref[0]**(o-1)+(-0.387015926056)
        ref[1,2,1]=o*(0.845826744282)*x_ref[1]**(o-1)+(0.380626980034)
        ref[1,3,0]=o*(-0.284167706438)*x_ref[0]**(o-1)+(0.730574128045)
        ref[1,3,1]=o*(0.823389953051)*x_ref[1]**(o-1)+(0.59483032515)
        ref[1,4,0]=o*(-0.545147810354)*x_ref[0]**(o-1)+(-0.632923743497)
        ref[1,4,1]=o*(0.619596244096)*x_ref[1]**(o-1)+(-0.816596365827)
        ref[2,0,0]=o*(0.161289895106)*x_ref[0]**(o-1)+(0.435472177094)
        ref[2,0,1]=o*(0.125591455076)*x_ref[1]**(o-1)+(0.0207320895937)
        ref[2,1,0]=o*(0.639457164801)*x_ref[0]**(o-1)+(0.2631408826)
        ref[2,1,1]=o*(0.00255095677899)*x_ref[1]**(o-1)+(0.732164720004)
        ref[2,2,0]=o*(-0.152014092979)*x_ref[0]**(o-1)+(0.347024615036)
        ref[2,2,1]=o*(0.94212294017)*x_ref[1]**(o-1)+(0.173939724203)
        ref[2,3,0]=o*(-0.0324872553975)*x_ref[0]**(o-1)+(-0.969492827039)
        ref[2,3,1]=o*(-0.830876162378)*x_ref[1]**(o-1)+(0.601308513636)
        ref[2,4,0]=o*(0.932999309515)*x_ref[0]**(o-1)+(0.0320937041249)
        ref[2,4,1]=o*(-0.354492894455)*x_ref[1]**(o-1)+(-0.752837146132)
        ref[3,0,0]=o*(0.995147840382)*x_ref[0]**(o-1)+(-0.276974278934)
        ref[3,0,1]=o*(0.905272259272)*x_ref[1]**(o-1)+(-0.433178159448)
        ref[3,1,0]=o*(-0.291708633096)*x_ref[0]**(o-1)+(-0.542000555721)
        ref[3,1,1]=o*(0.372404748364)*x_ref[1]**(o-1)+(0.23140215774)
        ref[3,2,0]=o*(0.88604771071)*x_ref[0]**(o-1)+(-0.753078356803)
        ref[3,2,1]=o*(-0.265302521749)*x_ref[1]**(o-1)+(-0.948898850214)
        ref[3,3,0]=o*(-0.429496844082)*x_ref[0]**(o-1)+(0.0731695917195)
        ref[3,3,1]=o*(0.291387878783)*x_ref[1]**(o-1)+(-0.078863825039)
        ref[3,4,0]=o*(-0.0067999635984)*x_ref[0]**(o-1)+(0.488018496019)
        ref[3,4,1]=o*(-0.875646054486)*x_ref[1]**(o-1)+(0.629803284064)
      else:
        arg[0,0]=(-0.512668741406)*x[0]**o+(-0.73142104842)*x[0]+(0.489050993226)*x[1]**o+(0.771941571419)*x[1]+(-0.538092512577)*x[2]**o+(-0.457812389328)*x[2]
        arg[0,1]=(0.258566436386)*x[0]**o+(-0.572006860118)*x[0]+(-0.0370305840623)*x[1]**o+(0.328810934717)*x[1]+(-0.652873123265)*x[2]**o+(0.734929914374)*x[2]
        arg[0,2]=(-0.228462027625)*x[0]**o+(-0.873359484139)*x[0]+(0.902575701641)*x[1]**o+(0.428758160478)*x[1]+(-0.377042847702)*x[2]**o+(-0.161364604019)*x[2]
        arg[0,3]=(0.31338185054)*x[0]**o+(0.973849539932)*x[0]+(-0.403569231167)*x[1]**o+(-0.110026269257)*x[1]+(0.603101392345)*x[2]**o+(0.499232034173)*x[2]
        arg[0,4]=(0.468416317974)*x[0]**o+(-0.0154203909264)*x[0]+(0.409228888268)*x[1]**o+(0.358886480877)*x[1]+(-0.00677343401301)*x[2]**o+(-0.0567153195302)*x[2]
        arg[1,0]=(0.614929795628)*x[0]**o+(0.54911679279)*x[0]+(0.527138394555)*x[1]**o+(-0.84070465171)*x[1]+(0.421928294706)*x[2]**o+(-0.871734463848)*x[2]
        arg[1,1]=(-0.498770327357)*x[0]**o+(0.294610855431)*x[0]+(-0.815129074288)*x[1]**o+(0.395332176796)*x[1]+(-0.555947305527)*x[2]**o+(-0.623550814988)*x[2]
        arg[1,2]=(-0.369404109593)*x[0]**o+(-0.662927702362)*x[0]+(0.886789099892)*x[1]**o+(-0.815728539359)*x[1]+(-0.843861162574)*x[2]**o+(-0.510039607856)*x[2]
        arg[1,3]=(-0.747305678574)*x[0]**o+(0.350337618157)*x[0]+(0.313491051893)*x[1]**o+(-0.554840960712)*x[1]+(-0.259760094369)*x[2]**o+(0.904246223768)*x[2]
        arg[1,4]=(0.485859115418)*x[0]**o+(-0.751900240375)*x[0]+(-0.233384727142)*x[1]**o+(-0.466677891749)*x[1]+(-0.771311310708)*x[2]**o+(0.380515497846)*x[2]
        arg[2,0]=(0.628529887016)*x[0]**o+(0.778810609149)*x[0]+(-0.95647127374)*x[1]**o+(-0.887895309984)*x[1]+(-0.926017554164)*x[2]**o+(0.0922106889932)*x[2]
        arg[2,1]=(-0.981382750453)*x[0]**o+(0.858805367823)*x[0]+(-0.202352133378)*x[1]**o+(0.527229765084)*x[1]+(-0.0862956097426)*x[2]**o+(-0.76269735579)*x[2]
        arg[2,2]=(0.591276459941)*x[0]**o+(0.554086306469)*x[0]+(-0.175045397917)*x[1]**o+(0.319542375346)*x[1]+(0.328810955576)*x[2]**o+(-0.245381809232)*x[2]
        arg[2,3]=(-0.826436985287)*x[0]**o+(-0.931352535769)*x[0]+(-0.480801956568)*x[1]**o+(0.830773509938)*x[1]+(0.157556801982)*x[2]**o+(0.982505166946)*x[2]
        arg[2,4]=(-0.0992736644463)*x[0]**o+(-0.147525710611)*x[0]+(-0.723322157973)*x[1]**o+(0.0527438662093)*x[1]+(-0.338847227522)*x[2]**o+(-0.704585609897)*x[2]
        arg[3,0]=(-0.45250460769)*x[0]**o+(-0.323299292565)*x[0]+(-0.435540041421)*x[1]**o+(0.418330181239)*x[1]+(0.526739853977)*x[2]**o+(-0.00513414974056)*x[2]
        arg[3,1]=(0.289534600598)*x[0]**o+(0.467498987256)*x[0]+(0.27851903721)*x[1]**o+(0.908249187588)*x[1]+(-0.84695489717)*x[2]**o+(-0.799372710877)*x[2]
        arg[3,2]=(0.00338011092566)*x[0]**o+(0.819216072868)*x[0]+(0.115437140937)*x[1]**o+(0.0968835984948)*x[1]+(-0.633253222581)*x[2]**o+(-0.0224890801282)*x[2]
        arg[3,3]=(0.578331574145)*x[0]**o+(0.328247456626)*x[0]+(-0.725914955131)*x[1]**o+(-0.863562766028)*x[1]+(-0.283057959979)*x[2]**o+(-0.521563360583)*x[2]
        arg[3,4]=(0.345725352698)*x[0]**o+(0.614912618718)*x[0]+(-0.218084952915)*x[1]**o+(-0.629514485943)*x[1]+(-0.207605980168)*x[2]**o+(-0.0950042458608)*x[2]
        ref[0,0,0]=o*(-0.512668741406)*x_ref[0]**(o-1)+(-0.73142104842)
        ref[0,0,1]=o*(0.489050993226)*x_ref[1]**(o-1)+(0.771941571419)
        ref[0,0,2]=o*(-0.538092512577)*x_ref[2]**(o-1)+(-0.457812389328)
        ref[0,1,0]=o*(0.258566436386)*x_ref[0]**(o-1)+(-0.572006860118)
        ref[0,1,1]=o*(-0.0370305840623)*x_ref[1]**(o-1)+(0.328810934717)
        ref[0,1,2]=o*(-0.652873123265)*x_ref[2]**(o-1)+(0.734929914374)
        ref[0,2,0]=o*(-0.228462027625)*x_ref[0]**(o-1)+(-0.873359484139)
        ref[0,2,1]=o*(0.902575701641)*x_ref[1]**(o-1)+(0.428758160478)
        ref[0,2,2]=o*(-0.377042847702)*x_ref[2]**(o-1)+(-0.161364604019)
        ref[0,3,0]=o*(0.31338185054)*x_ref[0]**(o-1)+(0.973849539932)
        ref[0,3,1]=o*(-0.403569231167)*x_ref[1]**(o-1)+(-0.110026269257)
        ref[0,3,2]=o*(0.603101392345)*x_ref[2]**(o-1)+(0.499232034173)
        ref[0,4,0]=o*(0.468416317974)*x_ref[0]**(o-1)+(-0.0154203909264)
        ref[0,4,1]=o*(0.409228888268)*x_ref[1]**(o-1)+(0.358886480877)
        ref[0,4,2]=o*(-0.00677343401301)*x_ref[2]**(o-1)+(-0.0567153195302)
        ref[1,0,0]=o*(0.614929795628)*x_ref[0]**(o-1)+(0.54911679279)
        ref[1,0,1]=o*(0.527138394555)*x_ref[1]**(o-1)+(-0.84070465171)
        ref[1,0,2]=o*(0.421928294706)*x_ref[2]**(o-1)+(-0.871734463848)
        ref[1,1,0]=o*(-0.498770327357)*x_ref[0]**(o-1)+(0.294610855431)
        ref[1,1,1]=o*(-0.815129074288)*x_ref[1]**(o-1)+(0.395332176796)
        ref[1,1,2]=o*(-0.555947305527)*x_ref[2]**(o-1)+(-0.623550814988)
        ref[1,2,0]=o*(-0.369404109593)*x_ref[0]**(o-1)+(-0.662927702362)
        ref[1,2,1]=o*(0.886789099892)*x_ref[1]**(o-1)+(-0.815728539359)
        ref[1,2,2]=o*(-0.843861162574)*x_ref[2]**(o-1)+(-0.510039607856)
        ref[1,3,0]=o*(-0.747305678574)*x_ref[0]**(o-1)+(0.350337618157)
        ref[1,3,1]=o*(0.313491051893)*x_ref[1]**(o-1)+(-0.554840960712)
        ref[1,3,2]=o*(-0.259760094369)*x_ref[2]**(o-1)+(0.904246223768)
        ref[1,4,0]=o*(0.485859115418)*x_ref[0]**(o-1)+(-0.751900240375)
        ref[1,4,1]=o*(-0.233384727142)*x_ref[1]**(o-1)+(-0.466677891749)
        ref[1,4,2]=o*(-0.771311310708)*x_ref[2]**(o-1)+(0.380515497846)
        ref[2,0,0]=o*(0.628529887016)*x_ref[0]**(o-1)+(0.778810609149)
        ref[2,0,1]=o*(-0.95647127374)*x_ref[1]**(o-1)+(-0.887895309984)
        ref[2,0,2]=o*(-0.926017554164)*x_ref[2]**(o-1)+(0.0922106889932)
        ref[2,1,0]=o*(-0.981382750453)*x_ref[0]**(o-1)+(0.858805367823)
        ref[2,1,1]=o*(-0.202352133378)*x_ref[1]**(o-1)+(0.527229765084)
        ref[2,1,2]=o*(-0.0862956097426)*x_ref[2]**(o-1)+(-0.76269735579)
        ref[2,2,0]=o*(0.591276459941)*x_ref[0]**(o-1)+(0.554086306469)
        ref[2,2,1]=o*(-0.175045397917)*x_ref[1]**(o-1)+(0.319542375346)
        ref[2,2,2]=o*(0.328810955576)*x_ref[2]**(o-1)+(-0.245381809232)
        ref[2,3,0]=o*(-0.826436985287)*x_ref[0]**(o-1)+(-0.931352535769)
        ref[2,3,1]=o*(-0.480801956568)*x_ref[1]**(o-1)+(0.830773509938)
        ref[2,3,2]=o*(0.157556801982)*x_ref[2]**(o-1)+(0.982505166946)
        ref[2,4,0]=o*(-0.0992736644463)*x_ref[0]**(o-1)+(-0.147525710611)
        ref[2,4,1]=o*(-0.723322157973)*x_ref[1]**(o-1)+(0.0527438662093)
        ref[2,4,2]=o*(-0.338847227522)*x_ref[2]**(o-1)+(-0.704585609897)
        ref[3,0,0]=o*(-0.45250460769)*x_ref[0]**(o-1)+(-0.323299292565)
        ref[3,0,1]=o*(-0.435540041421)*x_ref[1]**(o-1)+(0.418330181239)
        ref[3,0,2]=o*(0.526739853977)*x_ref[2]**(o-1)+(-0.00513414974056)
        ref[3,1,0]=o*(0.289534600598)*x_ref[0]**(o-1)+(0.467498987256)
        ref[3,1,1]=o*(0.27851903721)*x_ref[1]**(o-1)+(0.908249187588)
        ref[3,1,2]=o*(-0.84695489717)*x_ref[2]**(o-1)+(-0.799372710877)
        ref[3,2,0]=o*(0.00338011092566)*x_ref[0]**(o-1)+(0.819216072868)
        ref[3,2,1]=o*(0.115437140937)*x_ref[1]**(o-1)+(0.0968835984948)
        ref[3,2,2]=o*(-0.633253222581)*x_ref[2]**(o-1)+(-0.0224890801282)
        ref[3,3,0]=o*(0.578331574145)*x_ref[0]**(o-1)+(0.328247456626)
        ref[3,3,1]=o*(-0.725914955131)*x_ref[1]**(o-1)+(-0.863562766028)
        ref[3,3,2]=o*(-0.283057959979)*x_ref[2]**(o-1)+(-0.521563360583)
        ref[3,4,0]=o*(0.345725352698)*x_ref[0]**(o-1)+(0.614912618718)
        ref[3,4,1]=o*(-0.218084952915)*x_ref[1]**(o-1)+(-0.629514485943)
        ref[3,4,2]=o*(-0.207605980168)*x_ref[2]**(o-1)+(-0.0950042458608)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_Solution_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.353390112117)*x[0]**o+(0.121267964347)*x[0]+(0.095925646757)*x[1]**o+(-0.00141503525661)*x[1]
        arg[0,0,1]=(0.610556742471)*x[0]**o+(0.542948343923)*x[0]+(0.124762431991)*x[1]**o+(0.00216142978003)*x[1]
        arg[0,1,0]=(0.261559018753)*x[0]**o+(-0.881689419767)*x[0]+(-0.009352418091)*x[1]**o+(0.333357441723)*x[1]
        arg[0,1,1]=(-0.25379927884)*x[0]**o+(-0.124076833523)*x[0]+(-0.167690772835)*x[1]**o+(0.89804632709)*x[1]
        arg[1,0,0]=(-0.39353207502)*x[0]**o+(0.838815579522)*x[0]+(0.829467918805)*x[1]**o+(0.614403367394)*x[1]
        arg[1,0,1]=(-0.698946975112)*x[0]**o+(-0.551313390457)*x[0]+(0.965127390549)*x[1]**o+(-0.819660095397)*x[1]
        arg[1,1,0]=(-0.656237793952)*x[0]**o+(0.727261124295)*x[0]+(0.602446266985)*x[1]**o+(-0.320796826351)*x[1]
        arg[1,1,1]=(0.111875111713)*x[0]**o+(0.000360626785094)*x[0]+(-0.802516890555)*x[1]**o+(0.420759758492)*x[1]
        arg[2,0,0]=(-0.421920068352)*x[0]**o+(-0.022970702876)*x[0]+(-0.0257219638216)*x[1]**o+(-0.36544576073)*x[1]
        arg[2,0,1]=(-0.459431228769)*x[0]**o+(0.150702477357)*x[0]+(-0.869898945031)*x[1]**o+(0.261651620176)*x[1]
        arg[2,1,0]=(-0.866165026127)*x[0]**o+(0.149307656472)*x[0]+(-0.270058622191)*x[1]**o+(0.0873163553474)*x[1]
        arg[2,1,1]=(-0.871128617394)*x[0]**o+(-0.774356000237)*x[0]+(0.337383167457)*x[1]**o+(-0.453335744484)*x[1]
        arg[3,0,0]=(0.404195477481)*x[0]**o+(-0.350041445198)*x[0]+(-0.583794909281)*x[1]**o+(-0.895439618872)*x[1]
        arg[3,0,1]=(0.693567867164)*x[0]**o+(0.989642722856)*x[0]+(0.814104467522)*x[1]**o+(-0.126302588221)*x[1]
        arg[3,1,0]=(0.935509747673)*x[0]**o+(0.0104612091188)*x[0]+(0.304974305936)*x[1]**o+(0.658660310426)*x[1]
        arg[3,1,1]=(0.677420202064)*x[0]**o+(-0.160834685914)*x[0]+(0.241578146196)*x[1]**o+(-0.182223867353)*x[1]
        arg[4,0,0]=(-0.908542015864)*x[0]**o+(0.125595344482)*x[0]+(-0.396864254557)*x[1]**o+(0.408182849245)*x[1]
        arg[4,0,1]=(0.436101221257)*x[0]**o+(0.946667075656)*x[0]+(-0.293523459017)*x[1]**o+(0.363130416513)*x[1]
        arg[4,1,0]=(0.583416010605)*x[0]**o+(0.531883961319)*x[0]+(-0.418905217485)*x[1]**o+(0.271442034667)*x[1]
        arg[4,1,1]=(-0.708766397827)*x[0]**o+(0.0266251980073)*x[0]+(-0.0929293524384)*x[1]**o+(-0.661095295605)*x[1]
        arg[5,0,0]=(0.693424423443)*x[0]**o+(-0.201539674621)*x[0]+(-0.221922493999)*x[1]**o+(-0.218359043699)*x[1]
        arg[5,0,1]=(-0.478707796293)*x[0]**o+(0.791533578559)*x[0]+(0.280213321002)*x[1]**o+(-0.60809908629)*x[1]
        arg[5,1,0]=(-0.517992572451)*x[0]**o+(0.591995363281)*x[0]+(0.480220427965)*x[1]**o+(0.134542063762)*x[1]
        arg[5,1,1]=(0.415861674391)*x[0]**o+(0.773614372731)*x[0]+(-0.984515861984)*x[1]**o+(0.971916685266)*x[1]
        ref[0,0,0,0]=o*(-0.353390112117)*x_ref[0]**(o-1)+(0.121267964347)
        ref[0,0,0,1]=o*(0.095925646757)*x_ref[1]**(o-1)+(-0.00141503525661)
        ref[0,0,1,0]=o*(0.610556742471)*x_ref[0]**(o-1)+(0.542948343923)
        ref[0,0,1,1]=o*(0.124762431991)*x_ref[1]**(o-1)+(0.00216142978003)
        ref[0,1,0,0]=o*(0.261559018753)*x_ref[0]**(o-1)+(-0.881689419767)
        ref[0,1,0,1]=o*(-0.009352418091)*x_ref[1]**(o-1)+(0.333357441723)
        ref[0,1,1,0]=o*(-0.25379927884)*x_ref[0]**(o-1)+(-0.124076833523)
        ref[0,1,1,1]=o*(-0.167690772835)*x_ref[1]**(o-1)+(0.89804632709)
        ref[1,0,0,0]=o*(-0.39353207502)*x_ref[0]**(o-1)+(0.838815579522)
        ref[1,0,0,1]=o*(0.829467918805)*x_ref[1]**(o-1)+(0.614403367394)
        ref[1,0,1,0]=o*(-0.698946975112)*x_ref[0]**(o-1)+(-0.551313390457)
        ref[1,0,1,1]=o*(0.965127390549)*x_ref[1]**(o-1)+(-0.819660095397)
        ref[1,1,0,0]=o*(-0.656237793952)*x_ref[0]**(o-1)+(0.727261124295)
        ref[1,1,0,1]=o*(0.602446266985)*x_ref[1]**(o-1)+(-0.320796826351)
        ref[1,1,1,0]=o*(0.111875111713)*x_ref[0]**(o-1)+(0.000360626785094)
        ref[1,1,1,1]=o*(-0.802516890555)*x_ref[1]**(o-1)+(0.420759758492)
        ref[2,0,0,0]=o*(-0.421920068352)*x_ref[0]**(o-1)+(-0.022970702876)
        ref[2,0,0,1]=o*(-0.0257219638216)*x_ref[1]**(o-1)+(-0.36544576073)
        ref[2,0,1,0]=o*(-0.459431228769)*x_ref[0]**(o-1)+(0.150702477357)
        ref[2,0,1,1]=o*(-0.869898945031)*x_ref[1]**(o-1)+(0.261651620176)
        ref[2,1,0,0]=o*(-0.866165026127)*x_ref[0]**(o-1)+(0.149307656472)
        ref[2,1,0,1]=o*(-0.270058622191)*x_ref[1]**(o-1)+(0.0873163553474)
        ref[2,1,1,0]=o*(-0.871128617394)*x_ref[0]**(o-1)+(-0.774356000237)
        ref[2,1,1,1]=o*(0.337383167457)*x_ref[1]**(o-1)+(-0.453335744484)
        ref[3,0,0,0]=o*(0.404195477481)*x_ref[0]**(o-1)+(-0.350041445198)
        ref[3,0,0,1]=o*(-0.583794909281)*x_ref[1]**(o-1)+(-0.895439618872)
        ref[3,0,1,0]=o*(0.693567867164)*x_ref[0]**(o-1)+(0.989642722856)
        ref[3,0,1,1]=o*(0.814104467522)*x_ref[1]**(o-1)+(-0.126302588221)
        ref[3,1,0,0]=o*(0.935509747673)*x_ref[0]**(o-1)+(0.0104612091188)
        ref[3,1,0,1]=o*(0.304974305936)*x_ref[1]**(o-1)+(0.658660310426)
        ref[3,1,1,0]=o*(0.677420202064)*x_ref[0]**(o-1)+(-0.160834685914)
        ref[3,1,1,1]=o*(0.241578146196)*x_ref[1]**(o-1)+(-0.182223867353)
        ref[4,0,0,0]=o*(-0.908542015864)*x_ref[0]**(o-1)+(0.125595344482)
        ref[4,0,0,1]=o*(-0.396864254557)*x_ref[1]**(o-1)+(0.408182849245)
        ref[4,0,1,0]=o*(0.436101221257)*x_ref[0]**(o-1)+(0.946667075656)
        ref[4,0,1,1]=o*(-0.293523459017)*x_ref[1]**(o-1)+(0.363130416513)
        ref[4,1,0,0]=o*(0.583416010605)*x_ref[0]**(o-1)+(0.531883961319)
        ref[4,1,0,1]=o*(-0.418905217485)*x_ref[1]**(o-1)+(0.271442034667)
        ref[4,1,1,0]=o*(-0.708766397827)*x_ref[0]**(o-1)+(0.0266251980073)
        ref[4,1,1,1]=o*(-0.0929293524384)*x_ref[1]**(o-1)+(-0.661095295605)
        ref[5,0,0,0]=o*(0.693424423443)*x_ref[0]**(o-1)+(-0.201539674621)
        ref[5,0,0,1]=o*(-0.221922493999)*x_ref[1]**(o-1)+(-0.218359043699)
        ref[5,0,1,0]=o*(-0.478707796293)*x_ref[0]**(o-1)+(0.791533578559)
        ref[5,0,1,1]=o*(0.280213321002)*x_ref[1]**(o-1)+(-0.60809908629)
        ref[5,1,0,0]=o*(-0.517992572451)*x_ref[0]**(o-1)+(0.591995363281)
        ref[5,1,0,1]=o*(0.480220427965)*x_ref[1]**(o-1)+(0.134542063762)
        ref[5,1,1,0]=o*(0.415861674391)*x_ref[0]**(o-1)+(0.773614372731)
        ref[5,1,1,1]=o*(-0.984515861984)*x_ref[1]**(o-1)+(0.971916685266)
      else:
        arg[0,0,0]=(0.157678109652)*x[0]**o+(-0.939095069874)*x[0]+(0.340072659801)*x[1]**o+(0.806723712633)*x[1]+(0.0175654201162)*x[2]**o+(-0.695887679405)*x[2]
        arg[0,0,1]=(0.148444957808)*x[0]**o+(0.769780554196)*x[0]+(-0.999778710821)*x[1]**o+(0.765998508134)*x[1]+(0.489498280892)*x[2]**o+(0.272011172233)*x[2]
        arg[0,1,0]=(-0.78853623669)*x[0]**o+(-0.889333747099)*x[0]+(0.897684283327)*x[1]**o+(0.878046029595)*x[1]+(-0.355566201991)*x[2]**o+(-0.656187599294)*x[2]
        arg[0,1,1]=(-0.903496685895)*x[0]**o+(0.335762730702)*x[0]+(0.264330545204)*x[1]**o+(-0.369300150725)*x[1]+(0.00819339769162)*x[2]**o+(0.0466274018566)*x[2]
        arg[1,0,0]=(0.490893502666)*x[0]**o+(0.853253169019)*x[0]+(-0.706120536551)*x[1]**o+(-0.810665092749)*x[1]+(0.877124690496)*x[2]**o+(-0.663316848595)*x[2]
        arg[1,0,1]=(-0.897301110118)*x[0]**o+(-0.571577796347)*x[0]+(-0.00912216314091)*x[1]**o+(0.479091266837)*x[1]+(-0.951994896347)*x[2]**o+(0.293801256264)*x[2]
        arg[1,1,0]=(-0.0843442242335)*x[0]**o+(0.436592099426)*x[0]+(-0.961835284742)*x[1]**o+(0.453758903497)*x[1]+(0.759916173417)*x[2]**o+(-0.356674191439)*x[2]
        arg[1,1,1]=(0.368337095889)*x[0]**o+(-0.180552461333)*x[0]+(-0.814727889141)*x[1]**o+(-0.226678543654)*x[1]+(0.629233877979)*x[2]**o+(0.934452743806)*x[2]
        arg[2,0,0]=(0.335801632975)*x[0]**o+(0.4853793666)*x[0]+(-0.603075878381)*x[1]**o+(-0.464873655094)*x[1]+(-0.304107114487)*x[2]**o+(0.768564020927)*x[2]
        arg[2,0,1]=(-0.843867260693)*x[0]**o+(-0.758493294188)*x[0]+(0.801820428023)*x[1]**o+(0.0909753752871)*x[1]+(-0.208134327922)*x[2]**o+(0.343325582627)*x[2]
        arg[2,1,0]=(-0.143364423799)*x[0]**o+(-0.191201452642)*x[0]+(0.356480498629)*x[1]**o+(0.0853546566948)*x[1]+(0.234781870759)*x[2]**o+(-0.702906217789)*x[2]
        arg[2,1,1]=(0.81324792898)*x[0]**o+(0.565754643122)*x[0]+(-0.0246314241495)*x[1]**o+(-0.83548496516)*x[1]+(0.730246143504)*x[2]**o+(-0.835539994852)*x[2]
        arg[3,0,0]=(0.877992836393)*x[0]**o+(-0.137135423487)*x[0]+(0.534511286076)*x[1]**o+(-0.0584370047297)*x[1]+(-0.0497236607574)*x[2]**o+(0.408414670632)*x[2]
        arg[3,0,1]=(-0.959771372305)*x[0]**o+(0.587216149992)*x[0]+(-0.873216249139)*x[1]**o+(0.0703350087653)*x[1]+(0.452900901257)*x[2]**o+(-0.180788315472)*x[2]
        arg[3,1,0]=(0.477701526969)*x[0]**o+(0.16165566827)*x[0]+(0.219901428845)*x[1]**o+(-0.0816687269195)*x[1]+(-0.379553348394)*x[2]**o+(-0.764672519612)*x[2]
        arg[3,1,1]=(0.274518799686)*x[0]**o+(0.945009722336)*x[0]+(-0.979787014238)*x[1]**o+(0.587704213162)*x[1]+(0.7959640984)*x[2]**o+(-0.0811112488533)*x[2]
        arg[4,0,0]=(-0.579121056378)*x[0]**o+(0.971799195206)*x[0]+(-0.646271217985)*x[1]**o+(0.369751123548)*x[1]+(0.792425864156)*x[2]**o+(-0.170799235786)*x[2]
        arg[4,0,1]=(0.0939380562355)*x[0]**o+(0.857041359029)*x[0]+(-0.379025882204)*x[1]**o+(-0.571968622949)*x[1]+(-0.156853183697)*x[2]**o+(0.643214485063)*x[2]
        arg[4,1,0]=(-0.715210364871)*x[0]**o+(0.150341666571)*x[0]+(0.325165277323)*x[1]**o+(-0.322744240034)*x[1]+(-0.622094139461)*x[2]**o+(-0.830479146327)*x[2]
        arg[4,1,1]=(-0.697109424406)*x[0]**o+(-0.730318259042)*x[0]+(-0.135119012822)*x[1]**o+(0.410502029229)*x[1]+(0.547987909743)*x[2]**o+(0.144877544254)*x[2]
        arg[5,0,0]=(0.787147597087)*x[0]**o+(-0.528180541295)*x[0]+(-0.251578089108)*x[1]**o+(-0.170766337864)*x[1]+(-0.123341129221)*x[2]**o+(-0.815622069347)*x[2]
        arg[5,0,1]=(0.193873484005)*x[0]**o+(0.270190252915)*x[0]+(0.807946985621)*x[1]**o+(0.184522999461)*x[1]+(0.97429597406)*x[2]**o+(0.532931028021)*x[2]
        arg[5,1,0]=(0.555644676015)*x[0]**o+(-0.225593442089)*x[0]+(0.877869009767)*x[1]**o+(-0.750894245506)*x[1]+(0.888501611997)*x[2]**o+(-0.84027021889)*x[2]
        arg[5,1,1]=(0.76909422867)*x[0]**o+(0.406697645173)*x[0]+(-0.233710263631)*x[1]**o+(-0.432942345121)*x[1]+(0.270484113666)*x[2]**o+(0.707302418115)*x[2]
        ref[0,0,0,0]=o*(0.157678109652)*x_ref[0]**(o-1)+(-0.939095069874)
        ref[0,0,0,1]=o*(0.340072659801)*x_ref[1]**(o-1)+(0.806723712633)
        ref[0,0,0,2]=o*(0.0175654201162)*x_ref[2]**(o-1)+(-0.695887679405)
        ref[0,0,1,0]=o*(0.148444957808)*x_ref[0]**(o-1)+(0.769780554196)
        ref[0,0,1,1]=o*(-0.999778710821)*x_ref[1]**(o-1)+(0.765998508134)
        ref[0,0,1,2]=o*(0.489498280892)*x_ref[2]**(o-1)+(0.272011172233)
        ref[0,1,0,0]=o*(-0.78853623669)*x_ref[0]**(o-1)+(-0.889333747099)
        ref[0,1,0,1]=o*(0.897684283327)*x_ref[1]**(o-1)+(0.878046029595)
        ref[0,1,0,2]=o*(-0.355566201991)*x_ref[2]**(o-1)+(-0.656187599294)
        ref[0,1,1,0]=o*(-0.903496685895)*x_ref[0]**(o-1)+(0.335762730702)
        ref[0,1,1,1]=o*(0.264330545204)*x_ref[1]**(o-1)+(-0.369300150725)
        ref[0,1,1,2]=o*(0.00819339769162)*x_ref[2]**(o-1)+(0.0466274018566)
        ref[1,0,0,0]=o*(0.490893502666)*x_ref[0]**(o-1)+(0.853253169019)
        ref[1,0,0,1]=o*(-0.706120536551)*x_ref[1]**(o-1)+(-0.810665092749)
        ref[1,0,0,2]=o*(0.877124690496)*x_ref[2]**(o-1)+(-0.663316848595)
        ref[1,0,1,0]=o*(-0.897301110118)*x_ref[0]**(o-1)+(-0.571577796347)
        ref[1,0,1,1]=o*(-0.00912216314091)*x_ref[1]**(o-1)+(0.479091266837)
        ref[1,0,1,2]=o*(-0.951994896347)*x_ref[2]**(o-1)+(0.293801256264)
        ref[1,1,0,0]=o*(-0.0843442242335)*x_ref[0]**(o-1)+(0.436592099426)
        ref[1,1,0,1]=o*(-0.961835284742)*x_ref[1]**(o-1)+(0.453758903497)
        ref[1,1,0,2]=o*(0.759916173417)*x_ref[2]**(o-1)+(-0.356674191439)
        ref[1,1,1,0]=o*(0.368337095889)*x_ref[0]**(o-1)+(-0.180552461333)
        ref[1,1,1,1]=o*(-0.814727889141)*x_ref[1]**(o-1)+(-0.226678543654)
        ref[1,1,1,2]=o*(0.629233877979)*x_ref[2]**(o-1)+(0.934452743806)
        ref[2,0,0,0]=o*(0.335801632975)*x_ref[0]**(o-1)+(0.4853793666)
        ref[2,0,0,1]=o*(-0.603075878381)*x_ref[1]**(o-1)+(-0.464873655094)
        ref[2,0,0,2]=o*(-0.304107114487)*x_ref[2]**(o-1)+(0.768564020927)
        ref[2,0,1,0]=o*(-0.843867260693)*x_ref[0]**(o-1)+(-0.758493294188)
        ref[2,0,1,1]=o*(0.801820428023)*x_ref[1]**(o-1)+(0.0909753752871)
        ref[2,0,1,2]=o*(-0.208134327922)*x_ref[2]**(o-1)+(0.343325582627)
        ref[2,1,0,0]=o*(-0.143364423799)*x_ref[0]**(o-1)+(-0.191201452642)
        ref[2,1,0,1]=o*(0.356480498629)*x_ref[1]**(o-1)+(0.0853546566948)
        ref[2,1,0,2]=o*(0.234781870759)*x_ref[2]**(o-1)+(-0.702906217789)
        ref[2,1,1,0]=o*(0.81324792898)*x_ref[0]**(o-1)+(0.565754643122)
        ref[2,1,1,1]=o*(-0.0246314241495)*x_ref[1]**(o-1)+(-0.83548496516)
        ref[2,1,1,2]=o*(0.730246143504)*x_ref[2]**(o-1)+(-0.835539994852)
        ref[3,0,0,0]=o*(0.877992836393)*x_ref[0]**(o-1)+(-0.137135423487)
        ref[3,0,0,1]=o*(0.534511286076)*x_ref[1]**(o-1)+(-0.0584370047297)
        ref[3,0,0,2]=o*(-0.0497236607574)*x_ref[2]**(o-1)+(0.408414670632)
        ref[3,0,1,0]=o*(-0.959771372305)*x_ref[0]**(o-1)+(0.587216149992)
        ref[3,0,1,1]=o*(-0.873216249139)*x_ref[1]**(o-1)+(0.0703350087653)
        ref[3,0,1,2]=o*(0.452900901257)*x_ref[2]**(o-1)+(-0.180788315472)
        ref[3,1,0,0]=o*(0.477701526969)*x_ref[0]**(o-1)+(0.16165566827)
        ref[3,1,0,1]=o*(0.219901428845)*x_ref[1]**(o-1)+(-0.0816687269195)
        ref[3,1,0,2]=o*(-0.379553348394)*x_ref[2]**(o-1)+(-0.764672519612)
        ref[3,1,1,0]=o*(0.274518799686)*x_ref[0]**(o-1)+(0.945009722336)
        ref[3,1,1,1]=o*(-0.979787014238)*x_ref[1]**(o-1)+(0.587704213162)
        ref[3,1,1,2]=o*(0.7959640984)*x_ref[2]**(o-1)+(-0.0811112488533)
        ref[4,0,0,0]=o*(-0.579121056378)*x_ref[0]**(o-1)+(0.971799195206)
        ref[4,0,0,1]=o*(-0.646271217985)*x_ref[1]**(o-1)+(0.369751123548)
        ref[4,0,0,2]=o*(0.792425864156)*x_ref[2]**(o-1)+(-0.170799235786)
        ref[4,0,1,0]=o*(0.0939380562355)*x_ref[0]**(o-1)+(0.857041359029)
        ref[4,0,1,1]=o*(-0.379025882204)*x_ref[1]**(o-1)+(-0.571968622949)
        ref[4,0,1,2]=o*(-0.156853183697)*x_ref[2]**(o-1)+(0.643214485063)
        ref[4,1,0,0]=o*(-0.715210364871)*x_ref[0]**(o-1)+(0.150341666571)
        ref[4,1,0,1]=o*(0.325165277323)*x_ref[1]**(o-1)+(-0.322744240034)
        ref[4,1,0,2]=o*(-0.622094139461)*x_ref[2]**(o-1)+(-0.830479146327)
        ref[4,1,1,0]=o*(-0.697109424406)*x_ref[0]**(o-1)+(-0.730318259042)
        ref[4,1,1,1]=o*(-0.135119012822)*x_ref[1]**(o-1)+(0.410502029229)
        ref[4,1,1,2]=o*(0.547987909743)*x_ref[2]**(o-1)+(0.144877544254)
        ref[5,0,0,0]=o*(0.787147597087)*x_ref[0]**(o-1)+(-0.528180541295)
        ref[5,0,0,1]=o*(-0.251578089108)*x_ref[1]**(o-1)+(-0.170766337864)
        ref[5,0,0,2]=o*(-0.123341129221)*x_ref[2]**(o-1)+(-0.815622069347)
        ref[5,0,1,0]=o*(0.193873484005)*x_ref[0]**(o-1)+(0.270190252915)
        ref[5,0,1,1]=o*(0.807946985621)*x_ref[1]**(o-1)+(0.184522999461)
        ref[5,0,1,2]=o*(0.97429597406)*x_ref[2]**(o-1)+(0.532931028021)
        ref[5,1,0,0]=o*(0.555644676015)*x_ref[0]**(o-1)+(-0.225593442089)
        ref[5,1,0,1]=o*(0.877869009767)*x_ref[1]**(o-1)+(-0.750894245506)
        ref[5,1,0,2]=o*(0.888501611997)*x_ref[2]**(o-1)+(-0.84027021889)
        ref[5,1,1,0]=o*(0.76909422867)*x_ref[0]**(o-1)+(0.406697645173)
        ref[5,1,1,1]=o*(-0.233710263631)*x_ref[1]**(o-1)+(-0.432942345121)
        ref[5,1,1,2]=o*(0.270484113666)*x_ref[2]**(o-1)+(0.707302418115)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedSolution_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.662618392141)*x[0]**o+(-0.42617914202)*x[0]+(-0.419315426864)*x[1]**o+(-0.63613129268)*x[1]
        ref[0]=o*(-0.662618392141)*x_ref[0]**(o-1)+(-0.42617914202)
        ref[1]=o*(-0.419315426864)*x_ref[1]**(o-1)+(-0.63613129268)
      else:
        s=(-0.762028198747)*x[0]**o+(0.14263183612)*x[0]+(0.185743923374)*x[1]**o+(-0.75903441887)*x[1]+(0.77116503084)*x[2]**o+(0.404561994654)*x[2]
        ref[0]=o*(-0.762028198747)*x_ref[0]**(o-1)+(0.14263183612)
        ref[1]=o*(0.185743923374)*x_ref[1]**(o-1)+(-0.75903441887)
        ref[2]=o*(0.77116503084)*x_ref[2]**(o-1)+(0.404561994654)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(0.876893099705)*x[0]**o+(-0.177079224798)*x[0]+(-0.527304645017)*x[1]**o+(-0.0349448735698)*x[1]
        s[1]=(-0.534225377579)*x[0]**o+(-0.280049692517)*x[0]+(0.912290702798)*x[1]**o+(-0.236862709157)*x[1]
        ref[0,0]=o*(0.876893099705)*x_ref[0]**(o-1)+(-0.177079224798)
        ref[0,1]=o*(-0.527304645017)*x_ref[1]**(o-1)+(-0.0349448735698)
        ref[1,0]=o*(-0.534225377579)*x_ref[0]**(o-1)+(-0.280049692517)
        ref[1,1]=o*(0.912290702798)*x_ref[1]**(o-1)+(-0.236862709157)
      else:
        s[0]=(-0.254550799071)*x[0]**o+(0.163641443345)*x[0]+(0.224143736485)*x[1]**o+(0.536806485124)*x[1]+(0.195320362214)*x[2]**o+(-0.0914207464578)*x[2]
        s[1]=(0.708798134546)*x[0]**o+(-0.572385005996)*x[0]+(-0.28510012443)*x[1]**o+(-0.305494018321)*x[1]+(-0.944999853945)*x[2]**o+(0.690936294379)*x[2]
        ref[0,0]=o*(-0.254550799071)*x_ref[0]**(o-1)+(0.163641443345)
        ref[0,1]=o*(0.224143736485)*x_ref[1]**(o-1)+(0.536806485124)
        ref[0,2]=o*(0.195320362214)*x_ref[2]**(o-1)+(-0.0914207464578)
        ref[1,0]=o*(0.708798134546)*x_ref[0]**(o-1)+(-0.572385005996)
        ref[1,1]=o*(-0.28510012443)*x_ref[1]**(o-1)+(-0.305494018321)
        ref[1,2]=o*(-0.944999853945)*x_ref[2]**(o-1)+(0.690936294379)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.555539333621)*x[0]**o+(0.568177319612)*x[0]+(-0.515690028057)*x[1]**o+(0.293250590603)*x[1]
        s[0,1]=(0.7906503945)*x[0]**o+(-0.167319950519)*x[0]+(-0.554762741045)*x[1]**o+(-0.669125892384)*x[1]
        s[0,2]=(-0.309865666451)*x[0]**o+(0.419499273794)*x[0]+(-0.563409144055)*x[1]**o+(0.734002122712)*x[1]
        s[0,3]=(-0.93657260436)*x[0]**o+(0.586601294944)*x[0]+(-0.0753458445127)*x[1]**o+(-0.429871038581)*x[1]
        s[0,4]=(-0.198003364647)*x[0]**o+(-0.76482793062)*x[0]+(-0.125188294649)*x[1]**o+(0.385219545236)*x[1]
        s[1,0]=(-0.179923035467)*x[0]**o+(0.962110768245)*x[0]+(0.422898666863)*x[1]**o+(-0.92193819062)*x[1]
        s[1,1]=(-0.490610670554)*x[0]**o+(-0.876474842232)*x[0]+(-0.870535201491)*x[1]**o+(-0.208576590799)*x[1]
        s[1,2]=(0.625747905975)*x[0]**o+(0.766151918225)*x[0]+(-0.865271283252)*x[1]**o+(-0.984130648049)*x[1]
        s[1,3]=(0.21674453499)*x[0]**o+(0.74684159223)*x[0]+(0.968938733366)*x[1]**o+(-0.980174574707)*x[1]
        s[1,4]=(0.15542486304)*x[0]**o+(-0.288729546241)*x[0]+(0.780533404805)*x[1]**o+(0.201416001866)*x[1]
        s[2,0]=(0.353102047108)*x[0]**o+(-0.388306135155)*x[0]+(0.837210776206)*x[1]**o+(0.775323252612)*x[1]
        s[2,1]=(-0.407280915312)*x[0]**o+(-0.53768541262)*x[0]+(0.893214604403)*x[1]**o+(-0.69705043663)*x[1]
        s[2,2]=(-0.447169656337)*x[0]**o+(0.728088534193)*x[0]+(0.0485994009494)*x[1]**o+(-0.503223824878)*x[1]
        s[2,3]=(-0.19757745902)*x[0]**o+(-0.627533475183)*x[0]+(0.962502499281)*x[1]**o+(0.859404568594)*x[1]
        s[2,4]=(-0.796987077625)*x[0]**o+(0.0568885495838)*x[0]+(0.84204928286)*x[1]**o+(0.521081919225)*x[1]
        s[3,0]=(-0.310704277911)*x[0]**o+(0.768329284718)*x[0]+(0.348908427651)*x[1]**o+(0.403870270277)*x[1]
        s[3,1]=(0.655277014208)*x[0]**o+(-0.770568143234)*x[0]+(0.101447715545)*x[1]**o+(0.525372367142)*x[1]
        s[3,2]=(-0.450815491992)*x[0]**o+(0.267341563784)*x[0]+(0.666360532863)*x[1]**o+(0.25468870732)*x[1]
        s[3,3]=(-0.448798233791)*x[0]**o+(0.825032894501)*x[0]+(0.0555349523209)*x[1]**o+(0.168286007255)*x[1]
        s[3,4]=(-0.260484305029)*x[0]**o+(0.265463408379)*x[0]+(0.147356758821)*x[1]**o+(-0.916280010272)*x[1]
        ref[0,0,0]=o*(-0.555539333621)*x_ref[0]**(o-1)+(0.568177319612)
        ref[0,0,1]=o*(-0.515690028057)*x_ref[1]**(o-1)+(0.293250590603)
        ref[0,1,0]=o*(0.7906503945)*x_ref[0]**(o-1)+(-0.167319950519)
        ref[0,1,1]=o*(-0.554762741045)*x_ref[1]**(o-1)+(-0.669125892384)
        ref[0,2,0]=o*(-0.309865666451)*x_ref[0]**(o-1)+(0.419499273794)
        ref[0,2,1]=o*(-0.563409144055)*x_ref[1]**(o-1)+(0.734002122712)
        ref[0,3,0]=o*(-0.93657260436)*x_ref[0]**(o-1)+(0.586601294944)
        ref[0,3,1]=o*(-0.0753458445127)*x_ref[1]**(o-1)+(-0.429871038581)
        ref[0,4,0]=o*(-0.198003364647)*x_ref[0]**(o-1)+(-0.76482793062)
        ref[0,4,1]=o*(-0.125188294649)*x_ref[1]**(o-1)+(0.385219545236)
        ref[1,0,0]=o*(-0.179923035467)*x_ref[0]**(o-1)+(0.962110768245)
        ref[1,0,1]=o*(0.422898666863)*x_ref[1]**(o-1)+(-0.92193819062)
        ref[1,1,0]=o*(-0.490610670554)*x_ref[0]**(o-1)+(-0.876474842232)
        ref[1,1,1]=o*(-0.870535201491)*x_ref[1]**(o-1)+(-0.208576590799)
        ref[1,2,0]=o*(0.625747905975)*x_ref[0]**(o-1)+(0.766151918225)
        ref[1,2,1]=o*(-0.865271283252)*x_ref[1]**(o-1)+(-0.984130648049)
        ref[1,3,0]=o*(0.21674453499)*x_ref[0]**(o-1)+(0.74684159223)
        ref[1,3,1]=o*(0.968938733366)*x_ref[1]**(o-1)+(-0.980174574707)
        ref[1,4,0]=o*(0.15542486304)*x_ref[0]**(o-1)+(-0.288729546241)
        ref[1,4,1]=o*(0.780533404805)*x_ref[1]**(o-1)+(0.201416001866)
        ref[2,0,0]=o*(0.353102047108)*x_ref[0]**(o-1)+(-0.388306135155)
        ref[2,0,1]=o*(0.837210776206)*x_ref[1]**(o-1)+(0.775323252612)
        ref[2,1,0]=o*(-0.407280915312)*x_ref[0]**(o-1)+(-0.53768541262)
        ref[2,1,1]=o*(0.893214604403)*x_ref[1]**(o-1)+(-0.69705043663)
        ref[2,2,0]=o*(-0.447169656337)*x_ref[0]**(o-1)+(0.728088534193)
        ref[2,2,1]=o*(0.0485994009494)*x_ref[1]**(o-1)+(-0.503223824878)
        ref[2,3,0]=o*(-0.19757745902)*x_ref[0]**(o-1)+(-0.627533475183)
        ref[2,3,1]=o*(0.962502499281)*x_ref[1]**(o-1)+(0.859404568594)
        ref[2,4,0]=o*(-0.796987077625)*x_ref[0]**(o-1)+(0.0568885495838)
        ref[2,4,1]=o*(0.84204928286)*x_ref[1]**(o-1)+(0.521081919225)
        ref[3,0,0]=o*(-0.310704277911)*x_ref[0]**(o-1)+(0.768329284718)
        ref[3,0,1]=o*(0.348908427651)*x_ref[1]**(o-1)+(0.403870270277)
        ref[3,1,0]=o*(0.655277014208)*x_ref[0]**(o-1)+(-0.770568143234)
        ref[3,1,1]=o*(0.101447715545)*x_ref[1]**(o-1)+(0.525372367142)
        ref[3,2,0]=o*(-0.450815491992)*x_ref[0]**(o-1)+(0.267341563784)
        ref[3,2,1]=o*(0.666360532863)*x_ref[1]**(o-1)+(0.25468870732)
        ref[3,3,0]=o*(-0.448798233791)*x_ref[0]**(o-1)+(0.825032894501)
        ref[3,3,1]=o*(0.0555349523209)*x_ref[1]**(o-1)+(0.168286007255)
        ref[3,4,0]=o*(-0.260484305029)*x_ref[0]**(o-1)+(0.265463408379)
        ref[3,4,1]=o*(0.147356758821)*x_ref[1]**(o-1)+(-0.916280010272)
      else:
        s[0,0]=(0.344242828216)*x[0]**o+(0.491144523585)*x[0]+(-0.92238663597)*x[1]**o+(0.655901250824)*x[1]+(0.683755311378)*x[2]**o+(-0.499214028484)*x[2]
        s[0,1]=(0.246950598177)*x[0]**o+(0.690983811313)*x[0]+(0.0441182184921)*x[1]**o+(-0.236074195046)*x[1]+(-0.365664761621)*x[2]**o+(0.406464337338)*x[2]
        s[0,2]=(-0.327739540606)*x[0]**o+(-0.96200067317)*x[0]+(-0.634811990063)*x[1]**o+(-0.0028455698091)*x[1]+(-0.773566753548)*x[2]**o+(-0.944614851063)*x[2]
        s[0,3]=(-0.423726389514)*x[0]**o+(0.930379169254)*x[0]+(0.551259733694)*x[1]**o+(-0.277405976565)*x[1]+(0.429454682659)*x[2]**o+(0.665041921999)*x[2]
        s[0,4]=(-0.808919025592)*x[0]**o+(0.527988478329)*x[0]+(0.598427398457)*x[1]**o+(-0.740120252678)*x[1]+(-0.57495557526)*x[2]**o+(-0.09358843573)*x[2]
        s[1,0]=(0.139579066196)*x[0]**o+(0.64838866949)*x[0]+(-0.310649190227)*x[1]**o+(-0.947998845563)*x[1]+(0.15735741361)*x[2]**o+(0.521031688881)*x[2]
        s[1,1]=(0.3564671327)*x[0]**o+(0.44368001644)*x[0]+(-0.867803053218)*x[1]**o+(-0.808576642073)*x[1]+(0.164399621384)*x[2]**o+(-0.929324362594)*x[2]
        s[1,2]=(-0.719971382108)*x[0]**o+(-0.509264166588)*x[0]+(0.880679383007)*x[1]**o+(0.150377261228)*x[1]+(-0.695573153702)*x[2]**o+(-0.504859578896)*x[2]
        s[1,3]=(0.19999237262)*x[0]**o+(0.301010290809)*x[0]+(-0.761383660349)*x[1]**o+(-0.943121991098)*x[1]+(-0.824426064219)*x[2]**o+(-0.288809145407)*x[2]
        s[1,4]=(0.0209332296639)*x[0]**o+(-0.745572124803)*x[0]+(0.649794969638)*x[1]**o+(0.997896690337)*x[1]+(0.245116895509)*x[2]**o+(0.934427411594)*x[2]
        s[2,0]=(-0.415649481447)*x[0]**o+(0.967011773751)*x[0]+(0.850275812024)*x[1]**o+(0.368460051013)*x[1]+(-0.895648484555)*x[2]**o+(-0.748318156303)*x[2]
        s[2,1]=(0.822253194356)*x[0]**o+(-0.209207542819)*x[0]+(-0.323961739942)*x[1]**o+(-0.108506915189)*x[1]+(0.39419174146)*x[2]**o+(-0.880778942413)*x[2]
        s[2,2]=(0.730106025416)*x[0]**o+(-0.0921212828406)*x[0]+(-0.0790190498056)*x[1]**o+(0.456083884786)*x[1]+(-0.226537921015)*x[2]**o+(0.293694285152)*x[2]
        s[2,3]=(-0.978825730702)*x[0]**o+(-0.914905098143)*x[0]+(-0.110991764399)*x[1]**o+(-0.144744770355)*x[1]+(0.625189542344)*x[2]**o+(0.804751738687)*x[2]
        s[2,4]=(0.862409094298)*x[0]**o+(-0.953204890595)*x[0]+(-0.66083629488)*x[1]**o+(0.892835328867)*x[1]+(-0.724419944382)*x[2]**o+(-0.352952976454)*x[2]
        s[3,0]=(0.518573942986)*x[0]**o+(0.525366624319)*x[0]+(-0.184521233411)*x[1]**o+(-0.577732664805)*x[1]+(0.194495181041)*x[2]**o+(-0.651069449215)*x[2]
        s[3,1]=(-0.900134328609)*x[0]**o+(-0.502594184837)*x[0]+(0.972006928979)*x[1]**o+(-0.58560991354)*x[1]+(0.423531955361)*x[2]**o+(-0.149983453241)*x[2]
        s[3,2]=(-0.949776951808)*x[0]**o+(-0.557083489688)*x[0]+(-0.860426216679)*x[1]**o+(0.740745372034)*x[1]+(-0.770598389143)*x[2]**o+(0.576226272811)*x[2]
        s[3,3]=(0.947181609851)*x[0]**o+(0.624290827777)*x[0]+(0.163876818237)*x[1]**o+(-0.881648295139)*x[1]+(0.460250347501)*x[2]**o+(0.278384518072)*x[2]
        s[3,4]=(-0.235115463614)*x[0]**o+(-0.608162950212)*x[0]+(-0.942527781961)*x[1]**o+(-0.473027994027)*x[1]+(0.784701664346)*x[2]**o+(0.67497154466)*x[2]
        ref[0,0,0]=o*(0.344242828216)*x_ref[0]**(o-1)+(0.491144523585)
        ref[0,0,1]=o*(-0.92238663597)*x_ref[1]**(o-1)+(0.655901250824)
        ref[0,0,2]=o*(0.683755311378)*x_ref[2]**(o-1)+(-0.499214028484)
        ref[0,1,0]=o*(0.246950598177)*x_ref[0]**(o-1)+(0.690983811313)
        ref[0,1,1]=o*(0.0441182184921)*x_ref[1]**(o-1)+(-0.236074195046)
        ref[0,1,2]=o*(-0.365664761621)*x_ref[2]**(o-1)+(0.406464337338)
        ref[0,2,0]=o*(-0.327739540606)*x_ref[0]**(o-1)+(-0.96200067317)
        ref[0,2,1]=o*(-0.634811990063)*x_ref[1]**(o-1)+(-0.0028455698091)
        ref[0,2,2]=o*(-0.773566753548)*x_ref[2]**(o-1)+(-0.944614851063)
        ref[0,3,0]=o*(-0.423726389514)*x_ref[0]**(o-1)+(0.930379169254)
        ref[0,3,1]=o*(0.551259733694)*x_ref[1]**(o-1)+(-0.277405976565)
        ref[0,3,2]=o*(0.429454682659)*x_ref[2]**(o-1)+(0.665041921999)
        ref[0,4,0]=o*(-0.808919025592)*x_ref[0]**(o-1)+(0.527988478329)
        ref[0,4,1]=o*(0.598427398457)*x_ref[1]**(o-1)+(-0.740120252678)
        ref[0,4,2]=o*(-0.57495557526)*x_ref[2]**(o-1)+(-0.09358843573)
        ref[1,0,0]=o*(0.139579066196)*x_ref[0]**(o-1)+(0.64838866949)
        ref[1,0,1]=o*(-0.310649190227)*x_ref[1]**(o-1)+(-0.947998845563)
        ref[1,0,2]=o*(0.15735741361)*x_ref[2]**(o-1)+(0.521031688881)
        ref[1,1,0]=o*(0.3564671327)*x_ref[0]**(o-1)+(0.44368001644)
        ref[1,1,1]=o*(-0.867803053218)*x_ref[1]**(o-1)+(-0.808576642073)
        ref[1,1,2]=o*(0.164399621384)*x_ref[2]**(o-1)+(-0.929324362594)
        ref[1,2,0]=o*(-0.719971382108)*x_ref[0]**(o-1)+(-0.509264166588)
        ref[1,2,1]=o*(0.880679383007)*x_ref[1]**(o-1)+(0.150377261228)
        ref[1,2,2]=o*(-0.695573153702)*x_ref[2]**(o-1)+(-0.504859578896)
        ref[1,3,0]=o*(0.19999237262)*x_ref[0]**(o-1)+(0.301010290809)
        ref[1,3,1]=o*(-0.761383660349)*x_ref[1]**(o-1)+(-0.943121991098)
        ref[1,3,2]=o*(-0.824426064219)*x_ref[2]**(o-1)+(-0.288809145407)
        ref[1,4,0]=o*(0.0209332296639)*x_ref[0]**(o-1)+(-0.745572124803)
        ref[1,4,1]=o*(0.649794969638)*x_ref[1]**(o-1)+(0.997896690337)
        ref[1,4,2]=o*(0.245116895509)*x_ref[2]**(o-1)+(0.934427411594)
        ref[2,0,0]=o*(-0.415649481447)*x_ref[0]**(o-1)+(0.967011773751)
        ref[2,0,1]=o*(0.850275812024)*x_ref[1]**(o-1)+(0.368460051013)
        ref[2,0,2]=o*(-0.895648484555)*x_ref[2]**(o-1)+(-0.748318156303)
        ref[2,1,0]=o*(0.822253194356)*x_ref[0]**(o-1)+(-0.209207542819)
        ref[2,1,1]=o*(-0.323961739942)*x_ref[1]**(o-1)+(-0.108506915189)
        ref[2,1,2]=o*(0.39419174146)*x_ref[2]**(o-1)+(-0.880778942413)
        ref[2,2,0]=o*(0.730106025416)*x_ref[0]**(o-1)+(-0.0921212828406)
        ref[2,2,1]=o*(-0.0790190498056)*x_ref[1]**(o-1)+(0.456083884786)
        ref[2,2,2]=o*(-0.226537921015)*x_ref[2]**(o-1)+(0.293694285152)
        ref[2,3,0]=o*(-0.978825730702)*x_ref[0]**(o-1)+(-0.914905098143)
        ref[2,3,1]=o*(-0.110991764399)*x_ref[1]**(o-1)+(-0.144744770355)
        ref[2,3,2]=o*(0.625189542344)*x_ref[2]**(o-1)+(0.804751738687)
        ref[2,4,0]=o*(0.862409094298)*x_ref[0]**(o-1)+(-0.953204890595)
        ref[2,4,1]=o*(-0.66083629488)*x_ref[1]**(o-1)+(0.892835328867)
        ref[2,4,2]=o*(-0.724419944382)*x_ref[2]**(o-1)+(-0.352952976454)
        ref[3,0,0]=o*(0.518573942986)*x_ref[0]**(o-1)+(0.525366624319)
        ref[3,0,1]=o*(-0.184521233411)*x_ref[1]**(o-1)+(-0.577732664805)
        ref[3,0,2]=o*(0.194495181041)*x_ref[2]**(o-1)+(-0.651069449215)
        ref[3,1,0]=o*(-0.900134328609)*x_ref[0]**(o-1)+(-0.502594184837)
        ref[3,1,1]=o*(0.972006928979)*x_ref[1]**(o-1)+(-0.58560991354)
        ref[3,1,2]=o*(0.423531955361)*x_ref[2]**(o-1)+(-0.149983453241)
        ref[3,2,0]=o*(-0.949776951808)*x_ref[0]**(o-1)+(-0.557083489688)
        ref[3,2,1]=o*(-0.860426216679)*x_ref[1]**(o-1)+(0.740745372034)
        ref[3,2,2]=o*(-0.770598389143)*x_ref[2]**(o-1)+(0.576226272811)
        ref[3,3,0]=o*(0.947181609851)*x_ref[0]**(o-1)+(0.624290827777)
        ref[3,3,1]=o*(0.163876818237)*x_ref[1]**(o-1)+(-0.881648295139)
        ref[3,3,2]=o*(0.460250347501)*x_ref[2]**(o-1)+(0.278384518072)
        ref[3,4,0]=o*(-0.235115463614)*x_ref[0]**(o-1)+(-0.608162950212)
        ref[3,4,1]=o*(-0.942527781961)*x_ref[1]**(o-1)+(-0.473027994027)
        ref[3,4,2]=o*(0.784701664346)*x_ref[2]**(o-1)+(0.67497154466)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(-0.19873307647)*x[0]**o+(-0.402324928722)*x[0]+(0.78581260176)*x[1]**o+(0.855692742935)*x[1]
        s[0,0,1]=(-0.976563405673)*x[0]**o+(0.100149365141)*x[0]+(0.630689651275)*x[1]**o+(0.601727769629)*x[1]
        s[0,1,0]=(-0.832835486573)*x[0]**o+(0.0671587043511)*x[0]+(-0.610210594342)*x[1]**o+(-0.436026877569)*x[1]
        s[0,1,1]=(0.292942538448)*x[0]**o+(0.0385459359421)*x[0]+(0.790299248934)*x[1]**o+(-0.319413494032)*x[1]
        s[1,0,0]=(0.58698003273)*x[0]**o+(0.613088892876)*x[0]+(0.552617483007)*x[1]**o+(0.720306398451)*x[1]
        s[1,0,1]=(0.955352789796)*x[0]**o+(0.174378913925)*x[0]+(-0.0314000136685)*x[1]**o+(0.959578237088)*x[1]
        s[1,1,0]=(0.281022820802)*x[0]**o+(0.326052665369)*x[0]+(-0.492525348509)*x[1]**o+(-0.466634839923)*x[1]
        s[1,1,1]=(0.479963735062)*x[0]**o+(-0.779437072555)*x[0]+(-0.208397089582)*x[1]**o+(-0.303640352239)*x[1]
        s[2,0,0]=(0.434668114042)*x[0]**o+(0.0691902144608)*x[0]+(-0.177103703295)*x[1]**o+(0.91338709789)*x[1]
        s[2,0,1]=(-0.476825738275)*x[0]**o+(0.819380857167)*x[0]+(-0.661832108828)*x[1]**o+(0.873879046431)*x[1]
        s[2,1,0]=(0.30919914009)*x[0]**o+(-0.0366057876782)*x[0]+(-0.966956881285)*x[1]**o+(-0.829788683577)*x[1]
        s[2,1,1]=(-0.168382165296)*x[0]**o+(-0.682198962507)*x[0]+(0.190572223366)*x[1]**o+(-0.667568574676)*x[1]
        s[3,0,0]=(0.668658559933)*x[0]**o+(-0.253060573754)*x[0]+(-0.864766645952)*x[1]**o+(-0.714197991697)*x[1]
        s[3,0,1]=(0.538191340464)*x[0]**o+(0.806266803991)*x[0]+(0.437103643795)*x[1]**o+(0.972593986026)*x[1]
        s[3,1,0]=(-0.120672797742)*x[0]**o+(0.840236886959)*x[0]+(-0.279061306399)*x[1]**o+(0.119529093461)*x[1]
        s[3,1,1]=(-0.405723067833)*x[0]**o+(-0.964291783264)*x[0]+(0.201561553981)*x[1]**o+(0.882148297909)*x[1]
        s[4,0,0]=(-0.501360654417)*x[0]**o+(-0.977210439162)*x[0]+(0.0742786636276)*x[1]**o+(-0.831462191269)*x[1]
        s[4,0,1]=(-0.354758190164)*x[0]**o+(0.737302507124)*x[0]+(-0.221806769305)*x[1]**o+(-0.0261943694435)*x[1]
        s[4,1,0]=(-0.921367050665)*x[0]**o+(-0.77028348328)*x[0]+(-0.799765110774)*x[1]**o+(-0.708684233564)*x[1]
        s[4,1,1]=(-0.764817738089)*x[0]**o+(0.13308831245)*x[0]+(-0.0835084988678)*x[1]**o+(0.181328374835)*x[1]
        s[5,0,0]=(0.0795012634978)*x[0]**o+(0.527681277019)*x[0]+(-0.0656230383346)*x[1]**o+(0.0884801523765)*x[1]
        s[5,0,1]=(-0.316260178996)*x[0]**o+(-0.372232635711)*x[0]+(0.868288352534)*x[1]**o+(0.22823979163)*x[1]
        s[5,1,0]=(0.655698586327)*x[0]**o+(-0.734551617528)*x[0]+(-0.627595993352)*x[1]**o+(0.485738750199)*x[1]
        s[5,1,1]=(-0.522687357394)*x[0]**o+(-0.232523849025)*x[0]+(0.559349644178)*x[1]**o+(-0.640468984193)*x[1]
        ref[0,0,0,0]=o*(-0.19873307647)*x_ref[0]**(o-1)+(-0.402324928722)
        ref[0,0,0,1]=o*(0.78581260176)*x_ref[1]**(o-1)+(0.855692742935)
        ref[0,0,1,0]=o*(-0.976563405673)*x_ref[0]**(o-1)+(0.100149365141)
        ref[0,0,1,1]=o*(0.630689651275)*x_ref[1]**(o-1)+(0.601727769629)
        ref[0,1,0,0]=o*(-0.832835486573)*x_ref[0]**(o-1)+(0.0671587043511)
        ref[0,1,0,1]=o*(-0.610210594342)*x_ref[1]**(o-1)+(-0.436026877569)
        ref[0,1,1,0]=o*(0.292942538448)*x_ref[0]**(o-1)+(0.0385459359421)
        ref[0,1,1,1]=o*(0.790299248934)*x_ref[1]**(o-1)+(-0.319413494032)
        ref[1,0,0,0]=o*(0.58698003273)*x_ref[0]**(o-1)+(0.613088892876)
        ref[1,0,0,1]=o*(0.552617483007)*x_ref[1]**(o-1)+(0.720306398451)
        ref[1,0,1,0]=o*(0.955352789796)*x_ref[0]**(o-1)+(0.174378913925)
        ref[1,0,1,1]=o*(-0.0314000136685)*x_ref[1]**(o-1)+(0.959578237088)
        ref[1,1,0,0]=o*(0.281022820802)*x_ref[0]**(o-1)+(0.326052665369)
        ref[1,1,0,1]=o*(-0.492525348509)*x_ref[1]**(o-1)+(-0.466634839923)
        ref[1,1,1,0]=o*(0.479963735062)*x_ref[0]**(o-1)+(-0.779437072555)
        ref[1,1,1,1]=o*(-0.208397089582)*x_ref[1]**(o-1)+(-0.303640352239)
        ref[2,0,0,0]=o*(0.434668114042)*x_ref[0]**(o-1)+(0.0691902144608)
        ref[2,0,0,1]=o*(-0.177103703295)*x_ref[1]**(o-1)+(0.91338709789)
        ref[2,0,1,0]=o*(-0.476825738275)*x_ref[0]**(o-1)+(0.819380857167)
        ref[2,0,1,1]=o*(-0.661832108828)*x_ref[1]**(o-1)+(0.873879046431)
        ref[2,1,0,0]=o*(0.30919914009)*x_ref[0]**(o-1)+(-0.0366057876782)
        ref[2,1,0,1]=o*(-0.966956881285)*x_ref[1]**(o-1)+(-0.829788683577)
        ref[2,1,1,0]=o*(-0.168382165296)*x_ref[0]**(o-1)+(-0.682198962507)
        ref[2,1,1,1]=o*(0.190572223366)*x_ref[1]**(o-1)+(-0.667568574676)
        ref[3,0,0,0]=o*(0.668658559933)*x_ref[0]**(o-1)+(-0.253060573754)
        ref[3,0,0,1]=o*(-0.864766645952)*x_ref[1]**(o-1)+(-0.714197991697)
        ref[3,0,1,0]=o*(0.538191340464)*x_ref[0]**(o-1)+(0.806266803991)
        ref[3,0,1,1]=o*(0.437103643795)*x_ref[1]**(o-1)+(0.972593986026)
        ref[3,1,0,0]=o*(-0.120672797742)*x_ref[0]**(o-1)+(0.840236886959)
        ref[3,1,0,1]=o*(-0.279061306399)*x_ref[1]**(o-1)+(0.119529093461)
        ref[3,1,1,0]=o*(-0.405723067833)*x_ref[0]**(o-1)+(-0.964291783264)
        ref[3,1,1,1]=o*(0.201561553981)*x_ref[1]**(o-1)+(0.882148297909)
        ref[4,0,0,0]=o*(-0.501360654417)*x_ref[0]**(o-1)+(-0.977210439162)
        ref[4,0,0,1]=o*(0.0742786636276)*x_ref[1]**(o-1)+(-0.831462191269)
        ref[4,0,1,0]=o*(-0.354758190164)*x_ref[0]**(o-1)+(0.737302507124)
        ref[4,0,1,1]=o*(-0.221806769305)*x_ref[1]**(o-1)+(-0.0261943694435)
        ref[4,1,0,0]=o*(-0.921367050665)*x_ref[0]**(o-1)+(-0.77028348328)
        ref[4,1,0,1]=o*(-0.799765110774)*x_ref[1]**(o-1)+(-0.708684233564)
        ref[4,1,1,0]=o*(-0.764817738089)*x_ref[0]**(o-1)+(0.13308831245)
        ref[4,1,1,1]=o*(-0.0835084988678)*x_ref[1]**(o-1)+(0.181328374835)
        ref[5,0,0,0]=o*(0.0795012634978)*x_ref[0]**(o-1)+(0.527681277019)
        ref[5,0,0,1]=o*(-0.0656230383346)*x_ref[1]**(o-1)+(0.0884801523765)
        ref[5,0,1,0]=o*(-0.316260178996)*x_ref[0]**(o-1)+(-0.372232635711)
        ref[5,0,1,1]=o*(0.868288352534)*x_ref[1]**(o-1)+(0.22823979163)
        ref[5,1,0,0]=o*(0.655698586327)*x_ref[0]**(o-1)+(-0.734551617528)
        ref[5,1,0,1]=o*(-0.627595993352)*x_ref[1]**(o-1)+(0.485738750199)
        ref[5,1,1,0]=o*(-0.522687357394)*x_ref[0]**(o-1)+(-0.232523849025)
        ref[5,1,1,1]=o*(0.559349644178)*x_ref[1]**(o-1)+(-0.640468984193)
      else:
        s[0,0,0]=(-0.252254362053)*x[0]**o+(-0.73201258434)*x[0]+(0.129279219807)*x[1]**o+(0.739511847421)*x[1]+(-0.335757163655)*x[2]**o+(0.645975465365)*x[2]
        s[0,0,1]=(-0.735055416054)*x[0]**o+(-0.392972979059)*x[0]+(-0.400312505739)*x[1]**o+(-0.431371712491)*x[1]+(0.570395624547)*x[2]**o+(0.138472351049)*x[2]
        s[0,1,0]=(-0.000809920184319)*x[0]**o+(0.39610549699)*x[0]+(0.46183657032)*x[1]**o+(-0.90835208084)*x[1]+(0.359160987844)*x[2]**o+(0.63185061127)*x[2]
        s[0,1,1]=(0.627804065013)*x[0]**o+(-0.616378680528)*x[0]+(-0.623715533085)*x[1]**o+(-0.457832375425)*x[1]+(-0.274712768232)*x[2]**o+(0.425333514519)*x[2]
        s[1,0,0]=(-0.901107526813)*x[0]**o+(0.0680326855019)*x[0]+(-0.24023047937)*x[1]**o+(0.33244773534)*x[1]+(0.649715117184)*x[2]**o+(-0.832757442484)*x[2]
        s[1,0,1]=(0.419698786285)*x[0]**o+(-0.929558931125)*x[0]+(-0.211068611062)*x[1]**o+(-0.308568292995)*x[1]+(-0.302460643894)*x[2]**o+(-0.792090369968)*x[2]
        s[1,1,0]=(-0.350071906565)*x[0]**o+(0.915476959141)*x[0]+(0.511206427652)*x[1]**o+(0.925740134945)*x[1]+(-0.163642337237)*x[2]**o+(0.466249765823)*x[2]
        s[1,1,1]=(-0.00679707570295)*x[0]**o+(-0.20076295244)*x[0]+(0.264205579203)*x[1]**o+(-0.411292187293)*x[1]+(-0.934377917835)*x[2]**o+(0.634215851155)*x[2]
        s[2,0,0]=(-0.324841664958)*x[0]**o+(-0.956529004217)*x[0]+(-0.918338753073)*x[1]**o+(0.626923365653)*x[1]+(-0.799444332565)*x[2]**o+(-0.0695464844552)*x[2]
        s[2,0,1]=(0.506623201982)*x[0]**o+(-0.149035848663)*x[0]+(-0.238527222625)*x[1]**o+(0.179414470428)*x[1]+(-0.830587331222)*x[2]**o+(-0.771223952622)*x[2]
        s[2,1,0]=(-0.600971886655)*x[0]**o+(0.578997478246)*x[0]+(0.307807603149)*x[1]**o+(-0.187632470569)*x[1]+(0.715704085441)*x[2]**o+(0.666008020335)*x[2]
        s[2,1,1]=(-0.927640726378)*x[0]**o+(-0.615258128661)*x[0]+(0.0458637924464)*x[1]**o+(0.316776923148)*x[1]+(-0.816443096595)*x[2]**o+(0.586050614098)*x[2]
        s[3,0,0]=(0.0116773788954)*x[0]**o+(-0.450016183617)*x[0]+(-0.722050331425)*x[1]**o+(-0.752741628658)*x[1]+(0.0807023425045)*x[2]**o+(-0.888353661711)*x[2]
        s[3,0,1]=(0.774410128748)*x[0]**o+(0.00319037654241)*x[0]+(-0.872810066503)*x[1]**o+(-0.667066572773)*x[1]+(0.87558896843)*x[2]**o+(-0.691829241879)*x[2]
        s[3,1,0]=(-0.81762346888)*x[0]**o+(0.117005969026)*x[0]+(-0.745835352632)*x[1]**o+(-0.162993479637)*x[1]+(0.259977485657)*x[2]**o+(-0.4154446656)*x[2]
        s[3,1,1]=(0.0202189083822)*x[0]**o+(0.276300397225)*x[0]+(-0.473252990801)*x[1]**o+(0.139373130343)*x[1]+(0.236401212883)*x[2]**o+(0.250980379862)*x[2]
        s[4,0,0]=(0.114221091454)*x[0]**o+(0.956199921345)*x[0]+(0.920335553649)*x[1]**o+(0.953720056733)*x[1]+(-0.869205945463)*x[2]**o+(-0.702596947421)*x[2]
        s[4,0,1]=(0.242788249724)*x[0]**o+(0.484618105306)*x[0]+(-0.0856836792276)*x[1]**o+(0.314175782784)*x[1]+(-0.462603325333)*x[2]**o+(-0.527952799625)*x[2]
        s[4,1,0]=(0.124168860106)*x[0]**o+(-0.846086223242)*x[0]+(-0.536186421788)*x[1]**o+(-0.103140610306)*x[1]+(-0.115670483732)*x[2]**o+(0.771594201067)*x[2]
        s[4,1,1]=(0.649717322503)*x[0]**o+(0.911557314842)*x[0]+(-0.498767014909)*x[1]**o+(0.647065328805)*x[1]+(0.722903839932)*x[2]**o+(0.514148167148)*x[2]
        s[5,0,0]=(-0.380297540184)*x[0]**o+(-0.485514278978)*x[0]+(0.785198036928)*x[1]**o+(-0.0846164275939)*x[1]+(-0.252406925767)*x[2]**o+(-0.223499486716)*x[2]
        s[5,0,1]=(-0.184173490174)*x[0]**o+(0.569523639256)*x[0]+(-0.294243155077)*x[1]**o+(-0.323071302538)*x[1]+(0.661442924711)*x[2]**o+(0.640705614167)*x[2]
        s[5,1,0]=(0.110804523334)*x[0]**o+(-0.933148550901)*x[0]+(-0.664439117013)*x[1]**o+(-0.33712595594)*x[1]+(0.0561563633834)*x[2]**o+(-0.728089270393)*x[2]
        s[5,1,1]=(-0.474826954962)*x[0]**o+(0.953336718849)*x[0]+(0.616224037612)*x[1]**o+(0.596223200958)*x[1]+(-0.790460006091)*x[2]**o+(0.580623251887)*x[2]
        ref[0,0,0,0]=o*(-0.252254362053)*x_ref[0]**(o-1)+(-0.73201258434)
        ref[0,0,0,1]=o*(0.129279219807)*x_ref[1]**(o-1)+(0.739511847421)
        ref[0,0,0,2]=o*(-0.335757163655)*x_ref[2]**(o-1)+(0.645975465365)
        ref[0,0,1,0]=o*(-0.735055416054)*x_ref[0]**(o-1)+(-0.392972979059)
        ref[0,0,1,1]=o*(-0.400312505739)*x_ref[1]**(o-1)+(-0.431371712491)
        ref[0,0,1,2]=o*(0.570395624547)*x_ref[2]**(o-1)+(0.138472351049)
        ref[0,1,0,0]=o*(-0.000809920184319)*x_ref[0]**(o-1)+(0.39610549699)
        ref[0,1,0,1]=o*(0.46183657032)*x_ref[1]**(o-1)+(-0.90835208084)
        ref[0,1,0,2]=o*(0.359160987844)*x_ref[2]**(o-1)+(0.63185061127)
        ref[0,1,1,0]=o*(0.627804065013)*x_ref[0]**(o-1)+(-0.616378680528)
        ref[0,1,1,1]=o*(-0.623715533085)*x_ref[1]**(o-1)+(-0.457832375425)
        ref[0,1,1,2]=o*(-0.274712768232)*x_ref[2]**(o-1)+(0.425333514519)
        ref[1,0,0,0]=o*(-0.901107526813)*x_ref[0]**(o-1)+(0.0680326855019)
        ref[1,0,0,1]=o*(-0.24023047937)*x_ref[1]**(o-1)+(0.33244773534)
        ref[1,0,0,2]=o*(0.649715117184)*x_ref[2]**(o-1)+(-0.832757442484)
        ref[1,0,1,0]=o*(0.419698786285)*x_ref[0]**(o-1)+(-0.929558931125)
        ref[1,0,1,1]=o*(-0.211068611062)*x_ref[1]**(o-1)+(-0.308568292995)
        ref[1,0,1,2]=o*(-0.302460643894)*x_ref[2]**(o-1)+(-0.792090369968)
        ref[1,1,0,0]=o*(-0.350071906565)*x_ref[0]**(o-1)+(0.915476959141)
        ref[1,1,0,1]=o*(0.511206427652)*x_ref[1]**(o-1)+(0.925740134945)
        ref[1,1,0,2]=o*(-0.163642337237)*x_ref[2]**(o-1)+(0.466249765823)
        ref[1,1,1,0]=o*(-0.00679707570295)*x_ref[0]**(o-1)+(-0.20076295244)
        ref[1,1,1,1]=o*(0.264205579203)*x_ref[1]**(o-1)+(-0.411292187293)
        ref[1,1,1,2]=o*(-0.934377917835)*x_ref[2]**(o-1)+(0.634215851155)
        ref[2,0,0,0]=o*(-0.324841664958)*x_ref[0]**(o-1)+(-0.956529004217)
        ref[2,0,0,1]=o*(-0.918338753073)*x_ref[1]**(o-1)+(0.626923365653)
        ref[2,0,0,2]=o*(-0.799444332565)*x_ref[2]**(o-1)+(-0.0695464844552)
        ref[2,0,1,0]=o*(0.506623201982)*x_ref[0]**(o-1)+(-0.149035848663)
        ref[2,0,1,1]=o*(-0.238527222625)*x_ref[1]**(o-1)+(0.179414470428)
        ref[2,0,1,2]=o*(-0.830587331222)*x_ref[2]**(o-1)+(-0.771223952622)
        ref[2,1,0,0]=o*(-0.600971886655)*x_ref[0]**(o-1)+(0.578997478246)
        ref[2,1,0,1]=o*(0.307807603149)*x_ref[1]**(o-1)+(-0.187632470569)
        ref[2,1,0,2]=o*(0.715704085441)*x_ref[2]**(o-1)+(0.666008020335)
        ref[2,1,1,0]=o*(-0.927640726378)*x_ref[0]**(o-1)+(-0.615258128661)
        ref[2,1,1,1]=o*(0.0458637924464)*x_ref[1]**(o-1)+(0.316776923148)
        ref[2,1,1,2]=o*(-0.816443096595)*x_ref[2]**(o-1)+(0.586050614098)
        ref[3,0,0,0]=o*(0.0116773788954)*x_ref[0]**(o-1)+(-0.450016183617)
        ref[3,0,0,1]=o*(-0.722050331425)*x_ref[1]**(o-1)+(-0.752741628658)
        ref[3,0,0,2]=o*(0.0807023425045)*x_ref[2]**(o-1)+(-0.888353661711)
        ref[3,0,1,0]=o*(0.774410128748)*x_ref[0]**(o-1)+(0.00319037654241)
        ref[3,0,1,1]=o*(-0.872810066503)*x_ref[1]**(o-1)+(-0.667066572773)
        ref[3,0,1,2]=o*(0.87558896843)*x_ref[2]**(o-1)+(-0.691829241879)
        ref[3,1,0,0]=o*(-0.81762346888)*x_ref[0]**(o-1)+(0.117005969026)
        ref[3,1,0,1]=o*(-0.745835352632)*x_ref[1]**(o-1)+(-0.162993479637)
        ref[3,1,0,2]=o*(0.259977485657)*x_ref[2]**(o-1)+(-0.4154446656)
        ref[3,1,1,0]=o*(0.0202189083822)*x_ref[0]**(o-1)+(0.276300397225)
        ref[3,1,1,1]=o*(-0.473252990801)*x_ref[1]**(o-1)+(0.139373130343)
        ref[3,1,1,2]=o*(0.236401212883)*x_ref[2]**(o-1)+(0.250980379862)
        ref[4,0,0,0]=o*(0.114221091454)*x_ref[0]**(o-1)+(0.956199921345)
        ref[4,0,0,1]=o*(0.920335553649)*x_ref[1]**(o-1)+(0.953720056733)
        ref[4,0,0,2]=o*(-0.869205945463)*x_ref[2]**(o-1)+(-0.702596947421)
        ref[4,0,1,0]=o*(0.242788249724)*x_ref[0]**(o-1)+(0.484618105306)
        ref[4,0,1,1]=o*(-0.0856836792276)*x_ref[1]**(o-1)+(0.314175782784)
        ref[4,0,1,2]=o*(-0.462603325333)*x_ref[2]**(o-1)+(-0.527952799625)
        ref[4,1,0,0]=o*(0.124168860106)*x_ref[0]**(o-1)+(-0.846086223242)
        ref[4,1,0,1]=o*(-0.536186421788)*x_ref[1]**(o-1)+(-0.103140610306)
        ref[4,1,0,2]=o*(-0.115670483732)*x_ref[2]**(o-1)+(0.771594201067)
        ref[4,1,1,0]=o*(0.649717322503)*x_ref[0]**(o-1)+(0.911557314842)
        ref[4,1,1,1]=o*(-0.498767014909)*x_ref[1]**(o-1)+(0.647065328805)
        ref[4,1,1,2]=o*(0.722903839932)*x_ref[2]**(o-1)+(0.514148167148)
        ref[5,0,0,0]=o*(-0.380297540184)*x_ref[0]**(o-1)+(-0.485514278978)
        ref[5,0,0,1]=o*(0.785198036928)*x_ref[1]**(o-1)+(-0.0846164275939)
        ref[5,0,0,2]=o*(-0.252406925767)*x_ref[2]**(o-1)+(-0.223499486716)
        ref[5,0,1,0]=o*(-0.184173490174)*x_ref[0]**(o-1)+(0.569523639256)
        ref[5,0,1,1]=o*(-0.294243155077)*x_ref[1]**(o-1)+(-0.323071302538)
        ref[5,0,1,2]=o*(0.661442924711)*x_ref[2]**(o-1)+(0.640705614167)
        ref[5,1,0,0]=o*(0.110804523334)*x_ref[0]**(o-1)+(-0.933148550901)
        ref[5,1,0,1]=o*(-0.664439117013)*x_ref[1]**(o-1)+(-0.33712595594)
        ref[5,1,0,2]=o*(0.0561563633834)*x_ref[2]**(o-1)+(-0.728089270393)
        ref[5,1,1,0]=o*(-0.474826954962)*x_ref[0]**(o-1)+(0.953336718849)
        ref[5,1,1,1]=o*(0.616224037612)*x_ref[1]**(o-1)+(0.596223200958)
        ref[5,1,1,2]=o*(-0.790460006091)*x_ref[2]**(o-1)+(0.580623251887)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_Solution_rank0(self):
      """
      tests gradient for rank 0 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.661534580714)*x[0]**o+(0.108482570281)*x[0]+(0.930475222346)*x[1]**o+(-0.560831828628)*x[1]
        ref[0]=o*(-0.661534580714)*x_ref[0]**(o-1)+(0.108482570281)
        ref[1]=o*(0.930475222346)*x_ref[1]**(o-1)+(-0.560831828628)
      else:
        s=(0.661787087481)*x[0]**o+(0.789038424245)*x[0]+(-0.172842194551)*x[1]**o+(-0.313156164165)*x[1]+(-0.521097764776)*x[2]**o+(0.879698074812)*x[2]
        ref[0]=o*(0.661787087481)*x_ref[0]**(o-1)+(0.789038424245)
        ref[1]=o*(-0.172842194551)*x_ref[1]**(o-1)+(-0.313156164165)
        ref[2]=o*(-0.521097764776)*x_ref[2]**(o-1)+(0.879698074812)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_Solution_rank1(self):
      """
      tests gradient for rank 1 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(-0.190630799823)*x[0]**o+(0.235686163485)*x[0]+(0.568780014934)*x[1]**o+(0.869208291526)*x[1]
        s[1]=(-0.637753770052)*x[0]**o+(-0.184551176704)*x[0]+(0.678386051731)*x[1]**o+(0.103311313677)*x[1]
        ref[0,0]=o*(-0.190630799823)*x_ref[0]**(o-1)+(0.235686163485)
        ref[0,1]=o*(0.568780014934)*x_ref[1]**(o-1)+(0.869208291526)
        ref[1,0]=o*(-0.637753770052)*x_ref[0]**(o-1)+(-0.184551176704)
        ref[1,1]=o*(0.678386051731)*x_ref[1]**(o-1)+(0.103311313677)
      else:
        s[0]=(0.65617389354)*x[0]**o+(-0.41233134159)*x[0]+(-0.471086791763)*x[1]**o+(0.399070328851)*x[1]+(-0.30038453779)*x[2]**o+(0.637544697206)*x[2]
        s[1]=(0.0974925575768)*x[0]**o+(0.968900636228)*x[0]+(-0.386954772478)*x[1]**o+(0.234655265071)*x[1]+(0.755549468261)*x[2]**o+(-0.119269665245)*x[2]
        ref[0,0]=o*(0.65617389354)*x_ref[0]**(o-1)+(-0.41233134159)
        ref[0,1]=o*(-0.471086791763)*x_ref[1]**(o-1)+(0.399070328851)
        ref[0,2]=o*(-0.30038453779)*x_ref[2]**(o-1)+(0.637544697206)
        ref[1,0]=o*(0.0974925575768)*x_ref[0]**(o-1)+(0.968900636228)
        ref[1,1]=o*(-0.386954772478)*x_ref[1]**(o-1)+(0.234655265071)
        ref[1,2]=o*(0.755549468261)*x_ref[2]**(o-1)+(-0.119269665245)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_Solution_rank2(self):
      """
      tests gradient for rank 2 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.300429908481)*x[0]**o+(-0.550176774021)*x[0]+(0.313544523598)*x[1]**o+(0.191557196291)*x[1]
        s[0,1]=(0.00116867375625)*x[0]**o+(0.126481551395)*x[0]+(-0.525739733277)*x[1]**o+(0.628228355029)*x[1]
        s[0,2]=(-0.790055765486)*x[0]**o+(-0.118003096828)*x[0]+(-0.620290760565)*x[1]**o+(-0.460021810788)*x[1]
        s[0,3]=(0.800717644254)*x[0]**o+(-0.685505835346)*x[0]+(-0.382046994517)*x[1]**o+(-0.765159598284)*x[1]
        s[0,4]=(-0.295987131319)*x[0]**o+(-0.0324335264934)*x[0]+(-0.583454515945)*x[1]**o+(-0.137940341278)*x[1]
        s[1,0]=(-0.186271508459)*x[0]**o+(-0.0457260753159)*x[0]+(-0.83853233149)*x[1]**o+(0.730348417106)*x[1]
        s[1,1]=(-0.678758709357)*x[0]**o+(-0.352619765194)*x[0]+(0.850836393219)*x[1]**o+(-0.749995270113)*x[1]
        s[1,2]=(0.598004774999)*x[0]**o+(-0.764313433057)*x[0]+(-0.892691377674)*x[1]**o+(0.436164793316)*x[1]
        s[1,3]=(-0.114398805139)*x[0]**o+(0.170882318781)*x[0]+(-0.159446706481)*x[1]**o+(0.168268554115)*x[1]
        s[1,4]=(0.245440053023)*x[0]**o+(-0.391101810385)*x[0]+(-0.0621967792961)*x[1]**o+(-0.25914597073)*x[1]
        s[2,0]=(0.758930908002)*x[0]**o+(-0.137399611576)*x[0]+(-0.524730401698)*x[1]**o+(-0.68595790282)*x[1]
        s[2,1]=(-0.956320437139)*x[0]**o+(-0.741920593412)*x[0]+(0.470411496813)*x[1]**o+(-0.912492685238)*x[1]
        s[2,2]=(0.386762534234)*x[0]**o+(0.297029046975)*x[0]+(-0.999949621409)*x[1]**o+(-0.273952756751)*x[1]
        s[2,3]=(-0.137635508969)*x[0]**o+(0.0502825343025)*x[0]+(-0.856424864287)*x[1]**o+(0.501432472431)*x[1]
        s[2,4]=(0.927974425929)*x[0]**o+(-0.602904477506)*x[0]+(0.766583044002)*x[1]**o+(0.070024577589)*x[1]
        s[3,0]=(0.336864984737)*x[0]**o+(0.0974502498452)*x[0]+(-0.820387387723)*x[1]**o+(0.022457213421)*x[1]
        s[3,1]=(0.879517107747)*x[0]**o+(0.267883865556)*x[0]+(-0.795732666749)*x[1]**o+(0.81701302209)*x[1]
        s[3,2]=(-0.971853752113)*x[0]**o+(0.210482589767)*x[0]+(0.422058370831)*x[1]**o+(-0.0945771197357)*x[1]
        s[3,3]=(0.0130393893213)*x[0]**o+(0.331617264696)*x[0]+(-0.0460157855829)*x[1]**o+(0.513357385084)*x[1]
        s[3,4]=(-0.466291714481)*x[0]**o+(0.877178907345)*x[0]+(0.356001871079)*x[1]**o+(-0.755177274311)*x[1]
        ref[0,0,0]=o*(-0.300429908481)*x_ref[0]**(o-1)+(-0.550176774021)
        ref[0,0,1]=o*(0.313544523598)*x_ref[1]**(o-1)+(0.191557196291)
        ref[0,1,0]=o*(0.00116867375625)*x_ref[0]**(o-1)+(0.126481551395)
        ref[0,1,1]=o*(-0.525739733277)*x_ref[1]**(o-1)+(0.628228355029)
        ref[0,2,0]=o*(-0.790055765486)*x_ref[0]**(o-1)+(-0.118003096828)
        ref[0,2,1]=o*(-0.620290760565)*x_ref[1]**(o-1)+(-0.460021810788)
        ref[0,3,0]=o*(0.800717644254)*x_ref[0]**(o-1)+(-0.685505835346)
        ref[0,3,1]=o*(-0.382046994517)*x_ref[1]**(o-1)+(-0.765159598284)
        ref[0,4,0]=o*(-0.295987131319)*x_ref[0]**(o-1)+(-0.0324335264934)
        ref[0,4,1]=o*(-0.583454515945)*x_ref[1]**(o-1)+(-0.137940341278)
        ref[1,0,0]=o*(-0.186271508459)*x_ref[0]**(o-1)+(-0.0457260753159)
        ref[1,0,1]=o*(-0.83853233149)*x_ref[1]**(o-1)+(0.730348417106)
        ref[1,1,0]=o*(-0.678758709357)*x_ref[0]**(o-1)+(-0.352619765194)
        ref[1,1,1]=o*(0.850836393219)*x_ref[1]**(o-1)+(-0.749995270113)
        ref[1,2,0]=o*(0.598004774999)*x_ref[0]**(o-1)+(-0.764313433057)
        ref[1,2,1]=o*(-0.892691377674)*x_ref[1]**(o-1)+(0.436164793316)
        ref[1,3,0]=o*(-0.114398805139)*x_ref[0]**(o-1)+(0.170882318781)
        ref[1,3,1]=o*(-0.159446706481)*x_ref[1]**(o-1)+(0.168268554115)
        ref[1,4,0]=o*(0.245440053023)*x_ref[0]**(o-1)+(-0.391101810385)
        ref[1,4,1]=o*(-0.0621967792961)*x_ref[1]**(o-1)+(-0.25914597073)
        ref[2,0,0]=o*(0.758930908002)*x_ref[0]**(o-1)+(-0.137399611576)
        ref[2,0,1]=o*(-0.524730401698)*x_ref[1]**(o-1)+(-0.68595790282)
        ref[2,1,0]=o*(-0.956320437139)*x_ref[0]**(o-1)+(-0.741920593412)
        ref[2,1,1]=o*(0.470411496813)*x_ref[1]**(o-1)+(-0.912492685238)
        ref[2,2,0]=o*(0.386762534234)*x_ref[0]**(o-1)+(0.297029046975)
        ref[2,2,1]=o*(-0.999949621409)*x_ref[1]**(o-1)+(-0.273952756751)
        ref[2,3,0]=o*(-0.137635508969)*x_ref[0]**(o-1)+(0.0502825343025)
        ref[2,3,1]=o*(-0.856424864287)*x_ref[1]**(o-1)+(0.501432472431)
        ref[2,4,0]=o*(0.927974425929)*x_ref[0]**(o-1)+(-0.602904477506)
        ref[2,4,1]=o*(0.766583044002)*x_ref[1]**(o-1)+(0.070024577589)
        ref[3,0,0]=o*(0.336864984737)*x_ref[0]**(o-1)+(0.0974502498452)
        ref[3,0,1]=o*(-0.820387387723)*x_ref[1]**(o-1)+(0.022457213421)
        ref[3,1,0]=o*(0.879517107747)*x_ref[0]**(o-1)+(0.267883865556)
        ref[3,1,1]=o*(-0.795732666749)*x_ref[1]**(o-1)+(0.81701302209)
        ref[3,2,0]=o*(-0.971853752113)*x_ref[0]**(o-1)+(0.210482589767)
        ref[3,2,1]=o*(0.422058370831)*x_ref[1]**(o-1)+(-0.0945771197357)
        ref[3,3,0]=o*(0.0130393893213)*x_ref[0]**(o-1)+(0.331617264696)
        ref[3,3,1]=o*(-0.0460157855829)*x_ref[1]**(o-1)+(0.513357385084)
        ref[3,4,0]=o*(-0.466291714481)*x_ref[0]**(o-1)+(0.877178907345)
        ref[3,4,1]=o*(0.356001871079)*x_ref[1]**(o-1)+(-0.755177274311)
      else:
        s[0,0]=(-0.582848481634)*x[0]**o+(-0.799173040171)*x[0]+(-0.402728303622)*x[1]**o+(-0.294649773058)*x[1]+(-0.248183917161)*x[2]**o+(-0.182025430599)*x[2]
        s[0,1]=(0.421723448079)*x[0]**o+(-0.881083512259)*x[0]+(-0.411609580693)*x[1]**o+(-0.60866119666)*x[1]+(0.447246876403)*x[2]**o+(-0.566693550375)*x[2]
        s[0,2]=(0.676859700236)*x[0]**o+(-0.0721141912983)*x[0]+(-0.930202067587)*x[1]**o+(-0.779660291535)*x[1]+(-0.846576347271)*x[2]**o+(0.182815431369)*x[2]
        s[0,3]=(0.640898585384)*x[0]**o+(0.133026195724)*x[0]+(-0.264924898824)*x[1]**o+(0.984926461364)*x[1]+(-0.366473159006)*x[2]**o+(-0.893150501229)*x[2]
        s[0,4]=(-0.0617160479105)*x[0]**o+(-0.781163890754)*x[0]+(0.482406386837)*x[1]**o+(0.242984596118)*x[1]+(0.387455073331)*x[2]**o+(0.713852275793)*x[2]
        s[1,0]=(-0.261960056141)*x[0]**o+(0.289706307634)*x[0]+(0.33502315284)*x[1]**o+(0.373867353003)*x[1]+(0.988176700778)*x[2]**o+(0.821674109963)*x[2]
        s[1,1]=(-0.564538779546)*x[0]**o+(0.512720259636)*x[0]+(0.519794117449)*x[1]**o+(0.140310457627)*x[1]+(-0.987992481444)*x[2]**o+(-0.658619677691)*x[2]
        s[1,2]=(-0.94468459827)*x[0]**o+(-0.59503451145)*x[0]+(0.638928053728)*x[1]**o+(-0.426823410254)*x[1]+(0.698221019291)*x[2]**o+(0.584939961146)*x[2]
        s[1,3]=(-0.736608632072)*x[0]**o+(0.121237394007)*x[0]+(0.965644483539)*x[1]**o+(-0.69597048224)*x[1]+(-0.131814696207)*x[2]**o+(0.218156302554)*x[2]
        s[1,4]=(-0.68564146602)*x[0]**o+(-0.613073625086)*x[0]+(-0.480606344797)*x[1]**o+(0.305215204518)*x[1]+(-0.932537843561)*x[2]**o+(0.753998778964)*x[2]
        s[2,0]=(0.149366204787)*x[0]**o+(0.518985565336)*x[0]+(-0.108643244187)*x[1]**o+(-0.133769184691)*x[1]+(-0.91950647261)*x[2]**o+(-0.165713076313)*x[2]
        s[2,1]=(0.377574358299)*x[0]**o+(-0.881166371065)*x[0]+(0.859757091349)*x[1]**o+(-0.571419705867)*x[1]+(0.688875894062)*x[2]**o+(-0.475017852628)*x[2]
        s[2,2]=(-0.293556901343)*x[0]**o+(-0.21565327941)*x[0]+(-0.661622857503)*x[1]**o+(-0.822811400487)*x[1]+(0.773358448616)*x[2]**o+(0.687967954749)*x[2]
        s[2,3]=(-0.0678537895443)*x[0]**o+(-0.620817206574)*x[0]+(0.195444962807)*x[1]**o+(-0.802482456128)*x[1]+(0.839892027886)*x[2]**o+(0.53033825313)*x[2]
        s[2,4]=(-0.174133277232)*x[0]**o+(-0.11357878673)*x[0]+(0.896953765213)*x[1]**o+(-0.111411131996)*x[1]+(-0.911590351769)*x[2]**o+(-0.867631474956)*x[2]
        s[3,0]=(0.619809270826)*x[0]**o+(0.234416234461)*x[0]+(-0.265703869206)*x[1]**o+(-0.179792194117)*x[1]+(0.650878614932)*x[2]**o+(0.832933790635)*x[2]
        s[3,1]=(-0.464423972484)*x[0]**o+(-0.0792284263184)*x[0]+(0.591682473906)*x[1]**o+(-0.69612265297)*x[1]+(-0.847084735807)*x[2]**o+(0.406411159143)*x[2]
        s[3,2]=(-0.206502883044)*x[0]**o+(0.266143672088)*x[0]+(0.742487243777)*x[1]**o+(0.652888436702)*x[1]+(-0.638601136445)*x[2]**o+(-0.411202594715)*x[2]
        s[3,3]=(0.749860017366)*x[0]**o+(0.0712552207004)*x[0]+(0.932270746555)*x[1]**o+(-0.743243258158)*x[1]+(0.77742513611)*x[2]**o+(-0.0951940129286)*x[2]
        s[3,4]=(-0.736329018198)*x[0]**o+(-0.346737131331)*x[0]+(0.15155322637)*x[1]**o+(0.209410136584)*x[1]+(-0.283669919947)*x[2]**o+(-0.736677347286)*x[2]
        ref[0,0,0]=o*(-0.582848481634)*x_ref[0]**(o-1)+(-0.799173040171)
        ref[0,0,1]=o*(-0.402728303622)*x_ref[1]**(o-1)+(-0.294649773058)
        ref[0,0,2]=o*(-0.248183917161)*x_ref[2]**(o-1)+(-0.182025430599)
        ref[0,1,0]=o*(0.421723448079)*x_ref[0]**(o-1)+(-0.881083512259)
        ref[0,1,1]=o*(-0.411609580693)*x_ref[1]**(o-1)+(-0.60866119666)
        ref[0,1,2]=o*(0.447246876403)*x_ref[2]**(o-1)+(-0.566693550375)
        ref[0,2,0]=o*(0.676859700236)*x_ref[0]**(o-1)+(-0.0721141912983)
        ref[0,2,1]=o*(-0.930202067587)*x_ref[1]**(o-1)+(-0.779660291535)
        ref[0,2,2]=o*(-0.846576347271)*x_ref[2]**(o-1)+(0.182815431369)
        ref[0,3,0]=o*(0.640898585384)*x_ref[0]**(o-1)+(0.133026195724)
        ref[0,3,1]=o*(-0.264924898824)*x_ref[1]**(o-1)+(0.984926461364)
        ref[0,3,2]=o*(-0.366473159006)*x_ref[2]**(o-1)+(-0.893150501229)
        ref[0,4,0]=o*(-0.0617160479105)*x_ref[0]**(o-1)+(-0.781163890754)
        ref[0,4,1]=o*(0.482406386837)*x_ref[1]**(o-1)+(0.242984596118)
        ref[0,4,2]=o*(0.387455073331)*x_ref[2]**(o-1)+(0.713852275793)
        ref[1,0,0]=o*(-0.261960056141)*x_ref[0]**(o-1)+(0.289706307634)
        ref[1,0,1]=o*(0.33502315284)*x_ref[1]**(o-1)+(0.373867353003)
        ref[1,0,2]=o*(0.988176700778)*x_ref[2]**(o-1)+(0.821674109963)
        ref[1,1,0]=o*(-0.564538779546)*x_ref[0]**(o-1)+(0.512720259636)
        ref[1,1,1]=o*(0.519794117449)*x_ref[1]**(o-1)+(0.140310457627)
        ref[1,1,2]=o*(-0.987992481444)*x_ref[2]**(o-1)+(-0.658619677691)
        ref[1,2,0]=o*(-0.94468459827)*x_ref[0]**(o-1)+(-0.59503451145)
        ref[1,2,1]=o*(0.638928053728)*x_ref[1]**(o-1)+(-0.426823410254)
        ref[1,2,2]=o*(0.698221019291)*x_ref[2]**(o-1)+(0.584939961146)
        ref[1,3,0]=o*(-0.736608632072)*x_ref[0]**(o-1)+(0.121237394007)
        ref[1,3,1]=o*(0.965644483539)*x_ref[1]**(o-1)+(-0.69597048224)
        ref[1,3,2]=o*(-0.131814696207)*x_ref[2]**(o-1)+(0.218156302554)
        ref[1,4,0]=o*(-0.68564146602)*x_ref[0]**(o-1)+(-0.613073625086)
        ref[1,4,1]=o*(-0.480606344797)*x_ref[1]**(o-1)+(0.305215204518)
        ref[1,4,2]=o*(-0.932537843561)*x_ref[2]**(o-1)+(0.753998778964)
        ref[2,0,0]=o*(0.149366204787)*x_ref[0]**(o-1)+(0.518985565336)
        ref[2,0,1]=o*(-0.108643244187)*x_ref[1]**(o-1)+(-0.133769184691)
        ref[2,0,2]=o*(-0.91950647261)*x_ref[2]**(o-1)+(-0.165713076313)
        ref[2,1,0]=o*(0.377574358299)*x_ref[0]**(o-1)+(-0.881166371065)
        ref[2,1,1]=o*(0.859757091349)*x_ref[1]**(o-1)+(-0.571419705867)
        ref[2,1,2]=o*(0.688875894062)*x_ref[2]**(o-1)+(-0.475017852628)
        ref[2,2,0]=o*(-0.293556901343)*x_ref[0]**(o-1)+(-0.21565327941)
        ref[2,2,1]=o*(-0.661622857503)*x_ref[1]**(o-1)+(-0.822811400487)
        ref[2,2,2]=o*(0.773358448616)*x_ref[2]**(o-1)+(0.687967954749)
        ref[2,3,0]=o*(-0.0678537895443)*x_ref[0]**(o-1)+(-0.620817206574)
        ref[2,3,1]=o*(0.195444962807)*x_ref[1]**(o-1)+(-0.802482456128)
        ref[2,3,2]=o*(0.839892027886)*x_ref[2]**(o-1)+(0.53033825313)
        ref[2,4,0]=o*(-0.174133277232)*x_ref[0]**(o-1)+(-0.11357878673)
        ref[2,4,1]=o*(0.896953765213)*x_ref[1]**(o-1)+(-0.111411131996)
        ref[2,4,2]=o*(-0.911590351769)*x_ref[2]**(o-1)+(-0.867631474956)
        ref[3,0,0]=o*(0.619809270826)*x_ref[0]**(o-1)+(0.234416234461)
        ref[3,0,1]=o*(-0.265703869206)*x_ref[1]**(o-1)+(-0.179792194117)
        ref[3,0,2]=o*(0.650878614932)*x_ref[2]**(o-1)+(0.832933790635)
        ref[3,1,0]=o*(-0.464423972484)*x_ref[0]**(o-1)+(-0.0792284263184)
        ref[3,1,1]=o*(0.591682473906)*x_ref[1]**(o-1)+(-0.69612265297)
        ref[3,1,2]=o*(-0.847084735807)*x_ref[2]**(o-1)+(0.406411159143)
        ref[3,2,0]=o*(-0.206502883044)*x_ref[0]**(o-1)+(0.266143672088)
        ref[3,2,1]=o*(0.742487243777)*x_ref[1]**(o-1)+(0.652888436702)
        ref[3,2,2]=o*(-0.638601136445)*x_ref[2]**(o-1)+(-0.411202594715)
        ref[3,3,0]=o*(0.749860017366)*x_ref[0]**(o-1)+(0.0712552207004)
        ref[3,3,1]=o*(0.932270746555)*x_ref[1]**(o-1)+(-0.743243258158)
        ref[3,3,2]=o*(0.77742513611)*x_ref[2]**(o-1)+(-0.0951940129286)
        ref[3,4,0]=o*(-0.736329018198)*x_ref[0]**(o-1)+(-0.346737131331)
        ref[3,4,1]=o*(0.15155322637)*x_ref[1]**(o-1)+(0.209410136584)
        ref[3,4,2]=o*(-0.283669919947)*x_ref[2]**(o-1)+(-0.736677347286)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_Solution_rank3(self):
      """
      tests gradient for rank 3 Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(0.285489768558)*x[0]**o+(-0.814169283089)*x[0]+(-0.146214432938)*x[1]**o+(0.971176476923)*x[1]
        s[0,0,1]=(0.796197012783)*x[0]**o+(0.138854173842)*x[0]+(0.911141134854)*x[1]**o+(0.24230493445)*x[1]
        s[0,1,0]=(0.335306616869)*x[0]**o+(-0.346373461144)*x[0]+(0.485198289812)*x[1]**o+(-0.557460170056)*x[1]
        s[0,1,1]=(-0.202832958777)*x[0]**o+(0.202612852038)*x[0]+(-0.382959990066)*x[1]**o+(0.982713356847)*x[1]
        s[1,0,0]=(-0.732852201528)*x[0]**o+(-0.52380849964)*x[0]+(-0.423867496739)*x[1]**o+(-0.476139964916)*x[1]
        s[1,0,1]=(0.730537196604)*x[0]**o+(0.782053455575)*x[0]+(-0.496379081802)*x[1]**o+(-0.610796298113)*x[1]
        s[1,1,0]=(-0.518151715463)*x[0]**o+(0.466253389588)*x[0]+(0.792730016645)*x[1]**o+(-0.0928866491573)*x[1]
        s[1,1,1]=(0.721737904289)*x[0]**o+(0.662197543309)*x[0]+(0.402567060485)*x[1]**o+(-0.0537539240196)*x[1]
        s[2,0,0]=(0.264010742903)*x[0]**o+(-0.209394534976)*x[0]+(0.901338534988)*x[1]**o+(0.751548611498)*x[1]
        s[2,0,1]=(0.255790620214)*x[0]**o+(0.979495958037)*x[0]+(-0.00862935506386)*x[1]**o+(0.472608624671)*x[1]
        s[2,1,0]=(0.348157499697)*x[0]**o+(-0.489084366673)*x[0]+(-0.797020984331)*x[1]**o+(0.850642513227)*x[1]
        s[2,1,1]=(0.619156982334)*x[0]**o+(-0.871389250139)*x[0]+(0.838613332974)*x[1]**o+(0.865428066011)*x[1]
        s[3,0,0]=(-0.498572545085)*x[0]**o+(0.26752755104)*x[0]+(-0.742506202612)*x[1]**o+(-0.867811055537)*x[1]
        s[3,0,1]=(-0.469062081737)*x[0]**o+(-0.896231281785)*x[0]+(-0.0537944852375)*x[1]**o+(-0.566775495197)*x[1]
        s[3,1,0]=(0.960466927125)*x[0]**o+(-0.790015287315)*x[0]+(-0.249364909212)*x[1]**o+(0.55000656331)*x[1]
        s[3,1,1]=(0.107190484709)*x[0]**o+(0.206656185528)*x[0]+(0.445355269686)*x[1]**o+(0.177791381802)*x[1]
        s[4,0,0]=(0.926353234834)*x[0]**o+(0.0187238838755)*x[0]+(-0.0335895202391)*x[1]**o+(0.396027343683)*x[1]
        s[4,0,1]=(-0.122883815734)*x[0]**o+(0.114014252654)*x[0]+(0.390958937743)*x[1]**o+(0.743802098345)*x[1]
        s[4,1,0]=(0.0434615831789)*x[0]**o+(-0.0061546333485)*x[0]+(-0.941682112962)*x[1]**o+(0.546094645223)*x[1]
        s[4,1,1]=(-0.876438895739)*x[0]**o+(-0.307842960731)*x[0]+(-0.958906847802)*x[1]**o+(-0.698921900635)*x[1]
        s[5,0,0]=(-0.0294848627519)*x[0]**o+(-0.108277485049)*x[0]+(0.653184058519)*x[1]**o+(0.268533892275)*x[1]
        s[5,0,1]=(0.0643758658725)*x[0]**o+(0.403982947726)*x[0]+(-0.966297836476)*x[1]**o+(-0.934899934913)*x[1]
        s[5,1,0]=(-0.354834560075)*x[0]**o+(-0.641982491187)*x[0]+(0.231470813627)*x[1]**o+(0.597773058655)*x[1]
        s[5,1,1]=(0.457345532548)*x[0]**o+(0.930909964958)*x[0]+(-0.190821688076)*x[1]**o+(0.763332121356)*x[1]
        ref[0,0,0,0]=o*(0.285489768558)*x_ref[0]**(o-1)+(-0.814169283089)
        ref[0,0,0,1]=o*(-0.146214432938)*x_ref[1]**(o-1)+(0.971176476923)
        ref[0,0,1,0]=o*(0.796197012783)*x_ref[0]**(o-1)+(0.138854173842)
        ref[0,0,1,1]=o*(0.911141134854)*x_ref[1]**(o-1)+(0.24230493445)
        ref[0,1,0,0]=o*(0.335306616869)*x_ref[0]**(o-1)+(-0.346373461144)
        ref[0,1,0,1]=o*(0.485198289812)*x_ref[1]**(o-1)+(-0.557460170056)
        ref[0,1,1,0]=o*(-0.202832958777)*x_ref[0]**(o-1)+(0.202612852038)
        ref[0,1,1,1]=o*(-0.382959990066)*x_ref[1]**(o-1)+(0.982713356847)
        ref[1,0,0,0]=o*(-0.732852201528)*x_ref[0]**(o-1)+(-0.52380849964)
        ref[1,0,0,1]=o*(-0.423867496739)*x_ref[1]**(o-1)+(-0.476139964916)
        ref[1,0,1,0]=o*(0.730537196604)*x_ref[0]**(o-1)+(0.782053455575)
        ref[1,0,1,1]=o*(-0.496379081802)*x_ref[1]**(o-1)+(-0.610796298113)
        ref[1,1,0,0]=o*(-0.518151715463)*x_ref[0]**(o-1)+(0.466253389588)
        ref[1,1,0,1]=o*(0.792730016645)*x_ref[1]**(o-1)+(-0.0928866491573)
        ref[1,1,1,0]=o*(0.721737904289)*x_ref[0]**(o-1)+(0.662197543309)
        ref[1,1,1,1]=o*(0.402567060485)*x_ref[1]**(o-1)+(-0.0537539240196)
        ref[2,0,0,0]=o*(0.264010742903)*x_ref[0]**(o-1)+(-0.209394534976)
        ref[2,0,0,1]=o*(0.901338534988)*x_ref[1]**(o-1)+(0.751548611498)
        ref[2,0,1,0]=o*(0.255790620214)*x_ref[0]**(o-1)+(0.979495958037)
        ref[2,0,1,1]=o*(-0.00862935506386)*x_ref[1]**(o-1)+(0.472608624671)
        ref[2,1,0,0]=o*(0.348157499697)*x_ref[0]**(o-1)+(-0.489084366673)
        ref[2,1,0,1]=o*(-0.797020984331)*x_ref[1]**(o-1)+(0.850642513227)
        ref[2,1,1,0]=o*(0.619156982334)*x_ref[0]**(o-1)+(-0.871389250139)
        ref[2,1,1,1]=o*(0.838613332974)*x_ref[1]**(o-1)+(0.865428066011)
        ref[3,0,0,0]=o*(-0.498572545085)*x_ref[0]**(o-1)+(0.26752755104)
        ref[3,0,0,1]=o*(-0.742506202612)*x_ref[1]**(o-1)+(-0.867811055537)
        ref[3,0,1,0]=o*(-0.469062081737)*x_ref[0]**(o-1)+(-0.896231281785)
        ref[3,0,1,1]=o*(-0.0537944852375)*x_ref[1]**(o-1)+(-0.566775495197)
        ref[3,1,0,0]=o*(0.960466927125)*x_ref[0]**(o-1)+(-0.790015287315)
        ref[3,1,0,1]=o*(-0.249364909212)*x_ref[1]**(o-1)+(0.55000656331)
        ref[3,1,1,0]=o*(0.107190484709)*x_ref[0]**(o-1)+(0.206656185528)
        ref[3,1,1,1]=o*(0.445355269686)*x_ref[1]**(o-1)+(0.177791381802)
        ref[4,0,0,0]=o*(0.926353234834)*x_ref[0]**(o-1)+(0.0187238838755)
        ref[4,0,0,1]=o*(-0.0335895202391)*x_ref[1]**(o-1)+(0.396027343683)
        ref[4,0,1,0]=o*(-0.122883815734)*x_ref[0]**(o-1)+(0.114014252654)
        ref[4,0,1,1]=o*(0.390958937743)*x_ref[1]**(o-1)+(0.743802098345)
        ref[4,1,0,0]=o*(0.0434615831789)*x_ref[0]**(o-1)+(-0.0061546333485)
        ref[4,1,0,1]=o*(-0.941682112962)*x_ref[1]**(o-1)+(0.546094645223)
        ref[4,1,1,0]=o*(-0.876438895739)*x_ref[0]**(o-1)+(-0.307842960731)
        ref[4,1,1,1]=o*(-0.958906847802)*x_ref[1]**(o-1)+(-0.698921900635)
        ref[5,0,0,0]=o*(-0.0294848627519)*x_ref[0]**(o-1)+(-0.108277485049)
        ref[5,0,0,1]=o*(0.653184058519)*x_ref[1]**(o-1)+(0.268533892275)
        ref[5,0,1,0]=o*(0.0643758658725)*x_ref[0]**(o-1)+(0.403982947726)
        ref[5,0,1,1]=o*(-0.966297836476)*x_ref[1]**(o-1)+(-0.934899934913)
        ref[5,1,0,0]=o*(-0.354834560075)*x_ref[0]**(o-1)+(-0.641982491187)
        ref[5,1,0,1]=o*(0.231470813627)*x_ref[1]**(o-1)+(0.597773058655)
        ref[5,1,1,0]=o*(0.457345532548)*x_ref[0]**(o-1)+(0.930909964958)
        ref[5,1,1,1]=o*(-0.190821688076)*x_ref[1]**(o-1)+(0.763332121356)
      else:
        s[0,0,0]=(0.0799106193628)*x[0]**o+(-0.385218529142)*x[0]+(0.423551372182)*x[1]**o+(-0.357661802001)*x[1]+(0.523815428088)*x[2]**o+(-0.867664901871)*x[2]
        s[0,0,1]=(-0.223298613593)*x[0]**o+(-0.563675059615)*x[0]+(0.854553636152)*x[1]**o+(0.00619936897032)*x[1]+(-0.168119043675)*x[2]**o+(-0.959967076262)*x[2]
        s[0,1,0]=(-0.538044222292)*x[0]**o+(0.519685043014)*x[0]+(-0.351144688067)*x[1]**o+(0.887365493845)*x[1]+(0.801566727487)*x[2]**o+(0.571280510487)*x[2]
        s[0,1,1]=(0.784344723309)*x[0]**o+(-0.0907512562255)*x[0]+(-0.24324080888)*x[1]**o+(-0.199412485507)*x[1]+(0.642344630349)*x[2]**o+(-0.735616137201)*x[2]
        s[1,0,0]=(-0.880206903193)*x[0]**o+(-0.749123123033)*x[0]+(0.104561575813)*x[1]**o+(-0.647441987473)*x[1]+(0.182201321281)*x[2]**o+(0.0102973222322)*x[2]
        s[1,0,1]=(0.915893198392)*x[0]**o+(-0.612952390562)*x[0]+(0.803094622661)*x[1]**o+(-0.140410725667)*x[1]+(-0.247855770678)*x[2]**o+(0.399254178737)*x[2]
        s[1,1,0]=(0.818321238866)*x[0]**o+(0.214893239645)*x[0]+(0.909338463913)*x[1]**o+(0.855186366612)*x[1]+(-0.835375145102)*x[2]**o+(-0.807936215386)*x[2]
        s[1,1,1]=(0.257588723885)*x[0]**o+(-0.952059092336)*x[0]+(-0.28964814303)*x[1]**o+(0.0129418032921)*x[1]+(0.854165719561)*x[2]**o+(0.16196867256)*x[2]
        s[2,0,0]=(-0.291805757446)*x[0]**o+(0.609410741859)*x[0]+(0.541295607728)*x[1]**o+(-0.220049109439)*x[1]+(0.773352330852)*x[2]**o+(-0.0999084978714)*x[2]
        s[2,0,1]=(-0.094477913729)*x[0]**o+(-0.764264730185)*x[0]+(-0.671804542061)*x[1]**o+(0.368646822044)*x[1]+(0.448237271807)*x[2]**o+(0.565106260459)*x[2]
        s[2,1,0]=(-0.240257182637)*x[0]**o+(0.859112706839)*x[0]+(-0.575504060232)*x[1]**o+(-0.611929236978)*x[1]+(0.675448692587)*x[2]**o+(0.31890080218)*x[2]
        s[2,1,1]=(-0.39222652102)*x[0]**o+(0.677458518423)*x[0]+(-0.412612361541)*x[1]**o+(-0.374686862277)*x[1]+(-0.737213784962)*x[2]**o+(-0.486920606866)*x[2]
        s[3,0,0]=(0.635190942363)*x[0]**o+(-0.330731024263)*x[0]+(-0.690112375174)*x[1]**o+(-0.446491601534)*x[1]+(-0.748307399262)*x[2]**o+(-0.578747596175)*x[2]
        s[3,0,1]=(-0.419848395854)*x[0]**o+(-0.31014199789)*x[0]+(0.818235061383)*x[1]**o+(0.0547944058433)*x[1]+(0.855573639674)*x[2]**o+(-0.740418501951)*x[2]
        s[3,1,0]=(-0.0448849811136)*x[0]**o+(-0.337048179302)*x[0]+(-0.504885976058)*x[1]**o+(-0.445450937517)*x[1]+(0.567028111275)*x[2]**o+(0.123980108452)*x[2]
        s[3,1,1]=(-0.438038042566)*x[0]**o+(-0.427397249753)*x[0]+(0.574195396827)*x[1]**o+(0.824799756958)*x[1]+(-0.886957824366)*x[2]**o+(-0.192557581967)*x[2]
        s[4,0,0]=(0.961581428919)*x[0]**o+(-0.666268681241)*x[0]+(0.916527151658)*x[1]**o+(0.879372699554)*x[1]+(-0.648140763963)*x[2]**o+(-0.692256048468)*x[2]
        s[4,0,1]=(-0.162077276705)*x[0]**o+(-0.306833194309)*x[0]+(0.611399808831)*x[1]**o+(-0.508477663404)*x[1]+(-0.380712354501)*x[2]**o+(0.490458516352)*x[2]
        s[4,1,0]=(-0.64539443571)*x[0]**o+(-0.570490580768)*x[0]+(-0.612017756646)*x[1]**o+(0.979359105468)*x[1]+(0.901671131984)*x[2]**o+(-0.57388809992)*x[2]
        s[4,1,1]=(0.0239934065942)*x[0]**o+(0.358848963609)*x[0]+(-0.262596433728)*x[1]**o+(0.966612108153)*x[1]+(0.048144124831)*x[2]**o+(0.106585203917)*x[2]
        s[5,0,0]=(-0.894844992009)*x[0]**o+(-0.991143708359)*x[0]+(-0.719802159076)*x[1]**o+(0.609678198738)*x[1]+(0.934875984305)*x[2]**o+(0.210728029537)*x[2]
        s[5,0,1]=(-0.0669111311933)*x[0]**o+(-0.145495535442)*x[0]+(-0.988259077226)*x[1]**o+(-0.736349726987)*x[1]+(-0.164335654094)*x[2]**o+(-0.364166925171)*x[2]
        s[5,1,0]=(-0.279004883801)*x[0]**o+(0.96466483765)*x[0]+(-0.445599169347)*x[1]**o+(0.866248162976)*x[1]+(0.13706485262)*x[2]**o+(-0.78704568071)*x[2]
        s[5,1,1]=(0.452358514637)*x[0]**o+(-0.456298799376)*x[0]+(0.80484494772)*x[1]**o+(-0.949154901869)*x[1]+(-0.106819359297)*x[2]**o+(-0.710983814219)*x[2]
        ref[0,0,0,0]=o*(0.0799106193628)*x_ref[0]**(o-1)+(-0.385218529142)
        ref[0,0,0,1]=o*(0.423551372182)*x_ref[1]**(o-1)+(-0.357661802001)
        ref[0,0,0,2]=o*(0.523815428088)*x_ref[2]**(o-1)+(-0.867664901871)
        ref[0,0,1,0]=o*(-0.223298613593)*x_ref[0]**(o-1)+(-0.563675059615)
        ref[0,0,1,1]=o*(0.854553636152)*x_ref[1]**(o-1)+(0.00619936897032)
        ref[0,0,1,2]=o*(-0.168119043675)*x_ref[2]**(o-1)+(-0.959967076262)
        ref[0,1,0,0]=o*(-0.538044222292)*x_ref[0]**(o-1)+(0.519685043014)
        ref[0,1,0,1]=o*(-0.351144688067)*x_ref[1]**(o-1)+(0.887365493845)
        ref[0,1,0,2]=o*(0.801566727487)*x_ref[2]**(o-1)+(0.571280510487)
        ref[0,1,1,0]=o*(0.784344723309)*x_ref[0]**(o-1)+(-0.0907512562255)
        ref[0,1,1,1]=o*(-0.24324080888)*x_ref[1]**(o-1)+(-0.199412485507)
        ref[0,1,1,2]=o*(0.642344630349)*x_ref[2]**(o-1)+(-0.735616137201)
        ref[1,0,0,0]=o*(-0.880206903193)*x_ref[0]**(o-1)+(-0.749123123033)
        ref[1,0,0,1]=o*(0.104561575813)*x_ref[1]**(o-1)+(-0.647441987473)
        ref[1,0,0,2]=o*(0.182201321281)*x_ref[2]**(o-1)+(0.0102973222322)
        ref[1,0,1,0]=o*(0.915893198392)*x_ref[0]**(o-1)+(-0.612952390562)
        ref[1,0,1,1]=o*(0.803094622661)*x_ref[1]**(o-1)+(-0.140410725667)
        ref[1,0,1,2]=o*(-0.247855770678)*x_ref[2]**(o-1)+(0.399254178737)
        ref[1,1,0,0]=o*(0.818321238866)*x_ref[0]**(o-1)+(0.214893239645)
        ref[1,1,0,1]=o*(0.909338463913)*x_ref[1]**(o-1)+(0.855186366612)
        ref[1,1,0,2]=o*(-0.835375145102)*x_ref[2]**(o-1)+(-0.807936215386)
        ref[1,1,1,0]=o*(0.257588723885)*x_ref[0]**(o-1)+(-0.952059092336)
        ref[1,1,1,1]=o*(-0.28964814303)*x_ref[1]**(o-1)+(0.0129418032921)
        ref[1,1,1,2]=o*(0.854165719561)*x_ref[2]**(o-1)+(0.16196867256)
        ref[2,0,0,0]=o*(-0.291805757446)*x_ref[0]**(o-1)+(0.609410741859)
        ref[2,0,0,1]=o*(0.541295607728)*x_ref[1]**(o-1)+(-0.220049109439)
        ref[2,0,0,2]=o*(0.773352330852)*x_ref[2]**(o-1)+(-0.0999084978714)
        ref[2,0,1,0]=o*(-0.094477913729)*x_ref[0]**(o-1)+(-0.764264730185)
        ref[2,0,1,1]=o*(-0.671804542061)*x_ref[1]**(o-1)+(0.368646822044)
        ref[2,0,1,2]=o*(0.448237271807)*x_ref[2]**(o-1)+(0.565106260459)
        ref[2,1,0,0]=o*(-0.240257182637)*x_ref[0]**(o-1)+(0.859112706839)
        ref[2,1,0,1]=o*(-0.575504060232)*x_ref[1]**(o-1)+(-0.611929236978)
        ref[2,1,0,2]=o*(0.675448692587)*x_ref[2]**(o-1)+(0.31890080218)
        ref[2,1,1,0]=o*(-0.39222652102)*x_ref[0]**(o-1)+(0.677458518423)
        ref[2,1,1,1]=o*(-0.412612361541)*x_ref[1]**(o-1)+(-0.374686862277)
        ref[2,1,1,2]=o*(-0.737213784962)*x_ref[2]**(o-1)+(-0.486920606866)
        ref[3,0,0,0]=o*(0.635190942363)*x_ref[0]**(o-1)+(-0.330731024263)
        ref[3,0,0,1]=o*(-0.690112375174)*x_ref[1]**(o-1)+(-0.446491601534)
        ref[3,0,0,2]=o*(-0.748307399262)*x_ref[2]**(o-1)+(-0.578747596175)
        ref[3,0,1,0]=o*(-0.419848395854)*x_ref[0]**(o-1)+(-0.31014199789)
        ref[3,0,1,1]=o*(0.818235061383)*x_ref[1]**(o-1)+(0.0547944058433)
        ref[3,0,1,2]=o*(0.855573639674)*x_ref[2]**(o-1)+(-0.740418501951)
        ref[3,1,0,0]=o*(-0.0448849811136)*x_ref[0]**(o-1)+(-0.337048179302)
        ref[3,1,0,1]=o*(-0.504885976058)*x_ref[1]**(o-1)+(-0.445450937517)
        ref[3,1,0,2]=o*(0.567028111275)*x_ref[2]**(o-1)+(0.123980108452)
        ref[3,1,1,0]=o*(-0.438038042566)*x_ref[0]**(o-1)+(-0.427397249753)
        ref[3,1,1,1]=o*(0.574195396827)*x_ref[1]**(o-1)+(0.824799756958)
        ref[3,1,1,2]=o*(-0.886957824366)*x_ref[2]**(o-1)+(-0.192557581967)
        ref[4,0,0,0]=o*(0.961581428919)*x_ref[0]**(o-1)+(-0.666268681241)
        ref[4,0,0,1]=o*(0.916527151658)*x_ref[1]**(o-1)+(0.879372699554)
        ref[4,0,0,2]=o*(-0.648140763963)*x_ref[2]**(o-1)+(-0.692256048468)
        ref[4,0,1,0]=o*(-0.162077276705)*x_ref[0]**(o-1)+(-0.306833194309)
        ref[4,0,1,1]=o*(0.611399808831)*x_ref[1]**(o-1)+(-0.508477663404)
        ref[4,0,1,2]=o*(-0.380712354501)*x_ref[2]**(o-1)+(0.490458516352)
        ref[4,1,0,0]=o*(-0.64539443571)*x_ref[0]**(o-1)+(-0.570490580768)
        ref[4,1,0,1]=o*(-0.612017756646)*x_ref[1]**(o-1)+(0.979359105468)
        ref[4,1,0,2]=o*(0.901671131984)*x_ref[2]**(o-1)+(-0.57388809992)
        ref[4,1,1,0]=o*(0.0239934065942)*x_ref[0]**(o-1)+(0.358848963609)
        ref[4,1,1,1]=o*(-0.262596433728)*x_ref[1]**(o-1)+(0.966612108153)
        ref[4,1,1,2]=o*(0.048144124831)*x_ref[2]**(o-1)+(0.106585203917)
        ref[5,0,0,0]=o*(-0.894844992009)*x_ref[0]**(o-1)+(-0.991143708359)
        ref[5,0,0,1]=o*(-0.719802159076)*x_ref[1]**(o-1)+(0.609678198738)
        ref[5,0,0,2]=o*(0.934875984305)*x_ref[2]**(o-1)+(0.210728029537)
        ref[5,0,1,0]=o*(-0.0669111311933)*x_ref[0]**(o-1)+(-0.145495535442)
        ref[5,0,1,1]=o*(-0.988259077226)*x_ref[1]**(o-1)+(-0.736349726987)
        ref[5,0,1,2]=o*(-0.164335654094)*x_ref[2]**(o-1)+(-0.364166925171)
        ref[5,1,0,0]=o*(-0.279004883801)*x_ref[0]**(o-1)+(0.96466483765)
        ref[5,1,0,1]=o*(-0.445599169347)*x_ref[1]**(o-1)+(0.866248162976)
        ref[5,1,0,2]=o*(0.13706485262)*x_ref[2]**(o-1)+(-0.78704568071)
        ref[5,1,1,0]=o*(0.452358514637)*x_ref[0]**(o-1)+(-0.456298799376)
        ref[5,1,1,1]=o*(0.80484494772)*x_ref[1]**(o-1)+(-0.949154901869)
        ref[5,1,1,2]=o*(-0.106819359297)*x_ref[2]**(o-1)+(-0.710983814219)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests gradient for rank 0 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.495552973369)*x[0]+(0.874390589312)*x[1]
        ref[0]=(-0.495552973369)
        ref[1]=(0.874390589312)
      else:
        s=(-0.354636799848)*x[0]+(0.555572828684)*x[1]+(-0.270947199201)*x[2]
        ref[0]=(-0.354636799848)
        ref[1]=(0.555572828684)
        ref[2]=(-0.270947199201)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests gradient for rank 1 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(-0.187161756695)*x[0]+(-0.776766204086)*x[1]
        s[1]=(0.724774116602)*x[0]+(0.446775762856)*x[1]
        ref[0,0]=(-0.187161756695)
        ref[0,1]=(-0.776766204086)
        ref[1,0]=(0.724774116602)
        ref[1,1]=(0.446775762856)
      else:
        s[0]=(-0.0251530250542)*x[0]+(-1.23786881883)*x[1]+(1.1819624619)*x[2]
        s[1]=(0.801009266263)*x[0]+(-0.210512005637)*x[1]+(-1.42548669085)*x[2]
        ref[0,0]=(-0.0251530250542)
        ref[0,1]=(-1.23786881883)
        ref[0,2]=(1.1819624619)
        ref[1,0]=(0.801009266263)
        ref[1,1]=(-0.210512005637)
        ref[1,2]=(-1.42548669085)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests gradient for rank 2 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.105140628344)*x[0]+(0.681061911151)*x[1]
        s[0,1]=(1.12568473028)*x[0]+(-0.142806409661)*x[1]
        s[0,2]=(-0.709551703126)*x[0]+(0.470204858155)*x[1]
        s[0,3]=(-0.457889995297)*x[0]+(0.119831077838)*x[1]
        s[0,4]=(-0.137584694085)*x[0]+(-0.815364572262)*x[1]
        s[1,0]=(-0.747275494188)*x[0]+(0.419390416238)*x[1]
        s[1,1]=(-0.55554960963)*x[0]+(1.68766532296)*x[1]
        s[1,2]=(0.0380274635555)*x[0]+(0.903518514201)*x[1]
        s[1,3]=(1.30478047839)*x[0]+(0.576658405128)*x[1]
        s[1,4]=(-0.786884541126)*x[0]+(0.481365349672)*x[1]
        s[2,0]=(1.30445628093)*x[0]+(1.06765714413)*x[1]
        s[2,1]=(0.7269884164)*x[0]+(-1.34914325614)*x[1]
        s[2,2]=(-0.827940783305)*x[0]+(1.05409369275)*x[1]
        s[2,3]=(0.57117918118)*x[0]+(-0.857074951332)*x[1]
        s[2,4]=(0.808649779643)*x[0]+(-0.746017104416)*x[1]
        s[3,0]=(1.06443222736)*x[0]+(-0.458402762622)*x[1]
        s[3,1]=(-0.328990411524)*x[0]+(1.10697234223)*x[1]
        s[3,2]=(-0.0777059446026)*x[0]+(1.04837766698)*x[1]
        s[3,3]=(0.0542117671967)*x[0]+(-1.24129430221)*x[1]
        s[3,4]=(0.176491870492)*x[0]+(0.310238631995)*x[1]
        ref[0,0,0]=(-0.105140628344)
        ref[0,0,1]=(0.681061911151)
        ref[0,1,0]=(1.12568473028)
        ref[0,1,1]=(-0.142806409661)
        ref[0,2,0]=(-0.709551703126)
        ref[0,2,1]=(0.470204858155)
        ref[0,3,0]=(-0.457889995297)
        ref[0,3,1]=(0.119831077838)
        ref[0,4,0]=(-0.137584694085)
        ref[0,4,1]=(-0.815364572262)
        ref[1,0,0]=(-0.747275494188)
        ref[1,0,1]=(0.419390416238)
        ref[1,1,0]=(-0.55554960963)
        ref[1,1,1]=(1.68766532296)
        ref[1,2,0]=(0.0380274635555)
        ref[1,2,1]=(0.903518514201)
        ref[1,3,0]=(1.30478047839)
        ref[1,3,1]=(0.576658405128)
        ref[1,4,0]=(-0.786884541126)
        ref[1,4,1]=(0.481365349672)
        ref[2,0,0]=(1.30445628093)
        ref[2,0,1]=(1.06765714413)
        ref[2,1,0]=(0.7269884164)
        ref[2,1,1]=(-1.34914325614)
        ref[2,2,0]=(-0.827940783305)
        ref[2,2,1]=(1.05409369275)
        ref[2,3,0]=(0.57117918118)
        ref[2,3,1]=(-0.857074951332)
        ref[2,4,0]=(0.808649779643)
        ref[2,4,1]=(-0.746017104416)
        ref[3,0,0]=(1.06443222736)
        ref[3,0,1]=(-0.458402762622)
        ref[3,1,0]=(-0.328990411524)
        ref[3,1,1]=(1.10697234223)
        ref[3,2,0]=(-0.0777059446026)
        ref[3,2,1]=(1.04837766698)
        ref[3,3,0]=(0.0542117671967)
        ref[3,3,1]=(-1.24129430221)
        ref[3,4,0]=(0.176491870492)
        ref[3,4,1]=(0.310238631995)
      else:
        s[0,0]=(-0.835695322993)*x[0]+(0.168930135561)*x[1]+(-1.47180800481)*x[2]
        s[0,1]=(0.2001836256)*x[0]+(1.27356263599)*x[1]+(1.07485276442)*x[2]
        s[0,2]=(-0.178446572149)*x[0]+(-0.420078906939)*x[1]+(0.535633192895)*x[2]
        s[0,3]=(-0.0910549937921)*x[0]+(0.594316242371)*x[1]+(-1.06833005008)*x[2]
        s[0,4]=(0.722857786921)*x[0]+(-0.332473817629)*x[1]+(0.79883607252)*x[2]
        s[1,0]=(-1.02192468444)*x[0]+(1.2735296387)*x[1]+(-1.34574119314)*x[2]
        s[1,1]=(-1.20222077943)*x[0]+(0.180006226359)*x[1]+(0.582580396324)*x[2]
        s[1,2]=(0.292773497866)*x[0]+(0.503071430702)*x[1]+(1.25357442113)*x[2]
        s[1,3]=(-0.579871204353)*x[0]+(0.143994355108)*x[1]+(0.50466279341)*x[2]
        s[1,4]=(1.00530781493)*x[0]+(-0.646038982555)*x[1]+(0.315868158577)*x[2]
        s[2,0]=(0.0109517370271)*x[0]+(-1.27094713044)*x[1]+(0.647777134752)*x[2]
        s[2,1]=(0.615504753112)*x[0]+(0.479866346157)*x[1]+(1.72115593961)*x[2]
        s[2,2]=(-0.307778323155)*x[0]+(0.307520362533)*x[1]+(0.0061507801029)*x[2]
        s[2,3]=(-0.901393000596)*x[0]+(0.912379707397)*x[1]+(-0.0683825966669)*x[2]
        s[2,4]=(0.558642962243)*x[0]+(0.650906862857)*x[1]+(-1.44796931057)*x[2]
        s[3,0]=(1.08498680281)*x[0]+(-1.28372501216)*x[1]+(-1.27799790624)*x[2]
        s[3,1]=(0.45408516398)*x[0]+(-0.950068476479)*x[1]+(0.634133482831)*x[2]
        s[3,2]=(0.824015492184)*x[0]+(0.148535318857)*x[1]+(0.10609796765)*x[2]
        s[3,3]=(1.14342153343)*x[0]+(-0.0689548622931)*x[1]+(-0.166489981522)*x[2]
        s[3,4]=(0.995987765181)*x[0]+(0.915359249488)*x[1]+(-0.583052373261)*x[2]
        ref[0,0,0]=(-0.835695322993)
        ref[0,0,1]=(0.168930135561)
        ref[0,0,2]=(-1.47180800481)
        ref[0,1,0]=(0.2001836256)
        ref[0,1,1]=(1.27356263599)
        ref[0,1,2]=(1.07485276442)
        ref[0,2,0]=(-0.178446572149)
        ref[0,2,1]=(-0.420078906939)
        ref[0,2,2]=(0.535633192895)
        ref[0,3,0]=(-0.0910549937921)
        ref[0,3,1]=(0.594316242371)
        ref[0,3,2]=(-1.06833005008)
        ref[0,4,0]=(0.722857786921)
        ref[0,4,1]=(-0.332473817629)
        ref[0,4,2]=(0.79883607252)
        ref[1,0,0]=(-1.02192468444)
        ref[1,0,1]=(1.2735296387)
        ref[1,0,2]=(-1.34574119314)
        ref[1,1,0]=(-1.20222077943)
        ref[1,1,1]=(0.180006226359)
        ref[1,1,2]=(0.582580396324)
        ref[1,2,0]=(0.292773497866)
        ref[1,2,1]=(0.503071430702)
        ref[1,2,2]=(1.25357442113)
        ref[1,3,0]=(-0.579871204353)
        ref[1,3,1]=(0.143994355108)
        ref[1,3,2]=(0.50466279341)
        ref[1,4,0]=(1.00530781493)
        ref[1,4,1]=(-0.646038982555)
        ref[1,4,2]=(0.315868158577)
        ref[2,0,0]=(0.0109517370271)
        ref[2,0,1]=(-1.27094713044)
        ref[2,0,2]=(0.647777134752)
        ref[2,1,0]=(0.615504753112)
        ref[2,1,1]=(0.479866346157)
        ref[2,1,2]=(1.72115593961)
        ref[2,2,0]=(-0.307778323155)
        ref[2,2,1]=(0.307520362533)
        ref[2,2,2]=(0.0061507801029)
        ref[2,3,0]=(-0.901393000596)
        ref[2,3,1]=(0.912379707397)
        ref[2,3,2]=(-0.0683825966669)
        ref[2,4,0]=(0.558642962243)
        ref[2,4,1]=(0.650906862857)
        ref[2,4,2]=(-1.44796931057)
        ref[3,0,0]=(1.08498680281)
        ref[3,0,1]=(-1.28372501216)
        ref[3,0,2]=(-1.27799790624)
        ref[3,1,0]=(0.45408516398)
        ref[3,1,1]=(-0.950068476479)
        ref[3,1,2]=(0.634133482831)
        ref[3,2,0]=(0.824015492184)
        ref[3,2,1]=(0.148535318857)
        ref[3,2,2]=(0.10609796765)
        ref[3,3,0]=(1.14342153343)
        ref[3,3,1]=(-0.0689548622931)
        ref[3,3,2]=(-0.166489981522)
        ref[3,4,0]=(0.995987765181)
        ref[3,4,1]=(0.915359249488)
        ref[3,4,2]=(-0.583052373261)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests gradient for rank 3 Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(0.116652834373)*x[0]+(1.4681512344)*x[1]
        s[0,0,1]=(1.89418390988)*x[0]+(-0.548065006619)*x[1]
        s[0,1,0]=(-0.07602392145)*x[0]+(0.0796515016952)*x[1]
        s[0,1,1]=(1.50333606195)*x[0]+(-0.211673198103)*x[1]
        s[1,0,0]=(0.172419295955)*x[0]+(-0.586603468045)*x[1]
        s[1,0,1]=(0.212952657549)*x[0]+(-0.100672655087)*x[1]
        s[1,1,0]=(-1.11939406352)*x[0]+(-0.589882254225)*x[1]
        s[1,1,1]=(0.318138253303)*x[0]+(0.409850277983)*x[1]
        s[2,0,0]=(-0.0669759465557)*x[0]+(0.525669761096)*x[1]
        s[2,0,1]=(0.349019390883)*x[0]+(0.154604501057)*x[1]
        s[2,1,0]=(1.95277106546)*x[0]+(0.495701766483)*x[1]
        s[2,1,1]=(-0.0986210008015)*x[0]+(1.12636829398)*x[1]
        s[3,0,0]=(0.971837535878)*x[0]+(-0.738253595188)*x[1]
        s[3,0,1]=(0.0314513203112)*x[0]+(-1.16646217783)*x[1]
        s[3,1,0]=(0.485751127243)*x[0]+(-0.340178033471)*x[1]
        s[3,1,1]=(0.408742867205)*x[0]+(0.151827789426)*x[1]
        s[4,0,0]=(-0.0134678971538)*x[0]+(0.884924133141)*x[1]
        s[4,0,1]=(-0.66899222691)*x[0]+(-0.959185584071)*x[1]
        s[4,1,0]=(1.05311767886)*x[0]+(0.64093792324)*x[1]
        s[4,1,1]=(0.562875594668)*x[0]+(0.539525481691)*x[1]
        s[5,0,0]=(0.310908751279)*x[0]+(0.497074267163)*x[1]
        s[5,0,1]=(-0.804186513465)*x[0]+(0.329762438933)*x[1]
        s[5,1,0]=(-0.946808260459)*x[0]+(-0.00260213368693)*x[1]
        s[5,1,1]=(0.63415870202)*x[0]+(1.38103994652)*x[1]
        ref[0,0,0,0]=(0.116652834373)
        ref[0,0,0,1]=(1.4681512344)
        ref[0,0,1,0]=(1.89418390988)
        ref[0,0,1,1]=(-0.548065006619)
        ref[0,1,0,0]=(-0.07602392145)
        ref[0,1,0,1]=(0.0796515016952)
        ref[0,1,1,0]=(1.50333606195)
        ref[0,1,1,1]=(-0.211673198103)
        ref[1,0,0,0]=(0.172419295955)
        ref[1,0,0,1]=(-0.586603468045)
        ref[1,0,1,0]=(0.212952657549)
        ref[1,0,1,1]=(-0.100672655087)
        ref[1,1,0,0]=(-1.11939406352)
        ref[1,1,0,1]=(-0.589882254225)
        ref[1,1,1,0]=(0.318138253303)
        ref[1,1,1,1]=(0.409850277983)
        ref[2,0,0,0]=(-0.0669759465557)
        ref[2,0,0,1]=(0.525669761096)
        ref[2,0,1,0]=(0.349019390883)
        ref[2,0,1,1]=(0.154604501057)
        ref[2,1,0,0]=(1.95277106546)
        ref[2,1,0,1]=(0.495701766483)
        ref[2,1,1,0]=(-0.0986210008015)
        ref[2,1,1,1]=(1.12636829398)
        ref[3,0,0,0]=(0.971837535878)
        ref[3,0,0,1]=(-0.738253595188)
        ref[3,0,1,0]=(0.0314513203112)
        ref[3,0,1,1]=(-1.16646217783)
        ref[3,1,0,0]=(0.485751127243)
        ref[3,1,0,1]=(-0.340178033471)
        ref[3,1,1,0]=(0.408742867205)
        ref[3,1,1,1]=(0.151827789426)
        ref[4,0,0,0]=(-0.0134678971538)
        ref[4,0,0,1]=(0.884924133141)
        ref[4,0,1,0]=(-0.66899222691)
        ref[4,0,1,1]=(-0.959185584071)
        ref[4,1,0,0]=(1.05311767886)
        ref[4,1,0,1]=(0.64093792324)
        ref[4,1,1,0]=(0.562875594668)
        ref[4,1,1,1]=(0.539525481691)
        ref[5,0,0,0]=(0.310908751279)
        ref[5,0,0,1]=(0.497074267163)
        ref[5,0,1,0]=(-0.804186513465)
        ref[5,0,1,1]=(0.329762438933)
        ref[5,1,0,0]=(-0.946808260459)
        ref[5,1,0,1]=(-0.00260213368693)
        ref[5,1,1,0]=(0.63415870202)
        ref[5,1,1,1]=(1.38103994652)
      else:
        s[0,0,0]=(-0.0840260391925)*x[0]+(0.385117859595)*x[1]+(1.36354425949)*x[2]
        s[0,0,1]=(0.682480367755)*x[0]+(-0.658980348414)*x[1]+(-0.129734588977)*x[2]
        s[0,1,0]=(-0.213790105123)*x[0]+(-0.774270821015)*x[1]+(0.160024784274)*x[2]
        s[0,1,1]=(-1.02270895286)*x[0]+(1.02607310825)*x[1]+(0.35056114816)*x[2]
        s[1,0,0]=(0.201792116067)*x[0]+(0.223564282304)*x[1]+(0.986785519545)*x[2]
        s[1,0,1]=(0.250819712775)*x[0]+(0.287643279417)*x[1]+(0.18295590528)*x[2]
        s[1,1,0]=(-1.47569225207)*x[0]+(-0.886933933801)*x[1]+(0.525380474082)*x[2]
        s[1,1,1]=(0.680401800239)*x[0]+(1.36155081028)*x[1]+(-1.76198171789)*x[2]
        s[2,0,0]=(-0.784439188011)*x[0]+(-0.797291895631)*x[1]+(-0.425550387518)*x[2]
        s[2,0,1]=(1.62016996989)*x[0]+(1.02824549576)*x[1]+(-0.157995046652)*x[2]
        s[2,1,0]=(-0.437412344413)*x[0]+(0.97005624264)*x[1]+(-0.321998550341)*x[2]
        s[2,1,1]=(0.473879660993)*x[0]+(0.483144823024)*x[1]+(0.168449443711)*x[2]
        s[3,0,0]=(-1.59822104736)*x[0]+(-1.24012649834)*x[1]+(0.695408081789)*x[2]
        s[3,0,1]=(0.827901715486)*x[0]+(-0.230621260402)*x[1]+(-0.789726238476)*x[2]
        s[3,1,0]=(-0.279064257268)*x[0]+(-1.35315417734)*x[1]+(0.593090655629)*x[2]
        s[3,1,1]=(0.387685237641)*x[0]+(1.60258880823)*x[1]+(-1.10088045432)*x[2]
        s[4,0,0]=(0.771374132686)*x[0]+(0.530393756307)*x[1]+(-0.366697715551)*x[2]
        s[4,0,1]=(-0.182374799863)*x[0]+(-0.125579433837)*x[1]+(-1.2028145445)*x[2]
        s[4,1,0]=(-0.208766654808)*x[0]+(0.645251941741)*x[1]+(-0.534068338041)*x[2]
        s[4,1,1]=(-0.064786779421)*x[0]+(0.0354509511326)*x[1]+(-1.32632049137)*x[2]
        s[5,0,0]=(0.833438445425)*x[0]+(0.770186199496)*x[1]+(0.112178380686)*x[2]
        s[5,0,1]=(1.36614714953)*x[0]+(-0.367243359207)*x[1]+(0.812867174908)*x[2]
        s[5,1,0]=(0.0359079603828)*x[0]+(0.622094474303)*x[1]+(-0.33137187739)*x[2]
        s[5,1,1]=(-0.479714378899)*x[0]+(-0.959098499695)*x[1]+(0.843889761479)*x[2]
        ref[0,0,0,0]=(-0.0840260391925)
        ref[0,0,0,1]=(0.385117859595)
        ref[0,0,0,2]=(1.36354425949)
        ref[0,0,1,0]=(0.682480367755)
        ref[0,0,1,1]=(-0.658980348414)
        ref[0,0,1,2]=(-0.129734588977)
        ref[0,1,0,0]=(-0.213790105123)
        ref[0,1,0,1]=(-0.774270821015)
        ref[0,1,0,2]=(0.160024784274)
        ref[0,1,1,0]=(-1.02270895286)
        ref[0,1,1,1]=(1.02607310825)
        ref[0,1,1,2]=(0.35056114816)
        ref[1,0,0,0]=(0.201792116067)
        ref[1,0,0,1]=(0.223564282304)
        ref[1,0,0,2]=(0.986785519545)
        ref[1,0,1,0]=(0.250819712775)
        ref[1,0,1,1]=(0.287643279417)
        ref[1,0,1,2]=(0.18295590528)
        ref[1,1,0,0]=(-1.47569225207)
        ref[1,1,0,1]=(-0.886933933801)
        ref[1,1,0,2]=(0.525380474082)
        ref[1,1,1,0]=(0.680401800239)
        ref[1,1,1,1]=(1.36155081028)
        ref[1,1,1,2]=(-1.76198171789)
        ref[2,0,0,0]=(-0.784439188011)
        ref[2,0,0,1]=(-0.797291895631)
        ref[2,0,0,2]=(-0.425550387518)
        ref[2,0,1,0]=(1.62016996989)
        ref[2,0,1,1]=(1.02824549576)
        ref[2,0,1,2]=(-0.157995046652)
        ref[2,1,0,0]=(-0.437412344413)
        ref[2,1,0,1]=(0.97005624264)
        ref[2,1,0,2]=(-0.321998550341)
        ref[2,1,1,0]=(0.473879660993)
        ref[2,1,1,1]=(0.483144823024)
        ref[2,1,1,2]=(0.168449443711)
        ref[3,0,0,0]=(-1.59822104736)
        ref[3,0,0,1]=(-1.24012649834)
        ref[3,0,0,2]=(0.695408081789)
        ref[3,0,1,0]=(0.827901715486)
        ref[3,0,1,1]=(-0.230621260402)
        ref[3,0,1,2]=(-0.789726238476)
        ref[3,1,0,0]=(-0.279064257268)
        ref[3,1,0,1]=(-1.35315417734)
        ref[3,1,0,2]=(0.593090655629)
        ref[3,1,1,0]=(0.387685237641)
        ref[3,1,1,1]=(1.60258880823)
        ref[3,1,1,2]=(-1.10088045432)
        ref[4,0,0,0]=(0.771374132686)
        ref[4,0,0,1]=(0.530393756307)
        ref[4,0,0,2]=(-0.366697715551)
        ref[4,0,1,0]=(-0.182374799863)
        ref[4,0,1,1]=(-0.125579433837)
        ref[4,0,1,2]=(-1.2028145445)
        ref[4,1,0,0]=(-0.208766654808)
        ref[4,1,0,1]=(0.645251941741)
        ref[4,1,0,2]=(-0.534068338041)
        ref[4,1,1,0]=(-0.064786779421)
        ref[4,1,1,1]=(0.0354509511326)
        ref[4,1,1,2]=(-1.32632049137)
        ref[5,0,0,0]=(0.833438445425)
        ref[5,0,0,1]=(0.770186199496)
        ref[5,0,0,2]=(0.112178380686)
        ref[5,0,1,0]=(1.36614714953)
        ref[5,0,1,1]=(-0.367243359207)
        ref[5,0,1,2]=(0.812867174908)
        ref[5,1,0,0]=(0.0359079603828)
        ref[5,1,0,1]=(0.622094474303)
        ref[5,1,0,2]=(-0.33137187739)
        ref[5,1,1,0]=(-0.479714378899)
        ref[5,1,1,1]=(-0.959098499695)
        ref[5,1,1,2]=(0.843889761479)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.495552973369)*x[0]+(0.874390589312)*x[1]
        ref[0]=(-0.495552973369)
        ref[1]=(0.874390589312)
      else:
        s=(-0.354636799848)*x[0]+(0.555572828684)*x[1]+(-0.270947199201)*x[2]
        ref[0]=(-0.354636799848)
        ref[1]=(0.555572828684)
        ref[2]=(-0.270947199201)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(-0.187161756695)*x[0]+(-0.776766204086)*x[1]
        s[1]=(0.724774116602)*x[0]+(0.446775762856)*x[1]
        ref[0,0]=(-0.187161756695)
        ref[0,1]=(-0.776766204086)
        ref[1,0]=(0.724774116602)
        ref[1,1]=(0.446775762856)
      else:
        s[0]=(-0.0251530250542)*x[0]+(-1.23786881883)*x[1]+(1.1819624619)*x[2]
        s[1]=(0.801009266263)*x[0]+(-0.210512005637)*x[1]+(-1.42548669085)*x[2]
        ref[0,0]=(-0.0251530250542)
        ref[0,1]=(-1.23786881883)
        ref[0,2]=(1.1819624619)
        ref[1,0]=(0.801009266263)
        ref[1,1]=(-0.210512005637)
        ref[1,2]=(-1.42548669085)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.105140628344)*x[0]+(0.681061911151)*x[1]
        s[0,1]=(1.12568473028)*x[0]+(-0.142806409661)*x[1]
        s[0,2]=(-0.709551703126)*x[0]+(0.470204858155)*x[1]
        s[0,3]=(-0.457889995297)*x[0]+(0.119831077838)*x[1]
        s[0,4]=(-0.137584694085)*x[0]+(-0.815364572262)*x[1]
        s[1,0]=(-0.747275494188)*x[0]+(0.419390416238)*x[1]
        s[1,1]=(-0.55554960963)*x[0]+(1.68766532296)*x[1]
        s[1,2]=(0.0380274635555)*x[0]+(0.903518514201)*x[1]
        s[1,3]=(1.30478047839)*x[0]+(0.576658405128)*x[1]
        s[1,4]=(-0.786884541126)*x[0]+(0.481365349672)*x[1]
        s[2,0]=(1.30445628093)*x[0]+(1.06765714413)*x[1]
        s[2,1]=(0.7269884164)*x[0]+(-1.34914325614)*x[1]
        s[2,2]=(-0.827940783305)*x[0]+(1.05409369275)*x[1]
        s[2,3]=(0.57117918118)*x[0]+(-0.857074951332)*x[1]
        s[2,4]=(0.808649779643)*x[0]+(-0.746017104416)*x[1]
        s[3,0]=(1.06443222736)*x[0]+(-0.458402762622)*x[1]
        s[3,1]=(-0.328990411524)*x[0]+(1.10697234223)*x[1]
        s[3,2]=(-0.0777059446026)*x[0]+(1.04837766698)*x[1]
        s[3,3]=(0.0542117671967)*x[0]+(-1.24129430221)*x[1]
        s[3,4]=(0.176491870492)*x[0]+(0.310238631995)*x[1]
        ref[0,0,0]=(-0.105140628344)
        ref[0,0,1]=(0.681061911151)
        ref[0,1,0]=(1.12568473028)
        ref[0,1,1]=(-0.142806409661)
        ref[0,2,0]=(-0.709551703126)
        ref[0,2,1]=(0.470204858155)
        ref[0,3,0]=(-0.457889995297)
        ref[0,3,1]=(0.119831077838)
        ref[0,4,0]=(-0.137584694085)
        ref[0,4,1]=(-0.815364572262)
        ref[1,0,0]=(-0.747275494188)
        ref[1,0,1]=(0.419390416238)
        ref[1,1,0]=(-0.55554960963)
        ref[1,1,1]=(1.68766532296)
        ref[1,2,0]=(0.0380274635555)
        ref[1,2,1]=(0.903518514201)
        ref[1,3,0]=(1.30478047839)
        ref[1,3,1]=(0.576658405128)
        ref[1,4,0]=(-0.786884541126)
        ref[1,4,1]=(0.481365349672)
        ref[2,0,0]=(1.30445628093)
        ref[2,0,1]=(1.06765714413)
        ref[2,1,0]=(0.7269884164)
        ref[2,1,1]=(-1.34914325614)
        ref[2,2,0]=(-0.827940783305)
        ref[2,2,1]=(1.05409369275)
        ref[2,3,0]=(0.57117918118)
        ref[2,3,1]=(-0.857074951332)
        ref[2,4,0]=(0.808649779643)
        ref[2,4,1]=(-0.746017104416)
        ref[3,0,0]=(1.06443222736)
        ref[3,0,1]=(-0.458402762622)
        ref[3,1,0]=(-0.328990411524)
        ref[3,1,1]=(1.10697234223)
        ref[3,2,0]=(-0.0777059446026)
        ref[3,2,1]=(1.04837766698)
        ref[3,3,0]=(0.0542117671967)
        ref[3,3,1]=(-1.24129430221)
        ref[3,4,0]=(0.176491870492)
        ref[3,4,1]=(0.310238631995)
      else:
        s[0,0]=(-0.835695322993)*x[0]+(0.168930135561)*x[1]+(-1.47180800481)*x[2]
        s[0,1]=(0.2001836256)*x[0]+(1.27356263599)*x[1]+(1.07485276442)*x[2]
        s[0,2]=(-0.178446572149)*x[0]+(-0.420078906939)*x[1]+(0.535633192895)*x[2]
        s[0,3]=(-0.0910549937921)*x[0]+(0.594316242371)*x[1]+(-1.06833005008)*x[2]
        s[0,4]=(0.722857786921)*x[0]+(-0.332473817629)*x[1]+(0.79883607252)*x[2]
        s[1,0]=(-1.02192468444)*x[0]+(1.2735296387)*x[1]+(-1.34574119314)*x[2]
        s[1,1]=(-1.20222077943)*x[0]+(0.180006226359)*x[1]+(0.582580396324)*x[2]
        s[1,2]=(0.292773497866)*x[0]+(0.503071430702)*x[1]+(1.25357442113)*x[2]
        s[1,3]=(-0.579871204353)*x[0]+(0.143994355108)*x[1]+(0.50466279341)*x[2]
        s[1,4]=(1.00530781493)*x[0]+(-0.646038982555)*x[1]+(0.315868158577)*x[2]
        s[2,0]=(0.0109517370271)*x[0]+(-1.27094713044)*x[1]+(0.647777134752)*x[2]
        s[2,1]=(0.615504753112)*x[0]+(0.479866346157)*x[1]+(1.72115593961)*x[2]
        s[2,2]=(-0.307778323155)*x[0]+(0.307520362533)*x[1]+(0.0061507801029)*x[2]
        s[2,3]=(-0.901393000596)*x[0]+(0.912379707397)*x[1]+(-0.0683825966669)*x[2]
        s[2,4]=(0.558642962243)*x[0]+(0.650906862857)*x[1]+(-1.44796931057)*x[2]
        s[3,0]=(1.08498680281)*x[0]+(-1.28372501216)*x[1]+(-1.27799790624)*x[2]
        s[3,1]=(0.45408516398)*x[0]+(-0.950068476479)*x[1]+(0.634133482831)*x[2]
        s[3,2]=(0.824015492184)*x[0]+(0.148535318857)*x[1]+(0.10609796765)*x[2]
        s[3,3]=(1.14342153343)*x[0]+(-0.0689548622931)*x[1]+(-0.166489981522)*x[2]
        s[3,4]=(0.995987765181)*x[0]+(0.915359249488)*x[1]+(-0.583052373261)*x[2]
        ref[0,0,0]=(-0.835695322993)
        ref[0,0,1]=(0.168930135561)
        ref[0,0,2]=(-1.47180800481)
        ref[0,1,0]=(0.2001836256)
        ref[0,1,1]=(1.27356263599)
        ref[0,1,2]=(1.07485276442)
        ref[0,2,0]=(-0.178446572149)
        ref[0,2,1]=(-0.420078906939)
        ref[0,2,2]=(0.535633192895)
        ref[0,3,0]=(-0.0910549937921)
        ref[0,3,1]=(0.594316242371)
        ref[0,3,2]=(-1.06833005008)
        ref[0,4,0]=(0.722857786921)
        ref[0,4,1]=(-0.332473817629)
        ref[0,4,2]=(0.79883607252)
        ref[1,0,0]=(-1.02192468444)
        ref[1,0,1]=(1.2735296387)
        ref[1,0,2]=(-1.34574119314)
        ref[1,1,0]=(-1.20222077943)
        ref[1,1,1]=(0.180006226359)
        ref[1,1,2]=(0.582580396324)
        ref[1,2,0]=(0.292773497866)
        ref[1,2,1]=(0.503071430702)
        ref[1,2,2]=(1.25357442113)
        ref[1,3,0]=(-0.579871204353)
        ref[1,3,1]=(0.143994355108)
        ref[1,3,2]=(0.50466279341)
        ref[1,4,0]=(1.00530781493)
        ref[1,4,1]=(-0.646038982555)
        ref[1,4,2]=(0.315868158577)
        ref[2,0,0]=(0.0109517370271)
        ref[2,0,1]=(-1.27094713044)
        ref[2,0,2]=(0.647777134752)
        ref[2,1,0]=(0.615504753112)
        ref[2,1,1]=(0.479866346157)
        ref[2,1,2]=(1.72115593961)
        ref[2,2,0]=(-0.307778323155)
        ref[2,2,1]=(0.307520362533)
        ref[2,2,2]=(0.0061507801029)
        ref[2,3,0]=(-0.901393000596)
        ref[2,3,1]=(0.912379707397)
        ref[2,3,2]=(-0.0683825966669)
        ref[2,4,0]=(0.558642962243)
        ref[2,4,1]=(0.650906862857)
        ref[2,4,2]=(-1.44796931057)
        ref[3,0,0]=(1.08498680281)
        ref[3,0,1]=(-1.28372501216)
        ref[3,0,2]=(-1.27799790624)
        ref[3,1,0]=(0.45408516398)
        ref[3,1,1]=(-0.950068476479)
        ref[3,1,2]=(0.634133482831)
        ref[3,2,0]=(0.824015492184)
        ref[3,2,1]=(0.148535318857)
        ref[3,2,2]=(0.10609796765)
        ref[3,3,0]=(1.14342153343)
        ref[3,3,1]=(-0.0689548622931)
        ref[3,3,2]=(-0.166489981522)
        ref[3,4,0]=(0.995987765181)
        ref[3,4,1]=(0.915359249488)
        ref[3,4,2]=(-0.583052373261)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onFunction_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(0.116652834373)*x[0]+(1.4681512344)*x[1]
        s[0,0,1]=(1.89418390988)*x[0]+(-0.548065006619)*x[1]
        s[0,1,0]=(-0.07602392145)*x[0]+(0.0796515016952)*x[1]
        s[0,1,1]=(1.50333606195)*x[0]+(-0.211673198103)*x[1]
        s[1,0,0]=(0.172419295955)*x[0]+(-0.586603468045)*x[1]
        s[1,0,1]=(0.212952657549)*x[0]+(-0.100672655087)*x[1]
        s[1,1,0]=(-1.11939406352)*x[0]+(-0.589882254225)*x[1]
        s[1,1,1]=(0.318138253303)*x[0]+(0.409850277983)*x[1]
        s[2,0,0]=(-0.0669759465557)*x[0]+(0.525669761096)*x[1]
        s[2,0,1]=(0.349019390883)*x[0]+(0.154604501057)*x[1]
        s[2,1,0]=(1.95277106546)*x[0]+(0.495701766483)*x[1]
        s[2,1,1]=(-0.0986210008015)*x[0]+(1.12636829398)*x[1]
        s[3,0,0]=(0.971837535878)*x[0]+(-0.738253595188)*x[1]
        s[3,0,1]=(0.0314513203112)*x[0]+(-1.16646217783)*x[1]
        s[3,1,0]=(0.485751127243)*x[0]+(-0.340178033471)*x[1]
        s[3,1,1]=(0.408742867205)*x[0]+(0.151827789426)*x[1]
        s[4,0,0]=(-0.0134678971538)*x[0]+(0.884924133141)*x[1]
        s[4,0,1]=(-0.66899222691)*x[0]+(-0.959185584071)*x[1]
        s[4,1,0]=(1.05311767886)*x[0]+(0.64093792324)*x[1]
        s[4,1,1]=(0.562875594668)*x[0]+(0.539525481691)*x[1]
        s[5,0,0]=(0.310908751279)*x[0]+(0.497074267163)*x[1]
        s[5,0,1]=(-0.804186513465)*x[0]+(0.329762438933)*x[1]
        s[5,1,0]=(-0.946808260459)*x[0]+(-0.00260213368693)*x[1]
        s[5,1,1]=(0.63415870202)*x[0]+(1.38103994652)*x[1]
        ref[0,0,0,0]=(0.116652834373)
        ref[0,0,0,1]=(1.4681512344)
        ref[0,0,1,0]=(1.89418390988)
        ref[0,0,1,1]=(-0.548065006619)
        ref[0,1,0,0]=(-0.07602392145)
        ref[0,1,0,1]=(0.0796515016952)
        ref[0,1,1,0]=(1.50333606195)
        ref[0,1,1,1]=(-0.211673198103)
        ref[1,0,0,0]=(0.172419295955)
        ref[1,0,0,1]=(-0.586603468045)
        ref[1,0,1,0]=(0.212952657549)
        ref[1,0,1,1]=(-0.100672655087)
        ref[1,1,0,0]=(-1.11939406352)
        ref[1,1,0,1]=(-0.589882254225)
        ref[1,1,1,0]=(0.318138253303)
        ref[1,1,1,1]=(0.409850277983)
        ref[2,0,0,0]=(-0.0669759465557)
        ref[2,0,0,1]=(0.525669761096)
        ref[2,0,1,0]=(0.349019390883)
        ref[2,0,1,1]=(0.154604501057)
        ref[2,1,0,0]=(1.95277106546)
        ref[2,1,0,1]=(0.495701766483)
        ref[2,1,1,0]=(-0.0986210008015)
        ref[2,1,1,1]=(1.12636829398)
        ref[3,0,0,0]=(0.971837535878)
        ref[3,0,0,1]=(-0.738253595188)
        ref[3,0,1,0]=(0.0314513203112)
        ref[3,0,1,1]=(-1.16646217783)
        ref[3,1,0,0]=(0.485751127243)
        ref[3,1,0,1]=(-0.340178033471)
        ref[3,1,1,0]=(0.408742867205)
        ref[3,1,1,1]=(0.151827789426)
        ref[4,0,0,0]=(-0.0134678971538)
        ref[4,0,0,1]=(0.884924133141)
        ref[4,0,1,0]=(-0.66899222691)
        ref[4,0,1,1]=(-0.959185584071)
        ref[4,1,0,0]=(1.05311767886)
        ref[4,1,0,1]=(0.64093792324)
        ref[4,1,1,0]=(0.562875594668)
        ref[4,1,1,1]=(0.539525481691)
        ref[5,0,0,0]=(0.310908751279)
        ref[5,0,0,1]=(0.497074267163)
        ref[5,0,1,0]=(-0.804186513465)
        ref[5,0,1,1]=(0.329762438933)
        ref[5,1,0,0]=(-0.946808260459)
        ref[5,1,0,1]=(-0.00260213368693)
        ref[5,1,1,0]=(0.63415870202)
        ref[5,1,1,1]=(1.38103994652)
      else:
        s[0,0,0]=(-0.0840260391925)*x[0]+(0.385117859595)*x[1]+(1.36354425949)*x[2]
        s[0,0,1]=(0.682480367755)*x[0]+(-0.658980348414)*x[1]+(-0.129734588977)*x[2]
        s[0,1,0]=(-0.213790105123)*x[0]+(-0.774270821015)*x[1]+(0.160024784274)*x[2]
        s[0,1,1]=(-1.02270895286)*x[0]+(1.02607310825)*x[1]+(0.35056114816)*x[2]
        s[1,0,0]=(0.201792116067)*x[0]+(0.223564282304)*x[1]+(0.986785519545)*x[2]
        s[1,0,1]=(0.250819712775)*x[0]+(0.287643279417)*x[1]+(0.18295590528)*x[2]
        s[1,1,0]=(-1.47569225207)*x[0]+(-0.886933933801)*x[1]+(0.525380474082)*x[2]
        s[1,1,1]=(0.680401800239)*x[0]+(1.36155081028)*x[1]+(-1.76198171789)*x[2]
        s[2,0,0]=(-0.784439188011)*x[0]+(-0.797291895631)*x[1]+(-0.425550387518)*x[2]
        s[2,0,1]=(1.62016996989)*x[0]+(1.02824549576)*x[1]+(-0.157995046652)*x[2]
        s[2,1,0]=(-0.437412344413)*x[0]+(0.97005624264)*x[1]+(-0.321998550341)*x[2]
        s[2,1,1]=(0.473879660993)*x[0]+(0.483144823024)*x[1]+(0.168449443711)*x[2]
        s[3,0,0]=(-1.59822104736)*x[0]+(-1.24012649834)*x[1]+(0.695408081789)*x[2]
        s[3,0,1]=(0.827901715486)*x[0]+(-0.230621260402)*x[1]+(-0.789726238476)*x[2]
        s[3,1,0]=(-0.279064257268)*x[0]+(-1.35315417734)*x[1]+(0.593090655629)*x[2]
        s[3,1,1]=(0.387685237641)*x[0]+(1.60258880823)*x[1]+(-1.10088045432)*x[2]
        s[4,0,0]=(0.771374132686)*x[0]+(0.530393756307)*x[1]+(-0.366697715551)*x[2]
        s[4,0,1]=(-0.182374799863)*x[0]+(-0.125579433837)*x[1]+(-1.2028145445)*x[2]
        s[4,1,0]=(-0.208766654808)*x[0]+(0.645251941741)*x[1]+(-0.534068338041)*x[2]
        s[4,1,1]=(-0.064786779421)*x[0]+(0.0354509511326)*x[1]+(-1.32632049137)*x[2]
        s[5,0,0]=(0.833438445425)*x[0]+(0.770186199496)*x[1]+(0.112178380686)*x[2]
        s[5,0,1]=(1.36614714953)*x[0]+(-0.367243359207)*x[1]+(0.812867174908)*x[2]
        s[5,1,0]=(0.0359079603828)*x[0]+(0.622094474303)*x[1]+(-0.33137187739)*x[2]
        s[5,1,1]=(-0.479714378899)*x[0]+(-0.959098499695)*x[1]+(0.843889761479)*x[2]
        ref[0,0,0,0]=(-0.0840260391925)
        ref[0,0,0,1]=(0.385117859595)
        ref[0,0,0,2]=(1.36354425949)
        ref[0,0,1,0]=(0.682480367755)
        ref[0,0,1,1]=(-0.658980348414)
        ref[0,0,1,2]=(-0.129734588977)
        ref[0,1,0,0]=(-0.213790105123)
        ref[0,1,0,1]=(-0.774270821015)
        ref[0,1,0,2]=(0.160024784274)
        ref[0,1,1,0]=(-1.02270895286)
        ref[0,1,1,1]=(1.02607310825)
        ref[0,1,1,2]=(0.35056114816)
        ref[1,0,0,0]=(0.201792116067)
        ref[1,0,0,1]=(0.223564282304)
        ref[1,0,0,2]=(0.986785519545)
        ref[1,0,1,0]=(0.250819712775)
        ref[1,0,1,1]=(0.287643279417)
        ref[1,0,1,2]=(0.18295590528)
        ref[1,1,0,0]=(-1.47569225207)
        ref[1,1,0,1]=(-0.886933933801)
        ref[1,1,0,2]=(0.525380474082)
        ref[1,1,1,0]=(0.680401800239)
        ref[1,1,1,1]=(1.36155081028)
        ref[1,1,1,2]=(-1.76198171789)
        ref[2,0,0,0]=(-0.784439188011)
        ref[2,0,0,1]=(-0.797291895631)
        ref[2,0,0,2]=(-0.425550387518)
        ref[2,0,1,0]=(1.62016996989)
        ref[2,0,1,1]=(1.02824549576)
        ref[2,0,1,2]=(-0.157995046652)
        ref[2,1,0,0]=(-0.437412344413)
        ref[2,1,0,1]=(0.97005624264)
        ref[2,1,0,2]=(-0.321998550341)
        ref[2,1,1,0]=(0.473879660993)
        ref[2,1,1,1]=(0.483144823024)
        ref[2,1,1,2]=(0.168449443711)
        ref[3,0,0,0]=(-1.59822104736)
        ref[3,0,0,1]=(-1.24012649834)
        ref[3,0,0,2]=(0.695408081789)
        ref[3,0,1,0]=(0.827901715486)
        ref[3,0,1,1]=(-0.230621260402)
        ref[3,0,1,2]=(-0.789726238476)
        ref[3,1,0,0]=(-0.279064257268)
        ref[3,1,0,1]=(-1.35315417734)
        ref[3,1,0,2]=(0.593090655629)
        ref[3,1,1,0]=(0.387685237641)
        ref[3,1,1,1]=(1.60258880823)
        ref[3,1,1,2]=(-1.10088045432)
        ref[4,0,0,0]=(0.771374132686)
        ref[4,0,0,1]=(0.530393756307)
        ref[4,0,0,2]=(-0.366697715551)
        ref[4,0,1,0]=(-0.182374799863)
        ref[4,0,1,1]=(-0.125579433837)
        ref[4,0,1,2]=(-1.2028145445)
        ref[4,1,0,0]=(-0.208766654808)
        ref[4,1,0,1]=(0.645251941741)
        ref[4,1,0,2]=(-0.534068338041)
        ref[4,1,1,0]=(-0.064786779421)
        ref[4,1,1,1]=(0.0354509511326)
        ref[4,1,1,2]=(-1.32632049137)
        ref[5,0,0,0]=(0.833438445425)
        ref[5,0,0,1]=(0.770186199496)
        ref[5,0,0,2]=(0.112178380686)
        ref[5,0,1,0]=(1.36614714953)
        ref[5,0,1,1]=(-0.367243359207)
        ref[5,0,1,2]=(0.812867174908)
        ref[5,1,0,0]=(0.0359079603828)
        ref[5,1,0,1]=(0.622094474303)
        ref[5,1,0,2]=(-0.33137187739)
        ref[5,1,1,0]=(-0.479714378899)
        ref[5,1,1,1]=(-0.959098499695)
        ref[5,1,1,2]=(0.843889761479)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.260967575068)*x[0]**o+(-0.456559015376)*x[0]+(0.663023090785)*x[1]**o+(-0.000383723402374)*x[1]
        ref[0]=o*(0.260967575068)*x_ref[0]**(o-1)+(-0.456559015376)
        ref[1]=o*(0.663023090785)*x_ref[1]**(o-1)+(-0.000383723402374)
      else:
        arg=(-0.679453642968)*x[0]**o+(-0.836558889087)*x[0]+(0.938200435255)*x[1]**o+(-0.368159735907)*x[1]+(-0.0546850112561)*x[2]**o+(0.931986983035)*x[2]
        ref[0]=o*(-0.679453642968)*x_ref[0]**(o-1)+(-0.836558889087)
        ref[1]=o*(0.938200435255)*x_ref[1]**(o-1)+(-0.368159735907)
        ref[2]=o*(-0.0546850112561)*x_ref[2]**(o-1)+(0.931986983035)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.221278607378)*x[0]**o+(-0.25051444588)*x[0]+(0.752161895582)*x[1]**o+(0.662434618125)*x[1]
        arg[1]=(0.984829669888)*x[0]**o+(0.00500258390598)*x[0]+(-0.292523259369)*x[1]**o+(-0.238277884466)*x[1]
        ref[0,0]=o*(0.221278607378)*x_ref[0]**(o-1)+(-0.25051444588)
        ref[0,1]=o*(0.752161895582)*x_ref[1]**(o-1)+(0.662434618125)
        ref[1,0]=o*(0.984829669888)*x_ref[0]**(o-1)+(0.00500258390598)
        ref[1,1]=o*(-0.292523259369)*x_ref[1]**(o-1)+(-0.238277884466)
      else:
        arg[0]=(0.815886350601)*x[0]**o+(-0.41721938077)*x[0]+(0.478358181839)*x[1]**o+(-0.458860964588)*x[1]+(0.575706872631)*x[2]**o+(-0.965426161177)*x[2]
        arg[1]=(-0.159203003073)*x[0]**o+(-0.985356409774)*x[0]+(0.504125935966)*x[1]**o+(0.703505637664)*x[1]+(-0.480016288837)*x[2]**o+(0.544400209204)*x[2]
        ref[0,0]=o*(0.815886350601)*x_ref[0]**(o-1)+(-0.41721938077)
        ref[0,1]=o*(0.478358181839)*x_ref[1]**(o-1)+(-0.458860964588)
        ref[0,2]=o*(0.575706872631)*x_ref[2]**(o-1)+(-0.965426161177)
        ref[1,0]=o*(-0.159203003073)*x_ref[0]**(o-1)+(-0.985356409774)
        ref[1,1]=o*(0.504125935966)*x_ref[1]**(o-1)+(0.703505637664)
        ref[1,2]=o*(-0.480016288837)*x_ref[2]**(o-1)+(0.544400209204)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(-0.722004066753)*x[0]**o+(-0.911439550186)*x[0]+(0.0793361579222)*x[1]**o+(0.958526722957)*x[1]
        arg[0,1]=(-0.0237328054252)*x[0]**o+(0.832621528642)*x[0]+(0.168484408262)*x[1]**o+(-0.181293214021)*x[1]
        arg[0,2]=(0.0284992392733)*x[0]**o+(0.884411090784)*x[0]+(-0.615596355135)*x[1]**o+(-0.564819619851)*x[1]
        arg[0,3]=(0.37408647134)*x[0]**o+(-0.177901551674)*x[0]+(-0.466165048356)*x[1]**o+(-0.278511961588)*x[1]
        arg[0,4]=(-0.24695435531)*x[0]**o+(-0.624082916034)*x[0]+(-0.343849543401)*x[1]**o+(0.661610450445)*x[1]
        arg[1,0]=(-0.176271330569)*x[0]**o+(-0.661612282712)*x[0]+(-0.937163166613)*x[1]**o+(0.376473179587)*x[1]
        arg[1,1]=(0.964111839461)*x[0]**o+(0.576837069403)*x[0]+(-0.718562476684)*x[1]**o+(0.154883495525)*x[1]
        arg[1,2]=(-0.771556809734)*x[0]**o+(0.720054023732)*x[0]+(-0.390732462938)*x[1]**o+(-0.484359019049)*x[1]
        arg[1,3]=(0.565910084656)*x[0]**o+(0.69105192657)*x[0]+(-0.449408955892)*x[1]**o+(-0.789780518242)*x[1]
        arg[1,4]=(0.602485034177)*x[0]**o+(-0.345139282323)*x[0]+(-0.951334152921)*x[1]**o+(-0.478700081579)*x[1]
        arg[2,0]=(0.00114729937901)*x[0]**o+(0.295356090995)*x[0]+(0.982261544035)*x[1]**o+(0.357076141518)*x[1]
        arg[2,1]=(0.653029330145)*x[0]**o+(-0.0178804068723)*x[0]+(-0.0238303862422)*x[1]**o+(0.983033975535)*x[1]
        arg[2,2]=(0.680878878444)*x[0]**o+(-0.404330541218)*x[0]+(0.856544883866)*x[1]**o+(0.71899071821)*x[1]
        arg[2,3]=(0.0674687202072)*x[0]**o+(-0.748990717637)*x[0]+(0.185597242309)*x[1]**o+(0.623849127358)*x[1]
        arg[2,4]=(0.302869699915)*x[0]**o+(-0.852621552847)*x[0]+(0.115557112172)*x[1]**o+(-0.0827214208715)*x[1]
        arg[3,0]=(0.0929072570643)*x[0]**o+(0.533152382923)*x[0]+(-0.0913754984708)*x[1]**o+(-0.25865969337)*x[1]
        arg[3,1]=(-0.794799886368)*x[0]**o+(-0.865263962776)*x[0]+(0.441343356932)*x[1]**o+(-0.541975483415)*x[1]
        arg[3,2]=(0.261341948686)*x[0]**o+(0.378142729621)*x[0]+(0.0556054089673)*x[1]**o+(-0.854195169318)*x[1]
        arg[3,3]=(0.177222467454)*x[0]**o+(-0.400820197499)*x[0]+(0.88513075927)*x[1]**o+(0.0420082325044)*x[1]
        arg[3,4]=(0.307178422145)*x[0]**o+(-0.203765349758)*x[0]+(-0.546052990937)*x[1]**o+(-0.960312136292)*x[1]
        ref[0,0,0]=o*(-0.722004066753)*x_ref[0]**(o-1)+(-0.911439550186)
        ref[0,0,1]=o*(0.0793361579222)*x_ref[1]**(o-1)+(0.958526722957)
        ref[0,1,0]=o*(-0.0237328054252)*x_ref[0]**(o-1)+(0.832621528642)
        ref[0,1,1]=o*(0.168484408262)*x_ref[1]**(o-1)+(-0.181293214021)
        ref[0,2,0]=o*(0.0284992392733)*x_ref[0]**(o-1)+(0.884411090784)
        ref[0,2,1]=o*(-0.615596355135)*x_ref[1]**(o-1)+(-0.564819619851)
        ref[0,3,0]=o*(0.37408647134)*x_ref[0]**(o-1)+(-0.177901551674)
        ref[0,3,1]=o*(-0.466165048356)*x_ref[1]**(o-1)+(-0.278511961588)
        ref[0,4,0]=o*(-0.24695435531)*x_ref[0]**(o-1)+(-0.624082916034)
        ref[0,4,1]=o*(-0.343849543401)*x_ref[1]**(o-1)+(0.661610450445)
        ref[1,0,0]=o*(-0.176271330569)*x_ref[0]**(o-1)+(-0.661612282712)
        ref[1,0,1]=o*(-0.937163166613)*x_ref[1]**(o-1)+(0.376473179587)
        ref[1,1,0]=o*(0.964111839461)*x_ref[0]**(o-1)+(0.576837069403)
        ref[1,1,1]=o*(-0.718562476684)*x_ref[1]**(o-1)+(0.154883495525)
        ref[1,2,0]=o*(-0.771556809734)*x_ref[0]**(o-1)+(0.720054023732)
        ref[1,2,1]=o*(-0.390732462938)*x_ref[1]**(o-1)+(-0.484359019049)
        ref[1,3,0]=o*(0.565910084656)*x_ref[0]**(o-1)+(0.69105192657)
        ref[1,3,1]=o*(-0.449408955892)*x_ref[1]**(o-1)+(-0.789780518242)
        ref[1,4,0]=o*(0.602485034177)*x_ref[0]**(o-1)+(-0.345139282323)
        ref[1,4,1]=o*(-0.951334152921)*x_ref[1]**(o-1)+(-0.478700081579)
        ref[2,0,0]=o*(0.00114729937901)*x_ref[0]**(o-1)+(0.295356090995)
        ref[2,0,1]=o*(0.982261544035)*x_ref[1]**(o-1)+(0.357076141518)
        ref[2,1,0]=o*(0.653029330145)*x_ref[0]**(o-1)+(-0.0178804068723)
        ref[2,1,1]=o*(-0.0238303862422)*x_ref[1]**(o-1)+(0.983033975535)
        ref[2,2,0]=o*(0.680878878444)*x_ref[0]**(o-1)+(-0.404330541218)
        ref[2,2,1]=o*(0.856544883866)*x_ref[1]**(o-1)+(0.71899071821)
        ref[2,3,0]=o*(0.0674687202072)*x_ref[0]**(o-1)+(-0.748990717637)
        ref[2,3,1]=o*(0.185597242309)*x_ref[1]**(o-1)+(0.623849127358)
        ref[2,4,0]=o*(0.302869699915)*x_ref[0]**(o-1)+(-0.852621552847)
        ref[2,4,1]=o*(0.115557112172)*x_ref[1]**(o-1)+(-0.0827214208715)
        ref[3,0,0]=o*(0.0929072570643)*x_ref[0]**(o-1)+(0.533152382923)
        ref[3,0,1]=o*(-0.0913754984708)*x_ref[1]**(o-1)+(-0.25865969337)
        ref[3,1,0]=o*(-0.794799886368)*x_ref[0]**(o-1)+(-0.865263962776)
        ref[3,1,1]=o*(0.441343356932)*x_ref[1]**(o-1)+(-0.541975483415)
        ref[3,2,0]=o*(0.261341948686)*x_ref[0]**(o-1)+(0.378142729621)
        ref[3,2,1]=o*(0.0556054089673)*x_ref[1]**(o-1)+(-0.854195169318)
        ref[3,3,0]=o*(0.177222467454)*x_ref[0]**(o-1)+(-0.400820197499)
        ref[3,3,1]=o*(0.88513075927)*x_ref[1]**(o-1)+(0.0420082325044)
        ref[3,4,0]=o*(0.307178422145)*x_ref[0]**(o-1)+(-0.203765349758)
        ref[3,4,1]=o*(-0.546052990937)*x_ref[1]**(o-1)+(-0.960312136292)
      else:
        arg[0,0]=(0.767891828291)*x[0]**o+(-0.0840162180616)*x[0]+(-0.942354055114)*x[1]**o+(-0.458324281794)*x[1]+(-0.837589052489)*x[2]**o+(-0.49741065871)*x[2]
        arg[0,1]=(-0.770574732815)*x[0]**o+(0.949362675967)*x[0]+(0.334062389906)*x[1]**o+(0.589134582043)*x[1]+(0.611876443107)*x[2]**o+(-0.476746595186)*x[2]
        arg[0,2]=(-0.113076234857)*x[0]**o+(0.207116555982)*x[0]+(-0.434423505296)*x[1]**o+(0.582451788291)*x[1]+(-0.858451242875)*x[2]**o+(-0.229934112115)*x[2]
        arg[0,3]=(0.55494281628)*x[0]**o+(0.53116989281)*x[0]+(0.150549353151)*x[1]**o+(0.70886785918)*x[1]+(-0.148436740612)*x[2]**o+(0.755584812901)*x[2]
        arg[0,4]=(0.0855991607862)*x[0]**o+(0.509120320151)*x[0]+(-0.303374431746)*x[1]**o+(0.1813545791)*x[1]+(-0.0434667822132)*x[2]**o+(0.395447849194)*x[2]
        arg[1,0]=(-0.779720617706)*x[0]**o+(0.730668258583)*x[0]+(-0.0315884698035)*x[1]**o+(-0.0929845369431)*x[1]+(0.0853332239352)*x[2]**o+(0.303372304789)*x[2]
        arg[1,1]=(-0.687021672626)*x[0]**o+(-0.380139153436)*x[0]+(-0.624774641851)*x[1]**o+(0.331833510141)*x[1]+(-0.747115143343)*x[2]**o+(-0.901384102505)*x[2]
        arg[1,2]=(-0.0528543338921)*x[0]**o+(-0.683764012224)*x[0]+(0.498886779776)*x[1]**o+(-0.396777215862)*x[1]+(0.777051418657)*x[2]**o+(-0.786345142641)*x[2]
        arg[1,3]=(-0.823112017253)*x[0]**o+(0.278818264102)*x[0]+(-0.79372160065)*x[1]**o+(0.254044599609)*x[1]+(-0.372925254827)*x[2]**o+(-0.85054008508)*x[2]
        arg[1,4]=(-0.247156845703)*x[0]**o+(-0.944335229412)*x[0]+(-0.572532409179)*x[1]**o+(0.289091364421)*x[1]+(0.848874739954)*x[2]**o+(0.705098999827)*x[2]
        arg[2,0]=(0.0958493700534)*x[0]**o+(-0.334070579338)*x[0]+(-0.74117408785)*x[1]**o+(0.0609646344894)*x[1]+(-0.328888690751)*x[2]**o+(-0.0430772626841)*x[2]
        arg[2,1]=(0.846871401827)*x[0]**o+(0.379912211333)*x[0]+(-0.197881866966)*x[1]**o+(-0.551498190852)*x[1]+(0.416977164384)*x[2]**o+(0.09398324743)*x[2]
        arg[2,2]=(0.737261473925)*x[0]**o+(0.696028328637)*x[0]+(-0.735110324145)*x[1]**o+(0.949453963809)*x[1]+(-0.894722167055)*x[2]**o+(-0.16124057159)*x[2]
        arg[2,3]=(0.267556059297)*x[0]**o+(-0.633581895464)*x[0]+(-0.176232065927)*x[1]**o+(0.98533106984)*x[1]+(0.404671434387)*x[2]**o+(0.343560595562)*x[2]
        arg[2,4]=(0.502981833324)*x[0]**o+(0.716659442696)*x[0]+(-0.349106283426)*x[1]**o+(0.0726954534053)*x[1]+(0.468516366697)*x[2]**o+(0.328926164383)*x[2]
        arg[3,0]=(0.846509624251)*x[0]**o+(-0.34161660121)*x[0]+(0.25149102353)*x[1]**o+(-0.0130496471853)*x[1]+(-0.556826239055)*x[2]**o+(0.774658667066)*x[2]
        arg[3,1]=(-0.0777319662414)*x[0]**o+(0.992622365536)*x[0]+(0.720504365851)*x[1]**o+(0.0701085839568)*x[1]+(0.475573965193)*x[2]**o+(0.402599619829)*x[2]
        arg[3,2]=(-0.311228599853)*x[0]**o+(-0.732684576916)*x[0]+(0.308213192232)*x[1]**o+(-0.131325871896)*x[1]+(0.964631653839)*x[2]**o+(-0.600646614776)*x[2]
        arg[3,3]=(0.656589254036)*x[0]**o+(-0.275449493207)*x[0]+(-0.611075965309)*x[1]**o+(-0.460879628412)*x[1]+(-0.820861932527)*x[2]**o+(0.424066222721)*x[2]
        arg[3,4]=(0.303269333663)*x[0]**o+(-0.745774722275)*x[0]+(0.82933539351)*x[1]**o+(-0.295464064484)*x[1]+(0.818135297647)*x[2]**o+(-0.989600307642)*x[2]
        ref[0,0,0]=o*(0.767891828291)*x_ref[0]**(o-1)+(-0.0840162180616)
        ref[0,0,1]=o*(-0.942354055114)*x_ref[1]**(o-1)+(-0.458324281794)
        ref[0,0,2]=o*(-0.837589052489)*x_ref[2]**(o-1)+(-0.49741065871)
        ref[0,1,0]=o*(-0.770574732815)*x_ref[0]**(o-1)+(0.949362675967)
        ref[0,1,1]=o*(0.334062389906)*x_ref[1]**(o-1)+(0.589134582043)
        ref[0,1,2]=o*(0.611876443107)*x_ref[2]**(o-1)+(-0.476746595186)
        ref[0,2,0]=o*(-0.113076234857)*x_ref[0]**(o-1)+(0.207116555982)
        ref[0,2,1]=o*(-0.434423505296)*x_ref[1]**(o-1)+(0.582451788291)
        ref[0,2,2]=o*(-0.858451242875)*x_ref[2]**(o-1)+(-0.229934112115)
        ref[0,3,0]=o*(0.55494281628)*x_ref[0]**(o-1)+(0.53116989281)
        ref[0,3,1]=o*(0.150549353151)*x_ref[1]**(o-1)+(0.70886785918)
        ref[0,3,2]=o*(-0.148436740612)*x_ref[2]**(o-1)+(0.755584812901)
        ref[0,4,0]=o*(0.0855991607862)*x_ref[0]**(o-1)+(0.509120320151)
        ref[0,4,1]=o*(-0.303374431746)*x_ref[1]**(o-1)+(0.1813545791)
        ref[0,4,2]=o*(-0.0434667822132)*x_ref[2]**(o-1)+(0.395447849194)
        ref[1,0,0]=o*(-0.779720617706)*x_ref[0]**(o-1)+(0.730668258583)
        ref[1,0,1]=o*(-0.0315884698035)*x_ref[1]**(o-1)+(-0.0929845369431)
        ref[1,0,2]=o*(0.0853332239352)*x_ref[2]**(o-1)+(0.303372304789)
        ref[1,1,0]=o*(-0.687021672626)*x_ref[0]**(o-1)+(-0.380139153436)
        ref[1,1,1]=o*(-0.624774641851)*x_ref[1]**(o-1)+(0.331833510141)
        ref[1,1,2]=o*(-0.747115143343)*x_ref[2]**(o-1)+(-0.901384102505)
        ref[1,2,0]=o*(-0.0528543338921)*x_ref[0]**(o-1)+(-0.683764012224)
        ref[1,2,1]=o*(0.498886779776)*x_ref[1]**(o-1)+(-0.396777215862)
        ref[1,2,2]=o*(0.777051418657)*x_ref[2]**(o-1)+(-0.786345142641)
        ref[1,3,0]=o*(-0.823112017253)*x_ref[0]**(o-1)+(0.278818264102)
        ref[1,3,1]=o*(-0.79372160065)*x_ref[1]**(o-1)+(0.254044599609)
        ref[1,3,2]=o*(-0.372925254827)*x_ref[2]**(o-1)+(-0.85054008508)
        ref[1,4,0]=o*(-0.247156845703)*x_ref[0]**(o-1)+(-0.944335229412)
        ref[1,4,1]=o*(-0.572532409179)*x_ref[1]**(o-1)+(0.289091364421)
        ref[1,4,2]=o*(0.848874739954)*x_ref[2]**(o-1)+(0.705098999827)
        ref[2,0,0]=o*(0.0958493700534)*x_ref[0]**(o-1)+(-0.334070579338)
        ref[2,0,1]=o*(-0.74117408785)*x_ref[1]**(o-1)+(0.0609646344894)
        ref[2,0,2]=o*(-0.328888690751)*x_ref[2]**(o-1)+(-0.0430772626841)
        ref[2,1,0]=o*(0.846871401827)*x_ref[0]**(o-1)+(0.379912211333)
        ref[2,1,1]=o*(-0.197881866966)*x_ref[1]**(o-1)+(-0.551498190852)
        ref[2,1,2]=o*(0.416977164384)*x_ref[2]**(o-1)+(0.09398324743)
        ref[2,2,0]=o*(0.737261473925)*x_ref[0]**(o-1)+(0.696028328637)
        ref[2,2,1]=o*(-0.735110324145)*x_ref[1]**(o-1)+(0.949453963809)
        ref[2,2,2]=o*(-0.894722167055)*x_ref[2]**(o-1)+(-0.16124057159)
        ref[2,3,0]=o*(0.267556059297)*x_ref[0]**(o-1)+(-0.633581895464)
        ref[2,3,1]=o*(-0.176232065927)*x_ref[1]**(o-1)+(0.98533106984)
        ref[2,3,2]=o*(0.404671434387)*x_ref[2]**(o-1)+(0.343560595562)
        ref[2,4,0]=o*(0.502981833324)*x_ref[0]**(o-1)+(0.716659442696)
        ref[2,4,1]=o*(-0.349106283426)*x_ref[1]**(o-1)+(0.0726954534053)
        ref[2,4,2]=o*(0.468516366697)*x_ref[2]**(o-1)+(0.328926164383)
        ref[3,0,0]=o*(0.846509624251)*x_ref[0]**(o-1)+(-0.34161660121)
        ref[3,0,1]=o*(0.25149102353)*x_ref[1]**(o-1)+(-0.0130496471853)
        ref[3,0,2]=o*(-0.556826239055)*x_ref[2]**(o-1)+(0.774658667066)
        ref[3,1,0]=o*(-0.0777319662414)*x_ref[0]**(o-1)+(0.992622365536)
        ref[3,1,1]=o*(0.720504365851)*x_ref[1]**(o-1)+(0.0701085839568)
        ref[3,1,2]=o*(0.475573965193)*x_ref[2]**(o-1)+(0.402599619829)
        ref[3,2,0]=o*(-0.311228599853)*x_ref[0]**(o-1)+(-0.732684576916)
        ref[3,2,1]=o*(0.308213192232)*x_ref[1]**(o-1)+(-0.131325871896)
        ref[3,2,2]=o*(0.964631653839)*x_ref[2]**(o-1)+(-0.600646614776)
        ref[3,3,0]=o*(0.656589254036)*x_ref[0]**(o-1)+(-0.275449493207)
        ref[3,3,1]=o*(-0.611075965309)*x_ref[1]**(o-1)+(-0.460879628412)
        ref[3,3,2]=o*(-0.820861932527)*x_ref[2]**(o-1)+(0.424066222721)
        ref[3,4,0]=o*(0.303269333663)*x_ref[0]**(o-1)+(-0.745774722275)
        ref[3,4,1]=o*(0.82933539351)*x_ref[1]**(o-1)+(-0.295464064484)
        ref[3,4,2]=o*(0.818135297647)*x_ref[2]**(o-1)+(-0.989600307642)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.572693829631)*x[0]**o+(0.583556749477)*x[0]+(0.22594627003)*x[1]**o+(0.0199162626119)*x[1]
        arg[0,0,1]=(0.196266833125)*x[0]**o+(0.544824498113)*x[0]+(0.150736772157)*x[1]**o+(0.362174902716)*x[1]
        arg[0,1,0]=(0.169481539273)*x[0]**o+(0.0355504217105)*x[0]+(-0.156746013218)*x[1]**o+(0.879476257954)*x[1]
        arg[0,1,1]=(0.824826637612)*x[0]**o+(-0.732645177858)*x[0]+(0.732873648361)*x[1]**o+(-0.40090515696)*x[1]
        arg[1,0,0]=(0.77348720999)*x[0]**o+(0.229475702878)*x[0]+(0.686324947328)*x[1]**o+(0.313197971717)*x[1]
        arg[1,0,1]=(-0.966626441904)*x[0]**o+(0.989778282376)*x[0]+(0.0447214020611)*x[1]**o+(0.439337525343)*x[1]
        arg[1,1,0]=(-0.774806567769)*x[0]**o+(0.634591112025)*x[0]+(0.797720133135)*x[1]**o+(0.579137230562)*x[1]
        arg[1,1,1]=(-0.118703583742)*x[0]**o+(0.482822282284)*x[0]+(0.141839968056)*x[1]**o+(0.916615496002)*x[1]
        arg[2,0,0]=(0.7419378407)*x[0]**o+(0.754978933397)*x[0]+(-0.839586800033)*x[1]**o+(-0.525830008753)*x[1]
        arg[2,0,1]=(-0.855040380309)*x[0]**o+(0.423661210447)*x[0]+(0.538277902172)*x[1]**o+(0.127617006384)*x[1]
        arg[2,1,0]=(0.107349578938)*x[0]**o+(0.697738123598)*x[0]+(0.862106802027)*x[1]**o+(-0.829827651082)*x[1]
        arg[2,1,1]=(-0.970208311818)*x[0]**o+(0.806698181512)*x[0]+(-0.374007524144)*x[1]**o+(-0.0600501198736)*x[1]
        arg[3,0,0]=(0.223312188411)*x[0]**o+(-0.130776319489)*x[0]+(0.616021313401)*x[1]**o+(-0.552848902138)*x[1]
        arg[3,0,1]=(-0.461837450406)*x[0]**o+(-0.895912538242)*x[0]+(0.477606324718)*x[1]**o+(0.220148409572)*x[1]
        arg[3,1,0]=(0.0783005249695)*x[0]**o+(-0.762350707337)*x[0]+(0.616470467018)*x[1]**o+(-0.154358859634)*x[1]
        arg[3,1,1]=(-0.604945998091)*x[0]**o+(0.23801834237)*x[0]+(0.500666300529)*x[1]**o+(0.709482270038)*x[1]
        arg[4,0,0]=(0.746627054515)*x[0]**o+(0.550282853502)*x[0]+(-0.223527434147)*x[1]**o+(-0.291559486198)*x[1]
        arg[4,0,1]=(0.512462457431)*x[0]**o+(0.0908594735188)*x[0]+(-0.888862229845)*x[1]**o+(-0.969985508695)*x[1]
        arg[4,1,0]=(0.606213701819)*x[0]**o+(0.147049558356)*x[0]+(-0.0365908849884)*x[1]**o+(-0.584510320466)*x[1]
        arg[4,1,1]=(-0.110180223906)*x[0]**o+(-0.83442554862)*x[0]+(-0.0516097061459)*x[1]**o+(-0.155615857347)*x[1]
        arg[5,0,0]=(0.603753207781)*x[0]**o+(0.302032475639)*x[0]+(-0.68719165993)*x[1]**o+(-0.375829665239)*x[1]
        arg[5,0,1]=(-0.599357977805)*x[0]**o+(0.562486062075)*x[0]+(0.102714751356)*x[1]**o+(-0.114179799185)*x[1]
        arg[5,1,0]=(0.874756501256)*x[0]**o+(0.550416007969)*x[0]+(0.669721605367)*x[1]**o+(0.24372600222)*x[1]
        arg[5,1,1]=(0.651356889701)*x[0]**o+(0.178366196323)*x[0]+(0.762268544189)*x[1]**o+(0.515014711815)*x[1]
        ref[0,0,0,0]=o*(-0.572693829631)*x_ref[0]**(o-1)+(0.583556749477)
        ref[0,0,0,1]=o*(0.22594627003)*x_ref[1]**(o-1)+(0.0199162626119)
        ref[0,0,1,0]=o*(0.196266833125)*x_ref[0]**(o-1)+(0.544824498113)
        ref[0,0,1,1]=o*(0.150736772157)*x_ref[1]**(o-1)+(0.362174902716)
        ref[0,1,0,0]=o*(0.169481539273)*x_ref[0]**(o-1)+(0.0355504217105)
        ref[0,1,0,1]=o*(-0.156746013218)*x_ref[1]**(o-1)+(0.879476257954)
        ref[0,1,1,0]=o*(0.824826637612)*x_ref[0]**(o-1)+(-0.732645177858)
        ref[0,1,1,1]=o*(0.732873648361)*x_ref[1]**(o-1)+(-0.40090515696)
        ref[1,0,0,0]=o*(0.77348720999)*x_ref[0]**(o-1)+(0.229475702878)
        ref[1,0,0,1]=o*(0.686324947328)*x_ref[1]**(o-1)+(0.313197971717)
        ref[1,0,1,0]=o*(-0.966626441904)*x_ref[0]**(o-1)+(0.989778282376)
        ref[1,0,1,1]=o*(0.0447214020611)*x_ref[1]**(o-1)+(0.439337525343)
        ref[1,1,0,0]=o*(-0.774806567769)*x_ref[0]**(o-1)+(0.634591112025)
        ref[1,1,0,1]=o*(0.797720133135)*x_ref[1]**(o-1)+(0.579137230562)
        ref[1,1,1,0]=o*(-0.118703583742)*x_ref[0]**(o-1)+(0.482822282284)
        ref[1,1,1,1]=o*(0.141839968056)*x_ref[1]**(o-1)+(0.916615496002)
        ref[2,0,0,0]=o*(0.7419378407)*x_ref[0]**(o-1)+(0.754978933397)
        ref[2,0,0,1]=o*(-0.839586800033)*x_ref[1]**(o-1)+(-0.525830008753)
        ref[2,0,1,0]=o*(-0.855040380309)*x_ref[0]**(o-1)+(0.423661210447)
        ref[2,0,1,1]=o*(0.538277902172)*x_ref[1]**(o-1)+(0.127617006384)
        ref[2,1,0,0]=o*(0.107349578938)*x_ref[0]**(o-1)+(0.697738123598)
        ref[2,1,0,1]=o*(0.862106802027)*x_ref[1]**(o-1)+(-0.829827651082)
        ref[2,1,1,0]=o*(-0.970208311818)*x_ref[0]**(o-1)+(0.806698181512)
        ref[2,1,1,1]=o*(-0.374007524144)*x_ref[1]**(o-1)+(-0.0600501198736)
        ref[3,0,0,0]=o*(0.223312188411)*x_ref[0]**(o-1)+(-0.130776319489)
        ref[3,0,0,1]=o*(0.616021313401)*x_ref[1]**(o-1)+(-0.552848902138)
        ref[3,0,1,0]=o*(-0.461837450406)*x_ref[0]**(o-1)+(-0.895912538242)
        ref[3,0,1,1]=o*(0.477606324718)*x_ref[1]**(o-1)+(0.220148409572)
        ref[3,1,0,0]=o*(0.0783005249695)*x_ref[0]**(o-1)+(-0.762350707337)
        ref[3,1,0,1]=o*(0.616470467018)*x_ref[1]**(o-1)+(-0.154358859634)
        ref[3,1,1,0]=o*(-0.604945998091)*x_ref[0]**(o-1)+(0.23801834237)
        ref[3,1,1,1]=o*(0.500666300529)*x_ref[1]**(o-1)+(0.709482270038)
        ref[4,0,0,0]=o*(0.746627054515)*x_ref[0]**(o-1)+(0.550282853502)
        ref[4,0,0,1]=o*(-0.223527434147)*x_ref[1]**(o-1)+(-0.291559486198)
        ref[4,0,1,0]=o*(0.512462457431)*x_ref[0]**(o-1)+(0.0908594735188)
        ref[4,0,1,1]=o*(-0.888862229845)*x_ref[1]**(o-1)+(-0.969985508695)
        ref[4,1,0,0]=o*(0.606213701819)*x_ref[0]**(o-1)+(0.147049558356)
        ref[4,1,0,1]=o*(-0.0365908849884)*x_ref[1]**(o-1)+(-0.584510320466)
        ref[4,1,1,0]=o*(-0.110180223906)*x_ref[0]**(o-1)+(-0.83442554862)
        ref[4,1,1,1]=o*(-0.0516097061459)*x_ref[1]**(o-1)+(-0.155615857347)
        ref[5,0,0,0]=o*(0.603753207781)*x_ref[0]**(o-1)+(0.302032475639)
        ref[5,0,0,1]=o*(-0.68719165993)*x_ref[1]**(o-1)+(-0.375829665239)
        ref[5,0,1,0]=o*(-0.599357977805)*x_ref[0]**(o-1)+(0.562486062075)
        ref[5,0,1,1]=o*(0.102714751356)*x_ref[1]**(o-1)+(-0.114179799185)
        ref[5,1,0,0]=o*(0.874756501256)*x_ref[0]**(o-1)+(0.550416007969)
        ref[5,1,0,1]=o*(0.669721605367)*x_ref[1]**(o-1)+(0.24372600222)
        ref[5,1,1,0]=o*(0.651356889701)*x_ref[0]**(o-1)+(0.178366196323)
        ref[5,1,1,1]=o*(0.762268544189)*x_ref[1]**(o-1)+(0.515014711815)
      else:
        arg[0,0,0]=(-0.634739027866)*x[0]**o+(-0.696370463772)*x[0]+(0.503815931458)*x[1]**o+(-0.457824888032)*x[1]+(0.655074800235)*x[2]**o+(-0.456111009206)*x[2]
        arg[0,0,1]=(-0.533710147141)*x[0]**o+(-0.0814085290298)*x[0]+(-0.0894565935545)*x[1]**o+(-0.3775655102)*x[1]+(-0.134051544262)*x[2]**o+(0.000713749886013)*x[2]
        arg[0,1,0]=(-0.64394535332)*x[0]**o+(0.430741797171)*x[0]+(0.985157009863)*x[1]**o+(-0.655331266432)*x[1]+(-0.436629135988)*x[2]**o+(-0.404480211702)*x[2]
        arg[0,1,1]=(-0.0858143551662)*x[0]**o+(0.0452769288327)*x[0]+(0.930452744895)*x[1]**o+(-0.877149214078)*x[1]+(0.675472222413)*x[2]**o+(0.917497654616)*x[2]
        arg[1,0,0]=(0.88346309442)*x[0]**o+(0.892116796676)*x[0]+(0.22765643526)*x[1]**o+(-0.934442833938)*x[1]+(0.0781755340006)*x[2]**o+(0.106009847768)*x[2]
        arg[1,0,1]=(0.20824563298)*x[0]**o+(0.613584602599)*x[0]+(-0.482386438919)*x[1]**o+(-0.155840561301)*x[1]+(0.817890842096)*x[2]**o+(0.798519843541)*x[2]
        arg[1,1,0]=(-0.915853569389)*x[0]**o+(0.573041720657)*x[0]+(0.236967108991)*x[1]**o+(0.421805849528)*x[1]+(-0.576407564364)*x[2]**o+(0.0709741076428)*x[2]
        arg[1,1,1]=(0.530417188741)*x[0]**o+(0.872073272277)*x[0]+(-0.21047376763)*x[1]**o+(0.88490064655)*x[1]+(0.523085602451)*x[2]**o+(-0.926133382905)*x[2]
        arg[2,0,0]=(-0.974186019136)*x[0]**o+(-0.332433414417)*x[0]+(-0.500938330608)*x[1]**o+(-0.113188574497)*x[1]+(0.228656301416)*x[2]**o+(-0.375756425313)*x[2]
        arg[2,0,1]=(0.246057421673)*x[0]**o+(0.231587929687)*x[0]+(0.398413883486)*x[1]**o+(-0.390998026525)*x[1]+(0.273401149062)*x[2]**o+(-0.296142647578)*x[2]
        arg[2,1,0]=(0.283717204372)*x[0]**o+(-0.545852417255)*x[0]+(-0.616345461169)*x[1]**o+(-0.705051061202)*x[1]+(0.664377876346)*x[2]**o+(0.338382503272)*x[2]
        arg[2,1,1]=(-0.433297970157)*x[0]**o+(-0.716600154712)*x[0]+(0.836194932687)*x[1]**o+(0.0903372116119)*x[1]+(0.113374499353)*x[2]**o+(-0.588810627425)*x[2]
        arg[3,0,0]=(0.404388255474)*x[0]**o+(0.521816615546)*x[0]+(-0.0976581170689)*x[1]**o+(-0.678563989037)*x[1]+(0.550953471125)*x[2]**o+(-0.668793844235)*x[2]
        arg[3,0,1]=(-0.66023047868)*x[0]**o+(0.822794930168)*x[0]+(-0.284215677966)*x[1]**o+(-0.779237447318)*x[1]+(-0.191727752719)*x[2]**o+(-0.748596484262)*x[2]
        arg[3,1,0]=(-0.0610807571905)*x[0]**o+(0.044623781518)*x[0]+(0.851226664753)*x[1]**o+(0.921090244094)*x[1]+(0.494414571241)*x[2]**o+(-0.473831333684)*x[2]
        arg[3,1,1]=(0.54651807228)*x[0]**o+(-0.580525990778)*x[0]+(-0.822936606301)*x[1]**o+(-0.0169562878937)*x[1]+(-0.732219249555)*x[2]**o+(-0.435156155008)*x[2]
        arg[4,0,0]=(0.830021032139)*x[0]**o+(0.976836986062)*x[0]+(0.534162405111)*x[1]**o+(-0.286292103539)*x[1]+(0.820276894761)*x[2]**o+(0.245897387706)*x[2]
        arg[4,0,1]=(0.362201493311)*x[0]**o+(-0.186328147108)*x[0]+(-0.91611458572)*x[1]**o+(-0.965258904175)*x[1]+(0.61850292372)*x[2]**o+(0.410229835478)*x[2]
        arg[4,1,0]=(0.0726313013515)*x[0]**o+(-0.974714544914)*x[0]+(0.46313543268)*x[1]**o+(0.100486704072)*x[1]+(-0.149124857233)*x[2]**o+(0.455733236337)*x[2]
        arg[4,1,1]=(-0.875220220201)*x[0]**o+(-0.416152301789)*x[0]+(-0.0145627466423)*x[1]**o+(0.267856009119)*x[1]+(-0.604260174049)*x[2]**o+(0.689107997051)*x[2]
        arg[5,0,0]=(0.196182127971)*x[0]**o+(0.337209503833)*x[0]+(0.595342239158)*x[1]**o+(0.507423300857)*x[1]+(0.934760469389)*x[2]**o+(-0.198686443328)*x[2]
        arg[5,0,1]=(0.77210909967)*x[0]**o+(-0.310778136545)*x[0]+(0.512326103142)*x[1]**o+(-0.501056206243)*x[1]+(0.98755961257)*x[2]**o+(-0.658414340038)*x[2]
        arg[5,1,0]=(0.375277702198)*x[0]**o+(-0.603850047717)*x[0]+(0.677227868069)*x[1]**o+(-0.784640615159)*x[1]+(0.177350721639)*x[2]**o+(0.241053657333)*x[2]
        arg[5,1,1]=(-0.368625522736)*x[0]**o+(0.196977476909)*x[0]+(-0.582451887206)*x[1]**o+(0.598240904158)*x[1]+(0.102240230796)*x[2]**o+(0.487322529158)*x[2]
        ref[0,0,0,0]=o*(-0.634739027866)*x_ref[0]**(o-1)+(-0.696370463772)
        ref[0,0,0,1]=o*(0.503815931458)*x_ref[1]**(o-1)+(-0.457824888032)
        ref[0,0,0,2]=o*(0.655074800235)*x_ref[2]**(o-1)+(-0.456111009206)
        ref[0,0,1,0]=o*(-0.533710147141)*x_ref[0]**(o-1)+(-0.0814085290298)
        ref[0,0,1,1]=o*(-0.0894565935545)*x_ref[1]**(o-1)+(-0.3775655102)
        ref[0,0,1,2]=o*(-0.134051544262)*x_ref[2]**(o-1)+(0.000713749886013)
        ref[0,1,0,0]=o*(-0.64394535332)*x_ref[0]**(o-1)+(0.430741797171)
        ref[0,1,0,1]=o*(0.985157009863)*x_ref[1]**(o-1)+(-0.655331266432)
        ref[0,1,0,2]=o*(-0.436629135988)*x_ref[2]**(o-1)+(-0.404480211702)
        ref[0,1,1,0]=o*(-0.0858143551662)*x_ref[0]**(o-1)+(0.0452769288327)
        ref[0,1,1,1]=o*(0.930452744895)*x_ref[1]**(o-1)+(-0.877149214078)
        ref[0,1,1,2]=o*(0.675472222413)*x_ref[2]**(o-1)+(0.917497654616)
        ref[1,0,0,0]=o*(0.88346309442)*x_ref[0]**(o-1)+(0.892116796676)
        ref[1,0,0,1]=o*(0.22765643526)*x_ref[1]**(o-1)+(-0.934442833938)
        ref[1,0,0,2]=o*(0.0781755340006)*x_ref[2]**(o-1)+(0.106009847768)
        ref[1,0,1,0]=o*(0.20824563298)*x_ref[0]**(o-1)+(0.613584602599)
        ref[1,0,1,1]=o*(-0.482386438919)*x_ref[1]**(o-1)+(-0.155840561301)
        ref[1,0,1,2]=o*(0.817890842096)*x_ref[2]**(o-1)+(0.798519843541)
        ref[1,1,0,0]=o*(-0.915853569389)*x_ref[0]**(o-1)+(0.573041720657)
        ref[1,1,0,1]=o*(0.236967108991)*x_ref[1]**(o-1)+(0.421805849528)
        ref[1,1,0,2]=o*(-0.576407564364)*x_ref[2]**(o-1)+(0.0709741076428)
        ref[1,1,1,0]=o*(0.530417188741)*x_ref[0]**(o-1)+(0.872073272277)
        ref[1,1,1,1]=o*(-0.21047376763)*x_ref[1]**(o-1)+(0.88490064655)
        ref[1,1,1,2]=o*(0.523085602451)*x_ref[2]**(o-1)+(-0.926133382905)
        ref[2,0,0,0]=o*(-0.974186019136)*x_ref[0]**(o-1)+(-0.332433414417)
        ref[2,0,0,1]=o*(-0.500938330608)*x_ref[1]**(o-1)+(-0.113188574497)
        ref[2,0,0,2]=o*(0.228656301416)*x_ref[2]**(o-1)+(-0.375756425313)
        ref[2,0,1,0]=o*(0.246057421673)*x_ref[0]**(o-1)+(0.231587929687)
        ref[2,0,1,1]=o*(0.398413883486)*x_ref[1]**(o-1)+(-0.390998026525)
        ref[2,0,1,2]=o*(0.273401149062)*x_ref[2]**(o-1)+(-0.296142647578)
        ref[2,1,0,0]=o*(0.283717204372)*x_ref[0]**(o-1)+(-0.545852417255)
        ref[2,1,0,1]=o*(-0.616345461169)*x_ref[1]**(o-1)+(-0.705051061202)
        ref[2,1,0,2]=o*(0.664377876346)*x_ref[2]**(o-1)+(0.338382503272)
        ref[2,1,1,0]=o*(-0.433297970157)*x_ref[0]**(o-1)+(-0.716600154712)
        ref[2,1,1,1]=o*(0.836194932687)*x_ref[1]**(o-1)+(0.0903372116119)
        ref[2,1,1,2]=o*(0.113374499353)*x_ref[2]**(o-1)+(-0.588810627425)
        ref[3,0,0,0]=o*(0.404388255474)*x_ref[0]**(o-1)+(0.521816615546)
        ref[3,0,0,1]=o*(-0.0976581170689)*x_ref[1]**(o-1)+(-0.678563989037)
        ref[3,0,0,2]=o*(0.550953471125)*x_ref[2]**(o-1)+(-0.668793844235)
        ref[3,0,1,0]=o*(-0.66023047868)*x_ref[0]**(o-1)+(0.822794930168)
        ref[3,0,1,1]=o*(-0.284215677966)*x_ref[1]**(o-1)+(-0.779237447318)
        ref[3,0,1,2]=o*(-0.191727752719)*x_ref[2]**(o-1)+(-0.748596484262)
        ref[3,1,0,0]=o*(-0.0610807571905)*x_ref[0]**(o-1)+(0.044623781518)
        ref[3,1,0,1]=o*(0.851226664753)*x_ref[1]**(o-1)+(0.921090244094)
        ref[3,1,0,2]=o*(0.494414571241)*x_ref[2]**(o-1)+(-0.473831333684)
        ref[3,1,1,0]=o*(0.54651807228)*x_ref[0]**(o-1)+(-0.580525990778)
        ref[3,1,1,1]=o*(-0.822936606301)*x_ref[1]**(o-1)+(-0.0169562878937)
        ref[3,1,1,2]=o*(-0.732219249555)*x_ref[2]**(o-1)+(-0.435156155008)
        ref[4,0,0,0]=o*(0.830021032139)*x_ref[0]**(o-1)+(0.976836986062)
        ref[4,0,0,1]=o*(0.534162405111)*x_ref[1]**(o-1)+(-0.286292103539)
        ref[4,0,0,2]=o*(0.820276894761)*x_ref[2]**(o-1)+(0.245897387706)
        ref[4,0,1,0]=o*(0.362201493311)*x_ref[0]**(o-1)+(-0.186328147108)
        ref[4,0,1,1]=o*(-0.91611458572)*x_ref[1]**(o-1)+(-0.965258904175)
        ref[4,0,1,2]=o*(0.61850292372)*x_ref[2]**(o-1)+(0.410229835478)
        ref[4,1,0,0]=o*(0.0726313013515)*x_ref[0]**(o-1)+(-0.974714544914)
        ref[4,1,0,1]=o*(0.46313543268)*x_ref[1]**(o-1)+(0.100486704072)
        ref[4,1,0,2]=o*(-0.149124857233)*x_ref[2]**(o-1)+(0.455733236337)
        ref[4,1,1,0]=o*(-0.875220220201)*x_ref[0]**(o-1)+(-0.416152301789)
        ref[4,1,1,1]=o*(-0.0145627466423)*x_ref[1]**(o-1)+(0.267856009119)
        ref[4,1,1,2]=o*(-0.604260174049)*x_ref[2]**(o-1)+(0.689107997051)
        ref[5,0,0,0]=o*(0.196182127971)*x_ref[0]**(o-1)+(0.337209503833)
        ref[5,0,0,1]=o*(0.595342239158)*x_ref[1]**(o-1)+(0.507423300857)
        ref[5,0,0,2]=o*(0.934760469389)*x_ref[2]**(o-1)+(-0.198686443328)
        ref[5,0,1,0]=o*(0.77210909967)*x_ref[0]**(o-1)+(-0.310778136545)
        ref[5,0,1,1]=o*(0.512326103142)*x_ref[1]**(o-1)+(-0.501056206243)
        ref[5,0,1,2]=o*(0.98755961257)*x_ref[2]**(o-1)+(-0.658414340038)
        ref[5,1,0,0]=o*(0.375277702198)*x_ref[0]**(o-1)+(-0.603850047717)
        ref[5,1,0,1]=o*(0.677227868069)*x_ref[1]**(o-1)+(-0.784640615159)
        ref[5,1,0,2]=o*(0.177350721639)*x_ref[2]**(o-1)+(0.241053657333)
        ref[5,1,1,0]=o*(-0.368625522736)*x_ref[0]**(o-1)+(0.196977476909)
        ref[5,1,1,1]=o*(-0.582451887206)*x_ref[1]**(o-1)+(0.598240904158)
        ref[5,1,1,2]=o*(0.102240230796)*x_ref[2]**(o-1)+(0.487322529158)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.851643614423)*x[0]**o+(-0.77638754265)*x[0]+(-0.216746360457)*x[1]**o+(-0.117807364973)*x[1]
        ref[0]=o*(0.851643614423)*x_ref[0]**(o-1)+(-0.77638754265)
        ref[1]=o*(-0.216746360457)*x_ref[1]**(o-1)+(-0.117807364973)
      else:
        arg=(0.846271449213)*x[0]**o+(0.895578416904)*x[0]+(-0.518297490388)*x[1]**o+(0.272255265951)*x[1]+(0.537822992528)*x[2]**o+(0.74226977002)*x[2]
        ref[0]=o*(0.846271449213)*x_ref[0]**(o-1)+(0.895578416904)
        ref[1]=o*(-0.518297490388)*x_ref[1]**(o-1)+(0.272255265951)
        ref[2]=o*(0.537822992528)*x_ref[2]**(o-1)+(0.74226977002)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(0.365601318395)*x[0]**o+(0.295624555802)*x[0]+(0.540618182217)*x[1]**o+(0.0458416078386)*x[1]
        arg[1]=(0.332276039998)*x[0]**o+(0.422606936253)*x[0]+(-0.221603594854)*x[1]**o+(0.879863420258)*x[1]
        ref[0,0]=o*(0.365601318395)*x_ref[0]**(o-1)+(0.295624555802)
        ref[0,1]=o*(0.540618182217)*x_ref[1]**(o-1)+(0.0458416078386)
        ref[1,0]=o*(0.332276039998)*x_ref[0]**(o-1)+(0.422606936253)
        ref[1,1]=o*(-0.221603594854)*x_ref[1]**(o-1)+(0.879863420258)
      else:
        arg[0]=(-0.258072277129)*x[0]**o+(-0.833085462518)*x[0]+(-0.124240965858)*x[1]**o+(-0.121853073047)*x[1]+(-0.877974471569)*x[2]**o+(0.929660451331)*x[2]
        arg[1]=(0.306520203535)*x[0]**o+(0.635444320723)*x[0]+(-0.766579509915)*x[1]**o+(-0.381189007175)*x[1]+(-0.76608952351)*x[2]**o+(0.00718583364881)*x[2]
        ref[0,0]=o*(-0.258072277129)*x_ref[0]**(o-1)+(-0.833085462518)
        ref[0,1]=o*(-0.124240965858)*x_ref[1]**(o-1)+(-0.121853073047)
        ref[0,2]=o*(-0.877974471569)*x_ref[2]**(o-1)+(0.929660451331)
        ref[1,0]=o*(0.306520203535)*x_ref[0]**(o-1)+(0.635444320723)
        ref[1,1]=o*(-0.766579509915)*x_ref[1]**(o-1)+(-0.381189007175)
        ref[1,2]=o*(-0.76608952351)*x_ref[2]**(o-1)+(0.00718583364881)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.403088506806)*x[0]**o+(-0.436942773988)*x[0]+(-0.0594011537346)*x[1]**o+(0.855063195595)*x[1]
        arg[0,1]=(-0.516992303892)*x[0]**o+(-0.132586737299)*x[0]+(-0.854639239618)*x[1]**o+(0.241200454455)*x[1]
        arg[0,2]=(-0.373941155498)*x[0]**o+(0.933328225549)*x[0]+(-0.986870460316)*x[1]**o+(-0.051265908773)*x[1]
        arg[0,3]=(-0.175701653328)*x[0]**o+(-0.726542005356)*x[0]+(-0.553505635409)*x[1]**o+(-0.199806969718)*x[1]
        arg[0,4]=(-0.750832127973)*x[0]**o+(0.404511387161)*x[0]+(0.874423404897)*x[1]**o+(0.35360129734)*x[1]
        arg[1,0]=(-0.000191271001285)*x[0]**o+(-0.319365399066)*x[0]+(0.398444976089)*x[1]**o+(-0.966391658006)*x[1]
        arg[1,1]=(-0.58562890965)*x[0]**o+(-0.623992315262)*x[0]+(-0.913704774126)*x[1]**o+(0.910305409081)*x[1]
        arg[1,2]=(-0.480491011401)*x[0]**o+(-0.387015926056)*x[0]+(0.845826744282)*x[1]**o+(0.380626980034)*x[1]
        arg[1,3]=(-0.284167706438)*x[0]**o+(0.730574128045)*x[0]+(0.823389953051)*x[1]**o+(0.59483032515)*x[1]
        arg[1,4]=(-0.545147810354)*x[0]**o+(-0.632923743497)*x[0]+(0.619596244096)*x[1]**o+(-0.816596365827)*x[1]
        arg[2,0]=(0.161289895106)*x[0]**o+(0.435472177094)*x[0]+(0.125591455076)*x[1]**o+(0.0207320895937)*x[1]
        arg[2,1]=(0.639457164801)*x[0]**o+(0.2631408826)*x[0]+(0.00255095677899)*x[1]**o+(0.732164720004)*x[1]
        arg[2,2]=(-0.152014092979)*x[0]**o+(0.347024615036)*x[0]+(0.94212294017)*x[1]**o+(0.173939724203)*x[1]
        arg[2,3]=(-0.0324872553975)*x[0]**o+(-0.969492827039)*x[0]+(-0.830876162378)*x[1]**o+(0.601308513636)*x[1]
        arg[2,4]=(0.932999309515)*x[0]**o+(0.0320937041249)*x[0]+(-0.354492894455)*x[1]**o+(-0.752837146132)*x[1]
        arg[3,0]=(0.995147840382)*x[0]**o+(-0.276974278934)*x[0]+(0.905272259272)*x[1]**o+(-0.433178159448)*x[1]
        arg[3,1]=(-0.291708633096)*x[0]**o+(-0.542000555721)*x[0]+(0.372404748364)*x[1]**o+(0.23140215774)*x[1]
        arg[3,2]=(0.88604771071)*x[0]**o+(-0.753078356803)*x[0]+(-0.265302521749)*x[1]**o+(-0.948898850214)*x[1]
        arg[3,3]=(-0.429496844082)*x[0]**o+(0.0731695917195)*x[0]+(0.291387878783)*x[1]**o+(-0.078863825039)*x[1]
        arg[3,4]=(-0.0067999635984)*x[0]**o+(0.488018496019)*x[0]+(-0.875646054486)*x[1]**o+(0.629803284064)*x[1]
        ref[0,0,0]=o*(0.403088506806)*x_ref[0]**(o-1)+(-0.436942773988)
        ref[0,0,1]=o*(-0.0594011537346)*x_ref[1]**(o-1)+(0.855063195595)
        ref[0,1,0]=o*(-0.516992303892)*x_ref[0]**(o-1)+(-0.132586737299)
        ref[0,1,1]=o*(-0.854639239618)*x_ref[1]**(o-1)+(0.241200454455)
        ref[0,2,0]=o*(-0.373941155498)*x_ref[0]**(o-1)+(0.933328225549)
        ref[0,2,1]=o*(-0.986870460316)*x_ref[1]**(o-1)+(-0.051265908773)
        ref[0,3,0]=o*(-0.175701653328)*x_ref[0]**(o-1)+(-0.726542005356)
        ref[0,3,1]=o*(-0.553505635409)*x_ref[1]**(o-1)+(-0.199806969718)
        ref[0,4,0]=o*(-0.750832127973)*x_ref[0]**(o-1)+(0.404511387161)
        ref[0,4,1]=o*(0.874423404897)*x_ref[1]**(o-1)+(0.35360129734)
        ref[1,0,0]=o*(-0.000191271001285)*x_ref[0]**(o-1)+(-0.319365399066)
        ref[1,0,1]=o*(0.398444976089)*x_ref[1]**(o-1)+(-0.966391658006)
        ref[1,1,0]=o*(-0.58562890965)*x_ref[0]**(o-1)+(-0.623992315262)
        ref[1,1,1]=o*(-0.913704774126)*x_ref[1]**(o-1)+(0.910305409081)
        ref[1,2,0]=o*(-0.480491011401)*x_ref[0]**(o-1)+(-0.387015926056)
        ref[1,2,1]=o*(0.845826744282)*x_ref[1]**(o-1)+(0.380626980034)
        ref[1,3,0]=o*(-0.284167706438)*x_ref[0]**(o-1)+(0.730574128045)
        ref[1,3,1]=o*(0.823389953051)*x_ref[1]**(o-1)+(0.59483032515)
        ref[1,4,0]=o*(-0.545147810354)*x_ref[0]**(o-1)+(-0.632923743497)
        ref[1,4,1]=o*(0.619596244096)*x_ref[1]**(o-1)+(-0.816596365827)
        ref[2,0,0]=o*(0.161289895106)*x_ref[0]**(o-1)+(0.435472177094)
        ref[2,0,1]=o*(0.125591455076)*x_ref[1]**(o-1)+(0.0207320895937)
        ref[2,1,0]=o*(0.639457164801)*x_ref[0]**(o-1)+(0.2631408826)
        ref[2,1,1]=o*(0.00255095677899)*x_ref[1]**(o-1)+(0.732164720004)
        ref[2,2,0]=o*(-0.152014092979)*x_ref[0]**(o-1)+(0.347024615036)
        ref[2,2,1]=o*(0.94212294017)*x_ref[1]**(o-1)+(0.173939724203)
        ref[2,3,0]=o*(-0.0324872553975)*x_ref[0]**(o-1)+(-0.969492827039)
        ref[2,3,1]=o*(-0.830876162378)*x_ref[1]**(o-1)+(0.601308513636)
        ref[2,4,0]=o*(0.932999309515)*x_ref[0]**(o-1)+(0.0320937041249)
        ref[2,4,1]=o*(-0.354492894455)*x_ref[1]**(o-1)+(-0.752837146132)
        ref[3,0,0]=o*(0.995147840382)*x_ref[0]**(o-1)+(-0.276974278934)
        ref[3,0,1]=o*(0.905272259272)*x_ref[1]**(o-1)+(-0.433178159448)
        ref[3,1,0]=o*(-0.291708633096)*x_ref[0]**(o-1)+(-0.542000555721)
        ref[3,1,1]=o*(0.372404748364)*x_ref[1]**(o-1)+(0.23140215774)
        ref[3,2,0]=o*(0.88604771071)*x_ref[0]**(o-1)+(-0.753078356803)
        ref[3,2,1]=o*(-0.265302521749)*x_ref[1]**(o-1)+(-0.948898850214)
        ref[3,3,0]=o*(-0.429496844082)*x_ref[0]**(o-1)+(0.0731695917195)
        ref[3,3,1]=o*(0.291387878783)*x_ref[1]**(o-1)+(-0.078863825039)
        ref[3,4,0]=o*(-0.0067999635984)*x_ref[0]**(o-1)+(0.488018496019)
        ref[3,4,1]=o*(-0.875646054486)*x_ref[1]**(o-1)+(0.629803284064)
      else:
        arg[0,0]=(-0.512668741406)*x[0]**o+(-0.73142104842)*x[0]+(0.489050993226)*x[1]**o+(0.771941571419)*x[1]+(-0.538092512577)*x[2]**o+(-0.457812389328)*x[2]
        arg[0,1]=(0.258566436386)*x[0]**o+(-0.572006860118)*x[0]+(-0.0370305840623)*x[1]**o+(0.328810934717)*x[1]+(-0.652873123265)*x[2]**o+(0.734929914374)*x[2]
        arg[0,2]=(-0.228462027625)*x[0]**o+(-0.873359484139)*x[0]+(0.902575701641)*x[1]**o+(0.428758160478)*x[1]+(-0.377042847702)*x[2]**o+(-0.161364604019)*x[2]
        arg[0,3]=(0.31338185054)*x[0]**o+(0.973849539932)*x[0]+(-0.403569231167)*x[1]**o+(-0.110026269257)*x[1]+(0.603101392345)*x[2]**o+(0.499232034173)*x[2]
        arg[0,4]=(0.468416317974)*x[0]**o+(-0.0154203909264)*x[0]+(0.409228888268)*x[1]**o+(0.358886480877)*x[1]+(-0.00677343401301)*x[2]**o+(-0.0567153195302)*x[2]
        arg[1,0]=(0.614929795628)*x[0]**o+(0.54911679279)*x[0]+(0.527138394555)*x[1]**o+(-0.84070465171)*x[1]+(0.421928294706)*x[2]**o+(-0.871734463848)*x[2]
        arg[1,1]=(-0.498770327357)*x[0]**o+(0.294610855431)*x[0]+(-0.815129074288)*x[1]**o+(0.395332176796)*x[1]+(-0.555947305527)*x[2]**o+(-0.623550814988)*x[2]
        arg[1,2]=(-0.369404109593)*x[0]**o+(-0.662927702362)*x[0]+(0.886789099892)*x[1]**o+(-0.815728539359)*x[1]+(-0.843861162574)*x[2]**o+(-0.510039607856)*x[2]
        arg[1,3]=(-0.747305678574)*x[0]**o+(0.350337618157)*x[0]+(0.313491051893)*x[1]**o+(-0.554840960712)*x[1]+(-0.259760094369)*x[2]**o+(0.904246223768)*x[2]
        arg[1,4]=(0.485859115418)*x[0]**o+(-0.751900240375)*x[0]+(-0.233384727142)*x[1]**o+(-0.466677891749)*x[1]+(-0.771311310708)*x[2]**o+(0.380515497846)*x[2]
        arg[2,0]=(0.628529887016)*x[0]**o+(0.778810609149)*x[0]+(-0.95647127374)*x[1]**o+(-0.887895309984)*x[1]+(-0.926017554164)*x[2]**o+(0.0922106889932)*x[2]
        arg[2,1]=(-0.981382750453)*x[0]**o+(0.858805367823)*x[0]+(-0.202352133378)*x[1]**o+(0.527229765084)*x[1]+(-0.0862956097426)*x[2]**o+(-0.76269735579)*x[2]
        arg[2,2]=(0.591276459941)*x[0]**o+(0.554086306469)*x[0]+(-0.175045397917)*x[1]**o+(0.319542375346)*x[1]+(0.328810955576)*x[2]**o+(-0.245381809232)*x[2]
        arg[2,3]=(-0.826436985287)*x[0]**o+(-0.931352535769)*x[0]+(-0.480801956568)*x[1]**o+(0.830773509938)*x[1]+(0.157556801982)*x[2]**o+(0.982505166946)*x[2]
        arg[2,4]=(-0.0992736644463)*x[0]**o+(-0.147525710611)*x[0]+(-0.723322157973)*x[1]**o+(0.0527438662093)*x[1]+(-0.338847227522)*x[2]**o+(-0.704585609897)*x[2]
        arg[3,0]=(-0.45250460769)*x[0]**o+(-0.323299292565)*x[0]+(-0.435540041421)*x[1]**o+(0.418330181239)*x[1]+(0.526739853977)*x[2]**o+(-0.00513414974056)*x[2]
        arg[3,1]=(0.289534600598)*x[0]**o+(0.467498987256)*x[0]+(0.27851903721)*x[1]**o+(0.908249187588)*x[1]+(-0.84695489717)*x[2]**o+(-0.799372710877)*x[2]
        arg[3,2]=(0.00338011092566)*x[0]**o+(0.819216072868)*x[0]+(0.115437140937)*x[1]**o+(0.0968835984948)*x[1]+(-0.633253222581)*x[2]**o+(-0.0224890801282)*x[2]
        arg[3,3]=(0.578331574145)*x[0]**o+(0.328247456626)*x[0]+(-0.725914955131)*x[1]**o+(-0.863562766028)*x[1]+(-0.283057959979)*x[2]**o+(-0.521563360583)*x[2]
        arg[3,4]=(0.345725352698)*x[0]**o+(0.614912618718)*x[0]+(-0.218084952915)*x[1]**o+(-0.629514485943)*x[1]+(-0.207605980168)*x[2]**o+(-0.0950042458608)*x[2]
        ref[0,0,0]=o*(-0.512668741406)*x_ref[0]**(o-1)+(-0.73142104842)
        ref[0,0,1]=o*(0.489050993226)*x_ref[1]**(o-1)+(0.771941571419)
        ref[0,0,2]=o*(-0.538092512577)*x_ref[2]**(o-1)+(-0.457812389328)
        ref[0,1,0]=o*(0.258566436386)*x_ref[0]**(o-1)+(-0.572006860118)
        ref[0,1,1]=o*(-0.0370305840623)*x_ref[1]**(o-1)+(0.328810934717)
        ref[0,1,2]=o*(-0.652873123265)*x_ref[2]**(o-1)+(0.734929914374)
        ref[0,2,0]=o*(-0.228462027625)*x_ref[0]**(o-1)+(-0.873359484139)
        ref[0,2,1]=o*(0.902575701641)*x_ref[1]**(o-1)+(0.428758160478)
        ref[0,2,2]=o*(-0.377042847702)*x_ref[2]**(o-1)+(-0.161364604019)
        ref[0,3,0]=o*(0.31338185054)*x_ref[0]**(o-1)+(0.973849539932)
        ref[0,3,1]=o*(-0.403569231167)*x_ref[1]**(o-1)+(-0.110026269257)
        ref[0,3,2]=o*(0.603101392345)*x_ref[2]**(o-1)+(0.499232034173)
        ref[0,4,0]=o*(0.468416317974)*x_ref[0]**(o-1)+(-0.0154203909264)
        ref[0,4,1]=o*(0.409228888268)*x_ref[1]**(o-1)+(0.358886480877)
        ref[0,4,2]=o*(-0.00677343401301)*x_ref[2]**(o-1)+(-0.0567153195302)
        ref[1,0,0]=o*(0.614929795628)*x_ref[0]**(o-1)+(0.54911679279)
        ref[1,0,1]=o*(0.527138394555)*x_ref[1]**(o-1)+(-0.84070465171)
        ref[1,0,2]=o*(0.421928294706)*x_ref[2]**(o-1)+(-0.871734463848)
        ref[1,1,0]=o*(-0.498770327357)*x_ref[0]**(o-1)+(0.294610855431)
        ref[1,1,1]=o*(-0.815129074288)*x_ref[1]**(o-1)+(0.395332176796)
        ref[1,1,2]=o*(-0.555947305527)*x_ref[2]**(o-1)+(-0.623550814988)
        ref[1,2,0]=o*(-0.369404109593)*x_ref[0]**(o-1)+(-0.662927702362)
        ref[1,2,1]=o*(0.886789099892)*x_ref[1]**(o-1)+(-0.815728539359)
        ref[1,2,2]=o*(-0.843861162574)*x_ref[2]**(o-1)+(-0.510039607856)
        ref[1,3,0]=o*(-0.747305678574)*x_ref[0]**(o-1)+(0.350337618157)
        ref[1,3,1]=o*(0.313491051893)*x_ref[1]**(o-1)+(-0.554840960712)
        ref[1,3,2]=o*(-0.259760094369)*x_ref[2]**(o-1)+(0.904246223768)
        ref[1,4,0]=o*(0.485859115418)*x_ref[0]**(o-1)+(-0.751900240375)
        ref[1,4,1]=o*(-0.233384727142)*x_ref[1]**(o-1)+(-0.466677891749)
        ref[1,4,2]=o*(-0.771311310708)*x_ref[2]**(o-1)+(0.380515497846)
        ref[2,0,0]=o*(0.628529887016)*x_ref[0]**(o-1)+(0.778810609149)
        ref[2,0,1]=o*(-0.95647127374)*x_ref[1]**(o-1)+(-0.887895309984)
        ref[2,0,2]=o*(-0.926017554164)*x_ref[2]**(o-1)+(0.0922106889932)
        ref[2,1,0]=o*(-0.981382750453)*x_ref[0]**(o-1)+(0.858805367823)
        ref[2,1,1]=o*(-0.202352133378)*x_ref[1]**(o-1)+(0.527229765084)
        ref[2,1,2]=o*(-0.0862956097426)*x_ref[2]**(o-1)+(-0.76269735579)
        ref[2,2,0]=o*(0.591276459941)*x_ref[0]**(o-1)+(0.554086306469)
        ref[2,2,1]=o*(-0.175045397917)*x_ref[1]**(o-1)+(0.319542375346)
        ref[2,2,2]=o*(0.328810955576)*x_ref[2]**(o-1)+(-0.245381809232)
        ref[2,3,0]=o*(-0.826436985287)*x_ref[0]**(o-1)+(-0.931352535769)
        ref[2,3,1]=o*(-0.480801956568)*x_ref[1]**(o-1)+(0.830773509938)
        ref[2,3,2]=o*(0.157556801982)*x_ref[2]**(o-1)+(0.982505166946)
        ref[2,4,0]=o*(-0.0992736644463)*x_ref[0]**(o-1)+(-0.147525710611)
        ref[2,4,1]=o*(-0.723322157973)*x_ref[1]**(o-1)+(0.0527438662093)
        ref[2,4,2]=o*(-0.338847227522)*x_ref[2]**(o-1)+(-0.704585609897)
        ref[3,0,0]=o*(-0.45250460769)*x_ref[0]**(o-1)+(-0.323299292565)
        ref[3,0,1]=o*(-0.435540041421)*x_ref[1]**(o-1)+(0.418330181239)
        ref[3,0,2]=o*(0.526739853977)*x_ref[2]**(o-1)+(-0.00513414974056)
        ref[3,1,0]=o*(0.289534600598)*x_ref[0]**(o-1)+(0.467498987256)
        ref[3,1,1]=o*(0.27851903721)*x_ref[1]**(o-1)+(0.908249187588)
        ref[3,1,2]=o*(-0.84695489717)*x_ref[2]**(o-1)+(-0.799372710877)
        ref[3,2,0]=o*(0.00338011092566)*x_ref[0]**(o-1)+(0.819216072868)
        ref[3,2,1]=o*(0.115437140937)*x_ref[1]**(o-1)+(0.0968835984948)
        ref[3,2,2]=o*(-0.633253222581)*x_ref[2]**(o-1)+(-0.0224890801282)
        ref[3,3,0]=o*(0.578331574145)*x_ref[0]**(o-1)+(0.328247456626)
        ref[3,3,1]=o*(-0.725914955131)*x_ref[1]**(o-1)+(-0.863562766028)
        ref[3,3,2]=o*(-0.283057959979)*x_ref[2]**(o-1)+(-0.521563360583)
        ref[3,4,0]=o*(0.345725352698)*x_ref[0]**(o-1)+(0.614912618718)
        ref[3,4,1]=o*(-0.218084952915)*x_ref[1]**(o-1)+(-0.629514485943)
        ref[3,4,2]=o*(-0.207605980168)*x_ref[2]**(o-1)+(-0.0950042458608)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_Solution_rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.353390112117)*x[0]**o+(0.121267964347)*x[0]+(0.095925646757)*x[1]**o+(-0.00141503525661)*x[1]
        arg[0,0,1]=(0.610556742471)*x[0]**o+(0.542948343923)*x[0]+(0.124762431991)*x[1]**o+(0.00216142978003)*x[1]
        arg[0,1,0]=(0.261559018753)*x[0]**o+(-0.881689419767)*x[0]+(-0.009352418091)*x[1]**o+(0.333357441723)*x[1]
        arg[0,1,1]=(-0.25379927884)*x[0]**o+(-0.124076833523)*x[0]+(-0.167690772835)*x[1]**o+(0.89804632709)*x[1]
        arg[1,0,0]=(-0.39353207502)*x[0]**o+(0.838815579522)*x[0]+(0.829467918805)*x[1]**o+(0.614403367394)*x[1]
        arg[1,0,1]=(-0.698946975112)*x[0]**o+(-0.551313390457)*x[0]+(0.965127390549)*x[1]**o+(-0.819660095397)*x[1]
        arg[1,1,0]=(-0.656237793952)*x[0]**o+(0.727261124295)*x[0]+(0.602446266985)*x[1]**o+(-0.320796826351)*x[1]
        arg[1,1,1]=(0.111875111713)*x[0]**o+(0.000360626785094)*x[0]+(-0.802516890555)*x[1]**o+(0.420759758492)*x[1]
        arg[2,0,0]=(-0.421920068352)*x[0]**o+(-0.022970702876)*x[0]+(-0.0257219638216)*x[1]**o+(-0.36544576073)*x[1]
        arg[2,0,1]=(-0.459431228769)*x[0]**o+(0.150702477357)*x[0]+(-0.869898945031)*x[1]**o+(0.261651620176)*x[1]
        arg[2,1,0]=(-0.866165026127)*x[0]**o+(0.149307656472)*x[0]+(-0.270058622191)*x[1]**o+(0.0873163553474)*x[1]
        arg[2,1,1]=(-0.871128617394)*x[0]**o+(-0.774356000237)*x[0]+(0.337383167457)*x[1]**o+(-0.453335744484)*x[1]
        arg[3,0,0]=(0.404195477481)*x[0]**o+(-0.350041445198)*x[0]+(-0.583794909281)*x[1]**o+(-0.895439618872)*x[1]
        arg[3,0,1]=(0.693567867164)*x[0]**o+(0.989642722856)*x[0]+(0.814104467522)*x[1]**o+(-0.126302588221)*x[1]
        arg[3,1,0]=(0.935509747673)*x[0]**o+(0.0104612091188)*x[0]+(0.304974305936)*x[1]**o+(0.658660310426)*x[1]
        arg[3,1,1]=(0.677420202064)*x[0]**o+(-0.160834685914)*x[0]+(0.241578146196)*x[1]**o+(-0.182223867353)*x[1]
        arg[4,0,0]=(-0.908542015864)*x[0]**o+(0.125595344482)*x[0]+(-0.396864254557)*x[1]**o+(0.408182849245)*x[1]
        arg[4,0,1]=(0.436101221257)*x[0]**o+(0.946667075656)*x[0]+(-0.293523459017)*x[1]**o+(0.363130416513)*x[1]
        arg[4,1,0]=(0.583416010605)*x[0]**o+(0.531883961319)*x[0]+(-0.418905217485)*x[1]**o+(0.271442034667)*x[1]
        arg[4,1,1]=(-0.708766397827)*x[0]**o+(0.0266251980073)*x[0]+(-0.0929293524384)*x[1]**o+(-0.661095295605)*x[1]
        arg[5,0,0]=(0.693424423443)*x[0]**o+(-0.201539674621)*x[0]+(-0.221922493999)*x[1]**o+(-0.218359043699)*x[1]
        arg[5,0,1]=(-0.478707796293)*x[0]**o+(0.791533578559)*x[0]+(0.280213321002)*x[1]**o+(-0.60809908629)*x[1]
        arg[5,1,0]=(-0.517992572451)*x[0]**o+(0.591995363281)*x[0]+(0.480220427965)*x[1]**o+(0.134542063762)*x[1]
        arg[5,1,1]=(0.415861674391)*x[0]**o+(0.773614372731)*x[0]+(-0.984515861984)*x[1]**o+(0.971916685266)*x[1]
        ref[0,0,0,0]=o*(-0.353390112117)*x_ref[0]**(o-1)+(0.121267964347)
        ref[0,0,0,1]=o*(0.095925646757)*x_ref[1]**(o-1)+(-0.00141503525661)
        ref[0,0,1,0]=o*(0.610556742471)*x_ref[0]**(o-1)+(0.542948343923)
        ref[0,0,1,1]=o*(0.124762431991)*x_ref[1]**(o-1)+(0.00216142978003)
        ref[0,1,0,0]=o*(0.261559018753)*x_ref[0]**(o-1)+(-0.881689419767)
        ref[0,1,0,1]=o*(-0.009352418091)*x_ref[1]**(o-1)+(0.333357441723)
        ref[0,1,1,0]=o*(-0.25379927884)*x_ref[0]**(o-1)+(-0.124076833523)
        ref[0,1,1,1]=o*(-0.167690772835)*x_ref[1]**(o-1)+(0.89804632709)
        ref[1,0,0,0]=o*(-0.39353207502)*x_ref[0]**(o-1)+(0.838815579522)
        ref[1,0,0,1]=o*(0.829467918805)*x_ref[1]**(o-1)+(0.614403367394)
        ref[1,0,1,0]=o*(-0.698946975112)*x_ref[0]**(o-1)+(-0.551313390457)
        ref[1,0,1,1]=o*(0.965127390549)*x_ref[1]**(o-1)+(-0.819660095397)
        ref[1,1,0,0]=o*(-0.656237793952)*x_ref[0]**(o-1)+(0.727261124295)
        ref[1,1,0,1]=o*(0.602446266985)*x_ref[1]**(o-1)+(-0.320796826351)
        ref[1,1,1,0]=o*(0.111875111713)*x_ref[0]**(o-1)+(0.000360626785094)
        ref[1,1,1,1]=o*(-0.802516890555)*x_ref[1]**(o-1)+(0.420759758492)
        ref[2,0,0,0]=o*(-0.421920068352)*x_ref[0]**(o-1)+(-0.022970702876)
        ref[2,0,0,1]=o*(-0.0257219638216)*x_ref[1]**(o-1)+(-0.36544576073)
        ref[2,0,1,0]=o*(-0.459431228769)*x_ref[0]**(o-1)+(0.150702477357)
        ref[2,0,1,1]=o*(-0.869898945031)*x_ref[1]**(o-1)+(0.261651620176)
        ref[2,1,0,0]=o*(-0.866165026127)*x_ref[0]**(o-1)+(0.149307656472)
        ref[2,1,0,1]=o*(-0.270058622191)*x_ref[1]**(o-1)+(0.0873163553474)
        ref[2,1,1,0]=o*(-0.871128617394)*x_ref[0]**(o-1)+(-0.774356000237)
        ref[2,1,1,1]=o*(0.337383167457)*x_ref[1]**(o-1)+(-0.453335744484)
        ref[3,0,0,0]=o*(0.404195477481)*x_ref[0]**(o-1)+(-0.350041445198)
        ref[3,0,0,1]=o*(-0.583794909281)*x_ref[1]**(o-1)+(-0.895439618872)
        ref[3,0,1,0]=o*(0.693567867164)*x_ref[0]**(o-1)+(0.989642722856)
        ref[3,0,1,1]=o*(0.814104467522)*x_ref[1]**(o-1)+(-0.126302588221)
        ref[3,1,0,0]=o*(0.935509747673)*x_ref[0]**(o-1)+(0.0104612091188)
        ref[3,1,0,1]=o*(0.304974305936)*x_ref[1]**(o-1)+(0.658660310426)
        ref[3,1,1,0]=o*(0.677420202064)*x_ref[0]**(o-1)+(-0.160834685914)
        ref[3,1,1,1]=o*(0.241578146196)*x_ref[1]**(o-1)+(-0.182223867353)
        ref[4,0,0,0]=o*(-0.908542015864)*x_ref[0]**(o-1)+(0.125595344482)
        ref[4,0,0,1]=o*(-0.396864254557)*x_ref[1]**(o-1)+(0.408182849245)
        ref[4,0,1,0]=o*(0.436101221257)*x_ref[0]**(o-1)+(0.946667075656)
        ref[4,0,1,1]=o*(-0.293523459017)*x_ref[1]**(o-1)+(0.363130416513)
        ref[4,1,0,0]=o*(0.583416010605)*x_ref[0]**(o-1)+(0.531883961319)
        ref[4,1,0,1]=o*(-0.418905217485)*x_ref[1]**(o-1)+(0.271442034667)
        ref[4,1,1,0]=o*(-0.708766397827)*x_ref[0]**(o-1)+(0.0266251980073)
        ref[4,1,1,1]=o*(-0.0929293524384)*x_ref[1]**(o-1)+(-0.661095295605)
        ref[5,0,0,0]=o*(0.693424423443)*x_ref[0]**(o-1)+(-0.201539674621)
        ref[5,0,0,1]=o*(-0.221922493999)*x_ref[1]**(o-1)+(-0.218359043699)
        ref[5,0,1,0]=o*(-0.478707796293)*x_ref[0]**(o-1)+(0.791533578559)
        ref[5,0,1,1]=o*(0.280213321002)*x_ref[1]**(o-1)+(-0.60809908629)
        ref[5,1,0,0]=o*(-0.517992572451)*x_ref[0]**(o-1)+(0.591995363281)
        ref[5,1,0,1]=o*(0.480220427965)*x_ref[1]**(o-1)+(0.134542063762)
        ref[5,1,1,0]=o*(0.415861674391)*x_ref[0]**(o-1)+(0.773614372731)
        ref[5,1,1,1]=o*(-0.984515861984)*x_ref[1]**(o-1)+(0.971916685266)
      else:
        arg[0,0,0]=(0.157678109652)*x[0]**o+(-0.939095069874)*x[0]+(0.340072659801)*x[1]**o+(0.806723712633)*x[1]+(0.0175654201162)*x[2]**o+(-0.695887679405)*x[2]
        arg[0,0,1]=(0.148444957808)*x[0]**o+(0.769780554196)*x[0]+(-0.999778710821)*x[1]**o+(0.765998508134)*x[1]+(0.489498280892)*x[2]**o+(0.272011172233)*x[2]
        arg[0,1,0]=(-0.78853623669)*x[0]**o+(-0.889333747099)*x[0]+(0.897684283327)*x[1]**o+(0.878046029595)*x[1]+(-0.355566201991)*x[2]**o+(-0.656187599294)*x[2]
        arg[0,1,1]=(-0.903496685895)*x[0]**o+(0.335762730702)*x[0]+(0.264330545204)*x[1]**o+(-0.369300150725)*x[1]+(0.00819339769162)*x[2]**o+(0.0466274018566)*x[2]
        arg[1,0,0]=(0.490893502666)*x[0]**o+(0.853253169019)*x[0]+(-0.706120536551)*x[1]**o+(-0.810665092749)*x[1]+(0.877124690496)*x[2]**o+(-0.663316848595)*x[2]
        arg[1,0,1]=(-0.897301110118)*x[0]**o+(-0.571577796347)*x[0]+(-0.00912216314091)*x[1]**o+(0.479091266837)*x[1]+(-0.951994896347)*x[2]**o+(0.293801256264)*x[2]
        arg[1,1,0]=(-0.0843442242335)*x[0]**o+(0.436592099426)*x[0]+(-0.961835284742)*x[1]**o+(0.453758903497)*x[1]+(0.759916173417)*x[2]**o+(-0.356674191439)*x[2]
        arg[1,1,1]=(0.368337095889)*x[0]**o+(-0.180552461333)*x[0]+(-0.814727889141)*x[1]**o+(-0.226678543654)*x[1]+(0.629233877979)*x[2]**o+(0.934452743806)*x[2]
        arg[2,0,0]=(0.335801632975)*x[0]**o+(0.4853793666)*x[0]+(-0.603075878381)*x[1]**o+(-0.464873655094)*x[1]+(-0.304107114487)*x[2]**o+(0.768564020927)*x[2]
        arg[2,0,1]=(-0.843867260693)*x[0]**o+(-0.758493294188)*x[0]+(0.801820428023)*x[1]**o+(0.0909753752871)*x[1]+(-0.208134327922)*x[2]**o+(0.343325582627)*x[2]
        arg[2,1,0]=(-0.143364423799)*x[0]**o+(-0.191201452642)*x[0]+(0.356480498629)*x[1]**o+(0.0853546566948)*x[1]+(0.234781870759)*x[2]**o+(-0.702906217789)*x[2]
        arg[2,1,1]=(0.81324792898)*x[0]**o+(0.565754643122)*x[0]+(-0.0246314241495)*x[1]**o+(-0.83548496516)*x[1]+(0.730246143504)*x[2]**o+(-0.835539994852)*x[2]
        arg[3,0,0]=(0.877992836393)*x[0]**o+(-0.137135423487)*x[0]+(0.534511286076)*x[1]**o+(-0.0584370047297)*x[1]+(-0.0497236607574)*x[2]**o+(0.408414670632)*x[2]
        arg[3,0,1]=(-0.959771372305)*x[0]**o+(0.587216149992)*x[0]+(-0.873216249139)*x[1]**o+(0.0703350087653)*x[1]+(0.452900901257)*x[2]**o+(-0.180788315472)*x[2]
        arg[3,1,0]=(0.477701526969)*x[0]**o+(0.16165566827)*x[0]+(0.219901428845)*x[1]**o+(-0.0816687269195)*x[1]+(-0.379553348394)*x[2]**o+(-0.764672519612)*x[2]
        arg[3,1,1]=(0.274518799686)*x[0]**o+(0.945009722336)*x[0]+(-0.979787014238)*x[1]**o+(0.587704213162)*x[1]+(0.7959640984)*x[2]**o+(-0.0811112488533)*x[2]
        arg[4,0,0]=(-0.579121056378)*x[0]**o+(0.971799195206)*x[0]+(-0.646271217985)*x[1]**o+(0.369751123548)*x[1]+(0.792425864156)*x[2]**o+(-0.170799235786)*x[2]
        arg[4,0,1]=(0.0939380562355)*x[0]**o+(0.857041359029)*x[0]+(-0.379025882204)*x[1]**o+(-0.571968622949)*x[1]+(-0.156853183697)*x[2]**o+(0.643214485063)*x[2]
        arg[4,1,0]=(-0.715210364871)*x[0]**o+(0.150341666571)*x[0]+(0.325165277323)*x[1]**o+(-0.322744240034)*x[1]+(-0.622094139461)*x[2]**o+(-0.830479146327)*x[2]
        arg[4,1,1]=(-0.697109424406)*x[0]**o+(-0.730318259042)*x[0]+(-0.135119012822)*x[1]**o+(0.410502029229)*x[1]+(0.547987909743)*x[2]**o+(0.144877544254)*x[2]
        arg[5,0,0]=(0.787147597087)*x[0]**o+(-0.528180541295)*x[0]+(-0.251578089108)*x[1]**o+(-0.170766337864)*x[1]+(-0.123341129221)*x[2]**o+(-0.815622069347)*x[2]
        arg[5,0,1]=(0.193873484005)*x[0]**o+(0.270190252915)*x[0]+(0.807946985621)*x[1]**o+(0.184522999461)*x[1]+(0.97429597406)*x[2]**o+(0.532931028021)*x[2]
        arg[5,1,0]=(0.555644676015)*x[0]**o+(-0.225593442089)*x[0]+(0.877869009767)*x[1]**o+(-0.750894245506)*x[1]+(0.888501611997)*x[2]**o+(-0.84027021889)*x[2]
        arg[5,1,1]=(0.76909422867)*x[0]**o+(0.406697645173)*x[0]+(-0.233710263631)*x[1]**o+(-0.432942345121)*x[1]+(0.270484113666)*x[2]**o+(0.707302418115)*x[2]
        ref[0,0,0,0]=o*(0.157678109652)*x_ref[0]**(o-1)+(-0.939095069874)
        ref[0,0,0,1]=o*(0.340072659801)*x_ref[1]**(o-1)+(0.806723712633)
        ref[0,0,0,2]=o*(0.0175654201162)*x_ref[2]**(o-1)+(-0.695887679405)
        ref[0,0,1,0]=o*(0.148444957808)*x_ref[0]**(o-1)+(0.769780554196)
        ref[0,0,1,1]=o*(-0.999778710821)*x_ref[1]**(o-1)+(0.765998508134)
        ref[0,0,1,2]=o*(0.489498280892)*x_ref[2]**(o-1)+(0.272011172233)
        ref[0,1,0,0]=o*(-0.78853623669)*x_ref[0]**(o-1)+(-0.889333747099)
        ref[0,1,0,1]=o*(0.897684283327)*x_ref[1]**(o-1)+(0.878046029595)
        ref[0,1,0,2]=o*(-0.355566201991)*x_ref[2]**(o-1)+(-0.656187599294)
        ref[0,1,1,0]=o*(-0.903496685895)*x_ref[0]**(o-1)+(0.335762730702)
        ref[0,1,1,1]=o*(0.264330545204)*x_ref[1]**(o-1)+(-0.369300150725)
        ref[0,1,1,2]=o*(0.00819339769162)*x_ref[2]**(o-1)+(0.0466274018566)
        ref[1,0,0,0]=o*(0.490893502666)*x_ref[0]**(o-1)+(0.853253169019)
        ref[1,0,0,1]=o*(-0.706120536551)*x_ref[1]**(o-1)+(-0.810665092749)
        ref[1,0,0,2]=o*(0.877124690496)*x_ref[2]**(o-1)+(-0.663316848595)
        ref[1,0,1,0]=o*(-0.897301110118)*x_ref[0]**(o-1)+(-0.571577796347)
        ref[1,0,1,1]=o*(-0.00912216314091)*x_ref[1]**(o-1)+(0.479091266837)
        ref[1,0,1,2]=o*(-0.951994896347)*x_ref[2]**(o-1)+(0.293801256264)
        ref[1,1,0,0]=o*(-0.0843442242335)*x_ref[0]**(o-1)+(0.436592099426)
        ref[1,1,0,1]=o*(-0.961835284742)*x_ref[1]**(o-1)+(0.453758903497)
        ref[1,1,0,2]=o*(0.759916173417)*x_ref[2]**(o-1)+(-0.356674191439)
        ref[1,1,1,0]=o*(0.368337095889)*x_ref[0]**(o-1)+(-0.180552461333)
        ref[1,1,1,1]=o*(-0.814727889141)*x_ref[1]**(o-1)+(-0.226678543654)
        ref[1,1,1,2]=o*(0.629233877979)*x_ref[2]**(o-1)+(0.934452743806)
        ref[2,0,0,0]=o*(0.335801632975)*x_ref[0]**(o-1)+(0.4853793666)
        ref[2,0,0,1]=o*(-0.603075878381)*x_ref[1]**(o-1)+(-0.464873655094)
        ref[2,0,0,2]=o*(-0.304107114487)*x_ref[2]**(o-1)+(0.768564020927)
        ref[2,0,1,0]=o*(-0.843867260693)*x_ref[0]**(o-1)+(-0.758493294188)
        ref[2,0,1,1]=o*(0.801820428023)*x_ref[1]**(o-1)+(0.0909753752871)
        ref[2,0,1,2]=o*(-0.208134327922)*x_ref[2]**(o-1)+(0.343325582627)
        ref[2,1,0,0]=o*(-0.143364423799)*x_ref[0]**(o-1)+(-0.191201452642)
        ref[2,1,0,1]=o*(0.356480498629)*x_ref[1]**(o-1)+(0.0853546566948)
        ref[2,1,0,2]=o*(0.234781870759)*x_ref[2]**(o-1)+(-0.702906217789)
        ref[2,1,1,0]=o*(0.81324792898)*x_ref[0]**(o-1)+(0.565754643122)
        ref[2,1,1,1]=o*(-0.0246314241495)*x_ref[1]**(o-1)+(-0.83548496516)
        ref[2,1,1,2]=o*(0.730246143504)*x_ref[2]**(o-1)+(-0.835539994852)
        ref[3,0,0,0]=o*(0.877992836393)*x_ref[0]**(o-1)+(-0.137135423487)
        ref[3,0,0,1]=o*(0.534511286076)*x_ref[1]**(o-1)+(-0.0584370047297)
        ref[3,0,0,2]=o*(-0.0497236607574)*x_ref[2]**(o-1)+(0.408414670632)
        ref[3,0,1,0]=o*(-0.959771372305)*x_ref[0]**(o-1)+(0.587216149992)
        ref[3,0,1,1]=o*(-0.873216249139)*x_ref[1]**(o-1)+(0.0703350087653)
        ref[3,0,1,2]=o*(0.452900901257)*x_ref[2]**(o-1)+(-0.180788315472)
        ref[3,1,0,0]=o*(0.477701526969)*x_ref[0]**(o-1)+(0.16165566827)
        ref[3,1,0,1]=o*(0.219901428845)*x_ref[1]**(o-1)+(-0.0816687269195)
        ref[3,1,0,2]=o*(-0.379553348394)*x_ref[2]**(o-1)+(-0.764672519612)
        ref[3,1,1,0]=o*(0.274518799686)*x_ref[0]**(o-1)+(0.945009722336)
        ref[3,1,1,1]=o*(-0.979787014238)*x_ref[1]**(o-1)+(0.587704213162)
        ref[3,1,1,2]=o*(0.7959640984)*x_ref[2]**(o-1)+(-0.0811112488533)
        ref[4,0,0,0]=o*(-0.579121056378)*x_ref[0]**(o-1)+(0.971799195206)
        ref[4,0,0,1]=o*(-0.646271217985)*x_ref[1]**(o-1)+(0.369751123548)
        ref[4,0,0,2]=o*(0.792425864156)*x_ref[2]**(o-1)+(-0.170799235786)
        ref[4,0,1,0]=o*(0.0939380562355)*x_ref[0]**(o-1)+(0.857041359029)
        ref[4,0,1,1]=o*(-0.379025882204)*x_ref[1]**(o-1)+(-0.571968622949)
        ref[4,0,1,2]=o*(-0.156853183697)*x_ref[2]**(o-1)+(0.643214485063)
        ref[4,1,0,0]=o*(-0.715210364871)*x_ref[0]**(o-1)+(0.150341666571)
        ref[4,1,0,1]=o*(0.325165277323)*x_ref[1]**(o-1)+(-0.322744240034)
        ref[4,1,0,2]=o*(-0.622094139461)*x_ref[2]**(o-1)+(-0.830479146327)
        ref[4,1,1,0]=o*(-0.697109424406)*x_ref[0]**(o-1)+(-0.730318259042)
        ref[4,1,1,1]=o*(-0.135119012822)*x_ref[1]**(o-1)+(0.410502029229)
        ref[4,1,1,2]=o*(0.547987909743)*x_ref[2]**(o-1)+(0.144877544254)
        ref[5,0,0,0]=o*(0.787147597087)*x_ref[0]**(o-1)+(-0.528180541295)
        ref[5,0,0,1]=o*(-0.251578089108)*x_ref[1]**(o-1)+(-0.170766337864)
        ref[5,0,0,2]=o*(-0.123341129221)*x_ref[2]**(o-1)+(-0.815622069347)
        ref[5,0,1,0]=o*(0.193873484005)*x_ref[0]**(o-1)+(0.270190252915)
        ref[5,0,1,1]=o*(0.807946985621)*x_ref[1]**(o-1)+(0.184522999461)
        ref[5,0,1,2]=o*(0.97429597406)*x_ref[2]**(o-1)+(0.532931028021)
        ref[5,1,0,0]=o*(0.555644676015)*x_ref[0]**(o-1)+(-0.225593442089)
        ref[5,1,0,1]=o*(0.877869009767)*x_ref[1]**(o-1)+(-0.750894245506)
        ref[5,1,0,2]=o*(0.888501611997)*x_ref[2]**(o-1)+(-0.84027021889)
        ref[5,1,1,0]=o*(0.76909422867)*x_ref[0]**(o-1)+(0.406697645173)
        ref[5,1,1,1]=o*(-0.233710263631)*x_ref[1]**(o-1)+(-0.432942345121)
        ref[5,1,1,2]=o*(0.270484113666)*x_ref[2]**(o-1)+(0.707302418115)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedSolution_rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank0(self):
      """
      tests gradient for rank 0 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        arg=(0.390595923727)*x[0]+(0.870359886354)*x[1]
        ref[0]=(0.390595923727)
        ref[1]=(0.870359886354)
      else:
        arg=(1.4237066625)*x[0]+(0.191112329766)*x[1]+(0.342817637558)*x[2]
        ref[0]=(1.4237066625)
        ref[1]=(0.191112329766)
        ref[2]=(0.342817637558)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank1(self):
      """
      tests gradient for rank 1 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        arg[0]=(1.24969786007)*x[0]+(-0.503239623541)*x[1]
        arg[1]=(0.00152786815487)*x[0]+(0.486128960541)*x[1]
        ref[0,0]=(1.24969786007)
        ref[0,1]=(-0.503239623541)
        ref[1,0]=(0.00152786815487)
        ref[1,1]=(0.486128960541)
      else:
        arg[0]=(1.16486019581)*x[0]+(-0.244560585954)*x[1]+(1.11539615819)*x[2]
        arg[1]=(-0.831039160237)*x[0]+(0.822276239005)*x[1]+(0.421468778379)*x[2]
        ref[0,0]=(1.16486019581)
        ref[0,1]=(-0.244560585954)
        ref[0,2]=(1.11539615819)
        ref[1,0]=(-0.831039160237)
        ref[1,1]=(0.822276239005)
        ref[1,2]=(0.421468778379)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank2(self):
      """
      tests gradient for rank 2 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        arg[0,0]=(0.189442592013)*x[0]+(-1.60645028651)*x[1]
        arg[0,1]=(0.0100077079631)*x[0]+(-1.17387880858)*x[1]
        arg[0,2]=(0.0275548120659)*x[0]+(0.653617350295)*x[1]
        arg[0,3]=(0.05829133816)*x[0]+(-0.373563514457)*x[1]
        arg[0,4]=(0.779779569696)*x[0]+(-1.26967503826)*x[1]
        arg[1,0]=(-1.76287668803)*x[0]+(1.09016960913)*x[1]
        arg[1,1]=(-0.37217349459)*x[0]+(-0.241280622851)*x[1]
        arg[1,2]=(-0.590745342582)*x[0]+(0.258098534691)*x[1]
        arg[1,3]=(0.481754329217)*x[0]+(1.9686931184)*x[1]
        arg[1,4]=(1.62318899699)*x[0]+(-0.634644801985)*x[1]
        arg[2,0]=(-0.0113400901549)*x[0]+(0.448967376428)*x[1]
        arg[2,1]=(0.0239277722956)*x[0]+(0.248367174694)*x[1]
        arg[2,2]=(1.23484142782)*x[0]+(-0.408724357055)*x[1]
        arg[2,3]=(0.115412214024)*x[0]+(0.721349239053)*x[1]
        arg[2,4]=(-1.58878458768)*x[0]+(-0.297328738789)*x[1]
        arg[3,0]=(-0.536715000453)*x[0]+(-0.0610624173989)*x[1]
        arg[3,1]=(-0.365981795847)*x[0]+(0.710935216782)*x[1]
        arg[3,2]=(1.1343806743)*x[0]+(-1.12662285264)*x[1]
        arg[3,3]=(-1.56463302165)*x[0]+(0.701651939794)*x[1]
        arg[3,4]=(1.38464443939)*x[0]+(-0.389328824078)*x[1]
        ref[0,0,0]=(0.189442592013)
        ref[0,0,1]=(-1.60645028651)
        ref[0,1,0]=(0.0100077079631)
        ref[0,1,1]=(-1.17387880858)
        ref[0,2,0]=(0.0275548120659)
        ref[0,2,1]=(0.653617350295)
        ref[0,3,0]=(0.05829133816)
        ref[0,3,1]=(-0.373563514457)
        ref[0,4,0]=(0.779779569696)
        ref[0,4,1]=(-1.26967503826)
        ref[1,0,0]=(-1.76287668803)
        ref[1,0,1]=(1.09016960913)
        ref[1,1,0]=(-0.37217349459)
        ref[1,1,1]=(-0.241280622851)
        ref[1,2,0]=(-0.590745342582)
        ref[1,2,1]=(0.258098534691)
        ref[1,3,0]=(0.481754329217)
        ref[1,3,1]=(1.9686931184)
        ref[1,4,0]=(1.62318899699)
        ref[1,4,1]=(-0.634644801985)
        ref[2,0,0]=(-0.0113400901549)
        ref[2,0,1]=(0.448967376428)
        ref[2,1,0]=(0.0239277722956)
        ref[2,1,1]=(0.248367174694)
        ref[2,2,0]=(1.23484142782)
        ref[2,2,1]=(-0.408724357055)
        ref[2,3,0]=(0.115412214024)
        ref[2,3,1]=(0.721349239053)
        ref[2,4,0]=(-1.58878458768)
        ref[2,4,1]=(-0.297328738789)
        ref[3,0,0]=(-0.536715000453)
        ref[3,0,1]=(-0.0610624173989)
        ref[3,1,0]=(-0.365981795847)
        ref[3,1,1]=(0.710935216782)
        ref[3,2,0]=(1.1343806743)
        ref[3,2,1]=(-1.12662285264)
        ref[3,3,0]=(-1.56463302165)
        ref[3,3,1]=(0.701651939794)
        ref[3,4,0]=(1.38464443939)
        ref[3,4,1]=(-0.389328824078)
      else:
        arg[0,0]=(-0.0712739739657)*x[0]+(-0.876441668189)*x[1]+(0.0470586478107)*x[2]
        arg[0,1]=(0.162070750778)*x[0]+(-0.672482005621)*x[1]+(-1.03342931552)*x[2]
        arg[0,2]=(1.06895949733)*x[0]+(-1.10495954108)*x[1]+(-0.689435346922)*x[2]
        arg[0,3]=(-1.64208651194)*x[0]+(0.957715915835)*x[1]+(-0.693562002749)*x[2]
        arg[0,4]=(0.438930190966)*x[0]+(-0.696372759112)*x[1]+(-0.158769579398)*x[2]
        arg[1,0]=(-0.953682332409)*x[0]+(-1.06837831005)*x[1]+(1.19564302893)*x[2]
        arg[1,1]=(0.277654186276)*x[0]+(-0.4746702788)*x[1]+(0.173080688542)*x[2]
        arg[1,2]=(-0.0080483840277)*x[0]+(-0.606889636532)*x[1]+(-0.124482488485)*x[2]
        arg[1,3]=(1.6007700604)*x[0]+(-0.375573101659)*x[1]+(-0.0476989341458)*x[2]
        arg[1,4]=(-0.797213964765)*x[0]+(-1.71649956799)*x[1]+(-0.661820920574)*x[2]
        arg[2,0]=(-0.32770535636)*x[0]+(0.705676452368)*x[1]+(0.0527214815247)*x[2]
        arg[2,1]=(-0.279796835093)*x[0]+(-0.277264709729)*x[1]+(0.278447243546)*x[2]
        arg[2,2]=(-0.633674196984)*x[0]+(-0.0772663082858)*x[1]+(0.973414222834)*x[2]
        arg[2,3]=(0.0946037203409)*x[0]+(-0.556948102801)*x[1]+(0.521264798082)*x[2]
        arg[2,4]=(0.566962914588)*x[0]+(-0.942523827817)*x[1]+(-0.216364192998)*x[2]
        arg[3,0]=(-0.839702033371)*x[0]+(-1.54418367928)*x[1]+(-0.184693363932)*x[2]
        arg[3,1]=(-0.509575652351)*x[0]+(0.0710239576211)*x[1]+(0.758695521775)*x[2]
        arg[3,2]=(0.465980018625)*x[0]+(-0.907796828587)*x[1]+(1.00785361438)*x[2]
        arg[3,3]=(0.778181969842)*x[0]+(-1.45634928421)*x[1]+(1.20377448321)*x[2]
        arg[3,4]=(1.53649652976)*x[0]+(-0.769556905214)*x[1]+(0.736590460805)*x[2]
        ref[0,0,0]=(-0.0712739739657)
        ref[0,0,1]=(-0.876441668189)
        ref[0,0,2]=(0.0470586478107)
        ref[0,1,0]=(0.162070750778)
        ref[0,1,1]=(-0.672482005621)
        ref[0,1,2]=(-1.03342931552)
        ref[0,2,0]=(1.06895949733)
        ref[0,2,1]=(-1.10495954108)
        ref[0,2,2]=(-0.689435346922)
        ref[0,3,0]=(-1.64208651194)
        ref[0,3,1]=(0.957715915835)
        ref[0,3,2]=(-0.693562002749)
        ref[0,4,0]=(0.438930190966)
        ref[0,4,1]=(-0.696372759112)
        ref[0,4,2]=(-0.158769579398)
        ref[1,0,0]=(-0.953682332409)
        ref[1,0,1]=(-1.06837831005)
        ref[1,0,2]=(1.19564302893)
        ref[1,1,0]=(0.277654186276)
        ref[1,1,1]=(-0.4746702788)
        ref[1,1,2]=(0.173080688542)
        ref[1,2,0]=(-0.0080483840277)
        ref[1,2,1]=(-0.606889636532)
        ref[1,2,2]=(-0.124482488485)
        ref[1,3,0]=(1.6007700604)
        ref[1,3,1]=(-0.375573101659)
        ref[1,3,2]=(-0.0476989341458)
        ref[1,4,0]=(-0.797213964765)
        ref[1,4,1]=(-1.71649956799)
        ref[1,4,2]=(-0.661820920574)
        ref[2,0,0]=(-0.32770535636)
        ref[2,0,1]=(0.705676452368)
        ref[2,0,2]=(0.0527214815247)
        ref[2,1,0]=(-0.279796835093)
        ref[2,1,1]=(-0.277264709729)
        ref[2,1,2]=(0.278447243546)
        ref[2,2,0]=(-0.633674196984)
        ref[2,2,1]=(-0.0772663082858)
        ref[2,2,2]=(0.973414222834)
        ref[2,3,0]=(0.0946037203409)
        ref[2,3,1]=(-0.556948102801)
        ref[2,3,2]=(0.521264798082)
        ref[2,4,0]=(0.566962914588)
        ref[2,4,1]=(-0.942523827817)
        ref[2,4,2]=(-0.216364192998)
        ref[3,0,0]=(-0.839702033371)
        ref[3,0,1]=(-1.54418367928)
        ref[3,0,2]=(-0.184693363932)
        ref[3,1,0]=(-0.509575652351)
        ref[3,1,1]=(0.0710239576211)
        ref[3,1,2]=(0.758695521775)
        ref[3,2,0]=(0.465980018625)
        ref[3,2,1]=(-0.907796828587)
        ref[3,2,2]=(1.00785361438)
        ref[3,3,0]=(0.778181969842)
        ref[3,3,1]=(-1.45634928421)
        ref[3,3,2]=(1.20377448321)
        ref[3,4,0]=(1.53649652976)
        ref[3,4,1]=(-0.769556905214)
        ref[3,4,2]=(0.736590460805)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromData_ReducedContinuousFunction__rank3(self):
      """
      tests gradient for rank 3 Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        arg[0,0,0]=(1.39942183968)*x[0]+(0.930250163424)*x[1]
        arg[0,0,1]=(0.784529427298)*x[0]+(0.701356076527)*x[1]
        arg[0,1,0]=(0.555731012919)*x[0]+(0.344587262019)*x[1]
        arg[0,1,1]=(-0.276715609625)*x[0]+(-1.11338983237)*x[1]
        arg[1,0,0]=(-0.769761090097)*x[0]+(0.992057132466)*x[1]
        arg[1,0,1]=(-0.551932009885)*x[0]+(0.838811831137)*x[1]
        arg[1,1,0]=(-0.0342537649661)*x[0]+(-0.121298284661)*x[1]
        arg[1,1,1]=(-0.520900867)*x[0]+(0.332690359269)*x[1]
        arg[2,0,0]=(-0.11096511324)*x[0]+(1.5033138393)*x[1]
        arg[2,0,1]=(0.42710234817)*x[0]+(1.69083527833)*x[1]
        arg[2,1,0]=(-0.721628841933)*x[0]+(-0.142173751601)*x[1]
        arg[2,1,1]=(1.60406199163)*x[0]+(0.641129401426)*x[1]
        arg[3,0,0]=(-0.731738224766)*x[0]+(-0.900425804879)*x[1]
        arg[3,0,1]=(0.691108140204)*x[0]+(1.34043965547)*x[1]
        arg[3,1,0]=(0.0588648299821)*x[0]+(1.39496787747)*x[1]
        arg[3,1,1]=(0.308498671612)*x[0]+(-0.382011042003)*x[1]
        arg[4,0,0]=(0.267730803561)*x[0]+(1.76842225328)*x[1]
        arg[4,0,1]=(0.817593687794)*x[0]+(0.563964345071)*x[1]
        arg[4,1,0]=(-0.081090160095)*x[0]+(0.52343934951)*x[1]
        arg[4,1,1]=(-1.02492946795)*x[0]+(-0.725641086216)*x[1]
        arg[5,0,0]=(-0.294202699958)*x[0]+(0.288095939599)*x[1]
        arg[5,0,1]=(0.110479098614)*x[0]+(-0.113972943318)*x[1]
        arg[5,1,0]=(0.772912536924)*x[0]+(-0.883345081065)*x[1]
        arg[5,1,1]=(-0.17097935184)*x[0]+(0.645719745504)*x[1]
        ref[0,0,0,0]=(1.39942183968)
        ref[0,0,0,1]=(0.930250163424)
        ref[0,0,1,0]=(0.784529427298)
        ref[0,0,1,1]=(0.701356076527)
        ref[0,1,0,0]=(0.555731012919)
        ref[0,1,0,1]=(0.344587262019)
        ref[0,1,1,0]=(-0.276715609625)
        ref[0,1,1,1]=(-1.11338983237)
        ref[1,0,0,0]=(-0.769761090097)
        ref[1,0,0,1]=(0.992057132466)
        ref[1,0,1,0]=(-0.551932009885)
        ref[1,0,1,1]=(0.838811831137)
        ref[1,1,0,0]=(-0.0342537649661)
        ref[1,1,0,1]=(-0.121298284661)
        ref[1,1,1,0]=(-0.520900867)
        ref[1,1,1,1]=(0.332690359269)
        ref[2,0,0,0]=(-0.11096511324)
        ref[2,0,0,1]=(1.5033138393)
        ref[2,0,1,0]=(0.42710234817)
        ref[2,0,1,1]=(1.69083527833)
        ref[2,1,0,0]=(-0.721628841933)
        ref[2,1,0,1]=(-0.142173751601)
        ref[2,1,1,0]=(1.60406199163)
        ref[2,1,1,1]=(0.641129401426)
        ref[3,0,0,0]=(-0.731738224766)
        ref[3,0,0,1]=(-0.900425804879)
        ref[3,0,1,0]=(0.691108140204)
        ref[3,0,1,1]=(1.34043965547)
        ref[3,1,0,0]=(0.0588648299821)
        ref[3,1,0,1]=(1.39496787747)
        ref[3,1,1,0]=(0.308498671612)
        ref[3,1,1,1]=(-0.382011042003)
        ref[4,0,0,0]=(0.267730803561)
        ref[4,0,0,1]=(1.76842225328)
        ref[4,0,1,0]=(0.817593687794)
        ref[4,0,1,1]=(0.563964345071)
        ref[4,1,0,0]=(-0.081090160095)
        ref[4,1,0,1]=(0.52343934951)
        ref[4,1,1,0]=(-1.02492946795)
        ref[4,1,1,1]=(-0.725641086216)
        ref[5,0,0,0]=(-0.294202699958)
        ref[5,0,0,1]=(0.288095939599)
        ref[5,0,1,0]=(0.110479098614)
        ref[5,0,1,1]=(-0.113972943318)
        ref[5,1,0,0]=(0.772912536924)
        ref[5,1,0,1]=(-0.883345081065)
        ref[5,1,1,0]=(-0.17097935184)
        ref[5,1,1,1]=(0.645719745504)
      else:
        arg[0,0,0]=(1.30850237165)*x[0]+(-0.0995793053275)*x[1]+(0.787886140134)*x[2]
        arg[0,0,1]=(-1.50381324493)*x[0]+(1.26609518824)*x[1]+(-0.520747554357)*x[2]
        arg[0,1,0]=(1.29770238108)*x[0]+(-1.46648704595)*x[1]+(-0.917683003599)*x[2]
        arg[0,1,1]=(-1.13525521581)*x[0]+(1.13359501358)*x[1]+(-0.222600272477)*x[2]
        arg[1,0,0]=(0.0106829054463)*x[0]+(-0.00676483709573)*x[1]+(0.109412126311)*x[2]
        arg[1,0,1]=(0.0246128795814)*x[0]+(-0.356585346511)*x[1]+(0.479157114293)*x[2]
        arg[1,1,0]=(0.541787350557)*x[0]+(-0.594934006419)*x[1]+(0.464164867381)*x[2]
        arg[1,1,1]=(-1.12100758712)*x[0]+(1.22017329117)*x[1]+(0.417565754303)*x[2]
        arg[2,0,0]=(-0.902343328228)*x[0]+(-0.118486423959)*x[1]+(0.983457082831)*x[2]
        arg[2,0,1]=(-1.32314473272)*x[0]+(1.09473600277)*x[1]+(0.526840449987)*x[2]
        arg[2,1,0]=(1.02149157674)*x[0]+(0.6830798576)*x[1]+(-0.694642330424)*x[2]
        arg[2,1,1]=(0.471980048259)*x[0]+(1.43051661856)*x[1]+(-0.591151392642)*x[2]
        arg[3,0,0]=(-0.207925182408)*x[0]+(0.644497833444)*x[1]+(0.0356376370926)*x[2]
        arg[3,0,1]=(1.91147142951)*x[0]+(-0.872318227607)*x[1]+(0.714610341379)*x[2]
        arg[3,1,0]=(-0.197729735501)*x[0]+(0.812673023114)*x[1]+(0.637138094981)*x[2]
        arg[3,1,1]=(0.103430412074)*x[0]+(0.69310435757)*x[1]+(1.54404820552)*x[2]
        arg[4,0,0]=(1.04427320711)*x[0]+(-0.655287212961)*x[1]+(0.674873379725)*x[2]
        arg[4,0,1]=(0.0300148606336)*x[0]+(-0.336581334229)*x[1]+(-0.26036930034)*x[2]
        arg[4,1,0]=(0.338077518919)*x[0]+(-0.965310125748)*x[1]+(1.66372736241)*x[2]
        arg[4,1,1]=(-0.417143696516)*x[0]+(-0.764906477186)*x[1]+(-0.406563170367)*x[2]
        arg[5,0,0]=(-0.894638979617)*x[0]+(-1.03496915396)*x[1]+(0.102679227173)*x[2]
        arg[5,0,1]=(0.777823649077)*x[0]+(0.138595909746)*x[1]+(0.570318195247)*x[2]
        arg[5,1,0]=(0.387449776869)*x[0]+(0.782752492387)*x[1]+(0.175162039427)*x[2]
        arg[5,1,1]=(-0.467369934942)*x[0]+(0.343929304184)*x[1]+(0.42227012208)*x[2]
        ref[0,0,0,0]=(1.30850237165)
        ref[0,0,0,1]=(-0.0995793053275)
        ref[0,0,0,2]=(0.787886140134)
        ref[0,0,1,0]=(-1.50381324493)
        ref[0,0,1,1]=(1.26609518824)
        ref[0,0,1,2]=(-0.520747554357)
        ref[0,1,0,0]=(1.29770238108)
        ref[0,1,0,1]=(-1.46648704595)
        ref[0,1,0,2]=(-0.917683003599)
        ref[0,1,1,0]=(-1.13525521581)
        ref[0,1,1,1]=(1.13359501358)
        ref[0,1,1,2]=(-0.222600272477)
        ref[1,0,0,0]=(0.0106829054463)
        ref[1,0,0,1]=(-0.00676483709573)
        ref[1,0,0,2]=(0.109412126311)
        ref[1,0,1,0]=(0.0246128795814)
        ref[1,0,1,1]=(-0.356585346511)
        ref[1,0,1,2]=(0.479157114293)
        ref[1,1,0,0]=(0.541787350557)
        ref[1,1,0,1]=(-0.594934006419)
        ref[1,1,0,2]=(0.464164867381)
        ref[1,1,1,0]=(-1.12100758712)
        ref[1,1,1,1]=(1.22017329117)
        ref[1,1,1,2]=(0.417565754303)
        ref[2,0,0,0]=(-0.902343328228)
        ref[2,0,0,1]=(-0.118486423959)
        ref[2,0,0,2]=(0.983457082831)
        ref[2,0,1,0]=(-1.32314473272)
        ref[2,0,1,1]=(1.09473600277)
        ref[2,0,1,2]=(0.526840449987)
        ref[2,1,0,0]=(1.02149157674)
        ref[2,1,0,1]=(0.6830798576)
        ref[2,1,0,2]=(-0.694642330424)
        ref[2,1,1,0]=(0.471980048259)
        ref[2,1,1,1]=(1.43051661856)
        ref[2,1,1,2]=(-0.591151392642)
        ref[3,0,0,0]=(-0.207925182408)
        ref[3,0,0,1]=(0.644497833444)
        ref[3,0,0,2]=(0.0356376370926)
        ref[3,0,1,0]=(1.91147142951)
        ref[3,0,1,1]=(-0.872318227607)
        ref[3,0,1,2]=(0.714610341379)
        ref[3,1,0,0]=(-0.197729735501)
        ref[3,1,0,1]=(0.812673023114)
        ref[3,1,0,2]=(0.637138094981)
        ref[3,1,1,0]=(0.103430412074)
        ref[3,1,1,1]=(0.69310435757)
        ref[3,1,1,2]=(1.54404820552)
        ref[4,0,0,0]=(1.04427320711)
        ref[4,0,0,1]=(-0.655287212961)
        ref[4,0,0,2]=(0.674873379725)
        ref[4,0,1,0]=(0.0300148606336)
        ref[4,0,1,1]=(-0.336581334229)
        ref[4,0,1,2]=(-0.26036930034)
        ref[4,1,0,0]=(0.338077518919)
        ref[4,1,0,1]=(-0.965310125748)
        ref[4,1,0,2]=(1.66372736241)
        ref[4,1,1,0]=(-0.417143696516)
        ref[4,1,1,1]=(-0.764906477186)
        ref[4,1,1,2]=(-0.406563170367)
        ref[5,0,0,0]=(-0.894638979617)
        ref[5,0,0,1]=(-1.03496915396)
        ref[5,0,0,2]=(0.102679227173)
        ref[5,0,1,0]=(0.777823649077)
        ref[5,0,1,1]=(0.138595909746)
        ref[5,0,1,2]=(0.570318195247)
        ref[5,1,0,0]=(0.387449776869)
        ref[5,1,0,1]=(0.782752492387)
        ref[5,1,0,2]=(0.175162039427)
        ref[5,1,1,0]=(-0.467369934942)
        ref[5,1,1,1]=(0.343929304184)
        ref[5,1,1,2]=(0.42227012208)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.662618392141)*x[0]**o+(-0.42617914202)*x[0]+(-0.419315426864)*x[1]**o+(-0.63613129268)*x[1]
        ref[0]=o*(-0.662618392141)*x_ref[0]**(o-1)+(-0.42617914202)
        ref[1]=o*(-0.419315426864)*x_ref[1]**(o-1)+(-0.63613129268)
      else:
        s=(-0.762028198747)*x[0]**o+(0.14263183612)*x[0]+(0.185743923374)*x[1]**o+(-0.75903441887)*x[1]+(0.77116503084)*x[2]**o+(0.404561994654)*x[2]
        ref[0]=o*(-0.762028198747)*x_ref[0]**(o-1)+(0.14263183612)
        ref[1]=o*(0.185743923374)*x_ref[1]**(o-1)+(-0.75903441887)
        ref[2]=o*(0.77116503084)*x_ref[2]**(o-1)+(0.404561994654)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(0.876893099705)*x[0]**o+(-0.177079224798)*x[0]+(-0.527304645017)*x[1]**o+(-0.0349448735698)*x[1]
        s[1]=(-0.534225377579)*x[0]**o+(-0.280049692517)*x[0]+(0.912290702798)*x[1]**o+(-0.236862709157)*x[1]
        ref[0,0]=o*(0.876893099705)*x_ref[0]**(o-1)+(-0.177079224798)
        ref[0,1]=o*(-0.527304645017)*x_ref[1]**(o-1)+(-0.0349448735698)
        ref[1,0]=o*(-0.534225377579)*x_ref[0]**(o-1)+(-0.280049692517)
        ref[1,1]=o*(0.912290702798)*x_ref[1]**(o-1)+(-0.236862709157)
      else:
        s[0]=(-0.254550799071)*x[0]**o+(0.163641443345)*x[0]+(0.224143736485)*x[1]**o+(0.536806485124)*x[1]+(0.195320362214)*x[2]**o+(-0.0914207464578)*x[2]
        s[1]=(0.708798134546)*x[0]**o+(-0.572385005996)*x[0]+(-0.28510012443)*x[1]**o+(-0.305494018321)*x[1]+(-0.944999853945)*x[2]**o+(0.690936294379)*x[2]
        ref[0,0]=o*(-0.254550799071)*x_ref[0]**(o-1)+(0.163641443345)
        ref[0,1]=o*(0.224143736485)*x_ref[1]**(o-1)+(0.536806485124)
        ref[0,2]=o*(0.195320362214)*x_ref[2]**(o-1)+(-0.0914207464578)
        ref[1,0]=o*(0.708798134546)*x_ref[0]**(o-1)+(-0.572385005996)
        ref[1,1]=o*(-0.28510012443)*x_ref[1]**(o-1)+(-0.305494018321)
        ref[1,2]=o*(-0.944999853945)*x_ref[2]**(o-1)+(0.690936294379)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.555539333621)*x[0]**o+(0.568177319612)*x[0]+(-0.515690028057)*x[1]**o+(0.293250590603)*x[1]
        s[0,1]=(0.7906503945)*x[0]**o+(-0.167319950519)*x[0]+(-0.554762741045)*x[1]**o+(-0.669125892384)*x[1]
        s[0,2]=(-0.309865666451)*x[0]**o+(0.419499273794)*x[0]+(-0.563409144055)*x[1]**o+(0.734002122712)*x[1]
        s[0,3]=(-0.93657260436)*x[0]**o+(0.586601294944)*x[0]+(-0.0753458445127)*x[1]**o+(-0.429871038581)*x[1]
        s[0,4]=(-0.198003364647)*x[0]**o+(-0.76482793062)*x[0]+(-0.125188294649)*x[1]**o+(0.385219545236)*x[1]
        s[1,0]=(-0.179923035467)*x[0]**o+(0.962110768245)*x[0]+(0.422898666863)*x[1]**o+(-0.92193819062)*x[1]
        s[1,1]=(-0.490610670554)*x[0]**o+(-0.876474842232)*x[0]+(-0.870535201491)*x[1]**o+(-0.208576590799)*x[1]
        s[1,2]=(0.625747905975)*x[0]**o+(0.766151918225)*x[0]+(-0.865271283252)*x[1]**o+(-0.984130648049)*x[1]
        s[1,3]=(0.21674453499)*x[0]**o+(0.74684159223)*x[0]+(0.968938733366)*x[1]**o+(-0.980174574707)*x[1]
        s[1,4]=(0.15542486304)*x[0]**o+(-0.288729546241)*x[0]+(0.780533404805)*x[1]**o+(0.201416001866)*x[1]
        s[2,0]=(0.353102047108)*x[0]**o+(-0.388306135155)*x[0]+(0.837210776206)*x[1]**o+(0.775323252612)*x[1]
        s[2,1]=(-0.407280915312)*x[0]**o+(-0.53768541262)*x[0]+(0.893214604403)*x[1]**o+(-0.69705043663)*x[1]
        s[2,2]=(-0.447169656337)*x[0]**o+(0.728088534193)*x[0]+(0.0485994009494)*x[1]**o+(-0.503223824878)*x[1]
        s[2,3]=(-0.19757745902)*x[0]**o+(-0.627533475183)*x[0]+(0.962502499281)*x[1]**o+(0.859404568594)*x[1]
        s[2,4]=(-0.796987077625)*x[0]**o+(0.0568885495838)*x[0]+(0.84204928286)*x[1]**o+(0.521081919225)*x[1]
        s[3,0]=(-0.310704277911)*x[0]**o+(0.768329284718)*x[0]+(0.348908427651)*x[1]**o+(0.403870270277)*x[1]
        s[3,1]=(0.655277014208)*x[0]**o+(-0.770568143234)*x[0]+(0.101447715545)*x[1]**o+(0.525372367142)*x[1]
        s[3,2]=(-0.450815491992)*x[0]**o+(0.267341563784)*x[0]+(0.666360532863)*x[1]**o+(0.25468870732)*x[1]
        s[3,3]=(-0.448798233791)*x[0]**o+(0.825032894501)*x[0]+(0.0555349523209)*x[1]**o+(0.168286007255)*x[1]
        s[3,4]=(-0.260484305029)*x[0]**o+(0.265463408379)*x[0]+(0.147356758821)*x[1]**o+(-0.916280010272)*x[1]
        ref[0,0,0]=o*(-0.555539333621)*x_ref[0]**(o-1)+(0.568177319612)
        ref[0,0,1]=o*(-0.515690028057)*x_ref[1]**(o-1)+(0.293250590603)
        ref[0,1,0]=o*(0.7906503945)*x_ref[0]**(o-1)+(-0.167319950519)
        ref[0,1,1]=o*(-0.554762741045)*x_ref[1]**(o-1)+(-0.669125892384)
        ref[0,2,0]=o*(-0.309865666451)*x_ref[0]**(o-1)+(0.419499273794)
        ref[0,2,1]=o*(-0.563409144055)*x_ref[1]**(o-1)+(0.734002122712)
        ref[0,3,0]=o*(-0.93657260436)*x_ref[0]**(o-1)+(0.586601294944)
        ref[0,3,1]=o*(-0.0753458445127)*x_ref[1]**(o-1)+(-0.429871038581)
        ref[0,4,0]=o*(-0.198003364647)*x_ref[0]**(o-1)+(-0.76482793062)
        ref[0,4,1]=o*(-0.125188294649)*x_ref[1]**(o-1)+(0.385219545236)
        ref[1,0,0]=o*(-0.179923035467)*x_ref[0]**(o-1)+(0.962110768245)
        ref[1,0,1]=o*(0.422898666863)*x_ref[1]**(o-1)+(-0.92193819062)
        ref[1,1,0]=o*(-0.490610670554)*x_ref[0]**(o-1)+(-0.876474842232)
        ref[1,1,1]=o*(-0.870535201491)*x_ref[1]**(o-1)+(-0.208576590799)
        ref[1,2,0]=o*(0.625747905975)*x_ref[0]**(o-1)+(0.766151918225)
        ref[1,2,1]=o*(-0.865271283252)*x_ref[1]**(o-1)+(-0.984130648049)
        ref[1,3,0]=o*(0.21674453499)*x_ref[0]**(o-1)+(0.74684159223)
        ref[1,3,1]=o*(0.968938733366)*x_ref[1]**(o-1)+(-0.980174574707)
        ref[1,4,0]=o*(0.15542486304)*x_ref[0]**(o-1)+(-0.288729546241)
        ref[1,4,1]=o*(0.780533404805)*x_ref[1]**(o-1)+(0.201416001866)
        ref[2,0,0]=o*(0.353102047108)*x_ref[0]**(o-1)+(-0.388306135155)
        ref[2,0,1]=o*(0.837210776206)*x_ref[1]**(o-1)+(0.775323252612)
        ref[2,1,0]=o*(-0.407280915312)*x_ref[0]**(o-1)+(-0.53768541262)
        ref[2,1,1]=o*(0.893214604403)*x_ref[1]**(o-1)+(-0.69705043663)
        ref[2,2,0]=o*(-0.447169656337)*x_ref[0]**(o-1)+(0.728088534193)
        ref[2,2,1]=o*(0.0485994009494)*x_ref[1]**(o-1)+(-0.503223824878)
        ref[2,3,0]=o*(-0.19757745902)*x_ref[0]**(o-1)+(-0.627533475183)
        ref[2,3,1]=o*(0.962502499281)*x_ref[1]**(o-1)+(0.859404568594)
        ref[2,4,0]=o*(-0.796987077625)*x_ref[0]**(o-1)+(0.0568885495838)
        ref[2,4,1]=o*(0.84204928286)*x_ref[1]**(o-1)+(0.521081919225)
        ref[3,0,0]=o*(-0.310704277911)*x_ref[0]**(o-1)+(0.768329284718)
        ref[3,0,1]=o*(0.348908427651)*x_ref[1]**(o-1)+(0.403870270277)
        ref[3,1,0]=o*(0.655277014208)*x_ref[0]**(o-1)+(-0.770568143234)
        ref[3,1,1]=o*(0.101447715545)*x_ref[1]**(o-1)+(0.525372367142)
        ref[3,2,0]=o*(-0.450815491992)*x_ref[0]**(o-1)+(0.267341563784)
        ref[3,2,1]=o*(0.666360532863)*x_ref[1]**(o-1)+(0.25468870732)
        ref[3,3,0]=o*(-0.448798233791)*x_ref[0]**(o-1)+(0.825032894501)
        ref[3,3,1]=o*(0.0555349523209)*x_ref[1]**(o-1)+(0.168286007255)
        ref[3,4,0]=o*(-0.260484305029)*x_ref[0]**(o-1)+(0.265463408379)
        ref[3,4,1]=o*(0.147356758821)*x_ref[1]**(o-1)+(-0.916280010272)
      else:
        s[0,0]=(0.344242828216)*x[0]**o+(0.491144523585)*x[0]+(-0.92238663597)*x[1]**o+(0.655901250824)*x[1]+(0.683755311378)*x[2]**o+(-0.499214028484)*x[2]
        s[0,1]=(0.246950598177)*x[0]**o+(0.690983811313)*x[0]+(0.0441182184921)*x[1]**o+(-0.236074195046)*x[1]+(-0.365664761621)*x[2]**o+(0.406464337338)*x[2]
        s[0,2]=(-0.327739540606)*x[0]**o+(-0.96200067317)*x[0]+(-0.634811990063)*x[1]**o+(-0.0028455698091)*x[1]+(-0.773566753548)*x[2]**o+(-0.944614851063)*x[2]
        s[0,3]=(-0.423726389514)*x[0]**o+(0.930379169254)*x[0]+(0.551259733694)*x[1]**o+(-0.277405976565)*x[1]+(0.429454682659)*x[2]**o+(0.665041921999)*x[2]
        s[0,4]=(-0.808919025592)*x[0]**o+(0.527988478329)*x[0]+(0.598427398457)*x[1]**o+(-0.740120252678)*x[1]+(-0.57495557526)*x[2]**o+(-0.09358843573)*x[2]
        s[1,0]=(0.139579066196)*x[0]**o+(0.64838866949)*x[0]+(-0.310649190227)*x[1]**o+(-0.947998845563)*x[1]+(0.15735741361)*x[2]**o+(0.521031688881)*x[2]
        s[1,1]=(0.3564671327)*x[0]**o+(0.44368001644)*x[0]+(-0.867803053218)*x[1]**o+(-0.808576642073)*x[1]+(0.164399621384)*x[2]**o+(-0.929324362594)*x[2]
        s[1,2]=(-0.719971382108)*x[0]**o+(-0.509264166588)*x[0]+(0.880679383007)*x[1]**o+(0.150377261228)*x[1]+(-0.695573153702)*x[2]**o+(-0.504859578896)*x[2]
        s[1,3]=(0.19999237262)*x[0]**o+(0.301010290809)*x[0]+(-0.761383660349)*x[1]**o+(-0.943121991098)*x[1]+(-0.824426064219)*x[2]**o+(-0.288809145407)*x[2]
        s[1,4]=(0.0209332296639)*x[0]**o+(-0.745572124803)*x[0]+(0.649794969638)*x[1]**o+(0.997896690337)*x[1]+(0.245116895509)*x[2]**o+(0.934427411594)*x[2]
        s[2,0]=(-0.415649481447)*x[0]**o+(0.967011773751)*x[0]+(0.850275812024)*x[1]**o+(0.368460051013)*x[1]+(-0.895648484555)*x[2]**o+(-0.748318156303)*x[2]
        s[2,1]=(0.822253194356)*x[0]**o+(-0.209207542819)*x[0]+(-0.323961739942)*x[1]**o+(-0.108506915189)*x[1]+(0.39419174146)*x[2]**o+(-0.880778942413)*x[2]
        s[2,2]=(0.730106025416)*x[0]**o+(-0.0921212828406)*x[0]+(-0.0790190498056)*x[1]**o+(0.456083884786)*x[1]+(-0.226537921015)*x[2]**o+(0.293694285152)*x[2]
        s[2,3]=(-0.978825730702)*x[0]**o+(-0.914905098143)*x[0]+(-0.110991764399)*x[1]**o+(-0.144744770355)*x[1]+(0.625189542344)*x[2]**o+(0.804751738687)*x[2]
        s[2,4]=(0.862409094298)*x[0]**o+(-0.953204890595)*x[0]+(-0.66083629488)*x[1]**o+(0.892835328867)*x[1]+(-0.724419944382)*x[2]**o+(-0.352952976454)*x[2]
        s[3,0]=(0.518573942986)*x[0]**o+(0.525366624319)*x[0]+(-0.184521233411)*x[1]**o+(-0.577732664805)*x[1]+(0.194495181041)*x[2]**o+(-0.651069449215)*x[2]
        s[3,1]=(-0.900134328609)*x[0]**o+(-0.502594184837)*x[0]+(0.972006928979)*x[1]**o+(-0.58560991354)*x[1]+(0.423531955361)*x[2]**o+(-0.149983453241)*x[2]
        s[3,2]=(-0.949776951808)*x[0]**o+(-0.557083489688)*x[0]+(-0.860426216679)*x[1]**o+(0.740745372034)*x[1]+(-0.770598389143)*x[2]**o+(0.576226272811)*x[2]
        s[3,3]=(0.947181609851)*x[0]**o+(0.624290827777)*x[0]+(0.163876818237)*x[1]**o+(-0.881648295139)*x[1]+(0.460250347501)*x[2]**o+(0.278384518072)*x[2]
        s[3,4]=(-0.235115463614)*x[0]**o+(-0.608162950212)*x[0]+(-0.942527781961)*x[1]**o+(-0.473027994027)*x[1]+(0.784701664346)*x[2]**o+(0.67497154466)*x[2]
        ref[0,0,0]=o*(0.344242828216)*x_ref[0]**(o-1)+(0.491144523585)
        ref[0,0,1]=o*(-0.92238663597)*x_ref[1]**(o-1)+(0.655901250824)
        ref[0,0,2]=o*(0.683755311378)*x_ref[2]**(o-1)+(-0.499214028484)
        ref[0,1,0]=o*(0.246950598177)*x_ref[0]**(o-1)+(0.690983811313)
        ref[0,1,1]=o*(0.0441182184921)*x_ref[1]**(o-1)+(-0.236074195046)
        ref[0,1,2]=o*(-0.365664761621)*x_ref[2]**(o-1)+(0.406464337338)
        ref[0,2,0]=o*(-0.327739540606)*x_ref[0]**(o-1)+(-0.96200067317)
        ref[0,2,1]=o*(-0.634811990063)*x_ref[1]**(o-1)+(-0.0028455698091)
        ref[0,2,2]=o*(-0.773566753548)*x_ref[2]**(o-1)+(-0.944614851063)
        ref[0,3,0]=o*(-0.423726389514)*x_ref[0]**(o-1)+(0.930379169254)
        ref[0,3,1]=o*(0.551259733694)*x_ref[1]**(o-1)+(-0.277405976565)
        ref[0,3,2]=o*(0.429454682659)*x_ref[2]**(o-1)+(0.665041921999)
        ref[0,4,0]=o*(-0.808919025592)*x_ref[0]**(o-1)+(0.527988478329)
        ref[0,4,1]=o*(0.598427398457)*x_ref[1]**(o-1)+(-0.740120252678)
        ref[0,4,2]=o*(-0.57495557526)*x_ref[2]**(o-1)+(-0.09358843573)
        ref[1,0,0]=o*(0.139579066196)*x_ref[0]**(o-1)+(0.64838866949)
        ref[1,0,1]=o*(-0.310649190227)*x_ref[1]**(o-1)+(-0.947998845563)
        ref[1,0,2]=o*(0.15735741361)*x_ref[2]**(o-1)+(0.521031688881)
        ref[1,1,0]=o*(0.3564671327)*x_ref[0]**(o-1)+(0.44368001644)
        ref[1,1,1]=o*(-0.867803053218)*x_ref[1]**(o-1)+(-0.808576642073)
        ref[1,1,2]=o*(0.164399621384)*x_ref[2]**(o-1)+(-0.929324362594)
        ref[1,2,0]=o*(-0.719971382108)*x_ref[0]**(o-1)+(-0.509264166588)
        ref[1,2,1]=o*(0.880679383007)*x_ref[1]**(o-1)+(0.150377261228)
        ref[1,2,2]=o*(-0.695573153702)*x_ref[2]**(o-1)+(-0.504859578896)
        ref[1,3,0]=o*(0.19999237262)*x_ref[0]**(o-1)+(0.301010290809)
        ref[1,3,1]=o*(-0.761383660349)*x_ref[1]**(o-1)+(-0.943121991098)
        ref[1,3,2]=o*(-0.824426064219)*x_ref[2]**(o-1)+(-0.288809145407)
        ref[1,4,0]=o*(0.0209332296639)*x_ref[0]**(o-1)+(-0.745572124803)
        ref[1,4,1]=o*(0.649794969638)*x_ref[1]**(o-1)+(0.997896690337)
        ref[1,4,2]=o*(0.245116895509)*x_ref[2]**(o-1)+(0.934427411594)
        ref[2,0,0]=o*(-0.415649481447)*x_ref[0]**(o-1)+(0.967011773751)
        ref[2,0,1]=o*(0.850275812024)*x_ref[1]**(o-1)+(0.368460051013)
        ref[2,0,2]=o*(-0.895648484555)*x_ref[2]**(o-1)+(-0.748318156303)
        ref[2,1,0]=o*(0.822253194356)*x_ref[0]**(o-1)+(-0.209207542819)
        ref[2,1,1]=o*(-0.323961739942)*x_ref[1]**(o-1)+(-0.108506915189)
        ref[2,1,2]=o*(0.39419174146)*x_ref[2]**(o-1)+(-0.880778942413)
        ref[2,2,0]=o*(0.730106025416)*x_ref[0]**(o-1)+(-0.0921212828406)
        ref[2,2,1]=o*(-0.0790190498056)*x_ref[1]**(o-1)+(0.456083884786)
        ref[2,2,2]=o*(-0.226537921015)*x_ref[2]**(o-1)+(0.293694285152)
        ref[2,3,0]=o*(-0.978825730702)*x_ref[0]**(o-1)+(-0.914905098143)
        ref[2,3,1]=o*(-0.110991764399)*x_ref[1]**(o-1)+(-0.144744770355)
        ref[2,3,2]=o*(0.625189542344)*x_ref[2]**(o-1)+(0.804751738687)
        ref[2,4,0]=o*(0.862409094298)*x_ref[0]**(o-1)+(-0.953204890595)
        ref[2,4,1]=o*(-0.66083629488)*x_ref[1]**(o-1)+(0.892835328867)
        ref[2,4,2]=o*(-0.724419944382)*x_ref[2]**(o-1)+(-0.352952976454)
        ref[3,0,0]=o*(0.518573942986)*x_ref[0]**(o-1)+(0.525366624319)
        ref[3,0,1]=o*(-0.184521233411)*x_ref[1]**(o-1)+(-0.577732664805)
        ref[3,0,2]=o*(0.194495181041)*x_ref[2]**(o-1)+(-0.651069449215)
        ref[3,1,0]=o*(-0.900134328609)*x_ref[0]**(o-1)+(-0.502594184837)
        ref[3,1,1]=o*(0.972006928979)*x_ref[1]**(o-1)+(-0.58560991354)
        ref[3,1,2]=o*(0.423531955361)*x_ref[2]**(o-1)+(-0.149983453241)
        ref[3,2,0]=o*(-0.949776951808)*x_ref[0]**(o-1)+(-0.557083489688)
        ref[3,2,1]=o*(-0.860426216679)*x_ref[1]**(o-1)+(0.740745372034)
        ref[3,2,2]=o*(-0.770598389143)*x_ref[2]**(o-1)+(0.576226272811)
        ref[3,3,0]=o*(0.947181609851)*x_ref[0]**(o-1)+(0.624290827777)
        ref[3,3,1]=o*(0.163876818237)*x_ref[1]**(o-1)+(-0.881648295139)
        ref[3,3,2]=o*(0.460250347501)*x_ref[2]**(o-1)+(0.278384518072)
        ref[3,4,0]=o*(-0.235115463614)*x_ref[0]**(o-1)+(-0.608162950212)
        ref[3,4,1]=o*(-0.942527781961)*x_ref[1]**(o-1)+(-0.473027994027)
        ref[3,4,2]=o*(0.784701664346)*x_ref[2]**(o-1)+(0.67497154466)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(-0.19873307647)*x[0]**o+(-0.402324928722)*x[0]+(0.78581260176)*x[1]**o+(0.855692742935)*x[1]
        s[0,0,1]=(-0.976563405673)*x[0]**o+(0.100149365141)*x[0]+(0.630689651275)*x[1]**o+(0.601727769629)*x[1]
        s[0,1,0]=(-0.832835486573)*x[0]**o+(0.0671587043511)*x[0]+(-0.610210594342)*x[1]**o+(-0.436026877569)*x[1]
        s[0,1,1]=(0.292942538448)*x[0]**o+(0.0385459359421)*x[0]+(0.790299248934)*x[1]**o+(-0.319413494032)*x[1]
        s[1,0,0]=(0.58698003273)*x[0]**o+(0.613088892876)*x[0]+(0.552617483007)*x[1]**o+(0.720306398451)*x[1]
        s[1,0,1]=(0.955352789796)*x[0]**o+(0.174378913925)*x[0]+(-0.0314000136685)*x[1]**o+(0.959578237088)*x[1]
        s[1,1,0]=(0.281022820802)*x[0]**o+(0.326052665369)*x[0]+(-0.492525348509)*x[1]**o+(-0.466634839923)*x[1]
        s[1,1,1]=(0.479963735062)*x[0]**o+(-0.779437072555)*x[0]+(-0.208397089582)*x[1]**o+(-0.303640352239)*x[1]
        s[2,0,0]=(0.434668114042)*x[0]**o+(0.0691902144608)*x[0]+(-0.177103703295)*x[1]**o+(0.91338709789)*x[1]
        s[2,0,1]=(-0.476825738275)*x[0]**o+(0.819380857167)*x[0]+(-0.661832108828)*x[1]**o+(0.873879046431)*x[1]
        s[2,1,0]=(0.30919914009)*x[0]**o+(-0.0366057876782)*x[0]+(-0.966956881285)*x[1]**o+(-0.829788683577)*x[1]
        s[2,1,1]=(-0.168382165296)*x[0]**o+(-0.682198962507)*x[0]+(0.190572223366)*x[1]**o+(-0.667568574676)*x[1]
        s[3,0,0]=(0.668658559933)*x[0]**o+(-0.253060573754)*x[0]+(-0.864766645952)*x[1]**o+(-0.714197991697)*x[1]
        s[3,0,1]=(0.538191340464)*x[0]**o+(0.806266803991)*x[0]+(0.437103643795)*x[1]**o+(0.972593986026)*x[1]
        s[3,1,0]=(-0.120672797742)*x[0]**o+(0.840236886959)*x[0]+(-0.279061306399)*x[1]**o+(0.119529093461)*x[1]
        s[3,1,1]=(-0.405723067833)*x[0]**o+(-0.964291783264)*x[0]+(0.201561553981)*x[1]**o+(0.882148297909)*x[1]
        s[4,0,0]=(-0.501360654417)*x[0]**o+(-0.977210439162)*x[0]+(0.0742786636276)*x[1]**o+(-0.831462191269)*x[1]
        s[4,0,1]=(-0.354758190164)*x[0]**o+(0.737302507124)*x[0]+(-0.221806769305)*x[1]**o+(-0.0261943694435)*x[1]
        s[4,1,0]=(-0.921367050665)*x[0]**o+(-0.77028348328)*x[0]+(-0.799765110774)*x[1]**o+(-0.708684233564)*x[1]
        s[4,1,1]=(-0.764817738089)*x[0]**o+(0.13308831245)*x[0]+(-0.0835084988678)*x[1]**o+(0.181328374835)*x[1]
        s[5,0,0]=(0.0795012634978)*x[0]**o+(0.527681277019)*x[0]+(-0.0656230383346)*x[1]**o+(0.0884801523765)*x[1]
        s[5,0,1]=(-0.316260178996)*x[0]**o+(-0.372232635711)*x[0]+(0.868288352534)*x[1]**o+(0.22823979163)*x[1]
        s[5,1,0]=(0.655698586327)*x[0]**o+(-0.734551617528)*x[0]+(-0.627595993352)*x[1]**o+(0.485738750199)*x[1]
        s[5,1,1]=(-0.522687357394)*x[0]**o+(-0.232523849025)*x[0]+(0.559349644178)*x[1]**o+(-0.640468984193)*x[1]
        ref[0,0,0,0]=o*(-0.19873307647)*x_ref[0]**(o-1)+(-0.402324928722)
        ref[0,0,0,1]=o*(0.78581260176)*x_ref[1]**(o-1)+(0.855692742935)
        ref[0,0,1,0]=o*(-0.976563405673)*x_ref[0]**(o-1)+(0.100149365141)
        ref[0,0,1,1]=o*(0.630689651275)*x_ref[1]**(o-1)+(0.601727769629)
        ref[0,1,0,0]=o*(-0.832835486573)*x_ref[0]**(o-1)+(0.0671587043511)
        ref[0,1,0,1]=o*(-0.610210594342)*x_ref[1]**(o-1)+(-0.436026877569)
        ref[0,1,1,0]=o*(0.292942538448)*x_ref[0]**(o-1)+(0.0385459359421)
        ref[0,1,1,1]=o*(0.790299248934)*x_ref[1]**(o-1)+(-0.319413494032)
        ref[1,0,0,0]=o*(0.58698003273)*x_ref[0]**(o-1)+(0.613088892876)
        ref[1,0,0,1]=o*(0.552617483007)*x_ref[1]**(o-1)+(0.720306398451)
        ref[1,0,1,0]=o*(0.955352789796)*x_ref[0]**(o-1)+(0.174378913925)
        ref[1,0,1,1]=o*(-0.0314000136685)*x_ref[1]**(o-1)+(0.959578237088)
        ref[1,1,0,0]=o*(0.281022820802)*x_ref[0]**(o-1)+(0.326052665369)
        ref[1,1,0,1]=o*(-0.492525348509)*x_ref[1]**(o-1)+(-0.466634839923)
        ref[1,1,1,0]=o*(0.479963735062)*x_ref[0]**(o-1)+(-0.779437072555)
        ref[1,1,1,1]=o*(-0.208397089582)*x_ref[1]**(o-1)+(-0.303640352239)
        ref[2,0,0,0]=o*(0.434668114042)*x_ref[0]**(o-1)+(0.0691902144608)
        ref[2,0,0,1]=o*(-0.177103703295)*x_ref[1]**(o-1)+(0.91338709789)
        ref[2,0,1,0]=o*(-0.476825738275)*x_ref[0]**(o-1)+(0.819380857167)
        ref[2,0,1,1]=o*(-0.661832108828)*x_ref[1]**(o-1)+(0.873879046431)
        ref[2,1,0,0]=o*(0.30919914009)*x_ref[0]**(o-1)+(-0.0366057876782)
        ref[2,1,0,1]=o*(-0.966956881285)*x_ref[1]**(o-1)+(-0.829788683577)
        ref[2,1,1,0]=o*(-0.168382165296)*x_ref[0]**(o-1)+(-0.682198962507)
        ref[2,1,1,1]=o*(0.190572223366)*x_ref[1]**(o-1)+(-0.667568574676)
        ref[3,0,0,0]=o*(0.668658559933)*x_ref[0]**(o-1)+(-0.253060573754)
        ref[3,0,0,1]=o*(-0.864766645952)*x_ref[1]**(o-1)+(-0.714197991697)
        ref[3,0,1,0]=o*(0.538191340464)*x_ref[0]**(o-1)+(0.806266803991)
        ref[3,0,1,1]=o*(0.437103643795)*x_ref[1]**(o-1)+(0.972593986026)
        ref[3,1,0,0]=o*(-0.120672797742)*x_ref[0]**(o-1)+(0.840236886959)
        ref[3,1,0,1]=o*(-0.279061306399)*x_ref[1]**(o-1)+(0.119529093461)
        ref[3,1,1,0]=o*(-0.405723067833)*x_ref[0]**(o-1)+(-0.964291783264)
        ref[3,1,1,1]=o*(0.201561553981)*x_ref[1]**(o-1)+(0.882148297909)
        ref[4,0,0,0]=o*(-0.501360654417)*x_ref[0]**(o-1)+(-0.977210439162)
        ref[4,0,0,1]=o*(0.0742786636276)*x_ref[1]**(o-1)+(-0.831462191269)
        ref[4,0,1,0]=o*(-0.354758190164)*x_ref[0]**(o-1)+(0.737302507124)
        ref[4,0,1,1]=o*(-0.221806769305)*x_ref[1]**(o-1)+(-0.0261943694435)
        ref[4,1,0,0]=o*(-0.921367050665)*x_ref[0]**(o-1)+(-0.77028348328)
        ref[4,1,0,1]=o*(-0.799765110774)*x_ref[1]**(o-1)+(-0.708684233564)
        ref[4,1,1,0]=o*(-0.764817738089)*x_ref[0]**(o-1)+(0.13308831245)
        ref[4,1,1,1]=o*(-0.0835084988678)*x_ref[1]**(o-1)+(0.181328374835)
        ref[5,0,0,0]=o*(0.0795012634978)*x_ref[0]**(o-1)+(0.527681277019)
        ref[5,0,0,1]=o*(-0.0656230383346)*x_ref[1]**(o-1)+(0.0884801523765)
        ref[5,0,1,0]=o*(-0.316260178996)*x_ref[0]**(o-1)+(-0.372232635711)
        ref[5,0,1,1]=o*(0.868288352534)*x_ref[1]**(o-1)+(0.22823979163)
        ref[5,1,0,0]=o*(0.655698586327)*x_ref[0]**(o-1)+(-0.734551617528)
        ref[5,1,0,1]=o*(-0.627595993352)*x_ref[1]**(o-1)+(0.485738750199)
        ref[5,1,1,0]=o*(-0.522687357394)*x_ref[0]**(o-1)+(-0.232523849025)
        ref[5,1,1,1]=o*(0.559349644178)*x_ref[1]**(o-1)+(-0.640468984193)
      else:
        s[0,0,0]=(-0.252254362053)*x[0]**o+(-0.73201258434)*x[0]+(0.129279219807)*x[1]**o+(0.739511847421)*x[1]+(-0.335757163655)*x[2]**o+(0.645975465365)*x[2]
        s[0,0,1]=(-0.735055416054)*x[0]**o+(-0.392972979059)*x[0]+(-0.400312505739)*x[1]**o+(-0.431371712491)*x[1]+(0.570395624547)*x[2]**o+(0.138472351049)*x[2]
        s[0,1,0]=(-0.000809920184319)*x[0]**o+(0.39610549699)*x[0]+(0.46183657032)*x[1]**o+(-0.90835208084)*x[1]+(0.359160987844)*x[2]**o+(0.63185061127)*x[2]
        s[0,1,1]=(0.627804065013)*x[0]**o+(-0.616378680528)*x[0]+(-0.623715533085)*x[1]**o+(-0.457832375425)*x[1]+(-0.274712768232)*x[2]**o+(0.425333514519)*x[2]
        s[1,0,0]=(-0.901107526813)*x[0]**o+(0.0680326855019)*x[0]+(-0.24023047937)*x[1]**o+(0.33244773534)*x[1]+(0.649715117184)*x[2]**o+(-0.832757442484)*x[2]
        s[1,0,1]=(0.419698786285)*x[0]**o+(-0.929558931125)*x[0]+(-0.211068611062)*x[1]**o+(-0.308568292995)*x[1]+(-0.302460643894)*x[2]**o+(-0.792090369968)*x[2]
        s[1,1,0]=(-0.350071906565)*x[0]**o+(0.915476959141)*x[0]+(0.511206427652)*x[1]**o+(0.925740134945)*x[1]+(-0.163642337237)*x[2]**o+(0.466249765823)*x[2]
        s[1,1,1]=(-0.00679707570295)*x[0]**o+(-0.20076295244)*x[0]+(0.264205579203)*x[1]**o+(-0.411292187293)*x[1]+(-0.934377917835)*x[2]**o+(0.634215851155)*x[2]
        s[2,0,0]=(-0.324841664958)*x[0]**o+(-0.956529004217)*x[0]+(-0.918338753073)*x[1]**o+(0.626923365653)*x[1]+(-0.799444332565)*x[2]**o+(-0.0695464844552)*x[2]
        s[2,0,1]=(0.506623201982)*x[0]**o+(-0.149035848663)*x[0]+(-0.238527222625)*x[1]**o+(0.179414470428)*x[1]+(-0.830587331222)*x[2]**o+(-0.771223952622)*x[2]
        s[2,1,0]=(-0.600971886655)*x[0]**o+(0.578997478246)*x[0]+(0.307807603149)*x[1]**o+(-0.187632470569)*x[1]+(0.715704085441)*x[2]**o+(0.666008020335)*x[2]
        s[2,1,1]=(-0.927640726378)*x[0]**o+(-0.615258128661)*x[0]+(0.0458637924464)*x[1]**o+(0.316776923148)*x[1]+(-0.816443096595)*x[2]**o+(0.586050614098)*x[2]
        s[3,0,0]=(0.0116773788954)*x[0]**o+(-0.450016183617)*x[0]+(-0.722050331425)*x[1]**o+(-0.752741628658)*x[1]+(0.0807023425045)*x[2]**o+(-0.888353661711)*x[2]
        s[3,0,1]=(0.774410128748)*x[0]**o+(0.00319037654241)*x[0]+(-0.872810066503)*x[1]**o+(-0.667066572773)*x[1]+(0.87558896843)*x[2]**o+(-0.691829241879)*x[2]
        s[3,1,0]=(-0.81762346888)*x[0]**o+(0.117005969026)*x[0]+(-0.745835352632)*x[1]**o+(-0.162993479637)*x[1]+(0.259977485657)*x[2]**o+(-0.4154446656)*x[2]
        s[3,1,1]=(0.0202189083822)*x[0]**o+(0.276300397225)*x[0]+(-0.473252990801)*x[1]**o+(0.139373130343)*x[1]+(0.236401212883)*x[2]**o+(0.250980379862)*x[2]
        s[4,0,0]=(0.114221091454)*x[0]**o+(0.956199921345)*x[0]+(0.920335553649)*x[1]**o+(0.953720056733)*x[1]+(-0.869205945463)*x[2]**o+(-0.702596947421)*x[2]
        s[4,0,1]=(0.242788249724)*x[0]**o+(0.484618105306)*x[0]+(-0.0856836792276)*x[1]**o+(0.314175782784)*x[1]+(-0.462603325333)*x[2]**o+(-0.527952799625)*x[2]
        s[4,1,0]=(0.124168860106)*x[0]**o+(-0.846086223242)*x[0]+(-0.536186421788)*x[1]**o+(-0.103140610306)*x[1]+(-0.115670483732)*x[2]**o+(0.771594201067)*x[2]
        s[4,1,1]=(0.649717322503)*x[0]**o+(0.911557314842)*x[0]+(-0.498767014909)*x[1]**o+(0.647065328805)*x[1]+(0.722903839932)*x[2]**o+(0.514148167148)*x[2]
        s[5,0,0]=(-0.380297540184)*x[0]**o+(-0.485514278978)*x[0]+(0.785198036928)*x[1]**o+(-0.0846164275939)*x[1]+(-0.252406925767)*x[2]**o+(-0.223499486716)*x[2]
        s[5,0,1]=(-0.184173490174)*x[0]**o+(0.569523639256)*x[0]+(-0.294243155077)*x[1]**o+(-0.323071302538)*x[1]+(0.661442924711)*x[2]**o+(0.640705614167)*x[2]
        s[5,1,0]=(0.110804523334)*x[0]**o+(-0.933148550901)*x[0]+(-0.664439117013)*x[1]**o+(-0.33712595594)*x[1]+(0.0561563633834)*x[2]**o+(-0.728089270393)*x[2]
        s[5,1,1]=(-0.474826954962)*x[0]**o+(0.953336718849)*x[0]+(0.616224037612)*x[1]**o+(0.596223200958)*x[1]+(-0.790460006091)*x[2]**o+(0.580623251887)*x[2]
        ref[0,0,0,0]=o*(-0.252254362053)*x_ref[0]**(o-1)+(-0.73201258434)
        ref[0,0,0,1]=o*(0.129279219807)*x_ref[1]**(o-1)+(0.739511847421)
        ref[0,0,0,2]=o*(-0.335757163655)*x_ref[2]**(o-1)+(0.645975465365)
        ref[0,0,1,0]=o*(-0.735055416054)*x_ref[0]**(o-1)+(-0.392972979059)
        ref[0,0,1,1]=o*(-0.400312505739)*x_ref[1]**(o-1)+(-0.431371712491)
        ref[0,0,1,2]=o*(0.570395624547)*x_ref[2]**(o-1)+(0.138472351049)
        ref[0,1,0,0]=o*(-0.000809920184319)*x_ref[0]**(o-1)+(0.39610549699)
        ref[0,1,0,1]=o*(0.46183657032)*x_ref[1]**(o-1)+(-0.90835208084)
        ref[0,1,0,2]=o*(0.359160987844)*x_ref[2]**(o-1)+(0.63185061127)
        ref[0,1,1,0]=o*(0.627804065013)*x_ref[0]**(o-1)+(-0.616378680528)
        ref[0,1,1,1]=o*(-0.623715533085)*x_ref[1]**(o-1)+(-0.457832375425)
        ref[0,1,1,2]=o*(-0.274712768232)*x_ref[2]**(o-1)+(0.425333514519)
        ref[1,0,0,0]=o*(-0.901107526813)*x_ref[0]**(o-1)+(0.0680326855019)
        ref[1,0,0,1]=o*(-0.24023047937)*x_ref[1]**(o-1)+(0.33244773534)
        ref[1,0,0,2]=o*(0.649715117184)*x_ref[2]**(o-1)+(-0.832757442484)
        ref[1,0,1,0]=o*(0.419698786285)*x_ref[0]**(o-1)+(-0.929558931125)
        ref[1,0,1,1]=o*(-0.211068611062)*x_ref[1]**(o-1)+(-0.308568292995)
        ref[1,0,1,2]=o*(-0.302460643894)*x_ref[2]**(o-1)+(-0.792090369968)
        ref[1,1,0,0]=o*(-0.350071906565)*x_ref[0]**(o-1)+(0.915476959141)
        ref[1,1,0,1]=o*(0.511206427652)*x_ref[1]**(o-1)+(0.925740134945)
        ref[1,1,0,2]=o*(-0.163642337237)*x_ref[2]**(o-1)+(0.466249765823)
        ref[1,1,1,0]=o*(-0.00679707570295)*x_ref[0]**(o-1)+(-0.20076295244)
        ref[1,1,1,1]=o*(0.264205579203)*x_ref[1]**(o-1)+(-0.411292187293)
        ref[1,1,1,2]=o*(-0.934377917835)*x_ref[2]**(o-1)+(0.634215851155)
        ref[2,0,0,0]=o*(-0.324841664958)*x_ref[0]**(o-1)+(-0.956529004217)
        ref[2,0,0,1]=o*(-0.918338753073)*x_ref[1]**(o-1)+(0.626923365653)
        ref[2,0,0,2]=o*(-0.799444332565)*x_ref[2]**(o-1)+(-0.0695464844552)
        ref[2,0,1,0]=o*(0.506623201982)*x_ref[0]**(o-1)+(-0.149035848663)
        ref[2,0,1,1]=o*(-0.238527222625)*x_ref[1]**(o-1)+(0.179414470428)
        ref[2,0,1,2]=o*(-0.830587331222)*x_ref[2]**(o-1)+(-0.771223952622)
        ref[2,1,0,0]=o*(-0.600971886655)*x_ref[0]**(o-1)+(0.578997478246)
        ref[2,1,0,1]=o*(0.307807603149)*x_ref[1]**(o-1)+(-0.187632470569)
        ref[2,1,0,2]=o*(0.715704085441)*x_ref[2]**(o-1)+(0.666008020335)
        ref[2,1,1,0]=o*(-0.927640726378)*x_ref[0]**(o-1)+(-0.615258128661)
        ref[2,1,1,1]=o*(0.0458637924464)*x_ref[1]**(o-1)+(0.316776923148)
        ref[2,1,1,2]=o*(-0.816443096595)*x_ref[2]**(o-1)+(0.586050614098)
        ref[3,0,0,0]=o*(0.0116773788954)*x_ref[0]**(o-1)+(-0.450016183617)
        ref[3,0,0,1]=o*(-0.722050331425)*x_ref[1]**(o-1)+(-0.752741628658)
        ref[3,0,0,2]=o*(0.0807023425045)*x_ref[2]**(o-1)+(-0.888353661711)
        ref[3,0,1,0]=o*(0.774410128748)*x_ref[0]**(o-1)+(0.00319037654241)
        ref[3,0,1,1]=o*(-0.872810066503)*x_ref[1]**(o-1)+(-0.667066572773)
        ref[3,0,1,2]=o*(0.87558896843)*x_ref[2]**(o-1)+(-0.691829241879)
        ref[3,1,0,0]=o*(-0.81762346888)*x_ref[0]**(o-1)+(0.117005969026)
        ref[3,1,0,1]=o*(-0.745835352632)*x_ref[1]**(o-1)+(-0.162993479637)
        ref[3,1,0,2]=o*(0.259977485657)*x_ref[2]**(o-1)+(-0.4154446656)
        ref[3,1,1,0]=o*(0.0202189083822)*x_ref[0]**(o-1)+(0.276300397225)
        ref[3,1,1,1]=o*(-0.473252990801)*x_ref[1]**(o-1)+(0.139373130343)
        ref[3,1,1,2]=o*(0.236401212883)*x_ref[2]**(o-1)+(0.250980379862)
        ref[4,0,0,0]=o*(0.114221091454)*x_ref[0]**(o-1)+(0.956199921345)
        ref[4,0,0,1]=o*(0.920335553649)*x_ref[1]**(o-1)+(0.953720056733)
        ref[4,0,0,2]=o*(-0.869205945463)*x_ref[2]**(o-1)+(-0.702596947421)
        ref[4,0,1,0]=o*(0.242788249724)*x_ref[0]**(o-1)+(0.484618105306)
        ref[4,0,1,1]=o*(-0.0856836792276)*x_ref[1]**(o-1)+(0.314175782784)
        ref[4,0,1,2]=o*(-0.462603325333)*x_ref[2]**(o-1)+(-0.527952799625)
        ref[4,1,0,0]=o*(0.124168860106)*x_ref[0]**(o-1)+(-0.846086223242)
        ref[4,1,0,1]=o*(-0.536186421788)*x_ref[1]**(o-1)+(-0.103140610306)
        ref[4,1,0,2]=o*(-0.115670483732)*x_ref[2]**(o-1)+(0.771594201067)
        ref[4,1,1,0]=o*(0.649717322503)*x_ref[0]**(o-1)+(0.911557314842)
        ref[4,1,1,1]=o*(-0.498767014909)*x_ref[1]**(o-1)+(0.647065328805)
        ref[4,1,1,2]=o*(0.722903839932)*x_ref[2]**(o-1)+(0.514148167148)
        ref[5,0,0,0]=o*(-0.380297540184)*x_ref[0]**(o-1)+(-0.485514278978)
        ref[5,0,0,1]=o*(0.785198036928)*x_ref[1]**(o-1)+(-0.0846164275939)
        ref[5,0,0,2]=o*(-0.252406925767)*x_ref[2]**(o-1)+(-0.223499486716)
        ref[5,0,1,0]=o*(-0.184173490174)*x_ref[0]**(o-1)+(0.569523639256)
        ref[5,0,1,1]=o*(-0.294243155077)*x_ref[1]**(o-1)+(-0.323071302538)
        ref[5,0,1,2]=o*(0.661442924711)*x_ref[2]**(o-1)+(0.640705614167)
        ref[5,1,0,0]=o*(0.110804523334)*x_ref[0]**(o-1)+(-0.933148550901)
        ref[5,1,0,1]=o*(-0.664439117013)*x_ref[1]**(o-1)+(-0.33712595594)
        ref[5,1,0,2]=o*(0.0561563633834)*x_ref[2]**(o-1)+(-0.728089270393)
        ref[5,1,1,0]=o*(-0.474826954962)*x_ref[0]**(o-1)+(0.953336718849)
        ref[5,1,1,1]=o*(0.616224037612)*x_ref[1]**(o-1)+(0.596223200958)
        ref[5,1,1,2]=o*(-0.790460006091)*x_ref[2]**(o-1)+(0.580623251887)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_Solution_rank0(self):
      """
      tests gradient for rank 0 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.661534580714)*x[0]**o+(0.108482570281)*x[0]+(0.930475222346)*x[1]**o+(-0.560831828628)*x[1]
        ref[0]=o*(-0.661534580714)*x_ref[0]**(o-1)+(0.108482570281)
        ref[1]=o*(0.930475222346)*x_ref[1]**(o-1)+(-0.560831828628)
      else:
        s=(0.661787087481)*x[0]**o+(0.789038424245)*x[0]+(-0.172842194551)*x[1]**o+(-0.313156164165)*x[1]+(-0.521097764776)*x[2]**o+(0.879698074812)*x[2]
        ref[0]=o*(0.661787087481)*x_ref[0]**(o-1)+(0.789038424245)
        ref[1]=o*(-0.172842194551)*x_ref[1]**(o-1)+(-0.313156164165)
        ref[2]=o*(-0.521097764776)*x_ref[2]**(o-1)+(0.879698074812)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_Solution_rank1(self):
      """
      tests gradient for rank 1 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(-0.190630799823)*x[0]**o+(0.235686163485)*x[0]+(0.568780014934)*x[1]**o+(0.869208291526)*x[1]
        s[1]=(-0.637753770052)*x[0]**o+(-0.184551176704)*x[0]+(0.678386051731)*x[1]**o+(0.103311313677)*x[1]
        ref[0,0]=o*(-0.190630799823)*x_ref[0]**(o-1)+(0.235686163485)
        ref[0,1]=o*(0.568780014934)*x_ref[1]**(o-1)+(0.869208291526)
        ref[1,0]=o*(-0.637753770052)*x_ref[0]**(o-1)+(-0.184551176704)
        ref[1,1]=o*(0.678386051731)*x_ref[1]**(o-1)+(0.103311313677)
      else:
        s[0]=(0.65617389354)*x[0]**o+(-0.41233134159)*x[0]+(-0.471086791763)*x[1]**o+(0.399070328851)*x[1]+(-0.30038453779)*x[2]**o+(0.637544697206)*x[2]
        s[1]=(0.0974925575768)*x[0]**o+(0.968900636228)*x[0]+(-0.386954772478)*x[1]**o+(0.234655265071)*x[1]+(0.755549468261)*x[2]**o+(-0.119269665245)*x[2]
        ref[0,0]=o*(0.65617389354)*x_ref[0]**(o-1)+(-0.41233134159)
        ref[0,1]=o*(-0.471086791763)*x_ref[1]**(o-1)+(0.399070328851)
        ref[0,2]=o*(-0.30038453779)*x_ref[2]**(o-1)+(0.637544697206)
        ref[1,0]=o*(0.0974925575768)*x_ref[0]**(o-1)+(0.968900636228)
        ref[1,1]=o*(-0.386954772478)*x_ref[1]**(o-1)+(0.234655265071)
        ref[1,2]=o*(0.755549468261)*x_ref[2]**(o-1)+(-0.119269665245)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_Solution_rank2(self):
      """
      tests gradient for rank 2 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.300429908481)*x[0]**o+(-0.550176774021)*x[0]+(0.313544523598)*x[1]**o+(0.191557196291)*x[1]
        s[0,1]=(0.00116867375625)*x[0]**o+(0.126481551395)*x[0]+(-0.525739733277)*x[1]**o+(0.628228355029)*x[1]
        s[0,2]=(-0.790055765486)*x[0]**o+(-0.118003096828)*x[0]+(-0.620290760565)*x[1]**o+(-0.460021810788)*x[1]
        s[0,3]=(0.800717644254)*x[0]**o+(-0.685505835346)*x[0]+(-0.382046994517)*x[1]**o+(-0.765159598284)*x[1]
        s[0,4]=(-0.295987131319)*x[0]**o+(-0.0324335264934)*x[0]+(-0.583454515945)*x[1]**o+(-0.137940341278)*x[1]
        s[1,0]=(-0.186271508459)*x[0]**o+(-0.0457260753159)*x[0]+(-0.83853233149)*x[1]**o+(0.730348417106)*x[1]
        s[1,1]=(-0.678758709357)*x[0]**o+(-0.352619765194)*x[0]+(0.850836393219)*x[1]**o+(-0.749995270113)*x[1]
        s[1,2]=(0.598004774999)*x[0]**o+(-0.764313433057)*x[0]+(-0.892691377674)*x[1]**o+(0.436164793316)*x[1]
        s[1,3]=(-0.114398805139)*x[0]**o+(0.170882318781)*x[0]+(-0.159446706481)*x[1]**o+(0.168268554115)*x[1]
        s[1,4]=(0.245440053023)*x[0]**o+(-0.391101810385)*x[0]+(-0.0621967792961)*x[1]**o+(-0.25914597073)*x[1]
        s[2,0]=(0.758930908002)*x[0]**o+(-0.137399611576)*x[0]+(-0.524730401698)*x[1]**o+(-0.68595790282)*x[1]
        s[2,1]=(-0.956320437139)*x[0]**o+(-0.741920593412)*x[0]+(0.470411496813)*x[1]**o+(-0.912492685238)*x[1]
        s[2,2]=(0.386762534234)*x[0]**o+(0.297029046975)*x[0]+(-0.999949621409)*x[1]**o+(-0.273952756751)*x[1]
        s[2,3]=(-0.137635508969)*x[0]**o+(0.0502825343025)*x[0]+(-0.856424864287)*x[1]**o+(0.501432472431)*x[1]
        s[2,4]=(0.927974425929)*x[0]**o+(-0.602904477506)*x[0]+(0.766583044002)*x[1]**o+(0.070024577589)*x[1]
        s[3,0]=(0.336864984737)*x[0]**o+(0.0974502498452)*x[0]+(-0.820387387723)*x[1]**o+(0.022457213421)*x[1]
        s[3,1]=(0.879517107747)*x[0]**o+(0.267883865556)*x[0]+(-0.795732666749)*x[1]**o+(0.81701302209)*x[1]
        s[3,2]=(-0.971853752113)*x[0]**o+(0.210482589767)*x[0]+(0.422058370831)*x[1]**o+(-0.0945771197357)*x[1]
        s[3,3]=(0.0130393893213)*x[0]**o+(0.331617264696)*x[0]+(-0.0460157855829)*x[1]**o+(0.513357385084)*x[1]
        s[3,4]=(-0.466291714481)*x[0]**o+(0.877178907345)*x[0]+(0.356001871079)*x[1]**o+(-0.755177274311)*x[1]
        ref[0,0,0]=o*(-0.300429908481)*x_ref[0]**(o-1)+(-0.550176774021)
        ref[0,0,1]=o*(0.313544523598)*x_ref[1]**(o-1)+(0.191557196291)
        ref[0,1,0]=o*(0.00116867375625)*x_ref[0]**(o-1)+(0.126481551395)
        ref[0,1,1]=o*(-0.525739733277)*x_ref[1]**(o-1)+(0.628228355029)
        ref[0,2,0]=o*(-0.790055765486)*x_ref[0]**(o-1)+(-0.118003096828)
        ref[0,2,1]=o*(-0.620290760565)*x_ref[1]**(o-1)+(-0.460021810788)
        ref[0,3,0]=o*(0.800717644254)*x_ref[0]**(o-1)+(-0.685505835346)
        ref[0,3,1]=o*(-0.382046994517)*x_ref[1]**(o-1)+(-0.765159598284)
        ref[0,4,0]=o*(-0.295987131319)*x_ref[0]**(o-1)+(-0.0324335264934)
        ref[0,4,1]=o*(-0.583454515945)*x_ref[1]**(o-1)+(-0.137940341278)
        ref[1,0,0]=o*(-0.186271508459)*x_ref[0]**(o-1)+(-0.0457260753159)
        ref[1,0,1]=o*(-0.83853233149)*x_ref[1]**(o-1)+(0.730348417106)
        ref[1,1,0]=o*(-0.678758709357)*x_ref[0]**(o-1)+(-0.352619765194)
        ref[1,1,1]=o*(0.850836393219)*x_ref[1]**(o-1)+(-0.749995270113)
        ref[1,2,0]=o*(0.598004774999)*x_ref[0]**(o-1)+(-0.764313433057)
        ref[1,2,1]=o*(-0.892691377674)*x_ref[1]**(o-1)+(0.436164793316)
        ref[1,3,0]=o*(-0.114398805139)*x_ref[0]**(o-1)+(0.170882318781)
        ref[1,3,1]=o*(-0.159446706481)*x_ref[1]**(o-1)+(0.168268554115)
        ref[1,4,0]=o*(0.245440053023)*x_ref[0]**(o-1)+(-0.391101810385)
        ref[1,4,1]=o*(-0.0621967792961)*x_ref[1]**(o-1)+(-0.25914597073)
        ref[2,0,0]=o*(0.758930908002)*x_ref[0]**(o-1)+(-0.137399611576)
        ref[2,0,1]=o*(-0.524730401698)*x_ref[1]**(o-1)+(-0.68595790282)
        ref[2,1,0]=o*(-0.956320437139)*x_ref[0]**(o-1)+(-0.741920593412)
        ref[2,1,1]=o*(0.470411496813)*x_ref[1]**(o-1)+(-0.912492685238)
        ref[2,2,0]=o*(0.386762534234)*x_ref[0]**(o-1)+(0.297029046975)
        ref[2,2,1]=o*(-0.999949621409)*x_ref[1]**(o-1)+(-0.273952756751)
        ref[2,3,0]=o*(-0.137635508969)*x_ref[0]**(o-1)+(0.0502825343025)
        ref[2,3,1]=o*(-0.856424864287)*x_ref[1]**(o-1)+(0.501432472431)
        ref[2,4,0]=o*(0.927974425929)*x_ref[0]**(o-1)+(-0.602904477506)
        ref[2,4,1]=o*(0.766583044002)*x_ref[1]**(o-1)+(0.070024577589)
        ref[3,0,0]=o*(0.336864984737)*x_ref[0]**(o-1)+(0.0974502498452)
        ref[3,0,1]=o*(-0.820387387723)*x_ref[1]**(o-1)+(0.022457213421)
        ref[3,1,0]=o*(0.879517107747)*x_ref[0]**(o-1)+(0.267883865556)
        ref[3,1,1]=o*(-0.795732666749)*x_ref[1]**(o-1)+(0.81701302209)
        ref[3,2,0]=o*(-0.971853752113)*x_ref[0]**(o-1)+(0.210482589767)
        ref[3,2,1]=o*(0.422058370831)*x_ref[1]**(o-1)+(-0.0945771197357)
        ref[3,3,0]=o*(0.0130393893213)*x_ref[0]**(o-1)+(0.331617264696)
        ref[3,3,1]=o*(-0.0460157855829)*x_ref[1]**(o-1)+(0.513357385084)
        ref[3,4,0]=o*(-0.466291714481)*x_ref[0]**(o-1)+(0.877178907345)
        ref[3,4,1]=o*(0.356001871079)*x_ref[1]**(o-1)+(-0.755177274311)
      else:
        s[0,0]=(-0.582848481634)*x[0]**o+(-0.799173040171)*x[0]+(-0.402728303622)*x[1]**o+(-0.294649773058)*x[1]+(-0.248183917161)*x[2]**o+(-0.182025430599)*x[2]
        s[0,1]=(0.421723448079)*x[0]**o+(-0.881083512259)*x[0]+(-0.411609580693)*x[1]**o+(-0.60866119666)*x[1]+(0.447246876403)*x[2]**o+(-0.566693550375)*x[2]
        s[0,2]=(0.676859700236)*x[0]**o+(-0.0721141912983)*x[0]+(-0.930202067587)*x[1]**o+(-0.779660291535)*x[1]+(-0.846576347271)*x[2]**o+(0.182815431369)*x[2]
        s[0,3]=(0.640898585384)*x[0]**o+(0.133026195724)*x[0]+(-0.264924898824)*x[1]**o+(0.984926461364)*x[1]+(-0.366473159006)*x[2]**o+(-0.893150501229)*x[2]
        s[0,4]=(-0.0617160479105)*x[0]**o+(-0.781163890754)*x[0]+(0.482406386837)*x[1]**o+(0.242984596118)*x[1]+(0.387455073331)*x[2]**o+(0.713852275793)*x[2]
        s[1,0]=(-0.261960056141)*x[0]**o+(0.289706307634)*x[0]+(0.33502315284)*x[1]**o+(0.373867353003)*x[1]+(0.988176700778)*x[2]**o+(0.821674109963)*x[2]
        s[1,1]=(-0.564538779546)*x[0]**o+(0.512720259636)*x[0]+(0.519794117449)*x[1]**o+(0.140310457627)*x[1]+(-0.987992481444)*x[2]**o+(-0.658619677691)*x[2]
        s[1,2]=(-0.94468459827)*x[0]**o+(-0.59503451145)*x[0]+(0.638928053728)*x[1]**o+(-0.426823410254)*x[1]+(0.698221019291)*x[2]**o+(0.584939961146)*x[2]
        s[1,3]=(-0.736608632072)*x[0]**o+(0.121237394007)*x[0]+(0.965644483539)*x[1]**o+(-0.69597048224)*x[1]+(-0.131814696207)*x[2]**o+(0.218156302554)*x[2]
        s[1,4]=(-0.68564146602)*x[0]**o+(-0.613073625086)*x[0]+(-0.480606344797)*x[1]**o+(0.305215204518)*x[1]+(-0.932537843561)*x[2]**o+(0.753998778964)*x[2]
        s[2,0]=(0.149366204787)*x[0]**o+(0.518985565336)*x[0]+(-0.108643244187)*x[1]**o+(-0.133769184691)*x[1]+(-0.91950647261)*x[2]**o+(-0.165713076313)*x[2]
        s[2,1]=(0.377574358299)*x[0]**o+(-0.881166371065)*x[0]+(0.859757091349)*x[1]**o+(-0.571419705867)*x[1]+(0.688875894062)*x[2]**o+(-0.475017852628)*x[2]
        s[2,2]=(-0.293556901343)*x[0]**o+(-0.21565327941)*x[0]+(-0.661622857503)*x[1]**o+(-0.822811400487)*x[1]+(0.773358448616)*x[2]**o+(0.687967954749)*x[2]
        s[2,3]=(-0.0678537895443)*x[0]**o+(-0.620817206574)*x[0]+(0.195444962807)*x[1]**o+(-0.802482456128)*x[1]+(0.839892027886)*x[2]**o+(0.53033825313)*x[2]
        s[2,4]=(-0.174133277232)*x[0]**o+(-0.11357878673)*x[0]+(0.896953765213)*x[1]**o+(-0.111411131996)*x[1]+(-0.911590351769)*x[2]**o+(-0.867631474956)*x[2]
        s[3,0]=(0.619809270826)*x[0]**o+(0.234416234461)*x[0]+(-0.265703869206)*x[1]**o+(-0.179792194117)*x[1]+(0.650878614932)*x[2]**o+(0.832933790635)*x[2]
        s[3,1]=(-0.464423972484)*x[0]**o+(-0.0792284263184)*x[0]+(0.591682473906)*x[1]**o+(-0.69612265297)*x[1]+(-0.847084735807)*x[2]**o+(0.406411159143)*x[2]
        s[3,2]=(-0.206502883044)*x[0]**o+(0.266143672088)*x[0]+(0.742487243777)*x[1]**o+(0.652888436702)*x[1]+(-0.638601136445)*x[2]**o+(-0.411202594715)*x[2]
        s[3,3]=(0.749860017366)*x[0]**o+(0.0712552207004)*x[0]+(0.932270746555)*x[1]**o+(-0.743243258158)*x[1]+(0.77742513611)*x[2]**o+(-0.0951940129286)*x[2]
        s[3,4]=(-0.736329018198)*x[0]**o+(-0.346737131331)*x[0]+(0.15155322637)*x[1]**o+(0.209410136584)*x[1]+(-0.283669919947)*x[2]**o+(-0.736677347286)*x[2]
        ref[0,0,0]=o*(-0.582848481634)*x_ref[0]**(o-1)+(-0.799173040171)
        ref[0,0,1]=o*(-0.402728303622)*x_ref[1]**(o-1)+(-0.294649773058)
        ref[0,0,2]=o*(-0.248183917161)*x_ref[2]**(o-1)+(-0.182025430599)
        ref[0,1,0]=o*(0.421723448079)*x_ref[0]**(o-1)+(-0.881083512259)
        ref[0,1,1]=o*(-0.411609580693)*x_ref[1]**(o-1)+(-0.60866119666)
        ref[0,1,2]=o*(0.447246876403)*x_ref[2]**(o-1)+(-0.566693550375)
        ref[0,2,0]=o*(0.676859700236)*x_ref[0]**(o-1)+(-0.0721141912983)
        ref[0,2,1]=o*(-0.930202067587)*x_ref[1]**(o-1)+(-0.779660291535)
        ref[0,2,2]=o*(-0.846576347271)*x_ref[2]**(o-1)+(0.182815431369)
        ref[0,3,0]=o*(0.640898585384)*x_ref[0]**(o-1)+(0.133026195724)
        ref[0,3,1]=o*(-0.264924898824)*x_ref[1]**(o-1)+(0.984926461364)
        ref[0,3,2]=o*(-0.366473159006)*x_ref[2]**(o-1)+(-0.893150501229)
        ref[0,4,0]=o*(-0.0617160479105)*x_ref[0]**(o-1)+(-0.781163890754)
        ref[0,4,1]=o*(0.482406386837)*x_ref[1]**(o-1)+(0.242984596118)
        ref[0,4,2]=o*(0.387455073331)*x_ref[2]**(o-1)+(0.713852275793)
        ref[1,0,0]=o*(-0.261960056141)*x_ref[0]**(o-1)+(0.289706307634)
        ref[1,0,1]=o*(0.33502315284)*x_ref[1]**(o-1)+(0.373867353003)
        ref[1,0,2]=o*(0.988176700778)*x_ref[2]**(o-1)+(0.821674109963)
        ref[1,1,0]=o*(-0.564538779546)*x_ref[0]**(o-1)+(0.512720259636)
        ref[1,1,1]=o*(0.519794117449)*x_ref[1]**(o-1)+(0.140310457627)
        ref[1,1,2]=o*(-0.987992481444)*x_ref[2]**(o-1)+(-0.658619677691)
        ref[1,2,0]=o*(-0.94468459827)*x_ref[0]**(o-1)+(-0.59503451145)
        ref[1,2,1]=o*(0.638928053728)*x_ref[1]**(o-1)+(-0.426823410254)
        ref[1,2,2]=o*(0.698221019291)*x_ref[2]**(o-1)+(0.584939961146)
        ref[1,3,0]=o*(-0.736608632072)*x_ref[0]**(o-1)+(0.121237394007)
        ref[1,3,1]=o*(0.965644483539)*x_ref[1]**(o-1)+(-0.69597048224)
        ref[1,3,2]=o*(-0.131814696207)*x_ref[2]**(o-1)+(0.218156302554)
        ref[1,4,0]=o*(-0.68564146602)*x_ref[0]**(o-1)+(-0.613073625086)
        ref[1,4,1]=o*(-0.480606344797)*x_ref[1]**(o-1)+(0.305215204518)
        ref[1,4,2]=o*(-0.932537843561)*x_ref[2]**(o-1)+(0.753998778964)
        ref[2,0,0]=o*(0.149366204787)*x_ref[0]**(o-1)+(0.518985565336)
        ref[2,0,1]=o*(-0.108643244187)*x_ref[1]**(o-1)+(-0.133769184691)
        ref[2,0,2]=o*(-0.91950647261)*x_ref[2]**(o-1)+(-0.165713076313)
        ref[2,1,0]=o*(0.377574358299)*x_ref[0]**(o-1)+(-0.881166371065)
        ref[2,1,1]=o*(0.859757091349)*x_ref[1]**(o-1)+(-0.571419705867)
        ref[2,1,2]=o*(0.688875894062)*x_ref[2]**(o-1)+(-0.475017852628)
        ref[2,2,0]=o*(-0.293556901343)*x_ref[0]**(o-1)+(-0.21565327941)
        ref[2,2,1]=o*(-0.661622857503)*x_ref[1]**(o-1)+(-0.822811400487)
        ref[2,2,2]=o*(0.773358448616)*x_ref[2]**(o-1)+(0.687967954749)
        ref[2,3,0]=o*(-0.0678537895443)*x_ref[0]**(o-1)+(-0.620817206574)
        ref[2,3,1]=o*(0.195444962807)*x_ref[1]**(o-1)+(-0.802482456128)
        ref[2,3,2]=o*(0.839892027886)*x_ref[2]**(o-1)+(0.53033825313)
        ref[2,4,0]=o*(-0.174133277232)*x_ref[0]**(o-1)+(-0.11357878673)
        ref[2,4,1]=o*(0.896953765213)*x_ref[1]**(o-1)+(-0.111411131996)
        ref[2,4,2]=o*(-0.911590351769)*x_ref[2]**(o-1)+(-0.867631474956)
        ref[3,0,0]=o*(0.619809270826)*x_ref[0]**(o-1)+(0.234416234461)
        ref[3,0,1]=o*(-0.265703869206)*x_ref[1]**(o-1)+(-0.179792194117)
        ref[3,0,2]=o*(0.650878614932)*x_ref[2]**(o-1)+(0.832933790635)
        ref[3,1,0]=o*(-0.464423972484)*x_ref[0]**(o-1)+(-0.0792284263184)
        ref[3,1,1]=o*(0.591682473906)*x_ref[1]**(o-1)+(-0.69612265297)
        ref[3,1,2]=o*(-0.847084735807)*x_ref[2]**(o-1)+(0.406411159143)
        ref[3,2,0]=o*(-0.206502883044)*x_ref[0]**(o-1)+(0.266143672088)
        ref[3,2,1]=o*(0.742487243777)*x_ref[1]**(o-1)+(0.652888436702)
        ref[3,2,2]=o*(-0.638601136445)*x_ref[2]**(o-1)+(-0.411202594715)
        ref[3,3,0]=o*(0.749860017366)*x_ref[0]**(o-1)+(0.0712552207004)
        ref[3,3,1]=o*(0.932270746555)*x_ref[1]**(o-1)+(-0.743243258158)
        ref[3,3,2]=o*(0.77742513611)*x_ref[2]**(o-1)+(-0.0951940129286)
        ref[3,4,0]=o*(-0.736329018198)*x_ref[0]**(o-1)+(-0.346737131331)
        ref[3,4,1]=o*(0.15155322637)*x_ref[1]**(o-1)+(0.209410136584)
        ref[3,4,2]=o*(-0.283669919947)*x_ref[2]**(o-1)+(-0.736677347286)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_Solution_rank3(self):
      """
      tests gradient for rank 3 Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(0.285489768558)*x[0]**o+(-0.814169283089)*x[0]+(-0.146214432938)*x[1]**o+(0.971176476923)*x[1]
        s[0,0,1]=(0.796197012783)*x[0]**o+(0.138854173842)*x[0]+(0.911141134854)*x[1]**o+(0.24230493445)*x[1]
        s[0,1,0]=(0.335306616869)*x[0]**o+(-0.346373461144)*x[0]+(0.485198289812)*x[1]**o+(-0.557460170056)*x[1]
        s[0,1,1]=(-0.202832958777)*x[0]**o+(0.202612852038)*x[0]+(-0.382959990066)*x[1]**o+(0.982713356847)*x[1]
        s[1,0,0]=(-0.732852201528)*x[0]**o+(-0.52380849964)*x[0]+(-0.423867496739)*x[1]**o+(-0.476139964916)*x[1]
        s[1,0,1]=(0.730537196604)*x[0]**o+(0.782053455575)*x[0]+(-0.496379081802)*x[1]**o+(-0.610796298113)*x[1]
        s[1,1,0]=(-0.518151715463)*x[0]**o+(0.466253389588)*x[0]+(0.792730016645)*x[1]**o+(-0.0928866491573)*x[1]
        s[1,1,1]=(0.721737904289)*x[0]**o+(0.662197543309)*x[0]+(0.402567060485)*x[1]**o+(-0.0537539240196)*x[1]
        s[2,0,0]=(0.264010742903)*x[0]**o+(-0.209394534976)*x[0]+(0.901338534988)*x[1]**o+(0.751548611498)*x[1]
        s[2,0,1]=(0.255790620214)*x[0]**o+(0.979495958037)*x[0]+(-0.00862935506386)*x[1]**o+(0.472608624671)*x[1]
        s[2,1,0]=(0.348157499697)*x[0]**o+(-0.489084366673)*x[0]+(-0.797020984331)*x[1]**o+(0.850642513227)*x[1]
        s[2,1,1]=(0.619156982334)*x[0]**o+(-0.871389250139)*x[0]+(0.838613332974)*x[1]**o+(0.865428066011)*x[1]
        s[3,0,0]=(-0.498572545085)*x[0]**o+(0.26752755104)*x[0]+(-0.742506202612)*x[1]**o+(-0.867811055537)*x[1]
        s[3,0,1]=(-0.469062081737)*x[0]**o+(-0.896231281785)*x[0]+(-0.0537944852375)*x[1]**o+(-0.566775495197)*x[1]
        s[3,1,0]=(0.960466927125)*x[0]**o+(-0.790015287315)*x[0]+(-0.249364909212)*x[1]**o+(0.55000656331)*x[1]
        s[3,1,1]=(0.107190484709)*x[0]**o+(0.206656185528)*x[0]+(0.445355269686)*x[1]**o+(0.177791381802)*x[1]
        s[4,0,0]=(0.926353234834)*x[0]**o+(0.0187238838755)*x[0]+(-0.0335895202391)*x[1]**o+(0.396027343683)*x[1]
        s[4,0,1]=(-0.122883815734)*x[0]**o+(0.114014252654)*x[0]+(0.390958937743)*x[1]**o+(0.743802098345)*x[1]
        s[4,1,0]=(0.0434615831789)*x[0]**o+(-0.0061546333485)*x[0]+(-0.941682112962)*x[1]**o+(0.546094645223)*x[1]
        s[4,1,1]=(-0.876438895739)*x[0]**o+(-0.307842960731)*x[0]+(-0.958906847802)*x[1]**o+(-0.698921900635)*x[1]
        s[5,0,0]=(-0.0294848627519)*x[0]**o+(-0.108277485049)*x[0]+(0.653184058519)*x[1]**o+(0.268533892275)*x[1]
        s[5,0,1]=(0.0643758658725)*x[0]**o+(0.403982947726)*x[0]+(-0.966297836476)*x[1]**o+(-0.934899934913)*x[1]
        s[5,1,0]=(-0.354834560075)*x[0]**o+(-0.641982491187)*x[0]+(0.231470813627)*x[1]**o+(0.597773058655)*x[1]
        s[5,1,1]=(0.457345532548)*x[0]**o+(0.930909964958)*x[0]+(-0.190821688076)*x[1]**o+(0.763332121356)*x[1]
        ref[0,0,0,0]=o*(0.285489768558)*x_ref[0]**(o-1)+(-0.814169283089)
        ref[0,0,0,1]=o*(-0.146214432938)*x_ref[1]**(o-1)+(0.971176476923)
        ref[0,0,1,0]=o*(0.796197012783)*x_ref[0]**(o-1)+(0.138854173842)
        ref[0,0,1,1]=o*(0.911141134854)*x_ref[1]**(o-1)+(0.24230493445)
        ref[0,1,0,0]=o*(0.335306616869)*x_ref[0]**(o-1)+(-0.346373461144)
        ref[0,1,0,1]=o*(0.485198289812)*x_ref[1]**(o-1)+(-0.557460170056)
        ref[0,1,1,0]=o*(-0.202832958777)*x_ref[0]**(o-1)+(0.202612852038)
        ref[0,1,1,1]=o*(-0.382959990066)*x_ref[1]**(o-1)+(0.982713356847)
        ref[1,0,0,0]=o*(-0.732852201528)*x_ref[0]**(o-1)+(-0.52380849964)
        ref[1,0,0,1]=o*(-0.423867496739)*x_ref[1]**(o-1)+(-0.476139964916)
        ref[1,0,1,0]=o*(0.730537196604)*x_ref[0]**(o-1)+(0.782053455575)
        ref[1,0,1,1]=o*(-0.496379081802)*x_ref[1]**(o-1)+(-0.610796298113)
        ref[1,1,0,0]=o*(-0.518151715463)*x_ref[0]**(o-1)+(0.466253389588)
        ref[1,1,0,1]=o*(0.792730016645)*x_ref[1]**(o-1)+(-0.0928866491573)
        ref[1,1,1,0]=o*(0.721737904289)*x_ref[0]**(o-1)+(0.662197543309)
        ref[1,1,1,1]=o*(0.402567060485)*x_ref[1]**(o-1)+(-0.0537539240196)
        ref[2,0,0,0]=o*(0.264010742903)*x_ref[0]**(o-1)+(-0.209394534976)
        ref[2,0,0,1]=o*(0.901338534988)*x_ref[1]**(o-1)+(0.751548611498)
        ref[2,0,1,0]=o*(0.255790620214)*x_ref[0]**(o-1)+(0.979495958037)
        ref[2,0,1,1]=o*(-0.00862935506386)*x_ref[1]**(o-1)+(0.472608624671)
        ref[2,1,0,0]=o*(0.348157499697)*x_ref[0]**(o-1)+(-0.489084366673)
        ref[2,1,0,1]=o*(-0.797020984331)*x_ref[1]**(o-1)+(0.850642513227)
        ref[2,1,1,0]=o*(0.619156982334)*x_ref[0]**(o-1)+(-0.871389250139)
        ref[2,1,1,1]=o*(0.838613332974)*x_ref[1]**(o-1)+(0.865428066011)
        ref[3,0,0,0]=o*(-0.498572545085)*x_ref[0]**(o-1)+(0.26752755104)
        ref[3,0,0,1]=o*(-0.742506202612)*x_ref[1]**(o-1)+(-0.867811055537)
        ref[3,0,1,0]=o*(-0.469062081737)*x_ref[0]**(o-1)+(-0.896231281785)
        ref[3,0,1,1]=o*(-0.0537944852375)*x_ref[1]**(o-1)+(-0.566775495197)
        ref[3,1,0,0]=o*(0.960466927125)*x_ref[0]**(o-1)+(-0.790015287315)
        ref[3,1,0,1]=o*(-0.249364909212)*x_ref[1]**(o-1)+(0.55000656331)
        ref[3,1,1,0]=o*(0.107190484709)*x_ref[0]**(o-1)+(0.206656185528)
        ref[3,1,1,1]=o*(0.445355269686)*x_ref[1]**(o-1)+(0.177791381802)
        ref[4,0,0,0]=o*(0.926353234834)*x_ref[0]**(o-1)+(0.0187238838755)
        ref[4,0,0,1]=o*(-0.0335895202391)*x_ref[1]**(o-1)+(0.396027343683)
        ref[4,0,1,0]=o*(-0.122883815734)*x_ref[0]**(o-1)+(0.114014252654)
        ref[4,0,1,1]=o*(0.390958937743)*x_ref[1]**(o-1)+(0.743802098345)
        ref[4,1,0,0]=o*(0.0434615831789)*x_ref[0]**(o-1)+(-0.0061546333485)
        ref[4,1,0,1]=o*(-0.941682112962)*x_ref[1]**(o-1)+(0.546094645223)
        ref[4,1,1,0]=o*(-0.876438895739)*x_ref[0]**(o-1)+(-0.307842960731)
        ref[4,1,1,1]=o*(-0.958906847802)*x_ref[1]**(o-1)+(-0.698921900635)
        ref[5,0,0,0]=o*(-0.0294848627519)*x_ref[0]**(o-1)+(-0.108277485049)
        ref[5,0,0,1]=o*(0.653184058519)*x_ref[1]**(o-1)+(0.268533892275)
        ref[5,0,1,0]=o*(0.0643758658725)*x_ref[0]**(o-1)+(0.403982947726)
        ref[5,0,1,1]=o*(-0.966297836476)*x_ref[1]**(o-1)+(-0.934899934913)
        ref[5,1,0,0]=o*(-0.354834560075)*x_ref[0]**(o-1)+(-0.641982491187)
        ref[5,1,0,1]=o*(0.231470813627)*x_ref[1]**(o-1)+(0.597773058655)
        ref[5,1,1,0]=o*(0.457345532548)*x_ref[0]**(o-1)+(0.930909964958)
        ref[5,1,1,1]=o*(-0.190821688076)*x_ref[1]**(o-1)+(0.763332121356)
      else:
        s[0,0,0]=(0.0799106193628)*x[0]**o+(-0.385218529142)*x[0]+(0.423551372182)*x[1]**o+(-0.357661802001)*x[1]+(0.523815428088)*x[2]**o+(-0.867664901871)*x[2]
        s[0,0,1]=(-0.223298613593)*x[0]**o+(-0.563675059615)*x[0]+(0.854553636152)*x[1]**o+(0.00619936897032)*x[1]+(-0.168119043675)*x[2]**o+(-0.959967076262)*x[2]
        s[0,1,0]=(-0.538044222292)*x[0]**o+(0.519685043014)*x[0]+(-0.351144688067)*x[1]**o+(0.887365493845)*x[1]+(0.801566727487)*x[2]**o+(0.571280510487)*x[2]
        s[0,1,1]=(0.784344723309)*x[0]**o+(-0.0907512562255)*x[0]+(-0.24324080888)*x[1]**o+(-0.199412485507)*x[1]+(0.642344630349)*x[2]**o+(-0.735616137201)*x[2]
        s[1,0,0]=(-0.880206903193)*x[0]**o+(-0.749123123033)*x[0]+(0.104561575813)*x[1]**o+(-0.647441987473)*x[1]+(0.182201321281)*x[2]**o+(0.0102973222322)*x[2]
        s[1,0,1]=(0.915893198392)*x[0]**o+(-0.612952390562)*x[0]+(0.803094622661)*x[1]**o+(-0.140410725667)*x[1]+(-0.247855770678)*x[2]**o+(0.399254178737)*x[2]
        s[1,1,0]=(0.818321238866)*x[0]**o+(0.214893239645)*x[0]+(0.909338463913)*x[1]**o+(0.855186366612)*x[1]+(-0.835375145102)*x[2]**o+(-0.807936215386)*x[2]
        s[1,1,1]=(0.257588723885)*x[0]**o+(-0.952059092336)*x[0]+(-0.28964814303)*x[1]**o+(0.0129418032921)*x[1]+(0.854165719561)*x[2]**o+(0.16196867256)*x[2]
        s[2,0,0]=(-0.291805757446)*x[0]**o+(0.609410741859)*x[0]+(0.541295607728)*x[1]**o+(-0.220049109439)*x[1]+(0.773352330852)*x[2]**o+(-0.0999084978714)*x[2]
        s[2,0,1]=(-0.094477913729)*x[0]**o+(-0.764264730185)*x[0]+(-0.671804542061)*x[1]**o+(0.368646822044)*x[1]+(0.448237271807)*x[2]**o+(0.565106260459)*x[2]
        s[2,1,0]=(-0.240257182637)*x[0]**o+(0.859112706839)*x[0]+(-0.575504060232)*x[1]**o+(-0.611929236978)*x[1]+(0.675448692587)*x[2]**o+(0.31890080218)*x[2]
        s[2,1,1]=(-0.39222652102)*x[0]**o+(0.677458518423)*x[0]+(-0.412612361541)*x[1]**o+(-0.374686862277)*x[1]+(-0.737213784962)*x[2]**o+(-0.486920606866)*x[2]
        s[3,0,0]=(0.635190942363)*x[0]**o+(-0.330731024263)*x[0]+(-0.690112375174)*x[1]**o+(-0.446491601534)*x[1]+(-0.748307399262)*x[2]**o+(-0.578747596175)*x[2]
        s[3,0,1]=(-0.419848395854)*x[0]**o+(-0.31014199789)*x[0]+(0.818235061383)*x[1]**o+(0.0547944058433)*x[1]+(0.855573639674)*x[2]**o+(-0.740418501951)*x[2]
        s[3,1,0]=(-0.0448849811136)*x[0]**o+(-0.337048179302)*x[0]+(-0.504885976058)*x[1]**o+(-0.445450937517)*x[1]+(0.567028111275)*x[2]**o+(0.123980108452)*x[2]
        s[3,1,1]=(-0.438038042566)*x[0]**o+(-0.427397249753)*x[0]+(0.574195396827)*x[1]**o+(0.824799756958)*x[1]+(-0.886957824366)*x[2]**o+(-0.192557581967)*x[2]
        s[4,0,0]=(0.961581428919)*x[0]**o+(-0.666268681241)*x[0]+(0.916527151658)*x[1]**o+(0.879372699554)*x[1]+(-0.648140763963)*x[2]**o+(-0.692256048468)*x[2]
        s[4,0,1]=(-0.162077276705)*x[0]**o+(-0.306833194309)*x[0]+(0.611399808831)*x[1]**o+(-0.508477663404)*x[1]+(-0.380712354501)*x[2]**o+(0.490458516352)*x[2]
        s[4,1,0]=(-0.64539443571)*x[0]**o+(-0.570490580768)*x[0]+(-0.612017756646)*x[1]**o+(0.979359105468)*x[1]+(0.901671131984)*x[2]**o+(-0.57388809992)*x[2]
        s[4,1,1]=(0.0239934065942)*x[0]**o+(0.358848963609)*x[0]+(-0.262596433728)*x[1]**o+(0.966612108153)*x[1]+(0.048144124831)*x[2]**o+(0.106585203917)*x[2]
        s[5,0,0]=(-0.894844992009)*x[0]**o+(-0.991143708359)*x[0]+(-0.719802159076)*x[1]**o+(0.609678198738)*x[1]+(0.934875984305)*x[2]**o+(0.210728029537)*x[2]
        s[5,0,1]=(-0.0669111311933)*x[0]**o+(-0.145495535442)*x[0]+(-0.988259077226)*x[1]**o+(-0.736349726987)*x[1]+(-0.164335654094)*x[2]**o+(-0.364166925171)*x[2]
        s[5,1,0]=(-0.279004883801)*x[0]**o+(0.96466483765)*x[0]+(-0.445599169347)*x[1]**o+(0.866248162976)*x[1]+(0.13706485262)*x[2]**o+(-0.78704568071)*x[2]
        s[5,1,1]=(0.452358514637)*x[0]**o+(-0.456298799376)*x[0]+(0.80484494772)*x[1]**o+(-0.949154901869)*x[1]+(-0.106819359297)*x[2]**o+(-0.710983814219)*x[2]
        ref[0,0,0,0]=o*(0.0799106193628)*x_ref[0]**(o-1)+(-0.385218529142)
        ref[0,0,0,1]=o*(0.423551372182)*x_ref[1]**(o-1)+(-0.357661802001)
        ref[0,0,0,2]=o*(0.523815428088)*x_ref[2]**(o-1)+(-0.867664901871)
        ref[0,0,1,0]=o*(-0.223298613593)*x_ref[0]**(o-1)+(-0.563675059615)
        ref[0,0,1,1]=o*(0.854553636152)*x_ref[1]**(o-1)+(0.00619936897032)
        ref[0,0,1,2]=o*(-0.168119043675)*x_ref[2]**(o-1)+(-0.959967076262)
        ref[0,1,0,0]=o*(-0.538044222292)*x_ref[0]**(o-1)+(0.519685043014)
        ref[0,1,0,1]=o*(-0.351144688067)*x_ref[1]**(o-1)+(0.887365493845)
        ref[0,1,0,2]=o*(0.801566727487)*x_ref[2]**(o-1)+(0.571280510487)
        ref[0,1,1,0]=o*(0.784344723309)*x_ref[0]**(o-1)+(-0.0907512562255)
        ref[0,1,1,1]=o*(-0.24324080888)*x_ref[1]**(o-1)+(-0.199412485507)
        ref[0,1,1,2]=o*(0.642344630349)*x_ref[2]**(o-1)+(-0.735616137201)
        ref[1,0,0,0]=o*(-0.880206903193)*x_ref[0]**(o-1)+(-0.749123123033)
        ref[1,0,0,1]=o*(0.104561575813)*x_ref[1]**(o-1)+(-0.647441987473)
        ref[1,0,0,2]=o*(0.182201321281)*x_ref[2]**(o-1)+(0.0102973222322)
        ref[1,0,1,0]=o*(0.915893198392)*x_ref[0]**(o-1)+(-0.612952390562)
        ref[1,0,1,1]=o*(0.803094622661)*x_ref[1]**(o-1)+(-0.140410725667)
        ref[1,0,1,2]=o*(-0.247855770678)*x_ref[2]**(o-1)+(0.399254178737)
        ref[1,1,0,0]=o*(0.818321238866)*x_ref[0]**(o-1)+(0.214893239645)
        ref[1,1,0,1]=o*(0.909338463913)*x_ref[1]**(o-1)+(0.855186366612)
        ref[1,1,0,2]=o*(-0.835375145102)*x_ref[2]**(o-1)+(-0.807936215386)
        ref[1,1,1,0]=o*(0.257588723885)*x_ref[0]**(o-1)+(-0.952059092336)
        ref[1,1,1,1]=o*(-0.28964814303)*x_ref[1]**(o-1)+(0.0129418032921)
        ref[1,1,1,2]=o*(0.854165719561)*x_ref[2]**(o-1)+(0.16196867256)
        ref[2,0,0,0]=o*(-0.291805757446)*x_ref[0]**(o-1)+(0.609410741859)
        ref[2,0,0,1]=o*(0.541295607728)*x_ref[1]**(o-1)+(-0.220049109439)
        ref[2,0,0,2]=o*(0.773352330852)*x_ref[2]**(o-1)+(-0.0999084978714)
        ref[2,0,1,0]=o*(-0.094477913729)*x_ref[0]**(o-1)+(-0.764264730185)
        ref[2,0,1,1]=o*(-0.671804542061)*x_ref[1]**(o-1)+(0.368646822044)
        ref[2,0,1,2]=o*(0.448237271807)*x_ref[2]**(o-1)+(0.565106260459)
        ref[2,1,0,0]=o*(-0.240257182637)*x_ref[0]**(o-1)+(0.859112706839)
        ref[2,1,0,1]=o*(-0.575504060232)*x_ref[1]**(o-1)+(-0.611929236978)
        ref[2,1,0,2]=o*(0.675448692587)*x_ref[2]**(o-1)+(0.31890080218)
        ref[2,1,1,0]=o*(-0.39222652102)*x_ref[0]**(o-1)+(0.677458518423)
        ref[2,1,1,1]=o*(-0.412612361541)*x_ref[1]**(o-1)+(-0.374686862277)
        ref[2,1,1,2]=o*(-0.737213784962)*x_ref[2]**(o-1)+(-0.486920606866)
        ref[3,0,0,0]=o*(0.635190942363)*x_ref[0]**(o-1)+(-0.330731024263)
        ref[3,0,0,1]=o*(-0.690112375174)*x_ref[1]**(o-1)+(-0.446491601534)
        ref[3,0,0,2]=o*(-0.748307399262)*x_ref[2]**(o-1)+(-0.578747596175)
        ref[3,0,1,0]=o*(-0.419848395854)*x_ref[0]**(o-1)+(-0.31014199789)
        ref[3,0,1,1]=o*(0.818235061383)*x_ref[1]**(o-1)+(0.0547944058433)
        ref[3,0,1,2]=o*(0.855573639674)*x_ref[2]**(o-1)+(-0.740418501951)
        ref[3,1,0,0]=o*(-0.0448849811136)*x_ref[0]**(o-1)+(-0.337048179302)
        ref[3,1,0,1]=o*(-0.504885976058)*x_ref[1]**(o-1)+(-0.445450937517)
        ref[3,1,0,2]=o*(0.567028111275)*x_ref[2]**(o-1)+(0.123980108452)
        ref[3,1,1,0]=o*(-0.438038042566)*x_ref[0]**(o-1)+(-0.427397249753)
        ref[3,1,1,1]=o*(0.574195396827)*x_ref[1]**(o-1)+(0.824799756958)
        ref[3,1,1,2]=o*(-0.886957824366)*x_ref[2]**(o-1)+(-0.192557581967)
        ref[4,0,0,0]=o*(0.961581428919)*x_ref[0]**(o-1)+(-0.666268681241)
        ref[4,0,0,1]=o*(0.916527151658)*x_ref[1]**(o-1)+(0.879372699554)
        ref[4,0,0,2]=o*(-0.648140763963)*x_ref[2]**(o-1)+(-0.692256048468)
        ref[4,0,1,0]=o*(-0.162077276705)*x_ref[0]**(o-1)+(-0.306833194309)
        ref[4,0,1,1]=o*(0.611399808831)*x_ref[1]**(o-1)+(-0.508477663404)
        ref[4,0,1,2]=o*(-0.380712354501)*x_ref[2]**(o-1)+(0.490458516352)
        ref[4,1,0,0]=o*(-0.64539443571)*x_ref[0]**(o-1)+(-0.570490580768)
        ref[4,1,0,1]=o*(-0.612017756646)*x_ref[1]**(o-1)+(0.979359105468)
        ref[4,1,0,2]=o*(0.901671131984)*x_ref[2]**(o-1)+(-0.57388809992)
        ref[4,1,1,0]=o*(0.0239934065942)*x_ref[0]**(o-1)+(0.358848963609)
        ref[4,1,1,1]=o*(-0.262596433728)*x_ref[1]**(o-1)+(0.966612108153)
        ref[4,1,1,2]=o*(0.048144124831)*x_ref[2]**(o-1)+(0.106585203917)
        ref[5,0,0,0]=o*(-0.894844992009)*x_ref[0]**(o-1)+(-0.991143708359)
        ref[5,0,0,1]=o*(-0.719802159076)*x_ref[1]**(o-1)+(0.609678198738)
        ref[5,0,0,2]=o*(0.934875984305)*x_ref[2]**(o-1)+(0.210728029537)
        ref[5,0,1,0]=o*(-0.0669111311933)*x_ref[0]**(o-1)+(-0.145495535442)
        ref[5,0,1,1]=o*(-0.988259077226)*x_ref[1]**(o-1)+(-0.736349726987)
        ref[5,0,1,2]=o*(-0.164335654094)*x_ref[2]**(o-1)+(-0.364166925171)
        ref[5,1,0,0]=o*(-0.279004883801)*x_ref[0]**(o-1)+(0.96466483765)
        ref[5,1,0,1]=o*(-0.445599169347)*x_ref[1]**(o-1)+(0.866248162976)
        ref[5,1,0,2]=o*(0.13706485262)*x_ref[2]**(o-1)+(-0.78704568071)
        ref[5,1,1,0]=o*(0.452358514637)*x_ref[0]**(o-1)+(-0.456298799376)
        ref[5,1,1,1]=o*(0.80484494772)*x_ref[1]**(o-1)+(-0.949154901869)
        ref[5,1,1,2]=o*(-0.106819359297)*x_ref[2]**(o-1)+(-0.710983814219)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests gradient for rank 0 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.495552973369)*x[0]+(0.874390589312)*x[1]
        ref[0]=(-0.495552973369)
        ref[1]=(0.874390589312)
      else:
        s=(-0.354636799848)*x[0]+(0.555572828684)*x[1]+(-0.270947199201)*x[2]
        ref[0]=(-0.354636799848)
        ref[1]=(0.555572828684)
        ref[2]=(-0.270947199201)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests gradient for rank 1 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(-0.187161756695)*x[0]+(-0.776766204086)*x[1]
        s[1]=(0.724774116602)*x[0]+(0.446775762856)*x[1]
        ref[0,0]=(-0.187161756695)
        ref[0,1]=(-0.776766204086)
        ref[1,0]=(0.724774116602)
        ref[1,1]=(0.446775762856)
      else:
        s[0]=(-0.0251530250542)*x[0]+(-1.23786881883)*x[1]+(1.1819624619)*x[2]
        s[1]=(0.801009266263)*x[0]+(-0.210512005637)*x[1]+(-1.42548669085)*x[2]
        ref[0,0]=(-0.0251530250542)
        ref[0,1]=(-1.23786881883)
        ref[0,2]=(1.1819624619)
        ref[1,0]=(0.801009266263)
        ref[1,1]=(-0.210512005637)
        ref[1,2]=(-1.42548669085)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests gradient for rank 2 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.105140628344)*x[0]+(0.681061911151)*x[1]
        s[0,1]=(1.12568473028)*x[0]+(-0.142806409661)*x[1]
        s[0,2]=(-0.709551703126)*x[0]+(0.470204858155)*x[1]
        s[0,3]=(-0.457889995297)*x[0]+(0.119831077838)*x[1]
        s[0,4]=(-0.137584694085)*x[0]+(-0.815364572262)*x[1]
        s[1,0]=(-0.747275494188)*x[0]+(0.419390416238)*x[1]
        s[1,1]=(-0.55554960963)*x[0]+(1.68766532296)*x[1]
        s[1,2]=(0.0380274635555)*x[0]+(0.903518514201)*x[1]
        s[1,3]=(1.30478047839)*x[0]+(0.576658405128)*x[1]
        s[1,4]=(-0.786884541126)*x[0]+(0.481365349672)*x[1]
        s[2,0]=(1.30445628093)*x[0]+(1.06765714413)*x[1]
        s[2,1]=(0.7269884164)*x[0]+(-1.34914325614)*x[1]
        s[2,2]=(-0.827940783305)*x[0]+(1.05409369275)*x[1]
        s[2,3]=(0.57117918118)*x[0]+(-0.857074951332)*x[1]
        s[2,4]=(0.808649779643)*x[0]+(-0.746017104416)*x[1]
        s[3,0]=(1.06443222736)*x[0]+(-0.458402762622)*x[1]
        s[3,1]=(-0.328990411524)*x[0]+(1.10697234223)*x[1]
        s[3,2]=(-0.0777059446026)*x[0]+(1.04837766698)*x[1]
        s[3,3]=(0.0542117671967)*x[0]+(-1.24129430221)*x[1]
        s[3,4]=(0.176491870492)*x[0]+(0.310238631995)*x[1]
        ref[0,0,0]=(-0.105140628344)
        ref[0,0,1]=(0.681061911151)
        ref[0,1,0]=(1.12568473028)
        ref[0,1,1]=(-0.142806409661)
        ref[0,2,0]=(-0.709551703126)
        ref[0,2,1]=(0.470204858155)
        ref[0,3,0]=(-0.457889995297)
        ref[0,3,1]=(0.119831077838)
        ref[0,4,0]=(-0.137584694085)
        ref[0,4,1]=(-0.815364572262)
        ref[1,0,0]=(-0.747275494188)
        ref[1,0,1]=(0.419390416238)
        ref[1,1,0]=(-0.55554960963)
        ref[1,1,1]=(1.68766532296)
        ref[1,2,0]=(0.0380274635555)
        ref[1,2,1]=(0.903518514201)
        ref[1,3,0]=(1.30478047839)
        ref[1,3,1]=(0.576658405128)
        ref[1,4,0]=(-0.786884541126)
        ref[1,4,1]=(0.481365349672)
        ref[2,0,0]=(1.30445628093)
        ref[2,0,1]=(1.06765714413)
        ref[2,1,0]=(0.7269884164)
        ref[2,1,1]=(-1.34914325614)
        ref[2,2,0]=(-0.827940783305)
        ref[2,2,1]=(1.05409369275)
        ref[2,3,0]=(0.57117918118)
        ref[2,3,1]=(-0.857074951332)
        ref[2,4,0]=(0.808649779643)
        ref[2,4,1]=(-0.746017104416)
        ref[3,0,0]=(1.06443222736)
        ref[3,0,1]=(-0.458402762622)
        ref[3,1,0]=(-0.328990411524)
        ref[3,1,1]=(1.10697234223)
        ref[3,2,0]=(-0.0777059446026)
        ref[3,2,1]=(1.04837766698)
        ref[3,3,0]=(0.0542117671967)
        ref[3,3,1]=(-1.24129430221)
        ref[3,4,0]=(0.176491870492)
        ref[3,4,1]=(0.310238631995)
      else:
        s[0,0]=(-0.835695322993)*x[0]+(0.168930135561)*x[1]+(-1.47180800481)*x[2]
        s[0,1]=(0.2001836256)*x[0]+(1.27356263599)*x[1]+(1.07485276442)*x[2]
        s[0,2]=(-0.178446572149)*x[0]+(-0.420078906939)*x[1]+(0.535633192895)*x[2]
        s[0,3]=(-0.0910549937921)*x[0]+(0.594316242371)*x[1]+(-1.06833005008)*x[2]
        s[0,4]=(0.722857786921)*x[0]+(-0.332473817629)*x[1]+(0.79883607252)*x[2]
        s[1,0]=(-1.02192468444)*x[0]+(1.2735296387)*x[1]+(-1.34574119314)*x[2]
        s[1,1]=(-1.20222077943)*x[0]+(0.180006226359)*x[1]+(0.582580396324)*x[2]
        s[1,2]=(0.292773497866)*x[0]+(0.503071430702)*x[1]+(1.25357442113)*x[2]
        s[1,3]=(-0.579871204353)*x[0]+(0.143994355108)*x[1]+(0.50466279341)*x[2]
        s[1,4]=(1.00530781493)*x[0]+(-0.646038982555)*x[1]+(0.315868158577)*x[2]
        s[2,0]=(0.0109517370271)*x[0]+(-1.27094713044)*x[1]+(0.647777134752)*x[2]
        s[2,1]=(0.615504753112)*x[0]+(0.479866346157)*x[1]+(1.72115593961)*x[2]
        s[2,2]=(-0.307778323155)*x[0]+(0.307520362533)*x[1]+(0.0061507801029)*x[2]
        s[2,3]=(-0.901393000596)*x[0]+(0.912379707397)*x[1]+(-0.0683825966669)*x[2]
        s[2,4]=(0.558642962243)*x[0]+(0.650906862857)*x[1]+(-1.44796931057)*x[2]
        s[3,0]=(1.08498680281)*x[0]+(-1.28372501216)*x[1]+(-1.27799790624)*x[2]
        s[3,1]=(0.45408516398)*x[0]+(-0.950068476479)*x[1]+(0.634133482831)*x[2]
        s[3,2]=(0.824015492184)*x[0]+(0.148535318857)*x[1]+(0.10609796765)*x[2]
        s[3,3]=(1.14342153343)*x[0]+(-0.0689548622931)*x[1]+(-0.166489981522)*x[2]
        s[3,4]=(0.995987765181)*x[0]+(0.915359249488)*x[1]+(-0.583052373261)*x[2]
        ref[0,0,0]=(-0.835695322993)
        ref[0,0,1]=(0.168930135561)
        ref[0,0,2]=(-1.47180800481)
        ref[0,1,0]=(0.2001836256)
        ref[0,1,1]=(1.27356263599)
        ref[0,1,2]=(1.07485276442)
        ref[0,2,0]=(-0.178446572149)
        ref[0,2,1]=(-0.420078906939)
        ref[0,2,2]=(0.535633192895)
        ref[0,3,0]=(-0.0910549937921)
        ref[0,3,1]=(0.594316242371)
        ref[0,3,2]=(-1.06833005008)
        ref[0,4,0]=(0.722857786921)
        ref[0,4,1]=(-0.332473817629)
        ref[0,4,2]=(0.79883607252)
        ref[1,0,0]=(-1.02192468444)
        ref[1,0,1]=(1.2735296387)
        ref[1,0,2]=(-1.34574119314)
        ref[1,1,0]=(-1.20222077943)
        ref[1,1,1]=(0.180006226359)
        ref[1,1,2]=(0.582580396324)
        ref[1,2,0]=(0.292773497866)
        ref[1,2,1]=(0.503071430702)
        ref[1,2,2]=(1.25357442113)
        ref[1,3,0]=(-0.579871204353)
        ref[1,3,1]=(0.143994355108)
        ref[1,3,2]=(0.50466279341)
        ref[1,4,0]=(1.00530781493)
        ref[1,4,1]=(-0.646038982555)
        ref[1,4,2]=(0.315868158577)
        ref[2,0,0]=(0.0109517370271)
        ref[2,0,1]=(-1.27094713044)
        ref[2,0,2]=(0.647777134752)
        ref[2,1,0]=(0.615504753112)
        ref[2,1,1]=(0.479866346157)
        ref[2,1,2]=(1.72115593961)
        ref[2,2,0]=(-0.307778323155)
        ref[2,2,1]=(0.307520362533)
        ref[2,2,2]=(0.0061507801029)
        ref[2,3,0]=(-0.901393000596)
        ref[2,3,1]=(0.912379707397)
        ref[2,3,2]=(-0.0683825966669)
        ref[2,4,0]=(0.558642962243)
        ref[2,4,1]=(0.650906862857)
        ref[2,4,2]=(-1.44796931057)
        ref[3,0,0]=(1.08498680281)
        ref[3,0,1]=(-1.28372501216)
        ref[3,0,2]=(-1.27799790624)
        ref[3,1,0]=(0.45408516398)
        ref[3,1,1]=(-0.950068476479)
        ref[3,1,2]=(0.634133482831)
        ref[3,2,0]=(0.824015492184)
        ref[3,2,1]=(0.148535318857)
        ref[3,2,2]=(0.10609796765)
        ref[3,3,0]=(1.14342153343)
        ref[3,3,1]=(-0.0689548622931)
        ref[3,3,2]=(-0.166489981522)
        ref[3,4,0]=(0.995987765181)
        ref[3,4,1]=(0.915359249488)
        ref[3,4,2]=(-0.583052373261)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests gradient for rank 3 Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(0.116652834373)*x[0]+(1.4681512344)*x[1]
        s[0,0,1]=(1.89418390988)*x[0]+(-0.548065006619)*x[1]
        s[0,1,0]=(-0.07602392145)*x[0]+(0.0796515016952)*x[1]
        s[0,1,1]=(1.50333606195)*x[0]+(-0.211673198103)*x[1]
        s[1,0,0]=(0.172419295955)*x[0]+(-0.586603468045)*x[1]
        s[1,0,1]=(0.212952657549)*x[0]+(-0.100672655087)*x[1]
        s[1,1,0]=(-1.11939406352)*x[0]+(-0.589882254225)*x[1]
        s[1,1,1]=(0.318138253303)*x[0]+(0.409850277983)*x[1]
        s[2,0,0]=(-0.0669759465557)*x[0]+(0.525669761096)*x[1]
        s[2,0,1]=(0.349019390883)*x[0]+(0.154604501057)*x[1]
        s[2,1,0]=(1.95277106546)*x[0]+(0.495701766483)*x[1]
        s[2,1,1]=(-0.0986210008015)*x[0]+(1.12636829398)*x[1]
        s[3,0,0]=(0.971837535878)*x[0]+(-0.738253595188)*x[1]
        s[3,0,1]=(0.0314513203112)*x[0]+(-1.16646217783)*x[1]
        s[3,1,0]=(0.485751127243)*x[0]+(-0.340178033471)*x[1]
        s[3,1,1]=(0.408742867205)*x[0]+(0.151827789426)*x[1]
        s[4,0,0]=(-0.0134678971538)*x[0]+(0.884924133141)*x[1]
        s[4,0,1]=(-0.66899222691)*x[0]+(-0.959185584071)*x[1]
        s[4,1,0]=(1.05311767886)*x[0]+(0.64093792324)*x[1]
        s[4,1,1]=(0.562875594668)*x[0]+(0.539525481691)*x[1]
        s[5,0,0]=(0.310908751279)*x[0]+(0.497074267163)*x[1]
        s[5,0,1]=(-0.804186513465)*x[0]+(0.329762438933)*x[1]
        s[5,1,0]=(-0.946808260459)*x[0]+(-0.00260213368693)*x[1]
        s[5,1,1]=(0.63415870202)*x[0]+(1.38103994652)*x[1]
        ref[0,0,0,0]=(0.116652834373)
        ref[0,0,0,1]=(1.4681512344)
        ref[0,0,1,0]=(1.89418390988)
        ref[0,0,1,1]=(-0.548065006619)
        ref[0,1,0,0]=(-0.07602392145)
        ref[0,1,0,1]=(0.0796515016952)
        ref[0,1,1,0]=(1.50333606195)
        ref[0,1,1,1]=(-0.211673198103)
        ref[1,0,0,0]=(0.172419295955)
        ref[1,0,0,1]=(-0.586603468045)
        ref[1,0,1,0]=(0.212952657549)
        ref[1,0,1,1]=(-0.100672655087)
        ref[1,1,0,0]=(-1.11939406352)
        ref[1,1,0,1]=(-0.589882254225)
        ref[1,1,1,0]=(0.318138253303)
        ref[1,1,1,1]=(0.409850277983)
        ref[2,0,0,0]=(-0.0669759465557)
        ref[2,0,0,1]=(0.525669761096)
        ref[2,0,1,0]=(0.349019390883)
        ref[2,0,1,1]=(0.154604501057)
        ref[2,1,0,0]=(1.95277106546)
        ref[2,1,0,1]=(0.495701766483)
        ref[2,1,1,0]=(-0.0986210008015)
        ref[2,1,1,1]=(1.12636829398)
        ref[3,0,0,0]=(0.971837535878)
        ref[3,0,0,1]=(-0.738253595188)
        ref[3,0,1,0]=(0.0314513203112)
        ref[3,0,1,1]=(-1.16646217783)
        ref[3,1,0,0]=(0.485751127243)
        ref[3,1,0,1]=(-0.340178033471)
        ref[3,1,1,0]=(0.408742867205)
        ref[3,1,1,1]=(0.151827789426)
        ref[4,0,0,0]=(-0.0134678971538)
        ref[4,0,0,1]=(0.884924133141)
        ref[4,0,1,0]=(-0.66899222691)
        ref[4,0,1,1]=(-0.959185584071)
        ref[4,1,0,0]=(1.05311767886)
        ref[4,1,0,1]=(0.64093792324)
        ref[4,1,1,0]=(0.562875594668)
        ref[4,1,1,1]=(0.539525481691)
        ref[5,0,0,0]=(0.310908751279)
        ref[5,0,0,1]=(0.497074267163)
        ref[5,0,1,0]=(-0.804186513465)
        ref[5,0,1,1]=(0.329762438933)
        ref[5,1,0,0]=(-0.946808260459)
        ref[5,1,0,1]=(-0.00260213368693)
        ref[5,1,1,0]=(0.63415870202)
        ref[5,1,1,1]=(1.38103994652)
      else:
        s[0,0,0]=(-0.0840260391925)*x[0]+(0.385117859595)*x[1]+(1.36354425949)*x[2]
        s[0,0,1]=(0.682480367755)*x[0]+(-0.658980348414)*x[1]+(-0.129734588977)*x[2]
        s[0,1,0]=(-0.213790105123)*x[0]+(-0.774270821015)*x[1]+(0.160024784274)*x[2]
        s[0,1,1]=(-1.02270895286)*x[0]+(1.02607310825)*x[1]+(0.35056114816)*x[2]
        s[1,0,0]=(0.201792116067)*x[0]+(0.223564282304)*x[1]+(0.986785519545)*x[2]
        s[1,0,1]=(0.250819712775)*x[0]+(0.287643279417)*x[1]+(0.18295590528)*x[2]
        s[1,1,0]=(-1.47569225207)*x[0]+(-0.886933933801)*x[1]+(0.525380474082)*x[2]
        s[1,1,1]=(0.680401800239)*x[0]+(1.36155081028)*x[1]+(-1.76198171789)*x[2]
        s[2,0,0]=(-0.784439188011)*x[0]+(-0.797291895631)*x[1]+(-0.425550387518)*x[2]
        s[2,0,1]=(1.62016996989)*x[0]+(1.02824549576)*x[1]+(-0.157995046652)*x[2]
        s[2,1,0]=(-0.437412344413)*x[0]+(0.97005624264)*x[1]+(-0.321998550341)*x[2]
        s[2,1,1]=(0.473879660993)*x[0]+(0.483144823024)*x[1]+(0.168449443711)*x[2]
        s[3,0,0]=(-1.59822104736)*x[0]+(-1.24012649834)*x[1]+(0.695408081789)*x[2]
        s[3,0,1]=(0.827901715486)*x[0]+(-0.230621260402)*x[1]+(-0.789726238476)*x[2]
        s[3,1,0]=(-0.279064257268)*x[0]+(-1.35315417734)*x[1]+(0.593090655629)*x[2]
        s[3,1,1]=(0.387685237641)*x[0]+(1.60258880823)*x[1]+(-1.10088045432)*x[2]
        s[4,0,0]=(0.771374132686)*x[0]+(0.530393756307)*x[1]+(-0.366697715551)*x[2]
        s[4,0,1]=(-0.182374799863)*x[0]+(-0.125579433837)*x[1]+(-1.2028145445)*x[2]
        s[4,1,0]=(-0.208766654808)*x[0]+(0.645251941741)*x[1]+(-0.534068338041)*x[2]
        s[4,1,1]=(-0.064786779421)*x[0]+(0.0354509511326)*x[1]+(-1.32632049137)*x[2]
        s[5,0,0]=(0.833438445425)*x[0]+(0.770186199496)*x[1]+(0.112178380686)*x[2]
        s[5,0,1]=(1.36614714953)*x[0]+(-0.367243359207)*x[1]+(0.812867174908)*x[2]
        s[5,1,0]=(0.0359079603828)*x[0]+(0.622094474303)*x[1]+(-0.33137187739)*x[2]
        s[5,1,1]=(-0.479714378899)*x[0]+(-0.959098499695)*x[1]+(0.843889761479)*x[2]
        ref[0,0,0,0]=(-0.0840260391925)
        ref[0,0,0,1]=(0.385117859595)
        ref[0,0,0,2]=(1.36354425949)
        ref[0,0,1,0]=(0.682480367755)
        ref[0,0,1,1]=(-0.658980348414)
        ref[0,0,1,2]=(-0.129734588977)
        ref[0,1,0,0]=(-0.213790105123)
        ref[0,1,0,1]=(-0.774270821015)
        ref[0,1,0,2]=(0.160024784274)
        ref[0,1,1,0]=(-1.02270895286)
        ref[0,1,1,1]=(1.02607310825)
        ref[0,1,1,2]=(0.35056114816)
        ref[1,0,0,0]=(0.201792116067)
        ref[1,0,0,1]=(0.223564282304)
        ref[1,0,0,2]=(0.986785519545)
        ref[1,0,1,0]=(0.250819712775)
        ref[1,0,1,1]=(0.287643279417)
        ref[1,0,1,2]=(0.18295590528)
        ref[1,1,0,0]=(-1.47569225207)
        ref[1,1,0,1]=(-0.886933933801)
        ref[1,1,0,2]=(0.525380474082)
        ref[1,1,1,0]=(0.680401800239)
        ref[1,1,1,1]=(1.36155081028)
        ref[1,1,1,2]=(-1.76198171789)
        ref[2,0,0,0]=(-0.784439188011)
        ref[2,0,0,1]=(-0.797291895631)
        ref[2,0,0,2]=(-0.425550387518)
        ref[2,0,1,0]=(1.62016996989)
        ref[2,0,1,1]=(1.02824549576)
        ref[2,0,1,2]=(-0.157995046652)
        ref[2,1,0,0]=(-0.437412344413)
        ref[2,1,0,1]=(0.97005624264)
        ref[2,1,0,2]=(-0.321998550341)
        ref[2,1,1,0]=(0.473879660993)
        ref[2,1,1,1]=(0.483144823024)
        ref[2,1,1,2]=(0.168449443711)
        ref[3,0,0,0]=(-1.59822104736)
        ref[3,0,0,1]=(-1.24012649834)
        ref[3,0,0,2]=(0.695408081789)
        ref[3,0,1,0]=(0.827901715486)
        ref[3,0,1,1]=(-0.230621260402)
        ref[3,0,1,2]=(-0.789726238476)
        ref[3,1,0,0]=(-0.279064257268)
        ref[3,1,0,1]=(-1.35315417734)
        ref[3,1,0,2]=(0.593090655629)
        ref[3,1,1,0]=(0.387685237641)
        ref[3,1,1,1]=(1.60258880823)
        ref[3,1,1,2]=(-1.10088045432)
        ref[4,0,0,0]=(0.771374132686)
        ref[4,0,0,1]=(0.530393756307)
        ref[4,0,0,2]=(-0.366697715551)
        ref[4,0,1,0]=(-0.182374799863)
        ref[4,0,1,1]=(-0.125579433837)
        ref[4,0,1,2]=(-1.2028145445)
        ref[4,1,0,0]=(-0.208766654808)
        ref[4,1,0,1]=(0.645251941741)
        ref[4,1,0,2]=(-0.534068338041)
        ref[4,1,1,0]=(-0.064786779421)
        ref[4,1,1,1]=(0.0354509511326)
        ref[4,1,1,2]=(-1.32632049137)
        ref[5,0,0,0]=(0.833438445425)
        ref[5,0,0,1]=(0.770186199496)
        ref[5,0,0,2]=(0.112178380686)
        ref[5,0,1,0]=(1.36614714953)
        ref[5,0,1,1]=(-0.367243359207)
        ref[5,0,1,2]=(0.812867174908)
        ref[5,1,0,0]=(0.0359079603828)
        ref[5,1,0,1]=(0.622094474303)
        ref[5,1,0,2]=(-0.33137187739)
        ref[5,1,1,0]=(-0.479714378899)
        ref[5,1,1,1]=(-0.959098499695)
        ref[5,1,1,2]=(0.843889761479)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests gradient for rank 0 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,()+(dim,),w_ref)
      if dim==2:
        s=(-0.495552973369)*x[0]+(0.874390589312)*x[1]
        ref[0]=(-0.495552973369)
        ref[1]=(0.874390589312)
      else:
        s=(-0.354636799848)*x[0]+(0.555572828684)*x[1]+(-0.270947199201)*x[2]
        ref[0]=(-0.354636799848)
        ref[1]=(0.555572828684)
        ref[2]=(-0.270947199201)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),()+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests gradient for rank 1 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,)+(dim,),w_ref)
      if dim==2:
        s[0]=(-0.187161756695)*x[0]+(-0.776766204086)*x[1]
        s[1]=(0.724774116602)*x[0]+(0.446775762856)*x[1]
        ref[0,0]=(-0.187161756695)
        ref[0,1]=(-0.776766204086)
        ref[1,0]=(0.724774116602)
        ref[1,1]=(0.446775762856)
      else:
        s[0]=(-0.0251530250542)*x[0]+(-1.23786881883)*x[1]+(1.1819624619)*x[2]
        s[1]=(0.801009266263)*x[0]+(-0.210512005637)*x[1]+(-1.42548669085)*x[2]
        ref[0,0]=(-0.0251530250542)
        ref[0,1]=(-1.23786881883)
        ref[0,2]=(1.1819624619)
        ref[1,0]=(0.801009266263)
        ref[1,1]=(-0.210512005637)
        ref[1,2]=(-1.42548669085)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(2,)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests gradient for rank 2 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5)+(dim,),w_ref)
      if dim==2:
        s[0,0]=(-0.105140628344)*x[0]+(0.681061911151)*x[1]
        s[0,1]=(1.12568473028)*x[0]+(-0.142806409661)*x[1]
        s[0,2]=(-0.709551703126)*x[0]+(0.470204858155)*x[1]
        s[0,3]=(-0.457889995297)*x[0]+(0.119831077838)*x[1]
        s[0,4]=(-0.137584694085)*x[0]+(-0.815364572262)*x[1]
        s[1,0]=(-0.747275494188)*x[0]+(0.419390416238)*x[1]
        s[1,1]=(-0.55554960963)*x[0]+(1.68766532296)*x[1]
        s[1,2]=(0.0380274635555)*x[0]+(0.903518514201)*x[1]
        s[1,3]=(1.30478047839)*x[0]+(0.576658405128)*x[1]
        s[1,4]=(-0.786884541126)*x[0]+(0.481365349672)*x[1]
        s[2,0]=(1.30445628093)*x[0]+(1.06765714413)*x[1]
        s[2,1]=(0.7269884164)*x[0]+(-1.34914325614)*x[1]
        s[2,2]=(-0.827940783305)*x[0]+(1.05409369275)*x[1]
        s[2,3]=(0.57117918118)*x[0]+(-0.857074951332)*x[1]
        s[2,4]=(0.808649779643)*x[0]+(-0.746017104416)*x[1]
        s[3,0]=(1.06443222736)*x[0]+(-0.458402762622)*x[1]
        s[3,1]=(-0.328990411524)*x[0]+(1.10697234223)*x[1]
        s[3,2]=(-0.0777059446026)*x[0]+(1.04837766698)*x[1]
        s[3,3]=(0.0542117671967)*x[0]+(-1.24129430221)*x[1]
        s[3,4]=(0.176491870492)*x[0]+(0.310238631995)*x[1]
        ref[0,0,0]=(-0.105140628344)
        ref[0,0,1]=(0.681061911151)
        ref[0,1,0]=(1.12568473028)
        ref[0,1,1]=(-0.142806409661)
        ref[0,2,0]=(-0.709551703126)
        ref[0,2,1]=(0.470204858155)
        ref[0,3,0]=(-0.457889995297)
        ref[0,3,1]=(0.119831077838)
        ref[0,4,0]=(-0.137584694085)
        ref[0,4,1]=(-0.815364572262)
        ref[1,0,0]=(-0.747275494188)
        ref[1,0,1]=(0.419390416238)
        ref[1,1,0]=(-0.55554960963)
        ref[1,1,1]=(1.68766532296)
        ref[1,2,0]=(0.0380274635555)
        ref[1,2,1]=(0.903518514201)
        ref[1,3,0]=(1.30478047839)
        ref[1,3,1]=(0.576658405128)
        ref[1,4,0]=(-0.786884541126)
        ref[1,4,1]=(0.481365349672)
        ref[2,0,0]=(1.30445628093)
        ref[2,0,1]=(1.06765714413)
        ref[2,1,0]=(0.7269884164)
        ref[2,1,1]=(-1.34914325614)
        ref[2,2,0]=(-0.827940783305)
        ref[2,2,1]=(1.05409369275)
        ref[2,3,0]=(0.57117918118)
        ref[2,3,1]=(-0.857074951332)
        ref[2,4,0]=(0.808649779643)
        ref[2,4,1]=(-0.746017104416)
        ref[3,0,0]=(1.06443222736)
        ref[3,0,1]=(-0.458402762622)
        ref[3,1,0]=(-0.328990411524)
        ref[3,1,1]=(1.10697234223)
        ref[3,2,0]=(-0.0777059446026)
        ref[3,2,1]=(1.04837766698)
        ref[3,3,0]=(0.0542117671967)
        ref[3,3,1]=(-1.24129430221)
        ref[3,4,0]=(0.176491870492)
        ref[3,4,1]=(0.310238631995)
      else:
        s[0,0]=(-0.835695322993)*x[0]+(0.168930135561)*x[1]+(-1.47180800481)*x[2]
        s[0,1]=(0.2001836256)*x[0]+(1.27356263599)*x[1]+(1.07485276442)*x[2]
        s[0,2]=(-0.178446572149)*x[0]+(-0.420078906939)*x[1]+(0.535633192895)*x[2]
        s[0,3]=(-0.0910549937921)*x[0]+(0.594316242371)*x[1]+(-1.06833005008)*x[2]
        s[0,4]=(0.722857786921)*x[0]+(-0.332473817629)*x[1]+(0.79883607252)*x[2]
        s[1,0]=(-1.02192468444)*x[0]+(1.2735296387)*x[1]+(-1.34574119314)*x[2]
        s[1,1]=(-1.20222077943)*x[0]+(0.180006226359)*x[1]+(0.582580396324)*x[2]
        s[1,2]=(0.292773497866)*x[0]+(0.503071430702)*x[1]+(1.25357442113)*x[2]
        s[1,3]=(-0.579871204353)*x[0]+(0.143994355108)*x[1]+(0.50466279341)*x[2]
        s[1,4]=(1.00530781493)*x[0]+(-0.646038982555)*x[1]+(0.315868158577)*x[2]
        s[2,0]=(0.0109517370271)*x[0]+(-1.27094713044)*x[1]+(0.647777134752)*x[2]
        s[2,1]=(0.615504753112)*x[0]+(0.479866346157)*x[1]+(1.72115593961)*x[2]
        s[2,2]=(-0.307778323155)*x[0]+(0.307520362533)*x[1]+(0.0061507801029)*x[2]
        s[2,3]=(-0.901393000596)*x[0]+(0.912379707397)*x[1]+(-0.0683825966669)*x[2]
        s[2,4]=(0.558642962243)*x[0]+(0.650906862857)*x[1]+(-1.44796931057)*x[2]
        s[3,0]=(1.08498680281)*x[0]+(-1.28372501216)*x[1]+(-1.27799790624)*x[2]
        s[3,1]=(0.45408516398)*x[0]+(-0.950068476479)*x[1]+(0.634133482831)*x[2]
        s[3,2]=(0.824015492184)*x[0]+(0.148535318857)*x[1]+(0.10609796765)*x[2]
        s[3,3]=(1.14342153343)*x[0]+(-0.0689548622931)*x[1]+(-0.166489981522)*x[2]
        s[3,4]=(0.995987765181)*x[0]+(0.915359249488)*x[1]+(-0.583052373261)*x[2]
        ref[0,0,0]=(-0.835695322993)
        ref[0,0,1]=(0.168930135561)
        ref[0,0,2]=(-1.47180800481)
        ref[0,1,0]=(0.2001836256)
        ref[0,1,1]=(1.27356263599)
        ref[0,1,2]=(1.07485276442)
        ref[0,2,0]=(-0.178446572149)
        ref[0,2,1]=(-0.420078906939)
        ref[0,2,2]=(0.535633192895)
        ref[0,3,0]=(-0.0910549937921)
        ref[0,3,1]=(0.594316242371)
        ref[0,3,2]=(-1.06833005008)
        ref[0,4,0]=(0.722857786921)
        ref[0,4,1]=(-0.332473817629)
        ref[0,4,2]=(0.79883607252)
        ref[1,0,0]=(-1.02192468444)
        ref[1,0,1]=(1.2735296387)
        ref[1,0,2]=(-1.34574119314)
        ref[1,1,0]=(-1.20222077943)
        ref[1,1,1]=(0.180006226359)
        ref[1,1,2]=(0.582580396324)
        ref[1,2,0]=(0.292773497866)
        ref[1,2,1]=(0.503071430702)
        ref[1,2,2]=(1.25357442113)
        ref[1,3,0]=(-0.579871204353)
        ref[1,3,1]=(0.143994355108)
        ref[1,3,2]=(0.50466279341)
        ref[1,4,0]=(1.00530781493)
        ref[1,4,1]=(-0.646038982555)
        ref[1,4,2]=(0.315868158577)
        ref[2,0,0]=(0.0109517370271)
        ref[2,0,1]=(-1.27094713044)
        ref[2,0,2]=(0.647777134752)
        ref[2,1,0]=(0.615504753112)
        ref[2,1,1]=(0.479866346157)
        ref[2,1,2]=(1.72115593961)
        ref[2,2,0]=(-0.307778323155)
        ref[2,2,1]=(0.307520362533)
        ref[2,2,2]=(0.0061507801029)
        ref[2,3,0]=(-0.901393000596)
        ref[2,3,1]=(0.912379707397)
        ref[2,3,2]=(-0.0683825966669)
        ref[2,4,0]=(0.558642962243)
        ref[2,4,1]=(0.650906862857)
        ref[2,4,2]=(-1.44796931057)
        ref[3,0,0]=(1.08498680281)
        ref[3,0,1]=(-1.28372501216)
        ref[3,0,2]=(-1.27799790624)
        ref[3,1,0]=(0.45408516398)
        ref[3,1,1]=(-0.950068476479)
        ref[3,1,2]=(0.634133482831)
        ref[3,2,0]=(0.824015492184)
        ref[3,2,1]=(0.148535318857)
        ref[3,2,2]=(0.10609796765)
        ref[3,3,0]=(1.14342153343)
        ref[3,3,1]=(-0.0689548622931)
        ref[3,3,2]=(-0.166489981522)
        ref[3,4,0]=(0.995987765181)
        ref[3,4,1]=(0.915359249488)
        ref[3,4,2]=(-0.583052373261)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(4, 5)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_grad_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests gradient for rank 3 Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports gardient on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2)+(dim,),w_ref)
      if dim==2:
        s[0,0,0]=(0.116652834373)*x[0]+(1.4681512344)*x[1]
        s[0,0,1]=(1.89418390988)*x[0]+(-0.548065006619)*x[1]
        s[0,1,0]=(-0.07602392145)*x[0]+(0.0796515016952)*x[1]
        s[0,1,1]=(1.50333606195)*x[0]+(-0.211673198103)*x[1]
        s[1,0,0]=(0.172419295955)*x[0]+(-0.586603468045)*x[1]
        s[1,0,1]=(0.212952657549)*x[0]+(-0.100672655087)*x[1]
        s[1,1,0]=(-1.11939406352)*x[0]+(-0.589882254225)*x[1]
        s[1,1,1]=(0.318138253303)*x[0]+(0.409850277983)*x[1]
        s[2,0,0]=(-0.0669759465557)*x[0]+(0.525669761096)*x[1]
        s[2,0,1]=(0.349019390883)*x[0]+(0.154604501057)*x[1]
        s[2,1,0]=(1.95277106546)*x[0]+(0.495701766483)*x[1]
        s[2,1,1]=(-0.0986210008015)*x[0]+(1.12636829398)*x[1]
        s[3,0,0]=(0.971837535878)*x[0]+(-0.738253595188)*x[1]
        s[3,0,1]=(0.0314513203112)*x[0]+(-1.16646217783)*x[1]
        s[3,1,0]=(0.485751127243)*x[0]+(-0.340178033471)*x[1]
        s[3,1,1]=(0.408742867205)*x[0]+(0.151827789426)*x[1]
        s[4,0,0]=(-0.0134678971538)*x[0]+(0.884924133141)*x[1]
        s[4,0,1]=(-0.66899222691)*x[0]+(-0.959185584071)*x[1]
        s[4,1,0]=(1.05311767886)*x[0]+(0.64093792324)*x[1]
        s[4,1,1]=(0.562875594668)*x[0]+(0.539525481691)*x[1]
        s[5,0,0]=(0.310908751279)*x[0]+(0.497074267163)*x[1]
        s[5,0,1]=(-0.804186513465)*x[0]+(0.329762438933)*x[1]
        s[5,1,0]=(-0.946808260459)*x[0]+(-0.00260213368693)*x[1]
        s[5,1,1]=(0.63415870202)*x[0]+(1.38103994652)*x[1]
        ref[0,0,0,0]=(0.116652834373)
        ref[0,0,0,1]=(1.4681512344)
        ref[0,0,1,0]=(1.89418390988)
        ref[0,0,1,1]=(-0.548065006619)
        ref[0,1,0,0]=(-0.07602392145)
        ref[0,1,0,1]=(0.0796515016952)
        ref[0,1,1,0]=(1.50333606195)
        ref[0,1,1,1]=(-0.211673198103)
        ref[1,0,0,0]=(0.172419295955)
        ref[1,0,0,1]=(-0.586603468045)
        ref[1,0,1,0]=(0.212952657549)
        ref[1,0,1,1]=(-0.100672655087)
        ref[1,1,0,0]=(-1.11939406352)
        ref[1,1,0,1]=(-0.589882254225)
        ref[1,1,1,0]=(0.318138253303)
        ref[1,1,1,1]=(0.409850277983)
        ref[2,0,0,0]=(-0.0669759465557)
        ref[2,0,0,1]=(0.525669761096)
        ref[2,0,1,0]=(0.349019390883)
        ref[2,0,1,1]=(0.154604501057)
        ref[2,1,0,0]=(1.95277106546)
        ref[2,1,0,1]=(0.495701766483)
        ref[2,1,1,0]=(-0.0986210008015)
        ref[2,1,1,1]=(1.12636829398)
        ref[3,0,0,0]=(0.971837535878)
        ref[3,0,0,1]=(-0.738253595188)
        ref[3,0,1,0]=(0.0314513203112)
        ref[3,0,1,1]=(-1.16646217783)
        ref[3,1,0,0]=(0.485751127243)
        ref[3,1,0,1]=(-0.340178033471)
        ref[3,1,1,0]=(0.408742867205)
        ref[3,1,1,1]=(0.151827789426)
        ref[4,0,0,0]=(-0.0134678971538)
        ref[4,0,0,1]=(0.884924133141)
        ref[4,0,1,0]=(-0.66899222691)
        ref[4,0,1,1]=(-0.959185584071)
        ref[4,1,0,0]=(1.05311767886)
        ref[4,1,0,1]=(0.64093792324)
        ref[4,1,1,0]=(0.562875594668)
        ref[4,1,1,1]=(0.539525481691)
        ref[5,0,0,0]=(0.310908751279)
        ref[5,0,0,1]=(0.497074267163)
        ref[5,0,1,0]=(-0.804186513465)
        ref[5,0,1,1]=(0.329762438933)
        ref[5,1,0,0]=(-0.946808260459)
        ref[5,1,0,1]=(-0.00260213368693)
        ref[5,1,1,0]=(0.63415870202)
        ref[5,1,1,1]=(1.38103994652)
      else:
        s[0,0,0]=(-0.0840260391925)*x[0]+(0.385117859595)*x[1]+(1.36354425949)*x[2]
        s[0,0,1]=(0.682480367755)*x[0]+(-0.658980348414)*x[1]+(-0.129734588977)*x[2]
        s[0,1,0]=(-0.213790105123)*x[0]+(-0.774270821015)*x[1]+(0.160024784274)*x[2]
        s[0,1,1]=(-1.02270895286)*x[0]+(1.02607310825)*x[1]+(0.35056114816)*x[2]
        s[1,0,0]=(0.201792116067)*x[0]+(0.223564282304)*x[1]+(0.986785519545)*x[2]
        s[1,0,1]=(0.250819712775)*x[0]+(0.287643279417)*x[1]+(0.18295590528)*x[2]
        s[1,1,0]=(-1.47569225207)*x[0]+(-0.886933933801)*x[1]+(0.525380474082)*x[2]
        s[1,1,1]=(0.680401800239)*x[0]+(1.36155081028)*x[1]+(-1.76198171789)*x[2]
        s[2,0,0]=(-0.784439188011)*x[0]+(-0.797291895631)*x[1]+(-0.425550387518)*x[2]
        s[2,0,1]=(1.62016996989)*x[0]+(1.02824549576)*x[1]+(-0.157995046652)*x[2]
        s[2,1,0]=(-0.437412344413)*x[0]+(0.97005624264)*x[1]+(-0.321998550341)*x[2]
        s[2,1,1]=(0.473879660993)*x[0]+(0.483144823024)*x[1]+(0.168449443711)*x[2]
        s[3,0,0]=(-1.59822104736)*x[0]+(-1.24012649834)*x[1]+(0.695408081789)*x[2]
        s[3,0,1]=(0.827901715486)*x[0]+(-0.230621260402)*x[1]+(-0.789726238476)*x[2]
        s[3,1,0]=(-0.279064257268)*x[0]+(-1.35315417734)*x[1]+(0.593090655629)*x[2]
        s[3,1,1]=(0.387685237641)*x[0]+(1.60258880823)*x[1]+(-1.10088045432)*x[2]
        s[4,0,0]=(0.771374132686)*x[0]+(0.530393756307)*x[1]+(-0.366697715551)*x[2]
        s[4,0,1]=(-0.182374799863)*x[0]+(-0.125579433837)*x[1]+(-1.2028145445)*x[2]
        s[4,1,0]=(-0.208766654808)*x[0]+(0.645251941741)*x[1]+(-0.534068338041)*x[2]
        s[4,1,1]=(-0.064786779421)*x[0]+(0.0354509511326)*x[1]+(-1.32632049137)*x[2]
        s[5,0,0]=(0.833438445425)*x[0]+(0.770186199496)*x[1]+(0.112178380686)*x[2]
        s[5,0,1]=(1.36614714953)*x[0]+(-0.367243359207)*x[1]+(0.812867174908)*x[2]
        s[5,1,0]=(0.0359079603828)*x[0]+(0.622094474303)*x[1]+(-0.33137187739)*x[2]
        s[5,1,1]=(-0.479714378899)*x[0]+(-0.959098499695)*x[1]+(0.843889761479)*x[2]
        ref[0,0,0,0]=(-0.0840260391925)
        ref[0,0,0,1]=(0.385117859595)
        ref[0,0,0,2]=(1.36354425949)
        ref[0,0,1,0]=(0.682480367755)
        ref[0,0,1,1]=(-0.658980348414)
        ref[0,0,1,2]=(-0.129734588977)
        ref[0,1,0,0]=(-0.213790105123)
        ref[0,1,0,1]=(-0.774270821015)
        ref[0,1,0,2]=(0.160024784274)
        ref[0,1,1,0]=(-1.02270895286)
        ref[0,1,1,1]=(1.02607310825)
        ref[0,1,1,2]=(0.35056114816)
        ref[1,0,0,0]=(0.201792116067)
        ref[1,0,0,1]=(0.223564282304)
        ref[1,0,0,2]=(0.986785519545)
        ref[1,0,1,0]=(0.250819712775)
        ref[1,0,1,1]=(0.287643279417)
        ref[1,0,1,2]=(0.18295590528)
        ref[1,1,0,0]=(-1.47569225207)
        ref[1,1,0,1]=(-0.886933933801)
        ref[1,1,0,2]=(0.525380474082)
        ref[1,1,1,0]=(0.680401800239)
        ref[1,1,1,1]=(1.36155081028)
        ref[1,1,1,2]=(-1.76198171789)
        ref[2,0,0,0]=(-0.784439188011)
        ref[2,0,0,1]=(-0.797291895631)
        ref[2,0,0,2]=(-0.425550387518)
        ref[2,0,1,0]=(1.62016996989)
        ref[2,0,1,1]=(1.02824549576)
        ref[2,0,1,2]=(-0.157995046652)
        ref[2,1,0,0]=(-0.437412344413)
        ref[2,1,0,1]=(0.97005624264)
        ref[2,1,0,2]=(-0.321998550341)
        ref[2,1,1,0]=(0.473879660993)
        ref[2,1,1,1]=(0.483144823024)
        ref[2,1,1,2]=(0.168449443711)
        ref[3,0,0,0]=(-1.59822104736)
        ref[3,0,0,1]=(-1.24012649834)
        ref[3,0,0,2]=(0.695408081789)
        ref[3,0,1,0]=(0.827901715486)
        ref[3,0,1,1]=(-0.230621260402)
        ref[3,0,1,2]=(-0.789726238476)
        ref[3,1,0,0]=(-0.279064257268)
        ref[3,1,0,1]=(-1.35315417734)
        ref[3,1,0,2]=(0.593090655629)
        ref[3,1,1,0]=(0.387685237641)
        ref[3,1,1,1]=(1.60258880823)
        ref[3,1,1,2]=(-1.10088045432)
        ref[4,0,0,0]=(0.771374132686)
        ref[4,0,0,1]=(0.530393756307)
        ref[4,0,0,2]=(-0.366697715551)
        ref[4,0,1,0]=(-0.182374799863)
        ref[4,0,1,1]=(-0.125579433837)
        ref[4,0,1,2]=(-1.2028145445)
        ref[4,1,0,0]=(-0.208766654808)
        ref[4,1,0,1]=(0.645251941741)
        ref[4,1,0,2]=(-0.534068338041)
        ref[4,1,1,0]=(-0.064786779421)
        ref[4,1,1,1]=(0.0354509511326)
        ref[4,1,1,2]=(-1.32632049137)
        ref[5,0,0,0]=(0.833438445425)
        ref[5,0,0,1]=(0.770186199496)
        ref[5,0,0,2]=(0.112178380686)
        ref[5,0,1,0]=(1.36614714953)
        ref[5,0,1,1]=(-0.367243359207)
        ref[5,0,1,2]=(0.812867174908)
        ref[5,1,0,0]=(0.0359079603828)
        ref[5,1,0,1]=(0.622094474303)
        ref[5,1,0,2]=(-0.33137187739)
        ref[5,1,1,0]=(-0.479714378899)
        ref[5,1,1,1]=(-0.959098499695)
        ref[5,1,1,2]=(0.843889761479)
      res=grad(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(6, 2, 2)+(dim,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.350515208335)*x[0]**o+(-0.604596801095)*x[0]+(0.145267491062)*x[1]**o+(-0.422271555369)*x[1]
        ref=(-0.350515208335)*x_ref[0]**o+(-0.604596801095)*x_ref[0]+(0.145267491062)*x_ref[1]**o+(-0.422271555369)*x_ref[1]
      else:
        arg=(-0.162304501731)*x[0]**o+(0.342262454165)*x[0]+(0.693988751897)*x[1]**o+(-0.315225717754)*x[1]+(-0.938405302425)*x[2]**o+(-0.0929881908885)*x[2]
        ref=(-0.162304501731)*x_ref[0]**o+(0.342262454165)*x_ref[0]+(0.693988751897)*x_ref[1]**o+(-0.315225717754)*x_ref[1]+(-0.938405302425)*x_ref[2]**o+(-0.0929881908885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.104854565192)*x[0]**o+(-0.596688162793)*x[0]+(-0.0286471957779)*x[1]**o+(-0.157282508059)*x[1]
        arg[1]=(0.619817759027)*x[0]**o+(-0.832815965127)*x[0]+(-0.638126497014)*x[1]**o+(0.771271149341)*x[1]
        ref[0]=(0.104854565192)*x_ref[0]**o+(-0.596688162793)*x_ref[0]+(-0.0286471957779)*x_ref[1]**o+(-0.157282508059)*x_ref[1]
        ref[1]=(0.619817759027)*x_ref[0]**o+(-0.832815965127)*x_ref[0]+(-0.638126497014)*x_ref[1]**o+(0.771271149341)*x_ref[1]
      else:
        arg[0]=(-0.358142932888)*x[0]**o+(0.758685606027)*x[0]+(-0.0716050574096)*x[1]**o+(-0.723587074286)*x[1]+(0.387240173671)*x[2]**o+(0.544679811421)*x[2]
        arg[1]=(0.864154242034)*x[0]**o+(-0.0740598679188)*x[0]+(-0.0865611132887)*x[1]**o+(0.0414815431481)*x[1]+(-0.897395775334)*x[2]**o+(-0.580401889318)*x[2]
        ref[0]=(-0.358142932888)*x_ref[0]**o+(0.758685606027)*x_ref[0]+(-0.0716050574096)*x_ref[1]**o+(-0.723587074286)*x_ref[1]+(0.387240173671)*x_ref[2]**o+(0.544679811421)*x_ref[2]
        ref[1]=(0.864154242034)*x_ref[0]**o+(-0.0740598679188)*x_ref[0]+(-0.0865611132887)*x_ref[1]**o+(0.0414815431481)*x_ref[1]+(-0.897395775334)*x_ref[2]**o+(-0.580401889318)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.0571385613076)*x[0]**o+(-0.296048600622)*x[0]+(0.300914212927)*x[1]**o+(0.81578919689)*x[1]
        arg[0,1]=(0.136796732287)*x[0]**o+(0.169384864236)*x[0]+(-0.874585047899)*x[1]**o+(-0.67613629157)*x[1]
        arg[0,2]=(-0.947048556597)*x[0]**o+(-0.807647782941)*x[0]+(-0.53233790031)*x[1]**o+(0.618861393596)*x[1]
        arg[0,3]=(-0.533009370674)*x[0]**o+(-0.732258917532)*x[0]+(0.735058784793)*x[1]**o+(0.313284184128)*x[1]
        arg[0,4]=(0.678291805831)*x[0]**o+(0.892272189132)*x[0]+(-0.902740106267)*x[1]**o+(0.902593764933)*x[1]
        arg[1,0]=(0.101446968367)*x[0]**o+(0.758270058166)*x[0]+(-0.511416064913)*x[1]**o+(-0.464147847068)*x[1]
        arg[1,1]=(-0.0131419066862)*x[0]**o+(-0.942460391354)*x[0]+(0.479955107486)*x[1]**o+(-0.311304181244)*x[1]
        arg[1,2]=(0.623958732361)*x[0]**o+(-0.313196338837)*x[0]+(0.413428996617)*x[1]**o+(0.348327566672)*x[1]
        arg[1,3]=(-0.8413636876)*x[0]**o+(-0.572573098032)*x[0]+(-0.678663954356)*x[1]**o+(-0.419716403314)*x[1]
        arg[1,4]=(-0.580466216486)*x[0]**o+(0.384226238557)*x[0]+(-0.732722613215)*x[1]**o+(-0.0238970013828)*x[1]
        arg[2,0]=(-0.248986933717)*x[0]**o+(0.264019654506)*x[0]+(0.0314811616292)*x[1]**o+(0.631481081422)*x[1]
        arg[2,1]=(0.643576426484)*x[0]**o+(-0.313599735242)*x[0]+(0.908780876791)*x[1]**o+(-0.698972268201)*x[1]
        arg[2,2]=(-0.623208321993)*x[0]**o+(0.649637800083)*x[0]+(-0.878484988208)*x[1]**o+(0.368793479691)*x[1]
        arg[2,3]=(0.239333407828)*x[0]**o+(-0.293558915555)*x[0]+(0.271630238733)*x[1]**o+(0.741304510722)*x[1]
        arg[2,4]=(0.175914942104)*x[0]**o+(0.336116863935)*x[0]+(-0.799168065807)*x[1]**o+(0.0264962830729)*x[1]
        arg[3,0]=(0.777665875019)*x[0]**o+(-0.13232137001)*x[0]+(0.682046389381)*x[1]**o+(0.589069623109)*x[1]
        arg[3,1]=(0.7607621575)*x[0]**o+(-0.81285460551)*x[0]+(0.0523010659976)*x[1]**o+(-0.913765717778)*x[1]
        arg[3,2]=(0.822107753118)*x[0]**o+(0.508869893322)*x[0]+(0.679583348422)*x[1]**o+(-0.597814039932)*x[1]
        arg[3,3]=(-0.77699176705)*x[0]**o+(-0.267758166676)*x[0]+(0.661073127871)*x[1]**o+(-0.41434545147)*x[1]
        arg[3,4]=(0.223013380161)*x[0]**o+(-0.558077040245)*x[0]+(-0.450831764747)*x[1]**o+(-0.0897765635191)*x[1]
        ref[0,0]=(-0.0571385613076)*x_ref[0]**o+(-0.296048600622)*x_ref[0]+(0.300914212927)*x_ref[1]**o+(0.81578919689)*x_ref[1]
        ref[0,1]=(0.136796732287)*x_ref[0]**o+(0.169384864236)*x_ref[0]+(-0.874585047899)*x_ref[1]**o+(-0.67613629157)*x_ref[1]
        ref[0,2]=(-0.947048556597)*x_ref[0]**o+(-0.807647782941)*x_ref[0]+(-0.53233790031)*x_ref[1]**o+(0.618861393596)*x_ref[1]
        ref[0,3]=(-0.533009370674)*x_ref[0]**o+(-0.732258917532)*x_ref[0]+(0.735058784793)*x_ref[1]**o+(0.313284184128)*x_ref[1]
        ref[0,4]=(0.678291805831)*x_ref[0]**o+(0.892272189132)*x_ref[0]+(-0.902740106267)*x_ref[1]**o+(0.902593764933)*x_ref[1]
        ref[1,0]=(0.101446968367)*x_ref[0]**o+(0.758270058166)*x_ref[0]+(-0.511416064913)*x_ref[1]**o+(-0.464147847068)*x_ref[1]
        ref[1,1]=(-0.0131419066862)*x_ref[0]**o+(-0.942460391354)*x_ref[0]+(0.479955107486)*x_ref[1]**o+(-0.311304181244)*x_ref[1]
        ref[1,2]=(0.623958732361)*x_ref[0]**o+(-0.313196338837)*x_ref[0]+(0.413428996617)*x_ref[1]**o+(0.348327566672)*x_ref[1]
        ref[1,3]=(-0.8413636876)*x_ref[0]**o+(-0.572573098032)*x_ref[0]+(-0.678663954356)*x_ref[1]**o+(-0.419716403314)*x_ref[1]
        ref[1,4]=(-0.580466216486)*x_ref[0]**o+(0.384226238557)*x_ref[0]+(-0.732722613215)*x_ref[1]**o+(-0.0238970013828)*x_ref[1]
        ref[2,0]=(-0.248986933717)*x_ref[0]**o+(0.264019654506)*x_ref[0]+(0.0314811616292)*x_ref[1]**o+(0.631481081422)*x_ref[1]
        ref[2,1]=(0.643576426484)*x_ref[0]**o+(-0.313599735242)*x_ref[0]+(0.908780876791)*x_ref[1]**o+(-0.698972268201)*x_ref[1]
        ref[2,2]=(-0.623208321993)*x_ref[0]**o+(0.649637800083)*x_ref[0]+(-0.878484988208)*x_ref[1]**o+(0.368793479691)*x_ref[1]
        ref[2,3]=(0.239333407828)*x_ref[0]**o+(-0.293558915555)*x_ref[0]+(0.271630238733)*x_ref[1]**o+(0.741304510722)*x_ref[1]
        ref[2,4]=(0.175914942104)*x_ref[0]**o+(0.336116863935)*x_ref[0]+(-0.799168065807)*x_ref[1]**o+(0.0264962830729)*x_ref[1]
        ref[3,0]=(0.777665875019)*x_ref[0]**o+(-0.13232137001)*x_ref[0]+(0.682046389381)*x_ref[1]**o+(0.589069623109)*x_ref[1]
        ref[3,1]=(0.7607621575)*x_ref[0]**o+(-0.81285460551)*x_ref[0]+(0.0523010659976)*x_ref[1]**o+(-0.913765717778)*x_ref[1]
        ref[3,2]=(0.822107753118)*x_ref[0]**o+(0.508869893322)*x_ref[0]+(0.679583348422)*x_ref[1]**o+(-0.597814039932)*x_ref[1]
        ref[3,3]=(-0.77699176705)*x_ref[0]**o+(-0.267758166676)*x_ref[0]+(0.661073127871)*x_ref[1]**o+(-0.41434545147)*x_ref[1]
        ref[3,4]=(0.223013380161)*x_ref[0]**o+(-0.558077040245)*x_ref[0]+(-0.450831764747)*x_ref[1]**o+(-0.0897765635191)*x_ref[1]
      else:
        arg[0,0]=(0.487697656095)*x[0]**o+(0.362509734966)*x[0]+(-0.40438484582)*x[1]**o+(-0.358984481681)*x[1]+(0.157670938098)*x[2]**o+(-0.631406375093)*x[2]
        arg[0,1]=(-0.00251154956677)*x[0]**o+(0.671447769464)*x[0]+(0.535541351526)*x[1]**o+(0.433508112485)*x[1]+(-0.897668718638)*x[2]**o+(0.141331586102)*x[2]
        arg[0,2]=(-0.964778505658)*x[0]**o+(0.522815935313)*x[0]+(-0.750318077186)*x[1]**o+(0.0595240411196)*x[1]+(-0.384021192511)*x[2]**o+(-0.802529657745)*x[2]
        arg[0,3]=(0.264781810221)*x[0]**o+(-0.654277693956)*x[0]+(0.890264421741)*x[1]**o+(-0.657743774786)*x[1]+(0.320234835664)*x[2]**o+(0.0637070094165)*x[2]
        arg[0,4]=(0.712229359775)*x[0]**o+(-0.646586719826)*x[0]+(0.876519277606)*x[1]**o+(0.473649630238)*x[1]+(0.332259187685)*x[2]**o+(-0.331489699179)*x[2]
        arg[1,0]=(0.808534274392)*x[0]**o+(-0.296462057802)*x[0]+(0.152536716655)*x[1]**o+(-0.990267302005)*x[1]+(-0.823462709316)*x[2]**o+(-0.51608634537)*x[2]
        arg[1,1]=(-0.745238552081)*x[0]**o+(-0.892086591023)*x[0]+(0.691429994644)*x[1]**o+(-0.52538570831)*x[1]+(-0.668959934842)*x[2]**o+(-0.5178179109)*x[2]
        arg[1,2]=(0.10173204456)*x[0]**o+(-0.853001564748)*x[0]+(0.617252846191)*x[1]**o+(-0.518228673297)*x[1]+(-0.995447484095)*x[2]**o+(0.56679145486)*x[2]
        arg[1,3]=(0.285145125389)*x[0]**o+(0.773362877732)*x[0]+(-0.457364007088)*x[1]**o+(0.226257217659)*x[1]+(0.439832733067)*x[2]**o+(0.872812128578)*x[2]
        arg[1,4]=(-0.581535304023)*x[0]**o+(-0.944599120493)*x[0]+(-0.720009453699)*x[1]**o+(-0.29624047891)*x[1]+(0.966528522407)*x[2]**o+(0.818945905214)*x[2]
        arg[2,0]=(-0.650732348774)*x[0]**o+(0.368909306978)*x[0]+(-0.608327978493)*x[1]**o+(0.301600904054)*x[1]+(-0.747648472883)*x[2]**o+(0.846568601061)*x[2]
        arg[2,1]=(-0.0221803961531)*x[0]**o+(-0.858842969825)*x[0]+(-0.40141447334)*x[1]**o+(0.408164536045)*x[1]+(-0.0131899768237)*x[2]**o+(-0.0983376041104)*x[2]
        arg[2,2]=(0.0855129797683)*x[0]**o+(0.0645200342375)*x[0]+(0.295916169508)*x[1]**o+(0.343169626483)*x[1]+(0.973477621396)*x[2]**o+(0.179505855846)*x[2]
        arg[2,3]=(-0.115714157482)*x[0]**o+(-0.00785628190789)*x[0]+(0.406498238895)*x[1]**o+(-0.954353977843)*x[1]+(0.751544148524)*x[2]**o+(-0.092307218389)*x[2]
        arg[2,4]=(0.807583813259)*x[0]**o+(-0.0150562208564)*x[0]+(0.967290753425)*x[1]**o+(0.960049095802)*x[1]+(0.41935259669)*x[2]**o+(0.878823415691)*x[2]
        arg[3,0]=(-0.162221321908)*x[0]**o+(0.0142404015457)*x[0]+(0.769189546147)*x[1]**o+(-0.251240080116)*x[1]+(0.980441953151)*x[2]**o+(-0.183639594778)*x[2]
        arg[3,1]=(0.270231471303)*x[0]**o+(-0.527471406108)*x[0]+(-0.255144134035)*x[1]**o+(0.268061599615)*x[1]+(-0.396373327005)*x[2]**o+(0.105905036411)*x[2]
        arg[3,2]=(0.5159632953)*x[0]**o+(-0.87397274133)*x[0]+(0.493876004047)*x[1]**o+(0.320996363656)*x[1]+(0.884345148533)*x[2]**o+(0.116262647026)*x[2]
        arg[3,3]=(-0.354053396048)*x[0]**o+(-0.130121845174)*x[0]+(-0.33114435645)*x[1]**o+(-0.73998189126)*x[1]+(0.362417530484)*x[2]**o+(0.676820875084)*x[2]
        arg[3,4]=(-0.53273881792)*x[0]**o+(-0.0624208514424)*x[0]+(0.680486778939)*x[1]**o+(-0.370666453954)*x[1]+(0.294345859395)*x[2]**o+(-0.679169213305)*x[2]
        ref[0,0]=(0.487697656095)*x_ref[0]**o+(0.362509734966)*x_ref[0]+(-0.40438484582)*x_ref[1]**o+(-0.358984481681)*x_ref[1]+(0.157670938098)*x_ref[2]**o+(-0.631406375093)*x_ref[2]
        ref[0,1]=(-0.00251154956677)*x_ref[0]**o+(0.671447769464)*x_ref[0]+(0.535541351526)*x_ref[1]**o+(0.433508112485)*x_ref[1]+(-0.897668718638)*x_ref[2]**o+(0.141331586102)*x_ref[2]
        ref[0,2]=(-0.964778505658)*x_ref[0]**o+(0.522815935313)*x_ref[0]+(-0.750318077186)*x_ref[1]**o+(0.0595240411196)*x_ref[1]+(-0.384021192511)*x_ref[2]**o+(-0.802529657745)*x_ref[2]
        ref[0,3]=(0.264781810221)*x_ref[0]**o+(-0.654277693956)*x_ref[0]+(0.890264421741)*x_ref[1]**o+(-0.657743774786)*x_ref[1]+(0.320234835664)*x_ref[2]**o+(0.0637070094165)*x_ref[2]
        ref[0,4]=(0.712229359775)*x_ref[0]**o+(-0.646586719826)*x_ref[0]+(0.876519277606)*x_ref[1]**o+(0.473649630238)*x_ref[1]+(0.332259187685)*x_ref[2]**o+(-0.331489699179)*x_ref[2]
        ref[1,0]=(0.808534274392)*x_ref[0]**o+(-0.296462057802)*x_ref[0]+(0.152536716655)*x_ref[1]**o+(-0.990267302005)*x_ref[1]+(-0.823462709316)*x_ref[2]**o+(-0.51608634537)*x_ref[2]
        ref[1,1]=(-0.745238552081)*x_ref[0]**o+(-0.892086591023)*x_ref[0]+(0.691429994644)*x_ref[1]**o+(-0.52538570831)*x_ref[1]+(-0.668959934842)*x_ref[2]**o+(-0.5178179109)*x_ref[2]
        ref[1,2]=(0.10173204456)*x_ref[0]**o+(-0.853001564748)*x_ref[0]+(0.617252846191)*x_ref[1]**o+(-0.518228673297)*x_ref[1]+(-0.995447484095)*x_ref[2]**o+(0.56679145486)*x_ref[2]
        ref[1,3]=(0.285145125389)*x_ref[0]**o+(0.773362877732)*x_ref[0]+(-0.457364007088)*x_ref[1]**o+(0.226257217659)*x_ref[1]+(0.439832733067)*x_ref[2]**o+(0.872812128578)*x_ref[2]
        ref[1,4]=(-0.581535304023)*x_ref[0]**o+(-0.944599120493)*x_ref[0]+(-0.720009453699)*x_ref[1]**o+(-0.29624047891)*x_ref[1]+(0.966528522407)*x_ref[2]**o+(0.818945905214)*x_ref[2]
        ref[2,0]=(-0.650732348774)*x_ref[0]**o+(0.368909306978)*x_ref[0]+(-0.608327978493)*x_ref[1]**o+(0.301600904054)*x_ref[1]+(-0.747648472883)*x_ref[2]**o+(0.846568601061)*x_ref[2]
        ref[2,1]=(-0.0221803961531)*x_ref[0]**o+(-0.858842969825)*x_ref[0]+(-0.40141447334)*x_ref[1]**o+(0.408164536045)*x_ref[1]+(-0.0131899768237)*x_ref[2]**o+(-0.0983376041104)*x_ref[2]
        ref[2,2]=(0.0855129797683)*x_ref[0]**o+(0.0645200342375)*x_ref[0]+(0.295916169508)*x_ref[1]**o+(0.343169626483)*x_ref[1]+(0.973477621396)*x_ref[2]**o+(0.179505855846)*x_ref[2]
        ref[2,3]=(-0.115714157482)*x_ref[0]**o+(-0.00785628190789)*x_ref[0]+(0.406498238895)*x_ref[1]**o+(-0.954353977843)*x_ref[1]+(0.751544148524)*x_ref[2]**o+(-0.092307218389)*x_ref[2]
        ref[2,4]=(0.807583813259)*x_ref[0]**o+(-0.0150562208564)*x_ref[0]+(0.967290753425)*x_ref[1]**o+(0.960049095802)*x_ref[1]+(0.41935259669)*x_ref[2]**o+(0.878823415691)*x_ref[2]
        ref[3,0]=(-0.162221321908)*x_ref[0]**o+(0.0142404015457)*x_ref[0]+(0.769189546147)*x_ref[1]**o+(-0.251240080116)*x_ref[1]+(0.980441953151)*x_ref[2]**o+(-0.183639594778)*x_ref[2]
        ref[3,1]=(0.270231471303)*x_ref[0]**o+(-0.527471406108)*x_ref[0]+(-0.255144134035)*x_ref[1]**o+(0.268061599615)*x_ref[1]+(-0.396373327005)*x_ref[2]**o+(0.105905036411)*x_ref[2]
        ref[3,2]=(0.5159632953)*x_ref[0]**o+(-0.87397274133)*x_ref[0]+(0.493876004047)*x_ref[1]**o+(0.320996363656)*x_ref[1]+(0.884345148533)*x_ref[2]**o+(0.116262647026)*x_ref[2]
        ref[3,3]=(-0.354053396048)*x_ref[0]**o+(-0.130121845174)*x_ref[0]+(-0.33114435645)*x_ref[1]**o+(-0.73998189126)*x_ref[1]+(0.362417530484)*x_ref[2]**o+(0.676820875084)*x_ref[2]
        ref[3,4]=(-0.53273881792)*x_ref[0]**o+(-0.0624208514424)*x_ref[0]+(0.680486778939)*x_ref[1]**o+(-0.370666453954)*x_ref[1]+(0.294345859395)*x_ref[2]**o+(-0.679169213305)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0906971676514)*x[0]**o+(-0.0640252128559)*x[0]+(-0.0108470120234)*x[1]**o+(-0.62531206416)*x[1]
        arg[0,0,1]=(-0.0186703191671)*x[0]**o+(0.256044044276)*x[0]+(-0.420130556042)*x[1]**o+(-0.241203381189)*x[1]
        arg[0,1,0]=(-0.287587258734)*x[0]**o+(-0.152487596827)*x[0]+(-0.554975489568)*x[1]**o+(0.713650618367)*x[1]
        arg[0,1,1]=(0.467597882001)*x[0]**o+(0.617291271522)*x[0]+(0.839893220642)*x[1]**o+(0.477553624935)*x[1]
        arg[1,0,0]=(0.292433268891)*x[0]**o+(0.0420372671524)*x[0]+(-0.342734694042)*x[1]**o+(0.892580432212)*x[1]
        arg[1,0,1]=(0.191399101606)*x[0]**o+(0.413781516804)*x[0]+(-0.346483382584)*x[1]**o+(0.805718474386)*x[1]
        arg[1,1,0]=(0.917644938285)*x[0]**o+(0.60457272962)*x[0]+(-0.0949129141155)*x[1]**o+(0.705971564397)*x[1]
        arg[1,1,1]=(0.327738300229)*x[0]**o+(-0.738951673731)*x[0]+(0.335869509312)*x[1]**o+(0.392167761004)*x[1]
        arg[2,0,0]=(0.460678533261)*x[0]**o+(-0.10665029868)*x[0]+(0.658148262201)*x[1]**o+(-0.75975084995)*x[1]
        arg[2,0,1]=(0.307057582515)*x[0]**o+(0.0755776536138)*x[0]+(0.866743758656)*x[1]**o+(-0.463537240812)*x[1]
        arg[2,1,0]=(-0.470241867396)*x[0]**o+(-0.0316070952691)*x[0]+(0.877786692055)*x[1]**o+(-0.750713976717)*x[1]
        arg[2,1,1]=(0.755903505788)*x[0]**o+(0.398472877018)*x[0]+(-0.372159129627)*x[1]**o+(0.170870288116)*x[1]
        arg[3,0,0]=(0.0310115714805)*x[0]**o+(0.822261427932)*x[0]+(-0.0979068243755)*x[1]**o+(0.61848414895)*x[1]
        arg[3,0,1]=(-0.442666645023)*x[0]**o+(0.759423291727)*x[0]+(0.771826722418)*x[1]**o+(-0.342283533203)*x[1]
        arg[3,1,0]=(0.292608821995)*x[0]**o+(0.301553824784)*x[0]+(0.445054135435)*x[1]**o+(0.990249711651)*x[1]
        arg[3,1,1]=(0.835714722542)*x[0]**o+(0.888569910076)*x[0]+(0.76442321847)*x[1]**o+(-0.272880995105)*x[1]
        arg[4,0,0]=(0.434738367563)*x[0]**o+(-0.801063604881)*x[0]+(-0.0152782263323)*x[1]**o+(-0.309467614762)*x[1]
        arg[4,0,1]=(-0.647140151015)*x[0]**o+(-0.0810020795911)*x[0]+(0.356402263616)*x[1]**o+(0.861692124435)*x[1]
        arg[4,1,0]=(-0.958501155541)*x[0]**o+(0.611720259342)*x[0]+(-0.631450146062)*x[1]**o+(-0.512668667768)*x[1]
        arg[4,1,1]=(0.771988547023)*x[0]**o+(0.466474334634)*x[0]+(-0.380283529607)*x[1]**o+(-0.619497160805)*x[1]
        arg[5,0,0]=(0.427386565443)*x[0]**o+(-0.178637075872)*x[0]+(0.81988184595)*x[1]**o+(-0.407798714033)*x[1]
        arg[5,0,1]=(-0.850161084571)*x[0]**o+(0.183440360862)*x[0]+(-0.83848298192)*x[1]**o+(-0.222248160259)*x[1]
        arg[5,1,0]=(-0.930623680879)*x[0]**o+(0.694991147173)*x[0]+(-0.299904374239)*x[1]**o+(0.552324198856)*x[1]
        arg[5,1,1]=(0.740715083219)*x[0]**o+(0.438768785937)*x[0]+(0.977523774483)*x[1]**o+(-0.668723413125)*x[1]
        ref[0,0,0]=(-0.0906971676514)*x_ref[0]**o+(-0.0640252128559)*x_ref[0]+(-0.0108470120234)*x_ref[1]**o+(-0.62531206416)*x_ref[1]
        ref[0,0,1]=(-0.0186703191671)*x_ref[0]**o+(0.256044044276)*x_ref[0]+(-0.420130556042)*x_ref[1]**o+(-0.241203381189)*x_ref[1]
        ref[0,1,0]=(-0.287587258734)*x_ref[0]**o+(-0.152487596827)*x_ref[0]+(-0.554975489568)*x_ref[1]**o+(0.713650618367)*x_ref[1]
        ref[0,1,1]=(0.467597882001)*x_ref[0]**o+(0.617291271522)*x_ref[0]+(0.839893220642)*x_ref[1]**o+(0.477553624935)*x_ref[1]
        ref[1,0,0]=(0.292433268891)*x_ref[0]**o+(0.0420372671524)*x_ref[0]+(-0.342734694042)*x_ref[1]**o+(0.892580432212)*x_ref[1]
        ref[1,0,1]=(0.191399101606)*x_ref[0]**o+(0.413781516804)*x_ref[0]+(-0.346483382584)*x_ref[1]**o+(0.805718474386)*x_ref[1]
        ref[1,1,0]=(0.917644938285)*x_ref[0]**o+(0.60457272962)*x_ref[0]+(-0.0949129141155)*x_ref[1]**o+(0.705971564397)*x_ref[1]
        ref[1,1,1]=(0.327738300229)*x_ref[0]**o+(-0.738951673731)*x_ref[0]+(0.335869509312)*x_ref[1]**o+(0.392167761004)*x_ref[1]
        ref[2,0,0]=(0.460678533261)*x_ref[0]**o+(-0.10665029868)*x_ref[0]+(0.658148262201)*x_ref[1]**o+(-0.75975084995)*x_ref[1]
        ref[2,0,1]=(0.307057582515)*x_ref[0]**o+(0.0755776536138)*x_ref[0]+(0.866743758656)*x_ref[1]**o+(-0.463537240812)*x_ref[1]
        ref[2,1,0]=(-0.470241867396)*x_ref[0]**o+(-0.0316070952691)*x_ref[0]+(0.877786692055)*x_ref[1]**o+(-0.750713976717)*x_ref[1]
        ref[2,1,1]=(0.755903505788)*x_ref[0]**o+(0.398472877018)*x_ref[0]+(-0.372159129627)*x_ref[1]**o+(0.170870288116)*x_ref[1]
        ref[3,0,0]=(0.0310115714805)*x_ref[0]**o+(0.822261427932)*x_ref[0]+(-0.0979068243755)*x_ref[1]**o+(0.61848414895)*x_ref[1]
        ref[3,0,1]=(-0.442666645023)*x_ref[0]**o+(0.759423291727)*x_ref[0]+(0.771826722418)*x_ref[1]**o+(-0.342283533203)*x_ref[1]
        ref[3,1,0]=(0.292608821995)*x_ref[0]**o+(0.301553824784)*x_ref[0]+(0.445054135435)*x_ref[1]**o+(0.990249711651)*x_ref[1]
        ref[3,1,1]=(0.835714722542)*x_ref[0]**o+(0.888569910076)*x_ref[0]+(0.76442321847)*x_ref[1]**o+(-0.272880995105)*x_ref[1]
        ref[4,0,0]=(0.434738367563)*x_ref[0]**o+(-0.801063604881)*x_ref[0]+(-0.0152782263323)*x_ref[1]**o+(-0.309467614762)*x_ref[1]
        ref[4,0,1]=(-0.647140151015)*x_ref[0]**o+(-0.0810020795911)*x_ref[0]+(0.356402263616)*x_ref[1]**o+(0.861692124435)*x_ref[1]
        ref[4,1,0]=(-0.958501155541)*x_ref[0]**o+(0.611720259342)*x_ref[0]+(-0.631450146062)*x_ref[1]**o+(-0.512668667768)*x_ref[1]
        ref[4,1,1]=(0.771988547023)*x_ref[0]**o+(0.466474334634)*x_ref[0]+(-0.380283529607)*x_ref[1]**o+(-0.619497160805)*x_ref[1]
        ref[5,0,0]=(0.427386565443)*x_ref[0]**o+(-0.178637075872)*x_ref[0]+(0.81988184595)*x_ref[1]**o+(-0.407798714033)*x_ref[1]
        ref[5,0,1]=(-0.850161084571)*x_ref[0]**o+(0.183440360862)*x_ref[0]+(-0.83848298192)*x_ref[1]**o+(-0.222248160259)*x_ref[1]
        ref[5,1,0]=(-0.930623680879)*x_ref[0]**o+(0.694991147173)*x_ref[0]+(-0.299904374239)*x_ref[1]**o+(0.552324198856)*x_ref[1]
        ref[5,1,1]=(0.740715083219)*x_ref[0]**o+(0.438768785937)*x_ref[0]+(0.977523774483)*x_ref[1]**o+(-0.668723413125)*x_ref[1]
      else:
        arg[0,0,0]=(0.0408197213005)*x[0]**o+(-0.992920054273)*x[0]+(0.122468233435)*x[1]**o+(-0.435942398199)*x[1]+(0.822235014621)*x[2]**o+(0.425925287623)*x[2]
        arg[0,0,1]=(0.474354220259)*x[0]**o+(0.287892840646)*x[0]+(-0.519818501933)*x[1]**o+(0.412301993774)*x[1]+(-0.242320576454)*x[2]**o+(0.758556860596)*x[2]
        arg[0,1,0]=(0.588735088154)*x[0]**o+(-0.948279493327)*x[0]+(0.194404465814)*x[1]**o+(-0.00761227943336)*x[1]+(-0.117173968067)*x[2]**o+(-0.228746139749)*x[2]
        arg[0,1,1]=(0.166448081217)*x[0]**o+(-0.228754542424)*x[0]+(-0.87739172573)*x[1]**o+(-0.592282861607)*x[1]+(-0.836732991313)*x[2]**o+(-0.20226535214)*x[2]
        arg[1,0,0]=(0.2683388552)*x[0]**o+(-0.733516131653)*x[0]+(-0.900887340495)*x[1]**o+(-0.614908158304)*x[1]+(0.422189467695)*x[2]**o+(0.642276887435)*x[2]
        arg[1,0,1]=(0.965294152134)*x[0]**o+(-0.326731862045)*x[0]+(-0.797754723728)*x[1]**o+(-0.678993698958)*x[1]+(0.256815891879)*x[2]**o+(0.48610716642)*x[2]
        arg[1,1,0]=(0.71383480118)*x[0]**o+(-0.949206151211)*x[0]+(0.40802633055)*x[1]**o+(-0.512457451544)*x[1]+(0.83647020826)*x[2]**o+(0.330550196742)*x[2]
        arg[1,1,1]=(-0.108602934331)*x[0]**o+(-0.585532075482)*x[0]+(0.302790940309)*x[1]**o+(-0.201628116557)*x[1]+(0.523440894937)*x[2]**o+(-0.421867218852)*x[2]
        arg[2,0,0]=(-0.911096130025)*x[0]**o+(-0.944970520542)*x[0]+(0.597179186191)*x[1]**o+(0.530539729209)*x[1]+(-0.538576718159)*x[2]**o+(-0.908863106348)*x[2]
        arg[2,0,1]=(0.828174522282)*x[0]**o+(-0.132534369339)*x[0]+(0.851849613713)*x[1]**o+(0.191494183049)*x[1]+(-0.953488097214)*x[2]**o+(0.423553741843)*x[2]
        arg[2,1,0]=(0.458631350156)*x[0]**o+(0.21755431391)*x[0]+(-0.128541259216)*x[1]**o+(-0.172831640065)*x[1]+(-0.62567145453)*x[2]**o+(-0.345626622399)*x[2]
        arg[2,1,1]=(0.49890508395)*x[0]**o+(-0.395084188812)*x[0]+(-0.653164498387)*x[1]**o+(-0.656416446654)*x[1]+(-0.844855683273)*x[2]**o+(-0.813755115861)*x[2]
        arg[3,0,0]=(-0.252710828698)*x[0]**o+(-0.0100429888666)*x[0]+(0.952930817098)*x[1]**o+(-0.950777038172)*x[1]+(0.244588721191)*x[2]**o+(-0.400442332557)*x[2]
        arg[3,0,1]=(-0.505331423555)*x[0]**o+(-0.535163478725)*x[0]+(0.924129836434)*x[1]**o+(-0.285813201016)*x[1]+(0.225037301542)*x[2]**o+(-0.331424134564)*x[2]
        arg[3,1,0]=(-0.556397106971)*x[0]**o+(0.26005371138)*x[0]+(0.902941320217)*x[1]**o+(0.70375648044)*x[1]+(-0.559371940837)*x[2]**o+(-0.628051748121)*x[2]
        arg[3,1,1]=(-0.644440017615)*x[0]**o+(-0.13415915731)*x[0]+(-0.752973804206)*x[1]**o+(-0.430304389409)*x[1]+(0.892979644749)*x[2]**o+(0.0618885081881)*x[2]
        arg[4,0,0]=(0.388902949971)*x[0]**o+(-0.413207160819)*x[0]+(-0.597892565551)*x[1]**o+(-0.433017605538)*x[1]+(0.369120396114)*x[2]**o+(-0.714185662873)*x[2]
        arg[4,0,1]=(0.140621693105)*x[0]**o+(0.910612147355)*x[0]+(0.404844175588)*x[1]**o+(0.963351416048)*x[1]+(0.545142156468)*x[2]**o+(-0.513154196146)*x[2]
        arg[4,1,0]=(-0.6045448854)*x[0]**o+(0.364599037279)*x[0]+(-0.165333666972)*x[1]**o+(-0.454715984949)*x[1]+(-0.123749300084)*x[2]**o+(0.23666517892)*x[2]
        arg[4,1,1]=(-0.701819332028)*x[0]**o+(-0.945269094128)*x[0]+(0.313745376139)*x[1]**o+(-0.0508160648325)*x[1]+(0.273150632944)*x[2]**o+(0.363230489878)*x[2]
        arg[5,0,0]=(-0.899477778287)*x[0]**o+(0.408300340496)*x[0]+(-0.501090863132)*x[1]**o+(-0.142172524799)*x[1]+(0.840930927028)*x[2]**o+(0.0786174984564)*x[2]
        arg[5,0,1]=(0.816863720765)*x[0]**o+(0.421613919321)*x[0]+(0.619636513942)*x[1]**o+(-0.867504595488)*x[1]+(-0.377345269009)*x[2]**o+(-0.884630348781)*x[2]
        arg[5,1,0]=(0.972307172208)*x[0]**o+(0.704673083035)*x[0]+(0.546566839332)*x[1]**o+(-0.303145527888)*x[1]+(-0.203985539635)*x[2]**o+(0.984097584463)*x[2]
        arg[5,1,1]=(0.67971485539)*x[0]**o+(0.492284313707)*x[0]+(0.139113989973)*x[1]**o+(-0.981414708174)*x[1]+(0.585926444541)*x[2]**o+(-0.790952419012)*x[2]
        ref[0,0,0]=(0.0408197213005)*x_ref[0]**o+(-0.992920054273)*x_ref[0]+(0.122468233435)*x_ref[1]**o+(-0.435942398199)*x_ref[1]+(0.822235014621)*x_ref[2]**o+(0.425925287623)*x_ref[2]
        ref[0,0,1]=(0.474354220259)*x_ref[0]**o+(0.287892840646)*x_ref[0]+(-0.519818501933)*x_ref[1]**o+(0.412301993774)*x_ref[1]+(-0.242320576454)*x_ref[2]**o+(0.758556860596)*x_ref[2]
        ref[0,1,0]=(0.588735088154)*x_ref[0]**o+(-0.948279493327)*x_ref[0]+(0.194404465814)*x_ref[1]**o+(-0.00761227943336)*x_ref[1]+(-0.117173968067)*x_ref[2]**o+(-0.228746139749)*x_ref[2]
        ref[0,1,1]=(0.166448081217)*x_ref[0]**o+(-0.228754542424)*x_ref[0]+(-0.87739172573)*x_ref[1]**o+(-0.592282861607)*x_ref[1]+(-0.836732991313)*x_ref[2]**o+(-0.20226535214)*x_ref[2]
        ref[1,0,0]=(0.2683388552)*x_ref[0]**o+(-0.733516131653)*x_ref[0]+(-0.900887340495)*x_ref[1]**o+(-0.614908158304)*x_ref[1]+(0.422189467695)*x_ref[2]**o+(0.642276887435)*x_ref[2]
        ref[1,0,1]=(0.965294152134)*x_ref[0]**o+(-0.326731862045)*x_ref[0]+(-0.797754723728)*x_ref[1]**o+(-0.678993698958)*x_ref[1]+(0.256815891879)*x_ref[2]**o+(0.48610716642)*x_ref[2]
        ref[1,1,0]=(0.71383480118)*x_ref[0]**o+(-0.949206151211)*x_ref[0]+(0.40802633055)*x_ref[1]**o+(-0.512457451544)*x_ref[1]+(0.83647020826)*x_ref[2]**o+(0.330550196742)*x_ref[2]
        ref[1,1,1]=(-0.108602934331)*x_ref[0]**o+(-0.585532075482)*x_ref[0]+(0.302790940309)*x_ref[1]**o+(-0.201628116557)*x_ref[1]+(0.523440894937)*x_ref[2]**o+(-0.421867218852)*x_ref[2]
        ref[2,0,0]=(-0.911096130025)*x_ref[0]**o+(-0.944970520542)*x_ref[0]+(0.597179186191)*x_ref[1]**o+(0.530539729209)*x_ref[1]+(-0.538576718159)*x_ref[2]**o+(-0.908863106348)*x_ref[2]
        ref[2,0,1]=(0.828174522282)*x_ref[0]**o+(-0.132534369339)*x_ref[0]+(0.851849613713)*x_ref[1]**o+(0.191494183049)*x_ref[1]+(-0.953488097214)*x_ref[2]**o+(0.423553741843)*x_ref[2]
        ref[2,1,0]=(0.458631350156)*x_ref[0]**o+(0.21755431391)*x_ref[0]+(-0.128541259216)*x_ref[1]**o+(-0.172831640065)*x_ref[1]+(-0.62567145453)*x_ref[2]**o+(-0.345626622399)*x_ref[2]
        ref[2,1,1]=(0.49890508395)*x_ref[0]**o+(-0.395084188812)*x_ref[0]+(-0.653164498387)*x_ref[1]**o+(-0.656416446654)*x_ref[1]+(-0.844855683273)*x_ref[2]**o+(-0.813755115861)*x_ref[2]
        ref[3,0,0]=(-0.252710828698)*x_ref[0]**o+(-0.0100429888666)*x_ref[0]+(0.952930817098)*x_ref[1]**o+(-0.950777038172)*x_ref[1]+(0.244588721191)*x_ref[2]**o+(-0.400442332557)*x_ref[2]
        ref[3,0,1]=(-0.505331423555)*x_ref[0]**o+(-0.535163478725)*x_ref[0]+(0.924129836434)*x_ref[1]**o+(-0.285813201016)*x_ref[1]+(0.225037301542)*x_ref[2]**o+(-0.331424134564)*x_ref[2]
        ref[3,1,0]=(-0.556397106971)*x_ref[0]**o+(0.26005371138)*x_ref[0]+(0.902941320217)*x_ref[1]**o+(0.70375648044)*x_ref[1]+(-0.559371940837)*x_ref[2]**o+(-0.628051748121)*x_ref[2]
        ref[3,1,1]=(-0.644440017615)*x_ref[0]**o+(-0.13415915731)*x_ref[0]+(-0.752973804206)*x_ref[1]**o+(-0.430304389409)*x_ref[1]+(0.892979644749)*x_ref[2]**o+(0.0618885081881)*x_ref[2]
        ref[4,0,0]=(0.388902949971)*x_ref[0]**o+(-0.413207160819)*x_ref[0]+(-0.597892565551)*x_ref[1]**o+(-0.433017605538)*x_ref[1]+(0.369120396114)*x_ref[2]**o+(-0.714185662873)*x_ref[2]
        ref[4,0,1]=(0.140621693105)*x_ref[0]**o+(0.910612147355)*x_ref[0]+(0.404844175588)*x_ref[1]**o+(0.963351416048)*x_ref[1]+(0.545142156468)*x_ref[2]**o+(-0.513154196146)*x_ref[2]
        ref[4,1,0]=(-0.6045448854)*x_ref[0]**o+(0.364599037279)*x_ref[0]+(-0.165333666972)*x_ref[1]**o+(-0.454715984949)*x_ref[1]+(-0.123749300084)*x_ref[2]**o+(0.23666517892)*x_ref[2]
        ref[4,1,1]=(-0.701819332028)*x_ref[0]**o+(-0.945269094128)*x_ref[0]+(0.313745376139)*x_ref[1]**o+(-0.0508160648325)*x_ref[1]+(0.273150632944)*x_ref[2]**o+(0.363230489878)*x_ref[2]
        ref[5,0,0]=(-0.899477778287)*x_ref[0]**o+(0.408300340496)*x_ref[0]+(-0.501090863132)*x_ref[1]**o+(-0.142172524799)*x_ref[1]+(0.840930927028)*x_ref[2]**o+(0.0786174984564)*x_ref[2]
        ref[5,0,1]=(0.816863720765)*x_ref[0]**o+(0.421613919321)*x_ref[0]+(0.619636513942)*x_ref[1]**o+(-0.867504595488)*x_ref[1]+(-0.377345269009)*x_ref[2]**o+(-0.884630348781)*x_ref[2]
        ref[5,1,0]=(0.972307172208)*x_ref[0]**o+(0.704673083035)*x_ref[0]+(0.546566839332)*x_ref[1]**o+(-0.303145527888)*x_ref[1]+(-0.203985539635)*x_ref[2]**o+(0.984097584463)*x_ref[2]
        ref[5,1,1]=(0.67971485539)*x_ref[0]**o+(0.492284313707)*x_ref[0]+(0.139113989973)*x_ref[1]**o+(-0.981414708174)*x_ref[1]+(0.585926444541)*x_ref[2]**o+(-0.790952419012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.717741572212)*x[0]**o+(-0.771801794802)*x[0]+(-0.782201793954)*x[1]**o+(0.125253178673)*x[1]
        arg[0,0,0,1]=(-0.483339499341)*x[0]**o+(0.0177945970437)*x[0]+(0.233532060933)*x[1]**o+(-0.465948890703)*x[1]
        arg[0,0,1,0]=(0.140544455071)*x[0]**o+(-0.120145399579)*x[0]+(-0.0732045826929)*x[1]**o+(-0.738165902415)*x[1]
        arg[0,0,1,1]=(0.98057407477)*x[0]**o+(0.346125329427)*x[0]+(0.697386820108)*x[1]**o+(0.402406370862)*x[1]
        arg[0,0,2,0]=(0.347641628204)*x[0]**o+(-0.217938924727)*x[0]+(-0.0572727430966)*x[1]**o+(0.170631149745)*x[1]
        arg[0,0,2,1]=(0.621840702108)*x[0]**o+(-0.109219941839)*x[0]+(0.894052962811)*x[1]**o+(0.423894680808)*x[1]
        arg[0,1,0,0]=(0.0339315069567)*x[0]**o+(-0.655546201957)*x[0]+(-0.271776247317)*x[1]**o+(0.4760527815)*x[1]
        arg[0,1,0,1]=(0.904216237912)*x[0]**o+(0.974869309983)*x[0]+(0.678379481202)*x[1]**o+(0.602534020957)*x[1]
        arg[0,1,1,0]=(-0.98805028173)*x[0]**o+(0.255251214419)*x[0]+(0.659010647624)*x[1]**o+(-0.45360685971)*x[1]
        arg[0,1,1,1]=(0.0884339867328)*x[0]**o+(-0.325188714625)*x[0]+(0.158562643552)*x[1]**o+(-0.788600388567)*x[1]
        arg[0,1,2,0]=(0.976404407593)*x[0]**o+(0.353231650963)*x[0]+(0.546951413477)*x[1]**o+(0.0738098551899)*x[1]
        arg[0,1,2,1]=(-0.812156797654)*x[0]**o+(-0.0180014270879)*x[0]+(0.686011517957)*x[1]**o+(0.148522434365)*x[1]
        arg[0,2,0,0]=(-0.40018516407)*x[0]**o+(-0.0512695503575)*x[0]+(-0.715512784143)*x[1]**o+(0.111064716031)*x[1]
        arg[0,2,0,1]=(-0.963397739298)*x[0]**o+(0.406951519848)*x[0]+(-0.293539028967)*x[1]**o+(-0.549328925146)*x[1]
        arg[0,2,1,0]=(-0.00827402449089)*x[0]**o+(-0.657982352244)*x[0]+(-0.515966831052)*x[1]**o+(0.531853371259)*x[1]
        arg[0,2,1,1]=(0.0379862601491)*x[0]**o+(0.988691147833)*x[0]+(0.820623911484)*x[1]**o+(0.689064076022)*x[1]
        arg[0,2,2,0]=(-0.857638654259)*x[0]**o+(0.664968676908)*x[0]+(-0.228424264597)*x[1]**o+(0.740003679791)*x[1]
        arg[0,2,2,1]=(0.456486069586)*x[0]**o+(-0.510130176146)*x[0]+(-0.189525160745)*x[1]**o+(0.497975981413)*x[1]
        arg[0,3,0,0]=(-0.981633737105)*x[0]**o+(0.25636883881)*x[0]+(-0.0477676975467)*x[1]**o+(0.654475880014)*x[1]
        arg[0,3,0,1]=(-0.0710284929969)*x[0]**o+(0.338590765191)*x[0]+(0.639048913997)*x[1]**o+(-0.24999606762)*x[1]
        arg[0,3,1,0]=(-0.731484482659)*x[0]**o+(-0.808253846389)*x[0]+(-0.911723431862)*x[1]**o+(-0.266029942425)*x[1]
        arg[0,3,1,1]=(0.909037857376)*x[0]**o+(-0.164525039433)*x[0]+(-0.924332370345)*x[1]**o+(0.476479399362)*x[1]
        arg[0,3,2,0]=(0.00534173833653)*x[0]**o+(-0.435160515142)*x[0]+(0.402701511687)*x[1]**o+(0.161631276653)*x[1]
        arg[0,3,2,1]=(-0.595791669027)*x[0]**o+(0.170580206948)*x[0]+(0.468802250813)*x[1]**o+(-0.881472636722)*x[1]
        arg[0,4,0,0]=(-0.0739506349513)*x[0]**o+(0.513385505957)*x[0]+(0.046977740371)*x[1]**o+(0.95763159987)*x[1]
        arg[0,4,0,1]=(-0.286789006223)*x[0]**o+(0.017476293519)*x[0]+(0.637378781092)*x[1]**o+(-0.670201122126)*x[1]
        arg[0,4,1,0]=(0.159607100455)*x[0]**o+(0.715341910227)*x[0]+(0.949546645819)*x[1]**o+(0.362982182775)*x[1]
        arg[0,4,1,1]=(-0.237702024358)*x[0]**o+(-0.520488841454)*x[0]+(0.606333933569)*x[1]**o+(-0.00662757701049)*x[1]
        arg[0,4,2,0]=(-0.415753817901)*x[0]**o+(0.731285464784)*x[0]+(0.15987336144)*x[1]**o+(0.0697706236338)*x[1]
        arg[0,4,2,1]=(0.0467482158427)*x[0]**o+(0.306004722736)*x[0]+(-0.905184725692)*x[1]**o+(-0.158017889366)*x[1]
        arg[1,0,0,0]=(-0.785535082426)*x[0]**o+(-0.189285100452)*x[0]+(-0.766349940057)*x[1]**o+(0.531469308995)*x[1]
        arg[1,0,0,1]=(0.522213196477)*x[0]**o+(-0.661381741437)*x[0]+(0.816089886956)*x[1]**o+(0.139829382305)*x[1]
        arg[1,0,1,0]=(0.0957295570417)*x[0]**o+(-0.105252674536)*x[0]+(-0.569429953382)*x[1]**o+(-0.801934666657)*x[1]
        arg[1,0,1,1]=(-0.557307051058)*x[0]**o+(0.218852576416)*x[0]+(0.430167216464)*x[1]**o+(-0.112285681761)*x[1]
        arg[1,0,2,0]=(0.288472422477)*x[0]**o+(0.294377994763)*x[0]+(-0.809503961511)*x[1]**o+(0.357020226991)*x[1]
        arg[1,0,2,1]=(-0.148113477705)*x[0]**o+(0.164468646294)*x[0]+(0.7977142476)*x[1]**o+(-0.407274933384)*x[1]
        arg[1,1,0,0]=(-0.926374434822)*x[0]**o+(-0.107713463193)*x[0]+(-0.511635723869)*x[1]**o+(0.259319717689)*x[1]
        arg[1,1,0,1]=(-0.18636582243)*x[0]**o+(0.698258973628)*x[0]+(0.234379281397)*x[1]**o+(0.596195875495)*x[1]
        arg[1,1,1,0]=(-0.180395892006)*x[0]**o+(0.631248500084)*x[0]+(0.104751616377)*x[1]**o+(0.204500206653)*x[1]
        arg[1,1,1,1]=(-0.517119564238)*x[0]**o+(-0.16634015634)*x[0]+(0.556785586675)*x[1]**o+(0.42659306392)*x[1]
        arg[1,1,2,0]=(0.163807238958)*x[0]**o+(-0.534619478446)*x[0]+(-0.700353460753)*x[1]**o+(0.121393131648)*x[1]
        arg[1,1,2,1]=(0.964276278961)*x[0]**o+(0.846030751304)*x[0]+(-0.745301894392)*x[1]**o+(0.33821485558)*x[1]
        arg[1,2,0,0]=(0.333173101086)*x[0]**o+(0.774848048743)*x[0]+(-0.659091916357)*x[1]**o+(-0.797247144497)*x[1]
        arg[1,2,0,1]=(0.0581394534285)*x[0]**o+(-0.524771663031)*x[0]+(-0.519465753317)*x[1]**o+(0.71748795226)*x[1]
        arg[1,2,1,0]=(0.224935262833)*x[0]**o+(-0.393399819306)*x[0]+(0.199349075224)*x[1]**o+(-0.808847009412)*x[1]
        arg[1,2,1,1]=(0.0385750795521)*x[0]**o+(-8.09949553151e-05)*x[0]+(0.496860964506)*x[1]**o+(0.0119507407829)*x[1]
        arg[1,2,2,0]=(0.876341556167)*x[0]**o+(0.18082766372)*x[0]+(0.401349871827)*x[1]**o+(-0.543439037303)*x[1]
        arg[1,2,2,1]=(-0.894525004507)*x[0]**o+(-0.516771143695)*x[0]+(0.0410916490263)*x[1]**o+(0.753841618439)*x[1]
        arg[1,3,0,0]=(0.251979970924)*x[0]**o+(0.790276550438)*x[0]+(-0.155708793888)*x[1]**o+(0.28971278336)*x[1]
        arg[1,3,0,1]=(0.258336265436)*x[0]**o+(0.696642875149)*x[0]+(0.0561355915649)*x[1]**o+(0.601839744417)*x[1]
        arg[1,3,1,0]=(-0.978729102785)*x[0]**o+(0.335280613262)*x[0]+(0.89228798232)*x[1]**o+(-0.693050679348)*x[1]
        arg[1,3,1,1]=(-0.885817753431)*x[0]**o+(-0.0641571535708)*x[0]+(-0.649578576181)*x[1]**o+(-0.95773137602)*x[1]
        arg[1,3,2,0]=(-0.333891001461)*x[0]**o+(-0.496064308564)*x[0]+(0.438549603377)*x[1]**o+(0.0772468142672)*x[1]
        arg[1,3,2,1]=(-0.16171788255)*x[0]**o+(-0.939293453222)*x[0]+(-0.284522648546)*x[1]**o+(0.744642328375)*x[1]
        arg[1,4,0,0]=(0.806133262332)*x[0]**o+(-0.0225320206432)*x[0]+(0.125958642606)*x[1]**o+(0.610630249467)*x[1]
        arg[1,4,0,1]=(0.657971870946)*x[0]**o+(0.698894701525)*x[0]+(-0.595952612496)*x[1]**o+(-0.38268841292)*x[1]
        arg[1,4,1,0]=(-0.277274411331)*x[0]**o+(-0.483183004514)*x[0]+(0.77656657009)*x[1]**o+(-0.959290466645)*x[1]
        arg[1,4,1,1]=(-0.96384356305)*x[0]**o+(-0.00383301781476)*x[0]+(0.788010409608)*x[1]**o+(0.135378744316)*x[1]
        arg[1,4,2,0]=(0.986761958017)*x[0]**o+(-0.570276378038)*x[0]+(-0.515784028315)*x[1]**o+(0.487125123439)*x[1]
        arg[1,4,2,1]=(-0.954735054702)*x[0]**o+(-0.0610323074217)*x[0]+(-0.255284859281)*x[1]**o+(0.277795725698)*x[1]
        arg[2,0,0,0]=(-0.708301219324)*x[0]**o+(-0.970166781553)*x[0]+(-0.0836451425855)*x[1]**o+(0.525854202995)*x[1]
        arg[2,0,0,1]=(-0.0995393865183)*x[0]**o+(-0.0156624820324)*x[0]+(0.225529316197)*x[1]**o+(0.142617834056)*x[1]
        arg[2,0,1,0]=(-0.585329524103)*x[0]**o+(0.590285924441)*x[0]+(0.201483726322)*x[1]**o+(-0.735262197689)*x[1]
        arg[2,0,1,1]=(-0.682877182963)*x[0]**o+(-0.162682262799)*x[0]+(-0.890051459843)*x[1]**o+(-0.0188015866765)*x[1]
        arg[2,0,2,0]=(0.868935476294)*x[0]**o+(-0.629891068854)*x[0]+(0.335632865663)*x[1]**o+(-0.955849055716)*x[1]
        arg[2,0,2,1]=(-0.907933228238)*x[0]**o+(-0.105861987974)*x[0]+(-0.131372383042)*x[1]**o+(-0.970114136854)*x[1]
        arg[2,1,0,0]=(-0.275788501814)*x[0]**o+(-0.992861867033)*x[0]+(-0.80676324962)*x[1]**o+(0.269244680224)*x[1]
        arg[2,1,0,1]=(-0.704704138774)*x[0]**o+(-0.371803193009)*x[0]+(-0.0190586405922)*x[1]**o+(-0.0828837624698)*x[1]
        arg[2,1,1,0]=(-0.532645566013)*x[0]**o+(0.571754809868)*x[0]+(0.656737493773)*x[1]**o+(-0.548229591842)*x[1]
        arg[2,1,1,1]=(0.23836921087)*x[0]**o+(0.707475737785)*x[0]+(-0.59660155968)*x[1]**o+(-0.450204951071)*x[1]
        arg[2,1,2,0]=(-0.534027348014)*x[0]**o+(0.332100193727)*x[0]+(-0.864157138567)*x[1]**o+(-0.41833267066)*x[1]
        arg[2,1,2,1]=(0.459878593269)*x[0]**o+(0.446259140228)*x[0]+(0.571242887502)*x[1]**o+(-0.0390525669486)*x[1]
        arg[2,2,0,0]=(0.322052799527)*x[0]**o+(-0.915265209304)*x[0]+(-0.99942007618)*x[1]**o+(0.482294513743)*x[1]
        arg[2,2,0,1]=(0.558489194456)*x[0]**o+(0.00400576554086)*x[0]+(-0.87484330165)*x[1]**o+(-0.664504218977)*x[1]
        arg[2,2,1,0]=(0.669621607461)*x[0]**o+(-0.0450960554319)*x[0]+(-0.221151480246)*x[1]**o+(0.0123621737211)*x[1]
        arg[2,2,1,1]=(0.476743013029)*x[0]**o+(0.11118371822)*x[0]+(-0.202074315677)*x[1]**o+(0.377649433458)*x[1]
        arg[2,2,2,0]=(0.667910543701)*x[0]**o+(0.035703922547)*x[0]+(-0.274733967095)*x[1]**o+(-0.998934720182)*x[1]
        arg[2,2,2,1]=(0.578909660087)*x[0]**o+(0.149900086514)*x[0]+(-0.226597651858)*x[1]**o+(0.169175812287)*x[1]
        arg[2,3,0,0]=(0.158370246586)*x[0]**o+(0.595643112285)*x[0]+(-0.476875564052)*x[1]**o+(0.869752657517)*x[1]
        arg[2,3,0,1]=(0.99718847714)*x[0]**o+(0.581531822297)*x[0]+(-0.607102775564)*x[1]**o+(-0.8317970284)*x[1]
        arg[2,3,1,0]=(-0.27390195745)*x[0]**o+(0.0512853330414)*x[0]+(0.465190666302)*x[1]**o+(0.042533388788)*x[1]
        arg[2,3,1,1]=(-0.183679504962)*x[0]**o+(-0.0522782076121)*x[0]+(0.921423118529)*x[1]**o+(-0.0315080275225)*x[1]
        arg[2,3,2,0]=(0.0421842348778)*x[0]**o+(0.462023049702)*x[0]+(-0.888342606803)*x[1]**o+(-0.956503447884)*x[1]
        arg[2,3,2,1]=(-0.950820288444)*x[0]**o+(0.104271990168)*x[0]+(-0.607212909582)*x[1]**o+(-0.16283931259)*x[1]
        arg[2,4,0,0]=(0.113505933533)*x[0]**o+(0.27873514806)*x[0]+(0.736699443523)*x[1]**o+(-0.951737555693)*x[1]
        arg[2,4,0,1]=(0.264170584849)*x[0]**o+(-0.226649236617)*x[0]+(0.489902180874)*x[1]**o+(-0.191536591762)*x[1]
        arg[2,4,1,0]=(0.712322285985)*x[0]**o+(-0.323395455815)*x[0]+(-0.298356681586)*x[1]**o+(0.699891974541)*x[1]
        arg[2,4,1,1]=(-0.631057763255)*x[0]**o+(-0.489374933445)*x[0]+(-0.021591806874)*x[1]**o+(-0.692800124585)*x[1]
        arg[2,4,2,0]=(0.916068935184)*x[0]**o+(0.33779688858)*x[0]+(0.06300443559)*x[1]**o+(-0.231385203399)*x[1]
        arg[2,4,2,1]=(-0.283485044625)*x[0]**o+(0.548366574442)*x[0]+(-0.543953132872)*x[1]**o+(-0.956662823037)*x[1]
        arg[3,0,0,0]=(0.709349274221)*x[0]**o+(-0.342981009263)*x[0]+(-0.682331761929)*x[1]**o+(0.673169884813)*x[1]
        arg[3,0,0,1]=(0.236249836593)*x[0]**o+(-0.682845054434)*x[0]+(-0.349169988218)*x[1]**o+(0.502684425824)*x[1]
        arg[3,0,1,0]=(-0.561655748793)*x[0]**o+(0.120183772326)*x[0]+(-0.966535325583)*x[1]**o+(-0.0585117264949)*x[1]
        arg[3,0,1,1]=(-0.661030287189)*x[0]**o+(-0.322924086505)*x[0]+(-0.986892059552)*x[1]**o+(-0.778969361365)*x[1]
        arg[3,0,2,0]=(-0.166307107307)*x[0]**o+(0.612931303617)*x[0]+(-0.283925165456)*x[1]**o+(0.5998038814)*x[1]
        arg[3,0,2,1]=(0.205099355916)*x[0]**o+(-0.520970536278)*x[0]+(0.900416833183)*x[1]**o+(0.981826306069)*x[1]
        arg[3,1,0,0]=(-0.875306666443)*x[0]**o+(0.0189358479861)*x[0]+(0.101187857445)*x[1]**o+(-0.873584307534)*x[1]
        arg[3,1,0,1]=(-0.0203656235757)*x[0]**o+(-0.579229777042)*x[0]+(-0.0799345631826)*x[1]**o+(0.896315444069)*x[1]
        arg[3,1,1,0]=(0.576791459982)*x[0]**o+(-0.86541815506)*x[0]+(-0.582764176207)*x[1]**o+(-0.249437395011)*x[1]
        arg[3,1,1,1]=(0.525953680497)*x[0]**o+(-0.222445717402)*x[0]+(-0.87710202063)*x[1]**o+(-0.762921821973)*x[1]
        arg[3,1,2,0]=(-0.883737185908)*x[0]**o+(0.254053241277)*x[0]+(0.559803298059)*x[1]**o+(-0.577376956856)*x[1]
        arg[3,1,2,1]=(-0.125792083387)*x[0]**o+(-0.771549307509)*x[0]+(0.142493480552)*x[1]**o+(0.459996962548)*x[1]
        arg[3,2,0,0]=(0.0531598927049)*x[0]**o+(-0.950343229037)*x[0]+(-0.748240453154)*x[1]**o+(0.850555892342)*x[1]
        arg[3,2,0,1]=(-0.660852152041)*x[0]**o+(-0.763119278604)*x[0]+(-0.811925409663)*x[1]**o+(-0.178765071495)*x[1]
        arg[3,2,1,0]=(-0.189124223305)*x[0]**o+(-0.410392909042)*x[0]+(0.993247167522)*x[1]**o+(0.291451796298)*x[1]
        arg[3,2,1,1]=(0.83496133421)*x[0]**o+(0.695147480265)*x[0]+(0.444420524946)*x[1]**o+(0.107454620114)*x[1]
        arg[3,2,2,0]=(0.94687479313)*x[0]**o+(-0.140702324395)*x[0]+(0.149203502428)*x[1]**o+(-0.338246126345)*x[1]
        arg[3,2,2,1]=(0.147609394709)*x[0]**o+(-0.549219765239)*x[0]+(0.397308993347)*x[1]**o+(0.0402189312685)*x[1]
        arg[3,3,0,0]=(0.27400981409)*x[0]**o+(-0.516367266257)*x[0]+(-0.784095952814)*x[1]**o+(0.32858899009)*x[1]
        arg[3,3,0,1]=(-0.393457382163)*x[0]**o+(0.590406794051)*x[0]+(0.328523918357)*x[1]**o+(0.899537513355)*x[1]
        arg[3,3,1,0]=(-0.851038623366)*x[0]**o+(0.596974864726)*x[0]+(-0.338630687551)*x[1]**o+(0.634961435042)*x[1]
        arg[3,3,1,1]=(0.569166143918)*x[0]**o+(-0.388162282491)*x[0]+(0.764830536733)*x[1]**o+(-0.426737342542)*x[1]
        arg[3,3,2,0]=(0.502221903448)*x[0]**o+(0.940382225897)*x[0]+(0.563444730579)*x[1]**o+(0.023210869151)*x[1]
        arg[3,3,2,1]=(0.061695403852)*x[0]**o+(0.868302188758)*x[0]+(0.0855803325758)*x[1]**o+(-0.24366173638)*x[1]
        arg[3,4,0,0]=(-0.533038182387)*x[0]**o+(0.0222388915737)*x[0]+(0.529037752769)*x[1]**o+(-0.0665320895494)*x[1]
        arg[3,4,0,1]=(0.714374055253)*x[0]**o+(-0.623670743333)*x[0]+(-0.866024792929)*x[1]**o+(-0.0673124896111)*x[1]
        arg[3,4,1,0]=(-0.0523466615045)*x[0]**o+(-0.472563426622)*x[0]+(0.139892673626)*x[1]**o+(0.539655417178)*x[1]
        arg[3,4,1,1]=(-0.473032580514)*x[0]**o+(0.692812093618)*x[0]+(0.664400517742)*x[1]**o+(0.184125423364)*x[1]
        arg[3,4,2,0]=(-0.0340798058725)*x[0]**o+(-0.58900562863)*x[0]+(0.396856384717)*x[1]**o+(-0.559175428019)*x[1]
        arg[3,4,2,1]=(-0.217679376634)*x[0]**o+(-0.90389036667)*x[0]+(0.967372942873)*x[1]**o+(-0.482913619277)*x[1]
        ref[0,0,0,0]=(-0.717741572212)*x_ref[0]**o+(-0.771801794802)*x_ref[0]+(-0.782201793954)*x_ref[1]**o+(0.125253178673)*x_ref[1]
        ref[0,0,0,1]=(-0.483339499341)*x_ref[0]**o+(0.0177945970437)*x_ref[0]+(0.233532060933)*x_ref[1]**o+(-0.465948890703)*x_ref[1]
        ref[0,0,1,0]=(0.140544455071)*x_ref[0]**o+(-0.120145399579)*x_ref[0]+(-0.0732045826929)*x_ref[1]**o+(-0.738165902415)*x_ref[1]
        ref[0,0,1,1]=(0.98057407477)*x_ref[0]**o+(0.346125329427)*x_ref[0]+(0.697386820108)*x_ref[1]**o+(0.402406370862)*x_ref[1]
        ref[0,0,2,0]=(0.347641628204)*x_ref[0]**o+(-0.217938924727)*x_ref[0]+(-0.0572727430966)*x_ref[1]**o+(0.170631149745)*x_ref[1]
        ref[0,0,2,1]=(0.621840702108)*x_ref[0]**o+(-0.109219941839)*x_ref[0]+(0.894052962811)*x_ref[1]**o+(0.423894680808)*x_ref[1]
        ref[0,1,0,0]=(0.0339315069567)*x_ref[0]**o+(-0.655546201957)*x_ref[0]+(-0.271776247317)*x_ref[1]**o+(0.4760527815)*x_ref[1]
        ref[0,1,0,1]=(0.904216237912)*x_ref[0]**o+(0.974869309983)*x_ref[0]+(0.678379481202)*x_ref[1]**o+(0.602534020957)*x_ref[1]
        ref[0,1,1,0]=(-0.98805028173)*x_ref[0]**o+(0.255251214419)*x_ref[0]+(0.659010647624)*x_ref[1]**o+(-0.45360685971)*x_ref[1]
        ref[0,1,1,1]=(0.0884339867328)*x_ref[0]**o+(-0.325188714625)*x_ref[0]+(0.158562643552)*x_ref[1]**o+(-0.788600388567)*x_ref[1]
        ref[0,1,2,0]=(0.976404407593)*x_ref[0]**o+(0.353231650963)*x_ref[0]+(0.546951413477)*x_ref[1]**o+(0.0738098551899)*x_ref[1]
        ref[0,1,2,1]=(-0.812156797654)*x_ref[0]**o+(-0.0180014270879)*x_ref[0]+(0.686011517957)*x_ref[1]**o+(0.148522434365)*x_ref[1]
        ref[0,2,0,0]=(-0.40018516407)*x_ref[0]**o+(-0.0512695503575)*x_ref[0]+(-0.715512784143)*x_ref[1]**o+(0.111064716031)*x_ref[1]
        ref[0,2,0,1]=(-0.963397739298)*x_ref[0]**o+(0.406951519848)*x_ref[0]+(-0.293539028967)*x_ref[1]**o+(-0.549328925146)*x_ref[1]
        ref[0,2,1,0]=(-0.00827402449089)*x_ref[0]**o+(-0.657982352244)*x_ref[0]+(-0.515966831052)*x_ref[1]**o+(0.531853371259)*x_ref[1]
        ref[0,2,1,1]=(0.0379862601491)*x_ref[0]**o+(0.988691147833)*x_ref[0]+(0.820623911484)*x_ref[1]**o+(0.689064076022)*x_ref[1]
        ref[0,2,2,0]=(-0.857638654259)*x_ref[0]**o+(0.664968676908)*x_ref[0]+(-0.228424264597)*x_ref[1]**o+(0.740003679791)*x_ref[1]
        ref[0,2,2,1]=(0.456486069586)*x_ref[0]**o+(-0.510130176146)*x_ref[0]+(-0.189525160745)*x_ref[1]**o+(0.497975981413)*x_ref[1]
        ref[0,3,0,0]=(-0.981633737105)*x_ref[0]**o+(0.25636883881)*x_ref[0]+(-0.0477676975467)*x_ref[1]**o+(0.654475880014)*x_ref[1]
        ref[0,3,0,1]=(-0.0710284929969)*x_ref[0]**o+(0.338590765191)*x_ref[0]+(0.639048913997)*x_ref[1]**o+(-0.24999606762)*x_ref[1]
        ref[0,3,1,0]=(-0.731484482659)*x_ref[0]**o+(-0.808253846389)*x_ref[0]+(-0.911723431862)*x_ref[1]**o+(-0.266029942425)*x_ref[1]
        ref[0,3,1,1]=(0.909037857376)*x_ref[0]**o+(-0.164525039433)*x_ref[0]+(-0.924332370345)*x_ref[1]**o+(0.476479399362)*x_ref[1]
        ref[0,3,2,0]=(0.00534173833653)*x_ref[0]**o+(-0.435160515142)*x_ref[0]+(0.402701511687)*x_ref[1]**o+(0.161631276653)*x_ref[1]
        ref[0,3,2,1]=(-0.595791669027)*x_ref[0]**o+(0.170580206948)*x_ref[0]+(0.468802250813)*x_ref[1]**o+(-0.881472636722)*x_ref[1]
        ref[0,4,0,0]=(-0.0739506349513)*x_ref[0]**o+(0.513385505957)*x_ref[0]+(0.046977740371)*x_ref[1]**o+(0.95763159987)*x_ref[1]
        ref[0,4,0,1]=(-0.286789006223)*x_ref[0]**o+(0.017476293519)*x_ref[0]+(0.637378781092)*x_ref[1]**o+(-0.670201122126)*x_ref[1]
        ref[0,4,1,0]=(0.159607100455)*x_ref[0]**o+(0.715341910227)*x_ref[0]+(0.949546645819)*x_ref[1]**o+(0.362982182775)*x_ref[1]
        ref[0,4,1,1]=(-0.237702024358)*x_ref[0]**o+(-0.520488841454)*x_ref[0]+(0.606333933569)*x_ref[1]**o+(-0.00662757701049)*x_ref[1]
        ref[0,4,2,0]=(-0.415753817901)*x_ref[0]**o+(0.731285464784)*x_ref[0]+(0.15987336144)*x_ref[1]**o+(0.0697706236338)*x_ref[1]
        ref[0,4,2,1]=(0.0467482158427)*x_ref[0]**o+(0.306004722736)*x_ref[0]+(-0.905184725692)*x_ref[1]**o+(-0.158017889366)*x_ref[1]
        ref[1,0,0,0]=(-0.785535082426)*x_ref[0]**o+(-0.189285100452)*x_ref[0]+(-0.766349940057)*x_ref[1]**o+(0.531469308995)*x_ref[1]
        ref[1,0,0,1]=(0.522213196477)*x_ref[0]**o+(-0.661381741437)*x_ref[0]+(0.816089886956)*x_ref[1]**o+(0.139829382305)*x_ref[1]
        ref[1,0,1,0]=(0.0957295570417)*x_ref[0]**o+(-0.105252674536)*x_ref[0]+(-0.569429953382)*x_ref[1]**o+(-0.801934666657)*x_ref[1]
        ref[1,0,1,1]=(-0.557307051058)*x_ref[0]**o+(0.218852576416)*x_ref[0]+(0.430167216464)*x_ref[1]**o+(-0.112285681761)*x_ref[1]
        ref[1,0,2,0]=(0.288472422477)*x_ref[0]**o+(0.294377994763)*x_ref[0]+(-0.809503961511)*x_ref[1]**o+(0.357020226991)*x_ref[1]
        ref[1,0,2,1]=(-0.148113477705)*x_ref[0]**o+(0.164468646294)*x_ref[0]+(0.7977142476)*x_ref[1]**o+(-0.407274933384)*x_ref[1]
        ref[1,1,0,0]=(-0.926374434822)*x_ref[0]**o+(-0.107713463193)*x_ref[0]+(-0.511635723869)*x_ref[1]**o+(0.259319717689)*x_ref[1]
        ref[1,1,0,1]=(-0.18636582243)*x_ref[0]**o+(0.698258973628)*x_ref[0]+(0.234379281397)*x_ref[1]**o+(0.596195875495)*x_ref[1]
        ref[1,1,1,0]=(-0.180395892006)*x_ref[0]**o+(0.631248500084)*x_ref[0]+(0.104751616377)*x_ref[1]**o+(0.204500206653)*x_ref[1]
        ref[1,1,1,1]=(-0.517119564238)*x_ref[0]**o+(-0.16634015634)*x_ref[0]+(0.556785586675)*x_ref[1]**o+(0.42659306392)*x_ref[1]
        ref[1,1,2,0]=(0.163807238958)*x_ref[0]**o+(-0.534619478446)*x_ref[0]+(-0.700353460753)*x_ref[1]**o+(0.121393131648)*x_ref[1]
        ref[1,1,2,1]=(0.964276278961)*x_ref[0]**o+(0.846030751304)*x_ref[0]+(-0.745301894392)*x_ref[1]**o+(0.33821485558)*x_ref[1]
        ref[1,2,0,0]=(0.333173101086)*x_ref[0]**o+(0.774848048743)*x_ref[0]+(-0.659091916357)*x_ref[1]**o+(-0.797247144497)*x_ref[1]
        ref[1,2,0,1]=(0.0581394534285)*x_ref[0]**o+(-0.524771663031)*x_ref[0]+(-0.519465753317)*x_ref[1]**o+(0.71748795226)*x_ref[1]
        ref[1,2,1,0]=(0.224935262833)*x_ref[0]**o+(-0.393399819306)*x_ref[0]+(0.199349075224)*x_ref[1]**o+(-0.808847009412)*x_ref[1]
        ref[1,2,1,1]=(0.0385750795521)*x_ref[0]**o+(-8.09949553151e-05)*x_ref[0]+(0.496860964506)*x_ref[1]**o+(0.0119507407829)*x_ref[1]
        ref[1,2,2,0]=(0.876341556167)*x_ref[0]**o+(0.18082766372)*x_ref[0]+(0.401349871827)*x_ref[1]**o+(-0.543439037303)*x_ref[1]
        ref[1,2,2,1]=(-0.894525004507)*x_ref[0]**o+(-0.516771143695)*x_ref[0]+(0.0410916490263)*x_ref[1]**o+(0.753841618439)*x_ref[1]
        ref[1,3,0,0]=(0.251979970924)*x_ref[0]**o+(0.790276550438)*x_ref[0]+(-0.155708793888)*x_ref[1]**o+(0.28971278336)*x_ref[1]
        ref[1,3,0,1]=(0.258336265436)*x_ref[0]**o+(0.696642875149)*x_ref[0]+(0.0561355915649)*x_ref[1]**o+(0.601839744417)*x_ref[1]
        ref[1,3,1,0]=(-0.978729102785)*x_ref[0]**o+(0.335280613262)*x_ref[0]+(0.89228798232)*x_ref[1]**o+(-0.693050679348)*x_ref[1]
        ref[1,3,1,1]=(-0.885817753431)*x_ref[0]**o+(-0.0641571535708)*x_ref[0]+(-0.649578576181)*x_ref[1]**o+(-0.95773137602)*x_ref[1]
        ref[1,3,2,0]=(-0.333891001461)*x_ref[0]**o+(-0.496064308564)*x_ref[0]+(0.438549603377)*x_ref[1]**o+(0.0772468142672)*x_ref[1]
        ref[1,3,2,1]=(-0.16171788255)*x_ref[0]**o+(-0.939293453222)*x_ref[0]+(-0.284522648546)*x_ref[1]**o+(0.744642328375)*x_ref[1]
        ref[1,4,0,0]=(0.806133262332)*x_ref[0]**o+(-0.0225320206432)*x_ref[0]+(0.125958642606)*x_ref[1]**o+(0.610630249467)*x_ref[1]
        ref[1,4,0,1]=(0.657971870946)*x_ref[0]**o+(0.698894701525)*x_ref[0]+(-0.595952612496)*x_ref[1]**o+(-0.38268841292)*x_ref[1]
        ref[1,4,1,0]=(-0.277274411331)*x_ref[0]**o+(-0.483183004514)*x_ref[0]+(0.77656657009)*x_ref[1]**o+(-0.959290466645)*x_ref[1]
        ref[1,4,1,1]=(-0.96384356305)*x_ref[0]**o+(-0.00383301781476)*x_ref[0]+(0.788010409608)*x_ref[1]**o+(0.135378744316)*x_ref[1]
        ref[1,4,2,0]=(0.986761958017)*x_ref[0]**o+(-0.570276378038)*x_ref[0]+(-0.515784028315)*x_ref[1]**o+(0.487125123439)*x_ref[1]
        ref[1,4,2,1]=(-0.954735054702)*x_ref[0]**o+(-0.0610323074217)*x_ref[0]+(-0.255284859281)*x_ref[1]**o+(0.277795725698)*x_ref[1]
        ref[2,0,0,0]=(-0.708301219324)*x_ref[0]**o+(-0.970166781553)*x_ref[0]+(-0.0836451425855)*x_ref[1]**o+(0.525854202995)*x_ref[1]
        ref[2,0,0,1]=(-0.0995393865183)*x_ref[0]**o+(-0.0156624820324)*x_ref[0]+(0.225529316197)*x_ref[1]**o+(0.142617834056)*x_ref[1]
        ref[2,0,1,0]=(-0.585329524103)*x_ref[0]**o+(0.590285924441)*x_ref[0]+(0.201483726322)*x_ref[1]**o+(-0.735262197689)*x_ref[1]
        ref[2,0,1,1]=(-0.682877182963)*x_ref[0]**o+(-0.162682262799)*x_ref[0]+(-0.890051459843)*x_ref[1]**o+(-0.0188015866765)*x_ref[1]
        ref[2,0,2,0]=(0.868935476294)*x_ref[0]**o+(-0.629891068854)*x_ref[0]+(0.335632865663)*x_ref[1]**o+(-0.955849055716)*x_ref[1]
        ref[2,0,2,1]=(-0.907933228238)*x_ref[0]**o+(-0.105861987974)*x_ref[0]+(-0.131372383042)*x_ref[1]**o+(-0.970114136854)*x_ref[1]
        ref[2,1,0,0]=(-0.275788501814)*x_ref[0]**o+(-0.992861867033)*x_ref[0]+(-0.80676324962)*x_ref[1]**o+(0.269244680224)*x_ref[1]
        ref[2,1,0,1]=(-0.704704138774)*x_ref[0]**o+(-0.371803193009)*x_ref[0]+(-0.0190586405922)*x_ref[1]**o+(-0.0828837624698)*x_ref[1]
        ref[2,1,1,0]=(-0.532645566013)*x_ref[0]**o+(0.571754809868)*x_ref[0]+(0.656737493773)*x_ref[1]**o+(-0.548229591842)*x_ref[1]
        ref[2,1,1,1]=(0.23836921087)*x_ref[0]**o+(0.707475737785)*x_ref[0]+(-0.59660155968)*x_ref[1]**o+(-0.450204951071)*x_ref[1]
        ref[2,1,2,0]=(-0.534027348014)*x_ref[0]**o+(0.332100193727)*x_ref[0]+(-0.864157138567)*x_ref[1]**o+(-0.41833267066)*x_ref[1]
        ref[2,1,2,1]=(0.459878593269)*x_ref[0]**o+(0.446259140228)*x_ref[0]+(0.571242887502)*x_ref[1]**o+(-0.0390525669486)*x_ref[1]
        ref[2,2,0,0]=(0.322052799527)*x_ref[0]**o+(-0.915265209304)*x_ref[0]+(-0.99942007618)*x_ref[1]**o+(0.482294513743)*x_ref[1]
        ref[2,2,0,1]=(0.558489194456)*x_ref[0]**o+(0.00400576554086)*x_ref[0]+(-0.87484330165)*x_ref[1]**o+(-0.664504218977)*x_ref[1]
        ref[2,2,1,0]=(0.669621607461)*x_ref[0]**o+(-0.0450960554319)*x_ref[0]+(-0.221151480246)*x_ref[1]**o+(0.0123621737211)*x_ref[1]
        ref[2,2,1,1]=(0.476743013029)*x_ref[0]**o+(0.11118371822)*x_ref[0]+(-0.202074315677)*x_ref[1]**o+(0.377649433458)*x_ref[1]
        ref[2,2,2,0]=(0.667910543701)*x_ref[0]**o+(0.035703922547)*x_ref[0]+(-0.274733967095)*x_ref[1]**o+(-0.998934720182)*x_ref[1]
        ref[2,2,2,1]=(0.578909660087)*x_ref[0]**o+(0.149900086514)*x_ref[0]+(-0.226597651858)*x_ref[1]**o+(0.169175812287)*x_ref[1]
        ref[2,3,0,0]=(0.158370246586)*x_ref[0]**o+(0.595643112285)*x_ref[0]+(-0.476875564052)*x_ref[1]**o+(0.869752657517)*x_ref[1]
        ref[2,3,0,1]=(0.99718847714)*x_ref[0]**o+(0.581531822297)*x_ref[0]+(-0.607102775564)*x_ref[1]**o+(-0.8317970284)*x_ref[1]
        ref[2,3,1,0]=(-0.27390195745)*x_ref[0]**o+(0.0512853330414)*x_ref[0]+(0.465190666302)*x_ref[1]**o+(0.042533388788)*x_ref[1]
        ref[2,3,1,1]=(-0.183679504962)*x_ref[0]**o+(-0.0522782076121)*x_ref[0]+(0.921423118529)*x_ref[1]**o+(-0.0315080275225)*x_ref[1]
        ref[2,3,2,0]=(0.0421842348778)*x_ref[0]**o+(0.462023049702)*x_ref[0]+(-0.888342606803)*x_ref[1]**o+(-0.956503447884)*x_ref[1]
        ref[2,3,2,1]=(-0.950820288444)*x_ref[0]**o+(0.104271990168)*x_ref[0]+(-0.607212909582)*x_ref[1]**o+(-0.16283931259)*x_ref[1]
        ref[2,4,0,0]=(0.113505933533)*x_ref[0]**o+(0.27873514806)*x_ref[0]+(0.736699443523)*x_ref[1]**o+(-0.951737555693)*x_ref[1]
        ref[2,4,0,1]=(0.264170584849)*x_ref[0]**o+(-0.226649236617)*x_ref[0]+(0.489902180874)*x_ref[1]**o+(-0.191536591762)*x_ref[1]
        ref[2,4,1,0]=(0.712322285985)*x_ref[0]**o+(-0.323395455815)*x_ref[0]+(-0.298356681586)*x_ref[1]**o+(0.699891974541)*x_ref[1]
        ref[2,4,1,1]=(-0.631057763255)*x_ref[0]**o+(-0.489374933445)*x_ref[0]+(-0.021591806874)*x_ref[1]**o+(-0.692800124585)*x_ref[1]
        ref[2,4,2,0]=(0.916068935184)*x_ref[0]**o+(0.33779688858)*x_ref[0]+(0.06300443559)*x_ref[1]**o+(-0.231385203399)*x_ref[1]
        ref[2,4,2,1]=(-0.283485044625)*x_ref[0]**o+(0.548366574442)*x_ref[0]+(-0.543953132872)*x_ref[1]**o+(-0.956662823037)*x_ref[1]
        ref[3,0,0,0]=(0.709349274221)*x_ref[0]**o+(-0.342981009263)*x_ref[0]+(-0.682331761929)*x_ref[1]**o+(0.673169884813)*x_ref[1]
        ref[3,0,0,1]=(0.236249836593)*x_ref[0]**o+(-0.682845054434)*x_ref[0]+(-0.349169988218)*x_ref[1]**o+(0.502684425824)*x_ref[1]
        ref[3,0,1,0]=(-0.561655748793)*x_ref[0]**o+(0.120183772326)*x_ref[0]+(-0.966535325583)*x_ref[1]**o+(-0.0585117264949)*x_ref[1]
        ref[3,0,1,1]=(-0.661030287189)*x_ref[0]**o+(-0.322924086505)*x_ref[0]+(-0.986892059552)*x_ref[1]**o+(-0.778969361365)*x_ref[1]
        ref[3,0,2,0]=(-0.166307107307)*x_ref[0]**o+(0.612931303617)*x_ref[0]+(-0.283925165456)*x_ref[1]**o+(0.5998038814)*x_ref[1]
        ref[3,0,2,1]=(0.205099355916)*x_ref[0]**o+(-0.520970536278)*x_ref[0]+(0.900416833183)*x_ref[1]**o+(0.981826306069)*x_ref[1]
        ref[3,1,0,0]=(-0.875306666443)*x_ref[0]**o+(0.0189358479861)*x_ref[0]+(0.101187857445)*x_ref[1]**o+(-0.873584307534)*x_ref[1]
        ref[3,1,0,1]=(-0.0203656235757)*x_ref[0]**o+(-0.579229777042)*x_ref[0]+(-0.0799345631826)*x_ref[1]**o+(0.896315444069)*x_ref[1]
        ref[3,1,1,0]=(0.576791459982)*x_ref[0]**o+(-0.86541815506)*x_ref[0]+(-0.582764176207)*x_ref[1]**o+(-0.249437395011)*x_ref[1]
        ref[3,1,1,1]=(0.525953680497)*x_ref[0]**o+(-0.222445717402)*x_ref[0]+(-0.87710202063)*x_ref[1]**o+(-0.762921821973)*x_ref[1]
        ref[3,1,2,0]=(-0.883737185908)*x_ref[0]**o+(0.254053241277)*x_ref[0]+(0.559803298059)*x_ref[1]**o+(-0.577376956856)*x_ref[1]
        ref[3,1,2,1]=(-0.125792083387)*x_ref[0]**o+(-0.771549307509)*x_ref[0]+(0.142493480552)*x_ref[1]**o+(0.459996962548)*x_ref[1]
        ref[3,2,0,0]=(0.0531598927049)*x_ref[0]**o+(-0.950343229037)*x_ref[0]+(-0.748240453154)*x_ref[1]**o+(0.850555892342)*x_ref[1]
        ref[3,2,0,1]=(-0.660852152041)*x_ref[0]**o+(-0.763119278604)*x_ref[0]+(-0.811925409663)*x_ref[1]**o+(-0.178765071495)*x_ref[1]
        ref[3,2,1,0]=(-0.189124223305)*x_ref[0]**o+(-0.410392909042)*x_ref[0]+(0.993247167522)*x_ref[1]**o+(0.291451796298)*x_ref[1]
        ref[3,2,1,1]=(0.83496133421)*x_ref[0]**o+(0.695147480265)*x_ref[0]+(0.444420524946)*x_ref[1]**o+(0.107454620114)*x_ref[1]
        ref[3,2,2,0]=(0.94687479313)*x_ref[0]**o+(-0.140702324395)*x_ref[0]+(0.149203502428)*x_ref[1]**o+(-0.338246126345)*x_ref[1]
        ref[3,2,2,1]=(0.147609394709)*x_ref[0]**o+(-0.549219765239)*x_ref[0]+(0.397308993347)*x_ref[1]**o+(0.0402189312685)*x_ref[1]
        ref[3,3,0,0]=(0.27400981409)*x_ref[0]**o+(-0.516367266257)*x_ref[0]+(-0.784095952814)*x_ref[1]**o+(0.32858899009)*x_ref[1]
        ref[3,3,0,1]=(-0.393457382163)*x_ref[0]**o+(0.590406794051)*x_ref[0]+(0.328523918357)*x_ref[1]**o+(0.899537513355)*x_ref[1]
        ref[3,3,1,0]=(-0.851038623366)*x_ref[0]**o+(0.596974864726)*x_ref[0]+(-0.338630687551)*x_ref[1]**o+(0.634961435042)*x_ref[1]
        ref[3,3,1,1]=(0.569166143918)*x_ref[0]**o+(-0.388162282491)*x_ref[0]+(0.764830536733)*x_ref[1]**o+(-0.426737342542)*x_ref[1]
        ref[3,3,2,0]=(0.502221903448)*x_ref[0]**o+(0.940382225897)*x_ref[0]+(0.563444730579)*x_ref[1]**o+(0.023210869151)*x_ref[1]
        ref[3,3,2,1]=(0.061695403852)*x_ref[0]**o+(0.868302188758)*x_ref[0]+(0.0855803325758)*x_ref[1]**o+(-0.24366173638)*x_ref[1]
        ref[3,4,0,0]=(-0.533038182387)*x_ref[0]**o+(0.0222388915737)*x_ref[0]+(0.529037752769)*x_ref[1]**o+(-0.0665320895494)*x_ref[1]
        ref[3,4,0,1]=(0.714374055253)*x_ref[0]**o+(-0.623670743333)*x_ref[0]+(-0.866024792929)*x_ref[1]**o+(-0.0673124896111)*x_ref[1]
        ref[3,4,1,0]=(-0.0523466615045)*x_ref[0]**o+(-0.472563426622)*x_ref[0]+(0.139892673626)*x_ref[1]**o+(0.539655417178)*x_ref[1]
        ref[3,4,1,1]=(-0.473032580514)*x_ref[0]**o+(0.692812093618)*x_ref[0]+(0.664400517742)*x_ref[1]**o+(0.184125423364)*x_ref[1]
        ref[3,4,2,0]=(-0.0340798058725)*x_ref[0]**o+(-0.58900562863)*x_ref[0]+(0.396856384717)*x_ref[1]**o+(-0.559175428019)*x_ref[1]
        ref[3,4,2,1]=(-0.217679376634)*x_ref[0]**o+(-0.90389036667)*x_ref[0]+(0.967372942873)*x_ref[1]**o+(-0.482913619277)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.93229223906)*x[0]**o+(0.176839936782)*x[0]+(0.986271000359)*x[1]**o+(-0.449322385723)*x[1]+(-0.856450437823)*x[2]**o+(0.31473579221)*x[2]
        arg[0,0,0,1]=(0.714047232765)*x[0]**o+(-0.323471323003)*x[0]+(0.628080480343)*x[1]**o+(-0.992007912045)*x[1]+(-0.11019723051)*x[2]**o+(-0.104735408233)*x[2]
        arg[0,0,1,0]=(-0.437250443294)*x[0]**o+(-0.741469652781)*x[0]+(0.650921038465)*x[1]**o+(0.679622741243)*x[1]+(0.604005894895)*x[2]**o+(0.598155624876)*x[2]
        arg[0,0,1,1]=(-0.574872168761)*x[0]**o+(0.796207672391)*x[0]+(-0.478046646159)*x[1]**o+(0.604661326741)*x[1]+(-0.898457020186)*x[2]**o+(0.43234750827)*x[2]
        arg[0,0,2,0]=(-0.674008357742)*x[0]**o+(0.699614371016)*x[0]+(0.441321981559)*x[1]**o+(-0.619688689942)*x[1]+(-0.0874579208832)*x[2]**o+(-0.925753840555)*x[2]
        arg[0,0,2,1]=(-0.114329474892)*x[0]**o+(0.699501071503)*x[0]+(-0.618807420683)*x[1]**o+(0.411584695002)*x[1]+(-0.234620261556)*x[2]**o+(0.585720100967)*x[2]
        arg[0,1,0,0]=(0.617432097253)*x[0]**o+(-0.00304868863676)*x[0]+(0.571021785645)*x[1]**o+(0.457197205728)*x[1]+(0.462895867813)*x[2]**o+(0.278022956316)*x[2]
        arg[0,1,0,1]=(-0.659136231188)*x[0]**o+(-0.388212920906)*x[0]+(-0.982152608232)*x[1]**o+(-0.665797156013)*x[1]+(-0.76141890649)*x[2]**o+(-0.591458323836)*x[2]
        arg[0,1,1,0]=(-0.488016386893)*x[0]**o+(0.965660401944)*x[0]+(0.178400189819)*x[1]**o+(0.578417314499)*x[1]+(-0.398083955217)*x[2]**o+(0.460338187535)*x[2]
        arg[0,1,1,1]=(-0.182086458235)*x[0]**o+(0.0842368498629)*x[0]+(-0.642008026184)*x[1]**o+(-0.0449408131957)*x[1]+(-0.927397573018)*x[2]**o+(0.854313729816)*x[2]
        arg[0,1,2,0]=(-0.201475872769)*x[0]**o+(0.098190765292)*x[0]+(0.808624244922)*x[1]**o+(0.0726268899898)*x[1]+(-0.605334327027)*x[2]**o+(-0.999151255934)*x[2]
        arg[0,1,2,1]=(0.672344024218)*x[0]**o+(-0.638554833481)*x[0]+(-0.212231801634)*x[1]**o+(-0.492765366968)*x[1]+(0.311273771921)*x[2]**o+(-0.504370655006)*x[2]
        arg[0,2,0,0]=(0.9944935907)*x[0]**o+(-0.365261996611)*x[0]+(-0.105667939789)*x[1]**o+(0.886935751977)*x[1]+(0.263336157757)*x[2]**o+(-0.880316523113)*x[2]
        arg[0,2,0,1]=(-0.659779300787)*x[0]**o+(-0.728590180152)*x[0]+(0.732078508037)*x[1]**o+(0.684963412098)*x[1]+(0.822970700012)*x[2]**o+(-0.680001879674)*x[2]
        arg[0,2,1,0]=(-0.207855412893)*x[0]**o+(0.438275426529)*x[0]+(0.87046682066)*x[1]**o+(-0.960322013848)*x[1]+(-0.273728198625)*x[2]**o+(-0.960137228004)*x[2]
        arg[0,2,1,1]=(-0.189232680002)*x[0]**o+(0.866417302002)*x[0]+(-0.93403311248)*x[1]**o+(-0.55773104878)*x[1]+(0.94062895773)*x[2]**o+(-0.758052194805)*x[2]
        arg[0,2,2,0]=(0.100859944394)*x[0]**o+(-0.20292135909)*x[0]+(-0.713339941261)*x[1]**o+(-0.615106454411)*x[1]+(0.791245995152)*x[2]**o+(-0.81051791594)*x[2]
        arg[0,2,2,1]=(0.595743424517)*x[0]**o+(-0.179075876404)*x[0]+(-0.799005601715)*x[1]**o+(-0.0554635073569)*x[1]+(-0.850510701985)*x[2]**o+(0.234516538169)*x[2]
        arg[0,3,0,0]=(-0.361667738861)*x[0]**o+(-0.551837498696)*x[0]+(-0.834377202094)*x[1]**o+(0.814168224265)*x[1]+(-0.00479533268377)*x[2]**o+(0.843007652951)*x[2]
        arg[0,3,0,1]=(-0.448417439465)*x[0]**o+(0.431496158388)*x[0]+(-0.547416443433)*x[1]**o+(-0.0746785704695)*x[1]+(-0.646238182944)*x[2]**o+(-0.318608714595)*x[2]
        arg[0,3,1,0]=(-0.275647366007)*x[0]**o+(-0.410594785047)*x[0]+(-0.858282378196)*x[1]**o+(-0.755586690837)*x[1]+(-0.420481091654)*x[2]**o+(-0.0996487371831)*x[2]
        arg[0,3,1,1]=(-0.395961249009)*x[0]**o+(0.0860896467863)*x[0]+(0.482719525748)*x[1]**o+(0.319368178133)*x[1]+(0.516449153414)*x[2]**o+(-0.246749862576)*x[2]
        arg[0,3,2,0]=(-0.367322027122)*x[0]**o+(0.687206412524)*x[0]+(0.616265233972)*x[1]**o+(-0.32660589112)*x[1]+(-0.944607343096)*x[2]**o+(0.123982537951)*x[2]
        arg[0,3,2,1]=(0.700497166804)*x[0]**o+(0.553961329865)*x[0]+(-0.34020003754)*x[1]**o+(0.295870992783)*x[1]+(-0.220275880046)*x[2]**o+(0.766121035825)*x[2]
        arg[0,4,0,0]=(0.405469789242)*x[0]**o+(-0.338347662873)*x[0]+(0.0761796434233)*x[1]**o+(-0.120599175033)*x[1]+(0.107865585379)*x[2]**o+(0.0711941007838)*x[2]
        arg[0,4,0,1]=(0.614166154899)*x[0]**o+(0.195780702963)*x[0]+(0.718935985215)*x[1]**o+(0.214577296792)*x[1]+(-0.366985242107)*x[2]**o+(-0.509584787633)*x[2]
        arg[0,4,1,0]=(-0.638646370981)*x[0]**o+(-0.769955899284)*x[0]+(-0.887407540503)*x[1]**o+(0.0901245429998)*x[1]+(0.786085641321)*x[2]**o+(0.6248068552)*x[2]
        arg[0,4,1,1]=(-0.145426222043)*x[0]**o+(-0.494114545016)*x[0]+(0.422153721335)*x[1]**o+(0.617022143901)*x[1]+(-0.665972324723)*x[2]**o+(-0.476724432799)*x[2]
        arg[0,4,2,0]=(0.556105252092)*x[0]**o+(-0.893268696809)*x[0]+(0.146441889489)*x[1]**o+(-0.170512113283)*x[1]+(-0.869313262875)*x[2]**o+(-0.976874109392)*x[2]
        arg[0,4,2,1]=(0.858927206499)*x[0]**o+(0.198998355146)*x[0]+(0.278125296049)*x[1]**o+(-0.0442518828093)*x[1]+(0.274852724203)*x[2]**o+(-0.167649193314)*x[2]
        arg[1,0,0,0]=(-0.0429420289353)*x[0]**o+(-0.15764693929)*x[0]+(-0.859599917512)*x[1]**o+(-0.904178643575)*x[1]+(0.519560711493)*x[2]**o+(0.258644717907)*x[2]
        arg[1,0,0,1]=(-0.185321114264)*x[0]**o+(-0.794493830004)*x[0]+(0.136084421162)*x[1]**o+(0.926912021419)*x[1]+(0.346392120388)*x[2]**o+(-0.183908371454)*x[2]
        arg[1,0,1,0]=(-0.730675509877)*x[0]**o+(0.865034802778)*x[0]+(-0.506315431878)*x[1]**o+(-0.94607761051)*x[1]+(0.195194792801)*x[2]**o+(-0.865129258879)*x[2]
        arg[1,0,1,1]=(-0.0597026648107)*x[0]**o+(-0.884790497)*x[0]+(-0.0631278742098)*x[1]**o+(-0.808364940737)*x[1]+(-0.968952604837)*x[2]**o+(0.142246702632)*x[2]
        arg[1,0,2,0]=(0.648472894563)*x[0]**o+(0.145861644073)*x[0]+(0.501227520574)*x[1]**o+(0.265886482306)*x[1]+(-0.0472772173573)*x[2]**o+(-0.364169655969)*x[2]
        arg[1,0,2,1]=(0.847863600695)*x[0]**o+(-0.288841685199)*x[0]+(-0.500273502795)*x[1]**o+(0.627379277592)*x[1]+(0.326814024625)*x[2]**o+(-0.514712825895)*x[2]
        arg[1,1,0,0]=(-0.947845713712)*x[0]**o+(-0.372115616305)*x[0]+(-0.572858279998)*x[1]**o+(0.226237627692)*x[1]+(0.210201947182)*x[2]**o+(0.0147616538594)*x[2]
        arg[1,1,0,1]=(-0.39813630466)*x[0]**o+(-0.658496395731)*x[0]+(-0.363906465005)*x[1]**o+(-0.816719661238)*x[1]+(-0.478106835032)*x[2]**o+(0.0819042309589)*x[2]
        arg[1,1,1,0]=(-0.287160278593)*x[0]**o+(-0.0281402640978)*x[0]+(0.41559795378)*x[1]**o+(0.158274458097)*x[1]+(-0.798429481502)*x[2]**o+(-0.278585334624)*x[2]
        arg[1,1,1,1]=(0.809550480973)*x[0]**o+(0.192026915854)*x[0]+(-0.891168838599)*x[1]**o+(0.0482621192882)*x[1]+(-0.35339708159)*x[2]**o+(-0.133161551553)*x[2]
        arg[1,1,2,0]=(0.939582328872)*x[0]**o+(0.561963790516)*x[0]+(0.885264722949)*x[1]**o+(-0.468042305266)*x[1]+(0.232880002429)*x[2]**o+(-0.772896005701)*x[2]
        arg[1,1,2,1]=(-0.297333303209)*x[0]**o+(-0.414050924207)*x[0]+(-0.278903698873)*x[1]**o+(0.345370620869)*x[1]+(-0.0570151259027)*x[2]**o+(0.786449641378)*x[2]
        arg[1,2,0,0]=(0.644047138589)*x[0]**o+(-0.480660438838)*x[0]+(0.32526499462)*x[1]**o+(-0.775988282127)*x[1]+(0.108374143932)*x[2]**o+(0.955463667181)*x[2]
        arg[1,2,0,1]=(-0.468236790339)*x[0]**o+(0.844197711724)*x[0]+(0.808973429831)*x[1]**o+(-0.647699550005)*x[1]+(0.552555959813)*x[2]**o+(-0.3403079515)*x[2]
        arg[1,2,1,0]=(0.864686362308)*x[0]**o+(0.248923646116)*x[0]+(0.0287042863842)*x[1]**o+(0.404974282153)*x[1]+(0.113501438234)*x[2]**o+(-0.619364894397)*x[2]
        arg[1,2,1,1]=(-0.625146992499)*x[0]**o+(-0.0480602452983)*x[0]+(0.168910550914)*x[1]**o+(-0.891584218123)*x[1]+(0.0609357618263)*x[2]**o+(0.0451921002161)*x[2]
        arg[1,2,2,0]=(-0.478728628752)*x[0]**o+(0.428376547798)*x[0]+(0.817203376818)*x[1]**o+(-0.856056266953)*x[1]+(0.354960313382)*x[2]**o+(-0.611468960461)*x[2]
        arg[1,2,2,1]=(0.49500742113)*x[0]**o+(0.902948130263)*x[0]+(-0.810748543744)*x[1]**o+(-0.696150924858)*x[1]+(-0.748278516161)*x[2]**o+(0.694815527058)*x[2]
        arg[1,3,0,0]=(-0.361960150296)*x[0]**o+(-0.845139780461)*x[0]+(-0.0624499309665)*x[1]**o+(-0.0515558058489)*x[1]+(0.400690983017)*x[2]**o+(-0.744608407974)*x[2]
        arg[1,3,0,1]=(-0.452453617467)*x[0]**o+(0.845396139593)*x[0]+(0.292319469936)*x[1]**o+(-0.833547518634)*x[1]+(-0.511797700353)*x[2]**o+(0.857891454773)*x[2]
        arg[1,3,1,0]=(0.779504092912)*x[0]**o+(0.500524773615)*x[0]+(0.590542659382)*x[1]**o+(0.479310630519)*x[1]+(0.509960641956)*x[2]**o+(-0.564980432959)*x[2]
        arg[1,3,1,1]=(-0.145745189298)*x[0]**o+(-0.256064576359)*x[0]+(0.640890837126)*x[1]**o+(-0.718802035152)*x[1]+(-0.322216873583)*x[2]**o+(-0.0246526526334)*x[2]
        arg[1,3,2,0]=(-0.463434285098)*x[0]**o+(0.116722448643)*x[0]+(-0.690378381342)*x[1]**o+(0.429197465885)*x[1]+(-0.353019731384)*x[2]**o+(0.57397712918)*x[2]
        arg[1,3,2,1]=(0.169676853383)*x[0]**o+(0.815225472623)*x[0]+(-0.11125527552)*x[1]**o+(0.391513831525)*x[1]+(-0.587041361585)*x[2]**o+(0.255194151616)*x[2]
        arg[1,4,0,0]=(0.325351608628)*x[0]**o+(-0.57751715453)*x[0]+(0.145167511303)*x[1]**o+(-0.969286030742)*x[1]+(0.917370398478)*x[2]**o+(0.691483691598)*x[2]
        arg[1,4,0,1]=(0.0997998298271)*x[0]**o+(-0.101248202247)*x[0]+(-0.650649653848)*x[1]**o+(-0.157887947288)*x[1]+(0.517501747765)*x[2]**o+(-0.0852921926298)*x[2]
        arg[1,4,1,0]=(0.984595420407)*x[0]**o+(0.499293499076)*x[0]+(0.352193115874)*x[1]**o+(-0.684522909647)*x[1]+(-0.391538373605)*x[2]**o+(0.404635595162)*x[2]
        arg[1,4,1,1]=(-0.231310482173)*x[0]**o+(-0.0385089836559)*x[0]+(0.247671691888)*x[1]**o+(0.883971955027)*x[1]+(-0.0856624709734)*x[2]**o+(0.942940875904)*x[2]
        arg[1,4,2,0]=(-0.471467257434)*x[0]**o+(0.835634240897)*x[0]+(-0.594421294562)*x[1]**o+(-0.187584387111)*x[1]+(0.859867472765)*x[2]**o+(0.930951154576)*x[2]
        arg[1,4,2,1]=(0.504416975601)*x[0]**o+(0.435692923635)*x[0]+(-0.296945944313)*x[1]**o+(0.815156218765)*x[1]+(0.464464247442)*x[2]**o+(-0.378809453041)*x[2]
        arg[2,0,0,0]=(0.213087423107)*x[0]**o+(-0.504681665801)*x[0]+(0.796352190405)*x[1]**o+(0.546752751136)*x[1]+(0.57347644664)*x[2]**o+(0.715856963195)*x[2]
        arg[2,0,0,1]=(-0.28455266769)*x[0]**o+(0.853162220766)*x[0]+(0.164930982863)*x[1]**o+(-0.0136996498765)*x[1]+(0.293717915079)*x[2]**o+(-0.216124158473)*x[2]
        arg[2,0,1,0]=(-0.328115989441)*x[0]**o+(-0.529441496488)*x[0]+(-0.43579479677)*x[1]**o+(-0.851333801194)*x[1]+(-0.278406102972)*x[2]**o+(0.947777471824)*x[2]
        arg[2,0,1,1]=(-0.756438938789)*x[0]**o+(-0.624332272735)*x[0]+(-0.591638429565)*x[1]**o+(-0.0608281771566)*x[1]+(-0.852078784085)*x[2]**o+(-0.526558964866)*x[2]
        arg[2,0,2,0]=(-0.923892426969)*x[0]**o+(0.788121899878)*x[0]+(-0.446006331202)*x[1]**o+(0.997210871574)*x[1]+(0.356998606103)*x[2]**o+(0.972643822823)*x[2]
        arg[2,0,2,1]=(0.225183308571)*x[0]**o+(-0.757505414777)*x[0]+(-0.59573192074)*x[1]**o+(0.57166154987)*x[1]+(0.622343198809)*x[2]**o+(0.60343970503)*x[2]
        arg[2,1,0,0]=(0.51027702934)*x[0]**o+(-0.711737679522)*x[0]+(-0.211712579627)*x[1]**o+(-0.495817605314)*x[1]+(0.924425804272)*x[2]**o+(0.753424835275)*x[2]
        arg[2,1,0,1]=(0.476218252577)*x[0]**o+(0.187579891557)*x[0]+(0.727359583623)*x[1]**o+(-0.389875559955)*x[1]+(-0.352889326477)*x[2]**o+(0.365143943637)*x[2]
        arg[2,1,1,0]=(-0.815704312943)*x[0]**o+(0.315217470245)*x[0]+(-0.535234555065)*x[1]**o+(-0.433408988501)*x[1]+(-0.676694087389)*x[2]**o+(0.384097177718)*x[2]
        arg[2,1,1,1]=(-0.0368452411345)*x[0]**o+(0.514540686444)*x[0]+(-0.527100066965)*x[1]**o+(0.0291015562912)*x[1]+(-0.605646905004)*x[2]**o+(0.931528599981)*x[2]
        arg[2,1,2,0]=(-0.0994015945266)*x[0]**o+(-0.762310347434)*x[0]+(-0.845934950466)*x[1]**o+(-0.235758967923)*x[1]+(0.755531422698)*x[2]**o+(0.108353867777)*x[2]
        arg[2,1,2,1]=(-0.868779052889)*x[0]**o+(0.783878941075)*x[0]+(0.900089560473)*x[1]**o+(-0.836779405877)*x[1]+(-0.799571882449)*x[2]**o+(0.999710029157)*x[2]
        arg[2,2,0,0]=(0.541465277143)*x[0]**o+(-0.840477797945)*x[0]+(-0.362577721098)*x[1]**o+(0.00680131806718)*x[1]+(-0.338485239867)*x[2]**o+(0.828213756751)*x[2]
        arg[2,2,0,1]=(-0.927347423175)*x[0]**o+(-0.936191125785)*x[0]+(-0.411819804833)*x[1]**o+(-0.0222067637269)*x[1]+(-0.104411832586)*x[2]**o+(-0.79937203081)*x[2]
        arg[2,2,1,0]=(0.273051712848)*x[0]**o+(-0.802468557957)*x[0]+(-0.019392898867)*x[1]**o+(-0.612803074784)*x[1]+(-0.269705464223)*x[2]**o+(-0.244405147706)*x[2]
        arg[2,2,1,1]=(-0.08869649035)*x[0]**o+(0.809761782447)*x[0]+(0.761420080249)*x[1]**o+(0.16590038557)*x[1]+(-0.405657593445)*x[2]**o+(0.123593603037)*x[2]
        arg[2,2,2,0]=(-0.799691160987)*x[0]**o+(0.834835690475)*x[0]+(-0.0863569221737)*x[1]**o+(0.988343215152)*x[1]+(0.294423283474)*x[2]**o+(0.325725138203)*x[2]
        arg[2,2,2,1]=(0.648897496592)*x[0]**o+(-0.474329387314)*x[0]+(-0.904613697104)*x[1]**o+(0.382201388286)*x[1]+(-0.596889148339)*x[2]**o+(0.705132157684)*x[2]
        arg[2,3,0,0]=(0.906738144045)*x[0]**o+(-0.365392026426)*x[0]+(0.730557428733)*x[1]**o+(-0.857864471189)*x[1]+(-0.160930657186)*x[2]**o+(0.885067383334)*x[2]
        arg[2,3,0,1]=(0.992764702472)*x[0]**o+(0.720688532948)*x[0]+(0.318869772818)*x[1]**o+(0.995508347248)*x[1]+(0.108569456996)*x[2]**o+(-0.990939292995)*x[2]
        arg[2,3,1,0]=(0.690584992096)*x[0]**o+(0.605885242246)*x[0]+(0.72371794265)*x[1]**o+(-0.632609325879)*x[1]+(0.520273588734)*x[2]**o+(0.885804143155)*x[2]
        arg[2,3,1,1]=(0.485231542482)*x[0]**o+(0.866920225418)*x[0]+(-0.814964945972)*x[1]**o+(0.544132513209)*x[1]+(-0.155071376494)*x[2]**o+(-0.941300746237)*x[2]
        arg[2,3,2,0]=(0.748232238388)*x[0]**o+(0.179397724419)*x[0]+(-0.43881886248)*x[1]**o+(-0.124941799624)*x[1]+(0.14679335819)*x[2]**o+(-0.758738846688)*x[2]
        arg[2,3,2,1]=(-0.0951333655337)*x[0]**o+(0.37270972476)*x[0]+(-0.651977554115)*x[1]**o+(-0.88918597476)*x[1]+(-0.730858685576)*x[2]**o+(0.685881924459)*x[2]
        arg[2,4,0,0]=(-0.823487466802)*x[0]**o+(-0.583630855643)*x[0]+(-0.664814562747)*x[1]**o+(0.860952605956)*x[1]+(-0.249114322668)*x[2]**o+(-0.968356584237)*x[2]
        arg[2,4,0,1]=(0.44040146252)*x[0]**o+(-0.246827565902)*x[0]+(0.596300980733)*x[1]**o+(-0.825818200414)*x[1]+(0.0715512875967)*x[2]**o+(0.996064679803)*x[2]
        arg[2,4,1,0]=(-0.287806453021)*x[0]**o+(0.704331542927)*x[0]+(-0.346139754521)*x[1]**o+(-0.766667037125)*x[1]+(0.283874132105)*x[2]**o+(-0.613145979204)*x[2]
        arg[2,4,1,1]=(-0.116649875156)*x[0]**o+(0.291380048239)*x[0]+(0.689809221905)*x[1]**o+(0.758935910592)*x[1]+(0.505142787603)*x[2]**o+(-0.113611480503)*x[2]
        arg[2,4,2,0]=(-0.836908572692)*x[0]**o+(-0.21892828725)*x[0]+(-0.935780103868)*x[1]**o+(0.106204643268)*x[1]+(-0.894474423807)*x[2]**o+(0.296554773217)*x[2]
        arg[2,4,2,1]=(-0.89413495023)*x[0]**o+(0.6080813058)*x[0]+(0.087863889878)*x[1]**o+(0.213373665612)*x[1]+(0.17085345025)*x[2]**o+(-0.967625138712)*x[2]
        arg[3,0,0,0]=(0.327537677409)*x[0]**o+(-0.299373059373)*x[0]+(-0.43223648542)*x[1]**o+(-0.499115781051)*x[1]+(0.145464567449)*x[2]**o+(0.580930584563)*x[2]
        arg[3,0,0,1]=(-0.122622543351)*x[0]**o+(-0.220547821081)*x[0]+(0.892679827803)*x[1]**o+(0.436584604381)*x[1]+(-0.156093580288)*x[2]**o+(0.464069069075)*x[2]
        arg[3,0,1,0]=(0.590288663203)*x[0]**o+(-0.212929445251)*x[0]+(0.0899671639061)*x[1]**o+(-0.87729854872)*x[1]+(0.789506558716)*x[2]**o+(-0.68867690907)*x[2]
        arg[3,0,1,1]=(-0.913651849012)*x[0]**o+(-0.430938918892)*x[0]+(-0.351186012371)*x[1]**o+(0.500306994748)*x[1]+(0.145381845281)*x[2]**o+(0.624993552333)*x[2]
        arg[3,0,2,0]=(0.281393881265)*x[0]**o+(-0.704727129826)*x[0]+(-0.0102091836449)*x[1]**o+(0.311857501353)*x[1]+(0.300732947904)*x[2]**o+(0.923973499548)*x[2]
        arg[3,0,2,1]=(-0.584907526993)*x[0]**o+(0.625732316822)*x[0]+(-0.609691930408)*x[1]**o+(0.056160226956)*x[1]+(-0.766583561947)*x[2]**o+(0.307692190194)*x[2]
        arg[3,1,0,0]=(0.731231379445)*x[0]**o+(-0.0485980527115)*x[0]+(0.908932758852)*x[1]**o+(0.188918035886)*x[1]+(0.685837743688)*x[2]**o+(-0.299987553989)*x[2]
        arg[3,1,0,1]=(-0.784006561231)*x[0]**o+(-0.809789430022)*x[0]+(-0.846391792735)*x[1]**o+(-0.211583304405)*x[1]+(-0.00198590063687)*x[2]**o+(-0.853253255224)*x[2]
        arg[3,1,1,0]=(0.595567299132)*x[0]**o+(-0.581734416878)*x[0]+(0.88144481486)*x[1]**o+(0.776876564509)*x[1]+(-0.367477971386)*x[2]**o+(0.0806320164453)*x[2]
        arg[3,1,1,1]=(0.440836187611)*x[0]**o+(-0.578925091757)*x[0]+(0.923367189832)*x[1]**o+(-0.993888173929)*x[1]+(-0.645514380411)*x[2]**o+(0.0267439044199)*x[2]
        arg[3,1,2,0]=(-0.305182288788)*x[0]**o+(-0.503372502553)*x[0]+(0.740948657389)*x[1]**o+(0.285661052275)*x[1]+(-0.634510984218)*x[2]**o+(-0.123979746456)*x[2]
        arg[3,1,2,1]=(-0.897274928495)*x[0]**o+(0.971356796164)*x[0]+(0.940768290313)*x[1]**o+(-0.552941985667)*x[1]+(0.172857451423)*x[2]**o+(0.844758237523)*x[2]
        arg[3,2,0,0]=(-0.903542725333)*x[0]**o+(0.936704774551)*x[0]+(-0.398085875228)*x[1]**o+(-0.304061573512)*x[1]+(-0.13362037229)*x[2]**o+(-0.405043350728)*x[2]
        arg[3,2,0,1]=(-0.0208590751801)*x[0]**o+(-0.354086381885)*x[0]+(0.510884698744)*x[1]**o+(-0.275504880518)*x[1]+(0.459476496356)*x[2]**o+(0.25358097976)*x[2]
        arg[3,2,1,0]=(-0.0171081877107)*x[0]**o+(-0.856081626112)*x[0]+(-0.0698686571004)*x[1]**o+(0.0954918361122)*x[1]+(0.938494954069)*x[2]**o+(-0.824928688248)*x[2]
        arg[3,2,1,1]=(-0.277307915098)*x[0]**o+(0.737855779658)*x[0]+(-0.466761948855)*x[1]**o+(-0.821498435809)*x[1]+(-0.391009601442)*x[2]**o+(0.643834542048)*x[2]
        arg[3,2,2,0]=(0.379015007885)*x[0]**o+(0.348646179575)*x[0]+(0.733765133901)*x[1]**o+(-0.722363571789)*x[1]+(0.488531791931)*x[2]**o+(0.279813923679)*x[2]
        arg[3,2,2,1]=(0.516708382119)*x[0]**o+(0.0426665103314)*x[0]+(-0.168751315718)*x[1]**o+(-0.124069236608)*x[1]+(-0.778866001674)*x[2]**o+(-0.362031001279)*x[2]
        arg[3,3,0,0]=(0.995284803465)*x[0]**o+(-0.442327755879)*x[0]+(-0.13110270925)*x[1]**o+(-0.172857543102)*x[1]+(-0.910522664379)*x[2]**o+(0.516931833163)*x[2]
        arg[3,3,0,1]=(-0.889320295626)*x[0]**o+(-0.728330379534)*x[0]+(-0.584091700627)*x[1]**o+(0.469914082337)*x[1]+(0.18874127891)*x[2]**o+(0.0846532039249)*x[2]
        arg[3,3,1,0]=(0.911523535462)*x[0]**o+(0.0924361220354)*x[0]+(-0.0822504494228)*x[1]**o+(0.00630249023591)*x[1]+(-0.180497330322)*x[2]**o+(0.817405577154)*x[2]
        arg[3,3,1,1]=(0.681872747813)*x[0]**o+(0.840680190206)*x[0]+(-0.477660295206)*x[1]**o+(0.0399086968056)*x[1]+(0.440403204247)*x[2]**o+(-0.220656197016)*x[2]
        arg[3,3,2,0]=(-0.967524407336)*x[0]**o+(0.896652274462)*x[0]+(0.735269328519)*x[1]**o+(-0.249017384614)*x[1]+(-0.573222311563)*x[2]**o+(-0.653800334954)*x[2]
        arg[3,3,2,1]=(0.10041632804)*x[0]**o+(0.167657178348)*x[0]+(0.990621134594)*x[1]**o+(-0.309063473348)*x[1]+(-0.688206827575)*x[2]**o+(0.0926806818663)*x[2]
        arg[3,4,0,0]=(-0.128678339518)*x[0]**o+(-0.974238160059)*x[0]+(-0.762599464772)*x[1]**o+(0.0233286358698)*x[1]+(-0.622481697973)*x[2]**o+(-0.080495303798)*x[2]
        arg[3,4,0,1]=(-0.380318230058)*x[0]**o+(0.568425682425)*x[0]+(0.787369916262)*x[1]**o+(0.296745245123)*x[1]+(-0.917621002414)*x[2]**o+(-0.696209404954)*x[2]
        arg[3,4,1,0]=(-0.382961468898)*x[0]**o+(0.672614453621)*x[0]+(0.565371503874)*x[1]**o+(-0.597331999912)*x[1]+(-0.153637846704)*x[2]**o+(-0.438159587906)*x[2]
        arg[3,4,1,1]=(-0.541043007799)*x[0]**o+(0.415600628159)*x[0]+(-0.793716270785)*x[1]**o+(0.780110667203)*x[1]+(-0.101309486185)*x[2]**o+(-0.138868352939)*x[2]
        arg[3,4,2,0]=(0.679815917569)*x[0]**o+(-0.800657537776)*x[0]+(-0.711956069893)*x[1]**o+(-0.575626146908)*x[1]+(0.184247083249)*x[2]**o+(-0.728118084396)*x[2]
        arg[3,4,2,1]=(-0.608089432748)*x[0]**o+(0.356378798828)*x[0]+(-0.575957599599)*x[1]**o+(0.599586825137)*x[1]+(-0.160475743299)*x[2]**o+(-0.070762356177)*x[2]
        ref[0,0,0,0]=(-0.93229223906)*x_ref[0]**o+(0.176839936782)*x_ref[0]+(0.986271000359)*x_ref[1]**o+(-0.449322385723)*x_ref[1]+(-0.856450437823)*x_ref[2]**o+(0.31473579221)*x_ref[2]
        ref[0,0,0,1]=(0.714047232765)*x_ref[0]**o+(-0.323471323003)*x_ref[0]+(0.628080480343)*x_ref[1]**o+(-0.992007912045)*x_ref[1]+(-0.11019723051)*x_ref[2]**o+(-0.104735408233)*x_ref[2]
        ref[0,0,1,0]=(-0.437250443294)*x_ref[0]**o+(-0.741469652781)*x_ref[0]+(0.650921038465)*x_ref[1]**o+(0.679622741243)*x_ref[1]+(0.604005894895)*x_ref[2]**o+(0.598155624876)*x_ref[2]
        ref[0,0,1,1]=(-0.574872168761)*x_ref[0]**o+(0.796207672391)*x_ref[0]+(-0.478046646159)*x_ref[1]**o+(0.604661326741)*x_ref[1]+(-0.898457020186)*x_ref[2]**o+(0.43234750827)*x_ref[2]
        ref[0,0,2,0]=(-0.674008357742)*x_ref[0]**o+(0.699614371016)*x_ref[0]+(0.441321981559)*x_ref[1]**o+(-0.619688689942)*x_ref[1]+(-0.0874579208832)*x_ref[2]**o+(-0.925753840555)*x_ref[2]
        ref[0,0,2,1]=(-0.114329474892)*x_ref[0]**o+(0.699501071503)*x_ref[0]+(-0.618807420683)*x_ref[1]**o+(0.411584695002)*x_ref[1]+(-0.234620261556)*x_ref[2]**o+(0.585720100967)*x_ref[2]
        ref[0,1,0,0]=(0.617432097253)*x_ref[0]**o+(-0.00304868863676)*x_ref[0]+(0.571021785645)*x_ref[1]**o+(0.457197205728)*x_ref[1]+(0.462895867813)*x_ref[2]**o+(0.278022956316)*x_ref[2]
        ref[0,1,0,1]=(-0.659136231188)*x_ref[0]**o+(-0.388212920906)*x_ref[0]+(-0.982152608232)*x_ref[1]**o+(-0.665797156013)*x_ref[1]+(-0.76141890649)*x_ref[2]**o+(-0.591458323836)*x_ref[2]
        ref[0,1,1,0]=(-0.488016386893)*x_ref[0]**o+(0.965660401944)*x_ref[0]+(0.178400189819)*x_ref[1]**o+(0.578417314499)*x_ref[1]+(-0.398083955217)*x_ref[2]**o+(0.460338187535)*x_ref[2]
        ref[0,1,1,1]=(-0.182086458235)*x_ref[0]**o+(0.0842368498629)*x_ref[0]+(-0.642008026184)*x_ref[1]**o+(-0.0449408131957)*x_ref[1]+(-0.927397573018)*x_ref[2]**o+(0.854313729816)*x_ref[2]
        ref[0,1,2,0]=(-0.201475872769)*x_ref[0]**o+(0.098190765292)*x_ref[0]+(0.808624244922)*x_ref[1]**o+(0.0726268899898)*x_ref[1]+(-0.605334327027)*x_ref[2]**o+(-0.999151255934)*x_ref[2]
        ref[0,1,2,1]=(0.672344024218)*x_ref[0]**o+(-0.638554833481)*x_ref[0]+(-0.212231801634)*x_ref[1]**o+(-0.492765366968)*x_ref[1]+(0.311273771921)*x_ref[2]**o+(-0.504370655006)*x_ref[2]
        ref[0,2,0,0]=(0.9944935907)*x_ref[0]**o+(-0.365261996611)*x_ref[0]+(-0.105667939789)*x_ref[1]**o+(0.886935751977)*x_ref[1]+(0.263336157757)*x_ref[2]**o+(-0.880316523113)*x_ref[2]
        ref[0,2,0,1]=(-0.659779300787)*x_ref[0]**o+(-0.728590180152)*x_ref[0]+(0.732078508037)*x_ref[1]**o+(0.684963412098)*x_ref[1]+(0.822970700012)*x_ref[2]**o+(-0.680001879674)*x_ref[2]
        ref[0,2,1,0]=(-0.207855412893)*x_ref[0]**o+(0.438275426529)*x_ref[0]+(0.87046682066)*x_ref[1]**o+(-0.960322013848)*x_ref[1]+(-0.273728198625)*x_ref[2]**o+(-0.960137228004)*x_ref[2]
        ref[0,2,1,1]=(-0.189232680002)*x_ref[0]**o+(0.866417302002)*x_ref[0]+(-0.93403311248)*x_ref[1]**o+(-0.55773104878)*x_ref[1]+(0.94062895773)*x_ref[2]**o+(-0.758052194805)*x_ref[2]
        ref[0,2,2,0]=(0.100859944394)*x_ref[0]**o+(-0.20292135909)*x_ref[0]+(-0.713339941261)*x_ref[1]**o+(-0.615106454411)*x_ref[1]+(0.791245995152)*x_ref[2]**o+(-0.81051791594)*x_ref[2]
        ref[0,2,2,1]=(0.595743424517)*x_ref[0]**o+(-0.179075876404)*x_ref[0]+(-0.799005601715)*x_ref[1]**o+(-0.0554635073569)*x_ref[1]+(-0.850510701985)*x_ref[2]**o+(0.234516538169)*x_ref[2]
        ref[0,3,0,0]=(-0.361667738861)*x_ref[0]**o+(-0.551837498696)*x_ref[0]+(-0.834377202094)*x_ref[1]**o+(0.814168224265)*x_ref[1]+(-0.00479533268377)*x_ref[2]**o+(0.843007652951)*x_ref[2]
        ref[0,3,0,1]=(-0.448417439465)*x_ref[0]**o+(0.431496158388)*x_ref[0]+(-0.547416443433)*x_ref[1]**o+(-0.0746785704695)*x_ref[1]+(-0.646238182944)*x_ref[2]**o+(-0.318608714595)*x_ref[2]
        ref[0,3,1,0]=(-0.275647366007)*x_ref[0]**o+(-0.410594785047)*x_ref[0]+(-0.858282378196)*x_ref[1]**o+(-0.755586690837)*x_ref[1]+(-0.420481091654)*x_ref[2]**o+(-0.0996487371831)*x_ref[2]
        ref[0,3,1,1]=(-0.395961249009)*x_ref[0]**o+(0.0860896467863)*x_ref[0]+(0.482719525748)*x_ref[1]**o+(0.319368178133)*x_ref[1]+(0.516449153414)*x_ref[2]**o+(-0.246749862576)*x_ref[2]
        ref[0,3,2,0]=(-0.367322027122)*x_ref[0]**o+(0.687206412524)*x_ref[0]+(0.616265233972)*x_ref[1]**o+(-0.32660589112)*x_ref[1]+(-0.944607343096)*x_ref[2]**o+(0.123982537951)*x_ref[2]
        ref[0,3,2,1]=(0.700497166804)*x_ref[0]**o+(0.553961329865)*x_ref[0]+(-0.34020003754)*x_ref[1]**o+(0.295870992783)*x_ref[1]+(-0.220275880046)*x_ref[2]**o+(0.766121035825)*x_ref[2]
        ref[0,4,0,0]=(0.405469789242)*x_ref[0]**o+(-0.338347662873)*x_ref[0]+(0.0761796434233)*x_ref[1]**o+(-0.120599175033)*x_ref[1]+(0.107865585379)*x_ref[2]**o+(0.0711941007838)*x_ref[2]
        ref[0,4,0,1]=(0.614166154899)*x_ref[0]**o+(0.195780702963)*x_ref[0]+(0.718935985215)*x_ref[1]**o+(0.214577296792)*x_ref[1]+(-0.366985242107)*x_ref[2]**o+(-0.509584787633)*x_ref[2]
        ref[0,4,1,0]=(-0.638646370981)*x_ref[0]**o+(-0.769955899284)*x_ref[0]+(-0.887407540503)*x_ref[1]**o+(0.0901245429998)*x_ref[1]+(0.786085641321)*x_ref[2]**o+(0.6248068552)*x_ref[2]
        ref[0,4,1,1]=(-0.145426222043)*x_ref[0]**o+(-0.494114545016)*x_ref[0]+(0.422153721335)*x_ref[1]**o+(0.617022143901)*x_ref[1]+(-0.665972324723)*x_ref[2]**o+(-0.476724432799)*x_ref[2]
        ref[0,4,2,0]=(0.556105252092)*x_ref[0]**o+(-0.893268696809)*x_ref[0]+(0.146441889489)*x_ref[1]**o+(-0.170512113283)*x_ref[1]+(-0.869313262875)*x_ref[2]**o+(-0.976874109392)*x_ref[2]
        ref[0,4,2,1]=(0.858927206499)*x_ref[0]**o+(0.198998355146)*x_ref[0]+(0.278125296049)*x_ref[1]**o+(-0.0442518828093)*x_ref[1]+(0.274852724203)*x_ref[2]**o+(-0.167649193314)*x_ref[2]
        ref[1,0,0,0]=(-0.0429420289353)*x_ref[0]**o+(-0.15764693929)*x_ref[0]+(-0.859599917512)*x_ref[1]**o+(-0.904178643575)*x_ref[1]+(0.519560711493)*x_ref[2]**o+(0.258644717907)*x_ref[2]
        ref[1,0,0,1]=(-0.185321114264)*x_ref[0]**o+(-0.794493830004)*x_ref[0]+(0.136084421162)*x_ref[1]**o+(0.926912021419)*x_ref[1]+(0.346392120388)*x_ref[2]**o+(-0.183908371454)*x_ref[2]
        ref[1,0,1,0]=(-0.730675509877)*x_ref[0]**o+(0.865034802778)*x_ref[0]+(-0.506315431878)*x_ref[1]**o+(-0.94607761051)*x_ref[1]+(0.195194792801)*x_ref[2]**o+(-0.865129258879)*x_ref[2]
        ref[1,0,1,1]=(-0.0597026648107)*x_ref[0]**o+(-0.884790497)*x_ref[0]+(-0.0631278742098)*x_ref[1]**o+(-0.808364940737)*x_ref[1]+(-0.968952604837)*x_ref[2]**o+(0.142246702632)*x_ref[2]
        ref[1,0,2,0]=(0.648472894563)*x_ref[0]**o+(0.145861644073)*x_ref[0]+(0.501227520574)*x_ref[1]**o+(0.265886482306)*x_ref[1]+(-0.0472772173573)*x_ref[2]**o+(-0.364169655969)*x_ref[2]
        ref[1,0,2,1]=(0.847863600695)*x_ref[0]**o+(-0.288841685199)*x_ref[0]+(-0.500273502795)*x_ref[1]**o+(0.627379277592)*x_ref[1]+(0.326814024625)*x_ref[2]**o+(-0.514712825895)*x_ref[2]
        ref[1,1,0,0]=(-0.947845713712)*x_ref[0]**o+(-0.372115616305)*x_ref[0]+(-0.572858279998)*x_ref[1]**o+(0.226237627692)*x_ref[1]+(0.210201947182)*x_ref[2]**o+(0.0147616538594)*x_ref[2]
        ref[1,1,0,1]=(-0.39813630466)*x_ref[0]**o+(-0.658496395731)*x_ref[0]+(-0.363906465005)*x_ref[1]**o+(-0.816719661238)*x_ref[1]+(-0.478106835032)*x_ref[2]**o+(0.0819042309589)*x_ref[2]
        ref[1,1,1,0]=(-0.287160278593)*x_ref[0]**o+(-0.0281402640978)*x_ref[0]+(0.41559795378)*x_ref[1]**o+(0.158274458097)*x_ref[1]+(-0.798429481502)*x_ref[2]**o+(-0.278585334624)*x_ref[2]
        ref[1,1,1,1]=(0.809550480973)*x_ref[0]**o+(0.192026915854)*x_ref[0]+(-0.891168838599)*x_ref[1]**o+(0.0482621192882)*x_ref[1]+(-0.35339708159)*x_ref[2]**o+(-0.133161551553)*x_ref[2]
        ref[1,1,2,0]=(0.939582328872)*x_ref[0]**o+(0.561963790516)*x_ref[0]+(0.885264722949)*x_ref[1]**o+(-0.468042305266)*x_ref[1]+(0.232880002429)*x_ref[2]**o+(-0.772896005701)*x_ref[2]
        ref[1,1,2,1]=(-0.297333303209)*x_ref[0]**o+(-0.414050924207)*x_ref[0]+(-0.278903698873)*x_ref[1]**o+(0.345370620869)*x_ref[1]+(-0.0570151259027)*x_ref[2]**o+(0.786449641378)*x_ref[2]
        ref[1,2,0,0]=(0.644047138589)*x_ref[0]**o+(-0.480660438838)*x_ref[0]+(0.32526499462)*x_ref[1]**o+(-0.775988282127)*x_ref[1]+(0.108374143932)*x_ref[2]**o+(0.955463667181)*x_ref[2]
        ref[1,2,0,1]=(-0.468236790339)*x_ref[0]**o+(0.844197711724)*x_ref[0]+(0.808973429831)*x_ref[1]**o+(-0.647699550005)*x_ref[1]+(0.552555959813)*x_ref[2]**o+(-0.3403079515)*x_ref[2]
        ref[1,2,1,0]=(0.864686362308)*x_ref[0]**o+(0.248923646116)*x_ref[0]+(0.0287042863842)*x_ref[1]**o+(0.404974282153)*x_ref[1]+(0.113501438234)*x_ref[2]**o+(-0.619364894397)*x_ref[2]
        ref[1,2,1,1]=(-0.625146992499)*x_ref[0]**o+(-0.0480602452983)*x_ref[0]+(0.168910550914)*x_ref[1]**o+(-0.891584218123)*x_ref[1]+(0.0609357618263)*x_ref[2]**o+(0.0451921002161)*x_ref[2]
        ref[1,2,2,0]=(-0.478728628752)*x_ref[0]**o+(0.428376547798)*x_ref[0]+(0.817203376818)*x_ref[1]**o+(-0.856056266953)*x_ref[1]+(0.354960313382)*x_ref[2]**o+(-0.611468960461)*x_ref[2]
        ref[1,2,2,1]=(0.49500742113)*x_ref[0]**o+(0.902948130263)*x_ref[0]+(-0.810748543744)*x_ref[1]**o+(-0.696150924858)*x_ref[1]+(-0.748278516161)*x_ref[2]**o+(0.694815527058)*x_ref[2]
        ref[1,3,0,0]=(-0.361960150296)*x_ref[0]**o+(-0.845139780461)*x_ref[0]+(-0.0624499309665)*x_ref[1]**o+(-0.0515558058489)*x_ref[1]+(0.400690983017)*x_ref[2]**o+(-0.744608407974)*x_ref[2]
        ref[1,3,0,1]=(-0.452453617467)*x_ref[0]**o+(0.845396139593)*x_ref[0]+(0.292319469936)*x_ref[1]**o+(-0.833547518634)*x_ref[1]+(-0.511797700353)*x_ref[2]**o+(0.857891454773)*x_ref[2]
        ref[1,3,1,0]=(0.779504092912)*x_ref[0]**o+(0.500524773615)*x_ref[0]+(0.590542659382)*x_ref[1]**o+(0.479310630519)*x_ref[1]+(0.509960641956)*x_ref[2]**o+(-0.564980432959)*x_ref[2]
        ref[1,3,1,1]=(-0.145745189298)*x_ref[0]**o+(-0.256064576359)*x_ref[0]+(0.640890837126)*x_ref[1]**o+(-0.718802035152)*x_ref[1]+(-0.322216873583)*x_ref[2]**o+(-0.0246526526334)*x_ref[2]
        ref[1,3,2,0]=(-0.463434285098)*x_ref[0]**o+(0.116722448643)*x_ref[0]+(-0.690378381342)*x_ref[1]**o+(0.429197465885)*x_ref[1]+(-0.353019731384)*x_ref[2]**o+(0.57397712918)*x_ref[2]
        ref[1,3,2,1]=(0.169676853383)*x_ref[0]**o+(0.815225472623)*x_ref[0]+(-0.11125527552)*x_ref[1]**o+(0.391513831525)*x_ref[1]+(-0.587041361585)*x_ref[2]**o+(0.255194151616)*x_ref[2]
        ref[1,4,0,0]=(0.325351608628)*x_ref[0]**o+(-0.57751715453)*x_ref[0]+(0.145167511303)*x_ref[1]**o+(-0.969286030742)*x_ref[1]+(0.917370398478)*x_ref[2]**o+(0.691483691598)*x_ref[2]
        ref[1,4,0,1]=(0.0997998298271)*x_ref[0]**o+(-0.101248202247)*x_ref[0]+(-0.650649653848)*x_ref[1]**o+(-0.157887947288)*x_ref[1]+(0.517501747765)*x_ref[2]**o+(-0.0852921926298)*x_ref[2]
        ref[1,4,1,0]=(0.984595420407)*x_ref[0]**o+(0.499293499076)*x_ref[0]+(0.352193115874)*x_ref[1]**o+(-0.684522909647)*x_ref[1]+(-0.391538373605)*x_ref[2]**o+(0.404635595162)*x_ref[2]
        ref[1,4,1,1]=(-0.231310482173)*x_ref[0]**o+(-0.0385089836559)*x_ref[0]+(0.247671691888)*x_ref[1]**o+(0.883971955027)*x_ref[1]+(-0.0856624709734)*x_ref[2]**o+(0.942940875904)*x_ref[2]
        ref[1,4,2,0]=(-0.471467257434)*x_ref[0]**o+(0.835634240897)*x_ref[0]+(-0.594421294562)*x_ref[1]**o+(-0.187584387111)*x_ref[1]+(0.859867472765)*x_ref[2]**o+(0.930951154576)*x_ref[2]
        ref[1,4,2,1]=(0.504416975601)*x_ref[0]**o+(0.435692923635)*x_ref[0]+(-0.296945944313)*x_ref[1]**o+(0.815156218765)*x_ref[1]+(0.464464247442)*x_ref[2]**o+(-0.378809453041)*x_ref[2]
        ref[2,0,0,0]=(0.213087423107)*x_ref[0]**o+(-0.504681665801)*x_ref[0]+(0.796352190405)*x_ref[1]**o+(0.546752751136)*x_ref[1]+(0.57347644664)*x_ref[2]**o+(0.715856963195)*x_ref[2]
        ref[2,0,0,1]=(-0.28455266769)*x_ref[0]**o+(0.853162220766)*x_ref[0]+(0.164930982863)*x_ref[1]**o+(-0.0136996498765)*x_ref[1]+(0.293717915079)*x_ref[2]**o+(-0.216124158473)*x_ref[2]
        ref[2,0,1,0]=(-0.328115989441)*x_ref[0]**o+(-0.529441496488)*x_ref[0]+(-0.43579479677)*x_ref[1]**o+(-0.851333801194)*x_ref[1]+(-0.278406102972)*x_ref[2]**o+(0.947777471824)*x_ref[2]
        ref[2,0,1,1]=(-0.756438938789)*x_ref[0]**o+(-0.624332272735)*x_ref[0]+(-0.591638429565)*x_ref[1]**o+(-0.0608281771566)*x_ref[1]+(-0.852078784085)*x_ref[2]**o+(-0.526558964866)*x_ref[2]
        ref[2,0,2,0]=(-0.923892426969)*x_ref[0]**o+(0.788121899878)*x_ref[0]+(-0.446006331202)*x_ref[1]**o+(0.997210871574)*x_ref[1]+(0.356998606103)*x_ref[2]**o+(0.972643822823)*x_ref[2]
        ref[2,0,2,1]=(0.225183308571)*x_ref[0]**o+(-0.757505414777)*x_ref[0]+(-0.59573192074)*x_ref[1]**o+(0.57166154987)*x_ref[1]+(0.622343198809)*x_ref[2]**o+(0.60343970503)*x_ref[2]
        ref[2,1,0,0]=(0.51027702934)*x_ref[0]**o+(-0.711737679522)*x_ref[0]+(-0.211712579627)*x_ref[1]**o+(-0.495817605314)*x_ref[1]+(0.924425804272)*x_ref[2]**o+(0.753424835275)*x_ref[2]
        ref[2,1,0,1]=(0.476218252577)*x_ref[0]**o+(0.187579891557)*x_ref[0]+(0.727359583623)*x_ref[1]**o+(-0.389875559955)*x_ref[1]+(-0.352889326477)*x_ref[2]**o+(0.365143943637)*x_ref[2]
        ref[2,1,1,0]=(-0.815704312943)*x_ref[0]**o+(0.315217470245)*x_ref[0]+(-0.535234555065)*x_ref[1]**o+(-0.433408988501)*x_ref[1]+(-0.676694087389)*x_ref[2]**o+(0.384097177718)*x_ref[2]
        ref[2,1,1,1]=(-0.0368452411345)*x_ref[0]**o+(0.514540686444)*x_ref[0]+(-0.527100066965)*x_ref[1]**o+(0.0291015562912)*x_ref[1]+(-0.605646905004)*x_ref[2]**o+(0.931528599981)*x_ref[2]
        ref[2,1,2,0]=(-0.0994015945266)*x_ref[0]**o+(-0.762310347434)*x_ref[0]+(-0.845934950466)*x_ref[1]**o+(-0.235758967923)*x_ref[1]+(0.755531422698)*x_ref[2]**o+(0.108353867777)*x_ref[2]
        ref[2,1,2,1]=(-0.868779052889)*x_ref[0]**o+(0.783878941075)*x_ref[0]+(0.900089560473)*x_ref[1]**o+(-0.836779405877)*x_ref[1]+(-0.799571882449)*x_ref[2]**o+(0.999710029157)*x_ref[2]
        ref[2,2,0,0]=(0.541465277143)*x_ref[0]**o+(-0.840477797945)*x_ref[0]+(-0.362577721098)*x_ref[1]**o+(0.00680131806718)*x_ref[1]+(-0.338485239867)*x_ref[2]**o+(0.828213756751)*x_ref[2]
        ref[2,2,0,1]=(-0.927347423175)*x_ref[0]**o+(-0.936191125785)*x_ref[0]+(-0.411819804833)*x_ref[1]**o+(-0.0222067637269)*x_ref[1]+(-0.104411832586)*x_ref[2]**o+(-0.79937203081)*x_ref[2]
        ref[2,2,1,0]=(0.273051712848)*x_ref[0]**o+(-0.802468557957)*x_ref[0]+(-0.019392898867)*x_ref[1]**o+(-0.612803074784)*x_ref[1]+(-0.269705464223)*x_ref[2]**o+(-0.244405147706)*x_ref[2]
        ref[2,2,1,1]=(-0.08869649035)*x_ref[0]**o+(0.809761782447)*x_ref[0]+(0.761420080249)*x_ref[1]**o+(0.16590038557)*x_ref[1]+(-0.405657593445)*x_ref[2]**o+(0.123593603037)*x_ref[2]
        ref[2,2,2,0]=(-0.799691160987)*x_ref[0]**o+(0.834835690475)*x_ref[0]+(-0.0863569221737)*x_ref[1]**o+(0.988343215152)*x_ref[1]+(0.294423283474)*x_ref[2]**o+(0.325725138203)*x_ref[2]
        ref[2,2,2,1]=(0.648897496592)*x_ref[0]**o+(-0.474329387314)*x_ref[0]+(-0.904613697104)*x_ref[1]**o+(0.382201388286)*x_ref[1]+(-0.596889148339)*x_ref[2]**o+(0.705132157684)*x_ref[2]
        ref[2,3,0,0]=(0.906738144045)*x_ref[0]**o+(-0.365392026426)*x_ref[0]+(0.730557428733)*x_ref[1]**o+(-0.857864471189)*x_ref[1]+(-0.160930657186)*x_ref[2]**o+(0.885067383334)*x_ref[2]
        ref[2,3,0,1]=(0.992764702472)*x_ref[0]**o+(0.720688532948)*x_ref[0]+(0.318869772818)*x_ref[1]**o+(0.995508347248)*x_ref[1]+(0.108569456996)*x_ref[2]**o+(-0.990939292995)*x_ref[2]
        ref[2,3,1,0]=(0.690584992096)*x_ref[0]**o+(0.605885242246)*x_ref[0]+(0.72371794265)*x_ref[1]**o+(-0.632609325879)*x_ref[1]+(0.520273588734)*x_ref[2]**o+(0.885804143155)*x_ref[2]
        ref[2,3,1,1]=(0.485231542482)*x_ref[0]**o+(0.866920225418)*x_ref[0]+(-0.814964945972)*x_ref[1]**o+(0.544132513209)*x_ref[1]+(-0.155071376494)*x_ref[2]**o+(-0.941300746237)*x_ref[2]
        ref[2,3,2,0]=(0.748232238388)*x_ref[0]**o+(0.179397724419)*x_ref[0]+(-0.43881886248)*x_ref[1]**o+(-0.124941799624)*x_ref[1]+(0.14679335819)*x_ref[2]**o+(-0.758738846688)*x_ref[2]
        ref[2,3,2,1]=(-0.0951333655337)*x_ref[0]**o+(0.37270972476)*x_ref[0]+(-0.651977554115)*x_ref[1]**o+(-0.88918597476)*x_ref[1]+(-0.730858685576)*x_ref[2]**o+(0.685881924459)*x_ref[2]
        ref[2,4,0,0]=(-0.823487466802)*x_ref[0]**o+(-0.583630855643)*x_ref[0]+(-0.664814562747)*x_ref[1]**o+(0.860952605956)*x_ref[1]+(-0.249114322668)*x_ref[2]**o+(-0.968356584237)*x_ref[2]
        ref[2,4,0,1]=(0.44040146252)*x_ref[0]**o+(-0.246827565902)*x_ref[0]+(0.596300980733)*x_ref[1]**o+(-0.825818200414)*x_ref[1]+(0.0715512875967)*x_ref[2]**o+(0.996064679803)*x_ref[2]
        ref[2,4,1,0]=(-0.287806453021)*x_ref[0]**o+(0.704331542927)*x_ref[0]+(-0.346139754521)*x_ref[1]**o+(-0.766667037125)*x_ref[1]+(0.283874132105)*x_ref[2]**o+(-0.613145979204)*x_ref[2]
        ref[2,4,1,1]=(-0.116649875156)*x_ref[0]**o+(0.291380048239)*x_ref[0]+(0.689809221905)*x_ref[1]**o+(0.758935910592)*x_ref[1]+(0.505142787603)*x_ref[2]**o+(-0.113611480503)*x_ref[2]
        ref[2,4,2,0]=(-0.836908572692)*x_ref[0]**o+(-0.21892828725)*x_ref[0]+(-0.935780103868)*x_ref[1]**o+(0.106204643268)*x_ref[1]+(-0.894474423807)*x_ref[2]**o+(0.296554773217)*x_ref[2]
        ref[2,4,2,1]=(-0.89413495023)*x_ref[0]**o+(0.6080813058)*x_ref[0]+(0.087863889878)*x_ref[1]**o+(0.213373665612)*x_ref[1]+(0.17085345025)*x_ref[2]**o+(-0.967625138712)*x_ref[2]
        ref[3,0,0,0]=(0.327537677409)*x_ref[0]**o+(-0.299373059373)*x_ref[0]+(-0.43223648542)*x_ref[1]**o+(-0.499115781051)*x_ref[1]+(0.145464567449)*x_ref[2]**o+(0.580930584563)*x_ref[2]
        ref[3,0,0,1]=(-0.122622543351)*x_ref[0]**o+(-0.220547821081)*x_ref[0]+(0.892679827803)*x_ref[1]**o+(0.436584604381)*x_ref[1]+(-0.156093580288)*x_ref[2]**o+(0.464069069075)*x_ref[2]
        ref[3,0,1,0]=(0.590288663203)*x_ref[0]**o+(-0.212929445251)*x_ref[0]+(0.0899671639061)*x_ref[1]**o+(-0.87729854872)*x_ref[1]+(0.789506558716)*x_ref[2]**o+(-0.68867690907)*x_ref[2]
        ref[3,0,1,1]=(-0.913651849012)*x_ref[0]**o+(-0.430938918892)*x_ref[0]+(-0.351186012371)*x_ref[1]**o+(0.500306994748)*x_ref[1]+(0.145381845281)*x_ref[2]**o+(0.624993552333)*x_ref[2]
        ref[3,0,2,0]=(0.281393881265)*x_ref[0]**o+(-0.704727129826)*x_ref[0]+(-0.0102091836449)*x_ref[1]**o+(0.311857501353)*x_ref[1]+(0.300732947904)*x_ref[2]**o+(0.923973499548)*x_ref[2]
        ref[3,0,2,1]=(-0.584907526993)*x_ref[0]**o+(0.625732316822)*x_ref[0]+(-0.609691930408)*x_ref[1]**o+(0.056160226956)*x_ref[1]+(-0.766583561947)*x_ref[2]**o+(0.307692190194)*x_ref[2]
        ref[3,1,0,0]=(0.731231379445)*x_ref[0]**o+(-0.0485980527115)*x_ref[0]+(0.908932758852)*x_ref[1]**o+(0.188918035886)*x_ref[1]+(0.685837743688)*x_ref[2]**o+(-0.299987553989)*x_ref[2]
        ref[3,1,0,1]=(-0.784006561231)*x_ref[0]**o+(-0.809789430022)*x_ref[0]+(-0.846391792735)*x_ref[1]**o+(-0.211583304405)*x_ref[1]+(-0.00198590063687)*x_ref[2]**o+(-0.853253255224)*x_ref[2]
        ref[3,1,1,0]=(0.595567299132)*x_ref[0]**o+(-0.581734416878)*x_ref[0]+(0.88144481486)*x_ref[1]**o+(0.776876564509)*x_ref[1]+(-0.367477971386)*x_ref[2]**o+(0.0806320164453)*x_ref[2]
        ref[3,1,1,1]=(0.440836187611)*x_ref[0]**o+(-0.578925091757)*x_ref[0]+(0.923367189832)*x_ref[1]**o+(-0.993888173929)*x_ref[1]+(-0.645514380411)*x_ref[2]**o+(0.0267439044199)*x_ref[2]
        ref[3,1,2,0]=(-0.305182288788)*x_ref[0]**o+(-0.503372502553)*x_ref[0]+(0.740948657389)*x_ref[1]**o+(0.285661052275)*x_ref[1]+(-0.634510984218)*x_ref[2]**o+(-0.123979746456)*x_ref[2]
        ref[3,1,2,1]=(-0.897274928495)*x_ref[0]**o+(0.971356796164)*x_ref[0]+(0.940768290313)*x_ref[1]**o+(-0.552941985667)*x_ref[1]+(0.172857451423)*x_ref[2]**o+(0.844758237523)*x_ref[2]
        ref[3,2,0,0]=(-0.903542725333)*x_ref[0]**o+(0.936704774551)*x_ref[0]+(-0.398085875228)*x_ref[1]**o+(-0.304061573512)*x_ref[1]+(-0.13362037229)*x_ref[2]**o+(-0.405043350728)*x_ref[2]
        ref[3,2,0,1]=(-0.0208590751801)*x_ref[0]**o+(-0.354086381885)*x_ref[0]+(0.510884698744)*x_ref[1]**o+(-0.275504880518)*x_ref[1]+(0.459476496356)*x_ref[2]**o+(0.25358097976)*x_ref[2]
        ref[3,2,1,0]=(-0.0171081877107)*x_ref[0]**o+(-0.856081626112)*x_ref[0]+(-0.0698686571004)*x_ref[1]**o+(0.0954918361122)*x_ref[1]+(0.938494954069)*x_ref[2]**o+(-0.824928688248)*x_ref[2]
        ref[3,2,1,1]=(-0.277307915098)*x_ref[0]**o+(0.737855779658)*x_ref[0]+(-0.466761948855)*x_ref[1]**o+(-0.821498435809)*x_ref[1]+(-0.391009601442)*x_ref[2]**o+(0.643834542048)*x_ref[2]
        ref[3,2,2,0]=(0.379015007885)*x_ref[0]**o+(0.348646179575)*x_ref[0]+(0.733765133901)*x_ref[1]**o+(-0.722363571789)*x_ref[1]+(0.488531791931)*x_ref[2]**o+(0.279813923679)*x_ref[2]
        ref[3,2,2,1]=(0.516708382119)*x_ref[0]**o+(0.0426665103314)*x_ref[0]+(-0.168751315718)*x_ref[1]**o+(-0.124069236608)*x_ref[1]+(-0.778866001674)*x_ref[2]**o+(-0.362031001279)*x_ref[2]
        ref[3,3,0,0]=(0.995284803465)*x_ref[0]**o+(-0.442327755879)*x_ref[0]+(-0.13110270925)*x_ref[1]**o+(-0.172857543102)*x_ref[1]+(-0.910522664379)*x_ref[2]**o+(0.516931833163)*x_ref[2]
        ref[3,3,0,1]=(-0.889320295626)*x_ref[0]**o+(-0.728330379534)*x_ref[0]+(-0.584091700627)*x_ref[1]**o+(0.469914082337)*x_ref[1]+(0.18874127891)*x_ref[2]**o+(0.0846532039249)*x_ref[2]
        ref[3,3,1,0]=(0.911523535462)*x_ref[0]**o+(0.0924361220354)*x_ref[0]+(-0.0822504494228)*x_ref[1]**o+(0.00630249023591)*x_ref[1]+(-0.180497330322)*x_ref[2]**o+(0.817405577154)*x_ref[2]
        ref[3,3,1,1]=(0.681872747813)*x_ref[0]**o+(0.840680190206)*x_ref[0]+(-0.477660295206)*x_ref[1]**o+(0.0399086968056)*x_ref[1]+(0.440403204247)*x_ref[2]**o+(-0.220656197016)*x_ref[2]
        ref[3,3,2,0]=(-0.967524407336)*x_ref[0]**o+(0.896652274462)*x_ref[0]+(0.735269328519)*x_ref[1]**o+(-0.249017384614)*x_ref[1]+(-0.573222311563)*x_ref[2]**o+(-0.653800334954)*x_ref[2]
        ref[3,3,2,1]=(0.10041632804)*x_ref[0]**o+(0.167657178348)*x_ref[0]+(0.990621134594)*x_ref[1]**o+(-0.309063473348)*x_ref[1]+(-0.688206827575)*x_ref[2]**o+(0.0926806818663)*x_ref[2]
        ref[3,4,0,0]=(-0.128678339518)*x_ref[0]**o+(-0.974238160059)*x_ref[0]+(-0.762599464772)*x_ref[1]**o+(0.0233286358698)*x_ref[1]+(-0.622481697973)*x_ref[2]**o+(-0.080495303798)*x_ref[2]
        ref[3,4,0,1]=(-0.380318230058)*x_ref[0]**o+(0.568425682425)*x_ref[0]+(0.787369916262)*x_ref[1]**o+(0.296745245123)*x_ref[1]+(-0.917621002414)*x_ref[2]**o+(-0.696209404954)*x_ref[2]
        ref[3,4,1,0]=(-0.382961468898)*x_ref[0]**o+(0.672614453621)*x_ref[0]+(0.565371503874)*x_ref[1]**o+(-0.597331999912)*x_ref[1]+(-0.153637846704)*x_ref[2]**o+(-0.438159587906)*x_ref[2]
        ref[3,4,1,1]=(-0.541043007799)*x_ref[0]**o+(0.415600628159)*x_ref[0]+(-0.793716270785)*x_ref[1]**o+(0.780110667203)*x_ref[1]+(-0.101309486185)*x_ref[2]**o+(-0.138868352939)*x_ref[2]
        ref[3,4,2,0]=(0.679815917569)*x_ref[0]**o+(-0.800657537776)*x_ref[0]+(-0.711956069893)*x_ref[1]**o+(-0.575626146908)*x_ref[1]+(0.184247083249)*x_ref[2]**o+(-0.728118084396)*x_ref[2]
        ref[3,4,2,1]=(-0.608089432748)*x_ref[0]**o+(0.356378798828)*x_ref[0]+(-0.575957599599)*x_ref[1]**o+(0.599586825137)*x_ref[1]+(-0.160475743299)*x_ref[2]**o+(-0.070762356177)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.52738736047)*x[0]**o+(-0.513351179641)*x[0]+(0.466974564277)*x[1]**o+(0.247665236281)*x[1]
        ref=(0.52738736047)*x_ref[0]**o+(-0.513351179641)*x_ref[0]+(0.466974564277)*x_ref[1]**o+(0.247665236281)*x_ref[1]
      else:
        arg=(0.345612544612)*x[0]**o+(0.480207854215)*x[0]+(0.196672493786)*x[1]**o+(-0.891043131902)*x[1]+(0.106095952827)*x[2]**o+(0.436049931061)*x[2]
        ref=(0.345612544612)*x_ref[0]**o+(0.480207854215)*x_ref[0]+(0.196672493786)*x_ref[1]**o+(-0.891043131902)*x_ref[1]+(0.106095952827)*x_ref[2]**o+(0.436049931061)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.738262816349)*x[0]**o+(-0.650523560236)*x[0]+(-0.344136056546)*x[1]**o+(-0.0693168099812)*x[1]
        arg[1]=(-0.565224044511)*x[0]**o+(0.94461728731)*x[0]+(0.234800891741)*x[1]**o+(0.836341748852)*x[1]
        ref[0]=(-0.738262816349)*x_ref[0]**o+(-0.650523560236)*x_ref[0]+(-0.344136056546)*x_ref[1]**o+(-0.0693168099812)*x_ref[1]
        ref[1]=(-0.565224044511)*x_ref[0]**o+(0.94461728731)*x_ref[0]+(0.234800891741)*x_ref[1]**o+(0.836341748852)*x_ref[1]
      else:
        arg[0]=(-0.76925788678)*x[0]**o+(-0.391133765534)*x[0]+(0.89961543807)*x[1]**o+(-0.355090830059)*x[1]+(-0.663969382551)*x[2]**o+(0.0324773110608)*x[2]
        arg[1]=(-0.640271335706)*x[0]**o+(-0.0771899290795)*x[0]+(0.923847889924)*x[1]**o+(-0.732484895713)*x[1]+(-0.782800597476)*x[2]**o+(-0.402230990251)*x[2]
        ref[0]=(-0.76925788678)*x_ref[0]**o+(-0.391133765534)*x_ref[0]+(0.89961543807)*x_ref[1]**o+(-0.355090830059)*x_ref[1]+(-0.663969382551)*x_ref[2]**o+(0.0324773110608)*x_ref[2]
        ref[1]=(-0.640271335706)*x_ref[0]**o+(-0.0771899290795)*x_ref[0]+(0.923847889924)*x_ref[1]**o+(-0.732484895713)*x_ref[1]+(-0.782800597476)*x_ref[2]**o+(-0.402230990251)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.631349529615)*x[0]**o+(0.79216890406)*x[0]+(0.835421397222)*x[1]**o+(-0.209368192458)*x[1]
        arg[0,1]=(-0.166321945093)*x[0]**o+(0.701821475941)*x[0]+(-0.141804772049)*x[1]**o+(-0.860056917696)*x[1]
        arg[0,2]=(0.761659971379)*x[0]**o+(-0.39989119778)*x[0]+(-0.761140588562)*x[1]**o+(-0.595816184788)*x[1]
        arg[0,3]=(0.378518095375)*x[0]**o+(-0.880144120304)*x[0]+(0.814409656431)*x[1]**o+(0.0464425920173)*x[1]
        arg[0,4]=(-0.24431693596)*x[0]**o+(-0.840538735294)*x[0]+(0.0157782225732)*x[1]**o+(0.951311376091)*x[1]
        arg[1,0]=(0.0710042795971)*x[0]**o+(0.936106708209)*x[0]+(-0.580035969982)*x[1]**o+(0.740790378647)*x[1]
        arg[1,1]=(-0.807938035414)*x[0]**o+(0.8723236137)*x[0]+(-0.600615307712)*x[1]**o+(0.71483191921)*x[1]
        arg[1,2]=(0.454239081542)*x[0]**o+(0.966475397079)*x[0]+(-0.184649384525)*x[1]**o+(0.417686859721)*x[1]
        arg[1,3]=(-0.394999073966)*x[0]**o+(-0.0445825855756)*x[0]+(0.0389091203001)*x[1]**o+(-0.265569544304)*x[1]
        arg[1,4]=(0.625419856529)*x[0]**o+(-0.297955079048)*x[0]+(0.0092181343855)*x[1]**o+(-0.423768484981)*x[1]
        arg[2,0]=(0.836466109456)*x[0]**o+(-0.12515777587)*x[0]+(-0.0908815675065)*x[1]**o+(0.825332480799)*x[1]
        arg[2,1]=(-0.454612353892)*x[0]**o+(0.952942366642)*x[0]+(0.538638303761)*x[1]**o+(0.377957206795)*x[1]
        arg[2,2]=(-0.229405757447)*x[0]**o+(-0.125249012236)*x[0]+(0.194040765277)*x[1]**o+(-0.77353897054)*x[1]
        arg[2,3]=(-0.220514078279)*x[0]**o+(-0.937043129926)*x[0]+(0.661270617907)*x[1]**o+(0.342981486241)*x[1]
        arg[2,4]=(0.58858161977)*x[0]**o+(0.324813259161)*x[0]+(0.00439371077545)*x[1]**o+(0.517353738278)*x[1]
        arg[3,0]=(0.130001725331)*x[0]**o+(0.512570729718)*x[0]+(0.971292679928)*x[1]**o+(0.302530140373)*x[1]
        arg[3,1]=(0.999138842976)*x[0]**o+(0.325794749172)*x[0]+(0.94900982582)*x[1]**o+(0.56933825455)*x[1]
        arg[3,2]=(0.965201839092)*x[0]**o+(0.434187660328)*x[0]+(0.842238794411)*x[1]**o+(-0.754540335284)*x[1]
        arg[3,3]=(-0.539162954093)*x[0]**o+(-0.373779084363)*x[0]+(-0.426189858865)*x[1]**o+(-0.812351658629)*x[1]
        arg[3,4]=(0.871516777541)*x[0]**o+(-0.357393696308)*x[0]+(0.267358318776)*x[1]**o+(0.796701523616)*x[1]
        ref[0,0]=(-0.631349529615)*x_ref[0]**o+(0.79216890406)*x_ref[0]+(0.835421397222)*x_ref[1]**o+(-0.209368192458)*x_ref[1]
        ref[0,1]=(-0.166321945093)*x_ref[0]**o+(0.701821475941)*x_ref[0]+(-0.141804772049)*x_ref[1]**o+(-0.860056917696)*x_ref[1]
        ref[0,2]=(0.761659971379)*x_ref[0]**o+(-0.39989119778)*x_ref[0]+(-0.761140588562)*x_ref[1]**o+(-0.595816184788)*x_ref[1]
        ref[0,3]=(0.378518095375)*x_ref[0]**o+(-0.880144120304)*x_ref[0]+(0.814409656431)*x_ref[1]**o+(0.0464425920173)*x_ref[1]
        ref[0,4]=(-0.24431693596)*x_ref[0]**o+(-0.840538735294)*x_ref[0]+(0.0157782225732)*x_ref[1]**o+(0.951311376091)*x_ref[1]
        ref[1,0]=(0.0710042795971)*x_ref[0]**o+(0.936106708209)*x_ref[0]+(-0.580035969982)*x_ref[1]**o+(0.740790378647)*x_ref[1]
        ref[1,1]=(-0.807938035414)*x_ref[0]**o+(0.8723236137)*x_ref[0]+(-0.600615307712)*x_ref[1]**o+(0.71483191921)*x_ref[1]
        ref[1,2]=(0.454239081542)*x_ref[0]**o+(0.966475397079)*x_ref[0]+(-0.184649384525)*x_ref[1]**o+(0.417686859721)*x_ref[1]
        ref[1,3]=(-0.394999073966)*x_ref[0]**o+(-0.0445825855756)*x_ref[0]+(0.0389091203001)*x_ref[1]**o+(-0.265569544304)*x_ref[1]
        ref[1,4]=(0.625419856529)*x_ref[0]**o+(-0.297955079048)*x_ref[0]+(0.0092181343855)*x_ref[1]**o+(-0.423768484981)*x_ref[1]
        ref[2,0]=(0.836466109456)*x_ref[0]**o+(-0.12515777587)*x_ref[0]+(-0.0908815675065)*x_ref[1]**o+(0.825332480799)*x_ref[1]
        ref[2,1]=(-0.454612353892)*x_ref[0]**o+(0.952942366642)*x_ref[0]+(0.538638303761)*x_ref[1]**o+(0.377957206795)*x_ref[1]
        ref[2,2]=(-0.229405757447)*x_ref[0]**o+(-0.125249012236)*x_ref[0]+(0.194040765277)*x_ref[1]**o+(-0.77353897054)*x_ref[1]
        ref[2,3]=(-0.220514078279)*x_ref[0]**o+(-0.937043129926)*x_ref[0]+(0.661270617907)*x_ref[1]**o+(0.342981486241)*x_ref[1]
        ref[2,4]=(0.58858161977)*x_ref[0]**o+(0.324813259161)*x_ref[0]+(0.00439371077545)*x_ref[1]**o+(0.517353738278)*x_ref[1]
        ref[3,0]=(0.130001725331)*x_ref[0]**o+(0.512570729718)*x_ref[0]+(0.971292679928)*x_ref[1]**o+(0.302530140373)*x_ref[1]
        ref[3,1]=(0.999138842976)*x_ref[0]**o+(0.325794749172)*x_ref[0]+(0.94900982582)*x_ref[1]**o+(0.56933825455)*x_ref[1]
        ref[3,2]=(0.965201839092)*x_ref[0]**o+(0.434187660328)*x_ref[0]+(0.842238794411)*x_ref[1]**o+(-0.754540335284)*x_ref[1]
        ref[3,3]=(-0.539162954093)*x_ref[0]**o+(-0.373779084363)*x_ref[0]+(-0.426189858865)*x_ref[1]**o+(-0.812351658629)*x_ref[1]
        ref[3,4]=(0.871516777541)*x_ref[0]**o+(-0.357393696308)*x_ref[0]+(0.267358318776)*x_ref[1]**o+(0.796701523616)*x_ref[1]
      else:
        arg[0,0]=(-0.811142944724)*x[0]**o+(-0.65659407492)*x[0]+(-0.534808740191)*x[1]**o+(0.602252100451)*x[1]+(-0.158845636056)*x[2]**o+(-0.781826300399)*x[2]
        arg[0,1]=(-0.396870942248)*x[0]**o+(0.285698519063)*x[0]+(-0.962418369859)*x[1]**o+(-0.729381989807)*x[1]+(-0.0500038851534)*x[2]**o+(-0.567828852526)*x[2]
        arg[0,2]=(0.346399694372)*x[0]**o+(-0.170517233373)*x[0]+(-0.443275920724)*x[1]**o+(0.581464075295)*x[1]+(0.197230389253)*x[2]**o+(-0.613092248913)*x[2]
        arg[0,3]=(0.920174147113)*x[0]**o+(0.429847401539)*x[0]+(0.112888577662)*x[1]**o+(-0.0540034503714)*x[1]+(-0.0323244631044)*x[2]**o+(-0.293609544741)*x[2]
        arg[0,4]=(-0.0177585560206)*x[0]**o+(-0.806033755558)*x[0]+(-0.363921399918)*x[1]**o+(-0.415721267967)*x[1]+(0.832302676307)*x[2]**o+(-0.102510351741)*x[2]
        arg[1,0]=(-0.769760277763)*x[0]**o+(0.818187369213)*x[0]+(0.454400795864)*x[1]**o+(-0.220119708184)*x[1]+(-0.637117949075)*x[2]**o+(-0.661398404641)*x[2]
        arg[1,1]=(0.828424952723)*x[0]**o+(0.677022266764)*x[0]+(-0.958176510992)*x[1]**o+(-0.732014776269)*x[1]+(-0.495308153354)*x[2]**o+(-0.30698327016)*x[2]
        arg[1,2]=(-0.454106379068)*x[0]**o+(0.686920350455)*x[0]+(0.560344054206)*x[1]**o+(0.355495451775)*x[1]+(0.256969568039)*x[2]**o+(-0.737279585211)*x[2]
        arg[1,3]=(-0.976488579746)*x[0]**o+(-0.652812816758)*x[0]+(-0.996243459964)*x[1]**o+(0.631565668939)*x[1]+(0.996259991179)*x[2]**o+(-0.796610599165)*x[2]
        arg[1,4]=(0.676024716135)*x[0]**o+(-0.179835226483)*x[0]+(0.497541872254)*x[1]**o+(-0.678345104068)*x[1]+(0.945491022682)*x[2]**o+(-0.358403760669)*x[2]
        arg[2,0]=(0.166901676041)*x[0]**o+(0.833434173485)*x[0]+(0.622153751777)*x[1]**o+(0.767679765059)*x[1]+(-0.82271335395)*x[2]**o+(0.667651471482)*x[2]
        arg[2,1]=(-0.405015366041)*x[0]**o+(0.0548291583154)*x[0]+(0.894878845805)*x[1]**o+(-0.0955982667798)*x[1]+(0.468927799304)*x[2]**o+(0.547421481564)*x[2]
        arg[2,2]=(0.893319250823)*x[0]**o+(0.771073524295)*x[0]+(0.993876715991)*x[1]**o+(0.0832975372419)*x[1]+(-0.717061635157)*x[2]**o+(0.303760934227)*x[2]
        arg[2,3]=(0.247006018272)*x[0]**o+(0.329618310414)*x[0]+(-0.824176211089)*x[1]**o+(-0.678238702323)*x[1]+(0.218736302876)*x[2]**o+(0.8652063227)*x[2]
        arg[2,4]=(0.502380810223)*x[0]**o+(-0.94917091185)*x[0]+(-0.0392659831743)*x[1]**o+(-0.308291910503)*x[1]+(-0.498112428439)*x[2]**o+(0.217915791465)*x[2]
        arg[3,0]=(-0.018497051159)*x[0]**o+(-0.543005709743)*x[0]+(-0.433773552307)*x[1]**o+(-0.604850191015)*x[1]+(0.285970270117)*x[2]**o+(0.150938308283)*x[2]
        arg[3,1]=(0.185537796564)*x[0]**o+(0.663947686119)*x[0]+(-0.846597672873)*x[1]**o+(-0.798399069258)*x[1]+(0.0215106260597)*x[2]**o+(-0.388760513164)*x[2]
        arg[3,2]=(-0.378320218313)*x[0]**o+(-0.385079307834)*x[0]+(-0.392389046498)*x[1]**o+(0.323305501943)*x[1]+(0.496285813211)*x[2]**o+(-0.0269835921214)*x[2]
        arg[3,3]=(0.779877220393)*x[0]**o+(0.968536241033)*x[0]+(0.786166009582)*x[1]**o+(-0.974838905578)*x[1]+(0.4579382269)*x[2]**o+(0.178464213473)*x[2]
        arg[3,4]=(-0.149895758057)*x[0]**o+(0.3529889311)*x[0]+(0.223924981441)*x[1]**o+(0.10143440581)*x[1]+(0.162222746832)*x[2]**o+(0.421433640986)*x[2]
        ref[0,0]=(-0.811142944724)*x_ref[0]**o+(-0.65659407492)*x_ref[0]+(-0.534808740191)*x_ref[1]**o+(0.602252100451)*x_ref[1]+(-0.158845636056)*x_ref[2]**o+(-0.781826300399)*x_ref[2]
        ref[0,1]=(-0.396870942248)*x_ref[0]**o+(0.285698519063)*x_ref[0]+(-0.962418369859)*x_ref[1]**o+(-0.729381989807)*x_ref[1]+(-0.0500038851534)*x_ref[2]**o+(-0.567828852526)*x_ref[2]
        ref[0,2]=(0.346399694372)*x_ref[0]**o+(-0.170517233373)*x_ref[0]+(-0.443275920724)*x_ref[1]**o+(0.581464075295)*x_ref[1]+(0.197230389253)*x_ref[2]**o+(-0.613092248913)*x_ref[2]
        ref[0,3]=(0.920174147113)*x_ref[0]**o+(0.429847401539)*x_ref[0]+(0.112888577662)*x_ref[1]**o+(-0.0540034503714)*x_ref[1]+(-0.0323244631044)*x_ref[2]**o+(-0.293609544741)*x_ref[2]
        ref[0,4]=(-0.0177585560206)*x_ref[0]**o+(-0.806033755558)*x_ref[0]+(-0.363921399918)*x_ref[1]**o+(-0.415721267967)*x_ref[1]+(0.832302676307)*x_ref[2]**o+(-0.102510351741)*x_ref[2]
        ref[1,0]=(-0.769760277763)*x_ref[0]**o+(0.818187369213)*x_ref[0]+(0.454400795864)*x_ref[1]**o+(-0.220119708184)*x_ref[1]+(-0.637117949075)*x_ref[2]**o+(-0.661398404641)*x_ref[2]
        ref[1,1]=(0.828424952723)*x_ref[0]**o+(0.677022266764)*x_ref[0]+(-0.958176510992)*x_ref[1]**o+(-0.732014776269)*x_ref[1]+(-0.495308153354)*x_ref[2]**o+(-0.30698327016)*x_ref[2]
        ref[1,2]=(-0.454106379068)*x_ref[0]**o+(0.686920350455)*x_ref[0]+(0.560344054206)*x_ref[1]**o+(0.355495451775)*x_ref[1]+(0.256969568039)*x_ref[2]**o+(-0.737279585211)*x_ref[2]
        ref[1,3]=(-0.976488579746)*x_ref[0]**o+(-0.652812816758)*x_ref[0]+(-0.996243459964)*x_ref[1]**o+(0.631565668939)*x_ref[1]+(0.996259991179)*x_ref[2]**o+(-0.796610599165)*x_ref[2]
        ref[1,4]=(0.676024716135)*x_ref[0]**o+(-0.179835226483)*x_ref[0]+(0.497541872254)*x_ref[1]**o+(-0.678345104068)*x_ref[1]+(0.945491022682)*x_ref[2]**o+(-0.358403760669)*x_ref[2]
        ref[2,0]=(0.166901676041)*x_ref[0]**o+(0.833434173485)*x_ref[0]+(0.622153751777)*x_ref[1]**o+(0.767679765059)*x_ref[1]+(-0.82271335395)*x_ref[2]**o+(0.667651471482)*x_ref[2]
        ref[2,1]=(-0.405015366041)*x_ref[0]**o+(0.0548291583154)*x_ref[0]+(0.894878845805)*x_ref[1]**o+(-0.0955982667798)*x_ref[1]+(0.468927799304)*x_ref[2]**o+(0.547421481564)*x_ref[2]
        ref[2,2]=(0.893319250823)*x_ref[0]**o+(0.771073524295)*x_ref[0]+(0.993876715991)*x_ref[1]**o+(0.0832975372419)*x_ref[1]+(-0.717061635157)*x_ref[2]**o+(0.303760934227)*x_ref[2]
        ref[2,3]=(0.247006018272)*x_ref[0]**o+(0.329618310414)*x_ref[0]+(-0.824176211089)*x_ref[1]**o+(-0.678238702323)*x_ref[1]+(0.218736302876)*x_ref[2]**o+(0.8652063227)*x_ref[2]
        ref[2,4]=(0.502380810223)*x_ref[0]**o+(-0.94917091185)*x_ref[0]+(-0.0392659831743)*x_ref[1]**o+(-0.308291910503)*x_ref[1]+(-0.498112428439)*x_ref[2]**o+(0.217915791465)*x_ref[2]
        ref[3,0]=(-0.018497051159)*x_ref[0]**o+(-0.543005709743)*x_ref[0]+(-0.433773552307)*x_ref[1]**o+(-0.604850191015)*x_ref[1]+(0.285970270117)*x_ref[2]**o+(0.150938308283)*x_ref[2]
        ref[3,1]=(0.185537796564)*x_ref[0]**o+(0.663947686119)*x_ref[0]+(-0.846597672873)*x_ref[1]**o+(-0.798399069258)*x_ref[1]+(0.0215106260597)*x_ref[2]**o+(-0.388760513164)*x_ref[2]
        ref[3,2]=(-0.378320218313)*x_ref[0]**o+(-0.385079307834)*x_ref[0]+(-0.392389046498)*x_ref[1]**o+(0.323305501943)*x_ref[1]+(0.496285813211)*x_ref[2]**o+(-0.0269835921214)*x_ref[2]
        ref[3,3]=(0.779877220393)*x_ref[0]**o+(0.968536241033)*x_ref[0]+(0.786166009582)*x_ref[1]**o+(-0.974838905578)*x_ref[1]+(0.4579382269)*x_ref[2]**o+(0.178464213473)*x_ref[2]
        ref[3,4]=(-0.149895758057)*x_ref[0]**o+(0.3529889311)*x_ref[0]+(0.223924981441)*x_ref[1]**o+(0.10143440581)*x_ref[1]+(0.162222746832)*x_ref[2]**o+(0.421433640986)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.232971584272)*x[0]**o+(0.391351964909)*x[0]+(-0.832254448983)*x[1]**o+(0.880762034713)*x[1]
        arg[0,0,1]=(0.473097471101)*x[0]**o+(-0.660862772376)*x[0]+(0.72974853865)*x[1]**o+(-0.328183435687)*x[1]
        arg[0,1,0]=(-0.749483736482)*x[0]**o+(0.0872454020187)*x[0]+(0.6916024322)*x[1]**o+(0.82757019273)*x[1]
        arg[0,1,1]=(-0.325601366346)*x[0]**o+(0.918246327277)*x[0]+(0.855963838252)*x[1]**o+(0.448567464706)*x[1]
        arg[1,0,0]=(-0.6642143628)*x[0]**o+(-0.82649833447)*x[0]+(0.822778465465)*x[1]**o+(-0.926669226842)*x[1]
        arg[1,0,1]=(-0.644490899373)*x[0]**o+(0.0644054772987)*x[0]+(-0.914922609343)*x[1]**o+(0.898880463842)*x[1]
        arg[1,1,0]=(0.40963096867)*x[0]**o+(-0.347418076162)*x[0]+(0.857966073118)*x[1]**o+(-0.284521442085)*x[1]
        arg[1,1,1]=(-0.00998062171478)*x[0]**o+(-0.234122404174)*x[0]+(-0.149500951208)*x[1]**o+(0.680038328937)*x[1]
        arg[2,0,0]=(-0.963777985833)*x[0]**o+(0.710727608341)*x[0]+(-0.646898391724)*x[1]**o+(-0.464095155503)*x[1]
        arg[2,0,1]=(-0.617670394851)*x[0]**o+(0.592337321587)*x[0]+(-0.0881426217074)*x[1]**o+(-0.00431912882662)*x[1]
        arg[2,1,0]=(0.871279828121)*x[0]**o+(-0.858267190204)*x[0]+(0.440863109038)*x[1]**o+(-0.912521398099)*x[1]
        arg[2,1,1]=(0.17802472174)*x[0]**o+(0.948587960042)*x[0]+(-0.837945850976)*x[1]**o+(0.13308730872)*x[1]
        arg[3,0,0]=(0.0482409149433)*x[0]**o+(-0.198100088194)*x[0]+(0.504789334224)*x[1]**o+(0.296276808608)*x[1]
        arg[3,0,1]=(-0.30515363985)*x[0]**o+(-0.355293167256)*x[0]+(-0.577592997685)*x[1]**o+(-0.882408725943)*x[1]
        arg[3,1,0]=(0.76102937294)*x[0]**o+(-0.210169673144)*x[0]+(-0.181755140891)*x[1]**o+(0.865448484187)*x[1]
        arg[3,1,1]=(-0.894294091863)*x[0]**o+(-0.712245666103)*x[0]+(0.151657455329)*x[1]**o+(-0.54019436763)*x[1]
        arg[4,0,0]=(0.860662861318)*x[0]**o+(-0.36787473166)*x[0]+(0.240406823985)*x[1]**o+(-0.381822992473)*x[1]
        arg[4,0,1]=(-0.109197125112)*x[0]**o+(0.962595232385)*x[0]+(-0.323876506847)*x[1]**o+(0.178613363564)*x[1]
        arg[4,1,0]=(0.486980313424)*x[0]**o+(-0.611294566285)*x[0]+(-0.774813150316)*x[1]**o+(0.335411840405)*x[1]
        arg[4,1,1]=(-0.622163740867)*x[0]**o+(0.426305234663)*x[0]+(0.243557551101)*x[1]**o+(0.930601340004)*x[1]
        arg[5,0,0]=(-0.232795962027)*x[0]**o+(0.49010767268)*x[0]+(-0.503693305594)*x[1]**o+(-0.915931668186)*x[1]
        arg[5,0,1]=(-0.304614759138)*x[0]**o+(0.797445765321)*x[0]+(-0.00820825778888)*x[1]**o+(-0.196762463972)*x[1]
        arg[5,1,0]=(-0.516700339456)*x[0]**o+(-0.440953905414)*x[0]+(0.875266018165)*x[1]**o+(0.780895354543)*x[1]
        arg[5,1,1]=(0.219827931399)*x[0]**o+(-0.950514094131)*x[0]+(-0.120763970425)*x[1]**o+(0.0941571726596)*x[1]
        ref[0,0,0]=(-0.232971584272)*x_ref[0]**o+(0.391351964909)*x_ref[0]+(-0.832254448983)*x_ref[1]**o+(0.880762034713)*x_ref[1]
        ref[0,0,1]=(0.473097471101)*x_ref[0]**o+(-0.660862772376)*x_ref[0]+(0.72974853865)*x_ref[1]**o+(-0.328183435687)*x_ref[1]
        ref[0,1,0]=(-0.749483736482)*x_ref[0]**o+(0.0872454020187)*x_ref[0]+(0.6916024322)*x_ref[1]**o+(0.82757019273)*x_ref[1]
        ref[0,1,1]=(-0.325601366346)*x_ref[0]**o+(0.918246327277)*x_ref[0]+(0.855963838252)*x_ref[1]**o+(0.448567464706)*x_ref[1]
        ref[1,0,0]=(-0.6642143628)*x_ref[0]**o+(-0.82649833447)*x_ref[0]+(0.822778465465)*x_ref[1]**o+(-0.926669226842)*x_ref[1]
        ref[1,0,1]=(-0.644490899373)*x_ref[0]**o+(0.0644054772987)*x_ref[0]+(-0.914922609343)*x_ref[1]**o+(0.898880463842)*x_ref[1]
        ref[1,1,0]=(0.40963096867)*x_ref[0]**o+(-0.347418076162)*x_ref[0]+(0.857966073118)*x_ref[1]**o+(-0.284521442085)*x_ref[1]
        ref[1,1,1]=(-0.00998062171478)*x_ref[0]**o+(-0.234122404174)*x_ref[0]+(-0.149500951208)*x_ref[1]**o+(0.680038328937)*x_ref[1]
        ref[2,0,0]=(-0.963777985833)*x_ref[0]**o+(0.710727608341)*x_ref[0]+(-0.646898391724)*x_ref[1]**o+(-0.464095155503)*x_ref[1]
        ref[2,0,1]=(-0.617670394851)*x_ref[0]**o+(0.592337321587)*x_ref[0]+(-0.0881426217074)*x_ref[1]**o+(-0.00431912882662)*x_ref[1]
        ref[2,1,0]=(0.871279828121)*x_ref[0]**o+(-0.858267190204)*x_ref[0]+(0.440863109038)*x_ref[1]**o+(-0.912521398099)*x_ref[1]
        ref[2,1,1]=(0.17802472174)*x_ref[0]**o+(0.948587960042)*x_ref[0]+(-0.837945850976)*x_ref[1]**o+(0.13308730872)*x_ref[1]
        ref[3,0,0]=(0.0482409149433)*x_ref[0]**o+(-0.198100088194)*x_ref[0]+(0.504789334224)*x_ref[1]**o+(0.296276808608)*x_ref[1]
        ref[3,0,1]=(-0.30515363985)*x_ref[0]**o+(-0.355293167256)*x_ref[0]+(-0.577592997685)*x_ref[1]**o+(-0.882408725943)*x_ref[1]
        ref[3,1,0]=(0.76102937294)*x_ref[0]**o+(-0.210169673144)*x_ref[0]+(-0.181755140891)*x_ref[1]**o+(0.865448484187)*x_ref[1]
        ref[3,1,1]=(-0.894294091863)*x_ref[0]**o+(-0.712245666103)*x_ref[0]+(0.151657455329)*x_ref[1]**o+(-0.54019436763)*x_ref[1]
        ref[4,0,0]=(0.860662861318)*x_ref[0]**o+(-0.36787473166)*x_ref[0]+(0.240406823985)*x_ref[1]**o+(-0.381822992473)*x_ref[1]
        ref[4,0,1]=(-0.109197125112)*x_ref[0]**o+(0.962595232385)*x_ref[0]+(-0.323876506847)*x_ref[1]**o+(0.178613363564)*x_ref[1]
        ref[4,1,0]=(0.486980313424)*x_ref[0]**o+(-0.611294566285)*x_ref[0]+(-0.774813150316)*x_ref[1]**o+(0.335411840405)*x_ref[1]
        ref[4,1,1]=(-0.622163740867)*x_ref[0]**o+(0.426305234663)*x_ref[0]+(0.243557551101)*x_ref[1]**o+(0.930601340004)*x_ref[1]
        ref[5,0,0]=(-0.232795962027)*x_ref[0]**o+(0.49010767268)*x_ref[0]+(-0.503693305594)*x_ref[1]**o+(-0.915931668186)*x_ref[1]
        ref[5,0,1]=(-0.304614759138)*x_ref[0]**o+(0.797445765321)*x_ref[0]+(-0.00820825778888)*x_ref[1]**o+(-0.196762463972)*x_ref[1]
        ref[5,1,0]=(-0.516700339456)*x_ref[0]**o+(-0.440953905414)*x_ref[0]+(0.875266018165)*x_ref[1]**o+(0.780895354543)*x_ref[1]
        ref[5,1,1]=(0.219827931399)*x_ref[0]**o+(-0.950514094131)*x_ref[0]+(-0.120763970425)*x_ref[1]**o+(0.0941571726596)*x_ref[1]
      else:
        arg[0,0,0]=(0.772105181243)*x[0]**o+(-0.548341656718)*x[0]+(0.358327458717)*x[1]**o+(0.07838537297)*x[1]+(-0.221432264133)*x[2]**o+(-0.0831433214109)*x[2]
        arg[0,0,1]=(-0.920280406855)*x[0]**o+(0.288008584982)*x[0]+(0.273175197348)*x[1]**o+(0.71005891586)*x[1]+(-0.372322860924)*x[2]**o+(0.801820758328)*x[2]
        arg[0,1,0]=(0.610498050037)*x[0]**o+(-0.539393441621)*x[0]+(0.587033652815)*x[1]**o+(0.237738229824)*x[1]+(-0.79411765083)*x[2]**o+(-0.519989054973)*x[2]
        arg[0,1,1]=(0.5412774887)*x[0]**o+(-0.851503637634)*x[0]+(-0.71956793056)*x[1]**o+(0.166039550456)*x[1]+(-0.274290267921)*x[2]**o+(-0.909146653718)*x[2]
        arg[1,0,0]=(-0.00458109759775)*x[0]**o+(0.487263914082)*x[0]+(0.0171216930245)*x[1]**o+(0.691063106679)*x[1]+(-0.96202326744)*x[2]**o+(0.361419732405)*x[2]
        arg[1,0,1]=(-0.784311451677)*x[0]**o+(-0.862560770078)*x[0]+(-0.0702620499924)*x[1]**o+(0.183400389223)*x[1]+(0.364905596839)*x[2]**o+(0.874230430841)*x[2]
        arg[1,1,0]=(0.315952660065)*x[0]**o+(-0.0168150653709)*x[0]+(0.372807793658)*x[1]**o+(0.350959428104)*x[1]+(0.292074907124)*x[2]**o+(0.38176271099)*x[2]
        arg[1,1,1]=(-0.941288490366)*x[0]**o+(0.159864446187)*x[0]+(0.616516779388)*x[1]**o+(0.0940731931915)*x[1]+(-0.966416702566)*x[2]**o+(-0.716669694143)*x[2]
        arg[2,0,0]=(0.308619329706)*x[0]**o+(0.68203622136)*x[0]+(-0.732163723834)*x[1]**o+(-0.222214324226)*x[1]+(0.89540478356)*x[2]**o+(-0.0648727721753)*x[2]
        arg[2,0,1]=(-0.829002921476)*x[0]**o+(0.274738860807)*x[0]+(0.944302912883)*x[1]**o+(0.67434413206)*x[1]+(-0.861085567199)*x[2]**o+(0.994494977369)*x[2]
        arg[2,1,0]=(-0.871499634896)*x[0]**o+(-0.716060125296)*x[0]+(-0.655428500508)*x[1]**o+(-0.478135345667)*x[1]+(-0.858927907345)*x[2]**o+(0.718884168928)*x[2]
        arg[2,1,1]=(0.843612813106)*x[0]**o+(-0.221966655825)*x[0]+(-0.0770757351)*x[1]**o+(-0.678171902341)*x[1]+(0.814209490483)*x[2]**o+(0.409183122676)*x[2]
        arg[3,0,0]=(-0.226810706884)*x[0]**o+(-0.167181963701)*x[0]+(0.810478050432)*x[1]**o+(-0.335099187817)*x[1]+(0.210621363393)*x[2]**o+(0.429297116788)*x[2]
        arg[3,0,1]=(0.597117204298)*x[0]**o+(-0.000147371125322)*x[0]+(-0.557842700291)*x[1]**o+(0.329529225331)*x[1]+(-0.208265244071)*x[2]**o+(-0.977698555357)*x[2]
        arg[3,1,0]=(-0.163068070947)*x[0]**o+(0.640075503516)*x[0]+(0.425653990651)*x[1]**o+(0.650655294789)*x[1]+(-0.342488862403)*x[2]**o+(0.227438788286)*x[2]
        arg[3,1,1]=(-0.511981920097)*x[0]**o+(-0.18099315756)*x[0]+(-0.969881140182)*x[1]**o+(-0.418742703662)*x[1]+(-0.271422367765)*x[2]**o+(0.0812058381877)*x[2]
        arg[4,0,0]=(0.569514749505)*x[0]**o+(-0.711561480184)*x[0]+(-0.587159226794)*x[1]**o+(-0.240459786489)*x[1]+(-0.380983629494)*x[2]**o+(-0.550385464258)*x[2]
        arg[4,0,1]=(0.257375332759)*x[0]**o+(-0.477083226492)*x[0]+(0.0180233249756)*x[1]**o+(0.612470082967)*x[1]+(0.99566078545)*x[2]**o+(-0.551544637583)*x[2]
        arg[4,1,0]=(0.702787194057)*x[0]**o+(0.922262573534)*x[0]+(-0.465910368685)*x[1]**o+(0.748743004032)*x[1]+(0.642852066952)*x[2]**o+(0.24283802317)*x[2]
        arg[4,1,1]=(0.745762578199)*x[0]**o+(0.240631495429)*x[0]+(0.413077167171)*x[1]**o+(0.692304442104)*x[1]+(0.150058687546)*x[2]**o+(-0.133413399526)*x[2]
        arg[5,0,0]=(0.0244935245274)*x[0]**o+(-0.866653903557)*x[0]+(-0.682252235104)*x[1]**o+(-0.525092961647)*x[1]+(-0.87552740837)*x[2]**o+(-0.780785295196)*x[2]
        arg[5,0,1]=(0.403317147145)*x[0]**o+(-0.846638525869)*x[0]+(-0.798622632135)*x[1]**o+(-0.199698251819)*x[1]+(-0.852399643056)*x[2]**o+(-0.0618243339514)*x[2]
        arg[5,1,0]=(-0.94578445284)*x[0]**o+(-0.715570943968)*x[0]+(-0.193533970382)*x[1]**o+(0.0325400011289)*x[1]+(0.137922635356)*x[2]**o+(0.0192967461384)*x[2]
        arg[5,1,1]=(-0.706542583366)*x[0]**o+(0.412536423598)*x[0]+(-0.27322227526)*x[1]**o+(0.339116123776)*x[1]+(-0.856871339939)*x[2]**o+(0.140603890427)*x[2]
        ref[0,0,0]=(0.772105181243)*x_ref[0]**o+(-0.548341656718)*x_ref[0]+(0.358327458717)*x_ref[1]**o+(0.07838537297)*x_ref[1]+(-0.221432264133)*x_ref[2]**o+(-0.0831433214109)*x_ref[2]
        ref[0,0,1]=(-0.920280406855)*x_ref[0]**o+(0.288008584982)*x_ref[0]+(0.273175197348)*x_ref[1]**o+(0.71005891586)*x_ref[1]+(-0.372322860924)*x_ref[2]**o+(0.801820758328)*x_ref[2]
        ref[0,1,0]=(0.610498050037)*x_ref[0]**o+(-0.539393441621)*x_ref[0]+(0.587033652815)*x_ref[1]**o+(0.237738229824)*x_ref[1]+(-0.79411765083)*x_ref[2]**o+(-0.519989054973)*x_ref[2]
        ref[0,1,1]=(0.5412774887)*x_ref[0]**o+(-0.851503637634)*x_ref[0]+(-0.71956793056)*x_ref[1]**o+(0.166039550456)*x_ref[1]+(-0.274290267921)*x_ref[2]**o+(-0.909146653718)*x_ref[2]
        ref[1,0,0]=(-0.00458109759775)*x_ref[0]**o+(0.487263914082)*x_ref[0]+(0.0171216930245)*x_ref[1]**o+(0.691063106679)*x_ref[1]+(-0.96202326744)*x_ref[2]**o+(0.361419732405)*x_ref[2]
        ref[1,0,1]=(-0.784311451677)*x_ref[0]**o+(-0.862560770078)*x_ref[0]+(-0.0702620499924)*x_ref[1]**o+(0.183400389223)*x_ref[1]+(0.364905596839)*x_ref[2]**o+(0.874230430841)*x_ref[2]
        ref[1,1,0]=(0.315952660065)*x_ref[0]**o+(-0.0168150653709)*x_ref[0]+(0.372807793658)*x_ref[1]**o+(0.350959428104)*x_ref[1]+(0.292074907124)*x_ref[2]**o+(0.38176271099)*x_ref[2]
        ref[1,1,1]=(-0.941288490366)*x_ref[0]**o+(0.159864446187)*x_ref[0]+(0.616516779388)*x_ref[1]**o+(0.0940731931915)*x_ref[1]+(-0.966416702566)*x_ref[2]**o+(-0.716669694143)*x_ref[2]
        ref[2,0,0]=(0.308619329706)*x_ref[0]**o+(0.68203622136)*x_ref[0]+(-0.732163723834)*x_ref[1]**o+(-0.222214324226)*x_ref[1]+(0.89540478356)*x_ref[2]**o+(-0.0648727721753)*x_ref[2]
        ref[2,0,1]=(-0.829002921476)*x_ref[0]**o+(0.274738860807)*x_ref[0]+(0.944302912883)*x_ref[1]**o+(0.67434413206)*x_ref[1]+(-0.861085567199)*x_ref[2]**o+(0.994494977369)*x_ref[2]
        ref[2,1,0]=(-0.871499634896)*x_ref[0]**o+(-0.716060125296)*x_ref[0]+(-0.655428500508)*x_ref[1]**o+(-0.478135345667)*x_ref[1]+(-0.858927907345)*x_ref[2]**o+(0.718884168928)*x_ref[2]
        ref[2,1,1]=(0.843612813106)*x_ref[0]**o+(-0.221966655825)*x_ref[0]+(-0.0770757351)*x_ref[1]**o+(-0.678171902341)*x_ref[1]+(0.814209490483)*x_ref[2]**o+(0.409183122676)*x_ref[2]
        ref[3,0,0]=(-0.226810706884)*x_ref[0]**o+(-0.167181963701)*x_ref[0]+(0.810478050432)*x_ref[1]**o+(-0.335099187817)*x_ref[1]+(0.210621363393)*x_ref[2]**o+(0.429297116788)*x_ref[2]
        ref[3,0,1]=(0.597117204298)*x_ref[0]**o+(-0.000147371125322)*x_ref[0]+(-0.557842700291)*x_ref[1]**o+(0.329529225331)*x_ref[1]+(-0.208265244071)*x_ref[2]**o+(-0.977698555357)*x_ref[2]
        ref[3,1,0]=(-0.163068070947)*x_ref[0]**o+(0.640075503516)*x_ref[0]+(0.425653990651)*x_ref[1]**o+(0.650655294789)*x_ref[1]+(-0.342488862403)*x_ref[2]**o+(0.227438788286)*x_ref[2]
        ref[3,1,1]=(-0.511981920097)*x_ref[0]**o+(-0.18099315756)*x_ref[0]+(-0.969881140182)*x_ref[1]**o+(-0.418742703662)*x_ref[1]+(-0.271422367765)*x_ref[2]**o+(0.0812058381877)*x_ref[2]
        ref[4,0,0]=(0.569514749505)*x_ref[0]**o+(-0.711561480184)*x_ref[0]+(-0.587159226794)*x_ref[1]**o+(-0.240459786489)*x_ref[1]+(-0.380983629494)*x_ref[2]**o+(-0.550385464258)*x_ref[2]
        ref[4,0,1]=(0.257375332759)*x_ref[0]**o+(-0.477083226492)*x_ref[0]+(0.0180233249756)*x_ref[1]**o+(0.612470082967)*x_ref[1]+(0.99566078545)*x_ref[2]**o+(-0.551544637583)*x_ref[2]
        ref[4,1,0]=(0.702787194057)*x_ref[0]**o+(0.922262573534)*x_ref[0]+(-0.465910368685)*x_ref[1]**o+(0.748743004032)*x_ref[1]+(0.642852066952)*x_ref[2]**o+(0.24283802317)*x_ref[2]
        ref[4,1,1]=(0.745762578199)*x_ref[0]**o+(0.240631495429)*x_ref[0]+(0.413077167171)*x_ref[1]**o+(0.692304442104)*x_ref[1]+(0.150058687546)*x_ref[2]**o+(-0.133413399526)*x_ref[2]
        ref[5,0,0]=(0.0244935245274)*x_ref[0]**o+(-0.866653903557)*x_ref[0]+(-0.682252235104)*x_ref[1]**o+(-0.525092961647)*x_ref[1]+(-0.87552740837)*x_ref[2]**o+(-0.780785295196)*x_ref[2]
        ref[5,0,1]=(0.403317147145)*x_ref[0]**o+(-0.846638525869)*x_ref[0]+(-0.798622632135)*x_ref[1]**o+(-0.199698251819)*x_ref[1]+(-0.852399643056)*x_ref[2]**o+(-0.0618243339514)*x_ref[2]
        ref[5,1,0]=(-0.94578445284)*x_ref[0]**o+(-0.715570943968)*x_ref[0]+(-0.193533970382)*x_ref[1]**o+(0.0325400011289)*x_ref[1]+(0.137922635356)*x_ref[2]**o+(0.0192967461384)*x_ref[2]
        ref[5,1,1]=(-0.706542583366)*x_ref[0]**o+(0.412536423598)*x_ref[0]+(-0.27322227526)*x_ref[1]**o+(0.339116123776)*x_ref[1]+(-0.856871339939)*x_ref[2]**o+(0.140603890427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0577426785287)*x[0]**o+(0.469845292021)*x[0]+(-0.641706722819)*x[1]**o+(0.129652001842)*x[1]
        arg[0,0,0,1]=(-0.87686951076)*x[0]**o+(-0.981093040999)*x[0]+(0.654637319552)*x[1]**o+(-0.953104888811)*x[1]
        arg[0,0,1,0]=(-0.486386436415)*x[0]**o+(0.91083470956)*x[0]+(-0.392432937401)*x[1]**o+(0.494919250884)*x[1]
        arg[0,0,1,1]=(0.299928502076)*x[0]**o+(-0.0787848191157)*x[0]+(-0.541863592475)*x[1]**o+(0.435934816687)*x[1]
        arg[0,0,2,0]=(-0.0500083438365)*x[0]**o+(-0.882581886623)*x[0]+(-0.965347357063)*x[1]**o+(-0.270254041774)*x[1]
        arg[0,0,2,1]=(-0.18974861517)*x[0]**o+(0.341406736671)*x[0]+(0.847020774229)*x[1]**o+(0.650201215849)*x[1]
        arg[0,1,0,0]=(0.0823017084719)*x[0]**o+(0.964585359235)*x[0]+(0.357658660436)*x[1]**o+(0.182585271812)*x[1]
        arg[0,1,0,1]=(0.863411821905)*x[0]**o+(0.313374566426)*x[0]+(-0.273547644827)*x[1]**o+(0.429051854479)*x[1]
        arg[0,1,1,0]=(0.167608645068)*x[0]**o+(0.926908094967)*x[0]+(-0.853415240465)*x[1]**o+(-0.376634525241)*x[1]
        arg[0,1,1,1]=(-0.555491961591)*x[0]**o+(-0.64621137308)*x[0]+(-0.315877124991)*x[1]**o+(-0.962839286133)*x[1]
        arg[0,1,2,0]=(0.909012065667)*x[0]**o+(-0.656739022976)*x[0]+(-0.458377773547)*x[1]**o+(0.394223574009)*x[1]
        arg[0,1,2,1]=(-0.550560496714)*x[0]**o+(0.216941875936)*x[0]+(-0.634745794924)*x[1]**o+(0.887912543217)*x[1]
        arg[0,2,0,0]=(-0.961535901493)*x[0]**o+(0.980336384835)*x[0]+(-0.151755000128)*x[1]**o+(0.0402841925797)*x[1]
        arg[0,2,0,1]=(0.571527088727)*x[0]**o+(-0.475013184811)*x[0]+(0.80051818192)*x[1]**o+(-0.589499747694)*x[1]
        arg[0,2,1,0]=(0.545840648612)*x[0]**o+(0.888313209612)*x[0]+(-0.568938733446)*x[1]**o+(-0.289030374342)*x[1]
        arg[0,2,1,1]=(-0.0631638607979)*x[0]**o+(-0.173754439841)*x[0]+(0.143503554071)*x[1]**o+(-0.0850087835084)*x[1]
        arg[0,2,2,0]=(0.316299661412)*x[0]**o+(0.260890197379)*x[0]+(-0.803519867776)*x[1]**o+(0.36434810185)*x[1]
        arg[0,2,2,1]=(-0.656013716479)*x[0]**o+(0.810499942514)*x[0]+(0.783455480815)*x[1]**o+(-0.943085466999)*x[1]
        arg[0,3,0,0]=(0.128153439722)*x[0]**o+(-0.404740042302)*x[0]+(0.785717360548)*x[1]**o+(-0.710818900841)*x[1]
        arg[0,3,0,1]=(-0.097059274787)*x[0]**o+(0.578757823228)*x[0]+(0.186146155178)*x[1]**o+(0.274602030707)*x[1]
        arg[0,3,1,0]=(0.00522392753954)*x[0]**o+(0.015773606982)*x[0]+(-0.327110065344)*x[1]**o+(0.15571120542)*x[1]
        arg[0,3,1,1]=(-0.14009116305)*x[0]**o+(-0.659975337811)*x[0]+(-0.297582833489)*x[1]**o+(-0.600935015071)*x[1]
        arg[0,3,2,0]=(0.749388604357)*x[0]**o+(0.762968650778)*x[0]+(-0.623477518339)*x[1]**o+(-0.931762943135)*x[1]
        arg[0,3,2,1]=(0.748396100632)*x[0]**o+(-0.841314360401)*x[0]+(0.432649446216)*x[1]**o+(-0.971300843628)*x[1]
        arg[0,4,0,0]=(-0.217249090357)*x[0]**o+(-0.833633675601)*x[0]+(0.169984305627)*x[1]**o+(-0.831738272666)*x[1]
        arg[0,4,0,1]=(0.501571101669)*x[0]**o+(-0.223897334788)*x[0]+(0.24178181976)*x[1]**o+(0.307921789017)*x[1]
        arg[0,4,1,0]=(-0.162064610564)*x[0]**o+(0.217669674603)*x[0]+(0.995517866562)*x[1]**o+(0.903061734589)*x[1]
        arg[0,4,1,1]=(-0.837150436928)*x[0]**o+(0.465269468525)*x[0]+(0.522034182082)*x[1]**o+(-0.725469921625)*x[1]
        arg[0,4,2,0]=(-0.295739906169)*x[0]**o+(0.961959220618)*x[0]+(0.577799918942)*x[1]**o+(-0.313255108148)*x[1]
        arg[0,4,2,1]=(-0.455635927062)*x[0]**o+(0.667038701641)*x[0]+(-0.151183772961)*x[1]**o+(-0.550197828316)*x[1]
        arg[1,0,0,0]=(0.585510684472)*x[0]**o+(-0.64754219493)*x[0]+(0.781495111863)*x[1]**o+(-0.792246214695)*x[1]
        arg[1,0,0,1]=(0.586346471397)*x[0]**o+(0.291794158138)*x[0]+(-0.614723564286)*x[1]**o+(-0.151451723805)*x[1]
        arg[1,0,1,0]=(-0.472877220222)*x[0]**o+(-0.0118252391089)*x[0]+(-0.596943135354)*x[1]**o+(-0.413992986066)*x[1]
        arg[1,0,1,1]=(-0.715007718152)*x[0]**o+(0.032964784403)*x[0]+(-0.350680745946)*x[1]**o+(0.678841285054)*x[1]
        arg[1,0,2,0]=(0.673910318085)*x[0]**o+(-0.15303471314)*x[0]+(-0.0889279944344)*x[1]**o+(-0.0356474498002)*x[1]
        arg[1,0,2,1]=(-0.875691348779)*x[0]**o+(0.547300853671)*x[0]+(-0.362646169132)*x[1]**o+(-0.588521964671)*x[1]
        arg[1,1,0,0]=(-0.0988912551145)*x[0]**o+(-0.710944394432)*x[0]+(0.153875249337)*x[1]**o+(0.274427434926)*x[1]
        arg[1,1,0,1]=(-0.0854272185591)*x[0]**o+(-0.21840195791)*x[0]+(0.906982245924)*x[1]**o+(0.570990116211)*x[1]
        arg[1,1,1,0]=(-0.851259116331)*x[0]**o+(0.309829955487)*x[0]+(0.115725082696)*x[1]**o+(-0.91590292189)*x[1]
        arg[1,1,1,1]=(-0.637464462703)*x[0]**o+(-0.280971672162)*x[0]+(0.270985403603)*x[1]**o+(-0.534115311397)*x[1]
        arg[1,1,2,0]=(0.670179973397)*x[0]**o+(0.847483793904)*x[0]+(0.0681088634526)*x[1]**o+(0.412176359729)*x[1]
        arg[1,1,2,1]=(0.0378205172485)*x[0]**o+(0.974150495292)*x[0]+(-0.720821275299)*x[1]**o+(-0.552588296494)*x[1]
        arg[1,2,0,0]=(0.151367609975)*x[0]**o+(0.832981586349)*x[0]+(0.419488381843)*x[1]**o+(0.561052438283)*x[1]
        arg[1,2,0,1]=(0.267839010425)*x[0]**o+(0.47455974064)*x[0]+(0.498826652567)*x[1]**o+(-0.51906391205)*x[1]
        arg[1,2,1,0]=(-0.350105107527)*x[0]**o+(-0.489326689724)*x[0]+(-0.156970199109)*x[1]**o+(-0.175565230472)*x[1]
        arg[1,2,1,1]=(0.754379973459)*x[0]**o+(-0.70691373234)*x[0]+(-0.964497356137)*x[1]**o+(-0.694989442192)*x[1]
        arg[1,2,2,0]=(-0.491090361655)*x[0]**o+(0.154379441501)*x[0]+(0.403809695039)*x[1]**o+(-0.084183660554)*x[1]
        arg[1,2,2,1]=(-0.164538235223)*x[0]**o+(0.628058472821)*x[0]+(0.331956969519)*x[1]**o+(0.828755643202)*x[1]
        arg[1,3,0,0]=(0.103479416566)*x[0]**o+(0.380691196177)*x[0]+(-0.350375360146)*x[1]**o+(0.18528492785)*x[1]
        arg[1,3,0,1]=(-0.0744423760559)*x[0]**o+(0.912275380092)*x[0]+(0.867254399179)*x[1]**o+(0.410307184344)*x[1]
        arg[1,3,1,0]=(-0.480324052626)*x[0]**o+(-0.438544962062)*x[0]+(0.749575808465)*x[1]**o+(-0.0689669475685)*x[1]
        arg[1,3,1,1]=(0.723325499611)*x[0]**o+(-0.368668363075)*x[0]+(0.404302895836)*x[1]**o+(0.66913878408)*x[1]
        arg[1,3,2,0]=(-0.122189168877)*x[0]**o+(0.603362404944)*x[0]+(0.184163482697)*x[1]**o+(0.728136755361)*x[1]
        arg[1,3,2,1]=(-0.149430048386)*x[0]**o+(-0.478108072433)*x[0]+(0.342479834611)*x[1]**o+(0.596757508109)*x[1]
        arg[1,4,0,0]=(0.334694167334)*x[0]**o+(-0.815793032292)*x[0]+(0.394250112116)*x[1]**o+(0.269197166808)*x[1]
        arg[1,4,0,1]=(-0.19210954081)*x[0]**o+(0.416736181257)*x[0]+(0.146217029416)*x[1]**o+(0.906316717458)*x[1]
        arg[1,4,1,0]=(-0.708061018673)*x[0]**o+(0.299200707691)*x[0]+(0.665111947965)*x[1]**o+(-0.0340170473314)*x[1]
        arg[1,4,1,1]=(0.0353929744202)*x[0]**o+(0.0493194235067)*x[0]+(-0.12215424813)*x[1]**o+(-0.758677440837)*x[1]
        arg[1,4,2,0]=(0.769702957701)*x[0]**o+(0.901805654682)*x[0]+(0.260222241019)*x[1]**o+(-0.471997705392)*x[1]
        arg[1,4,2,1]=(0.204258068316)*x[0]**o+(-0.173085114573)*x[0]+(-0.0253164178046)*x[1]**o+(0.84254484259)*x[1]
        arg[2,0,0,0]=(-0.688170206644)*x[0]**o+(-0.718673702314)*x[0]+(0.834265379201)*x[1]**o+(-0.879925159193)*x[1]
        arg[2,0,0,1]=(-0.515401334666)*x[0]**o+(-0.810523180967)*x[0]+(0.725182741017)*x[1]**o+(-0.74319471007)*x[1]
        arg[2,0,1,0]=(-0.799698108602)*x[0]**o+(0.289222327549)*x[0]+(0.0589402409999)*x[1]**o+(-0.437669765839)*x[1]
        arg[2,0,1,1]=(-0.734933454418)*x[0]**o+(0.371391146956)*x[0]+(-0.17797293763)*x[1]**o+(-0.499412028554)*x[1]
        arg[2,0,2,0]=(0.217283159505)*x[0]**o+(-0.184871188378)*x[0]+(-0.764095670777)*x[1]**o+(0.103705507755)*x[1]
        arg[2,0,2,1]=(0.578071527416)*x[0]**o+(0.845365737459)*x[0]+(-0.173187973137)*x[1]**o+(-0.827819925403)*x[1]
        arg[2,1,0,0]=(-0.258505119864)*x[0]**o+(0.640873475074)*x[0]+(-0.993841847675)*x[1]**o+(0.347490565123)*x[1]
        arg[2,1,0,1]=(0.224893445547)*x[0]**o+(0.827208591218)*x[0]+(0.0391563892525)*x[1]**o+(-0.320939200596)*x[1]
        arg[2,1,1,0]=(-0.993044884903)*x[0]**o+(-0.395235414086)*x[0]+(-0.447460653655)*x[1]**o+(-0.114585622808)*x[1]
        arg[2,1,1,1]=(0.908711735866)*x[0]**o+(-0.911405008413)*x[0]+(-0.798601192961)*x[1]**o+(-0.432937813587)*x[1]
        arg[2,1,2,0]=(-0.270424920067)*x[0]**o+(-0.533648944143)*x[0]+(0.184307763968)*x[1]**o+(-0.802221601505)*x[1]
        arg[2,1,2,1]=(0.154432841861)*x[0]**o+(0.306458633794)*x[0]+(0.0343421658454)*x[1]**o+(-0.768559457414)*x[1]
        arg[2,2,0,0]=(-0.497828132445)*x[0]**o+(0.484920248315)*x[0]+(-0.188239341875)*x[1]**o+(0.100360622012)*x[1]
        arg[2,2,0,1]=(-0.0433518780446)*x[0]**o+(0.119470819101)*x[0]+(0.344025017428)*x[1]**o+(0.667394126898)*x[1]
        arg[2,2,1,0]=(0.512271795846)*x[0]**o+(0.567024910623)*x[0]+(-0.316191920076)*x[1]**o+(-0.202135136851)*x[1]
        arg[2,2,1,1]=(0.781678855797)*x[0]**o+(0.159208533718)*x[0]+(-0.930150837465)*x[1]**o+(-0.666903895504)*x[1]
        arg[2,2,2,0]=(0.292270965732)*x[0]**o+(-0.112877441048)*x[0]+(-0.643355582049)*x[1]**o+(-0.0482560996171)*x[1]
        arg[2,2,2,1]=(0.638688065872)*x[0]**o+(0.787719224138)*x[0]+(0.815312681592)*x[1]**o+(0.334735776221)*x[1]
        arg[2,3,0,0]=(0.409920519672)*x[0]**o+(-0.191886243537)*x[0]+(-0.348205383522)*x[1]**o+(-0.651065510647)*x[1]
        arg[2,3,0,1]=(0.824428500178)*x[0]**o+(-0.909533490456)*x[0]+(-0.455079541262)*x[1]**o+(-0.613994310127)*x[1]
        arg[2,3,1,0]=(0.225908995914)*x[0]**o+(0.258098741048)*x[0]+(-0.858905205327)*x[1]**o+(0.535845140874)*x[1]
        arg[2,3,1,1]=(-0.183600641837)*x[0]**o+(-0.0436432564818)*x[0]+(0.306524301109)*x[1]**o+(0.662513699164)*x[1]
        arg[2,3,2,0]=(-0.678973103555)*x[0]**o+(0.674110571991)*x[0]+(-0.129743880618)*x[1]**o+(-0.210759231178)*x[1]
        arg[2,3,2,1]=(-0.594989617102)*x[0]**o+(0.0557145839856)*x[0]+(-0.379516802094)*x[1]**o+(-0.496127170036)*x[1]
        arg[2,4,0,0]=(0.401088886175)*x[0]**o+(-0.208580896076)*x[0]+(-0.655582275521)*x[1]**o+(-0.522892532154)*x[1]
        arg[2,4,0,1]=(0.609111916382)*x[0]**o+(-0.864603847696)*x[0]+(-0.373482060442)*x[1]**o+(0.048777005392)*x[1]
        arg[2,4,1,0]=(-0.503935244799)*x[0]**o+(0.320517360825)*x[0]+(-0.658994963813)*x[1]**o+(0.00465823412972)*x[1]
        arg[2,4,1,1]=(-0.202767966863)*x[0]**o+(0.737239608437)*x[0]+(-0.633206522406)*x[1]**o+(-0.408531988781)*x[1]
        arg[2,4,2,0]=(0.660048575091)*x[0]**o+(-0.399137043056)*x[0]+(-0.0977181117895)*x[1]**o+(0.148903352776)*x[1]
        arg[2,4,2,1]=(-0.396768795712)*x[0]**o+(-0.550388681028)*x[0]+(0.0946256561554)*x[1]**o+(-0.0566941754267)*x[1]
        arg[3,0,0,0]=(0.0647180650687)*x[0]**o+(0.920788893299)*x[0]+(0.131306350033)*x[1]**o+(-0.840527983774)*x[1]
        arg[3,0,0,1]=(0.451700262969)*x[0]**o+(-0.16104676139)*x[0]+(0.114145999713)*x[1]**o+(-0.949265482931)*x[1]
        arg[3,0,1,0]=(-0.219444617083)*x[0]**o+(0.846470673388)*x[0]+(-0.341392225388)*x[1]**o+(-0.721912717627)*x[1]
        arg[3,0,1,1]=(-0.199502120929)*x[0]**o+(-0.00662410675554)*x[0]+(0.0529013037694)*x[1]**o+(-0.022278580228)*x[1]
        arg[3,0,2,0]=(0.803500732535)*x[0]**o+(0.754138333448)*x[0]+(0.822404687478)*x[1]**o+(0.900099584934)*x[1]
        arg[3,0,2,1]=(0.84674757393)*x[0]**o+(0.326823539079)*x[0]+(-0.824446392056)*x[1]**o+(-0.506583766552)*x[1]
        arg[3,1,0,0]=(0.816775849327)*x[0]**o+(0.957523430402)*x[0]+(-0.822535686534)*x[1]**o+(-0.583389123083)*x[1]
        arg[3,1,0,1]=(0.518120693337)*x[0]**o+(0.0946135115511)*x[0]+(0.577364260402)*x[1]**o+(-0.677435109716)*x[1]
        arg[3,1,1,0]=(-0.559041961891)*x[0]**o+(0.436791134026)*x[0]+(0.287609046653)*x[1]**o+(-0.484573207793)*x[1]
        arg[3,1,1,1]=(-0.0786041693053)*x[0]**o+(0.793526931222)*x[0]+(0.055780633982)*x[1]**o+(0.0447074319839)*x[1]
        arg[3,1,2,0]=(-0.190231018951)*x[0]**o+(-0.61318830169)*x[0]+(-0.157202158534)*x[1]**o+(-0.823043159126)*x[1]
        arg[3,1,2,1]=(-0.502637901786)*x[0]**o+(0.781083145697)*x[0]+(-0.663013942651)*x[1]**o+(-0.305726874525)*x[1]
        arg[3,2,0,0]=(0.566955929517)*x[0]**o+(-0.506981068575)*x[0]+(-0.831439289654)*x[1]**o+(-0.337491442765)*x[1]
        arg[3,2,0,1]=(-0.713648844422)*x[0]**o+(-0.00815764297609)*x[0]+(0.829706206618)*x[1]**o+(0.96471376965)*x[1]
        arg[3,2,1,0]=(-0.944705561715)*x[0]**o+(-0.87962603172)*x[0]+(-0.0525190846342)*x[1]**o+(-0.0457463400515)*x[1]
        arg[3,2,1,1]=(0.54922446571)*x[0]**o+(0.651923573438)*x[0]+(-0.00229398468376)*x[1]**o+(-0.00170863787749)*x[1]
        arg[3,2,2,0]=(0.622372079685)*x[0]**o+(-0.553578221845)*x[0]+(-0.737877450857)*x[1]**o+(0.459729854002)*x[1]
        arg[3,2,2,1]=(0.826155732508)*x[0]**o+(-0.000468898258991)*x[0]+(-0.255179342686)*x[1]**o+(-0.80957312045)*x[1]
        arg[3,3,0,0]=(-0.195212469839)*x[0]**o+(-0.0235216449547)*x[0]+(-0.175192639485)*x[1]**o+(-0.58937661213)*x[1]
        arg[3,3,0,1]=(-0.931538910404)*x[0]**o+(0.649282880007)*x[0]+(-0.237841482637)*x[1]**o+(-0.0220516220802)*x[1]
        arg[3,3,1,0]=(0.420911057166)*x[0]**o+(0.200126728184)*x[0]+(-0.829130861656)*x[1]**o+(-0.563462916661)*x[1]
        arg[3,3,1,1]=(-0.148199974575)*x[0]**o+(-0.248825389533)*x[0]+(-0.346424894584)*x[1]**o+(0.339758232978)*x[1]
        arg[3,3,2,0]=(0.249372496215)*x[0]**o+(-0.343626915746)*x[0]+(0.0961351845087)*x[1]**o+(0.612679249068)*x[1]
        arg[3,3,2,1]=(0.985380582259)*x[0]**o+(-0.603514318626)*x[0]+(-0.131105546364)*x[1]**o+(-0.0798158697238)*x[1]
        arg[3,4,0,0]=(0.478635365792)*x[0]**o+(-0.964424450881)*x[0]+(-0.469995535544)*x[1]**o+(-0.72871149515)*x[1]
        arg[3,4,0,1]=(-0.264802297415)*x[0]**o+(-0.926518282507)*x[0]+(0.803390182149)*x[1]**o+(0.387910762706)*x[1]
        arg[3,4,1,0]=(0.0452448579909)*x[0]**o+(-0.0482221515826)*x[0]+(-0.728020494982)*x[1]**o+(-0.288823633249)*x[1]
        arg[3,4,1,1]=(-0.640459595265)*x[0]**o+(0.759525453728)*x[0]+(-0.0280300707884)*x[1]**o+(0.079438187123)*x[1]
        arg[3,4,2,0]=(-0.14185171171)*x[0]**o+(-0.810989314052)*x[0]+(0.921902325773)*x[1]**o+(0.885920658784)*x[1]
        arg[3,4,2,1]=(-0.0101259901393)*x[0]**o+(-0.696541046542)*x[0]+(0.127932822496)*x[1]**o+(-0.822265796323)*x[1]
        ref[0,0,0,0]=(0.0577426785287)*x_ref[0]**o+(0.469845292021)*x_ref[0]+(-0.641706722819)*x_ref[1]**o+(0.129652001842)*x_ref[1]
        ref[0,0,0,1]=(-0.87686951076)*x_ref[0]**o+(-0.981093040999)*x_ref[0]+(0.654637319552)*x_ref[1]**o+(-0.953104888811)*x_ref[1]
        ref[0,0,1,0]=(-0.486386436415)*x_ref[0]**o+(0.91083470956)*x_ref[0]+(-0.392432937401)*x_ref[1]**o+(0.494919250884)*x_ref[1]
        ref[0,0,1,1]=(0.299928502076)*x_ref[0]**o+(-0.0787848191157)*x_ref[0]+(-0.541863592475)*x_ref[1]**o+(0.435934816687)*x_ref[1]
        ref[0,0,2,0]=(-0.0500083438365)*x_ref[0]**o+(-0.882581886623)*x_ref[0]+(-0.965347357063)*x_ref[1]**o+(-0.270254041774)*x_ref[1]
        ref[0,0,2,1]=(-0.18974861517)*x_ref[0]**o+(0.341406736671)*x_ref[0]+(0.847020774229)*x_ref[1]**o+(0.650201215849)*x_ref[1]
        ref[0,1,0,0]=(0.0823017084719)*x_ref[0]**o+(0.964585359235)*x_ref[0]+(0.357658660436)*x_ref[1]**o+(0.182585271812)*x_ref[1]
        ref[0,1,0,1]=(0.863411821905)*x_ref[0]**o+(0.313374566426)*x_ref[0]+(-0.273547644827)*x_ref[1]**o+(0.429051854479)*x_ref[1]
        ref[0,1,1,0]=(0.167608645068)*x_ref[0]**o+(0.926908094967)*x_ref[0]+(-0.853415240465)*x_ref[1]**o+(-0.376634525241)*x_ref[1]
        ref[0,1,1,1]=(-0.555491961591)*x_ref[0]**o+(-0.64621137308)*x_ref[0]+(-0.315877124991)*x_ref[1]**o+(-0.962839286133)*x_ref[1]
        ref[0,1,2,0]=(0.909012065667)*x_ref[0]**o+(-0.656739022976)*x_ref[0]+(-0.458377773547)*x_ref[1]**o+(0.394223574009)*x_ref[1]
        ref[0,1,2,1]=(-0.550560496714)*x_ref[0]**o+(0.216941875936)*x_ref[0]+(-0.634745794924)*x_ref[1]**o+(0.887912543217)*x_ref[1]
        ref[0,2,0,0]=(-0.961535901493)*x_ref[0]**o+(0.980336384835)*x_ref[0]+(-0.151755000128)*x_ref[1]**o+(0.0402841925797)*x_ref[1]
        ref[0,2,0,1]=(0.571527088727)*x_ref[0]**o+(-0.475013184811)*x_ref[0]+(0.80051818192)*x_ref[1]**o+(-0.589499747694)*x_ref[1]
        ref[0,2,1,0]=(0.545840648612)*x_ref[0]**o+(0.888313209612)*x_ref[0]+(-0.568938733446)*x_ref[1]**o+(-0.289030374342)*x_ref[1]
        ref[0,2,1,1]=(-0.0631638607979)*x_ref[0]**o+(-0.173754439841)*x_ref[0]+(0.143503554071)*x_ref[1]**o+(-0.0850087835084)*x_ref[1]
        ref[0,2,2,0]=(0.316299661412)*x_ref[0]**o+(0.260890197379)*x_ref[0]+(-0.803519867776)*x_ref[1]**o+(0.36434810185)*x_ref[1]
        ref[0,2,2,1]=(-0.656013716479)*x_ref[0]**o+(0.810499942514)*x_ref[0]+(0.783455480815)*x_ref[1]**o+(-0.943085466999)*x_ref[1]
        ref[0,3,0,0]=(0.128153439722)*x_ref[0]**o+(-0.404740042302)*x_ref[0]+(0.785717360548)*x_ref[1]**o+(-0.710818900841)*x_ref[1]
        ref[0,3,0,1]=(-0.097059274787)*x_ref[0]**o+(0.578757823228)*x_ref[0]+(0.186146155178)*x_ref[1]**o+(0.274602030707)*x_ref[1]
        ref[0,3,1,0]=(0.00522392753954)*x_ref[0]**o+(0.015773606982)*x_ref[0]+(-0.327110065344)*x_ref[1]**o+(0.15571120542)*x_ref[1]
        ref[0,3,1,1]=(-0.14009116305)*x_ref[0]**o+(-0.659975337811)*x_ref[0]+(-0.297582833489)*x_ref[1]**o+(-0.600935015071)*x_ref[1]
        ref[0,3,2,0]=(0.749388604357)*x_ref[0]**o+(0.762968650778)*x_ref[0]+(-0.623477518339)*x_ref[1]**o+(-0.931762943135)*x_ref[1]
        ref[0,3,2,1]=(0.748396100632)*x_ref[0]**o+(-0.841314360401)*x_ref[0]+(0.432649446216)*x_ref[1]**o+(-0.971300843628)*x_ref[1]
        ref[0,4,0,0]=(-0.217249090357)*x_ref[0]**o+(-0.833633675601)*x_ref[0]+(0.169984305627)*x_ref[1]**o+(-0.831738272666)*x_ref[1]
        ref[0,4,0,1]=(0.501571101669)*x_ref[0]**o+(-0.223897334788)*x_ref[0]+(0.24178181976)*x_ref[1]**o+(0.307921789017)*x_ref[1]
        ref[0,4,1,0]=(-0.162064610564)*x_ref[0]**o+(0.217669674603)*x_ref[0]+(0.995517866562)*x_ref[1]**o+(0.903061734589)*x_ref[1]
        ref[0,4,1,1]=(-0.837150436928)*x_ref[0]**o+(0.465269468525)*x_ref[0]+(0.522034182082)*x_ref[1]**o+(-0.725469921625)*x_ref[1]
        ref[0,4,2,0]=(-0.295739906169)*x_ref[0]**o+(0.961959220618)*x_ref[0]+(0.577799918942)*x_ref[1]**o+(-0.313255108148)*x_ref[1]
        ref[0,4,2,1]=(-0.455635927062)*x_ref[0]**o+(0.667038701641)*x_ref[0]+(-0.151183772961)*x_ref[1]**o+(-0.550197828316)*x_ref[1]
        ref[1,0,0,0]=(0.585510684472)*x_ref[0]**o+(-0.64754219493)*x_ref[0]+(0.781495111863)*x_ref[1]**o+(-0.792246214695)*x_ref[1]
        ref[1,0,0,1]=(0.586346471397)*x_ref[0]**o+(0.291794158138)*x_ref[0]+(-0.614723564286)*x_ref[1]**o+(-0.151451723805)*x_ref[1]
        ref[1,0,1,0]=(-0.472877220222)*x_ref[0]**o+(-0.0118252391089)*x_ref[0]+(-0.596943135354)*x_ref[1]**o+(-0.413992986066)*x_ref[1]
        ref[1,0,1,1]=(-0.715007718152)*x_ref[0]**o+(0.032964784403)*x_ref[0]+(-0.350680745946)*x_ref[1]**o+(0.678841285054)*x_ref[1]
        ref[1,0,2,0]=(0.673910318085)*x_ref[0]**o+(-0.15303471314)*x_ref[0]+(-0.0889279944344)*x_ref[1]**o+(-0.0356474498002)*x_ref[1]
        ref[1,0,2,1]=(-0.875691348779)*x_ref[0]**o+(0.547300853671)*x_ref[0]+(-0.362646169132)*x_ref[1]**o+(-0.588521964671)*x_ref[1]
        ref[1,1,0,0]=(-0.0988912551145)*x_ref[0]**o+(-0.710944394432)*x_ref[0]+(0.153875249337)*x_ref[1]**o+(0.274427434926)*x_ref[1]
        ref[1,1,0,1]=(-0.0854272185591)*x_ref[0]**o+(-0.21840195791)*x_ref[0]+(0.906982245924)*x_ref[1]**o+(0.570990116211)*x_ref[1]
        ref[1,1,1,0]=(-0.851259116331)*x_ref[0]**o+(0.309829955487)*x_ref[0]+(0.115725082696)*x_ref[1]**o+(-0.91590292189)*x_ref[1]
        ref[1,1,1,1]=(-0.637464462703)*x_ref[0]**o+(-0.280971672162)*x_ref[0]+(0.270985403603)*x_ref[1]**o+(-0.534115311397)*x_ref[1]
        ref[1,1,2,0]=(0.670179973397)*x_ref[0]**o+(0.847483793904)*x_ref[0]+(0.0681088634526)*x_ref[1]**o+(0.412176359729)*x_ref[1]
        ref[1,1,2,1]=(0.0378205172485)*x_ref[0]**o+(0.974150495292)*x_ref[0]+(-0.720821275299)*x_ref[1]**o+(-0.552588296494)*x_ref[1]
        ref[1,2,0,0]=(0.151367609975)*x_ref[0]**o+(0.832981586349)*x_ref[0]+(0.419488381843)*x_ref[1]**o+(0.561052438283)*x_ref[1]
        ref[1,2,0,1]=(0.267839010425)*x_ref[0]**o+(0.47455974064)*x_ref[0]+(0.498826652567)*x_ref[1]**o+(-0.51906391205)*x_ref[1]
        ref[1,2,1,0]=(-0.350105107527)*x_ref[0]**o+(-0.489326689724)*x_ref[0]+(-0.156970199109)*x_ref[1]**o+(-0.175565230472)*x_ref[1]
        ref[1,2,1,1]=(0.754379973459)*x_ref[0]**o+(-0.70691373234)*x_ref[0]+(-0.964497356137)*x_ref[1]**o+(-0.694989442192)*x_ref[1]
        ref[1,2,2,0]=(-0.491090361655)*x_ref[0]**o+(0.154379441501)*x_ref[0]+(0.403809695039)*x_ref[1]**o+(-0.084183660554)*x_ref[1]
        ref[1,2,2,1]=(-0.164538235223)*x_ref[0]**o+(0.628058472821)*x_ref[0]+(0.331956969519)*x_ref[1]**o+(0.828755643202)*x_ref[1]
        ref[1,3,0,0]=(0.103479416566)*x_ref[0]**o+(0.380691196177)*x_ref[0]+(-0.350375360146)*x_ref[1]**o+(0.18528492785)*x_ref[1]
        ref[1,3,0,1]=(-0.0744423760559)*x_ref[0]**o+(0.912275380092)*x_ref[0]+(0.867254399179)*x_ref[1]**o+(0.410307184344)*x_ref[1]
        ref[1,3,1,0]=(-0.480324052626)*x_ref[0]**o+(-0.438544962062)*x_ref[0]+(0.749575808465)*x_ref[1]**o+(-0.0689669475685)*x_ref[1]
        ref[1,3,1,1]=(0.723325499611)*x_ref[0]**o+(-0.368668363075)*x_ref[0]+(0.404302895836)*x_ref[1]**o+(0.66913878408)*x_ref[1]
        ref[1,3,2,0]=(-0.122189168877)*x_ref[0]**o+(0.603362404944)*x_ref[0]+(0.184163482697)*x_ref[1]**o+(0.728136755361)*x_ref[1]
        ref[1,3,2,1]=(-0.149430048386)*x_ref[0]**o+(-0.478108072433)*x_ref[0]+(0.342479834611)*x_ref[1]**o+(0.596757508109)*x_ref[1]
        ref[1,4,0,0]=(0.334694167334)*x_ref[0]**o+(-0.815793032292)*x_ref[0]+(0.394250112116)*x_ref[1]**o+(0.269197166808)*x_ref[1]
        ref[1,4,0,1]=(-0.19210954081)*x_ref[0]**o+(0.416736181257)*x_ref[0]+(0.146217029416)*x_ref[1]**o+(0.906316717458)*x_ref[1]
        ref[1,4,1,0]=(-0.708061018673)*x_ref[0]**o+(0.299200707691)*x_ref[0]+(0.665111947965)*x_ref[1]**o+(-0.0340170473314)*x_ref[1]
        ref[1,4,1,1]=(0.0353929744202)*x_ref[0]**o+(0.0493194235067)*x_ref[0]+(-0.12215424813)*x_ref[1]**o+(-0.758677440837)*x_ref[1]
        ref[1,4,2,0]=(0.769702957701)*x_ref[0]**o+(0.901805654682)*x_ref[0]+(0.260222241019)*x_ref[1]**o+(-0.471997705392)*x_ref[1]
        ref[1,4,2,1]=(0.204258068316)*x_ref[0]**o+(-0.173085114573)*x_ref[0]+(-0.0253164178046)*x_ref[1]**o+(0.84254484259)*x_ref[1]
        ref[2,0,0,0]=(-0.688170206644)*x_ref[0]**o+(-0.718673702314)*x_ref[0]+(0.834265379201)*x_ref[1]**o+(-0.879925159193)*x_ref[1]
        ref[2,0,0,1]=(-0.515401334666)*x_ref[0]**o+(-0.810523180967)*x_ref[0]+(0.725182741017)*x_ref[1]**o+(-0.74319471007)*x_ref[1]
        ref[2,0,1,0]=(-0.799698108602)*x_ref[0]**o+(0.289222327549)*x_ref[0]+(0.0589402409999)*x_ref[1]**o+(-0.437669765839)*x_ref[1]
        ref[2,0,1,1]=(-0.734933454418)*x_ref[0]**o+(0.371391146956)*x_ref[0]+(-0.17797293763)*x_ref[1]**o+(-0.499412028554)*x_ref[1]
        ref[2,0,2,0]=(0.217283159505)*x_ref[0]**o+(-0.184871188378)*x_ref[0]+(-0.764095670777)*x_ref[1]**o+(0.103705507755)*x_ref[1]
        ref[2,0,2,1]=(0.578071527416)*x_ref[0]**o+(0.845365737459)*x_ref[0]+(-0.173187973137)*x_ref[1]**o+(-0.827819925403)*x_ref[1]
        ref[2,1,0,0]=(-0.258505119864)*x_ref[0]**o+(0.640873475074)*x_ref[0]+(-0.993841847675)*x_ref[1]**o+(0.347490565123)*x_ref[1]
        ref[2,1,0,1]=(0.224893445547)*x_ref[0]**o+(0.827208591218)*x_ref[0]+(0.0391563892525)*x_ref[1]**o+(-0.320939200596)*x_ref[1]
        ref[2,1,1,0]=(-0.993044884903)*x_ref[0]**o+(-0.395235414086)*x_ref[0]+(-0.447460653655)*x_ref[1]**o+(-0.114585622808)*x_ref[1]
        ref[2,1,1,1]=(0.908711735866)*x_ref[0]**o+(-0.911405008413)*x_ref[0]+(-0.798601192961)*x_ref[1]**o+(-0.432937813587)*x_ref[1]
        ref[2,1,2,0]=(-0.270424920067)*x_ref[0]**o+(-0.533648944143)*x_ref[0]+(0.184307763968)*x_ref[1]**o+(-0.802221601505)*x_ref[1]
        ref[2,1,2,1]=(0.154432841861)*x_ref[0]**o+(0.306458633794)*x_ref[0]+(0.0343421658454)*x_ref[1]**o+(-0.768559457414)*x_ref[1]
        ref[2,2,0,0]=(-0.497828132445)*x_ref[0]**o+(0.484920248315)*x_ref[0]+(-0.188239341875)*x_ref[1]**o+(0.100360622012)*x_ref[1]
        ref[2,2,0,1]=(-0.0433518780446)*x_ref[0]**o+(0.119470819101)*x_ref[0]+(0.344025017428)*x_ref[1]**o+(0.667394126898)*x_ref[1]
        ref[2,2,1,0]=(0.512271795846)*x_ref[0]**o+(0.567024910623)*x_ref[0]+(-0.316191920076)*x_ref[1]**o+(-0.202135136851)*x_ref[1]
        ref[2,2,1,1]=(0.781678855797)*x_ref[0]**o+(0.159208533718)*x_ref[0]+(-0.930150837465)*x_ref[1]**o+(-0.666903895504)*x_ref[1]
        ref[2,2,2,0]=(0.292270965732)*x_ref[0]**o+(-0.112877441048)*x_ref[0]+(-0.643355582049)*x_ref[1]**o+(-0.0482560996171)*x_ref[1]
        ref[2,2,2,1]=(0.638688065872)*x_ref[0]**o+(0.787719224138)*x_ref[0]+(0.815312681592)*x_ref[1]**o+(0.334735776221)*x_ref[1]
        ref[2,3,0,0]=(0.409920519672)*x_ref[0]**o+(-0.191886243537)*x_ref[0]+(-0.348205383522)*x_ref[1]**o+(-0.651065510647)*x_ref[1]
        ref[2,3,0,1]=(0.824428500178)*x_ref[0]**o+(-0.909533490456)*x_ref[0]+(-0.455079541262)*x_ref[1]**o+(-0.613994310127)*x_ref[1]
        ref[2,3,1,0]=(0.225908995914)*x_ref[0]**o+(0.258098741048)*x_ref[0]+(-0.858905205327)*x_ref[1]**o+(0.535845140874)*x_ref[1]
        ref[2,3,1,1]=(-0.183600641837)*x_ref[0]**o+(-0.0436432564818)*x_ref[0]+(0.306524301109)*x_ref[1]**o+(0.662513699164)*x_ref[1]
        ref[2,3,2,0]=(-0.678973103555)*x_ref[0]**o+(0.674110571991)*x_ref[0]+(-0.129743880618)*x_ref[1]**o+(-0.210759231178)*x_ref[1]
        ref[2,3,2,1]=(-0.594989617102)*x_ref[0]**o+(0.0557145839856)*x_ref[0]+(-0.379516802094)*x_ref[1]**o+(-0.496127170036)*x_ref[1]
        ref[2,4,0,0]=(0.401088886175)*x_ref[0]**o+(-0.208580896076)*x_ref[0]+(-0.655582275521)*x_ref[1]**o+(-0.522892532154)*x_ref[1]
        ref[2,4,0,1]=(0.609111916382)*x_ref[0]**o+(-0.864603847696)*x_ref[0]+(-0.373482060442)*x_ref[1]**o+(0.048777005392)*x_ref[1]
        ref[2,4,1,0]=(-0.503935244799)*x_ref[0]**o+(0.320517360825)*x_ref[0]+(-0.658994963813)*x_ref[1]**o+(0.00465823412972)*x_ref[1]
        ref[2,4,1,1]=(-0.202767966863)*x_ref[0]**o+(0.737239608437)*x_ref[0]+(-0.633206522406)*x_ref[1]**o+(-0.408531988781)*x_ref[1]
        ref[2,4,2,0]=(0.660048575091)*x_ref[0]**o+(-0.399137043056)*x_ref[0]+(-0.0977181117895)*x_ref[1]**o+(0.148903352776)*x_ref[1]
        ref[2,4,2,1]=(-0.396768795712)*x_ref[0]**o+(-0.550388681028)*x_ref[0]+(0.0946256561554)*x_ref[1]**o+(-0.0566941754267)*x_ref[1]
        ref[3,0,0,0]=(0.0647180650687)*x_ref[0]**o+(0.920788893299)*x_ref[0]+(0.131306350033)*x_ref[1]**o+(-0.840527983774)*x_ref[1]
        ref[3,0,0,1]=(0.451700262969)*x_ref[0]**o+(-0.16104676139)*x_ref[0]+(0.114145999713)*x_ref[1]**o+(-0.949265482931)*x_ref[1]
        ref[3,0,1,0]=(-0.219444617083)*x_ref[0]**o+(0.846470673388)*x_ref[0]+(-0.341392225388)*x_ref[1]**o+(-0.721912717627)*x_ref[1]
        ref[3,0,1,1]=(-0.199502120929)*x_ref[0]**o+(-0.00662410675554)*x_ref[0]+(0.0529013037694)*x_ref[1]**o+(-0.022278580228)*x_ref[1]
        ref[3,0,2,0]=(0.803500732535)*x_ref[0]**o+(0.754138333448)*x_ref[0]+(0.822404687478)*x_ref[1]**o+(0.900099584934)*x_ref[1]
        ref[3,0,2,1]=(0.84674757393)*x_ref[0]**o+(0.326823539079)*x_ref[0]+(-0.824446392056)*x_ref[1]**o+(-0.506583766552)*x_ref[1]
        ref[3,1,0,0]=(0.816775849327)*x_ref[0]**o+(0.957523430402)*x_ref[0]+(-0.822535686534)*x_ref[1]**o+(-0.583389123083)*x_ref[1]
        ref[3,1,0,1]=(0.518120693337)*x_ref[0]**o+(0.0946135115511)*x_ref[0]+(0.577364260402)*x_ref[1]**o+(-0.677435109716)*x_ref[1]
        ref[3,1,1,0]=(-0.559041961891)*x_ref[0]**o+(0.436791134026)*x_ref[0]+(0.287609046653)*x_ref[1]**o+(-0.484573207793)*x_ref[1]
        ref[3,1,1,1]=(-0.0786041693053)*x_ref[0]**o+(0.793526931222)*x_ref[0]+(0.055780633982)*x_ref[1]**o+(0.0447074319839)*x_ref[1]
        ref[3,1,2,0]=(-0.190231018951)*x_ref[0]**o+(-0.61318830169)*x_ref[0]+(-0.157202158534)*x_ref[1]**o+(-0.823043159126)*x_ref[1]
        ref[3,1,2,1]=(-0.502637901786)*x_ref[0]**o+(0.781083145697)*x_ref[0]+(-0.663013942651)*x_ref[1]**o+(-0.305726874525)*x_ref[1]
        ref[3,2,0,0]=(0.566955929517)*x_ref[0]**o+(-0.506981068575)*x_ref[0]+(-0.831439289654)*x_ref[1]**o+(-0.337491442765)*x_ref[1]
        ref[3,2,0,1]=(-0.713648844422)*x_ref[0]**o+(-0.00815764297609)*x_ref[0]+(0.829706206618)*x_ref[1]**o+(0.96471376965)*x_ref[1]
        ref[3,2,1,0]=(-0.944705561715)*x_ref[0]**o+(-0.87962603172)*x_ref[0]+(-0.0525190846342)*x_ref[1]**o+(-0.0457463400515)*x_ref[1]
        ref[3,2,1,1]=(0.54922446571)*x_ref[0]**o+(0.651923573438)*x_ref[0]+(-0.00229398468376)*x_ref[1]**o+(-0.00170863787749)*x_ref[1]
        ref[3,2,2,0]=(0.622372079685)*x_ref[0]**o+(-0.553578221845)*x_ref[0]+(-0.737877450857)*x_ref[1]**o+(0.459729854002)*x_ref[1]
        ref[3,2,2,1]=(0.826155732508)*x_ref[0]**o+(-0.000468898258991)*x_ref[0]+(-0.255179342686)*x_ref[1]**o+(-0.80957312045)*x_ref[1]
        ref[3,3,0,0]=(-0.195212469839)*x_ref[0]**o+(-0.0235216449547)*x_ref[0]+(-0.175192639485)*x_ref[1]**o+(-0.58937661213)*x_ref[1]
        ref[3,3,0,1]=(-0.931538910404)*x_ref[0]**o+(0.649282880007)*x_ref[0]+(-0.237841482637)*x_ref[1]**o+(-0.0220516220802)*x_ref[1]
        ref[3,3,1,0]=(0.420911057166)*x_ref[0]**o+(0.200126728184)*x_ref[0]+(-0.829130861656)*x_ref[1]**o+(-0.563462916661)*x_ref[1]
        ref[3,3,1,1]=(-0.148199974575)*x_ref[0]**o+(-0.248825389533)*x_ref[0]+(-0.346424894584)*x_ref[1]**o+(0.339758232978)*x_ref[1]
        ref[3,3,2,0]=(0.249372496215)*x_ref[0]**o+(-0.343626915746)*x_ref[0]+(0.0961351845087)*x_ref[1]**o+(0.612679249068)*x_ref[1]
        ref[3,3,2,1]=(0.985380582259)*x_ref[0]**o+(-0.603514318626)*x_ref[0]+(-0.131105546364)*x_ref[1]**o+(-0.0798158697238)*x_ref[1]
        ref[3,4,0,0]=(0.478635365792)*x_ref[0]**o+(-0.964424450881)*x_ref[0]+(-0.469995535544)*x_ref[1]**o+(-0.72871149515)*x_ref[1]
        ref[3,4,0,1]=(-0.264802297415)*x_ref[0]**o+(-0.926518282507)*x_ref[0]+(0.803390182149)*x_ref[1]**o+(0.387910762706)*x_ref[1]
        ref[3,4,1,0]=(0.0452448579909)*x_ref[0]**o+(-0.0482221515826)*x_ref[0]+(-0.728020494982)*x_ref[1]**o+(-0.288823633249)*x_ref[1]
        ref[3,4,1,1]=(-0.640459595265)*x_ref[0]**o+(0.759525453728)*x_ref[0]+(-0.0280300707884)*x_ref[1]**o+(0.079438187123)*x_ref[1]
        ref[3,4,2,0]=(-0.14185171171)*x_ref[0]**o+(-0.810989314052)*x_ref[0]+(0.921902325773)*x_ref[1]**o+(0.885920658784)*x_ref[1]
        ref[3,4,2,1]=(-0.0101259901393)*x_ref[0]**o+(-0.696541046542)*x_ref[0]+(0.127932822496)*x_ref[1]**o+(-0.822265796323)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.657703355307)*x[0]**o+(0.397813831843)*x[0]+(0.364460538717)*x[1]**o+(-0.845626669617)*x[1]+(-0.987144254126)*x[2]**o+(-0.732384980726)*x[2]
        arg[0,0,0,1]=(-0.13280304886)*x[0]**o+(-0.953796807758)*x[0]+(0.875723746654)*x[1]**o+(0.0764745117242)*x[1]+(-0.536085676221)*x[2]**o+(-0.520611959678)*x[2]
        arg[0,0,1,0]=(0.716429446578)*x[0]**o+(-0.729164693441)*x[0]+(-0.832505557428)*x[1]**o+(-0.329200605611)*x[1]+(0.0310162060964)*x[2]**o+(-0.869978941999)*x[2]
        arg[0,0,1,1]=(0.110253530584)*x[0]**o+(0.995528486822)*x[0]+(0.838577797395)*x[1]**o+(0.443987752225)*x[1]+(0.387337959057)*x[2]**o+(0.548448752612)*x[2]
        arg[0,0,2,0]=(-0.995737515431)*x[0]**o+(-0.899807241717)*x[0]+(-0.0803287764138)*x[1]**o+(0.242680051556)*x[1]+(0.58974836823)*x[2]**o+(-0.498450188944)*x[2]
        arg[0,0,2,1]=(0.341561541555)*x[0]**o+(-0.00541988184597)*x[0]+(-0.204832898701)*x[1]**o+(-0.957247935865)*x[1]+(-0.441002902064)*x[2]**o+(0.756843480304)*x[2]
        arg[0,1,0,0]=(0.136201526244)*x[0]**o+(0.517677304193)*x[0]+(0.523031610748)*x[1]**o+(-0.908998053383)*x[1]+(-0.629649678416)*x[2]**o+(0.762977917512)*x[2]
        arg[0,1,0,1]=(-0.103521080864)*x[0]**o+(-0.434410222595)*x[0]+(0.418239530241)*x[1]**o+(0.65638475119)*x[1]+(0.413036393129)*x[2]**o+(-0.220365518353)*x[2]
        arg[0,1,1,0]=(-0.415690201483)*x[0]**o+(-0.969323394689)*x[0]+(-0.295639420195)*x[1]**o+(-0.24915002749)*x[1]+(-0.00927627101467)*x[2]**o+(0.520335261634)*x[2]
        arg[0,1,1,1]=(-0.262653741797)*x[0]**o+(-0.463590591666)*x[0]+(0.568893042923)*x[1]**o+(-0.869851332898)*x[1]+(-0.463261399415)*x[2]**o+(0.0281915686202)*x[2]
        arg[0,1,2,0]=(0.377159503058)*x[0]**o+(0.484512261943)*x[0]+(0.0303594628925)*x[1]**o+(0.482051243004)*x[1]+(0.240643013949)*x[2]**o+(-0.943908617705)*x[2]
        arg[0,1,2,1]=(0.843854614097)*x[0]**o+(0.018041534553)*x[0]+(0.264198703361)*x[1]**o+(0.732889517628)*x[1]+(-0.154807384394)*x[2]**o+(0.60513622187)*x[2]
        arg[0,2,0,0]=(-0.234758118286)*x[0]**o+(-0.77807396119)*x[0]+(-0.97783091103)*x[1]**o+(-0.716185223828)*x[1]+(-0.573126474259)*x[2]**o+(0.403110701555)*x[2]
        arg[0,2,0,1]=(-0.853878344804)*x[0]**o+(-0.482644826561)*x[0]+(-0.826041149012)*x[1]**o+(0.596376234798)*x[1]+(-0.216400517274)*x[2]**o+(-0.323127194259)*x[2]
        arg[0,2,1,0]=(0.931551121492)*x[0]**o+(-0.527132174379)*x[0]+(0.373496704463)*x[1]**o+(-0.375215826463)*x[1]+(-0.639167297984)*x[2]**o+(0.143608733189)*x[2]
        arg[0,2,1,1]=(-0.982206000879)*x[0]**o+(0.836589596534)*x[0]+(0.606275000138)*x[1]**o+(-0.821469453892)*x[1]+(-0.633828380948)*x[2]**o+(-0.637688777216)*x[2]
        arg[0,2,2,0]=(0.441649557529)*x[0]**o+(-0.34978365051)*x[0]+(0.514175365405)*x[1]**o+(-0.903212271234)*x[1]+(0.70236887156)*x[2]**o+(0.72953045189)*x[2]
        arg[0,2,2,1]=(0.302787896978)*x[0]**o+(0.157598071021)*x[0]+(-0.357101947632)*x[1]**o+(0.381098383156)*x[1]+(-0.552546036689)*x[2]**o+(0.448181721984)*x[2]
        arg[0,3,0,0]=(0.662181077399)*x[0]**o+(-0.283755239518)*x[0]+(0.33694741278)*x[1]**o+(0.275801685868)*x[1]+(0.526155259101)*x[2]**o+(-0.378241855559)*x[2]
        arg[0,3,0,1]=(0.157822133581)*x[0]**o+(0.25402072747)*x[0]+(0.522842123099)*x[1]**o+(0.510595773584)*x[1]+(-0.829239126666)*x[2]**o+(0.990879342044)*x[2]
        arg[0,3,1,0]=(-0.775925280059)*x[0]**o+(-0.726971388023)*x[0]+(-0.630088730753)*x[1]**o+(-0.731024178237)*x[1]+(-0.288643541196)*x[2]**o+(-0.283708428793)*x[2]
        arg[0,3,1,1]=(-0.883176640445)*x[0]**o+(-0.0810211976762)*x[0]+(-0.102919217784)*x[1]**o+(0.271384866385)*x[1]+(0.980796900477)*x[2]**o+(0.439294038672)*x[2]
        arg[0,3,2,0]=(-0.826688221502)*x[0]**o+(-0.909388347347)*x[0]+(0.164406927722)*x[1]**o+(0.60157189364)*x[1]+(-0.885578176748)*x[2]**o+(0.610008674247)*x[2]
        arg[0,3,2,1]=(0.130154197193)*x[0]**o+(-0.754703602975)*x[0]+(0.00463710497017)*x[1]**o+(0.502162916123)*x[1]+(-0.749124548541)*x[2]**o+(0.946592962092)*x[2]
        arg[0,4,0,0]=(-0.0478347448906)*x[0]**o+(-0.812937871882)*x[0]+(-0.460982012551)*x[1]**o+(0.317485998991)*x[1]+(-0.593671683477)*x[2]**o+(0.551385718668)*x[2]
        arg[0,4,0,1]=(0.281998767502)*x[0]**o+(-0.0613830311997)*x[0]+(0.981429151019)*x[1]**o+(-0.580567714008)*x[1]+(0.0369483005331)*x[2]**o+(-0.74132358588)*x[2]
        arg[0,4,1,0]=(0.525845217483)*x[0]**o+(0.907803915216)*x[0]+(0.111429079263)*x[1]**o+(-0.653030401377)*x[1]+(0.0814793913696)*x[2]**o+(-0.419201413875)*x[2]
        arg[0,4,1,1]=(0.862892518463)*x[0]**o+(0.34033875398)*x[0]+(0.0371376953103)*x[1]**o+(0.568160310726)*x[1]+(-0.237231036717)*x[2]**o+(-0.130046899302)*x[2]
        arg[0,4,2,0]=(0.223651181698)*x[0]**o+(0.730991398203)*x[0]+(0.133551525027)*x[1]**o+(0.685783540126)*x[1]+(0.339586395814)*x[2]**o+(-0.524371194863)*x[2]
        arg[0,4,2,1]=(0.220110871383)*x[0]**o+(-0.37385773877)*x[0]+(0.773372000545)*x[1]**o+(-0.4261356565)*x[1]+(-0.619438438945)*x[2]**o+(0.126288949127)*x[2]
        arg[1,0,0,0]=(0.727680196631)*x[0]**o+(-0.876280912599)*x[0]+(0.576553782915)*x[1]**o+(-0.0497780322367)*x[1]+(-0.690257457282)*x[2]**o+(-0.11130953054)*x[2]
        arg[1,0,0,1]=(0.331863815938)*x[0]**o+(0.169379824389)*x[0]+(0.665507756012)*x[1]**o+(0.94933357273)*x[1]+(-0.610506070093)*x[2]**o+(0.752476708027)*x[2]
        arg[1,0,1,0]=(0.851320583415)*x[0]**o+(-0.697235996161)*x[0]+(0.235192165807)*x[1]**o+(0.325473517022)*x[1]+(-0.424683549673)*x[2]**o+(0.228953796486)*x[2]
        arg[1,0,1,1]=(0.804192098517)*x[0]**o+(-0.684337197059)*x[0]+(-0.349251615715)*x[1]**o+(0.596165705743)*x[1]+(-0.636033221182)*x[2]**o+(-0.304367076307)*x[2]
        arg[1,0,2,0]=(-0.237305477187)*x[0]**o+(-0.997679652719)*x[0]+(0.485994068389)*x[1]**o+(0.44508530534)*x[1]+(0.940139935628)*x[2]**o+(0.0789221540062)*x[2]
        arg[1,0,2,1]=(0.814928463364)*x[0]**o+(-0.257693939789)*x[0]+(0.179643785703)*x[1]**o+(-0.696436530567)*x[1]+(-0.592226945554)*x[2]**o+(-0.454620411499)*x[2]
        arg[1,1,0,0]=(-0.933925672439)*x[0]**o+(0.212298180817)*x[0]+(-0.447836316801)*x[1]**o+(0.180796376138)*x[1]+(-0.749910125701)*x[2]**o+(0.134540433851)*x[2]
        arg[1,1,0,1]=(-0.649009346978)*x[0]**o+(0.0316819485953)*x[0]+(-0.718001767835)*x[1]**o+(0.289608435124)*x[1]+(-0.906045493384)*x[2]**o+(0.851021484125)*x[2]
        arg[1,1,1,0]=(-0.216265404566)*x[0]**o+(0.657073282172)*x[0]+(-0.796155541438)*x[1]**o+(0.338980498824)*x[1]+(0.477444448193)*x[2]**o+(0.674278909522)*x[2]
        arg[1,1,1,1]=(-0.67605942954)*x[0]**o+(0.457997253216)*x[0]+(-0.973576381948)*x[1]**o+(0.430432753773)*x[1]+(0.0329169564451)*x[2]**o+(-0.422512702485)*x[2]
        arg[1,1,2,0]=(-0.32458256139)*x[0]**o+(0.289568625922)*x[0]+(-0.529564027476)*x[1]**o+(0.596449833118)*x[1]+(0.960461031989)*x[2]**o+(0.00801425321069)*x[2]
        arg[1,1,2,1]=(-0.0425883615923)*x[0]**o+(-0.313926807333)*x[0]+(0.766062773671)*x[1]**o+(-0.857784560309)*x[1]+(0.914867381682)*x[2]**o+(-0.312493334639)*x[2]
        arg[1,2,0,0]=(-0.637151767355)*x[0]**o+(-0.679947136952)*x[0]+(-0.0838349105494)*x[1]**o+(0.00800835879666)*x[1]+(0.385024194519)*x[2]**o+(-0.627828236571)*x[2]
        arg[1,2,0,1]=(-0.62580153007)*x[0]**o+(-0.183834415542)*x[0]+(0.322406823575)*x[1]**o+(0.302908755868)*x[1]+(0.727234001542)*x[2]**o+(-0.933944878425)*x[2]
        arg[1,2,1,0]=(-0.897198312615)*x[0]**o+(0.825086526945)*x[0]+(-0.031001758991)*x[1]**o+(0.987845551304)*x[1]+(-0.551212418179)*x[2]**o+(0.657813241764)*x[2]
        arg[1,2,1,1]=(-0.544595822278)*x[0]**o+(0.595858429591)*x[0]+(-0.557503238514)*x[1]**o+(0.507560179011)*x[1]+(-0.519438270771)*x[2]**o+(-0.471678264995)*x[2]
        arg[1,2,2,0]=(0.135616278099)*x[0]**o+(0.40676743342)*x[0]+(0.458810719213)*x[1]**o+(-0.881620410228)*x[1]+(-0.897267732846)*x[2]**o+(0.665458992487)*x[2]
        arg[1,2,2,1]=(0.83854610457)*x[0]**o+(0.670263196782)*x[0]+(0.551767734306)*x[1]**o+(0.883934821215)*x[1]+(0.455375261768)*x[2]**o+(0.95604480045)*x[2]
        arg[1,3,0,0]=(0.705853698882)*x[0]**o+(0.84681314245)*x[0]+(-0.0984020252408)*x[1]**o+(0.943165990555)*x[1]+(-0.428955841327)*x[2]**o+(0.0961618993003)*x[2]
        arg[1,3,0,1]=(0.583731800337)*x[0]**o+(-0.738079255003)*x[0]+(0.315556768944)*x[1]**o+(-0.489631118687)*x[1]+(0.371775779104)*x[2]**o+(0.82029490872)*x[2]
        arg[1,3,1,0]=(-0.885657545142)*x[0]**o+(0.542706113286)*x[0]+(0.800319452136)*x[1]**o+(0.314194184897)*x[1]+(-0.320905532427)*x[2]**o+(-0.314307365683)*x[2]
        arg[1,3,1,1]=(-0.217030771566)*x[0]**o+(0.260108519031)*x[0]+(-0.686624725975)*x[1]**o+(-0.869322439609)*x[1]+(0.558806239315)*x[2]**o+(-0.440661121246)*x[2]
        arg[1,3,2,0]=(-0.839335541382)*x[0]**o+(0.228726621639)*x[0]+(0.354403444002)*x[1]**o+(0.215616048959)*x[1]+(0.49622647562)*x[2]**o+(-0.725639112729)*x[2]
        arg[1,3,2,1]=(-0.439687404632)*x[0]**o+(-0.820583921269)*x[0]+(-0.863730409631)*x[1]**o+(0.119773401911)*x[1]+(0.460158483187)*x[2]**o+(0.757153436592)*x[2]
        arg[1,4,0,0]=(-0.77288839701)*x[0]**o+(-0.570203714622)*x[0]+(0.93832772305)*x[1]**o+(-0.759089944596)*x[1]+(0.000174937961704)*x[2]**o+(0.812917484286)*x[2]
        arg[1,4,0,1]=(-0.25535378432)*x[0]**o+(-0.0938207170929)*x[0]+(0.0545321717685)*x[1]**o+(-0.414414792674)*x[1]+(-0.297698896712)*x[2]**o+(-0.367166660571)*x[2]
        arg[1,4,1,0]=(0.743408112248)*x[0]**o+(0.852863814785)*x[0]+(0.259483951249)*x[1]**o+(-0.519288864163)*x[1]+(-0.305772267727)*x[2]**o+(-0.401384318883)*x[2]
        arg[1,4,1,1]=(0.194889815352)*x[0]**o+(0.522093986804)*x[0]+(-0.343320977396)*x[1]**o+(0.21840236597)*x[1]+(-0.672107409136)*x[2]**o+(-0.600203805002)*x[2]
        arg[1,4,2,0]=(0.420860556361)*x[0]**o+(0.789879869049)*x[0]+(0.100987805744)*x[1]**o+(0.802794815123)*x[1]+(-0.69466030201)*x[2]**o+(-0.746043845966)*x[2]
        arg[1,4,2,1]=(-0.306621579211)*x[0]**o+(-0.0676982346933)*x[0]+(0.568861573715)*x[1]**o+(0.0823366334578)*x[1]+(0.501187255641)*x[2]**o+(-0.538615164478)*x[2]
        arg[2,0,0,0]=(-0.231052916706)*x[0]**o+(0.127790831537)*x[0]+(-0.475600435038)*x[1]**o+(0.431586151473)*x[1]+(-0.882365547235)*x[2]**o+(0.778027701879)*x[2]
        arg[2,0,0,1]=(0.0709490835498)*x[0]**o+(0.442614693755)*x[0]+(0.870405247692)*x[1]**o+(-0.648538779515)*x[1]+(0.751107652054)*x[2]**o+(0.0122956931127)*x[2]
        arg[2,0,1,0]=(0.703128329145)*x[0]**o+(0.259775843631)*x[0]+(0.311958888594)*x[1]**o+(-0.0445092402277)*x[1]+(0.338222457799)*x[2]**o+(0.845995300456)*x[2]
        arg[2,0,1,1]=(-0.653977898919)*x[0]**o+(0.0495539350857)*x[0]+(0.4013623408)*x[1]**o+(0.490747039887)*x[1]+(-0.480487636729)*x[2]**o+(0.361660115775)*x[2]
        arg[2,0,2,0]=(0.215778506945)*x[0]**o+(0.316853875053)*x[0]+(0.556770639829)*x[1]**o+(-0.499729882355)*x[1]+(0.692012327582)*x[2]**o+(0.00681314946195)*x[2]
        arg[2,0,2,1]=(0.432077857223)*x[0]**o+(-0.0497579036988)*x[0]+(-0.465403408565)*x[1]**o+(0.125421502956)*x[1]+(-0.313541879877)*x[2]**o+(0.113957941731)*x[2]
        arg[2,1,0,0]=(-0.528126103192)*x[0]**o+(-0.141306512096)*x[0]+(-0.505835313198)*x[1]**o+(-0.0433171534408)*x[1]+(-0.522465687628)*x[2]**o+(-0.16728285732)*x[2]
        arg[2,1,0,1]=(0.892019660948)*x[0]**o+(0.646944130193)*x[0]+(0.865495756748)*x[1]**o+(0.949095276008)*x[1]+(0.786394755068)*x[2]**o+(0.482441081687)*x[2]
        arg[2,1,1,0]=(-0.0485295711204)*x[0]**o+(0.250163927911)*x[0]+(0.0387861125998)*x[1]**o+(-0.0789121929817)*x[1]+(-0.982330126583)*x[2]**o+(0.66328870814)*x[2]
        arg[2,1,1,1]=(-0.494726556678)*x[0]**o+(-0.961342091528)*x[0]+(0.211410563083)*x[1]**o+(-0.884287887286)*x[1]+(-0.583401908135)*x[2]**o+(-0.361423905219)*x[2]
        arg[2,1,2,0]=(-0.326785246692)*x[0]**o+(0.596272945325)*x[0]+(0.625039539234)*x[1]**o+(0.88699688986)*x[1]+(-0.649803810419)*x[2]**o+(0.890718271469)*x[2]
        arg[2,1,2,1]=(0.702110072391)*x[0]**o+(0.72009439563)*x[0]+(-0.930941572511)*x[1]**o+(0.955877158334)*x[1]+(0.1211281332)*x[2]**o+(0.658738626852)*x[2]
        arg[2,2,0,0]=(-0.177402589697)*x[0]**o+(0.701090894256)*x[0]+(-0.00205333079339)*x[1]**o+(-0.427862310859)*x[1]+(0.529186374803)*x[2]**o+(-0.110462035551)*x[2]
        arg[2,2,0,1]=(-0.104112165367)*x[0]**o+(0.043539351452)*x[0]+(0.757384935306)*x[1]**o+(-0.451086492844)*x[1]+(0.0549627297033)*x[2]**o+(-0.311718955315)*x[2]
        arg[2,2,1,0]=(-0.579986090808)*x[0]**o+(-0.922872917924)*x[0]+(0.66422700207)*x[1]**o+(0.947304780214)*x[1]+(-0.0492553517704)*x[2]**o+(0.817225619153)*x[2]
        arg[2,2,1,1]=(0.306298476233)*x[0]**o+(-0.620190960783)*x[0]+(-0.684876074013)*x[1]**o+(-0.371829616982)*x[1]+(0.904709743321)*x[2]**o+(0.170802273852)*x[2]
        arg[2,2,2,0]=(0.317943399758)*x[0]**o+(0.518026054212)*x[0]+(-0.801462311167)*x[1]**o+(0.767424157898)*x[1]+(-0.547271995492)*x[2]**o+(-0.912619177188)*x[2]
        arg[2,2,2,1]=(-0.848711258358)*x[0]**o+(0.456489097468)*x[0]+(-0.310979078367)*x[1]**o+(-0.946717007726)*x[1]+(0.454363833646)*x[2]**o+(-0.715701991077)*x[2]
        arg[2,3,0,0]=(0.605493533014)*x[0]**o+(0.478437880368)*x[0]+(-0.672727491605)*x[1]**o+(0.831259256009)*x[1]+(-0.434153354239)*x[2]**o+(-0.505363218578)*x[2]
        arg[2,3,0,1]=(-0.0703539676643)*x[0]**o+(-0.452352344215)*x[0]+(0.0403568720903)*x[1]**o+(0.959415939826)*x[1]+(-0.784519698477)*x[2]**o+(-0.691664442958)*x[2]
        arg[2,3,1,0]=(0.696353768719)*x[0]**o+(0.295477297618)*x[0]+(0.580540484825)*x[1]**o+(0.098954967023)*x[1]+(-0.547594444517)*x[2]**o+(-0.574783665389)*x[2]
        arg[2,3,1,1]=(0.997841887408)*x[0]**o+(0.857864040549)*x[0]+(-0.413188343971)*x[1]**o+(0.344770631687)*x[1]+(-0.127257058552)*x[2]**o+(0.132618819456)*x[2]
        arg[2,3,2,0]=(0.174045751083)*x[0]**o+(0.866724759298)*x[0]+(0.0818516103455)*x[1]**o+(-0.891670609599)*x[1]+(0.982721717044)*x[2]**o+(0.378710155856)*x[2]
        arg[2,3,2,1]=(0.7760778906)*x[0]**o+(-0.710104290654)*x[0]+(-0.471259845977)*x[1]**o+(-0.791979211724)*x[1]+(-0.845652753706)*x[2]**o+(-0.301795837852)*x[2]
        arg[2,4,0,0]=(-0.698193658421)*x[0]**o+(-0.831956135385)*x[0]+(-0.442637915133)*x[1]**o+(0.108715813637)*x[1]+(-0.888348149694)*x[2]**o+(-0.90115888496)*x[2]
        arg[2,4,0,1]=(0.770237705577)*x[0]**o+(0.889304763618)*x[0]+(-0.139196183564)*x[1]**o+(0.332516232239)*x[1]+(-0.0218103253052)*x[2]**o+(-0.368033928854)*x[2]
        arg[2,4,1,0]=(0.675479351844)*x[0]**o+(0.880998418303)*x[0]+(0.408381837105)*x[1]**o+(0.717402364084)*x[1]+(-0.303902947422)*x[2]**o+(-0.322350049962)*x[2]
        arg[2,4,1,1]=(-0.480881673289)*x[0]**o+(0.30290141529)*x[0]+(-0.538411390886)*x[1]**o+(0.260090848702)*x[1]+(0.638130764808)*x[2]**o+(0.756338975575)*x[2]
        arg[2,4,2,0]=(-0.299233221889)*x[0]**o+(-0.26471083761)*x[0]+(0.717191172884)*x[1]**o+(0.488553216118)*x[1]+(0.718956726644)*x[2]**o+(0.370889025703)*x[2]
        arg[2,4,2,1]=(-0.989020272269)*x[0]**o+(-0.891423884362)*x[0]+(0.486153100693)*x[1]**o+(-0.971171395246)*x[1]+(-0.766942968184)*x[2]**o+(0.728728968862)*x[2]
        arg[3,0,0,0]=(0.756186246227)*x[0]**o+(0.76507856268)*x[0]+(0.771165000414)*x[1]**o+(0.673968873856)*x[1]+(0.827045846253)*x[2]**o+(-0.788265752412)*x[2]
        arg[3,0,0,1]=(-0.66925731439)*x[0]**o+(-0.0110127596435)*x[0]+(0.944388255414)*x[1]**o+(-0.613832072486)*x[1]+(0.0977542120724)*x[2]**o+(0.100776856689)*x[2]
        arg[3,0,1,0]=(0.905138713064)*x[0]**o+(0.673051338153)*x[0]+(0.240614253653)*x[1]**o+(-0.964132330033)*x[1]+(-0.114961018838)*x[2]**o+(-0.438756962988)*x[2]
        arg[3,0,1,1]=(-0.192259182094)*x[0]**o+(0.12315590119)*x[0]+(-0.907744420226)*x[1]**o+(-0.483981269809)*x[1]+(-0.169699431472)*x[2]**o+(-0.45231635406)*x[2]
        arg[3,0,2,0]=(0.895387255885)*x[0]**o+(0.23892997411)*x[0]+(-0.396010446618)*x[1]**o+(0.94853984413)*x[1]+(-0.0165259637241)*x[2]**o+(-0.312558539149)*x[2]
        arg[3,0,2,1]=(0.710478713361)*x[0]**o+(-0.109316102216)*x[0]+(-0.042711568986)*x[1]**o+(-0.448508085742)*x[1]+(-0.0979193346011)*x[2]**o+(0.93156672211)*x[2]
        arg[3,1,0,0]=(-0.21605108221)*x[0]**o+(0.410466481712)*x[0]+(0.799288386783)*x[1]**o+(0.622866017113)*x[1]+(-0.236698938159)*x[2]**o+(0.246625241928)*x[2]
        arg[3,1,0,1]=(0.989568527762)*x[0]**o+(0.0212906378151)*x[0]+(0.311279947324)*x[1]**o+(-0.231412902902)*x[1]+(-0.444760786142)*x[2]**o+(0.908171985421)*x[2]
        arg[3,1,1,0]=(0.42605234231)*x[0]**o+(-0.58629332944)*x[0]+(0.322063640761)*x[1]**o+(0.837771339825)*x[1]+(-0.911083211394)*x[2]**o+(-0.0699607581992)*x[2]
        arg[3,1,1,1]=(0.75449445591)*x[0]**o+(0.722970800307)*x[0]+(-0.582970919465)*x[1]**o+(-0.868818538212)*x[1]+(0.940067636181)*x[2]**o+(-0.659469743617)*x[2]
        arg[3,1,2,0]=(-0.391224924663)*x[0]**o+(-0.768173177737)*x[0]+(-0.856876347997)*x[1]**o+(-0.0451180499647)*x[1]+(0.198590679246)*x[2]**o+(0.187329095476)*x[2]
        arg[3,1,2,1]=(0.483309063092)*x[0]**o+(-0.0777241026893)*x[0]+(0.0438427457286)*x[1]**o+(-0.73195704867)*x[1]+(-0.439660745725)*x[2]**o+(0.279501672861)*x[2]
        arg[3,2,0,0]=(0.799935089139)*x[0]**o+(-0.306860809857)*x[0]+(0.169104975039)*x[1]**o+(-0.3480717525)*x[1]+(0.215231688)*x[2]**o+(-0.528775376945)*x[2]
        arg[3,2,0,1]=(0.917815789587)*x[0]**o+(0.0672910338102)*x[0]+(-0.186550440011)*x[1]**o+(0.922579714252)*x[1]+(0.924162379857)*x[2]**o+(-0.769320483618)*x[2]
        arg[3,2,1,0]=(0.604324468058)*x[0]**o+(-0.0109731419053)*x[0]+(0.229447318325)*x[1]**o+(-0.480693435232)*x[1]+(-0.881779256999)*x[2]**o+(0.288656394771)*x[2]
        arg[3,2,1,1]=(0.934680723352)*x[0]**o+(0.993838528376)*x[0]+(0.957781588502)*x[1]**o+(-0.859837057866)*x[1]+(-0.702139964493)*x[2]**o+(0.0538177124609)*x[2]
        arg[3,2,2,0]=(-0.142809047814)*x[0]**o+(-0.472594415134)*x[0]+(0.0381782768377)*x[1]**o+(0.727331491769)*x[1]+(0.0910992897581)*x[2]**o+(-0.845316063851)*x[2]
        arg[3,2,2,1]=(-0.217483358272)*x[0]**o+(0.991535092285)*x[0]+(-0.399175896468)*x[1]**o+(-0.665836512895)*x[1]+(0.0983700911493)*x[2]**o+(0.321811300155)*x[2]
        arg[3,3,0,0]=(0.797049585464)*x[0]**o+(-0.724467959299)*x[0]+(0.19971589973)*x[1]**o+(0.845928461864)*x[1]+(-0.365158866359)*x[2]**o+(-0.698274270132)*x[2]
        arg[3,3,0,1]=(-0.638474704878)*x[0]**o+(0.551866433352)*x[0]+(0.192309452979)*x[1]**o+(-0.703649325501)*x[1]+(0.823910578138)*x[2]**o+(-0.917446989634)*x[2]
        arg[3,3,1,0]=(0.217577282169)*x[0]**o+(0.521716803668)*x[0]+(-0.116587783856)*x[1]**o+(0.529672391417)*x[1]+(-0.798918158284)*x[2]**o+(-0.0996139511955)*x[2]
        arg[3,3,1,1]=(0.498965049079)*x[0]**o+(0.316489921186)*x[0]+(-0.0988355471817)*x[1]**o+(-0.0477508287681)*x[1]+(-0.778914523993)*x[2]**o+(0.189397812723)*x[2]
        arg[3,3,2,0]=(0.203823698152)*x[0]**o+(-0.340499414266)*x[0]+(0.353374693517)*x[1]**o+(-0.91035264494)*x[1]+(-0.779204084198)*x[2]**o+(-0.00276454165248)*x[2]
        arg[3,3,2,1]=(0.295232007123)*x[0]**o+(0.271766544502)*x[0]+(0.431956430336)*x[1]**o+(0.896481311536)*x[1]+(-0.325014902487)*x[2]**o+(0.70836170284)*x[2]
        arg[3,4,0,0]=(-0.407535202468)*x[0]**o+(-0.33055681091)*x[0]+(0.129762143151)*x[1]**o+(0.352087436228)*x[1]+(-0.538770401565)*x[2]**o+(-0.519071077045)*x[2]
        arg[3,4,0,1]=(0.846011435499)*x[0]**o+(0.172822133672)*x[0]+(-0.740732553847)*x[1]**o+(0.230062837231)*x[1]+(-0.231810007237)*x[2]**o+(-0.568366338923)*x[2]
        arg[3,4,1,0]=(-0.198091119057)*x[0]**o+(0.957989508794)*x[0]+(0.861739071175)*x[1]**o+(0.923450697883)*x[1]+(-0.755262504937)*x[2]**o+(0.102897579936)*x[2]
        arg[3,4,1,1]=(0.198829545353)*x[0]**o+(-0.757182006818)*x[0]+(0.254092337808)*x[1]**o+(-0.00815072620097)*x[1]+(0.347421988308)*x[2]**o+(-0.0215277616923)*x[2]
        arg[3,4,2,0]=(0.775062112529)*x[0]**o+(-0.77746792774)*x[0]+(-0.666490294869)*x[1]**o+(-0.679971069878)*x[1]+(-0.972450779894)*x[2]**o+(-0.705322282477)*x[2]
        arg[3,4,2,1]=(-0.0994927587382)*x[0]**o+(0.293071774786)*x[0]+(0.370784399397)*x[1]**o+(0.931411225946)*x[1]+(-0.55789163294)*x[2]**o+(0.0831890658592)*x[2]
        ref[0,0,0,0]=(-0.657703355307)*x_ref[0]**o+(0.397813831843)*x_ref[0]+(0.364460538717)*x_ref[1]**o+(-0.845626669617)*x_ref[1]+(-0.987144254126)*x_ref[2]**o+(-0.732384980726)*x_ref[2]
        ref[0,0,0,1]=(-0.13280304886)*x_ref[0]**o+(-0.953796807758)*x_ref[0]+(0.875723746654)*x_ref[1]**o+(0.0764745117242)*x_ref[1]+(-0.536085676221)*x_ref[2]**o+(-0.520611959678)*x_ref[2]
        ref[0,0,1,0]=(0.716429446578)*x_ref[0]**o+(-0.729164693441)*x_ref[0]+(-0.832505557428)*x_ref[1]**o+(-0.329200605611)*x_ref[1]+(0.0310162060964)*x_ref[2]**o+(-0.869978941999)*x_ref[2]
        ref[0,0,1,1]=(0.110253530584)*x_ref[0]**o+(0.995528486822)*x_ref[0]+(0.838577797395)*x_ref[1]**o+(0.443987752225)*x_ref[1]+(0.387337959057)*x_ref[2]**o+(0.548448752612)*x_ref[2]
        ref[0,0,2,0]=(-0.995737515431)*x_ref[0]**o+(-0.899807241717)*x_ref[0]+(-0.0803287764138)*x_ref[1]**o+(0.242680051556)*x_ref[1]+(0.58974836823)*x_ref[2]**o+(-0.498450188944)*x_ref[2]
        ref[0,0,2,1]=(0.341561541555)*x_ref[0]**o+(-0.00541988184597)*x_ref[0]+(-0.204832898701)*x_ref[1]**o+(-0.957247935865)*x_ref[1]+(-0.441002902064)*x_ref[2]**o+(0.756843480304)*x_ref[2]
        ref[0,1,0,0]=(0.136201526244)*x_ref[0]**o+(0.517677304193)*x_ref[0]+(0.523031610748)*x_ref[1]**o+(-0.908998053383)*x_ref[1]+(-0.629649678416)*x_ref[2]**o+(0.762977917512)*x_ref[2]
        ref[0,1,0,1]=(-0.103521080864)*x_ref[0]**o+(-0.434410222595)*x_ref[0]+(0.418239530241)*x_ref[1]**o+(0.65638475119)*x_ref[1]+(0.413036393129)*x_ref[2]**o+(-0.220365518353)*x_ref[2]
        ref[0,1,1,0]=(-0.415690201483)*x_ref[0]**o+(-0.969323394689)*x_ref[0]+(-0.295639420195)*x_ref[1]**o+(-0.24915002749)*x_ref[1]+(-0.00927627101467)*x_ref[2]**o+(0.520335261634)*x_ref[2]
        ref[0,1,1,1]=(-0.262653741797)*x_ref[0]**o+(-0.463590591666)*x_ref[0]+(0.568893042923)*x_ref[1]**o+(-0.869851332898)*x_ref[1]+(-0.463261399415)*x_ref[2]**o+(0.0281915686202)*x_ref[2]
        ref[0,1,2,0]=(0.377159503058)*x_ref[0]**o+(0.484512261943)*x_ref[0]+(0.0303594628925)*x_ref[1]**o+(0.482051243004)*x_ref[1]+(0.240643013949)*x_ref[2]**o+(-0.943908617705)*x_ref[2]
        ref[0,1,2,1]=(0.843854614097)*x_ref[0]**o+(0.018041534553)*x_ref[0]+(0.264198703361)*x_ref[1]**o+(0.732889517628)*x_ref[1]+(-0.154807384394)*x_ref[2]**o+(0.60513622187)*x_ref[2]
        ref[0,2,0,0]=(-0.234758118286)*x_ref[0]**o+(-0.77807396119)*x_ref[0]+(-0.97783091103)*x_ref[1]**o+(-0.716185223828)*x_ref[1]+(-0.573126474259)*x_ref[2]**o+(0.403110701555)*x_ref[2]
        ref[0,2,0,1]=(-0.853878344804)*x_ref[0]**o+(-0.482644826561)*x_ref[0]+(-0.826041149012)*x_ref[1]**o+(0.596376234798)*x_ref[1]+(-0.216400517274)*x_ref[2]**o+(-0.323127194259)*x_ref[2]
        ref[0,2,1,0]=(0.931551121492)*x_ref[0]**o+(-0.527132174379)*x_ref[0]+(0.373496704463)*x_ref[1]**o+(-0.375215826463)*x_ref[1]+(-0.639167297984)*x_ref[2]**o+(0.143608733189)*x_ref[2]
        ref[0,2,1,1]=(-0.982206000879)*x_ref[0]**o+(0.836589596534)*x_ref[0]+(0.606275000138)*x_ref[1]**o+(-0.821469453892)*x_ref[1]+(-0.633828380948)*x_ref[2]**o+(-0.637688777216)*x_ref[2]
        ref[0,2,2,0]=(0.441649557529)*x_ref[0]**o+(-0.34978365051)*x_ref[0]+(0.514175365405)*x_ref[1]**o+(-0.903212271234)*x_ref[1]+(0.70236887156)*x_ref[2]**o+(0.72953045189)*x_ref[2]
        ref[0,2,2,1]=(0.302787896978)*x_ref[0]**o+(0.157598071021)*x_ref[0]+(-0.357101947632)*x_ref[1]**o+(0.381098383156)*x_ref[1]+(-0.552546036689)*x_ref[2]**o+(0.448181721984)*x_ref[2]
        ref[0,3,0,0]=(0.662181077399)*x_ref[0]**o+(-0.283755239518)*x_ref[0]+(0.33694741278)*x_ref[1]**o+(0.275801685868)*x_ref[1]+(0.526155259101)*x_ref[2]**o+(-0.378241855559)*x_ref[2]
        ref[0,3,0,1]=(0.157822133581)*x_ref[0]**o+(0.25402072747)*x_ref[0]+(0.522842123099)*x_ref[1]**o+(0.510595773584)*x_ref[1]+(-0.829239126666)*x_ref[2]**o+(0.990879342044)*x_ref[2]
        ref[0,3,1,0]=(-0.775925280059)*x_ref[0]**o+(-0.726971388023)*x_ref[0]+(-0.630088730753)*x_ref[1]**o+(-0.731024178237)*x_ref[1]+(-0.288643541196)*x_ref[2]**o+(-0.283708428793)*x_ref[2]
        ref[0,3,1,1]=(-0.883176640445)*x_ref[0]**o+(-0.0810211976762)*x_ref[0]+(-0.102919217784)*x_ref[1]**o+(0.271384866385)*x_ref[1]+(0.980796900477)*x_ref[2]**o+(0.439294038672)*x_ref[2]
        ref[0,3,2,0]=(-0.826688221502)*x_ref[0]**o+(-0.909388347347)*x_ref[0]+(0.164406927722)*x_ref[1]**o+(0.60157189364)*x_ref[1]+(-0.885578176748)*x_ref[2]**o+(0.610008674247)*x_ref[2]
        ref[0,3,2,1]=(0.130154197193)*x_ref[0]**o+(-0.754703602975)*x_ref[0]+(0.00463710497017)*x_ref[1]**o+(0.502162916123)*x_ref[1]+(-0.749124548541)*x_ref[2]**o+(0.946592962092)*x_ref[2]
        ref[0,4,0,0]=(-0.0478347448906)*x_ref[0]**o+(-0.812937871882)*x_ref[0]+(-0.460982012551)*x_ref[1]**o+(0.317485998991)*x_ref[1]+(-0.593671683477)*x_ref[2]**o+(0.551385718668)*x_ref[2]
        ref[0,4,0,1]=(0.281998767502)*x_ref[0]**o+(-0.0613830311997)*x_ref[0]+(0.981429151019)*x_ref[1]**o+(-0.580567714008)*x_ref[1]+(0.0369483005331)*x_ref[2]**o+(-0.74132358588)*x_ref[2]
        ref[0,4,1,0]=(0.525845217483)*x_ref[0]**o+(0.907803915216)*x_ref[0]+(0.111429079263)*x_ref[1]**o+(-0.653030401377)*x_ref[1]+(0.0814793913696)*x_ref[2]**o+(-0.419201413875)*x_ref[2]
        ref[0,4,1,1]=(0.862892518463)*x_ref[0]**o+(0.34033875398)*x_ref[0]+(0.0371376953103)*x_ref[1]**o+(0.568160310726)*x_ref[1]+(-0.237231036717)*x_ref[2]**o+(-0.130046899302)*x_ref[2]
        ref[0,4,2,0]=(0.223651181698)*x_ref[0]**o+(0.730991398203)*x_ref[0]+(0.133551525027)*x_ref[1]**o+(0.685783540126)*x_ref[1]+(0.339586395814)*x_ref[2]**o+(-0.524371194863)*x_ref[2]
        ref[0,4,2,1]=(0.220110871383)*x_ref[0]**o+(-0.37385773877)*x_ref[0]+(0.773372000545)*x_ref[1]**o+(-0.4261356565)*x_ref[1]+(-0.619438438945)*x_ref[2]**o+(0.126288949127)*x_ref[2]
        ref[1,0,0,0]=(0.727680196631)*x_ref[0]**o+(-0.876280912599)*x_ref[0]+(0.576553782915)*x_ref[1]**o+(-0.0497780322367)*x_ref[1]+(-0.690257457282)*x_ref[2]**o+(-0.11130953054)*x_ref[2]
        ref[1,0,0,1]=(0.331863815938)*x_ref[0]**o+(0.169379824389)*x_ref[0]+(0.665507756012)*x_ref[1]**o+(0.94933357273)*x_ref[1]+(-0.610506070093)*x_ref[2]**o+(0.752476708027)*x_ref[2]
        ref[1,0,1,0]=(0.851320583415)*x_ref[0]**o+(-0.697235996161)*x_ref[0]+(0.235192165807)*x_ref[1]**o+(0.325473517022)*x_ref[1]+(-0.424683549673)*x_ref[2]**o+(0.228953796486)*x_ref[2]
        ref[1,0,1,1]=(0.804192098517)*x_ref[0]**o+(-0.684337197059)*x_ref[0]+(-0.349251615715)*x_ref[1]**o+(0.596165705743)*x_ref[1]+(-0.636033221182)*x_ref[2]**o+(-0.304367076307)*x_ref[2]
        ref[1,0,2,0]=(-0.237305477187)*x_ref[0]**o+(-0.997679652719)*x_ref[0]+(0.485994068389)*x_ref[1]**o+(0.44508530534)*x_ref[1]+(0.940139935628)*x_ref[2]**o+(0.0789221540062)*x_ref[2]
        ref[1,0,2,1]=(0.814928463364)*x_ref[0]**o+(-0.257693939789)*x_ref[0]+(0.179643785703)*x_ref[1]**o+(-0.696436530567)*x_ref[1]+(-0.592226945554)*x_ref[2]**o+(-0.454620411499)*x_ref[2]
        ref[1,1,0,0]=(-0.933925672439)*x_ref[0]**o+(0.212298180817)*x_ref[0]+(-0.447836316801)*x_ref[1]**o+(0.180796376138)*x_ref[1]+(-0.749910125701)*x_ref[2]**o+(0.134540433851)*x_ref[2]
        ref[1,1,0,1]=(-0.649009346978)*x_ref[0]**o+(0.0316819485953)*x_ref[0]+(-0.718001767835)*x_ref[1]**o+(0.289608435124)*x_ref[1]+(-0.906045493384)*x_ref[2]**o+(0.851021484125)*x_ref[2]
        ref[1,1,1,0]=(-0.216265404566)*x_ref[0]**o+(0.657073282172)*x_ref[0]+(-0.796155541438)*x_ref[1]**o+(0.338980498824)*x_ref[1]+(0.477444448193)*x_ref[2]**o+(0.674278909522)*x_ref[2]
        ref[1,1,1,1]=(-0.67605942954)*x_ref[0]**o+(0.457997253216)*x_ref[0]+(-0.973576381948)*x_ref[1]**o+(0.430432753773)*x_ref[1]+(0.0329169564451)*x_ref[2]**o+(-0.422512702485)*x_ref[2]
        ref[1,1,2,0]=(-0.32458256139)*x_ref[0]**o+(0.289568625922)*x_ref[0]+(-0.529564027476)*x_ref[1]**o+(0.596449833118)*x_ref[1]+(0.960461031989)*x_ref[2]**o+(0.00801425321069)*x_ref[2]
        ref[1,1,2,1]=(-0.0425883615923)*x_ref[0]**o+(-0.313926807333)*x_ref[0]+(0.766062773671)*x_ref[1]**o+(-0.857784560309)*x_ref[1]+(0.914867381682)*x_ref[2]**o+(-0.312493334639)*x_ref[2]
        ref[1,2,0,0]=(-0.637151767355)*x_ref[0]**o+(-0.679947136952)*x_ref[0]+(-0.0838349105494)*x_ref[1]**o+(0.00800835879666)*x_ref[1]+(0.385024194519)*x_ref[2]**o+(-0.627828236571)*x_ref[2]
        ref[1,2,0,1]=(-0.62580153007)*x_ref[0]**o+(-0.183834415542)*x_ref[0]+(0.322406823575)*x_ref[1]**o+(0.302908755868)*x_ref[1]+(0.727234001542)*x_ref[2]**o+(-0.933944878425)*x_ref[2]
        ref[1,2,1,0]=(-0.897198312615)*x_ref[0]**o+(0.825086526945)*x_ref[0]+(-0.031001758991)*x_ref[1]**o+(0.987845551304)*x_ref[1]+(-0.551212418179)*x_ref[2]**o+(0.657813241764)*x_ref[2]
        ref[1,2,1,1]=(-0.544595822278)*x_ref[0]**o+(0.595858429591)*x_ref[0]+(-0.557503238514)*x_ref[1]**o+(0.507560179011)*x_ref[1]+(-0.519438270771)*x_ref[2]**o+(-0.471678264995)*x_ref[2]
        ref[1,2,2,0]=(0.135616278099)*x_ref[0]**o+(0.40676743342)*x_ref[0]+(0.458810719213)*x_ref[1]**o+(-0.881620410228)*x_ref[1]+(-0.897267732846)*x_ref[2]**o+(0.665458992487)*x_ref[2]
        ref[1,2,2,1]=(0.83854610457)*x_ref[0]**o+(0.670263196782)*x_ref[0]+(0.551767734306)*x_ref[1]**o+(0.883934821215)*x_ref[1]+(0.455375261768)*x_ref[2]**o+(0.95604480045)*x_ref[2]
        ref[1,3,0,0]=(0.705853698882)*x_ref[0]**o+(0.84681314245)*x_ref[0]+(-0.0984020252408)*x_ref[1]**o+(0.943165990555)*x_ref[1]+(-0.428955841327)*x_ref[2]**o+(0.0961618993003)*x_ref[2]
        ref[1,3,0,1]=(0.583731800337)*x_ref[0]**o+(-0.738079255003)*x_ref[0]+(0.315556768944)*x_ref[1]**o+(-0.489631118687)*x_ref[1]+(0.371775779104)*x_ref[2]**o+(0.82029490872)*x_ref[2]
        ref[1,3,1,0]=(-0.885657545142)*x_ref[0]**o+(0.542706113286)*x_ref[0]+(0.800319452136)*x_ref[1]**o+(0.314194184897)*x_ref[1]+(-0.320905532427)*x_ref[2]**o+(-0.314307365683)*x_ref[2]
        ref[1,3,1,1]=(-0.217030771566)*x_ref[0]**o+(0.260108519031)*x_ref[0]+(-0.686624725975)*x_ref[1]**o+(-0.869322439609)*x_ref[1]+(0.558806239315)*x_ref[2]**o+(-0.440661121246)*x_ref[2]
        ref[1,3,2,0]=(-0.839335541382)*x_ref[0]**o+(0.228726621639)*x_ref[0]+(0.354403444002)*x_ref[1]**o+(0.215616048959)*x_ref[1]+(0.49622647562)*x_ref[2]**o+(-0.725639112729)*x_ref[2]
        ref[1,3,2,1]=(-0.439687404632)*x_ref[0]**o+(-0.820583921269)*x_ref[0]+(-0.863730409631)*x_ref[1]**o+(0.119773401911)*x_ref[1]+(0.460158483187)*x_ref[2]**o+(0.757153436592)*x_ref[2]
        ref[1,4,0,0]=(-0.77288839701)*x_ref[0]**o+(-0.570203714622)*x_ref[0]+(0.93832772305)*x_ref[1]**o+(-0.759089944596)*x_ref[1]+(0.000174937961704)*x_ref[2]**o+(0.812917484286)*x_ref[2]
        ref[1,4,0,1]=(-0.25535378432)*x_ref[0]**o+(-0.0938207170929)*x_ref[0]+(0.0545321717685)*x_ref[1]**o+(-0.414414792674)*x_ref[1]+(-0.297698896712)*x_ref[2]**o+(-0.367166660571)*x_ref[2]
        ref[1,4,1,0]=(0.743408112248)*x_ref[0]**o+(0.852863814785)*x_ref[0]+(0.259483951249)*x_ref[1]**o+(-0.519288864163)*x_ref[1]+(-0.305772267727)*x_ref[2]**o+(-0.401384318883)*x_ref[2]
        ref[1,4,1,1]=(0.194889815352)*x_ref[0]**o+(0.522093986804)*x_ref[0]+(-0.343320977396)*x_ref[1]**o+(0.21840236597)*x_ref[1]+(-0.672107409136)*x_ref[2]**o+(-0.600203805002)*x_ref[2]
        ref[1,4,2,0]=(0.420860556361)*x_ref[0]**o+(0.789879869049)*x_ref[0]+(0.100987805744)*x_ref[1]**o+(0.802794815123)*x_ref[1]+(-0.69466030201)*x_ref[2]**o+(-0.746043845966)*x_ref[2]
        ref[1,4,2,1]=(-0.306621579211)*x_ref[0]**o+(-0.0676982346933)*x_ref[0]+(0.568861573715)*x_ref[1]**o+(0.0823366334578)*x_ref[1]+(0.501187255641)*x_ref[2]**o+(-0.538615164478)*x_ref[2]
        ref[2,0,0,0]=(-0.231052916706)*x_ref[0]**o+(0.127790831537)*x_ref[0]+(-0.475600435038)*x_ref[1]**o+(0.431586151473)*x_ref[1]+(-0.882365547235)*x_ref[2]**o+(0.778027701879)*x_ref[2]
        ref[2,0,0,1]=(0.0709490835498)*x_ref[0]**o+(0.442614693755)*x_ref[0]+(0.870405247692)*x_ref[1]**o+(-0.648538779515)*x_ref[1]+(0.751107652054)*x_ref[2]**o+(0.0122956931127)*x_ref[2]
        ref[2,0,1,0]=(0.703128329145)*x_ref[0]**o+(0.259775843631)*x_ref[0]+(0.311958888594)*x_ref[1]**o+(-0.0445092402277)*x_ref[1]+(0.338222457799)*x_ref[2]**o+(0.845995300456)*x_ref[2]
        ref[2,0,1,1]=(-0.653977898919)*x_ref[0]**o+(0.0495539350857)*x_ref[0]+(0.4013623408)*x_ref[1]**o+(0.490747039887)*x_ref[1]+(-0.480487636729)*x_ref[2]**o+(0.361660115775)*x_ref[2]
        ref[2,0,2,0]=(0.215778506945)*x_ref[0]**o+(0.316853875053)*x_ref[0]+(0.556770639829)*x_ref[1]**o+(-0.499729882355)*x_ref[1]+(0.692012327582)*x_ref[2]**o+(0.00681314946195)*x_ref[2]
        ref[2,0,2,1]=(0.432077857223)*x_ref[0]**o+(-0.0497579036988)*x_ref[0]+(-0.465403408565)*x_ref[1]**o+(0.125421502956)*x_ref[1]+(-0.313541879877)*x_ref[2]**o+(0.113957941731)*x_ref[2]
        ref[2,1,0,0]=(-0.528126103192)*x_ref[0]**o+(-0.141306512096)*x_ref[0]+(-0.505835313198)*x_ref[1]**o+(-0.0433171534408)*x_ref[1]+(-0.522465687628)*x_ref[2]**o+(-0.16728285732)*x_ref[2]
        ref[2,1,0,1]=(0.892019660948)*x_ref[0]**o+(0.646944130193)*x_ref[0]+(0.865495756748)*x_ref[1]**o+(0.949095276008)*x_ref[1]+(0.786394755068)*x_ref[2]**o+(0.482441081687)*x_ref[2]
        ref[2,1,1,0]=(-0.0485295711204)*x_ref[0]**o+(0.250163927911)*x_ref[0]+(0.0387861125998)*x_ref[1]**o+(-0.0789121929817)*x_ref[1]+(-0.982330126583)*x_ref[2]**o+(0.66328870814)*x_ref[2]
        ref[2,1,1,1]=(-0.494726556678)*x_ref[0]**o+(-0.961342091528)*x_ref[0]+(0.211410563083)*x_ref[1]**o+(-0.884287887286)*x_ref[1]+(-0.583401908135)*x_ref[2]**o+(-0.361423905219)*x_ref[2]
        ref[2,1,2,0]=(-0.326785246692)*x_ref[0]**o+(0.596272945325)*x_ref[0]+(0.625039539234)*x_ref[1]**o+(0.88699688986)*x_ref[1]+(-0.649803810419)*x_ref[2]**o+(0.890718271469)*x_ref[2]
        ref[2,1,2,1]=(0.702110072391)*x_ref[0]**o+(0.72009439563)*x_ref[0]+(-0.930941572511)*x_ref[1]**o+(0.955877158334)*x_ref[1]+(0.1211281332)*x_ref[2]**o+(0.658738626852)*x_ref[2]
        ref[2,2,0,0]=(-0.177402589697)*x_ref[0]**o+(0.701090894256)*x_ref[0]+(-0.00205333079339)*x_ref[1]**o+(-0.427862310859)*x_ref[1]+(0.529186374803)*x_ref[2]**o+(-0.110462035551)*x_ref[2]
        ref[2,2,0,1]=(-0.104112165367)*x_ref[0]**o+(0.043539351452)*x_ref[0]+(0.757384935306)*x_ref[1]**o+(-0.451086492844)*x_ref[1]+(0.0549627297033)*x_ref[2]**o+(-0.311718955315)*x_ref[2]
        ref[2,2,1,0]=(-0.579986090808)*x_ref[0]**o+(-0.922872917924)*x_ref[0]+(0.66422700207)*x_ref[1]**o+(0.947304780214)*x_ref[1]+(-0.0492553517704)*x_ref[2]**o+(0.817225619153)*x_ref[2]
        ref[2,2,1,1]=(0.306298476233)*x_ref[0]**o+(-0.620190960783)*x_ref[0]+(-0.684876074013)*x_ref[1]**o+(-0.371829616982)*x_ref[1]+(0.904709743321)*x_ref[2]**o+(0.170802273852)*x_ref[2]
        ref[2,2,2,0]=(0.317943399758)*x_ref[0]**o+(0.518026054212)*x_ref[0]+(-0.801462311167)*x_ref[1]**o+(0.767424157898)*x_ref[1]+(-0.547271995492)*x_ref[2]**o+(-0.912619177188)*x_ref[2]
        ref[2,2,2,1]=(-0.848711258358)*x_ref[0]**o+(0.456489097468)*x_ref[0]+(-0.310979078367)*x_ref[1]**o+(-0.946717007726)*x_ref[1]+(0.454363833646)*x_ref[2]**o+(-0.715701991077)*x_ref[2]
        ref[2,3,0,0]=(0.605493533014)*x_ref[0]**o+(0.478437880368)*x_ref[0]+(-0.672727491605)*x_ref[1]**o+(0.831259256009)*x_ref[1]+(-0.434153354239)*x_ref[2]**o+(-0.505363218578)*x_ref[2]
        ref[2,3,0,1]=(-0.0703539676643)*x_ref[0]**o+(-0.452352344215)*x_ref[0]+(0.0403568720903)*x_ref[1]**o+(0.959415939826)*x_ref[1]+(-0.784519698477)*x_ref[2]**o+(-0.691664442958)*x_ref[2]
        ref[2,3,1,0]=(0.696353768719)*x_ref[0]**o+(0.295477297618)*x_ref[0]+(0.580540484825)*x_ref[1]**o+(0.098954967023)*x_ref[1]+(-0.547594444517)*x_ref[2]**o+(-0.574783665389)*x_ref[2]
        ref[2,3,1,1]=(0.997841887408)*x_ref[0]**o+(0.857864040549)*x_ref[0]+(-0.413188343971)*x_ref[1]**o+(0.344770631687)*x_ref[1]+(-0.127257058552)*x_ref[2]**o+(0.132618819456)*x_ref[2]
        ref[2,3,2,0]=(0.174045751083)*x_ref[0]**o+(0.866724759298)*x_ref[0]+(0.0818516103455)*x_ref[1]**o+(-0.891670609599)*x_ref[1]+(0.982721717044)*x_ref[2]**o+(0.378710155856)*x_ref[2]
        ref[2,3,2,1]=(0.7760778906)*x_ref[0]**o+(-0.710104290654)*x_ref[0]+(-0.471259845977)*x_ref[1]**o+(-0.791979211724)*x_ref[1]+(-0.845652753706)*x_ref[2]**o+(-0.301795837852)*x_ref[2]
        ref[2,4,0,0]=(-0.698193658421)*x_ref[0]**o+(-0.831956135385)*x_ref[0]+(-0.442637915133)*x_ref[1]**o+(0.108715813637)*x_ref[1]+(-0.888348149694)*x_ref[2]**o+(-0.90115888496)*x_ref[2]
        ref[2,4,0,1]=(0.770237705577)*x_ref[0]**o+(0.889304763618)*x_ref[0]+(-0.139196183564)*x_ref[1]**o+(0.332516232239)*x_ref[1]+(-0.0218103253052)*x_ref[2]**o+(-0.368033928854)*x_ref[2]
        ref[2,4,1,0]=(0.675479351844)*x_ref[0]**o+(0.880998418303)*x_ref[0]+(0.408381837105)*x_ref[1]**o+(0.717402364084)*x_ref[1]+(-0.303902947422)*x_ref[2]**o+(-0.322350049962)*x_ref[2]
        ref[2,4,1,1]=(-0.480881673289)*x_ref[0]**o+(0.30290141529)*x_ref[0]+(-0.538411390886)*x_ref[1]**o+(0.260090848702)*x_ref[1]+(0.638130764808)*x_ref[2]**o+(0.756338975575)*x_ref[2]
        ref[2,4,2,0]=(-0.299233221889)*x_ref[0]**o+(-0.26471083761)*x_ref[0]+(0.717191172884)*x_ref[1]**o+(0.488553216118)*x_ref[1]+(0.718956726644)*x_ref[2]**o+(0.370889025703)*x_ref[2]
        ref[2,4,2,1]=(-0.989020272269)*x_ref[0]**o+(-0.891423884362)*x_ref[0]+(0.486153100693)*x_ref[1]**o+(-0.971171395246)*x_ref[1]+(-0.766942968184)*x_ref[2]**o+(0.728728968862)*x_ref[2]
        ref[3,0,0,0]=(0.756186246227)*x_ref[0]**o+(0.76507856268)*x_ref[0]+(0.771165000414)*x_ref[1]**o+(0.673968873856)*x_ref[1]+(0.827045846253)*x_ref[2]**o+(-0.788265752412)*x_ref[2]
        ref[3,0,0,1]=(-0.66925731439)*x_ref[0]**o+(-0.0110127596435)*x_ref[0]+(0.944388255414)*x_ref[1]**o+(-0.613832072486)*x_ref[1]+(0.0977542120724)*x_ref[2]**o+(0.100776856689)*x_ref[2]
        ref[3,0,1,0]=(0.905138713064)*x_ref[0]**o+(0.673051338153)*x_ref[0]+(0.240614253653)*x_ref[1]**o+(-0.964132330033)*x_ref[1]+(-0.114961018838)*x_ref[2]**o+(-0.438756962988)*x_ref[2]
        ref[3,0,1,1]=(-0.192259182094)*x_ref[0]**o+(0.12315590119)*x_ref[0]+(-0.907744420226)*x_ref[1]**o+(-0.483981269809)*x_ref[1]+(-0.169699431472)*x_ref[2]**o+(-0.45231635406)*x_ref[2]
        ref[3,0,2,0]=(0.895387255885)*x_ref[0]**o+(0.23892997411)*x_ref[0]+(-0.396010446618)*x_ref[1]**o+(0.94853984413)*x_ref[1]+(-0.0165259637241)*x_ref[2]**o+(-0.312558539149)*x_ref[2]
        ref[3,0,2,1]=(0.710478713361)*x_ref[0]**o+(-0.109316102216)*x_ref[0]+(-0.042711568986)*x_ref[1]**o+(-0.448508085742)*x_ref[1]+(-0.0979193346011)*x_ref[2]**o+(0.93156672211)*x_ref[2]
        ref[3,1,0,0]=(-0.21605108221)*x_ref[0]**o+(0.410466481712)*x_ref[0]+(0.799288386783)*x_ref[1]**o+(0.622866017113)*x_ref[1]+(-0.236698938159)*x_ref[2]**o+(0.246625241928)*x_ref[2]
        ref[3,1,0,1]=(0.989568527762)*x_ref[0]**o+(0.0212906378151)*x_ref[0]+(0.311279947324)*x_ref[1]**o+(-0.231412902902)*x_ref[1]+(-0.444760786142)*x_ref[2]**o+(0.908171985421)*x_ref[2]
        ref[3,1,1,0]=(0.42605234231)*x_ref[0]**o+(-0.58629332944)*x_ref[0]+(0.322063640761)*x_ref[1]**o+(0.837771339825)*x_ref[1]+(-0.911083211394)*x_ref[2]**o+(-0.0699607581992)*x_ref[2]
        ref[3,1,1,1]=(0.75449445591)*x_ref[0]**o+(0.722970800307)*x_ref[0]+(-0.582970919465)*x_ref[1]**o+(-0.868818538212)*x_ref[1]+(0.940067636181)*x_ref[2]**o+(-0.659469743617)*x_ref[2]
        ref[3,1,2,0]=(-0.391224924663)*x_ref[0]**o+(-0.768173177737)*x_ref[0]+(-0.856876347997)*x_ref[1]**o+(-0.0451180499647)*x_ref[1]+(0.198590679246)*x_ref[2]**o+(0.187329095476)*x_ref[2]
        ref[3,1,2,1]=(0.483309063092)*x_ref[0]**o+(-0.0777241026893)*x_ref[0]+(0.0438427457286)*x_ref[1]**o+(-0.73195704867)*x_ref[1]+(-0.439660745725)*x_ref[2]**o+(0.279501672861)*x_ref[2]
        ref[3,2,0,0]=(0.799935089139)*x_ref[0]**o+(-0.306860809857)*x_ref[0]+(0.169104975039)*x_ref[1]**o+(-0.3480717525)*x_ref[1]+(0.215231688)*x_ref[2]**o+(-0.528775376945)*x_ref[2]
        ref[3,2,0,1]=(0.917815789587)*x_ref[0]**o+(0.0672910338102)*x_ref[0]+(-0.186550440011)*x_ref[1]**o+(0.922579714252)*x_ref[1]+(0.924162379857)*x_ref[2]**o+(-0.769320483618)*x_ref[2]
        ref[3,2,1,0]=(0.604324468058)*x_ref[0]**o+(-0.0109731419053)*x_ref[0]+(0.229447318325)*x_ref[1]**o+(-0.480693435232)*x_ref[1]+(-0.881779256999)*x_ref[2]**o+(0.288656394771)*x_ref[2]
        ref[3,2,1,1]=(0.934680723352)*x_ref[0]**o+(0.993838528376)*x_ref[0]+(0.957781588502)*x_ref[1]**o+(-0.859837057866)*x_ref[1]+(-0.702139964493)*x_ref[2]**o+(0.0538177124609)*x_ref[2]
        ref[3,2,2,0]=(-0.142809047814)*x_ref[0]**o+(-0.472594415134)*x_ref[0]+(0.0381782768377)*x_ref[1]**o+(0.727331491769)*x_ref[1]+(0.0910992897581)*x_ref[2]**o+(-0.845316063851)*x_ref[2]
        ref[3,2,2,1]=(-0.217483358272)*x_ref[0]**o+(0.991535092285)*x_ref[0]+(-0.399175896468)*x_ref[1]**o+(-0.665836512895)*x_ref[1]+(0.0983700911493)*x_ref[2]**o+(0.321811300155)*x_ref[2]
        ref[3,3,0,0]=(0.797049585464)*x_ref[0]**o+(-0.724467959299)*x_ref[0]+(0.19971589973)*x_ref[1]**o+(0.845928461864)*x_ref[1]+(-0.365158866359)*x_ref[2]**o+(-0.698274270132)*x_ref[2]
        ref[3,3,0,1]=(-0.638474704878)*x_ref[0]**o+(0.551866433352)*x_ref[0]+(0.192309452979)*x_ref[1]**o+(-0.703649325501)*x_ref[1]+(0.823910578138)*x_ref[2]**o+(-0.917446989634)*x_ref[2]
        ref[3,3,1,0]=(0.217577282169)*x_ref[0]**o+(0.521716803668)*x_ref[0]+(-0.116587783856)*x_ref[1]**o+(0.529672391417)*x_ref[1]+(-0.798918158284)*x_ref[2]**o+(-0.0996139511955)*x_ref[2]
        ref[3,3,1,1]=(0.498965049079)*x_ref[0]**o+(0.316489921186)*x_ref[0]+(-0.0988355471817)*x_ref[1]**o+(-0.0477508287681)*x_ref[1]+(-0.778914523993)*x_ref[2]**o+(0.189397812723)*x_ref[2]
        ref[3,3,2,0]=(0.203823698152)*x_ref[0]**o+(-0.340499414266)*x_ref[0]+(0.353374693517)*x_ref[1]**o+(-0.91035264494)*x_ref[1]+(-0.779204084198)*x_ref[2]**o+(-0.00276454165248)*x_ref[2]
        ref[3,3,2,1]=(0.295232007123)*x_ref[0]**o+(0.271766544502)*x_ref[0]+(0.431956430336)*x_ref[1]**o+(0.896481311536)*x_ref[1]+(-0.325014902487)*x_ref[2]**o+(0.70836170284)*x_ref[2]
        ref[3,4,0,0]=(-0.407535202468)*x_ref[0]**o+(-0.33055681091)*x_ref[0]+(0.129762143151)*x_ref[1]**o+(0.352087436228)*x_ref[1]+(-0.538770401565)*x_ref[2]**o+(-0.519071077045)*x_ref[2]
        ref[3,4,0,1]=(0.846011435499)*x_ref[0]**o+(0.172822133672)*x_ref[0]+(-0.740732553847)*x_ref[1]**o+(0.230062837231)*x_ref[1]+(-0.231810007237)*x_ref[2]**o+(-0.568366338923)*x_ref[2]
        ref[3,4,1,0]=(-0.198091119057)*x_ref[0]**o+(0.957989508794)*x_ref[0]+(0.861739071175)*x_ref[1]**o+(0.923450697883)*x_ref[1]+(-0.755262504937)*x_ref[2]**o+(0.102897579936)*x_ref[2]
        ref[3,4,1,1]=(0.198829545353)*x_ref[0]**o+(-0.757182006818)*x_ref[0]+(0.254092337808)*x_ref[1]**o+(-0.00815072620097)*x_ref[1]+(0.347421988308)*x_ref[2]**o+(-0.0215277616923)*x_ref[2]
        ref[3,4,2,0]=(0.775062112529)*x_ref[0]**o+(-0.77746792774)*x_ref[0]+(-0.666490294869)*x_ref[1]**o+(-0.679971069878)*x_ref[1]+(-0.972450779894)*x_ref[2]**o+(-0.705322282477)*x_ref[2]
        ref[3,4,2,1]=(-0.0994927587382)*x_ref[0]**o+(0.293071774786)*x_ref[0]+(0.370784399397)*x_ref[1]**o+(0.931411225946)*x_ref[1]+(-0.55789163294)*x_ref[2]**o+(0.0831890658592)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0969659653911)*x[0]**o+(-0.0778467995236)*x[0]+(-0.896520347956)*x[1]**o+(0.452579859858)*x[1]
        ref=(0.0969659653911)*x_ref[0]**o+(-0.0778467995236)*x_ref[0]+(-0.896520347956)*x_ref[1]**o+(0.452579859858)*x_ref[1]
      else:
        arg=(-0.62506591451)*x[0]**o+(0.970735975624)*x[0]+(-0.563055321362)*x[1]**o+(-0.281482915069)*x[1]+(-0.6461258967)*x[2]**o+(0.786610350609)*x[2]
        ref=(-0.62506591451)*x_ref[0]**o+(0.970735975624)*x_ref[0]+(-0.563055321362)*x_ref[1]**o+(-0.281482915069)*x_ref[1]+(-0.6461258967)*x_ref[2]**o+(0.786610350609)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.725240039321)*x[0]**o+(0.326302203613)*x[0]+(0.356072156898)*x[1]**o+(0.980066388307)*x[1]
        arg[1]=(0.171551389095)*x[0]**o+(0.379133470853)*x[0]+(0.120608077039)*x[1]**o+(-0.413701037132)*x[1]
        ref[0]=(0.725240039321)*x_ref[0]**o+(0.326302203613)*x_ref[0]+(0.356072156898)*x_ref[1]**o+(0.980066388307)*x_ref[1]
        ref[1]=(0.171551389095)*x_ref[0]**o+(0.379133470853)*x_ref[0]+(0.120608077039)*x_ref[1]**o+(-0.413701037132)*x_ref[1]
      else:
        arg[0]=(-0.656095745624)*x[0]**o+(-0.540430373724)*x[0]+(-0.109962817674)*x[1]**o+(-0.969692783268)*x[1]+(-0.377994841279)*x[2]**o+(-0.664035030664)*x[2]
        arg[1]=(0.768498465456)*x[0]**o+(0.647294499552)*x[0]+(0.824672861058)*x[1]**o+(-0.848088780836)*x[1]+(0.045558138597)*x[2]**o+(-0.209365900036)*x[2]
        ref[0]=(-0.656095745624)*x_ref[0]**o+(-0.540430373724)*x_ref[0]+(-0.109962817674)*x_ref[1]**o+(-0.969692783268)*x_ref[1]+(-0.377994841279)*x_ref[2]**o+(-0.664035030664)*x_ref[2]
        ref[1]=(0.768498465456)*x_ref[0]**o+(0.647294499552)*x_ref[0]+(0.824672861058)*x_ref[1]**o+(-0.848088780836)*x_ref[1]+(0.045558138597)*x_ref[2]**o+(-0.209365900036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.694866182164)*x[0]**o+(0.759641527396)*x[0]+(-0.696641139757)*x[1]**o+(0.617978953453)*x[1]
        arg[0,1]=(0.75058922213)*x[0]**o+(0.580832851854)*x[0]+(-0.237039031011)*x[1]**o+(0.53518450519)*x[1]
        arg[0,2]=(-0.191933019173)*x[0]**o+(-0.0638088686782)*x[0]+(0.871645336149)*x[1]**o+(0.256502442034)*x[1]
        arg[0,3]=(0.829189580591)*x[0]**o+(0.186064016504)*x[0]+(-0.204496850484)*x[1]**o+(-0.314535598074)*x[1]
        arg[0,4]=(0.170968249419)*x[0]**o+(-0.719959835105)*x[0]+(0.367000223059)*x[1]**o+(-0.210239222752)*x[1]
        arg[1,0]=(-0.424311493894)*x[0]**o+(-0.381097143579)*x[0]+(-0.54925292526)*x[1]**o+(-0.393349131169)*x[1]
        arg[1,1]=(0.148181508143)*x[0]**o+(-0.224530937841)*x[0]+(0.778748719265)*x[1]**o+(-0.522104185968)*x[1]
        arg[1,2]=(0.134819103337)*x[0]**o+(0.0388165865124)*x[0]+(-0.772318490115)*x[1]**o+(-0.928611227188)*x[1]
        arg[1,3]=(0.790807030518)*x[0]**o+(-0.849964960983)*x[0]+(-0.79407080953)*x[1]**o+(-0.626315155005)*x[1]
        arg[1,4]=(-0.998842104726)*x[0]**o+(-0.325938405818)*x[0]+(-0.81445420822)*x[1]**o+(0.868281547664)*x[1]
        arg[2,0]=(-0.975572739854)*x[0]**o+(-0.0630975707316)*x[0]+(0.0688031679656)*x[1]**o+(0.557152464527)*x[1]
        arg[2,1]=(0.389814261767)*x[0]**o+(0.523891881188)*x[0]+(-0.558388242102)*x[1]**o+(-0.45720905598)*x[1]
        arg[2,2]=(0.763327517145)*x[0]**o+(0.637954909904)*x[0]+(0.407602154355)*x[1]**o+(0.707118271445)*x[1]
        arg[2,3]=(-0.823371482965)*x[0]**o+(0.918685041464)*x[0]+(0.991991462022)*x[1]**o+(0.489892661208)*x[1]
        arg[2,4]=(-0.661567455889)*x[0]**o+(-0.00566795402284)*x[0]+(0.86705493299)*x[1]**o+(0.00158321121677)*x[1]
        arg[3,0]=(0.365456232671)*x[0]**o+(-0.587373381962)*x[0]+(0.751291631507)*x[1]**o+(-0.0605248375222)*x[1]
        arg[3,1]=(0.546630294768)*x[0]**o+(-0.903068297271)*x[0]+(0.486313468693)*x[1]**o+(0.355322343068)*x[1]
        arg[3,2]=(0.233755056849)*x[0]**o+(-0.350375425427)*x[0]+(0.0115435986629)*x[1]**o+(-0.713895312694)*x[1]
        arg[3,3]=(0.385411302842)*x[0]**o+(0.0733743700003)*x[0]+(0.439219708301)*x[1]**o+(0.129391539278)*x[1]
        arg[3,4]=(-0.194268604537)*x[0]**o+(0.175829932738)*x[0]+(0.866342013931)*x[1]**o+(-0.213891435501)*x[1]
        ref[0,0]=(0.694866182164)*x_ref[0]**o+(0.759641527396)*x_ref[0]+(-0.696641139757)*x_ref[1]**o+(0.617978953453)*x_ref[1]
        ref[0,1]=(0.75058922213)*x_ref[0]**o+(0.580832851854)*x_ref[0]+(-0.237039031011)*x_ref[1]**o+(0.53518450519)*x_ref[1]
        ref[0,2]=(-0.191933019173)*x_ref[0]**o+(-0.0638088686782)*x_ref[0]+(0.871645336149)*x_ref[1]**o+(0.256502442034)*x_ref[1]
        ref[0,3]=(0.829189580591)*x_ref[0]**o+(0.186064016504)*x_ref[0]+(-0.204496850484)*x_ref[1]**o+(-0.314535598074)*x_ref[1]
        ref[0,4]=(0.170968249419)*x_ref[0]**o+(-0.719959835105)*x_ref[0]+(0.367000223059)*x_ref[1]**o+(-0.210239222752)*x_ref[1]
        ref[1,0]=(-0.424311493894)*x_ref[0]**o+(-0.381097143579)*x_ref[0]+(-0.54925292526)*x_ref[1]**o+(-0.393349131169)*x_ref[1]
        ref[1,1]=(0.148181508143)*x_ref[0]**o+(-0.224530937841)*x_ref[0]+(0.778748719265)*x_ref[1]**o+(-0.522104185968)*x_ref[1]
        ref[1,2]=(0.134819103337)*x_ref[0]**o+(0.0388165865124)*x_ref[0]+(-0.772318490115)*x_ref[1]**o+(-0.928611227188)*x_ref[1]
        ref[1,3]=(0.790807030518)*x_ref[0]**o+(-0.849964960983)*x_ref[0]+(-0.79407080953)*x_ref[1]**o+(-0.626315155005)*x_ref[1]
        ref[1,4]=(-0.998842104726)*x_ref[0]**o+(-0.325938405818)*x_ref[0]+(-0.81445420822)*x_ref[1]**o+(0.868281547664)*x_ref[1]
        ref[2,0]=(-0.975572739854)*x_ref[0]**o+(-0.0630975707316)*x_ref[0]+(0.0688031679656)*x_ref[1]**o+(0.557152464527)*x_ref[1]
        ref[2,1]=(0.389814261767)*x_ref[0]**o+(0.523891881188)*x_ref[0]+(-0.558388242102)*x_ref[1]**o+(-0.45720905598)*x_ref[1]
        ref[2,2]=(0.763327517145)*x_ref[0]**o+(0.637954909904)*x_ref[0]+(0.407602154355)*x_ref[1]**o+(0.707118271445)*x_ref[1]
        ref[2,3]=(-0.823371482965)*x_ref[0]**o+(0.918685041464)*x_ref[0]+(0.991991462022)*x_ref[1]**o+(0.489892661208)*x_ref[1]
        ref[2,4]=(-0.661567455889)*x_ref[0]**o+(-0.00566795402284)*x_ref[0]+(0.86705493299)*x_ref[1]**o+(0.00158321121677)*x_ref[1]
        ref[3,0]=(0.365456232671)*x_ref[0]**o+(-0.587373381962)*x_ref[0]+(0.751291631507)*x_ref[1]**o+(-0.0605248375222)*x_ref[1]
        ref[3,1]=(0.546630294768)*x_ref[0]**o+(-0.903068297271)*x_ref[0]+(0.486313468693)*x_ref[1]**o+(0.355322343068)*x_ref[1]
        ref[3,2]=(0.233755056849)*x_ref[0]**o+(-0.350375425427)*x_ref[0]+(0.0115435986629)*x_ref[1]**o+(-0.713895312694)*x_ref[1]
        ref[3,3]=(0.385411302842)*x_ref[0]**o+(0.0733743700003)*x_ref[0]+(0.439219708301)*x_ref[1]**o+(0.129391539278)*x_ref[1]
        ref[3,4]=(-0.194268604537)*x_ref[0]**o+(0.175829932738)*x_ref[0]+(0.866342013931)*x_ref[1]**o+(-0.213891435501)*x_ref[1]
      else:
        arg[0,0]=(-0.479138946804)*x[0]**o+(0.240408443007)*x[0]+(0.309155796205)*x[1]**o+(-0.124360991957)*x[1]+(0.00301047542037)*x[2]**o+(-0.039439802799)*x[2]
        arg[0,1]=(0.41979402771)*x[0]**o+(0.798738902847)*x[0]+(0.619960746922)*x[1]**o+(-0.551662130664)*x[1]+(0.914293623695)*x[2]**o+(0.6859768812)*x[2]
        arg[0,2]=(0.744593492696)*x[0]**o+(-0.31743031745)*x[0]+(0.389545123971)*x[1]**o+(0.592024987288)*x[1]+(0.63456095022)*x[2]**o+(-0.746571796061)*x[2]
        arg[0,3]=(-0.632549503649)*x[0]**o+(-0.687745599598)*x[0]+(0.794176042123)*x[1]**o+(-0.971466375953)*x[1]+(-0.086939632062)*x[2]**o+(0.373736042024)*x[2]
        arg[0,4]=(-0.966879538217)*x[0]**o+(-0.609291970974)*x[0]+(0.000409196978095)*x[1]**o+(0.623934484362)*x[1]+(-0.1694749999)*x[2]**o+(0.131778917979)*x[2]
        arg[1,0]=(0.763452444611)*x[0]**o+(0.805147582411)*x[0]+(0.999704124146)*x[1]**o+(0.0635253753138)*x[1]+(-0.717889410548)*x[2]**o+(-0.840765968423)*x[2]
        arg[1,1]=(-0.26910479702)*x[0]**o+(0.594093239771)*x[0]+(-0.536150372834)*x[1]**o+(-0.768506263316)*x[1]+(-0.924510063594)*x[2]**o+(-0.950568839784)*x[2]
        arg[1,2]=(-0.452317334239)*x[0]**o+(0.186343827048)*x[0]+(0.921015823436)*x[1]**o+(-0.656809197675)*x[1]+(-0.841898648461)*x[2]**o+(0.267827136605)*x[2]
        arg[1,3]=(-0.880873336233)*x[0]**o+(-0.915350352959)*x[0]+(0.92197459758)*x[1]**o+(0.20861139921)*x[1]+(0.776743964515)*x[2]**o+(0.858569288507)*x[2]
        arg[1,4]=(0.422773256864)*x[0]**o+(0.397743860671)*x[0]+(0.726470981309)*x[1]**o+(-0.233534612019)*x[1]+(-0.350345927201)*x[2]**o+(-0.192294374344)*x[2]
        arg[2,0]=(0.0706785590906)*x[0]**o+(-0.102098736751)*x[0]+(-0.858298802695)*x[1]**o+(-0.1390708698)*x[1]+(0.221129390442)*x[2]**o+(0.770148745174)*x[2]
        arg[2,1]=(0.619559141603)*x[0]**o+(-0.969165124767)*x[0]+(0.36662839455)*x[1]**o+(0.505630485999)*x[1]+(0.170716012487)*x[2]**o+(0.355457960368)*x[2]
        arg[2,2]=(0.254098998564)*x[0]**o+(0.725662782597)*x[0]+(0.274108937869)*x[1]**o+(0.682336753679)*x[1]+(-0.858192561327)*x[2]**o+(-0.538785359078)*x[2]
        arg[2,3]=(0.25513729674)*x[0]**o+(-0.594968324608)*x[0]+(-0.306612686559)*x[1]**o+(-0.953891015684)*x[1]+(-0.00631658753287)*x[2]**o+(-0.889694464702)*x[2]
        arg[2,4]=(0.654463492789)*x[0]**o+(-0.42220702047)*x[0]+(-0.690125229151)*x[1]**o+(0.318923837754)*x[1]+(-0.691341299676)*x[2]**o+(0.824631778823)*x[2]
        arg[3,0]=(-0.576563725107)*x[0]**o+(-0.0354219245402)*x[0]+(0.650854296808)*x[1]**o+(0.930072736771)*x[1]+(0.8383657117)*x[2]**o+(0.191293272146)*x[2]
        arg[3,1]=(0.104382485594)*x[0]**o+(-0.718025390561)*x[0]+(-0.822188446684)*x[1]**o+(-0.0772502664782)*x[1]+(-0.513331680922)*x[2]**o+(0.164340503698)*x[2]
        arg[3,2]=(0.685178273481)*x[0]**o+(-0.241726920352)*x[0]+(-0.891512838953)*x[1]**o+(-0.428833037662)*x[1]+(-0.910220899771)*x[2]**o+(-0.0109246344661)*x[2]
        arg[3,3]=(-0.294769406995)*x[0]**o+(0.643502505933)*x[0]+(0.309444282121)*x[1]**o+(-0.935318923262)*x[1]+(0.242426961867)*x[2]**o+(-0.417268637067)*x[2]
        arg[3,4]=(0.0382306334841)*x[0]**o+(-0.564858278904)*x[0]+(-0.594566652327)*x[1]**o+(0.911894434818)*x[1]+(-0.314178793024)*x[2]**o+(0.302730333267)*x[2]
        ref[0,0]=(-0.479138946804)*x_ref[0]**o+(0.240408443007)*x_ref[0]+(0.309155796205)*x_ref[1]**o+(-0.124360991957)*x_ref[1]+(0.00301047542037)*x_ref[2]**o+(-0.039439802799)*x_ref[2]
        ref[0,1]=(0.41979402771)*x_ref[0]**o+(0.798738902847)*x_ref[0]+(0.619960746922)*x_ref[1]**o+(-0.551662130664)*x_ref[1]+(0.914293623695)*x_ref[2]**o+(0.6859768812)*x_ref[2]
        ref[0,2]=(0.744593492696)*x_ref[0]**o+(-0.31743031745)*x_ref[0]+(0.389545123971)*x_ref[1]**o+(0.592024987288)*x_ref[1]+(0.63456095022)*x_ref[2]**o+(-0.746571796061)*x_ref[2]
        ref[0,3]=(-0.632549503649)*x_ref[0]**o+(-0.687745599598)*x_ref[0]+(0.794176042123)*x_ref[1]**o+(-0.971466375953)*x_ref[1]+(-0.086939632062)*x_ref[2]**o+(0.373736042024)*x_ref[2]
        ref[0,4]=(-0.966879538217)*x_ref[0]**o+(-0.609291970974)*x_ref[0]+(0.000409196978095)*x_ref[1]**o+(0.623934484362)*x_ref[1]+(-0.1694749999)*x_ref[2]**o+(0.131778917979)*x_ref[2]
        ref[1,0]=(0.763452444611)*x_ref[0]**o+(0.805147582411)*x_ref[0]+(0.999704124146)*x_ref[1]**o+(0.0635253753138)*x_ref[1]+(-0.717889410548)*x_ref[2]**o+(-0.840765968423)*x_ref[2]
        ref[1,1]=(-0.26910479702)*x_ref[0]**o+(0.594093239771)*x_ref[0]+(-0.536150372834)*x_ref[1]**o+(-0.768506263316)*x_ref[1]+(-0.924510063594)*x_ref[2]**o+(-0.950568839784)*x_ref[2]
        ref[1,2]=(-0.452317334239)*x_ref[0]**o+(0.186343827048)*x_ref[0]+(0.921015823436)*x_ref[1]**o+(-0.656809197675)*x_ref[1]+(-0.841898648461)*x_ref[2]**o+(0.267827136605)*x_ref[2]
        ref[1,3]=(-0.880873336233)*x_ref[0]**o+(-0.915350352959)*x_ref[0]+(0.92197459758)*x_ref[1]**o+(0.20861139921)*x_ref[1]+(0.776743964515)*x_ref[2]**o+(0.858569288507)*x_ref[2]
        ref[1,4]=(0.422773256864)*x_ref[0]**o+(0.397743860671)*x_ref[0]+(0.726470981309)*x_ref[1]**o+(-0.233534612019)*x_ref[1]+(-0.350345927201)*x_ref[2]**o+(-0.192294374344)*x_ref[2]
        ref[2,0]=(0.0706785590906)*x_ref[0]**o+(-0.102098736751)*x_ref[0]+(-0.858298802695)*x_ref[1]**o+(-0.1390708698)*x_ref[1]+(0.221129390442)*x_ref[2]**o+(0.770148745174)*x_ref[2]
        ref[2,1]=(0.619559141603)*x_ref[0]**o+(-0.969165124767)*x_ref[0]+(0.36662839455)*x_ref[1]**o+(0.505630485999)*x_ref[1]+(0.170716012487)*x_ref[2]**o+(0.355457960368)*x_ref[2]
        ref[2,2]=(0.254098998564)*x_ref[0]**o+(0.725662782597)*x_ref[0]+(0.274108937869)*x_ref[1]**o+(0.682336753679)*x_ref[1]+(-0.858192561327)*x_ref[2]**o+(-0.538785359078)*x_ref[2]
        ref[2,3]=(0.25513729674)*x_ref[0]**o+(-0.594968324608)*x_ref[0]+(-0.306612686559)*x_ref[1]**o+(-0.953891015684)*x_ref[1]+(-0.00631658753287)*x_ref[2]**o+(-0.889694464702)*x_ref[2]
        ref[2,4]=(0.654463492789)*x_ref[0]**o+(-0.42220702047)*x_ref[0]+(-0.690125229151)*x_ref[1]**o+(0.318923837754)*x_ref[1]+(-0.691341299676)*x_ref[2]**o+(0.824631778823)*x_ref[2]
        ref[3,0]=(-0.576563725107)*x_ref[0]**o+(-0.0354219245402)*x_ref[0]+(0.650854296808)*x_ref[1]**o+(0.930072736771)*x_ref[1]+(0.8383657117)*x_ref[2]**o+(0.191293272146)*x_ref[2]
        ref[3,1]=(0.104382485594)*x_ref[0]**o+(-0.718025390561)*x_ref[0]+(-0.822188446684)*x_ref[1]**o+(-0.0772502664782)*x_ref[1]+(-0.513331680922)*x_ref[2]**o+(0.164340503698)*x_ref[2]
        ref[3,2]=(0.685178273481)*x_ref[0]**o+(-0.241726920352)*x_ref[0]+(-0.891512838953)*x_ref[1]**o+(-0.428833037662)*x_ref[1]+(-0.910220899771)*x_ref[2]**o+(-0.0109246344661)*x_ref[2]
        ref[3,3]=(-0.294769406995)*x_ref[0]**o+(0.643502505933)*x_ref[0]+(0.309444282121)*x_ref[1]**o+(-0.935318923262)*x_ref[1]+(0.242426961867)*x_ref[2]**o+(-0.417268637067)*x_ref[2]
        ref[3,4]=(0.0382306334841)*x_ref[0]**o+(-0.564858278904)*x_ref[0]+(-0.594566652327)*x_ref[1]**o+(0.911894434818)*x_ref[1]+(-0.314178793024)*x_ref[2]**o+(0.302730333267)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.298635657968)*x[0]**o+(0.917751825722)*x[0]+(-0.000157823722007)*x[1]**o+(-0.202721972906)*x[1]
        arg[0,0,1]=(0.164880328673)*x[0]**o+(-0.310233508297)*x[0]+(-0.976270095138)*x[1]**o+(-0.86689594754)*x[1]
        arg[0,1,0]=(-0.130835262631)*x[0]**o+(-0.485555484157)*x[0]+(0.055655815912)*x[1]**o+(-0.488201458657)*x[1]
        arg[0,1,1]=(-0.339982923518)*x[0]**o+(0.435464515598)*x[0]+(-0.986955506076)*x[1]**o+(0.590966000031)*x[1]
        arg[1,0,0]=(0.150280127463)*x[0]**o+(0.827183876844)*x[0]+(0.468401701763)*x[1]**o+(-0.879871376091)*x[1]
        arg[1,0,1]=(0.48434831025)*x[0]**o+(-0.623958069024)*x[0]+(-0.781911271849)*x[1]**o+(0.34398679645)*x[1]
        arg[1,1,0]=(0.124273770782)*x[0]**o+(-0.683096046651)*x[0]+(0.769636378285)*x[1]**o+(-0.471970832437)*x[1]
        arg[1,1,1]=(0.769777912195)*x[0]**o+(-0.740802020319)*x[0]+(0.659109761568)*x[1]**o+(-0.495444655614)*x[1]
        arg[2,0,0]=(0.270063777356)*x[0]**o+(0.558849439723)*x[0]+(0.128946408169)*x[1]**o+(-0.962495994445)*x[1]
        arg[2,0,1]=(-0.0403460883009)*x[0]**o+(0.084983334662)*x[0]+(0.297573700577)*x[1]**o+(0.911774703678)*x[1]
        arg[2,1,0]=(0.821680378088)*x[0]**o+(0.10823172035)*x[0]+(-0.150213295203)*x[1]**o+(0.310471151961)*x[1]
        arg[2,1,1]=(-0.407033195691)*x[0]**o+(0.716262964777)*x[0]+(-0.262475432397)*x[1]**o+(-0.234158898682)*x[1]
        arg[3,0,0]=(-0.109193787139)*x[0]**o+(0.403892847875)*x[0]+(0.794082322801)*x[1]**o+(0.669070572253)*x[1]
        arg[3,0,1]=(0.969770217228)*x[0]**o+(0.735280190411)*x[0]+(-0.339891263613)*x[1]**o+(0.859989056027)*x[1]
        arg[3,1,0]=(-0.906193851712)*x[0]**o+(-0.330178132059)*x[0]+(-0.48035114847)*x[1]**o+(-0.48050067188)*x[1]
        arg[3,1,1]=(0.922778761752)*x[0]**o+(0.899145572602)*x[0]+(-0.991922151176)*x[1]**o+(0.615744852656)*x[1]
        arg[4,0,0]=(0.457092295352)*x[0]**o+(-0.120904430848)*x[0]+(0.575186782188)*x[1]**o+(-0.685522788659)*x[1]
        arg[4,0,1]=(0.254934861907)*x[0]**o+(-0.208850206761)*x[0]+(-0.0933987467304)*x[1]**o+(0.687715419212)*x[1]
        arg[4,1,0]=(0.483890412252)*x[0]**o+(-0.192055355216)*x[0]+(0.0891595367214)*x[1]**o+(-0.835900121805)*x[1]
        arg[4,1,1]=(-0.281109655935)*x[0]**o+(0.279867178407)*x[0]+(0.843094337157)*x[1]**o+(-0.457583293166)*x[1]
        arg[5,0,0]=(0.595078633674)*x[0]**o+(-0.994652086537)*x[0]+(0.111354924733)*x[1]**o+(0.715580298167)*x[1]
        arg[5,0,1]=(0.823409068766)*x[0]**o+(0.41047089933)*x[0]+(-0.610763277348)*x[1]**o+(0.0991108428868)*x[1]
        arg[5,1,0]=(-0.0291539407811)*x[0]**o+(0.371799932333)*x[0]+(0.631702366805)*x[1]**o+(-0.858415923309)*x[1]
        arg[5,1,1]=(-0.636146064991)*x[0]**o+(0.407248062196)*x[0]+(-0.507069143933)*x[1]**o+(-0.0509951705166)*x[1]
        ref[0,0,0]=(0.298635657968)*x_ref[0]**o+(0.917751825722)*x_ref[0]+(-0.000157823722007)*x_ref[1]**o+(-0.202721972906)*x_ref[1]
        ref[0,0,1]=(0.164880328673)*x_ref[0]**o+(-0.310233508297)*x_ref[0]+(-0.976270095138)*x_ref[1]**o+(-0.86689594754)*x_ref[1]
        ref[0,1,0]=(-0.130835262631)*x_ref[0]**o+(-0.485555484157)*x_ref[0]+(0.055655815912)*x_ref[1]**o+(-0.488201458657)*x_ref[1]
        ref[0,1,1]=(-0.339982923518)*x_ref[0]**o+(0.435464515598)*x_ref[0]+(-0.986955506076)*x_ref[1]**o+(0.590966000031)*x_ref[1]
        ref[1,0,0]=(0.150280127463)*x_ref[0]**o+(0.827183876844)*x_ref[0]+(0.468401701763)*x_ref[1]**o+(-0.879871376091)*x_ref[1]
        ref[1,0,1]=(0.48434831025)*x_ref[0]**o+(-0.623958069024)*x_ref[0]+(-0.781911271849)*x_ref[1]**o+(0.34398679645)*x_ref[1]
        ref[1,1,0]=(0.124273770782)*x_ref[0]**o+(-0.683096046651)*x_ref[0]+(0.769636378285)*x_ref[1]**o+(-0.471970832437)*x_ref[1]
        ref[1,1,1]=(0.769777912195)*x_ref[0]**o+(-0.740802020319)*x_ref[0]+(0.659109761568)*x_ref[1]**o+(-0.495444655614)*x_ref[1]
        ref[2,0,0]=(0.270063777356)*x_ref[0]**o+(0.558849439723)*x_ref[0]+(0.128946408169)*x_ref[1]**o+(-0.962495994445)*x_ref[1]
        ref[2,0,1]=(-0.0403460883009)*x_ref[0]**o+(0.084983334662)*x_ref[0]+(0.297573700577)*x_ref[1]**o+(0.911774703678)*x_ref[1]
        ref[2,1,0]=(0.821680378088)*x_ref[0]**o+(0.10823172035)*x_ref[0]+(-0.150213295203)*x_ref[1]**o+(0.310471151961)*x_ref[1]
        ref[2,1,1]=(-0.407033195691)*x_ref[0]**o+(0.716262964777)*x_ref[0]+(-0.262475432397)*x_ref[1]**o+(-0.234158898682)*x_ref[1]
        ref[3,0,0]=(-0.109193787139)*x_ref[0]**o+(0.403892847875)*x_ref[0]+(0.794082322801)*x_ref[1]**o+(0.669070572253)*x_ref[1]
        ref[3,0,1]=(0.969770217228)*x_ref[0]**o+(0.735280190411)*x_ref[0]+(-0.339891263613)*x_ref[1]**o+(0.859989056027)*x_ref[1]
        ref[3,1,0]=(-0.906193851712)*x_ref[0]**o+(-0.330178132059)*x_ref[0]+(-0.48035114847)*x_ref[1]**o+(-0.48050067188)*x_ref[1]
        ref[3,1,1]=(0.922778761752)*x_ref[0]**o+(0.899145572602)*x_ref[0]+(-0.991922151176)*x_ref[1]**o+(0.615744852656)*x_ref[1]
        ref[4,0,0]=(0.457092295352)*x_ref[0]**o+(-0.120904430848)*x_ref[0]+(0.575186782188)*x_ref[1]**o+(-0.685522788659)*x_ref[1]
        ref[4,0,1]=(0.254934861907)*x_ref[0]**o+(-0.208850206761)*x_ref[0]+(-0.0933987467304)*x_ref[1]**o+(0.687715419212)*x_ref[1]
        ref[4,1,0]=(0.483890412252)*x_ref[0]**o+(-0.192055355216)*x_ref[0]+(0.0891595367214)*x_ref[1]**o+(-0.835900121805)*x_ref[1]
        ref[4,1,1]=(-0.281109655935)*x_ref[0]**o+(0.279867178407)*x_ref[0]+(0.843094337157)*x_ref[1]**o+(-0.457583293166)*x_ref[1]
        ref[5,0,0]=(0.595078633674)*x_ref[0]**o+(-0.994652086537)*x_ref[0]+(0.111354924733)*x_ref[1]**o+(0.715580298167)*x_ref[1]
        ref[5,0,1]=(0.823409068766)*x_ref[0]**o+(0.41047089933)*x_ref[0]+(-0.610763277348)*x_ref[1]**o+(0.0991108428868)*x_ref[1]
        ref[5,1,0]=(-0.0291539407811)*x_ref[0]**o+(0.371799932333)*x_ref[0]+(0.631702366805)*x_ref[1]**o+(-0.858415923309)*x_ref[1]
        ref[5,1,1]=(-0.636146064991)*x_ref[0]**o+(0.407248062196)*x_ref[0]+(-0.507069143933)*x_ref[1]**o+(-0.0509951705166)*x_ref[1]
      else:
        arg[0,0,0]=(-0.662043178282)*x[0]**o+(0.600130780119)*x[0]+(-0.52233323628)*x[1]**o+(-0.463302513654)*x[1]+(0.714996279403)*x[2]**o+(0.113219137203)*x[2]
        arg[0,0,1]=(0.94317347224)*x[0]**o+(0.987584533548)*x[0]+(0.116577593127)*x[1]**o+(-0.515267022711)*x[1]+(0.856578953903)*x[2]**o+(0.70372338917)*x[2]
        arg[0,1,0]=(-0.405265998966)*x[0]**o+(-0.656404120416)*x[0]+(0.270655604701)*x[1]**o+(0.0638534875176)*x[1]+(0.816834973788)*x[2]**o+(0.918436000225)*x[2]
        arg[0,1,1]=(-0.37299950332)*x[0]**o+(-0.0367737761004)*x[0]+(0.176531372245)*x[1]**o+(0.849773143988)*x[1]+(-0.572911529869)*x[2]**o+(-0.140163625628)*x[2]
        arg[1,0,0]=(0.263616091516)*x[0]**o+(-0.65550051998)*x[0]+(-0.118310211881)*x[1]**o+(-0.695263983587)*x[1]+(-0.49011649182)*x[2]**o+(-0.553284002669)*x[2]
        arg[1,0,1]=(0.859640205193)*x[0]**o+(-0.569681205214)*x[0]+(0.459979003108)*x[1]**o+(0.672561515814)*x[1]+(-0.854289983441)*x[2]**o+(-0.445344129294)*x[2]
        arg[1,1,0]=(-0.904355809101)*x[0]**o+(0.540513356972)*x[0]+(-0.377422260181)*x[1]**o+(0.937285151058)*x[1]+(0.148943597435)*x[2]**o+(0.992256197076)*x[2]
        arg[1,1,1]=(-0.868336238952)*x[0]**o+(-0.713337888363)*x[0]+(0.800427728006)*x[1]**o+(-0.83502690568)*x[1]+(0.365707922084)*x[2]**o+(0.218740577318)*x[2]
        arg[2,0,0]=(-0.0382672594668)*x[0]**o+(0.463831755073)*x[0]+(-0.104957678792)*x[1]**o+(-0.561738607653)*x[1]+(-0.660128314406)*x[2]**o+(0.594649133189)*x[2]
        arg[2,0,1]=(-0.399297921751)*x[0]**o+(0.550876554224)*x[0]+(0.150599666563)*x[1]**o+(-0.748400310268)*x[1]+(0.658318134651)*x[2]**o+(0.49075091442)*x[2]
        arg[2,1,0]=(-0.393769876286)*x[0]**o+(0.328125610064)*x[0]+(-0.703385111404)*x[1]**o+(0.0418077854723)*x[1]+(0.918774663106)*x[2]**o+(-0.119085951508)*x[2]
        arg[2,1,1]=(-0.549250924758)*x[0]**o+(0.169148558793)*x[0]+(0.664162931811)*x[1]**o+(0.207136816188)*x[1]+(-0.17207817969)*x[2]**o+(0.34258020774)*x[2]
        arg[3,0,0]=(0.446467472517)*x[0]**o+(-0.358128380843)*x[0]+(0.146916048359)*x[1]**o+(-0.606692124285)*x[1]+(0.828655704727)*x[2]**o+(0.067139299396)*x[2]
        arg[3,0,1]=(-0.357052242351)*x[0]**o+(-0.528255164843)*x[0]+(0.140204673142)*x[1]**o+(0.674282235672)*x[1]+(-0.469118454039)*x[2]**o+(0.388103121466)*x[2]
        arg[3,1,0]=(-0.359629791333)*x[0]**o+(-0.625108929158)*x[0]+(0.855034791748)*x[1]**o+(-0.343451063341)*x[1]+(-0.681557972105)*x[2]**o+(0.576310870238)*x[2]
        arg[3,1,1]=(0.00546530278132)*x[0]**o+(0.359221089344)*x[0]+(-0.57416247885)*x[1]**o+(-0.521306138323)*x[1]+(-0.618281954095)*x[2]**o+(-0.862162207352)*x[2]
        arg[4,0,0]=(-0.196356423488)*x[0]**o+(-0.871482508533)*x[0]+(-0.367390886081)*x[1]**o+(0.580310108591)*x[1]+(0.857970671482)*x[2]**o+(0.123685673359)*x[2]
        arg[4,0,1]=(0.715104674022)*x[0]**o+(-0.00444438628932)*x[0]+(0.538883210054)*x[1]**o+(0.179340705463)*x[1]+(-0.605249436909)*x[2]**o+(-0.744253826562)*x[2]
        arg[4,1,0]=(-0.315453893647)*x[0]**o+(-0.560946546082)*x[0]+(-0.674448928458)*x[1]**o+(-0.510243444449)*x[1]+(0.592615876779)*x[2]**o+(0.218024801263)*x[2]
        arg[4,1,1]=(-0.46902473326)*x[0]**o+(-0.573926691452)*x[0]+(0.827898114814)*x[1]**o+(-0.644208082857)*x[1]+(-0.956169254426)*x[2]**o+(-0.297101737561)*x[2]
        arg[5,0,0]=(0.45818164662)*x[0]**o+(0.219667021464)*x[0]+(-0.453125753418)*x[1]**o+(-0.201214929832)*x[1]+(0.135363067088)*x[2]**o+(0.850636247323)*x[2]
        arg[5,0,1]=(0.296120304926)*x[0]**o+(-0.139971296351)*x[0]+(0.448115780401)*x[1]**o+(-0.300602343399)*x[1]+(-0.165568753274)*x[2]**o+(0.984788641639)*x[2]
        arg[5,1,0]=(0.321875942152)*x[0]**o+(-0.403122417353)*x[0]+(-0.148664639927)*x[1]**o+(0.644143662303)*x[1]+(-0.718311175542)*x[2]**o+(-0.850248996913)*x[2]
        arg[5,1,1]=(0.44869757406)*x[0]**o+(0.0292094503086)*x[0]+(-0.159644215279)*x[1]**o+(0.695976060359)*x[1]+(-0.221444458713)*x[2]**o+(-0.42974111206)*x[2]
        ref[0,0,0]=(-0.662043178282)*x_ref[0]**o+(0.600130780119)*x_ref[0]+(-0.52233323628)*x_ref[1]**o+(-0.463302513654)*x_ref[1]+(0.714996279403)*x_ref[2]**o+(0.113219137203)*x_ref[2]
        ref[0,0,1]=(0.94317347224)*x_ref[0]**o+(0.987584533548)*x_ref[0]+(0.116577593127)*x_ref[1]**o+(-0.515267022711)*x_ref[1]+(0.856578953903)*x_ref[2]**o+(0.70372338917)*x_ref[2]
        ref[0,1,0]=(-0.405265998966)*x_ref[0]**o+(-0.656404120416)*x_ref[0]+(0.270655604701)*x_ref[1]**o+(0.0638534875176)*x_ref[1]+(0.816834973788)*x_ref[2]**o+(0.918436000225)*x_ref[2]
        ref[0,1,1]=(-0.37299950332)*x_ref[0]**o+(-0.0367737761004)*x_ref[0]+(0.176531372245)*x_ref[1]**o+(0.849773143988)*x_ref[1]+(-0.572911529869)*x_ref[2]**o+(-0.140163625628)*x_ref[2]
        ref[1,0,0]=(0.263616091516)*x_ref[0]**o+(-0.65550051998)*x_ref[0]+(-0.118310211881)*x_ref[1]**o+(-0.695263983587)*x_ref[1]+(-0.49011649182)*x_ref[2]**o+(-0.553284002669)*x_ref[2]
        ref[1,0,1]=(0.859640205193)*x_ref[0]**o+(-0.569681205214)*x_ref[0]+(0.459979003108)*x_ref[1]**o+(0.672561515814)*x_ref[1]+(-0.854289983441)*x_ref[2]**o+(-0.445344129294)*x_ref[2]
        ref[1,1,0]=(-0.904355809101)*x_ref[0]**o+(0.540513356972)*x_ref[0]+(-0.377422260181)*x_ref[1]**o+(0.937285151058)*x_ref[1]+(0.148943597435)*x_ref[2]**o+(0.992256197076)*x_ref[2]
        ref[1,1,1]=(-0.868336238952)*x_ref[0]**o+(-0.713337888363)*x_ref[0]+(0.800427728006)*x_ref[1]**o+(-0.83502690568)*x_ref[1]+(0.365707922084)*x_ref[2]**o+(0.218740577318)*x_ref[2]
        ref[2,0,0]=(-0.0382672594668)*x_ref[0]**o+(0.463831755073)*x_ref[0]+(-0.104957678792)*x_ref[1]**o+(-0.561738607653)*x_ref[1]+(-0.660128314406)*x_ref[2]**o+(0.594649133189)*x_ref[2]
        ref[2,0,1]=(-0.399297921751)*x_ref[0]**o+(0.550876554224)*x_ref[0]+(0.150599666563)*x_ref[1]**o+(-0.748400310268)*x_ref[1]+(0.658318134651)*x_ref[2]**o+(0.49075091442)*x_ref[2]
        ref[2,1,0]=(-0.393769876286)*x_ref[0]**o+(0.328125610064)*x_ref[0]+(-0.703385111404)*x_ref[1]**o+(0.0418077854723)*x_ref[1]+(0.918774663106)*x_ref[2]**o+(-0.119085951508)*x_ref[2]
        ref[2,1,1]=(-0.549250924758)*x_ref[0]**o+(0.169148558793)*x_ref[0]+(0.664162931811)*x_ref[1]**o+(0.207136816188)*x_ref[1]+(-0.17207817969)*x_ref[2]**o+(0.34258020774)*x_ref[2]
        ref[3,0,0]=(0.446467472517)*x_ref[0]**o+(-0.358128380843)*x_ref[0]+(0.146916048359)*x_ref[1]**o+(-0.606692124285)*x_ref[1]+(0.828655704727)*x_ref[2]**o+(0.067139299396)*x_ref[2]
        ref[3,0,1]=(-0.357052242351)*x_ref[0]**o+(-0.528255164843)*x_ref[0]+(0.140204673142)*x_ref[1]**o+(0.674282235672)*x_ref[1]+(-0.469118454039)*x_ref[2]**o+(0.388103121466)*x_ref[2]
        ref[3,1,0]=(-0.359629791333)*x_ref[0]**o+(-0.625108929158)*x_ref[0]+(0.855034791748)*x_ref[1]**o+(-0.343451063341)*x_ref[1]+(-0.681557972105)*x_ref[2]**o+(0.576310870238)*x_ref[2]
        ref[3,1,1]=(0.00546530278132)*x_ref[0]**o+(0.359221089344)*x_ref[0]+(-0.57416247885)*x_ref[1]**o+(-0.521306138323)*x_ref[1]+(-0.618281954095)*x_ref[2]**o+(-0.862162207352)*x_ref[2]
        ref[4,0,0]=(-0.196356423488)*x_ref[0]**o+(-0.871482508533)*x_ref[0]+(-0.367390886081)*x_ref[1]**o+(0.580310108591)*x_ref[1]+(0.857970671482)*x_ref[2]**o+(0.123685673359)*x_ref[2]
        ref[4,0,1]=(0.715104674022)*x_ref[0]**o+(-0.00444438628932)*x_ref[0]+(0.538883210054)*x_ref[1]**o+(0.179340705463)*x_ref[1]+(-0.605249436909)*x_ref[2]**o+(-0.744253826562)*x_ref[2]
        ref[4,1,0]=(-0.315453893647)*x_ref[0]**o+(-0.560946546082)*x_ref[0]+(-0.674448928458)*x_ref[1]**o+(-0.510243444449)*x_ref[1]+(0.592615876779)*x_ref[2]**o+(0.218024801263)*x_ref[2]
        ref[4,1,1]=(-0.46902473326)*x_ref[0]**o+(-0.573926691452)*x_ref[0]+(0.827898114814)*x_ref[1]**o+(-0.644208082857)*x_ref[1]+(-0.956169254426)*x_ref[2]**o+(-0.297101737561)*x_ref[2]
        ref[5,0,0]=(0.45818164662)*x_ref[0]**o+(0.219667021464)*x_ref[0]+(-0.453125753418)*x_ref[1]**o+(-0.201214929832)*x_ref[1]+(0.135363067088)*x_ref[2]**o+(0.850636247323)*x_ref[2]
        ref[5,0,1]=(0.296120304926)*x_ref[0]**o+(-0.139971296351)*x_ref[0]+(0.448115780401)*x_ref[1]**o+(-0.300602343399)*x_ref[1]+(-0.165568753274)*x_ref[2]**o+(0.984788641639)*x_ref[2]
        ref[5,1,0]=(0.321875942152)*x_ref[0]**o+(-0.403122417353)*x_ref[0]+(-0.148664639927)*x_ref[1]**o+(0.644143662303)*x_ref[1]+(-0.718311175542)*x_ref[2]**o+(-0.850248996913)*x_ref[2]
        ref[5,1,1]=(0.44869757406)*x_ref[0]**o+(0.0292094503086)*x_ref[0]+(-0.159644215279)*x_ref[1]**o+(0.695976060359)*x_ref[1]+(-0.221444458713)*x_ref[2]**o+(-0.42974111206)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromData_Function_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0737428272134)*x[0]**o+(0.841780036645)*x[0]+(-0.110824257531)*x[1]**o+(-0.957726444468)*x[1]
        arg[0,0,0,1]=(-0.934840410671)*x[0]**o+(-0.83301064321)*x[0]+(-0.352287650266)*x[1]**o+(-0.961111564077)*x[1]
        arg[0,0,1,0]=(-0.829500940226)*x[0]**o+(0.644171710931)*x[0]+(0.232545739722)*x[1]**o+(0.584336012862)*x[1]
        arg[0,0,1,1]=(0.426298830976)*x[0]**o+(-0.615281197268)*x[0]+(0.206334026192)*x[1]**o+(-0.443698947044)*x[1]
        arg[0,0,2,0]=(-0.236495891617)*x[0]**o+(0.0883376851179)*x[0]+(-0.176541434356)*x[1]**o+(0.297175301617)*x[1]
        arg[0,0,2,1]=(-0.879955324373)*x[0]**o+(-0.244778913771)*x[0]+(0.244131188594)*x[1]**o+(-0.334739112198)*x[1]
        arg[0,1,0,0]=(-0.64887128006)*x[0]**o+(-0.489951454377)*x[0]+(-0.328402433668)*x[1]**o+(0.229814237117)*x[1]
        arg[0,1,0,1]=(-0.339451656097)*x[0]**o+(-0.409320682304)*x[0]+(0.216482338677)*x[1]**o+(-0.10594373055)*x[1]
        arg[0,1,1,0]=(-0.350538479425)*x[0]**o+(0.466450046985)*x[0]+(-0.779124795292)*x[1]**o+(-0.654562596704)*x[1]
        arg[0,1,1,1]=(0.00350092879868)*x[0]**o+(-0.214643088647)*x[0]+(0.786934081411)*x[1]**o+(-0.296853380043)*x[1]
        arg[0,1,2,0]=(-0.640183955648)*x[0]**o+(0.568647968237)*x[0]+(0.575533476901)*x[1]**o+(0.854540594085)*x[1]
        arg[0,1,2,1]=(0.633674899518)*x[0]**o+(0.829255679443)*x[0]+(0.537888424303)*x[1]**o+(-0.495082587617)*x[1]
        arg[0,2,0,0]=(-0.700436798342)*x[0]**o+(-0.732581859026)*x[0]+(-0.508905353053)*x[1]**o+(0.664601151676)*x[1]
        arg[0,2,0,1]=(0.720335764302)*x[0]**o+(0.941159917988)*x[0]+(0.2817440547)*x[1]**o+(0.553712160058)*x[1]
        arg[0,2,1,0]=(-0.937114312262)*x[0]**o+(-0.115517396201)*x[0]+(-0.672697125905)*x[1]**o+(-0.221320095369)*x[1]
        arg[0,2,1,1]=(-0.52059341062)*x[0]**o+(-0.653834647475)*x[0]+(-0.0851100861135)*x[1]**o+(-0.668300210067)*x[1]
        arg[0,2,2,0]=(-0.943613035509)*x[0]**o+(-0.294843145822)*x[0]+(0.496278748831)*x[1]**o+(0.901154419425)*x[1]
        arg[0,2,2,1]=(-0.421698818203)*x[0]**o+(-0.721918177544)*x[0]+(-0.197855702215)*x[1]**o+(0.50882827715)*x[1]
        arg[0,3,0,0]=(0.931990780298)*x[0]**o+(-0.550332606951)*x[0]+(0.401584654642)*x[1]**o+(0.0965259747601)*x[1]
        arg[0,3,0,1]=(0.881357708445)*x[0]**o+(0.564624034197)*x[0]+(-0.458191897924)*x[1]**o+(-0.427701112807)*x[1]
        arg[0,3,1,0]=(0.624777088081)*x[0]**o+(0.862654721208)*x[0]+(-0.707618093268)*x[1]**o+(0.743304953313)*x[1]
        arg[0,3,1,1]=(0.4510836974)*x[0]**o+(-0.357023031453)*x[0]+(-0.00503924010273)*x[1]**o+(0.545024336011)*x[1]
        arg[0,3,2,0]=(0.109003851094)*x[0]**o+(0.168659376881)*x[0]+(-0.0578953780655)*x[1]**o+(-0.481212462712)*x[1]
        arg[0,3,2,1]=(-0.191183501636)*x[0]**o+(-0.316145516377)*x[0]+(0.402460103016)*x[1]**o+(-0.163612464438)*x[1]
        arg[0,4,0,0]=(0.489815345033)*x[0]**o+(0.999138280765)*x[0]+(-0.626627766812)*x[1]**o+(-0.339653992568)*x[1]
        arg[0,4,0,1]=(0.616350228791)*x[0]**o+(0.933446438765)*x[0]+(0.944244035512)*x[1]**o+(0.829407789353)*x[1]
        arg[0,4,1,0]=(0.619918547863)*x[0]**o+(0.995488719729)*x[0]+(-0.78847429581)*x[1]**o+(0.542645674462)*x[1]
        arg[0,4,1,1]=(-0.149244819992)*x[0]**o+(0.921766299925)*x[0]+(0.435255339567)*x[1]**o+(-0.0596561300741)*x[1]
        arg[0,4,2,0]=(-0.816364336598)*x[0]**o+(0.701160246658)*x[0]+(-0.616385505652)*x[1]**o+(-0.558625482105)*x[1]
        arg[0,4,2,1]=(-0.36944364693)*x[0]**o+(0.243426977623)*x[0]+(-0.61578435391)*x[1]**o+(0.932570326139)*x[1]
        arg[1,0,0,0]=(-0.36839962807)*x[0]**o+(0.590534439534)*x[0]+(-0.615158899289)*x[1]**o+(0.973649158296)*x[1]
        arg[1,0,0,1]=(0.761515772771)*x[0]**o+(0.706406307012)*x[0]+(0.328712319043)*x[1]**o+(0.438600140517)*x[1]
        arg[1,0,1,0]=(0.895080579207)*x[0]**o+(-0.166789328225)*x[0]+(0.807540897371)*x[1]**o+(0.0345522692723)*x[1]
        arg[1,0,1,1]=(-0.460062661172)*x[0]**o+(0.505167337562)*x[0]+(0.25155527144)*x[1]**o+(0.601093887451)*x[1]
        arg[1,0,2,0]=(0.382129818527)*x[0]**o+(0.694676536542)*x[0]+(0.702794119083)*x[1]**o+(0.704094013056)*x[1]
        arg[1,0,2,1]=(0.265476982913)*x[0]**o+(-0.134338744881)*x[0]+(-0.0938910335356)*x[1]**o+(-0.560073316097)*x[1]
        arg[1,1,0,0]=(0.977368328792)*x[0]**o+(0.084482890741)*x[0]+(0.57068389287)*x[1]**o+(-0.242087071526)*x[1]
        arg[1,1,0,1]=(-0.36067775897)*x[0]**o+(0.768187470286)*x[0]+(-0.881150642619)*x[1]**o+(-0.15839498619)*x[1]
        arg[1,1,1,0]=(-0.808263913645)*x[0]**o+(0.788403321664)*x[0]+(0.976206012343)*x[1]**o+(-0.636933353579)*x[1]
        arg[1,1,1,1]=(0.699185445546)*x[0]**o+(0.251811206876)*x[0]+(-0.146320762812)*x[1]**o+(-0.248033048423)*x[1]
        arg[1,1,2,0]=(-0.782940880812)*x[0]**o+(0.789401686528)*x[0]+(-0.806883904562)*x[1]**o+(0.83689696772)*x[1]
        arg[1,1,2,1]=(0.652404993342)*x[0]**o+(-0.135318598212)*x[0]+(0.138928760301)*x[1]**o+(0.675593554805)*x[1]
        arg[1,2,0,0]=(0.471746379585)*x[0]**o+(-0.989783415847)*x[0]+(-0.0935052296418)*x[1]**o+(-0.532245102898)*x[1]
        arg[1,2,0,1]=(-0.44541059979)*x[0]**o+(0.593734266824)*x[0]+(-0.616647011751)*x[1]**o+(0.461140297024)*x[1]
        arg[1,2,1,0]=(-0.9144210459)*x[0]**o+(0.116806032197)*x[0]+(0.935610395306)*x[1]**o+(0.550884822213)*x[1]
        arg[1,2,1,1]=(-0.962444455944)*x[0]**o+(-0.654328013652)*x[0]+(-0.545330994236)*x[1]**o+(-0.919705456162)*x[1]
        arg[1,2,2,0]=(-0.724491374462)*x[0]**o+(-0.820402375498)*x[0]+(-0.953956431336)*x[1]**o+(-0.201450523182)*x[1]
        arg[1,2,2,1]=(0.723637497881)*x[0]**o+(-0.144483238523)*x[0]+(0.00379745988102)*x[1]**o+(-0.838616087543)*x[1]
        arg[1,3,0,0]=(-0.962256046042)*x[0]**o+(-0.453580322779)*x[0]+(0.887446007424)*x[1]**o+(-0.404454768131)*x[1]
        arg[1,3,0,1]=(0.829189872775)*x[0]**o+(0.327985850399)*x[0]+(0.113600094307)*x[1]**o+(-0.338775377551)*x[1]
        arg[1,3,1,0]=(0.912320027833)*x[0]**o+(0.0658152563951)*x[0]+(-0.843491517568)*x[1]**o+(0.858344434434)*x[1]
        arg[1,3,1,1]=(-0.133918678317)*x[0]**o+(0.644272729149)*x[0]+(0.140663927965)*x[1]**o+(0.853385355163)*x[1]
        arg[1,3,2,0]=(0.37733357322)*x[0]**o+(0.255105934063)*x[0]+(0.88881892183)*x[1]**o+(-0.722661948279)*x[1]
        arg[1,3,2,1]=(-0.2827466345)*x[0]**o+(-0.518815029512)*x[0]+(-0.995043824929)*x[1]**o+(0.973473354481)*x[1]
        arg[1,4,0,0]=(0.044904732821)*x[0]**o+(-0.0373604593083)*x[0]+(0.184254626748)*x[1]**o+(0.031832590233)*x[1]
        arg[1,4,0,1]=(0.388699053911)*x[0]**o+(-0.0619606394414)*x[0]+(0.477947819745)*x[1]**o+(0.982909216221)*x[1]
        arg[1,4,1,0]=(0.854348679973)*x[0]**o+(-0.822052612379)*x[0]+(-0.950305089968)*x[1]**o+(0.0100277256384)*x[1]
        arg[1,4,1,1]=(0.265075277179)*x[0]**o+(0.489846993407)*x[0]+(-0.961111499722)*x[1]**o+(-0.0457909055602)*x[1]
        arg[1,4,2,0]=(0.701951191951)*x[0]**o+(0.675232620259)*x[0]+(-0.141848599227)*x[1]**o+(0.737150257564)*x[1]
        arg[1,4,2,1]=(-0.450684564415)*x[0]**o+(0.854849922021)*x[0]+(0.729670746723)*x[1]**o+(0.530372628879)*x[1]
        arg[2,0,0,0]=(-0.131153262675)*x[0]**o+(-0.477683407979)*x[0]+(0.231416752516)*x[1]**o+(0.790897064402)*x[1]
        arg[2,0,0,1]=(0.0956557004382)*x[0]**o+(0.209084913471)*x[0]+(-0.328544695733)*x[1]**o+(0.701245770328)*x[1]
        arg[2,0,1,0]=(-0.476403718361)*x[0]**o+(0.689023533158)*x[0]+(0.108732189071)*x[1]**o+(0.243379977453)*x[1]
        arg[2,0,1,1]=(0.414917801157)*x[0]**o+(-0.843544596785)*x[0]+(0.772236792935)*x[1]**o+(0.836028831232)*x[1]
        arg[2,0,2,0]=(0.754970600617)*x[0]**o+(-0.422318891122)*x[0]+(0.770767911534)*x[1]**o+(-0.446737894095)*x[1]
        arg[2,0,2,1]=(0.0139989331471)*x[0]**o+(0.262286042291)*x[0]+(-0.440109329114)*x[1]**o+(-0.844965114722)*x[1]
        arg[2,1,0,0]=(-0.758977153009)*x[0]**o+(-0.991192011039)*x[0]+(0.642267239475)*x[1]**o+(-0.701360491163)*x[1]
        arg[2,1,0,1]=(-0.123019007509)*x[0]**o+(-0.622323252979)*x[0]+(0.441364031664)*x[1]**o+(0.311433222821)*x[1]
        arg[2,1,1,0]=(-0.0446644182928)*x[0]**o+(0.614952704493)*x[0]+(-0.228284494629)*x[1]**o+(0.077603612991)*x[1]
        arg[2,1,1,1]=(0.712924108225)*x[0]**o+(-0.153966300537)*x[0]+(-0.584450014127)*x[1]**o+(0.476329149833)*x[1]
        arg[2,1,2,0]=(0.554026516037)*x[0]**o+(0.730575821847)*x[0]+(-0.247529915107)*x[1]**o+(0.210060454223)*x[1]
        arg[2,1,2,1]=(0.877017208752)*x[0]**o+(-0.800916200997)*x[0]+(0.0996710064295)*x[1]**o+(-0.0890375735808)*x[1]
        arg[2,2,0,0]=(-0.198026186062)*x[0]**o+(-0.416160224681)*x[0]+(-0.626491103997)*x[1]**o+(-0.950177752043)*x[1]
        arg[2,2,0,1]=(-0.401319234479)*x[0]**o+(0.948120359627)*x[0]+(0.420351950404)*x[1]**o+(-0.242850312594)*x[1]
        arg[2,2,1,0]=(-0.239652069063)*x[0]**o+(-0.894811076801)*x[0]+(0.979721413509)*x[1]**o+(-0.712493893566)*x[1]
        arg[2,2,1,1]=(0.239416494209)*x[0]**o+(0.449047649538)*x[0]+(0.0503751073583)*x[1]**o+(0.739949331521)*x[1]
        arg[2,2,2,0]=(-0.318067701034)*x[0]**o+(-0.00208209268356)*x[0]+(-0.513821624843)*x[1]**o+(-0.202688815501)*x[1]
        arg[2,2,2,1]=(-0.194111524544)*x[0]**o+(-0.843213241929)*x[0]+(0.124223747087)*x[1]**o+(-0.988305059412)*x[1]
        arg[2,3,0,0]=(0.450894655873)*x[0]**o+(-0.809351657482)*x[0]+(-0.80555128185)*x[1]**o+(-0.922123806508)*x[1]
        arg[2,3,0,1]=(-0.462822602909)*x[0]**o+(0.504388766027)*x[0]+(-0.152725064915)*x[1]**o+(0.988745142638)*x[1]
        arg[2,3,1,0]=(-0.717873944493)*x[0]**o+(0.987011714613)*x[0]+(0.908500737929)*x[1]**o+(0.612497707425)*x[1]
        arg[2,3,1,1]=(0.221645363399)*x[0]**o+(0.219147687681)*x[0]+(-0.195831118343)*x[1]**o+(-0.59541920919)*x[1]
        arg[2,3,2,0]=(-0.289487342206)*x[0]**o+(-0.96904981397)*x[0]+(-0.844880792969)*x[1]**o+(0.803145388038)*x[1]
        arg[2,3,2,1]=(-0.0234295826132)*x[0]**o+(-0.839069887016)*x[0]+(0.591231341444)*x[1]**o+(-0.678605318787)*x[1]
        arg[2,4,0,0]=(0.717943320641)*x[0]**o+(0.625728011485)*x[0]+(0.456836257053)*x[1]**o+(-0.00523685585287)*x[1]
        arg[2,4,0,1]=(-0.549526374185)*x[0]**o+(0.57441917876)*x[0]+(-0.913459106781)*x[1]**o+(-0.610369375195)*x[1]
        arg[2,4,1,0]=(-0.27166563459)*x[0]**o+(0.146710581194)*x[0]+(0.297934124718)*x[1]**o+(0.989561209048)*x[1]
        arg[2,4,1,1]=(-0.0973858709778)*x[0]**o+(-0.909133394381)*x[0]+(0.634182571287)*x[1]**o+(0.477163052831)*x[1]
        arg[2,4,2,0]=(0.0878998038301)*x[0]**o+(-0.90568487021)*x[0]+(0.798978355456)*x[1]**o+(0.977402478037)*x[1]
        arg[2,4,2,1]=(-0.456868538667)*x[0]**o+(0.0723272702119)*x[0]+(0.143461337462)*x[1]**o+(-0.923713568343)*x[1]
        arg[3,0,0,0]=(-0.154410689788)*x[0]**o+(0.634925838303)*x[0]+(-0.0508765762892)*x[1]**o+(-0.889731180834)*x[1]
        arg[3,0,0,1]=(-0.611336323518)*x[0]**o+(0.748058372943)*x[0]+(-0.395060568704)*x[1]**o+(-0.413772262434)*x[1]
        arg[3,0,1,0]=(-0.76924573047)*x[0]**o+(0.61286303551)*x[0]+(0.810883211884)*x[1]**o+(-0.908861022244)*x[1]
        arg[3,0,1,1]=(0.269999661235)*x[0]**o+(0.0330954575695)*x[0]+(0.560272546054)*x[1]**o+(0.877730763728)*x[1]
        arg[3,0,2,0]=(-0.617658215856)*x[0]**o+(0.758414649277)*x[0]+(-0.223497113863)*x[1]**o+(-0.0706589108357)*x[1]
        arg[3,0,2,1]=(0.47956571804)*x[0]**o+(0.938375191345)*x[0]+(-0.193997511219)*x[1]**o+(0.500063249143)*x[1]
        arg[3,1,0,0]=(0.994033773649)*x[0]**o+(0.203022945646)*x[0]+(0.439296640709)*x[1]**o+(0.283679657893)*x[1]
        arg[3,1,0,1]=(0.32495149734)*x[0]**o+(-0.871627347694)*x[0]+(-0.438159387872)*x[1]**o+(0.169112665588)*x[1]
        arg[3,1,1,0]=(-0.0979322945554)*x[0]**o+(-0.936928842822)*x[0]+(0.0402738011011)*x[1]**o+(-0.445031618539)*x[1]
        arg[3,1,1,1]=(-0.511218949907)*x[0]**o+(0.255146779315)*x[0]+(-0.802012163396)*x[1]**o+(0.361819705056)*x[1]
        arg[3,1,2,0]=(-0.902499203011)*x[0]**o+(0.76589952141)*x[0]+(0.52498929732)*x[1]**o+(0.493894292321)*x[1]
        arg[3,1,2,1]=(-0.475117609018)*x[0]**o+(0.682205309303)*x[0]+(-0.661824086649)*x[1]**o+(0.985521553774)*x[1]
        arg[3,2,0,0]=(0.937321686473)*x[0]**o+(-0.518305775925)*x[0]+(0.171689574942)*x[1]**o+(-0.751579960819)*x[1]
        arg[3,2,0,1]=(-0.607004597985)*x[0]**o+(0.342330208344)*x[0]+(0.692094243326)*x[1]**o+(-0.457756057357)*x[1]
        arg[3,2,1,0]=(0.126171621354)*x[0]**o+(0.0499938423847)*x[0]+(0.572006209576)*x[1]**o+(0.121601182464)*x[1]
        arg[3,2,1,1]=(0.337936456181)*x[0]**o+(-0.940186907514)*x[0]+(-0.965354931124)*x[1]**o+(-0.0503295532873)*x[1]
        arg[3,2,2,0]=(0.439590816231)*x[0]**o+(-0.145314466072)*x[0]+(-0.847175364201)*x[1]**o+(-0.570000251495)*x[1]
        arg[3,2,2,1]=(0.401069335123)*x[0]**o+(-0.117139167508)*x[0]+(0.192907527853)*x[1]**o+(0.573283391581)*x[1]
        arg[3,3,0,0]=(-0.434029411666)*x[0]**o+(-0.689115317009)*x[0]+(0.63450218888)*x[1]**o+(0.492858023519)*x[1]
        arg[3,3,0,1]=(-0.678722283369)*x[0]**o+(-0.62031826393)*x[0]+(-0.458646022327)*x[1]**o+(-0.650857410927)*x[1]
        arg[3,3,1,0]=(0.413477874719)*x[0]**o+(-0.903116096879)*x[0]+(-0.505961511722)*x[1]**o+(0.649554735459)*x[1]
        arg[3,3,1,1]=(-0.58459645458)*x[0]**o+(0.861318565628)*x[0]+(0.378511063123)*x[1]**o+(-0.694820011654)*x[1]
        arg[3,3,2,0]=(0.82844733258)*x[0]**o+(-0.151929131446)*x[0]+(0.322848411081)*x[1]**o+(0.436062233263)*x[1]
        arg[3,3,2,1]=(0.269627291792)*x[0]**o+(0.618140998132)*x[0]+(0.851057431407)*x[1]**o+(0.716088832502)*x[1]
        arg[3,4,0,0]=(0.0557605432613)*x[0]**o+(0.0689569775656)*x[0]+(0.489781121774)*x[1]**o+(0.214459998733)*x[1]
        arg[3,4,0,1]=(0.429111134289)*x[0]**o+(0.648324951051)*x[0]+(0.00596799425797)*x[1]**o+(-0.752637714965)*x[1]
        arg[3,4,1,0]=(-0.0394648097179)*x[0]**o+(0.842965554058)*x[0]+(-0.124493160479)*x[1]**o+(0.81914053244)*x[1]
        arg[3,4,1,1]=(0.309465325874)*x[0]**o+(0.631381272937)*x[0]+(-0.164415508845)*x[1]**o+(0.316908911599)*x[1]
        arg[3,4,2,0]=(-0.615108314792)*x[0]**o+(0.300936829452)*x[0]+(-0.289514252333)*x[1]**o+(0.844949030767)*x[1]
        arg[3,4,2,1]=(0.80117516909)*x[0]**o+(-0.602632221639)*x[0]+(-0.68193581457)*x[1]**o+(0.203319693067)*x[1]
        ref[0,0,0,0]=(0.0737428272134)*x_ref[0]**o+(0.841780036645)*x_ref[0]+(-0.110824257531)*x_ref[1]**o+(-0.957726444468)*x_ref[1]
        ref[0,0,0,1]=(-0.934840410671)*x_ref[0]**o+(-0.83301064321)*x_ref[0]+(-0.352287650266)*x_ref[1]**o+(-0.961111564077)*x_ref[1]
        ref[0,0,1,0]=(-0.829500940226)*x_ref[0]**o+(0.644171710931)*x_ref[0]+(0.232545739722)*x_ref[1]**o+(0.584336012862)*x_ref[1]
        ref[0,0,1,1]=(0.426298830976)*x_ref[0]**o+(-0.615281197268)*x_ref[0]+(0.206334026192)*x_ref[1]**o+(-0.443698947044)*x_ref[1]
        ref[0,0,2,0]=(-0.236495891617)*x_ref[0]**o+(0.0883376851179)*x_ref[0]+(-0.176541434356)*x_ref[1]**o+(0.297175301617)*x_ref[1]
        ref[0,0,2,1]=(-0.879955324373)*x_ref[0]**o+(-0.244778913771)*x_ref[0]+(0.244131188594)*x_ref[1]**o+(-0.334739112198)*x_ref[1]
        ref[0,1,0,0]=(-0.64887128006)*x_ref[0]**o+(-0.489951454377)*x_ref[0]+(-0.328402433668)*x_ref[1]**o+(0.229814237117)*x_ref[1]
        ref[0,1,0,1]=(-0.339451656097)*x_ref[0]**o+(-0.409320682304)*x_ref[0]+(0.216482338677)*x_ref[1]**o+(-0.10594373055)*x_ref[1]
        ref[0,1,1,0]=(-0.350538479425)*x_ref[0]**o+(0.466450046985)*x_ref[0]+(-0.779124795292)*x_ref[1]**o+(-0.654562596704)*x_ref[1]
        ref[0,1,1,1]=(0.00350092879868)*x_ref[0]**o+(-0.214643088647)*x_ref[0]+(0.786934081411)*x_ref[1]**o+(-0.296853380043)*x_ref[1]
        ref[0,1,2,0]=(-0.640183955648)*x_ref[0]**o+(0.568647968237)*x_ref[0]+(0.575533476901)*x_ref[1]**o+(0.854540594085)*x_ref[1]
        ref[0,1,2,1]=(0.633674899518)*x_ref[0]**o+(0.829255679443)*x_ref[0]+(0.537888424303)*x_ref[1]**o+(-0.495082587617)*x_ref[1]
        ref[0,2,0,0]=(-0.700436798342)*x_ref[0]**o+(-0.732581859026)*x_ref[0]+(-0.508905353053)*x_ref[1]**o+(0.664601151676)*x_ref[1]
        ref[0,2,0,1]=(0.720335764302)*x_ref[0]**o+(0.941159917988)*x_ref[0]+(0.2817440547)*x_ref[1]**o+(0.553712160058)*x_ref[1]
        ref[0,2,1,0]=(-0.937114312262)*x_ref[0]**o+(-0.115517396201)*x_ref[0]+(-0.672697125905)*x_ref[1]**o+(-0.221320095369)*x_ref[1]
        ref[0,2,1,1]=(-0.52059341062)*x_ref[0]**o+(-0.653834647475)*x_ref[0]+(-0.0851100861135)*x_ref[1]**o+(-0.668300210067)*x_ref[1]
        ref[0,2,2,0]=(-0.943613035509)*x_ref[0]**o+(-0.294843145822)*x_ref[0]+(0.496278748831)*x_ref[1]**o+(0.901154419425)*x_ref[1]
        ref[0,2,2,1]=(-0.421698818203)*x_ref[0]**o+(-0.721918177544)*x_ref[0]+(-0.197855702215)*x_ref[1]**o+(0.50882827715)*x_ref[1]
        ref[0,3,0,0]=(0.931990780298)*x_ref[0]**o+(-0.550332606951)*x_ref[0]+(0.401584654642)*x_ref[1]**o+(0.0965259747601)*x_ref[1]
        ref[0,3,0,1]=(0.881357708445)*x_ref[0]**o+(0.564624034197)*x_ref[0]+(-0.458191897924)*x_ref[1]**o+(-0.427701112807)*x_ref[1]
        ref[0,3,1,0]=(0.624777088081)*x_ref[0]**o+(0.862654721208)*x_ref[0]+(-0.707618093268)*x_ref[1]**o+(0.743304953313)*x_ref[1]
        ref[0,3,1,1]=(0.4510836974)*x_ref[0]**o+(-0.357023031453)*x_ref[0]+(-0.00503924010273)*x_ref[1]**o+(0.545024336011)*x_ref[1]
        ref[0,3,2,0]=(0.109003851094)*x_ref[0]**o+(0.168659376881)*x_ref[0]+(-0.0578953780655)*x_ref[1]**o+(-0.481212462712)*x_ref[1]
        ref[0,3,2,1]=(-0.191183501636)*x_ref[0]**o+(-0.316145516377)*x_ref[0]+(0.402460103016)*x_ref[1]**o+(-0.163612464438)*x_ref[1]
        ref[0,4,0,0]=(0.489815345033)*x_ref[0]**o+(0.999138280765)*x_ref[0]+(-0.626627766812)*x_ref[1]**o+(-0.339653992568)*x_ref[1]
        ref[0,4,0,1]=(0.616350228791)*x_ref[0]**o+(0.933446438765)*x_ref[0]+(0.944244035512)*x_ref[1]**o+(0.829407789353)*x_ref[1]
        ref[0,4,1,0]=(0.619918547863)*x_ref[0]**o+(0.995488719729)*x_ref[0]+(-0.78847429581)*x_ref[1]**o+(0.542645674462)*x_ref[1]
        ref[0,4,1,1]=(-0.149244819992)*x_ref[0]**o+(0.921766299925)*x_ref[0]+(0.435255339567)*x_ref[1]**o+(-0.0596561300741)*x_ref[1]
        ref[0,4,2,0]=(-0.816364336598)*x_ref[0]**o+(0.701160246658)*x_ref[0]+(-0.616385505652)*x_ref[1]**o+(-0.558625482105)*x_ref[1]
        ref[0,4,2,1]=(-0.36944364693)*x_ref[0]**o+(0.243426977623)*x_ref[0]+(-0.61578435391)*x_ref[1]**o+(0.932570326139)*x_ref[1]
        ref[1,0,0,0]=(-0.36839962807)*x_ref[0]**o+(0.590534439534)*x_ref[0]+(-0.615158899289)*x_ref[1]**o+(0.973649158296)*x_ref[1]
        ref[1,0,0,1]=(0.761515772771)*x_ref[0]**o+(0.706406307012)*x_ref[0]+(0.328712319043)*x_ref[1]**o+(0.438600140517)*x_ref[1]
        ref[1,0,1,0]=(0.895080579207)*x_ref[0]**o+(-0.166789328225)*x_ref[0]+(0.807540897371)*x_ref[1]**o+(0.0345522692723)*x_ref[1]
        ref[1,0,1,1]=(-0.460062661172)*x_ref[0]**o+(0.505167337562)*x_ref[0]+(0.25155527144)*x_ref[1]**o+(0.601093887451)*x_ref[1]
        ref[1,0,2,0]=(0.382129818527)*x_ref[0]**o+(0.694676536542)*x_ref[0]+(0.702794119083)*x_ref[1]**o+(0.704094013056)*x_ref[1]
        ref[1,0,2,1]=(0.265476982913)*x_ref[0]**o+(-0.134338744881)*x_ref[0]+(-0.0938910335356)*x_ref[1]**o+(-0.560073316097)*x_ref[1]
        ref[1,1,0,0]=(0.977368328792)*x_ref[0]**o+(0.084482890741)*x_ref[0]+(0.57068389287)*x_ref[1]**o+(-0.242087071526)*x_ref[1]
        ref[1,1,0,1]=(-0.36067775897)*x_ref[0]**o+(0.768187470286)*x_ref[0]+(-0.881150642619)*x_ref[1]**o+(-0.15839498619)*x_ref[1]
        ref[1,1,1,0]=(-0.808263913645)*x_ref[0]**o+(0.788403321664)*x_ref[0]+(0.976206012343)*x_ref[1]**o+(-0.636933353579)*x_ref[1]
        ref[1,1,1,1]=(0.699185445546)*x_ref[0]**o+(0.251811206876)*x_ref[0]+(-0.146320762812)*x_ref[1]**o+(-0.248033048423)*x_ref[1]
        ref[1,1,2,0]=(-0.782940880812)*x_ref[0]**o+(0.789401686528)*x_ref[0]+(-0.806883904562)*x_ref[1]**o+(0.83689696772)*x_ref[1]
        ref[1,1,2,1]=(0.652404993342)*x_ref[0]**o+(-0.135318598212)*x_ref[0]+(0.138928760301)*x_ref[1]**o+(0.675593554805)*x_ref[1]
        ref[1,2,0,0]=(0.471746379585)*x_ref[0]**o+(-0.989783415847)*x_ref[0]+(-0.0935052296418)*x_ref[1]**o+(-0.532245102898)*x_ref[1]
        ref[1,2,0,1]=(-0.44541059979)*x_ref[0]**o+(0.593734266824)*x_ref[0]+(-0.616647011751)*x_ref[1]**o+(0.461140297024)*x_ref[1]
        ref[1,2,1,0]=(-0.9144210459)*x_ref[0]**o+(0.116806032197)*x_ref[0]+(0.935610395306)*x_ref[1]**o+(0.550884822213)*x_ref[1]
        ref[1,2,1,1]=(-0.962444455944)*x_ref[0]**o+(-0.654328013652)*x_ref[0]+(-0.545330994236)*x_ref[1]**o+(-0.919705456162)*x_ref[1]
        ref[1,2,2,0]=(-0.724491374462)*x_ref[0]**o+(-0.820402375498)*x_ref[0]+(-0.953956431336)*x_ref[1]**o+(-0.201450523182)*x_ref[1]
        ref[1,2,2,1]=(0.723637497881)*x_ref[0]**o+(-0.144483238523)*x_ref[0]+(0.00379745988102)*x_ref[1]**o+(-0.838616087543)*x_ref[1]
        ref[1,3,0,0]=(-0.962256046042)*x_ref[0]**o+(-0.453580322779)*x_ref[0]+(0.887446007424)*x_ref[1]**o+(-0.404454768131)*x_ref[1]
        ref[1,3,0,1]=(0.829189872775)*x_ref[0]**o+(0.327985850399)*x_ref[0]+(0.113600094307)*x_ref[1]**o+(-0.338775377551)*x_ref[1]
        ref[1,3,1,0]=(0.912320027833)*x_ref[0]**o+(0.0658152563951)*x_ref[0]+(-0.843491517568)*x_ref[1]**o+(0.858344434434)*x_ref[1]
        ref[1,3,1,1]=(-0.133918678317)*x_ref[0]**o+(0.644272729149)*x_ref[0]+(0.140663927965)*x_ref[1]**o+(0.853385355163)*x_ref[1]
        ref[1,3,2,0]=(0.37733357322)*x_ref[0]**o+(0.255105934063)*x_ref[0]+(0.88881892183)*x_ref[1]**o+(-0.722661948279)*x_ref[1]
        ref[1,3,2,1]=(-0.2827466345)*x_ref[0]**o+(-0.518815029512)*x_ref[0]+(-0.995043824929)*x_ref[1]**o+(0.973473354481)*x_ref[1]
        ref[1,4,0,0]=(0.044904732821)*x_ref[0]**o+(-0.0373604593083)*x_ref[0]+(0.184254626748)*x_ref[1]**o+(0.031832590233)*x_ref[1]
        ref[1,4,0,1]=(0.388699053911)*x_ref[0]**o+(-0.0619606394414)*x_ref[0]+(0.477947819745)*x_ref[1]**o+(0.982909216221)*x_ref[1]
        ref[1,4,1,0]=(0.854348679973)*x_ref[0]**o+(-0.822052612379)*x_ref[0]+(-0.950305089968)*x_ref[1]**o+(0.0100277256384)*x_ref[1]
        ref[1,4,1,1]=(0.265075277179)*x_ref[0]**o+(0.489846993407)*x_ref[0]+(-0.961111499722)*x_ref[1]**o+(-0.0457909055602)*x_ref[1]
        ref[1,4,2,0]=(0.701951191951)*x_ref[0]**o+(0.675232620259)*x_ref[0]+(-0.141848599227)*x_ref[1]**o+(0.737150257564)*x_ref[1]
        ref[1,4,2,1]=(-0.450684564415)*x_ref[0]**o+(0.854849922021)*x_ref[0]+(0.729670746723)*x_ref[1]**o+(0.530372628879)*x_ref[1]
        ref[2,0,0,0]=(-0.131153262675)*x_ref[0]**o+(-0.477683407979)*x_ref[0]+(0.231416752516)*x_ref[1]**o+(0.790897064402)*x_ref[1]
        ref[2,0,0,1]=(0.0956557004382)*x_ref[0]**o+(0.209084913471)*x_ref[0]+(-0.328544695733)*x_ref[1]**o+(0.701245770328)*x_ref[1]
        ref[2,0,1,0]=(-0.476403718361)*x_ref[0]**o+(0.689023533158)*x_ref[0]+(0.108732189071)*x_ref[1]**o+(0.243379977453)*x_ref[1]
        ref[2,0,1,1]=(0.414917801157)*x_ref[0]**o+(-0.843544596785)*x_ref[0]+(0.772236792935)*x_ref[1]**o+(0.836028831232)*x_ref[1]
        ref[2,0,2,0]=(0.754970600617)*x_ref[0]**o+(-0.422318891122)*x_ref[0]+(0.770767911534)*x_ref[1]**o+(-0.446737894095)*x_ref[1]
        ref[2,0,2,1]=(0.0139989331471)*x_ref[0]**o+(0.262286042291)*x_ref[0]+(-0.440109329114)*x_ref[1]**o+(-0.844965114722)*x_ref[1]
        ref[2,1,0,0]=(-0.758977153009)*x_ref[0]**o+(-0.991192011039)*x_ref[0]+(0.642267239475)*x_ref[1]**o+(-0.701360491163)*x_ref[1]
        ref[2,1,0,1]=(-0.123019007509)*x_ref[0]**o+(-0.622323252979)*x_ref[0]+(0.441364031664)*x_ref[1]**o+(0.311433222821)*x_ref[1]
        ref[2,1,1,0]=(-0.0446644182928)*x_ref[0]**o+(0.614952704493)*x_ref[0]+(-0.228284494629)*x_ref[1]**o+(0.077603612991)*x_ref[1]
        ref[2,1,1,1]=(0.712924108225)*x_ref[0]**o+(-0.153966300537)*x_ref[0]+(-0.584450014127)*x_ref[1]**o+(0.476329149833)*x_ref[1]
        ref[2,1,2,0]=(0.554026516037)*x_ref[0]**o+(0.730575821847)*x_ref[0]+(-0.247529915107)*x_ref[1]**o+(0.210060454223)*x_ref[1]
        ref[2,1,2,1]=(0.877017208752)*x_ref[0]**o+(-0.800916200997)*x_ref[0]+(0.0996710064295)*x_ref[1]**o+(-0.0890375735808)*x_ref[1]
        ref[2,2,0,0]=(-0.198026186062)*x_ref[0]**o+(-0.416160224681)*x_ref[0]+(-0.626491103997)*x_ref[1]**o+(-0.950177752043)*x_ref[1]
        ref[2,2,0,1]=(-0.401319234479)*x_ref[0]**o+(0.948120359627)*x_ref[0]+(0.420351950404)*x_ref[1]**o+(-0.242850312594)*x_ref[1]
        ref[2,2,1,0]=(-0.239652069063)*x_ref[0]**o+(-0.894811076801)*x_ref[0]+(0.979721413509)*x_ref[1]**o+(-0.712493893566)*x_ref[1]
        ref[2,2,1,1]=(0.239416494209)*x_ref[0]**o+(0.449047649538)*x_ref[0]+(0.0503751073583)*x_ref[1]**o+(0.739949331521)*x_ref[1]
        ref[2,2,2,0]=(-0.318067701034)*x_ref[0]**o+(-0.00208209268356)*x_ref[0]+(-0.513821624843)*x_ref[1]**o+(-0.202688815501)*x_ref[1]
        ref[2,2,2,1]=(-0.194111524544)*x_ref[0]**o+(-0.843213241929)*x_ref[0]+(0.124223747087)*x_ref[1]**o+(-0.988305059412)*x_ref[1]
        ref[2,3,0,0]=(0.450894655873)*x_ref[0]**o+(-0.809351657482)*x_ref[0]+(-0.80555128185)*x_ref[1]**o+(-0.922123806508)*x_ref[1]
        ref[2,3,0,1]=(-0.462822602909)*x_ref[0]**o+(0.504388766027)*x_ref[0]+(-0.152725064915)*x_ref[1]**o+(0.988745142638)*x_ref[1]
        ref[2,3,1,0]=(-0.717873944493)*x_ref[0]**o+(0.987011714613)*x_ref[0]+(0.908500737929)*x_ref[1]**o+(0.612497707425)*x_ref[1]
        ref[2,3,1,1]=(0.221645363399)*x_ref[0]**o+(0.219147687681)*x_ref[0]+(-0.195831118343)*x_ref[1]**o+(-0.59541920919)*x_ref[1]
        ref[2,3,2,0]=(-0.289487342206)*x_ref[0]**o+(-0.96904981397)*x_ref[0]+(-0.844880792969)*x_ref[1]**o+(0.803145388038)*x_ref[1]
        ref[2,3,2,1]=(-0.0234295826132)*x_ref[0]**o+(-0.839069887016)*x_ref[0]+(0.591231341444)*x_ref[1]**o+(-0.678605318787)*x_ref[1]
        ref[2,4,0,0]=(0.717943320641)*x_ref[0]**o+(0.625728011485)*x_ref[0]+(0.456836257053)*x_ref[1]**o+(-0.00523685585287)*x_ref[1]
        ref[2,4,0,1]=(-0.549526374185)*x_ref[0]**o+(0.57441917876)*x_ref[0]+(-0.913459106781)*x_ref[1]**o+(-0.610369375195)*x_ref[1]
        ref[2,4,1,0]=(-0.27166563459)*x_ref[0]**o+(0.146710581194)*x_ref[0]+(0.297934124718)*x_ref[1]**o+(0.989561209048)*x_ref[1]
        ref[2,4,1,1]=(-0.0973858709778)*x_ref[0]**o+(-0.909133394381)*x_ref[0]+(0.634182571287)*x_ref[1]**o+(0.477163052831)*x_ref[1]
        ref[2,4,2,0]=(0.0878998038301)*x_ref[0]**o+(-0.90568487021)*x_ref[0]+(0.798978355456)*x_ref[1]**o+(0.977402478037)*x_ref[1]
        ref[2,4,2,1]=(-0.456868538667)*x_ref[0]**o+(0.0723272702119)*x_ref[0]+(0.143461337462)*x_ref[1]**o+(-0.923713568343)*x_ref[1]
        ref[3,0,0,0]=(-0.154410689788)*x_ref[0]**o+(0.634925838303)*x_ref[0]+(-0.0508765762892)*x_ref[1]**o+(-0.889731180834)*x_ref[1]
        ref[3,0,0,1]=(-0.611336323518)*x_ref[0]**o+(0.748058372943)*x_ref[0]+(-0.395060568704)*x_ref[1]**o+(-0.413772262434)*x_ref[1]
        ref[3,0,1,0]=(-0.76924573047)*x_ref[0]**o+(0.61286303551)*x_ref[0]+(0.810883211884)*x_ref[1]**o+(-0.908861022244)*x_ref[1]
        ref[3,0,1,1]=(0.269999661235)*x_ref[0]**o+(0.0330954575695)*x_ref[0]+(0.560272546054)*x_ref[1]**o+(0.877730763728)*x_ref[1]
        ref[3,0,2,0]=(-0.617658215856)*x_ref[0]**o+(0.758414649277)*x_ref[0]+(-0.223497113863)*x_ref[1]**o+(-0.0706589108357)*x_ref[1]
        ref[3,0,2,1]=(0.47956571804)*x_ref[0]**o+(0.938375191345)*x_ref[0]+(-0.193997511219)*x_ref[1]**o+(0.500063249143)*x_ref[1]
        ref[3,1,0,0]=(0.994033773649)*x_ref[0]**o+(0.203022945646)*x_ref[0]+(0.439296640709)*x_ref[1]**o+(0.283679657893)*x_ref[1]
        ref[3,1,0,1]=(0.32495149734)*x_ref[0]**o+(-0.871627347694)*x_ref[0]+(-0.438159387872)*x_ref[1]**o+(0.169112665588)*x_ref[1]
        ref[3,1,1,0]=(-0.0979322945554)*x_ref[0]**o+(-0.936928842822)*x_ref[0]+(0.0402738011011)*x_ref[1]**o+(-0.445031618539)*x_ref[1]
        ref[3,1,1,1]=(-0.511218949907)*x_ref[0]**o+(0.255146779315)*x_ref[0]+(-0.802012163396)*x_ref[1]**o+(0.361819705056)*x_ref[1]
        ref[3,1,2,0]=(-0.902499203011)*x_ref[0]**o+(0.76589952141)*x_ref[0]+(0.52498929732)*x_ref[1]**o+(0.493894292321)*x_ref[1]
        ref[3,1,2,1]=(-0.475117609018)*x_ref[0]**o+(0.682205309303)*x_ref[0]+(-0.661824086649)*x_ref[1]**o+(0.985521553774)*x_ref[1]
        ref[3,2,0,0]=(0.937321686473)*x_ref[0]**o+(-0.518305775925)*x_ref[0]+(0.171689574942)*x_ref[1]**o+(-0.751579960819)*x_ref[1]
        ref[3,2,0,1]=(-0.607004597985)*x_ref[0]**o+(0.342330208344)*x_ref[0]+(0.692094243326)*x_ref[1]**o+(-0.457756057357)*x_ref[1]
        ref[3,2,1,0]=(0.126171621354)*x_ref[0]**o+(0.0499938423847)*x_ref[0]+(0.572006209576)*x_ref[1]**o+(0.121601182464)*x_ref[1]
        ref[3,2,1,1]=(0.337936456181)*x_ref[0]**o+(-0.940186907514)*x_ref[0]+(-0.965354931124)*x_ref[1]**o+(-0.0503295532873)*x_ref[1]
        ref[3,2,2,0]=(0.439590816231)*x_ref[0]**o+(-0.145314466072)*x_ref[0]+(-0.847175364201)*x_ref[1]**o+(-0.570000251495)*x_ref[1]
        ref[3,2,2,1]=(0.401069335123)*x_ref[0]**o+(-0.117139167508)*x_ref[0]+(0.192907527853)*x_ref[1]**o+(0.573283391581)*x_ref[1]
        ref[3,3,0,0]=(-0.434029411666)*x_ref[0]**o+(-0.689115317009)*x_ref[0]+(0.63450218888)*x_ref[1]**o+(0.492858023519)*x_ref[1]
        ref[3,3,0,1]=(-0.678722283369)*x_ref[0]**o+(-0.62031826393)*x_ref[0]+(-0.458646022327)*x_ref[1]**o+(-0.650857410927)*x_ref[1]
        ref[3,3,1,0]=(0.413477874719)*x_ref[0]**o+(-0.903116096879)*x_ref[0]+(-0.505961511722)*x_ref[1]**o+(0.649554735459)*x_ref[1]
        ref[3,3,1,1]=(-0.58459645458)*x_ref[0]**o+(0.861318565628)*x_ref[0]+(0.378511063123)*x_ref[1]**o+(-0.694820011654)*x_ref[1]
        ref[3,3,2,0]=(0.82844733258)*x_ref[0]**o+(-0.151929131446)*x_ref[0]+(0.322848411081)*x_ref[1]**o+(0.436062233263)*x_ref[1]
        ref[3,3,2,1]=(0.269627291792)*x_ref[0]**o+(0.618140998132)*x_ref[0]+(0.851057431407)*x_ref[1]**o+(0.716088832502)*x_ref[1]
        ref[3,4,0,0]=(0.0557605432613)*x_ref[0]**o+(0.0689569775656)*x_ref[0]+(0.489781121774)*x_ref[1]**o+(0.214459998733)*x_ref[1]
        ref[3,4,0,1]=(0.429111134289)*x_ref[0]**o+(0.648324951051)*x_ref[0]+(0.00596799425797)*x_ref[1]**o+(-0.752637714965)*x_ref[1]
        ref[3,4,1,0]=(-0.0394648097179)*x_ref[0]**o+(0.842965554058)*x_ref[0]+(-0.124493160479)*x_ref[1]**o+(0.81914053244)*x_ref[1]
        ref[3,4,1,1]=(0.309465325874)*x_ref[0]**o+(0.631381272937)*x_ref[0]+(-0.164415508845)*x_ref[1]**o+(0.316908911599)*x_ref[1]
        ref[3,4,2,0]=(-0.615108314792)*x_ref[0]**o+(0.300936829452)*x_ref[0]+(-0.289514252333)*x_ref[1]**o+(0.844949030767)*x_ref[1]
        ref[3,4,2,1]=(0.80117516909)*x_ref[0]**o+(-0.602632221639)*x_ref[0]+(-0.68193581457)*x_ref[1]**o+(0.203319693067)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.349437323395)*x[0]**o+(0.817821891642)*x[0]+(-0.357998520848)*x[1]**o+(0.349311985429)*x[1]+(0.470351907942)*x[2]**o+(0.264007843212)*x[2]
        arg[0,0,0,1]=(-0.747824148035)*x[0]**o+(0.551817215468)*x[0]+(0.980476067966)*x[1]**o+(0.521120126285)*x[1]+(-0.472357720779)*x[2]**o+(-0.379502066444)*x[2]
        arg[0,0,1,0]=(0.302984899868)*x[0]**o+(0.556601505796)*x[0]+(-0.698948476338)*x[1]**o+(-0.587379572024)*x[1]+(-0.625760359909)*x[2]**o+(0.231956543432)*x[2]
        arg[0,0,1,1]=(-0.335943560463)*x[0]**o+(-0.497643152155)*x[0]+(-0.849829024888)*x[1]**o+(-0.388751135318)*x[1]+(-0.517696637743)*x[2]**o+(-0.966210901)*x[2]
        arg[0,0,2,0]=(-0.0953058405321)*x[0]**o+(-0.864580918175)*x[0]+(-0.732086619685)*x[1]**o+(0.872044654816)*x[1]+(-0.225850391375)*x[2]**o+(-0.502203654371)*x[2]
        arg[0,0,2,1]=(0.228408069101)*x[0]**o+(0.267590362458)*x[0]+(0.127397041031)*x[1]**o+(0.0142393720782)*x[1]+(-0.257563412478)*x[2]**o+(0.0211877877507)*x[2]
        arg[0,1,0,0]=(0.921125039994)*x[0]**o+(0.372548304521)*x[0]+(-0.66574364308)*x[1]**o+(0.0895600219155)*x[1]+(-0.656870562037)*x[2]**o+(-0.608819638268)*x[2]
        arg[0,1,0,1]=(-0.396625252579)*x[0]**o+(0.655542320616)*x[0]+(-0.542629502059)*x[1]**o+(0.587317299384)*x[1]+(0.96380848967)*x[2]**o+(0.121341827766)*x[2]
        arg[0,1,1,0]=(0.0908208346003)*x[0]**o+(-0.0398227321951)*x[0]+(0.907607555288)*x[1]**o+(0.527230589013)*x[1]+(0.519287416653)*x[2]**o+(0.103180358249)*x[2]
        arg[0,1,1,1]=(0.0615195625837)*x[0]**o+(0.695720276474)*x[0]+(0.0862101684049)*x[1]**o+(0.517150419955)*x[1]+(0.511207669575)*x[2]**o+(-0.9556472362)*x[2]
        arg[0,1,2,0]=(0.70040395652)*x[0]**o+(-0.576266027865)*x[0]+(-0.133947982034)*x[1]**o+(-0.764976368456)*x[1]+(-0.860780506782)*x[2]**o+(-0.137946340663)*x[2]
        arg[0,1,2,1]=(0.356017596832)*x[0]**o+(0.94912783604)*x[0]+(0.325951760294)*x[1]**o+(0.0623695610721)*x[1]+(-0.106031051895)*x[2]**o+(-0.339314970581)*x[2]
        arg[0,2,0,0]=(-0.124954395016)*x[0]**o+(0.558233271782)*x[0]+(-0.956846820394)*x[1]**o+(-0.599332648704)*x[1]+(-0.322444201477)*x[2]**o+(0.0833108732532)*x[2]
        arg[0,2,0,1]=(-0.713065259589)*x[0]**o+(-0.597746726477)*x[0]+(-0.976654446591)*x[1]**o+(0.977779698697)*x[1]+(0.694427679422)*x[2]**o+(-0.364463986712)*x[2]
        arg[0,2,1,0]=(-0.415861468352)*x[0]**o+(-0.159802964908)*x[0]+(-0.132331280778)*x[1]**o+(0.120829594848)*x[1]+(-0.114181951611)*x[2]**o+(-0.810187313564)*x[2]
        arg[0,2,1,1]=(-0.869925695451)*x[0]**o+(-0.535444611026)*x[0]+(0.62908837594)*x[1]**o+(-0.488583607314)*x[1]+(0.8461615049)*x[2]**o+(-0.558010294727)*x[2]
        arg[0,2,2,0]=(-0.0822232217466)*x[0]**o+(-0.673631382141)*x[0]+(0.803717792107)*x[1]**o+(0.339934715615)*x[1]+(-0.886365924996)*x[2]**o+(0.0206517750753)*x[2]
        arg[0,2,2,1]=(0.367656249869)*x[0]**o+(0.000987963028539)*x[0]+(0.399182215451)*x[1]**o+(0.923259635138)*x[1]+(0.8573338122)*x[2]**o+(0.679082213415)*x[2]
        arg[0,3,0,0]=(-0.722680496644)*x[0]**o+(0.199427912651)*x[0]+(-0.351799245099)*x[1]**o+(0.240285416231)*x[1]+(-0.701704868242)*x[2]**o+(0.265610436018)*x[2]
        arg[0,3,0,1]=(-0.14479043328)*x[0]**o+(-0.192670959504)*x[0]+(0.327412080296)*x[1]**o+(-0.390894637576)*x[1]+(-0.401862813677)*x[2]**o+(-0.87150071561)*x[2]
        arg[0,3,1,0]=(0.430039400239)*x[0]**o+(-0.0691443536291)*x[0]+(0.205492039071)*x[1]**o+(-0.391580258314)*x[1]+(0.791748125642)*x[2]**o+(-0.0896722225431)*x[2]
        arg[0,3,1,1]=(-0.177554766913)*x[0]**o+(-0.399389195454)*x[0]+(-0.304228953477)*x[1]**o+(0.00462595697167)*x[1]+(0.271814579884)*x[2]**o+(-0.17305639368)*x[2]
        arg[0,3,2,0]=(-0.105869589157)*x[0]**o+(0.803488342924)*x[0]+(-0.65117861655)*x[1]**o+(-0.27529059731)*x[1]+(-0.777107387582)*x[2]**o+(-0.690980328369)*x[2]
        arg[0,3,2,1]=(-0.13792230295)*x[0]**o+(0.649721420509)*x[0]+(0.401441756822)*x[1]**o+(0.468949007973)*x[1]+(-0.762015699412)*x[2]**o+(0.389781367434)*x[2]
        arg[0,4,0,0]=(0.873653819021)*x[0]**o+(0.539406798913)*x[0]+(0.802177801924)*x[1]**o+(-0.861314989922)*x[1]+(0.847509960118)*x[2]**o+(-0.41381965291)*x[2]
        arg[0,4,0,1]=(0.0416858647633)*x[0]**o+(0.866825467268)*x[0]+(0.89776163679)*x[1]**o+(-0.978326218858)*x[1]+(-0.409795413277)*x[2]**o+(0.982449327977)*x[2]
        arg[0,4,1,0]=(-0.114530966075)*x[0]**o+(-0.182414280034)*x[0]+(0.154385814218)*x[1]**o+(0.869624332271)*x[1]+(0.477213291128)*x[2]**o+(-0.137123375612)*x[2]
        arg[0,4,1,1]=(-0.52963019184)*x[0]**o+(0.0217955878025)*x[0]+(-0.174531798693)*x[1]**o+(-0.738947178141)*x[1]+(0.198130667502)*x[2]**o+(-0.449952166994)*x[2]
        arg[0,4,2,0]=(0.998209708224)*x[0]**o+(0.734906841374)*x[0]+(0.225596363429)*x[1]**o+(0.862772101878)*x[1]+(-0.133563132638)*x[2]**o+(-0.78059821331)*x[2]
        arg[0,4,2,1]=(0.384580893901)*x[0]**o+(-0.881918368669)*x[0]+(-0.950768236608)*x[1]**o+(-0.452275072093)*x[1]+(0.592779023744)*x[2]**o+(0.798735695084)*x[2]
        arg[1,0,0,0]=(0.786827894179)*x[0]**o+(-0.233241290567)*x[0]+(0.946717698992)*x[1]**o+(0.179368479607)*x[1]+(-0.10409968587)*x[2]**o+(-0.770115022756)*x[2]
        arg[1,0,0,1]=(-0.355389699793)*x[0]**o+(0.356407960494)*x[0]+(-0.371831767713)*x[1]**o+(0.139120259823)*x[1]+(0.910761795881)*x[2]**o+(0.725966050775)*x[2]
        arg[1,0,1,0]=(0.316372320948)*x[0]**o+(0.432472869369)*x[0]+(0.733187356771)*x[1]**o+(0.133915829584)*x[1]+(0.380713482112)*x[2]**o+(-0.874827460114)*x[2]
        arg[1,0,1,1]=(-0.0625786527083)*x[0]**o+(-0.469518973019)*x[0]+(-0.318610570327)*x[1]**o+(-0.304412247656)*x[1]+(0.844057224107)*x[2]**o+(0.508926494536)*x[2]
        arg[1,0,2,0]=(-0.759697281694)*x[0]**o+(0.701456398418)*x[0]+(-0.671612899124)*x[1]**o+(0.292474197717)*x[1]+(-0.445886723952)*x[2]**o+(-0.319953406782)*x[2]
        arg[1,0,2,1]=(-0.865703619781)*x[0]**o+(0.115639820869)*x[0]+(0.493634056223)*x[1]**o+(0.0343245670418)*x[1]+(0.748200872819)*x[2]**o+(-0.208438395054)*x[2]
        arg[1,1,0,0]=(0.767661930656)*x[0]**o+(-0.578385615919)*x[0]+(0.241313268557)*x[1]**o+(0.74276605359)*x[1]+(0.373638986644)*x[2]**o+(0.475479466784)*x[2]
        arg[1,1,0,1]=(0.723228882826)*x[0]**o+(-0.988524686322)*x[0]+(0.333664096597)*x[1]**o+(-0.756435323592)*x[1]+(0.771173510221)*x[2]**o+(-0.879337556561)*x[2]
        arg[1,1,1,0]=(0.939682105701)*x[0]**o+(-0.172311578817)*x[0]+(-0.407241492831)*x[1]**o+(0.446660064834)*x[1]+(0.0371203367053)*x[2]**o+(-0.302853430561)*x[2]
        arg[1,1,1,1]=(0.1093213068)*x[0]**o+(0.31232933407)*x[0]+(0.47474581408)*x[1]**o+(-0.602763201317)*x[1]+(0.803198523872)*x[2]**o+(-0.178088233667)*x[2]
        arg[1,1,2,0]=(-0.217141686114)*x[0]**o+(0.287363815362)*x[0]+(-0.352552947551)*x[1]**o+(0.106445580983)*x[1]+(0.340814562712)*x[2]**o+(-0.963691249733)*x[2]
        arg[1,1,2,1]=(0.612302241126)*x[0]**o+(-0.771765095612)*x[0]+(-0.18662763339)*x[1]**o+(0.10259882275)*x[1]+(-0.29535785913)*x[2]**o+(-0.577876877465)*x[2]
        arg[1,2,0,0]=(0.497131459554)*x[0]**o+(0.912133134267)*x[0]+(0.944369633113)*x[1]**o+(0.163408823815)*x[1]+(0.85636766551)*x[2]**o+(-0.00282458078743)*x[2]
        arg[1,2,0,1]=(0.362490806338)*x[0]**o+(0.22490192994)*x[0]+(-0.913665309243)*x[1]**o+(-0.88010395132)*x[1]+(-0.926205930269)*x[2]**o+(-0.231830082111)*x[2]
        arg[1,2,1,0]=(-0.974792784505)*x[0]**o+(-0.765605618923)*x[0]+(0.241470054253)*x[1]**o+(0.445863114014)*x[1]+(0.561062749944)*x[2]**o+(-0.215674406206)*x[2]
        arg[1,2,1,1]=(0.827013456823)*x[0]**o+(-0.191864800977)*x[0]+(-0.158099224794)*x[1]**o+(-0.921959562711)*x[1]+(0.930693535629)*x[2]**o+(-0.352967183792)*x[2]
        arg[1,2,2,0]=(0.258970603305)*x[0]**o+(0.586779837275)*x[0]+(0.82707087137)*x[1]**o+(-0.215210535461)*x[1]+(0.36890299705)*x[2]**o+(-0.724385890088)*x[2]
        arg[1,2,2,1]=(-0.125969269863)*x[0]**o+(-0.505216539273)*x[0]+(0.43242008053)*x[1]**o+(0.212079268112)*x[1]+(0.345759190477)*x[2]**o+(-0.98505468199)*x[2]
        arg[1,3,0,0]=(-0.647118215098)*x[0]**o+(0.318566502446)*x[0]+(-0.151406735047)*x[1]**o+(0.253202745125)*x[1]+(-0.404823111329)*x[2]**o+(-0.520797366557)*x[2]
        arg[1,3,0,1]=(-0.946886311645)*x[0]**o+(0.733699524424)*x[0]+(-0.615470980807)*x[1]**o+(-0.103976314414)*x[1]+(-0.932104431883)*x[2]**o+(0.157563806559)*x[2]
        arg[1,3,1,0]=(-0.289471503181)*x[0]**o+(0.749124853135)*x[0]+(-0.578861403092)*x[1]**o+(-0.136231378985)*x[1]+(0.532719246315)*x[2]**o+(0.542455831657)*x[2]
        arg[1,3,1,1]=(-0.343842490684)*x[0]**o+(0.110039182926)*x[0]+(0.901234650534)*x[1]**o+(-0.513362876241)*x[1]+(0.371952189126)*x[2]**o+(-0.169400349862)*x[2]
        arg[1,3,2,0]=(-0.766791944503)*x[0]**o+(0.0267491122166)*x[0]+(-0.699293509597)*x[1]**o+(-0.697363669971)*x[1]+(0.567854170308)*x[2]**o+(-0.516301389629)*x[2]
        arg[1,3,2,1]=(-0.288508579358)*x[0]**o+(0.239321880772)*x[0]+(0.677249635795)*x[1]**o+(0.48369084768)*x[1]+(0.681460633595)*x[2]**o+(-0.80218734437)*x[2]
        arg[1,4,0,0]=(0.510273388936)*x[0]**o+(0.35600020965)*x[0]+(0.271077500616)*x[1]**o+(-0.00479335810576)*x[1]+(-0.348457583464)*x[2]**o+(-0.965895122152)*x[2]
        arg[1,4,0,1]=(0.0134532748813)*x[0]**o+(-0.396954380686)*x[0]+(-0.176682002946)*x[1]**o+(0.971224261138)*x[1]+(-0.359602180261)*x[2]**o+(0.373748703363)*x[2]
        arg[1,4,1,0]=(-0.519413160925)*x[0]**o+(0.856346255965)*x[0]+(0.159167243935)*x[1]**o+(0.92068779838)*x[1]+(0.415122182947)*x[2]**o+(-0.595702311496)*x[2]
        arg[1,4,1,1]=(-0.608969598184)*x[0]**o+(-0.434750765676)*x[0]+(0.74471475701)*x[1]**o+(-0.229333246615)*x[1]+(-0.678138032736)*x[2]**o+(-0.514112926798)*x[2]
        arg[1,4,2,0]=(0.0633453561735)*x[0]**o+(-0.278569473082)*x[0]+(-0.842227144473)*x[1]**o+(-0.00289799815652)*x[1]+(0.153392291113)*x[2]**o+(-0.61448940993)*x[2]
        arg[1,4,2,1]=(0.0897863401674)*x[0]**o+(-0.487876787706)*x[0]+(-0.464008634014)*x[1]**o+(0.576466896344)*x[1]+(0.962345361537)*x[2]**o+(0.580878819156)*x[2]
        arg[2,0,0,0]=(0.316582275434)*x[0]**o+(0.0607872538221)*x[0]+(0.642492366848)*x[1]**o+(0.200300167843)*x[1]+(-0.402802396184)*x[2]**o+(-0.834254866233)*x[2]
        arg[2,0,0,1]=(0.644614184272)*x[0]**o+(-0.086367798694)*x[0]+(-0.428671462705)*x[1]**o+(-0.517827373575)*x[1]+(0.528197945165)*x[2]**o+(-0.523092888005)*x[2]
        arg[2,0,1,0]=(0.41376145869)*x[0]**o+(-0.27425322418)*x[0]+(0.816942168928)*x[1]**o+(-0.882784454814)*x[1]+(-0.270205535673)*x[2]**o+(0.642372102899)*x[2]
        arg[2,0,1,1]=(0.584931736844)*x[0]**o+(0.622511792702)*x[0]+(0.76174625736)*x[1]**o+(0.0561486448557)*x[1]+(-0.588861269552)*x[2]**o+(-0.341896323569)*x[2]
        arg[2,0,2,0]=(0.635215108566)*x[0]**o+(0.299769133671)*x[0]+(0.141053572388)*x[1]**o+(0.102844255879)*x[1]+(-0.995154777534)*x[2]**o+(0.501319388994)*x[2]
        arg[2,0,2,1]=(0.750089643836)*x[0]**o+(0.57570708561)*x[0]+(-0.449935823319)*x[1]**o+(-0.902982109411)*x[1]+(0.561185740222)*x[2]**o+(0.422654056847)*x[2]
        arg[2,1,0,0]=(0.953986415316)*x[0]**o+(-0.7883326473)*x[0]+(-0.723690234059)*x[1]**o+(-0.537242781471)*x[1]+(-0.503466717519)*x[2]**o+(-0.777865169241)*x[2]
        arg[2,1,0,1]=(0.317357554441)*x[0]**o+(-0.326828957855)*x[0]+(-0.609992651237)*x[1]**o+(0.811066934743)*x[1]+(-0.90668441447)*x[2]**o+(0.92581926687)*x[2]
        arg[2,1,1,0]=(-0.0801236189368)*x[0]**o+(0.49151264813)*x[0]+(-0.0770654446151)*x[1]**o+(-0.314981217276)*x[1]+(-0.743180837478)*x[2]**o+(-0.554491413146)*x[2]
        arg[2,1,1,1]=(0.831017344937)*x[0]**o+(-0.61064793935)*x[0]+(0.652976737027)*x[1]**o+(0.242104127969)*x[1]+(0.593330958678)*x[2]**o+(-0.35899584298)*x[2]
        arg[2,1,2,0]=(-0.580618798556)*x[0]**o+(-0.936176474082)*x[0]+(-0.768716477288)*x[1]**o+(-0.353687473159)*x[1]+(-0.261509314019)*x[2]**o+(-0.442139199257)*x[2]
        arg[2,1,2,1]=(0.945760793889)*x[0]**o+(-0.646958670891)*x[0]+(-0.440341726063)*x[1]**o+(0.223283751014)*x[1]+(0.799798944743)*x[2]**o+(0.591902067633)*x[2]
        arg[2,2,0,0]=(-0.267699746405)*x[0]**o+(0.271554172399)*x[0]+(0.747278903642)*x[1]**o+(-0.832877459834)*x[1]+(-0.640373164716)*x[2]**o+(-0.666487164763)*x[2]
        arg[2,2,0,1]=(0.795630715924)*x[0]**o+(0.682973866186)*x[0]+(-0.750818874926)*x[1]**o+(-0.904782408217)*x[1]+(-0.578138718379)*x[2]**o+(-0.369846516992)*x[2]
        arg[2,2,1,0]=(0.165654497062)*x[0]**o+(-0.329827871165)*x[0]+(-0.808981530672)*x[1]**o+(0.350106763785)*x[1]+(0.106638087689)*x[2]**o+(0.425516820398)*x[2]
        arg[2,2,1,1]=(0.017419119421)*x[0]**o+(0.500359047996)*x[0]+(-0.888509940049)*x[1]**o+(-0.375469124891)*x[1]+(0.863718970618)*x[2]**o+(-0.807667604763)*x[2]
        arg[2,2,2,0]=(0.860662652165)*x[0]**o+(0.162530730499)*x[0]+(-0.657510966961)*x[1]**o+(0.215905840286)*x[1]+(0.97380371264)*x[2]**o+(-0.431027671897)*x[2]
        arg[2,2,2,1]=(-0.649151503991)*x[0]**o+(0.484589779517)*x[0]+(0.905080495553)*x[1]**o+(0.997504119204)*x[1]+(-0.4268802512)*x[2]**o+(-0.687236308612)*x[2]
        arg[2,3,0,0]=(-0.60485902589)*x[0]**o+(0.49500777286)*x[0]+(0.33977532104)*x[1]**o+(-0.219209845338)*x[1]+(-0.197311627683)*x[2]**o+(0.285859591044)*x[2]
        arg[2,3,0,1]=(0.117438340483)*x[0]**o+(-0.562954894877)*x[0]+(-0.302982842797)*x[1]**o+(0.977923453436)*x[1]+(0.300127749784)*x[2]**o+(0.45784201373)*x[2]
        arg[2,3,1,0]=(-0.910061442266)*x[0]**o+(0.723114553174)*x[0]+(-0.435693478652)*x[1]**o+(-0.584326584996)*x[1]+(-0.51000911177)*x[2]**o+(0.506615483889)*x[2]
        arg[2,3,1,1]=(-0.50179741763)*x[0]**o+(-0.954551290001)*x[0]+(0.375839266192)*x[1]**o+(0.122120198059)*x[1]+(-0.593050674641)*x[2]**o+(0.549989594312)*x[2]
        arg[2,3,2,0]=(-0.889519666705)*x[0]**o+(-0.815572507674)*x[0]+(0.453677651057)*x[1]**o+(0.450910828217)*x[1]+(-0.00465373008682)*x[2]**o+(0.731899900165)*x[2]
        arg[2,3,2,1]=(-0.969971151589)*x[0]**o+(-0.156948104522)*x[0]+(0.492885029353)*x[1]**o+(-0.0310677847833)*x[1]+(0.327614233837)*x[2]**o+(-0.274974040707)*x[2]
        arg[2,4,0,0]=(0.844667900826)*x[0]**o+(0.863374870522)*x[0]+(0.251456923241)*x[1]**o+(0.434916005183)*x[1]+(-0.139023370677)*x[2]**o+(0.0655176616053)*x[2]
        arg[2,4,0,1]=(0.831755059318)*x[0]**o+(0.0897642674136)*x[0]+(0.190414878812)*x[1]**o+(0.976565327031)*x[1]+(0.413321767371)*x[2]**o+(0.755395415341)*x[2]
        arg[2,4,1,0]=(-0.790364876377)*x[0]**o+(0.829442094248)*x[0]+(0.556251403891)*x[1]**o+(-0.953487711658)*x[1]+(-0.610781905401)*x[2]**o+(-0.320764072326)*x[2]
        arg[2,4,1,1]=(-0.757101565136)*x[0]**o+(0.789340839395)*x[0]+(-0.615189514438)*x[1]**o+(-0.910992676978)*x[1]+(0.693736813401)*x[2]**o+(0.0538707588266)*x[2]
        arg[2,4,2,0]=(0.000335625616611)*x[0]**o+(-0.195171446476)*x[0]+(0.0281149133705)*x[1]**o+(-0.362294159243)*x[1]+(0.830520288743)*x[2]**o+(-0.431967694247)*x[2]
        arg[2,4,2,1]=(-0.232668364843)*x[0]**o+(0.655302452302)*x[0]+(-0.0125806408337)*x[1]**o+(-0.103979202109)*x[1]+(-0.277450383033)*x[2]**o+(0.708914971885)*x[2]
        arg[3,0,0,0]=(-0.813447875186)*x[0]**o+(0.133491078101)*x[0]+(-0.675510725453)*x[1]**o+(-0.161489202717)*x[1]+(0.0474778506607)*x[2]**o+(0.513729029652)*x[2]
        arg[3,0,0,1]=(0.45388386743)*x[0]**o+(-0.909074944715)*x[0]+(-0.540576382376)*x[1]**o+(-0.849753706068)*x[1]+(-0.496320191697)*x[2]**o+(-0.930725196363)*x[2]
        arg[3,0,1,0]=(-0.145526830518)*x[0]**o+(-0.34717641097)*x[0]+(-0.818886792022)*x[1]**o+(0.574315644784)*x[1]+(-0.780431329222)*x[2]**o+(0.892840747651)*x[2]
        arg[3,0,1,1]=(-0.415342962745)*x[0]**o+(-0.899768945518)*x[0]+(0.798573910731)*x[1]**o+(-0.0652893933479)*x[1]+(-0.444459519602)*x[2]**o+(-0.0100131702565)*x[2]
        arg[3,0,2,0]=(-0.997936676013)*x[0]**o+(0.822000858069)*x[0]+(-0.553794408949)*x[1]**o+(0.829037287423)*x[1]+(-0.700288437806)*x[2]**o+(-0.449584600891)*x[2]
        arg[3,0,2,1]=(-0.70979868598)*x[0]**o+(0.689868687655)*x[0]+(0.151257146397)*x[1]**o+(0.32895215144)*x[1]+(0.286043871383)*x[2]**o+(-0.604421567781)*x[2]
        arg[3,1,0,0]=(0.60431849542)*x[0]**o+(0.518618291627)*x[0]+(-0.471868097013)*x[1]**o+(-0.639110899151)*x[1]+(-0.0904184061601)*x[2]**o+(0.113762528304)*x[2]
        arg[3,1,0,1]=(0.307624103157)*x[0]**o+(-0.388281302162)*x[0]+(0.772429578102)*x[1]**o+(-0.197799629585)*x[1]+(-0.149120667036)*x[2]**o+(0.830154780264)*x[2]
        arg[3,1,1,0]=(0.957127413133)*x[0]**o+(0.730080945389)*x[0]+(0.497191948774)*x[1]**o+(-0.586439221663)*x[1]+(-0.56223071093)*x[2]**o+(0.162899395804)*x[2]
        arg[3,1,1,1]=(-0.285740818451)*x[0]**o+(-0.573919970411)*x[0]+(-0.400833211669)*x[1]**o+(0.84414534634)*x[1]+(-0.124352096731)*x[2]**o+(0.907013800834)*x[2]
        arg[3,1,2,0]=(0.186947347989)*x[0]**o+(0.499582449213)*x[0]+(0.525473636695)*x[1]**o+(-0.926500953998)*x[1]+(-0.0392615366415)*x[2]**o+(0.909002911192)*x[2]
        arg[3,1,2,1]=(0.634350440683)*x[0]**o+(0.901699291665)*x[0]+(-0.683899667575)*x[1]**o+(0.411642037744)*x[1]+(-0.72791611585)*x[2]**o+(-0.013293689839)*x[2]
        arg[3,2,0,0]=(0.185164429507)*x[0]**o+(-0.671011652257)*x[0]+(0.973361040829)*x[1]**o+(0.989738565899)*x[1]+(-0.501245887091)*x[2]**o+(-0.405445444386)*x[2]
        arg[3,2,0,1]=(-0.311554020884)*x[0]**o+(-0.929020990333)*x[0]+(0.366016375797)*x[1]**o+(0.838234804773)*x[1]+(0.939602632409)*x[2]**o+(-0.885524260372)*x[2]
        arg[3,2,1,0]=(-0.0101604655589)*x[0]**o+(0.670274078493)*x[0]+(0.198135855354)*x[1]**o+(-0.000275422963644)*x[1]+(0.136674233404)*x[2]**o+(-0.442510118497)*x[2]
        arg[3,2,1,1]=(-0.677948769785)*x[0]**o+(-0.165403610813)*x[0]+(-0.219353502981)*x[1]**o+(0.664297154009)*x[1]+(0.158536302892)*x[2]**o+(0.293940884925)*x[2]
        arg[3,2,2,0]=(0.666669534708)*x[0]**o+(0.609459298502)*x[0]+(0.297058070582)*x[1]**o+(-0.0809555825821)*x[1]+(0.139430791042)*x[2]**o+(-0.253514861624)*x[2]
        arg[3,2,2,1]=(-0.577932413274)*x[0]**o+(-0.583353108895)*x[0]+(-0.839262034394)*x[1]**o+(0.795642236106)*x[1]+(0.0882627381543)*x[2]**o+(-0.382970623122)*x[2]
        arg[3,3,0,0]=(0.187941627429)*x[0]**o+(0.552040688531)*x[0]+(0.837458832716)*x[1]**o+(0.296793121897)*x[1]+(-0.665699511216)*x[2]**o+(-0.318565991052)*x[2]
        arg[3,3,0,1]=(-0.78770176804)*x[0]**o+(0.807509260396)*x[0]+(0.46241086085)*x[1]**o+(0.668506908751)*x[1]+(-0.988954457343)*x[2]**o+(0.0896365679915)*x[2]
        arg[3,3,1,0]=(0.727665746433)*x[0]**o+(-0.819391676739)*x[0]+(-0.118648934607)*x[1]**o+(-0.433862939031)*x[1]+(0.827788711313)*x[2]**o+(-0.615751446784)*x[2]
        arg[3,3,1,1]=(-0.00427901841482)*x[0]**o+(-0.38545713036)*x[0]+(0.570160083404)*x[1]**o+(-0.235505162795)*x[1]+(0.514578304371)*x[2]**o+(-0.152692113279)*x[2]
        arg[3,3,2,0]=(-0.396900613831)*x[0]**o+(0.161875314543)*x[0]+(0.28471295984)*x[1]**o+(-0.623276596728)*x[1]+(-0.748772373219)*x[2]**o+(0.584246898704)*x[2]
        arg[3,3,2,1]=(0.210026027009)*x[0]**o+(-0.530437530387)*x[0]+(0.437155825716)*x[1]**o+(0.984102407204)*x[1]+(0.125030773605)*x[2]**o+(-0.821764907318)*x[2]
        arg[3,4,0,0]=(-0.819129946783)*x[0]**o+(-0.729517565451)*x[0]+(-0.126013244602)*x[1]**o+(0.592832125254)*x[1]+(-0.850426487488)*x[2]**o+(0.271550912174)*x[2]
        arg[3,4,0,1]=(-0.573554440921)*x[0]**o+(0.232106829584)*x[0]+(0.810201614042)*x[1]**o+(-0.85518164211)*x[1]+(0.249535563949)*x[2]**o+(-0.69728963017)*x[2]
        arg[3,4,1,0]=(0.785661643628)*x[0]**o+(0.936424600424)*x[0]+(0.321363811055)*x[1]**o+(0.362341458592)*x[1]+(-0.593576579363)*x[2]**o+(0.0601278435924)*x[2]
        arg[3,4,1,1]=(-0.854420686885)*x[0]**o+(0.993352097501)*x[0]+(-0.572116112224)*x[1]**o+(0.308008845182)*x[1]+(0.0443339578639)*x[2]**o+(-0.1469857285)*x[2]
        arg[3,4,2,0]=(0.696373642151)*x[0]**o+(0.82425126832)*x[0]+(-0.0891735358455)*x[1]**o+(-0.22218993613)*x[1]+(-0.238393149017)*x[2]**o+(-0.631352718194)*x[2]
        arg[3,4,2,1]=(-0.0932690431928)*x[0]**o+(-0.391900483919)*x[0]+(0.236843565361)*x[1]**o+(-0.164197648532)*x[1]+(0.235502596353)*x[2]**o+(0.0110131867917)*x[2]
        ref[0,0,0,0]=(0.349437323395)*x_ref[0]**o+(0.817821891642)*x_ref[0]+(-0.357998520848)*x_ref[1]**o+(0.349311985429)*x_ref[1]+(0.470351907942)*x_ref[2]**o+(0.264007843212)*x_ref[2]
        ref[0,0,0,1]=(-0.747824148035)*x_ref[0]**o+(0.551817215468)*x_ref[0]+(0.980476067966)*x_ref[1]**o+(0.521120126285)*x_ref[1]+(-0.472357720779)*x_ref[2]**o+(-0.379502066444)*x_ref[2]
        ref[0,0,1,0]=(0.302984899868)*x_ref[0]**o+(0.556601505796)*x_ref[0]+(-0.698948476338)*x_ref[1]**o+(-0.587379572024)*x_ref[1]+(-0.625760359909)*x_ref[2]**o+(0.231956543432)*x_ref[2]
        ref[0,0,1,1]=(-0.335943560463)*x_ref[0]**o+(-0.497643152155)*x_ref[0]+(-0.849829024888)*x_ref[1]**o+(-0.388751135318)*x_ref[1]+(-0.517696637743)*x_ref[2]**o+(-0.966210901)*x_ref[2]
        ref[0,0,2,0]=(-0.0953058405321)*x_ref[0]**o+(-0.864580918175)*x_ref[0]+(-0.732086619685)*x_ref[1]**o+(0.872044654816)*x_ref[1]+(-0.225850391375)*x_ref[2]**o+(-0.502203654371)*x_ref[2]
        ref[0,0,2,1]=(0.228408069101)*x_ref[0]**o+(0.267590362458)*x_ref[0]+(0.127397041031)*x_ref[1]**o+(0.0142393720782)*x_ref[1]+(-0.257563412478)*x_ref[2]**o+(0.0211877877507)*x_ref[2]
        ref[0,1,0,0]=(0.921125039994)*x_ref[0]**o+(0.372548304521)*x_ref[0]+(-0.66574364308)*x_ref[1]**o+(0.0895600219155)*x_ref[1]+(-0.656870562037)*x_ref[2]**o+(-0.608819638268)*x_ref[2]
        ref[0,1,0,1]=(-0.396625252579)*x_ref[0]**o+(0.655542320616)*x_ref[0]+(-0.542629502059)*x_ref[1]**o+(0.587317299384)*x_ref[1]+(0.96380848967)*x_ref[2]**o+(0.121341827766)*x_ref[2]
        ref[0,1,1,0]=(0.0908208346003)*x_ref[0]**o+(-0.0398227321951)*x_ref[0]+(0.907607555288)*x_ref[1]**o+(0.527230589013)*x_ref[1]+(0.519287416653)*x_ref[2]**o+(0.103180358249)*x_ref[2]
        ref[0,1,1,1]=(0.0615195625837)*x_ref[0]**o+(0.695720276474)*x_ref[0]+(0.0862101684049)*x_ref[1]**o+(0.517150419955)*x_ref[1]+(0.511207669575)*x_ref[2]**o+(-0.9556472362)*x_ref[2]
        ref[0,1,2,0]=(0.70040395652)*x_ref[0]**o+(-0.576266027865)*x_ref[0]+(-0.133947982034)*x_ref[1]**o+(-0.764976368456)*x_ref[1]+(-0.860780506782)*x_ref[2]**o+(-0.137946340663)*x_ref[2]
        ref[0,1,2,1]=(0.356017596832)*x_ref[0]**o+(0.94912783604)*x_ref[0]+(0.325951760294)*x_ref[1]**o+(0.0623695610721)*x_ref[1]+(-0.106031051895)*x_ref[2]**o+(-0.339314970581)*x_ref[2]
        ref[0,2,0,0]=(-0.124954395016)*x_ref[0]**o+(0.558233271782)*x_ref[0]+(-0.956846820394)*x_ref[1]**o+(-0.599332648704)*x_ref[1]+(-0.322444201477)*x_ref[2]**o+(0.0833108732532)*x_ref[2]
        ref[0,2,0,1]=(-0.713065259589)*x_ref[0]**o+(-0.597746726477)*x_ref[0]+(-0.976654446591)*x_ref[1]**o+(0.977779698697)*x_ref[1]+(0.694427679422)*x_ref[2]**o+(-0.364463986712)*x_ref[2]
        ref[0,2,1,0]=(-0.415861468352)*x_ref[0]**o+(-0.159802964908)*x_ref[0]+(-0.132331280778)*x_ref[1]**o+(0.120829594848)*x_ref[1]+(-0.114181951611)*x_ref[2]**o+(-0.810187313564)*x_ref[2]
        ref[0,2,1,1]=(-0.869925695451)*x_ref[0]**o+(-0.535444611026)*x_ref[0]+(0.62908837594)*x_ref[1]**o+(-0.488583607314)*x_ref[1]+(0.8461615049)*x_ref[2]**o+(-0.558010294727)*x_ref[2]
        ref[0,2,2,0]=(-0.0822232217466)*x_ref[0]**o+(-0.673631382141)*x_ref[0]+(0.803717792107)*x_ref[1]**o+(0.339934715615)*x_ref[1]+(-0.886365924996)*x_ref[2]**o+(0.0206517750753)*x_ref[2]
        ref[0,2,2,1]=(0.367656249869)*x_ref[0]**o+(0.000987963028539)*x_ref[0]+(0.399182215451)*x_ref[1]**o+(0.923259635138)*x_ref[1]+(0.8573338122)*x_ref[2]**o+(0.679082213415)*x_ref[2]
        ref[0,3,0,0]=(-0.722680496644)*x_ref[0]**o+(0.199427912651)*x_ref[0]+(-0.351799245099)*x_ref[1]**o+(0.240285416231)*x_ref[1]+(-0.701704868242)*x_ref[2]**o+(0.265610436018)*x_ref[2]
        ref[0,3,0,1]=(-0.14479043328)*x_ref[0]**o+(-0.192670959504)*x_ref[0]+(0.327412080296)*x_ref[1]**o+(-0.390894637576)*x_ref[1]+(-0.401862813677)*x_ref[2]**o+(-0.87150071561)*x_ref[2]
        ref[0,3,1,0]=(0.430039400239)*x_ref[0]**o+(-0.0691443536291)*x_ref[0]+(0.205492039071)*x_ref[1]**o+(-0.391580258314)*x_ref[1]+(0.791748125642)*x_ref[2]**o+(-0.0896722225431)*x_ref[2]
        ref[0,3,1,1]=(-0.177554766913)*x_ref[0]**o+(-0.399389195454)*x_ref[0]+(-0.304228953477)*x_ref[1]**o+(0.00462595697167)*x_ref[1]+(0.271814579884)*x_ref[2]**o+(-0.17305639368)*x_ref[2]
        ref[0,3,2,0]=(-0.105869589157)*x_ref[0]**o+(0.803488342924)*x_ref[0]+(-0.65117861655)*x_ref[1]**o+(-0.27529059731)*x_ref[1]+(-0.777107387582)*x_ref[2]**o+(-0.690980328369)*x_ref[2]
        ref[0,3,2,1]=(-0.13792230295)*x_ref[0]**o+(0.649721420509)*x_ref[0]+(0.401441756822)*x_ref[1]**o+(0.468949007973)*x_ref[1]+(-0.762015699412)*x_ref[2]**o+(0.389781367434)*x_ref[2]
        ref[0,4,0,0]=(0.873653819021)*x_ref[0]**o+(0.539406798913)*x_ref[0]+(0.802177801924)*x_ref[1]**o+(-0.861314989922)*x_ref[1]+(0.847509960118)*x_ref[2]**o+(-0.41381965291)*x_ref[2]
        ref[0,4,0,1]=(0.0416858647633)*x_ref[0]**o+(0.866825467268)*x_ref[0]+(0.89776163679)*x_ref[1]**o+(-0.978326218858)*x_ref[1]+(-0.409795413277)*x_ref[2]**o+(0.982449327977)*x_ref[2]
        ref[0,4,1,0]=(-0.114530966075)*x_ref[0]**o+(-0.182414280034)*x_ref[0]+(0.154385814218)*x_ref[1]**o+(0.869624332271)*x_ref[1]+(0.477213291128)*x_ref[2]**o+(-0.137123375612)*x_ref[2]
        ref[0,4,1,1]=(-0.52963019184)*x_ref[0]**o+(0.0217955878025)*x_ref[0]+(-0.174531798693)*x_ref[1]**o+(-0.738947178141)*x_ref[1]+(0.198130667502)*x_ref[2]**o+(-0.449952166994)*x_ref[2]
        ref[0,4,2,0]=(0.998209708224)*x_ref[0]**o+(0.734906841374)*x_ref[0]+(0.225596363429)*x_ref[1]**o+(0.862772101878)*x_ref[1]+(-0.133563132638)*x_ref[2]**o+(-0.78059821331)*x_ref[2]
        ref[0,4,2,1]=(0.384580893901)*x_ref[0]**o+(-0.881918368669)*x_ref[0]+(-0.950768236608)*x_ref[1]**o+(-0.452275072093)*x_ref[1]+(0.592779023744)*x_ref[2]**o+(0.798735695084)*x_ref[2]
        ref[1,0,0,0]=(0.786827894179)*x_ref[0]**o+(-0.233241290567)*x_ref[0]+(0.946717698992)*x_ref[1]**o+(0.179368479607)*x_ref[1]+(-0.10409968587)*x_ref[2]**o+(-0.770115022756)*x_ref[2]
        ref[1,0,0,1]=(-0.355389699793)*x_ref[0]**o+(0.356407960494)*x_ref[0]+(-0.371831767713)*x_ref[1]**o+(0.139120259823)*x_ref[1]+(0.910761795881)*x_ref[2]**o+(0.725966050775)*x_ref[2]
        ref[1,0,1,0]=(0.316372320948)*x_ref[0]**o+(0.432472869369)*x_ref[0]+(0.733187356771)*x_ref[1]**o+(0.133915829584)*x_ref[1]+(0.380713482112)*x_ref[2]**o+(-0.874827460114)*x_ref[2]
        ref[1,0,1,1]=(-0.0625786527083)*x_ref[0]**o+(-0.469518973019)*x_ref[0]+(-0.318610570327)*x_ref[1]**o+(-0.304412247656)*x_ref[1]+(0.844057224107)*x_ref[2]**o+(0.508926494536)*x_ref[2]
        ref[1,0,2,0]=(-0.759697281694)*x_ref[0]**o+(0.701456398418)*x_ref[0]+(-0.671612899124)*x_ref[1]**o+(0.292474197717)*x_ref[1]+(-0.445886723952)*x_ref[2]**o+(-0.319953406782)*x_ref[2]
        ref[1,0,2,1]=(-0.865703619781)*x_ref[0]**o+(0.115639820869)*x_ref[0]+(0.493634056223)*x_ref[1]**o+(0.0343245670418)*x_ref[1]+(0.748200872819)*x_ref[2]**o+(-0.208438395054)*x_ref[2]
        ref[1,1,0,0]=(0.767661930656)*x_ref[0]**o+(-0.578385615919)*x_ref[0]+(0.241313268557)*x_ref[1]**o+(0.74276605359)*x_ref[1]+(0.373638986644)*x_ref[2]**o+(0.475479466784)*x_ref[2]
        ref[1,1,0,1]=(0.723228882826)*x_ref[0]**o+(-0.988524686322)*x_ref[0]+(0.333664096597)*x_ref[1]**o+(-0.756435323592)*x_ref[1]+(0.771173510221)*x_ref[2]**o+(-0.879337556561)*x_ref[2]
        ref[1,1,1,0]=(0.939682105701)*x_ref[0]**o+(-0.172311578817)*x_ref[0]+(-0.407241492831)*x_ref[1]**o+(0.446660064834)*x_ref[1]+(0.0371203367053)*x_ref[2]**o+(-0.302853430561)*x_ref[2]
        ref[1,1,1,1]=(0.1093213068)*x_ref[0]**o+(0.31232933407)*x_ref[0]+(0.47474581408)*x_ref[1]**o+(-0.602763201317)*x_ref[1]+(0.803198523872)*x_ref[2]**o+(-0.178088233667)*x_ref[2]
        ref[1,1,2,0]=(-0.217141686114)*x_ref[0]**o+(0.287363815362)*x_ref[0]+(-0.352552947551)*x_ref[1]**o+(0.106445580983)*x_ref[1]+(0.340814562712)*x_ref[2]**o+(-0.963691249733)*x_ref[2]
        ref[1,1,2,1]=(0.612302241126)*x_ref[0]**o+(-0.771765095612)*x_ref[0]+(-0.18662763339)*x_ref[1]**o+(0.10259882275)*x_ref[1]+(-0.29535785913)*x_ref[2]**o+(-0.577876877465)*x_ref[2]
        ref[1,2,0,0]=(0.497131459554)*x_ref[0]**o+(0.912133134267)*x_ref[0]+(0.944369633113)*x_ref[1]**o+(0.163408823815)*x_ref[1]+(0.85636766551)*x_ref[2]**o+(-0.00282458078743)*x_ref[2]
        ref[1,2,0,1]=(0.362490806338)*x_ref[0]**o+(0.22490192994)*x_ref[0]+(-0.913665309243)*x_ref[1]**o+(-0.88010395132)*x_ref[1]+(-0.926205930269)*x_ref[2]**o+(-0.231830082111)*x_ref[2]
        ref[1,2,1,0]=(-0.974792784505)*x_ref[0]**o+(-0.765605618923)*x_ref[0]+(0.241470054253)*x_ref[1]**o+(0.445863114014)*x_ref[1]+(0.561062749944)*x_ref[2]**o+(-0.215674406206)*x_ref[2]
        ref[1,2,1,1]=(0.827013456823)*x_ref[0]**o+(-0.191864800977)*x_ref[0]+(-0.158099224794)*x_ref[1]**o+(-0.921959562711)*x_ref[1]+(0.930693535629)*x_ref[2]**o+(-0.352967183792)*x_ref[2]
        ref[1,2,2,0]=(0.258970603305)*x_ref[0]**o+(0.586779837275)*x_ref[0]+(0.82707087137)*x_ref[1]**o+(-0.215210535461)*x_ref[1]+(0.36890299705)*x_ref[2]**o+(-0.724385890088)*x_ref[2]
        ref[1,2,2,1]=(-0.125969269863)*x_ref[0]**o+(-0.505216539273)*x_ref[0]+(0.43242008053)*x_ref[1]**o+(0.212079268112)*x_ref[1]+(0.345759190477)*x_ref[2]**o+(-0.98505468199)*x_ref[2]
        ref[1,3,0,0]=(-0.647118215098)*x_ref[0]**o+(0.318566502446)*x_ref[0]+(-0.151406735047)*x_ref[1]**o+(0.253202745125)*x_ref[1]+(-0.404823111329)*x_ref[2]**o+(-0.520797366557)*x_ref[2]
        ref[1,3,0,1]=(-0.946886311645)*x_ref[0]**o+(0.733699524424)*x_ref[0]+(-0.615470980807)*x_ref[1]**o+(-0.103976314414)*x_ref[1]+(-0.932104431883)*x_ref[2]**o+(0.157563806559)*x_ref[2]
        ref[1,3,1,0]=(-0.289471503181)*x_ref[0]**o+(0.749124853135)*x_ref[0]+(-0.578861403092)*x_ref[1]**o+(-0.136231378985)*x_ref[1]+(0.532719246315)*x_ref[2]**o+(0.542455831657)*x_ref[2]
        ref[1,3,1,1]=(-0.343842490684)*x_ref[0]**o+(0.110039182926)*x_ref[0]+(0.901234650534)*x_ref[1]**o+(-0.513362876241)*x_ref[1]+(0.371952189126)*x_ref[2]**o+(-0.169400349862)*x_ref[2]
        ref[1,3,2,0]=(-0.766791944503)*x_ref[0]**o+(0.0267491122166)*x_ref[0]+(-0.699293509597)*x_ref[1]**o+(-0.697363669971)*x_ref[1]+(0.567854170308)*x_ref[2]**o+(-0.516301389629)*x_ref[2]
        ref[1,3,2,1]=(-0.288508579358)*x_ref[0]**o+(0.239321880772)*x_ref[0]+(0.677249635795)*x_ref[1]**o+(0.48369084768)*x_ref[1]+(0.681460633595)*x_ref[2]**o+(-0.80218734437)*x_ref[2]
        ref[1,4,0,0]=(0.510273388936)*x_ref[0]**o+(0.35600020965)*x_ref[0]+(0.271077500616)*x_ref[1]**o+(-0.00479335810576)*x_ref[1]+(-0.348457583464)*x_ref[2]**o+(-0.965895122152)*x_ref[2]
        ref[1,4,0,1]=(0.0134532748813)*x_ref[0]**o+(-0.396954380686)*x_ref[0]+(-0.176682002946)*x_ref[1]**o+(0.971224261138)*x_ref[1]+(-0.359602180261)*x_ref[2]**o+(0.373748703363)*x_ref[2]
        ref[1,4,1,0]=(-0.519413160925)*x_ref[0]**o+(0.856346255965)*x_ref[0]+(0.159167243935)*x_ref[1]**o+(0.92068779838)*x_ref[1]+(0.415122182947)*x_ref[2]**o+(-0.595702311496)*x_ref[2]
        ref[1,4,1,1]=(-0.608969598184)*x_ref[0]**o+(-0.434750765676)*x_ref[0]+(0.74471475701)*x_ref[1]**o+(-0.229333246615)*x_ref[1]+(-0.678138032736)*x_ref[2]**o+(-0.514112926798)*x_ref[2]
        ref[1,4,2,0]=(0.0633453561735)*x_ref[0]**o+(-0.278569473082)*x_ref[0]+(-0.842227144473)*x_ref[1]**o+(-0.00289799815652)*x_ref[1]+(0.153392291113)*x_ref[2]**o+(-0.61448940993)*x_ref[2]
        ref[1,4,2,1]=(0.0897863401674)*x_ref[0]**o+(-0.487876787706)*x_ref[0]+(-0.464008634014)*x_ref[1]**o+(0.576466896344)*x_ref[1]+(0.962345361537)*x_ref[2]**o+(0.580878819156)*x_ref[2]
        ref[2,0,0,0]=(0.316582275434)*x_ref[0]**o+(0.0607872538221)*x_ref[0]+(0.642492366848)*x_ref[1]**o+(0.200300167843)*x_ref[1]+(-0.402802396184)*x_ref[2]**o+(-0.834254866233)*x_ref[2]
        ref[2,0,0,1]=(0.644614184272)*x_ref[0]**o+(-0.086367798694)*x_ref[0]+(-0.428671462705)*x_ref[1]**o+(-0.517827373575)*x_ref[1]+(0.528197945165)*x_ref[2]**o+(-0.523092888005)*x_ref[2]
        ref[2,0,1,0]=(0.41376145869)*x_ref[0]**o+(-0.27425322418)*x_ref[0]+(0.816942168928)*x_ref[1]**o+(-0.882784454814)*x_ref[1]+(-0.270205535673)*x_ref[2]**o+(0.642372102899)*x_ref[2]
        ref[2,0,1,1]=(0.584931736844)*x_ref[0]**o+(0.622511792702)*x_ref[0]+(0.76174625736)*x_ref[1]**o+(0.0561486448557)*x_ref[1]+(-0.588861269552)*x_ref[2]**o+(-0.341896323569)*x_ref[2]
        ref[2,0,2,0]=(0.635215108566)*x_ref[0]**o+(0.299769133671)*x_ref[0]+(0.141053572388)*x_ref[1]**o+(0.102844255879)*x_ref[1]+(-0.995154777534)*x_ref[2]**o+(0.501319388994)*x_ref[2]
        ref[2,0,2,1]=(0.750089643836)*x_ref[0]**o+(0.57570708561)*x_ref[0]+(-0.449935823319)*x_ref[1]**o+(-0.902982109411)*x_ref[1]+(0.561185740222)*x_ref[2]**o+(0.422654056847)*x_ref[2]
        ref[2,1,0,0]=(0.953986415316)*x_ref[0]**o+(-0.7883326473)*x_ref[0]+(-0.723690234059)*x_ref[1]**o+(-0.537242781471)*x_ref[1]+(-0.503466717519)*x_ref[2]**o+(-0.777865169241)*x_ref[2]
        ref[2,1,0,1]=(0.317357554441)*x_ref[0]**o+(-0.326828957855)*x_ref[0]+(-0.609992651237)*x_ref[1]**o+(0.811066934743)*x_ref[1]+(-0.90668441447)*x_ref[2]**o+(0.92581926687)*x_ref[2]
        ref[2,1,1,0]=(-0.0801236189368)*x_ref[0]**o+(0.49151264813)*x_ref[0]+(-0.0770654446151)*x_ref[1]**o+(-0.314981217276)*x_ref[1]+(-0.743180837478)*x_ref[2]**o+(-0.554491413146)*x_ref[2]
        ref[2,1,1,1]=(0.831017344937)*x_ref[0]**o+(-0.61064793935)*x_ref[0]+(0.652976737027)*x_ref[1]**o+(0.242104127969)*x_ref[1]+(0.593330958678)*x_ref[2]**o+(-0.35899584298)*x_ref[2]
        ref[2,1,2,0]=(-0.580618798556)*x_ref[0]**o+(-0.936176474082)*x_ref[0]+(-0.768716477288)*x_ref[1]**o+(-0.353687473159)*x_ref[1]+(-0.261509314019)*x_ref[2]**o+(-0.442139199257)*x_ref[2]
        ref[2,1,2,1]=(0.945760793889)*x_ref[0]**o+(-0.646958670891)*x_ref[0]+(-0.440341726063)*x_ref[1]**o+(0.223283751014)*x_ref[1]+(0.799798944743)*x_ref[2]**o+(0.591902067633)*x_ref[2]
        ref[2,2,0,0]=(-0.267699746405)*x_ref[0]**o+(0.271554172399)*x_ref[0]+(0.747278903642)*x_ref[1]**o+(-0.832877459834)*x_ref[1]+(-0.640373164716)*x_ref[2]**o+(-0.666487164763)*x_ref[2]
        ref[2,2,0,1]=(0.795630715924)*x_ref[0]**o+(0.682973866186)*x_ref[0]+(-0.750818874926)*x_ref[1]**o+(-0.904782408217)*x_ref[1]+(-0.578138718379)*x_ref[2]**o+(-0.369846516992)*x_ref[2]
        ref[2,2,1,0]=(0.165654497062)*x_ref[0]**o+(-0.329827871165)*x_ref[0]+(-0.808981530672)*x_ref[1]**o+(0.350106763785)*x_ref[1]+(0.106638087689)*x_ref[2]**o+(0.425516820398)*x_ref[2]
        ref[2,2,1,1]=(0.017419119421)*x_ref[0]**o+(0.500359047996)*x_ref[0]+(-0.888509940049)*x_ref[1]**o+(-0.375469124891)*x_ref[1]+(0.863718970618)*x_ref[2]**o+(-0.807667604763)*x_ref[2]
        ref[2,2,2,0]=(0.860662652165)*x_ref[0]**o+(0.162530730499)*x_ref[0]+(-0.657510966961)*x_ref[1]**o+(0.215905840286)*x_ref[1]+(0.97380371264)*x_ref[2]**o+(-0.431027671897)*x_ref[2]
        ref[2,2,2,1]=(-0.649151503991)*x_ref[0]**o+(0.484589779517)*x_ref[0]+(0.905080495553)*x_ref[1]**o+(0.997504119204)*x_ref[1]+(-0.4268802512)*x_ref[2]**o+(-0.687236308612)*x_ref[2]
        ref[2,3,0,0]=(-0.60485902589)*x_ref[0]**o+(0.49500777286)*x_ref[0]+(0.33977532104)*x_ref[1]**o+(-0.219209845338)*x_ref[1]+(-0.197311627683)*x_ref[2]**o+(0.285859591044)*x_ref[2]
        ref[2,3,0,1]=(0.117438340483)*x_ref[0]**o+(-0.562954894877)*x_ref[0]+(-0.302982842797)*x_ref[1]**o+(0.977923453436)*x_ref[1]+(0.300127749784)*x_ref[2]**o+(0.45784201373)*x_ref[2]
        ref[2,3,1,0]=(-0.910061442266)*x_ref[0]**o+(0.723114553174)*x_ref[0]+(-0.435693478652)*x_ref[1]**o+(-0.584326584996)*x_ref[1]+(-0.51000911177)*x_ref[2]**o+(0.506615483889)*x_ref[2]
        ref[2,3,1,1]=(-0.50179741763)*x_ref[0]**o+(-0.954551290001)*x_ref[0]+(0.375839266192)*x_ref[1]**o+(0.122120198059)*x_ref[1]+(-0.593050674641)*x_ref[2]**o+(0.549989594312)*x_ref[2]
        ref[2,3,2,0]=(-0.889519666705)*x_ref[0]**o+(-0.815572507674)*x_ref[0]+(0.453677651057)*x_ref[1]**o+(0.450910828217)*x_ref[1]+(-0.00465373008682)*x_ref[2]**o+(0.731899900165)*x_ref[2]
        ref[2,3,2,1]=(-0.969971151589)*x_ref[0]**o+(-0.156948104522)*x_ref[0]+(0.492885029353)*x_ref[1]**o+(-0.0310677847833)*x_ref[1]+(0.327614233837)*x_ref[2]**o+(-0.274974040707)*x_ref[2]
        ref[2,4,0,0]=(0.844667900826)*x_ref[0]**o+(0.863374870522)*x_ref[0]+(0.251456923241)*x_ref[1]**o+(0.434916005183)*x_ref[1]+(-0.139023370677)*x_ref[2]**o+(0.0655176616053)*x_ref[2]
        ref[2,4,0,1]=(0.831755059318)*x_ref[0]**o+(0.0897642674136)*x_ref[0]+(0.190414878812)*x_ref[1]**o+(0.976565327031)*x_ref[1]+(0.413321767371)*x_ref[2]**o+(0.755395415341)*x_ref[2]
        ref[2,4,1,0]=(-0.790364876377)*x_ref[0]**o+(0.829442094248)*x_ref[0]+(0.556251403891)*x_ref[1]**o+(-0.953487711658)*x_ref[1]+(-0.610781905401)*x_ref[2]**o+(-0.320764072326)*x_ref[2]
        ref[2,4,1,1]=(-0.757101565136)*x_ref[0]**o+(0.789340839395)*x_ref[0]+(-0.615189514438)*x_ref[1]**o+(-0.910992676978)*x_ref[1]+(0.693736813401)*x_ref[2]**o+(0.0538707588266)*x_ref[2]
        ref[2,4,2,0]=(0.000335625616611)*x_ref[0]**o+(-0.195171446476)*x_ref[0]+(0.0281149133705)*x_ref[1]**o+(-0.362294159243)*x_ref[1]+(0.830520288743)*x_ref[2]**o+(-0.431967694247)*x_ref[2]
        ref[2,4,2,1]=(-0.232668364843)*x_ref[0]**o+(0.655302452302)*x_ref[0]+(-0.0125806408337)*x_ref[1]**o+(-0.103979202109)*x_ref[1]+(-0.277450383033)*x_ref[2]**o+(0.708914971885)*x_ref[2]
        ref[3,0,0,0]=(-0.813447875186)*x_ref[0]**o+(0.133491078101)*x_ref[0]+(-0.675510725453)*x_ref[1]**o+(-0.161489202717)*x_ref[1]+(0.0474778506607)*x_ref[2]**o+(0.513729029652)*x_ref[2]
        ref[3,0,0,1]=(0.45388386743)*x_ref[0]**o+(-0.909074944715)*x_ref[0]+(-0.540576382376)*x_ref[1]**o+(-0.849753706068)*x_ref[1]+(-0.496320191697)*x_ref[2]**o+(-0.930725196363)*x_ref[2]
        ref[3,0,1,0]=(-0.145526830518)*x_ref[0]**o+(-0.34717641097)*x_ref[0]+(-0.818886792022)*x_ref[1]**o+(0.574315644784)*x_ref[1]+(-0.780431329222)*x_ref[2]**o+(0.892840747651)*x_ref[2]
        ref[3,0,1,1]=(-0.415342962745)*x_ref[0]**o+(-0.899768945518)*x_ref[0]+(0.798573910731)*x_ref[1]**o+(-0.0652893933479)*x_ref[1]+(-0.444459519602)*x_ref[2]**o+(-0.0100131702565)*x_ref[2]
        ref[3,0,2,0]=(-0.997936676013)*x_ref[0]**o+(0.822000858069)*x_ref[0]+(-0.553794408949)*x_ref[1]**o+(0.829037287423)*x_ref[1]+(-0.700288437806)*x_ref[2]**o+(-0.449584600891)*x_ref[2]
        ref[3,0,2,1]=(-0.70979868598)*x_ref[0]**o+(0.689868687655)*x_ref[0]+(0.151257146397)*x_ref[1]**o+(0.32895215144)*x_ref[1]+(0.286043871383)*x_ref[2]**o+(-0.604421567781)*x_ref[2]
        ref[3,1,0,0]=(0.60431849542)*x_ref[0]**o+(0.518618291627)*x_ref[0]+(-0.471868097013)*x_ref[1]**o+(-0.639110899151)*x_ref[1]+(-0.0904184061601)*x_ref[2]**o+(0.113762528304)*x_ref[2]
        ref[3,1,0,1]=(0.307624103157)*x_ref[0]**o+(-0.388281302162)*x_ref[0]+(0.772429578102)*x_ref[1]**o+(-0.197799629585)*x_ref[1]+(-0.149120667036)*x_ref[2]**o+(0.830154780264)*x_ref[2]
        ref[3,1,1,0]=(0.957127413133)*x_ref[0]**o+(0.730080945389)*x_ref[0]+(0.497191948774)*x_ref[1]**o+(-0.586439221663)*x_ref[1]+(-0.56223071093)*x_ref[2]**o+(0.162899395804)*x_ref[2]
        ref[3,1,1,1]=(-0.285740818451)*x_ref[0]**o+(-0.573919970411)*x_ref[0]+(-0.400833211669)*x_ref[1]**o+(0.84414534634)*x_ref[1]+(-0.124352096731)*x_ref[2]**o+(0.907013800834)*x_ref[2]
        ref[3,1,2,0]=(0.186947347989)*x_ref[0]**o+(0.499582449213)*x_ref[0]+(0.525473636695)*x_ref[1]**o+(-0.926500953998)*x_ref[1]+(-0.0392615366415)*x_ref[2]**o+(0.909002911192)*x_ref[2]
        ref[3,1,2,1]=(0.634350440683)*x_ref[0]**o+(0.901699291665)*x_ref[0]+(-0.683899667575)*x_ref[1]**o+(0.411642037744)*x_ref[1]+(-0.72791611585)*x_ref[2]**o+(-0.013293689839)*x_ref[2]
        ref[3,2,0,0]=(0.185164429507)*x_ref[0]**o+(-0.671011652257)*x_ref[0]+(0.973361040829)*x_ref[1]**o+(0.989738565899)*x_ref[1]+(-0.501245887091)*x_ref[2]**o+(-0.405445444386)*x_ref[2]
        ref[3,2,0,1]=(-0.311554020884)*x_ref[0]**o+(-0.929020990333)*x_ref[0]+(0.366016375797)*x_ref[1]**o+(0.838234804773)*x_ref[1]+(0.939602632409)*x_ref[2]**o+(-0.885524260372)*x_ref[2]
        ref[3,2,1,0]=(-0.0101604655589)*x_ref[0]**o+(0.670274078493)*x_ref[0]+(0.198135855354)*x_ref[1]**o+(-0.000275422963644)*x_ref[1]+(0.136674233404)*x_ref[2]**o+(-0.442510118497)*x_ref[2]
        ref[3,2,1,1]=(-0.677948769785)*x_ref[0]**o+(-0.165403610813)*x_ref[0]+(-0.219353502981)*x_ref[1]**o+(0.664297154009)*x_ref[1]+(0.158536302892)*x_ref[2]**o+(0.293940884925)*x_ref[2]
        ref[3,2,2,0]=(0.666669534708)*x_ref[0]**o+(0.609459298502)*x_ref[0]+(0.297058070582)*x_ref[1]**o+(-0.0809555825821)*x_ref[1]+(0.139430791042)*x_ref[2]**o+(-0.253514861624)*x_ref[2]
        ref[3,2,2,1]=(-0.577932413274)*x_ref[0]**o+(-0.583353108895)*x_ref[0]+(-0.839262034394)*x_ref[1]**o+(0.795642236106)*x_ref[1]+(0.0882627381543)*x_ref[2]**o+(-0.382970623122)*x_ref[2]
        ref[3,3,0,0]=(0.187941627429)*x_ref[0]**o+(0.552040688531)*x_ref[0]+(0.837458832716)*x_ref[1]**o+(0.296793121897)*x_ref[1]+(-0.665699511216)*x_ref[2]**o+(-0.318565991052)*x_ref[2]
        ref[3,3,0,1]=(-0.78770176804)*x_ref[0]**o+(0.807509260396)*x_ref[0]+(0.46241086085)*x_ref[1]**o+(0.668506908751)*x_ref[1]+(-0.988954457343)*x_ref[2]**o+(0.0896365679915)*x_ref[2]
        ref[3,3,1,0]=(0.727665746433)*x_ref[0]**o+(-0.819391676739)*x_ref[0]+(-0.118648934607)*x_ref[1]**o+(-0.433862939031)*x_ref[1]+(0.827788711313)*x_ref[2]**o+(-0.615751446784)*x_ref[2]
        ref[3,3,1,1]=(-0.00427901841482)*x_ref[0]**o+(-0.38545713036)*x_ref[0]+(0.570160083404)*x_ref[1]**o+(-0.235505162795)*x_ref[1]+(0.514578304371)*x_ref[2]**o+(-0.152692113279)*x_ref[2]
        ref[3,3,2,0]=(-0.396900613831)*x_ref[0]**o+(0.161875314543)*x_ref[0]+(0.28471295984)*x_ref[1]**o+(-0.623276596728)*x_ref[1]+(-0.748772373219)*x_ref[2]**o+(0.584246898704)*x_ref[2]
        ref[3,3,2,1]=(0.210026027009)*x_ref[0]**o+(-0.530437530387)*x_ref[0]+(0.437155825716)*x_ref[1]**o+(0.984102407204)*x_ref[1]+(0.125030773605)*x_ref[2]**o+(-0.821764907318)*x_ref[2]
        ref[3,4,0,0]=(-0.819129946783)*x_ref[0]**o+(-0.729517565451)*x_ref[0]+(-0.126013244602)*x_ref[1]**o+(0.592832125254)*x_ref[1]+(-0.850426487488)*x_ref[2]**o+(0.271550912174)*x_ref[2]
        ref[3,4,0,1]=(-0.573554440921)*x_ref[0]**o+(0.232106829584)*x_ref[0]+(0.810201614042)*x_ref[1]**o+(-0.85518164211)*x_ref[1]+(0.249535563949)*x_ref[2]**o+(-0.69728963017)*x_ref[2]
        ref[3,4,1,0]=(0.785661643628)*x_ref[0]**o+(0.936424600424)*x_ref[0]+(0.321363811055)*x_ref[1]**o+(0.362341458592)*x_ref[1]+(-0.593576579363)*x_ref[2]**o+(0.0601278435924)*x_ref[2]
        ref[3,4,1,1]=(-0.854420686885)*x_ref[0]**o+(0.993352097501)*x_ref[0]+(-0.572116112224)*x_ref[1]**o+(0.308008845182)*x_ref[1]+(0.0443339578639)*x_ref[2]**o+(-0.1469857285)*x_ref[2]
        ref[3,4,2,0]=(0.696373642151)*x_ref[0]**o+(0.82425126832)*x_ref[0]+(-0.0891735358455)*x_ref[1]**o+(-0.22218993613)*x_ref[1]+(-0.238393149017)*x_ref[2]**o+(-0.631352718194)*x_ref[2]
        ref[3,4,2,1]=(-0.0932690431928)*x_ref[0]**o+(-0.391900483919)*x_ref[0]+(0.236843565361)*x_ref[1]**o+(-0.164197648532)*x_ref[1]+(0.235502596353)*x_ref[2]**o+(0.0110131867917)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.592324160495)*x[0]**o+(-0.0471428147361)*x[0]+(0.756452478895)*x[1]**o+(-0.531318490356)*x[1]
        ref=(-0.592324160495)*x_ref[0]**o+(-0.0471428147361)*x_ref[0]+(0.756452478895)*x_ref[1]**o+(-0.531318490356)*x_ref[1]
      else:
        s=(0.935975918364)*x[0]**o+(-0.0518611788088)*x[0]+(0.198153832638)*x[1]**o+(0.0915392556709)*x[1]+(-0.740073697392)*x[2]**o+(0.0344998139597)*x[2]
        ref=(0.935975918364)*x_ref[0]**o+(-0.0518611788088)*x_ref[0]+(0.198153832638)*x_ref[1]**o+(0.0915392556709)*x_ref[1]+(-0.740073697392)*x_ref[2]**o+(0.0344998139597)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.115732010972)*x[0]**o+(-0.244595740544)*x[0]+(-0.627201201607)*x[1]**o+(-0.0597882966954)*x[1]
        s[1]=(0.0303875091965)*x[0]**o+(0.474828783855)*x[0]+(0.0299053774471)*x[1]**o+(0.364050762964)*x[1]
        ref[0]=(0.115732010972)*x_ref[0]**o+(-0.244595740544)*x_ref[0]+(-0.627201201607)*x_ref[1]**o+(-0.0597882966954)*x_ref[1]
        ref[1]=(0.0303875091965)*x_ref[0]**o+(0.474828783855)*x_ref[0]+(0.0299053774471)*x_ref[1]**o+(0.364050762964)*x_ref[1]
      else:
        s[0]=(0.645056057607)*x[0]**o+(0.154023831925)*x[0]+(0.121000626682)*x[1]**o+(-0.962672948686)*x[1]+(0.997016990346)*x[2]**o+(-0.746450513341)*x[2]
        s[1]=(0.880153248744)*x[0]**o+(0.399741188143)*x[0]+(-0.862453529609)*x[1]**o+(0.777403830466)*x[1]+(-0.759299926221)*x[2]**o+(0.271056515624)*x[2]
        ref[0]=(0.645056057607)*x_ref[0]**o+(0.154023831925)*x_ref[0]+(0.121000626682)*x_ref[1]**o+(-0.962672948686)*x_ref[1]+(0.997016990346)*x_ref[2]**o+(-0.746450513341)*x_ref[2]
        ref[1]=(0.880153248744)*x_ref[0]**o+(0.399741188143)*x_ref[0]+(-0.862453529609)*x_ref[1]**o+(0.777403830466)*x_ref[1]+(-0.759299926221)*x_ref[2]**o+(0.271056515624)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.163301494275)*x[0]**o+(-0.571871996499)*x[0]+(-0.126318698091)*x[1]**o+(0.594323280745)*x[1]
        s[0,1]=(-0.722891165327)*x[0]**o+(0.539458396201)*x[0]+(-0.338410805174)*x[1]**o+(-0.0441039302518)*x[1]
        s[0,2]=(0.987165940018)*x[0]**o+(0.277540309671)*x[0]+(-0.194843676722)*x[1]**o+(-0.838659115747)*x[1]
        s[0,3]=(0.668488628751)*x[0]**o+(0.848828025004)*x[0]+(-0.749811369279)*x[1]**o+(-0.857737537329)*x[1]
        s[0,4]=(0.292839673829)*x[0]**o+(-0.196773033161)*x[0]+(-0.365394929162)*x[1]**o+(-0.304841309615)*x[1]
        s[1,0]=(-0.244415346064)*x[0]**o+(0.0653469288689)*x[0]+(-0.405830560862)*x[1]**o+(0.958797302829)*x[1]
        s[1,1]=(-0.420934154847)*x[0]**o+(0.532576137523)*x[0]+(0.65082076651)*x[1]**o+(-0.264312562332)*x[1]
        s[1,2]=(-0.133788641889)*x[0]**o+(0.556829684305)*x[0]+(0.134872826121)*x[1]**o+(0.716478729883)*x[1]
        s[1,3]=(-0.345881339901)*x[0]**o+(-0.319261340187)*x[0]+(-0.114438699999)*x[1]**o+(0.0825491045988)*x[1]
        s[1,4]=(0.630228774591)*x[0]**o+(-0.858970963301)*x[0]+(0.708103720535)*x[1]**o+(-0.542940693024)*x[1]
        s[2,0]=(-0.683466738531)*x[0]**o+(-0.145169931761)*x[0]+(0.449697093379)*x[1]**o+(-0.907773514556)*x[1]
        s[2,1]=(-0.0705140461777)*x[0]**o+(0.374565128403)*x[0]+(-0.219001949605)*x[1]**o+(-0.239599954648)*x[1]
        s[2,2]=(-0.435803146278)*x[0]**o+(-0.150900699809)*x[0]+(-0.028936013795)*x[1]**o+(-0.169287689744)*x[1]
        s[2,3]=(-0.976991446034)*x[0]**o+(-0.975014510986)*x[0]+(0.314904815148)*x[1]**o+(0.479090592274)*x[1]
        s[2,4]=(-0.955674787191)*x[0]**o+(0.633046166846)*x[0]+(0.680202620704)*x[1]**o+(0.0533385507777)*x[1]
        s[3,0]=(0.876058021021)*x[0]**o+(0.294827067859)*x[0]+(0.482052318125)*x[1]**o+(-0.365719292507)*x[1]
        s[3,1]=(0.530559404731)*x[0]**o+(-0.313386503724)*x[0]+(-0.371352352152)*x[1]**o+(-0.682915252526)*x[1]
        s[3,2]=(-0.0522489801875)*x[0]**o+(0.631979705829)*x[0]+(-0.989995736543)*x[1]**o+(-0.386092008201)*x[1]
        s[3,3]=(-0.360737462637)*x[0]**o+(-0.0286143460674)*x[0]+(-0.743283421495)*x[1]**o+(0.346219067885)*x[1]
        s[3,4]=(-0.880285923018)*x[0]**o+(0.492777568952)*x[0]+(0.572215124514)*x[1]**o+(0.201965538371)*x[1]
        ref[0,0]=(-0.163301494275)*x_ref[0]**o+(-0.571871996499)*x_ref[0]+(-0.126318698091)*x_ref[1]**o+(0.594323280745)*x_ref[1]
        ref[0,1]=(-0.722891165327)*x_ref[0]**o+(0.539458396201)*x_ref[0]+(-0.338410805174)*x_ref[1]**o+(-0.0441039302518)*x_ref[1]
        ref[0,2]=(0.987165940018)*x_ref[0]**o+(0.277540309671)*x_ref[0]+(-0.194843676722)*x_ref[1]**o+(-0.838659115747)*x_ref[1]
        ref[0,3]=(0.668488628751)*x_ref[0]**o+(0.848828025004)*x_ref[0]+(-0.749811369279)*x_ref[1]**o+(-0.857737537329)*x_ref[1]
        ref[0,4]=(0.292839673829)*x_ref[0]**o+(-0.196773033161)*x_ref[0]+(-0.365394929162)*x_ref[1]**o+(-0.304841309615)*x_ref[1]
        ref[1,0]=(-0.244415346064)*x_ref[0]**o+(0.0653469288689)*x_ref[0]+(-0.405830560862)*x_ref[1]**o+(0.958797302829)*x_ref[1]
        ref[1,1]=(-0.420934154847)*x_ref[0]**o+(0.532576137523)*x_ref[0]+(0.65082076651)*x_ref[1]**o+(-0.264312562332)*x_ref[1]
        ref[1,2]=(-0.133788641889)*x_ref[0]**o+(0.556829684305)*x_ref[0]+(0.134872826121)*x_ref[1]**o+(0.716478729883)*x_ref[1]
        ref[1,3]=(-0.345881339901)*x_ref[0]**o+(-0.319261340187)*x_ref[0]+(-0.114438699999)*x_ref[1]**o+(0.0825491045988)*x_ref[1]
        ref[1,4]=(0.630228774591)*x_ref[0]**o+(-0.858970963301)*x_ref[0]+(0.708103720535)*x_ref[1]**o+(-0.542940693024)*x_ref[1]
        ref[2,0]=(-0.683466738531)*x_ref[0]**o+(-0.145169931761)*x_ref[0]+(0.449697093379)*x_ref[1]**o+(-0.907773514556)*x_ref[1]
        ref[2,1]=(-0.0705140461777)*x_ref[0]**o+(0.374565128403)*x_ref[0]+(-0.219001949605)*x_ref[1]**o+(-0.239599954648)*x_ref[1]
        ref[2,2]=(-0.435803146278)*x_ref[0]**o+(-0.150900699809)*x_ref[0]+(-0.028936013795)*x_ref[1]**o+(-0.169287689744)*x_ref[1]
        ref[2,3]=(-0.976991446034)*x_ref[0]**o+(-0.975014510986)*x_ref[0]+(0.314904815148)*x_ref[1]**o+(0.479090592274)*x_ref[1]
        ref[2,4]=(-0.955674787191)*x_ref[0]**o+(0.633046166846)*x_ref[0]+(0.680202620704)*x_ref[1]**o+(0.0533385507777)*x_ref[1]
        ref[3,0]=(0.876058021021)*x_ref[0]**o+(0.294827067859)*x_ref[0]+(0.482052318125)*x_ref[1]**o+(-0.365719292507)*x_ref[1]
        ref[3,1]=(0.530559404731)*x_ref[0]**o+(-0.313386503724)*x_ref[0]+(-0.371352352152)*x_ref[1]**o+(-0.682915252526)*x_ref[1]
        ref[3,2]=(-0.0522489801875)*x_ref[0]**o+(0.631979705829)*x_ref[0]+(-0.989995736543)*x_ref[1]**o+(-0.386092008201)*x_ref[1]
        ref[3,3]=(-0.360737462637)*x_ref[0]**o+(-0.0286143460674)*x_ref[0]+(-0.743283421495)*x_ref[1]**o+(0.346219067885)*x_ref[1]
        ref[3,4]=(-0.880285923018)*x_ref[0]**o+(0.492777568952)*x_ref[0]+(0.572215124514)*x_ref[1]**o+(0.201965538371)*x_ref[1]
      else:
        s[0,0]=(0.459801313828)*x[0]**o+(0.874491492894)*x[0]+(-0.676707065181)*x[1]**o+(-0.106095988336)*x[1]+(-0.171044270123)*x[2]**o+(0.0638260656588)*x[2]
        s[0,1]=(0.596371056873)*x[0]**o+(0.316401622905)*x[0]+(0.238994026229)*x[1]**o+(-0.462592805759)*x[1]+(-0.237565256761)*x[2]**o+(-0.100115327497)*x[2]
        s[0,2]=(-0.423227643937)*x[0]**o+(0.706315080194)*x[0]+(-0.43521651208)*x[1]**o+(-0.0219505777477)*x[1]+(-0.248035954152)*x[2]**o+(-0.0426762513079)*x[2]
        s[0,3]=(-0.807609617307)*x[0]**o+(0.870379040362)*x[0]+(0.0450254577238)*x[1]**o+(0.409422364523)*x[1]+(-0.0747425184319)*x[2]**o+(0.714168287946)*x[2]
        s[0,4]=(0.229697696099)*x[0]**o+(-0.193953210678)*x[0]+(0.821756203411)*x[1]**o+(-0.686241275184)*x[1]+(-0.461813580643)*x[2]**o+(-0.111679064974)*x[2]
        s[1,0]=(0.09199980955)*x[0]**o+(-0.596786310748)*x[0]+(-0.519482393268)*x[1]**o+(0.494226697149)*x[1]+(-0.533138224953)*x[2]**o+(-0.442965941122)*x[2]
        s[1,1]=(-0.499699311698)*x[0]**o+(-0.300118366174)*x[0]+(-0.0622786566862)*x[1]**o+(0.389573411943)*x[1]+(0.812614223101)*x[2]**o+(0.0241052242281)*x[2]
        s[1,2]=(-0.937872631185)*x[0]**o+(0.693932819382)*x[0]+(-0.940606874805)*x[1]**o+(0.0565548603525)*x[1]+(0.488404217301)*x[2]**o+(-0.54908344074)*x[2]
        s[1,3]=(0.187440914227)*x[0]**o+(0.513746783521)*x[0]+(0.975266666494)*x[1]**o+(-0.298164635346)*x[1]+(0.796028384674)*x[2]**o+(-0.555474012189)*x[2]
        s[1,4]=(-0.542284827327)*x[0]**o+(0.72336133394)*x[0]+(-0.382404238652)*x[1]**o+(-0.966114351457)*x[1]+(-0.0713526699547)*x[2]**o+(-0.969056495871)*x[2]
        s[2,0]=(-0.669587562542)*x[0]**o+(0.927249904009)*x[0]+(0.0901118246496)*x[1]**o+(0.0105771498119)*x[1]+(0.0347578359214)*x[2]**o+(0.306378930574)*x[2]
        s[2,1]=(0.552110347651)*x[0]**o+(0.534155224195)*x[0]+(-0.94964594078)*x[1]**o+(-0.867398879865)*x[1]+(0.529146894499)*x[2]**o+(-0.478336329324)*x[2]
        s[2,2]=(0.0250007510203)*x[0]**o+(-0.249616427955)*x[0]+(0.606560318424)*x[1]**o+(-0.921833180264)*x[1]+(-0.15817883795)*x[2]**o+(0.430892612477)*x[2]
        s[2,3]=(0.186816270003)*x[0]**o+(0.408732439837)*x[0]+(-0.89708625066)*x[1]**o+(0.0672189023842)*x[1]+(-0.502969902157)*x[2]**o+(0.0172554167453)*x[2]
        s[2,4]=(0.458126497715)*x[0]**o+(-0.929705410355)*x[0]+(0.493346469334)*x[1]**o+(-0.432232217413)*x[1]+(-0.579275855459)*x[2]**o+(0.756337283238)*x[2]
        s[3,0]=(0.933176387441)*x[0]**o+(0.24114580865)*x[0]+(-0.830402155007)*x[1]**o+(0.925370556667)*x[1]+(0.242042958988)*x[2]**o+(0.341138866916)*x[2]
        s[3,1]=(0.0660813821046)*x[0]**o+(-0.492921246892)*x[0]+(0.728036967007)*x[1]**o+(0.241487680125)*x[1]+(-0.705909691396)*x[2]**o+(0.812820536373)*x[2]
        s[3,2]=(-0.71644646201)*x[0]**o+(-0.428574354372)*x[0]+(0.928810019254)*x[1]**o+(0.0142522119308)*x[1]+(0.641710384867)*x[2]**o+(-0.512508294632)*x[2]
        s[3,3]=(-0.185856758974)*x[0]**o+(-0.265588242215)*x[0]+(-0.0750623283309)*x[1]**o+(-0.166114006253)*x[1]+(-0.986359701771)*x[2]**o+(0.237860767478)*x[2]
        s[3,4]=(0.0423246107159)*x[0]**o+(0.216782521414)*x[0]+(-0.536875736799)*x[1]**o+(0.0489914911388)*x[1]+(0.759222862013)*x[2]**o+(0.743454532517)*x[2]
        ref[0,0]=(0.459801313828)*x_ref[0]**o+(0.874491492894)*x_ref[0]+(-0.676707065181)*x_ref[1]**o+(-0.106095988336)*x_ref[1]+(-0.171044270123)*x_ref[2]**o+(0.0638260656588)*x_ref[2]
        ref[0,1]=(0.596371056873)*x_ref[0]**o+(0.316401622905)*x_ref[0]+(0.238994026229)*x_ref[1]**o+(-0.462592805759)*x_ref[1]+(-0.237565256761)*x_ref[2]**o+(-0.100115327497)*x_ref[2]
        ref[0,2]=(-0.423227643937)*x_ref[0]**o+(0.706315080194)*x_ref[0]+(-0.43521651208)*x_ref[1]**o+(-0.0219505777477)*x_ref[1]+(-0.248035954152)*x_ref[2]**o+(-0.0426762513079)*x_ref[2]
        ref[0,3]=(-0.807609617307)*x_ref[0]**o+(0.870379040362)*x_ref[0]+(0.0450254577238)*x_ref[1]**o+(0.409422364523)*x_ref[1]+(-0.0747425184319)*x_ref[2]**o+(0.714168287946)*x_ref[2]
        ref[0,4]=(0.229697696099)*x_ref[0]**o+(-0.193953210678)*x_ref[0]+(0.821756203411)*x_ref[1]**o+(-0.686241275184)*x_ref[1]+(-0.461813580643)*x_ref[2]**o+(-0.111679064974)*x_ref[2]
        ref[1,0]=(0.09199980955)*x_ref[0]**o+(-0.596786310748)*x_ref[0]+(-0.519482393268)*x_ref[1]**o+(0.494226697149)*x_ref[1]+(-0.533138224953)*x_ref[2]**o+(-0.442965941122)*x_ref[2]
        ref[1,1]=(-0.499699311698)*x_ref[0]**o+(-0.300118366174)*x_ref[0]+(-0.0622786566862)*x_ref[1]**o+(0.389573411943)*x_ref[1]+(0.812614223101)*x_ref[2]**o+(0.0241052242281)*x_ref[2]
        ref[1,2]=(-0.937872631185)*x_ref[0]**o+(0.693932819382)*x_ref[0]+(-0.940606874805)*x_ref[1]**o+(0.0565548603525)*x_ref[1]+(0.488404217301)*x_ref[2]**o+(-0.54908344074)*x_ref[2]
        ref[1,3]=(0.187440914227)*x_ref[0]**o+(0.513746783521)*x_ref[0]+(0.975266666494)*x_ref[1]**o+(-0.298164635346)*x_ref[1]+(0.796028384674)*x_ref[2]**o+(-0.555474012189)*x_ref[2]
        ref[1,4]=(-0.542284827327)*x_ref[0]**o+(0.72336133394)*x_ref[0]+(-0.382404238652)*x_ref[1]**o+(-0.966114351457)*x_ref[1]+(-0.0713526699547)*x_ref[2]**o+(-0.969056495871)*x_ref[2]
        ref[2,0]=(-0.669587562542)*x_ref[0]**o+(0.927249904009)*x_ref[0]+(0.0901118246496)*x_ref[1]**o+(0.0105771498119)*x_ref[1]+(0.0347578359214)*x_ref[2]**o+(0.306378930574)*x_ref[2]
        ref[2,1]=(0.552110347651)*x_ref[0]**o+(0.534155224195)*x_ref[0]+(-0.94964594078)*x_ref[1]**o+(-0.867398879865)*x_ref[1]+(0.529146894499)*x_ref[2]**o+(-0.478336329324)*x_ref[2]
        ref[2,2]=(0.0250007510203)*x_ref[0]**o+(-0.249616427955)*x_ref[0]+(0.606560318424)*x_ref[1]**o+(-0.921833180264)*x_ref[1]+(-0.15817883795)*x_ref[2]**o+(0.430892612477)*x_ref[2]
        ref[2,3]=(0.186816270003)*x_ref[0]**o+(0.408732439837)*x_ref[0]+(-0.89708625066)*x_ref[1]**o+(0.0672189023842)*x_ref[1]+(-0.502969902157)*x_ref[2]**o+(0.0172554167453)*x_ref[2]
        ref[2,4]=(0.458126497715)*x_ref[0]**o+(-0.929705410355)*x_ref[0]+(0.493346469334)*x_ref[1]**o+(-0.432232217413)*x_ref[1]+(-0.579275855459)*x_ref[2]**o+(0.756337283238)*x_ref[2]
        ref[3,0]=(0.933176387441)*x_ref[0]**o+(0.24114580865)*x_ref[0]+(-0.830402155007)*x_ref[1]**o+(0.925370556667)*x_ref[1]+(0.242042958988)*x_ref[2]**o+(0.341138866916)*x_ref[2]
        ref[3,1]=(0.0660813821046)*x_ref[0]**o+(-0.492921246892)*x_ref[0]+(0.728036967007)*x_ref[1]**o+(0.241487680125)*x_ref[1]+(-0.705909691396)*x_ref[2]**o+(0.812820536373)*x_ref[2]
        ref[3,2]=(-0.71644646201)*x_ref[0]**o+(-0.428574354372)*x_ref[0]+(0.928810019254)*x_ref[1]**o+(0.0142522119308)*x_ref[1]+(0.641710384867)*x_ref[2]**o+(-0.512508294632)*x_ref[2]
        ref[3,3]=(-0.185856758974)*x_ref[0]**o+(-0.265588242215)*x_ref[0]+(-0.0750623283309)*x_ref[1]**o+(-0.166114006253)*x_ref[1]+(-0.986359701771)*x_ref[2]**o+(0.237860767478)*x_ref[2]
        ref[3,4]=(0.0423246107159)*x_ref[0]**o+(0.216782521414)*x_ref[0]+(-0.536875736799)*x_ref[1]**o+(0.0489914911388)*x_ref[1]+(0.759222862013)*x_ref[2]**o+(0.743454532517)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.598481759913)*x[0]**o+(-0.272925610474)*x[0]+(0.551460738992)*x[1]**o+(-0.609542685963)*x[1]
        s[0,0,1]=(0.645588310869)*x[0]**o+(0.983661063707)*x[0]+(-0.585913047558)*x[1]**o+(-0.700447569863)*x[1]
        s[0,1,0]=(0.0735002177703)*x[0]**o+(0.640392078189)*x[0]+(-0.640531878864)*x[1]**o+(0.927957489213)*x[1]
        s[0,1,1]=(-0.91764393635)*x[0]**o+(-0.234646513068)*x[0]+(0.434778070289)*x[1]**o+(0.676898688957)*x[1]
        s[1,0,0]=(0.900485566233)*x[0]**o+(0.17381724123)*x[0]+(-0.0472757726724)*x[1]**o+(-0.603802455434)*x[1]
        s[1,0,1]=(0.0111284144129)*x[0]**o+(0.729344462439)*x[0]+(0.650641541889)*x[1]**o+(-0.382226116166)*x[1]
        s[1,1,0]=(0.866905202972)*x[0]**o+(0.586112114242)*x[0]+(0.132836683118)*x[1]**o+(0.685439975799)*x[1]
        s[1,1,1]=(0.0893846587203)*x[0]**o+(0.475293339259)*x[0]+(-0.524027603353)*x[1]**o+(-0.427042646418)*x[1]
        s[2,0,0]=(0.696717009328)*x[0]**o+(0.17836429668)*x[0]+(0.339729683456)*x[1]**o+(0.425401166923)*x[1]
        s[2,0,1]=(-0.578374482181)*x[0]**o+(-0.06883462626)*x[0]+(-0.62708396204)*x[1]**o+(-0.732305455774)*x[1]
        s[2,1,0]=(-0.170455966767)*x[0]**o+(-0.19206482096)*x[0]+(-0.291324644463)*x[1]**o+(0.0296539206517)*x[1]
        s[2,1,1]=(0.209978578865)*x[0]**o+(-0.026213873389)*x[0]+(-0.248910233717)*x[1]**o+(-0.855378045214)*x[1]
        s[3,0,0]=(-0.926098922919)*x[0]**o+(-0.12977446027)*x[0]+(-0.706408024307)*x[1]**o+(0.00802672230618)*x[1]
        s[3,0,1]=(-0.677659537077)*x[0]**o+(0.319391836395)*x[0]+(0.876973869047)*x[1]**o+(0.569113379857)*x[1]
        s[3,1,0]=(-0.486504252182)*x[0]**o+(-0.0946134881675)*x[0]+(0.490827332525)*x[1]**o+(-0.392910432859)*x[1]
        s[3,1,1]=(0.784833517776)*x[0]**o+(-0.576527986841)*x[0]+(0.439029003252)*x[1]**o+(-0.307033754704)*x[1]
        s[4,0,0]=(0.290446522592)*x[0]**o+(0.441136393638)*x[0]+(-0.372607076196)*x[1]**o+(-0.159674601004)*x[1]
        s[4,0,1]=(0.440515634069)*x[0]**o+(-0.42368755307)*x[0]+(0.370018085845)*x[1]**o+(0.0235740884719)*x[1]
        s[4,1,0]=(-0.697567283199)*x[0]**o+(0.327933485486)*x[0]+(-0.36825773126)*x[1]**o+(0.016215691731)*x[1]
        s[4,1,1]=(0.572572720028)*x[0]**o+(0.0840839889325)*x[0]+(0.826717033934)*x[1]**o+(-0.0474809600044)*x[1]
        s[5,0,0]=(-0.16977415879)*x[0]**o+(-0.448717259264)*x[0]+(-0.595995298169)*x[1]**o+(-0.331364898672)*x[1]
        s[5,0,1]=(-0.0068969930042)*x[0]**o+(-0.599597651559)*x[0]+(0.912012252833)*x[1]**o+(-0.0802048593495)*x[1]
        s[5,1,0]=(-0.897450521147)*x[0]**o+(-0.126994660738)*x[0]+(0.939958074736)*x[1]**o+(-0.600360767914)*x[1]
        s[5,1,1]=(-0.733313832271)*x[0]**o+(-0.53733764856)*x[0]+(-0.770590244589)*x[1]**o+(0.62787073827)*x[1]
        ref[0,0,0]=(-0.598481759913)*x_ref[0]**o+(-0.272925610474)*x_ref[0]+(0.551460738992)*x_ref[1]**o+(-0.609542685963)*x_ref[1]
        ref[0,0,1]=(0.645588310869)*x_ref[0]**o+(0.983661063707)*x_ref[0]+(-0.585913047558)*x_ref[1]**o+(-0.700447569863)*x_ref[1]
        ref[0,1,0]=(0.0735002177703)*x_ref[0]**o+(0.640392078189)*x_ref[0]+(-0.640531878864)*x_ref[1]**o+(0.927957489213)*x_ref[1]
        ref[0,1,1]=(-0.91764393635)*x_ref[0]**o+(-0.234646513068)*x_ref[0]+(0.434778070289)*x_ref[1]**o+(0.676898688957)*x_ref[1]
        ref[1,0,0]=(0.900485566233)*x_ref[0]**o+(0.17381724123)*x_ref[0]+(-0.0472757726724)*x_ref[1]**o+(-0.603802455434)*x_ref[1]
        ref[1,0,1]=(0.0111284144129)*x_ref[0]**o+(0.729344462439)*x_ref[0]+(0.650641541889)*x_ref[1]**o+(-0.382226116166)*x_ref[1]
        ref[1,1,0]=(0.866905202972)*x_ref[0]**o+(0.586112114242)*x_ref[0]+(0.132836683118)*x_ref[1]**o+(0.685439975799)*x_ref[1]
        ref[1,1,1]=(0.0893846587203)*x_ref[0]**o+(0.475293339259)*x_ref[0]+(-0.524027603353)*x_ref[1]**o+(-0.427042646418)*x_ref[1]
        ref[2,0,0]=(0.696717009328)*x_ref[0]**o+(0.17836429668)*x_ref[0]+(0.339729683456)*x_ref[1]**o+(0.425401166923)*x_ref[1]
        ref[2,0,1]=(-0.578374482181)*x_ref[0]**o+(-0.06883462626)*x_ref[0]+(-0.62708396204)*x_ref[1]**o+(-0.732305455774)*x_ref[1]
        ref[2,1,0]=(-0.170455966767)*x_ref[0]**o+(-0.19206482096)*x_ref[0]+(-0.291324644463)*x_ref[1]**o+(0.0296539206517)*x_ref[1]
        ref[2,1,1]=(0.209978578865)*x_ref[0]**o+(-0.026213873389)*x_ref[0]+(-0.248910233717)*x_ref[1]**o+(-0.855378045214)*x_ref[1]
        ref[3,0,0]=(-0.926098922919)*x_ref[0]**o+(-0.12977446027)*x_ref[0]+(-0.706408024307)*x_ref[1]**o+(0.00802672230618)*x_ref[1]
        ref[3,0,1]=(-0.677659537077)*x_ref[0]**o+(0.319391836395)*x_ref[0]+(0.876973869047)*x_ref[1]**o+(0.569113379857)*x_ref[1]
        ref[3,1,0]=(-0.486504252182)*x_ref[0]**o+(-0.0946134881675)*x_ref[0]+(0.490827332525)*x_ref[1]**o+(-0.392910432859)*x_ref[1]
        ref[3,1,1]=(0.784833517776)*x_ref[0]**o+(-0.576527986841)*x_ref[0]+(0.439029003252)*x_ref[1]**o+(-0.307033754704)*x_ref[1]
        ref[4,0,0]=(0.290446522592)*x_ref[0]**o+(0.441136393638)*x_ref[0]+(-0.372607076196)*x_ref[1]**o+(-0.159674601004)*x_ref[1]
        ref[4,0,1]=(0.440515634069)*x_ref[0]**o+(-0.42368755307)*x_ref[0]+(0.370018085845)*x_ref[1]**o+(0.0235740884719)*x_ref[1]
        ref[4,1,0]=(-0.697567283199)*x_ref[0]**o+(0.327933485486)*x_ref[0]+(-0.36825773126)*x_ref[1]**o+(0.016215691731)*x_ref[1]
        ref[4,1,1]=(0.572572720028)*x_ref[0]**o+(0.0840839889325)*x_ref[0]+(0.826717033934)*x_ref[1]**o+(-0.0474809600044)*x_ref[1]
        ref[5,0,0]=(-0.16977415879)*x_ref[0]**o+(-0.448717259264)*x_ref[0]+(-0.595995298169)*x_ref[1]**o+(-0.331364898672)*x_ref[1]
        ref[5,0,1]=(-0.0068969930042)*x_ref[0]**o+(-0.599597651559)*x_ref[0]+(0.912012252833)*x_ref[1]**o+(-0.0802048593495)*x_ref[1]
        ref[5,1,0]=(-0.897450521147)*x_ref[0]**o+(-0.126994660738)*x_ref[0]+(0.939958074736)*x_ref[1]**o+(-0.600360767914)*x_ref[1]
        ref[5,1,1]=(-0.733313832271)*x_ref[0]**o+(-0.53733764856)*x_ref[0]+(-0.770590244589)*x_ref[1]**o+(0.62787073827)*x_ref[1]
      else:
        s[0,0,0]=(0.49027687523)*x[0]**o+(-0.945396297892)*x[0]+(-0.649119833947)*x[1]**o+(-0.193510024763)*x[1]+(0.672065654766)*x[2]**o+(0.396941813883)*x[2]
        s[0,0,1]=(-0.379608112905)*x[0]**o+(-0.321849772246)*x[0]+(-0.695093455872)*x[1]**o+(-0.114263310683)*x[1]+(0.425556105367)*x[2]**o+(0.737954023041)*x[2]
        s[0,1,0]=(0.597295505367)*x[0]**o+(0.285081105939)*x[0]+(-0.690746281653)*x[1]**o+(0.431456884332)*x[1]+(-0.569785421553)*x[2]**o+(0.218907517674)*x[2]
        s[0,1,1]=(-0.743660806659)*x[0]**o+(0.916636055834)*x[0]+(0.747852198685)*x[1]**o+(-0.304767976464)*x[1]+(0.737051048825)*x[2]**o+(0.108795699665)*x[2]
        s[1,0,0]=(0.0322299276093)*x[0]**o+(-0.507367087704)*x[0]+(-0.154613332688)*x[1]**o+(-0.642692955185)*x[1]+(0.921618425322)*x[2]**o+(-0.89234021469)*x[2]
        s[1,0,1]=(0.250927374346)*x[0]**o+(-0.713045403997)*x[0]+(0.0229252996751)*x[1]**o+(0.251843016211)*x[1]+(0.683324421319)*x[2]**o+(0.466509616753)*x[2]
        s[1,1,0]=(0.610653309325)*x[0]**o+(0.503568507563)*x[0]+(0.519292621274)*x[1]**o+(-0.452962008782)*x[1]+(0.908988844098)*x[2]**o+(-0.933406144464)*x[2]
        s[1,1,1]=(0.271788487178)*x[0]**o+(0.341989918651)*x[0]+(-0.406971961331)*x[1]**o+(-0.928647769169)*x[1]+(0.662902796818)*x[2]**o+(0.655876724787)*x[2]
        s[2,0,0]=(-0.775353204706)*x[0]**o+(0.278726576715)*x[0]+(0.448754566832)*x[1]**o+(-0.943150229959)*x[1]+(0.0806002558903)*x[2]**o+(-0.350491928882)*x[2]
        s[2,0,1]=(-0.899719123904)*x[0]**o+(-0.89394097319)*x[0]+(0.136672330017)*x[1]**o+(0.941198317525)*x[1]+(0.945970508521)*x[2]**o+(0.232574429377)*x[2]
        s[2,1,0]=(0.102227090178)*x[0]**o+(0.498258334271)*x[0]+(-0.369219100853)*x[1]**o+(0.965612831948)*x[1]+(-0.580434278342)*x[2]**o+(0.0747696579664)*x[2]
        s[2,1,1]=(0.0680171783989)*x[0]**o+(-0.655506093893)*x[0]+(0.0854275560955)*x[1]**o+(0.788063374419)*x[1]+(0.3912557517)*x[2]**o+(-0.911795675902)*x[2]
        s[3,0,0]=(0.941143798566)*x[0]**o+(0.324489623574)*x[0]+(-0.259012903836)*x[1]**o+(-0.779051329011)*x[1]+(0.673245283112)*x[2]**o+(-0.888729936858)*x[2]
        s[3,0,1]=(0.904124513313)*x[0]**o+(0.564171149844)*x[0]+(0.799169083649)*x[1]**o+(-0.172322806388)*x[1]+(-0.0754065342235)*x[2]**o+(0.0419393221333)*x[2]
        s[3,1,0]=(-0.200531316713)*x[0]**o+(0.786777777323)*x[0]+(-0.222824591798)*x[1]**o+(0.32280872616)*x[1]+(-0.418791963436)*x[2]**o+(-0.549019641386)*x[2]
        s[3,1,1]=(0.0519992199857)*x[0]**o+(0.278310751677)*x[0]+(0.590847528061)*x[1]**o+(0.258564224896)*x[1]+(-0.500091685923)*x[2]**o+(0.511781801585)*x[2]
        s[4,0,0]=(0.459040882538)*x[0]**o+(-0.481335872602)*x[0]+(0.0556391566559)*x[1]**o+(-0.208836885623)*x[1]+(-0.11717985744)*x[2]**o+(-0.544122946452)*x[2]
        s[4,0,1]=(-0.255258906635)*x[0]**o+(0.860765893836)*x[0]+(0.389276109422)*x[1]**o+(-0.999502217953)*x[1]+(-0.715269648526)*x[2]**o+(0.451562561245)*x[2]
        s[4,1,0]=(-0.579717249622)*x[0]**o+(0.984974487836)*x[0]+(-0.803917364532)*x[1]**o+(-0.764878964835)*x[1]+(-0.0642799485519)*x[2]**o+(-0.917022175106)*x[2]
        s[4,1,1]=(-0.873863405322)*x[0]**o+(-0.872375014129)*x[0]+(0.981355044198)*x[1]**o+(0.40407870155)*x[1]+(0.796125441505)*x[2]**o+(0.486002048777)*x[2]
        s[5,0,0]=(0.92716952184)*x[0]**o+(-0.280973691624)*x[0]+(0.815340507346)*x[1]**o+(-0.555904571967)*x[1]+(0.765200732877)*x[2]**o+(-0.48251397176)*x[2]
        s[5,0,1]=(-0.6976034156)*x[0]**o+(-0.342128490335)*x[0]+(-0.273657764836)*x[1]**o+(-0.393975979448)*x[1]+(0.522428789629)*x[2]**o+(-0.210033793616)*x[2]
        s[5,1,0]=(-0.64289167402)*x[0]**o+(-0.772752345352)*x[0]+(0.900114631593)*x[1]**o+(0.510514117797)*x[1]+(-0.984639269239)*x[2]**o+(0.243489379424)*x[2]
        s[5,1,1]=(-0.551740975948)*x[0]**o+(-0.645067068599)*x[0]+(-0.238391882646)*x[1]**o+(-0.959940660719)*x[1]+(-0.020674238581)*x[2]**o+(0.0667084698715)*x[2]
        ref[0,0,0]=(0.49027687523)*x_ref[0]**o+(-0.945396297892)*x_ref[0]+(-0.649119833947)*x_ref[1]**o+(-0.193510024763)*x_ref[1]+(0.672065654766)*x_ref[2]**o+(0.396941813883)*x_ref[2]
        ref[0,0,1]=(-0.379608112905)*x_ref[0]**o+(-0.321849772246)*x_ref[0]+(-0.695093455872)*x_ref[1]**o+(-0.114263310683)*x_ref[1]+(0.425556105367)*x_ref[2]**o+(0.737954023041)*x_ref[2]
        ref[0,1,0]=(0.597295505367)*x_ref[0]**o+(0.285081105939)*x_ref[0]+(-0.690746281653)*x_ref[1]**o+(0.431456884332)*x_ref[1]+(-0.569785421553)*x_ref[2]**o+(0.218907517674)*x_ref[2]
        ref[0,1,1]=(-0.743660806659)*x_ref[0]**o+(0.916636055834)*x_ref[0]+(0.747852198685)*x_ref[1]**o+(-0.304767976464)*x_ref[1]+(0.737051048825)*x_ref[2]**o+(0.108795699665)*x_ref[2]
        ref[1,0,0]=(0.0322299276093)*x_ref[0]**o+(-0.507367087704)*x_ref[0]+(-0.154613332688)*x_ref[1]**o+(-0.642692955185)*x_ref[1]+(0.921618425322)*x_ref[2]**o+(-0.89234021469)*x_ref[2]
        ref[1,0,1]=(0.250927374346)*x_ref[0]**o+(-0.713045403997)*x_ref[0]+(0.0229252996751)*x_ref[1]**o+(0.251843016211)*x_ref[1]+(0.683324421319)*x_ref[2]**o+(0.466509616753)*x_ref[2]
        ref[1,1,0]=(0.610653309325)*x_ref[0]**o+(0.503568507563)*x_ref[0]+(0.519292621274)*x_ref[1]**o+(-0.452962008782)*x_ref[1]+(0.908988844098)*x_ref[2]**o+(-0.933406144464)*x_ref[2]
        ref[1,1,1]=(0.271788487178)*x_ref[0]**o+(0.341989918651)*x_ref[0]+(-0.406971961331)*x_ref[1]**o+(-0.928647769169)*x_ref[1]+(0.662902796818)*x_ref[2]**o+(0.655876724787)*x_ref[2]
        ref[2,0,0]=(-0.775353204706)*x_ref[0]**o+(0.278726576715)*x_ref[0]+(0.448754566832)*x_ref[1]**o+(-0.943150229959)*x_ref[1]+(0.0806002558903)*x_ref[2]**o+(-0.350491928882)*x_ref[2]
        ref[2,0,1]=(-0.899719123904)*x_ref[0]**o+(-0.89394097319)*x_ref[0]+(0.136672330017)*x_ref[1]**o+(0.941198317525)*x_ref[1]+(0.945970508521)*x_ref[2]**o+(0.232574429377)*x_ref[2]
        ref[2,1,0]=(0.102227090178)*x_ref[0]**o+(0.498258334271)*x_ref[0]+(-0.369219100853)*x_ref[1]**o+(0.965612831948)*x_ref[1]+(-0.580434278342)*x_ref[2]**o+(0.0747696579664)*x_ref[2]
        ref[2,1,1]=(0.0680171783989)*x_ref[0]**o+(-0.655506093893)*x_ref[0]+(0.0854275560955)*x_ref[1]**o+(0.788063374419)*x_ref[1]+(0.3912557517)*x_ref[2]**o+(-0.911795675902)*x_ref[2]
        ref[3,0,0]=(0.941143798566)*x_ref[0]**o+(0.324489623574)*x_ref[0]+(-0.259012903836)*x_ref[1]**o+(-0.779051329011)*x_ref[1]+(0.673245283112)*x_ref[2]**o+(-0.888729936858)*x_ref[2]
        ref[3,0,1]=(0.904124513313)*x_ref[0]**o+(0.564171149844)*x_ref[0]+(0.799169083649)*x_ref[1]**o+(-0.172322806388)*x_ref[1]+(-0.0754065342235)*x_ref[2]**o+(0.0419393221333)*x_ref[2]
        ref[3,1,0]=(-0.200531316713)*x_ref[0]**o+(0.786777777323)*x_ref[0]+(-0.222824591798)*x_ref[1]**o+(0.32280872616)*x_ref[1]+(-0.418791963436)*x_ref[2]**o+(-0.549019641386)*x_ref[2]
        ref[3,1,1]=(0.0519992199857)*x_ref[0]**o+(0.278310751677)*x_ref[0]+(0.590847528061)*x_ref[1]**o+(0.258564224896)*x_ref[1]+(-0.500091685923)*x_ref[2]**o+(0.511781801585)*x_ref[2]
        ref[4,0,0]=(0.459040882538)*x_ref[0]**o+(-0.481335872602)*x_ref[0]+(0.0556391566559)*x_ref[1]**o+(-0.208836885623)*x_ref[1]+(-0.11717985744)*x_ref[2]**o+(-0.544122946452)*x_ref[2]
        ref[4,0,1]=(-0.255258906635)*x_ref[0]**o+(0.860765893836)*x_ref[0]+(0.389276109422)*x_ref[1]**o+(-0.999502217953)*x_ref[1]+(-0.715269648526)*x_ref[2]**o+(0.451562561245)*x_ref[2]
        ref[4,1,0]=(-0.579717249622)*x_ref[0]**o+(0.984974487836)*x_ref[0]+(-0.803917364532)*x_ref[1]**o+(-0.764878964835)*x_ref[1]+(-0.0642799485519)*x_ref[2]**o+(-0.917022175106)*x_ref[2]
        ref[4,1,1]=(-0.873863405322)*x_ref[0]**o+(-0.872375014129)*x_ref[0]+(0.981355044198)*x_ref[1]**o+(0.40407870155)*x_ref[1]+(0.796125441505)*x_ref[2]**o+(0.486002048777)*x_ref[2]
        ref[5,0,0]=(0.92716952184)*x_ref[0]**o+(-0.280973691624)*x_ref[0]+(0.815340507346)*x_ref[1]**o+(-0.555904571967)*x_ref[1]+(0.765200732877)*x_ref[2]**o+(-0.48251397176)*x_ref[2]
        ref[5,0,1]=(-0.6976034156)*x_ref[0]**o+(-0.342128490335)*x_ref[0]+(-0.273657764836)*x_ref[1]**o+(-0.393975979448)*x_ref[1]+(0.522428789629)*x_ref[2]**o+(-0.210033793616)*x_ref[2]
        ref[5,1,0]=(-0.64289167402)*x_ref[0]**o+(-0.772752345352)*x_ref[0]+(0.900114631593)*x_ref[1]**o+(0.510514117797)*x_ref[1]+(-0.984639269239)*x_ref[2]**o+(0.243489379424)*x_ref[2]
        ref[5,1,1]=(-0.551740975948)*x_ref[0]**o+(-0.645067068599)*x_ref[0]+(-0.238391882646)*x_ref[1]**o+(-0.959940660719)*x_ref[1]+(-0.020674238581)*x_ref[2]**o+(0.0667084698715)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.711963461677)*x[0]**o+(0.542670298734)*x[0]+(0.28075382249)*x[1]**o+(0.667437625224)*x[1]
        s[0,0,0,1]=(0.573441578286)*x[0]**o+(0.506095821827)*x[0]+(0.150993391009)*x[1]**o+(0.526409669633)*x[1]
        s[0,0,1,0]=(0.706627280956)*x[0]**o+(0.57780902263)*x[0]+(0.270996212375)*x[1]**o+(0.938868396109)*x[1]
        s[0,0,1,1]=(0.970926483868)*x[0]**o+(0.0967183437714)*x[0]+(0.305310109546)*x[1]**o+(-0.746529372684)*x[1]
        s[0,0,2,0]=(-0.0598871068472)*x[0]**o+(0.441097065872)*x[0]+(0.240831657822)*x[1]**o+(-0.781319614159)*x[1]
        s[0,0,2,1]=(0.0738716023753)*x[0]**o+(-0.280473432627)*x[0]+(0.351688231305)*x[1]**o+(0.287190769396)*x[1]
        s[0,1,0,0]=(-0.662029332598)*x[0]**o+(0.769980048213)*x[0]+(-0.316622870493)*x[1]**o+(0.756839554902)*x[1]
        s[0,1,0,1]=(-0.931130133674)*x[0]**o+(-0.70789066479)*x[0]+(0.350649077237)*x[1]**o+(-0.340577841771)*x[1]
        s[0,1,1,0]=(0.541533987605)*x[0]**o+(-0.871063950931)*x[0]+(0.198020061012)*x[1]**o+(0.802998878081)*x[1]
        s[0,1,1,1]=(0.264037408365)*x[0]**o+(-0.0557314225594)*x[0]+(-0.684018523812)*x[1]**o+(-0.229282658383)*x[1]
        s[0,1,2,0]=(0.637810239569)*x[0]**o+(0.552431819202)*x[0]+(0.0939985703867)*x[1]**o+(0.828751113775)*x[1]
        s[0,1,2,1]=(-0.374671159094)*x[0]**o+(0.664088936153)*x[0]+(-0.345847704492)*x[1]**o+(-0.364075081521)*x[1]
        s[0,2,0,0]=(0.00895781289078)*x[0]**o+(0.399614286422)*x[0]+(0.341683450789)*x[1]**o+(0.167206336911)*x[1]
        s[0,2,0,1]=(-0.104604431902)*x[0]**o+(-0.659107995643)*x[0]+(0.680691727783)*x[1]**o+(-0.684349104067)*x[1]
        s[0,2,1,0]=(0.710804511082)*x[0]**o+(-0.972641516063)*x[0]+(0.330095093159)*x[1]**o+(0.937337443354)*x[1]
        s[0,2,1,1]=(-0.294227883996)*x[0]**o+(0.965392046246)*x[0]+(0.714517712165)*x[1]**o+(0.449886137172)*x[1]
        s[0,2,2,0]=(-0.825491552072)*x[0]**o+(-0.638472841016)*x[0]+(-0.191336005365)*x[1]**o+(-0.422472261204)*x[1]
        s[0,2,2,1]=(-0.23092704399)*x[0]**o+(-0.637860425921)*x[0]+(0.840387209559)*x[1]**o+(0.165561089405)*x[1]
        s[0,3,0,0]=(0.420745779316)*x[0]**o+(-0.311588614725)*x[0]+(-0.553702597164)*x[1]**o+(0.272368731382)*x[1]
        s[0,3,0,1]=(0.327243090728)*x[0]**o+(-0.386774129727)*x[0]+(-0.950289816609)*x[1]**o+(0.662365527837)*x[1]
        s[0,3,1,0]=(-0.889641002357)*x[0]**o+(-0.195916682305)*x[0]+(-0.60085377073)*x[1]**o+(-0.191887529995)*x[1]
        s[0,3,1,1]=(0.987376581443)*x[0]**o+(-0.0350561501803)*x[0]+(0.704905530249)*x[1]**o+(-0.735687404507)*x[1]
        s[0,3,2,0]=(0.420809308115)*x[0]**o+(-0.339666948082)*x[0]+(0.996296976276)*x[1]**o+(0.506174123674)*x[1]
        s[0,3,2,1]=(0.426714798574)*x[0]**o+(0.620612017549)*x[0]+(0.60846811755)*x[1]**o+(0.561758397959)*x[1]
        s[0,4,0,0]=(0.0193321433149)*x[0]**o+(0.395174248733)*x[0]+(-0.631710108856)*x[1]**o+(0.562496027755)*x[1]
        s[0,4,0,1]=(0.936397656738)*x[0]**o+(-0.620128461169)*x[0]+(0.424872968744)*x[1]**o+(0.936360345969)*x[1]
        s[0,4,1,0]=(-0.823655625312)*x[0]**o+(0.110021832964)*x[0]+(-0.480484367041)*x[1]**o+(0.613626788942)*x[1]
        s[0,4,1,1]=(0.516537102221)*x[0]**o+(-0.465693898711)*x[0]+(-0.0649481633029)*x[1]**o+(0.498343667101)*x[1]
        s[0,4,2,0]=(0.364997630891)*x[0]**o+(0.726894447639)*x[0]+(0.729067579856)*x[1]**o+(-0.786376612921)*x[1]
        s[0,4,2,1]=(-0.574801923675)*x[0]**o+(-0.769120062116)*x[0]+(0.0802603764785)*x[1]**o+(0.322665268053)*x[1]
        s[1,0,0,0]=(-0.497916263484)*x[0]**o+(0.0689477158621)*x[0]+(-0.59065778662)*x[1]**o+(0.620475916092)*x[1]
        s[1,0,0,1]=(0.928072689885)*x[0]**o+(0.550220523492)*x[0]+(0.384056461821)*x[1]**o+(0.991195851819)*x[1]
        s[1,0,1,0]=(-0.171652790551)*x[0]**o+(-0.310223770891)*x[0]+(0.505801949231)*x[1]**o+(-0.809872602671)*x[1]
        s[1,0,1,1]=(-0.843616998175)*x[0]**o+(-0.514683249854)*x[0]+(0.22719571088)*x[1]**o+(-0.801964932709)*x[1]
        s[1,0,2,0]=(-0.816576821632)*x[0]**o+(-0.629668603878)*x[0]+(-0.296562767682)*x[1]**o+(-0.0121908797722)*x[1]
        s[1,0,2,1]=(-0.816221470582)*x[0]**o+(0.101678348599)*x[0]+(0.0841692807823)*x[1]**o+(0.265517698128)*x[1]
        s[1,1,0,0]=(0.74252349936)*x[0]**o+(0.488296875375)*x[0]+(0.553091389727)*x[1]**o+(0.0951295346041)*x[1]
        s[1,1,0,1]=(0.573858954955)*x[0]**o+(-0.0210888969717)*x[0]+(-0.813464794029)*x[1]**o+(0.760502343882)*x[1]
        s[1,1,1,0]=(0.806002928884)*x[0]**o+(0.191962011356)*x[0]+(0.788308484735)*x[1]**o+(0.325349383493)*x[1]
        s[1,1,1,1]=(-0.975008178175)*x[0]**o+(-0.525299302456)*x[0]+(-0.718959301598)*x[1]**o+(0.821616294375)*x[1]
        s[1,1,2,0]=(-0.441476557684)*x[0]**o+(0.442725962693)*x[0]+(0.605338741812)*x[1]**o+(-0.0356460149592)*x[1]
        s[1,1,2,1]=(0.75296734605)*x[0]**o+(0.0218514182228)*x[0]+(-0.408708955555)*x[1]**o+(-0.729191065301)*x[1]
        s[1,2,0,0]=(0.24165794152)*x[0]**o+(-0.33924002011)*x[0]+(0.357553913734)*x[1]**o+(0.166312753501)*x[1]
        s[1,2,0,1]=(0.835534214783)*x[0]**o+(-0.688462973774)*x[0]+(0.779388430668)*x[1]**o+(-0.869108490882)*x[1]
        s[1,2,1,0]=(0.172345662475)*x[0]**o+(0.323742121947)*x[0]+(-0.296938383275)*x[1]**o+(0.0359645605525)*x[1]
        s[1,2,1,1]=(-0.596600537083)*x[0]**o+(0.644152951661)*x[0]+(0.609094045392)*x[1]**o+(-0.312234180863)*x[1]
        s[1,2,2,0]=(0.484589912205)*x[0]**o+(-0.906567695494)*x[0]+(0.811265520788)*x[1]**o+(0.444799696239)*x[1]
        s[1,2,2,1]=(0.0254720755485)*x[0]**o+(-0.41174472105)*x[0]+(0.90099646911)*x[1]**o+(-0.33088554417)*x[1]
        s[1,3,0,0]=(-0.281155224735)*x[0]**o+(0.274137552899)*x[0]+(0.0974242478848)*x[1]**o+(-0.995002288445)*x[1]
        s[1,3,0,1]=(0.907653805131)*x[0]**o+(0.743984237516)*x[0]+(-0.776028924136)*x[1]**o+(0.209784894077)*x[1]
        s[1,3,1,0]=(0.0470679202395)*x[0]**o+(-0.809937294063)*x[0]+(0.956184337312)*x[1]**o+(-0.539155228812)*x[1]
        s[1,3,1,1]=(0.720979862876)*x[0]**o+(0.412993117346)*x[0]+(0.259998599343)*x[1]**o+(-0.739191117502)*x[1]
        s[1,3,2,0]=(0.403493857435)*x[0]**o+(0.0582403515426)*x[0]+(0.0311147864546)*x[1]**o+(0.397740429479)*x[1]
        s[1,3,2,1]=(0.51510647545)*x[0]**o+(0.179454689412)*x[0]+(-0.864589732719)*x[1]**o+(-0.919542815326)*x[1]
        s[1,4,0,0]=(0.176385360146)*x[0]**o+(0.666029075488)*x[0]+(-0.765133182482)*x[1]**o+(0.506226091771)*x[1]
        s[1,4,0,1]=(-0.217264015218)*x[0]**o+(0.997799494004)*x[0]+(0.706270854724)*x[1]**o+(-0.440854336576)*x[1]
        s[1,4,1,0]=(-0.127481889699)*x[0]**o+(-0.959927193428)*x[0]+(-0.314190555296)*x[1]**o+(-0.0254770240779)*x[1]
        s[1,4,1,1]=(0.943927827542)*x[0]**o+(0.695303968245)*x[0]+(0.537165245123)*x[1]**o+(-0.414505300627)*x[1]
        s[1,4,2,0]=(-0.583859597164)*x[0]**o+(-0.71993694674)*x[0]+(0.0390198271077)*x[1]**o+(0.917816363923)*x[1]
        s[1,4,2,1]=(0.752163371766)*x[0]**o+(0.47299437482)*x[0]+(-0.499822965518)*x[1]**o+(-0.629775835404)*x[1]
        s[2,0,0,0]=(0.441638332575)*x[0]**o+(0.183300755854)*x[0]+(0.635094026763)*x[1]**o+(-0.981751131017)*x[1]
        s[2,0,0,1]=(0.564782099962)*x[0]**o+(-0.00355099531182)*x[0]+(-0.309065488817)*x[1]**o+(-0.805388517953)*x[1]
        s[2,0,1,0]=(-0.152274122381)*x[0]**o+(-0.884657204578)*x[0]+(-0.362170305747)*x[1]**o+(0.688413345566)*x[1]
        s[2,0,1,1]=(0.433159898578)*x[0]**o+(-0.926183453021)*x[0]+(-0.990900351089)*x[1]**o+(-0.558683257136)*x[1]
        s[2,0,2,0]=(0.378679325956)*x[0]**o+(-0.178401685533)*x[0]+(0.294582559995)*x[1]**o+(0.324129328685)*x[1]
        s[2,0,2,1]=(0.191782222208)*x[0]**o+(-0.718853143054)*x[0]+(-0.1561779489)*x[1]**o+(0.122297214902)*x[1]
        s[2,1,0,0]=(-0.745888747317)*x[0]**o+(0.124662622215)*x[0]+(0.61762073845)*x[1]**o+(-0.543673681976)*x[1]
        s[2,1,0,1]=(0.696268656575)*x[0]**o+(0.879682905689)*x[0]+(0.101543002235)*x[1]**o+(0.795349066813)*x[1]
        s[2,1,1,0]=(-0.595042825753)*x[0]**o+(0.244946186339)*x[0]+(0.107868704293)*x[1]**o+(0.20527282324)*x[1]
        s[2,1,1,1]=(0.702092863468)*x[0]**o+(-0.95898866728)*x[0]+(0.415299739291)*x[1]**o+(-0.602967505073)*x[1]
        s[2,1,2,0]=(0.242441579412)*x[0]**o+(-0.621648829236)*x[0]+(0.00279977134428)*x[1]**o+(0.20444394565)*x[1]
        s[2,1,2,1]=(-0.248160979861)*x[0]**o+(0.510934355012)*x[0]+(-0.888013212236)*x[1]**o+(0.348960154636)*x[1]
        s[2,2,0,0]=(-0.484727669828)*x[0]**o+(0.23183410781)*x[0]+(0.771832910785)*x[1]**o+(-0.259023608026)*x[1]
        s[2,2,0,1]=(-0.0293853663853)*x[0]**o+(0.132402850017)*x[0]+(0.460425782419)*x[1]**o+(0.918697618402)*x[1]
        s[2,2,1,0]=(-0.960506935656)*x[0]**o+(0.302982411255)*x[0]+(0.927623802726)*x[1]**o+(0.446529083445)*x[1]
        s[2,2,1,1]=(-0.0791528123579)*x[0]**o+(-0.963662594671)*x[0]+(0.445231487496)*x[1]**o+(-0.635556750912)*x[1]
        s[2,2,2,0]=(0.119522639306)*x[0]**o+(0.999196602488)*x[0]+(0.75598380813)*x[1]**o+(0.159012027458)*x[1]
        s[2,2,2,1]=(0.671543289198)*x[0]**o+(-0.897542325495)*x[0]+(0.184767661479)*x[1]**o+(-0.679101227287)*x[1]
        s[2,3,0,0]=(0.705751399559)*x[0]**o+(0.0343552696754)*x[0]+(0.295737350313)*x[1]**o+(0.603031693221)*x[1]
        s[2,3,0,1]=(-0.568041111686)*x[0]**o+(0.795100811027)*x[0]+(-0.853308139159)*x[1]**o+(-0.198019398328)*x[1]
        s[2,3,1,0]=(-0.568416608732)*x[0]**o+(-0.785696005424)*x[0]+(0.734563901681)*x[1]**o+(0.337683654534)*x[1]
        s[2,3,1,1]=(0.750616777015)*x[0]**o+(-0.243919199772)*x[0]+(0.47894230818)*x[1]**o+(0.117853316761)*x[1]
        s[2,3,2,0]=(0.595655270733)*x[0]**o+(-0.131631796604)*x[0]+(-0.363246505069)*x[1]**o+(0.0399311321376)*x[1]
        s[2,3,2,1]=(-0.871124094437)*x[0]**o+(-0.495878262772)*x[0]+(-0.59370100135)*x[1]**o+(0.745382881709)*x[1]
        s[2,4,0,0]=(0.479665892387)*x[0]**o+(-0.393886135683)*x[0]+(-0.969679118858)*x[1]**o+(-0.720467296702)*x[1]
        s[2,4,0,1]=(0.111063753944)*x[0]**o+(-0.759833999129)*x[0]+(-0.123061597258)*x[1]**o+(-0.982096877401)*x[1]
        s[2,4,1,0]=(-0.901470811302)*x[0]**o+(0.395087698063)*x[0]+(0.553738458787)*x[1]**o+(-0.112328034611)*x[1]
        s[2,4,1,1]=(0.39514286938)*x[0]**o+(-0.923250016486)*x[0]+(0.361513401174)*x[1]**o+(-0.272660910785)*x[1]
        s[2,4,2,0]=(0.454263938343)*x[0]**o+(-0.0921743296726)*x[0]+(-0.867658322099)*x[1]**o+(-0.469179181142)*x[1]
        s[2,4,2,1]=(0.00363399429065)*x[0]**o+(0.295615173633)*x[0]+(-0.619556300313)*x[1]**o+(0.385912799652)*x[1]
        s[3,0,0,0]=(-0.176638548482)*x[0]**o+(0.716817113557)*x[0]+(0.713969909735)*x[1]**o+(-0.315381536549)*x[1]
        s[3,0,0,1]=(-0.981761588756)*x[0]**o+(-0.264195260433)*x[0]+(0.514278247605)*x[1]**o+(0.620939057298)*x[1]
        s[3,0,1,0]=(-0.599489307325)*x[0]**o+(0.213895727472)*x[0]+(0.986896902529)*x[1]**o+(0.423516351738)*x[1]
        s[3,0,1,1]=(0.282281951751)*x[0]**o+(0.198509073556)*x[0]+(0.261590765595)*x[1]**o+(0.753982289812)*x[1]
        s[3,0,2,0]=(0.78593458655)*x[0]**o+(-0.52358431526)*x[0]+(0.479533127253)*x[1]**o+(0.120268547872)*x[1]
        s[3,0,2,1]=(0.229057209441)*x[0]**o+(-0.200667079463)*x[0]+(0.713824885941)*x[1]**o+(-0.317631878875)*x[1]
        s[3,1,0,0]=(-0.621475342986)*x[0]**o+(-0.776830750036)*x[0]+(0.74327888403)*x[1]**o+(0.77144000144)*x[1]
        s[3,1,0,1]=(-0.0701637401121)*x[0]**o+(0.315253792233)*x[0]+(0.714679121123)*x[1]**o+(-0.124524161083)*x[1]
        s[3,1,1,0]=(0.58339165807)*x[0]**o+(0.754756142322)*x[0]+(-0.552898566299)*x[1]**o+(0.0401072453256)*x[1]
        s[3,1,1,1]=(-0.704826353788)*x[0]**o+(-0.850442055451)*x[0]+(0.0185807835768)*x[1]**o+(0.798634373727)*x[1]
        s[3,1,2,0]=(-0.304968678423)*x[0]**o+(0.882631700011)*x[0]+(-0.00670582763206)*x[1]**o+(-0.790997780888)*x[1]
        s[3,1,2,1]=(0.978052946899)*x[0]**o+(0.0822751454714)*x[0]+(-0.141545511663)*x[1]**o+(0.804142415097)*x[1]
        s[3,2,0,0]=(-0.233834449356)*x[0]**o+(-0.349324841655)*x[0]+(0.633317011256)*x[1]**o+(-0.682052363655)*x[1]
        s[3,2,0,1]=(-0.828772677459)*x[0]**o+(-0.407293570997)*x[0]+(0.0378365988776)*x[1]**o+(0.575150943785)*x[1]
        s[3,2,1,0]=(0.00530675292728)*x[0]**o+(-0.583321711314)*x[0]+(-0.461755026035)*x[1]**o+(-0.234209819371)*x[1]
        s[3,2,1,1]=(-0.529179272486)*x[0]**o+(-0.126586669403)*x[0]+(0.199780481034)*x[1]**o+(-0.475384002433)*x[1]
        s[3,2,2,0]=(-0.794495647543)*x[0]**o+(0.586459171803)*x[0]+(-0.417091337526)*x[1]**o+(-0.920839445366)*x[1]
        s[3,2,2,1]=(-0.543510366902)*x[0]**o+(0.9390916149)*x[0]+(-0.378905656214)*x[1]**o+(-0.580144827894)*x[1]
        s[3,3,0,0]=(-0.218318771891)*x[0]**o+(-0.741715169744)*x[0]+(0.471332960959)*x[1]**o+(-0.942331201293)*x[1]
        s[3,3,0,1]=(-0.639400890732)*x[0]**o+(-0.35086784511)*x[0]+(-0.662448518836)*x[1]**o+(-0.109260566936)*x[1]
        s[3,3,1,0]=(0.00137953261949)*x[0]**o+(-0.527278610973)*x[0]+(0.547316965109)*x[1]**o+(0.747307046881)*x[1]
        s[3,3,1,1]=(-0.110490640565)*x[0]**o+(0.30599876694)*x[0]+(0.464447167957)*x[1]**o+(-0.650215871098)*x[1]
        s[3,3,2,0]=(-0.0752004672453)*x[0]**o+(-0.476012931477)*x[0]+(-0.463766892732)*x[1]**o+(-0.864312392032)*x[1]
        s[3,3,2,1]=(-0.614219705651)*x[0]**o+(0.579350508101)*x[0]+(-0.307032385253)*x[1]**o+(0.609948162009)*x[1]
        s[3,4,0,0]=(-0.946145025239)*x[0]**o+(0.383874203169)*x[0]+(-0.128251366911)*x[1]**o+(-0.592420767396)*x[1]
        s[3,4,0,1]=(0.0656946533978)*x[0]**o+(-0.400244444789)*x[0]+(-0.263415867899)*x[1]**o+(-0.241065146194)*x[1]
        s[3,4,1,0]=(-0.44363374655)*x[0]**o+(-0.702360780698)*x[0]+(0.39242809706)*x[1]**o+(-0.499711507536)*x[1]
        s[3,4,1,1]=(-0.790376230857)*x[0]**o+(0.271619243859)*x[0]+(0.221426582034)*x[1]**o+(-0.244571958075)*x[1]
        s[3,4,2,0]=(-0.599565364787)*x[0]**o+(0.0637522137783)*x[0]+(0.362816245637)*x[1]**o+(-0.52786138669)*x[1]
        s[3,4,2,1]=(0.955472544947)*x[0]**o+(-0.368132248377)*x[0]+(0.262396158645)*x[1]**o+(-0.252857678444)*x[1]
        ref[0,0,0,0]=(-0.711963461677)*x_ref[0]**o+(0.542670298734)*x_ref[0]+(0.28075382249)*x_ref[1]**o+(0.667437625224)*x_ref[1]
        ref[0,0,0,1]=(0.573441578286)*x_ref[0]**o+(0.506095821827)*x_ref[0]+(0.150993391009)*x_ref[1]**o+(0.526409669633)*x_ref[1]
        ref[0,0,1,0]=(0.706627280956)*x_ref[0]**o+(0.57780902263)*x_ref[0]+(0.270996212375)*x_ref[1]**o+(0.938868396109)*x_ref[1]
        ref[0,0,1,1]=(0.970926483868)*x_ref[0]**o+(0.0967183437714)*x_ref[0]+(0.305310109546)*x_ref[1]**o+(-0.746529372684)*x_ref[1]
        ref[0,0,2,0]=(-0.0598871068472)*x_ref[0]**o+(0.441097065872)*x_ref[0]+(0.240831657822)*x_ref[1]**o+(-0.781319614159)*x_ref[1]
        ref[0,0,2,1]=(0.0738716023753)*x_ref[0]**o+(-0.280473432627)*x_ref[0]+(0.351688231305)*x_ref[1]**o+(0.287190769396)*x_ref[1]
        ref[0,1,0,0]=(-0.662029332598)*x_ref[0]**o+(0.769980048213)*x_ref[0]+(-0.316622870493)*x_ref[1]**o+(0.756839554902)*x_ref[1]
        ref[0,1,0,1]=(-0.931130133674)*x_ref[0]**o+(-0.70789066479)*x_ref[0]+(0.350649077237)*x_ref[1]**o+(-0.340577841771)*x_ref[1]
        ref[0,1,1,0]=(0.541533987605)*x_ref[0]**o+(-0.871063950931)*x_ref[0]+(0.198020061012)*x_ref[1]**o+(0.802998878081)*x_ref[1]
        ref[0,1,1,1]=(0.264037408365)*x_ref[0]**o+(-0.0557314225594)*x_ref[0]+(-0.684018523812)*x_ref[1]**o+(-0.229282658383)*x_ref[1]
        ref[0,1,2,0]=(0.637810239569)*x_ref[0]**o+(0.552431819202)*x_ref[0]+(0.0939985703867)*x_ref[1]**o+(0.828751113775)*x_ref[1]
        ref[0,1,2,1]=(-0.374671159094)*x_ref[0]**o+(0.664088936153)*x_ref[0]+(-0.345847704492)*x_ref[1]**o+(-0.364075081521)*x_ref[1]
        ref[0,2,0,0]=(0.00895781289078)*x_ref[0]**o+(0.399614286422)*x_ref[0]+(0.341683450789)*x_ref[1]**o+(0.167206336911)*x_ref[1]
        ref[0,2,0,1]=(-0.104604431902)*x_ref[0]**o+(-0.659107995643)*x_ref[0]+(0.680691727783)*x_ref[1]**o+(-0.684349104067)*x_ref[1]
        ref[0,2,1,0]=(0.710804511082)*x_ref[0]**o+(-0.972641516063)*x_ref[0]+(0.330095093159)*x_ref[1]**o+(0.937337443354)*x_ref[1]
        ref[0,2,1,1]=(-0.294227883996)*x_ref[0]**o+(0.965392046246)*x_ref[0]+(0.714517712165)*x_ref[1]**o+(0.449886137172)*x_ref[1]
        ref[0,2,2,0]=(-0.825491552072)*x_ref[0]**o+(-0.638472841016)*x_ref[0]+(-0.191336005365)*x_ref[1]**o+(-0.422472261204)*x_ref[1]
        ref[0,2,2,1]=(-0.23092704399)*x_ref[0]**o+(-0.637860425921)*x_ref[0]+(0.840387209559)*x_ref[1]**o+(0.165561089405)*x_ref[1]
        ref[0,3,0,0]=(0.420745779316)*x_ref[0]**o+(-0.311588614725)*x_ref[0]+(-0.553702597164)*x_ref[1]**o+(0.272368731382)*x_ref[1]
        ref[0,3,0,1]=(0.327243090728)*x_ref[0]**o+(-0.386774129727)*x_ref[0]+(-0.950289816609)*x_ref[1]**o+(0.662365527837)*x_ref[1]
        ref[0,3,1,0]=(-0.889641002357)*x_ref[0]**o+(-0.195916682305)*x_ref[0]+(-0.60085377073)*x_ref[1]**o+(-0.191887529995)*x_ref[1]
        ref[0,3,1,1]=(0.987376581443)*x_ref[0]**o+(-0.0350561501803)*x_ref[0]+(0.704905530249)*x_ref[1]**o+(-0.735687404507)*x_ref[1]
        ref[0,3,2,0]=(0.420809308115)*x_ref[0]**o+(-0.339666948082)*x_ref[0]+(0.996296976276)*x_ref[1]**o+(0.506174123674)*x_ref[1]
        ref[0,3,2,1]=(0.426714798574)*x_ref[0]**o+(0.620612017549)*x_ref[0]+(0.60846811755)*x_ref[1]**o+(0.561758397959)*x_ref[1]
        ref[0,4,0,0]=(0.0193321433149)*x_ref[0]**o+(0.395174248733)*x_ref[0]+(-0.631710108856)*x_ref[1]**o+(0.562496027755)*x_ref[1]
        ref[0,4,0,1]=(0.936397656738)*x_ref[0]**o+(-0.620128461169)*x_ref[0]+(0.424872968744)*x_ref[1]**o+(0.936360345969)*x_ref[1]
        ref[0,4,1,0]=(-0.823655625312)*x_ref[0]**o+(0.110021832964)*x_ref[0]+(-0.480484367041)*x_ref[1]**o+(0.613626788942)*x_ref[1]
        ref[0,4,1,1]=(0.516537102221)*x_ref[0]**o+(-0.465693898711)*x_ref[0]+(-0.0649481633029)*x_ref[1]**o+(0.498343667101)*x_ref[1]
        ref[0,4,2,0]=(0.364997630891)*x_ref[0]**o+(0.726894447639)*x_ref[0]+(0.729067579856)*x_ref[1]**o+(-0.786376612921)*x_ref[1]
        ref[0,4,2,1]=(-0.574801923675)*x_ref[0]**o+(-0.769120062116)*x_ref[0]+(0.0802603764785)*x_ref[1]**o+(0.322665268053)*x_ref[1]
        ref[1,0,0,0]=(-0.497916263484)*x_ref[0]**o+(0.0689477158621)*x_ref[0]+(-0.59065778662)*x_ref[1]**o+(0.620475916092)*x_ref[1]
        ref[1,0,0,1]=(0.928072689885)*x_ref[0]**o+(0.550220523492)*x_ref[0]+(0.384056461821)*x_ref[1]**o+(0.991195851819)*x_ref[1]
        ref[1,0,1,0]=(-0.171652790551)*x_ref[0]**o+(-0.310223770891)*x_ref[0]+(0.505801949231)*x_ref[1]**o+(-0.809872602671)*x_ref[1]
        ref[1,0,1,1]=(-0.843616998175)*x_ref[0]**o+(-0.514683249854)*x_ref[0]+(0.22719571088)*x_ref[1]**o+(-0.801964932709)*x_ref[1]
        ref[1,0,2,0]=(-0.816576821632)*x_ref[0]**o+(-0.629668603878)*x_ref[0]+(-0.296562767682)*x_ref[1]**o+(-0.0121908797722)*x_ref[1]
        ref[1,0,2,1]=(-0.816221470582)*x_ref[0]**o+(0.101678348599)*x_ref[0]+(0.0841692807823)*x_ref[1]**o+(0.265517698128)*x_ref[1]
        ref[1,1,0,0]=(0.74252349936)*x_ref[0]**o+(0.488296875375)*x_ref[0]+(0.553091389727)*x_ref[1]**o+(0.0951295346041)*x_ref[1]
        ref[1,1,0,1]=(0.573858954955)*x_ref[0]**o+(-0.0210888969717)*x_ref[0]+(-0.813464794029)*x_ref[1]**o+(0.760502343882)*x_ref[1]
        ref[1,1,1,0]=(0.806002928884)*x_ref[0]**o+(0.191962011356)*x_ref[0]+(0.788308484735)*x_ref[1]**o+(0.325349383493)*x_ref[1]
        ref[1,1,1,1]=(-0.975008178175)*x_ref[0]**o+(-0.525299302456)*x_ref[0]+(-0.718959301598)*x_ref[1]**o+(0.821616294375)*x_ref[1]
        ref[1,1,2,0]=(-0.441476557684)*x_ref[0]**o+(0.442725962693)*x_ref[0]+(0.605338741812)*x_ref[1]**o+(-0.0356460149592)*x_ref[1]
        ref[1,1,2,1]=(0.75296734605)*x_ref[0]**o+(0.0218514182228)*x_ref[0]+(-0.408708955555)*x_ref[1]**o+(-0.729191065301)*x_ref[1]
        ref[1,2,0,0]=(0.24165794152)*x_ref[0]**o+(-0.33924002011)*x_ref[0]+(0.357553913734)*x_ref[1]**o+(0.166312753501)*x_ref[1]
        ref[1,2,0,1]=(0.835534214783)*x_ref[0]**o+(-0.688462973774)*x_ref[0]+(0.779388430668)*x_ref[1]**o+(-0.869108490882)*x_ref[1]
        ref[1,2,1,0]=(0.172345662475)*x_ref[0]**o+(0.323742121947)*x_ref[0]+(-0.296938383275)*x_ref[1]**o+(0.0359645605525)*x_ref[1]
        ref[1,2,1,1]=(-0.596600537083)*x_ref[0]**o+(0.644152951661)*x_ref[0]+(0.609094045392)*x_ref[1]**o+(-0.312234180863)*x_ref[1]
        ref[1,2,2,0]=(0.484589912205)*x_ref[0]**o+(-0.906567695494)*x_ref[0]+(0.811265520788)*x_ref[1]**o+(0.444799696239)*x_ref[1]
        ref[1,2,2,1]=(0.0254720755485)*x_ref[0]**o+(-0.41174472105)*x_ref[0]+(0.90099646911)*x_ref[1]**o+(-0.33088554417)*x_ref[1]
        ref[1,3,0,0]=(-0.281155224735)*x_ref[0]**o+(0.274137552899)*x_ref[0]+(0.0974242478848)*x_ref[1]**o+(-0.995002288445)*x_ref[1]
        ref[1,3,0,1]=(0.907653805131)*x_ref[0]**o+(0.743984237516)*x_ref[0]+(-0.776028924136)*x_ref[1]**o+(0.209784894077)*x_ref[1]
        ref[1,3,1,0]=(0.0470679202395)*x_ref[0]**o+(-0.809937294063)*x_ref[0]+(0.956184337312)*x_ref[1]**o+(-0.539155228812)*x_ref[1]
        ref[1,3,1,1]=(0.720979862876)*x_ref[0]**o+(0.412993117346)*x_ref[0]+(0.259998599343)*x_ref[1]**o+(-0.739191117502)*x_ref[1]
        ref[1,3,2,0]=(0.403493857435)*x_ref[0]**o+(0.0582403515426)*x_ref[0]+(0.0311147864546)*x_ref[1]**o+(0.397740429479)*x_ref[1]
        ref[1,3,2,1]=(0.51510647545)*x_ref[0]**o+(0.179454689412)*x_ref[0]+(-0.864589732719)*x_ref[1]**o+(-0.919542815326)*x_ref[1]
        ref[1,4,0,0]=(0.176385360146)*x_ref[0]**o+(0.666029075488)*x_ref[0]+(-0.765133182482)*x_ref[1]**o+(0.506226091771)*x_ref[1]
        ref[1,4,0,1]=(-0.217264015218)*x_ref[0]**o+(0.997799494004)*x_ref[0]+(0.706270854724)*x_ref[1]**o+(-0.440854336576)*x_ref[1]
        ref[1,4,1,0]=(-0.127481889699)*x_ref[0]**o+(-0.959927193428)*x_ref[0]+(-0.314190555296)*x_ref[1]**o+(-0.0254770240779)*x_ref[1]
        ref[1,4,1,1]=(0.943927827542)*x_ref[0]**o+(0.695303968245)*x_ref[0]+(0.537165245123)*x_ref[1]**o+(-0.414505300627)*x_ref[1]
        ref[1,4,2,0]=(-0.583859597164)*x_ref[0]**o+(-0.71993694674)*x_ref[0]+(0.0390198271077)*x_ref[1]**o+(0.917816363923)*x_ref[1]
        ref[1,4,2,1]=(0.752163371766)*x_ref[0]**o+(0.47299437482)*x_ref[0]+(-0.499822965518)*x_ref[1]**o+(-0.629775835404)*x_ref[1]
        ref[2,0,0,0]=(0.441638332575)*x_ref[0]**o+(0.183300755854)*x_ref[0]+(0.635094026763)*x_ref[1]**o+(-0.981751131017)*x_ref[1]
        ref[2,0,0,1]=(0.564782099962)*x_ref[0]**o+(-0.00355099531182)*x_ref[0]+(-0.309065488817)*x_ref[1]**o+(-0.805388517953)*x_ref[1]
        ref[2,0,1,0]=(-0.152274122381)*x_ref[0]**o+(-0.884657204578)*x_ref[0]+(-0.362170305747)*x_ref[1]**o+(0.688413345566)*x_ref[1]
        ref[2,0,1,1]=(0.433159898578)*x_ref[0]**o+(-0.926183453021)*x_ref[0]+(-0.990900351089)*x_ref[1]**o+(-0.558683257136)*x_ref[1]
        ref[2,0,2,0]=(0.378679325956)*x_ref[0]**o+(-0.178401685533)*x_ref[0]+(0.294582559995)*x_ref[1]**o+(0.324129328685)*x_ref[1]
        ref[2,0,2,1]=(0.191782222208)*x_ref[0]**o+(-0.718853143054)*x_ref[0]+(-0.1561779489)*x_ref[1]**o+(0.122297214902)*x_ref[1]
        ref[2,1,0,0]=(-0.745888747317)*x_ref[0]**o+(0.124662622215)*x_ref[0]+(0.61762073845)*x_ref[1]**o+(-0.543673681976)*x_ref[1]
        ref[2,1,0,1]=(0.696268656575)*x_ref[0]**o+(0.879682905689)*x_ref[0]+(0.101543002235)*x_ref[1]**o+(0.795349066813)*x_ref[1]
        ref[2,1,1,0]=(-0.595042825753)*x_ref[0]**o+(0.244946186339)*x_ref[0]+(0.107868704293)*x_ref[1]**o+(0.20527282324)*x_ref[1]
        ref[2,1,1,1]=(0.702092863468)*x_ref[0]**o+(-0.95898866728)*x_ref[0]+(0.415299739291)*x_ref[1]**o+(-0.602967505073)*x_ref[1]
        ref[2,1,2,0]=(0.242441579412)*x_ref[0]**o+(-0.621648829236)*x_ref[0]+(0.00279977134428)*x_ref[1]**o+(0.20444394565)*x_ref[1]
        ref[2,1,2,1]=(-0.248160979861)*x_ref[0]**o+(0.510934355012)*x_ref[0]+(-0.888013212236)*x_ref[1]**o+(0.348960154636)*x_ref[1]
        ref[2,2,0,0]=(-0.484727669828)*x_ref[0]**o+(0.23183410781)*x_ref[0]+(0.771832910785)*x_ref[1]**o+(-0.259023608026)*x_ref[1]
        ref[2,2,0,1]=(-0.0293853663853)*x_ref[0]**o+(0.132402850017)*x_ref[0]+(0.460425782419)*x_ref[1]**o+(0.918697618402)*x_ref[1]
        ref[2,2,1,0]=(-0.960506935656)*x_ref[0]**o+(0.302982411255)*x_ref[0]+(0.927623802726)*x_ref[1]**o+(0.446529083445)*x_ref[1]
        ref[2,2,1,1]=(-0.0791528123579)*x_ref[0]**o+(-0.963662594671)*x_ref[0]+(0.445231487496)*x_ref[1]**o+(-0.635556750912)*x_ref[1]
        ref[2,2,2,0]=(0.119522639306)*x_ref[0]**o+(0.999196602488)*x_ref[0]+(0.75598380813)*x_ref[1]**o+(0.159012027458)*x_ref[1]
        ref[2,2,2,1]=(0.671543289198)*x_ref[0]**o+(-0.897542325495)*x_ref[0]+(0.184767661479)*x_ref[1]**o+(-0.679101227287)*x_ref[1]
        ref[2,3,0,0]=(0.705751399559)*x_ref[0]**o+(0.0343552696754)*x_ref[0]+(0.295737350313)*x_ref[1]**o+(0.603031693221)*x_ref[1]
        ref[2,3,0,1]=(-0.568041111686)*x_ref[0]**o+(0.795100811027)*x_ref[0]+(-0.853308139159)*x_ref[1]**o+(-0.198019398328)*x_ref[1]
        ref[2,3,1,0]=(-0.568416608732)*x_ref[0]**o+(-0.785696005424)*x_ref[0]+(0.734563901681)*x_ref[1]**o+(0.337683654534)*x_ref[1]
        ref[2,3,1,1]=(0.750616777015)*x_ref[0]**o+(-0.243919199772)*x_ref[0]+(0.47894230818)*x_ref[1]**o+(0.117853316761)*x_ref[1]
        ref[2,3,2,0]=(0.595655270733)*x_ref[0]**o+(-0.131631796604)*x_ref[0]+(-0.363246505069)*x_ref[1]**o+(0.0399311321376)*x_ref[1]
        ref[2,3,2,1]=(-0.871124094437)*x_ref[0]**o+(-0.495878262772)*x_ref[0]+(-0.59370100135)*x_ref[1]**o+(0.745382881709)*x_ref[1]
        ref[2,4,0,0]=(0.479665892387)*x_ref[0]**o+(-0.393886135683)*x_ref[0]+(-0.969679118858)*x_ref[1]**o+(-0.720467296702)*x_ref[1]
        ref[2,4,0,1]=(0.111063753944)*x_ref[0]**o+(-0.759833999129)*x_ref[0]+(-0.123061597258)*x_ref[1]**o+(-0.982096877401)*x_ref[1]
        ref[2,4,1,0]=(-0.901470811302)*x_ref[0]**o+(0.395087698063)*x_ref[0]+(0.553738458787)*x_ref[1]**o+(-0.112328034611)*x_ref[1]
        ref[2,4,1,1]=(0.39514286938)*x_ref[0]**o+(-0.923250016486)*x_ref[0]+(0.361513401174)*x_ref[1]**o+(-0.272660910785)*x_ref[1]
        ref[2,4,2,0]=(0.454263938343)*x_ref[0]**o+(-0.0921743296726)*x_ref[0]+(-0.867658322099)*x_ref[1]**o+(-0.469179181142)*x_ref[1]
        ref[2,4,2,1]=(0.00363399429065)*x_ref[0]**o+(0.295615173633)*x_ref[0]+(-0.619556300313)*x_ref[1]**o+(0.385912799652)*x_ref[1]
        ref[3,0,0,0]=(-0.176638548482)*x_ref[0]**o+(0.716817113557)*x_ref[0]+(0.713969909735)*x_ref[1]**o+(-0.315381536549)*x_ref[1]
        ref[3,0,0,1]=(-0.981761588756)*x_ref[0]**o+(-0.264195260433)*x_ref[0]+(0.514278247605)*x_ref[1]**o+(0.620939057298)*x_ref[1]
        ref[3,0,1,0]=(-0.599489307325)*x_ref[0]**o+(0.213895727472)*x_ref[0]+(0.986896902529)*x_ref[1]**o+(0.423516351738)*x_ref[1]
        ref[3,0,1,1]=(0.282281951751)*x_ref[0]**o+(0.198509073556)*x_ref[0]+(0.261590765595)*x_ref[1]**o+(0.753982289812)*x_ref[1]
        ref[3,0,2,0]=(0.78593458655)*x_ref[0]**o+(-0.52358431526)*x_ref[0]+(0.479533127253)*x_ref[1]**o+(0.120268547872)*x_ref[1]
        ref[3,0,2,1]=(0.229057209441)*x_ref[0]**o+(-0.200667079463)*x_ref[0]+(0.713824885941)*x_ref[1]**o+(-0.317631878875)*x_ref[1]
        ref[3,1,0,0]=(-0.621475342986)*x_ref[0]**o+(-0.776830750036)*x_ref[0]+(0.74327888403)*x_ref[1]**o+(0.77144000144)*x_ref[1]
        ref[3,1,0,1]=(-0.0701637401121)*x_ref[0]**o+(0.315253792233)*x_ref[0]+(0.714679121123)*x_ref[1]**o+(-0.124524161083)*x_ref[1]
        ref[3,1,1,0]=(0.58339165807)*x_ref[0]**o+(0.754756142322)*x_ref[0]+(-0.552898566299)*x_ref[1]**o+(0.0401072453256)*x_ref[1]
        ref[3,1,1,1]=(-0.704826353788)*x_ref[0]**o+(-0.850442055451)*x_ref[0]+(0.0185807835768)*x_ref[1]**o+(0.798634373727)*x_ref[1]
        ref[3,1,2,0]=(-0.304968678423)*x_ref[0]**o+(0.882631700011)*x_ref[0]+(-0.00670582763206)*x_ref[1]**o+(-0.790997780888)*x_ref[1]
        ref[3,1,2,1]=(0.978052946899)*x_ref[0]**o+(0.0822751454714)*x_ref[0]+(-0.141545511663)*x_ref[1]**o+(0.804142415097)*x_ref[1]
        ref[3,2,0,0]=(-0.233834449356)*x_ref[0]**o+(-0.349324841655)*x_ref[0]+(0.633317011256)*x_ref[1]**o+(-0.682052363655)*x_ref[1]
        ref[3,2,0,1]=(-0.828772677459)*x_ref[0]**o+(-0.407293570997)*x_ref[0]+(0.0378365988776)*x_ref[1]**o+(0.575150943785)*x_ref[1]
        ref[3,2,1,0]=(0.00530675292728)*x_ref[0]**o+(-0.583321711314)*x_ref[0]+(-0.461755026035)*x_ref[1]**o+(-0.234209819371)*x_ref[1]
        ref[3,2,1,1]=(-0.529179272486)*x_ref[0]**o+(-0.126586669403)*x_ref[0]+(0.199780481034)*x_ref[1]**o+(-0.475384002433)*x_ref[1]
        ref[3,2,2,0]=(-0.794495647543)*x_ref[0]**o+(0.586459171803)*x_ref[0]+(-0.417091337526)*x_ref[1]**o+(-0.920839445366)*x_ref[1]
        ref[3,2,2,1]=(-0.543510366902)*x_ref[0]**o+(0.9390916149)*x_ref[0]+(-0.378905656214)*x_ref[1]**o+(-0.580144827894)*x_ref[1]
        ref[3,3,0,0]=(-0.218318771891)*x_ref[0]**o+(-0.741715169744)*x_ref[0]+(0.471332960959)*x_ref[1]**o+(-0.942331201293)*x_ref[1]
        ref[3,3,0,1]=(-0.639400890732)*x_ref[0]**o+(-0.35086784511)*x_ref[0]+(-0.662448518836)*x_ref[1]**o+(-0.109260566936)*x_ref[1]
        ref[3,3,1,0]=(0.00137953261949)*x_ref[0]**o+(-0.527278610973)*x_ref[0]+(0.547316965109)*x_ref[1]**o+(0.747307046881)*x_ref[1]
        ref[3,3,1,1]=(-0.110490640565)*x_ref[0]**o+(0.30599876694)*x_ref[0]+(0.464447167957)*x_ref[1]**o+(-0.650215871098)*x_ref[1]
        ref[3,3,2,0]=(-0.0752004672453)*x_ref[0]**o+(-0.476012931477)*x_ref[0]+(-0.463766892732)*x_ref[1]**o+(-0.864312392032)*x_ref[1]
        ref[3,3,2,1]=(-0.614219705651)*x_ref[0]**o+(0.579350508101)*x_ref[0]+(-0.307032385253)*x_ref[1]**o+(0.609948162009)*x_ref[1]
        ref[3,4,0,0]=(-0.946145025239)*x_ref[0]**o+(0.383874203169)*x_ref[0]+(-0.128251366911)*x_ref[1]**o+(-0.592420767396)*x_ref[1]
        ref[3,4,0,1]=(0.0656946533978)*x_ref[0]**o+(-0.400244444789)*x_ref[0]+(-0.263415867899)*x_ref[1]**o+(-0.241065146194)*x_ref[1]
        ref[3,4,1,0]=(-0.44363374655)*x_ref[0]**o+(-0.702360780698)*x_ref[0]+(0.39242809706)*x_ref[1]**o+(-0.499711507536)*x_ref[1]
        ref[3,4,1,1]=(-0.790376230857)*x_ref[0]**o+(0.271619243859)*x_ref[0]+(0.221426582034)*x_ref[1]**o+(-0.244571958075)*x_ref[1]
        ref[3,4,2,0]=(-0.599565364787)*x_ref[0]**o+(0.0637522137783)*x_ref[0]+(0.362816245637)*x_ref[1]**o+(-0.52786138669)*x_ref[1]
        ref[3,4,2,1]=(0.955472544947)*x_ref[0]**o+(-0.368132248377)*x_ref[0]+(0.262396158645)*x_ref[1]**o+(-0.252857678444)*x_ref[1]
      else:
        s[0,0,0,0]=(0.18541544761)*x[0]**o+(0.21364695446)*x[0]+(-0.472182966156)*x[1]**o+(0.896603283439)*x[1]+(-0.0688785608079)*x[2]**o+(-0.909852023221)*x[2]
        s[0,0,0,1]=(-0.415942079174)*x[0]**o+(0.875121980159)*x[0]+(0.310534797766)*x[1]**o+(-0.308596058529)*x[1]+(0.108266854467)*x[2]**o+(0.0333359884244)*x[2]
        s[0,0,1,0]=(-0.640298530291)*x[0]**o+(0.0117522671103)*x[0]+(-0.0110933087631)*x[1]**o+(0.265510073738)*x[1]+(0.533125824444)*x[2]**o+(0.608040031534)*x[2]
        s[0,0,1,1]=(-0.614118844709)*x[0]**o+(0.220136724383)*x[0]+(0.500278681433)*x[1]**o+(0.812249768469)*x[1]+(-0.694030974501)*x[2]**o+(-0.0657321516154)*x[2]
        s[0,0,2,0]=(-0.450201716565)*x[0]**o+(0.334136354848)*x[0]+(-0.832011262075)*x[1]**o+(-0.302792705771)*x[1]+(0.143656982651)*x[2]**o+(0.122111775743)*x[2]
        s[0,0,2,1]=(-0.527491596817)*x[0]**o+(0.454281406832)*x[0]+(0.0271798764911)*x[1]**o+(0.101232125515)*x[1]+(-0.16747525537)*x[2]**o+(0.901067653324)*x[2]
        s[0,1,0,0]=(0.124726861265)*x[0]**o+(0.694941253649)*x[0]+(-0.429194779514)*x[1]**o+(-0.857934333758)*x[1]+(-0.156711811013)*x[2]**o+(0.972951797234)*x[2]
        s[0,1,0,1]=(0.342428292381)*x[0]**o+(-0.880577787437)*x[0]+(0.769255197783)*x[1]**o+(0.369977548898)*x[1]+(0.0857387440023)*x[2]**o+(0.679043219801)*x[2]
        s[0,1,1,0]=(-0.393854353246)*x[0]**o+(-0.55861356234)*x[0]+(-0.103173283242)*x[1]**o+(-0.330435961835)*x[1]+(0.169248307533)*x[2]**o+(-0.807306530736)*x[2]
        s[0,1,1,1]=(0.460877309723)*x[0]**o+(-0.318462045167)*x[0]+(-0.397527462539)*x[1]**o+(0.915376261877)*x[1]+(-0.152840673613)*x[2]**o+(-0.877711938515)*x[2]
        s[0,1,2,0]=(-0.709963198272)*x[0]**o+(0.419992355176)*x[0]+(0.453556747804)*x[1]**o+(0.416257482776)*x[1]+(0.157460132998)*x[2]**o+(-0.820379434298)*x[2]
        s[0,1,2,1]=(-0.721496163985)*x[0]**o+(-0.420196812696)*x[0]+(0.494326744722)*x[1]**o+(-0.862937430197)*x[1]+(0.268700000282)*x[2]**o+(0.0964679940875)*x[2]
        s[0,2,0,0]=(-0.72785754306)*x[0]**o+(0.483405287113)*x[0]+(-0.0380836637151)*x[1]**o+(-0.750831233018)*x[1]+(0.890060303635)*x[2]**o+(-0.0889399861718)*x[2]
        s[0,2,0,1]=(-0.020133113466)*x[0]**o+(-0.111314198984)*x[0]+(0.246881700618)*x[1]**o+(-0.0463604335025)*x[1]+(0.340080199964)*x[2]**o+(0.546847705923)*x[2]
        s[0,2,1,0]=(0.44979764213)*x[0]**o+(0.136717449657)*x[0]+(0.885280875379)*x[1]**o+(-0.357374019051)*x[1]+(0.0780825418244)*x[2]**o+(0.407971535477)*x[2]
        s[0,2,1,1]=(-0.97479874766)*x[0]**o+(0.0206982137259)*x[0]+(0.522901951355)*x[1]**o+(-0.455258114156)*x[1]+(-0.0253655613024)*x[2]**o+(0.437644792848)*x[2]
        s[0,2,2,0]=(-0.704117519396)*x[0]**o+(0.529911491014)*x[0]+(0.468182929176)*x[1]**o+(-0.0350363879761)*x[1]+(0.752339337583)*x[2]**o+(-0.216693692569)*x[2]
        s[0,2,2,1]=(-0.684051531073)*x[0]**o+(-0.925299650276)*x[0]+(0.703519111239)*x[1]**o+(0.0606067581517)*x[1]+(-0.0425874860265)*x[2]**o+(0.510164371094)*x[2]
        s[0,3,0,0]=(0.694322896391)*x[0]**o+(0.510283487094)*x[0]+(-0.170009483148)*x[1]**o+(-0.977620371318)*x[1]+(-0.245739552357)*x[2]**o+(0.705681250186)*x[2]
        s[0,3,0,1]=(-0.949754702137)*x[0]**o+(0.866426989442)*x[0]+(-0.86696547128)*x[1]**o+(-0.588446790814)*x[1]+(0.395890571504)*x[2]**o+(-0.423457961295)*x[2]
        s[0,3,1,0]=(0.959539309072)*x[0]**o+(-0.0887741681458)*x[0]+(-0.428764678901)*x[1]**o+(0.181556980404)*x[1]+(0.147860645259)*x[2]**o+(0.608439127363)*x[2]
        s[0,3,1,1]=(-0.4122627159)*x[0]**o+(0.276340029649)*x[0]+(0.514501999828)*x[1]**o+(0.348320416875)*x[1]+(-0.440025117506)*x[2]**o+(-0.402728176651)*x[2]
        s[0,3,2,0]=(-0.19724428313)*x[0]**o+(-0.98620491494)*x[0]+(-0.927442513317)*x[1]**o+(-0.105595731292)*x[1]+(0.265791342515)*x[2]**o+(-0.488717755361)*x[2]
        s[0,3,2,1]=(0.346092758704)*x[0]**o+(0.471986850699)*x[0]+(-0.620240703772)*x[1]**o+(-0.655317618692)*x[1]+(0.552548181694)*x[2]**o+(-0.893031037271)*x[2]
        s[0,4,0,0]=(-0.598836349815)*x[0]**o+(-0.805034268296)*x[0]+(-0.24807713651)*x[1]**o+(0.502403721932)*x[1]+(-0.871875426638)*x[2]**o+(0.975048359849)*x[2]
        s[0,4,0,1]=(-0.848727481846)*x[0]**o+(0.289751356976)*x[0]+(0.866460492643)*x[1]**o+(-0.299228878853)*x[1]+(0.353468499277)*x[2]**o+(0.726026383162)*x[2]
        s[0,4,1,0]=(-0.433026970636)*x[0]**o+(0.952884497458)*x[0]+(0.390828895167)*x[1]**o+(0.460241307465)*x[1]+(-0.826148904445)*x[2]**o+(-0.907790760228)*x[2]
        s[0,4,1,1]=(-0.911683982834)*x[0]**o+(0.713610726367)*x[0]+(0.208806166291)*x[1]**o+(0.0880377414163)*x[1]+(-0.542366224153)*x[2]**o+(0.664202617762)*x[2]
        s[0,4,2,0]=(-0.484661301801)*x[0]**o+(0.636638351317)*x[0]+(0.807801531569)*x[1]**o+(-0.0265380392548)*x[1]+(0.726678256591)*x[2]**o+(0.348640889897)*x[2]
        s[0,4,2,1]=(-0.153834824583)*x[0]**o+(-0.855877123511)*x[0]+(0.539564337213)*x[1]**o+(-0.0218740334893)*x[1]+(0.437159750607)*x[2]**o+(0.482829271016)*x[2]
        s[1,0,0,0]=(-0.0168099698768)*x[0]**o+(-0.989521551028)*x[0]+(0.817466057301)*x[1]**o+(-0.21953132457)*x[1]+(-0.730490846394)*x[2]**o+(0.169063128153)*x[2]
        s[1,0,0,1]=(0.06972569424)*x[0]**o+(-0.775322425858)*x[0]+(-0.671461733727)*x[1]**o+(-0.431923083922)*x[1]+(0.812299957548)*x[2]**o+(0.966461396717)*x[2]
        s[1,0,1,0]=(0.456671962378)*x[0]**o+(0.359745823125)*x[0]+(0.63578449252)*x[1]**o+(0.900009101048)*x[1]+(-0.208011722813)*x[2]**o+(0.604028542303)*x[2]
        s[1,0,1,1]=(-0.0942278500641)*x[0]**o+(-0.54503821106)*x[0]+(-0.240087623461)*x[1]**o+(0.900303365317)*x[1]+(-0.00754914184533)*x[2]**o+(-0.836352541169)*x[2]
        s[1,0,2,0]=(-0.154823915066)*x[0]**o+(-0.984429514577)*x[0]+(0.309393138159)*x[1]**o+(-0.898889250398)*x[1]+(-0.597610752962)*x[2]**o+(-0.595043417245)*x[2]
        s[1,0,2,1]=(-0.452012178484)*x[0]**o+(0.406812802283)*x[0]+(-0.0252372070515)*x[1]**o+(0.599365873757)*x[1]+(0.696137317421)*x[2]**o+(0.39882469864)*x[2]
        s[1,1,0,0]=(0.960061217889)*x[0]**o+(-0.46292996833)*x[0]+(-0.908851982171)*x[1]**o+(-0.35478509058)*x[1]+(-0.0891426046361)*x[2]**o+(0.504413928397)*x[2]
        s[1,1,0,1]=(0.184040383397)*x[0]**o+(-0.125184201347)*x[0]+(0.411368449993)*x[1]**o+(0.76097933208)*x[1]+(0.85874596874)*x[2]**o+(0.0754186631504)*x[2]
        s[1,1,1,0]=(0.670739292338)*x[0]**o+(0.391393649635)*x[0]+(0.171246535424)*x[1]**o+(-0.195052345183)*x[1]+(-0.81284256555)*x[2]**o+(0.729283799083)*x[2]
        s[1,1,1,1]=(-0.456981034012)*x[0]**o+(0.155704355345)*x[0]+(0.480506857776)*x[1]**o+(0.951868527916)*x[1]+(-0.632815963451)*x[2]**o+(-0.281974721192)*x[2]
        s[1,1,2,0]=(0.509907581268)*x[0]**o+(0.974002899767)*x[0]+(-0.487852409248)*x[1]**o+(-0.994031112124)*x[1]+(-0.338175212869)*x[2]**o+(0.0922405530126)*x[2]
        s[1,1,2,1]=(0.240744658624)*x[0]**o+(-0.781993603247)*x[0]+(-0.857979982963)*x[1]**o+(0.138259946342)*x[1]+(-0.0257373573435)*x[2]**o+(-0.549948531512)*x[2]
        s[1,2,0,0]=(0.880823067307)*x[0]**o+(0.769790088552)*x[0]+(0.9242309604)*x[1]**o+(0.710853877125)*x[1]+(0.401127571007)*x[2]**o+(-0.601797131968)*x[2]
        s[1,2,0,1]=(-0.250958562933)*x[0]**o+(0.240041223327)*x[0]+(-0.0868736376912)*x[1]**o+(0.250994177285)*x[1]+(-0.198684086146)*x[2]**o+(0.752504023325)*x[2]
        s[1,2,1,0]=(-0.625906698075)*x[0]**o+(0.0837653273654)*x[0]+(-0.912953810309)*x[1]**o+(-0.591346082423)*x[1]+(0.396077142183)*x[2]**o+(0.199486858965)*x[2]
        s[1,2,1,1]=(-0.870955204115)*x[0]**o+(0.291646033459)*x[0]+(0.0845308439963)*x[1]**o+(-0.427803300865)*x[1]+(0.76119438126)*x[2]**o+(-0.8056578936)*x[2]
        s[1,2,2,0]=(0.649263352322)*x[0]**o+(0.968952679331)*x[0]+(0.473809168621)*x[1]**o+(0.129679304698)*x[1]+(-0.434791766847)*x[2]**o+(-0.141531952455)*x[2]
        s[1,2,2,1]=(-0.680138603152)*x[0]**o+(0.481602504342)*x[0]+(0.0411112054712)*x[1]**o+(-0.335246451045)*x[1]+(-0.532198210944)*x[2]**o+(0.278248356925)*x[2]
        s[1,3,0,0]=(0.983866320172)*x[0]**o+(0.779388801006)*x[0]+(0.943099409418)*x[1]**o+(0.29867320629)*x[1]+(0.920899852268)*x[2]**o+(-0.90449127693)*x[2]
        s[1,3,0,1]=(0.888499557305)*x[0]**o+(-0.0555515101983)*x[0]+(0.833149202095)*x[1]**o+(0.590559034771)*x[1]+(0.765622961327)*x[2]**o+(-0.325502160283)*x[2]
        s[1,3,1,0]=(-0.630488855411)*x[0]**o+(-0.453307600004)*x[0]+(0.773558050428)*x[1]**o+(-0.520443804373)*x[1]+(0.612407257369)*x[2]**o+(0.993705347884)*x[2]
        s[1,3,1,1]=(-0.785312491822)*x[0]**o+(-0.760712669832)*x[0]+(-0.572813097554)*x[1]**o+(0.151654230301)*x[1]+(0.770476334769)*x[2]**o+(-0.663724452997)*x[2]
        s[1,3,2,0]=(0.52562535089)*x[0]**o+(0.137778695278)*x[0]+(0.355394127523)*x[1]**o+(-0.241900298662)*x[1]+(-0.0311531414204)*x[2]**o+(0.61419880417)*x[2]
        s[1,3,2,1]=(0.642395964205)*x[0]**o+(0.938741084322)*x[0]+(-0.0385956334857)*x[1]**o+(0.44257567849)*x[1]+(-0.355655993218)*x[2]**o+(0.060707347503)*x[2]
        s[1,4,0,0]=(-0.363374730402)*x[0]**o+(0.985470783307)*x[0]+(-0.966941636823)*x[1]**o+(-0.0300574872162)*x[1]+(-0.513481421962)*x[2]**o+(0.869607357947)*x[2]
        s[1,4,0,1]=(0.646448255445)*x[0]**o+(-0.19357727026)*x[0]+(0.12200322941)*x[1]**o+(0.560785915298)*x[1]+(0.103714094141)*x[2]**o+(0.0235574798014)*x[2]
        s[1,4,1,0]=(0.658100608473)*x[0]**o+(0.294472394081)*x[0]+(-0.5469350794)*x[1]**o+(0.650910226682)*x[1]+(0.574983452761)*x[2]**o+(-0.885589741854)*x[2]
        s[1,4,1,1]=(-0.833278077163)*x[0]**o+(-0.612685520243)*x[0]+(0.145850763252)*x[1]**o+(0.710544228623)*x[1]+(-0.237536168655)*x[2]**o+(-0.864417422252)*x[2]
        s[1,4,2,0]=(0.547214540024)*x[0]**o+(0.259625305608)*x[0]+(0.972583304662)*x[1]**o+(0.0205295851227)*x[1]+(-0.383299704447)*x[2]**o+(0.899668950656)*x[2]
        s[1,4,2,1]=(-0.0898305090604)*x[0]**o+(0.0583491290278)*x[0]+(0.612069962579)*x[1]**o+(0.00451030954439)*x[1]+(-0.543702915529)*x[2]**o+(0.455742731079)*x[2]
        s[2,0,0,0]=(0.869996673689)*x[0]**o+(0.786072517277)*x[0]+(0.784064187563)*x[1]**o+(0.215121486636)*x[1]+(0.0167704070196)*x[2]**o+(0.510818786007)*x[2]
        s[2,0,0,1]=(0.70636305186)*x[0]**o+(-0.774931605014)*x[0]+(-0.566769222836)*x[1]**o+(-0.490039037574)*x[1]+(0.84779476702)*x[2]**o+(-0.277061243254)*x[2]
        s[2,0,1,0]=(0.71021869545)*x[0]**o+(0.297318884767)*x[0]+(0.393052893616)*x[1]**o+(0.126298772061)*x[1]+(-0.919383761879)*x[2]**o+(0.993331773902)*x[2]
        s[2,0,1,1]=(-0.422176053109)*x[0]**o+(0.1573461669)*x[0]+(-0.383086926798)*x[1]**o+(0.521025942156)*x[1]+(0.573775622558)*x[2]**o+(-0.424831674319)*x[2]
        s[2,0,2,0]=(-0.555531945291)*x[0]**o+(0.174293068827)*x[0]+(-0.937417557253)*x[1]**o+(0.0307979382142)*x[1]+(0.0250257663963)*x[2]**o+(-0.0601088874906)*x[2]
        s[2,0,2,1]=(-0.231539926473)*x[0]**o+(0.943546667559)*x[0]+(-0.453034497139)*x[1]**o+(-0.799323498413)*x[1]+(0.626571998737)*x[2]**o+(0.256254025304)*x[2]
        s[2,1,0,0]=(0.348839391625)*x[0]**o+(0.747443708496)*x[0]+(-0.632807812622)*x[1]**o+(0.30083165118)*x[1]+(-0.694353354492)*x[2]**o+(-0.384450833928)*x[2]
        s[2,1,0,1]=(-0.107737403038)*x[0]**o+(0.298410093423)*x[0]+(0.160830729598)*x[1]**o+(-0.627822690248)*x[1]+(0.511195156768)*x[2]**o+(-0.115525478726)*x[2]
        s[2,1,1,0]=(-0.188820324857)*x[0]**o+(-0.0663606843968)*x[0]+(0.151440137108)*x[1]**o+(0.946232159278)*x[1]+(-0.282849712427)*x[2]**o+(-0.439152308605)*x[2]
        s[2,1,1,1]=(-0.374881709538)*x[0]**o+(0.745922133487)*x[0]+(-0.0242470036044)*x[1]**o+(-0.711004108426)*x[1]+(0.417440756886)*x[2]**o+(-0.510006752436)*x[2]
        s[2,1,2,0]=(0.864353278124)*x[0]**o+(0.0809895383191)*x[0]+(0.457075492213)*x[1]**o+(-0.14392908535)*x[1]+(0.784143277535)*x[2]**o+(-0.15341048191)*x[2]
        s[2,1,2,1]=(0.269079647662)*x[0]**o+(0.0402843915276)*x[0]+(-0.683925387125)*x[1]**o+(0.729804106713)*x[1]+(0.0732202476811)*x[2]**o+(-0.547958146499)*x[2]
        s[2,2,0,0]=(0.272934512065)*x[0]**o+(-0.882225076784)*x[0]+(-0.405012739084)*x[1]**o+(-0.989715644128)*x[1]+(-0.155608583603)*x[2]**o+(-0.722826494052)*x[2]
        s[2,2,0,1]=(0.339423408668)*x[0]**o+(-0.330465697311)*x[0]+(-0.692428979863)*x[1]**o+(0.322765870878)*x[1]+(-0.384201212534)*x[2]**o+(-0.0946217242584)*x[2]
        s[2,2,1,0]=(-0.793318742542)*x[0]**o+(-0.375149299842)*x[0]+(0.793898380613)*x[1]**o+(-0.593499973718)*x[1]+(-0.65815800334)*x[2]**o+(-0.0744951302063)*x[2]
        s[2,2,1,1]=(0.0279370862836)*x[0]**o+(-0.196535478436)*x[0]+(-0.942547794666)*x[1]**o+(0.994269193299)*x[1]+(0.781181650698)*x[2]**o+(-0.221591113746)*x[2]
        s[2,2,2,0]=(0.268315410937)*x[0]**o+(-0.53373824299)*x[0]+(0.777981431744)*x[1]**o+(0.226918217122)*x[1]+(-0.561960454199)*x[2]**o+(-0.273776681445)*x[2]
        s[2,2,2,1]=(0.454454381954)*x[0]**o+(-0.391271318249)*x[0]+(-0.835917284669)*x[1]**o+(-0.42457026252)*x[1]+(-0.848426604454)*x[2]**o+(0.285767008408)*x[2]
        s[2,3,0,0]=(0.771743337775)*x[0]**o+(-0.157913284247)*x[0]+(0.00417714721394)*x[1]**o+(0.875095301539)*x[1]+(0.965135621963)*x[2]**o+(-0.12209631704)*x[2]
        s[2,3,0,1]=(0.00527421717368)*x[0]**o+(0.119837014154)*x[0]+(-0.326451619959)*x[1]**o+(-0.179736365312)*x[1]+(-0.659003332514)*x[2]**o+(0.44402878832)*x[2]
        s[2,3,1,0]=(0.080864510177)*x[0]**o+(-0.910530923746)*x[0]+(-0.065442686402)*x[1]**o+(-0.162655506011)*x[1]+(-0.559755580552)*x[2]**o+(-0.447763051956)*x[2]
        s[2,3,1,1]=(0.45313869387)*x[0]**o+(0.478878867458)*x[0]+(0.0959944374683)*x[1]**o+(-0.0379945879624)*x[1]+(0.877383707005)*x[2]**o+(0.0037296134062)*x[2]
        s[2,3,2,0]=(0.187710780263)*x[0]**o+(0.722012532502)*x[0]+(-0.599862911914)*x[1]**o+(-0.974072653552)*x[1]+(0.632960753399)*x[2]**o+(-0.249506081862)*x[2]
        s[2,3,2,1]=(0.915570663708)*x[0]**o+(-0.590657458642)*x[0]+(-0.221387098604)*x[1]**o+(-0.619974779079)*x[1]+(-0.156470444965)*x[2]**o+(-0.712127719679)*x[2]
        s[2,4,0,0]=(0.57214571034)*x[0]**o+(0.444107582229)*x[0]+(-0.455561895744)*x[1]**o+(0.350956075511)*x[1]+(-0.060920804543)*x[2]**o+(0.102498779175)*x[2]
        s[2,4,0,1]=(-0.117233414571)*x[0]**o+(0.336917030674)*x[0]+(-0.544691169707)*x[1]**o+(0.506633531456)*x[1]+(0.193173288491)*x[2]**o+(0.690898822609)*x[2]
        s[2,4,1,0]=(-0.28899417241)*x[0]**o+(0.930513954761)*x[0]+(-0.564925441731)*x[1]**o+(-0.082228884842)*x[1]+(0.271797239833)*x[2]**o+(-0.293458041224)*x[2]
        s[2,4,1,1]=(0.303746570859)*x[0]**o+(-0.668303989547)*x[0]+(-0.845649340659)*x[1]**o+(0.826906198662)*x[1]+(-0.499957471874)*x[2]**o+(0.540857798511)*x[2]
        s[2,4,2,0]=(0.215998594196)*x[0]**o+(0.80715947344)*x[0]+(0.440392341629)*x[1]**o+(0.429949814054)*x[1]+(-0.476601527283)*x[2]**o+(0.21482287961)*x[2]
        s[2,4,2,1]=(0.744163797227)*x[0]**o+(0.485491624827)*x[0]+(0.0113951682847)*x[1]**o+(-0.219871501999)*x[1]+(-0.960000031149)*x[2]**o+(0.514892677016)*x[2]
        s[3,0,0,0]=(0.955379977654)*x[0]**o+(0.88166711804)*x[0]+(-0.212371823264)*x[1]**o+(0.61537617198)*x[1]+(0.290444741617)*x[2]**o+(0.669594654028)*x[2]
        s[3,0,0,1]=(-0.422932658626)*x[0]**o+(0.898327577684)*x[0]+(0.821836925105)*x[1]**o+(-0.777406619005)*x[1]+(0.836622617839)*x[2]**o+(-0.414551285496)*x[2]
        s[3,0,1,0]=(-0.0922163192963)*x[0]**o+(-0.426319353836)*x[0]+(-0.265988919456)*x[1]**o+(0.0364005746661)*x[1]+(-0.75420756707)*x[2]**o+(0.957027945483)*x[2]
        s[3,0,1,1]=(0.0501511171161)*x[0]**o+(-0.960573277539)*x[0]+(-0.484587580032)*x[1]**o+(-0.229640708516)*x[1]+(-0.652459896537)*x[2]**o+(-0.59981248766)*x[2]
        s[3,0,2,0]=(-0.378646761494)*x[0]**o+(-0.0165689331328)*x[0]+(-0.883277009771)*x[1]**o+(-0.947577778882)*x[1]+(0.188169490614)*x[2]**o+(0.079992543794)*x[2]
        s[3,0,2,1]=(-0.688755441231)*x[0]**o+(0.290810540822)*x[0]+(-0.272544759374)*x[1]**o+(0.412904132844)*x[1]+(-0.851170264556)*x[2]**o+(-0.257140479814)*x[2]
        s[3,1,0,0]=(0.69577276718)*x[0]**o+(0.777468411586)*x[0]+(0.0291906552252)*x[1]**o+(-0.465217099179)*x[1]+(-0.265314241222)*x[2]**o+(-0.914393443339)*x[2]
        s[3,1,0,1]=(-0.605826705227)*x[0]**o+(0.786198128441)*x[0]+(-0.413639776886)*x[1]**o+(0.9484840564)*x[1]+(0.660128431709)*x[2]**o+(0.294794064545)*x[2]
        s[3,1,1,0]=(0.470527104285)*x[0]**o+(-0.673225307232)*x[0]+(0.17528543617)*x[1]**o+(-0.61453857548)*x[1]+(-0.112025849328)*x[2]**o+(0.3719665037)*x[2]
        s[3,1,1,1]=(0.738351761472)*x[0]**o+(-0.692489261597)*x[0]+(-0.00750387276605)*x[1]**o+(-0.194526069313)*x[1]+(0.206427047745)*x[2]**o+(0.936451437125)*x[2]
        s[3,1,2,0]=(0.258949551337)*x[0]**o+(0.772440182785)*x[0]+(0.416917261494)*x[1]**o+(0.61679695051)*x[1]+(0.25182107102)*x[2]**o+(0.404770153496)*x[2]
        s[3,1,2,1]=(0.788357790268)*x[0]**o+(-0.561028019813)*x[0]+(0.842701777125)*x[1]**o+(0.861240216256)*x[1]+(0.633704418848)*x[2]**o+(0.660953804821)*x[2]
        s[3,2,0,0]=(-0.954508015564)*x[0]**o+(0.770831368915)*x[0]+(0.71068286494)*x[1]**o+(-0.706262271265)*x[1]+(0.000711549487662)*x[2]**o+(0.343413819166)*x[2]
        s[3,2,0,1]=(-0.486904541331)*x[0]**o+(-0.535805915583)*x[0]+(-0.697643420586)*x[1]**o+(-0.350844416469)*x[1]+(-0.570854441331)*x[2]**o+(0.844325362411)*x[2]
        s[3,2,1,0]=(0.989057215778)*x[0]**o+(0.822742140223)*x[0]+(0.791720783337)*x[1]**o+(0.0884164131725)*x[1]+(0.337584265788)*x[2]**o+(0.291148268923)*x[2]
        s[3,2,1,1]=(-0.938554739295)*x[0]**o+(0.154409183584)*x[0]+(0.296554553803)*x[1]**o+(0.0130418768484)*x[1]+(0.440811811112)*x[2]**o+(0.0891377617855)*x[2]
        s[3,2,2,0]=(0.0966486195221)*x[0]**o+(-0.228831238429)*x[0]+(-0.00903762540107)*x[1]**o+(0.687947867973)*x[1]+(0.29955036712)*x[2]**o+(-0.153152667621)*x[2]
        s[3,2,2,1]=(0.994414965494)*x[0]**o+(0.787224423165)*x[0]+(-0.572935987357)*x[1]**o+(-0.605628735246)*x[1]+(0.268503770734)*x[2]**o+(-0.32711743813)*x[2]
        s[3,3,0,0]=(-0.539283649631)*x[0]**o+(0.34809741571)*x[0]+(-0.792527713661)*x[1]**o+(-0.636087526521)*x[1]+(0.765496317046)*x[2]**o+(-0.908214892663)*x[2]
        s[3,3,0,1]=(0.903206720181)*x[0]**o+(0.761511999734)*x[0]+(0.649680135137)*x[1]**o+(-0.36181233615)*x[1]+(0.563162618539)*x[2]**o+(-0.810859258189)*x[2]
        s[3,3,1,0]=(-0.0915176526383)*x[0]**o+(0.674948774943)*x[0]+(0.105537180084)*x[1]**o+(0.765853705339)*x[1]+(-0.706270111874)*x[2]**o+(-0.899798116939)*x[2]
        s[3,3,1,1]=(0.247148082226)*x[0]**o+(-0.252654356068)*x[0]+(0.347296182632)*x[1]**o+(0.809639778301)*x[1]+(0.130358549062)*x[2]**o+(0.772783598594)*x[2]
        s[3,3,2,0]=(0.282056095613)*x[0]**o+(0.563274162153)*x[0]+(-0.469420374642)*x[1]**o+(-0.00929250493535)*x[1]+(-0.509250535991)*x[2]**o+(0.747602253057)*x[2]
        s[3,3,2,1]=(-0.40151566811)*x[0]**o+(0.878418940192)*x[0]+(0.548958199553)*x[1]**o+(0.556900329931)*x[1]+(-0.406381288534)*x[2]**o+(-0.767206838398)*x[2]
        s[3,4,0,0]=(-0.672264988169)*x[0]**o+(0.422502707117)*x[0]+(0.459427006408)*x[1]**o+(0.871928411825)*x[1]+(0.85683710126)*x[2]**o+(0.242796550494)*x[2]
        s[3,4,0,1]=(-0.821076005446)*x[0]**o+(-0.337616881342)*x[0]+(0.435309707923)*x[1]**o+(0.989009740264)*x[1]+(-0.470842629471)*x[2]**o+(-0.94476330486)*x[2]
        s[3,4,1,0]=(0.0488301102108)*x[0]**o+(-0.905860578484)*x[0]+(-0.986441486712)*x[1]**o+(0.819623730826)*x[1]+(0.75012214057)*x[2]**o+(-0.460862091553)*x[2]
        s[3,4,1,1]=(-0.932638473603)*x[0]**o+(-0.882471440278)*x[0]+(-0.498748850986)*x[1]**o+(-0.499262858754)*x[1]+(-0.748745750147)*x[2]**o+(-0.168661287934)*x[2]
        s[3,4,2,0]=(0.515883010306)*x[0]**o+(0.752629892309)*x[0]+(-0.340202982107)*x[1]**o+(-0.120916620105)*x[1]+(0.193421000969)*x[2]**o+(0.815950266515)*x[2]
        s[3,4,2,1]=(-0.397114672795)*x[0]**o+(-0.737395536831)*x[0]+(0.589874124917)*x[1]**o+(-0.476304220998)*x[1]+(-0.177825589992)*x[2]**o+(-0.776118337595)*x[2]
        ref[0,0,0,0]=(0.18541544761)*x_ref[0]**o+(0.21364695446)*x_ref[0]+(-0.472182966156)*x_ref[1]**o+(0.896603283439)*x_ref[1]+(-0.0688785608079)*x_ref[2]**o+(-0.909852023221)*x_ref[2]
        ref[0,0,0,1]=(-0.415942079174)*x_ref[0]**o+(0.875121980159)*x_ref[0]+(0.310534797766)*x_ref[1]**o+(-0.308596058529)*x_ref[1]+(0.108266854467)*x_ref[2]**o+(0.0333359884244)*x_ref[2]
        ref[0,0,1,0]=(-0.640298530291)*x_ref[0]**o+(0.0117522671103)*x_ref[0]+(-0.0110933087631)*x_ref[1]**o+(0.265510073738)*x_ref[1]+(0.533125824444)*x_ref[2]**o+(0.608040031534)*x_ref[2]
        ref[0,0,1,1]=(-0.614118844709)*x_ref[0]**o+(0.220136724383)*x_ref[0]+(0.500278681433)*x_ref[1]**o+(0.812249768469)*x_ref[1]+(-0.694030974501)*x_ref[2]**o+(-0.0657321516154)*x_ref[2]
        ref[0,0,2,0]=(-0.450201716565)*x_ref[0]**o+(0.334136354848)*x_ref[0]+(-0.832011262075)*x_ref[1]**o+(-0.302792705771)*x_ref[1]+(0.143656982651)*x_ref[2]**o+(0.122111775743)*x_ref[2]
        ref[0,0,2,1]=(-0.527491596817)*x_ref[0]**o+(0.454281406832)*x_ref[0]+(0.0271798764911)*x_ref[1]**o+(0.101232125515)*x_ref[1]+(-0.16747525537)*x_ref[2]**o+(0.901067653324)*x_ref[2]
        ref[0,1,0,0]=(0.124726861265)*x_ref[0]**o+(0.694941253649)*x_ref[0]+(-0.429194779514)*x_ref[1]**o+(-0.857934333758)*x_ref[1]+(-0.156711811013)*x_ref[2]**o+(0.972951797234)*x_ref[2]
        ref[0,1,0,1]=(0.342428292381)*x_ref[0]**o+(-0.880577787437)*x_ref[0]+(0.769255197783)*x_ref[1]**o+(0.369977548898)*x_ref[1]+(0.0857387440023)*x_ref[2]**o+(0.679043219801)*x_ref[2]
        ref[0,1,1,0]=(-0.393854353246)*x_ref[0]**o+(-0.55861356234)*x_ref[0]+(-0.103173283242)*x_ref[1]**o+(-0.330435961835)*x_ref[1]+(0.169248307533)*x_ref[2]**o+(-0.807306530736)*x_ref[2]
        ref[0,1,1,1]=(0.460877309723)*x_ref[0]**o+(-0.318462045167)*x_ref[0]+(-0.397527462539)*x_ref[1]**o+(0.915376261877)*x_ref[1]+(-0.152840673613)*x_ref[2]**o+(-0.877711938515)*x_ref[2]
        ref[0,1,2,0]=(-0.709963198272)*x_ref[0]**o+(0.419992355176)*x_ref[0]+(0.453556747804)*x_ref[1]**o+(0.416257482776)*x_ref[1]+(0.157460132998)*x_ref[2]**o+(-0.820379434298)*x_ref[2]
        ref[0,1,2,1]=(-0.721496163985)*x_ref[0]**o+(-0.420196812696)*x_ref[0]+(0.494326744722)*x_ref[1]**o+(-0.862937430197)*x_ref[1]+(0.268700000282)*x_ref[2]**o+(0.0964679940875)*x_ref[2]
        ref[0,2,0,0]=(-0.72785754306)*x_ref[0]**o+(0.483405287113)*x_ref[0]+(-0.0380836637151)*x_ref[1]**o+(-0.750831233018)*x_ref[1]+(0.890060303635)*x_ref[2]**o+(-0.0889399861718)*x_ref[2]
        ref[0,2,0,1]=(-0.020133113466)*x_ref[0]**o+(-0.111314198984)*x_ref[0]+(0.246881700618)*x_ref[1]**o+(-0.0463604335025)*x_ref[1]+(0.340080199964)*x_ref[2]**o+(0.546847705923)*x_ref[2]
        ref[0,2,1,0]=(0.44979764213)*x_ref[0]**o+(0.136717449657)*x_ref[0]+(0.885280875379)*x_ref[1]**o+(-0.357374019051)*x_ref[1]+(0.0780825418244)*x_ref[2]**o+(0.407971535477)*x_ref[2]
        ref[0,2,1,1]=(-0.97479874766)*x_ref[0]**o+(0.0206982137259)*x_ref[0]+(0.522901951355)*x_ref[1]**o+(-0.455258114156)*x_ref[1]+(-0.0253655613024)*x_ref[2]**o+(0.437644792848)*x_ref[2]
        ref[0,2,2,0]=(-0.704117519396)*x_ref[0]**o+(0.529911491014)*x_ref[0]+(0.468182929176)*x_ref[1]**o+(-0.0350363879761)*x_ref[1]+(0.752339337583)*x_ref[2]**o+(-0.216693692569)*x_ref[2]
        ref[0,2,2,1]=(-0.684051531073)*x_ref[0]**o+(-0.925299650276)*x_ref[0]+(0.703519111239)*x_ref[1]**o+(0.0606067581517)*x_ref[1]+(-0.0425874860265)*x_ref[2]**o+(0.510164371094)*x_ref[2]
        ref[0,3,0,0]=(0.694322896391)*x_ref[0]**o+(0.510283487094)*x_ref[0]+(-0.170009483148)*x_ref[1]**o+(-0.977620371318)*x_ref[1]+(-0.245739552357)*x_ref[2]**o+(0.705681250186)*x_ref[2]
        ref[0,3,0,1]=(-0.949754702137)*x_ref[0]**o+(0.866426989442)*x_ref[0]+(-0.86696547128)*x_ref[1]**o+(-0.588446790814)*x_ref[1]+(0.395890571504)*x_ref[2]**o+(-0.423457961295)*x_ref[2]
        ref[0,3,1,0]=(0.959539309072)*x_ref[0]**o+(-0.0887741681458)*x_ref[0]+(-0.428764678901)*x_ref[1]**o+(0.181556980404)*x_ref[1]+(0.147860645259)*x_ref[2]**o+(0.608439127363)*x_ref[2]
        ref[0,3,1,1]=(-0.4122627159)*x_ref[0]**o+(0.276340029649)*x_ref[0]+(0.514501999828)*x_ref[1]**o+(0.348320416875)*x_ref[1]+(-0.440025117506)*x_ref[2]**o+(-0.402728176651)*x_ref[2]
        ref[0,3,2,0]=(-0.19724428313)*x_ref[0]**o+(-0.98620491494)*x_ref[0]+(-0.927442513317)*x_ref[1]**o+(-0.105595731292)*x_ref[1]+(0.265791342515)*x_ref[2]**o+(-0.488717755361)*x_ref[2]
        ref[0,3,2,1]=(0.346092758704)*x_ref[0]**o+(0.471986850699)*x_ref[0]+(-0.620240703772)*x_ref[1]**o+(-0.655317618692)*x_ref[1]+(0.552548181694)*x_ref[2]**o+(-0.893031037271)*x_ref[2]
        ref[0,4,0,0]=(-0.598836349815)*x_ref[0]**o+(-0.805034268296)*x_ref[0]+(-0.24807713651)*x_ref[1]**o+(0.502403721932)*x_ref[1]+(-0.871875426638)*x_ref[2]**o+(0.975048359849)*x_ref[2]
        ref[0,4,0,1]=(-0.848727481846)*x_ref[0]**o+(0.289751356976)*x_ref[0]+(0.866460492643)*x_ref[1]**o+(-0.299228878853)*x_ref[1]+(0.353468499277)*x_ref[2]**o+(0.726026383162)*x_ref[2]
        ref[0,4,1,0]=(-0.433026970636)*x_ref[0]**o+(0.952884497458)*x_ref[0]+(0.390828895167)*x_ref[1]**o+(0.460241307465)*x_ref[1]+(-0.826148904445)*x_ref[2]**o+(-0.907790760228)*x_ref[2]
        ref[0,4,1,1]=(-0.911683982834)*x_ref[0]**o+(0.713610726367)*x_ref[0]+(0.208806166291)*x_ref[1]**o+(0.0880377414163)*x_ref[1]+(-0.542366224153)*x_ref[2]**o+(0.664202617762)*x_ref[2]
        ref[0,4,2,0]=(-0.484661301801)*x_ref[0]**o+(0.636638351317)*x_ref[0]+(0.807801531569)*x_ref[1]**o+(-0.0265380392548)*x_ref[1]+(0.726678256591)*x_ref[2]**o+(0.348640889897)*x_ref[2]
        ref[0,4,2,1]=(-0.153834824583)*x_ref[0]**o+(-0.855877123511)*x_ref[0]+(0.539564337213)*x_ref[1]**o+(-0.0218740334893)*x_ref[1]+(0.437159750607)*x_ref[2]**o+(0.482829271016)*x_ref[2]
        ref[1,0,0,0]=(-0.0168099698768)*x_ref[0]**o+(-0.989521551028)*x_ref[0]+(0.817466057301)*x_ref[1]**o+(-0.21953132457)*x_ref[1]+(-0.730490846394)*x_ref[2]**o+(0.169063128153)*x_ref[2]
        ref[1,0,0,1]=(0.06972569424)*x_ref[0]**o+(-0.775322425858)*x_ref[0]+(-0.671461733727)*x_ref[1]**o+(-0.431923083922)*x_ref[1]+(0.812299957548)*x_ref[2]**o+(0.966461396717)*x_ref[2]
        ref[1,0,1,0]=(0.456671962378)*x_ref[0]**o+(0.359745823125)*x_ref[0]+(0.63578449252)*x_ref[1]**o+(0.900009101048)*x_ref[1]+(-0.208011722813)*x_ref[2]**o+(0.604028542303)*x_ref[2]
        ref[1,0,1,1]=(-0.0942278500641)*x_ref[0]**o+(-0.54503821106)*x_ref[0]+(-0.240087623461)*x_ref[1]**o+(0.900303365317)*x_ref[1]+(-0.00754914184533)*x_ref[2]**o+(-0.836352541169)*x_ref[2]
        ref[1,0,2,0]=(-0.154823915066)*x_ref[0]**o+(-0.984429514577)*x_ref[0]+(0.309393138159)*x_ref[1]**o+(-0.898889250398)*x_ref[1]+(-0.597610752962)*x_ref[2]**o+(-0.595043417245)*x_ref[2]
        ref[1,0,2,1]=(-0.452012178484)*x_ref[0]**o+(0.406812802283)*x_ref[0]+(-0.0252372070515)*x_ref[1]**o+(0.599365873757)*x_ref[1]+(0.696137317421)*x_ref[2]**o+(0.39882469864)*x_ref[2]
        ref[1,1,0,0]=(0.960061217889)*x_ref[0]**o+(-0.46292996833)*x_ref[0]+(-0.908851982171)*x_ref[1]**o+(-0.35478509058)*x_ref[1]+(-0.0891426046361)*x_ref[2]**o+(0.504413928397)*x_ref[2]
        ref[1,1,0,1]=(0.184040383397)*x_ref[0]**o+(-0.125184201347)*x_ref[0]+(0.411368449993)*x_ref[1]**o+(0.76097933208)*x_ref[1]+(0.85874596874)*x_ref[2]**o+(0.0754186631504)*x_ref[2]
        ref[1,1,1,0]=(0.670739292338)*x_ref[0]**o+(0.391393649635)*x_ref[0]+(0.171246535424)*x_ref[1]**o+(-0.195052345183)*x_ref[1]+(-0.81284256555)*x_ref[2]**o+(0.729283799083)*x_ref[2]
        ref[1,1,1,1]=(-0.456981034012)*x_ref[0]**o+(0.155704355345)*x_ref[0]+(0.480506857776)*x_ref[1]**o+(0.951868527916)*x_ref[1]+(-0.632815963451)*x_ref[2]**o+(-0.281974721192)*x_ref[2]
        ref[1,1,2,0]=(0.509907581268)*x_ref[0]**o+(0.974002899767)*x_ref[0]+(-0.487852409248)*x_ref[1]**o+(-0.994031112124)*x_ref[1]+(-0.338175212869)*x_ref[2]**o+(0.0922405530126)*x_ref[2]
        ref[1,1,2,1]=(0.240744658624)*x_ref[0]**o+(-0.781993603247)*x_ref[0]+(-0.857979982963)*x_ref[1]**o+(0.138259946342)*x_ref[1]+(-0.0257373573435)*x_ref[2]**o+(-0.549948531512)*x_ref[2]
        ref[1,2,0,0]=(0.880823067307)*x_ref[0]**o+(0.769790088552)*x_ref[0]+(0.9242309604)*x_ref[1]**o+(0.710853877125)*x_ref[1]+(0.401127571007)*x_ref[2]**o+(-0.601797131968)*x_ref[2]
        ref[1,2,0,1]=(-0.250958562933)*x_ref[0]**o+(0.240041223327)*x_ref[0]+(-0.0868736376912)*x_ref[1]**o+(0.250994177285)*x_ref[1]+(-0.198684086146)*x_ref[2]**o+(0.752504023325)*x_ref[2]
        ref[1,2,1,0]=(-0.625906698075)*x_ref[0]**o+(0.0837653273654)*x_ref[0]+(-0.912953810309)*x_ref[1]**o+(-0.591346082423)*x_ref[1]+(0.396077142183)*x_ref[2]**o+(0.199486858965)*x_ref[2]
        ref[1,2,1,1]=(-0.870955204115)*x_ref[0]**o+(0.291646033459)*x_ref[0]+(0.0845308439963)*x_ref[1]**o+(-0.427803300865)*x_ref[1]+(0.76119438126)*x_ref[2]**o+(-0.8056578936)*x_ref[2]
        ref[1,2,2,0]=(0.649263352322)*x_ref[0]**o+(0.968952679331)*x_ref[0]+(0.473809168621)*x_ref[1]**o+(0.129679304698)*x_ref[1]+(-0.434791766847)*x_ref[2]**o+(-0.141531952455)*x_ref[2]
        ref[1,2,2,1]=(-0.680138603152)*x_ref[0]**o+(0.481602504342)*x_ref[0]+(0.0411112054712)*x_ref[1]**o+(-0.335246451045)*x_ref[1]+(-0.532198210944)*x_ref[2]**o+(0.278248356925)*x_ref[2]
        ref[1,3,0,0]=(0.983866320172)*x_ref[0]**o+(0.779388801006)*x_ref[0]+(0.943099409418)*x_ref[1]**o+(0.29867320629)*x_ref[1]+(0.920899852268)*x_ref[2]**o+(-0.90449127693)*x_ref[2]
        ref[1,3,0,1]=(0.888499557305)*x_ref[0]**o+(-0.0555515101983)*x_ref[0]+(0.833149202095)*x_ref[1]**o+(0.590559034771)*x_ref[1]+(0.765622961327)*x_ref[2]**o+(-0.325502160283)*x_ref[2]
        ref[1,3,1,0]=(-0.630488855411)*x_ref[0]**o+(-0.453307600004)*x_ref[0]+(0.773558050428)*x_ref[1]**o+(-0.520443804373)*x_ref[1]+(0.612407257369)*x_ref[2]**o+(0.993705347884)*x_ref[2]
        ref[1,3,1,1]=(-0.785312491822)*x_ref[0]**o+(-0.760712669832)*x_ref[0]+(-0.572813097554)*x_ref[1]**o+(0.151654230301)*x_ref[1]+(0.770476334769)*x_ref[2]**o+(-0.663724452997)*x_ref[2]
        ref[1,3,2,0]=(0.52562535089)*x_ref[0]**o+(0.137778695278)*x_ref[0]+(0.355394127523)*x_ref[1]**o+(-0.241900298662)*x_ref[1]+(-0.0311531414204)*x_ref[2]**o+(0.61419880417)*x_ref[2]
        ref[1,3,2,1]=(0.642395964205)*x_ref[0]**o+(0.938741084322)*x_ref[0]+(-0.0385956334857)*x_ref[1]**o+(0.44257567849)*x_ref[1]+(-0.355655993218)*x_ref[2]**o+(0.060707347503)*x_ref[2]
        ref[1,4,0,0]=(-0.363374730402)*x_ref[0]**o+(0.985470783307)*x_ref[0]+(-0.966941636823)*x_ref[1]**o+(-0.0300574872162)*x_ref[1]+(-0.513481421962)*x_ref[2]**o+(0.869607357947)*x_ref[2]
        ref[1,4,0,1]=(0.646448255445)*x_ref[0]**o+(-0.19357727026)*x_ref[0]+(0.12200322941)*x_ref[1]**o+(0.560785915298)*x_ref[1]+(0.103714094141)*x_ref[2]**o+(0.0235574798014)*x_ref[2]
        ref[1,4,1,0]=(0.658100608473)*x_ref[0]**o+(0.294472394081)*x_ref[0]+(-0.5469350794)*x_ref[1]**o+(0.650910226682)*x_ref[1]+(0.574983452761)*x_ref[2]**o+(-0.885589741854)*x_ref[2]
        ref[1,4,1,1]=(-0.833278077163)*x_ref[0]**o+(-0.612685520243)*x_ref[0]+(0.145850763252)*x_ref[1]**o+(0.710544228623)*x_ref[1]+(-0.237536168655)*x_ref[2]**o+(-0.864417422252)*x_ref[2]
        ref[1,4,2,0]=(0.547214540024)*x_ref[0]**o+(0.259625305608)*x_ref[0]+(0.972583304662)*x_ref[1]**o+(0.0205295851227)*x_ref[1]+(-0.383299704447)*x_ref[2]**o+(0.899668950656)*x_ref[2]
        ref[1,4,2,1]=(-0.0898305090604)*x_ref[0]**o+(0.0583491290278)*x_ref[0]+(0.612069962579)*x_ref[1]**o+(0.00451030954439)*x_ref[1]+(-0.543702915529)*x_ref[2]**o+(0.455742731079)*x_ref[2]
        ref[2,0,0,0]=(0.869996673689)*x_ref[0]**o+(0.786072517277)*x_ref[0]+(0.784064187563)*x_ref[1]**o+(0.215121486636)*x_ref[1]+(0.0167704070196)*x_ref[2]**o+(0.510818786007)*x_ref[2]
        ref[2,0,0,1]=(0.70636305186)*x_ref[0]**o+(-0.774931605014)*x_ref[0]+(-0.566769222836)*x_ref[1]**o+(-0.490039037574)*x_ref[1]+(0.84779476702)*x_ref[2]**o+(-0.277061243254)*x_ref[2]
        ref[2,0,1,0]=(0.71021869545)*x_ref[0]**o+(0.297318884767)*x_ref[0]+(0.393052893616)*x_ref[1]**o+(0.126298772061)*x_ref[1]+(-0.919383761879)*x_ref[2]**o+(0.993331773902)*x_ref[2]
        ref[2,0,1,1]=(-0.422176053109)*x_ref[0]**o+(0.1573461669)*x_ref[0]+(-0.383086926798)*x_ref[1]**o+(0.521025942156)*x_ref[1]+(0.573775622558)*x_ref[2]**o+(-0.424831674319)*x_ref[2]
        ref[2,0,2,0]=(-0.555531945291)*x_ref[0]**o+(0.174293068827)*x_ref[0]+(-0.937417557253)*x_ref[1]**o+(0.0307979382142)*x_ref[1]+(0.0250257663963)*x_ref[2]**o+(-0.0601088874906)*x_ref[2]
        ref[2,0,2,1]=(-0.231539926473)*x_ref[0]**o+(0.943546667559)*x_ref[0]+(-0.453034497139)*x_ref[1]**o+(-0.799323498413)*x_ref[1]+(0.626571998737)*x_ref[2]**o+(0.256254025304)*x_ref[2]
        ref[2,1,0,0]=(0.348839391625)*x_ref[0]**o+(0.747443708496)*x_ref[0]+(-0.632807812622)*x_ref[1]**o+(0.30083165118)*x_ref[1]+(-0.694353354492)*x_ref[2]**o+(-0.384450833928)*x_ref[2]
        ref[2,1,0,1]=(-0.107737403038)*x_ref[0]**o+(0.298410093423)*x_ref[0]+(0.160830729598)*x_ref[1]**o+(-0.627822690248)*x_ref[1]+(0.511195156768)*x_ref[2]**o+(-0.115525478726)*x_ref[2]
        ref[2,1,1,0]=(-0.188820324857)*x_ref[0]**o+(-0.0663606843968)*x_ref[0]+(0.151440137108)*x_ref[1]**o+(0.946232159278)*x_ref[1]+(-0.282849712427)*x_ref[2]**o+(-0.439152308605)*x_ref[2]
        ref[2,1,1,1]=(-0.374881709538)*x_ref[0]**o+(0.745922133487)*x_ref[0]+(-0.0242470036044)*x_ref[1]**o+(-0.711004108426)*x_ref[1]+(0.417440756886)*x_ref[2]**o+(-0.510006752436)*x_ref[2]
        ref[2,1,2,0]=(0.864353278124)*x_ref[0]**o+(0.0809895383191)*x_ref[0]+(0.457075492213)*x_ref[1]**o+(-0.14392908535)*x_ref[1]+(0.784143277535)*x_ref[2]**o+(-0.15341048191)*x_ref[2]
        ref[2,1,2,1]=(0.269079647662)*x_ref[0]**o+(0.0402843915276)*x_ref[0]+(-0.683925387125)*x_ref[1]**o+(0.729804106713)*x_ref[1]+(0.0732202476811)*x_ref[2]**o+(-0.547958146499)*x_ref[2]
        ref[2,2,0,0]=(0.272934512065)*x_ref[0]**o+(-0.882225076784)*x_ref[0]+(-0.405012739084)*x_ref[1]**o+(-0.989715644128)*x_ref[1]+(-0.155608583603)*x_ref[2]**o+(-0.722826494052)*x_ref[2]
        ref[2,2,0,1]=(0.339423408668)*x_ref[0]**o+(-0.330465697311)*x_ref[0]+(-0.692428979863)*x_ref[1]**o+(0.322765870878)*x_ref[1]+(-0.384201212534)*x_ref[2]**o+(-0.0946217242584)*x_ref[2]
        ref[2,2,1,0]=(-0.793318742542)*x_ref[0]**o+(-0.375149299842)*x_ref[0]+(0.793898380613)*x_ref[1]**o+(-0.593499973718)*x_ref[1]+(-0.65815800334)*x_ref[2]**o+(-0.0744951302063)*x_ref[2]
        ref[2,2,1,1]=(0.0279370862836)*x_ref[0]**o+(-0.196535478436)*x_ref[0]+(-0.942547794666)*x_ref[1]**o+(0.994269193299)*x_ref[1]+(0.781181650698)*x_ref[2]**o+(-0.221591113746)*x_ref[2]
        ref[2,2,2,0]=(0.268315410937)*x_ref[0]**o+(-0.53373824299)*x_ref[0]+(0.777981431744)*x_ref[1]**o+(0.226918217122)*x_ref[1]+(-0.561960454199)*x_ref[2]**o+(-0.273776681445)*x_ref[2]
        ref[2,2,2,1]=(0.454454381954)*x_ref[0]**o+(-0.391271318249)*x_ref[0]+(-0.835917284669)*x_ref[1]**o+(-0.42457026252)*x_ref[1]+(-0.848426604454)*x_ref[2]**o+(0.285767008408)*x_ref[2]
        ref[2,3,0,0]=(0.771743337775)*x_ref[0]**o+(-0.157913284247)*x_ref[0]+(0.00417714721394)*x_ref[1]**o+(0.875095301539)*x_ref[1]+(0.965135621963)*x_ref[2]**o+(-0.12209631704)*x_ref[2]
        ref[2,3,0,1]=(0.00527421717368)*x_ref[0]**o+(0.119837014154)*x_ref[0]+(-0.326451619959)*x_ref[1]**o+(-0.179736365312)*x_ref[1]+(-0.659003332514)*x_ref[2]**o+(0.44402878832)*x_ref[2]
        ref[2,3,1,0]=(0.080864510177)*x_ref[0]**o+(-0.910530923746)*x_ref[0]+(-0.065442686402)*x_ref[1]**o+(-0.162655506011)*x_ref[1]+(-0.559755580552)*x_ref[2]**o+(-0.447763051956)*x_ref[2]
        ref[2,3,1,1]=(0.45313869387)*x_ref[0]**o+(0.478878867458)*x_ref[0]+(0.0959944374683)*x_ref[1]**o+(-0.0379945879624)*x_ref[1]+(0.877383707005)*x_ref[2]**o+(0.0037296134062)*x_ref[2]
        ref[2,3,2,0]=(0.187710780263)*x_ref[0]**o+(0.722012532502)*x_ref[0]+(-0.599862911914)*x_ref[1]**o+(-0.974072653552)*x_ref[1]+(0.632960753399)*x_ref[2]**o+(-0.249506081862)*x_ref[2]
        ref[2,3,2,1]=(0.915570663708)*x_ref[0]**o+(-0.590657458642)*x_ref[0]+(-0.221387098604)*x_ref[1]**o+(-0.619974779079)*x_ref[1]+(-0.156470444965)*x_ref[2]**o+(-0.712127719679)*x_ref[2]
        ref[2,4,0,0]=(0.57214571034)*x_ref[0]**o+(0.444107582229)*x_ref[0]+(-0.455561895744)*x_ref[1]**o+(0.350956075511)*x_ref[1]+(-0.060920804543)*x_ref[2]**o+(0.102498779175)*x_ref[2]
        ref[2,4,0,1]=(-0.117233414571)*x_ref[0]**o+(0.336917030674)*x_ref[0]+(-0.544691169707)*x_ref[1]**o+(0.506633531456)*x_ref[1]+(0.193173288491)*x_ref[2]**o+(0.690898822609)*x_ref[2]
        ref[2,4,1,0]=(-0.28899417241)*x_ref[0]**o+(0.930513954761)*x_ref[0]+(-0.564925441731)*x_ref[1]**o+(-0.082228884842)*x_ref[1]+(0.271797239833)*x_ref[2]**o+(-0.293458041224)*x_ref[2]
        ref[2,4,1,1]=(0.303746570859)*x_ref[0]**o+(-0.668303989547)*x_ref[0]+(-0.845649340659)*x_ref[1]**o+(0.826906198662)*x_ref[1]+(-0.499957471874)*x_ref[2]**o+(0.540857798511)*x_ref[2]
        ref[2,4,2,0]=(0.215998594196)*x_ref[0]**o+(0.80715947344)*x_ref[0]+(0.440392341629)*x_ref[1]**o+(0.429949814054)*x_ref[1]+(-0.476601527283)*x_ref[2]**o+(0.21482287961)*x_ref[2]
        ref[2,4,2,1]=(0.744163797227)*x_ref[0]**o+(0.485491624827)*x_ref[0]+(0.0113951682847)*x_ref[1]**o+(-0.219871501999)*x_ref[1]+(-0.960000031149)*x_ref[2]**o+(0.514892677016)*x_ref[2]
        ref[3,0,0,0]=(0.955379977654)*x_ref[0]**o+(0.88166711804)*x_ref[0]+(-0.212371823264)*x_ref[1]**o+(0.61537617198)*x_ref[1]+(0.290444741617)*x_ref[2]**o+(0.669594654028)*x_ref[2]
        ref[3,0,0,1]=(-0.422932658626)*x_ref[0]**o+(0.898327577684)*x_ref[0]+(0.821836925105)*x_ref[1]**o+(-0.777406619005)*x_ref[1]+(0.836622617839)*x_ref[2]**o+(-0.414551285496)*x_ref[2]
        ref[3,0,1,0]=(-0.0922163192963)*x_ref[0]**o+(-0.426319353836)*x_ref[0]+(-0.265988919456)*x_ref[1]**o+(0.0364005746661)*x_ref[1]+(-0.75420756707)*x_ref[2]**o+(0.957027945483)*x_ref[2]
        ref[3,0,1,1]=(0.0501511171161)*x_ref[0]**o+(-0.960573277539)*x_ref[0]+(-0.484587580032)*x_ref[1]**o+(-0.229640708516)*x_ref[1]+(-0.652459896537)*x_ref[2]**o+(-0.59981248766)*x_ref[2]
        ref[3,0,2,0]=(-0.378646761494)*x_ref[0]**o+(-0.0165689331328)*x_ref[0]+(-0.883277009771)*x_ref[1]**o+(-0.947577778882)*x_ref[1]+(0.188169490614)*x_ref[2]**o+(0.079992543794)*x_ref[2]
        ref[3,0,2,1]=(-0.688755441231)*x_ref[0]**o+(0.290810540822)*x_ref[0]+(-0.272544759374)*x_ref[1]**o+(0.412904132844)*x_ref[1]+(-0.851170264556)*x_ref[2]**o+(-0.257140479814)*x_ref[2]
        ref[3,1,0,0]=(0.69577276718)*x_ref[0]**o+(0.777468411586)*x_ref[0]+(0.0291906552252)*x_ref[1]**o+(-0.465217099179)*x_ref[1]+(-0.265314241222)*x_ref[2]**o+(-0.914393443339)*x_ref[2]
        ref[3,1,0,1]=(-0.605826705227)*x_ref[0]**o+(0.786198128441)*x_ref[0]+(-0.413639776886)*x_ref[1]**o+(0.9484840564)*x_ref[1]+(0.660128431709)*x_ref[2]**o+(0.294794064545)*x_ref[2]
        ref[3,1,1,0]=(0.470527104285)*x_ref[0]**o+(-0.673225307232)*x_ref[0]+(0.17528543617)*x_ref[1]**o+(-0.61453857548)*x_ref[1]+(-0.112025849328)*x_ref[2]**o+(0.3719665037)*x_ref[2]
        ref[3,1,1,1]=(0.738351761472)*x_ref[0]**o+(-0.692489261597)*x_ref[0]+(-0.00750387276605)*x_ref[1]**o+(-0.194526069313)*x_ref[1]+(0.206427047745)*x_ref[2]**o+(0.936451437125)*x_ref[2]
        ref[3,1,2,0]=(0.258949551337)*x_ref[0]**o+(0.772440182785)*x_ref[0]+(0.416917261494)*x_ref[1]**o+(0.61679695051)*x_ref[1]+(0.25182107102)*x_ref[2]**o+(0.404770153496)*x_ref[2]
        ref[3,1,2,1]=(0.788357790268)*x_ref[0]**o+(-0.561028019813)*x_ref[0]+(0.842701777125)*x_ref[1]**o+(0.861240216256)*x_ref[1]+(0.633704418848)*x_ref[2]**o+(0.660953804821)*x_ref[2]
        ref[3,2,0,0]=(-0.954508015564)*x_ref[0]**o+(0.770831368915)*x_ref[0]+(0.71068286494)*x_ref[1]**o+(-0.706262271265)*x_ref[1]+(0.000711549487662)*x_ref[2]**o+(0.343413819166)*x_ref[2]
        ref[3,2,0,1]=(-0.486904541331)*x_ref[0]**o+(-0.535805915583)*x_ref[0]+(-0.697643420586)*x_ref[1]**o+(-0.350844416469)*x_ref[1]+(-0.570854441331)*x_ref[2]**o+(0.844325362411)*x_ref[2]
        ref[3,2,1,0]=(0.989057215778)*x_ref[0]**o+(0.822742140223)*x_ref[0]+(0.791720783337)*x_ref[1]**o+(0.0884164131725)*x_ref[1]+(0.337584265788)*x_ref[2]**o+(0.291148268923)*x_ref[2]
        ref[3,2,1,1]=(-0.938554739295)*x_ref[0]**o+(0.154409183584)*x_ref[0]+(0.296554553803)*x_ref[1]**o+(0.0130418768484)*x_ref[1]+(0.440811811112)*x_ref[2]**o+(0.0891377617855)*x_ref[2]
        ref[3,2,2,0]=(0.0966486195221)*x_ref[0]**o+(-0.228831238429)*x_ref[0]+(-0.00903762540107)*x_ref[1]**o+(0.687947867973)*x_ref[1]+(0.29955036712)*x_ref[2]**o+(-0.153152667621)*x_ref[2]
        ref[3,2,2,1]=(0.994414965494)*x_ref[0]**o+(0.787224423165)*x_ref[0]+(-0.572935987357)*x_ref[1]**o+(-0.605628735246)*x_ref[1]+(0.268503770734)*x_ref[2]**o+(-0.32711743813)*x_ref[2]
        ref[3,3,0,0]=(-0.539283649631)*x_ref[0]**o+(0.34809741571)*x_ref[0]+(-0.792527713661)*x_ref[1]**o+(-0.636087526521)*x_ref[1]+(0.765496317046)*x_ref[2]**o+(-0.908214892663)*x_ref[2]
        ref[3,3,0,1]=(0.903206720181)*x_ref[0]**o+(0.761511999734)*x_ref[0]+(0.649680135137)*x_ref[1]**o+(-0.36181233615)*x_ref[1]+(0.563162618539)*x_ref[2]**o+(-0.810859258189)*x_ref[2]
        ref[3,3,1,0]=(-0.0915176526383)*x_ref[0]**o+(0.674948774943)*x_ref[0]+(0.105537180084)*x_ref[1]**o+(0.765853705339)*x_ref[1]+(-0.706270111874)*x_ref[2]**o+(-0.899798116939)*x_ref[2]
        ref[3,3,1,1]=(0.247148082226)*x_ref[0]**o+(-0.252654356068)*x_ref[0]+(0.347296182632)*x_ref[1]**o+(0.809639778301)*x_ref[1]+(0.130358549062)*x_ref[2]**o+(0.772783598594)*x_ref[2]
        ref[3,3,2,0]=(0.282056095613)*x_ref[0]**o+(0.563274162153)*x_ref[0]+(-0.469420374642)*x_ref[1]**o+(-0.00929250493535)*x_ref[1]+(-0.509250535991)*x_ref[2]**o+(0.747602253057)*x_ref[2]
        ref[3,3,2,1]=(-0.40151566811)*x_ref[0]**o+(0.878418940192)*x_ref[0]+(0.548958199553)*x_ref[1]**o+(0.556900329931)*x_ref[1]+(-0.406381288534)*x_ref[2]**o+(-0.767206838398)*x_ref[2]
        ref[3,4,0,0]=(-0.672264988169)*x_ref[0]**o+(0.422502707117)*x_ref[0]+(0.459427006408)*x_ref[1]**o+(0.871928411825)*x_ref[1]+(0.85683710126)*x_ref[2]**o+(0.242796550494)*x_ref[2]
        ref[3,4,0,1]=(-0.821076005446)*x_ref[0]**o+(-0.337616881342)*x_ref[0]+(0.435309707923)*x_ref[1]**o+(0.989009740264)*x_ref[1]+(-0.470842629471)*x_ref[2]**o+(-0.94476330486)*x_ref[2]
        ref[3,4,1,0]=(0.0488301102108)*x_ref[0]**o+(-0.905860578484)*x_ref[0]+(-0.986441486712)*x_ref[1]**o+(0.819623730826)*x_ref[1]+(0.75012214057)*x_ref[2]**o+(-0.460862091553)*x_ref[2]
        ref[3,4,1,1]=(-0.932638473603)*x_ref[0]**o+(-0.882471440278)*x_ref[0]+(-0.498748850986)*x_ref[1]**o+(-0.499262858754)*x_ref[1]+(-0.748745750147)*x_ref[2]**o+(-0.168661287934)*x_ref[2]
        ref[3,4,2,0]=(0.515883010306)*x_ref[0]**o+(0.752629892309)*x_ref[0]+(-0.340202982107)*x_ref[1]**o+(-0.120916620105)*x_ref[1]+(0.193421000969)*x_ref[2]**o+(0.815950266515)*x_ref[2]
        ref[3,4,2,1]=(-0.397114672795)*x_ref[0]**o+(-0.737395536831)*x_ref[0]+(0.589874124917)*x_ref[1]**o+(-0.476304220998)*x_ref[1]+(-0.177825589992)*x_ref[2]**o+(-0.776118337595)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.458775377574)*x[0]**o+(-0.142783482982)*x[0]+(-0.254317031169)*x[1]**o+(0.702505689234)*x[1]
        ref=(0.458775377574)*x_ref[0]**o+(-0.142783482982)*x_ref[0]+(-0.254317031169)*x_ref[1]**o+(0.702505689234)*x_ref[1]
      else:
        s=(0.00533071834267)*x[0]**o+(0.986105059315)*x[0]+(-0.0825201956456)*x[1]**o+(-0.961255985926)*x[1]+(-0.336427126531)*x[2]**o+(0.0885714255159)*x[2]
        ref=(0.00533071834267)*x_ref[0]**o+(0.986105059315)*x_ref[0]+(-0.0825201956456)*x_ref[1]**o+(-0.961255985926)*x_ref[1]+(-0.336427126531)*x_ref[2]**o+(0.0885714255159)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.694815627598)*x[0]**o+(0.644063606633)*x[0]+(0.773087712154)*x[1]**o+(-0.962214094823)*x[1]
        s[1]=(-0.490711112677)*x[0]**o+(-0.87388065943)*x[0]+(-0.21628327344)*x[1]**o+(0.772088460047)*x[1]
        ref[0]=(-0.694815627598)*x_ref[0]**o+(0.644063606633)*x_ref[0]+(0.773087712154)*x_ref[1]**o+(-0.962214094823)*x_ref[1]
        ref[1]=(-0.490711112677)*x_ref[0]**o+(-0.87388065943)*x_ref[0]+(-0.21628327344)*x_ref[1]**o+(0.772088460047)*x_ref[1]
      else:
        s[0]=(-0.330435350709)*x[0]**o+(-0.359970727709)*x[0]+(0.642617648247)*x[1]**o+(-0.0394251455843)*x[1]+(-0.376622312269)*x[2]**o+(-0.541590231201)*x[2]
        s[1]=(0.463360592825)*x[0]**o+(0.498364920452)*x[0]+(0.755121144761)*x[1]**o+(0.743552227571)*x[1]+(0.168251767468)*x[2]**o+(0.584887196144)*x[2]
        ref[0]=(-0.330435350709)*x_ref[0]**o+(-0.359970727709)*x_ref[0]+(0.642617648247)*x_ref[1]**o+(-0.0394251455843)*x_ref[1]+(-0.376622312269)*x_ref[2]**o+(-0.541590231201)*x_ref[2]
        ref[1]=(0.463360592825)*x_ref[0]**o+(0.498364920452)*x_ref[0]+(0.755121144761)*x_ref[1]**o+(0.743552227571)*x_ref[1]+(0.168251767468)*x_ref[2]**o+(0.584887196144)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.488185951644)*x[0]**o+(0.597167519115)*x[0]+(0.943246675163)*x[1]**o+(-0.955863359088)*x[1]
        s[0,1]=(0.178292948477)*x[0]**o+(0.148038709562)*x[0]+(-0.950623527857)*x[1]**o+(0.411442651276)*x[1]
        s[0,2]=(0.134333500047)*x[0]**o+(0.6729047386)*x[0]+(-0.0897085832883)*x[1]**o+(0.372249915499)*x[1]
        s[0,3]=(-0.483653997546)*x[0]**o+(0.407897915806)*x[0]+(0.367153337432)*x[1]**o+(-0.698138025547)*x[1]
        s[0,4]=(-0.842134883349)*x[0]**o+(0.722923275602)*x[0]+(-0.74193204191)*x[1]**o+(-0.0669061392828)*x[1]
        s[1,0]=(0.264069997282)*x[0]**o+(-0.0754854815395)*x[0]+(-0.779935831011)*x[1]**o+(-0.114316965672)*x[1]
        s[1,1]=(0.179470045091)*x[0]**o+(-0.437165723748)*x[0]+(-0.294258032415)*x[1]**o+(0.899095305499)*x[1]
        s[1,2]=(-0.831740706096)*x[0]**o+(-0.149736068183)*x[0]+(-0.73474463786)*x[1]**o+(0.717231742118)*x[1]
        s[1,3]=(0.844816529175)*x[0]**o+(-0.0920760676122)*x[0]+(-0.2115275026)*x[1]**o+(-0.225889157104)*x[1]
        s[1,4]=(-0.0205843755762)*x[0]**o+(-0.718205564182)*x[0]+(-0.819513689268)*x[1]**o+(-0.0732333163823)*x[1]
        s[2,0]=(-0.85716147564)*x[0]**o+(-0.498404753235)*x[0]+(-0.326491535106)*x[1]**o+(0.0574234719803)*x[1]
        s[2,1]=(-0.952095737467)*x[0]**o+(0.473884826053)*x[0]+(-0.531204698326)*x[1]**o+(-0.174344061567)*x[1]
        s[2,2]=(0.999365057788)*x[0]**o+(-0.142639750172)*x[0]+(0.230480966664)*x[1]**o+(0.685914853943)*x[1]
        s[2,3]=(0.166345823031)*x[0]**o+(0.648257656524)*x[0]+(0.654487571777)*x[1]**o+(0.774745511568)*x[1]
        s[2,4]=(-0.0288937922147)*x[0]**o+(-0.979202627577)*x[0]+(0.228136074988)*x[1]**o+(0.738593896247)*x[1]
        s[3,0]=(0.774577798208)*x[0]**o+(0.874351290495)*x[0]+(0.490940628382)*x[1]**o+(-0.526685544309)*x[1]
        s[3,1]=(-0.708233931806)*x[0]**o+(-0.432420027826)*x[0]+(-0.883862641256)*x[1]**o+(0.291054377884)*x[1]
        s[3,2]=(0.359193932932)*x[0]**o+(0.527103835826)*x[0]+(0.072550559316)*x[1]**o+(0.818996368479)*x[1]
        s[3,3]=(-0.39082954898)*x[0]**o+(0.903749706745)*x[0]+(-0.0673256061393)*x[1]**o+(0.714848975884)*x[1]
        s[3,4]=(0.322408322579)*x[0]**o+(-0.429593260425)*x[0]+(0.982604999319)*x[1]**o+(-0.190161113783)*x[1]
        ref[0,0]=(-0.488185951644)*x_ref[0]**o+(0.597167519115)*x_ref[0]+(0.943246675163)*x_ref[1]**o+(-0.955863359088)*x_ref[1]
        ref[0,1]=(0.178292948477)*x_ref[0]**o+(0.148038709562)*x_ref[0]+(-0.950623527857)*x_ref[1]**o+(0.411442651276)*x_ref[1]
        ref[0,2]=(0.134333500047)*x_ref[0]**o+(0.6729047386)*x_ref[0]+(-0.0897085832883)*x_ref[1]**o+(0.372249915499)*x_ref[1]
        ref[0,3]=(-0.483653997546)*x_ref[0]**o+(0.407897915806)*x_ref[0]+(0.367153337432)*x_ref[1]**o+(-0.698138025547)*x_ref[1]
        ref[0,4]=(-0.842134883349)*x_ref[0]**o+(0.722923275602)*x_ref[0]+(-0.74193204191)*x_ref[1]**o+(-0.0669061392828)*x_ref[1]
        ref[1,0]=(0.264069997282)*x_ref[0]**o+(-0.0754854815395)*x_ref[0]+(-0.779935831011)*x_ref[1]**o+(-0.114316965672)*x_ref[1]
        ref[1,1]=(0.179470045091)*x_ref[0]**o+(-0.437165723748)*x_ref[0]+(-0.294258032415)*x_ref[1]**o+(0.899095305499)*x_ref[1]
        ref[1,2]=(-0.831740706096)*x_ref[0]**o+(-0.149736068183)*x_ref[0]+(-0.73474463786)*x_ref[1]**o+(0.717231742118)*x_ref[1]
        ref[1,3]=(0.844816529175)*x_ref[0]**o+(-0.0920760676122)*x_ref[0]+(-0.2115275026)*x_ref[1]**o+(-0.225889157104)*x_ref[1]
        ref[1,4]=(-0.0205843755762)*x_ref[0]**o+(-0.718205564182)*x_ref[0]+(-0.819513689268)*x_ref[1]**o+(-0.0732333163823)*x_ref[1]
        ref[2,0]=(-0.85716147564)*x_ref[0]**o+(-0.498404753235)*x_ref[0]+(-0.326491535106)*x_ref[1]**o+(0.0574234719803)*x_ref[1]
        ref[2,1]=(-0.952095737467)*x_ref[0]**o+(0.473884826053)*x_ref[0]+(-0.531204698326)*x_ref[1]**o+(-0.174344061567)*x_ref[1]
        ref[2,2]=(0.999365057788)*x_ref[0]**o+(-0.142639750172)*x_ref[0]+(0.230480966664)*x_ref[1]**o+(0.685914853943)*x_ref[1]
        ref[2,3]=(0.166345823031)*x_ref[0]**o+(0.648257656524)*x_ref[0]+(0.654487571777)*x_ref[1]**o+(0.774745511568)*x_ref[1]
        ref[2,4]=(-0.0288937922147)*x_ref[0]**o+(-0.979202627577)*x_ref[0]+(0.228136074988)*x_ref[1]**o+(0.738593896247)*x_ref[1]
        ref[3,0]=(0.774577798208)*x_ref[0]**o+(0.874351290495)*x_ref[0]+(0.490940628382)*x_ref[1]**o+(-0.526685544309)*x_ref[1]
        ref[3,1]=(-0.708233931806)*x_ref[0]**o+(-0.432420027826)*x_ref[0]+(-0.883862641256)*x_ref[1]**o+(0.291054377884)*x_ref[1]
        ref[3,2]=(0.359193932932)*x_ref[0]**o+(0.527103835826)*x_ref[0]+(0.072550559316)*x_ref[1]**o+(0.818996368479)*x_ref[1]
        ref[3,3]=(-0.39082954898)*x_ref[0]**o+(0.903749706745)*x_ref[0]+(-0.0673256061393)*x_ref[1]**o+(0.714848975884)*x_ref[1]
        ref[3,4]=(0.322408322579)*x_ref[0]**o+(-0.429593260425)*x_ref[0]+(0.982604999319)*x_ref[1]**o+(-0.190161113783)*x_ref[1]
      else:
        s[0,0]=(0.232804655799)*x[0]**o+(0.880141618164)*x[0]+(0.223221634005)*x[1]**o+(0.452983026652)*x[1]+(-0.573854985537)*x[2]**o+(-0.653809009284)*x[2]
        s[0,1]=(0.198858642702)*x[0]**o+(-0.314878524755)*x[0]+(-0.255927629298)*x[1]**o+(0.928004497503)*x[1]+(0.506963465945)*x[2]**o+(-0.47102323414)*x[2]
        s[0,2]=(0.308455355109)*x[0]**o+(-0.880701737691)*x[0]+(0.381469789283)*x[1]**o+(0.801918317713)*x[1]+(0.834098938062)*x[2]**o+(0.600601216943)*x[2]
        s[0,3]=(-0.469143012993)*x[0]**o+(-0.325961388166)*x[0]+(-0.0776077288313)*x[1]**o+(-0.0488948311893)*x[1]+(-0.0667465969735)*x[2]**o+(0.590630229618)*x[2]
        s[0,4]=(0.460884530892)*x[0]**o+(0.981078628528)*x[0]+(0.819191927543)*x[1]**o+(-0.225345065568)*x[1]+(-0.461177702761)*x[2]**o+(0.973442359563)*x[2]
        s[1,0]=(0.277068448313)*x[0]**o+(0.955364825669)*x[0]+(-0.912489837333)*x[1]**o+(-0.267268491011)*x[1]+(-0.847314081076)*x[2]**o+(-0.975760667649)*x[2]
        s[1,1]=(0.257880013047)*x[0]**o+(-0.343779864363)*x[0]+(0.616718460823)*x[1]**o+(-0.739452662888)*x[1]+(0.347386315181)*x[2]**o+(0.615160648765)*x[2]
        s[1,2]=(0.988114419561)*x[0]**o+(0.882801837855)*x[0]+(0.617420713454)*x[1]**o+(0.653402104375)*x[1]+(-0.0585299332344)*x[2]**o+(-0.400646895518)*x[2]
        s[1,3]=(-0.421049210036)*x[0]**o+(-0.0090620420347)*x[0]+(-0.54358224088)*x[1]**o+(0.254744898368)*x[1]+(-0.356411594187)*x[2]**o+(0.490942960051)*x[2]
        s[1,4]=(0.266613733548)*x[0]**o+(-0.953213213138)*x[0]+(0.659951033053)*x[1]**o+(0.0887943186214)*x[1]+(0.232221116015)*x[2]**o+(0.921710239665)*x[2]
        s[2,0]=(-0.278908005525)*x[0]**o+(-0.67161337084)*x[0]+(0.682065318715)*x[1]**o+(-0.184380015721)*x[1]+(0.72644685348)*x[2]**o+(0.951051205648)*x[2]
        s[2,1]=(-0.90781744456)*x[0]**o+(0.0077008898356)*x[0]+(0.671790725713)*x[1]**o+(0.853379797278)*x[1]+(-0.0764999700365)*x[2]**o+(-0.98659879948)*x[2]
        s[2,2]=(-0.913635850698)*x[0]**o+(-0.590635380841)*x[0]+(-0.663640142768)*x[1]**o+(-0.406156532072)*x[1]+(-0.784260406208)*x[2]**o+(-0.645167192291)*x[2]
        s[2,3]=(0.860741367256)*x[0]**o+(-0.0755200276519)*x[0]+(-0.501330034283)*x[1]**o+(0.774551262816)*x[1]+(-0.140777760645)*x[2]**o+(0.412398487198)*x[2]
        s[2,4]=(0.755189880126)*x[0]**o+(-0.688143708643)*x[0]+(0.497944488811)*x[1]**o+(-0.957053638137)*x[1]+(-0.116257980976)*x[2]**o+(0.356020915088)*x[2]
        s[3,0]=(-0.856430646919)*x[0]**o+(-0.0955546373651)*x[0]+(-0.588927939248)*x[1]**o+(0.207587246072)*x[1]+(-0.424773091145)*x[2]**o+(-0.0793349369172)*x[2]
        s[3,1]=(-0.0814274211195)*x[0]**o+(-0.103989324245)*x[0]+(-0.651528552136)*x[1]**o+(-0.757919391603)*x[1]+(0.142468149009)*x[2]**o+(0.642940283508)*x[2]
        s[3,2]=(0.266690596451)*x[0]**o+(-0.309503686303)*x[0]+(0.210541256761)*x[1]**o+(-0.363519101414)*x[1]+(-0.703156838318)*x[2]**o+(0.089872265658)*x[2]
        s[3,3]=(0.772726367524)*x[0]**o+(0.621843084065)*x[0]+(0.347047977396)*x[1]**o+(0.124103441993)*x[1]+(-0.626843590322)*x[2]**o+(0.215729189789)*x[2]
        s[3,4]=(-0.622360125477)*x[0]**o+(0.260262244242)*x[0]+(-0.312822631976)*x[1]**o+(-0.559674699162)*x[1]+(-0.0776492554888)*x[2]**o+(-0.156182340995)*x[2]
        ref[0,0]=(0.232804655799)*x_ref[0]**o+(0.880141618164)*x_ref[0]+(0.223221634005)*x_ref[1]**o+(0.452983026652)*x_ref[1]+(-0.573854985537)*x_ref[2]**o+(-0.653809009284)*x_ref[2]
        ref[0,1]=(0.198858642702)*x_ref[0]**o+(-0.314878524755)*x_ref[0]+(-0.255927629298)*x_ref[1]**o+(0.928004497503)*x_ref[1]+(0.506963465945)*x_ref[2]**o+(-0.47102323414)*x_ref[2]
        ref[0,2]=(0.308455355109)*x_ref[0]**o+(-0.880701737691)*x_ref[0]+(0.381469789283)*x_ref[1]**o+(0.801918317713)*x_ref[1]+(0.834098938062)*x_ref[2]**o+(0.600601216943)*x_ref[2]
        ref[0,3]=(-0.469143012993)*x_ref[0]**o+(-0.325961388166)*x_ref[0]+(-0.0776077288313)*x_ref[1]**o+(-0.0488948311893)*x_ref[1]+(-0.0667465969735)*x_ref[2]**o+(0.590630229618)*x_ref[2]
        ref[0,4]=(0.460884530892)*x_ref[0]**o+(0.981078628528)*x_ref[0]+(0.819191927543)*x_ref[1]**o+(-0.225345065568)*x_ref[1]+(-0.461177702761)*x_ref[2]**o+(0.973442359563)*x_ref[2]
        ref[1,0]=(0.277068448313)*x_ref[0]**o+(0.955364825669)*x_ref[0]+(-0.912489837333)*x_ref[1]**o+(-0.267268491011)*x_ref[1]+(-0.847314081076)*x_ref[2]**o+(-0.975760667649)*x_ref[2]
        ref[1,1]=(0.257880013047)*x_ref[0]**o+(-0.343779864363)*x_ref[0]+(0.616718460823)*x_ref[1]**o+(-0.739452662888)*x_ref[1]+(0.347386315181)*x_ref[2]**o+(0.615160648765)*x_ref[2]
        ref[1,2]=(0.988114419561)*x_ref[0]**o+(0.882801837855)*x_ref[0]+(0.617420713454)*x_ref[1]**o+(0.653402104375)*x_ref[1]+(-0.0585299332344)*x_ref[2]**o+(-0.400646895518)*x_ref[2]
        ref[1,3]=(-0.421049210036)*x_ref[0]**o+(-0.0090620420347)*x_ref[0]+(-0.54358224088)*x_ref[1]**o+(0.254744898368)*x_ref[1]+(-0.356411594187)*x_ref[2]**o+(0.490942960051)*x_ref[2]
        ref[1,4]=(0.266613733548)*x_ref[0]**o+(-0.953213213138)*x_ref[0]+(0.659951033053)*x_ref[1]**o+(0.0887943186214)*x_ref[1]+(0.232221116015)*x_ref[2]**o+(0.921710239665)*x_ref[2]
        ref[2,0]=(-0.278908005525)*x_ref[0]**o+(-0.67161337084)*x_ref[0]+(0.682065318715)*x_ref[1]**o+(-0.184380015721)*x_ref[1]+(0.72644685348)*x_ref[2]**o+(0.951051205648)*x_ref[2]
        ref[2,1]=(-0.90781744456)*x_ref[0]**o+(0.0077008898356)*x_ref[0]+(0.671790725713)*x_ref[1]**o+(0.853379797278)*x_ref[1]+(-0.0764999700365)*x_ref[2]**o+(-0.98659879948)*x_ref[2]
        ref[2,2]=(-0.913635850698)*x_ref[0]**o+(-0.590635380841)*x_ref[0]+(-0.663640142768)*x_ref[1]**o+(-0.406156532072)*x_ref[1]+(-0.784260406208)*x_ref[2]**o+(-0.645167192291)*x_ref[2]
        ref[2,3]=(0.860741367256)*x_ref[0]**o+(-0.0755200276519)*x_ref[0]+(-0.501330034283)*x_ref[1]**o+(0.774551262816)*x_ref[1]+(-0.140777760645)*x_ref[2]**o+(0.412398487198)*x_ref[2]
        ref[2,4]=(0.755189880126)*x_ref[0]**o+(-0.688143708643)*x_ref[0]+(0.497944488811)*x_ref[1]**o+(-0.957053638137)*x_ref[1]+(-0.116257980976)*x_ref[2]**o+(0.356020915088)*x_ref[2]
        ref[3,0]=(-0.856430646919)*x_ref[0]**o+(-0.0955546373651)*x_ref[0]+(-0.588927939248)*x_ref[1]**o+(0.207587246072)*x_ref[1]+(-0.424773091145)*x_ref[2]**o+(-0.0793349369172)*x_ref[2]
        ref[3,1]=(-0.0814274211195)*x_ref[0]**o+(-0.103989324245)*x_ref[0]+(-0.651528552136)*x_ref[1]**o+(-0.757919391603)*x_ref[1]+(0.142468149009)*x_ref[2]**o+(0.642940283508)*x_ref[2]
        ref[3,2]=(0.266690596451)*x_ref[0]**o+(-0.309503686303)*x_ref[0]+(0.210541256761)*x_ref[1]**o+(-0.363519101414)*x_ref[1]+(-0.703156838318)*x_ref[2]**o+(0.089872265658)*x_ref[2]
        ref[3,3]=(0.772726367524)*x_ref[0]**o+(0.621843084065)*x_ref[0]+(0.347047977396)*x_ref[1]**o+(0.124103441993)*x_ref[1]+(-0.626843590322)*x_ref[2]**o+(0.215729189789)*x_ref[2]
        ref[3,4]=(-0.622360125477)*x_ref[0]**o+(0.260262244242)*x_ref[0]+(-0.312822631976)*x_ref[1]**o+(-0.559674699162)*x_ref[1]+(-0.0776492554888)*x_ref[2]**o+(-0.156182340995)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.87064070758)*x[0]**o+(-0.566055900642)*x[0]+(-0.917420111598)*x[1]**o+(0.0551635734945)*x[1]
        s[0,0,1]=(-0.117176717885)*x[0]**o+(-0.822793678972)*x[0]+(-0.41376673807)*x[1]**o+(-0.782915075242)*x[1]
        s[0,1,0]=(0.309660501821)*x[0]**o+(-0.705257783376)*x[0]+(0.436897623093)*x[1]**o+(0.624205776859)*x[1]
        s[0,1,1]=(-0.338652342709)*x[0]**o+(-0.783666910448)*x[0]+(0.869895700694)*x[1]**o+(-0.106196197398)*x[1]
        s[1,0,0]=(-0.418615528975)*x[0]**o+(-0.821973439514)*x[0]+(0.715494828047)*x[1]**o+(0.717827721559)*x[1]
        s[1,0,1]=(-0.0319361959791)*x[0]**o+(0.60209433014)*x[0]+(-0.297989638002)*x[1]**o+(0.133472987028)*x[1]
        s[1,1,0]=(0.874169575419)*x[0]**o+(-0.804771710143)*x[0]+(-0.956980034287)*x[1]**o+(0.60848322694)*x[1]
        s[1,1,1]=(-0.96607032622)*x[0]**o+(0.58719800455)*x[0]+(0.432119827191)*x[1]**o+(-0.0515514190014)*x[1]
        s[2,0,0]=(0.92256117661)*x[0]**o+(0.286836406901)*x[0]+(-0.566218833636)*x[1]**o+(-0.318500128005)*x[1]
        s[2,0,1]=(-0.295696171332)*x[0]**o+(-0.847067252777)*x[0]+(-0.0214499844044)*x[1]**o+(0.56142047851)*x[1]
        s[2,1,0]=(-0.278598238494)*x[0]**o+(0.775672819752)*x[0]+(-0.594303030563)*x[1]**o+(-0.764128845442)*x[1]
        s[2,1,1]=(0.195219876382)*x[0]**o+(0.31911545942)*x[0]+(0.414261884794)*x[1]**o+(0.430367930637)*x[1]
        s[3,0,0]=(-0.957282546866)*x[0]**o+(-0.227277741511)*x[0]+(0.386584028319)*x[1]**o+(0.166263307275)*x[1]
        s[3,0,1]=(-0.682200893175)*x[0]**o+(-0.845531035272)*x[0]+(0.938083890385)*x[1]**o+(-0.674378235835)*x[1]
        s[3,1,0]=(0.544986524999)*x[0]**o+(0.529710178839)*x[0]+(-0.489344340247)*x[1]**o+(0.476288384952)*x[1]
        s[3,1,1]=(-0.680566273064)*x[0]**o+(0.806369975569)*x[0]+(0.305516322897)*x[1]**o+(-0.51330834934)*x[1]
        s[4,0,0]=(-0.465052752387)*x[0]**o+(-0.00313229607036)*x[0]+(0.499351385096)*x[1]**o+(-0.401491711188)*x[1]
        s[4,0,1]=(0.27236428546)*x[0]**o+(-0.278717867273)*x[0]+(-0.96726186532)*x[1]**o+(-0.317275683828)*x[1]
        s[4,1,0]=(0.477975953682)*x[0]**o+(0.772356007674)*x[0]+(0.285581555335)*x[1]**o+(0.872551473549)*x[1]
        s[4,1,1]=(-0.181196107871)*x[0]**o+(-0.473439467597)*x[0]+(0.90210175622)*x[1]**o+(-0.378511577829)*x[1]
        s[5,0,0]=(0.633625574073)*x[0]**o+(-0.563894010102)*x[0]+(-0.302946833681)*x[1]**o+(-0.601584229143)*x[1]
        s[5,0,1]=(-0.754395228954)*x[0]**o+(0.317899599612)*x[0]+(-0.612069470764)*x[1]**o+(0.937528317073)*x[1]
        s[5,1,0]=(0.0360449364159)*x[0]**o+(0.0944318504395)*x[0]+(0.745586397044)*x[1]**o+(-0.839531808187)*x[1]
        s[5,1,1]=(0.620453262074)*x[0]**o+(-0.199107011047)*x[0]+(0.754391775519)*x[1]**o+(0.853229319102)*x[1]
        ref[0,0,0]=(0.87064070758)*x_ref[0]**o+(-0.566055900642)*x_ref[0]+(-0.917420111598)*x_ref[1]**o+(0.0551635734945)*x_ref[1]
        ref[0,0,1]=(-0.117176717885)*x_ref[0]**o+(-0.822793678972)*x_ref[0]+(-0.41376673807)*x_ref[1]**o+(-0.782915075242)*x_ref[1]
        ref[0,1,0]=(0.309660501821)*x_ref[0]**o+(-0.705257783376)*x_ref[0]+(0.436897623093)*x_ref[1]**o+(0.624205776859)*x_ref[1]
        ref[0,1,1]=(-0.338652342709)*x_ref[0]**o+(-0.783666910448)*x_ref[0]+(0.869895700694)*x_ref[1]**o+(-0.106196197398)*x_ref[1]
        ref[1,0,0]=(-0.418615528975)*x_ref[0]**o+(-0.821973439514)*x_ref[0]+(0.715494828047)*x_ref[1]**o+(0.717827721559)*x_ref[1]
        ref[1,0,1]=(-0.0319361959791)*x_ref[0]**o+(0.60209433014)*x_ref[0]+(-0.297989638002)*x_ref[1]**o+(0.133472987028)*x_ref[1]
        ref[1,1,0]=(0.874169575419)*x_ref[0]**o+(-0.804771710143)*x_ref[0]+(-0.956980034287)*x_ref[1]**o+(0.60848322694)*x_ref[1]
        ref[1,1,1]=(-0.96607032622)*x_ref[0]**o+(0.58719800455)*x_ref[0]+(0.432119827191)*x_ref[1]**o+(-0.0515514190014)*x_ref[1]
        ref[2,0,0]=(0.92256117661)*x_ref[0]**o+(0.286836406901)*x_ref[0]+(-0.566218833636)*x_ref[1]**o+(-0.318500128005)*x_ref[1]
        ref[2,0,1]=(-0.295696171332)*x_ref[0]**o+(-0.847067252777)*x_ref[0]+(-0.0214499844044)*x_ref[1]**o+(0.56142047851)*x_ref[1]
        ref[2,1,0]=(-0.278598238494)*x_ref[0]**o+(0.775672819752)*x_ref[0]+(-0.594303030563)*x_ref[1]**o+(-0.764128845442)*x_ref[1]
        ref[2,1,1]=(0.195219876382)*x_ref[0]**o+(0.31911545942)*x_ref[0]+(0.414261884794)*x_ref[1]**o+(0.430367930637)*x_ref[1]
        ref[3,0,0]=(-0.957282546866)*x_ref[0]**o+(-0.227277741511)*x_ref[0]+(0.386584028319)*x_ref[1]**o+(0.166263307275)*x_ref[1]
        ref[3,0,1]=(-0.682200893175)*x_ref[0]**o+(-0.845531035272)*x_ref[0]+(0.938083890385)*x_ref[1]**o+(-0.674378235835)*x_ref[1]
        ref[3,1,0]=(0.544986524999)*x_ref[0]**o+(0.529710178839)*x_ref[0]+(-0.489344340247)*x_ref[1]**o+(0.476288384952)*x_ref[1]
        ref[3,1,1]=(-0.680566273064)*x_ref[0]**o+(0.806369975569)*x_ref[0]+(0.305516322897)*x_ref[1]**o+(-0.51330834934)*x_ref[1]
        ref[4,0,0]=(-0.465052752387)*x_ref[0]**o+(-0.00313229607036)*x_ref[0]+(0.499351385096)*x_ref[1]**o+(-0.401491711188)*x_ref[1]
        ref[4,0,1]=(0.27236428546)*x_ref[0]**o+(-0.278717867273)*x_ref[0]+(-0.96726186532)*x_ref[1]**o+(-0.317275683828)*x_ref[1]
        ref[4,1,0]=(0.477975953682)*x_ref[0]**o+(0.772356007674)*x_ref[0]+(0.285581555335)*x_ref[1]**o+(0.872551473549)*x_ref[1]
        ref[4,1,1]=(-0.181196107871)*x_ref[0]**o+(-0.473439467597)*x_ref[0]+(0.90210175622)*x_ref[1]**o+(-0.378511577829)*x_ref[1]
        ref[5,0,0]=(0.633625574073)*x_ref[0]**o+(-0.563894010102)*x_ref[0]+(-0.302946833681)*x_ref[1]**o+(-0.601584229143)*x_ref[1]
        ref[5,0,1]=(-0.754395228954)*x_ref[0]**o+(0.317899599612)*x_ref[0]+(-0.612069470764)*x_ref[1]**o+(0.937528317073)*x_ref[1]
        ref[5,1,0]=(0.0360449364159)*x_ref[0]**o+(0.0944318504395)*x_ref[0]+(0.745586397044)*x_ref[1]**o+(-0.839531808187)*x_ref[1]
        ref[5,1,1]=(0.620453262074)*x_ref[0]**o+(-0.199107011047)*x_ref[0]+(0.754391775519)*x_ref[1]**o+(0.853229319102)*x_ref[1]
      else:
        s[0,0,0]=(-0.384948393134)*x[0]**o+(-0.961967178877)*x[0]+(0.801941664706)*x[1]**o+(0.341078763066)*x[1]+(-0.965751483242)*x[2]**o+(-0.335041713825)*x[2]
        s[0,0,1]=(-0.67569110343)*x[0]**o+(-0.607731651311)*x[0]+(-0.769155006252)*x[1]**o+(-0.951474117563)*x[1]+(0.346757382654)*x[2]**o+(0.468605905517)*x[2]
        s[0,1,0]=(0.357827714378)*x[0]**o+(0.531884637396)*x[0]+(-0.340932021448)*x[1]**o+(0.107907346017)*x[1]+(0.899991044816)*x[2]**o+(-0.619238986401)*x[2]
        s[0,1,1]=(0.92143345512)*x[0]**o+(0.0319943191208)*x[0]+(0.143931354465)*x[1]**o+(-0.827543395412)*x[1]+(0.670291174904)*x[2]**o+(0.323131021196)*x[2]
        s[1,0,0]=(0.253706012955)*x[0]**o+(0.920629791188)*x[0]+(-0.221548307026)*x[1]**o+(0.920275162623)*x[1]+(0.159384070106)*x[2]**o+(0.914780524212)*x[2]
        s[1,0,1]=(-0.145041193856)*x[0]**o+(-0.777916320748)*x[0]+(-0.270423261412)*x[1]**o+(0.948002693835)*x[1]+(-0.872437269572)*x[2]**o+(-0.389969462287)*x[2]
        s[1,1,0]=(-0.0258988000508)*x[0]**o+(0.27202895614)*x[0]+(-0.979886715231)*x[1]**o+(-0.102396555104)*x[1]+(-0.97303335804)*x[2]**o+(0.0848739951879)*x[2]
        s[1,1,1]=(-0.171608662455)*x[0]**o+(-0.0499287952471)*x[0]+(-0.780462044785)*x[1]**o+(0.506400671006)*x[1]+(-0.678439420533)*x[2]**o+(-0.609358013123)*x[2]
        s[2,0,0]=(0.715810623102)*x[0]**o+(0.478297229523)*x[0]+(-0.700770389598)*x[1]**o+(0.279096508315)*x[1]+(-0.385130762383)*x[2]**o+(0.46599268203)*x[2]
        s[2,0,1]=(0.47346229532)*x[0]**o+(0.304487177002)*x[0]+(0.0242783060008)*x[1]**o+(0.772667749027)*x[1]+(-0.42037502877)*x[2]**o+(-0.912335922325)*x[2]
        s[2,1,0]=(0.509168300969)*x[0]**o+(-0.746969033155)*x[0]+(-0.907693621932)*x[1]**o+(-0.642615676429)*x[1]+(-0.439912454318)*x[2]**o+(-0.452216863742)*x[2]
        s[2,1,1]=(-0.410444089831)*x[0]**o+(-0.147372964881)*x[0]+(0.850150090851)*x[1]**o+(-0.988475742361)*x[1]+(-0.462654428176)*x[2]**o+(-0.940338508948)*x[2]
        s[3,0,0]=(-0.794220586578)*x[0]**o+(0.922257088538)*x[0]+(0.0134033910944)*x[1]**o+(-0.95038627404)*x[1]+(-0.609227492432)*x[2]**o+(0.558321765879)*x[2]
        s[3,0,1]=(-0.304420694648)*x[0]**o+(0.244625817097)*x[0]+(-0.0933192976308)*x[1]**o+(-0.874643761558)*x[1]+(0.936742259878)*x[2]**o+(0.895320681558)*x[2]
        s[3,1,0]=(-0.202315611337)*x[0]**o+(-0.2365832224)*x[0]+(-0.480962593573)*x[1]**o+(-0.210912260068)*x[1]+(-0.000213415776588)*x[2]**o+(0.259609090936)*x[2]
        s[3,1,1]=(-0.612351015469)*x[0]**o+(0.546469916272)*x[0]+(-0.166009868547)*x[1]**o+(0.931501987972)*x[1]+(-0.330366993065)*x[2]**o+(0.435322489369)*x[2]
        s[4,0,0]=(0.124588792911)*x[0]**o+(0.0472809655836)*x[0]+(0.126649125977)*x[1]**o+(-0.980286766046)*x[1]+(-0.708140145708)*x[2]**o+(0.105934454391)*x[2]
        s[4,0,1]=(-0.967814569406)*x[0]**o+(-0.606176160393)*x[0]+(0.0252434349925)*x[1]**o+(0.606791676748)*x[1]+(-0.615219396992)*x[2]**o+(0.462327857208)*x[2]
        s[4,1,0]=(0.47926323984)*x[0]**o+(0.493300730667)*x[0]+(0.940191066665)*x[1]**o+(0.245800539336)*x[1]+(-0.694101640383)*x[2]**o+(0.200348996082)*x[2]
        s[4,1,1]=(0.552067136819)*x[0]**o+(0.351346598128)*x[0]+(0.39731191883)*x[1]**o+(-0.70236991903)*x[1]+(-0.625528596695)*x[2]**o+(0.798534719203)*x[2]
        s[5,0,0]=(0.167335359474)*x[0]**o+(-0.232843920341)*x[0]+(-0.126773325792)*x[1]**o+(-0.203898990499)*x[1]+(0.839140245989)*x[2]**o+(-0.728050229083)*x[2]
        s[5,0,1]=(-0.664841432025)*x[0]**o+(-0.0489116955008)*x[0]+(0.685379040894)*x[1]**o+(0.370521174212)*x[1]+(-0.742375864799)*x[2]**o+(-0.880322708097)*x[2]
        s[5,1,0]=(0.653110528414)*x[0]**o+(0.602025573916)*x[0]+(0.912361317481)*x[1]**o+(0.534616035231)*x[1]+(-0.103102947904)*x[2]**o+(0.531938071791)*x[2]
        s[5,1,1]=(-0.217856022054)*x[0]**o+(0.11102291479)*x[0]+(-0.166664940768)*x[1]**o+(-0.812763150681)*x[1]+(0.442963442721)*x[2]**o+(0.706518891053)*x[2]
        ref[0,0,0]=(-0.384948393134)*x_ref[0]**o+(-0.961967178877)*x_ref[0]+(0.801941664706)*x_ref[1]**o+(0.341078763066)*x_ref[1]+(-0.965751483242)*x_ref[2]**o+(-0.335041713825)*x_ref[2]
        ref[0,0,1]=(-0.67569110343)*x_ref[0]**o+(-0.607731651311)*x_ref[0]+(-0.769155006252)*x_ref[1]**o+(-0.951474117563)*x_ref[1]+(0.346757382654)*x_ref[2]**o+(0.468605905517)*x_ref[2]
        ref[0,1,0]=(0.357827714378)*x_ref[0]**o+(0.531884637396)*x_ref[0]+(-0.340932021448)*x_ref[1]**o+(0.107907346017)*x_ref[1]+(0.899991044816)*x_ref[2]**o+(-0.619238986401)*x_ref[2]
        ref[0,1,1]=(0.92143345512)*x_ref[0]**o+(0.0319943191208)*x_ref[0]+(0.143931354465)*x_ref[1]**o+(-0.827543395412)*x_ref[1]+(0.670291174904)*x_ref[2]**o+(0.323131021196)*x_ref[2]
        ref[1,0,0]=(0.253706012955)*x_ref[0]**o+(0.920629791188)*x_ref[0]+(-0.221548307026)*x_ref[1]**o+(0.920275162623)*x_ref[1]+(0.159384070106)*x_ref[2]**o+(0.914780524212)*x_ref[2]
        ref[1,0,1]=(-0.145041193856)*x_ref[0]**o+(-0.777916320748)*x_ref[0]+(-0.270423261412)*x_ref[1]**o+(0.948002693835)*x_ref[1]+(-0.872437269572)*x_ref[2]**o+(-0.389969462287)*x_ref[2]
        ref[1,1,0]=(-0.0258988000508)*x_ref[0]**o+(0.27202895614)*x_ref[0]+(-0.979886715231)*x_ref[1]**o+(-0.102396555104)*x_ref[1]+(-0.97303335804)*x_ref[2]**o+(0.0848739951879)*x_ref[2]
        ref[1,1,1]=(-0.171608662455)*x_ref[0]**o+(-0.0499287952471)*x_ref[0]+(-0.780462044785)*x_ref[1]**o+(0.506400671006)*x_ref[1]+(-0.678439420533)*x_ref[2]**o+(-0.609358013123)*x_ref[2]
        ref[2,0,0]=(0.715810623102)*x_ref[0]**o+(0.478297229523)*x_ref[0]+(-0.700770389598)*x_ref[1]**o+(0.279096508315)*x_ref[1]+(-0.385130762383)*x_ref[2]**o+(0.46599268203)*x_ref[2]
        ref[2,0,1]=(0.47346229532)*x_ref[0]**o+(0.304487177002)*x_ref[0]+(0.0242783060008)*x_ref[1]**o+(0.772667749027)*x_ref[1]+(-0.42037502877)*x_ref[2]**o+(-0.912335922325)*x_ref[2]
        ref[2,1,0]=(0.509168300969)*x_ref[0]**o+(-0.746969033155)*x_ref[0]+(-0.907693621932)*x_ref[1]**o+(-0.642615676429)*x_ref[1]+(-0.439912454318)*x_ref[2]**o+(-0.452216863742)*x_ref[2]
        ref[2,1,1]=(-0.410444089831)*x_ref[0]**o+(-0.147372964881)*x_ref[0]+(0.850150090851)*x_ref[1]**o+(-0.988475742361)*x_ref[1]+(-0.462654428176)*x_ref[2]**o+(-0.940338508948)*x_ref[2]
        ref[3,0,0]=(-0.794220586578)*x_ref[0]**o+(0.922257088538)*x_ref[0]+(0.0134033910944)*x_ref[1]**o+(-0.95038627404)*x_ref[1]+(-0.609227492432)*x_ref[2]**o+(0.558321765879)*x_ref[2]
        ref[3,0,1]=(-0.304420694648)*x_ref[0]**o+(0.244625817097)*x_ref[0]+(-0.0933192976308)*x_ref[1]**o+(-0.874643761558)*x_ref[1]+(0.936742259878)*x_ref[2]**o+(0.895320681558)*x_ref[2]
        ref[3,1,0]=(-0.202315611337)*x_ref[0]**o+(-0.2365832224)*x_ref[0]+(-0.480962593573)*x_ref[1]**o+(-0.210912260068)*x_ref[1]+(-0.000213415776588)*x_ref[2]**o+(0.259609090936)*x_ref[2]
        ref[3,1,1]=(-0.612351015469)*x_ref[0]**o+(0.546469916272)*x_ref[0]+(-0.166009868547)*x_ref[1]**o+(0.931501987972)*x_ref[1]+(-0.330366993065)*x_ref[2]**o+(0.435322489369)*x_ref[2]
        ref[4,0,0]=(0.124588792911)*x_ref[0]**o+(0.0472809655836)*x_ref[0]+(0.126649125977)*x_ref[1]**o+(-0.980286766046)*x_ref[1]+(-0.708140145708)*x_ref[2]**o+(0.105934454391)*x_ref[2]
        ref[4,0,1]=(-0.967814569406)*x_ref[0]**o+(-0.606176160393)*x_ref[0]+(0.0252434349925)*x_ref[1]**o+(0.606791676748)*x_ref[1]+(-0.615219396992)*x_ref[2]**o+(0.462327857208)*x_ref[2]
        ref[4,1,0]=(0.47926323984)*x_ref[0]**o+(0.493300730667)*x_ref[0]+(0.940191066665)*x_ref[1]**o+(0.245800539336)*x_ref[1]+(-0.694101640383)*x_ref[2]**o+(0.200348996082)*x_ref[2]
        ref[4,1,1]=(0.552067136819)*x_ref[0]**o+(0.351346598128)*x_ref[0]+(0.39731191883)*x_ref[1]**o+(-0.70236991903)*x_ref[1]+(-0.625528596695)*x_ref[2]**o+(0.798534719203)*x_ref[2]
        ref[5,0,0]=(0.167335359474)*x_ref[0]**o+(-0.232843920341)*x_ref[0]+(-0.126773325792)*x_ref[1]**o+(-0.203898990499)*x_ref[1]+(0.839140245989)*x_ref[2]**o+(-0.728050229083)*x_ref[2]
        ref[5,0,1]=(-0.664841432025)*x_ref[0]**o+(-0.0489116955008)*x_ref[0]+(0.685379040894)*x_ref[1]**o+(0.370521174212)*x_ref[1]+(-0.742375864799)*x_ref[2]**o+(-0.880322708097)*x_ref[2]
        ref[5,1,0]=(0.653110528414)*x_ref[0]**o+(0.602025573916)*x_ref[0]+(0.912361317481)*x_ref[1]**o+(0.534616035231)*x_ref[1]+(-0.103102947904)*x_ref[2]**o+(0.531938071791)*x_ref[2]
        ref[5,1,1]=(-0.217856022054)*x_ref[0]**o+(0.11102291479)*x_ref[0]+(-0.166664940768)*x_ref[1]**o+(-0.812763150681)*x_ref[1]+(0.442963442721)*x_ref[2]**o+(0.706518891053)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Solution onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.209596991985)*x[0]**o+(0.997116487512)*x[0]+(0.0205082690151)*x[1]**o+(0.781596313955)*x[1]
        s[0,0,0,1]=(-0.475035630436)*x[0]**o+(0.476939038583)*x[0]+(-0.403224734103)*x[1]**o+(-0.726327526537)*x[1]
        s[0,0,1,0]=(0.165692569014)*x[0]**o+(-0.958953962148)*x[0]+(0.778520844819)*x[1]**o+(-0.371211501917)*x[1]
        s[0,0,1,1]=(-0.0863741752107)*x[0]**o+(0.64735697812)*x[0]+(-0.717371101374)*x[1]**o+(-0.517476214609)*x[1]
        s[0,0,2,0]=(0.414678358278)*x[0]**o+(0.766044973671)*x[0]+(0.473600649794)*x[1]**o+(-0.350863429165)*x[1]
        s[0,0,2,1]=(0.0391148157551)*x[0]**o+(0.550258520615)*x[0]+(-0.0952445590164)*x[1]**o+(0.799829721692)*x[1]
        s[0,1,0,0]=(-0.910760056281)*x[0]**o+(0.319346354593)*x[0]+(0.640854539851)*x[1]**o+(0.0622173716277)*x[1]
        s[0,1,0,1]=(0.761672466091)*x[0]**o+(0.266521259527)*x[0]+(-0.567155906289)*x[1]**o+(0.124135553341)*x[1]
        s[0,1,1,0]=(-0.681712184552)*x[0]**o+(0.388484728164)*x[0]+(-0.412243885661)*x[1]**o+(0.637954103055)*x[1]
        s[0,1,1,1]=(-0.485017958909)*x[0]**o+(0.736606905633)*x[0]+(-0.299204062504)*x[1]**o+(0.0840021848608)*x[1]
        s[0,1,2,0]=(0.955598154983)*x[0]**o+(-0.0204689463318)*x[0]+(-0.737123328202)*x[1]**o+(0.708418145874)*x[1]
        s[0,1,2,1]=(0.541565827885)*x[0]**o+(0.97493788494)*x[0]+(-0.450344230708)*x[1]**o+(-0.0231871887964)*x[1]
        s[0,2,0,0]=(-0.338442282639)*x[0]**o+(0.0746189409911)*x[0]+(0.76030098768)*x[1]**o+(0.629119828718)*x[1]
        s[0,2,0,1]=(-0.900573785768)*x[0]**o+(-0.971041944198)*x[0]+(0.637105978641)*x[1]**o+(0.387639692227)*x[1]
        s[0,2,1,0]=(0.0889093715568)*x[0]**o+(0.260479957021)*x[0]+(-0.908420084369)*x[1]**o+(0.6517961193)*x[1]
        s[0,2,1,1]=(0.683959794134)*x[0]**o+(-0.84665688176)*x[0]+(0.710853815216)*x[1]**o+(0.536738240986)*x[1]
        s[0,2,2,0]=(0.365754968462)*x[0]**o+(0.994576642301)*x[0]+(0.110702356612)*x[1]**o+(0.68992003076)*x[1]
        s[0,2,2,1]=(-0.238463519384)*x[0]**o+(-0.91372863405)*x[0]+(-0.334884010749)*x[1]**o+(0.481631381739)*x[1]
        s[0,3,0,0]=(-0.580344649945)*x[0]**o+(0.557953798016)*x[0]+(0.835117346932)*x[1]**o+(-0.196440751777)*x[1]
        s[0,3,0,1]=(0.339197786593)*x[0]**o+(0.231910549016)*x[0]+(0.515605654416)*x[1]**o+(-0.546885567308)*x[1]
        s[0,3,1,0]=(-0.562198188719)*x[0]**o+(0.410791865348)*x[0]+(-0.0823054523407)*x[1]**o+(0.789832279038)*x[1]
        s[0,3,1,1]=(-0.519245133296)*x[0]**o+(0.622234324355)*x[0]+(-0.89225914065)*x[1]**o+(-0.0376143576483)*x[1]
        s[0,3,2,0]=(-0.884895063274)*x[0]**o+(-0.569137436872)*x[0]+(0.107652574639)*x[1]**o+(0.936768318078)*x[1]
        s[0,3,2,1]=(-0.209493065726)*x[0]**o+(0.830161297986)*x[0]+(-0.811634331709)*x[1]**o+(0.0673874073431)*x[1]
        s[0,4,0,0]=(-0.518639168744)*x[0]**o+(-0.953627666411)*x[0]+(0.00153430886561)*x[1]**o+(0.118432410415)*x[1]
        s[0,4,0,1]=(0.901325505931)*x[0]**o+(0.276592227771)*x[0]+(0.906505574302)*x[1]**o+(-0.604032185175)*x[1]
        s[0,4,1,0]=(0.539336544226)*x[0]**o+(-0.518177781519)*x[0]+(-0.397453923803)*x[1]**o+(0.0661677991096)*x[1]
        s[0,4,1,1]=(0.31900579667)*x[0]**o+(-0.950735779053)*x[0]+(0.495414320287)*x[1]**o+(0.310834200079)*x[1]
        s[0,4,2,0]=(0.567277111907)*x[0]**o+(0.124915250004)*x[0]+(0.699531698929)*x[1]**o+(-0.434718433784)*x[1]
        s[0,4,2,1]=(0.607190906026)*x[0]**o+(0.411045834019)*x[0]+(-0.116848105246)*x[1]**o+(-0.343872842277)*x[1]
        s[1,0,0,0]=(0.263188443069)*x[0]**o+(0.782586354151)*x[0]+(-0.0593809262828)*x[1]**o+(-0.516894145417)*x[1]
        s[1,0,0,1]=(-0.155136250976)*x[0]**o+(0.155174971237)*x[0]+(0.452174963131)*x[1]**o+(-0.811875854704)*x[1]
        s[1,0,1,0]=(-0.161366848419)*x[0]**o+(-0.674425732875)*x[0]+(-0.99246147148)*x[1]**o+(0.618506213959)*x[1]
        s[1,0,1,1]=(0.64592094187)*x[0]**o+(0.444912658041)*x[0]+(0.963730302998)*x[1]**o+(-0.227756330634)*x[1]
        s[1,0,2,0]=(0.199217092911)*x[0]**o+(-0.118577255523)*x[0]+(-0.560600457675)*x[1]**o+(-0.233703068897)*x[1]
        s[1,0,2,1]=(0.200710250009)*x[0]**o+(0.366088123353)*x[0]+(0.419249984559)*x[1]**o+(-0.0188044094656)*x[1]
        s[1,1,0,0]=(-0.990545888549)*x[0]**o+(-0.100521244971)*x[0]+(-0.0134093118672)*x[1]**o+(0.582791526008)*x[1]
        s[1,1,0,1]=(-0.932085864166)*x[0]**o+(-0.402198623203)*x[0]+(-0.769996422181)*x[1]**o+(0.701839645425)*x[1]
        s[1,1,1,0]=(-0.0546050562235)*x[0]**o+(-0.821680588129)*x[0]+(0.254170421382)*x[1]**o+(0.0419062526906)*x[1]
        s[1,1,1,1]=(0.0361631230129)*x[0]**o+(0.155809438009)*x[0]+(-0.797631293341)*x[1]**o+(-0.0764161686239)*x[1]
        s[1,1,2,0]=(-0.679858025056)*x[0]**o+(-0.0411819071479)*x[0]+(0.033731324617)*x[1]**o+(0.32496792339)*x[1]
        s[1,1,2,1]=(0.424198647091)*x[0]**o+(-0.332229088063)*x[0]+(-0.225491160735)*x[1]**o+(-0.833706567263)*x[1]
        s[1,2,0,0]=(0.701510843723)*x[0]**o+(0.780699569051)*x[0]+(0.369319568113)*x[1]**o+(-0.865220461023)*x[1]
        s[1,2,0,1]=(-0.388700890661)*x[0]**o+(-0.550094979352)*x[0]+(0.585925695487)*x[1]**o+(0.812085187015)*x[1]
        s[1,2,1,0]=(-0.825827041568)*x[0]**o+(-0.281570175483)*x[0]+(-0.352446259467)*x[1]**o+(-0.162431177857)*x[1]
        s[1,2,1,1]=(-0.615917011781)*x[0]**o+(-0.218604155424)*x[0]+(-0.66718327064)*x[1]**o+(-0.723246262472)*x[1]
        s[1,2,2,0]=(-0.794217383815)*x[0]**o+(-0.939716907399)*x[0]+(-0.748413832758)*x[1]**o+(-0.196912090362)*x[1]
        s[1,2,2,1]=(0.447998156083)*x[0]**o+(-0.384539204664)*x[0]+(0.324989614124)*x[1]**o+(0.388489258939)*x[1]
        s[1,3,0,0]=(0.156094378734)*x[0]**o+(0.0533068325426)*x[0]+(0.268202796609)*x[1]**o+(-0.255422374899)*x[1]
        s[1,3,0,1]=(0.739277170457)*x[0]**o+(-0.290832216981)*x[0]+(0.350183930218)*x[1]**o+(-0.408368578636)*x[1]
        s[1,3,1,0]=(-0.0712530751134)*x[0]**o+(0.0658634843799)*x[0]+(0.099290893895)*x[1]**o+(0.647215326641)*x[1]
        s[1,3,1,1]=(0.0186407323716)*x[0]**o+(0.976985705627)*x[0]+(-0.871306465979)*x[1]**o+(-0.348450333839)*x[1]
        s[1,3,2,0]=(0.654402359016)*x[0]**o+(-0.633876744297)*x[0]+(0.160936266611)*x[1]**o+(0.476046463717)*x[1]
        s[1,3,2,1]=(-0.211306213041)*x[0]**o+(-0.296909064821)*x[0]+(-0.0612316728533)*x[1]**o+(0.901013542531)*x[1]
        s[1,4,0,0]=(0.974517622467)*x[0]**o+(0.998168509173)*x[0]+(-0.65931551157)*x[1]**o+(0.52094937925)*x[1]
        s[1,4,0,1]=(-0.145944864102)*x[0]**o+(-0.220443830385)*x[0]+(-0.380011620278)*x[1]**o+(0.315714506907)*x[1]
        s[1,4,1,0]=(0.00531622403943)*x[0]**o+(0.835826085399)*x[0]+(0.940648822215)*x[1]**o+(-0.505845376528)*x[1]
        s[1,4,1,1]=(0.794860008787)*x[0]**o+(-0.0885596053398)*x[0]+(-0.997579283268)*x[1]**o+(-0.339054217487)*x[1]
        s[1,4,2,0]=(-0.767918730098)*x[0]**o+(-0.023203645481)*x[0]+(-0.328338152731)*x[1]**o+(-0.232724548626)*x[1]
        s[1,4,2,1]=(0.211271068544)*x[0]**o+(-0.615899439167)*x[0]+(-0.0550973662218)*x[1]**o+(0.298493847127)*x[1]
        s[2,0,0,0]=(-0.911192683424)*x[0]**o+(0.00941390836017)*x[0]+(0.613172977466)*x[1]**o+(-0.788194982428)*x[1]
        s[2,0,0,1]=(-0.611555394987)*x[0]**o+(-0.577205119049)*x[0]+(0.319048321193)*x[1]**o+(-0.292972092026)*x[1]
        s[2,0,1,0]=(-0.669546200077)*x[0]**o+(-0.329254481941)*x[0]+(-0.841878266845)*x[1]**o+(-0.430302552502)*x[1]
        s[2,0,1,1]=(0.717338343232)*x[0]**o+(-0.184491535462)*x[0]+(-0.279163935394)*x[1]**o+(0.0902583733781)*x[1]
        s[2,0,2,0]=(0.74607828062)*x[0]**o+(0.220552976213)*x[0]+(-0.706261725925)*x[1]**o+(0.623659385726)*x[1]
        s[2,0,2,1]=(-0.673037605536)*x[0]**o+(-0.339261766556)*x[0]+(0.320833099845)*x[1]**o+(0.52076779438)*x[1]
        s[2,1,0,0]=(-0.765649699147)*x[0]**o+(0.916686538205)*x[0]+(0.97387546141)*x[1]**o+(0.553218952221)*x[1]
        s[2,1,0,1]=(0.973872792159)*x[0]**o+(0.517984046533)*x[0]+(0.51777252219)*x[1]**o+(-0.930026053905)*x[1]
        s[2,1,1,0]=(-0.948456218104)*x[0]**o+(-0.589599908446)*x[0]+(-0.467553533038)*x[1]**o+(0.274344478481)*x[1]
        s[2,1,1,1]=(0.881166959527)*x[0]**o+(0.49970488032)*x[0]+(-0.145215907327)*x[1]**o+(-0.781851254801)*x[1]
        s[2,1,2,0]=(0.249782338934)*x[0]**o+(-0.0658588005808)*x[0]+(0.812406146808)*x[1]**o+(0.692856516166)*x[1]
        s[2,1,2,1]=(-0.421909165386)*x[0]**o+(0.0386490156907)*x[0]+(-0.777622310099)*x[1]**o+(0.814181162845)*x[1]
        s[2,2,0,0]=(0.788918999327)*x[0]**o+(0.89626709297)*x[0]+(-0.0416920496206)*x[1]**o+(0.534755088018)*x[1]
        s[2,2,0,1]=(-0.723430509139)*x[0]**o+(0.937773302416)*x[0]+(0.120624514522)*x[1]**o+(0.712526712703)*x[1]
        s[2,2,1,0]=(0.755999533415)*x[0]**o+(-0.25818410984)*x[0]+(-0.833677276054)*x[1]**o+(0.389537968068)*x[1]
        s[2,2,1,1]=(-0.405091108345)*x[0]**o+(0.960266750195)*x[0]+(0.805114171635)*x[1]**o+(0.782424140168)*x[1]
        s[2,2,2,0]=(-0.00378071602137)*x[0]**o+(-0.774268061846)*x[0]+(0.317996842663)*x[1]**o+(-0.298998768873)*x[1]
        s[2,2,2,1]=(-0.640713160714)*x[0]**o+(-0.602638574906)*x[0]+(0.846395131454)*x[1]**o+(0.810407357088)*x[1]
        s[2,3,0,0]=(0.52309664987)*x[0]**o+(0.340015333205)*x[0]+(-0.463305517289)*x[1]**o+(0.377383683585)*x[1]
        s[2,3,0,1]=(-0.200960363239)*x[0]**o+(-0.343104026793)*x[0]+(0.72971674385)*x[1]**o+(-0.250972070079)*x[1]
        s[2,3,1,0]=(-0.198674852354)*x[0]**o+(0.376975952368)*x[0]+(-0.289846962669)*x[1]**o+(0.503874399797)*x[1]
        s[2,3,1,1]=(-0.500646179446)*x[0]**o+(0.999980656019)*x[0]+(-0.22225911248)*x[1]**o+(0.59532135129)*x[1]
        s[2,3,2,0]=(-0.0479308919198)*x[0]**o+(-0.286156974259)*x[0]+(-0.565672171896)*x[1]**o+(-0.496164919028)*x[1]
        s[2,3,2,1]=(0.283126731716)*x[0]**o+(-0.847220465645)*x[0]+(-0.202433223165)*x[1]**o+(0.733249392104)*x[1]
        s[2,4,0,0]=(0.379082415544)*x[0]**o+(0.619696544807)*x[0]+(0.30046651915)*x[1]**o+(-0.0956554688639)*x[1]
        s[2,4,0,1]=(0.645993528534)*x[0]**o+(0.636401374002)*x[0]+(-0.48531244649)*x[1]**o+(-0.566808631893)*x[1]
        s[2,4,1,0]=(-0.27936019151)*x[0]**o+(-0.0453561833702)*x[0]+(-0.38173600111)*x[1]**o+(0.0436591877645)*x[1]
        s[2,4,1,1]=(-0.305384262353)*x[0]**o+(-0.541085058351)*x[0]+(-0.055581601052)*x[1]**o+(0.128158662772)*x[1]
        s[2,4,2,0]=(-0.725495356548)*x[0]**o+(0.29235879065)*x[0]+(0.468015633053)*x[1]**o+(0.907396937488)*x[1]
        s[2,4,2,1]=(-0.191459527415)*x[0]**o+(0.337550470203)*x[0]+(0.846947249886)*x[1]**o+(0.115611176928)*x[1]
        s[3,0,0,0]=(-0.33176415824)*x[0]**o+(0.334074664888)*x[0]+(0.71328836071)*x[1]**o+(-0.108353365639)*x[1]
        s[3,0,0,1]=(0.189471905167)*x[0]**o+(-0.735388054498)*x[0]+(-0.145791816558)*x[1]**o+(0.1166770253)*x[1]
        s[3,0,1,0]=(0.130681026646)*x[0]**o+(-0.429395309447)*x[0]+(-0.960782097087)*x[1]**o+(-0.831182265535)*x[1]
        s[3,0,1,1]=(0.17875090433)*x[0]**o+(0.117342315307)*x[0]+(0.712145561102)*x[1]**o+(0.980052033498)*x[1]
        s[3,0,2,0]=(0.350160192028)*x[0]**o+(0.319028577789)*x[0]+(-0.390278261047)*x[1]**o+(-0.563618392837)*x[1]
        s[3,0,2,1]=(0.612683052829)*x[0]**o+(-0.808481977369)*x[0]+(-0.275769366104)*x[1]**o+(-0.16847980965)*x[1]
        s[3,1,0,0]=(0.441331101931)*x[0]**o+(0.631654604311)*x[0]+(-0.417381407441)*x[1]**o+(0.623209016434)*x[1]
        s[3,1,0,1]=(-0.316799359576)*x[0]**o+(0.0469238577708)*x[0]+(0.453846283394)*x[1]**o+(0.235093194654)*x[1]
        s[3,1,1,0]=(0.709151428964)*x[0]**o+(-0.66563806393)*x[0]+(0.957204730095)*x[1]**o+(0.757289506743)*x[1]
        s[3,1,1,1]=(0.651240972131)*x[0]**o+(0.125595241682)*x[0]+(0.673153945584)*x[1]**o+(-0.0741235551187)*x[1]
        s[3,1,2,0]=(0.109062793745)*x[0]**o+(0.603138432718)*x[0]+(0.626347779202)*x[1]**o+(0.747924284641)*x[1]
        s[3,1,2,1]=(-0.708014322028)*x[0]**o+(0.656641954147)*x[0]+(-0.5750166511)*x[1]**o+(0.971548956135)*x[1]
        s[3,2,0,0]=(0.308690818385)*x[0]**o+(-0.0743510684794)*x[0]+(0.503445484884)*x[1]**o+(0.122671270558)*x[1]
        s[3,2,0,1]=(-0.605770669307)*x[0]**o+(0.235294066964)*x[0]+(0.328631665933)*x[1]**o+(-0.713722501714)*x[1]
        s[3,2,1,0]=(0.772400565953)*x[0]**o+(0.889908318804)*x[0]+(0.166946692512)*x[1]**o+(0.207246011251)*x[1]
        s[3,2,1,1]=(-0.0256795661601)*x[0]**o+(-0.438929207275)*x[0]+(-0.586749375325)*x[1]**o+(-0.365377718456)*x[1]
        s[3,2,2,0]=(-0.395900535643)*x[0]**o+(0.814855175613)*x[0]+(0.139931347448)*x[1]**o+(-0.305268254265)*x[1]
        s[3,2,2,1]=(-0.613972524997)*x[0]**o+(0.151282941113)*x[0]+(0.137716512378)*x[1]**o+(0.987022640532)*x[1]
        s[3,3,0,0]=(0.054989906601)*x[0]**o+(0.629145288873)*x[0]+(-0.844944582016)*x[1]**o+(0.146581752248)*x[1]
        s[3,3,0,1]=(-0.14487953322)*x[0]**o+(0.618818776941)*x[0]+(0.924407652252)*x[1]**o+(0.563487350843)*x[1]
        s[3,3,1,0]=(0.58912173549)*x[0]**o+(-0.0901432408625)*x[0]+(0.873874316115)*x[1]**o+(0.881058120153)*x[1]
        s[3,3,1,1]=(0.576631116996)*x[0]**o+(-0.686918072799)*x[0]+(-0.50936823904)*x[1]**o+(0.585732475064)*x[1]
        s[3,3,2,0]=(0.306914981486)*x[0]**o+(-0.433296796977)*x[0]+(0.194615724385)*x[1]**o+(-0.98720631317)*x[1]
        s[3,3,2,1]=(0.48136043634)*x[0]**o+(0.38409585721)*x[0]+(0.809958268113)*x[1]**o+(-0.197158044454)*x[1]
        s[3,4,0,0]=(0.562436637234)*x[0]**o+(-0.729750092898)*x[0]+(-0.638146901507)*x[1]**o+(-0.629609335189)*x[1]
        s[3,4,0,1]=(0.610239954639)*x[0]**o+(-0.796185555272)*x[0]+(0.618408850895)*x[1]**o+(0.409684187852)*x[1]
        s[3,4,1,0]=(0.176527890224)*x[0]**o+(-0.680125370142)*x[0]+(0.297855195077)*x[1]**o+(0.416241653852)*x[1]
        s[3,4,1,1]=(-0.656877186833)*x[0]**o+(-0.622659703955)*x[0]+(0.92765398041)*x[1]**o+(-0.381997501461)*x[1]
        s[3,4,2,0]=(-0.206401064204)*x[0]**o+(0.496539998899)*x[0]+(0.303242800413)*x[1]**o+(-0.648983010434)*x[1]
        s[3,4,2,1]=(-0.850103219136)*x[0]**o+(-0.54057890339)*x[0]+(0.673138252663)*x[1]**o+(0.0181453330374)*x[1]
        ref[0,0,0,0]=(0.209596991985)*x_ref[0]**o+(0.997116487512)*x_ref[0]+(0.0205082690151)*x_ref[1]**o+(0.781596313955)*x_ref[1]
        ref[0,0,0,1]=(-0.475035630436)*x_ref[0]**o+(0.476939038583)*x_ref[0]+(-0.403224734103)*x_ref[1]**o+(-0.726327526537)*x_ref[1]
        ref[0,0,1,0]=(0.165692569014)*x_ref[0]**o+(-0.958953962148)*x_ref[0]+(0.778520844819)*x_ref[1]**o+(-0.371211501917)*x_ref[1]
        ref[0,0,1,1]=(-0.0863741752107)*x_ref[0]**o+(0.64735697812)*x_ref[0]+(-0.717371101374)*x_ref[1]**o+(-0.517476214609)*x_ref[1]
        ref[0,0,2,0]=(0.414678358278)*x_ref[0]**o+(0.766044973671)*x_ref[0]+(0.473600649794)*x_ref[1]**o+(-0.350863429165)*x_ref[1]
        ref[0,0,2,1]=(0.0391148157551)*x_ref[0]**o+(0.550258520615)*x_ref[0]+(-0.0952445590164)*x_ref[1]**o+(0.799829721692)*x_ref[1]
        ref[0,1,0,0]=(-0.910760056281)*x_ref[0]**o+(0.319346354593)*x_ref[0]+(0.640854539851)*x_ref[1]**o+(0.0622173716277)*x_ref[1]
        ref[0,1,0,1]=(0.761672466091)*x_ref[0]**o+(0.266521259527)*x_ref[0]+(-0.567155906289)*x_ref[1]**o+(0.124135553341)*x_ref[1]
        ref[0,1,1,0]=(-0.681712184552)*x_ref[0]**o+(0.388484728164)*x_ref[0]+(-0.412243885661)*x_ref[1]**o+(0.637954103055)*x_ref[1]
        ref[0,1,1,1]=(-0.485017958909)*x_ref[0]**o+(0.736606905633)*x_ref[0]+(-0.299204062504)*x_ref[1]**o+(0.0840021848608)*x_ref[1]
        ref[0,1,2,0]=(0.955598154983)*x_ref[0]**o+(-0.0204689463318)*x_ref[0]+(-0.737123328202)*x_ref[1]**o+(0.708418145874)*x_ref[1]
        ref[0,1,2,1]=(0.541565827885)*x_ref[0]**o+(0.97493788494)*x_ref[0]+(-0.450344230708)*x_ref[1]**o+(-0.0231871887964)*x_ref[1]
        ref[0,2,0,0]=(-0.338442282639)*x_ref[0]**o+(0.0746189409911)*x_ref[0]+(0.76030098768)*x_ref[1]**o+(0.629119828718)*x_ref[1]
        ref[0,2,0,1]=(-0.900573785768)*x_ref[0]**o+(-0.971041944198)*x_ref[0]+(0.637105978641)*x_ref[1]**o+(0.387639692227)*x_ref[1]
        ref[0,2,1,0]=(0.0889093715568)*x_ref[0]**o+(0.260479957021)*x_ref[0]+(-0.908420084369)*x_ref[1]**o+(0.6517961193)*x_ref[1]
        ref[0,2,1,1]=(0.683959794134)*x_ref[0]**o+(-0.84665688176)*x_ref[0]+(0.710853815216)*x_ref[1]**o+(0.536738240986)*x_ref[1]
        ref[0,2,2,0]=(0.365754968462)*x_ref[0]**o+(0.994576642301)*x_ref[0]+(0.110702356612)*x_ref[1]**o+(0.68992003076)*x_ref[1]
        ref[0,2,2,1]=(-0.238463519384)*x_ref[0]**o+(-0.91372863405)*x_ref[0]+(-0.334884010749)*x_ref[1]**o+(0.481631381739)*x_ref[1]
        ref[0,3,0,0]=(-0.580344649945)*x_ref[0]**o+(0.557953798016)*x_ref[0]+(0.835117346932)*x_ref[1]**o+(-0.196440751777)*x_ref[1]
        ref[0,3,0,1]=(0.339197786593)*x_ref[0]**o+(0.231910549016)*x_ref[0]+(0.515605654416)*x_ref[1]**o+(-0.546885567308)*x_ref[1]
        ref[0,3,1,0]=(-0.562198188719)*x_ref[0]**o+(0.410791865348)*x_ref[0]+(-0.0823054523407)*x_ref[1]**o+(0.789832279038)*x_ref[1]
        ref[0,3,1,1]=(-0.519245133296)*x_ref[0]**o+(0.622234324355)*x_ref[0]+(-0.89225914065)*x_ref[1]**o+(-0.0376143576483)*x_ref[1]
        ref[0,3,2,0]=(-0.884895063274)*x_ref[0]**o+(-0.569137436872)*x_ref[0]+(0.107652574639)*x_ref[1]**o+(0.936768318078)*x_ref[1]
        ref[0,3,2,1]=(-0.209493065726)*x_ref[0]**o+(0.830161297986)*x_ref[0]+(-0.811634331709)*x_ref[1]**o+(0.0673874073431)*x_ref[1]
        ref[0,4,0,0]=(-0.518639168744)*x_ref[0]**o+(-0.953627666411)*x_ref[0]+(0.00153430886561)*x_ref[1]**o+(0.118432410415)*x_ref[1]
        ref[0,4,0,1]=(0.901325505931)*x_ref[0]**o+(0.276592227771)*x_ref[0]+(0.906505574302)*x_ref[1]**o+(-0.604032185175)*x_ref[1]
        ref[0,4,1,0]=(0.539336544226)*x_ref[0]**o+(-0.518177781519)*x_ref[0]+(-0.397453923803)*x_ref[1]**o+(0.0661677991096)*x_ref[1]
        ref[0,4,1,1]=(0.31900579667)*x_ref[0]**o+(-0.950735779053)*x_ref[0]+(0.495414320287)*x_ref[1]**o+(0.310834200079)*x_ref[1]
        ref[0,4,2,0]=(0.567277111907)*x_ref[0]**o+(0.124915250004)*x_ref[0]+(0.699531698929)*x_ref[1]**o+(-0.434718433784)*x_ref[1]
        ref[0,4,2,1]=(0.607190906026)*x_ref[0]**o+(0.411045834019)*x_ref[0]+(-0.116848105246)*x_ref[1]**o+(-0.343872842277)*x_ref[1]
        ref[1,0,0,0]=(0.263188443069)*x_ref[0]**o+(0.782586354151)*x_ref[0]+(-0.0593809262828)*x_ref[1]**o+(-0.516894145417)*x_ref[1]
        ref[1,0,0,1]=(-0.155136250976)*x_ref[0]**o+(0.155174971237)*x_ref[0]+(0.452174963131)*x_ref[1]**o+(-0.811875854704)*x_ref[1]
        ref[1,0,1,0]=(-0.161366848419)*x_ref[0]**o+(-0.674425732875)*x_ref[0]+(-0.99246147148)*x_ref[1]**o+(0.618506213959)*x_ref[1]
        ref[1,0,1,1]=(0.64592094187)*x_ref[0]**o+(0.444912658041)*x_ref[0]+(0.963730302998)*x_ref[1]**o+(-0.227756330634)*x_ref[1]
        ref[1,0,2,0]=(0.199217092911)*x_ref[0]**o+(-0.118577255523)*x_ref[0]+(-0.560600457675)*x_ref[1]**o+(-0.233703068897)*x_ref[1]
        ref[1,0,2,1]=(0.200710250009)*x_ref[0]**o+(0.366088123353)*x_ref[0]+(0.419249984559)*x_ref[1]**o+(-0.0188044094656)*x_ref[1]
        ref[1,1,0,0]=(-0.990545888549)*x_ref[0]**o+(-0.100521244971)*x_ref[0]+(-0.0134093118672)*x_ref[1]**o+(0.582791526008)*x_ref[1]
        ref[1,1,0,1]=(-0.932085864166)*x_ref[0]**o+(-0.402198623203)*x_ref[0]+(-0.769996422181)*x_ref[1]**o+(0.701839645425)*x_ref[1]
        ref[1,1,1,0]=(-0.0546050562235)*x_ref[0]**o+(-0.821680588129)*x_ref[0]+(0.254170421382)*x_ref[1]**o+(0.0419062526906)*x_ref[1]
        ref[1,1,1,1]=(0.0361631230129)*x_ref[0]**o+(0.155809438009)*x_ref[0]+(-0.797631293341)*x_ref[1]**o+(-0.0764161686239)*x_ref[1]
        ref[1,1,2,0]=(-0.679858025056)*x_ref[0]**o+(-0.0411819071479)*x_ref[0]+(0.033731324617)*x_ref[1]**o+(0.32496792339)*x_ref[1]
        ref[1,1,2,1]=(0.424198647091)*x_ref[0]**o+(-0.332229088063)*x_ref[0]+(-0.225491160735)*x_ref[1]**o+(-0.833706567263)*x_ref[1]
        ref[1,2,0,0]=(0.701510843723)*x_ref[0]**o+(0.780699569051)*x_ref[0]+(0.369319568113)*x_ref[1]**o+(-0.865220461023)*x_ref[1]
        ref[1,2,0,1]=(-0.388700890661)*x_ref[0]**o+(-0.550094979352)*x_ref[0]+(0.585925695487)*x_ref[1]**o+(0.812085187015)*x_ref[1]
        ref[1,2,1,0]=(-0.825827041568)*x_ref[0]**o+(-0.281570175483)*x_ref[0]+(-0.352446259467)*x_ref[1]**o+(-0.162431177857)*x_ref[1]
        ref[1,2,1,1]=(-0.615917011781)*x_ref[0]**o+(-0.218604155424)*x_ref[0]+(-0.66718327064)*x_ref[1]**o+(-0.723246262472)*x_ref[1]
        ref[1,2,2,0]=(-0.794217383815)*x_ref[0]**o+(-0.939716907399)*x_ref[0]+(-0.748413832758)*x_ref[1]**o+(-0.196912090362)*x_ref[1]
        ref[1,2,2,1]=(0.447998156083)*x_ref[0]**o+(-0.384539204664)*x_ref[0]+(0.324989614124)*x_ref[1]**o+(0.388489258939)*x_ref[1]
        ref[1,3,0,0]=(0.156094378734)*x_ref[0]**o+(0.0533068325426)*x_ref[0]+(0.268202796609)*x_ref[1]**o+(-0.255422374899)*x_ref[1]
        ref[1,3,0,1]=(0.739277170457)*x_ref[0]**o+(-0.290832216981)*x_ref[0]+(0.350183930218)*x_ref[1]**o+(-0.408368578636)*x_ref[1]
        ref[1,3,1,0]=(-0.0712530751134)*x_ref[0]**o+(0.0658634843799)*x_ref[0]+(0.099290893895)*x_ref[1]**o+(0.647215326641)*x_ref[1]
        ref[1,3,1,1]=(0.0186407323716)*x_ref[0]**o+(0.976985705627)*x_ref[0]+(-0.871306465979)*x_ref[1]**o+(-0.348450333839)*x_ref[1]
        ref[1,3,2,0]=(0.654402359016)*x_ref[0]**o+(-0.633876744297)*x_ref[0]+(0.160936266611)*x_ref[1]**o+(0.476046463717)*x_ref[1]
        ref[1,3,2,1]=(-0.211306213041)*x_ref[0]**o+(-0.296909064821)*x_ref[0]+(-0.0612316728533)*x_ref[1]**o+(0.901013542531)*x_ref[1]
        ref[1,4,0,0]=(0.974517622467)*x_ref[0]**o+(0.998168509173)*x_ref[0]+(-0.65931551157)*x_ref[1]**o+(0.52094937925)*x_ref[1]
        ref[1,4,0,1]=(-0.145944864102)*x_ref[0]**o+(-0.220443830385)*x_ref[0]+(-0.380011620278)*x_ref[1]**o+(0.315714506907)*x_ref[1]
        ref[1,4,1,0]=(0.00531622403943)*x_ref[0]**o+(0.835826085399)*x_ref[0]+(0.940648822215)*x_ref[1]**o+(-0.505845376528)*x_ref[1]
        ref[1,4,1,1]=(0.794860008787)*x_ref[0]**o+(-0.0885596053398)*x_ref[0]+(-0.997579283268)*x_ref[1]**o+(-0.339054217487)*x_ref[1]
        ref[1,4,2,0]=(-0.767918730098)*x_ref[0]**o+(-0.023203645481)*x_ref[0]+(-0.328338152731)*x_ref[1]**o+(-0.232724548626)*x_ref[1]
        ref[1,4,2,1]=(0.211271068544)*x_ref[0]**o+(-0.615899439167)*x_ref[0]+(-0.0550973662218)*x_ref[1]**o+(0.298493847127)*x_ref[1]
        ref[2,0,0,0]=(-0.911192683424)*x_ref[0]**o+(0.00941390836017)*x_ref[0]+(0.613172977466)*x_ref[1]**o+(-0.788194982428)*x_ref[1]
        ref[2,0,0,1]=(-0.611555394987)*x_ref[0]**o+(-0.577205119049)*x_ref[0]+(0.319048321193)*x_ref[1]**o+(-0.292972092026)*x_ref[1]
        ref[2,0,1,0]=(-0.669546200077)*x_ref[0]**o+(-0.329254481941)*x_ref[0]+(-0.841878266845)*x_ref[1]**o+(-0.430302552502)*x_ref[1]
        ref[2,0,1,1]=(0.717338343232)*x_ref[0]**o+(-0.184491535462)*x_ref[0]+(-0.279163935394)*x_ref[1]**o+(0.0902583733781)*x_ref[1]
        ref[2,0,2,0]=(0.74607828062)*x_ref[0]**o+(0.220552976213)*x_ref[0]+(-0.706261725925)*x_ref[1]**o+(0.623659385726)*x_ref[1]
        ref[2,0,2,1]=(-0.673037605536)*x_ref[0]**o+(-0.339261766556)*x_ref[0]+(0.320833099845)*x_ref[1]**o+(0.52076779438)*x_ref[1]
        ref[2,1,0,0]=(-0.765649699147)*x_ref[0]**o+(0.916686538205)*x_ref[0]+(0.97387546141)*x_ref[1]**o+(0.553218952221)*x_ref[1]
        ref[2,1,0,1]=(0.973872792159)*x_ref[0]**o+(0.517984046533)*x_ref[0]+(0.51777252219)*x_ref[1]**o+(-0.930026053905)*x_ref[1]
        ref[2,1,1,0]=(-0.948456218104)*x_ref[0]**o+(-0.589599908446)*x_ref[0]+(-0.467553533038)*x_ref[1]**o+(0.274344478481)*x_ref[1]
        ref[2,1,1,1]=(0.881166959527)*x_ref[0]**o+(0.49970488032)*x_ref[0]+(-0.145215907327)*x_ref[1]**o+(-0.781851254801)*x_ref[1]
        ref[2,1,2,0]=(0.249782338934)*x_ref[0]**o+(-0.0658588005808)*x_ref[0]+(0.812406146808)*x_ref[1]**o+(0.692856516166)*x_ref[1]
        ref[2,1,2,1]=(-0.421909165386)*x_ref[0]**o+(0.0386490156907)*x_ref[0]+(-0.777622310099)*x_ref[1]**o+(0.814181162845)*x_ref[1]
        ref[2,2,0,0]=(0.788918999327)*x_ref[0]**o+(0.89626709297)*x_ref[0]+(-0.0416920496206)*x_ref[1]**o+(0.534755088018)*x_ref[1]
        ref[2,2,0,1]=(-0.723430509139)*x_ref[0]**o+(0.937773302416)*x_ref[0]+(0.120624514522)*x_ref[1]**o+(0.712526712703)*x_ref[1]
        ref[2,2,1,0]=(0.755999533415)*x_ref[0]**o+(-0.25818410984)*x_ref[0]+(-0.833677276054)*x_ref[1]**o+(0.389537968068)*x_ref[1]
        ref[2,2,1,1]=(-0.405091108345)*x_ref[0]**o+(0.960266750195)*x_ref[0]+(0.805114171635)*x_ref[1]**o+(0.782424140168)*x_ref[1]
        ref[2,2,2,0]=(-0.00378071602137)*x_ref[0]**o+(-0.774268061846)*x_ref[0]+(0.317996842663)*x_ref[1]**o+(-0.298998768873)*x_ref[1]
        ref[2,2,2,1]=(-0.640713160714)*x_ref[0]**o+(-0.602638574906)*x_ref[0]+(0.846395131454)*x_ref[1]**o+(0.810407357088)*x_ref[1]
        ref[2,3,0,0]=(0.52309664987)*x_ref[0]**o+(0.340015333205)*x_ref[0]+(-0.463305517289)*x_ref[1]**o+(0.377383683585)*x_ref[1]
        ref[2,3,0,1]=(-0.200960363239)*x_ref[0]**o+(-0.343104026793)*x_ref[0]+(0.72971674385)*x_ref[1]**o+(-0.250972070079)*x_ref[1]
        ref[2,3,1,0]=(-0.198674852354)*x_ref[0]**o+(0.376975952368)*x_ref[0]+(-0.289846962669)*x_ref[1]**o+(0.503874399797)*x_ref[1]
        ref[2,3,1,1]=(-0.500646179446)*x_ref[0]**o+(0.999980656019)*x_ref[0]+(-0.22225911248)*x_ref[1]**o+(0.59532135129)*x_ref[1]
        ref[2,3,2,0]=(-0.0479308919198)*x_ref[0]**o+(-0.286156974259)*x_ref[0]+(-0.565672171896)*x_ref[1]**o+(-0.496164919028)*x_ref[1]
        ref[2,3,2,1]=(0.283126731716)*x_ref[0]**o+(-0.847220465645)*x_ref[0]+(-0.202433223165)*x_ref[1]**o+(0.733249392104)*x_ref[1]
        ref[2,4,0,0]=(0.379082415544)*x_ref[0]**o+(0.619696544807)*x_ref[0]+(0.30046651915)*x_ref[1]**o+(-0.0956554688639)*x_ref[1]
        ref[2,4,0,1]=(0.645993528534)*x_ref[0]**o+(0.636401374002)*x_ref[0]+(-0.48531244649)*x_ref[1]**o+(-0.566808631893)*x_ref[1]
        ref[2,4,1,0]=(-0.27936019151)*x_ref[0]**o+(-0.0453561833702)*x_ref[0]+(-0.38173600111)*x_ref[1]**o+(0.0436591877645)*x_ref[1]
        ref[2,4,1,1]=(-0.305384262353)*x_ref[0]**o+(-0.541085058351)*x_ref[0]+(-0.055581601052)*x_ref[1]**o+(0.128158662772)*x_ref[1]
        ref[2,4,2,0]=(-0.725495356548)*x_ref[0]**o+(0.29235879065)*x_ref[0]+(0.468015633053)*x_ref[1]**o+(0.907396937488)*x_ref[1]
        ref[2,4,2,1]=(-0.191459527415)*x_ref[0]**o+(0.337550470203)*x_ref[0]+(0.846947249886)*x_ref[1]**o+(0.115611176928)*x_ref[1]
        ref[3,0,0,0]=(-0.33176415824)*x_ref[0]**o+(0.334074664888)*x_ref[0]+(0.71328836071)*x_ref[1]**o+(-0.108353365639)*x_ref[1]
        ref[3,0,0,1]=(0.189471905167)*x_ref[0]**o+(-0.735388054498)*x_ref[0]+(-0.145791816558)*x_ref[1]**o+(0.1166770253)*x_ref[1]
        ref[3,0,1,0]=(0.130681026646)*x_ref[0]**o+(-0.429395309447)*x_ref[0]+(-0.960782097087)*x_ref[1]**o+(-0.831182265535)*x_ref[1]
        ref[3,0,1,1]=(0.17875090433)*x_ref[0]**o+(0.117342315307)*x_ref[0]+(0.712145561102)*x_ref[1]**o+(0.980052033498)*x_ref[1]
        ref[3,0,2,0]=(0.350160192028)*x_ref[0]**o+(0.319028577789)*x_ref[0]+(-0.390278261047)*x_ref[1]**o+(-0.563618392837)*x_ref[1]
        ref[3,0,2,1]=(0.612683052829)*x_ref[0]**o+(-0.808481977369)*x_ref[0]+(-0.275769366104)*x_ref[1]**o+(-0.16847980965)*x_ref[1]
        ref[3,1,0,0]=(0.441331101931)*x_ref[0]**o+(0.631654604311)*x_ref[0]+(-0.417381407441)*x_ref[1]**o+(0.623209016434)*x_ref[1]
        ref[3,1,0,1]=(-0.316799359576)*x_ref[0]**o+(0.0469238577708)*x_ref[0]+(0.453846283394)*x_ref[1]**o+(0.235093194654)*x_ref[1]
        ref[3,1,1,0]=(0.709151428964)*x_ref[0]**o+(-0.66563806393)*x_ref[0]+(0.957204730095)*x_ref[1]**o+(0.757289506743)*x_ref[1]
        ref[3,1,1,1]=(0.651240972131)*x_ref[0]**o+(0.125595241682)*x_ref[0]+(0.673153945584)*x_ref[1]**o+(-0.0741235551187)*x_ref[1]
        ref[3,1,2,0]=(0.109062793745)*x_ref[0]**o+(0.603138432718)*x_ref[0]+(0.626347779202)*x_ref[1]**o+(0.747924284641)*x_ref[1]
        ref[3,1,2,1]=(-0.708014322028)*x_ref[0]**o+(0.656641954147)*x_ref[0]+(-0.5750166511)*x_ref[1]**o+(0.971548956135)*x_ref[1]
        ref[3,2,0,0]=(0.308690818385)*x_ref[0]**o+(-0.0743510684794)*x_ref[0]+(0.503445484884)*x_ref[1]**o+(0.122671270558)*x_ref[1]
        ref[3,2,0,1]=(-0.605770669307)*x_ref[0]**o+(0.235294066964)*x_ref[0]+(0.328631665933)*x_ref[1]**o+(-0.713722501714)*x_ref[1]
        ref[3,2,1,0]=(0.772400565953)*x_ref[0]**o+(0.889908318804)*x_ref[0]+(0.166946692512)*x_ref[1]**o+(0.207246011251)*x_ref[1]
        ref[3,2,1,1]=(-0.0256795661601)*x_ref[0]**o+(-0.438929207275)*x_ref[0]+(-0.586749375325)*x_ref[1]**o+(-0.365377718456)*x_ref[1]
        ref[3,2,2,0]=(-0.395900535643)*x_ref[0]**o+(0.814855175613)*x_ref[0]+(0.139931347448)*x_ref[1]**o+(-0.305268254265)*x_ref[1]
        ref[3,2,2,1]=(-0.613972524997)*x_ref[0]**o+(0.151282941113)*x_ref[0]+(0.137716512378)*x_ref[1]**o+(0.987022640532)*x_ref[1]
        ref[3,3,0,0]=(0.054989906601)*x_ref[0]**o+(0.629145288873)*x_ref[0]+(-0.844944582016)*x_ref[1]**o+(0.146581752248)*x_ref[1]
        ref[3,3,0,1]=(-0.14487953322)*x_ref[0]**o+(0.618818776941)*x_ref[0]+(0.924407652252)*x_ref[1]**o+(0.563487350843)*x_ref[1]
        ref[3,3,1,0]=(0.58912173549)*x_ref[0]**o+(-0.0901432408625)*x_ref[0]+(0.873874316115)*x_ref[1]**o+(0.881058120153)*x_ref[1]
        ref[3,3,1,1]=(0.576631116996)*x_ref[0]**o+(-0.686918072799)*x_ref[0]+(-0.50936823904)*x_ref[1]**o+(0.585732475064)*x_ref[1]
        ref[3,3,2,0]=(0.306914981486)*x_ref[0]**o+(-0.433296796977)*x_ref[0]+(0.194615724385)*x_ref[1]**o+(-0.98720631317)*x_ref[1]
        ref[3,3,2,1]=(0.48136043634)*x_ref[0]**o+(0.38409585721)*x_ref[0]+(0.809958268113)*x_ref[1]**o+(-0.197158044454)*x_ref[1]
        ref[3,4,0,0]=(0.562436637234)*x_ref[0]**o+(-0.729750092898)*x_ref[0]+(-0.638146901507)*x_ref[1]**o+(-0.629609335189)*x_ref[1]
        ref[3,4,0,1]=(0.610239954639)*x_ref[0]**o+(-0.796185555272)*x_ref[0]+(0.618408850895)*x_ref[1]**o+(0.409684187852)*x_ref[1]
        ref[3,4,1,0]=(0.176527890224)*x_ref[0]**o+(-0.680125370142)*x_ref[0]+(0.297855195077)*x_ref[1]**o+(0.416241653852)*x_ref[1]
        ref[3,4,1,1]=(-0.656877186833)*x_ref[0]**o+(-0.622659703955)*x_ref[0]+(0.92765398041)*x_ref[1]**o+(-0.381997501461)*x_ref[1]
        ref[3,4,2,0]=(-0.206401064204)*x_ref[0]**o+(0.496539998899)*x_ref[0]+(0.303242800413)*x_ref[1]**o+(-0.648983010434)*x_ref[1]
        ref[3,4,2,1]=(-0.850103219136)*x_ref[0]**o+(-0.54057890339)*x_ref[0]+(0.673138252663)*x_ref[1]**o+(0.0181453330374)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.183051403523)*x[0]**o+(0.675163647622)*x[0]+(0.383724078835)*x[1]**o+(-0.863174107932)*x[1]+(0.752053635179)*x[2]**o+(-0.442218538219)*x[2]
        s[0,0,0,1]=(0.145022844971)*x[0]**o+(-0.390513961466)*x[0]+(0.112480666084)*x[1]**o+(-0.484118377018)*x[1]+(0.804050685467)*x[2]**o+(-0.278608094628)*x[2]
        s[0,0,1,0]=(0.897135266304)*x[0]**o+(-0.761774008569)*x[0]+(-0.997851989059)*x[1]**o+(0.658068087366)*x[1]+(-0.80701588118)*x[2]**o+(0.596880821142)*x[2]
        s[0,0,1,1]=(-0.706353681044)*x[0]**o+(-0.19299984915)*x[0]+(0.0183270582241)*x[1]**o+(0.0848075866792)*x[1]+(-0.380936183304)*x[2]**o+(0.404471922242)*x[2]
        s[0,0,2,0]=(-0.382331572203)*x[0]**o+(-0.862239031672)*x[0]+(-0.664876503393)*x[1]**o+(0.721170780428)*x[1]+(0.936933909848)*x[2]**o+(0.553724430131)*x[2]
        s[0,0,2,1]=(0.41570549153)*x[0]**o+(-0.668260289502)*x[0]+(-0.118888455714)*x[1]**o+(0.454665663519)*x[1]+(-0.470911434781)*x[2]**o+(-0.677230009093)*x[2]
        s[0,1,0,0]=(-0.246347502133)*x[0]**o+(0.124690136148)*x[0]+(0.257467906886)*x[1]**o+(0.393404860933)*x[1]+(-0.216777928803)*x[2]**o+(-0.893375700242)*x[2]
        s[0,1,0,1]=(0.853598016536)*x[0]**o+(-0.614278282211)*x[0]+(0.958798548557)*x[1]**o+(0.539228268539)*x[1]+(-0.893749389287)*x[2]**o+(-0.421872187434)*x[2]
        s[0,1,1,0]=(-0.554158393064)*x[0]**o+(-0.0302602851801)*x[0]+(-0.627598510228)*x[1]**o+(-0.14419255857)*x[1]+(-0.319740995473)*x[2]**o+(-0.829801924596)*x[2]
        s[0,1,1,1]=(0.736629752286)*x[0]**o+(0.104716405421)*x[0]+(-0.340008364599)*x[1]**o+(-0.276509803031)*x[1]+(-0.65815905914)*x[2]**o+(-0.862207388004)*x[2]
        s[0,1,2,0]=(-0.963285485998)*x[0]**o+(-0.932861475962)*x[0]+(0.0893861922413)*x[1]**o+(-0.284311111286)*x[1]+(-0.549298067285)*x[2]**o+(0.990727028557)*x[2]
        s[0,1,2,1]=(0.354221245392)*x[0]**o+(-0.259651087945)*x[0]+(0.386589396916)*x[1]**o+(-0.315236532701)*x[1]+(0.0276667266652)*x[2]**o+(0.462795309548)*x[2]
        s[0,2,0,0]=(-0.488345436139)*x[0]**o+(0.0812145763632)*x[0]+(-0.766180568398)*x[1]**o+(0.52489429503)*x[1]+(-0.369663190972)*x[2]**o+(0.214012502771)*x[2]
        s[0,2,0,1]=(-0.447889959663)*x[0]**o+(0.815971127504)*x[0]+(-0.43620764595)*x[1]**o+(0.974646962464)*x[1]+(-0.0947298803439)*x[2]**o+(0.57765513293)*x[2]
        s[0,2,1,0]=(0.90082728964)*x[0]**o+(-0.471148200008)*x[0]+(0.130856999133)*x[1]**o+(0.94489744998)*x[1]+(0.97494516676)*x[2]**o+(0.379281707318)*x[2]
        s[0,2,1,1]=(-0.741008766911)*x[0]**o+(0.125378297796)*x[0]+(0.196554106007)*x[1]**o+(0.67968014258)*x[1]+(-0.22331716546)*x[2]**o+(0.69307086692)*x[2]
        s[0,2,2,0]=(0.731935335415)*x[0]**o+(0.203487322565)*x[0]+(-0.824149733689)*x[1]**o+(-0.765020086954)*x[1]+(0.652715725055)*x[2]**o+(-0.731883148625)*x[2]
        s[0,2,2,1]=(-0.845146208295)*x[0]**o+(-0.226473926648)*x[0]+(0.683088958387)*x[1]**o+(0.588656694973)*x[1]+(0.223845818469)*x[2]**o+(-0.311691745089)*x[2]
        s[0,3,0,0]=(-0.390979632965)*x[0]**o+(0.712430823379)*x[0]+(0.250097638139)*x[1]**o+(-0.0573974192609)*x[1]+(-0.104674518818)*x[2]**o+(-0.19187273988)*x[2]
        s[0,3,0,1]=(-0.422704759794)*x[0]**o+(-0.356438670304)*x[0]+(-0.99949139973)*x[1]**o+(-0.923501978496)*x[1]+(0.95094078991)*x[2]**o+(-0.726323072853)*x[2]
        s[0,3,1,0]=(-0.752686683053)*x[0]**o+(0.426969213595)*x[0]+(-0.884811980577)*x[1]**o+(-0.248298773095)*x[1]+(-0.248150259844)*x[2]**o+(0.307320547671)*x[2]
        s[0,3,1,1]=(0.444548575453)*x[0]**o+(-0.425971241634)*x[0]+(0.291461522783)*x[1]**o+(0.7897751016)*x[1]+(-0.7201321357)*x[2]**o+(0.335574613087)*x[2]
        s[0,3,2,0]=(0.320923109599)*x[0]**o+(0.768023804425)*x[0]+(0.786842381217)*x[1]**o+(0.508810585491)*x[1]+(0.565569549043)*x[2]**o+(-0.543149420951)*x[2]
        s[0,3,2,1]=(-0.434748955301)*x[0]**o+(0.753997363265)*x[0]+(-0.998212958853)*x[1]**o+(0.308236036581)*x[1]+(-0.604346477377)*x[2]**o+(0.0456731524173)*x[2]
        s[0,4,0,0]=(0.19259119938)*x[0]**o+(-0.676730834051)*x[0]+(0.782503706074)*x[1]**o+(-0.356602170048)*x[1]+(0.242151748383)*x[2]**o+(-0.555208960512)*x[2]
        s[0,4,0,1]=(-0.899694378162)*x[0]**o+(-0.998178345345)*x[0]+(-0.0299332693574)*x[1]**o+(-0.524636203176)*x[1]+(-0.090669349917)*x[2]**o+(-0.167591132189)*x[2]
        s[0,4,1,0]=(-0.68812275824)*x[0]**o+(0.973257165291)*x[0]+(0.714426301535)*x[1]**o+(0.913858401922)*x[1]+(0.59761682967)*x[2]**o+(-0.0363844534093)*x[2]
        s[0,4,1,1]=(0.479806493127)*x[0]**o+(-0.345277338487)*x[0]+(-0.0440371196931)*x[1]**o+(-0.393339132322)*x[1]+(0.0828262968561)*x[2]**o+(-0.492220459324)*x[2]
        s[0,4,2,0]=(-0.516900343789)*x[0]**o+(0.52589827237)*x[0]+(0.0897613840474)*x[1]**o+(0.0310617959058)*x[1]+(-0.0539611989165)*x[2]**o+(0.694823328714)*x[2]
        s[0,4,2,1]=(0.381866427224)*x[0]**o+(-0.136004696065)*x[0]+(0.127661553496)*x[1]**o+(-0.659183794977)*x[1]+(0.644887418078)*x[2]**o+(-0.378485836785)*x[2]
        s[1,0,0,0]=(0.533070302858)*x[0]**o+(0.0486822371539)*x[0]+(0.108020635139)*x[1]**o+(0.459684900644)*x[1]+(0.927900808007)*x[2]**o+(0.27439510523)*x[2]
        s[1,0,0,1]=(0.167370720476)*x[0]**o+(0.196355013651)*x[0]+(-0.727282731548)*x[1]**o+(0.911074385588)*x[1]+(0.0825461328514)*x[2]**o+(-0.352988731028)*x[2]
        s[1,0,1,0]=(-0.468001332309)*x[0]**o+(0.245999394647)*x[0]+(0.698629778456)*x[1]**o+(0.725916422941)*x[1]+(-0.486956268801)*x[2]**o+(-0.56567979109)*x[2]
        s[1,0,1,1]=(-0.738621292386)*x[0]**o+(0.943952843499)*x[0]+(-0.244367535526)*x[1]**o+(-0.984213076842)*x[1]+(0.429661592984)*x[2]**o+(-0.424380411926)*x[2]
        s[1,0,2,0]=(0.239251156871)*x[0]**o+(0.484710141522)*x[0]+(-0.748317757318)*x[1]**o+(0.448311080458)*x[1]+(-0.539095491323)*x[2]**o+(0.408909869581)*x[2]
        s[1,0,2,1]=(0.137854759003)*x[0]**o+(-0.683196454909)*x[0]+(-0.313962804942)*x[1]**o+(-0.72203426292)*x[1]+(-0.73815484699)*x[2]**o+(-0.363352540518)*x[2]
        s[1,1,0,0]=(-0.0285460639018)*x[0]**o+(-0.698099710448)*x[0]+(0.914032704615)*x[1]**o+(0.903165323006)*x[1]+(-0.943251579494)*x[2]**o+(0.342550252155)*x[2]
        s[1,1,0,1]=(0.467266211318)*x[0]**o+(0.274280471823)*x[0]+(-0.214518600035)*x[1]**o+(0.590589154176)*x[1]+(-0.641849036312)*x[2]**o+(-0.791498351136)*x[2]
        s[1,1,1,0]=(-0.871908333966)*x[0]**o+(0.0185255578465)*x[0]+(-0.872206753541)*x[1]**o+(-0.411775119904)*x[1]+(-0.806967989381)*x[2]**o+(0.193227663136)*x[2]
        s[1,1,1,1]=(0.952880647773)*x[0]**o+(0.611625705075)*x[0]+(0.14426391661)*x[1]**o+(0.62264475033)*x[1]+(0.0544261355307)*x[2]**o+(0.169080721619)*x[2]
        s[1,1,2,0]=(-0.734065499268)*x[0]**o+(-0.500982105171)*x[0]+(0.0974889823271)*x[1]**o+(0.920634492824)*x[1]+(0.146444953125)*x[2]**o+(0.968257204573)*x[2]
        s[1,1,2,1]=(0.210841124176)*x[0]**o+(-0.0353348787064)*x[0]+(0.704879863852)*x[1]**o+(-0.376074578373)*x[1]+(-0.562677666612)*x[2]**o+(-0.562469297446)*x[2]
        s[1,2,0,0]=(0.670790719895)*x[0]**o+(-0.568997121688)*x[0]+(0.662949640213)*x[1]**o+(0.770243690967)*x[1]+(-0.967323252654)*x[2]**o+(0.453018057272)*x[2]
        s[1,2,0,1]=(-0.724419676112)*x[0]**o+(0.437279546951)*x[0]+(0.52128041598)*x[1]**o+(-0.594484377874)*x[1]+(0.0245512079914)*x[2]**o+(-0.836825535903)*x[2]
        s[1,2,1,0]=(-0.126043799205)*x[0]**o+(0.410880843631)*x[0]+(0.509268071512)*x[1]**o+(-0.408676470752)*x[1]+(-0.241389348552)*x[2]**o+(0.21098949668)*x[2]
        s[1,2,1,1]=(0.0714381616781)*x[0]**o+(0.845308210908)*x[0]+(-0.156244949835)*x[1]**o+(-0.349790094709)*x[1]+(-0.373247740966)*x[2]**o+(0.844274382162)*x[2]
        s[1,2,2,0]=(0.392643676508)*x[0]**o+(0.379574595297)*x[0]+(-0.312714958636)*x[1]**o+(-0.697892817722)*x[1]+(0.319938083594)*x[2]**o+(0.848010553529)*x[2]
        s[1,2,2,1]=(0.0165719844498)*x[0]**o+(0.777792194538)*x[0]+(-0.581257805743)*x[1]**o+(0.576469709555)*x[1]+(0.881761997352)*x[2]**o+(-0.852836910827)*x[2]
        s[1,3,0,0]=(-0.368761641941)*x[0]**o+(-0.341981426952)*x[0]+(0.0850222818871)*x[1]**o+(-0.161499189059)*x[1]+(0.681488711691)*x[2]**o+(-0.220250578178)*x[2]
        s[1,3,0,1]=(-0.978416057456)*x[0]**o+(0.587189510525)*x[0]+(0.121078873003)*x[1]**o+(-0.499166770082)*x[1]+(-0.60948578652)*x[2]**o+(0.139721741069)*x[2]
        s[1,3,1,0]=(-0.248229483054)*x[0]**o+(0.396024076796)*x[0]+(0.955385247577)*x[1]**o+(0.0363257962548)*x[1]+(0.473586138101)*x[2]**o+(-0.978818683913)*x[2]
        s[1,3,1,1]=(-0.203828777074)*x[0]**o+(0.323222541043)*x[0]+(-0.0611217538705)*x[1]**o+(0.688466815312)*x[1]+(0.95854862113)*x[2]**o+(0.424013804278)*x[2]
        s[1,3,2,0]=(-0.734238871098)*x[0]**o+(0.987103771892)*x[0]+(-0.536241083275)*x[1]**o+(0.956192017815)*x[1]+(-0.458846574286)*x[2]**o+(0.0465550451449)*x[2]
        s[1,3,2,1]=(-0.6383215636)*x[0]**o+(-0.325193071614)*x[0]+(-0.91650917726)*x[1]**o+(0.693580745379)*x[1]+(-0.766712955542)*x[2]**o+(-0.929524610715)*x[2]
        s[1,4,0,0]=(-0.990767390655)*x[0]**o+(0.518778324025)*x[0]+(0.728020751614)*x[1]**o+(-0.755324391775)*x[1]+(0.186405771099)*x[2]**o+(0.632953433339)*x[2]
        s[1,4,0,1]=(0.592323495191)*x[0]**o+(0.409731821366)*x[0]+(0.964158523996)*x[1]**o+(0.424299771487)*x[1]+(0.660072793012)*x[2]**o+(-0.899939765137)*x[2]
        s[1,4,1,0]=(-0.303351442424)*x[0]**o+(-0.88332718624)*x[0]+(0.602724455773)*x[1]**o+(0.232452946908)*x[1]+(0.985569023425)*x[2]**o+(0.49272814355)*x[2]
        s[1,4,1,1]=(0.851768453481)*x[0]**o+(0.909492002062)*x[0]+(-0.810487634736)*x[1]**o+(0.919297425734)*x[1]+(0.909780009526)*x[2]**o+(0.369012187671)*x[2]
        s[1,4,2,0]=(-0.0531952997141)*x[0]**o+(-0.224430435352)*x[0]+(-0.578404055096)*x[1]**o+(0.33447800953)*x[1]+(-0.496336772814)*x[2]**o+(0.725685669073)*x[2]
        s[1,4,2,1]=(0.541476219067)*x[0]**o+(0.107132182409)*x[0]+(-0.335093661175)*x[1]**o+(0.227203070125)*x[1]+(0.216737562875)*x[2]**o+(-0.622943563245)*x[2]
        s[2,0,0,0]=(-0.756128419319)*x[0]**o+(0.547579725255)*x[0]+(0.687267799491)*x[1]**o+(0.819030721043)*x[1]+(-0.173876155736)*x[2]**o+(-0.0301920564367)*x[2]
        s[2,0,0,1]=(-0.0759002903983)*x[0]**o+(-0.438645875523)*x[0]+(-0.783783394278)*x[1]**o+(0.939615801765)*x[1]+(-0.0129095800279)*x[2]**o+(-0.723293627815)*x[2]
        s[2,0,1,0]=(-0.783224383717)*x[0]**o+(-0.137339879229)*x[0]+(-0.168000348627)*x[1]**o+(0.223734275088)*x[1]+(-0.582857962024)*x[2]**o+(-0.215263259255)*x[2]
        s[2,0,1,1]=(0.866554819231)*x[0]**o+(0.942792854767)*x[0]+(0.506186289419)*x[1]**o+(0.0775869106357)*x[1]+(0.877213607927)*x[2]**o+(-0.792692659916)*x[2]
        s[2,0,2,0]=(0.00917554286226)*x[0]**o+(-0.5183359551)*x[0]+(0.15698167339)*x[1]**o+(-0.489805488541)*x[1]+(-0.592166661492)*x[2]**o+(0.391650695037)*x[2]
        s[2,0,2,1]=(0.356986311456)*x[0]**o+(-0.14389994339)*x[0]+(0.838182818055)*x[1]**o+(-0.735112928546)*x[1]+(-0.0333560634355)*x[2]**o+(0.0855151256652)*x[2]
        s[2,1,0,0]=(0.357398992706)*x[0]**o+(-0.733457701676)*x[0]+(-0.142200448651)*x[1]**o+(0.844914653536)*x[1]+(-0.336954698769)*x[2]**o+(-0.296222395555)*x[2]
        s[2,1,0,1]=(-0.950414726707)*x[0]**o+(-0.134623616883)*x[0]+(-0.403232115759)*x[1]**o+(-0.539055139116)*x[1]+(-0.0298791725363)*x[2]**o+(-0.829664568373)*x[2]
        s[2,1,1,0]=(-0.656224623199)*x[0]**o+(0.0872237997616)*x[0]+(0.682387976548)*x[1]**o+(-0.225630652343)*x[1]+(-0.477261038385)*x[2]**o+(-0.57098588683)*x[2]
        s[2,1,1,1]=(-0.504013815131)*x[0]**o+(-0.0309238842436)*x[0]+(0.350436304599)*x[1]**o+(-0.891903027142)*x[1]+(-0.352433668387)*x[2]**o+(-0.603914711075)*x[2]
        s[2,1,2,0]=(0.505422678232)*x[0]**o+(-0.36241271654)*x[0]+(0.418645226594)*x[1]**o+(-0.392311515242)*x[1]+(-0.740372176332)*x[2]**o+(-0.861742253361)*x[2]
        s[2,1,2,1]=(-0.853192617592)*x[0]**o+(-0.269855659669)*x[0]+(0.439592720974)*x[1]**o+(0.62371411753)*x[1]+(0.77806877626)*x[2]**o+(0.923146838795)*x[2]
        s[2,2,0,0]=(0.228191521853)*x[0]**o+(0.286276293845)*x[0]+(0.205603167312)*x[1]**o+(-0.208236801784)*x[1]+(-0.787902086996)*x[2]**o+(0.676274514448)*x[2]
        s[2,2,0,1]=(-0.75952147535)*x[0]**o+(0.685910476453)*x[0]+(-0.945474186679)*x[1]**o+(0.257392146079)*x[1]+(-0.502240261104)*x[2]**o+(-0.446181698029)*x[2]
        s[2,2,1,0]=(-0.347627224549)*x[0]**o+(0.425329908635)*x[0]+(0.350298538346)*x[1]**o+(-0.484357417885)*x[1]+(0.538076599416)*x[2]**o+(-0.0530149980096)*x[2]
        s[2,2,1,1]=(0.0539971839295)*x[0]**o+(0.0721806780674)*x[0]+(-0.309030912724)*x[1]**o+(0.169874235759)*x[1]+(-0.433682959518)*x[2]**o+(0.813732426585)*x[2]
        s[2,2,2,0]=(0.336471441117)*x[0]**o+(-0.484255919842)*x[0]+(0.968308411862)*x[1]**o+(-0.521533322453)*x[1]+(0.633627764624)*x[2]**o+(0.13497783288)*x[2]
        s[2,2,2,1]=(-0.602946926541)*x[0]**o+(-0.67535176256)*x[0]+(-0.527632780342)*x[1]**o+(0.987852520439)*x[1]+(0.426772082556)*x[2]**o+(0.00805954755153)*x[2]
        s[2,3,0,0]=(0.834197069571)*x[0]**o+(-0.213325947671)*x[0]+(0.26098973561)*x[1]**o+(-0.892672004839)*x[1]+(0.690037689842)*x[2]**o+(-0.554911545988)*x[2]
        s[2,3,0,1]=(-0.917087293134)*x[0]**o+(-0.226247312973)*x[0]+(0.901194492616)*x[1]**o+(0.879779745124)*x[1]+(-0.501788697362)*x[2]**o+(0.62176481794)*x[2]
        s[2,3,1,0]=(-0.78132288334)*x[0]**o+(0.249383051506)*x[0]+(-0.276401042016)*x[1]**o+(0.0913832980328)*x[1]+(-0.199057369246)*x[2]**o+(0.229465279886)*x[2]
        s[2,3,1,1]=(0.33249458134)*x[0]**o+(-0.79179194291)*x[0]+(0.146436436451)*x[1]**o+(0.858879181025)*x[1]+(0.069509608477)*x[2]**o+(0.993263657)*x[2]
        s[2,3,2,0]=(0.103735002938)*x[0]**o+(-0.326548520992)*x[0]+(-0.826992179127)*x[1]**o+(-0.883356562305)*x[1]+(-0.340900003747)*x[2]**o+(-0.134490572338)*x[2]
        s[2,3,2,1]=(0.10363967847)*x[0]**o+(-0.171531371867)*x[0]+(-0.929315735292)*x[1]**o+(-0.719614719828)*x[1]+(-0.592883409628)*x[2]**o+(-0.641629610424)*x[2]
        s[2,4,0,0]=(-0.297999082535)*x[0]**o+(0.439209856881)*x[0]+(-0.751163554256)*x[1]**o+(-0.189087784579)*x[1]+(-0.795871127446)*x[2]**o+(0.130133923162)*x[2]
        s[2,4,0,1]=(0.721793867996)*x[0]**o+(-0.635599886382)*x[0]+(0.801889483297)*x[1]**o+(0.211160357461)*x[1]+(0.237091518418)*x[2]**o+(0.992344543271)*x[2]
        s[2,4,1,0]=(-0.775779912316)*x[0]**o+(-0.376074847935)*x[0]+(0.212663936842)*x[1]**o+(-0.549546184149)*x[1]+(-0.0854498425298)*x[2]**o+(0.542469273724)*x[2]
        s[2,4,1,1]=(0.363553642873)*x[0]**o+(-0.311431185265)*x[0]+(-0.375820595299)*x[1]**o+(-0.845438661463)*x[1]+(-0.478899940153)*x[2]**o+(0.0806897601294)*x[2]
        s[2,4,2,0]=(-0.489422873739)*x[0]**o+(-0.157008373165)*x[0]+(-0.267263773527)*x[1]**o+(-0.53340253927)*x[1]+(0.356782019115)*x[2]**o+(0.117093532841)*x[2]
        s[2,4,2,1]=(0.106435699048)*x[0]**o+(0.457452797294)*x[0]+(-0.640139127483)*x[1]**o+(-0.378451058366)*x[1]+(-0.232397757243)*x[2]**o+(-0.395640130256)*x[2]
        s[3,0,0,0]=(0.360807863609)*x[0]**o+(0.689413996889)*x[0]+(-0.267242935468)*x[1]**o+(0.341694875336)*x[1]+(0.955444260018)*x[2]**o+(0.940815872612)*x[2]
        s[3,0,0,1]=(0.0640653983699)*x[0]**o+(0.17954797305)*x[0]+(0.739660609779)*x[1]**o+(-0.174212756014)*x[1]+(0.190134414562)*x[2]**o+(-0.45741222058)*x[2]
        s[3,0,1,0]=(-0.642372239311)*x[0]**o+(-0.518079894531)*x[0]+(-0.843773918922)*x[1]**o+(0.407352407859)*x[1]+(0.0823899653044)*x[2]**o+(0.655867974962)*x[2]
        s[3,0,1,1]=(-0.545184505438)*x[0]**o+(-0.20749144757)*x[0]+(0.656821663662)*x[1]**o+(0.238823262425)*x[1]+(0.645114250743)*x[2]**o+(0.931193868381)*x[2]
        s[3,0,2,0]=(0.715326090131)*x[0]**o+(-0.623038006294)*x[0]+(-0.0272773549244)*x[1]**o+(0.0443002059335)*x[1]+(-0.639961942663)*x[2]**o+(-0.863662064393)*x[2]
        s[3,0,2,1]=(-0.150998418072)*x[0]**o+(-0.858320880863)*x[0]+(-0.711824804062)*x[1]**o+(-0.257945841504)*x[1]+(-0.070002743188)*x[2]**o+(-0.967136937623)*x[2]
        s[3,1,0,0]=(-0.5416631385)*x[0]**o+(-0.315276228939)*x[0]+(0.327960879182)*x[1]**o+(0.309383847119)*x[1]+(-0.531771040881)*x[2]**o+(-0.503014905853)*x[2]
        s[3,1,0,1]=(-0.93431005785)*x[0]**o+(-0.416716442301)*x[0]+(0.737336965624)*x[1]**o+(-0.676338424415)*x[1]+(-0.638729889321)*x[2]**o+(0.779664539115)*x[2]
        s[3,1,1,0]=(0.994892412301)*x[0]**o+(0.0257912160747)*x[0]+(-0.829652002123)*x[1]**o+(0.412033947586)*x[1]+(0.808588738484)*x[2]**o+(-0.537100420921)*x[2]
        s[3,1,1,1]=(-0.0459132262405)*x[0]**o+(-0.517692240206)*x[0]+(0.66461361799)*x[1]**o+(0.0202376762776)*x[1]+(0.561286210305)*x[2]**o+(-0.462378162744)*x[2]
        s[3,1,2,0]=(0.909401279733)*x[0]**o+(-0.092132877706)*x[0]+(-0.224813536034)*x[1]**o+(0.318594065478)*x[1]+(-0.959865958362)*x[2]**o+(0.458327172218)*x[2]
        s[3,1,2,1]=(0.174314370931)*x[0]**o+(-0.0149495584725)*x[0]+(0.719658326164)*x[1]**o+(0.416208781614)*x[1]+(0.922363382172)*x[2]**o+(-0.123220562614)*x[2]
        s[3,2,0,0]=(-0.362407421114)*x[0]**o+(-0.35287313685)*x[0]+(-0.712118894359)*x[1]**o+(-0.749389429473)*x[1]+(-0.520706676442)*x[2]**o+(0.15576625076)*x[2]
        s[3,2,0,1]=(-0.620594009087)*x[0]**o+(-0.866479846742)*x[0]+(-0.429532862455)*x[1]**o+(-0.736262913146)*x[1]+(0.14914428425)*x[2]**o+(0.00655098493102)*x[2]
        s[3,2,1,0]=(0.47291961358)*x[0]**o+(-0.223022599565)*x[0]+(-0.489191311572)*x[1]**o+(-0.616554022459)*x[1]+(0.525481658641)*x[2]**o+(0.999870533503)*x[2]
        s[3,2,1,1]=(0.315715505936)*x[0]**o+(-0.640835348385)*x[0]+(0.768958709128)*x[1]**o+(0.587161832357)*x[1]+(0.454427338357)*x[2]**o+(-0.133391799438)*x[2]
        s[3,2,2,0]=(-0.273381601564)*x[0]**o+(-0.527114000062)*x[0]+(-0.241611283455)*x[1]**o+(0.511784453537)*x[1]+(-0.558176752993)*x[2]**o+(0.563602029554)*x[2]
        s[3,2,2,1]=(-0.554552149574)*x[0]**o+(0.277189312969)*x[0]+(0.528236290321)*x[1]**o+(-0.56161965546)*x[1]+(0.186320859613)*x[2]**o+(-0.462794918456)*x[2]
        s[3,3,0,0]=(-0.379669006639)*x[0]**o+(-0.290542994989)*x[0]+(0.950886714589)*x[1]**o+(0.113833056597)*x[1]+(-0.395921494129)*x[2]**o+(-0.530197898233)*x[2]
        s[3,3,0,1]=(-0.37250317146)*x[0]**o+(0.266496776367)*x[0]+(-0.992301806744)*x[1]**o+(-0.215688062569)*x[1]+(0.86298623844)*x[2]**o+(0.932822343278)*x[2]
        s[3,3,1,0]=(-0.354585401042)*x[0]**o+(0.472315228214)*x[0]+(-0.604904767558)*x[1]**o+(-0.868411440101)*x[1]+(-0.434629251045)*x[2]**o+(-0.4582356693)*x[2]
        s[3,3,1,1]=(-0.931569757365)*x[0]**o+(0.062290982568)*x[0]+(-0.402530852638)*x[1]**o+(0.13606876168)*x[1]+(-0.421824162622)*x[2]**o+(0.754078648605)*x[2]
        s[3,3,2,0]=(0.370446872991)*x[0]**o+(-0.678219724124)*x[0]+(0.208629662303)*x[1]**o+(0.742029980811)*x[1]+(0.311836736294)*x[2]**o+(-0.102558272951)*x[2]
        s[3,3,2,1]=(-0.15347338298)*x[0]**o+(-0.798719756084)*x[0]+(0.884663004721)*x[1]**o+(-0.658935444548)*x[1]+(-0.721604057065)*x[2]**o+(0.299122598053)*x[2]
        s[3,4,0,0]=(0.544627747607)*x[0]**o+(0.513489478068)*x[0]+(-0.410347246011)*x[1]**o+(0.610018445392)*x[1]+(0.87473885042)*x[2]**o+(0.914293254335)*x[2]
        s[3,4,0,1]=(-0.236207879972)*x[0]**o+(0.46233924924)*x[0]+(-0.472291476346)*x[1]**o+(0.480397963815)*x[1]+(-0.976610220911)*x[2]**o+(-0.442554954343)*x[2]
        s[3,4,1,0]=(0.332927708298)*x[0]**o+(0.737580232392)*x[0]+(-0.56615273435)*x[1]**o+(0.487228009962)*x[1]+(-0.620250989525)*x[2]**o+(0.699377689126)*x[2]
        s[3,4,1,1]=(-0.574680118107)*x[0]**o+(-0.645001786533)*x[0]+(-0.0623752350971)*x[1]**o+(-0.943489194832)*x[1]+(0.690135931129)*x[2]**o+(0.047173398961)*x[2]
        s[3,4,2,0]=(0.218423477936)*x[0]**o+(0.892507556482)*x[0]+(-0.621739148946)*x[1]**o+(-0.253753368776)*x[1]+(-0.88306720835)*x[2]**o+(0.0426804971625)*x[2]
        s[3,4,2,1]=(-0.901566218988)*x[0]**o+(-0.0843229102423)*x[0]+(-0.206812141605)*x[1]**o+(-0.321102598716)*x[1]+(-0.873774065466)*x[2]**o+(-0.0507766082592)*x[2]
        ref[0,0,0,0]=(-0.183051403523)*x_ref[0]**o+(0.675163647622)*x_ref[0]+(0.383724078835)*x_ref[1]**o+(-0.863174107932)*x_ref[1]+(0.752053635179)*x_ref[2]**o+(-0.442218538219)*x_ref[2]
        ref[0,0,0,1]=(0.145022844971)*x_ref[0]**o+(-0.390513961466)*x_ref[0]+(0.112480666084)*x_ref[1]**o+(-0.484118377018)*x_ref[1]+(0.804050685467)*x_ref[2]**o+(-0.278608094628)*x_ref[2]
        ref[0,0,1,0]=(0.897135266304)*x_ref[0]**o+(-0.761774008569)*x_ref[0]+(-0.997851989059)*x_ref[1]**o+(0.658068087366)*x_ref[1]+(-0.80701588118)*x_ref[2]**o+(0.596880821142)*x_ref[2]
        ref[0,0,1,1]=(-0.706353681044)*x_ref[0]**o+(-0.19299984915)*x_ref[0]+(0.0183270582241)*x_ref[1]**o+(0.0848075866792)*x_ref[1]+(-0.380936183304)*x_ref[2]**o+(0.404471922242)*x_ref[2]
        ref[0,0,2,0]=(-0.382331572203)*x_ref[0]**o+(-0.862239031672)*x_ref[0]+(-0.664876503393)*x_ref[1]**o+(0.721170780428)*x_ref[1]+(0.936933909848)*x_ref[2]**o+(0.553724430131)*x_ref[2]
        ref[0,0,2,1]=(0.41570549153)*x_ref[0]**o+(-0.668260289502)*x_ref[0]+(-0.118888455714)*x_ref[1]**o+(0.454665663519)*x_ref[1]+(-0.470911434781)*x_ref[2]**o+(-0.677230009093)*x_ref[2]
        ref[0,1,0,0]=(-0.246347502133)*x_ref[0]**o+(0.124690136148)*x_ref[0]+(0.257467906886)*x_ref[1]**o+(0.393404860933)*x_ref[1]+(-0.216777928803)*x_ref[2]**o+(-0.893375700242)*x_ref[2]
        ref[0,1,0,1]=(0.853598016536)*x_ref[0]**o+(-0.614278282211)*x_ref[0]+(0.958798548557)*x_ref[1]**o+(0.539228268539)*x_ref[1]+(-0.893749389287)*x_ref[2]**o+(-0.421872187434)*x_ref[2]
        ref[0,1,1,0]=(-0.554158393064)*x_ref[0]**o+(-0.0302602851801)*x_ref[0]+(-0.627598510228)*x_ref[1]**o+(-0.14419255857)*x_ref[1]+(-0.319740995473)*x_ref[2]**o+(-0.829801924596)*x_ref[2]
        ref[0,1,1,1]=(0.736629752286)*x_ref[0]**o+(0.104716405421)*x_ref[0]+(-0.340008364599)*x_ref[1]**o+(-0.276509803031)*x_ref[1]+(-0.65815905914)*x_ref[2]**o+(-0.862207388004)*x_ref[2]
        ref[0,1,2,0]=(-0.963285485998)*x_ref[0]**o+(-0.932861475962)*x_ref[0]+(0.0893861922413)*x_ref[1]**o+(-0.284311111286)*x_ref[1]+(-0.549298067285)*x_ref[2]**o+(0.990727028557)*x_ref[2]
        ref[0,1,2,1]=(0.354221245392)*x_ref[0]**o+(-0.259651087945)*x_ref[0]+(0.386589396916)*x_ref[1]**o+(-0.315236532701)*x_ref[1]+(0.0276667266652)*x_ref[2]**o+(0.462795309548)*x_ref[2]
        ref[0,2,0,0]=(-0.488345436139)*x_ref[0]**o+(0.0812145763632)*x_ref[0]+(-0.766180568398)*x_ref[1]**o+(0.52489429503)*x_ref[1]+(-0.369663190972)*x_ref[2]**o+(0.214012502771)*x_ref[2]
        ref[0,2,0,1]=(-0.447889959663)*x_ref[0]**o+(0.815971127504)*x_ref[0]+(-0.43620764595)*x_ref[1]**o+(0.974646962464)*x_ref[1]+(-0.0947298803439)*x_ref[2]**o+(0.57765513293)*x_ref[2]
        ref[0,2,1,0]=(0.90082728964)*x_ref[0]**o+(-0.471148200008)*x_ref[0]+(0.130856999133)*x_ref[1]**o+(0.94489744998)*x_ref[1]+(0.97494516676)*x_ref[2]**o+(0.379281707318)*x_ref[2]
        ref[0,2,1,1]=(-0.741008766911)*x_ref[0]**o+(0.125378297796)*x_ref[0]+(0.196554106007)*x_ref[1]**o+(0.67968014258)*x_ref[1]+(-0.22331716546)*x_ref[2]**o+(0.69307086692)*x_ref[2]
        ref[0,2,2,0]=(0.731935335415)*x_ref[0]**o+(0.203487322565)*x_ref[0]+(-0.824149733689)*x_ref[1]**o+(-0.765020086954)*x_ref[1]+(0.652715725055)*x_ref[2]**o+(-0.731883148625)*x_ref[2]
        ref[0,2,2,1]=(-0.845146208295)*x_ref[0]**o+(-0.226473926648)*x_ref[0]+(0.683088958387)*x_ref[1]**o+(0.588656694973)*x_ref[1]+(0.223845818469)*x_ref[2]**o+(-0.311691745089)*x_ref[2]
        ref[0,3,0,0]=(-0.390979632965)*x_ref[0]**o+(0.712430823379)*x_ref[0]+(0.250097638139)*x_ref[1]**o+(-0.0573974192609)*x_ref[1]+(-0.104674518818)*x_ref[2]**o+(-0.19187273988)*x_ref[2]
        ref[0,3,0,1]=(-0.422704759794)*x_ref[0]**o+(-0.356438670304)*x_ref[0]+(-0.99949139973)*x_ref[1]**o+(-0.923501978496)*x_ref[1]+(0.95094078991)*x_ref[2]**o+(-0.726323072853)*x_ref[2]
        ref[0,3,1,0]=(-0.752686683053)*x_ref[0]**o+(0.426969213595)*x_ref[0]+(-0.884811980577)*x_ref[1]**o+(-0.248298773095)*x_ref[1]+(-0.248150259844)*x_ref[2]**o+(0.307320547671)*x_ref[2]
        ref[0,3,1,1]=(0.444548575453)*x_ref[0]**o+(-0.425971241634)*x_ref[0]+(0.291461522783)*x_ref[1]**o+(0.7897751016)*x_ref[1]+(-0.7201321357)*x_ref[2]**o+(0.335574613087)*x_ref[2]
        ref[0,3,2,0]=(0.320923109599)*x_ref[0]**o+(0.768023804425)*x_ref[0]+(0.786842381217)*x_ref[1]**o+(0.508810585491)*x_ref[1]+(0.565569549043)*x_ref[2]**o+(-0.543149420951)*x_ref[2]
        ref[0,3,2,1]=(-0.434748955301)*x_ref[0]**o+(0.753997363265)*x_ref[0]+(-0.998212958853)*x_ref[1]**o+(0.308236036581)*x_ref[1]+(-0.604346477377)*x_ref[2]**o+(0.0456731524173)*x_ref[2]
        ref[0,4,0,0]=(0.19259119938)*x_ref[0]**o+(-0.676730834051)*x_ref[0]+(0.782503706074)*x_ref[1]**o+(-0.356602170048)*x_ref[1]+(0.242151748383)*x_ref[2]**o+(-0.555208960512)*x_ref[2]
        ref[0,4,0,1]=(-0.899694378162)*x_ref[0]**o+(-0.998178345345)*x_ref[0]+(-0.0299332693574)*x_ref[1]**o+(-0.524636203176)*x_ref[1]+(-0.090669349917)*x_ref[2]**o+(-0.167591132189)*x_ref[2]
        ref[0,4,1,0]=(-0.68812275824)*x_ref[0]**o+(0.973257165291)*x_ref[0]+(0.714426301535)*x_ref[1]**o+(0.913858401922)*x_ref[1]+(0.59761682967)*x_ref[2]**o+(-0.0363844534093)*x_ref[2]
        ref[0,4,1,1]=(0.479806493127)*x_ref[0]**o+(-0.345277338487)*x_ref[0]+(-0.0440371196931)*x_ref[1]**o+(-0.393339132322)*x_ref[1]+(0.0828262968561)*x_ref[2]**o+(-0.492220459324)*x_ref[2]
        ref[0,4,2,0]=(-0.516900343789)*x_ref[0]**o+(0.52589827237)*x_ref[0]+(0.0897613840474)*x_ref[1]**o+(0.0310617959058)*x_ref[1]+(-0.0539611989165)*x_ref[2]**o+(0.694823328714)*x_ref[2]
        ref[0,4,2,1]=(0.381866427224)*x_ref[0]**o+(-0.136004696065)*x_ref[0]+(0.127661553496)*x_ref[1]**o+(-0.659183794977)*x_ref[1]+(0.644887418078)*x_ref[2]**o+(-0.378485836785)*x_ref[2]
        ref[1,0,0,0]=(0.533070302858)*x_ref[0]**o+(0.0486822371539)*x_ref[0]+(0.108020635139)*x_ref[1]**o+(0.459684900644)*x_ref[1]+(0.927900808007)*x_ref[2]**o+(0.27439510523)*x_ref[2]
        ref[1,0,0,1]=(0.167370720476)*x_ref[0]**o+(0.196355013651)*x_ref[0]+(-0.727282731548)*x_ref[1]**o+(0.911074385588)*x_ref[1]+(0.0825461328514)*x_ref[2]**o+(-0.352988731028)*x_ref[2]
        ref[1,0,1,0]=(-0.468001332309)*x_ref[0]**o+(0.245999394647)*x_ref[0]+(0.698629778456)*x_ref[1]**o+(0.725916422941)*x_ref[1]+(-0.486956268801)*x_ref[2]**o+(-0.56567979109)*x_ref[2]
        ref[1,0,1,1]=(-0.738621292386)*x_ref[0]**o+(0.943952843499)*x_ref[0]+(-0.244367535526)*x_ref[1]**o+(-0.984213076842)*x_ref[1]+(0.429661592984)*x_ref[2]**o+(-0.424380411926)*x_ref[2]
        ref[1,0,2,0]=(0.239251156871)*x_ref[0]**o+(0.484710141522)*x_ref[0]+(-0.748317757318)*x_ref[1]**o+(0.448311080458)*x_ref[1]+(-0.539095491323)*x_ref[2]**o+(0.408909869581)*x_ref[2]
        ref[1,0,2,1]=(0.137854759003)*x_ref[0]**o+(-0.683196454909)*x_ref[0]+(-0.313962804942)*x_ref[1]**o+(-0.72203426292)*x_ref[1]+(-0.73815484699)*x_ref[2]**o+(-0.363352540518)*x_ref[2]
        ref[1,1,0,0]=(-0.0285460639018)*x_ref[0]**o+(-0.698099710448)*x_ref[0]+(0.914032704615)*x_ref[1]**o+(0.903165323006)*x_ref[1]+(-0.943251579494)*x_ref[2]**o+(0.342550252155)*x_ref[2]
        ref[1,1,0,1]=(0.467266211318)*x_ref[0]**o+(0.274280471823)*x_ref[0]+(-0.214518600035)*x_ref[1]**o+(0.590589154176)*x_ref[1]+(-0.641849036312)*x_ref[2]**o+(-0.791498351136)*x_ref[2]
        ref[1,1,1,0]=(-0.871908333966)*x_ref[0]**o+(0.0185255578465)*x_ref[0]+(-0.872206753541)*x_ref[1]**o+(-0.411775119904)*x_ref[1]+(-0.806967989381)*x_ref[2]**o+(0.193227663136)*x_ref[2]
        ref[1,1,1,1]=(0.952880647773)*x_ref[0]**o+(0.611625705075)*x_ref[0]+(0.14426391661)*x_ref[1]**o+(0.62264475033)*x_ref[1]+(0.0544261355307)*x_ref[2]**o+(0.169080721619)*x_ref[2]
        ref[1,1,2,0]=(-0.734065499268)*x_ref[0]**o+(-0.500982105171)*x_ref[0]+(0.0974889823271)*x_ref[1]**o+(0.920634492824)*x_ref[1]+(0.146444953125)*x_ref[2]**o+(0.968257204573)*x_ref[2]
        ref[1,1,2,1]=(0.210841124176)*x_ref[0]**o+(-0.0353348787064)*x_ref[0]+(0.704879863852)*x_ref[1]**o+(-0.376074578373)*x_ref[1]+(-0.562677666612)*x_ref[2]**o+(-0.562469297446)*x_ref[2]
        ref[1,2,0,0]=(0.670790719895)*x_ref[0]**o+(-0.568997121688)*x_ref[0]+(0.662949640213)*x_ref[1]**o+(0.770243690967)*x_ref[1]+(-0.967323252654)*x_ref[2]**o+(0.453018057272)*x_ref[2]
        ref[1,2,0,1]=(-0.724419676112)*x_ref[0]**o+(0.437279546951)*x_ref[0]+(0.52128041598)*x_ref[1]**o+(-0.594484377874)*x_ref[1]+(0.0245512079914)*x_ref[2]**o+(-0.836825535903)*x_ref[2]
        ref[1,2,1,0]=(-0.126043799205)*x_ref[0]**o+(0.410880843631)*x_ref[0]+(0.509268071512)*x_ref[1]**o+(-0.408676470752)*x_ref[1]+(-0.241389348552)*x_ref[2]**o+(0.21098949668)*x_ref[2]
        ref[1,2,1,1]=(0.0714381616781)*x_ref[0]**o+(0.845308210908)*x_ref[0]+(-0.156244949835)*x_ref[1]**o+(-0.349790094709)*x_ref[1]+(-0.373247740966)*x_ref[2]**o+(0.844274382162)*x_ref[2]
        ref[1,2,2,0]=(0.392643676508)*x_ref[0]**o+(0.379574595297)*x_ref[0]+(-0.312714958636)*x_ref[1]**o+(-0.697892817722)*x_ref[1]+(0.319938083594)*x_ref[2]**o+(0.848010553529)*x_ref[2]
        ref[1,2,2,1]=(0.0165719844498)*x_ref[0]**o+(0.777792194538)*x_ref[0]+(-0.581257805743)*x_ref[1]**o+(0.576469709555)*x_ref[1]+(0.881761997352)*x_ref[2]**o+(-0.852836910827)*x_ref[2]
        ref[1,3,0,0]=(-0.368761641941)*x_ref[0]**o+(-0.341981426952)*x_ref[0]+(0.0850222818871)*x_ref[1]**o+(-0.161499189059)*x_ref[1]+(0.681488711691)*x_ref[2]**o+(-0.220250578178)*x_ref[2]
        ref[1,3,0,1]=(-0.978416057456)*x_ref[0]**o+(0.587189510525)*x_ref[0]+(0.121078873003)*x_ref[1]**o+(-0.499166770082)*x_ref[1]+(-0.60948578652)*x_ref[2]**o+(0.139721741069)*x_ref[2]
        ref[1,3,1,0]=(-0.248229483054)*x_ref[0]**o+(0.396024076796)*x_ref[0]+(0.955385247577)*x_ref[1]**o+(0.0363257962548)*x_ref[1]+(0.473586138101)*x_ref[2]**o+(-0.978818683913)*x_ref[2]
        ref[1,3,1,1]=(-0.203828777074)*x_ref[0]**o+(0.323222541043)*x_ref[0]+(-0.0611217538705)*x_ref[1]**o+(0.688466815312)*x_ref[1]+(0.95854862113)*x_ref[2]**o+(0.424013804278)*x_ref[2]
        ref[1,3,2,0]=(-0.734238871098)*x_ref[0]**o+(0.987103771892)*x_ref[0]+(-0.536241083275)*x_ref[1]**o+(0.956192017815)*x_ref[1]+(-0.458846574286)*x_ref[2]**o+(0.0465550451449)*x_ref[2]
        ref[1,3,2,1]=(-0.6383215636)*x_ref[0]**o+(-0.325193071614)*x_ref[0]+(-0.91650917726)*x_ref[1]**o+(0.693580745379)*x_ref[1]+(-0.766712955542)*x_ref[2]**o+(-0.929524610715)*x_ref[2]
        ref[1,4,0,0]=(-0.990767390655)*x_ref[0]**o+(0.518778324025)*x_ref[0]+(0.728020751614)*x_ref[1]**o+(-0.755324391775)*x_ref[1]+(0.186405771099)*x_ref[2]**o+(0.632953433339)*x_ref[2]
        ref[1,4,0,1]=(0.592323495191)*x_ref[0]**o+(0.409731821366)*x_ref[0]+(0.964158523996)*x_ref[1]**o+(0.424299771487)*x_ref[1]+(0.660072793012)*x_ref[2]**o+(-0.899939765137)*x_ref[2]
        ref[1,4,1,0]=(-0.303351442424)*x_ref[0]**o+(-0.88332718624)*x_ref[0]+(0.602724455773)*x_ref[1]**o+(0.232452946908)*x_ref[1]+(0.985569023425)*x_ref[2]**o+(0.49272814355)*x_ref[2]
        ref[1,4,1,1]=(0.851768453481)*x_ref[0]**o+(0.909492002062)*x_ref[0]+(-0.810487634736)*x_ref[1]**o+(0.919297425734)*x_ref[1]+(0.909780009526)*x_ref[2]**o+(0.369012187671)*x_ref[2]
        ref[1,4,2,0]=(-0.0531952997141)*x_ref[0]**o+(-0.224430435352)*x_ref[0]+(-0.578404055096)*x_ref[1]**o+(0.33447800953)*x_ref[1]+(-0.496336772814)*x_ref[2]**o+(0.725685669073)*x_ref[2]
        ref[1,4,2,1]=(0.541476219067)*x_ref[0]**o+(0.107132182409)*x_ref[0]+(-0.335093661175)*x_ref[1]**o+(0.227203070125)*x_ref[1]+(0.216737562875)*x_ref[2]**o+(-0.622943563245)*x_ref[2]
        ref[2,0,0,0]=(-0.756128419319)*x_ref[0]**o+(0.547579725255)*x_ref[0]+(0.687267799491)*x_ref[1]**o+(0.819030721043)*x_ref[1]+(-0.173876155736)*x_ref[2]**o+(-0.0301920564367)*x_ref[2]
        ref[2,0,0,1]=(-0.0759002903983)*x_ref[0]**o+(-0.438645875523)*x_ref[0]+(-0.783783394278)*x_ref[1]**o+(0.939615801765)*x_ref[1]+(-0.0129095800279)*x_ref[2]**o+(-0.723293627815)*x_ref[2]
        ref[2,0,1,0]=(-0.783224383717)*x_ref[0]**o+(-0.137339879229)*x_ref[0]+(-0.168000348627)*x_ref[1]**o+(0.223734275088)*x_ref[1]+(-0.582857962024)*x_ref[2]**o+(-0.215263259255)*x_ref[2]
        ref[2,0,1,1]=(0.866554819231)*x_ref[0]**o+(0.942792854767)*x_ref[0]+(0.506186289419)*x_ref[1]**o+(0.0775869106357)*x_ref[1]+(0.877213607927)*x_ref[2]**o+(-0.792692659916)*x_ref[2]
        ref[2,0,2,0]=(0.00917554286226)*x_ref[0]**o+(-0.5183359551)*x_ref[0]+(0.15698167339)*x_ref[1]**o+(-0.489805488541)*x_ref[1]+(-0.592166661492)*x_ref[2]**o+(0.391650695037)*x_ref[2]
        ref[2,0,2,1]=(0.356986311456)*x_ref[0]**o+(-0.14389994339)*x_ref[0]+(0.838182818055)*x_ref[1]**o+(-0.735112928546)*x_ref[1]+(-0.0333560634355)*x_ref[2]**o+(0.0855151256652)*x_ref[2]
        ref[2,1,0,0]=(0.357398992706)*x_ref[0]**o+(-0.733457701676)*x_ref[0]+(-0.142200448651)*x_ref[1]**o+(0.844914653536)*x_ref[1]+(-0.336954698769)*x_ref[2]**o+(-0.296222395555)*x_ref[2]
        ref[2,1,0,1]=(-0.950414726707)*x_ref[0]**o+(-0.134623616883)*x_ref[0]+(-0.403232115759)*x_ref[1]**o+(-0.539055139116)*x_ref[1]+(-0.0298791725363)*x_ref[2]**o+(-0.829664568373)*x_ref[2]
        ref[2,1,1,0]=(-0.656224623199)*x_ref[0]**o+(0.0872237997616)*x_ref[0]+(0.682387976548)*x_ref[1]**o+(-0.225630652343)*x_ref[1]+(-0.477261038385)*x_ref[2]**o+(-0.57098588683)*x_ref[2]
        ref[2,1,1,1]=(-0.504013815131)*x_ref[0]**o+(-0.0309238842436)*x_ref[0]+(0.350436304599)*x_ref[1]**o+(-0.891903027142)*x_ref[1]+(-0.352433668387)*x_ref[2]**o+(-0.603914711075)*x_ref[2]
        ref[2,1,2,0]=(0.505422678232)*x_ref[0]**o+(-0.36241271654)*x_ref[0]+(0.418645226594)*x_ref[1]**o+(-0.392311515242)*x_ref[1]+(-0.740372176332)*x_ref[2]**o+(-0.861742253361)*x_ref[2]
        ref[2,1,2,1]=(-0.853192617592)*x_ref[0]**o+(-0.269855659669)*x_ref[0]+(0.439592720974)*x_ref[1]**o+(0.62371411753)*x_ref[1]+(0.77806877626)*x_ref[2]**o+(0.923146838795)*x_ref[2]
        ref[2,2,0,0]=(0.228191521853)*x_ref[0]**o+(0.286276293845)*x_ref[0]+(0.205603167312)*x_ref[1]**o+(-0.208236801784)*x_ref[1]+(-0.787902086996)*x_ref[2]**o+(0.676274514448)*x_ref[2]
        ref[2,2,0,1]=(-0.75952147535)*x_ref[0]**o+(0.685910476453)*x_ref[0]+(-0.945474186679)*x_ref[1]**o+(0.257392146079)*x_ref[1]+(-0.502240261104)*x_ref[2]**o+(-0.446181698029)*x_ref[2]
        ref[2,2,1,0]=(-0.347627224549)*x_ref[0]**o+(0.425329908635)*x_ref[0]+(0.350298538346)*x_ref[1]**o+(-0.484357417885)*x_ref[1]+(0.538076599416)*x_ref[2]**o+(-0.0530149980096)*x_ref[2]
        ref[2,2,1,1]=(0.0539971839295)*x_ref[0]**o+(0.0721806780674)*x_ref[0]+(-0.309030912724)*x_ref[1]**o+(0.169874235759)*x_ref[1]+(-0.433682959518)*x_ref[2]**o+(0.813732426585)*x_ref[2]
        ref[2,2,2,0]=(0.336471441117)*x_ref[0]**o+(-0.484255919842)*x_ref[0]+(0.968308411862)*x_ref[1]**o+(-0.521533322453)*x_ref[1]+(0.633627764624)*x_ref[2]**o+(0.13497783288)*x_ref[2]
        ref[2,2,2,1]=(-0.602946926541)*x_ref[0]**o+(-0.67535176256)*x_ref[0]+(-0.527632780342)*x_ref[1]**o+(0.987852520439)*x_ref[1]+(0.426772082556)*x_ref[2]**o+(0.00805954755153)*x_ref[2]
        ref[2,3,0,0]=(0.834197069571)*x_ref[0]**o+(-0.213325947671)*x_ref[0]+(0.26098973561)*x_ref[1]**o+(-0.892672004839)*x_ref[1]+(0.690037689842)*x_ref[2]**o+(-0.554911545988)*x_ref[2]
        ref[2,3,0,1]=(-0.917087293134)*x_ref[0]**o+(-0.226247312973)*x_ref[0]+(0.901194492616)*x_ref[1]**o+(0.879779745124)*x_ref[1]+(-0.501788697362)*x_ref[2]**o+(0.62176481794)*x_ref[2]
        ref[2,3,1,0]=(-0.78132288334)*x_ref[0]**o+(0.249383051506)*x_ref[0]+(-0.276401042016)*x_ref[1]**o+(0.0913832980328)*x_ref[1]+(-0.199057369246)*x_ref[2]**o+(0.229465279886)*x_ref[2]
        ref[2,3,1,1]=(0.33249458134)*x_ref[0]**o+(-0.79179194291)*x_ref[0]+(0.146436436451)*x_ref[1]**o+(0.858879181025)*x_ref[1]+(0.069509608477)*x_ref[2]**o+(0.993263657)*x_ref[2]
        ref[2,3,2,0]=(0.103735002938)*x_ref[0]**o+(-0.326548520992)*x_ref[0]+(-0.826992179127)*x_ref[1]**o+(-0.883356562305)*x_ref[1]+(-0.340900003747)*x_ref[2]**o+(-0.134490572338)*x_ref[2]
        ref[2,3,2,1]=(0.10363967847)*x_ref[0]**o+(-0.171531371867)*x_ref[0]+(-0.929315735292)*x_ref[1]**o+(-0.719614719828)*x_ref[1]+(-0.592883409628)*x_ref[2]**o+(-0.641629610424)*x_ref[2]
        ref[2,4,0,0]=(-0.297999082535)*x_ref[0]**o+(0.439209856881)*x_ref[0]+(-0.751163554256)*x_ref[1]**o+(-0.189087784579)*x_ref[1]+(-0.795871127446)*x_ref[2]**o+(0.130133923162)*x_ref[2]
        ref[2,4,0,1]=(0.721793867996)*x_ref[0]**o+(-0.635599886382)*x_ref[0]+(0.801889483297)*x_ref[1]**o+(0.211160357461)*x_ref[1]+(0.237091518418)*x_ref[2]**o+(0.992344543271)*x_ref[2]
        ref[2,4,1,0]=(-0.775779912316)*x_ref[0]**o+(-0.376074847935)*x_ref[0]+(0.212663936842)*x_ref[1]**o+(-0.549546184149)*x_ref[1]+(-0.0854498425298)*x_ref[2]**o+(0.542469273724)*x_ref[2]
        ref[2,4,1,1]=(0.363553642873)*x_ref[0]**o+(-0.311431185265)*x_ref[0]+(-0.375820595299)*x_ref[1]**o+(-0.845438661463)*x_ref[1]+(-0.478899940153)*x_ref[2]**o+(0.0806897601294)*x_ref[2]
        ref[2,4,2,0]=(-0.489422873739)*x_ref[0]**o+(-0.157008373165)*x_ref[0]+(-0.267263773527)*x_ref[1]**o+(-0.53340253927)*x_ref[1]+(0.356782019115)*x_ref[2]**o+(0.117093532841)*x_ref[2]
        ref[2,4,2,1]=(0.106435699048)*x_ref[0]**o+(0.457452797294)*x_ref[0]+(-0.640139127483)*x_ref[1]**o+(-0.378451058366)*x_ref[1]+(-0.232397757243)*x_ref[2]**o+(-0.395640130256)*x_ref[2]
        ref[3,0,0,0]=(0.360807863609)*x_ref[0]**o+(0.689413996889)*x_ref[0]+(-0.267242935468)*x_ref[1]**o+(0.341694875336)*x_ref[1]+(0.955444260018)*x_ref[2]**o+(0.940815872612)*x_ref[2]
        ref[3,0,0,1]=(0.0640653983699)*x_ref[0]**o+(0.17954797305)*x_ref[0]+(0.739660609779)*x_ref[1]**o+(-0.174212756014)*x_ref[1]+(0.190134414562)*x_ref[2]**o+(-0.45741222058)*x_ref[2]
        ref[3,0,1,0]=(-0.642372239311)*x_ref[0]**o+(-0.518079894531)*x_ref[0]+(-0.843773918922)*x_ref[1]**o+(0.407352407859)*x_ref[1]+(0.0823899653044)*x_ref[2]**o+(0.655867974962)*x_ref[2]
        ref[3,0,1,1]=(-0.545184505438)*x_ref[0]**o+(-0.20749144757)*x_ref[0]+(0.656821663662)*x_ref[1]**o+(0.238823262425)*x_ref[1]+(0.645114250743)*x_ref[2]**o+(0.931193868381)*x_ref[2]
        ref[3,0,2,0]=(0.715326090131)*x_ref[0]**o+(-0.623038006294)*x_ref[0]+(-0.0272773549244)*x_ref[1]**o+(0.0443002059335)*x_ref[1]+(-0.639961942663)*x_ref[2]**o+(-0.863662064393)*x_ref[2]
        ref[3,0,2,1]=(-0.150998418072)*x_ref[0]**o+(-0.858320880863)*x_ref[0]+(-0.711824804062)*x_ref[1]**o+(-0.257945841504)*x_ref[1]+(-0.070002743188)*x_ref[2]**o+(-0.967136937623)*x_ref[2]
        ref[3,1,0,0]=(-0.5416631385)*x_ref[0]**o+(-0.315276228939)*x_ref[0]+(0.327960879182)*x_ref[1]**o+(0.309383847119)*x_ref[1]+(-0.531771040881)*x_ref[2]**o+(-0.503014905853)*x_ref[2]
        ref[3,1,0,1]=(-0.93431005785)*x_ref[0]**o+(-0.416716442301)*x_ref[0]+(0.737336965624)*x_ref[1]**o+(-0.676338424415)*x_ref[1]+(-0.638729889321)*x_ref[2]**o+(0.779664539115)*x_ref[2]
        ref[3,1,1,0]=(0.994892412301)*x_ref[0]**o+(0.0257912160747)*x_ref[0]+(-0.829652002123)*x_ref[1]**o+(0.412033947586)*x_ref[1]+(0.808588738484)*x_ref[2]**o+(-0.537100420921)*x_ref[2]
        ref[3,1,1,1]=(-0.0459132262405)*x_ref[0]**o+(-0.517692240206)*x_ref[0]+(0.66461361799)*x_ref[1]**o+(0.0202376762776)*x_ref[1]+(0.561286210305)*x_ref[2]**o+(-0.462378162744)*x_ref[2]
        ref[3,1,2,0]=(0.909401279733)*x_ref[0]**o+(-0.092132877706)*x_ref[0]+(-0.224813536034)*x_ref[1]**o+(0.318594065478)*x_ref[1]+(-0.959865958362)*x_ref[2]**o+(0.458327172218)*x_ref[2]
        ref[3,1,2,1]=(0.174314370931)*x_ref[0]**o+(-0.0149495584725)*x_ref[0]+(0.719658326164)*x_ref[1]**o+(0.416208781614)*x_ref[1]+(0.922363382172)*x_ref[2]**o+(-0.123220562614)*x_ref[2]
        ref[3,2,0,0]=(-0.362407421114)*x_ref[0]**o+(-0.35287313685)*x_ref[0]+(-0.712118894359)*x_ref[1]**o+(-0.749389429473)*x_ref[1]+(-0.520706676442)*x_ref[2]**o+(0.15576625076)*x_ref[2]
        ref[3,2,0,1]=(-0.620594009087)*x_ref[0]**o+(-0.866479846742)*x_ref[0]+(-0.429532862455)*x_ref[1]**o+(-0.736262913146)*x_ref[1]+(0.14914428425)*x_ref[2]**o+(0.00655098493102)*x_ref[2]
        ref[3,2,1,0]=(0.47291961358)*x_ref[0]**o+(-0.223022599565)*x_ref[0]+(-0.489191311572)*x_ref[1]**o+(-0.616554022459)*x_ref[1]+(0.525481658641)*x_ref[2]**o+(0.999870533503)*x_ref[2]
        ref[3,2,1,1]=(0.315715505936)*x_ref[0]**o+(-0.640835348385)*x_ref[0]+(0.768958709128)*x_ref[1]**o+(0.587161832357)*x_ref[1]+(0.454427338357)*x_ref[2]**o+(-0.133391799438)*x_ref[2]
        ref[3,2,2,0]=(-0.273381601564)*x_ref[0]**o+(-0.527114000062)*x_ref[0]+(-0.241611283455)*x_ref[1]**o+(0.511784453537)*x_ref[1]+(-0.558176752993)*x_ref[2]**o+(0.563602029554)*x_ref[2]
        ref[3,2,2,1]=(-0.554552149574)*x_ref[0]**o+(0.277189312969)*x_ref[0]+(0.528236290321)*x_ref[1]**o+(-0.56161965546)*x_ref[1]+(0.186320859613)*x_ref[2]**o+(-0.462794918456)*x_ref[2]
        ref[3,3,0,0]=(-0.379669006639)*x_ref[0]**o+(-0.290542994989)*x_ref[0]+(0.950886714589)*x_ref[1]**o+(0.113833056597)*x_ref[1]+(-0.395921494129)*x_ref[2]**o+(-0.530197898233)*x_ref[2]
        ref[3,3,0,1]=(-0.37250317146)*x_ref[0]**o+(0.266496776367)*x_ref[0]+(-0.992301806744)*x_ref[1]**o+(-0.215688062569)*x_ref[1]+(0.86298623844)*x_ref[2]**o+(0.932822343278)*x_ref[2]
        ref[3,3,1,0]=(-0.354585401042)*x_ref[0]**o+(0.472315228214)*x_ref[0]+(-0.604904767558)*x_ref[1]**o+(-0.868411440101)*x_ref[1]+(-0.434629251045)*x_ref[2]**o+(-0.4582356693)*x_ref[2]
        ref[3,3,1,1]=(-0.931569757365)*x_ref[0]**o+(0.062290982568)*x_ref[0]+(-0.402530852638)*x_ref[1]**o+(0.13606876168)*x_ref[1]+(-0.421824162622)*x_ref[2]**o+(0.754078648605)*x_ref[2]
        ref[3,3,2,0]=(0.370446872991)*x_ref[0]**o+(-0.678219724124)*x_ref[0]+(0.208629662303)*x_ref[1]**o+(0.742029980811)*x_ref[1]+(0.311836736294)*x_ref[2]**o+(-0.102558272951)*x_ref[2]
        ref[3,3,2,1]=(-0.15347338298)*x_ref[0]**o+(-0.798719756084)*x_ref[0]+(0.884663004721)*x_ref[1]**o+(-0.658935444548)*x_ref[1]+(-0.721604057065)*x_ref[2]**o+(0.299122598053)*x_ref[2]
        ref[3,4,0,0]=(0.544627747607)*x_ref[0]**o+(0.513489478068)*x_ref[0]+(-0.410347246011)*x_ref[1]**o+(0.610018445392)*x_ref[1]+(0.87473885042)*x_ref[2]**o+(0.914293254335)*x_ref[2]
        ref[3,4,0,1]=(-0.236207879972)*x_ref[0]**o+(0.46233924924)*x_ref[0]+(-0.472291476346)*x_ref[1]**o+(0.480397963815)*x_ref[1]+(-0.976610220911)*x_ref[2]**o+(-0.442554954343)*x_ref[2]
        ref[3,4,1,0]=(0.332927708298)*x_ref[0]**o+(0.737580232392)*x_ref[0]+(-0.56615273435)*x_ref[1]**o+(0.487228009962)*x_ref[1]+(-0.620250989525)*x_ref[2]**o+(0.699377689126)*x_ref[2]
        ref[3,4,1,1]=(-0.574680118107)*x_ref[0]**o+(-0.645001786533)*x_ref[0]+(-0.0623752350971)*x_ref[1]**o+(-0.943489194832)*x_ref[1]+(0.690135931129)*x_ref[2]**o+(0.047173398961)*x_ref[2]
        ref[3,4,2,0]=(0.218423477936)*x_ref[0]**o+(0.892507556482)*x_ref[0]+(-0.621739148946)*x_ref[1]**o+(-0.253753368776)*x_ref[1]+(-0.88306720835)*x_ref[2]**o+(0.0426804971625)*x_ref[2]
        ref[3,4,2,1]=(-0.901566218988)*x_ref[0]**o+(-0.0843229102423)*x_ref[0]+(-0.206812141605)*x_ref[1]**o+(-0.321102598716)*x_ref[1]+(-0.873774065466)*x_ref[2]**o+(-0.0507766082592)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.32868090106)*x[0]+(-0.840160165417)*x[1]
        ref=(-0.32868090106)*x_ref[0]+(-0.840160165417)*x_ref[1]
      else:
        s=(-1.32732649563)*x[0]+(0.20149624757)*x[1]+(0.701715136384)*x[2]
        ref=(-1.32732649563)*x_ref[0]+(0.20149624757)*x_ref[1]+(0.701715136384)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        s[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
        ref[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        ref[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
      else:
        s[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        s[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
        ref[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        ref[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.199811526227)*x[0]+(1.25013707432)*x[1]
        s[0,1]=(0.277539536007)*x[0]+(-0.023732930558)*x[1]
        s[0,2]=(-0.244973175438)*x[0]+(-0.501902270422)*x[1]
        s[0,3]=(0.249200744572)*x[0]+(0.520843463388)*x[1]
        s[0,4]=(-0.995168115444)*x[0]+(-0.259093571207)*x[1]
        s[1,0]=(0.179469283663)*x[0]+(-0.461808204127)*x[1]
        s[1,1]=(0.431508044494)*x[0]+(-0.423277002443)*x[1]
        s[1,2]=(0.732281368811)*x[0]+(-0.490387490227)*x[1]
        s[1,3]=(-0.665998043548)*x[0]+(-0.180996501441)*x[1]
        s[1,4]=(-0.773309943626)*x[0]+(-0.27171870837)*x[1]
        s[2,0]=(0.975757439909)*x[0]+(-0.980217145831)*x[1]
        s[2,1]=(-0.898690989301)*x[0]+(0.696886429725)*x[1]
        s[2,2]=(-0.568728495444)*x[0]+(0.171459808085)*x[1]
        s[2,3]=(-0.509734570618)*x[0]+(-0.318239573518)*x[1]
        s[2,4]=(-0.868166262762)*x[0]+(0.669164551211)*x[1]
        s[3,0]=(-1.11259074191)*x[0]+(-0.317962607489)*x[1]
        s[3,1]=(-0.735838783493)*x[0]+(-0.15621592888)*x[1]
        s[3,2]=(-0.348618152308)*x[0]+(0.346356772969)*x[1]
        s[3,3]=(1.58890219057)*x[0]+(-0.106915533345)*x[1]
        s[3,4]=(-0.739997195498)*x[0]+(-0.668986698058)*x[1]
        ref[0,0]=(0.199811526227)*x_ref[0]+(1.25013707432)*x_ref[1]
        ref[0,1]=(0.277539536007)*x_ref[0]+(-0.023732930558)*x_ref[1]
        ref[0,2]=(-0.244973175438)*x_ref[0]+(-0.501902270422)*x_ref[1]
        ref[0,3]=(0.249200744572)*x_ref[0]+(0.520843463388)*x_ref[1]
        ref[0,4]=(-0.995168115444)*x_ref[0]+(-0.259093571207)*x_ref[1]
        ref[1,0]=(0.179469283663)*x_ref[0]+(-0.461808204127)*x_ref[1]
        ref[1,1]=(0.431508044494)*x_ref[0]+(-0.423277002443)*x_ref[1]
        ref[1,2]=(0.732281368811)*x_ref[0]+(-0.490387490227)*x_ref[1]
        ref[1,3]=(-0.665998043548)*x_ref[0]+(-0.180996501441)*x_ref[1]
        ref[1,4]=(-0.773309943626)*x_ref[0]+(-0.27171870837)*x_ref[1]
        ref[2,0]=(0.975757439909)*x_ref[0]+(-0.980217145831)*x_ref[1]
        ref[2,1]=(-0.898690989301)*x_ref[0]+(0.696886429725)*x_ref[1]
        ref[2,2]=(-0.568728495444)*x_ref[0]+(0.171459808085)*x_ref[1]
        ref[2,3]=(-0.509734570618)*x_ref[0]+(-0.318239573518)*x_ref[1]
        ref[2,4]=(-0.868166262762)*x_ref[0]+(0.669164551211)*x_ref[1]
        ref[3,0]=(-1.11259074191)*x_ref[0]+(-0.317962607489)*x_ref[1]
        ref[3,1]=(-0.735838783493)*x_ref[0]+(-0.15621592888)*x_ref[1]
        ref[3,2]=(-0.348618152308)*x_ref[0]+(0.346356772969)*x_ref[1]
        ref[3,3]=(1.58890219057)*x_ref[0]+(-0.106915533345)*x_ref[1]
        ref[3,4]=(-0.739997195498)*x_ref[0]+(-0.668986698058)*x_ref[1]
      else:
        s[0,0]=(1.00508479945)*x[0]+(0.671813078579)*x[1]+(0.0352167301347)*x[2]
        s[0,1]=(0.832418208523)*x[0]+(-0.228971976603)*x[1]+(0.472369173779)*x[2]
        s[0,2]=(0.495212675956)*x[0]+(-0.850840698084)*x[1]+(-0.480464122957)*x[2]
        s[0,3]=(1.09800821588)*x[0]+(-1.16790708068)*x[1]+(-0.16448878954)*x[2]
        s[0,4]=(-0.142901980481)*x[0]+(-0.456358812317)*x[1]+(0.688171420193)*x[2]
        s[1,0]=(-0.452209235863)*x[0]+(0.835019801073)*x[1]+(-0.537900952591)*x[2]
        s[1,1]=(1.24188029232)*x[0]+(0.537899388384)*x[1]+(0.825366946283)*x[2]
        s[1,2]=(0.0568959695268)*x[0]+(-0.530615837561)*x[1]+(0.208390615403)*x[2]
        s[1,3]=(0.704415982126)*x[0]+(0.557947101676)*x[1]+(-1.41348269833)*x[2]
        s[1,4]=(0.23735208213)*x[0]+(0.188600394593)*x[1]+(-0.587309525946)*x[2]
        s[2,0]=(0.202262121276)*x[0]+(1.34583368863)*x[1]+(-1.416408583)*x[2]
        s[2,1]=(-0.889020760737)*x[0]+(1.18733643427)*x[1]+(1.11796325716)*x[2]
        s[2,2]=(0.0729529752138)*x[0]+(0.467461724385)*x[1]+(-1.26467465607)*x[2]
        s[2,3]=(0.553563173427)*x[0]+(-0.96991696383)*x[1]+(1.21489979267)*x[2]
        s[2,4]=(0.504157770518)*x[0]+(-0.0966112701707)*x[1]+(0.820591191307)*x[2]
        s[3,0]=(0.438942088634)*x[0]+(0.737754158644)*x[1]+(0.327747630162)*x[2]
        s[3,1]=(-0.966999874402)*x[0]+(0.218515238902)*x[1]+(-1.85502435057)*x[2]
        s[3,2]=(0.632366690747)*x[0]+(0.710212883666)*x[1]+(1.52843521678)*x[2]
        s[3,3]=(-0.139393405525)*x[0]+(-1.020102199)*x[1]+(0.485417385052)*x[2]
        s[3,4]=(-0.753676205348)*x[0]+(0.783454715482)*x[1]+(0.568363667465)*x[2]
        ref[0,0]=(1.00508479945)*x_ref[0]+(0.671813078579)*x_ref[1]+(0.0352167301347)*x_ref[2]
        ref[0,1]=(0.832418208523)*x_ref[0]+(-0.228971976603)*x_ref[1]+(0.472369173779)*x_ref[2]
        ref[0,2]=(0.495212675956)*x_ref[0]+(-0.850840698084)*x_ref[1]+(-0.480464122957)*x_ref[2]
        ref[0,3]=(1.09800821588)*x_ref[0]+(-1.16790708068)*x_ref[1]+(-0.16448878954)*x_ref[2]
        ref[0,4]=(-0.142901980481)*x_ref[0]+(-0.456358812317)*x_ref[1]+(0.688171420193)*x_ref[2]
        ref[1,0]=(-0.452209235863)*x_ref[0]+(0.835019801073)*x_ref[1]+(-0.537900952591)*x_ref[2]
        ref[1,1]=(1.24188029232)*x_ref[0]+(0.537899388384)*x_ref[1]+(0.825366946283)*x_ref[2]
        ref[1,2]=(0.0568959695268)*x_ref[0]+(-0.530615837561)*x_ref[1]+(0.208390615403)*x_ref[2]
        ref[1,3]=(0.704415982126)*x_ref[0]+(0.557947101676)*x_ref[1]+(-1.41348269833)*x_ref[2]
        ref[1,4]=(0.23735208213)*x_ref[0]+(0.188600394593)*x_ref[1]+(-0.587309525946)*x_ref[2]
        ref[2,0]=(0.202262121276)*x_ref[0]+(1.34583368863)*x_ref[1]+(-1.416408583)*x_ref[2]
        ref[2,1]=(-0.889020760737)*x_ref[0]+(1.18733643427)*x_ref[1]+(1.11796325716)*x_ref[2]
        ref[2,2]=(0.0729529752138)*x_ref[0]+(0.467461724385)*x_ref[1]+(-1.26467465607)*x_ref[2]
        ref[2,3]=(0.553563173427)*x_ref[0]+(-0.96991696383)*x_ref[1]+(1.21489979267)*x_ref[2]
        ref[2,4]=(0.504157770518)*x_ref[0]+(-0.0966112701707)*x_ref[1]+(0.820591191307)*x_ref[2]
        ref[3,0]=(0.438942088634)*x_ref[0]+(0.737754158644)*x_ref[1]+(0.327747630162)*x_ref[2]
        ref[3,1]=(-0.966999874402)*x_ref[0]+(0.218515238902)*x_ref[1]+(-1.85502435057)*x_ref[2]
        ref[3,2]=(0.632366690747)*x_ref[0]+(0.710212883666)*x_ref[1]+(1.52843521678)*x_ref[2]
        ref[3,3]=(-0.139393405525)*x_ref[0]+(-1.020102199)*x_ref[1]+(0.485417385052)*x_ref[2]
        ref[3,4]=(-0.753676205348)*x_ref[0]+(0.783454715482)*x_ref[1]+(0.568363667465)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.734232753652)*x[0]+(0.800930015514)*x[1]
        s[0,0,1]=(-0.219864689475)*x[0]+(0.157713027162)*x[1]
        s[0,1,0]=(-0.00664406256638)*x[0]+(0.872104848272)*x[1]
        s[0,1,1]=(-1.5575275283)*x[0]+(-1.43701883839)*x[1]
        s[1,0,0]=(1.60056664833)*x[0]+(1.6215424255)*x[1]
        s[1,0,1]=(1.5172017347)*x[0]+(-0.022919888148)*x[1]
        s[1,1,0]=(0.042690381655)*x[0]+(1.30019922)*x[1]
        s[1,1,1]=(0.105731664319)*x[0]+(-1.87027128366)*x[1]
        s[2,0,0]=(1.59675688345)*x[0]+(0.0608090042054)*x[1]
        s[2,0,1]=(-1.3520682478)*x[0]+(-0.881995801453)*x[1]
        s[2,1,0]=(1.08974009958)*x[0]+(0.13217220748)*x[1]
        s[2,1,1]=(-0.631056656103)*x[0]+(-0.548103546753)*x[1]
        s[3,0,0]=(1.59082745694)*x[0]+(0.101815532645)*x[1]
        s[3,0,1]=(-0.390741362781)*x[0]+(0.525254974689)*x[1]
        s[3,1,0]=(0.439652119858)*x[0]+(1.25127795686)*x[1]
        s[3,1,1]=(-1.10496039621)*x[0]+(0.0889406484341)*x[1]
        s[4,0,0]=(1.76164121088)*x[0]+(0.790919865199)*x[1]
        s[4,0,1]=(1.53483109)*x[0]+(1.15931695117)*x[1]
        s[4,1,0]=(-1.4226648704)*x[0]+(0.0380245720655)*x[1]
        s[4,1,1]=(-0.158958015165)*x[0]+(0.359005195468)*x[1]
        s[5,0,0]=(-0.0531318028725)*x[0]+(-0.0716200483232)*x[1]
        s[5,0,1]=(-1.64720026194)*x[0]+(-1.13699910338)*x[1]
        s[5,1,0]=(1.79036772333)*x[0]+(0.414445739998)*x[1]
        s[5,1,1]=(-0.168931889239)*x[0]+(1.55913218176)*x[1]
        ref[0,0,0]=(-0.734232753652)*x_ref[0]+(0.800930015514)*x_ref[1]
        ref[0,0,1]=(-0.219864689475)*x_ref[0]+(0.157713027162)*x_ref[1]
        ref[0,1,0]=(-0.00664406256638)*x_ref[0]+(0.872104848272)*x_ref[1]
        ref[0,1,1]=(-1.5575275283)*x_ref[0]+(-1.43701883839)*x_ref[1]
        ref[1,0,0]=(1.60056664833)*x_ref[0]+(1.6215424255)*x_ref[1]
        ref[1,0,1]=(1.5172017347)*x_ref[0]+(-0.022919888148)*x_ref[1]
        ref[1,1,0]=(0.042690381655)*x_ref[0]+(1.30019922)*x_ref[1]
        ref[1,1,1]=(0.105731664319)*x_ref[0]+(-1.87027128366)*x_ref[1]
        ref[2,0,0]=(1.59675688345)*x_ref[0]+(0.0608090042054)*x_ref[1]
        ref[2,0,1]=(-1.3520682478)*x_ref[0]+(-0.881995801453)*x_ref[1]
        ref[2,1,0]=(1.08974009958)*x_ref[0]+(0.13217220748)*x_ref[1]
        ref[2,1,1]=(-0.631056656103)*x_ref[0]+(-0.548103546753)*x_ref[1]
        ref[3,0,0]=(1.59082745694)*x_ref[0]+(0.101815532645)*x_ref[1]
        ref[3,0,1]=(-0.390741362781)*x_ref[0]+(0.525254974689)*x_ref[1]
        ref[3,1,0]=(0.439652119858)*x_ref[0]+(1.25127795686)*x_ref[1]
        ref[3,1,1]=(-1.10496039621)*x_ref[0]+(0.0889406484341)*x_ref[1]
        ref[4,0,0]=(1.76164121088)*x_ref[0]+(0.790919865199)*x_ref[1]
        ref[4,0,1]=(1.53483109)*x_ref[0]+(1.15931695117)*x_ref[1]
        ref[4,1,0]=(-1.4226648704)*x_ref[0]+(0.0380245720655)*x_ref[1]
        ref[4,1,1]=(-0.158958015165)*x_ref[0]+(0.359005195468)*x_ref[1]
        ref[5,0,0]=(-0.0531318028725)*x_ref[0]+(-0.0716200483232)*x_ref[1]
        ref[5,0,1]=(-1.64720026194)*x_ref[0]+(-1.13699910338)*x_ref[1]
        ref[5,1,0]=(1.79036772333)*x_ref[0]+(0.414445739998)*x_ref[1]
        ref[5,1,1]=(-0.168931889239)*x_ref[0]+(1.55913218176)*x_ref[1]
      else:
        s[0,0,0]=(0.030748782631)*x[0]+(0.0965401372343)*x[1]+(-1.19722517556)*x[2]
        s[0,0,1]=(-1.25981749671)*x[0]+(0.375214736326)*x[1]+(0.78094551805)*x[2]
        s[0,1,0]=(0.0859894648882)*x[0]+(-1.02857522767)*x[1]+(0.752176468687)*x[2]
        s[0,1,1]=(0.857538315926)*x[0]+(1.20788107416)*x[1]+(-1.00656644081)*x[2]
        s[1,0,0]=(-0.0208770456539)*x[0]+(0.296657331936)*x[1]+(1.49455132218)*x[2]
        s[1,0,1]=(-1.1213775487)*x[0]+(1.12351512934)*x[1]+(0.661972762047)*x[2]
        s[1,1,0]=(0.75336533272)*x[0]+(1.33059115242)*x[1]+(-0.43556744027)*x[2]
        s[1,1,1]=(0.467898749534)*x[0]+(1.53277916261)*x[1]+(-1.36041296872)*x[2]
        s[2,0,0]=(0.390444800695)*x[0]+(-0.0902034979398)*x[1]+(1.79547473252)*x[2]
        s[2,0,1]=(0.0488678241039)*x[0]+(0.000207248527408)*x[1]+(-0.231246108981)*x[2]
        s[2,1,0]=(0.907328661278)*x[0]+(-0.777592463716)*x[1]+(-0.949764289407)*x[2]
        s[2,1,1]=(-0.508314624748)*x[0]+(-0.0150508206233)*x[1]+(0.786266513715)*x[2]
        s[3,0,0]=(-0.264992651778)*x[0]+(1.23980464726)*x[1]+(0.480251929476)*x[2]
        s[3,0,1]=(-0.519373168965)*x[0]+(-0.091129773203)*x[1]+(-1.12961715008)*x[2]
        s[3,1,0]=(0.737417275735)*x[0]+(0.00073881678009)*x[1]+(-0.913229240795)*x[2]
        s[3,1,1]=(-0.565248614525)*x[0]+(0.232901289197)*x[1]+(-0.852815115095)*x[2]
        s[4,0,0]=(-0.513590194423)*x[0]+(-1.10607554932)*x[1]+(1.3269295152)*x[2]
        s[4,0,1]=(-0.0108694161983)*x[0]+(-1.39154818645)*x[1]+(1.5448779549)*x[2]
        s[4,1,0]=(1.25248373617)*x[0]+(-0.444522495013)*x[1]+(0.791060985516)*x[2]
        s[4,1,1]=(0.287880876137)*x[0]+(1.42265824896)*x[1]+(-1.16312992573)*x[2]
        s[5,0,0]=(-1.25472591758)*x[0]+(-1.48986095574)*x[1]+(1.49212797538)*x[2]
        s[5,0,1]=(-0.774594904232)*x[0]+(0.582557149894)*x[1]+(-0.460088096141)*x[2]
        s[5,1,0]=(0.404584427176)*x[0]+(-0.0808051689456)*x[1]+(0.522819318812)*x[2]
        s[5,1,1]=(1.15513940465)*x[0]+(1.09285861353)*x[1]+(-0.166414260437)*x[2]
        ref[0,0,0]=(0.030748782631)*x_ref[0]+(0.0965401372343)*x_ref[1]+(-1.19722517556)*x_ref[2]
        ref[0,0,1]=(-1.25981749671)*x_ref[0]+(0.375214736326)*x_ref[1]+(0.78094551805)*x_ref[2]
        ref[0,1,0]=(0.0859894648882)*x_ref[0]+(-1.02857522767)*x_ref[1]+(0.752176468687)*x_ref[2]
        ref[0,1,1]=(0.857538315926)*x_ref[0]+(1.20788107416)*x_ref[1]+(-1.00656644081)*x_ref[2]
        ref[1,0,0]=(-0.0208770456539)*x_ref[0]+(0.296657331936)*x_ref[1]+(1.49455132218)*x_ref[2]
        ref[1,0,1]=(-1.1213775487)*x_ref[0]+(1.12351512934)*x_ref[1]+(0.661972762047)*x_ref[2]
        ref[1,1,0]=(0.75336533272)*x_ref[0]+(1.33059115242)*x_ref[1]+(-0.43556744027)*x_ref[2]
        ref[1,1,1]=(0.467898749534)*x_ref[0]+(1.53277916261)*x_ref[1]+(-1.36041296872)*x_ref[2]
        ref[2,0,0]=(0.390444800695)*x_ref[0]+(-0.0902034979398)*x_ref[1]+(1.79547473252)*x_ref[2]
        ref[2,0,1]=(0.0488678241039)*x_ref[0]+(0.000207248527408)*x_ref[1]+(-0.231246108981)*x_ref[2]
        ref[2,1,0]=(0.907328661278)*x_ref[0]+(-0.777592463716)*x_ref[1]+(-0.949764289407)*x_ref[2]
        ref[2,1,1]=(-0.508314624748)*x_ref[0]+(-0.0150508206233)*x_ref[1]+(0.786266513715)*x_ref[2]
        ref[3,0,0]=(-0.264992651778)*x_ref[0]+(1.23980464726)*x_ref[1]+(0.480251929476)*x_ref[2]
        ref[3,0,1]=(-0.519373168965)*x_ref[0]+(-0.091129773203)*x_ref[1]+(-1.12961715008)*x_ref[2]
        ref[3,1,0]=(0.737417275735)*x_ref[0]+(0.00073881678009)*x_ref[1]+(-0.913229240795)*x_ref[2]
        ref[3,1,1]=(-0.565248614525)*x_ref[0]+(0.232901289197)*x_ref[1]+(-0.852815115095)*x_ref[2]
        ref[4,0,0]=(-0.513590194423)*x_ref[0]+(-1.10607554932)*x_ref[1]+(1.3269295152)*x_ref[2]
        ref[4,0,1]=(-0.0108694161983)*x_ref[0]+(-1.39154818645)*x_ref[1]+(1.5448779549)*x_ref[2]
        ref[4,1,0]=(1.25248373617)*x_ref[0]+(-0.444522495013)*x_ref[1]+(0.791060985516)*x_ref[2]
        ref[4,1,1]=(0.287880876137)*x_ref[0]+(1.42265824896)*x_ref[1]+(-1.16312992573)*x_ref[2]
        ref[5,0,0]=(-1.25472591758)*x_ref[0]+(-1.48986095574)*x_ref[1]+(1.49212797538)*x_ref[2]
        ref[5,0,1]=(-0.774594904232)*x_ref[0]+(0.582557149894)*x_ref[1]+(-0.460088096141)*x_ref[2]
        ref[5,1,0]=(0.404584427176)*x_ref[0]+(-0.0808051689456)*x_ref[1]+(0.522819318812)*x_ref[2]
        ref[5,1,1]=(1.15513940465)*x_ref[0]+(1.09285861353)*x_ref[1]+(-0.166414260437)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedSolution onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-1.26340514811)*x[0]+(0.779300786262)*x[1]
        s[0,0,0,1]=(0.184428277326)*x[0]+(-0.831497209177)*x[1]
        s[0,0,1,0]=(1.35946842566)*x[0]+(0.273881497376)*x[1]
        s[0,0,1,1]=(0.117792005007)*x[0]+(1.06185199224)*x[1]
        s[0,0,2,0]=(0.810342149523)*x[0]+(-0.271662508688)*x[1]
        s[0,0,2,1]=(-0.332968381669)*x[0]+(0.0660545206117)*x[1]
        s[0,1,0,0]=(0.413613455431)*x[0]+(0.641032202343)*x[1]
        s[0,1,0,1]=(0.773447709702)*x[0]+(0.48460517354)*x[1]
        s[0,1,1,0]=(0.173632755729)*x[0]+(0.0635245280713)*x[1]
        s[0,1,1,1]=(0.731914799404)*x[0]+(1.42146728852)*x[1]
        s[0,1,2,0]=(0.688894013616)*x[0]+(1.0339010436)*x[1]
        s[0,1,2,1]=(-0.534272384342)*x[0]+(0.186670150961)*x[1]
        s[0,2,0,0]=(-0.318755700985)*x[0]+(-1.35142313256)*x[1]
        s[0,2,0,1]=(0.973640957973)*x[0]+(-0.103253061427)*x[1]
        s[0,2,1,0]=(1.14046612599)*x[0]+(0.238774677715)*x[1]
        s[0,2,1,1]=(-0.383452508275)*x[0]+(-1.62240702377)*x[1]
        s[0,2,2,0]=(0.761272968366)*x[0]+(1.446646048)*x[1]
        s[0,2,2,1]=(0.183707769875)*x[0]+(-0.630747785283)*x[1]
        s[0,3,0,0]=(-0.474486603803)*x[0]+(-0.0855459523088)*x[1]
        s[0,3,0,1]=(-0.991769725003)*x[0]+(1.004616452)*x[1]
        s[0,3,1,0]=(0.261836325547)*x[0]+(0.353357820559)*x[1]
        s[0,3,1,1]=(0.808880823505)*x[0]+(-0.495896346674)*x[1]
        s[0,3,2,0]=(1.06375339919)*x[0]+(0.00408804480046)*x[1]
        s[0,3,2,1]=(-1.32927483163)*x[0]+(-0.0833136005835)*x[1]
        s[0,4,0,0]=(1.13794063387)*x[0]+(-1.1458884083)*x[1]
        s[0,4,0,1]=(0.73883590181)*x[0]+(1.45878975539)*x[1]
        s[0,4,1,0]=(0.428378040368)*x[0]+(-1.25324354587)*x[1]
        s[0,4,1,1]=(1.19772514118)*x[0]+(0.457504803527)*x[1]
        s[0,4,2,0]=(-0.646891239861)*x[0]+(1.14291257798)*x[1]
        s[0,4,2,1]=(-0.579096313066)*x[0]+(-0.0276859933543)*x[1]
        s[1,0,0,0]=(-0.48935920813)*x[0]+(-0.327683586833)*x[1]
        s[1,0,0,1]=(0.59804841111)*x[0]+(-0.452980496343)*x[1]
        s[1,0,1,0]=(0.608017924036)*x[0]+(0.561553199982)*x[1]
        s[1,0,1,1]=(-0.00361014464582)*x[0]+(0.151090357053)*x[1]
        s[1,0,2,0]=(1.24385396019)*x[0]+(-0.0213646134992)*x[1]
        s[1,0,2,1]=(-0.219031743102)*x[0]+(-0.402476762676)*x[1]
        s[1,1,0,0]=(-0.196889997171)*x[0]+(0.675626499802)*x[1]
        s[1,1,0,1]=(1.23421246519)*x[0]+(-1.72367262236)*x[1]
        s[1,1,1,0]=(1.35571029115)*x[0]+(0.506057274132)*x[1]
        s[1,1,1,1]=(0.310699908104)*x[0]+(-0.292826394288)*x[1]
        s[1,1,2,0]=(0.122289986824)*x[0]+(-0.15947928964)*x[1]
        s[1,1,2,1]=(0.662808118991)*x[0]+(-0.927311967494)*x[1]
        s[1,2,0,0]=(-0.215898581104)*x[0]+(0.32902895413)*x[1]
        s[1,2,0,1]=(0.393145660435)*x[0]+(1.49040999822)*x[1]
        s[1,2,1,0]=(0.904144553172)*x[0]+(0.199605443931)*x[1]
        s[1,2,1,1]=(0.535979155678)*x[0]+(1.78418572873)*x[1]
        s[1,2,2,0]=(1.15578250862)*x[0]+(0.167811579149)*x[1]
        s[1,2,2,1]=(-0.0844120860286)*x[0]+(-0.56656817877)*x[1]
        s[1,3,0,0]=(0.760581325264)*x[0]+(1.00766443232)*x[1]
        s[1,3,0,1]=(-0.199244602194)*x[0]+(-1.54110932148)*x[1]
        s[1,3,1,0]=(1.40934237781)*x[0]+(0.700934912313)*x[1]
        s[1,3,1,1]=(-0.702507416172)*x[0]+(-0.206530323456)*x[1]
        s[1,3,2,0]=(-0.234301811191)*x[0]+(1.41101365089)*x[1]
        s[1,3,2,1]=(1.19609435879)*x[0]+(0.295560183318)*x[1]
        s[1,4,0,0]=(1.15287205838)*x[0]+(0.630564383314)*x[1]
        s[1,4,0,1]=(-0.993907995704)*x[0]+(-0.30836527499)*x[1]
        s[1,4,1,0]=(-0.265158946854)*x[0]+(0.689120014144)*x[1]
        s[1,4,1,1]=(0.503151681106)*x[0]+(-0.928730587416)*x[1]
        s[1,4,2,0]=(-1.0599860135)*x[0]+(-0.696932543908)*x[1]
        s[1,4,2,1]=(-0.365096260865)*x[0]+(0.443129268519)*x[1]
        s[2,0,0,0]=(0.412604298623)*x[0]+(0.591441458509)*x[1]
        s[2,0,0,1]=(0.536034627643)*x[0]+(-0.735261974283)*x[1]
        s[2,0,1,0]=(-0.336900405282)*x[0]+(-1.38493113859)*x[1]
        s[2,0,1,1]=(-0.50634870942)*x[0]+(0.0376544898311)*x[1]
        s[2,0,2,0]=(-0.0595134635196)*x[0]+(0.067285997446)*x[1]
        s[2,0,2,1]=(0.308894832145)*x[0]+(-1.94600763342)*x[1]
        s[2,1,0,0]=(0.443019228067)*x[0]+(-0.68858070865)*x[1]
        s[2,1,0,1]=(1.4461941226)*x[0]+(0.711186579745)*x[1]
        s[2,1,1,0]=(0.178777491574)*x[0]+(-0.746259056845)*x[1]
        s[2,1,1,1]=(1.24682492601)*x[0]+(-0.987057469956)*x[1]
        s[2,1,2,0]=(1.1863711951)*x[0]+(-0.853660028797)*x[1]
        s[2,1,2,1]=(1.13176028078)*x[0]+(-0.887521745194)*x[1]
        s[2,2,0,0]=(0.37579402707)*x[0]+(-0.462533529495)*x[1]
        s[2,2,0,1]=(-1.27778045393)*x[0]+(0.383310656984)*x[1]
        s[2,2,1,0]=(-0.0871651954146)*x[0]+(0.0266612015617)*x[1]
        s[2,2,1,1]=(1.26455972911)*x[0]+(0.597556950133)*x[1]
        s[2,2,2,0]=(0.397071934531)*x[0]+(-0.792146197976)*x[1]
        s[2,2,2,1]=(-0.531842419771)*x[0]+(-0.437895983867)*x[1]
        s[2,3,0,0]=(0.253602494754)*x[0]+(-0.943660021452)*x[1]
        s[2,3,0,1]=(0.381472581352)*x[0]+(1.36193258173)*x[1]
        s[2,3,1,0]=(0.530962583291)*x[0]+(0.853717668569)*x[1]
        s[2,3,1,1]=(-0.233925714076)*x[0]+(0.20197385961)*x[1]
        s[2,3,2,0]=(0.148804235592)*x[0]+(-0.116553369538)*x[1]
        s[2,3,2,1]=(-1.79360831582)*x[0]+(1.62247077485)*x[1]
        s[2,4,0,0]=(0.617987809937)*x[0]+(1.27685238582)*x[1]
        s[2,4,0,1]=(-0.359480838494)*x[0]+(-0.11716388114)*x[1]
        s[2,4,1,0]=(0.204308581462)*x[0]+(0.0241091749363)*x[1]
        s[2,4,1,1]=(0.424779981556)*x[0]+(-1.01811123317)*x[1]
        s[2,4,2,0]=(-0.708751234946)*x[0]+(0.714679051838)*x[1]
        s[2,4,2,1]=(-0.352350728857)*x[0]+(1.47198192719)*x[1]
        s[3,0,0,0]=(-0.0592432179871)*x[0]+(0.0900836798021)*x[1]
        s[3,0,0,1]=(-0.192266211858)*x[0]+(0.561246968252)*x[1]
        s[3,0,1,0]=(0.30878942472)*x[0]+(-0.578076897247)*x[1]
        s[3,0,1,1]=(-0.246226203035)*x[0]+(0.68840556577)*x[1]
        s[3,0,2,0]=(-0.42433037357)*x[0]+(-0.168333732109)*x[1]
        s[3,0,2,1]=(0.501931661511)*x[0]+(1.53237442299)*x[1]
        s[3,1,0,0]=(-0.835618728532)*x[0]+(-0.823254922183)*x[1]
        s[3,1,0,1]=(-0.609502259942)*x[0]+(0.139549090645)*x[1]
        s[3,1,1,0]=(-0.690897498494)*x[0]+(0.280254985362)*x[1]
        s[3,1,1,1]=(-0.89040023732)*x[0]+(0.616118941926)*x[1]
        s[3,1,2,0]=(-0.90402399371)*x[0]+(-0.399494199588)*x[1]
        s[3,1,2,1]=(1.31002070769)*x[0]+(-0.26001592275)*x[1]
        s[3,2,0,0]=(-0.741667891927)*x[0]+(-1.13827783123)*x[1]
        s[3,2,0,1]=(0.879168335296)*x[0]+(0.232648259015)*x[1]
        s[3,2,1,0]=(0.737095222363)*x[0]+(-1.26232530994)*x[1]
        s[3,2,1,1]=(-0.523976448837)*x[0]+(0.480347424944)*x[1]
        s[3,2,2,0]=(-1.61727611988)*x[0]+(-0.464564622523)*x[1]
        s[3,2,2,1]=(0.928282179697)*x[0]+(0.629575166525)*x[1]
        s[3,3,0,0]=(0.748679757793)*x[0]+(1.26002219505)*x[1]
        s[3,3,0,1]=(-0.497730460683)*x[0]+(0.963512065302)*x[1]
        s[3,3,1,0]=(-0.999580985099)*x[0]+(0.933789269497)*x[1]
        s[3,3,1,1]=(-0.151096110981)*x[0]+(-0.739374723252)*x[1]
        s[3,3,2,0]=(0.257880503573)*x[0]+(-1.07115978896)*x[1]
        s[3,3,2,1]=(0.188404246313)*x[0]+(0.365000571954)*x[1]
        s[3,4,0,0]=(1.63625895457)*x[0]+(-0.685581590196)*x[1]
        s[3,4,0,1]=(-0.665388096854)*x[0]+(-0.9121332766)*x[1]
        s[3,4,1,0]=(1.76715282849)*x[0]+(-1.15317031503)*x[1]
        s[3,4,1,1]=(-0.378469229808)*x[0]+(-0.897880026236)*x[1]
        s[3,4,2,0]=(0.183976703061)*x[0]+(1.76560300417)*x[1]
        s[3,4,2,1]=(-0.569272512314)*x[0]+(1.09830336519)*x[1]
        ref[0,0,0,0]=(-1.26340514811)*x_ref[0]+(0.779300786262)*x_ref[1]
        ref[0,0,0,1]=(0.184428277326)*x_ref[0]+(-0.831497209177)*x_ref[1]
        ref[0,0,1,0]=(1.35946842566)*x_ref[0]+(0.273881497376)*x_ref[1]
        ref[0,0,1,1]=(0.117792005007)*x_ref[0]+(1.06185199224)*x_ref[1]
        ref[0,0,2,0]=(0.810342149523)*x_ref[0]+(-0.271662508688)*x_ref[1]
        ref[0,0,2,1]=(-0.332968381669)*x_ref[0]+(0.0660545206117)*x_ref[1]
        ref[0,1,0,0]=(0.413613455431)*x_ref[0]+(0.641032202343)*x_ref[1]
        ref[0,1,0,1]=(0.773447709702)*x_ref[0]+(0.48460517354)*x_ref[1]
        ref[0,1,1,0]=(0.173632755729)*x_ref[0]+(0.0635245280713)*x_ref[1]
        ref[0,1,1,1]=(0.731914799404)*x_ref[0]+(1.42146728852)*x_ref[1]
        ref[0,1,2,0]=(0.688894013616)*x_ref[0]+(1.0339010436)*x_ref[1]
        ref[0,1,2,1]=(-0.534272384342)*x_ref[0]+(0.186670150961)*x_ref[1]
        ref[0,2,0,0]=(-0.318755700985)*x_ref[0]+(-1.35142313256)*x_ref[1]
        ref[0,2,0,1]=(0.973640957973)*x_ref[0]+(-0.103253061427)*x_ref[1]
        ref[0,2,1,0]=(1.14046612599)*x_ref[0]+(0.238774677715)*x_ref[1]
        ref[0,2,1,1]=(-0.383452508275)*x_ref[0]+(-1.62240702377)*x_ref[1]
        ref[0,2,2,0]=(0.761272968366)*x_ref[0]+(1.446646048)*x_ref[1]
        ref[0,2,2,1]=(0.183707769875)*x_ref[0]+(-0.630747785283)*x_ref[1]
        ref[0,3,0,0]=(-0.474486603803)*x_ref[0]+(-0.0855459523088)*x_ref[1]
        ref[0,3,0,1]=(-0.991769725003)*x_ref[0]+(1.004616452)*x_ref[1]
        ref[0,3,1,0]=(0.261836325547)*x_ref[0]+(0.353357820559)*x_ref[1]
        ref[0,3,1,1]=(0.808880823505)*x_ref[0]+(-0.495896346674)*x_ref[1]
        ref[0,3,2,0]=(1.06375339919)*x_ref[0]+(0.00408804480046)*x_ref[1]
        ref[0,3,2,1]=(-1.32927483163)*x_ref[0]+(-0.0833136005835)*x_ref[1]
        ref[0,4,0,0]=(1.13794063387)*x_ref[0]+(-1.1458884083)*x_ref[1]
        ref[0,4,0,1]=(0.73883590181)*x_ref[0]+(1.45878975539)*x_ref[1]
        ref[0,4,1,0]=(0.428378040368)*x_ref[0]+(-1.25324354587)*x_ref[1]
        ref[0,4,1,1]=(1.19772514118)*x_ref[0]+(0.457504803527)*x_ref[1]
        ref[0,4,2,0]=(-0.646891239861)*x_ref[0]+(1.14291257798)*x_ref[1]
        ref[0,4,2,1]=(-0.579096313066)*x_ref[0]+(-0.0276859933543)*x_ref[1]
        ref[1,0,0,0]=(-0.48935920813)*x_ref[0]+(-0.327683586833)*x_ref[1]
        ref[1,0,0,1]=(0.59804841111)*x_ref[0]+(-0.452980496343)*x_ref[1]
        ref[1,0,1,0]=(0.608017924036)*x_ref[0]+(0.561553199982)*x_ref[1]
        ref[1,0,1,1]=(-0.00361014464582)*x_ref[0]+(0.151090357053)*x_ref[1]
        ref[1,0,2,0]=(1.24385396019)*x_ref[0]+(-0.0213646134992)*x_ref[1]
        ref[1,0,2,1]=(-0.219031743102)*x_ref[0]+(-0.402476762676)*x_ref[1]
        ref[1,1,0,0]=(-0.196889997171)*x_ref[0]+(0.675626499802)*x_ref[1]
        ref[1,1,0,1]=(1.23421246519)*x_ref[0]+(-1.72367262236)*x_ref[1]
        ref[1,1,1,0]=(1.35571029115)*x_ref[0]+(0.506057274132)*x_ref[1]
        ref[1,1,1,1]=(0.310699908104)*x_ref[0]+(-0.292826394288)*x_ref[1]
        ref[1,1,2,0]=(0.122289986824)*x_ref[0]+(-0.15947928964)*x_ref[1]
        ref[1,1,2,1]=(0.662808118991)*x_ref[0]+(-0.927311967494)*x_ref[1]
        ref[1,2,0,0]=(-0.215898581104)*x_ref[0]+(0.32902895413)*x_ref[1]
        ref[1,2,0,1]=(0.393145660435)*x_ref[0]+(1.49040999822)*x_ref[1]
        ref[1,2,1,0]=(0.904144553172)*x_ref[0]+(0.199605443931)*x_ref[1]
        ref[1,2,1,1]=(0.535979155678)*x_ref[0]+(1.78418572873)*x_ref[1]
        ref[1,2,2,0]=(1.15578250862)*x_ref[0]+(0.167811579149)*x_ref[1]
        ref[1,2,2,1]=(-0.0844120860286)*x_ref[0]+(-0.56656817877)*x_ref[1]
        ref[1,3,0,0]=(0.760581325264)*x_ref[0]+(1.00766443232)*x_ref[1]
        ref[1,3,0,1]=(-0.199244602194)*x_ref[0]+(-1.54110932148)*x_ref[1]
        ref[1,3,1,0]=(1.40934237781)*x_ref[0]+(0.700934912313)*x_ref[1]
        ref[1,3,1,1]=(-0.702507416172)*x_ref[0]+(-0.206530323456)*x_ref[1]
        ref[1,3,2,0]=(-0.234301811191)*x_ref[0]+(1.41101365089)*x_ref[1]
        ref[1,3,2,1]=(1.19609435879)*x_ref[0]+(0.295560183318)*x_ref[1]
        ref[1,4,0,0]=(1.15287205838)*x_ref[0]+(0.630564383314)*x_ref[1]
        ref[1,4,0,1]=(-0.993907995704)*x_ref[0]+(-0.30836527499)*x_ref[1]
        ref[1,4,1,0]=(-0.265158946854)*x_ref[0]+(0.689120014144)*x_ref[1]
        ref[1,4,1,1]=(0.503151681106)*x_ref[0]+(-0.928730587416)*x_ref[1]
        ref[1,4,2,0]=(-1.0599860135)*x_ref[0]+(-0.696932543908)*x_ref[1]
        ref[1,4,2,1]=(-0.365096260865)*x_ref[0]+(0.443129268519)*x_ref[1]
        ref[2,0,0,0]=(0.412604298623)*x_ref[0]+(0.591441458509)*x_ref[1]
        ref[2,0,0,1]=(0.536034627643)*x_ref[0]+(-0.735261974283)*x_ref[1]
        ref[2,0,1,0]=(-0.336900405282)*x_ref[0]+(-1.38493113859)*x_ref[1]
        ref[2,0,1,1]=(-0.50634870942)*x_ref[0]+(0.0376544898311)*x_ref[1]
        ref[2,0,2,0]=(-0.0595134635196)*x_ref[0]+(0.067285997446)*x_ref[1]
        ref[2,0,2,1]=(0.308894832145)*x_ref[0]+(-1.94600763342)*x_ref[1]
        ref[2,1,0,0]=(0.443019228067)*x_ref[0]+(-0.68858070865)*x_ref[1]
        ref[2,1,0,1]=(1.4461941226)*x_ref[0]+(0.711186579745)*x_ref[1]
        ref[2,1,1,0]=(0.178777491574)*x_ref[0]+(-0.746259056845)*x_ref[1]
        ref[2,1,1,1]=(1.24682492601)*x_ref[0]+(-0.987057469956)*x_ref[1]
        ref[2,1,2,0]=(1.1863711951)*x_ref[0]+(-0.853660028797)*x_ref[1]
        ref[2,1,2,1]=(1.13176028078)*x_ref[0]+(-0.887521745194)*x_ref[1]
        ref[2,2,0,0]=(0.37579402707)*x_ref[0]+(-0.462533529495)*x_ref[1]
        ref[2,2,0,1]=(-1.27778045393)*x_ref[0]+(0.383310656984)*x_ref[1]
        ref[2,2,1,0]=(-0.0871651954146)*x_ref[0]+(0.0266612015617)*x_ref[1]
        ref[2,2,1,1]=(1.26455972911)*x_ref[0]+(0.597556950133)*x_ref[1]
        ref[2,2,2,0]=(0.397071934531)*x_ref[0]+(-0.792146197976)*x_ref[1]
        ref[2,2,2,1]=(-0.531842419771)*x_ref[0]+(-0.437895983867)*x_ref[1]
        ref[2,3,0,0]=(0.253602494754)*x_ref[0]+(-0.943660021452)*x_ref[1]
        ref[2,3,0,1]=(0.381472581352)*x_ref[0]+(1.36193258173)*x_ref[1]
        ref[2,3,1,0]=(0.530962583291)*x_ref[0]+(0.853717668569)*x_ref[1]
        ref[2,3,1,1]=(-0.233925714076)*x_ref[0]+(0.20197385961)*x_ref[1]
        ref[2,3,2,0]=(0.148804235592)*x_ref[0]+(-0.116553369538)*x_ref[1]
        ref[2,3,2,1]=(-1.79360831582)*x_ref[0]+(1.62247077485)*x_ref[1]
        ref[2,4,0,0]=(0.617987809937)*x_ref[0]+(1.27685238582)*x_ref[1]
        ref[2,4,0,1]=(-0.359480838494)*x_ref[0]+(-0.11716388114)*x_ref[1]
        ref[2,4,1,0]=(0.204308581462)*x_ref[0]+(0.0241091749363)*x_ref[1]
        ref[2,4,1,1]=(0.424779981556)*x_ref[0]+(-1.01811123317)*x_ref[1]
        ref[2,4,2,0]=(-0.708751234946)*x_ref[0]+(0.714679051838)*x_ref[1]
        ref[2,4,2,1]=(-0.352350728857)*x_ref[0]+(1.47198192719)*x_ref[1]
        ref[3,0,0,0]=(-0.0592432179871)*x_ref[0]+(0.0900836798021)*x_ref[1]
        ref[3,0,0,1]=(-0.192266211858)*x_ref[0]+(0.561246968252)*x_ref[1]
        ref[3,0,1,0]=(0.30878942472)*x_ref[0]+(-0.578076897247)*x_ref[1]
        ref[3,0,1,1]=(-0.246226203035)*x_ref[0]+(0.68840556577)*x_ref[1]
        ref[3,0,2,0]=(-0.42433037357)*x_ref[0]+(-0.168333732109)*x_ref[1]
        ref[3,0,2,1]=(0.501931661511)*x_ref[0]+(1.53237442299)*x_ref[1]
        ref[3,1,0,0]=(-0.835618728532)*x_ref[0]+(-0.823254922183)*x_ref[1]
        ref[3,1,0,1]=(-0.609502259942)*x_ref[0]+(0.139549090645)*x_ref[1]
        ref[3,1,1,0]=(-0.690897498494)*x_ref[0]+(0.280254985362)*x_ref[1]
        ref[3,1,1,1]=(-0.89040023732)*x_ref[0]+(0.616118941926)*x_ref[1]
        ref[3,1,2,0]=(-0.90402399371)*x_ref[0]+(-0.399494199588)*x_ref[1]
        ref[3,1,2,1]=(1.31002070769)*x_ref[0]+(-0.26001592275)*x_ref[1]
        ref[3,2,0,0]=(-0.741667891927)*x_ref[0]+(-1.13827783123)*x_ref[1]
        ref[3,2,0,1]=(0.879168335296)*x_ref[0]+(0.232648259015)*x_ref[1]
        ref[3,2,1,0]=(0.737095222363)*x_ref[0]+(-1.26232530994)*x_ref[1]
        ref[3,2,1,1]=(-0.523976448837)*x_ref[0]+(0.480347424944)*x_ref[1]
        ref[3,2,2,0]=(-1.61727611988)*x_ref[0]+(-0.464564622523)*x_ref[1]
        ref[3,2,2,1]=(0.928282179697)*x_ref[0]+(0.629575166525)*x_ref[1]
        ref[3,3,0,0]=(0.748679757793)*x_ref[0]+(1.26002219505)*x_ref[1]
        ref[3,3,0,1]=(-0.497730460683)*x_ref[0]+(0.963512065302)*x_ref[1]
        ref[3,3,1,0]=(-0.999580985099)*x_ref[0]+(0.933789269497)*x_ref[1]
        ref[3,3,1,1]=(-0.151096110981)*x_ref[0]+(-0.739374723252)*x_ref[1]
        ref[3,3,2,0]=(0.257880503573)*x_ref[0]+(-1.07115978896)*x_ref[1]
        ref[3,3,2,1]=(0.188404246313)*x_ref[0]+(0.365000571954)*x_ref[1]
        ref[3,4,0,0]=(1.63625895457)*x_ref[0]+(-0.685581590196)*x_ref[1]
        ref[3,4,0,1]=(-0.665388096854)*x_ref[0]+(-0.9121332766)*x_ref[1]
        ref[3,4,1,0]=(1.76715282849)*x_ref[0]+(-1.15317031503)*x_ref[1]
        ref[3,4,1,1]=(-0.378469229808)*x_ref[0]+(-0.897880026236)*x_ref[1]
        ref[3,4,2,0]=(0.183976703061)*x_ref[0]+(1.76560300417)*x_ref[1]
        ref[3,4,2,1]=(-0.569272512314)*x_ref[0]+(1.09830336519)*x_ref[1]
      else:
        s[0,0,0,0]=(0.146894168765)*x[0]+(1.25822855062)*x[1]+(0.088660814219)*x[2]
        s[0,0,0,1]=(0.572576296123)*x[0]+(0.670191219644)*x[1]+(-0.280619677889)*x[2]
        s[0,0,1,0]=(-1.19817665348)*x[0]+(1.35545322011)*x[1]+(-0.279751222491)*x[2]
        s[0,0,1,1]=(-0.0558236682639)*x[0]+(-1.61042107055)*x[1]+(0.622456635293)*x[2]
        s[0,0,2,0]=(-0.154784205079)*x[0]+(-0.52314996688)*x[1]+(1.0590614932)*x[2]
        s[0,0,2,1]=(1.58101401758)*x[0]+(-0.610736715272)*x[1]+(-0.487233776225)*x[2]
        s[0,1,0,0]=(0.0117365392099)*x[0]+(1.02397373706)*x[1]+(0.407377316737)*x[2]
        s[0,1,0,1]=(0.460387220384)*x[0]+(1.10329384655)*x[1]+(0.252209465365)*x[2]
        s[0,1,1,0]=(-1.33153199048)*x[0]+(-0.759669253534)*x[1]+(-0.856223614724)*x[2]
        s[0,1,1,1]=(-1.00599668771)*x[0]+(0.321343926218)*x[1]+(-0.198860330101)*x[2]
        s[0,1,2,0]=(1.00741183642)*x[0]+(-0.148473398917)*x[1]+(-0.484181353094)*x[2]
        s[0,1,2,1]=(-1.46490648953)*x[0]+(0.621077170698)*x[1]+(0.802865695373)*x[2]
        s[0,2,0,0]=(1.18498612655)*x[0]+(-0.511705644809)*x[1]+(0.563568352278)*x[2]
        s[0,2,0,1]=(-0.807427977876)*x[0]+(-0.252372201752)*x[1]+(0.251722661898)*x[2]
        s[0,2,1,0]=(0.524572175367)*x[0]+(0.49906104527)*x[1]+(-0.601836374873)*x[2]
        s[0,2,1,1]=(-0.318976775137)*x[0]+(-0.153065654914)*x[1]+(-0.458888325378)*x[2]
        s[0,2,2,0]=(-1.45513892332)*x[0]+(-0.252635212625)*x[1]+(0.490153491036)*x[2]
        s[0,2,2,1]=(-0.439422986461)*x[0]+(-0.937409556421)*x[1]+(0.209106988492)*x[2]
        s[0,3,0,0]=(-0.26828401981)*x[0]+(1.70808529771)*x[1]+(0.431245149358)*x[2]
        s[0,3,0,1]=(-0.831846912527)*x[0]+(0.572822507173)*x[1]+(-0.283130753547)*x[2]
        s[0,3,1,0]=(-0.625617835832)*x[0]+(-1.25495997626)*x[1]+(1.28743784867)*x[2]
        s[0,3,1,1]=(1.47223047075)*x[0]+(0.274365456026)*x[1]+(1.34355769078)*x[2]
        s[0,3,2,0]=(-0.0506967319243)*x[0]+(1.16569056569)*x[1]+(-0.965008332019)*x[2]
        s[0,3,2,1]=(0.0106139754664)*x[0]+(0.274252087281)*x[1]+(-0.501970174754)*x[2]
        s[0,4,0,0]=(0.717476010799)*x[0]+(0.380533443018)*x[1]+(-0.304127891816)*x[2]
        s[0,4,0,1]=(-1.29133512255)*x[0]+(-1.16667771199)*x[1]+(0.275809086731)*x[2]
        s[0,4,1,0]=(0.137939049278)*x[0]+(-0.302950342448)*x[1]+(-0.28865204469)*x[2]
        s[0,4,1,1]=(-0.857457318663)*x[0]+(0.423143726447)*x[1]+(0.749889416366)*x[2]
        s[0,4,2,0]=(-0.994428027329)*x[0]+(-0.377780341115)*x[1]+(-0.997480394895)*x[2]
        s[0,4,2,1]=(-0.681653236311)*x[0]+(-0.091905911233)*x[1]+(-0.877007024631)*x[2]
        s[1,0,0,0]=(0.33375482228)*x[0]+(-1.08816236114)*x[1]+(0.154701633038)*x[2]
        s[1,0,0,1]=(0.979227592935)*x[0]+(0.571415419963)*x[1]+(-0.20312466228)*x[2]
        s[1,0,1,0]=(1.59673013189)*x[0]+(1.01626905603)*x[1]+(0.216806440372)*x[2]
        s[1,0,1,1]=(-0.276034446694)*x[0]+(-0.168494544051)*x[1]+(0.981220114947)*x[2]
        s[1,0,2,0]=(0.205085469341)*x[0]+(-0.507970934322)*x[1]+(-0.724745229941)*x[2]
        s[1,0,2,1]=(-0.629331749902)*x[0]+(-1.27086230989)*x[1]+(-1.08479549542)*x[2]
        s[1,1,0,0]=(-0.786670678498)*x[0]+(1.24461814343)*x[1]+(0.644579154985)*x[2]
        s[1,1,0,1]=(1.05414628662)*x[0]+(0.901590532372)*x[1]+(0.809519110188)*x[2]
        s[1,1,1,0]=(-0.679152609307)*x[0]+(0.84101357998)*x[1]+(1.14312188299)*x[2]
        s[1,1,1,1]=(0.137715540853)*x[0]+(1.42995779339)*x[1]+(-1.10477793041)*x[2]
        s[1,1,2,0]=(-0.0647098241647)*x[0]+(1.18580807605)*x[1]+(0.605076723926)*x[2]
        s[1,1,2,1]=(0.744824848213)*x[0]+(0.532106337173)*x[1]+(0.68490109894)*x[2]
        s[1,2,0,0]=(-0.557916122207)*x[0]+(1.06859221643)*x[1]+(0.823215271018)*x[2]
        s[1,2,0,1]=(0.00598565716481)*x[0]+(0.884417903372)*x[1]+(1.7921906066)*x[2]
        s[1,2,1,0]=(-0.47245962749)*x[0]+(-0.906624987584)*x[1]+(0.137792563282)*x[2]
        s[1,2,1,1]=(0.948613971685)*x[0]+(0.753281322508)*x[1]+(0.110929011008)*x[2]
        s[1,2,2,0]=(0.175178588707)*x[0]+(-0.796494364923)*x[1]+(0.390555269738)*x[2]
        s[1,2,2,1]=(-1.04516910973)*x[0]+(-0.921757404949)*x[1]+(0.845139972079)*x[2]
        s[1,3,0,0]=(-1.42392853478)*x[0]+(-0.458461019373)*x[1]+(-0.213286746949)*x[2]
        s[1,3,0,1]=(0.13996349254)*x[0]+(0.108946526608)*x[1]+(-0.848132079317)*x[2]
        s[1,3,1,0]=(-1.44188153548)*x[0]+(-0.108156975475)*x[1]+(0.0715891132579)*x[2]
        s[1,3,1,1]=(0.525856508836)*x[0]+(0.950609265517)*x[1]+(0.0127362240999)*x[2]
        s[1,3,2,0]=(-1.0845975079)*x[0]+(-1.24949492248)*x[1]+(-0.00240035631582)*x[2]
        s[1,3,2,1]=(0.642060593331)*x[0]+(1.46218158975)*x[1]+(1.32385980887)*x[2]
        s[1,4,0,0]=(0.0119712366759)*x[0]+(1.39772768926)*x[1]+(-0.570649410717)*x[2]
        s[1,4,0,1]=(0.320030773751)*x[0]+(0.358086736384)*x[1]+(-0.859721999492)*x[2]
        s[1,4,1,0]=(0.574417642497)*x[0]+(-0.421844733062)*x[1]+(0.611638265232)*x[2]
        s[1,4,1,1]=(-0.228413169897)*x[0]+(-0.0391189339225)*x[1]+(0.635031723033)*x[2]
        s[1,4,2,0]=(0.0424678701733)*x[0]+(0.261307736177)*x[1]+(-0.133428864005)*x[2]
        s[1,4,2,1]=(0.169996877966)*x[0]+(1.30730453844)*x[1]+(-1.07718562596)*x[2]
        s[2,0,0,0]=(0.147232851651)*x[0]+(-0.134174394382)*x[1]+(-0.785323160299)*x[2]
        s[2,0,0,1]=(1.29417433206)*x[0]+(-1.39241927595)*x[1]+(-0.224395758464)*x[2]
        s[2,0,1,0]=(1.60938474462)*x[0]+(0.68821531646)*x[1]+(0.453462382464)*x[2]
        s[2,0,1,1]=(0.63619844559)*x[0]+(-0.956987170375)*x[1]+(-1.24753100984)*x[2]
        s[2,0,2,0]=(1.14194732577)*x[0]+(0.973489037101)*x[1]+(0.187588261589)*x[2]
        s[2,0,2,1]=(-0.898956363323)*x[0]+(0.108722086141)*x[1]+(-1.62445422766)*x[2]
        s[2,1,0,0]=(0.127827687335)*x[0]+(0.216922403148)*x[1]+(-1.5099539458)*x[2]
        s[2,1,0,1]=(-0.769830743998)*x[0]+(1.46060444448)*x[1]+(-0.318171252653)*x[2]
        s[2,1,1,0]=(-0.311976479418)*x[0]+(-0.949061813767)*x[1]+(-0.237478430906)*x[2]
        s[2,1,1,1]=(1.27107354882)*x[0]+(-0.427981502964)*x[1]+(-0.457224808787)*x[2]
        s[2,1,2,0]=(-0.633029461167)*x[0]+(0.255322271264)*x[1]+(-1.79431610464)*x[2]
        s[2,1,2,1]=(1.05355586142)*x[0]+(-1.69722728443)*x[1]+(-1.11928223816)*x[2]
        s[2,2,0,0]=(1.15538109557)*x[0]+(-0.52692609967)*x[1]+(-0.686266572053)*x[2]
        s[2,2,0,1]=(-0.327531879955)*x[0]+(0.365760847752)*x[1]+(-0.0600793962176)*x[2]
        s[2,2,1,0]=(1.16249877038)*x[0]+(0.710438081472)*x[1]+(0.986719636688)*x[2]
        s[2,2,1,1]=(-1.22799525195)*x[0]+(0.748811223969)*x[1]+(0.370195043623)*x[2]
        s[2,2,2,0]=(0.976246115046)*x[0]+(1.10537245041)*x[1]+(-0.759735725531)*x[2]
        s[2,2,2,1]=(1.01411361404)*x[0]+(-0.000596974234967)*x[1]+(-1.39135228207)*x[2]
        s[2,3,0,0]=(-0.581025188916)*x[0]+(-1.43405762699)*x[1]+(-0.912365741093)*x[2]
        s[2,3,0,1]=(-0.0511110676196)*x[0]+(0.230142796427)*x[1]+(-1.58165960293)*x[2]
        s[2,3,1,0]=(-1.46709370576)*x[0]+(0.425035892802)*x[1]+(0.282281348864)*x[2]
        s[2,3,1,1]=(-0.35514320269)*x[0]+(-0.39238669714)*x[1]+(-0.0118498421145)*x[2]
        s[2,3,2,0]=(0.317197568125)*x[0]+(0.689803280183)*x[1]+(-0.238327580547)*x[2]
        s[2,3,2,1]=(1.31241465464)*x[0]+(0.548355955144)*x[1]+(0.0218217641073)*x[2]
        s[2,4,0,0]=(0.787240379652)*x[0]+(-0.906804255968)*x[1]+(-1.17498478521)*x[2]
        s[2,4,0,1]=(1.19721624988)*x[0]+(0.567119981635)*x[1]+(1.42310267091)*x[2]
        s[2,4,1,0]=(-0.199829256778)*x[0]+(0.28908416123)*x[1]+(0.491497583863)*x[2]
        s[2,4,1,1]=(0.336357763916)*x[0]+(-0.750103385624)*x[1]+(-1.19492486127)*x[2]
        s[2,4,2,0]=(1.509049274)*x[0]+(-1.28641285242)*x[1]+(-0.250003617687)*x[2]
        s[2,4,2,1]=(-1.23803199733)*x[0]+(0.688217821134)*x[1]+(-0.436478772021)*x[2]
        s[3,0,0,0]=(0.634885383655)*x[0]+(1.14728850769)*x[1]+(0.417496697665)*x[2]
        s[3,0,0,1]=(-0.326274842766)*x[0]+(-1.47136469943)*x[1]+(1.87373171786)*x[2]
        s[3,0,1,0]=(-0.140107043914)*x[0]+(-0.261565104318)*x[1]+(0.0780015778888)*x[2]
        s[3,0,1,1]=(0.770979582613)*x[0]+(-0.899853446798)*x[1]+(1.45115909188)*x[2]
        s[3,0,2,0]=(0.477865962499)*x[0]+(-0.0621575016815)*x[1]+(-1.28975967232)*x[2]
        s[3,0,2,1]=(-1.41031619045)*x[0]+(1.50155797656)*x[1]+(-1.44890527708)*x[2]
        s[3,1,0,0]=(-1.04632613119)*x[0]+(0.00196059522552)*x[1]+(0.139147980023)*x[2]
        s[3,1,0,1]=(0.217272687162)*x[0]+(0.295773880887)*x[1]+(0.474090779139)*x[2]
        s[3,1,1,0]=(1.50572386246)*x[0]+(0.868579168785)*x[1]+(-0.12394808575)*x[2]
        s[3,1,1,1]=(0.882176073715)*x[0]+(0.144766914417)*x[1]+(-0.408510755049)*x[2]
        s[3,1,2,0]=(-1.21767372555)*x[0]+(0.195682844047)*x[1]+(-1.70399790168)*x[2]
        s[3,1,2,1]=(-0.110311118233)*x[0]+(-0.721806016748)*x[1]+(1.81495268748)*x[2]
        s[3,2,0,0]=(-0.0441997228291)*x[0]+(0.0754984632459)*x[1]+(0.274177049743)*x[2]
        s[3,2,0,1]=(-0.48089133278)*x[0]+(-0.00712717635695)*x[1]+(0.242678324953)*x[2]
        s[3,2,1,0]=(0.48377042087)*x[0]+(1.4374627607)*x[1]+(-1.63994815623)*x[2]
        s[3,2,1,1]=(0.548308948512)*x[0]+(-0.587878136236)*x[1]+(-0.636242676063)*x[2]
        s[3,2,2,0]=(-1.62147144921)*x[0]+(-0.937713365221)*x[1]+(0.416583670897)*x[2]
        s[3,2,2,1]=(1.50617037978)*x[0]+(0.849964476529)*x[1]+(-0.421030022756)*x[2]
        s[3,3,0,0]=(-0.0258240269876)*x[0]+(1.09488855424)*x[1]+(-0.864078888453)*x[2]
        s[3,3,0,1]=(0.121803754668)*x[0]+(1.47064577586)*x[1]+(-1.15267381807)*x[2]
        s[3,3,1,0]=(0.7847153928)*x[0]+(1.19923938448)*x[1]+(-0.220983796213)*x[2]
        s[3,3,1,1]=(0.591086037588)*x[0]+(1.49019206942)*x[1]+(-0.764272587507)*x[2]
        s[3,3,2,0]=(-0.751721311032)*x[0]+(-1.41725473775)*x[1]+(-0.182681015042)*x[2]
        s[3,3,2,1]=(1.18789607758)*x[0]+(0.321544556274)*x[1]+(-0.37682906633)*x[2]
        s[3,4,0,0]=(-0.314136382522)*x[0]+(0.190342372946)*x[1]+(-0.0583598827644)*x[2]
        s[3,4,0,1]=(1.87879524007)*x[0]+(0.0472715596509)*x[1]+(0.849571713097)*x[2]
        s[3,4,1,0]=(0.816120005278)*x[0]+(0.296980756484)*x[1]+(0.251581386555)*x[2]
        s[3,4,1,1]=(0.12560195013)*x[0]+(-0.293571823053)*x[1]+(-1.82952568911)*x[2]
        s[3,4,2,0]=(1.03923438653)*x[0]+(-0.899000072461)*x[1]+(-1.48431797596)*x[2]
        s[3,4,2,1]=(-1.12443865282)*x[0]+(-0.200378451615)*x[1]+(-1.41505465511)*x[2]
        ref[0,0,0,0]=(0.146894168765)*x_ref[0]+(1.25822855062)*x_ref[1]+(0.088660814219)*x_ref[2]
        ref[0,0,0,1]=(0.572576296123)*x_ref[0]+(0.670191219644)*x_ref[1]+(-0.280619677889)*x_ref[2]
        ref[0,0,1,0]=(-1.19817665348)*x_ref[0]+(1.35545322011)*x_ref[1]+(-0.279751222491)*x_ref[2]
        ref[0,0,1,1]=(-0.0558236682639)*x_ref[0]+(-1.61042107055)*x_ref[1]+(0.622456635293)*x_ref[2]
        ref[0,0,2,0]=(-0.154784205079)*x_ref[0]+(-0.52314996688)*x_ref[1]+(1.0590614932)*x_ref[2]
        ref[0,0,2,1]=(1.58101401758)*x_ref[0]+(-0.610736715272)*x_ref[1]+(-0.487233776225)*x_ref[2]
        ref[0,1,0,0]=(0.0117365392099)*x_ref[0]+(1.02397373706)*x_ref[1]+(0.407377316737)*x_ref[2]
        ref[0,1,0,1]=(0.460387220384)*x_ref[0]+(1.10329384655)*x_ref[1]+(0.252209465365)*x_ref[2]
        ref[0,1,1,0]=(-1.33153199048)*x_ref[0]+(-0.759669253534)*x_ref[1]+(-0.856223614724)*x_ref[2]
        ref[0,1,1,1]=(-1.00599668771)*x_ref[0]+(0.321343926218)*x_ref[1]+(-0.198860330101)*x_ref[2]
        ref[0,1,2,0]=(1.00741183642)*x_ref[0]+(-0.148473398917)*x_ref[1]+(-0.484181353094)*x_ref[2]
        ref[0,1,2,1]=(-1.46490648953)*x_ref[0]+(0.621077170698)*x_ref[1]+(0.802865695373)*x_ref[2]
        ref[0,2,0,0]=(1.18498612655)*x_ref[0]+(-0.511705644809)*x_ref[1]+(0.563568352278)*x_ref[2]
        ref[0,2,0,1]=(-0.807427977876)*x_ref[0]+(-0.252372201752)*x_ref[1]+(0.251722661898)*x_ref[2]
        ref[0,2,1,0]=(0.524572175367)*x_ref[0]+(0.49906104527)*x_ref[1]+(-0.601836374873)*x_ref[2]
        ref[0,2,1,1]=(-0.318976775137)*x_ref[0]+(-0.153065654914)*x_ref[1]+(-0.458888325378)*x_ref[2]
        ref[0,2,2,0]=(-1.45513892332)*x_ref[0]+(-0.252635212625)*x_ref[1]+(0.490153491036)*x_ref[2]
        ref[0,2,2,1]=(-0.439422986461)*x_ref[0]+(-0.937409556421)*x_ref[1]+(0.209106988492)*x_ref[2]
        ref[0,3,0,0]=(-0.26828401981)*x_ref[0]+(1.70808529771)*x_ref[1]+(0.431245149358)*x_ref[2]
        ref[0,3,0,1]=(-0.831846912527)*x_ref[0]+(0.572822507173)*x_ref[1]+(-0.283130753547)*x_ref[2]
        ref[0,3,1,0]=(-0.625617835832)*x_ref[0]+(-1.25495997626)*x_ref[1]+(1.28743784867)*x_ref[2]
        ref[0,3,1,1]=(1.47223047075)*x_ref[0]+(0.274365456026)*x_ref[1]+(1.34355769078)*x_ref[2]
        ref[0,3,2,0]=(-0.0506967319243)*x_ref[0]+(1.16569056569)*x_ref[1]+(-0.965008332019)*x_ref[2]
        ref[0,3,2,1]=(0.0106139754664)*x_ref[0]+(0.274252087281)*x_ref[1]+(-0.501970174754)*x_ref[2]
        ref[0,4,0,0]=(0.717476010799)*x_ref[0]+(0.380533443018)*x_ref[1]+(-0.304127891816)*x_ref[2]
        ref[0,4,0,1]=(-1.29133512255)*x_ref[0]+(-1.16667771199)*x_ref[1]+(0.275809086731)*x_ref[2]
        ref[0,4,1,0]=(0.137939049278)*x_ref[0]+(-0.302950342448)*x_ref[1]+(-0.28865204469)*x_ref[2]
        ref[0,4,1,1]=(-0.857457318663)*x_ref[0]+(0.423143726447)*x_ref[1]+(0.749889416366)*x_ref[2]
        ref[0,4,2,0]=(-0.994428027329)*x_ref[0]+(-0.377780341115)*x_ref[1]+(-0.997480394895)*x_ref[2]
        ref[0,4,2,1]=(-0.681653236311)*x_ref[0]+(-0.091905911233)*x_ref[1]+(-0.877007024631)*x_ref[2]
        ref[1,0,0,0]=(0.33375482228)*x_ref[0]+(-1.08816236114)*x_ref[1]+(0.154701633038)*x_ref[2]
        ref[1,0,0,1]=(0.979227592935)*x_ref[0]+(0.571415419963)*x_ref[1]+(-0.20312466228)*x_ref[2]
        ref[1,0,1,0]=(1.59673013189)*x_ref[0]+(1.01626905603)*x_ref[1]+(0.216806440372)*x_ref[2]
        ref[1,0,1,1]=(-0.276034446694)*x_ref[0]+(-0.168494544051)*x_ref[1]+(0.981220114947)*x_ref[2]
        ref[1,0,2,0]=(0.205085469341)*x_ref[0]+(-0.507970934322)*x_ref[1]+(-0.724745229941)*x_ref[2]
        ref[1,0,2,1]=(-0.629331749902)*x_ref[0]+(-1.27086230989)*x_ref[1]+(-1.08479549542)*x_ref[2]
        ref[1,1,0,0]=(-0.786670678498)*x_ref[0]+(1.24461814343)*x_ref[1]+(0.644579154985)*x_ref[2]
        ref[1,1,0,1]=(1.05414628662)*x_ref[0]+(0.901590532372)*x_ref[1]+(0.809519110188)*x_ref[2]
        ref[1,1,1,0]=(-0.679152609307)*x_ref[0]+(0.84101357998)*x_ref[1]+(1.14312188299)*x_ref[2]
        ref[1,1,1,1]=(0.137715540853)*x_ref[0]+(1.42995779339)*x_ref[1]+(-1.10477793041)*x_ref[2]
        ref[1,1,2,0]=(-0.0647098241647)*x_ref[0]+(1.18580807605)*x_ref[1]+(0.605076723926)*x_ref[2]
        ref[1,1,2,1]=(0.744824848213)*x_ref[0]+(0.532106337173)*x_ref[1]+(0.68490109894)*x_ref[2]
        ref[1,2,0,0]=(-0.557916122207)*x_ref[0]+(1.06859221643)*x_ref[1]+(0.823215271018)*x_ref[2]
        ref[1,2,0,1]=(0.00598565716481)*x_ref[0]+(0.884417903372)*x_ref[1]+(1.7921906066)*x_ref[2]
        ref[1,2,1,0]=(-0.47245962749)*x_ref[0]+(-0.906624987584)*x_ref[1]+(0.137792563282)*x_ref[2]
        ref[1,2,1,1]=(0.948613971685)*x_ref[0]+(0.753281322508)*x_ref[1]+(0.110929011008)*x_ref[2]
        ref[1,2,2,0]=(0.175178588707)*x_ref[0]+(-0.796494364923)*x_ref[1]+(0.390555269738)*x_ref[2]
        ref[1,2,2,1]=(-1.04516910973)*x_ref[0]+(-0.921757404949)*x_ref[1]+(0.845139972079)*x_ref[2]
        ref[1,3,0,0]=(-1.42392853478)*x_ref[0]+(-0.458461019373)*x_ref[1]+(-0.213286746949)*x_ref[2]
        ref[1,3,0,1]=(0.13996349254)*x_ref[0]+(0.108946526608)*x_ref[1]+(-0.848132079317)*x_ref[2]
        ref[1,3,1,0]=(-1.44188153548)*x_ref[0]+(-0.108156975475)*x_ref[1]+(0.0715891132579)*x_ref[2]
        ref[1,3,1,1]=(0.525856508836)*x_ref[0]+(0.950609265517)*x_ref[1]+(0.0127362240999)*x_ref[2]
        ref[1,3,2,0]=(-1.0845975079)*x_ref[0]+(-1.24949492248)*x_ref[1]+(-0.00240035631582)*x_ref[2]
        ref[1,3,2,1]=(0.642060593331)*x_ref[0]+(1.46218158975)*x_ref[1]+(1.32385980887)*x_ref[2]
        ref[1,4,0,0]=(0.0119712366759)*x_ref[0]+(1.39772768926)*x_ref[1]+(-0.570649410717)*x_ref[2]
        ref[1,4,0,1]=(0.320030773751)*x_ref[0]+(0.358086736384)*x_ref[1]+(-0.859721999492)*x_ref[2]
        ref[1,4,1,0]=(0.574417642497)*x_ref[0]+(-0.421844733062)*x_ref[1]+(0.611638265232)*x_ref[2]
        ref[1,4,1,1]=(-0.228413169897)*x_ref[0]+(-0.0391189339225)*x_ref[1]+(0.635031723033)*x_ref[2]
        ref[1,4,2,0]=(0.0424678701733)*x_ref[0]+(0.261307736177)*x_ref[1]+(-0.133428864005)*x_ref[2]
        ref[1,4,2,1]=(0.169996877966)*x_ref[0]+(1.30730453844)*x_ref[1]+(-1.07718562596)*x_ref[2]
        ref[2,0,0,0]=(0.147232851651)*x_ref[0]+(-0.134174394382)*x_ref[1]+(-0.785323160299)*x_ref[2]
        ref[2,0,0,1]=(1.29417433206)*x_ref[0]+(-1.39241927595)*x_ref[1]+(-0.224395758464)*x_ref[2]
        ref[2,0,1,0]=(1.60938474462)*x_ref[0]+(0.68821531646)*x_ref[1]+(0.453462382464)*x_ref[2]
        ref[2,0,1,1]=(0.63619844559)*x_ref[0]+(-0.956987170375)*x_ref[1]+(-1.24753100984)*x_ref[2]
        ref[2,0,2,0]=(1.14194732577)*x_ref[0]+(0.973489037101)*x_ref[1]+(0.187588261589)*x_ref[2]
        ref[2,0,2,1]=(-0.898956363323)*x_ref[0]+(0.108722086141)*x_ref[1]+(-1.62445422766)*x_ref[2]
        ref[2,1,0,0]=(0.127827687335)*x_ref[0]+(0.216922403148)*x_ref[1]+(-1.5099539458)*x_ref[2]
        ref[2,1,0,1]=(-0.769830743998)*x_ref[0]+(1.46060444448)*x_ref[1]+(-0.318171252653)*x_ref[2]
        ref[2,1,1,0]=(-0.311976479418)*x_ref[0]+(-0.949061813767)*x_ref[1]+(-0.237478430906)*x_ref[2]
        ref[2,1,1,1]=(1.27107354882)*x_ref[0]+(-0.427981502964)*x_ref[1]+(-0.457224808787)*x_ref[2]
        ref[2,1,2,0]=(-0.633029461167)*x_ref[0]+(0.255322271264)*x_ref[1]+(-1.79431610464)*x_ref[2]
        ref[2,1,2,1]=(1.05355586142)*x_ref[0]+(-1.69722728443)*x_ref[1]+(-1.11928223816)*x_ref[2]
        ref[2,2,0,0]=(1.15538109557)*x_ref[0]+(-0.52692609967)*x_ref[1]+(-0.686266572053)*x_ref[2]
        ref[2,2,0,1]=(-0.327531879955)*x_ref[0]+(0.365760847752)*x_ref[1]+(-0.0600793962176)*x_ref[2]
        ref[2,2,1,0]=(1.16249877038)*x_ref[0]+(0.710438081472)*x_ref[1]+(0.986719636688)*x_ref[2]
        ref[2,2,1,1]=(-1.22799525195)*x_ref[0]+(0.748811223969)*x_ref[1]+(0.370195043623)*x_ref[2]
        ref[2,2,2,0]=(0.976246115046)*x_ref[0]+(1.10537245041)*x_ref[1]+(-0.759735725531)*x_ref[2]
        ref[2,2,2,1]=(1.01411361404)*x_ref[0]+(-0.000596974234967)*x_ref[1]+(-1.39135228207)*x_ref[2]
        ref[2,3,0,0]=(-0.581025188916)*x_ref[0]+(-1.43405762699)*x_ref[1]+(-0.912365741093)*x_ref[2]
        ref[2,3,0,1]=(-0.0511110676196)*x_ref[0]+(0.230142796427)*x_ref[1]+(-1.58165960293)*x_ref[2]
        ref[2,3,1,0]=(-1.46709370576)*x_ref[0]+(0.425035892802)*x_ref[1]+(0.282281348864)*x_ref[2]
        ref[2,3,1,1]=(-0.35514320269)*x_ref[0]+(-0.39238669714)*x_ref[1]+(-0.0118498421145)*x_ref[2]
        ref[2,3,2,0]=(0.317197568125)*x_ref[0]+(0.689803280183)*x_ref[1]+(-0.238327580547)*x_ref[2]
        ref[2,3,2,1]=(1.31241465464)*x_ref[0]+(0.548355955144)*x_ref[1]+(0.0218217641073)*x_ref[2]
        ref[2,4,0,0]=(0.787240379652)*x_ref[0]+(-0.906804255968)*x_ref[1]+(-1.17498478521)*x_ref[2]
        ref[2,4,0,1]=(1.19721624988)*x_ref[0]+(0.567119981635)*x_ref[1]+(1.42310267091)*x_ref[2]
        ref[2,4,1,0]=(-0.199829256778)*x_ref[0]+(0.28908416123)*x_ref[1]+(0.491497583863)*x_ref[2]
        ref[2,4,1,1]=(0.336357763916)*x_ref[0]+(-0.750103385624)*x_ref[1]+(-1.19492486127)*x_ref[2]
        ref[2,4,2,0]=(1.509049274)*x_ref[0]+(-1.28641285242)*x_ref[1]+(-0.250003617687)*x_ref[2]
        ref[2,4,2,1]=(-1.23803199733)*x_ref[0]+(0.688217821134)*x_ref[1]+(-0.436478772021)*x_ref[2]
        ref[3,0,0,0]=(0.634885383655)*x_ref[0]+(1.14728850769)*x_ref[1]+(0.417496697665)*x_ref[2]
        ref[3,0,0,1]=(-0.326274842766)*x_ref[0]+(-1.47136469943)*x_ref[1]+(1.87373171786)*x_ref[2]
        ref[3,0,1,0]=(-0.140107043914)*x_ref[0]+(-0.261565104318)*x_ref[1]+(0.0780015778888)*x_ref[2]
        ref[3,0,1,1]=(0.770979582613)*x_ref[0]+(-0.899853446798)*x_ref[1]+(1.45115909188)*x_ref[2]
        ref[3,0,2,0]=(0.477865962499)*x_ref[0]+(-0.0621575016815)*x_ref[1]+(-1.28975967232)*x_ref[2]
        ref[3,0,2,1]=(-1.41031619045)*x_ref[0]+(1.50155797656)*x_ref[1]+(-1.44890527708)*x_ref[2]
        ref[3,1,0,0]=(-1.04632613119)*x_ref[0]+(0.00196059522552)*x_ref[1]+(0.139147980023)*x_ref[2]
        ref[3,1,0,1]=(0.217272687162)*x_ref[0]+(0.295773880887)*x_ref[1]+(0.474090779139)*x_ref[2]
        ref[3,1,1,0]=(1.50572386246)*x_ref[0]+(0.868579168785)*x_ref[1]+(-0.12394808575)*x_ref[2]
        ref[3,1,1,1]=(0.882176073715)*x_ref[0]+(0.144766914417)*x_ref[1]+(-0.408510755049)*x_ref[2]
        ref[3,1,2,0]=(-1.21767372555)*x_ref[0]+(0.195682844047)*x_ref[1]+(-1.70399790168)*x_ref[2]
        ref[3,1,2,1]=(-0.110311118233)*x_ref[0]+(-0.721806016748)*x_ref[1]+(1.81495268748)*x_ref[2]
        ref[3,2,0,0]=(-0.0441997228291)*x_ref[0]+(0.0754984632459)*x_ref[1]+(0.274177049743)*x_ref[2]
        ref[3,2,0,1]=(-0.48089133278)*x_ref[0]+(-0.00712717635695)*x_ref[1]+(0.242678324953)*x_ref[2]
        ref[3,2,1,0]=(0.48377042087)*x_ref[0]+(1.4374627607)*x_ref[1]+(-1.63994815623)*x_ref[2]
        ref[3,2,1,1]=(0.548308948512)*x_ref[0]+(-0.587878136236)*x_ref[1]+(-0.636242676063)*x_ref[2]
        ref[3,2,2,0]=(-1.62147144921)*x_ref[0]+(-0.937713365221)*x_ref[1]+(0.416583670897)*x_ref[2]
        ref[3,2,2,1]=(1.50617037978)*x_ref[0]+(0.849964476529)*x_ref[1]+(-0.421030022756)*x_ref[2]
        ref[3,3,0,0]=(-0.0258240269876)*x_ref[0]+(1.09488855424)*x_ref[1]+(-0.864078888453)*x_ref[2]
        ref[3,3,0,1]=(0.121803754668)*x_ref[0]+(1.47064577586)*x_ref[1]+(-1.15267381807)*x_ref[2]
        ref[3,3,1,0]=(0.7847153928)*x_ref[0]+(1.19923938448)*x_ref[1]+(-0.220983796213)*x_ref[2]
        ref[3,3,1,1]=(0.591086037588)*x_ref[0]+(1.49019206942)*x_ref[1]+(-0.764272587507)*x_ref[2]
        ref[3,3,2,0]=(-0.751721311032)*x_ref[0]+(-1.41725473775)*x_ref[1]+(-0.182681015042)*x_ref[2]
        ref[3,3,2,1]=(1.18789607758)*x_ref[0]+(0.321544556274)*x_ref[1]+(-0.37682906633)*x_ref[2]
        ref[3,4,0,0]=(-0.314136382522)*x_ref[0]+(0.190342372946)*x_ref[1]+(-0.0583598827644)*x_ref[2]
        ref[3,4,0,1]=(1.87879524007)*x_ref[0]+(0.0472715596509)*x_ref[1]+(0.849571713097)*x_ref[2]
        ref[3,4,1,0]=(0.816120005278)*x_ref[0]+(0.296980756484)*x_ref[1]+(0.251581386555)*x_ref[2]
        ref[3,4,1,1]=(0.12560195013)*x_ref[0]+(-0.293571823053)*x_ref[1]+(-1.82952568911)*x_ref[2]
        ref[3,4,2,0]=(1.03923438653)*x_ref[0]+(-0.899000072461)*x_ref[1]+(-1.48431797596)*x_ref[2]
        ref[3,4,2,1]=(-1.12443865282)*x_ref[0]+(-0.200378451615)*x_ref[1]+(-1.41505465511)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.32868090106)*x[0]+(-0.840160165417)*x[1]
        ref=(-0.32868090106)*x_ref[0]+(-0.840160165417)*x_ref[1]
      else:
        s=(-1.32732649563)*x[0]+(0.20149624757)*x[1]+(0.701715136384)*x[2]
        ref=(-1.32732649563)*x_ref[0]+(0.20149624757)*x_ref[1]+(0.701715136384)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        s[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
        ref[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        ref[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
      else:
        s[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        s[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
        ref[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        ref[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.199811526227)*x[0]+(1.25013707432)*x[1]
        s[0,1]=(0.277539536007)*x[0]+(-0.023732930558)*x[1]
        s[0,2]=(-0.244973175438)*x[0]+(-0.501902270422)*x[1]
        s[0,3]=(0.249200744572)*x[0]+(0.520843463388)*x[1]
        s[0,4]=(-0.995168115444)*x[0]+(-0.259093571207)*x[1]
        s[1,0]=(0.179469283663)*x[0]+(-0.461808204127)*x[1]
        s[1,1]=(0.431508044494)*x[0]+(-0.423277002443)*x[1]
        s[1,2]=(0.732281368811)*x[0]+(-0.490387490227)*x[1]
        s[1,3]=(-0.665998043548)*x[0]+(-0.180996501441)*x[1]
        s[1,4]=(-0.773309943626)*x[0]+(-0.27171870837)*x[1]
        s[2,0]=(0.975757439909)*x[0]+(-0.980217145831)*x[1]
        s[2,1]=(-0.898690989301)*x[0]+(0.696886429725)*x[1]
        s[2,2]=(-0.568728495444)*x[0]+(0.171459808085)*x[1]
        s[2,3]=(-0.509734570618)*x[0]+(-0.318239573518)*x[1]
        s[2,4]=(-0.868166262762)*x[0]+(0.669164551211)*x[1]
        s[3,0]=(-1.11259074191)*x[0]+(-0.317962607489)*x[1]
        s[3,1]=(-0.735838783493)*x[0]+(-0.15621592888)*x[1]
        s[3,2]=(-0.348618152308)*x[0]+(0.346356772969)*x[1]
        s[3,3]=(1.58890219057)*x[0]+(-0.106915533345)*x[1]
        s[3,4]=(-0.739997195498)*x[0]+(-0.668986698058)*x[1]
        ref[0,0]=(0.199811526227)*x_ref[0]+(1.25013707432)*x_ref[1]
        ref[0,1]=(0.277539536007)*x_ref[0]+(-0.023732930558)*x_ref[1]
        ref[0,2]=(-0.244973175438)*x_ref[0]+(-0.501902270422)*x_ref[1]
        ref[0,3]=(0.249200744572)*x_ref[0]+(0.520843463388)*x_ref[1]
        ref[0,4]=(-0.995168115444)*x_ref[0]+(-0.259093571207)*x_ref[1]
        ref[1,0]=(0.179469283663)*x_ref[0]+(-0.461808204127)*x_ref[1]
        ref[1,1]=(0.431508044494)*x_ref[0]+(-0.423277002443)*x_ref[1]
        ref[1,2]=(0.732281368811)*x_ref[0]+(-0.490387490227)*x_ref[1]
        ref[1,3]=(-0.665998043548)*x_ref[0]+(-0.180996501441)*x_ref[1]
        ref[1,4]=(-0.773309943626)*x_ref[0]+(-0.27171870837)*x_ref[1]
        ref[2,0]=(0.975757439909)*x_ref[0]+(-0.980217145831)*x_ref[1]
        ref[2,1]=(-0.898690989301)*x_ref[0]+(0.696886429725)*x_ref[1]
        ref[2,2]=(-0.568728495444)*x_ref[0]+(0.171459808085)*x_ref[1]
        ref[2,3]=(-0.509734570618)*x_ref[0]+(-0.318239573518)*x_ref[1]
        ref[2,4]=(-0.868166262762)*x_ref[0]+(0.669164551211)*x_ref[1]
        ref[3,0]=(-1.11259074191)*x_ref[0]+(-0.317962607489)*x_ref[1]
        ref[3,1]=(-0.735838783493)*x_ref[0]+(-0.15621592888)*x_ref[1]
        ref[3,2]=(-0.348618152308)*x_ref[0]+(0.346356772969)*x_ref[1]
        ref[3,3]=(1.58890219057)*x_ref[0]+(-0.106915533345)*x_ref[1]
        ref[3,4]=(-0.739997195498)*x_ref[0]+(-0.668986698058)*x_ref[1]
      else:
        s[0,0]=(1.00508479945)*x[0]+(0.671813078579)*x[1]+(0.0352167301347)*x[2]
        s[0,1]=(0.832418208523)*x[0]+(-0.228971976603)*x[1]+(0.472369173779)*x[2]
        s[0,2]=(0.495212675956)*x[0]+(-0.850840698084)*x[1]+(-0.480464122957)*x[2]
        s[0,3]=(1.09800821588)*x[0]+(-1.16790708068)*x[1]+(-0.16448878954)*x[2]
        s[0,4]=(-0.142901980481)*x[0]+(-0.456358812317)*x[1]+(0.688171420193)*x[2]
        s[1,0]=(-0.452209235863)*x[0]+(0.835019801073)*x[1]+(-0.537900952591)*x[2]
        s[1,1]=(1.24188029232)*x[0]+(0.537899388384)*x[1]+(0.825366946283)*x[2]
        s[1,2]=(0.0568959695268)*x[0]+(-0.530615837561)*x[1]+(0.208390615403)*x[2]
        s[1,3]=(0.704415982126)*x[0]+(0.557947101676)*x[1]+(-1.41348269833)*x[2]
        s[1,4]=(0.23735208213)*x[0]+(0.188600394593)*x[1]+(-0.587309525946)*x[2]
        s[2,0]=(0.202262121276)*x[0]+(1.34583368863)*x[1]+(-1.416408583)*x[2]
        s[2,1]=(-0.889020760737)*x[0]+(1.18733643427)*x[1]+(1.11796325716)*x[2]
        s[2,2]=(0.0729529752138)*x[0]+(0.467461724385)*x[1]+(-1.26467465607)*x[2]
        s[2,3]=(0.553563173427)*x[0]+(-0.96991696383)*x[1]+(1.21489979267)*x[2]
        s[2,4]=(0.504157770518)*x[0]+(-0.0966112701707)*x[1]+(0.820591191307)*x[2]
        s[3,0]=(0.438942088634)*x[0]+(0.737754158644)*x[1]+(0.327747630162)*x[2]
        s[3,1]=(-0.966999874402)*x[0]+(0.218515238902)*x[1]+(-1.85502435057)*x[2]
        s[3,2]=(0.632366690747)*x[0]+(0.710212883666)*x[1]+(1.52843521678)*x[2]
        s[3,3]=(-0.139393405525)*x[0]+(-1.020102199)*x[1]+(0.485417385052)*x[2]
        s[3,4]=(-0.753676205348)*x[0]+(0.783454715482)*x[1]+(0.568363667465)*x[2]
        ref[0,0]=(1.00508479945)*x_ref[0]+(0.671813078579)*x_ref[1]+(0.0352167301347)*x_ref[2]
        ref[0,1]=(0.832418208523)*x_ref[0]+(-0.228971976603)*x_ref[1]+(0.472369173779)*x_ref[2]
        ref[0,2]=(0.495212675956)*x_ref[0]+(-0.850840698084)*x_ref[1]+(-0.480464122957)*x_ref[2]
        ref[0,3]=(1.09800821588)*x_ref[0]+(-1.16790708068)*x_ref[1]+(-0.16448878954)*x_ref[2]
        ref[0,4]=(-0.142901980481)*x_ref[0]+(-0.456358812317)*x_ref[1]+(0.688171420193)*x_ref[2]
        ref[1,0]=(-0.452209235863)*x_ref[0]+(0.835019801073)*x_ref[1]+(-0.537900952591)*x_ref[2]
        ref[1,1]=(1.24188029232)*x_ref[0]+(0.537899388384)*x_ref[1]+(0.825366946283)*x_ref[2]
        ref[1,2]=(0.0568959695268)*x_ref[0]+(-0.530615837561)*x_ref[1]+(0.208390615403)*x_ref[2]
        ref[1,3]=(0.704415982126)*x_ref[0]+(0.557947101676)*x_ref[1]+(-1.41348269833)*x_ref[2]
        ref[1,4]=(0.23735208213)*x_ref[0]+(0.188600394593)*x_ref[1]+(-0.587309525946)*x_ref[2]
        ref[2,0]=(0.202262121276)*x_ref[0]+(1.34583368863)*x_ref[1]+(-1.416408583)*x_ref[2]
        ref[2,1]=(-0.889020760737)*x_ref[0]+(1.18733643427)*x_ref[1]+(1.11796325716)*x_ref[2]
        ref[2,2]=(0.0729529752138)*x_ref[0]+(0.467461724385)*x_ref[1]+(-1.26467465607)*x_ref[2]
        ref[2,3]=(0.553563173427)*x_ref[0]+(-0.96991696383)*x_ref[1]+(1.21489979267)*x_ref[2]
        ref[2,4]=(0.504157770518)*x_ref[0]+(-0.0966112701707)*x_ref[1]+(0.820591191307)*x_ref[2]
        ref[3,0]=(0.438942088634)*x_ref[0]+(0.737754158644)*x_ref[1]+(0.327747630162)*x_ref[2]
        ref[3,1]=(-0.966999874402)*x_ref[0]+(0.218515238902)*x_ref[1]+(-1.85502435057)*x_ref[2]
        ref[3,2]=(0.632366690747)*x_ref[0]+(0.710212883666)*x_ref[1]+(1.52843521678)*x_ref[2]
        ref[3,3]=(-0.139393405525)*x_ref[0]+(-1.020102199)*x_ref[1]+(0.485417385052)*x_ref[2]
        ref[3,4]=(-0.753676205348)*x_ref[0]+(0.783454715482)*x_ref[1]+(0.568363667465)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.734232753652)*x[0]+(0.800930015514)*x[1]
        s[0,0,1]=(-0.219864689475)*x[0]+(0.157713027162)*x[1]
        s[0,1,0]=(-0.00664406256638)*x[0]+(0.872104848272)*x[1]
        s[0,1,1]=(-1.5575275283)*x[0]+(-1.43701883839)*x[1]
        s[1,0,0]=(1.60056664833)*x[0]+(1.6215424255)*x[1]
        s[1,0,1]=(1.5172017347)*x[0]+(-0.022919888148)*x[1]
        s[1,1,0]=(0.042690381655)*x[0]+(1.30019922)*x[1]
        s[1,1,1]=(0.105731664319)*x[0]+(-1.87027128366)*x[1]
        s[2,0,0]=(1.59675688345)*x[0]+(0.0608090042054)*x[1]
        s[2,0,1]=(-1.3520682478)*x[0]+(-0.881995801453)*x[1]
        s[2,1,0]=(1.08974009958)*x[0]+(0.13217220748)*x[1]
        s[2,1,1]=(-0.631056656103)*x[0]+(-0.548103546753)*x[1]
        s[3,0,0]=(1.59082745694)*x[0]+(0.101815532645)*x[1]
        s[3,0,1]=(-0.390741362781)*x[0]+(0.525254974689)*x[1]
        s[3,1,0]=(0.439652119858)*x[0]+(1.25127795686)*x[1]
        s[3,1,1]=(-1.10496039621)*x[0]+(0.0889406484341)*x[1]
        s[4,0,0]=(1.76164121088)*x[0]+(0.790919865199)*x[1]
        s[4,0,1]=(1.53483109)*x[0]+(1.15931695117)*x[1]
        s[4,1,0]=(-1.4226648704)*x[0]+(0.0380245720655)*x[1]
        s[4,1,1]=(-0.158958015165)*x[0]+(0.359005195468)*x[1]
        s[5,0,0]=(-0.0531318028725)*x[0]+(-0.0716200483232)*x[1]
        s[5,0,1]=(-1.64720026194)*x[0]+(-1.13699910338)*x[1]
        s[5,1,0]=(1.79036772333)*x[0]+(0.414445739998)*x[1]
        s[5,1,1]=(-0.168931889239)*x[0]+(1.55913218176)*x[1]
        ref[0,0,0]=(-0.734232753652)*x_ref[0]+(0.800930015514)*x_ref[1]
        ref[0,0,1]=(-0.219864689475)*x_ref[0]+(0.157713027162)*x_ref[1]
        ref[0,1,0]=(-0.00664406256638)*x_ref[0]+(0.872104848272)*x_ref[1]
        ref[0,1,1]=(-1.5575275283)*x_ref[0]+(-1.43701883839)*x_ref[1]
        ref[1,0,0]=(1.60056664833)*x_ref[0]+(1.6215424255)*x_ref[1]
        ref[1,0,1]=(1.5172017347)*x_ref[0]+(-0.022919888148)*x_ref[1]
        ref[1,1,0]=(0.042690381655)*x_ref[0]+(1.30019922)*x_ref[1]
        ref[1,1,1]=(0.105731664319)*x_ref[0]+(-1.87027128366)*x_ref[1]
        ref[2,0,0]=(1.59675688345)*x_ref[0]+(0.0608090042054)*x_ref[1]
        ref[2,0,1]=(-1.3520682478)*x_ref[0]+(-0.881995801453)*x_ref[1]
        ref[2,1,0]=(1.08974009958)*x_ref[0]+(0.13217220748)*x_ref[1]
        ref[2,1,1]=(-0.631056656103)*x_ref[0]+(-0.548103546753)*x_ref[1]
        ref[3,0,0]=(1.59082745694)*x_ref[0]+(0.101815532645)*x_ref[1]
        ref[3,0,1]=(-0.390741362781)*x_ref[0]+(0.525254974689)*x_ref[1]
        ref[3,1,0]=(0.439652119858)*x_ref[0]+(1.25127795686)*x_ref[1]
        ref[3,1,1]=(-1.10496039621)*x_ref[0]+(0.0889406484341)*x_ref[1]
        ref[4,0,0]=(1.76164121088)*x_ref[0]+(0.790919865199)*x_ref[1]
        ref[4,0,1]=(1.53483109)*x_ref[0]+(1.15931695117)*x_ref[1]
        ref[4,1,0]=(-1.4226648704)*x_ref[0]+(0.0380245720655)*x_ref[1]
        ref[4,1,1]=(-0.158958015165)*x_ref[0]+(0.359005195468)*x_ref[1]
        ref[5,0,0]=(-0.0531318028725)*x_ref[0]+(-0.0716200483232)*x_ref[1]
        ref[5,0,1]=(-1.64720026194)*x_ref[0]+(-1.13699910338)*x_ref[1]
        ref[5,1,0]=(1.79036772333)*x_ref[0]+(0.414445739998)*x_ref[1]
        ref[5,1,1]=(-0.168931889239)*x_ref[0]+(1.55913218176)*x_ref[1]
      else:
        s[0,0,0]=(0.030748782631)*x[0]+(0.0965401372343)*x[1]+(-1.19722517556)*x[2]
        s[0,0,1]=(-1.25981749671)*x[0]+(0.375214736326)*x[1]+(0.78094551805)*x[2]
        s[0,1,0]=(0.0859894648882)*x[0]+(-1.02857522767)*x[1]+(0.752176468687)*x[2]
        s[0,1,1]=(0.857538315926)*x[0]+(1.20788107416)*x[1]+(-1.00656644081)*x[2]
        s[1,0,0]=(-0.0208770456539)*x[0]+(0.296657331936)*x[1]+(1.49455132218)*x[2]
        s[1,0,1]=(-1.1213775487)*x[0]+(1.12351512934)*x[1]+(0.661972762047)*x[2]
        s[1,1,0]=(0.75336533272)*x[0]+(1.33059115242)*x[1]+(-0.43556744027)*x[2]
        s[1,1,1]=(0.467898749534)*x[0]+(1.53277916261)*x[1]+(-1.36041296872)*x[2]
        s[2,0,0]=(0.390444800695)*x[0]+(-0.0902034979398)*x[1]+(1.79547473252)*x[2]
        s[2,0,1]=(0.0488678241039)*x[0]+(0.000207248527408)*x[1]+(-0.231246108981)*x[2]
        s[2,1,0]=(0.907328661278)*x[0]+(-0.777592463716)*x[1]+(-0.949764289407)*x[2]
        s[2,1,1]=(-0.508314624748)*x[0]+(-0.0150508206233)*x[1]+(0.786266513715)*x[2]
        s[3,0,0]=(-0.264992651778)*x[0]+(1.23980464726)*x[1]+(0.480251929476)*x[2]
        s[3,0,1]=(-0.519373168965)*x[0]+(-0.091129773203)*x[1]+(-1.12961715008)*x[2]
        s[3,1,0]=(0.737417275735)*x[0]+(0.00073881678009)*x[1]+(-0.913229240795)*x[2]
        s[3,1,1]=(-0.565248614525)*x[0]+(0.232901289197)*x[1]+(-0.852815115095)*x[2]
        s[4,0,0]=(-0.513590194423)*x[0]+(-1.10607554932)*x[1]+(1.3269295152)*x[2]
        s[4,0,1]=(-0.0108694161983)*x[0]+(-1.39154818645)*x[1]+(1.5448779549)*x[2]
        s[4,1,0]=(1.25248373617)*x[0]+(-0.444522495013)*x[1]+(0.791060985516)*x[2]
        s[4,1,1]=(0.287880876137)*x[0]+(1.42265824896)*x[1]+(-1.16312992573)*x[2]
        s[5,0,0]=(-1.25472591758)*x[0]+(-1.48986095574)*x[1]+(1.49212797538)*x[2]
        s[5,0,1]=(-0.774594904232)*x[0]+(0.582557149894)*x[1]+(-0.460088096141)*x[2]
        s[5,1,0]=(0.404584427176)*x[0]+(-0.0808051689456)*x[1]+(0.522819318812)*x[2]
        s[5,1,1]=(1.15513940465)*x[0]+(1.09285861353)*x[1]+(-0.166414260437)*x[2]
        ref[0,0,0]=(0.030748782631)*x_ref[0]+(0.0965401372343)*x_ref[1]+(-1.19722517556)*x_ref[2]
        ref[0,0,1]=(-1.25981749671)*x_ref[0]+(0.375214736326)*x_ref[1]+(0.78094551805)*x_ref[2]
        ref[0,1,0]=(0.0859894648882)*x_ref[0]+(-1.02857522767)*x_ref[1]+(0.752176468687)*x_ref[2]
        ref[0,1,1]=(0.857538315926)*x_ref[0]+(1.20788107416)*x_ref[1]+(-1.00656644081)*x_ref[2]
        ref[1,0,0]=(-0.0208770456539)*x_ref[0]+(0.296657331936)*x_ref[1]+(1.49455132218)*x_ref[2]
        ref[1,0,1]=(-1.1213775487)*x_ref[0]+(1.12351512934)*x_ref[1]+(0.661972762047)*x_ref[2]
        ref[1,1,0]=(0.75336533272)*x_ref[0]+(1.33059115242)*x_ref[1]+(-0.43556744027)*x_ref[2]
        ref[1,1,1]=(0.467898749534)*x_ref[0]+(1.53277916261)*x_ref[1]+(-1.36041296872)*x_ref[2]
        ref[2,0,0]=(0.390444800695)*x_ref[0]+(-0.0902034979398)*x_ref[1]+(1.79547473252)*x_ref[2]
        ref[2,0,1]=(0.0488678241039)*x_ref[0]+(0.000207248527408)*x_ref[1]+(-0.231246108981)*x_ref[2]
        ref[2,1,0]=(0.907328661278)*x_ref[0]+(-0.777592463716)*x_ref[1]+(-0.949764289407)*x_ref[2]
        ref[2,1,1]=(-0.508314624748)*x_ref[0]+(-0.0150508206233)*x_ref[1]+(0.786266513715)*x_ref[2]
        ref[3,0,0]=(-0.264992651778)*x_ref[0]+(1.23980464726)*x_ref[1]+(0.480251929476)*x_ref[2]
        ref[3,0,1]=(-0.519373168965)*x_ref[0]+(-0.091129773203)*x_ref[1]+(-1.12961715008)*x_ref[2]
        ref[3,1,0]=(0.737417275735)*x_ref[0]+(0.00073881678009)*x_ref[1]+(-0.913229240795)*x_ref[2]
        ref[3,1,1]=(-0.565248614525)*x_ref[0]+(0.232901289197)*x_ref[1]+(-0.852815115095)*x_ref[2]
        ref[4,0,0]=(-0.513590194423)*x_ref[0]+(-1.10607554932)*x_ref[1]+(1.3269295152)*x_ref[2]
        ref[4,0,1]=(-0.0108694161983)*x_ref[0]+(-1.39154818645)*x_ref[1]+(1.5448779549)*x_ref[2]
        ref[4,1,0]=(1.25248373617)*x_ref[0]+(-0.444522495013)*x_ref[1]+(0.791060985516)*x_ref[2]
        ref[4,1,1]=(0.287880876137)*x_ref[0]+(1.42265824896)*x_ref[1]+(-1.16312992573)*x_ref[2]
        ref[5,0,0]=(-1.25472591758)*x_ref[0]+(-1.48986095574)*x_ref[1]+(1.49212797538)*x_ref[2]
        ref[5,0,1]=(-0.774594904232)*x_ref[0]+(0.582557149894)*x_ref[1]+(-0.460088096141)*x_ref[2]
        ref[5,1,0]=(0.404584427176)*x_ref[0]+(-0.0808051689456)*x_ref[1]+(0.522819318812)*x_ref[2]
        ref[5,1,1]=(1.15513940465)*x_ref[0]+(1.09285861353)*x_ref[1]+(-0.166414260437)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-1.26340514811)*x[0]+(0.779300786262)*x[1]
        s[0,0,0,1]=(0.184428277326)*x[0]+(-0.831497209177)*x[1]
        s[0,0,1,0]=(1.35946842566)*x[0]+(0.273881497376)*x[1]
        s[0,0,1,1]=(0.117792005007)*x[0]+(1.06185199224)*x[1]
        s[0,0,2,0]=(0.810342149523)*x[0]+(-0.271662508688)*x[1]
        s[0,0,2,1]=(-0.332968381669)*x[0]+(0.0660545206117)*x[1]
        s[0,1,0,0]=(0.413613455431)*x[0]+(0.641032202343)*x[1]
        s[0,1,0,1]=(0.773447709702)*x[0]+(0.48460517354)*x[1]
        s[0,1,1,0]=(0.173632755729)*x[0]+(0.0635245280713)*x[1]
        s[0,1,1,1]=(0.731914799404)*x[0]+(1.42146728852)*x[1]
        s[0,1,2,0]=(0.688894013616)*x[0]+(1.0339010436)*x[1]
        s[0,1,2,1]=(-0.534272384342)*x[0]+(0.186670150961)*x[1]
        s[0,2,0,0]=(-0.318755700985)*x[0]+(-1.35142313256)*x[1]
        s[0,2,0,1]=(0.973640957973)*x[0]+(-0.103253061427)*x[1]
        s[0,2,1,0]=(1.14046612599)*x[0]+(0.238774677715)*x[1]
        s[0,2,1,1]=(-0.383452508275)*x[0]+(-1.62240702377)*x[1]
        s[0,2,2,0]=(0.761272968366)*x[0]+(1.446646048)*x[1]
        s[0,2,2,1]=(0.183707769875)*x[0]+(-0.630747785283)*x[1]
        s[0,3,0,0]=(-0.474486603803)*x[0]+(-0.0855459523088)*x[1]
        s[0,3,0,1]=(-0.991769725003)*x[0]+(1.004616452)*x[1]
        s[0,3,1,0]=(0.261836325547)*x[0]+(0.353357820559)*x[1]
        s[0,3,1,1]=(0.808880823505)*x[0]+(-0.495896346674)*x[1]
        s[0,3,2,0]=(1.06375339919)*x[0]+(0.00408804480046)*x[1]
        s[0,3,2,1]=(-1.32927483163)*x[0]+(-0.0833136005835)*x[1]
        s[0,4,0,0]=(1.13794063387)*x[0]+(-1.1458884083)*x[1]
        s[0,4,0,1]=(0.73883590181)*x[0]+(1.45878975539)*x[1]
        s[0,4,1,0]=(0.428378040368)*x[0]+(-1.25324354587)*x[1]
        s[0,4,1,1]=(1.19772514118)*x[0]+(0.457504803527)*x[1]
        s[0,4,2,0]=(-0.646891239861)*x[0]+(1.14291257798)*x[1]
        s[0,4,2,1]=(-0.579096313066)*x[0]+(-0.0276859933543)*x[1]
        s[1,0,0,0]=(-0.48935920813)*x[0]+(-0.327683586833)*x[1]
        s[1,0,0,1]=(0.59804841111)*x[0]+(-0.452980496343)*x[1]
        s[1,0,1,0]=(0.608017924036)*x[0]+(0.561553199982)*x[1]
        s[1,0,1,1]=(-0.00361014464582)*x[0]+(0.151090357053)*x[1]
        s[1,0,2,0]=(1.24385396019)*x[0]+(-0.0213646134992)*x[1]
        s[1,0,2,1]=(-0.219031743102)*x[0]+(-0.402476762676)*x[1]
        s[1,1,0,0]=(-0.196889997171)*x[0]+(0.675626499802)*x[1]
        s[1,1,0,1]=(1.23421246519)*x[0]+(-1.72367262236)*x[1]
        s[1,1,1,0]=(1.35571029115)*x[0]+(0.506057274132)*x[1]
        s[1,1,1,1]=(0.310699908104)*x[0]+(-0.292826394288)*x[1]
        s[1,1,2,0]=(0.122289986824)*x[0]+(-0.15947928964)*x[1]
        s[1,1,2,1]=(0.662808118991)*x[0]+(-0.927311967494)*x[1]
        s[1,2,0,0]=(-0.215898581104)*x[0]+(0.32902895413)*x[1]
        s[1,2,0,1]=(0.393145660435)*x[0]+(1.49040999822)*x[1]
        s[1,2,1,0]=(0.904144553172)*x[0]+(0.199605443931)*x[1]
        s[1,2,1,1]=(0.535979155678)*x[0]+(1.78418572873)*x[1]
        s[1,2,2,0]=(1.15578250862)*x[0]+(0.167811579149)*x[1]
        s[1,2,2,1]=(-0.0844120860286)*x[0]+(-0.56656817877)*x[1]
        s[1,3,0,0]=(0.760581325264)*x[0]+(1.00766443232)*x[1]
        s[1,3,0,1]=(-0.199244602194)*x[0]+(-1.54110932148)*x[1]
        s[1,3,1,0]=(1.40934237781)*x[0]+(0.700934912313)*x[1]
        s[1,3,1,1]=(-0.702507416172)*x[0]+(-0.206530323456)*x[1]
        s[1,3,2,0]=(-0.234301811191)*x[0]+(1.41101365089)*x[1]
        s[1,3,2,1]=(1.19609435879)*x[0]+(0.295560183318)*x[1]
        s[1,4,0,0]=(1.15287205838)*x[0]+(0.630564383314)*x[1]
        s[1,4,0,1]=(-0.993907995704)*x[0]+(-0.30836527499)*x[1]
        s[1,4,1,0]=(-0.265158946854)*x[0]+(0.689120014144)*x[1]
        s[1,4,1,1]=(0.503151681106)*x[0]+(-0.928730587416)*x[1]
        s[1,4,2,0]=(-1.0599860135)*x[0]+(-0.696932543908)*x[1]
        s[1,4,2,1]=(-0.365096260865)*x[0]+(0.443129268519)*x[1]
        s[2,0,0,0]=(0.412604298623)*x[0]+(0.591441458509)*x[1]
        s[2,0,0,1]=(0.536034627643)*x[0]+(-0.735261974283)*x[1]
        s[2,0,1,0]=(-0.336900405282)*x[0]+(-1.38493113859)*x[1]
        s[2,0,1,1]=(-0.50634870942)*x[0]+(0.0376544898311)*x[1]
        s[2,0,2,0]=(-0.0595134635196)*x[0]+(0.067285997446)*x[1]
        s[2,0,2,1]=(0.308894832145)*x[0]+(-1.94600763342)*x[1]
        s[2,1,0,0]=(0.443019228067)*x[0]+(-0.68858070865)*x[1]
        s[2,1,0,1]=(1.4461941226)*x[0]+(0.711186579745)*x[1]
        s[2,1,1,0]=(0.178777491574)*x[0]+(-0.746259056845)*x[1]
        s[2,1,1,1]=(1.24682492601)*x[0]+(-0.987057469956)*x[1]
        s[2,1,2,0]=(1.1863711951)*x[0]+(-0.853660028797)*x[1]
        s[2,1,2,1]=(1.13176028078)*x[0]+(-0.887521745194)*x[1]
        s[2,2,0,0]=(0.37579402707)*x[0]+(-0.462533529495)*x[1]
        s[2,2,0,1]=(-1.27778045393)*x[0]+(0.383310656984)*x[1]
        s[2,2,1,0]=(-0.0871651954146)*x[0]+(0.0266612015617)*x[1]
        s[2,2,1,1]=(1.26455972911)*x[0]+(0.597556950133)*x[1]
        s[2,2,2,0]=(0.397071934531)*x[0]+(-0.792146197976)*x[1]
        s[2,2,2,1]=(-0.531842419771)*x[0]+(-0.437895983867)*x[1]
        s[2,3,0,0]=(0.253602494754)*x[0]+(-0.943660021452)*x[1]
        s[2,3,0,1]=(0.381472581352)*x[0]+(1.36193258173)*x[1]
        s[2,3,1,0]=(0.530962583291)*x[0]+(0.853717668569)*x[1]
        s[2,3,1,1]=(-0.233925714076)*x[0]+(0.20197385961)*x[1]
        s[2,3,2,0]=(0.148804235592)*x[0]+(-0.116553369538)*x[1]
        s[2,3,2,1]=(-1.79360831582)*x[0]+(1.62247077485)*x[1]
        s[2,4,0,0]=(0.617987809937)*x[0]+(1.27685238582)*x[1]
        s[2,4,0,1]=(-0.359480838494)*x[0]+(-0.11716388114)*x[1]
        s[2,4,1,0]=(0.204308581462)*x[0]+(0.0241091749363)*x[1]
        s[2,4,1,1]=(0.424779981556)*x[0]+(-1.01811123317)*x[1]
        s[2,4,2,0]=(-0.708751234946)*x[0]+(0.714679051838)*x[1]
        s[2,4,2,1]=(-0.352350728857)*x[0]+(1.47198192719)*x[1]
        s[3,0,0,0]=(-0.0592432179871)*x[0]+(0.0900836798021)*x[1]
        s[3,0,0,1]=(-0.192266211858)*x[0]+(0.561246968252)*x[1]
        s[3,0,1,0]=(0.30878942472)*x[0]+(-0.578076897247)*x[1]
        s[3,0,1,1]=(-0.246226203035)*x[0]+(0.68840556577)*x[1]
        s[3,0,2,0]=(-0.42433037357)*x[0]+(-0.168333732109)*x[1]
        s[3,0,2,1]=(0.501931661511)*x[0]+(1.53237442299)*x[1]
        s[3,1,0,0]=(-0.835618728532)*x[0]+(-0.823254922183)*x[1]
        s[3,1,0,1]=(-0.609502259942)*x[0]+(0.139549090645)*x[1]
        s[3,1,1,0]=(-0.690897498494)*x[0]+(0.280254985362)*x[1]
        s[3,1,1,1]=(-0.89040023732)*x[0]+(0.616118941926)*x[1]
        s[3,1,2,0]=(-0.90402399371)*x[0]+(-0.399494199588)*x[1]
        s[3,1,2,1]=(1.31002070769)*x[0]+(-0.26001592275)*x[1]
        s[3,2,0,0]=(-0.741667891927)*x[0]+(-1.13827783123)*x[1]
        s[3,2,0,1]=(0.879168335296)*x[0]+(0.232648259015)*x[1]
        s[3,2,1,0]=(0.737095222363)*x[0]+(-1.26232530994)*x[1]
        s[3,2,1,1]=(-0.523976448837)*x[0]+(0.480347424944)*x[1]
        s[3,2,2,0]=(-1.61727611988)*x[0]+(-0.464564622523)*x[1]
        s[3,2,2,1]=(0.928282179697)*x[0]+(0.629575166525)*x[1]
        s[3,3,0,0]=(0.748679757793)*x[0]+(1.26002219505)*x[1]
        s[3,3,0,1]=(-0.497730460683)*x[0]+(0.963512065302)*x[1]
        s[3,3,1,0]=(-0.999580985099)*x[0]+(0.933789269497)*x[1]
        s[3,3,1,1]=(-0.151096110981)*x[0]+(-0.739374723252)*x[1]
        s[3,3,2,0]=(0.257880503573)*x[0]+(-1.07115978896)*x[1]
        s[3,3,2,1]=(0.188404246313)*x[0]+(0.365000571954)*x[1]
        s[3,4,0,0]=(1.63625895457)*x[0]+(-0.685581590196)*x[1]
        s[3,4,0,1]=(-0.665388096854)*x[0]+(-0.9121332766)*x[1]
        s[3,4,1,0]=(1.76715282849)*x[0]+(-1.15317031503)*x[1]
        s[3,4,1,1]=(-0.378469229808)*x[0]+(-0.897880026236)*x[1]
        s[3,4,2,0]=(0.183976703061)*x[0]+(1.76560300417)*x[1]
        s[3,4,2,1]=(-0.569272512314)*x[0]+(1.09830336519)*x[1]
        ref[0,0,0,0]=(-1.26340514811)*x_ref[0]+(0.779300786262)*x_ref[1]
        ref[0,0,0,1]=(0.184428277326)*x_ref[0]+(-0.831497209177)*x_ref[1]
        ref[0,0,1,0]=(1.35946842566)*x_ref[0]+(0.273881497376)*x_ref[1]
        ref[0,0,1,1]=(0.117792005007)*x_ref[0]+(1.06185199224)*x_ref[1]
        ref[0,0,2,0]=(0.810342149523)*x_ref[0]+(-0.271662508688)*x_ref[1]
        ref[0,0,2,1]=(-0.332968381669)*x_ref[0]+(0.0660545206117)*x_ref[1]
        ref[0,1,0,0]=(0.413613455431)*x_ref[0]+(0.641032202343)*x_ref[1]
        ref[0,1,0,1]=(0.773447709702)*x_ref[0]+(0.48460517354)*x_ref[1]
        ref[0,1,1,0]=(0.173632755729)*x_ref[0]+(0.0635245280713)*x_ref[1]
        ref[0,1,1,1]=(0.731914799404)*x_ref[0]+(1.42146728852)*x_ref[1]
        ref[0,1,2,0]=(0.688894013616)*x_ref[0]+(1.0339010436)*x_ref[1]
        ref[0,1,2,1]=(-0.534272384342)*x_ref[0]+(0.186670150961)*x_ref[1]
        ref[0,2,0,0]=(-0.318755700985)*x_ref[0]+(-1.35142313256)*x_ref[1]
        ref[0,2,0,1]=(0.973640957973)*x_ref[0]+(-0.103253061427)*x_ref[1]
        ref[0,2,1,0]=(1.14046612599)*x_ref[0]+(0.238774677715)*x_ref[1]
        ref[0,2,1,1]=(-0.383452508275)*x_ref[0]+(-1.62240702377)*x_ref[1]
        ref[0,2,2,0]=(0.761272968366)*x_ref[0]+(1.446646048)*x_ref[1]
        ref[0,2,2,1]=(0.183707769875)*x_ref[0]+(-0.630747785283)*x_ref[1]
        ref[0,3,0,0]=(-0.474486603803)*x_ref[0]+(-0.0855459523088)*x_ref[1]
        ref[0,3,0,1]=(-0.991769725003)*x_ref[0]+(1.004616452)*x_ref[1]
        ref[0,3,1,0]=(0.261836325547)*x_ref[0]+(0.353357820559)*x_ref[1]
        ref[0,3,1,1]=(0.808880823505)*x_ref[0]+(-0.495896346674)*x_ref[1]
        ref[0,3,2,0]=(1.06375339919)*x_ref[0]+(0.00408804480046)*x_ref[1]
        ref[0,3,2,1]=(-1.32927483163)*x_ref[0]+(-0.0833136005835)*x_ref[1]
        ref[0,4,0,0]=(1.13794063387)*x_ref[0]+(-1.1458884083)*x_ref[1]
        ref[0,4,0,1]=(0.73883590181)*x_ref[0]+(1.45878975539)*x_ref[1]
        ref[0,4,1,0]=(0.428378040368)*x_ref[0]+(-1.25324354587)*x_ref[1]
        ref[0,4,1,1]=(1.19772514118)*x_ref[0]+(0.457504803527)*x_ref[1]
        ref[0,4,2,0]=(-0.646891239861)*x_ref[0]+(1.14291257798)*x_ref[1]
        ref[0,4,2,1]=(-0.579096313066)*x_ref[0]+(-0.0276859933543)*x_ref[1]
        ref[1,0,0,0]=(-0.48935920813)*x_ref[0]+(-0.327683586833)*x_ref[1]
        ref[1,0,0,1]=(0.59804841111)*x_ref[0]+(-0.452980496343)*x_ref[1]
        ref[1,0,1,0]=(0.608017924036)*x_ref[0]+(0.561553199982)*x_ref[1]
        ref[1,0,1,1]=(-0.00361014464582)*x_ref[0]+(0.151090357053)*x_ref[1]
        ref[1,0,2,0]=(1.24385396019)*x_ref[0]+(-0.0213646134992)*x_ref[1]
        ref[1,0,2,1]=(-0.219031743102)*x_ref[0]+(-0.402476762676)*x_ref[1]
        ref[1,1,0,0]=(-0.196889997171)*x_ref[0]+(0.675626499802)*x_ref[1]
        ref[1,1,0,1]=(1.23421246519)*x_ref[0]+(-1.72367262236)*x_ref[1]
        ref[1,1,1,0]=(1.35571029115)*x_ref[0]+(0.506057274132)*x_ref[1]
        ref[1,1,1,1]=(0.310699908104)*x_ref[0]+(-0.292826394288)*x_ref[1]
        ref[1,1,2,0]=(0.122289986824)*x_ref[0]+(-0.15947928964)*x_ref[1]
        ref[1,1,2,1]=(0.662808118991)*x_ref[0]+(-0.927311967494)*x_ref[1]
        ref[1,2,0,0]=(-0.215898581104)*x_ref[0]+(0.32902895413)*x_ref[1]
        ref[1,2,0,1]=(0.393145660435)*x_ref[0]+(1.49040999822)*x_ref[1]
        ref[1,2,1,0]=(0.904144553172)*x_ref[0]+(0.199605443931)*x_ref[1]
        ref[1,2,1,1]=(0.535979155678)*x_ref[0]+(1.78418572873)*x_ref[1]
        ref[1,2,2,0]=(1.15578250862)*x_ref[0]+(0.167811579149)*x_ref[1]
        ref[1,2,2,1]=(-0.0844120860286)*x_ref[0]+(-0.56656817877)*x_ref[1]
        ref[1,3,0,0]=(0.760581325264)*x_ref[0]+(1.00766443232)*x_ref[1]
        ref[1,3,0,1]=(-0.199244602194)*x_ref[0]+(-1.54110932148)*x_ref[1]
        ref[1,3,1,0]=(1.40934237781)*x_ref[0]+(0.700934912313)*x_ref[1]
        ref[1,3,1,1]=(-0.702507416172)*x_ref[0]+(-0.206530323456)*x_ref[1]
        ref[1,3,2,0]=(-0.234301811191)*x_ref[0]+(1.41101365089)*x_ref[1]
        ref[1,3,2,1]=(1.19609435879)*x_ref[0]+(0.295560183318)*x_ref[1]
        ref[1,4,0,0]=(1.15287205838)*x_ref[0]+(0.630564383314)*x_ref[1]
        ref[1,4,0,1]=(-0.993907995704)*x_ref[0]+(-0.30836527499)*x_ref[1]
        ref[1,4,1,0]=(-0.265158946854)*x_ref[0]+(0.689120014144)*x_ref[1]
        ref[1,4,1,1]=(0.503151681106)*x_ref[0]+(-0.928730587416)*x_ref[1]
        ref[1,4,2,0]=(-1.0599860135)*x_ref[0]+(-0.696932543908)*x_ref[1]
        ref[1,4,2,1]=(-0.365096260865)*x_ref[0]+(0.443129268519)*x_ref[1]
        ref[2,0,0,0]=(0.412604298623)*x_ref[0]+(0.591441458509)*x_ref[1]
        ref[2,0,0,1]=(0.536034627643)*x_ref[0]+(-0.735261974283)*x_ref[1]
        ref[2,0,1,0]=(-0.336900405282)*x_ref[0]+(-1.38493113859)*x_ref[1]
        ref[2,0,1,1]=(-0.50634870942)*x_ref[0]+(0.0376544898311)*x_ref[1]
        ref[2,0,2,0]=(-0.0595134635196)*x_ref[0]+(0.067285997446)*x_ref[1]
        ref[2,0,2,1]=(0.308894832145)*x_ref[0]+(-1.94600763342)*x_ref[1]
        ref[2,1,0,0]=(0.443019228067)*x_ref[0]+(-0.68858070865)*x_ref[1]
        ref[2,1,0,1]=(1.4461941226)*x_ref[0]+(0.711186579745)*x_ref[1]
        ref[2,1,1,0]=(0.178777491574)*x_ref[0]+(-0.746259056845)*x_ref[1]
        ref[2,1,1,1]=(1.24682492601)*x_ref[0]+(-0.987057469956)*x_ref[1]
        ref[2,1,2,0]=(1.1863711951)*x_ref[0]+(-0.853660028797)*x_ref[1]
        ref[2,1,2,1]=(1.13176028078)*x_ref[0]+(-0.887521745194)*x_ref[1]
        ref[2,2,0,0]=(0.37579402707)*x_ref[0]+(-0.462533529495)*x_ref[1]
        ref[2,2,0,1]=(-1.27778045393)*x_ref[0]+(0.383310656984)*x_ref[1]
        ref[2,2,1,0]=(-0.0871651954146)*x_ref[0]+(0.0266612015617)*x_ref[1]
        ref[2,2,1,1]=(1.26455972911)*x_ref[0]+(0.597556950133)*x_ref[1]
        ref[2,2,2,0]=(0.397071934531)*x_ref[0]+(-0.792146197976)*x_ref[1]
        ref[2,2,2,1]=(-0.531842419771)*x_ref[0]+(-0.437895983867)*x_ref[1]
        ref[2,3,0,0]=(0.253602494754)*x_ref[0]+(-0.943660021452)*x_ref[1]
        ref[2,3,0,1]=(0.381472581352)*x_ref[0]+(1.36193258173)*x_ref[1]
        ref[2,3,1,0]=(0.530962583291)*x_ref[0]+(0.853717668569)*x_ref[1]
        ref[2,3,1,1]=(-0.233925714076)*x_ref[0]+(0.20197385961)*x_ref[1]
        ref[2,3,2,0]=(0.148804235592)*x_ref[0]+(-0.116553369538)*x_ref[1]
        ref[2,3,2,1]=(-1.79360831582)*x_ref[0]+(1.62247077485)*x_ref[1]
        ref[2,4,0,0]=(0.617987809937)*x_ref[0]+(1.27685238582)*x_ref[1]
        ref[2,4,0,1]=(-0.359480838494)*x_ref[0]+(-0.11716388114)*x_ref[1]
        ref[2,4,1,0]=(0.204308581462)*x_ref[0]+(0.0241091749363)*x_ref[1]
        ref[2,4,1,1]=(0.424779981556)*x_ref[0]+(-1.01811123317)*x_ref[1]
        ref[2,4,2,0]=(-0.708751234946)*x_ref[0]+(0.714679051838)*x_ref[1]
        ref[2,4,2,1]=(-0.352350728857)*x_ref[0]+(1.47198192719)*x_ref[1]
        ref[3,0,0,0]=(-0.0592432179871)*x_ref[0]+(0.0900836798021)*x_ref[1]
        ref[3,0,0,1]=(-0.192266211858)*x_ref[0]+(0.561246968252)*x_ref[1]
        ref[3,0,1,0]=(0.30878942472)*x_ref[0]+(-0.578076897247)*x_ref[1]
        ref[3,0,1,1]=(-0.246226203035)*x_ref[0]+(0.68840556577)*x_ref[1]
        ref[3,0,2,0]=(-0.42433037357)*x_ref[0]+(-0.168333732109)*x_ref[1]
        ref[3,0,2,1]=(0.501931661511)*x_ref[0]+(1.53237442299)*x_ref[1]
        ref[3,1,0,0]=(-0.835618728532)*x_ref[0]+(-0.823254922183)*x_ref[1]
        ref[3,1,0,1]=(-0.609502259942)*x_ref[0]+(0.139549090645)*x_ref[1]
        ref[3,1,1,0]=(-0.690897498494)*x_ref[0]+(0.280254985362)*x_ref[1]
        ref[3,1,1,1]=(-0.89040023732)*x_ref[0]+(0.616118941926)*x_ref[1]
        ref[3,1,2,0]=(-0.90402399371)*x_ref[0]+(-0.399494199588)*x_ref[1]
        ref[3,1,2,1]=(1.31002070769)*x_ref[0]+(-0.26001592275)*x_ref[1]
        ref[3,2,0,0]=(-0.741667891927)*x_ref[0]+(-1.13827783123)*x_ref[1]
        ref[3,2,0,1]=(0.879168335296)*x_ref[0]+(0.232648259015)*x_ref[1]
        ref[3,2,1,0]=(0.737095222363)*x_ref[0]+(-1.26232530994)*x_ref[1]
        ref[3,2,1,1]=(-0.523976448837)*x_ref[0]+(0.480347424944)*x_ref[1]
        ref[3,2,2,0]=(-1.61727611988)*x_ref[0]+(-0.464564622523)*x_ref[1]
        ref[3,2,2,1]=(0.928282179697)*x_ref[0]+(0.629575166525)*x_ref[1]
        ref[3,3,0,0]=(0.748679757793)*x_ref[0]+(1.26002219505)*x_ref[1]
        ref[3,3,0,1]=(-0.497730460683)*x_ref[0]+(0.963512065302)*x_ref[1]
        ref[3,3,1,0]=(-0.999580985099)*x_ref[0]+(0.933789269497)*x_ref[1]
        ref[3,3,1,1]=(-0.151096110981)*x_ref[0]+(-0.739374723252)*x_ref[1]
        ref[3,3,2,0]=(0.257880503573)*x_ref[0]+(-1.07115978896)*x_ref[1]
        ref[3,3,2,1]=(0.188404246313)*x_ref[0]+(0.365000571954)*x_ref[1]
        ref[3,4,0,0]=(1.63625895457)*x_ref[0]+(-0.685581590196)*x_ref[1]
        ref[3,4,0,1]=(-0.665388096854)*x_ref[0]+(-0.9121332766)*x_ref[1]
        ref[3,4,1,0]=(1.76715282849)*x_ref[0]+(-1.15317031503)*x_ref[1]
        ref[3,4,1,1]=(-0.378469229808)*x_ref[0]+(-0.897880026236)*x_ref[1]
        ref[3,4,2,0]=(0.183976703061)*x_ref[0]+(1.76560300417)*x_ref[1]
        ref[3,4,2,1]=(-0.569272512314)*x_ref[0]+(1.09830336519)*x_ref[1]
      else:
        s[0,0,0,0]=(0.146894168765)*x[0]+(1.25822855062)*x[1]+(0.088660814219)*x[2]
        s[0,0,0,1]=(0.572576296123)*x[0]+(0.670191219644)*x[1]+(-0.280619677889)*x[2]
        s[0,0,1,0]=(-1.19817665348)*x[0]+(1.35545322011)*x[1]+(-0.279751222491)*x[2]
        s[0,0,1,1]=(-0.0558236682639)*x[0]+(-1.61042107055)*x[1]+(0.622456635293)*x[2]
        s[0,0,2,0]=(-0.154784205079)*x[0]+(-0.52314996688)*x[1]+(1.0590614932)*x[2]
        s[0,0,2,1]=(1.58101401758)*x[0]+(-0.610736715272)*x[1]+(-0.487233776225)*x[2]
        s[0,1,0,0]=(0.0117365392099)*x[0]+(1.02397373706)*x[1]+(0.407377316737)*x[2]
        s[0,1,0,1]=(0.460387220384)*x[0]+(1.10329384655)*x[1]+(0.252209465365)*x[2]
        s[0,1,1,0]=(-1.33153199048)*x[0]+(-0.759669253534)*x[1]+(-0.856223614724)*x[2]
        s[0,1,1,1]=(-1.00599668771)*x[0]+(0.321343926218)*x[1]+(-0.198860330101)*x[2]
        s[0,1,2,0]=(1.00741183642)*x[0]+(-0.148473398917)*x[1]+(-0.484181353094)*x[2]
        s[0,1,2,1]=(-1.46490648953)*x[0]+(0.621077170698)*x[1]+(0.802865695373)*x[2]
        s[0,2,0,0]=(1.18498612655)*x[0]+(-0.511705644809)*x[1]+(0.563568352278)*x[2]
        s[0,2,0,1]=(-0.807427977876)*x[0]+(-0.252372201752)*x[1]+(0.251722661898)*x[2]
        s[0,2,1,0]=(0.524572175367)*x[0]+(0.49906104527)*x[1]+(-0.601836374873)*x[2]
        s[0,2,1,1]=(-0.318976775137)*x[0]+(-0.153065654914)*x[1]+(-0.458888325378)*x[2]
        s[0,2,2,0]=(-1.45513892332)*x[0]+(-0.252635212625)*x[1]+(0.490153491036)*x[2]
        s[0,2,2,1]=(-0.439422986461)*x[0]+(-0.937409556421)*x[1]+(0.209106988492)*x[2]
        s[0,3,0,0]=(-0.26828401981)*x[0]+(1.70808529771)*x[1]+(0.431245149358)*x[2]
        s[0,3,0,1]=(-0.831846912527)*x[0]+(0.572822507173)*x[1]+(-0.283130753547)*x[2]
        s[0,3,1,0]=(-0.625617835832)*x[0]+(-1.25495997626)*x[1]+(1.28743784867)*x[2]
        s[0,3,1,1]=(1.47223047075)*x[0]+(0.274365456026)*x[1]+(1.34355769078)*x[2]
        s[0,3,2,0]=(-0.0506967319243)*x[0]+(1.16569056569)*x[1]+(-0.965008332019)*x[2]
        s[0,3,2,1]=(0.0106139754664)*x[0]+(0.274252087281)*x[1]+(-0.501970174754)*x[2]
        s[0,4,0,0]=(0.717476010799)*x[0]+(0.380533443018)*x[1]+(-0.304127891816)*x[2]
        s[0,4,0,1]=(-1.29133512255)*x[0]+(-1.16667771199)*x[1]+(0.275809086731)*x[2]
        s[0,4,1,0]=(0.137939049278)*x[0]+(-0.302950342448)*x[1]+(-0.28865204469)*x[2]
        s[0,4,1,1]=(-0.857457318663)*x[0]+(0.423143726447)*x[1]+(0.749889416366)*x[2]
        s[0,4,2,0]=(-0.994428027329)*x[0]+(-0.377780341115)*x[1]+(-0.997480394895)*x[2]
        s[0,4,2,1]=(-0.681653236311)*x[0]+(-0.091905911233)*x[1]+(-0.877007024631)*x[2]
        s[1,0,0,0]=(0.33375482228)*x[0]+(-1.08816236114)*x[1]+(0.154701633038)*x[2]
        s[1,0,0,1]=(0.979227592935)*x[0]+(0.571415419963)*x[1]+(-0.20312466228)*x[2]
        s[1,0,1,0]=(1.59673013189)*x[0]+(1.01626905603)*x[1]+(0.216806440372)*x[2]
        s[1,0,1,1]=(-0.276034446694)*x[0]+(-0.168494544051)*x[1]+(0.981220114947)*x[2]
        s[1,0,2,0]=(0.205085469341)*x[0]+(-0.507970934322)*x[1]+(-0.724745229941)*x[2]
        s[1,0,2,1]=(-0.629331749902)*x[0]+(-1.27086230989)*x[1]+(-1.08479549542)*x[2]
        s[1,1,0,0]=(-0.786670678498)*x[0]+(1.24461814343)*x[1]+(0.644579154985)*x[2]
        s[1,1,0,1]=(1.05414628662)*x[0]+(0.901590532372)*x[1]+(0.809519110188)*x[2]
        s[1,1,1,0]=(-0.679152609307)*x[0]+(0.84101357998)*x[1]+(1.14312188299)*x[2]
        s[1,1,1,1]=(0.137715540853)*x[0]+(1.42995779339)*x[1]+(-1.10477793041)*x[2]
        s[1,1,2,0]=(-0.0647098241647)*x[0]+(1.18580807605)*x[1]+(0.605076723926)*x[2]
        s[1,1,2,1]=(0.744824848213)*x[0]+(0.532106337173)*x[1]+(0.68490109894)*x[2]
        s[1,2,0,0]=(-0.557916122207)*x[0]+(1.06859221643)*x[1]+(0.823215271018)*x[2]
        s[1,2,0,1]=(0.00598565716481)*x[0]+(0.884417903372)*x[1]+(1.7921906066)*x[2]
        s[1,2,1,0]=(-0.47245962749)*x[0]+(-0.906624987584)*x[1]+(0.137792563282)*x[2]
        s[1,2,1,1]=(0.948613971685)*x[0]+(0.753281322508)*x[1]+(0.110929011008)*x[2]
        s[1,2,2,0]=(0.175178588707)*x[0]+(-0.796494364923)*x[1]+(0.390555269738)*x[2]
        s[1,2,2,1]=(-1.04516910973)*x[0]+(-0.921757404949)*x[1]+(0.845139972079)*x[2]
        s[1,3,0,0]=(-1.42392853478)*x[0]+(-0.458461019373)*x[1]+(-0.213286746949)*x[2]
        s[1,3,0,1]=(0.13996349254)*x[0]+(0.108946526608)*x[1]+(-0.848132079317)*x[2]
        s[1,3,1,0]=(-1.44188153548)*x[0]+(-0.108156975475)*x[1]+(0.0715891132579)*x[2]
        s[1,3,1,1]=(0.525856508836)*x[0]+(0.950609265517)*x[1]+(0.0127362240999)*x[2]
        s[1,3,2,0]=(-1.0845975079)*x[0]+(-1.24949492248)*x[1]+(-0.00240035631582)*x[2]
        s[1,3,2,1]=(0.642060593331)*x[0]+(1.46218158975)*x[1]+(1.32385980887)*x[2]
        s[1,4,0,0]=(0.0119712366759)*x[0]+(1.39772768926)*x[1]+(-0.570649410717)*x[2]
        s[1,4,0,1]=(0.320030773751)*x[0]+(0.358086736384)*x[1]+(-0.859721999492)*x[2]
        s[1,4,1,0]=(0.574417642497)*x[0]+(-0.421844733062)*x[1]+(0.611638265232)*x[2]
        s[1,4,1,1]=(-0.228413169897)*x[0]+(-0.0391189339225)*x[1]+(0.635031723033)*x[2]
        s[1,4,2,0]=(0.0424678701733)*x[0]+(0.261307736177)*x[1]+(-0.133428864005)*x[2]
        s[1,4,2,1]=(0.169996877966)*x[0]+(1.30730453844)*x[1]+(-1.07718562596)*x[2]
        s[2,0,0,0]=(0.147232851651)*x[0]+(-0.134174394382)*x[1]+(-0.785323160299)*x[2]
        s[2,0,0,1]=(1.29417433206)*x[0]+(-1.39241927595)*x[1]+(-0.224395758464)*x[2]
        s[2,0,1,0]=(1.60938474462)*x[0]+(0.68821531646)*x[1]+(0.453462382464)*x[2]
        s[2,0,1,1]=(0.63619844559)*x[0]+(-0.956987170375)*x[1]+(-1.24753100984)*x[2]
        s[2,0,2,0]=(1.14194732577)*x[0]+(0.973489037101)*x[1]+(0.187588261589)*x[2]
        s[2,0,2,1]=(-0.898956363323)*x[0]+(0.108722086141)*x[1]+(-1.62445422766)*x[2]
        s[2,1,0,0]=(0.127827687335)*x[0]+(0.216922403148)*x[1]+(-1.5099539458)*x[2]
        s[2,1,0,1]=(-0.769830743998)*x[0]+(1.46060444448)*x[1]+(-0.318171252653)*x[2]
        s[2,1,1,0]=(-0.311976479418)*x[0]+(-0.949061813767)*x[1]+(-0.237478430906)*x[2]
        s[2,1,1,1]=(1.27107354882)*x[0]+(-0.427981502964)*x[1]+(-0.457224808787)*x[2]
        s[2,1,2,0]=(-0.633029461167)*x[0]+(0.255322271264)*x[1]+(-1.79431610464)*x[2]
        s[2,1,2,1]=(1.05355586142)*x[0]+(-1.69722728443)*x[1]+(-1.11928223816)*x[2]
        s[2,2,0,0]=(1.15538109557)*x[0]+(-0.52692609967)*x[1]+(-0.686266572053)*x[2]
        s[2,2,0,1]=(-0.327531879955)*x[0]+(0.365760847752)*x[1]+(-0.0600793962176)*x[2]
        s[2,2,1,0]=(1.16249877038)*x[0]+(0.710438081472)*x[1]+(0.986719636688)*x[2]
        s[2,2,1,1]=(-1.22799525195)*x[0]+(0.748811223969)*x[1]+(0.370195043623)*x[2]
        s[2,2,2,0]=(0.976246115046)*x[0]+(1.10537245041)*x[1]+(-0.759735725531)*x[2]
        s[2,2,2,1]=(1.01411361404)*x[0]+(-0.000596974234967)*x[1]+(-1.39135228207)*x[2]
        s[2,3,0,0]=(-0.581025188916)*x[0]+(-1.43405762699)*x[1]+(-0.912365741093)*x[2]
        s[2,3,0,1]=(-0.0511110676196)*x[0]+(0.230142796427)*x[1]+(-1.58165960293)*x[2]
        s[2,3,1,0]=(-1.46709370576)*x[0]+(0.425035892802)*x[1]+(0.282281348864)*x[2]
        s[2,3,1,1]=(-0.35514320269)*x[0]+(-0.39238669714)*x[1]+(-0.0118498421145)*x[2]
        s[2,3,2,0]=(0.317197568125)*x[0]+(0.689803280183)*x[1]+(-0.238327580547)*x[2]
        s[2,3,2,1]=(1.31241465464)*x[0]+(0.548355955144)*x[1]+(0.0218217641073)*x[2]
        s[2,4,0,0]=(0.787240379652)*x[0]+(-0.906804255968)*x[1]+(-1.17498478521)*x[2]
        s[2,4,0,1]=(1.19721624988)*x[0]+(0.567119981635)*x[1]+(1.42310267091)*x[2]
        s[2,4,1,0]=(-0.199829256778)*x[0]+(0.28908416123)*x[1]+(0.491497583863)*x[2]
        s[2,4,1,1]=(0.336357763916)*x[0]+(-0.750103385624)*x[1]+(-1.19492486127)*x[2]
        s[2,4,2,0]=(1.509049274)*x[0]+(-1.28641285242)*x[1]+(-0.250003617687)*x[2]
        s[2,4,2,1]=(-1.23803199733)*x[0]+(0.688217821134)*x[1]+(-0.436478772021)*x[2]
        s[3,0,0,0]=(0.634885383655)*x[0]+(1.14728850769)*x[1]+(0.417496697665)*x[2]
        s[3,0,0,1]=(-0.326274842766)*x[0]+(-1.47136469943)*x[1]+(1.87373171786)*x[2]
        s[3,0,1,0]=(-0.140107043914)*x[0]+(-0.261565104318)*x[1]+(0.0780015778888)*x[2]
        s[3,0,1,1]=(0.770979582613)*x[0]+(-0.899853446798)*x[1]+(1.45115909188)*x[2]
        s[3,0,2,0]=(0.477865962499)*x[0]+(-0.0621575016815)*x[1]+(-1.28975967232)*x[2]
        s[3,0,2,1]=(-1.41031619045)*x[0]+(1.50155797656)*x[1]+(-1.44890527708)*x[2]
        s[3,1,0,0]=(-1.04632613119)*x[0]+(0.00196059522552)*x[1]+(0.139147980023)*x[2]
        s[3,1,0,1]=(0.217272687162)*x[0]+(0.295773880887)*x[1]+(0.474090779139)*x[2]
        s[3,1,1,0]=(1.50572386246)*x[0]+(0.868579168785)*x[1]+(-0.12394808575)*x[2]
        s[3,1,1,1]=(0.882176073715)*x[0]+(0.144766914417)*x[1]+(-0.408510755049)*x[2]
        s[3,1,2,0]=(-1.21767372555)*x[0]+(0.195682844047)*x[1]+(-1.70399790168)*x[2]
        s[3,1,2,1]=(-0.110311118233)*x[0]+(-0.721806016748)*x[1]+(1.81495268748)*x[2]
        s[3,2,0,0]=(-0.0441997228291)*x[0]+(0.0754984632459)*x[1]+(0.274177049743)*x[2]
        s[3,2,0,1]=(-0.48089133278)*x[0]+(-0.00712717635695)*x[1]+(0.242678324953)*x[2]
        s[3,2,1,0]=(0.48377042087)*x[0]+(1.4374627607)*x[1]+(-1.63994815623)*x[2]
        s[3,2,1,1]=(0.548308948512)*x[0]+(-0.587878136236)*x[1]+(-0.636242676063)*x[2]
        s[3,2,2,0]=(-1.62147144921)*x[0]+(-0.937713365221)*x[1]+(0.416583670897)*x[2]
        s[3,2,2,1]=(1.50617037978)*x[0]+(0.849964476529)*x[1]+(-0.421030022756)*x[2]
        s[3,3,0,0]=(-0.0258240269876)*x[0]+(1.09488855424)*x[1]+(-0.864078888453)*x[2]
        s[3,3,0,1]=(0.121803754668)*x[0]+(1.47064577586)*x[1]+(-1.15267381807)*x[2]
        s[3,3,1,0]=(0.7847153928)*x[0]+(1.19923938448)*x[1]+(-0.220983796213)*x[2]
        s[3,3,1,1]=(0.591086037588)*x[0]+(1.49019206942)*x[1]+(-0.764272587507)*x[2]
        s[3,3,2,0]=(-0.751721311032)*x[0]+(-1.41725473775)*x[1]+(-0.182681015042)*x[2]
        s[3,3,2,1]=(1.18789607758)*x[0]+(0.321544556274)*x[1]+(-0.37682906633)*x[2]
        s[3,4,0,0]=(-0.314136382522)*x[0]+(0.190342372946)*x[1]+(-0.0583598827644)*x[2]
        s[3,4,0,1]=(1.87879524007)*x[0]+(0.0472715596509)*x[1]+(0.849571713097)*x[2]
        s[3,4,1,0]=(0.816120005278)*x[0]+(0.296980756484)*x[1]+(0.251581386555)*x[2]
        s[3,4,1,1]=(0.12560195013)*x[0]+(-0.293571823053)*x[1]+(-1.82952568911)*x[2]
        s[3,4,2,0]=(1.03923438653)*x[0]+(-0.899000072461)*x[1]+(-1.48431797596)*x[2]
        s[3,4,2,1]=(-1.12443865282)*x[0]+(-0.200378451615)*x[1]+(-1.41505465511)*x[2]
        ref[0,0,0,0]=(0.146894168765)*x_ref[0]+(1.25822855062)*x_ref[1]+(0.088660814219)*x_ref[2]
        ref[0,0,0,1]=(0.572576296123)*x_ref[0]+(0.670191219644)*x_ref[1]+(-0.280619677889)*x_ref[2]
        ref[0,0,1,0]=(-1.19817665348)*x_ref[0]+(1.35545322011)*x_ref[1]+(-0.279751222491)*x_ref[2]
        ref[0,0,1,1]=(-0.0558236682639)*x_ref[0]+(-1.61042107055)*x_ref[1]+(0.622456635293)*x_ref[2]
        ref[0,0,2,0]=(-0.154784205079)*x_ref[0]+(-0.52314996688)*x_ref[1]+(1.0590614932)*x_ref[2]
        ref[0,0,2,1]=(1.58101401758)*x_ref[0]+(-0.610736715272)*x_ref[1]+(-0.487233776225)*x_ref[2]
        ref[0,1,0,0]=(0.0117365392099)*x_ref[0]+(1.02397373706)*x_ref[1]+(0.407377316737)*x_ref[2]
        ref[0,1,0,1]=(0.460387220384)*x_ref[0]+(1.10329384655)*x_ref[1]+(0.252209465365)*x_ref[2]
        ref[0,1,1,0]=(-1.33153199048)*x_ref[0]+(-0.759669253534)*x_ref[1]+(-0.856223614724)*x_ref[2]
        ref[0,1,1,1]=(-1.00599668771)*x_ref[0]+(0.321343926218)*x_ref[1]+(-0.198860330101)*x_ref[2]
        ref[0,1,2,0]=(1.00741183642)*x_ref[0]+(-0.148473398917)*x_ref[1]+(-0.484181353094)*x_ref[2]
        ref[0,1,2,1]=(-1.46490648953)*x_ref[0]+(0.621077170698)*x_ref[1]+(0.802865695373)*x_ref[2]
        ref[0,2,0,0]=(1.18498612655)*x_ref[0]+(-0.511705644809)*x_ref[1]+(0.563568352278)*x_ref[2]
        ref[0,2,0,1]=(-0.807427977876)*x_ref[0]+(-0.252372201752)*x_ref[1]+(0.251722661898)*x_ref[2]
        ref[0,2,1,0]=(0.524572175367)*x_ref[0]+(0.49906104527)*x_ref[1]+(-0.601836374873)*x_ref[2]
        ref[0,2,1,1]=(-0.318976775137)*x_ref[0]+(-0.153065654914)*x_ref[1]+(-0.458888325378)*x_ref[2]
        ref[0,2,2,0]=(-1.45513892332)*x_ref[0]+(-0.252635212625)*x_ref[1]+(0.490153491036)*x_ref[2]
        ref[0,2,2,1]=(-0.439422986461)*x_ref[0]+(-0.937409556421)*x_ref[1]+(0.209106988492)*x_ref[2]
        ref[0,3,0,0]=(-0.26828401981)*x_ref[0]+(1.70808529771)*x_ref[1]+(0.431245149358)*x_ref[2]
        ref[0,3,0,1]=(-0.831846912527)*x_ref[0]+(0.572822507173)*x_ref[1]+(-0.283130753547)*x_ref[2]
        ref[0,3,1,0]=(-0.625617835832)*x_ref[0]+(-1.25495997626)*x_ref[1]+(1.28743784867)*x_ref[2]
        ref[0,3,1,1]=(1.47223047075)*x_ref[0]+(0.274365456026)*x_ref[1]+(1.34355769078)*x_ref[2]
        ref[0,3,2,0]=(-0.0506967319243)*x_ref[0]+(1.16569056569)*x_ref[1]+(-0.965008332019)*x_ref[2]
        ref[0,3,2,1]=(0.0106139754664)*x_ref[0]+(0.274252087281)*x_ref[1]+(-0.501970174754)*x_ref[2]
        ref[0,4,0,0]=(0.717476010799)*x_ref[0]+(0.380533443018)*x_ref[1]+(-0.304127891816)*x_ref[2]
        ref[0,4,0,1]=(-1.29133512255)*x_ref[0]+(-1.16667771199)*x_ref[1]+(0.275809086731)*x_ref[2]
        ref[0,4,1,0]=(0.137939049278)*x_ref[0]+(-0.302950342448)*x_ref[1]+(-0.28865204469)*x_ref[2]
        ref[0,4,1,1]=(-0.857457318663)*x_ref[0]+(0.423143726447)*x_ref[1]+(0.749889416366)*x_ref[2]
        ref[0,4,2,0]=(-0.994428027329)*x_ref[0]+(-0.377780341115)*x_ref[1]+(-0.997480394895)*x_ref[2]
        ref[0,4,2,1]=(-0.681653236311)*x_ref[0]+(-0.091905911233)*x_ref[1]+(-0.877007024631)*x_ref[2]
        ref[1,0,0,0]=(0.33375482228)*x_ref[0]+(-1.08816236114)*x_ref[1]+(0.154701633038)*x_ref[2]
        ref[1,0,0,1]=(0.979227592935)*x_ref[0]+(0.571415419963)*x_ref[1]+(-0.20312466228)*x_ref[2]
        ref[1,0,1,0]=(1.59673013189)*x_ref[0]+(1.01626905603)*x_ref[1]+(0.216806440372)*x_ref[2]
        ref[1,0,1,1]=(-0.276034446694)*x_ref[0]+(-0.168494544051)*x_ref[1]+(0.981220114947)*x_ref[2]
        ref[1,0,2,0]=(0.205085469341)*x_ref[0]+(-0.507970934322)*x_ref[1]+(-0.724745229941)*x_ref[2]
        ref[1,0,2,1]=(-0.629331749902)*x_ref[0]+(-1.27086230989)*x_ref[1]+(-1.08479549542)*x_ref[2]
        ref[1,1,0,0]=(-0.786670678498)*x_ref[0]+(1.24461814343)*x_ref[1]+(0.644579154985)*x_ref[2]
        ref[1,1,0,1]=(1.05414628662)*x_ref[0]+(0.901590532372)*x_ref[1]+(0.809519110188)*x_ref[2]
        ref[1,1,1,0]=(-0.679152609307)*x_ref[0]+(0.84101357998)*x_ref[1]+(1.14312188299)*x_ref[2]
        ref[1,1,1,1]=(0.137715540853)*x_ref[0]+(1.42995779339)*x_ref[1]+(-1.10477793041)*x_ref[2]
        ref[1,1,2,0]=(-0.0647098241647)*x_ref[0]+(1.18580807605)*x_ref[1]+(0.605076723926)*x_ref[2]
        ref[1,1,2,1]=(0.744824848213)*x_ref[0]+(0.532106337173)*x_ref[1]+(0.68490109894)*x_ref[2]
        ref[1,2,0,0]=(-0.557916122207)*x_ref[0]+(1.06859221643)*x_ref[1]+(0.823215271018)*x_ref[2]
        ref[1,2,0,1]=(0.00598565716481)*x_ref[0]+(0.884417903372)*x_ref[1]+(1.7921906066)*x_ref[2]
        ref[1,2,1,0]=(-0.47245962749)*x_ref[0]+(-0.906624987584)*x_ref[1]+(0.137792563282)*x_ref[2]
        ref[1,2,1,1]=(0.948613971685)*x_ref[0]+(0.753281322508)*x_ref[1]+(0.110929011008)*x_ref[2]
        ref[1,2,2,0]=(0.175178588707)*x_ref[0]+(-0.796494364923)*x_ref[1]+(0.390555269738)*x_ref[2]
        ref[1,2,2,1]=(-1.04516910973)*x_ref[0]+(-0.921757404949)*x_ref[1]+(0.845139972079)*x_ref[2]
        ref[1,3,0,0]=(-1.42392853478)*x_ref[0]+(-0.458461019373)*x_ref[1]+(-0.213286746949)*x_ref[2]
        ref[1,3,0,1]=(0.13996349254)*x_ref[0]+(0.108946526608)*x_ref[1]+(-0.848132079317)*x_ref[2]
        ref[1,3,1,0]=(-1.44188153548)*x_ref[0]+(-0.108156975475)*x_ref[1]+(0.0715891132579)*x_ref[2]
        ref[1,3,1,1]=(0.525856508836)*x_ref[0]+(0.950609265517)*x_ref[1]+(0.0127362240999)*x_ref[2]
        ref[1,3,2,0]=(-1.0845975079)*x_ref[0]+(-1.24949492248)*x_ref[1]+(-0.00240035631582)*x_ref[2]
        ref[1,3,2,1]=(0.642060593331)*x_ref[0]+(1.46218158975)*x_ref[1]+(1.32385980887)*x_ref[2]
        ref[1,4,0,0]=(0.0119712366759)*x_ref[0]+(1.39772768926)*x_ref[1]+(-0.570649410717)*x_ref[2]
        ref[1,4,0,1]=(0.320030773751)*x_ref[0]+(0.358086736384)*x_ref[1]+(-0.859721999492)*x_ref[2]
        ref[1,4,1,0]=(0.574417642497)*x_ref[0]+(-0.421844733062)*x_ref[1]+(0.611638265232)*x_ref[2]
        ref[1,4,1,1]=(-0.228413169897)*x_ref[0]+(-0.0391189339225)*x_ref[1]+(0.635031723033)*x_ref[2]
        ref[1,4,2,0]=(0.0424678701733)*x_ref[0]+(0.261307736177)*x_ref[1]+(-0.133428864005)*x_ref[2]
        ref[1,4,2,1]=(0.169996877966)*x_ref[0]+(1.30730453844)*x_ref[1]+(-1.07718562596)*x_ref[2]
        ref[2,0,0,0]=(0.147232851651)*x_ref[0]+(-0.134174394382)*x_ref[1]+(-0.785323160299)*x_ref[2]
        ref[2,0,0,1]=(1.29417433206)*x_ref[0]+(-1.39241927595)*x_ref[1]+(-0.224395758464)*x_ref[2]
        ref[2,0,1,0]=(1.60938474462)*x_ref[0]+(0.68821531646)*x_ref[1]+(0.453462382464)*x_ref[2]
        ref[2,0,1,1]=(0.63619844559)*x_ref[0]+(-0.956987170375)*x_ref[1]+(-1.24753100984)*x_ref[2]
        ref[2,0,2,0]=(1.14194732577)*x_ref[0]+(0.973489037101)*x_ref[1]+(0.187588261589)*x_ref[2]
        ref[2,0,2,1]=(-0.898956363323)*x_ref[0]+(0.108722086141)*x_ref[1]+(-1.62445422766)*x_ref[2]
        ref[2,1,0,0]=(0.127827687335)*x_ref[0]+(0.216922403148)*x_ref[1]+(-1.5099539458)*x_ref[2]
        ref[2,1,0,1]=(-0.769830743998)*x_ref[0]+(1.46060444448)*x_ref[1]+(-0.318171252653)*x_ref[2]
        ref[2,1,1,0]=(-0.311976479418)*x_ref[0]+(-0.949061813767)*x_ref[1]+(-0.237478430906)*x_ref[2]
        ref[2,1,1,1]=(1.27107354882)*x_ref[0]+(-0.427981502964)*x_ref[1]+(-0.457224808787)*x_ref[2]
        ref[2,1,2,0]=(-0.633029461167)*x_ref[0]+(0.255322271264)*x_ref[1]+(-1.79431610464)*x_ref[2]
        ref[2,1,2,1]=(1.05355586142)*x_ref[0]+(-1.69722728443)*x_ref[1]+(-1.11928223816)*x_ref[2]
        ref[2,2,0,0]=(1.15538109557)*x_ref[0]+(-0.52692609967)*x_ref[1]+(-0.686266572053)*x_ref[2]
        ref[2,2,0,1]=(-0.327531879955)*x_ref[0]+(0.365760847752)*x_ref[1]+(-0.0600793962176)*x_ref[2]
        ref[2,2,1,0]=(1.16249877038)*x_ref[0]+(0.710438081472)*x_ref[1]+(0.986719636688)*x_ref[2]
        ref[2,2,1,1]=(-1.22799525195)*x_ref[0]+(0.748811223969)*x_ref[1]+(0.370195043623)*x_ref[2]
        ref[2,2,2,0]=(0.976246115046)*x_ref[0]+(1.10537245041)*x_ref[1]+(-0.759735725531)*x_ref[2]
        ref[2,2,2,1]=(1.01411361404)*x_ref[0]+(-0.000596974234967)*x_ref[1]+(-1.39135228207)*x_ref[2]
        ref[2,3,0,0]=(-0.581025188916)*x_ref[0]+(-1.43405762699)*x_ref[1]+(-0.912365741093)*x_ref[2]
        ref[2,3,0,1]=(-0.0511110676196)*x_ref[0]+(0.230142796427)*x_ref[1]+(-1.58165960293)*x_ref[2]
        ref[2,3,1,0]=(-1.46709370576)*x_ref[0]+(0.425035892802)*x_ref[1]+(0.282281348864)*x_ref[2]
        ref[2,3,1,1]=(-0.35514320269)*x_ref[0]+(-0.39238669714)*x_ref[1]+(-0.0118498421145)*x_ref[2]
        ref[2,3,2,0]=(0.317197568125)*x_ref[0]+(0.689803280183)*x_ref[1]+(-0.238327580547)*x_ref[2]
        ref[2,3,2,1]=(1.31241465464)*x_ref[0]+(0.548355955144)*x_ref[1]+(0.0218217641073)*x_ref[2]
        ref[2,4,0,0]=(0.787240379652)*x_ref[0]+(-0.906804255968)*x_ref[1]+(-1.17498478521)*x_ref[2]
        ref[2,4,0,1]=(1.19721624988)*x_ref[0]+(0.567119981635)*x_ref[1]+(1.42310267091)*x_ref[2]
        ref[2,4,1,0]=(-0.199829256778)*x_ref[0]+(0.28908416123)*x_ref[1]+(0.491497583863)*x_ref[2]
        ref[2,4,1,1]=(0.336357763916)*x_ref[0]+(-0.750103385624)*x_ref[1]+(-1.19492486127)*x_ref[2]
        ref[2,4,2,0]=(1.509049274)*x_ref[0]+(-1.28641285242)*x_ref[1]+(-0.250003617687)*x_ref[2]
        ref[2,4,2,1]=(-1.23803199733)*x_ref[0]+(0.688217821134)*x_ref[1]+(-0.436478772021)*x_ref[2]
        ref[3,0,0,0]=(0.634885383655)*x_ref[0]+(1.14728850769)*x_ref[1]+(0.417496697665)*x_ref[2]
        ref[3,0,0,1]=(-0.326274842766)*x_ref[0]+(-1.47136469943)*x_ref[1]+(1.87373171786)*x_ref[2]
        ref[3,0,1,0]=(-0.140107043914)*x_ref[0]+(-0.261565104318)*x_ref[1]+(0.0780015778888)*x_ref[2]
        ref[3,0,1,1]=(0.770979582613)*x_ref[0]+(-0.899853446798)*x_ref[1]+(1.45115909188)*x_ref[2]
        ref[3,0,2,0]=(0.477865962499)*x_ref[0]+(-0.0621575016815)*x_ref[1]+(-1.28975967232)*x_ref[2]
        ref[3,0,2,1]=(-1.41031619045)*x_ref[0]+(1.50155797656)*x_ref[1]+(-1.44890527708)*x_ref[2]
        ref[3,1,0,0]=(-1.04632613119)*x_ref[0]+(0.00196059522552)*x_ref[1]+(0.139147980023)*x_ref[2]
        ref[3,1,0,1]=(0.217272687162)*x_ref[0]+(0.295773880887)*x_ref[1]+(0.474090779139)*x_ref[2]
        ref[3,1,1,0]=(1.50572386246)*x_ref[0]+(0.868579168785)*x_ref[1]+(-0.12394808575)*x_ref[2]
        ref[3,1,1,1]=(0.882176073715)*x_ref[0]+(0.144766914417)*x_ref[1]+(-0.408510755049)*x_ref[2]
        ref[3,1,2,0]=(-1.21767372555)*x_ref[0]+(0.195682844047)*x_ref[1]+(-1.70399790168)*x_ref[2]
        ref[3,1,2,1]=(-0.110311118233)*x_ref[0]+(-0.721806016748)*x_ref[1]+(1.81495268748)*x_ref[2]
        ref[3,2,0,0]=(-0.0441997228291)*x_ref[0]+(0.0754984632459)*x_ref[1]+(0.274177049743)*x_ref[2]
        ref[3,2,0,1]=(-0.48089133278)*x_ref[0]+(-0.00712717635695)*x_ref[1]+(0.242678324953)*x_ref[2]
        ref[3,2,1,0]=(0.48377042087)*x_ref[0]+(1.4374627607)*x_ref[1]+(-1.63994815623)*x_ref[2]
        ref[3,2,1,1]=(0.548308948512)*x_ref[0]+(-0.587878136236)*x_ref[1]+(-0.636242676063)*x_ref[2]
        ref[3,2,2,0]=(-1.62147144921)*x_ref[0]+(-0.937713365221)*x_ref[1]+(0.416583670897)*x_ref[2]
        ref[3,2,2,1]=(1.50617037978)*x_ref[0]+(0.849964476529)*x_ref[1]+(-0.421030022756)*x_ref[2]
        ref[3,3,0,0]=(-0.0258240269876)*x_ref[0]+(1.09488855424)*x_ref[1]+(-0.864078888453)*x_ref[2]
        ref[3,3,0,1]=(0.121803754668)*x_ref[0]+(1.47064577586)*x_ref[1]+(-1.15267381807)*x_ref[2]
        ref[3,3,1,0]=(0.7847153928)*x_ref[0]+(1.19923938448)*x_ref[1]+(-0.220983796213)*x_ref[2]
        ref[3,3,1,1]=(0.591086037588)*x_ref[0]+(1.49019206942)*x_ref[1]+(-0.764272587507)*x_ref[2]
        ref[3,3,2,0]=(-0.751721311032)*x_ref[0]+(-1.41725473775)*x_ref[1]+(-0.182681015042)*x_ref[2]
        ref[3,3,2,1]=(1.18789607758)*x_ref[0]+(0.321544556274)*x_ref[1]+(-0.37682906633)*x_ref[2]
        ref[3,4,0,0]=(-0.314136382522)*x_ref[0]+(0.190342372946)*x_ref[1]+(-0.0583598827644)*x_ref[2]
        ref[3,4,0,1]=(1.87879524007)*x_ref[0]+(0.0472715596509)*x_ref[1]+(0.849571713097)*x_ref[2]
        ref[3,4,1,0]=(0.816120005278)*x_ref[0]+(0.296980756484)*x_ref[1]+(0.251581386555)*x_ref[2]
        ref[3,4,1,1]=(0.12560195013)*x_ref[0]+(-0.293571823053)*x_ref[1]+(-1.82952568911)*x_ref[2]
        ref[3,4,2,0]=(1.03923438653)*x_ref[0]+(-0.899000072461)*x_ref[1]+(-1.48431797596)*x_ref[2]
        ref[3,4,2,1]=(-1.12443865282)*x_ref[0]+(-0.200378451615)*x_ref[1]+(-1.41505465511)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Function_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.775940204341)*x[0]**o+(0.408012597308)*x[0]+(-0.526684561152)*x[1]**o+(-0.900147960936)*x[1]
        ref=(0.775940204341)*x_ref[0]**o+(0.408012597308)*x_ref[0]+(-0.526684561152)*x_ref[1]**o+(-0.900147960936)*x_ref[1]
      else:
        s=(-0.0117662561618)*x[0]**o+(0.813445281753)*x[0]+(-0.179209673046)*x[1]**o+(-0.841498005448)*x[1]+(0.660691793582)*x[2]**o+(0.149938868852)*x[2]
        ref=(-0.0117662561618)*x_ref[0]**o+(0.813445281753)*x_ref[0]+(-0.179209673046)*x_ref[1]**o+(-0.841498005448)*x_ref[1]+(0.660691793582)*x_ref[2]**o+(0.149938868852)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Function_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.940436734602)*x[0]**o+(-0.499319247438)*x[0]+(-0.87358475558)*x[1]**o+(-0.243771941219)*x[1]
        s[1]=(0.289360868897)*x[0]**o+(0.808574135723)*x[0]+(0.881242535917)*x[1]**o+(0.452398398881)*x[1]
        ref[0]=(0.940436734602)*x_ref[0]**o+(-0.499319247438)*x_ref[0]+(-0.87358475558)*x_ref[1]**o+(-0.243771941219)*x_ref[1]
        ref[1]=(0.289360868897)*x_ref[0]**o+(0.808574135723)*x_ref[0]+(0.881242535917)*x_ref[1]**o+(0.452398398881)*x_ref[1]
      else:
        s[0]=(-0.354278440772)*x[0]**o+(-0.0252551558693)*x[0]+(-0.955840831777)*x[1]**o+(-0.183292049624)*x[1]+(0.132001820383)*x[2]**o+(0.968364405641)*x[2]
        s[1]=(0.00318403616402)*x[0]**o+(0.278370320241)*x[0]+(0.451122966354)*x[1]**o+(0.0275444352047)*x[1]+(-0.0499740581816)*x[2]**o+(0.402684856351)*x[2]
        ref[0]=(-0.354278440772)*x_ref[0]**o+(-0.0252551558693)*x_ref[0]+(-0.955840831777)*x_ref[1]**o+(-0.183292049624)*x_ref[1]+(0.132001820383)*x_ref[2]**o+(0.968364405641)*x_ref[2]
        ref[1]=(0.00318403616402)*x_ref[0]**o+(0.278370320241)*x_ref[0]+(0.451122966354)*x_ref[1]**o+(0.0275444352047)*x_ref[1]+(-0.0499740581816)*x_ref[2]**o+(0.402684856351)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Function_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.877859091933)*x[0]**o+(-0.169762567932)*x[0]+(0.552958849443)*x[1]**o+(-0.11599142298)*x[1]
        s[0,1]=(0.471005726569)*x[0]**o+(-0.632583456078)*x[0]+(-0.360082824936)*x[1]**o+(-0.967781840973)*x[1]
        s[0,2]=(0.923269093961)*x[0]**o+(0.0574842386662)*x[0]+(-0.530617507423)*x[1]**o+(0.0473937183531)*x[1]
        s[0,3]=(0.824862043209)*x[0]**o+(0.698896112144)*x[0]+(-0.383808246808)*x[1]**o+(0.24923604362)*x[1]
        s[0,4]=(0.297428271453)*x[0]**o+(-0.18963891196)*x[0]+(-0.16010009844)*x[1]**o+(-0.0635747494778)*x[1]
        s[1,0]=(0.212999154122)*x[0]**o+(0.495137279523)*x[0]+(0.532511060201)*x[1]**o+(0.320057408221)*x[1]
        s[1,1]=(-0.994527399384)*x[0]**o+(0.0991063567708)*x[0]+(0.0375586255455)*x[1]**o+(-0.306126578081)*x[1]
        s[1,2]=(-0.140745202296)*x[0]**o+(-0.967168602143)*x[0]+(0.977938661252)*x[1]**o+(0.97796652405)*x[1]
        s[1,3]=(0.0508848805668)*x[0]**o+(-0.370907761322)*x[0]+(0.979063533914)*x[1]**o+(0.922258254977)*x[1]
        s[1,4]=(0.630127237807)*x[0]**o+(0.712161343099)*x[0]+(-0.15778527275)*x[1]**o+(0.590043042293)*x[1]
        s[2,0]=(-0.531086529572)*x[0]**o+(0.467450157091)*x[0]+(0.234837506275)*x[1]**o+(0.407064919776)*x[1]
        s[2,1]=(-0.056859895221)*x[0]**o+(0.533125022835)*x[0]+(0.904938824716)*x[1]**o+(0.793055507712)*x[1]
        s[2,2]=(-0.551130390502)*x[0]**o+(-0.01664955762)*x[0]+(0.788924790297)*x[1]**o+(-0.518737057199)*x[1]
        s[2,3]=(-0.593665554555)*x[0]**o+(-0.936497023625)*x[0]+(0.553767265739)*x[1]**o+(0.650790290426)*x[1]
        s[2,4]=(0.295642143189)*x[0]**o+(0.996496426422)*x[0]+(0.269107887571)*x[1]**o+(0.441099854152)*x[1]
        s[3,0]=(0.348944839773)*x[0]**o+(0.0364801849524)*x[0]+(0.00330101221443)*x[1]**o+(0.806022278452)*x[1]
        s[3,1]=(-0.620797900903)*x[0]**o+(0.516037923834)*x[0]+(0.861530193737)*x[1]**o+(0.466776529186)*x[1]
        s[3,2]=(0.055887897583)*x[0]**o+(-0.536336196157)*x[0]+(-0.889596296008)*x[1]**o+(0.299571474478)*x[1]
        s[3,3]=(-0.862274293699)*x[0]**o+(0.740404601136)*x[0]+(0.20012481659)*x[1]**o+(0.0655004723729)*x[1]
        s[3,4]=(0.561396042945)*x[0]**o+(-0.58695912231)*x[0]+(0.843743224177)*x[1]**o+(0.549571256677)*x[1]
        ref[0,0]=(0.877859091933)*x_ref[0]**o+(-0.169762567932)*x_ref[0]+(0.552958849443)*x_ref[1]**o+(-0.11599142298)*x_ref[1]
        ref[0,1]=(0.471005726569)*x_ref[0]**o+(-0.632583456078)*x_ref[0]+(-0.360082824936)*x_ref[1]**o+(-0.967781840973)*x_ref[1]
        ref[0,2]=(0.923269093961)*x_ref[0]**o+(0.0574842386662)*x_ref[0]+(-0.530617507423)*x_ref[1]**o+(0.0473937183531)*x_ref[1]
        ref[0,3]=(0.824862043209)*x_ref[0]**o+(0.698896112144)*x_ref[0]+(-0.383808246808)*x_ref[1]**o+(0.24923604362)*x_ref[1]
        ref[0,4]=(0.297428271453)*x_ref[0]**o+(-0.18963891196)*x_ref[0]+(-0.16010009844)*x_ref[1]**o+(-0.0635747494778)*x_ref[1]
        ref[1,0]=(0.212999154122)*x_ref[0]**o+(0.495137279523)*x_ref[0]+(0.532511060201)*x_ref[1]**o+(0.320057408221)*x_ref[1]
        ref[1,1]=(-0.994527399384)*x_ref[0]**o+(0.0991063567708)*x_ref[0]+(0.0375586255455)*x_ref[1]**o+(-0.306126578081)*x_ref[1]
        ref[1,2]=(-0.140745202296)*x_ref[0]**o+(-0.967168602143)*x_ref[0]+(0.977938661252)*x_ref[1]**o+(0.97796652405)*x_ref[1]
        ref[1,3]=(0.0508848805668)*x_ref[0]**o+(-0.370907761322)*x_ref[0]+(0.979063533914)*x_ref[1]**o+(0.922258254977)*x_ref[1]
        ref[1,4]=(0.630127237807)*x_ref[0]**o+(0.712161343099)*x_ref[0]+(-0.15778527275)*x_ref[1]**o+(0.590043042293)*x_ref[1]
        ref[2,0]=(-0.531086529572)*x_ref[0]**o+(0.467450157091)*x_ref[0]+(0.234837506275)*x_ref[1]**o+(0.407064919776)*x_ref[1]
        ref[2,1]=(-0.056859895221)*x_ref[0]**o+(0.533125022835)*x_ref[0]+(0.904938824716)*x_ref[1]**o+(0.793055507712)*x_ref[1]
        ref[2,2]=(-0.551130390502)*x_ref[0]**o+(-0.01664955762)*x_ref[0]+(0.788924790297)*x_ref[1]**o+(-0.518737057199)*x_ref[1]
        ref[2,3]=(-0.593665554555)*x_ref[0]**o+(-0.936497023625)*x_ref[0]+(0.553767265739)*x_ref[1]**o+(0.650790290426)*x_ref[1]
        ref[2,4]=(0.295642143189)*x_ref[0]**o+(0.996496426422)*x_ref[0]+(0.269107887571)*x_ref[1]**o+(0.441099854152)*x_ref[1]
        ref[3,0]=(0.348944839773)*x_ref[0]**o+(0.0364801849524)*x_ref[0]+(0.00330101221443)*x_ref[1]**o+(0.806022278452)*x_ref[1]
        ref[3,1]=(-0.620797900903)*x_ref[0]**o+(0.516037923834)*x_ref[0]+(0.861530193737)*x_ref[1]**o+(0.466776529186)*x_ref[1]
        ref[3,2]=(0.055887897583)*x_ref[0]**o+(-0.536336196157)*x_ref[0]+(-0.889596296008)*x_ref[1]**o+(0.299571474478)*x_ref[1]
        ref[3,3]=(-0.862274293699)*x_ref[0]**o+(0.740404601136)*x_ref[0]+(0.20012481659)*x_ref[1]**o+(0.0655004723729)*x_ref[1]
        ref[3,4]=(0.561396042945)*x_ref[0]**o+(-0.58695912231)*x_ref[0]+(0.843743224177)*x_ref[1]**o+(0.549571256677)*x_ref[1]
      else:
        s[0,0]=(-0.169971719578)*x[0]**o+(-0.490891054826)*x[0]+(0.744482943882)*x[1]**o+(0.353623823086)*x[1]+(0.265501620199)*x[2]**o+(-0.629048564143)*x[2]
        s[0,1]=(0.213635035052)*x[0]**o+(0.584035034456)*x[0]+(0.133486854141)*x[1]**o+(-0.563257345568)*x[1]+(-0.846099536223)*x[2]**o+(-0.854728289866)*x[2]
        s[0,2]=(0.342184893432)*x[0]**o+(0.515624951667)*x[0]+(-0.136008144001)*x[1]**o+(-0.280682548245)*x[1]+(-0.927880146409)*x[2]**o+(0.23737193095)*x[2]
        s[0,3]=(0.0547841640747)*x[0]**o+(-0.969984349739)*x[0]+(-0.155447158154)*x[1]**o+(-0.495088348804)*x[1]+(0.879915433385)*x[2]**o+(0.438389550488)*x[2]
        s[0,4]=(0.15107258681)*x[0]**o+(0.878036814589)*x[0]+(0.875901664353)*x[1]**o+(0.455383770416)*x[1]+(0.316791478263)*x[2]**o+(0.13991508459)*x[2]
        s[1,0]=(0.627219372033)*x[0]**o+(0.0514202591569)*x[0]+(0.696459316229)*x[1]**o+(0.232864735057)*x[1]+(0.516267189254)*x[2]**o+(-0.277533957737)*x[2]
        s[1,1]=(-0.474165959949)*x[0]**o+(0.875608881493)*x[0]+(-0.432866681632)*x[1]**o+(0.165933079742)*x[1]+(-0.0745331372502)*x[2]**o+(-0.0168977805625)*x[2]
        s[1,2]=(-0.228391157708)*x[0]**o+(-0.294221312175)*x[0]+(-0.121429334894)*x[1]**o+(-0.557107244165)*x[1]+(-0.97514025916)*x[2]**o+(0.041970524119)*x[2]
        s[1,3]=(0.344978415951)*x[0]**o+(0.319420517862)*x[0]+(0.802947508962)*x[1]**o+(0.169628742731)*x[1]+(-0.380928178408)*x[2]**o+(0.272551034104)*x[2]
        s[1,4]=(0.613832101706)*x[0]**o+(-0.332306428069)*x[0]+(0.02563364621)*x[1]**o+(-0.415952850048)*x[1]+(0.673350831947)*x[2]**o+(0.713348183687)*x[2]
        s[2,0]=(0.290558527133)*x[0]**o+(0.135335532216)*x[0]+(-0.352459421266)*x[1]**o+(-0.854183560051)*x[1]+(0.581843223923)*x[2]**o+(0.97378057053)*x[2]
        s[2,1]=(-0.328513503613)*x[0]**o+(-0.205196873021)*x[0]+(-0.910190961873)*x[1]**o+(0.568820988101)*x[1]+(0.495810061604)*x[2]**o+(-0.761749884838)*x[2]
        s[2,2]=(-0.00783916993558)*x[0]**o+(0.898831843878)*x[0]+(0.26265892346)*x[1]**o+(-0.654638473159)*x[1]+(-0.713198105003)*x[2]**o+(-0.59350752607)*x[2]
        s[2,3]=(-0.881651184273)*x[0]**o+(-0.834213060843)*x[0]+(-0.885142407575)*x[1]**o+(0.206850906485)*x[1]+(0.372454818886)*x[2]**o+(0.159769678242)*x[2]
        s[2,4]=(-0.937735543761)*x[0]**o+(0.529592561341)*x[0]+(-0.936499401425)*x[1]**o+(0.638973583422)*x[1]+(0.333562215313)*x[2]**o+(0.770257986138)*x[2]
        s[3,0]=(0.866403926159)*x[0]**o+(-0.567381195064)*x[0]+(-0.689200549972)*x[1]**o+(0.0278246156735)*x[1]+(-0.267522213281)*x[2]**o+(0.536314895418)*x[2]
        s[3,1]=(-0.561464080998)*x[0]**o+(-0.544070025131)*x[0]+(0.73146085432)*x[1]**o+(0.181244005603)*x[1]+(0.63008027542)*x[2]**o+(0.442771325928)*x[2]
        s[3,2]=(0.12906843064)*x[0]**o+(-0.736188526885)*x[0]+(0.675691533379)*x[1]**o+(0.975452480209)*x[1]+(0.877415310424)*x[2]**o+(-0.281346067495)*x[2]
        s[3,3]=(0.895295566131)*x[0]**o+(0.384502745469)*x[0]+(-0.6558585454)*x[1]**o+(-0.990764124757)*x[1]+(0.089950807223)*x[2]**o+(-0.117514839208)*x[2]
        s[3,4]=(0.908567029179)*x[0]**o+(0.130934285429)*x[0]+(-0.0913907665269)*x[1]**o+(-0.483606581717)*x[1]+(0.316018801782)*x[2]**o+(-0.371754102357)*x[2]
        ref[0,0]=(-0.169971719578)*x_ref[0]**o+(-0.490891054826)*x_ref[0]+(0.744482943882)*x_ref[1]**o+(0.353623823086)*x_ref[1]+(0.265501620199)*x_ref[2]**o+(-0.629048564143)*x_ref[2]
        ref[0,1]=(0.213635035052)*x_ref[0]**o+(0.584035034456)*x_ref[0]+(0.133486854141)*x_ref[1]**o+(-0.563257345568)*x_ref[1]+(-0.846099536223)*x_ref[2]**o+(-0.854728289866)*x_ref[2]
        ref[0,2]=(0.342184893432)*x_ref[0]**o+(0.515624951667)*x_ref[0]+(-0.136008144001)*x_ref[1]**o+(-0.280682548245)*x_ref[1]+(-0.927880146409)*x_ref[2]**o+(0.23737193095)*x_ref[2]
        ref[0,3]=(0.0547841640747)*x_ref[0]**o+(-0.969984349739)*x_ref[0]+(-0.155447158154)*x_ref[1]**o+(-0.495088348804)*x_ref[1]+(0.879915433385)*x_ref[2]**o+(0.438389550488)*x_ref[2]
        ref[0,4]=(0.15107258681)*x_ref[0]**o+(0.878036814589)*x_ref[0]+(0.875901664353)*x_ref[1]**o+(0.455383770416)*x_ref[1]+(0.316791478263)*x_ref[2]**o+(0.13991508459)*x_ref[2]
        ref[1,0]=(0.627219372033)*x_ref[0]**o+(0.0514202591569)*x_ref[0]+(0.696459316229)*x_ref[1]**o+(0.232864735057)*x_ref[1]+(0.516267189254)*x_ref[2]**o+(-0.277533957737)*x_ref[2]
        ref[1,1]=(-0.474165959949)*x_ref[0]**o+(0.875608881493)*x_ref[0]+(-0.432866681632)*x_ref[1]**o+(0.165933079742)*x_ref[1]+(-0.0745331372502)*x_ref[2]**o+(-0.0168977805625)*x_ref[2]
        ref[1,2]=(-0.228391157708)*x_ref[0]**o+(-0.294221312175)*x_ref[0]+(-0.121429334894)*x_ref[1]**o+(-0.557107244165)*x_ref[1]+(-0.97514025916)*x_ref[2]**o+(0.041970524119)*x_ref[2]
        ref[1,3]=(0.344978415951)*x_ref[0]**o+(0.319420517862)*x_ref[0]+(0.802947508962)*x_ref[1]**o+(0.169628742731)*x_ref[1]+(-0.380928178408)*x_ref[2]**o+(0.272551034104)*x_ref[2]
        ref[1,4]=(0.613832101706)*x_ref[0]**o+(-0.332306428069)*x_ref[0]+(0.02563364621)*x_ref[1]**o+(-0.415952850048)*x_ref[1]+(0.673350831947)*x_ref[2]**o+(0.713348183687)*x_ref[2]
        ref[2,0]=(0.290558527133)*x_ref[0]**o+(0.135335532216)*x_ref[0]+(-0.352459421266)*x_ref[1]**o+(-0.854183560051)*x_ref[1]+(0.581843223923)*x_ref[2]**o+(0.97378057053)*x_ref[2]
        ref[2,1]=(-0.328513503613)*x_ref[0]**o+(-0.205196873021)*x_ref[0]+(-0.910190961873)*x_ref[1]**o+(0.568820988101)*x_ref[1]+(0.495810061604)*x_ref[2]**o+(-0.761749884838)*x_ref[2]
        ref[2,2]=(-0.00783916993558)*x_ref[0]**o+(0.898831843878)*x_ref[0]+(0.26265892346)*x_ref[1]**o+(-0.654638473159)*x_ref[1]+(-0.713198105003)*x_ref[2]**o+(-0.59350752607)*x_ref[2]
        ref[2,3]=(-0.881651184273)*x_ref[0]**o+(-0.834213060843)*x_ref[0]+(-0.885142407575)*x_ref[1]**o+(0.206850906485)*x_ref[1]+(0.372454818886)*x_ref[2]**o+(0.159769678242)*x_ref[2]
        ref[2,4]=(-0.937735543761)*x_ref[0]**o+(0.529592561341)*x_ref[0]+(-0.936499401425)*x_ref[1]**o+(0.638973583422)*x_ref[1]+(0.333562215313)*x_ref[2]**o+(0.770257986138)*x_ref[2]
        ref[3,0]=(0.866403926159)*x_ref[0]**o+(-0.567381195064)*x_ref[0]+(-0.689200549972)*x_ref[1]**o+(0.0278246156735)*x_ref[1]+(-0.267522213281)*x_ref[2]**o+(0.536314895418)*x_ref[2]
        ref[3,1]=(-0.561464080998)*x_ref[0]**o+(-0.544070025131)*x_ref[0]+(0.73146085432)*x_ref[1]**o+(0.181244005603)*x_ref[1]+(0.63008027542)*x_ref[2]**o+(0.442771325928)*x_ref[2]
        ref[3,2]=(0.12906843064)*x_ref[0]**o+(-0.736188526885)*x_ref[0]+(0.675691533379)*x_ref[1]**o+(0.975452480209)*x_ref[1]+(0.877415310424)*x_ref[2]**o+(-0.281346067495)*x_ref[2]
        ref[3,3]=(0.895295566131)*x_ref[0]**o+(0.384502745469)*x_ref[0]+(-0.6558585454)*x_ref[1]**o+(-0.990764124757)*x_ref[1]+(0.089950807223)*x_ref[2]**o+(-0.117514839208)*x_ref[2]
        ref[3,4]=(0.908567029179)*x_ref[0]**o+(0.130934285429)*x_ref[0]+(-0.0913907665269)*x_ref[1]**o+(-0.483606581717)*x_ref[1]+(0.316018801782)*x_ref[2]**o+(-0.371754102357)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Function_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.922012175796)*x[0]**o+(-0.704323474251)*x[0]+(-0.772095928607)*x[1]**o+(-0.102522378969)*x[1]
        s[0,0,1]=(-0.756476066423)*x[0]**o+(0.476014645849)*x[0]+(0.746838535595)*x[1]**o+(-0.580996711866)*x[1]
        s[0,1,0]=(-0.372958624489)*x[0]**o+(-0.742768798844)*x[0]+(0.930234188365)*x[1]**o+(-0.934316477902)*x[1]
        s[0,1,1]=(-0.887657339887)*x[0]**o+(0.477506024138)*x[0]+(-0.072611336226)*x[1]**o+(-0.632380868263)*x[1]
        s[1,0,0]=(0.783838363367)*x[0]**o+(-0.231553922468)*x[0]+(0.747645980538)*x[1]**o+(0.480696124395)*x[1]
        s[1,0,1]=(0.916049259239)*x[0]**o+(-0.334020782789)*x[0]+(0.250380400053)*x[1]**o+(0.192676718299)*x[1]
        s[1,1,0]=(0.242999551485)*x[0]**o+(0.651169244253)*x[0]+(0.979851149428)*x[1]**o+(-0.551088471757)*x[1]
        s[1,1,1]=(0.998442854355)*x[0]**o+(-0.97084316076)*x[0]+(0.887766823857)*x[1]**o+(-0.115041414343)*x[1]
        s[2,0,0]=(0.0439895718917)*x[0]**o+(0.214997856989)*x[0]+(0.920934212621)*x[1]**o+(-0.202158287769)*x[1]
        s[2,0,1]=(-0.134517914198)*x[0]**o+(0.153074355415)*x[0]+(-0.459946249543)*x[1]**o+(0.635578282444)*x[1]
        s[2,1,0]=(-0.0910166438835)*x[0]**o+(0.802617171383)*x[0]+(0.247569766531)*x[1]**o+(0.951985690421)*x[1]
        s[2,1,1]=(0.264891721801)*x[0]**o+(-0.199605387174)*x[0]+(-0.833082379901)*x[1]**o+(0.721123597626)*x[1]
        s[3,0,0]=(-0.703356543249)*x[0]**o+(-0.146515334407)*x[0]+(-0.713251719666)*x[1]**o+(0.0740766553512)*x[1]
        s[3,0,1]=(0.623941214153)*x[0]**o+(-0.884052095344)*x[0]+(-0.976500944913)*x[1]**o+(0.985013528937)*x[1]
        s[3,1,0]=(0.794339208094)*x[0]**o+(-0.542029674325)*x[0]+(-0.427049948772)*x[1]**o+(0.37932594769)*x[1]
        s[3,1,1]=(0.563582163907)*x[0]**o+(-0.458198108954)*x[0]+(0.303347268915)*x[1]**o+(0.342283816391)*x[1]
        s[4,0,0]=(-0.717532671202)*x[0]**o+(-0.878860514987)*x[0]+(-0.0976608307102)*x[1]**o+(0.250567801094)*x[1]
        s[4,0,1]=(0.397375802895)*x[0]**o+(-0.212057004666)*x[0]+(0.558358501772)*x[1]**o+(-0.134232085341)*x[1]
        s[4,1,0]=(-0.460059564884)*x[0]**o+(-0.628865924343)*x[0]+(0.209325635948)*x[1]**o+(0.122601801028)*x[1]
        s[4,1,1]=(-0.528413992704)*x[0]**o+(-0.275278112162)*x[0]+(-0.904021647214)*x[1]**o+(0.171477127541)*x[1]
        s[5,0,0]=(-0.590272357875)*x[0]**o+(0.787036563397)*x[0]+(-0.821383663122)*x[1]**o+(-0.43696048132)*x[1]
        s[5,0,1]=(0.63774872195)*x[0]**o+(0.57629113566)*x[0]+(-0.860489105031)*x[1]**o+(0.548377309092)*x[1]
        s[5,1,0]=(0.584516519044)*x[0]**o+(0.292577411211)*x[0]+(-0.277291138474)*x[1]**o+(0.36529689442)*x[1]
        s[5,1,1]=(-0.58877604496)*x[0]**o+(-0.340786512436)*x[0]+(0.5949521732)*x[1]**o+(0.302898904668)*x[1]
        ref[0,0,0]=(0.922012175796)*x_ref[0]**o+(-0.704323474251)*x_ref[0]+(-0.772095928607)*x_ref[1]**o+(-0.102522378969)*x_ref[1]
        ref[0,0,1]=(-0.756476066423)*x_ref[0]**o+(0.476014645849)*x_ref[0]+(0.746838535595)*x_ref[1]**o+(-0.580996711866)*x_ref[1]
        ref[0,1,0]=(-0.372958624489)*x_ref[0]**o+(-0.742768798844)*x_ref[0]+(0.930234188365)*x_ref[1]**o+(-0.934316477902)*x_ref[1]
        ref[0,1,1]=(-0.887657339887)*x_ref[0]**o+(0.477506024138)*x_ref[0]+(-0.072611336226)*x_ref[1]**o+(-0.632380868263)*x_ref[1]
        ref[1,0,0]=(0.783838363367)*x_ref[0]**o+(-0.231553922468)*x_ref[0]+(0.747645980538)*x_ref[1]**o+(0.480696124395)*x_ref[1]
        ref[1,0,1]=(0.916049259239)*x_ref[0]**o+(-0.334020782789)*x_ref[0]+(0.250380400053)*x_ref[1]**o+(0.192676718299)*x_ref[1]
        ref[1,1,0]=(0.242999551485)*x_ref[0]**o+(0.651169244253)*x_ref[0]+(0.979851149428)*x_ref[1]**o+(-0.551088471757)*x_ref[1]
        ref[1,1,1]=(0.998442854355)*x_ref[0]**o+(-0.97084316076)*x_ref[0]+(0.887766823857)*x_ref[1]**o+(-0.115041414343)*x_ref[1]
        ref[2,0,0]=(0.0439895718917)*x_ref[0]**o+(0.214997856989)*x_ref[0]+(0.920934212621)*x_ref[1]**o+(-0.202158287769)*x_ref[1]
        ref[2,0,1]=(-0.134517914198)*x_ref[0]**o+(0.153074355415)*x_ref[0]+(-0.459946249543)*x_ref[1]**o+(0.635578282444)*x_ref[1]
        ref[2,1,0]=(-0.0910166438835)*x_ref[0]**o+(0.802617171383)*x_ref[0]+(0.247569766531)*x_ref[1]**o+(0.951985690421)*x_ref[1]
        ref[2,1,1]=(0.264891721801)*x_ref[0]**o+(-0.199605387174)*x_ref[0]+(-0.833082379901)*x_ref[1]**o+(0.721123597626)*x_ref[1]
        ref[3,0,0]=(-0.703356543249)*x_ref[0]**o+(-0.146515334407)*x_ref[0]+(-0.713251719666)*x_ref[1]**o+(0.0740766553512)*x_ref[1]
        ref[3,0,1]=(0.623941214153)*x_ref[0]**o+(-0.884052095344)*x_ref[0]+(-0.976500944913)*x_ref[1]**o+(0.985013528937)*x_ref[1]
        ref[3,1,0]=(0.794339208094)*x_ref[0]**o+(-0.542029674325)*x_ref[0]+(-0.427049948772)*x_ref[1]**o+(0.37932594769)*x_ref[1]
        ref[3,1,1]=(0.563582163907)*x_ref[0]**o+(-0.458198108954)*x_ref[0]+(0.303347268915)*x_ref[1]**o+(0.342283816391)*x_ref[1]
        ref[4,0,0]=(-0.717532671202)*x_ref[0]**o+(-0.878860514987)*x_ref[0]+(-0.0976608307102)*x_ref[1]**o+(0.250567801094)*x_ref[1]
        ref[4,0,1]=(0.397375802895)*x_ref[0]**o+(-0.212057004666)*x_ref[0]+(0.558358501772)*x_ref[1]**o+(-0.134232085341)*x_ref[1]
        ref[4,1,0]=(-0.460059564884)*x_ref[0]**o+(-0.628865924343)*x_ref[0]+(0.209325635948)*x_ref[1]**o+(0.122601801028)*x_ref[1]
        ref[4,1,1]=(-0.528413992704)*x_ref[0]**o+(-0.275278112162)*x_ref[0]+(-0.904021647214)*x_ref[1]**o+(0.171477127541)*x_ref[1]
        ref[5,0,0]=(-0.590272357875)*x_ref[0]**o+(0.787036563397)*x_ref[0]+(-0.821383663122)*x_ref[1]**o+(-0.43696048132)*x_ref[1]
        ref[5,0,1]=(0.63774872195)*x_ref[0]**o+(0.57629113566)*x_ref[0]+(-0.860489105031)*x_ref[1]**o+(0.548377309092)*x_ref[1]
        ref[5,1,0]=(0.584516519044)*x_ref[0]**o+(0.292577411211)*x_ref[0]+(-0.277291138474)*x_ref[1]**o+(0.36529689442)*x_ref[1]
        ref[5,1,1]=(-0.58877604496)*x_ref[0]**o+(-0.340786512436)*x_ref[0]+(0.5949521732)*x_ref[1]**o+(0.302898904668)*x_ref[1]
      else:
        s[0,0,0]=(0.103127120457)*x[0]**o+(-0.790648806335)*x[0]+(0.32847672023)*x[1]**o+(-0.690237185512)*x[1]+(-0.317184442145)*x[2]**o+(0.537453763779)*x[2]
        s[0,0,1]=(-0.36596240665)*x[0]**o+(0.256732865132)*x[0]+(-0.139947303843)*x[1]**o+(0.230496874435)*x[1]+(-0.616356953667)*x[2]**o+(-0.59441989496)*x[2]
        s[0,1,0]=(0.513873980476)*x[0]**o+(0.484045483315)*x[0]+(-0.273179211414)*x[1]**o+(-0.604138913893)*x[1]+(-0.818085004879)*x[2]**o+(-0.829509682818)*x[2]
        s[0,1,1]=(-0.797011406939)*x[0]**o+(-0.515647970266)*x[0]+(0.466120391094)*x[1]**o+(-0.996260216826)*x[1]+(-0.42533407671)*x[2]**o+(-0.081427515701)*x[2]
        s[1,0,0]=(0.840036174626)*x[0]**o+(-0.974272190945)*x[0]+(-0.912824754225)*x[1]**o+(-0.0576196265976)*x[1]+(0.311357427135)*x[2]**o+(0.0483209368492)*x[2]
        s[1,0,1]=(0.76482565945)*x[0]**o+(0.0613089443427)*x[0]+(0.907470294652)*x[1]**o+(0.720216160172)*x[1]+(-0.848264518892)*x[2]**o+(-0.622655911866)*x[2]
        s[1,1,0]=(-0.291088924515)*x[0]**o+(-0.273460845284)*x[0]+(-0.860316787156)*x[1]**o+(-0.117032009627)*x[1]+(0.0487955308094)*x[2]**o+(-0.701553638107)*x[2]
        s[1,1,1]=(-0.471933516494)*x[0]**o+(0.635615053447)*x[0]+(0.471737879913)*x[1]**o+(0.313165588781)*x[1]+(-0.15916085554)*x[2]**o+(-0.825577908132)*x[2]
        s[2,0,0]=(-0.0524254116203)*x[0]**o+(0.88748910061)*x[0]+(-0.522351516282)*x[1]**o+(-0.284852401017)*x[1]+(0.0158220147327)*x[2]**o+(0.00646469511963)*x[2]
        s[2,0,1]=(0.775455989914)*x[0]**o+(-0.592268899844)*x[0]+(-0.395459506696)*x[1]**o+(-0.778966655151)*x[1]+(-0.608812345651)*x[2]**o+(0.431749899976)*x[2]
        s[2,1,0]=(-0.472052989428)*x[0]**o+(-0.653779618462)*x[0]+(0.726973003838)*x[1]**o+(0.615103945489)*x[1]+(-0.635243574086)*x[2]**o+(-0.944283499648)*x[2]
        s[2,1,1]=(-0.538096393025)*x[0]**o+(-0.736062177106)*x[0]+(0.54984272471)*x[1]**o+(-0.252874828323)*x[1]+(0.520232358093)*x[2]**o+(0.658820792593)*x[2]
        s[3,0,0]=(0.0189624424339)*x[0]**o+(-0.928190186927)*x[0]+(-0.102190963399)*x[1]**o+(0.0985219931729)*x[1]+(0.964850306827)*x[2]**o+(-0.384616188398)*x[2]
        s[3,0,1]=(-0.122886102392)*x[0]**o+(0.916297286756)*x[0]+(0.0549430981199)*x[1]**o+(0.911993790356)*x[1]+(0.214659705881)*x[2]**o+(-0.69412668699)*x[2]
        s[3,1,0]=(0.902798669922)*x[0]**o+(-0.518265765234)*x[0]+(0.719822200918)*x[1]**o+(-0.962846306331)*x[1]+(-0.603216741191)*x[2]**o+(-0.225102184332)*x[2]
        s[3,1,1]=(0.947054397721)*x[0]**o+(0.799669807545)*x[0]+(-0.210074700167)*x[1]**o+(-0.776975337853)*x[1]+(0.750882895589)*x[2]**o+(-0.292157154938)*x[2]
        s[4,0,0]=(-0.00199580734612)*x[0]**o+(-0.894095448382)*x[0]+(-0.17192951285)*x[1]**o+(0.579608423454)*x[1]+(-0.947392406137)*x[2]**o+(-0.133887792274)*x[2]
        s[4,0,1]=(-0.395728778674)*x[0]**o+(-0.59395401109)*x[0]+(-0.334793568635)*x[1]**o+(0.848150466553)*x[1]+(0.449812215037)*x[2]**o+(-0.577515589907)*x[2]
        s[4,1,0]=(-0.815755197958)*x[0]**o+(-0.15839764311)*x[0]+(-0.829730261843)*x[1]**o+(0.891820791252)*x[1]+(0.903790315572)*x[2]**o+(-0.295122999627)*x[2]
        s[4,1,1]=(0.763626481603)*x[0]**o+(0.801670589231)*x[0]+(-0.779956910059)*x[1]**o+(0.593907194188)*x[1]+(-0.542116631989)*x[2]**o+(-0.845496315821)*x[2]
        s[5,0,0]=(-0.088706022005)*x[0]**o+(0.903197426259)*x[0]+(0.562811562755)*x[1]**o+(-0.999408930701)*x[1]+(0.647113624796)*x[2]**o+(0.917310785712)*x[2]
        s[5,0,1]=(-0.202743164028)*x[0]**o+(0.71872847025)*x[0]+(-0.709895205028)*x[1]**o+(0.294942879757)*x[1]+(-0.83533110878)*x[2]**o+(0.683966938727)*x[2]
        s[5,1,0]=(-0.0399613077171)*x[0]**o+(-0.504881528731)*x[0]+(-0.554451064799)*x[1]**o+(0.104714660691)*x[1]+(0.443770831652)*x[2]**o+(-0.673196707698)*x[2]
        s[5,1,1]=(0.993420509994)*x[0]**o+(0.717025596885)*x[0]+(-0.939900190744)*x[1]**o+(0.957571831696)*x[1]+(-0.712509953461)*x[2]**o+(0.150276310366)*x[2]
        ref[0,0,0]=(0.103127120457)*x_ref[0]**o+(-0.790648806335)*x_ref[0]+(0.32847672023)*x_ref[1]**o+(-0.690237185512)*x_ref[1]+(-0.317184442145)*x_ref[2]**o+(0.537453763779)*x_ref[2]
        ref[0,0,1]=(-0.36596240665)*x_ref[0]**o+(0.256732865132)*x_ref[0]+(-0.139947303843)*x_ref[1]**o+(0.230496874435)*x_ref[1]+(-0.616356953667)*x_ref[2]**o+(-0.59441989496)*x_ref[2]
        ref[0,1,0]=(0.513873980476)*x_ref[0]**o+(0.484045483315)*x_ref[0]+(-0.273179211414)*x_ref[1]**o+(-0.604138913893)*x_ref[1]+(-0.818085004879)*x_ref[2]**o+(-0.829509682818)*x_ref[2]
        ref[0,1,1]=(-0.797011406939)*x_ref[0]**o+(-0.515647970266)*x_ref[0]+(0.466120391094)*x_ref[1]**o+(-0.996260216826)*x_ref[1]+(-0.42533407671)*x_ref[2]**o+(-0.081427515701)*x_ref[2]
        ref[1,0,0]=(0.840036174626)*x_ref[0]**o+(-0.974272190945)*x_ref[0]+(-0.912824754225)*x_ref[1]**o+(-0.0576196265976)*x_ref[1]+(0.311357427135)*x_ref[2]**o+(0.0483209368492)*x_ref[2]
        ref[1,0,1]=(0.76482565945)*x_ref[0]**o+(0.0613089443427)*x_ref[0]+(0.907470294652)*x_ref[1]**o+(0.720216160172)*x_ref[1]+(-0.848264518892)*x_ref[2]**o+(-0.622655911866)*x_ref[2]
        ref[1,1,0]=(-0.291088924515)*x_ref[0]**o+(-0.273460845284)*x_ref[0]+(-0.860316787156)*x_ref[1]**o+(-0.117032009627)*x_ref[1]+(0.0487955308094)*x_ref[2]**o+(-0.701553638107)*x_ref[2]
        ref[1,1,1]=(-0.471933516494)*x_ref[0]**o+(0.635615053447)*x_ref[0]+(0.471737879913)*x_ref[1]**o+(0.313165588781)*x_ref[1]+(-0.15916085554)*x_ref[2]**o+(-0.825577908132)*x_ref[2]
        ref[2,0,0]=(-0.0524254116203)*x_ref[0]**o+(0.88748910061)*x_ref[0]+(-0.522351516282)*x_ref[1]**o+(-0.284852401017)*x_ref[1]+(0.0158220147327)*x_ref[2]**o+(0.00646469511963)*x_ref[2]
        ref[2,0,1]=(0.775455989914)*x_ref[0]**o+(-0.592268899844)*x_ref[0]+(-0.395459506696)*x_ref[1]**o+(-0.778966655151)*x_ref[1]+(-0.608812345651)*x_ref[2]**o+(0.431749899976)*x_ref[2]
        ref[2,1,0]=(-0.472052989428)*x_ref[0]**o+(-0.653779618462)*x_ref[0]+(0.726973003838)*x_ref[1]**o+(0.615103945489)*x_ref[1]+(-0.635243574086)*x_ref[2]**o+(-0.944283499648)*x_ref[2]
        ref[2,1,1]=(-0.538096393025)*x_ref[0]**o+(-0.736062177106)*x_ref[0]+(0.54984272471)*x_ref[1]**o+(-0.252874828323)*x_ref[1]+(0.520232358093)*x_ref[2]**o+(0.658820792593)*x_ref[2]
        ref[3,0,0]=(0.0189624424339)*x_ref[0]**o+(-0.928190186927)*x_ref[0]+(-0.102190963399)*x_ref[1]**o+(0.0985219931729)*x_ref[1]+(0.964850306827)*x_ref[2]**o+(-0.384616188398)*x_ref[2]
        ref[3,0,1]=(-0.122886102392)*x_ref[0]**o+(0.916297286756)*x_ref[0]+(0.0549430981199)*x_ref[1]**o+(0.911993790356)*x_ref[1]+(0.214659705881)*x_ref[2]**o+(-0.69412668699)*x_ref[2]
        ref[3,1,0]=(0.902798669922)*x_ref[0]**o+(-0.518265765234)*x_ref[0]+(0.719822200918)*x_ref[1]**o+(-0.962846306331)*x_ref[1]+(-0.603216741191)*x_ref[2]**o+(-0.225102184332)*x_ref[2]
        ref[3,1,1]=(0.947054397721)*x_ref[0]**o+(0.799669807545)*x_ref[0]+(-0.210074700167)*x_ref[1]**o+(-0.776975337853)*x_ref[1]+(0.750882895589)*x_ref[2]**o+(-0.292157154938)*x_ref[2]
        ref[4,0,0]=(-0.00199580734612)*x_ref[0]**o+(-0.894095448382)*x_ref[0]+(-0.17192951285)*x_ref[1]**o+(0.579608423454)*x_ref[1]+(-0.947392406137)*x_ref[2]**o+(-0.133887792274)*x_ref[2]
        ref[4,0,1]=(-0.395728778674)*x_ref[0]**o+(-0.59395401109)*x_ref[0]+(-0.334793568635)*x_ref[1]**o+(0.848150466553)*x_ref[1]+(0.449812215037)*x_ref[2]**o+(-0.577515589907)*x_ref[2]
        ref[4,1,0]=(-0.815755197958)*x_ref[0]**o+(-0.15839764311)*x_ref[0]+(-0.829730261843)*x_ref[1]**o+(0.891820791252)*x_ref[1]+(0.903790315572)*x_ref[2]**o+(-0.295122999627)*x_ref[2]
        ref[4,1,1]=(0.763626481603)*x_ref[0]**o+(0.801670589231)*x_ref[0]+(-0.779956910059)*x_ref[1]**o+(0.593907194188)*x_ref[1]+(-0.542116631989)*x_ref[2]**o+(-0.845496315821)*x_ref[2]
        ref[5,0,0]=(-0.088706022005)*x_ref[0]**o+(0.903197426259)*x_ref[0]+(0.562811562755)*x_ref[1]**o+(-0.999408930701)*x_ref[1]+(0.647113624796)*x_ref[2]**o+(0.917310785712)*x_ref[2]
        ref[5,0,1]=(-0.202743164028)*x_ref[0]**o+(0.71872847025)*x_ref[0]+(-0.709895205028)*x_ref[1]**o+(0.294942879757)*x_ref[1]+(-0.83533110878)*x_ref[2]**o+(0.683966938727)*x_ref[2]
        ref[5,1,0]=(-0.0399613077171)*x_ref[0]**o+(-0.504881528731)*x_ref[0]+(-0.554451064799)*x_ref[1]**o+(0.104714660691)*x_ref[1]+(0.443770831652)*x_ref[2]**o+(-0.673196707698)*x_ref[2]
        ref[5,1,1]=(0.993420509994)*x_ref[0]**o+(0.717025596885)*x_ref[0]+(-0.939900190744)*x_ref[1]**o+(0.957571831696)*x_ref[1]+(-0.712509953461)*x_ref[2]**o+(0.150276310366)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunction_fromSymbol_Function_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Function

      assumptions: self.domain supports inpterpolation from Function onto Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Function(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.889687592166)*x[0]**o+(-0.244091857538)*x[0]+(-0.218564463727)*x[1]**o+(-0.937124237311)*x[1]
        s[0,0,0,1]=(0.440232360433)*x[0]**o+(0.441448419004)*x[0]+(-0.469778033804)*x[1]**o+(0.561615436148)*x[1]
        s[0,0,1,0]=(-0.704681295174)*x[0]**o+(-0.698896667728)*x[0]+(-0.999233531685)*x[1]**o+(0.287081983675)*x[1]
        s[0,0,1,1]=(0.899708232838)*x[0]**o+(0.521718791408)*x[0]+(-0.225938242874)*x[1]**o+(0.150056519691)*x[1]
        s[0,0,2,0]=(-0.00550720503612)*x[0]**o+(-0.521199335637)*x[0]+(0.943701370941)*x[1]**o+(-0.285282621044)*x[1]
        s[0,0,2,1]=(0.758068587203)*x[0]**o+(-0.247932113502)*x[0]+(-0.441683356848)*x[1]**o+(0.598148360435)*x[1]
        s[0,1,0,0]=(0.252073761781)*x[0]**o+(0.803649353117)*x[0]+(0.939591428513)*x[1]**o+(0.254473527243)*x[1]
        s[0,1,0,1]=(0.652883529471)*x[0]**o+(0.832856006165)*x[0]+(-0.754806656391)*x[1]**o+(0.477988652237)*x[1]
        s[0,1,1,0]=(0.256691612427)*x[0]**o+(-0.827547687439)*x[0]+(-0.426446760113)*x[1]**o+(-0.709508780384)*x[1]
        s[0,1,1,1]=(-0.66459132575)*x[0]**o+(0.82036954645)*x[0]+(-0.154864164844)*x[1]**o+(0.967900851407)*x[1]
        s[0,1,2,0]=(0.82202922242)*x[0]**o+(-0.564876598949)*x[0]+(0.842505409303)*x[1]**o+(-0.710636181332)*x[1]
        s[0,1,2,1]=(0.83881620922)*x[0]**o+(-0.237072167776)*x[0]+(-0.442181682029)*x[1]**o+(0.771550454195)*x[1]
        s[0,2,0,0]=(0.474850766572)*x[0]**o+(-0.857346378644)*x[0]+(0.919175013993)*x[1]**o+(0.629995461581)*x[1]
        s[0,2,0,1]=(0.523096282888)*x[0]**o+(-0.510665736272)*x[0]+(0.145171493458)*x[1]**o+(-0.103513634427)*x[1]
        s[0,2,1,0]=(-0.423486426499)*x[0]**o+(-0.49703927479)*x[0]+(0.174127212691)*x[1]**o+(0.217944977415)*x[1]
        s[0,2,1,1]=(-0.631505727799)*x[0]**o+(0.588807237677)*x[0]+(0.668437404223)*x[1]**o+(0.552798016144)*x[1]
        s[0,2,2,0]=(0.734466686613)*x[0]**o+(0.142930362034)*x[0]+(0.490607179637)*x[1]**o+(0.0445798411208)*x[1]
        s[0,2,2,1]=(0.59894699236)*x[0]**o+(-0.80614241159)*x[0]+(-0.966344783722)*x[1]**o+(-0.537851881337)*x[1]
        s[0,3,0,0]=(0.635428545228)*x[0]**o+(0.458465797988)*x[0]+(0.431015659184)*x[1]**o+(-0.59835525481)*x[1]
        s[0,3,0,1]=(0.0494865211965)*x[0]**o+(-0.776500412974)*x[0]+(0.932578121115)*x[1]**o+(0.914133479331)*x[1]
        s[0,3,1,0]=(0.231396476054)*x[0]**o+(-0.265428688952)*x[0]+(-0.344497302174)*x[1]**o+(-0.757106502174)*x[1]
        s[0,3,1,1]=(0.551883714399)*x[0]**o+(0.832451314453)*x[0]+(-0.899905963729)*x[1]**o+(-0.79240620147)*x[1]
        s[0,3,2,0]=(0.564173236354)*x[0]**o+(-0.218691242318)*x[0]+(0.974498273818)*x[1]**o+(-0.712174073532)*x[1]
        s[0,3,2,1]=(-0.756975354573)*x[0]**o+(0.99073177566)*x[0]+(-0.680264934169)*x[1]**o+(-0.551951700356)*x[1]
        s[0,4,0,0]=(0.342204785324)*x[0]**o+(-0.332826079413)*x[0]+(0.827106843714)*x[1]**o+(0.147296550674)*x[1]
        s[0,4,0,1]=(0.718074130442)*x[0]**o+(0.360060294518)*x[0]+(-0.751135064333)*x[1]**o+(0.217154371958)*x[1]
        s[0,4,1,0]=(-0.930762040509)*x[0]**o+(0.368919544302)*x[0]+(0.883523093869)*x[1]**o+(-0.904151741067)*x[1]
        s[0,4,1,1]=(0.518896572101)*x[0]**o+(-0.342606634816)*x[0]+(-0.129850163393)*x[1]**o+(0.633244155895)*x[1]
        s[0,4,2,0]=(0.392116617423)*x[0]**o+(-0.173211325829)*x[0]+(0.400087049088)*x[1]**o+(0.7891166848)*x[1]
        s[0,4,2,1]=(0.716606140961)*x[0]**o+(0.496596392306)*x[0]+(-0.671375280598)*x[1]**o+(-0.528611363434)*x[1]
        s[1,0,0,0]=(0.618225203091)*x[0]**o+(0.384959538542)*x[0]+(-0.0428815543398)*x[1]**o+(0.161091350961)*x[1]
        s[1,0,0,1]=(0.443417981463)*x[0]**o+(-0.162056175957)*x[0]+(0.207624970868)*x[1]**o+(-0.121433587807)*x[1]
        s[1,0,1,0]=(0.776131229022)*x[0]**o+(-0.871658190605)*x[0]+(0.54507490858)*x[1]**o+(0.159672390575)*x[1]
        s[1,0,1,1]=(-0.361120028772)*x[0]**o+(-0.406707123035)*x[0]+(-0.54341861971)*x[1]**o+(-0.0392836956005)*x[1]
        s[1,0,2,0]=(0.944116160979)*x[0]**o+(0.0890185721642)*x[0]+(0.73533144751)*x[1]**o+(0.701431910124)*x[1]
        s[1,0,2,1]=(0.722492883973)*x[0]**o+(0.351843053475)*x[0]+(-0.1756241753)*x[1]**o+(-0.703690677653)*x[1]
        s[1,1,0,0]=(0.262345197739)*x[0]**o+(0.683913892378)*x[0]+(0.322362020764)*x[1]**o+(0.851902848166)*x[1]
        s[1,1,0,1]=(-0.0509963903106)*x[0]**o+(0.387065915322)*x[0]+(0.0727012685343)*x[1]**o+(0.666978314077)*x[1]
        s[1,1,1,0]=(-0.518122762009)*x[0]**o+(0.69514335574)*x[0]+(0.34513236897)*x[1]**o+(0.641034103704)*x[1]
        s[1,1,1,1]=(0.0658005558873)*x[0]**o+(-0.305147050226)*x[0]+(-0.707909968597)*x[1]**o+(0.0682220439272)*x[1]
        s[1,1,2,0]=(-0.443540005123)*x[0]**o+(0.179307931754)*x[0]+(0.982374202912)*x[1]**o+(0.398796141179)*x[1]
        s[1,1,2,1]=(-0.52173751599)*x[0]**o+(-0.523741826056)*x[0]+(0.172601588482)*x[1]**o+(0.969133917994)*x[1]
        s[1,2,0,0]=(-0.807607635564)*x[0]**o+(-0.525440627018)*x[0]+(-0.868823879608)*x[1]**o+(0.913203387198)*x[1]
        s[1,2,0,1]=(0.84988195907)*x[0]**o+(-0.394750851572)*x[0]+(0.740132509878)*x[1]**o+(0.855925902936)*x[1]
        s[1,2,1,0]=(0.055360229349)*x[0]**o+(0.299990998824)*x[0]+(0.558216562537)*x[1]**o+(-0.382651884902)*x[1]
        s[1,2,1,1]=(0.423603813726)*x[0]**o+(-0.907945615918)*x[0]+(-0.117306295879)*x[1]**o+(-0.0502458178048)*x[1]
        s[1,2,2,0]=(0.06196107837)*x[0]**o+(-0.612619509899)*x[0]+(0.169179798359)*x[1]**o+(-0.27854976134)*x[1]
        s[1,2,2,1]=(-0.506388098563)*x[0]**o+(0.125008314945)*x[0]+(0.509571389625)*x[1]**o+(0.442027080754)*x[1]
        s[1,3,0,0]=(0.304208463979)*x[0]**o+(-0.379194737667)*x[0]+(0.978868674946)*x[1]**o+(0.011573322151)*x[1]
        s[1,3,0,1]=(-0.151986748739)*x[0]**o+(0.153593063241)*x[0]+(0.644694257244)*x[1]**o+(0.414073745353)*x[1]
        s[1,3,1,0]=(0.924109632511)*x[0]**o+(0.398532810155)*x[0]+(-0.321086484535)*x[1]**o+(0.0802378501316)*x[1]
        s[1,3,1,1]=(-0.175639876422)*x[0]**o+(-0.109223049695)*x[0]+(-0.206450491082)*x[1]**o+(0.462124555092)*x[1]
        s[1,3,2,0]=(0.379156310804)*x[0]**o+(0.629232237742)*x[0]+(0.0918877201433)*x[1]**o+(-0.0490282902402)*x[1]
        s[1,3,2,1]=(-0.616126549785)*x[0]**o+(0.528823818571)*x[0]+(-0.947449800384)*x[1]**o+(0.145045253788)*x[1]
        s[1,4,0,0]=(-0.623165695489)*x[0]**o+(-0.861583499935)*x[0]+(0.219861821844)*x[1]**o+(-0.962135814763)*x[1]
        s[1,4,0,1]=(0.467875500099)*x[0]**o+(0.906948957958)*x[0]+(0.606958286586)*x[1]**o+(-0.423219122678)*x[1]
        s[1,4,1,0]=(0.200880358483)*x[0]**o+(0.571730447827)*x[0]+(0.908300268754)*x[1]**o+(0.19978633475)*x[1]
        s[1,4,1,1]=(0.55118731313)*x[0]**o+(-0.89903633572)*x[0]+(0.43836310821)*x[1]**o+(0.404697808939)*x[1]
        s[1,4,2,0]=(0.536587233299)*x[0]**o+(-0.0789531097738)*x[0]+(-0.775848629826)*x[1]**o+(-0.11020318415)*x[1]
        s[1,4,2,1]=(-0.648091804458)*x[0]**o+(-0.0282126867936)*x[0]+(0.0182411167897)*x[1]**o+(0.922414038295)*x[1]
        s[2,0,0,0]=(0.67385583226)*x[0]**o+(-0.993060381925)*x[0]+(0.731462156889)*x[1]**o+(0.183718928132)*x[1]
        s[2,0,0,1]=(0.881124628544)*x[0]**o+(0.828720578546)*x[0]+(-0.324258860918)*x[1]**o+(-0.836625964944)*x[1]
        s[2,0,1,0]=(0.944307497546)*x[0]**o+(0.546752005636)*x[0]+(0.334176020601)*x[1]**o+(0.654940705861)*x[1]
        s[2,0,1,1]=(-0.414075359332)*x[0]**o+(-0.782021485027)*x[0]+(-0.233424028825)*x[1]**o+(0.652900422088)*x[1]
        s[2,0,2,0]=(0.355942393341)*x[0]**o+(-0.727957138948)*x[0]+(-0.924406373726)*x[1]**o+(-0.806178313048)*x[1]
        s[2,0,2,1]=(-0.791343585742)*x[0]**o+(-0.61722808774)*x[0]+(-0.627757910074)*x[1]**o+(0.838801750555)*x[1]
        s[2,1,0,0]=(-0.866831381931)*x[0]**o+(-0.00498093548417)*x[0]+(-0.568387329284)*x[1]**o+(-0.428222759011)*x[1]
        s[2,1,0,1]=(-0.245588724312)*x[0]**o+(0.571669044307)*x[0]+(0.832968046978)*x[1]**o+(-0.0595908787599)*x[1]
        s[2,1,1,0]=(0.454631203504)*x[0]**o+(-0.827357560144)*x[0]+(0.876484630903)*x[1]**o+(0.316668730296)*x[1]
        s[2,1,1,1]=(-0.802326431878)*x[0]**o+(-0.547113643585)*x[0]+(0.619358075866)*x[1]**o+(0.387898635274)*x[1]
        s[2,1,2,0]=(0.204730319481)*x[0]**o+(0.462536054065)*x[0]+(-0.267050942194)*x[1]**o+(-0.529151462961)*x[1]
        s[2,1,2,1]=(-0.354690492675)*x[0]**o+(0.518914850203)*x[0]+(-0.283765250756)*x[1]**o+(0.595912311106)*x[1]
        s[2,2,0,0]=(0.329350522628)*x[0]**o+(0.689800893689)*x[0]+(-0.316097521431)*x[1]**o+(0.567054800242)*x[1]
        s[2,2,0,1]=(-0.68603272612)*x[0]**o+(-0.458451752241)*x[0]+(0.442247854284)*x[1]**o+(0.342266856887)*x[1]
        s[2,2,1,0]=(0.0562781380442)*x[0]**o+(0.615293898746)*x[0]+(-0.446430775323)*x[1]**o+(-0.720310464006)*x[1]
        s[2,2,1,1]=(-0.620026567808)*x[0]**o+(0.64935057185)*x[0]+(-0.402630635809)*x[1]**o+(0.984397383657)*x[1]
        s[2,2,2,0]=(-0.565466963253)*x[0]**o+(-0.983283191302)*x[0]+(0.0517997901365)*x[1]**o+(-0.00459375732072)*x[1]
        s[2,2,2,1]=(-0.544844126966)*x[0]**o+(-0.993222368569)*x[0]+(-0.0761437907904)*x[1]**o+(0.959202974542)*x[1]
        s[2,3,0,0]=(-0.542176217992)*x[0]**o+(-0.210550241274)*x[0]+(0.447732450287)*x[1]**o+(-0.0802692719358)*x[1]
        s[2,3,0,1]=(0.339422485349)*x[0]**o+(0.581409183541)*x[0]+(-0.255731595248)*x[1]**o+(-0.196957128507)*x[1]
        s[2,3,1,0]=(0.938606287977)*x[0]**o+(0.396408957375)*x[0]+(-0.53613392333)*x[1]**o+(-0.201412967427)*x[1]
        s[2,3,1,1]=(0.482931877644)*x[0]**o+(-0.395740204538)*x[0]+(-0.816216426797)*x[1]**o+(-0.390149600366)*x[1]
        s[2,3,2,0]=(0.507838191989)*x[0]**o+(0.590218664943)*x[0]+(0.987052963874)*x[1]**o+(-0.39735585208)*x[1]
        s[2,3,2,1]=(-0.036562453009)*x[0]**o+(0.299852609043)*x[0]+(0.194160324844)*x[1]**o+(-0.629803094543)*x[1]
        s[2,4,0,0]=(0.128664458677)*x[0]**o+(0.956037220714)*x[0]+(0.770331226849)*x[1]**o+(-0.208007445309)*x[1]
        s[2,4,0,1]=(0.526755720598)*x[0]**o+(-0.334650615987)*x[0]+(-0.245693030991)*x[1]**o+(-0.712134261554)*x[1]
        s[2,4,1,0]=(0.827770360865)*x[0]**o+(0.982906973774)*x[0]+(0.471438390576)*x[1]**o+(0.691215965917)*x[1]
        s[2,4,1,1]=(-0.206715480236)*x[0]**o+(0.488812089863)*x[0]+(0.777672390635)*x[1]**o+(-0.496681850755)*x[1]
        s[2,4,2,0]=(0.183938264875)*x[0]**o+(-0.247363713904)*x[0]+(0.962928810844)*x[1]**o+(0.642706194018)*x[1]
        s[2,4,2,1]=(0.758025105584)*x[0]**o+(-0.484793869316)*x[0]+(-0.431089127161)*x[1]**o+(0.127528834965)*x[1]
        s[3,0,0,0]=(-0.130978215422)*x[0]**o+(0.688819586128)*x[0]+(0.517162828175)*x[1]**o+(-0.979632247431)*x[1]
        s[3,0,0,1]=(0.130939629053)*x[0]**o+(0.685452224086)*x[0]+(-0.438454293093)*x[1]**o+(-0.727705397967)*x[1]
        s[3,0,1,0]=(-0.31299808275)*x[0]**o+(0.266897949461)*x[0]+(0.117566429054)*x[1]**o+(0.536075754797)*x[1]
        s[3,0,1,1]=(-0.401764430381)*x[0]**o+(-0.205983383212)*x[0]+(0.461300727769)*x[1]**o+(-0.580777919325)*x[1]
        s[3,0,2,0]=(-0.63526664417)*x[0]**o+(-0.878687700719)*x[0]+(0.104141198375)*x[1]**o+(-0.539520033395)*x[1]
        s[3,0,2,1]=(-0.592054773422)*x[0]**o+(-0.10222451991)*x[0]+(-0.385428960408)*x[1]**o+(-0.497192270367)*x[1]
        s[3,1,0,0]=(0.17159987373)*x[0]**o+(-0.215728352568)*x[0]+(0.188842078024)*x[1]**o+(0.417647732421)*x[1]
        s[3,1,0,1]=(0.548015992401)*x[0]**o+(-0.669157307443)*x[0]+(-0.796467777561)*x[1]**o+(0.163833145799)*x[1]
        s[3,1,1,0]=(-0.643762450865)*x[0]**o+(0.548750992658)*x[0]+(0.409211186464)*x[1]**o+(-0.986601038131)*x[1]
        s[3,1,1,1]=(-0.882340195766)*x[0]**o+(-0.185979117217)*x[0]+(-0.00826877676749)*x[1]**o+(0.0657705611536)*x[1]
        s[3,1,2,0]=(-0.0644145920386)*x[0]**o+(0.18740710155)*x[0]+(-0.690647279595)*x[1]**o+(-0.185556727699)*x[1]
        s[3,1,2,1]=(0.114366852618)*x[0]**o+(-0.388821787469)*x[0]+(-0.195241613331)*x[1]**o+(0.871862920816)*x[1]
        s[3,2,0,0]=(-0.673828733159)*x[0]**o+(-0.172601630353)*x[0]+(-0.488329313517)*x[1]**o+(0.0809307315053)*x[1]
        s[3,2,0,1]=(0.565273426718)*x[0]**o+(-0.610423824924)*x[0]+(0.856623167832)*x[1]**o+(-0.148845473418)*x[1]
        s[3,2,1,0]=(-0.3472751889)*x[0]**o+(0.908614631864)*x[0]+(0.445654124992)*x[1]**o+(0.719429313439)*x[1]
        s[3,2,1,1]=(0.24891832441)*x[0]**o+(-0.650081727005)*x[0]+(0.0980208592314)*x[1]**o+(0.575231534075)*x[1]
        s[3,2,2,0]=(-0.737935474986)*x[0]**o+(-0.874708935739)*x[0]+(0.856037958766)*x[1]**o+(-0.420041922056)*x[1]
        s[3,2,2,1]=(0.106345127894)*x[0]**o+(0.363060447152)*x[0]+(-0.472262909157)*x[1]**o+(-0.231475435242)*x[1]
        s[3,3,0,0]=(-0.3699135022)*x[0]**o+(-0.167427922355)*x[0]+(-0.911307660539)*x[1]**o+(-0.445726901976)*x[1]
        s[3,3,0,1]=(0.505339735087)*x[0]**o+(0.410541984072)*x[0]+(0.847415917938)*x[1]**o+(0.434325090295)*x[1]
        s[3,3,1,0]=(0.2589814894)*x[0]**o+(-0.087780879445)*x[0]+(0.847904119817)*x[1]**o+(0.462585909981)*x[1]
        s[3,3,1,1]=(0.283347628256)*x[0]**o+(0.637029684847)*x[0]+(0.174554376017)*x[1]**o+(0.642427538978)*x[1]
        s[3,3,2,0]=(-0.599348189011)*x[0]**o+(0.087002991683)*x[0]+(0.823297612508)*x[1]**o+(0.84180029964)*x[1]
        s[3,3,2,1]=(0.846090423225)*x[0]**o+(-0.465062745856)*x[0]+(0.498670439686)*x[1]**o+(0.721831089205)*x[1]
        s[3,4,0,0]=(0.0344263713162)*x[0]**o+(0.789039070423)*x[0]+(-0.518898476617)*x[1]**o+(0.112666233369)*x[1]
        s[3,4,0,1]=(0.752855926629)*x[0]**o+(-0.028671093428)*x[0]+(0.457225098582)*x[1]**o+(0.311273673822)*x[1]
        s[3,4,1,0]=(-0.623675057135)*x[0]**o+(-0.19730173917)*x[0]+(-0.328378661061)*x[1]**o+(-0.196083495589)*x[1]
        s[3,4,1,1]=(0.722548420108)*x[0]**o+(0.471733735786)*x[0]+(-0.151931791011)*x[1]**o+(0.0147226290316)*x[1]
        s[3,4,2,0]=(0.512232160342)*x[0]**o+(-0.339879887275)*x[0]+(-0.261940015624)*x[1]**o+(-0.589559690156)*x[1]
        s[3,4,2,1]=(0.387751869794)*x[0]**o+(0.284749443635)*x[0]+(0.764067938503)*x[1]**o+(0.20999603354)*x[1]
        ref[0,0,0,0]=(-0.889687592166)*x_ref[0]**o+(-0.244091857538)*x_ref[0]+(-0.218564463727)*x_ref[1]**o+(-0.937124237311)*x_ref[1]
        ref[0,0,0,1]=(0.440232360433)*x_ref[0]**o+(0.441448419004)*x_ref[0]+(-0.469778033804)*x_ref[1]**o+(0.561615436148)*x_ref[1]
        ref[0,0,1,0]=(-0.704681295174)*x_ref[0]**o+(-0.698896667728)*x_ref[0]+(-0.999233531685)*x_ref[1]**o+(0.287081983675)*x_ref[1]
        ref[0,0,1,1]=(0.899708232838)*x_ref[0]**o+(0.521718791408)*x_ref[0]+(-0.225938242874)*x_ref[1]**o+(0.150056519691)*x_ref[1]
        ref[0,0,2,0]=(-0.00550720503612)*x_ref[0]**o+(-0.521199335637)*x_ref[0]+(0.943701370941)*x_ref[1]**o+(-0.285282621044)*x_ref[1]
        ref[0,0,2,1]=(0.758068587203)*x_ref[0]**o+(-0.247932113502)*x_ref[0]+(-0.441683356848)*x_ref[1]**o+(0.598148360435)*x_ref[1]
        ref[0,1,0,0]=(0.252073761781)*x_ref[0]**o+(0.803649353117)*x_ref[0]+(0.939591428513)*x_ref[1]**o+(0.254473527243)*x_ref[1]
        ref[0,1,0,1]=(0.652883529471)*x_ref[0]**o+(0.832856006165)*x_ref[0]+(-0.754806656391)*x_ref[1]**o+(0.477988652237)*x_ref[1]
        ref[0,1,1,0]=(0.256691612427)*x_ref[0]**o+(-0.827547687439)*x_ref[0]+(-0.426446760113)*x_ref[1]**o+(-0.709508780384)*x_ref[1]
        ref[0,1,1,1]=(-0.66459132575)*x_ref[0]**o+(0.82036954645)*x_ref[0]+(-0.154864164844)*x_ref[1]**o+(0.967900851407)*x_ref[1]
        ref[0,1,2,0]=(0.82202922242)*x_ref[0]**o+(-0.564876598949)*x_ref[0]+(0.842505409303)*x_ref[1]**o+(-0.710636181332)*x_ref[1]
        ref[0,1,2,1]=(0.83881620922)*x_ref[0]**o+(-0.237072167776)*x_ref[0]+(-0.442181682029)*x_ref[1]**o+(0.771550454195)*x_ref[1]
        ref[0,2,0,0]=(0.474850766572)*x_ref[0]**o+(-0.857346378644)*x_ref[0]+(0.919175013993)*x_ref[1]**o+(0.629995461581)*x_ref[1]
        ref[0,2,0,1]=(0.523096282888)*x_ref[0]**o+(-0.510665736272)*x_ref[0]+(0.145171493458)*x_ref[1]**o+(-0.103513634427)*x_ref[1]
        ref[0,2,1,0]=(-0.423486426499)*x_ref[0]**o+(-0.49703927479)*x_ref[0]+(0.174127212691)*x_ref[1]**o+(0.217944977415)*x_ref[1]
        ref[0,2,1,1]=(-0.631505727799)*x_ref[0]**o+(0.588807237677)*x_ref[0]+(0.668437404223)*x_ref[1]**o+(0.552798016144)*x_ref[1]
        ref[0,2,2,0]=(0.734466686613)*x_ref[0]**o+(0.142930362034)*x_ref[0]+(0.490607179637)*x_ref[1]**o+(0.0445798411208)*x_ref[1]
        ref[0,2,2,1]=(0.59894699236)*x_ref[0]**o+(-0.80614241159)*x_ref[0]+(-0.966344783722)*x_ref[1]**o+(-0.537851881337)*x_ref[1]
        ref[0,3,0,0]=(0.635428545228)*x_ref[0]**o+(0.458465797988)*x_ref[0]+(0.431015659184)*x_ref[1]**o+(-0.59835525481)*x_ref[1]
        ref[0,3,0,1]=(0.0494865211965)*x_ref[0]**o+(-0.776500412974)*x_ref[0]+(0.932578121115)*x_ref[1]**o+(0.914133479331)*x_ref[1]
        ref[0,3,1,0]=(0.231396476054)*x_ref[0]**o+(-0.265428688952)*x_ref[0]+(-0.344497302174)*x_ref[1]**o+(-0.757106502174)*x_ref[1]
        ref[0,3,1,1]=(0.551883714399)*x_ref[0]**o+(0.832451314453)*x_ref[0]+(-0.899905963729)*x_ref[1]**o+(-0.79240620147)*x_ref[1]
        ref[0,3,2,0]=(0.564173236354)*x_ref[0]**o+(-0.218691242318)*x_ref[0]+(0.974498273818)*x_ref[1]**o+(-0.712174073532)*x_ref[1]
        ref[0,3,2,1]=(-0.756975354573)*x_ref[0]**o+(0.99073177566)*x_ref[0]+(-0.680264934169)*x_ref[1]**o+(-0.551951700356)*x_ref[1]
        ref[0,4,0,0]=(0.342204785324)*x_ref[0]**o+(-0.332826079413)*x_ref[0]+(0.827106843714)*x_ref[1]**o+(0.147296550674)*x_ref[1]
        ref[0,4,0,1]=(0.718074130442)*x_ref[0]**o+(0.360060294518)*x_ref[0]+(-0.751135064333)*x_ref[1]**o+(0.217154371958)*x_ref[1]
        ref[0,4,1,0]=(-0.930762040509)*x_ref[0]**o+(0.368919544302)*x_ref[0]+(0.883523093869)*x_ref[1]**o+(-0.904151741067)*x_ref[1]
        ref[0,4,1,1]=(0.518896572101)*x_ref[0]**o+(-0.342606634816)*x_ref[0]+(-0.129850163393)*x_ref[1]**o+(0.633244155895)*x_ref[1]
        ref[0,4,2,0]=(0.392116617423)*x_ref[0]**o+(-0.173211325829)*x_ref[0]+(0.400087049088)*x_ref[1]**o+(0.7891166848)*x_ref[1]
        ref[0,4,2,1]=(0.716606140961)*x_ref[0]**o+(0.496596392306)*x_ref[0]+(-0.671375280598)*x_ref[1]**o+(-0.528611363434)*x_ref[1]
        ref[1,0,0,0]=(0.618225203091)*x_ref[0]**o+(0.384959538542)*x_ref[0]+(-0.0428815543398)*x_ref[1]**o+(0.161091350961)*x_ref[1]
        ref[1,0,0,1]=(0.443417981463)*x_ref[0]**o+(-0.162056175957)*x_ref[0]+(0.207624970868)*x_ref[1]**o+(-0.121433587807)*x_ref[1]
        ref[1,0,1,0]=(0.776131229022)*x_ref[0]**o+(-0.871658190605)*x_ref[0]+(0.54507490858)*x_ref[1]**o+(0.159672390575)*x_ref[1]
        ref[1,0,1,1]=(-0.361120028772)*x_ref[0]**o+(-0.406707123035)*x_ref[0]+(-0.54341861971)*x_ref[1]**o+(-0.0392836956005)*x_ref[1]
        ref[1,0,2,0]=(0.944116160979)*x_ref[0]**o+(0.0890185721642)*x_ref[0]+(0.73533144751)*x_ref[1]**o+(0.701431910124)*x_ref[1]
        ref[1,0,2,1]=(0.722492883973)*x_ref[0]**o+(0.351843053475)*x_ref[0]+(-0.1756241753)*x_ref[1]**o+(-0.703690677653)*x_ref[1]
        ref[1,1,0,0]=(0.262345197739)*x_ref[0]**o+(0.683913892378)*x_ref[0]+(0.322362020764)*x_ref[1]**o+(0.851902848166)*x_ref[1]
        ref[1,1,0,1]=(-0.0509963903106)*x_ref[0]**o+(0.387065915322)*x_ref[0]+(0.0727012685343)*x_ref[1]**o+(0.666978314077)*x_ref[1]
        ref[1,1,1,0]=(-0.518122762009)*x_ref[0]**o+(0.69514335574)*x_ref[0]+(0.34513236897)*x_ref[1]**o+(0.641034103704)*x_ref[1]
        ref[1,1,1,1]=(0.0658005558873)*x_ref[0]**o+(-0.305147050226)*x_ref[0]+(-0.707909968597)*x_ref[1]**o+(0.0682220439272)*x_ref[1]
        ref[1,1,2,0]=(-0.443540005123)*x_ref[0]**o+(0.179307931754)*x_ref[0]+(0.982374202912)*x_ref[1]**o+(0.398796141179)*x_ref[1]
        ref[1,1,2,1]=(-0.52173751599)*x_ref[0]**o+(-0.523741826056)*x_ref[0]+(0.172601588482)*x_ref[1]**o+(0.969133917994)*x_ref[1]
        ref[1,2,0,0]=(-0.807607635564)*x_ref[0]**o+(-0.525440627018)*x_ref[0]+(-0.868823879608)*x_ref[1]**o+(0.913203387198)*x_ref[1]
        ref[1,2,0,1]=(0.84988195907)*x_ref[0]**o+(-0.394750851572)*x_ref[0]+(0.740132509878)*x_ref[1]**o+(0.855925902936)*x_ref[1]
        ref[1,2,1,0]=(0.055360229349)*x_ref[0]**o+(0.299990998824)*x_ref[0]+(0.558216562537)*x_ref[1]**o+(-0.382651884902)*x_ref[1]
        ref[1,2,1,1]=(0.423603813726)*x_ref[0]**o+(-0.907945615918)*x_ref[0]+(-0.117306295879)*x_ref[1]**o+(-0.0502458178048)*x_ref[1]
        ref[1,2,2,0]=(0.06196107837)*x_ref[0]**o+(-0.612619509899)*x_ref[0]+(0.169179798359)*x_ref[1]**o+(-0.27854976134)*x_ref[1]
        ref[1,2,2,1]=(-0.506388098563)*x_ref[0]**o+(0.125008314945)*x_ref[0]+(0.509571389625)*x_ref[1]**o+(0.442027080754)*x_ref[1]
        ref[1,3,0,0]=(0.304208463979)*x_ref[0]**o+(-0.379194737667)*x_ref[0]+(0.978868674946)*x_ref[1]**o+(0.011573322151)*x_ref[1]
        ref[1,3,0,1]=(-0.151986748739)*x_ref[0]**o+(0.153593063241)*x_ref[0]+(0.644694257244)*x_ref[1]**o+(0.414073745353)*x_ref[1]
        ref[1,3,1,0]=(0.924109632511)*x_ref[0]**o+(0.398532810155)*x_ref[0]+(-0.321086484535)*x_ref[1]**o+(0.0802378501316)*x_ref[1]
        ref[1,3,1,1]=(-0.175639876422)*x_ref[0]**o+(-0.109223049695)*x_ref[0]+(-0.206450491082)*x_ref[1]**o+(0.462124555092)*x_ref[1]
        ref[1,3,2,0]=(0.379156310804)*x_ref[0]**o+(0.629232237742)*x_ref[0]+(0.0918877201433)*x_ref[1]**o+(-0.0490282902402)*x_ref[1]
        ref[1,3,2,1]=(-0.616126549785)*x_ref[0]**o+(0.528823818571)*x_ref[0]+(-0.947449800384)*x_ref[1]**o+(0.145045253788)*x_ref[1]
        ref[1,4,0,0]=(-0.623165695489)*x_ref[0]**o+(-0.861583499935)*x_ref[0]+(0.219861821844)*x_ref[1]**o+(-0.962135814763)*x_ref[1]
        ref[1,4,0,1]=(0.467875500099)*x_ref[0]**o+(0.906948957958)*x_ref[0]+(0.606958286586)*x_ref[1]**o+(-0.423219122678)*x_ref[1]
        ref[1,4,1,0]=(0.200880358483)*x_ref[0]**o+(0.571730447827)*x_ref[0]+(0.908300268754)*x_ref[1]**o+(0.19978633475)*x_ref[1]
        ref[1,4,1,1]=(0.55118731313)*x_ref[0]**o+(-0.89903633572)*x_ref[0]+(0.43836310821)*x_ref[1]**o+(0.404697808939)*x_ref[1]
        ref[1,4,2,0]=(0.536587233299)*x_ref[0]**o+(-0.0789531097738)*x_ref[0]+(-0.775848629826)*x_ref[1]**o+(-0.11020318415)*x_ref[1]
        ref[1,4,2,1]=(-0.648091804458)*x_ref[0]**o+(-0.0282126867936)*x_ref[0]+(0.0182411167897)*x_ref[1]**o+(0.922414038295)*x_ref[1]
        ref[2,0,0,0]=(0.67385583226)*x_ref[0]**o+(-0.993060381925)*x_ref[0]+(0.731462156889)*x_ref[1]**o+(0.183718928132)*x_ref[1]
        ref[2,0,0,1]=(0.881124628544)*x_ref[0]**o+(0.828720578546)*x_ref[0]+(-0.324258860918)*x_ref[1]**o+(-0.836625964944)*x_ref[1]
        ref[2,0,1,0]=(0.944307497546)*x_ref[0]**o+(0.546752005636)*x_ref[0]+(0.334176020601)*x_ref[1]**o+(0.654940705861)*x_ref[1]
        ref[2,0,1,1]=(-0.414075359332)*x_ref[0]**o+(-0.782021485027)*x_ref[0]+(-0.233424028825)*x_ref[1]**o+(0.652900422088)*x_ref[1]
        ref[2,0,2,0]=(0.355942393341)*x_ref[0]**o+(-0.727957138948)*x_ref[0]+(-0.924406373726)*x_ref[1]**o+(-0.806178313048)*x_ref[1]
        ref[2,0,2,1]=(-0.791343585742)*x_ref[0]**o+(-0.61722808774)*x_ref[0]+(-0.627757910074)*x_ref[1]**o+(0.838801750555)*x_ref[1]
        ref[2,1,0,0]=(-0.866831381931)*x_ref[0]**o+(-0.00498093548417)*x_ref[0]+(-0.568387329284)*x_ref[1]**o+(-0.428222759011)*x_ref[1]
        ref[2,1,0,1]=(-0.245588724312)*x_ref[0]**o+(0.571669044307)*x_ref[0]+(0.832968046978)*x_ref[1]**o+(-0.0595908787599)*x_ref[1]
        ref[2,1,1,0]=(0.454631203504)*x_ref[0]**o+(-0.827357560144)*x_ref[0]+(0.876484630903)*x_ref[1]**o+(0.316668730296)*x_ref[1]
        ref[2,1,1,1]=(-0.802326431878)*x_ref[0]**o+(-0.547113643585)*x_ref[0]+(0.619358075866)*x_ref[1]**o+(0.387898635274)*x_ref[1]
        ref[2,1,2,0]=(0.204730319481)*x_ref[0]**o+(0.462536054065)*x_ref[0]+(-0.267050942194)*x_ref[1]**o+(-0.529151462961)*x_ref[1]
        ref[2,1,2,1]=(-0.354690492675)*x_ref[0]**o+(0.518914850203)*x_ref[0]+(-0.283765250756)*x_ref[1]**o+(0.595912311106)*x_ref[1]
        ref[2,2,0,0]=(0.329350522628)*x_ref[0]**o+(0.689800893689)*x_ref[0]+(-0.316097521431)*x_ref[1]**o+(0.567054800242)*x_ref[1]
        ref[2,2,0,1]=(-0.68603272612)*x_ref[0]**o+(-0.458451752241)*x_ref[0]+(0.442247854284)*x_ref[1]**o+(0.342266856887)*x_ref[1]
        ref[2,2,1,0]=(0.0562781380442)*x_ref[0]**o+(0.615293898746)*x_ref[0]+(-0.446430775323)*x_ref[1]**o+(-0.720310464006)*x_ref[1]
        ref[2,2,1,1]=(-0.620026567808)*x_ref[0]**o+(0.64935057185)*x_ref[0]+(-0.402630635809)*x_ref[1]**o+(0.984397383657)*x_ref[1]
        ref[2,2,2,0]=(-0.565466963253)*x_ref[0]**o+(-0.983283191302)*x_ref[0]+(0.0517997901365)*x_ref[1]**o+(-0.00459375732072)*x_ref[1]
        ref[2,2,2,1]=(-0.544844126966)*x_ref[0]**o+(-0.993222368569)*x_ref[0]+(-0.0761437907904)*x_ref[1]**o+(0.959202974542)*x_ref[1]
        ref[2,3,0,0]=(-0.542176217992)*x_ref[0]**o+(-0.210550241274)*x_ref[0]+(0.447732450287)*x_ref[1]**o+(-0.0802692719358)*x_ref[1]
        ref[2,3,0,1]=(0.339422485349)*x_ref[0]**o+(0.581409183541)*x_ref[0]+(-0.255731595248)*x_ref[1]**o+(-0.196957128507)*x_ref[1]
        ref[2,3,1,0]=(0.938606287977)*x_ref[0]**o+(0.396408957375)*x_ref[0]+(-0.53613392333)*x_ref[1]**o+(-0.201412967427)*x_ref[1]
        ref[2,3,1,1]=(0.482931877644)*x_ref[0]**o+(-0.395740204538)*x_ref[0]+(-0.816216426797)*x_ref[1]**o+(-0.390149600366)*x_ref[1]
        ref[2,3,2,0]=(0.507838191989)*x_ref[0]**o+(0.590218664943)*x_ref[0]+(0.987052963874)*x_ref[1]**o+(-0.39735585208)*x_ref[1]
        ref[2,3,2,1]=(-0.036562453009)*x_ref[0]**o+(0.299852609043)*x_ref[0]+(0.194160324844)*x_ref[1]**o+(-0.629803094543)*x_ref[1]
        ref[2,4,0,0]=(0.128664458677)*x_ref[0]**o+(0.956037220714)*x_ref[0]+(0.770331226849)*x_ref[1]**o+(-0.208007445309)*x_ref[1]
        ref[2,4,0,1]=(0.526755720598)*x_ref[0]**o+(-0.334650615987)*x_ref[0]+(-0.245693030991)*x_ref[1]**o+(-0.712134261554)*x_ref[1]
        ref[2,4,1,0]=(0.827770360865)*x_ref[0]**o+(0.982906973774)*x_ref[0]+(0.471438390576)*x_ref[1]**o+(0.691215965917)*x_ref[1]
        ref[2,4,1,1]=(-0.206715480236)*x_ref[0]**o+(0.488812089863)*x_ref[0]+(0.777672390635)*x_ref[1]**o+(-0.496681850755)*x_ref[1]
        ref[2,4,2,0]=(0.183938264875)*x_ref[0]**o+(-0.247363713904)*x_ref[0]+(0.962928810844)*x_ref[1]**o+(0.642706194018)*x_ref[1]
        ref[2,4,2,1]=(0.758025105584)*x_ref[0]**o+(-0.484793869316)*x_ref[0]+(-0.431089127161)*x_ref[1]**o+(0.127528834965)*x_ref[1]
        ref[3,0,0,0]=(-0.130978215422)*x_ref[0]**o+(0.688819586128)*x_ref[0]+(0.517162828175)*x_ref[1]**o+(-0.979632247431)*x_ref[1]
        ref[3,0,0,1]=(0.130939629053)*x_ref[0]**o+(0.685452224086)*x_ref[0]+(-0.438454293093)*x_ref[1]**o+(-0.727705397967)*x_ref[1]
        ref[3,0,1,0]=(-0.31299808275)*x_ref[0]**o+(0.266897949461)*x_ref[0]+(0.117566429054)*x_ref[1]**o+(0.536075754797)*x_ref[1]
        ref[3,0,1,1]=(-0.401764430381)*x_ref[0]**o+(-0.205983383212)*x_ref[0]+(0.461300727769)*x_ref[1]**o+(-0.580777919325)*x_ref[1]
        ref[3,0,2,0]=(-0.63526664417)*x_ref[0]**o+(-0.878687700719)*x_ref[0]+(0.104141198375)*x_ref[1]**o+(-0.539520033395)*x_ref[1]
        ref[3,0,2,1]=(-0.592054773422)*x_ref[0]**o+(-0.10222451991)*x_ref[0]+(-0.385428960408)*x_ref[1]**o+(-0.497192270367)*x_ref[1]
        ref[3,1,0,0]=(0.17159987373)*x_ref[0]**o+(-0.215728352568)*x_ref[0]+(0.188842078024)*x_ref[1]**o+(0.417647732421)*x_ref[1]
        ref[3,1,0,1]=(0.548015992401)*x_ref[0]**o+(-0.669157307443)*x_ref[0]+(-0.796467777561)*x_ref[1]**o+(0.163833145799)*x_ref[1]
        ref[3,1,1,0]=(-0.643762450865)*x_ref[0]**o+(0.548750992658)*x_ref[0]+(0.409211186464)*x_ref[1]**o+(-0.986601038131)*x_ref[1]
        ref[3,1,1,1]=(-0.882340195766)*x_ref[0]**o+(-0.185979117217)*x_ref[0]+(-0.00826877676749)*x_ref[1]**o+(0.0657705611536)*x_ref[1]
        ref[3,1,2,0]=(-0.0644145920386)*x_ref[0]**o+(0.18740710155)*x_ref[0]+(-0.690647279595)*x_ref[1]**o+(-0.185556727699)*x_ref[1]
        ref[3,1,2,1]=(0.114366852618)*x_ref[0]**o+(-0.388821787469)*x_ref[0]+(-0.195241613331)*x_ref[1]**o+(0.871862920816)*x_ref[1]
        ref[3,2,0,0]=(-0.673828733159)*x_ref[0]**o+(-0.172601630353)*x_ref[0]+(-0.488329313517)*x_ref[1]**o+(0.0809307315053)*x_ref[1]
        ref[3,2,0,1]=(0.565273426718)*x_ref[0]**o+(-0.610423824924)*x_ref[0]+(0.856623167832)*x_ref[1]**o+(-0.148845473418)*x_ref[1]
        ref[3,2,1,0]=(-0.3472751889)*x_ref[0]**o+(0.908614631864)*x_ref[0]+(0.445654124992)*x_ref[1]**o+(0.719429313439)*x_ref[1]
        ref[3,2,1,1]=(0.24891832441)*x_ref[0]**o+(-0.650081727005)*x_ref[0]+(0.0980208592314)*x_ref[1]**o+(0.575231534075)*x_ref[1]
        ref[3,2,2,0]=(-0.737935474986)*x_ref[0]**o+(-0.874708935739)*x_ref[0]+(0.856037958766)*x_ref[1]**o+(-0.420041922056)*x_ref[1]
        ref[3,2,2,1]=(0.106345127894)*x_ref[0]**o+(0.363060447152)*x_ref[0]+(-0.472262909157)*x_ref[1]**o+(-0.231475435242)*x_ref[1]
        ref[3,3,0,0]=(-0.3699135022)*x_ref[0]**o+(-0.167427922355)*x_ref[0]+(-0.911307660539)*x_ref[1]**o+(-0.445726901976)*x_ref[1]
        ref[3,3,0,1]=(0.505339735087)*x_ref[0]**o+(0.410541984072)*x_ref[0]+(0.847415917938)*x_ref[1]**o+(0.434325090295)*x_ref[1]
        ref[3,3,1,0]=(0.2589814894)*x_ref[0]**o+(-0.087780879445)*x_ref[0]+(0.847904119817)*x_ref[1]**o+(0.462585909981)*x_ref[1]
        ref[3,3,1,1]=(0.283347628256)*x_ref[0]**o+(0.637029684847)*x_ref[0]+(0.174554376017)*x_ref[1]**o+(0.642427538978)*x_ref[1]
        ref[3,3,2,0]=(-0.599348189011)*x_ref[0]**o+(0.087002991683)*x_ref[0]+(0.823297612508)*x_ref[1]**o+(0.84180029964)*x_ref[1]
        ref[3,3,2,1]=(0.846090423225)*x_ref[0]**o+(-0.465062745856)*x_ref[0]+(0.498670439686)*x_ref[1]**o+(0.721831089205)*x_ref[1]
        ref[3,4,0,0]=(0.0344263713162)*x_ref[0]**o+(0.789039070423)*x_ref[0]+(-0.518898476617)*x_ref[1]**o+(0.112666233369)*x_ref[1]
        ref[3,4,0,1]=(0.752855926629)*x_ref[0]**o+(-0.028671093428)*x_ref[0]+(0.457225098582)*x_ref[1]**o+(0.311273673822)*x_ref[1]
        ref[3,4,1,0]=(-0.623675057135)*x_ref[0]**o+(-0.19730173917)*x_ref[0]+(-0.328378661061)*x_ref[1]**o+(-0.196083495589)*x_ref[1]
        ref[3,4,1,1]=(0.722548420108)*x_ref[0]**o+(0.471733735786)*x_ref[0]+(-0.151931791011)*x_ref[1]**o+(0.0147226290316)*x_ref[1]
        ref[3,4,2,0]=(0.512232160342)*x_ref[0]**o+(-0.339879887275)*x_ref[0]+(-0.261940015624)*x_ref[1]**o+(-0.589559690156)*x_ref[1]
        ref[3,4,2,1]=(0.387751869794)*x_ref[0]**o+(0.284749443635)*x_ref[0]+(0.764067938503)*x_ref[1]**o+(0.20999603354)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.844350537532)*x[0]**o+(0.981166530382)*x[0]+(0.255920408286)*x[1]**o+(0.483711485104)*x[1]+(0.00959687239474)*x[2]**o+(0.306323594446)*x[2]
        s[0,0,0,1]=(0.740076186971)*x[0]**o+(-0.491347138384)*x[0]+(0.25175965888)*x[1]**o+(-0.623273215882)*x[1]+(0.51258068256)*x[2]**o+(-0.546204518632)*x[2]
        s[0,0,1,0]=(0.135495798592)*x[0]**o+(-0.928404807377)*x[0]+(-0.385528692626)*x[1]**o+(0.533554015352)*x[1]+(0.117724513259)*x[2]**o+(0.572637990898)*x[2]
        s[0,0,1,1]=(-0.850632098694)*x[0]**o+(-0.488169472233)*x[0]+(0.503260539233)*x[1]**o+(-0.859633434356)*x[1]+(0.415712182486)*x[2]**o+(0.339616059154)*x[2]
        s[0,0,2,0]=(-0.018555867134)*x[0]**o+(0.861452555488)*x[0]+(0.381089258556)*x[1]**o+(0.101855012068)*x[1]+(-0.15705292069)*x[2]**o+(-0.826396062146)*x[2]
        s[0,0,2,1]=(0.160998814621)*x[0]**o+(-0.410483868941)*x[0]+(-0.159057698561)*x[1]**o+(0.78589010204)*x[1]+(0.766903367421)*x[2]**o+(0.102093838786)*x[2]
        s[0,1,0,0]=(-0.445025622437)*x[0]**o+(-0.683407434608)*x[0]+(0.415296785191)*x[1]**o+(-0.720473598436)*x[1]+(0.892104458869)*x[2]**o+(0.678624532111)*x[2]
        s[0,1,0,1]=(0.494607245554)*x[0]**o+(-0.261983074835)*x[0]+(0.659330653523)*x[1]**o+(0.517010331237)*x[1]+(0.33680183722)*x[2]**o+(0.984150461948)*x[2]
        s[0,1,1,0]=(-0.461365247044)*x[0]**o+(-0.913532792608)*x[0]+(0.0548715431963)*x[1]**o+(-0.195304655857)*x[1]+(-0.680191903401)*x[2]**o+(-0.344258045633)*x[2]
        s[0,1,1,1]=(-0.366551699639)*x[0]**o+(0.0380843408355)*x[0]+(-0.959343312241)*x[1]**o+(-0.0236083869931)*x[1]+(-0.728804551514)*x[2]**o+(0.411917097114)*x[2]
        s[0,1,2,0]=(-0.755044693231)*x[0]**o+(-0.589493821888)*x[0]+(-0.675570377341)*x[1]**o+(0.712529396708)*x[1]+(0.121878507837)*x[2]**o+(-0.500132877994)*x[2]
        s[0,1,2,1]=(-0.613779986247)*x[0]**o+(0.977105357645)*x[0]+(-0.82406258445)*x[1]**o+(0.109110319693)*x[1]+(-0.384891270465)*x[2]**o+(0.0513108774825)*x[2]
        s[0,2,0,0]=(-0.215753544701)*x[0]**o+(-0.485302830852)*x[0]+(0.305187675609)*x[1]**o+(-0.852320135062)*x[1]+(0.479016046098)*x[2]**o+(0.904887182515)*x[2]
        s[0,2,0,1]=(-0.902665442283)*x[0]**o+(0.339847683115)*x[0]+(-0.201409198091)*x[1]**o+(-0.28916884043)*x[1]+(0.0122973146477)*x[2]**o+(0.902178823571)*x[2]
        s[0,2,1,0]=(-0.345222234379)*x[0]**o+(-0.402351849375)*x[0]+(-0.830448872698)*x[1]**o+(-0.00616252140166)*x[1]+(-0.973740549537)*x[2]**o+(-0.430178208352)*x[2]
        s[0,2,1,1]=(0.602597376682)*x[0]**o+(0.19777615169)*x[0]+(0.432285684454)*x[1]**o+(0.837105284275)*x[1]+(0.674505043227)*x[2]**o+(0.908555029565)*x[2]
        s[0,2,2,0]=(0.05197142873)*x[0]**o+(0.23981420979)*x[0]+(-0.630581895848)*x[1]**o+(-0.823407671488)*x[1]+(-0.819697754602)*x[2]**o+(-0.0606600491833)*x[2]
        s[0,2,2,1]=(0.671981483754)*x[0]**o+(0.273934751949)*x[0]+(0.680845824642)*x[1]**o+(-0.243860132213)*x[1]+(0.166045811143)*x[2]**o+(0.661029385371)*x[2]
        s[0,3,0,0]=(-0.599325667157)*x[0]**o+(0.295726750611)*x[0]+(-0.446959020679)*x[1]**o+(-0.934383765903)*x[1]+(-0.142759892188)*x[2]**o+(0.892076097753)*x[2]
        s[0,3,0,1]=(-0.081435415733)*x[0]**o+(-0.914632751608)*x[0]+(-0.0356505325211)*x[1]**o+(-0.594158265622)*x[1]+(-0.183149255345)*x[2]**o+(0.992685124274)*x[2]
        s[0,3,1,0]=(-0.52041548403)*x[0]**o+(0.84346502916)*x[0]+(0.589777480478)*x[1]**o+(-0.395593542394)*x[1]+(0.670347108347)*x[2]**o+(-0.755732891394)*x[2]
        s[0,3,1,1]=(-0.0843404796224)*x[0]**o+(-0.350884489082)*x[0]+(-0.230005943865)*x[1]**o+(0.775052308561)*x[1]+(-0.722944185393)*x[2]**o+(-0.661807415523)*x[2]
        s[0,3,2,0]=(0.562202744376)*x[0]**o+(0.755433320508)*x[0]+(0.983360558953)*x[1]**o+(-0.18576911233)*x[1]+(-0.546634474194)*x[2]**o+(-0.448629525918)*x[2]
        s[0,3,2,1]=(0.418182343511)*x[0]**o+(-0.804470815559)*x[0]+(0.328075033205)*x[1]**o+(0.794791705586)*x[1]+(0.955923558558)*x[2]**o+(0.477204042313)*x[2]
        s[0,4,0,0]=(-0.473795825706)*x[0]**o+(0.0613706961191)*x[0]+(0.728253884932)*x[1]**o+(0.922149327337)*x[1]+(-0.514934433516)*x[2]**o+(-0.350899100204)*x[2]
        s[0,4,0,1]=(0.403705624315)*x[0]**o+(0.332463550032)*x[0]+(0.0413433808287)*x[1]**o+(-0.88184294823)*x[1]+(0.898896470347)*x[2]**o+(-0.1692829666)*x[2]
        s[0,4,1,0]=(-0.0696293975354)*x[0]**o+(0.697969617481)*x[0]+(-0.872250160629)*x[1]**o+(-0.396555707549)*x[1]+(0.398974483685)*x[2]**o+(-0.38500615055)*x[2]
        s[0,4,1,1]=(-0.818302505854)*x[0]**o+(-0.660712096214)*x[0]+(-0.687407491883)*x[1]**o+(0.662996835898)*x[1]+(-0.777440711816)*x[2]**o+(0.901132206169)*x[2]
        s[0,4,2,0]=(-0.98506239403)*x[0]**o+(-0.457191412508)*x[0]+(0.203796091348)*x[1]**o+(-0.0304382176582)*x[1]+(0.375197892442)*x[2]**o+(0.0794166875091)*x[2]
        s[0,4,2,1]=(-0.939331221009)*x[0]**o+(0.662619061409)*x[0]+(0.550834749841)*x[1]**o+(0.895500286676)*x[1]+(-0.886048606057)*x[2]**o+(-0.15220276644)*x[2]
        s[1,0,0,0]=(-0.553168089515)*x[0]**o+(-0.631287474707)*x[0]+(0.236518377648)*x[1]**o+(0.798404510441)*x[1]+(0.33391551032)*x[2]**o+(0.723204832044)*x[2]
        s[1,0,0,1]=(0.618774989743)*x[0]**o+(-0.625192149767)*x[0]+(-0.423599611797)*x[1]**o+(0.0627868455822)*x[1]+(0.248591794171)*x[2]**o+(-0.568908092813)*x[2]
        s[1,0,1,0]=(-0.856803065565)*x[0]**o+(0.615617576001)*x[0]+(-0.689638395609)*x[1]**o+(-0.756681900602)*x[1]+(-0.226065925851)*x[2]**o+(0.348925448185)*x[2]
        s[1,0,1,1]=(0.801728371493)*x[0]**o+(-0.436203474211)*x[0]+(-0.0202077001179)*x[1]**o+(0.652651006485)*x[1]+(-0.230375932312)*x[2]**o+(-0.890003876378)*x[2]
        s[1,0,2,0]=(0.29719598244)*x[0]**o+(-0.15276643068)*x[0]+(-0.753013966455)*x[1]**o+(-0.517288289121)*x[1]+(-0.218994832652)*x[2]**o+(-0.146759568045)*x[2]
        s[1,0,2,1]=(-0.823649332214)*x[0]**o+(-0.843687314415)*x[0]+(-0.0638848545941)*x[1]**o+(0.722449122523)*x[1]+(0.491191260847)*x[2]**o+(-0.752018096725)*x[2]
        s[1,1,0,0]=(0.605980591392)*x[0]**o+(0.202355210592)*x[0]+(0.379171937943)*x[1]**o+(0.063757128966)*x[1]+(-0.971141711866)*x[2]**o+(-0.632361023174)*x[2]
        s[1,1,0,1]=(-0.962798234484)*x[0]**o+(0.446686088092)*x[0]+(-0.521649313187)*x[1]**o+(-0.756613553959)*x[1]+(0.598374656129)*x[2]**o+(0.203329622383)*x[2]
        s[1,1,1,0]=(0.300526220295)*x[0]**o+(-0.821407258021)*x[0]+(-0.519739408983)*x[1]**o+(-0.300945102018)*x[1]+(0.592444505162)*x[2]**o+(0.093896688069)*x[2]
        s[1,1,1,1]=(-0.683108409273)*x[0]**o+(-0.724514513218)*x[0]+(-0.672965343753)*x[1]**o+(-0.990600638457)*x[1]+(0.288343195848)*x[2]**o+(0.836765793149)*x[2]
        s[1,1,2,0]=(-0.938695582412)*x[0]**o+(0.524424434979)*x[0]+(-0.230446441603)*x[1]**o+(-0.660840814146)*x[1]+(0.410079472542)*x[2]**o+(-0.501145734383)*x[2]
        s[1,1,2,1]=(-0.934625690674)*x[0]**o+(-0.355510911308)*x[0]+(0.112799513937)*x[1]**o+(0.0550081941593)*x[1]+(-0.564671701833)*x[2]**o+(-0.435422588591)*x[2]
        s[1,2,0,0]=(-0.329336582979)*x[0]**o+(0.557448032082)*x[0]+(0.660452325378)*x[1]**o+(0.986140231524)*x[1]+(-0.343267425889)*x[2]**o+(-0.124929320203)*x[2]
        s[1,2,0,1]=(0.141679080878)*x[0]**o+(-0.929990482459)*x[0]+(-0.0124869738018)*x[1]**o+(-0.297954361644)*x[1]+(0.237086926195)*x[2]**o+(-0.163693767697)*x[2]
        s[1,2,1,0]=(0.840523338196)*x[0]**o+(0.292937319051)*x[0]+(-0.647937602564)*x[1]**o+(0.599823564217)*x[1]+(0.966887229971)*x[2]**o+(-0.301069847494)*x[2]
        s[1,2,1,1]=(0.75606641666)*x[0]**o+(0.694906659273)*x[0]+(0.894582185091)*x[1]**o+(0.431150046844)*x[1]+(0.988226420754)*x[2]**o+(-0.860220033688)*x[2]
        s[1,2,2,0]=(-0.716168748278)*x[0]**o+(0.464108589188)*x[0]+(0.631539541512)*x[1]**o+(-0.992859266146)*x[1]+(-0.669793952605)*x[2]**o+(-0.743066646537)*x[2]
        s[1,2,2,1]=(0.695010429853)*x[0]**o+(-0.56668549077)*x[0]+(0.242451432508)*x[1]**o+(0.155033232196)*x[1]+(-0.362798375291)*x[2]**o+(-0.488187632585)*x[2]
        s[1,3,0,0]=(-0.943147179882)*x[0]**o+(0.71046937082)*x[0]+(0.852528992189)*x[1]**o+(-0.193096337802)*x[1]+(-0.787307409902)*x[2]**o+(0.251655133658)*x[2]
        s[1,3,0,1]=(-0.208552630722)*x[0]**o+(0.346012986086)*x[0]+(0.551360403822)*x[1]**o+(-0.348440170895)*x[1]+(0.363393290055)*x[2]**o+(0.0470822356286)*x[2]
        s[1,3,1,0]=(0.113226447773)*x[0]**o+(0.522623018987)*x[0]+(-0.0796736494068)*x[1]**o+(0.789297223214)*x[1]+(-0.779573310177)*x[2]**o+(-0.743401293513)*x[2]
        s[1,3,1,1]=(-0.466035390795)*x[0]**o+(0.511762679374)*x[0]+(-0.682170406028)*x[1]**o+(-0.949494996805)*x[1]+(-0.0648915682154)*x[2]**o+(0.495239120908)*x[2]
        s[1,3,2,0]=(-0.274148045419)*x[0]**o+(0.0815161328992)*x[0]+(0.358273465722)*x[1]**o+(0.636995118303)*x[1]+(-0.535665346021)*x[2]**o+(0.823457431671)*x[2]
        s[1,3,2,1]=(-0.716452730034)*x[0]**o+(-0.483846622879)*x[0]+(-0.608042660277)*x[1]**o+(0.0278424023059)*x[1]+(0.843814111075)*x[2]**o+(0.955433558186)*x[2]
        s[1,4,0,0]=(0.455371216931)*x[0]**o+(0.602586579764)*x[0]+(0.158622918138)*x[1]**o+(-0.706234646453)*x[1]+(-0.475064482018)*x[2]**o+(0.146600391902)*x[2]
        s[1,4,0,1]=(0.980992566699)*x[0]**o+(0.117314780276)*x[0]+(-0.158086008164)*x[1]**o+(0.632974744247)*x[1]+(0.978173293529)*x[2]**o+(0.885266669359)*x[2]
        s[1,4,1,0]=(0.279463579513)*x[0]**o+(0.402328966807)*x[0]+(0.301400738565)*x[1]**o+(-0.955899292995)*x[1]+(-0.500757799904)*x[2]**o+(0.281493491795)*x[2]
        s[1,4,1,1]=(-0.822078727162)*x[0]**o+(0.369502628318)*x[0]+(-0.928613467846)*x[1]**o+(-0.350001970293)*x[1]+(-0.968163790337)*x[2]**o+(-0.310282764085)*x[2]
        s[1,4,2,0]=(0.3290445083)*x[0]**o+(-0.985109449578)*x[0]+(-0.632388659807)*x[1]**o+(0.466010510408)*x[1]+(0.460974186748)*x[2]**o+(-0.716019421615)*x[2]
        s[1,4,2,1]=(0.72236273453)*x[0]**o+(-0.0121296025323)*x[0]+(-0.846055371974)*x[1]**o+(0.984741432507)*x[1]+(-0.522899574925)*x[2]**o+(-0.21853151342)*x[2]
        s[2,0,0,0]=(0.0109795220027)*x[0]**o+(0.556433154187)*x[0]+(-0.770646566445)*x[1]**o+(-0.736722512756)*x[1]+(0.668238695822)*x[2]**o+(0.180946691645)*x[2]
        s[2,0,0,1]=(0.719005782249)*x[0]**o+(-0.872006790171)*x[0]+(-0.957147163648)*x[1]**o+(-0.0682732175112)*x[1]+(0.583562455856)*x[2]**o+(-0.30991382684)*x[2]
        s[2,0,1,0]=(0.622360614797)*x[0]**o+(-0.00796068126083)*x[0]+(0.3225801025)*x[1]**o+(-0.919808515024)*x[1]+(-0.919597324996)*x[2]**o+(0.373038030942)*x[2]
        s[2,0,1,1]=(0.405340603348)*x[0]**o+(-0.0495755391664)*x[0]+(-0.47787299709)*x[1]**o+(-0.254771467734)*x[1]+(-0.591700035231)*x[2]**o+(-0.164217809972)*x[2]
        s[2,0,2,0]=(0.087396380793)*x[0]**o+(-0.374710687926)*x[0]+(0.0575627557138)*x[1]**o+(-0.70523617076)*x[1]+(-0.484089875239)*x[2]**o+(0.789318249087)*x[2]
        s[2,0,2,1]=(0.778697410279)*x[0]**o+(0.519333641831)*x[0]+(-0.123493167274)*x[1]**o+(0.366951337816)*x[1]+(-0.537900250619)*x[2]**o+(0.627028303525)*x[2]
        s[2,1,0,0]=(-0.745940888357)*x[0]**o+(0.711657374433)*x[0]+(0.836411404243)*x[1]**o+(-0.174311748628)*x[1]+(0.776203832353)*x[2]**o+(-0.706629687992)*x[2]
        s[2,1,0,1]=(-0.642638091614)*x[0]**o+(0.172213012143)*x[0]+(0.0413400352844)*x[1]**o+(-0.69201275421)*x[1]+(-0.824191389562)*x[2]**o+(0.251335644925)*x[2]
        s[2,1,1,0]=(-0.749920542498)*x[0]**o+(0.389193707611)*x[0]+(0.413903115322)*x[1]**o+(0.74472403825)*x[1]+(-0.591260331172)*x[2]**o+(0.276810972178)*x[2]
        s[2,1,1,1]=(0.439899556721)*x[0]**o+(0.566349932103)*x[0]+(-0.741172831484)*x[1]**o+(-0.983770395204)*x[1]+(-0.337604226356)*x[2]**o+(-0.984256469612)*x[2]
        s[2,1,2,0]=(-0.744905118064)*x[0]**o+(0.0758118730978)*x[0]+(-0.954906020107)*x[1]**o+(-0.321181040222)*x[1]+(-0.319406466958)*x[2]**o+(-0.462108015037)*x[2]
        s[2,1,2,1]=(0.495181652645)*x[0]**o+(0.321912068502)*x[0]+(0.625371783418)*x[1]**o+(0.661392879428)*x[1]+(0.95878113647)*x[2]**o+(-0.00161261625228)*x[2]
        s[2,2,0,0]=(-0.747880472493)*x[0]**o+(0.133266267487)*x[0]+(0.594232553468)*x[1]**o+(0.164357025178)*x[1]+(0.183282050202)*x[2]**o+(0.510098487732)*x[2]
        s[2,2,0,1]=(-0.2990096515)*x[0]**o+(-0.530653405949)*x[0]+(-0.774239397996)*x[1]**o+(0.513406622276)*x[1]+(0.68969837178)*x[2]**o+(0.0542139499089)*x[2]
        s[2,2,1,0]=(0.803320557325)*x[0]**o+(-0.53977892905)*x[0]+(0.602881824151)*x[1]**o+(0.151232673003)*x[1]+(0.933315439891)*x[2]**o+(-0.0915356088824)*x[2]
        s[2,2,1,1]=(-0.573022070609)*x[0]**o+(0.394673455213)*x[0]+(0.290948264558)*x[1]**o+(0.482816198889)*x[1]+(0.457587694713)*x[2]**o+(0.146402139866)*x[2]
        s[2,2,2,0]=(0.0984322495009)*x[0]**o+(-0.607655184921)*x[0]+(-0.165102103099)*x[1]**o+(-0.949911109313)*x[1]+(0.807082265921)*x[2]**o+(-0.201828673075)*x[2]
        s[2,2,2,1]=(0.387692723867)*x[0]**o+(-0.350126071452)*x[0]+(-0.840248594934)*x[1]**o+(-0.754015389174)*x[1]+(0.82690792431)*x[2]**o+(0.984937777626)*x[2]
        s[2,3,0,0]=(-0.22249653492)*x[0]**o+(0.95589235945)*x[0]+(0.748620345408)*x[1]**o+(-0.544563975574)*x[1]+(0.399260091432)*x[2]**o+(-0.664314569966)*x[2]
        s[2,3,0,1]=(-0.881513041393)*x[0]**o+(-0.286371325145)*x[0]+(0.207149857405)*x[1]**o+(0.962030218672)*x[1]+(-0.621096606228)*x[2]**o+(0.321143840842)*x[2]
        s[2,3,1,0]=(-0.294822275479)*x[0]**o+(0.952468891463)*x[0]+(0.890797831223)*x[1]**o+(-0.154408717393)*x[1]+(0.653171114568)*x[2]**o+(-0.335321213379)*x[2]
        s[2,3,1,1]=(0.391962263482)*x[0]**o+(0.567275318333)*x[0]+(-0.185626734066)*x[1]**o+(-0.377715271388)*x[1]+(0.739574492513)*x[2]**o+(-0.603873103899)*x[2]
        s[2,3,2,0]=(0.428479962249)*x[0]**o+(0.127318777736)*x[0]+(-0.42333631821)*x[1]**o+(-0.169907154638)*x[1]+(0.725857578617)*x[2]**o+(-0.697584549589)*x[2]
        s[2,3,2,1]=(0.372507332734)*x[0]**o+(0.381914565887)*x[0]+(0.218450152368)*x[1]**o+(-0.802951368728)*x[1]+(-0.00721884700159)*x[2]**o+(0.071666772532)*x[2]
        s[2,4,0,0]=(-0.520915873118)*x[0]**o+(-0.904770241319)*x[0]+(0.594347214938)*x[1]**o+(-0.681345109104)*x[1]+(0.209360616612)*x[2]**o+(0.424835494205)*x[2]
        s[2,4,0,1]=(-0.727883496639)*x[0]**o+(-0.668568348644)*x[0]+(0.833770360827)*x[1]**o+(-0.0255151478099)*x[1]+(-0.193240025032)*x[2]**o+(0.197362196363)*x[2]
        s[2,4,1,0]=(0.742869556969)*x[0]**o+(0.475432034925)*x[0]+(0.357310924868)*x[1]**o+(-0.905361920184)*x[1]+(-0.0284188203545)*x[2]**o+(-0.028888633895)*x[2]
        s[2,4,1,1]=(-0.331466533146)*x[0]**o+(0.544600410552)*x[0]+(0.959927301686)*x[1]**o+(-0.875596269307)*x[1]+(0.0907336453125)*x[2]**o+(0.638768254635)*x[2]
        s[2,4,2,0]=(-0.140519776938)*x[0]**o+(-0.341221155863)*x[0]+(0.476853206406)*x[1]**o+(-0.376640947632)*x[1]+(0.222188072931)*x[2]**o+(-0.600634989398)*x[2]
        s[2,4,2,1]=(0.436824352085)*x[0]**o+(0.0902891669206)*x[0]+(0.950721335793)*x[1]**o+(0.578125459387)*x[1]+(-0.482219010223)*x[2]**o+(-0.232985764124)*x[2]
        s[3,0,0,0]=(-0.571010773207)*x[0]**o+(-0.80484920281)*x[0]+(-0.823071194682)*x[1]**o+(-0.896108871228)*x[1]+(-0.631430205832)*x[2]**o+(-0.704868332657)*x[2]
        s[3,0,0,1]=(0.273328305518)*x[0]**o+(0.391247949879)*x[0]+(-0.577768086094)*x[1]**o+(-0.0268817540424)*x[1]+(-0.807017351875)*x[2]**o+(0.855714026407)*x[2]
        s[3,0,1,0]=(-0.783392552436)*x[0]**o+(-0.606379147005)*x[0]+(0.596212011594)*x[1]**o+(0.837204930781)*x[1]+(-0.444629434195)*x[2]**o+(0.818388831162)*x[2]
        s[3,0,1,1]=(-0.7632340237)*x[0]**o+(-0.525136694101)*x[0]+(0.79942659492)*x[1]**o+(-0.404546089352)*x[1]+(-0.728139040689)*x[2]**o+(-0.998101874873)*x[2]
        s[3,0,2,0]=(-0.509268960989)*x[0]**o+(-0.676093609952)*x[0]+(0.861170135902)*x[1]**o+(0.577425218873)*x[1]+(0.577209383315)*x[2]**o+(-0.384779054446)*x[2]
        s[3,0,2,1]=(0.011431602973)*x[0]**o+(-0.334328421599)*x[0]+(0.252604277681)*x[1]**o+(-0.603913209811)*x[1]+(0.00917009818384)*x[2]**o+(0.301867115072)*x[2]
        s[3,1,0,0]=(-0.620651658587)*x[0]**o+(0.814248186892)*x[0]+(0.231571604024)*x[1]**o+(-0.208758448156)*x[1]+(0.514917171695)*x[2]**o+(0.795545898743)*x[2]
        s[3,1,0,1]=(-0.358180244409)*x[0]**o+(0.762252417714)*x[0]+(0.235442173264)*x[1]**o+(0.681677639507)*x[1]+(-0.610029281327)*x[2]**o+(0.255430019725)*x[2]
        s[3,1,1,0]=(-0.337744414049)*x[0]**o+(0.659764964399)*x[0]+(-0.93916869727)*x[1]**o+(0.0920440667279)*x[1]+(0.9429852742)*x[2]**o+(-0.152643224429)*x[2]
        s[3,1,1,1]=(-0.850080149157)*x[0]**o+(0.150991562708)*x[0]+(-0.265559714378)*x[1]**o+(0.0825721664033)*x[1]+(0.0577388830574)*x[2]**o+(-0.0168755109359)*x[2]
        s[3,1,2,0]=(0.736517689601)*x[0]**o+(-0.299770623584)*x[0]+(-0.642480331197)*x[1]**o+(0.507838878495)*x[1]+(0.102240842692)*x[2]**o+(-0.474668897224)*x[2]
        s[3,1,2,1]=(0.718251907196)*x[0]**o+(-0.732535005154)*x[0]+(-0.0595472398188)*x[1]**o+(-0.809928140752)*x[1]+(-0.949798402613)*x[2]**o+(-0.817043878578)*x[2]
        s[3,2,0,0]=(-0.226866840941)*x[0]**o+(0.981200696871)*x[0]+(0.329563738019)*x[1]**o+(-0.859364161406)*x[1]+(-0.0889989821964)*x[2]**o+(-0.891130810171)*x[2]
        s[3,2,0,1]=(0.448731899011)*x[0]**o+(-0.295976710644)*x[0]+(-0.0631712972169)*x[1]**o+(-0.0910278410217)*x[1]+(-0.568889472919)*x[2]**o+(0.763619087805)*x[2]
        s[3,2,1,0]=(0.383474130267)*x[0]**o+(0.73677599326)*x[0]+(-0.208757244921)*x[1]**o+(-0.278191993802)*x[1]+(0.795441377078)*x[2]**o+(0.905216800847)*x[2]
        s[3,2,1,1]=(0.0725196158569)*x[0]**o+(-0.943723180474)*x[0]+(0.699282898995)*x[1]**o+(0.388214849035)*x[1]+(0.596007988212)*x[2]**o+(0.736772507174)*x[2]
        s[3,2,2,0]=(-0.448938945133)*x[0]**o+(0.384414869647)*x[0]+(0.527747975961)*x[1]**o+(0.127653616477)*x[1]+(0.377808521053)*x[2]**o+(-0.252616720439)*x[2]
        s[3,2,2,1]=(-0.376250853312)*x[0]**o+(-0.707650293785)*x[0]+(0.672982484116)*x[1]**o+(0.990265204194)*x[1]+(0.529775927528)*x[2]**o+(-0.409195688811)*x[2]
        s[3,3,0,0]=(-0.530159451732)*x[0]**o+(-0.955185089804)*x[0]+(-0.748512782868)*x[1]**o+(-0.618229320649)*x[1]+(0.627829579969)*x[2]**o+(-0.333717653758)*x[2]
        s[3,3,0,1]=(0.966906267367)*x[0]**o+(-0.0951932785216)*x[0]+(-0.61695602507)*x[1]**o+(0.446458255991)*x[1]+(0.207784087287)*x[2]**o+(-0.288316765862)*x[2]
        s[3,3,1,0]=(-0.996153491218)*x[0]**o+(0.0588370396384)*x[0]+(-0.294277377632)*x[1]**o+(0.689632857934)*x[1]+(-0.143160246331)*x[2]**o+(0.833437162663)*x[2]
        s[3,3,1,1]=(0.906322391848)*x[0]**o+(0.492397941858)*x[0]+(-0.0266527290051)*x[1]**o+(-0.172670670559)*x[1]+(-0.826003550146)*x[2]**o+(-0.103742447179)*x[2]
        s[3,3,2,0]=(-0.479133669526)*x[0]**o+(0.577467684597)*x[0]+(-0.214066405495)*x[1]**o+(0.0589162518627)*x[1]+(-0.618757108816)*x[2]**o+(-0.0596902292359)*x[2]
        s[3,3,2,1]=(0.501813126681)*x[0]**o+(-0.0527123864794)*x[0]+(0.442612384316)*x[1]**o+(0.587529973102)*x[1]+(0.96334743671)*x[2]**o+(-0.878243725936)*x[2]
        s[3,4,0,0]=(-0.283828850179)*x[0]**o+(0.553981240206)*x[0]+(0.526151552782)*x[1]**o+(-0.822056178815)*x[1]+(-0.922360625781)*x[2]**o+(-0.0972249338686)*x[2]
        s[3,4,0,1]=(0.513911875447)*x[0]**o+(0.66906817932)*x[0]+(-0.843990607619)*x[1]**o+(-0.873810309081)*x[1]+(0.9732978504)*x[2]**o+(-0.767991735083)*x[2]
        s[3,4,1,0]=(-0.918970664697)*x[0]**o+(0.409739379673)*x[0]+(0.856834850961)*x[1]**o+(-0.149117026418)*x[1]+(-0.215870849261)*x[2]**o+(-0.139382801219)*x[2]
        s[3,4,1,1]=(0.869749963776)*x[0]**o+(-0.0551502951625)*x[0]+(-0.211684312437)*x[1]**o+(0.674587225438)*x[1]+(0.713205722506)*x[2]**o+(0.943169316201)*x[2]
        s[3,4,2,0]=(-0.0849462931018)*x[0]**o+(0.131197062211)*x[0]+(-0.462104529576)*x[1]**o+(0.620371430368)*x[1]+(0.858403152981)*x[2]**o+(0.102462631271)*x[2]
        s[3,4,2,1]=(-0.402094215027)*x[0]**o+(-0.246651318354)*x[0]+(-0.741149807021)*x[1]**o+(-0.616686358212)*x[1]+(-0.606948106264)*x[2]**o+(-0.911839049257)*x[2]
        ref[0,0,0,0]=(-0.844350537532)*x_ref[0]**o+(0.981166530382)*x_ref[0]+(0.255920408286)*x_ref[1]**o+(0.483711485104)*x_ref[1]+(0.00959687239474)*x_ref[2]**o+(0.306323594446)*x_ref[2]
        ref[0,0,0,1]=(0.740076186971)*x_ref[0]**o+(-0.491347138384)*x_ref[0]+(0.25175965888)*x_ref[1]**o+(-0.623273215882)*x_ref[1]+(0.51258068256)*x_ref[2]**o+(-0.546204518632)*x_ref[2]
        ref[0,0,1,0]=(0.135495798592)*x_ref[0]**o+(-0.928404807377)*x_ref[0]+(-0.385528692626)*x_ref[1]**o+(0.533554015352)*x_ref[1]+(0.117724513259)*x_ref[2]**o+(0.572637990898)*x_ref[2]
        ref[0,0,1,1]=(-0.850632098694)*x_ref[0]**o+(-0.488169472233)*x_ref[0]+(0.503260539233)*x_ref[1]**o+(-0.859633434356)*x_ref[1]+(0.415712182486)*x_ref[2]**o+(0.339616059154)*x_ref[2]
        ref[0,0,2,0]=(-0.018555867134)*x_ref[0]**o+(0.861452555488)*x_ref[0]+(0.381089258556)*x_ref[1]**o+(0.101855012068)*x_ref[1]+(-0.15705292069)*x_ref[2]**o+(-0.826396062146)*x_ref[2]
        ref[0,0,2,1]=(0.160998814621)*x_ref[0]**o+(-0.410483868941)*x_ref[0]+(-0.159057698561)*x_ref[1]**o+(0.78589010204)*x_ref[1]+(0.766903367421)*x_ref[2]**o+(0.102093838786)*x_ref[2]
        ref[0,1,0,0]=(-0.445025622437)*x_ref[0]**o+(-0.683407434608)*x_ref[0]+(0.415296785191)*x_ref[1]**o+(-0.720473598436)*x_ref[1]+(0.892104458869)*x_ref[2]**o+(0.678624532111)*x_ref[2]
        ref[0,1,0,1]=(0.494607245554)*x_ref[0]**o+(-0.261983074835)*x_ref[0]+(0.659330653523)*x_ref[1]**o+(0.517010331237)*x_ref[1]+(0.33680183722)*x_ref[2]**o+(0.984150461948)*x_ref[2]
        ref[0,1,1,0]=(-0.461365247044)*x_ref[0]**o+(-0.913532792608)*x_ref[0]+(0.0548715431963)*x_ref[1]**o+(-0.195304655857)*x_ref[1]+(-0.680191903401)*x_ref[2]**o+(-0.344258045633)*x_ref[2]
        ref[0,1,1,1]=(-0.366551699639)*x_ref[0]**o+(0.0380843408355)*x_ref[0]+(-0.959343312241)*x_ref[1]**o+(-0.0236083869931)*x_ref[1]+(-0.728804551514)*x_ref[2]**o+(0.411917097114)*x_ref[2]
        ref[0,1,2,0]=(-0.755044693231)*x_ref[0]**o+(-0.589493821888)*x_ref[0]+(-0.675570377341)*x_ref[1]**o+(0.712529396708)*x_ref[1]+(0.121878507837)*x_ref[2]**o+(-0.500132877994)*x_ref[2]
        ref[0,1,2,1]=(-0.613779986247)*x_ref[0]**o+(0.977105357645)*x_ref[0]+(-0.82406258445)*x_ref[1]**o+(0.109110319693)*x_ref[1]+(-0.384891270465)*x_ref[2]**o+(0.0513108774825)*x_ref[2]
        ref[0,2,0,0]=(-0.215753544701)*x_ref[0]**o+(-0.485302830852)*x_ref[0]+(0.305187675609)*x_ref[1]**o+(-0.852320135062)*x_ref[1]+(0.479016046098)*x_ref[2]**o+(0.904887182515)*x_ref[2]
        ref[0,2,0,1]=(-0.902665442283)*x_ref[0]**o+(0.339847683115)*x_ref[0]+(-0.201409198091)*x_ref[1]**o+(-0.28916884043)*x_ref[1]+(0.0122973146477)*x_ref[2]**o+(0.902178823571)*x_ref[2]
        ref[0,2,1,0]=(-0.345222234379)*x_ref[0]**o+(-0.402351849375)*x_ref[0]+(-0.830448872698)*x_ref[1]**o+(-0.00616252140166)*x_ref[1]+(-0.973740549537)*x_ref[2]**o+(-0.430178208352)*x_ref[2]
        ref[0,2,1,1]=(0.602597376682)*x_ref[0]**o+(0.19777615169)*x_ref[0]+(0.432285684454)*x_ref[1]**o+(0.837105284275)*x_ref[1]+(0.674505043227)*x_ref[2]**o+(0.908555029565)*x_ref[2]
        ref[0,2,2,0]=(0.05197142873)*x_ref[0]**o+(0.23981420979)*x_ref[0]+(-0.630581895848)*x_ref[1]**o+(-0.823407671488)*x_ref[1]+(-0.819697754602)*x_ref[2]**o+(-0.0606600491833)*x_ref[2]
        ref[0,2,2,1]=(0.671981483754)*x_ref[0]**o+(0.273934751949)*x_ref[0]+(0.680845824642)*x_ref[1]**o+(-0.243860132213)*x_ref[1]+(0.166045811143)*x_ref[2]**o+(0.661029385371)*x_ref[2]
        ref[0,3,0,0]=(-0.599325667157)*x_ref[0]**o+(0.295726750611)*x_ref[0]+(-0.446959020679)*x_ref[1]**o+(-0.934383765903)*x_ref[1]+(-0.142759892188)*x_ref[2]**o+(0.892076097753)*x_ref[2]
        ref[0,3,0,1]=(-0.081435415733)*x_ref[0]**o+(-0.914632751608)*x_ref[0]+(-0.0356505325211)*x_ref[1]**o+(-0.594158265622)*x_ref[1]+(-0.183149255345)*x_ref[2]**o+(0.992685124274)*x_ref[2]
        ref[0,3,1,0]=(-0.52041548403)*x_ref[0]**o+(0.84346502916)*x_ref[0]+(0.589777480478)*x_ref[1]**o+(-0.395593542394)*x_ref[1]+(0.670347108347)*x_ref[2]**o+(-0.755732891394)*x_ref[2]
        ref[0,3,1,1]=(-0.0843404796224)*x_ref[0]**o+(-0.350884489082)*x_ref[0]+(-0.230005943865)*x_ref[1]**o+(0.775052308561)*x_ref[1]+(-0.722944185393)*x_ref[2]**o+(-0.661807415523)*x_ref[2]
        ref[0,3,2,0]=(0.562202744376)*x_ref[0]**o+(0.755433320508)*x_ref[0]+(0.983360558953)*x_ref[1]**o+(-0.18576911233)*x_ref[1]+(-0.546634474194)*x_ref[2]**o+(-0.448629525918)*x_ref[2]
        ref[0,3,2,1]=(0.418182343511)*x_ref[0]**o+(-0.804470815559)*x_ref[0]+(0.328075033205)*x_ref[1]**o+(0.794791705586)*x_ref[1]+(0.955923558558)*x_ref[2]**o+(0.477204042313)*x_ref[2]
        ref[0,4,0,0]=(-0.473795825706)*x_ref[0]**o+(0.0613706961191)*x_ref[0]+(0.728253884932)*x_ref[1]**o+(0.922149327337)*x_ref[1]+(-0.514934433516)*x_ref[2]**o+(-0.350899100204)*x_ref[2]
        ref[0,4,0,1]=(0.403705624315)*x_ref[0]**o+(0.332463550032)*x_ref[0]+(0.0413433808287)*x_ref[1]**o+(-0.88184294823)*x_ref[1]+(0.898896470347)*x_ref[2]**o+(-0.1692829666)*x_ref[2]
        ref[0,4,1,0]=(-0.0696293975354)*x_ref[0]**o+(0.697969617481)*x_ref[0]+(-0.872250160629)*x_ref[1]**o+(-0.396555707549)*x_ref[1]+(0.398974483685)*x_ref[2]**o+(-0.38500615055)*x_ref[2]
        ref[0,4,1,1]=(-0.818302505854)*x_ref[0]**o+(-0.660712096214)*x_ref[0]+(-0.687407491883)*x_ref[1]**o+(0.662996835898)*x_ref[1]+(-0.777440711816)*x_ref[2]**o+(0.901132206169)*x_ref[2]
        ref[0,4,2,0]=(-0.98506239403)*x_ref[0]**o+(-0.457191412508)*x_ref[0]+(0.203796091348)*x_ref[1]**o+(-0.0304382176582)*x_ref[1]+(0.375197892442)*x_ref[2]**o+(0.0794166875091)*x_ref[2]
        ref[0,4,2,1]=(-0.939331221009)*x_ref[0]**o+(0.662619061409)*x_ref[0]+(0.550834749841)*x_ref[1]**o+(0.895500286676)*x_ref[1]+(-0.886048606057)*x_ref[2]**o+(-0.15220276644)*x_ref[2]
        ref[1,0,0,0]=(-0.553168089515)*x_ref[0]**o+(-0.631287474707)*x_ref[0]+(0.236518377648)*x_ref[1]**o+(0.798404510441)*x_ref[1]+(0.33391551032)*x_ref[2]**o+(0.723204832044)*x_ref[2]
        ref[1,0,0,1]=(0.618774989743)*x_ref[0]**o+(-0.625192149767)*x_ref[0]+(-0.423599611797)*x_ref[1]**o+(0.0627868455822)*x_ref[1]+(0.248591794171)*x_ref[2]**o+(-0.568908092813)*x_ref[2]
        ref[1,0,1,0]=(-0.856803065565)*x_ref[0]**o+(0.615617576001)*x_ref[0]+(-0.689638395609)*x_ref[1]**o+(-0.756681900602)*x_ref[1]+(-0.226065925851)*x_ref[2]**o+(0.348925448185)*x_ref[2]
        ref[1,0,1,1]=(0.801728371493)*x_ref[0]**o+(-0.436203474211)*x_ref[0]+(-0.0202077001179)*x_ref[1]**o+(0.652651006485)*x_ref[1]+(-0.230375932312)*x_ref[2]**o+(-0.890003876378)*x_ref[2]
        ref[1,0,2,0]=(0.29719598244)*x_ref[0]**o+(-0.15276643068)*x_ref[0]+(-0.753013966455)*x_ref[1]**o+(-0.517288289121)*x_ref[1]+(-0.218994832652)*x_ref[2]**o+(-0.146759568045)*x_ref[2]
        ref[1,0,2,1]=(-0.823649332214)*x_ref[0]**o+(-0.843687314415)*x_ref[0]+(-0.0638848545941)*x_ref[1]**o+(0.722449122523)*x_ref[1]+(0.491191260847)*x_ref[2]**o+(-0.752018096725)*x_ref[2]
        ref[1,1,0,0]=(0.605980591392)*x_ref[0]**o+(0.202355210592)*x_ref[0]+(0.379171937943)*x_ref[1]**o+(0.063757128966)*x_ref[1]+(-0.971141711866)*x_ref[2]**o+(-0.632361023174)*x_ref[2]
        ref[1,1,0,1]=(-0.962798234484)*x_ref[0]**o+(0.446686088092)*x_ref[0]+(-0.521649313187)*x_ref[1]**o+(-0.756613553959)*x_ref[1]+(0.598374656129)*x_ref[2]**o+(0.203329622383)*x_ref[2]
        ref[1,1,1,0]=(0.300526220295)*x_ref[0]**o+(-0.821407258021)*x_ref[0]+(-0.519739408983)*x_ref[1]**o+(-0.300945102018)*x_ref[1]+(0.592444505162)*x_ref[2]**o+(0.093896688069)*x_ref[2]
        ref[1,1,1,1]=(-0.683108409273)*x_ref[0]**o+(-0.724514513218)*x_ref[0]+(-0.672965343753)*x_ref[1]**o+(-0.990600638457)*x_ref[1]+(0.288343195848)*x_ref[2]**o+(0.836765793149)*x_ref[2]
        ref[1,1,2,0]=(-0.938695582412)*x_ref[0]**o+(0.524424434979)*x_ref[0]+(-0.230446441603)*x_ref[1]**o+(-0.660840814146)*x_ref[1]+(0.410079472542)*x_ref[2]**o+(-0.501145734383)*x_ref[2]
        ref[1,1,2,1]=(-0.934625690674)*x_ref[0]**o+(-0.355510911308)*x_ref[0]+(0.112799513937)*x_ref[1]**o+(0.0550081941593)*x_ref[1]+(-0.564671701833)*x_ref[2]**o+(-0.435422588591)*x_ref[2]
        ref[1,2,0,0]=(-0.329336582979)*x_ref[0]**o+(0.557448032082)*x_ref[0]+(0.660452325378)*x_ref[1]**o+(0.986140231524)*x_ref[1]+(-0.343267425889)*x_ref[2]**o+(-0.124929320203)*x_ref[2]
        ref[1,2,0,1]=(0.141679080878)*x_ref[0]**o+(-0.929990482459)*x_ref[0]+(-0.0124869738018)*x_ref[1]**o+(-0.297954361644)*x_ref[1]+(0.237086926195)*x_ref[2]**o+(-0.163693767697)*x_ref[2]
        ref[1,2,1,0]=(0.840523338196)*x_ref[0]**o+(0.292937319051)*x_ref[0]+(-0.647937602564)*x_ref[1]**o+(0.599823564217)*x_ref[1]+(0.966887229971)*x_ref[2]**o+(-0.301069847494)*x_ref[2]
        ref[1,2,1,1]=(0.75606641666)*x_ref[0]**o+(0.694906659273)*x_ref[0]+(0.894582185091)*x_ref[1]**o+(0.431150046844)*x_ref[1]+(0.988226420754)*x_ref[2]**o+(-0.860220033688)*x_ref[2]
        ref[1,2,2,0]=(-0.716168748278)*x_ref[0]**o+(0.464108589188)*x_ref[0]+(0.631539541512)*x_ref[1]**o+(-0.992859266146)*x_ref[1]+(-0.669793952605)*x_ref[2]**o+(-0.743066646537)*x_ref[2]
        ref[1,2,2,1]=(0.695010429853)*x_ref[0]**o+(-0.56668549077)*x_ref[0]+(0.242451432508)*x_ref[1]**o+(0.155033232196)*x_ref[1]+(-0.362798375291)*x_ref[2]**o+(-0.488187632585)*x_ref[2]
        ref[1,3,0,0]=(-0.943147179882)*x_ref[0]**o+(0.71046937082)*x_ref[0]+(0.852528992189)*x_ref[1]**o+(-0.193096337802)*x_ref[1]+(-0.787307409902)*x_ref[2]**o+(0.251655133658)*x_ref[2]
        ref[1,3,0,1]=(-0.208552630722)*x_ref[0]**o+(0.346012986086)*x_ref[0]+(0.551360403822)*x_ref[1]**o+(-0.348440170895)*x_ref[1]+(0.363393290055)*x_ref[2]**o+(0.0470822356286)*x_ref[2]
        ref[1,3,1,0]=(0.113226447773)*x_ref[0]**o+(0.522623018987)*x_ref[0]+(-0.0796736494068)*x_ref[1]**o+(0.789297223214)*x_ref[1]+(-0.779573310177)*x_ref[2]**o+(-0.743401293513)*x_ref[2]
        ref[1,3,1,1]=(-0.466035390795)*x_ref[0]**o+(0.511762679374)*x_ref[0]+(-0.682170406028)*x_ref[1]**o+(-0.949494996805)*x_ref[1]+(-0.0648915682154)*x_ref[2]**o+(0.495239120908)*x_ref[2]
        ref[1,3,2,0]=(-0.274148045419)*x_ref[0]**o+(0.0815161328992)*x_ref[0]+(0.358273465722)*x_ref[1]**o+(0.636995118303)*x_ref[1]+(-0.535665346021)*x_ref[2]**o+(0.823457431671)*x_ref[2]
        ref[1,3,2,1]=(-0.716452730034)*x_ref[0]**o+(-0.483846622879)*x_ref[0]+(-0.608042660277)*x_ref[1]**o+(0.0278424023059)*x_ref[1]+(0.843814111075)*x_ref[2]**o+(0.955433558186)*x_ref[2]
        ref[1,4,0,0]=(0.455371216931)*x_ref[0]**o+(0.602586579764)*x_ref[0]+(0.158622918138)*x_ref[1]**o+(-0.706234646453)*x_ref[1]+(-0.475064482018)*x_ref[2]**o+(0.146600391902)*x_ref[2]
        ref[1,4,0,1]=(0.980992566699)*x_ref[0]**o+(0.117314780276)*x_ref[0]+(-0.158086008164)*x_ref[1]**o+(0.632974744247)*x_ref[1]+(0.978173293529)*x_ref[2]**o+(0.885266669359)*x_ref[2]
        ref[1,4,1,0]=(0.279463579513)*x_ref[0]**o+(0.402328966807)*x_ref[0]+(0.301400738565)*x_ref[1]**o+(-0.955899292995)*x_ref[1]+(-0.500757799904)*x_ref[2]**o+(0.281493491795)*x_ref[2]
        ref[1,4,1,1]=(-0.822078727162)*x_ref[0]**o+(0.369502628318)*x_ref[0]+(-0.928613467846)*x_ref[1]**o+(-0.350001970293)*x_ref[1]+(-0.968163790337)*x_ref[2]**o+(-0.310282764085)*x_ref[2]
        ref[1,4,2,0]=(0.3290445083)*x_ref[0]**o+(-0.985109449578)*x_ref[0]+(-0.632388659807)*x_ref[1]**o+(0.466010510408)*x_ref[1]+(0.460974186748)*x_ref[2]**o+(-0.716019421615)*x_ref[2]
        ref[1,4,2,1]=(0.72236273453)*x_ref[0]**o+(-0.0121296025323)*x_ref[0]+(-0.846055371974)*x_ref[1]**o+(0.984741432507)*x_ref[1]+(-0.522899574925)*x_ref[2]**o+(-0.21853151342)*x_ref[2]
        ref[2,0,0,0]=(0.0109795220027)*x_ref[0]**o+(0.556433154187)*x_ref[0]+(-0.770646566445)*x_ref[1]**o+(-0.736722512756)*x_ref[1]+(0.668238695822)*x_ref[2]**o+(0.180946691645)*x_ref[2]
        ref[2,0,0,1]=(0.719005782249)*x_ref[0]**o+(-0.872006790171)*x_ref[0]+(-0.957147163648)*x_ref[1]**o+(-0.0682732175112)*x_ref[1]+(0.583562455856)*x_ref[2]**o+(-0.30991382684)*x_ref[2]
        ref[2,0,1,0]=(0.622360614797)*x_ref[0]**o+(-0.00796068126083)*x_ref[0]+(0.3225801025)*x_ref[1]**o+(-0.919808515024)*x_ref[1]+(-0.919597324996)*x_ref[2]**o+(0.373038030942)*x_ref[2]
        ref[2,0,1,1]=(0.405340603348)*x_ref[0]**o+(-0.0495755391664)*x_ref[0]+(-0.47787299709)*x_ref[1]**o+(-0.254771467734)*x_ref[1]+(-0.591700035231)*x_ref[2]**o+(-0.164217809972)*x_ref[2]
        ref[2,0,2,0]=(0.087396380793)*x_ref[0]**o+(-0.374710687926)*x_ref[0]+(0.0575627557138)*x_ref[1]**o+(-0.70523617076)*x_ref[1]+(-0.484089875239)*x_ref[2]**o+(0.789318249087)*x_ref[2]
        ref[2,0,2,1]=(0.778697410279)*x_ref[0]**o+(0.519333641831)*x_ref[0]+(-0.123493167274)*x_ref[1]**o+(0.366951337816)*x_ref[1]+(-0.537900250619)*x_ref[2]**o+(0.627028303525)*x_ref[2]
        ref[2,1,0,0]=(-0.745940888357)*x_ref[0]**o+(0.711657374433)*x_ref[0]+(0.836411404243)*x_ref[1]**o+(-0.174311748628)*x_ref[1]+(0.776203832353)*x_ref[2]**o+(-0.706629687992)*x_ref[2]
        ref[2,1,0,1]=(-0.642638091614)*x_ref[0]**o+(0.172213012143)*x_ref[0]+(0.0413400352844)*x_ref[1]**o+(-0.69201275421)*x_ref[1]+(-0.824191389562)*x_ref[2]**o+(0.251335644925)*x_ref[2]
        ref[2,1,1,0]=(-0.749920542498)*x_ref[0]**o+(0.389193707611)*x_ref[0]+(0.413903115322)*x_ref[1]**o+(0.74472403825)*x_ref[1]+(-0.591260331172)*x_ref[2]**o+(0.276810972178)*x_ref[2]
        ref[2,1,1,1]=(0.439899556721)*x_ref[0]**o+(0.566349932103)*x_ref[0]+(-0.741172831484)*x_ref[1]**o+(-0.983770395204)*x_ref[1]+(-0.337604226356)*x_ref[2]**o+(-0.984256469612)*x_ref[2]
        ref[2,1,2,0]=(-0.744905118064)*x_ref[0]**o+(0.0758118730978)*x_ref[0]+(-0.954906020107)*x_ref[1]**o+(-0.321181040222)*x_ref[1]+(-0.319406466958)*x_ref[2]**o+(-0.462108015037)*x_ref[2]
        ref[2,1,2,1]=(0.495181652645)*x_ref[0]**o+(0.321912068502)*x_ref[0]+(0.625371783418)*x_ref[1]**o+(0.661392879428)*x_ref[1]+(0.95878113647)*x_ref[2]**o+(-0.00161261625228)*x_ref[2]
        ref[2,2,0,0]=(-0.747880472493)*x_ref[0]**o+(0.133266267487)*x_ref[0]+(0.594232553468)*x_ref[1]**o+(0.164357025178)*x_ref[1]+(0.183282050202)*x_ref[2]**o+(0.510098487732)*x_ref[2]
        ref[2,2,0,1]=(-0.2990096515)*x_ref[0]**o+(-0.530653405949)*x_ref[0]+(-0.774239397996)*x_ref[1]**o+(0.513406622276)*x_ref[1]+(0.68969837178)*x_ref[2]**o+(0.0542139499089)*x_ref[2]
        ref[2,2,1,0]=(0.803320557325)*x_ref[0]**o+(-0.53977892905)*x_ref[0]+(0.602881824151)*x_ref[1]**o+(0.151232673003)*x_ref[1]+(0.933315439891)*x_ref[2]**o+(-0.0915356088824)*x_ref[2]
        ref[2,2,1,1]=(-0.573022070609)*x_ref[0]**o+(0.394673455213)*x_ref[0]+(0.290948264558)*x_ref[1]**o+(0.482816198889)*x_ref[1]+(0.457587694713)*x_ref[2]**o+(0.146402139866)*x_ref[2]
        ref[2,2,2,0]=(0.0984322495009)*x_ref[0]**o+(-0.607655184921)*x_ref[0]+(-0.165102103099)*x_ref[1]**o+(-0.949911109313)*x_ref[1]+(0.807082265921)*x_ref[2]**o+(-0.201828673075)*x_ref[2]
        ref[2,2,2,1]=(0.387692723867)*x_ref[0]**o+(-0.350126071452)*x_ref[0]+(-0.840248594934)*x_ref[1]**o+(-0.754015389174)*x_ref[1]+(0.82690792431)*x_ref[2]**o+(0.984937777626)*x_ref[2]
        ref[2,3,0,0]=(-0.22249653492)*x_ref[0]**o+(0.95589235945)*x_ref[0]+(0.748620345408)*x_ref[1]**o+(-0.544563975574)*x_ref[1]+(0.399260091432)*x_ref[2]**o+(-0.664314569966)*x_ref[2]
        ref[2,3,0,1]=(-0.881513041393)*x_ref[0]**o+(-0.286371325145)*x_ref[0]+(0.207149857405)*x_ref[1]**o+(0.962030218672)*x_ref[1]+(-0.621096606228)*x_ref[2]**o+(0.321143840842)*x_ref[2]
        ref[2,3,1,0]=(-0.294822275479)*x_ref[0]**o+(0.952468891463)*x_ref[0]+(0.890797831223)*x_ref[1]**o+(-0.154408717393)*x_ref[1]+(0.653171114568)*x_ref[2]**o+(-0.335321213379)*x_ref[2]
        ref[2,3,1,1]=(0.391962263482)*x_ref[0]**o+(0.567275318333)*x_ref[0]+(-0.185626734066)*x_ref[1]**o+(-0.377715271388)*x_ref[1]+(0.739574492513)*x_ref[2]**o+(-0.603873103899)*x_ref[2]
        ref[2,3,2,0]=(0.428479962249)*x_ref[0]**o+(0.127318777736)*x_ref[0]+(-0.42333631821)*x_ref[1]**o+(-0.169907154638)*x_ref[1]+(0.725857578617)*x_ref[2]**o+(-0.697584549589)*x_ref[2]
        ref[2,3,2,1]=(0.372507332734)*x_ref[0]**o+(0.381914565887)*x_ref[0]+(0.218450152368)*x_ref[1]**o+(-0.802951368728)*x_ref[1]+(-0.00721884700159)*x_ref[2]**o+(0.071666772532)*x_ref[2]
        ref[2,4,0,0]=(-0.520915873118)*x_ref[0]**o+(-0.904770241319)*x_ref[0]+(0.594347214938)*x_ref[1]**o+(-0.681345109104)*x_ref[1]+(0.209360616612)*x_ref[2]**o+(0.424835494205)*x_ref[2]
        ref[2,4,0,1]=(-0.727883496639)*x_ref[0]**o+(-0.668568348644)*x_ref[0]+(0.833770360827)*x_ref[1]**o+(-0.0255151478099)*x_ref[1]+(-0.193240025032)*x_ref[2]**o+(0.197362196363)*x_ref[2]
        ref[2,4,1,0]=(0.742869556969)*x_ref[0]**o+(0.475432034925)*x_ref[0]+(0.357310924868)*x_ref[1]**o+(-0.905361920184)*x_ref[1]+(-0.0284188203545)*x_ref[2]**o+(-0.028888633895)*x_ref[2]
        ref[2,4,1,1]=(-0.331466533146)*x_ref[0]**o+(0.544600410552)*x_ref[0]+(0.959927301686)*x_ref[1]**o+(-0.875596269307)*x_ref[1]+(0.0907336453125)*x_ref[2]**o+(0.638768254635)*x_ref[2]
        ref[2,4,2,0]=(-0.140519776938)*x_ref[0]**o+(-0.341221155863)*x_ref[0]+(0.476853206406)*x_ref[1]**o+(-0.376640947632)*x_ref[1]+(0.222188072931)*x_ref[2]**o+(-0.600634989398)*x_ref[2]
        ref[2,4,2,1]=(0.436824352085)*x_ref[0]**o+(0.0902891669206)*x_ref[0]+(0.950721335793)*x_ref[1]**o+(0.578125459387)*x_ref[1]+(-0.482219010223)*x_ref[2]**o+(-0.232985764124)*x_ref[2]
        ref[3,0,0,0]=(-0.571010773207)*x_ref[0]**o+(-0.80484920281)*x_ref[0]+(-0.823071194682)*x_ref[1]**o+(-0.896108871228)*x_ref[1]+(-0.631430205832)*x_ref[2]**o+(-0.704868332657)*x_ref[2]
        ref[3,0,0,1]=(0.273328305518)*x_ref[0]**o+(0.391247949879)*x_ref[0]+(-0.577768086094)*x_ref[1]**o+(-0.0268817540424)*x_ref[1]+(-0.807017351875)*x_ref[2]**o+(0.855714026407)*x_ref[2]
        ref[3,0,1,0]=(-0.783392552436)*x_ref[0]**o+(-0.606379147005)*x_ref[0]+(0.596212011594)*x_ref[1]**o+(0.837204930781)*x_ref[1]+(-0.444629434195)*x_ref[2]**o+(0.818388831162)*x_ref[2]
        ref[3,0,1,1]=(-0.7632340237)*x_ref[0]**o+(-0.525136694101)*x_ref[0]+(0.79942659492)*x_ref[1]**o+(-0.404546089352)*x_ref[1]+(-0.728139040689)*x_ref[2]**o+(-0.998101874873)*x_ref[2]
        ref[3,0,2,0]=(-0.509268960989)*x_ref[0]**o+(-0.676093609952)*x_ref[0]+(0.861170135902)*x_ref[1]**o+(0.577425218873)*x_ref[1]+(0.577209383315)*x_ref[2]**o+(-0.384779054446)*x_ref[2]
        ref[3,0,2,1]=(0.011431602973)*x_ref[0]**o+(-0.334328421599)*x_ref[0]+(0.252604277681)*x_ref[1]**o+(-0.603913209811)*x_ref[1]+(0.00917009818384)*x_ref[2]**o+(0.301867115072)*x_ref[2]
        ref[3,1,0,0]=(-0.620651658587)*x_ref[0]**o+(0.814248186892)*x_ref[0]+(0.231571604024)*x_ref[1]**o+(-0.208758448156)*x_ref[1]+(0.514917171695)*x_ref[2]**o+(0.795545898743)*x_ref[2]
        ref[3,1,0,1]=(-0.358180244409)*x_ref[0]**o+(0.762252417714)*x_ref[0]+(0.235442173264)*x_ref[1]**o+(0.681677639507)*x_ref[1]+(-0.610029281327)*x_ref[2]**o+(0.255430019725)*x_ref[2]
        ref[3,1,1,0]=(-0.337744414049)*x_ref[0]**o+(0.659764964399)*x_ref[0]+(-0.93916869727)*x_ref[1]**o+(0.0920440667279)*x_ref[1]+(0.9429852742)*x_ref[2]**o+(-0.152643224429)*x_ref[2]
        ref[3,1,1,1]=(-0.850080149157)*x_ref[0]**o+(0.150991562708)*x_ref[0]+(-0.265559714378)*x_ref[1]**o+(0.0825721664033)*x_ref[1]+(0.0577388830574)*x_ref[2]**o+(-0.0168755109359)*x_ref[2]
        ref[3,1,2,0]=(0.736517689601)*x_ref[0]**o+(-0.299770623584)*x_ref[0]+(-0.642480331197)*x_ref[1]**o+(0.507838878495)*x_ref[1]+(0.102240842692)*x_ref[2]**o+(-0.474668897224)*x_ref[2]
        ref[3,1,2,1]=(0.718251907196)*x_ref[0]**o+(-0.732535005154)*x_ref[0]+(-0.0595472398188)*x_ref[1]**o+(-0.809928140752)*x_ref[1]+(-0.949798402613)*x_ref[2]**o+(-0.817043878578)*x_ref[2]
        ref[3,2,0,0]=(-0.226866840941)*x_ref[0]**o+(0.981200696871)*x_ref[0]+(0.329563738019)*x_ref[1]**o+(-0.859364161406)*x_ref[1]+(-0.0889989821964)*x_ref[2]**o+(-0.891130810171)*x_ref[2]
        ref[3,2,0,1]=(0.448731899011)*x_ref[0]**o+(-0.295976710644)*x_ref[0]+(-0.0631712972169)*x_ref[1]**o+(-0.0910278410217)*x_ref[1]+(-0.568889472919)*x_ref[2]**o+(0.763619087805)*x_ref[2]
        ref[3,2,1,0]=(0.383474130267)*x_ref[0]**o+(0.73677599326)*x_ref[0]+(-0.208757244921)*x_ref[1]**o+(-0.278191993802)*x_ref[1]+(0.795441377078)*x_ref[2]**o+(0.905216800847)*x_ref[2]
        ref[3,2,1,1]=(0.0725196158569)*x_ref[0]**o+(-0.943723180474)*x_ref[0]+(0.699282898995)*x_ref[1]**o+(0.388214849035)*x_ref[1]+(0.596007988212)*x_ref[2]**o+(0.736772507174)*x_ref[2]
        ref[3,2,2,0]=(-0.448938945133)*x_ref[0]**o+(0.384414869647)*x_ref[0]+(0.527747975961)*x_ref[1]**o+(0.127653616477)*x_ref[1]+(0.377808521053)*x_ref[2]**o+(-0.252616720439)*x_ref[2]
        ref[3,2,2,1]=(-0.376250853312)*x_ref[0]**o+(-0.707650293785)*x_ref[0]+(0.672982484116)*x_ref[1]**o+(0.990265204194)*x_ref[1]+(0.529775927528)*x_ref[2]**o+(-0.409195688811)*x_ref[2]
        ref[3,3,0,0]=(-0.530159451732)*x_ref[0]**o+(-0.955185089804)*x_ref[0]+(-0.748512782868)*x_ref[1]**o+(-0.618229320649)*x_ref[1]+(0.627829579969)*x_ref[2]**o+(-0.333717653758)*x_ref[2]
        ref[3,3,0,1]=(0.966906267367)*x_ref[0]**o+(-0.0951932785216)*x_ref[0]+(-0.61695602507)*x_ref[1]**o+(0.446458255991)*x_ref[1]+(0.207784087287)*x_ref[2]**o+(-0.288316765862)*x_ref[2]
        ref[3,3,1,0]=(-0.996153491218)*x_ref[0]**o+(0.0588370396384)*x_ref[0]+(-0.294277377632)*x_ref[1]**o+(0.689632857934)*x_ref[1]+(-0.143160246331)*x_ref[2]**o+(0.833437162663)*x_ref[2]
        ref[3,3,1,1]=(0.906322391848)*x_ref[0]**o+(0.492397941858)*x_ref[0]+(-0.0266527290051)*x_ref[1]**o+(-0.172670670559)*x_ref[1]+(-0.826003550146)*x_ref[2]**o+(-0.103742447179)*x_ref[2]
        ref[3,3,2,0]=(-0.479133669526)*x_ref[0]**o+(0.577467684597)*x_ref[0]+(-0.214066405495)*x_ref[1]**o+(0.0589162518627)*x_ref[1]+(-0.618757108816)*x_ref[2]**o+(-0.0596902292359)*x_ref[2]
        ref[3,3,2,1]=(0.501813126681)*x_ref[0]**o+(-0.0527123864794)*x_ref[0]+(0.442612384316)*x_ref[1]**o+(0.587529973102)*x_ref[1]+(0.96334743671)*x_ref[2]**o+(-0.878243725936)*x_ref[2]
        ref[3,4,0,0]=(-0.283828850179)*x_ref[0]**o+(0.553981240206)*x_ref[0]+(0.526151552782)*x_ref[1]**o+(-0.822056178815)*x_ref[1]+(-0.922360625781)*x_ref[2]**o+(-0.0972249338686)*x_ref[2]
        ref[3,4,0,1]=(0.513911875447)*x_ref[0]**o+(0.66906817932)*x_ref[0]+(-0.843990607619)*x_ref[1]**o+(-0.873810309081)*x_ref[1]+(0.9732978504)*x_ref[2]**o+(-0.767991735083)*x_ref[2]
        ref[3,4,1,0]=(-0.918970664697)*x_ref[0]**o+(0.409739379673)*x_ref[0]+(0.856834850961)*x_ref[1]**o+(-0.149117026418)*x_ref[1]+(-0.215870849261)*x_ref[2]**o+(-0.139382801219)*x_ref[2]
        ref[3,4,1,1]=(0.869749963776)*x_ref[0]**o+(-0.0551502951625)*x_ref[0]+(-0.211684312437)*x_ref[1]**o+(0.674587225438)*x_ref[1]+(0.713205722506)*x_ref[2]**o+(0.943169316201)*x_ref[2]
        ref[3,4,2,0]=(-0.0849462931018)*x_ref[0]**o+(0.131197062211)*x_ref[0]+(-0.462104529576)*x_ref[1]**o+(0.620371430368)*x_ref[1]+(0.858403152981)*x_ref[2]**o+(0.102462631271)*x_ref[2]
        ref[3,4,2,1]=(-0.402094215027)*x_ref[0]**o+(-0.246651318354)*x_ref[0]+(-0.741149807021)*x_ref[1]**o+(-0.616686358212)*x_ref[1]+(-0.606948106264)*x_ref[2]**o+(-0.911839049257)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.521673635599)*x[0]**o+(-0.288664980886)*x[0]+(0.57565949318)*x[1]**o+(-0.555613874185)*x[1]
        ref=(-0.521673635599)*x_ref[0]**o+(-0.288664980886)*x_ref[0]+(0.57565949318)*x_ref[1]**o+(-0.555613874185)*x_ref[1]
      else:
        arg=(0.520429804914)*x[0]**o+(-0.000683147004841)*x[0]+(0.676879975826)*x[1]**o+(0.760707578765)*x[1]+(0.479225734268)*x[2]**o+(0.132006876687)*x[2]
        ref=(0.520429804914)*x_ref[0]**o+(-0.000683147004841)*x_ref[0]+(0.676879975826)*x_ref[1]**o+(0.760707578765)*x_ref[1]+(0.479225734268)*x_ref[2]**o+(0.132006876687)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.350515208335)*x[0]**o+(-0.604596801095)*x[0]+(0.145267491062)*x[1]**o+(-0.422271555369)*x[1]
        ref=(-0.350515208335)*x_ref[0]**o+(-0.604596801095)*x_ref[0]+(0.145267491062)*x_ref[1]**o+(-0.422271555369)*x_ref[1]
      else:
        arg=(-0.162304501731)*x[0]**o+(0.342262454165)*x[0]+(0.693988751897)*x[1]**o+(-0.315225717754)*x[1]+(-0.938405302425)*x[2]**o+(-0.0929881908885)*x[2]
        ref=(-0.162304501731)*x_ref[0]**o+(0.342262454165)*x_ref[0]+(0.693988751897)*x_ref[1]**o+(-0.315225717754)*x_ref[1]+(-0.938405302425)*x_ref[2]**o+(-0.0929881908885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.104854565192)*x[0]**o+(-0.596688162793)*x[0]+(-0.0286471957779)*x[1]**o+(-0.157282508059)*x[1]
        arg[1]=(0.619817759027)*x[0]**o+(-0.832815965127)*x[0]+(-0.638126497014)*x[1]**o+(0.771271149341)*x[1]
        ref[0]=(0.104854565192)*x_ref[0]**o+(-0.596688162793)*x_ref[0]+(-0.0286471957779)*x_ref[1]**o+(-0.157282508059)*x_ref[1]
        ref[1]=(0.619817759027)*x_ref[0]**o+(-0.832815965127)*x_ref[0]+(-0.638126497014)*x_ref[1]**o+(0.771271149341)*x_ref[1]
      else:
        arg[0]=(-0.358142932888)*x[0]**o+(0.758685606027)*x[0]+(-0.0716050574096)*x[1]**o+(-0.723587074286)*x[1]+(0.387240173671)*x[2]**o+(0.544679811421)*x[2]
        arg[1]=(0.864154242034)*x[0]**o+(-0.0740598679188)*x[0]+(-0.0865611132887)*x[1]**o+(0.0414815431481)*x[1]+(-0.897395775334)*x[2]**o+(-0.580401889318)*x[2]
        ref[0]=(-0.358142932888)*x_ref[0]**o+(0.758685606027)*x_ref[0]+(-0.0716050574096)*x_ref[1]**o+(-0.723587074286)*x_ref[1]+(0.387240173671)*x_ref[2]**o+(0.544679811421)*x_ref[2]
        ref[1]=(0.864154242034)*x_ref[0]**o+(-0.0740598679188)*x_ref[0]+(-0.0865611132887)*x_ref[1]**o+(0.0414815431481)*x_ref[1]+(-0.897395775334)*x_ref[2]**o+(-0.580401889318)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.0571385613076)*x[0]**o+(-0.296048600622)*x[0]+(0.300914212927)*x[1]**o+(0.81578919689)*x[1]
        arg[0,1]=(0.136796732287)*x[0]**o+(0.169384864236)*x[0]+(-0.874585047899)*x[1]**o+(-0.67613629157)*x[1]
        arg[0,2]=(-0.947048556597)*x[0]**o+(-0.807647782941)*x[0]+(-0.53233790031)*x[1]**o+(0.618861393596)*x[1]
        arg[0,3]=(-0.533009370674)*x[0]**o+(-0.732258917532)*x[0]+(0.735058784793)*x[1]**o+(0.313284184128)*x[1]
        arg[0,4]=(0.678291805831)*x[0]**o+(0.892272189132)*x[0]+(-0.902740106267)*x[1]**o+(0.902593764933)*x[1]
        arg[1,0]=(0.101446968367)*x[0]**o+(0.758270058166)*x[0]+(-0.511416064913)*x[1]**o+(-0.464147847068)*x[1]
        arg[1,1]=(-0.0131419066862)*x[0]**o+(-0.942460391354)*x[0]+(0.479955107486)*x[1]**o+(-0.311304181244)*x[1]
        arg[1,2]=(0.623958732361)*x[0]**o+(-0.313196338837)*x[0]+(0.413428996617)*x[1]**o+(0.348327566672)*x[1]
        arg[1,3]=(-0.8413636876)*x[0]**o+(-0.572573098032)*x[0]+(-0.678663954356)*x[1]**o+(-0.419716403314)*x[1]
        arg[1,4]=(-0.580466216486)*x[0]**o+(0.384226238557)*x[0]+(-0.732722613215)*x[1]**o+(-0.0238970013828)*x[1]
        arg[2,0]=(-0.248986933717)*x[0]**o+(0.264019654506)*x[0]+(0.0314811616292)*x[1]**o+(0.631481081422)*x[1]
        arg[2,1]=(0.643576426484)*x[0]**o+(-0.313599735242)*x[0]+(0.908780876791)*x[1]**o+(-0.698972268201)*x[1]
        arg[2,2]=(-0.623208321993)*x[0]**o+(0.649637800083)*x[0]+(-0.878484988208)*x[1]**o+(0.368793479691)*x[1]
        arg[2,3]=(0.239333407828)*x[0]**o+(-0.293558915555)*x[0]+(0.271630238733)*x[1]**o+(0.741304510722)*x[1]
        arg[2,4]=(0.175914942104)*x[0]**o+(0.336116863935)*x[0]+(-0.799168065807)*x[1]**o+(0.0264962830729)*x[1]
        arg[3,0]=(0.777665875019)*x[0]**o+(-0.13232137001)*x[0]+(0.682046389381)*x[1]**o+(0.589069623109)*x[1]
        arg[3,1]=(0.7607621575)*x[0]**o+(-0.81285460551)*x[0]+(0.0523010659976)*x[1]**o+(-0.913765717778)*x[1]
        arg[3,2]=(0.822107753118)*x[0]**o+(0.508869893322)*x[0]+(0.679583348422)*x[1]**o+(-0.597814039932)*x[1]
        arg[3,3]=(-0.77699176705)*x[0]**o+(-0.267758166676)*x[0]+(0.661073127871)*x[1]**o+(-0.41434545147)*x[1]
        arg[3,4]=(0.223013380161)*x[0]**o+(-0.558077040245)*x[0]+(-0.450831764747)*x[1]**o+(-0.0897765635191)*x[1]
        ref[0,0]=(-0.0571385613076)*x_ref[0]**o+(-0.296048600622)*x_ref[0]+(0.300914212927)*x_ref[1]**o+(0.81578919689)*x_ref[1]
        ref[0,1]=(0.136796732287)*x_ref[0]**o+(0.169384864236)*x_ref[0]+(-0.874585047899)*x_ref[1]**o+(-0.67613629157)*x_ref[1]
        ref[0,2]=(-0.947048556597)*x_ref[0]**o+(-0.807647782941)*x_ref[0]+(-0.53233790031)*x_ref[1]**o+(0.618861393596)*x_ref[1]
        ref[0,3]=(-0.533009370674)*x_ref[0]**o+(-0.732258917532)*x_ref[0]+(0.735058784793)*x_ref[1]**o+(0.313284184128)*x_ref[1]
        ref[0,4]=(0.678291805831)*x_ref[0]**o+(0.892272189132)*x_ref[0]+(-0.902740106267)*x_ref[1]**o+(0.902593764933)*x_ref[1]
        ref[1,0]=(0.101446968367)*x_ref[0]**o+(0.758270058166)*x_ref[0]+(-0.511416064913)*x_ref[1]**o+(-0.464147847068)*x_ref[1]
        ref[1,1]=(-0.0131419066862)*x_ref[0]**o+(-0.942460391354)*x_ref[0]+(0.479955107486)*x_ref[1]**o+(-0.311304181244)*x_ref[1]
        ref[1,2]=(0.623958732361)*x_ref[0]**o+(-0.313196338837)*x_ref[0]+(0.413428996617)*x_ref[1]**o+(0.348327566672)*x_ref[1]
        ref[1,3]=(-0.8413636876)*x_ref[0]**o+(-0.572573098032)*x_ref[0]+(-0.678663954356)*x_ref[1]**o+(-0.419716403314)*x_ref[1]
        ref[1,4]=(-0.580466216486)*x_ref[0]**o+(0.384226238557)*x_ref[0]+(-0.732722613215)*x_ref[1]**o+(-0.0238970013828)*x_ref[1]
        ref[2,0]=(-0.248986933717)*x_ref[0]**o+(0.264019654506)*x_ref[0]+(0.0314811616292)*x_ref[1]**o+(0.631481081422)*x_ref[1]
        ref[2,1]=(0.643576426484)*x_ref[0]**o+(-0.313599735242)*x_ref[0]+(0.908780876791)*x_ref[1]**o+(-0.698972268201)*x_ref[1]
        ref[2,2]=(-0.623208321993)*x_ref[0]**o+(0.649637800083)*x_ref[0]+(-0.878484988208)*x_ref[1]**o+(0.368793479691)*x_ref[1]
        ref[2,3]=(0.239333407828)*x_ref[0]**o+(-0.293558915555)*x_ref[0]+(0.271630238733)*x_ref[1]**o+(0.741304510722)*x_ref[1]
        ref[2,4]=(0.175914942104)*x_ref[0]**o+(0.336116863935)*x_ref[0]+(-0.799168065807)*x_ref[1]**o+(0.0264962830729)*x_ref[1]
        ref[3,0]=(0.777665875019)*x_ref[0]**o+(-0.13232137001)*x_ref[0]+(0.682046389381)*x_ref[1]**o+(0.589069623109)*x_ref[1]
        ref[3,1]=(0.7607621575)*x_ref[0]**o+(-0.81285460551)*x_ref[0]+(0.0523010659976)*x_ref[1]**o+(-0.913765717778)*x_ref[1]
        ref[3,2]=(0.822107753118)*x_ref[0]**o+(0.508869893322)*x_ref[0]+(0.679583348422)*x_ref[1]**o+(-0.597814039932)*x_ref[1]
        ref[3,3]=(-0.77699176705)*x_ref[0]**o+(-0.267758166676)*x_ref[0]+(0.661073127871)*x_ref[1]**o+(-0.41434545147)*x_ref[1]
        ref[3,4]=(0.223013380161)*x_ref[0]**o+(-0.558077040245)*x_ref[0]+(-0.450831764747)*x_ref[1]**o+(-0.0897765635191)*x_ref[1]
      else:
        arg[0,0]=(0.487697656095)*x[0]**o+(0.362509734966)*x[0]+(-0.40438484582)*x[1]**o+(-0.358984481681)*x[1]+(0.157670938098)*x[2]**o+(-0.631406375093)*x[2]
        arg[0,1]=(-0.00251154956677)*x[0]**o+(0.671447769464)*x[0]+(0.535541351526)*x[1]**o+(0.433508112485)*x[1]+(-0.897668718638)*x[2]**o+(0.141331586102)*x[2]
        arg[0,2]=(-0.964778505658)*x[0]**o+(0.522815935313)*x[0]+(-0.750318077186)*x[1]**o+(0.0595240411196)*x[1]+(-0.384021192511)*x[2]**o+(-0.802529657745)*x[2]
        arg[0,3]=(0.264781810221)*x[0]**o+(-0.654277693956)*x[0]+(0.890264421741)*x[1]**o+(-0.657743774786)*x[1]+(0.320234835664)*x[2]**o+(0.0637070094165)*x[2]
        arg[0,4]=(0.712229359775)*x[0]**o+(-0.646586719826)*x[0]+(0.876519277606)*x[1]**o+(0.473649630238)*x[1]+(0.332259187685)*x[2]**o+(-0.331489699179)*x[2]
        arg[1,0]=(0.808534274392)*x[0]**o+(-0.296462057802)*x[0]+(0.152536716655)*x[1]**o+(-0.990267302005)*x[1]+(-0.823462709316)*x[2]**o+(-0.51608634537)*x[2]
        arg[1,1]=(-0.745238552081)*x[0]**o+(-0.892086591023)*x[0]+(0.691429994644)*x[1]**o+(-0.52538570831)*x[1]+(-0.668959934842)*x[2]**o+(-0.5178179109)*x[2]
        arg[1,2]=(0.10173204456)*x[0]**o+(-0.853001564748)*x[0]+(0.617252846191)*x[1]**o+(-0.518228673297)*x[1]+(-0.995447484095)*x[2]**o+(0.56679145486)*x[2]
        arg[1,3]=(0.285145125389)*x[0]**o+(0.773362877732)*x[0]+(-0.457364007088)*x[1]**o+(0.226257217659)*x[1]+(0.439832733067)*x[2]**o+(0.872812128578)*x[2]
        arg[1,4]=(-0.581535304023)*x[0]**o+(-0.944599120493)*x[0]+(-0.720009453699)*x[1]**o+(-0.29624047891)*x[1]+(0.966528522407)*x[2]**o+(0.818945905214)*x[2]
        arg[2,0]=(-0.650732348774)*x[0]**o+(0.368909306978)*x[0]+(-0.608327978493)*x[1]**o+(0.301600904054)*x[1]+(-0.747648472883)*x[2]**o+(0.846568601061)*x[2]
        arg[2,1]=(-0.0221803961531)*x[0]**o+(-0.858842969825)*x[0]+(-0.40141447334)*x[1]**o+(0.408164536045)*x[1]+(-0.0131899768237)*x[2]**o+(-0.0983376041104)*x[2]
        arg[2,2]=(0.0855129797683)*x[0]**o+(0.0645200342375)*x[0]+(0.295916169508)*x[1]**o+(0.343169626483)*x[1]+(0.973477621396)*x[2]**o+(0.179505855846)*x[2]
        arg[2,3]=(-0.115714157482)*x[0]**o+(-0.00785628190789)*x[0]+(0.406498238895)*x[1]**o+(-0.954353977843)*x[1]+(0.751544148524)*x[2]**o+(-0.092307218389)*x[2]
        arg[2,4]=(0.807583813259)*x[0]**o+(-0.0150562208564)*x[0]+(0.967290753425)*x[1]**o+(0.960049095802)*x[1]+(0.41935259669)*x[2]**o+(0.878823415691)*x[2]
        arg[3,0]=(-0.162221321908)*x[0]**o+(0.0142404015457)*x[0]+(0.769189546147)*x[1]**o+(-0.251240080116)*x[1]+(0.980441953151)*x[2]**o+(-0.183639594778)*x[2]
        arg[3,1]=(0.270231471303)*x[0]**o+(-0.527471406108)*x[0]+(-0.255144134035)*x[1]**o+(0.268061599615)*x[1]+(-0.396373327005)*x[2]**o+(0.105905036411)*x[2]
        arg[3,2]=(0.5159632953)*x[0]**o+(-0.87397274133)*x[0]+(0.493876004047)*x[1]**o+(0.320996363656)*x[1]+(0.884345148533)*x[2]**o+(0.116262647026)*x[2]
        arg[3,3]=(-0.354053396048)*x[0]**o+(-0.130121845174)*x[0]+(-0.33114435645)*x[1]**o+(-0.73998189126)*x[1]+(0.362417530484)*x[2]**o+(0.676820875084)*x[2]
        arg[3,4]=(-0.53273881792)*x[0]**o+(-0.0624208514424)*x[0]+(0.680486778939)*x[1]**o+(-0.370666453954)*x[1]+(0.294345859395)*x[2]**o+(-0.679169213305)*x[2]
        ref[0,0]=(0.487697656095)*x_ref[0]**o+(0.362509734966)*x_ref[0]+(-0.40438484582)*x_ref[1]**o+(-0.358984481681)*x_ref[1]+(0.157670938098)*x_ref[2]**o+(-0.631406375093)*x_ref[2]
        ref[0,1]=(-0.00251154956677)*x_ref[0]**o+(0.671447769464)*x_ref[0]+(0.535541351526)*x_ref[1]**o+(0.433508112485)*x_ref[1]+(-0.897668718638)*x_ref[2]**o+(0.141331586102)*x_ref[2]
        ref[0,2]=(-0.964778505658)*x_ref[0]**o+(0.522815935313)*x_ref[0]+(-0.750318077186)*x_ref[1]**o+(0.0595240411196)*x_ref[1]+(-0.384021192511)*x_ref[2]**o+(-0.802529657745)*x_ref[2]
        ref[0,3]=(0.264781810221)*x_ref[0]**o+(-0.654277693956)*x_ref[0]+(0.890264421741)*x_ref[1]**o+(-0.657743774786)*x_ref[1]+(0.320234835664)*x_ref[2]**o+(0.0637070094165)*x_ref[2]
        ref[0,4]=(0.712229359775)*x_ref[0]**o+(-0.646586719826)*x_ref[0]+(0.876519277606)*x_ref[1]**o+(0.473649630238)*x_ref[1]+(0.332259187685)*x_ref[2]**o+(-0.331489699179)*x_ref[2]
        ref[1,0]=(0.808534274392)*x_ref[0]**o+(-0.296462057802)*x_ref[0]+(0.152536716655)*x_ref[1]**o+(-0.990267302005)*x_ref[1]+(-0.823462709316)*x_ref[2]**o+(-0.51608634537)*x_ref[2]
        ref[1,1]=(-0.745238552081)*x_ref[0]**o+(-0.892086591023)*x_ref[0]+(0.691429994644)*x_ref[1]**o+(-0.52538570831)*x_ref[1]+(-0.668959934842)*x_ref[2]**o+(-0.5178179109)*x_ref[2]
        ref[1,2]=(0.10173204456)*x_ref[0]**o+(-0.853001564748)*x_ref[0]+(0.617252846191)*x_ref[1]**o+(-0.518228673297)*x_ref[1]+(-0.995447484095)*x_ref[2]**o+(0.56679145486)*x_ref[2]
        ref[1,3]=(0.285145125389)*x_ref[0]**o+(0.773362877732)*x_ref[0]+(-0.457364007088)*x_ref[1]**o+(0.226257217659)*x_ref[1]+(0.439832733067)*x_ref[2]**o+(0.872812128578)*x_ref[2]
        ref[1,4]=(-0.581535304023)*x_ref[0]**o+(-0.944599120493)*x_ref[0]+(-0.720009453699)*x_ref[1]**o+(-0.29624047891)*x_ref[1]+(0.966528522407)*x_ref[2]**o+(0.818945905214)*x_ref[2]
        ref[2,0]=(-0.650732348774)*x_ref[0]**o+(0.368909306978)*x_ref[0]+(-0.608327978493)*x_ref[1]**o+(0.301600904054)*x_ref[1]+(-0.747648472883)*x_ref[2]**o+(0.846568601061)*x_ref[2]
        ref[2,1]=(-0.0221803961531)*x_ref[0]**o+(-0.858842969825)*x_ref[0]+(-0.40141447334)*x_ref[1]**o+(0.408164536045)*x_ref[1]+(-0.0131899768237)*x_ref[2]**o+(-0.0983376041104)*x_ref[2]
        ref[2,2]=(0.0855129797683)*x_ref[0]**o+(0.0645200342375)*x_ref[0]+(0.295916169508)*x_ref[1]**o+(0.343169626483)*x_ref[1]+(0.973477621396)*x_ref[2]**o+(0.179505855846)*x_ref[2]
        ref[2,3]=(-0.115714157482)*x_ref[0]**o+(-0.00785628190789)*x_ref[0]+(0.406498238895)*x_ref[1]**o+(-0.954353977843)*x_ref[1]+(0.751544148524)*x_ref[2]**o+(-0.092307218389)*x_ref[2]
        ref[2,4]=(0.807583813259)*x_ref[0]**o+(-0.0150562208564)*x_ref[0]+(0.967290753425)*x_ref[1]**o+(0.960049095802)*x_ref[1]+(0.41935259669)*x_ref[2]**o+(0.878823415691)*x_ref[2]
        ref[3,0]=(-0.162221321908)*x_ref[0]**o+(0.0142404015457)*x_ref[0]+(0.769189546147)*x_ref[1]**o+(-0.251240080116)*x_ref[1]+(0.980441953151)*x_ref[2]**o+(-0.183639594778)*x_ref[2]
        ref[3,1]=(0.270231471303)*x_ref[0]**o+(-0.527471406108)*x_ref[0]+(-0.255144134035)*x_ref[1]**o+(0.268061599615)*x_ref[1]+(-0.396373327005)*x_ref[2]**o+(0.105905036411)*x_ref[2]
        ref[3,2]=(0.5159632953)*x_ref[0]**o+(-0.87397274133)*x_ref[0]+(0.493876004047)*x_ref[1]**o+(0.320996363656)*x_ref[1]+(0.884345148533)*x_ref[2]**o+(0.116262647026)*x_ref[2]
        ref[3,3]=(-0.354053396048)*x_ref[0]**o+(-0.130121845174)*x_ref[0]+(-0.33114435645)*x_ref[1]**o+(-0.73998189126)*x_ref[1]+(0.362417530484)*x_ref[2]**o+(0.676820875084)*x_ref[2]
        ref[3,4]=(-0.53273881792)*x_ref[0]**o+(-0.0624208514424)*x_ref[0]+(0.680486778939)*x_ref[1]**o+(-0.370666453954)*x_ref[1]+(0.294345859395)*x_ref[2]**o+(-0.679169213305)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0906971676514)*x[0]**o+(-0.0640252128559)*x[0]+(-0.0108470120234)*x[1]**o+(-0.62531206416)*x[1]
        arg[0,0,1]=(-0.0186703191671)*x[0]**o+(0.256044044276)*x[0]+(-0.420130556042)*x[1]**o+(-0.241203381189)*x[1]
        arg[0,1,0]=(-0.287587258734)*x[0]**o+(-0.152487596827)*x[0]+(-0.554975489568)*x[1]**o+(0.713650618367)*x[1]
        arg[0,1,1]=(0.467597882001)*x[0]**o+(0.617291271522)*x[0]+(0.839893220642)*x[1]**o+(0.477553624935)*x[1]
        arg[1,0,0]=(0.292433268891)*x[0]**o+(0.0420372671524)*x[0]+(-0.342734694042)*x[1]**o+(0.892580432212)*x[1]
        arg[1,0,1]=(0.191399101606)*x[0]**o+(0.413781516804)*x[0]+(-0.346483382584)*x[1]**o+(0.805718474386)*x[1]
        arg[1,1,0]=(0.917644938285)*x[0]**o+(0.60457272962)*x[0]+(-0.0949129141155)*x[1]**o+(0.705971564397)*x[1]
        arg[1,1,1]=(0.327738300229)*x[0]**o+(-0.738951673731)*x[0]+(0.335869509312)*x[1]**o+(0.392167761004)*x[1]
        arg[2,0,0]=(0.460678533261)*x[0]**o+(-0.10665029868)*x[0]+(0.658148262201)*x[1]**o+(-0.75975084995)*x[1]
        arg[2,0,1]=(0.307057582515)*x[0]**o+(0.0755776536138)*x[0]+(0.866743758656)*x[1]**o+(-0.463537240812)*x[1]
        arg[2,1,0]=(-0.470241867396)*x[0]**o+(-0.0316070952691)*x[0]+(0.877786692055)*x[1]**o+(-0.750713976717)*x[1]
        arg[2,1,1]=(0.755903505788)*x[0]**o+(0.398472877018)*x[0]+(-0.372159129627)*x[1]**o+(0.170870288116)*x[1]
        arg[3,0,0]=(0.0310115714805)*x[0]**o+(0.822261427932)*x[0]+(-0.0979068243755)*x[1]**o+(0.61848414895)*x[1]
        arg[3,0,1]=(-0.442666645023)*x[0]**o+(0.759423291727)*x[0]+(0.771826722418)*x[1]**o+(-0.342283533203)*x[1]
        arg[3,1,0]=(0.292608821995)*x[0]**o+(0.301553824784)*x[0]+(0.445054135435)*x[1]**o+(0.990249711651)*x[1]
        arg[3,1,1]=(0.835714722542)*x[0]**o+(0.888569910076)*x[0]+(0.76442321847)*x[1]**o+(-0.272880995105)*x[1]
        arg[4,0,0]=(0.434738367563)*x[0]**o+(-0.801063604881)*x[0]+(-0.0152782263323)*x[1]**o+(-0.309467614762)*x[1]
        arg[4,0,1]=(-0.647140151015)*x[0]**o+(-0.0810020795911)*x[0]+(0.356402263616)*x[1]**o+(0.861692124435)*x[1]
        arg[4,1,0]=(-0.958501155541)*x[0]**o+(0.611720259342)*x[0]+(-0.631450146062)*x[1]**o+(-0.512668667768)*x[1]
        arg[4,1,1]=(0.771988547023)*x[0]**o+(0.466474334634)*x[0]+(-0.380283529607)*x[1]**o+(-0.619497160805)*x[1]
        arg[5,0,0]=(0.427386565443)*x[0]**o+(-0.178637075872)*x[0]+(0.81988184595)*x[1]**o+(-0.407798714033)*x[1]
        arg[5,0,1]=(-0.850161084571)*x[0]**o+(0.183440360862)*x[0]+(-0.83848298192)*x[1]**o+(-0.222248160259)*x[1]
        arg[5,1,0]=(-0.930623680879)*x[0]**o+(0.694991147173)*x[0]+(-0.299904374239)*x[1]**o+(0.552324198856)*x[1]
        arg[5,1,1]=(0.740715083219)*x[0]**o+(0.438768785937)*x[0]+(0.977523774483)*x[1]**o+(-0.668723413125)*x[1]
        ref[0,0,0]=(-0.0906971676514)*x_ref[0]**o+(-0.0640252128559)*x_ref[0]+(-0.0108470120234)*x_ref[1]**o+(-0.62531206416)*x_ref[1]
        ref[0,0,1]=(-0.0186703191671)*x_ref[0]**o+(0.256044044276)*x_ref[0]+(-0.420130556042)*x_ref[1]**o+(-0.241203381189)*x_ref[1]
        ref[0,1,0]=(-0.287587258734)*x_ref[0]**o+(-0.152487596827)*x_ref[0]+(-0.554975489568)*x_ref[1]**o+(0.713650618367)*x_ref[1]
        ref[0,1,1]=(0.467597882001)*x_ref[0]**o+(0.617291271522)*x_ref[0]+(0.839893220642)*x_ref[1]**o+(0.477553624935)*x_ref[1]
        ref[1,0,0]=(0.292433268891)*x_ref[0]**o+(0.0420372671524)*x_ref[0]+(-0.342734694042)*x_ref[1]**o+(0.892580432212)*x_ref[1]
        ref[1,0,1]=(0.191399101606)*x_ref[0]**o+(0.413781516804)*x_ref[0]+(-0.346483382584)*x_ref[1]**o+(0.805718474386)*x_ref[1]
        ref[1,1,0]=(0.917644938285)*x_ref[0]**o+(0.60457272962)*x_ref[0]+(-0.0949129141155)*x_ref[1]**o+(0.705971564397)*x_ref[1]
        ref[1,1,1]=(0.327738300229)*x_ref[0]**o+(-0.738951673731)*x_ref[0]+(0.335869509312)*x_ref[1]**o+(0.392167761004)*x_ref[1]
        ref[2,0,0]=(0.460678533261)*x_ref[0]**o+(-0.10665029868)*x_ref[0]+(0.658148262201)*x_ref[1]**o+(-0.75975084995)*x_ref[1]
        ref[2,0,1]=(0.307057582515)*x_ref[0]**o+(0.0755776536138)*x_ref[0]+(0.866743758656)*x_ref[1]**o+(-0.463537240812)*x_ref[1]
        ref[2,1,0]=(-0.470241867396)*x_ref[0]**o+(-0.0316070952691)*x_ref[0]+(0.877786692055)*x_ref[1]**o+(-0.750713976717)*x_ref[1]
        ref[2,1,1]=(0.755903505788)*x_ref[0]**o+(0.398472877018)*x_ref[0]+(-0.372159129627)*x_ref[1]**o+(0.170870288116)*x_ref[1]
        ref[3,0,0]=(0.0310115714805)*x_ref[0]**o+(0.822261427932)*x_ref[0]+(-0.0979068243755)*x_ref[1]**o+(0.61848414895)*x_ref[1]
        ref[3,0,1]=(-0.442666645023)*x_ref[0]**o+(0.759423291727)*x_ref[0]+(0.771826722418)*x_ref[1]**o+(-0.342283533203)*x_ref[1]
        ref[3,1,0]=(0.292608821995)*x_ref[0]**o+(0.301553824784)*x_ref[0]+(0.445054135435)*x_ref[1]**o+(0.990249711651)*x_ref[1]
        ref[3,1,1]=(0.835714722542)*x_ref[0]**o+(0.888569910076)*x_ref[0]+(0.76442321847)*x_ref[1]**o+(-0.272880995105)*x_ref[1]
        ref[4,0,0]=(0.434738367563)*x_ref[0]**o+(-0.801063604881)*x_ref[0]+(-0.0152782263323)*x_ref[1]**o+(-0.309467614762)*x_ref[1]
        ref[4,0,1]=(-0.647140151015)*x_ref[0]**o+(-0.0810020795911)*x_ref[0]+(0.356402263616)*x_ref[1]**o+(0.861692124435)*x_ref[1]
        ref[4,1,0]=(-0.958501155541)*x_ref[0]**o+(0.611720259342)*x_ref[0]+(-0.631450146062)*x_ref[1]**o+(-0.512668667768)*x_ref[1]
        ref[4,1,1]=(0.771988547023)*x_ref[0]**o+(0.466474334634)*x_ref[0]+(-0.380283529607)*x_ref[1]**o+(-0.619497160805)*x_ref[1]
        ref[5,0,0]=(0.427386565443)*x_ref[0]**o+(-0.178637075872)*x_ref[0]+(0.81988184595)*x_ref[1]**o+(-0.407798714033)*x_ref[1]
        ref[5,0,1]=(-0.850161084571)*x_ref[0]**o+(0.183440360862)*x_ref[0]+(-0.83848298192)*x_ref[1]**o+(-0.222248160259)*x_ref[1]
        ref[5,1,0]=(-0.930623680879)*x_ref[0]**o+(0.694991147173)*x_ref[0]+(-0.299904374239)*x_ref[1]**o+(0.552324198856)*x_ref[1]
        ref[5,1,1]=(0.740715083219)*x_ref[0]**o+(0.438768785937)*x_ref[0]+(0.977523774483)*x_ref[1]**o+(-0.668723413125)*x_ref[1]
      else:
        arg[0,0,0]=(0.0408197213005)*x[0]**o+(-0.992920054273)*x[0]+(0.122468233435)*x[1]**o+(-0.435942398199)*x[1]+(0.822235014621)*x[2]**o+(0.425925287623)*x[2]
        arg[0,0,1]=(0.474354220259)*x[0]**o+(0.287892840646)*x[0]+(-0.519818501933)*x[1]**o+(0.412301993774)*x[1]+(-0.242320576454)*x[2]**o+(0.758556860596)*x[2]
        arg[0,1,0]=(0.588735088154)*x[0]**o+(-0.948279493327)*x[0]+(0.194404465814)*x[1]**o+(-0.00761227943336)*x[1]+(-0.117173968067)*x[2]**o+(-0.228746139749)*x[2]
        arg[0,1,1]=(0.166448081217)*x[0]**o+(-0.228754542424)*x[0]+(-0.87739172573)*x[1]**o+(-0.592282861607)*x[1]+(-0.836732991313)*x[2]**o+(-0.20226535214)*x[2]
        arg[1,0,0]=(0.2683388552)*x[0]**o+(-0.733516131653)*x[0]+(-0.900887340495)*x[1]**o+(-0.614908158304)*x[1]+(0.422189467695)*x[2]**o+(0.642276887435)*x[2]
        arg[1,0,1]=(0.965294152134)*x[0]**o+(-0.326731862045)*x[0]+(-0.797754723728)*x[1]**o+(-0.678993698958)*x[1]+(0.256815891879)*x[2]**o+(0.48610716642)*x[2]
        arg[1,1,0]=(0.71383480118)*x[0]**o+(-0.949206151211)*x[0]+(0.40802633055)*x[1]**o+(-0.512457451544)*x[1]+(0.83647020826)*x[2]**o+(0.330550196742)*x[2]
        arg[1,1,1]=(-0.108602934331)*x[0]**o+(-0.585532075482)*x[0]+(0.302790940309)*x[1]**o+(-0.201628116557)*x[1]+(0.523440894937)*x[2]**o+(-0.421867218852)*x[2]
        arg[2,0,0]=(-0.911096130025)*x[0]**o+(-0.944970520542)*x[0]+(0.597179186191)*x[1]**o+(0.530539729209)*x[1]+(-0.538576718159)*x[2]**o+(-0.908863106348)*x[2]
        arg[2,0,1]=(0.828174522282)*x[0]**o+(-0.132534369339)*x[0]+(0.851849613713)*x[1]**o+(0.191494183049)*x[1]+(-0.953488097214)*x[2]**o+(0.423553741843)*x[2]
        arg[2,1,0]=(0.458631350156)*x[0]**o+(0.21755431391)*x[0]+(-0.128541259216)*x[1]**o+(-0.172831640065)*x[1]+(-0.62567145453)*x[2]**o+(-0.345626622399)*x[2]
        arg[2,1,1]=(0.49890508395)*x[0]**o+(-0.395084188812)*x[0]+(-0.653164498387)*x[1]**o+(-0.656416446654)*x[1]+(-0.844855683273)*x[2]**o+(-0.813755115861)*x[2]
        arg[3,0,0]=(-0.252710828698)*x[0]**o+(-0.0100429888666)*x[0]+(0.952930817098)*x[1]**o+(-0.950777038172)*x[1]+(0.244588721191)*x[2]**o+(-0.400442332557)*x[2]
        arg[3,0,1]=(-0.505331423555)*x[0]**o+(-0.535163478725)*x[0]+(0.924129836434)*x[1]**o+(-0.285813201016)*x[1]+(0.225037301542)*x[2]**o+(-0.331424134564)*x[2]
        arg[3,1,0]=(-0.556397106971)*x[0]**o+(0.26005371138)*x[0]+(0.902941320217)*x[1]**o+(0.70375648044)*x[1]+(-0.559371940837)*x[2]**o+(-0.628051748121)*x[2]
        arg[3,1,1]=(-0.644440017615)*x[0]**o+(-0.13415915731)*x[0]+(-0.752973804206)*x[1]**o+(-0.430304389409)*x[1]+(0.892979644749)*x[2]**o+(0.0618885081881)*x[2]
        arg[4,0,0]=(0.388902949971)*x[0]**o+(-0.413207160819)*x[0]+(-0.597892565551)*x[1]**o+(-0.433017605538)*x[1]+(0.369120396114)*x[2]**o+(-0.714185662873)*x[2]
        arg[4,0,1]=(0.140621693105)*x[0]**o+(0.910612147355)*x[0]+(0.404844175588)*x[1]**o+(0.963351416048)*x[1]+(0.545142156468)*x[2]**o+(-0.513154196146)*x[2]
        arg[4,1,0]=(-0.6045448854)*x[0]**o+(0.364599037279)*x[0]+(-0.165333666972)*x[1]**o+(-0.454715984949)*x[1]+(-0.123749300084)*x[2]**o+(0.23666517892)*x[2]
        arg[4,1,1]=(-0.701819332028)*x[0]**o+(-0.945269094128)*x[0]+(0.313745376139)*x[1]**o+(-0.0508160648325)*x[1]+(0.273150632944)*x[2]**o+(0.363230489878)*x[2]
        arg[5,0,0]=(-0.899477778287)*x[0]**o+(0.408300340496)*x[0]+(-0.501090863132)*x[1]**o+(-0.142172524799)*x[1]+(0.840930927028)*x[2]**o+(0.0786174984564)*x[2]
        arg[5,0,1]=(0.816863720765)*x[0]**o+(0.421613919321)*x[0]+(0.619636513942)*x[1]**o+(-0.867504595488)*x[1]+(-0.377345269009)*x[2]**o+(-0.884630348781)*x[2]
        arg[5,1,0]=(0.972307172208)*x[0]**o+(0.704673083035)*x[0]+(0.546566839332)*x[1]**o+(-0.303145527888)*x[1]+(-0.203985539635)*x[2]**o+(0.984097584463)*x[2]
        arg[5,1,1]=(0.67971485539)*x[0]**o+(0.492284313707)*x[0]+(0.139113989973)*x[1]**o+(-0.981414708174)*x[1]+(0.585926444541)*x[2]**o+(-0.790952419012)*x[2]
        ref[0,0,0]=(0.0408197213005)*x_ref[0]**o+(-0.992920054273)*x_ref[0]+(0.122468233435)*x_ref[1]**o+(-0.435942398199)*x_ref[1]+(0.822235014621)*x_ref[2]**o+(0.425925287623)*x_ref[2]
        ref[0,0,1]=(0.474354220259)*x_ref[0]**o+(0.287892840646)*x_ref[0]+(-0.519818501933)*x_ref[1]**o+(0.412301993774)*x_ref[1]+(-0.242320576454)*x_ref[2]**o+(0.758556860596)*x_ref[2]
        ref[0,1,0]=(0.588735088154)*x_ref[0]**o+(-0.948279493327)*x_ref[0]+(0.194404465814)*x_ref[1]**o+(-0.00761227943336)*x_ref[1]+(-0.117173968067)*x_ref[2]**o+(-0.228746139749)*x_ref[2]
        ref[0,1,1]=(0.166448081217)*x_ref[0]**o+(-0.228754542424)*x_ref[0]+(-0.87739172573)*x_ref[1]**o+(-0.592282861607)*x_ref[1]+(-0.836732991313)*x_ref[2]**o+(-0.20226535214)*x_ref[2]
        ref[1,0,0]=(0.2683388552)*x_ref[0]**o+(-0.733516131653)*x_ref[0]+(-0.900887340495)*x_ref[1]**o+(-0.614908158304)*x_ref[1]+(0.422189467695)*x_ref[2]**o+(0.642276887435)*x_ref[2]
        ref[1,0,1]=(0.965294152134)*x_ref[0]**o+(-0.326731862045)*x_ref[0]+(-0.797754723728)*x_ref[1]**o+(-0.678993698958)*x_ref[1]+(0.256815891879)*x_ref[2]**o+(0.48610716642)*x_ref[2]
        ref[1,1,0]=(0.71383480118)*x_ref[0]**o+(-0.949206151211)*x_ref[0]+(0.40802633055)*x_ref[1]**o+(-0.512457451544)*x_ref[1]+(0.83647020826)*x_ref[2]**o+(0.330550196742)*x_ref[2]
        ref[1,1,1]=(-0.108602934331)*x_ref[0]**o+(-0.585532075482)*x_ref[0]+(0.302790940309)*x_ref[1]**o+(-0.201628116557)*x_ref[1]+(0.523440894937)*x_ref[2]**o+(-0.421867218852)*x_ref[2]
        ref[2,0,0]=(-0.911096130025)*x_ref[0]**o+(-0.944970520542)*x_ref[0]+(0.597179186191)*x_ref[1]**o+(0.530539729209)*x_ref[1]+(-0.538576718159)*x_ref[2]**o+(-0.908863106348)*x_ref[2]
        ref[2,0,1]=(0.828174522282)*x_ref[0]**o+(-0.132534369339)*x_ref[0]+(0.851849613713)*x_ref[1]**o+(0.191494183049)*x_ref[1]+(-0.953488097214)*x_ref[2]**o+(0.423553741843)*x_ref[2]
        ref[2,1,0]=(0.458631350156)*x_ref[0]**o+(0.21755431391)*x_ref[0]+(-0.128541259216)*x_ref[1]**o+(-0.172831640065)*x_ref[1]+(-0.62567145453)*x_ref[2]**o+(-0.345626622399)*x_ref[2]
        ref[2,1,1]=(0.49890508395)*x_ref[0]**o+(-0.395084188812)*x_ref[0]+(-0.653164498387)*x_ref[1]**o+(-0.656416446654)*x_ref[1]+(-0.844855683273)*x_ref[2]**o+(-0.813755115861)*x_ref[2]
        ref[3,0,0]=(-0.252710828698)*x_ref[0]**o+(-0.0100429888666)*x_ref[0]+(0.952930817098)*x_ref[1]**o+(-0.950777038172)*x_ref[1]+(0.244588721191)*x_ref[2]**o+(-0.400442332557)*x_ref[2]
        ref[3,0,1]=(-0.505331423555)*x_ref[0]**o+(-0.535163478725)*x_ref[0]+(0.924129836434)*x_ref[1]**o+(-0.285813201016)*x_ref[1]+(0.225037301542)*x_ref[2]**o+(-0.331424134564)*x_ref[2]
        ref[3,1,0]=(-0.556397106971)*x_ref[0]**o+(0.26005371138)*x_ref[0]+(0.902941320217)*x_ref[1]**o+(0.70375648044)*x_ref[1]+(-0.559371940837)*x_ref[2]**o+(-0.628051748121)*x_ref[2]
        ref[3,1,1]=(-0.644440017615)*x_ref[0]**o+(-0.13415915731)*x_ref[0]+(-0.752973804206)*x_ref[1]**o+(-0.430304389409)*x_ref[1]+(0.892979644749)*x_ref[2]**o+(0.0618885081881)*x_ref[2]
        ref[4,0,0]=(0.388902949971)*x_ref[0]**o+(-0.413207160819)*x_ref[0]+(-0.597892565551)*x_ref[1]**o+(-0.433017605538)*x_ref[1]+(0.369120396114)*x_ref[2]**o+(-0.714185662873)*x_ref[2]
        ref[4,0,1]=(0.140621693105)*x_ref[0]**o+(0.910612147355)*x_ref[0]+(0.404844175588)*x_ref[1]**o+(0.963351416048)*x_ref[1]+(0.545142156468)*x_ref[2]**o+(-0.513154196146)*x_ref[2]
        ref[4,1,0]=(-0.6045448854)*x_ref[0]**o+(0.364599037279)*x_ref[0]+(-0.165333666972)*x_ref[1]**o+(-0.454715984949)*x_ref[1]+(-0.123749300084)*x_ref[2]**o+(0.23666517892)*x_ref[2]
        ref[4,1,1]=(-0.701819332028)*x_ref[0]**o+(-0.945269094128)*x_ref[0]+(0.313745376139)*x_ref[1]**o+(-0.0508160648325)*x_ref[1]+(0.273150632944)*x_ref[2]**o+(0.363230489878)*x_ref[2]
        ref[5,0,0]=(-0.899477778287)*x_ref[0]**o+(0.408300340496)*x_ref[0]+(-0.501090863132)*x_ref[1]**o+(-0.142172524799)*x_ref[1]+(0.840930927028)*x_ref[2]**o+(0.0786174984564)*x_ref[2]
        ref[5,0,1]=(0.816863720765)*x_ref[0]**o+(0.421613919321)*x_ref[0]+(0.619636513942)*x_ref[1]**o+(-0.867504595488)*x_ref[1]+(-0.377345269009)*x_ref[2]**o+(-0.884630348781)*x_ref[2]
        ref[5,1,0]=(0.972307172208)*x_ref[0]**o+(0.704673083035)*x_ref[0]+(0.546566839332)*x_ref[1]**o+(-0.303145527888)*x_ref[1]+(-0.203985539635)*x_ref[2]**o+(0.984097584463)*x_ref[2]
        ref[5,1,1]=(0.67971485539)*x_ref[0]**o+(0.492284313707)*x_ref[0]+(0.139113989973)*x_ref[1]**o+(-0.981414708174)*x_ref[1]+(0.585926444541)*x_ref[2]**o+(-0.790952419012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.717741572212)*x[0]**o+(-0.771801794802)*x[0]+(-0.782201793954)*x[1]**o+(0.125253178673)*x[1]
        arg[0,0,0,1]=(-0.483339499341)*x[0]**o+(0.0177945970437)*x[0]+(0.233532060933)*x[1]**o+(-0.465948890703)*x[1]
        arg[0,0,1,0]=(0.140544455071)*x[0]**o+(-0.120145399579)*x[0]+(-0.0732045826929)*x[1]**o+(-0.738165902415)*x[1]
        arg[0,0,1,1]=(0.98057407477)*x[0]**o+(0.346125329427)*x[0]+(0.697386820108)*x[1]**o+(0.402406370862)*x[1]
        arg[0,0,2,0]=(0.347641628204)*x[0]**o+(-0.217938924727)*x[0]+(-0.0572727430966)*x[1]**o+(0.170631149745)*x[1]
        arg[0,0,2,1]=(0.621840702108)*x[0]**o+(-0.109219941839)*x[0]+(0.894052962811)*x[1]**o+(0.423894680808)*x[1]
        arg[0,1,0,0]=(0.0339315069567)*x[0]**o+(-0.655546201957)*x[0]+(-0.271776247317)*x[1]**o+(0.4760527815)*x[1]
        arg[0,1,0,1]=(0.904216237912)*x[0]**o+(0.974869309983)*x[0]+(0.678379481202)*x[1]**o+(0.602534020957)*x[1]
        arg[0,1,1,0]=(-0.98805028173)*x[0]**o+(0.255251214419)*x[0]+(0.659010647624)*x[1]**o+(-0.45360685971)*x[1]
        arg[0,1,1,1]=(0.0884339867328)*x[0]**o+(-0.325188714625)*x[0]+(0.158562643552)*x[1]**o+(-0.788600388567)*x[1]
        arg[0,1,2,0]=(0.976404407593)*x[0]**o+(0.353231650963)*x[0]+(0.546951413477)*x[1]**o+(0.0738098551899)*x[1]
        arg[0,1,2,1]=(-0.812156797654)*x[0]**o+(-0.0180014270879)*x[0]+(0.686011517957)*x[1]**o+(0.148522434365)*x[1]
        arg[0,2,0,0]=(-0.40018516407)*x[0]**o+(-0.0512695503575)*x[0]+(-0.715512784143)*x[1]**o+(0.111064716031)*x[1]
        arg[0,2,0,1]=(-0.963397739298)*x[0]**o+(0.406951519848)*x[0]+(-0.293539028967)*x[1]**o+(-0.549328925146)*x[1]
        arg[0,2,1,0]=(-0.00827402449089)*x[0]**o+(-0.657982352244)*x[0]+(-0.515966831052)*x[1]**o+(0.531853371259)*x[1]
        arg[0,2,1,1]=(0.0379862601491)*x[0]**o+(0.988691147833)*x[0]+(0.820623911484)*x[1]**o+(0.689064076022)*x[1]
        arg[0,2,2,0]=(-0.857638654259)*x[0]**o+(0.664968676908)*x[0]+(-0.228424264597)*x[1]**o+(0.740003679791)*x[1]
        arg[0,2,2,1]=(0.456486069586)*x[0]**o+(-0.510130176146)*x[0]+(-0.189525160745)*x[1]**o+(0.497975981413)*x[1]
        arg[0,3,0,0]=(-0.981633737105)*x[0]**o+(0.25636883881)*x[0]+(-0.0477676975467)*x[1]**o+(0.654475880014)*x[1]
        arg[0,3,0,1]=(-0.0710284929969)*x[0]**o+(0.338590765191)*x[0]+(0.639048913997)*x[1]**o+(-0.24999606762)*x[1]
        arg[0,3,1,0]=(-0.731484482659)*x[0]**o+(-0.808253846389)*x[0]+(-0.911723431862)*x[1]**o+(-0.266029942425)*x[1]
        arg[0,3,1,1]=(0.909037857376)*x[0]**o+(-0.164525039433)*x[0]+(-0.924332370345)*x[1]**o+(0.476479399362)*x[1]
        arg[0,3,2,0]=(0.00534173833653)*x[0]**o+(-0.435160515142)*x[0]+(0.402701511687)*x[1]**o+(0.161631276653)*x[1]
        arg[0,3,2,1]=(-0.595791669027)*x[0]**o+(0.170580206948)*x[0]+(0.468802250813)*x[1]**o+(-0.881472636722)*x[1]
        arg[0,4,0,0]=(-0.0739506349513)*x[0]**o+(0.513385505957)*x[0]+(0.046977740371)*x[1]**o+(0.95763159987)*x[1]
        arg[0,4,0,1]=(-0.286789006223)*x[0]**o+(0.017476293519)*x[0]+(0.637378781092)*x[1]**o+(-0.670201122126)*x[1]
        arg[0,4,1,0]=(0.159607100455)*x[0]**o+(0.715341910227)*x[0]+(0.949546645819)*x[1]**o+(0.362982182775)*x[1]
        arg[0,4,1,1]=(-0.237702024358)*x[0]**o+(-0.520488841454)*x[0]+(0.606333933569)*x[1]**o+(-0.00662757701049)*x[1]
        arg[0,4,2,0]=(-0.415753817901)*x[0]**o+(0.731285464784)*x[0]+(0.15987336144)*x[1]**o+(0.0697706236338)*x[1]
        arg[0,4,2,1]=(0.0467482158427)*x[0]**o+(0.306004722736)*x[0]+(-0.905184725692)*x[1]**o+(-0.158017889366)*x[1]
        arg[1,0,0,0]=(-0.785535082426)*x[0]**o+(-0.189285100452)*x[0]+(-0.766349940057)*x[1]**o+(0.531469308995)*x[1]
        arg[1,0,0,1]=(0.522213196477)*x[0]**o+(-0.661381741437)*x[0]+(0.816089886956)*x[1]**o+(0.139829382305)*x[1]
        arg[1,0,1,0]=(0.0957295570417)*x[0]**o+(-0.105252674536)*x[0]+(-0.569429953382)*x[1]**o+(-0.801934666657)*x[1]
        arg[1,0,1,1]=(-0.557307051058)*x[0]**o+(0.218852576416)*x[0]+(0.430167216464)*x[1]**o+(-0.112285681761)*x[1]
        arg[1,0,2,0]=(0.288472422477)*x[0]**o+(0.294377994763)*x[0]+(-0.809503961511)*x[1]**o+(0.357020226991)*x[1]
        arg[1,0,2,1]=(-0.148113477705)*x[0]**o+(0.164468646294)*x[0]+(0.7977142476)*x[1]**o+(-0.407274933384)*x[1]
        arg[1,1,0,0]=(-0.926374434822)*x[0]**o+(-0.107713463193)*x[0]+(-0.511635723869)*x[1]**o+(0.259319717689)*x[1]
        arg[1,1,0,1]=(-0.18636582243)*x[0]**o+(0.698258973628)*x[0]+(0.234379281397)*x[1]**o+(0.596195875495)*x[1]
        arg[1,1,1,0]=(-0.180395892006)*x[0]**o+(0.631248500084)*x[0]+(0.104751616377)*x[1]**o+(0.204500206653)*x[1]
        arg[1,1,1,1]=(-0.517119564238)*x[0]**o+(-0.16634015634)*x[0]+(0.556785586675)*x[1]**o+(0.42659306392)*x[1]
        arg[1,1,2,0]=(0.163807238958)*x[0]**o+(-0.534619478446)*x[0]+(-0.700353460753)*x[1]**o+(0.121393131648)*x[1]
        arg[1,1,2,1]=(0.964276278961)*x[0]**o+(0.846030751304)*x[0]+(-0.745301894392)*x[1]**o+(0.33821485558)*x[1]
        arg[1,2,0,0]=(0.333173101086)*x[0]**o+(0.774848048743)*x[0]+(-0.659091916357)*x[1]**o+(-0.797247144497)*x[1]
        arg[1,2,0,1]=(0.0581394534285)*x[0]**o+(-0.524771663031)*x[0]+(-0.519465753317)*x[1]**o+(0.71748795226)*x[1]
        arg[1,2,1,0]=(0.224935262833)*x[0]**o+(-0.393399819306)*x[0]+(0.199349075224)*x[1]**o+(-0.808847009412)*x[1]
        arg[1,2,1,1]=(0.0385750795521)*x[0]**o+(-8.09949553151e-05)*x[0]+(0.496860964506)*x[1]**o+(0.0119507407829)*x[1]
        arg[1,2,2,0]=(0.876341556167)*x[0]**o+(0.18082766372)*x[0]+(0.401349871827)*x[1]**o+(-0.543439037303)*x[1]
        arg[1,2,2,1]=(-0.894525004507)*x[0]**o+(-0.516771143695)*x[0]+(0.0410916490263)*x[1]**o+(0.753841618439)*x[1]
        arg[1,3,0,0]=(0.251979970924)*x[0]**o+(0.790276550438)*x[0]+(-0.155708793888)*x[1]**o+(0.28971278336)*x[1]
        arg[1,3,0,1]=(0.258336265436)*x[0]**o+(0.696642875149)*x[0]+(0.0561355915649)*x[1]**o+(0.601839744417)*x[1]
        arg[1,3,1,0]=(-0.978729102785)*x[0]**o+(0.335280613262)*x[0]+(0.89228798232)*x[1]**o+(-0.693050679348)*x[1]
        arg[1,3,1,1]=(-0.885817753431)*x[0]**o+(-0.0641571535708)*x[0]+(-0.649578576181)*x[1]**o+(-0.95773137602)*x[1]
        arg[1,3,2,0]=(-0.333891001461)*x[0]**o+(-0.496064308564)*x[0]+(0.438549603377)*x[1]**o+(0.0772468142672)*x[1]
        arg[1,3,2,1]=(-0.16171788255)*x[0]**o+(-0.939293453222)*x[0]+(-0.284522648546)*x[1]**o+(0.744642328375)*x[1]
        arg[1,4,0,0]=(0.806133262332)*x[0]**o+(-0.0225320206432)*x[0]+(0.125958642606)*x[1]**o+(0.610630249467)*x[1]
        arg[1,4,0,1]=(0.657971870946)*x[0]**o+(0.698894701525)*x[0]+(-0.595952612496)*x[1]**o+(-0.38268841292)*x[1]
        arg[1,4,1,0]=(-0.277274411331)*x[0]**o+(-0.483183004514)*x[0]+(0.77656657009)*x[1]**o+(-0.959290466645)*x[1]
        arg[1,4,1,1]=(-0.96384356305)*x[0]**o+(-0.00383301781476)*x[0]+(0.788010409608)*x[1]**o+(0.135378744316)*x[1]
        arg[1,4,2,0]=(0.986761958017)*x[0]**o+(-0.570276378038)*x[0]+(-0.515784028315)*x[1]**o+(0.487125123439)*x[1]
        arg[1,4,2,1]=(-0.954735054702)*x[0]**o+(-0.0610323074217)*x[0]+(-0.255284859281)*x[1]**o+(0.277795725698)*x[1]
        arg[2,0,0,0]=(-0.708301219324)*x[0]**o+(-0.970166781553)*x[0]+(-0.0836451425855)*x[1]**o+(0.525854202995)*x[1]
        arg[2,0,0,1]=(-0.0995393865183)*x[0]**o+(-0.0156624820324)*x[0]+(0.225529316197)*x[1]**o+(0.142617834056)*x[1]
        arg[2,0,1,0]=(-0.585329524103)*x[0]**o+(0.590285924441)*x[0]+(0.201483726322)*x[1]**o+(-0.735262197689)*x[1]
        arg[2,0,1,1]=(-0.682877182963)*x[0]**o+(-0.162682262799)*x[0]+(-0.890051459843)*x[1]**o+(-0.0188015866765)*x[1]
        arg[2,0,2,0]=(0.868935476294)*x[0]**o+(-0.629891068854)*x[0]+(0.335632865663)*x[1]**o+(-0.955849055716)*x[1]
        arg[2,0,2,1]=(-0.907933228238)*x[0]**o+(-0.105861987974)*x[0]+(-0.131372383042)*x[1]**o+(-0.970114136854)*x[1]
        arg[2,1,0,0]=(-0.275788501814)*x[0]**o+(-0.992861867033)*x[0]+(-0.80676324962)*x[1]**o+(0.269244680224)*x[1]
        arg[2,1,0,1]=(-0.704704138774)*x[0]**o+(-0.371803193009)*x[0]+(-0.0190586405922)*x[1]**o+(-0.0828837624698)*x[1]
        arg[2,1,1,0]=(-0.532645566013)*x[0]**o+(0.571754809868)*x[0]+(0.656737493773)*x[1]**o+(-0.548229591842)*x[1]
        arg[2,1,1,1]=(0.23836921087)*x[0]**o+(0.707475737785)*x[0]+(-0.59660155968)*x[1]**o+(-0.450204951071)*x[1]
        arg[2,1,2,0]=(-0.534027348014)*x[0]**o+(0.332100193727)*x[0]+(-0.864157138567)*x[1]**o+(-0.41833267066)*x[1]
        arg[2,1,2,1]=(0.459878593269)*x[0]**o+(0.446259140228)*x[0]+(0.571242887502)*x[1]**o+(-0.0390525669486)*x[1]
        arg[2,2,0,0]=(0.322052799527)*x[0]**o+(-0.915265209304)*x[0]+(-0.99942007618)*x[1]**o+(0.482294513743)*x[1]
        arg[2,2,0,1]=(0.558489194456)*x[0]**o+(0.00400576554086)*x[0]+(-0.87484330165)*x[1]**o+(-0.664504218977)*x[1]
        arg[2,2,1,0]=(0.669621607461)*x[0]**o+(-0.0450960554319)*x[0]+(-0.221151480246)*x[1]**o+(0.0123621737211)*x[1]
        arg[2,2,1,1]=(0.476743013029)*x[0]**o+(0.11118371822)*x[0]+(-0.202074315677)*x[1]**o+(0.377649433458)*x[1]
        arg[2,2,2,0]=(0.667910543701)*x[0]**o+(0.035703922547)*x[0]+(-0.274733967095)*x[1]**o+(-0.998934720182)*x[1]
        arg[2,2,2,1]=(0.578909660087)*x[0]**o+(0.149900086514)*x[0]+(-0.226597651858)*x[1]**o+(0.169175812287)*x[1]
        arg[2,3,0,0]=(0.158370246586)*x[0]**o+(0.595643112285)*x[0]+(-0.476875564052)*x[1]**o+(0.869752657517)*x[1]
        arg[2,3,0,1]=(0.99718847714)*x[0]**o+(0.581531822297)*x[0]+(-0.607102775564)*x[1]**o+(-0.8317970284)*x[1]
        arg[2,3,1,0]=(-0.27390195745)*x[0]**o+(0.0512853330414)*x[0]+(0.465190666302)*x[1]**o+(0.042533388788)*x[1]
        arg[2,3,1,1]=(-0.183679504962)*x[0]**o+(-0.0522782076121)*x[0]+(0.921423118529)*x[1]**o+(-0.0315080275225)*x[1]
        arg[2,3,2,0]=(0.0421842348778)*x[0]**o+(0.462023049702)*x[0]+(-0.888342606803)*x[1]**o+(-0.956503447884)*x[1]
        arg[2,3,2,1]=(-0.950820288444)*x[0]**o+(0.104271990168)*x[0]+(-0.607212909582)*x[1]**o+(-0.16283931259)*x[1]
        arg[2,4,0,0]=(0.113505933533)*x[0]**o+(0.27873514806)*x[0]+(0.736699443523)*x[1]**o+(-0.951737555693)*x[1]
        arg[2,4,0,1]=(0.264170584849)*x[0]**o+(-0.226649236617)*x[0]+(0.489902180874)*x[1]**o+(-0.191536591762)*x[1]
        arg[2,4,1,0]=(0.712322285985)*x[0]**o+(-0.323395455815)*x[0]+(-0.298356681586)*x[1]**o+(0.699891974541)*x[1]
        arg[2,4,1,1]=(-0.631057763255)*x[0]**o+(-0.489374933445)*x[0]+(-0.021591806874)*x[1]**o+(-0.692800124585)*x[1]
        arg[2,4,2,0]=(0.916068935184)*x[0]**o+(0.33779688858)*x[0]+(0.06300443559)*x[1]**o+(-0.231385203399)*x[1]
        arg[2,4,2,1]=(-0.283485044625)*x[0]**o+(0.548366574442)*x[0]+(-0.543953132872)*x[1]**o+(-0.956662823037)*x[1]
        arg[3,0,0,0]=(0.709349274221)*x[0]**o+(-0.342981009263)*x[0]+(-0.682331761929)*x[1]**o+(0.673169884813)*x[1]
        arg[3,0,0,1]=(0.236249836593)*x[0]**o+(-0.682845054434)*x[0]+(-0.349169988218)*x[1]**o+(0.502684425824)*x[1]
        arg[3,0,1,0]=(-0.561655748793)*x[0]**o+(0.120183772326)*x[0]+(-0.966535325583)*x[1]**o+(-0.0585117264949)*x[1]
        arg[3,0,1,1]=(-0.661030287189)*x[0]**o+(-0.322924086505)*x[0]+(-0.986892059552)*x[1]**o+(-0.778969361365)*x[1]
        arg[3,0,2,0]=(-0.166307107307)*x[0]**o+(0.612931303617)*x[0]+(-0.283925165456)*x[1]**o+(0.5998038814)*x[1]
        arg[3,0,2,1]=(0.205099355916)*x[0]**o+(-0.520970536278)*x[0]+(0.900416833183)*x[1]**o+(0.981826306069)*x[1]
        arg[3,1,0,0]=(-0.875306666443)*x[0]**o+(0.0189358479861)*x[0]+(0.101187857445)*x[1]**o+(-0.873584307534)*x[1]
        arg[3,1,0,1]=(-0.0203656235757)*x[0]**o+(-0.579229777042)*x[0]+(-0.0799345631826)*x[1]**o+(0.896315444069)*x[1]
        arg[3,1,1,0]=(0.576791459982)*x[0]**o+(-0.86541815506)*x[0]+(-0.582764176207)*x[1]**o+(-0.249437395011)*x[1]
        arg[3,1,1,1]=(0.525953680497)*x[0]**o+(-0.222445717402)*x[0]+(-0.87710202063)*x[1]**o+(-0.762921821973)*x[1]
        arg[3,1,2,0]=(-0.883737185908)*x[0]**o+(0.254053241277)*x[0]+(0.559803298059)*x[1]**o+(-0.577376956856)*x[1]
        arg[3,1,2,1]=(-0.125792083387)*x[0]**o+(-0.771549307509)*x[0]+(0.142493480552)*x[1]**o+(0.459996962548)*x[1]
        arg[3,2,0,0]=(0.0531598927049)*x[0]**o+(-0.950343229037)*x[0]+(-0.748240453154)*x[1]**o+(0.850555892342)*x[1]
        arg[3,2,0,1]=(-0.660852152041)*x[0]**o+(-0.763119278604)*x[0]+(-0.811925409663)*x[1]**o+(-0.178765071495)*x[1]
        arg[3,2,1,0]=(-0.189124223305)*x[0]**o+(-0.410392909042)*x[0]+(0.993247167522)*x[1]**o+(0.291451796298)*x[1]
        arg[3,2,1,1]=(0.83496133421)*x[0]**o+(0.695147480265)*x[0]+(0.444420524946)*x[1]**o+(0.107454620114)*x[1]
        arg[3,2,2,0]=(0.94687479313)*x[0]**o+(-0.140702324395)*x[0]+(0.149203502428)*x[1]**o+(-0.338246126345)*x[1]
        arg[3,2,2,1]=(0.147609394709)*x[0]**o+(-0.549219765239)*x[0]+(0.397308993347)*x[1]**o+(0.0402189312685)*x[1]
        arg[3,3,0,0]=(0.27400981409)*x[0]**o+(-0.516367266257)*x[0]+(-0.784095952814)*x[1]**o+(0.32858899009)*x[1]
        arg[3,3,0,1]=(-0.393457382163)*x[0]**o+(0.590406794051)*x[0]+(0.328523918357)*x[1]**o+(0.899537513355)*x[1]
        arg[3,3,1,0]=(-0.851038623366)*x[0]**o+(0.596974864726)*x[0]+(-0.338630687551)*x[1]**o+(0.634961435042)*x[1]
        arg[3,3,1,1]=(0.569166143918)*x[0]**o+(-0.388162282491)*x[0]+(0.764830536733)*x[1]**o+(-0.426737342542)*x[1]
        arg[3,3,2,0]=(0.502221903448)*x[0]**o+(0.940382225897)*x[0]+(0.563444730579)*x[1]**o+(0.023210869151)*x[1]
        arg[3,3,2,1]=(0.061695403852)*x[0]**o+(0.868302188758)*x[0]+(0.0855803325758)*x[1]**o+(-0.24366173638)*x[1]
        arg[3,4,0,0]=(-0.533038182387)*x[0]**o+(0.0222388915737)*x[0]+(0.529037752769)*x[1]**o+(-0.0665320895494)*x[1]
        arg[3,4,0,1]=(0.714374055253)*x[0]**o+(-0.623670743333)*x[0]+(-0.866024792929)*x[1]**o+(-0.0673124896111)*x[1]
        arg[3,4,1,0]=(-0.0523466615045)*x[0]**o+(-0.472563426622)*x[0]+(0.139892673626)*x[1]**o+(0.539655417178)*x[1]
        arg[3,4,1,1]=(-0.473032580514)*x[0]**o+(0.692812093618)*x[0]+(0.664400517742)*x[1]**o+(0.184125423364)*x[1]
        arg[3,4,2,0]=(-0.0340798058725)*x[0]**o+(-0.58900562863)*x[0]+(0.396856384717)*x[1]**o+(-0.559175428019)*x[1]
        arg[3,4,2,1]=(-0.217679376634)*x[0]**o+(-0.90389036667)*x[0]+(0.967372942873)*x[1]**o+(-0.482913619277)*x[1]
        ref[0,0,0,0]=(-0.717741572212)*x_ref[0]**o+(-0.771801794802)*x_ref[0]+(-0.782201793954)*x_ref[1]**o+(0.125253178673)*x_ref[1]
        ref[0,0,0,1]=(-0.483339499341)*x_ref[0]**o+(0.0177945970437)*x_ref[0]+(0.233532060933)*x_ref[1]**o+(-0.465948890703)*x_ref[1]
        ref[0,0,1,0]=(0.140544455071)*x_ref[0]**o+(-0.120145399579)*x_ref[0]+(-0.0732045826929)*x_ref[1]**o+(-0.738165902415)*x_ref[1]
        ref[0,0,1,1]=(0.98057407477)*x_ref[0]**o+(0.346125329427)*x_ref[0]+(0.697386820108)*x_ref[1]**o+(0.402406370862)*x_ref[1]
        ref[0,0,2,0]=(0.347641628204)*x_ref[0]**o+(-0.217938924727)*x_ref[0]+(-0.0572727430966)*x_ref[1]**o+(0.170631149745)*x_ref[1]
        ref[0,0,2,1]=(0.621840702108)*x_ref[0]**o+(-0.109219941839)*x_ref[0]+(0.894052962811)*x_ref[1]**o+(0.423894680808)*x_ref[1]
        ref[0,1,0,0]=(0.0339315069567)*x_ref[0]**o+(-0.655546201957)*x_ref[0]+(-0.271776247317)*x_ref[1]**o+(0.4760527815)*x_ref[1]
        ref[0,1,0,1]=(0.904216237912)*x_ref[0]**o+(0.974869309983)*x_ref[0]+(0.678379481202)*x_ref[1]**o+(0.602534020957)*x_ref[1]
        ref[0,1,1,0]=(-0.98805028173)*x_ref[0]**o+(0.255251214419)*x_ref[0]+(0.659010647624)*x_ref[1]**o+(-0.45360685971)*x_ref[1]
        ref[0,1,1,1]=(0.0884339867328)*x_ref[0]**o+(-0.325188714625)*x_ref[0]+(0.158562643552)*x_ref[1]**o+(-0.788600388567)*x_ref[1]
        ref[0,1,2,0]=(0.976404407593)*x_ref[0]**o+(0.353231650963)*x_ref[0]+(0.546951413477)*x_ref[1]**o+(0.0738098551899)*x_ref[1]
        ref[0,1,2,1]=(-0.812156797654)*x_ref[0]**o+(-0.0180014270879)*x_ref[0]+(0.686011517957)*x_ref[1]**o+(0.148522434365)*x_ref[1]
        ref[0,2,0,0]=(-0.40018516407)*x_ref[0]**o+(-0.0512695503575)*x_ref[0]+(-0.715512784143)*x_ref[1]**o+(0.111064716031)*x_ref[1]
        ref[0,2,0,1]=(-0.963397739298)*x_ref[0]**o+(0.406951519848)*x_ref[0]+(-0.293539028967)*x_ref[1]**o+(-0.549328925146)*x_ref[1]
        ref[0,2,1,0]=(-0.00827402449089)*x_ref[0]**o+(-0.657982352244)*x_ref[0]+(-0.515966831052)*x_ref[1]**o+(0.531853371259)*x_ref[1]
        ref[0,2,1,1]=(0.0379862601491)*x_ref[0]**o+(0.988691147833)*x_ref[0]+(0.820623911484)*x_ref[1]**o+(0.689064076022)*x_ref[1]
        ref[0,2,2,0]=(-0.857638654259)*x_ref[0]**o+(0.664968676908)*x_ref[0]+(-0.228424264597)*x_ref[1]**o+(0.740003679791)*x_ref[1]
        ref[0,2,2,1]=(0.456486069586)*x_ref[0]**o+(-0.510130176146)*x_ref[0]+(-0.189525160745)*x_ref[1]**o+(0.497975981413)*x_ref[1]
        ref[0,3,0,0]=(-0.981633737105)*x_ref[0]**o+(0.25636883881)*x_ref[0]+(-0.0477676975467)*x_ref[1]**o+(0.654475880014)*x_ref[1]
        ref[0,3,0,1]=(-0.0710284929969)*x_ref[0]**o+(0.338590765191)*x_ref[0]+(0.639048913997)*x_ref[1]**o+(-0.24999606762)*x_ref[1]
        ref[0,3,1,0]=(-0.731484482659)*x_ref[0]**o+(-0.808253846389)*x_ref[0]+(-0.911723431862)*x_ref[1]**o+(-0.266029942425)*x_ref[1]
        ref[0,3,1,1]=(0.909037857376)*x_ref[0]**o+(-0.164525039433)*x_ref[0]+(-0.924332370345)*x_ref[1]**o+(0.476479399362)*x_ref[1]
        ref[0,3,2,0]=(0.00534173833653)*x_ref[0]**o+(-0.435160515142)*x_ref[0]+(0.402701511687)*x_ref[1]**o+(0.161631276653)*x_ref[1]
        ref[0,3,2,1]=(-0.595791669027)*x_ref[0]**o+(0.170580206948)*x_ref[0]+(0.468802250813)*x_ref[1]**o+(-0.881472636722)*x_ref[1]
        ref[0,4,0,0]=(-0.0739506349513)*x_ref[0]**o+(0.513385505957)*x_ref[0]+(0.046977740371)*x_ref[1]**o+(0.95763159987)*x_ref[1]
        ref[0,4,0,1]=(-0.286789006223)*x_ref[0]**o+(0.017476293519)*x_ref[0]+(0.637378781092)*x_ref[1]**o+(-0.670201122126)*x_ref[1]
        ref[0,4,1,0]=(0.159607100455)*x_ref[0]**o+(0.715341910227)*x_ref[0]+(0.949546645819)*x_ref[1]**o+(0.362982182775)*x_ref[1]
        ref[0,4,1,1]=(-0.237702024358)*x_ref[0]**o+(-0.520488841454)*x_ref[0]+(0.606333933569)*x_ref[1]**o+(-0.00662757701049)*x_ref[1]
        ref[0,4,2,0]=(-0.415753817901)*x_ref[0]**o+(0.731285464784)*x_ref[0]+(0.15987336144)*x_ref[1]**o+(0.0697706236338)*x_ref[1]
        ref[0,4,2,1]=(0.0467482158427)*x_ref[0]**o+(0.306004722736)*x_ref[0]+(-0.905184725692)*x_ref[1]**o+(-0.158017889366)*x_ref[1]
        ref[1,0,0,0]=(-0.785535082426)*x_ref[0]**o+(-0.189285100452)*x_ref[0]+(-0.766349940057)*x_ref[1]**o+(0.531469308995)*x_ref[1]
        ref[1,0,0,1]=(0.522213196477)*x_ref[0]**o+(-0.661381741437)*x_ref[0]+(0.816089886956)*x_ref[1]**o+(0.139829382305)*x_ref[1]
        ref[1,0,1,0]=(0.0957295570417)*x_ref[0]**o+(-0.105252674536)*x_ref[0]+(-0.569429953382)*x_ref[1]**o+(-0.801934666657)*x_ref[1]
        ref[1,0,1,1]=(-0.557307051058)*x_ref[0]**o+(0.218852576416)*x_ref[0]+(0.430167216464)*x_ref[1]**o+(-0.112285681761)*x_ref[1]
        ref[1,0,2,0]=(0.288472422477)*x_ref[0]**o+(0.294377994763)*x_ref[0]+(-0.809503961511)*x_ref[1]**o+(0.357020226991)*x_ref[1]
        ref[1,0,2,1]=(-0.148113477705)*x_ref[0]**o+(0.164468646294)*x_ref[0]+(0.7977142476)*x_ref[1]**o+(-0.407274933384)*x_ref[1]
        ref[1,1,0,0]=(-0.926374434822)*x_ref[0]**o+(-0.107713463193)*x_ref[0]+(-0.511635723869)*x_ref[1]**o+(0.259319717689)*x_ref[1]
        ref[1,1,0,1]=(-0.18636582243)*x_ref[0]**o+(0.698258973628)*x_ref[0]+(0.234379281397)*x_ref[1]**o+(0.596195875495)*x_ref[1]
        ref[1,1,1,0]=(-0.180395892006)*x_ref[0]**o+(0.631248500084)*x_ref[0]+(0.104751616377)*x_ref[1]**o+(0.204500206653)*x_ref[1]
        ref[1,1,1,1]=(-0.517119564238)*x_ref[0]**o+(-0.16634015634)*x_ref[0]+(0.556785586675)*x_ref[1]**o+(0.42659306392)*x_ref[1]
        ref[1,1,2,0]=(0.163807238958)*x_ref[0]**o+(-0.534619478446)*x_ref[0]+(-0.700353460753)*x_ref[1]**o+(0.121393131648)*x_ref[1]
        ref[1,1,2,1]=(0.964276278961)*x_ref[0]**o+(0.846030751304)*x_ref[0]+(-0.745301894392)*x_ref[1]**o+(0.33821485558)*x_ref[1]
        ref[1,2,0,0]=(0.333173101086)*x_ref[0]**o+(0.774848048743)*x_ref[0]+(-0.659091916357)*x_ref[1]**o+(-0.797247144497)*x_ref[1]
        ref[1,2,0,1]=(0.0581394534285)*x_ref[0]**o+(-0.524771663031)*x_ref[0]+(-0.519465753317)*x_ref[1]**o+(0.71748795226)*x_ref[1]
        ref[1,2,1,0]=(0.224935262833)*x_ref[0]**o+(-0.393399819306)*x_ref[0]+(0.199349075224)*x_ref[1]**o+(-0.808847009412)*x_ref[1]
        ref[1,2,1,1]=(0.0385750795521)*x_ref[0]**o+(-8.09949553151e-05)*x_ref[0]+(0.496860964506)*x_ref[1]**o+(0.0119507407829)*x_ref[1]
        ref[1,2,2,0]=(0.876341556167)*x_ref[0]**o+(0.18082766372)*x_ref[0]+(0.401349871827)*x_ref[1]**o+(-0.543439037303)*x_ref[1]
        ref[1,2,2,1]=(-0.894525004507)*x_ref[0]**o+(-0.516771143695)*x_ref[0]+(0.0410916490263)*x_ref[1]**o+(0.753841618439)*x_ref[1]
        ref[1,3,0,0]=(0.251979970924)*x_ref[0]**o+(0.790276550438)*x_ref[0]+(-0.155708793888)*x_ref[1]**o+(0.28971278336)*x_ref[1]
        ref[1,3,0,1]=(0.258336265436)*x_ref[0]**o+(0.696642875149)*x_ref[0]+(0.0561355915649)*x_ref[1]**o+(0.601839744417)*x_ref[1]
        ref[1,3,1,0]=(-0.978729102785)*x_ref[0]**o+(0.335280613262)*x_ref[0]+(0.89228798232)*x_ref[1]**o+(-0.693050679348)*x_ref[1]
        ref[1,3,1,1]=(-0.885817753431)*x_ref[0]**o+(-0.0641571535708)*x_ref[0]+(-0.649578576181)*x_ref[1]**o+(-0.95773137602)*x_ref[1]
        ref[1,3,2,0]=(-0.333891001461)*x_ref[0]**o+(-0.496064308564)*x_ref[0]+(0.438549603377)*x_ref[1]**o+(0.0772468142672)*x_ref[1]
        ref[1,3,2,1]=(-0.16171788255)*x_ref[0]**o+(-0.939293453222)*x_ref[0]+(-0.284522648546)*x_ref[1]**o+(0.744642328375)*x_ref[1]
        ref[1,4,0,0]=(0.806133262332)*x_ref[0]**o+(-0.0225320206432)*x_ref[0]+(0.125958642606)*x_ref[1]**o+(0.610630249467)*x_ref[1]
        ref[1,4,0,1]=(0.657971870946)*x_ref[0]**o+(0.698894701525)*x_ref[0]+(-0.595952612496)*x_ref[1]**o+(-0.38268841292)*x_ref[1]
        ref[1,4,1,0]=(-0.277274411331)*x_ref[0]**o+(-0.483183004514)*x_ref[0]+(0.77656657009)*x_ref[1]**o+(-0.959290466645)*x_ref[1]
        ref[1,4,1,1]=(-0.96384356305)*x_ref[0]**o+(-0.00383301781476)*x_ref[0]+(0.788010409608)*x_ref[1]**o+(0.135378744316)*x_ref[1]
        ref[1,4,2,0]=(0.986761958017)*x_ref[0]**o+(-0.570276378038)*x_ref[0]+(-0.515784028315)*x_ref[1]**o+(0.487125123439)*x_ref[1]
        ref[1,4,2,1]=(-0.954735054702)*x_ref[0]**o+(-0.0610323074217)*x_ref[0]+(-0.255284859281)*x_ref[1]**o+(0.277795725698)*x_ref[1]
        ref[2,0,0,0]=(-0.708301219324)*x_ref[0]**o+(-0.970166781553)*x_ref[0]+(-0.0836451425855)*x_ref[1]**o+(0.525854202995)*x_ref[1]
        ref[2,0,0,1]=(-0.0995393865183)*x_ref[0]**o+(-0.0156624820324)*x_ref[0]+(0.225529316197)*x_ref[1]**o+(0.142617834056)*x_ref[1]
        ref[2,0,1,0]=(-0.585329524103)*x_ref[0]**o+(0.590285924441)*x_ref[0]+(0.201483726322)*x_ref[1]**o+(-0.735262197689)*x_ref[1]
        ref[2,0,1,1]=(-0.682877182963)*x_ref[0]**o+(-0.162682262799)*x_ref[0]+(-0.890051459843)*x_ref[1]**o+(-0.0188015866765)*x_ref[1]
        ref[2,0,2,0]=(0.868935476294)*x_ref[0]**o+(-0.629891068854)*x_ref[0]+(0.335632865663)*x_ref[1]**o+(-0.955849055716)*x_ref[1]
        ref[2,0,2,1]=(-0.907933228238)*x_ref[0]**o+(-0.105861987974)*x_ref[0]+(-0.131372383042)*x_ref[1]**o+(-0.970114136854)*x_ref[1]
        ref[2,1,0,0]=(-0.275788501814)*x_ref[0]**o+(-0.992861867033)*x_ref[0]+(-0.80676324962)*x_ref[1]**o+(0.269244680224)*x_ref[1]
        ref[2,1,0,1]=(-0.704704138774)*x_ref[0]**o+(-0.371803193009)*x_ref[0]+(-0.0190586405922)*x_ref[1]**o+(-0.0828837624698)*x_ref[1]
        ref[2,1,1,0]=(-0.532645566013)*x_ref[0]**o+(0.571754809868)*x_ref[0]+(0.656737493773)*x_ref[1]**o+(-0.548229591842)*x_ref[1]
        ref[2,1,1,1]=(0.23836921087)*x_ref[0]**o+(0.707475737785)*x_ref[0]+(-0.59660155968)*x_ref[1]**o+(-0.450204951071)*x_ref[1]
        ref[2,1,2,0]=(-0.534027348014)*x_ref[0]**o+(0.332100193727)*x_ref[0]+(-0.864157138567)*x_ref[1]**o+(-0.41833267066)*x_ref[1]
        ref[2,1,2,1]=(0.459878593269)*x_ref[0]**o+(0.446259140228)*x_ref[0]+(0.571242887502)*x_ref[1]**o+(-0.0390525669486)*x_ref[1]
        ref[2,2,0,0]=(0.322052799527)*x_ref[0]**o+(-0.915265209304)*x_ref[0]+(-0.99942007618)*x_ref[1]**o+(0.482294513743)*x_ref[1]
        ref[2,2,0,1]=(0.558489194456)*x_ref[0]**o+(0.00400576554086)*x_ref[0]+(-0.87484330165)*x_ref[1]**o+(-0.664504218977)*x_ref[1]
        ref[2,2,1,0]=(0.669621607461)*x_ref[0]**o+(-0.0450960554319)*x_ref[0]+(-0.221151480246)*x_ref[1]**o+(0.0123621737211)*x_ref[1]
        ref[2,2,1,1]=(0.476743013029)*x_ref[0]**o+(0.11118371822)*x_ref[0]+(-0.202074315677)*x_ref[1]**o+(0.377649433458)*x_ref[1]
        ref[2,2,2,0]=(0.667910543701)*x_ref[0]**o+(0.035703922547)*x_ref[0]+(-0.274733967095)*x_ref[1]**o+(-0.998934720182)*x_ref[1]
        ref[2,2,2,1]=(0.578909660087)*x_ref[0]**o+(0.149900086514)*x_ref[0]+(-0.226597651858)*x_ref[1]**o+(0.169175812287)*x_ref[1]
        ref[2,3,0,0]=(0.158370246586)*x_ref[0]**o+(0.595643112285)*x_ref[0]+(-0.476875564052)*x_ref[1]**o+(0.869752657517)*x_ref[1]
        ref[2,3,0,1]=(0.99718847714)*x_ref[0]**o+(0.581531822297)*x_ref[0]+(-0.607102775564)*x_ref[1]**o+(-0.8317970284)*x_ref[1]
        ref[2,3,1,0]=(-0.27390195745)*x_ref[0]**o+(0.0512853330414)*x_ref[0]+(0.465190666302)*x_ref[1]**o+(0.042533388788)*x_ref[1]
        ref[2,3,1,1]=(-0.183679504962)*x_ref[0]**o+(-0.0522782076121)*x_ref[0]+(0.921423118529)*x_ref[1]**o+(-0.0315080275225)*x_ref[1]
        ref[2,3,2,0]=(0.0421842348778)*x_ref[0]**o+(0.462023049702)*x_ref[0]+(-0.888342606803)*x_ref[1]**o+(-0.956503447884)*x_ref[1]
        ref[2,3,2,1]=(-0.950820288444)*x_ref[0]**o+(0.104271990168)*x_ref[0]+(-0.607212909582)*x_ref[1]**o+(-0.16283931259)*x_ref[1]
        ref[2,4,0,0]=(0.113505933533)*x_ref[0]**o+(0.27873514806)*x_ref[0]+(0.736699443523)*x_ref[1]**o+(-0.951737555693)*x_ref[1]
        ref[2,4,0,1]=(0.264170584849)*x_ref[0]**o+(-0.226649236617)*x_ref[0]+(0.489902180874)*x_ref[1]**o+(-0.191536591762)*x_ref[1]
        ref[2,4,1,0]=(0.712322285985)*x_ref[0]**o+(-0.323395455815)*x_ref[0]+(-0.298356681586)*x_ref[1]**o+(0.699891974541)*x_ref[1]
        ref[2,4,1,1]=(-0.631057763255)*x_ref[0]**o+(-0.489374933445)*x_ref[0]+(-0.021591806874)*x_ref[1]**o+(-0.692800124585)*x_ref[1]
        ref[2,4,2,0]=(0.916068935184)*x_ref[0]**o+(0.33779688858)*x_ref[0]+(0.06300443559)*x_ref[1]**o+(-0.231385203399)*x_ref[1]
        ref[2,4,2,1]=(-0.283485044625)*x_ref[0]**o+(0.548366574442)*x_ref[0]+(-0.543953132872)*x_ref[1]**o+(-0.956662823037)*x_ref[1]
        ref[3,0,0,0]=(0.709349274221)*x_ref[0]**o+(-0.342981009263)*x_ref[0]+(-0.682331761929)*x_ref[1]**o+(0.673169884813)*x_ref[1]
        ref[3,0,0,1]=(0.236249836593)*x_ref[0]**o+(-0.682845054434)*x_ref[0]+(-0.349169988218)*x_ref[1]**o+(0.502684425824)*x_ref[1]
        ref[3,0,1,0]=(-0.561655748793)*x_ref[0]**o+(0.120183772326)*x_ref[0]+(-0.966535325583)*x_ref[1]**o+(-0.0585117264949)*x_ref[1]
        ref[3,0,1,1]=(-0.661030287189)*x_ref[0]**o+(-0.322924086505)*x_ref[0]+(-0.986892059552)*x_ref[1]**o+(-0.778969361365)*x_ref[1]
        ref[3,0,2,0]=(-0.166307107307)*x_ref[0]**o+(0.612931303617)*x_ref[0]+(-0.283925165456)*x_ref[1]**o+(0.5998038814)*x_ref[1]
        ref[3,0,2,1]=(0.205099355916)*x_ref[0]**o+(-0.520970536278)*x_ref[0]+(0.900416833183)*x_ref[1]**o+(0.981826306069)*x_ref[1]
        ref[3,1,0,0]=(-0.875306666443)*x_ref[0]**o+(0.0189358479861)*x_ref[0]+(0.101187857445)*x_ref[1]**o+(-0.873584307534)*x_ref[1]
        ref[3,1,0,1]=(-0.0203656235757)*x_ref[0]**o+(-0.579229777042)*x_ref[0]+(-0.0799345631826)*x_ref[1]**o+(0.896315444069)*x_ref[1]
        ref[3,1,1,0]=(0.576791459982)*x_ref[0]**o+(-0.86541815506)*x_ref[0]+(-0.582764176207)*x_ref[1]**o+(-0.249437395011)*x_ref[1]
        ref[3,1,1,1]=(0.525953680497)*x_ref[0]**o+(-0.222445717402)*x_ref[0]+(-0.87710202063)*x_ref[1]**o+(-0.762921821973)*x_ref[1]
        ref[3,1,2,0]=(-0.883737185908)*x_ref[0]**o+(0.254053241277)*x_ref[0]+(0.559803298059)*x_ref[1]**o+(-0.577376956856)*x_ref[1]
        ref[3,1,2,1]=(-0.125792083387)*x_ref[0]**o+(-0.771549307509)*x_ref[0]+(0.142493480552)*x_ref[1]**o+(0.459996962548)*x_ref[1]
        ref[3,2,0,0]=(0.0531598927049)*x_ref[0]**o+(-0.950343229037)*x_ref[0]+(-0.748240453154)*x_ref[1]**o+(0.850555892342)*x_ref[1]
        ref[3,2,0,1]=(-0.660852152041)*x_ref[0]**o+(-0.763119278604)*x_ref[0]+(-0.811925409663)*x_ref[1]**o+(-0.178765071495)*x_ref[1]
        ref[3,2,1,0]=(-0.189124223305)*x_ref[0]**o+(-0.410392909042)*x_ref[0]+(0.993247167522)*x_ref[1]**o+(0.291451796298)*x_ref[1]
        ref[3,2,1,1]=(0.83496133421)*x_ref[0]**o+(0.695147480265)*x_ref[0]+(0.444420524946)*x_ref[1]**o+(0.107454620114)*x_ref[1]
        ref[3,2,2,0]=(0.94687479313)*x_ref[0]**o+(-0.140702324395)*x_ref[0]+(0.149203502428)*x_ref[1]**o+(-0.338246126345)*x_ref[1]
        ref[3,2,2,1]=(0.147609394709)*x_ref[0]**o+(-0.549219765239)*x_ref[0]+(0.397308993347)*x_ref[1]**o+(0.0402189312685)*x_ref[1]
        ref[3,3,0,0]=(0.27400981409)*x_ref[0]**o+(-0.516367266257)*x_ref[0]+(-0.784095952814)*x_ref[1]**o+(0.32858899009)*x_ref[1]
        ref[3,3,0,1]=(-0.393457382163)*x_ref[0]**o+(0.590406794051)*x_ref[0]+(0.328523918357)*x_ref[1]**o+(0.899537513355)*x_ref[1]
        ref[3,3,1,0]=(-0.851038623366)*x_ref[0]**o+(0.596974864726)*x_ref[0]+(-0.338630687551)*x_ref[1]**o+(0.634961435042)*x_ref[1]
        ref[3,3,1,1]=(0.569166143918)*x_ref[0]**o+(-0.388162282491)*x_ref[0]+(0.764830536733)*x_ref[1]**o+(-0.426737342542)*x_ref[1]
        ref[3,3,2,0]=(0.502221903448)*x_ref[0]**o+(0.940382225897)*x_ref[0]+(0.563444730579)*x_ref[1]**o+(0.023210869151)*x_ref[1]
        ref[3,3,2,1]=(0.061695403852)*x_ref[0]**o+(0.868302188758)*x_ref[0]+(0.0855803325758)*x_ref[1]**o+(-0.24366173638)*x_ref[1]
        ref[3,4,0,0]=(-0.533038182387)*x_ref[0]**o+(0.0222388915737)*x_ref[0]+(0.529037752769)*x_ref[1]**o+(-0.0665320895494)*x_ref[1]
        ref[3,4,0,1]=(0.714374055253)*x_ref[0]**o+(-0.623670743333)*x_ref[0]+(-0.866024792929)*x_ref[1]**o+(-0.0673124896111)*x_ref[1]
        ref[3,4,1,0]=(-0.0523466615045)*x_ref[0]**o+(-0.472563426622)*x_ref[0]+(0.139892673626)*x_ref[1]**o+(0.539655417178)*x_ref[1]
        ref[3,4,1,1]=(-0.473032580514)*x_ref[0]**o+(0.692812093618)*x_ref[0]+(0.664400517742)*x_ref[1]**o+(0.184125423364)*x_ref[1]
        ref[3,4,2,0]=(-0.0340798058725)*x_ref[0]**o+(-0.58900562863)*x_ref[0]+(0.396856384717)*x_ref[1]**o+(-0.559175428019)*x_ref[1]
        ref[3,4,2,1]=(-0.217679376634)*x_ref[0]**o+(-0.90389036667)*x_ref[0]+(0.967372942873)*x_ref[1]**o+(-0.482913619277)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.93229223906)*x[0]**o+(0.176839936782)*x[0]+(0.986271000359)*x[1]**o+(-0.449322385723)*x[1]+(-0.856450437823)*x[2]**o+(0.31473579221)*x[2]
        arg[0,0,0,1]=(0.714047232765)*x[0]**o+(-0.323471323003)*x[0]+(0.628080480343)*x[1]**o+(-0.992007912045)*x[1]+(-0.11019723051)*x[2]**o+(-0.104735408233)*x[2]
        arg[0,0,1,0]=(-0.437250443294)*x[0]**o+(-0.741469652781)*x[0]+(0.650921038465)*x[1]**o+(0.679622741243)*x[1]+(0.604005894895)*x[2]**o+(0.598155624876)*x[2]
        arg[0,0,1,1]=(-0.574872168761)*x[0]**o+(0.796207672391)*x[0]+(-0.478046646159)*x[1]**o+(0.604661326741)*x[1]+(-0.898457020186)*x[2]**o+(0.43234750827)*x[2]
        arg[0,0,2,0]=(-0.674008357742)*x[0]**o+(0.699614371016)*x[0]+(0.441321981559)*x[1]**o+(-0.619688689942)*x[1]+(-0.0874579208832)*x[2]**o+(-0.925753840555)*x[2]
        arg[0,0,2,1]=(-0.114329474892)*x[0]**o+(0.699501071503)*x[0]+(-0.618807420683)*x[1]**o+(0.411584695002)*x[1]+(-0.234620261556)*x[2]**o+(0.585720100967)*x[2]
        arg[0,1,0,0]=(0.617432097253)*x[0]**o+(-0.00304868863676)*x[0]+(0.571021785645)*x[1]**o+(0.457197205728)*x[1]+(0.462895867813)*x[2]**o+(0.278022956316)*x[2]
        arg[0,1,0,1]=(-0.659136231188)*x[0]**o+(-0.388212920906)*x[0]+(-0.982152608232)*x[1]**o+(-0.665797156013)*x[1]+(-0.76141890649)*x[2]**o+(-0.591458323836)*x[2]
        arg[0,1,1,0]=(-0.488016386893)*x[0]**o+(0.965660401944)*x[0]+(0.178400189819)*x[1]**o+(0.578417314499)*x[1]+(-0.398083955217)*x[2]**o+(0.460338187535)*x[2]
        arg[0,1,1,1]=(-0.182086458235)*x[0]**o+(0.0842368498629)*x[0]+(-0.642008026184)*x[1]**o+(-0.0449408131957)*x[1]+(-0.927397573018)*x[2]**o+(0.854313729816)*x[2]
        arg[0,1,2,0]=(-0.201475872769)*x[0]**o+(0.098190765292)*x[0]+(0.808624244922)*x[1]**o+(0.0726268899898)*x[1]+(-0.605334327027)*x[2]**o+(-0.999151255934)*x[2]
        arg[0,1,2,1]=(0.672344024218)*x[0]**o+(-0.638554833481)*x[0]+(-0.212231801634)*x[1]**o+(-0.492765366968)*x[1]+(0.311273771921)*x[2]**o+(-0.504370655006)*x[2]
        arg[0,2,0,0]=(0.9944935907)*x[0]**o+(-0.365261996611)*x[0]+(-0.105667939789)*x[1]**o+(0.886935751977)*x[1]+(0.263336157757)*x[2]**o+(-0.880316523113)*x[2]
        arg[0,2,0,1]=(-0.659779300787)*x[0]**o+(-0.728590180152)*x[0]+(0.732078508037)*x[1]**o+(0.684963412098)*x[1]+(0.822970700012)*x[2]**o+(-0.680001879674)*x[2]
        arg[0,2,1,0]=(-0.207855412893)*x[0]**o+(0.438275426529)*x[0]+(0.87046682066)*x[1]**o+(-0.960322013848)*x[1]+(-0.273728198625)*x[2]**o+(-0.960137228004)*x[2]
        arg[0,2,1,1]=(-0.189232680002)*x[0]**o+(0.866417302002)*x[0]+(-0.93403311248)*x[1]**o+(-0.55773104878)*x[1]+(0.94062895773)*x[2]**o+(-0.758052194805)*x[2]
        arg[0,2,2,0]=(0.100859944394)*x[0]**o+(-0.20292135909)*x[0]+(-0.713339941261)*x[1]**o+(-0.615106454411)*x[1]+(0.791245995152)*x[2]**o+(-0.81051791594)*x[2]
        arg[0,2,2,1]=(0.595743424517)*x[0]**o+(-0.179075876404)*x[0]+(-0.799005601715)*x[1]**o+(-0.0554635073569)*x[1]+(-0.850510701985)*x[2]**o+(0.234516538169)*x[2]
        arg[0,3,0,0]=(-0.361667738861)*x[0]**o+(-0.551837498696)*x[0]+(-0.834377202094)*x[1]**o+(0.814168224265)*x[1]+(-0.00479533268377)*x[2]**o+(0.843007652951)*x[2]
        arg[0,3,0,1]=(-0.448417439465)*x[0]**o+(0.431496158388)*x[0]+(-0.547416443433)*x[1]**o+(-0.0746785704695)*x[1]+(-0.646238182944)*x[2]**o+(-0.318608714595)*x[2]
        arg[0,3,1,0]=(-0.275647366007)*x[0]**o+(-0.410594785047)*x[0]+(-0.858282378196)*x[1]**o+(-0.755586690837)*x[1]+(-0.420481091654)*x[2]**o+(-0.0996487371831)*x[2]
        arg[0,3,1,1]=(-0.395961249009)*x[0]**o+(0.0860896467863)*x[0]+(0.482719525748)*x[1]**o+(0.319368178133)*x[1]+(0.516449153414)*x[2]**o+(-0.246749862576)*x[2]
        arg[0,3,2,0]=(-0.367322027122)*x[0]**o+(0.687206412524)*x[0]+(0.616265233972)*x[1]**o+(-0.32660589112)*x[1]+(-0.944607343096)*x[2]**o+(0.123982537951)*x[2]
        arg[0,3,2,1]=(0.700497166804)*x[0]**o+(0.553961329865)*x[0]+(-0.34020003754)*x[1]**o+(0.295870992783)*x[1]+(-0.220275880046)*x[2]**o+(0.766121035825)*x[2]
        arg[0,4,0,0]=(0.405469789242)*x[0]**o+(-0.338347662873)*x[0]+(0.0761796434233)*x[1]**o+(-0.120599175033)*x[1]+(0.107865585379)*x[2]**o+(0.0711941007838)*x[2]
        arg[0,4,0,1]=(0.614166154899)*x[0]**o+(0.195780702963)*x[0]+(0.718935985215)*x[1]**o+(0.214577296792)*x[1]+(-0.366985242107)*x[2]**o+(-0.509584787633)*x[2]
        arg[0,4,1,0]=(-0.638646370981)*x[0]**o+(-0.769955899284)*x[0]+(-0.887407540503)*x[1]**o+(0.0901245429998)*x[1]+(0.786085641321)*x[2]**o+(0.6248068552)*x[2]
        arg[0,4,1,1]=(-0.145426222043)*x[0]**o+(-0.494114545016)*x[0]+(0.422153721335)*x[1]**o+(0.617022143901)*x[1]+(-0.665972324723)*x[2]**o+(-0.476724432799)*x[2]
        arg[0,4,2,0]=(0.556105252092)*x[0]**o+(-0.893268696809)*x[0]+(0.146441889489)*x[1]**o+(-0.170512113283)*x[1]+(-0.869313262875)*x[2]**o+(-0.976874109392)*x[2]
        arg[0,4,2,1]=(0.858927206499)*x[0]**o+(0.198998355146)*x[0]+(0.278125296049)*x[1]**o+(-0.0442518828093)*x[1]+(0.274852724203)*x[2]**o+(-0.167649193314)*x[2]
        arg[1,0,0,0]=(-0.0429420289353)*x[0]**o+(-0.15764693929)*x[0]+(-0.859599917512)*x[1]**o+(-0.904178643575)*x[1]+(0.519560711493)*x[2]**o+(0.258644717907)*x[2]
        arg[1,0,0,1]=(-0.185321114264)*x[0]**o+(-0.794493830004)*x[0]+(0.136084421162)*x[1]**o+(0.926912021419)*x[1]+(0.346392120388)*x[2]**o+(-0.183908371454)*x[2]
        arg[1,0,1,0]=(-0.730675509877)*x[0]**o+(0.865034802778)*x[0]+(-0.506315431878)*x[1]**o+(-0.94607761051)*x[1]+(0.195194792801)*x[2]**o+(-0.865129258879)*x[2]
        arg[1,0,1,1]=(-0.0597026648107)*x[0]**o+(-0.884790497)*x[0]+(-0.0631278742098)*x[1]**o+(-0.808364940737)*x[1]+(-0.968952604837)*x[2]**o+(0.142246702632)*x[2]
        arg[1,0,2,0]=(0.648472894563)*x[0]**o+(0.145861644073)*x[0]+(0.501227520574)*x[1]**o+(0.265886482306)*x[1]+(-0.0472772173573)*x[2]**o+(-0.364169655969)*x[2]
        arg[1,0,2,1]=(0.847863600695)*x[0]**o+(-0.288841685199)*x[0]+(-0.500273502795)*x[1]**o+(0.627379277592)*x[1]+(0.326814024625)*x[2]**o+(-0.514712825895)*x[2]
        arg[1,1,0,0]=(-0.947845713712)*x[0]**o+(-0.372115616305)*x[0]+(-0.572858279998)*x[1]**o+(0.226237627692)*x[1]+(0.210201947182)*x[2]**o+(0.0147616538594)*x[2]
        arg[1,1,0,1]=(-0.39813630466)*x[0]**o+(-0.658496395731)*x[0]+(-0.363906465005)*x[1]**o+(-0.816719661238)*x[1]+(-0.478106835032)*x[2]**o+(0.0819042309589)*x[2]
        arg[1,1,1,0]=(-0.287160278593)*x[0]**o+(-0.0281402640978)*x[0]+(0.41559795378)*x[1]**o+(0.158274458097)*x[1]+(-0.798429481502)*x[2]**o+(-0.278585334624)*x[2]
        arg[1,1,1,1]=(0.809550480973)*x[0]**o+(0.192026915854)*x[0]+(-0.891168838599)*x[1]**o+(0.0482621192882)*x[1]+(-0.35339708159)*x[2]**o+(-0.133161551553)*x[2]
        arg[1,1,2,0]=(0.939582328872)*x[0]**o+(0.561963790516)*x[0]+(0.885264722949)*x[1]**o+(-0.468042305266)*x[1]+(0.232880002429)*x[2]**o+(-0.772896005701)*x[2]
        arg[1,1,2,1]=(-0.297333303209)*x[0]**o+(-0.414050924207)*x[0]+(-0.278903698873)*x[1]**o+(0.345370620869)*x[1]+(-0.0570151259027)*x[2]**o+(0.786449641378)*x[2]
        arg[1,2,0,0]=(0.644047138589)*x[0]**o+(-0.480660438838)*x[0]+(0.32526499462)*x[1]**o+(-0.775988282127)*x[1]+(0.108374143932)*x[2]**o+(0.955463667181)*x[2]
        arg[1,2,0,1]=(-0.468236790339)*x[0]**o+(0.844197711724)*x[0]+(0.808973429831)*x[1]**o+(-0.647699550005)*x[1]+(0.552555959813)*x[2]**o+(-0.3403079515)*x[2]
        arg[1,2,1,0]=(0.864686362308)*x[0]**o+(0.248923646116)*x[0]+(0.0287042863842)*x[1]**o+(0.404974282153)*x[1]+(0.113501438234)*x[2]**o+(-0.619364894397)*x[2]
        arg[1,2,1,1]=(-0.625146992499)*x[0]**o+(-0.0480602452983)*x[0]+(0.168910550914)*x[1]**o+(-0.891584218123)*x[1]+(0.0609357618263)*x[2]**o+(0.0451921002161)*x[2]
        arg[1,2,2,0]=(-0.478728628752)*x[0]**o+(0.428376547798)*x[0]+(0.817203376818)*x[1]**o+(-0.856056266953)*x[1]+(0.354960313382)*x[2]**o+(-0.611468960461)*x[2]
        arg[1,2,2,1]=(0.49500742113)*x[0]**o+(0.902948130263)*x[0]+(-0.810748543744)*x[1]**o+(-0.696150924858)*x[1]+(-0.748278516161)*x[2]**o+(0.694815527058)*x[2]
        arg[1,3,0,0]=(-0.361960150296)*x[0]**o+(-0.845139780461)*x[0]+(-0.0624499309665)*x[1]**o+(-0.0515558058489)*x[1]+(0.400690983017)*x[2]**o+(-0.744608407974)*x[2]
        arg[1,3,0,1]=(-0.452453617467)*x[0]**o+(0.845396139593)*x[0]+(0.292319469936)*x[1]**o+(-0.833547518634)*x[1]+(-0.511797700353)*x[2]**o+(0.857891454773)*x[2]
        arg[1,3,1,0]=(0.779504092912)*x[0]**o+(0.500524773615)*x[0]+(0.590542659382)*x[1]**o+(0.479310630519)*x[1]+(0.509960641956)*x[2]**o+(-0.564980432959)*x[2]
        arg[1,3,1,1]=(-0.145745189298)*x[0]**o+(-0.256064576359)*x[0]+(0.640890837126)*x[1]**o+(-0.718802035152)*x[1]+(-0.322216873583)*x[2]**o+(-0.0246526526334)*x[2]
        arg[1,3,2,0]=(-0.463434285098)*x[0]**o+(0.116722448643)*x[0]+(-0.690378381342)*x[1]**o+(0.429197465885)*x[1]+(-0.353019731384)*x[2]**o+(0.57397712918)*x[2]
        arg[1,3,2,1]=(0.169676853383)*x[0]**o+(0.815225472623)*x[0]+(-0.11125527552)*x[1]**o+(0.391513831525)*x[1]+(-0.587041361585)*x[2]**o+(0.255194151616)*x[2]
        arg[1,4,0,0]=(0.325351608628)*x[0]**o+(-0.57751715453)*x[0]+(0.145167511303)*x[1]**o+(-0.969286030742)*x[1]+(0.917370398478)*x[2]**o+(0.691483691598)*x[2]
        arg[1,4,0,1]=(0.0997998298271)*x[0]**o+(-0.101248202247)*x[0]+(-0.650649653848)*x[1]**o+(-0.157887947288)*x[1]+(0.517501747765)*x[2]**o+(-0.0852921926298)*x[2]
        arg[1,4,1,0]=(0.984595420407)*x[0]**o+(0.499293499076)*x[0]+(0.352193115874)*x[1]**o+(-0.684522909647)*x[1]+(-0.391538373605)*x[2]**o+(0.404635595162)*x[2]
        arg[1,4,1,1]=(-0.231310482173)*x[0]**o+(-0.0385089836559)*x[0]+(0.247671691888)*x[1]**o+(0.883971955027)*x[1]+(-0.0856624709734)*x[2]**o+(0.942940875904)*x[2]
        arg[1,4,2,0]=(-0.471467257434)*x[0]**o+(0.835634240897)*x[0]+(-0.594421294562)*x[1]**o+(-0.187584387111)*x[1]+(0.859867472765)*x[2]**o+(0.930951154576)*x[2]
        arg[1,4,2,1]=(0.504416975601)*x[0]**o+(0.435692923635)*x[0]+(-0.296945944313)*x[1]**o+(0.815156218765)*x[1]+(0.464464247442)*x[2]**o+(-0.378809453041)*x[2]
        arg[2,0,0,0]=(0.213087423107)*x[0]**o+(-0.504681665801)*x[0]+(0.796352190405)*x[1]**o+(0.546752751136)*x[1]+(0.57347644664)*x[2]**o+(0.715856963195)*x[2]
        arg[2,0,0,1]=(-0.28455266769)*x[0]**o+(0.853162220766)*x[0]+(0.164930982863)*x[1]**o+(-0.0136996498765)*x[1]+(0.293717915079)*x[2]**o+(-0.216124158473)*x[2]
        arg[2,0,1,0]=(-0.328115989441)*x[0]**o+(-0.529441496488)*x[0]+(-0.43579479677)*x[1]**o+(-0.851333801194)*x[1]+(-0.278406102972)*x[2]**o+(0.947777471824)*x[2]
        arg[2,0,1,1]=(-0.756438938789)*x[0]**o+(-0.624332272735)*x[0]+(-0.591638429565)*x[1]**o+(-0.0608281771566)*x[1]+(-0.852078784085)*x[2]**o+(-0.526558964866)*x[2]
        arg[2,0,2,0]=(-0.923892426969)*x[0]**o+(0.788121899878)*x[0]+(-0.446006331202)*x[1]**o+(0.997210871574)*x[1]+(0.356998606103)*x[2]**o+(0.972643822823)*x[2]
        arg[2,0,2,1]=(0.225183308571)*x[0]**o+(-0.757505414777)*x[0]+(-0.59573192074)*x[1]**o+(0.57166154987)*x[1]+(0.622343198809)*x[2]**o+(0.60343970503)*x[2]
        arg[2,1,0,0]=(0.51027702934)*x[0]**o+(-0.711737679522)*x[0]+(-0.211712579627)*x[1]**o+(-0.495817605314)*x[1]+(0.924425804272)*x[2]**o+(0.753424835275)*x[2]
        arg[2,1,0,1]=(0.476218252577)*x[0]**o+(0.187579891557)*x[0]+(0.727359583623)*x[1]**o+(-0.389875559955)*x[1]+(-0.352889326477)*x[2]**o+(0.365143943637)*x[2]
        arg[2,1,1,0]=(-0.815704312943)*x[0]**o+(0.315217470245)*x[0]+(-0.535234555065)*x[1]**o+(-0.433408988501)*x[1]+(-0.676694087389)*x[2]**o+(0.384097177718)*x[2]
        arg[2,1,1,1]=(-0.0368452411345)*x[0]**o+(0.514540686444)*x[0]+(-0.527100066965)*x[1]**o+(0.0291015562912)*x[1]+(-0.605646905004)*x[2]**o+(0.931528599981)*x[2]
        arg[2,1,2,0]=(-0.0994015945266)*x[0]**o+(-0.762310347434)*x[0]+(-0.845934950466)*x[1]**o+(-0.235758967923)*x[1]+(0.755531422698)*x[2]**o+(0.108353867777)*x[2]
        arg[2,1,2,1]=(-0.868779052889)*x[0]**o+(0.783878941075)*x[0]+(0.900089560473)*x[1]**o+(-0.836779405877)*x[1]+(-0.799571882449)*x[2]**o+(0.999710029157)*x[2]
        arg[2,2,0,0]=(0.541465277143)*x[0]**o+(-0.840477797945)*x[0]+(-0.362577721098)*x[1]**o+(0.00680131806718)*x[1]+(-0.338485239867)*x[2]**o+(0.828213756751)*x[2]
        arg[2,2,0,1]=(-0.927347423175)*x[0]**o+(-0.936191125785)*x[0]+(-0.411819804833)*x[1]**o+(-0.0222067637269)*x[1]+(-0.104411832586)*x[2]**o+(-0.79937203081)*x[2]
        arg[2,2,1,0]=(0.273051712848)*x[0]**o+(-0.802468557957)*x[0]+(-0.019392898867)*x[1]**o+(-0.612803074784)*x[1]+(-0.269705464223)*x[2]**o+(-0.244405147706)*x[2]
        arg[2,2,1,1]=(-0.08869649035)*x[0]**o+(0.809761782447)*x[0]+(0.761420080249)*x[1]**o+(0.16590038557)*x[1]+(-0.405657593445)*x[2]**o+(0.123593603037)*x[2]
        arg[2,2,2,0]=(-0.799691160987)*x[0]**o+(0.834835690475)*x[0]+(-0.0863569221737)*x[1]**o+(0.988343215152)*x[1]+(0.294423283474)*x[2]**o+(0.325725138203)*x[2]
        arg[2,2,2,1]=(0.648897496592)*x[0]**o+(-0.474329387314)*x[0]+(-0.904613697104)*x[1]**o+(0.382201388286)*x[1]+(-0.596889148339)*x[2]**o+(0.705132157684)*x[2]
        arg[2,3,0,0]=(0.906738144045)*x[0]**o+(-0.365392026426)*x[0]+(0.730557428733)*x[1]**o+(-0.857864471189)*x[1]+(-0.160930657186)*x[2]**o+(0.885067383334)*x[2]
        arg[2,3,0,1]=(0.992764702472)*x[0]**o+(0.720688532948)*x[0]+(0.318869772818)*x[1]**o+(0.995508347248)*x[1]+(0.108569456996)*x[2]**o+(-0.990939292995)*x[2]
        arg[2,3,1,0]=(0.690584992096)*x[0]**o+(0.605885242246)*x[0]+(0.72371794265)*x[1]**o+(-0.632609325879)*x[1]+(0.520273588734)*x[2]**o+(0.885804143155)*x[2]
        arg[2,3,1,1]=(0.485231542482)*x[0]**o+(0.866920225418)*x[0]+(-0.814964945972)*x[1]**o+(0.544132513209)*x[1]+(-0.155071376494)*x[2]**o+(-0.941300746237)*x[2]
        arg[2,3,2,0]=(0.748232238388)*x[0]**o+(0.179397724419)*x[0]+(-0.43881886248)*x[1]**o+(-0.124941799624)*x[1]+(0.14679335819)*x[2]**o+(-0.758738846688)*x[2]
        arg[2,3,2,1]=(-0.0951333655337)*x[0]**o+(0.37270972476)*x[0]+(-0.651977554115)*x[1]**o+(-0.88918597476)*x[1]+(-0.730858685576)*x[2]**o+(0.685881924459)*x[2]
        arg[2,4,0,0]=(-0.823487466802)*x[0]**o+(-0.583630855643)*x[0]+(-0.664814562747)*x[1]**o+(0.860952605956)*x[1]+(-0.249114322668)*x[2]**o+(-0.968356584237)*x[2]
        arg[2,4,0,1]=(0.44040146252)*x[0]**o+(-0.246827565902)*x[0]+(0.596300980733)*x[1]**o+(-0.825818200414)*x[1]+(0.0715512875967)*x[2]**o+(0.996064679803)*x[2]
        arg[2,4,1,0]=(-0.287806453021)*x[0]**o+(0.704331542927)*x[0]+(-0.346139754521)*x[1]**o+(-0.766667037125)*x[1]+(0.283874132105)*x[2]**o+(-0.613145979204)*x[2]
        arg[2,4,1,1]=(-0.116649875156)*x[0]**o+(0.291380048239)*x[0]+(0.689809221905)*x[1]**o+(0.758935910592)*x[1]+(0.505142787603)*x[2]**o+(-0.113611480503)*x[2]
        arg[2,4,2,0]=(-0.836908572692)*x[0]**o+(-0.21892828725)*x[0]+(-0.935780103868)*x[1]**o+(0.106204643268)*x[1]+(-0.894474423807)*x[2]**o+(0.296554773217)*x[2]
        arg[2,4,2,1]=(-0.89413495023)*x[0]**o+(0.6080813058)*x[0]+(0.087863889878)*x[1]**o+(0.213373665612)*x[1]+(0.17085345025)*x[2]**o+(-0.967625138712)*x[2]
        arg[3,0,0,0]=(0.327537677409)*x[0]**o+(-0.299373059373)*x[0]+(-0.43223648542)*x[1]**o+(-0.499115781051)*x[1]+(0.145464567449)*x[2]**o+(0.580930584563)*x[2]
        arg[3,0,0,1]=(-0.122622543351)*x[0]**o+(-0.220547821081)*x[0]+(0.892679827803)*x[1]**o+(0.436584604381)*x[1]+(-0.156093580288)*x[2]**o+(0.464069069075)*x[2]
        arg[3,0,1,0]=(0.590288663203)*x[0]**o+(-0.212929445251)*x[0]+(0.0899671639061)*x[1]**o+(-0.87729854872)*x[1]+(0.789506558716)*x[2]**o+(-0.68867690907)*x[2]
        arg[3,0,1,1]=(-0.913651849012)*x[0]**o+(-0.430938918892)*x[0]+(-0.351186012371)*x[1]**o+(0.500306994748)*x[1]+(0.145381845281)*x[2]**o+(0.624993552333)*x[2]
        arg[3,0,2,0]=(0.281393881265)*x[0]**o+(-0.704727129826)*x[0]+(-0.0102091836449)*x[1]**o+(0.311857501353)*x[1]+(0.300732947904)*x[2]**o+(0.923973499548)*x[2]
        arg[3,0,2,1]=(-0.584907526993)*x[0]**o+(0.625732316822)*x[0]+(-0.609691930408)*x[1]**o+(0.056160226956)*x[1]+(-0.766583561947)*x[2]**o+(0.307692190194)*x[2]
        arg[3,1,0,0]=(0.731231379445)*x[0]**o+(-0.0485980527115)*x[0]+(0.908932758852)*x[1]**o+(0.188918035886)*x[1]+(0.685837743688)*x[2]**o+(-0.299987553989)*x[2]
        arg[3,1,0,1]=(-0.784006561231)*x[0]**o+(-0.809789430022)*x[0]+(-0.846391792735)*x[1]**o+(-0.211583304405)*x[1]+(-0.00198590063687)*x[2]**o+(-0.853253255224)*x[2]
        arg[3,1,1,0]=(0.595567299132)*x[0]**o+(-0.581734416878)*x[0]+(0.88144481486)*x[1]**o+(0.776876564509)*x[1]+(-0.367477971386)*x[2]**o+(0.0806320164453)*x[2]
        arg[3,1,1,1]=(0.440836187611)*x[0]**o+(-0.578925091757)*x[0]+(0.923367189832)*x[1]**o+(-0.993888173929)*x[1]+(-0.645514380411)*x[2]**o+(0.0267439044199)*x[2]
        arg[3,1,2,0]=(-0.305182288788)*x[0]**o+(-0.503372502553)*x[0]+(0.740948657389)*x[1]**o+(0.285661052275)*x[1]+(-0.634510984218)*x[2]**o+(-0.123979746456)*x[2]
        arg[3,1,2,1]=(-0.897274928495)*x[0]**o+(0.971356796164)*x[0]+(0.940768290313)*x[1]**o+(-0.552941985667)*x[1]+(0.172857451423)*x[2]**o+(0.844758237523)*x[2]
        arg[3,2,0,0]=(-0.903542725333)*x[0]**o+(0.936704774551)*x[0]+(-0.398085875228)*x[1]**o+(-0.304061573512)*x[1]+(-0.13362037229)*x[2]**o+(-0.405043350728)*x[2]
        arg[3,2,0,1]=(-0.0208590751801)*x[0]**o+(-0.354086381885)*x[0]+(0.510884698744)*x[1]**o+(-0.275504880518)*x[1]+(0.459476496356)*x[2]**o+(0.25358097976)*x[2]
        arg[3,2,1,0]=(-0.0171081877107)*x[0]**o+(-0.856081626112)*x[0]+(-0.0698686571004)*x[1]**o+(0.0954918361122)*x[1]+(0.938494954069)*x[2]**o+(-0.824928688248)*x[2]
        arg[3,2,1,1]=(-0.277307915098)*x[0]**o+(0.737855779658)*x[0]+(-0.466761948855)*x[1]**o+(-0.821498435809)*x[1]+(-0.391009601442)*x[2]**o+(0.643834542048)*x[2]
        arg[3,2,2,0]=(0.379015007885)*x[0]**o+(0.348646179575)*x[0]+(0.733765133901)*x[1]**o+(-0.722363571789)*x[1]+(0.488531791931)*x[2]**o+(0.279813923679)*x[2]
        arg[3,2,2,1]=(0.516708382119)*x[0]**o+(0.0426665103314)*x[0]+(-0.168751315718)*x[1]**o+(-0.124069236608)*x[1]+(-0.778866001674)*x[2]**o+(-0.362031001279)*x[2]
        arg[3,3,0,0]=(0.995284803465)*x[0]**o+(-0.442327755879)*x[0]+(-0.13110270925)*x[1]**o+(-0.172857543102)*x[1]+(-0.910522664379)*x[2]**o+(0.516931833163)*x[2]
        arg[3,3,0,1]=(-0.889320295626)*x[0]**o+(-0.728330379534)*x[0]+(-0.584091700627)*x[1]**o+(0.469914082337)*x[1]+(0.18874127891)*x[2]**o+(0.0846532039249)*x[2]
        arg[3,3,1,0]=(0.911523535462)*x[0]**o+(0.0924361220354)*x[0]+(-0.0822504494228)*x[1]**o+(0.00630249023591)*x[1]+(-0.180497330322)*x[2]**o+(0.817405577154)*x[2]
        arg[3,3,1,1]=(0.681872747813)*x[0]**o+(0.840680190206)*x[0]+(-0.477660295206)*x[1]**o+(0.0399086968056)*x[1]+(0.440403204247)*x[2]**o+(-0.220656197016)*x[2]
        arg[3,3,2,0]=(-0.967524407336)*x[0]**o+(0.896652274462)*x[0]+(0.735269328519)*x[1]**o+(-0.249017384614)*x[1]+(-0.573222311563)*x[2]**o+(-0.653800334954)*x[2]
        arg[3,3,2,1]=(0.10041632804)*x[0]**o+(0.167657178348)*x[0]+(0.990621134594)*x[1]**o+(-0.309063473348)*x[1]+(-0.688206827575)*x[2]**o+(0.0926806818663)*x[2]
        arg[3,4,0,0]=(-0.128678339518)*x[0]**o+(-0.974238160059)*x[0]+(-0.762599464772)*x[1]**o+(0.0233286358698)*x[1]+(-0.622481697973)*x[2]**o+(-0.080495303798)*x[2]
        arg[3,4,0,1]=(-0.380318230058)*x[0]**o+(0.568425682425)*x[0]+(0.787369916262)*x[1]**o+(0.296745245123)*x[1]+(-0.917621002414)*x[2]**o+(-0.696209404954)*x[2]
        arg[3,4,1,0]=(-0.382961468898)*x[0]**o+(0.672614453621)*x[0]+(0.565371503874)*x[1]**o+(-0.597331999912)*x[1]+(-0.153637846704)*x[2]**o+(-0.438159587906)*x[2]
        arg[3,4,1,1]=(-0.541043007799)*x[0]**o+(0.415600628159)*x[0]+(-0.793716270785)*x[1]**o+(0.780110667203)*x[1]+(-0.101309486185)*x[2]**o+(-0.138868352939)*x[2]
        arg[3,4,2,0]=(0.679815917569)*x[0]**o+(-0.800657537776)*x[0]+(-0.711956069893)*x[1]**o+(-0.575626146908)*x[1]+(0.184247083249)*x[2]**o+(-0.728118084396)*x[2]
        arg[3,4,2,1]=(-0.608089432748)*x[0]**o+(0.356378798828)*x[0]+(-0.575957599599)*x[1]**o+(0.599586825137)*x[1]+(-0.160475743299)*x[2]**o+(-0.070762356177)*x[2]
        ref[0,0,0,0]=(-0.93229223906)*x_ref[0]**o+(0.176839936782)*x_ref[0]+(0.986271000359)*x_ref[1]**o+(-0.449322385723)*x_ref[1]+(-0.856450437823)*x_ref[2]**o+(0.31473579221)*x_ref[2]
        ref[0,0,0,1]=(0.714047232765)*x_ref[0]**o+(-0.323471323003)*x_ref[0]+(0.628080480343)*x_ref[1]**o+(-0.992007912045)*x_ref[1]+(-0.11019723051)*x_ref[2]**o+(-0.104735408233)*x_ref[2]
        ref[0,0,1,0]=(-0.437250443294)*x_ref[0]**o+(-0.741469652781)*x_ref[0]+(0.650921038465)*x_ref[1]**o+(0.679622741243)*x_ref[1]+(0.604005894895)*x_ref[2]**o+(0.598155624876)*x_ref[2]
        ref[0,0,1,1]=(-0.574872168761)*x_ref[0]**o+(0.796207672391)*x_ref[0]+(-0.478046646159)*x_ref[1]**o+(0.604661326741)*x_ref[1]+(-0.898457020186)*x_ref[2]**o+(0.43234750827)*x_ref[2]
        ref[0,0,2,0]=(-0.674008357742)*x_ref[0]**o+(0.699614371016)*x_ref[0]+(0.441321981559)*x_ref[1]**o+(-0.619688689942)*x_ref[1]+(-0.0874579208832)*x_ref[2]**o+(-0.925753840555)*x_ref[2]
        ref[0,0,2,1]=(-0.114329474892)*x_ref[0]**o+(0.699501071503)*x_ref[0]+(-0.618807420683)*x_ref[1]**o+(0.411584695002)*x_ref[1]+(-0.234620261556)*x_ref[2]**o+(0.585720100967)*x_ref[2]
        ref[0,1,0,0]=(0.617432097253)*x_ref[0]**o+(-0.00304868863676)*x_ref[0]+(0.571021785645)*x_ref[1]**o+(0.457197205728)*x_ref[1]+(0.462895867813)*x_ref[2]**o+(0.278022956316)*x_ref[2]
        ref[0,1,0,1]=(-0.659136231188)*x_ref[0]**o+(-0.388212920906)*x_ref[0]+(-0.982152608232)*x_ref[1]**o+(-0.665797156013)*x_ref[1]+(-0.76141890649)*x_ref[2]**o+(-0.591458323836)*x_ref[2]
        ref[0,1,1,0]=(-0.488016386893)*x_ref[0]**o+(0.965660401944)*x_ref[0]+(0.178400189819)*x_ref[1]**o+(0.578417314499)*x_ref[1]+(-0.398083955217)*x_ref[2]**o+(0.460338187535)*x_ref[2]
        ref[0,1,1,1]=(-0.182086458235)*x_ref[0]**o+(0.0842368498629)*x_ref[0]+(-0.642008026184)*x_ref[1]**o+(-0.0449408131957)*x_ref[1]+(-0.927397573018)*x_ref[2]**o+(0.854313729816)*x_ref[2]
        ref[0,1,2,0]=(-0.201475872769)*x_ref[0]**o+(0.098190765292)*x_ref[0]+(0.808624244922)*x_ref[1]**o+(0.0726268899898)*x_ref[1]+(-0.605334327027)*x_ref[2]**o+(-0.999151255934)*x_ref[2]
        ref[0,1,2,1]=(0.672344024218)*x_ref[0]**o+(-0.638554833481)*x_ref[0]+(-0.212231801634)*x_ref[1]**o+(-0.492765366968)*x_ref[1]+(0.311273771921)*x_ref[2]**o+(-0.504370655006)*x_ref[2]
        ref[0,2,0,0]=(0.9944935907)*x_ref[0]**o+(-0.365261996611)*x_ref[0]+(-0.105667939789)*x_ref[1]**o+(0.886935751977)*x_ref[1]+(0.263336157757)*x_ref[2]**o+(-0.880316523113)*x_ref[2]
        ref[0,2,0,1]=(-0.659779300787)*x_ref[0]**o+(-0.728590180152)*x_ref[0]+(0.732078508037)*x_ref[1]**o+(0.684963412098)*x_ref[1]+(0.822970700012)*x_ref[2]**o+(-0.680001879674)*x_ref[2]
        ref[0,2,1,0]=(-0.207855412893)*x_ref[0]**o+(0.438275426529)*x_ref[0]+(0.87046682066)*x_ref[1]**o+(-0.960322013848)*x_ref[1]+(-0.273728198625)*x_ref[2]**o+(-0.960137228004)*x_ref[2]
        ref[0,2,1,1]=(-0.189232680002)*x_ref[0]**o+(0.866417302002)*x_ref[0]+(-0.93403311248)*x_ref[1]**o+(-0.55773104878)*x_ref[1]+(0.94062895773)*x_ref[2]**o+(-0.758052194805)*x_ref[2]
        ref[0,2,2,0]=(0.100859944394)*x_ref[0]**o+(-0.20292135909)*x_ref[0]+(-0.713339941261)*x_ref[1]**o+(-0.615106454411)*x_ref[1]+(0.791245995152)*x_ref[2]**o+(-0.81051791594)*x_ref[2]
        ref[0,2,2,1]=(0.595743424517)*x_ref[0]**o+(-0.179075876404)*x_ref[0]+(-0.799005601715)*x_ref[1]**o+(-0.0554635073569)*x_ref[1]+(-0.850510701985)*x_ref[2]**o+(0.234516538169)*x_ref[2]
        ref[0,3,0,0]=(-0.361667738861)*x_ref[0]**o+(-0.551837498696)*x_ref[0]+(-0.834377202094)*x_ref[1]**o+(0.814168224265)*x_ref[1]+(-0.00479533268377)*x_ref[2]**o+(0.843007652951)*x_ref[2]
        ref[0,3,0,1]=(-0.448417439465)*x_ref[0]**o+(0.431496158388)*x_ref[0]+(-0.547416443433)*x_ref[1]**o+(-0.0746785704695)*x_ref[1]+(-0.646238182944)*x_ref[2]**o+(-0.318608714595)*x_ref[2]
        ref[0,3,1,0]=(-0.275647366007)*x_ref[0]**o+(-0.410594785047)*x_ref[0]+(-0.858282378196)*x_ref[1]**o+(-0.755586690837)*x_ref[1]+(-0.420481091654)*x_ref[2]**o+(-0.0996487371831)*x_ref[2]
        ref[0,3,1,1]=(-0.395961249009)*x_ref[0]**o+(0.0860896467863)*x_ref[0]+(0.482719525748)*x_ref[1]**o+(0.319368178133)*x_ref[1]+(0.516449153414)*x_ref[2]**o+(-0.246749862576)*x_ref[2]
        ref[0,3,2,0]=(-0.367322027122)*x_ref[0]**o+(0.687206412524)*x_ref[0]+(0.616265233972)*x_ref[1]**o+(-0.32660589112)*x_ref[1]+(-0.944607343096)*x_ref[2]**o+(0.123982537951)*x_ref[2]
        ref[0,3,2,1]=(0.700497166804)*x_ref[0]**o+(0.553961329865)*x_ref[0]+(-0.34020003754)*x_ref[1]**o+(0.295870992783)*x_ref[1]+(-0.220275880046)*x_ref[2]**o+(0.766121035825)*x_ref[2]
        ref[0,4,0,0]=(0.405469789242)*x_ref[0]**o+(-0.338347662873)*x_ref[0]+(0.0761796434233)*x_ref[1]**o+(-0.120599175033)*x_ref[1]+(0.107865585379)*x_ref[2]**o+(0.0711941007838)*x_ref[2]
        ref[0,4,0,1]=(0.614166154899)*x_ref[0]**o+(0.195780702963)*x_ref[0]+(0.718935985215)*x_ref[1]**o+(0.214577296792)*x_ref[1]+(-0.366985242107)*x_ref[2]**o+(-0.509584787633)*x_ref[2]
        ref[0,4,1,0]=(-0.638646370981)*x_ref[0]**o+(-0.769955899284)*x_ref[0]+(-0.887407540503)*x_ref[1]**o+(0.0901245429998)*x_ref[1]+(0.786085641321)*x_ref[2]**o+(0.6248068552)*x_ref[2]
        ref[0,4,1,1]=(-0.145426222043)*x_ref[0]**o+(-0.494114545016)*x_ref[0]+(0.422153721335)*x_ref[1]**o+(0.617022143901)*x_ref[1]+(-0.665972324723)*x_ref[2]**o+(-0.476724432799)*x_ref[2]
        ref[0,4,2,0]=(0.556105252092)*x_ref[0]**o+(-0.893268696809)*x_ref[0]+(0.146441889489)*x_ref[1]**o+(-0.170512113283)*x_ref[1]+(-0.869313262875)*x_ref[2]**o+(-0.976874109392)*x_ref[2]
        ref[0,4,2,1]=(0.858927206499)*x_ref[0]**o+(0.198998355146)*x_ref[0]+(0.278125296049)*x_ref[1]**o+(-0.0442518828093)*x_ref[1]+(0.274852724203)*x_ref[2]**o+(-0.167649193314)*x_ref[2]
        ref[1,0,0,0]=(-0.0429420289353)*x_ref[0]**o+(-0.15764693929)*x_ref[0]+(-0.859599917512)*x_ref[1]**o+(-0.904178643575)*x_ref[1]+(0.519560711493)*x_ref[2]**o+(0.258644717907)*x_ref[2]
        ref[1,0,0,1]=(-0.185321114264)*x_ref[0]**o+(-0.794493830004)*x_ref[0]+(0.136084421162)*x_ref[1]**o+(0.926912021419)*x_ref[1]+(0.346392120388)*x_ref[2]**o+(-0.183908371454)*x_ref[2]
        ref[1,0,1,0]=(-0.730675509877)*x_ref[0]**o+(0.865034802778)*x_ref[0]+(-0.506315431878)*x_ref[1]**o+(-0.94607761051)*x_ref[1]+(0.195194792801)*x_ref[2]**o+(-0.865129258879)*x_ref[2]
        ref[1,0,1,1]=(-0.0597026648107)*x_ref[0]**o+(-0.884790497)*x_ref[0]+(-0.0631278742098)*x_ref[1]**o+(-0.808364940737)*x_ref[1]+(-0.968952604837)*x_ref[2]**o+(0.142246702632)*x_ref[2]
        ref[1,0,2,0]=(0.648472894563)*x_ref[0]**o+(0.145861644073)*x_ref[0]+(0.501227520574)*x_ref[1]**o+(0.265886482306)*x_ref[1]+(-0.0472772173573)*x_ref[2]**o+(-0.364169655969)*x_ref[2]
        ref[1,0,2,1]=(0.847863600695)*x_ref[0]**o+(-0.288841685199)*x_ref[0]+(-0.500273502795)*x_ref[1]**o+(0.627379277592)*x_ref[1]+(0.326814024625)*x_ref[2]**o+(-0.514712825895)*x_ref[2]
        ref[1,1,0,0]=(-0.947845713712)*x_ref[0]**o+(-0.372115616305)*x_ref[0]+(-0.572858279998)*x_ref[1]**o+(0.226237627692)*x_ref[1]+(0.210201947182)*x_ref[2]**o+(0.0147616538594)*x_ref[2]
        ref[1,1,0,1]=(-0.39813630466)*x_ref[0]**o+(-0.658496395731)*x_ref[0]+(-0.363906465005)*x_ref[1]**o+(-0.816719661238)*x_ref[1]+(-0.478106835032)*x_ref[2]**o+(0.0819042309589)*x_ref[2]
        ref[1,1,1,0]=(-0.287160278593)*x_ref[0]**o+(-0.0281402640978)*x_ref[0]+(0.41559795378)*x_ref[1]**o+(0.158274458097)*x_ref[1]+(-0.798429481502)*x_ref[2]**o+(-0.278585334624)*x_ref[2]
        ref[1,1,1,1]=(0.809550480973)*x_ref[0]**o+(0.192026915854)*x_ref[0]+(-0.891168838599)*x_ref[1]**o+(0.0482621192882)*x_ref[1]+(-0.35339708159)*x_ref[2]**o+(-0.133161551553)*x_ref[2]
        ref[1,1,2,0]=(0.939582328872)*x_ref[0]**o+(0.561963790516)*x_ref[0]+(0.885264722949)*x_ref[1]**o+(-0.468042305266)*x_ref[1]+(0.232880002429)*x_ref[2]**o+(-0.772896005701)*x_ref[2]
        ref[1,1,2,1]=(-0.297333303209)*x_ref[0]**o+(-0.414050924207)*x_ref[0]+(-0.278903698873)*x_ref[1]**o+(0.345370620869)*x_ref[1]+(-0.0570151259027)*x_ref[2]**o+(0.786449641378)*x_ref[2]
        ref[1,2,0,0]=(0.644047138589)*x_ref[0]**o+(-0.480660438838)*x_ref[0]+(0.32526499462)*x_ref[1]**o+(-0.775988282127)*x_ref[1]+(0.108374143932)*x_ref[2]**o+(0.955463667181)*x_ref[2]
        ref[1,2,0,1]=(-0.468236790339)*x_ref[0]**o+(0.844197711724)*x_ref[0]+(0.808973429831)*x_ref[1]**o+(-0.647699550005)*x_ref[1]+(0.552555959813)*x_ref[2]**o+(-0.3403079515)*x_ref[2]
        ref[1,2,1,0]=(0.864686362308)*x_ref[0]**o+(0.248923646116)*x_ref[0]+(0.0287042863842)*x_ref[1]**o+(0.404974282153)*x_ref[1]+(0.113501438234)*x_ref[2]**o+(-0.619364894397)*x_ref[2]
        ref[1,2,1,1]=(-0.625146992499)*x_ref[0]**o+(-0.0480602452983)*x_ref[0]+(0.168910550914)*x_ref[1]**o+(-0.891584218123)*x_ref[1]+(0.0609357618263)*x_ref[2]**o+(0.0451921002161)*x_ref[2]
        ref[1,2,2,0]=(-0.478728628752)*x_ref[0]**o+(0.428376547798)*x_ref[0]+(0.817203376818)*x_ref[1]**o+(-0.856056266953)*x_ref[1]+(0.354960313382)*x_ref[2]**o+(-0.611468960461)*x_ref[2]
        ref[1,2,2,1]=(0.49500742113)*x_ref[0]**o+(0.902948130263)*x_ref[0]+(-0.810748543744)*x_ref[1]**o+(-0.696150924858)*x_ref[1]+(-0.748278516161)*x_ref[2]**o+(0.694815527058)*x_ref[2]
        ref[1,3,0,0]=(-0.361960150296)*x_ref[0]**o+(-0.845139780461)*x_ref[0]+(-0.0624499309665)*x_ref[1]**o+(-0.0515558058489)*x_ref[1]+(0.400690983017)*x_ref[2]**o+(-0.744608407974)*x_ref[2]
        ref[1,3,0,1]=(-0.452453617467)*x_ref[0]**o+(0.845396139593)*x_ref[0]+(0.292319469936)*x_ref[1]**o+(-0.833547518634)*x_ref[1]+(-0.511797700353)*x_ref[2]**o+(0.857891454773)*x_ref[2]
        ref[1,3,1,0]=(0.779504092912)*x_ref[0]**o+(0.500524773615)*x_ref[0]+(0.590542659382)*x_ref[1]**o+(0.479310630519)*x_ref[1]+(0.509960641956)*x_ref[2]**o+(-0.564980432959)*x_ref[2]
        ref[1,3,1,1]=(-0.145745189298)*x_ref[0]**o+(-0.256064576359)*x_ref[0]+(0.640890837126)*x_ref[1]**o+(-0.718802035152)*x_ref[1]+(-0.322216873583)*x_ref[2]**o+(-0.0246526526334)*x_ref[2]
        ref[1,3,2,0]=(-0.463434285098)*x_ref[0]**o+(0.116722448643)*x_ref[0]+(-0.690378381342)*x_ref[1]**o+(0.429197465885)*x_ref[1]+(-0.353019731384)*x_ref[2]**o+(0.57397712918)*x_ref[2]
        ref[1,3,2,1]=(0.169676853383)*x_ref[0]**o+(0.815225472623)*x_ref[0]+(-0.11125527552)*x_ref[1]**o+(0.391513831525)*x_ref[1]+(-0.587041361585)*x_ref[2]**o+(0.255194151616)*x_ref[2]
        ref[1,4,0,0]=(0.325351608628)*x_ref[0]**o+(-0.57751715453)*x_ref[0]+(0.145167511303)*x_ref[1]**o+(-0.969286030742)*x_ref[1]+(0.917370398478)*x_ref[2]**o+(0.691483691598)*x_ref[2]
        ref[1,4,0,1]=(0.0997998298271)*x_ref[0]**o+(-0.101248202247)*x_ref[0]+(-0.650649653848)*x_ref[1]**o+(-0.157887947288)*x_ref[1]+(0.517501747765)*x_ref[2]**o+(-0.0852921926298)*x_ref[2]
        ref[1,4,1,0]=(0.984595420407)*x_ref[0]**o+(0.499293499076)*x_ref[0]+(0.352193115874)*x_ref[1]**o+(-0.684522909647)*x_ref[1]+(-0.391538373605)*x_ref[2]**o+(0.404635595162)*x_ref[2]
        ref[1,4,1,1]=(-0.231310482173)*x_ref[0]**o+(-0.0385089836559)*x_ref[0]+(0.247671691888)*x_ref[1]**o+(0.883971955027)*x_ref[1]+(-0.0856624709734)*x_ref[2]**o+(0.942940875904)*x_ref[2]
        ref[1,4,2,0]=(-0.471467257434)*x_ref[0]**o+(0.835634240897)*x_ref[0]+(-0.594421294562)*x_ref[1]**o+(-0.187584387111)*x_ref[1]+(0.859867472765)*x_ref[2]**o+(0.930951154576)*x_ref[2]
        ref[1,4,2,1]=(0.504416975601)*x_ref[0]**o+(0.435692923635)*x_ref[0]+(-0.296945944313)*x_ref[1]**o+(0.815156218765)*x_ref[1]+(0.464464247442)*x_ref[2]**o+(-0.378809453041)*x_ref[2]
        ref[2,0,0,0]=(0.213087423107)*x_ref[0]**o+(-0.504681665801)*x_ref[0]+(0.796352190405)*x_ref[1]**o+(0.546752751136)*x_ref[1]+(0.57347644664)*x_ref[2]**o+(0.715856963195)*x_ref[2]
        ref[2,0,0,1]=(-0.28455266769)*x_ref[0]**o+(0.853162220766)*x_ref[0]+(0.164930982863)*x_ref[1]**o+(-0.0136996498765)*x_ref[1]+(0.293717915079)*x_ref[2]**o+(-0.216124158473)*x_ref[2]
        ref[2,0,1,0]=(-0.328115989441)*x_ref[0]**o+(-0.529441496488)*x_ref[0]+(-0.43579479677)*x_ref[1]**o+(-0.851333801194)*x_ref[1]+(-0.278406102972)*x_ref[2]**o+(0.947777471824)*x_ref[2]
        ref[2,0,1,1]=(-0.756438938789)*x_ref[0]**o+(-0.624332272735)*x_ref[0]+(-0.591638429565)*x_ref[1]**o+(-0.0608281771566)*x_ref[1]+(-0.852078784085)*x_ref[2]**o+(-0.526558964866)*x_ref[2]
        ref[2,0,2,0]=(-0.923892426969)*x_ref[0]**o+(0.788121899878)*x_ref[0]+(-0.446006331202)*x_ref[1]**o+(0.997210871574)*x_ref[1]+(0.356998606103)*x_ref[2]**o+(0.972643822823)*x_ref[2]
        ref[2,0,2,1]=(0.225183308571)*x_ref[0]**o+(-0.757505414777)*x_ref[0]+(-0.59573192074)*x_ref[1]**o+(0.57166154987)*x_ref[1]+(0.622343198809)*x_ref[2]**o+(0.60343970503)*x_ref[2]
        ref[2,1,0,0]=(0.51027702934)*x_ref[0]**o+(-0.711737679522)*x_ref[0]+(-0.211712579627)*x_ref[1]**o+(-0.495817605314)*x_ref[1]+(0.924425804272)*x_ref[2]**o+(0.753424835275)*x_ref[2]
        ref[2,1,0,1]=(0.476218252577)*x_ref[0]**o+(0.187579891557)*x_ref[0]+(0.727359583623)*x_ref[1]**o+(-0.389875559955)*x_ref[1]+(-0.352889326477)*x_ref[2]**o+(0.365143943637)*x_ref[2]
        ref[2,1,1,0]=(-0.815704312943)*x_ref[0]**o+(0.315217470245)*x_ref[0]+(-0.535234555065)*x_ref[1]**o+(-0.433408988501)*x_ref[1]+(-0.676694087389)*x_ref[2]**o+(0.384097177718)*x_ref[2]
        ref[2,1,1,1]=(-0.0368452411345)*x_ref[0]**o+(0.514540686444)*x_ref[0]+(-0.527100066965)*x_ref[1]**o+(0.0291015562912)*x_ref[1]+(-0.605646905004)*x_ref[2]**o+(0.931528599981)*x_ref[2]
        ref[2,1,2,0]=(-0.0994015945266)*x_ref[0]**o+(-0.762310347434)*x_ref[0]+(-0.845934950466)*x_ref[1]**o+(-0.235758967923)*x_ref[1]+(0.755531422698)*x_ref[2]**o+(0.108353867777)*x_ref[2]
        ref[2,1,2,1]=(-0.868779052889)*x_ref[0]**o+(0.783878941075)*x_ref[0]+(0.900089560473)*x_ref[1]**o+(-0.836779405877)*x_ref[1]+(-0.799571882449)*x_ref[2]**o+(0.999710029157)*x_ref[2]
        ref[2,2,0,0]=(0.541465277143)*x_ref[0]**o+(-0.840477797945)*x_ref[0]+(-0.362577721098)*x_ref[1]**o+(0.00680131806718)*x_ref[1]+(-0.338485239867)*x_ref[2]**o+(0.828213756751)*x_ref[2]
        ref[2,2,0,1]=(-0.927347423175)*x_ref[0]**o+(-0.936191125785)*x_ref[0]+(-0.411819804833)*x_ref[1]**o+(-0.0222067637269)*x_ref[1]+(-0.104411832586)*x_ref[2]**o+(-0.79937203081)*x_ref[2]
        ref[2,2,1,0]=(0.273051712848)*x_ref[0]**o+(-0.802468557957)*x_ref[0]+(-0.019392898867)*x_ref[1]**o+(-0.612803074784)*x_ref[1]+(-0.269705464223)*x_ref[2]**o+(-0.244405147706)*x_ref[2]
        ref[2,2,1,1]=(-0.08869649035)*x_ref[0]**o+(0.809761782447)*x_ref[0]+(0.761420080249)*x_ref[1]**o+(0.16590038557)*x_ref[1]+(-0.405657593445)*x_ref[2]**o+(0.123593603037)*x_ref[2]
        ref[2,2,2,0]=(-0.799691160987)*x_ref[0]**o+(0.834835690475)*x_ref[0]+(-0.0863569221737)*x_ref[1]**o+(0.988343215152)*x_ref[1]+(0.294423283474)*x_ref[2]**o+(0.325725138203)*x_ref[2]
        ref[2,2,2,1]=(0.648897496592)*x_ref[0]**o+(-0.474329387314)*x_ref[0]+(-0.904613697104)*x_ref[1]**o+(0.382201388286)*x_ref[1]+(-0.596889148339)*x_ref[2]**o+(0.705132157684)*x_ref[2]
        ref[2,3,0,0]=(0.906738144045)*x_ref[0]**o+(-0.365392026426)*x_ref[0]+(0.730557428733)*x_ref[1]**o+(-0.857864471189)*x_ref[1]+(-0.160930657186)*x_ref[2]**o+(0.885067383334)*x_ref[2]
        ref[2,3,0,1]=(0.992764702472)*x_ref[0]**o+(0.720688532948)*x_ref[0]+(0.318869772818)*x_ref[1]**o+(0.995508347248)*x_ref[1]+(0.108569456996)*x_ref[2]**o+(-0.990939292995)*x_ref[2]
        ref[2,3,1,0]=(0.690584992096)*x_ref[0]**o+(0.605885242246)*x_ref[0]+(0.72371794265)*x_ref[1]**o+(-0.632609325879)*x_ref[1]+(0.520273588734)*x_ref[2]**o+(0.885804143155)*x_ref[2]
        ref[2,3,1,1]=(0.485231542482)*x_ref[0]**o+(0.866920225418)*x_ref[0]+(-0.814964945972)*x_ref[1]**o+(0.544132513209)*x_ref[1]+(-0.155071376494)*x_ref[2]**o+(-0.941300746237)*x_ref[2]
        ref[2,3,2,0]=(0.748232238388)*x_ref[0]**o+(0.179397724419)*x_ref[0]+(-0.43881886248)*x_ref[1]**o+(-0.124941799624)*x_ref[1]+(0.14679335819)*x_ref[2]**o+(-0.758738846688)*x_ref[2]
        ref[2,3,2,1]=(-0.0951333655337)*x_ref[0]**o+(0.37270972476)*x_ref[0]+(-0.651977554115)*x_ref[1]**o+(-0.88918597476)*x_ref[1]+(-0.730858685576)*x_ref[2]**o+(0.685881924459)*x_ref[2]
        ref[2,4,0,0]=(-0.823487466802)*x_ref[0]**o+(-0.583630855643)*x_ref[0]+(-0.664814562747)*x_ref[1]**o+(0.860952605956)*x_ref[1]+(-0.249114322668)*x_ref[2]**o+(-0.968356584237)*x_ref[2]
        ref[2,4,0,1]=(0.44040146252)*x_ref[0]**o+(-0.246827565902)*x_ref[0]+(0.596300980733)*x_ref[1]**o+(-0.825818200414)*x_ref[1]+(0.0715512875967)*x_ref[2]**o+(0.996064679803)*x_ref[2]
        ref[2,4,1,0]=(-0.287806453021)*x_ref[0]**o+(0.704331542927)*x_ref[0]+(-0.346139754521)*x_ref[1]**o+(-0.766667037125)*x_ref[1]+(0.283874132105)*x_ref[2]**o+(-0.613145979204)*x_ref[2]
        ref[2,4,1,1]=(-0.116649875156)*x_ref[0]**o+(0.291380048239)*x_ref[0]+(0.689809221905)*x_ref[1]**o+(0.758935910592)*x_ref[1]+(0.505142787603)*x_ref[2]**o+(-0.113611480503)*x_ref[2]
        ref[2,4,2,0]=(-0.836908572692)*x_ref[0]**o+(-0.21892828725)*x_ref[0]+(-0.935780103868)*x_ref[1]**o+(0.106204643268)*x_ref[1]+(-0.894474423807)*x_ref[2]**o+(0.296554773217)*x_ref[2]
        ref[2,4,2,1]=(-0.89413495023)*x_ref[0]**o+(0.6080813058)*x_ref[0]+(0.087863889878)*x_ref[1]**o+(0.213373665612)*x_ref[1]+(0.17085345025)*x_ref[2]**o+(-0.967625138712)*x_ref[2]
        ref[3,0,0,0]=(0.327537677409)*x_ref[0]**o+(-0.299373059373)*x_ref[0]+(-0.43223648542)*x_ref[1]**o+(-0.499115781051)*x_ref[1]+(0.145464567449)*x_ref[2]**o+(0.580930584563)*x_ref[2]
        ref[3,0,0,1]=(-0.122622543351)*x_ref[0]**o+(-0.220547821081)*x_ref[0]+(0.892679827803)*x_ref[1]**o+(0.436584604381)*x_ref[1]+(-0.156093580288)*x_ref[2]**o+(0.464069069075)*x_ref[2]
        ref[3,0,1,0]=(0.590288663203)*x_ref[0]**o+(-0.212929445251)*x_ref[0]+(0.0899671639061)*x_ref[1]**o+(-0.87729854872)*x_ref[1]+(0.789506558716)*x_ref[2]**o+(-0.68867690907)*x_ref[2]
        ref[3,0,1,1]=(-0.913651849012)*x_ref[0]**o+(-0.430938918892)*x_ref[0]+(-0.351186012371)*x_ref[1]**o+(0.500306994748)*x_ref[1]+(0.145381845281)*x_ref[2]**o+(0.624993552333)*x_ref[2]
        ref[3,0,2,0]=(0.281393881265)*x_ref[0]**o+(-0.704727129826)*x_ref[0]+(-0.0102091836449)*x_ref[1]**o+(0.311857501353)*x_ref[1]+(0.300732947904)*x_ref[2]**o+(0.923973499548)*x_ref[2]
        ref[3,0,2,1]=(-0.584907526993)*x_ref[0]**o+(0.625732316822)*x_ref[0]+(-0.609691930408)*x_ref[1]**o+(0.056160226956)*x_ref[1]+(-0.766583561947)*x_ref[2]**o+(0.307692190194)*x_ref[2]
        ref[3,1,0,0]=(0.731231379445)*x_ref[0]**o+(-0.0485980527115)*x_ref[0]+(0.908932758852)*x_ref[1]**o+(0.188918035886)*x_ref[1]+(0.685837743688)*x_ref[2]**o+(-0.299987553989)*x_ref[2]
        ref[3,1,0,1]=(-0.784006561231)*x_ref[0]**o+(-0.809789430022)*x_ref[0]+(-0.846391792735)*x_ref[1]**o+(-0.211583304405)*x_ref[1]+(-0.00198590063687)*x_ref[2]**o+(-0.853253255224)*x_ref[2]
        ref[3,1,1,0]=(0.595567299132)*x_ref[0]**o+(-0.581734416878)*x_ref[0]+(0.88144481486)*x_ref[1]**o+(0.776876564509)*x_ref[1]+(-0.367477971386)*x_ref[2]**o+(0.0806320164453)*x_ref[2]
        ref[3,1,1,1]=(0.440836187611)*x_ref[0]**o+(-0.578925091757)*x_ref[0]+(0.923367189832)*x_ref[1]**o+(-0.993888173929)*x_ref[1]+(-0.645514380411)*x_ref[2]**o+(0.0267439044199)*x_ref[2]
        ref[3,1,2,0]=(-0.305182288788)*x_ref[0]**o+(-0.503372502553)*x_ref[0]+(0.740948657389)*x_ref[1]**o+(0.285661052275)*x_ref[1]+(-0.634510984218)*x_ref[2]**o+(-0.123979746456)*x_ref[2]
        ref[3,1,2,1]=(-0.897274928495)*x_ref[0]**o+(0.971356796164)*x_ref[0]+(0.940768290313)*x_ref[1]**o+(-0.552941985667)*x_ref[1]+(0.172857451423)*x_ref[2]**o+(0.844758237523)*x_ref[2]
        ref[3,2,0,0]=(-0.903542725333)*x_ref[0]**o+(0.936704774551)*x_ref[0]+(-0.398085875228)*x_ref[1]**o+(-0.304061573512)*x_ref[1]+(-0.13362037229)*x_ref[2]**o+(-0.405043350728)*x_ref[2]
        ref[3,2,0,1]=(-0.0208590751801)*x_ref[0]**o+(-0.354086381885)*x_ref[0]+(0.510884698744)*x_ref[1]**o+(-0.275504880518)*x_ref[1]+(0.459476496356)*x_ref[2]**o+(0.25358097976)*x_ref[2]
        ref[3,2,1,0]=(-0.0171081877107)*x_ref[0]**o+(-0.856081626112)*x_ref[0]+(-0.0698686571004)*x_ref[1]**o+(0.0954918361122)*x_ref[1]+(0.938494954069)*x_ref[2]**o+(-0.824928688248)*x_ref[2]
        ref[3,2,1,1]=(-0.277307915098)*x_ref[0]**o+(0.737855779658)*x_ref[0]+(-0.466761948855)*x_ref[1]**o+(-0.821498435809)*x_ref[1]+(-0.391009601442)*x_ref[2]**o+(0.643834542048)*x_ref[2]
        ref[3,2,2,0]=(0.379015007885)*x_ref[0]**o+(0.348646179575)*x_ref[0]+(0.733765133901)*x_ref[1]**o+(-0.722363571789)*x_ref[1]+(0.488531791931)*x_ref[2]**o+(0.279813923679)*x_ref[2]
        ref[3,2,2,1]=(0.516708382119)*x_ref[0]**o+(0.0426665103314)*x_ref[0]+(-0.168751315718)*x_ref[1]**o+(-0.124069236608)*x_ref[1]+(-0.778866001674)*x_ref[2]**o+(-0.362031001279)*x_ref[2]
        ref[3,3,0,0]=(0.995284803465)*x_ref[0]**o+(-0.442327755879)*x_ref[0]+(-0.13110270925)*x_ref[1]**o+(-0.172857543102)*x_ref[1]+(-0.910522664379)*x_ref[2]**o+(0.516931833163)*x_ref[2]
        ref[3,3,0,1]=(-0.889320295626)*x_ref[0]**o+(-0.728330379534)*x_ref[0]+(-0.584091700627)*x_ref[1]**o+(0.469914082337)*x_ref[1]+(0.18874127891)*x_ref[2]**o+(0.0846532039249)*x_ref[2]
        ref[3,3,1,0]=(0.911523535462)*x_ref[0]**o+(0.0924361220354)*x_ref[0]+(-0.0822504494228)*x_ref[1]**o+(0.00630249023591)*x_ref[1]+(-0.180497330322)*x_ref[2]**o+(0.817405577154)*x_ref[2]
        ref[3,3,1,1]=(0.681872747813)*x_ref[0]**o+(0.840680190206)*x_ref[0]+(-0.477660295206)*x_ref[1]**o+(0.0399086968056)*x_ref[1]+(0.440403204247)*x_ref[2]**o+(-0.220656197016)*x_ref[2]
        ref[3,3,2,0]=(-0.967524407336)*x_ref[0]**o+(0.896652274462)*x_ref[0]+(0.735269328519)*x_ref[1]**o+(-0.249017384614)*x_ref[1]+(-0.573222311563)*x_ref[2]**o+(-0.653800334954)*x_ref[2]
        ref[3,3,2,1]=(0.10041632804)*x_ref[0]**o+(0.167657178348)*x_ref[0]+(0.990621134594)*x_ref[1]**o+(-0.309063473348)*x_ref[1]+(-0.688206827575)*x_ref[2]**o+(0.0926806818663)*x_ref[2]
        ref[3,4,0,0]=(-0.128678339518)*x_ref[0]**o+(-0.974238160059)*x_ref[0]+(-0.762599464772)*x_ref[1]**o+(0.0233286358698)*x_ref[1]+(-0.622481697973)*x_ref[2]**o+(-0.080495303798)*x_ref[2]
        ref[3,4,0,1]=(-0.380318230058)*x_ref[0]**o+(0.568425682425)*x_ref[0]+(0.787369916262)*x_ref[1]**o+(0.296745245123)*x_ref[1]+(-0.917621002414)*x_ref[2]**o+(-0.696209404954)*x_ref[2]
        ref[3,4,1,0]=(-0.382961468898)*x_ref[0]**o+(0.672614453621)*x_ref[0]+(0.565371503874)*x_ref[1]**o+(-0.597331999912)*x_ref[1]+(-0.153637846704)*x_ref[2]**o+(-0.438159587906)*x_ref[2]
        ref[3,4,1,1]=(-0.541043007799)*x_ref[0]**o+(0.415600628159)*x_ref[0]+(-0.793716270785)*x_ref[1]**o+(0.780110667203)*x_ref[1]+(-0.101309486185)*x_ref[2]**o+(-0.138868352939)*x_ref[2]
        ref[3,4,2,0]=(0.679815917569)*x_ref[0]**o+(-0.800657537776)*x_ref[0]+(-0.711956069893)*x_ref[1]**o+(-0.575626146908)*x_ref[1]+(0.184247083249)*x_ref[2]**o+(-0.728118084396)*x_ref[2]
        ref[3,4,2,1]=(-0.608089432748)*x_ref[0]**o+(0.356378798828)*x_ref[0]+(-0.575957599599)*x_ref[1]**o+(0.599586825137)*x_ref[1]+(-0.160475743299)*x_ref[2]**o+(-0.070762356177)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.52738736047)*x[0]**o+(-0.513351179641)*x[0]+(0.466974564277)*x[1]**o+(0.247665236281)*x[1]
        ref=(0.52738736047)*x_ref[0]**o+(-0.513351179641)*x_ref[0]+(0.466974564277)*x_ref[1]**o+(0.247665236281)*x_ref[1]
      else:
        arg=(0.345612544612)*x[0]**o+(0.480207854215)*x[0]+(0.196672493786)*x[1]**o+(-0.891043131902)*x[1]+(0.106095952827)*x[2]**o+(0.436049931061)*x[2]
        ref=(0.345612544612)*x_ref[0]**o+(0.480207854215)*x_ref[0]+(0.196672493786)*x_ref[1]**o+(-0.891043131902)*x_ref[1]+(0.106095952827)*x_ref[2]**o+(0.436049931061)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.738262816349)*x[0]**o+(-0.650523560236)*x[0]+(-0.344136056546)*x[1]**o+(-0.0693168099812)*x[1]
        arg[1]=(-0.565224044511)*x[0]**o+(0.94461728731)*x[0]+(0.234800891741)*x[1]**o+(0.836341748852)*x[1]
        ref[0]=(-0.738262816349)*x_ref[0]**o+(-0.650523560236)*x_ref[0]+(-0.344136056546)*x_ref[1]**o+(-0.0693168099812)*x_ref[1]
        ref[1]=(-0.565224044511)*x_ref[0]**o+(0.94461728731)*x_ref[0]+(0.234800891741)*x_ref[1]**o+(0.836341748852)*x_ref[1]
      else:
        arg[0]=(-0.76925788678)*x[0]**o+(-0.391133765534)*x[0]+(0.89961543807)*x[1]**o+(-0.355090830059)*x[1]+(-0.663969382551)*x[2]**o+(0.0324773110608)*x[2]
        arg[1]=(-0.640271335706)*x[0]**o+(-0.0771899290795)*x[0]+(0.923847889924)*x[1]**o+(-0.732484895713)*x[1]+(-0.782800597476)*x[2]**o+(-0.402230990251)*x[2]
        ref[0]=(-0.76925788678)*x_ref[0]**o+(-0.391133765534)*x_ref[0]+(0.89961543807)*x_ref[1]**o+(-0.355090830059)*x_ref[1]+(-0.663969382551)*x_ref[2]**o+(0.0324773110608)*x_ref[2]
        ref[1]=(-0.640271335706)*x_ref[0]**o+(-0.0771899290795)*x_ref[0]+(0.923847889924)*x_ref[1]**o+(-0.732484895713)*x_ref[1]+(-0.782800597476)*x_ref[2]**o+(-0.402230990251)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.631349529615)*x[0]**o+(0.79216890406)*x[0]+(0.835421397222)*x[1]**o+(-0.209368192458)*x[1]
        arg[0,1]=(-0.166321945093)*x[0]**o+(0.701821475941)*x[0]+(-0.141804772049)*x[1]**o+(-0.860056917696)*x[1]
        arg[0,2]=(0.761659971379)*x[0]**o+(-0.39989119778)*x[0]+(-0.761140588562)*x[1]**o+(-0.595816184788)*x[1]
        arg[0,3]=(0.378518095375)*x[0]**o+(-0.880144120304)*x[0]+(0.814409656431)*x[1]**o+(0.0464425920173)*x[1]
        arg[0,4]=(-0.24431693596)*x[0]**o+(-0.840538735294)*x[0]+(0.0157782225732)*x[1]**o+(0.951311376091)*x[1]
        arg[1,0]=(0.0710042795971)*x[0]**o+(0.936106708209)*x[0]+(-0.580035969982)*x[1]**o+(0.740790378647)*x[1]
        arg[1,1]=(-0.807938035414)*x[0]**o+(0.8723236137)*x[0]+(-0.600615307712)*x[1]**o+(0.71483191921)*x[1]
        arg[1,2]=(0.454239081542)*x[0]**o+(0.966475397079)*x[0]+(-0.184649384525)*x[1]**o+(0.417686859721)*x[1]
        arg[1,3]=(-0.394999073966)*x[0]**o+(-0.0445825855756)*x[0]+(0.0389091203001)*x[1]**o+(-0.265569544304)*x[1]
        arg[1,4]=(0.625419856529)*x[0]**o+(-0.297955079048)*x[0]+(0.0092181343855)*x[1]**o+(-0.423768484981)*x[1]
        arg[2,0]=(0.836466109456)*x[0]**o+(-0.12515777587)*x[0]+(-0.0908815675065)*x[1]**o+(0.825332480799)*x[1]
        arg[2,1]=(-0.454612353892)*x[0]**o+(0.952942366642)*x[0]+(0.538638303761)*x[1]**o+(0.377957206795)*x[1]
        arg[2,2]=(-0.229405757447)*x[0]**o+(-0.125249012236)*x[0]+(0.194040765277)*x[1]**o+(-0.77353897054)*x[1]
        arg[2,3]=(-0.220514078279)*x[0]**o+(-0.937043129926)*x[0]+(0.661270617907)*x[1]**o+(0.342981486241)*x[1]
        arg[2,4]=(0.58858161977)*x[0]**o+(0.324813259161)*x[0]+(0.00439371077545)*x[1]**o+(0.517353738278)*x[1]
        arg[3,0]=(0.130001725331)*x[0]**o+(0.512570729718)*x[0]+(0.971292679928)*x[1]**o+(0.302530140373)*x[1]
        arg[3,1]=(0.999138842976)*x[0]**o+(0.325794749172)*x[0]+(0.94900982582)*x[1]**o+(0.56933825455)*x[1]
        arg[3,2]=(0.965201839092)*x[0]**o+(0.434187660328)*x[0]+(0.842238794411)*x[1]**o+(-0.754540335284)*x[1]
        arg[3,3]=(-0.539162954093)*x[0]**o+(-0.373779084363)*x[0]+(-0.426189858865)*x[1]**o+(-0.812351658629)*x[1]
        arg[3,4]=(0.871516777541)*x[0]**o+(-0.357393696308)*x[0]+(0.267358318776)*x[1]**o+(0.796701523616)*x[1]
        ref[0,0]=(-0.631349529615)*x_ref[0]**o+(0.79216890406)*x_ref[0]+(0.835421397222)*x_ref[1]**o+(-0.209368192458)*x_ref[1]
        ref[0,1]=(-0.166321945093)*x_ref[0]**o+(0.701821475941)*x_ref[0]+(-0.141804772049)*x_ref[1]**o+(-0.860056917696)*x_ref[1]
        ref[0,2]=(0.761659971379)*x_ref[0]**o+(-0.39989119778)*x_ref[0]+(-0.761140588562)*x_ref[1]**o+(-0.595816184788)*x_ref[1]
        ref[0,3]=(0.378518095375)*x_ref[0]**o+(-0.880144120304)*x_ref[0]+(0.814409656431)*x_ref[1]**o+(0.0464425920173)*x_ref[1]
        ref[0,4]=(-0.24431693596)*x_ref[0]**o+(-0.840538735294)*x_ref[0]+(0.0157782225732)*x_ref[1]**o+(0.951311376091)*x_ref[1]
        ref[1,0]=(0.0710042795971)*x_ref[0]**o+(0.936106708209)*x_ref[0]+(-0.580035969982)*x_ref[1]**o+(0.740790378647)*x_ref[1]
        ref[1,1]=(-0.807938035414)*x_ref[0]**o+(0.8723236137)*x_ref[0]+(-0.600615307712)*x_ref[1]**o+(0.71483191921)*x_ref[1]
        ref[1,2]=(0.454239081542)*x_ref[0]**o+(0.966475397079)*x_ref[0]+(-0.184649384525)*x_ref[1]**o+(0.417686859721)*x_ref[1]
        ref[1,3]=(-0.394999073966)*x_ref[0]**o+(-0.0445825855756)*x_ref[0]+(0.0389091203001)*x_ref[1]**o+(-0.265569544304)*x_ref[1]
        ref[1,4]=(0.625419856529)*x_ref[0]**o+(-0.297955079048)*x_ref[0]+(0.0092181343855)*x_ref[1]**o+(-0.423768484981)*x_ref[1]
        ref[2,0]=(0.836466109456)*x_ref[0]**o+(-0.12515777587)*x_ref[0]+(-0.0908815675065)*x_ref[1]**o+(0.825332480799)*x_ref[1]
        ref[2,1]=(-0.454612353892)*x_ref[0]**o+(0.952942366642)*x_ref[0]+(0.538638303761)*x_ref[1]**o+(0.377957206795)*x_ref[1]
        ref[2,2]=(-0.229405757447)*x_ref[0]**o+(-0.125249012236)*x_ref[0]+(0.194040765277)*x_ref[1]**o+(-0.77353897054)*x_ref[1]
        ref[2,3]=(-0.220514078279)*x_ref[0]**o+(-0.937043129926)*x_ref[0]+(0.661270617907)*x_ref[1]**o+(0.342981486241)*x_ref[1]
        ref[2,4]=(0.58858161977)*x_ref[0]**o+(0.324813259161)*x_ref[0]+(0.00439371077545)*x_ref[1]**o+(0.517353738278)*x_ref[1]
        ref[3,0]=(0.130001725331)*x_ref[0]**o+(0.512570729718)*x_ref[0]+(0.971292679928)*x_ref[1]**o+(0.302530140373)*x_ref[1]
        ref[3,1]=(0.999138842976)*x_ref[0]**o+(0.325794749172)*x_ref[0]+(0.94900982582)*x_ref[1]**o+(0.56933825455)*x_ref[1]
        ref[3,2]=(0.965201839092)*x_ref[0]**o+(0.434187660328)*x_ref[0]+(0.842238794411)*x_ref[1]**o+(-0.754540335284)*x_ref[1]
        ref[3,3]=(-0.539162954093)*x_ref[0]**o+(-0.373779084363)*x_ref[0]+(-0.426189858865)*x_ref[1]**o+(-0.812351658629)*x_ref[1]
        ref[3,4]=(0.871516777541)*x_ref[0]**o+(-0.357393696308)*x_ref[0]+(0.267358318776)*x_ref[1]**o+(0.796701523616)*x_ref[1]
      else:
        arg[0,0]=(-0.811142944724)*x[0]**o+(-0.65659407492)*x[0]+(-0.534808740191)*x[1]**o+(0.602252100451)*x[1]+(-0.158845636056)*x[2]**o+(-0.781826300399)*x[2]
        arg[0,1]=(-0.396870942248)*x[0]**o+(0.285698519063)*x[0]+(-0.962418369859)*x[1]**o+(-0.729381989807)*x[1]+(-0.0500038851534)*x[2]**o+(-0.567828852526)*x[2]
        arg[0,2]=(0.346399694372)*x[0]**o+(-0.170517233373)*x[0]+(-0.443275920724)*x[1]**o+(0.581464075295)*x[1]+(0.197230389253)*x[2]**o+(-0.613092248913)*x[2]
        arg[0,3]=(0.920174147113)*x[0]**o+(0.429847401539)*x[0]+(0.112888577662)*x[1]**o+(-0.0540034503714)*x[1]+(-0.0323244631044)*x[2]**o+(-0.293609544741)*x[2]
        arg[0,4]=(-0.0177585560206)*x[0]**o+(-0.806033755558)*x[0]+(-0.363921399918)*x[1]**o+(-0.415721267967)*x[1]+(0.832302676307)*x[2]**o+(-0.102510351741)*x[2]
        arg[1,0]=(-0.769760277763)*x[0]**o+(0.818187369213)*x[0]+(0.454400795864)*x[1]**o+(-0.220119708184)*x[1]+(-0.637117949075)*x[2]**o+(-0.661398404641)*x[2]
        arg[1,1]=(0.828424952723)*x[0]**o+(0.677022266764)*x[0]+(-0.958176510992)*x[1]**o+(-0.732014776269)*x[1]+(-0.495308153354)*x[2]**o+(-0.30698327016)*x[2]
        arg[1,2]=(-0.454106379068)*x[0]**o+(0.686920350455)*x[0]+(0.560344054206)*x[1]**o+(0.355495451775)*x[1]+(0.256969568039)*x[2]**o+(-0.737279585211)*x[2]
        arg[1,3]=(-0.976488579746)*x[0]**o+(-0.652812816758)*x[0]+(-0.996243459964)*x[1]**o+(0.631565668939)*x[1]+(0.996259991179)*x[2]**o+(-0.796610599165)*x[2]
        arg[1,4]=(0.676024716135)*x[0]**o+(-0.179835226483)*x[0]+(0.497541872254)*x[1]**o+(-0.678345104068)*x[1]+(0.945491022682)*x[2]**o+(-0.358403760669)*x[2]
        arg[2,0]=(0.166901676041)*x[0]**o+(0.833434173485)*x[0]+(0.622153751777)*x[1]**o+(0.767679765059)*x[1]+(-0.82271335395)*x[2]**o+(0.667651471482)*x[2]
        arg[2,1]=(-0.405015366041)*x[0]**o+(0.0548291583154)*x[0]+(0.894878845805)*x[1]**o+(-0.0955982667798)*x[1]+(0.468927799304)*x[2]**o+(0.547421481564)*x[2]
        arg[2,2]=(0.893319250823)*x[0]**o+(0.771073524295)*x[0]+(0.993876715991)*x[1]**o+(0.0832975372419)*x[1]+(-0.717061635157)*x[2]**o+(0.303760934227)*x[2]
        arg[2,3]=(0.247006018272)*x[0]**o+(0.329618310414)*x[0]+(-0.824176211089)*x[1]**o+(-0.678238702323)*x[1]+(0.218736302876)*x[2]**o+(0.8652063227)*x[2]
        arg[2,4]=(0.502380810223)*x[0]**o+(-0.94917091185)*x[0]+(-0.0392659831743)*x[1]**o+(-0.308291910503)*x[1]+(-0.498112428439)*x[2]**o+(0.217915791465)*x[2]
        arg[3,0]=(-0.018497051159)*x[0]**o+(-0.543005709743)*x[0]+(-0.433773552307)*x[1]**o+(-0.604850191015)*x[1]+(0.285970270117)*x[2]**o+(0.150938308283)*x[2]
        arg[3,1]=(0.185537796564)*x[0]**o+(0.663947686119)*x[0]+(-0.846597672873)*x[1]**o+(-0.798399069258)*x[1]+(0.0215106260597)*x[2]**o+(-0.388760513164)*x[2]
        arg[3,2]=(-0.378320218313)*x[0]**o+(-0.385079307834)*x[0]+(-0.392389046498)*x[1]**o+(0.323305501943)*x[1]+(0.496285813211)*x[2]**o+(-0.0269835921214)*x[2]
        arg[3,3]=(0.779877220393)*x[0]**o+(0.968536241033)*x[0]+(0.786166009582)*x[1]**o+(-0.974838905578)*x[1]+(0.4579382269)*x[2]**o+(0.178464213473)*x[2]
        arg[3,4]=(-0.149895758057)*x[0]**o+(0.3529889311)*x[0]+(0.223924981441)*x[1]**o+(0.10143440581)*x[1]+(0.162222746832)*x[2]**o+(0.421433640986)*x[2]
        ref[0,0]=(-0.811142944724)*x_ref[0]**o+(-0.65659407492)*x_ref[0]+(-0.534808740191)*x_ref[1]**o+(0.602252100451)*x_ref[1]+(-0.158845636056)*x_ref[2]**o+(-0.781826300399)*x_ref[2]
        ref[0,1]=(-0.396870942248)*x_ref[0]**o+(0.285698519063)*x_ref[0]+(-0.962418369859)*x_ref[1]**o+(-0.729381989807)*x_ref[1]+(-0.0500038851534)*x_ref[2]**o+(-0.567828852526)*x_ref[2]
        ref[0,2]=(0.346399694372)*x_ref[0]**o+(-0.170517233373)*x_ref[0]+(-0.443275920724)*x_ref[1]**o+(0.581464075295)*x_ref[1]+(0.197230389253)*x_ref[2]**o+(-0.613092248913)*x_ref[2]
        ref[0,3]=(0.920174147113)*x_ref[0]**o+(0.429847401539)*x_ref[0]+(0.112888577662)*x_ref[1]**o+(-0.0540034503714)*x_ref[1]+(-0.0323244631044)*x_ref[2]**o+(-0.293609544741)*x_ref[2]
        ref[0,4]=(-0.0177585560206)*x_ref[0]**o+(-0.806033755558)*x_ref[0]+(-0.363921399918)*x_ref[1]**o+(-0.415721267967)*x_ref[1]+(0.832302676307)*x_ref[2]**o+(-0.102510351741)*x_ref[2]
        ref[1,0]=(-0.769760277763)*x_ref[0]**o+(0.818187369213)*x_ref[0]+(0.454400795864)*x_ref[1]**o+(-0.220119708184)*x_ref[1]+(-0.637117949075)*x_ref[2]**o+(-0.661398404641)*x_ref[2]
        ref[1,1]=(0.828424952723)*x_ref[0]**o+(0.677022266764)*x_ref[0]+(-0.958176510992)*x_ref[1]**o+(-0.732014776269)*x_ref[1]+(-0.495308153354)*x_ref[2]**o+(-0.30698327016)*x_ref[2]
        ref[1,2]=(-0.454106379068)*x_ref[0]**o+(0.686920350455)*x_ref[0]+(0.560344054206)*x_ref[1]**o+(0.355495451775)*x_ref[1]+(0.256969568039)*x_ref[2]**o+(-0.737279585211)*x_ref[2]
        ref[1,3]=(-0.976488579746)*x_ref[0]**o+(-0.652812816758)*x_ref[0]+(-0.996243459964)*x_ref[1]**o+(0.631565668939)*x_ref[1]+(0.996259991179)*x_ref[2]**o+(-0.796610599165)*x_ref[2]
        ref[1,4]=(0.676024716135)*x_ref[0]**o+(-0.179835226483)*x_ref[0]+(0.497541872254)*x_ref[1]**o+(-0.678345104068)*x_ref[1]+(0.945491022682)*x_ref[2]**o+(-0.358403760669)*x_ref[2]
        ref[2,0]=(0.166901676041)*x_ref[0]**o+(0.833434173485)*x_ref[0]+(0.622153751777)*x_ref[1]**o+(0.767679765059)*x_ref[1]+(-0.82271335395)*x_ref[2]**o+(0.667651471482)*x_ref[2]
        ref[2,1]=(-0.405015366041)*x_ref[0]**o+(0.0548291583154)*x_ref[0]+(0.894878845805)*x_ref[1]**o+(-0.0955982667798)*x_ref[1]+(0.468927799304)*x_ref[2]**o+(0.547421481564)*x_ref[2]
        ref[2,2]=(0.893319250823)*x_ref[0]**o+(0.771073524295)*x_ref[0]+(0.993876715991)*x_ref[1]**o+(0.0832975372419)*x_ref[1]+(-0.717061635157)*x_ref[2]**o+(0.303760934227)*x_ref[2]
        ref[2,3]=(0.247006018272)*x_ref[0]**o+(0.329618310414)*x_ref[0]+(-0.824176211089)*x_ref[1]**o+(-0.678238702323)*x_ref[1]+(0.218736302876)*x_ref[2]**o+(0.8652063227)*x_ref[2]
        ref[2,4]=(0.502380810223)*x_ref[0]**o+(-0.94917091185)*x_ref[0]+(-0.0392659831743)*x_ref[1]**o+(-0.308291910503)*x_ref[1]+(-0.498112428439)*x_ref[2]**o+(0.217915791465)*x_ref[2]
        ref[3,0]=(-0.018497051159)*x_ref[0]**o+(-0.543005709743)*x_ref[0]+(-0.433773552307)*x_ref[1]**o+(-0.604850191015)*x_ref[1]+(0.285970270117)*x_ref[2]**o+(0.150938308283)*x_ref[2]
        ref[3,1]=(0.185537796564)*x_ref[0]**o+(0.663947686119)*x_ref[0]+(-0.846597672873)*x_ref[1]**o+(-0.798399069258)*x_ref[1]+(0.0215106260597)*x_ref[2]**o+(-0.388760513164)*x_ref[2]
        ref[3,2]=(-0.378320218313)*x_ref[0]**o+(-0.385079307834)*x_ref[0]+(-0.392389046498)*x_ref[1]**o+(0.323305501943)*x_ref[1]+(0.496285813211)*x_ref[2]**o+(-0.0269835921214)*x_ref[2]
        ref[3,3]=(0.779877220393)*x_ref[0]**o+(0.968536241033)*x_ref[0]+(0.786166009582)*x_ref[1]**o+(-0.974838905578)*x_ref[1]+(0.4579382269)*x_ref[2]**o+(0.178464213473)*x_ref[2]
        ref[3,4]=(-0.149895758057)*x_ref[0]**o+(0.3529889311)*x_ref[0]+(0.223924981441)*x_ref[1]**o+(0.10143440581)*x_ref[1]+(0.162222746832)*x_ref[2]**o+(0.421433640986)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.232971584272)*x[0]**o+(0.391351964909)*x[0]+(-0.832254448983)*x[1]**o+(0.880762034713)*x[1]
        arg[0,0,1]=(0.473097471101)*x[0]**o+(-0.660862772376)*x[0]+(0.72974853865)*x[1]**o+(-0.328183435687)*x[1]
        arg[0,1,0]=(-0.749483736482)*x[0]**o+(0.0872454020187)*x[0]+(0.6916024322)*x[1]**o+(0.82757019273)*x[1]
        arg[0,1,1]=(-0.325601366346)*x[0]**o+(0.918246327277)*x[0]+(0.855963838252)*x[1]**o+(0.448567464706)*x[1]
        arg[1,0,0]=(-0.6642143628)*x[0]**o+(-0.82649833447)*x[0]+(0.822778465465)*x[1]**o+(-0.926669226842)*x[1]
        arg[1,0,1]=(-0.644490899373)*x[0]**o+(0.0644054772987)*x[0]+(-0.914922609343)*x[1]**o+(0.898880463842)*x[1]
        arg[1,1,0]=(0.40963096867)*x[0]**o+(-0.347418076162)*x[0]+(0.857966073118)*x[1]**o+(-0.284521442085)*x[1]
        arg[1,1,1]=(-0.00998062171478)*x[0]**o+(-0.234122404174)*x[0]+(-0.149500951208)*x[1]**o+(0.680038328937)*x[1]
        arg[2,0,0]=(-0.963777985833)*x[0]**o+(0.710727608341)*x[0]+(-0.646898391724)*x[1]**o+(-0.464095155503)*x[1]
        arg[2,0,1]=(-0.617670394851)*x[0]**o+(0.592337321587)*x[0]+(-0.0881426217074)*x[1]**o+(-0.00431912882662)*x[1]
        arg[2,1,0]=(0.871279828121)*x[0]**o+(-0.858267190204)*x[0]+(0.440863109038)*x[1]**o+(-0.912521398099)*x[1]
        arg[2,1,1]=(0.17802472174)*x[0]**o+(0.948587960042)*x[0]+(-0.837945850976)*x[1]**o+(0.13308730872)*x[1]
        arg[3,0,0]=(0.0482409149433)*x[0]**o+(-0.198100088194)*x[0]+(0.504789334224)*x[1]**o+(0.296276808608)*x[1]
        arg[3,0,1]=(-0.30515363985)*x[0]**o+(-0.355293167256)*x[0]+(-0.577592997685)*x[1]**o+(-0.882408725943)*x[1]
        arg[3,1,0]=(0.76102937294)*x[0]**o+(-0.210169673144)*x[0]+(-0.181755140891)*x[1]**o+(0.865448484187)*x[1]
        arg[3,1,1]=(-0.894294091863)*x[0]**o+(-0.712245666103)*x[0]+(0.151657455329)*x[1]**o+(-0.54019436763)*x[1]
        arg[4,0,0]=(0.860662861318)*x[0]**o+(-0.36787473166)*x[0]+(0.240406823985)*x[1]**o+(-0.381822992473)*x[1]
        arg[4,0,1]=(-0.109197125112)*x[0]**o+(0.962595232385)*x[0]+(-0.323876506847)*x[1]**o+(0.178613363564)*x[1]
        arg[4,1,0]=(0.486980313424)*x[0]**o+(-0.611294566285)*x[0]+(-0.774813150316)*x[1]**o+(0.335411840405)*x[1]
        arg[4,1,1]=(-0.622163740867)*x[0]**o+(0.426305234663)*x[0]+(0.243557551101)*x[1]**o+(0.930601340004)*x[1]
        arg[5,0,0]=(-0.232795962027)*x[0]**o+(0.49010767268)*x[0]+(-0.503693305594)*x[1]**o+(-0.915931668186)*x[1]
        arg[5,0,1]=(-0.304614759138)*x[0]**o+(0.797445765321)*x[0]+(-0.00820825778888)*x[1]**o+(-0.196762463972)*x[1]
        arg[5,1,0]=(-0.516700339456)*x[0]**o+(-0.440953905414)*x[0]+(0.875266018165)*x[1]**o+(0.780895354543)*x[1]
        arg[5,1,1]=(0.219827931399)*x[0]**o+(-0.950514094131)*x[0]+(-0.120763970425)*x[1]**o+(0.0941571726596)*x[1]
        ref[0,0,0]=(-0.232971584272)*x_ref[0]**o+(0.391351964909)*x_ref[0]+(-0.832254448983)*x_ref[1]**o+(0.880762034713)*x_ref[1]
        ref[0,0,1]=(0.473097471101)*x_ref[0]**o+(-0.660862772376)*x_ref[0]+(0.72974853865)*x_ref[1]**o+(-0.328183435687)*x_ref[1]
        ref[0,1,0]=(-0.749483736482)*x_ref[0]**o+(0.0872454020187)*x_ref[0]+(0.6916024322)*x_ref[1]**o+(0.82757019273)*x_ref[1]
        ref[0,1,1]=(-0.325601366346)*x_ref[0]**o+(0.918246327277)*x_ref[0]+(0.855963838252)*x_ref[1]**o+(0.448567464706)*x_ref[1]
        ref[1,0,0]=(-0.6642143628)*x_ref[0]**o+(-0.82649833447)*x_ref[0]+(0.822778465465)*x_ref[1]**o+(-0.926669226842)*x_ref[1]
        ref[1,0,1]=(-0.644490899373)*x_ref[0]**o+(0.0644054772987)*x_ref[0]+(-0.914922609343)*x_ref[1]**o+(0.898880463842)*x_ref[1]
        ref[1,1,0]=(0.40963096867)*x_ref[0]**o+(-0.347418076162)*x_ref[0]+(0.857966073118)*x_ref[1]**o+(-0.284521442085)*x_ref[1]
        ref[1,1,1]=(-0.00998062171478)*x_ref[0]**o+(-0.234122404174)*x_ref[0]+(-0.149500951208)*x_ref[1]**o+(0.680038328937)*x_ref[1]
        ref[2,0,0]=(-0.963777985833)*x_ref[0]**o+(0.710727608341)*x_ref[0]+(-0.646898391724)*x_ref[1]**o+(-0.464095155503)*x_ref[1]
        ref[2,0,1]=(-0.617670394851)*x_ref[0]**o+(0.592337321587)*x_ref[0]+(-0.0881426217074)*x_ref[1]**o+(-0.00431912882662)*x_ref[1]
        ref[2,1,0]=(0.871279828121)*x_ref[0]**o+(-0.858267190204)*x_ref[0]+(0.440863109038)*x_ref[1]**o+(-0.912521398099)*x_ref[1]
        ref[2,1,1]=(0.17802472174)*x_ref[0]**o+(0.948587960042)*x_ref[0]+(-0.837945850976)*x_ref[1]**o+(0.13308730872)*x_ref[1]
        ref[3,0,0]=(0.0482409149433)*x_ref[0]**o+(-0.198100088194)*x_ref[0]+(0.504789334224)*x_ref[1]**o+(0.296276808608)*x_ref[1]
        ref[3,0,1]=(-0.30515363985)*x_ref[0]**o+(-0.355293167256)*x_ref[0]+(-0.577592997685)*x_ref[1]**o+(-0.882408725943)*x_ref[1]
        ref[3,1,0]=(0.76102937294)*x_ref[0]**o+(-0.210169673144)*x_ref[0]+(-0.181755140891)*x_ref[1]**o+(0.865448484187)*x_ref[1]
        ref[3,1,1]=(-0.894294091863)*x_ref[0]**o+(-0.712245666103)*x_ref[0]+(0.151657455329)*x_ref[1]**o+(-0.54019436763)*x_ref[1]
        ref[4,0,0]=(0.860662861318)*x_ref[0]**o+(-0.36787473166)*x_ref[0]+(0.240406823985)*x_ref[1]**o+(-0.381822992473)*x_ref[1]
        ref[4,0,1]=(-0.109197125112)*x_ref[0]**o+(0.962595232385)*x_ref[0]+(-0.323876506847)*x_ref[1]**o+(0.178613363564)*x_ref[1]
        ref[4,1,0]=(0.486980313424)*x_ref[0]**o+(-0.611294566285)*x_ref[0]+(-0.774813150316)*x_ref[1]**o+(0.335411840405)*x_ref[1]
        ref[4,1,1]=(-0.622163740867)*x_ref[0]**o+(0.426305234663)*x_ref[0]+(0.243557551101)*x_ref[1]**o+(0.930601340004)*x_ref[1]
        ref[5,0,0]=(-0.232795962027)*x_ref[0]**o+(0.49010767268)*x_ref[0]+(-0.503693305594)*x_ref[1]**o+(-0.915931668186)*x_ref[1]
        ref[5,0,1]=(-0.304614759138)*x_ref[0]**o+(0.797445765321)*x_ref[0]+(-0.00820825778888)*x_ref[1]**o+(-0.196762463972)*x_ref[1]
        ref[5,1,0]=(-0.516700339456)*x_ref[0]**o+(-0.440953905414)*x_ref[0]+(0.875266018165)*x_ref[1]**o+(0.780895354543)*x_ref[1]
        ref[5,1,1]=(0.219827931399)*x_ref[0]**o+(-0.950514094131)*x_ref[0]+(-0.120763970425)*x_ref[1]**o+(0.0941571726596)*x_ref[1]
      else:
        arg[0,0,0]=(0.772105181243)*x[0]**o+(-0.548341656718)*x[0]+(0.358327458717)*x[1]**o+(0.07838537297)*x[1]+(-0.221432264133)*x[2]**o+(-0.0831433214109)*x[2]
        arg[0,0,1]=(-0.920280406855)*x[0]**o+(0.288008584982)*x[0]+(0.273175197348)*x[1]**o+(0.71005891586)*x[1]+(-0.372322860924)*x[2]**o+(0.801820758328)*x[2]
        arg[0,1,0]=(0.610498050037)*x[0]**o+(-0.539393441621)*x[0]+(0.587033652815)*x[1]**o+(0.237738229824)*x[1]+(-0.79411765083)*x[2]**o+(-0.519989054973)*x[2]
        arg[0,1,1]=(0.5412774887)*x[0]**o+(-0.851503637634)*x[0]+(-0.71956793056)*x[1]**o+(0.166039550456)*x[1]+(-0.274290267921)*x[2]**o+(-0.909146653718)*x[2]
        arg[1,0,0]=(-0.00458109759775)*x[0]**o+(0.487263914082)*x[0]+(0.0171216930245)*x[1]**o+(0.691063106679)*x[1]+(-0.96202326744)*x[2]**o+(0.361419732405)*x[2]
        arg[1,0,1]=(-0.784311451677)*x[0]**o+(-0.862560770078)*x[0]+(-0.0702620499924)*x[1]**o+(0.183400389223)*x[1]+(0.364905596839)*x[2]**o+(0.874230430841)*x[2]
        arg[1,1,0]=(0.315952660065)*x[0]**o+(-0.0168150653709)*x[0]+(0.372807793658)*x[1]**o+(0.350959428104)*x[1]+(0.292074907124)*x[2]**o+(0.38176271099)*x[2]
        arg[1,1,1]=(-0.941288490366)*x[0]**o+(0.159864446187)*x[0]+(0.616516779388)*x[1]**o+(0.0940731931915)*x[1]+(-0.966416702566)*x[2]**o+(-0.716669694143)*x[2]
        arg[2,0,0]=(0.308619329706)*x[0]**o+(0.68203622136)*x[0]+(-0.732163723834)*x[1]**o+(-0.222214324226)*x[1]+(0.89540478356)*x[2]**o+(-0.0648727721753)*x[2]
        arg[2,0,1]=(-0.829002921476)*x[0]**o+(0.274738860807)*x[0]+(0.944302912883)*x[1]**o+(0.67434413206)*x[1]+(-0.861085567199)*x[2]**o+(0.994494977369)*x[2]
        arg[2,1,0]=(-0.871499634896)*x[0]**o+(-0.716060125296)*x[0]+(-0.655428500508)*x[1]**o+(-0.478135345667)*x[1]+(-0.858927907345)*x[2]**o+(0.718884168928)*x[2]
        arg[2,1,1]=(0.843612813106)*x[0]**o+(-0.221966655825)*x[0]+(-0.0770757351)*x[1]**o+(-0.678171902341)*x[1]+(0.814209490483)*x[2]**o+(0.409183122676)*x[2]
        arg[3,0,0]=(-0.226810706884)*x[0]**o+(-0.167181963701)*x[0]+(0.810478050432)*x[1]**o+(-0.335099187817)*x[1]+(0.210621363393)*x[2]**o+(0.429297116788)*x[2]
        arg[3,0,1]=(0.597117204298)*x[0]**o+(-0.000147371125322)*x[0]+(-0.557842700291)*x[1]**o+(0.329529225331)*x[1]+(-0.208265244071)*x[2]**o+(-0.977698555357)*x[2]
        arg[3,1,0]=(-0.163068070947)*x[0]**o+(0.640075503516)*x[0]+(0.425653990651)*x[1]**o+(0.650655294789)*x[1]+(-0.342488862403)*x[2]**o+(0.227438788286)*x[2]
        arg[3,1,1]=(-0.511981920097)*x[0]**o+(-0.18099315756)*x[0]+(-0.969881140182)*x[1]**o+(-0.418742703662)*x[1]+(-0.271422367765)*x[2]**o+(0.0812058381877)*x[2]
        arg[4,0,0]=(0.569514749505)*x[0]**o+(-0.711561480184)*x[0]+(-0.587159226794)*x[1]**o+(-0.240459786489)*x[1]+(-0.380983629494)*x[2]**o+(-0.550385464258)*x[2]
        arg[4,0,1]=(0.257375332759)*x[0]**o+(-0.477083226492)*x[0]+(0.0180233249756)*x[1]**o+(0.612470082967)*x[1]+(0.99566078545)*x[2]**o+(-0.551544637583)*x[2]
        arg[4,1,0]=(0.702787194057)*x[0]**o+(0.922262573534)*x[0]+(-0.465910368685)*x[1]**o+(0.748743004032)*x[1]+(0.642852066952)*x[2]**o+(0.24283802317)*x[2]
        arg[4,1,1]=(0.745762578199)*x[0]**o+(0.240631495429)*x[0]+(0.413077167171)*x[1]**o+(0.692304442104)*x[1]+(0.150058687546)*x[2]**o+(-0.133413399526)*x[2]
        arg[5,0,0]=(0.0244935245274)*x[0]**o+(-0.866653903557)*x[0]+(-0.682252235104)*x[1]**o+(-0.525092961647)*x[1]+(-0.87552740837)*x[2]**o+(-0.780785295196)*x[2]
        arg[5,0,1]=(0.403317147145)*x[0]**o+(-0.846638525869)*x[0]+(-0.798622632135)*x[1]**o+(-0.199698251819)*x[1]+(-0.852399643056)*x[2]**o+(-0.0618243339514)*x[2]
        arg[5,1,0]=(-0.94578445284)*x[0]**o+(-0.715570943968)*x[0]+(-0.193533970382)*x[1]**o+(0.0325400011289)*x[1]+(0.137922635356)*x[2]**o+(0.0192967461384)*x[2]
        arg[5,1,1]=(-0.706542583366)*x[0]**o+(0.412536423598)*x[0]+(-0.27322227526)*x[1]**o+(0.339116123776)*x[1]+(-0.856871339939)*x[2]**o+(0.140603890427)*x[2]
        ref[0,0,0]=(0.772105181243)*x_ref[0]**o+(-0.548341656718)*x_ref[0]+(0.358327458717)*x_ref[1]**o+(0.07838537297)*x_ref[1]+(-0.221432264133)*x_ref[2]**o+(-0.0831433214109)*x_ref[2]
        ref[0,0,1]=(-0.920280406855)*x_ref[0]**o+(0.288008584982)*x_ref[0]+(0.273175197348)*x_ref[1]**o+(0.71005891586)*x_ref[1]+(-0.372322860924)*x_ref[2]**o+(0.801820758328)*x_ref[2]
        ref[0,1,0]=(0.610498050037)*x_ref[0]**o+(-0.539393441621)*x_ref[0]+(0.587033652815)*x_ref[1]**o+(0.237738229824)*x_ref[1]+(-0.79411765083)*x_ref[2]**o+(-0.519989054973)*x_ref[2]
        ref[0,1,1]=(0.5412774887)*x_ref[0]**o+(-0.851503637634)*x_ref[0]+(-0.71956793056)*x_ref[1]**o+(0.166039550456)*x_ref[1]+(-0.274290267921)*x_ref[2]**o+(-0.909146653718)*x_ref[2]
        ref[1,0,0]=(-0.00458109759775)*x_ref[0]**o+(0.487263914082)*x_ref[0]+(0.0171216930245)*x_ref[1]**o+(0.691063106679)*x_ref[1]+(-0.96202326744)*x_ref[2]**o+(0.361419732405)*x_ref[2]
        ref[1,0,1]=(-0.784311451677)*x_ref[0]**o+(-0.862560770078)*x_ref[0]+(-0.0702620499924)*x_ref[1]**o+(0.183400389223)*x_ref[1]+(0.364905596839)*x_ref[2]**o+(0.874230430841)*x_ref[2]
        ref[1,1,0]=(0.315952660065)*x_ref[0]**o+(-0.0168150653709)*x_ref[0]+(0.372807793658)*x_ref[1]**o+(0.350959428104)*x_ref[1]+(0.292074907124)*x_ref[2]**o+(0.38176271099)*x_ref[2]
        ref[1,1,1]=(-0.941288490366)*x_ref[0]**o+(0.159864446187)*x_ref[0]+(0.616516779388)*x_ref[1]**o+(0.0940731931915)*x_ref[1]+(-0.966416702566)*x_ref[2]**o+(-0.716669694143)*x_ref[2]
        ref[2,0,0]=(0.308619329706)*x_ref[0]**o+(0.68203622136)*x_ref[0]+(-0.732163723834)*x_ref[1]**o+(-0.222214324226)*x_ref[1]+(0.89540478356)*x_ref[2]**o+(-0.0648727721753)*x_ref[2]
        ref[2,0,1]=(-0.829002921476)*x_ref[0]**o+(0.274738860807)*x_ref[0]+(0.944302912883)*x_ref[1]**o+(0.67434413206)*x_ref[1]+(-0.861085567199)*x_ref[2]**o+(0.994494977369)*x_ref[2]
        ref[2,1,0]=(-0.871499634896)*x_ref[0]**o+(-0.716060125296)*x_ref[0]+(-0.655428500508)*x_ref[1]**o+(-0.478135345667)*x_ref[1]+(-0.858927907345)*x_ref[2]**o+(0.718884168928)*x_ref[2]
        ref[2,1,1]=(0.843612813106)*x_ref[0]**o+(-0.221966655825)*x_ref[0]+(-0.0770757351)*x_ref[1]**o+(-0.678171902341)*x_ref[1]+(0.814209490483)*x_ref[2]**o+(0.409183122676)*x_ref[2]
        ref[3,0,0]=(-0.226810706884)*x_ref[0]**o+(-0.167181963701)*x_ref[0]+(0.810478050432)*x_ref[1]**o+(-0.335099187817)*x_ref[1]+(0.210621363393)*x_ref[2]**o+(0.429297116788)*x_ref[2]
        ref[3,0,1]=(0.597117204298)*x_ref[0]**o+(-0.000147371125322)*x_ref[0]+(-0.557842700291)*x_ref[1]**o+(0.329529225331)*x_ref[1]+(-0.208265244071)*x_ref[2]**o+(-0.977698555357)*x_ref[2]
        ref[3,1,0]=(-0.163068070947)*x_ref[0]**o+(0.640075503516)*x_ref[0]+(0.425653990651)*x_ref[1]**o+(0.650655294789)*x_ref[1]+(-0.342488862403)*x_ref[2]**o+(0.227438788286)*x_ref[2]
        ref[3,1,1]=(-0.511981920097)*x_ref[0]**o+(-0.18099315756)*x_ref[0]+(-0.969881140182)*x_ref[1]**o+(-0.418742703662)*x_ref[1]+(-0.271422367765)*x_ref[2]**o+(0.0812058381877)*x_ref[2]
        ref[4,0,0]=(0.569514749505)*x_ref[0]**o+(-0.711561480184)*x_ref[0]+(-0.587159226794)*x_ref[1]**o+(-0.240459786489)*x_ref[1]+(-0.380983629494)*x_ref[2]**o+(-0.550385464258)*x_ref[2]
        ref[4,0,1]=(0.257375332759)*x_ref[0]**o+(-0.477083226492)*x_ref[0]+(0.0180233249756)*x_ref[1]**o+(0.612470082967)*x_ref[1]+(0.99566078545)*x_ref[2]**o+(-0.551544637583)*x_ref[2]
        ref[4,1,0]=(0.702787194057)*x_ref[0]**o+(0.922262573534)*x_ref[0]+(-0.465910368685)*x_ref[1]**o+(0.748743004032)*x_ref[1]+(0.642852066952)*x_ref[2]**o+(0.24283802317)*x_ref[2]
        ref[4,1,1]=(0.745762578199)*x_ref[0]**o+(0.240631495429)*x_ref[0]+(0.413077167171)*x_ref[1]**o+(0.692304442104)*x_ref[1]+(0.150058687546)*x_ref[2]**o+(-0.133413399526)*x_ref[2]
        ref[5,0,0]=(0.0244935245274)*x_ref[0]**o+(-0.866653903557)*x_ref[0]+(-0.682252235104)*x_ref[1]**o+(-0.525092961647)*x_ref[1]+(-0.87552740837)*x_ref[2]**o+(-0.780785295196)*x_ref[2]
        ref[5,0,1]=(0.403317147145)*x_ref[0]**o+(-0.846638525869)*x_ref[0]+(-0.798622632135)*x_ref[1]**o+(-0.199698251819)*x_ref[1]+(-0.852399643056)*x_ref[2]**o+(-0.0618243339514)*x_ref[2]
        ref[5,1,0]=(-0.94578445284)*x_ref[0]**o+(-0.715570943968)*x_ref[0]+(-0.193533970382)*x_ref[1]**o+(0.0325400011289)*x_ref[1]+(0.137922635356)*x_ref[2]**o+(0.0192967461384)*x_ref[2]
        ref[5,1,1]=(-0.706542583366)*x_ref[0]**o+(0.412536423598)*x_ref[0]+(-0.27322227526)*x_ref[1]**o+(0.339116123776)*x_ref[1]+(-0.856871339939)*x_ref[2]**o+(0.140603890427)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0577426785287)*x[0]**o+(0.469845292021)*x[0]+(-0.641706722819)*x[1]**o+(0.129652001842)*x[1]
        arg[0,0,0,1]=(-0.87686951076)*x[0]**o+(-0.981093040999)*x[0]+(0.654637319552)*x[1]**o+(-0.953104888811)*x[1]
        arg[0,0,1,0]=(-0.486386436415)*x[0]**o+(0.91083470956)*x[0]+(-0.392432937401)*x[1]**o+(0.494919250884)*x[1]
        arg[0,0,1,1]=(0.299928502076)*x[0]**o+(-0.0787848191157)*x[0]+(-0.541863592475)*x[1]**o+(0.435934816687)*x[1]
        arg[0,0,2,0]=(-0.0500083438365)*x[0]**o+(-0.882581886623)*x[0]+(-0.965347357063)*x[1]**o+(-0.270254041774)*x[1]
        arg[0,0,2,1]=(-0.18974861517)*x[0]**o+(0.341406736671)*x[0]+(0.847020774229)*x[1]**o+(0.650201215849)*x[1]
        arg[0,1,0,0]=(0.0823017084719)*x[0]**o+(0.964585359235)*x[0]+(0.357658660436)*x[1]**o+(0.182585271812)*x[1]
        arg[0,1,0,1]=(0.863411821905)*x[0]**o+(0.313374566426)*x[0]+(-0.273547644827)*x[1]**o+(0.429051854479)*x[1]
        arg[0,1,1,0]=(0.167608645068)*x[0]**o+(0.926908094967)*x[0]+(-0.853415240465)*x[1]**o+(-0.376634525241)*x[1]
        arg[0,1,1,1]=(-0.555491961591)*x[0]**o+(-0.64621137308)*x[0]+(-0.315877124991)*x[1]**o+(-0.962839286133)*x[1]
        arg[0,1,2,0]=(0.909012065667)*x[0]**o+(-0.656739022976)*x[0]+(-0.458377773547)*x[1]**o+(0.394223574009)*x[1]
        arg[0,1,2,1]=(-0.550560496714)*x[0]**o+(0.216941875936)*x[0]+(-0.634745794924)*x[1]**o+(0.887912543217)*x[1]
        arg[0,2,0,0]=(-0.961535901493)*x[0]**o+(0.980336384835)*x[0]+(-0.151755000128)*x[1]**o+(0.0402841925797)*x[1]
        arg[0,2,0,1]=(0.571527088727)*x[0]**o+(-0.475013184811)*x[0]+(0.80051818192)*x[1]**o+(-0.589499747694)*x[1]
        arg[0,2,1,0]=(0.545840648612)*x[0]**o+(0.888313209612)*x[0]+(-0.568938733446)*x[1]**o+(-0.289030374342)*x[1]
        arg[0,2,1,1]=(-0.0631638607979)*x[0]**o+(-0.173754439841)*x[0]+(0.143503554071)*x[1]**o+(-0.0850087835084)*x[1]
        arg[0,2,2,0]=(0.316299661412)*x[0]**o+(0.260890197379)*x[0]+(-0.803519867776)*x[1]**o+(0.36434810185)*x[1]
        arg[0,2,2,1]=(-0.656013716479)*x[0]**o+(0.810499942514)*x[0]+(0.783455480815)*x[1]**o+(-0.943085466999)*x[1]
        arg[0,3,0,0]=(0.128153439722)*x[0]**o+(-0.404740042302)*x[0]+(0.785717360548)*x[1]**o+(-0.710818900841)*x[1]
        arg[0,3,0,1]=(-0.097059274787)*x[0]**o+(0.578757823228)*x[0]+(0.186146155178)*x[1]**o+(0.274602030707)*x[1]
        arg[0,3,1,0]=(0.00522392753954)*x[0]**o+(0.015773606982)*x[0]+(-0.327110065344)*x[1]**o+(0.15571120542)*x[1]
        arg[0,3,1,1]=(-0.14009116305)*x[0]**o+(-0.659975337811)*x[0]+(-0.297582833489)*x[1]**o+(-0.600935015071)*x[1]
        arg[0,3,2,0]=(0.749388604357)*x[0]**o+(0.762968650778)*x[0]+(-0.623477518339)*x[1]**o+(-0.931762943135)*x[1]
        arg[0,3,2,1]=(0.748396100632)*x[0]**o+(-0.841314360401)*x[0]+(0.432649446216)*x[1]**o+(-0.971300843628)*x[1]
        arg[0,4,0,0]=(-0.217249090357)*x[0]**o+(-0.833633675601)*x[0]+(0.169984305627)*x[1]**o+(-0.831738272666)*x[1]
        arg[0,4,0,1]=(0.501571101669)*x[0]**o+(-0.223897334788)*x[0]+(0.24178181976)*x[1]**o+(0.307921789017)*x[1]
        arg[0,4,1,0]=(-0.162064610564)*x[0]**o+(0.217669674603)*x[0]+(0.995517866562)*x[1]**o+(0.903061734589)*x[1]
        arg[0,4,1,1]=(-0.837150436928)*x[0]**o+(0.465269468525)*x[0]+(0.522034182082)*x[1]**o+(-0.725469921625)*x[1]
        arg[0,4,2,0]=(-0.295739906169)*x[0]**o+(0.961959220618)*x[0]+(0.577799918942)*x[1]**o+(-0.313255108148)*x[1]
        arg[0,4,2,1]=(-0.455635927062)*x[0]**o+(0.667038701641)*x[0]+(-0.151183772961)*x[1]**o+(-0.550197828316)*x[1]
        arg[1,0,0,0]=(0.585510684472)*x[0]**o+(-0.64754219493)*x[0]+(0.781495111863)*x[1]**o+(-0.792246214695)*x[1]
        arg[1,0,0,1]=(0.586346471397)*x[0]**o+(0.291794158138)*x[0]+(-0.614723564286)*x[1]**o+(-0.151451723805)*x[1]
        arg[1,0,1,0]=(-0.472877220222)*x[0]**o+(-0.0118252391089)*x[0]+(-0.596943135354)*x[1]**o+(-0.413992986066)*x[1]
        arg[1,0,1,1]=(-0.715007718152)*x[0]**o+(0.032964784403)*x[0]+(-0.350680745946)*x[1]**o+(0.678841285054)*x[1]
        arg[1,0,2,0]=(0.673910318085)*x[0]**o+(-0.15303471314)*x[0]+(-0.0889279944344)*x[1]**o+(-0.0356474498002)*x[1]
        arg[1,0,2,1]=(-0.875691348779)*x[0]**o+(0.547300853671)*x[0]+(-0.362646169132)*x[1]**o+(-0.588521964671)*x[1]
        arg[1,1,0,0]=(-0.0988912551145)*x[0]**o+(-0.710944394432)*x[0]+(0.153875249337)*x[1]**o+(0.274427434926)*x[1]
        arg[1,1,0,1]=(-0.0854272185591)*x[0]**o+(-0.21840195791)*x[0]+(0.906982245924)*x[1]**o+(0.570990116211)*x[1]
        arg[1,1,1,0]=(-0.851259116331)*x[0]**o+(0.309829955487)*x[0]+(0.115725082696)*x[1]**o+(-0.91590292189)*x[1]
        arg[1,1,1,1]=(-0.637464462703)*x[0]**o+(-0.280971672162)*x[0]+(0.270985403603)*x[1]**o+(-0.534115311397)*x[1]
        arg[1,1,2,0]=(0.670179973397)*x[0]**o+(0.847483793904)*x[0]+(0.0681088634526)*x[1]**o+(0.412176359729)*x[1]
        arg[1,1,2,1]=(0.0378205172485)*x[0]**o+(0.974150495292)*x[0]+(-0.720821275299)*x[1]**o+(-0.552588296494)*x[1]
        arg[1,2,0,0]=(0.151367609975)*x[0]**o+(0.832981586349)*x[0]+(0.419488381843)*x[1]**o+(0.561052438283)*x[1]
        arg[1,2,0,1]=(0.267839010425)*x[0]**o+(0.47455974064)*x[0]+(0.498826652567)*x[1]**o+(-0.51906391205)*x[1]
        arg[1,2,1,0]=(-0.350105107527)*x[0]**o+(-0.489326689724)*x[0]+(-0.156970199109)*x[1]**o+(-0.175565230472)*x[1]
        arg[1,2,1,1]=(0.754379973459)*x[0]**o+(-0.70691373234)*x[0]+(-0.964497356137)*x[1]**o+(-0.694989442192)*x[1]
        arg[1,2,2,0]=(-0.491090361655)*x[0]**o+(0.154379441501)*x[0]+(0.403809695039)*x[1]**o+(-0.084183660554)*x[1]
        arg[1,2,2,1]=(-0.164538235223)*x[0]**o+(0.628058472821)*x[0]+(0.331956969519)*x[1]**o+(0.828755643202)*x[1]
        arg[1,3,0,0]=(0.103479416566)*x[0]**o+(0.380691196177)*x[0]+(-0.350375360146)*x[1]**o+(0.18528492785)*x[1]
        arg[1,3,0,1]=(-0.0744423760559)*x[0]**o+(0.912275380092)*x[0]+(0.867254399179)*x[1]**o+(0.410307184344)*x[1]
        arg[1,3,1,0]=(-0.480324052626)*x[0]**o+(-0.438544962062)*x[0]+(0.749575808465)*x[1]**o+(-0.0689669475685)*x[1]
        arg[1,3,1,1]=(0.723325499611)*x[0]**o+(-0.368668363075)*x[0]+(0.404302895836)*x[1]**o+(0.66913878408)*x[1]
        arg[1,3,2,0]=(-0.122189168877)*x[0]**o+(0.603362404944)*x[0]+(0.184163482697)*x[1]**o+(0.728136755361)*x[1]
        arg[1,3,2,1]=(-0.149430048386)*x[0]**o+(-0.478108072433)*x[0]+(0.342479834611)*x[1]**o+(0.596757508109)*x[1]
        arg[1,4,0,0]=(0.334694167334)*x[0]**o+(-0.815793032292)*x[0]+(0.394250112116)*x[1]**o+(0.269197166808)*x[1]
        arg[1,4,0,1]=(-0.19210954081)*x[0]**o+(0.416736181257)*x[0]+(0.146217029416)*x[1]**o+(0.906316717458)*x[1]
        arg[1,4,1,0]=(-0.708061018673)*x[0]**o+(0.299200707691)*x[0]+(0.665111947965)*x[1]**o+(-0.0340170473314)*x[1]
        arg[1,4,1,1]=(0.0353929744202)*x[0]**o+(0.0493194235067)*x[0]+(-0.12215424813)*x[1]**o+(-0.758677440837)*x[1]
        arg[1,4,2,0]=(0.769702957701)*x[0]**o+(0.901805654682)*x[0]+(0.260222241019)*x[1]**o+(-0.471997705392)*x[1]
        arg[1,4,2,1]=(0.204258068316)*x[0]**o+(-0.173085114573)*x[0]+(-0.0253164178046)*x[1]**o+(0.84254484259)*x[1]
        arg[2,0,0,0]=(-0.688170206644)*x[0]**o+(-0.718673702314)*x[0]+(0.834265379201)*x[1]**o+(-0.879925159193)*x[1]
        arg[2,0,0,1]=(-0.515401334666)*x[0]**o+(-0.810523180967)*x[0]+(0.725182741017)*x[1]**o+(-0.74319471007)*x[1]
        arg[2,0,1,0]=(-0.799698108602)*x[0]**o+(0.289222327549)*x[0]+(0.0589402409999)*x[1]**o+(-0.437669765839)*x[1]
        arg[2,0,1,1]=(-0.734933454418)*x[0]**o+(0.371391146956)*x[0]+(-0.17797293763)*x[1]**o+(-0.499412028554)*x[1]
        arg[2,0,2,0]=(0.217283159505)*x[0]**o+(-0.184871188378)*x[0]+(-0.764095670777)*x[1]**o+(0.103705507755)*x[1]
        arg[2,0,2,1]=(0.578071527416)*x[0]**o+(0.845365737459)*x[0]+(-0.173187973137)*x[1]**o+(-0.827819925403)*x[1]
        arg[2,1,0,0]=(-0.258505119864)*x[0]**o+(0.640873475074)*x[0]+(-0.993841847675)*x[1]**o+(0.347490565123)*x[1]
        arg[2,1,0,1]=(0.224893445547)*x[0]**o+(0.827208591218)*x[0]+(0.0391563892525)*x[1]**o+(-0.320939200596)*x[1]
        arg[2,1,1,0]=(-0.993044884903)*x[0]**o+(-0.395235414086)*x[0]+(-0.447460653655)*x[1]**o+(-0.114585622808)*x[1]
        arg[2,1,1,1]=(0.908711735866)*x[0]**o+(-0.911405008413)*x[0]+(-0.798601192961)*x[1]**o+(-0.432937813587)*x[1]
        arg[2,1,2,0]=(-0.270424920067)*x[0]**o+(-0.533648944143)*x[0]+(0.184307763968)*x[1]**o+(-0.802221601505)*x[1]
        arg[2,1,2,1]=(0.154432841861)*x[0]**o+(0.306458633794)*x[0]+(0.0343421658454)*x[1]**o+(-0.768559457414)*x[1]
        arg[2,2,0,0]=(-0.497828132445)*x[0]**o+(0.484920248315)*x[0]+(-0.188239341875)*x[1]**o+(0.100360622012)*x[1]
        arg[2,2,0,1]=(-0.0433518780446)*x[0]**o+(0.119470819101)*x[0]+(0.344025017428)*x[1]**o+(0.667394126898)*x[1]
        arg[2,2,1,0]=(0.512271795846)*x[0]**o+(0.567024910623)*x[0]+(-0.316191920076)*x[1]**o+(-0.202135136851)*x[1]
        arg[2,2,1,1]=(0.781678855797)*x[0]**o+(0.159208533718)*x[0]+(-0.930150837465)*x[1]**o+(-0.666903895504)*x[1]
        arg[2,2,2,0]=(0.292270965732)*x[0]**o+(-0.112877441048)*x[0]+(-0.643355582049)*x[1]**o+(-0.0482560996171)*x[1]
        arg[2,2,2,1]=(0.638688065872)*x[0]**o+(0.787719224138)*x[0]+(0.815312681592)*x[1]**o+(0.334735776221)*x[1]
        arg[2,3,0,0]=(0.409920519672)*x[0]**o+(-0.191886243537)*x[0]+(-0.348205383522)*x[1]**o+(-0.651065510647)*x[1]
        arg[2,3,0,1]=(0.824428500178)*x[0]**o+(-0.909533490456)*x[0]+(-0.455079541262)*x[1]**o+(-0.613994310127)*x[1]
        arg[2,3,1,0]=(0.225908995914)*x[0]**o+(0.258098741048)*x[0]+(-0.858905205327)*x[1]**o+(0.535845140874)*x[1]
        arg[2,3,1,1]=(-0.183600641837)*x[0]**o+(-0.0436432564818)*x[0]+(0.306524301109)*x[1]**o+(0.662513699164)*x[1]
        arg[2,3,2,0]=(-0.678973103555)*x[0]**o+(0.674110571991)*x[0]+(-0.129743880618)*x[1]**o+(-0.210759231178)*x[1]
        arg[2,3,2,1]=(-0.594989617102)*x[0]**o+(0.0557145839856)*x[0]+(-0.379516802094)*x[1]**o+(-0.496127170036)*x[1]
        arg[2,4,0,0]=(0.401088886175)*x[0]**o+(-0.208580896076)*x[0]+(-0.655582275521)*x[1]**o+(-0.522892532154)*x[1]
        arg[2,4,0,1]=(0.609111916382)*x[0]**o+(-0.864603847696)*x[0]+(-0.373482060442)*x[1]**o+(0.048777005392)*x[1]
        arg[2,4,1,0]=(-0.503935244799)*x[0]**o+(0.320517360825)*x[0]+(-0.658994963813)*x[1]**o+(0.00465823412972)*x[1]
        arg[2,4,1,1]=(-0.202767966863)*x[0]**o+(0.737239608437)*x[0]+(-0.633206522406)*x[1]**o+(-0.408531988781)*x[1]
        arg[2,4,2,0]=(0.660048575091)*x[0]**o+(-0.399137043056)*x[0]+(-0.0977181117895)*x[1]**o+(0.148903352776)*x[1]
        arg[2,4,2,1]=(-0.396768795712)*x[0]**o+(-0.550388681028)*x[0]+(0.0946256561554)*x[1]**o+(-0.0566941754267)*x[1]
        arg[3,0,0,0]=(0.0647180650687)*x[0]**o+(0.920788893299)*x[0]+(0.131306350033)*x[1]**o+(-0.840527983774)*x[1]
        arg[3,0,0,1]=(0.451700262969)*x[0]**o+(-0.16104676139)*x[0]+(0.114145999713)*x[1]**o+(-0.949265482931)*x[1]
        arg[3,0,1,0]=(-0.219444617083)*x[0]**o+(0.846470673388)*x[0]+(-0.341392225388)*x[1]**o+(-0.721912717627)*x[1]
        arg[3,0,1,1]=(-0.199502120929)*x[0]**o+(-0.00662410675554)*x[0]+(0.0529013037694)*x[1]**o+(-0.022278580228)*x[1]
        arg[3,0,2,0]=(0.803500732535)*x[0]**o+(0.754138333448)*x[0]+(0.822404687478)*x[1]**o+(0.900099584934)*x[1]
        arg[3,0,2,1]=(0.84674757393)*x[0]**o+(0.326823539079)*x[0]+(-0.824446392056)*x[1]**o+(-0.506583766552)*x[1]
        arg[3,1,0,0]=(0.816775849327)*x[0]**o+(0.957523430402)*x[0]+(-0.822535686534)*x[1]**o+(-0.583389123083)*x[1]
        arg[3,1,0,1]=(0.518120693337)*x[0]**o+(0.0946135115511)*x[0]+(0.577364260402)*x[1]**o+(-0.677435109716)*x[1]
        arg[3,1,1,0]=(-0.559041961891)*x[0]**o+(0.436791134026)*x[0]+(0.287609046653)*x[1]**o+(-0.484573207793)*x[1]
        arg[3,1,1,1]=(-0.0786041693053)*x[0]**o+(0.793526931222)*x[0]+(0.055780633982)*x[1]**o+(0.0447074319839)*x[1]
        arg[3,1,2,0]=(-0.190231018951)*x[0]**o+(-0.61318830169)*x[0]+(-0.157202158534)*x[1]**o+(-0.823043159126)*x[1]
        arg[3,1,2,1]=(-0.502637901786)*x[0]**o+(0.781083145697)*x[0]+(-0.663013942651)*x[1]**o+(-0.305726874525)*x[1]
        arg[3,2,0,0]=(0.566955929517)*x[0]**o+(-0.506981068575)*x[0]+(-0.831439289654)*x[1]**o+(-0.337491442765)*x[1]
        arg[3,2,0,1]=(-0.713648844422)*x[0]**o+(-0.00815764297609)*x[0]+(0.829706206618)*x[1]**o+(0.96471376965)*x[1]
        arg[3,2,1,0]=(-0.944705561715)*x[0]**o+(-0.87962603172)*x[0]+(-0.0525190846342)*x[1]**o+(-0.0457463400515)*x[1]
        arg[3,2,1,1]=(0.54922446571)*x[0]**o+(0.651923573438)*x[0]+(-0.00229398468376)*x[1]**o+(-0.00170863787749)*x[1]
        arg[3,2,2,0]=(0.622372079685)*x[0]**o+(-0.553578221845)*x[0]+(-0.737877450857)*x[1]**o+(0.459729854002)*x[1]
        arg[3,2,2,1]=(0.826155732508)*x[0]**o+(-0.000468898258991)*x[0]+(-0.255179342686)*x[1]**o+(-0.80957312045)*x[1]
        arg[3,3,0,0]=(-0.195212469839)*x[0]**o+(-0.0235216449547)*x[0]+(-0.175192639485)*x[1]**o+(-0.58937661213)*x[1]
        arg[3,3,0,1]=(-0.931538910404)*x[0]**o+(0.649282880007)*x[0]+(-0.237841482637)*x[1]**o+(-0.0220516220802)*x[1]
        arg[3,3,1,0]=(0.420911057166)*x[0]**o+(0.200126728184)*x[0]+(-0.829130861656)*x[1]**o+(-0.563462916661)*x[1]
        arg[3,3,1,1]=(-0.148199974575)*x[0]**o+(-0.248825389533)*x[0]+(-0.346424894584)*x[1]**o+(0.339758232978)*x[1]
        arg[3,3,2,0]=(0.249372496215)*x[0]**o+(-0.343626915746)*x[0]+(0.0961351845087)*x[1]**o+(0.612679249068)*x[1]
        arg[3,3,2,1]=(0.985380582259)*x[0]**o+(-0.603514318626)*x[0]+(-0.131105546364)*x[1]**o+(-0.0798158697238)*x[1]
        arg[3,4,0,0]=(0.478635365792)*x[0]**o+(-0.964424450881)*x[0]+(-0.469995535544)*x[1]**o+(-0.72871149515)*x[1]
        arg[3,4,0,1]=(-0.264802297415)*x[0]**o+(-0.926518282507)*x[0]+(0.803390182149)*x[1]**o+(0.387910762706)*x[1]
        arg[3,4,1,0]=(0.0452448579909)*x[0]**o+(-0.0482221515826)*x[0]+(-0.728020494982)*x[1]**o+(-0.288823633249)*x[1]
        arg[3,4,1,1]=(-0.640459595265)*x[0]**o+(0.759525453728)*x[0]+(-0.0280300707884)*x[1]**o+(0.079438187123)*x[1]
        arg[3,4,2,0]=(-0.14185171171)*x[0]**o+(-0.810989314052)*x[0]+(0.921902325773)*x[1]**o+(0.885920658784)*x[1]
        arg[3,4,2,1]=(-0.0101259901393)*x[0]**o+(-0.696541046542)*x[0]+(0.127932822496)*x[1]**o+(-0.822265796323)*x[1]
        ref[0,0,0,0]=(0.0577426785287)*x_ref[0]**o+(0.469845292021)*x_ref[0]+(-0.641706722819)*x_ref[1]**o+(0.129652001842)*x_ref[1]
        ref[0,0,0,1]=(-0.87686951076)*x_ref[0]**o+(-0.981093040999)*x_ref[0]+(0.654637319552)*x_ref[1]**o+(-0.953104888811)*x_ref[1]
        ref[0,0,1,0]=(-0.486386436415)*x_ref[0]**o+(0.91083470956)*x_ref[0]+(-0.392432937401)*x_ref[1]**o+(0.494919250884)*x_ref[1]
        ref[0,0,1,1]=(0.299928502076)*x_ref[0]**o+(-0.0787848191157)*x_ref[0]+(-0.541863592475)*x_ref[1]**o+(0.435934816687)*x_ref[1]
        ref[0,0,2,0]=(-0.0500083438365)*x_ref[0]**o+(-0.882581886623)*x_ref[0]+(-0.965347357063)*x_ref[1]**o+(-0.270254041774)*x_ref[1]
        ref[0,0,2,1]=(-0.18974861517)*x_ref[0]**o+(0.341406736671)*x_ref[0]+(0.847020774229)*x_ref[1]**o+(0.650201215849)*x_ref[1]
        ref[0,1,0,0]=(0.0823017084719)*x_ref[0]**o+(0.964585359235)*x_ref[0]+(0.357658660436)*x_ref[1]**o+(0.182585271812)*x_ref[1]
        ref[0,1,0,1]=(0.863411821905)*x_ref[0]**o+(0.313374566426)*x_ref[0]+(-0.273547644827)*x_ref[1]**o+(0.429051854479)*x_ref[1]
        ref[0,1,1,0]=(0.167608645068)*x_ref[0]**o+(0.926908094967)*x_ref[0]+(-0.853415240465)*x_ref[1]**o+(-0.376634525241)*x_ref[1]
        ref[0,1,1,1]=(-0.555491961591)*x_ref[0]**o+(-0.64621137308)*x_ref[0]+(-0.315877124991)*x_ref[1]**o+(-0.962839286133)*x_ref[1]
        ref[0,1,2,0]=(0.909012065667)*x_ref[0]**o+(-0.656739022976)*x_ref[0]+(-0.458377773547)*x_ref[1]**o+(0.394223574009)*x_ref[1]
        ref[0,1,2,1]=(-0.550560496714)*x_ref[0]**o+(0.216941875936)*x_ref[0]+(-0.634745794924)*x_ref[1]**o+(0.887912543217)*x_ref[1]
        ref[0,2,0,0]=(-0.961535901493)*x_ref[0]**o+(0.980336384835)*x_ref[0]+(-0.151755000128)*x_ref[1]**o+(0.0402841925797)*x_ref[1]
        ref[0,2,0,1]=(0.571527088727)*x_ref[0]**o+(-0.475013184811)*x_ref[0]+(0.80051818192)*x_ref[1]**o+(-0.589499747694)*x_ref[1]
        ref[0,2,1,0]=(0.545840648612)*x_ref[0]**o+(0.888313209612)*x_ref[0]+(-0.568938733446)*x_ref[1]**o+(-0.289030374342)*x_ref[1]
        ref[0,2,1,1]=(-0.0631638607979)*x_ref[0]**o+(-0.173754439841)*x_ref[0]+(0.143503554071)*x_ref[1]**o+(-0.0850087835084)*x_ref[1]
        ref[0,2,2,0]=(0.316299661412)*x_ref[0]**o+(0.260890197379)*x_ref[0]+(-0.803519867776)*x_ref[1]**o+(0.36434810185)*x_ref[1]
        ref[0,2,2,1]=(-0.656013716479)*x_ref[0]**o+(0.810499942514)*x_ref[0]+(0.783455480815)*x_ref[1]**o+(-0.943085466999)*x_ref[1]
        ref[0,3,0,0]=(0.128153439722)*x_ref[0]**o+(-0.404740042302)*x_ref[0]+(0.785717360548)*x_ref[1]**o+(-0.710818900841)*x_ref[1]
        ref[0,3,0,1]=(-0.097059274787)*x_ref[0]**o+(0.578757823228)*x_ref[0]+(0.186146155178)*x_ref[1]**o+(0.274602030707)*x_ref[1]
        ref[0,3,1,0]=(0.00522392753954)*x_ref[0]**o+(0.015773606982)*x_ref[0]+(-0.327110065344)*x_ref[1]**o+(0.15571120542)*x_ref[1]
        ref[0,3,1,1]=(-0.14009116305)*x_ref[0]**o+(-0.659975337811)*x_ref[0]+(-0.297582833489)*x_ref[1]**o+(-0.600935015071)*x_ref[1]
        ref[0,3,2,0]=(0.749388604357)*x_ref[0]**o+(0.762968650778)*x_ref[0]+(-0.623477518339)*x_ref[1]**o+(-0.931762943135)*x_ref[1]
        ref[0,3,2,1]=(0.748396100632)*x_ref[0]**o+(-0.841314360401)*x_ref[0]+(0.432649446216)*x_ref[1]**o+(-0.971300843628)*x_ref[1]
        ref[0,4,0,0]=(-0.217249090357)*x_ref[0]**o+(-0.833633675601)*x_ref[0]+(0.169984305627)*x_ref[1]**o+(-0.831738272666)*x_ref[1]
        ref[0,4,0,1]=(0.501571101669)*x_ref[0]**o+(-0.223897334788)*x_ref[0]+(0.24178181976)*x_ref[1]**o+(0.307921789017)*x_ref[1]
        ref[0,4,1,0]=(-0.162064610564)*x_ref[0]**o+(0.217669674603)*x_ref[0]+(0.995517866562)*x_ref[1]**o+(0.903061734589)*x_ref[1]
        ref[0,4,1,1]=(-0.837150436928)*x_ref[0]**o+(0.465269468525)*x_ref[0]+(0.522034182082)*x_ref[1]**o+(-0.725469921625)*x_ref[1]
        ref[0,4,2,0]=(-0.295739906169)*x_ref[0]**o+(0.961959220618)*x_ref[0]+(0.577799918942)*x_ref[1]**o+(-0.313255108148)*x_ref[1]
        ref[0,4,2,1]=(-0.455635927062)*x_ref[0]**o+(0.667038701641)*x_ref[0]+(-0.151183772961)*x_ref[1]**o+(-0.550197828316)*x_ref[1]
        ref[1,0,0,0]=(0.585510684472)*x_ref[0]**o+(-0.64754219493)*x_ref[0]+(0.781495111863)*x_ref[1]**o+(-0.792246214695)*x_ref[1]
        ref[1,0,0,1]=(0.586346471397)*x_ref[0]**o+(0.291794158138)*x_ref[0]+(-0.614723564286)*x_ref[1]**o+(-0.151451723805)*x_ref[1]
        ref[1,0,1,0]=(-0.472877220222)*x_ref[0]**o+(-0.0118252391089)*x_ref[0]+(-0.596943135354)*x_ref[1]**o+(-0.413992986066)*x_ref[1]
        ref[1,0,1,1]=(-0.715007718152)*x_ref[0]**o+(0.032964784403)*x_ref[0]+(-0.350680745946)*x_ref[1]**o+(0.678841285054)*x_ref[1]
        ref[1,0,2,0]=(0.673910318085)*x_ref[0]**o+(-0.15303471314)*x_ref[0]+(-0.0889279944344)*x_ref[1]**o+(-0.0356474498002)*x_ref[1]
        ref[1,0,2,1]=(-0.875691348779)*x_ref[0]**o+(0.547300853671)*x_ref[0]+(-0.362646169132)*x_ref[1]**o+(-0.588521964671)*x_ref[1]
        ref[1,1,0,0]=(-0.0988912551145)*x_ref[0]**o+(-0.710944394432)*x_ref[0]+(0.153875249337)*x_ref[1]**o+(0.274427434926)*x_ref[1]
        ref[1,1,0,1]=(-0.0854272185591)*x_ref[0]**o+(-0.21840195791)*x_ref[0]+(0.906982245924)*x_ref[1]**o+(0.570990116211)*x_ref[1]
        ref[1,1,1,0]=(-0.851259116331)*x_ref[0]**o+(0.309829955487)*x_ref[0]+(0.115725082696)*x_ref[1]**o+(-0.91590292189)*x_ref[1]
        ref[1,1,1,1]=(-0.637464462703)*x_ref[0]**o+(-0.280971672162)*x_ref[0]+(0.270985403603)*x_ref[1]**o+(-0.534115311397)*x_ref[1]
        ref[1,1,2,0]=(0.670179973397)*x_ref[0]**o+(0.847483793904)*x_ref[0]+(0.0681088634526)*x_ref[1]**o+(0.412176359729)*x_ref[1]
        ref[1,1,2,1]=(0.0378205172485)*x_ref[0]**o+(0.974150495292)*x_ref[0]+(-0.720821275299)*x_ref[1]**o+(-0.552588296494)*x_ref[1]
        ref[1,2,0,0]=(0.151367609975)*x_ref[0]**o+(0.832981586349)*x_ref[0]+(0.419488381843)*x_ref[1]**o+(0.561052438283)*x_ref[1]
        ref[1,2,0,1]=(0.267839010425)*x_ref[0]**o+(0.47455974064)*x_ref[0]+(0.498826652567)*x_ref[1]**o+(-0.51906391205)*x_ref[1]
        ref[1,2,1,0]=(-0.350105107527)*x_ref[0]**o+(-0.489326689724)*x_ref[0]+(-0.156970199109)*x_ref[1]**o+(-0.175565230472)*x_ref[1]
        ref[1,2,1,1]=(0.754379973459)*x_ref[0]**o+(-0.70691373234)*x_ref[0]+(-0.964497356137)*x_ref[1]**o+(-0.694989442192)*x_ref[1]
        ref[1,2,2,0]=(-0.491090361655)*x_ref[0]**o+(0.154379441501)*x_ref[0]+(0.403809695039)*x_ref[1]**o+(-0.084183660554)*x_ref[1]
        ref[1,2,2,1]=(-0.164538235223)*x_ref[0]**o+(0.628058472821)*x_ref[0]+(0.331956969519)*x_ref[1]**o+(0.828755643202)*x_ref[1]
        ref[1,3,0,0]=(0.103479416566)*x_ref[0]**o+(0.380691196177)*x_ref[0]+(-0.350375360146)*x_ref[1]**o+(0.18528492785)*x_ref[1]
        ref[1,3,0,1]=(-0.0744423760559)*x_ref[0]**o+(0.912275380092)*x_ref[0]+(0.867254399179)*x_ref[1]**o+(0.410307184344)*x_ref[1]
        ref[1,3,1,0]=(-0.480324052626)*x_ref[0]**o+(-0.438544962062)*x_ref[0]+(0.749575808465)*x_ref[1]**o+(-0.0689669475685)*x_ref[1]
        ref[1,3,1,1]=(0.723325499611)*x_ref[0]**o+(-0.368668363075)*x_ref[0]+(0.404302895836)*x_ref[1]**o+(0.66913878408)*x_ref[1]
        ref[1,3,2,0]=(-0.122189168877)*x_ref[0]**o+(0.603362404944)*x_ref[0]+(0.184163482697)*x_ref[1]**o+(0.728136755361)*x_ref[1]
        ref[1,3,2,1]=(-0.149430048386)*x_ref[0]**o+(-0.478108072433)*x_ref[0]+(0.342479834611)*x_ref[1]**o+(0.596757508109)*x_ref[1]
        ref[1,4,0,0]=(0.334694167334)*x_ref[0]**o+(-0.815793032292)*x_ref[0]+(0.394250112116)*x_ref[1]**o+(0.269197166808)*x_ref[1]
        ref[1,4,0,1]=(-0.19210954081)*x_ref[0]**o+(0.416736181257)*x_ref[0]+(0.146217029416)*x_ref[1]**o+(0.906316717458)*x_ref[1]
        ref[1,4,1,0]=(-0.708061018673)*x_ref[0]**o+(0.299200707691)*x_ref[0]+(0.665111947965)*x_ref[1]**o+(-0.0340170473314)*x_ref[1]
        ref[1,4,1,1]=(0.0353929744202)*x_ref[0]**o+(0.0493194235067)*x_ref[0]+(-0.12215424813)*x_ref[1]**o+(-0.758677440837)*x_ref[1]
        ref[1,4,2,0]=(0.769702957701)*x_ref[0]**o+(0.901805654682)*x_ref[0]+(0.260222241019)*x_ref[1]**o+(-0.471997705392)*x_ref[1]
        ref[1,4,2,1]=(0.204258068316)*x_ref[0]**o+(-0.173085114573)*x_ref[0]+(-0.0253164178046)*x_ref[1]**o+(0.84254484259)*x_ref[1]
        ref[2,0,0,0]=(-0.688170206644)*x_ref[0]**o+(-0.718673702314)*x_ref[0]+(0.834265379201)*x_ref[1]**o+(-0.879925159193)*x_ref[1]
        ref[2,0,0,1]=(-0.515401334666)*x_ref[0]**o+(-0.810523180967)*x_ref[0]+(0.725182741017)*x_ref[1]**o+(-0.74319471007)*x_ref[1]
        ref[2,0,1,0]=(-0.799698108602)*x_ref[0]**o+(0.289222327549)*x_ref[0]+(0.0589402409999)*x_ref[1]**o+(-0.437669765839)*x_ref[1]
        ref[2,0,1,1]=(-0.734933454418)*x_ref[0]**o+(0.371391146956)*x_ref[0]+(-0.17797293763)*x_ref[1]**o+(-0.499412028554)*x_ref[1]
        ref[2,0,2,0]=(0.217283159505)*x_ref[0]**o+(-0.184871188378)*x_ref[0]+(-0.764095670777)*x_ref[1]**o+(0.103705507755)*x_ref[1]
        ref[2,0,2,1]=(0.578071527416)*x_ref[0]**o+(0.845365737459)*x_ref[0]+(-0.173187973137)*x_ref[1]**o+(-0.827819925403)*x_ref[1]
        ref[2,1,0,0]=(-0.258505119864)*x_ref[0]**o+(0.640873475074)*x_ref[0]+(-0.993841847675)*x_ref[1]**o+(0.347490565123)*x_ref[1]
        ref[2,1,0,1]=(0.224893445547)*x_ref[0]**o+(0.827208591218)*x_ref[0]+(0.0391563892525)*x_ref[1]**o+(-0.320939200596)*x_ref[1]
        ref[2,1,1,0]=(-0.993044884903)*x_ref[0]**o+(-0.395235414086)*x_ref[0]+(-0.447460653655)*x_ref[1]**o+(-0.114585622808)*x_ref[1]
        ref[2,1,1,1]=(0.908711735866)*x_ref[0]**o+(-0.911405008413)*x_ref[0]+(-0.798601192961)*x_ref[1]**o+(-0.432937813587)*x_ref[1]
        ref[2,1,2,0]=(-0.270424920067)*x_ref[0]**o+(-0.533648944143)*x_ref[0]+(0.184307763968)*x_ref[1]**o+(-0.802221601505)*x_ref[1]
        ref[2,1,2,1]=(0.154432841861)*x_ref[0]**o+(0.306458633794)*x_ref[0]+(0.0343421658454)*x_ref[1]**o+(-0.768559457414)*x_ref[1]
        ref[2,2,0,0]=(-0.497828132445)*x_ref[0]**o+(0.484920248315)*x_ref[0]+(-0.188239341875)*x_ref[1]**o+(0.100360622012)*x_ref[1]
        ref[2,2,0,1]=(-0.0433518780446)*x_ref[0]**o+(0.119470819101)*x_ref[0]+(0.344025017428)*x_ref[1]**o+(0.667394126898)*x_ref[1]
        ref[2,2,1,0]=(0.512271795846)*x_ref[0]**o+(0.567024910623)*x_ref[0]+(-0.316191920076)*x_ref[1]**o+(-0.202135136851)*x_ref[1]
        ref[2,2,1,1]=(0.781678855797)*x_ref[0]**o+(0.159208533718)*x_ref[0]+(-0.930150837465)*x_ref[1]**o+(-0.666903895504)*x_ref[1]
        ref[2,2,2,0]=(0.292270965732)*x_ref[0]**o+(-0.112877441048)*x_ref[0]+(-0.643355582049)*x_ref[1]**o+(-0.0482560996171)*x_ref[1]
        ref[2,2,2,1]=(0.638688065872)*x_ref[0]**o+(0.787719224138)*x_ref[0]+(0.815312681592)*x_ref[1]**o+(0.334735776221)*x_ref[1]
        ref[2,3,0,0]=(0.409920519672)*x_ref[0]**o+(-0.191886243537)*x_ref[0]+(-0.348205383522)*x_ref[1]**o+(-0.651065510647)*x_ref[1]
        ref[2,3,0,1]=(0.824428500178)*x_ref[0]**o+(-0.909533490456)*x_ref[0]+(-0.455079541262)*x_ref[1]**o+(-0.613994310127)*x_ref[1]
        ref[2,3,1,0]=(0.225908995914)*x_ref[0]**o+(0.258098741048)*x_ref[0]+(-0.858905205327)*x_ref[1]**o+(0.535845140874)*x_ref[1]
        ref[2,3,1,1]=(-0.183600641837)*x_ref[0]**o+(-0.0436432564818)*x_ref[0]+(0.306524301109)*x_ref[1]**o+(0.662513699164)*x_ref[1]
        ref[2,3,2,0]=(-0.678973103555)*x_ref[0]**o+(0.674110571991)*x_ref[0]+(-0.129743880618)*x_ref[1]**o+(-0.210759231178)*x_ref[1]
        ref[2,3,2,1]=(-0.594989617102)*x_ref[0]**o+(0.0557145839856)*x_ref[0]+(-0.379516802094)*x_ref[1]**o+(-0.496127170036)*x_ref[1]
        ref[2,4,0,0]=(0.401088886175)*x_ref[0]**o+(-0.208580896076)*x_ref[0]+(-0.655582275521)*x_ref[1]**o+(-0.522892532154)*x_ref[1]
        ref[2,4,0,1]=(0.609111916382)*x_ref[0]**o+(-0.864603847696)*x_ref[0]+(-0.373482060442)*x_ref[1]**o+(0.048777005392)*x_ref[1]
        ref[2,4,1,0]=(-0.503935244799)*x_ref[0]**o+(0.320517360825)*x_ref[0]+(-0.658994963813)*x_ref[1]**o+(0.00465823412972)*x_ref[1]
        ref[2,4,1,1]=(-0.202767966863)*x_ref[0]**o+(0.737239608437)*x_ref[0]+(-0.633206522406)*x_ref[1]**o+(-0.408531988781)*x_ref[1]
        ref[2,4,2,0]=(0.660048575091)*x_ref[0]**o+(-0.399137043056)*x_ref[0]+(-0.0977181117895)*x_ref[1]**o+(0.148903352776)*x_ref[1]
        ref[2,4,2,1]=(-0.396768795712)*x_ref[0]**o+(-0.550388681028)*x_ref[0]+(0.0946256561554)*x_ref[1]**o+(-0.0566941754267)*x_ref[1]
        ref[3,0,0,0]=(0.0647180650687)*x_ref[0]**o+(0.920788893299)*x_ref[0]+(0.131306350033)*x_ref[1]**o+(-0.840527983774)*x_ref[1]
        ref[3,0,0,1]=(0.451700262969)*x_ref[0]**o+(-0.16104676139)*x_ref[0]+(0.114145999713)*x_ref[1]**o+(-0.949265482931)*x_ref[1]
        ref[3,0,1,0]=(-0.219444617083)*x_ref[0]**o+(0.846470673388)*x_ref[0]+(-0.341392225388)*x_ref[1]**o+(-0.721912717627)*x_ref[1]
        ref[3,0,1,1]=(-0.199502120929)*x_ref[0]**o+(-0.00662410675554)*x_ref[0]+(0.0529013037694)*x_ref[1]**o+(-0.022278580228)*x_ref[1]
        ref[3,0,2,0]=(0.803500732535)*x_ref[0]**o+(0.754138333448)*x_ref[0]+(0.822404687478)*x_ref[1]**o+(0.900099584934)*x_ref[1]
        ref[3,0,2,1]=(0.84674757393)*x_ref[0]**o+(0.326823539079)*x_ref[0]+(-0.824446392056)*x_ref[1]**o+(-0.506583766552)*x_ref[1]
        ref[3,1,0,0]=(0.816775849327)*x_ref[0]**o+(0.957523430402)*x_ref[0]+(-0.822535686534)*x_ref[1]**o+(-0.583389123083)*x_ref[1]
        ref[3,1,0,1]=(0.518120693337)*x_ref[0]**o+(0.0946135115511)*x_ref[0]+(0.577364260402)*x_ref[1]**o+(-0.677435109716)*x_ref[1]
        ref[3,1,1,0]=(-0.559041961891)*x_ref[0]**o+(0.436791134026)*x_ref[0]+(0.287609046653)*x_ref[1]**o+(-0.484573207793)*x_ref[1]
        ref[3,1,1,1]=(-0.0786041693053)*x_ref[0]**o+(0.793526931222)*x_ref[0]+(0.055780633982)*x_ref[1]**o+(0.0447074319839)*x_ref[1]
        ref[3,1,2,0]=(-0.190231018951)*x_ref[0]**o+(-0.61318830169)*x_ref[0]+(-0.157202158534)*x_ref[1]**o+(-0.823043159126)*x_ref[1]
        ref[3,1,2,1]=(-0.502637901786)*x_ref[0]**o+(0.781083145697)*x_ref[0]+(-0.663013942651)*x_ref[1]**o+(-0.305726874525)*x_ref[1]
        ref[3,2,0,0]=(0.566955929517)*x_ref[0]**o+(-0.506981068575)*x_ref[0]+(-0.831439289654)*x_ref[1]**o+(-0.337491442765)*x_ref[1]
        ref[3,2,0,1]=(-0.713648844422)*x_ref[0]**o+(-0.00815764297609)*x_ref[0]+(0.829706206618)*x_ref[1]**o+(0.96471376965)*x_ref[1]
        ref[3,2,1,0]=(-0.944705561715)*x_ref[0]**o+(-0.87962603172)*x_ref[0]+(-0.0525190846342)*x_ref[1]**o+(-0.0457463400515)*x_ref[1]
        ref[3,2,1,1]=(0.54922446571)*x_ref[0]**o+(0.651923573438)*x_ref[0]+(-0.00229398468376)*x_ref[1]**o+(-0.00170863787749)*x_ref[1]
        ref[3,2,2,0]=(0.622372079685)*x_ref[0]**o+(-0.553578221845)*x_ref[0]+(-0.737877450857)*x_ref[1]**o+(0.459729854002)*x_ref[1]
        ref[3,2,2,1]=(0.826155732508)*x_ref[0]**o+(-0.000468898258991)*x_ref[0]+(-0.255179342686)*x_ref[1]**o+(-0.80957312045)*x_ref[1]
        ref[3,3,0,0]=(-0.195212469839)*x_ref[0]**o+(-0.0235216449547)*x_ref[0]+(-0.175192639485)*x_ref[1]**o+(-0.58937661213)*x_ref[1]
        ref[3,3,0,1]=(-0.931538910404)*x_ref[0]**o+(0.649282880007)*x_ref[0]+(-0.237841482637)*x_ref[1]**o+(-0.0220516220802)*x_ref[1]
        ref[3,3,1,0]=(0.420911057166)*x_ref[0]**o+(0.200126728184)*x_ref[0]+(-0.829130861656)*x_ref[1]**o+(-0.563462916661)*x_ref[1]
        ref[3,3,1,1]=(-0.148199974575)*x_ref[0]**o+(-0.248825389533)*x_ref[0]+(-0.346424894584)*x_ref[1]**o+(0.339758232978)*x_ref[1]
        ref[3,3,2,0]=(0.249372496215)*x_ref[0]**o+(-0.343626915746)*x_ref[0]+(0.0961351845087)*x_ref[1]**o+(0.612679249068)*x_ref[1]
        ref[3,3,2,1]=(0.985380582259)*x_ref[0]**o+(-0.603514318626)*x_ref[0]+(-0.131105546364)*x_ref[1]**o+(-0.0798158697238)*x_ref[1]
        ref[3,4,0,0]=(0.478635365792)*x_ref[0]**o+(-0.964424450881)*x_ref[0]+(-0.469995535544)*x_ref[1]**o+(-0.72871149515)*x_ref[1]
        ref[3,4,0,1]=(-0.264802297415)*x_ref[0]**o+(-0.926518282507)*x_ref[0]+(0.803390182149)*x_ref[1]**o+(0.387910762706)*x_ref[1]
        ref[3,4,1,0]=(0.0452448579909)*x_ref[0]**o+(-0.0482221515826)*x_ref[0]+(-0.728020494982)*x_ref[1]**o+(-0.288823633249)*x_ref[1]
        ref[3,4,1,1]=(-0.640459595265)*x_ref[0]**o+(0.759525453728)*x_ref[0]+(-0.0280300707884)*x_ref[1]**o+(0.079438187123)*x_ref[1]
        ref[3,4,2,0]=(-0.14185171171)*x_ref[0]**o+(-0.810989314052)*x_ref[0]+(0.921902325773)*x_ref[1]**o+(0.885920658784)*x_ref[1]
        ref[3,4,2,1]=(-0.0101259901393)*x_ref[0]**o+(-0.696541046542)*x_ref[0]+(0.127932822496)*x_ref[1]**o+(-0.822265796323)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.657703355307)*x[0]**o+(0.397813831843)*x[0]+(0.364460538717)*x[1]**o+(-0.845626669617)*x[1]+(-0.987144254126)*x[2]**o+(-0.732384980726)*x[2]
        arg[0,0,0,1]=(-0.13280304886)*x[0]**o+(-0.953796807758)*x[0]+(0.875723746654)*x[1]**o+(0.0764745117242)*x[1]+(-0.536085676221)*x[2]**o+(-0.520611959678)*x[2]
        arg[0,0,1,0]=(0.716429446578)*x[0]**o+(-0.729164693441)*x[0]+(-0.832505557428)*x[1]**o+(-0.329200605611)*x[1]+(0.0310162060964)*x[2]**o+(-0.869978941999)*x[2]
        arg[0,0,1,1]=(0.110253530584)*x[0]**o+(0.995528486822)*x[0]+(0.838577797395)*x[1]**o+(0.443987752225)*x[1]+(0.387337959057)*x[2]**o+(0.548448752612)*x[2]
        arg[0,0,2,0]=(-0.995737515431)*x[0]**o+(-0.899807241717)*x[0]+(-0.0803287764138)*x[1]**o+(0.242680051556)*x[1]+(0.58974836823)*x[2]**o+(-0.498450188944)*x[2]
        arg[0,0,2,1]=(0.341561541555)*x[0]**o+(-0.00541988184597)*x[0]+(-0.204832898701)*x[1]**o+(-0.957247935865)*x[1]+(-0.441002902064)*x[2]**o+(0.756843480304)*x[2]
        arg[0,1,0,0]=(0.136201526244)*x[0]**o+(0.517677304193)*x[0]+(0.523031610748)*x[1]**o+(-0.908998053383)*x[1]+(-0.629649678416)*x[2]**o+(0.762977917512)*x[2]
        arg[0,1,0,1]=(-0.103521080864)*x[0]**o+(-0.434410222595)*x[0]+(0.418239530241)*x[1]**o+(0.65638475119)*x[1]+(0.413036393129)*x[2]**o+(-0.220365518353)*x[2]
        arg[0,1,1,0]=(-0.415690201483)*x[0]**o+(-0.969323394689)*x[0]+(-0.295639420195)*x[1]**o+(-0.24915002749)*x[1]+(-0.00927627101467)*x[2]**o+(0.520335261634)*x[2]
        arg[0,1,1,1]=(-0.262653741797)*x[0]**o+(-0.463590591666)*x[0]+(0.568893042923)*x[1]**o+(-0.869851332898)*x[1]+(-0.463261399415)*x[2]**o+(0.0281915686202)*x[2]
        arg[0,1,2,0]=(0.377159503058)*x[0]**o+(0.484512261943)*x[0]+(0.0303594628925)*x[1]**o+(0.482051243004)*x[1]+(0.240643013949)*x[2]**o+(-0.943908617705)*x[2]
        arg[0,1,2,1]=(0.843854614097)*x[0]**o+(0.018041534553)*x[0]+(0.264198703361)*x[1]**o+(0.732889517628)*x[1]+(-0.154807384394)*x[2]**o+(0.60513622187)*x[2]
        arg[0,2,0,0]=(-0.234758118286)*x[0]**o+(-0.77807396119)*x[0]+(-0.97783091103)*x[1]**o+(-0.716185223828)*x[1]+(-0.573126474259)*x[2]**o+(0.403110701555)*x[2]
        arg[0,2,0,1]=(-0.853878344804)*x[0]**o+(-0.482644826561)*x[0]+(-0.826041149012)*x[1]**o+(0.596376234798)*x[1]+(-0.216400517274)*x[2]**o+(-0.323127194259)*x[2]
        arg[0,2,1,0]=(0.931551121492)*x[0]**o+(-0.527132174379)*x[0]+(0.373496704463)*x[1]**o+(-0.375215826463)*x[1]+(-0.639167297984)*x[2]**o+(0.143608733189)*x[2]
        arg[0,2,1,1]=(-0.982206000879)*x[0]**o+(0.836589596534)*x[0]+(0.606275000138)*x[1]**o+(-0.821469453892)*x[1]+(-0.633828380948)*x[2]**o+(-0.637688777216)*x[2]
        arg[0,2,2,0]=(0.441649557529)*x[0]**o+(-0.34978365051)*x[0]+(0.514175365405)*x[1]**o+(-0.903212271234)*x[1]+(0.70236887156)*x[2]**o+(0.72953045189)*x[2]
        arg[0,2,2,1]=(0.302787896978)*x[0]**o+(0.157598071021)*x[0]+(-0.357101947632)*x[1]**o+(0.381098383156)*x[1]+(-0.552546036689)*x[2]**o+(0.448181721984)*x[2]
        arg[0,3,0,0]=(0.662181077399)*x[0]**o+(-0.283755239518)*x[0]+(0.33694741278)*x[1]**o+(0.275801685868)*x[1]+(0.526155259101)*x[2]**o+(-0.378241855559)*x[2]
        arg[0,3,0,1]=(0.157822133581)*x[0]**o+(0.25402072747)*x[0]+(0.522842123099)*x[1]**o+(0.510595773584)*x[1]+(-0.829239126666)*x[2]**o+(0.990879342044)*x[2]
        arg[0,3,1,0]=(-0.775925280059)*x[0]**o+(-0.726971388023)*x[0]+(-0.630088730753)*x[1]**o+(-0.731024178237)*x[1]+(-0.288643541196)*x[2]**o+(-0.283708428793)*x[2]
        arg[0,3,1,1]=(-0.883176640445)*x[0]**o+(-0.0810211976762)*x[0]+(-0.102919217784)*x[1]**o+(0.271384866385)*x[1]+(0.980796900477)*x[2]**o+(0.439294038672)*x[2]
        arg[0,3,2,0]=(-0.826688221502)*x[0]**o+(-0.909388347347)*x[0]+(0.164406927722)*x[1]**o+(0.60157189364)*x[1]+(-0.885578176748)*x[2]**o+(0.610008674247)*x[2]
        arg[0,3,2,1]=(0.130154197193)*x[0]**o+(-0.754703602975)*x[0]+(0.00463710497017)*x[1]**o+(0.502162916123)*x[1]+(-0.749124548541)*x[2]**o+(0.946592962092)*x[2]
        arg[0,4,0,0]=(-0.0478347448906)*x[0]**o+(-0.812937871882)*x[0]+(-0.460982012551)*x[1]**o+(0.317485998991)*x[1]+(-0.593671683477)*x[2]**o+(0.551385718668)*x[2]
        arg[0,4,0,1]=(0.281998767502)*x[0]**o+(-0.0613830311997)*x[0]+(0.981429151019)*x[1]**o+(-0.580567714008)*x[1]+(0.0369483005331)*x[2]**o+(-0.74132358588)*x[2]
        arg[0,4,1,0]=(0.525845217483)*x[0]**o+(0.907803915216)*x[0]+(0.111429079263)*x[1]**o+(-0.653030401377)*x[1]+(0.0814793913696)*x[2]**o+(-0.419201413875)*x[2]
        arg[0,4,1,1]=(0.862892518463)*x[0]**o+(0.34033875398)*x[0]+(0.0371376953103)*x[1]**o+(0.568160310726)*x[1]+(-0.237231036717)*x[2]**o+(-0.130046899302)*x[2]
        arg[0,4,2,0]=(0.223651181698)*x[0]**o+(0.730991398203)*x[0]+(0.133551525027)*x[1]**o+(0.685783540126)*x[1]+(0.339586395814)*x[2]**o+(-0.524371194863)*x[2]
        arg[0,4,2,1]=(0.220110871383)*x[0]**o+(-0.37385773877)*x[0]+(0.773372000545)*x[1]**o+(-0.4261356565)*x[1]+(-0.619438438945)*x[2]**o+(0.126288949127)*x[2]
        arg[1,0,0,0]=(0.727680196631)*x[0]**o+(-0.876280912599)*x[0]+(0.576553782915)*x[1]**o+(-0.0497780322367)*x[1]+(-0.690257457282)*x[2]**o+(-0.11130953054)*x[2]
        arg[1,0,0,1]=(0.331863815938)*x[0]**o+(0.169379824389)*x[0]+(0.665507756012)*x[1]**o+(0.94933357273)*x[1]+(-0.610506070093)*x[2]**o+(0.752476708027)*x[2]
        arg[1,0,1,0]=(0.851320583415)*x[0]**o+(-0.697235996161)*x[0]+(0.235192165807)*x[1]**o+(0.325473517022)*x[1]+(-0.424683549673)*x[2]**o+(0.228953796486)*x[2]
        arg[1,0,1,1]=(0.804192098517)*x[0]**o+(-0.684337197059)*x[0]+(-0.349251615715)*x[1]**o+(0.596165705743)*x[1]+(-0.636033221182)*x[2]**o+(-0.304367076307)*x[2]
        arg[1,0,2,0]=(-0.237305477187)*x[0]**o+(-0.997679652719)*x[0]+(0.485994068389)*x[1]**o+(0.44508530534)*x[1]+(0.940139935628)*x[2]**o+(0.0789221540062)*x[2]
        arg[1,0,2,1]=(0.814928463364)*x[0]**o+(-0.257693939789)*x[0]+(0.179643785703)*x[1]**o+(-0.696436530567)*x[1]+(-0.592226945554)*x[2]**o+(-0.454620411499)*x[2]
        arg[1,1,0,0]=(-0.933925672439)*x[0]**o+(0.212298180817)*x[0]+(-0.447836316801)*x[1]**o+(0.180796376138)*x[1]+(-0.749910125701)*x[2]**o+(0.134540433851)*x[2]
        arg[1,1,0,1]=(-0.649009346978)*x[0]**o+(0.0316819485953)*x[0]+(-0.718001767835)*x[1]**o+(0.289608435124)*x[1]+(-0.906045493384)*x[2]**o+(0.851021484125)*x[2]
        arg[1,1,1,0]=(-0.216265404566)*x[0]**o+(0.657073282172)*x[0]+(-0.796155541438)*x[1]**o+(0.338980498824)*x[1]+(0.477444448193)*x[2]**o+(0.674278909522)*x[2]
        arg[1,1,1,1]=(-0.67605942954)*x[0]**o+(0.457997253216)*x[0]+(-0.973576381948)*x[1]**o+(0.430432753773)*x[1]+(0.0329169564451)*x[2]**o+(-0.422512702485)*x[2]
        arg[1,1,2,0]=(-0.32458256139)*x[0]**o+(0.289568625922)*x[0]+(-0.529564027476)*x[1]**o+(0.596449833118)*x[1]+(0.960461031989)*x[2]**o+(0.00801425321069)*x[2]
        arg[1,1,2,1]=(-0.0425883615923)*x[0]**o+(-0.313926807333)*x[0]+(0.766062773671)*x[1]**o+(-0.857784560309)*x[1]+(0.914867381682)*x[2]**o+(-0.312493334639)*x[2]
        arg[1,2,0,0]=(-0.637151767355)*x[0]**o+(-0.679947136952)*x[0]+(-0.0838349105494)*x[1]**o+(0.00800835879666)*x[1]+(0.385024194519)*x[2]**o+(-0.627828236571)*x[2]
        arg[1,2,0,1]=(-0.62580153007)*x[0]**o+(-0.183834415542)*x[0]+(0.322406823575)*x[1]**o+(0.302908755868)*x[1]+(0.727234001542)*x[2]**o+(-0.933944878425)*x[2]
        arg[1,2,1,0]=(-0.897198312615)*x[0]**o+(0.825086526945)*x[0]+(-0.031001758991)*x[1]**o+(0.987845551304)*x[1]+(-0.551212418179)*x[2]**o+(0.657813241764)*x[2]
        arg[1,2,1,1]=(-0.544595822278)*x[0]**o+(0.595858429591)*x[0]+(-0.557503238514)*x[1]**o+(0.507560179011)*x[1]+(-0.519438270771)*x[2]**o+(-0.471678264995)*x[2]
        arg[1,2,2,0]=(0.135616278099)*x[0]**o+(0.40676743342)*x[0]+(0.458810719213)*x[1]**o+(-0.881620410228)*x[1]+(-0.897267732846)*x[2]**o+(0.665458992487)*x[2]
        arg[1,2,2,1]=(0.83854610457)*x[0]**o+(0.670263196782)*x[0]+(0.551767734306)*x[1]**o+(0.883934821215)*x[1]+(0.455375261768)*x[2]**o+(0.95604480045)*x[2]
        arg[1,3,0,0]=(0.705853698882)*x[0]**o+(0.84681314245)*x[0]+(-0.0984020252408)*x[1]**o+(0.943165990555)*x[1]+(-0.428955841327)*x[2]**o+(0.0961618993003)*x[2]
        arg[1,3,0,1]=(0.583731800337)*x[0]**o+(-0.738079255003)*x[0]+(0.315556768944)*x[1]**o+(-0.489631118687)*x[1]+(0.371775779104)*x[2]**o+(0.82029490872)*x[2]
        arg[1,3,1,0]=(-0.885657545142)*x[0]**o+(0.542706113286)*x[0]+(0.800319452136)*x[1]**o+(0.314194184897)*x[1]+(-0.320905532427)*x[2]**o+(-0.314307365683)*x[2]
        arg[1,3,1,1]=(-0.217030771566)*x[0]**o+(0.260108519031)*x[0]+(-0.686624725975)*x[1]**o+(-0.869322439609)*x[1]+(0.558806239315)*x[2]**o+(-0.440661121246)*x[2]
        arg[1,3,2,0]=(-0.839335541382)*x[0]**o+(0.228726621639)*x[0]+(0.354403444002)*x[1]**o+(0.215616048959)*x[1]+(0.49622647562)*x[2]**o+(-0.725639112729)*x[2]
        arg[1,3,2,1]=(-0.439687404632)*x[0]**o+(-0.820583921269)*x[0]+(-0.863730409631)*x[1]**o+(0.119773401911)*x[1]+(0.460158483187)*x[2]**o+(0.757153436592)*x[2]
        arg[1,4,0,0]=(-0.77288839701)*x[0]**o+(-0.570203714622)*x[0]+(0.93832772305)*x[1]**o+(-0.759089944596)*x[1]+(0.000174937961704)*x[2]**o+(0.812917484286)*x[2]
        arg[1,4,0,1]=(-0.25535378432)*x[0]**o+(-0.0938207170929)*x[0]+(0.0545321717685)*x[1]**o+(-0.414414792674)*x[1]+(-0.297698896712)*x[2]**o+(-0.367166660571)*x[2]
        arg[1,4,1,0]=(0.743408112248)*x[0]**o+(0.852863814785)*x[0]+(0.259483951249)*x[1]**o+(-0.519288864163)*x[1]+(-0.305772267727)*x[2]**o+(-0.401384318883)*x[2]
        arg[1,4,1,1]=(0.194889815352)*x[0]**o+(0.522093986804)*x[0]+(-0.343320977396)*x[1]**o+(0.21840236597)*x[1]+(-0.672107409136)*x[2]**o+(-0.600203805002)*x[2]
        arg[1,4,2,0]=(0.420860556361)*x[0]**o+(0.789879869049)*x[0]+(0.100987805744)*x[1]**o+(0.802794815123)*x[1]+(-0.69466030201)*x[2]**o+(-0.746043845966)*x[2]
        arg[1,4,2,1]=(-0.306621579211)*x[0]**o+(-0.0676982346933)*x[0]+(0.568861573715)*x[1]**o+(0.0823366334578)*x[1]+(0.501187255641)*x[2]**o+(-0.538615164478)*x[2]
        arg[2,0,0,0]=(-0.231052916706)*x[0]**o+(0.127790831537)*x[0]+(-0.475600435038)*x[1]**o+(0.431586151473)*x[1]+(-0.882365547235)*x[2]**o+(0.778027701879)*x[2]
        arg[2,0,0,1]=(0.0709490835498)*x[0]**o+(0.442614693755)*x[0]+(0.870405247692)*x[1]**o+(-0.648538779515)*x[1]+(0.751107652054)*x[2]**o+(0.0122956931127)*x[2]
        arg[2,0,1,0]=(0.703128329145)*x[0]**o+(0.259775843631)*x[0]+(0.311958888594)*x[1]**o+(-0.0445092402277)*x[1]+(0.338222457799)*x[2]**o+(0.845995300456)*x[2]
        arg[2,0,1,1]=(-0.653977898919)*x[0]**o+(0.0495539350857)*x[0]+(0.4013623408)*x[1]**o+(0.490747039887)*x[1]+(-0.480487636729)*x[2]**o+(0.361660115775)*x[2]
        arg[2,0,2,0]=(0.215778506945)*x[0]**o+(0.316853875053)*x[0]+(0.556770639829)*x[1]**o+(-0.499729882355)*x[1]+(0.692012327582)*x[2]**o+(0.00681314946195)*x[2]
        arg[2,0,2,1]=(0.432077857223)*x[0]**o+(-0.0497579036988)*x[0]+(-0.465403408565)*x[1]**o+(0.125421502956)*x[1]+(-0.313541879877)*x[2]**o+(0.113957941731)*x[2]
        arg[2,1,0,0]=(-0.528126103192)*x[0]**o+(-0.141306512096)*x[0]+(-0.505835313198)*x[1]**o+(-0.0433171534408)*x[1]+(-0.522465687628)*x[2]**o+(-0.16728285732)*x[2]
        arg[2,1,0,1]=(0.892019660948)*x[0]**o+(0.646944130193)*x[0]+(0.865495756748)*x[1]**o+(0.949095276008)*x[1]+(0.786394755068)*x[2]**o+(0.482441081687)*x[2]
        arg[2,1,1,0]=(-0.0485295711204)*x[0]**o+(0.250163927911)*x[0]+(0.0387861125998)*x[1]**o+(-0.0789121929817)*x[1]+(-0.982330126583)*x[2]**o+(0.66328870814)*x[2]
        arg[2,1,1,1]=(-0.494726556678)*x[0]**o+(-0.961342091528)*x[0]+(0.211410563083)*x[1]**o+(-0.884287887286)*x[1]+(-0.583401908135)*x[2]**o+(-0.361423905219)*x[2]
        arg[2,1,2,0]=(-0.326785246692)*x[0]**o+(0.596272945325)*x[0]+(0.625039539234)*x[1]**o+(0.88699688986)*x[1]+(-0.649803810419)*x[2]**o+(0.890718271469)*x[2]
        arg[2,1,2,1]=(0.702110072391)*x[0]**o+(0.72009439563)*x[0]+(-0.930941572511)*x[1]**o+(0.955877158334)*x[1]+(0.1211281332)*x[2]**o+(0.658738626852)*x[2]
        arg[2,2,0,0]=(-0.177402589697)*x[0]**o+(0.701090894256)*x[0]+(-0.00205333079339)*x[1]**o+(-0.427862310859)*x[1]+(0.529186374803)*x[2]**o+(-0.110462035551)*x[2]
        arg[2,2,0,1]=(-0.104112165367)*x[0]**o+(0.043539351452)*x[0]+(0.757384935306)*x[1]**o+(-0.451086492844)*x[1]+(0.0549627297033)*x[2]**o+(-0.311718955315)*x[2]
        arg[2,2,1,0]=(-0.579986090808)*x[0]**o+(-0.922872917924)*x[0]+(0.66422700207)*x[1]**o+(0.947304780214)*x[1]+(-0.0492553517704)*x[2]**o+(0.817225619153)*x[2]
        arg[2,2,1,1]=(0.306298476233)*x[0]**o+(-0.620190960783)*x[0]+(-0.684876074013)*x[1]**o+(-0.371829616982)*x[1]+(0.904709743321)*x[2]**o+(0.170802273852)*x[2]
        arg[2,2,2,0]=(0.317943399758)*x[0]**o+(0.518026054212)*x[0]+(-0.801462311167)*x[1]**o+(0.767424157898)*x[1]+(-0.547271995492)*x[2]**o+(-0.912619177188)*x[2]
        arg[2,2,2,1]=(-0.848711258358)*x[0]**o+(0.456489097468)*x[0]+(-0.310979078367)*x[1]**o+(-0.946717007726)*x[1]+(0.454363833646)*x[2]**o+(-0.715701991077)*x[2]
        arg[2,3,0,0]=(0.605493533014)*x[0]**o+(0.478437880368)*x[0]+(-0.672727491605)*x[1]**o+(0.831259256009)*x[1]+(-0.434153354239)*x[2]**o+(-0.505363218578)*x[2]
        arg[2,3,0,1]=(-0.0703539676643)*x[0]**o+(-0.452352344215)*x[0]+(0.0403568720903)*x[1]**o+(0.959415939826)*x[1]+(-0.784519698477)*x[2]**o+(-0.691664442958)*x[2]
        arg[2,3,1,0]=(0.696353768719)*x[0]**o+(0.295477297618)*x[0]+(0.580540484825)*x[1]**o+(0.098954967023)*x[1]+(-0.547594444517)*x[2]**o+(-0.574783665389)*x[2]
        arg[2,3,1,1]=(0.997841887408)*x[0]**o+(0.857864040549)*x[0]+(-0.413188343971)*x[1]**o+(0.344770631687)*x[1]+(-0.127257058552)*x[2]**o+(0.132618819456)*x[2]
        arg[2,3,2,0]=(0.174045751083)*x[0]**o+(0.866724759298)*x[0]+(0.0818516103455)*x[1]**o+(-0.891670609599)*x[1]+(0.982721717044)*x[2]**o+(0.378710155856)*x[2]
        arg[2,3,2,1]=(0.7760778906)*x[0]**o+(-0.710104290654)*x[0]+(-0.471259845977)*x[1]**o+(-0.791979211724)*x[1]+(-0.845652753706)*x[2]**o+(-0.301795837852)*x[2]
        arg[2,4,0,0]=(-0.698193658421)*x[0]**o+(-0.831956135385)*x[0]+(-0.442637915133)*x[1]**o+(0.108715813637)*x[1]+(-0.888348149694)*x[2]**o+(-0.90115888496)*x[2]
        arg[2,4,0,1]=(0.770237705577)*x[0]**o+(0.889304763618)*x[0]+(-0.139196183564)*x[1]**o+(0.332516232239)*x[1]+(-0.0218103253052)*x[2]**o+(-0.368033928854)*x[2]
        arg[2,4,1,0]=(0.675479351844)*x[0]**o+(0.880998418303)*x[0]+(0.408381837105)*x[1]**o+(0.717402364084)*x[1]+(-0.303902947422)*x[2]**o+(-0.322350049962)*x[2]
        arg[2,4,1,1]=(-0.480881673289)*x[0]**o+(0.30290141529)*x[0]+(-0.538411390886)*x[1]**o+(0.260090848702)*x[1]+(0.638130764808)*x[2]**o+(0.756338975575)*x[2]
        arg[2,4,2,0]=(-0.299233221889)*x[0]**o+(-0.26471083761)*x[0]+(0.717191172884)*x[1]**o+(0.488553216118)*x[1]+(0.718956726644)*x[2]**o+(0.370889025703)*x[2]
        arg[2,4,2,1]=(-0.989020272269)*x[0]**o+(-0.891423884362)*x[0]+(0.486153100693)*x[1]**o+(-0.971171395246)*x[1]+(-0.766942968184)*x[2]**o+(0.728728968862)*x[2]
        arg[3,0,0,0]=(0.756186246227)*x[0]**o+(0.76507856268)*x[0]+(0.771165000414)*x[1]**o+(0.673968873856)*x[1]+(0.827045846253)*x[2]**o+(-0.788265752412)*x[2]
        arg[3,0,0,1]=(-0.66925731439)*x[0]**o+(-0.0110127596435)*x[0]+(0.944388255414)*x[1]**o+(-0.613832072486)*x[1]+(0.0977542120724)*x[2]**o+(0.100776856689)*x[2]
        arg[3,0,1,0]=(0.905138713064)*x[0]**o+(0.673051338153)*x[0]+(0.240614253653)*x[1]**o+(-0.964132330033)*x[1]+(-0.114961018838)*x[2]**o+(-0.438756962988)*x[2]
        arg[3,0,1,1]=(-0.192259182094)*x[0]**o+(0.12315590119)*x[0]+(-0.907744420226)*x[1]**o+(-0.483981269809)*x[1]+(-0.169699431472)*x[2]**o+(-0.45231635406)*x[2]
        arg[3,0,2,0]=(0.895387255885)*x[0]**o+(0.23892997411)*x[0]+(-0.396010446618)*x[1]**o+(0.94853984413)*x[1]+(-0.0165259637241)*x[2]**o+(-0.312558539149)*x[2]
        arg[3,0,2,1]=(0.710478713361)*x[0]**o+(-0.109316102216)*x[0]+(-0.042711568986)*x[1]**o+(-0.448508085742)*x[1]+(-0.0979193346011)*x[2]**o+(0.93156672211)*x[2]
        arg[3,1,0,0]=(-0.21605108221)*x[0]**o+(0.410466481712)*x[0]+(0.799288386783)*x[1]**o+(0.622866017113)*x[1]+(-0.236698938159)*x[2]**o+(0.246625241928)*x[2]
        arg[3,1,0,1]=(0.989568527762)*x[0]**o+(0.0212906378151)*x[0]+(0.311279947324)*x[1]**o+(-0.231412902902)*x[1]+(-0.444760786142)*x[2]**o+(0.908171985421)*x[2]
        arg[3,1,1,0]=(0.42605234231)*x[0]**o+(-0.58629332944)*x[0]+(0.322063640761)*x[1]**o+(0.837771339825)*x[1]+(-0.911083211394)*x[2]**o+(-0.0699607581992)*x[2]
        arg[3,1,1,1]=(0.75449445591)*x[0]**o+(0.722970800307)*x[0]+(-0.582970919465)*x[1]**o+(-0.868818538212)*x[1]+(0.940067636181)*x[2]**o+(-0.659469743617)*x[2]
        arg[3,1,2,0]=(-0.391224924663)*x[0]**o+(-0.768173177737)*x[0]+(-0.856876347997)*x[1]**o+(-0.0451180499647)*x[1]+(0.198590679246)*x[2]**o+(0.187329095476)*x[2]
        arg[3,1,2,1]=(0.483309063092)*x[0]**o+(-0.0777241026893)*x[0]+(0.0438427457286)*x[1]**o+(-0.73195704867)*x[1]+(-0.439660745725)*x[2]**o+(0.279501672861)*x[2]
        arg[3,2,0,0]=(0.799935089139)*x[0]**o+(-0.306860809857)*x[0]+(0.169104975039)*x[1]**o+(-0.3480717525)*x[1]+(0.215231688)*x[2]**o+(-0.528775376945)*x[2]
        arg[3,2,0,1]=(0.917815789587)*x[0]**o+(0.0672910338102)*x[0]+(-0.186550440011)*x[1]**o+(0.922579714252)*x[1]+(0.924162379857)*x[2]**o+(-0.769320483618)*x[2]
        arg[3,2,1,0]=(0.604324468058)*x[0]**o+(-0.0109731419053)*x[0]+(0.229447318325)*x[1]**o+(-0.480693435232)*x[1]+(-0.881779256999)*x[2]**o+(0.288656394771)*x[2]
        arg[3,2,1,1]=(0.934680723352)*x[0]**o+(0.993838528376)*x[0]+(0.957781588502)*x[1]**o+(-0.859837057866)*x[1]+(-0.702139964493)*x[2]**o+(0.0538177124609)*x[2]
        arg[3,2,2,0]=(-0.142809047814)*x[0]**o+(-0.472594415134)*x[0]+(0.0381782768377)*x[1]**o+(0.727331491769)*x[1]+(0.0910992897581)*x[2]**o+(-0.845316063851)*x[2]
        arg[3,2,2,1]=(-0.217483358272)*x[0]**o+(0.991535092285)*x[0]+(-0.399175896468)*x[1]**o+(-0.665836512895)*x[1]+(0.0983700911493)*x[2]**o+(0.321811300155)*x[2]
        arg[3,3,0,0]=(0.797049585464)*x[0]**o+(-0.724467959299)*x[0]+(0.19971589973)*x[1]**o+(0.845928461864)*x[1]+(-0.365158866359)*x[2]**o+(-0.698274270132)*x[2]
        arg[3,3,0,1]=(-0.638474704878)*x[0]**o+(0.551866433352)*x[0]+(0.192309452979)*x[1]**o+(-0.703649325501)*x[1]+(0.823910578138)*x[2]**o+(-0.917446989634)*x[2]
        arg[3,3,1,0]=(0.217577282169)*x[0]**o+(0.521716803668)*x[0]+(-0.116587783856)*x[1]**o+(0.529672391417)*x[1]+(-0.798918158284)*x[2]**o+(-0.0996139511955)*x[2]
        arg[3,3,1,1]=(0.498965049079)*x[0]**o+(0.316489921186)*x[0]+(-0.0988355471817)*x[1]**o+(-0.0477508287681)*x[1]+(-0.778914523993)*x[2]**o+(0.189397812723)*x[2]
        arg[3,3,2,0]=(0.203823698152)*x[0]**o+(-0.340499414266)*x[0]+(0.353374693517)*x[1]**o+(-0.91035264494)*x[1]+(-0.779204084198)*x[2]**o+(-0.00276454165248)*x[2]
        arg[3,3,2,1]=(0.295232007123)*x[0]**o+(0.271766544502)*x[0]+(0.431956430336)*x[1]**o+(0.896481311536)*x[1]+(-0.325014902487)*x[2]**o+(0.70836170284)*x[2]
        arg[3,4,0,0]=(-0.407535202468)*x[0]**o+(-0.33055681091)*x[0]+(0.129762143151)*x[1]**o+(0.352087436228)*x[1]+(-0.538770401565)*x[2]**o+(-0.519071077045)*x[2]
        arg[3,4,0,1]=(0.846011435499)*x[0]**o+(0.172822133672)*x[0]+(-0.740732553847)*x[1]**o+(0.230062837231)*x[1]+(-0.231810007237)*x[2]**o+(-0.568366338923)*x[2]
        arg[3,4,1,0]=(-0.198091119057)*x[0]**o+(0.957989508794)*x[0]+(0.861739071175)*x[1]**o+(0.923450697883)*x[1]+(-0.755262504937)*x[2]**o+(0.102897579936)*x[2]
        arg[3,4,1,1]=(0.198829545353)*x[0]**o+(-0.757182006818)*x[0]+(0.254092337808)*x[1]**o+(-0.00815072620097)*x[1]+(0.347421988308)*x[2]**o+(-0.0215277616923)*x[2]
        arg[3,4,2,0]=(0.775062112529)*x[0]**o+(-0.77746792774)*x[0]+(-0.666490294869)*x[1]**o+(-0.679971069878)*x[1]+(-0.972450779894)*x[2]**o+(-0.705322282477)*x[2]
        arg[3,4,2,1]=(-0.0994927587382)*x[0]**o+(0.293071774786)*x[0]+(0.370784399397)*x[1]**o+(0.931411225946)*x[1]+(-0.55789163294)*x[2]**o+(0.0831890658592)*x[2]
        ref[0,0,0,0]=(-0.657703355307)*x_ref[0]**o+(0.397813831843)*x_ref[0]+(0.364460538717)*x_ref[1]**o+(-0.845626669617)*x_ref[1]+(-0.987144254126)*x_ref[2]**o+(-0.732384980726)*x_ref[2]
        ref[0,0,0,1]=(-0.13280304886)*x_ref[0]**o+(-0.953796807758)*x_ref[0]+(0.875723746654)*x_ref[1]**o+(0.0764745117242)*x_ref[1]+(-0.536085676221)*x_ref[2]**o+(-0.520611959678)*x_ref[2]
        ref[0,0,1,0]=(0.716429446578)*x_ref[0]**o+(-0.729164693441)*x_ref[0]+(-0.832505557428)*x_ref[1]**o+(-0.329200605611)*x_ref[1]+(0.0310162060964)*x_ref[2]**o+(-0.869978941999)*x_ref[2]
        ref[0,0,1,1]=(0.110253530584)*x_ref[0]**o+(0.995528486822)*x_ref[0]+(0.838577797395)*x_ref[1]**o+(0.443987752225)*x_ref[1]+(0.387337959057)*x_ref[2]**o+(0.548448752612)*x_ref[2]
        ref[0,0,2,0]=(-0.995737515431)*x_ref[0]**o+(-0.899807241717)*x_ref[0]+(-0.0803287764138)*x_ref[1]**o+(0.242680051556)*x_ref[1]+(0.58974836823)*x_ref[2]**o+(-0.498450188944)*x_ref[2]
        ref[0,0,2,1]=(0.341561541555)*x_ref[0]**o+(-0.00541988184597)*x_ref[0]+(-0.204832898701)*x_ref[1]**o+(-0.957247935865)*x_ref[1]+(-0.441002902064)*x_ref[2]**o+(0.756843480304)*x_ref[2]
        ref[0,1,0,0]=(0.136201526244)*x_ref[0]**o+(0.517677304193)*x_ref[0]+(0.523031610748)*x_ref[1]**o+(-0.908998053383)*x_ref[1]+(-0.629649678416)*x_ref[2]**o+(0.762977917512)*x_ref[2]
        ref[0,1,0,1]=(-0.103521080864)*x_ref[0]**o+(-0.434410222595)*x_ref[0]+(0.418239530241)*x_ref[1]**o+(0.65638475119)*x_ref[1]+(0.413036393129)*x_ref[2]**o+(-0.220365518353)*x_ref[2]
        ref[0,1,1,0]=(-0.415690201483)*x_ref[0]**o+(-0.969323394689)*x_ref[0]+(-0.295639420195)*x_ref[1]**o+(-0.24915002749)*x_ref[1]+(-0.00927627101467)*x_ref[2]**o+(0.520335261634)*x_ref[2]
        ref[0,1,1,1]=(-0.262653741797)*x_ref[0]**o+(-0.463590591666)*x_ref[0]+(0.568893042923)*x_ref[1]**o+(-0.869851332898)*x_ref[1]+(-0.463261399415)*x_ref[2]**o+(0.0281915686202)*x_ref[2]
        ref[0,1,2,0]=(0.377159503058)*x_ref[0]**o+(0.484512261943)*x_ref[0]+(0.0303594628925)*x_ref[1]**o+(0.482051243004)*x_ref[1]+(0.240643013949)*x_ref[2]**o+(-0.943908617705)*x_ref[2]
        ref[0,1,2,1]=(0.843854614097)*x_ref[0]**o+(0.018041534553)*x_ref[0]+(0.264198703361)*x_ref[1]**o+(0.732889517628)*x_ref[1]+(-0.154807384394)*x_ref[2]**o+(0.60513622187)*x_ref[2]
        ref[0,2,0,0]=(-0.234758118286)*x_ref[0]**o+(-0.77807396119)*x_ref[0]+(-0.97783091103)*x_ref[1]**o+(-0.716185223828)*x_ref[1]+(-0.573126474259)*x_ref[2]**o+(0.403110701555)*x_ref[2]
        ref[0,2,0,1]=(-0.853878344804)*x_ref[0]**o+(-0.482644826561)*x_ref[0]+(-0.826041149012)*x_ref[1]**o+(0.596376234798)*x_ref[1]+(-0.216400517274)*x_ref[2]**o+(-0.323127194259)*x_ref[2]
        ref[0,2,1,0]=(0.931551121492)*x_ref[0]**o+(-0.527132174379)*x_ref[0]+(0.373496704463)*x_ref[1]**o+(-0.375215826463)*x_ref[1]+(-0.639167297984)*x_ref[2]**o+(0.143608733189)*x_ref[2]
        ref[0,2,1,1]=(-0.982206000879)*x_ref[0]**o+(0.836589596534)*x_ref[0]+(0.606275000138)*x_ref[1]**o+(-0.821469453892)*x_ref[1]+(-0.633828380948)*x_ref[2]**o+(-0.637688777216)*x_ref[2]
        ref[0,2,2,0]=(0.441649557529)*x_ref[0]**o+(-0.34978365051)*x_ref[0]+(0.514175365405)*x_ref[1]**o+(-0.903212271234)*x_ref[1]+(0.70236887156)*x_ref[2]**o+(0.72953045189)*x_ref[2]
        ref[0,2,2,1]=(0.302787896978)*x_ref[0]**o+(0.157598071021)*x_ref[0]+(-0.357101947632)*x_ref[1]**o+(0.381098383156)*x_ref[1]+(-0.552546036689)*x_ref[2]**o+(0.448181721984)*x_ref[2]
        ref[0,3,0,0]=(0.662181077399)*x_ref[0]**o+(-0.283755239518)*x_ref[0]+(0.33694741278)*x_ref[1]**o+(0.275801685868)*x_ref[1]+(0.526155259101)*x_ref[2]**o+(-0.378241855559)*x_ref[2]
        ref[0,3,0,1]=(0.157822133581)*x_ref[0]**o+(0.25402072747)*x_ref[0]+(0.522842123099)*x_ref[1]**o+(0.510595773584)*x_ref[1]+(-0.829239126666)*x_ref[2]**o+(0.990879342044)*x_ref[2]
        ref[0,3,1,0]=(-0.775925280059)*x_ref[0]**o+(-0.726971388023)*x_ref[0]+(-0.630088730753)*x_ref[1]**o+(-0.731024178237)*x_ref[1]+(-0.288643541196)*x_ref[2]**o+(-0.283708428793)*x_ref[2]
        ref[0,3,1,1]=(-0.883176640445)*x_ref[0]**o+(-0.0810211976762)*x_ref[0]+(-0.102919217784)*x_ref[1]**o+(0.271384866385)*x_ref[1]+(0.980796900477)*x_ref[2]**o+(0.439294038672)*x_ref[2]
        ref[0,3,2,0]=(-0.826688221502)*x_ref[0]**o+(-0.909388347347)*x_ref[0]+(0.164406927722)*x_ref[1]**o+(0.60157189364)*x_ref[1]+(-0.885578176748)*x_ref[2]**o+(0.610008674247)*x_ref[2]
        ref[0,3,2,1]=(0.130154197193)*x_ref[0]**o+(-0.754703602975)*x_ref[0]+(0.00463710497017)*x_ref[1]**o+(0.502162916123)*x_ref[1]+(-0.749124548541)*x_ref[2]**o+(0.946592962092)*x_ref[2]
        ref[0,4,0,0]=(-0.0478347448906)*x_ref[0]**o+(-0.812937871882)*x_ref[0]+(-0.460982012551)*x_ref[1]**o+(0.317485998991)*x_ref[1]+(-0.593671683477)*x_ref[2]**o+(0.551385718668)*x_ref[2]
        ref[0,4,0,1]=(0.281998767502)*x_ref[0]**o+(-0.0613830311997)*x_ref[0]+(0.981429151019)*x_ref[1]**o+(-0.580567714008)*x_ref[1]+(0.0369483005331)*x_ref[2]**o+(-0.74132358588)*x_ref[2]
        ref[0,4,1,0]=(0.525845217483)*x_ref[0]**o+(0.907803915216)*x_ref[0]+(0.111429079263)*x_ref[1]**o+(-0.653030401377)*x_ref[1]+(0.0814793913696)*x_ref[2]**o+(-0.419201413875)*x_ref[2]
        ref[0,4,1,1]=(0.862892518463)*x_ref[0]**o+(0.34033875398)*x_ref[0]+(0.0371376953103)*x_ref[1]**o+(0.568160310726)*x_ref[1]+(-0.237231036717)*x_ref[2]**o+(-0.130046899302)*x_ref[2]
        ref[0,4,2,0]=(0.223651181698)*x_ref[0]**o+(0.730991398203)*x_ref[0]+(0.133551525027)*x_ref[1]**o+(0.685783540126)*x_ref[1]+(0.339586395814)*x_ref[2]**o+(-0.524371194863)*x_ref[2]
        ref[0,4,2,1]=(0.220110871383)*x_ref[0]**o+(-0.37385773877)*x_ref[0]+(0.773372000545)*x_ref[1]**o+(-0.4261356565)*x_ref[1]+(-0.619438438945)*x_ref[2]**o+(0.126288949127)*x_ref[2]
        ref[1,0,0,0]=(0.727680196631)*x_ref[0]**o+(-0.876280912599)*x_ref[0]+(0.576553782915)*x_ref[1]**o+(-0.0497780322367)*x_ref[1]+(-0.690257457282)*x_ref[2]**o+(-0.11130953054)*x_ref[2]
        ref[1,0,0,1]=(0.331863815938)*x_ref[0]**o+(0.169379824389)*x_ref[0]+(0.665507756012)*x_ref[1]**o+(0.94933357273)*x_ref[1]+(-0.610506070093)*x_ref[2]**o+(0.752476708027)*x_ref[2]
        ref[1,0,1,0]=(0.851320583415)*x_ref[0]**o+(-0.697235996161)*x_ref[0]+(0.235192165807)*x_ref[1]**o+(0.325473517022)*x_ref[1]+(-0.424683549673)*x_ref[2]**o+(0.228953796486)*x_ref[2]
        ref[1,0,1,1]=(0.804192098517)*x_ref[0]**o+(-0.684337197059)*x_ref[0]+(-0.349251615715)*x_ref[1]**o+(0.596165705743)*x_ref[1]+(-0.636033221182)*x_ref[2]**o+(-0.304367076307)*x_ref[2]
        ref[1,0,2,0]=(-0.237305477187)*x_ref[0]**o+(-0.997679652719)*x_ref[0]+(0.485994068389)*x_ref[1]**o+(0.44508530534)*x_ref[1]+(0.940139935628)*x_ref[2]**o+(0.0789221540062)*x_ref[2]
        ref[1,0,2,1]=(0.814928463364)*x_ref[0]**o+(-0.257693939789)*x_ref[0]+(0.179643785703)*x_ref[1]**o+(-0.696436530567)*x_ref[1]+(-0.592226945554)*x_ref[2]**o+(-0.454620411499)*x_ref[2]
        ref[1,1,0,0]=(-0.933925672439)*x_ref[0]**o+(0.212298180817)*x_ref[0]+(-0.447836316801)*x_ref[1]**o+(0.180796376138)*x_ref[1]+(-0.749910125701)*x_ref[2]**o+(0.134540433851)*x_ref[2]
        ref[1,1,0,1]=(-0.649009346978)*x_ref[0]**o+(0.0316819485953)*x_ref[0]+(-0.718001767835)*x_ref[1]**o+(0.289608435124)*x_ref[1]+(-0.906045493384)*x_ref[2]**o+(0.851021484125)*x_ref[2]
        ref[1,1,1,0]=(-0.216265404566)*x_ref[0]**o+(0.657073282172)*x_ref[0]+(-0.796155541438)*x_ref[1]**o+(0.338980498824)*x_ref[1]+(0.477444448193)*x_ref[2]**o+(0.674278909522)*x_ref[2]
        ref[1,1,1,1]=(-0.67605942954)*x_ref[0]**o+(0.457997253216)*x_ref[0]+(-0.973576381948)*x_ref[1]**o+(0.430432753773)*x_ref[1]+(0.0329169564451)*x_ref[2]**o+(-0.422512702485)*x_ref[2]
        ref[1,1,2,0]=(-0.32458256139)*x_ref[0]**o+(0.289568625922)*x_ref[0]+(-0.529564027476)*x_ref[1]**o+(0.596449833118)*x_ref[1]+(0.960461031989)*x_ref[2]**o+(0.00801425321069)*x_ref[2]
        ref[1,1,2,1]=(-0.0425883615923)*x_ref[0]**o+(-0.313926807333)*x_ref[0]+(0.766062773671)*x_ref[1]**o+(-0.857784560309)*x_ref[1]+(0.914867381682)*x_ref[2]**o+(-0.312493334639)*x_ref[2]
        ref[1,2,0,0]=(-0.637151767355)*x_ref[0]**o+(-0.679947136952)*x_ref[0]+(-0.0838349105494)*x_ref[1]**o+(0.00800835879666)*x_ref[1]+(0.385024194519)*x_ref[2]**o+(-0.627828236571)*x_ref[2]
        ref[1,2,0,1]=(-0.62580153007)*x_ref[0]**o+(-0.183834415542)*x_ref[0]+(0.322406823575)*x_ref[1]**o+(0.302908755868)*x_ref[1]+(0.727234001542)*x_ref[2]**o+(-0.933944878425)*x_ref[2]
        ref[1,2,1,0]=(-0.897198312615)*x_ref[0]**o+(0.825086526945)*x_ref[0]+(-0.031001758991)*x_ref[1]**o+(0.987845551304)*x_ref[1]+(-0.551212418179)*x_ref[2]**o+(0.657813241764)*x_ref[2]
        ref[1,2,1,1]=(-0.544595822278)*x_ref[0]**o+(0.595858429591)*x_ref[0]+(-0.557503238514)*x_ref[1]**o+(0.507560179011)*x_ref[1]+(-0.519438270771)*x_ref[2]**o+(-0.471678264995)*x_ref[2]
        ref[1,2,2,0]=(0.135616278099)*x_ref[0]**o+(0.40676743342)*x_ref[0]+(0.458810719213)*x_ref[1]**o+(-0.881620410228)*x_ref[1]+(-0.897267732846)*x_ref[2]**o+(0.665458992487)*x_ref[2]
        ref[1,2,2,1]=(0.83854610457)*x_ref[0]**o+(0.670263196782)*x_ref[0]+(0.551767734306)*x_ref[1]**o+(0.883934821215)*x_ref[1]+(0.455375261768)*x_ref[2]**o+(0.95604480045)*x_ref[2]
        ref[1,3,0,0]=(0.705853698882)*x_ref[0]**o+(0.84681314245)*x_ref[0]+(-0.0984020252408)*x_ref[1]**o+(0.943165990555)*x_ref[1]+(-0.428955841327)*x_ref[2]**o+(0.0961618993003)*x_ref[2]
        ref[1,3,0,1]=(0.583731800337)*x_ref[0]**o+(-0.738079255003)*x_ref[0]+(0.315556768944)*x_ref[1]**o+(-0.489631118687)*x_ref[1]+(0.371775779104)*x_ref[2]**o+(0.82029490872)*x_ref[2]
        ref[1,3,1,0]=(-0.885657545142)*x_ref[0]**o+(0.542706113286)*x_ref[0]+(0.800319452136)*x_ref[1]**o+(0.314194184897)*x_ref[1]+(-0.320905532427)*x_ref[2]**o+(-0.314307365683)*x_ref[2]
        ref[1,3,1,1]=(-0.217030771566)*x_ref[0]**o+(0.260108519031)*x_ref[0]+(-0.686624725975)*x_ref[1]**o+(-0.869322439609)*x_ref[1]+(0.558806239315)*x_ref[2]**o+(-0.440661121246)*x_ref[2]
        ref[1,3,2,0]=(-0.839335541382)*x_ref[0]**o+(0.228726621639)*x_ref[0]+(0.354403444002)*x_ref[1]**o+(0.215616048959)*x_ref[1]+(0.49622647562)*x_ref[2]**o+(-0.725639112729)*x_ref[2]
        ref[1,3,2,1]=(-0.439687404632)*x_ref[0]**o+(-0.820583921269)*x_ref[0]+(-0.863730409631)*x_ref[1]**o+(0.119773401911)*x_ref[1]+(0.460158483187)*x_ref[2]**o+(0.757153436592)*x_ref[2]
        ref[1,4,0,0]=(-0.77288839701)*x_ref[0]**o+(-0.570203714622)*x_ref[0]+(0.93832772305)*x_ref[1]**o+(-0.759089944596)*x_ref[1]+(0.000174937961704)*x_ref[2]**o+(0.812917484286)*x_ref[2]
        ref[1,4,0,1]=(-0.25535378432)*x_ref[0]**o+(-0.0938207170929)*x_ref[0]+(0.0545321717685)*x_ref[1]**o+(-0.414414792674)*x_ref[1]+(-0.297698896712)*x_ref[2]**o+(-0.367166660571)*x_ref[2]
        ref[1,4,1,0]=(0.743408112248)*x_ref[0]**o+(0.852863814785)*x_ref[0]+(0.259483951249)*x_ref[1]**o+(-0.519288864163)*x_ref[1]+(-0.305772267727)*x_ref[2]**o+(-0.401384318883)*x_ref[2]
        ref[1,4,1,1]=(0.194889815352)*x_ref[0]**o+(0.522093986804)*x_ref[0]+(-0.343320977396)*x_ref[1]**o+(0.21840236597)*x_ref[1]+(-0.672107409136)*x_ref[2]**o+(-0.600203805002)*x_ref[2]
        ref[1,4,2,0]=(0.420860556361)*x_ref[0]**o+(0.789879869049)*x_ref[0]+(0.100987805744)*x_ref[1]**o+(0.802794815123)*x_ref[1]+(-0.69466030201)*x_ref[2]**o+(-0.746043845966)*x_ref[2]
        ref[1,4,2,1]=(-0.306621579211)*x_ref[0]**o+(-0.0676982346933)*x_ref[0]+(0.568861573715)*x_ref[1]**o+(0.0823366334578)*x_ref[1]+(0.501187255641)*x_ref[2]**o+(-0.538615164478)*x_ref[2]
        ref[2,0,0,0]=(-0.231052916706)*x_ref[0]**o+(0.127790831537)*x_ref[0]+(-0.475600435038)*x_ref[1]**o+(0.431586151473)*x_ref[1]+(-0.882365547235)*x_ref[2]**o+(0.778027701879)*x_ref[2]
        ref[2,0,0,1]=(0.0709490835498)*x_ref[0]**o+(0.442614693755)*x_ref[0]+(0.870405247692)*x_ref[1]**o+(-0.648538779515)*x_ref[1]+(0.751107652054)*x_ref[2]**o+(0.0122956931127)*x_ref[2]
        ref[2,0,1,0]=(0.703128329145)*x_ref[0]**o+(0.259775843631)*x_ref[0]+(0.311958888594)*x_ref[1]**o+(-0.0445092402277)*x_ref[1]+(0.338222457799)*x_ref[2]**o+(0.845995300456)*x_ref[2]
        ref[2,0,1,1]=(-0.653977898919)*x_ref[0]**o+(0.0495539350857)*x_ref[0]+(0.4013623408)*x_ref[1]**o+(0.490747039887)*x_ref[1]+(-0.480487636729)*x_ref[2]**o+(0.361660115775)*x_ref[2]
        ref[2,0,2,0]=(0.215778506945)*x_ref[0]**o+(0.316853875053)*x_ref[0]+(0.556770639829)*x_ref[1]**o+(-0.499729882355)*x_ref[1]+(0.692012327582)*x_ref[2]**o+(0.00681314946195)*x_ref[2]
        ref[2,0,2,1]=(0.432077857223)*x_ref[0]**o+(-0.0497579036988)*x_ref[0]+(-0.465403408565)*x_ref[1]**o+(0.125421502956)*x_ref[1]+(-0.313541879877)*x_ref[2]**o+(0.113957941731)*x_ref[2]
        ref[2,1,0,0]=(-0.528126103192)*x_ref[0]**o+(-0.141306512096)*x_ref[0]+(-0.505835313198)*x_ref[1]**o+(-0.0433171534408)*x_ref[1]+(-0.522465687628)*x_ref[2]**o+(-0.16728285732)*x_ref[2]
        ref[2,1,0,1]=(0.892019660948)*x_ref[0]**o+(0.646944130193)*x_ref[0]+(0.865495756748)*x_ref[1]**o+(0.949095276008)*x_ref[1]+(0.786394755068)*x_ref[2]**o+(0.482441081687)*x_ref[2]
        ref[2,1,1,0]=(-0.0485295711204)*x_ref[0]**o+(0.250163927911)*x_ref[0]+(0.0387861125998)*x_ref[1]**o+(-0.0789121929817)*x_ref[1]+(-0.982330126583)*x_ref[2]**o+(0.66328870814)*x_ref[2]
        ref[2,1,1,1]=(-0.494726556678)*x_ref[0]**o+(-0.961342091528)*x_ref[0]+(0.211410563083)*x_ref[1]**o+(-0.884287887286)*x_ref[1]+(-0.583401908135)*x_ref[2]**o+(-0.361423905219)*x_ref[2]
        ref[2,1,2,0]=(-0.326785246692)*x_ref[0]**o+(0.596272945325)*x_ref[0]+(0.625039539234)*x_ref[1]**o+(0.88699688986)*x_ref[1]+(-0.649803810419)*x_ref[2]**o+(0.890718271469)*x_ref[2]
        ref[2,1,2,1]=(0.702110072391)*x_ref[0]**o+(0.72009439563)*x_ref[0]+(-0.930941572511)*x_ref[1]**o+(0.955877158334)*x_ref[1]+(0.1211281332)*x_ref[2]**o+(0.658738626852)*x_ref[2]
        ref[2,2,0,0]=(-0.177402589697)*x_ref[0]**o+(0.701090894256)*x_ref[0]+(-0.00205333079339)*x_ref[1]**o+(-0.427862310859)*x_ref[1]+(0.529186374803)*x_ref[2]**o+(-0.110462035551)*x_ref[2]
        ref[2,2,0,1]=(-0.104112165367)*x_ref[0]**o+(0.043539351452)*x_ref[0]+(0.757384935306)*x_ref[1]**o+(-0.451086492844)*x_ref[1]+(0.0549627297033)*x_ref[2]**o+(-0.311718955315)*x_ref[2]
        ref[2,2,1,0]=(-0.579986090808)*x_ref[0]**o+(-0.922872917924)*x_ref[0]+(0.66422700207)*x_ref[1]**o+(0.947304780214)*x_ref[1]+(-0.0492553517704)*x_ref[2]**o+(0.817225619153)*x_ref[2]
        ref[2,2,1,1]=(0.306298476233)*x_ref[0]**o+(-0.620190960783)*x_ref[0]+(-0.684876074013)*x_ref[1]**o+(-0.371829616982)*x_ref[1]+(0.904709743321)*x_ref[2]**o+(0.170802273852)*x_ref[2]
        ref[2,2,2,0]=(0.317943399758)*x_ref[0]**o+(0.518026054212)*x_ref[0]+(-0.801462311167)*x_ref[1]**o+(0.767424157898)*x_ref[1]+(-0.547271995492)*x_ref[2]**o+(-0.912619177188)*x_ref[2]
        ref[2,2,2,1]=(-0.848711258358)*x_ref[0]**o+(0.456489097468)*x_ref[0]+(-0.310979078367)*x_ref[1]**o+(-0.946717007726)*x_ref[1]+(0.454363833646)*x_ref[2]**o+(-0.715701991077)*x_ref[2]
        ref[2,3,0,0]=(0.605493533014)*x_ref[0]**o+(0.478437880368)*x_ref[0]+(-0.672727491605)*x_ref[1]**o+(0.831259256009)*x_ref[1]+(-0.434153354239)*x_ref[2]**o+(-0.505363218578)*x_ref[2]
        ref[2,3,0,1]=(-0.0703539676643)*x_ref[0]**o+(-0.452352344215)*x_ref[0]+(0.0403568720903)*x_ref[1]**o+(0.959415939826)*x_ref[1]+(-0.784519698477)*x_ref[2]**o+(-0.691664442958)*x_ref[2]
        ref[2,3,1,0]=(0.696353768719)*x_ref[0]**o+(0.295477297618)*x_ref[0]+(0.580540484825)*x_ref[1]**o+(0.098954967023)*x_ref[1]+(-0.547594444517)*x_ref[2]**o+(-0.574783665389)*x_ref[2]
        ref[2,3,1,1]=(0.997841887408)*x_ref[0]**o+(0.857864040549)*x_ref[0]+(-0.413188343971)*x_ref[1]**o+(0.344770631687)*x_ref[1]+(-0.127257058552)*x_ref[2]**o+(0.132618819456)*x_ref[2]
        ref[2,3,2,0]=(0.174045751083)*x_ref[0]**o+(0.866724759298)*x_ref[0]+(0.0818516103455)*x_ref[1]**o+(-0.891670609599)*x_ref[1]+(0.982721717044)*x_ref[2]**o+(0.378710155856)*x_ref[2]
        ref[2,3,2,1]=(0.7760778906)*x_ref[0]**o+(-0.710104290654)*x_ref[0]+(-0.471259845977)*x_ref[1]**o+(-0.791979211724)*x_ref[1]+(-0.845652753706)*x_ref[2]**o+(-0.301795837852)*x_ref[2]
        ref[2,4,0,0]=(-0.698193658421)*x_ref[0]**o+(-0.831956135385)*x_ref[0]+(-0.442637915133)*x_ref[1]**o+(0.108715813637)*x_ref[1]+(-0.888348149694)*x_ref[2]**o+(-0.90115888496)*x_ref[2]
        ref[2,4,0,1]=(0.770237705577)*x_ref[0]**o+(0.889304763618)*x_ref[0]+(-0.139196183564)*x_ref[1]**o+(0.332516232239)*x_ref[1]+(-0.0218103253052)*x_ref[2]**o+(-0.368033928854)*x_ref[2]
        ref[2,4,1,0]=(0.675479351844)*x_ref[0]**o+(0.880998418303)*x_ref[0]+(0.408381837105)*x_ref[1]**o+(0.717402364084)*x_ref[1]+(-0.303902947422)*x_ref[2]**o+(-0.322350049962)*x_ref[2]
        ref[2,4,1,1]=(-0.480881673289)*x_ref[0]**o+(0.30290141529)*x_ref[0]+(-0.538411390886)*x_ref[1]**o+(0.260090848702)*x_ref[1]+(0.638130764808)*x_ref[2]**o+(0.756338975575)*x_ref[2]
        ref[2,4,2,0]=(-0.299233221889)*x_ref[0]**o+(-0.26471083761)*x_ref[0]+(0.717191172884)*x_ref[1]**o+(0.488553216118)*x_ref[1]+(0.718956726644)*x_ref[2]**o+(0.370889025703)*x_ref[2]
        ref[2,4,2,1]=(-0.989020272269)*x_ref[0]**o+(-0.891423884362)*x_ref[0]+(0.486153100693)*x_ref[1]**o+(-0.971171395246)*x_ref[1]+(-0.766942968184)*x_ref[2]**o+(0.728728968862)*x_ref[2]
        ref[3,0,0,0]=(0.756186246227)*x_ref[0]**o+(0.76507856268)*x_ref[0]+(0.771165000414)*x_ref[1]**o+(0.673968873856)*x_ref[1]+(0.827045846253)*x_ref[2]**o+(-0.788265752412)*x_ref[2]
        ref[3,0,0,1]=(-0.66925731439)*x_ref[0]**o+(-0.0110127596435)*x_ref[0]+(0.944388255414)*x_ref[1]**o+(-0.613832072486)*x_ref[1]+(0.0977542120724)*x_ref[2]**o+(0.100776856689)*x_ref[2]
        ref[3,0,1,0]=(0.905138713064)*x_ref[0]**o+(0.673051338153)*x_ref[0]+(0.240614253653)*x_ref[1]**o+(-0.964132330033)*x_ref[1]+(-0.114961018838)*x_ref[2]**o+(-0.438756962988)*x_ref[2]
        ref[3,0,1,1]=(-0.192259182094)*x_ref[0]**o+(0.12315590119)*x_ref[0]+(-0.907744420226)*x_ref[1]**o+(-0.483981269809)*x_ref[1]+(-0.169699431472)*x_ref[2]**o+(-0.45231635406)*x_ref[2]
        ref[3,0,2,0]=(0.895387255885)*x_ref[0]**o+(0.23892997411)*x_ref[0]+(-0.396010446618)*x_ref[1]**o+(0.94853984413)*x_ref[1]+(-0.0165259637241)*x_ref[2]**o+(-0.312558539149)*x_ref[2]
        ref[3,0,2,1]=(0.710478713361)*x_ref[0]**o+(-0.109316102216)*x_ref[0]+(-0.042711568986)*x_ref[1]**o+(-0.448508085742)*x_ref[1]+(-0.0979193346011)*x_ref[2]**o+(0.93156672211)*x_ref[2]
        ref[3,1,0,0]=(-0.21605108221)*x_ref[0]**o+(0.410466481712)*x_ref[0]+(0.799288386783)*x_ref[1]**o+(0.622866017113)*x_ref[1]+(-0.236698938159)*x_ref[2]**o+(0.246625241928)*x_ref[2]
        ref[3,1,0,1]=(0.989568527762)*x_ref[0]**o+(0.0212906378151)*x_ref[0]+(0.311279947324)*x_ref[1]**o+(-0.231412902902)*x_ref[1]+(-0.444760786142)*x_ref[2]**o+(0.908171985421)*x_ref[2]
        ref[3,1,1,0]=(0.42605234231)*x_ref[0]**o+(-0.58629332944)*x_ref[0]+(0.322063640761)*x_ref[1]**o+(0.837771339825)*x_ref[1]+(-0.911083211394)*x_ref[2]**o+(-0.0699607581992)*x_ref[2]
        ref[3,1,1,1]=(0.75449445591)*x_ref[0]**o+(0.722970800307)*x_ref[0]+(-0.582970919465)*x_ref[1]**o+(-0.868818538212)*x_ref[1]+(0.940067636181)*x_ref[2]**o+(-0.659469743617)*x_ref[2]
        ref[3,1,2,0]=(-0.391224924663)*x_ref[0]**o+(-0.768173177737)*x_ref[0]+(-0.856876347997)*x_ref[1]**o+(-0.0451180499647)*x_ref[1]+(0.198590679246)*x_ref[2]**o+(0.187329095476)*x_ref[2]
        ref[3,1,2,1]=(0.483309063092)*x_ref[0]**o+(-0.0777241026893)*x_ref[0]+(0.0438427457286)*x_ref[1]**o+(-0.73195704867)*x_ref[1]+(-0.439660745725)*x_ref[2]**o+(0.279501672861)*x_ref[2]
        ref[3,2,0,0]=(0.799935089139)*x_ref[0]**o+(-0.306860809857)*x_ref[0]+(0.169104975039)*x_ref[1]**o+(-0.3480717525)*x_ref[1]+(0.215231688)*x_ref[2]**o+(-0.528775376945)*x_ref[2]
        ref[3,2,0,1]=(0.917815789587)*x_ref[0]**o+(0.0672910338102)*x_ref[0]+(-0.186550440011)*x_ref[1]**o+(0.922579714252)*x_ref[1]+(0.924162379857)*x_ref[2]**o+(-0.769320483618)*x_ref[2]
        ref[3,2,1,0]=(0.604324468058)*x_ref[0]**o+(-0.0109731419053)*x_ref[0]+(0.229447318325)*x_ref[1]**o+(-0.480693435232)*x_ref[1]+(-0.881779256999)*x_ref[2]**o+(0.288656394771)*x_ref[2]
        ref[3,2,1,1]=(0.934680723352)*x_ref[0]**o+(0.993838528376)*x_ref[0]+(0.957781588502)*x_ref[1]**o+(-0.859837057866)*x_ref[1]+(-0.702139964493)*x_ref[2]**o+(0.0538177124609)*x_ref[2]
        ref[3,2,2,0]=(-0.142809047814)*x_ref[0]**o+(-0.472594415134)*x_ref[0]+(0.0381782768377)*x_ref[1]**o+(0.727331491769)*x_ref[1]+(0.0910992897581)*x_ref[2]**o+(-0.845316063851)*x_ref[2]
        ref[3,2,2,1]=(-0.217483358272)*x_ref[0]**o+(0.991535092285)*x_ref[0]+(-0.399175896468)*x_ref[1]**o+(-0.665836512895)*x_ref[1]+(0.0983700911493)*x_ref[2]**o+(0.321811300155)*x_ref[2]
        ref[3,3,0,0]=(0.797049585464)*x_ref[0]**o+(-0.724467959299)*x_ref[0]+(0.19971589973)*x_ref[1]**o+(0.845928461864)*x_ref[1]+(-0.365158866359)*x_ref[2]**o+(-0.698274270132)*x_ref[2]
        ref[3,3,0,1]=(-0.638474704878)*x_ref[0]**o+(0.551866433352)*x_ref[0]+(0.192309452979)*x_ref[1]**o+(-0.703649325501)*x_ref[1]+(0.823910578138)*x_ref[2]**o+(-0.917446989634)*x_ref[2]
        ref[3,3,1,0]=(0.217577282169)*x_ref[0]**o+(0.521716803668)*x_ref[0]+(-0.116587783856)*x_ref[1]**o+(0.529672391417)*x_ref[1]+(-0.798918158284)*x_ref[2]**o+(-0.0996139511955)*x_ref[2]
        ref[3,3,1,1]=(0.498965049079)*x_ref[0]**o+(0.316489921186)*x_ref[0]+(-0.0988355471817)*x_ref[1]**o+(-0.0477508287681)*x_ref[1]+(-0.778914523993)*x_ref[2]**o+(0.189397812723)*x_ref[2]
        ref[3,3,2,0]=(0.203823698152)*x_ref[0]**o+(-0.340499414266)*x_ref[0]+(0.353374693517)*x_ref[1]**o+(-0.91035264494)*x_ref[1]+(-0.779204084198)*x_ref[2]**o+(-0.00276454165248)*x_ref[2]
        ref[3,3,2,1]=(0.295232007123)*x_ref[0]**o+(0.271766544502)*x_ref[0]+(0.431956430336)*x_ref[1]**o+(0.896481311536)*x_ref[1]+(-0.325014902487)*x_ref[2]**o+(0.70836170284)*x_ref[2]
        ref[3,4,0,0]=(-0.407535202468)*x_ref[0]**o+(-0.33055681091)*x_ref[0]+(0.129762143151)*x_ref[1]**o+(0.352087436228)*x_ref[1]+(-0.538770401565)*x_ref[2]**o+(-0.519071077045)*x_ref[2]
        ref[3,4,0,1]=(0.846011435499)*x_ref[0]**o+(0.172822133672)*x_ref[0]+(-0.740732553847)*x_ref[1]**o+(0.230062837231)*x_ref[1]+(-0.231810007237)*x_ref[2]**o+(-0.568366338923)*x_ref[2]
        ref[3,4,1,0]=(-0.198091119057)*x_ref[0]**o+(0.957989508794)*x_ref[0]+(0.861739071175)*x_ref[1]**o+(0.923450697883)*x_ref[1]+(-0.755262504937)*x_ref[2]**o+(0.102897579936)*x_ref[2]
        ref[3,4,1,1]=(0.198829545353)*x_ref[0]**o+(-0.757182006818)*x_ref[0]+(0.254092337808)*x_ref[1]**o+(-0.00815072620097)*x_ref[1]+(0.347421988308)*x_ref[2]**o+(-0.0215277616923)*x_ref[2]
        ref[3,4,2,0]=(0.775062112529)*x_ref[0]**o+(-0.77746792774)*x_ref[0]+(-0.666490294869)*x_ref[1]**o+(-0.679971069878)*x_ref[1]+(-0.972450779894)*x_ref[2]**o+(-0.705322282477)*x_ref[2]
        ref[3,4,2,1]=(-0.0994927587382)*x_ref[0]**o+(0.293071774786)*x_ref[0]+(0.370784399397)*x_ref[1]**o+(0.931411225946)*x_ref[1]+(-0.55789163294)*x_ref[2]**o+(0.0831890658592)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(1.66854814732)*x[0]+(1.38199165128)*x[1]
        ref=(1.66854814732)*x_ref[0]+(1.38199165128)*x_ref[1]
      else:
        arg=(-0.426142624852)*x[0]+(0.0492229050252)*x[1]+(-0.821661899519)*x[2]
        ref=(-0.426142624852)*x_ref[0]+(0.0492229050252)*x_ref[1]+(-0.821661899519)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        arg[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
        ref[0]=(0.287050581427)*x[0]+(0.0745644988438)*x[1]
        ref[1]=(-0.279090829341)*x[0]+(-0.100758565381)*x[1]
      else:
        arg[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        arg[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
        ref[0]=(-0.367806366245)*x[0]+(0.0185557433717)*x[1]+(1.14806419827)*x[2]
        ref[1]=(-0.588016618937)*x[0]+(-1.03057423712)*x[1]+(-0.0359765566708)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.178459116054)*x[0]+(-0.908190366917)*x[1]
        arg[0,1]=(-0.365861276878)*x[0]+(-1.57557211363)*x[1]
        arg[0,2]=(-1.22266296874)*x[0]+(0.476090041507)*x[1]
        arg[0,3]=(0.0677955238995)*x[0]+(-0.899620825938)*x[1]
        arg[0,4]=(-0.29315929952)*x[0]+(-0.0589414650397)*x[1]
        arg[1,0]=(-0.510486453858)*x[0]+(0.960724573137)*x[1]
        arg[1,1]=(-0.316955543)*x[0]+(-0.160275601404)*x[1]
        arg[1,2]=(-0.960162300557)*x[0]+(0.426685726779)*x[1]
        arg[1,3]=(-0.458645373633)*x[0]+(0.623646469232)*x[1]
        arg[1,4]=(-0.831993675808)*x[0]+(-0.380162434858)*x[1]
        arg[2,0]=(-0.0614446129662)*x[0]+(0.442134320884)*x[1]
        arg[2,1]=(-0.0466135888936)*x[0]+(0.465021672347)*x[1]
        arg[2,2]=(-1.69520946599)*x[0]+(-1.69200166455)*x[1]
        arg[2,3]=(0.928809309584)*x[0]+(-0.344012063367)*x[1]
        arg[2,4]=(0.557754072268)*x[0]+(-0.0104949991147)*x[1]
        arg[3,0]=(0.0683969753885)*x[0]+(1.59892831147)*x[1]
        arg[3,1]=(0.90059768275)*x[0]+(1.4136644906)*x[1]
        arg[3,2]=(-0.397873855228)*x[0]+(-0.876192378488)*x[1]
        arg[3,3]=(-1.71837532502)*x[0]+(-0.911585940448)*x[1]
        arg[3,4]=(-1.619565287)*x[0]+(-0.911830282675)*x[1]
        ref[0,0]=(-0.178459116054)*x_ref[0]+(-0.908190366917)*x_ref[1]
        ref[0,1]=(-0.365861276878)*x_ref[0]+(-1.57557211363)*x_ref[1]
        ref[0,2]=(-1.22266296874)*x_ref[0]+(0.476090041507)*x_ref[1]
        ref[0,3]=(0.0677955238995)*x_ref[0]+(-0.899620825938)*x_ref[1]
        ref[0,4]=(-0.29315929952)*x_ref[0]+(-0.0589414650397)*x_ref[1]
        ref[1,0]=(-0.510486453858)*x_ref[0]+(0.960724573137)*x_ref[1]
        ref[1,1]=(-0.316955543)*x_ref[0]+(-0.160275601404)*x_ref[1]
        ref[1,2]=(-0.960162300557)*x_ref[0]+(0.426685726779)*x_ref[1]
        ref[1,3]=(-0.458645373633)*x_ref[0]+(0.623646469232)*x_ref[1]
        ref[1,4]=(-0.831993675808)*x_ref[0]+(-0.380162434858)*x_ref[1]
        ref[2,0]=(-0.0614446129662)*x_ref[0]+(0.442134320884)*x_ref[1]
        ref[2,1]=(-0.0466135888936)*x_ref[0]+(0.465021672347)*x_ref[1]
        ref[2,2]=(-1.69520946599)*x_ref[0]+(-1.69200166455)*x_ref[1]
        ref[2,3]=(0.928809309584)*x_ref[0]+(-0.344012063367)*x_ref[1]
        ref[2,4]=(0.557754072268)*x_ref[0]+(-0.0104949991147)*x_ref[1]
        ref[3,0]=(0.0683969753885)*x_ref[0]+(1.59892831147)*x_ref[1]
        ref[3,1]=(0.90059768275)*x_ref[0]+(1.4136644906)*x_ref[1]
        ref[3,2]=(-0.397873855228)*x_ref[0]+(-0.876192378488)*x_ref[1]
        ref[3,3]=(-1.71837532502)*x_ref[0]+(-0.911585940448)*x_ref[1]
        ref[3,4]=(-1.619565287)*x_ref[0]+(-0.911830282675)*x_ref[1]
      else:
        arg[0,0]=(-0.822599790066)*x[0]+(-1.22492139307)*x[1]+(-0.696878025616)*x[2]
        arg[0,1]=(0.600762284033)*x[0]+(-0.157734608915)*x[1]+(-0.729304703393)*x[2]
        arg[0,2]=(-1.42910707426)*x[0]+(1.46395956796)*x[1]+(1.38800770798)*x[2]
        arg[0,3]=(1.2198448221)*x[0]+(0.0898753488356)*x[1]+(0.976038279019)*x[2]
        arg[0,4]=(-0.616927403743)*x[0]+(1.41159934363)*x[1]+(-1.74783039584)*x[2]
        arg[1,0]=(-0.101387750238)*x[0]+(-0.294997280251)*x[1]+(0.247605331103)*x[2]
        arg[1,1]=(-0.320864335195)*x[0]+(-0.202828013922)*x[1]+(-0.366246434001)*x[2]
        arg[1,2]=(-0.827685495006)*x[0]+(-0.236835428287)*x[1]+(-1.48668813903)*x[2]
        arg[1,3]=(0.56131602615)*x[0]+(-0.572858337349)*x[1]+(0.701775515666)*x[2]
        arg[1,4]=(-1.09228098916)*x[0]+(-0.0251387524711)*x[1]+(1.91587136325)*x[2]
        arg[2,0]=(-0.0758503728265)*x[0]+(0.236248603718)*x[1]+(0.0446858988194)*x[2]
        arg[2,1]=(-1.1745073907)*x[0]+(-0.519328441654)*x[1]+(-1.19738117811)*x[2]
        arg[2,2]=(-0.018812738143)*x[0]+(1.00051415805)*x[1]+(-0.620776442731)*x[2]
        arg[2,3]=(-0.430333705208)*x[0]+(0.136368267739)*x[1]+(0.59767563861)*x[2]
        arg[2,4]=(1.54336203988)*x[0]+(0.106189845211)*x[1]+(1.53097571242)*x[2]
        arg[3,0]=(-0.00451956527502)*x[0]+(0.710318142757)*x[1]+(0.386947317474)*x[2]
        arg[3,1]=(1.22013531098)*x[0]+(-0.735641779691)*x[1]+(-0.0137925239152)*x[2]
        arg[3,2]=(0.0949363128405)*x[0]+(-0.0778790744582)*x[1]+(-0.466004036944)*x[2]
        arg[3,3]=(0.656445527911)*x[0]+(1.67275242688)*x[1]+(-0.0946012389031)*x[2]
        arg[3,4]=(0.927582093503)*x[0]+(-0.812773507566)*x[1]+(1.14365402059)*x[2]
        ref[0,0]=(-0.822599790066)*x_ref[0]+(-1.22492139307)*x_ref[1]+(-0.696878025616)*x_ref[2]
        ref[0,1]=(0.600762284033)*x_ref[0]+(-0.157734608915)*x_ref[1]+(-0.729304703393)*x_ref[2]
        ref[0,2]=(-1.42910707426)*x_ref[0]+(1.46395956796)*x_ref[1]+(1.38800770798)*x_ref[2]
        ref[0,3]=(1.2198448221)*x_ref[0]+(0.0898753488356)*x_ref[1]+(0.976038279019)*x_ref[2]
        ref[0,4]=(-0.616927403743)*x_ref[0]+(1.41159934363)*x_ref[1]+(-1.74783039584)*x_ref[2]
        ref[1,0]=(-0.101387750238)*x_ref[0]+(-0.294997280251)*x_ref[1]+(0.247605331103)*x_ref[2]
        ref[1,1]=(-0.320864335195)*x_ref[0]+(-0.202828013922)*x_ref[1]+(-0.366246434001)*x_ref[2]
        ref[1,2]=(-0.827685495006)*x_ref[0]+(-0.236835428287)*x_ref[1]+(-1.48668813903)*x_ref[2]
        ref[1,3]=(0.56131602615)*x_ref[0]+(-0.572858337349)*x_ref[1]+(0.701775515666)*x_ref[2]
        ref[1,4]=(-1.09228098916)*x_ref[0]+(-0.0251387524711)*x_ref[1]+(1.91587136325)*x_ref[2]
        ref[2,0]=(-0.0758503728265)*x_ref[0]+(0.236248603718)*x_ref[1]+(0.0446858988194)*x_ref[2]
        ref[2,1]=(-1.1745073907)*x_ref[0]+(-0.519328441654)*x_ref[1]+(-1.19738117811)*x_ref[2]
        ref[2,2]=(-0.018812738143)*x_ref[0]+(1.00051415805)*x_ref[1]+(-0.620776442731)*x_ref[2]
        ref[2,3]=(-0.430333705208)*x_ref[0]+(0.136368267739)*x_ref[1]+(0.59767563861)*x_ref[2]
        ref[2,4]=(1.54336203988)*x_ref[0]+(0.106189845211)*x_ref[1]+(1.53097571242)*x_ref[2]
        ref[3,0]=(-0.00451956527502)*x_ref[0]+(0.710318142757)*x_ref[1]+(0.386947317474)*x_ref[2]
        ref[3,1]=(1.22013531098)*x_ref[0]+(-0.735641779691)*x_ref[1]+(-0.0137925239152)*x_ref[2]
        ref[3,2]=(0.0949363128405)*x_ref[0]+(-0.0778790744582)*x_ref[1]+(-0.466004036944)*x_ref[2]
        ref[3,3]=(0.656445527911)*x_ref[0]+(1.67275242688)*x_ref[1]+(-0.0946012389031)*x_ref[2]
        ref[3,4]=(0.927582093503)*x_ref[0]+(-0.812773507566)*x_ref[1]+(1.14365402059)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(1.53783653017)*x[0]+(-0.262380397029)*x[1]
        arg[0,0,1]=(0.0150173886328)*x[0]+(1.16643858096)*x[1]
        arg[0,1,0]=(0.886211915116)*x[0]+(0.0707779287946)*x[1]
        arg[0,1,1]=(-1.38689895895)*x[0]+(0.970706440444)*x[1]
        arg[1,0,0]=(-0.923778075186)*x[0]+(1.93261160466)*x[1]
        arg[1,0,1]=(1.03640566039)*x[0]+(-0.511763660443)*x[1]
        arg[1,1,0]=(0.127401766462)*x[0]+(-0.472120435035)*x[1]
        arg[1,1,1]=(-0.366402024047)*x[0]+(-0.396644605167)*x[1]
        arg[2,0,0]=(-0.715041877609)*x[0]+(-1.00214766039)*x[1]
        arg[2,0,1]=(-1.52452556675)*x[0]+(-0.570352600973)*x[1]
        arg[2,1,0]=(0.730426967288)*x[0]+(0.158290182237)*x[1]
        arg[2,1,1]=(-0.152190333299)*x[0]+(-0.254128104593)*x[1]
        arg[3,0,0]=(0.835319813606)*x[0]+(-0.142845950855)*x[1]
        arg[3,0,1]=(0.80118569153)*x[0]+(0.260295390539)*x[1]
        arg[3,1,0]=(0.283151732104)*x[0]+(0.0284945097388)*x[1]
        arg[3,1,1]=(-0.715345696074)*x[0]+(0.563181102176)*x[1]
        arg[4,0,0]=(-0.481733499899)*x[0]+(1.30638761744)*x[1]
        arg[4,0,1]=(-0.515540152341)*x[0]+(1.47724553596)*x[1]
        arg[4,1,0]=(0.631902493814)*x[0]+(0.279373852162)*x[1]
        arg[4,1,1]=(0.666932982001)*x[0]+(1.02356936117)*x[1]
        arg[5,0,0]=(-1.15844673154)*x[0]+(-0.472595660401)*x[1]
        arg[5,0,1]=(-0.469626313646)*x[0]+(0.464876870861)*x[1]
        arg[5,1,0]=(-0.0934345630024)*x[0]+(0.92605991751)*x[1]
        arg[5,1,1]=(-0.0528050605159)*x[0]+(0.984329702885)*x[1]
        ref[0,0,0]=(1.53783653017)*x_ref[0]+(-0.262380397029)*x_ref[1]
        ref[0,0,1]=(0.0150173886328)*x_ref[0]+(1.16643858096)*x_ref[1]
        ref[0,1,0]=(0.886211915116)*x_ref[0]+(0.0707779287946)*x_ref[1]
        ref[0,1,1]=(-1.38689895895)*x_ref[0]+(0.970706440444)*x_ref[1]
        ref[1,0,0]=(-0.923778075186)*x_ref[0]+(1.93261160466)*x_ref[1]
        ref[1,0,1]=(1.03640566039)*x_ref[0]+(-0.511763660443)*x_ref[1]
        ref[1,1,0]=(0.127401766462)*x_ref[0]+(-0.472120435035)*x_ref[1]
        ref[1,1,1]=(-0.366402024047)*x_ref[0]+(-0.396644605167)*x_ref[1]
        ref[2,0,0]=(-0.715041877609)*x_ref[0]+(-1.00214766039)*x_ref[1]
        ref[2,0,1]=(-1.52452556675)*x_ref[0]+(-0.570352600973)*x_ref[1]
        ref[2,1,0]=(0.730426967288)*x_ref[0]+(0.158290182237)*x_ref[1]
        ref[2,1,1]=(-0.152190333299)*x_ref[0]+(-0.254128104593)*x_ref[1]
        ref[3,0,0]=(0.835319813606)*x_ref[0]+(-0.142845950855)*x_ref[1]
        ref[3,0,1]=(0.80118569153)*x_ref[0]+(0.260295390539)*x_ref[1]
        ref[3,1,0]=(0.283151732104)*x_ref[0]+(0.0284945097388)*x_ref[1]
        ref[3,1,1]=(-0.715345696074)*x_ref[0]+(0.563181102176)*x_ref[1]
        ref[4,0,0]=(-0.481733499899)*x_ref[0]+(1.30638761744)*x_ref[1]
        ref[4,0,1]=(-0.515540152341)*x_ref[0]+(1.47724553596)*x_ref[1]
        ref[4,1,0]=(0.631902493814)*x_ref[0]+(0.279373852162)*x_ref[1]
        ref[4,1,1]=(0.666932982001)*x_ref[0]+(1.02356936117)*x_ref[1]
        ref[5,0,0]=(-1.15844673154)*x_ref[0]+(-0.472595660401)*x_ref[1]
        ref[5,0,1]=(-0.469626313646)*x_ref[0]+(0.464876870861)*x_ref[1]
        ref[5,1,0]=(-0.0934345630024)*x_ref[0]+(0.92605991751)*x_ref[1]
        ref[5,1,1]=(-0.0528050605159)*x_ref[0]+(0.984329702885)*x_ref[1]
      else:
        arg[0,0,0]=(-0.713551995443)*x[0]+(0.694017368402)*x[1]+(-0.0738591792901)*x[2]
        arg[0,0,1]=(1.56876736864)*x[0]+(-0.345130721117)*x[1]+(1.1690777905)*x[2]
        arg[0,1,0]=(-0.62505379409)*x[0]+(-0.143697617605)*x[1]+(0.132102288091)*x[2]
        arg[0,1,1]=(0.873577158168)*x[0]+(-0.967200786585)*x[1]+(1.2903546666)*x[2]
        arg[1,0,0]=(-0.192658558584)*x[0]+(0.506709431455)*x[1]+(-0.0952065550445)*x[2]
        arg[1,0,1]=(1.30984695498)*x[0]+(-0.469661654194)*x[1]+(1.39212979011)*x[2]
        arg[1,1,0]=(-0.49094299988)*x[0]+(1.02571925005)*x[1]+(0.125396191195)*x[2]
        arg[1,1,1]=(-0.062798346747)*x[0]+(-0.789359897908)*x[1]+(1.77093308656)*x[2]
        arg[2,0,0]=(1.61875859199)*x[0]+(0.594855189809)*x[1]+(0.42445154554)*x[2]
        arg[2,0,1]=(-0.856240554561)*x[0]+(0.100699001565)*x[1]+(-0.27854306919)*x[2]
        arg[2,1,0]=(0.0478958246378)*x[0]+(0.695175069462)*x[1]+(-0.847667978286)*x[2]
        arg[2,1,1]=(1.60820982887)*x[0]+(-0.251793277319)*x[1]+(0.143173366678)*x[2]
        arg[3,0,0]=(-0.292866056474)*x[0]+(1.01748459143)*x[1]+(-0.484686028756)*x[2]
        arg[3,0,1]=(-0.408126207016)*x[0]+(-1.31011612654)*x[1]+(-0.955847365219)*x[2]
        arg[3,1,0]=(0.0396674474263)*x[0]+(0.354470432765)*x[1]+(-0.291731786972)*x[2]
        arg[3,1,1]=(0.752098712767)*x[0]+(-1.83970538813)*x[1]+(0.0227500647892)*x[2]
        arg[4,0,0]=(0.398849520011)*x[0]+(-1.38237112549)*x[1]+(-0.800550650133)*x[2]
        arg[4,0,1]=(0.102399478593)*x[0]+(0.812513136316)*x[1]+(1.25442670221)*x[2]
        arg[4,1,0]=(0.224914729031)*x[0]+(0.0706934334606)*x[1]+(0.640265153974)*x[2]
        arg[4,1,1]=(0.278731023964)*x[0]+(-0.766892970502)*x[1]+(1.62378242532)*x[2]
        arg[5,0,0]=(-0.874002481668)*x[0]+(0.168391989524)*x[1]+(-0.0150981524408)*x[2]
        arg[5,0,1]=(1.03369744159)*x[0]+(-0.363152709617)*x[1]+(-0.429141439899)*x[2]
        arg[5,1,0]=(1.74293672288)*x[0]+(1.91064523276)*x[1]+(0.159389690962)*x[2]
        arg[5,1,1]=(0.769432577623)*x[0]+(0.929942365281)*x[1]+(0.170908152574)*x[2]
        ref[0,0,0]=(-0.713551995443)*x_ref[0]+(0.694017368402)*x_ref[1]+(-0.0738591792901)*x_ref[2]
        ref[0,0,1]=(1.56876736864)*x_ref[0]+(-0.345130721117)*x_ref[1]+(1.1690777905)*x_ref[2]
        ref[0,1,0]=(-0.62505379409)*x_ref[0]+(-0.143697617605)*x_ref[1]+(0.132102288091)*x_ref[2]
        ref[0,1,1]=(0.873577158168)*x_ref[0]+(-0.967200786585)*x_ref[1]+(1.2903546666)*x_ref[2]
        ref[1,0,0]=(-0.192658558584)*x_ref[0]+(0.506709431455)*x_ref[1]+(-0.0952065550445)*x_ref[2]
        ref[1,0,1]=(1.30984695498)*x_ref[0]+(-0.469661654194)*x_ref[1]+(1.39212979011)*x_ref[2]
        ref[1,1,0]=(-0.49094299988)*x_ref[0]+(1.02571925005)*x_ref[1]+(0.125396191195)*x_ref[2]
        ref[1,1,1]=(-0.062798346747)*x_ref[0]+(-0.789359897908)*x_ref[1]+(1.77093308656)*x_ref[2]
        ref[2,0,0]=(1.61875859199)*x_ref[0]+(0.594855189809)*x_ref[1]+(0.42445154554)*x_ref[2]
        ref[2,0,1]=(-0.856240554561)*x_ref[0]+(0.100699001565)*x_ref[1]+(-0.27854306919)*x_ref[2]
        ref[2,1,0]=(0.0478958246378)*x_ref[0]+(0.695175069462)*x_ref[1]+(-0.847667978286)*x_ref[2]
        ref[2,1,1]=(1.60820982887)*x_ref[0]+(-0.251793277319)*x_ref[1]+(0.143173366678)*x_ref[2]
        ref[3,0,0]=(-0.292866056474)*x_ref[0]+(1.01748459143)*x_ref[1]+(-0.484686028756)*x_ref[2]
        ref[3,0,1]=(-0.408126207016)*x_ref[0]+(-1.31011612654)*x_ref[1]+(-0.955847365219)*x_ref[2]
        ref[3,1,0]=(0.0396674474263)*x_ref[0]+(0.354470432765)*x_ref[1]+(-0.291731786972)*x_ref[2]
        ref[3,1,1]=(0.752098712767)*x_ref[0]+(-1.83970538813)*x_ref[1]+(0.0227500647892)*x_ref[2]
        ref[4,0,0]=(0.398849520011)*x_ref[0]+(-1.38237112549)*x_ref[1]+(-0.800550650133)*x_ref[2]
        ref[4,0,1]=(0.102399478593)*x_ref[0]+(0.812513136316)*x_ref[1]+(1.25442670221)*x_ref[2]
        ref[4,1,0]=(0.224914729031)*x_ref[0]+(0.0706934334606)*x_ref[1]+(0.640265153974)*x_ref[2]
        ref[4,1,1]=(0.278731023964)*x_ref[0]+(-0.766892970502)*x_ref[1]+(1.62378242532)*x_ref[2]
        ref[5,0,0]=(-0.874002481668)*x_ref[0]+(0.168391989524)*x_ref[1]+(-0.0150981524408)*x_ref[2]
        ref[5,0,1]=(1.03369744159)*x_ref[0]+(-0.363152709617)*x_ref[1]+(-0.429141439899)*x_ref[2]
        ref[5,1,0]=(1.74293672288)*x_ref[0]+(1.91064523276)*x_ref[1]+(0.159389690962)*x_ref[2]
        ref[5,1,1]=(0.769432577623)*x_ref[0]+(0.929942365281)*x_ref[1]+(0.170908152574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-1.15521896363)*x[0]+(0.8731619129)*x[1]
        arg[0,0,0,1]=(0.597656443464)*x[0]+(0.341688152811)*x[1]
        arg[0,0,1,0]=(-0.422350580427)*x[0]+(0.728011338687)*x[1]
        arg[0,0,1,1]=(-1.72046294064)*x[0]+(1.13345636724)*x[1]
        arg[0,0,2,0]=(0.150530941523)*x[0]+(0.437417461641)*x[1]
        arg[0,0,2,1]=(-0.413133183531)*x[0]+(-0.268630939964)*x[1]
        arg[0,1,0,0]=(-0.220368905416)*x[0]+(1.10085894684)*x[1]
        arg[0,1,0,1]=(-0.943367491188)*x[0]+(0.80942793885)*x[1]
        arg[0,1,1,0]=(0.906220152303)*x[0]+(0.801214136703)*x[1]
        arg[0,1,1,1]=(-1.3650200124)*x[0]+(-0.373454973523)*x[1]
        arg[0,1,2,0]=(-0.277067377427)*x[0]+(-1.62229711799)*x[1]
        arg[0,1,2,1]=(0.787076886379)*x[0]+(0.496906919368)*x[1]
        arg[0,2,0,0]=(-0.0770154256748)*x[0]+(1.5002648939)*x[1]
        arg[0,2,0,1]=(0.225572553862)*x[0]+(1.39471337012)*x[1]
        arg[0,2,1,0]=(-0.54799237637)*x[0]+(-0.607785237418)*x[1]
        arg[0,2,1,1]=(0.384591801229)*x[0]+(1.30760620469)*x[1]
        arg[0,2,2,0]=(0.228248139163)*x[0]+(0.151268778109)*x[1]
        arg[0,2,2,1]=(-1.43978230199)*x[0]+(0.958903347524)*x[1]
        arg[0,3,0,0]=(-0.568133098072)*x[0]+(0.128510595507)*x[1]
        arg[0,3,0,1]=(-0.304029315535)*x[0]+(1.19478464847)*x[1]
        arg[0,3,1,0]=(0.0996876882275)*x[0]+(-0.255346045102)*x[1]
        arg[0,3,1,1]=(0.414708254109)*x[0]+(-0.148700489536)*x[1]
        arg[0,3,2,0]=(-0.740719850161)*x[0]+(0.375907815982)*x[1]
        arg[0,3,2,1]=(-0.184189058365)*x[0]+(-0.274477516444)*x[1]
        arg[0,4,0,0]=(-0.565848858136)*x[0]+(-0.206818417121)*x[1]
        arg[0,4,0,1]=(-0.983652399085)*x[0]+(1.21445827986)*x[1]
        arg[0,4,1,0]=(1.6425686123)*x[0]+(-0.621952337654)*x[1]
        arg[0,4,1,1]=(-1.07526530849)*x[0]+(-0.0915810578715)*x[1]
        arg[0,4,2,0]=(1.4812278738)*x[0]+(1.61991326423)*x[1]
        arg[0,4,2,1]=(-1.4147420413)*x[0]+(1.34428787131)*x[1]
        arg[1,0,0,0]=(-0.910373659777)*x[0]+(1.46920176435)*x[1]
        arg[1,0,0,1]=(0.074473571609)*x[0]+(-0.336642181231)*x[1]
        arg[1,0,1,0]=(-0.141813739296)*x[0]+(-0.534432398621)*x[1]
        arg[1,0,1,1]=(0.728030938984)*x[0]+(0.532542738961)*x[1]
        arg[1,0,2,0]=(0.20676316374)*x[0]+(-0.065235999066)*x[1]
        arg[1,0,2,1]=(-1.09054680794)*x[0]+(0.537754476805)*x[1]
        arg[1,1,0,0]=(0.523878896271)*x[0]+(-0.0561828109964)*x[1]
        arg[1,1,0,1]=(1.27754653741)*x[0]+(0.119734333004)*x[1]
        arg[1,1,1,0]=(0.0933160885942)*x[0]+(0.390183021193)*x[1]
        arg[1,1,1,1]=(0.548273953924)*x[0]+(-0.552588185124)*x[1]
        arg[1,1,2,0]=(0.576803722241)*x[0]+(0.60898781381)*x[1]
        arg[1,1,2,1]=(-0.307231650547)*x[0]+(0.439111970656)*x[1]
        arg[1,2,0,0]=(0.172113300736)*x[0]+(-0.0153971432613)*x[1]
        arg[1,2,0,1]=(-0.883753634465)*x[0]+(1.35320294175)*x[1]
        arg[1,2,1,0]=(0.153937139638)*x[0]+(-0.865654430739)*x[1]
        arg[1,2,1,1]=(0.970377377543)*x[0]+(-1.11418225395)*x[1]
        arg[1,2,2,0]=(-1.11106779953)*x[0]+(-0.357700480625)*x[1]
        arg[1,2,2,1]=(0.475529404147)*x[0]+(-0.121369909464)*x[1]
        arg[1,3,0,0]=(-0.867370155191)*x[0]+(0.625161512036)*x[1]
        arg[1,3,0,1]=(0.252873452474)*x[0]+(1.11368263907)*x[1]
        arg[1,3,1,0]=(-1.25698605702)*x[0]+(-0.381254857218)*x[1]
        arg[1,3,1,1]=(1.12038052674)*x[0]+(-0.113429410248)*x[1]
        arg[1,3,2,0]=(0.0821799570201)*x[0]+(-0.167383600252)*x[1]
        arg[1,3,2,1]=(-0.534217900599)*x[0]+(-1.26144431009)*x[1]
        arg[1,4,0,0]=(0.639186045888)*x[0]+(0.258686782482)*x[1]
        arg[1,4,0,1]=(-0.256515011769)*x[0]+(0.797453395844)*x[1]
        arg[1,4,1,0]=(-0.427518421999)*x[0]+(-0.101439725502)*x[1]
        arg[1,4,1,1]=(-0.251303248239)*x[0]+(0.361194806374)*x[1]
        arg[1,4,2,0]=(0.415313798163)*x[0]+(-0.909602300262)*x[1]
        arg[1,4,2,1]=(0.118542894027)*x[0]+(-0.28224555744)*x[1]
        arg[2,0,0,0]=(0.700687096184)*x[0]+(-1.69964436969)*x[1]
        arg[2,0,0,1]=(0.880393551605)*x[0]+(-0.571112060462)*x[1]
        arg[2,0,1,0]=(1.35990706973)*x[0]+(0.260154297077)*x[1]
        arg[2,0,1,1]=(0.294201987212)*x[0]+(-0.379007248662)*x[1]
        arg[2,0,2,0]=(-1.42914701121)*x[0]+(0.88289145627)*x[1]
        arg[2,0,2,1]=(-0.343976881604)*x[0]+(0.0642493086788)*x[1]
        arg[2,1,0,0]=(0.479103547757)*x[0]+(-1.60276044525)*x[1]
        arg[2,1,0,1]=(0.944321957371)*x[0]+(0.313721173145)*x[1]
        arg[2,1,1,0]=(0.233589885494)*x[0]+(-0.427865810692)*x[1]
        arg[2,1,1,1]=(-0.954518596356)*x[0]+(-0.153033245523)*x[1]
        arg[2,1,2,0]=(1.14079619235)*x[0]+(0.136503305207)*x[1]
        arg[2,1,2,1]=(-0.715601080868)*x[0]+(-1.74677798058)*x[1]
        arg[2,2,0,0]=(0.211908166714)*x[0]+(0.379390512119)*x[1]
        arg[2,2,0,1]=(0.97243389801)*x[0]+(-0.988193298566)*x[1]
        arg[2,2,1,0]=(0.8990152555)*x[0]+(0.22398476583)*x[1]
        arg[2,2,1,1]=(-0.748858338529)*x[0]+(-1.00255098852)*x[1]
        arg[2,2,2,0]=(0.863334969354)*x[0]+(1.14334958607)*x[1]
        arg[2,2,2,1]=(-0.807672728362)*x[0]+(0.837576150239)*x[1]
        arg[2,3,0,0]=(-0.310471066952)*x[0]+(0.261468387883)*x[1]
        arg[2,3,0,1]=(-1.61325161386)*x[0]+(0.016614907729)*x[1]
        arg[2,3,1,0]=(-0.742577714565)*x[0]+(-1.54936077725)*x[1]
        arg[2,3,1,1]=(-0.393825799694)*x[0]+(0.66004298074)*x[1]
        arg[2,3,2,0]=(-1.03965632182)*x[0]+(-0.863160703719)*x[1]
        arg[2,3,2,1]=(0.673788653196)*x[0]+(0.740311339524)*x[1]
        arg[2,4,0,0]=(1.59272250531)*x[0]+(1.27845229216)*x[1]
        arg[2,4,0,1]=(0.112947869874)*x[0]+(1.06626388923)*x[1]
        arg[2,4,1,0]=(-0.935174560087)*x[0]+(0.729112002619)*x[1]
        arg[2,4,1,1]=(1.28167575095)*x[0]+(-1.6626881011)*x[1]
        arg[2,4,2,0]=(-1.01382997014)*x[0]+(0.40870705964)*x[1]
        arg[2,4,2,1]=(-0.903293274454)*x[0]+(0.451662927854)*x[1]
        arg[3,0,0,0]=(0.414933474895)*x[0]+(0.274208992931)*x[1]
        arg[3,0,0,1]=(-0.140384007335)*x[0]+(-0.201270983912)*x[1]
        arg[3,0,1,0]=(-1.34339529394)*x[0]+(-1.26275414913)*x[1]
        arg[3,0,1,1]=(0.34482759487)*x[0]+(0.992140984097)*x[1]
        arg[3,0,2,0]=(-1.56575003976)*x[0]+(-0.514557317682)*x[1]
        arg[3,0,2,1]=(-0.203196026518)*x[0]+(-0.50777550196)*x[1]
        arg[3,1,0,0]=(-0.465027052452)*x[0]+(0.165299984247)*x[1]
        arg[3,1,0,1]=(-0.876537945705)*x[0]+(0.569062140702)*x[1]
        arg[3,1,1,0]=(0.212854890028)*x[0]+(0.772436691343)*x[1]
        arg[3,1,1,1]=(0.376912060989)*x[0]+(-1.59078116917)*x[1]
        arg[3,1,2,0]=(-1.19949528387)*x[0]+(0.508138309995)*x[1]
        arg[3,1,2,1]=(0.535067293288)*x[0]+(0.483486234966)*x[1]
        arg[3,2,0,0]=(-0.366454519079)*x[0]+(0.880937088999)*x[1]
        arg[3,2,0,1]=(-0.164158778522)*x[0]+(-0.0933716527947)*x[1]
        arg[3,2,1,0]=(-0.338148774302)*x[0]+(0.389103301223)*x[1]
        arg[3,2,1,1]=(-1.17323344849)*x[0]+(-0.286752347393)*x[1]
        arg[3,2,2,0]=(0.43651426746)*x[0]+(0.0751512323008)*x[1]
        arg[3,2,2,1]=(1.44242451962)*x[0]+(-0.853169055909)*x[1]
        arg[3,3,0,0]=(0.268862750262)*x[0]+(0.668643005784)*x[1]
        arg[3,3,0,1]=(-1.47078437928)*x[0]+(1.25865018681)*x[1]
        arg[3,3,1,0]=(-0.218849811361)*x[0]+(0.147601849599)*x[1]
        arg[3,3,1,1]=(1.66493534918)*x[0]+(0.368711512357)*x[1]
        arg[3,3,2,0]=(0.259910144631)*x[0]+(-0.977975383491)*x[1]
        arg[3,3,2,1]=(1.87696180213)*x[0]+(-0.583679438031)*x[1]
        arg[3,4,0,0]=(0.226114159843)*x[0]+(0.121547181819)*x[1]
        arg[3,4,0,1]=(1.21215141305)*x[0]+(-0.175673309633)*x[1]
        arg[3,4,1,0]=(0.644340639306)*x[0]+(0.35372514848)*x[1]
        arg[3,4,1,1]=(0.452987488144)*x[0]+(0.838110546855)*x[1]
        arg[3,4,2,0]=(-0.434414511976)*x[0]+(-0.879038818613)*x[1]
        arg[3,4,2,1]=(-0.387240543602)*x[0]+(0.187795787877)*x[1]
        ref[0,0,0,0]=(-1.15521896363)*x_ref[0]+(0.8731619129)*x_ref[1]
        ref[0,0,0,1]=(0.597656443464)*x_ref[0]+(0.341688152811)*x_ref[1]
        ref[0,0,1,0]=(-0.422350580427)*x_ref[0]+(0.728011338687)*x_ref[1]
        ref[0,0,1,1]=(-1.72046294064)*x_ref[0]+(1.13345636724)*x_ref[1]
        ref[0,0,2,0]=(0.150530941523)*x_ref[0]+(0.437417461641)*x_ref[1]
        ref[0,0,2,1]=(-0.413133183531)*x_ref[0]+(-0.268630939964)*x_ref[1]
        ref[0,1,0,0]=(-0.220368905416)*x_ref[0]+(1.10085894684)*x_ref[1]
        ref[0,1,0,1]=(-0.943367491188)*x_ref[0]+(0.80942793885)*x_ref[1]
        ref[0,1,1,0]=(0.906220152303)*x_ref[0]+(0.801214136703)*x_ref[1]
        ref[0,1,1,1]=(-1.3650200124)*x_ref[0]+(-0.373454973523)*x_ref[1]
        ref[0,1,2,0]=(-0.277067377427)*x_ref[0]+(-1.62229711799)*x_ref[1]
        ref[0,1,2,1]=(0.787076886379)*x_ref[0]+(0.496906919368)*x_ref[1]
        ref[0,2,0,0]=(-0.0770154256748)*x_ref[0]+(1.5002648939)*x_ref[1]
        ref[0,2,0,1]=(0.225572553862)*x_ref[0]+(1.39471337012)*x_ref[1]
        ref[0,2,1,0]=(-0.54799237637)*x_ref[0]+(-0.607785237418)*x_ref[1]
        ref[0,2,1,1]=(0.384591801229)*x_ref[0]+(1.30760620469)*x_ref[1]
        ref[0,2,2,0]=(0.228248139163)*x_ref[0]+(0.151268778109)*x_ref[1]
        ref[0,2,2,1]=(-1.43978230199)*x_ref[0]+(0.958903347524)*x_ref[1]
        ref[0,3,0,0]=(-0.568133098072)*x_ref[0]+(0.128510595507)*x_ref[1]
        ref[0,3,0,1]=(-0.304029315535)*x_ref[0]+(1.19478464847)*x_ref[1]
        ref[0,3,1,0]=(0.0996876882275)*x_ref[0]+(-0.255346045102)*x_ref[1]
        ref[0,3,1,1]=(0.414708254109)*x_ref[0]+(-0.148700489536)*x_ref[1]
        ref[0,3,2,0]=(-0.740719850161)*x_ref[0]+(0.375907815982)*x_ref[1]
        ref[0,3,2,1]=(-0.184189058365)*x_ref[0]+(-0.274477516444)*x_ref[1]
        ref[0,4,0,0]=(-0.565848858136)*x_ref[0]+(-0.206818417121)*x_ref[1]
        ref[0,4,0,1]=(-0.983652399085)*x_ref[0]+(1.21445827986)*x_ref[1]
        ref[0,4,1,0]=(1.6425686123)*x_ref[0]+(-0.621952337654)*x_ref[1]
        ref[0,4,1,1]=(-1.07526530849)*x_ref[0]+(-0.0915810578715)*x_ref[1]
        ref[0,4,2,0]=(1.4812278738)*x_ref[0]+(1.61991326423)*x_ref[1]
        ref[0,4,2,1]=(-1.4147420413)*x_ref[0]+(1.34428787131)*x_ref[1]
        ref[1,0,0,0]=(-0.910373659777)*x_ref[0]+(1.46920176435)*x_ref[1]
        ref[1,0,0,1]=(0.074473571609)*x_ref[0]+(-0.336642181231)*x_ref[1]
        ref[1,0,1,0]=(-0.141813739296)*x_ref[0]+(-0.534432398621)*x_ref[1]
        ref[1,0,1,1]=(0.728030938984)*x_ref[0]+(0.532542738961)*x_ref[1]
        ref[1,0,2,0]=(0.20676316374)*x_ref[0]+(-0.065235999066)*x_ref[1]
        ref[1,0,2,1]=(-1.09054680794)*x_ref[0]+(0.537754476805)*x_ref[1]
        ref[1,1,0,0]=(0.523878896271)*x_ref[0]+(-0.0561828109964)*x_ref[1]
        ref[1,1,0,1]=(1.27754653741)*x_ref[0]+(0.119734333004)*x_ref[1]
        ref[1,1,1,0]=(0.0933160885942)*x_ref[0]+(0.390183021193)*x_ref[1]
        ref[1,1,1,1]=(0.548273953924)*x_ref[0]+(-0.552588185124)*x_ref[1]
        ref[1,1,2,0]=(0.576803722241)*x_ref[0]+(0.60898781381)*x_ref[1]
        ref[1,1,2,1]=(-0.307231650547)*x_ref[0]+(0.439111970656)*x_ref[1]
        ref[1,2,0,0]=(0.172113300736)*x_ref[0]+(-0.0153971432613)*x_ref[1]
        ref[1,2,0,1]=(-0.883753634465)*x_ref[0]+(1.35320294175)*x_ref[1]
        ref[1,2,1,0]=(0.153937139638)*x_ref[0]+(-0.865654430739)*x_ref[1]
        ref[1,2,1,1]=(0.970377377543)*x_ref[0]+(-1.11418225395)*x_ref[1]
        ref[1,2,2,0]=(-1.11106779953)*x_ref[0]+(-0.357700480625)*x_ref[1]
        ref[1,2,2,1]=(0.475529404147)*x_ref[0]+(-0.121369909464)*x_ref[1]
        ref[1,3,0,0]=(-0.867370155191)*x_ref[0]+(0.625161512036)*x_ref[1]
        ref[1,3,0,1]=(0.252873452474)*x_ref[0]+(1.11368263907)*x_ref[1]
        ref[1,3,1,0]=(-1.25698605702)*x_ref[0]+(-0.381254857218)*x_ref[1]
        ref[1,3,1,1]=(1.12038052674)*x_ref[0]+(-0.113429410248)*x_ref[1]
        ref[1,3,2,0]=(0.0821799570201)*x_ref[0]+(-0.167383600252)*x_ref[1]
        ref[1,3,2,1]=(-0.534217900599)*x_ref[0]+(-1.26144431009)*x_ref[1]
        ref[1,4,0,0]=(0.639186045888)*x_ref[0]+(0.258686782482)*x_ref[1]
        ref[1,4,0,1]=(-0.256515011769)*x_ref[0]+(0.797453395844)*x_ref[1]
        ref[1,4,1,0]=(-0.427518421999)*x_ref[0]+(-0.101439725502)*x_ref[1]
        ref[1,4,1,1]=(-0.251303248239)*x_ref[0]+(0.361194806374)*x_ref[1]
        ref[1,4,2,0]=(0.415313798163)*x_ref[0]+(-0.909602300262)*x_ref[1]
        ref[1,4,2,1]=(0.118542894027)*x_ref[0]+(-0.28224555744)*x_ref[1]
        ref[2,0,0,0]=(0.700687096184)*x_ref[0]+(-1.69964436969)*x_ref[1]
        ref[2,0,0,1]=(0.880393551605)*x_ref[0]+(-0.571112060462)*x_ref[1]
        ref[2,0,1,0]=(1.35990706973)*x_ref[0]+(0.260154297077)*x_ref[1]
        ref[2,0,1,1]=(0.294201987212)*x_ref[0]+(-0.379007248662)*x_ref[1]
        ref[2,0,2,0]=(-1.42914701121)*x_ref[0]+(0.88289145627)*x_ref[1]
        ref[2,0,2,1]=(-0.343976881604)*x_ref[0]+(0.0642493086788)*x_ref[1]
        ref[2,1,0,0]=(0.479103547757)*x_ref[0]+(-1.60276044525)*x_ref[1]
        ref[2,1,0,1]=(0.944321957371)*x_ref[0]+(0.313721173145)*x_ref[1]
        ref[2,1,1,0]=(0.233589885494)*x_ref[0]+(-0.427865810692)*x_ref[1]
        ref[2,1,1,1]=(-0.954518596356)*x_ref[0]+(-0.153033245523)*x_ref[1]
        ref[2,1,2,0]=(1.14079619235)*x_ref[0]+(0.136503305207)*x_ref[1]
        ref[2,1,2,1]=(-0.715601080868)*x_ref[0]+(-1.74677798058)*x_ref[1]
        ref[2,2,0,0]=(0.211908166714)*x_ref[0]+(0.379390512119)*x_ref[1]
        ref[2,2,0,1]=(0.97243389801)*x_ref[0]+(-0.988193298566)*x_ref[1]
        ref[2,2,1,0]=(0.8990152555)*x_ref[0]+(0.22398476583)*x_ref[1]
        ref[2,2,1,1]=(-0.748858338529)*x_ref[0]+(-1.00255098852)*x_ref[1]
        ref[2,2,2,0]=(0.863334969354)*x_ref[0]+(1.14334958607)*x_ref[1]
        ref[2,2,2,1]=(-0.807672728362)*x_ref[0]+(0.837576150239)*x_ref[1]
        ref[2,3,0,0]=(-0.310471066952)*x_ref[0]+(0.261468387883)*x_ref[1]
        ref[2,3,0,1]=(-1.61325161386)*x_ref[0]+(0.016614907729)*x_ref[1]
        ref[2,3,1,0]=(-0.742577714565)*x_ref[0]+(-1.54936077725)*x_ref[1]
        ref[2,3,1,1]=(-0.393825799694)*x_ref[0]+(0.66004298074)*x_ref[1]
        ref[2,3,2,0]=(-1.03965632182)*x_ref[0]+(-0.863160703719)*x_ref[1]
        ref[2,3,2,1]=(0.673788653196)*x_ref[0]+(0.740311339524)*x_ref[1]
        ref[2,4,0,0]=(1.59272250531)*x_ref[0]+(1.27845229216)*x_ref[1]
        ref[2,4,0,1]=(0.112947869874)*x_ref[0]+(1.06626388923)*x_ref[1]
        ref[2,4,1,0]=(-0.935174560087)*x_ref[0]+(0.729112002619)*x_ref[1]
        ref[2,4,1,1]=(1.28167575095)*x_ref[0]+(-1.6626881011)*x_ref[1]
        ref[2,4,2,0]=(-1.01382997014)*x_ref[0]+(0.40870705964)*x_ref[1]
        ref[2,4,2,1]=(-0.903293274454)*x_ref[0]+(0.451662927854)*x_ref[1]
        ref[3,0,0,0]=(0.414933474895)*x_ref[0]+(0.274208992931)*x_ref[1]
        ref[3,0,0,1]=(-0.140384007335)*x_ref[0]+(-0.201270983912)*x_ref[1]
        ref[3,0,1,0]=(-1.34339529394)*x_ref[0]+(-1.26275414913)*x_ref[1]
        ref[3,0,1,1]=(0.34482759487)*x_ref[0]+(0.992140984097)*x_ref[1]
        ref[3,0,2,0]=(-1.56575003976)*x_ref[0]+(-0.514557317682)*x_ref[1]
        ref[3,0,2,1]=(-0.203196026518)*x_ref[0]+(-0.50777550196)*x_ref[1]
        ref[3,1,0,0]=(-0.465027052452)*x_ref[0]+(0.165299984247)*x_ref[1]
        ref[3,1,0,1]=(-0.876537945705)*x_ref[0]+(0.569062140702)*x_ref[1]
        ref[3,1,1,0]=(0.212854890028)*x_ref[0]+(0.772436691343)*x_ref[1]
        ref[3,1,1,1]=(0.376912060989)*x_ref[0]+(-1.59078116917)*x_ref[1]
        ref[3,1,2,0]=(-1.19949528387)*x_ref[0]+(0.508138309995)*x_ref[1]
        ref[3,1,2,1]=(0.535067293288)*x_ref[0]+(0.483486234966)*x_ref[1]
        ref[3,2,0,0]=(-0.366454519079)*x_ref[0]+(0.880937088999)*x_ref[1]
        ref[3,2,0,1]=(-0.164158778522)*x_ref[0]+(-0.0933716527947)*x_ref[1]
        ref[3,2,1,0]=(-0.338148774302)*x_ref[0]+(0.389103301223)*x_ref[1]
        ref[3,2,1,1]=(-1.17323344849)*x_ref[0]+(-0.286752347393)*x_ref[1]
        ref[3,2,2,0]=(0.43651426746)*x_ref[0]+(0.0751512323008)*x_ref[1]
        ref[3,2,2,1]=(1.44242451962)*x_ref[0]+(-0.853169055909)*x_ref[1]
        ref[3,3,0,0]=(0.268862750262)*x_ref[0]+(0.668643005784)*x_ref[1]
        ref[3,3,0,1]=(-1.47078437928)*x_ref[0]+(1.25865018681)*x_ref[1]
        ref[3,3,1,0]=(-0.218849811361)*x_ref[0]+(0.147601849599)*x_ref[1]
        ref[3,3,1,1]=(1.66493534918)*x_ref[0]+(0.368711512357)*x_ref[1]
        ref[3,3,2,0]=(0.259910144631)*x_ref[0]+(-0.977975383491)*x_ref[1]
        ref[3,3,2,1]=(1.87696180213)*x_ref[0]+(-0.583679438031)*x_ref[1]
        ref[3,4,0,0]=(0.226114159843)*x_ref[0]+(0.121547181819)*x_ref[1]
        ref[3,4,0,1]=(1.21215141305)*x_ref[0]+(-0.175673309633)*x_ref[1]
        ref[3,4,1,0]=(0.644340639306)*x_ref[0]+(0.35372514848)*x_ref[1]
        ref[3,4,1,1]=(0.452987488144)*x_ref[0]+(0.838110546855)*x_ref[1]
        ref[3,4,2,0]=(-0.434414511976)*x_ref[0]+(-0.879038818613)*x_ref[1]
        ref[3,4,2,1]=(-0.387240543602)*x_ref[0]+(0.187795787877)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.523374388946)*x[0]+(-0.462029311047)*x[1]+(-0.47732568751)*x[2]
        arg[0,0,0,1]=(0.641437288282)*x[0]+(-0.245967769804)*x[1]+(-0.882711488334)*x[2]
        arg[0,0,1,0]=(-0.039075405101)*x[0]+(0.233479040723)*x[1]+(-0.450244284473)*x[2]
        arg[0,0,1,1]=(-1.00711630706)*x[0]+(-0.247720707527)*x[1]+(-0.597073422746)*x[2]
        arg[0,0,2,0]=(0.141738833983)*x[0]+(-1.34846391496)*x[1]+(0.214891803475)*x[2]
        arg[0,0,2,1]=(0.276967719755)*x[0]+(-0.64300940909)*x[1]+(-1.10068852616)*x[2]
        arg[0,1,0,0]=(0.438911917989)*x[0]+(0.588137625057)*x[1]+(0.744848057181)*x[2]
        arg[0,1,0,1]=(0.266343859518)*x[0]+(1.24092635599)*x[1]+(0.477206170905)*x[2]
        arg[0,1,1,0]=(0.328857688358)*x[0]+(-1.17720038645)*x[1]+(0.285045368661)*x[2]
        arg[0,1,1,1]=(-0.0985505878353)*x[0]+(1.11228710542)*x[1]+(0.159347839514)*x[2]
        arg[0,1,2,0]=(0.125378210814)*x[0]+(1.51252392793)*x[1]+(-0.0402281595871)*x[2]
        arg[0,1,2,1]=(-1.41964435548)*x[0]+(0.825108246489)*x[1]+(0.31679169508)*x[2]
        arg[0,2,0,0]=(1.47728541117)*x[0]+(0.957708443424)*x[1]+(-0.187892562217)*x[2]
        arg[0,2,0,1]=(0.0614505001405)*x[0]+(0.144444072992)*x[1]+(0.287839563179)*x[2]
        arg[0,2,1,0]=(0.512698173313)*x[0]+(0.316852993542)*x[1]+(0.538020705489)*x[2]
        arg[0,2,1,1]=(-0.349895838053)*x[0]+(1.50602760725)*x[1]+(-0.959979406715)*x[2]
        arg[0,2,2,0]=(-0.618802027626)*x[0]+(0.71544179402)*x[1]+(1.09297898989)*x[2]
        arg[0,2,2,1]=(-0.264932530879)*x[0]+(-1.4100323815)*x[1]+(-1.28423925034)*x[2]
        arg[0,3,0,0]=(0.481724893561)*x[0]+(-0.546248388253)*x[1]+(-0.214332802735)*x[2]
        arg[0,3,0,1]=(-1.33293479098)*x[0]+(0.989294320484)*x[1]+(-0.561974021785)*x[2]
        arg[0,3,1,0]=(-0.394718780001)*x[0]+(1.0043343416)*x[1]+(1.51370839793)*x[2]
        arg[0,3,1,1]=(1.08238957373)*x[0]+(0.591631122094)*x[1]+(0.075283954318)*x[2]
        arg[0,3,2,0]=(-1.0250149161)*x[0]+(-0.739848118412)*x[1]+(0.736595112789)*x[2]
        arg[0,3,2,1]=(-1.08033009366)*x[0]+(-0.574926147021)*x[1]+(0.421251948724)*x[2]
        arg[0,4,0,0]=(-0.0894466241931)*x[0]+(1.14935896473)*x[1]+(-0.490410015077)*x[2]
        arg[0,4,0,1]=(1.07777309372)*x[0]+(1.0672971936)*x[1]+(0.340143160972)*x[2]
        arg[0,4,1,0]=(0.905553949676)*x[0]+(0.171320798438)*x[1]+(1.83634812891)*x[2]
        arg[0,4,1,1]=(-0.663516264963)*x[0]+(0.14871937304)*x[1]+(-0.697949666962)*x[2]
        arg[0,4,2,0]=(1.41806579095)*x[0]+(0.0357313086623)*x[1]+(0.110846683649)*x[2]
        arg[0,4,2,1]=(-0.31919270211)*x[0]+(-1.651049084)*x[1]+(-1.47528307262)*x[2]
        arg[1,0,0,0]=(0.167209321535)*x[0]+(-1.13066332212)*x[1]+(0.209489173491)*x[2]
        arg[1,0,0,1]=(1.45770122794)*x[0]+(1.12024118263)*x[1]+(0.0203526002602)*x[2]
        arg[1,0,1,0]=(0.180403238696)*x[0]+(0.580625819717)*x[1]+(-1.00882441351)*x[2]
        arg[1,0,1,1]=(-0.20671421669)*x[0]+(-0.199445466326)*x[1]+(0.887365974398)*x[2]
        arg[1,0,2,0]=(-0.0109233374932)*x[0]+(0.835241335188)*x[1]+(-0.810658050514)*x[2]
        arg[1,0,2,1]=(-1.45991974539)*x[0]+(0.955399748215)*x[1]+(0.398673469474)*x[2]
        arg[1,1,0,0]=(0.726687414113)*x[0]+(-1.35343793545)*x[1]+(0.750079540417)*x[2]
        arg[1,1,0,1]=(-0.677871860083)*x[0]+(-1.56756818572)*x[1]+(-0.240923946416)*x[2]
        arg[1,1,1,0]=(-1.05791511835)*x[0]+(-0.425827952234)*x[1]+(-1.82875340474)*x[2]
        arg[1,1,1,1]=(-0.430783664297)*x[0]+(0.552238591434)*x[1]+(0.502702385375)*x[2]
        arg[1,1,2,0]=(0.240203734777)*x[0]+(-0.0650121194654)*x[1]+(1.04662204076)*x[2]
        arg[1,1,2,1]=(-0.25305133215)*x[0]+(-0.31486026571)*x[1]+(-1.83073697314)*x[2]
        arg[1,2,0,0]=(-1.09643062195)*x[0]+(-0.437887555247)*x[1]+(-1.33970467276)*x[2]
        arg[1,2,0,1]=(1.24489646204)*x[0]+(1.0714412157)*x[1]+(-0.192090595585)*x[2]
        arg[1,2,1,0]=(0.029290840641)*x[0]+(-0.940655160496)*x[1]+(-0.820414570983)*x[2]
        arg[1,2,1,1]=(-0.677190307386)*x[0]+(-0.00164385522847)*x[1]+(-0.46460277293)*x[2]
        arg[1,2,2,0]=(-0.320006508895)*x[0]+(1.13412133331)*x[1]+(-0.619705535377)*x[2]
        arg[1,2,2,1]=(-0.0220330715863)*x[0]+(-0.64687567378)*x[1]+(0.0957350181768)*x[2]
        arg[1,3,0,0]=(-0.25073330162)*x[0]+(0.85795600432)*x[1]+(1.53293778947)*x[2]
        arg[1,3,0,1]=(-0.183346360622)*x[0]+(0.266035391885)*x[1]+(-0.422228348627)*x[2]
        arg[1,3,1,0]=(-0.41587781442)*x[0]+(-0.0103068212789)*x[1]+(0.440665738388)*x[2]
        arg[1,3,1,1]=(1.41825912495)*x[0]+(-0.0611650685825)*x[1]+(0.972511387093)*x[2]
        arg[1,3,2,0]=(-0.336532230508)*x[0]+(-0.0123275905127)*x[1]+(-0.254643420879)*x[2]
        arg[1,3,2,1]=(-0.365437122858)*x[0]+(0.71384297093)*x[1]+(0.447802028475)*x[2]
        arg[1,4,0,0]=(1.33371172009)*x[0]+(1.16948649723)*x[1]+(-0.624453779138)*x[2]
        arg[1,4,0,1]=(0.0919647304743)*x[0]+(1.20509307321)*x[1]+(0.454668856107)*x[2]
        arg[1,4,1,0]=(0.179837036326)*x[0]+(1.27470487309)*x[1]+(0.248127557917)*x[2]
        arg[1,4,1,1]=(-0.0163592123585)*x[0]+(-1.09373628426)*x[1]+(-1.07284707536)*x[2]
        arg[1,4,2,0]=(-0.0617954247638)*x[0]+(0.180336071139)*x[1]+(0.0715760255435)*x[2]
        arg[1,4,2,1]=(0.449148667011)*x[0]+(0.476092516356)*x[1]+(-0.235672424702)*x[2]
        arg[2,0,0,0]=(1.06662096238)*x[0]+(1.52079387843)*x[1]+(-0.465087313097)*x[2]
        arg[2,0,0,1]=(-0.0113962755891)*x[0]+(0.459517781034)*x[1]+(0.423861505703)*x[2]
        arg[2,0,1,0]=(1.19571956088)*x[0]+(1.02444853474)*x[1]+(-1.31054246173)*x[2]
        arg[2,0,1,1]=(-1.39536067914)*x[0]+(0.884184554143)*x[1]+(-0.650926411157)*x[2]
        arg[2,0,2,0]=(-0.328672728946)*x[0]+(1.07247782219)*x[1]+(-0.61103916867)*x[2]
        arg[2,0,2,1]=(0.155304775433)*x[0]+(-1.34794080729)*x[1]+(-0.931891854948)*x[2]
        arg[2,1,0,0]=(-0.953726622792)*x[0]+(0.483614421199)*x[1]+(1.07026616662)*x[2]
        arg[2,1,0,1]=(0.720797127905)*x[0]+(1.31309704242)*x[1]+(0.534213389945)*x[2]
        arg[2,1,1,0]=(1.12642665973)*x[0]+(-0.375134467285)*x[1]+(0.223566393552)*x[2]
        arg[2,1,1,1]=(-0.380892997026)*x[0]+(-0.661627351046)*x[1]+(0.596842673155)*x[2]
        arg[2,1,2,0]=(0.233212348554)*x[0]+(0.106673294307)*x[1]+(-0.22268427628)*x[2]
        arg[2,1,2,1]=(0.67982223365)*x[0]+(-1.53842407155)*x[1]+(0.995328875213)*x[2]
        arg[2,2,0,0]=(1.5835734841)*x[0]+(0.178824495609)*x[1]+(-0.31332770355)*x[2]
        arg[2,2,0,1]=(0.101063103927)*x[0]+(-0.0437670737377)*x[1]+(-0.109514170393)*x[2]
        arg[2,2,1,0]=(-0.892614006193)*x[0]+(0.952803608394)*x[1]+(-0.643308315526)*x[2]
        arg[2,2,1,1]=(-0.0606310193159)*x[0]+(-0.768822197122)*x[1]+(-1.43018654257)*x[2]
        arg[2,2,2,0]=(0.107463856971)*x[0]+(1.27309352669)*x[1]+(1.65234673057)*x[2]
        arg[2,2,2,1]=(0.699315319729)*x[0]+(-0.345960580793)*x[1]+(-1.61621444026)*x[2]
        arg[2,3,0,0]=(0.9761936338)*x[0]+(0.665547749873)*x[1]+(-1.07512721449)*x[2]
        arg[2,3,0,1]=(1.26182572924)*x[0]+(-0.416232568884)*x[1]+(-0.356737405174)*x[2]
        arg[2,3,1,0]=(-0.477958253427)*x[0]+(0.198049156904)*x[1]+(-0.295022782364)*x[2]
        arg[2,3,1,1]=(0.412187471103)*x[0]+(0.335655056628)*x[1]+(-0.0873658679329)*x[2]
        arg[2,3,2,0]=(0.168439409327)*x[0]+(-0.131948711409)*x[1]+(1.66777364846)*x[2]
        arg[2,3,2,1]=(0.508795552378)*x[0]+(-0.390505460243)*x[1]+(0.13190717559)*x[2]
        arg[2,4,0,0]=(0.446543094067)*x[0]+(-1.49620720862)*x[1]+(-1.23581717894)*x[2]
        arg[2,4,0,1]=(-0.964404449625)*x[0]+(0.229523164331)*x[1]+(-0.0829574431644)*x[2]
        arg[2,4,1,0]=(-0.0222885761611)*x[0]+(-0.151419169433)*x[1]+(0.0521007301341)*x[2]
        arg[2,4,1,1]=(-0.299232945099)*x[0]+(0.831745448311)*x[1]+(0.0354480204711)*x[2]
        arg[2,4,2,0]=(0.350303784953)*x[0]+(0.0608299778701)*x[1]+(1.10333111365)*x[2]
        arg[2,4,2,1]=(0.719573610926)*x[0]+(-0.0868905011544)*x[1]+(0.852804030732)*x[2]
        arg[3,0,0,0]=(0.643070908885)*x[0]+(0.846705112025)*x[1]+(-0.340473447494)*x[2]
        arg[3,0,0,1]=(-0.180433440729)*x[0]+(1.20490149625)*x[1]+(1.65034499912)*x[2]
        arg[3,0,1,0]=(-0.454239570444)*x[0]+(0.425446741192)*x[1]+(0.469566134089)*x[2]
        arg[3,0,1,1]=(-0.080272513364)*x[0]+(-0.437416002114)*x[1]+(0.741423922011)*x[2]
        arg[3,0,2,0]=(0.0202920395247)*x[0]+(0.724451134981)*x[1]+(-0.628631137287)*x[2]
        arg[3,0,2,1]=(1.27329890828)*x[0]+(1.32598594097)*x[1]+(-1.03893516914)*x[2]
        arg[3,1,0,0]=(-0.316679599208)*x[0]+(1.37172239358)*x[1]+(-0.333448830595)*x[2]
        arg[3,1,0,1]=(-0.97422120848)*x[0]+(-0.800596641862)*x[1]+(1.21296910893)*x[2]
        arg[3,1,1,0]=(-0.791725431754)*x[0]+(0.173363698424)*x[1]+(-0.137920990647)*x[2]
        arg[3,1,1,1]=(-0.104259554446)*x[0]+(1.14966042893)*x[1]+(0.694287361582)*x[2]
        arg[3,1,2,0]=(0.130345878498)*x[0]+(0.186936528352)*x[1]+(-1.74442226692)*x[2]
        arg[3,1,2,1]=(-0.952031053033)*x[0]+(0.718038193789)*x[1]+(-0.912958209036)*x[2]
        arg[3,2,0,0]=(-1.16180973047)*x[0]+(0.0107173530275)*x[1]+(1.30278786691)*x[2]
        arg[3,2,0,1]=(-0.224509227379)*x[0]+(0.0996327632409)*x[1]+(-0.590950424929)*x[2]
        arg[3,2,1,0]=(1.37202300472)*x[0]+(-1.46745698646)*x[1]+(-0.811483806572)*x[2]
        arg[3,2,1,1]=(1.4828065241)*x[0]+(-0.177365694124)*x[1]+(0.723931620237)*x[2]
        arg[3,2,2,0]=(0.230529059637)*x[0]+(0.728445639896)*x[1]+(-0.566988770509)*x[2]
        arg[3,2,2,1]=(0.797697737992)*x[0]+(-1.13322423945)*x[1]+(-1.1409387296)*x[2]
        arg[3,3,0,0]=(0.217205133789)*x[0]+(-0.510677603419)*x[1]+(-0.580535665335)*x[2]
        arg[3,3,0,1]=(-0.388656568244)*x[0]+(-0.906189671302)*x[1]+(0.44802765855)*x[2]
        arg[3,3,1,0]=(0.13492738758)*x[0]+(-0.0523341874741)*x[1]+(1.42846176777)*x[2]
        arg[3,3,1,1]=(0.243377010926)*x[0]+(-0.701174291473)*x[1]+(-0.756059212704)*x[2]
        arg[3,3,2,0]=(1.48359083301)*x[0]+(1.17692500152)*x[1]+(-0.184437593764)*x[2]
        arg[3,3,2,1]=(-0.105457636222)*x[0]+(-0.0190796656068)*x[1]+(-0.938560206034)*x[2]
        arg[3,4,0,0]=(-0.101684388703)*x[0]+(-1.19809056647)*x[1]+(1.1989510907)*x[2]
        arg[3,4,0,1]=(-1.0745487924)*x[0]+(-0.37560532919)*x[1]+(-0.967022985879)*x[2]
        arg[3,4,1,0]=(0.315585027641)*x[0]+(-0.238716421886)*x[1]+(-0.275850551625)*x[2]
        arg[3,4,1,1]=(-0.824271180248)*x[0]+(-0.0649845724269)*x[1]+(0.530472696035)*x[2]
        arg[3,4,2,0]=(-0.518165130932)*x[0]+(1.44722596827)*x[1]+(0.072367534289)*x[2]
        arg[3,4,2,1]=(-0.247477354356)*x[0]+(0.988584526142)*x[1]+(0.0491978506036)*x[2]
        ref[0,0,0,0]=(0.523374388946)*x_ref[0]+(-0.462029311047)*x_ref[1]+(-0.47732568751)*x_ref[2]
        ref[0,0,0,1]=(0.641437288282)*x_ref[0]+(-0.245967769804)*x_ref[1]+(-0.882711488334)*x_ref[2]
        ref[0,0,1,0]=(-0.039075405101)*x_ref[0]+(0.233479040723)*x_ref[1]+(-0.450244284473)*x_ref[2]
        ref[0,0,1,1]=(-1.00711630706)*x_ref[0]+(-0.247720707527)*x_ref[1]+(-0.597073422746)*x_ref[2]
        ref[0,0,2,0]=(0.141738833983)*x_ref[0]+(-1.34846391496)*x_ref[1]+(0.214891803475)*x_ref[2]
        ref[0,0,2,1]=(0.276967719755)*x_ref[0]+(-0.64300940909)*x_ref[1]+(-1.10068852616)*x_ref[2]
        ref[0,1,0,0]=(0.438911917989)*x_ref[0]+(0.588137625057)*x_ref[1]+(0.744848057181)*x_ref[2]
        ref[0,1,0,1]=(0.266343859518)*x_ref[0]+(1.24092635599)*x_ref[1]+(0.477206170905)*x_ref[2]
        ref[0,1,1,0]=(0.328857688358)*x_ref[0]+(-1.17720038645)*x_ref[1]+(0.285045368661)*x_ref[2]
        ref[0,1,1,1]=(-0.0985505878353)*x_ref[0]+(1.11228710542)*x_ref[1]+(0.159347839514)*x_ref[2]
        ref[0,1,2,0]=(0.125378210814)*x_ref[0]+(1.51252392793)*x_ref[1]+(-0.0402281595871)*x_ref[2]
        ref[0,1,2,1]=(-1.41964435548)*x_ref[0]+(0.825108246489)*x_ref[1]+(0.31679169508)*x_ref[2]
        ref[0,2,0,0]=(1.47728541117)*x_ref[0]+(0.957708443424)*x_ref[1]+(-0.187892562217)*x_ref[2]
        ref[0,2,0,1]=(0.0614505001405)*x_ref[0]+(0.144444072992)*x_ref[1]+(0.287839563179)*x_ref[2]
        ref[0,2,1,0]=(0.512698173313)*x_ref[0]+(0.316852993542)*x_ref[1]+(0.538020705489)*x_ref[2]
        ref[0,2,1,1]=(-0.349895838053)*x_ref[0]+(1.50602760725)*x_ref[1]+(-0.959979406715)*x_ref[2]
        ref[0,2,2,0]=(-0.618802027626)*x_ref[0]+(0.71544179402)*x_ref[1]+(1.09297898989)*x_ref[2]
        ref[0,2,2,1]=(-0.264932530879)*x_ref[0]+(-1.4100323815)*x_ref[1]+(-1.28423925034)*x_ref[2]
        ref[0,3,0,0]=(0.481724893561)*x_ref[0]+(-0.546248388253)*x_ref[1]+(-0.214332802735)*x_ref[2]
        ref[0,3,0,1]=(-1.33293479098)*x_ref[0]+(0.989294320484)*x_ref[1]+(-0.561974021785)*x_ref[2]
        ref[0,3,1,0]=(-0.394718780001)*x_ref[0]+(1.0043343416)*x_ref[1]+(1.51370839793)*x_ref[2]
        ref[0,3,1,1]=(1.08238957373)*x_ref[0]+(0.591631122094)*x_ref[1]+(0.075283954318)*x_ref[2]
        ref[0,3,2,0]=(-1.0250149161)*x_ref[0]+(-0.739848118412)*x_ref[1]+(0.736595112789)*x_ref[2]
        ref[0,3,2,1]=(-1.08033009366)*x_ref[0]+(-0.574926147021)*x_ref[1]+(0.421251948724)*x_ref[2]
        ref[0,4,0,0]=(-0.0894466241931)*x_ref[0]+(1.14935896473)*x_ref[1]+(-0.490410015077)*x_ref[2]
        ref[0,4,0,1]=(1.07777309372)*x_ref[0]+(1.0672971936)*x_ref[1]+(0.340143160972)*x_ref[2]
        ref[0,4,1,0]=(0.905553949676)*x_ref[0]+(0.171320798438)*x_ref[1]+(1.83634812891)*x_ref[2]
        ref[0,4,1,1]=(-0.663516264963)*x_ref[0]+(0.14871937304)*x_ref[1]+(-0.697949666962)*x_ref[2]
        ref[0,4,2,0]=(1.41806579095)*x_ref[0]+(0.0357313086623)*x_ref[1]+(0.110846683649)*x_ref[2]
        ref[0,4,2,1]=(-0.31919270211)*x_ref[0]+(-1.651049084)*x_ref[1]+(-1.47528307262)*x_ref[2]
        ref[1,0,0,0]=(0.167209321535)*x_ref[0]+(-1.13066332212)*x_ref[1]+(0.209489173491)*x_ref[2]
        ref[1,0,0,1]=(1.45770122794)*x_ref[0]+(1.12024118263)*x_ref[1]+(0.0203526002602)*x_ref[2]
        ref[1,0,1,0]=(0.180403238696)*x_ref[0]+(0.580625819717)*x_ref[1]+(-1.00882441351)*x_ref[2]
        ref[1,0,1,1]=(-0.20671421669)*x_ref[0]+(-0.199445466326)*x_ref[1]+(0.887365974398)*x_ref[2]
        ref[1,0,2,0]=(-0.0109233374932)*x_ref[0]+(0.835241335188)*x_ref[1]+(-0.810658050514)*x_ref[2]
        ref[1,0,2,1]=(-1.45991974539)*x_ref[0]+(0.955399748215)*x_ref[1]+(0.398673469474)*x_ref[2]
        ref[1,1,0,0]=(0.726687414113)*x_ref[0]+(-1.35343793545)*x_ref[1]+(0.750079540417)*x_ref[2]
        ref[1,1,0,1]=(-0.677871860083)*x_ref[0]+(-1.56756818572)*x_ref[1]+(-0.240923946416)*x_ref[2]
        ref[1,1,1,0]=(-1.05791511835)*x_ref[0]+(-0.425827952234)*x_ref[1]+(-1.82875340474)*x_ref[2]
        ref[1,1,1,1]=(-0.430783664297)*x_ref[0]+(0.552238591434)*x_ref[1]+(0.502702385375)*x_ref[2]
        ref[1,1,2,0]=(0.240203734777)*x_ref[0]+(-0.0650121194654)*x_ref[1]+(1.04662204076)*x_ref[2]
        ref[1,1,2,1]=(-0.25305133215)*x_ref[0]+(-0.31486026571)*x_ref[1]+(-1.83073697314)*x_ref[2]
        ref[1,2,0,0]=(-1.09643062195)*x_ref[0]+(-0.437887555247)*x_ref[1]+(-1.33970467276)*x_ref[2]
        ref[1,2,0,1]=(1.24489646204)*x_ref[0]+(1.0714412157)*x_ref[1]+(-0.192090595585)*x_ref[2]
        ref[1,2,1,0]=(0.029290840641)*x_ref[0]+(-0.940655160496)*x_ref[1]+(-0.820414570983)*x_ref[2]
        ref[1,2,1,1]=(-0.677190307386)*x_ref[0]+(-0.00164385522847)*x_ref[1]+(-0.46460277293)*x_ref[2]
        ref[1,2,2,0]=(-0.320006508895)*x_ref[0]+(1.13412133331)*x_ref[1]+(-0.619705535377)*x_ref[2]
        ref[1,2,2,1]=(-0.0220330715863)*x_ref[0]+(-0.64687567378)*x_ref[1]+(0.0957350181768)*x_ref[2]
        ref[1,3,0,0]=(-0.25073330162)*x_ref[0]+(0.85795600432)*x_ref[1]+(1.53293778947)*x_ref[2]
        ref[1,3,0,1]=(-0.183346360622)*x_ref[0]+(0.266035391885)*x_ref[1]+(-0.422228348627)*x_ref[2]
        ref[1,3,1,0]=(-0.41587781442)*x_ref[0]+(-0.0103068212789)*x_ref[1]+(0.440665738388)*x_ref[2]
        ref[1,3,1,1]=(1.41825912495)*x_ref[0]+(-0.0611650685825)*x_ref[1]+(0.972511387093)*x_ref[2]
        ref[1,3,2,0]=(-0.336532230508)*x_ref[0]+(-0.0123275905127)*x_ref[1]+(-0.254643420879)*x_ref[2]
        ref[1,3,2,1]=(-0.365437122858)*x_ref[0]+(0.71384297093)*x_ref[1]+(0.447802028475)*x_ref[2]
        ref[1,4,0,0]=(1.33371172009)*x_ref[0]+(1.16948649723)*x_ref[1]+(-0.624453779138)*x_ref[2]
        ref[1,4,0,1]=(0.0919647304743)*x_ref[0]+(1.20509307321)*x_ref[1]+(0.454668856107)*x_ref[2]
        ref[1,4,1,0]=(0.179837036326)*x_ref[0]+(1.27470487309)*x_ref[1]+(0.248127557917)*x_ref[2]
        ref[1,4,1,1]=(-0.0163592123585)*x_ref[0]+(-1.09373628426)*x_ref[1]+(-1.07284707536)*x_ref[2]
        ref[1,4,2,0]=(-0.0617954247638)*x_ref[0]+(0.180336071139)*x_ref[1]+(0.0715760255435)*x_ref[2]
        ref[1,4,2,1]=(0.449148667011)*x_ref[0]+(0.476092516356)*x_ref[1]+(-0.235672424702)*x_ref[2]
        ref[2,0,0,0]=(1.06662096238)*x_ref[0]+(1.52079387843)*x_ref[1]+(-0.465087313097)*x_ref[2]
        ref[2,0,0,1]=(-0.0113962755891)*x_ref[0]+(0.459517781034)*x_ref[1]+(0.423861505703)*x_ref[2]
        ref[2,0,1,0]=(1.19571956088)*x_ref[0]+(1.02444853474)*x_ref[1]+(-1.31054246173)*x_ref[2]
        ref[2,0,1,1]=(-1.39536067914)*x_ref[0]+(0.884184554143)*x_ref[1]+(-0.650926411157)*x_ref[2]
        ref[2,0,2,0]=(-0.328672728946)*x_ref[0]+(1.07247782219)*x_ref[1]+(-0.61103916867)*x_ref[2]
        ref[2,0,2,1]=(0.155304775433)*x_ref[0]+(-1.34794080729)*x_ref[1]+(-0.931891854948)*x_ref[2]
        ref[2,1,0,0]=(-0.953726622792)*x_ref[0]+(0.483614421199)*x_ref[1]+(1.07026616662)*x_ref[2]
        ref[2,1,0,1]=(0.720797127905)*x_ref[0]+(1.31309704242)*x_ref[1]+(0.534213389945)*x_ref[2]
        ref[2,1,1,0]=(1.12642665973)*x_ref[0]+(-0.375134467285)*x_ref[1]+(0.223566393552)*x_ref[2]
        ref[2,1,1,1]=(-0.380892997026)*x_ref[0]+(-0.661627351046)*x_ref[1]+(0.596842673155)*x_ref[2]
        ref[2,1,2,0]=(0.233212348554)*x_ref[0]+(0.106673294307)*x_ref[1]+(-0.22268427628)*x_ref[2]
        ref[2,1,2,1]=(0.67982223365)*x_ref[0]+(-1.53842407155)*x_ref[1]+(0.995328875213)*x_ref[2]
        ref[2,2,0,0]=(1.5835734841)*x_ref[0]+(0.178824495609)*x_ref[1]+(-0.31332770355)*x_ref[2]
        ref[2,2,0,1]=(0.101063103927)*x_ref[0]+(-0.0437670737377)*x_ref[1]+(-0.109514170393)*x_ref[2]
        ref[2,2,1,0]=(-0.892614006193)*x_ref[0]+(0.952803608394)*x_ref[1]+(-0.643308315526)*x_ref[2]
        ref[2,2,1,1]=(-0.0606310193159)*x_ref[0]+(-0.768822197122)*x_ref[1]+(-1.43018654257)*x_ref[2]
        ref[2,2,2,0]=(0.107463856971)*x_ref[0]+(1.27309352669)*x_ref[1]+(1.65234673057)*x_ref[2]
        ref[2,2,2,1]=(0.699315319729)*x_ref[0]+(-0.345960580793)*x_ref[1]+(-1.61621444026)*x_ref[2]
        ref[2,3,0,0]=(0.9761936338)*x_ref[0]+(0.665547749873)*x_ref[1]+(-1.07512721449)*x_ref[2]
        ref[2,3,0,1]=(1.26182572924)*x_ref[0]+(-0.416232568884)*x_ref[1]+(-0.356737405174)*x_ref[2]
        ref[2,3,1,0]=(-0.477958253427)*x_ref[0]+(0.198049156904)*x_ref[1]+(-0.295022782364)*x_ref[2]
        ref[2,3,1,1]=(0.412187471103)*x_ref[0]+(0.335655056628)*x_ref[1]+(-0.0873658679329)*x_ref[2]
        ref[2,3,2,0]=(0.168439409327)*x_ref[0]+(-0.131948711409)*x_ref[1]+(1.66777364846)*x_ref[2]
        ref[2,3,2,1]=(0.508795552378)*x_ref[0]+(-0.390505460243)*x_ref[1]+(0.13190717559)*x_ref[2]
        ref[2,4,0,0]=(0.446543094067)*x_ref[0]+(-1.49620720862)*x_ref[1]+(-1.23581717894)*x_ref[2]
        ref[2,4,0,1]=(-0.964404449625)*x_ref[0]+(0.229523164331)*x_ref[1]+(-0.0829574431644)*x_ref[2]
        ref[2,4,1,0]=(-0.0222885761611)*x_ref[0]+(-0.151419169433)*x_ref[1]+(0.0521007301341)*x_ref[2]
        ref[2,4,1,1]=(-0.299232945099)*x_ref[0]+(0.831745448311)*x_ref[1]+(0.0354480204711)*x_ref[2]
        ref[2,4,2,0]=(0.350303784953)*x_ref[0]+(0.0608299778701)*x_ref[1]+(1.10333111365)*x_ref[2]
        ref[2,4,2,1]=(0.719573610926)*x_ref[0]+(-0.0868905011544)*x_ref[1]+(0.852804030732)*x_ref[2]
        ref[3,0,0,0]=(0.643070908885)*x_ref[0]+(0.846705112025)*x_ref[1]+(-0.340473447494)*x_ref[2]
        ref[3,0,0,1]=(-0.180433440729)*x_ref[0]+(1.20490149625)*x_ref[1]+(1.65034499912)*x_ref[2]
        ref[3,0,1,0]=(-0.454239570444)*x_ref[0]+(0.425446741192)*x_ref[1]+(0.469566134089)*x_ref[2]
        ref[3,0,1,1]=(-0.080272513364)*x_ref[0]+(-0.437416002114)*x_ref[1]+(0.741423922011)*x_ref[2]
        ref[3,0,2,0]=(0.0202920395247)*x_ref[0]+(0.724451134981)*x_ref[1]+(-0.628631137287)*x_ref[2]
        ref[3,0,2,1]=(1.27329890828)*x_ref[0]+(1.32598594097)*x_ref[1]+(-1.03893516914)*x_ref[2]
        ref[3,1,0,0]=(-0.316679599208)*x_ref[0]+(1.37172239358)*x_ref[1]+(-0.333448830595)*x_ref[2]
        ref[3,1,0,1]=(-0.97422120848)*x_ref[0]+(-0.800596641862)*x_ref[1]+(1.21296910893)*x_ref[2]
        ref[3,1,1,0]=(-0.791725431754)*x_ref[0]+(0.173363698424)*x_ref[1]+(-0.137920990647)*x_ref[2]
        ref[3,1,1,1]=(-0.104259554446)*x_ref[0]+(1.14966042893)*x_ref[1]+(0.694287361582)*x_ref[2]
        ref[3,1,2,0]=(0.130345878498)*x_ref[0]+(0.186936528352)*x_ref[1]+(-1.74442226692)*x_ref[2]
        ref[3,1,2,1]=(-0.952031053033)*x_ref[0]+(0.718038193789)*x_ref[1]+(-0.912958209036)*x_ref[2]
        ref[3,2,0,0]=(-1.16180973047)*x_ref[0]+(0.0107173530275)*x_ref[1]+(1.30278786691)*x_ref[2]
        ref[3,2,0,1]=(-0.224509227379)*x_ref[0]+(0.0996327632409)*x_ref[1]+(-0.590950424929)*x_ref[2]
        ref[3,2,1,0]=(1.37202300472)*x_ref[0]+(-1.46745698646)*x_ref[1]+(-0.811483806572)*x_ref[2]
        ref[3,2,1,1]=(1.4828065241)*x_ref[0]+(-0.177365694124)*x_ref[1]+(0.723931620237)*x_ref[2]
        ref[3,2,2,0]=(0.230529059637)*x_ref[0]+(0.728445639896)*x_ref[1]+(-0.566988770509)*x_ref[2]
        ref[3,2,2,1]=(0.797697737992)*x_ref[0]+(-1.13322423945)*x_ref[1]+(-1.1409387296)*x_ref[2]
        ref[3,3,0,0]=(0.217205133789)*x_ref[0]+(-0.510677603419)*x_ref[1]+(-0.580535665335)*x_ref[2]
        ref[3,3,0,1]=(-0.388656568244)*x_ref[0]+(-0.906189671302)*x_ref[1]+(0.44802765855)*x_ref[2]
        ref[3,3,1,0]=(0.13492738758)*x_ref[0]+(-0.0523341874741)*x_ref[1]+(1.42846176777)*x_ref[2]
        ref[3,3,1,1]=(0.243377010926)*x_ref[0]+(-0.701174291473)*x_ref[1]+(-0.756059212704)*x_ref[2]
        ref[3,3,2,0]=(1.48359083301)*x_ref[0]+(1.17692500152)*x_ref[1]+(-0.184437593764)*x_ref[2]
        ref[3,3,2,1]=(-0.105457636222)*x_ref[0]+(-0.0190796656068)*x_ref[1]+(-0.938560206034)*x_ref[2]
        ref[3,4,0,0]=(-0.101684388703)*x_ref[0]+(-1.19809056647)*x_ref[1]+(1.1989510907)*x_ref[2]
        ref[3,4,0,1]=(-1.0745487924)*x_ref[0]+(-0.37560532919)*x_ref[1]+(-0.967022985879)*x_ref[2]
        ref[3,4,1,0]=(0.315585027641)*x_ref[0]+(-0.238716421886)*x_ref[1]+(-0.275850551625)*x_ref[2]
        ref[3,4,1,1]=(-0.824271180248)*x_ref[0]+(-0.0649845724269)*x_ref[1]+(0.530472696035)*x_ref[2]
        ref[3,4,2,0]=(-0.518165130932)*x_ref[0]+(1.44722596827)*x_ref[1]+(0.072367534289)*x_ref[2]
        ref[3,4,2,1]=(-0.247477354356)*x_ref[0]+(0.988584526142)*x_ref[1]+(0.0491978506036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0969659653911)*x[0]**o+(-0.0778467995236)*x[0]+(-0.896520347956)*x[1]**o+(0.452579859858)*x[1]
        ref=(0.0969659653911)*x_ref[0]**o+(-0.0778467995236)*x_ref[0]+(-0.896520347956)*x_ref[1]**o+(0.452579859858)*x_ref[1]
      else:
        arg=(-0.62506591451)*x[0]**o+(0.970735975624)*x[0]+(-0.563055321362)*x[1]**o+(-0.281482915069)*x[1]+(-0.6461258967)*x[2]**o+(0.786610350609)*x[2]
        ref=(-0.62506591451)*x_ref[0]**o+(0.970735975624)*x_ref[0]+(-0.563055321362)*x_ref[1]**o+(-0.281482915069)*x_ref[1]+(-0.6461258967)*x_ref[2]**o+(0.786610350609)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.725240039321)*x[0]**o+(0.326302203613)*x[0]+(0.356072156898)*x[1]**o+(0.980066388307)*x[1]
        arg[1]=(0.171551389095)*x[0]**o+(0.379133470853)*x[0]+(0.120608077039)*x[1]**o+(-0.413701037132)*x[1]
        ref[0]=(0.725240039321)*x_ref[0]**o+(0.326302203613)*x_ref[0]+(0.356072156898)*x_ref[1]**o+(0.980066388307)*x_ref[1]
        ref[1]=(0.171551389095)*x_ref[0]**o+(0.379133470853)*x_ref[0]+(0.120608077039)*x_ref[1]**o+(-0.413701037132)*x_ref[1]
      else:
        arg[0]=(-0.656095745624)*x[0]**o+(-0.540430373724)*x[0]+(-0.109962817674)*x[1]**o+(-0.969692783268)*x[1]+(-0.377994841279)*x[2]**o+(-0.664035030664)*x[2]
        arg[1]=(0.768498465456)*x[0]**o+(0.647294499552)*x[0]+(0.824672861058)*x[1]**o+(-0.848088780836)*x[1]+(0.045558138597)*x[2]**o+(-0.209365900036)*x[2]
        ref[0]=(-0.656095745624)*x_ref[0]**o+(-0.540430373724)*x_ref[0]+(-0.109962817674)*x_ref[1]**o+(-0.969692783268)*x_ref[1]+(-0.377994841279)*x_ref[2]**o+(-0.664035030664)*x_ref[2]
        ref[1]=(0.768498465456)*x_ref[0]**o+(0.647294499552)*x_ref[0]+(0.824672861058)*x_ref[1]**o+(-0.848088780836)*x_ref[1]+(0.045558138597)*x_ref[2]**o+(-0.209365900036)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.694866182164)*x[0]**o+(0.759641527396)*x[0]+(-0.696641139757)*x[1]**o+(0.617978953453)*x[1]
        arg[0,1]=(0.75058922213)*x[0]**o+(0.580832851854)*x[0]+(-0.237039031011)*x[1]**o+(0.53518450519)*x[1]
        arg[0,2]=(-0.191933019173)*x[0]**o+(-0.0638088686782)*x[0]+(0.871645336149)*x[1]**o+(0.256502442034)*x[1]
        arg[0,3]=(0.829189580591)*x[0]**o+(0.186064016504)*x[0]+(-0.204496850484)*x[1]**o+(-0.314535598074)*x[1]
        arg[0,4]=(0.170968249419)*x[0]**o+(-0.719959835105)*x[0]+(0.367000223059)*x[1]**o+(-0.210239222752)*x[1]
        arg[1,0]=(-0.424311493894)*x[0]**o+(-0.381097143579)*x[0]+(-0.54925292526)*x[1]**o+(-0.393349131169)*x[1]
        arg[1,1]=(0.148181508143)*x[0]**o+(-0.224530937841)*x[0]+(0.778748719265)*x[1]**o+(-0.522104185968)*x[1]
        arg[1,2]=(0.134819103337)*x[0]**o+(0.0388165865124)*x[0]+(-0.772318490115)*x[1]**o+(-0.928611227188)*x[1]
        arg[1,3]=(0.790807030518)*x[0]**o+(-0.849964960983)*x[0]+(-0.79407080953)*x[1]**o+(-0.626315155005)*x[1]
        arg[1,4]=(-0.998842104726)*x[0]**o+(-0.325938405818)*x[0]+(-0.81445420822)*x[1]**o+(0.868281547664)*x[1]
        arg[2,0]=(-0.975572739854)*x[0]**o+(-0.0630975707316)*x[0]+(0.0688031679656)*x[1]**o+(0.557152464527)*x[1]
        arg[2,1]=(0.389814261767)*x[0]**o+(0.523891881188)*x[0]+(-0.558388242102)*x[1]**o+(-0.45720905598)*x[1]
        arg[2,2]=(0.763327517145)*x[0]**o+(0.637954909904)*x[0]+(0.407602154355)*x[1]**o+(0.707118271445)*x[1]
        arg[2,3]=(-0.823371482965)*x[0]**o+(0.918685041464)*x[0]+(0.991991462022)*x[1]**o+(0.489892661208)*x[1]
        arg[2,4]=(-0.661567455889)*x[0]**o+(-0.00566795402284)*x[0]+(0.86705493299)*x[1]**o+(0.00158321121677)*x[1]
        arg[3,0]=(0.365456232671)*x[0]**o+(-0.587373381962)*x[0]+(0.751291631507)*x[1]**o+(-0.0605248375222)*x[1]
        arg[3,1]=(0.546630294768)*x[0]**o+(-0.903068297271)*x[0]+(0.486313468693)*x[1]**o+(0.355322343068)*x[1]
        arg[3,2]=(0.233755056849)*x[0]**o+(-0.350375425427)*x[0]+(0.0115435986629)*x[1]**o+(-0.713895312694)*x[1]
        arg[3,3]=(0.385411302842)*x[0]**o+(0.0733743700003)*x[0]+(0.439219708301)*x[1]**o+(0.129391539278)*x[1]
        arg[3,4]=(-0.194268604537)*x[0]**o+(0.175829932738)*x[0]+(0.866342013931)*x[1]**o+(-0.213891435501)*x[1]
        ref[0,0]=(0.694866182164)*x_ref[0]**o+(0.759641527396)*x_ref[0]+(-0.696641139757)*x_ref[1]**o+(0.617978953453)*x_ref[1]
        ref[0,1]=(0.75058922213)*x_ref[0]**o+(0.580832851854)*x_ref[0]+(-0.237039031011)*x_ref[1]**o+(0.53518450519)*x_ref[1]
        ref[0,2]=(-0.191933019173)*x_ref[0]**o+(-0.0638088686782)*x_ref[0]+(0.871645336149)*x_ref[1]**o+(0.256502442034)*x_ref[1]
        ref[0,3]=(0.829189580591)*x_ref[0]**o+(0.186064016504)*x_ref[0]+(-0.204496850484)*x_ref[1]**o+(-0.314535598074)*x_ref[1]
        ref[0,4]=(0.170968249419)*x_ref[0]**o+(-0.719959835105)*x_ref[0]+(0.367000223059)*x_ref[1]**o+(-0.210239222752)*x_ref[1]
        ref[1,0]=(-0.424311493894)*x_ref[0]**o+(-0.381097143579)*x_ref[0]+(-0.54925292526)*x_ref[1]**o+(-0.393349131169)*x_ref[1]
        ref[1,1]=(0.148181508143)*x_ref[0]**o+(-0.224530937841)*x_ref[0]+(0.778748719265)*x_ref[1]**o+(-0.522104185968)*x_ref[1]
        ref[1,2]=(0.134819103337)*x_ref[0]**o+(0.0388165865124)*x_ref[0]+(-0.772318490115)*x_ref[1]**o+(-0.928611227188)*x_ref[1]
        ref[1,3]=(0.790807030518)*x_ref[0]**o+(-0.849964960983)*x_ref[0]+(-0.79407080953)*x_ref[1]**o+(-0.626315155005)*x_ref[1]
        ref[1,4]=(-0.998842104726)*x_ref[0]**o+(-0.325938405818)*x_ref[0]+(-0.81445420822)*x_ref[1]**o+(0.868281547664)*x_ref[1]
        ref[2,0]=(-0.975572739854)*x_ref[0]**o+(-0.0630975707316)*x_ref[0]+(0.0688031679656)*x_ref[1]**o+(0.557152464527)*x_ref[1]
        ref[2,1]=(0.389814261767)*x_ref[0]**o+(0.523891881188)*x_ref[0]+(-0.558388242102)*x_ref[1]**o+(-0.45720905598)*x_ref[1]
        ref[2,2]=(0.763327517145)*x_ref[0]**o+(0.637954909904)*x_ref[0]+(0.407602154355)*x_ref[1]**o+(0.707118271445)*x_ref[1]
        ref[2,3]=(-0.823371482965)*x_ref[0]**o+(0.918685041464)*x_ref[0]+(0.991991462022)*x_ref[1]**o+(0.489892661208)*x_ref[1]
        ref[2,4]=(-0.661567455889)*x_ref[0]**o+(-0.00566795402284)*x_ref[0]+(0.86705493299)*x_ref[1]**o+(0.00158321121677)*x_ref[1]
        ref[3,0]=(0.365456232671)*x_ref[0]**o+(-0.587373381962)*x_ref[0]+(0.751291631507)*x_ref[1]**o+(-0.0605248375222)*x_ref[1]
        ref[3,1]=(0.546630294768)*x_ref[0]**o+(-0.903068297271)*x_ref[0]+(0.486313468693)*x_ref[1]**o+(0.355322343068)*x_ref[1]
        ref[3,2]=(0.233755056849)*x_ref[0]**o+(-0.350375425427)*x_ref[0]+(0.0115435986629)*x_ref[1]**o+(-0.713895312694)*x_ref[1]
        ref[3,3]=(0.385411302842)*x_ref[0]**o+(0.0733743700003)*x_ref[0]+(0.439219708301)*x_ref[1]**o+(0.129391539278)*x_ref[1]
        ref[3,4]=(-0.194268604537)*x_ref[0]**o+(0.175829932738)*x_ref[0]+(0.866342013931)*x_ref[1]**o+(-0.213891435501)*x_ref[1]
      else:
        arg[0,0]=(-0.479138946804)*x[0]**o+(0.240408443007)*x[0]+(0.309155796205)*x[1]**o+(-0.124360991957)*x[1]+(0.00301047542037)*x[2]**o+(-0.039439802799)*x[2]
        arg[0,1]=(0.41979402771)*x[0]**o+(0.798738902847)*x[0]+(0.619960746922)*x[1]**o+(-0.551662130664)*x[1]+(0.914293623695)*x[2]**o+(0.6859768812)*x[2]
        arg[0,2]=(0.744593492696)*x[0]**o+(-0.31743031745)*x[0]+(0.389545123971)*x[1]**o+(0.592024987288)*x[1]+(0.63456095022)*x[2]**o+(-0.746571796061)*x[2]
        arg[0,3]=(-0.632549503649)*x[0]**o+(-0.687745599598)*x[0]+(0.794176042123)*x[1]**o+(-0.971466375953)*x[1]+(-0.086939632062)*x[2]**o+(0.373736042024)*x[2]
        arg[0,4]=(-0.966879538217)*x[0]**o+(-0.609291970974)*x[0]+(0.000409196978095)*x[1]**o+(0.623934484362)*x[1]+(-0.1694749999)*x[2]**o+(0.131778917979)*x[2]
        arg[1,0]=(0.763452444611)*x[0]**o+(0.805147582411)*x[0]+(0.999704124146)*x[1]**o+(0.0635253753138)*x[1]+(-0.717889410548)*x[2]**o+(-0.840765968423)*x[2]
        arg[1,1]=(-0.26910479702)*x[0]**o+(0.594093239771)*x[0]+(-0.536150372834)*x[1]**o+(-0.768506263316)*x[1]+(-0.924510063594)*x[2]**o+(-0.950568839784)*x[2]
        arg[1,2]=(-0.452317334239)*x[0]**o+(0.186343827048)*x[0]+(0.921015823436)*x[1]**o+(-0.656809197675)*x[1]+(-0.841898648461)*x[2]**o+(0.267827136605)*x[2]
        arg[1,3]=(-0.880873336233)*x[0]**o+(-0.915350352959)*x[0]+(0.92197459758)*x[1]**o+(0.20861139921)*x[1]+(0.776743964515)*x[2]**o+(0.858569288507)*x[2]
        arg[1,4]=(0.422773256864)*x[0]**o+(0.397743860671)*x[0]+(0.726470981309)*x[1]**o+(-0.233534612019)*x[1]+(-0.350345927201)*x[2]**o+(-0.192294374344)*x[2]
        arg[2,0]=(0.0706785590906)*x[0]**o+(-0.102098736751)*x[0]+(-0.858298802695)*x[1]**o+(-0.1390708698)*x[1]+(0.221129390442)*x[2]**o+(0.770148745174)*x[2]
        arg[2,1]=(0.619559141603)*x[0]**o+(-0.969165124767)*x[0]+(0.36662839455)*x[1]**o+(0.505630485999)*x[1]+(0.170716012487)*x[2]**o+(0.355457960368)*x[2]
        arg[2,2]=(0.254098998564)*x[0]**o+(0.725662782597)*x[0]+(0.274108937869)*x[1]**o+(0.682336753679)*x[1]+(-0.858192561327)*x[2]**o+(-0.538785359078)*x[2]
        arg[2,3]=(0.25513729674)*x[0]**o+(-0.594968324608)*x[0]+(-0.306612686559)*x[1]**o+(-0.953891015684)*x[1]+(-0.00631658753287)*x[2]**o+(-0.889694464702)*x[2]
        arg[2,4]=(0.654463492789)*x[0]**o+(-0.42220702047)*x[0]+(-0.690125229151)*x[1]**o+(0.318923837754)*x[1]+(-0.691341299676)*x[2]**o+(0.824631778823)*x[2]
        arg[3,0]=(-0.576563725107)*x[0]**o+(-0.0354219245402)*x[0]+(0.650854296808)*x[1]**o+(0.930072736771)*x[1]+(0.8383657117)*x[2]**o+(0.191293272146)*x[2]
        arg[3,1]=(0.104382485594)*x[0]**o+(-0.718025390561)*x[0]+(-0.822188446684)*x[1]**o+(-0.0772502664782)*x[1]+(-0.513331680922)*x[2]**o+(0.164340503698)*x[2]
        arg[3,2]=(0.685178273481)*x[0]**o+(-0.241726920352)*x[0]+(-0.891512838953)*x[1]**o+(-0.428833037662)*x[1]+(-0.910220899771)*x[2]**o+(-0.0109246344661)*x[2]
        arg[3,3]=(-0.294769406995)*x[0]**o+(0.643502505933)*x[0]+(0.309444282121)*x[1]**o+(-0.935318923262)*x[1]+(0.242426961867)*x[2]**o+(-0.417268637067)*x[2]
        arg[3,4]=(0.0382306334841)*x[0]**o+(-0.564858278904)*x[0]+(-0.594566652327)*x[1]**o+(0.911894434818)*x[1]+(-0.314178793024)*x[2]**o+(0.302730333267)*x[2]
        ref[0,0]=(-0.479138946804)*x_ref[0]**o+(0.240408443007)*x_ref[0]+(0.309155796205)*x_ref[1]**o+(-0.124360991957)*x_ref[1]+(0.00301047542037)*x_ref[2]**o+(-0.039439802799)*x_ref[2]
        ref[0,1]=(0.41979402771)*x_ref[0]**o+(0.798738902847)*x_ref[0]+(0.619960746922)*x_ref[1]**o+(-0.551662130664)*x_ref[1]+(0.914293623695)*x_ref[2]**o+(0.6859768812)*x_ref[2]
        ref[0,2]=(0.744593492696)*x_ref[0]**o+(-0.31743031745)*x_ref[0]+(0.389545123971)*x_ref[1]**o+(0.592024987288)*x_ref[1]+(0.63456095022)*x_ref[2]**o+(-0.746571796061)*x_ref[2]
        ref[0,3]=(-0.632549503649)*x_ref[0]**o+(-0.687745599598)*x_ref[0]+(0.794176042123)*x_ref[1]**o+(-0.971466375953)*x_ref[1]+(-0.086939632062)*x_ref[2]**o+(0.373736042024)*x_ref[2]
        ref[0,4]=(-0.966879538217)*x_ref[0]**o+(-0.609291970974)*x_ref[0]+(0.000409196978095)*x_ref[1]**o+(0.623934484362)*x_ref[1]+(-0.1694749999)*x_ref[2]**o+(0.131778917979)*x_ref[2]
        ref[1,0]=(0.763452444611)*x_ref[0]**o+(0.805147582411)*x_ref[0]+(0.999704124146)*x_ref[1]**o+(0.0635253753138)*x_ref[1]+(-0.717889410548)*x_ref[2]**o+(-0.840765968423)*x_ref[2]
        ref[1,1]=(-0.26910479702)*x_ref[0]**o+(0.594093239771)*x_ref[0]+(-0.536150372834)*x_ref[1]**o+(-0.768506263316)*x_ref[1]+(-0.924510063594)*x_ref[2]**o+(-0.950568839784)*x_ref[2]
        ref[1,2]=(-0.452317334239)*x_ref[0]**o+(0.186343827048)*x_ref[0]+(0.921015823436)*x_ref[1]**o+(-0.656809197675)*x_ref[1]+(-0.841898648461)*x_ref[2]**o+(0.267827136605)*x_ref[2]
        ref[1,3]=(-0.880873336233)*x_ref[0]**o+(-0.915350352959)*x_ref[0]+(0.92197459758)*x_ref[1]**o+(0.20861139921)*x_ref[1]+(0.776743964515)*x_ref[2]**o+(0.858569288507)*x_ref[2]
        ref[1,4]=(0.422773256864)*x_ref[0]**o+(0.397743860671)*x_ref[0]+(0.726470981309)*x_ref[1]**o+(-0.233534612019)*x_ref[1]+(-0.350345927201)*x_ref[2]**o+(-0.192294374344)*x_ref[2]
        ref[2,0]=(0.0706785590906)*x_ref[0]**o+(-0.102098736751)*x_ref[0]+(-0.858298802695)*x_ref[1]**o+(-0.1390708698)*x_ref[1]+(0.221129390442)*x_ref[2]**o+(0.770148745174)*x_ref[2]
        ref[2,1]=(0.619559141603)*x_ref[0]**o+(-0.969165124767)*x_ref[0]+(0.36662839455)*x_ref[1]**o+(0.505630485999)*x_ref[1]+(0.170716012487)*x_ref[2]**o+(0.355457960368)*x_ref[2]
        ref[2,2]=(0.254098998564)*x_ref[0]**o+(0.725662782597)*x_ref[0]+(0.274108937869)*x_ref[1]**o+(0.682336753679)*x_ref[1]+(-0.858192561327)*x_ref[2]**o+(-0.538785359078)*x_ref[2]
        ref[2,3]=(0.25513729674)*x_ref[0]**o+(-0.594968324608)*x_ref[0]+(-0.306612686559)*x_ref[1]**o+(-0.953891015684)*x_ref[1]+(-0.00631658753287)*x_ref[2]**o+(-0.889694464702)*x_ref[2]
        ref[2,4]=(0.654463492789)*x_ref[0]**o+(-0.42220702047)*x_ref[0]+(-0.690125229151)*x_ref[1]**o+(0.318923837754)*x_ref[1]+(-0.691341299676)*x_ref[2]**o+(0.824631778823)*x_ref[2]
        ref[3,0]=(-0.576563725107)*x_ref[0]**o+(-0.0354219245402)*x_ref[0]+(0.650854296808)*x_ref[1]**o+(0.930072736771)*x_ref[1]+(0.8383657117)*x_ref[2]**o+(0.191293272146)*x_ref[2]
        ref[3,1]=(0.104382485594)*x_ref[0]**o+(-0.718025390561)*x_ref[0]+(-0.822188446684)*x_ref[1]**o+(-0.0772502664782)*x_ref[1]+(-0.513331680922)*x_ref[2]**o+(0.164340503698)*x_ref[2]
        ref[3,2]=(0.685178273481)*x_ref[0]**o+(-0.241726920352)*x_ref[0]+(-0.891512838953)*x_ref[1]**o+(-0.428833037662)*x_ref[1]+(-0.910220899771)*x_ref[2]**o+(-0.0109246344661)*x_ref[2]
        ref[3,3]=(-0.294769406995)*x_ref[0]**o+(0.643502505933)*x_ref[0]+(0.309444282121)*x_ref[1]**o+(-0.935318923262)*x_ref[1]+(0.242426961867)*x_ref[2]**o+(-0.417268637067)*x_ref[2]
        ref[3,4]=(0.0382306334841)*x_ref[0]**o+(-0.564858278904)*x_ref[0]+(-0.594566652327)*x_ref[1]**o+(0.911894434818)*x_ref[1]+(-0.314178793024)*x_ref[2]**o+(0.302730333267)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.298635657968)*x[0]**o+(0.917751825722)*x[0]+(-0.000157823722007)*x[1]**o+(-0.202721972906)*x[1]
        arg[0,0,1]=(0.164880328673)*x[0]**o+(-0.310233508297)*x[0]+(-0.976270095138)*x[1]**o+(-0.86689594754)*x[1]
        arg[0,1,0]=(-0.130835262631)*x[0]**o+(-0.485555484157)*x[0]+(0.055655815912)*x[1]**o+(-0.488201458657)*x[1]
        arg[0,1,1]=(-0.339982923518)*x[0]**o+(0.435464515598)*x[0]+(-0.986955506076)*x[1]**o+(0.590966000031)*x[1]
        arg[1,0,0]=(0.150280127463)*x[0]**o+(0.827183876844)*x[0]+(0.468401701763)*x[1]**o+(-0.879871376091)*x[1]
        arg[1,0,1]=(0.48434831025)*x[0]**o+(-0.623958069024)*x[0]+(-0.781911271849)*x[1]**o+(0.34398679645)*x[1]
        arg[1,1,0]=(0.124273770782)*x[0]**o+(-0.683096046651)*x[0]+(0.769636378285)*x[1]**o+(-0.471970832437)*x[1]
        arg[1,1,1]=(0.769777912195)*x[0]**o+(-0.740802020319)*x[0]+(0.659109761568)*x[1]**o+(-0.495444655614)*x[1]
        arg[2,0,0]=(0.270063777356)*x[0]**o+(0.558849439723)*x[0]+(0.128946408169)*x[1]**o+(-0.962495994445)*x[1]
        arg[2,0,1]=(-0.0403460883009)*x[0]**o+(0.084983334662)*x[0]+(0.297573700577)*x[1]**o+(0.911774703678)*x[1]
        arg[2,1,0]=(0.821680378088)*x[0]**o+(0.10823172035)*x[0]+(-0.150213295203)*x[1]**o+(0.310471151961)*x[1]
        arg[2,1,1]=(-0.407033195691)*x[0]**o+(0.716262964777)*x[0]+(-0.262475432397)*x[1]**o+(-0.234158898682)*x[1]
        arg[3,0,0]=(-0.109193787139)*x[0]**o+(0.403892847875)*x[0]+(0.794082322801)*x[1]**o+(0.669070572253)*x[1]
        arg[3,0,1]=(0.969770217228)*x[0]**o+(0.735280190411)*x[0]+(-0.339891263613)*x[1]**o+(0.859989056027)*x[1]
        arg[3,1,0]=(-0.906193851712)*x[0]**o+(-0.330178132059)*x[0]+(-0.48035114847)*x[1]**o+(-0.48050067188)*x[1]
        arg[3,1,1]=(0.922778761752)*x[0]**o+(0.899145572602)*x[0]+(-0.991922151176)*x[1]**o+(0.615744852656)*x[1]
        arg[4,0,0]=(0.457092295352)*x[0]**o+(-0.120904430848)*x[0]+(0.575186782188)*x[1]**o+(-0.685522788659)*x[1]
        arg[4,0,1]=(0.254934861907)*x[0]**o+(-0.208850206761)*x[0]+(-0.0933987467304)*x[1]**o+(0.687715419212)*x[1]
        arg[4,1,0]=(0.483890412252)*x[0]**o+(-0.192055355216)*x[0]+(0.0891595367214)*x[1]**o+(-0.835900121805)*x[1]
        arg[4,1,1]=(-0.281109655935)*x[0]**o+(0.279867178407)*x[0]+(0.843094337157)*x[1]**o+(-0.457583293166)*x[1]
        arg[5,0,0]=(0.595078633674)*x[0]**o+(-0.994652086537)*x[0]+(0.111354924733)*x[1]**o+(0.715580298167)*x[1]
        arg[5,0,1]=(0.823409068766)*x[0]**o+(0.41047089933)*x[0]+(-0.610763277348)*x[1]**o+(0.0991108428868)*x[1]
        arg[5,1,0]=(-0.0291539407811)*x[0]**o+(0.371799932333)*x[0]+(0.631702366805)*x[1]**o+(-0.858415923309)*x[1]
        arg[5,1,1]=(-0.636146064991)*x[0]**o+(0.407248062196)*x[0]+(-0.507069143933)*x[1]**o+(-0.0509951705166)*x[1]
        ref[0,0,0]=(0.298635657968)*x_ref[0]**o+(0.917751825722)*x_ref[0]+(-0.000157823722007)*x_ref[1]**o+(-0.202721972906)*x_ref[1]
        ref[0,0,1]=(0.164880328673)*x_ref[0]**o+(-0.310233508297)*x_ref[0]+(-0.976270095138)*x_ref[1]**o+(-0.86689594754)*x_ref[1]
        ref[0,1,0]=(-0.130835262631)*x_ref[0]**o+(-0.485555484157)*x_ref[0]+(0.055655815912)*x_ref[1]**o+(-0.488201458657)*x_ref[1]
        ref[0,1,1]=(-0.339982923518)*x_ref[0]**o+(0.435464515598)*x_ref[0]+(-0.986955506076)*x_ref[1]**o+(0.590966000031)*x_ref[1]
        ref[1,0,0]=(0.150280127463)*x_ref[0]**o+(0.827183876844)*x_ref[0]+(0.468401701763)*x_ref[1]**o+(-0.879871376091)*x_ref[1]
        ref[1,0,1]=(0.48434831025)*x_ref[0]**o+(-0.623958069024)*x_ref[0]+(-0.781911271849)*x_ref[1]**o+(0.34398679645)*x_ref[1]
        ref[1,1,0]=(0.124273770782)*x_ref[0]**o+(-0.683096046651)*x_ref[0]+(0.769636378285)*x_ref[1]**o+(-0.471970832437)*x_ref[1]
        ref[1,1,1]=(0.769777912195)*x_ref[0]**o+(-0.740802020319)*x_ref[0]+(0.659109761568)*x_ref[1]**o+(-0.495444655614)*x_ref[1]
        ref[2,0,0]=(0.270063777356)*x_ref[0]**o+(0.558849439723)*x_ref[0]+(0.128946408169)*x_ref[1]**o+(-0.962495994445)*x_ref[1]
        ref[2,0,1]=(-0.0403460883009)*x_ref[0]**o+(0.084983334662)*x_ref[0]+(0.297573700577)*x_ref[1]**o+(0.911774703678)*x_ref[1]
        ref[2,1,0]=(0.821680378088)*x_ref[0]**o+(0.10823172035)*x_ref[0]+(-0.150213295203)*x_ref[1]**o+(0.310471151961)*x_ref[1]
        ref[2,1,1]=(-0.407033195691)*x_ref[0]**o+(0.716262964777)*x_ref[0]+(-0.262475432397)*x_ref[1]**o+(-0.234158898682)*x_ref[1]
        ref[3,0,0]=(-0.109193787139)*x_ref[0]**o+(0.403892847875)*x_ref[0]+(0.794082322801)*x_ref[1]**o+(0.669070572253)*x_ref[1]
        ref[3,0,1]=(0.969770217228)*x_ref[0]**o+(0.735280190411)*x_ref[0]+(-0.339891263613)*x_ref[1]**o+(0.859989056027)*x_ref[1]
        ref[3,1,0]=(-0.906193851712)*x_ref[0]**o+(-0.330178132059)*x_ref[0]+(-0.48035114847)*x_ref[1]**o+(-0.48050067188)*x_ref[1]
        ref[3,1,1]=(0.922778761752)*x_ref[0]**o+(0.899145572602)*x_ref[0]+(-0.991922151176)*x_ref[1]**o+(0.615744852656)*x_ref[1]
        ref[4,0,0]=(0.457092295352)*x_ref[0]**o+(-0.120904430848)*x_ref[0]+(0.575186782188)*x_ref[1]**o+(-0.685522788659)*x_ref[1]
        ref[4,0,1]=(0.254934861907)*x_ref[0]**o+(-0.208850206761)*x_ref[0]+(-0.0933987467304)*x_ref[1]**o+(0.687715419212)*x_ref[1]
        ref[4,1,0]=(0.483890412252)*x_ref[0]**o+(-0.192055355216)*x_ref[0]+(0.0891595367214)*x_ref[1]**o+(-0.835900121805)*x_ref[1]
        ref[4,1,1]=(-0.281109655935)*x_ref[0]**o+(0.279867178407)*x_ref[0]+(0.843094337157)*x_ref[1]**o+(-0.457583293166)*x_ref[1]
        ref[5,0,0]=(0.595078633674)*x_ref[0]**o+(-0.994652086537)*x_ref[0]+(0.111354924733)*x_ref[1]**o+(0.715580298167)*x_ref[1]
        ref[5,0,1]=(0.823409068766)*x_ref[0]**o+(0.41047089933)*x_ref[0]+(-0.610763277348)*x_ref[1]**o+(0.0991108428868)*x_ref[1]
        ref[5,1,0]=(-0.0291539407811)*x_ref[0]**o+(0.371799932333)*x_ref[0]+(0.631702366805)*x_ref[1]**o+(-0.858415923309)*x_ref[1]
        ref[5,1,1]=(-0.636146064991)*x_ref[0]**o+(0.407248062196)*x_ref[0]+(-0.507069143933)*x_ref[1]**o+(-0.0509951705166)*x_ref[1]
      else:
        arg[0,0,0]=(-0.662043178282)*x[0]**o+(0.600130780119)*x[0]+(-0.52233323628)*x[1]**o+(-0.463302513654)*x[1]+(0.714996279403)*x[2]**o+(0.113219137203)*x[2]
        arg[0,0,1]=(0.94317347224)*x[0]**o+(0.987584533548)*x[0]+(0.116577593127)*x[1]**o+(-0.515267022711)*x[1]+(0.856578953903)*x[2]**o+(0.70372338917)*x[2]
        arg[0,1,0]=(-0.405265998966)*x[0]**o+(-0.656404120416)*x[0]+(0.270655604701)*x[1]**o+(0.0638534875176)*x[1]+(0.816834973788)*x[2]**o+(0.918436000225)*x[2]
        arg[0,1,1]=(-0.37299950332)*x[0]**o+(-0.0367737761004)*x[0]+(0.176531372245)*x[1]**o+(0.849773143988)*x[1]+(-0.572911529869)*x[2]**o+(-0.140163625628)*x[2]
        arg[1,0,0]=(0.263616091516)*x[0]**o+(-0.65550051998)*x[0]+(-0.118310211881)*x[1]**o+(-0.695263983587)*x[1]+(-0.49011649182)*x[2]**o+(-0.553284002669)*x[2]
        arg[1,0,1]=(0.859640205193)*x[0]**o+(-0.569681205214)*x[0]+(0.459979003108)*x[1]**o+(0.672561515814)*x[1]+(-0.854289983441)*x[2]**o+(-0.445344129294)*x[2]
        arg[1,1,0]=(-0.904355809101)*x[0]**o+(0.540513356972)*x[0]+(-0.377422260181)*x[1]**o+(0.937285151058)*x[1]+(0.148943597435)*x[2]**o+(0.992256197076)*x[2]
        arg[1,1,1]=(-0.868336238952)*x[0]**o+(-0.713337888363)*x[0]+(0.800427728006)*x[1]**o+(-0.83502690568)*x[1]+(0.365707922084)*x[2]**o+(0.218740577318)*x[2]
        arg[2,0,0]=(-0.0382672594668)*x[0]**o+(0.463831755073)*x[0]+(-0.104957678792)*x[1]**o+(-0.561738607653)*x[1]+(-0.660128314406)*x[2]**o+(0.594649133189)*x[2]
        arg[2,0,1]=(-0.399297921751)*x[0]**o+(0.550876554224)*x[0]+(0.150599666563)*x[1]**o+(-0.748400310268)*x[1]+(0.658318134651)*x[2]**o+(0.49075091442)*x[2]
        arg[2,1,0]=(-0.393769876286)*x[0]**o+(0.328125610064)*x[0]+(-0.703385111404)*x[1]**o+(0.0418077854723)*x[1]+(0.918774663106)*x[2]**o+(-0.119085951508)*x[2]
        arg[2,1,1]=(-0.549250924758)*x[0]**o+(0.169148558793)*x[0]+(0.664162931811)*x[1]**o+(0.207136816188)*x[1]+(-0.17207817969)*x[2]**o+(0.34258020774)*x[2]
        arg[3,0,0]=(0.446467472517)*x[0]**o+(-0.358128380843)*x[0]+(0.146916048359)*x[1]**o+(-0.606692124285)*x[1]+(0.828655704727)*x[2]**o+(0.067139299396)*x[2]
        arg[3,0,1]=(-0.357052242351)*x[0]**o+(-0.528255164843)*x[0]+(0.140204673142)*x[1]**o+(0.674282235672)*x[1]+(-0.469118454039)*x[2]**o+(0.388103121466)*x[2]
        arg[3,1,0]=(-0.359629791333)*x[0]**o+(-0.625108929158)*x[0]+(0.855034791748)*x[1]**o+(-0.343451063341)*x[1]+(-0.681557972105)*x[2]**o+(0.576310870238)*x[2]
        arg[3,1,1]=(0.00546530278132)*x[0]**o+(0.359221089344)*x[0]+(-0.57416247885)*x[1]**o+(-0.521306138323)*x[1]+(-0.618281954095)*x[2]**o+(-0.862162207352)*x[2]
        arg[4,0,0]=(-0.196356423488)*x[0]**o+(-0.871482508533)*x[0]+(-0.367390886081)*x[1]**o+(0.580310108591)*x[1]+(0.857970671482)*x[2]**o+(0.123685673359)*x[2]
        arg[4,0,1]=(0.715104674022)*x[0]**o+(-0.00444438628932)*x[0]+(0.538883210054)*x[1]**o+(0.179340705463)*x[1]+(-0.605249436909)*x[2]**o+(-0.744253826562)*x[2]
        arg[4,1,0]=(-0.315453893647)*x[0]**o+(-0.560946546082)*x[0]+(-0.674448928458)*x[1]**o+(-0.510243444449)*x[1]+(0.592615876779)*x[2]**o+(0.218024801263)*x[2]
        arg[4,1,1]=(-0.46902473326)*x[0]**o+(-0.573926691452)*x[0]+(0.827898114814)*x[1]**o+(-0.644208082857)*x[1]+(-0.956169254426)*x[2]**o+(-0.297101737561)*x[2]
        arg[5,0,0]=(0.45818164662)*x[0]**o+(0.219667021464)*x[0]+(-0.453125753418)*x[1]**o+(-0.201214929832)*x[1]+(0.135363067088)*x[2]**o+(0.850636247323)*x[2]
        arg[5,0,1]=(0.296120304926)*x[0]**o+(-0.139971296351)*x[0]+(0.448115780401)*x[1]**o+(-0.300602343399)*x[1]+(-0.165568753274)*x[2]**o+(0.984788641639)*x[2]
        arg[5,1,0]=(0.321875942152)*x[0]**o+(-0.403122417353)*x[0]+(-0.148664639927)*x[1]**o+(0.644143662303)*x[1]+(-0.718311175542)*x[2]**o+(-0.850248996913)*x[2]
        arg[5,1,1]=(0.44869757406)*x[0]**o+(0.0292094503086)*x[0]+(-0.159644215279)*x[1]**o+(0.695976060359)*x[1]+(-0.221444458713)*x[2]**o+(-0.42974111206)*x[2]
        ref[0,0,0]=(-0.662043178282)*x_ref[0]**o+(0.600130780119)*x_ref[0]+(-0.52233323628)*x_ref[1]**o+(-0.463302513654)*x_ref[1]+(0.714996279403)*x_ref[2]**o+(0.113219137203)*x_ref[2]
        ref[0,0,1]=(0.94317347224)*x_ref[0]**o+(0.987584533548)*x_ref[0]+(0.116577593127)*x_ref[1]**o+(-0.515267022711)*x_ref[1]+(0.856578953903)*x_ref[2]**o+(0.70372338917)*x_ref[2]
        ref[0,1,0]=(-0.405265998966)*x_ref[0]**o+(-0.656404120416)*x_ref[0]+(0.270655604701)*x_ref[1]**o+(0.0638534875176)*x_ref[1]+(0.816834973788)*x_ref[2]**o+(0.918436000225)*x_ref[2]
        ref[0,1,1]=(-0.37299950332)*x_ref[0]**o+(-0.0367737761004)*x_ref[0]+(0.176531372245)*x_ref[1]**o+(0.849773143988)*x_ref[1]+(-0.572911529869)*x_ref[2]**o+(-0.140163625628)*x_ref[2]
        ref[1,0,0]=(0.263616091516)*x_ref[0]**o+(-0.65550051998)*x_ref[0]+(-0.118310211881)*x_ref[1]**o+(-0.695263983587)*x_ref[1]+(-0.49011649182)*x_ref[2]**o+(-0.553284002669)*x_ref[2]
        ref[1,0,1]=(0.859640205193)*x_ref[0]**o+(-0.569681205214)*x_ref[0]+(0.459979003108)*x_ref[1]**o+(0.672561515814)*x_ref[1]+(-0.854289983441)*x_ref[2]**o+(-0.445344129294)*x_ref[2]
        ref[1,1,0]=(-0.904355809101)*x_ref[0]**o+(0.540513356972)*x_ref[0]+(-0.377422260181)*x_ref[1]**o+(0.937285151058)*x_ref[1]+(0.148943597435)*x_ref[2]**o+(0.992256197076)*x_ref[2]
        ref[1,1,1]=(-0.868336238952)*x_ref[0]**o+(-0.713337888363)*x_ref[0]+(0.800427728006)*x_ref[1]**o+(-0.83502690568)*x_ref[1]+(0.365707922084)*x_ref[2]**o+(0.218740577318)*x_ref[2]
        ref[2,0,0]=(-0.0382672594668)*x_ref[0]**o+(0.463831755073)*x_ref[0]+(-0.104957678792)*x_ref[1]**o+(-0.561738607653)*x_ref[1]+(-0.660128314406)*x_ref[2]**o+(0.594649133189)*x_ref[2]
        ref[2,0,1]=(-0.399297921751)*x_ref[0]**o+(0.550876554224)*x_ref[0]+(0.150599666563)*x_ref[1]**o+(-0.748400310268)*x_ref[1]+(0.658318134651)*x_ref[2]**o+(0.49075091442)*x_ref[2]
        ref[2,1,0]=(-0.393769876286)*x_ref[0]**o+(0.328125610064)*x_ref[0]+(-0.703385111404)*x_ref[1]**o+(0.0418077854723)*x_ref[1]+(0.918774663106)*x_ref[2]**o+(-0.119085951508)*x_ref[2]
        ref[2,1,1]=(-0.549250924758)*x_ref[0]**o+(0.169148558793)*x_ref[0]+(0.664162931811)*x_ref[1]**o+(0.207136816188)*x_ref[1]+(-0.17207817969)*x_ref[2]**o+(0.34258020774)*x_ref[2]
        ref[3,0,0]=(0.446467472517)*x_ref[0]**o+(-0.358128380843)*x_ref[0]+(0.146916048359)*x_ref[1]**o+(-0.606692124285)*x_ref[1]+(0.828655704727)*x_ref[2]**o+(0.067139299396)*x_ref[2]
        ref[3,0,1]=(-0.357052242351)*x_ref[0]**o+(-0.528255164843)*x_ref[0]+(0.140204673142)*x_ref[1]**o+(0.674282235672)*x_ref[1]+(-0.469118454039)*x_ref[2]**o+(0.388103121466)*x_ref[2]
        ref[3,1,0]=(-0.359629791333)*x_ref[0]**o+(-0.625108929158)*x_ref[0]+(0.855034791748)*x_ref[1]**o+(-0.343451063341)*x_ref[1]+(-0.681557972105)*x_ref[2]**o+(0.576310870238)*x_ref[2]
        ref[3,1,1]=(0.00546530278132)*x_ref[0]**o+(0.359221089344)*x_ref[0]+(-0.57416247885)*x_ref[1]**o+(-0.521306138323)*x_ref[1]+(-0.618281954095)*x_ref[2]**o+(-0.862162207352)*x_ref[2]
        ref[4,0,0]=(-0.196356423488)*x_ref[0]**o+(-0.871482508533)*x_ref[0]+(-0.367390886081)*x_ref[1]**o+(0.580310108591)*x_ref[1]+(0.857970671482)*x_ref[2]**o+(0.123685673359)*x_ref[2]
        ref[4,0,1]=(0.715104674022)*x_ref[0]**o+(-0.00444438628932)*x_ref[0]+(0.538883210054)*x_ref[1]**o+(0.179340705463)*x_ref[1]+(-0.605249436909)*x_ref[2]**o+(-0.744253826562)*x_ref[2]
        ref[4,1,0]=(-0.315453893647)*x_ref[0]**o+(-0.560946546082)*x_ref[0]+(-0.674448928458)*x_ref[1]**o+(-0.510243444449)*x_ref[1]+(0.592615876779)*x_ref[2]**o+(0.218024801263)*x_ref[2]
        ref[4,1,1]=(-0.46902473326)*x_ref[0]**o+(-0.573926691452)*x_ref[0]+(0.827898114814)*x_ref[1]**o+(-0.644208082857)*x_ref[1]+(-0.956169254426)*x_ref[2]**o+(-0.297101737561)*x_ref[2]
        ref[5,0,0]=(0.45818164662)*x_ref[0]**o+(0.219667021464)*x_ref[0]+(-0.453125753418)*x_ref[1]**o+(-0.201214929832)*x_ref[1]+(0.135363067088)*x_ref[2]**o+(0.850636247323)*x_ref[2]
        ref[5,0,1]=(0.296120304926)*x_ref[0]**o+(-0.139971296351)*x_ref[0]+(0.448115780401)*x_ref[1]**o+(-0.300602343399)*x_ref[1]+(-0.165568753274)*x_ref[2]**o+(0.984788641639)*x_ref[2]
        ref[5,1,0]=(0.321875942152)*x_ref[0]**o+(-0.403122417353)*x_ref[0]+(-0.148664639927)*x_ref[1]**o+(0.644143662303)*x_ref[1]+(-0.718311175542)*x_ref[2]**o+(-0.850248996913)*x_ref[2]
        ref[5,1,1]=(0.44869757406)*x_ref[0]**o+(0.0292094503086)*x_ref[0]+(-0.159644215279)*x_ref[1]**o+(0.695976060359)*x_ref[1]+(-0.221444458713)*x_ref[2]**o+(-0.42974111206)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromData_ReducedFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.0737428272134)*x[0]**o+(0.841780036645)*x[0]+(-0.110824257531)*x[1]**o+(-0.957726444468)*x[1]
        arg[0,0,0,1]=(-0.934840410671)*x[0]**o+(-0.83301064321)*x[0]+(-0.352287650266)*x[1]**o+(-0.961111564077)*x[1]
        arg[0,0,1,0]=(-0.829500940226)*x[0]**o+(0.644171710931)*x[0]+(0.232545739722)*x[1]**o+(0.584336012862)*x[1]
        arg[0,0,1,1]=(0.426298830976)*x[0]**o+(-0.615281197268)*x[0]+(0.206334026192)*x[1]**o+(-0.443698947044)*x[1]
        arg[0,0,2,0]=(-0.236495891617)*x[0]**o+(0.0883376851179)*x[0]+(-0.176541434356)*x[1]**o+(0.297175301617)*x[1]
        arg[0,0,2,1]=(-0.879955324373)*x[0]**o+(-0.244778913771)*x[0]+(0.244131188594)*x[1]**o+(-0.334739112198)*x[1]
        arg[0,1,0,0]=(-0.64887128006)*x[0]**o+(-0.489951454377)*x[0]+(-0.328402433668)*x[1]**o+(0.229814237117)*x[1]
        arg[0,1,0,1]=(-0.339451656097)*x[0]**o+(-0.409320682304)*x[0]+(0.216482338677)*x[1]**o+(-0.10594373055)*x[1]
        arg[0,1,1,0]=(-0.350538479425)*x[0]**o+(0.466450046985)*x[0]+(-0.779124795292)*x[1]**o+(-0.654562596704)*x[1]
        arg[0,1,1,1]=(0.00350092879868)*x[0]**o+(-0.214643088647)*x[0]+(0.786934081411)*x[1]**o+(-0.296853380043)*x[1]
        arg[0,1,2,0]=(-0.640183955648)*x[0]**o+(0.568647968237)*x[0]+(0.575533476901)*x[1]**o+(0.854540594085)*x[1]
        arg[0,1,2,1]=(0.633674899518)*x[0]**o+(0.829255679443)*x[0]+(0.537888424303)*x[1]**o+(-0.495082587617)*x[1]
        arg[0,2,0,0]=(-0.700436798342)*x[0]**o+(-0.732581859026)*x[0]+(-0.508905353053)*x[1]**o+(0.664601151676)*x[1]
        arg[0,2,0,1]=(0.720335764302)*x[0]**o+(0.941159917988)*x[0]+(0.2817440547)*x[1]**o+(0.553712160058)*x[1]
        arg[0,2,1,0]=(-0.937114312262)*x[0]**o+(-0.115517396201)*x[0]+(-0.672697125905)*x[1]**o+(-0.221320095369)*x[1]
        arg[0,2,1,1]=(-0.52059341062)*x[0]**o+(-0.653834647475)*x[0]+(-0.0851100861135)*x[1]**o+(-0.668300210067)*x[1]
        arg[0,2,2,0]=(-0.943613035509)*x[0]**o+(-0.294843145822)*x[0]+(0.496278748831)*x[1]**o+(0.901154419425)*x[1]
        arg[0,2,2,1]=(-0.421698818203)*x[0]**o+(-0.721918177544)*x[0]+(-0.197855702215)*x[1]**o+(0.50882827715)*x[1]
        arg[0,3,0,0]=(0.931990780298)*x[0]**o+(-0.550332606951)*x[0]+(0.401584654642)*x[1]**o+(0.0965259747601)*x[1]
        arg[0,3,0,1]=(0.881357708445)*x[0]**o+(0.564624034197)*x[0]+(-0.458191897924)*x[1]**o+(-0.427701112807)*x[1]
        arg[0,3,1,0]=(0.624777088081)*x[0]**o+(0.862654721208)*x[0]+(-0.707618093268)*x[1]**o+(0.743304953313)*x[1]
        arg[0,3,1,1]=(0.4510836974)*x[0]**o+(-0.357023031453)*x[0]+(-0.00503924010273)*x[1]**o+(0.545024336011)*x[1]
        arg[0,3,2,0]=(0.109003851094)*x[0]**o+(0.168659376881)*x[0]+(-0.0578953780655)*x[1]**o+(-0.481212462712)*x[1]
        arg[0,3,2,1]=(-0.191183501636)*x[0]**o+(-0.316145516377)*x[0]+(0.402460103016)*x[1]**o+(-0.163612464438)*x[1]
        arg[0,4,0,0]=(0.489815345033)*x[0]**o+(0.999138280765)*x[0]+(-0.626627766812)*x[1]**o+(-0.339653992568)*x[1]
        arg[0,4,0,1]=(0.616350228791)*x[0]**o+(0.933446438765)*x[0]+(0.944244035512)*x[1]**o+(0.829407789353)*x[1]
        arg[0,4,1,0]=(0.619918547863)*x[0]**o+(0.995488719729)*x[0]+(-0.78847429581)*x[1]**o+(0.542645674462)*x[1]
        arg[0,4,1,1]=(-0.149244819992)*x[0]**o+(0.921766299925)*x[0]+(0.435255339567)*x[1]**o+(-0.0596561300741)*x[1]
        arg[0,4,2,0]=(-0.816364336598)*x[0]**o+(0.701160246658)*x[0]+(-0.616385505652)*x[1]**o+(-0.558625482105)*x[1]
        arg[0,4,2,1]=(-0.36944364693)*x[0]**o+(0.243426977623)*x[0]+(-0.61578435391)*x[1]**o+(0.932570326139)*x[1]
        arg[1,0,0,0]=(-0.36839962807)*x[0]**o+(0.590534439534)*x[0]+(-0.615158899289)*x[1]**o+(0.973649158296)*x[1]
        arg[1,0,0,1]=(0.761515772771)*x[0]**o+(0.706406307012)*x[0]+(0.328712319043)*x[1]**o+(0.438600140517)*x[1]
        arg[1,0,1,0]=(0.895080579207)*x[0]**o+(-0.166789328225)*x[0]+(0.807540897371)*x[1]**o+(0.0345522692723)*x[1]
        arg[1,0,1,1]=(-0.460062661172)*x[0]**o+(0.505167337562)*x[0]+(0.25155527144)*x[1]**o+(0.601093887451)*x[1]
        arg[1,0,2,0]=(0.382129818527)*x[0]**o+(0.694676536542)*x[0]+(0.702794119083)*x[1]**o+(0.704094013056)*x[1]
        arg[1,0,2,1]=(0.265476982913)*x[0]**o+(-0.134338744881)*x[0]+(-0.0938910335356)*x[1]**o+(-0.560073316097)*x[1]
        arg[1,1,0,0]=(0.977368328792)*x[0]**o+(0.084482890741)*x[0]+(0.57068389287)*x[1]**o+(-0.242087071526)*x[1]
        arg[1,1,0,1]=(-0.36067775897)*x[0]**o+(0.768187470286)*x[0]+(-0.881150642619)*x[1]**o+(-0.15839498619)*x[1]
        arg[1,1,1,0]=(-0.808263913645)*x[0]**o+(0.788403321664)*x[0]+(0.976206012343)*x[1]**o+(-0.636933353579)*x[1]
        arg[1,1,1,1]=(0.699185445546)*x[0]**o+(0.251811206876)*x[0]+(-0.146320762812)*x[1]**o+(-0.248033048423)*x[1]
        arg[1,1,2,0]=(-0.782940880812)*x[0]**o+(0.789401686528)*x[0]+(-0.806883904562)*x[1]**o+(0.83689696772)*x[1]
        arg[1,1,2,1]=(0.652404993342)*x[0]**o+(-0.135318598212)*x[0]+(0.138928760301)*x[1]**o+(0.675593554805)*x[1]
        arg[1,2,0,0]=(0.471746379585)*x[0]**o+(-0.989783415847)*x[0]+(-0.0935052296418)*x[1]**o+(-0.532245102898)*x[1]
        arg[1,2,0,1]=(-0.44541059979)*x[0]**o+(0.593734266824)*x[0]+(-0.616647011751)*x[1]**o+(0.461140297024)*x[1]
        arg[1,2,1,0]=(-0.9144210459)*x[0]**o+(0.116806032197)*x[0]+(0.935610395306)*x[1]**o+(0.550884822213)*x[1]
        arg[1,2,1,1]=(-0.962444455944)*x[0]**o+(-0.654328013652)*x[0]+(-0.545330994236)*x[1]**o+(-0.919705456162)*x[1]
        arg[1,2,2,0]=(-0.724491374462)*x[0]**o+(-0.820402375498)*x[0]+(-0.953956431336)*x[1]**o+(-0.201450523182)*x[1]
        arg[1,2,2,1]=(0.723637497881)*x[0]**o+(-0.144483238523)*x[0]+(0.00379745988102)*x[1]**o+(-0.838616087543)*x[1]
        arg[1,3,0,0]=(-0.962256046042)*x[0]**o+(-0.453580322779)*x[0]+(0.887446007424)*x[1]**o+(-0.404454768131)*x[1]
        arg[1,3,0,1]=(0.829189872775)*x[0]**o+(0.327985850399)*x[0]+(0.113600094307)*x[1]**o+(-0.338775377551)*x[1]
        arg[1,3,1,0]=(0.912320027833)*x[0]**o+(0.0658152563951)*x[0]+(-0.843491517568)*x[1]**o+(0.858344434434)*x[1]
        arg[1,3,1,1]=(-0.133918678317)*x[0]**o+(0.644272729149)*x[0]+(0.140663927965)*x[1]**o+(0.853385355163)*x[1]
        arg[1,3,2,0]=(0.37733357322)*x[0]**o+(0.255105934063)*x[0]+(0.88881892183)*x[1]**o+(-0.722661948279)*x[1]
        arg[1,3,2,1]=(-0.2827466345)*x[0]**o+(-0.518815029512)*x[0]+(-0.995043824929)*x[1]**o+(0.973473354481)*x[1]
        arg[1,4,0,0]=(0.044904732821)*x[0]**o+(-0.0373604593083)*x[0]+(0.184254626748)*x[1]**o+(0.031832590233)*x[1]
        arg[1,4,0,1]=(0.388699053911)*x[0]**o+(-0.0619606394414)*x[0]+(0.477947819745)*x[1]**o+(0.982909216221)*x[1]
        arg[1,4,1,0]=(0.854348679973)*x[0]**o+(-0.822052612379)*x[0]+(-0.950305089968)*x[1]**o+(0.0100277256384)*x[1]
        arg[1,4,1,1]=(0.265075277179)*x[0]**o+(0.489846993407)*x[0]+(-0.961111499722)*x[1]**o+(-0.0457909055602)*x[1]
        arg[1,4,2,0]=(0.701951191951)*x[0]**o+(0.675232620259)*x[0]+(-0.141848599227)*x[1]**o+(0.737150257564)*x[1]
        arg[1,4,2,1]=(-0.450684564415)*x[0]**o+(0.854849922021)*x[0]+(0.729670746723)*x[1]**o+(0.530372628879)*x[1]
        arg[2,0,0,0]=(-0.131153262675)*x[0]**o+(-0.477683407979)*x[0]+(0.231416752516)*x[1]**o+(0.790897064402)*x[1]
        arg[2,0,0,1]=(0.0956557004382)*x[0]**o+(0.209084913471)*x[0]+(-0.328544695733)*x[1]**o+(0.701245770328)*x[1]
        arg[2,0,1,0]=(-0.476403718361)*x[0]**o+(0.689023533158)*x[0]+(0.108732189071)*x[1]**o+(0.243379977453)*x[1]
        arg[2,0,1,1]=(0.414917801157)*x[0]**o+(-0.843544596785)*x[0]+(0.772236792935)*x[1]**o+(0.836028831232)*x[1]
        arg[2,0,2,0]=(0.754970600617)*x[0]**o+(-0.422318891122)*x[0]+(0.770767911534)*x[1]**o+(-0.446737894095)*x[1]
        arg[2,0,2,1]=(0.0139989331471)*x[0]**o+(0.262286042291)*x[0]+(-0.440109329114)*x[1]**o+(-0.844965114722)*x[1]
        arg[2,1,0,0]=(-0.758977153009)*x[0]**o+(-0.991192011039)*x[0]+(0.642267239475)*x[1]**o+(-0.701360491163)*x[1]
        arg[2,1,0,1]=(-0.123019007509)*x[0]**o+(-0.622323252979)*x[0]+(0.441364031664)*x[1]**o+(0.311433222821)*x[1]
        arg[2,1,1,0]=(-0.0446644182928)*x[0]**o+(0.614952704493)*x[0]+(-0.228284494629)*x[1]**o+(0.077603612991)*x[1]
        arg[2,1,1,1]=(0.712924108225)*x[0]**o+(-0.153966300537)*x[0]+(-0.584450014127)*x[1]**o+(0.476329149833)*x[1]
        arg[2,1,2,0]=(0.554026516037)*x[0]**o+(0.730575821847)*x[0]+(-0.247529915107)*x[1]**o+(0.210060454223)*x[1]
        arg[2,1,2,1]=(0.877017208752)*x[0]**o+(-0.800916200997)*x[0]+(0.0996710064295)*x[1]**o+(-0.0890375735808)*x[1]
        arg[2,2,0,0]=(-0.198026186062)*x[0]**o+(-0.416160224681)*x[0]+(-0.626491103997)*x[1]**o+(-0.950177752043)*x[1]
        arg[2,2,0,1]=(-0.401319234479)*x[0]**o+(0.948120359627)*x[0]+(0.420351950404)*x[1]**o+(-0.242850312594)*x[1]
        arg[2,2,1,0]=(-0.239652069063)*x[0]**o+(-0.894811076801)*x[0]+(0.979721413509)*x[1]**o+(-0.712493893566)*x[1]
        arg[2,2,1,1]=(0.239416494209)*x[0]**o+(0.449047649538)*x[0]+(0.0503751073583)*x[1]**o+(0.739949331521)*x[1]
        arg[2,2,2,0]=(-0.318067701034)*x[0]**o+(-0.00208209268356)*x[0]+(-0.513821624843)*x[1]**o+(-0.202688815501)*x[1]
        arg[2,2,2,1]=(-0.194111524544)*x[0]**o+(-0.843213241929)*x[0]+(0.124223747087)*x[1]**o+(-0.988305059412)*x[1]
        arg[2,3,0,0]=(0.450894655873)*x[0]**o+(-0.809351657482)*x[0]+(-0.80555128185)*x[1]**o+(-0.922123806508)*x[1]
        arg[2,3,0,1]=(-0.462822602909)*x[0]**o+(0.504388766027)*x[0]+(-0.152725064915)*x[1]**o+(0.988745142638)*x[1]
        arg[2,3,1,0]=(-0.717873944493)*x[0]**o+(0.987011714613)*x[0]+(0.908500737929)*x[1]**o+(0.612497707425)*x[1]
        arg[2,3,1,1]=(0.221645363399)*x[0]**o+(0.219147687681)*x[0]+(-0.195831118343)*x[1]**o+(-0.59541920919)*x[1]
        arg[2,3,2,0]=(-0.289487342206)*x[0]**o+(-0.96904981397)*x[0]+(-0.844880792969)*x[1]**o+(0.803145388038)*x[1]
        arg[2,3,2,1]=(-0.0234295826132)*x[0]**o+(-0.839069887016)*x[0]+(0.591231341444)*x[1]**o+(-0.678605318787)*x[1]
        arg[2,4,0,0]=(0.717943320641)*x[0]**o+(0.625728011485)*x[0]+(0.456836257053)*x[1]**o+(-0.00523685585287)*x[1]
        arg[2,4,0,1]=(-0.549526374185)*x[0]**o+(0.57441917876)*x[0]+(-0.913459106781)*x[1]**o+(-0.610369375195)*x[1]
        arg[2,4,1,0]=(-0.27166563459)*x[0]**o+(0.146710581194)*x[0]+(0.297934124718)*x[1]**o+(0.989561209048)*x[1]
        arg[2,4,1,1]=(-0.0973858709778)*x[0]**o+(-0.909133394381)*x[0]+(0.634182571287)*x[1]**o+(0.477163052831)*x[1]
        arg[2,4,2,0]=(0.0878998038301)*x[0]**o+(-0.90568487021)*x[0]+(0.798978355456)*x[1]**o+(0.977402478037)*x[1]
        arg[2,4,2,1]=(-0.456868538667)*x[0]**o+(0.0723272702119)*x[0]+(0.143461337462)*x[1]**o+(-0.923713568343)*x[1]
        arg[3,0,0,0]=(-0.154410689788)*x[0]**o+(0.634925838303)*x[0]+(-0.0508765762892)*x[1]**o+(-0.889731180834)*x[1]
        arg[3,0,0,1]=(-0.611336323518)*x[0]**o+(0.748058372943)*x[0]+(-0.395060568704)*x[1]**o+(-0.413772262434)*x[1]
        arg[3,0,1,0]=(-0.76924573047)*x[0]**o+(0.61286303551)*x[0]+(0.810883211884)*x[1]**o+(-0.908861022244)*x[1]
        arg[3,0,1,1]=(0.269999661235)*x[0]**o+(0.0330954575695)*x[0]+(0.560272546054)*x[1]**o+(0.877730763728)*x[1]
        arg[3,0,2,0]=(-0.617658215856)*x[0]**o+(0.758414649277)*x[0]+(-0.223497113863)*x[1]**o+(-0.0706589108357)*x[1]
        arg[3,0,2,1]=(0.47956571804)*x[0]**o+(0.938375191345)*x[0]+(-0.193997511219)*x[1]**o+(0.500063249143)*x[1]
        arg[3,1,0,0]=(0.994033773649)*x[0]**o+(0.203022945646)*x[0]+(0.439296640709)*x[1]**o+(0.283679657893)*x[1]
        arg[3,1,0,1]=(0.32495149734)*x[0]**o+(-0.871627347694)*x[0]+(-0.438159387872)*x[1]**o+(0.169112665588)*x[1]
        arg[3,1,1,0]=(-0.0979322945554)*x[0]**o+(-0.936928842822)*x[0]+(0.0402738011011)*x[1]**o+(-0.445031618539)*x[1]
        arg[3,1,1,1]=(-0.511218949907)*x[0]**o+(0.255146779315)*x[0]+(-0.802012163396)*x[1]**o+(0.361819705056)*x[1]
        arg[3,1,2,0]=(-0.902499203011)*x[0]**o+(0.76589952141)*x[0]+(0.52498929732)*x[1]**o+(0.493894292321)*x[1]
        arg[3,1,2,1]=(-0.475117609018)*x[0]**o+(0.682205309303)*x[0]+(-0.661824086649)*x[1]**o+(0.985521553774)*x[1]
        arg[3,2,0,0]=(0.937321686473)*x[0]**o+(-0.518305775925)*x[0]+(0.171689574942)*x[1]**o+(-0.751579960819)*x[1]
        arg[3,2,0,1]=(-0.607004597985)*x[0]**o+(0.342330208344)*x[0]+(0.692094243326)*x[1]**o+(-0.457756057357)*x[1]
        arg[3,2,1,0]=(0.126171621354)*x[0]**o+(0.0499938423847)*x[0]+(0.572006209576)*x[1]**o+(0.121601182464)*x[1]
        arg[3,2,1,1]=(0.337936456181)*x[0]**o+(-0.940186907514)*x[0]+(-0.965354931124)*x[1]**o+(-0.0503295532873)*x[1]
        arg[3,2,2,0]=(0.439590816231)*x[0]**o+(-0.145314466072)*x[0]+(-0.847175364201)*x[1]**o+(-0.570000251495)*x[1]
        arg[3,2,2,1]=(0.401069335123)*x[0]**o+(-0.117139167508)*x[0]+(0.192907527853)*x[1]**o+(0.573283391581)*x[1]
        arg[3,3,0,0]=(-0.434029411666)*x[0]**o+(-0.689115317009)*x[0]+(0.63450218888)*x[1]**o+(0.492858023519)*x[1]
        arg[3,3,0,1]=(-0.678722283369)*x[0]**o+(-0.62031826393)*x[0]+(-0.458646022327)*x[1]**o+(-0.650857410927)*x[1]
        arg[3,3,1,0]=(0.413477874719)*x[0]**o+(-0.903116096879)*x[0]+(-0.505961511722)*x[1]**o+(0.649554735459)*x[1]
        arg[3,3,1,1]=(-0.58459645458)*x[0]**o+(0.861318565628)*x[0]+(0.378511063123)*x[1]**o+(-0.694820011654)*x[1]
        arg[3,3,2,0]=(0.82844733258)*x[0]**o+(-0.151929131446)*x[0]+(0.322848411081)*x[1]**o+(0.436062233263)*x[1]
        arg[3,3,2,1]=(0.269627291792)*x[0]**o+(0.618140998132)*x[0]+(0.851057431407)*x[1]**o+(0.716088832502)*x[1]
        arg[3,4,0,0]=(0.0557605432613)*x[0]**o+(0.0689569775656)*x[0]+(0.489781121774)*x[1]**o+(0.214459998733)*x[1]
        arg[3,4,0,1]=(0.429111134289)*x[0]**o+(0.648324951051)*x[0]+(0.00596799425797)*x[1]**o+(-0.752637714965)*x[1]
        arg[3,4,1,0]=(-0.0394648097179)*x[0]**o+(0.842965554058)*x[0]+(-0.124493160479)*x[1]**o+(0.81914053244)*x[1]
        arg[3,4,1,1]=(0.309465325874)*x[0]**o+(0.631381272937)*x[0]+(-0.164415508845)*x[1]**o+(0.316908911599)*x[1]
        arg[3,4,2,0]=(-0.615108314792)*x[0]**o+(0.300936829452)*x[0]+(-0.289514252333)*x[1]**o+(0.844949030767)*x[1]
        arg[3,4,2,1]=(0.80117516909)*x[0]**o+(-0.602632221639)*x[0]+(-0.68193581457)*x[1]**o+(0.203319693067)*x[1]
        ref[0,0,0,0]=(0.0737428272134)*x_ref[0]**o+(0.841780036645)*x_ref[0]+(-0.110824257531)*x_ref[1]**o+(-0.957726444468)*x_ref[1]
        ref[0,0,0,1]=(-0.934840410671)*x_ref[0]**o+(-0.83301064321)*x_ref[0]+(-0.352287650266)*x_ref[1]**o+(-0.961111564077)*x_ref[1]
        ref[0,0,1,0]=(-0.829500940226)*x_ref[0]**o+(0.644171710931)*x_ref[0]+(0.232545739722)*x_ref[1]**o+(0.584336012862)*x_ref[1]
        ref[0,0,1,1]=(0.426298830976)*x_ref[0]**o+(-0.615281197268)*x_ref[0]+(0.206334026192)*x_ref[1]**o+(-0.443698947044)*x_ref[1]
        ref[0,0,2,0]=(-0.236495891617)*x_ref[0]**o+(0.0883376851179)*x_ref[0]+(-0.176541434356)*x_ref[1]**o+(0.297175301617)*x_ref[1]
        ref[0,0,2,1]=(-0.879955324373)*x_ref[0]**o+(-0.244778913771)*x_ref[0]+(0.244131188594)*x_ref[1]**o+(-0.334739112198)*x_ref[1]
        ref[0,1,0,0]=(-0.64887128006)*x_ref[0]**o+(-0.489951454377)*x_ref[0]+(-0.328402433668)*x_ref[1]**o+(0.229814237117)*x_ref[1]
        ref[0,1,0,1]=(-0.339451656097)*x_ref[0]**o+(-0.409320682304)*x_ref[0]+(0.216482338677)*x_ref[1]**o+(-0.10594373055)*x_ref[1]
        ref[0,1,1,0]=(-0.350538479425)*x_ref[0]**o+(0.466450046985)*x_ref[0]+(-0.779124795292)*x_ref[1]**o+(-0.654562596704)*x_ref[1]
        ref[0,1,1,1]=(0.00350092879868)*x_ref[0]**o+(-0.214643088647)*x_ref[0]+(0.786934081411)*x_ref[1]**o+(-0.296853380043)*x_ref[1]
        ref[0,1,2,0]=(-0.640183955648)*x_ref[0]**o+(0.568647968237)*x_ref[0]+(0.575533476901)*x_ref[1]**o+(0.854540594085)*x_ref[1]
        ref[0,1,2,1]=(0.633674899518)*x_ref[0]**o+(0.829255679443)*x_ref[0]+(0.537888424303)*x_ref[1]**o+(-0.495082587617)*x_ref[1]
        ref[0,2,0,0]=(-0.700436798342)*x_ref[0]**o+(-0.732581859026)*x_ref[0]+(-0.508905353053)*x_ref[1]**o+(0.664601151676)*x_ref[1]
        ref[0,2,0,1]=(0.720335764302)*x_ref[0]**o+(0.941159917988)*x_ref[0]+(0.2817440547)*x_ref[1]**o+(0.553712160058)*x_ref[1]
        ref[0,2,1,0]=(-0.937114312262)*x_ref[0]**o+(-0.115517396201)*x_ref[0]+(-0.672697125905)*x_ref[1]**o+(-0.221320095369)*x_ref[1]
        ref[0,2,1,1]=(-0.52059341062)*x_ref[0]**o+(-0.653834647475)*x_ref[0]+(-0.0851100861135)*x_ref[1]**o+(-0.668300210067)*x_ref[1]
        ref[0,2,2,0]=(-0.943613035509)*x_ref[0]**o+(-0.294843145822)*x_ref[0]+(0.496278748831)*x_ref[1]**o+(0.901154419425)*x_ref[1]
        ref[0,2,2,1]=(-0.421698818203)*x_ref[0]**o+(-0.721918177544)*x_ref[0]+(-0.197855702215)*x_ref[1]**o+(0.50882827715)*x_ref[1]
        ref[0,3,0,0]=(0.931990780298)*x_ref[0]**o+(-0.550332606951)*x_ref[0]+(0.401584654642)*x_ref[1]**o+(0.0965259747601)*x_ref[1]
        ref[0,3,0,1]=(0.881357708445)*x_ref[0]**o+(0.564624034197)*x_ref[0]+(-0.458191897924)*x_ref[1]**o+(-0.427701112807)*x_ref[1]
        ref[0,3,1,0]=(0.624777088081)*x_ref[0]**o+(0.862654721208)*x_ref[0]+(-0.707618093268)*x_ref[1]**o+(0.743304953313)*x_ref[1]
        ref[0,3,1,1]=(0.4510836974)*x_ref[0]**o+(-0.357023031453)*x_ref[0]+(-0.00503924010273)*x_ref[1]**o+(0.545024336011)*x_ref[1]
        ref[0,3,2,0]=(0.109003851094)*x_ref[0]**o+(0.168659376881)*x_ref[0]+(-0.0578953780655)*x_ref[1]**o+(-0.481212462712)*x_ref[1]
        ref[0,3,2,1]=(-0.191183501636)*x_ref[0]**o+(-0.316145516377)*x_ref[0]+(0.402460103016)*x_ref[1]**o+(-0.163612464438)*x_ref[1]
        ref[0,4,0,0]=(0.489815345033)*x_ref[0]**o+(0.999138280765)*x_ref[0]+(-0.626627766812)*x_ref[1]**o+(-0.339653992568)*x_ref[1]
        ref[0,4,0,1]=(0.616350228791)*x_ref[0]**o+(0.933446438765)*x_ref[0]+(0.944244035512)*x_ref[1]**o+(0.829407789353)*x_ref[1]
        ref[0,4,1,0]=(0.619918547863)*x_ref[0]**o+(0.995488719729)*x_ref[0]+(-0.78847429581)*x_ref[1]**o+(0.542645674462)*x_ref[1]
        ref[0,4,1,1]=(-0.149244819992)*x_ref[0]**o+(0.921766299925)*x_ref[0]+(0.435255339567)*x_ref[1]**o+(-0.0596561300741)*x_ref[1]
        ref[0,4,2,0]=(-0.816364336598)*x_ref[0]**o+(0.701160246658)*x_ref[0]+(-0.616385505652)*x_ref[1]**o+(-0.558625482105)*x_ref[1]
        ref[0,4,2,1]=(-0.36944364693)*x_ref[0]**o+(0.243426977623)*x_ref[0]+(-0.61578435391)*x_ref[1]**o+(0.932570326139)*x_ref[1]
        ref[1,0,0,0]=(-0.36839962807)*x_ref[0]**o+(0.590534439534)*x_ref[0]+(-0.615158899289)*x_ref[1]**o+(0.973649158296)*x_ref[1]
        ref[1,0,0,1]=(0.761515772771)*x_ref[0]**o+(0.706406307012)*x_ref[0]+(0.328712319043)*x_ref[1]**o+(0.438600140517)*x_ref[1]
        ref[1,0,1,0]=(0.895080579207)*x_ref[0]**o+(-0.166789328225)*x_ref[0]+(0.807540897371)*x_ref[1]**o+(0.0345522692723)*x_ref[1]
        ref[1,0,1,1]=(-0.460062661172)*x_ref[0]**o+(0.505167337562)*x_ref[0]+(0.25155527144)*x_ref[1]**o+(0.601093887451)*x_ref[1]
        ref[1,0,2,0]=(0.382129818527)*x_ref[0]**o+(0.694676536542)*x_ref[0]+(0.702794119083)*x_ref[1]**o+(0.704094013056)*x_ref[1]
        ref[1,0,2,1]=(0.265476982913)*x_ref[0]**o+(-0.134338744881)*x_ref[0]+(-0.0938910335356)*x_ref[1]**o+(-0.560073316097)*x_ref[1]
        ref[1,1,0,0]=(0.977368328792)*x_ref[0]**o+(0.084482890741)*x_ref[0]+(0.57068389287)*x_ref[1]**o+(-0.242087071526)*x_ref[1]
        ref[1,1,0,1]=(-0.36067775897)*x_ref[0]**o+(0.768187470286)*x_ref[0]+(-0.881150642619)*x_ref[1]**o+(-0.15839498619)*x_ref[1]
        ref[1,1,1,0]=(-0.808263913645)*x_ref[0]**o+(0.788403321664)*x_ref[0]+(0.976206012343)*x_ref[1]**o+(-0.636933353579)*x_ref[1]
        ref[1,1,1,1]=(0.699185445546)*x_ref[0]**o+(0.251811206876)*x_ref[0]+(-0.146320762812)*x_ref[1]**o+(-0.248033048423)*x_ref[1]
        ref[1,1,2,0]=(-0.782940880812)*x_ref[0]**o+(0.789401686528)*x_ref[0]+(-0.806883904562)*x_ref[1]**o+(0.83689696772)*x_ref[1]
        ref[1,1,2,1]=(0.652404993342)*x_ref[0]**o+(-0.135318598212)*x_ref[0]+(0.138928760301)*x_ref[1]**o+(0.675593554805)*x_ref[1]
        ref[1,2,0,0]=(0.471746379585)*x_ref[0]**o+(-0.989783415847)*x_ref[0]+(-0.0935052296418)*x_ref[1]**o+(-0.532245102898)*x_ref[1]
        ref[1,2,0,1]=(-0.44541059979)*x_ref[0]**o+(0.593734266824)*x_ref[0]+(-0.616647011751)*x_ref[1]**o+(0.461140297024)*x_ref[1]
        ref[1,2,1,0]=(-0.9144210459)*x_ref[0]**o+(0.116806032197)*x_ref[0]+(0.935610395306)*x_ref[1]**o+(0.550884822213)*x_ref[1]
        ref[1,2,1,1]=(-0.962444455944)*x_ref[0]**o+(-0.654328013652)*x_ref[0]+(-0.545330994236)*x_ref[1]**o+(-0.919705456162)*x_ref[1]
        ref[1,2,2,0]=(-0.724491374462)*x_ref[0]**o+(-0.820402375498)*x_ref[0]+(-0.953956431336)*x_ref[1]**o+(-0.201450523182)*x_ref[1]
        ref[1,2,2,1]=(0.723637497881)*x_ref[0]**o+(-0.144483238523)*x_ref[0]+(0.00379745988102)*x_ref[1]**o+(-0.838616087543)*x_ref[1]
        ref[1,3,0,0]=(-0.962256046042)*x_ref[0]**o+(-0.453580322779)*x_ref[0]+(0.887446007424)*x_ref[1]**o+(-0.404454768131)*x_ref[1]
        ref[1,3,0,1]=(0.829189872775)*x_ref[0]**o+(0.327985850399)*x_ref[0]+(0.113600094307)*x_ref[1]**o+(-0.338775377551)*x_ref[1]
        ref[1,3,1,0]=(0.912320027833)*x_ref[0]**o+(0.0658152563951)*x_ref[0]+(-0.843491517568)*x_ref[1]**o+(0.858344434434)*x_ref[1]
        ref[1,3,1,1]=(-0.133918678317)*x_ref[0]**o+(0.644272729149)*x_ref[0]+(0.140663927965)*x_ref[1]**o+(0.853385355163)*x_ref[1]
        ref[1,3,2,0]=(0.37733357322)*x_ref[0]**o+(0.255105934063)*x_ref[0]+(0.88881892183)*x_ref[1]**o+(-0.722661948279)*x_ref[1]
        ref[1,3,2,1]=(-0.2827466345)*x_ref[0]**o+(-0.518815029512)*x_ref[0]+(-0.995043824929)*x_ref[1]**o+(0.973473354481)*x_ref[1]
        ref[1,4,0,0]=(0.044904732821)*x_ref[0]**o+(-0.0373604593083)*x_ref[0]+(0.184254626748)*x_ref[1]**o+(0.031832590233)*x_ref[1]
        ref[1,4,0,1]=(0.388699053911)*x_ref[0]**o+(-0.0619606394414)*x_ref[0]+(0.477947819745)*x_ref[1]**o+(0.982909216221)*x_ref[1]
        ref[1,4,1,0]=(0.854348679973)*x_ref[0]**o+(-0.822052612379)*x_ref[0]+(-0.950305089968)*x_ref[1]**o+(0.0100277256384)*x_ref[1]
        ref[1,4,1,1]=(0.265075277179)*x_ref[0]**o+(0.489846993407)*x_ref[0]+(-0.961111499722)*x_ref[1]**o+(-0.0457909055602)*x_ref[1]
        ref[1,4,2,0]=(0.701951191951)*x_ref[0]**o+(0.675232620259)*x_ref[0]+(-0.141848599227)*x_ref[1]**o+(0.737150257564)*x_ref[1]
        ref[1,4,2,1]=(-0.450684564415)*x_ref[0]**o+(0.854849922021)*x_ref[0]+(0.729670746723)*x_ref[1]**o+(0.530372628879)*x_ref[1]
        ref[2,0,0,0]=(-0.131153262675)*x_ref[0]**o+(-0.477683407979)*x_ref[0]+(0.231416752516)*x_ref[1]**o+(0.790897064402)*x_ref[1]
        ref[2,0,0,1]=(0.0956557004382)*x_ref[0]**o+(0.209084913471)*x_ref[0]+(-0.328544695733)*x_ref[1]**o+(0.701245770328)*x_ref[1]
        ref[2,0,1,0]=(-0.476403718361)*x_ref[0]**o+(0.689023533158)*x_ref[0]+(0.108732189071)*x_ref[1]**o+(0.243379977453)*x_ref[1]
        ref[2,0,1,1]=(0.414917801157)*x_ref[0]**o+(-0.843544596785)*x_ref[0]+(0.772236792935)*x_ref[1]**o+(0.836028831232)*x_ref[1]
        ref[2,0,2,0]=(0.754970600617)*x_ref[0]**o+(-0.422318891122)*x_ref[0]+(0.770767911534)*x_ref[1]**o+(-0.446737894095)*x_ref[1]
        ref[2,0,2,1]=(0.0139989331471)*x_ref[0]**o+(0.262286042291)*x_ref[0]+(-0.440109329114)*x_ref[1]**o+(-0.844965114722)*x_ref[1]
        ref[2,1,0,0]=(-0.758977153009)*x_ref[0]**o+(-0.991192011039)*x_ref[0]+(0.642267239475)*x_ref[1]**o+(-0.701360491163)*x_ref[1]
        ref[2,1,0,1]=(-0.123019007509)*x_ref[0]**o+(-0.622323252979)*x_ref[0]+(0.441364031664)*x_ref[1]**o+(0.311433222821)*x_ref[1]
        ref[2,1,1,0]=(-0.0446644182928)*x_ref[0]**o+(0.614952704493)*x_ref[0]+(-0.228284494629)*x_ref[1]**o+(0.077603612991)*x_ref[1]
        ref[2,1,1,1]=(0.712924108225)*x_ref[0]**o+(-0.153966300537)*x_ref[0]+(-0.584450014127)*x_ref[1]**o+(0.476329149833)*x_ref[1]
        ref[2,1,2,0]=(0.554026516037)*x_ref[0]**o+(0.730575821847)*x_ref[0]+(-0.247529915107)*x_ref[1]**o+(0.210060454223)*x_ref[1]
        ref[2,1,2,1]=(0.877017208752)*x_ref[0]**o+(-0.800916200997)*x_ref[0]+(0.0996710064295)*x_ref[1]**o+(-0.0890375735808)*x_ref[1]
        ref[2,2,0,0]=(-0.198026186062)*x_ref[0]**o+(-0.416160224681)*x_ref[0]+(-0.626491103997)*x_ref[1]**o+(-0.950177752043)*x_ref[1]
        ref[2,2,0,1]=(-0.401319234479)*x_ref[0]**o+(0.948120359627)*x_ref[0]+(0.420351950404)*x_ref[1]**o+(-0.242850312594)*x_ref[1]
        ref[2,2,1,0]=(-0.239652069063)*x_ref[0]**o+(-0.894811076801)*x_ref[0]+(0.979721413509)*x_ref[1]**o+(-0.712493893566)*x_ref[1]
        ref[2,2,1,1]=(0.239416494209)*x_ref[0]**o+(0.449047649538)*x_ref[0]+(0.0503751073583)*x_ref[1]**o+(0.739949331521)*x_ref[1]
        ref[2,2,2,0]=(-0.318067701034)*x_ref[0]**o+(-0.00208209268356)*x_ref[0]+(-0.513821624843)*x_ref[1]**o+(-0.202688815501)*x_ref[1]
        ref[2,2,2,1]=(-0.194111524544)*x_ref[0]**o+(-0.843213241929)*x_ref[0]+(0.124223747087)*x_ref[1]**o+(-0.988305059412)*x_ref[1]
        ref[2,3,0,0]=(0.450894655873)*x_ref[0]**o+(-0.809351657482)*x_ref[0]+(-0.80555128185)*x_ref[1]**o+(-0.922123806508)*x_ref[1]
        ref[2,3,0,1]=(-0.462822602909)*x_ref[0]**o+(0.504388766027)*x_ref[0]+(-0.152725064915)*x_ref[1]**o+(0.988745142638)*x_ref[1]
        ref[2,3,1,0]=(-0.717873944493)*x_ref[0]**o+(0.987011714613)*x_ref[0]+(0.908500737929)*x_ref[1]**o+(0.612497707425)*x_ref[1]
        ref[2,3,1,1]=(0.221645363399)*x_ref[0]**o+(0.219147687681)*x_ref[0]+(-0.195831118343)*x_ref[1]**o+(-0.59541920919)*x_ref[1]
        ref[2,3,2,0]=(-0.289487342206)*x_ref[0]**o+(-0.96904981397)*x_ref[0]+(-0.844880792969)*x_ref[1]**o+(0.803145388038)*x_ref[1]
        ref[2,3,2,1]=(-0.0234295826132)*x_ref[0]**o+(-0.839069887016)*x_ref[0]+(0.591231341444)*x_ref[1]**o+(-0.678605318787)*x_ref[1]
        ref[2,4,0,0]=(0.717943320641)*x_ref[0]**o+(0.625728011485)*x_ref[0]+(0.456836257053)*x_ref[1]**o+(-0.00523685585287)*x_ref[1]
        ref[2,4,0,1]=(-0.549526374185)*x_ref[0]**o+(0.57441917876)*x_ref[0]+(-0.913459106781)*x_ref[1]**o+(-0.610369375195)*x_ref[1]
        ref[2,4,1,0]=(-0.27166563459)*x_ref[0]**o+(0.146710581194)*x_ref[0]+(0.297934124718)*x_ref[1]**o+(0.989561209048)*x_ref[1]
        ref[2,4,1,1]=(-0.0973858709778)*x_ref[0]**o+(-0.909133394381)*x_ref[0]+(0.634182571287)*x_ref[1]**o+(0.477163052831)*x_ref[1]
        ref[2,4,2,0]=(0.0878998038301)*x_ref[0]**o+(-0.90568487021)*x_ref[0]+(0.798978355456)*x_ref[1]**o+(0.977402478037)*x_ref[1]
        ref[2,4,2,1]=(-0.456868538667)*x_ref[0]**o+(0.0723272702119)*x_ref[0]+(0.143461337462)*x_ref[1]**o+(-0.923713568343)*x_ref[1]
        ref[3,0,0,0]=(-0.154410689788)*x_ref[0]**o+(0.634925838303)*x_ref[0]+(-0.0508765762892)*x_ref[1]**o+(-0.889731180834)*x_ref[1]
        ref[3,0,0,1]=(-0.611336323518)*x_ref[0]**o+(0.748058372943)*x_ref[0]+(-0.395060568704)*x_ref[1]**o+(-0.413772262434)*x_ref[1]
        ref[3,0,1,0]=(-0.76924573047)*x_ref[0]**o+(0.61286303551)*x_ref[0]+(0.810883211884)*x_ref[1]**o+(-0.908861022244)*x_ref[1]
        ref[3,0,1,1]=(0.269999661235)*x_ref[0]**o+(0.0330954575695)*x_ref[0]+(0.560272546054)*x_ref[1]**o+(0.877730763728)*x_ref[1]
        ref[3,0,2,0]=(-0.617658215856)*x_ref[0]**o+(0.758414649277)*x_ref[0]+(-0.223497113863)*x_ref[1]**o+(-0.0706589108357)*x_ref[1]
        ref[3,0,2,1]=(0.47956571804)*x_ref[0]**o+(0.938375191345)*x_ref[0]+(-0.193997511219)*x_ref[1]**o+(0.500063249143)*x_ref[1]
        ref[3,1,0,0]=(0.994033773649)*x_ref[0]**o+(0.203022945646)*x_ref[0]+(0.439296640709)*x_ref[1]**o+(0.283679657893)*x_ref[1]
        ref[3,1,0,1]=(0.32495149734)*x_ref[0]**o+(-0.871627347694)*x_ref[0]+(-0.438159387872)*x_ref[1]**o+(0.169112665588)*x_ref[1]
        ref[3,1,1,0]=(-0.0979322945554)*x_ref[0]**o+(-0.936928842822)*x_ref[0]+(0.0402738011011)*x_ref[1]**o+(-0.445031618539)*x_ref[1]
        ref[3,1,1,1]=(-0.511218949907)*x_ref[0]**o+(0.255146779315)*x_ref[0]+(-0.802012163396)*x_ref[1]**o+(0.361819705056)*x_ref[1]
        ref[3,1,2,0]=(-0.902499203011)*x_ref[0]**o+(0.76589952141)*x_ref[0]+(0.52498929732)*x_ref[1]**o+(0.493894292321)*x_ref[1]
        ref[3,1,2,1]=(-0.475117609018)*x_ref[0]**o+(0.682205309303)*x_ref[0]+(-0.661824086649)*x_ref[1]**o+(0.985521553774)*x_ref[1]
        ref[3,2,0,0]=(0.937321686473)*x_ref[0]**o+(-0.518305775925)*x_ref[0]+(0.171689574942)*x_ref[1]**o+(-0.751579960819)*x_ref[1]
        ref[3,2,0,1]=(-0.607004597985)*x_ref[0]**o+(0.342330208344)*x_ref[0]+(0.692094243326)*x_ref[1]**o+(-0.457756057357)*x_ref[1]
        ref[3,2,1,0]=(0.126171621354)*x_ref[0]**o+(0.0499938423847)*x_ref[0]+(0.572006209576)*x_ref[1]**o+(0.121601182464)*x_ref[1]
        ref[3,2,1,1]=(0.337936456181)*x_ref[0]**o+(-0.940186907514)*x_ref[0]+(-0.965354931124)*x_ref[1]**o+(-0.0503295532873)*x_ref[1]
        ref[3,2,2,0]=(0.439590816231)*x_ref[0]**o+(-0.145314466072)*x_ref[0]+(-0.847175364201)*x_ref[1]**o+(-0.570000251495)*x_ref[1]
        ref[3,2,2,1]=(0.401069335123)*x_ref[0]**o+(-0.117139167508)*x_ref[0]+(0.192907527853)*x_ref[1]**o+(0.573283391581)*x_ref[1]
        ref[3,3,0,0]=(-0.434029411666)*x_ref[0]**o+(-0.689115317009)*x_ref[0]+(0.63450218888)*x_ref[1]**o+(0.492858023519)*x_ref[1]
        ref[3,3,0,1]=(-0.678722283369)*x_ref[0]**o+(-0.62031826393)*x_ref[0]+(-0.458646022327)*x_ref[1]**o+(-0.650857410927)*x_ref[1]
        ref[3,3,1,0]=(0.413477874719)*x_ref[0]**o+(-0.903116096879)*x_ref[0]+(-0.505961511722)*x_ref[1]**o+(0.649554735459)*x_ref[1]
        ref[3,3,1,1]=(-0.58459645458)*x_ref[0]**o+(0.861318565628)*x_ref[0]+(0.378511063123)*x_ref[1]**o+(-0.694820011654)*x_ref[1]
        ref[3,3,2,0]=(0.82844733258)*x_ref[0]**o+(-0.151929131446)*x_ref[0]+(0.322848411081)*x_ref[1]**o+(0.436062233263)*x_ref[1]
        ref[3,3,2,1]=(0.269627291792)*x_ref[0]**o+(0.618140998132)*x_ref[0]+(0.851057431407)*x_ref[1]**o+(0.716088832502)*x_ref[1]
        ref[3,4,0,0]=(0.0557605432613)*x_ref[0]**o+(0.0689569775656)*x_ref[0]+(0.489781121774)*x_ref[1]**o+(0.214459998733)*x_ref[1]
        ref[3,4,0,1]=(0.429111134289)*x_ref[0]**o+(0.648324951051)*x_ref[0]+(0.00596799425797)*x_ref[1]**o+(-0.752637714965)*x_ref[1]
        ref[3,4,1,0]=(-0.0394648097179)*x_ref[0]**o+(0.842965554058)*x_ref[0]+(-0.124493160479)*x_ref[1]**o+(0.81914053244)*x_ref[1]
        ref[3,4,1,1]=(0.309465325874)*x_ref[0]**o+(0.631381272937)*x_ref[0]+(-0.164415508845)*x_ref[1]**o+(0.316908911599)*x_ref[1]
        ref[3,4,2,0]=(-0.615108314792)*x_ref[0]**o+(0.300936829452)*x_ref[0]+(-0.289514252333)*x_ref[1]**o+(0.844949030767)*x_ref[1]
        ref[3,4,2,1]=(0.80117516909)*x_ref[0]**o+(-0.602632221639)*x_ref[0]+(-0.68193581457)*x_ref[1]**o+(0.203319693067)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.349437323395)*x[0]**o+(0.817821891642)*x[0]+(-0.357998520848)*x[1]**o+(0.349311985429)*x[1]+(0.470351907942)*x[2]**o+(0.264007843212)*x[2]
        arg[0,0,0,1]=(-0.747824148035)*x[0]**o+(0.551817215468)*x[0]+(0.980476067966)*x[1]**o+(0.521120126285)*x[1]+(-0.472357720779)*x[2]**o+(-0.379502066444)*x[2]
        arg[0,0,1,0]=(0.302984899868)*x[0]**o+(0.556601505796)*x[0]+(-0.698948476338)*x[1]**o+(-0.587379572024)*x[1]+(-0.625760359909)*x[2]**o+(0.231956543432)*x[2]
        arg[0,0,1,1]=(-0.335943560463)*x[0]**o+(-0.497643152155)*x[0]+(-0.849829024888)*x[1]**o+(-0.388751135318)*x[1]+(-0.517696637743)*x[2]**o+(-0.966210901)*x[2]
        arg[0,0,2,0]=(-0.0953058405321)*x[0]**o+(-0.864580918175)*x[0]+(-0.732086619685)*x[1]**o+(0.872044654816)*x[1]+(-0.225850391375)*x[2]**o+(-0.502203654371)*x[2]
        arg[0,0,2,1]=(0.228408069101)*x[0]**o+(0.267590362458)*x[0]+(0.127397041031)*x[1]**o+(0.0142393720782)*x[1]+(-0.257563412478)*x[2]**o+(0.0211877877507)*x[2]
        arg[0,1,0,0]=(0.921125039994)*x[0]**o+(0.372548304521)*x[0]+(-0.66574364308)*x[1]**o+(0.0895600219155)*x[1]+(-0.656870562037)*x[2]**o+(-0.608819638268)*x[2]
        arg[0,1,0,1]=(-0.396625252579)*x[0]**o+(0.655542320616)*x[0]+(-0.542629502059)*x[1]**o+(0.587317299384)*x[1]+(0.96380848967)*x[2]**o+(0.121341827766)*x[2]
        arg[0,1,1,0]=(0.0908208346003)*x[0]**o+(-0.0398227321951)*x[0]+(0.907607555288)*x[1]**o+(0.527230589013)*x[1]+(0.519287416653)*x[2]**o+(0.103180358249)*x[2]
        arg[0,1,1,1]=(0.0615195625837)*x[0]**o+(0.695720276474)*x[0]+(0.0862101684049)*x[1]**o+(0.517150419955)*x[1]+(0.511207669575)*x[2]**o+(-0.9556472362)*x[2]
        arg[0,1,2,0]=(0.70040395652)*x[0]**o+(-0.576266027865)*x[0]+(-0.133947982034)*x[1]**o+(-0.764976368456)*x[1]+(-0.860780506782)*x[2]**o+(-0.137946340663)*x[2]
        arg[0,1,2,1]=(0.356017596832)*x[0]**o+(0.94912783604)*x[0]+(0.325951760294)*x[1]**o+(0.0623695610721)*x[1]+(-0.106031051895)*x[2]**o+(-0.339314970581)*x[2]
        arg[0,2,0,0]=(-0.124954395016)*x[0]**o+(0.558233271782)*x[0]+(-0.956846820394)*x[1]**o+(-0.599332648704)*x[1]+(-0.322444201477)*x[2]**o+(0.0833108732532)*x[2]
        arg[0,2,0,1]=(-0.713065259589)*x[0]**o+(-0.597746726477)*x[0]+(-0.976654446591)*x[1]**o+(0.977779698697)*x[1]+(0.694427679422)*x[2]**o+(-0.364463986712)*x[2]
        arg[0,2,1,0]=(-0.415861468352)*x[0]**o+(-0.159802964908)*x[0]+(-0.132331280778)*x[1]**o+(0.120829594848)*x[1]+(-0.114181951611)*x[2]**o+(-0.810187313564)*x[2]
        arg[0,2,1,1]=(-0.869925695451)*x[0]**o+(-0.535444611026)*x[0]+(0.62908837594)*x[1]**o+(-0.488583607314)*x[1]+(0.8461615049)*x[2]**o+(-0.558010294727)*x[2]
        arg[0,2,2,0]=(-0.0822232217466)*x[0]**o+(-0.673631382141)*x[0]+(0.803717792107)*x[1]**o+(0.339934715615)*x[1]+(-0.886365924996)*x[2]**o+(0.0206517750753)*x[2]
        arg[0,2,2,1]=(0.367656249869)*x[0]**o+(0.000987963028539)*x[0]+(0.399182215451)*x[1]**o+(0.923259635138)*x[1]+(0.8573338122)*x[2]**o+(0.679082213415)*x[2]
        arg[0,3,0,0]=(-0.722680496644)*x[0]**o+(0.199427912651)*x[0]+(-0.351799245099)*x[1]**o+(0.240285416231)*x[1]+(-0.701704868242)*x[2]**o+(0.265610436018)*x[2]
        arg[0,3,0,1]=(-0.14479043328)*x[0]**o+(-0.192670959504)*x[0]+(0.327412080296)*x[1]**o+(-0.390894637576)*x[1]+(-0.401862813677)*x[2]**o+(-0.87150071561)*x[2]
        arg[0,3,1,0]=(0.430039400239)*x[0]**o+(-0.0691443536291)*x[0]+(0.205492039071)*x[1]**o+(-0.391580258314)*x[1]+(0.791748125642)*x[2]**o+(-0.0896722225431)*x[2]
        arg[0,3,1,1]=(-0.177554766913)*x[0]**o+(-0.399389195454)*x[0]+(-0.304228953477)*x[1]**o+(0.00462595697167)*x[1]+(0.271814579884)*x[2]**o+(-0.17305639368)*x[2]
        arg[0,3,2,0]=(-0.105869589157)*x[0]**o+(0.803488342924)*x[0]+(-0.65117861655)*x[1]**o+(-0.27529059731)*x[1]+(-0.777107387582)*x[2]**o+(-0.690980328369)*x[2]
        arg[0,3,2,1]=(-0.13792230295)*x[0]**o+(0.649721420509)*x[0]+(0.401441756822)*x[1]**o+(0.468949007973)*x[1]+(-0.762015699412)*x[2]**o+(0.389781367434)*x[2]
        arg[0,4,0,0]=(0.873653819021)*x[0]**o+(0.539406798913)*x[0]+(0.802177801924)*x[1]**o+(-0.861314989922)*x[1]+(0.847509960118)*x[2]**o+(-0.41381965291)*x[2]
        arg[0,4,0,1]=(0.0416858647633)*x[0]**o+(0.866825467268)*x[0]+(0.89776163679)*x[1]**o+(-0.978326218858)*x[1]+(-0.409795413277)*x[2]**o+(0.982449327977)*x[2]
        arg[0,4,1,0]=(-0.114530966075)*x[0]**o+(-0.182414280034)*x[0]+(0.154385814218)*x[1]**o+(0.869624332271)*x[1]+(0.477213291128)*x[2]**o+(-0.137123375612)*x[2]
        arg[0,4,1,1]=(-0.52963019184)*x[0]**o+(0.0217955878025)*x[0]+(-0.174531798693)*x[1]**o+(-0.738947178141)*x[1]+(0.198130667502)*x[2]**o+(-0.449952166994)*x[2]
        arg[0,4,2,0]=(0.998209708224)*x[0]**o+(0.734906841374)*x[0]+(0.225596363429)*x[1]**o+(0.862772101878)*x[1]+(-0.133563132638)*x[2]**o+(-0.78059821331)*x[2]
        arg[0,4,2,1]=(0.384580893901)*x[0]**o+(-0.881918368669)*x[0]+(-0.950768236608)*x[1]**o+(-0.452275072093)*x[1]+(0.592779023744)*x[2]**o+(0.798735695084)*x[2]
        arg[1,0,0,0]=(0.786827894179)*x[0]**o+(-0.233241290567)*x[0]+(0.946717698992)*x[1]**o+(0.179368479607)*x[1]+(-0.10409968587)*x[2]**o+(-0.770115022756)*x[2]
        arg[1,0,0,1]=(-0.355389699793)*x[0]**o+(0.356407960494)*x[0]+(-0.371831767713)*x[1]**o+(0.139120259823)*x[1]+(0.910761795881)*x[2]**o+(0.725966050775)*x[2]
        arg[1,0,1,0]=(0.316372320948)*x[0]**o+(0.432472869369)*x[0]+(0.733187356771)*x[1]**o+(0.133915829584)*x[1]+(0.380713482112)*x[2]**o+(-0.874827460114)*x[2]
        arg[1,0,1,1]=(-0.0625786527083)*x[0]**o+(-0.469518973019)*x[0]+(-0.318610570327)*x[1]**o+(-0.304412247656)*x[1]+(0.844057224107)*x[2]**o+(0.508926494536)*x[2]
        arg[1,0,2,0]=(-0.759697281694)*x[0]**o+(0.701456398418)*x[0]+(-0.671612899124)*x[1]**o+(0.292474197717)*x[1]+(-0.445886723952)*x[2]**o+(-0.319953406782)*x[2]
        arg[1,0,2,1]=(-0.865703619781)*x[0]**o+(0.115639820869)*x[0]+(0.493634056223)*x[1]**o+(0.0343245670418)*x[1]+(0.748200872819)*x[2]**o+(-0.208438395054)*x[2]
        arg[1,1,0,0]=(0.767661930656)*x[0]**o+(-0.578385615919)*x[0]+(0.241313268557)*x[1]**o+(0.74276605359)*x[1]+(0.373638986644)*x[2]**o+(0.475479466784)*x[2]
        arg[1,1,0,1]=(0.723228882826)*x[0]**o+(-0.988524686322)*x[0]+(0.333664096597)*x[1]**o+(-0.756435323592)*x[1]+(0.771173510221)*x[2]**o+(-0.879337556561)*x[2]
        arg[1,1,1,0]=(0.939682105701)*x[0]**o+(-0.172311578817)*x[0]+(-0.407241492831)*x[1]**o+(0.446660064834)*x[1]+(0.0371203367053)*x[2]**o+(-0.302853430561)*x[2]
        arg[1,1,1,1]=(0.1093213068)*x[0]**o+(0.31232933407)*x[0]+(0.47474581408)*x[1]**o+(-0.602763201317)*x[1]+(0.803198523872)*x[2]**o+(-0.178088233667)*x[2]
        arg[1,1,2,0]=(-0.217141686114)*x[0]**o+(0.287363815362)*x[0]+(-0.352552947551)*x[1]**o+(0.106445580983)*x[1]+(0.340814562712)*x[2]**o+(-0.963691249733)*x[2]
        arg[1,1,2,1]=(0.612302241126)*x[0]**o+(-0.771765095612)*x[0]+(-0.18662763339)*x[1]**o+(0.10259882275)*x[1]+(-0.29535785913)*x[2]**o+(-0.577876877465)*x[2]
        arg[1,2,0,0]=(0.497131459554)*x[0]**o+(0.912133134267)*x[0]+(0.944369633113)*x[1]**o+(0.163408823815)*x[1]+(0.85636766551)*x[2]**o+(-0.00282458078743)*x[2]
        arg[1,2,0,1]=(0.362490806338)*x[0]**o+(0.22490192994)*x[0]+(-0.913665309243)*x[1]**o+(-0.88010395132)*x[1]+(-0.926205930269)*x[2]**o+(-0.231830082111)*x[2]
        arg[1,2,1,0]=(-0.974792784505)*x[0]**o+(-0.765605618923)*x[0]+(0.241470054253)*x[1]**o+(0.445863114014)*x[1]+(0.561062749944)*x[2]**o+(-0.215674406206)*x[2]
        arg[1,2,1,1]=(0.827013456823)*x[0]**o+(-0.191864800977)*x[0]+(-0.158099224794)*x[1]**o+(-0.921959562711)*x[1]+(0.930693535629)*x[2]**o+(-0.352967183792)*x[2]
        arg[1,2,2,0]=(0.258970603305)*x[0]**o+(0.586779837275)*x[0]+(0.82707087137)*x[1]**o+(-0.215210535461)*x[1]+(0.36890299705)*x[2]**o+(-0.724385890088)*x[2]
        arg[1,2,2,1]=(-0.125969269863)*x[0]**o+(-0.505216539273)*x[0]+(0.43242008053)*x[1]**o+(0.212079268112)*x[1]+(0.345759190477)*x[2]**o+(-0.98505468199)*x[2]
        arg[1,3,0,0]=(-0.647118215098)*x[0]**o+(0.318566502446)*x[0]+(-0.151406735047)*x[1]**o+(0.253202745125)*x[1]+(-0.404823111329)*x[2]**o+(-0.520797366557)*x[2]
        arg[1,3,0,1]=(-0.946886311645)*x[0]**o+(0.733699524424)*x[0]+(-0.615470980807)*x[1]**o+(-0.103976314414)*x[1]+(-0.932104431883)*x[2]**o+(0.157563806559)*x[2]
        arg[1,3,1,0]=(-0.289471503181)*x[0]**o+(0.749124853135)*x[0]+(-0.578861403092)*x[1]**o+(-0.136231378985)*x[1]+(0.532719246315)*x[2]**o+(0.542455831657)*x[2]
        arg[1,3,1,1]=(-0.343842490684)*x[0]**o+(0.110039182926)*x[0]+(0.901234650534)*x[1]**o+(-0.513362876241)*x[1]+(0.371952189126)*x[2]**o+(-0.169400349862)*x[2]
        arg[1,3,2,0]=(-0.766791944503)*x[0]**o+(0.0267491122166)*x[0]+(-0.699293509597)*x[1]**o+(-0.697363669971)*x[1]+(0.567854170308)*x[2]**o+(-0.516301389629)*x[2]
        arg[1,3,2,1]=(-0.288508579358)*x[0]**o+(0.239321880772)*x[0]+(0.677249635795)*x[1]**o+(0.48369084768)*x[1]+(0.681460633595)*x[2]**o+(-0.80218734437)*x[2]
        arg[1,4,0,0]=(0.510273388936)*x[0]**o+(0.35600020965)*x[0]+(0.271077500616)*x[1]**o+(-0.00479335810576)*x[1]+(-0.348457583464)*x[2]**o+(-0.965895122152)*x[2]
        arg[1,4,0,1]=(0.0134532748813)*x[0]**o+(-0.396954380686)*x[0]+(-0.176682002946)*x[1]**o+(0.971224261138)*x[1]+(-0.359602180261)*x[2]**o+(0.373748703363)*x[2]
        arg[1,4,1,0]=(-0.519413160925)*x[0]**o+(0.856346255965)*x[0]+(0.159167243935)*x[1]**o+(0.92068779838)*x[1]+(0.415122182947)*x[2]**o+(-0.595702311496)*x[2]
        arg[1,4,1,1]=(-0.608969598184)*x[0]**o+(-0.434750765676)*x[0]+(0.74471475701)*x[1]**o+(-0.229333246615)*x[1]+(-0.678138032736)*x[2]**o+(-0.514112926798)*x[2]
        arg[1,4,2,0]=(0.0633453561735)*x[0]**o+(-0.278569473082)*x[0]+(-0.842227144473)*x[1]**o+(-0.00289799815652)*x[1]+(0.153392291113)*x[2]**o+(-0.61448940993)*x[2]
        arg[1,4,2,1]=(0.0897863401674)*x[0]**o+(-0.487876787706)*x[0]+(-0.464008634014)*x[1]**o+(0.576466896344)*x[1]+(0.962345361537)*x[2]**o+(0.580878819156)*x[2]
        arg[2,0,0,0]=(0.316582275434)*x[0]**o+(0.0607872538221)*x[0]+(0.642492366848)*x[1]**o+(0.200300167843)*x[1]+(-0.402802396184)*x[2]**o+(-0.834254866233)*x[2]
        arg[2,0,0,1]=(0.644614184272)*x[0]**o+(-0.086367798694)*x[0]+(-0.428671462705)*x[1]**o+(-0.517827373575)*x[1]+(0.528197945165)*x[2]**o+(-0.523092888005)*x[2]
        arg[2,0,1,0]=(0.41376145869)*x[0]**o+(-0.27425322418)*x[0]+(0.816942168928)*x[1]**o+(-0.882784454814)*x[1]+(-0.270205535673)*x[2]**o+(0.642372102899)*x[2]
        arg[2,0,1,1]=(0.584931736844)*x[0]**o+(0.622511792702)*x[0]+(0.76174625736)*x[1]**o+(0.0561486448557)*x[1]+(-0.588861269552)*x[2]**o+(-0.341896323569)*x[2]
        arg[2,0,2,0]=(0.635215108566)*x[0]**o+(0.299769133671)*x[0]+(0.141053572388)*x[1]**o+(0.102844255879)*x[1]+(-0.995154777534)*x[2]**o+(0.501319388994)*x[2]
        arg[2,0,2,1]=(0.750089643836)*x[0]**o+(0.57570708561)*x[0]+(-0.449935823319)*x[1]**o+(-0.902982109411)*x[1]+(0.561185740222)*x[2]**o+(0.422654056847)*x[2]
        arg[2,1,0,0]=(0.953986415316)*x[0]**o+(-0.7883326473)*x[0]+(-0.723690234059)*x[1]**o+(-0.537242781471)*x[1]+(-0.503466717519)*x[2]**o+(-0.777865169241)*x[2]
        arg[2,1,0,1]=(0.317357554441)*x[0]**o+(-0.326828957855)*x[0]+(-0.609992651237)*x[1]**o+(0.811066934743)*x[1]+(-0.90668441447)*x[2]**o+(0.92581926687)*x[2]
        arg[2,1,1,0]=(-0.0801236189368)*x[0]**o+(0.49151264813)*x[0]+(-0.0770654446151)*x[1]**o+(-0.314981217276)*x[1]+(-0.743180837478)*x[2]**o+(-0.554491413146)*x[2]
        arg[2,1,1,1]=(0.831017344937)*x[0]**o+(-0.61064793935)*x[0]+(0.652976737027)*x[1]**o+(0.242104127969)*x[1]+(0.593330958678)*x[2]**o+(-0.35899584298)*x[2]
        arg[2,1,2,0]=(-0.580618798556)*x[0]**o+(-0.936176474082)*x[0]+(-0.768716477288)*x[1]**o+(-0.353687473159)*x[1]+(-0.261509314019)*x[2]**o+(-0.442139199257)*x[2]
        arg[2,1,2,1]=(0.945760793889)*x[0]**o+(-0.646958670891)*x[0]+(-0.440341726063)*x[1]**o+(0.223283751014)*x[1]+(0.799798944743)*x[2]**o+(0.591902067633)*x[2]
        arg[2,2,0,0]=(-0.267699746405)*x[0]**o+(0.271554172399)*x[0]+(0.747278903642)*x[1]**o+(-0.832877459834)*x[1]+(-0.640373164716)*x[2]**o+(-0.666487164763)*x[2]
        arg[2,2,0,1]=(0.795630715924)*x[0]**o+(0.682973866186)*x[0]+(-0.750818874926)*x[1]**o+(-0.904782408217)*x[1]+(-0.578138718379)*x[2]**o+(-0.369846516992)*x[2]
        arg[2,2,1,0]=(0.165654497062)*x[0]**o+(-0.329827871165)*x[0]+(-0.808981530672)*x[1]**o+(0.350106763785)*x[1]+(0.106638087689)*x[2]**o+(0.425516820398)*x[2]
        arg[2,2,1,1]=(0.017419119421)*x[0]**o+(0.500359047996)*x[0]+(-0.888509940049)*x[1]**o+(-0.375469124891)*x[1]+(0.863718970618)*x[2]**o+(-0.807667604763)*x[2]
        arg[2,2,2,0]=(0.860662652165)*x[0]**o+(0.162530730499)*x[0]+(-0.657510966961)*x[1]**o+(0.215905840286)*x[1]+(0.97380371264)*x[2]**o+(-0.431027671897)*x[2]
        arg[2,2,2,1]=(-0.649151503991)*x[0]**o+(0.484589779517)*x[0]+(0.905080495553)*x[1]**o+(0.997504119204)*x[1]+(-0.4268802512)*x[2]**o+(-0.687236308612)*x[2]
        arg[2,3,0,0]=(-0.60485902589)*x[0]**o+(0.49500777286)*x[0]+(0.33977532104)*x[1]**o+(-0.219209845338)*x[1]+(-0.197311627683)*x[2]**o+(0.285859591044)*x[2]
        arg[2,3,0,1]=(0.117438340483)*x[0]**o+(-0.562954894877)*x[0]+(-0.302982842797)*x[1]**o+(0.977923453436)*x[1]+(0.300127749784)*x[2]**o+(0.45784201373)*x[2]
        arg[2,3,1,0]=(-0.910061442266)*x[0]**o+(0.723114553174)*x[0]+(-0.435693478652)*x[1]**o+(-0.584326584996)*x[1]+(-0.51000911177)*x[2]**o+(0.506615483889)*x[2]
        arg[2,3,1,1]=(-0.50179741763)*x[0]**o+(-0.954551290001)*x[0]+(0.375839266192)*x[1]**o+(0.122120198059)*x[1]+(-0.593050674641)*x[2]**o+(0.549989594312)*x[2]
        arg[2,3,2,0]=(-0.889519666705)*x[0]**o+(-0.815572507674)*x[0]+(0.453677651057)*x[1]**o+(0.450910828217)*x[1]+(-0.00465373008682)*x[2]**o+(0.731899900165)*x[2]
        arg[2,3,2,1]=(-0.969971151589)*x[0]**o+(-0.156948104522)*x[0]+(0.492885029353)*x[1]**o+(-0.0310677847833)*x[1]+(0.327614233837)*x[2]**o+(-0.274974040707)*x[2]
        arg[2,4,0,0]=(0.844667900826)*x[0]**o+(0.863374870522)*x[0]+(0.251456923241)*x[1]**o+(0.434916005183)*x[1]+(-0.139023370677)*x[2]**o+(0.0655176616053)*x[2]
        arg[2,4,0,1]=(0.831755059318)*x[0]**o+(0.0897642674136)*x[0]+(0.190414878812)*x[1]**o+(0.976565327031)*x[1]+(0.413321767371)*x[2]**o+(0.755395415341)*x[2]
        arg[2,4,1,0]=(-0.790364876377)*x[0]**o+(0.829442094248)*x[0]+(0.556251403891)*x[1]**o+(-0.953487711658)*x[1]+(-0.610781905401)*x[2]**o+(-0.320764072326)*x[2]
        arg[2,4,1,1]=(-0.757101565136)*x[0]**o+(0.789340839395)*x[0]+(-0.615189514438)*x[1]**o+(-0.910992676978)*x[1]+(0.693736813401)*x[2]**o+(0.0538707588266)*x[2]
        arg[2,4,2,0]=(0.000335625616611)*x[0]**o+(-0.195171446476)*x[0]+(0.0281149133705)*x[1]**o+(-0.362294159243)*x[1]+(0.830520288743)*x[2]**o+(-0.431967694247)*x[2]
        arg[2,4,2,1]=(-0.232668364843)*x[0]**o+(0.655302452302)*x[0]+(-0.0125806408337)*x[1]**o+(-0.103979202109)*x[1]+(-0.277450383033)*x[2]**o+(0.708914971885)*x[2]
        arg[3,0,0,0]=(-0.813447875186)*x[0]**o+(0.133491078101)*x[0]+(-0.675510725453)*x[1]**o+(-0.161489202717)*x[1]+(0.0474778506607)*x[2]**o+(0.513729029652)*x[2]
        arg[3,0,0,1]=(0.45388386743)*x[0]**o+(-0.909074944715)*x[0]+(-0.540576382376)*x[1]**o+(-0.849753706068)*x[1]+(-0.496320191697)*x[2]**o+(-0.930725196363)*x[2]
        arg[3,0,1,0]=(-0.145526830518)*x[0]**o+(-0.34717641097)*x[0]+(-0.818886792022)*x[1]**o+(0.574315644784)*x[1]+(-0.780431329222)*x[2]**o+(0.892840747651)*x[2]
        arg[3,0,1,1]=(-0.415342962745)*x[0]**o+(-0.899768945518)*x[0]+(0.798573910731)*x[1]**o+(-0.0652893933479)*x[1]+(-0.444459519602)*x[2]**o+(-0.0100131702565)*x[2]
        arg[3,0,2,0]=(-0.997936676013)*x[0]**o+(0.822000858069)*x[0]+(-0.553794408949)*x[1]**o+(0.829037287423)*x[1]+(-0.700288437806)*x[2]**o+(-0.449584600891)*x[2]
        arg[3,0,2,1]=(-0.70979868598)*x[0]**o+(0.689868687655)*x[0]+(0.151257146397)*x[1]**o+(0.32895215144)*x[1]+(0.286043871383)*x[2]**o+(-0.604421567781)*x[2]
        arg[3,1,0,0]=(0.60431849542)*x[0]**o+(0.518618291627)*x[0]+(-0.471868097013)*x[1]**o+(-0.639110899151)*x[1]+(-0.0904184061601)*x[2]**o+(0.113762528304)*x[2]
        arg[3,1,0,1]=(0.307624103157)*x[0]**o+(-0.388281302162)*x[0]+(0.772429578102)*x[1]**o+(-0.197799629585)*x[1]+(-0.149120667036)*x[2]**o+(0.830154780264)*x[2]
        arg[3,1,1,0]=(0.957127413133)*x[0]**o+(0.730080945389)*x[0]+(0.497191948774)*x[1]**o+(-0.586439221663)*x[1]+(-0.56223071093)*x[2]**o+(0.162899395804)*x[2]
        arg[3,1,1,1]=(-0.285740818451)*x[0]**o+(-0.573919970411)*x[0]+(-0.400833211669)*x[1]**o+(0.84414534634)*x[1]+(-0.124352096731)*x[2]**o+(0.907013800834)*x[2]
        arg[3,1,2,0]=(0.186947347989)*x[0]**o+(0.499582449213)*x[0]+(0.525473636695)*x[1]**o+(-0.926500953998)*x[1]+(-0.0392615366415)*x[2]**o+(0.909002911192)*x[2]
        arg[3,1,2,1]=(0.634350440683)*x[0]**o+(0.901699291665)*x[0]+(-0.683899667575)*x[1]**o+(0.411642037744)*x[1]+(-0.72791611585)*x[2]**o+(-0.013293689839)*x[2]
        arg[3,2,0,0]=(0.185164429507)*x[0]**o+(-0.671011652257)*x[0]+(0.973361040829)*x[1]**o+(0.989738565899)*x[1]+(-0.501245887091)*x[2]**o+(-0.405445444386)*x[2]
        arg[3,2,0,1]=(-0.311554020884)*x[0]**o+(-0.929020990333)*x[0]+(0.366016375797)*x[1]**o+(0.838234804773)*x[1]+(0.939602632409)*x[2]**o+(-0.885524260372)*x[2]
        arg[3,2,1,0]=(-0.0101604655589)*x[0]**o+(0.670274078493)*x[0]+(0.198135855354)*x[1]**o+(-0.000275422963644)*x[1]+(0.136674233404)*x[2]**o+(-0.442510118497)*x[2]
        arg[3,2,1,1]=(-0.677948769785)*x[0]**o+(-0.165403610813)*x[0]+(-0.219353502981)*x[1]**o+(0.664297154009)*x[1]+(0.158536302892)*x[2]**o+(0.293940884925)*x[2]
        arg[3,2,2,0]=(0.666669534708)*x[0]**o+(0.609459298502)*x[0]+(0.297058070582)*x[1]**o+(-0.0809555825821)*x[1]+(0.139430791042)*x[2]**o+(-0.253514861624)*x[2]
        arg[3,2,2,1]=(-0.577932413274)*x[0]**o+(-0.583353108895)*x[0]+(-0.839262034394)*x[1]**o+(0.795642236106)*x[1]+(0.0882627381543)*x[2]**o+(-0.382970623122)*x[2]
        arg[3,3,0,0]=(0.187941627429)*x[0]**o+(0.552040688531)*x[0]+(0.837458832716)*x[1]**o+(0.296793121897)*x[1]+(-0.665699511216)*x[2]**o+(-0.318565991052)*x[2]
        arg[3,3,0,1]=(-0.78770176804)*x[0]**o+(0.807509260396)*x[0]+(0.46241086085)*x[1]**o+(0.668506908751)*x[1]+(-0.988954457343)*x[2]**o+(0.0896365679915)*x[2]
        arg[3,3,1,0]=(0.727665746433)*x[0]**o+(-0.819391676739)*x[0]+(-0.118648934607)*x[1]**o+(-0.433862939031)*x[1]+(0.827788711313)*x[2]**o+(-0.615751446784)*x[2]
        arg[3,3,1,1]=(-0.00427901841482)*x[0]**o+(-0.38545713036)*x[0]+(0.570160083404)*x[1]**o+(-0.235505162795)*x[1]+(0.514578304371)*x[2]**o+(-0.152692113279)*x[2]
        arg[3,3,2,0]=(-0.396900613831)*x[0]**o+(0.161875314543)*x[0]+(0.28471295984)*x[1]**o+(-0.623276596728)*x[1]+(-0.748772373219)*x[2]**o+(0.584246898704)*x[2]
        arg[3,3,2,1]=(0.210026027009)*x[0]**o+(-0.530437530387)*x[0]+(0.437155825716)*x[1]**o+(0.984102407204)*x[1]+(0.125030773605)*x[2]**o+(-0.821764907318)*x[2]
        arg[3,4,0,0]=(-0.819129946783)*x[0]**o+(-0.729517565451)*x[0]+(-0.126013244602)*x[1]**o+(0.592832125254)*x[1]+(-0.850426487488)*x[2]**o+(0.271550912174)*x[2]
        arg[3,4,0,1]=(-0.573554440921)*x[0]**o+(0.232106829584)*x[0]+(0.810201614042)*x[1]**o+(-0.85518164211)*x[1]+(0.249535563949)*x[2]**o+(-0.69728963017)*x[2]
        arg[3,4,1,0]=(0.785661643628)*x[0]**o+(0.936424600424)*x[0]+(0.321363811055)*x[1]**o+(0.362341458592)*x[1]+(-0.593576579363)*x[2]**o+(0.0601278435924)*x[2]
        arg[3,4,1,1]=(-0.854420686885)*x[0]**o+(0.993352097501)*x[0]+(-0.572116112224)*x[1]**o+(0.308008845182)*x[1]+(0.0443339578639)*x[2]**o+(-0.1469857285)*x[2]
        arg[3,4,2,0]=(0.696373642151)*x[0]**o+(0.82425126832)*x[0]+(-0.0891735358455)*x[1]**o+(-0.22218993613)*x[1]+(-0.238393149017)*x[2]**o+(-0.631352718194)*x[2]
        arg[3,4,2,1]=(-0.0932690431928)*x[0]**o+(-0.391900483919)*x[0]+(0.236843565361)*x[1]**o+(-0.164197648532)*x[1]+(0.235502596353)*x[2]**o+(0.0110131867917)*x[2]
        ref[0,0,0,0]=(0.349437323395)*x_ref[0]**o+(0.817821891642)*x_ref[0]+(-0.357998520848)*x_ref[1]**o+(0.349311985429)*x_ref[1]+(0.470351907942)*x_ref[2]**o+(0.264007843212)*x_ref[2]
        ref[0,0,0,1]=(-0.747824148035)*x_ref[0]**o+(0.551817215468)*x_ref[0]+(0.980476067966)*x_ref[1]**o+(0.521120126285)*x_ref[1]+(-0.472357720779)*x_ref[2]**o+(-0.379502066444)*x_ref[2]
        ref[0,0,1,0]=(0.302984899868)*x_ref[0]**o+(0.556601505796)*x_ref[0]+(-0.698948476338)*x_ref[1]**o+(-0.587379572024)*x_ref[1]+(-0.625760359909)*x_ref[2]**o+(0.231956543432)*x_ref[2]
        ref[0,0,1,1]=(-0.335943560463)*x_ref[0]**o+(-0.497643152155)*x_ref[0]+(-0.849829024888)*x_ref[1]**o+(-0.388751135318)*x_ref[1]+(-0.517696637743)*x_ref[2]**o+(-0.966210901)*x_ref[2]
        ref[0,0,2,0]=(-0.0953058405321)*x_ref[0]**o+(-0.864580918175)*x_ref[0]+(-0.732086619685)*x_ref[1]**o+(0.872044654816)*x_ref[1]+(-0.225850391375)*x_ref[2]**o+(-0.502203654371)*x_ref[2]
        ref[0,0,2,1]=(0.228408069101)*x_ref[0]**o+(0.267590362458)*x_ref[0]+(0.127397041031)*x_ref[1]**o+(0.0142393720782)*x_ref[1]+(-0.257563412478)*x_ref[2]**o+(0.0211877877507)*x_ref[2]
        ref[0,1,0,0]=(0.921125039994)*x_ref[0]**o+(0.372548304521)*x_ref[0]+(-0.66574364308)*x_ref[1]**o+(0.0895600219155)*x_ref[1]+(-0.656870562037)*x_ref[2]**o+(-0.608819638268)*x_ref[2]
        ref[0,1,0,1]=(-0.396625252579)*x_ref[0]**o+(0.655542320616)*x_ref[0]+(-0.542629502059)*x_ref[1]**o+(0.587317299384)*x_ref[1]+(0.96380848967)*x_ref[2]**o+(0.121341827766)*x_ref[2]
        ref[0,1,1,0]=(0.0908208346003)*x_ref[0]**o+(-0.0398227321951)*x_ref[0]+(0.907607555288)*x_ref[1]**o+(0.527230589013)*x_ref[1]+(0.519287416653)*x_ref[2]**o+(0.103180358249)*x_ref[2]
        ref[0,1,1,1]=(0.0615195625837)*x_ref[0]**o+(0.695720276474)*x_ref[0]+(0.0862101684049)*x_ref[1]**o+(0.517150419955)*x_ref[1]+(0.511207669575)*x_ref[2]**o+(-0.9556472362)*x_ref[2]
        ref[0,1,2,0]=(0.70040395652)*x_ref[0]**o+(-0.576266027865)*x_ref[0]+(-0.133947982034)*x_ref[1]**o+(-0.764976368456)*x_ref[1]+(-0.860780506782)*x_ref[2]**o+(-0.137946340663)*x_ref[2]
        ref[0,1,2,1]=(0.356017596832)*x_ref[0]**o+(0.94912783604)*x_ref[0]+(0.325951760294)*x_ref[1]**o+(0.0623695610721)*x_ref[1]+(-0.106031051895)*x_ref[2]**o+(-0.339314970581)*x_ref[2]
        ref[0,2,0,0]=(-0.124954395016)*x_ref[0]**o+(0.558233271782)*x_ref[0]+(-0.956846820394)*x_ref[1]**o+(-0.599332648704)*x_ref[1]+(-0.322444201477)*x_ref[2]**o+(0.0833108732532)*x_ref[2]
        ref[0,2,0,1]=(-0.713065259589)*x_ref[0]**o+(-0.597746726477)*x_ref[0]+(-0.976654446591)*x_ref[1]**o+(0.977779698697)*x_ref[1]+(0.694427679422)*x_ref[2]**o+(-0.364463986712)*x_ref[2]
        ref[0,2,1,0]=(-0.415861468352)*x_ref[0]**o+(-0.159802964908)*x_ref[0]+(-0.132331280778)*x_ref[1]**o+(0.120829594848)*x_ref[1]+(-0.114181951611)*x_ref[2]**o+(-0.810187313564)*x_ref[2]
        ref[0,2,1,1]=(-0.869925695451)*x_ref[0]**o+(-0.535444611026)*x_ref[0]+(0.62908837594)*x_ref[1]**o+(-0.488583607314)*x_ref[1]+(0.8461615049)*x_ref[2]**o+(-0.558010294727)*x_ref[2]
        ref[0,2,2,0]=(-0.0822232217466)*x_ref[0]**o+(-0.673631382141)*x_ref[0]+(0.803717792107)*x_ref[1]**o+(0.339934715615)*x_ref[1]+(-0.886365924996)*x_ref[2]**o+(0.0206517750753)*x_ref[2]
        ref[0,2,2,1]=(0.367656249869)*x_ref[0]**o+(0.000987963028539)*x_ref[0]+(0.399182215451)*x_ref[1]**o+(0.923259635138)*x_ref[1]+(0.8573338122)*x_ref[2]**o+(0.679082213415)*x_ref[2]
        ref[0,3,0,0]=(-0.722680496644)*x_ref[0]**o+(0.199427912651)*x_ref[0]+(-0.351799245099)*x_ref[1]**o+(0.240285416231)*x_ref[1]+(-0.701704868242)*x_ref[2]**o+(0.265610436018)*x_ref[2]
        ref[0,3,0,1]=(-0.14479043328)*x_ref[0]**o+(-0.192670959504)*x_ref[0]+(0.327412080296)*x_ref[1]**o+(-0.390894637576)*x_ref[1]+(-0.401862813677)*x_ref[2]**o+(-0.87150071561)*x_ref[2]
        ref[0,3,1,0]=(0.430039400239)*x_ref[0]**o+(-0.0691443536291)*x_ref[0]+(0.205492039071)*x_ref[1]**o+(-0.391580258314)*x_ref[1]+(0.791748125642)*x_ref[2]**o+(-0.0896722225431)*x_ref[2]
        ref[0,3,1,1]=(-0.177554766913)*x_ref[0]**o+(-0.399389195454)*x_ref[0]+(-0.304228953477)*x_ref[1]**o+(0.00462595697167)*x_ref[1]+(0.271814579884)*x_ref[2]**o+(-0.17305639368)*x_ref[2]
        ref[0,3,2,0]=(-0.105869589157)*x_ref[0]**o+(0.803488342924)*x_ref[0]+(-0.65117861655)*x_ref[1]**o+(-0.27529059731)*x_ref[1]+(-0.777107387582)*x_ref[2]**o+(-0.690980328369)*x_ref[2]
        ref[0,3,2,1]=(-0.13792230295)*x_ref[0]**o+(0.649721420509)*x_ref[0]+(0.401441756822)*x_ref[1]**o+(0.468949007973)*x_ref[1]+(-0.762015699412)*x_ref[2]**o+(0.389781367434)*x_ref[2]
        ref[0,4,0,0]=(0.873653819021)*x_ref[0]**o+(0.539406798913)*x_ref[0]+(0.802177801924)*x_ref[1]**o+(-0.861314989922)*x_ref[1]+(0.847509960118)*x_ref[2]**o+(-0.41381965291)*x_ref[2]
        ref[0,4,0,1]=(0.0416858647633)*x_ref[0]**o+(0.866825467268)*x_ref[0]+(0.89776163679)*x_ref[1]**o+(-0.978326218858)*x_ref[1]+(-0.409795413277)*x_ref[2]**o+(0.982449327977)*x_ref[2]
        ref[0,4,1,0]=(-0.114530966075)*x_ref[0]**o+(-0.182414280034)*x_ref[0]+(0.154385814218)*x_ref[1]**o+(0.869624332271)*x_ref[1]+(0.477213291128)*x_ref[2]**o+(-0.137123375612)*x_ref[2]
        ref[0,4,1,1]=(-0.52963019184)*x_ref[0]**o+(0.0217955878025)*x_ref[0]+(-0.174531798693)*x_ref[1]**o+(-0.738947178141)*x_ref[1]+(0.198130667502)*x_ref[2]**o+(-0.449952166994)*x_ref[2]
        ref[0,4,2,0]=(0.998209708224)*x_ref[0]**o+(0.734906841374)*x_ref[0]+(0.225596363429)*x_ref[1]**o+(0.862772101878)*x_ref[1]+(-0.133563132638)*x_ref[2]**o+(-0.78059821331)*x_ref[2]
        ref[0,4,2,1]=(0.384580893901)*x_ref[0]**o+(-0.881918368669)*x_ref[0]+(-0.950768236608)*x_ref[1]**o+(-0.452275072093)*x_ref[1]+(0.592779023744)*x_ref[2]**o+(0.798735695084)*x_ref[2]
        ref[1,0,0,0]=(0.786827894179)*x_ref[0]**o+(-0.233241290567)*x_ref[0]+(0.946717698992)*x_ref[1]**o+(0.179368479607)*x_ref[1]+(-0.10409968587)*x_ref[2]**o+(-0.770115022756)*x_ref[2]
        ref[1,0,0,1]=(-0.355389699793)*x_ref[0]**o+(0.356407960494)*x_ref[0]+(-0.371831767713)*x_ref[1]**o+(0.139120259823)*x_ref[1]+(0.910761795881)*x_ref[2]**o+(0.725966050775)*x_ref[2]
        ref[1,0,1,0]=(0.316372320948)*x_ref[0]**o+(0.432472869369)*x_ref[0]+(0.733187356771)*x_ref[1]**o+(0.133915829584)*x_ref[1]+(0.380713482112)*x_ref[2]**o+(-0.874827460114)*x_ref[2]
        ref[1,0,1,1]=(-0.0625786527083)*x_ref[0]**o+(-0.469518973019)*x_ref[0]+(-0.318610570327)*x_ref[1]**o+(-0.304412247656)*x_ref[1]+(0.844057224107)*x_ref[2]**o+(0.508926494536)*x_ref[2]
        ref[1,0,2,0]=(-0.759697281694)*x_ref[0]**o+(0.701456398418)*x_ref[0]+(-0.671612899124)*x_ref[1]**o+(0.292474197717)*x_ref[1]+(-0.445886723952)*x_ref[2]**o+(-0.319953406782)*x_ref[2]
        ref[1,0,2,1]=(-0.865703619781)*x_ref[0]**o+(0.115639820869)*x_ref[0]+(0.493634056223)*x_ref[1]**o+(0.0343245670418)*x_ref[1]+(0.748200872819)*x_ref[2]**o+(-0.208438395054)*x_ref[2]
        ref[1,1,0,0]=(0.767661930656)*x_ref[0]**o+(-0.578385615919)*x_ref[0]+(0.241313268557)*x_ref[1]**o+(0.74276605359)*x_ref[1]+(0.373638986644)*x_ref[2]**o+(0.475479466784)*x_ref[2]
        ref[1,1,0,1]=(0.723228882826)*x_ref[0]**o+(-0.988524686322)*x_ref[0]+(0.333664096597)*x_ref[1]**o+(-0.756435323592)*x_ref[1]+(0.771173510221)*x_ref[2]**o+(-0.879337556561)*x_ref[2]
        ref[1,1,1,0]=(0.939682105701)*x_ref[0]**o+(-0.172311578817)*x_ref[0]+(-0.407241492831)*x_ref[1]**o+(0.446660064834)*x_ref[1]+(0.0371203367053)*x_ref[2]**o+(-0.302853430561)*x_ref[2]
        ref[1,1,1,1]=(0.1093213068)*x_ref[0]**o+(0.31232933407)*x_ref[0]+(0.47474581408)*x_ref[1]**o+(-0.602763201317)*x_ref[1]+(0.803198523872)*x_ref[2]**o+(-0.178088233667)*x_ref[2]
        ref[1,1,2,0]=(-0.217141686114)*x_ref[0]**o+(0.287363815362)*x_ref[0]+(-0.352552947551)*x_ref[1]**o+(0.106445580983)*x_ref[1]+(0.340814562712)*x_ref[2]**o+(-0.963691249733)*x_ref[2]
        ref[1,1,2,1]=(0.612302241126)*x_ref[0]**o+(-0.771765095612)*x_ref[0]+(-0.18662763339)*x_ref[1]**o+(0.10259882275)*x_ref[1]+(-0.29535785913)*x_ref[2]**o+(-0.577876877465)*x_ref[2]
        ref[1,2,0,0]=(0.497131459554)*x_ref[0]**o+(0.912133134267)*x_ref[0]+(0.944369633113)*x_ref[1]**o+(0.163408823815)*x_ref[1]+(0.85636766551)*x_ref[2]**o+(-0.00282458078743)*x_ref[2]
        ref[1,2,0,1]=(0.362490806338)*x_ref[0]**o+(0.22490192994)*x_ref[0]+(-0.913665309243)*x_ref[1]**o+(-0.88010395132)*x_ref[1]+(-0.926205930269)*x_ref[2]**o+(-0.231830082111)*x_ref[2]
        ref[1,2,1,0]=(-0.974792784505)*x_ref[0]**o+(-0.765605618923)*x_ref[0]+(0.241470054253)*x_ref[1]**o+(0.445863114014)*x_ref[1]+(0.561062749944)*x_ref[2]**o+(-0.215674406206)*x_ref[2]
        ref[1,2,1,1]=(0.827013456823)*x_ref[0]**o+(-0.191864800977)*x_ref[0]+(-0.158099224794)*x_ref[1]**o+(-0.921959562711)*x_ref[1]+(0.930693535629)*x_ref[2]**o+(-0.352967183792)*x_ref[2]
        ref[1,2,2,0]=(0.258970603305)*x_ref[0]**o+(0.586779837275)*x_ref[0]+(0.82707087137)*x_ref[1]**o+(-0.215210535461)*x_ref[1]+(0.36890299705)*x_ref[2]**o+(-0.724385890088)*x_ref[2]
        ref[1,2,2,1]=(-0.125969269863)*x_ref[0]**o+(-0.505216539273)*x_ref[0]+(0.43242008053)*x_ref[1]**o+(0.212079268112)*x_ref[1]+(0.345759190477)*x_ref[2]**o+(-0.98505468199)*x_ref[2]
        ref[1,3,0,0]=(-0.647118215098)*x_ref[0]**o+(0.318566502446)*x_ref[0]+(-0.151406735047)*x_ref[1]**o+(0.253202745125)*x_ref[1]+(-0.404823111329)*x_ref[2]**o+(-0.520797366557)*x_ref[2]
        ref[1,3,0,1]=(-0.946886311645)*x_ref[0]**o+(0.733699524424)*x_ref[0]+(-0.615470980807)*x_ref[1]**o+(-0.103976314414)*x_ref[1]+(-0.932104431883)*x_ref[2]**o+(0.157563806559)*x_ref[2]
        ref[1,3,1,0]=(-0.289471503181)*x_ref[0]**o+(0.749124853135)*x_ref[0]+(-0.578861403092)*x_ref[1]**o+(-0.136231378985)*x_ref[1]+(0.532719246315)*x_ref[2]**o+(0.542455831657)*x_ref[2]
        ref[1,3,1,1]=(-0.343842490684)*x_ref[0]**o+(0.110039182926)*x_ref[0]+(0.901234650534)*x_ref[1]**o+(-0.513362876241)*x_ref[1]+(0.371952189126)*x_ref[2]**o+(-0.169400349862)*x_ref[2]
        ref[1,3,2,0]=(-0.766791944503)*x_ref[0]**o+(0.0267491122166)*x_ref[0]+(-0.699293509597)*x_ref[1]**o+(-0.697363669971)*x_ref[1]+(0.567854170308)*x_ref[2]**o+(-0.516301389629)*x_ref[2]
        ref[1,3,2,1]=(-0.288508579358)*x_ref[0]**o+(0.239321880772)*x_ref[0]+(0.677249635795)*x_ref[1]**o+(0.48369084768)*x_ref[1]+(0.681460633595)*x_ref[2]**o+(-0.80218734437)*x_ref[2]
        ref[1,4,0,0]=(0.510273388936)*x_ref[0]**o+(0.35600020965)*x_ref[0]+(0.271077500616)*x_ref[1]**o+(-0.00479335810576)*x_ref[1]+(-0.348457583464)*x_ref[2]**o+(-0.965895122152)*x_ref[2]
        ref[1,4,0,1]=(0.0134532748813)*x_ref[0]**o+(-0.396954380686)*x_ref[0]+(-0.176682002946)*x_ref[1]**o+(0.971224261138)*x_ref[1]+(-0.359602180261)*x_ref[2]**o+(0.373748703363)*x_ref[2]
        ref[1,4,1,0]=(-0.519413160925)*x_ref[0]**o+(0.856346255965)*x_ref[0]+(0.159167243935)*x_ref[1]**o+(0.92068779838)*x_ref[1]+(0.415122182947)*x_ref[2]**o+(-0.595702311496)*x_ref[2]
        ref[1,4,1,1]=(-0.608969598184)*x_ref[0]**o+(-0.434750765676)*x_ref[0]+(0.74471475701)*x_ref[1]**o+(-0.229333246615)*x_ref[1]+(-0.678138032736)*x_ref[2]**o+(-0.514112926798)*x_ref[2]
        ref[1,4,2,0]=(0.0633453561735)*x_ref[0]**o+(-0.278569473082)*x_ref[0]+(-0.842227144473)*x_ref[1]**o+(-0.00289799815652)*x_ref[1]+(0.153392291113)*x_ref[2]**o+(-0.61448940993)*x_ref[2]
        ref[1,4,2,1]=(0.0897863401674)*x_ref[0]**o+(-0.487876787706)*x_ref[0]+(-0.464008634014)*x_ref[1]**o+(0.576466896344)*x_ref[1]+(0.962345361537)*x_ref[2]**o+(0.580878819156)*x_ref[2]
        ref[2,0,0,0]=(0.316582275434)*x_ref[0]**o+(0.0607872538221)*x_ref[0]+(0.642492366848)*x_ref[1]**o+(0.200300167843)*x_ref[1]+(-0.402802396184)*x_ref[2]**o+(-0.834254866233)*x_ref[2]
        ref[2,0,0,1]=(0.644614184272)*x_ref[0]**o+(-0.086367798694)*x_ref[0]+(-0.428671462705)*x_ref[1]**o+(-0.517827373575)*x_ref[1]+(0.528197945165)*x_ref[2]**o+(-0.523092888005)*x_ref[2]
        ref[2,0,1,0]=(0.41376145869)*x_ref[0]**o+(-0.27425322418)*x_ref[0]+(0.816942168928)*x_ref[1]**o+(-0.882784454814)*x_ref[1]+(-0.270205535673)*x_ref[2]**o+(0.642372102899)*x_ref[2]
        ref[2,0,1,1]=(0.584931736844)*x_ref[0]**o+(0.622511792702)*x_ref[0]+(0.76174625736)*x_ref[1]**o+(0.0561486448557)*x_ref[1]+(-0.588861269552)*x_ref[2]**o+(-0.341896323569)*x_ref[2]
        ref[2,0,2,0]=(0.635215108566)*x_ref[0]**o+(0.299769133671)*x_ref[0]+(0.141053572388)*x_ref[1]**o+(0.102844255879)*x_ref[1]+(-0.995154777534)*x_ref[2]**o+(0.501319388994)*x_ref[2]
        ref[2,0,2,1]=(0.750089643836)*x_ref[0]**o+(0.57570708561)*x_ref[0]+(-0.449935823319)*x_ref[1]**o+(-0.902982109411)*x_ref[1]+(0.561185740222)*x_ref[2]**o+(0.422654056847)*x_ref[2]
        ref[2,1,0,0]=(0.953986415316)*x_ref[0]**o+(-0.7883326473)*x_ref[0]+(-0.723690234059)*x_ref[1]**o+(-0.537242781471)*x_ref[1]+(-0.503466717519)*x_ref[2]**o+(-0.777865169241)*x_ref[2]
        ref[2,1,0,1]=(0.317357554441)*x_ref[0]**o+(-0.326828957855)*x_ref[0]+(-0.609992651237)*x_ref[1]**o+(0.811066934743)*x_ref[1]+(-0.90668441447)*x_ref[2]**o+(0.92581926687)*x_ref[2]
        ref[2,1,1,0]=(-0.0801236189368)*x_ref[0]**o+(0.49151264813)*x_ref[0]+(-0.0770654446151)*x_ref[1]**o+(-0.314981217276)*x_ref[1]+(-0.743180837478)*x_ref[2]**o+(-0.554491413146)*x_ref[2]
        ref[2,1,1,1]=(0.831017344937)*x_ref[0]**o+(-0.61064793935)*x_ref[0]+(0.652976737027)*x_ref[1]**o+(0.242104127969)*x_ref[1]+(0.593330958678)*x_ref[2]**o+(-0.35899584298)*x_ref[2]
        ref[2,1,2,0]=(-0.580618798556)*x_ref[0]**o+(-0.936176474082)*x_ref[0]+(-0.768716477288)*x_ref[1]**o+(-0.353687473159)*x_ref[1]+(-0.261509314019)*x_ref[2]**o+(-0.442139199257)*x_ref[2]
        ref[2,1,2,1]=(0.945760793889)*x_ref[0]**o+(-0.646958670891)*x_ref[0]+(-0.440341726063)*x_ref[1]**o+(0.223283751014)*x_ref[1]+(0.799798944743)*x_ref[2]**o+(0.591902067633)*x_ref[2]
        ref[2,2,0,0]=(-0.267699746405)*x_ref[0]**o+(0.271554172399)*x_ref[0]+(0.747278903642)*x_ref[1]**o+(-0.832877459834)*x_ref[1]+(-0.640373164716)*x_ref[2]**o+(-0.666487164763)*x_ref[2]
        ref[2,2,0,1]=(0.795630715924)*x_ref[0]**o+(0.682973866186)*x_ref[0]+(-0.750818874926)*x_ref[1]**o+(-0.904782408217)*x_ref[1]+(-0.578138718379)*x_ref[2]**o+(-0.369846516992)*x_ref[2]
        ref[2,2,1,0]=(0.165654497062)*x_ref[0]**o+(-0.329827871165)*x_ref[0]+(-0.808981530672)*x_ref[1]**o+(0.350106763785)*x_ref[1]+(0.106638087689)*x_ref[2]**o+(0.425516820398)*x_ref[2]
        ref[2,2,1,1]=(0.017419119421)*x_ref[0]**o+(0.500359047996)*x_ref[0]+(-0.888509940049)*x_ref[1]**o+(-0.375469124891)*x_ref[1]+(0.863718970618)*x_ref[2]**o+(-0.807667604763)*x_ref[2]
        ref[2,2,2,0]=(0.860662652165)*x_ref[0]**o+(0.162530730499)*x_ref[0]+(-0.657510966961)*x_ref[1]**o+(0.215905840286)*x_ref[1]+(0.97380371264)*x_ref[2]**o+(-0.431027671897)*x_ref[2]
        ref[2,2,2,1]=(-0.649151503991)*x_ref[0]**o+(0.484589779517)*x_ref[0]+(0.905080495553)*x_ref[1]**o+(0.997504119204)*x_ref[1]+(-0.4268802512)*x_ref[2]**o+(-0.687236308612)*x_ref[2]
        ref[2,3,0,0]=(-0.60485902589)*x_ref[0]**o+(0.49500777286)*x_ref[0]+(0.33977532104)*x_ref[1]**o+(-0.219209845338)*x_ref[1]+(-0.197311627683)*x_ref[2]**o+(0.285859591044)*x_ref[2]
        ref[2,3,0,1]=(0.117438340483)*x_ref[0]**o+(-0.562954894877)*x_ref[0]+(-0.302982842797)*x_ref[1]**o+(0.977923453436)*x_ref[1]+(0.300127749784)*x_ref[2]**o+(0.45784201373)*x_ref[2]
        ref[2,3,1,0]=(-0.910061442266)*x_ref[0]**o+(0.723114553174)*x_ref[0]+(-0.435693478652)*x_ref[1]**o+(-0.584326584996)*x_ref[1]+(-0.51000911177)*x_ref[2]**o+(0.506615483889)*x_ref[2]
        ref[2,3,1,1]=(-0.50179741763)*x_ref[0]**o+(-0.954551290001)*x_ref[0]+(0.375839266192)*x_ref[1]**o+(0.122120198059)*x_ref[1]+(-0.593050674641)*x_ref[2]**o+(0.549989594312)*x_ref[2]
        ref[2,3,2,0]=(-0.889519666705)*x_ref[0]**o+(-0.815572507674)*x_ref[0]+(0.453677651057)*x_ref[1]**o+(0.450910828217)*x_ref[1]+(-0.00465373008682)*x_ref[2]**o+(0.731899900165)*x_ref[2]
        ref[2,3,2,1]=(-0.969971151589)*x_ref[0]**o+(-0.156948104522)*x_ref[0]+(0.492885029353)*x_ref[1]**o+(-0.0310677847833)*x_ref[1]+(0.327614233837)*x_ref[2]**o+(-0.274974040707)*x_ref[2]
        ref[2,4,0,0]=(0.844667900826)*x_ref[0]**o+(0.863374870522)*x_ref[0]+(0.251456923241)*x_ref[1]**o+(0.434916005183)*x_ref[1]+(-0.139023370677)*x_ref[2]**o+(0.0655176616053)*x_ref[2]
        ref[2,4,0,1]=(0.831755059318)*x_ref[0]**o+(0.0897642674136)*x_ref[0]+(0.190414878812)*x_ref[1]**o+(0.976565327031)*x_ref[1]+(0.413321767371)*x_ref[2]**o+(0.755395415341)*x_ref[2]
        ref[2,4,1,0]=(-0.790364876377)*x_ref[0]**o+(0.829442094248)*x_ref[0]+(0.556251403891)*x_ref[1]**o+(-0.953487711658)*x_ref[1]+(-0.610781905401)*x_ref[2]**o+(-0.320764072326)*x_ref[2]
        ref[2,4,1,1]=(-0.757101565136)*x_ref[0]**o+(0.789340839395)*x_ref[0]+(-0.615189514438)*x_ref[1]**o+(-0.910992676978)*x_ref[1]+(0.693736813401)*x_ref[2]**o+(0.0538707588266)*x_ref[2]
        ref[2,4,2,0]=(0.000335625616611)*x_ref[0]**o+(-0.195171446476)*x_ref[0]+(0.0281149133705)*x_ref[1]**o+(-0.362294159243)*x_ref[1]+(0.830520288743)*x_ref[2]**o+(-0.431967694247)*x_ref[2]
        ref[2,4,2,1]=(-0.232668364843)*x_ref[0]**o+(0.655302452302)*x_ref[0]+(-0.0125806408337)*x_ref[1]**o+(-0.103979202109)*x_ref[1]+(-0.277450383033)*x_ref[2]**o+(0.708914971885)*x_ref[2]
        ref[3,0,0,0]=(-0.813447875186)*x_ref[0]**o+(0.133491078101)*x_ref[0]+(-0.675510725453)*x_ref[1]**o+(-0.161489202717)*x_ref[1]+(0.0474778506607)*x_ref[2]**o+(0.513729029652)*x_ref[2]
        ref[3,0,0,1]=(0.45388386743)*x_ref[0]**o+(-0.909074944715)*x_ref[0]+(-0.540576382376)*x_ref[1]**o+(-0.849753706068)*x_ref[1]+(-0.496320191697)*x_ref[2]**o+(-0.930725196363)*x_ref[2]
        ref[3,0,1,0]=(-0.145526830518)*x_ref[0]**o+(-0.34717641097)*x_ref[0]+(-0.818886792022)*x_ref[1]**o+(0.574315644784)*x_ref[1]+(-0.780431329222)*x_ref[2]**o+(0.892840747651)*x_ref[2]
        ref[3,0,1,1]=(-0.415342962745)*x_ref[0]**o+(-0.899768945518)*x_ref[0]+(0.798573910731)*x_ref[1]**o+(-0.0652893933479)*x_ref[1]+(-0.444459519602)*x_ref[2]**o+(-0.0100131702565)*x_ref[2]
        ref[3,0,2,0]=(-0.997936676013)*x_ref[0]**o+(0.822000858069)*x_ref[0]+(-0.553794408949)*x_ref[1]**o+(0.829037287423)*x_ref[1]+(-0.700288437806)*x_ref[2]**o+(-0.449584600891)*x_ref[2]
        ref[3,0,2,1]=(-0.70979868598)*x_ref[0]**o+(0.689868687655)*x_ref[0]+(0.151257146397)*x_ref[1]**o+(0.32895215144)*x_ref[1]+(0.286043871383)*x_ref[2]**o+(-0.604421567781)*x_ref[2]
        ref[3,1,0,0]=(0.60431849542)*x_ref[0]**o+(0.518618291627)*x_ref[0]+(-0.471868097013)*x_ref[1]**o+(-0.639110899151)*x_ref[1]+(-0.0904184061601)*x_ref[2]**o+(0.113762528304)*x_ref[2]
        ref[3,1,0,1]=(0.307624103157)*x_ref[0]**o+(-0.388281302162)*x_ref[0]+(0.772429578102)*x_ref[1]**o+(-0.197799629585)*x_ref[1]+(-0.149120667036)*x_ref[2]**o+(0.830154780264)*x_ref[2]
        ref[3,1,1,0]=(0.957127413133)*x_ref[0]**o+(0.730080945389)*x_ref[0]+(0.497191948774)*x_ref[1]**o+(-0.586439221663)*x_ref[1]+(-0.56223071093)*x_ref[2]**o+(0.162899395804)*x_ref[2]
        ref[3,1,1,1]=(-0.285740818451)*x_ref[0]**o+(-0.573919970411)*x_ref[0]+(-0.400833211669)*x_ref[1]**o+(0.84414534634)*x_ref[1]+(-0.124352096731)*x_ref[2]**o+(0.907013800834)*x_ref[2]
        ref[3,1,2,0]=(0.186947347989)*x_ref[0]**o+(0.499582449213)*x_ref[0]+(0.525473636695)*x_ref[1]**o+(-0.926500953998)*x_ref[1]+(-0.0392615366415)*x_ref[2]**o+(0.909002911192)*x_ref[2]
        ref[3,1,2,1]=(0.634350440683)*x_ref[0]**o+(0.901699291665)*x_ref[0]+(-0.683899667575)*x_ref[1]**o+(0.411642037744)*x_ref[1]+(-0.72791611585)*x_ref[2]**o+(-0.013293689839)*x_ref[2]
        ref[3,2,0,0]=(0.185164429507)*x_ref[0]**o+(-0.671011652257)*x_ref[0]+(0.973361040829)*x_ref[1]**o+(0.989738565899)*x_ref[1]+(-0.501245887091)*x_ref[2]**o+(-0.405445444386)*x_ref[2]
        ref[3,2,0,1]=(-0.311554020884)*x_ref[0]**o+(-0.929020990333)*x_ref[0]+(0.366016375797)*x_ref[1]**o+(0.838234804773)*x_ref[1]+(0.939602632409)*x_ref[2]**o+(-0.885524260372)*x_ref[2]
        ref[3,2,1,0]=(-0.0101604655589)*x_ref[0]**o+(0.670274078493)*x_ref[0]+(0.198135855354)*x_ref[1]**o+(-0.000275422963644)*x_ref[1]+(0.136674233404)*x_ref[2]**o+(-0.442510118497)*x_ref[2]
        ref[3,2,1,1]=(-0.677948769785)*x_ref[0]**o+(-0.165403610813)*x_ref[0]+(-0.219353502981)*x_ref[1]**o+(0.664297154009)*x_ref[1]+(0.158536302892)*x_ref[2]**o+(0.293940884925)*x_ref[2]
        ref[3,2,2,0]=(0.666669534708)*x_ref[0]**o+(0.609459298502)*x_ref[0]+(0.297058070582)*x_ref[1]**o+(-0.0809555825821)*x_ref[1]+(0.139430791042)*x_ref[2]**o+(-0.253514861624)*x_ref[2]
        ref[3,2,2,1]=(-0.577932413274)*x_ref[0]**o+(-0.583353108895)*x_ref[0]+(-0.839262034394)*x_ref[1]**o+(0.795642236106)*x_ref[1]+(0.0882627381543)*x_ref[2]**o+(-0.382970623122)*x_ref[2]
        ref[3,3,0,0]=(0.187941627429)*x_ref[0]**o+(0.552040688531)*x_ref[0]+(0.837458832716)*x_ref[1]**o+(0.296793121897)*x_ref[1]+(-0.665699511216)*x_ref[2]**o+(-0.318565991052)*x_ref[2]
        ref[3,3,0,1]=(-0.78770176804)*x_ref[0]**o+(0.807509260396)*x_ref[0]+(0.46241086085)*x_ref[1]**o+(0.668506908751)*x_ref[1]+(-0.988954457343)*x_ref[2]**o+(0.0896365679915)*x_ref[2]
        ref[3,3,1,0]=(0.727665746433)*x_ref[0]**o+(-0.819391676739)*x_ref[0]+(-0.118648934607)*x_ref[1]**o+(-0.433862939031)*x_ref[1]+(0.827788711313)*x_ref[2]**o+(-0.615751446784)*x_ref[2]
        ref[3,3,1,1]=(-0.00427901841482)*x_ref[0]**o+(-0.38545713036)*x_ref[0]+(0.570160083404)*x_ref[1]**o+(-0.235505162795)*x_ref[1]+(0.514578304371)*x_ref[2]**o+(-0.152692113279)*x_ref[2]
        ref[3,3,2,0]=(-0.396900613831)*x_ref[0]**o+(0.161875314543)*x_ref[0]+(0.28471295984)*x_ref[1]**o+(-0.623276596728)*x_ref[1]+(-0.748772373219)*x_ref[2]**o+(0.584246898704)*x_ref[2]
        ref[3,3,2,1]=(0.210026027009)*x_ref[0]**o+(-0.530437530387)*x_ref[0]+(0.437155825716)*x_ref[1]**o+(0.984102407204)*x_ref[1]+(0.125030773605)*x_ref[2]**o+(-0.821764907318)*x_ref[2]
        ref[3,4,0,0]=(-0.819129946783)*x_ref[0]**o+(-0.729517565451)*x_ref[0]+(-0.126013244602)*x_ref[1]**o+(0.592832125254)*x_ref[1]+(-0.850426487488)*x_ref[2]**o+(0.271550912174)*x_ref[2]
        ref[3,4,0,1]=(-0.573554440921)*x_ref[0]**o+(0.232106829584)*x_ref[0]+(0.810201614042)*x_ref[1]**o+(-0.85518164211)*x_ref[1]+(0.249535563949)*x_ref[2]**o+(-0.69728963017)*x_ref[2]
        ref[3,4,1,0]=(0.785661643628)*x_ref[0]**o+(0.936424600424)*x_ref[0]+(0.321363811055)*x_ref[1]**o+(0.362341458592)*x_ref[1]+(-0.593576579363)*x_ref[2]**o+(0.0601278435924)*x_ref[2]
        ref[3,4,1,1]=(-0.854420686885)*x_ref[0]**o+(0.993352097501)*x_ref[0]+(-0.572116112224)*x_ref[1]**o+(0.308008845182)*x_ref[1]+(0.0443339578639)*x_ref[2]**o+(-0.1469857285)*x_ref[2]
        ref[3,4,2,0]=(0.696373642151)*x_ref[0]**o+(0.82425126832)*x_ref[0]+(-0.0891735358455)*x_ref[1]**o+(-0.22218993613)*x_ref[1]+(-0.238393149017)*x_ref[2]**o+(-0.631352718194)*x_ref[2]
        ref[3,4,2,1]=(-0.0932690431928)*x_ref[0]**o+(-0.391900483919)*x_ref[0]+(0.236843565361)*x_ref[1]**o+(-0.164197648532)*x_ref[1]+(0.235502596353)*x_ref[2]**o+(0.0110131867917)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.592324160495)*x[0]**o+(-0.0471428147361)*x[0]+(0.756452478895)*x[1]**o+(-0.531318490356)*x[1]
        ref=(-0.592324160495)*x_ref[0]**o+(-0.0471428147361)*x_ref[0]+(0.756452478895)*x_ref[1]**o+(-0.531318490356)*x_ref[1]
      else:
        s=(0.935975918364)*x[0]**o+(-0.0518611788088)*x[0]+(0.198153832638)*x[1]**o+(0.0915392556709)*x[1]+(-0.740073697392)*x[2]**o+(0.0344998139597)*x[2]
        ref=(0.935975918364)*x_ref[0]**o+(-0.0518611788088)*x_ref[0]+(0.198153832638)*x_ref[1]**o+(0.0915392556709)*x_ref[1]+(-0.740073697392)*x_ref[2]**o+(0.0344998139597)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.115732010972)*x[0]**o+(-0.244595740544)*x[0]+(-0.627201201607)*x[1]**o+(-0.0597882966954)*x[1]
        s[1]=(0.0303875091965)*x[0]**o+(0.474828783855)*x[0]+(0.0299053774471)*x[1]**o+(0.364050762964)*x[1]
        ref[0]=(0.115732010972)*x_ref[0]**o+(-0.244595740544)*x_ref[0]+(-0.627201201607)*x_ref[1]**o+(-0.0597882966954)*x_ref[1]
        ref[1]=(0.0303875091965)*x_ref[0]**o+(0.474828783855)*x_ref[0]+(0.0299053774471)*x_ref[1]**o+(0.364050762964)*x_ref[1]
      else:
        s[0]=(0.645056057607)*x[0]**o+(0.154023831925)*x[0]+(0.121000626682)*x[1]**o+(-0.962672948686)*x[1]+(0.997016990346)*x[2]**o+(-0.746450513341)*x[2]
        s[1]=(0.880153248744)*x[0]**o+(0.399741188143)*x[0]+(-0.862453529609)*x[1]**o+(0.777403830466)*x[1]+(-0.759299926221)*x[2]**o+(0.271056515624)*x[2]
        ref[0]=(0.645056057607)*x_ref[0]**o+(0.154023831925)*x_ref[0]+(0.121000626682)*x_ref[1]**o+(-0.962672948686)*x_ref[1]+(0.997016990346)*x_ref[2]**o+(-0.746450513341)*x_ref[2]
        ref[1]=(0.880153248744)*x_ref[0]**o+(0.399741188143)*x_ref[0]+(-0.862453529609)*x_ref[1]**o+(0.777403830466)*x_ref[1]+(-0.759299926221)*x_ref[2]**o+(0.271056515624)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.163301494275)*x[0]**o+(-0.571871996499)*x[0]+(-0.126318698091)*x[1]**o+(0.594323280745)*x[1]
        s[0,1]=(-0.722891165327)*x[0]**o+(0.539458396201)*x[0]+(-0.338410805174)*x[1]**o+(-0.0441039302518)*x[1]
        s[0,2]=(0.987165940018)*x[0]**o+(0.277540309671)*x[0]+(-0.194843676722)*x[1]**o+(-0.838659115747)*x[1]
        s[0,3]=(0.668488628751)*x[0]**o+(0.848828025004)*x[0]+(-0.749811369279)*x[1]**o+(-0.857737537329)*x[1]
        s[0,4]=(0.292839673829)*x[0]**o+(-0.196773033161)*x[0]+(-0.365394929162)*x[1]**o+(-0.304841309615)*x[1]
        s[1,0]=(-0.244415346064)*x[0]**o+(0.0653469288689)*x[0]+(-0.405830560862)*x[1]**o+(0.958797302829)*x[1]
        s[1,1]=(-0.420934154847)*x[0]**o+(0.532576137523)*x[0]+(0.65082076651)*x[1]**o+(-0.264312562332)*x[1]
        s[1,2]=(-0.133788641889)*x[0]**o+(0.556829684305)*x[0]+(0.134872826121)*x[1]**o+(0.716478729883)*x[1]
        s[1,3]=(-0.345881339901)*x[0]**o+(-0.319261340187)*x[0]+(-0.114438699999)*x[1]**o+(0.0825491045988)*x[1]
        s[1,4]=(0.630228774591)*x[0]**o+(-0.858970963301)*x[0]+(0.708103720535)*x[1]**o+(-0.542940693024)*x[1]
        s[2,0]=(-0.683466738531)*x[0]**o+(-0.145169931761)*x[0]+(0.449697093379)*x[1]**o+(-0.907773514556)*x[1]
        s[2,1]=(-0.0705140461777)*x[0]**o+(0.374565128403)*x[0]+(-0.219001949605)*x[1]**o+(-0.239599954648)*x[1]
        s[2,2]=(-0.435803146278)*x[0]**o+(-0.150900699809)*x[0]+(-0.028936013795)*x[1]**o+(-0.169287689744)*x[1]
        s[2,3]=(-0.976991446034)*x[0]**o+(-0.975014510986)*x[0]+(0.314904815148)*x[1]**o+(0.479090592274)*x[1]
        s[2,4]=(-0.955674787191)*x[0]**o+(0.633046166846)*x[0]+(0.680202620704)*x[1]**o+(0.0533385507777)*x[1]
        s[3,0]=(0.876058021021)*x[0]**o+(0.294827067859)*x[0]+(0.482052318125)*x[1]**o+(-0.365719292507)*x[1]
        s[3,1]=(0.530559404731)*x[0]**o+(-0.313386503724)*x[0]+(-0.371352352152)*x[1]**o+(-0.682915252526)*x[1]
        s[3,2]=(-0.0522489801875)*x[0]**o+(0.631979705829)*x[0]+(-0.989995736543)*x[1]**o+(-0.386092008201)*x[1]
        s[3,3]=(-0.360737462637)*x[0]**o+(-0.0286143460674)*x[0]+(-0.743283421495)*x[1]**o+(0.346219067885)*x[1]
        s[3,4]=(-0.880285923018)*x[0]**o+(0.492777568952)*x[0]+(0.572215124514)*x[1]**o+(0.201965538371)*x[1]
        ref[0,0]=(-0.163301494275)*x_ref[0]**o+(-0.571871996499)*x_ref[0]+(-0.126318698091)*x_ref[1]**o+(0.594323280745)*x_ref[1]
        ref[0,1]=(-0.722891165327)*x_ref[0]**o+(0.539458396201)*x_ref[0]+(-0.338410805174)*x_ref[1]**o+(-0.0441039302518)*x_ref[1]
        ref[0,2]=(0.987165940018)*x_ref[0]**o+(0.277540309671)*x_ref[0]+(-0.194843676722)*x_ref[1]**o+(-0.838659115747)*x_ref[1]
        ref[0,3]=(0.668488628751)*x_ref[0]**o+(0.848828025004)*x_ref[0]+(-0.749811369279)*x_ref[1]**o+(-0.857737537329)*x_ref[1]
        ref[0,4]=(0.292839673829)*x_ref[0]**o+(-0.196773033161)*x_ref[0]+(-0.365394929162)*x_ref[1]**o+(-0.304841309615)*x_ref[1]
        ref[1,0]=(-0.244415346064)*x_ref[0]**o+(0.0653469288689)*x_ref[0]+(-0.405830560862)*x_ref[1]**o+(0.958797302829)*x_ref[1]
        ref[1,1]=(-0.420934154847)*x_ref[0]**o+(0.532576137523)*x_ref[0]+(0.65082076651)*x_ref[1]**o+(-0.264312562332)*x_ref[1]
        ref[1,2]=(-0.133788641889)*x_ref[0]**o+(0.556829684305)*x_ref[0]+(0.134872826121)*x_ref[1]**o+(0.716478729883)*x_ref[1]
        ref[1,3]=(-0.345881339901)*x_ref[0]**o+(-0.319261340187)*x_ref[0]+(-0.114438699999)*x_ref[1]**o+(0.0825491045988)*x_ref[1]
        ref[1,4]=(0.630228774591)*x_ref[0]**o+(-0.858970963301)*x_ref[0]+(0.708103720535)*x_ref[1]**o+(-0.542940693024)*x_ref[1]
        ref[2,0]=(-0.683466738531)*x_ref[0]**o+(-0.145169931761)*x_ref[0]+(0.449697093379)*x_ref[1]**o+(-0.907773514556)*x_ref[1]
        ref[2,1]=(-0.0705140461777)*x_ref[0]**o+(0.374565128403)*x_ref[0]+(-0.219001949605)*x_ref[1]**o+(-0.239599954648)*x_ref[1]
        ref[2,2]=(-0.435803146278)*x_ref[0]**o+(-0.150900699809)*x_ref[0]+(-0.028936013795)*x_ref[1]**o+(-0.169287689744)*x_ref[1]
        ref[2,3]=(-0.976991446034)*x_ref[0]**o+(-0.975014510986)*x_ref[0]+(0.314904815148)*x_ref[1]**o+(0.479090592274)*x_ref[1]
        ref[2,4]=(-0.955674787191)*x_ref[0]**o+(0.633046166846)*x_ref[0]+(0.680202620704)*x_ref[1]**o+(0.0533385507777)*x_ref[1]
        ref[3,0]=(0.876058021021)*x_ref[0]**o+(0.294827067859)*x_ref[0]+(0.482052318125)*x_ref[1]**o+(-0.365719292507)*x_ref[1]
        ref[3,1]=(0.530559404731)*x_ref[0]**o+(-0.313386503724)*x_ref[0]+(-0.371352352152)*x_ref[1]**o+(-0.682915252526)*x_ref[1]
        ref[3,2]=(-0.0522489801875)*x_ref[0]**o+(0.631979705829)*x_ref[0]+(-0.989995736543)*x_ref[1]**o+(-0.386092008201)*x_ref[1]
        ref[3,3]=(-0.360737462637)*x_ref[0]**o+(-0.0286143460674)*x_ref[0]+(-0.743283421495)*x_ref[1]**o+(0.346219067885)*x_ref[1]
        ref[3,4]=(-0.880285923018)*x_ref[0]**o+(0.492777568952)*x_ref[0]+(0.572215124514)*x_ref[1]**o+(0.201965538371)*x_ref[1]
      else:
        s[0,0]=(0.459801313828)*x[0]**o+(0.874491492894)*x[0]+(-0.676707065181)*x[1]**o+(-0.106095988336)*x[1]+(-0.171044270123)*x[2]**o+(0.0638260656588)*x[2]
        s[0,1]=(0.596371056873)*x[0]**o+(0.316401622905)*x[0]+(0.238994026229)*x[1]**o+(-0.462592805759)*x[1]+(-0.237565256761)*x[2]**o+(-0.100115327497)*x[2]
        s[0,2]=(-0.423227643937)*x[0]**o+(0.706315080194)*x[0]+(-0.43521651208)*x[1]**o+(-0.0219505777477)*x[1]+(-0.248035954152)*x[2]**o+(-0.0426762513079)*x[2]
        s[0,3]=(-0.807609617307)*x[0]**o+(0.870379040362)*x[0]+(0.0450254577238)*x[1]**o+(0.409422364523)*x[1]+(-0.0747425184319)*x[2]**o+(0.714168287946)*x[2]
        s[0,4]=(0.229697696099)*x[0]**o+(-0.193953210678)*x[0]+(0.821756203411)*x[1]**o+(-0.686241275184)*x[1]+(-0.461813580643)*x[2]**o+(-0.111679064974)*x[2]
        s[1,0]=(0.09199980955)*x[0]**o+(-0.596786310748)*x[0]+(-0.519482393268)*x[1]**o+(0.494226697149)*x[1]+(-0.533138224953)*x[2]**o+(-0.442965941122)*x[2]
        s[1,1]=(-0.499699311698)*x[0]**o+(-0.300118366174)*x[0]+(-0.0622786566862)*x[1]**o+(0.389573411943)*x[1]+(0.812614223101)*x[2]**o+(0.0241052242281)*x[2]
        s[1,2]=(-0.937872631185)*x[0]**o+(0.693932819382)*x[0]+(-0.940606874805)*x[1]**o+(0.0565548603525)*x[1]+(0.488404217301)*x[2]**o+(-0.54908344074)*x[2]
        s[1,3]=(0.187440914227)*x[0]**o+(0.513746783521)*x[0]+(0.975266666494)*x[1]**o+(-0.298164635346)*x[1]+(0.796028384674)*x[2]**o+(-0.555474012189)*x[2]
        s[1,4]=(-0.542284827327)*x[0]**o+(0.72336133394)*x[0]+(-0.382404238652)*x[1]**o+(-0.966114351457)*x[1]+(-0.0713526699547)*x[2]**o+(-0.969056495871)*x[2]
        s[2,0]=(-0.669587562542)*x[0]**o+(0.927249904009)*x[0]+(0.0901118246496)*x[1]**o+(0.0105771498119)*x[1]+(0.0347578359214)*x[2]**o+(0.306378930574)*x[2]
        s[2,1]=(0.552110347651)*x[0]**o+(0.534155224195)*x[0]+(-0.94964594078)*x[1]**o+(-0.867398879865)*x[1]+(0.529146894499)*x[2]**o+(-0.478336329324)*x[2]
        s[2,2]=(0.0250007510203)*x[0]**o+(-0.249616427955)*x[0]+(0.606560318424)*x[1]**o+(-0.921833180264)*x[1]+(-0.15817883795)*x[2]**o+(0.430892612477)*x[2]
        s[2,3]=(0.186816270003)*x[0]**o+(0.408732439837)*x[0]+(-0.89708625066)*x[1]**o+(0.0672189023842)*x[1]+(-0.502969902157)*x[2]**o+(0.0172554167453)*x[2]
        s[2,4]=(0.458126497715)*x[0]**o+(-0.929705410355)*x[0]+(0.493346469334)*x[1]**o+(-0.432232217413)*x[1]+(-0.579275855459)*x[2]**o+(0.756337283238)*x[2]
        s[3,0]=(0.933176387441)*x[0]**o+(0.24114580865)*x[0]+(-0.830402155007)*x[1]**o+(0.925370556667)*x[1]+(0.242042958988)*x[2]**o+(0.341138866916)*x[2]
        s[3,1]=(0.0660813821046)*x[0]**o+(-0.492921246892)*x[0]+(0.728036967007)*x[1]**o+(0.241487680125)*x[1]+(-0.705909691396)*x[2]**o+(0.812820536373)*x[2]
        s[3,2]=(-0.71644646201)*x[0]**o+(-0.428574354372)*x[0]+(0.928810019254)*x[1]**o+(0.0142522119308)*x[1]+(0.641710384867)*x[2]**o+(-0.512508294632)*x[2]
        s[3,3]=(-0.185856758974)*x[0]**o+(-0.265588242215)*x[0]+(-0.0750623283309)*x[1]**o+(-0.166114006253)*x[1]+(-0.986359701771)*x[2]**o+(0.237860767478)*x[2]
        s[3,4]=(0.0423246107159)*x[0]**o+(0.216782521414)*x[0]+(-0.536875736799)*x[1]**o+(0.0489914911388)*x[1]+(0.759222862013)*x[2]**o+(0.743454532517)*x[2]
        ref[0,0]=(0.459801313828)*x_ref[0]**o+(0.874491492894)*x_ref[0]+(-0.676707065181)*x_ref[1]**o+(-0.106095988336)*x_ref[1]+(-0.171044270123)*x_ref[2]**o+(0.0638260656588)*x_ref[2]
        ref[0,1]=(0.596371056873)*x_ref[0]**o+(0.316401622905)*x_ref[0]+(0.238994026229)*x_ref[1]**o+(-0.462592805759)*x_ref[1]+(-0.237565256761)*x_ref[2]**o+(-0.100115327497)*x_ref[2]
        ref[0,2]=(-0.423227643937)*x_ref[0]**o+(0.706315080194)*x_ref[0]+(-0.43521651208)*x_ref[1]**o+(-0.0219505777477)*x_ref[1]+(-0.248035954152)*x_ref[2]**o+(-0.0426762513079)*x_ref[2]
        ref[0,3]=(-0.807609617307)*x_ref[0]**o+(0.870379040362)*x_ref[0]+(0.0450254577238)*x_ref[1]**o+(0.409422364523)*x_ref[1]+(-0.0747425184319)*x_ref[2]**o+(0.714168287946)*x_ref[2]
        ref[0,4]=(0.229697696099)*x_ref[0]**o+(-0.193953210678)*x_ref[0]+(0.821756203411)*x_ref[1]**o+(-0.686241275184)*x_ref[1]+(-0.461813580643)*x_ref[2]**o+(-0.111679064974)*x_ref[2]
        ref[1,0]=(0.09199980955)*x_ref[0]**o+(-0.596786310748)*x_ref[0]+(-0.519482393268)*x_ref[1]**o+(0.494226697149)*x_ref[1]+(-0.533138224953)*x_ref[2]**o+(-0.442965941122)*x_ref[2]
        ref[1,1]=(-0.499699311698)*x_ref[0]**o+(-0.300118366174)*x_ref[0]+(-0.0622786566862)*x_ref[1]**o+(0.389573411943)*x_ref[1]+(0.812614223101)*x_ref[2]**o+(0.0241052242281)*x_ref[2]
        ref[1,2]=(-0.937872631185)*x_ref[0]**o+(0.693932819382)*x_ref[0]+(-0.940606874805)*x_ref[1]**o+(0.0565548603525)*x_ref[1]+(0.488404217301)*x_ref[2]**o+(-0.54908344074)*x_ref[2]
        ref[1,3]=(0.187440914227)*x_ref[0]**o+(0.513746783521)*x_ref[0]+(0.975266666494)*x_ref[1]**o+(-0.298164635346)*x_ref[1]+(0.796028384674)*x_ref[2]**o+(-0.555474012189)*x_ref[2]
        ref[1,4]=(-0.542284827327)*x_ref[0]**o+(0.72336133394)*x_ref[0]+(-0.382404238652)*x_ref[1]**o+(-0.966114351457)*x_ref[1]+(-0.0713526699547)*x_ref[2]**o+(-0.969056495871)*x_ref[2]
        ref[2,0]=(-0.669587562542)*x_ref[0]**o+(0.927249904009)*x_ref[0]+(0.0901118246496)*x_ref[1]**o+(0.0105771498119)*x_ref[1]+(0.0347578359214)*x_ref[2]**o+(0.306378930574)*x_ref[2]
        ref[2,1]=(0.552110347651)*x_ref[0]**o+(0.534155224195)*x_ref[0]+(-0.94964594078)*x_ref[1]**o+(-0.867398879865)*x_ref[1]+(0.529146894499)*x_ref[2]**o+(-0.478336329324)*x_ref[2]
        ref[2,2]=(0.0250007510203)*x_ref[0]**o+(-0.249616427955)*x_ref[0]+(0.606560318424)*x_ref[1]**o+(-0.921833180264)*x_ref[1]+(-0.15817883795)*x_ref[2]**o+(0.430892612477)*x_ref[2]
        ref[2,3]=(0.186816270003)*x_ref[0]**o+(0.408732439837)*x_ref[0]+(-0.89708625066)*x_ref[1]**o+(0.0672189023842)*x_ref[1]+(-0.502969902157)*x_ref[2]**o+(0.0172554167453)*x_ref[2]
        ref[2,4]=(0.458126497715)*x_ref[0]**o+(-0.929705410355)*x_ref[0]+(0.493346469334)*x_ref[1]**o+(-0.432232217413)*x_ref[1]+(-0.579275855459)*x_ref[2]**o+(0.756337283238)*x_ref[2]
        ref[3,0]=(0.933176387441)*x_ref[0]**o+(0.24114580865)*x_ref[0]+(-0.830402155007)*x_ref[1]**o+(0.925370556667)*x_ref[1]+(0.242042958988)*x_ref[2]**o+(0.341138866916)*x_ref[2]
        ref[3,1]=(0.0660813821046)*x_ref[0]**o+(-0.492921246892)*x_ref[0]+(0.728036967007)*x_ref[1]**o+(0.241487680125)*x_ref[1]+(-0.705909691396)*x_ref[2]**o+(0.812820536373)*x_ref[2]
        ref[3,2]=(-0.71644646201)*x_ref[0]**o+(-0.428574354372)*x_ref[0]+(0.928810019254)*x_ref[1]**o+(0.0142522119308)*x_ref[1]+(0.641710384867)*x_ref[2]**o+(-0.512508294632)*x_ref[2]
        ref[3,3]=(-0.185856758974)*x_ref[0]**o+(-0.265588242215)*x_ref[0]+(-0.0750623283309)*x_ref[1]**o+(-0.166114006253)*x_ref[1]+(-0.986359701771)*x_ref[2]**o+(0.237860767478)*x_ref[2]
        ref[3,4]=(0.0423246107159)*x_ref[0]**o+(0.216782521414)*x_ref[0]+(-0.536875736799)*x_ref[1]**o+(0.0489914911388)*x_ref[1]+(0.759222862013)*x_ref[2]**o+(0.743454532517)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.598481759913)*x[0]**o+(-0.272925610474)*x[0]+(0.551460738992)*x[1]**o+(-0.609542685963)*x[1]
        s[0,0,1]=(0.645588310869)*x[0]**o+(0.983661063707)*x[0]+(-0.585913047558)*x[1]**o+(-0.700447569863)*x[1]
        s[0,1,0]=(0.0735002177703)*x[0]**o+(0.640392078189)*x[0]+(-0.640531878864)*x[1]**o+(0.927957489213)*x[1]
        s[0,1,1]=(-0.91764393635)*x[0]**o+(-0.234646513068)*x[0]+(0.434778070289)*x[1]**o+(0.676898688957)*x[1]
        s[1,0,0]=(0.900485566233)*x[0]**o+(0.17381724123)*x[0]+(-0.0472757726724)*x[1]**o+(-0.603802455434)*x[1]
        s[1,0,1]=(0.0111284144129)*x[0]**o+(0.729344462439)*x[0]+(0.650641541889)*x[1]**o+(-0.382226116166)*x[1]
        s[1,1,0]=(0.866905202972)*x[0]**o+(0.586112114242)*x[0]+(0.132836683118)*x[1]**o+(0.685439975799)*x[1]
        s[1,1,1]=(0.0893846587203)*x[0]**o+(0.475293339259)*x[0]+(-0.524027603353)*x[1]**o+(-0.427042646418)*x[1]
        s[2,0,0]=(0.696717009328)*x[0]**o+(0.17836429668)*x[0]+(0.339729683456)*x[1]**o+(0.425401166923)*x[1]
        s[2,0,1]=(-0.578374482181)*x[0]**o+(-0.06883462626)*x[0]+(-0.62708396204)*x[1]**o+(-0.732305455774)*x[1]
        s[2,1,0]=(-0.170455966767)*x[0]**o+(-0.19206482096)*x[0]+(-0.291324644463)*x[1]**o+(0.0296539206517)*x[1]
        s[2,1,1]=(0.209978578865)*x[0]**o+(-0.026213873389)*x[0]+(-0.248910233717)*x[1]**o+(-0.855378045214)*x[1]
        s[3,0,0]=(-0.926098922919)*x[0]**o+(-0.12977446027)*x[0]+(-0.706408024307)*x[1]**o+(0.00802672230618)*x[1]
        s[3,0,1]=(-0.677659537077)*x[0]**o+(0.319391836395)*x[0]+(0.876973869047)*x[1]**o+(0.569113379857)*x[1]
        s[3,1,0]=(-0.486504252182)*x[0]**o+(-0.0946134881675)*x[0]+(0.490827332525)*x[1]**o+(-0.392910432859)*x[1]
        s[3,1,1]=(0.784833517776)*x[0]**o+(-0.576527986841)*x[0]+(0.439029003252)*x[1]**o+(-0.307033754704)*x[1]
        s[4,0,0]=(0.290446522592)*x[0]**o+(0.441136393638)*x[0]+(-0.372607076196)*x[1]**o+(-0.159674601004)*x[1]
        s[4,0,1]=(0.440515634069)*x[0]**o+(-0.42368755307)*x[0]+(0.370018085845)*x[1]**o+(0.0235740884719)*x[1]
        s[4,1,0]=(-0.697567283199)*x[0]**o+(0.327933485486)*x[0]+(-0.36825773126)*x[1]**o+(0.016215691731)*x[1]
        s[4,1,1]=(0.572572720028)*x[0]**o+(0.0840839889325)*x[0]+(0.826717033934)*x[1]**o+(-0.0474809600044)*x[1]
        s[5,0,0]=(-0.16977415879)*x[0]**o+(-0.448717259264)*x[0]+(-0.595995298169)*x[1]**o+(-0.331364898672)*x[1]
        s[5,0,1]=(-0.0068969930042)*x[0]**o+(-0.599597651559)*x[0]+(0.912012252833)*x[1]**o+(-0.0802048593495)*x[1]
        s[5,1,0]=(-0.897450521147)*x[0]**o+(-0.126994660738)*x[0]+(0.939958074736)*x[1]**o+(-0.600360767914)*x[1]
        s[5,1,1]=(-0.733313832271)*x[0]**o+(-0.53733764856)*x[0]+(-0.770590244589)*x[1]**o+(0.62787073827)*x[1]
        ref[0,0,0]=(-0.598481759913)*x_ref[0]**o+(-0.272925610474)*x_ref[0]+(0.551460738992)*x_ref[1]**o+(-0.609542685963)*x_ref[1]
        ref[0,0,1]=(0.645588310869)*x_ref[0]**o+(0.983661063707)*x_ref[0]+(-0.585913047558)*x_ref[1]**o+(-0.700447569863)*x_ref[1]
        ref[0,1,0]=(0.0735002177703)*x_ref[0]**o+(0.640392078189)*x_ref[0]+(-0.640531878864)*x_ref[1]**o+(0.927957489213)*x_ref[1]
        ref[0,1,1]=(-0.91764393635)*x_ref[0]**o+(-0.234646513068)*x_ref[0]+(0.434778070289)*x_ref[1]**o+(0.676898688957)*x_ref[1]
        ref[1,0,0]=(0.900485566233)*x_ref[0]**o+(0.17381724123)*x_ref[0]+(-0.0472757726724)*x_ref[1]**o+(-0.603802455434)*x_ref[1]
        ref[1,0,1]=(0.0111284144129)*x_ref[0]**o+(0.729344462439)*x_ref[0]+(0.650641541889)*x_ref[1]**o+(-0.382226116166)*x_ref[1]
        ref[1,1,0]=(0.866905202972)*x_ref[0]**o+(0.586112114242)*x_ref[0]+(0.132836683118)*x_ref[1]**o+(0.685439975799)*x_ref[1]
        ref[1,1,1]=(0.0893846587203)*x_ref[0]**o+(0.475293339259)*x_ref[0]+(-0.524027603353)*x_ref[1]**o+(-0.427042646418)*x_ref[1]
        ref[2,0,0]=(0.696717009328)*x_ref[0]**o+(0.17836429668)*x_ref[0]+(0.339729683456)*x_ref[1]**o+(0.425401166923)*x_ref[1]
        ref[2,0,1]=(-0.578374482181)*x_ref[0]**o+(-0.06883462626)*x_ref[0]+(-0.62708396204)*x_ref[1]**o+(-0.732305455774)*x_ref[1]
        ref[2,1,0]=(-0.170455966767)*x_ref[0]**o+(-0.19206482096)*x_ref[0]+(-0.291324644463)*x_ref[1]**o+(0.0296539206517)*x_ref[1]
        ref[2,1,1]=(0.209978578865)*x_ref[0]**o+(-0.026213873389)*x_ref[0]+(-0.248910233717)*x_ref[1]**o+(-0.855378045214)*x_ref[1]
        ref[3,0,0]=(-0.926098922919)*x_ref[0]**o+(-0.12977446027)*x_ref[0]+(-0.706408024307)*x_ref[1]**o+(0.00802672230618)*x_ref[1]
        ref[3,0,1]=(-0.677659537077)*x_ref[0]**o+(0.319391836395)*x_ref[0]+(0.876973869047)*x_ref[1]**o+(0.569113379857)*x_ref[1]
        ref[3,1,0]=(-0.486504252182)*x_ref[0]**o+(-0.0946134881675)*x_ref[0]+(0.490827332525)*x_ref[1]**o+(-0.392910432859)*x_ref[1]
        ref[3,1,1]=(0.784833517776)*x_ref[0]**o+(-0.576527986841)*x_ref[0]+(0.439029003252)*x_ref[1]**o+(-0.307033754704)*x_ref[1]
        ref[4,0,0]=(0.290446522592)*x_ref[0]**o+(0.441136393638)*x_ref[0]+(-0.372607076196)*x_ref[1]**o+(-0.159674601004)*x_ref[1]
        ref[4,0,1]=(0.440515634069)*x_ref[0]**o+(-0.42368755307)*x_ref[0]+(0.370018085845)*x_ref[1]**o+(0.0235740884719)*x_ref[1]
        ref[4,1,0]=(-0.697567283199)*x_ref[0]**o+(0.327933485486)*x_ref[0]+(-0.36825773126)*x_ref[1]**o+(0.016215691731)*x_ref[1]
        ref[4,1,1]=(0.572572720028)*x_ref[0]**o+(0.0840839889325)*x_ref[0]+(0.826717033934)*x_ref[1]**o+(-0.0474809600044)*x_ref[1]
        ref[5,0,0]=(-0.16977415879)*x_ref[0]**o+(-0.448717259264)*x_ref[0]+(-0.595995298169)*x_ref[1]**o+(-0.331364898672)*x_ref[1]
        ref[5,0,1]=(-0.0068969930042)*x_ref[0]**o+(-0.599597651559)*x_ref[0]+(0.912012252833)*x_ref[1]**o+(-0.0802048593495)*x_ref[1]
        ref[5,1,0]=(-0.897450521147)*x_ref[0]**o+(-0.126994660738)*x_ref[0]+(0.939958074736)*x_ref[1]**o+(-0.600360767914)*x_ref[1]
        ref[5,1,1]=(-0.733313832271)*x_ref[0]**o+(-0.53733764856)*x_ref[0]+(-0.770590244589)*x_ref[1]**o+(0.62787073827)*x_ref[1]
      else:
        s[0,0,0]=(0.49027687523)*x[0]**o+(-0.945396297892)*x[0]+(-0.649119833947)*x[1]**o+(-0.193510024763)*x[1]+(0.672065654766)*x[2]**o+(0.396941813883)*x[2]
        s[0,0,1]=(-0.379608112905)*x[0]**o+(-0.321849772246)*x[0]+(-0.695093455872)*x[1]**o+(-0.114263310683)*x[1]+(0.425556105367)*x[2]**o+(0.737954023041)*x[2]
        s[0,1,0]=(0.597295505367)*x[0]**o+(0.285081105939)*x[0]+(-0.690746281653)*x[1]**o+(0.431456884332)*x[1]+(-0.569785421553)*x[2]**o+(0.218907517674)*x[2]
        s[0,1,1]=(-0.743660806659)*x[0]**o+(0.916636055834)*x[0]+(0.747852198685)*x[1]**o+(-0.304767976464)*x[1]+(0.737051048825)*x[2]**o+(0.108795699665)*x[2]
        s[1,0,0]=(0.0322299276093)*x[0]**o+(-0.507367087704)*x[0]+(-0.154613332688)*x[1]**o+(-0.642692955185)*x[1]+(0.921618425322)*x[2]**o+(-0.89234021469)*x[2]
        s[1,0,1]=(0.250927374346)*x[0]**o+(-0.713045403997)*x[0]+(0.0229252996751)*x[1]**o+(0.251843016211)*x[1]+(0.683324421319)*x[2]**o+(0.466509616753)*x[2]
        s[1,1,0]=(0.610653309325)*x[0]**o+(0.503568507563)*x[0]+(0.519292621274)*x[1]**o+(-0.452962008782)*x[1]+(0.908988844098)*x[2]**o+(-0.933406144464)*x[2]
        s[1,1,1]=(0.271788487178)*x[0]**o+(0.341989918651)*x[0]+(-0.406971961331)*x[1]**o+(-0.928647769169)*x[1]+(0.662902796818)*x[2]**o+(0.655876724787)*x[2]
        s[2,0,0]=(-0.775353204706)*x[0]**o+(0.278726576715)*x[0]+(0.448754566832)*x[1]**o+(-0.943150229959)*x[1]+(0.0806002558903)*x[2]**o+(-0.350491928882)*x[2]
        s[2,0,1]=(-0.899719123904)*x[0]**o+(-0.89394097319)*x[0]+(0.136672330017)*x[1]**o+(0.941198317525)*x[1]+(0.945970508521)*x[2]**o+(0.232574429377)*x[2]
        s[2,1,0]=(0.102227090178)*x[0]**o+(0.498258334271)*x[0]+(-0.369219100853)*x[1]**o+(0.965612831948)*x[1]+(-0.580434278342)*x[2]**o+(0.0747696579664)*x[2]
        s[2,1,1]=(0.0680171783989)*x[0]**o+(-0.655506093893)*x[0]+(0.0854275560955)*x[1]**o+(0.788063374419)*x[1]+(0.3912557517)*x[2]**o+(-0.911795675902)*x[2]
        s[3,0,0]=(0.941143798566)*x[0]**o+(0.324489623574)*x[0]+(-0.259012903836)*x[1]**o+(-0.779051329011)*x[1]+(0.673245283112)*x[2]**o+(-0.888729936858)*x[2]
        s[3,0,1]=(0.904124513313)*x[0]**o+(0.564171149844)*x[0]+(0.799169083649)*x[1]**o+(-0.172322806388)*x[1]+(-0.0754065342235)*x[2]**o+(0.0419393221333)*x[2]
        s[3,1,0]=(-0.200531316713)*x[0]**o+(0.786777777323)*x[0]+(-0.222824591798)*x[1]**o+(0.32280872616)*x[1]+(-0.418791963436)*x[2]**o+(-0.549019641386)*x[2]
        s[3,1,1]=(0.0519992199857)*x[0]**o+(0.278310751677)*x[0]+(0.590847528061)*x[1]**o+(0.258564224896)*x[1]+(-0.500091685923)*x[2]**o+(0.511781801585)*x[2]
        s[4,0,0]=(0.459040882538)*x[0]**o+(-0.481335872602)*x[0]+(0.0556391566559)*x[1]**o+(-0.208836885623)*x[1]+(-0.11717985744)*x[2]**o+(-0.544122946452)*x[2]
        s[4,0,1]=(-0.255258906635)*x[0]**o+(0.860765893836)*x[0]+(0.389276109422)*x[1]**o+(-0.999502217953)*x[1]+(-0.715269648526)*x[2]**o+(0.451562561245)*x[2]
        s[4,1,0]=(-0.579717249622)*x[0]**o+(0.984974487836)*x[0]+(-0.803917364532)*x[1]**o+(-0.764878964835)*x[1]+(-0.0642799485519)*x[2]**o+(-0.917022175106)*x[2]
        s[4,1,1]=(-0.873863405322)*x[0]**o+(-0.872375014129)*x[0]+(0.981355044198)*x[1]**o+(0.40407870155)*x[1]+(0.796125441505)*x[2]**o+(0.486002048777)*x[2]
        s[5,0,0]=(0.92716952184)*x[0]**o+(-0.280973691624)*x[0]+(0.815340507346)*x[1]**o+(-0.555904571967)*x[1]+(0.765200732877)*x[2]**o+(-0.48251397176)*x[2]
        s[5,0,1]=(-0.6976034156)*x[0]**o+(-0.342128490335)*x[0]+(-0.273657764836)*x[1]**o+(-0.393975979448)*x[1]+(0.522428789629)*x[2]**o+(-0.210033793616)*x[2]
        s[5,1,0]=(-0.64289167402)*x[0]**o+(-0.772752345352)*x[0]+(0.900114631593)*x[1]**o+(0.510514117797)*x[1]+(-0.984639269239)*x[2]**o+(0.243489379424)*x[2]
        s[5,1,1]=(-0.551740975948)*x[0]**o+(-0.645067068599)*x[0]+(-0.238391882646)*x[1]**o+(-0.959940660719)*x[1]+(-0.020674238581)*x[2]**o+(0.0667084698715)*x[2]
        ref[0,0,0]=(0.49027687523)*x_ref[0]**o+(-0.945396297892)*x_ref[0]+(-0.649119833947)*x_ref[1]**o+(-0.193510024763)*x_ref[1]+(0.672065654766)*x_ref[2]**o+(0.396941813883)*x_ref[2]
        ref[0,0,1]=(-0.379608112905)*x_ref[0]**o+(-0.321849772246)*x_ref[0]+(-0.695093455872)*x_ref[1]**o+(-0.114263310683)*x_ref[1]+(0.425556105367)*x_ref[2]**o+(0.737954023041)*x_ref[2]
        ref[0,1,0]=(0.597295505367)*x_ref[0]**o+(0.285081105939)*x_ref[0]+(-0.690746281653)*x_ref[1]**o+(0.431456884332)*x_ref[1]+(-0.569785421553)*x_ref[2]**o+(0.218907517674)*x_ref[2]
        ref[0,1,1]=(-0.743660806659)*x_ref[0]**o+(0.916636055834)*x_ref[0]+(0.747852198685)*x_ref[1]**o+(-0.304767976464)*x_ref[1]+(0.737051048825)*x_ref[2]**o+(0.108795699665)*x_ref[2]
        ref[1,0,0]=(0.0322299276093)*x_ref[0]**o+(-0.507367087704)*x_ref[0]+(-0.154613332688)*x_ref[1]**o+(-0.642692955185)*x_ref[1]+(0.921618425322)*x_ref[2]**o+(-0.89234021469)*x_ref[2]
        ref[1,0,1]=(0.250927374346)*x_ref[0]**o+(-0.713045403997)*x_ref[0]+(0.0229252996751)*x_ref[1]**o+(0.251843016211)*x_ref[1]+(0.683324421319)*x_ref[2]**o+(0.466509616753)*x_ref[2]
        ref[1,1,0]=(0.610653309325)*x_ref[0]**o+(0.503568507563)*x_ref[0]+(0.519292621274)*x_ref[1]**o+(-0.452962008782)*x_ref[1]+(0.908988844098)*x_ref[2]**o+(-0.933406144464)*x_ref[2]
        ref[1,1,1]=(0.271788487178)*x_ref[0]**o+(0.341989918651)*x_ref[0]+(-0.406971961331)*x_ref[1]**o+(-0.928647769169)*x_ref[1]+(0.662902796818)*x_ref[2]**o+(0.655876724787)*x_ref[2]
        ref[2,0,0]=(-0.775353204706)*x_ref[0]**o+(0.278726576715)*x_ref[0]+(0.448754566832)*x_ref[1]**o+(-0.943150229959)*x_ref[1]+(0.0806002558903)*x_ref[2]**o+(-0.350491928882)*x_ref[2]
        ref[2,0,1]=(-0.899719123904)*x_ref[0]**o+(-0.89394097319)*x_ref[0]+(0.136672330017)*x_ref[1]**o+(0.941198317525)*x_ref[1]+(0.945970508521)*x_ref[2]**o+(0.232574429377)*x_ref[2]
        ref[2,1,0]=(0.102227090178)*x_ref[0]**o+(0.498258334271)*x_ref[0]+(-0.369219100853)*x_ref[1]**o+(0.965612831948)*x_ref[1]+(-0.580434278342)*x_ref[2]**o+(0.0747696579664)*x_ref[2]
        ref[2,1,1]=(0.0680171783989)*x_ref[0]**o+(-0.655506093893)*x_ref[0]+(0.0854275560955)*x_ref[1]**o+(0.788063374419)*x_ref[1]+(0.3912557517)*x_ref[2]**o+(-0.911795675902)*x_ref[2]
        ref[3,0,0]=(0.941143798566)*x_ref[0]**o+(0.324489623574)*x_ref[0]+(-0.259012903836)*x_ref[1]**o+(-0.779051329011)*x_ref[1]+(0.673245283112)*x_ref[2]**o+(-0.888729936858)*x_ref[2]
        ref[3,0,1]=(0.904124513313)*x_ref[0]**o+(0.564171149844)*x_ref[0]+(0.799169083649)*x_ref[1]**o+(-0.172322806388)*x_ref[1]+(-0.0754065342235)*x_ref[2]**o+(0.0419393221333)*x_ref[2]
        ref[3,1,0]=(-0.200531316713)*x_ref[0]**o+(0.786777777323)*x_ref[0]+(-0.222824591798)*x_ref[1]**o+(0.32280872616)*x_ref[1]+(-0.418791963436)*x_ref[2]**o+(-0.549019641386)*x_ref[2]
        ref[3,1,1]=(0.0519992199857)*x_ref[0]**o+(0.278310751677)*x_ref[0]+(0.590847528061)*x_ref[1]**o+(0.258564224896)*x_ref[1]+(-0.500091685923)*x_ref[2]**o+(0.511781801585)*x_ref[2]
        ref[4,0,0]=(0.459040882538)*x_ref[0]**o+(-0.481335872602)*x_ref[0]+(0.0556391566559)*x_ref[1]**o+(-0.208836885623)*x_ref[1]+(-0.11717985744)*x_ref[2]**o+(-0.544122946452)*x_ref[2]
        ref[4,0,1]=(-0.255258906635)*x_ref[0]**o+(0.860765893836)*x_ref[0]+(0.389276109422)*x_ref[1]**o+(-0.999502217953)*x_ref[1]+(-0.715269648526)*x_ref[2]**o+(0.451562561245)*x_ref[2]
        ref[4,1,0]=(-0.579717249622)*x_ref[0]**o+(0.984974487836)*x_ref[0]+(-0.803917364532)*x_ref[1]**o+(-0.764878964835)*x_ref[1]+(-0.0642799485519)*x_ref[2]**o+(-0.917022175106)*x_ref[2]
        ref[4,1,1]=(-0.873863405322)*x_ref[0]**o+(-0.872375014129)*x_ref[0]+(0.981355044198)*x_ref[1]**o+(0.40407870155)*x_ref[1]+(0.796125441505)*x_ref[2]**o+(0.486002048777)*x_ref[2]
        ref[5,0,0]=(0.92716952184)*x_ref[0]**o+(-0.280973691624)*x_ref[0]+(0.815340507346)*x_ref[1]**o+(-0.555904571967)*x_ref[1]+(0.765200732877)*x_ref[2]**o+(-0.48251397176)*x_ref[2]
        ref[5,0,1]=(-0.6976034156)*x_ref[0]**o+(-0.342128490335)*x_ref[0]+(-0.273657764836)*x_ref[1]**o+(-0.393975979448)*x_ref[1]+(0.522428789629)*x_ref[2]**o+(-0.210033793616)*x_ref[2]
        ref[5,1,0]=(-0.64289167402)*x_ref[0]**o+(-0.772752345352)*x_ref[0]+(0.900114631593)*x_ref[1]**o+(0.510514117797)*x_ref[1]+(-0.984639269239)*x_ref[2]**o+(0.243489379424)*x_ref[2]
        ref[5,1,1]=(-0.551740975948)*x_ref[0]**o+(-0.645067068599)*x_ref[0]+(-0.238391882646)*x_ref[1]**o+(-0.959940660719)*x_ref[1]+(-0.020674238581)*x_ref[2]**o+(0.0667084698715)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.711963461677)*x[0]**o+(0.542670298734)*x[0]+(0.28075382249)*x[1]**o+(0.667437625224)*x[1]
        s[0,0,0,1]=(0.573441578286)*x[0]**o+(0.506095821827)*x[0]+(0.150993391009)*x[1]**o+(0.526409669633)*x[1]
        s[0,0,1,0]=(0.706627280956)*x[0]**o+(0.57780902263)*x[0]+(0.270996212375)*x[1]**o+(0.938868396109)*x[1]
        s[0,0,1,1]=(0.970926483868)*x[0]**o+(0.0967183437714)*x[0]+(0.305310109546)*x[1]**o+(-0.746529372684)*x[1]
        s[0,0,2,0]=(-0.0598871068472)*x[0]**o+(0.441097065872)*x[0]+(0.240831657822)*x[1]**o+(-0.781319614159)*x[1]
        s[0,0,2,1]=(0.0738716023753)*x[0]**o+(-0.280473432627)*x[0]+(0.351688231305)*x[1]**o+(0.287190769396)*x[1]
        s[0,1,0,0]=(-0.662029332598)*x[0]**o+(0.769980048213)*x[0]+(-0.316622870493)*x[1]**o+(0.756839554902)*x[1]
        s[0,1,0,1]=(-0.931130133674)*x[0]**o+(-0.70789066479)*x[0]+(0.350649077237)*x[1]**o+(-0.340577841771)*x[1]
        s[0,1,1,0]=(0.541533987605)*x[0]**o+(-0.871063950931)*x[0]+(0.198020061012)*x[1]**o+(0.802998878081)*x[1]
        s[0,1,1,1]=(0.264037408365)*x[0]**o+(-0.0557314225594)*x[0]+(-0.684018523812)*x[1]**o+(-0.229282658383)*x[1]
        s[0,1,2,0]=(0.637810239569)*x[0]**o+(0.552431819202)*x[0]+(0.0939985703867)*x[1]**o+(0.828751113775)*x[1]
        s[0,1,2,1]=(-0.374671159094)*x[0]**o+(0.664088936153)*x[0]+(-0.345847704492)*x[1]**o+(-0.364075081521)*x[1]
        s[0,2,0,0]=(0.00895781289078)*x[0]**o+(0.399614286422)*x[0]+(0.341683450789)*x[1]**o+(0.167206336911)*x[1]
        s[0,2,0,1]=(-0.104604431902)*x[0]**o+(-0.659107995643)*x[0]+(0.680691727783)*x[1]**o+(-0.684349104067)*x[1]
        s[0,2,1,0]=(0.710804511082)*x[0]**o+(-0.972641516063)*x[0]+(0.330095093159)*x[1]**o+(0.937337443354)*x[1]
        s[0,2,1,1]=(-0.294227883996)*x[0]**o+(0.965392046246)*x[0]+(0.714517712165)*x[1]**o+(0.449886137172)*x[1]
        s[0,2,2,0]=(-0.825491552072)*x[0]**o+(-0.638472841016)*x[0]+(-0.191336005365)*x[1]**o+(-0.422472261204)*x[1]
        s[0,2,2,1]=(-0.23092704399)*x[0]**o+(-0.637860425921)*x[0]+(0.840387209559)*x[1]**o+(0.165561089405)*x[1]
        s[0,3,0,0]=(0.420745779316)*x[0]**o+(-0.311588614725)*x[0]+(-0.553702597164)*x[1]**o+(0.272368731382)*x[1]
        s[0,3,0,1]=(0.327243090728)*x[0]**o+(-0.386774129727)*x[0]+(-0.950289816609)*x[1]**o+(0.662365527837)*x[1]
        s[0,3,1,0]=(-0.889641002357)*x[0]**o+(-0.195916682305)*x[0]+(-0.60085377073)*x[1]**o+(-0.191887529995)*x[1]
        s[0,3,1,1]=(0.987376581443)*x[0]**o+(-0.0350561501803)*x[0]+(0.704905530249)*x[1]**o+(-0.735687404507)*x[1]
        s[0,3,2,0]=(0.420809308115)*x[0]**o+(-0.339666948082)*x[0]+(0.996296976276)*x[1]**o+(0.506174123674)*x[1]
        s[0,3,2,1]=(0.426714798574)*x[0]**o+(0.620612017549)*x[0]+(0.60846811755)*x[1]**o+(0.561758397959)*x[1]
        s[0,4,0,0]=(0.0193321433149)*x[0]**o+(0.395174248733)*x[0]+(-0.631710108856)*x[1]**o+(0.562496027755)*x[1]
        s[0,4,0,1]=(0.936397656738)*x[0]**o+(-0.620128461169)*x[0]+(0.424872968744)*x[1]**o+(0.936360345969)*x[1]
        s[0,4,1,0]=(-0.823655625312)*x[0]**o+(0.110021832964)*x[0]+(-0.480484367041)*x[1]**o+(0.613626788942)*x[1]
        s[0,4,1,1]=(0.516537102221)*x[0]**o+(-0.465693898711)*x[0]+(-0.0649481633029)*x[1]**o+(0.498343667101)*x[1]
        s[0,4,2,0]=(0.364997630891)*x[0]**o+(0.726894447639)*x[0]+(0.729067579856)*x[1]**o+(-0.786376612921)*x[1]
        s[0,4,2,1]=(-0.574801923675)*x[0]**o+(-0.769120062116)*x[0]+(0.0802603764785)*x[1]**o+(0.322665268053)*x[1]
        s[1,0,0,0]=(-0.497916263484)*x[0]**o+(0.0689477158621)*x[0]+(-0.59065778662)*x[1]**o+(0.620475916092)*x[1]
        s[1,0,0,1]=(0.928072689885)*x[0]**o+(0.550220523492)*x[0]+(0.384056461821)*x[1]**o+(0.991195851819)*x[1]
        s[1,0,1,0]=(-0.171652790551)*x[0]**o+(-0.310223770891)*x[0]+(0.505801949231)*x[1]**o+(-0.809872602671)*x[1]
        s[1,0,1,1]=(-0.843616998175)*x[0]**o+(-0.514683249854)*x[0]+(0.22719571088)*x[1]**o+(-0.801964932709)*x[1]
        s[1,0,2,0]=(-0.816576821632)*x[0]**o+(-0.629668603878)*x[0]+(-0.296562767682)*x[1]**o+(-0.0121908797722)*x[1]
        s[1,0,2,1]=(-0.816221470582)*x[0]**o+(0.101678348599)*x[0]+(0.0841692807823)*x[1]**o+(0.265517698128)*x[1]
        s[1,1,0,0]=(0.74252349936)*x[0]**o+(0.488296875375)*x[0]+(0.553091389727)*x[1]**o+(0.0951295346041)*x[1]
        s[1,1,0,1]=(0.573858954955)*x[0]**o+(-0.0210888969717)*x[0]+(-0.813464794029)*x[1]**o+(0.760502343882)*x[1]
        s[1,1,1,0]=(0.806002928884)*x[0]**o+(0.191962011356)*x[0]+(0.788308484735)*x[1]**o+(0.325349383493)*x[1]
        s[1,1,1,1]=(-0.975008178175)*x[0]**o+(-0.525299302456)*x[0]+(-0.718959301598)*x[1]**o+(0.821616294375)*x[1]
        s[1,1,2,0]=(-0.441476557684)*x[0]**o+(0.442725962693)*x[0]+(0.605338741812)*x[1]**o+(-0.0356460149592)*x[1]
        s[1,1,2,1]=(0.75296734605)*x[0]**o+(0.0218514182228)*x[0]+(-0.408708955555)*x[1]**o+(-0.729191065301)*x[1]
        s[1,2,0,0]=(0.24165794152)*x[0]**o+(-0.33924002011)*x[0]+(0.357553913734)*x[1]**o+(0.166312753501)*x[1]
        s[1,2,0,1]=(0.835534214783)*x[0]**o+(-0.688462973774)*x[0]+(0.779388430668)*x[1]**o+(-0.869108490882)*x[1]
        s[1,2,1,0]=(0.172345662475)*x[0]**o+(0.323742121947)*x[0]+(-0.296938383275)*x[1]**o+(0.0359645605525)*x[1]
        s[1,2,1,1]=(-0.596600537083)*x[0]**o+(0.644152951661)*x[0]+(0.609094045392)*x[1]**o+(-0.312234180863)*x[1]
        s[1,2,2,0]=(0.484589912205)*x[0]**o+(-0.906567695494)*x[0]+(0.811265520788)*x[1]**o+(0.444799696239)*x[1]
        s[1,2,2,1]=(0.0254720755485)*x[0]**o+(-0.41174472105)*x[0]+(0.90099646911)*x[1]**o+(-0.33088554417)*x[1]
        s[1,3,0,0]=(-0.281155224735)*x[0]**o+(0.274137552899)*x[0]+(0.0974242478848)*x[1]**o+(-0.995002288445)*x[1]
        s[1,3,0,1]=(0.907653805131)*x[0]**o+(0.743984237516)*x[0]+(-0.776028924136)*x[1]**o+(0.209784894077)*x[1]
        s[1,3,1,0]=(0.0470679202395)*x[0]**o+(-0.809937294063)*x[0]+(0.956184337312)*x[1]**o+(-0.539155228812)*x[1]
        s[1,3,1,1]=(0.720979862876)*x[0]**o+(0.412993117346)*x[0]+(0.259998599343)*x[1]**o+(-0.739191117502)*x[1]
        s[1,3,2,0]=(0.403493857435)*x[0]**o+(0.0582403515426)*x[0]+(0.0311147864546)*x[1]**o+(0.397740429479)*x[1]
        s[1,3,2,1]=(0.51510647545)*x[0]**o+(0.179454689412)*x[0]+(-0.864589732719)*x[1]**o+(-0.919542815326)*x[1]
        s[1,4,0,0]=(0.176385360146)*x[0]**o+(0.666029075488)*x[0]+(-0.765133182482)*x[1]**o+(0.506226091771)*x[1]
        s[1,4,0,1]=(-0.217264015218)*x[0]**o+(0.997799494004)*x[0]+(0.706270854724)*x[1]**o+(-0.440854336576)*x[1]
        s[1,4,1,0]=(-0.127481889699)*x[0]**o+(-0.959927193428)*x[0]+(-0.314190555296)*x[1]**o+(-0.0254770240779)*x[1]
        s[1,4,1,1]=(0.943927827542)*x[0]**o+(0.695303968245)*x[0]+(0.537165245123)*x[1]**o+(-0.414505300627)*x[1]
        s[1,4,2,0]=(-0.583859597164)*x[0]**o+(-0.71993694674)*x[0]+(0.0390198271077)*x[1]**o+(0.917816363923)*x[1]
        s[1,4,2,1]=(0.752163371766)*x[0]**o+(0.47299437482)*x[0]+(-0.499822965518)*x[1]**o+(-0.629775835404)*x[1]
        s[2,0,0,0]=(0.441638332575)*x[0]**o+(0.183300755854)*x[0]+(0.635094026763)*x[1]**o+(-0.981751131017)*x[1]
        s[2,0,0,1]=(0.564782099962)*x[0]**o+(-0.00355099531182)*x[0]+(-0.309065488817)*x[1]**o+(-0.805388517953)*x[1]
        s[2,0,1,0]=(-0.152274122381)*x[0]**o+(-0.884657204578)*x[0]+(-0.362170305747)*x[1]**o+(0.688413345566)*x[1]
        s[2,0,1,1]=(0.433159898578)*x[0]**o+(-0.926183453021)*x[0]+(-0.990900351089)*x[1]**o+(-0.558683257136)*x[1]
        s[2,0,2,0]=(0.378679325956)*x[0]**o+(-0.178401685533)*x[0]+(0.294582559995)*x[1]**o+(0.324129328685)*x[1]
        s[2,0,2,1]=(0.191782222208)*x[0]**o+(-0.718853143054)*x[0]+(-0.1561779489)*x[1]**o+(0.122297214902)*x[1]
        s[2,1,0,0]=(-0.745888747317)*x[0]**o+(0.124662622215)*x[0]+(0.61762073845)*x[1]**o+(-0.543673681976)*x[1]
        s[2,1,0,1]=(0.696268656575)*x[0]**o+(0.879682905689)*x[0]+(0.101543002235)*x[1]**o+(0.795349066813)*x[1]
        s[2,1,1,0]=(-0.595042825753)*x[0]**o+(0.244946186339)*x[0]+(0.107868704293)*x[1]**o+(0.20527282324)*x[1]
        s[2,1,1,1]=(0.702092863468)*x[0]**o+(-0.95898866728)*x[0]+(0.415299739291)*x[1]**o+(-0.602967505073)*x[1]
        s[2,1,2,0]=(0.242441579412)*x[0]**o+(-0.621648829236)*x[0]+(0.00279977134428)*x[1]**o+(0.20444394565)*x[1]
        s[2,1,2,1]=(-0.248160979861)*x[0]**o+(0.510934355012)*x[0]+(-0.888013212236)*x[1]**o+(0.348960154636)*x[1]
        s[2,2,0,0]=(-0.484727669828)*x[0]**o+(0.23183410781)*x[0]+(0.771832910785)*x[1]**o+(-0.259023608026)*x[1]
        s[2,2,0,1]=(-0.0293853663853)*x[0]**o+(0.132402850017)*x[0]+(0.460425782419)*x[1]**o+(0.918697618402)*x[1]
        s[2,2,1,0]=(-0.960506935656)*x[0]**o+(0.302982411255)*x[0]+(0.927623802726)*x[1]**o+(0.446529083445)*x[1]
        s[2,2,1,1]=(-0.0791528123579)*x[0]**o+(-0.963662594671)*x[0]+(0.445231487496)*x[1]**o+(-0.635556750912)*x[1]
        s[2,2,2,0]=(0.119522639306)*x[0]**o+(0.999196602488)*x[0]+(0.75598380813)*x[1]**o+(0.159012027458)*x[1]
        s[2,2,2,1]=(0.671543289198)*x[0]**o+(-0.897542325495)*x[0]+(0.184767661479)*x[1]**o+(-0.679101227287)*x[1]
        s[2,3,0,0]=(0.705751399559)*x[0]**o+(0.0343552696754)*x[0]+(0.295737350313)*x[1]**o+(0.603031693221)*x[1]
        s[2,3,0,1]=(-0.568041111686)*x[0]**o+(0.795100811027)*x[0]+(-0.853308139159)*x[1]**o+(-0.198019398328)*x[1]
        s[2,3,1,0]=(-0.568416608732)*x[0]**o+(-0.785696005424)*x[0]+(0.734563901681)*x[1]**o+(0.337683654534)*x[1]
        s[2,3,1,1]=(0.750616777015)*x[0]**o+(-0.243919199772)*x[0]+(0.47894230818)*x[1]**o+(0.117853316761)*x[1]
        s[2,3,2,0]=(0.595655270733)*x[0]**o+(-0.131631796604)*x[0]+(-0.363246505069)*x[1]**o+(0.0399311321376)*x[1]
        s[2,3,2,1]=(-0.871124094437)*x[0]**o+(-0.495878262772)*x[0]+(-0.59370100135)*x[1]**o+(0.745382881709)*x[1]
        s[2,4,0,0]=(0.479665892387)*x[0]**o+(-0.393886135683)*x[0]+(-0.969679118858)*x[1]**o+(-0.720467296702)*x[1]
        s[2,4,0,1]=(0.111063753944)*x[0]**o+(-0.759833999129)*x[0]+(-0.123061597258)*x[1]**o+(-0.982096877401)*x[1]
        s[2,4,1,0]=(-0.901470811302)*x[0]**o+(0.395087698063)*x[0]+(0.553738458787)*x[1]**o+(-0.112328034611)*x[1]
        s[2,4,1,1]=(0.39514286938)*x[0]**o+(-0.923250016486)*x[0]+(0.361513401174)*x[1]**o+(-0.272660910785)*x[1]
        s[2,4,2,0]=(0.454263938343)*x[0]**o+(-0.0921743296726)*x[0]+(-0.867658322099)*x[1]**o+(-0.469179181142)*x[1]
        s[2,4,2,1]=(0.00363399429065)*x[0]**o+(0.295615173633)*x[0]+(-0.619556300313)*x[1]**o+(0.385912799652)*x[1]
        s[3,0,0,0]=(-0.176638548482)*x[0]**o+(0.716817113557)*x[0]+(0.713969909735)*x[1]**o+(-0.315381536549)*x[1]
        s[3,0,0,1]=(-0.981761588756)*x[0]**o+(-0.264195260433)*x[0]+(0.514278247605)*x[1]**o+(0.620939057298)*x[1]
        s[3,0,1,0]=(-0.599489307325)*x[0]**o+(0.213895727472)*x[0]+(0.986896902529)*x[1]**o+(0.423516351738)*x[1]
        s[3,0,1,1]=(0.282281951751)*x[0]**o+(0.198509073556)*x[0]+(0.261590765595)*x[1]**o+(0.753982289812)*x[1]
        s[3,0,2,0]=(0.78593458655)*x[0]**o+(-0.52358431526)*x[0]+(0.479533127253)*x[1]**o+(0.120268547872)*x[1]
        s[3,0,2,1]=(0.229057209441)*x[0]**o+(-0.200667079463)*x[0]+(0.713824885941)*x[1]**o+(-0.317631878875)*x[1]
        s[3,1,0,0]=(-0.621475342986)*x[0]**o+(-0.776830750036)*x[0]+(0.74327888403)*x[1]**o+(0.77144000144)*x[1]
        s[3,1,0,1]=(-0.0701637401121)*x[0]**o+(0.315253792233)*x[0]+(0.714679121123)*x[1]**o+(-0.124524161083)*x[1]
        s[3,1,1,0]=(0.58339165807)*x[0]**o+(0.754756142322)*x[0]+(-0.552898566299)*x[1]**o+(0.0401072453256)*x[1]
        s[3,1,1,1]=(-0.704826353788)*x[0]**o+(-0.850442055451)*x[0]+(0.0185807835768)*x[1]**o+(0.798634373727)*x[1]
        s[3,1,2,0]=(-0.304968678423)*x[0]**o+(0.882631700011)*x[0]+(-0.00670582763206)*x[1]**o+(-0.790997780888)*x[1]
        s[3,1,2,1]=(0.978052946899)*x[0]**o+(0.0822751454714)*x[0]+(-0.141545511663)*x[1]**o+(0.804142415097)*x[1]
        s[3,2,0,0]=(-0.233834449356)*x[0]**o+(-0.349324841655)*x[0]+(0.633317011256)*x[1]**o+(-0.682052363655)*x[1]
        s[3,2,0,1]=(-0.828772677459)*x[0]**o+(-0.407293570997)*x[0]+(0.0378365988776)*x[1]**o+(0.575150943785)*x[1]
        s[3,2,1,0]=(0.00530675292728)*x[0]**o+(-0.583321711314)*x[0]+(-0.461755026035)*x[1]**o+(-0.234209819371)*x[1]
        s[3,2,1,1]=(-0.529179272486)*x[0]**o+(-0.126586669403)*x[0]+(0.199780481034)*x[1]**o+(-0.475384002433)*x[1]
        s[3,2,2,0]=(-0.794495647543)*x[0]**o+(0.586459171803)*x[0]+(-0.417091337526)*x[1]**o+(-0.920839445366)*x[1]
        s[3,2,2,1]=(-0.543510366902)*x[0]**o+(0.9390916149)*x[0]+(-0.378905656214)*x[1]**o+(-0.580144827894)*x[1]
        s[3,3,0,0]=(-0.218318771891)*x[0]**o+(-0.741715169744)*x[0]+(0.471332960959)*x[1]**o+(-0.942331201293)*x[1]
        s[3,3,0,1]=(-0.639400890732)*x[0]**o+(-0.35086784511)*x[0]+(-0.662448518836)*x[1]**o+(-0.109260566936)*x[1]
        s[3,3,1,0]=(0.00137953261949)*x[0]**o+(-0.527278610973)*x[0]+(0.547316965109)*x[1]**o+(0.747307046881)*x[1]
        s[3,3,1,1]=(-0.110490640565)*x[0]**o+(0.30599876694)*x[0]+(0.464447167957)*x[1]**o+(-0.650215871098)*x[1]
        s[3,3,2,0]=(-0.0752004672453)*x[0]**o+(-0.476012931477)*x[0]+(-0.463766892732)*x[1]**o+(-0.864312392032)*x[1]
        s[3,3,2,1]=(-0.614219705651)*x[0]**o+(0.579350508101)*x[0]+(-0.307032385253)*x[1]**o+(0.609948162009)*x[1]
        s[3,4,0,0]=(-0.946145025239)*x[0]**o+(0.383874203169)*x[0]+(-0.128251366911)*x[1]**o+(-0.592420767396)*x[1]
        s[3,4,0,1]=(0.0656946533978)*x[0]**o+(-0.400244444789)*x[0]+(-0.263415867899)*x[1]**o+(-0.241065146194)*x[1]
        s[3,4,1,0]=(-0.44363374655)*x[0]**o+(-0.702360780698)*x[0]+(0.39242809706)*x[1]**o+(-0.499711507536)*x[1]
        s[3,4,1,1]=(-0.790376230857)*x[0]**o+(0.271619243859)*x[0]+(0.221426582034)*x[1]**o+(-0.244571958075)*x[1]
        s[3,4,2,0]=(-0.599565364787)*x[0]**o+(0.0637522137783)*x[0]+(0.362816245637)*x[1]**o+(-0.52786138669)*x[1]
        s[3,4,2,1]=(0.955472544947)*x[0]**o+(-0.368132248377)*x[0]+(0.262396158645)*x[1]**o+(-0.252857678444)*x[1]
        ref[0,0,0,0]=(-0.711963461677)*x_ref[0]**o+(0.542670298734)*x_ref[0]+(0.28075382249)*x_ref[1]**o+(0.667437625224)*x_ref[1]
        ref[0,0,0,1]=(0.573441578286)*x_ref[0]**o+(0.506095821827)*x_ref[0]+(0.150993391009)*x_ref[1]**o+(0.526409669633)*x_ref[1]
        ref[0,0,1,0]=(0.706627280956)*x_ref[0]**o+(0.57780902263)*x_ref[0]+(0.270996212375)*x_ref[1]**o+(0.938868396109)*x_ref[1]
        ref[0,0,1,1]=(0.970926483868)*x_ref[0]**o+(0.0967183437714)*x_ref[0]+(0.305310109546)*x_ref[1]**o+(-0.746529372684)*x_ref[1]
        ref[0,0,2,0]=(-0.0598871068472)*x_ref[0]**o+(0.441097065872)*x_ref[0]+(0.240831657822)*x_ref[1]**o+(-0.781319614159)*x_ref[1]
        ref[0,0,2,1]=(0.0738716023753)*x_ref[0]**o+(-0.280473432627)*x_ref[0]+(0.351688231305)*x_ref[1]**o+(0.287190769396)*x_ref[1]
        ref[0,1,0,0]=(-0.662029332598)*x_ref[0]**o+(0.769980048213)*x_ref[0]+(-0.316622870493)*x_ref[1]**o+(0.756839554902)*x_ref[1]
        ref[0,1,0,1]=(-0.931130133674)*x_ref[0]**o+(-0.70789066479)*x_ref[0]+(0.350649077237)*x_ref[1]**o+(-0.340577841771)*x_ref[1]
        ref[0,1,1,0]=(0.541533987605)*x_ref[0]**o+(-0.871063950931)*x_ref[0]+(0.198020061012)*x_ref[1]**o+(0.802998878081)*x_ref[1]
        ref[0,1,1,1]=(0.264037408365)*x_ref[0]**o+(-0.0557314225594)*x_ref[0]+(-0.684018523812)*x_ref[1]**o+(-0.229282658383)*x_ref[1]
        ref[0,1,2,0]=(0.637810239569)*x_ref[0]**o+(0.552431819202)*x_ref[0]+(0.0939985703867)*x_ref[1]**o+(0.828751113775)*x_ref[1]
        ref[0,1,2,1]=(-0.374671159094)*x_ref[0]**o+(0.664088936153)*x_ref[0]+(-0.345847704492)*x_ref[1]**o+(-0.364075081521)*x_ref[1]
        ref[0,2,0,0]=(0.00895781289078)*x_ref[0]**o+(0.399614286422)*x_ref[0]+(0.341683450789)*x_ref[1]**o+(0.167206336911)*x_ref[1]
        ref[0,2,0,1]=(-0.104604431902)*x_ref[0]**o+(-0.659107995643)*x_ref[0]+(0.680691727783)*x_ref[1]**o+(-0.684349104067)*x_ref[1]
        ref[0,2,1,0]=(0.710804511082)*x_ref[0]**o+(-0.972641516063)*x_ref[0]+(0.330095093159)*x_ref[1]**o+(0.937337443354)*x_ref[1]
        ref[0,2,1,1]=(-0.294227883996)*x_ref[0]**o+(0.965392046246)*x_ref[0]+(0.714517712165)*x_ref[1]**o+(0.449886137172)*x_ref[1]
        ref[0,2,2,0]=(-0.825491552072)*x_ref[0]**o+(-0.638472841016)*x_ref[0]+(-0.191336005365)*x_ref[1]**o+(-0.422472261204)*x_ref[1]
        ref[0,2,2,1]=(-0.23092704399)*x_ref[0]**o+(-0.637860425921)*x_ref[0]+(0.840387209559)*x_ref[1]**o+(0.165561089405)*x_ref[1]
        ref[0,3,0,0]=(0.420745779316)*x_ref[0]**o+(-0.311588614725)*x_ref[0]+(-0.553702597164)*x_ref[1]**o+(0.272368731382)*x_ref[1]
        ref[0,3,0,1]=(0.327243090728)*x_ref[0]**o+(-0.386774129727)*x_ref[0]+(-0.950289816609)*x_ref[1]**o+(0.662365527837)*x_ref[1]
        ref[0,3,1,0]=(-0.889641002357)*x_ref[0]**o+(-0.195916682305)*x_ref[0]+(-0.60085377073)*x_ref[1]**o+(-0.191887529995)*x_ref[1]
        ref[0,3,1,1]=(0.987376581443)*x_ref[0]**o+(-0.0350561501803)*x_ref[0]+(0.704905530249)*x_ref[1]**o+(-0.735687404507)*x_ref[1]
        ref[0,3,2,0]=(0.420809308115)*x_ref[0]**o+(-0.339666948082)*x_ref[0]+(0.996296976276)*x_ref[1]**o+(0.506174123674)*x_ref[1]
        ref[0,3,2,1]=(0.426714798574)*x_ref[0]**o+(0.620612017549)*x_ref[0]+(0.60846811755)*x_ref[1]**o+(0.561758397959)*x_ref[1]
        ref[0,4,0,0]=(0.0193321433149)*x_ref[0]**o+(0.395174248733)*x_ref[0]+(-0.631710108856)*x_ref[1]**o+(0.562496027755)*x_ref[1]
        ref[0,4,0,1]=(0.936397656738)*x_ref[0]**o+(-0.620128461169)*x_ref[0]+(0.424872968744)*x_ref[1]**o+(0.936360345969)*x_ref[1]
        ref[0,4,1,0]=(-0.823655625312)*x_ref[0]**o+(0.110021832964)*x_ref[0]+(-0.480484367041)*x_ref[1]**o+(0.613626788942)*x_ref[1]
        ref[0,4,1,1]=(0.516537102221)*x_ref[0]**o+(-0.465693898711)*x_ref[0]+(-0.0649481633029)*x_ref[1]**o+(0.498343667101)*x_ref[1]
        ref[0,4,2,0]=(0.364997630891)*x_ref[0]**o+(0.726894447639)*x_ref[0]+(0.729067579856)*x_ref[1]**o+(-0.786376612921)*x_ref[1]
        ref[0,4,2,1]=(-0.574801923675)*x_ref[0]**o+(-0.769120062116)*x_ref[0]+(0.0802603764785)*x_ref[1]**o+(0.322665268053)*x_ref[1]
        ref[1,0,0,0]=(-0.497916263484)*x_ref[0]**o+(0.0689477158621)*x_ref[0]+(-0.59065778662)*x_ref[1]**o+(0.620475916092)*x_ref[1]
        ref[1,0,0,1]=(0.928072689885)*x_ref[0]**o+(0.550220523492)*x_ref[0]+(0.384056461821)*x_ref[1]**o+(0.991195851819)*x_ref[1]
        ref[1,0,1,0]=(-0.171652790551)*x_ref[0]**o+(-0.310223770891)*x_ref[0]+(0.505801949231)*x_ref[1]**o+(-0.809872602671)*x_ref[1]
        ref[1,0,1,1]=(-0.843616998175)*x_ref[0]**o+(-0.514683249854)*x_ref[0]+(0.22719571088)*x_ref[1]**o+(-0.801964932709)*x_ref[1]
        ref[1,0,2,0]=(-0.816576821632)*x_ref[0]**o+(-0.629668603878)*x_ref[0]+(-0.296562767682)*x_ref[1]**o+(-0.0121908797722)*x_ref[1]
        ref[1,0,2,1]=(-0.816221470582)*x_ref[0]**o+(0.101678348599)*x_ref[0]+(0.0841692807823)*x_ref[1]**o+(0.265517698128)*x_ref[1]
        ref[1,1,0,0]=(0.74252349936)*x_ref[0]**o+(0.488296875375)*x_ref[0]+(0.553091389727)*x_ref[1]**o+(0.0951295346041)*x_ref[1]
        ref[1,1,0,1]=(0.573858954955)*x_ref[0]**o+(-0.0210888969717)*x_ref[0]+(-0.813464794029)*x_ref[1]**o+(0.760502343882)*x_ref[1]
        ref[1,1,1,0]=(0.806002928884)*x_ref[0]**o+(0.191962011356)*x_ref[0]+(0.788308484735)*x_ref[1]**o+(0.325349383493)*x_ref[1]
        ref[1,1,1,1]=(-0.975008178175)*x_ref[0]**o+(-0.525299302456)*x_ref[0]+(-0.718959301598)*x_ref[1]**o+(0.821616294375)*x_ref[1]
        ref[1,1,2,0]=(-0.441476557684)*x_ref[0]**o+(0.442725962693)*x_ref[0]+(0.605338741812)*x_ref[1]**o+(-0.0356460149592)*x_ref[1]
        ref[1,1,2,1]=(0.75296734605)*x_ref[0]**o+(0.0218514182228)*x_ref[0]+(-0.408708955555)*x_ref[1]**o+(-0.729191065301)*x_ref[1]
        ref[1,2,0,0]=(0.24165794152)*x_ref[0]**o+(-0.33924002011)*x_ref[0]+(0.357553913734)*x_ref[1]**o+(0.166312753501)*x_ref[1]
        ref[1,2,0,1]=(0.835534214783)*x_ref[0]**o+(-0.688462973774)*x_ref[0]+(0.779388430668)*x_ref[1]**o+(-0.869108490882)*x_ref[1]
        ref[1,2,1,0]=(0.172345662475)*x_ref[0]**o+(0.323742121947)*x_ref[0]+(-0.296938383275)*x_ref[1]**o+(0.0359645605525)*x_ref[1]
        ref[1,2,1,1]=(-0.596600537083)*x_ref[0]**o+(0.644152951661)*x_ref[0]+(0.609094045392)*x_ref[1]**o+(-0.312234180863)*x_ref[1]
        ref[1,2,2,0]=(0.484589912205)*x_ref[0]**o+(-0.906567695494)*x_ref[0]+(0.811265520788)*x_ref[1]**o+(0.444799696239)*x_ref[1]
        ref[1,2,2,1]=(0.0254720755485)*x_ref[0]**o+(-0.41174472105)*x_ref[0]+(0.90099646911)*x_ref[1]**o+(-0.33088554417)*x_ref[1]
        ref[1,3,0,0]=(-0.281155224735)*x_ref[0]**o+(0.274137552899)*x_ref[0]+(0.0974242478848)*x_ref[1]**o+(-0.995002288445)*x_ref[1]
        ref[1,3,0,1]=(0.907653805131)*x_ref[0]**o+(0.743984237516)*x_ref[0]+(-0.776028924136)*x_ref[1]**o+(0.209784894077)*x_ref[1]
        ref[1,3,1,0]=(0.0470679202395)*x_ref[0]**o+(-0.809937294063)*x_ref[0]+(0.956184337312)*x_ref[1]**o+(-0.539155228812)*x_ref[1]
        ref[1,3,1,1]=(0.720979862876)*x_ref[0]**o+(0.412993117346)*x_ref[0]+(0.259998599343)*x_ref[1]**o+(-0.739191117502)*x_ref[1]
        ref[1,3,2,0]=(0.403493857435)*x_ref[0]**o+(0.0582403515426)*x_ref[0]+(0.0311147864546)*x_ref[1]**o+(0.397740429479)*x_ref[1]
        ref[1,3,2,1]=(0.51510647545)*x_ref[0]**o+(0.179454689412)*x_ref[0]+(-0.864589732719)*x_ref[1]**o+(-0.919542815326)*x_ref[1]
        ref[1,4,0,0]=(0.176385360146)*x_ref[0]**o+(0.666029075488)*x_ref[0]+(-0.765133182482)*x_ref[1]**o+(0.506226091771)*x_ref[1]
        ref[1,4,0,1]=(-0.217264015218)*x_ref[0]**o+(0.997799494004)*x_ref[0]+(0.706270854724)*x_ref[1]**o+(-0.440854336576)*x_ref[1]
        ref[1,4,1,0]=(-0.127481889699)*x_ref[0]**o+(-0.959927193428)*x_ref[0]+(-0.314190555296)*x_ref[1]**o+(-0.0254770240779)*x_ref[1]
        ref[1,4,1,1]=(0.943927827542)*x_ref[0]**o+(0.695303968245)*x_ref[0]+(0.537165245123)*x_ref[1]**o+(-0.414505300627)*x_ref[1]
        ref[1,4,2,0]=(-0.583859597164)*x_ref[0]**o+(-0.71993694674)*x_ref[0]+(0.0390198271077)*x_ref[1]**o+(0.917816363923)*x_ref[1]
        ref[1,4,2,1]=(0.752163371766)*x_ref[0]**o+(0.47299437482)*x_ref[0]+(-0.499822965518)*x_ref[1]**o+(-0.629775835404)*x_ref[1]
        ref[2,0,0,0]=(0.441638332575)*x_ref[0]**o+(0.183300755854)*x_ref[0]+(0.635094026763)*x_ref[1]**o+(-0.981751131017)*x_ref[1]
        ref[2,0,0,1]=(0.564782099962)*x_ref[0]**o+(-0.00355099531182)*x_ref[0]+(-0.309065488817)*x_ref[1]**o+(-0.805388517953)*x_ref[1]
        ref[2,0,1,0]=(-0.152274122381)*x_ref[0]**o+(-0.884657204578)*x_ref[0]+(-0.362170305747)*x_ref[1]**o+(0.688413345566)*x_ref[1]
        ref[2,0,1,1]=(0.433159898578)*x_ref[0]**o+(-0.926183453021)*x_ref[0]+(-0.990900351089)*x_ref[1]**o+(-0.558683257136)*x_ref[1]
        ref[2,0,2,0]=(0.378679325956)*x_ref[0]**o+(-0.178401685533)*x_ref[0]+(0.294582559995)*x_ref[1]**o+(0.324129328685)*x_ref[1]
        ref[2,0,2,1]=(0.191782222208)*x_ref[0]**o+(-0.718853143054)*x_ref[0]+(-0.1561779489)*x_ref[1]**o+(0.122297214902)*x_ref[1]
        ref[2,1,0,0]=(-0.745888747317)*x_ref[0]**o+(0.124662622215)*x_ref[0]+(0.61762073845)*x_ref[1]**o+(-0.543673681976)*x_ref[1]
        ref[2,1,0,1]=(0.696268656575)*x_ref[0]**o+(0.879682905689)*x_ref[0]+(0.101543002235)*x_ref[1]**o+(0.795349066813)*x_ref[1]
        ref[2,1,1,0]=(-0.595042825753)*x_ref[0]**o+(0.244946186339)*x_ref[0]+(0.107868704293)*x_ref[1]**o+(0.20527282324)*x_ref[1]
        ref[2,1,1,1]=(0.702092863468)*x_ref[0]**o+(-0.95898866728)*x_ref[0]+(0.415299739291)*x_ref[1]**o+(-0.602967505073)*x_ref[1]
        ref[2,1,2,0]=(0.242441579412)*x_ref[0]**o+(-0.621648829236)*x_ref[0]+(0.00279977134428)*x_ref[1]**o+(0.20444394565)*x_ref[1]
        ref[2,1,2,1]=(-0.248160979861)*x_ref[0]**o+(0.510934355012)*x_ref[0]+(-0.888013212236)*x_ref[1]**o+(0.348960154636)*x_ref[1]
        ref[2,2,0,0]=(-0.484727669828)*x_ref[0]**o+(0.23183410781)*x_ref[0]+(0.771832910785)*x_ref[1]**o+(-0.259023608026)*x_ref[1]
        ref[2,2,0,1]=(-0.0293853663853)*x_ref[0]**o+(0.132402850017)*x_ref[0]+(0.460425782419)*x_ref[1]**o+(0.918697618402)*x_ref[1]
        ref[2,2,1,0]=(-0.960506935656)*x_ref[0]**o+(0.302982411255)*x_ref[0]+(0.927623802726)*x_ref[1]**o+(0.446529083445)*x_ref[1]
        ref[2,2,1,1]=(-0.0791528123579)*x_ref[0]**o+(-0.963662594671)*x_ref[0]+(0.445231487496)*x_ref[1]**o+(-0.635556750912)*x_ref[1]
        ref[2,2,2,0]=(0.119522639306)*x_ref[0]**o+(0.999196602488)*x_ref[0]+(0.75598380813)*x_ref[1]**o+(0.159012027458)*x_ref[1]
        ref[2,2,2,1]=(0.671543289198)*x_ref[0]**o+(-0.897542325495)*x_ref[0]+(0.184767661479)*x_ref[1]**o+(-0.679101227287)*x_ref[1]
        ref[2,3,0,0]=(0.705751399559)*x_ref[0]**o+(0.0343552696754)*x_ref[0]+(0.295737350313)*x_ref[1]**o+(0.603031693221)*x_ref[1]
        ref[2,3,0,1]=(-0.568041111686)*x_ref[0]**o+(0.795100811027)*x_ref[0]+(-0.853308139159)*x_ref[1]**o+(-0.198019398328)*x_ref[1]
        ref[2,3,1,0]=(-0.568416608732)*x_ref[0]**o+(-0.785696005424)*x_ref[0]+(0.734563901681)*x_ref[1]**o+(0.337683654534)*x_ref[1]
        ref[2,3,1,1]=(0.750616777015)*x_ref[0]**o+(-0.243919199772)*x_ref[0]+(0.47894230818)*x_ref[1]**o+(0.117853316761)*x_ref[1]
        ref[2,3,2,0]=(0.595655270733)*x_ref[0]**o+(-0.131631796604)*x_ref[0]+(-0.363246505069)*x_ref[1]**o+(0.0399311321376)*x_ref[1]
        ref[2,3,2,1]=(-0.871124094437)*x_ref[0]**o+(-0.495878262772)*x_ref[0]+(-0.59370100135)*x_ref[1]**o+(0.745382881709)*x_ref[1]
        ref[2,4,0,0]=(0.479665892387)*x_ref[0]**o+(-0.393886135683)*x_ref[0]+(-0.969679118858)*x_ref[1]**o+(-0.720467296702)*x_ref[1]
        ref[2,4,0,1]=(0.111063753944)*x_ref[0]**o+(-0.759833999129)*x_ref[0]+(-0.123061597258)*x_ref[1]**o+(-0.982096877401)*x_ref[1]
        ref[2,4,1,0]=(-0.901470811302)*x_ref[0]**o+(0.395087698063)*x_ref[0]+(0.553738458787)*x_ref[1]**o+(-0.112328034611)*x_ref[1]
        ref[2,4,1,1]=(0.39514286938)*x_ref[0]**o+(-0.923250016486)*x_ref[0]+(0.361513401174)*x_ref[1]**o+(-0.272660910785)*x_ref[1]
        ref[2,4,2,0]=(0.454263938343)*x_ref[0]**o+(-0.0921743296726)*x_ref[0]+(-0.867658322099)*x_ref[1]**o+(-0.469179181142)*x_ref[1]
        ref[2,4,2,1]=(0.00363399429065)*x_ref[0]**o+(0.295615173633)*x_ref[0]+(-0.619556300313)*x_ref[1]**o+(0.385912799652)*x_ref[1]
        ref[3,0,0,0]=(-0.176638548482)*x_ref[0]**o+(0.716817113557)*x_ref[0]+(0.713969909735)*x_ref[1]**o+(-0.315381536549)*x_ref[1]
        ref[3,0,0,1]=(-0.981761588756)*x_ref[0]**o+(-0.264195260433)*x_ref[0]+(0.514278247605)*x_ref[1]**o+(0.620939057298)*x_ref[1]
        ref[3,0,1,0]=(-0.599489307325)*x_ref[0]**o+(0.213895727472)*x_ref[0]+(0.986896902529)*x_ref[1]**o+(0.423516351738)*x_ref[1]
        ref[3,0,1,1]=(0.282281951751)*x_ref[0]**o+(0.198509073556)*x_ref[0]+(0.261590765595)*x_ref[1]**o+(0.753982289812)*x_ref[1]
        ref[3,0,2,0]=(0.78593458655)*x_ref[0]**o+(-0.52358431526)*x_ref[0]+(0.479533127253)*x_ref[1]**o+(0.120268547872)*x_ref[1]
        ref[3,0,2,1]=(0.229057209441)*x_ref[0]**o+(-0.200667079463)*x_ref[0]+(0.713824885941)*x_ref[1]**o+(-0.317631878875)*x_ref[1]
        ref[3,1,0,0]=(-0.621475342986)*x_ref[0]**o+(-0.776830750036)*x_ref[0]+(0.74327888403)*x_ref[1]**o+(0.77144000144)*x_ref[1]
        ref[3,1,0,1]=(-0.0701637401121)*x_ref[0]**o+(0.315253792233)*x_ref[0]+(0.714679121123)*x_ref[1]**o+(-0.124524161083)*x_ref[1]
        ref[3,1,1,0]=(0.58339165807)*x_ref[0]**o+(0.754756142322)*x_ref[0]+(-0.552898566299)*x_ref[1]**o+(0.0401072453256)*x_ref[1]
        ref[3,1,1,1]=(-0.704826353788)*x_ref[0]**o+(-0.850442055451)*x_ref[0]+(0.0185807835768)*x_ref[1]**o+(0.798634373727)*x_ref[1]
        ref[3,1,2,0]=(-0.304968678423)*x_ref[0]**o+(0.882631700011)*x_ref[0]+(-0.00670582763206)*x_ref[1]**o+(-0.790997780888)*x_ref[1]
        ref[3,1,2,1]=(0.978052946899)*x_ref[0]**o+(0.0822751454714)*x_ref[0]+(-0.141545511663)*x_ref[1]**o+(0.804142415097)*x_ref[1]
        ref[3,2,0,0]=(-0.233834449356)*x_ref[0]**o+(-0.349324841655)*x_ref[0]+(0.633317011256)*x_ref[1]**o+(-0.682052363655)*x_ref[1]
        ref[3,2,0,1]=(-0.828772677459)*x_ref[0]**o+(-0.407293570997)*x_ref[0]+(0.0378365988776)*x_ref[1]**o+(0.575150943785)*x_ref[1]
        ref[3,2,1,0]=(0.00530675292728)*x_ref[0]**o+(-0.583321711314)*x_ref[0]+(-0.461755026035)*x_ref[1]**o+(-0.234209819371)*x_ref[1]
        ref[3,2,1,1]=(-0.529179272486)*x_ref[0]**o+(-0.126586669403)*x_ref[0]+(0.199780481034)*x_ref[1]**o+(-0.475384002433)*x_ref[1]
        ref[3,2,2,0]=(-0.794495647543)*x_ref[0]**o+(0.586459171803)*x_ref[0]+(-0.417091337526)*x_ref[1]**o+(-0.920839445366)*x_ref[1]
        ref[3,2,2,1]=(-0.543510366902)*x_ref[0]**o+(0.9390916149)*x_ref[0]+(-0.378905656214)*x_ref[1]**o+(-0.580144827894)*x_ref[1]
        ref[3,3,0,0]=(-0.218318771891)*x_ref[0]**o+(-0.741715169744)*x_ref[0]+(0.471332960959)*x_ref[1]**o+(-0.942331201293)*x_ref[1]
        ref[3,3,0,1]=(-0.639400890732)*x_ref[0]**o+(-0.35086784511)*x_ref[0]+(-0.662448518836)*x_ref[1]**o+(-0.109260566936)*x_ref[1]
        ref[3,3,1,0]=(0.00137953261949)*x_ref[0]**o+(-0.527278610973)*x_ref[0]+(0.547316965109)*x_ref[1]**o+(0.747307046881)*x_ref[1]
        ref[3,3,1,1]=(-0.110490640565)*x_ref[0]**o+(0.30599876694)*x_ref[0]+(0.464447167957)*x_ref[1]**o+(-0.650215871098)*x_ref[1]
        ref[3,3,2,0]=(-0.0752004672453)*x_ref[0]**o+(-0.476012931477)*x_ref[0]+(-0.463766892732)*x_ref[1]**o+(-0.864312392032)*x_ref[1]
        ref[3,3,2,1]=(-0.614219705651)*x_ref[0]**o+(0.579350508101)*x_ref[0]+(-0.307032385253)*x_ref[1]**o+(0.609948162009)*x_ref[1]
        ref[3,4,0,0]=(-0.946145025239)*x_ref[0]**o+(0.383874203169)*x_ref[0]+(-0.128251366911)*x_ref[1]**o+(-0.592420767396)*x_ref[1]
        ref[3,4,0,1]=(0.0656946533978)*x_ref[0]**o+(-0.400244444789)*x_ref[0]+(-0.263415867899)*x_ref[1]**o+(-0.241065146194)*x_ref[1]
        ref[3,4,1,0]=(-0.44363374655)*x_ref[0]**o+(-0.702360780698)*x_ref[0]+(0.39242809706)*x_ref[1]**o+(-0.499711507536)*x_ref[1]
        ref[3,4,1,1]=(-0.790376230857)*x_ref[0]**o+(0.271619243859)*x_ref[0]+(0.221426582034)*x_ref[1]**o+(-0.244571958075)*x_ref[1]
        ref[3,4,2,0]=(-0.599565364787)*x_ref[0]**o+(0.0637522137783)*x_ref[0]+(0.362816245637)*x_ref[1]**o+(-0.52786138669)*x_ref[1]
        ref[3,4,2,1]=(0.955472544947)*x_ref[0]**o+(-0.368132248377)*x_ref[0]+(0.262396158645)*x_ref[1]**o+(-0.252857678444)*x_ref[1]
      else:
        s[0,0,0,0]=(0.18541544761)*x[0]**o+(0.21364695446)*x[0]+(-0.472182966156)*x[1]**o+(0.896603283439)*x[1]+(-0.0688785608079)*x[2]**o+(-0.909852023221)*x[2]
        s[0,0,0,1]=(-0.415942079174)*x[0]**o+(0.875121980159)*x[0]+(0.310534797766)*x[1]**o+(-0.308596058529)*x[1]+(0.108266854467)*x[2]**o+(0.0333359884244)*x[2]
        s[0,0,1,0]=(-0.640298530291)*x[0]**o+(0.0117522671103)*x[0]+(-0.0110933087631)*x[1]**o+(0.265510073738)*x[1]+(0.533125824444)*x[2]**o+(0.608040031534)*x[2]
        s[0,0,1,1]=(-0.614118844709)*x[0]**o+(0.220136724383)*x[0]+(0.500278681433)*x[1]**o+(0.812249768469)*x[1]+(-0.694030974501)*x[2]**o+(-0.0657321516154)*x[2]
        s[0,0,2,0]=(-0.450201716565)*x[0]**o+(0.334136354848)*x[0]+(-0.832011262075)*x[1]**o+(-0.302792705771)*x[1]+(0.143656982651)*x[2]**o+(0.122111775743)*x[2]
        s[0,0,2,1]=(-0.527491596817)*x[0]**o+(0.454281406832)*x[0]+(0.0271798764911)*x[1]**o+(0.101232125515)*x[1]+(-0.16747525537)*x[2]**o+(0.901067653324)*x[2]
        s[0,1,0,0]=(0.124726861265)*x[0]**o+(0.694941253649)*x[0]+(-0.429194779514)*x[1]**o+(-0.857934333758)*x[1]+(-0.156711811013)*x[2]**o+(0.972951797234)*x[2]
        s[0,1,0,1]=(0.342428292381)*x[0]**o+(-0.880577787437)*x[0]+(0.769255197783)*x[1]**o+(0.369977548898)*x[1]+(0.0857387440023)*x[2]**o+(0.679043219801)*x[2]
        s[0,1,1,0]=(-0.393854353246)*x[0]**o+(-0.55861356234)*x[0]+(-0.103173283242)*x[1]**o+(-0.330435961835)*x[1]+(0.169248307533)*x[2]**o+(-0.807306530736)*x[2]
        s[0,1,1,1]=(0.460877309723)*x[0]**o+(-0.318462045167)*x[0]+(-0.397527462539)*x[1]**o+(0.915376261877)*x[1]+(-0.152840673613)*x[2]**o+(-0.877711938515)*x[2]
        s[0,1,2,0]=(-0.709963198272)*x[0]**o+(0.419992355176)*x[0]+(0.453556747804)*x[1]**o+(0.416257482776)*x[1]+(0.157460132998)*x[2]**o+(-0.820379434298)*x[2]
        s[0,1,2,1]=(-0.721496163985)*x[0]**o+(-0.420196812696)*x[0]+(0.494326744722)*x[1]**o+(-0.862937430197)*x[1]+(0.268700000282)*x[2]**o+(0.0964679940875)*x[2]
        s[0,2,0,0]=(-0.72785754306)*x[0]**o+(0.483405287113)*x[0]+(-0.0380836637151)*x[1]**o+(-0.750831233018)*x[1]+(0.890060303635)*x[2]**o+(-0.0889399861718)*x[2]
        s[0,2,0,1]=(-0.020133113466)*x[0]**o+(-0.111314198984)*x[0]+(0.246881700618)*x[1]**o+(-0.0463604335025)*x[1]+(0.340080199964)*x[2]**o+(0.546847705923)*x[2]
        s[0,2,1,0]=(0.44979764213)*x[0]**o+(0.136717449657)*x[0]+(0.885280875379)*x[1]**o+(-0.357374019051)*x[1]+(0.0780825418244)*x[2]**o+(0.407971535477)*x[2]
        s[0,2,1,1]=(-0.97479874766)*x[0]**o+(0.0206982137259)*x[0]+(0.522901951355)*x[1]**o+(-0.455258114156)*x[1]+(-0.0253655613024)*x[2]**o+(0.437644792848)*x[2]
        s[0,2,2,0]=(-0.704117519396)*x[0]**o+(0.529911491014)*x[0]+(0.468182929176)*x[1]**o+(-0.0350363879761)*x[1]+(0.752339337583)*x[2]**o+(-0.216693692569)*x[2]
        s[0,2,2,1]=(-0.684051531073)*x[0]**o+(-0.925299650276)*x[0]+(0.703519111239)*x[1]**o+(0.0606067581517)*x[1]+(-0.0425874860265)*x[2]**o+(0.510164371094)*x[2]
        s[0,3,0,0]=(0.694322896391)*x[0]**o+(0.510283487094)*x[0]+(-0.170009483148)*x[1]**o+(-0.977620371318)*x[1]+(-0.245739552357)*x[2]**o+(0.705681250186)*x[2]
        s[0,3,0,1]=(-0.949754702137)*x[0]**o+(0.866426989442)*x[0]+(-0.86696547128)*x[1]**o+(-0.588446790814)*x[1]+(0.395890571504)*x[2]**o+(-0.423457961295)*x[2]
        s[0,3,1,0]=(0.959539309072)*x[0]**o+(-0.0887741681458)*x[0]+(-0.428764678901)*x[1]**o+(0.181556980404)*x[1]+(0.147860645259)*x[2]**o+(0.608439127363)*x[2]
        s[0,3,1,1]=(-0.4122627159)*x[0]**o+(0.276340029649)*x[0]+(0.514501999828)*x[1]**o+(0.348320416875)*x[1]+(-0.440025117506)*x[2]**o+(-0.402728176651)*x[2]
        s[0,3,2,0]=(-0.19724428313)*x[0]**o+(-0.98620491494)*x[0]+(-0.927442513317)*x[1]**o+(-0.105595731292)*x[1]+(0.265791342515)*x[2]**o+(-0.488717755361)*x[2]
        s[0,3,2,1]=(0.346092758704)*x[0]**o+(0.471986850699)*x[0]+(-0.620240703772)*x[1]**o+(-0.655317618692)*x[1]+(0.552548181694)*x[2]**o+(-0.893031037271)*x[2]
        s[0,4,0,0]=(-0.598836349815)*x[0]**o+(-0.805034268296)*x[0]+(-0.24807713651)*x[1]**o+(0.502403721932)*x[1]+(-0.871875426638)*x[2]**o+(0.975048359849)*x[2]
        s[0,4,0,1]=(-0.848727481846)*x[0]**o+(0.289751356976)*x[0]+(0.866460492643)*x[1]**o+(-0.299228878853)*x[1]+(0.353468499277)*x[2]**o+(0.726026383162)*x[2]
        s[0,4,1,0]=(-0.433026970636)*x[0]**o+(0.952884497458)*x[0]+(0.390828895167)*x[1]**o+(0.460241307465)*x[1]+(-0.826148904445)*x[2]**o+(-0.907790760228)*x[2]
        s[0,4,1,1]=(-0.911683982834)*x[0]**o+(0.713610726367)*x[0]+(0.208806166291)*x[1]**o+(0.0880377414163)*x[1]+(-0.542366224153)*x[2]**o+(0.664202617762)*x[2]
        s[0,4,2,0]=(-0.484661301801)*x[0]**o+(0.636638351317)*x[0]+(0.807801531569)*x[1]**o+(-0.0265380392548)*x[1]+(0.726678256591)*x[2]**o+(0.348640889897)*x[2]
        s[0,4,2,1]=(-0.153834824583)*x[0]**o+(-0.855877123511)*x[0]+(0.539564337213)*x[1]**o+(-0.0218740334893)*x[1]+(0.437159750607)*x[2]**o+(0.482829271016)*x[2]
        s[1,0,0,0]=(-0.0168099698768)*x[0]**o+(-0.989521551028)*x[0]+(0.817466057301)*x[1]**o+(-0.21953132457)*x[1]+(-0.730490846394)*x[2]**o+(0.169063128153)*x[2]
        s[1,0,0,1]=(0.06972569424)*x[0]**o+(-0.775322425858)*x[0]+(-0.671461733727)*x[1]**o+(-0.431923083922)*x[1]+(0.812299957548)*x[2]**o+(0.966461396717)*x[2]
        s[1,0,1,0]=(0.456671962378)*x[0]**o+(0.359745823125)*x[0]+(0.63578449252)*x[1]**o+(0.900009101048)*x[1]+(-0.208011722813)*x[2]**o+(0.604028542303)*x[2]
        s[1,0,1,1]=(-0.0942278500641)*x[0]**o+(-0.54503821106)*x[0]+(-0.240087623461)*x[1]**o+(0.900303365317)*x[1]+(-0.00754914184533)*x[2]**o+(-0.836352541169)*x[2]
        s[1,0,2,0]=(-0.154823915066)*x[0]**o+(-0.984429514577)*x[0]+(0.309393138159)*x[1]**o+(-0.898889250398)*x[1]+(-0.597610752962)*x[2]**o+(-0.595043417245)*x[2]
        s[1,0,2,1]=(-0.452012178484)*x[0]**o+(0.406812802283)*x[0]+(-0.0252372070515)*x[1]**o+(0.599365873757)*x[1]+(0.696137317421)*x[2]**o+(0.39882469864)*x[2]
        s[1,1,0,0]=(0.960061217889)*x[0]**o+(-0.46292996833)*x[0]+(-0.908851982171)*x[1]**o+(-0.35478509058)*x[1]+(-0.0891426046361)*x[2]**o+(0.504413928397)*x[2]
        s[1,1,0,1]=(0.184040383397)*x[0]**o+(-0.125184201347)*x[0]+(0.411368449993)*x[1]**o+(0.76097933208)*x[1]+(0.85874596874)*x[2]**o+(0.0754186631504)*x[2]
        s[1,1,1,0]=(0.670739292338)*x[0]**o+(0.391393649635)*x[0]+(0.171246535424)*x[1]**o+(-0.195052345183)*x[1]+(-0.81284256555)*x[2]**o+(0.729283799083)*x[2]
        s[1,1,1,1]=(-0.456981034012)*x[0]**o+(0.155704355345)*x[0]+(0.480506857776)*x[1]**o+(0.951868527916)*x[1]+(-0.632815963451)*x[2]**o+(-0.281974721192)*x[2]
        s[1,1,2,0]=(0.509907581268)*x[0]**o+(0.974002899767)*x[0]+(-0.487852409248)*x[1]**o+(-0.994031112124)*x[1]+(-0.338175212869)*x[2]**o+(0.0922405530126)*x[2]
        s[1,1,2,1]=(0.240744658624)*x[0]**o+(-0.781993603247)*x[0]+(-0.857979982963)*x[1]**o+(0.138259946342)*x[1]+(-0.0257373573435)*x[2]**o+(-0.549948531512)*x[2]
        s[1,2,0,0]=(0.880823067307)*x[0]**o+(0.769790088552)*x[0]+(0.9242309604)*x[1]**o+(0.710853877125)*x[1]+(0.401127571007)*x[2]**o+(-0.601797131968)*x[2]
        s[1,2,0,1]=(-0.250958562933)*x[0]**o+(0.240041223327)*x[0]+(-0.0868736376912)*x[1]**o+(0.250994177285)*x[1]+(-0.198684086146)*x[2]**o+(0.752504023325)*x[2]
        s[1,2,1,0]=(-0.625906698075)*x[0]**o+(0.0837653273654)*x[0]+(-0.912953810309)*x[1]**o+(-0.591346082423)*x[1]+(0.396077142183)*x[2]**o+(0.199486858965)*x[2]
        s[1,2,1,1]=(-0.870955204115)*x[0]**o+(0.291646033459)*x[0]+(0.0845308439963)*x[1]**o+(-0.427803300865)*x[1]+(0.76119438126)*x[2]**o+(-0.8056578936)*x[2]
        s[1,2,2,0]=(0.649263352322)*x[0]**o+(0.968952679331)*x[0]+(0.473809168621)*x[1]**o+(0.129679304698)*x[1]+(-0.434791766847)*x[2]**o+(-0.141531952455)*x[2]
        s[1,2,2,1]=(-0.680138603152)*x[0]**o+(0.481602504342)*x[0]+(0.0411112054712)*x[1]**o+(-0.335246451045)*x[1]+(-0.532198210944)*x[2]**o+(0.278248356925)*x[2]
        s[1,3,0,0]=(0.983866320172)*x[0]**o+(0.779388801006)*x[0]+(0.943099409418)*x[1]**o+(0.29867320629)*x[1]+(0.920899852268)*x[2]**o+(-0.90449127693)*x[2]
        s[1,3,0,1]=(0.888499557305)*x[0]**o+(-0.0555515101983)*x[0]+(0.833149202095)*x[1]**o+(0.590559034771)*x[1]+(0.765622961327)*x[2]**o+(-0.325502160283)*x[2]
        s[1,3,1,0]=(-0.630488855411)*x[0]**o+(-0.453307600004)*x[0]+(0.773558050428)*x[1]**o+(-0.520443804373)*x[1]+(0.612407257369)*x[2]**o+(0.993705347884)*x[2]
        s[1,3,1,1]=(-0.785312491822)*x[0]**o+(-0.760712669832)*x[0]+(-0.572813097554)*x[1]**o+(0.151654230301)*x[1]+(0.770476334769)*x[2]**o+(-0.663724452997)*x[2]
        s[1,3,2,0]=(0.52562535089)*x[0]**o+(0.137778695278)*x[0]+(0.355394127523)*x[1]**o+(-0.241900298662)*x[1]+(-0.0311531414204)*x[2]**o+(0.61419880417)*x[2]
        s[1,3,2,1]=(0.642395964205)*x[0]**o+(0.938741084322)*x[0]+(-0.0385956334857)*x[1]**o+(0.44257567849)*x[1]+(-0.355655993218)*x[2]**o+(0.060707347503)*x[2]
        s[1,4,0,0]=(-0.363374730402)*x[0]**o+(0.985470783307)*x[0]+(-0.966941636823)*x[1]**o+(-0.0300574872162)*x[1]+(-0.513481421962)*x[2]**o+(0.869607357947)*x[2]
        s[1,4,0,1]=(0.646448255445)*x[0]**o+(-0.19357727026)*x[0]+(0.12200322941)*x[1]**o+(0.560785915298)*x[1]+(0.103714094141)*x[2]**o+(0.0235574798014)*x[2]
        s[1,4,1,0]=(0.658100608473)*x[0]**o+(0.294472394081)*x[0]+(-0.5469350794)*x[1]**o+(0.650910226682)*x[1]+(0.574983452761)*x[2]**o+(-0.885589741854)*x[2]
        s[1,4,1,1]=(-0.833278077163)*x[0]**o+(-0.612685520243)*x[0]+(0.145850763252)*x[1]**o+(0.710544228623)*x[1]+(-0.237536168655)*x[2]**o+(-0.864417422252)*x[2]
        s[1,4,2,0]=(0.547214540024)*x[0]**o+(0.259625305608)*x[0]+(0.972583304662)*x[1]**o+(0.0205295851227)*x[1]+(-0.383299704447)*x[2]**o+(0.899668950656)*x[2]
        s[1,4,2,1]=(-0.0898305090604)*x[0]**o+(0.0583491290278)*x[0]+(0.612069962579)*x[1]**o+(0.00451030954439)*x[1]+(-0.543702915529)*x[2]**o+(0.455742731079)*x[2]
        s[2,0,0,0]=(0.869996673689)*x[0]**o+(0.786072517277)*x[0]+(0.784064187563)*x[1]**o+(0.215121486636)*x[1]+(0.0167704070196)*x[2]**o+(0.510818786007)*x[2]
        s[2,0,0,1]=(0.70636305186)*x[0]**o+(-0.774931605014)*x[0]+(-0.566769222836)*x[1]**o+(-0.490039037574)*x[1]+(0.84779476702)*x[2]**o+(-0.277061243254)*x[2]
        s[2,0,1,0]=(0.71021869545)*x[0]**o+(0.297318884767)*x[0]+(0.393052893616)*x[1]**o+(0.126298772061)*x[1]+(-0.919383761879)*x[2]**o+(0.993331773902)*x[2]
        s[2,0,1,1]=(-0.422176053109)*x[0]**o+(0.1573461669)*x[0]+(-0.383086926798)*x[1]**o+(0.521025942156)*x[1]+(0.573775622558)*x[2]**o+(-0.424831674319)*x[2]
        s[2,0,2,0]=(-0.555531945291)*x[0]**o+(0.174293068827)*x[0]+(-0.937417557253)*x[1]**o+(0.0307979382142)*x[1]+(0.0250257663963)*x[2]**o+(-0.0601088874906)*x[2]
        s[2,0,2,1]=(-0.231539926473)*x[0]**o+(0.943546667559)*x[0]+(-0.453034497139)*x[1]**o+(-0.799323498413)*x[1]+(0.626571998737)*x[2]**o+(0.256254025304)*x[2]
        s[2,1,0,0]=(0.348839391625)*x[0]**o+(0.747443708496)*x[0]+(-0.632807812622)*x[1]**o+(0.30083165118)*x[1]+(-0.694353354492)*x[2]**o+(-0.384450833928)*x[2]
        s[2,1,0,1]=(-0.107737403038)*x[0]**o+(0.298410093423)*x[0]+(0.160830729598)*x[1]**o+(-0.627822690248)*x[1]+(0.511195156768)*x[2]**o+(-0.115525478726)*x[2]
        s[2,1,1,0]=(-0.188820324857)*x[0]**o+(-0.0663606843968)*x[0]+(0.151440137108)*x[1]**o+(0.946232159278)*x[1]+(-0.282849712427)*x[2]**o+(-0.439152308605)*x[2]
        s[2,1,1,1]=(-0.374881709538)*x[0]**o+(0.745922133487)*x[0]+(-0.0242470036044)*x[1]**o+(-0.711004108426)*x[1]+(0.417440756886)*x[2]**o+(-0.510006752436)*x[2]
        s[2,1,2,0]=(0.864353278124)*x[0]**o+(0.0809895383191)*x[0]+(0.457075492213)*x[1]**o+(-0.14392908535)*x[1]+(0.784143277535)*x[2]**o+(-0.15341048191)*x[2]
        s[2,1,2,1]=(0.269079647662)*x[0]**o+(0.0402843915276)*x[0]+(-0.683925387125)*x[1]**o+(0.729804106713)*x[1]+(0.0732202476811)*x[2]**o+(-0.547958146499)*x[2]
        s[2,2,0,0]=(0.272934512065)*x[0]**o+(-0.882225076784)*x[0]+(-0.405012739084)*x[1]**o+(-0.989715644128)*x[1]+(-0.155608583603)*x[2]**o+(-0.722826494052)*x[2]
        s[2,2,0,1]=(0.339423408668)*x[0]**o+(-0.330465697311)*x[0]+(-0.692428979863)*x[1]**o+(0.322765870878)*x[1]+(-0.384201212534)*x[2]**o+(-0.0946217242584)*x[2]
        s[2,2,1,0]=(-0.793318742542)*x[0]**o+(-0.375149299842)*x[0]+(0.793898380613)*x[1]**o+(-0.593499973718)*x[1]+(-0.65815800334)*x[2]**o+(-0.0744951302063)*x[2]
        s[2,2,1,1]=(0.0279370862836)*x[0]**o+(-0.196535478436)*x[0]+(-0.942547794666)*x[1]**o+(0.994269193299)*x[1]+(0.781181650698)*x[2]**o+(-0.221591113746)*x[2]
        s[2,2,2,0]=(0.268315410937)*x[0]**o+(-0.53373824299)*x[0]+(0.777981431744)*x[1]**o+(0.226918217122)*x[1]+(-0.561960454199)*x[2]**o+(-0.273776681445)*x[2]
        s[2,2,2,1]=(0.454454381954)*x[0]**o+(-0.391271318249)*x[0]+(-0.835917284669)*x[1]**o+(-0.42457026252)*x[1]+(-0.848426604454)*x[2]**o+(0.285767008408)*x[2]
        s[2,3,0,0]=(0.771743337775)*x[0]**o+(-0.157913284247)*x[0]+(0.00417714721394)*x[1]**o+(0.875095301539)*x[1]+(0.965135621963)*x[2]**o+(-0.12209631704)*x[2]
        s[2,3,0,1]=(0.00527421717368)*x[0]**o+(0.119837014154)*x[0]+(-0.326451619959)*x[1]**o+(-0.179736365312)*x[1]+(-0.659003332514)*x[2]**o+(0.44402878832)*x[2]
        s[2,3,1,0]=(0.080864510177)*x[0]**o+(-0.910530923746)*x[0]+(-0.065442686402)*x[1]**o+(-0.162655506011)*x[1]+(-0.559755580552)*x[2]**o+(-0.447763051956)*x[2]
        s[2,3,1,1]=(0.45313869387)*x[0]**o+(0.478878867458)*x[0]+(0.0959944374683)*x[1]**o+(-0.0379945879624)*x[1]+(0.877383707005)*x[2]**o+(0.0037296134062)*x[2]
        s[2,3,2,0]=(0.187710780263)*x[0]**o+(0.722012532502)*x[0]+(-0.599862911914)*x[1]**o+(-0.974072653552)*x[1]+(0.632960753399)*x[2]**o+(-0.249506081862)*x[2]
        s[2,3,2,1]=(0.915570663708)*x[0]**o+(-0.590657458642)*x[0]+(-0.221387098604)*x[1]**o+(-0.619974779079)*x[1]+(-0.156470444965)*x[2]**o+(-0.712127719679)*x[2]
        s[2,4,0,0]=(0.57214571034)*x[0]**o+(0.444107582229)*x[0]+(-0.455561895744)*x[1]**o+(0.350956075511)*x[1]+(-0.060920804543)*x[2]**o+(0.102498779175)*x[2]
        s[2,4,0,1]=(-0.117233414571)*x[0]**o+(0.336917030674)*x[0]+(-0.544691169707)*x[1]**o+(0.506633531456)*x[1]+(0.193173288491)*x[2]**o+(0.690898822609)*x[2]
        s[2,4,1,0]=(-0.28899417241)*x[0]**o+(0.930513954761)*x[0]+(-0.564925441731)*x[1]**o+(-0.082228884842)*x[1]+(0.271797239833)*x[2]**o+(-0.293458041224)*x[2]
        s[2,4,1,1]=(0.303746570859)*x[0]**o+(-0.668303989547)*x[0]+(-0.845649340659)*x[1]**o+(0.826906198662)*x[1]+(-0.499957471874)*x[2]**o+(0.540857798511)*x[2]
        s[2,4,2,0]=(0.215998594196)*x[0]**o+(0.80715947344)*x[0]+(0.440392341629)*x[1]**o+(0.429949814054)*x[1]+(-0.476601527283)*x[2]**o+(0.21482287961)*x[2]
        s[2,4,2,1]=(0.744163797227)*x[0]**o+(0.485491624827)*x[0]+(0.0113951682847)*x[1]**o+(-0.219871501999)*x[1]+(-0.960000031149)*x[2]**o+(0.514892677016)*x[2]
        s[3,0,0,0]=(0.955379977654)*x[0]**o+(0.88166711804)*x[0]+(-0.212371823264)*x[1]**o+(0.61537617198)*x[1]+(0.290444741617)*x[2]**o+(0.669594654028)*x[2]
        s[3,0,0,1]=(-0.422932658626)*x[0]**o+(0.898327577684)*x[0]+(0.821836925105)*x[1]**o+(-0.777406619005)*x[1]+(0.836622617839)*x[2]**o+(-0.414551285496)*x[2]
        s[3,0,1,0]=(-0.0922163192963)*x[0]**o+(-0.426319353836)*x[0]+(-0.265988919456)*x[1]**o+(0.0364005746661)*x[1]+(-0.75420756707)*x[2]**o+(0.957027945483)*x[2]
        s[3,0,1,1]=(0.0501511171161)*x[0]**o+(-0.960573277539)*x[0]+(-0.484587580032)*x[1]**o+(-0.229640708516)*x[1]+(-0.652459896537)*x[2]**o+(-0.59981248766)*x[2]
        s[3,0,2,0]=(-0.378646761494)*x[0]**o+(-0.0165689331328)*x[0]+(-0.883277009771)*x[1]**o+(-0.947577778882)*x[1]+(0.188169490614)*x[2]**o+(0.079992543794)*x[2]
        s[3,0,2,1]=(-0.688755441231)*x[0]**o+(0.290810540822)*x[0]+(-0.272544759374)*x[1]**o+(0.412904132844)*x[1]+(-0.851170264556)*x[2]**o+(-0.257140479814)*x[2]
        s[3,1,0,0]=(0.69577276718)*x[0]**o+(0.777468411586)*x[0]+(0.0291906552252)*x[1]**o+(-0.465217099179)*x[1]+(-0.265314241222)*x[2]**o+(-0.914393443339)*x[2]
        s[3,1,0,1]=(-0.605826705227)*x[0]**o+(0.786198128441)*x[0]+(-0.413639776886)*x[1]**o+(0.9484840564)*x[1]+(0.660128431709)*x[2]**o+(0.294794064545)*x[2]
        s[3,1,1,0]=(0.470527104285)*x[0]**o+(-0.673225307232)*x[0]+(0.17528543617)*x[1]**o+(-0.61453857548)*x[1]+(-0.112025849328)*x[2]**o+(0.3719665037)*x[2]
        s[3,1,1,1]=(0.738351761472)*x[0]**o+(-0.692489261597)*x[0]+(-0.00750387276605)*x[1]**o+(-0.194526069313)*x[1]+(0.206427047745)*x[2]**o+(0.936451437125)*x[2]
        s[3,1,2,0]=(0.258949551337)*x[0]**o+(0.772440182785)*x[0]+(0.416917261494)*x[1]**o+(0.61679695051)*x[1]+(0.25182107102)*x[2]**o+(0.404770153496)*x[2]
        s[3,1,2,1]=(0.788357790268)*x[0]**o+(-0.561028019813)*x[0]+(0.842701777125)*x[1]**o+(0.861240216256)*x[1]+(0.633704418848)*x[2]**o+(0.660953804821)*x[2]
        s[3,2,0,0]=(-0.954508015564)*x[0]**o+(0.770831368915)*x[0]+(0.71068286494)*x[1]**o+(-0.706262271265)*x[1]+(0.000711549487662)*x[2]**o+(0.343413819166)*x[2]
        s[3,2,0,1]=(-0.486904541331)*x[0]**o+(-0.535805915583)*x[0]+(-0.697643420586)*x[1]**o+(-0.350844416469)*x[1]+(-0.570854441331)*x[2]**o+(0.844325362411)*x[2]
        s[3,2,1,0]=(0.989057215778)*x[0]**o+(0.822742140223)*x[0]+(0.791720783337)*x[1]**o+(0.0884164131725)*x[1]+(0.337584265788)*x[2]**o+(0.291148268923)*x[2]
        s[3,2,1,1]=(-0.938554739295)*x[0]**o+(0.154409183584)*x[0]+(0.296554553803)*x[1]**o+(0.0130418768484)*x[1]+(0.440811811112)*x[2]**o+(0.0891377617855)*x[2]
        s[3,2,2,0]=(0.0966486195221)*x[0]**o+(-0.228831238429)*x[0]+(-0.00903762540107)*x[1]**o+(0.687947867973)*x[1]+(0.29955036712)*x[2]**o+(-0.153152667621)*x[2]
        s[3,2,2,1]=(0.994414965494)*x[0]**o+(0.787224423165)*x[0]+(-0.572935987357)*x[1]**o+(-0.605628735246)*x[1]+(0.268503770734)*x[2]**o+(-0.32711743813)*x[2]
        s[3,3,0,0]=(-0.539283649631)*x[0]**o+(0.34809741571)*x[0]+(-0.792527713661)*x[1]**o+(-0.636087526521)*x[1]+(0.765496317046)*x[2]**o+(-0.908214892663)*x[2]
        s[3,3,0,1]=(0.903206720181)*x[0]**o+(0.761511999734)*x[0]+(0.649680135137)*x[1]**o+(-0.36181233615)*x[1]+(0.563162618539)*x[2]**o+(-0.810859258189)*x[2]
        s[3,3,1,0]=(-0.0915176526383)*x[0]**o+(0.674948774943)*x[0]+(0.105537180084)*x[1]**o+(0.765853705339)*x[1]+(-0.706270111874)*x[2]**o+(-0.899798116939)*x[2]
        s[3,3,1,1]=(0.247148082226)*x[0]**o+(-0.252654356068)*x[0]+(0.347296182632)*x[1]**o+(0.809639778301)*x[1]+(0.130358549062)*x[2]**o+(0.772783598594)*x[2]
        s[3,3,2,0]=(0.282056095613)*x[0]**o+(0.563274162153)*x[0]+(-0.469420374642)*x[1]**o+(-0.00929250493535)*x[1]+(-0.509250535991)*x[2]**o+(0.747602253057)*x[2]
        s[3,3,2,1]=(-0.40151566811)*x[0]**o+(0.878418940192)*x[0]+(0.548958199553)*x[1]**o+(0.556900329931)*x[1]+(-0.406381288534)*x[2]**o+(-0.767206838398)*x[2]
        s[3,4,0,0]=(-0.672264988169)*x[0]**o+(0.422502707117)*x[0]+(0.459427006408)*x[1]**o+(0.871928411825)*x[1]+(0.85683710126)*x[2]**o+(0.242796550494)*x[2]
        s[3,4,0,1]=(-0.821076005446)*x[0]**o+(-0.337616881342)*x[0]+(0.435309707923)*x[1]**o+(0.989009740264)*x[1]+(-0.470842629471)*x[2]**o+(-0.94476330486)*x[2]
        s[3,4,1,0]=(0.0488301102108)*x[0]**o+(-0.905860578484)*x[0]+(-0.986441486712)*x[1]**o+(0.819623730826)*x[1]+(0.75012214057)*x[2]**o+(-0.460862091553)*x[2]
        s[3,4,1,1]=(-0.932638473603)*x[0]**o+(-0.882471440278)*x[0]+(-0.498748850986)*x[1]**o+(-0.499262858754)*x[1]+(-0.748745750147)*x[2]**o+(-0.168661287934)*x[2]
        s[3,4,2,0]=(0.515883010306)*x[0]**o+(0.752629892309)*x[0]+(-0.340202982107)*x[1]**o+(-0.120916620105)*x[1]+(0.193421000969)*x[2]**o+(0.815950266515)*x[2]
        s[3,4,2,1]=(-0.397114672795)*x[0]**o+(-0.737395536831)*x[0]+(0.589874124917)*x[1]**o+(-0.476304220998)*x[1]+(-0.177825589992)*x[2]**o+(-0.776118337595)*x[2]
        ref[0,0,0,0]=(0.18541544761)*x_ref[0]**o+(0.21364695446)*x_ref[0]+(-0.472182966156)*x_ref[1]**o+(0.896603283439)*x_ref[1]+(-0.0688785608079)*x_ref[2]**o+(-0.909852023221)*x_ref[2]
        ref[0,0,0,1]=(-0.415942079174)*x_ref[0]**o+(0.875121980159)*x_ref[0]+(0.310534797766)*x_ref[1]**o+(-0.308596058529)*x_ref[1]+(0.108266854467)*x_ref[2]**o+(0.0333359884244)*x_ref[2]
        ref[0,0,1,0]=(-0.640298530291)*x_ref[0]**o+(0.0117522671103)*x_ref[0]+(-0.0110933087631)*x_ref[1]**o+(0.265510073738)*x_ref[1]+(0.533125824444)*x_ref[2]**o+(0.608040031534)*x_ref[2]
        ref[0,0,1,1]=(-0.614118844709)*x_ref[0]**o+(0.220136724383)*x_ref[0]+(0.500278681433)*x_ref[1]**o+(0.812249768469)*x_ref[1]+(-0.694030974501)*x_ref[2]**o+(-0.0657321516154)*x_ref[2]
        ref[0,0,2,0]=(-0.450201716565)*x_ref[0]**o+(0.334136354848)*x_ref[0]+(-0.832011262075)*x_ref[1]**o+(-0.302792705771)*x_ref[1]+(0.143656982651)*x_ref[2]**o+(0.122111775743)*x_ref[2]
        ref[0,0,2,1]=(-0.527491596817)*x_ref[0]**o+(0.454281406832)*x_ref[0]+(0.0271798764911)*x_ref[1]**o+(0.101232125515)*x_ref[1]+(-0.16747525537)*x_ref[2]**o+(0.901067653324)*x_ref[2]
        ref[0,1,0,0]=(0.124726861265)*x_ref[0]**o+(0.694941253649)*x_ref[0]+(-0.429194779514)*x_ref[1]**o+(-0.857934333758)*x_ref[1]+(-0.156711811013)*x_ref[2]**o+(0.972951797234)*x_ref[2]
        ref[0,1,0,1]=(0.342428292381)*x_ref[0]**o+(-0.880577787437)*x_ref[0]+(0.769255197783)*x_ref[1]**o+(0.369977548898)*x_ref[1]+(0.0857387440023)*x_ref[2]**o+(0.679043219801)*x_ref[2]
        ref[0,1,1,0]=(-0.393854353246)*x_ref[0]**o+(-0.55861356234)*x_ref[0]+(-0.103173283242)*x_ref[1]**o+(-0.330435961835)*x_ref[1]+(0.169248307533)*x_ref[2]**o+(-0.807306530736)*x_ref[2]
        ref[0,1,1,1]=(0.460877309723)*x_ref[0]**o+(-0.318462045167)*x_ref[0]+(-0.397527462539)*x_ref[1]**o+(0.915376261877)*x_ref[1]+(-0.152840673613)*x_ref[2]**o+(-0.877711938515)*x_ref[2]
        ref[0,1,2,0]=(-0.709963198272)*x_ref[0]**o+(0.419992355176)*x_ref[0]+(0.453556747804)*x_ref[1]**o+(0.416257482776)*x_ref[1]+(0.157460132998)*x_ref[2]**o+(-0.820379434298)*x_ref[2]
        ref[0,1,2,1]=(-0.721496163985)*x_ref[0]**o+(-0.420196812696)*x_ref[0]+(0.494326744722)*x_ref[1]**o+(-0.862937430197)*x_ref[1]+(0.268700000282)*x_ref[2]**o+(0.0964679940875)*x_ref[2]
        ref[0,2,0,0]=(-0.72785754306)*x_ref[0]**o+(0.483405287113)*x_ref[0]+(-0.0380836637151)*x_ref[1]**o+(-0.750831233018)*x_ref[1]+(0.890060303635)*x_ref[2]**o+(-0.0889399861718)*x_ref[2]
        ref[0,2,0,1]=(-0.020133113466)*x_ref[0]**o+(-0.111314198984)*x_ref[0]+(0.246881700618)*x_ref[1]**o+(-0.0463604335025)*x_ref[1]+(0.340080199964)*x_ref[2]**o+(0.546847705923)*x_ref[2]
        ref[0,2,1,0]=(0.44979764213)*x_ref[0]**o+(0.136717449657)*x_ref[0]+(0.885280875379)*x_ref[1]**o+(-0.357374019051)*x_ref[1]+(0.0780825418244)*x_ref[2]**o+(0.407971535477)*x_ref[2]
        ref[0,2,1,1]=(-0.97479874766)*x_ref[0]**o+(0.0206982137259)*x_ref[0]+(0.522901951355)*x_ref[1]**o+(-0.455258114156)*x_ref[1]+(-0.0253655613024)*x_ref[2]**o+(0.437644792848)*x_ref[2]
        ref[0,2,2,0]=(-0.704117519396)*x_ref[0]**o+(0.529911491014)*x_ref[0]+(0.468182929176)*x_ref[1]**o+(-0.0350363879761)*x_ref[1]+(0.752339337583)*x_ref[2]**o+(-0.216693692569)*x_ref[2]
        ref[0,2,2,1]=(-0.684051531073)*x_ref[0]**o+(-0.925299650276)*x_ref[0]+(0.703519111239)*x_ref[1]**o+(0.0606067581517)*x_ref[1]+(-0.0425874860265)*x_ref[2]**o+(0.510164371094)*x_ref[2]
        ref[0,3,0,0]=(0.694322896391)*x_ref[0]**o+(0.510283487094)*x_ref[0]+(-0.170009483148)*x_ref[1]**o+(-0.977620371318)*x_ref[1]+(-0.245739552357)*x_ref[2]**o+(0.705681250186)*x_ref[2]
        ref[0,3,0,1]=(-0.949754702137)*x_ref[0]**o+(0.866426989442)*x_ref[0]+(-0.86696547128)*x_ref[1]**o+(-0.588446790814)*x_ref[1]+(0.395890571504)*x_ref[2]**o+(-0.423457961295)*x_ref[2]
        ref[0,3,1,0]=(0.959539309072)*x_ref[0]**o+(-0.0887741681458)*x_ref[0]+(-0.428764678901)*x_ref[1]**o+(0.181556980404)*x_ref[1]+(0.147860645259)*x_ref[2]**o+(0.608439127363)*x_ref[2]
        ref[0,3,1,1]=(-0.4122627159)*x_ref[0]**o+(0.276340029649)*x_ref[0]+(0.514501999828)*x_ref[1]**o+(0.348320416875)*x_ref[1]+(-0.440025117506)*x_ref[2]**o+(-0.402728176651)*x_ref[2]
        ref[0,3,2,0]=(-0.19724428313)*x_ref[0]**o+(-0.98620491494)*x_ref[0]+(-0.927442513317)*x_ref[1]**o+(-0.105595731292)*x_ref[1]+(0.265791342515)*x_ref[2]**o+(-0.488717755361)*x_ref[2]
        ref[0,3,2,1]=(0.346092758704)*x_ref[0]**o+(0.471986850699)*x_ref[0]+(-0.620240703772)*x_ref[1]**o+(-0.655317618692)*x_ref[1]+(0.552548181694)*x_ref[2]**o+(-0.893031037271)*x_ref[2]
        ref[0,4,0,0]=(-0.598836349815)*x_ref[0]**o+(-0.805034268296)*x_ref[0]+(-0.24807713651)*x_ref[1]**o+(0.502403721932)*x_ref[1]+(-0.871875426638)*x_ref[2]**o+(0.975048359849)*x_ref[2]
        ref[0,4,0,1]=(-0.848727481846)*x_ref[0]**o+(0.289751356976)*x_ref[0]+(0.866460492643)*x_ref[1]**o+(-0.299228878853)*x_ref[1]+(0.353468499277)*x_ref[2]**o+(0.726026383162)*x_ref[2]
        ref[0,4,1,0]=(-0.433026970636)*x_ref[0]**o+(0.952884497458)*x_ref[0]+(0.390828895167)*x_ref[1]**o+(0.460241307465)*x_ref[1]+(-0.826148904445)*x_ref[2]**o+(-0.907790760228)*x_ref[2]
        ref[0,4,1,1]=(-0.911683982834)*x_ref[0]**o+(0.713610726367)*x_ref[0]+(0.208806166291)*x_ref[1]**o+(0.0880377414163)*x_ref[1]+(-0.542366224153)*x_ref[2]**o+(0.664202617762)*x_ref[2]
        ref[0,4,2,0]=(-0.484661301801)*x_ref[0]**o+(0.636638351317)*x_ref[0]+(0.807801531569)*x_ref[1]**o+(-0.0265380392548)*x_ref[1]+(0.726678256591)*x_ref[2]**o+(0.348640889897)*x_ref[2]
        ref[0,4,2,1]=(-0.153834824583)*x_ref[0]**o+(-0.855877123511)*x_ref[0]+(0.539564337213)*x_ref[1]**o+(-0.0218740334893)*x_ref[1]+(0.437159750607)*x_ref[2]**o+(0.482829271016)*x_ref[2]
        ref[1,0,0,0]=(-0.0168099698768)*x_ref[0]**o+(-0.989521551028)*x_ref[0]+(0.817466057301)*x_ref[1]**o+(-0.21953132457)*x_ref[1]+(-0.730490846394)*x_ref[2]**o+(0.169063128153)*x_ref[2]
        ref[1,0,0,1]=(0.06972569424)*x_ref[0]**o+(-0.775322425858)*x_ref[0]+(-0.671461733727)*x_ref[1]**o+(-0.431923083922)*x_ref[1]+(0.812299957548)*x_ref[2]**o+(0.966461396717)*x_ref[2]
        ref[1,0,1,0]=(0.456671962378)*x_ref[0]**o+(0.359745823125)*x_ref[0]+(0.63578449252)*x_ref[1]**o+(0.900009101048)*x_ref[1]+(-0.208011722813)*x_ref[2]**o+(0.604028542303)*x_ref[2]
        ref[1,0,1,1]=(-0.0942278500641)*x_ref[0]**o+(-0.54503821106)*x_ref[0]+(-0.240087623461)*x_ref[1]**o+(0.900303365317)*x_ref[1]+(-0.00754914184533)*x_ref[2]**o+(-0.836352541169)*x_ref[2]
        ref[1,0,2,0]=(-0.154823915066)*x_ref[0]**o+(-0.984429514577)*x_ref[0]+(0.309393138159)*x_ref[1]**o+(-0.898889250398)*x_ref[1]+(-0.597610752962)*x_ref[2]**o+(-0.595043417245)*x_ref[2]
        ref[1,0,2,1]=(-0.452012178484)*x_ref[0]**o+(0.406812802283)*x_ref[0]+(-0.0252372070515)*x_ref[1]**o+(0.599365873757)*x_ref[1]+(0.696137317421)*x_ref[2]**o+(0.39882469864)*x_ref[2]
        ref[1,1,0,0]=(0.960061217889)*x_ref[0]**o+(-0.46292996833)*x_ref[0]+(-0.908851982171)*x_ref[1]**o+(-0.35478509058)*x_ref[1]+(-0.0891426046361)*x_ref[2]**o+(0.504413928397)*x_ref[2]
        ref[1,1,0,1]=(0.184040383397)*x_ref[0]**o+(-0.125184201347)*x_ref[0]+(0.411368449993)*x_ref[1]**o+(0.76097933208)*x_ref[1]+(0.85874596874)*x_ref[2]**o+(0.0754186631504)*x_ref[2]
        ref[1,1,1,0]=(0.670739292338)*x_ref[0]**o+(0.391393649635)*x_ref[0]+(0.171246535424)*x_ref[1]**o+(-0.195052345183)*x_ref[1]+(-0.81284256555)*x_ref[2]**o+(0.729283799083)*x_ref[2]
        ref[1,1,1,1]=(-0.456981034012)*x_ref[0]**o+(0.155704355345)*x_ref[0]+(0.480506857776)*x_ref[1]**o+(0.951868527916)*x_ref[1]+(-0.632815963451)*x_ref[2]**o+(-0.281974721192)*x_ref[2]
        ref[1,1,2,0]=(0.509907581268)*x_ref[0]**o+(0.974002899767)*x_ref[0]+(-0.487852409248)*x_ref[1]**o+(-0.994031112124)*x_ref[1]+(-0.338175212869)*x_ref[2]**o+(0.0922405530126)*x_ref[2]
        ref[1,1,2,1]=(0.240744658624)*x_ref[0]**o+(-0.781993603247)*x_ref[0]+(-0.857979982963)*x_ref[1]**o+(0.138259946342)*x_ref[1]+(-0.0257373573435)*x_ref[2]**o+(-0.549948531512)*x_ref[2]
        ref[1,2,0,0]=(0.880823067307)*x_ref[0]**o+(0.769790088552)*x_ref[0]+(0.9242309604)*x_ref[1]**o+(0.710853877125)*x_ref[1]+(0.401127571007)*x_ref[2]**o+(-0.601797131968)*x_ref[2]
        ref[1,2,0,1]=(-0.250958562933)*x_ref[0]**o+(0.240041223327)*x_ref[0]+(-0.0868736376912)*x_ref[1]**o+(0.250994177285)*x_ref[1]+(-0.198684086146)*x_ref[2]**o+(0.752504023325)*x_ref[2]
        ref[1,2,1,0]=(-0.625906698075)*x_ref[0]**o+(0.0837653273654)*x_ref[0]+(-0.912953810309)*x_ref[1]**o+(-0.591346082423)*x_ref[1]+(0.396077142183)*x_ref[2]**o+(0.199486858965)*x_ref[2]
        ref[1,2,1,1]=(-0.870955204115)*x_ref[0]**o+(0.291646033459)*x_ref[0]+(0.0845308439963)*x_ref[1]**o+(-0.427803300865)*x_ref[1]+(0.76119438126)*x_ref[2]**o+(-0.8056578936)*x_ref[2]
        ref[1,2,2,0]=(0.649263352322)*x_ref[0]**o+(0.968952679331)*x_ref[0]+(0.473809168621)*x_ref[1]**o+(0.129679304698)*x_ref[1]+(-0.434791766847)*x_ref[2]**o+(-0.141531952455)*x_ref[2]
        ref[1,2,2,1]=(-0.680138603152)*x_ref[0]**o+(0.481602504342)*x_ref[0]+(0.0411112054712)*x_ref[1]**o+(-0.335246451045)*x_ref[1]+(-0.532198210944)*x_ref[2]**o+(0.278248356925)*x_ref[2]
        ref[1,3,0,0]=(0.983866320172)*x_ref[0]**o+(0.779388801006)*x_ref[0]+(0.943099409418)*x_ref[1]**o+(0.29867320629)*x_ref[1]+(0.920899852268)*x_ref[2]**o+(-0.90449127693)*x_ref[2]
        ref[1,3,0,1]=(0.888499557305)*x_ref[0]**o+(-0.0555515101983)*x_ref[0]+(0.833149202095)*x_ref[1]**o+(0.590559034771)*x_ref[1]+(0.765622961327)*x_ref[2]**o+(-0.325502160283)*x_ref[2]
        ref[1,3,1,0]=(-0.630488855411)*x_ref[0]**o+(-0.453307600004)*x_ref[0]+(0.773558050428)*x_ref[1]**o+(-0.520443804373)*x_ref[1]+(0.612407257369)*x_ref[2]**o+(0.993705347884)*x_ref[2]
        ref[1,3,1,1]=(-0.785312491822)*x_ref[0]**o+(-0.760712669832)*x_ref[0]+(-0.572813097554)*x_ref[1]**o+(0.151654230301)*x_ref[1]+(0.770476334769)*x_ref[2]**o+(-0.663724452997)*x_ref[2]
        ref[1,3,2,0]=(0.52562535089)*x_ref[0]**o+(0.137778695278)*x_ref[0]+(0.355394127523)*x_ref[1]**o+(-0.241900298662)*x_ref[1]+(-0.0311531414204)*x_ref[2]**o+(0.61419880417)*x_ref[2]
        ref[1,3,2,1]=(0.642395964205)*x_ref[0]**o+(0.938741084322)*x_ref[0]+(-0.0385956334857)*x_ref[1]**o+(0.44257567849)*x_ref[1]+(-0.355655993218)*x_ref[2]**o+(0.060707347503)*x_ref[2]
        ref[1,4,0,0]=(-0.363374730402)*x_ref[0]**o+(0.985470783307)*x_ref[0]+(-0.966941636823)*x_ref[1]**o+(-0.0300574872162)*x_ref[1]+(-0.513481421962)*x_ref[2]**o+(0.869607357947)*x_ref[2]
        ref[1,4,0,1]=(0.646448255445)*x_ref[0]**o+(-0.19357727026)*x_ref[0]+(0.12200322941)*x_ref[1]**o+(0.560785915298)*x_ref[1]+(0.103714094141)*x_ref[2]**o+(0.0235574798014)*x_ref[2]
        ref[1,4,1,0]=(0.658100608473)*x_ref[0]**o+(0.294472394081)*x_ref[0]+(-0.5469350794)*x_ref[1]**o+(0.650910226682)*x_ref[1]+(0.574983452761)*x_ref[2]**o+(-0.885589741854)*x_ref[2]
        ref[1,4,1,1]=(-0.833278077163)*x_ref[0]**o+(-0.612685520243)*x_ref[0]+(0.145850763252)*x_ref[1]**o+(0.710544228623)*x_ref[1]+(-0.237536168655)*x_ref[2]**o+(-0.864417422252)*x_ref[2]
        ref[1,4,2,0]=(0.547214540024)*x_ref[0]**o+(0.259625305608)*x_ref[0]+(0.972583304662)*x_ref[1]**o+(0.0205295851227)*x_ref[1]+(-0.383299704447)*x_ref[2]**o+(0.899668950656)*x_ref[2]
        ref[1,4,2,1]=(-0.0898305090604)*x_ref[0]**o+(0.0583491290278)*x_ref[0]+(0.612069962579)*x_ref[1]**o+(0.00451030954439)*x_ref[1]+(-0.543702915529)*x_ref[2]**o+(0.455742731079)*x_ref[2]
        ref[2,0,0,0]=(0.869996673689)*x_ref[0]**o+(0.786072517277)*x_ref[0]+(0.784064187563)*x_ref[1]**o+(0.215121486636)*x_ref[1]+(0.0167704070196)*x_ref[2]**o+(0.510818786007)*x_ref[2]
        ref[2,0,0,1]=(0.70636305186)*x_ref[0]**o+(-0.774931605014)*x_ref[0]+(-0.566769222836)*x_ref[1]**o+(-0.490039037574)*x_ref[1]+(0.84779476702)*x_ref[2]**o+(-0.277061243254)*x_ref[2]
        ref[2,0,1,0]=(0.71021869545)*x_ref[0]**o+(0.297318884767)*x_ref[0]+(0.393052893616)*x_ref[1]**o+(0.126298772061)*x_ref[1]+(-0.919383761879)*x_ref[2]**o+(0.993331773902)*x_ref[2]
        ref[2,0,1,1]=(-0.422176053109)*x_ref[0]**o+(0.1573461669)*x_ref[0]+(-0.383086926798)*x_ref[1]**o+(0.521025942156)*x_ref[1]+(0.573775622558)*x_ref[2]**o+(-0.424831674319)*x_ref[2]
        ref[2,0,2,0]=(-0.555531945291)*x_ref[0]**o+(0.174293068827)*x_ref[0]+(-0.937417557253)*x_ref[1]**o+(0.0307979382142)*x_ref[1]+(0.0250257663963)*x_ref[2]**o+(-0.0601088874906)*x_ref[2]
        ref[2,0,2,1]=(-0.231539926473)*x_ref[0]**o+(0.943546667559)*x_ref[0]+(-0.453034497139)*x_ref[1]**o+(-0.799323498413)*x_ref[1]+(0.626571998737)*x_ref[2]**o+(0.256254025304)*x_ref[2]
        ref[2,1,0,0]=(0.348839391625)*x_ref[0]**o+(0.747443708496)*x_ref[0]+(-0.632807812622)*x_ref[1]**o+(0.30083165118)*x_ref[1]+(-0.694353354492)*x_ref[2]**o+(-0.384450833928)*x_ref[2]
        ref[2,1,0,1]=(-0.107737403038)*x_ref[0]**o+(0.298410093423)*x_ref[0]+(0.160830729598)*x_ref[1]**o+(-0.627822690248)*x_ref[1]+(0.511195156768)*x_ref[2]**o+(-0.115525478726)*x_ref[2]
        ref[2,1,1,0]=(-0.188820324857)*x_ref[0]**o+(-0.0663606843968)*x_ref[0]+(0.151440137108)*x_ref[1]**o+(0.946232159278)*x_ref[1]+(-0.282849712427)*x_ref[2]**o+(-0.439152308605)*x_ref[2]
        ref[2,1,1,1]=(-0.374881709538)*x_ref[0]**o+(0.745922133487)*x_ref[0]+(-0.0242470036044)*x_ref[1]**o+(-0.711004108426)*x_ref[1]+(0.417440756886)*x_ref[2]**o+(-0.510006752436)*x_ref[2]
        ref[2,1,2,0]=(0.864353278124)*x_ref[0]**o+(0.0809895383191)*x_ref[0]+(0.457075492213)*x_ref[1]**o+(-0.14392908535)*x_ref[1]+(0.784143277535)*x_ref[2]**o+(-0.15341048191)*x_ref[2]
        ref[2,1,2,1]=(0.269079647662)*x_ref[0]**o+(0.0402843915276)*x_ref[0]+(-0.683925387125)*x_ref[1]**o+(0.729804106713)*x_ref[1]+(0.0732202476811)*x_ref[2]**o+(-0.547958146499)*x_ref[2]
        ref[2,2,0,0]=(0.272934512065)*x_ref[0]**o+(-0.882225076784)*x_ref[0]+(-0.405012739084)*x_ref[1]**o+(-0.989715644128)*x_ref[1]+(-0.155608583603)*x_ref[2]**o+(-0.722826494052)*x_ref[2]
        ref[2,2,0,1]=(0.339423408668)*x_ref[0]**o+(-0.330465697311)*x_ref[0]+(-0.692428979863)*x_ref[1]**o+(0.322765870878)*x_ref[1]+(-0.384201212534)*x_ref[2]**o+(-0.0946217242584)*x_ref[2]
        ref[2,2,1,0]=(-0.793318742542)*x_ref[0]**o+(-0.375149299842)*x_ref[0]+(0.793898380613)*x_ref[1]**o+(-0.593499973718)*x_ref[1]+(-0.65815800334)*x_ref[2]**o+(-0.0744951302063)*x_ref[2]
        ref[2,2,1,1]=(0.0279370862836)*x_ref[0]**o+(-0.196535478436)*x_ref[0]+(-0.942547794666)*x_ref[1]**o+(0.994269193299)*x_ref[1]+(0.781181650698)*x_ref[2]**o+(-0.221591113746)*x_ref[2]
        ref[2,2,2,0]=(0.268315410937)*x_ref[0]**o+(-0.53373824299)*x_ref[0]+(0.777981431744)*x_ref[1]**o+(0.226918217122)*x_ref[1]+(-0.561960454199)*x_ref[2]**o+(-0.273776681445)*x_ref[2]
        ref[2,2,2,1]=(0.454454381954)*x_ref[0]**o+(-0.391271318249)*x_ref[0]+(-0.835917284669)*x_ref[1]**o+(-0.42457026252)*x_ref[1]+(-0.848426604454)*x_ref[2]**o+(0.285767008408)*x_ref[2]
        ref[2,3,0,0]=(0.771743337775)*x_ref[0]**o+(-0.157913284247)*x_ref[0]+(0.00417714721394)*x_ref[1]**o+(0.875095301539)*x_ref[1]+(0.965135621963)*x_ref[2]**o+(-0.12209631704)*x_ref[2]
        ref[2,3,0,1]=(0.00527421717368)*x_ref[0]**o+(0.119837014154)*x_ref[0]+(-0.326451619959)*x_ref[1]**o+(-0.179736365312)*x_ref[1]+(-0.659003332514)*x_ref[2]**o+(0.44402878832)*x_ref[2]
        ref[2,3,1,0]=(0.080864510177)*x_ref[0]**o+(-0.910530923746)*x_ref[0]+(-0.065442686402)*x_ref[1]**o+(-0.162655506011)*x_ref[1]+(-0.559755580552)*x_ref[2]**o+(-0.447763051956)*x_ref[2]
        ref[2,3,1,1]=(0.45313869387)*x_ref[0]**o+(0.478878867458)*x_ref[0]+(0.0959944374683)*x_ref[1]**o+(-0.0379945879624)*x_ref[1]+(0.877383707005)*x_ref[2]**o+(0.0037296134062)*x_ref[2]
        ref[2,3,2,0]=(0.187710780263)*x_ref[0]**o+(0.722012532502)*x_ref[0]+(-0.599862911914)*x_ref[1]**o+(-0.974072653552)*x_ref[1]+(0.632960753399)*x_ref[2]**o+(-0.249506081862)*x_ref[2]
        ref[2,3,2,1]=(0.915570663708)*x_ref[0]**o+(-0.590657458642)*x_ref[0]+(-0.221387098604)*x_ref[1]**o+(-0.619974779079)*x_ref[1]+(-0.156470444965)*x_ref[2]**o+(-0.712127719679)*x_ref[2]
        ref[2,4,0,0]=(0.57214571034)*x_ref[0]**o+(0.444107582229)*x_ref[0]+(-0.455561895744)*x_ref[1]**o+(0.350956075511)*x_ref[1]+(-0.060920804543)*x_ref[2]**o+(0.102498779175)*x_ref[2]
        ref[2,4,0,1]=(-0.117233414571)*x_ref[0]**o+(0.336917030674)*x_ref[0]+(-0.544691169707)*x_ref[1]**o+(0.506633531456)*x_ref[1]+(0.193173288491)*x_ref[2]**o+(0.690898822609)*x_ref[2]
        ref[2,4,1,0]=(-0.28899417241)*x_ref[0]**o+(0.930513954761)*x_ref[0]+(-0.564925441731)*x_ref[1]**o+(-0.082228884842)*x_ref[1]+(0.271797239833)*x_ref[2]**o+(-0.293458041224)*x_ref[2]
        ref[2,4,1,1]=(0.303746570859)*x_ref[0]**o+(-0.668303989547)*x_ref[0]+(-0.845649340659)*x_ref[1]**o+(0.826906198662)*x_ref[1]+(-0.499957471874)*x_ref[2]**o+(0.540857798511)*x_ref[2]
        ref[2,4,2,0]=(0.215998594196)*x_ref[0]**o+(0.80715947344)*x_ref[0]+(0.440392341629)*x_ref[1]**o+(0.429949814054)*x_ref[1]+(-0.476601527283)*x_ref[2]**o+(0.21482287961)*x_ref[2]
        ref[2,4,2,1]=(0.744163797227)*x_ref[0]**o+(0.485491624827)*x_ref[0]+(0.0113951682847)*x_ref[1]**o+(-0.219871501999)*x_ref[1]+(-0.960000031149)*x_ref[2]**o+(0.514892677016)*x_ref[2]
        ref[3,0,0,0]=(0.955379977654)*x_ref[0]**o+(0.88166711804)*x_ref[0]+(-0.212371823264)*x_ref[1]**o+(0.61537617198)*x_ref[1]+(0.290444741617)*x_ref[2]**o+(0.669594654028)*x_ref[2]
        ref[3,0,0,1]=(-0.422932658626)*x_ref[0]**o+(0.898327577684)*x_ref[0]+(0.821836925105)*x_ref[1]**o+(-0.777406619005)*x_ref[1]+(0.836622617839)*x_ref[2]**o+(-0.414551285496)*x_ref[2]
        ref[3,0,1,0]=(-0.0922163192963)*x_ref[0]**o+(-0.426319353836)*x_ref[0]+(-0.265988919456)*x_ref[1]**o+(0.0364005746661)*x_ref[1]+(-0.75420756707)*x_ref[2]**o+(0.957027945483)*x_ref[2]
        ref[3,0,1,1]=(0.0501511171161)*x_ref[0]**o+(-0.960573277539)*x_ref[0]+(-0.484587580032)*x_ref[1]**o+(-0.229640708516)*x_ref[1]+(-0.652459896537)*x_ref[2]**o+(-0.59981248766)*x_ref[2]
        ref[3,0,2,0]=(-0.378646761494)*x_ref[0]**o+(-0.0165689331328)*x_ref[0]+(-0.883277009771)*x_ref[1]**o+(-0.947577778882)*x_ref[1]+(0.188169490614)*x_ref[2]**o+(0.079992543794)*x_ref[2]
        ref[3,0,2,1]=(-0.688755441231)*x_ref[0]**o+(0.290810540822)*x_ref[0]+(-0.272544759374)*x_ref[1]**o+(0.412904132844)*x_ref[1]+(-0.851170264556)*x_ref[2]**o+(-0.257140479814)*x_ref[2]
        ref[3,1,0,0]=(0.69577276718)*x_ref[0]**o+(0.777468411586)*x_ref[0]+(0.0291906552252)*x_ref[1]**o+(-0.465217099179)*x_ref[1]+(-0.265314241222)*x_ref[2]**o+(-0.914393443339)*x_ref[2]
        ref[3,1,0,1]=(-0.605826705227)*x_ref[0]**o+(0.786198128441)*x_ref[0]+(-0.413639776886)*x_ref[1]**o+(0.9484840564)*x_ref[1]+(0.660128431709)*x_ref[2]**o+(0.294794064545)*x_ref[2]
        ref[3,1,1,0]=(0.470527104285)*x_ref[0]**o+(-0.673225307232)*x_ref[0]+(0.17528543617)*x_ref[1]**o+(-0.61453857548)*x_ref[1]+(-0.112025849328)*x_ref[2]**o+(0.3719665037)*x_ref[2]
        ref[3,1,1,1]=(0.738351761472)*x_ref[0]**o+(-0.692489261597)*x_ref[0]+(-0.00750387276605)*x_ref[1]**o+(-0.194526069313)*x_ref[1]+(0.206427047745)*x_ref[2]**o+(0.936451437125)*x_ref[2]
        ref[3,1,2,0]=(0.258949551337)*x_ref[0]**o+(0.772440182785)*x_ref[0]+(0.416917261494)*x_ref[1]**o+(0.61679695051)*x_ref[1]+(0.25182107102)*x_ref[2]**o+(0.404770153496)*x_ref[2]
        ref[3,1,2,1]=(0.788357790268)*x_ref[0]**o+(-0.561028019813)*x_ref[0]+(0.842701777125)*x_ref[1]**o+(0.861240216256)*x_ref[1]+(0.633704418848)*x_ref[2]**o+(0.660953804821)*x_ref[2]
        ref[3,2,0,0]=(-0.954508015564)*x_ref[0]**o+(0.770831368915)*x_ref[0]+(0.71068286494)*x_ref[1]**o+(-0.706262271265)*x_ref[1]+(0.000711549487662)*x_ref[2]**o+(0.343413819166)*x_ref[2]
        ref[3,2,0,1]=(-0.486904541331)*x_ref[0]**o+(-0.535805915583)*x_ref[0]+(-0.697643420586)*x_ref[1]**o+(-0.350844416469)*x_ref[1]+(-0.570854441331)*x_ref[2]**o+(0.844325362411)*x_ref[2]
        ref[3,2,1,0]=(0.989057215778)*x_ref[0]**o+(0.822742140223)*x_ref[0]+(0.791720783337)*x_ref[1]**o+(0.0884164131725)*x_ref[1]+(0.337584265788)*x_ref[2]**o+(0.291148268923)*x_ref[2]
        ref[3,2,1,1]=(-0.938554739295)*x_ref[0]**o+(0.154409183584)*x_ref[0]+(0.296554553803)*x_ref[1]**o+(0.0130418768484)*x_ref[1]+(0.440811811112)*x_ref[2]**o+(0.0891377617855)*x_ref[2]
        ref[3,2,2,0]=(0.0966486195221)*x_ref[0]**o+(-0.228831238429)*x_ref[0]+(-0.00903762540107)*x_ref[1]**o+(0.687947867973)*x_ref[1]+(0.29955036712)*x_ref[2]**o+(-0.153152667621)*x_ref[2]
        ref[3,2,2,1]=(0.994414965494)*x_ref[0]**o+(0.787224423165)*x_ref[0]+(-0.572935987357)*x_ref[1]**o+(-0.605628735246)*x_ref[1]+(0.268503770734)*x_ref[2]**o+(-0.32711743813)*x_ref[2]
        ref[3,3,0,0]=(-0.539283649631)*x_ref[0]**o+(0.34809741571)*x_ref[0]+(-0.792527713661)*x_ref[1]**o+(-0.636087526521)*x_ref[1]+(0.765496317046)*x_ref[2]**o+(-0.908214892663)*x_ref[2]
        ref[3,3,0,1]=(0.903206720181)*x_ref[0]**o+(0.761511999734)*x_ref[0]+(0.649680135137)*x_ref[1]**o+(-0.36181233615)*x_ref[1]+(0.563162618539)*x_ref[2]**o+(-0.810859258189)*x_ref[2]
        ref[3,3,1,0]=(-0.0915176526383)*x_ref[0]**o+(0.674948774943)*x_ref[0]+(0.105537180084)*x_ref[1]**o+(0.765853705339)*x_ref[1]+(-0.706270111874)*x_ref[2]**o+(-0.899798116939)*x_ref[2]
        ref[3,3,1,1]=(0.247148082226)*x_ref[0]**o+(-0.252654356068)*x_ref[0]+(0.347296182632)*x_ref[1]**o+(0.809639778301)*x_ref[1]+(0.130358549062)*x_ref[2]**o+(0.772783598594)*x_ref[2]
        ref[3,3,2,0]=(0.282056095613)*x_ref[0]**o+(0.563274162153)*x_ref[0]+(-0.469420374642)*x_ref[1]**o+(-0.00929250493535)*x_ref[1]+(-0.509250535991)*x_ref[2]**o+(0.747602253057)*x_ref[2]
        ref[3,3,2,1]=(-0.40151566811)*x_ref[0]**o+(0.878418940192)*x_ref[0]+(0.548958199553)*x_ref[1]**o+(0.556900329931)*x_ref[1]+(-0.406381288534)*x_ref[2]**o+(-0.767206838398)*x_ref[2]
        ref[3,4,0,0]=(-0.672264988169)*x_ref[0]**o+(0.422502707117)*x_ref[0]+(0.459427006408)*x_ref[1]**o+(0.871928411825)*x_ref[1]+(0.85683710126)*x_ref[2]**o+(0.242796550494)*x_ref[2]
        ref[3,4,0,1]=(-0.821076005446)*x_ref[0]**o+(-0.337616881342)*x_ref[0]+(0.435309707923)*x_ref[1]**o+(0.989009740264)*x_ref[1]+(-0.470842629471)*x_ref[2]**o+(-0.94476330486)*x_ref[2]
        ref[3,4,1,0]=(0.0488301102108)*x_ref[0]**o+(-0.905860578484)*x_ref[0]+(-0.986441486712)*x_ref[1]**o+(0.819623730826)*x_ref[1]+(0.75012214057)*x_ref[2]**o+(-0.460862091553)*x_ref[2]
        ref[3,4,1,1]=(-0.932638473603)*x_ref[0]**o+(-0.882471440278)*x_ref[0]+(-0.498748850986)*x_ref[1]**o+(-0.499262858754)*x_ref[1]+(-0.748745750147)*x_ref[2]**o+(-0.168661287934)*x_ref[2]
        ref[3,4,2,0]=(0.515883010306)*x_ref[0]**o+(0.752629892309)*x_ref[0]+(-0.340202982107)*x_ref[1]**o+(-0.120916620105)*x_ref[1]+(0.193421000969)*x_ref[2]**o+(0.815950266515)*x_ref[2]
        ref[3,4,2,1]=(-0.397114672795)*x_ref[0]**o+(-0.737395536831)*x_ref[0]+(0.589874124917)*x_ref[1]**o+(-0.476304220998)*x_ref[1]+(-0.177825589992)*x_ref[2]**o+(-0.776118337595)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.458775377574)*x[0]**o+(-0.142783482982)*x[0]+(-0.254317031169)*x[1]**o+(0.702505689234)*x[1]
        ref=(0.458775377574)*x_ref[0]**o+(-0.142783482982)*x_ref[0]+(-0.254317031169)*x_ref[1]**o+(0.702505689234)*x_ref[1]
      else:
        s=(0.00533071834267)*x[0]**o+(0.986105059315)*x[0]+(-0.0825201956456)*x[1]**o+(-0.961255985926)*x[1]+(-0.336427126531)*x[2]**o+(0.0885714255159)*x[2]
        ref=(0.00533071834267)*x_ref[0]**o+(0.986105059315)*x_ref[0]+(-0.0825201956456)*x_ref[1]**o+(-0.961255985926)*x_ref[1]+(-0.336427126531)*x_ref[2]**o+(0.0885714255159)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.694815627598)*x[0]**o+(0.644063606633)*x[0]+(0.773087712154)*x[1]**o+(-0.962214094823)*x[1]
        s[1]=(-0.490711112677)*x[0]**o+(-0.87388065943)*x[0]+(-0.21628327344)*x[1]**o+(0.772088460047)*x[1]
        ref[0]=(-0.694815627598)*x_ref[0]**o+(0.644063606633)*x_ref[0]+(0.773087712154)*x_ref[1]**o+(-0.962214094823)*x_ref[1]
        ref[1]=(-0.490711112677)*x_ref[0]**o+(-0.87388065943)*x_ref[0]+(-0.21628327344)*x_ref[1]**o+(0.772088460047)*x_ref[1]
      else:
        s[0]=(-0.330435350709)*x[0]**o+(-0.359970727709)*x[0]+(0.642617648247)*x[1]**o+(-0.0394251455843)*x[1]+(-0.376622312269)*x[2]**o+(-0.541590231201)*x[2]
        s[1]=(0.463360592825)*x[0]**o+(0.498364920452)*x[0]+(0.755121144761)*x[1]**o+(0.743552227571)*x[1]+(0.168251767468)*x[2]**o+(0.584887196144)*x[2]
        ref[0]=(-0.330435350709)*x_ref[0]**o+(-0.359970727709)*x_ref[0]+(0.642617648247)*x_ref[1]**o+(-0.0394251455843)*x_ref[1]+(-0.376622312269)*x_ref[2]**o+(-0.541590231201)*x_ref[2]
        ref[1]=(0.463360592825)*x_ref[0]**o+(0.498364920452)*x_ref[0]+(0.755121144761)*x_ref[1]**o+(0.743552227571)*x_ref[1]+(0.168251767468)*x_ref[2]**o+(0.584887196144)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.488185951644)*x[0]**o+(0.597167519115)*x[0]+(0.943246675163)*x[1]**o+(-0.955863359088)*x[1]
        s[0,1]=(0.178292948477)*x[0]**o+(0.148038709562)*x[0]+(-0.950623527857)*x[1]**o+(0.411442651276)*x[1]
        s[0,2]=(0.134333500047)*x[0]**o+(0.6729047386)*x[0]+(-0.0897085832883)*x[1]**o+(0.372249915499)*x[1]
        s[0,3]=(-0.483653997546)*x[0]**o+(0.407897915806)*x[0]+(0.367153337432)*x[1]**o+(-0.698138025547)*x[1]
        s[0,4]=(-0.842134883349)*x[0]**o+(0.722923275602)*x[0]+(-0.74193204191)*x[1]**o+(-0.0669061392828)*x[1]
        s[1,0]=(0.264069997282)*x[0]**o+(-0.0754854815395)*x[0]+(-0.779935831011)*x[1]**o+(-0.114316965672)*x[1]
        s[1,1]=(0.179470045091)*x[0]**o+(-0.437165723748)*x[0]+(-0.294258032415)*x[1]**o+(0.899095305499)*x[1]
        s[1,2]=(-0.831740706096)*x[0]**o+(-0.149736068183)*x[0]+(-0.73474463786)*x[1]**o+(0.717231742118)*x[1]
        s[1,3]=(0.844816529175)*x[0]**o+(-0.0920760676122)*x[0]+(-0.2115275026)*x[1]**o+(-0.225889157104)*x[1]
        s[1,4]=(-0.0205843755762)*x[0]**o+(-0.718205564182)*x[0]+(-0.819513689268)*x[1]**o+(-0.0732333163823)*x[1]
        s[2,0]=(-0.85716147564)*x[0]**o+(-0.498404753235)*x[0]+(-0.326491535106)*x[1]**o+(0.0574234719803)*x[1]
        s[2,1]=(-0.952095737467)*x[0]**o+(0.473884826053)*x[0]+(-0.531204698326)*x[1]**o+(-0.174344061567)*x[1]
        s[2,2]=(0.999365057788)*x[0]**o+(-0.142639750172)*x[0]+(0.230480966664)*x[1]**o+(0.685914853943)*x[1]
        s[2,3]=(0.166345823031)*x[0]**o+(0.648257656524)*x[0]+(0.654487571777)*x[1]**o+(0.774745511568)*x[1]
        s[2,4]=(-0.0288937922147)*x[0]**o+(-0.979202627577)*x[0]+(0.228136074988)*x[1]**o+(0.738593896247)*x[1]
        s[3,0]=(0.774577798208)*x[0]**o+(0.874351290495)*x[0]+(0.490940628382)*x[1]**o+(-0.526685544309)*x[1]
        s[3,1]=(-0.708233931806)*x[0]**o+(-0.432420027826)*x[0]+(-0.883862641256)*x[1]**o+(0.291054377884)*x[1]
        s[3,2]=(0.359193932932)*x[0]**o+(0.527103835826)*x[0]+(0.072550559316)*x[1]**o+(0.818996368479)*x[1]
        s[3,3]=(-0.39082954898)*x[0]**o+(0.903749706745)*x[0]+(-0.0673256061393)*x[1]**o+(0.714848975884)*x[1]
        s[3,4]=(0.322408322579)*x[0]**o+(-0.429593260425)*x[0]+(0.982604999319)*x[1]**o+(-0.190161113783)*x[1]
        ref[0,0]=(-0.488185951644)*x_ref[0]**o+(0.597167519115)*x_ref[0]+(0.943246675163)*x_ref[1]**o+(-0.955863359088)*x_ref[1]
        ref[0,1]=(0.178292948477)*x_ref[0]**o+(0.148038709562)*x_ref[0]+(-0.950623527857)*x_ref[1]**o+(0.411442651276)*x_ref[1]
        ref[0,2]=(0.134333500047)*x_ref[0]**o+(0.6729047386)*x_ref[0]+(-0.0897085832883)*x_ref[1]**o+(0.372249915499)*x_ref[1]
        ref[0,3]=(-0.483653997546)*x_ref[0]**o+(0.407897915806)*x_ref[0]+(0.367153337432)*x_ref[1]**o+(-0.698138025547)*x_ref[1]
        ref[0,4]=(-0.842134883349)*x_ref[0]**o+(0.722923275602)*x_ref[0]+(-0.74193204191)*x_ref[1]**o+(-0.0669061392828)*x_ref[1]
        ref[1,0]=(0.264069997282)*x_ref[0]**o+(-0.0754854815395)*x_ref[0]+(-0.779935831011)*x_ref[1]**o+(-0.114316965672)*x_ref[1]
        ref[1,1]=(0.179470045091)*x_ref[0]**o+(-0.437165723748)*x_ref[0]+(-0.294258032415)*x_ref[1]**o+(0.899095305499)*x_ref[1]
        ref[1,2]=(-0.831740706096)*x_ref[0]**o+(-0.149736068183)*x_ref[0]+(-0.73474463786)*x_ref[1]**o+(0.717231742118)*x_ref[1]
        ref[1,3]=(0.844816529175)*x_ref[0]**o+(-0.0920760676122)*x_ref[0]+(-0.2115275026)*x_ref[1]**o+(-0.225889157104)*x_ref[1]
        ref[1,4]=(-0.0205843755762)*x_ref[0]**o+(-0.718205564182)*x_ref[0]+(-0.819513689268)*x_ref[1]**o+(-0.0732333163823)*x_ref[1]
        ref[2,0]=(-0.85716147564)*x_ref[0]**o+(-0.498404753235)*x_ref[0]+(-0.326491535106)*x_ref[1]**o+(0.0574234719803)*x_ref[1]
        ref[2,1]=(-0.952095737467)*x_ref[0]**o+(0.473884826053)*x_ref[0]+(-0.531204698326)*x_ref[1]**o+(-0.174344061567)*x_ref[1]
        ref[2,2]=(0.999365057788)*x_ref[0]**o+(-0.142639750172)*x_ref[0]+(0.230480966664)*x_ref[1]**o+(0.685914853943)*x_ref[1]
        ref[2,3]=(0.166345823031)*x_ref[0]**o+(0.648257656524)*x_ref[0]+(0.654487571777)*x_ref[1]**o+(0.774745511568)*x_ref[1]
        ref[2,4]=(-0.0288937922147)*x_ref[0]**o+(-0.979202627577)*x_ref[0]+(0.228136074988)*x_ref[1]**o+(0.738593896247)*x_ref[1]
        ref[3,0]=(0.774577798208)*x_ref[0]**o+(0.874351290495)*x_ref[0]+(0.490940628382)*x_ref[1]**o+(-0.526685544309)*x_ref[1]
        ref[3,1]=(-0.708233931806)*x_ref[0]**o+(-0.432420027826)*x_ref[0]+(-0.883862641256)*x_ref[1]**o+(0.291054377884)*x_ref[1]
        ref[3,2]=(0.359193932932)*x_ref[0]**o+(0.527103835826)*x_ref[0]+(0.072550559316)*x_ref[1]**o+(0.818996368479)*x_ref[1]
        ref[3,3]=(-0.39082954898)*x_ref[0]**o+(0.903749706745)*x_ref[0]+(-0.0673256061393)*x_ref[1]**o+(0.714848975884)*x_ref[1]
        ref[3,4]=(0.322408322579)*x_ref[0]**o+(-0.429593260425)*x_ref[0]+(0.982604999319)*x_ref[1]**o+(-0.190161113783)*x_ref[1]
      else:
        s[0,0]=(0.232804655799)*x[0]**o+(0.880141618164)*x[0]+(0.223221634005)*x[1]**o+(0.452983026652)*x[1]+(-0.573854985537)*x[2]**o+(-0.653809009284)*x[2]
        s[0,1]=(0.198858642702)*x[0]**o+(-0.314878524755)*x[0]+(-0.255927629298)*x[1]**o+(0.928004497503)*x[1]+(0.506963465945)*x[2]**o+(-0.47102323414)*x[2]
        s[0,2]=(0.308455355109)*x[0]**o+(-0.880701737691)*x[0]+(0.381469789283)*x[1]**o+(0.801918317713)*x[1]+(0.834098938062)*x[2]**o+(0.600601216943)*x[2]
        s[0,3]=(-0.469143012993)*x[0]**o+(-0.325961388166)*x[0]+(-0.0776077288313)*x[1]**o+(-0.0488948311893)*x[1]+(-0.0667465969735)*x[2]**o+(0.590630229618)*x[2]
        s[0,4]=(0.460884530892)*x[0]**o+(0.981078628528)*x[0]+(0.819191927543)*x[1]**o+(-0.225345065568)*x[1]+(-0.461177702761)*x[2]**o+(0.973442359563)*x[2]
        s[1,0]=(0.277068448313)*x[0]**o+(0.955364825669)*x[0]+(-0.912489837333)*x[1]**o+(-0.267268491011)*x[1]+(-0.847314081076)*x[2]**o+(-0.975760667649)*x[2]
        s[1,1]=(0.257880013047)*x[0]**o+(-0.343779864363)*x[0]+(0.616718460823)*x[1]**o+(-0.739452662888)*x[1]+(0.347386315181)*x[2]**o+(0.615160648765)*x[2]
        s[1,2]=(0.988114419561)*x[0]**o+(0.882801837855)*x[0]+(0.617420713454)*x[1]**o+(0.653402104375)*x[1]+(-0.0585299332344)*x[2]**o+(-0.400646895518)*x[2]
        s[1,3]=(-0.421049210036)*x[0]**o+(-0.0090620420347)*x[0]+(-0.54358224088)*x[1]**o+(0.254744898368)*x[1]+(-0.356411594187)*x[2]**o+(0.490942960051)*x[2]
        s[1,4]=(0.266613733548)*x[0]**o+(-0.953213213138)*x[0]+(0.659951033053)*x[1]**o+(0.0887943186214)*x[1]+(0.232221116015)*x[2]**o+(0.921710239665)*x[2]
        s[2,0]=(-0.278908005525)*x[0]**o+(-0.67161337084)*x[0]+(0.682065318715)*x[1]**o+(-0.184380015721)*x[1]+(0.72644685348)*x[2]**o+(0.951051205648)*x[2]
        s[2,1]=(-0.90781744456)*x[0]**o+(0.0077008898356)*x[0]+(0.671790725713)*x[1]**o+(0.853379797278)*x[1]+(-0.0764999700365)*x[2]**o+(-0.98659879948)*x[2]
        s[2,2]=(-0.913635850698)*x[0]**o+(-0.590635380841)*x[0]+(-0.663640142768)*x[1]**o+(-0.406156532072)*x[1]+(-0.784260406208)*x[2]**o+(-0.645167192291)*x[2]
        s[2,3]=(0.860741367256)*x[0]**o+(-0.0755200276519)*x[0]+(-0.501330034283)*x[1]**o+(0.774551262816)*x[1]+(-0.140777760645)*x[2]**o+(0.412398487198)*x[2]
        s[2,4]=(0.755189880126)*x[0]**o+(-0.688143708643)*x[0]+(0.497944488811)*x[1]**o+(-0.957053638137)*x[1]+(-0.116257980976)*x[2]**o+(0.356020915088)*x[2]
        s[3,0]=(-0.856430646919)*x[0]**o+(-0.0955546373651)*x[0]+(-0.588927939248)*x[1]**o+(0.207587246072)*x[1]+(-0.424773091145)*x[2]**o+(-0.0793349369172)*x[2]
        s[3,1]=(-0.0814274211195)*x[0]**o+(-0.103989324245)*x[0]+(-0.651528552136)*x[1]**o+(-0.757919391603)*x[1]+(0.142468149009)*x[2]**o+(0.642940283508)*x[2]
        s[3,2]=(0.266690596451)*x[0]**o+(-0.309503686303)*x[0]+(0.210541256761)*x[1]**o+(-0.363519101414)*x[1]+(-0.703156838318)*x[2]**o+(0.089872265658)*x[2]
        s[3,3]=(0.772726367524)*x[0]**o+(0.621843084065)*x[0]+(0.347047977396)*x[1]**o+(0.124103441993)*x[1]+(-0.626843590322)*x[2]**o+(0.215729189789)*x[2]
        s[3,4]=(-0.622360125477)*x[0]**o+(0.260262244242)*x[0]+(-0.312822631976)*x[1]**o+(-0.559674699162)*x[1]+(-0.0776492554888)*x[2]**o+(-0.156182340995)*x[2]
        ref[0,0]=(0.232804655799)*x_ref[0]**o+(0.880141618164)*x_ref[0]+(0.223221634005)*x_ref[1]**o+(0.452983026652)*x_ref[1]+(-0.573854985537)*x_ref[2]**o+(-0.653809009284)*x_ref[2]
        ref[0,1]=(0.198858642702)*x_ref[0]**o+(-0.314878524755)*x_ref[0]+(-0.255927629298)*x_ref[1]**o+(0.928004497503)*x_ref[1]+(0.506963465945)*x_ref[2]**o+(-0.47102323414)*x_ref[2]
        ref[0,2]=(0.308455355109)*x_ref[0]**o+(-0.880701737691)*x_ref[0]+(0.381469789283)*x_ref[1]**o+(0.801918317713)*x_ref[1]+(0.834098938062)*x_ref[2]**o+(0.600601216943)*x_ref[2]
        ref[0,3]=(-0.469143012993)*x_ref[0]**o+(-0.325961388166)*x_ref[0]+(-0.0776077288313)*x_ref[1]**o+(-0.0488948311893)*x_ref[1]+(-0.0667465969735)*x_ref[2]**o+(0.590630229618)*x_ref[2]
        ref[0,4]=(0.460884530892)*x_ref[0]**o+(0.981078628528)*x_ref[0]+(0.819191927543)*x_ref[1]**o+(-0.225345065568)*x_ref[1]+(-0.461177702761)*x_ref[2]**o+(0.973442359563)*x_ref[2]
        ref[1,0]=(0.277068448313)*x_ref[0]**o+(0.955364825669)*x_ref[0]+(-0.912489837333)*x_ref[1]**o+(-0.267268491011)*x_ref[1]+(-0.847314081076)*x_ref[2]**o+(-0.975760667649)*x_ref[2]
        ref[1,1]=(0.257880013047)*x_ref[0]**o+(-0.343779864363)*x_ref[0]+(0.616718460823)*x_ref[1]**o+(-0.739452662888)*x_ref[1]+(0.347386315181)*x_ref[2]**o+(0.615160648765)*x_ref[2]
        ref[1,2]=(0.988114419561)*x_ref[0]**o+(0.882801837855)*x_ref[0]+(0.617420713454)*x_ref[1]**o+(0.653402104375)*x_ref[1]+(-0.0585299332344)*x_ref[2]**o+(-0.400646895518)*x_ref[2]
        ref[1,3]=(-0.421049210036)*x_ref[0]**o+(-0.0090620420347)*x_ref[0]+(-0.54358224088)*x_ref[1]**o+(0.254744898368)*x_ref[1]+(-0.356411594187)*x_ref[2]**o+(0.490942960051)*x_ref[2]
        ref[1,4]=(0.266613733548)*x_ref[0]**o+(-0.953213213138)*x_ref[0]+(0.659951033053)*x_ref[1]**o+(0.0887943186214)*x_ref[1]+(0.232221116015)*x_ref[2]**o+(0.921710239665)*x_ref[2]
        ref[2,0]=(-0.278908005525)*x_ref[0]**o+(-0.67161337084)*x_ref[0]+(0.682065318715)*x_ref[1]**o+(-0.184380015721)*x_ref[1]+(0.72644685348)*x_ref[2]**o+(0.951051205648)*x_ref[2]
        ref[2,1]=(-0.90781744456)*x_ref[0]**o+(0.0077008898356)*x_ref[0]+(0.671790725713)*x_ref[1]**o+(0.853379797278)*x_ref[1]+(-0.0764999700365)*x_ref[2]**o+(-0.98659879948)*x_ref[2]
        ref[2,2]=(-0.913635850698)*x_ref[0]**o+(-0.590635380841)*x_ref[0]+(-0.663640142768)*x_ref[1]**o+(-0.406156532072)*x_ref[1]+(-0.784260406208)*x_ref[2]**o+(-0.645167192291)*x_ref[2]
        ref[2,3]=(0.860741367256)*x_ref[0]**o+(-0.0755200276519)*x_ref[0]+(-0.501330034283)*x_ref[1]**o+(0.774551262816)*x_ref[1]+(-0.140777760645)*x_ref[2]**o+(0.412398487198)*x_ref[2]
        ref[2,4]=(0.755189880126)*x_ref[0]**o+(-0.688143708643)*x_ref[0]+(0.497944488811)*x_ref[1]**o+(-0.957053638137)*x_ref[1]+(-0.116257980976)*x_ref[2]**o+(0.356020915088)*x_ref[2]
        ref[3,0]=(-0.856430646919)*x_ref[0]**o+(-0.0955546373651)*x_ref[0]+(-0.588927939248)*x_ref[1]**o+(0.207587246072)*x_ref[1]+(-0.424773091145)*x_ref[2]**o+(-0.0793349369172)*x_ref[2]
        ref[3,1]=(-0.0814274211195)*x_ref[0]**o+(-0.103989324245)*x_ref[0]+(-0.651528552136)*x_ref[1]**o+(-0.757919391603)*x_ref[1]+(0.142468149009)*x_ref[2]**o+(0.642940283508)*x_ref[2]
        ref[3,2]=(0.266690596451)*x_ref[0]**o+(-0.309503686303)*x_ref[0]+(0.210541256761)*x_ref[1]**o+(-0.363519101414)*x_ref[1]+(-0.703156838318)*x_ref[2]**o+(0.089872265658)*x_ref[2]
        ref[3,3]=(0.772726367524)*x_ref[0]**o+(0.621843084065)*x_ref[0]+(0.347047977396)*x_ref[1]**o+(0.124103441993)*x_ref[1]+(-0.626843590322)*x_ref[2]**o+(0.215729189789)*x_ref[2]
        ref[3,4]=(-0.622360125477)*x_ref[0]**o+(0.260262244242)*x_ref[0]+(-0.312822631976)*x_ref[1]**o+(-0.559674699162)*x_ref[1]+(-0.0776492554888)*x_ref[2]**o+(-0.156182340995)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.87064070758)*x[0]**o+(-0.566055900642)*x[0]+(-0.917420111598)*x[1]**o+(0.0551635734945)*x[1]
        s[0,0,1]=(-0.117176717885)*x[0]**o+(-0.822793678972)*x[0]+(-0.41376673807)*x[1]**o+(-0.782915075242)*x[1]
        s[0,1,0]=(0.309660501821)*x[0]**o+(-0.705257783376)*x[0]+(0.436897623093)*x[1]**o+(0.624205776859)*x[1]
        s[0,1,1]=(-0.338652342709)*x[0]**o+(-0.783666910448)*x[0]+(0.869895700694)*x[1]**o+(-0.106196197398)*x[1]
        s[1,0,0]=(-0.418615528975)*x[0]**o+(-0.821973439514)*x[0]+(0.715494828047)*x[1]**o+(0.717827721559)*x[1]
        s[1,0,1]=(-0.0319361959791)*x[0]**o+(0.60209433014)*x[0]+(-0.297989638002)*x[1]**o+(0.133472987028)*x[1]
        s[1,1,0]=(0.874169575419)*x[0]**o+(-0.804771710143)*x[0]+(-0.956980034287)*x[1]**o+(0.60848322694)*x[1]
        s[1,1,1]=(-0.96607032622)*x[0]**o+(0.58719800455)*x[0]+(0.432119827191)*x[1]**o+(-0.0515514190014)*x[1]
        s[2,0,0]=(0.92256117661)*x[0]**o+(0.286836406901)*x[0]+(-0.566218833636)*x[1]**o+(-0.318500128005)*x[1]
        s[2,0,1]=(-0.295696171332)*x[0]**o+(-0.847067252777)*x[0]+(-0.0214499844044)*x[1]**o+(0.56142047851)*x[1]
        s[2,1,0]=(-0.278598238494)*x[0]**o+(0.775672819752)*x[0]+(-0.594303030563)*x[1]**o+(-0.764128845442)*x[1]
        s[2,1,1]=(0.195219876382)*x[0]**o+(0.31911545942)*x[0]+(0.414261884794)*x[1]**o+(0.430367930637)*x[1]
        s[3,0,0]=(-0.957282546866)*x[0]**o+(-0.227277741511)*x[0]+(0.386584028319)*x[1]**o+(0.166263307275)*x[1]
        s[3,0,1]=(-0.682200893175)*x[0]**o+(-0.845531035272)*x[0]+(0.938083890385)*x[1]**o+(-0.674378235835)*x[1]
        s[3,1,0]=(0.544986524999)*x[0]**o+(0.529710178839)*x[0]+(-0.489344340247)*x[1]**o+(0.476288384952)*x[1]
        s[3,1,1]=(-0.680566273064)*x[0]**o+(0.806369975569)*x[0]+(0.305516322897)*x[1]**o+(-0.51330834934)*x[1]
        s[4,0,0]=(-0.465052752387)*x[0]**o+(-0.00313229607036)*x[0]+(0.499351385096)*x[1]**o+(-0.401491711188)*x[1]
        s[4,0,1]=(0.27236428546)*x[0]**o+(-0.278717867273)*x[0]+(-0.96726186532)*x[1]**o+(-0.317275683828)*x[1]
        s[4,1,0]=(0.477975953682)*x[0]**o+(0.772356007674)*x[0]+(0.285581555335)*x[1]**o+(0.872551473549)*x[1]
        s[4,1,1]=(-0.181196107871)*x[0]**o+(-0.473439467597)*x[0]+(0.90210175622)*x[1]**o+(-0.378511577829)*x[1]
        s[5,0,0]=(0.633625574073)*x[0]**o+(-0.563894010102)*x[0]+(-0.302946833681)*x[1]**o+(-0.601584229143)*x[1]
        s[5,0,1]=(-0.754395228954)*x[0]**o+(0.317899599612)*x[0]+(-0.612069470764)*x[1]**o+(0.937528317073)*x[1]
        s[5,1,0]=(0.0360449364159)*x[0]**o+(0.0944318504395)*x[0]+(0.745586397044)*x[1]**o+(-0.839531808187)*x[1]
        s[5,1,1]=(0.620453262074)*x[0]**o+(-0.199107011047)*x[0]+(0.754391775519)*x[1]**o+(0.853229319102)*x[1]
        ref[0,0,0]=(0.87064070758)*x_ref[0]**o+(-0.566055900642)*x_ref[0]+(-0.917420111598)*x_ref[1]**o+(0.0551635734945)*x_ref[1]
        ref[0,0,1]=(-0.117176717885)*x_ref[0]**o+(-0.822793678972)*x_ref[0]+(-0.41376673807)*x_ref[1]**o+(-0.782915075242)*x_ref[1]
        ref[0,1,0]=(0.309660501821)*x_ref[0]**o+(-0.705257783376)*x_ref[0]+(0.436897623093)*x_ref[1]**o+(0.624205776859)*x_ref[1]
        ref[0,1,1]=(-0.338652342709)*x_ref[0]**o+(-0.783666910448)*x_ref[0]+(0.869895700694)*x_ref[1]**o+(-0.106196197398)*x_ref[1]
        ref[1,0,0]=(-0.418615528975)*x_ref[0]**o+(-0.821973439514)*x_ref[0]+(0.715494828047)*x_ref[1]**o+(0.717827721559)*x_ref[1]
        ref[1,0,1]=(-0.0319361959791)*x_ref[0]**o+(0.60209433014)*x_ref[0]+(-0.297989638002)*x_ref[1]**o+(0.133472987028)*x_ref[1]
        ref[1,1,0]=(0.874169575419)*x_ref[0]**o+(-0.804771710143)*x_ref[0]+(-0.956980034287)*x_ref[1]**o+(0.60848322694)*x_ref[1]
        ref[1,1,1]=(-0.96607032622)*x_ref[0]**o+(0.58719800455)*x_ref[0]+(0.432119827191)*x_ref[1]**o+(-0.0515514190014)*x_ref[1]
        ref[2,0,0]=(0.92256117661)*x_ref[0]**o+(0.286836406901)*x_ref[0]+(-0.566218833636)*x_ref[1]**o+(-0.318500128005)*x_ref[1]
        ref[2,0,1]=(-0.295696171332)*x_ref[0]**o+(-0.847067252777)*x_ref[0]+(-0.0214499844044)*x_ref[1]**o+(0.56142047851)*x_ref[1]
        ref[2,1,0]=(-0.278598238494)*x_ref[0]**o+(0.775672819752)*x_ref[0]+(-0.594303030563)*x_ref[1]**o+(-0.764128845442)*x_ref[1]
        ref[2,1,1]=(0.195219876382)*x_ref[0]**o+(0.31911545942)*x_ref[0]+(0.414261884794)*x_ref[1]**o+(0.430367930637)*x_ref[1]
        ref[3,0,0]=(-0.957282546866)*x_ref[0]**o+(-0.227277741511)*x_ref[0]+(0.386584028319)*x_ref[1]**o+(0.166263307275)*x_ref[1]
        ref[3,0,1]=(-0.682200893175)*x_ref[0]**o+(-0.845531035272)*x_ref[0]+(0.938083890385)*x_ref[1]**o+(-0.674378235835)*x_ref[1]
        ref[3,1,0]=(0.544986524999)*x_ref[0]**o+(0.529710178839)*x_ref[0]+(-0.489344340247)*x_ref[1]**o+(0.476288384952)*x_ref[1]
        ref[3,1,1]=(-0.680566273064)*x_ref[0]**o+(0.806369975569)*x_ref[0]+(0.305516322897)*x_ref[1]**o+(-0.51330834934)*x_ref[1]
        ref[4,0,0]=(-0.465052752387)*x_ref[0]**o+(-0.00313229607036)*x_ref[0]+(0.499351385096)*x_ref[1]**o+(-0.401491711188)*x_ref[1]
        ref[4,0,1]=(0.27236428546)*x_ref[0]**o+(-0.278717867273)*x_ref[0]+(-0.96726186532)*x_ref[1]**o+(-0.317275683828)*x_ref[1]
        ref[4,1,0]=(0.477975953682)*x_ref[0]**o+(0.772356007674)*x_ref[0]+(0.285581555335)*x_ref[1]**o+(0.872551473549)*x_ref[1]
        ref[4,1,1]=(-0.181196107871)*x_ref[0]**o+(-0.473439467597)*x_ref[0]+(0.90210175622)*x_ref[1]**o+(-0.378511577829)*x_ref[1]
        ref[5,0,0]=(0.633625574073)*x_ref[0]**o+(-0.563894010102)*x_ref[0]+(-0.302946833681)*x_ref[1]**o+(-0.601584229143)*x_ref[1]
        ref[5,0,1]=(-0.754395228954)*x_ref[0]**o+(0.317899599612)*x_ref[0]+(-0.612069470764)*x_ref[1]**o+(0.937528317073)*x_ref[1]
        ref[5,1,0]=(0.0360449364159)*x_ref[0]**o+(0.0944318504395)*x_ref[0]+(0.745586397044)*x_ref[1]**o+(-0.839531808187)*x_ref[1]
        ref[5,1,1]=(0.620453262074)*x_ref[0]**o+(-0.199107011047)*x_ref[0]+(0.754391775519)*x_ref[1]**o+(0.853229319102)*x_ref[1]
      else:
        s[0,0,0]=(-0.384948393134)*x[0]**o+(-0.961967178877)*x[0]+(0.801941664706)*x[1]**o+(0.341078763066)*x[1]+(-0.965751483242)*x[2]**o+(-0.335041713825)*x[2]
        s[0,0,1]=(-0.67569110343)*x[0]**o+(-0.607731651311)*x[0]+(-0.769155006252)*x[1]**o+(-0.951474117563)*x[1]+(0.346757382654)*x[2]**o+(0.468605905517)*x[2]
        s[0,1,0]=(0.357827714378)*x[0]**o+(0.531884637396)*x[0]+(-0.340932021448)*x[1]**o+(0.107907346017)*x[1]+(0.899991044816)*x[2]**o+(-0.619238986401)*x[2]
        s[0,1,1]=(0.92143345512)*x[0]**o+(0.0319943191208)*x[0]+(0.143931354465)*x[1]**o+(-0.827543395412)*x[1]+(0.670291174904)*x[2]**o+(0.323131021196)*x[2]
        s[1,0,0]=(0.253706012955)*x[0]**o+(0.920629791188)*x[0]+(-0.221548307026)*x[1]**o+(0.920275162623)*x[1]+(0.159384070106)*x[2]**o+(0.914780524212)*x[2]
        s[1,0,1]=(-0.145041193856)*x[0]**o+(-0.777916320748)*x[0]+(-0.270423261412)*x[1]**o+(0.948002693835)*x[1]+(-0.872437269572)*x[2]**o+(-0.389969462287)*x[2]
        s[1,1,0]=(-0.0258988000508)*x[0]**o+(0.27202895614)*x[0]+(-0.979886715231)*x[1]**o+(-0.102396555104)*x[1]+(-0.97303335804)*x[2]**o+(0.0848739951879)*x[2]
        s[1,1,1]=(-0.171608662455)*x[0]**o+(-0.0499287952471)*x[0]+(-0.780462044785)*x[1]**o+(0.506400671006)*x[1]+(-0.678439420533)*x[2]**o+(-0.609358013123)*x[2]
        s[2,0,0]=(0.715810623102)*x[0]**o+(0.478297229523)*x[0]+(-0.700770389598)*x[1]**o+(0.279096508315)*x[1]+(-0.385130762383)*x[2]**o+(0.46599268203)*x[2]
        s[2,0,1]=(0.47346229532)*x[0]**o+(0.304487177002)*x[0]+(0.0242783060008)*x[1]**o+(0.772667749027)*x[1]+(-0.42037502877)*x[2]**o+(-0.912335922325)*x[2]
        s[2,1,0]=(0.509168300969)*x[0]**o+(-0.746969033155)*x[0]+(-0.907693621932)*x[1]**o+(-0.642615676429)*x[1]+(-0.439912454318)*x[2]**o+(-0.452216863742)*x[2]
        s[2,1,1]=(-0.410444089831)*x[0]**o+(-0.147372964881)*x[0]+(0.850150090851)*x[1]**o+(-0.988475742361)*x[1]+(-0.462654428176)*x[2]**o+(-0.940338508948)*x[2]
        s[3,0,0]=(-0.794220586578)*x[0]**o+(0.922257088538)*x[0]+(0.0134033910944)*x[1]**o+(-0.95038627404)*x[1]+(-0.609227492432)*x[2]**o+(0.558321765879)*x[2]
        s[3,0,1]=(-0.304420694648)*x[0]**o+(0.244625817097)*x[0]+(-0.0933192976308)*x[1]**o+(-0.874643761558)*x[1]+(0.936742259878)*x[2]**o+(0.895320681558)*x[2]
        s[3,1,0]=(-0.202315611337)*x[0]**o+(-0.2365832224)*x[0]+(-0.480962593573)*x[1]**o+(-0.210912260068)*x[1]+(-0.000213415776588)*x[2]**o+(0.259609090936)*x[2]
        s[3,1,1]=(-0.612351015469)*x[0]**o+(0.546469916272)*x[0]+(-0.166009868547)*x[1]**o+(0.931501987972)*x[1]+(-0.330366993065)*x[2]**o+(0.435322489369)*x[2]
        s[4,0,0]=(0.124588792911)*x[0]**o+(0.0472809655836)*x[0]+(0.126649125977)*x[1]**o+(-0.980286766046)*x[1]+(-0.708140145708)*x[2]**o+(0.105934454391)*x[2]
        s[4,0,1]=(-0.967814569406)*x[0]**o+(-0.606176160393)*x[0]+(0.0252434349925)*x[1]**o+(0.606791676748)*x[1]+(-0.615219396992)*x[2]**o+(0.462327857208)*x[2]
        s[4,1,0]=(0.47926323984)*x[0]**o+(0.493300730667)*x[0]+(0.940191066665)*x[1]**o+(0.245800539336)*x[1]+(-0.694101640383)*x[2]**o+(0.200348996082)*x[2]
        s[4,1,1]=(0.552067136819)*x[0]**o+(0.351346598128)*x[0]+(0.39731191883)*x[1]**o+(-0.70236991903)*x[1]+(-0.625528596695)*x[2]**o+(0.798534719203)*x[2]
        s[5,0,0]=(0.167335359474)*x[0]**o+(-0.232843920341)*x[0]+(-0.126773325792)*x[1]**o+(-0.203898990499)*x[1]+(0.839140245989)*x[2]**o+(-0.728050229083)*x[2]
        s[5,0,1]=(-0.664841432025)*x[0]**o+(-0.0489116955008)*x[0]+(0.685379040894)*x[1]**o+(0.370521174212)*x[1]+(-0.742375864799)*x[2]**o+(-0.880322708097)*x[2]
        s[5,1,0]=(0.653110528414)*x[0]**o+(0.602025573916)*x[0]+(0.912361317481)*x[1]**o+(0.534616035231)*x[1]+(-0.103102947904)*x[2]**o+(0.531938071791)*x[2]
        s[5,1,1]=(-0.217856022054)*x[0]**o+(0.11102291479)*x[0]+(-0.166664940768)*x[1]**o+(-0.812763150681)*x[1]+(0.442963442721)*x[2]**o+(0.706518891053)*x[2]
        ref[0,0,0]=(-0.384948393134)*x_ref[0]**o+(-0.961967178877)*x_ref[0]+(0.801941664706)*x_ref[1]**o+(0.341078763066)*x_ref[1]+(-0.965751483242)*x_ref[2]**o+(-0.335041713825)*x_ref[2]
        ref[0,0,1]=(-0.67569110343)*x_ref[0]**o+(-0.607731651311)*x_ref[0]+(-0.769155006252)*x_ref[1]**o+(-0.951474117563)*x_ref[1]+(0.346757382654)*x_ref[2]**o+(0.468605905517)*x_ref[2]
        ref[0,1,0]=(0.357827714378)*x_ref[0]**o+(0.531884637396)*x_ref[0]+(-0.340932021448)*x_ref[1]**o+(0.107907346017)*x_ref[1]+(0.899991044816)*x_ref[2]**o+(-0.619238986401)*x_ref[2]
        ref[0,1,1]=(0.92143345512)*x_ref[0]**o+(0.0319943191208)*x_ref[0]+(0.143931354465)*x_ref[1]**o+(-0.827543395412)*x_ref[1]+(0.670291174904)*x_ref[2]**o+(0.323131021196)*x_ref[2]
        ref[1,0,0]=(0.253706012955)*x_ref[0]**o+(0.920629791188)*x_ref[0]+(-0.221548307026)*x_ref[1]**o+(0.920275162623)*x_ref[1]+(0.159384070106)*x_ref[2]**o+(0.914780524212)*x_ref[2]
        ref[1,0,1]=(-0.145041193856)*x_ref[0]**o+(-0.777916320748)*x_ref[0]+(-0.270423261412)*x_ref[1]**o+(0.948002693835)*x_ref[1]+(-0.872437269572)*x_ref[2]**o+(-0.389969462287)*x_ref[2]
        ref[1,1,0]=(-0.0258988000508)*x_ref[0]**o+(0.27202895614)*x_ref[0]+(-0.979886715231)*x_ref[1]**o+(-0.102396555104)*x_ref[1]+(-0.97303335804)*x_ref[2]**o+(0.0848739951879)*x_ref[2]
        ref[1,1,1]=(-0.171608662455)*x_ref[0]**o+(-0.0499287952471)*x_ref[0]+(-0.780462044785)*x_ref[1]**o+(0.506400671006)*x_ref[1]+(-0.678439420533)*x_ref[2]**o+(-0.609358013123)*x_ref[2]
        ref[2,0,0]=(0.715810623102)*x_ref[0]**o+(0.478297229523)*x_ref[0]+(-0.700770389598)*x_ref[1]**o+(0.279096508315)*x_ref[1]+(-0.385130762383)*x_ref[2]**o+(0.46599268203)*x_ref[2]
        ref[2,0,1]=(0.47346229532)*x_ref[0]**o+(0.304487177002)*x_ref[0]+(0.0242783060008)*x_ref[1]**o+(0.772667749027)*x_ref[1]+(-0.42037502877)*x_ref[2]**o+(-0.912335922325)*x_ref[2]
        ref[2,1,0]=(0.509168300969)*x_ref[0]**o+(-0.746969033155)*x_ref[0]+(-0.907693621932)*x_ref[1]**o+(-0.642615676429)*x_ref[1]+(-0.439912454318)*x_ref[2]**o+(-0.452216863742)*x_ref[2]
        ref[2,1,1]=(-0.410444089831)*x_ref[0]**o+(-0.147372964881)*x_ref[0]+(0.850150090851)*x_ref[1]**o+(-0.988475742361)*x_ref[1]+(-0.462654428176)*x_ref[2]**o+(-0.940338508948)*x_ref[2]
        ref[3,0,0]=(-0.794220586578)*x_ref[0]**o+(0.922257088538)*x_ref[0]+(0.0134033910944)*x_ref[1]**o+(-0.95038627404)*x_ref[1]+(-0.609227492432)*x_ref[2]**o+(0.558321765879)*x_ref[2]
        ref[3,0,1]=(-0.304420694648)*x_ref[0]**o+(0.244625817097)*x_ref[0]+(-0.0933192976308)*x_ref[1]**o+(-0.874643761558)*x_ref[1]+(0.936742259878)*x_ref[2]**o+(0.895320681558)*x_ref[2]
        ref[3,1,0]=(-0.202315611337)*x_ref[0]**o+(-0.2365832224)*x_ref[0]+(-0.480962593573)*x_ref[1]**o+(-0.210912260068)*x_ref[1]+(-0.000213415776588)*x_ref[2]**o+(0.259609090936)*x_ref[2]
        ref[3,1,1]=(-0.612351015469)*x_ref[0]**o+(0.546469916272)*x_ref[0]+(-0.166009868547)*x_ref[1]**o+(0.931501987972)*x_ref[1]+(-0.330366993065)*x_ref[2]**o+(0.435322489369)*x_ref[2]
        ref[4,0,0]=(0.124588792911)*x_ref[0]**o+(0.0472809655836)*x_ref[0]+(0.126649125977)*x_ref[1]**o+(-0.980286766046)*x_ref[1]+(-0.708140145708)*x_ref[2]**o+(0.105934454391)*x_ref[2]
        ref[4,0,1]=(-0.967814569406)*x_ref[0]**o+(-0.606176160393)*x_ref[0]+(0.0252434349925)*x_ref[1]**o+(0.606791676748)*x_ref[1]+(-0.615219396992)*x_ref[2]**o+(0.462327857208)*x_ref[2]
        ref[4,1,0]=(0.47926323984)*x_ref[0]**o+(0.493300730667)*x_ref[0]+(0.940191066665)*x_ref[1]**o+(0.245800539336)*x_ref[1]+(-0.694101640383)*x_ref[2]**o+(0.200348996082)*x_ref[2]
        ref[4,1,1]=(0.552067136819)*x_ref[0]**o+(0.351346598128)*x_ref[0]+(0.39731191883)*x_ref[1]**o+(-0.70236991903)*x_ref[1]+(-0.625528596695)*x_ref[2]**o+(0.798534719203)*x_ref[2]
        ref[5,0,0]=(0.167335359474)*x_ref[0]**o+(-0.232843920341)*x_ref[0]+(-0.126773325792)*x_ref[1]**o+(-0.203898990499)*x_ref[1]+(0.839140245989)*x_ref[2]**o+(-0.728050229083)*x_ref[2]
        ref[5,0,1]=(-0.664841432025)*x_ref[0]**o+(-0.0489116955008)*x_ref[0]+(0.685379040894)*x_ref[1]**o+(0.370521174212)*x_ref[1]+(-0.742375864799)*x_ref[2]**o+(-0.880322708097)*x_ref[2]
        ref[5,1,0]=(0.653110528414)*x_ref[0]**o+(0.602025573916)*x_ref[0]+(0.912361317481)*x_ref[1]**o+(0.534616035231)*x_ref[1]+(-0.103102947904)*x_ref[2]**o+(0.531938071791)*x_ref[2]
        ref[5,1,1]=(-0.217856022054)*x_ref[0]**o+(0.11102291479)*x_ref[0]+(-0.166664940768)*x_ref[1]**o+(-0.812763150681)*x_ref[1]+(0.442963442721)*x_ref[2]**o+(0.706518891053)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.209596991985)*x[0]**o+(0.997116487512)*x[0]+(0.0205082690151)*x[1]**o+(0.781596313955)*x[1]
        s[0,0,0,1]=(-0.475035630436)*x[0]**o+(0.476939038583)*x[0]+(-0.403224734103)*x[1]**o+(-0.726327526537)*x[1]
        s[0,0,1,0]=(0.165692569014)*x[0]**o+(-0.958953962148)*x[0]+(0.778520844819)*x[1]**o+(-0.371211501917)*x[1]
        s[0,0,1,1]=(-0.0863741752107)*x[0]**o+(0.64735697812)*x[0]+(-0.717371101374)*x[1]**o+(-0.517476214609)*x[1]
        s[0,0,2,0]=(0.414678358278)*x[0]**o+(0.766044973671)*x[0]+(0.473600649794)*x[1]**o+(-0.350863429165)*x[1]
        s[0,0,2,1]=(0.0391148157551)*x[0]**o+(0.550258520615)*x[0]+(-0.0952445590164)*x[1]**o+(0.799829721692)*x[1]
        s[0,1,0,0]=(-0.910760056281)*x[0]**o+(0.319346354593)*x[0]+(0.640854539851)*x[1]**o+(0.0622173716277)*x[1]
        s[0,1,0,1]=(0.761672466091)*x[0]**o+(0.266521259527)*x[0]+(-0.567155906289)*x[1]**o+(0.124135553341)*x[1]
        s[0,1,1,0]=(-0.681712184552)*x[0]**o+(0.388484728164)*x[0]+(-0.412243885661)*x[1]**o+(0.637954103055)*x[1]
        s[0,1,1,1]=(-0.485017958909)*x[0]**o+(0.736606905633)*x[0]+(-0.299204062504)*x[1]**o+(0.0840021848608)*x[1]
        s[0,1,2,0]=(0.955598154983)*x[0]**o+(-0.0204689463318)*x[0]+(-0.737123328202)*x[1]**o+(0.708418145874)*x[1]
        s[0,1,2,1]=(0.541565827885)*x[0]**o+(0.97493788494)*x[0]+(-0.450344230708)*x[1]**o+(-0.0231871887964)*x[1]
        s[0,2,0,0]=(-0.338442282639)*x[0]**o+(0.0746189409911)*x[0]+(0.76030098768)*x[1]**o+(0.629119828718)*x[1]
        s[0,2,0,1]=(-0.900573785768)*x[0]**o+(-0.971041944198)*x[0]+(0.637105978641)*x[1]**o+(0.387639692227)*x[1]
        s[0,2,1,0]=(0.0889093715568)*x[0]**o+(0.260479957021)*x[0]+(-0.908420084369)*x[1]**o+(0.6517961193)*x[1]
        s[0,2,1,1]=(0.683959794134)*x[0]**o+(-0.84665688176)*x[0]+(0.710853815216)*x[1]**o+(0.536738240986)*x[1]
        s[0,2,2,0]=(0.365754968462)*x[0]**o+(0.994576642301)*x[0]+(0.110702356612)*x[1]**o+(0.68992003076)*x[1]
        s[0,2,2,1]=(-0.238463519384)*x[0]**o+(-0.91372863405)*x[0]+(-0.334884010749)*x[1]**o+(0.481631381739)*x[1]
        s[0,3,0,0]=(-0.580344649945)*x[0]**o+(0.557953798016)*x[0]+(0.835117346932)*x[1]**o+(-0.196440751777)*x[1]
        s[0,3,0,1]=(0.339197786593)*x[0]**o+(0.231910549016)*x[0]+(0.515605654416)*x[1]**o+(-0.546885567308)*x[1]
        s[0,3,1,0]=(-0.562198188719)*x[0]**o+(0.410791865348)*x[0]+(-0.0823054523407)*x[1]**o+(0.789832279038)*x[1]
        s[0,3,1,1]=(-0.519245133296)*x[0]**o+(0.622234324355)*x[0]+(-0.89225914065)*x[1]**o+(-0.0376143576483)*x[1]
        s[0,3,2,0]=(-0.884895063274)*x[0]**o+(-0.569137436872)*x[0]+(0.107652574639)*x[1]**o+(0.936768318078)*x[1]
        s[0,3,2,1]=(-0.209493065726)*x[0]**o+(0.830161297986)*x[0]+(-0.811634331709)*x[1]**o+(0.0673874073431)*x[1]
        s[0,4,0,0]=(-0.518639168744)*x[0]**o+(-0.953627666411)*x[0]+(0.00153430886561)*x[1]**o+(0.118432410415)*x[1]
        s[0,4,0,1]=(0.901325505931)*x[0]**o+(0.276592227771)*x[0]+(0.906505574302)*x[1]**o+(-0.604032185175)*x[1]
        s[0,4,1,0]=(0.539336544226)*x[0]**o+(-0.518177781519)*x[0]+(-0.397453923803)*x[1]**o+(0.0661677991096)*x[1]
        s[0,4,1,1]=(0.31900579667)*x[0]**o+(-0.950735779053)*x[0]+(0.495414320287)*x[1]**o+(0.310834200079)*x[1]
        s[0,4,2,0]=(0.567277111907)*x[0]**o+(0.124915250004)*x[0]+(0.699531698929)*x[1]**o+(-0.434718433784)*x[1]
        s[0,4,2,1]=(0.607190906026)*x[0]**o+(0.411045834019)*x[0]+(-0.116848105246)*x[1]**o+(-0.343872842277)*x[1]
        s[1,0,0,0]=(0.263188443069)*x[0]**o+(0.782586354151)*x[0]+(-0.0593809262828)*x[1]**o+(-0.516894145417)*x[1]
        s[1,0,0,1]=(-0.155136250976)*x[0]**o+(0.155174971237)*x[0]+(0.452174963131)*x[1]**o+(-0.811875854704)*x[1]
        s[1,0,1,0]=(-0.161366848419)*x[0]**o+(-0.674425732875)*x[0]+(-0.99246147148)*x[1]**o+(0.618506213959)*x[1]
        s[1,0,1,1]=(0.64592094187)*x[0]**o+(0.444912658041)*x[0]+(0.963730302998)*x[1]**o+(-0.227756330634)*x[1]
        s[1,0,2,0]=(0.199217092911)*x[0]**o+(-0.118577255523)*x[0]+(-0.560600457675)*x[1]**o+(-0.233703068897)*x[1]
        s[1,0,2,1]=(0.200710250009)*x[0]**o+(0.366088123353)*x[0]+(0.419249984559)*x[1]**o+(-0.0188044094656)*x[1]
        s[1,1,0,0]=(-0.990545888549)*x[0]**o+(-0.100521244971)*x[0]+(-0.0134093118672)*x[1]**o+(0.582791526008)*x[1]
        s[1,1,0,1]=(-0.932085864166)*x[0]**o+(-0.402198623203)*x[0]+(-0.769996422181)*x[1]**o+(0.701839645425)*x[1]
        s[1,1,1,0]=(-0.0546050562235)*x[0]**o+(-0.821680588129)*x[0]+(0.254170421382)*x[1]**o+(0.0419062526906)*x[1]
        s[1,1,1,1]=(0.0361631230129)*x[0]**o+(0.155809438009)*x[0]+(-0.797631293341)*x[1]**o+(-0.0764161686239)*x[1]
        s[1,1,2,0]=(-0.679858025056)*x[0]**o+(-0.0411819071479)*x[0]+(0.033731324617)*x[1]**o+(0.32496792339)*x[1]
        s[1,1,2,1]=(0.424198647091)*x[0]**o+(-0.332229088063)*x[0]+(-0.225491160735)*x[1]**o+(-0.833706567263)*x[1]
        s[1,2,0,0]=(0.701510843723)*x[0]**o+(0.780699569051)*x[0]+(0.369319568113)*x[1]**o+(-0.865220461023)*x[1]
        s[1,2,0,1]=(-0.388700890661)*x[0]**o+(-0.550094979352)*x[0]+(0.585925695487)*x[1]**o+(0.812085187015)*x[1]
        s[1,2,1,0]=(-0.825827041568)*x[0]**o+(-0.281570175483)*x[0]+(-0.352446259467)*x[1]**o+(-0.162431177857)*x[1]
        s[1,2,1,1]=(-0.615917011781)*x[0]**o+(-0.218604155424)*x[0]+(-0.66718327064)*x[1]**o+(-0.723246262472)*x[1]
        s[1,2,2,0]=(-0.794217383815)*x[0]**o+(-0.939716907399)*x[0]+(-0.748413832758)*x[1]**o+(-0.196912090362)*x[1]
        s[1,2,2,1]=(0.447998156083)*x[0]**o+(-0.384539204664)*x[0]+(0.324989614124)*x[1]**o+(0.388489258939)*x[1]
        s[1,3,0,0]=(0.156094378734)*x[0]**o+(0.0533068325426)*x[0]+(0.268202796609)*x[1]**o+(-0.255422374899)*x[1]
        s[1,3,0,1]=(0.739277170457)*x[0]**o+(-0.290832216981)*x[0]+(0.350183930218)*x[1]**o+(-0.408368578636)*x[1]
        s[1,3,1,0]=(-0.0712530751134)*x[0]**o+(0.0658634843799)*x[0]+(0.099290893895)*x[1]**o+(0.647215326641)*x[1]
        s[1,3,1,1]=(0.0186407323716)*x[0]**o+(0.976985705627)*x[0]+(-0.871306465979)*x[1]**o+(-0.348450333839)*x[1]
        s[1,3,2,0]=(0.654402359016)*x[0]**o+(-0.633876744297)*x[0]+(0.160936266611)*x[1]**o+(0.476046463717)*x[1]
        s[1,3,2,1]=(-0.211306213041)*x[0]**o+(-0.296909064821)*x[0]+(-0.0612316728533)*x[1]**o+(0.901013542531)*x[1]
        s[1,4,0,0]=(0.974517622467)*x[0]**o+(0.998168509173)*x[0]+(-0.65931551157)*x[1]**o+(0.52094937925)*x[1]
        s[1,4,0,1]=(-0.145944864102)*x[0]**o+(-0.220443830385)*x[0]+(-0.380011620278)*x[1]**o+(0.315714506907)*x[1]
        s[1,4,1,0]=(0.00531622403943)*x[0]**o+(0.835826085399)*x[0]+(0.940648822215)*x[1]**o+(-0.505845376528)*x[1]
        s[1,4,1,1]=(0.794860008787)*x[0]**o+(-0.0885596053398)*x[0]+(-0.997579283268)*x[1]**o+(-0.339054217487)*x[1]
        s[1,4,2,0]=(-0.767918730098)*x[0]**o+(-0.023203645481)*x[0]+(-0.328338152731)*x[1]**o+(-0.232724548626)*x[1]
        s[1,4,2,1]=(0.211271068544)*x[0]**o+(-0.615899439167)*x[0]+(-0.0550973662218)*x[1]**o+(0.298493847127)*x[1]
        s[2,0,0,0]=(-0.911192683424)*x[0]**o+(0.00941390836017)*x[0]+(0.613172977466)*x[1]**o+(-0.788194982428)*x[1]
        s[2,0,0,1]=(-0.611555394987)*x[0]**o+(-0.577205119049)*x[0]+(0.319048321193)*x[1]**o+(-0.292972092026)*x[1]
        s[2,0,1,0]=(-0.669546200077)*x[0]**o+(-0.329254481941)*x[0]+(-0.841878266845)*x[1]**o+(-0.430302552502)*x[1]
        s[2,0,1,1]=(0.717338343232)*x[0]**o+(-0.184491535462)*x[0]+(-0.279163935394)*x[1]**o+(0.0902583733781)*x[1]
        s[2,0,2,0]=(0.74607828062)*x[0]**o+(0.220552976213)*x[0]+(-0.706261725925)*x[1]**o+(0.623659385726)*x[1]
        s[2,0,2,1]=(-0.673037605536)*x[0]**o+(-0.339261766556)*x[0]+(0.320833099845)*x[1]**o+(0.52076779438)*x[1]
        s[2,1,0,0]=(-0.765649699147)*x[0]**o+(0.916686538205)*x[0]+(0.97387546141)*x[1]**o+(0.553218952221)*x[1]
        s[2,1,0,1]=(0.973872792159)*x[0]**o+(0.517984046533)*x[0]+(0.51777252219)*x[1]**o+(-0.930026053905)*x[1]
        s[2,1,1,0]=(-0.948456218104)*x[0]**o+(-0.589599908446)*x[0]+(-0.467553533038)*x[1]**o+(0.274344478481)*x[1]
        s[2,1,1,1]=(0.881166959527)*x[0]**o+(0.49970488032)*x[0]+(-0.145215907327)*x[1]**o+(-0.781851254801)*x[1]
        s[2,1,2,0]=(0.249782338934)*x[0]**o+(-0.0658588005808)*x[0]+(0.812406146808)*x[1]**o+(0.692856516166)*x[1]
        s[2,1,2,1]=(-0.421909165386)*x[0]**o+(0.0386490156907)*x[0]+(-0.777622310099)*x[1]**o+(0.814181162845)*x[1]
        s[2,2,0,0]=(0.788918999327)*x[0]**o+(0.89626709297)*x[0]+(-0.0416920496206)*x[1]**o+(0.534755088018)*x[1]
        s[2,2,0,1]=(-0.723430509139)*x[0]**o+(0.937773302416)*x[0]+(0.120624514522)*x[1]**o+(0.712526712703)*x[1]
        s[2,2,1,0]=(0.755999533415)*x[0]**o+(-0.25818410984)*x[0]+(-0.833677276054)*x[1]**o+(0.389537968068)*x[1]
        s[2,2,1,1]=(-0.405091108345)*x[0]**o+(0.960266750195)*x[0]+(0.805114171635)*x[1]**o+(0.782424140168)*x[1]
        s[2,2,2,0]=(-0.00378071602137)*x[0]**o+(-0.774268061846)*x[0]+(0.317996842663)*x[1]**o+(-0.298998768873)*x[1]
        s[2,2,2,1]=(-0.640713160714)*x[0]**o+(-0.602638574906)*x[0]+(0.846395131454)*x[1]**o+(0.810407357088)*x[1]
        s[2,3,0,0]=(0.52309664987)*x[0]**o+(0.340015333205)*x[0]+(-0.463305517289)*x[1]**o+(0.377383683585)*x[1]
        s[2,3,0,1]=(-0.200960363239)*x[0]**o+(-0.343104026793)*x[0]+(0.72971674385)*x[1]**o+(-0.250972070079)*x[1]
        s[2,3,1,0]=(-0.198674852354)*x[0]**o+(0.376975952368)*x[0]+(-0.289846962669)*x[1]**o+(0.503874399797)*x[1]
        s[2,3,1,1]=(-0.500646179446)*x[0]**o+(0.999980656019)*x[0]+(-0.22225911248)*x[1]**o+(0.59532135129)*x[1]
        s[2,3,2,0]=(-0.0479308919198)*x[0]**o+(-0.286156974259)*x[0]+(-0.565672171896)*x[1]**o+(-0.496164919028)*x[1]
        s[2,3,2,1]=(0.283126731716)*x[0]**o+(-0.847220465645)*x[0]+(-0.202433223165)*x[1]**o+(0.733249392104)*x[1]
        s[2,4,0,0]=(0.379082415544)*x[0]**o+(0.619696544807)*x[0]+(0.30046651915)*x[1]**o+(-0.0956554688639)*x[1]
        s[2,4,0,1]=(0.645993528534)*x[0]**o+(0.636401374002)*x[0]+(-0.48531244649)*x[1]**o+(-0.566808631893)*x[1]
        s[2,4,1,0]=(-0.27936019151)*x[0]**o+(-0.0453561833702)*x[0]+(-0.38173600111)*x[1]**o+(0.0436591877645)*x[1]
        s[2,4,1,1]=(-0.305384262353)*x[0]**o+(-0.541085058351)*x[0]+(-0.055581601052)*x[1]**o+(0.128158662772)*x[1]
        s[2,4,2,0]=(-0.725495356548)*x[0]**o+(0.29235879065)*x[0]+(0.468015633053)*x[1]**o+(0.907396937488)*x[1]
        s[2,4,2,1]=(-0.191459527415)*x[0]**o+(0.337550470203)*x[0]+(0.846947249886)*x[1]**o+(0.115611176928)*x[1]
        s[3,0,0,0]=(-0.33176415824)*x[0]**o+(0.334074664888)*x[0]+(0.71328836071)*x[1]**o+(-0.108353365639)*x[1]
        s[3,0,0,1]=(0.189471905167)*x[0]**o+(-0.735388054498)*x[0]+(-0.145791816558)*x[1]**o+(0.1166770253)*x[1]
        s[3,0,1,0]=(0.130681026646)*x[0]**o+(-0.429395309447)*x[0]+(-0.960782097087)*x[1]**o+(-0.831182265535)*x[1]
        s[3,0,1,1]=(0.17875090433)*x[0]**o+(0.117342315307)*x[0]+(0.712145561102)*x[1]**o+(0.980052033498)*x[1]
        s[3,0,2,0]=(0.350160192028)*x[0]**o+(0.319028577789)*x[0]+(-0.390278261047)*x[1]**o+(-0.563618392837)*x[1]
        s[3,0,2,1]=(0.612683052829)*x[0]**o+(-0.808481977369)*x[0]+(-0.275769366104)*x[1]**o+(-0.16847980965)*x[1]
        s[3,1,0,0]=(0.441331101931)*x[0]**o+(0.631654604311)*x[0]+(-0.417381407441)*x[1]**o+(0.623209016434)*x[1]
        s[3,1,0,1]=(-0.316799359576)*x[0]**o+(0.0469238577708)*x[0]+(0.453846283394)*x[1]**o+(0.235093194654)*x[1]
        s[3,1,1,0]=(0.709151428964)*x[0]**o+(-0.66563806393)*x[0]+(0.957204730095)*x[1]**o+(0.757289506743)*x[1]
        s[3,1,1,1]=(0.651240972131)*x[0]**o+(0.125595241682)*x[0]+(0.673153945584)*x[1]**o+(-0.0741235551187)*x[1]
        s[3,1,2,0]=(0.109062793745)*x[0]**o+(0.603138432718)*x[0]+(0.626347779202)*x[1]**o+(0.747924284641)*x[1]
        s[3,1,2,1]=(-0.708014322028)*x[0]**o+(0.656641954147)*x[0]+(-0.5750166511)*x[1]**o+(0.971548956135)*x[1]
        s[3,2,0,0]=(0.308690818385)*x[0]**o+(-0.0743510684794)*x[0]+(0.503445484884)*x[1]**o+(0.122671270558)*x[1]
        s[3,2,0,1]=(-0.605770669307)*x[0]**o+(0.235294066964)*x[0]+(0.328631665933)*x[1]**o+(-0.713722501714)*x[1]
        s[3,2,1,0]=(0.772400565953)*x[0]**o+(0.889908318804)*x[0]+(0.166946692512)*x[1]**o+(0.207246011251)*x[1]
        s[3,2,1,1]=(-0.0256795661601)*x[0]**o+(-0.438929207275)*x[0]+(-0.586749375325)*x[1]**o+(-0.365377718456)*x[1]
        s[3,2,2,0]=(-0.395900535643)*x[0]**o+(0.814855175613)*x[0]+(0.139931347448)*x[1]**o+(-0.305268254265)*x[1]
        s[3,2,2,1]=(-0.613972524997)*x[0]**o+(0.151282941113)*x[0]+(0.137716512378)*x[1]**o+(0.987022640532)*x[1]
        s[3,3,0,0]=(0.054989906601)*x[0]**o+(0.629145288873)*x[0]+(-0.844944582016)*x[1]**o+(0.146581752248)*x[1]
        s[3,3,0,1]=(-0.14487953322)*x[0]**o+(0.618818776941)*x[0]+(0.924407652252)*x[1]**o+(0.563487350843)*x[1]
        s[3,3,1,0]=(0.58912173549)*x[0]**o+(-0.0901432408625)*x[0]+(0.873874316115)*x[1]**o+(0.881058120153)*x[1]
        s[3,3,1,1]=(0.576631116996)*x[0]**o+(-0.686918072799)*x[0]+(-0.50936823904)*x[1]**o+(0.585732475064)*x[1]
        s[3,3,2,0]=(0.306914981486)*x[0]**o+(-0.433296796977)*x[0]+(0.194615724385)*x[1]**o+(-0.98720631317)*x[1]
        s[3,3,2,1]=(0.48136043634)*x[0]**o+(0.38409585721)*x[0]+(0.809958268113)*x[1]**o+(-0.197158044454)*x[1]
        s[3,4,0,0]=(0.562436637234)*x[0]**o+(-0.729750092898)*x[0]+(-0.638146901507)*x[1]**o+(-0.629609335189)*x[1]
        s[3,4,0,1]=(0.610239954639)*x[0]**o+(-0.796185555272)*x[0]+(0.618408850895)*x[1]**o+(0.409684187852)*x[1]
        s[3,4,1,0]=(0.176527890224)*x[0]**o+(-0.680125370142)*x[0]+(0.297855195077)*x[1]**o+(0.416241653852)*x[1]
        s[3,4,1,1]=(-0.656877186833)*x[0]**o+(-0.622659703955)*x[0]+(0.92765398041)*x[1]**o+(-0.381997501461)*x[1]
        s[3,4,2,0]=(-0.206401064204)*x[0]**o+(0.496539998899)*x[0]+(0.303242800413)*x[1]**o+(-0.648983010434)*x[1]
        s[3,4,2,1]=(-0.850103219136)*x[0]**o+(-0.54057890339)*x[0]+(0.673138252663)*x[1]**o+(0.0181453330374)*x[1]
        ref[0,0,0,0]=(0.209596991985)*x_ref[0]**o+(0.997116487512)*x_ref[0]+(0.0205082690151)*x_ref[1]**o+(0.781596313955)*x_ref[1]
        ref[0,0,0,1]=(-0.475035630436)*x_ref[0]**o+(0.476939038583)*x_ref[0]+(-0.403224734103)*x_ref[1]**o+(-0.726327526537)*x_ref[1]
        ref[0,0,1,0]=(0.165692569014)*x_ref[0]**o+(-0.958953962148)*x_ref[0]+(0.778520844819)*x_ref[1]**o+(-0.371211501917)*x_ref[1]
        ref[0,0,1,1]=(-0.0863741752107)*x_ref[0]**o+(0.64735697812)*x_ref[0]+(-0.717371101374)*x_ref[1]**o+(-0.517476214609)*x_ref[1]
        ref[0,0,2,0]=(0.414678358278)*x_ref[0]**o+(0.766044973671)*x_ref[0]+(0.473600649794)*x_ref[1]**o+(-0.350863429165)*x_ref[1]
        ref[0,0,2,1]=(0.0391148157551)*x_ref[0]**o+(0.550258520615)*x_ref[0]+(-0.0952445590164)*x_ref[1]**o+(0.799829721692)*x_ref[1]
        ref[0,1,0,0]=(-0.910760056281)*x_ref[0]**o+(0.319346354593)*x_ref[0]+(0.640854539851)*x_ref[1]**o+(0.0622173716277)*x_ref[1]
        ref[0,1,0,1]=(0.761672466091)*x_ref[0]**o+(0.266521259527)*x_ref[0]+(-0.567155906289)*x_ref[1]**o+(0.124135553341)*x_ref[1]
        ref[0,1,1,0]=(-0.681712184552)*x_ref[0]**o+(0.388484728164)*x_ref[0]+(-0.412243885661)*x_ref[1]**o+(0.637954103055)*x_ref[1]
        ref[0,1,1,1]=(-0.485017958909)*x_ref[0]**o+(0.736606905633)*x_ref[0]+(-0.299204062504)*x_ref[1]**o+(0.0840021848608)*x_ref[1]
        ref[0,1,2,0]=(0.955598154983)*x_ref[0]**o+(-0.0204689463318)*x_ref[0]+(-0.737123328202)*x_ref[1]**o+(0.708418145874)*x_ref[1]
        ref[0,1,2,1]=(0.541565827885)*x_ref[0]**o+(0.97493788494)*x_ref[0]+(-0.450344230708)*x_ref[1]**o+(-0.0231871887964)*x_ref[1]
        ref[0,2,0,0]=(-0.338442282639)*x_ref[0]**o+(0.0746189409911)*x_ref[0]+(0.76030098768)*x_ref[1]**o+(0.629119828718)*x_ref[1]
        ref[0,2,0,1]=(-0.900573785768)*x_ref[0]**o+(-0.971041944198)*x_ref[0]+(0.637105978641)*x_ref[1]**o+(0.387639692227)*x_ref[1]
        ref[0,2,1,0]=(0.0889093715568)*x_ref[0]**o+(0.260479957021)*x_ref[0]+(-0.908420084369)*x_ref[1]**o+(0.6517961193)*x_ref[1]
        ref[0,2,1,1]=(0.683959794134)*x_ref[0]**o+(-0.84665688176)*x_ref[0]+(0.710853815216)*x_ref[1]**o+(0.536738240986)*x_ref[1]
        ref[0,2,2,0]=(0.365754968462)*x_ref[0]**o+(0.994576642301)*x_ref[0]+(0.110702356612)*x_ref[1]**o+(0.68992003076)*x_ref[1]
        ref[0,2,2,1]=(-0.238463519384)*x_ref[0]**o+(-0.91372863405)*x_ref[0]+(-0.334884010749)*x_ref[1]**o+(0.481631381739)*x_ref[1]
        ref[0,3,0,0]=(-0.580344649945)*x_ref[0]**o+(0.557953798016)*x_ref[0]+(0.835117346932)*x_ref[1]**o+(-0.196440751777)*x_ref[1]
        ref[0,3,0,1]=(0.339197786593)*x_ref[0]**o+(0.231910549016)*x_ref[0]+(0.515605654416)*x_ref[1]**o+(-0.546885567308)*x_ref[1]
        ref[0,3,1,0]=(-0.562198188719)*x_ref[0]**o+(0.410791865348)*x_ref[0]+(-0.0823054523407)*x_ref[1]**o+(0.789832279038)*x_ref[1]
        ref[0,3,1,1]=(-0.519245133296)*x_ref[0]**o+(0.622234324355)*x_ref[0]+(-0.89225914065)*x_ref[1]**o+(-0.0376143576483)*x_ref[1]
        ref[0,3,2,0]=(-0.884895063274)*x_ref[0]**o+(-0.569137436872)*x_ref[0]+(0.107652574639)*x_ref[1]**o+(0.936768318078)*x_ref[1]
        ref[0,3,2,1]=(-0.209493065726)*x_ref[0]**o+(0.830161297986)*x_ref[0]+(-0.811634331709)*x_ref[1]**o+(0.0673874073431)*x_ref[1]
        ref[0,4,0,0]=(-0.518639168744)*x_ref[0]**o+(-0.953627666411)*x_ref[0]+(0.00153430886561)*x_ref[1]**o+(0.118432410415)*x_ref[1]
        ref[0,4,0,1]=(0.901325505931)*x_ref[0]**o+(0.276592227771)*x_ref[0]+(0.906505574302)*x_ref[1]**o+(-0.604032185175)*x_ref[1]
        ref[0,4,1,0]=(0.539336544226)*x_ref[0]**o+(-0.518177781519)*x_ref[0]+(-0.397453923803)*x_ref[1]**o+(0.0661677991096)*x_ref[1]
        ref[0,4,1,1]=(0.31900579667)*x_ref[0]**o+(-0.950735779053)*x_ref[0]+(0.495414320287)*x_ref[1]**o+(0.310834200079)*x_ref[1]
        ref[0,4,2,0]=(0.567277111907)*x_ref[0]**o+(0.124915250004)*x_ref[0]+(0.699531698929)*x_ref[1]**o+(-0.434718433784)*x_ref[1]
        ref[0,4,2,1]=(0.607190906026)*x_ref[0]**o+(0.411045834019)*x_ref[0]+(-0.116848105246)*x_ref[1]**o+(-0.343872842277)*x_ref[1]
        ref[1,0,0,0]=(0.263188443069)*x_ref[0]**o+(0.782586354151)*x_ref[0]+(-0.0593809262828)*x_ref[1]**o+(-0.516894145417)*x_ref[1]
        ref[1,0,0,1]=(-0.155136250976)*x_ref[0]**o+(0.155174971237)*x_ref[0]+(0.452174963131)*x_ref[1]**o+(-0.811875854704)*x_ref[1]
        ref[1,0,1,0]=(-0.161366848419)*x_ref[0]**o+(-0.674425732875)*x_ref[0]+(-0.99246147148)*x_ref[1]**o+(0.618506213959)*x_ref[1]
        ref[1,0,1,1]=(0.64592094187)*x_ref[0]**o+(0.444912658041)*x_ref[0]+(0.963730302998)*x_ref[1]**o+(-0.227756330634)*x_ref[1]
        ref[1,0,2,0]=(0.199217092911)*x_ref[0]**o+(-0.118577255523)*x_ref[0]+(-0.560600457675)*x_ref[1]**o+(-0.233703068897)*x_ref[1]
        ref[1,0,2,1]=(0.200710250009)*x_ref[0]**o+(0.366088123353)*x_ref[0]+(0.419249984559)*x_ref[1]**o+(-0.0188044094656)*x_ref[1]
        ref[1,1,0,0]=(-0.990545888549)*x_ref[0]**o+(-0.100521244971)*x_ref[0]+(-0.0134093118672)*x_ref[1]**o+(0.582791526008)*x_ref[1]
        ref[1,1,0,1]=(-0.932085864166)*x_ref[0]**o+(-0.402198623203)*x_ref[0]+(-0.769996422181)*x_ref[1]**o+(0.701839645425)*x_ref[1]
        ref[1,1,1,0]=(-0.0546050562235)*x_ref[0]**o+(-0.821680588129)*x_ref[0]+(0.254170421382)*x_ref[1]**o+(0.0419062526906)*x_ref[1]
        ref[1,1,1,1]=(0.0361631230129)*x_ref[0]**o+(0.155809438009)*x_ref[0]+(-0.797631293341)*x_ref[1]**o+(-0.0764161686239)*x_ref[1]
        ref[1,1,2,0]=(-0.679858025056)*x_ref[0]**o+(-0.0411819071479)*x_ref[0]+(0.033731324617)*x_ref[1]**o+(0.32496792339)*x_ref[1]
        ref[1,1,2,1]=(0.424198647091)*x_ref[0]**o+(-0.332229088063)*x_ref[0]+(-0.225491160735)*x_ref[1]**o+(-0.833706567263)*x_ref[1]
        ref[1,2,0,0]=(0.701510843723)*x_ref[0]**o+(0.780699569051)*x_ref[0]+(0.369319568113)*x_ref[1]**o+(-0.865220461023)*x_ref[1]
        ref[1,2,0,1]=(-0.388700890661)*x_ref[0]**o+(-0.550094979352)*x_ref[0]+(0.585925695487)*x_ref[1]**o+(0.812085187015)*x_ref[1]
        ref[1,2,1,0]=(-0.825827041568)*x_ref[0]**o+(-0.281570175483)*x_ref[0]+(-0.352446259467)*x_ref[1]**o+(-0.162431177857)*x_ref[1]
        ref[1,2,1,1]=(-0.615917011781)*x_ref[0]**o+(-0.218604155424)*x_ref[0]+(-0.66718327064)*x_ref[1]**o+(-0.723246262472)*x_ref[1]
        ref[1,2,2,0]=(-0.794217383815)*x_ref[0]**o+(-0.939716907399)*x_ref[0]+(-0.748413832758)*x_ref[1]**o+(-0.196912090362)*x_ref[1]
        ref[1,2,2,1]=(0.447998156083)*x_ref[0]**o+(-0.384539204664)*x_ref[0]+(0.324989614124)*x_ref[1]**o+(0.388489258939)*x_ref[1]
        ref[1,3,0,0]=(0.156094378734)*x_ref[0]**o+(0.0533068325426)*x_ref[0]+(0.268202796609)*x_ref[1]**o+(-0.255422374899)*x_ref[1]
        ref[1,3,0,1]=(0.739277170457)*x_ref[0]**o+(-0.290832216981)*x_ref[0]+(0.350183930218)*x_ref[1]**o+(-0.408368578636)*x_ref[1]
        ref[1,3,1,0]=(-0.0712530751134)*x_ref[0]**o+(0.0658634843799)*x_ref[0]+(0.099290893895)*x_ref[1]**o+(0.647215326641)*x_ref[1]
        ref[1,3,1,1]=(0.0186407323716)*x_ref[0]**o+(0.976985705627)*x_ref[0]+(-0.871306465979)*x_ref[1]**o+(-0.348450333839)*x_ref[1]
        ref[1,3,2,0]=(0.654402359016)*x_ref[0]**o+(-0.633876744297)*x_ref[0]+(0.160936266611)*x_ref[1]**o+(0.476046463717)*x_ref[1]
        ref[1,3,2,1]=(-0.211306213041)*x_ref[0]**o+(-0.296909064821)*x_ref[0]+(-0.0612316728533)*x_ref[1]**o+(0.901013542531)*x_ref[1]
        ref[1,4,0,0]=(0.974517622467)*x_ref[0]**o+(0.998168509173)*x_ref[0]+(-0.65931551157)*x_ref[1]**o+(0.52094937925)*x_ref[1]
        ref[1,4,0,1]=(-0.145944864102)*x_ref[0]**o+(-0.220443830385)*x_ref[0]+(-0.380011620278)*x_ref[1]**o+(0.315714506907)*x_ref[1]
        ref[1,4,1,0]=(0.00531622403943)*x_ref[0]**o+(0.835826085399)*x_ref[0]+(0.940648822215)*x_ref[1]**o+(-0.505845376528)*x_ref[1]
        ref[1,4,1,1]=(0.794860008787)*x_ref[0]**o+(-0.0885596053398)*x_ref[0]+(-0.997579283268)*x_ref[1]**o+(-0.339054217487)*x_ref[1]
        ref[1,4,2,0]=(-0.767918730098)*x_ref[0]**o+(-0.023203645481)*x_ref[0]+(-0.328338152731)*x_ref[1]**o+(-0.232724548626)*x_ref[1]
        ref[1,4,2,1]=(0.211271068544)*x_ref[0]**o+(-0.615899439167)*x_ref[0]+(-0.0550973662218)*x_ref[1]**o+(0.298493847127)*x_ref[1]
        ref[2,0,0,0]=(-0.911192683424)*x_ref[0]**o+(0.00941390836017)*x_ref[0]+(0.613172977466)*x_ref[1]**o+(-0.788194982428)*x_ref[1]
        ref[2,0,0,1]=(-0.611555394987)*x_ref[0]**o+(-0.577205119049)*x_ref[0]+(0.319048321193)*x_ref[1]**o+(-0.292972092026)*x_ref[1]
        ref[2,0,1,0]=(-0.669546200077)*x_ref[0]**o+(-0.329254481941)*x_ref[0]+(-0.841878266845)*x_ref[1]**o+(-0.430302552502)*x_ref[1]
        ref[2,0,1,1]=(0.717338343232)*x_ref[0]**o+(-0.184491535462)*x_ref[0]+(-0.279163935394)*x_ref[1]**o+(0.0902583733781)*x_ref[1]
        ref[2,0,2,0]=(0.74607828062)*x_ref[0]**o+(0.220552976213)*x_ref[0]+(-0.706261725925)*x_ref[1]**o+(0.623659385726)*x_ref[1]
        ref[2,0,2,1]=(-0.673037605536)*x_ref[0]**o+(-0.339261766556)*x_ref[0]+(0.320833099845)*x_ref[1]**o+(0.52076779438)*x_ref[1]
        ref[2,1,0,0]=(-0.765649699147)*x_ref[0]**o+(0.916686538205)*x_ref[0]+(0.97387546141)*x_ref[1]**o+(0.553218952221)*x_ref[1]
        ref[2,1,0,1]=(0.973872792159)*x_ref[0]**o+(0.517984046533)*x_ref[0]+(0.51777252219)*x_ref[1]**o+(-0.930026053905)*x_ref[1]
        ref[2,1,1,0]=(-0.948456218104)*x_ref[0]**o+(-0.589599908446)*x_ref[0]+(-0.467553533038)*x_ref[1]**o+(0.274344478481)*x_ref[1]
        ref[2,1,1,1]=(0.881166959527)*x_ref[0]**o+(0.49970488032)*x_ref[0]+(-0.145215907327)*x_ref[1]**o+(-0.781851254801)*x_ref[1]
        ref[2,1,2,0]=(0.249782338934)*x_ref[0]**o+(-0.0658588005808)*x_ref[0]+(0.812406146808)*x_ref[1]**o+(0.692856516166)*x_ref[1]
        ref[2,1,2,1]=(-0.421909165386)*x_ref[0]**o+(0.0386490156907)*x_ref[0]+(-0.777622310099)*x_ref[1]**o+(0.814181162845)*x_ref[1]
        ref[2,2,0,0]=(0.788918999327)*x_ref[0]**o+(0.89626709297)*x_ref[0]+(-0.0416920496206)*x_ref[1]**o+(0.534755088018)*x_ref[1]
        ref[2,2,0,1]=(-0.723430509139)*x_ref[0]**o+(0.937773302416)*x_ref[0]+(0.120624514522)*x_ref[1]**o+(0.712526712703)*x_ref[1]
        ref[2,2,1,0]=(0.755999533415)*x_ref[0]**o+(-0.25818410984)*x_ref[0]+(-0.833677276054)*x_ref[1]**o+(0.389537968068)*x_ref[1]
        ref[2,2,1,1]=(-0.405091108345)*x_ref[0]**o+(0.960266750195)*x_ref[0]+(0.805114171635)*x_ref[1]**o+(0.782424140168)*x_ref[1]
        ref[2,2,2,0]=(-0.00378071602137)*x_ref[0]**o+(-0.774268061846)*x_ref[0]+(0.317996842663)*x_ref[1]**o+(-0.298998768873)*x_ref[1]
        ref[2,2,2,1]=(-0.640713160714)*x_ref[0]**o+(-0.602638574906)*x_ref[0]+(0.846395131454)*x_ref[1]**o+(0.810407357088)*x_ref[1]
        ref[2,3,0,0]=(0.52309664987)*x_ref[0]**o+(0.340015333205)*x_ref[0]+(-0.463305517289)*x_ref[1]**o+(0.377383683585)*x_ref[1]
        ref[2,3,0,1]=(-0.200960363239)*x_ref[0]**o+(-0.343104026793)*x_ref[0]+(0.72971674385)*x_ref[1]**o+(-0.250972070079)*x_ref[1]
        ref[2,3,1,0]=(-0.198674852354)*x_ref[0]**o+(0.376975952368)*x_ref[0]+(-0.289846962669)*x_ref[1]**o+(0.503874399797)*x_ref[1]
        ref[2,3,1,1]=(-0.500646179446)*x_ref[0]**o+(0.999980656019)*x_ref[0]+(-0.22225911248)*x_ref[1]**o+(0.59532135129)*x_ref[1]
        ref[2,3,2,0]=(-0.0479308919198)*x_ref[0]**o+(-0.286156974259)*x_ref[0]+(-0.565672171896)*x_ref[1]**o+(-0.496164919028)*x_ref[1]
        ref[2,3,2,1]=(0.283126731716)*x_ref[0]**o+(-0.847220465645)*x_ref[0]+(-0.202433223165)*x_ref[1]**o+(0.733249392104)*x_ref[1]
        ref[2,4,0,0]=(0.379082415544)*x_ref[0]**o+(0.619696544807)*x_ref[0]+(0.30046651915)*x_ref[1]**o+(-0.0956554688639)*x_ref[1]
        ref[2,4,0,1]=(0.645993528534)*x_ref[0]**o+(0.636401374002)*x_ref[0]+(-0.48531244649)*x_ref[1]**o+(-0.566808631893)*x_ref[1]
        ref[2,4,1,0]=(-0.27936019151)*x_ref[0]**o+(-0.0453561833702)*x_ref[0]+(-0.38173600111)*x_ref[1]**o+(0.0436591877645)*x_ref[1]
        ref[2,4,1,1]=(-0.305384262353)*x_ref[0]**o+(-0.541085058351)*x_ref[0]+(-0.055581601052)*x_ref[1]**o+(0.128158662772)*x_ref[1]
        ref[2,4,2,0]=(-0.725495356548)*x_ref[0]**o+(0.29235879065)*x_ref[0]+(0.468015633053)*x_ref[1]**o+(0.907396937488)*x_ref[1]
        ref[2,4,2,1]=(-0.191459527415)*x_ref[0]**o+(0.337550470203)*x_ref[0]+(0.846947249886)*x_ref[1]**o+(0.115611176928)*x_ref[1]
        ref[3,0,0,0]=(-0.33176415824)*x_ref[0]**o+(0.334074664888)*x_ref[0]+(0.71328836071)*x_ref[1]**o+(-0.108353365639)*x_ref[1]
        ref[3,0,0,1]=(0.189471905167)*x_ref[0]**o+(-0.735388054498)*x_ref[0]+(-0.145791816558)*x_ref[1]**o+(0.1166770253)*x_ref[1]
        ref[3,0,1,0]=(0.130681026646)*x_ref[0]**o+(-0.429395309447)*x_ref[0]+(-0.960782097087)*x_ref[1]**o+(-0.831182265535)*x_ref[1]
        ref[3,0,1,1]=(0.17875090433)*x_ref[0]**o+(0.117342315307)*x_ref[0]+(0.712145561102)*x_ref[1]**o+(0.980052033498)*x_ref[1]
        ref[3,0,2,0]=(0.350160192028)*x_ref[0]**o+(0.319028577789)*x_ref[0]+(-0.390278261047)*x_ref[1]**o+(-0.563618392837)*x_ref[1]
        ref[3,0,2,1]=(0.612683052829)*x_ref[0]**o+(-0.808481977369)*x_ref[0]+(-0.275769366104)*x_ref[1]**o+(-0.16847980965)*x_ref[1]
        ref[3,1,0,0]=(0.441331101931)*x_ref[0]**o+(0.631654604311)*x_ref[0]+(-0.417381407441)*x_ref[1]**o+(0.623209016434)*x_ref[1]
        ref[3,1,0,1]=(-0.316799359576)*x_ref[0]**o+(0.0469238577708)*x_ref[0]+(0.453846283394)*x_ref[1]**o+(0.235093194654)*x_ref[1]
        ref[3,1,1,0]=(0.709151428964)*x_ref[0]**o+(-0.66563806393)*x_ref[0]+(0.957204730095)*x_ref[1]**o+(0.757289506743)*x_ref[1]
        ref[3,1,1,1]=(0.651240972131)*x_ref[0]**o+(0.125595241682)*x_ref[0]+(0.673153945584)*x_ref[1]**o+(-0.0741235551187)*x_ref[1]
        ref[3,1,2,0]=(0.109062793745)*x_ref[0]**o+(0.603138432718)*x_ref[0]+(0.626347779202)*x_ref[1]**o+(0.747924284641)*x_ref[1]
        ref[3,1,2,1]=(-0.708014322028)*x_ref[0]**o+(0.656641954147)*x_ref[0]+(-0.5750166511)*x_ref[1]**o+(0.971548956135)*x_ref[1]
        ref[3,2,0,0]=(0.308690818385)*x_ref[0]**o+(-0.0743510684794)*x_ref[0]+(0.503445484884)*x_ref[1]**o+(0.122671270558)*x_ref[1]
        ref[3,2,0,1]=(-0.605770669307)*x_ref[0]**o+(0.235294066964)*x_ref[0]+(0.328631665933)*x_ref[1]**o+(-0.713722501714)*x_ref[1]
        ref[3,2,1,0]=(0.772400565953)*x_ref[0]**o+(0.889908318804)*x_ref[0]+(0.166946692512)*x_ref[1]**o+(0.207246011251)*x_ref[1]
        ref[3,2,1,1]=(-0.0256795661601)*x_ref[0]**o+(-0.438929207275)*x_ref[0]+(-0.586749375325)*x_ref[1]**o+(-0.365377718456)*x_ref[1]
        ref[3,2,2,0]=(-0.395900535643)*x_ref[0]**o+(0.814855175613)*x_ref[0]+(0.139931347448)*x_ref[1]**o+(-0.305268254265)*x_ref[1]
        ref[3,2,2,1]=(-0.613972524997)*x_ref[0]**o+(0.151282941113)*x_ref[0]+(0.137716512378)*x_ref[1]**o+(0.987022640532)*x_ref[1]
        ref[3,3,0,0]=(0.054989906601)*x_ref[0]**o+(0.629145288873)*x_ref[0]+(-0.844944582016)*x_ref[1]**o+(0.146581752248)*x_ref[1]
        ref[3,3,0,1]=(-0.14487953322)*x_ref[0]**o+(0.618818776941)*x_ref[0]+(0.924407652252)*x_ref[1]**o+(0.563487350843)*x_ref[1]
        ref[3,3,1,0]=(0.58912173549)*x_ref[0]**o+(-0.0901432408625)*x_ref[0]+(0.873874316115)*x_ref[1]**o+(0.881058120153)*x_ref[1]
        ref[3,3,1,1]=(0.576631116996)*x_ref[0]**o+(-0.686918072799)*x_ref[0]+(-0.50936823904)*x_ref[1]**o+(0.585732475064)*x_ref[1]
        ref[3,3,2,0]=(0.306914981486)*x_ref[0]**o+(-0.433296796977)*x_ref[0]+(0.194615724385)*x_ref[1]**o+(-0.98720631317)*x_ref[1]
        ref[3,3,2,1]=(0.48136043634)*x_ref[0]**o+(0.38409585721)*x_ref[0]+(0.809958268113)*x_ref[1]**o+(-0.197158044454)*x_ref[1]
        ref[3,4,0,0]=(0.562436637234)*x_ref[0]**o+(-0.729750092898)*x_ref[0]+(-0.638146901507)*x_ref[1]**o+(-0.629609335189)*x_ref[1]
        ref[3,4,0,1]=(0.610239954639)*x_ref[0]**o+(-0.796185555272)*x_ref[0]+(0.618408850895)*x_ref[1]**o+(0.409684187852)*x_ref[1]
        ref[3,4,1,0]=(0.176527890224)*x_ref[0]**o+(-0.680125370142)*x_ref[0]+(0.297855195077)*x_ref[1]**o+(0.416241653852)*x_ref[1]
        ref[3,4,1,1]=(-0.656877186833)*x_ref[0]**o+(-0.622659703955)*x_ref[0]+(0.92765398041)*x_ref[1]**o+(-0.381997501461)*x_ref[1]
        ref[3,4,2,0]=(-0.206401064204)*x_ref[0]**o+(0.496539998899)*x_ref[0]+(0.303242800413)*x_ref[1]**o+(-0.648983010434)*x_ref[1]
        ref[3,4,2,1]=(-0.850103219136)*x_ref[0]**o+(-0.54057890339)*x_ref[0]+(0.673138252663)*x_ref[1]**o+(0.0181453330374)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.183051403523)*x[0]**o+(0.675163647622)*x[0]+(0.383724078835)*x[1]**o+(-0.863174107932)*x[1]+(0.752053635179)*x[2]**o+(-0.442218538219)*x[2]
        s[0,0,0,1]=(0.145022844971)*x[0]**o+(-0.390513961466)*x[0]+(0.112480666084)*x[1]**o+(-0.484118377018)*x[1]+(0.804050685467)*x[2]**o+(-0.278608094628)*x[2]
        s[0,0,1,0]=(0.897135266304)*x[0]**o+(-0.761774008569)*x[0]+(-0.997851989059)*x[1]**o+(0.658068087366)*x[1]+(-0.80701588118)*x[2]**o+(0.596880821142)*x[2]
        s[0,0,1,1]=(-0.706353681044)*x[0]**o+(-0.19299984915)*x[0]+(0.0183270582241)*x[1]**o+(0.0848075866792)*x[1]+(-0.380936183304)*x[2]**o+(0.404471922242)*x[2]
        s[0,0,2,0]=(-0.382331572203)*x[0]**o+(-0.862239031672)*x[0]+(-0.664876503393)*x[1]**o+(0.721170780428)*x[1]+(0.936933909848)*x[2]**o+(0.553724430131)*x[2]
        s[0,0,2,1]=(0.41570549153)*x[0]**o+(-0.668260289502)*x[0]+(-0.118888455714)*x[1]**o+(0.454665663519)*x[1]+(-0.470911434781)*x[2]**o+(-0.677230009093)*x[2]
        s[0,1,0,0]=(-0.246347502133)*x[0]**o+(0.124690136148)*x[0]+(0.257467906886)*x[1]**o+(0.393404860933)*x[1]+(-0.216777928803)*x[2]**o+(-0.893375700242)*x[2]
        s[0,1,0,1]=(0.853598016536)*x[0]**o+(-0.614278282211)*x[0]+(0.958798548557)*x[1]**o+(0.539228268539)*x[1]+(-0.893749389287)*x[2]**o+(-0.421872187434)*x[2]
        s[0,1,1,0]=(-0.554158393064)*x[0]**o+(-0.0302602851801)*x[0]+(-0.627598510228)*x[1]**o+(-0.14419255857)*x[1]+(-0.319740995473)*x[2]**o+(-0.829801924596)*x[2]
        s[0,1,1,1]=(0.736629752286)*x[0]**o+(0.104716405421)*x[0]+(-0.340008364599)*x[1]**o+(-0.276509803031)*x[1]+(-0.65815905914)*x[2]**o+(-0.862207388004)*x[2]
        s[0,1,2,0]=(-0.963285485998)*x[0]**o+(-0.932861475962)*x[0]+(0.0893861922413)*x[1]**o+(-0.284311111286)*x[1]+(-0.549298067285)*x[2]**o+(0.990727028557)*x[2]
        s[0,1,2,1]=(0.354221245392)*x[0]**o+(-0.259651087945)*x[0]+(0.386589396916)*x[1]**o+(-0.315236532701)*x[1]+(0.0276667266652)*x[2]**o+(0.462795309548)*x[2]
        s[0,2,0,0]=(-0.488345436139)*x[0]**o+(0.0812145763632)*x[0]+(-0.766180568398)*x[1]**o+(0.52489429503)*x[1]+(-0.369663190972)*x[2]**o+(0.214012502771)*x[2]
        s[0,2,0,1]=(-0.447889959663)*x[0]**o+(0.815971127504)*x[0]+(-0.43620764595)*x[1]**o+(0.974646962464)*x[1]+(-0.0947298803439)*x[2]**o+(0.57765513293)*x[2]
        s[0,2,1,0]=(0.90082728964)*x[0]**o+(-0.471148200008)*x[0]+(0.130856999133)*x[1]**o+(0.94489744998)*x[1]+(0.97494516676)*x[2]**o+(0.379281707318)*x[2]
        s[0,2,1,1]=(-0.741008766911)*x[0]**o+(0.125378297796)*x[0]+(0.196554106007)*x[1]**o+(0.67968014258)*x[1]+(-0.22331716546)*x[2]**o+(0.69307086692)*x[2]
        s[0,2,2,0]=(0.731935335415)*x[0]**o+(0.203487322565)*x[0]+(-0.824149733689)*x[1]**o+(-0.765020086954)*x[1]+(0.652715725055)*x[2]**o+(-0.731883148625)*x[2]
        s[0,2,2,1]=(-0.845146208295)*x[0]**o+(-0.226473926648)*x[0]+(0.683088958387)*x[1]**o+(0.588656694973)*x[1]+(0.223845818469)*x[2]**o+(-0.311691745089)*x[2]
        s[0,3,0,0]=(-0.390979632965)*x[0]**o+(0.712430823379)*x[0]+(0.250097638139)*x[1]**o+(-0.0573974192609)*x[1]+(-0.104674518818)*x[2]**o+(-0.19187273988)*x[2]
        s[0,3,0,1]=(-0.422704759794)*x[0]**o+(-0.356438670304)*x[0]+(-0.99949139973)*x[1]**o+(-0.923501978496)*x[1]+(0.95094078991)*x[2]**o+(-0.726323072853)*x[2]
        s[0,3,1,0]=(-0.752686683053)*x[0]**o+(0.426969213595)*x[0]+(-0.884811980577)*x[1]**o+(-0.248298773095)*x[1]+(-0.248150259844)*x[2]**o+(0.307320547671)*x[2]
        s[0,3,1,1]=(0.444548575453)*x[0]**o+(-0.425971241634)*x[0]+(0.291461522783)*x[1]**o+(0.7897751016)*x[1]+(-0.7201321357)*x[2]**o+(0.335574613087)*x[2]
        s[0,3,2,0]=(0.320923109599)*x[0]**o+(0.768023804425)*x[0]+(0.786842381217)*x[1]**o+(0.508810585491)*x[1]+(0.565569549043)*x[2]**o+(-0.543149420951)*x[2]
        s[0,3,2,1]=(-0.434748955301)*x[0]**o+(0.753997363265)*x[0]+(-0.998212958853)*x[1]**o+(0.308236036581)*x[1]+(-0.604346477377)*x[2]**o+(0.0456731524173)*x[2]
        s[0,4,0,0]=(0.19259119938)*x[0]**o+(-0.676730834051)*x[0]+(0.782503706074)*x[1]**o+(-0.356602170048)*x[1]+(0.242151748383)*x[2]**o+(-0.555208960512)*x[2]
        s[0,4,0,1]=(-0.899694378162)*x[0]**o+(-0.998178345345)*x[0]+(-0.0299332693574)*x[1]**o+(-0.524636203176)*x[1]+(-0.090669349917)*x[2]**o+(-0.167591132189)*x[2]
        s[0,4,1,0]=(-0.68812275824)*x[0]**o+(0.973257165291)*x[0]+(0.714426301535)*x[1]**o+(0.913858401922)*x[1]+(0.59761682967)*x[2]**o+(-0.0363844534093)*x[2]
        s[0,4,1,1]=(0.479806493127)*x[0]**o+(-0.345277338487)*x[0]+(-0.0440371196931)*x[1]**o+(-0.393339132322)*x[1]+(0.0828262968561)*x[2]**o+(-0.492220459324)*x[2]
        s[0,4,2,0]=(-0.516900343789)*x[0]**o+(0.52589827237)*x[0]+(0.0897613840474)*x[1]**o+(0.0310617959058)*x[1]+(-0.0539611989165)*x[2]**o+(0.694823328714)*x[2]
        s[0,4,2,1]=(0.381866427224)*x[0]**o+(-0.136004696065)*x[0]+(0.127661553496)*x[1]**o+(-0.659183794977)*x[1]+(0.644887418078)*x[2]**o+(-0.378485836785)*x[2]
        s[1,0,0,0]=(0.533070302858)*x[0]**o+(0.0486822371539)*x[0]+(0.108020635139)*x[1]**o+(0.459684900644)*x[1]+(0.927900808007)*x[2]**o+(0.27439510523)*x[2]
        s[1,0,0,1]=(0.167370720476)*x[0]**o+(0.196355013651)*x[0]+(-0.727282731548)*x[1]**o+(0.911074385588)*x[1]+(0.0825461328514)*x[2]**o+(-0.352988731028)*x[2]
        s[1,0,1,0]=(-0.468001332309)*x[0]**o+(0.245999394647)*x[0]+(0.698629778456)*x[1]**o+(0.725916422941)*x[1]+(-0.486956268801)*x[2]**o+(-0.56567979109)*x[2]
        s[1,0,1,1]=(-0.738621292386)*x[0]**o+(0.943952843499)*x[0]+(-0.244367535526)*x[1]**o+(-0.984213076842)*x[1]+(0.429661592984)*x[2]**o+(-0.424380411926)*x[2]
        s[1,0,2,0]=(0.239251156871)*x[0]**o+(0.484710141522)*x[0]+(-0.748317757318)*x[1]**o+(0.448311080458)*x[1]+(-0.539095491323)*x[2]**o+(0.408909869581)*x[2]
        s[1,0,2,1]=(0.137854759003)*x[0]**o+(-0.683196454909)*x[0]+(-0.313962804942)*x[1]**o+(-0.72203426292)*x[1]+(-0.73815484699)*x[2]**o+(-0.363352540518)*x[2]
        s[1,1,0,0]=(-0.0285460639018)*x[0]**o+(-0.698099710448)*x[0]+(0.914032704615)*x[1]**o+(0.903165323006)*x[1]+(-0.943251579494)*x[2]**o+(0.342550252155)*x[2]
        s[1,1,0,1]=(0.467266211318)*x[0]**o+(0.274280471823)*x[0]+(-0.214518600035)*x[1]**o+(0.590589154176)*x[1]+(-0.641849036312)*x[2]**o+(-0.791498351136)*x[2]
        s[1,1,1,0]=(-0.871908333966)*x[0]**o+(0.0185255578465)*x[0]+(-0.872206753541)*x[1]**o+(-0.411775119904)*x[1]+(-0.806967989381)*x[2]**o+(0.193227663136)*x[2]
        s[1,1,1,1]=(0.952880647773)*x[0]**o+(0.611625705075)*x[0]+(0.14426391661)*x[1]**o+(0.62264475033)*x[1]+(0.0544261355307)*x[2]**o+(0.169080721619)*x[2]
        s[1,1,2,0]=(-0.734065499268)*x[0]**o+(-0.500982105171)*x[0]+(0.0974889823271)*x[1]**o+(0.920634492824)*x[1]+(0.146444953125)*x[2]**o+(0.968257204573)*x[2]
        s[1,1,2,1]=(0.210841124176)*x[0]**o+(-0.0353348787064)*x[0]+(0.704879863852)*x[1]**o+(-0.376074578373)*x[1]+(-0.562677666612)*x[2]**o+(-0.562469297446)*x[2]
        s[1,2,0,0]=(0.670790719895)*x[0]**o+(-0.568997121688)*x[0]+(0.662949640213)*x[1]**o+(0.770243690967)*x[1]+(-0.967323252654)*x[2]**o+(0.453018057272)*x[2]
        s[1,2,0,1]=(-0.724419676112)*x[0]**o+(0.437279546951)*x[0]+(0.52128041598)*x[1]**o+(-0.594484377874)*x[1]+(0.0245512079914)*x[2]**o+(-0.836825535903)*x[2]
        s[1,2,1,0]=(-0.126043799205)*x[0]**o+(0.410880843631)*x[0]+(0.509268071512)*x[1]**o+(-0.408676470752)*x[1]+(-0.241389348552)*x[2]**o+(0.21098949668)*x[2]
        s[1,2,1,1]=(0.0714381616781)*x[0]**o+(0.845308210908)*x[0]+(-0.156244949835)*x[1]**o+(-0.349790094709)*x[1]+(-0.373247740966)*x[2]**o+(0.844274382162)*x[2]
        s[1,2,2,0]=(0.392643676508)*x[0]**o+(0.379574595297)*x[0]+(-0.312714958636)*x[1]**o+(-0.697892817722)*x[1]+(0.319938083594)*x[2]**o+(0.848010553529)*x[2]
        s[1,2,2,1]=(0.0165719844498)*x[0]**o+(0.777792194538)*x[0]+(-0.581257805743)*x[1]**o+(0.576469709555)*x[1]+(0.881761997352)*x[2]**o+(-0.852836910827)*x[2]
        s[1,3,0,0]=(-0.368761641941)*x[0]**o+(-0.341981426952)*x[0]+(0.0850222818871)*x[1]**o+(-0.161499189059)*x[1]+(0.681488711691)*x[2]**o+(-0.220250578178)*x[2]
        s[1,3,0,1]=(-0.978416057456)*x[0]**o+(0.587189510525)*x[0]+(0.121078873003)*x[1]**o+(-0.499166770082)*x[1]+(-0.60948578652)*x[2]**o+(0.139721741069)*x[2]
        s[1,3,1,0]=(-0.248229483054)*x[0]**o+(0.396024076796)*x[0]+(0.955385247577)*x[1]**o+(0.0363257962548)*x[1]+(0.473586138101)*x[2]**o+(-0.978818683913)*x[2]
        s[1,3,1,1]=(-0.203828777074)*x[0]**o+(0.323222541043)*x[0]+(-0.0611217538705)*x[1]**o+(0.688466815312)*x[1]+(0.95854862113)*x[2]**o+(0.424013804278)*x[2]
        s[1,3,2,0]=(-0.734238871098)*x[0]**o+(0.987103771892)*x[0]+(-0.536241083275)*x[1]**o+(0.956192017815)*x[1]+(-0.458846574286)*x[2]**o+(0.0465550451449)*x[2]
        s[1,3,2,1]=(-0.6383215636)*x[0]**o+(-0.325193071614)*x[0]+(-0.91650917726)*x[1]**o+(0.693580745379)*x[1]+(-0.766712955542)*x[2]**o+(-0.929524610715)*x[2]
        s[1,4,0,0]=(-0.990767390655)*x[0]**o+(0.518778324025)*x[0]+(0.728020751614)*x[1]**o+(-0.755324391775)*x[1]+(0.186405771099)*x[2]**o+(0.632953433339)*x[2]
        s[1,4,0,1]=(0.592323495191)*x[0]**o+(0.409731821366)*x[0]+(0.964158523996)*x[1]**o+(0.424299771487)*x[1]+(0.660072793012)*x[2]**o+(-0.899939765137)*x[2]
        s[1,4,1,0]=(-0.303351442424)*x[0]**o+(-0.88332718624)*x[0]+(0.602724455773)*x[1]**o+(0.232452946908)*x[1]+(0.985569023425)*x[2]**o+(0.49272814355)*x[2]
        s[1,4,1,1]=(0.851768453481)*x[0]**o+(0.909492002062)*x[0]+(-0.810487634736)*x[1]**o+(0.919297425734)*x[1]+(0.909780009526)*x[2]**o+(0.369012187671)*x[2]
        s[1,4,2,0]=(-0.0531952997141)*x[0]**o+(-0.224430435352)*x[0]+(-0.578404055096)*x[1]**o+(0.33447800953)*x[1]+(-0.496336772814)*x[2]**o+(0.725685669073)*x[2]
        s[1,4,2,1]=(0.541476219067)*x[0]**o+(0.107132182409)*x[0]+(-0.335093661175)*x[1]**o+(0.227203070125)*x[1]+(0.216737562875)*x[2]**o+(-0.622943563245)*x[2]
        s[2,0,0,0]=(-0.756128419319)*x[0]**o+(0.547579725255)*x[0]+(0.687267799491)*x[1]**o+(0.819030721043)*x[1]+(-0.173876155736)*x[2]**o+(-0.0301920564367)*x[2]
        s[2,0,0,1]=(-0.0759002903983)*x[0]**o+(-0.438645875523)*x[0]+(-0.783783394278)*x[1]**o+(0.939615801765)*x[1]+(-0.0129095800279)*x[2]**o+(-0.723293627815)*x[2]
        s[2,0,1,0]=(-0.783224383717)*x[0]**o+(-0.137339879229)*x[0]+(-0.168000348627)*x[1]**o+(0.223734275088)*x[1]+(-0.582857962024)*x[2]**o+(-0.215263259255)*x[2]
        s[2,0,1,1]=(0.866554819231)*x[0]**o+(0.942792854767)*x[0]+(0.506186289419)*x[1]**o+(0.0775869106357)*x[1]+(0.877213607927)*x[2]**o+(-0.792692659916)*x[2]
        s[2,0,2,0]=(0.00917554286226)*x[0]**o+(-0.5183359551)*x[0]+(0.15698167339)*x[1]**o+(-0.489805488541)*x[1]+(-0.592166661492)*x[2]**o+(0.391650695037)*x[2]
        s[2,0,2,1]=(0.356986311456)*x[0]**o+(-0.14389994339)*x[0]+(0.838182818055)*x[1]**o+(-0.735112928546)*x[1]+(-0.0333560634355)*x[2]**o+(0.0855151256652)*x[2]
        s[2,1,0,0]=(0.357398992706)*x[0]**o+(-0.733457701676)*x[0]+(-0.142200448651)*x[1]**o+(0.844914653536)*x[1]+(-0.336954698769)*x[2]**o+(-0.296222395555)*x[2]
        s[2,1,0,1]=(-0.950414726707)*x[0]**o+(-0.134623616883)*x[0]+(-0.403232115759)*x[1]**o+(-0.539055139116)*x[1]+(-0.0298791725363)*x[2]**o+(-0.829664568373)*x[2]
        s[2,1,1,0]=(-0.656224623199)*x[0]**o+(0.0872237997616)*x[0]+(0.682387976548)*x[1]**o+(-0.225630652343)*x[1]+(-0.477261038385)*x[2]**o+(-0.57098588683)*x[2]
        s[2,1,1,1]=(-0.504013815131)*x[0]**o+(-0.0309238842436)*x[0]+(0.350436304599)*x[1]**o+(-0.891903027142)*x[1]+(-0.352433668387)*x[2]**o+(-0.603914711075)*x[2]
        s[2,1,2,0]=(0.505422678232)*x[0]**o+(-0.36241271654)*x[0]+(0.418645226594)*x[1]**o+(-0.392311515242)*x[1]+(-0.740372176332)*x[2]**o+(-0.861742253361)*x[2]
        s[2,1,2,1]=(-0.853192617592)*x[0]**o+(-0.269855659669)*x[0]+(0.439592720974)*x[1]**o+(0.62371411753)*x[1]+(0.77806877626)*x[2]**o+(0.923146838795)*x[2]
        s[2,2,0,0]=(0.228191521853)*x[0]**o+(0.286276293845)*x[0]+(0.205603167312)*x[1]**o+(-0.208236801784)*x[1]+(-0.787902086996)*x[2]**o+(0.676274514448)*x[2]
        s[2,2,0,1]=(-0.75952147535)*x[0]**o+(0.685910476453)*x[0]+(-0.945474186679)*x[1]**o+(0.257392146079)*x[1]+(-0.502240261104)*x[2]**o+(-0.446181698029)*x[2]
        s[2,2,1,0]=(-0.347627224549)*x[0]**o+(0.425329908635)*x[0]+(0.350298538346)*x[1]**o+(-0.484357417885)*x[1]+(0.538076599416)*x[2]**o+(-0.0530149980096)*x[2]
        s[2,2,1,1]=(0.0539971839295)*x[0]**o+(0.0721806780674)*x[0]+(-0.309030912724)*x[1]**o+(0.169874235759)*x[1]+(-0.433682959518)*x[2]**o+(0.813732426585)*x[2]
        s[2,2,2,0]=(0.336471441117)*x[0]**o+(-0.484255919842)*x[0]+(0.968308411862)*x[1]**o+(-0.521533322453)*x[1]+(0.633627764624)*x[2]**o+(0.13497783288)*x[2]
        s[2,2,2,1]=(-0.602946926541)*x[0]**o+(-0.67535176256)*x[0]+(-0.527632780342)*x[1]**o+(0.987852520439)*x[1]+(0.426772082556)*x[2]**o+(0.00805954755153)*x[2]
        s[2,3,0,0]=(0.834197069571)*x[0]**o+(-0.213325947671)*x[0]+(0.26098973561)*x[1]**o+(-0.892672004839)*x[1]+(0.690037689842)*x[2]**o+(-0.554911545988)*x[2]
        s[2,3,0,1]=(-0.917087293134)*x[0]**o+(-0.226247312973)*x[0]+(0.901194492616)*x[1]**o+(0.879779745124)*x[1]+(-0.501788697362)*x[2]**o+(0.62176481794)*x[2]
        s[2,3,1,0]=(-0.78132288334)*x[0]**o+(0.249383051506)*x[0]+(-0.276401042016)*x[1]**o+(0.0913832980328)*x[1]+(-0.199057369246)*x[2]**o+(0.229465279886)*x[2]
        s[2,3,1,1]=(0.33249458134)*x[0]**o+(-0.79179194291)*x[0]+(0.146436436451)*x[1]**o+(0.858879181025)*x[1]+(0.069509608477)*x[2]**o+(0.993263657)*x[2]
        s[2,3,2,0]=(0.103735002938)*x[0]**o+(-0.326548520992)*x[0]+(-0.826992179127)*x[1]**o+(-0.883356562305)*x[1]+(-0.340900003747)*x[2]**o+(-0.134490572338)*x[2]
        s[2,3,2,1]=(0.10363967847)*x[0]**o+(-0.171531371867)*x[0]+(-0.929315735292)*x[1]**o+(-0.719614719828)*x[1]+(-0.592883409628)*x[2]**o+(-0.641629610424)*x[2]
        s[2,4,0,0]=(-0.297999082535)*x[0]**o+(0.439209856881)*x[0]+(-0.751163554256)*x[1]**o+(-0.189087784579)*x[1]+(-0.795871127446)*x[2]**o+(0.130133923162)*x[2]
        s[2,4,0,1]=(0.721793867996)*x[0]**o+(-0.635599886382)*x[0]+(0.801889483297)*x[1]**o+(0.211160357461)*x[1]+(0.237091518418)*x[2]**o+(0.992344543271)*x[2]
        s[2,4,1,0]=(-0.775779912316)*x[0]**o+(-0.376074847935)*x[0]+(0.212663936842)*x[1]**o+(-0.549546184149)*x[1]+(-0.0854498425298)*x[2]**o+(0.542469273724)*x[2]
        s[2,4,1,1]=(0.363553642873)*x[0]**o+(-0.311431185265)*x[0]+(-0.375820595299)*x[1]**o+(-0.845438661463)*x[1]+(-0.478899940153)*x[2]**o+(0.0806897601294)*x[2]
        s[2,4,2,0]=(-0.489422873739)*x[0]**o+(-0.157008373165)*x[0]+(-0.267263773527)*x[1]**o+(-0.53340253927)*x[1]+(0.356782019115)*x[2]**o+(0.117093532841)*x[2]
        s[2,4,2,1]=(0.106435699048)*x[0]**o+(0.457452797294)*x[0]+(-0.640139127483)*x[1]**o+(-0.378451058366)*x[1]+(-0.232397757243)*x[2]**o+(-0.395640130256)*x[2]
        s[3,0,0,0]=(0.360807863609)*x[0]**o+(0.689413996889)*x[0]+(-0.267242935468)*x[1]**o+(0.341694875336)*x[1]+(0.955444260018)*x[2]**o+(0.940815872612)*x[2]
        s[3,0,0,1]=(0.0640653983699)*x[0]**o+(0.17954797305)*x[0]+(0.739660609779)*x[1]**o+(-0.174212756014)*x[1]+(0.190134414562)*x[2]**o+(-0.45741222058)*x[2]
        s[3,0,1,0]=(-0.642372239311)*x[0]**o+(-0.518079894531)*x[0]+(-0.843773918922)*x[1]**o+(0.407352407859)*x[1]+(0.0823899653044)*x[2]**o+(0.655867974962)*x[2]
        s[3,0,1,1]=(-0.545184505438)*x[0]**o+(-0.20749144757)*x[0]+(0.656821663662)*x[1]**o+(0.238823262425)*x[1]+(0.645114250743)*x[2]**o+(0.931193868381)*x[2]
        s[3,0,2,0]=(0.715326090131)*x[0]**o+(-0.623038006294)*x[0]+(-0.0272773549244)*x[1]**o+(0.0443002059335)*x[1]+(-0.639961942663)*x[2]**o+(-0.863662064393)*x[2]
        s[3,0,2,1]=(-0.150998418072)*x[0]**o+(-0.858320880863)*x[0]+(-0.711824804062)*x[1]**o+(-0.257945841504)*x[1]+(-0.070002743188)*x[2]**o+(-0.967136937623)*x[2]
        s[3,1,0,0]=(-0.5416631385)*x[0]**o+(-0.315276228939)*x[0]+(0.327960879182)*x[1]**o+(0.309383847119)*x[1]+(-0.531771040881)*x[2]**o+(-0.503014905853)*x[2]
        s[3,1,0,1]=(-0.93431005785)*x[0]**o+(-0.416716442301)*x[0]+(0.737336965624)*x[1]**o+(-0.676338424415)*x[1]+(-0.638729889321)*x[2]**o+(0.779664539115)*x[2]
        s[3,1,1,0]=(0.994892412301)*x[0]**o+(0.0257912160747)*x[0]+(-0.829652002123)*x[1]**o+(0.412033947586)*x[1]+(0.808588738484)*x[2]**o+(-0.537100420921)*x[2]
        s[3,1,1,1]=(-0.0459132262405)*x[0]**o+(-0.517692240206)*x[0]+(0.66461361799)*x[1]**o+(0.0202376762776)*x[1]+(0.561286210305)*x[2]**o+(-0.462378162744)*x[2]
        s[3,1,2,0]=(0.909401279733)*x[0]**o+(-0.092132877706)*x[0]+(-0.224813536034)*x[1]**o+(0.318594065478)*x[1]+(-0.959865958362)*x[2]**o+(0.458327172218)*x[2]
        s[3,1,2,1]=(0.174314370931)*x[0]**o+(-0.0149495584725)*x[0]+(0.719658326164)*x[1]**o+(0.416208781614)*x[1]+(0.922363382172)*x[2]**o+(-0.123220562614)*x[2]
        s[3,2,0,0]=(-0.362407421114)*x[0]**o+(-0.35287313685)*x[0]+(-0.712118894359)*x[1]**o+(-0.749389429473)*x[1]+(-0.520706676442)*x[2]**o+(0.15576625076)*x[2]
        s[3,2,0,1]=(-0.620594009087)*x[0]**o+(-0.866479846742)*x[0]+(-0.429532862455)*x[1]**o+(-0.736262913146)*x[1]+(0.14914428425)*x[2]**o+(0.00655098493102)*x[2]
        s[3,2,1,0]=(0.47291961358)*x[0]**o+(-0.223022599565)*x[0]+(-0.489191311572)*x[1]**o+(-0.616554022459)*x[1]+(0.525481658641)*x[2]**o+(0.999870533503)*x[2]
        s[3,2,1,1]=(0.315715505936)*x[0]**o+(-0.640835348385)*x[0]+(0.768958709128)*x[1]**o+(0.587161832357)*x[1]+(0.454427338357)*x[2]**o+(-0.133391799438)*x[2]
        s[3,2,2,0]=(-0.273381601564)*x[0]**o+(-0.527114000062)*x[0]+(-0.241611283455)*x[1]**o+(0.511784453537)*x[1]+(-0.558176752993)*x[2]**o+(0.563602029554)*x[2]
        s[3,2,2,1]=(-0.554552149574)*x[0]**o+(0.277189312969)*x[0]+(0.528236290321)*x[1]**o+(-0.56161965546)*x[1]+(0.186320859613)*x[2]**o+(-0.462794918456)*x[2]
        s[3,3,0,0]=(-0.379669006639)*x[0]**o+(-0.290542994989)*x[0]+(0.950886714589)*x[1]**o+(0.113833056597)*x[1]+(-0.395921494129)*x[2]**o+(-0.530197898233)*x[2]
        s[3,3,0,1]=(-0.37250317146)*x[0]**o+(0.266496776367)*x[0]+(-0.992301806744)*x[1]**o+(-0.215688062569)*x[1]+(0.86298623844)*x[2]**o+(0.932822343278)*x[2]
        s[3,3,1,0]=(-0.354585401042)*x[0]**o+(0.472315228214)*x[0]+(-0.604904767558)*x[1]**o+(-0.868411440101)*x[1]+(-0.434629251045)*x[2]**o+(-0.4582356693)*x[2]
        s[3,3,1,1]=(-0.931569757365)*x[0]**o+(0.062290982568)*x[0]+(-0.402530852638)*x[1]**o+(0.13606876168)*x[1]+(-0.421824162622)*x[2]**o+(0.754078648605)*x[2]
        s[3,3,2,0]=(0.370446872991)*x[0]**o+(-0.678219724124)*x[0]+(0.208629662303)*x[1]**o+(0.742029980811)*x[1]+(0.311836736294)*x[2]**o+(-0.102558272951)*x[2]
        s[3,3,2,1]=(-0.15347338298)*x[0]**o+(-0.798719756084)*x[0]+(0.884663004721)*x[1]**o+(-0.658935444548)*x[1]+(-0.721604057065)*x[2]**o+(0.299122598053)*x[2]
        s[3,4,0,0]=(0.544627747607)*x[0]**o+(0.513489478068)*x[0]+(-0.410347246011)*x[1]**o+(0.610018445392)*x[1]+(0.87473885042)*x[2]**o+(0.914293254335)*x[2]
        s[3,4,0,1]=(-0.236207879972)*x[0]**o+(0.46233924924)*x[0]+(-0.472291476346)*x[1]**o+(0.480397963815)*x[1]+(-0.976610220911)*x[2]**o+(-0.442554954343)*x[2]
        s[3,4,1,0]=(0.332927708298)*x[0]**o+(0.737580232392)*x[0]+(-0.56615273435)*x[1]**o+(0.487228009962)*x[1]+(-0.620250989525)*x[2]**o+(0.699377689126)*x[2]
        s[3,4,1,1]=(-0.574680118107)*x[0]**o+(-0.645001786533)*x[0]+(-0.0623752350971)*x[1]**o+(-0.943489194832)*x[1]+(0.690135931129)*x[2]**o+(0.047173398961)*x[2]
        s[3,4,2,0]=(0.218423477936)*x[0]**o+(0.892507556482)*x[0]+(-0.621739148946)*x[1]**o+(-0.253753368776)*x[1]+(-0.88306720835)*x[2]**o+(0.0426804971625)*x[2]
        s[3,4,2,1]=(-0.901566218988)*x[0]**o+(-0.0843229102423)*x[0]+(-0.206812141605)*x[1]**o+(-0.321102598716)*x[1]+(-0.873774065466)*x[2]**o+(-0.0507766082592)*x[2]
        ref[0,0,0,0]=(-0.183051403523)*x_ref[0]**o+(0.675163647622)*x_ref[0]+(0.383724078835)*x_ref[1]**o+(-0.863174107932)*x_ref[1]+(0.752053635179)*x_ref[2]**o+(-0.442218538219)*x_ref[2]
        ref[0,0,0,1]=(0.145022844971)*x_ref[0]**o+(-0.390513961466)*x_ref[0]+(0.112480666084)*x_ref[1]**o+(-0.484118377018)*x_ref[1]+(0.804050685467)*x_ref[2]**o+(-0.278608094628)*x_ref[2]
        ref[0,0,1,0]=(0.897135266304)*x_ref[0]**o+(-0.761774008569)*x_ref[0]+(-0.997851989059)*x_ref[1]**o+(0.658068087366)*x_ref[1]+(-0.80701588118)*x_ref[2]**o+(0.596880821142)*x_ref[2]
        ref[0,0,1,1]=(-0.706353681044)*x_ref[0]**o+(-0.19299984915)*x_ref[0]+(0.0183270582241)*x_ref[1]**o+(0.0848075866792)*x_ref[1]+(-0.380936183304)*x_ref[2]**o+(0.404471922242)*x_ref[2]
        ref[0,0,2,0]=(-0.382331572203)*x_ref[0]**o+(-0.862239031672)*x_ref[0]+(-0.664876503393)*x_ref[1]**o+(0.721170780428)*x_ref[1]+(0.936933909848)*x_ref[2]**o+(0.553724430131)*x_ref[2]
        ref[0,0,2,1]=(0.41570549153)*x_ref[0]**o+(-0.668260289502)*x_ref[0]+(-0.118888455714)*x_ref[1]**o+(0.454665663519)*x_ref[1]+(-0.470911434781)*x_ref[2]**o+(-0.677230009093)*x_ref[2]
        ref[0,1,0,0]=(-0.246347502133)*x_ref[0]**o+(0.124690136148)*x_ref[0]+(0.257467906886)*x_ref[1]**o+(0.393404860933)*x_ref[1]+(-0.216777928803)*x_ref[2]**o+(-0.893375700242)*x_ref[2]
        ref[0,1,0,1]=(0.853598016536)*x_ref[0]**o+(-0.614278282211)*x_ref[0]+(0.958798548557)*x_ref[1]**o+(0.539228268539)*x_ref[1]+(-0.893749389287)*x_ref[2]**o+(-0.421872187434)*x_ref[2]
        ref[0,1,1,0]=(-0.554158393064)*x_ref[0]**o+(-0.0302602851801)*x_ref[0]+(-0.627598510228)*x_ref[1]**o+(-0.14419255857)*x_ref[1]+(-0.319740995473)*x_ref[2]**o+(-0.829801924596)*x_ref[2]
        ref[0,1,1,1]=(0.736629752286)*x_ref[0]**o+(0.104716405421)*x_ref[0]+(-0.340008364599)*x_ref[1]**o+(-0.276509803031)*x_ref[1]+(-0.65815905914)*x_ref[2]**o+(-0.862207388004)*x_ref[2]
        ref[0,1,2,0]=(-0.963285485998)*x_ref[0]**o+(-0.932861475962)*x_ref[0]+(0.0893861922413)*x_ref[1]**o+(-0.284311111286)*x_ref[1]+(-0.549298067285)*x_ref[2]**o+(0.990727028557)*x_ref[2]
        ref[0,1,2,1]=(0.354221245392)*x_ref[0]**o+(-0.259651087945)*x_ref[0]+(0.386589396916)*x_ref[1]**o+(-0.315236532701)*x_ref[1]+(0.0276667266652)*x_ref[2]**o+(0.462795309548)*x_ref[2]
        ref[0,2,0,0]=(-0.488345436139)*x_ref[0]**o+(0.0812145763632)*x_ref[0]+(-0.766180568398)*x_ref[1]**o+(0.52489429503)*x_ref[1]+(-0.369663190972)*x_ref[2]**o+(0.214012502771)*x_ref[2]
        ref[0,2,0,1]=(-0.447889959663)*x_ref[0]**o+(0.815971127504)*x_ref[0]+(-0.43620764595)*x_ref[1]**o+(0.974646962464)*x_ref[1]+(-0.0947298803439)*x_ref[2]**o+(0.57765513293)*x_ref[2]
        ref[0,2,1,0]=(0.90082728964)*x_ref[0]**o+(-0.471148200008)*x_ref[0]+(0.130856999133)*x_ref[1]**o+(0.94489744998)*x_ref[1]+(0.97494516676)*x_ref[2]**o+(0.379281707318)*x_ref[2]
        ref[0,2,1,1]=(-0.741008766911)*x_ref[0]**o+(0.125378297796)*x_ref[0]+(0.196554106007)*x_ref[1]**o+(0.67968014258)*x_ref[1]+(-0.22331716546)*x_ref[2]**o+(0.69307086692)*x_ref[2]
        ref[0,2,2,0]=(0.731935335415)*x_ref[0]**o+(0.203487322565)*x_ref[0]+(-0.824149733689)*x_ref[1]**o+(-0.765020086954)*x_ref[1]+(0.652715725055)*x_ref[2]**o+(-0.731883148625)*x_ref[2]
        ref[0,2,2,1]=(-0.845146208295)*x_ref[0]**o+(-0.226473926648)*x_ref[0]+(0.683088958387)*x_ref[1]**o+(0.588656694973)*x_ref[1]+(0.223845818469)*x_ref[2]**o+(-0.311691745089)*x_ref[2]
        ref[0,3,0,0]=(-0.390979632965)*x_ref[0]**o+(0.712430823379)*x_ref[0]+(0.250097638139)*x_ref[1]**o+(-0.0573974192609)*x_ref[1]+(-0.104674518818)*x_ref[2]**o+(-0.19187273988)*x_ref[2]
        ref[0,3,0,1]=(-0.422704759794)*x_ref[0]**o+(-0.356438670304)*x_ref[0]+(-0.99949139973)*x_ref[1]**o+(-0.923501978496)*x_ref[1]+(0.95094078991)*x_ref[2]**o+(-0.726323072853)*x_ref[2]
        ref[0,3,1,0]=(-0.752686683053)*x_ref[0]**o+(0.426969213595)*x_ref[0]+(-0.884811980577)*x_ref[1]**o+(-0.248298773095)*x_ref[1]+(-0.248150259844)*x_ref[2]**o+(0.307320547671)*x_ref[2]
        ref[0,3,1,1]=(0.444548575453)*x_ref[0]**o+(-0.425971241634)*x_ref[0]+(0.291461522783)*x_ref[1]**o+(0.7897751016)*x_ref[1]+(-0.7201321357)*x_ref[2]**o+(0.335574613087)*x_ref[2]
        ref[0,3,2,0]=(0.320923109599)*x_ref[0]**o+(0.768023804425)*x_ref[0]+(0.786842381217)*x_ref[1]**o+(0.508810585491)*x_ref[1]+(0.565569549043)*x_ref[2]**o+(-0.543149420951)*x_ref[2]
        ref[0,3,2,1]=(-0.434748955301)*x_ref[0]**o+(0.753997363265)*x_ref[0]+(-0.998212958853)*x_ref[1]**o+(0.308236036581)*x_ref[1]+(-0.604346477377)*x_ref[2]**o+(0.0456731524173)*x_ref[2]
        ref[0,4,0,0]=(0.19259119938)*x_ref[0]**o+(-0.676730834051)*x_ref[0]+(0.782503706074)*x_ref[1]**o+(-0.356602170048)*x_ref[1]+(0.242151748383)*x_ref[2]**o+(-0.555208960512)*x_ref[2]
        ref[0,4,0,1]=(-0.899694378162)*x_ref[0]**o+(-0.998178345345)*x_ref[0]+(-0.0299332693574)*x_ref[1]**o+(-0.524636203176)*x_ref[1]+(-0.090669349917)*x_ref[2]**o+(-0.167591132189)*x_ref[2]
        ref[0,4,1,0]=(-0.68812275824)*x_ref[0]**o+(0.973257165291)*x_ref[0]+(0.714426301535)*x_ref[1]**o+(0.913858401922)*x_ref[1]+(0.59761682967)*x_ref[2]**o+(-0.0363844534093)*x_ref[2]
        ref[0,4,1,1]=(0.479806493127)*x_ref[0]**o+(-0.345277338487)*x_ref[0]+(-0.0440371196931)*x_ref[1]**o+(-0.393339132322)*x_ref[1]+(0.0828262968561)*x_ref[2]**o+(-0.492220459324)*x_ref[2]
        ref[0,4,2,0]=(-0.516900343789)*x_ref[0]**o+(0.52589827237)*x_ref[0]+(0.0897613840474)*x_ref[1]**o+(0.0310617959058)*x_ref[1]+(-0.0539611989165)*x_ref[2]**o+(0.694823328714)*x_ref[2]
        ref[0,4,2,1]=(0.381866427224)*x_ref[0]**o+(-0.136004696065)*x_ref[0]+(0.127661553496)*x_ref[1]**o+(-0.659183794977)*x_ref[1]+(0.644887418078)*x_ref[2]**o+(-0.378485836785)*x_ref[2]
        ref[1,0,0,0]=(0.533070302858)*x_ref[0]**o+(0.0486822371539)*x_ref[0]+(0.108020635139)*x_ref[1]**o+(0.459684900644)*x_ref[1]+(0.927900808007)*x_ref[2]**o+(0.27439510523)*x_ref[2]
        ref[1,0,0,1]=(0.167370720476)*x_ref[0]**o+(0.196355013651)*x_ref[0]+(-0.727282731548)*x_ref[1]**o+(0.911074385588)*x_ref[1]+(0.0825461328514)*x_ref[2]**o+(-0.352988731028)*x_ref[2]
        ref[1,0,1,0]=(-0.468001332309)*x_ref[0]**o+(0.245999394647)*x_ref[0]+(0.698629778456)*x_ref[1]**o+(0.725916422941)*x_ref[1]+(-0.486956268801)*x_ref[2]**o+(-0.56567979109)*x_ref[2]
        ref[1,0,1,1]=(-0.738621292386)*x_ref[0]**o+(0.943952843499)*x_ref[0]+(-0.244367535526)*x_ref[1]**o+(-0.984213076842)*x_ref[1]+(0.429661592984)*x_ref[2]**o+(-0.424380411926)*x_ref[2]
        ref[1,0,2,0]=(0.239251156871)*x_ref[0]**o+(0.484710141522)*x_ref[0]+(-0.748317757318)*x_ref[1]**o+(0.448311080458)*x_ref[1]+(-0.539095491323)*x_ref[2]**o+(0.408909869581)*x_ref[2]
        ref[1,0,2,1]=(0.137854759003)*x_ref[0]**o+(-0.683196454909)*x_ref[0]+(-0.313962804942)*x_ref[1]**o+(-0.72203426292)*x_ref[1]+(-0.73815484699)*x_ref[2]**o+(-0.363352540518)*x_ref[2]
        ref[1,1,0,0]=(-0.0285460639018)*x_ref[0]**o+(-0.698099710448)*x_ref[0]+(0.914032704615)*x_ref[1]**o+(0.903165323006)*x_ref[1]+(-0.943251579494)*x_ref[2]**o+(0.342550252155)*x_ref[2]
        ref[1,1,0,1]=(0.467266211318)*x_ref[0]**o+(0.274280471823)*x_ref[0]+(-0.214518600035)*x_ref[1]**o+(0.590589154176)*x_ref[1]+(-0.641849036312)*x_ref[2]**o+(-0.791498351136)*x_ref[2]
        ref[1,1,1,0]=(-0.871908333966)*x_ref[0]**o+(0.0185255578465)*x_ref[0]+(-0.872206753541)*x_ref[1]**o+(-0.411775119904)*x_ref[1]+(-0.806967989381)*x_ref[2]**o+(0.193227663136)*x_ref[2]
        ref[1,1,1,1]=(0.952880647773)*x_ref[0]**o+(0.611625705075)*x_ref[0]+(0.14426391661)*x_ref[1]**o+(0.62264475033)*x_ref[1]+(0.0544261355307)*x_ref[2]**o+(0.169080721619)*x_ref[2]
        ref[1,1,2,0]=(-0.734065499268)*x_ref[0]**o+(-0.500982105171)*x_ref[0]+(0.0974889823271)*x_ref[1]**o+(0.920634492824)*x_ref[1]+(0.146444953125)*x_ref[2]**o+(0.968257204573)*x_ref[2]
        ref[1,1,2,1]=(0.210841124176)*x_ref[0]**o+(-0.0353348787064)*x_ref[0]+(0.704879863852)*x_ref[1]**o+(-0.376074578373)*x_ref[1]+(-0.562677666612)*x_ref[2]**o+(-0.562469297446)*x_ref[2]
        ref[1,2,0,0]=(0.670790719895)*x_ref[0]**o+(-0.568997121688)*x_ref[0]+(0.662949640213)*x_ref[1]**o+(0.770243690967)*x_ref[1]+(-0.967323252654)*x_ref[2]**o+(0.453018057272)*x_ref[2]
        ref[1,2,0,1]=(-0.724419676112)*x_ref[0]**o+(0.437279546951)*x_ref[0]+(0.52128041598)*x_ref[1]**o+(-0.594484377874)*x_ref[1]+(0.0245512079914)*x_ref[2]**o+(-0.836825535903)*x_ref[2]
        ref[1,2,1,0]=(-0.126043799205)*x_ref[0]**o+(0.410880843631)*x_ref[0]+(0.509268071512)*x_ref[1]**o+(-0.408676470752)*x_ref[1]+(-0.241389348552)*x_ref[2]**o+(0.21098949668)*x_ref[2]
        ref[1,2,1,1]=(0.0714381616781)*x_ref[0]**o+(0.845308210908)*x_ref[0]+(-0.156244949835)*x_ref[1]**o+(-0.349790094709)*x_ref[1]+(-0.373247740966)*x_ref[2]**o+(0.844274382162)*x_ref[2]
        ref[1,2,2,0]=(0.392643676508)*x_ref[0]**o+(0.379574595297)*x_ref[0]+(-0.312714958636)*x_ref[1]**o+(-0.697892817722)*x_ref[1]+(0.319938083594)*x_ref[2]**o+(0.848010553529)*x_ref[2]
        ref[1,2,2,1]=(0.0165719844498)*x_ref[0]**o+(0.777792194538)*x_ref[0]+(-0.581257805743)*x_ref[1]**o+(0.576469709555)*x_ref[1]+(0.881761997352)*x_ref[2]**o+(-0.852836910827)*x_ref[2]
        ref[1,3,0,0]=(-0.368761641941)*x_ref[0]**o+(-0.341981426952)*x_ref[0]+(0.0850222818871)*x_ref[1]**o+(-0.161499189059)*x_ref[1]+(0.681488711691)*x_ref[2]**o+(-0.220250578178)*x_ref[2]
        ref[1,3,0,1]=(-0.978416057456)*x_ref[0]**o+(0.587189510525)*x_ref[0]+(0.121078873003)*x_ref[1]**o+(-0.499166770082)*x_ref[1]+(-0.60948578652)*x_ref[2]**o+(0.139721741069)*x_ref[2]
        ref[1,3,1,0]=(-0.248229483054)*x_ref[0]**o+(0.396024076796)*x_ref[0]+(0.955385247577)*x_ref[1]**o+(0.0363257962548)*x_ref[1]+(0.473586138101)*x_ref[2]**o+(-0.978818683913)*x_ref[2]
        ref[1,3,1,1]=(-0.203828777074)*x_ref[0]**o+(0.323222541043)*x_ref[0]+(-0.0611217538705)*x_ref[1]**o+(0.688466815312)*x_ref[1]+(0.95854862113)*x_ref[2]**o+(0.424013804278)*x_ref[2]
        ref[1,3,2,0]=(-0.734238871098)*x_ref[0]**o+(0.987103771892)*x_ref[0]+(-0.536241083275)*x_ref[1]**o+(0.956192017815)*x_ref[1]+(-0.458846574286)*x_ref[2]**o+(0.0465550451449)*x_ref[2]
        ref[1,3,2,1]=(-0.6383215636)*x_ref[0]**o+(-0.325193071614)*x_ref[0]+(-0.91650917726)*x_ref[1]**o+(0.693580745379)*x_ref[1]+(-0.766712955542)*x_ref[2]**o+(-0.929524610715)*x_ref[2]
        ref[1,4,0,0]=(-0.990767390655)*x_ref[0]**o+(0.518778324025)*x_ref[0]+(0.728020751614)*x_ref[1]**o+(-0.755324391775)*x_ref[1]+(0.186405771099)*x_ref[2]**o+(0.632953433339)*x_ref[2]
        ref[1,4,0,1]=(0.592323495191)*x_ref[0]**o+(0.409731821366)*x_ref[0]+(0.964158523996)*x_ref[1]**o+(0.424299771487)*x_ref[1]+(0.660072793012)*x_ref[2]**o+(-0.899939765137)*x_ref[2]
        ref[1,4,1,0]=(-0.303351442424)*x_ref[0]**o+(-0.88332718624)*x_ref[0]+(0.602724455773)*x_ref[1]**o+(0.232452946908)*x_ref[1]+(0.985569023425)*x_ref[2]**o+(0.49272814355)*x_ref[2]
        ref[1,4,1,1]=(0.851768453481)*x_ref[0]**o+(0.909492002062)*x_ref[0]+(-0.810487634736)*x_ref[1]**o+(0.919297425734)*x_ref[1]+(0.909780009526)*x_ref[2]**o+(0.369012187671)*x_ref[2]
        ref[1,4,2,0]=(-0.0531952997141)*x_ref[0]**o+(-0.224430435352)*x_ref[0]+(-0.578404055096)*x_ref[1]**o+(0.33447800953)*x_ref[1]+(-0.496336772814)*x_ref[2]**o+(0.725685669073)*x_ref[2]
        ref[1,4,2,1]=(0.541476219067)*x_ref[0]**o+(0.107132182409)*x_ref[0]+(-0.335093661175)*x_ref[1]**o+(0.227203070125)*x_ref[1]+(0.216737562875)*x_ref[2]**o+(-0.622943563245)*x_ref[2]
        ref[2,0,0,0]=(-0.756128419319)*x_ref[0]**o+(0.547579725255)*x_ref[0]+(0.687267799491)*x_ref[1]**o+(0.819030721043)*x_ref[1]+(-0.173876155736)*x_ref[2]**o+(-0.0301920564367)*x_ref[2]
        ref[2,0,0,1]=(-0.0759002903983)*x_ref[0]**o+(-0.438645875523)*x_ref[0]+(-0.783783394278)*x_ref[1]**o+(0.939615801765)*x_ref[1]+(-0.0129095800279)*x_ref[2]**o+(-0.723293627815)*x_ref[2]
        ref[2,0,1,0]=(-0.783224383717)*x_ref[0]**o+(-0.137339879229)*x_ref[0]+(-0.168000348627)*x_ref[1]**o+(0.223734275088)*x_ref[1]+(-0.582857962024)*x_ref[2]**o+(-0.215263259255)*x_ref[2]
        ref[2,0,1,1]=(0.866554819231)*x_ref[0]**o+(0.942792854767)*x_ref[0]+(0.506186289419)*x_ref[1]**o+(0.0775869106357)*x_ref[1]+(0.877213607927)*x_ref[2]**o+(-0.792692659916)*x_ref[2]
        ref[2,0,2,0]=(0.00917554286226)*x_ref[0]**o+(-0.5183359551)*x_ref[0]+(0.15698167339)*x_ref[1]**o+(-0.489805488541)*x_ref[1]+(-0.592166661492)*x_ref[2]**o+(0.391650695037)*x_ref[2]
        ref[2,0,2,1]=(0.356986311456)*x_ref[0]**o+(-0.14389994339)*x_ref[0]+(0.838182818055)*x_ref[1]**o+(-0.735112928546)*x_ref[1]+(-0.0333560634355)*x_ref[2]**o+(0.0855151256652)*x_ref[2]
        ref[2,1,0,0]=(0.357398992706)*x_ref[0]**o+(-0.733457701676)*x_ref[0]+(-0.142200448651)*x_ref[1]**o+(0.844914653536)*x_ref[1]+(-0.336954698769)*x_ref[2]**o+(-0.296222395555)*x_ref[2]
        ref[2,1,0,1]=(-0.950414726707)*x_ref[0]**o+(-0.134623616883)*x_ref[0]+(-0.403232115759)*x_ref[1]**o+(-0.539055139116)*x_ref[1]+(-0.0298791725363)*x_ref[2]**o+(-0.829664568373)*x_ref[2]
        ref[2,1,1,0]=(-0.656224623199)*x_ref[0]**o+(0.0872237997616)*x_ref[0]+(0.682387976548)*x_ref[1]**o+(-0.225630652343)*x_ref[1]+(-0.477261038385)*x_ref[2]**o+(-0.57098588683)*x_ref[2]
        ref[2,1,1,1]=(-0.504013815131)*x_ref[0]**o+(-0.0309238842436)*x_ref[0]+(0.350436304599)*x_ref[1]**o+(-0.891903027142)*x_ref[1]+(-0.352433668387)*x_ref[2]**o+(-0.603914711075)*x_ref[2]
        ref[2,1,2,0]=(0.505422678232)*x_ref[0]**o+(-0.36241271654)*x_ref[0]+(0.418645226594)*x_ref[1]**o+(-0.392311515242)*x_ref[1]+(-0.740372176332)*x_ref[2]**o+(-0.861742253361)*x_ref[2]
        ref[2,1,2,1]=(-0.853192617592)*x_ref[0]**o+(-0.269855659669)*x_ref[0]+(0.439592720974)*x_ref[1]**o+(0.62371411753)*x_ref[1]+(0.77806877626)*x_ref[2]**o+(0.923146838795)*x_ref[2]
        ref[2,2,0,0]=(0.228191521853)*x_ref[0]**o+(0.286276293845)*x_ref[0]+(0.205603167312)*x_ref[1]**o+(-0.208236801784)*x_ref[1]+(-0.787902086996)*x_ref[2]**o+(0.676274514448)*x_ref[2]
        ref[2,2,0,1]=(-0.75952147535)*x_ref[0]**o+(0.685910476453)*x_ref[0]+(-0.945474186679)*x_ref[1]**o+(0.257392146079)*x_ref[1]+(-0.502240261104)*x_ref[2]**o+(-0.446181698029)*x_ref[2]
        ref[2,2,1,0]=(-0.347627224549)*x_ref[0]**o+(0.425329908635)*x_ref[0]+(0.350298538346)*x_ref[1]**o+(-0.484357417885)*x_ref[1]+(0.538076599416)*x_ref[2]**o+(-0.0530149980096)*x_ref[2]
        ref[2,2,1,1]=(0.0539971839295)*x_ref[0]**o+(0.0721806780674)*x_ref[0]+(-0.309030912724)*x_ref[1]**o+(0.169874235759)*x_ref[1]+(-0.433682959518)*x_ref[2]**o+(0.813732426585)*x_ref[2]
        ref[2,2,2,0]=(0.336471441117)*x_ref[0]**o+(-0.484255919842)*x_ref[0]+(0.968308411862)*x_ref[1]**o+(-0.521533322453)*x_ref[1]+(0.633627764624)*x_ref[2]**o+(0.13497783288)*x_ref[2]
        ref[2,2,2,1]=(-0.602946926541)*x_ref[0]**o+(-0.67535176256)*x_ref[0]+(-0.527632780342)*x_ref[1]**o+(0.987852520439)*x_ref[1]+(0.426772082556)*x_ref[2]**o+(0.00805954755153)*x_ref[2]
        ref[2,3,0,0]=(0.834197069571)*x_ref[0]**o+(-0.213325947671)*x_ref[0]+(0.26098973561)*x_ref[1]**o+(-0.892672004839)*x_ref[1]+(0.690037689842)*x_ref[2]**o+(-0.554911545988)*x_ref[2]
        ref[2,3,0,1]=(-0.917087293134)*x_ref[0]**o+(-0.226247312973)*x_ref[0]+(0.901194492616)*x_ref[1]**o+(0.879779745124)*x_ref[1]+(-0.501788697362)*x_ref[2]**o+(0.62176481794)*x_ref[2]
        ref[2,3,1,0]=(-0.78132288334)*x_ref[0]**o+(0.249383051506)*x_ref[0]+(-0.276401042016)*x_ref[1]**o+(0.0913832980328)*x_ref[1]+(-0.199057369246)*x_ref[2]**o+(0.229465279886)*x_ref[2]
        ref[2,3,1,1]=(0.33249458134)*x_ref[0]**o+(-0.79179194291)*x_ref[0]+(0.146436436451)*x_ref[1]**o+(0.858879181025)*x_ref[1]+(0.069509608477)*x_ref[2]**o+(0.993263657)*x_ref[2]
        ref[2,3,2,0]=(0.103735002938)*x_ref[0]**o+(-0.326548520992)*x_ref[0]+(-0.826992179127)*x_ref[1]**o+(-0.883356562305)*x_ref[1]+(-0.340900003747)*x_ref[2]**o+(-0.134490572338)*x_ref[2]
        ref[2,3,2,1]=(0.10363967847)*x_ref[0]**o+(-0.171531371867)*x_ref[0]+(-0.929315735292)*x_ref[1]**o+(-0.719614719828)*x_ref[1]+(-0.592883409628)*x_ref[2]**o+(-0.641629610424)*x_ref[2]
        ref[2,4,0,0]=(-0.297999082535)*x_ref[0]**o+(0.439209856881)*x_ref[0]+(-0.751163554256)*x_ref[1]**o+(-0.189087784579)*x_ref[1]+(-0.795871127446)*x_ref[2]**o+(0.130133923162)*x_ref[2]
        ref[2,4,0,1]=(0.721793867996)*x_ref[0]**o+(-0.635599886382)*x_ref[0]+(0.801889483297)*x_ref[1]**o+(0.211160357461)*x_ref[1]+(0.237091518418)*x_ref[2]**o+(0.992344543271)*x_ref[2]
        ref[2,4,1,0]=(-0.775779912316)*x_ref[0]**o+(-0.376074847935)*x_ref[0]+(0.212663936842)*x_ref[1]**o+(-0.549546184149)*x_ref[1]+(-0.0854498425298)*x_ref[2]**o+(0.542469273724)*x_ref[2]
        ref[2,4,1,1]=(0.363553642873)*x_ref[0]**o+(-0.311431185265)*x_ref[0]+(-0.375820595299)*x_ref[1]**o+(-0.845438661463)*x_ref[1]+(-0.478899940153)*x_ref[2]**o+(0.0806897601294)*x_ref[2]
        ref[2,4,2,0]=(-0.489422873739)*x_ref[0]**o+(-0.157008373165)*x_ref[0]+(-0.267263773527)*x_ref[1]**o+(-0.53340253927)*x_ref[1]+(0.356782019115)*x_ref[2]**o+(0.117093532841)*x_ref[2]
        ref[2,4,2,1]=(0.106435699048)*x_ref[0]**o+(0.457452797294)*x_ref[0]+(-0.640139127483)*x_ref[1]**o+(-0.378451058366)*x_ref[1]+(-0.232397757243)*x_ref[2]**o+(-0.395640130256)*x_ref[2]
        ref[3,0,0,0]=(0.360807863609)*x_ref[0]**o+(0.689413996889)*x_ref[0]+(-0.267242935468)*x_ref[1]**o+(0.341694875336)*x_ref[1]+(0.955444260018)*x_ref[2]**o+(0.940815872612)*x_ref[2]
        ref[3,0,0,1]=(0.0640653983699)*x_ref[0]**o+(0.17954797305)*x_ref[0]+(0.739660609779)*x_ref[1]**o+(-0.174212756014)*x_ref[1]+(0.190134414562)*x_ref[2]**o+(-0.45741222058)*x_ref[2]
        ref[3,0,1,0]=(-0.642372239311)*x_ref[0]**o+(-0.518079894531)*x_ref[0]+(-0.843773918922)*x_ref[1]**o+(0.407352407859)*x_ref[1]+(0.0823899653044)*x_ref[2]**o+(0.655867974962)*x_ref[2]
        ref[3,0,1,1]=(-0.545184505438)*x_ref[0]**o+(-0.20749144757)*x_ref[0]+(0.656821663662)*x_ref[1]**o+(0.238823262425)*x_ref[1]+(0.645114250743)*x_ref[2]**o+(0.931193868381)*x_ref[2]
        ref[3,0,2,0]=(0.715326090131)*x_ref[0]**o+(-0.623038006294)*x_ref[0]+(-0.0272773549244)*x_ref[1]**o+(0.0443002059335)*x_ref[1]+(-0.639961942663)*x_ref[2]**o+(-0.863662064393)*x_ref[2]
        ref[3,0,2,1]=(-0.150998418072)*x_ref[0]**o+(-0.858320880863)*x_ref[0]+(-0.711824804062)*x_ref[1]**o+(-0.257945841504)*x_ref[1]+(-0.070002743188)*x_ref[2]**o+(-0.967136937623)*x_ref[2]
        ref[3,1,0,0]=(-0.5416631385)*x_ref[0]**o+(-0.315276228939)*x_ref[0]+(0.327960879182)*x_ref[1]**o+(0.309383847119)*x_ref[1]+(-0.531771040881)*x_ref[2]**o+(-0.503014905853)*x_ref[2]
        ref[3,1,0,1]=(-0.93431005785)*x_ref[0]**o+(-0.416716442301)*x_ref[0]+(0.737336965624)*x_ref[1]**o+(-0.676338424415)*x_ref[1]+(-0.638729889321)*x_ref[2]**o+(0.779664539115)*x_ref[2]
        ref[3,1,1,0]=(0.994892412301)*x_ref[0]**o+(0.0257912160747)*x_ref[0]+(-0.829652002123)*x_ref[1]**o+(0.412033947586)*x_ref[1]+(0.808588738484)*x_ref[2]**o+(-0.537100420921)*x_ref[2]
        ref[3,1,1,1]=(-0.0459132262405)*x_ref[0]**o+(-0.517692240206)*x_ref[0]+(0.66461361799)*x_ref[1]**o+(0.0202376762776)*x_ref[1]+(0.561286210305)*x_ref[2]**o+(-0.462378162744)*x_ref[2]
        ref[3,1,2,0]=(0.909401279733)*x_ref[0]**o+(-0.092132877706)*x_ref[0]+(-0.224813536034)*x_ref[1]**o+(0.318594065478)*x_ref[1]+(-0.959865958362)*x_ref[2]**o+(0.458327172218)*x_ref[2]
        ref[3,1,2,1]=(0.174314370931)*x_ref[0]**o+(-0.0149495584725)*x_ref[0]+(0.719658326164)*x_ref[1]**o+(0.416208781614)*x_ref[1]+(0.922363382172)*x_ref[2]**o+(-0.123220562614)*x_ref[2]
        ref[3,2,0,0]=(-0.362407421114)*x_ref[0]**o+(-0.35287313685)*x_ref[0]+(-0.712118894359)*x_ref[1]**o+(-0.749389429473)*x_ref[1]+(-0.520706676442)*x_ref[2]**o+(0.15576625076)*x_ref[2]
        ref[3,2,0,1]=(-0.620594009087)*x_ref[0]**o+(-0.866479846742)*x_ref[0]+(-0.429532862455)*x_ref[1]**o+(-0.736262913146)*x_ref[1]+(0.14914428425)*x_ref[2]**o+(0.00655098493102)*x_ref[2]
        ref[3,2,1,0]=(0.47291961358)*x_ref[0]**o+(-0.223022599565)*x_ref[0]+(-0.489191311572)*x_ref[1]**o+(-0.616554022459)*x_ref[1]+(0.525481658641)*x_ref[2]**o+(0.999870533503)*x_ref[2]
        ref[3,2,1,1]=(0.315715505936)*x_ref[0]**o+(-0.640835348385)*x_ref[0]+(0.768958709128)*x_ref[1]**o+(0.587161832357)*x_ref[1]+(0.454427338357)*x_ref[2]**o+(-0.133391799438)*x_ref[2]
        ref[3,2,2,0]=(-0.273381601564)*x_ref[0]**o+(-0.527114000062)*x_ref[0]+(-0.241611283455)*x_ref[1]**o+(0.511784453537)*x_ref[1]+(-0.558176752993)*x_ref[2]**o+(0.563602029554)*x_ref[2]
        ref[3,2,2,1]=(-0.554552149574)*x_ref[0]**o+(0.277189312969)*x_ref[0]+(0.528236290321)*x_ref[1]**o+(-0.56161965546)*x_ref[1]+(0.186320859613)*x_ref[2]**o+(-0.462794918456)*x_ref[2]
        ref[3,3,0,0]=(-0.379669006639)*x_ref[0]**o+(-0.290542994989)*x_ref[0]+(0.950886714589)*x_ref[1]**o+(0.113833056597)*x_ref[1]+(-0.395921494129)*x_ref[2]**o+(-0.530197898233)*x_ref[2]
        ref[3,3,0,1]=(-0.37250317146)*x_ref[0]**o+(0.266496776367)*x_ref[0]+(-0.992301806744)*x_ref[1]**o+(-0.215688062569)*x_ref[1]+(0.86298623844)*x_ref[2]**o+(0.932822343278)*x_ref[2]
        ref[3,3,1,0]=(-0.354585401042)*x_ref[0]**o+(0.472315228214)*x_ref[0]+(-0.604904767558)*x_ref[1]**o+(-0.868411440101)*x_ref[1]+(-0.434629251045)*x_ref[2]**o+(-0.4582356693)*x_ref[2]
        ref[3,3,1,1]=(-0.931569757365)*x_ref[0]**o+(0.062290982568)*x_ref[0]+(-0.402530852638)*x_ref[1]**o+(0.13606876168)*x_ref[1]+(-0.421824162622)*x_ref[2]**o+(0.754078648605)*x_ref[2]
        ref[3,3,2,0]=(0.370446872991)*x_ref[0]**o+(-0.678219724124)*x_ref[0]+(0.208629662303)*x_ref[1]**o+(0.742029980811)*x_ref[1]+(0.311836736294)*x_ref[2]**o+(-0.102558272951)*x_ref[2]
        ref[3,3,2,1]=(-0.15347338298)*x_ref[0]**o+(-0.798719756084)*x_ref[0]+(0.884663004721)*x_ref[1]**o+(-0.658935444548)*x_ref[1]+(-0.721604057065)*x_ref[2]**o+(0.299122598053)*x_ref[2]
        ref[3,4,0,0]=(0.544627747607)*x_ref[0]**o+(0.513489478068)*x_ref[0]+(-0.410347246011)*x_ref[1]**o+(0.610018445392)*x_ref[1]+(0.87473885042)*x_ref[2]**o+(0.914293254335)*x_ref[2]
        ref[3,4,0,1]=(-0.236207879972)*x_ref[0]**o+(0.46233924924)*x_ref[0]+(-0.472291476346)*x_ref[1]**o+(0.480397963815)*x_ref[1]+(-0.976610220911)*x_ref[2]**o+(-0.442554954343)*x_ref[2]
        ref[3,4,1,0]=(0.332927708298)*x_ref[0]**o+(0.737580232392)*x_ref[0]+(-0.56615273435)*x_ref[1]**o+(0.487228009962)*x_ref[1]+(-0.620250989525)*x_ref[2]**o+(0.699377689126)*x_ref[2]
        ref[3,4,1,1]=(-0.574680118107)*x_ref[0]**o+(-0.645001786533)*x_ref[0]+(-0.0623752350971)*x_ref[1]**o+(-0.943489194832)*x_ref[1]+(0.690135931129)*x_ref[2]**o+(0.047173398961)*x_ref[2]
        ref[3,4,2,0]=(0.218423477936)*x_ref[0]**o+(0.892507556482)*x_ref[0]+(-0.621739148946)*x_ref[1]**o+(-0.253753368776)*x_ref[1]+(-0.88306720835)*x_ref[2]**o+(0.0426804971625)*x_ref[2]
        ref[3,4,2,1]=(-0.901566218988)*x_ref[0]**o+(-0.0843229102423)*x_ref[0]+(-0.206812141605)*x_ref[1]**o+(-0.321102598716)*x_ref[1]+(-0.873774065466)*x_ref[2]**o+(-0.0507766082592)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.32868090106)*x[0]+(-0.840160165417)*x[1]
        ref=(-0.32868090106)*x_ref[0]+(-0.840160165417)*x_ref[1]
      else:
        s=(-1.32732649563)*x[0]+(0.20149624757)*x[1]+(0.701715136384)*x[2]
        ref=(-1.32732649563)*x_ref[0]+(0.20149624757)*x_ref[1]+(0.701715136384)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        s[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
        ref[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        ref[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
      else:
        s[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        s[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
        ref[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        ref[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.199811526227)*x[0]+(1.25013707432)*x[1]
        s[0,1]=(0.277539536007)*x[0]+(-0.023732930558)*x[1]
        s[0,2]=(-0.244973175438)*x[0]+(-0.501902270422)*x[1]
        s[0,3]=(0.249200744572)*x[0]+(0.520843463388)*x[1]
        s[0,4]=(-0.995168115444)*x[0]+(-0.259093571207)*x[1]
        s[1,0]=(0.179469283663)*x[0]+(-0.461808204127)*x[1]
        s[1,1]=(0.431508044494)*x[0]+(-0.423277002443)*x[1]
        s[1,2]=(0.732281368811)*x[0]+(-0.490387490227)*x[1]
        s[1,3]=(-0.665998043548)*x[0]+(-0.180996501441)*x[1]
        s[1,4]=(-0.773309943626)*x[0]+(-0.27171870837)*x[1]
        s[2,0]=(0.975757439909)*x[0]+(-0.980217145831)*x[1]
        s[2,1]=(-0.898690989301)*x[0]+(0.696886429725)*x[1]
        s[2,2]=(-0.568728495444)*x[0]+(0.171459808085)*x[1]
        s[2,3]=(-0.509734570618)*x[0]+(-0.318239573518)*x[1]
        s[2,4]=(-0.868166262762)*x[0]+(0.669164551211)*x[1]
        s[3,0]=(-1.11259074191)*x[0]+(-0.317962607489)*x[1]
        s[3,1]=(-0.735838783493)*x[0]+(-0.15621592888)*x[1]
        s[3,2]=(-0.348618152308)*x[0]+(0.346356772969)*x[1]
        s[3,3]=(1.58890219057)*x[0]+(-0.106915533345)*x[1]
        s[3,4]=(-0.739997195498)*x[0]+(-0.668986698058)*x[1]
        ref[0,0]=(0.199811526227)*x_ref[0]+(1.25013707432)*x_ref[1]
        ref[0,1]=(0.277539536007)*x_ref[0]+(-0.023732930558)*x_ref[1]
        ref[0,2]=(-0.244973175438)*x_ref[0]+(-0.501902270422)*x_ref[1]
        ref[0,3]=(0.249200744572)*x_ref[0]+(0.520843463388)*x_ref[1]
        ref[0,4]=(-0.995168115444)*x_ref[0]+(-0.259093571207)*x_ref[1]
        ref[1,0]=(0.179469283663)*x_ref[0]+(-0.461808204127)*x_ref[1]
        ref[1,1]=(0.431508044494)*x_ref[0]+(-0.423277002443)*x_ref[1]
        ref[1,2]=(0.732281368811)*x_ref[0]+(-0.490387490227)*x_ref[1]
        ref[1,3]=(-0.665998043548)*x_ref[0]+(-0.180996501441)*x_ref[1]
        ref[1,4]=(-0.773309943626)*x_ref[0]+(-0.27171870837)*x_ref[1]
        ref[2,0]=(0.975757439909)*x_ref[0]+(-0.980217145831)*x_ref[1]
        ref[2,1]=(-0.898690989301)*x_ref[0]+(0.696886429725)*x_ref[1]
        ref[2,2]=(-0.568728495444)*x_ref[0]+(0.171459808085)*x_ref[1]
        ref[2,3]=(-0.509734570618)*x_ref[0]+(-0.318239573518)*x_ref[1]
        ref[2,4]=(-0.868166262762)*x_ref[0]+(0.669164551211)*x_ref[1]
        ref[3,0]=(-1.11259074191)*x_ref[0]+(-0.317962607489)*x_ref[1]
        ref[3,1]=(-0.735838783493)*x_ref[0]+(-0.15621592888)*x_ref[1]
        ref[3,2]=(-0.348618152308)*x_ref[0]+(0.346356772969)*x_ref[1]
        ref[3,3]=(1.58890219057)*x_ref[0]+(-0.106915533345)*x_ref[1]
        ref[3,4]=(-0.739997195498)*x_ref[0]+(-0.668986698058)*x_ref[1]
      else:
        s[0,0]=(1.00508479945)*x[0]+(0.671813078579)*x[1]+(0.0352167301347)*x[2]
        s[0,1]=(0.832418208523)*x[0]+(-0.228971976603)*x[1]+(0.472369173779)*x[2]
        s[0,2]=(0.495212675956)*x[0]+(-0.850840698084)*x[1]+(-0.480464122957)*x[2]
        s[0,3]=(1.09800821588)*x[0]+(-1.16790708068)*x[1]+(-0.16448878954)*x[2]
        s[0,4]=(-0.142901980481)*x[0]+(-0.456358812317)*x[1]+(0.688171420193)*x[2]
        s[1,0]=(-0.452209235863)*x[0]+(0.835019801073)*x[1]+(-0.537900952591)*x[2]
        s[1,1]=(1.24188029232)*x[0]+(0.537899388384)*x[1]+(0.825366946283)*x[2]
        s[1,2]=(0.0568959695268)*x[0]+(-0.530615837561)*x[1]+(0.208390615403)*x[2]
        s[1,3]=(0.704415982126)*x[0]+(0.557947101676)*x[1]+(-1.41348269833)*x[2]
        s[1,4]=(0.23735208213)*x[0]+(0.188600394593)*x[1]+(-0.587309525946)*x[2]
        s[2,0]=(0.202262121276)*x[0]+(1.34583368863)*x[1]+(-1.416408583)*x[2]
        s[2,1]=(-0.889020760737)*x[0]+(1.18733643427)*x[1]+(1.11796325716)*x[2]
        s[2,2]=(0.0729529752138)*x[0]+(0.467461724385)*x[1]+(-1.26467465607)*x[2]
        s[2,3]=(0.553563173427)*x[0]+(-0.96991696383)*x[1]+(1.21489979267)*x[2]
        s[2,4]=(0.504157770518)*x[0]+(-0.0966112701707)*x[1]+(0.820591191307)*x[2]
        s[3,0]=(0.438942088634)*x[0]+(0.737754158644)*x[1]+(0.327747630162)*x[2]
        s[3,1]=(-0.966999874402)*x[0]+(0.218515238902)*x[1]+(-1.85502435057)*x[2]
        s[3,2]=(0.632366690747)*x[0]+(0.710212883666)*x[1]+(1.52843521678)*x[2]
        s[3,3]=(-0.139393405525)*x[0]+(-1.020102199)*x[1]+(0.485417385052)*x[2]
        s[3,4]=(-0.753676205348)*x[0]+(0.783454715482)*x[1]+(0.568363667465)*x[2]
        ref[0,0]=(1.00508479945)*x_ref[0]+(0.671813078579)*x_ref[1]+(0.0352167301347)*x_ref[2]
        ref[0,1]=(0.832418208523)*x_ref[0]+(-0.228971976603)*x_ref[1]+(0.472369173779)*x_ref[2]
        ref[0,2]=(0.495212675956)*x_ref[0]+(-0.850840698084)*x_ref[1]+(-0.480464122957)*x_ref[2]
        ref[0,3]=(1.09800821588)*x_ref[0]+(-1.16790708068)*x_ref[1]+(-0.16448878954)*x_ref[2]
        ref[0,4]=(-0.142901980481)*x_ref[0]+(-0.456358812317)*x_ref[1]+(0.688171420193)*x_ref[2]
        ref[1,0]=(-0.452209235863)*x_ref[0]+(0.835019801073)*x_ref[1]+(-0.537900952591)*x_ref[2]
        ref[1,1]=(1.24188029232)*x_ref[0]+(0.537899388384)*x_ref[1]+(0.825366946283)*x_ref[2]
        ref[1,2]=(0.0568959695268)*x_ref[0]+(-0.530615837561)*x_ref[1]+(0.208390615403)*x_ref[2]
        ref[1,3]=(0.704415982126)*x_ref[0]+(0.557947101676)*x_ref[1]+(-1.41348269833)*x_ref[2]
        ref[1,4]=(0.23735208213)*x_ref[0]+(0.188600394593)*x_ref[1]+(-0.587309525946)*x_ref[2]
        ref[2,0]=(0.202262121276)*x_ref[0]+(1.34583368863)*x_ref[1]+(-1.416408583)*x_ref[2]
        ref[2,1]=(-0.889020760737)*x_ref[0]+(1.18733643427)*x_ref[1]+(1.11796325716)*x_ref[2]
        ref[2,2]=(0.0729529752138)*x_ref[0]+(0.467461724385)*x_ref[1]+(-1.26467465607)*x_ref[2]
        ref[2,3]=(0.553563173427)*x_ref[0]+(-0.96991696383)*x_ref[1]+(1.21489979267)*x_ref[2]
        ref[2,4]=(0.504157770518)*x_ref[0]+(-0.0966112701707)*x_ref[1]+(0.820591191307)*x_ref[2]
        ref[3,0]=(0.438942088634)*x_ref[0]+(0.737754158644)*x_ref[1]+(0.327747630162)*x_ref[2]
        ref[3,1]=(-0.966999874402)*x_ref[0]+(0.218515238902)*x_ref[1]+(-1.85502435057)*x_ref[2]
        ref[3,2]=(0.632366690747)*x_ref[0]+(0.710212883666)*x_ref[1]+(1.52843521678)*x_ref[2]
        ref[3,3]=(-0.139393405525)*x_ref[0]+(-1.020102199)*x_ref[1]+(0.485417385052)*x_ref[2]
        ref[3,4]=(-0.753676205348)*x_ref[0]+(0.783454715482)*x_ref[1]+(0.568363667465)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.734232753652)*x[0]+(0.800930015514)*x[1]
        s[0,0,1]=(-0.219864689475)*x[0]+(0.157713027162)*x[1]
        s[0,1,0]=(-0.00664406256638)*x[0]+(0.872104848272)*x[1]
        s[0,1,1]=(-1.5575275283)*x[0]+(-1.43701883839)*x[1]
        s[1,0,0]=(1.60056664833)*x[0]+(1.6215424255)*x[1]
        s[1,0,1]=(1.5172017347)*x[0]+(-0.022919888148)*x[1]
        s[1,1,0]=(0.042690381655)*x[0]+(1.30019922)*x[1]
        s[1,1,1]=(0.105731664319)*x[0]+(-1.87027128366)*x[1]
        s[2,0,0]=(1.59675688345)*x[0]+(0.0608090042054)*x[1]
        s[2,0,1]=(-1.3520682478)*x[0]+(-0.881995801453)*x[1]
        s[2,1,0]=(1.08974009958)*x[0]+(0.13217220748)*x[1]
        s[2,1,1]=(-0.631056656103)*x[0]+(-0.548103546753)*x[1]
        s[3,0,0]=(1.59082745694)*x[0]+(0.101815532645)*x[1]
        s[3,0,1]=(-0.390741362781)*x[0]+(0.525254974689)*x[1]
        s[3,1,0]=(0.439652119858)*x[0]+(1.25127795686)*x[1]
        s[3,1,1]=(-1.10496039621)*x[0]+(0.0889406484341)*x[1]
        s[4,0,0]=(1.76164121088)*x[0]+(0.790919865199)*x[1]
        s[4,0,1]=(1.53483109)*x[0]+(1.15931695117)*x[1]
        s[4,1,0]=(-1.4226648704)*x[0]+(0.0380245720655)*x[1]
        s[4,1,1]=(-0.158958015165)*x[0]+(0.359005195468)*x[1]
        s[5,0,0]=(-0.0531318028725)*x[0]+(-0.0716200483232)*x[1]
        s[5,0,1]=(-1.64720026194)*x[0]+(-1.13699910338)*x[1]
        s[5,1,0]=(1.79036772333)*x[0]+(0.414445739998)*x[1]
        s[5,1,1]=(-0.168931889239)*x[0]+(1.55913218176)*x[1]
        ref[0,0,0]=(-0.734232753652)*x_ref[0]+(0.800930015514)*x_ref[1]
        ref[0,0,1]=(-0.219864689475)*x_ref[0]+(0.157713027162)*x_ref[1]
        ref[0,1,0]=(-0.00664406256638)*x_ref[0]+(0.872104848272)*x_ref[1]
        ref[0,1,1]=(-1.5575275283)*x_ref[0]+(-1.43701883839)*x_ref[1]
        ref[1,0,0]=(1.60056664833)*x_ref[0]+(1.6215424255)*x_ref[1]
        ref[1,0,1]=(1.5172017347)*x_ref[0]+(-0.022919888148)*x_ref[1]
        ref[1,1,0]=(0.042690381655)*x_ref[0]+(1.30019922)*x_ref[1]
        ref[1,1,1]=(0.105731664319)*x_ref[0]+(-1.87027128366)*x_ref[1]
        ref[2,0,0]=(1.59675688345)*x_ref[0]+(0.0608090042054)*x_ref[1]
        ref[2,0,1]=(-1.3520682478)*x_ref[0]+(-0.881995801453)*x_ref[1]
        ref[2,1,0]=(1.08974009958)*x_ref[0]+(0.13217220748)*x_ref[1]
        ref[2,1,1]=(-0.631056656103)*x_ref[0]+(-0.548103546753)*x_ref[1]
        ref[3,0,0]=(1.59082745694)*x_ref[0]+(0.101815532645)*x_ref[1]
        ref[3,0,1]=(-0.390741362781)*x_ref[0]+(0.525254974689)*x_ref[1]
        ref[3,1,0]=(0.439652119858)*x_ref[0]+(1.25127795686)*x_ref[1]
        ref[3,1,1]=(-1.10496039621)*x_ref[0]+(0.0889406484341)*x_ref[1]
        ref[4,0,0]=(1.76164121088)*x_ref[0]+(0.790919865199)*x_ref[1]
        ref[4,0,1]=(1.53483109)*x_ref[0]+(1.15931695117)*x_ref[1]
        ref[4,1,0]=(-1.4226648704)*x_ref[0]+(0.0380245720655)*x_ref[1]
        ref[4,1,1]=(-0.158958015165)*x_ref[0]+(0.359005195468)*x_ref[1]
        ref[5,0,0]=(-0.0531318028725)*x_ref[0]+(-0.0716200483232)*x_ref[1]
        ref[5,0,1]=(-1.64720026194)*x_ref[0]+(-1.13699910338)*x_ref[1]
        ref[5,1,0]=(1.79036772333)*x_ref[0]+(0.414445739998)*x_ref[1]
        ref[5,1,1]=(-0.168931889239)*x_ref[0]+(1.55913218176)*x_ref[1]
      else:
        s[0,0,0]=(0.030748782631)*x[0]+(0.0965401372343)*x[1]+(-1.19722517556)*x[2]
        s[0,0,1]=(-1.25981749671)*x[0]+(0.375214736326)*x[1]+(0.78094551805)*x[2]
        s[0,1,0]=(0.0859894648882)*x[0]+(-1.02857522767)*x[1]+(0.752176468687)*x[2]
        s[0,1,1]=(0.857538315926)*x[0]+(1.20788107416)*x[1]+(-1.00656644081)*x[2]
        s[1,0,0]=(-0.0208770456539)*x[0]+(0.296657331936)*x[1]+(1.49455132218)*x[2]
        s[1,0,1]=(-1.1213775487)*x[0]+(1.12351512934)*x[1]+(0.661972762047)*x[2]
        s[1,1,0]=(0.75336533272)*x[0]+(1.33059115242)*x[1]+(-0.43556744027)*x[2]
        s[1,1,1]=(0.467898749534)*x[0]+(1.53277916261)*x[1]+(-1.36041296872)*x[2]
        s[2,0,0]=(0.390444800695)*x[0]+(-0.0902034979398)*x[1]+(1.79547473252)*x[2]
        s[2,0,1]=(0.0488678241039)*x[0]+(0.000207248527408)*x[1]+(-0.231246108981)*x[2]
        s[2,1,0]=(0.907328661278)*x[0]+(-0.777592463716)*x[1]+(-0.949764289407)*x[2]
        s[2,1,1]=(-0.508314624748)*x[0]+(-0.0150508206233)*x[1]+(0.786266513715)*x[2]
        s[3,0,0]=(-0.264992651778)*x[0]+(1.23980464726)*x[1]+(0.480251929476)*x[2]
        s[3,0,1]=(-0.519373168965)*x[0]+(-0.091129773203)*x[1]+(-1.12961715008)*x[2]
        s[3,1,0]=(0.737417275735)*x[0]+(0.00073881678009)*x[1]+(-0.913229240795)*x[2]
        s[3,1,1]=(-0.565248614525)*x[0]+(0.232901289197)*x[1]+(-0.852815115095)*x[2]
        s[4,0,0]=(-0.513590194423)*x[0]+(-1.10607554932)*x[1]+(1.3269295152)*x[2]
        s[4,0,1]=(-0.0108694161983)*x[0]+(-1.39154818645)*x[1]+(1.5448779549)*x[2]
        s[4,1,0]=(1.25248373617)*x[0]+(-0.444522495013)*x[1]+(0.791060985516)*x[2]
        s[4,1,1]=(0.287880876137)*x[0]+(1.42265824896)*x[1]+(-1.16312992573)*x[2]
        s[5,0,0]=(-1.25472591758)*x[0]+(-1.48986095574)*x[1]+(1.49212797538)*x[2]
        s[5,0,1]=(-0.774594904232)*x[0]+(0.582557149894)*x[1]+(-0.460088096141)*x[2]
        s[5,1,0]=(0.404584427176)*x[0]+(-0.0808051689456)*x[1]+(0.522819318812)*x[2]
        s[5,1,1]=(1.15513940465)*x[0]+(1.09285861353)*x[1]+(-0.166414260437)*x[2]
        ref[0,0,0]=(0.030748782631)*x_ref[0]+(0.0965401372343)*x_ref[1]+(-1.19722517556)*x_ref[2]
        ref[0,0,1]=(-1.25981749671)*x_ref[0]+(0.375214736326)*x_ref[1]+(0.78094551805)*x_ref[2]
        ref[0,1,0]=(0.0859894648882)*x_ref[0]+(-1.02857522767)*x_ref[1]+(0.752176468687)*x_ref[2]
        ref[0,1,1]=(0.857538315926)*x_ref[0]+(1.20788107416)*x_ref[1]+(-1.00656644081)*x_ref[2]
        ref[1,0,0]=(-0.0208770456539)*x_ref[0]+(0.296657331936)*x_ref[1]+(1.49455132218)*x_ref[2]
        ref[1,0,1]=(-1.1213775487)*x_ref[0]+(1.12351512934)*x_ref[1]+(0.661972762047)*x_ref[2]
        ref[1,1,0]=(0.75336533272)*x_ref[0]+(1.33059115242)*x_ref[1]+(-0.43556744027)*x_ref[2]
        ref[1,1,1]=(0.467898749534)*x_ref[0]+(1.53277916261)*x_ref[1]+(-1.36041296872)*x_ref[2]
        ref[2,0,0]=(0.390444800695)*x_ref[0]+(-0.0902034979398)*x_ref[1]+(1.79547473252)*x_ref[2]
        ref[2,0,1]=(0.0488678241039)*x_ref[0]+(0.000207248527408)*x_ref[1]+(-0.231246108981)*x_ref[2]
        ref[2,1,0]=(0.907328661278)*x_ref[0]+(-0.777592463716)*x_ref[1]+(-0.949764289407)*x_ref[2]
        ref[2,1,1]=(-0.508314624748)*x_ref[0]+(-0.0150508206233)*x_ref[1]+(0.786266513715)*x_ref[2]
        ref[3,0,0]=(-0.264992651778)*x_ref[0]+(1.23980464726)*x_ref[1]+(0.480251929476)*x_ref[2]
        ref[3,0,1]=(-0.519373168965)*x_ref[0]+(-0.091129773203)*x_ref[1]+(-1.12961715008)*x_ref[2]
        ref[3,1,0]=(0.737417275735)*x_ref[0]+(0.00073881678009)*x_ref[1]+(-0.913229240795)*x_ref[2]
        ref[3,1,1]=(-0.565248614525)*x_ref[0]+(0.232901289197)*x_ref[1]+(-0.852815115095)*x_ref[2]
        ref[4,0,0]=(-0.513590194423)*x_ref[0]+(-1.10607554932)*x_ref[1]+(1.3269295152)*x_ref[2]
        ref[4,0,1]=(-0.0108694161983)*x_ref[0]+(-1.39154818645)*x_ref[1]+(1.5448779549)*x_ref[2]
        ref[4,1,0]=(1.25248373617)*x_ref[0]+(-0.444522495013)*x_ref[1]+(0.791060985516)*x_ref[2]
        ref[4,1,1]=(0.287880876137)*x_ref[0]+(1.42265824896)*x_ref[1]+(-1.16312992573)*x_ref[2]
        ref[5,0,0]=(-1.25472591758)*x_ref[0]+(-1.48986095574)*x_ref[1]+(1.49212797538)*x_ref[2]
        ref[5,0,1]=(-0.774594904232)*x_ref[0]+(0.582557149894)*x_ref[1]+(-0.460088096141)*x_ref[2]
        ref[5,1,0]=(0.404584427176)*x_ref[0]+(-0.0808051689456)*x_ref[1]+(0.522819318812)*x_ref[2]
        ref[5,1,1]=(1.15513940465)*x_ref[0]+(1.09285861353)*x_ref[1]+(-0.166414260437)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-1.26340514811)*x[0]+(0.779300786262)*x[1]
        s[0,0,0,1]=(0.184428277326)*x[0]+(-0.831497209177)*x[1]
        s[0,0,1,0]=(1.35946842566)*x[0]+(0.273881497376)*x[1]
        s[0,0,1,1]=(0.117792005007)*x[0]+(1.06185199224)*x[1]
        s[0,0,2,0]=(0.810342149523)*x[0]+(-0.271662508688)*x[1]
        s[0,0,2,1]=(-0.332968381669)*x[0]+(0.0660545206117)*x[1]
        s[0,1,0,0]=(0.413613455431)*x[0]+(0.641032202343)*x[1]
        s[0,1,0,1]=(0.773447709702)*x[0]+(0.48460517354)*x[1]
        s[0,1,1,0]=(0.173632755729)*x[0]+(0.0635245280713)*x[1]
        s[0,1,1,1]=(0.731914799404)*x[0]+(1.42146728852)*x[1]
        s[0,1,2,0]=(0.688894013616)*x[0]+(1.0339010436)*x[1]
        s[0,1,2,1]=(-0.534272384342)*x[0]+(0.186670150961)*x[1]
        s[0,2,0,0]=(-0.318755700985)*x[0]+(-1.35142313256)*x[1]
        s[0,2,0,1]=(0.973640957973)*x[0]+(-0.103253061427)*x[1]
        s[0,2,1,0]=(1.14046612599)*x[0]+(0.238774677715)*x[1]
        s[0,2,1,1]=(-0.383452508275)*x[0]+(-1.62240702377)*x[1]
        s[0,2,2,0]=(0.761272968366)*x[0]+(1.446646048)*x[1]
        s[0,2,2,1]=(0.183707769875)*x[0]+(-0.630747785283)*x[1]
        s[0,3,0,0]=(-0.474486603803)*x[0]+(-0.0855459523088)*x[1]
        s[0,3,0,1]=(-0.991769725003)*x[0]+(1.004616452)*x[1]
        s[0,3,1,0]=(0.261836325547)*x[0]+(0.353357820559)*x[1]
        s[0,3,1,1]=(0.808880823505)*x[0]+(-0.495896346674)*x[1]
        s[0,3,2,0]=(1.06375339919)*x[0]+(0.00408804480046)*x[1]
        s[0,3,2,1]=(-1.32927483163)*x[0]+(-0.0833136005835)*x[1]
        s[0,4,0,0]=(1.13794063387)*x[0]+(-1.1458884083)*x[1]
        s[0,4,0,1]=(0.73883590181)*x[0]+(1.45878975539)*x[1]
        s[0,4,1,0]=(0.428378040368)*x[0]+(-1.25324354587)*x[1]
        s[0,4,1,1]=(1.19772514118)*x[0]+(0.457504803527)*x[1]
        s[0,4,2,0]=(-0.646891239861)*x[0]+(1.14291257798)*x[1]
        s[0,4,2,1]=(-0.579096313066)*x[0]+(-0.0276859933543)*x[1]
        s[1,0,0,0]=(-0.48935920813)*x[0]+(-0.327683586833)*x[1]
        s[1,0,0,1]=(0.59804841111)*x[0]+(-0.452980496343)*x[1]
        s[1,0,1,0]=(0.608017924036)*x[0]+(0.561553199982)*x[1]
        s[1,0,1,1]=(-0.00361014464582)*x[0]+(0.151090357053)*x[1]
        s[1,0,2,0]=(1.24385396019)*x[0]+(-0.0213646134992)*x[1]
        s[1,0,2,1]=(-0.219031743102)*x[0]+(-0.402476762676)*x[1]
        s[1,1,0,0]=(-0.196889997171)*x[0]+(0.675626499802)*x[1]
        s[1,1,0,1]=(1.23421246519)*x[0]+(-1.72367262236)*x[1]
        s[1,1,1,0]=(1.35571029115)*x[0]+(0.506057274132)*x[1]
        s[1,1,1,1]=(0.310699908104)*x[0]+(-0.292826394288)*x[1]
        s[1,1,2,0]=(0.122289986824)*x[0]+(-0.15947928964)*x[1]
        s[1,1,2,1]=(0.662808118991)*x[0]+(-0.927311967494)*x[1]
        s[1,2,0,0]=(-0.215898581104)*x[0]+(0.32902895413)*x[1]
        s[1,2,0,1]=(0.393145660435)*x[0]+(1.49040999822)*x[1]
        s[1,2,1,0]=(0.904144553172)*x[0]+(0.199605443931)*x[1]
        s[1,2,1,1]=(0.535979155678)*x[0]+(1.78418572873)*x[1]
        s[1,2,2,0]=(1.15578250862)*x[0]+(0.167811579149)*x[1]
        s[1,2,2,1]=(-0.0844120860286)*x[0]+(-0.56656817877)*x[1]
        s[1,3,0,0]=(0.760581325264)*x[0]+(1.00766443232)*x[1]
        s[1,3,0,1]=(-0.199244602194)*x[0]+(-1.54110932148)*x[1]
        s[1,3,1,0]=(1.40934237781)*x[0]+(0.700934912313)*x[1]
        s[1,3,1,1]=(-0.702507416172)*x[0]+(-0.206530323456)*x[1]
        s[1,3,2,0]=(-0.234301811191)*x[0]+(1.41101365089)*x[1]
        s[1,3,2,1]=(1.19609435879)*x[0]+(0.295560183318)*x[1]
        s[1,4,0,0]=(1.15287205838)*x[0]+(0.630564383314)*x[1]
        s[1,4,0,1]=(-0.993907995704)*x[0]+(-0.30836527499)*x[1]
        s[1,4,1,0]=(-0.265158946854)*x[0]+(0.689120014144)*x[1]
        s[1,4,1,1]=(0.503151681106)*x[0]+(-0.928730587416)*x[1]
        s[1,4,2,0]=(-1.0599860135)*x[0]+(-0.696932543908)*x[1]
        s[1,4,2,1]=(-0.365096260865)*x[0]+(0.443129268519)*x[1]
        s[2,0,0,0]=(0.412604298623)*x[0]+(0.591441458509)*x[1]
        s[2,0,0,1]=(0.536034627643)*x[0]+(-0.735261974283)*x[1]
        s[2,0,1,0]=(-0.336900405282)*x[0]+(-1.38493113859)*x[1]
        s[2,0,1,1]=(-0.50634870942)*x[0]+(0.0376544898311)*x[1]
        s[2,0,2,0]=(-0.0595134635196)*x[0]+(0.067285997446)*x[1]
        s[2,0,2,1]=(0.308894832145)*x[0]+(-1.94600763342)*x[1]
        s[2,1,0,0]=(0.443019228067)*x[0]+(-0.68858070865)*x[1]
        s[2,1,0,1]=(1.4461941226)*x[0]+(0.711186579745)*x[1]
        s[2,1,1,0]=(0.178777491574)*x[0]+(-0.746259056845)*x[1]
        s[2,1,1,1]=(1.24682492601)*x[0]+(-0.987057469956)*x[1]
        s[2,1,2,0]=(1.1863711951)*x[0]+(-0.853660028797)*x[1]
        s[2,1,2,1]=(1.13176028078)*x[0]+(-0.887521745194)*x[1]
        s[2,2,0,0]=(0.37579402707)*x[0]+(-0.462533529495)*x[1]
        s[2,2,0,1]=(-1.27778045393)*x[0]+(0.383310656984)*x[1]
        s[2,2,1,0]=(-0.0871651954146)*x[0]+(0.0266612015617)*x[1]
        s[2,2,1,1]=(1.26455972911)*x[0]+(0.597556950133)*x[1]
        s[2,2,2,0]=(0.397071934531)*x[0]+(-0.792146197976)*x[1]
        s[2,2,2,1]=(-0.531842419771)*x[0]+(-0.437895983867)*x[1]
        s[2,3,0,0]=(0.253602494754)*x[0]+(-0.943660021452)*x[1]
        s[2,3,0,1]=(0.381472581352)*x[0]+(1.36193258173)*x[1]
        s[2,3,1,0]=(0.530962583291)*x[0]+(0.853717668569)*x[1]
        s[2,3,1,1]=(-0.233925714076)*x[0]+(0.20197385961)*x[1]
        s[2,3,2,0]=(0.148804235592)*x[0]+(-0.116553369538)*x[1]
        s[2,3,2,1]=(-1.79360831582)*x[0]+(1.62247077485)*x[1]
        s[2,4,0,0]=(0.617987809937)*x[0]+(1.27685238582)*x[1]
        s[2,4,0,1]=(-0.359480838494)*x[0]+(-0.11716388114)*x[1]
        s[2,4,1,0]=(0.204308581462)*x[0]+(0.0241091749363)*x[1]
        s[2,4,1,1]=(0.424779981556)*x[0]+(-1.01811123317)*x[1]
        s[2,4,2,0]=(-0.708751234946)*x[0]+(0.714679051838)*x[1]
        s[2,4,2,1]=(-0.352350728857)*x[0]+(1.47198192719)*x[1]
        s[3,0,0,0]=(-0.0592432179871)*x[0]+(0.0900836798021)*x[1]
        s[3,0,0,1]=(-0.192266211858)*x[0]+(0.561246968252)*x[1]
        s[3,0,1,0]=(0.30878942472)*x[0]+(-0.578076897247)*x[1]
        s[3,0,1,1]=(-0.246226203035)*x[0]+(0.68840556577)*x[1]
        s[3,0,2,0]=(-0.42433037357)*x[0]+(-0.168333732109)*x[1]
        s[3,0,2,1]=(0.501931661511)*x[0]+(1.53237442299)*x[1]
        s[3,1,0,0]=(-0.835618728532)*x[0]+(-0.823254922183)*x[1]
        s[3,1,0,1]=(-0.609502259942)*x[0]+(0.139549090645)*x[1]
        s[3,1,1,0]=(-0.690897498494)*x[0]+(0.280254985362)*x[1]
        s[3,1,1,1]=(-0.89040023732)*x[0]+(0.616118941926)*x[1]
        s[3,1,2,0]=(-0.90402399371)*x[0]+(-0.399494199588)*x[1]
        s[3,1,2,1]=(1.31002070769)*x[0]+(-0.26001592275)*x[1]
        s[3,2,0,0]=(-0.741667891927)*x[0]+(-1.13827783123)*x[1]
        s[3,2,0,1]=(0.879168335296)*x[0]+(0.232648259015)*x[1]
        s[3,2,1,0]=(0.737095222363)*x[0]+(-1.26232530994)*x[1]
        s[3,2,1,1]=(-0.523976448837)*x[0]+(0.480347424944)*x[1]
        s[3,2,2,0]=(-1.61727611988)*x[0]+(-0.464564622523)*x[1]
        s[3,2,2,1]=(0.928282179697)*x[0]+(0.629575166525)*x[1]
        s[3,3,0,0]=(0.748679757793)*x[0]+(1.26002219505)*x[1]
        s[3,3,0,1]=(-0.497730460683)*x[0]+(0.963512065302)*x[1]
        s[3,3,1,0]=(-0.999580985099)*x[0]+(0.933789269497)*x[1]
        s[3,3,1,1]=(-0.151096110981)*x[0]+(-0.739374723252)*x[1]
        s[3,3,2,0]=(0.257880503573)*x[0]+(-1.07115978896)*x[1]
        s[3,3,2,1]=(0.188404246313)*x[0]+(0.365000571954)*x[1]
        s[3,4,0,0]=(1.63625895457)*x[0]+(-0.685581590196)*x[1]
        s[3,4,0,1]=(-0.665388096854)*x[0]+(-0.9121332766)*x[1]
        s[3,4,1,0]=(1.76715282849)*x[0]+(-1.15317031503)*x[1]
        s[3,4,1,1]=(-0.378469229808)*x[0]+(-0.897880026236)*x[1]
        s[3,4,2,0]=(0.183976703061)*x[0]+(1.76560300417)*x[1]
        s[3,4,2,1]=(-0.569272512314)*x[0]+(1.09830336519)*x[1]
        ref[0,0,0,0]=(-1.26340514811)*x_ref[0]+(0.779300786262)*x_ref[1]
        ref[0,0,0,1]=(0.184428277326)*x_ref[0]+(-0.831497209177)*x_ref[1]
        ref[0,0,1,0]=(1.35946842566)*x_ref[0]+(0.273881497376)*x_ref[1]
        ref[0,0,1,1]=(0.117792005007)*x_ref[0]+(1.06185199224)*x_ref[1]
        ref[0,0,2,0]=(0.810342149523)*x_ref[0]+(-0.271662508688)*x_ref[1]
        ref[0,0,2,1]=(-0.332968381669)*x_ref[0]+(0.0660545206117)*x_ref[1]
        ref[0,1,0,0]=(0.413613455431)*x_ref[0]+(0.641032202343)*x_ref[1]
        ref[0,1,0,1]=(0.773447709702)*x_ref[0]+(0.48460517354)*x_ref[1]
        ref[0,1,1,0]=(0.173632755729)*x_ref[0]+(0.0635245280713)*x_ref[1]
        ref[0,1,1,1]=(0.731914799404)*x_ref[0]+(1.42146728852)*x_ref[1]
        ref[0,1,2,0]=(0.688894013616)*x_ref[0]+(1.0339010436)*x_ref[1]
        ref[0,1,2,1]=(-0.534272384342)*x_ref[0]+(0.186670150961)*x_ref[1]
        ref[0,2,0,0]=(-0.318755700985)*x_ref[0]+(-1.35142313256)*x_ref[1]
        ref[0,2,0,1]=(0.973640957973)*x_ref[0]+(-0.103253061427)*x_ref[1]
        ref[0,2,1,0]=(1.14046612599)*x_ref[0]+(0.238774677715)*x_ref[1]
        ref[0,2,1,1]=(-0.383452508275)*x_ref[0]+(-1.62240702377)*x_ref[1]
        ref[0,2,2,0]=(0.761272968366)*x_ref[0]+(1.446646048)*x_ref[1]
        ref[0,2,2,1]=(0.183707769875)*x_ref[0]+(-0.630747785283)*x_ref[1]
        ref[0,3,0,0]=(-0.474486603803)*x_ref[0]+(-0.0855459523088)*x_ref[1]
        ref[0,3,0,1]=(-0.991769725003)*x_ref[0]+(1.004616452)*x_ref[1]
        ref[0,3,1,0]=(0.261836325547)*x_ref[0]+(0.353357820559)*x_ref[1]
        ref[0,3,1,1]=(0.808880823505)*x_ref[0]+(-0.495896346674)*x_ref[1]
        ref[0,3,2,0]=(1.06375339919)*x_ref[0]+(0.00408804480046)*x_ref[1]
        ref[0,3,2,1]=(-1.32927483163)*x_ref[0]+(-0.0833136005835)*x_ref[1]
        ref[0,4,0,0]=(1.13794063387)*x_ref[0]+(-1.1458884083)*x_ref[1]
        ref[0,4,0,1]=(0.73883590181)*x_ref[0]+(1.45878975539)*x_ref[1]
        ref[0,4,1,0]=(0.428378040368)*x_ref[0]+(-1.25324354587)*x_ref[1]
        ref[0,4,1,1]=(1.19772514118)*x_ref[0]+(0.457504803527)*x_ref[1]
        ref[0,4,2,0]=(-0.646891239861)*x_ref[0]+(1.14291257798)*x_ref[1]
        ref[0,4,2,1]=(-0.579096313066)*x_ref[0]+(-0.0276859933543)*x_ref[1]
        ref[1,0,0,0]=(-0.48935920813)*x_ref[0]+(-0.327683586833)*x_ref[1]
        ref[1,0,0,1]=(0.59804841111)*x_ref[0]+(-0.452980496343)*x_ref[1]
        ref[1,0,1,0]=(0.608017924036)*x_ref[0]+(0.561553199982)*x_ref[1]
        ref[1,0,1,1]=(-0.00361014464582)*x_ref[0]+(0.151090357053)*x_ref[1]
        ref[1,0,2,0]=(1.24385396019)*x_ref[0]+(-0.0213646134992)*x_ref[1]
        ref[1,0,2,1]=(-0.219031743102)*x_ref[0]+(-0.402476762676)*x_ref[1]
        ref[1,1,0,0]=(-0.196889997171)*x_ref[0]+(0.675626499802)*x_ref[1]
        ref[1,1,0,1]=(1.23421246519)*x_ref[0]+(-1.72367262236)*x_ref[1]
        ref[1,1,1,0]=(1.35571029115)*x_ref[0]+(0.506057274132)*x_ref[1]
        ref[1,1,1,1]=(0.310699908104)*x_ref[0]+(-0.292826394288)*x_ref[1]
        ref[1,1,2,0]=(0.122289986824)*x_ref[0]+(-0.15947928964)*x_ref[1]
        ref[1,1,2,1]=(0.662808118991)*x_ref[0]+(-0.927311967494)*x_ref[1]
        ref[1,2,0,0]=(-0.215898581104)*x_ref[0]+(0.32902895413)*x_ref[1]
        ref[1,2,0,1]=(0.393145660435)*x_ref[0]+(1.49040999822)*x_ref[1]
        ref[1,2,1,0]=(0.904144553172)*x_ref[0]+(0.199605443931)*x_ref[1]
        ref[1,2,1,1]=(0.535979155678)*x_ref[0]+(1.78418572873)*x_ref[1]
        ref[1,2,2,0]=(1.15578250862)*x_ref[0]+(0.167811579149)*x_ref[1]
        ref[1,2,2,1]=(-0.0844120860286)*x_ref[0]+(-0.56656817877)*x_ref[1]
        ref[1,3,0,0]=(0.760581325264)*x_ref[0]+(1.00766443232)*x_ref[1]
        ref[1,3,0,1]=(-0.199244602194)*x_ref[0]+(-1.54110932148)*x_ref[1]
        ref[1,3,1,0]=(1.40934237781)*x_ref[0]+(0.700934912313)*x_ref[1]
        ref[1,3,1,1]=(-0.702507416172)*x_ref[0]+(-0.206530323456)*x_ref[1]
        ref[1,3,2,0]=(-0.234301811191)*x_ref[0]+(1.41101365089)*x_ref[1]
        ref[1,3,2,1]=(1.19609435879)*x_ref[0]+(0.295560183318)*x_ref[1]
        ref[1,4,0,0]=(1.15287205838)*x_ref[0]+(0.630564383314)*x_ref[1]
        ref[1,4,0,1]=(-0.993907995704)*x_ref[0]+(-0.30836527499)*x_ref[1]
        ref[1,4,1,0]=(-0.265158946854)*x_ref[0]+(0.689120014144)*x_ref[1]
        ref[1,4,1,1]=(0.503151681106)*x_ref[0]+(-0.928730587416)*x_ref[1]
        ref[1,4,2,0]=(-1.0599860135)*x_ref[0]+(-0.696932543908)*x_ref[1]
        ref[1,4,2,1]=(-0.365096260865)*x_ref[0]+(0.443129268519)*x_ref[1]
        ref[2,0,0,0]=(0.412604298623)*x_ref[0]+(0.591441458509)*x_ref[1]
        ref[2,0,0,1]=(0.536034627643)*x_ref[0]+(-0.735261974283)*x_ref[1]
        ref[2,0,1,0]=(-0.336900405282)*x_ref[0]+(-1.38493113859)*x_ref[1]
        ref[2,0,1,1]=(-0.50634870942)*x_ref[0]+(0.0376544898311)*x_ref[1]
        ref[2,0,2,0]=(-0.0595134635196)*x_ref[0]+(0.067285997446)*x_ref[1]
        ref[2,0,2,1]=(0.308894832145)*x_ref[0]+(-1.94600763342)*x_ref[1]
        ref[2,1,0,0]=(0.443019228067)*x_ref[0]+(-0.68858070865)*x_ref[1]
        ref[2,1,0,1]=(1.4461941226)*x_ref[0]+(0.711186579745)*x_ref[1]
        ref[2,1,1,0]=(0.178777491574)*x_ref[0]+(-0.746259056845)*x_ref[1]
        ref[2,1,1,1]=(1.24682492601)*x_ref[0]+(-0.987057469956)*x_ref[1]
        ref[2,1,2,0]=(1.1863711951)*x_ref[0]+(-0.853660028797)*x_ref[1]
        ref[2,1,2,1]=(1.13176028078)*x_ref[0]+(-0.887521745194)*x_ref[1]
        ref[2,2,0,0]=(0.37579402707)*x_ref[0]+(-0.462533529495)*x_ref[1]
        ref[2,2,0,1]=(-1.27778045393)*x_ref[0]+(0.383310656984)*x_ref[1]
        ref[2,2,1,0]=(-0.0871651954146)*x_ref[0]+(0.0266612015617)*x_ref[1]
        ref[2,2,1,1]=(1.26455972911)*x_ref[0]+(0.597556950133)*x_ref[1]
        ref[2,2,2,0]=(0.397071934531)*x_ref[0]+(-0.792146197976)*x_ref[1]
        ref[2,2,2,1]=(-0.531842419771)*x_ref[0]+(-0.437895983867)*x_ref[1]
        ref[2,3,0,0]=(0.253602494754)*x_ref[0]+(-0.943660021452)*x_ref[1]
        ref[2,3,0,1]=(0.381472581352)*x_ref[0]+(1.36193258173)*x_ref[1]
        ref[2,3,1,0]=(0.530962583291)*x_ref[0]+(0.853717668569)*x_ref[1]
        ref[2,3,1,1]=(-0.233925714076)*x_ref[0]+(0.20197385961)*x_ref[1]
        ref[2,3,2,0]=(0.148804235592)*x_ref[0]+(-0.116553369538)*x_ref[1]
        ref[2,3,2,1]=(-1.79360831582)*x_ref[0]+(1.62247077485)*x_ref[1]
        ref[2,4,0,0]=(0.617987809937)*x_ref[0]+(1.27685238582)*x_ref[1]
        ref[2,4,0,1]=(-0.359480838494)*x_ref[0]+(-0.11716388114)*x_ref[1]
        ref[2,4,1,0]=(0.204308581462)*x_ref[0]+(0.0241091749363)*x_ref[1]
        ref[2,4,1,1]=(0.424779981556)*x_ref[0]+(-1.01811123317)*x_ref[1]
        ref[2,4,2,0]=(-0.708751234946)*x_ref[0]+(0.714679051838)*x_ref[1]
        ref[2,4,2,1]=(-0.352350728857)*x_ref[0]+(1.47198192719)*x_ref[1]
        ref[3,0,0,0]=(-0.0592432179871)*x_ref[0]+(0.0900836798021)*x_ref[1]
        ref[3,0,0,1]=(-0.192266211858)*x_ref[0]+(0.561246968252)*x_ref[1]
        ref[3,0,1,0]=(0.30878942472)*x_ref[0]+(-0.578076897247)*x_ref[1]
        ref[3,0,1,1]=(-0.246226203035)*x_ref[0]+(0.68840556577)*x_ref[1]
        ref[3,0,2,0]=(-0.42433037357)*x_ref[0]+(-0.168333732109)*x_ref[1]
        ref[3,0,2,1]=(0.501931661511)*x_ref[0]+(1.53237442299)*x_ref[1]
        ref[3,1,0,0]=(-0.835618728532)*x_ref[0]+(-0.823254922183)*x_ref[1]
        ref[3,1,0,1]=(-0.609502259942)*x_ref[0]+(0.139549090645)*x_ref[1]
        ref[3,1,1,0]=(-0.690897498494)*x_ref[0]+(0.280254985362)*x_ref[1]
        ref[3,1,1,1]=(-0.89040023732)*x_ref[0]+(0.616118941926)*x_ref[1]
        ref[3,1,2,0]=(-0.90402399371)*x_ref[0]+(-0.399494199588)*x_ref[1]
        ref[3,1,2,1]=(1.31002070769)*x_ref[0]+(-0.26001592275)*x_ref[1]
        ref[3,2,0,0]=(-0.741667891927)*x_ref[0]+(-1.13827783123)*x_ref[1]
        ref[3,2,0,1]=(0.879168335296)*x_ref[0]+(0.232648259015)*x_ref[1]
        ref[3,2,1,0]=(0.737095222363)*x_ref[0]+(-1.26232530994)*x_ref[1]
        ref[3,2,1,1]=(-0.523976448837)*x_ref[0]+(0.480347424944)*x_ref[1]
        ref[3,2,2,0]=(-1.61727611988)*x_ref[0]+(-0.464564622523)*x_ref[1]
        ref[3,2,2,1]=(0.928282179697)*x_ref[0]+(0.629575166525)*x_ref[1]
        ref[3,3,0,0]=(0.748679757793)*x_ref[0]+(1.26002219505)*x_ref[1]
        ref[3,3,0,1]=(-0.497730460683)*x_ref[0]+(0.963512065302)*x_ref[1]
        ref[3,3,1,0]=(-0.999580985099)*x_ref[0]+(0.933789269497)*x_ref[1]
        ref[3,3,1,1]=(-0.151096110981)*x_ref[0]+(-0.739374723252)*x_ref[1]
        ref[3,3,2,0]=(0.257880503573)*x_ref[0]+(-1.07115978896)*x_ref[1]
        ref[3,3,2,1]=(0.188404246313)*x_ref[0]+(0.365000571954)*x_ref[1]
        ref[3,4,0,0]=(1.63625895457)*x_ref[0]+(-0.685581590196)*x_ref[1]
        ref[3,4,0,1]=(-0.665388096854)*x_ref[0]+(-0.9121332766)*x_ref[1]
        ref[3,4,1,0]=(1.76715282849)*x_ref[0]+(-1.15317031503)*x_ref[1]
        ref[3,4,1,1]=(-0.378469229808)*x_ref[0]+(-0.897880026236)*x_ref[1]
        ref[3,4,2,0]=(0.183976703061)*x_ref[0]+(1.76560300417)*x_ref[1]
        ref[3,4,2,1]=(-0.569272512314)*x_ref[0]+(1.09830336519)*x_ref[1]
      else:
        s[0,0,0,0]=(0.146894168765)*x[0]+(1.25822855062)*x[1]+(0.088660814219)*x[2]
        s[0,0,0,1]=(0.572576296123)*x[0]+(0.670191219644)*x[1]+(-0.280619677889)*x[2]
        s[0,0,1,0]=(-1.19817665348)*x[0]+(1.35545322011)*x[1]+(-0.279751222491)*x[2]
        s[0,0,1,1]=(-0.0558236682639)*x[0]+(-1.61042107055)*x[1]+(0.622456635293)*x[2]
        s[0,0,2,0]=(-0.154784205079)*x[0]+(-0.52314996688)*x[1]+(1.0590614932)*x[2]
        s[0,0,2,1]=(1.58101401758)*x[0]+(-0.610736715272)*x[1]+(-0.487233776225)*x[2]
        s[0,1,0,0]=(0.0117365392099)*x[0]+(1.02397373706)*x[1]+(0.407377316737)*x[2]
        s[0,1,0,1]=(0.460387220384)*x[0]+(1.10329384655)*x[1]+(0.252209465365)*x[2]
        s[0,1,1,0]=(-1.33153199048)*x[0]+(-0.759669253534)*x[1]+(-0.856223614724)*x[2]
        s[0,1,1,1]=(-1.00599668771)*x[0]+(0.321343926218)*x[1]+(-0.198860330101)*x[2]
        s[0,1,2,0]=(1.00741183642)*x[0]+(-0.148473398917)*x[1]+(-0.484181353094)*x[2]
        s[0,1,2,1]=(-1.46490648953)*x[0]+(0.621077170698)*x[1]+(0.802865695373)*x[2]
        s[0,2,0,0]=(1.18498612655)*x[0]+(-0.511705644809)*x[1]+(0.563568352278)*x[2]
        s[0,2,0,1]=(-0.807427977876)*x[0]+(-0.252372201752)*x[1]+(0.251722661898)*x[2]
        s[0,2,1,0]=(0.524572175367)*x[0]+(0.49906104527)*x[1]+(-0.601836374873)*x[2]
        s[0,2,1,1]=(-0.318976775137)*x[0]+(-0.153065654914)*x[1]+(-0.458888325378)*x[2]
        s[0,2,2,0]=(-1.45513892332)*x[0]+(-0.252635212625)*x[1]+(0.490153491036)*x[2]
        s[0,2,2,1]=(-0.439422986461)*x[0]+(-0.937409556421)*x[1]+(0.209106988492)*x[2]
        s[0,3,0,0]=(-0.26828401981)*x[0]+(1.70808529771)*x[1]+(0.431245149358)*x[2]
        s[0,3,0,1]=(-0.831846912527)*x[0]+(0.572822507173)*x[1]+(-0.283130753547)*x[2]
        s[0,3,1,0]=(-0.625617835832)*x[0]+(-1.25495997626)*x[1]+(1.28743784867)*x[2]
        s[0,3,1,1]=(1.47223047075)*x[0]+(0.274365456026)*x[1]+(1.34355769078)*x[2]
        s[0,3,2,0]=(-0.0506967319243)*x[0]+(1.16569056569)*x[1]+(-0.965008332019)*x[2]
        s[0,3,2,1]=(0.0106139754664)*x[0]+(0.274252087281)*x[1]+(-0.501970174754)*x[2]
        s[0,4,0,0]=(0.717476010799)*x[0]+(0.380533443018)*x[1]+(-0.304127891816)*x[2]
        s[0,4,0,1]=(-1.29133512255)*x[0]+(-1.16667771199)*x[1]+(0.275809086731)*x[2]
        s[0,4,1,0]=(0.137939049278)*x[0]+(-0.302950342448)*x[1]+(-0.28865204469)*x[2]
        s[0,4,1,1]=(-0.857457318663)*x[0]+(0.423143726447)*x[1]+(0.749889416366)*x[2]
        s[0,4,2,0]=(-0.994428027329)*x[0]+(-0.377780341115)*x[1]+(-0.997480394895)*x[2]
        s[0,4,2,1]=(-0.681653236311)*x[0]+(-0.091905911233)*x[1]+(-0.877007024631)*x[2]
        s[1,0,0,0]=(0.33375482228)*x[0]+(-1.08816236114)*x[1]+(0.154701633038)*x[2]
        s[1,0,0,1]=(0.979227592935)*x[0]+(0.571415419963)*x[1]+(-0.20312466228)*x[2]
        s[1,0,1,0]=(1.59673013189)*x[0]+(1.01626905603)*x[1]+(0.216806440372)*x[2]
        s[1,0,1,1]=(-0.276034446694)*x[0]+(-0.168494544051)*x[1]+(0.981220114947)*x[2]
        s[1,0,2,0]=(0.205085469341)*x[0]+(-0.507970934322)*x[1]+(-0.724745229941)*x[2]
        s[1,0,2,1]=(-0.629331749902)*x[0]+(-1.27086230989)*x[1]+(-1.08479549542)*x[2]
        s[1,1,0,0]=(-0.786670678498)*x[0]+(1.24461814343)*x[1]+(0.644579154985)*x[2]
        s[1,1,0,1]=(1.05414628662)*x[0]+(0.901590532372)*x[1]+(0.809519110188)*x[2]
        s[1,1,1,0]=(-0.679152609307)*x[0]+(0.84101357998)*x[1]+(1.14312188299)*x[2]
        s[1,1,1,1]=(0.137715540853)*x[0]+(1.42995779339)*x[1]+(-1.10477793041)*x[2]
        s[1,1,2,0]=(-0.0647098241647)*x[0]+(1.18580807605)*x[1]+(0.605076723926)*x[2]
        s[1,1,2,1]=(0.744824848213)*x[0]+(0.532106337173)*x[1]+(0.68490109894)*x[2]
        s[1,2,0,0]=(-0.557916122207)*x[0]+(1.06859221643)*x[1]+(0.823215271018)*x[2]
        s[1,2,0,1]=(0.00598565716481)*x[0]+(0.884417903372)*x[1]+(1.7921906066)*x[2]
        s[1,2,1,0]=(-0.47245962749)*x[0]+(-0.906624987584)*x[1]+(0.137792563282)*x[2]
        s[1,2,1,1]=(0.948613971685)*x[0]+(0.753281322508)*x[1]+(0.110929011008)*x[2]
        s[1,2,2,0]=(0.175178588707)*x[0]+(-0.796494364923)*x[1]+(0.390555269738)*x[2]
        s[1,2,2,1]=(-1.04516910973)*x[0]+(-0.921757404949)*x[1]+(0.845139972079)*x[2]
        s[1,3,0,0]=(-1.42392853478)*x[0]+(-0.458461019373)*x[1]+(-0.213286746949)*x[2]
        s[1,3,0,1]=(0.13996349254)*x[0]+(0.108946526608)*x[1]+(-0.848132079317)*x[2]
        s[1,3,1,0]=(-1.44188153548)*x[0]+(-0.108156975475)*x[1]+(0.0715891132579)*x[2]
        s[1,3,1,1]=(0.525856508836)*x[0]+(0.950609265517)*x[1]+(0.0127362240999)*x[2]
        s[1,3,2,0]=(-1.0845975079)*x[0]+(-1.24949492248)*x[1]+(-0.00240035631582)*x[2]
        s[1,3,2,1]=(0.642060593331)*x[0]+(1.46218158975)*x[1]+(1.32385980887)*x[2]
        s[1,4,0,0]=(0.0119712366759)*x[0]+(1.39772768926)*x[1]+(-0.570649410717)*x[2]
        s[1,4,0,1]=(0.320030773751)*x[0]+(0.358086736384)*x[1]+(-0.859721999492)*x[2]
        s[1,4,1,0]=(0.574417642497)*x[0]+(-0.421844733062)*x[1]+(0.611638265232)*x[2]
        s[1,4,1,1]=(-0.228413169897)*x[0]+(-0.0391189339225)*x[1]+(0.635031723033)*x[2]
        s[1,4,2,0]=(0.0424678701733)*x[0]+(0.261307736177)*x[1]+(-0.133428864005)*x[2]
        s[1,4,2,1]=(0.169996877966)*x[0]+(1.30730453844)*x[1]+(-1.07718562596)*x[2]
        s[2,0,0,0]=(0.147232851651)*x[0]+(-0.134174394382)*x[1]+(-0.785323160299)*x[2]
        s[2,0,0,1]=(1.29417433206)*x[0]+(-1.39241927595)*x[1]+(-0.224395758464)*x[2]
        s[2,0,1,0]=(1.60938474462)*x[0]+(0.68821531646)*x[1]+(0.453462382464)*x[2]
        s[2,0,1,1]=(0.63619844559)*x[0]+(-0.956987170375)*x[1]+(-1.24753100984)*x[2]
        s[2,0,2,0]=(1.14194732577)*x[0]+(0.973489037101)*x[1]+(0.187588261589)*x[2]
        s[2,0,2,1]=(-0.898956363323)*x[0]+(0.108722086141)*x[1]+(-1.62445422766)*x[2]
        s[2,1,0,0]=(0.127827687335)*x[0]+(0.216922403148)*x[1]+(-1.5099539458)*x[2]
        s[2,1,0,1]=(-0.769830743998)*x[0]+(1.46060444448)*x[1]+(-0.318171252653)*x[2]
        s[2,1,1,0]=(-0.311976479418)*x[0]+(-0.949061813767)*x[1]+(-0.237478430906)*x[2]
        s[2,1,1,1]=(1.27107354882)*x[0]+(-0.427981502964)*x[1]+(-0.457224808787)*x[2]
        s[2,1,2,0]=(-0.633029461167)*x[0]+(0.255322271264)*x[1]+(-1.79431610464)*x[2]
        s[2,1,2,1]=(1.05355586142)*x[0]+(-1.69722728443)*x[1]+(-1.11928223816)*x[2]
        s[2,2,0,0]=(1.15538109557)*x[0]+(-0.52692609967)*x[1]+(-0.686266572053)*x[2]
        s[2,2,0,1]=(-0.327531879955)*x[0]+(0.365760847752)*x[1]+(-0.0600793962176)*x[2]
        s[2,2,1,0]=(1.16249877038)*x[0]+(0.710438081472)*x[1]+(0.986719636688)*x[2]
        s[2,2,1,1]=(-1.22799525195)*x[0]+(0.748811223969)*x[1]+(0.370195043623)*x[2]
        s[2,2,2,0]=(0.976246115046)*x[0]+(1.10537245041)*x[1]+(-0.759735725531)*x[2]
        s[2,2,2,1]=(1.01411361404)*x[0]+(-0.000596974234967)*x[1]+(-1.39135228207)*x[2]
        s[2,3,0,0]=(-0.581025188916)*x[0]+(-1.43405762699)*x[1]+(-0.912365741093)*x[2]
        s[2,3,0,1]=(-0.0511110676196)*x[0]+(0.230142796427)*x[1]+(-1.58165960293)*x[2]
        s[2,3,1,0]=(-1.46709370576)*x[0]+(0.425035892802)*x[1]+(0.282281348864)*x[2]
        s[2,3,1,1]=(-0.35514320269)*x[0]+(-0.39238669714)*x[1]+(-0.0118498421145)*x[2]
        s[2,3,2,0]=(0.317197568125)*x[0]+(0.689803280183)*x[1]+(-0.238327580547)*x[2]
        s[2,3,2,1]=(1.31241465464)*x[0]+(0.548355955144)*x[1]+(0.0218217641073)*x[2]
        s[2,4,0,0]=(0.787240379652)*x[0]+(-0.906804255968)*x[1]+(-1.17498478521)*x[2]
        s[2,4,0,1]=(1.19721624988)*x[0]+(0.567119981635)*x[1]+(1.42310267091)*x[2]
        s[2,4,1,0]=(-0.199829256778)*x[0]+(0.28908416123)*x[1]+(0.491497583863)*x[2]
        s[2,4,1,1]=(0.336357763916)*x[0]+(-0.750103385624)*x[1]+(-1.19492486127)*x[2]
        s[2,4,2,0]=(1.509049274)*x[0]+(-1.28641285242)*x[1]+(-0.250003617687)*x[2]
        s[2,4,2,1]=(-1.23803199733)*x[0]+(0.688217821134)*x[1]+(-0.436478772021)*x[2]
        s[3,0,0,0]=(0.634885383655)*x[0]+(1.14728850769)*x[1]+(0.417496697665)*x[2]
        s[3,0,0,1]=(-0.326274842766)*x[0]+(-1.47136469943)*x[1]+(1.87373171786)*x[2]
        s[3,0,1,0]=(-0.140107043914)*x[0]+(-0.261565104318)*x[1]+(0.0780015778888)*x[2]
        s[3,0,1,1]=(0.770979582613)*x[0]+(-0.899853446798)*x[1]+(1.45115909188)*x[2]
        s[3,0,2,0]=(0.477865962499)*x[0]+(-0.0621575016815)*x[1]+(-1.28975967232)*x[2]
        s[3,0,2,1]=(-1.41031619045)*x[0]+(1.50155797656)*x[1]+(-1.44890527708)*x[2]
        s[3,1,0,0]=(-1.04632613119)*x[0]+(0.00196059522552)*x[1]+(0.139147980023)*x[2]
        s[3,1,0,1]=(0.217272687162)*x[0]+(0.295773880887)*x[1]+(0.474090779139)*x[2]
        s[3,1,1,0]=(1.50572386246)*x[0]+(0.868579168785)*x[1]+(-0.12394808575)*x[2]
        s[3,1,1,1]=(0.882176073715)*x[0]+(0.144766914417)*x[1]+(-0.408510755049)*x[2]
        s[3,1,2,0]=(-1.21767372555)*x[0]+(0.195682844047)*x[1]+(-1.70399790168)*x[2]
        s[3,1,2,1]=(-0.110311118233)*x[0]+(-0.721806016748)*x[1]+(1.81495268748)*x[2]
        s[3,2,0,0]=(-0.0441997228291)*x[0]+(0.0754984632459)*x[1]+(0.274177049743)*x[2]
        s[3,2,0,1]=(-0.48089133278)*x[0]+(-0.00712717635695)*x[1]+(0.242678324953)*x[2]
        s[3,2,1,0]=(0.48377042087)*x[0]+(1.4374627607)*x[1]+(-1.63994815623)*x[2]
        s[3,2,1,1]=(0.548308948512)*x[0]+(-0.587878136236)*x[1]+(-0.636242676063)*x[2]
        s[3,2,2,0]=(-1.62147144921)*x[0]+(-0.937713365221)*x[1]+(0.416583670897)*x[2]
        s[3,2,2,1]=(1.50617037978)*x[0]+(0.849964476529)*x[1]+(-0.421030022756)*x[2]
        s[3,3,0,0]=(-0.0258240269876)*x[0]+(1.09488855424)*x[1]+(-0.864078888453)*x[2]
        s[3,3,0,1]=(0.121803754668)*x[0]+(1.47064577586)*x[1]+(-1.15267381807)*x[2]
        s[3,3,1,0]=(0.7847153928)*x[0]+(1.19923938448)*x[1]+(-0.220983796213)*x[2]
        s[3,3,1,1]=(0.591086037588)*x[0]+(1.49019206942)*x[1]+(-0.764272587507)*x[2]
        s[3,3,2,0]=(-0.751721311032)*x[0]+(-1.41725473775)*x[1]+(-0.182681015042)*x[2]
        s[3,3,2,1]=(1.18789607758)*x[0]+(0.321544556274)*x[1]+(-0.37682906633)*x[2]
        s[3,4,0,0]=(-0.314136382522)*x[0]+(0.190342372946)*x[1]+(-0.0583598827644)*x[2]
        s[3,4,0,1]=(1.87879524007)*x[0]+(0.0472715596509)*x[1]+(0.849571713097)*x[2]
        s[3,4,1,0]=(0.816120005278)*x[0]+(0.296980756484)*x[1]+(0.251581386555)*x[2]
        s[3,4,1,1]=(0.12560195013)*x[0]+(-0.293571823053)*x[1]+(-1.82952568911)*x[2]
        s[3,4,2,0]=(1.03923438653)*x[0]+(-0.899000072461)*x[1]+(-1.48431797596)*x[2]
        s[3,4,2,1]=(-1.12443865282)*x[0]+(-0.200378451615)*x[1]+(-1.41505465511)*x[2]
        ref[0,0,0,0]=(0.146894168765)*x_ref[0]+(1.25822855062)*x_ref[1]+(0.088660814219)*x_ref[2]
        ref[0,0,0,1]=(0.572576296123)*x_ref[0]+(0.670191219644)*x_ref[1]+(-0.280619677889)*x_ref[2]
        ref[0,0,1,0]=(-1.19817665348)*x_ref[0]+(1.35545322011)*x_ref[1]+(-0.279751222491)*x_ref[2]
        ref[0,0,1,1]=(-0.0558236682639)*x_ref[0]+(-1.61042107055)*x_ref[1]+(0.622456635293)*x_ref[2]
        ref[0,0,2,0]=(-0.154784205079)*x_ref[0]+(-0.52314996688)*x_ref[1]+(1.0590614932)*x_ref[2]
        ref[0,0,2,1]=(1.58101401758)*x_ref[0]+(-0.610736715272)*x_ref[1]+(-0.487233776225)*x_ref[2]
        ref[0,1,0,0]=(0.0117365392099)*x_ref[0]+(1.02397373706)*x_ref[1]+(0.407377316737)*x_ref[2]
        ref[0,1,0,1]=(0.460387220384)*x_ref[0]+(1.10329384655)*x_ref[1]+(0.252209465365)*x_ref[2]
        ref[0,1,1,0]=(-1.33153199048)*x_ref[0]+(-0.759669253534)*x_ref[1]+(-0.856223614724)*x_ref[2]
        ref[0,1,1,1]=(-1.00599668771)*x_ref[0]+(0.321343926218)*x_ref[1]+(-0.198860330101)*x_ref[2]
        ref[0,1,2,0]=(1.00741183642)*x_ref[0]+(-0.148473398917)*x_ref[1]+(-0.484181353094)*x_ref[2]
        ref[0,1,2,1]=(-1.46490648953)*x_ref[0]+(0.621077170698)*x_ref[1]+(0.802865695373)*x_ref[2]
        ref[0,2,0,0]=(1.18498612655)*x_ref[0]+(-0.511705644809)*x_ref[1]+(0.563568352278)*x_ref[2]
        ref[0,2,0,1]=(-0.807427977876)*x_ref[0]+(-0.252372201752)*x_ref[1]+(0.251722661898)*x_ref[2]
        ref[0,2,1,0]=(0.524572175367)*x_ref[0]+(0.49906104527)*x_ref[1]+(-0.601836374873)*x_ref[2]
        ref[0,2,1,1]=(-0.318976775137)*x_ref[0]+(-0.153065654914)*x_ref[1]+(-0.458888325378)*x_ref[2]
        ref[0,2,2,0]=(-1.45513892332)*x_ref[0]+(-0.252635212625)*x_ref[1]+(0.490153491036)*x_ref[2]
        ref[0,2,2,1]=(-0.439422986461)*x_ref[0]+(-0.937409556421)*x_ref[1]+(0.209106988492)*x_ref[2]
        ref[0,3,0,0]=(-0.26828401981)*x_ref[0]+(1.70808529771)*x_ref[1]+(0.431245149358)*x_ref[2]
        ref[0,3,0,1]=(-0.831846912527)*x_ref[0]+(0.572822507173)*x_ref[1]+(-0.283130753547)*x_ref[2]
        ref[0,3,1,0]=(-0.625617835832)*x_ref[0]+(-1.25495997626)*x_ref[1]+(1.28743784867)*x_ref[2]
        ref[0,3,1,1]=(1.47223047075)*x_ref[0]+(0.274365456026)*x_ref[1]+(1.34355769078)*x_ref[2]
        ref[0,3,2,0]=(-0.0506967319243)*x_ref[0]+(1.16569056569)*x_ref[1]+(-0.965008332019)*x_ref[2]
        ref[0,3,2,1]=(0.0106139754664)*x_ref[0]+(0.274252087281)*x_ref[1]+(-0.501970174754)*x_ref[2]
        ref[0,4,0,0]=(0.717476010799)*x_ref[0]+(0.380533443018)*x_ref[1]+(-0.304127891816)*x_ref[2]
        ref[0,4,0,1]=(-1.29133512255)*x_ref[0]+(-1.16667771199)*x_ref[1]+(0.275809086731)*x_ref[2]
        ref[0,4,1,0]=(0.137939049278)*x_ref[0]+(-0.302950342448)*x_ref[1]+(-0.28865204469)*x_ref[2]
        ref[0,4,1,1]=(-0.857457318663)*x_ref[0]+(0.423143726447)*x_ref[1]+(0.749889416366)*x_ref[2]
        ref[0,4,2,0]=(-0.994428027329)*x_ref[0]+(-0.377780341115)*x_ref[1]+(-0.997480394895)*x_ref[2]
        ref[0,4,2,1]=(-0.681653236311)*x_ref[0]+(-0.091905911233)*x_ref[1]+(-0.877007024631)*x_ref[2]
        ref[1,0,0,0]=(0.33375482228)*x_ref[0]+(-1.08816236114)*x_ref[1]+(0.154701633038)*x_ref[2]
        ref[1,0,0,1]=(0.979227592935)*x_ref[0]+(0.571415419963)*x_ref[1]+(-0.20312466228)*x_ref[2]
        ref[1,0,1,0]=(1.59673013189)*x_ref[0]+(1.01626905603)*x_ref[1]+(0.216806440372)*x_ref[2]
        ref[1,0,1,1]=(-0.276034446694)*x_ref[0]+(-0.168494544051)*x_ref[1]+(0.981220114947)*x_ref[2]
        ref[1,0,2,0]=(0.205085469341)*x_ref[0]+(-0.507970934322)*x_ref[1]+(-0.724745229941)*x_ref[2]
        ref[1,0,2,1]=(-0.629331749902)*x_ref[0]+(-1.27086230989)*x_ref[1]+(-1.08479549542)*x_ref[2]
        ref[1,1,0,0]=(-0.786670678498)*x_ref[0]+(1.24461814343)*x_ref[1]+(0.644579154985)*x_ref[2]
        ref[1,1,0,1]=(1.05414628662)*x_ref[0]+(0.901590532372)*x_ref[1]+(0.809519110188)*x_ref[2]
        ref[1,1,1,0]=(-0.679152609307)*x_ref[0]+(0.84101357998)*x_ref[1]+(1.14312188299)*x_ref[2]
        ref[1,1,1,1]=(0.137715540853)*x_ref[0]+(1.42995779339)*x_ref[1]+(-1.10477793041)*x_ref[2]
        ref[1,1,2,0]=(-0.0647098241647)*x_ref[0]+(1.18580807605)*x_ref[1]+(0.605076723926)*x_ref[2]
        ref[1,1,2,1]=(0.744824848213)*x_ref[0]+(0.532106337173)*x_ref[1]+(0.68490109894)*x_ref[2]
        ref[1,2,0,0]=(-0.557916122207)*x_ref[0]+(1.06859221643)*x_ref[1]+(0.823215271018)*x_ref[2]
        ref[1,2,0,1]=(0.00598565716481)*x_ref[0]+(0.884417903372)*x_ref[1]+(1.7921906066)*x_ref[2]
        ref[1,2,1,0]=(-0.47245962749)*x_ref[0]+(-0.906624987584)*x_ref[1]+(0.137792563282)*x_ref[2]
        ref[1,2,1,1]=(0.948613971685)*x_ref[0]+(0.753281322508)*x_ref[1]+(0.110929011008)*x_ref[2]
        ref[1,2,2,0]=(0.175178588707)*x_ref[0]+(-0.796494364923)*x_ref[1]+(0.390555269738)*x_ref[2]
        ref[1,2,2,1]=(-1.04516910973)*x_ref[0]+(-0.921757404949)*x_ref[1]+(0.845139972079)*x_ref[2]
        ref[1,3,0,0]=(-1.42392853478)*x_ref[0]+(-0.458461019373)*x_ref[1]+(-0.213286746949)*x_ref[2]
        ref[1,3,0,1]=(0.13996349254)*x_ref[0]+(0.108946526608)*x_ref[1]+(-0.848132079317)*x_ref[2]
        ref[1,3,1,0]=(-1.44188153548)*x_ref[0]+(-0.108156975475)*x_ref[1]+(0.0715891132579)*x_ref[2]
        ref[1,3,1,1]=(0.525856508836)*x_ref[0]+(0.950609265517)*x_ref[1]+(0.0127362240999)*x_ref[2]
        ref[1,3,2,0]=(-1.0845975079)*x_ref[0]+(-1.24949492248)*x_ref[1]+(-0.00240035631582)*x_ref[2]
        ref[1,3,2,1]=(0.642060593331)*x_ref[0]+(1.46218158975)*x_ref[1]+(1.32385980887)*x_ref[2]
        ref[1,4,0,0]=(0.0119712366759)*x_ref[0]+(1.39772768926)*x_ref[1]+(-0.570649410717)*x_ref[2]
        ref[1,4,0,1]=(0.320030773751)*x_ref[0]+(0.358086736384)*x_ref[1]+(-0.859721999492)*x_ref[2]
        ref[1,4,1,0]=(0.574417642497)*x_ref[0]+(-0.421844733062)*x_ref[1]+(0.611638265232)*x_ref[2]
        ref[1,4,1,1]=(-0.228413169897)*x_ref[0]+(-0.0391189339225)*x_ref[1]+(0.635031723033)*x_ref[2]
        ref[1,4,2,0]=(0.0424678701733)*x_ref[0]+(0.261307736177)*x_ref[1]+(-0.133428864005)*x_ref[2]
        ref[1,4,2,1]=(0.169996877966)*x_ref[0]+(1.30730453844)*x_ref[1]+(-1.07718562596)*x_ref[2]
        ref[2,0,0,0]=(0.147232851651)*x_ref[0]+(-0.134174394382)*x_ref[1]+(-0.785323160299)*x_ref[2]
        ref[2,0,0,1]=(1.29417433206)*x_ref[0]+(-1.39241927595)*x_ref[1]+(-0.224395758464)*x_ref[2]
        ref[2,0,1,0]=(1.60938474462)*x_ref[0]+(0.68821531646)*x_ref[1]+(0.453462382464)*x_ref[2]
        ref[2,0,1,1]=(0.63619844559)*x_ref[0]+(-0.956987170375)*x_ref[1]+(-1.24753100984)*x_ref[2]
        ref[2,0,2,0]=(1.14194732577)*x_ref[0]+(0.973489037101)*x_ref[1]+(0.187588261589)*x_ref[2]
        ref[2,0,2,1]=(-0.898956363323)*x_ref[0]+(0.108722086141)*x_ref[1]+(-1.62445422766)*x_ref[2]
        ref[2,1,0,0]=(0.127827687335)*x_ref[0]+(0.216922403148)*x_ref[1]+(-1.5099539458)*x_ref[2]
        ref[2,1,0,1]=(-0.769830743998)*x_ref[0]+(1.46060444448)*x_ref[1]+(-0.318171252653)*x_ref[2]
        ref[2,1,1,0]=(-0.311976479418)*x_ref[0]+(-0.949061813767)*x_ref[1]+(-0.237478430906)*x_ref[2]
        ref[2,1,1,1]=(1.27107354882)*x_ref[0]+(-0.427981502964)*x_ref[1]+(-0.457224808787)*x_ref[2]
        ref[2,1,2,0]=(-0.633029461167)*x_ref[0]+(0.255322271264)*x_ref[1]+(-1.79431610464)*x_ref[2]
        ref[2,1,2,1]=(1.05355586142)*x_ref[0]+(-1.69722728443)*x_ref[1]+(-1.11928223816)*x_ref[2]
        ref[2,2,0,0]=(1.15538109557)*x_ref[0]+(-0.52692609967)*x_ref[1]+(-0.686266572053)*x_ref[2]
        ref[2,2,0,1]=(-0.327531879955)*x_ref[0]+(0.365760847752)*x_ref[1]+(-0.0600793962176)*x_ref[2]
        ref[2,2,1,0]=(1.16249877038)*x_ref[0]+(0.710438081472)*x_ref[1]+(0.986719636688)*x_ref[2]
        ref[2,2,1,1]=(-1.22799525195)*x_ref[0]+(0.748811223969)*x_ref[1]+(0.370195043623)*x_ref[2]
        ref[2,2,2,0]=(0.976246115046)*x_ref[0]+(1.10537245041)*x_ref[1]+(-0.759735725531)*x_ref[2]
        ref[2,2,2,1]=(1.01411361404)*x_ref[0]+(-0.000596974234967)*x_ref[1]+(-1.39135228207)*x_ref[2]
        ref[2,3,0,0]=(-0.581025188916)*x_ref[0]+(-1.43405762699)*x_ref[1]+(-0.912365741093)*x_ref[2]
        ref[2,3,0,1]=(-0.0511110676196)*x_ref[0]+(0.230142796427)*x_ref[1]+(-1.58165960293)*x_ref[2]
        ref[2,3,1,0]=(-1.46709370576)*x_ref[0]+(0.425035892802)*x_ref[1]+(0.282281348864)*x_ref[2]
        ref[2,3,1,1]=(-0.35514320269)*x_ref[0]+(-0.39238669714)*x_ref[1]+(-0.0118498421145)*x_ref[2]
        ref[2,3,2,0]=(0.317197568125)*x_ref[0]+(0.689803280183)*x_ref[1]+(-0.238327580547)*x_ref[2]
        ref[2,3,2,1]=(1.31241465464)*x_ref[0]+(0.548355955144)*x_ref[1]+(0.0218217641073)*x_ref[2]
        ref[2,4,0,0]=(0.787240379652)*x_ref[0]+(-0.906804255968)*x_ref[1]+(-1.17498478521)*x_ref[2]
        ref[2,4,0,1]=(1.19721624988)*x_ref[0]+(0.567119981635)*x_ref[1]+(1.42310267091)*x_ref[2]
        ref[2,4,1,0]=(-0.199829256778)*x_ref[0]+(0.28908416123)*x_ref[1]+(0.491497583863)*x_ref[2]
        ref[2,4,1,1]=(0.336357763916)*x_ref[0]+(-0.750103385624)*x_ref[1]+(-1.19492486127)*x_ref[2]
        ref[2,4,2,0]=(1.509049274)*x_ref[0]+(-1.28641285242)*x_ref[1]+(-0.250003617687)*x_ref[2]
        ref[2,4,2,1]=(-1.23803199733)*x_ref[0]+(0.688217821134)*x_ref[1]+(-0.436478772021)*x_ref[2]
        ref[3,0,0,0]=(0.634885383655)*x_ref[0]+(1.14728850769)*x_ref[1]+(0.417496697665)*x_ref[2]
        ref[3,0,0,1]=(-0.326274842766)*x_ref[0]+(-1.47136469943)*x_ref[1]+(1.87373171786)*x_ref[2]
        ref[3,0,1,0]=(-0.140107043914)*x_ref[0]+(-0.261565104318)*x_ref[1]+(0.0780015778888)*x_ref[2]
        ref[3,0,1,1]=(0.770979582613)*x_ref[0]+(-0.899853446798)*x_ref[1]+(1.45115909188)*x_ref[2]
        ref[3,0,2,0]=(0.477865962499)*x_ref[0]+(-0.0621575016815)*x_ref[1]+(-1.28975967232)*x_ref[2]
        ref[3,0,2,1]=(-1.41031619045)*x_ref[0]+(1.50155797656)*x_ref[1]+(-1.44890527708)*x_ref[2]
        ref[3,1,0,0]=(-1.04632613119)*x_ref[0]+(0.00196059522552)*x_ref[1]+(0.139147980023)*x_ref[2]
        ref[3,1,0,1]=(0.217272687162)*x_ref[0]+(0.295773880887)*x_ref[1]+(0.474090779139)*x_ref[2]
        ref[3,1,1,0]=(1.50572386246)*x_ref[0]+(0.868579168785)*x_ref[1]+(-0.12394808575)*x_ref[2]
        ref[3,1,1,1]=(0.882176073715)*x_ref[0]+(0.144766914417)*x_ref[1]+(-0.408510755049)*x_ref[2]
        ref[3,1,2,0]=(-1.21767372555)*x_ref[0]+(0.195682844047)*x_ref[1]+(-1.70399790168)*x_ref[2]
        ref[3,1,2,1]=(-0.110311118233)*x_ref[0]+(-0.721806016748)*x_ref[1]+(1.81495268748)*x_ref[2]
        ref[3,2,0,0]=(-0.0441997228291)*x_ref[0]+(0.0754984632459)*x_ref[1]+(0.274177049743)*x_ref[2]
        ref[3,2,0,1]=(-0.48089133278)*x_ref[0]+(-0.00712717635695)*x_ref[1]+(0.242678324953)*x_ref[2]
        ref[3,2,1,0]=(0.48377042087)*x_ref[0]+(1.4374627607)*x_ref[1]+(-1.63994815623)*x_ref[2]
        ref[3,2,1,1]=(0.548308948512)*x_ref[0]+(-0.587878136236)*x_ref[1]+(-0.636242676063)*x_ref[2]
        ref[3,2,2,0]=(-1.62147144921)*x_ref[0]+(-0.937713365221)*x_ref[1]+(0.416583670897)*x_ref[2]
        ref[3,2,2,1]=(1.50617037978)*x_ref[0]+(0.849964476529)*x_ref[1]+(-0.421030022756)*x_ref[2]
        ref[3,3,0,0]=(-0.0258240269876)*x_ref[0]+(1.09488855424)*x_ref[1]+(-0.864078888453)*x_ref[2]
        ref[3,3,0,1]=(0.121803754668)*x_ref[0]+(1.47064577586)*x_ref[1]+(-1.15267381807)*x_ref[2]
        ref[3,3,1,0]=(0.7847153928)*x_ref[0]+(1.19923938448)*x_ref[1]+(-0.220983796213)*x_ref[2]
        ref[3,3,1,1]=(0.591086037588)*x_ref[0]+(1.49019206942)*x_ref[1]+(-0.764272587507)*x_ref[2]
        ref[3,3,2,0]=(-0.751721311032)*x_ref[0]+(-1.41725473775)*x_ref[1]+(-0.182681015042)*x_ref[2]
        ref[3,3,2,1]=(1.18789607758)*x_ref[0]+(0.321544556274)*x_ref[1]+(-0.37682906633)*x_ref[2]
        ref[3,4,0,0]=(-0.314136382522)*x_ref[0]+(0.190342372946)*x_ref[1]+(-0.0583598827644)*x_ref[2]
        ref[3,4,0,1]=(1.87879524007)*x_ref[0]+(0.0472715596509)*x_ref[1]+(0.849571713097)*x_ref[2]
        ref[3,4,1,0]=(0.816120005278)*x_ref[0]+(0.296980756484)*x_ref[1]+(0.251581386555)*x_ref[2]
        ref[3,4,1,1]=(0.12560195013)*x_ref[0]+(-0.293571823053)*x_ref[1]+(-1.82952568911)*x_ref[2]
        ref[3,4,2,0]=(1.03923438653)*x_ref[0]+(-0.899000072461)*x_ref[1]+(-1.48431797596)*x_ref[2]
        ref[3,4,2,1]=(-1.12443865282)*x_ref[0]+(-0.200378451615)*x_ref[1]+(-1.41505465511)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.32868090106)*x[0]+(-0.840160165417)*x[1]
        ref=(-0.32868090106)*x_ref[0]+(-0.840160165417)*x_ref[1]
      else:
        s=(-1.32732649563)*x[0]+(0.20149624757)*x[1]+(0.701715136384)*x[2]
        ref=(-1.32732649563)*x_ref[0]+(0.20149624757)*x_ref[1]+(0.701715136384)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        s[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
        ref[0]=(0.386938778323)*x[0]+(0.420564245379)*x[1]
        ref[1]=(-0.0278118108424)*x[0]+(1.37454230848)*x[1]
      else:
        s[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        s[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
        ref[0]=(1.10756282384)*x[0]+(-0.102116525275)*x[1]+(-0.680038229535)*x[2]
        ref[1]=(-0.804657774057)*x[0]+(0.937337454626)*x[1]+(0.214733790383)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.199811526227)*x[0]+(1.25013707432)*x[1]
        s[0,1]=(0.277539536007)*x[0]+(-0.023732930558)*x[1]
        s[0,2]=(-0.244973175438)*x[0]+(-0.501902270422)*x[1]
        s[0,3]=(0.249200744572)*x[0]+(0.520843463388)*x[1]
        s[0,4]=(-0.995168115444)*x[0]+(-0.259093571207)*x[1]
        s[1,0]=(0.179469283663)*x[0]+(-0.461808204127)*x[1]
        s[1,1]=(0.431508044494)*x[0]+(-0.423277002443)*x[1]
        s[1,2]=(0.732281368811)*x[0]+(-0.490387490227)*x[1]
        s[1,3]=(-0.665998043548)*x[0]+(-0.180996501441)*x[1]
        s[1,4]=(-0.773309943626)*x[0]+(-0.27171870837)*x[1]
        s[2,0]=(0.975757439909)*x[0]+(-0.980217145831)*x[1]
        s[2,1]=(-0.898690989301)*x[0]+(0.696886429725)*x[1]
        s[2,2]=(-0.568728495444)*x[0]+(0.171459808085)*x[1]
        s[2,3]=(-0.509734570618)*x[0]+(-0.318239573518)*x[1]
        s[2,4]=(-0.868166262762)*x[0]+(0.669164551211)*x[1]
        s[3,0]=(-1.11259074191)*x[0]+(-0.317962607489)*x[1]
        s[3,1]=(-0.735838783493)*x[0]+(-0.15621592888)*x[1]
        s[3,2]=(-0.348618152308)*x[0]+(0.346356772969)*x[1]
        s[3,3]=(1.58890219057)*x[0]+(-0.106915533345)*x[1]
        s[3,4]=(-0.739997195498)*x[0]+(-0.668986698058)*x[1]
        ref[0,0]=(0.199811526227)*x_ref[0]+(1.25013707432)*x_ref[1]
        ref[0,1]=(0.277539536007)*x_ref[0]+(-0.023732930558)*x_ref[1]
        ref[0,2]=(-0.244973175438)*x_ref[0]+(-0.501902270422)*x_ref[1]
        ref[0,3]=(0.249200744572)*x_ref[0]+(0.520843463388)*x_ref[1]
        ref[0,4]=(-0.995168115444)*x_ref[0]+(-0.259093571207)*x_ref[1]
        ref[1,0]=(0.179469283663)*x_ref[0]+(-0.461808204127)*x_ref[1]
        ref[1,1]=(0.431508044494)*x_ref[0]+(-0.423277002443)*x_ref[1]
        ref[1,2]=(0.732281368811)*x_ref[0]+(-0.490387490227)*x_ref[1]
        ref[1,3]=(-0.665998043548)*x_ref[0]+(-0.180996501441)*x_ref[1]
        ref[1,4]=(-0.773309943626)*x_ref[0]+(-0.27171870837)*x_ref[1]
        ref[2,0]=(0.975757439909)*x_ref[0]+(-0.980217145831)*x_ref[1]
        ref[2,1]=(-0.898690989301)*x_ref[0]+(0.696886429725)*x_ref[1]
        ref[2,2]=(-0.568728495444)*x_ref[0]+(0.171459808085)*x_ref[1]
        ref[2,3]=(-0.509734570618)*x_ref[0]+(-0.318239573518)*x_ref[1]
        ref[2,4]=(-0.868166262762)*x_ref[0]+(0.669164551211)*x_ref[1]
        ref[3,0]=(-1.11259074191)*x_ref[0]+(-0.317962607489)*x_ref[1]
        ref[3,1]=(-0.735838783493)*x_ref[0]+(-0.15621592888)*x_ref[1]
        ref[3,2]=(-0.348618152308)*x_ref[0]+(0.346356772969)*x_ref[1]
        ref[3,3]=(1.58890219057)*x_ref[0]+(-0.106915533345)*x_ref[1]
        ref[3,4]=(-0.739997195498)*x_ref[0]+(-0.668986698058)*x_ref[1]
      else:
        s[0,0]=(1.00508479945)*x[0]+(0.671813078579)*x[1]+(0.0352167301347)*x[2]
        s[0,1]=(0.832418208523)*x[0]+(-0.228971976603)*x[1]+(0.472369173779)*x[2]
        s[0,2]=(0.495212675956)*x[0]+(-0.850840698084)*x[1]+(-0.480464122957)*x[2]
        s[0,3]=(1.09800821588)*x[0]+(-1.16790708068)*x[1]+(-0.16448878954)*x[2]
        s[0,4]=(-0.142901980481)*x[0]+(-0.456358812317)*x[1]+(0.688171420193)*x[2]
        s[1,0]=(-0.452209235863)*x[0]+(0.835019801073)*x[1]+(-0.537900952591)*x[2]
        s[1,1]=(1.24188029232)*x[0]+(0.537899388384)*x[1]+(0.825366946283)*x[2]
        s[1,2]=(0.0568959695268)*x[0]+(-0.530615837561)*x[1]+(0.208390615403)*x[2]
        s[1,3]=(0.704415982126)*x[0]+(0.557947101676)*x[1]+(-1.41348269833)*x[2]
        s[1,4]=(0.23735208213)*x[0]+(0.188600394593)*x[1]+(-0.587309525946)*x[2]
        s[2,0]=(0.202262121276)*x[0]+(1.34583368863)*x[1]+(-1.416408583)*x[2]
        s[2,1]=(-0.889020760737)*x[0]+(1.18733643427)*x[1]+(1.11796325716)*x[2]
        s[2,2]=(0.0729529752138)*x[0]+(0.467461724385)*x[1]+(-1.26467465607)*x[2]
        s[2,3]=(0.553563173427)*x[0]+(-0.96991696383)*x[1]+(1.21489979267)*x[2]
        s[2,4]=(0.504157770518)*x[0]+(-0.0966112701707)*x[1]+(0.820591191307)*x[2]
        s[3,0]=(0.438942088634)*x[0]+(0.737754158644)*x[1]+(0.327747630162)*x[2]
        s[3,1]=(-0.966999874402)*x[0]+(0.218515238902)*x[1]+(-1.85502435057)*x[2]
        s[3,2]=(0.632366690747)*x[0]+(0.710212883666)*x[1]+(1.52843521678)*x[2]
        s[3,3]=(-0.139393405525)*x[0]+(-1.020102199)*x[1]+(0.485417385052)*x[2]
        s[3,4]=(-0.753676205348)*x[0]+(0.783454715482)*x[1]+(0.568363667465)*x[2]
        ref[0,0]=(1.00508479945)*x_ref[0]+(0.671813078579)*x_ref[1]+(0.0352167301347)*x_ref[2]
        ref[0,1]=(0.832418208523)*x_ref[0]+(-0.228971976603)*x_ref[1]+(0.472369173779)*x_ref[2]
        ref[0,2]=(0.495212675956)*x_ref[0]+(-0.850840698084)*x_ref[1]+(-0.480464122957)*x_ref[2]
        ref[0,3]=(1.09800821588)*x_ref[0]+(-1.16790708068)*x_ref[1]+(-0.16448878954)*x_ref[2]
        ref[0,4]=(-0.142901980481)*x_ref[0]+(-0.456358812317)*x_ref[1]+(0.688171420193)*x_ref[2]
        ref[1,0]=(-0.452209235863)*x_ref[0]+(0.835019801073)*x_ref[1]+(-0.537900952591)*x_ref[2]
        ref[1,1]=(1.24188029232)*x_ref[0]+(0.537899388384)*x_ref[1]+(0.825366946283)*x_ref[2]
        ref[1,2]=(0.0568959695268)*x_ref[0]+(-0.530615837561)*x_ref[1]+(0.208390615403)*x_ref[2]
        ref[1,3]=(0.704415982126)*x_ref[0]+(0.557947101676)*x_ref[1]+(-1.41348269833)*x_ref[2]
        ref[1,4]=(0.23735208213)*x_ref[0]+(0.188600394593)*x_ref[1]+(-0.587309525946)*x_ref[2]
        ref[2,0]=(0.202262121276)*x_ref[0]+(1.34583368863)*x_ref[1]+(-1.416408583)*x_ref[2]
        ref[2,1]=(-0.889020760737)*x_ref[0]+(1.18733643427)*x_ref[1]+(1.11796325716)*x_ref[2]
        ref[2,2]=(0.0729529752138)*x_ref[0]+(0.467461724385)*x_ref[1]+(-1.26467465607)*x_ref[2]
        ref[2,3]=(0.553563173427)*x_ref[0]+(-0.96991696383)*x_ref[1]+(1.21489979267)*x_ref[2]
        ref[2,4]=(0.504157770518)*x_ref[0]+(-0.0966112701707)*x_ref[1]+(0.820591191307)*x_ref[2]
        ref[3,0]=(0.438942088634)*x_ref[0]+(0.737754158644)*x_ref[1]+(0.327747630162)*x_ref[2]
        ref[3,1]=(-0.966999874402)*x_ref[0]+(0.218515238902)*x_ref[1]+(-1.85502435057)*x_ref[2]
        ref[3,2]=(0.632366690747)*x_ref[0]+(0.710212883666)*x_ref[1]+(1.52843521678)*x_ref[2]
        ref[3,3]=(-0.139393405525)*x_ref[0]+(-1.020102199)*x_ref[1]+(0.485417385052)*x_ref[2]
        ref[3,4]=(-0.753676205348)*x_ref[0]+(0.783454715482)*x_ref[1]+(0.568363667465)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.734232753652)*x[0]+(0.800930015514)*x[1]
        s[0,0,1]=(-0.219864689475)*x[0]+(0.157713027162)*x[1]
        s[0,1,0]=(-0.00664406256638)*x[0]+(0.872104848272)*x[1]
        s[0,1,1]=(-1.5575275283)*x[0]+(-1.43701883839)*x[1]
        s[1,0,0]=(1.60056664833)*x[0]+(1.6215424255)*x[1]
        s[1,0,1]=(1.5172017347)*x[0]+(-0.022919888148)*x[1]
        s[1,1,0]=(0.042690381655)*x[0]+(1.30019922)*x[1]
        s[1,1,1]=(0.105731664319)*x[0]+(-1.87027128366)*x[1]
        s[2,0,0]=(1.59675688345)*x[0]+(0.0608090042054)*x[1]
        s[2,0,1]=(-1.3520682478)*x[0]+(-0.881995801453)*x[1]
        s[2,1,0]=(1.08974009958)*x[0]+(0.13217220748)*x[1]
        s[2,1,1]=(-0.631056656103)*x[0]+(-0.548103546753)*x[1]
        s[3,0,0]=(1.59082745694)*x[0]+(0.101815532645)*x[1]
        s[3,0,1]=(-0.390741362781)*x[0]+(0.525254974689)*x[1]
        s[3,1,0]=(0.439652119858)*x[0]+(1.25127795686)*x[1]
        s[3,1,1]=(-1.10496039621)*x[0]+(0.0889406484341)*x[1]
        s[4,0,0]=(1.76164121088)*x[0]+(0.790919865199)*x[1]
        s[4,0,1]=(1.53483109)*x[0]+(1.15931695117)*x[1]
        s[4,1,0]=(-1.4226648704)*x[0]+(0.0380245720655)*x[1]
        s[4,1,1]=(-0.158958015165)*x[0]+(0.359005195468)*x[1]
        s[5,0,0]=(-0.0531318028725)*x[0]+(-0.0716200483232)*x[1]
        s[5,0,1]=(-1.64720026194)*x[0]+(-1.13699910338)*x[1]
        s[5,1,0]=(1.79036772333)*x[0]+(0.414445739998)*x[1]
        s[5,1,1]=(-0.168931889239)*x[0]+(1.55913218176)*x[1]
        ref[0,0,0]=(-0.734232753652)*x_ref[0]+(0.800930015514)*x_ref[1]
        ref[0,0,1]=(-0.219864689475)*x_ref[0]+(0.157713027162)*x_ref[1]
        ref[0,1,0]=(-0.00664406256638)*x_ref[0]+(0.872104848272)*x_ref[1]
        ref[0,1,1]=(-1.5575275283)*x_ref[0]+(-1.43701883839)*x_ref[1]
        ref[1,0,0]=(1.60056664833)*x_ref[0]+(1.6215424255)*x_ref[1]
        ref[1,0,1]=(1.5172017347)*x_ref[0]+(-0.022919888148)*x_ref[1]
        ref[1,1,0]=(0.042690381655)*x_ref[0]+(1.30019922)*x_ref[1]
        ref[1,1,1]=(0.105731664319)*x_ref[0]+(-1.87027128366)*x_ref[1]
        ref[2,0,0]=(1.59675688345)*x_ref[0]+(0.0608090042054)*x_ref[1]
        ref[2,0,1]=(-1.3520682478)*x_ref[0]+(-0.881995801453)*x_ref[1]
        ref[2,1,0]=(1.08974009958)*x_ref[0]+(0.13217220748)*x_ref[1]
        ref[2,1,1]=(-0.631056656103)*x_ref[0]+(-0.548103546753)*x_ref[1]
        ref[3,0,0]=(1.59082745694)*x_ref[0]+(0.101815532645)*x_ref[1]
        ref[3,0,1]=(-0.390741362781)*x_ref[0]+(0.525254974689)*x_ref[1]
        ref[3,1,0]=(0.439652119858)*x_ref[0]+(1.25127795686)*x_ref[1]
        ref[3,1,1]=(-1.10496039621)*x_ref[0]+(0.0889406484341)*x_ref[1]
        ref[4,0,0]=(1.76164121088)*x_ref[0]+(0.790919865199)*x_ref[1]
        ref[4,0,1]=(1.53483109)*x_ref[0]+(1.15931695117)*x_ref[1]
        ref[4,1,0]=(-1.4226648704)*x_ref[0]+(0.0380245720655)*x_ref[1]
        ref[4,1,1]=(-0.158958015165)*x_ref[0]+(0.359005195468)*x_ref[1]
        ref[5,0,0]=(-0.0531318028725)*x_ref[0]+(-0.0716200483232)*x_ref[1]
        ref[5,0,1]=(-1.64720026194)*x_ref[0]+(-1.13699910338)*x_ref[1]
        ref[5,1,0]=(1.79036772333)*x_ref[0]+(0.414445739998)*x_ref[1]
        ref[5,1,1]=(-0.168931889239)*x_ref[0]+(1.55913218176)*x_ref[1]
      else:
        s[0,0,0]=(0.030748782631)*x[0]+(0.0965401372343)*x[1]+(-1.19722517556)*x[2]
        s[0,0,1]=(-1.25981749671)*x[0]+(0.375214736326)*x[1]+(0.78094551805)*x[2]
        s[0,1,0]=(0.0859894648882)*x[0]+(-1.02857522767)*x[1]+(0.752176468687)*x[2]
        s[0,1,1]=(0.857538315926)*x[0]+(1.20788107416)*x[1]+(-1.00656644081)*x[2]
        s[1,0,0]=(-0.0208770456539)*x[0]+(0.296657331936)*x[1]+(1.49455132218)*x[2]
        s[1,0,1]=(-1.1213775487)*x[0]+(1.12351512934)*x[1]+(0.661972762047)*x[2]
        s[1,1,0]=(0.75336533272)*x[0]+(1.33059115242)*x[1]+(-0.43556744027)*x[2]
        s[1,1,1]=(0.467898749534)*x[0]+(1.53277916261)*x[1]+(-1.36041296872)*x[2]
        s[2,0,0]=(0.390444800695)*x[0]+(-0.0902034979398)*x[1]+(1.79547473252)*x[2]
        s[2,0,1]=(0.0488678241039)*x[0]+(0.000207248527408)*x[1]+(-0.231246108981)*x[2]
        s[2,1,0]=(0.907328661278)*x[0]+(-0.777592463716)*x[1]+(-0.949764289407)*x[2]
        s[2,1,1]=(-0.508314624748)*x[0]+(-0.0150508206233)*x[1]+(0.786266513715)*x[2]
        s[3,0,0]=(-0.264992651778)*x[0]+(1.23980464726)*x[1]+(0.480251929476)*x[2]
        s[3,0,1]=(-0.519373168965)*x[0]+(-0.091129773203)*x[1]+(-1.12961715008)*x[2]
        s[3,1,0]=(0.737417275735)*x[0]+(0.00073881678009)*x[1]+(-0.913229240795)*x[2]
        s[3,1,1]=(-0.565248614525)*x[0]+(0.232901289197)*x[1]+(-0.852815115095)*x[2]
        s[4,0,0]=(-0.513590194423)*x[0]+(-1.10607554932)*x[1]+(1.3269295152)*x[2]
        s[4,0,1]=(-0.0108694161983)*x[0]+(-1.39154818645)*x[1]+(1.5448779549)*x[2]
        s[4,1,0]=(1.25248373617)*x[0]+(-0.444522495013)*x[1]+(0.791060985516)*x[2]
        s[4,1,1]=(0.287880876137)*x[0]+(1.42265824896)*x[1]+(-1.16312992573)*x[2]
        s[5,0,0]=(-1.25472591758)*x[0]+(-1.48986095574)*x[1]+(1.49212797538)*x[2]
        s[5,0,1]=(-0.774594904232)*x[0]+(0.582557149894)*x[1]+(-0.460088096141)*x[2]
        s[5,1,0]=(0.404584427176)*x[0]+(-0.0808051689456)*x[1]+(0.522819318812)*x[2]
        s[5,1,1]=(1.15513940465)*x[0]+(1.09285861353)*x[1]+(-0.166414260437)*x[2]
        ref[0,0,0]=(0.030748782631)*x_ref[0]+(0.0965401372343)*x_ref[1]+(-1.19722517556)*x_ref[2]
        ref[0,0,1]=(-1.25981749671)*x_ref[0]+(0.375214736326)*x_ref[1]+(0.78094551805)*x_ref[2]
        ref[0,1,0]=(0.0859894648882)*x_ref[0]+(-1.02857522767)*x_ref[1]+(0.752176468687)*x_ref[2]
        ref[0,1,1]=(0.857538315926)*x_ref[0]+(1.20788107416)*x_ref[1]+(-1.00656644081)*x_ref[2]
        ref[1,0,0]=(-0.0208770456539)*x_ref[0]+(0.296657331936)*x_ref[1]+(1.49455132218)*x_ref[2]
        ref[1,0,1]=(-1.1213775487)*x_ref[0]+(1.12351512934)*x_ref[1]+(0.661972762047)*x_ref[2]
        ref[1,1,0]=(0.75336533272)*x_ref[0]+(1.33059115242)*x_ref[1]+(-0.43556744027)*x_ref[2]
        ref[1,1,1]=(0.467898749534)*x_ref[0]+(1.53277916261)*x_ref[1]+(-1.36041296872)*x_ref[2]
        ref[2,0,0]=(0.390444800695)*x_ref[0]+(-0.0902034979398)*x_ref[1]+(1.79547473252)*x_ref[2]
        ref[2,0,1]=(0.0488678241039)*x_ref[0]+(0.000207248527408)*x_ref[1]+(-0.231246108981)*x_ref[2]
        ref[2,1,0]=(0.907328661278)*x_ref[0]+(-0.777592463716)*x_ref[1]+(-0.949764289407)*x_ref[2]
        ref[2,1,1]=(-0.508314624748)*x_ref[0]+(-0.0150508206233)*x_ref[1]+(0.786266513715)*x_ref[2]
        ref[3,0,0]=(-0.264992651778)*x_ref[0]+(1.23980464726)*x_ref[1]+(0.480251929476)*x_ref[2]
        ref[3,0,1]=(-0.519373168965)*x_ref[0]+(-0.091129773203)*x_ref[1]+(-1.12961715008)*x_ref[2]
        ref[3,1,0]=(0.737417275735)*x_ref[0]+(0.00073881678009)*x_ref[1]+(-0.913229240795)*x_ref[2]
        ref[3,1,1]=(-0.565248614525)*x_ref[0]+(0.232901289197)*x_ref[1]+(-0.852815115095)*x_ref[2]
        ref[4,0,0]=(-0.513590194423)*x_ref[0]+(-1.10607554932)*x_ref[1]+(1.3269295152)*x_ref[2]
        ref[4,0,1]=(-0.0108694161983)*x_ref[0]+(-1.39154818645)*x_ref[1]+(1.5448779549)*x_ref[2]
        ref[4,1,0]=(1.25248373617)*x_ref[0]+(-0.444522495013)*x_ref[1]+(0.791060985516)*x_ref[2]
        ref[4,1,1]=(0.287880876137)*x_ref[0]+(1.42265824896)*x_ref[1]+(-1.16312992573)*x_ref[2]
        ref[5,0,0]=(-1.25472591758)*x_ref[0]+(-1.48986095574)*x_ref[1]+(1.49212797538)*x_ref[2]
        ref[5,0,1]=(-0.774594904232)*x_ref[0]+(0.582557149894)*x_ref[1]+(-0.460088096141)*x_ref[2]
        ref[5,1,0]=(0.404584427176)*x_ref[0]+(-0.0808051689456)*x_ref[1]+(0.522819318812)*x_ref[2]
        ref[5,1,1]=(1.15513940465)*x_ref[0]+(1.09285861353)*x_ref[1]+(-0.166414260437)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-1.26340514811)*x[0]+(0.779300786262)*x[1]
        s[0,0,0,1]=(0.184428277326)*x[0]+(-0.831497209177)*x[1]
        s[0,0,1,0]=(1.35946842566)*x[0]+(0.273881497376)*x[1]
        s[0,0,1,1]=(0.117792005007)*x[0]+(1.06185199224)*x[1]
        s[0,0,2,0]=(0.810342149523)*x[0]+(-0.271662508688)*x[1]
        s[0,0,2,1]=(-0.332968381669)*x[0]+(0.0660545206117)*x[1]
        s[0,1,0,0]=(0.413613455431)*x[0]+(0.641032202343)*x[1]
        s[0,1,0,1]=(0.773447709702)*x[0]+(0.48460517354)*x[1]
        s[0,1,1,0]=(0.173632755729)*x[0]+(0.0635245280713)*x[1]
        s[0,1,1,1]=(0.731914799404)*x[0]+(1.42146728852)*x[1]
        s[0,1,2,0]=(0.688894013616)*x[0]+(1.0339010436)*x[1]
        s[0,1,2,1]=(-0.534272384342)*x[0]+(0.186670150961)*x[1]
        s[0,2,0,0]=(-0.318755700985)*x[0]+(-1.35142313256)*x[1]
        s[0,2,0,1]=(0.973640957973)*x[0]+(-0.103253061427)*x[1]
        s[0,2,1,0]=(1.14046612599)*x[0]+(0.238774677715)*x[1]
        s[0,2,1,1]=(-0.383452508275)*x[0]+(-1.62240702377)*x[1]
        s[0,2,2,0]=(0.761272968366)*x[0]+(1.446646048)*x[1]
        s[0,2,2,1]=(0.183707769875)*x[0]+(-0.630747785283)*x[1]
        s[0,3,0,0]=(-0.474486603803)*x[0]+(-0.0855459523088)*x[1]
        s[0,3,0,1]=(-0.991769725003)*x[0]+(1.004616452)*x[1]
        s[0,3,1,0]=(0.261836325547)*x[0]+(0.353357820559)*x[1]
        s[0,3,1,1]=(0.808880823505)*x[0]+(-0.495896346674)*x[1]
        s[0,3,2,0]=(1.06375339919)*x[0]+(0.00408804480046)*x[1]
        s[0,3,2,1]=(-1.32927483163)*x[0]+(-0.0833136005835)*x[1]
        s[0,4,0,0]=(1.13794063387)*x[0]+(-1.1458884083)*x[1]
        s[0,4,0,1]=(0.73883590181)*x[0]+(1.45878975539)*x[1]
        s[0,4,1,0]=(0.428378040368)*x[0]+(-1.25324354587)*x[1]
        s[0,4,1,1]=(1.19772514118)*x[0]+(0.457504803527)*x[1]
        s[0,4,2,0]=(-0.646891239861)*x[0]+(1.14291257798)*x[1]
        s[0,4,2,1]=(-0.579096313066)*x[0]+(-0.0276859933543)*x[1]
        s[1,0,0,0]=(-0.48935920813)*x[0]+(-0.327683586833)*x[1]
        s[1,0,0,1]=(0.59804841111)*x[0]+(-0.452980496343)*x[1]
        s[1,0,1,0]=(0.608017924036)*x[0]+(0.561553199982)*x[1]
        s[1,0,1,1]=(-0.00361014464582)*x[0]+(0.151090357053)*x[1]
        s[1,0,2,0]=(1.24385396019)*x[0]+(-0.0213646134992)*x[1]
        s[1,0,2,1]=(-0.219031743102)*x[0]+(-0.402476762676)*x[1]
        s[1,1,0,0]=(-0.196889997171)*x[0]+(0.675626499802)*x[1]
        s[1,1,0,1]=(1.23421246519)*x[0]+(-1.72367262236)*x[1]
        s[1,1,1,0]=(1.35571029115)*x[0]+(0.506057274132)*x[1]
        s[1,1,1,1]=(0.310699908104)*x[0]+(-0.292826394288)*x[1]
        s[1,1,2,0]=(0.122289986824)*x[0]+(-0.15947928964)*x[1]
        s[1,1,2,1]=(0.662808118991)*x[0]+(-0.927311967494)*x[1]
        s[1,2,0,0]=(-0.215898581104)*x[0]+(0.32902895413)*x[1]
        s[1,2,0,1]=(0.393145660435)*x[0]+(1.49040999822)*x[1]
        s[1,2,1,0]=(0.904144553172)*x[0]+(0.199605443931)*x[1]
        s[1,2,1,1]=(0.535979155678)*x[0]+(1.78418572873)*x[1]
        s[1,2,2,0]=(1.15578250862)*x[0]+(0.167811579149)*x[1]
        s[1,2,2,1]=(-0.0844120860286)*x[0]+(-0.56656817877)*x[1]
        s[1,3,0,0]=(0.760581325264)*x[0]+(1.00766443232)*x[1]
        s[1,3,0,1]=(-0.199244602194)*x[0]+(-1.54110932148)*x[1]
        s[1,3,1,0]=(1.40934237781)*x[0]+(0.700934912313)*x[1]
        s[1,3,1,1]=(-0.702507416172)*x[0]+(-0.206530323456)*x[1]
        s[1,3,2,0]=(-0.234301811191)*x[0]+(1.41101365089)*x[1]
        s[1,3,2,1]=(1.19609435879)*x[0]+(0.295560183318)*x[1]
        s[1,4,0,0]=(1.15287205838)*x[0]+(0.630564383314)*x[1]
        s[1,4,0,1]=(-0.993907995704)*x[0]+(-0.30836527499)*x[1]
        s[1,4,1,0]=(-0.265158946854)*x[0]+(0.689120014144)*x[1]
        s[1,4,1,1]=(0.503151681106)*x[0]+(-0.928730587416)*x[1]
        s[1,4,2,0]=(-1.0599860135)*x[0]+(-0.696932543908)*x[1]
        s[1,4,2,1]=(-0.365096260865)*x[0]+(0.443129268519)*x[1]
        s[2,0,0,0]=(0.412604298623)*x[0]+(0.591441458509)*x[1]
        s[2,0,0,1]=(0.536034627643)*x[0]+(-0.735261974283)*x[1]
        s[2,0,1,0]=(-0.336900405282)*x[0]+(-1.38493113859)*x[1]
        s[2,0,1,1]=(-0.50634870942)*x[0]+(0.0376544898311)*x[1]
        s[2,0,2,0]=(-0.0595134635196)*x[0]+(0.067285997446)*x[1]
        s[2,0,2,1]=(0.308894832145)*x[0]+(-1.94600763342)*x[1]
        s[2,1,0,0]=(0.443019228067)*x[0]+(-0.68858070865)*x[1]
        s[2,1,0,1]=(1.4461941226)*x[0]+(0.711186579745)*x[1]
        s[2,1,1,0]=(0.178777491574)*x[0]+(-0.746259056845)*x[1]
        s[2,1,1,1]=(1.24682492601)*x[0]+(-0.987057469956)*x[1]
        s[2,1,2,0]=(1.1863711951)*x[0]+(-0.853660028797)*x[1]
        s[2,1,2,1]=(1.13176028078)*x[0]+(-0.887521745194)*x[1]
        s[2,2,0,0]=(0.37579402707)*x[0]+(-0.462533529495)*x[1]
        s[2,2,0,1]=(-1.27778045393)*x[0]+(0.383310656984)*x[1]
        s[2,2,1,0]=(-0.0871651954146)*x[0]+(0.0266612015617)*x[1]
        s[2,2,1,1]=(1.26455972911)*x[0]+(0.597556950133)*x[1]
        s[2,2,2,0]=(0.397071934531)*x[0]+(-0.792146197976)*x[1]
        s[2,2,2,1]=(-0.531842419771)*x[0]+(-0.437895983867)*x[1]
        s[2,3,0,0]=(0.253602494754)*x[0]+(-0.943660021452)*x[1]
        s[2,3,0,1]=(0.381472581352)*x[0]+(1.36193258173)*x[1]
        s[2,3,1,0]=(0.530962583291)*x[0]+(0.853717668569)*x[1]
        s[2,3,1,1]=(-0.233925714076)*x[0]+(0.20197385961)*x[1]
        s[2,3,2,0]=(0.148804235592)*x[0]+(-0.116553369538)*x[1]
        s[2,3,2,1]=(-1.79360831582)*x[0]+(1.62247077485)*x[1]
        s[2,4,0,0]=(0.617987809937)*x[0]+(1.27685238582)*x[1]
        s[2,4,0,1]=(-0.359480838494)*x[0]+(-0.11716388114)*x[1]
        s[2,4,1,0]=(0.204308581462)*x[0]+(0.0241091749363)*x[1]
        s[2,4,1,1]=(0.424779981556)*x[0]+(-1.01811123317)*x[1]
        s[2,4,2,0]=(-0.708751234946)*x[0]+(0.714679051838)*x[1]
        s[2,4,2,1]=(-0.352350728857)*x[0]+(1.47198192719)*x[1]
        s[3,0,0,0]=(-0.0592432179871)*x[0]+(0.0900836798021)*x[1]
        s[3,0,0,1]=(-0.192266211858)*x[0]+(0.561246968252)*x[1]
        s[3,0,1,0]=(0.30878942472)*x[0]+(-0.578076897247)*x[1]
        s[3,0,1,1]=(-0.246226203035)*x[0]+(0.68840556577)*x[1]
        s[3,0,2,0]=(-0.42433037357)*x[0]+(-0.168333732109)*x[1]
        s[3,0,2,1]=(0.501931661511)*x[0]+(1.53237442299)*x[1]
        s[3,1,0,0]=(-0.835618728532)*x[0]+(-0.823254922183)*x[1]
        s[3,1,0,1]=(-0.609502259942)*x[0]+(0.139549090645)*x[1]
        s[3,1,1,0]=(-0.690897498494)*x[0]+(0.280254985362)*x[1]
        s[3,1,1,1]=(-0.89040023732)*x[0]+(0.616118941926)*x[1]
        s[3,1,2,0]=(-0.90402399371)*x[0]+(-0.399494199588)*x[1]
        s[3,1,2,1]=(1.31002070769)*x[0]+(-0.26001592275)*x[1]
        s[3,2,0,0]=(-0.741667891927)*x[0]+(-1.13827783123)*x[1]
        s[3,2,0,1]=(0.879168335296)*x[0]+(0.232648259015)*x[1]
        s[3,2,1,0]=(0.737095222363)*x[0]+(-1.26232530994)*x[1]
        s[3,2,1,1]=(-0.523976448837)*x[0]+(0.480347424944)*x[1]
        s[3,2,2,0]=(-1.61727611988)*x[0]+(-0.464564622523)*x[1]
        s[3,2,2,1]=(0.928282179697)*x[0]+(0.629575166525)*x[1]
        s[3,3,0,0]=(0.748679757793)*x[0]+(1.26002219505)*x[1]
        s[3,3,0,1]=(-0.497730460683)*x[0]+(0.963512065302)*x[1]
        s[3,3,1,0]=(-0.999580985099)*x[0]+(0.933789269497)*x[1]
        s[3,3,1,1]=(-0.151096110981)*x[0]+(-0.739374723252)*x[1]
        s[3,3,2,0]=(0.257880503573)*x[0]+(-1.07115978896)*x[1]
        s[3,3,2,1]=(0.188404246313)*x[0]+(0.365000571954)*x[1]
        s[3,4,0,0]=(1.63625895457)*x[0]+(-0.685581590196)*x[1]
        s[3,4,0,1]=(-0.665388096854)*x[0]+(-0.9121332766)*x[1]
        s[3,4,1,0]=(1.76715282849)*x[0]+(-1.15317031503)*x[1]
        s[3,4,1,1]=(-0.378469229808)*x[0]+(-0.897880026236)*x[1]
        s[3,4,2,0]=(0.183976703061)*x[0]+(1.76560300417)*x[1]
        s[3,4,2,1]=(-0.569272512314)*x[0]+(1.09830336519)*x[1]
        ref[0,0,0,0]=(-1.26340514811)*x_ref[0]+(0.779300786262)*x_ref[1]
        ref[0,0,0,1]=(0.184428277326)*x_ref[0]+(-0.831497209177)*x_ref[1]
        ref[0,0,1,0]=(1.35946842566)*x_ref[0]+(0.273881497376)*x_ref[1]
        ref[0,0,1,1]=(0.117792005007)*x_ref[0]+(1.06185199224)*x_ref[1]
        ref[0,0,2,0]=(0.810342149523)*x_ref[0]+(-0.271662508688)*x_ref[1]
        ref[0,0,2,1]=(-0.332968381669)*x_ref[0]+(0.0660545206117)*x_ref[1]
        ref[0,1,0,0]=(0.413613455431)*x_ref[0]+(0.641032202343)*x_ref[1]
        ref[0,1,0,1]=(0.773447709702)*x_ref[0]+(0.48460517354)*x_ref[1]
        ref[0,1,1,0]=(0.173632755729)*x_ref[0]+(0.0635245280713)*x_ref[1]
        ref[0,1,1,1]=(0.731914799404)*x_ref[0]+(1.42146728852)*x_ref[1]
        ref[0,1,2,0]=(0.688894013616)*x_ref[0]+(1.0339010436)*x_ref[1]
        ref[0,1,2,1]=(-0.534272384342)*x_ref[0]+(0.186670150961)*x_ref[1]
        ref[0,2,0,0]=(-0.318755700985)*x_ref[0]+(-1.35142313256)*x_ref[1]
        ref[0,2,0,1]=(0.973640957973)*x_ref[0]+(-0.103253061427)*x_ref[1]
        ref[0,2,1,0]=(1.14046612599)*x_ref[0]+(0.238774677715)*x_ref[1]
        ref[0,2,1,1]=(-0.383452508275)*x_ref[0]+(-1.62240702377)*x_ref[1]
        ref[0,2,2,0]=(0.761272968366)*x_ref[0]+(1.446646048)*x_ref[1]
        ref[0,2,2,1]=(0.183707769875)*x_ref[0]+(-0.630747785283)*x_ref[1]
        ref[0,3,0,0]=(-0.474486603803)*x_ref[0]+(-0.0855459523088)*x_ref[1]
        ref[0,3,0,1]=(-0.991769725003)*x_ref[0]+(1.004616452)*x_ref[1]
        ref[0,3,1,0]=(0.261836325547)*x_ref[0]+(0.353357820559)*x_ref[1]
        ref[0,3,1,1]=(0.808880823505)*x_ref[0]+(-0.495896346674)*x_ref[1]
        ref[0,3,2,0]=(1.06375339919)*x_ref[0]+(0.00408804480046)*x_ref[1]
        ref[0,3,2,1]=(-1.32927483163)*x_ref[0]+(-0.0833136005835)*x_ref[1]
        ref[0,4,0,0]=(1.13794063387)*x_ref[0]+(-1.1458884083)*x_ref[1]
        ref[0,4,0,1]=(0.73883590181)*x_ref[0]+(1.45878975539)*x_ref[1]
        ref[0,4,1,0]=(0.428378040368)*x_ref[0]+(-1.25324354587)*x_ref[1]
        ref[0,4,1,1]=(1.19772514118)*x_ref[0]+(0.457504803527)*x_ref[1]
        ref[0,4,2,0]=(-0.646891239861)*x_ref[0]+(1.14291257798)*x_ref[1]
        ref[0,4,2,1]=(-0.579096313066)*x_ref[0]+(-0.0276859933543)*x_ref[1]
        ref[1,0,0,0]=(-0.48935920813)*x_ref[0]+(-0.327683586833)*x_ref[1]
        ref[1,0,0,1]=(0.59804841111)*x_ref[0]+(-0.452980496343)*x_ref[1]
        ref[1,0,1,0]=(0.608017924036)*x_ref[0]+(0.561553199982)*x_ref[1]
        ref[1,0,1,1]=(-0.00361014464582)*x_ref[0]+(0.151090357053)*x_ref[1]
        ref[1,0,2,0]=(1.24385396019)*x_ref[0]+(-0.0213646134992)*x_ref[1]
        ref[1,0,2,1]=(-0.219031743102)*x_ref[0]+(-0.402476762676)*x_ref[1]
        ref[1,1,0,0]=(-0.196889997171)*x_ref[0]+(0.675626499802)*x_ref[1]
        ref[1,1,0,1]=(1.23421246519)*x_ref[0]+(-1.72367262236)*x_ref[1]
        ref[1,1,1,0]=(1.35571029115)*x_ref[0]+(0.506057274132)*x_ref[1]
        ref[1,1,1,1]=(0.310699908104)*x_ref[0]+(-0.292826394288)*x_ref[1]
        ref[1,1,2,0]=(0.122289986824)*x_ref[0]+(-0.15947928964)*x_ref[1]
        ref[1,1,2,1]=(0.662808118991)*x_ref[0]+(-0.927311967494)*x_ref[1]
        ref[1,2,0,0]=(-0.215898581104)*x_ref[0]+(0.32902895413)*x_ref[1]
        ref[1,2,0,1]=(0.393145660435)*x_ref[0]+(1.49040999822)*x_ref[1]
        ref[1,2,1,0]=(0.904144553172)*x_ref[0]+(0.199605443931)*x_ref[1]
        ref[1,2,1,1]=(0.535979155678)*x_ref[0]+(1.78418572873)*x_ref[1]
        ref[1,2,2,0]=(1.15578250862)*x_ref[0]+(0.167811579149)*x_ref[1]
        ref[1,2,2,1]=(-0.0844120860286)*x_ref[0]+(-0.56656817877)*x_ref[1]
        ref[1,3,0,0]=(0.760581325264)*x_ref[0]+(1.00766443232)*x_ref[1]
        ref[1,3,0,1]=(-0.199244602194)*x_ref[0]+(-1.54110932148)*x_ref[1]
        ref[1,3,1,0]=(1.40934237781)*x_ref[0]+(0.700934912313)*x_ref[1]
        ref[1,3,1,1]=(-0.702507416172)*x_ref[0]+(-0.206530323456)*x_ref[1]
        ref[1,3,2,0]=(-0.234301811191)*x_ref[0]+(1.41101365089)*x_ref[1]
        ref[1,3,2,1]=(1.19609435879)*x_ref[0]+(0.295560183318)*x_ref[1]
        ref[1,4,0,0]=(1.15287205838)*x_ref[0]+(0.630564383314)*x_ref[1]
        ref[1,4,0,1]=(-0.993907995704)*x_ref[0]+(-0.30836527499)*x_ref[1]
        ref[1,4,1,0]=(-0.265158946854)*x_ref[0]+(0.689120014144)*x_ref[1]
        ref[1,4,1,1]=(0.503151681106)*x_ref[0]+(-0.928730587416)*x_ref[1]
        ref[1,4,2,0]=(-1.0599860135)*x_ref[0]+(-0.696932543908)*x_ref[1]
        ref[1,4,2,1]=(-0.365096260865)*x_ref[0]+(0.443129268519)*x_ref[1]
        ref[2,0,0,0]=(0.412604298623)*x_ref[0]+(0.591441458509)*x_ref[1]
        ref[2,0,0,1]=(0.536034627643)*x_ref[0]+(-0.735261974283)*x_ref[1]
        ref[2,0,1,0]=(-0.336900405282)*x_ref[0]+(-1.38493113859)*x_ref[1]
        ref[2,0,1,1]=(-0.50634870942)*x_ref[0]+(0.0376544898311)*x_ref[1]
        ref[2,0,2,0]=(-0.0595134635196)*x_ref[0]+(0.067285997446)*x_ref[1]
        ref[2,0,2,1]=(0.308894832145)*x_ref[0]+(-1.94600763342)*x_ref[1]
        ref[2,1,0,0]=(0.443019228067)*x_ref[0]+(-0.68858070865)*x_ref[1]
        ref[2,1,0,1]=(1.4461941226)*x_ref[0]+(0.711186579745)*x_ref[1]
        ref[2,1,1,0]=(0.178777491574)*x_ref[0]+(-0.746259056845)*x_ref[1]
        ref[2,1,1,1]=(1.24682492601)*x_ref[0]+(-0.987057469956)*x_ref[1]
        ref[2,1,2,0]=(1.1863711951)*x_ref[0]+(-0.853660028797)*x_ref[1]
        ref[2,1,2,1]=(1.13176028078)*x_ref[0]+(-0.887521745194)*x_ref[1]
        ref[2,2,0,0]=(0.37579402707)*x_ref[0]+(-0.462533529495)*x_ref[1]
        ref[2,2,0,1]=(-1.27778045393)*x_ref[0]+(0.383310656984)*x_ref[1]
        ref[2,2,1,0]=(-0.0871651954146)*x_ref[0]+(0.0266612015617)*x_ref[1]
        ref[2,2,1,1]=(1.26455972911)*x_ref[0]+(0.597556950133)*x_ref[1]
        ref[2,2,2,0]=(0.397071934531)*x_ref[0]+(-0.792146197976)*x_ref[1]
        ref[2,2,2,1]=(-0.531842419771)*x_ref[0]+(-0.437895983867)*x_ref[1]
        ref[2,3,0,0]=(0.253602494754)*x_ref[0]+(-0.943660021452)*x_ref[1]
        ref[2,3,0,1]=(0.381472581352)*x_ref[0]+(1.36193258173)*x_ref[1]
        ref[2,3,1,0]=(0.530962583291)*x_ref[0]+(0.853717668569)*x_ref[1]
        ref[2,3,1,1]=(-0.233925714076)*x_ref[0]+(0.20197385961)*x_ref[1]
        ref[2,3,2,0]=(0.148804235592)*x_ref[0]+(-0.116553369538)*x_ref[1]
        ref[2,3,2,1]=(-1.79360831582)*x_ref[0]+(1.62247077485)*x_ref[1]
        ref[2,4,0,0]=(0.617987809937)*x_ref[0]+(1.27685238582)*x_ref[1]
        ref[2,4,0,1]=(-0.359480838494)*x_ref[0]+(-0.11716388114)*x_ref[1]
        ref[2,4,1,0]=(0.204308581462)*x_ref[0]+(0.0241091749363)*x_ref[1]
        ref[2,4,1,1]=(0.424779981556)*x_ref[0]+(-1.01811123317)*x_ref[1]
        ref[2,4,2,0]=(-0.708751234946)*x_ref[0]+(0.714679051838)*x_ref[1]
        ref[2,4,2,1]=(-0.352350728857)*x_ref[0]+(1.47198192719)*x_ref[1]
        ref[3,0,0,0]=(-0.0592432179871)*x_ref[0]+(0.0900836798021)*x_ref[1]
        ref[3,0,0,1]=(-0.192266211858)*x_ref[0]+(0.561246968252)*x_ref[1]
        ref[3,0,1,0]=(0.30878942472)*x_ref[0]+(-0.578076897247)*x_ref[1]
        ref[3,0,1,1]=(-0.246226203035)*x_ref[0]+(0.68840556577)*x_ref[1]
        ref[3,0,2,0]=(-0.42433037357)*x_ref[0]+(-0.168333732109)*x_ref[1]
        ref[3,0,2,1]=(0.501931661511)*x_ref[0]+(1.53237442299)*x_ref[1]
        ref[3,1,0,0]=(-0.835618728532)*x_ref[0]+(-0.823254922183)*x_ref[1]
        ref[3,1,0,1]=(-0.609502259942)*x_ref[0]+(0.139549090645)*x_ref[1]
        ref[3,1,1,0]=(-0.690897498494)*x_ref[0]+(0.280254985362)*x_ref[1]
        ref[3,1,1,1]=(-0.89040023732)*x_ref[0]+(0.616118941926)*x_ref[1]
        ref[3,1,2,0]=(-0.90402399371)*x_ref[0]+(-0.399494199588)*x_ref[1]
        ref[3,1,2,1]=(1.31002070769)*x_ref[0]+(-0.26001592275)*x_ref[1]
        ref[3,2,0,0]=(-0.741667891927)*x_ref[0]+(-1.13827783123)*x_ref[1]
        ref[3,2,0,1]=(0.879168335296)*x_ref[0]+(0.232648259015)*x_ref[1]
        ref[3,2,1,0]=(0.737095222363)*x_ref[0]+(-1.26232530994)*x_ref[1]
        ref[3,2,1,1]=(-0.523976448837)*x_ref[0]+(0.480347424944)*x_ref[1]
        ref[3,2,2,0]=(-1.61727611988)*x_ref[0]+(-0.464564622523)*x_ref[1]
        ref[3,2,2,1]=(0.928282179697)*x_ref[0]+(0.629575166525)*x_ref[1]
        ref[3,3,0,0]=(0.748679757793)*x_ref[0]+(1.26002219505)*x_ref[1]
        ref[3,3,0,1]=(-0.497730460683)*x_ref[0]+(0.963512065302)*x_ref[1]
        ref[3,3,1,0]=(-0.999580985099)*x_ref[0]+(0.933789269497)*x_ref[1]
        ref[3,3,1,1]=(-0.151096110981)*x_ref[0]+(-0.739374723252)*x_ref[1]
        ref[3,3,2,0]=(0.257880503573)*x_ref[0]+(-1.07115978896)*x_ref[1]
        ref[3,3,2,1]=(0.188404246313)*x_ref[0]+(0.365000571954)*x_ref[1]
        ref[3,4,0,0]=(1.63625895457)*x_ref[0]+(-0.685581590196)*x_ref[1]
        ref[3,4,0,1]=(-0.665388096854)*x_ref[0]+(-0.9121332766)*x_ref[1]
        ref[3,4,1,0]=(1.76715282849)*x_ref[0]+(-1.15317031503)*x_ref[1]
        ref[3,4,1,1]=(-0.378469229808)*x_ref[0]+(-0.897880026236)*x_ref[1]
        ref[3,4,2,0]=(0.183976703061)*x_ref[0]+(1.76560300417)*x_ref[1]
        ref[3,4,2,1]=(-0.569272512314)*x_ref[0]+(1.09830336519)*x_ref[1]
      else:
        s[0,0,0,0]=(0.146894168765)*x[0]+(1.25822855062)*x[1]+(0.088660814219)*x[2]
        s[0,0,0,1]=(0.572576296123)*x[0]+(0.670191219644)*x[1]+(-0.280619677889)*x[2]
        s[0,0,1,0]=(-1.19817665348)*x[0]+(1.35545322011)*x[1]+(-0.279751222491)*x[2]
        s[0,0,1,1]=(-0.0558236682639)*x[0]+(-1.61042107055)*x[1]+(0.622456635293)*x[2]
        s[0,0,2,0]=(-0.154784205079)*x[0]+(-0.52314996688)*x[1]+(1.0590614932)*x[2]
        s[0,0,2,1]=(1.58101401758)*x[0]+(-0.610736715272)*x[1]+(-0.487233776225)*x[2]
        s[0,1,0,0]=(0.0117365392099)*x[0]+(1.02397373706)*x[1]+(0.407377316737)*x[2]
        s[0,1,0,1]=(0.460387220384)*x[0]+(1.10329384655)*x[1]+(0.252209465365)*x[2]
        s[0,1,1,0]=(-1.33153199048)*x[0]+(-0.759669253534)*x[1]+(-0.856223614724)*x[2]
        s[0,1,1,1]=(-1.00599668771)*x[0]+(0.321343926218)*x[1]+(-0.198860330101)*x[2]
        s[0,1,2,0]=(1.00741183642)*x[0]+(-0.148473398917)*x[1]+(-0.484181353094)*x[2]
        s[0,1,2,1]=(-1.46490648953)*x[0]+(0.621077170698)*x[1]+(0.802865695373)*x[2]
        s[0,2,0,0]=(1.18498612655)*x[0]+(-0.511705644809)*x[1]+(0.563568352278)*x[2]
        s[0,2,0,1]=(-0.807427977876)*x[0]+(-0.252372201752)*x[1]+(0.251722661898)*x[2]
        s[0,2,1,0]=(0.524572175367)*x[0]+(0.49906104527)*x[1]+(-0.601836374873)*x[2]
        s[0,2,1,1]=(-0.318976775137)*x[0]+(-0.153065654914)*x[1]+(-0.458888325378)*x[2]
        s[0,2,2,0]=(-1.45513892332)*x[0]+(-0.252635212625)*x[1]+(0.490153491036)*x[2]
        s[0,2,2,1]=(-0.439422986461)*x[0]+(-0.937409556421)*x[1]+(0.209106988492)*x[2]
        s[0,3,0,0]=(-0.26828401981)*x[0]+(1.70808529771)*x[1]+(0.431245149358)*x[2]
        s[0,3,0,1]=(-0.831846912527)*x[0]+(0.572822507173)*x[1]+(-0.283130753547)*x[2]
        s[0,3,1,0]=(-0.625617835832)*x[0]+(-1.25495997626)*x[1]+(1.28743784867)*x[2]
        s[0,3,1,1]=(1.47223047075)*x[0]+(0.274365456026)*x[1]+(1.34355769078)*x[2]
        s[0,3,2,0]=(-0.0506967319243)*x[0]+(1.16569056569)*x[1]+(-0.965008332019)*x[2]
        s[0,3,2,1]=(0.0106139754664)*x[0]+(0.274252087281)*x[1]+(-0.501970174754)*x[2]
        s[0,4,0,0]=(0.717476010799)*x[0]+(0.380533443018)*x[1]+(-0.304127891816)*x[2]
        s[0,4,0,1]=(-1.29133512255)*x[0]+(-1.16667771199)*x[1]+(0.275809086731)*x[2]
        s[0,4,1,0]=(0.137939049278)*x[0]+(-0.302950342448)*x[1]+(-0.28865204469)*x[2]
        s[0,4,1,1]=(-0.857457318663)*x[0]+(0.423143726447)*x[1]+(0.749889416366)*x[2]
        s[0,4,2,0]=(-0.994428027329)*x[0]+(-0.377780341115)*x[1]+(-0.997480394895)*x[2]
        s[0,4,2,1]=(-0.681653236311)*x[0]+(-0.091905911233)*x[1]+(-0.877007024631)*x[2]
        s[1,0,0,0]=(0.33375482228)*x[0]+(-1.08816236114)*x[1]+(0.154701633038)*x[2]
        s[1,0,0,1]=(0.979227592935)*x[0]+(0.571415419963)*x[1]+(-0.20312466228)*x[2]
        s[1,0,1,0]=(1.59673013189)*x[0]+(1.01626905603)*x[1]+(0.216806440372)*x[2]
        s[1,0,1,1]=(-0.276034446694)*x[0]+(-0.168494544051)*x[1]+(0.981220114947)*x[2]
        s[1,0,2,0]=(0.205085469341)*x[0]+(-0.507970934322)*x[1]+(-0.724745229941)*x[2]
        s[1,0,2,1]=(-0.629331749902)*x[0]+(-1.27086230989)*x[1]+(-1.08479549542)*x[2]
        s[1,1,0,0]=(-0.786670678498)*x[0]+(1.24461814343)*x[1]+(0.644579154985)*x[2]
        s[1,1,0,1]=(1.05414628662)*x[0]+(0.901590532372)*x[1]+(0.809519110188)*x[2]
        s[1,1,1,0]=(-0.679152609307)*x[0]+(0.84101357998)*x[1]+(1.14312188299)*x[2]
        s[1,1,1,1]=(0.137715540853)*x[0]+(1.42995779339)*x[1]+(-1.10477793041)*x[2]
        s[1,1,2,0]=(-0.0647098241647)*x[0]+(1.18580807605)*x[1]+(0.605076723926)*x[2]
        s[1,1,2,1]=(0.744824848213)*x[0]+(0.532106337173)*x[1]+(0.68490109894)*x[2]
        s[1,2,0,0]=(-0.557916122207)*x[0]+(1.06859221643)*x[1]+(0.823215271018)*x[2]
        s[1,2,0,1]=(0.00598565716481)*x[0]+(0.884417903372)*x[1]+(1.7921906066)*x[2]
        s[1,2,1,0]=(-0.47245962749)*x[0]+(-0.906624987584)*x[1]+(0.137792563282)*x[2]
        s[1,2,1,1]=(0.948613971685)*x[0]+(0.753281322508)*x[1]+(0.110929011008)*x[2]
        s[1,2,2,0]=(0.175178588707)*x[0]+(-0.796494364923)*x[1]+(0.390555269738)*x[2]
        s[1,2,2,1]=(-1.04516910973)*x[0]+(-0.921757404949)*x[1]+(0.845139972079)*x[2]
        s[1,3,0,0]=(-1.42392853478)*x[0]+(-0.458461019373)*x[1]+(-0.213286746949)*x[2]
        s[1,3,0,1]=(0.13996349254)*x[0]+(0.108946526608)*x[1]+(-0.848132079317)*x[2]
        s[1,3,1,0]=(-1.44188153548)*x[0]+(-0.108156975475)*x[1]+(0.0715891132579)*x[2]
        s[1,3,1,1]=(0.525856508836)*x[0]+(0.950609265517)*x[1]+(0.0127362240999)*x[2]
        s[1,3,2,0]=(-1.0845975079)*x[0]+(-1.24949492248)*x[1]+(-0.00240035631582)*x[2]
        s[1,3,2,1]=(0.642060593331)*x[0]+(1.46218158975)*x[1]+(1.32385980887)*x[2]
        s[1,4,0,0]=(0.0119712366759)*x[0]+(1.39772768926)*x[1]+(-0.570649410717)*x[2]
        s[1,4,0,1]=(0.320030773751)*x[0]+(0.358086736384)*x[1]+(-0.859721999492)*x[2]
        s[1,4,1,0]=(0.574417642497)*x[0]+(-0.421844733062)*x[1]+(0.611638265232)*x[2]
        s[1,4,1,1]=(-0.228413169897)*x[0]+(-0.0391189339225)*x[1]+(0.635031723033)*x[2]
        s[1,4,2,0]=(0.0424678701733)*x[0]+(0.261307736177)*x[1]+(-0.133428864005)*x[2]
        s[1,4,2,1]=(0.169996877966)*x[0]+(1.30730453844)*x[1]+(-1.07718562596)*x[2]
        s[2,0,0,0]=(0.147232851651)*x[0]+(-0.134174394382)*x[1]+(-0.785323160299)*x[2]
        s[2,0,0,1]=(1.29417433206)*x[0]+(-1.39241927595)*x[1]+(-0.224395758464)*x[2]
        s[2,0,1,0]=(1.60938474462)*x[0]+(0.68821531646)*x[1]+(0.453462382464)*x[2]
        s[2,0,1,1]=(0.63619844559)*x[0]+(-0.956987170375)*x[1]+(-1.24753100984)*x[2]
        s[2,0,2,0]=(1.14194732577)*x[0]+(0.973489037101)*x[1]+(0.187588261589)*x[2]
        s[2,0,2,1]=(-0.898956363323)*x[0]+(0.108722086141)*x[1]+(-1.62445422766)*x[2]
        s[2,1,0,0]=(0.127827687335)*x[0]+(0.216922403148)*x[1]+(-1.5099539458)*x[2]
        s[2,1,0,1]=(-0.769830743998)*x[0]+(1.46060444448)*x[1]+(-0.318171252653)*x[2]
        s[2,1,1,0]=(-0.311976479418)*x[0]+(-0.949061813767)*x[1]+(-0.237478430906)*x[2]
        s[2,1,1,1]=(1.27107354882)*x[0]+(-0.427981502964)*x[1]+(-0.457224808787)*x[2]
        s[2,1,2,0]=(-0.633029461167)*x[0]+(0.255322271264)*x[1]+(-1.79431610464)*x[2]
        s[2,1,2,1]=(1.05355586142)*x[0]+(-1.69722728443)*x[1]+(-1.11928223816)*x[2]
        s[2,2,0,0]=(1.15538109557)*x[0]+(-0.52692609967)*x[1]+(-0.686266572053)*x[2]
        s[2,2,0,1]=(-0.327531879955)*x[0]+(0.365760847752)*x[1]+(-0.0600793962176)*x[2]
        s[2,2,1,0]=(1.16249877038)*x[0]+(0.710438081472)*x[1]+(0.986719636688)*x[2]
        s[2,2,1,1]=(-1.22799525195)*x[0]+(0.748811223969)*x[1]+(0.370195043623)*x[2]
        s[2,2,2,0]=(0.976246115046)*x[0]+(1.10537245041)*x[1]+(-0.759735725531)*x[2]
        s[2,2,2,1]=(1.01411361404)*x[0]+(-0.000596974234967)*x[1]+(-1.39135228207)*x[2]
        s[2,3,0,0]=(-0.581025188916)*x[0]+(-1.43405762699)*x[1]+(-0.912365741093)*x[2]
        s[2,3,0,1]=(-0.0511110676196)*x[0]+(0.230142796427)*x[1]+(-1.58165960293)*x[2]
        s[2,3,1,0]=(-1.46709370576)*x[0]+(0.425035892802)*x[1]+(0.282281348864)*x[2]
        s[2,3,1,1]=(-0.35514320269)*x[0]+(-0.39238669714)*x[1]+(-0.0118498421145)*x[2]
        s[2,3,2,0]=(0.317197568125)*x[0]+(0.689803280183)*x[1]+(-0.238327580547)*x[2]
        s[2,3,2,1]=(1.31241465464)*x[0]+(0.548355955144)*x[1]+(0.0218217641073)*x[2]
        s[2,4,0,0]=(0.787240379652)*x[0]+(-0.906804255968)*x[1]+(-1.17498478521)*x[2]
        s[2,4,0,1]=(1.19721624988)*x[0]+(0.567119981635)*x[1]+(1.42310267091)*x[2]
        s[2,4,1,0]=(-0.199829256778)*x[0]+(0.28908416123)*x[1]+(0.491497583863)*x[2]
        s[2,4,1,1]=(0.336357763916)*x[0]+(-0.750103385624)*x[1]+(-1.19492486127)*x[2]
        s[2,4,2,0]=(1.509049274)*x[0]+(-1.28641285242)*x[1]+(-0.250003617687)*x[2]
        s[2,4,2,1]=(-1.23803199733)*x[0]+(0.688217821134)*x[1]+(-0.436478772021)*x[2]
        s[3,0,0,0]=(0.634885383655)*x[0]+(1.14728850769)*x[1]+(0.417496697665)*x[2]
        s[3,0,0,1]=(-0.326274842766)*x[0]+(-1.47136469943)*x[1]+(1.87373171786)*x[2]
        s[3,0,1,0]=(-0.140107043914)*x[0]+(-0.261565104318)*x[1]+(0.0780015778888)*x[2]
        s[3,0,1,1]=(0.770979582613)*x[0]+(-0.899853446798)*x[1]+(1.45115909188)*x[2]
        s[3,0,2,0]=(0.477865962499)*x[0]+(-0.0621575016815)*x[1]+(-1.28975967232)*x[2]
        s[3,0,2,1]=(-1.41031619045)*x[0]+(1.50155797656)*x[1]+(-1.44890527708)*x[2]
        s[3,1,0,0]=(-1.04632613119)*x[0]+(0.00196059522552)*x[1]+(0.139147980023)*x[2]
        s[3,1,0,1]=(0.217272687162)*x[0]+(0.295773880887)*x[1]+(0.474090779139)*x[2]
        s[3,1,1,0]=(1.50572386246)*x[0]+(0.868579168785)*x[1]+(-0.12394808575)*x[2]
        s[3,1,1,1]=(0.882176073715)*x[0]+(0.144766914417)*x[1]+(-0.408510755049)*x[2]
        s[3,1,2,0]=(-1.21767372555)*x[0]+(0.195682844047)*x[1]+(-1.70399790168)*x[2]
        s[3,1,2,1]=(-0.110311118233)*x[0]+(-0.721806016748)*x[1]+(1.81495268748)*x[2]
        s[3,2,0,0]=(-0.0441997228291)*x[0]+(0.0754984632459)*x[1]+(0.274177049743)*x[2]
        s[3,2,0,1]=(-0.48089133278)*x[0]+(-0.00712717635695)*x[1]+(0.242678324953)*x[2]
        s[3,2,1,0]=(0.48377042087)*x[0]+(1.4374627607)*x[1]+(-1.63994815623)*x[2]
        s[3,2,1,1]=(0.548308948512)*x[0]+(-0.587878136236)*x[1]+(-0.636242676063)*x[2]
        s[3,2,2,0]=(-1.62147144921)*x[0]+(-0.937713365221)*x[1]+(0.416583670897)*x[2]
        s[3,2,2,1]=(1.50617037978)*x[0]+(0.849964476529)*x[1]+(-0.421030022756)*x[2]
        s[3,3,0,0]=(-0.0258240269876)*x[0]+(1.09488855424)*x[1]+(-0.864078888453)*x[2]
        s[3,3,0,1]=(0.121803754668)*x[0]+(1.47064577586)*x[1]+(-1.15267381807)*x[2]
        s[3,3,1,0]=(0.7847153928)*x[0]+(1.19923938448)*x[1]+(-0.220983796213)*x[2]
        s[3,3,1,1]=(0.591086037588)*x[0]+(1.49019206942)*x[1]+(-0.764272587507)*x[2]
        s[3,3,2,0]=(-0.751721311032)*x[0]+(-1.41725473775)*x[1]+(-0.182681015042)*x[2]
        s[3,3,2,1]=(1.18789607758)*x[0]+(0.321544556274)*x[1]+(-0.37682906633)*x[2]
        s[3,4,0,0]=(-0.314136382522)*x[0]+(0.190342372946)*x[1]+(-0.0583598827644)*x[2]
        s[3,4,0,1]=(1.87879524007)*x[0]+(0.0472715596509)*x[1]+(0.849571713097)*x[2]
        s[3,4,1,0]=(0.816120005278)*x[0]+(0.296980756484)*x[1]+(0.251581386555)*x[2]
        s[3,4,1,1]=(0.12560195013)*x[0]+(-0.293571823053)*x[1]+(-1.82952568911)*x[2]
        s[3,4,2,0]=(1.03923438653)*x[0]+(-0.899000072461)*x[1]+(-1.48431797596)*x[2]
        s[3,4,2,1]=(-1.12443865282)*x[0]+(-0.200378451615)*x[1]+(-1.41505465511)*x[2]
        ref[0,0,0,0]=(0.146894168765)*x_ref[0]+(1.25822855062)*x_ref[1]+(0.088660814219)*x_ref[2]
        ref[0,0,0,1]=(0.572576296123)*x_ref[0]+(0.670191219644)*x_ref[1]+(-0.280619677889)*x_ref[2]
        ref[0,0,1,0]=(-1.19817665348)*x_ref[0]+(1.35545322011)*x_ref[1]+(-0.279751222491)*x_ref[2]
        ref[0,0,1,1]=(-0.0558236682639)*x_ref[0]+(-1.61042107055)*x_ref[1]+(0.622456635293)*x_ref[2]
        ref[0,0,2,0]=(-0.154784205079)*x_ref[0]+(-0.52314996688)*x_ref[1]+(1.0590614932)*x_ref[2]
        ref[0,0,2,1]=(1.58101401758)*x_ref[0]+(-0.610736715272)*x_ref[1]+(-0.487233776225)*x_ref[2]
        ref[0,1,0,0]=(0.0117365392099)*x_ref[0]+(1.02397373706)*x_ref[1]+(0.407377316737)*x_ref[2]
        ref[0,1,0,1]=(0.460387220384)*x_ref[0]+(1.10329384655)*x_ref[1]+(0.252209465365)*x_ref[2]
        ref[0,1,1,0]=(-1.33153199048)*x_ref[0]+(-0.759669253534)*x_ref[1]+(-0.856223614724)*x_ref[2]
        ref[0,1,1,1]=(-1.00599668771)*x_ref[0]+(0.321343926218)*x_ref[1]+(-0.198860330101)*x_ref[2]
        ref[0,1,2,0]=(1.00741183642)*x_ref[0]+(-0.148473398917)*x_ref[1]+(-0.484181353094)*x_ref[2]
        ref[0,1,2,1]=(-1.46490648953)*x_ref[0]+(0.621077170698)*x_ref[1]+(0.802865695373)*x_ref[2]
        ref[0,2,0,0]=(1.18498612655)*x_ref[0]+(-0.511705644809)*x_ref[1]+(0.563568352278)*x_ref[2]
        ref[0,2,0,1]=(-0.807427977876)*x_ref[0]+(-0.252372201752)*x_ref[1]+(0.251722661898)*x_ref[2]
        ref[0,2,1,0]=(0.524572175367)*x_ref[0]+(0.49906104527)*x_ref[1]+(-0.601836374873)*x_ref[2]
        ref[0,2,1,1]=(-0.318976775137)*x_ref[0]+(-0.153065654914)*x_ref[1]+(-0.458888325378)*x_ref[2]
        ref[0,2,2,0]=(-1.45513892332)*x_ref[0]+(-0.252635212625)*x_ref[1]+(0.490153491036)*x_ref[2]
        ref[0,2,2,1]=(-0.439422986461)*x_ref[0]+(-0.937409556421)*x_ref[1]+(0.209106988492)*x_ref[2]
        ref[0,3,0,0]=(-0.26828401981)*x_ref[0]+(1.70808529771)*x_ref[1]+(0.431245149358)*x_ref[2]
        ref[0,3,0,1]=(-0.831846912527)*x_ref[0]+(0.572822507173)*x_ref[1]+(-0.283130753547)*x_ref[2]
        ref[0,3,1,0]=(-0.625617835832)*x_ref[0]+(-1.25495997626)*x_ref[1]+(1.28743784867)*x_ref[2]
        ref[0,3,1,1]=(1.47223047075)*x_ref[0]+(0.274365456026)*x_ref[1]+(1.34355769078)*x_ref[2]
        ref[0,3,2,0]=(-0.0506967319243)*x_ref[0]+(1.16569056569)*x_ref[1]+(-0.965008332019)*x_ref[2]
        ref[0,3,2,1]=(0.0106139754664)*x_ref[0]+(0.274252087281)*x_ref[1]+(-0.501970174754)*x_ref[2]
        ref[0,4,0,0]=(0.717476010799)*x_ref[0]+(0.380533443018)*x_ref[1]+(-0.304127891816)*x_ref[2]
        ref[0,4,0,1]=(-1.29133512255)*x_ref[0]+(-1.16667771199)*x_ref[1]+(0.275809086731)*x_ref[2]
        ref[0,4,1,0]=(0.137939049278)*x_ref[0]+(-0.302950342448)*x_ref[1]+(-0.28865204469)*x_ref[2]
        ref[0,4,1,1]=(-0.857457318663)*x_ref[0]+(0.423143726447)*x_ref[1]+(0.749889416366)*x_ref[2]
        ref[0,4,2,0]=(-0.994428027329)*x_ref[0]+(-0.377780341115)*x_ref[1]+(-0.997480394895)*x_ref[2]
        ref[0,4,2,1]=(-0.681653236311)*x_ref[0]+(-0.091905911233)*x_ref[1]+(-0.877007024631)*x_ref[2]
        ref[1,0,0,0]=(0.33375482228)*x_ref[0]+(-1.08816236114)*x_ref[1]+(0.154701633038)*x_ref[2]
        ref[1,0,0,1]=(0.979227592935)*x_ref[0]+(0.571415419963)*x_ref[1]+(-0.20312466228)*x_ref[2]
        ref[1,0,1,0]=(1.59673013189)*x_ref[0]+(1.01626905603)*x_ref[1]+(0.216806440372)*x_ref[2]
        ref[1,0,1,1]=(-0.276034446694)*x_ref[0]+(-0.168494544051)*x_ref[1]+(0.981220114947)*x_ref[2]
        ref[1,0,2,0]=(0.205085469341)*x_ref[0]+(-0.507970934322)*x_ref[1]+(-0.724745229941)*x_ref[2]
        ref[1,0,2,1]=(-0.629331749902)*x_ref[0]+(-1.27086230989)*x_ref[1]+(-1.08479549542)*x_ref[2]
        ref[1,1,0,0]=(-0.786670678498)*x_ref[0]+(1.24461814343)*x_ref[1]+(0.644579154985)*x_ref[2]
        ref[1,1,0,1]=(1.05414628662)*x_ref[0]+(0.901590532372)*x_ref[1]+(0.809519110188)*x_ref[2]
        ref[1,1,1,0]=(-0.679152609307)*x_ref[0]+(0.84101357998)*x_ref[1]+(1.14312188299)*x_ref[2]
        ref[1,1,1,1]=(0.137715540853)*x_ref[0]+(1.42995779339)*x_ref[1]+(-1.10477793041)*x_ref[2]
        ref[1,1,2,0]=(-0.0647098241647)*x_ref[0]+(1.18580807605)*x_ref[1]+(0.605076723926)*x_ref[2]
        ref[1,1,2,1]=(0.744824848213)*x_ref[0]+(0.532106337173)*x_ref[1]+(0.68490109894)*x_ref[2]
        ref[1,2,0,0]=(-0.557916122207)*x_ref[0]+(1.06859221643)*x_ref[1]+(0.823215271018)*x_ref[2]
        ref[1,2,0,1]=(0.00598565716481)*x_ref[0]+(0.884417903372)*x_ref[1]+(1.7921906066)*x_ref[2]
        ref[1,2,1,0]=(-0.47245962749)*x_ref[0]+(-0.906624987584)*x_ref[1]+(0.137792563282)*x_ref[2]
        ref[1,2,1,1]=(0.948613971685)*x_ref[0]+(0.753281322508)*x_ref[1]+(0.110929011008)*x_ref[2]
        ref[1,2,2,0]=(0.175178588707)*x_ref[0]+(-0.796494364923)*x_ref[1]+(0.390555269738)*x_ref[2]
        ref[1,2,2,1]=(-1.04516910973)*x_ref[0]+(-0.921757404949)*x_ref[1]+(0.845139972079)*x_ref[2]
        ref[1,3,0,0]=(-1.42392853478)*x_ref[0]+(-0.458461019373)*x_ref[1]+(-0.213286746949)*x_ref[2]
        ref[1,3,0,1]=(0.13996349254)*x_ref[0]+(0.108946526608)*x_ref[1]+(-0.848132079317)*x_ref[2]
        ref[1,3,1,0]=(-1.44188153548)*x_ref[0]+(-0.108156975475)*x_ref[1]+(0.0715891132579)*x_ref[2]
        ref[1,3,1,1]=(0.525856508836)*x_ref[0]+(0.950609265517)*x_ref[1]+(0.0127362240999)*x_ref[2]
        ref[1,3,2,0]=(-1.0845975079)*x_ref[0]+(-1.24949492248)*x_ref[1]+(-0.00240035631582)*x_ref[2]
        ref[1,3,2,1]=(0.642060593331)*x_ref[0]+(1.46218158975)*x_ref[1]+(1.32385980887)*x_ref[2]
        ref[1,4,0,0]=(0.0119712366759)*x_ref[0]+(1.39772768926)*x_ref[1]+(-0.570649410717)*x_ref[2]
        ref[1,4,0,1]=(0.320030773751)*x_ref[0]+(0.358086736384)*x_ref[1]+(-0.859721999492)*x_ref[2]
        ref[1,4,1,0]=(0.574417642497)*x_ref[0]+(-0.421844733062)*x_ref[1]+(0.611638265232)*x_ref[2]
        ref[1,4,1,1]=(-0.228413169897)*x_ref[0]+(-0.0391189339225)*x_ref[1]+(0.635031723033)*x_ref[2]
        ref[1,4,2,0]=(0.0424678701733)*x_ref[0]+(0.261307736177)*x_ref[1]+(-0.133428864005)*x_ref[2]
        ref[1,4,2,1]=(0.169996877966)*x_ref[0]+(1.30730453844)*x_ref[1]+(-1.07718562596)*x_ref[2]
        ref[2,0,0,0]=(0.147232851651)*x_ref[0]+(-0.134174394382)*x_ref[1]+(-0.785323160299)*x_ref[2]
        ref[2,0,0,1]=(1.29417433206)*x_ref[0]+(-1.39241927595)*x_ref[1]+(-0.224395758464)*x_ref[2]
        ref[2,0,1,0]=(1.60938474462)*x_ref[0]+(0.68821531646)*x_ref[1]+(0.453462382464)*x_ref[2]
        ref[2,0,1,1]=(0.63619844559)*x_ref[0]+(-0.956987170375)*x_ref[1]+(-1.24753100984)*x_ref[2]
        ref[2,0,2,0]=(1.14194732577)*x_ref[0]+(0.973489037101)*x_ref[1]+(0.187588261589)*x_ref[2]
        ref[2,0,2,1]=(-0.898956363323)*x_ref[0]+(0.108722086141)*x_ref[1]+(-1.62445422766)*x_ref[2]
        ref[2,1,0,0]=(0.127827687335)*x_ref[0]+(0.216922403148)*x_ref[1]+(-1.5099539458)*x_ref[2]
        ref[2,1,0,1]=(-0.769830743998)*x_ref[0]+(1.46060444448)*x_ref[1]+(-0.318171252653)*x_ref[2]
        ref[2,1,1,0]=(-0.311976479418)*x_ref[0]+(-0.949061813767)*x_ref[1]+(-0.237478430906)*x_ref[2]
        ref[2,1,1,1]=(1.27107354882)*x_ref[0]+(-0.427981502964)*x_ref[1]+(-0.457224808787)*x_ref[2]
        ref[2,1,2,0]=(-0.633029461167)*x_ref[0]+(0.255322271264)*x_ref[1]+(-1.79431610464)*x_ref[2]
        ref[2,1,2,1]=(1.05355586142)*x_ref[0]+(-1.69722728443)*x_ref[1]+(-1.11928223816)*x_ref[2]
        ref[2,2,0,0]=(1.15538109557)*x_ref[0]+(-0.52692609967)*x_ref[1]+(-0.686266572053)*x_ref[2]
        ref[2,2,0,1]=(-0.327531879955)*x_ref[0]+(0.365760847752)*x_ref[1]+(-0.0600793962176)*x_ref[2]
        ref[2,2,1,0]=(1.16249877038)*x_ref[0]+(0.710438081472)*x_ref[1]+(0.986719636688)*x_ref[2]
        ref[2,2,1,1]=(-1.22799525195)*x_ref[0]+(0.748811223969)*x_ref[1]+(0.370195043623)*x_ref[2]
        ref[2,2,2,0]=(0.976246115046)*x_ref[0]+(1.10537245041)*x_ref[1]+(-0.759735725531)*x_ref[2]
        ref[2,2,2,1]=(1.01411361404)*x_ref[0]+(-0.000596974234967)*x_ref[1]+(-1.39135228207)*x_ref[2]
        ref[2,3,0,0]=(-0.581025188916)*x_ref[0]+(-1.43405762699)*x_ref[1]+(-0.912365741093)*x_ref[2]
        ref[2,3,0,1]=(-0.0511110676196)*x_ref[0]+(0.230142796427)*x_ref[1]+(-1.58165960293)*x_ref[2]
        ref[2,3,1,0]=(-1.46709370576)*x_ref[0]+(0.425035892802)*x_ref[1]+(0.282281348864)*x_ref[2]
        ref[2,3,1,1]=(-0.35514320269)*x_ref[0]+(-0.39238669714)*x_ref[1]+(-0.0118498421145)*x_ref[2]
        ref[2,3,2,0]=(0.317197568125)*x_ref[0]+(0.689803280183)*x_ref[1]+(-0.238327580547)*x_ref[2]
        ref[2,3,2,1]=(1.31241465464)*x_ref[0]+(0.548355955144)*x_ref[1]+(0.0218217641073)*x_ref[2]
        ref[2,4,0,0]=(0.787240379652)*x_ref[0]+(-0.906804255968)*x_ref[1]+(-1.17498478521)*x_ref[2]
        ref[2,4,0,1]=(1.19721624988)*x_ref[0]+(0.567119981635)*x_ref[1]+(1.42310267091)*x_ref[2]
        ref[2,4,1,0]=(-0.199829256778)*x_ref[0]+(0.28908416123)*x_ref[1]+(0.491497583863)*x_ref[2]
        ref[2,4,1,1]=(0.336357763916)*x_ref[0]+(-0.750103385624)*x_ref[1]+(-1.19492486127)*x_ref[2]
        ref[2,4,2,0]=(1.509049274)*x_ref[0]+(-1.28641285242)*x_ref[1]+(-0.250003617687)*x_ref[2]
        ref[2,4,2,1]=(-1.23803199733)*x_ref[0]+(0.688217821134)*x_ref[1]+(-0.436478772021)*x_ref[2]
        ref[3,0,0,0]=(0.634885383655)*x_ref[0]+(1.14728850769)*x_ref[1]+(0.417496697665)*x_ref[2]
        ref[3,0,0,1]=(-0.326274842766)*x_ref[0]+(-1.47136469943)*x_ref[1]+(1.87373171786)*x_ref[2]
        ref[3,0,1,0]=(-0.140107043914)*x_ref[0]+(-0.261565104318)*x_ref[1]+(0.0780015778888)*x_ref[2]
        ref[3,0,1,1]=(0.770979582613)*x_ref[0]+(-0.899853446798)*x_ref[1]+(1.45115909188)*x_ref[2]
        ref[3,0,2,0]=(0.477865962499)*x_ref[0]+(-0.0621575016815)*x_ref[1]+(-1.28975967232)*x_ref[2]
        ref[3,0,2,1]=(-1.41031619045)*x_ref[0]+(1.50155797656)*x_ref[1]+(-1.44890527708)*x_ref[2]
        ref[3,1,0,0]=(-1.04632613119)*x_ref[0]+(0.00196059522552)*x_ref[1]+(0.139147980023)*x_ref[2]
        ref[3,1,0,1]=(0.217272687162)*x_ref[0]+(0.295773880887)*x_ref[1]+(0.474090779139)*x_ref[2]
        ref[3,1,1,0]=(1.50572386246)*x_ref[0]+(0.868579168785)*x_ref[1]+(-0.12394808575)*x_ref[2]
        ref[3,1,1,1]=(0.882176073715)*x_ref[0]+(0.144766914417)*x_ref[1]+(-0.408510755049)*x_ref[2]
        ref[3,1,2,0]=(-1.21767372555)*x_ref[0]+(0.195682844047)*x_ref[1]+(-1.70399790168)*x_ref[2]
        ref[3,1,2,1]=(-0.110311118233)*x_ref[0]+(-0.721806016748)*x_ref[1]+(1.81495268748)*x_ref[2]
        ref[3,2,0,0]=(-0.0441997228291)*x_ref[0]+(0.0754984632459)*x_ref[1]+(0.274177049743)*x_ref[2]
        ref[3,2,0,1]=(-0.48089133278)*x_ref[0]+(-0.00712717635695)*x_ref[1]+(0.242678324953)*x_ref[2]
        ref[3,2,1,0]=(0.48377042087)*x_ref[0]+(1.4374627607)*x_ref[1]+(-1.63994815623)*x_ref[2]
        ref[3,2,1,1]=(0.548308948512)*x_ref[0]+(-0.587878136236)*x_ref[1]+(-0.636242676063)*x_ref[2]
        ref[3,2,2,0]=(-1.62147144921)*x_ref[0]+(-0.937713365221)*x_ref[1]+(0.416583670897)*x_ref[2]
        ref[3,2,2,1]=(1.50617037978)*x_ref[0]+(0.849964476529)*x_ref[1]+(-0.421030022756)*x_ref[2]
        ref[3,3,0,0]=(-0.0258240269876)*x_ref[0]+(1.09488855424)*x_ref[1]+(-0.864078888453)*x_ref[2]
        ref[3,3,0,1]=(0.121803754668)*x_ref[0]+(1.47064577586)*x_ref[1]+(-1.15267381807)*x_ref[2]
        ref[3,3,1,0]=(0.7847153928)*x_ref[0]+(1.19923938448)*x_ref[1]+(-0.220983796213)*x_ref[2]
        ref[3,3,1,1]=(0.591086037588)*x_ref[0]+(1.49019206942)*x_ref[1]+(-0.764272587507)*x_ref[2]
        ref[3,3,2,0]=(-0.751721311032)*x_ref[0]+(-1.41725473775)*x_ref[1]+(-0.182681015042)*x_ref[2]
        ref[3,3,2,1]=(1.18789607758)*x_ref[0]+(0.321544556274)*x_ref[1]+(-0.37682906633)*x_ref[2]
        ref[3,4,0,0]=(-0.314136382522)*x_ref[0]+(0.190342372946)*x_ref[1]+(-0.0583598827644)*x_ref[2]
        ref[3,4,0,1]=(1.87879524007)*x_ref[0]+(0.0472715596509)*x_ref[1]+(0.849571713097)*x_ref[2]
        ref[3,4,1,0]=(0.816120005278)*x_ref[0]+(0.296980756484)*x_ref[1]+(0.251581386555)*x_ref[2]
        ref[3,4,1,1]=(0.12560195013)*x_ref[0]+(-0.293571823053)*x_ref[1]+(-1.82952568911)*x_ref[2]
        ref[3,4,2,0]=(1.03923438653)*x_ref[0]+(-0.899000072461)*x_ref[1]+(-1.48431797596)*x_ref[2]
        ref[3,4,2,1]=(-1.12443865282)*x_ref[0]+(-0.200378451615)*x_ref[1]+(-1.41505465511)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.775940204341)*x[0]**o+(0.408012597308)*x[0]+(-0.526684561152)*x[1]**o+(-0.900147960936)*x[1]
        ref=(0.775940204341)*x_ref[0]**o+(0.408012597308)*x_ref[0]+(-0.526684561152)*x_ref[1]**o+(-0.900147960936)*x_ref[1]
      else:
        s=(-0.0117662561618)*x[0]**o+(0.813445281753)*x[0]+(-0.179209673046)*x[1]**o+(-0.841498005448)*x[1]+(0.660691793582)*x[2]**o+(0.149938868852)*x[2]
        ref=(-0.0117662561618)*x_ref[0]**o+(0.813445281753)*x_ref[0]+(-0.179209673046)*x_ref[1]**o+(-0.841498005448)*x_ref[1]+(0.660691793582)*x_ref[2]**o+(0.149938868852)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.940436734602)*x[0]**o+(-0.499319247438)*x[0]+(-0.87358475558)*x[1]**o+(-0.243771941219)*x[1]
        s[1]=(0.289360868897)*x[0]**o+(0.808574135723)*x[0]+(0.881242535917)*x[1]**o+(0.452398398881)*x[1]
        ref[0]=(0.940436734602)*x_ref[0]**o+(-0.499319247438)*x_ref[0]+(-0.87358475558)*x_ref[1]**o+(-0.243771941219)*x_ref[1]
        ref[1]=(0.289360868897)*x_ref[0]**o+(0.808574135723)*x_ref[0]+(0.881242535917)*x_ref[1]**o+(0.452398398881)*x_ref[1]
      else:
        s[0]=(-0.354278440772)*x[0]**o+(-0.0252551558693)*x[0]+(-0.955840831777)*x[1]**o+(-0.183292049624)*x[1]+(0.132001820383)*x[2]**o+(0.968364405641)*x[2]
        s[1]=(0.00318403616402)*x[0]**o+(0.278370320241)*x[0]+(0.451122966354)*x[1]**o+(0.0275444352047)*x[1]+(-0.0499740581816)*x[2]**o+(0.402684856351)*x[2]
        ref[0]=(-0.354278440772)*x_ref[0]**o+(-0.0252551558693)*x_ref[0]+(-0.955840831777)*x_ref[1]**o+(-0.183292049624)*x_ref[1]+(0.132001820383)*x_ref[2]**o+(0.968364405641)*x_ref[2]
        ref[1]=(0.00318403616402)*x_ref[0]**o+(0.278370320241)*x_ref[0]+(0.451122966354)*x_ref[1]**o+(0.0275444352047)*x_ref[1]+(-0.0499740581816)*x_ref[2]**o+(0.402684856351)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.877859091933)*x[0]**o+(-0.169762567932)*x[0]+(0.552958849443)*x[1]**o+(-0.11599142298)*x[1]
        s[0,1]=(0.471005726569)*x[0]**o+(-0.632583456078)*x[0]+(-0.360082824936)*x[1]**o+(-0.967781840973)*x[1]
        s[0,2]=(0.923269093961)*x[0]**o+(0.0574842386662)*x[0]+(-0.530617507423)*x[1]**o+(0.0473937183531)*x[1]
        s[0,3]=(0.824862043209)*x[0]**o+(0.698896112144)*x[0]+(-0.383808246808)*x[1]**o+(0.24923604362)*x[1]
        s[0,4]=(0.297428271453)*x[0]**o+(-0.18963891196)*x[0]+(-0.16010009844)*x[1]**o+(-0.0635747494778)*x[1]
        s[1,0]=(0.212999154122)*x[0]**o+(0.495137279523)*x[0]+(0.532511060201)*x[1]**o+(0.320057408221)*x[1]
        s[1,1]=(-0.994527399384)*x[0]**o+(0.0991063567708)*x[0]+(0.0375586255455)*x[1]**o+(-0.306126578081)*x[1]
        s[1,2]=(-0.140745202296)*x[0]**o+(-0.967168602143)*x[0]+(0.977938661252)*x[1]**o+(0.97796652405)*x[1]
        s[1,3]=(0.0508848805668)*x[0]**o+(-0.370907761322)*x[0]+(0.979063533914)*x[1]**o+(0.922258254977)*x[1]
        s[1,4]=(0.630127237807)*x[0]**o+(0.712161343099)*x[0]+(-0.15778527275)*x[1]**o+(0.590043042293)*x[1]
        s[2,0]=(-0.531086529572)*x[0]**o+(0.467450157091)*x[0]+(0.234837506275)*x[1]**o+(0.407064919776)*x[1]
        s[2,1]=(-0.056859895221)*x[0]**o+(0.533125022835)*x[0]+(0.904938824716)*x[1]**o+(0.793055507712)*x[1]
        s[2,2]=(-0.551130390502)*x[0]**o+(-0.01664955762)*x[0]+(0.788924790297)*x[1]**o+(-0.518737057199)*x[1]
        s[2,3]=(-0.593665554555)*x[0]**o+(-0.936497023625)*x[0]+(0.553767265739)*x[1]**o+(0.650790290426)*x[1]
        s[2,4]=(0.295642143189)*x[0]**o+(0.996496426422)*x[0]+(0.269107887571)*x[1]**o+(0.441099854152)*x[1]
        s[3,0]=(0.348944839773)*x[0]**o+(0.0364801849524)*x[0]+(0.00330101221443)*x[1]**o+(0.806022278452)*x[1]
        s[3,1]=(-0.620797900903)*x[0]**o+(0.516037923834)*x[0]+(0.861530193737)*x[1]**o+(0.466776529186)*x[1]
        s[3,2]=(0.055887897583)*x[0]**o+(-0.536336196157)*x[0]+(-0.889596296008)*x[1]**o+(0.299571474478)*x[1]
        s[3,3]=(-0.862274293699)*x[0]**o+(0.740404601136)*x[0]+(0.20012481659)*x[1]**o+(0.0655004723729)*x[1]
        s[3,4]=(0.561396042945)*x[0]**o+(-0.58695912231)*x[0]+(0.843743224177)*x[1]**o+(0.549571256677)*x[1]
        ref[0,0]=(0.877859091933)*x_ref[0]**o+(-0.169762567932)*x_ref[0]+(0.552958849443)*x_ref[1]**o+(-0.11599142298)*x_ref[1]
        ref[0,1]=(0.471005726569)*x_ref[0]**o+(-0.632583456078)*x_ref[0]+(-0.360082824936)*x_ref[1]**o+(-0.967781840973)*x_ref[1]
        ref[0,2]=(0.923269093961)*x_ref[0]**o+(0.0574842386662)*x_ref[0]+(-0.530617507423)*x_ref[1]**o+(0.0473937183531)*x_ref[1]
        ref[0,3]=(0.824862043209)*x_ref[0]**o+(0.698896112144)*x_ref[0]+(-0.383808246808)*x_ref[1]**o+(0.24923604362)*x_ref[1]
        ref[0,4]=(0.297428271453)*x_ref[0]**o+(-0.18963891196)*x_ref[0]+(-0.16010009844)*x_ref[1]**o+(-0.0635747494778)*x_ref[1]
        ref[1,0]=(0.212999154122)*x_ref[0]**o+(0.495137279523)*x_ref[0]+(0.532511060201)*x_ref[1]**o+(0.320057408221)*x_ref[1]
        ref[1,1]=(-0.994527399384)*x_ref[0]**o+(0.0991063567708)*x_ref[0]+(0.0375586255455)*x_ref[1]**o+(-0.306126578081)*x_ref[1]
        ref[1,2]=(-0.140745202296)*x_ref[0]**o+(-0.967168602143)*x_ref[0]+(0.977938661252)*x_ref[1]**o+(0.97796652405)*x_ref[1]
        ref[1,3]=(0.0508848805668)*x_ref[0]**o+(-0.370907761322)*x_ref[0]+(0.979063533914)*x_ref[1]**o+(0.922258254977)*x_ref[1]
        ref[1,4]=(0.630127237807)*x_ref[0]**o+(0.712161343099)*x_ref[0]+(-0.15778527275)*x_ref[1]**o+(0.590043042293)*x_ref[1]
        ref[2,0]=(-0.531086529572)*x_ref[0]**o+(0.467450157091)*x_ref[0]+(0.234837506275)*x_ref[1]**o+(0.407064919776)*x_ref[1]
        ref[2,1]=(-0.056859895221)*x_ref[0]**o+(0.533125022835)*x_ref[0]+(0.904938824716)*x_ref[1]**o+(0.793055507712)*x_ref[1]
        ref[2,2]=(-0.551130390502)*x_ref[0]**o+(-0.01664955762)*x_ref[0]+(0.788924790297)*x_ref[1]**o+(-0.518737057199)*x_ref[1]
        ref[2,3]=(-0.593665554555)*x_ref[0]**o+(-0.936497023625)*x_ref[0]+(0.553767265739)*x_ref[1]**o+(0.650790290426)*x_ref[1]
        ref[2,4]=(0.295642143189)*x_ref[0]**o+(0.996496426422)*x_ref[0]+(0.269107887571)*x_ref[1]**o+(0.441099854152)*x_ref[1]
        ref[3,0]=(0.348944839773)*x_ref[0]**o+(0.0364801849524)*x_ref[0]+(0.00330101221443)*x_ref[1]**o+(0.806022278452)*x_ref[1]
        ref[3,1]=(-0.620797900903)*x_ref[0]**o+(0.516037923834)*x_ref[0]+(0.861530193737)*x_ref[1]**o+(0.466776529186)*x_ref[1]
        ref[3,2]=(0.055887897583)*x_ref[0]**o+(-0.536336196157)*x_ref[0]+(-0.889596296008)*x_ref[1]**o+(0.299571474478)*x_ref[1]
        ref[3,3]=(-0.862274293699)*x_ref[0]**o+(0.740404601136)*x_ref[0]+(0.20012481659)*x_ref[1]**o+(0.0655004723729)*x_ref[1]
        ref[3,4]=(0.561396042945)*x_ref[0]**o+(-0.58695912231)*x_ref[0]+(0.843743224177)*x_ref[1]**o+(0.549571256677)*x_ref[1]
      else:
        s[0,0]=(-0.169971719578)*x[0]**o+(-0.490891054826)*x[0]+(0.744482943882)*x[1]**o+(0.353623823086)*x[1]+(0.265501620199)*x[2]**o+(-0.629048564143)*x[2]
        s[0,1]=(0.213635035052)*x[0]**o+(0.584035034456)*x[0]+(0.133486854141)*x[1]**o+(-0.563257345568)*x[1]+(-0.846099536223)*x[2]**o+(-0.854728289866)*x[2]
        s[0,2]=(0.342184893432)*x[0]**o+(0.515624951667)*x[0]+(-0.136008144001)*x[1]**o+(-0.280682548245)*x[1]+(-0.927880146409)*x[2]**o+(0.23737193095)*x[2]
        s[0,3]=(0.0547841640747)*x[0]**o+(-0.969984349739)*x[0]+(-0.155447158154)*x[1]**o+(-0.495088348804)*x[1]+(0.879915433385)*x[2]**o+(0.438389550488)*x[2]
        s[0,4]=(0.15107258681)*x[0]**o+(0.878036814589)*x[0]+(0.875901664353)*x[1]**o+(0.455383770416)*x[1]+(0.316791478263)*x[2]**o+(0.13991508459)*x[2]
        s[1,0]=(0.627219372033)*x[0]**o+(0.0514202591569)*x[0]+(0.696459316229)*x[1]**o+(0.232864735057)*x[1]+(0.516267189254)*x[2]**o+(-0.277533957737)*x[2]
        s[1,1]=(-0.474165959949)*x[0]**o+(0.875608881493)*x[0]+(-0.432866681632)*x[1]**o+(0.165933079742)*x[1]+(-0.0745331372502)*x[2]**o+(-0.0168977805625)*x[2]
        s[1,2]=(-0.228391157708)*x[0]**o+(-0.294221312175)*x[0]+(-0.121429334894)*x[1]**o+(-0.557107244165)*x[1]+(-0.97514025916)*x[2]**o+(0.041970524119)*x[2]
        s[1,3]=(0.344978415951)*x[0]**o+(0.319420517862)*x[0]+(0.802947508962)*x[1]**o+(0.169628742731)*x[1]+(-0.380928178408)*x[2]**o+(0.272551034104)*x[2]
        s[1,4]=(0.613832101706)*x[0]**o+(-0.332306428069)*x[0]+(0.02563364621)*x[1]**o+(-0.415952850048)*x[1]+(0.673350831947)*x[2]**o+(0.713348183687)*x[2]
        s[2,0]=(0.290558527133)*x[0]**o+(0.135335532216)*x[0]+(-0.352459421266)*x[1]**o+(-0.854183560051)*x[1]+(0.581843223923)*x[2]**o+(0.97378057053)*x[2]
        s[2,1]=(-0.328513503613)*x[0]**o+(-0.205196873021)*x[0]+(-0.910190961873)*x[1]**o+(0.568820988101)*x[1]+(0.495810061604)*x[2]**o+(-0.761749884838)*x[2]
        s[2,2]=(-0.00783916993558)*x[0]**o+(0.898831843878)*x[0]+(0.26265892346)*x[1]**o+(-0.654638473159)*x[1]+(-0.713198105003)*x[2]**o+(-0.59350752607)*x[2]
        s[2,3]=(-0.881651184273)*x[0]**o+(-0.834213060843)*x[0]+(-0.885142407575)*x[1]**o+(0.206850906485)*x[1]+(0.372454818886)*x[2]**o+(0.159769678242)*x[2]
        s[2,4]=(-0.937735543761)*x[0]**o+(0.529592561341)*x[0]+(-0.936499401425)*x[1]**o+(0.638973583422)*x[1]+(0.333562215313)*x[2]**o+(0.770257986138)*x[2]
        s[3,0]=(0.866403926159)*x[0]**o+(-0.567381195064)*x[0]+(-0.689200549972)*x[1]**o+(0.0278246156735)*x[1]+(-0.267522213281)*x[2]**o+(0.536314895418)*x[2]
        s[3,1]=(-0.561464080998)*x[0]**o+(-0.544070025131)*x[0]+(0.73146085432)*x[1]**o+(0.181244005603)*x[1]+(0.63008027542)*x[2]**o+(0.442771325928)*x[2]
        s[3,2]=(0.12906843064)*x[0]**o+(-0.736188526885)*x[0]+(0.675691533379)*x[1]**o+(0.975452480209)*x[1]+(0.877415310424)*x[2]**o+(-0.281346067495)*x[2]
        s[3,3]=(0.895295566131)*x[0]**o+(0.384502745469)*x[0]+(-0.6558585454)*x[1]**o+(-0.990764124757)*x[1]+(0.089950807223)*x[2]**o+(-0.117514839208)*x[2]
        s[3,4]=(0.908567029179)*x[0]**o+(0.130934285429)*x[0]+(-0.0913907665269)*x[1]**o+(-0.483606581717)*x[1]+(0.316018801782)*x[2]**o+(-0.371754102357)*x[2]
        ref[0,0]=(-0.169971719578)*x_ref[0]**o+(-0.490891054826)*x_ref[0]+(0.744482943882)*x_ref[1]**o+(0.353623823086)*x_ref[1]+(0.265501620199)*x_ref[2]**o+(-0.629048564143)*x_ref[2]
        ref[0,1]=(0.213635035052)*x_ref[0]**o+(0.584035034456)*x_ref[0]+(0.133486854141)*x_ref[1]**o+(-0.563257345568)*x_ref[1]+(-0.846099536223)*x_ref[2]**o+(-0.854728289866)*x_ref[2]
        ref[0,2]=(0.342184893432)*x_ref[0]**o+(0.515624951667)*x_ref[0]+(-0.136008144001)*x_ref[1]**o+(-0.280682548245)*x_ref[1]+(-0.927880146409)*x_ref[2]**o+(0.23737193095)*x_ref[2]
        ref[0,3]=(0.0547841640747)*x_ref[0]**o+(-0.969984349739)*x_ref[0]+(-0.155447158154)*x_ref[1]**o+(-0.495088348804)*x_ref[1]+(0.879915433385)*x_ref[2]**o+(0.438389550488)*x_ref[2]
        ref[0,4]=(0.15107258681)*x_ref[0]**o+(0.878036814589)*x_ref[0]+(0.875901664353)*x_ref[1]**o+(0.455383770416)*x_ref[1]+(0.316791478263)*x_ref[2]**o+(0.13991508459)*x_ref[2]
        ref[1,0]=(0.627219372033)*x_ref[0]**o+(0.0514202591569)*x_ref[0]+(0.696459316229)*x_ref[1]**o+(0.232864735057)*x_ref[1]+(0.516267189254)*x_ref[2]**o+(-0.277533957737)*x_ref[2]
        ref[1,1]=(-0.474165959949)*x_ref[0]**o+(0.875608881493)*x_ref[0]+(-0.432866681632)*x_ref[1]**o+(0.165933079742)*x_ref[1]+(-0.0745331372502)*x_ref[2]**o+(-0.0168977805625)*x_ref[2]
        ref[1,2]=(-0.228391157708)*x_ref[0]**o+(-0.294221312175)*x_ref[0]+(-0.121429334894)*x_ref[1]**o+(-0.557107244165)*x_ref[1]+(-0.97514025916)*x_ref[2]**o+(0.041970524119)*x_ref[2]
        ref[1,3]=(0.344978415951)*x_ref[0]**o+(0.319420517862)*x_ref[0]+(0.802947508962)*x_ref[1]**o+(0.169628742731)*x_ref[1]+(-0.380928178408)*x_ref[2]**o+(0.272551034104)*x_ref[2]
        ref[1,4]=(0.613832101706)*x_ref[0]**o+(-0.332306428069)*x_ref[0]+(0.02563364621)*x_ref[1]**o+(-0.415952850048)*x_ref[1]+(0.673350831947)*x_ref[2]**o+(0.713348183687)*x_ref[2]
        ref[2,0]=(0.290558527133)*x_ref[0]**o+(0.135335532216)*x_ref[0]+(-0.352459421266)*x_ref[1]**o+(-0.854183560051)*x_ref[1]+(0.581843223923)*x_ref[2]**o+(0.97378057053)*x_ref[2]
        ref[2,1]=(-0.328513503613)*x_ref[0]**o+(-0.205196873021)*x_ref[0]+(-0.910190961873)*x_ref[1]**o+(0.568820988101)*x_ref[1]+(0.495810061604)*x_ref[2]**o+(-0.761749884838)*x_ref[2]
        ref[2,2]=(-0.00783916993558)*x_ref[0]**o+(0.898831843878)*x_ref[0]+(0.26265892346)*x_ref[1]**o+(-0.654638473159)*x_ref[1]+(-0.713198105003)*x_ref[2]**o+(-0.59350752607)*x_ref[2]
        ref[2,3]=(-0.881651184273)*x_ref[0]**o+(-0.834213060843)*x_ref[0]+(-0.885142407575)*x_ref[1]**o+(0.206850906485)*x_ref[1]+(0.372454818886)*x_ref[2]**o+(0.159769678242)*x_ref[2]
        ref[2,4]=(-0.937735543761)*x_ref[0]**o+(0.529592561341)*x_ref[0]+(-0.936499401425)*x_ref[1]**o+(0.638973583422)*x_ref[1]+(0.333562215313)*x_ref[2]**o+(0.770257986138)*x_ref[2]
        ref[3,0]=(0.866403926159)*x_ref[0]**o+(-0.567381195064)*x_ref[0]+(-0.689200549972)*x_ref[1]**o+(0.0278246156735)*x_ref[1]+(-0.267522213281)*x_ref[2]**o+(0.536314895418)*x_ref[2]
        ref[3,1]=(-0.561464080998)*x_ref[0]**o+(-0.544070025131)*x_ref[0]+(0.73146085432)*x_ref[1]**o+(0.181244005603)*x_ref[1]+(0.63008027542)*x_ref[2]**o+(0.442771325928)*x_ref[2]
        ref[3,2]=(0.12906843064)*x_ref[0]**o+(-0.736188526885)*x_ref[0]+(0.675691533379)*x_ref[1]**o+(0.975452480209)*x_ref[1]+(0.877415310424)*x_ref[2]**o+(-0.281346067495)*x_ref[2]
        ref[3,3]=(0.895295566131)*x_ref[0]**o+(0.384502745469)*x_ref[0]+(-0.6558585454)*x_ref[1]**o+(-0.990764124757)*x_ref[1]+(0.089950807223)*x_ref[2]**o+(-0.117514839208)*x_ref[2]
        ref[3,4]=(0.908567029179)*x_ref[0]**o+(0.130934285429)*x_ref[0]+(-0.0913907665269)*x_ref[1]**o+(-0.483606581717)*x_ref[1]+(0.316018801782)*x_ref[2]**o+(-0.371754102357)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.922012175796)*x[0]**o+(-0.704323474251)*x[0]+(-0.772095928607)*x[1]**o+(-0.102522378969)*x[1]
        s[0,0,1]=(-0.756476066423)*x[0]**o+(0.476014645849)*x[0]+(0.746838535595)*x[1]**o+(-0.580996711866)*x[1]
        s[0,1,0]=(-0.372958624489)*x[0]**o+(-0.742768798844)*x[0]+(0.930234188365)*x[1]**o+(-0.934316477902)*x[1]
        s[0,1,1]=(-0.887657339887)*x[0]**o+(0.477506024138)*x[0]+(-0.072611336226)*x[1]**o+(-0.632380868263)*x[1]
        s[1,0,0]=(0.783838363367)*x[0]**o+(-0.231553922468)*x[0]+(0.747645980538)*x[1]**o+(0.480696124395)*x[1]
        s[1,0,1]=(0.916049259239)*x[0]**o+(-0.334020782789)*x[0]+(0.250380400053)*x[1]**o+(0.192676718299)*x[1]
        s[1,1,0]=(0.242999551485)*x[0]**o+(0.651169244253)*x[0]+(0.979851149428)*x[1]**o+(-0.551088471757)*x[1]
        s[1,1,1]=(0.998442854355)*x[0]**o+(-0.97084316076)*x[0]+(0.887766823857)*x[1]**o+(-0.115041414343)*x[1]
        s[2,0,0]=(0.0439895718917)*x[0]**o+(0.214997856989)*x[0]+(0.920934212621)*x[1]**o+(-0.202158287769)*x[1]
        s[2,0,1]=(-0.134517914198)*x[0]**o+(0.153074355415)*x[0]+(-0.459946249543)*x[1]**o+(0.635578282444)*x[1]
        s[2,1,0]=(-0.0910166438835)*x[0]**o+(0.802617171383)*x[0]+(0.247569766531)*x[1]**o+(0.951985690421)*x[1]
        s[2,1,1]=(0.264891721801)*x[0]**o+(-0.199605387174)*x[0]+(-0.833082379901)*x[1]**o+(0.721123597626)*x[1]
        s[3,0,0]=(-0.703356543249)*x[0]**o+(-0.146515334407)*x[0]+(-0.713251719666)*x[1]**o+(0.0740766553512)*x[1]
        s[3,0,1]=(0.623941214153)*x[0]**o+(-0.884052095344)*x[0]+(-0.976500944913)*x[1]**o+(0.985013528937)*x[1]
        s[3,1,0]=(0.794339208094)*x[0]**o+(-0.542029674325)*x[0]+(-0.427049948772)*x[1]**o+(0.37932594769)*x[1]
        s[3,1,1]=(0.563582163907)*x[0]**o+(-0.458198108954)*x[0]+(0.303347268915)*x[1]**o+(0.342283816391)*x[1]
        s[4,0,0]=(-0.717532671202)*x[0]**o+(-0.878860514987)*x[0]+(-0.0976608307102)*x[1]**o+(0.250567801094)*x[1]
        s[4,0,1]=(0.397375802895)*x[0]**o+(-0.212057004666)*x[0]+(0.558358501772)*x[1]**o+(-0.134232085341)*x[1]
        s[4,1,0]=(-0.460059564884)*x[0]**o+(-0.628865924343)*x[0]+(0.209325635948)*x[1]**o+(0.122601801028)*x[1]
        s[4,1,1]=(-0.528413992704)*x[0]**o+(-0.275278112162)*x[0]+(-0.904021647214)*x[1]**o+(0.171477127541)*x[1]
        s[5,0,0]=(-0.590272357875)*x[0]**o+(0.787036563397)*x[0]+(-0.821383663122)*x[1]**o+(-0.43696048132)*x[1]
        s[5,0,1]=(0.63774872195)*x[0]**o+(0.57629113566)*x[0]+(-0.860489105031)*x[1]**o+(0.548377309092)*x[1]
        s[5,1,0]=(0.584516519044)*x[0]**o+(0.292577411211)*x[0]+(-0.277291138474)*x[1]**o+(0.36529689442)*x[1]
        s[5,1,1]=(-0.58877604496)*x[0]**o+(-0.340786512436)*x[0]+(0.5949521732)*x[1]**o+(0.302898904668)*x[1]
        ref[0,0,0]=(0.922012175796)*x_ref[0]**o+(-0.704323474251)*x_ref[0]+(-0.772095928607)*x_ref[1]**o+(-0.102522378969)*x_ref[1]
        ref[0,0,1]=(-0.756476066423)*x_ref[0]**o+(0.476014645849)*x_ref[0]+(0.746838535595)*x_ref[1]**o+(-0.580996711866)*x_ref[1]
        ref[0,1,0]=(-0.372958624489)*x_ref[0]**o+(-0.742768798844)*x_ref[0]+(0.930234188365)*x_ref[1]**o+(-0.934316477902)*x_ref[1]
        ref[0,1,1]=(-0.887657339887)*x_ref[0]**o+(0.477506024138)*x_ref[0]+(-0.072611336226)*x_ref[1]**o+(-0.632380868263)*x_ref[1]
        ref[1,0,0]=(0.783838363367)*x_ref[0]**o+(-0.231553922468)*x_ref[0]+(0.747645980538)*x_ref[1]**o+(0.480696124395)*x_ref[1]
        ref[1,0,1]=(0.916049259239)*x_ref[0]**o+(-0.334020782789)*x_ref[0]+(0.250380400053)*x_ref[1]**o+(0.192676718299)*x_ref[1]
        ref[1,1,0]=(0.242999551485)*x_ref[0]**o+(0.651169244253)*x_ref[0]+(0.979851149428)*x_ref[1]**o+(-0.551088471757)*x_ref[1]
        ref[1,1,1]=(0.998442854355)*x_ref[0]**o+(-0.97084316076)*x_ref[0]+(0.887766823857)*x_ref[1]**o+(-0.115041414343)*x_ref[1]
        ref[2,0,0]=(0.0439895718917)*x_ref[0]**o+(0.214997856989)*x_ref[0]+(0.920934212621)*x_ref[1]**o+(-0.202158287769)*x_ref[1]
        ref[2,0,1]=(-0.134517914198)*x_ref[0]**o+(0.153074355415)*x_ref[0]+(-0.459946249543)*x_ref[1]**o+(0.635578282444)*x_ref[1]
        ref[2,1,0]=(-0.0910166438835)*x_ref[0]**o+(0.802617171383)*x_ref[0]+(0.247569766531)*x_ref[1]**o+(0.951985690421)*x_ref[1]
        ref[2,1,1]=(0.264891721801)*x_ref[0]**o+(-0.199605387174)*x_ref[0]+(-0.833082379901)*x_ref[1]**o+(0.721123597626)*x_ref[1]
        ref[3,0,0]=(-0.703356543249)*x_ref[0]**o+(-0.146515334407)*x_ref[0]+(-0.713251719666)*x_ref[1]**o+(0.0740766553512)*x_ref[1]
        ref[3,0,1]=(0.623941214153)*x_ref[0]**o+(-0.884052095344)*x_ref[0]+(-0.976500944913)*x_ref[1]**o+(0.985013528937)*x_ref[1]
        ref[3,1,0]=(0.794339208094)*x_ref[0]**o+(-0.542029674325)*x_ref[0]+(-0.427049948772)*x_ref[1]**o+(0.37932594769)*x_ref[1]
        ref[3,1,1]=(0.563582163907)*x_ref[0]**o+(-0.458198108954)*x_ref[0]+(0.303347268915)*x_ref[1]**o+(0.342283816391)*x_ref[1]
        ref[4,0,0]=(-0.717532671202)*x_ref[0]**o+(-0.878860514987)*x_ref[0]+(-0.0976608307102)*x_ref[1]**o+(0.250567801094)*x_ref[1]
        ref[4,0,1]=(0.397375802895)*x_ref[0]**o+(-0.212057004666)*x_ref[0]+(0.558358501772)*x_ref[1]**o+(-0.134232085341)*x_ref[1]
        ref[4,1,0]=(-0.460059564884)*x_ref[0]**o+(-0.628865924343)*x_ref[0]+(0.209325635948)*x_ref[1]**o+(0.122601801028)*x_ref[1]
        ref[4,1,1]=(-0.528413992704)*x_ref[0]**o+(-0.275278112162)*x_ref[0]+(-0.904021647214)*x_ref[1]**o+(0.171477127541)*x_ref[1]
        ref[5,0,0]=(-0.590272357875)*x_ref[0]**o+(0.787036563397)*x_ref[0]+(-0.821383663122)*x_ref[1]**o+(-0.43696048132)*x_ref[1]
        ref[5,0,1]=(0.63774872195)*x_ref[0]**o+(0.57629113566)*x_ref[0]+(-0.860489105031)*x_ref[1]**o+(0.548377309092)*x_ref[1]
        ref[5,1,0]=(0.584516519044)*x_ref[0]**o+(0.292577411211)*x_ref[0]+(-0.277291138474)*x_ref[1]**o+(0.36529689442)*x_ref[1]
        ref[5,1,1]=(-0.58877604496)*x_ref[0]**o+(-0.340786512436)*x_ref[0]+(0.5949521732)*x_ref[1]**o+(0.302898904668)*x_ref[1]
      else:
        s[0,0,0]=(0.103127120457)*x[0]**o+(-0.790648806335)*x[0]+(0.32847672023)*x[1]**o+(-0.690237185512)*x[1]+(-0.317184442145)*x[2]**o+(0.537453763779)*x[2]
        s[0,0,1]=(-0.36596240665)*x[0]**o+(0.256732865132)*x[0]+(-0.139947303843)*x[1]**o+(0.230496874435)*x[1]+(-0.616356953667)*x[2]**o+(-0.59441989496)*x[2]
        s[0,1,0]=(0.513873980476)*x[0]**o+(0.484045483315)*x[0]+(-0.273179211414)*x[1]**o+(-0.604138913893)*x[1]+(-0.818085004879)*x[2]**o+(-0.829509682818)*x[2]
        s[0,1,1]=(-0.797011406939)*x[0]**o+(-0.515647970266)*x[0]+(0.466120391094)*x[1]**o+(-0.996260216826)*x[1]+(-0.42533407671)*x[2]**o+(-0.081427515701)*x[2]
        s[1,0,0]=(0.840036174626)*x[0]**o+(-0.974272190945)*x[0]+(-0.912824754225)*x[1]**o+(-0.0576196265976)*x[1]+(0.311357427135)*x[2]**o+(0.0483209368492)*x[2]
        s[1,0,1]=(0.76482565945)*x[0]**o+(0.0613089443427)*x[0]+(0.907470294652)*x[1]**o+(0.720216160172)*x[1]+(-0.848264518892)*x[2]**o+(-0.622655911866)*x[2]
        s[1,1,0]=(-0.291088924515)*x[0]**o+(-0.273460845284)*x[0]+(-0.860316787156)*x[1]**o+(-0.117032009627)*x[1]+(0.0487955308094)*x[2]**o+(-0.701553638107)*x[2]
        s[1,1,1]=(-0.471933516494)*x[0]**o+(0.635615053447)*x[0]+(0.471737879913)*x[1]**o+(0.313165588781)*x[1]+(-0.15916085554)*x[2]**o+(-0.825577908132)*x[2]
        s[2,0,0]=(-0.0524254116203)*x[0]**o+(0.88748910061)*x[0]+(-0.522351516282)*x[1]**o+(-0.284852401017)*x[1]+(0.0158220147327)*x[2]**o+(0.00646469511963)*x[2]
        s[2,0,1]=(0.775455989914)*x[0]**o+(-0.592268899844)*x[0]+(-0.395459506696)*x[1]**o+(-0.778966655151)*x[1]+(-0.608812345651)*x[2]**o+(0.431749899976)*x[2]
        s[2,1,0]=(-0.472052989428)*x[0]**o+(-0.653779618462)*x[0]+(0.726973003838)*x[1]**o+(0.615103945489)*x[1]+(-0.635243574086)*x[2]**o+(-0.944283499648)*x[2]
        s[2,1,1]=(-0.538096393025)*x[0]**o+(-0.736062177106)*x[0]+(0.54984272471)*x[1]**o+(-0.252874828323)*x[1]+(0.520232358093)*x[2]**o+(0.658820792593)*x[2]
        s[3,0,0]=(0.0189624424339)*x[0]**o+(-0.928190186927)*x[0]+(-0.102190963399)*x[1]**o+(0.0985219931729)*x[1]+(0.964850306827)*x[2]**o+(-0.384616188398)*x[2]
        s[3,0,1]=(-0.122886102392)*x[0]**o+(0.916297286756)*x[0]+(0.0549430981199)*x[1]**o+(0.911993790356)*x[1]+(0.214659705881)*x[2]**o+(-0.69412668699)*x[2]
        s[3,1,0]=(0.902798669922)*x[0]**o+(-0.518265765234)*x[0]+(0.719822200918)*x[1]**o+(-0.962846306331)*x[1]+(-0.603216741191)*x[2]**o+(-0.225102184332)*x[2]
        s[3,1,1]=(0.947054397721)*x[0]**o+(0.799669807545)*x[0]+(-0.210074700167)*x[1]**o+(-0.776975337853)*x[1]+(0.750882895589)*x[2]**o+(-0.292157154938)*x[2]
        s[4,0,0]=(-0.00199580734612)*x[0]**o+(-0.894095448382)*x[0]+(-0.17192951285)*x[1]**o+(0.579608423454)*x[1]+(-0.947392406137)*x[2]**o+(-0.133887792274)*x[2]
        s[4,0,1]=(-0.395728778674)*x[0]**o+(-0.59395401109)*x[0]+(-0.334793568635)*x[1]**o+(0.848150466553)*x[1]+(0.449812215037)*x[2]**o+(-0.577515589907)*x[2]
        s[4,1,0]=(-0.815755197958)*x[0]**o+(-0.15839764311)*x[0]+(-0.829730261843)*x[1]**o+(0.891820791252)*x[1]+(0.903790315572)*x[2]**o+(-0.295122999627)*x[2]
        s[4,1,1]=(0.763626481603)*x[0]**o+(0.801670589231)*x[0]+(-0.779956910059)*x[1]**o+(0.593907194188)*x[1]+(-0.542116631989)*x[2]**o+(-0.845496315821)*x[2]
        s[5,0,0]=(-0.088706022005)*x[0]**o+(0.903197426259)*x[0]+(0.562811562755)*x[1]**o+(-0.999408930701)*x[1]+(0.647113624796)*x[2]**o+(0.917310785712)*x[2]
        s[5,0,1]=(-0.202743164028)*x[0]**o+(0.71872847025)*x[0]+(-0.709895205028)*x[1]**o+(0.294942879757)*x[1]+(-0.83533110878)*x[2]**o+(0.683966938727)*x[2]
        s[5,1,0]=(-0.0399613077171)*x[0]**o+(-0.504881528731)*x[0]+(-0.554451064799)*x[1]**o+(0.104714660691)*x[1]+(0.443770831652)*x[2]**o+(-0.673196707698)*x[2]
        s[5,1,1]=(0.993420509994)*x[0]**o+(0.717025596885)*x[0]+(-0.939900190744)*x[1]**o+(0.957571831696)*x[1]+(-0.712509953461)*x[2]**o+(0.150276310366)*x[2]
        ref[0,0,0]=(0.103127120457)*x_ref[0]**o+(-0.790648806335)*x_ref[0]+(0.32847672023)*x_ref[1]**o+(-0.690237185512)*x_ref[1]+(-0.317184442145)*x_ref[2]**o+(0.537453763779)*x_ref[2]
        ref[0,0,1]=(-0.36596240665)*x_ref[0]**o+(0.256732865132)*x_ref[0]+(-0.139947303843)*x_ref[1]**o+(0.230496874435)*x_ref[1]+(-0.616356953667)*x_ref[2]**o+(-0.59441989496)*x_ref[2]
        ref[0,1,0]=(0.513873980476)*x_ref[0]**o+(0.484045483315)*x_ref[0]+(-0.273179211414)*x_ref[1]**o+(-0.604138913893)*x_ref[1]+(-0.818085004879)*x_ref[2]**o+(-0.829509682818)*x_ref[2]
        ref[0,1,1]=(-0.797011406939)*x_ref[0]**o+(-0.515647970266)*x_ref[0]+(0.466120391094)*x_ref[1]**o+(-0.996260216826)*x_ref[1]+(-0.42533407671)*x_ref[2]**o+(-0.081427515701)*x_ref[2]
        ref[1,0,0]=(0.840036174626)*x_ref[0]**o+(-0.974272190945)*x_ref[0]+(-0.912824754225)*x_ref[1]**o+(-0.0576196265976)*x_ref[1]+(0.311357427135)*x_ref[2]**o+(0.0483209368492)*x_ref[2]
        ref[1,0,1]=(0.76482565945)*x_ref[0]**o+(0.0613089443427)*x_ref[0]+(0.907470294652)*x_ref[1]**o+(0.720216160172)*x_ref[1]+(-0.848264518892)*x_ref[2]**o+(-0.622655911866)*x_ref[2]
        ref[1,1,0]=(-0.291088924515)*x_ref[0]**o+(-0.273460845284)*x_ref[0]+(-0.860316787156)*x_ref[1]**o+(-0.117032009627)*x_ref[1]+(0.0487955308094)*x_ref[2]**o+(-0.701553638107)*x_ref[2]
        ref[1,1,1]=(-0.471933516494)*x_ref[0]**o+(0.635615053447)*x_ref[0]+(0.471737879913)*x_ref[1]**o+(0.313165588781)*x_ref[1]+(-0.15916085554)*x_ref[2]**o+(-0.825577908132)*x_ref[2]
        ref[2,0,0]=(-0.0524254116203)*x_ref[0]**o+(0.88748910061)*x_ref[0]+(-0.522351516282)*x_ref[1]**o+(-0.284852401017)*x_ref[1]+(0.0158220147327)*x_ref[2]**o+(0.00646469511963)*x_ref[2]
        ref[2,0,1]=(0.775455989914)*x_ref[0]**o+(-0.592268899844)*x_ref[0]+(-0.395459506696)*x_ref[1]**o+(-0.778966655151)*x_ref[1]+(-0.608812345651)*x_ref[2]**o+(0.431749899976)*x_ref[2]
        ref[2,1,0]=(-0.472052989428)*x_ref[0]**o+(-0.653779618462)*x_ref[0]+(0.726973003838)*x_ref[1]**o+(0.615103945489)*x_ref[1]+(-0.635243574086)*x_ref[2]**o+(-0.944283499648)*x_ref[2]
        ref[2,1,1]=(-0.538096393025)*x_ref[0]**o+(-0.736062177106)*x_ref[0]+(0.54984272471)*x_ref[1]**o+(-0.252874828323)*x_ref[1]+(0.520232358093)*x_ref[2]**o+(0.658820792593)*x_ref[2]
        ref[3,0,0]=(0.0189624424339)*x_ref[0]**o+(-0.928190186927)*x_ref[0]+(-0.102190963399)*x_ref[1]**o+(0.0985219931729)*x_ref[1]+(0.964850306827)*x_ref[2]**o+(-0.384616188398)*x_ref[2]
        ref[3,0,1]=(-0.122886102392)*x_ref[0]**o+(0.916297286756)*x_ref[0]+(0.0549430981199)*x_ref[1]**o+(0.911993790356)*x_ref[1]+(0.214659705881)*x_ref[2]**o+(-0.69412668699)*x_ref[2]
        ref[3,1,0]=(0.902798669922)*x_ref[0]**o+(-0.518265765234)*x_ref[0]+(0.719822200918)*x_ref[1]**o+(-0.962846306331)*x_ref[1]+(-0.603216741191)*x_ref[2]**o+(-0.225102184332)*x_ref[2]
        ref[3,1,1]=(0.947054397721)*x_ref[0]**o+(0.799669807545)*x_ref[0]+(-0.210074700167)*x_ref[1]**o+(-0.776975337853)*x_ref[1]+(0.750882895589)*x_ref[2]**o+(-0.292157154938)*x_ref[2]
        ref[4,0,0]=(-0.00199580734612)*x_ref[0]**o+(-0.894095448382)*x_ref[0]+(-0.17192951285)*x_ref[1]**o+(0.579608423454)*x_ref[1]+(-0.947392406137)*x_ref[2]**o+(-0.133887792274)*x_ref[2]
        ref[4,0,1]=(-0.395728778674)*x_ref[0]**o+(-0.59395401109)*x_ref[0]+(-0.334793568635)*x_ref[1]**o+(0.848150466553)*x_ref[1]+(0.449812215037)*x_ref[2]**o+(-0.577515589907)*x_ref[2]
        ref[4,1,0]=(-0.815755197958)*x_ref[0]**o+(-0.15839764311)*x_ref[0]+(-0.829730261843)*x_ref[1]**o+(0.891820791252)*x_ref[1]+(0.903790315572)*x_ref[2]**o+(-0.295122999627)*x_ref[2]
        ref[4,1,1]=(0.763626481603)*x_ref[0]**o+(0.801670589231)*x_ref[0]+(-0.779956910059)*x_ref[1]**o+(0.593907194188)*x_ref[1]+(-0.542116631989)*x_ref[2]**o+(-0.845496315821)*x_ref[2]
        ref[5,0,0]=(-0.088706022005)*x_ref[0]**o+(0.903197426259)*x_ref[0]+(0.562811562755)*x_ref[1]**o+(-0.999408930701)*x_ref[1]+(0.647113624796)*x_ref[2]**o+(0.917310785712)*x_ref[2]
        ref[5,0,1]=(-0.202743164028)*x_ref[0]**o+(0.71872847025)*x_ref[0]+(-0.709895205028)*x_ref[1]**o+(0.294942879757)*x_ref[1]+(-0.83533110878)*x_ref[2]**o+(0.683966938727)*x_ref[2]
        ref[5,1,0]=(-0.0399613077171)*x_ref[0]**o+(-0.504881528731)*x_ref[0]+(-0.554451064799)*x_ref[1]**o+(0.104714660691)*x_ref[1]+(0.443770831652)*x_ref[2]**o+(-0.673196707698)*x_ref[2]
        ref[5,1,1]=(0.993420509994)*x_ref[0]**o+(0.717025596885)*x_ref[0]+(-0.939900190744)*x_ref[1]**o+(0.957571831696)*x_ref[1]+(-0.712509953461)*x_ref[2]**o+(0.150276310366)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunction_fromSymbol_ReducedFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunction

      assumptions: self.domain supports inpterpolation from ReducedFunction onto ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.889687592166)*x[0]**o+(-0.244091857538)*x[0]+(-0.218564463727)*x[1]**o+(-0.937124237311)*x[1]
        s[0,0,0,1]=(0.440232360433)*x[0]**o+(0.441448419004)*x[0]+(-0.469778033804)*x[1]**o+(0.561615436148)*x[1]
        s[0,0,1,0]=(-0.704681295174)*x[0]**o+(-0.698896667728)*x[0]+(-0.999233531685)*x[1]**o+(0.287081983675)*x[1]
        s[0,0,1,1]=(0.899708232838)*x[0]**o+(0.521718791408)*x[0]+(-0.225938242874)*x[1]**o+(0.150056519691)*x[1]
        s[0,0,2,0]=(-0.00550720503612)*x[0]**o+(-0.521199335637)*x[0]+(0.943701370941)*x[1]**o+(-0.285282621044)*x[1]
        s[0,0,2,1]=(0.758068587203)*x[0]**o+(-0.247932113502)*x[0]+(-0.441683356848)*x[1]**o+(0.598148360435)*x[1]
        s[0,1,0,0]=(0.252073761781)*x[0]**o+(0.803649353117)*x[0]+(0.939591428513)*x[1]**o+(0.254473527243)*x[1]
        s[0,1,0,1]=(0.652883529471)*x[0]**o+(0.832856006165)*x[0]+(-0.754806656391)*x[1]**o+(0.477988652237)*x[1]
        s[0,1,1,0]=(0.256691612427)*x[0]**o+(-0.827547687439)*x[0]+(-0.426446760113)*x[1]**o+(-0.709508780384)*x[1]
        s[0,1,1,1]=(-0.66459132575)*x[0]**o+(0.82036954645)*x[0]+(-0.154864164844)*x[1]**o+(0.967900851407)*x[1]
        s[0,1,2,0]=(0.82202922242)*x[0]**o+(-0.564876598949)*x[0]+(0.842505409303)*x[1]**o+(-0.710636181332)*x[1]
        s[0,1,2,1]=(0.83881620922)*x[0]**o+(-0.237072167776)*x[0]+(-0.442181682029)*x[1]**o+(0.771550454195)*x[1]
        s[0,2,0,0]=(0.474850766572)*x[0]**o+(-0.857346378644)*x[0]+(0.919175013993)*x[1]**o+(0.629995461581)*x[1]
        s[0,2,0,1]=(0.523096282888)*x[0]**o+(-0.510665736272)*x[0]+(0.145171493458)*x[1]**o+(-0.103513634427)*x[1]
        s[0,2,1,0]=(-0.423486426499)*x[0]**o+(-0.49703927479)*x[0]+(0.174127212691)*x[1]**o+(0.217944977415)*x[1]
        s[0,2,1,1]=(-0.631505727799)*x[0]**o+(0.588807237677)*x[0]+(0.668437404223)*x[1]**o+(0.552798016144)*x[1]
        s[0,2,2,0]=(0.734466686613)*x[0]**o+(0.142930362034)*x[0]+(0.490607179637)*x[1]**o+(0.0445798411208)*x[1]
        s[0,2,2,1]=(0.59894699236)*x[0]**o+(-0.80614241159)*x[0]+(-0.966344783722)*x[1]**o+(-0.537851881337)*x[1]
        s[0,3,0,0]=(0.635428545228)*x[0]**o+(0.458465797988)*x[0]+(0.431015659184)*x[1]**o+(-0.59835525481)*x[1]
        s[0,3,0,1]=(0.0494865211965)*x[0]**o+(-0.776500412974)*x[0]+(0.932578121115)*x[1]**o+(0.914133479331)*x[1]
        s[0,3,1,0]=(0.231396476054)*x[0]**o+(-0.265428688952)*x[0]+(-0.344497302174)*x[1]**o+(-0.757106502174)*x[1]
        s[0,3,1,1]=(0.551883714399)*x[0]**o+(0.832451314453)*x[0]+(-0.899905963729)*x[1]**o+(-0.79240620147)*x[1]
        s[0,3,2,0]=(0.564173236354)*x[0]**o+(-0.218691242318)*x[0]+(0.974498273818)*x[1]**o+(-0.712174073532)*x[1]
        s[0,3,2,1]=(-0.756975354573)*x[0]**o+(0.99073177566)*x[0]+(-0.680264934169)*x[1]**o+(-0.551951700356)*x[1]
        s[0,4,0,0]=(0.342204785324)*x[0]**o+(-0.332826079413)*x[0]+(0.827106843714)*x[1]**o+(0.147296550674)*x[1]
        s[0,4,0,1]=(0.718074130442)*x[0]**o+(0.360060294518)*x[0]+(-0.751135064333)*x[1]**o+(0.217154371958)*x[1]
        s[0,4,1,0]=(-0.930762040509)*x[0]**o+(0.368919544302)*x[0]+(0.883523093869)*x[1]**o+(-0.904151741067)*x[1]
        s[0,4,1,1]=(0.518896572101)*x[0]**o+(-0.342606634816)*x[0]+(-0.129850163393)*x[1]**o+(0.633244155895)*x[1]
        s[0,4,2,0]=(0.392116617423)*x[0]**o+(-0.173211325829)*x[0]+(0.400087049088)*x[1]**o+(0.7891166848)*x[1]
        s[0,4,2,1]=(0.716606140961)*x[0]**o+(0.496596392306)*x[0]+(-0.671375280598)*x[1]**o+(-0.528611363434)*x[1]
        s[1,0,0,0]=(0.618225203091)*x[0]**o+(0.384959538542)*x[0]+(-0.0428815543398)*x[1]**o+(0.161091350961)*x[1]
        s[1,0,0,1]=(0.443417981463)*x[0]**o+(-0.162056175957)*x[0]+(0.207624970868)*x[1]**o+(-0.121433587807)*x[1]
        s[1,0,1,0]=(0.776131229022)*x[0]**o+(-0.871658190605)*x[0]+(0.54507490858)*x[1]**o+(0.159672390575)*x[1]
        s[1,0,1,1]=(-0.361120028772)*x[0]**o+(-0.406707123035)*x[0]+(-0.54341861971)*x[1]**o+(-0.0392836956005)*x[1]
        s[1,0,2,0]=(0.944116160979)*x[0]**o+(0.0890185721642)*x[0]+(0.73533144751)*x[1]**o+(0.701431910124)*x[1]
        s[1,0,2,1]=(0.722492883973)*x[0]**o+(0.351843053475)*x[0]+(-0.1756241753)*x[1]**o+(-0.703690677653)*x[1]
        s[1,1,0,0]=(0.262345197739)*x[0]**o+(0.683913892378)*x[0]+(0.322362020764)*x[1]**o+(0.851902848166)*x[1]
        s[1,1,0,1]=(-0.0509963903106)*x[0]**o+(0.387065915322)*x[0]+(0.0727012685343)*x[1]**o+(0.666978314077)*x[1]
        s[1,1,1,0]=(-0.518122762009)*x[0]**o+(0.69514335574)*x[0]+(0.34513236897)*x[1]**o+(0.641034103704)*x[1]
        s[1,1,1,1]=(0.0658005558873)*x[0]**o+(-0.305147050226)*x[0]+(-0.707909968597)*x[1]**o+(0.0682220439272)*x[1]
        s[1,1,2,0]=(-0.443540005123)*x[0]**o+(0.179307931754)*x[0]+(0.982374202912)*x[1]**o+(0.398796141179)*x[1]
        s[1,1,2,1]=(-0.52173751599)*x[0]**o+(-0.523741826056)*x[0]+(0.172601588482)*x[1]**o+(0.969133917994)*x[1]
        s[1,2,0,0]=(-0.807607635564)*x[0]**o+(-0.525440627018)*x[0]+(-0.868823879608)*x[1]**o+(0.913203387198)*x[1]
        s[1,2,0,1]=(0.84988195907)*x[0]**o+(-0.394750851572)*x[0]+(0.740132509878)*x[1]**o+(0.855925902936)*x[1]
        s[1,2,1,0]=(0.055360229349)*x[0]**o+(0.299990998824)*x[0]+(0.558216562537)*x[1]**o+(-0.382651884902)*x[1]
        s[1,2,1,1]=(0.423603813726)*x[0]**o+(-0.907945615918)*x[0]+(-0.117306295879)*x[1]**o+(-0.0502458178048)*x[1]
        s[1,2,2,0]=(0.06196107837)*x[0]**o+(-0.612619509899)*x[0]+(0.169179798359)*x[1]**o+(-0.27854976134)*x[1]
        s[1,2,2,1]=(-0.506388098563)*x[0]**o+(0.125008314945)*x[0]+(0.509571389625)*x[1]**o+(0.442027080754)*x[1]
        s[1,3,0,0]=(0.304208463979)*x[0]**o+(-0.379194737667)*x[0]+(0.978868674946)*x[1]**o+(0.011573322151)*x[1]
        s[1,3,0,1]=(-0.151986748739)*x[0]**o+(0.153593063241)*x[0]+(0.644694257244)*x[1]**o+(0.414073745353)*x[1]
        s[1,3,1,0]=(0.924109632511)*x[0]**o+(0.398532810155)*x[0]+(-0.321086484535)*x[1]**o+(0.0802378501316)*x[1]
        s[1,3,1,1]=(-0.175639876422)*x[0]**o+(-0.109223049695)*x[0]+(-0.206450491082)*x[1]**o+(0.462124555092)*x[1]
        s[1,3,2,0]=(0.379156310804)*x[0]**o+(0.629232237742)*x[0]+(0.0918877201433)*x[1]**o+(-0.0490282902402)*x[1]
        s[1,3,2,1]=(-0.616126549785)*x[0]**o+(0.528823818571)*x[0]+(-0.947449800384)*x[1]**o+(0.145045253788)*x[1]
        s[1,4,0,0]=(-0.623165695489)*x[0]**o+(-0.861583499935)*x[0]+(0.219861821844)*x[1]**o+(-0.962135814763)*x[1]
        s[1,4,0,1]=(0.467875500099)*x[0]**o+(0.906948957958)*x[0]+(0.606958286586)*x[1]**o+(-0.423219122678)*x[1]
        s[1,4,1,0]=(0.200880358483)*x[0]**o+(0.571730447827)*x[0]+(0.908300268754)*x[1]**o+(0.19978633475)*x[1]
        s[1,4,1,1]=(0.55118731313)*x[0]**o+(-0.89903633572)*x[0]+(0.43836310821)*x[1]**o+(0.404697808939)*x[1]
        s[1,4,2,0]=(0.536587233299)*x[0]**o+(-0.0789531097738)*x[0]+(-0.775848629826)*x[1]**o+(-0.11020318415)*x[1]
        s[1,4,2,1]=(-0.648091804458)*x[0]**o+(-0.0282126867936)*x[0]+(0.0182411167897)*x[1]**o+(0.922414038295)*x[1]
        s[2,0,0,0]=(0.67385583226)*x[0]**o+(-0.993060381925)*x[0]+(0.731462156889)*x[1]**o+(0.183718928132)*x[1]
        s[2,0,0,1]=(0.881124628544)*x[0]**o+(0.828720578546)*x[0]+(-0.324258860918)*x[1]**o+(-0.836625964944)*x[1]
        s[2,0,1,0]=(0.944307497546)*x[0]**o+(0.546752005636)*x[0]+(0.334176020601)*x[1]**o+(0.654940705861)*x[1]
        s[2,0,1,1]=(-0.414075359332)*x[0]**o+(-0.782021485027)*x[0]+(-0.233424028825)*x[1]**o+(0.652900422088)*x[1]
        s[2,0,2,0]=(0.355942393341)*x[0]**o+(-0.727957138948)*x[0]+(-0.924406373726)*x[1]**o+(-0.806178313048)*x[1]
        s[2,0,2,1]=(-0.791343585742)*x[0]**o+(-0.61722808774)*x[0]+(-0.627757910074)*x[1]**o+(0.838801750555)*x[1]
        s[2,1,0,0]=(-0.866831381931)*x[0]**o+(-0.00498093548417)*x[0]+(-0.568387329284)*x[1]**o+(-0.428222759011)*x[1]
        s[2,1,0,1]=(-0.245588724312)*x[0]**o+(0.571669044307)*x[0]+(0.832968046978)*x[1]**o+(-0.0595908787599)*x[1]
        s[2,1,1,0]=(0.454631203504)*x[0]**o+(-0.827357560144)*x[0]+(0.876484630903)*x[1]**o+(0.316668730296)*x[1]
        s[2,1,1,1]=(-0.802326431878)*x[0]**o+(-0.547113643585)*x[0]+(0.619358075866)*x[1]**o+(0.387898635274)*x[1]
        s[2,1,2,0]=(0.204730319481)*x[0]**o+(0.462536054065)*x[0]+(-0.267050942194)*x[1]**o+(-0.529151462961)*x[1]
        s[2,1,2,1]=(-0.354690492675)*x[0]**o+(0.518914850203)*x[0]+(-0.283765250756)*x[1]**o+(0.595912311106)*x[1]
        s[2,2,0,0]=(0.329350522628)*x[0]**o+(0.689800893689)*x[0]+(-0.316097521431)*x[1]**o+(0.567054800242)*x[1]
        s[2,2,0,1]=(-0.68603272612)*x[0]**o+(-0.458451752241)*x[0]+(0.442247854284)*x[1]**o+(0.342266856887)*x[1]
        s[2,2,1,0]=(0.0562781380442)*x[0]**o+(0.615293898746)*x[0]+(-0.446430775323)*x[1]**o+(-0.720310464006)*x[1]
        s[2,2,1,1]=(-0.620026567808)*x[0]**o+(0.64935057185)*x[0]+(-0.402630635809)*x[1]**o+(0.984397383657)*x[1]
        s[2,2,2,0]=(-0.565466963253)*x[0]**o+(-0.983283191302)*x[0]+(0.0517997901365)*x[1]**o+(-0.00459375732072)*x[1]
        s[2,2,2,1]=(-0.544844126966)*x[0]**o+(-0.993222368569)*x[0]+(-0.0761437907904)*x[1]**o+(0.959202974542)*x[1]
        s[2,3,0,0]=(-0.542176217992)*x[0]**o+(-0.210550241274)*x[0]+(0.447732450287)*x[1]**o+(-0.0802692719358)*x[1]
        s[2,3,0,1]=(0.339422485349)*x[0]**o+(0.581409183541)*x[0]+(-0.255731595248)*x[1]**o+(-0.196957128507)*x[1]
        s[2,3,1,0]=(0.938606287977)*x[0]**o+(0.396408957375)*x[0]+(-0.53613392333)*x[1]**o+(-0.201412967427)*x[1]
        s[2,3,1,1]=(0.482931877644)*x[0]**o+(-0.395740204538)*x[0]+(-0.816216426797)*x[1]**o+(-0.390149600366)*x[1]
        s[2,3,2,0]=(0.507838191989)*x[0]**o+(0.590218664943)*x[0]+(0.987052963874)*x[1]**o+(-0.39735585208)*x[1]
        s[2,3,2,1]=(-0.036562453009)*x[0]**o+(0.299852609043)*x[0]+(0.194160324844)*x[1]**o+(-0.629803094543)*x[1]
        s[2,4,0,0]=(0.128664458677)*x[0]**o+(0.956037220714)*x[0]+(0.770331226849)*x[1]**o+(-0.208007445309)*x[1]
        s[2,4,0,1]=(0.526755720598)*x[0]**o+(-0.334650615987)*x[0]+(-0.245693030991)*x[1]**o+(-0.712134261554)*x[1]
        s[2,4,1,0]=(0.827770360865)*x[0]**o+(0.982906973774)*x[0]+(0.471438390576)*x[1]**o+(0.691215965917)*x[1]
        s[2,4,1,1]=(-0.206715480236)*x[0]**o+(0.488812089863)*x[0]+(0.777672390635)*x[1]**o+(-0.496681850755)*x[1]
        s[2,4,2,0]=(0.183938264875)*x[0]**o+(-0.247363713904)*x[0]+(0.962928810844)*x[1]**o+(0.642706194018)*x[1]
        s[2,4,2,1]=(0.758025105584)*x[0]**o+(-0.484793869316)*x[0]+(-0.431089127161)*x[1]**o+(0.127528834965)*x[1]
        s[3,0,0,0]=(-0.130978215422)*x[0]**o+(0.688819586128)*x[0]+(0.517162828175)*x[1]**o+(-0.979632247431)*x[1]
        s[3,0,0,1]=(0.130939629053)*x[0]**o+(0.685452224086)*x[0]+(-0.438454293093)*x[1]**o+(-0.727705397967)*x[1]
        s[3,0,1,0]=(-0.31299808275)*x[0]**o+(0.266897949461)*x[0]+(0.117566429054)*x[1]**o+(0.536075754797)*x[1]
        s[3,0,1,1]=(-0.401764430381)*x[0]**o+(-0.205983383212)*x[0]+(0.461300727769)*x[1]**o+(-0.580777919325)*x[1]
        s[3,0,2,0]=(-0.63526664417)*x[0]**o+(-0.878687700719)*x[0]+(0.104141198375)*x[1]**o+(-0.539520033395)*x[1]
        s[3,0,2,1]=(-0.592054773422)*x[0]**o+(-0.10222451991)*x[0]+(-0.385428960408)*x[1]**o+(-0.497192270367)*x[1]
        s[3,1,0,0]=(0.17159987373)*x[0]**o+(-0.215728352568)*x[0]+(0.188842078024)*x[1]**o+(0.417647732421)*x[1]
        s[3,1,0,1]=(0.548015992401)*x[0]**o+(-0.669157307443)*x[0]+(-0.796467777561)*x[1]**o+(0.163833145799)*x[1]
        s[3,1,1,0]=(-0.643762450865)*x[0]**o+(0.548750992658)*x[0]+(0.409211186464)*x[1]**o+(-0.986601038131)*x[1]
        s[3,1,1,1]=(-0.882340195766)*x[0]**o+(-0.185979117217)*x[0]+(-0.00826877676749)*x[1]**o+(0.0657705611536)*x[1]
        s[3,1,2,0]=(-0.0644145920386)*x[0]**o+(0.18740710155)*x[0]+(-0.690647279595)*x[1]**o+(-0.185556727699)*x[1]
        s[3,1,2,1]=(0.114366852618)*x[0]**o+(-0.388821787469)*x[0]+(-0.195241613331)*x[1]**o+(0.871862920816)*x[1]
        s[3,2,0,0]=(-0.673828733159)*x[0]**o+(-0.172601630353)*x[0]+(-0.488329313517)*x[1]**o+(0.0809307315053)*x[1]
        s[3,2,0,1]=(0.565273426718)*x[0]**o+(-0.610423824924)*x[0]+(0.856623167832)*x[1]**o+(-0.148845473418)*x[1]
        s[3,2,1,0]=(-0.3472751889)*x[0]**o+(0.908614631864)*x[0]+(0.445654124992)*x[1]**o+(0.719429313439)*x[1]
        s[3,2,1,1]=(0.24891832441)*x[0]**o+(-0.650081727005)*x[0]+(0.0980208592314)*x[1]**o+(0.575231534075)*x[1]
        s[3,2,2,0]=(-0.737935474986)*x[0]**o+(-0.874708935739)*x[0]+(0.856037958766)*x[1]**o+(-0.420041922056)*x[1]
        s[3,2,2,1]=(0.106345127894)*x[0]**o+(0.363060447152)*x[0]+(-0.472262909157)*x[1]**o+(-0.231475435242)*x[1]
        s[3,3,0,0]=(-0.3699135022)*x[0]**o+(-0.167427922355)*x[0]+(-0.911307660539)*x[1]**o+(-0.445726901976)*x[1]
        s[3,3,0,1]=(0.505339735087)*x[0]**o+(0.410541984072)*x[0]+(0.847415917938)*x[1]**o+(0.434325090295)*x[1]
        s[3,3,1,0]=(0.2589814894)*x[0]**o+(-0.087780879445)*x[0]+(0.847904119817)*x[1]**o+(0.462585909981)*x[1]
        s[3,3,1,1]=(0.283347628256)*x[0]**o+(0.637029684847)*x[0]+(0.174554376017)*x[1]**o+(0.642427538978)*x[1]
        s[3,3,2,0]=(-0.599348189011)*x[0]**o+(0.087002991683)*x[0]+(0.823297612508)*x[1]**o+(0.84180029964)*x[1]
        s[3,3,2,1]=(0.846090423225)*x[0]**o+(-0.465062745856)*x[0]+(0.498670439686)*x[1]**o+(0.721831089205)*x[1]
        s[3,4,0,0]=(0.0344263713162)*x[0]**o+(0.789039070423)*x[0]+(-0.518898476617)*x[1]**o+(0.112666233369)*x[1]
        s[3,4,0,1]=(0.752855926629)*x[0]**o+(-0.028671093428)*x[0]+(0.457225098582)*x[1]**o+(0.311273673822)*x[1]
        s[3,4,1,0]=(-0.623675057135)*x[0]**o+(-0.19730173917)*x[0]+(-0.328378661061)*x[1]**o+(-0.196083495589)*x[1]
        s[3,4,1,1]=(0.722548420108)*x[0]**o+(0.471733735786)*x[0]+(-0.151931791011)*x[1]**o+(0.0147226290316)*x[1]
        s[3,4,2,0]=(0.512232160342)*x[0]**o+(-0.339879887275)*x[0]+(-0.261940015624)*x[1]**o+(-0.589559690156)*x[1]
        s[3,4,2,1]=(0.387751869794)*x[0]**o+(0.284749443635)*x[0]+(0.764067938503)*x[1]**o+(0.20999603354)*x[1]
        ref[0,0,0,0]=(-0.889687592166)*x_ref[0]**o+(-0.244091857538)*x_ref[0]+(-0.218564463727)*x_ref[1]**o+(-0.937124237311)*x_ref[1]
        ref[0,0,0,1]=(0.440232360433)*x_ref[0]**o+(0.441448419004)*x_ref[0]+(-0.469778033804)*x_ref[1]**o+(0.561615436148)*x_ref[1]
        ref[0,0,1,0]=(-0.704681295174)*x_ref[0]**o+(-0.698896667728)*x_ref[0]+(-0.999233531685)*x_ref[1]**o+(0.287081983675)*x_ref[1]
        ref[0,0,1,1]=(0.899708232838)*x_ref[0]**o+(0.521718791408)*x_ref[0]+(-0.225938242874)*x_ref[1]**o+(0.150056519691)*x_ref[1]
        ref[0,0,2,0]=(-0.00550720503612)*x_ref[0]**o+(-0.521199335637)*x_ref[0]+(0.943701370941)*x_ref[1]**o+(-0.285282621044)*x_ref[1]
        ref[0,0,2,1]=(0.758068587203)*x_ref[0]**o+(-0.247932113502)*x_ref[0]+(-0.441683356848)*x_ref[1]**o+(0.598148360435)*x_ref[1]
        ref[0,1,0,0]=(0.252073761781)*x_ref[0]**o+(0.803649353117)*x_ref[0]+(0.939591428513)*x_ref[1]**o+(0.254473527243)*x_ref[1]
        ref[0,1,0,1]=(0.652883529471)*x_ref[0]**o+(0.832856006165)*x_ref[0]+(-0.754806656391)*x_ref[1]**o+(0.477988652237)*x_ref[1]
        ref[0,1,1,0]=(0.256691612427)*x_ref[0]**o+(-0.827547687439)*x_ref[0]+(-0.426446760113)*x_ref[1]**o+(-0.709508780384)*x_ref[1]
        ref[0,1,1,1]=(-0.66459132575)*x_ref[0]**o+(0.82036954645)*x_ref[0]+(-0.154864164844)*x_ref[1]**o+(0.967900851407)*x_ref[1]
        ref[0,1,2,0]=(0.82202922242)*x_ref[0]**o+(-0.564876598949)*x_ref[0]+(0.842505409303)*x_ref[1]**o+(-0.710636181332)*x_ref[1]
        ref[0,1,2,1]=(0.83881620922)*x_ref[0]**o+(-0.237072167776)*x_ref[0]+(-0.442181682029)*x_ref[1]**o+(0.771550454195)*x_ref[1]
        ref[0,2,0,0]=(0.474850766572)*x_ref[0]**o+(-0.857346378644)*x_ref[0]+(0.919175013993)*x_ref[1]**o+(0.629995461581)*x_ref[1]
        ref[0,2,0,1]=(0.523096282888)*x_ref[0]**o+(-0.510665736272)*x_ref[0]+(0.145171493458)*x_ref[1]**o+(-0.103513634427)*x_ref[1]
        ref[0,2,1,0]=(-0.423486426499)*x_ref[0]**o+(-0.49703927479)*x_ref[0]+(0.174127212691)*x_ref[1]**o+(0.217944977415)*x_ref[1]
        ref[0,2,1,1]=(-0.631505727799)*x_ref[0]**o+(0.588807237677)*x_ref[0]+(0.668437404223)*x_ref[1]**o+(0.552798016144)*x_ref[1]
        ref[0,2,2,0]=(0.734466686613)*x_ref[0]**o+(0.142930362034)*x_ref[0]+(0.490607179637)*x_ref[1]**o+(0.0445798411208)*x_ref[1]
        ref[0,2,2,1]=(0.59894699236)*x_ref[0]**o+(-0.80614241159)*x_ref[0]+(-0.966344783722)*x_ref[1]**o+(-0.537851881337)*x_ref[1]
        ref[0,3,0,0]=(0.635428545228)*x_ref[0]**o+(0.458465797988)*x_ref[0]+(0.431015659184)*x_ref[1]**o+(-0.59835525481)*x_ref[1]
        ref[0,3,0,1]=(0.0494865211965)*x_ref[0]**o+(-0.776500412974)*x_ref[0]+(0.932578121115)*x_ref[1]**o+(0.914133479331)*x_ref[1]
        ref[0,3,1,0]=(0.231396476054)*x_ref[0]**o+(-0.265428688952)*x_ref[0]+(-0.344497302174)*x_ref[1]**o+(-0.757106502174)*x_ref[1]
        ref[0,3,1,1]=(0.551883714399)*x_ref[0]**o+(0.832451314453)*x_ref[0]+(-0.899905963729)*x_ref[1]**o+(-0.79240620147)*x_ref[1]
        ref[0,3,2,0]=(0.564173236354)*x_ref[0]**o+(-0.218691242318)*x_ref[0]+(0.974498273818)*x_ref[1]**o+(-0.712174073532)*x_ref[1]
        ref[0,3,2,1]=(-0.756975354573)*x_ref[0]**o+(0.99073177566)*x_ref[0]+(-0.680264934169)*x_ref[1]**o+(-0.551951700356)*x_ref[1]
        ref[0,4,0,0]=(0.342204785324)*x_ref[0]**o+(-0.332826079413)*x_ref[0]+(0.827106843714)*x_ref[1]**o+(0.147296550674)*x_ref[1]
        ref[0,4,0,1]=(0.718074130442)*x_ref[0]**o+(0.360060294518)*x_ref[0]+(-0.751135064333)*x_ref[1]**o+(0.217154371958)*x_ref[1]
        ref[0,4,1,0]=(-0.930762040509)*x_ref[0]**o+(0.368919544302)*x_ref[0]+(0.883523093869)*x_ref[1]**o+(-0.904151741067)*x_ref[1]
        ref[0,4,1,1]=(0.518896572101)*x_ref[0]**o+(-0.342606634816)*x_ref[0]+(-0.129850163393)*x_ref[1]**o+(0.633244155895)*x_ref[1]
        ref[0,4,2,0]=(0.392116617423)*x_ref[0]**o+(-0.173211325829)*x_ref[0]+(0.400087049088)*x_ref[1]**o+(0.7891166848)*x_ref[1]
        ref[0,4,2,1]=(0.716606140961)*x_ref[0]**o+(0.496596392306)*x_ref[0]+(-0.671375280598)*x_ref[1]**o+(-0.528611363434)*x_ref[1]
        ref[1,0,0,0]=(0.618225203091)*x_ref[0]**o+(0.384959538542)*x_ref[0]+(-0.0428815543398)*x_ref[1]**o+(0.161091350961)*x_ref[1]
        ref[1,0,0,1]=(0.443417981463)*x_ref[0]**o+(-0.162056175957)*x_ref[0]+(0.207624970868)*x_ref[1]**o+(-0.121433587807)*x_ref[1]
        ref[1,0,1,0]=(0.776131229022)*x_ref[0]**o+(-0.871658190605)*x_ref[0]+(0.54507490858)*x_ref[1]**o+(0.159672390575)*x_ref[1]
        ref[1,0,1,1]=(-0.361120028772)*x_ref[0]**o+(-0.406707123035)*x_ref[0]+(-0.54341861971)*x_ref[1]**o+(-0.0392836956005)*x_ref[1]
        ref[1,0,2,0]=(0.944116160979)*x_ref[0]**o+(0.0890185721642)*x_ref[0]+(0.73533144751)*x_ref[1]**o+(0.701431910124)*x_ref[1]
        ref[1,0,2,1]=(0.722492883973)*x_ref[0]**o+(0.351843053475)*x_ref[0]+(-0.1756241753)*x_ref[1]**o+(-0.703690677653)*x_ref[1]
        ref[1,1,0,0]=(0.262345197739)*x_ref[0]**o+(0.683913892378)*x_ref[0]+(0.322362020764)*x_ref[1]**o+(0.851902848166)*x_ref[1]
        ref[1,1,0,1]=(-0.0509963903106)*x_ref[0]**o+(0.387065915322)*x_ref[0]+(0.0727012685343)*x_ref[1]**o+(0.666978314077)*x_ref[1]
        ref[1,1,1,0]=(-0.518122762009)*x_ref[0]**o+(0.69514335574)*x_ref[0]+(0.34513236897)*x_ref[1]**o+(0.641034103704)*x_ref[1]
        ref[1,1,1,1]=(0.0658005558873)*x_ref[0]**o+(-0.305147050226)*x_ref[0]+(-0.707909968597)*x_ref[1]**o+(0.0682220439272)*x_ref[1]
        ref[1,1,2,0]=(-0.443540005123)*x_ref[0]**o+(0.179307931754)*x_ref[0]+(0.982374202912)*x_ref[1]**o+(0.398796141179)*x_ref[1]
        ref[1,1,2,1]=(-0.52173751599)*x_ref[0]**o+(-0.523741826056)*x_ref[0]+(0.172601588482)*x_ref[1]**o+(0.969133917994)*x_ref[1]
        ref[1,2,0,0]=(-0.807607635564)*x_ref[0]**o+(-0.525440627018)*x_ref[0]+(-0.868823879608)*x_ref[1]**o+(0.913203387198)*x_ref[1]
        ref[1,2,0,1]=(0.84988195907)*x_ref[0]**o+(-0.394750851572)*x_ref[0]+(0.740132509878)*x_ref[1]**o+(0.855925902936)*x_ref[1]
        ref[1,2,1,0]=(0.055360229349)*x_ref[0]**o+(0.299990998824)*x_ref[0]+(0.558216562537)*x_ref[1]**o+(-0.382651884902)*x_ref[1]
        ref[1,2,1,1]=(0.423603813726)*x_ref[0]**o+(-0.907945615918)*x_ref[0]+(-0.117306295879)*x_ref[1]**o+(-0.0502458178048)*x_ref[1]
        ref[1,2,2,0]=(0.06196107837)*x_ref[0]**o+(-0.612619509899)*x_ref[0]+(0.169179798359)*x_ref[1]**o+(-0.27854976134)*x_ref[1]
        ref[1,2,2,1]=(-0.506388098563)*x_ref[0]**o+(0.125008314945)*x_ref[0]+(0.509571389625)*x_ref[1]**o+(0.442027080754)*x_ref[1]
        ref[1,3,0,0]=(0.304208463979)*x_ref[0]**o+(-0.379194737667)*x_ref[0]+(0.978868674946)*x_ref[1]**o+(0.011573322151)*x_ref[1]
        ref[1,3,0,1]=(-0.151986748739)*x_ref[0]**o+(0.153593063241)*x_ref[0]+(0.644694257244)*x_ref[1]**o+(0.414073745353)*x_ref[1]
        ref[1,3,1,0]=(0.924109632511)*x_ref[0]**o+(0.398532810155)*x_ref[0]+(-0.321086484535)*x_ref[1]**o+(0.0802378501316)*x_ref[1]
        ref[1,3,1,1]=(-0.175639876422)*x_ref[0]**o+(-0.109223049695)*x_ref[0]+(-0.206450491082)*x_ref[1]**o+(0.462124555092)*x_ref[1]
        ref[1,3,2,0]=(0.379156310804)*x_ref[0]**o+(0.629232237742)*x_ref[0]+(0.0918877201433)*x_ref[1]**o+(-0.0490282902402)*x_ref[1]
        ref[1,3,2,1]=(-0.616126549785)*x_ref[0]**o+(0.528823818571)*x_ref[0]+(-0.947449800384)*x_ref[1]**o+(0.145045253788)*x_ref[1]
        ref[1,4,0,0]=(-0.623165695489)*x_ref[0]**o+(-0.861583499935)*x_ref[0]+(0.219861821844)*x_ref[1]**o+(-0.962135814763)*x_ref[1]
        ref[1,4,0,1]=(0.467875500099)*x_ref[0]**o+(0.906948957958)*x_ref[0]+(0.606958286586)*x_ref[1]**o+(-0.423219122678)*x_ref[1]
        ref[1,4,1,0]=(0.200880358483)*x_ref[0]**o+(0.571730447827)*x_ref[0]+(0.908300268754)*x_ref[1]**o+(0.19978633475)*x_ref[1]
        ref[1,4,1,1]=(0.55118731313)*x_ref[0]**o+(-0.89903633572)*x_ref[0]+(0.43836310821)*x_ref[1]**o+(0.404697808939)*x_ref[1]
        ref[1,4,2,0]=(0.536587233299)*x_ref[0]**o+(-0.0789531097738)*x_ref[0]+(-0.775848629826)*x_ref[1]**o+(-0.11020318415)*x_ref[1]
        ref[1,4,2,1]=(-0.648091804458)*x_ref[0]**o+(-0.0282126867936)*x_ref[0]+(0.0182411167897)*x_ref[1]**o+(0.922414038295)*x_ref[1]
        ref[2,0,0,0]=(0.67385583226)*x_ref[0]**o+(-0.993060381925)*x_ref[0]+(0.731462156889)*x_ref[1]**o+(0.183718928132)*x_ref[1]
        ref[2,0,0,1]=(0.881124628544)*x_ref[0]**o+(0.828720578546)*x_ref[0]+(-0.324258860918)*x_ref[1]**o+(-0.836625964944)*x_ref[1]
        ref[2,0,1,0]=(0.944307497546)*x_ref[0]**o+(0.546752005636)*x_ref[0]+(0.334176020601)*x_ref[1]**o+(0.654940705861)*x_ref[1]
        ref[2,0,1,1]=(-0.414075359332)*x_ref[0]**o+(-0.782021485027)*x_ref[0]+(-0.233424028825)*x_ref[1]**o+(0.652900422088)*x_ref[1]
        ref[2,0,2,0]=(0.355942393341)*x_ref[0]**o+(-0.727957138948)*x_ref[0]+(-0.924406373726)*x_ref[1]**o+(-0.806178313048)*x_ref[1]
        ref[2,0,2,1]=(-0.791343585742)*x_ref[0]**o+(-0.61722808774)*x_ref[0]+(-0.627757910074)*x_ref[1]**o+(0.838801750555)*x_ref[1]
        ref[2,1,0,0]=(-0.866831381931)*x_ref[0]**o+(-0.00498093548417)*x_ref[0]+(-0.568387329284)*x_ref[1]**o+(-0.428222759011)*x_ref[1]
        ref[2,1,0,1]=(-0.245588724312)*x_ref[0]**o+(0.571669044307)*x_ref[0]+(0.832968046978)*x_ref[1]**o+(-0.0595908787599)*x_ref[1]
        ref[2,1,1,0]=(0.454631203504)*x_ref[0]**o+(-0.827357560144)*x_ref[0]+(0.876484630903)*x_ref[1]**o+(0.316668730296)*x_ref[1]
        ref[2,1,1,1]=(-0.802326431878)*x_ref[0]**o+(-0.547113643585)*x_ref[0]+(0.619358075866)*x_ref[1]**o+(0.387898635274)*x_ref[1]
        ref[2,1,2,0]=(0.204730319481)*x_ref[0]**o+(0.462536054065)*x_ref[0]+(-0.267050942194)*x_ref[1]**o+(-0.529151462961)*x_ref[1]
        ref[2,1,2,1]=(-0.354690492675)*x_ref[0]**o+(0.518914850203)*x_ref[0]+(-0.283765250756)*x_ref[1]**o+(0.595912311106)*x_ref[1]
        ref[2,2,0,0]=(0.329350522628)*x_ref[0]**o+(0.689800893689)*x_ref[0]+(-0.316097521431)*x_ref[1]**o+(0.567054800242)*x_ref[1]
        ref[2,2,0,1]=(-0.68603272612)*x_ref[0]**o+(-0.458451752241)*x_ref[0]+(0.442247854284)*x_ref[1]**o+(0.342266856887)*x_ref[1]
        ref[2,2,1,0]=(0.0562781380442)*x_ref[0]**o+(0.615293898746)*x_ref[0]+(-0.446430775323)*x_ref[1]**o+(-0.720310464006)*x_ref[1]
        ref[2,2,1,1]=(-0.620026567808)*x_ref[0]**o+(0.64935057185)*x_ref[0]+(-0.402630635809)*x_ref[1]**o+(0.984397383657)*x_ref[1]
        ref[2,2,2,0]=(-0.565466963253)*x_ref[0]**o+(-0.983283191302)*x_ref[0]+(0.0517997901365)*x_ref[1]**o+(-0.00459375732072)*x_ref[1]
        ref[2,2,2,1]=(-0.544844126966)*x_ref[0]**o+(-0.993222368569)*x_ref[0]+(-0.0761437907904)*x_ref[1]**o+(0.959202974542)*x_ref[1]
        ref[2,3,0,0]=(-0.542176217992)*x_ref[0]**o+(-0.210550241274)*x_ref[0]+(0.447732450287)*x_ref[1]**o+(-0.0802692719358)*x_ref[1]
        ref[2,3,0,1]=(0.339422485349)*x_ref[0]**o+(0.581409183541)*x_ref[0]+(-0.255731595248)*x_ref[1]**o+(-0.196957128507)*x_ref[1]
        ref[2,3,1,0]=(0.938606287977)*x_ref[0]**o+(0.396408957375)*x_ref[0]+(-0.53613392333)*x_ref[1]**o+(-0.201412967427)*x_ref[1]
        ref[2,3,1,1]=(0.482931877644)*x_ref[0]**o+(-0.395740204538)*x_ref[0]+(-0.816216426797)*x_ref[1]**o+(-0.390149600366)*x_ref[1]
        ref[2,3,2,0]=(0.507838191989)*x_ref[0]**o+(0.590218664943)*x_ref[0]+(0.987052963874)*x_ref[1]**o+(-0.39735585208)*x_ref[1]
        ref[2,3,2,1]=(-0.036562453009)*x_ref[0]**o+(0.299852609043)*x_ref[0]+(0.194160324844)*x_ref[1]**o+(-0.629803094543)*x_ref[1]
        ref[2,4,0,0]=(0.128664458677)*x_ref[0]**o+(0.956037220714)*x_ref[0]+(0.770331226849)*x_ref[1]**o+(-0.208007445309)*x_ref[1]
        ref[2,4,0,1]=(0.526755720598)*x_ref[0]**o+(-0.334650615987)*x_ref[0]+(-0.245693030991)*x_ref[1]**o+(-0.712134261554)*x_ref[1]
        ref[2,4,1,0]=(0.827770360865)*x_ref[0]**o+(0.982906973774)*x_ref[0]+(0.471438390576)*x_ref[1]**o+(0.691215965917)*x_ref[1]
        ref[2,4,1,1]=(-0.206715480236)*x_ref[0]**o+(0.488812089863)*x_ref[0]+(0.777672390635)*x_ref[1]**o+(-0.496681850755)*x_ref[1]
        ref[2,4,2,0]=(0.183938264875)*x_ref[0]**o+(-0.247363713904)*x_ref[0]+(0.962928810844)*x_ref[1]**o+(0.642706194018)*x_ref[1]
        ref[2,4,2,1]=(0.758025105584)*x_ref[0]**o+(-0.484793869316)*x_ref[0]+(-0.431089127161)*x_ref[1]**o+(0.127528834965)*x_ref[1]
        ref[3,0,0,0]=(-0.130978215422)*x_ref[0]**o+(0.688819586128)*x_ref[0]+(0.517162828175)*x_ref[1]**o+(-0.979632247431)*x_ref[1]
        ref[3,0,0,1]=(0.130939629053)*x_ref[0]**o+(0.685452224086)*x_ref[0]+(-0.438454293093)*x_ref[1]**o+(-0.727705397967)*x_ref[1]
        ref[3,0,1,0]=(-0.31299808275)*x_ref[0]**o+(0.266897949461)*x_ref[0]+(0.117566429054)*x_ref[1]**o+(0.536075754797)*x_ref[1]
        ref[3,0,1,1]=(-0.401764430381)*x_ref[0]**o+(-0.205983383212)*x_ref[0]+(0.461300727769)*x_ref[1]**o+(-0.580777919325)*x_ref[1]
        ref[3,0,2,0]=(-0.63526664417)*x_ref[0]**o+(-0.878687700719)*x_ref[0]+(0.104141198375)*x_ref[1]**o+(-0.539520033395)*x_ref[1]
        ref[3,0,2,1]=(-0.592054773422)*x_ref[0]**o+(-0.10222451991)*x_ref[0]+(-0.385428960408)*x_ref[1]**o+(-0.497192270367)*x_ref[1]
        ref[3,1,0,0]=(0.17159987373)*x_ref[0]**o+(-0.215728352568)*x_ref[0]+(0.188842078024)*x_ref[1]**o+(0.417647732421)*x_ref[1]
        ref[3,1,0,1]=(0.548015992401)*x_ref[0]**o+(-0.669157307443)*x_ref[0]+(-0.796467777561)*x_ref[1]**o+(0.163833145799)*x_ref[1]
        ref[3,1,1,0]=(-0.643762450865)*x_ref[0]**o+(0.548750992658)*x_ref[0]+(0.409211186464)*x_ref[1]**o+(-0.986601038131)*x_ref[1]
        ref[3,1,1,1]=(-0.882340195766)*x_ref[0]**o+(-0.185979117217)*x_ref[0]+(-0.00826877676749)*x_ref[1]**o+(0.0657705611536)*x_ref[1]
        ref[3,1,2,0]=(-0.0644145920386)*x_ref[0]**o+(0.18740710155)*x_ref[0]+(-0.690647279595)*x_ref[1]**o+(-0.185556727699)*x_ref[1]
        ref[3,1,2,1]=(0.114366852618)*x_ref[0]**o+(-0.388821787469)*x_ref[0]+(-0.195241613331)*x_ref[1]**o+(0.871862920816)*x_ref[1]
        ref[3,2,0,0]=(-0.673828733159)*x_ref[0]**o+(-0.172601630353)*x_ref[0]+(-0.488329313517)*x_ref[1]**o+(0.0809307315053)*x_ref[1]
        ref[3,2,0,1]=(0.565273426718)*x_ref[0]**o+(-0.610423824924)*x_ref[0]+(0.856623167832)*x_ref[1]**o+(-0.148845473418)*x_ref[1]
        ref[3,2,1,0]=(-0.3472751889)*x_ref[0]**o+(0.908614631864)*x_ref[0]+(0.445654124992)*x_ref[1]**o+(0.719429313439)*x_ref[1]
        ref[3,2,1,1]=(0.24891832441)*x_ref[0]**o+(-0.650081727005)*x_ref[0]+(0.0980208592314)*x_ref[1]**o+(0.575231534075)*x_ref[1]
        ref[3,2,2,0]=(-0.737935474986)*x_ref[0]**o+(-0.874708935739)*x_ref[0]+(0.856037958766)*x_ref[1]**o+(-0.420041922056)*x_ref[1]
        ref[3,2,2,1]=(0.106345127894)*x_ref[0]**o+(0.363060447152)*x_ref[0]+(-0.472262909157)*x_ref[1]**o+(-0.231475435242)*x_ref[1]
        ref[3,3,0,0]=(-0.3699135022)*x_ref[0]**o+(-0.167427922355)*x_ref[0]+(-0.911307660539)*x_ref[1]**o+(-0.445726901976)*x_ref[1]
        ref[3,3,0,1]=(0.505339735087)*x_ref[0]**o+(0.410541984072)*x_ref[0]+(0.847415917938)*x_ref[1]**o+(0.434325090295)*x_ref[1]
        ref[3,3,1,0]=(0.2589814894)*x_ref[0]**o+(-0.087780879445)*x_ref[0]+(0.847904119817)*x_ref[1]**o+(0.462585909981)*x_ref[1]
        ref[3,3,1,1]=(0.283347628256)*x_ref[0]**o+(0.637029684847)*x_ref[0]+(0.174554376017)*x_ref[1]**o+(0.642427538978)*x_ref[1]
        ref[3,3,2,0]=(-0.599348189011)*x_ref[0]**o+(0.087002991683)*x_ref[0]+(0.823297612508)*x_ref[1]**o+(0.84180029964)*x_ref[1]
        ref[3,3,2,1]=(0.846090423225)*x_ref[0]**o+(-0.465062745856)*x_ref[0]+(0.498670439686)*x_ref[1]**o+(0.721831089205)*x_ref[1]
        ref[3,4,0,0]=(0.0344263713162)*x_ref[0]**o+(0.789039070423)*x_ref[0]+(-0.518898476617)*x_ref[1]**o+(0.112666233369)*x_ref[1]
        ref[3,4,0,1]=(0.752855926629)*x_ref[0]**o+(-0.028671093428)*x_ref[0]+(0.457225098582)*x_ref[1]**o+(0.311273673822)*x_ref[1]
        ref[3,4,1,0]=(-0.623675057135)*x_ref[0]**o+(-0.19730173917)*x_ref[0]+(-0.328378661061)*x_ref[1]**o+(-0.196083495589)*x_ref[1]
        ref[3,4,1,1]=(0.722548420108)*x_ref[0]**o+(0.471733735786)*x_ref[0]+(-0.151931791011)*x_ref[1]**o+(0.0147226290316)*x_ref[1]
        ref[3,4,2,0]=(0.512232160342)*x_ref[0]**o+(-0.339879887275)*x_ref[0]+(-0.261940015624)*x_ref[1]**o+(-0.589559690156)*x_ref[1]
        ref[3,4,2,1]=(0.387751869794)*x_ref[0]**o+(0.284749443635)*x_ref[0]+(0.764067938503)*x_ref[1]**o+(0.20999603354)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.844350537532)*x[0]**o+(0.981166530382)*x[0]+(0.255920408286)*x[1]**o+(0.483711485104)*x[1]+(0.00959687239474)*x[2]**o+(0.306323594446)*x[2]
        s[0,0,0,1]=(0.740076186971)*x[0]**o+(-0.491347138384)*x[0]+(0.25175965888)*x[1]**o+(-0.623273215882)*x[1]+(0.51258068256)*x[2]**o+(-0.546204518632)*x[2]
        s[0,0,1,0]=(0.135495798592)*x[0]**o+(-0.928404807377)*x[0]+(-0.385528692626)*x[1]**o+(0.533554015352)*x[1]+(0.117724513259)*x[2]**o+(0.572637990898)*x[2]
        s[0,0,1,1]=(-0.850632098694)*x[0]**o+(-0.488169472233)*x[0]+(0.503260539233)*x[1]**o+(-0.859633434356)*x[1]+(0.415712182486)*x[2]**o+(0.339616059154)*x[2]
        s[0,0,2,0]=(-0.018555867134)*x[0]**o+(0.861452555488)*x[0]+(0.381089258556)*x[1]**o+(0.101855012068)*x[1]+(-0.15705292069)*x[2]**o+(-0.826396062146)*x[2]
        s[0,0,2,1]=(0.160998814621)*x[0]**o+(-0.410483868941)*x[0]+(-0.159057698561)*x[1]**o+(0.78589010204)*x[1]+(0.766903367421)*x[2]**o+(0.102093838786)*x[2]
        s[0,1,0,0]=(-0.445025622437)*x[0]**o+(-0.683407434608)*x[0]+(0.415296785191)*x[1]**o+(-0.720473598436)*x[1]+(0.892104458869)*x[2]**o+(0.678624532111)*x[2]
        s[0,1,0,1]=(0.494607245554)*x[0]**o+(-0.261983074835)*x[0]+(0.659330653523)*x[1]**o+(0.517010331237)*x[1]+(0.33680183722)*x[2]**o+(0.984150461948)*x[2]
        s[0,1,1,0]=(-0.461365247044)*x[0]**o+(-0.913532792608)*x[0]+(0.0548715431963)*x[1]**o+(-0.195304655857)*x[1]+(-0.680191903401)*x[2]**o+(-0.344258045633)*x[2]
        s[0,1,1,1]=(-0.366551699639)*x[0]**o+(0.0380843408355)*x[0]+(-0.959343312241)*x[1]**o+(-0.0236083869931)*x[1]+(-0.728804551514)*x[2]**o+(0.411917097114)*x[2]
        s[0,1,2,0]=(-0.755044693231)*x[0]**o+(-0.589493821888)*x[0]+(-0.675570377341)*x[1]**o+(0.712529396708)*x[1]+(0.121878507837)*x[2]**o+(-0.500132877994)*x[2]
        s[0,1,2,1]=(-0.613779986247)*x[0]**o+(0.977105357645)*x[0]+(-0.82406258445)*x[1]**o+(0.109110319693)*x[1]+(-0.384891270465)*x[2]**o+(0.0513108774825)*x[2]
        s[0,2,0,0]=(-0.215753544701)*x[0]**o+(-0.485302830852)*x[0]+(0.305187675609)*x[1]**o+(-0.852320135062)*x[1]+(0.479016046098)*x[2]**o+(0.904887182515)*x[2]
        s[0,2,0,1]=(-0.902665442283)*x[0]**o+(0.339847683115)*x[0]+(-0.201409198091)*x[1]**o+(-0.28916884043)*x[1]+(0.0122973146477)*x[2]**o+(0.902178823571)*x[2]
        s[0,2,1,0]=(-0.345222234379)*x[0]**o+(-0.402351849375)*x[0]+(-0.830448872698)*x[1]**o+(-0.00616252140166)*x[1]+(-0.973740549537)*x[2]**o+(-0.430178208352)*x[2]
        s[0,2,1,1]=(0.602597376682)*x[0]**o+(0.19777615169)*x[0]+(0.432285684454)*x[1]**o+(0.837105284275)*x[1]+(0.674505043227)*x[2]**o+(0.908555029565)*x[2]
        s[0,2,2,0]=(0.05197142873)*x[0]**o+(0.23981420979)*x[0]+(-0.630581895848)*x[1]**o+(-0.823407671488)*x[1]+(-0.819697754602)*x[2]**o+(-0.0606600491833)*x[2]
        s[0,2,2,1]=(0.671981483754)*x[0]**o+(0.273934751949)*x[0]+(0.680845824642)*x[1]**o+(-0.243860132213)*x[1]+(0.166045811143)*x[2]**o+(0.661029385371)*x[2]
        s[0,3,0,0]=(-0.599325667157)*x[0]**o+(0.295726750611)*x[0]+(-0.446959020679)*x[1]**o+(-0.934383765903)*x[1]+(-0.142759892188)*x[2]**o+(0.892076097753)*x[2]
        s[0,3,0,1]=(-0.081435415733)*x[0]**o+(-0.914632751608)*x[0]+(-0.0356505325211)*x[1]**o+(-0.594158265622)*x[1]+(-0.183149255345)*x[2]**o+(0.992685124274)*x[2]
        s[0,3,1,0]=(-0.52041548403)*x[0]**o+(0.84346502916)*x[0]+(0.589777480478)*x[1]**o+(-0.395593542394)*x[1]+(0.670347108347)*x[2]**o+(-0.755732891394)*x[2]
        s[0,3,1,1]=(-0.0843404796224)*x[0]**o+(-0.350884489082)*x[0]+(-0.230005943865)*x[1]**o+(0.775052308561)*x[1]+(-0.722944185393)*x[2]**o+(-0.661807415523)*x[2]
        s[0,3,2,0]=(0.562202744376)*x[0]**o+(0.755433320508)*x[0]+(0.983360558953)*x[1]**o+(-0.18576911233)*x[1]+(-0.546634474194)*x[2]**o+(-0.448629525918)*x[2]
        s[0,3,2,1]=(0.418182343511)*x[0]**o+(-0.804470815559)*x[0]+(0.328075033205)*x[1]**o+(0.794791705586)*x[1]+(0.955923558558)*x[2]**o+(0.477204042313)*x[2]
        s[0,4,0,0]=(-0.473795825706)*x[0]**o+(0.0613706961191)*x[0]+(0.728253884932)*x[1]**o+(0.922149327337)*x[1]+(-0.514934433516)*x[2]**o+(-0.350899100204)*x[2]
        s[0,4,0,1]=(0.403705624315)*x[0]**o+(0.332463550032)*x[0]+(0.0413433808287)*x[1]**o+(-0.88184294823)*x[1]+(0.898896470347)*x[2]**o+(-0.1692829666)*x[2]
        s[0,4,1,0]=(-0.0696293975354)*x[0]**o+(0.697969617481)*x[0]+(-0.872250160629)*x[1]**o+(-0.396555707549)*x[1]+(0.398974483685)*x[2]**o+(-0.38500615055)*x[2]
        s[0,4,1,1]=(-0.818302505854)*x[0]**o+(-0.660712096214)*x[0]+(-0.687407491883)*x[1]**o+(0.662996835898)*x[1]+(-0.777440711816)*x[2]**o+(0.901132206169)*x[2]
        s[0,4,2,0]=(-0.98506239403)*x[0]**o+(-0.457191412508)*x[0]+(0.203796091348)*x[1]**o+(-0.0304382176582)*x[1]+(0.375197892442)*x[2]**o+(0.0794166875091)*x[2]
        s[0,4,2,1]=(-0.939331221009)*x[0]**o+(0.662619061409)*x[0]+(0.550834749841)*x[1]**o+(0.895500286676)*x[1]+(-0.886048606057)*x[2]**o+(-0.15220276644)*x[2]
        s[1,0,0,0]=(-0.553168089515)*x[0]**o+(-0.631287474707)*x[0]+(0.236518377648)*x[1]**o+(0.798404510441)*x[1]+(0.33391551032)*x[2]**o+(0.723204832044)*x[2]
        s[1,0,0,1]=(0.618774989743)*x[0]**o+(-0.625192149767)*x[0]+(-0.423599611797)*x[1]**o+(0.0627868455822)*x[1]+(0.248591794171)*x[2]**o+(-0.568908092813)*x[2]
        s[1,0,1,0]=(-0.856803065565)*x[0]**o+(0.615617576001)*x[0]+(-0.689638395609)*x[1]**o+(-0.756681900602)*x[1]+(-0.226065925851)*x[2]**o+(0.348925448185)*x[2]
        s[1,0,1,1]=(0.801728371493)*x[0]**o+(-0.436203474211)*x[0]+(-0.0202077001179)*x[1]**o+(0.652651006485)*x[1]+(-0.230375932312)*x[2]**o+(-0.890003876378)*x[2]
        s[1,0,2,0]=(0.29719598244)*x[0]**o+(-0.15276643068)*x[0]+(-0.753013966455)*x[1]**o+(-0.517288289121)*x[1]+(-0.218994832652)*x[2]**o+(-0.146759568045)*x[2]
        s[1,0,2,1]=(-0.823649332214)*x[0]**o+(-0.843687314415)*x[0]+(-0.0638848545941)*x[1]**o+(0.722449122523)*x[1]+(0.491191260847)*x[2]**o+(-0.752018096725)*x[2]
        s[1,1,0,0]=(0.605980591392)*x[0]**o+(0.202355210592)*x[0]+(0.379171937943)*x[1]**o+(0.063757128966)*x[1]+(-0.971141711866)*x[2]**o+(-0.632361023174)*x[2]
        s[1,1,0,1]=(-0.962798234484)*x[0]**o+(0.446686088092)*x[0]+(-0.521649313187)*x[1]**o+(-0.756613553959)*x[1]+(0.598374656129)*x[2]**o+(0.203329622383)*x[2]
        s[1,1,1,0]=(0.300526220295)*x[0]**o+(-0.821407258021)*x[0]+(-0.519739408983)*x[1]**o+(-0.300945102018)*x[1]+(0.592444505162)*x[2]**o+(0.093896688069)*x[2]
        s[1,1,1,1]=(-0.683108409273)*x[0]**o+(-0.724514513218)*x[0]+(-0.672965343753)*x[1]**o+(-0.990600638457)*x[1]+(0.288343195848)*x[2]**o+(0.836765793149)*x[2]
        s[1,1,2,0]=(-0.938695582412)*x[0]**o+(0.524424434979)*x[0]+(-0.230446441603)*x[1]**o+(-0.660840814146)*x[1]+(0.410079472542)*x[2]**o+(-0.501145734383)*x[2]
        s[1,1,2,1]=(-0.934625690674)*x[0]**o+(-0.355510911308)*x[0]+(0.112799513937)*x[1]**o+(0.0550081941593)*x[1]+(-0.564671701833)*x[2]**o+(-0.435422588591)*x[2]
        s[1,2,0,0]=(-0.329336582979)*x[0]**o+(0.557448032082)*x[0]+(0.660452325378)*x[1]**o+(0.986140231524)*x[1]+(-0.343267425889)*x[2]**o+(-0.124929320203)*x[2]
        s[1,2,0,1]=(0.141679080878)*x[0]**o+(-0.929990482459)*x[0]+(-0.0124869738018)*x[1]**o+(-0.297954361644)*x[1]+(0.237086926195)*x[2]**o+(-0.163693767697)*x[2]
        s[1,2,1,0]=(0.840523338196)*x[0]**o+(0.292937319051)*x[0]+(-0.647937602564)*x[1]**o+(0.599823564217)*x[1]+(0.966887229971)*x[2]**o+(-0.301069847494)*x[2]
        s[1,2,1,1]=(0.75606641666)*x[0]**o+(0.694906659273)*x[0]+(0.894582185091)*x[1]**o+(0.431150046844)*x[1]+(0.988226420754)*x[2]**o+(-0.860220033688)*x[2]
        s[1,2,2,0]=(-0.716168748278)*x[0]**o+(0.464108589188)*x[0]+(0.631539541512)*x[1]**o+(-0.992859266146)*x[1]+(-0.669793952605)*x[2]**o+(-0.743066646537)*x[2]
        s[1,2,2,1]=(0.695010429853)*x[0]**o+(-0.56668549077)*x[0]+(0.242451432508)*x[1]**o+(0.155033232196)*x[1]+(-0.362798375291)*x[2]**o+(-0.488187632585)*x[2]
        s[1,3,0,0]=(-0.943147179882)*x[0]**o+(0.71046937082)*x[0]+(0.852528992189)*x[1]**o+(-0.193096337802)*x[1]+(-0.787307409902)*x[2]**o+(0.251655133658)*x[2]
        s[1,3,0,1]=(-0.208552630722)*x[0]**o+(0.346012986086)*x[0]+(0.551360403822)*x[1]**o+(-0.348440170895)*x[1]+(0.363393290055)*x[2]**o+(0.0470822356286)*x[2]
        s[1,3,1,0]=(0.113226447773)*x[0]**o+(0.522623018987)*x[0]+(-0.0796736494068)*x[1]**o+(0.789297223214)*x[1]+(-0.779573310177)*x[2]**o+(-0.743401293513)*x[2]
        s[1,3,1,1]=(-0.466035390795)*x[0]**o+(0.511762679374)*x[0]+(-0.682170406028)*x[1]**o+(-0.949494996805)*x[1]+(-0.0648915682154)*x[2]**o+(0.495239120908)*x[2]
        s[1,3,2,0]=(-0.274148045419)*x[0]**o+(0.0815161328992)*x[0]+(0.358273465722)*x[1]**o+(0.636995118303)*x[1]+(-0.535665346021)*x[2]**o+(0.823457431671)*x[2]
        s[1,3,2,1]=(-0.716452730034)*x[0]**o+(-0.483846622879)*x[0]+(-0.608042660277)*x[1]**o+(0.0278424023059)*x[1]+(0.843814111075)*x[2]**o+(0.955433558186)*x[2]
        s[1,4,0,0]=(0.455371216931)*x[0]**o+(0.602586579764)*x[0]+(0.158622918138)*x[1]**o+(-0.706234646453)*x[1]+(-0.475064482018)*x[2]**o+(0.146600391902)*x[2]
        s[1,4,0,1]=(0.980992566699)*x[0]**o+(0.117314780276)*x[0]+(-0.158086008164)*x[1]**o+(0.632974744247)*x[1]+(0.978173293529)*x[2]**o+(0.885266669359)*x[2]
        s[1,4,1,0]=(0.279463579513)*x[0]**o+(0.402328966807)*x[0]+(0.301400738565)*x[1]**o+(-0.955899292995)*x[1]+(-0.500757799904)*x[2]**o+(0.281493491795)*x[2]
        s[1,4,1,1]=(-0.822078727162)*x[0]**o+(0.369502628318)*x[0]+(-0.928613467846)*x[1]**o+(-0.350001970293)*x[1]+(-0.968163790337)*x[2]**o+(-0.310282764085)*x[2]
        s[1,4,2,0]=(0.3290445083)*x[0]**o+(-0.985109449578)*x[0]+(-0.632388659807)*x[1]**o+(0.466010510408)*x[1]+(0.460974186748)*x[2]**o+(-0.716019421615)*x[2]
        s[1,4,2,1]=(0.72236273453)*x[0]**o+(-0.0121296025323)*x[0]+(-0.846055371974)*x[1]**o+(0.984741432507)*x[1]+(-0.522899574925)*x[2]**o+(-0.21853151342)*x[2]
        s[2,0,0,0]=(0.0109795220027)*x[0]**o+(0.556433154187)*x[0]+(-0.770646566445)*x[1]**o+(-0.736722512756)*x[1]+(0.668238695822)*x[2]**o+(0.180946691645)*x[2]
        s[2,0,0,1]=(0.719005782249)*x[0]**o+(-0.872006790171)*x[0]+(-0.957147163648)*x[1]**o+(-0.0682732175112)*x[1]+(0.583562455856)*x[2]**o+(-0.30991382684)*x[2]
        s[2,0,1,0]=(0.622360614797)*x[0]**o+(-0.00796068126083)*x[0]+(0.3225801025)*x[1]**o+(-0.919808515024)*x[1]+(-0.919597324996)*x[2]**o+(0.373038030942)*x[2]
        s[2,0,1,1]=(0.405340603348)*x[0]**o+(-0.0495755391664)*x[0]+(-0.47787299709)*x[1]**o+(-0.254771467734)*x[1]+(-0.591700035231)*x[2]**o+(-0.164217809972)*x[2]
        s[2,0,2,0]=(0.087396380793)*x[0]**o+(-0.374710687926)*x[0]+(0.0575627557138)*x[1]**o+(-0.70523617076)*x[1]+(-0.484089875239)*x[2]**o+(0.789318249087)*x[2]
        s[2,0,2,1]=(0.778697410279)*x[0]**o+(0.519333641831)*x[0]+(-0.123493167274)*x[1]**o+(0.366951337816)*x[1]+(-0.537900250619)*x[2]**o+(0.627028303525)*x[2]
        s[2,1,0,0]=(-0.745940888357)*x[0]**o+(0.711657374433)*x[0]+(0.836411404243)*x[1]**o+(-0.174311748628)*x[1]+(0.776203832353)*x[2]**o+(-0.706629687992)*x[2]
        s[2,1,0,1]=(-0.642638091614)*x[0]**o+(0.172213012143)*x[0]+(0.0413400352844)*x[1]**o+(-0.69201275421)*x[1]+(-0.824191389562)*x[2]**o+(0.251335644925)*x[2]
        s[2,1,1,0]=(-0.749920542498)*x[0]**o+(0.389193707611)*x[0]+(0.413903115322)*x[1]**o+(0.74472403825)*x[1]+(-0.591260331172)*x[2]**o+(0.276810972178)*x[2]
        s[2,1,1,1]=(0.439899556721)*x[0]**o+(0.566349932103)*x[0]+(-0.741172831484)*x[1]**o+(-0.983770395204)*x[1]+(-0.337604226356)*x[2]**o+(-0.984256469612)*x[2]
        s[2,1,2,0]=(-0.744905118064)*x[0]**o+(0.0758118730978)*x[0]+(-0.954906020107)*x[1]**o+(-0.321181040222)*x[1]+(-0.319406466958)*x[2]**o+(-0.462108015037)*x[2]
        s[2,1,2,1]=(0.495181652645)*x[0]**o+(0.321912068502)*x[0]+(0.625371783418)*x[1]**o+(0.661392879428)*x[1]+(0.95878113647)*x[2]**o+(-0.00161261625228)*x[2]
        s[2,2,0,0]=(-0.747880472493)*x[0]**o+(0.133266267487)*x[0]+(0.594232553468)*x[1]**o+(0.164357025178)*x[1]+(0.183282050202)*x[2]**o+(0.510098487732)*x[2]
        s[2,2,0,1]=(-0.2990096515)*x[0]**o+(-0.530653405949)*x[0]+(-0.774239397996)*x[1]**o+(0.513406622276)*x[1]+(0.68969837178)*x[2]**o+(0.0542139499089)*x[2]
        s[2,2,1,0]=(0.803320557325)*x[0]**o+(-0.53977892905)*x[0]+(0.602881824151)*x[1]**o+(0.151232673003)*x[1]+(0.933315439891)*x[2]**o+(-0.0915356088824)*x[2]
        s[2,2,1,1]=(-0.573022070609)*x[0]**o+(0.394673455213)*x[0]+(0.290948264558)*x[1]**o+(0.482816198889)*x[1]+(0.457587694713)*x[2]**o+(0.146402139866)*x[2]
        s[2,2,2,0]=(0.0984322495009)*x[0]**o+(-0.607655184921)*x[0]+(-0.165102103099)*x[1]**o+(-0.949911109313)*x[1]+(0.807082265921)*x[2]**o+(-0.201828673075)*x[2]
        s[2,2,2,1]=(0.387692723867)*x[0]**o+(-0.350126071452)*x[0]+(-0.840248594934)*x[1]**o+(-0.754015389174)*x[1]+(0.82690792431)*x[2]**o+(0.984937777626)*x[2]
        s[2,3,0,0]=(-0.22249653492)*x[0]**o+(0.95589235945)*x[0]+(0.748620345408)*x[1]**o+(-0.544563975574)*x[1]+(0.399260091432)*x[2]**o+(-0.664314569966)*x[2]
        s[2,3,0,1]=(-0.881513041393)*x[0]**o+(-0.286371325145)*x[0]+(0.207149857405)*x[1]**o+(0.962030218672)*x[1]+(-0.621096606228)*x[2]**o+(0.321143840842)*x[2]
        s[2,3,1,0]=(-0.294822275479)*x[0]**o+(0.952468891463)*x[0]+(0.890797831223)*x[1]**o+(-0.154408717393)*x[1]+(0.653171114568)*x[2]**o+(-0.335321213379)*x[2]
        s[2,3,1,1]=(0.391962263482)*x[0]**o+(0.567275318333)*x[0]+(-0.185626734066)*x[1]**o+(-0.377715271388)*x[1]+(0.739574492513)*x[2]**o+(-0.603873103899)*x[2]
        s[2,3,2,0]=(0.428479962249)*x[0]**o+(0.127318777736)*x[0]+(-0.42333631821)*x[1]**o+(-0.169907154638)*x[1]+(0.725857578617)*x[2]**o+(-0.697584549589)*x[2]
        s[2,3,2,1]=(0.372507332734)*x[0]**o+(0.381914565887)*x[0]+(0.218450152368)*x[1]**o+(-0.802951368728)*x[1]+(-0.00721884700159)*x[2]**o+(0.071666772532)*x[2]
        s[2,4,0,0]=(-0.520915873118)*x[0]**o+(-0.904770241319)*x[0]+(0.594347214938)*x[1]**o+(-0.681345109104)*x[1]+(0.209360616612)*x[2]**o+(0.424835494205)*x[2]
        s[2,4,0,1]=(-0.727883496639)*x[0]**o+(-0.668568348644)*x[0]+(0.833770360827)*x[1]**o+(-0.0255151478099)*x[1]+(-0.193240025032)*x[2]**o+(0.197362196363)*x[2]
        s[2,4,1,0]=(0.742869556969)*x[0]**o+(0.475432034925)*x[0]+(0.357310924868)*x[1]**o+(-0.905361920184)*x[1]+(-0.0284188203545)*x[2]**o+(-0.028888633895)*x[2]
        s[2,4,1,1]=(-0.331466533146)*x[0]**o+(0.544600410552)*x[0]+(0.959927301686)*x[1]**o+(-0.875596269307)*x[1]+(0.0907336453125)*x[2]**o+(0.638768254635)*x[2]
        s[2,4,2,0]=(-0.140519776938)*x[0]**o+(-0.341221155863)*x[0]+(0.476853206406)*x[1]**o+(-0.376640947632)*x[1]+(0.222188072931)*x[2]**o+(-0.600634989398)*x[2]
        s[2,4,2,1]=(0.436824352085)*x[0]**o+(0.0902891669206)*x[0]+(0.950721335793)*x[1]**o+(0.578125459387)*x[1]+(-0.482219010223)*x[2]**o+(-0.232985764124)*x[2]
        s[3,0,0,0]=(-0.571010773207)*x[0]**o+(-0.80484920281)*x[0]+(-0.823071194682)*x[1]**o+(-0.896108871228)*x[1]+(-0.631430205832)*x[2]**o+(-0.704868332657)*x[2]
        s[3,0,0,1]=(0.273328305518)*x[0]**o+(0.391247949879)*x[0]+(-0.577768086094)*x[1]**o+(-0.0268817540424)*x[1]+(-0.807017351875)*x[2]**o+(0.855714026407)*x[2]
        s[3,0,1,0]=(-0.783392552436)*x[0]**o+(-0.606379147005)*x[0]+(0.596212011594)*x[1]**o+(0.837204930781)*x[1]+(-0.444629434195)*x[2]**o+(0.818388831162)*x[2]
        s[3,0,1,1]=(-0.7632340237)*x[0]**o+(-0.525136694101)*x[0]+(0.79942659492)*x[1]**o+(-0.404546089352)*x[1]+(-0.728139040689)*x[2]**o+(-0.998101874873)*x[2]
        s[3,0,2,0]=(-0.509268960989)*x[0]**o+(-0.676093609952)*x[0]+(0.861170135902)*x[1]**o+(0.577425218873)*x[1]+(0.577209383315)*x[2]**o+(-0.384779054446)*x[2]
        s[3,0,2,1]=(0.011431602973)*x[0]**o+(-0.334328421599)*x[0]+(0.252604277681)*x[1]**o+(-0.603913209811)*x[1]+(0.00917009818384)*x[2]**o+(0.301867115072)*x[2]
        s[3,1,0,0]=(-0.620651658587)*x[0]**o+(0.814248186892)*x[0]+(0.231571604024)*x[1]**o+(-0.208758448156)*x[1]+(0.514917171695)*x[2]**o+(0.795545898743)*x[2]
        s[3,1,0,1]=(-0.358180244409)*x[0]**o+(0.762252417714)*x[0]+(0.235442173264)*x[1]**o+(0.681677639507)*x[1]+(-0.610029281327)*x[2]**o+(0.255430019725)*x[2]
        s[3,1,1,0]=(-0.337744414049)*x[0]**o+(0.659764964399)*x[0]+(-0.93916869727)*x[1]**o+(0.0920440667279)*x[1]+(0.9429852742)*x[2]**o+(-0.152643224429)*x[2]
        s[3,1,1,1]=(-0.850080149157)*x[0]**o+(0.150991562708)*x[0]+(-0.265559714378)*x[1]**o+(0.0825721664033)*x[1]+(0.0577388830574)*x[2]**o+(-0.0168755109359)*x[2]
        s[3,1,2,0]=(0.736517689601)*x[0]**o+(-0.299770623584)*x[0]+(-0.642480331197)*x[1]**o+(0.507838878495)*x[1]+(0.102240842692)*x[2]**o+(-0.474668897224)*x[2]
        s[3,1,2,1]=(0.718251907196)*x[0]**o+(-0.732535005154)*x[0]+(-0.0595472398188)*x[1]**o+(-0.809928140752)*x[1]+(-0.949798402613)*x[2]**o+(-0.817043878578)*x[2]
        s[3,2,0,0]=(-0.226866840941)*x[0]**o+(0.981200696871)*x[0]+(0.329563738019)*x[1]**o+(-0.859364161406)*x[1]+(-0.0889989821964)*x[2]**o+(-0.891130810171)*x[2]
        s[3,2,0,1]=(0.448731899011)*x[0]**o+(-0.295976710644)*x[0]+(-0.0631712972169)*x[1]**o+(-0.0910278410217)*x[1]+(-0.568889472919)*x[2]**o+(0.763619087805)*x[2]
        s[3,2,1,0]=(0.383474130267)*x[0]**o+(0.73677599326)*x[0]+(-0.208757244921)*x[1]**o+(-0.278191993802)*x[1]+(0.795441377078)*x[2]**o+(0.905216800847)*x[2]
        s[3,2,1,1]=(0.0725196158569)*x[0]**o+(-0.943723180474)*x[0]+(0.699282898995)*x[1]**o+(0.388214849035)*x[1]+(0.596007988212)*x[2]**o+(0.736772507174)*x[2]
        s[3,2,2,0]=(-0.448938945133)*x[0]**o+(0.384414869647)*x[0]+(0.527747975961)*x[1]**o+(0.127653616477)*x[1]+(0.377808521053)*x[2]**o+(-0.252616720439)*x[2]
        s[3,2,2,1]=(-0.376250853312)*x[0]**o+(-0.707650293785)*x[0]+(0.672982484116)*x[1]**o+(0.990265204194)*x[1]+(0.529775927528)*x[2]**o+(-0.409195688811)*x[2]
        s[3,3,0,0]=(-0.530159451732)*x[0]**o+(-0.955185089804)*x[0]+(-0.748512782868)*x[1]**o+(-0.618229320649)*x[1]+(0.627829579969)*x[2]**o+(-0.333717653758)*x[2]
        s[3,3,0,1]=(0.966906267367)*x[0]**o+(-0.0951932785216)*x[0]+(-0.61695602507)*x[1]**o+(0.446458255991)*x[1]+(0.207784087287)*x[2]**o+(-0.288316765862)*x[2]
        s[3,3,1,0]=(-0.996153491218)*x[0]**o+(0.0588370396384)*x[0]+(-0.294277377632)*x[1]**o+(0.689632857934)*x[1]+(-0.143160246331)*x[2]**o+(0.833437162663)*x[2]
        s[3,3,1,1]=(0.906322391848)*x[0]**o+(0.492397941858)*x[0]+(-0.0266527290051)*x[1]**o+(-0.172670670559)*x[1]+(-0.826003550146)*x[2]**o+(-0.103742447179)*x[2]
        s[3,3,2,0]=(-0.479133669526)*x[0]**o+(0.577467684597)*x[0]+(-0.214066405495)*x[1]**o+(0.0589162518627)*x[1]+(-0.618757108816)*x[2]**o+(-0.0596902292359)*x[2]
        s[3,3,2,1]=(0.501813126681)*x[0]**o+(-0.0527123864794)*x[0]+(0.442612384316)*x[1]**o+(0.587529973102)*x[1]+(0.96334743671)*x[2]**o+(-0.878243725936)*x[2]
        s[3,4,0,0]=(-0.283828850179)*x[0]**o+(0.553981240206)*x[0]+(0.526151552782)*x[1]**o+(-0.822056178815)*x[1]+(-0.922360625781)*x[2]**o+(-0.0972249338686)*x[2]
        s[3,4,0,1]=(0.513911875447)*x[0]**o+(0.66906817932)*x[0]+(-0.843990607619)*x[1]**o+(-0.873810309081)*x[1]+(0.9732978504)*x[2]**o+(-0.767991735083)*x[2]
        s[3,4,1,0]=(-0.918970664697)*x[0]**o+(0.409739379673)*x[0]+(0.856834850961)*x[1]**o+(-0.149117026418)*x[1]+(-0.215870849261)*x[2]**o+(-0.139382801219)*x[2]
        s[3,4,1,1]=(0.869749963776)*x[0]**o+(-0.0551502951625)*x[0]+(-0.211684312437)*x[1]**o+(0.674587225438)*x[1]+(0.713205722506)*x[2]**o+(0.943169316201)*x[2]
        s[3,4,2,0]=(-0.0849462931018)*x[0]**o+(0.131197062211)*x[0]+(-0.462104529576)*x[1]**o+(0.620371430368)*x[1]+(0.858403152981)*x[2]**o+(0.102462631271)*x[2]
        s[3,4,2,1]=(-0.402094215027)*x[0]**o+(-0.246651318354)*x[0]+(-0.741149807021)*x[1]**o+(-0.616686358212)*x[1]+(-0.606948106264)*x[2]**o+(-0.911839049257)*x[2]
        ref[0,0,0,0]=(-0.844350537532)*x_ref[0]**o+(0.981166530382)*x_ref[0]+(0.255920408286)*x_ref[1]**o+(0.483711485104)*x_ref[1]+(0.00959687239474)*x_ref[2]**o+(0.306323594446)*x_ref[2]
        ref[0,0,0,1]=(0.740076186971)*x_ref[0]**o+(-0.491347138384)*x_ref[0]+(0.25175965888)*x_ref[1]**o+(-0.623273215882)*x_ref[1]+(0.51258068256)*x_ref[2]**o+(-0.546204518632)*x_ref[2]
        ref[0,0,1,0]=(0.135495798592)*x_ref[0]**o+(-0.928404807377)*x_ref[0]+(-0.385528692626)*x_ref[1]**o+(0.533554015352)*x_ref[1]+(0.117724513259)*x_ref[2]**o+(0.572637990898)*x_ref[2]
        ref[0,0,1,1]=(-0.850632098694)*x_ref[0]**o+(-0.488169472233)*x_ref[0]+(0.503260539233)*x_ref[1]**o+(-0.859633434356)*x_ref[1]+(0.415712182486)*x_ref[2]**o+(0.339616059154)*x_ref[2]
        ref[0,0,2,0]=(-0.018555867134)*x_ref[0]**o+(0.861452555488)*x_ref[0]+(0.381089258556)*x_ref[1]**o+(0.101855012068)*x_ref[1]+(-0.15705292069)*x_ref[2]**o+(-0.826396062146)*x_ref[2]
        ref[0,0,2,1]=(0.160998814621)*x_ref[0]**o+(-0.410483868941)*x_ref[0]+(-0.159057698561)*x_ref[1]**o+(0.78589010204)*x_ref[1]+(0.766903367421)*x_ref[2]**o+(0.102093838786)*x_ref[2]
        ref[0,1,0,0]=(-0.445025622437)*x_ref[0]**o+(-0.683407434608)*x_ref[0]+(0.415296785191)*x_ref[1]**o+(-0.720473598436)*x_ref[1]+(0.892104458869)*x_ref[2]**o+(0.678624532111)*x_ref[2]
        ref[0,1,0,1]=(0.494607245554)*x_ref[0]**o+(-0.261983074835)*x_ref[0]+(0.659330653523)*x_ref[1]**o+(0.517010331237)*x_ref[1]+(0.33680183722)*x_ref[2]**o+(0.984150461948)*x_ref[2]
        ref[0,1,1,0]=(-0.461365247044)*x_ref[0]**o+(-0.913532792608)*x_ref[0]+(0.0548715431963)*x_ref[1]**o+(-0.195304655857)*x_ref[1]+(-0.680191903401)*x_ref[2]**o+(-0.344258045633)*x_ref[2]
        ref[0,1,1,1]=(-0.366551699639)*x_ref[0]**o+(0.0380843408355)*x_ref[0]+(-0.959343312241)*x_ref[1]**o+(-0.0236083869931)*x_ref[1]+(-0.728804551514)*x_ref[2]**o+(0.411917097114)*x_ref[2]
        ref[0,1,2,0]=(-0.755044693231)*x_ref[0]**o+(-0.589493821888)*x_ref[0]+(-0.675570377341)*x_ref[1]**o+(0.712529396708)*x_ref[1]+(0.121878507837)*x_ref[2]**o+(-0.500132877994)*x_ref[2]
        ref[0,1,2,1]=(-0.613779986247)*x_ref[0]**o+(0.977105357645)*x_ref[0]+(-0.82406258445)*x_ref[1]**o+(0.109110319693)*x_ref[1]+(-0.384891270465)*x_ref[2]**o+(0.0513108774825)*x_ref[2]
        ref[0,2,0,0]=(-0.215753544701)*x_ref[0]**o+(-0.485302830852)*x_ref[0]+(0.305187675609)*x_ref[1]**o+(-0.852320135062)*x_ref[1]+(0.479016046098)*x_ref[2]**o+(0.904887182515)*x_ref[2]
        ref[0,2,0,1]=(-0.902665442283)*x_ref[0]**o+(0.339847683115)*x_ref[0]+(-0.201409198091)*x_ref[1]**o+(-0.28916884043)*x_ref[1]+(0.0122973146477)*x_ref[2]**o+(0.902178823571)*x_ref[2]
        ref[0,2,1,0]=(-0.345222234379)*x_ref[0]**o+(-0.402351849375)*x_ref[0]+(-0.830448872698)*x_ref[1]**o+(-0.00616252140166)*x_ref[1]+(-0.973740549537)*x_ref[2]**o+(-0.430178208352)*x_ref[2]
        ref[0,2,1,1]=(0.602597376682)*x_ref[0]**o+(0.19777615169)*x_ref[0]+(0.432285684454)*x_ref[1]**o+(0.837105284275)*x_ref[1]+(0.674505043227)*x_ref[2]**o+(0.908555029565)*x_ref[2]
        ref[0,2,2,0]=(0.05197142873)*x_ref[0]**o+(0.23981420979)*x_ref[0]+(-0.630581895848)*x_ref[1]**o+(-0.823407671488)*x_ref[1]+(-0.819697754602)*x_ref[2]**o+(-0.0606600491833)*x_ref[2]
        ref[0,2,2,1]=(0.671981483754)*x_ref[0]**o+(0.273934751949)*x_ref[0]+(0.680845824642)*x_ref[1]**o+(-0.243860132213)*x_ref[1]+(0.166045811143)*x_ref[2]**o+(0.661029385371)*x_ref[2]
        ref[0,3,0,0]=(-0.599325667157)*x_ref[0]**o+(0.295726750611)*x_ref[0]+(-0.446959020679)*x_ref[1]**o+(-0.934383765903)*x_ref[1]+(-0.142759892188)*x_ref[2]**o+(0.892076097753)*x_ref[2]
        ref[0,3,0,1]=(-0.081435415733)*x_ref[0]**o+(-0.914632751608)*x_ref[0]+(-0.0356505325211)*x_ref[1]**o+(-0.594158265622)*x_ref[1]+(-0.183149255345)*x_ref[2]**o+(0.992685124274)*x_ref[2]
        ref[0,3,1,0]=(-0.52041548403)*x_ref[0]**o+(0.84346502916)*x_ref[0]+(0.589777480478)*x_ref[1]**o+(-0.395593542394)*x_ref[1]+(0.670347108347)*x_ref[2]**o+(-0.755732891394)*x_ref[2]
        ref[0,3,1,1]=(-0.0843404796224)*x_ref[0]**o+(-0.350884489082)*x_ref[0]+(-0.230005943865)*x_ref[1]**o+(0.775052308561)*x_ref[1]+(-0.722944185393)*x_ref[2]**o+(-0.661807415523)*x_ref[2]
        ref[0,3,2,0]=(0.562202744376)*x_ref[0]**o+(0.755433320508)*x_ref[0]+(0.983360558953)*x_ref[1]**o+(-0.18576911233)*x_ref[1]+(-0.546634474194)*x_ref[2]**o+(-0.448629525918)*x_ref[2]
        ref[0,3,2,1]=(0.418182343511)*x_ref[0]**o+(-0.804470815559)*x_ref[0]+(0.328075033205)*x_ref[1]**o+(0.794791705586)*x_ref[1]+(0.955923558558)*x_ref[2]**o+(0.477204042313)*x_ref[2]
        ref[0,4,0,0]=(-0.473795825706)*x_ref[0]**o+(0.0613706961191)*x_ref[0]+(0.728253884932)*x_ref[1]**o+(0.922149327337)*x_ref[1]+(-0.514934433516)*x_ref[2]**o+(-0.350899100204)*x_ref[2]
        ref[0,4,0,1]=(0.403705624315)*x_ref[0]**o+(0.332463550032)*x_ref[0]+(0.0413433808287)*x_ref[1]**o+(-0.88184294823)*x_ref[1]+(0.898896470347)*x_ref[2]**o+(-0.1692829666)*x_ref[2]
        ref[0,4,1,0]=(-0.0696293975354)*x_ref[0]**o+(0.697969617481)*x_ref[0]+(-0.872250160629)*x_ref[1]**o+(-0.396555707549)*x_ref[1]+(0.398974483685)*x_ref[2]**o+(-0.38500615055)*x_ref[2]
        ref[0,4,1,1]=(-0.818302505854)*x_ref[0]**o+(-0.660712096214)*x_ref[0]+(-0.687407491883)*x_ref[1]**o+(0.662996835898)*x_ref[1]+(-0.777440711816)*x_ref[2]**o+(0.901132206169)*x_ref[2]
        ref[0,4,2,0]=(-0.98506239403)*x_ref[0]**o+(-0.457191412508)*x_ref[0]+(0.203796091348)*x_ref[1]**o+(-0.0304382176582)*x_ref[1]+(0.375197892442)*x_ref[2]**o+(0.0794166875091)*x_ref[2]
        ref[0,4,2,1]=(-0.939331221009)*x_ref[0]**o+(0.662619061409)*x_ref[0]+(0.550834749841)*x_ref[1]**o+(0.895500286676)*x_ref[1]+(-0.886048606057)*x_ref[2]**o+(-0.15220276644)*x_ref[2]
        ref[1,0,0,0]=(-0.553168089515)*x_ref[0]**o+(-0.631287474707)*x_ref[0]+(0.236518377648)*x_ref[1]**o+(0.798404510441)*x_ref[1]+(0.33391551032)*x_ref[2]**o+(0.723204832044)*x_ref[2]
        ref[1,0,0,1]=(0.618774989743)*x_ref[0]**o+(-0.625192149767)*x_ref[0]+(-0.423599611797)*x_ref[1]**o+(0.0627868455822)*x_ref[1]+(0.248591794171)*x_ref[2]**o+(-0.568908092813)*x_ref[2]
        ref[1,0,1,0]=(-0.856803065565)*x_ref[0]**o+(0.615617576001)*x_ref[0]+(-0.689638395609)*x_ref[1]**o+(-0.756681900602)*x_ref[1]+(-0.226065925851)*x_ref[2]**o+(0.348925448185)*x_ref[2]
        ref[1,0,1,1]=(0.801728371493)*x_ref[0]**o+(-0.436203474211)*x_ref[0]+(-0.0202077001179)*x_ref[1]**o+(0.652651006485)*x_ref[1]+(-0.230375932312)*x_ref[2]**o+(-0.890003876378)*x_ref[2]
        ref[1,0,2,0]=(0.29719598244)*x_ref[0]**o+(-0.15276643068)*x_ref[0]+(-0.753013966455)*x_ref[1]**o+(-0.517288289121)*x_ref[1]+(-0.218994832652)*x_ref[2]**o+(-0.146759568045)*x_ref[2]
        ref[1,0,2,1]=(-0.823649332214)*x_ref[0]**o+(-0.843687314415)*x_ref[0]+(-0.0638848545941)*x_ref[1]**o+(0.722449122523)*x_ref[1]+(0.491191260847)*x_ref[2]**o+(-0.752018096725)*x_ref[2]
        ref[1,1,0,0]=(0.605980591392)*x_ref[0]**o+(0.202355210592)*x_ref[0]+(0.379171937943)*x_ref[1]**o+(0.063757128966)*x_ref[1]+(-0.971141711866)*x_ref[2]**o+(-0.632361023174)*x_ref[2]
        ref[1,1,0,1]=(-0.962798234484)*x_ref[0]**o+(0.446686088092)*x_ref[0]+(-0.521649313187)*x_ref[1]**o+(-0.756613553959)*x_ref[1]+(0.598374656129)*x_ref[2]**o+(0.203329622383)*x_ref[2]
        ref[1,1,1,0]=(0.300526220295)*x_ref[0]**o+(-0.821407258021)*x_ref[0]+(-0.519739408983)*x_ref[1]**o+(-0.300945102018)*x_ref[1]+(0.592444505162)*x_ref[2]**o+(0.093896688069)*x_ref[2]
        ref[1,1,1,1]=(-0.683108409273)*x_ref[0]**o+(-0.724514513218)*x_ref[0]+(-0.672965343753)*x_ref[1]**o+(-0.990600638457)*x_ref[1]+(0.288343195848)*x_ref[2]**o+(0.836765793149)*x_ref[2]
        ref[1,1,2,0]=(-0.938695582412)*x_ref[0]**o+(0.524424434979)*x_ref[0]+(-0.230446441603)*x_ref[1]**o+(-0.660840814146)*x_ref[1]+(0.410079472542)*x_ref[2]**o+(-0.501145734383)*x_ref[2]
        ref[1,1,2,1]=(-0.934625690674)*x_ref[0]**o+(-0.355510911308)*x_ref[0]+(0.112799513937)*x_ref[1]**o+(0.0550081941593)*x_ref[1]+(-0.564671701833)*x_ref[2]**o+(-0.435422588591)*x_ref[2]
        ref[1,2,0,0]=(-0.329336582979)*x_ref[0]**o+(0.557448032082)*x_ref[0]+(0.660452325378)*x_ref[1]**o+(0.986140231524)*x_ref[1]+(-0.343267425889)*x_ref[2]**o+(-0.124929320203)*x_ref[2]
        ref[1,2,0,1]=(0.141679080878)*x_ref[0]**o+(-0.929990482459)*x_ref[0]+(-0.0124869738018)*x_ref[1]**o+(-0.297954361644)*x_ref[1]+(0.237086926195)*x_ref[2]**o+(-0.163693767697)*x_ref[2]
        ref[1,2,1,0]=(0.840523338196)*x_ref[0]**o+(0.292937319051)*x_ref[0]+(-0.647937602564)*x_ref[1]**o+(0.599823564217)*x_ref[1]+(0.966887229971)*x_ref[2]**o+(-0.301069847494)*x_ref[2]
        ref[1,2,1,1]=(0.75606641666)*x_ref[0]**o+(0.694906659273)*x_ref[0]+(0.894582185091)*x_ref[1]**o+(0.431150046844)*x_ref[1]+(0.988226420754)*x_ref[2]**o+(-0.860220033688)*x_ref[2]
        ref[1,2,2,0]=(-0.716168748278)*x_ref[0]**o+(0.464108589188)*x_ref[0]+(0.631539541512)*x_ref[1]**o+(-0.992859266146)*x_ref[1]+(-0.669793952605)*x_ref[2]**o+(-0.743066646537)*x_ref[2]
        ref[1,2,2,1]=(0.695010429853)*x_ref[0]**o+(-0.56668549077)*x_ref[0]+(0.242451432508)*x_ref[1]**o+(0.155033232196)*x_ref[1]+(-0.362798375291)*x_ref[2]**o+(-0.488187632585)*x_ref[2]
        ref[1,3,0,0]=(-0.943147179882)*x_ref[0]**o+(0.71046937082)*x_ref[0]+(0.852528992189)*x_ref[1]**o+(-0.193096337802)*x_ref[1]+(-0.787307409902)*x_ref[2]**o+(0.251655133658)*x_ref[2]
        ref[1,3,0,1]=(-0.208552630722)*x_ref[0]**o+(0.346012986086)*x_ref[0]+(0.551360403822)*x_ref[1]**o+(-0.348440170895)*x_ref[1]+(0.363393290055)*x_ref[2]**o+(0.0470822356286)*x_ref[2]
        ref[1,3,1,0]=(0.113226447773)*x_ref[0]**o+(0.522623018987)*x_ref[0]+(-0.0796736494068)*x_ref[1]**o+(0.789297223214)*x_ref[1]+(-0.779573310177)*x_ref[2]**o+(-0.743401293513)*x_ref[2]
        ref[1,3,1,1]=(-0.466035390795)*x_ref[0]**o+(0.511762679374)*x_ref[0]+(-0.682170406028)*x_ref[1]**o+(-0.949494996805)*x_ref[1]+(-0.0648915682154)*x_ref[2]**o+(0.495239120908)*x_ref[2]
        ref[1,3,2,0]=(-0.274148045419)*x_ref[0]**o+(0.0815161328992)*x_ref[0]+(0.358273465722)*x_ref[1]**o+(0.636995118303)*x_ref[1]+(-0.535665346021)*x_ref[2]**o+(0.823457431671)*x_ref[2]
        ref[1,3,2,1]=(-0.716452730034)*x_ref[0]**o+(-0.483846622879)*x_ref[0]+(-0.608042660277)*x_ref[1]**o+(0.0278424023059)*x_ref[1]+(0.843814111075)*x_ref[2]**o+(0.955433558186)*x_ref[2]
        ref[1,4,0,0]=(0.455371216931)*x_ref[0]**o+(0.602586579764)*x_ref[0]+(0.158622918138)*x_ref[1]**o+(-0.706234646453)*x_ref[1]+(-0.475064482018)*x_ref[2]**o+(0.146600391902)*x_ref[2]
        ref[1,4,0,1]=(0.980992566699)*x_ref[0]**o+(0.117314780276)*x_ref[0]+(-0.158086008164)*x_ref[1]**o+(0.632974744247)*x_ref[1]+(0.978173293529)*x_ref[2]**o+(0.885266669359)*x_ref[2]
        ref[1,4,1,0]=(0.279463579513)*x_ref[0]**o+(0.402328966807)*x_ref[0]+(0.301400738565)*x_ref[1]**o+(-0.955899292995)*x_ref[1]+(-0.500757799904)*x_ref[2]**o+(0.281493491795)*x_ref[2]
        ref[1,4,1,1]=(-0.822078727162)*x_ref[0]**o+(0.369502628318)*x_ref[0]+(-0.928613467846)*x_ref[1]**o+(-0.350001970293)*x_ref[1]+(-0.968163790337)*x_ref[2]**o+(-0.310282764085)*x_ref[2]
        ref[1,4,2,0]=(0.3290445083)*x_ref[0]**o+(-0.985109449578)*x_ref[0]+(-0.632388659807)*x_ref[1]**o+(0.466010510408)*x_ref[1]+(0.460974186748)*x_ref[2]**o+(-0.716019421615)*x_ref[2]
        ref[1,4,2,1]=(0.72236273453)*x_ref[0]**o+(-0.0121296025323)*x_ref[0]+(-0.846055371974)*x_ref[1]**o+(0.984741432507)*x_ref[1]+(-0.522899574925)*x_ref[2]**o+(-0.21853151342)*x_ref[2]
        ref[2,0,0,0]=(0.0109795220027)*x_ref[0]**o+(0.556433154187)*x_ref[0]+(-0.770646566445)*x_ref[1]**o+(-0.736722512756)*x_ref[1]+(0.668238695822)*x_ref[2]**o+(0.180946691645)*x_ref[2]
        ref[2,0,0,1]=(0.719005782249)*x_ref[0]**o+(-0.872006790171)*x_ref[0]+(-0.957147163648)*x_ref[1]**o+(-0.0682732175112)*x_ref[1]+(0.583562455856)*x_ref[2]**o+(-0.30991382684)*x_ref[2]
        ref[2,0,1,0]=(0.622360614797)*x_ref[0]**o+(-0.00796068126083)*x_ref[0]+(0.3225801025)*x_ref[1]**o+(-0.919808515024)*x_ref[1]+(-0.919597324996)*x_ref[2]**o+(0.373038030942)*x_ref[2]
        ref[2,0,1,1]=(0.405340603348)*x_ref[0]**o+(-0.0495755391664)*x_ref[0]+(-0.47787299709)*x_ref[1]**o+(-0.254771467734)*x_ref[1]+(-0.591700035231)*x_ref[2]**o+(-0.164217809972)*x_ref[2]
        ref[2,0,2,0]=(0.087396380793)*x_ref[0]**o+(-0.374710687926)*x_ref[0]+(0.0575627557138)*x_ref[1]**o+(-0.70523617076)*x_ref[1]+(-0.484089875239)*x_ref[2]**o+(0.789318249087)*x_ref[2]
        ref[2,0,2,1]=(0.778697410279)*x_ref[0]**o+(0.519333641831)*x_ref[0]+(-0.123493167274)*x_ref[1]**o+(0.366951337816)*x_ref[1]+(-0.537900250619)*x_ref[2]**o+(0.627028303525)*x_ref[2]
        ref[2,1,0,0]=(-0.745940888357)*x_ref[0]**o+(0.711657374433)*x_ref[0]+(0.836411404243)*x_ref[1]**o+(-0.174311748628)*x_ref[1]+(0.776203832353)*x_ref[2]**o+(-0.706629687992)*x_ref[2]
        ref[2,1,0,1]=(-0.642638091614)*x_ref[0]**o+(0.172213012143)*x_ref[0]+(0.0413400352844)*x_ref[1]**o+(-0.69201275421)*x_ref[1]+(-0.824191389562)*x_ref[2]**o+(0.251335644925)*x_ref[2]
        ref[2,1,1,0]=(-0.749920542498)*x_ref[0]**o+(0.389193707611)*x_ref[0]+(0.413903115322)*x_ref[1]**o+(0.74472403825)*x_ref[1]+(-0.591260331172)*x_ref[2]**o+(0.276810972178)*x_ref[2]
        ref[2,1,1,1]=(0.439899556721)*x_ref[0]**o+(0.566349932103)*x_ref[0]+(-0.741172831484)*x_ref[1]**o+(-0.983770395204)*x_ref[1]+(-0.337604226356)*x_ref[2]**o+(-0.984256469612)*x_ref[2]
        ref[2,1,2,0]=(-0.744905118064)*x_ref[0]**o+(0.0758118730978)*x_ref[0]+(-0.954906020107)*x_ref[1]**o+(-0.321181040222)*x_ref[1]+(-0.319406466958)*x_ref[2]**o+(-0.462108015037)*x_ref[2]
        ref[2,1,2,1]=(0.495181652645)*x_ref[0]**o+(0.321912068502)*x_ref[0]+(0.625371783418)*x_ref[1]**o+(0.661392879428)*x_ref[1]+(0.95878113647)*x_ref[2]**o+(-0.00161261625228)*x_ref[2]
        ref[2,2,0,0]=(-0.747880472493)*x_ref[0]**o+(0.133266267487)*x_ref[0]+(0.594232553468)*x_ref[1]**o+(0.164357025178)*x_ref[1]+(0.183282050202)*x_ref[2]**o+(0.510098487732)*x_ref[2]
        ref[2,2,0,1]=(-0.2990096515)*x_ref[0]**o+(-0.530653405949)*x_ref[0]+(-0.774239397996)*x_ref[1]**o+(0.513406622276)*x_ref[1]+(0.68969837178)*x_ref[2]**o+(0.0542139499089)*x_ref[2]
        ref[2,2,1,0]=(0.803320557325)*x_ref[0]**o+(-0.53977892905)*x_ref[0]+(0.602881824151)*x_ref[1]**o+(0.151232673003)*x_ref[1]+(0.933315439891)*x_ref[2]**o+(-0.0915356088824)*x_ref[2]
        ref[2,2,1,1]=(-0.573022070609)*x_ref[0]**o+(0.394673455213)*x_ref[0]+(0.290948264558)*x_ref[1]**o+(0.482816198889)*x_ref[1]+(0.457587694713)*x_ref[2]**o+(0.146402139866)*x_ref[2]
        ref[2,2,2,0]=(0.0984322495009)*x_ref[0]**o+(-0.607655184921)*x_ref[0]+(-0.165102103099)*x_ref[1]**o+(-0.949911109313)*x_ref[1]+(0.807082265921)*x_ref[2]**o+(-0.201828673075)*x_ref[2]
        ref[2,2,2,1]=(0.387692723867)*x_ref[0]**o+(-0.350126071452)*x_ref[0]+(-0.840248594934)*x_ref[1]**o+(-0.754015389174)*x_ref[1]+(0.82690792431)*x_ref[2]**o+(0.984937777626)*x_ref[2]
        ref[2,3,0,0]=(-0.22249653492)*x_ref[0]**o+(0.95589235945)*x_ref[0]+(0.748620345408)*x_ref[1]**o+(-0.544563975574)*x_ref[1]+(0.399260091432)*x_ref[2]**o+(-0.664314569966)*x_ref[2]
        ref[2,3,0,1]=(-0.881513041393)*x_ref[0]**o+(-0.286371325145)*x_ref[0]+(0.207149857405)*x_ref[1]**o+(0.962030218672)*x_ref[1]+(-0.621096606228)*x_ref[2]**o+(0.321143840842)*x_ref[2]
        ref[2,3,1,0]=(-0.294822275479)*x_ref[0]**o+(0.952468891463)*x_ref[0]+(0.890797831223)*x_ref[1]**o+(-0.154408717393)*x_ref[1]+(0.653171114568)*x_ref[2]**o+(-0.335321213379)*x_ref[2]
        ref[2,3,1,1]=(0.391962263482)*x_ref[0]**o+(0.567275318333)*x_ref[0]+(-0.185626734066)*x_ref[1]**o+(-0.377715271388)*x_ref[1]+(0.739574492513)*x_ref[2]**o+(-0.603873103899)*x_ref[2]
        ref[2,3,2,0]=(0.428479962249)*x_ref[0]**o+(0.127318777736)*x_ref[0]+(-0.42333631821)*x_ref[1]**o+(-0.169907154638)*x_ref[1]+(0.725857578617)*x_ref[2]**o+(-0.697584549589)*x_ref[2]
        ref[2,3,2,1]=(0.372507332734)*x_ref[0]**o+(0.381914565887)*x_ref[0]+(0.218450152368)*x_ref[1]**o+(-0.802951368728)*x_ref[1]+(-0.00721884700159)*x_ref[2]**o+(0.071666772532)*x_ref[2]
        ref[2,4,0,0]=(-0.520915873118)*x_ref[0]**o+(-0.904770241319)*x_ref[0]+(0.594347214938)*x_ref[1]**o+(-0.681345109104)*x_ref[1]+(0.209360616612)*x_ref[2]**o+(0.424835494205)*x_ref[2]
        ref[2,4,0,1]=(-0.727883496639)*x_ref[0]**o+(-0.668568348644)*x_ref[0]+(0.833770360827)*x_ref[1]**o+(-0.0255151478099)*x_ref[1]+(-0.193240025032)*x_ref[2]**o+(0.197362196363)*x_ref[2]
        ref[2,4,1,0]=(0.742869556969)*x_ref[0]**o+(0.475432034925)*x_ref[0]+(0.357310924868)*x_ref[1]**o+(-0.905361920184)*x_ref[1]+(-0.0284188203545)*x_ref[2]**o+(-0.028888633895)*x_ref[2]
        ref[2,4,1,1]=(-0.331466533146)*x_ref[0]**o+(0.544600410552)*x_ref[0]+(0.959927301686)*x_ref[1]**o+(-0.875596269307)*x_ref[1]+(0.0907336453125)*x_ref[2]**o+(0.638768254635)*x_ref[2]
        ref[2,4,2,0]=(-0.140519776938)*x_ref[0]**o+(-0.341221155863)*x_ref[0]+(0.476853206406)*x_ref[1]**o+(-0.376640947632)*x_ref[1]+(0.222188072931)*x_ref[2]**o+(-0.600634989398)*x_ref[2]
        ref[2,4,2,1]=(0.436824352085)*x_ref[0]**o+(0.0902891669206)*x_ref[0]+(0.950721335793)*x_ref[1]**o+(0.578125459387)*x_ref[1]+(-0.482219010223)*x_ref[2]**o+(-0.232985764124)*x_ref[2]
        ref[3,0,0,0]=(-0.571010773207)*x_ref[0]**o+(-0.80484920281)*x_ref[0]+(-0.823071194682)*x_ref[1]**o+(-0.896108871228)*x_ref[1]+(-0.631430205832)*x_ref[2]**o+(-0.704868332657)*x_ref[2]
        ref[3,0,0,1]=(0.273328305518)*x_ref[0]**o+(0.391247949879)*x_ref[0]+(-0.577768086094)*x_ref[1]**o+(-0.0268817540424)*x_ref[1]+(-0.807017351875)*x_ref[2]**o+(0.855714026407)*x_ref[2]
        ref[3,0,1,0]=(-0.783392552436)*x_ref[0]**o+(-0.606379147005)*x_ref[0]+(0.596212011594)*x_ref[1]**o+(0.837204930781)*x_ref[1]+(-0.444629434195)*x_ref[2]**o+(0.818388831162)*x_ref[2]
        ref[3,0,1,1]=(-0.7632340237)*x_ref[0]**o+(-0.525136694101)*x_ref[0]+(0.79942659492)*x_ref[1]**o+(-0.404546089352)*x_ref[1]+(-0.728139040689)*x_ref[2]**o+(-0.998101874873)*x_ref[2]
        ref[3,0,2,0]=(-0.509268960989)*x_ref[0]**o+(-0.676093609952)*x_ref[0]+(0.861170135902)*x_ref[1]**o+(0.577425218873)*x_ref[1]+(0.577209383315)*x_ref[2]**o+(-0.384779054446)*x_ref[2]
        ref[3,0,2,1]=(0.011431602973)*x_ref[0]**o+(-0.334328421599)*x_ref[0]+(0.252604277681)*x_ref[1]**o+(-0.603913209811)*x_ref[1]+(0.00917009818384)*x_ref[2]**o+(0.301867115072)*x_ref[2]
        ref[3,1,0,0]=(-0.620651658587)*x_ref[0]**o+(0.814248186892)*x_ref[0]+(0.231571604024)*x_ref[1]**o+(-0.208758448156)*x_ref[1]+(0.514917171695)*x_ref[2]**o+(0.795545898743)*x_ref[2]
        ref[3,1,0,1]=(-0.358180244409)*x_ref[0]**o+(0.762252417714)*x_ref[0]+(0.235442173264)*x_ref[1]**o+(0.681677639507)*x_ref[1]+(-0.610029281327)*x_ref[2]**o+(0.255430019725)*x_ref[2]
        ref[3,1,1,0]=(-0.337744414049)*x_ref[0]**o+(0.659764964399)*x_ref[0]+(-0.93916869727)*x_ref[1]**o+(0.0920440667279)*x_ref[1]+(0.9429852742)*x_ref[2]**o+(-0.152643224429)*x_ref[2]
        ref[3,1,1,1]=(-0.850080149157)*x_ref[0]**o+(0.150991562708)*x_ref[0]+(-0.265559714378)*x_ref[1]**o+(0.0825721664033)*x_ref[1]+(0.0577388830574)*x_ref[2]**o+(-0.0168755109359)*x_ref[2]
        ref[3,1,2,0]=(0.736517689601)*x_ref[0]**o+(-0.299770623584)*x_ref[0]+(-0.642480331197)*x_ref[1]**o+(0.507838878495)*x_ref[1]+(0.102240842692)*x_ref[2]**o+(-0.474668897224)*x_ref[2]
        ref[3,1,2,1]=(0.718251907196)*x_ref[0]**o+(-0.732535005154)*x_ref[0]+(-0.0595472398188)*x_ref[1]**o+(-0.809928140752)*x_ref[1]+(-0.949798402613)*x_ref[2]**o+(-0.817043878578)*x_ref[2]
        ref[3,2,0,0]=(-0.226866840941)*x_ref[0]**o+(0.981200696871)*x_ref[0]+(0.329563738019)*x_ref[1]**o+(-0.859364161406)*x_ref[1]+(-0.0889989821964)*x_ref[2]**o+(-0.891130810171)*x_ref[2]
        ref[3,2,0,1]=(0.448731899011)*x_ref[0]**o+(-0.295976710644)*x_ref[0]+(-0.0631712972169)*x_ref[1]**o+(-0.0910278410217)*x_ref[1]+(-0.568889472919)*x_ref[2]**o+(0.763619087805)*x_ref[2]
        ref[3,2,1,0]=(0.383474130267)*x_ref[0]**o+(0.73677599326)*x_ref[0]+(-0.208757244921)*x_ref[1]**o+(-0.278191993802)*x_ref[1]+(0.795441377078)*x_ref[2]**o+(0.905216800847)*x_ref[2]
        ref[3,2,1,1]=(0.0725196158569)*x_ref[0]**o+(-0.943723180474)*x_ref[0]+(0.699282898995)*x_ref[1]**o+(0.388214849035)*x_ref[1]+(0.596007988212)*x_ref[2]**o+(0.736772507174)*x_ref[2]
        ref[3,2,2,0]=(-0.448938945133)*x_ref[0]**o+(0.384414869647)*x_ref[0]+(0.527747975961)*x_ref[1]**o+(0.127653616477)*x_ref[1]+(0.377808521053)*x_ref[2]**o+(-0.252616720439)*x_ref[2]
        ref[3,2,2,1]=(-0.376250853312)*x_ref[0]**o+(-0.707650293785)*x_ref[0]+(0.672982484116)*x_ref[1]**o+(0.990265204194)*x_ref[1]+(0.529775927528)*x_ref[2]**o+(-0.409195688811)*x_ref[2]
        ref[3,3,0,0]=(-0.530159451732)*x_ref[0]**o+(-0.955185089804)*x_ref[0]+(-0.748512782868)*x_ref[1]**o+(-0.618229320649)*x_ref[1]+(0.627829579969)*x_ref[2]**o+(-0.333717653758)*x_ref[2]
        ref[3,3,0,1]=(0.966906267367)*x_ref[0]**o+(-0.0951932785216)*x_ref[0]+(-0.61695602507)*x_ref[1]**o+(0.446458255991)*x_ref[1]+(0.207784087287)*x_ref[2]**o+(-0.288316765862)*x_ref[2]
        ref[3,3,1,0]=(-0.996153491218)*x_ref[0]**o+(0.0588370396384)*x_ref[0]+(-0.294277377632)*x_ref[1]**o+(0.689632857934)*x_ref[1]+(-0.143160246331)*x_ref[2]**o+(0.833437162663)*x_ref[2]
        ref[3,3,1,1]=(0.906322391848)*x_ref[0]**o+(0.492397941858)*x_ref[0]+(-0.0266527290051)*x_ref[1]**o+(-0.172670670559)*x_ref[1]+(-0.826003550146)*x_ref[2]**o+(-0.103742447179)*x_ref[2]
        ref[3,3,2,0]=(-0.479133669526)*x_ref[0]**o+(0.577467684597)*x_ref[0]+(-0.214066405495)*x_ref[1]**o+(0.0589162518627)*x_ref[1]+(-0.618757108816)*x_ref[2]**o+(-0.0596902292359)*x_ref[2]
        ref[3,3,2,1]=(0.501813126681)*x_ref[0]**o+(-0.0527123864794)*x_ref[0]+(0.442612384316)*x_ref[1]**o+(0.587529973102)*x_ref[1]+(0.96334743671)*x_ref[2]**o+(-0.878243725936)*x_ref[2]
        ref[3,4,0,0]=(-0.283828850179)*x_ref[0]**o+(0.553981240206)*x_ref[0]+(0.526151552782)*x_ref[1]**o+(-0.822056178815)*x_ref[1]+(-0.922360625781)*x_ref[2]**o+(-0.0972249338686)*x_ref[2]
        ref[3,4,0,1]=(0.513911875447)*x_ref[0]**o+(0.66906817932)*x_ref[0]+(-0.843990607619)*x_ref[1]**o+(-0.873810309081)*x_ref[1]+(0.9732978504)*x_ref[2]**o+(-0.767991735083)*x_ref[2]
        ref[3,4,1,0]=(-0.918970664697)*x_ref[0]**o+(0.409739379673)*x_ref[0]+(0.856834850961)*x_ref[1]**o+(-0.149117026418)*x_ref[1]+(-0.215870849261)*x_ref[2]**o+(-0.139382801219)*x_ref[2]
        ref[3,4,1,1]=(0.869749963776)*x_ref[0]**o+(-0.0551502951625)*x_ref[0]+(-0.211684312437)*x_ref[1]**o+(0.674587225438)*x_ref[1]+(0.713205722506)*x_ref[2]**o+(0.943169316201)*x_ref[2]
        ref[3,4,2,0]=(-0.0849462931018)*x_ref[0]**o+(0.131197062211)*x_ref[0]+(-0.462104529576)*x_ref[1]**o+(0.620371430368)*x_ref[1]+(0.858403152981)*x_ref[2]**o+(0.102462631271)*x_ref[2]
        ref[3,4,2,1]=(-0.402094215027)*x_ref[0]**o+(-0.246651318354)*x_ref[0]+(-0.741149807021)*x_ref[1]**o+(-0.616686358212)*x_ref[1]+(-0.606948106264)*x_ref[2]**o+(-0.911839049257)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.521673635599)*x[0]**o+(-0.288664980886)*x[0]+(0.57565949318)*x[1]**o+(-0.555613874185)*x[1]
        ref=(-0.521673635599)*x_ref[0]**o+(-0.288664980886)*x_ref[0]+(0.57565949318)*x_ref[1]**o+(-0.555613874185)*x_ref[1]
      else:
        arg=(0.520429804914)*x[0]**o+(-0.000683147004841)*x[0]+(0.676879975826)*x[1]**o+(0.760707578765)*x[1]+(0.479225734268)*x[2]**o+(0.132006876687)*x[2]
        ref=(0.520429804914)*x_ref[0]**o+(-0.000683147004841)*x_ref[0]+(0.676879975826)*x_ref[1]**o+(0.760707578765)*x_ref[1]+(0.479225734268)*x_ref[2]**o+(0.132006876687)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.264333591663)*x[0]**o+(-0.884594789709)*x[0]+(0.809768917911)*x[1]**o+(-0.999064121889)*x[1]
        arg[1]=(-0.476663257532)*x[0]**o+(0.703121978294)*x[0]+(0.768222344055)*x[1]**o+(0.0126615637065)*x[1]
        ref[0]=(0.264333591663)*x_ref[0]**o+(-0.884594789709)*x_ref[0]+(0.809768917911)*x_ref[1]**o+(-0.999064121889)*x_ref[1]
        ref[1]=(-0.476663257532)*x_ref[0]**o+(0.703121978294)*x_ref[0]+(0.768222344055)*x_ref[1]**o+(0.0126615637065)*x_ref[1]
      else:
        arg[0]=(-0.814801897766)*x[0]**o+(0.417782597744)*x[0]+(0.354691012333)*x[1]**o+(-0.0529647248706)*x[1]+(-0.455085903639)*x[2]**o+(-0.501751102496)*x[2]
        arg[1]=(0.165354246971)*x[0]**o+(-0.238754573504)*x[0]+(-0.293225479626)*x[1]**o+(-0.42192384305)*x[1]+(-0.109509840118)*x[2]**o+(0.511666964461)*x[2]
        ref[0]=(-0.814801897766)*x_ref[0]**o+(0.417782597744)*x_ref[0]+(0.354691012333)*x_ref[1]**o+(-0.0529647248706)*x_ref[1]+(-0.455085903639)*x_ref[2]**o+(-0.501751102496)*x_ref[2]
        ref[1]=(0.165354246971)*x_ref[0]**o+(-0.238754573504)*x_ref[0]+(-0.293225479626)*x_ref[1]**o+(-0.42192384305)*x_ref[1]+(-0.109509840118)*x_ref[2]**o+(0.511666964461)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.469225177227)*x[0]**o+(-0.312308556752)*x[0]+(0.0207638302484)*x[1]**o+(-0.42312789395)*x[1]
        arg[0,1]=(0.435465800442)*x[0]**o+(0.66981866069)*x[0]+(-0.563664236008)*x[1]**o+(-0.979640187021)*x[1]
        arg[0,2]=(0.761422556363)*x[0]**o+(-0.374824039734)*x[0]+(-0.258990622087)*x[1]**o+(-0.553555723801)*x[1]
        arg[0,3]=(-0.346737956865)*x[0]**o+(-0.509820084012)*x[0]+(-0.270279814718)*x[1]**o+(-0.77787682055)*x[1]
        arg[0,4]=(-0.607151504575)*x[0]**o+(0.0370061792286)*x[0]+(-0.597194755504)*x[1]**o+(0.463748874183)*x[1]
        arg[1,0]=(-0.614871053235)*x[0]**o+(0.984151090349)*x[0]+(0.176558255935)*x[1]**o+(-0.00143697546002)*x[1]
        arg[1,1]=(-0.678550535699)*x[0]**o+(-0.40068020768)*x[0]+(-0.73609233692)*x[1]**o+(0.204641737633)*x[1]
        arg[1,2]=(0.644566611818)*x[0]**o+(0.676838663571)*x[0]+(0.592803079261)*x[1]**o+(-0.257851556005)*x[1]
        arg[1,3]=(-0.811008507599)*x[0]**o+(-0.810627035248)*x[0]+(0.470993697001)*x[1]**o+(0.524157551045)*x[1]
        arg[1,4]=(0.507989473752)*x[0]**o+(0.474629400564)*x[0]+(0.972372960952)*x[1]**o+(-0.713136826767)*x[1]
        arg[2,0]=(-0.18215096122)*x[0]**o+(0.0926167766546)*x[0]+(-0.526818906453)*x[1]**o+(-0.169071065789)*x[1]
        arg[2,1]=(-0.402265903759)*x[0]**o+(-0.460219787922)*x[0]+(0.160493468523)*x[1]**o+(-0.816631415356)*x[1]
        arg[2,2]=(-0.826968582935)*x[0]**o+(-0.477060281375)*x[0]+(-0.753204398647)*x[1]**o+(0.340944419869)*x[1]
        arg[2,3]=(-0.289758403392)*x[0]**o+(-0.216496476621)*x[0]+(-0.0729120551729)*x[1]**o+(0.0206397202194)*x[1]
        arg[2,4]=(0.900190015504)*x[0]**o+(0.554334407718)*x[0]+(-0.188091500011)*x[1]**o+(-0.239344021421)*x[1]
        arg[3,0]=(-0.846253042969)*x[0]**o+(0.212019924467)*x[0]+(-0.209309311993)*x[1]**o+(0.47889463943)*x[1]
        arg[3,1]=(-0.796067196362)*x[0]**o+(0.570882191233)*x[0]+(0.789930258)*x[1]**o+(-0.687939656508)*x[1]
        arg[3,2]=(0.0603011963958)*x[0]**o+(-0.291794860225)*x[0]+(-0.572598612637)*x[1]**o+(0.308585554728)*x[1]
        arg[3,3]=(0.883976313141)*x[0]**o+(0.857027876956)*x[0]+(-0.991113135222)*x[1]**o+(-0.0649848727938)*x[1]
        arg[3,4]=(-0.420604320052)*x[0]**o+(-0.147041915772)*x[0]+(0.696294525095)*x[1]**o+(-0.45791291995)*x[1]
        ref[0,0]=(-0.469225177227)*x_ref[0]**o+(-0.312308556752)*x_ref[0]+(0.0207638302484)*x_ref[1]**o+(-0.42312789395)*x_ref[1]
        ref[0,1]=(0.435465800442)*x_ref[0]**o+(0.66981866069)*x_ref[0]+(-0.563664236008)*x_ref[1]**o+(-0.979640187021)*x_ref[1]
        ref[0,2]=(0.761422556363)*x_ref[0]**o+(-0.374824039734)*x_ref[0]+(-0.258990622087)*x_ref[1]**o+(-0.553555723801)*x_ref[1]
        ref[0,3]=(-0.346737956865)*x_ref[0]**o+(-0.509820084012)*x_ref[0]+(-0.270279814718)*x_ref[1]**o+(-0.77787682055)*x_ref[1]
        ref[0,4]=(-0.607151504575)*x_ref[0]**o+(0.0370061792286)*x_ref[0]+(-0.597194755504)*x_ref[1]**o+(0.463748874183)*x_ref[1]
        ref[1,0]=(-0.614871053235)*x_ref[0]**o+(0.984151090349)*x_ref[0]+(0.176558255935)*x_ref[1]**o+(-0.00143697546002)*x_ref[1]
        ref[1,1]=(-0.678550535699)*x_ref[0]**o+(-0.40068020768)*x_ref[0]+(-0.73609233692)*x_ref[1]**o+(0.204641737633)*x_ref[1]
        ref[1,2]=(0.644566611818)*x_ref[0]**o+(0.676838663571)*x_ref[0]+(0.592803079261)*x_ref[1]**o+(-0.257851556005)*x_ref[1]
        ref[1,3]=(-0.811008507599)*x_ref[0]**o+(-0.810627035248)*x_ref[0]+(0.470993697001)*x_ref[1]**o+(0.524157551045)*x_ref[1]
        ref[1,4]=(0.507989473752)*x_ref[0]**o+(0.474629400564)*x_ref[0]+(0.972372960952)*x_ref[1]**o+(-0.713136826767)*x_ref[1]
        ref[2,0]=(-0.18215096122)*x_ref[0]**o+(0.0926167766546)*x_ref[0]+(-0.526818906453)*x_ref[1]**o+(-0.169071065789)*x_ref[1]
        ref[2,1]=(-0.402265903759)*x_ref[0]**o+(-0.460219787922)*x_ref[0]+(0.160493468523)*x_ref[1]**o+(-0.816631415356)*x_ref[1]
        ref[2,2]=(-0.826968582935)*x_ref[0]**o+(-0.477060281375)*x_ref[0]+(-0.753204398647)*x_ref[1]**o+(0.340944419869)*x_ref[1]
        ref[2,3]=(-0.289758403392)*x_ref[0]**o+(-0.216496476621)*x_ref[0]+(-0.0729120551729)*x_ref[1]**o+(0.0206397202194)*x_ref[1]
        ref[2,4]=(0.900190015504)*x_ref[0]**o+(0.554334407718)*x_ref[0]+(-0.188091500011)*x_ref[1]**o+(-0.239344021421)*x_ref[1]
        ref[3,0]=(-0.846253042969)*x_ref[0]**o+(0.212019924467)*x_ref[0]+(-0.209309311993)*x_ref[1]**o+(0.47889463943)*x_ref[1]
        ref[3,1]=(-0.796067196362)*x_ref[0]**o+(0.570882191233)*x_ref[0]+(0.789930258)*x_ref[1]**o+(-0.687939656508)*x_ref[1]
        ref[3,2]=(0.0603011963958)*x_ref[0]**o+(-0.291794860225)*x_ref[0]+(-0.572598612637)*x_ref[1]**o+(0.308585554728)*x_ref[1]
        ref[3,3]=(0.883976313141)*x_ref[0]**o+(0.857027876956)*x_ref[0]+(-0.991113135222)*x_ref[1]**o+(-0.0649848727938)*x_ref[1]
        ref[3,4]=(-0.420604320052)*x_ref[0]**o+(-0.147041915772)*x_ref[0]+(0.696294525095)*x_ref[1]**o+(-0.45791291995)*x_ref[1]
      else:
        arg[0,0]=(0.650989082369)*x[0]**o+(-0.291482513935)*x[0]+(-0.791203000903)*x[1]**o+(0.297385821221)*x[1]+(-0.169814832171)*x[2]**o+(-0.125360282254)*x[2]
        arg[0,1]=(0.166614460758)*x[0]**o+(0.0840046062434)*x[0]+(-0.114646748006)*x[1]**o+(-0.872705089326)*x[1]+(-0.687707782796)*x[2]**o+(-0.230067474846)*x[2]
        arg[0,2]=(-0.219927466535)*x[0]**o+(-0.466007112543)*x[0]+(-0.255341401047)*x[1]**o+(0.615864283673)*x[1]+(-0.983984041665)*x[2]**o+(0.305409617185)*x[2]
        arg[0,3]=(-0.0879672823913)*x[0]**o+(-0.956382814278)*x[0]+(-0.3091322614)*x[1]**o+(0.147177613425)*x[1]+(-0.98911137256)*x[2]**o+(-0.750791857003)*x[2]
        arg[0,4]=(0.715277956634)*x[0]**o+(-0.73358535985)*x[0]+(0.14587372226)*x[1]**o+(0.411467948058)*x[1]+(0.925673970492)*x[2]**o+(-0.0678728528684)*x[2]
        arg[1,0]=(0.150067062674)*x[0]**o+(-0.522799232667)*x[0]+(-0.256261929148)*x[1]**o+(-0.316346866901)*x[1]+(-0.849523865324)*x[2]**o+(-0.585421363115)*x[2]
        arg[1,1]=(-0.49684130662)*x[0]**o+(0.868986326423)*x[0]+(0.650641361203)*x[1]**o+(0.558696373611)*x[1]+(0.945790183046)*x[2]**o+(0.379250336542)*x[2]
        arg[1,2]=(-0.67349806443)*x[0]**o+(-0.337586770393)*x[0]+(0.719579305241)*x[1]**o+(-0.300711356473)*x[1]+(-0.350130081791)*x[2]**o+(0.604354519926)*x[2]
        arg[1,3]=(-0.242995155153)*x[0]**o+(0.668842006698)*x[0]+(-0.981902121891)*x[1]**o+(-0.358779020262)*x[1]+(0.298801927955)*x[2]**o+(-0.859807798836)*x[2]
        arg[1,4]=(-0.248474519407)*x[0]**o+(-0.309793754496)*x[0]+(-0.967889559474)*x[1]**o+(0.541761608865)*x[1]+(0.573726565934)*x[2]**o+(-0.876387334383)*x[2]
        arg[2,0]=(0.879662395476)*x[0]**o+(0.172656747739)*x[0]+(0.74129137412)*x[1]**o+(0.236748254611)*x[1]+(0.764805785803)*x[2]**o+(-0.257470519211)*x[2]
        arg[2,1]=(0.44326363254)*x[0]**o+(0.121548286191)*x[0]+(0.605969252858)*x[1]**o+(0.261621973968)*x[1]+(-0.694403080782)*x[2]**o+(0.15039482013)*x[2]
        arg[2,2]=(-0.372758098923)*x[0]**o+(0.577481130716)*x[0]+(-0.120284392247)*x[1]**o+(-0.630815000386)*x[1]+(-0.601595664522)*x[2]**o+(0.806894830983)*x[2]
        arg[2,3]=(0.548413470095)*x[0]**o+(0.804535150265)*x[0]+(0.469766192947)*x[1]**o+(-0.796431483578)*x[1]+(0.434732585165)*x[2]**o+(-0.20040808975)*x[2]
        arg[2,4]=(-0.119065000638)*x[0]**o+(-0.651948510575)*x[0]+(0.868418571224)*x[1]**o+(-0.987758488584)*x[1]+(0.495264099949)*x[2]**o+(-0.928684424484)*x[2]
        arg[3,0]=(-0.712239839816)*x[0]**o+(0.90734601523)*x[0]+(0.901855037856)*x[1]**o+(0.773630983477)*x[1]+(0.599873877074)*x[2]**o+(-0.511864534352)*x[2]
        arg[3,1]=(0.0499812109271)*x[0]**o+(0.959791265887)*x[0]+(0.861696669737)*x[1]**o+(-0.887520952646)*x[1]+(-0.609317719697)*x[2]**o+(-0.314327480839)*x[2]
        arg[3,2]=(0.717930038261)*x[0]**o+(-0.480998377737)*x[0]+(0.890930499387)*x[1]**o+(-0.880493180277)*x[1]+(0.540792831261)*x[2]**o+(0.343985055993)*x[2]
        arg[3,3]=(0.20531032325)*x[0]**o+(0.0400856366105)*x[0]+(0.829583965232)*x[1]**o+(0.701555952938)*x[1]+(0.676177356543)*x[2]**o+(0.934250384088)*x[2]
        arg[3,4]=(-0.426083844591)*x[0]**o+(0.0732912754379)*x[0]+(0.555586413705)*x[1]**o+(0.379137165516)*x[1]+(-0.481638055369)*x[2]**o+(-0.156787271652)*x[2]
        ref[0,0]=(0.650989082369)*x_ref[0]**o+(-0.291482513935)*x_ref[0]+(-0.791203000903)*x_ref[1]**o+(0.297385821221)*x_ref[1]+(-0.169814832171)*x_ref[2]**o+(-0.125360282254)*x_ref[2]
        ref[0,1]=(0.166614460758)*x_ref[0]**o+(0.0840046062434)*x_ref[0]+(-0.114646748006)*x_ref[1]**o+(-0.872705089326)*x_ref[1]+(-0.687707782796)*x_ref[2]**o+(-0.230067474846)*x_ref[2]
        ref[0,2]=(-0.219927466535)*x_ref[0]**o+(-0.466007112543)*x_ref[0]+(-0.255341401047)*x_ref[1]**o+(0.615864283673)*x_ref[1]+(-0.983984041665)*x_ref[2]**o+(0.305409617185)*x_ref[2]
        ref[0,3]=(-0.0879672823913)*x_ref[0]**o+(-0.956382814278)*x_ref[0]+(-0.3091322614)*x_ref[1]**o+(0.147177613425)*x_ref[1]+(-0.98911137256)*x_ref[2]**o+(-0.750791857003)*x_ref[2]
        ref[0,4]=(0.715277956634)*x_ref[0]**o+(-0.73358535985)*x_ref[0]+(0.14587372226)*x_ref[1]**o+(0.411467948058)*x_ref[1]+(0.925673970492)*x_ref[2]**o+(-0.0678728528684)*x_ref[2]
        ref[1,0]=(0.150067062674)*x_ref[0]**o+(-0.522799232667)*x_ref[0]+(-0.256261929148)*x_ref[1]**o+(-0.316346866901)*x_ref[1]+(-0.849523865324)*x_ref[2]**o+(-0.585421363115)*x_ref[2]
        ref[1,1]=(-0.49684130662)*x_ref[0]**o+(0.868986326423)*x_ref[0]+(0.650641361203)*x_ref[1]**o+(0.558696373611)*x_ref[1]+(0.945790183046)*x_ref[2]**o+(0.379250336542)*x_ref[2]
        ref[1,2]=(-0.67349806443)*x_ref[0]**o+(-0.337586770393)*x_ref[0]+(0.719579305241)*x_ref[1]**o+(-0.300711356473)*x_ref[1]+(-0.350130081791)*x_ref[2]**o+(0.604354519926)*x_ref[2]
        ref[1,3]=(-0.242995155153)*x_ref[0]**o+(0.668842006698)*x_ref[0]+(-0.981902121891)*x_ref[1]**o+(-0.358779020262)*x_ref[1]+(0.298801927955)*x_ref[2]**o+(-0.859807798836)*x_ref[2]
        ref[1,4]=(-0.248474519407)*x_ref[0]**o+(-0.309793754496)*x_ref[0]+(-0.967889559474)*x_ref[1]**o+(0.541761608865)*x_ref[1]+(0.573726565934)*x_ref[2]**o+(-0.876387334383)*x_ref[2]
        ref[2,0]=(0.879662395476)*x_ref[0]**o+(0.172656747739)*x_ref[0]+(0.74129137412)*x_ref[1]**o+(0.236748254611)*x_ref[1]+(0.764805785803)*x_ref[2]**o+(-0.257470519211)*x_ref[2]
        ref[2,1]=(0.44326363254)*x_ref[0]**o+(0.121548286191)*x_ref[0]+(0.605969252858)*x_ref[1]**o+(0.261621973968)*x_ref[1]+(-0.694403080782)*x_ref[2]**o+(0.15039482013)*x_ref[2]
        ref[2,2]=(-0.372758098923)*x_ref[0]**o+(0.577481130716)*x_ref[0]+(-0.120284392247)*x_ref[1]**o+(-0.630815000386)*x_ref[1]+(-0.601595664522)*x_ref[2]**o+(0.806894830983)*x_ref[2]
        ref[2,3]=(0.548413470095)*x_ref[0]**o+(0.804535150265)*x_ref[0]+(0.469766192947)*x_ref[1]**o+(-0.796431483578)*x_ref[1]+(0.434732585165)*x_ref[2]**o+(-0.20040808975)*x_ref[2]
        ref[2,4]=(-0.119065000638)*x_ref[0]**o+(-0.651948510575)*x_ref[0]+(0.868418571224)*x_ref[1]**o+(-0.987758488584)*x_ref[1]+(0.495264099949)*x_ref[2]**o+(-0.928684424484)*x_ref[2]
        ref[3,0]=(-0.712239839816)*x_ref[0]**o+(0.90734601523)*x_ref[0]+(0.901855037856)*x_ref[1]**o+(0.773630983477)*x_ref[1]+(0.599873877074)*x_ref[2]**o+(-0.511864534352)*x_ref[2]
        ref[3,1]=(0.0499812109271)*x_ref[0]**o+(0.959791265887)*x_ref[0]+(0.861696669737)*x_ref[1]**o+(-0.887520952646)*x_ref[1]+(-0.609317719697)*x_ref[2]**o+(-0.314327480839)*x_ref[2]
        ref[3,2]=(0.717930038261)*x_ref[0]**o+(-0.480998377737)*x_ref[0]+(0.890930499387)*x_ref[1]**o+(-0.880493180277)*x_ref[1]+(0.540792831261)*x_ref[2]**o+(0.343985055993)*x_ref[2]
        ref[3,3]=(0.20531032325)*x_ref[0]**o+(0.0400856366105)*x_ref[0]+(0.829583965232)*x_ref[1]**o+(0.701555952938)*x_ref[1]+(0.676177356543)*x_ref[2]**o+(0.934250384088)*x_ref[2]
        ref[3,4]=(-0.426083844591)*x_ref[0]**o+(0.0732912754379)*x_ref[0]+(0.555586413705)*x_ref[1]**o+(0.379137165516)*x_ref[1]+(-0.481638055369)*x_ref[2]**o+(-0.156787271652)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.0451039988637)*x[0]**o+(0.159426346984)*x[0]+(0.693271264689)*x[1]**o+(-0.129057826038)*x[1]
        arg[0,0,1]=(0.165096887432)*x[0]**o+(0.421267288474)*x[0]+(0.259248346979)*x[1]**o+(-0.216110688389)*x[1]
        arg[0,1,0]=(-0.419500329289)*x[0]**o+(-0.516326782996)*x[0]+(-0.726209291778)*x[1]**o+(0.80434579811)*x[1]
        arg[0,1,1]=(-0.680429758627)*x[0]**o+(-0.539365634112)*x[0]+(-0.20379891956)*x[1]**o+(-0.760436888151)*x[1]
        arg[1,0,0]=(-0.472638888322)*x[0]**o+(0.982469790589)*x[0]+(-0.383590346736)*x[1]**o+(-0.69130786846)*x[1]
        arg[1,0,1]=(0.334498522873)*x[0]**o+(0.190404359107)*x[0]+(0.210624993429)*x[1]**o+(-0.744921567562)*x[1]
        arg[1,1,0]=(0.632591775382)*x[0]**o+(-0.405215541727)*x[0]+(-0.880471123936)*x[1]**o+(0.0942729925938)*x[1]
        arg[1,1,1]=(-0.837885933046)*x[0]**o+(-0.798882152889)*x[0]+(0.648322104814)*x[1]**o+(-0.357620132856)*x[1]
        arg[2,0,0]=(-0.490592533882)*x[0]**o+(0.540431305386)*x[0]+(-0.777299904284)*x[1]**o+(-0.308682146808)*x[1]
        arg[2,0,1]=(0.971265898941)*x[0]**o+(0.546875456819)*x[0]+(0.627171799906)*x[1]**o+(0.0732789392684)*x[1]
        arg[2,1,0]=(0.178076638358)*x[0]**o+(0.587800697151)*x[0]+(0.748336208169)*x[1]**o+(-0.0341258196613)*x[1]
        arg[2,1,1]=(0.36345594982)*x[0]**o+(-0.899402201831)*x[0]+(0.0206286665096)*x[1]**o+(0.303557722727)*x[1]
        arg[3,0,0]=(-0.544119125777)*x[0]**o+(-0.805891689123)*x[0]+(-0.687902925271)*x[1]**o+(0.346236636736)*x[1]
        arg[3,0,1]=(0.725902084271)*x[0]**o+(0.605461404125)*x[0]+(0.452767036006)*x[1]**o+(0.266584718883)*x[1]
        arg[3,1,0]=(0.322052654392)*x[0]**o+(-0.91949636775)*x[0]+(0.948097272559)*x[1]**o+(-0.463715765101)*x[1]
        arg[3,1,1]=(-0.369192142232)*x[0]**o+(0.616001201865)*x[0]+(-0.257989673651)*x[1]**o+(0.507550625249)*x[1]
        arg[4,0,0]=(0.270343836597)*x[0]**o+(0.851686984957)*x[0]+(0.838756000051)*x[1]**o+(-0.329625536762)*x[1]
        arg[4,0,1]=(0.00026238459417)*x[0]**o+(0.0274581707442)*x[0]+(0.770120998708)*x[1]**o+(0.574619799143)*x[1]
        arg[4,1,0]=(-0.709788768284)*x[0]**o+(0.0856634157603)*x[0]+(-0.294662391976)*x[1]**o+(-0.90639308515)*x[1]
        arg[4,1,1]=(0.223381811001)*x[0]**o+(0.941451590223)*x[0]+(0.917865807171)*x[1]**o+(0.912972271206)*x[1]
        arg[5,0,0]=(-0.731163054411)*x[0]**o+(0.185548632151)*x[0]+(0.907527980776)*x[1]**o+(0.816017248525)*x[1]
        arg[5,0,1]=(0.719177781335)*x[0]**o+(-0.788006096642)*x[0]+(-0.434802402615)*x[1]**o+(0.0249465251663)*x[1]
        arg[5,1,0]=(0.982421863977)*x[0]**o+(0.127223012805)*x[0]+(-0.858491508817)*x[1]**o+(-0.152253965468)*x[1]
        arg[5,1,1]=(-0.779596616421)*x[0]**o+(0.049815217296)*x[0]+(0.892121656728)*x[1]**o+(0.258587664554)*x[1]
        ref[0,0,0]=(0.0451039988637)*x_ref[0]**o+(0.159426346984)*x_ref[0]+(0.693271264689)*x_ref[1]**o+(-0.129057826038)*x_ref[1]
        ref[0,0,1]=(0.165096887432)*x_ref[0]**o+(0.421267288474)*x_ref[0]+(0.259248346979)*x_ref[1]**o+(-0.216110688389)*x_ref[1]
        ref[0,1,0]=(-0.419500329289)*x_ref[0]**o+(-0.516326782996)*x_ref[0]+(-0.726209291778)*x_ref[1]**o+(0.80434579811)*x_ref[1]
        ref[0,1,1]=(-0.680429758627)*x_ref[0]**o+(-0.539365634112)*x_ref[0]+(-0.20379891956)*x_ref[1]**o+(-0.760436888151)*x_ref[1]
        ref[1,0,0]=(-0.472638888322)*x_ref[0]**o+(0.982469790589)*x_ref[0]+(-0.383590346736)*x_ref[1]**o+(-0.69130786846)*x_ref[1]
        ref[1,0,1]=(0.334498522873)*x_ref[0]**o+(0.190404359107)*x_ref[0]+(0.210624993429)*x_ref[1]**o+(-0.744921567562)*x_ref[1]
        ref[1,1,0]=(0.632591775382)*x_ref[0]**o+(-0.405215541727)*x_ref[0]+(-0.880471123936)*x_ref[1]**o+(0.0942729925938)*x_ref[1]
        ref[1,1,1]=(-0.837885933046)*x_ref[0]**o+(-0.798882152889)*x_ref[0]+(0.648322104814)*x_ref[1]**o+(-0.357620132856)*x_ref[1]
        ref[2,0,0]=(-0.490592533882)*x_ref[0]**o+(0.540431305386)*x_ref[0]+(-0.777299904284)*x_ref[1]**o+(-0.308682146808)*x_ref[1]
        ref[2,0,1]=(0.971265898941)*x_ref[0]**o+(0.546875456819)*x_ref[0]+(0.627171799906)*x_ref[1]**o+(0.0732789392684)*x_ref[1]
        ref[2,1,0]=(0.178076638358)*x_ref[0]**o+(0.587800697151)*x_ref[0]+(0.748336208169)*x_ref[1]**o+(-0.0341258196613)*x_ref[1]
        ref[2,1,1]=(0.36345594982)*x_ref[0]**o+(-0.899402201831)*x_ref[0]+(0.0206286665096)*x_ref[1]**o+(0.303557722727)*x_ref[1]
        ref[3,0,0]=(-0.544119125777)*x_ref[0]**o+(-0.805891689123)*x_ref[0]+(-0.687902925271)*x_ref[1]**o+(0.346236636736)*x_ref[1]
        ref[3,0,1]=(0.725902084271)*x_ref[0]**o+(0.605461404125)*x_ref[0]+(0.452767036006)*x_ref[1]**o+(0.266584718883)*x_ref[1]
        ref[3,1,0]=(0.322052654392)*x_ref[0]**o+(-0.91949636775)*x_ref[0]+(0.948097272559)*x_ref[1]**o+(-0.463715765101)*x_ref[1]
        ref[3,1,1]=(-0.369192142232)*x_ref[0]**o+(0.616001201865)*x_ref[0]+(-0.257989673651)*x_ref[1]**o+(0.507550625249)*x_ref[1]
        ref[4,0,0]=(0.270343836597)*x_ref[0]**o+(0.851686984957)*x_ref[0]+(0.838756000051)*x_ref[1]**o+(-0.329625536762)*x_ref[1]
        ref[4,0,1]=(0.00026238459417)*x_ref[0]**o+(0.0274581707442)*x_ref[0]+(0.770120998708)*x_ref[1]**o+(0.574619799143)*x_ref[1]
        ref[4,1,0]=(-0.709788768284)*x_ref[0]**o+(0.0856634157603)*x_ref[0]+(-0.294662391976)*x_ref[1]**o+(-0.90639308515)*x_ref[1]
        ref[4,1,1]=(0.223381811001)*x_ref[0]**o+(0.941451590223)*x_ref[0]+(0.917865807171)*x_ref[1]**o+(0.912972271206)*x_ref[1]
        ref[5,0,0]=(-0.731163054411)*x_ref[0]**o+(0.185548632151)*x_ref[0]+(0.907527980776)*x_ref[1]**o+(0.816017248525)*x_ref[1]
        ref[5,0,1]=(0.719177781335)*x_ref[0]**o+(-0.788006096642)*x_ref[0]+(-0.434802402615)*x_ref[1]**o+(0.0249465251663)*x_ref[1]
        ref[5,1,0]=(0.982421863977)*x_ref[0]**o+(0.127223012805)*x_ref[0]+(-0.858491508817)*x_ref[1]**o+(-0.152253965468)*x_ref[1]
        ref[5,1,1]=(-0.779596616421)*x_ref[0]**o+(0.049815217296)*x_ref[0]+(0.892121656728)*x_ref[1]**o+(0.258587664554)*x_ref[1]
      else:
        arg[0,0,0]=(0.804406156662)*x[0]**o+(0.304298949556)*x[0]+(-9.23949223481e-05)*x[1]**o+(0.521245437597)*x[1]+(0.822885620595)*x[2]**o+(0.44472437921)*x[2]
        arg[0,0,1]=(0.208119461011)*x[0]**o+(-0.967847500694)*x[0]+(-0.182961335156)*x[1]**o+(-0.0510538508635)*x[1]+(0.932285255322)*x[2]**o+(-0.463525963095)*x[2]
        arg[0,1,0]=(-0.0514207812926)*x[0]**o+(-0.730723672364)*x[0]+(-0.471144684451)*x[1]**o+(0.0359308791998)*x[1]+(-0.935079157697)*x[2]**o+(0.841965819816)*x[2]
        arg[0,1,1]=(0.607427052267)*x[0]**o+(-0.344216102016)*x[0]+(0.967271353509)*x[1]**o+(-0.412288382331)*x[1]+(-0.0716617724983)*x[2]**o+(0.56895080823)*x[2]
        arg[1,0,0]=(-0.574314769591)*x[0]**o+(-0.819434106932)*x[0]+(0.143229095273)*x[1]**o+(0.505343958082)*x[1]+(0.497764227469)*x[2]**o+(0.592609757968)*x[2]
        arg[1,0,1]=(0.414984900558)*x[0]**o+(0.624574053857)*x[0]+(-0.367330776954)*x[1]**o+(-0.30093924896)*x[1]+(-0.99416380749)*x[2]**o+(0.492443599924)*x[2]
        arg[1,1,0]=(0.171673119954)*x[0]**o+(0.612986140068)*x[0]+(-0.671355907363)*x[1]**o+(-0.292080539973)*x[1]+(-0.450890819125)*x[2]**o+(-0.64742047997)*x[2]
        arg[1,1,1]=(0.452494887295)*x[0]**o+(0.187308285528)*x[0]+(-0.995316825981)*x[1]**o+(-0.825194018202)*x[1]+(-0.00651224511881)*x[2]**o+(-0.715747547803)*x[2]
        arg[2,0,0]=(0.546297169466)*x[0]**o+(-0.424626884041)*x[0]+(0.327647062351)*x[1]**o+(-0.050651105527)*x[1]+(-0.667117310429)*x[2]**o+(-0.37659304263)*x[2]
        arg[2,0,1]=(-0.323595958622)*x[0]**o+(0.541718593289)*x[0]+(0.6285998056)*x[1]**o+(0.488539567593)*x[1]+(0.983086232412)*x[2]**o+(0.692821558469)*x[2]
        arg[2,1,0]=(0.16009856161)*x[0]**o+(-0.776069988482)*x[0]+(0.750717144319)*x[1]**o+(-0.244287895989)*x[1]+(0.345242766297)*x[2]**o+(0.572063651895)*x[2]
        arg[2,1,1]=(-0.34626843638)*x[0]**o+(0.197125033681)*x[0]+(-0.570015603866)*x[1]**o+(-0.633161422872)*x[1]+(-0.602998606808)*x[2]**o+(-0.674829268895)*x[2]
        arg[3,0,0]=(-0.659696747189)*x[0]**o+(-0.571128578512)*x[0]+(-0.704414117225)*x[1]**o+(-0.334924171135)*x[1]+(-0.0200665796175)*x[2]**o+(-0.660688341854)*x[2]
        arg[3,0,1]=(-0.497019087713)*x[0]**o+(0.25850813993)*x[0]+(-0.713426904693)*x[1]**o+(-0.994061006282)*x[1]+(-0.669422282701)*x[2]**o+(-0.553981025574)*x[2]
        arg[3,1,0]=(-0.151249618276)*x[0]**o+(0.29432574729)*x[0]+(0.173047919293)*x[1]**o+(0.844456199843)*x[1]+(0.845402272715)*x[2]**o+(-0.115698924822)*x[2]
        arg[3,1,1]=(0.840440627264)*x[0]**o+(0.593400475213)*x[0]+(0.974677794562)*x[1]**o+(0.938845603167)*x[1]+(-0.944730436891)*x[2]**o+(-0.657543633218)*x[2]
        arg[4,0,0]=(0.968290161629)*x[0]**o+(0.578471458995)*x[0]+(0.946558149029)*x[1]**o+(-0.232692317459)*x[1]+(-0.97762469848)*x[2]**o+(0.660519364613)*x[2]
        arg[4,0,1]=(-0.187317070761)*x[0]**o+(-0.0363170179961)*x[0]+(0.278499613269)*x[1]**o+(-0.0685831047696)*x[1]+(-0.540282212359)*x[2]**o+(0.564656560066)*x[2]
        arg[4,1,0]=(-0.211825530749)*x[0]**o+(0.225546568724)*x[0]+(-0.83241974026)*x[1]**o+(-0.503828393954)*x[1]+(-0.635937384108)*x[2]**o+(-0.513183062177)*x[2]
        arg[4,1,1]=(-0.257123983215)*x[0]**o+(-0.0700924567638)*x[0]+(0.550859393713)*x[1]**o+(0.132070212001)*x[1]+(-0.478549332853)*x[2]**o+(-0.560087260102)*x[2]
        arg[5,0,0]=(0.677259763344)*x[0]**o+(-0.213422444013)*x[0]+(-0.294112138735)*x[1]**o+(0.083135232348)*x[1]+(-0.899953896237)*x[2]**o+(-0.117143433064)*x[2]
        arg[5,0,1]=(-0.489822592306)*x[0]**o+(0.368217759904)*x[0]+(-0.927837261915)*x[1]**o+(-0.708243327427)*x[1]+(-0.132256551163)*x[2]**o+(0.316204052014)*x[2]
        arg[5,1,0]=(0.314897334392)*x[0]**o+(-0.848168680639)*x[0]+(0.92421318346)*x[1]**o+(0.995505186786)*x[1]+(0.372381949158)*x[2]**o+(-0.563415970977)*x[2]
        arg[5,1,1]=(0.745011483885)*x[0]**o+(-0.6423283108)*x[0]+(-0.119592965986)*x[1]**o+(0.241068250891)*x[1]+(0.477172198315)*x[2]**o+(-0.468580220406)*x[2]
        ref[0,0,0]=(0.804406156662)*x_ref[0]**o+(0.304298949556)*x_ref[0]+(-9.23949223481e-05)*x_ref[1]**o+(0.521245437597)*x_ref[1]+(0.822885620595)*x_ref[2]**o+(0.44472437921)*x_ref[2]
        ref[0,0,1]=(0.208119461011)*x_ref[0]**o+(-0.967847500694)*x_ref[0]+(-0.182961335156)*x_ref[1]**o+(-0.0510538508635)*x_ref[1]+(0.932285255322)*x_ref[2]**o+(-0.463525963095)*x_ref[2]
        ref[0,1,0]=(-0.0514207812926)*x_ref[0]**o+(-0.730723672364)*x_ref[0]+(-0.471144684451)*x_ref[1]**o+(0.0359308791998)*x_ref[1]+(-0.935079157697)*x_ref[2]**o+(0.841965819816)*x_ref[2]
        ref[0,1,1]=(0.607427052267)*x_ref[0]**o+(-0.344216102016)*x_ref[0]+(0.967271353509)*x_ref[1]**o+(-0.412288382331)*x_ref[1]+(-0.0716617724983)*x_ref[2]**o+(0.56895080823)*x_ref[2]
        ref[1,0,0]=(-0.574314769591)*x_ref[0]**o+(-0.819434106932)*x_ref[0]+(0.143229095273)*x_ref[1]**o+(0.505343958082)*x_ref[1]+(0.497764227469)*x_ref[2]**o+(0.592609757968)*x_ref[2]
        ref[1,0,1]=(0.414984900558)*x_ref[0]**o+(0.624574053857)*x_ref[0]+(-0.367330776954)*x_ref[1]**o+(-0.30093924896)*x_ref[1]+(-0.99416380749)*x_ref[2]**o+(0.492443599924)*x_ref[2]
        ref[1,1,0]=(0.171673119954)*x_ref[0]**o+(0.612986140068)*x_ref[0]+(-0.671355907363)*x_ref[1]**o+(-0.292080539973)*x_ref[1]+(-0.450890819125)*x_ref[2]**o+(-0.64742047997)*x_ref[2]
        ref[1,1,1]=(0.452494887295)*x_ref[0]**o+(0.187308285528)*x_ref[0]+(-0.995316825981)*x_ref[1]**o+(-0.825194018202)*x_ref[1]+(-0.00651224511881)*x_ref[2]**o+(-0.715747547803)*x_ref[2]
        ref[2,0,0]=(0.546297169466)*x_ref[0]**o+(-0.424626884041)*x_ref[0]+(0.327647062351)*x_ref[1]**o+(-0.050651105527)*x_ref[1]+(-0.667117310429)*x_ref[2]**o+(-0.37659304263)*x_ref[2]
        ref[2,0,1]=(-0.323595958622)*x_ref[0]**o+(0.541718593289)*x_ref[0]+(0.6285998056)*x_ref[1]**o+(0.488539567593)*x_ref[1]+(0.983086232412)*x_ref[2]**o+(0.692821558469)*x_ref[2]
        ref[2,1,0]=(0.16009856161)*x_ref[0]**o+(-0.776069988482)*x_ref[0]+(0.750717144319)*x_ref[1]**o+(-0.244287895989)*x_ref[1]+(0.345242766297)*x_ref[2]**o+(0.572063651895)*x_ref[2]
        ref[2,1,1]=(-0.34626843638)*x_ref[0]**o+(0.197125033681)*x_ref[0]+(-0.570015603866)*x_ref[1]**o+(-0.633161422872)*x_ref[1]+(-0.602998606808)*x_ref[2]**o+(-0.674829268895)*x_ref[2]
        ref[3,0,0]=(-0.659696747189)*x_ref[0]**o+(-0.571128578512)*x_ref[0]+(-0.704414117225)*x_ref[1]**o+(-0.334924171135)*x_ref[1]+(-0.0200665796175)*x_ref[2]**o+(-0.660688341854)*x_ref[2]
        ref[3,0,1]=(-0.497019087713)*x_ref[0]**o+(0.25850813993)*x_ref[0]+(-0.713426904693)*x_ref[1]**o+(-0.994061006282)*x_ref[1]+(-0.669422282701)*x_ref[2]**o+(-0.553981025574)*x_ref[2]
        ref[3,1,0]=(-0.151249618276)*x_ref[0]**o+(0.29432574729)*x_ref[0]+(0.173047919293)*x_ref[1]**o+(0.844456199843)*x_ref[1]+(0.845402272715)*x_ref[2]**o+(-0.115698924822)*x_ref[2]
        ref[3,1,1]=(0.840440627264)*x_ref[0]**o+(0.593400475213)*x_ref[0]+(0.974677794562)*x_ref[1]**o+(0.938845603167)*x_ref[1]+(-0.944730436891)*x_ref[2]**o+(-0.657543633218)*x_ref[2]
        ref[4,0,0]=(0.968290161629)*x_ref[0]**o+(0.578471458995)*x_ref[0]+(0.946558149029)*x_ref[1]**o+(-0.232692317459)*x_ref[1]+(-0.97762469848)*x_ref[2]**o+(0.660519364613)*x_ref[2]
        ref[4,0,1]=(-0.187317070761)*x_ref[0]**o+(-0.0363170179961)*x_ref[0]+(0.278499613269)*x_ref[1]**o+(-0.0685831047696)*x_ref[1]+(-0.540282212359)*x_ref[2]**o+(0.564656560066)*x_ref[2]
        ref[4,1,0]=(-0.211825530749)*x_ref[0]**o+(0.225546568724)*x_ref[0]+(-0.83241974026)*x_ref[1]**o+(-0.503828393954)*x_ref[1]+(-0.635937384108)*x_ref[2]**o+(-0.513183062177)*x_ref[2]
        ref[4,1,1]=(-0.257123983215)*x_ref[0]**o+(-0.0700924567638)*x_ref[0]+(0.550859393713)*x_ref[1]**o+(0.132070212001)*x_ref[1]+(-0.478549332853)*x_ref[2]**o+(-0.560087260102)*x_ref[2]
        ref[5,0,0]=(0.677259763344)*x_ref[0]**o+(-0.213422444013)*x_ref[0]+(-0.294112138735)*x_ref[1]**o+(0.083135232348)*x_ref[1]+(-0.899953896237)*x_ref[2]**o+(-0.117143433064)*x_ref[2]
        ref[5,0,1]=(-0.489822592306)*x_ref[0]**o+(0.368217759904)*x_ref[0]+(-0.927837261915)*x_ref[1]**o+(-0.708243327427)*x_ref[1]+(-0.132256551163)*x_ref[2]**o+(0.316204052014)*x_ref[2]
        ref[5,1,0]=(0.314897334392)*x_ref[0]**o+(-0.848168680639)*x_ref[0]+(0.92421318346)*x_ref[1]**o+(0.995505186786)*x_ref[1]+(0.372381949158)*x_ref[2]**o+(-0.563415970977)*x_ref[2]
        ref[5,1,1]=(0.745011483885)*x_ref[0]**o+(-0.6423283108)*x_ref[0]+(-0.119592965986)*x_ref[1]**o+(0.241068250891)*x_ref[1]+(0.477172198315)*x_ref[2]**o+(-0.468580220406)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.633576516648)*x[0]**o+(0.77399256211)*x[0]+(-0.396696417195)*x[1]**o+(-0.604572540725)*x[1]
        arg[0,0,0,1]=(0.612672991978)*x[0]**o+(-0.331012302611)*x[0]+(0.885533578235)*x[1]**o+(0.334686818058)*x[1]
        arg[0,0,1,0]=(-0.982701069025)*x[0]**o+(0.467849033642)*x[0]+(0.525392773375)*x[1]**o+(0.355622324425)*x[1]
        arg[0,0,1,1]=(-0.0243671469666)*x[0]**o+(0.790044842383)*x[0]+(0.413384172033)*x[1]**o+(0.740805135968)*x[1]
        arg[0,0,2,0]=(0.0384789957691)*x[0]**o+(-0.317009899417)*x[0]+(0.0389710369615)*x[1]**o+(0.784549852598)*x[1]
        arg[0,0,2,1]=(0.361832259252)*x[0]**o+(0.410815792944)*x[0]+(0.39496145831)*x[1]**o+(-0.820784143964)*x[1]
        arg[0,1,0,0]=(0.224261088976)*x[0]**o+(-0.824694527256)*x[0]+(0.318674472135)*x[1]**o+(0.827696354338)*x[1]
        arg[0,1,0,1]=(-0.442154901359)*x[0]**o+(0.141775320862)*x[0]+(0.688711841268)*x[1]**o+(-0.27047542755)*x[1]
        arg[0,1,1,0]=(-0.570125802019)*x[0]**o+(0.050303762895)*x[0]+(-0.972329837417)*x[1]**o+(0.143700303765)*x[1]
        arg[0,1,1,1]=(0.852888550499)*x[0]**o+(-0.51216090329)*x[0]+(-0.895663922335)*x[1]**o+(-0.183466225547)*x[1]
        arg[0,1,2,0]=(0.863840920739)*x[0]**o+(-0.247789353524)*x[0]+(0.63056436157)*x[1]**o+(0.932522589942)*x[1]
        arg[0,1,2,1]=(-0.403832121794)*x[0]**o+(-0.769625259789)*x[0]+(0.507724303191)*x[1]**o+(0.314178775526)*x[1]
        arg[0,2,0,0]=(0.543867153515)*x[0]**o+(-0.414107817028)*x[0]+(0.423922720842)*x[1]**o+(0.166632407916)*x[1]
        arg[0,2,0,1]=(0.731968209929)*x[0]**o+(-0.114742736683)*x[0]+(0.112562415021)*x[1]**o+(0.166924480325)*x[1]
        arg[0,2,1,0]=(0.919472066242)*x[0]**o+(0.560565315205)*x[0]+(0.4967650028)*x[1]**o+(0.708376671625)*x[1]
        arg[0,2,1,1]=(0.147500239739)*x[0]**o+(-0.632670389919)*x[0]+(-0.992479063437)*x[1]**o+(0.258941122962)*x[1]
        arg[0,2,2,0]=(0.474481785329)*x[0]**o+(0.592902604094)*x[0]+(0.64058685211)*x[1]**o+(-0.477843307901)*x[1]
        arg[0,2,2,1]=(0.769898355146)*x[0]**o+(0.563458279487)*x[0]+(0.876276425449)*x[1]**o+(-0.0625727749869)*x[1]
        arg[0,3,0,0]=(-0.330055037059)*x[0]**o+(-0.727656105163)*x[0]+(-0.559901684392)*x[1]**o+(-0.433646878263)*x[1]
        arg[0,3,0,1]=(-0.725314328666)*x[0]**o+(0.638326609105)*x[0]+(-0.126619368417)*x[1]**o+(-0.232194758676)*x[1]
        arg[0,3,1,0]=(0.318096005545)*x[0]**o+(0.325247494289)*x[0]+(-0.514305253643)*x[1]**o+(0.108523350086)*x[1]
        arg[0,3,1,1]=(0.298034894956)*x[0]**o+(0.825714393651)*x[0]+(-0.67162850079)*x[1]**o+(0.681809083004)*x[1]
        arg[0,3,2,0]=(-0.00715890582648)*x[0]**o+(0.893900731227)*x[0]+(-0.773710274545)*x[1]**o+(0.415913728569)*x[1]
        arg[0,3,2,1]=(0.0853551940958)*x[0]**o+(-0.228876011127)*x[0]+(0.87699456762)*x[1]**o+(-0.744605229253)*x[1]
        arg[0,4,0,0]=(0.942646937302)*x[0]**o+(0.92717236759)*x[0]+(0.794806821569)*x[1]**o+(0.848766303526)*x[1]
        arg[0,4,0,1]=(0.20140672561)*x[0]**o+(0.974490372414)*x[0]+(-0.450332419587)*x[1]**o+(-0.859102184782)*x[1]
        arg[0,4,1,0]=(0.632153053193)*x[0]**o+(0.702336756277)*x[0]+(-0.571496076236)*x[1]**o+(-0.882472852219)*x[1]
        arg[0,4,1,1]=(0.366452829499)*x[0]**o+(-0.514660164741)*x[0]+(-0.152794832253)*x[1]**o+(-0.658688117796)*x[1]
        arg[0,4,2,0]=(0.406218394702)*x[0]**o+(0.782359698198)*x[0]+(-0.0582499587562)*x[1]**o+(0.664506852595)*x[1]
        arg[0,4,2,1]=(0.768240870343)*x[0]**o+(0.188470592904)*x[0]+(0.855933509171)*x[1]**o+(0.639174087541)*x[1]
        arg[1,0,0,0]=(-0.946841398046)*x[0]**o+(0.892729256183)*x[0]+(-0.15644094115)*x[1]**o+(-0.859371368233)*x[1]
        arg[1,0,0,1]=(0.238113647666)*x[0]**o+(0.173695470467)*x[0]+(-0.753529252313)*x[1]**o+(-0.243203902463)*x[1]
        arg[1,0,1,0]=(0.267937523443)*x[0]**o+(0.600200492374)*x[0]+(0.442653529257)*x[1]**o+(-0.394984210541)*x[1]
        arg[1,0,1,1]=(-0.495022907622)*x[0]**o+(-0.593273808093)*x[0]+(0.888764746608)*x[1]**o+(-0.745127825763)*x[1]
        arg[1,0,2,0]=(0.352827238496)*x[0]**o+(0.293200928839)*x[0]+(-0.00369812109034)*x[1]**o+(0.478875774193)*x[1]
        arg[1,0,2,1]=(-0.403358878981)*x[0]**o+(0.756947364667)*x[0]+(0.649335795039)*x[1]**o+(-0.339962796859)*x[1]
        arg[1,1,0,0]=(-0.607505349015)*x[0]**o+(-0.392535399582)*x[0]+(0.799782389634)*x[1]**o+(0.186339165863)*x[1]
        arg[1,1,0,1]=(0.997492304627)*x[0]**o+(-0.281456622581)*x[0]+(-0.959294693613)*x[1]**o+(-0.454812606674)*x[1]
        arg[1,1,1,0]=(0.454682037549)*x[0]**o+(0.797129935647)*x[0]+(-0.949539573053)*x[1]**o+(-0.918499382199)*x[1]
        arg[1,1,1,1]=(0.903016449758)*x[0]**o+(-0.575039046836)*x[0]+(0.957158277074)*x[1]**o+(-0.715797460906)*x[1]
        arg[1,1,2,0]=(0.92005547899)*x[0]**o+(-0.716975919302)*x[0]+(-0.900904311708)*x[1]**o+(0.989458809643)*x[1]
        arg[1,1,2,1]=(-0.633214991087)*x[0]**o+(0.680658578627)*x[0]+(0.552194054556)*x[1]**o+(0.343557072769)*x[1]
        arg[1,2,0,0]=(-0.0182903530463)*x[0]**o+(-0.0445569497267)*x[0]+(-0.259920069365)*x[1]**o+(0.118765787279)*x[1]
        arg[1,2,0,1]=(-0.667382174733)*x[0]**o+(-0.79898800024)*x[0]+(-0.217638556567)*x[1]**o+(-0.0288504127936)*x[1]
        arg[1,2,1,0]=(-0.437982445862)*x[0]**o+(-0.104054454808)*x[0]+(0.711429718088)*x[1]**o+(-0.607226296329)*x[1]
        arg[1,2,1,1]=(0.252861218037)*x[0]**o+(-0.635701456504)*x[0]+(-0.555222304592)*x[1]**o+(0.723744010092)*x[1]
        arg[1,2,2,0]=(-0.73532710885)*x[0]**o+(0.600599078741)*x[0]+(0.0465795173194)*x[1]**o+(0.524165544645)*x[1]
        arg[1,2,2,1]=(-0.025246019744)*x[0]**o+(-0.376526675619)*x[0]+(0.213011788698)*x[1]**o+(-0.477018225777)*x[1]
        arg[1,3,0,0]=(-0.52682690957)*x[0]**o+(0.0567472858499)*x[0]+(0.574116256102)*x[1]**o+(-0.130012791898)*x[1]
        arg[1,3,0,1]=(0.113502182936)*x[0]**o+(0.7643861313)*x[0]+(0.48414948847)*x[1]**o+(0.863168250318)*x[1]
        arg[1,3,1,0]=(-0.208102162378)*x[0]**o+(0.327770728981)*x[0]+(0.843778895278)*x[1]**o+(0.474843504177)*x[1]
        arg[1,3,1,1]=(-0.549818192184)*x[0]**o+(-0.863333816685)*x[0]+(-0.365631583957)*x[1]**o+(-0.491320441182)*x[1]
        arg[1,3,2,0]=(-0.265833845759)*x[0]**o+(-0.94416746081)*x[0]+(0.48465442593)*x[1]**o+(0.258284621408)*x[1]
        arg[1,3,2,1]=(-0.625716082268)*x[0]**o+(-0.234697130359)*x[0]+(-0.973704406131)*x[1]**o+(0.00141746117847)*x[1]
        arg[1,4,0,0]=(0.565044361964)*x[0]**o+(-0.0672474991748)*x[0]+(-0.125901930524)*x[1]**o+(-0.753103899732)*x[1]
        arg[1,4,0,1]=(-0.979743467289)*x[0]**o+(-0.25225085214)*x[0]+(-0.141635164635)*x[1]**o+(0.378153168667)*x[1]
        arg[1,4,1,0]=(0.97728399206)*x[0]**o+(-0.202043160532)*x[0]+(0.997445424954)*x[1]**o+(-0.100049830154)*x[1]
        arg[1,4,1,1]=(0.364840105486)*x[0]**o+(-0.682234327437)*x[0]+(0.252584592259)*x[1]**o+(-0.0785430242425)*x[1]
        arg[1,4,2,0]=(-0.0168764722616)*x[0]**o+(-0.997460139026)*x[0]+(0.560548548912)*x[1]**o+(0.220532224022)*x[1]
        arg[1,4,2,1]=(0.307927943715)*x[0]**o+(0.496390872708)*x[0]+(0.982750813274)*x[1]**o+(0.392968327778)*x[1]
        arg[2,0,0,0]=(0.645991188948)*x[0]**o+(0.104035822886)*x[0]+(0.18285829097)*x[1]**o+(0.51186286294)*x[1]
        arg[2,0,0,1]=(-0.996971717318)*x[0]**o+(0.126488047141)*x[0]+(0.123585312562)*x[1]**o+(-0.775854123529)*x[1]
        arg[2,0,1,0]=(-0.0258590124655)*x[0]**o+(0.0079454735671)*x[0]+(-0.42819024422)*x[1]**o+(0.12377360967)*x[1]
        arg[2,0,1,1]=(0.990441351225)*x[0]**o+(-0.453806376472)*x[0]+(0.134316618645)*x[1]**o+(-0.902666953282)*x[1]
        arg[2,0,2,0]=(0.149304385108)*x[0]**o+(-0.226132046999)*x[0]+(0.647186106203)*x[1]**o+(-0.269247029225)*x[1]
        arg[2,0,2,1]=(-0.159068164422)*x[0]**o+(-0.392275577366)*x[0]+(0.131423093404)*x[1]**o+(-0.353311282658)*x[1]
        arg[2,1,0,0]=(-0.2862918681)*x[0]**o+(-0.392264098939)*x[0]+(-0.344033237285)*x[1]**o+(0.644261713963)*x[1]
        arg[2,1,0,1]=(0.398022911658)*x[0]**o+(-0.84552546032)*x[0]+(-0.823861260943)*x[1]**o+(0.739994988755)*x[1]
        arg[2,1,1,0]=(-0.696854341234)*x[0]**o+(-0.810851183298)*x[0]+(-0.712182042896)*x[1]**o+(0.594887691783)*x[1]
        arg[2,1,1,1]=(-0.182583840117)*x[0]**o+(0.962594377088)*x[0]+(-0.0157575661453)*x[1]**o+(-0.432167995007)*x[1]
        arg[2,1,2,0]=(0.293029196216)*x[0]**o+(0.46839918212)*x[0]+(0.262375931577)*x[1]**o+(0.0348536956306)*x[1]
        arg[2,1,2,1]=(0.775632135095)*x[0]**o+(-0.874598534107)*x[0]+(0.711602976736)*x[1]**o+(0.37065543293)*x[1]
        arg[2,2,0,0]=(-0.103625214158)*x[0]**o+(0.553558117877)*x[0]+(-0.0522465477589)*x[1]**o+(-0.747139195529)*x[1]
        arg[2,2,0,1]=(0.507427663615)*x[0]**o+(-0.536361186134)*x[0]+(0.809054726193)*x[1]**o+(-0.979546347794)*x[1]
        arg[2,2,1,0]=(0.786210623833)*x[0]**o+(0.244510127091)*x[0]+(0.159799334111)*x[1]**o+(0.0499263839273)*x[1]
        arg[2,2,1,1]=(0.454205244371)*x[0]**o+(-0.650927822092)*x[0]+(-0.939252656087)*x[1]**o+(-0.894014755314)*x[1]
        arg[2,2,2,0]=(-0.829617920904)*x[0]**o+(0.764070900994)*x[0]+(0.972419443556)*x[1]**o+(-0.824913543157)*x[1]
        arg[2,2,2,1]=(0.694310748478)*x[0]**o+(-0.191413364091)*x[0]+(0.588258464472)*x[1]**o+(0.743973223755)*x[1]
        arg[2,3,0,0]=(-0.588397365184)*x[0]**o+(-0.634207034861)*x[0]+(0.0853498555546)*x[1]**o+(0.559933556865)*x[1]
        arg[2,3,0,1]=(0.626829165689)*x[0]**o+(0.942185296616)*x[0]+(0.0360770405565)*x[1]**o+(0.635966218547)*x[1]
        arg[2,3,1,0]=(-0.0103526873852)*x[0]**o+(0.546330663496)*x[0]+(0.799224103454)*x[1]**o+(0.791922324897)*x[1]
        arg[2,3,1,1]=(0.921564474495)*x[0]**o+(-0.420410835144)*x[0]+(-0.20482503843)*x[1]**o+(-0.483331643037)*x[1]
        arg[2,3,2,0]=(-0.903600870314)*x[0]**o+(0.901956716274)*x[0]+(-0.113926342023)*x[1]**o+(-0.061596184768)*x[1]
        arg[2,3,2,1]=(0.653250267262)*x[0]**o+(-0.13770933726)*x[0]+(-0.0791563636079)*x[1]**o+(-0.578313150893)*x[1]
        arg[2,4,0,0]=(0.993903405725)*x[0]**o+(0.694324226249)*x[0]+(0.203872883541)*x[1]**o+(-0.526880767264)*x[1]
        arg[2,4,0,1]=(0.0120058513647)*x[0]**o+(0.00907603337269)*x[0]+(0.0247958601677)*x[1]**o+(-0.996916120875)*x[1]
        arg[2,4,1,0]=(0.281518049478)*x[0]**o+(0.0773799323777)*x[0]+(-0.968565098773)*x[1]**o+(-0.275558076229)*x[1]
        arg[2,4,1,1]=(0.565275744757)*x[0]**o+(0.978636675869)*x[0]+(-0.352951657751)*x[1]**o+(-0.54170487914)*x[1]
        arg[2,4,2,0]=(-0.571207550678)*x[0]**o+(0.196754283318)*x[0]+(0.665558917127)*x[1]**o+(0.361915269355)*x[1]
        arg[2,4,2,1]=(0.54579191885)*x[0]**o+(0.798319371798)*x[0]+(0.961669996133)*x[1]**o+(-0.549305529342)*x[1]
        arg[3,0,0,0]=(0.264064946121)*x[0]**o+(-0.314396185593)*x[0]+(-0.017492206135)*x[1]**o+(0.190513393091)*x[1]
        arg[3,0,0,1]=(0.812213666919)*x[0]**o+(-0.757919357489)*x[0]+(0.294829087285)*x[1]**o+(0.916181101963)*x[1]
        arg[3,0,1,0]=(-0.529906091574)*x[0]**o+(0.553682716765)*x[0]+(-0.693177625995)*x[1]**o+(0.377145561683)*x[1]
        arg[3,0,1,1]=(-0.298942248169)*x[0]**o+(0.876878593705)*x[0]+(-0.609807982832)*x[1]**o+(-0.574954919539)*x[1]
        arg[3,0,2,0]=(-0.729324018245)*x[0]**o+(-0.495194565254)*x[0]+(-0.616162751356)*x[1]**o+(0.878549207376)*x[1]
        arg[3,0,2,1]=(0.625373988248)*x[0]**o+(0.106736108396)*x[0]+(-0.871225833944)*x[1]**o+(0.840614493429)*x[1]
        arg[3,1,0,0]=(-0.90880839589)*x[0]**o+(0.390298288126)*x[0]+(0.446872042362)*x[1]**o+(-0.0703294219415)*x[1]
        arg[3,1,0,1]=(-0.0706115592617)*x[0]**o+(-0.158386152093)*x[0]+(-0.687459181052)*x[1]**o+(-0.251170026598)*x[1]
        arg[3,1,1,0]=(-0.286907297081)*x[0]**o+(0.185480985579)*x[0]+(-0.49763263834)*x[1]**o+(-0.773056048642)*x[1]
        arg[3,1,1,1]=(-0.849572703562)*x[0]**o+(0.481739532516)*x[0]+(-0.227573063955)*x[1]**o+(-0.282367611381)*x[1]
        arg[3,1,2,0]=(-0.524975224603)*x[0]**o+(-0.795450756917)*x[0]+(-0.656865716441)*x[1]**o+(-0.0166902340102)*x[1]
        arg[3,1,2,1]=(-0.297441818145)*x[0]**o+(-0.726759852457)*x[0]+(0.146859888138)*x[1]**o+(-0.73690394554)*x[1]
        arg[3,2,0,0]=(-0.278794538596)*x[0]**o+(-0.463428732496)*x[0]+(-0.478500378607)*x[1]**o+(0.806643392698)*x[1]
        arg[3,2,0,1]=(0.573853299578)*x[0]**o+(-0.324830388013)*x[0]+(-0.185528736416)*x[1]**o+(0.444228024864)*x[1]
        arg[3,2,1,0]=(0.686054084195)*x[0]**o+(-0.914380567743)*x[0]+(0.751959744445)*x[1]**o+(0.531759675659)*x[1]
        arg[3,2,1,1]=(-0.876330263589)*x[0]**o+(0.566155144894)*x[0]+(-0.0233988901429)*x[1]**o+(0.42480184987)*x[1]
        arg[3,2,2,0]=(0.957053098936)*x[0]**o+(0.152292007141)*x[0]+(0.332521504492)*x[1]**o+(-0.289068173241)*x[1]
        arg[3,2,2,1]=(0.923170924493)*x[0]**o+(0.191154526504)*x[0]+(0.647848811717)*x[1]**o+(0.72509033544)*x[1]
        arg[3,3,0,0]=(0.388829993841)*x[0]**o+(0.243404004988)*x[0]+(-0.0313000323506)*x[1]**o+(0.558135056894)*x[1]
        arg[3,3,0,1]=(-0.784568790167)*x[0]**o+(0.742074534256)*x[0]+(-0.624869261724)*x[1]**o+(-0.412100194843)*x[1]
        arg[3,3,1,0]=(0.924173486785)*x[0]**o+(-0.966635363007)*x[0]+(0.395550623492)*x[1]**o+(0.469204722173)*x[1]
        arg[3,3,1,1]=(-0.842963105645)*x[0]**o+(0.685039360106)*x[0]+(0.279605123662)*x[1]**o+(-0.953092961091)*x[1]
        arg[3,3,2,0]=(0.598860382403)*x[0]**o+(-0.226413712435)*x[0]+(0.722250410625)*x[1]**o+(-0.944980469407)*x[1]
        arg[3,3,2,1]=(0.787289357927)*x[0]**o+(-0.270737247579)*x[0]+(-0.134351386793)*x[1]**o+(-0.0437387143766)*x[1]
        arg[3,4,0,0]=(-0.0601939617429)*x[0]**o+(-0.212346121247)*x[0]+(0.0137510326612)*x[1]**o+(0.749376479477)*x[1]
        arg[3,4,0,1]=(-0.515391847891)*x[0]**o+(-0.191867157458)*x[0]+(0.0883438399597)*x[1]**o+(-0.155095894625)*x[1]
        arg[3,4,1,0]=(0.172678225476)*x[0]**o+(-0.917390756059)*x[0]+(-0.498696112089)*x[1]**o+(-0.267655707135)*x[1]
        arg[3,4,1,1]=(-0.498294994476)*x[0]**o+(-0.301078252047)*x[0]+(0.100215523329)*x[1]**o+(-0.424039854141)*x[1]
        arg[3,4,2,0]=(0.904127003142)*x[0]**o+(-0.70704194874)*x[0]+(0.309638449466)*x[1]**o+(-0.290894515899)*x[1]
        arg[3,4,2,1]=(-0.924994398191)*x[0]**o+(0.624996423224)*x[0]+(0.461586866277)*x[1]**o+(-0.499709859121)*x[1]
        ref[0,0,0,0]=(-0.633576516648)*x_ref[0]**o+(0.77399256211)*x_ref[0]+(-0.396696417195)*x_ref[1]**o+(-0.604572540725)*x_ref[1]
        ref[0,0,0,1]=(0.612672991978)*x_ref[0]**o+(-0.331012302611)*x_ref[0]+(0.885533578235)*x_ref[1]**o+(0.334686818058)*x_ref[1]
        ref[0,0,1,0]=(-0.982701069025)*x_ref[0]**o+(0.467849033642)*x_ref[0]+(0.525392773375)*x_ref[1]**o+(0.355622324425)*x_ref[1]
        ref[0,0,1,1]=(-0.0243671469666)*x_ref[0]**o+(0.790044842383)*x_ref[0]+(0.413384172033)*x_ref[1]**o+(0.740805135968)*x_ref[1]
        ref[0,0,2,0]=(0.0384789957691)*x_ref[0]**o+(-0.317009899417)*x_ref[0]+(0.0389710369615)*x_ref[1]**o+(0.784549852598)*x_ref[1]
        ref[0,0,2,1]=(0.361832259252)*x_ref[0]**o+(0.410815792944)*x_ref[0]+(0.39496145831)*x_ref[1]**o+(-0.820784143964)*x_ref[1]
        ref[0,1,0,0]=(0.224261088976)*x_ref[0]**o+(-0.824694527256)*x_ref[0]+(0.318674472135)*x_ref[1]**o+(0.827696354338)*x_ref[1]
        ref[0,1,0,1]=(-0.442154901359)*x_ref[0]**o+(0.141775320862)*x_ref[0]+(0.688711841268)*x_ref[1]**o+(-0.27047542755)*x_ref[1]
        ref[0,1,1,0]=(-0.570125802019)*x_ref[0]**o+(0.050303762895)*x_ref[0]+(-0.972329837417)*x_ref[1]**o+(0.143700303765)*x_ref[1]
        ref[0,1,1,1]=(0.852888550499)*x_ref[0]**o+(-0.51216090329)*x_ref[0]+(-0.895663922335)*x_ref[1]**o+(-0.183466225547)*x_ref[1]
        ref[0,1,2,0]=(0.863840920739)*x_ref[0]**o+(-0.247789353524)*x_ref[0]+(0.63056436157)*x_ref[1]**o+(0.932522589942)*x_ref[1]
        ref[0,1,2,1]=(-0.403832121794)*x_ref[0]**o+(-0.769625259789)*x_ref[0]+(0.507724303191)*x_ref[1]**o+(0.314178775526)*x_ref[1]
        ref[0,2,0,0]=(0.543867153515)*x_ref[0]**o+(-0.414107817028)*x_ref[0]+(0.423922720842)*x_ref[1]**o+(0.166632407916)*x_ref[1]
        ref[0,2,0,1]=(0.731968209929)*x_ref[0]**o+(-0.114742736683)*x_ref[0]+(0.112562415021)*x_ref[1]**o+(0.166924480325)*x_ref[1]
        ref[0,2,1,0]=(0.919472066242)*x_ref[0]**o+(0.560565315205)*x_ref[0]+(0.4967650028)*x_ref[1]**o+(0.708376671625)*x_ref[1]
        ref[0,2,1,1]=(0.147500239739)*x_ref[0]**o+(-0.632670389919)*x_ref[0]+(-0.992479063437)*x_ref[1]**o+(0.258941122962)*x_ref[1]
        ref[0,2,2,0]=(0.474481785329)*x_ref[0]**o+(0.592902604094)*x_ref[0]+(0.64058685211)*x_ref[1]**o+(-0.477843307901)*x_ref[1]
        ref[0,2,2,1]=(0.769898355146)*x_ref[0]**o+(0.563458279487)*x_ref[0]+(0.876276425449)*x_ref[1]**o+(-0.0625727749869)*x_ref[1]
        ref[0,3,0,0]=(-0.330055037059)*x_ref[0]**o+(-0.727656105163)*x_ref[0]+(-0.559901684392)*x_ref[1]**o+(-0.433646878263)*x_ref[1]
        ref[0,3,0,1]=(-0.725314328666)*x_ref[0]**o+(0.638326609105)*x_ref[0]+(-0.126619368417)*x_ref[1]**o+(-0.232194758676)*x_ref[1]
        ref[0,3,1,0]=(0.318096005545)*x_ref[0]**o+(0.325247494289)*x_ref[0]+(-0.514305253643)*x_ref[1]**o+(0.108523350086)*x_ref[1]
        ref[0,3,1,1]=(0.298034894956)*x_ref[0]**o+(0.825714393651)*x_ref[0]+(-0.67162850079)*x_ref[1]**o+(0.681809083004)*x_ref[1]
        ref[0,3,2,0]=(-0.00715890582648)*x_ref[0]**o+(0.893900731227)*x_ref[0]+(-0.773710274545)*x_ref[1]**o+(0.415913728569)*x_ref[1]
        ref[0,3,2,1]=(0.0853551940958)*x_ref[0]**o+(-0.228876011127)*x_ref[0]+(0.87699456762)*x_ref[1]**o+(-0.744605229253)*x_ref[1]
        ref[0,4,0,0]=(0.942646937302)*x_ref[0]**o+(0.92717236759)*x_ref[0]+(0.794806821569)*x_ref[1]**o+(0.848766303526)*x_ref[1]
        ref[0,4,0,1]=(0.20140672561)*x_ref[0]**o+(0.974490372414)*x_ref[0]+(-0.450332419587)*x_ref[1]**o+(-0.859102184782)*x_ref[1]
        ref[0,4,1,0]=(0.632153053193)*x_ref[0]**o+(0.702336756277)*x_ref[0]+(-0.571496076236)*x_ref[1]**o+(-0.882472852219)*x_ref[1]
        ref[0,4,1,1]=(0.366452829499)*x_ref[0]**o+(-0.514660164741)*x_ref[0]+(-0.152794832253)*x_ref[1]**o+(-0.658688117796)*x_ref[1]
        ref[0,4,2,0]=(0.406218394702)*x_ref[0]**o+(0.782359698198)*x_ref[0]+(-0.0582499587562)*x_ref[1]**o+(0.664506852595)*x_ref[1]
        ref[0,4,2,1]=(0.768240870343)*x_ref[0]**o+(0.188470592904)*x_ref[0]+(0.855933509171)*x_ref[1]**o+(0.639174087541)*x_ref[1]
        ref[1,0,0,0]=(-0.946841398046)*x_ref[0]**o+(0.892729256183)*x_ref[0]+(-0.15644094115)*x_ref[1]**o+(-0.859371368233)*x_ref[1]
        ref[1,0,0,1]=(0.238113647666)*x_ref[0]**o+(0.173695470467)*x_ref[0]+(-0.753529252313)*x_ref[1]**o+(-0.243203902463)*x_ref[1]
        ref[1,0,1,0]=(0.267937523443)*x_ref[0]**o+(0.600200492374)*x_ref[0]+(0.442653529257)*x_ref[1]**o+(-0.394984210541)*x_ref[1]
        ref[1,0,1,1]=(-0.495022907622)*x_ref[0]**o+(-0.593273808093)*x_ref[0]+(0.888764746608)*x_ref[1]**o+(-0.745127825763)*x_ref[1]
        ref[1,0,2,0]=(0.352827238496)*x_ref[0]**o+(0.293200928839)*x_ref[0]+(-0.00369812109034)*x_ref[1]**o+(0.478875774193)*x_ref[1]
        ref[1,0,2,1]=(-0.403358878981)*x_ref[0]**o+(0.756947364667)*x_ref[0]+(0.649335795039)*x_ref[1]**o+(-0.339962796859)*x_ref[1]
        ref[1,1,0,0]=(-0.607505349015)*x_ref[0]**o+(-0.392535399582)*x_ref[0]+(0.799782389634)*x_ref[1]**o+(0.186339165863)*x_ref[1]
        ref[1,1,0,1]=(0.997492304627)*x_ref[0]**o+(-0.281456622581)*x_ref[0]+(-0.959294693613)*x_ref[1]**o+(-0.454812606674)*x_ref[1]
        ref[1,1,1,0]=(0.454682037549)*x_ref[0]**o+(0.797129935647)*x_ref[0]+(-0.949539573053)*x_ref[1]**o+(-0.918499382199)*x_ref[1]
        ref[1,1,1,1]=(0.903016449758)*x_ref[0]**o+(-0.575039046836)*x_ref[0]+(0.957158277074)*x_ref[1]**o+(-0.715797460906)*x_ref[1]
        ref[1,1,2,0]=(0.92005547899)*x_ref[0]**o+(-0.716975919302)*x_ref[0]+(-0.900904311708)*x_ref[1]**o+(0.989458809643)*x_ref[1]
        ref[1,1,2,1]=(-0.633214991087)*x_ref[0]**o+(0.680658578627)*x_ref[0]+(0.552194054556)*x_ref[1]**o+(0.343557072769)*x_ref[1]
        ref[1,2,0,0]=(-0.0182903530463)*x_ref[0]**o+(-0.0445569497267)*x_ref[0]+(-0.259920069365)*x_ref[1]**o+(0.118765787279)*x_ref[1]
        ref[1,2,0,1]=(-0.667382174733)*x_ref[0]**o+(-0.79898800024)*x_ref[0]+(-0.217638556567)*x_ref[1]**o+(-0.0288504127936)*x_ref[1]
        ref[1,2,1,0]=(-0.437982445862)*x_ref[0]**o+(-0.104054454808)*x_ref[0]+(0.711429718088)*x_ref[1]**o+(-0.607226296329)*x_ref[1]
        ref[1,2,1,1]=(0.252861218037)*x_ref[0]**o+(-0.635701456504)*x_ref[0]+(-0.555222304592)*x_ref[1]**o+(0.723744010092)*x_ref[1]
        ref[1,2,2,0]=(-0.73532710885)*x_ref[0]**o+(0.600599078741)*x_ref[0]+(0.0465795173194)*x_ref[1]**o+(0.524165544645)*x_ref[1]
        ref[1,2,2,1]=(-0.025246019744)*x_ref[0]**o+(-0.376526675619)*x_ref[0]+(0.213011788698)*x_ref[1]**o+(-0.477018225777)*x_ref[1]
        ref[1,3,0,0]=(-0.52682690957)*x_ref[0]**o+(0.0567472858499)*x_ref[0]+(0.574116256102)*x_ref[1]**o+(-0.130012791898)*x_ref[1]
        ref[1,3,0,1]=(0.113502182936)*x_ref[0]**o+(0.7643861313)*x_ref[0]+(0.48414948847)*x_ref[1]**o+(0.863168250318)*x_ref[1]
        ref[1,3,1,0]=(-0.208102162378)*x_ref[0]**o+(0.327770728981)*x_ref[0]+(0.843778895278)*x_ref[1]**o+(0.474843504177)*x_ref[1]
        ref[1,3,1,1]=(-0.549818192184)*x_ref[0]**o+(-0.863333816685)*x_ref[0]+(-0.365631583957)*x_ref[1]**o+(-0.491320441182)*x_ref[1]
        ref[1,3,2,0]=(-0.265833845759)*x_ref[0]**o+(-0.94416746081)*x_ref[0]+(0.48465442593)*x_ref[1]**o+(0.258284621408)*x_ref[1]
        ref[1,3,2,1]=(-0.625716082268)*x_ref[0]**o+(-0.234697130359)*x_ref[0]+(-0.973704406131)*x_ref[1]**o+(0.00141746117847)*x_ref[1]
        ref[1,4,0,0]=(0.565044361964)*x_ref[0]**o+(-0.0672474991748)*x_ref[0]+(-0.125901930524)*x_ref[1]**o+(-0.753103899732)*x_ref[1]
        ref[1,4,0,1]=(-0.979743467289)*x_ref[0]**o+(-0.25225085214)*x_ref[0]+(-0.141635164635)*x_ref[1]**o+(0.378153168667)*x_ref[1]
        ref[1,4,1,0]=(0.97728399206)*x_ref[0]**o+(-0.202043160532)*x_ref[0]+(0.997445424954)*x_ref[1]**o+(-0.100049830154)*x_ref[1]
        ref[1,4,1,1]=(0.364840105486)*x_ref[0]**o+(-0.682234327437)*x_ref[0]+(0.252584592259)*x_ref[1]**o+(-0.0785430242425)*x_ref[1]
        ref[1,4,2,0]=(-0.0168764722616)*x_ref[0]**o+(-0.997460139026)*x_ref[0]+(0.560548548912)*x_ref[1]**o+(0.220532224022)*x_ref[1]
        ref[1,4,2,1]=(0.307927943715)*x_ref[0]**o+(0.496390872708)*x_ref[0]+(0.982750813274)*x_ref[1]**o+(0.392968327778)*x_ref[1]
        ref[2,0,0,0]=(0.645991188948)*x_ref[0]**o+(0.104035822886)*x_ref[0]+(0.18285829097)*x_ref[1]**o+(0.51186286294)*x_ref[1]
        ref[2,0,0,1]=(-0.996971717318)*x_ref[0]**o+(0.126488047141)*x_ref[0]+(0.123585312562)*x_ref[1]**o+(-0.775854123529)*x_ref[1]
        ref[2,0,1,0]=(-0.0258590124655)*x_ref[0]**o+(0.0079454735671)*x_ref[0]+(-0.42819024422)*x_ref[1]**o+(0.12377360967)*x_ref[1]
        ref[2,0,1,1]=(0.990441351225)*x_ref[0]**o+(-0.453806376472)*x_ref[0]+(0.134316618645)*x_ref[1]**o+(-0.902666953282)*x_ref[1]
        ref[2,0,2,0]=(0.149304385108)*x_ref[0]**o+(-0.226132046999)*x_ref[0]+(0.647186106203)*x_ref[1]**o+(-0.269247029225)*x_ref[1]
        ref[2,0,2,1]=(-0.159068164422)*x_ref[0]**o+(-0.392275577366)*x_ref[0]+(0.131423093404)*x_ref[1]**o+(-0.353311282658)*x_ref[1]
        ref[2,1,0,0]=(-0.2862918681)*x_ref[0]**o+(-0.392264098939)*x_ref[0]+(-0.344033237285)*x_ref[1]**o+(0.644261713963)*x_ref[1]
        ref[2,1,0,1]=(0.398022911658)*x_ref[0]**o+(-0.84552546032)*x_ref[0]+(-0.823861260943)*x_ref[1]**o+(0.739994988755)*x_ref[1]
        ref[2,1,1,0]=(-0.696854341234)*x_ref[0]**o+(-0.810851183298)*x_ref[0]+(-0.712182042896)*x_ref[1]**o+(0.594887691783)*x_ref[1]
        ref[2,1,1,1]=(-0.182583840117)*x_ref[0]**o+(0.962594377088)*x_ref[0]+(-0.0157575661453)*x_ref[1]**o+(-0.432167995007)*x_ref[1]
        ref[2,1,2,0]=(0.293029196216)*x_ref[0]**o+(0.46839918212)*x_ref[0]+(0.262375931577)*x_ref[1]**o+(0.0348536956306)*x_ref[1]
        ref[2,1,2,1]=(0.775632135095)*x_ref[0]**o+(-0.874598534107)*x_ref[0]+(0.711602976736)*x_ref[1]**o+(0.37065543293)*x_ref[1]
        ref[2,2,0,0]=(-0.103625214158)*x_ref[0]**o+(0.553558117877)*x_ref[0]+(-0.0522465477589)*x_ref[1]**o+(-0.747139195529)*x_ref[1]
        ref[2,2,0,1]=(0.507427663615)*x_ref[0]**o+(-0.536361186134)*x_ref[0]+(0.809054726193)*x_ref[1]**o+(-0.979546347794)*x_ref[1]
        ref[2,2,1,0]=(0.786210623833)*x_ref[0]**o+(0.244510127091)*x_ref[0]+(0.159799334111)*x_ref[1]**o+(0.0499263839273)*x_ref[1]
        ref[2,2,1,1]=(0.454205244371)*x_ref[0]**o+(-0.650927822092)*x_ref[0]+(-0.939252656087)*x_ref[1]**o+(-0.894014755314)*x_ref[1]
        ref[2,2,2,0]=(-0.829617920904)*x_ref[0]**o+(0.764070900994)*x_ref[0]+(0.972419443556)*x_ref[1]**o+(-0.824913543157)*x_ref[1]
        ref[2,2,2,1]=(0.694310748478)*x_ref[0]**o+(-0.191413364091)*x_ref[0]+(0.588258464472)*x_ref[1]**o+(0.743973223755)*x_ref[1]
        ref[2,3,0,0]=(-0.588397365184)*x_ref[0]**o+(-0.634207034861)*x_ref[0]+(0.0853498555546)*x_ref[1]**o+(0.559933556865)*x_ref[1]
        ref[2,3,0,1]=(0.626829165689)*x_ref[0]**o+(0.942185296616)*x_ref[0]+(0.0360770405565)*x_ref[1]**o+(0.635966218547)*x_ref[1]
        ref[2,3,1,0]=(-0.0103526873852)*x_ref[0]**o+(0.546330663496)*x_ref[0]+(0.799224103454)*x_ref[1]**o+(0.791922324897)*x_ref[1]
        ref[2,3,1,1]=(0.921564474495)*x_ref[0]**o+(-0.420410835144)*x_ref[0]+(-0.20482503843)*x_ref[1]**o+(-0.483331643037)*x_ref[1]
        ref[2,3,2,0]=(-0.903600870314)*x_ref[0]**o+(0.901956716274)*x_ref[0]+(-0.113926342023)*x_ref[1]**o+(-0.061596184768)*x_ref[1]
        ref[2,3,2,1]=(0.653250267262)*x_ref[0]**o+(-0.13770933726)*x_ref[0]+(-0.0791563636079)*x_ref[1]**o+(-0.578313150893)*x_ref[1]
        ref[2,4,0,0]=(0.993903405725)*x_ref[0]**o+(0.694324226249)*x_ref[0]+(0.203872883541)*x_ref[1]**o+(-0.526880767264)*x_ref[1]
        ref[2,4,0,1]=(0.0120058513647)*x_ref[0]**o+(0.00907603337269)*x_ref[0]+(0.0247958601677)*x_ref[1]**o+(-0.996916120875)*x_ref[1]
        ref[2,4,1,0]=(0.281518049478)*x_ref[0]**o+(0.0773799323777)*x_ref[0]+(-0.968565098773)*x_ref[1]**o+(-0.275558076229)*x_ref[1]
        ref[2,4,1,1]=(0.565275744757)*x_ref[0]**o+(0.978636675869)*x_ref[0]+(-0.352951657751)*x_ref[1]**o+(-0.54170487914)*x_ref[1]
        ref[2,4,2,0]=(-0.571207550678)*x_ref[0]**o+(0.196754283318)*x_ref[0]+(0.665558917127)*x_ref[1]**o+(0.361915269355)*x_ref[1]
        ref[2,4,2,1]=(0.54579191885)*x_ref[0]**o+(0.798319371798)*x_ref[0]+(0.961669996133)*x_ref[1]**o+(-0.549305529342)*x_ref[1]
        ref[3,0,0,0]=(0.264064946121)*x_ref[0]**o+(-0.314396185593)*x_ref[0]+(-0.017492206135)*x_ref[1]**o+(0.190513393091)*x_ref[1]
        ref[3,0,0,1]=(0.812213666919)*x_ref[0]**o+(-0.757919357489)*x_ref[0]+(0.294829087285)*x_ref[1]**o+(0.916181101963)*x_ref[1]
        ref[3,0,1,0]=(-0.529906091574)*x_ref[0]**o+(0.553682716765)*x_ref[0]+(-0.693177625995)*x_ref[1]**o+(0.377145561683)*x_ref[1]
        ref[3,0,1,1]=(-0.298942248169)*x_ref[0]**o+(0.876878593705)*x_ref[0]+(-0.609807982832)*x_ref[1]**o+(-0.574954919539)*x_ref[1]
        ref[3,0,2,0]=(-0.729324018245)*x_ref[0]**o+(-0.495194565254)*x_ref[0]+(-0.616162751356)*x_ref[1]**o+(0.878549207376)*x_ref[1]
        ref[3,0,2,1]=(0.625373988248)*x_ref[0]**o+(0.106736108396)*x_ref[0]+(-0.871225833944)*x_ref[1]**o+(0.840614493429)*x_ref[1]
        ref[3,1,0,0]=(-0.90880839589)*x_ref[0]**o+(0.390298288126)*x_ref[0]+(0.446872042362)*x_ref[1]**o+(-0.0703294219415)*x_ref[1]
        ref[3,1,0,1]=(-0.0706115592617)*x_ref[0]**o+(-0.158386152093)*x_ref[0]+(-0.687459181052)*x_ref[1]**o+(-0.251170026598)*x_ref[1]
        ref[3,1,1,0]=(-0.286907297081)*x_ref[0]**o+(0.185480985579)*x_ref[0]+(-0.49763263834)*x_ref[1]**o+(-0.773056048642)*x_ref[1]
        ref[3,1,1,1]=(-0.849572703562)*x_ref[0]**o+(0.481739532516)*x_ref[0]+(-0.227573063955)*x_ref[1]**o+(-0.282367611381)*x_ref[1]
        ref[3,1,2,0]=(-0.524975224603)*x_ref[0]**o+(-0.795450756917)*x_ref[0]+(-0.656865716441)*x_ref[1]**o+(-0.0166902340102)*x_ref[1]
        ref[3,1,2,1]=(-0.297441818145)*x_ref[0]**o+(-0.726759852457)*x_ref[0]+(0.146859888138)*x_ref[1]**o+(-0.73690394554)*x_ref[1]
        ref[3,2,0,0]=(-0.278794538596)*x_ref[0]**o+(-0.463428732496)*x_ref[0]+(-0.478500378607)*x_ref[1]**o+(0.806643392698)*x_ref[1]
        ref[3,2,0,1]=(0.573853299578)*x_ref[0]**o+(-0.324830388013)*x_ref[0]+(-0.185528736416)*x_ref[1]**o+(0.444228024864)*x_ref[1]
        ref[3,2,1,0]=(0.686054084195)*x_ref[0]**o+(-0.914380567743)*x_ref[0]+(0.751959744445)*x_ref[1]**o+(0.531759675659)*x_ref[1]
        ref[3,2,1,1]=(-0.876330263589)*x_ref[0]**o+(0.566155144894)*x_ref[0]+(-0.0233988901429)*x_ref[1]**o+(0.42480184987)*x_ref[1]
        ref[3,2,2,0]=(0.957053098936)*x_ref[0]**o+(0.152292007141)*x_ref[0]+(0.332521504492)*x_ref[1]**o+(-0.289068173241)*x_ref[1]
        ref[3,2,2,1]=(0.923170924493)*x_ref[0]**o+(0.191154526504)*x_ref[0]+(0.647848811717)*x_ref[1]**o+(0.72509033544)*x_ref[1]
        ref[3,3,0,0]=(0.388829993841)*x_ref[0]**o+(0.243404004988)*x_ref[0]+(-0.0313000323506)*x_ref[1]**o+(0.558135056894)*x_ref[1]
        ref[3,3,0,1]=(-0.784568790167)*x_ref[0]**o+(0.742074534256)*x_ref[0]+(-0.624869261724)*x_ref[1]**o+(-0.412100194843)*x_ref[1]
        ref[3,3,1,0]=(0.924173486785)*x_ref[0]**o+(-0.966635363007)*x_ref[0]+(0.395550623492)*x_ref[1]**o+(0.469204722173)*x_ref[1]
        ref[3,3,1,1]=(-0.842963105645)*x_ref[0]**o+(0.685039360106)*x_ref[0]+(0.279605123662)*x_ref[1]**o+(-0.953092961091)*x_ref[1]
        ref[3,3,2,0]=(0.598860382403)*x_ref[0]**o+(-0.226413712435)*x_ref[0]+(0.722250410625)*x_ref[1]**o+(-0.944980469407)*x_ref[1]
        ref[3,3,2,1]=(0.787289357927)*x_ref[0]**o+(-0.270737247579)*x_ref[0]+(-0.134351386793)*x_ref[1]**o+(-0.0437387143766)*x_ref[1]
        ref[3,4,0,0]=(-0.0601939617429)*x_ref[0]**o+(-0.212346121247)*x_ref[0]+(0.0137510326612)*x_ref[1]**o+(0.749376479477)*x_ref[1]
        ref[3,4,0,1]=(-0.515391847891)*x_ref[0]**o+(-0.191867157458)*x_ref[0]+(0.0883438399597)*x_ref[1]**o+(-0.155095894625)*x_ref[1]
        ref[3,4,1,0]=(0.172678225476)*x_ref[0]**o+(-0.917390756059)*x_ref[0]+(-0.498696112089)*x_ref[1]**o+(-0.267655707135)*x_ref[1]
        ref[3,4,1,1]=(-0.498294994476)*x_ref[0]**o+(-0.301078252047)*x_ref[0]+(0.100215523329)*x_ref[1]**o+(-0.424039854141)*x_ref[1]
        ref[3,4,2,0]=(0.904127003142)*x_ref[0]**o+(-0.70704194874)*x_ref[0]+(0.309638449466)*x_ref[1]**o+(-0.290894515899)*x_ref[1]
        ref[3,4,2,1]=(-0.924994398191)*x_ref[0]**o+(0.624996423224)*x_ref[0]+(0.461586866277)*x_ref[1]**o+(-0.499709859121)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.0998996137438)*x[0]**o+(-0.648495171588)*x[0]+(-0.836177837758)*x[1]**o+(-0.9597381047)*x[1]+(-0.958069967907)*x[2]**o+(0.399097251753)*x[2]
        arg[0,0,0,1]=(0.625409406207)*x[0]**o+(-0.522476054218)*x[0]+(0.794724584797)*x[1]**o+(-0.450875661781)*x[1]+(-0.189505729613)*x[2]**o+(0.0910808218379)*x[2]
        arg[0,0,1,0]=(-0.36073198549)*x[0]**o+(-0.0286144078107)*x[0]+(-0.53827851785)*x[1]**o+(-0.563600303848)*x[1]+(0.970211345168)*x[2]**o+(-0.405253675856)*x[2]
        arg[0,0,1,1]=(-0.0204302558323)*x[0]**o+(0.927570053009)*x[0]+(-0.358571532503)*x[1]**o+(0.282513525404)*x[1]+(0.105530612734)*x[2]**o+(-0.156753729378)*x[2]
        arg[0,0,2,0]=(0.25049329858)*x[0]**o+(0.412299654624)*x[0]+(0.827683104949)*x[1]**o+(0.0195885473957)*x[1]+(0.849521623993)*x[2]**o+(-0.107922097347)*x[2]
        arg[0,0,2,1]=(0.824054630326)*x[0]**o+(0.315612619612)*x[0]+(-0.33528975625)*x[1]**o+(-0.848543364468)*x[1]+(0.67462279419)*x[2]**o+(-0.0387606720463)*x[2]
        arg[0,1,0,0]=(0.251613993994)*x[0]**o+(-0.871217704935)*x[0]+(-0.648182240752)*x[1]**o+(0.324536900202)*x[1]+(-0.884871901147)*x[2]**o+(0.353160185772)*x[2]
        arg[0,1,0,1]=(-0.629894976165)*x[0]**o+(0.99092287348)*x[0]+(0.87471936536)*x[1]**o+(0.732884844801)*x[1]+(0.423531105096)*x[2]**o+(0.648959408663)*x[2]
        arg[0,1,1,0]=(0.123366820062)*x[0]**o+(-0.856900246151)*x[0]+(-0.887819649211)*x[1]**o+(-0.901367472665)*x[1]+(0.222535500196)*x[2]**o+(0.42830671462)*x[2]
        arg[0,1,1,1]=(0.343789367141)*x[0]**o+(0.614284333526)*x[0]+(0.976225499577)*x[1]**o+(0.955115562801)*x[1]+(0.0700409126681)*x[2]**o+(-0.709602275548)*x[2]
        arg[0,1,2,0]=(-0.863502900367)*x[0]**o+(-0.452731447667)*x[0]+(0.495519161334)*x[1]**o+(0.0254261045314)*x[1]+(-0.484479110985)*x[2]**o+(0.615421138564)*x[2]
        arg[0,1,2,1]=(-0.861966549825)*x[0]**o+(0.957385177208)*x[0]+(0.710449966969)*x[1]**o+(-0.367149885665)*x[1]+(-0.100800220614)*x[2]**o+(-0.266781849874)*x[2]
        arg[0,2,0,0]=(-0.781309947495)*x[0]**o+(-0.709036446117)*x[0]+(0.86136976691)*x[1]**o+(-0.745743057084)*x[1]+(0.519795761984)*x[2]**o+(0.298838092751)*x[2]
        arg[0,2,0,1]=(-0.514037154361)*x[0]**o+(-0.155867632386)*x[0]+(-0.125008826668)*x[1]**o+(-0.138197902584)*x[1]+(0.360698647284)*x[2]**o+(0.432385729464)*x[2]
        arg[0,2,1,0]=(-0.19580514344)*x[0]**o+(-0.430510395896)*x[0]+(0.676498502854)*x[1]**o+(-0.472411864377)*x[1]+(-0.323468590594)*x[2]**o+(0.837971311859)*x[2]
        arg[0,2,1,1]=(-0.358714577548)*x[0]**o+(-0.583080123482)*x[0]+(-0.225887451598)*x[1]**o+(0.744101609163)*x[1]+(-0.0300539574094)*x[2]**o+(-0.985181847505)*x[2]
        arg[0,2,2,0]=(0.820610118447)*x[0]**o+(-0.560334915671)*x[0]+(0.478672329569)*x[1]**o+(-0.956892480643)*x[1]+(0.302100338619)*x[2]**o+(0.597584594886)*x[2]
        arg[0,2,2,1]=(-0.0658925658941)*x[0]**o+(-0.136738027019)*x[0]+(0.617259524472)*x[1]**o+(-0.486181324584)*x[1]+(0.225924222893)*x[2]**o+(-0.304893108343)*x[2]
        arg[0,3,0,0]=(0.0824288958004)*x[0]**o+(-0.67546530428)*x[0]+(-0.726843419073)*x[1]**o+(0.150753236521)*x[1]+(0.200621841908)*x[2]**o+(0.869623957921)*x[2]
        arg[0,3,0,1]=(0.66227806631)*x[0]**o+(-0.389229756136)*x[0]+(0.739548522805)*x[1]**o+(0.131026218939)*x[1]+(0.743271925877)*x[2]**o+(-0.910220233981)*x[2]
        arg[0,3,1,0]=(-0.259647259175)*x[0]**o+(-0.208585364012)*x[0]+(-0.323867955964)*x[1]**o+(0.778380345151)*x[1]+(-0.555780852614)*x[2]**o+(-0.16366483426)*x[2]
        arg[0,3,1,1]=(0.248968056109)*x[0]**o+(0.84999990087)*x[0]+(0.180937025847)*x[1]**o+(0.332075828156)*x[1]+(-0.00577042146131)*x[2]**o+(0.708188905312)*x[2]
        arg[0,3,2,0]=(-0.80327596428)*x[0]**o+(-0.131004563471)*x[0]+(0.443094741792)*x[1]**o+(0.515867966266)*x[1]+(-0.964003373109)*x[2]**o+(-0.580721642935)*x[2]
        arg[0,3,2,1]=(-0.107361389717)*x[0]**o+(-0.543617395264)*x[0]+(0.840867379108)*x[1]**o+(-0.459941493027)*x[1]+(-0.300033656162)*x[2]**o+(0.440763554498)*x[2]
        arg[0,4,0,0]=(0.272916362055)*x[0]**o+(-0.554210888243)*x[0]+(-0.458102512273)*x[1]**o+(-0.54594406835)*x[1]+(0.570670715852)*x[2]**o+(0.792731049718)*x[2]
        arg[0,4,0,1]=(0.896330380602)*x[0]**o+(-0.929757950797)*x[0]+(0.229049814758)*x[1]**o+(0.1245974562)*x[1]+(0.853250033203)*x[2]**o+(-0.877052530974)*x[2]
        arg[0,4,1,0]=(0.76027467594)*x[0]**o+(-0.510095378376)*x[0]+(0.697258966361)*x[1]**o+(-0.44013714543)*x[1]+(0.617675705023)*x[2]**o+(-0.00118327085347)*x[2]
        arg[0,4,1,1]=(0.92472377319)*x[0]**o+(0.27783405618)*x[0]+(-0.24317632451)*x[1]**o+(-0.0264727232824)*x[1]+(0.428751701408)*x[2]**o+(0.184142273303)*x[2]
        arg[0,4,2,0]=(0.619110983031)*x[0]**o+(-0.375037512361)*x[0]+(0.298576843317)*x[1]**o+(-0.218802335776)*x[1]+(-0.394889850336)*x[2]**o+(-0.842359750924)*x[2]
        arg[0,4,2,1]=(0.899714346356)*x[0]**o+(0.0630133184109)*x[0]+(0.0796796488932)*x[1]**o+(-0.693371400327)*x[1]+(-0.412090402419)*x[2]**o+(0.681871832039)*x[2]
        arg[1,0,0,0]=(-0.0367677770358)*x[0]**o+(-0.75352301678)*x[0]+(0.0705016755404)*x[1]**o+(0.870115630516)*x[1]+(0.136757694474)*x[2]**o+(-0.557722157484)*x[2]
        arg[1,0,0,1]=(0.550371937413)*x[0]**o+(-0.590052751632)*x[0]+(0.38665744963)*x[1]**o+(-0.896537565188)*x[1]+(0.550869514185)*x[2]**o+(0.693828237691)*x[2]
        arg[1,0,1,0]=(-0.150321579562)*x[0]**o+(0.111135028893)*x[0]+(0.582595079816)*x[1]**o+(-0.250920804447)*x[1]+(0.392998664266)*x[2]**o+(-0.226735735878)*x[2]
        arg[1,0,1,1]=(0.435196573865)*x[0]**o+(-0.31161203165)*x[0]+(-0.117580582163)*x[1]**o+(0.99583282896)*x[1]+(-0.164872214322)*x[2]**o+(0.243837126961)*x[2]
        arg[1,0,2,0]=(-0.320135099791)*x[0]**o+(0.339871290643)*x[0]+(0.593784119675)*x[1]**o+(0.695166367768)*x[1]+(0.407551427099)*x[2]**o+(-0.768445418933)*x[2]
        arg[1,0,2,1]=(-0.390718323619)*x[0]**o+(-0.0515766643887)*x[0]+(0.394112545837)*x[1]**o+(-0.379592335881)*x[1]+(-0.29680237927)*x[2]**o+(0.500583589142)*x[2]
        arg[1,1,0,0]=(-0.958930666062)*x[0]**o+(0.516874100468)*x[0]+(0.0859034666349)*x[1]**o+(-0.370561856015)*x[1]+(-0.175293417132)*x[2]**o+(0.672554615126)*x[2]
        arg[1,1,0,1]=(-0.38129381629)*x[0]**o+(-0.615636530444)*x[0]+(0.447284140424)*x[1]**o+(0.507583882952)*x[1]+(-0.942413761354)*x[2]**o+(-0.607928807162)*x[2]
        arg[1,1,1,0]=(-0.456462200294)*x[0]**o+(0.778557108965)*x[0]+(-0.939624730616)*x[1]**o+(0.965857713341)*x[1]+(-0.913772635427)*x[2]**o+(-0.459963514384)*x[2]
        arg[1,1,1,1]=(0.138896120228)*x[0]**o+(0.941491395046)*x[0]+(0.546538931051)*x[1]**o+(-0.526098363992)*x[1]+(0.32139556996)*x[2]**o+(0.899346403792)*x[2]
        arg[1,1,2,0]=(0.591597103411)*x[0]**o+(0.926278363416)*x[0]+(0.565926211046)*x[1]**o+(0.930516925784)*x[1]+(-0.498131779353)*x[2]**o+(0.358157310681)*x[2]
        arg[1,1,2,1]=(0.758395199797)*x[0]**o+(0.769578766255)*x[0]+(0.86819976051)*x[1]**o+(-0.380436310842)*x[1]+(-0.513012037048)*x[2]**o+(-0.984361866072)*x[2]
        arg[1,2,0,0]=(0.0845846076213)*x[0]**o+(0.987145343601)*x[0]+(-0.447680898266)*x[1]**o+(-0.30278197108)*x[1]+(0.271667503636)*x[2]**o+(0.698967822656)*x[2]
        arg[1,2,0,1]=(-0.314076111079)*x[0]**o+(0.507407943673)*x[0]+(-0.38467148586)*x[1]**o+(0.802764957621)*x[1]+(0.692949890402)*x[2]**o+(-0.965766183849)*x[2]
        arg[1,2,1,0]=(-0.855806343042)*x[0]**o+(0.753362713203)*x[0]+(-0.9877924012)*x[1]**o+(-0.54047626663)*x[1]+(-0.488859780266)*x[2]**o+(0.444279382381)*x[2]
        arg[1,2,1,1]=(-0.313257382113)*x[0]**o+(-0.926058887186)*x[0]+(0.509510780366)*x[1]**o+(0.738049597453)*x[1]+(0.191676625722)*x[2]**o+(-0.354148482258)*x[2]
        arg[1,2,2,0]=(-0.955204517346)*x[0]**o+(0.234220317948)*x[0]+(0.44435792198)*x[1]**o+(0.96099571541)*x[1]+(-0.422462745872)*x[2]**o+(0.808817733192)*x[2]
        arg[1,2,2,1]=(0.932454071169)*x[0]**o+(-0.449841123911)*x[0]+(-0.371587228288)*x[1]**o+(-0.552185751938)*x[1]+(-0.222070148528)*x[2]**o+(-0.911768306425)*x[2]
        arg[1,3,0,0]=(-0.355162862133)*x[0]**o+(-0.00562439573347)*x[0]+(0.982216168139)*x[1]**o+(0.0493986220989)*x[1]+(0.0242698649494)*x[2]**o+(0.593411693569)*x[2]
        arg[1,3,0,1]=(-0.982164819772)*x[0]**o+(-0.417475530445)*x[0]+(-0.421752719264)*x[1]**o+(-0.102522772133)*x[1]+(-0.384947583437)*x[2]**o+(-0.675090951882)*x[2]
        arg[1,3,1,0]=(0.862354670281)*x[0]**o+(0.0332806244503)*x[0]+(0.248152892705)*x[1]**o+(-0.380941068103)*x[1]+(0.499668747376)*x[2]**o+(-0.732344330916)*x[2]
        arg[1,3,1,1]=(-0.367669055003)*x[0]**o+(-0.28761578782)*x[0]+(0.10838982068)*x[1]**o+(-0.495999837684)*x[1]+(0.599492939122)*x[2]**o+(0.886609670648)*x[2]
        arg[1,3,2,0]=(0.707026122148)*x[0]**o+(-0.625314188271)*x[0]+(-0.734283031657)*x[1]**o+(0.779848908302)*x[1]+(-0.965798269237)*x[2]**o+(0.35160103487)*x[2]
        arg[1,3,2,1]=(0.198008431503)*x[0]**o+(0.145470568903)*x[0]+(-0.342584113223)*x[1]**o+(-0.976390968602)*x[1]+(-0.371678081685)*x[2]**o+(0.569699640624)*x[2]
        arg[1,4,0,0]=(-0.816642540092)*x[0]**o+(0.466127552066)*x[0]+(-0.491109738021)*x[1]**o+(-0.177222784312)*x[1]+(0.723624314914)*x[2]**o+(-0.103929106752)*x[2]
        arg[1,4,0,1]=(-0.461770336596)*x[0]**o+(-0.416473447722)*x[0]+(-0.844639456093)*x[1]**o+(-0.021509624606)*x[1]+(-0.507857805076)*x[2]**o+(-0.0132005113079)*x[2]
        arg[1,4,1,0]=(0.279614653607)*x[0]**o+(-0.547332204747)*x[0]+(0.329751635637)*x[1]**o+(0.813149533168)*x[1]+(0.0361758823053)*x[2]**o+(0.0635753923527)*x[2]
        arg[1,4,1,1]=(0.124940246141)*x[0]**o+(-0.470715329674)*x[0]+(0.401922174627)*x[1]**o+(-0.292371975658)*x[1]+(0.302438467505)*x[2]**o+(0.131045792461)*x[2]
        arg[1,4,2,0]=(0.817170444398)*x[0]**o+(0.153182033447)*x[0]+(-0.98908973731)*x[1]**o+(0.359609127275)*x[1]+(-0.891501663456)*x[2]**o+(0.34700998982)*x[2]
        arg[1,4,2,1]=(0.619050545861)*x[0]**o+(-0.784466574079)*x[0]+(-0.15316511355)*x[1]**o+(0.587255937366)*x[1]+(0.740644692907)*x[2]**o+(0.0651236072508)*x[2]
        arg[2,0,0,0]=(0.67904774175)*x[0]**o+(-0.24998331377)*x[0]+(-0.263635906274)*x[1]**o+(0.157008949549)*x[1]+(-0.905158896842)*x[2]**o+(-0.975062080366)*x[2]
        arg[2,0,0,1]=(0.724305963876)*x[0]**o+(0.666891161317)*x[0]+(0.769784501206)*x[1]**o+(-0.84353242369)*x[1]+(0.631055556571)*x[2]**o+(-0.433528873211)*x[2]
        arg[2,0,1,0]=(-0.545702871321)*x[0]**o+(-0.153233364062)*x[0]+(-0.892642075879)*x[1]**o+(-0.303965445232)*x[1]+(0.675938211305)*x[2]**o+(0.134016007164)*x[2]
        arg[2,0,1,1]=(0.612528479764)*x[0]**o+(0.444039154606)*x[0]+(-0.185281638413)*x[1]**o+(0.643315773689)*x[1]+(-0.5562387833)*x[2]**o+(-0.960249468718)*x[2]
        arg[2,0,2,0]=(-0.330135152971)*x[0]**o+(0.359951176686)*x[0]+(-0.797088515842)*x[1]**o+(-0.257909258598)*x[1]+(-0.927156080256)*x[2]**o+(-0.843405252125)*x[2]
        arg[2,0,2,1]=(0.447041226497)*x[0]**o+(0.691225329527)*x[0]+(0.462798475523)*x[1]**o+(0.157497098383)*x[1]+(-0.77785224695)*x[2]**o+(0.886430058902)*x[2]
        arg[2,1,0,0]=(0.397605326971)*x[0]**o+(0.66227843262)*x[0]+(-0.634922556623)*x[1]**o+(-0.263342020187)*x[1]+(-0.81486179124)*x[2]**o+(-0.216716259698)*x[2]
        arg[2,1,0,1]=(-0.868246547394)*x[0]**o+(-0.194983059368)*x[0]+(-0.667277750808)*x[1]**o+(0.284011270958)*x[1]+(-0.276295620425)*x[2]**o+(0.215159516398)*x[2]
        arg[2,1,1,0]=(0.990906320323)*x[0]**o+(0.526754062323)*x[0]+(-0.963398315893)*x[1]**o+(-0.72561275544)*x[1]+(0.539779065056)*x[2]**o+(0.37738462043)*x[2]
        arg[2,1,1,1]=(-0.668520028258)*x[0]**o+(-0.292883094355)*x[0]+(0.260949538172)*x[1]**o+(-0.952706524522)*x[1]+(-0.715386804875)*x[2]**o+(-0.683320136224)*x[2]
        arg[2,1,2,0]=(-0.646186301281)*x[0]**o+(-0.62747987465)*x[0]+(-0.858714171368)*x[1]**o+(0.46910094021)*x[1]+(0.468975652807)*x[2]**o+(0.242115907129)*x[2]
        arg[2,1,2,1]=(0.262205767991)*x[0]**o+(-0.460376423099)*x[0]+(0.88572453638)*x[1]**o+(0.611761462581)*x[1]+(0.763860971141)*x[2]**o+(0.108470198525)*x[2]
        arg[2,2,0,0]=(-0.0218691308052)*x[0]**o+(0.942271392012)*x[0]+(-0.778469966054)*x[1]**o+(0.0873305333765)*x[1]+(-0.300479183111)*x[2]**o+(-0.862547713455)*x[2]
        arg[2,2,0,1]=(-0.736808990311)*x[0]**o+(-0.200478457878)*x[0]+(0.272608362607)*x[1]**o+(-0.0651056524306)*x[1]+(0.616628495568)*x[2]**o+(-0.8623490351)*x[2]
        arg[2,2,1,0]=(0.931530668297)*x[0]**o+(0.39701649536)*x[0]+(0.552135074711)*x[1]**o+(-0.489731326344)*x[1]+(-0.864724191354)*x[2]**o+(0.369557062583)*x[2]
        arg[2,2,1,1]=(0.418780453395)*x[0]**o+(-0.4935990418)*x[0]+(0.34625989303)*x[1]**o+(0.979829617988)*x[1]+(0.853169987441)*x[2]**o+(-0.945958722805)*x[2]
        arg[2,2,2,0]=(0.973689940887)*x[0]**o+(-0.0585784323994)*x[0]+(-0.472419744635)*x[1]**o+(0.851814193674)*x[1]+(-0.393393340405)*x[2]**o+(-0.755376551633)*x[2]
        arg[2,2,2,1]=(-0.217049913648)*x[0]**o+(-0.262150796255)*x[0]+(-0.407805273745)*x[1]**o+(0.819646964669)*x[1]+(-0.0130316627236)*x[2]**o+(0.179034130621)*x[2]
        arg[2,3,0,0]=(-0.377708814803)*x[0]**o+(-0.943893058411)*x[0]+(-0.354341838264)*x[1]**o+(0.0696491295689)*x[1]+(-0.192838410156)*x[2]**o+(-0.22208932914)*x[2]
        arg[2,3,0,1]=(0.252429489766)*x[0]**o+(0.808052283089)*x[0]+(0.282941493712)*x[1]**o+(0.89816007357)*x[1]+(0.000979423783521)*x[2]**o+(0.645599553602)*x[2]
        arg[2,3,1,0]=(0.788089987784)*x[0]**o+(0.414254044423)*x[0]+(0.105204158639)*x[1]**o+(0.97527041289)*x[1]+(0.447463569594)*x[2]**o+(0.91465488394)*x[2]
        arg[2,3,1,1]=(0.882323493958)*x[0]**o+(-0.98135461385)*x[0]+(0.641881834427)*x[1]**o+(-0.888068999894)*x[1]+(0.705728558352)*x[2]**o+(0.295815435834)*x[2]
        arg[2,3,2,0]=(-0.570666876652)*x[0]**o+(0.52308582534)*x[0]+(-0.604213776925)*x[1]**o+(-0.568684610916)*x[1]+(-0.0378463784885)*x[2]**o+(0.557994858276)*x[2]
        arg[2,3,2,1]=(-0.497489149519)*x[0]**o+(0.739341493998)*x[0]+(-0.375869065986)*x[1]**o+(-0.895553158302)*x[1]+(0.916829505813)*x[2]**o+(-0.650522069672)*x[2]
        arg[2,4,0,0]=(0.627540843427)*x[0]**o+(-0.572125230014)*x[0]+(-0.404781893475)*x[1]**o+(-0.114896790942)*x[1]+(-0.108914764977)*x[2]**o+(0.739758413931)*x[2]
        arg[2,4,0,1]=(0.243205825969)*x[0]**o+(-0.989602234642)*x[0]+(-0.925730664677)*x[1]**o+(0.454801040316)*x[1]+(0.51758036389)*x[2]**o+(0.325160660047)*x[2]
        arg[2,4,1,0]=(0.559764013478)*x[0]**o+(2.69034311577e-05)*x[0]+(-0.691119937581)*x[1]**o+(-0.347457112767)*x[1]+(0.714683489268)*x[2]**o+(-0.589270137254)*x[2]
        arg[2,4,1,1]=(-0.657837524953)*x[0]**o+(0.071158671662)*x[0]+(-0.998677178035)*x[1]**o+(-0.544136374784)*x[1]+(0.6200901771)*x[2]**o+(-0.280453395396)*x[2]
        arg[2,4,2,0]=(0.692403743003)*x[0]**o+(-0.570086669428)*x[0]+(0.455036498939)*x[1]**o+(0.291433179453)*x[1]+(-0.203004457401)*x[2]**o+(0.0192489593981)*x[2]
        arg[2,4,2,1]=(-0.656783033826)*x[0]**o+(0.865748949755)*x[0]+(-0.297731124456)*x[1]**o+(-0.826786841513)*x[1]+(0.891374127144)*x[2]**o+(-0.736789456498)*x[2]
        arg[3,0,0,0]=(-0.963120586854)*x[0]**o+(0.466089565408)*x[0]+(-0.785137906849)*x[1]**o+(0.194880437113)*x[1]+(0.652057839639)*x[2]**o+(-0.63674893761)*x[2]
        arg[3,0,0,1]=(-0.923234625819)*x[0]**o+(-0.303847080864)*x[0]+(-0.374245465515)*x[1]**o+(0.131093877726)*x[1]+(-0.13469395748)*x[2]**o+(0.42721676585)*x[2]
        arg[3,0,1,0]=(-0.613510153481)*x[0]**o+(0.986696685318)*x[0]+(0.188583200151)*x[1]**o+(-0.895175969251)*x[1]+(-0.699510846377)*x[2]**o+(0.953157679279)*x[2]
        arg[3,0,1,1]=(-0.805809890231)*x[0]**o+(-0.786467529592)*x[0]+(-0.287240865489)*x[1]**o+(-0.78714682888)*x[1]+(-0.301034554724)*x[2]**o+(0.698100092246)*x[2]
        arg[3,0,2,0]=(-0.29320618172)*x[0]**o+(0.482436772935)*x[0]+(-0.126052929261)*x[1]**o+(-0.0638862140753)*x[1]+(0.42769411972)*x[2]**o+(0.741118745829)*x[2]
        arg[3,0,2,1]=(-0.380041427783)*x[0]**o+(-0.0347904596008)*x[0]+(-0.321599184546)*x[1]**o+(-0.061333518544)*x[1]+(-0.885193271364)*x[2]**o+(0.216506075886)*x[2]
        arg[3,1,0,0]=(-0.872457576682)*x[0]**o+(0.1368749584)*x[0]+(0.883209741333)*x[1]**o+(0.946644762296)*x[1]+(-0.941665219846)*x[2]**o+(0.683787986795)*x[2]
        arg[3,1,0,1]=(0.564843030077)*x[0]**o+(0.987198898299)*x[0]+(0.808116156661)*x[1]**o+(-0.712172259479)*x[1]+(0.942066890171)*x[2]**o+(-0.265957158917)*x[2]
        arg[3,1,1,0]=(-0.171018264631)*x[0]**o+(0.277202848962)*x[0]+(0.0726306773528)*x[1]**o+(-0.0265640725409)*x[1]+(-0.871118922929)*x[2]**o+(0.0562314512261)*x[2]
        arg[3,1,1,1]=(0.0861984079085)*x[0]**o+(-0.954818837051)*x[0]+(0.817362109646)*x[1]**o+(-0.513086270222)*x[1]+(-0.487472170741)*x[2]**o+(0.702514192773)*x[2]
        arg[3,1,2,0]=(0.272419057202)*x[0]**o+(-0.0454292509397)*x[0]+(-0.117471126603)*x[1]**o+(0.638237510553)*x[1]+(0.156858948691)*x[2]**o+(0.499459702141)*x[2]
        arg[3,1,2,1]=(-0.384525883334)*x[0]**o+(-0.795638828759)*x[0]+(-0.105795059118)*x[1]**o+(0.539873566967)*x[1]+(-0.402558868714)*x[2]**o+(-0.690578475243)*x[2]
        arg[3,2,0,0]=(0.776889517136)*x[0]**o+(0.710931840891)*x[0]+(0.335992677388)*x[1]**o+(-0.689182863932)*x[1]+(0.294149805572)*x[2]**o+(-0.144223257292)*x[2]
        arg[3,2,0,1]=(0.738703695733)*x[0]**o+(-0.523207061995)*x[0]+(0.707078352596)*x[1]**o+(-0.594373978096)*x[1]+(-0.833468457639)*x[2]**o+(-0.60181674446)*x[2]
        arg[3,2,1,0]=(-0.70243682985)*x[0]**o+(-0.557914174271)*x[0]+(-0.97153723053)*x[1]**o+(0.791692594024)*x[1]+(-0.442273808305)*x[2]**o+(-0.74069605321)*x[2]
        arg[3,2,1,1]=(0.136307039789)*x[0]**o+(0.820661597623)*x[0]+(0.379752060994)*x[1]**o+(-0.739083116742)*x[1]+(-0.964857055764)*x[2]**o+(0.523159928671)*x[2]
        arg[3,2,2,0]=(0.530883295074)*x[0]**o+(-0.859538509888)*x[0]+(-0.33917158254)*x[1]**o+(0.476510741055)*x[1]+(0.851975683024)*x[2]**o+(0.146277015139)*x[2]
        arg[3,2,2,1]=(-0.532984574212)*x[0]**o+(-0.413169979754)*x[0]+(0.0113284200501)*x[1]**o+(0.406768750567)*x[1]+(0.730968378807)*x[2]**o+(-0.74004422909)*x[2]
        arg[3,3,0,0]=(0.784793271465)*x[0]**o+(0.879569037436)*x[0]+(0.687804194006)*x[1]**o+(0.153617608524)*x[1]+(-0.410247334526)*x[2]**o+(0.147941373268)*x[2]
        arg[3,3,0,1]=(-0.896748874838)*x[0]**o+(-0.408404228864)*x[0]+(0.831967697989)*x[1]**o+(-0.290875629476)*x[1]+(-0.244101719785)*x[2]**o+(0.240453425519)*x[2]
        arg[3,3,1,0]=(-0.352293918285)*x[0]**o+(0.0341233590345)*x[0]+(-0.0165872699681)*x[1]**o+(-0.0668507705942)*x[1]+(-0.133733710061)*x[2]**o+(-0.00620237772745)*x[2]
        arg[3,3,1,1]=(0.173638606531)*x[0]**o+(0.619705705811)*x[0]+(0.279972531576)*x[1]**o+(-0.0811181576037)*x[1]+(-0.645673968637)*x[2]**o+(0.913753411758)*x[2]
        arg[3,3,2,0]=(0.0623921286142)*x[0]**o+(0.393414889204)*x[0]+(-0.300034851075)*x[1]**o+(-0.519515206457)*x[1]+(-0.0074981731879)*x[2]**o+(0.202996242895)*x[2]
        arg[3,3,2,1]=(0.583279597505)*x[0]**o+(-0.940603487209)*x[0]+(0.027966157338)*x[1]**o+(0.251466344692)*x[1]+(-0.97011033154)*x[2]**o+(0.46527056203)*x[2]
        arg[3,4,0,0]=(0.865170740218)*x[0]**o+(-0.466853211289)*x[0]+(0.0935492537278)*x[1]**o+(-0.632374632657)*x[1]+(0.673466346463)*x[2]**o+(-0.47934739873)*x[2]
        arg[3,4,0,1]=(0.0417451136027)*x[0]**o+(-0.781570733286)*x[0]+(0.547452389454)*x[1]**o+(-0.0286368211528)*x[1]+(-0.98907358808)*x[2]**o+(0.62749495057)*x[2]
        arg[3,4,1,0]=(-0.632067479664)*x[0]**o+(-0.872835386336)*x[0]+(-0.228785631078)*x[1]**o+(-0.224206127306)*x[1]+(0.769729051217)*x[2]**o+(-0.406750034352)*x[2]
        arg[3,4,1,1]=(0.216883071299)*x[0]**o+(-0.225632200121)*x[0]+(0.827186402743)*x[1]**o+(0.161488636886)*x[1]+(0.773801288836)*x[2]**o+(0.977006832166)*x[2]
        arg[3,4,2,0]=(-0.867098845884)*x[0]**o+(0.14989544838)*x[0]+(0.756720043262)*x[1]**o+(-0.496074302626)*x[1]+(-0.662787072451)*x[2]**o+(-0.142473973368)*x[2]
        arg[3,4,2,1]=(-0.294128396829)*x[0]**o+(-0.931379392091)*x[0]+(0.906552131161)*x[1]**o+(-0.422008485494)*x[1]+(0.806742455679)*x[2]**o+(0.940488803006)*x[2]
        ref[0,0,0,0]=(-0.0998996137438)*x_ref[0]**o+(-0.648495171588)*x_ref[0]+(-0.836177837758)*x_ref[1]**o+(-0.9597381047)*x_ref[1]+(-0.958069967907)*x_ref[2]**o+(0.399097251753)*x_ref[2]
        ref[0,0,0,1]=(0.625409406207)*x_ref[0]**o+(-0.522476054218)*x_ref[0]+(0.794724584797)*x_ref[1]**o+(-0.450875661781)*x_ref[1]+(-0.189505729613)*x_ref[2]**o+(0.0910808218379)*x_ref[2]
        ref[0,0,1,0]=(-0.36073198549)*x_ref[0]**o+(-0.0286144078107)*x_ref[0]+(-0.53827851785)*x_ref[1]**o+(-0.563600303848)*x_ref[1]+(0.970211345168)*x_ref[2]**o+(-0.405253675856)*x_ref[2]
        ref[0,0,1,1]=(-0.0204302558323)*x_ref[0]**o+(0.927570053009)*x_ref[0]+(-0.358571532503)*x_ref[1]**o+(0.282513525404)*x_ref[1]+(0.105530612734)*x_ref[2]**o+(-0.156753729378)*x_ref[2]
        ref[0,0,2,0]=(0.25049329858)*x_ref[0]**o+(0.412299654624)*x_ref[0]+(0.827683104949)*x_ref[1]**o+(0.0195885473957)*x_ref[1]+(0.849521623993)*x_ref[2]**o+(-0.107922097347)*x_ref[2]
        ref[0,0,2,1]=(0.824054630326)*x_ref[0]**o+(0.315612619612)*x_ref[0]+(-0.33528975625)*x_ref[1]**o+(-0.848543364468)*x_ref[1]+(0.67462279419)*x_ref[2]**o+(-0.0387606720463)*x_ref[2]
        ref[0,1,0,0]=(0.251613993994)*x_ref[0]**o+(-0.871217704935)*x_ref[0]+(-0.648182240752)*x_ref[1]**o+(0.324536900202)*x_ref[1]+(-0.884871901147)*x_ref[2]**o+(0.353160185772)*x_ref[2]
        ref[0,1,0,1]=(-0.629894976165)*x_ref[0]**o+(0.99092287348)*x_ref[0]+(0.87471936536)*x_ref[1]**o+(0.732884844801)*x_ref[1]+(0.423531105096)*x_ref[2]**o+(0.648959408663)*x_ref[2]
        ref[0,1,1,0]=(0.123366820062)*x_ref[0]**o+(-0.856900246151)*x_ref[0]+(-0.887819649211)*x_ref[1]**o+(-0.901367472665)*x_ref[1]+(0.222535500196)*x_ref[2]**o+(0.42830671462)*x_ref[2]
        ref[0,1,1,1]=(0.343789367141)*x_ref[0]**o+(0.614284333526)*x_ref[0]+(0.976225499577)*x_ref[1]**o+(0.955115562801)*x_ref[1]+(0.0700409126681)*x_ref[2]**o+(-0.709602275548)*x_ref[2]
        ref[0,1,2,0]=(-0.863502900367)*x_ref[0]**o+(-0.452731447667)*x_ref[0]+(0.495519161334)*x_ref[1]**o+(0.0254261045314)*x_ref[1]+(-0.484479110985)*x_ref[2]**o+(0.615421138564)*x_ref[2]
        ref[0,1,2,1]=(-0.861966549825)*x_ref[0]**o+(0.957385177208)*x_ref[0]+(0.710449966969)*x_ref[1]**o+(-0.367149885665)*x_ref[1]+(-0.100800220614)*x_ref[2]**o+(-0.266781849874)*x_ref[2]
        ref[0,2,0,0]=(-0.781309947495)*x_ref[0]**o+(-0.709036446117)*x_ref[0]+(0.86136976691)*x_ref[1]**o+(-0.745743057084)*x_ref[1]+(0.519795761984)*x_ref[2]**o+(0.298838092751)*x_ref[2]
        ref[0,2,0,1]=(-0.514037154361)*x_ref[0]**o+(-0.155867632386)*x_ref[0]+(-0.125008826668)*x_ref[1]**o+(-0.138197902584)*x_ref[1]+(0.360698647284)*x_ref[2]**o+(0.432385729464)*x_ref[2]
        ref[0,2,1,0]=(-0.19580514344)*x_ref[0]**o+(-0.430510395896)*x_ref[0]+(0.676498502854)*x_ref[1]**o+(-0.472411864377)*x_ref[1]+(-0.323468590594)*x_ref[2]**o+(0.837971311859)*x_ref[2]
        ref[0,2,1,1]=(-0.358714577548)*x_ref[0]**o+(-0.583080123482)*x_ref[0]+(-0.225887451598)*x_ref[1]**o+(0.744101609163)*x_ref[1]+(-0.0300539574094)*x_ref[2]**o+(-0.985181847505)*x_ref[2]
        ref[0,2,2,0]=(0.820610118447)*x_ref[0]**o+(-0.560334915671)*x_ref[0]+(0.478672329569)*x_ref[1]**o+(-0.956892480643)*x_ref[1]+(0.302100338619)*x_ref[2]**o+(0.597584594886)*x_ref[2]
        ref[0,2,2,1]=(-0.0658925658941)*x_ref[0]**o+(-0.136738027019)*x_ref[0]+(0.617259524472)*x_ref[1]**o+(-0.486181324584)*x_ref[1]+(0.225924222893)*x_ref[2]**o+(-0.304893108343)*x_ref[2]
        ref[0,3,0,0]=(0.0824288958004)*x_ref[0]**o+(-0.67546530428)*x_ref[0]+(-0.726843419073)*x_ref[1]**o+(0.150753236521)*x_ref[1]+(0.200621841908)*x_ref[2]**o+(0.869623957921)*x_ref[2]
        ref[0,3,0,1]=(0.66227806631)*x_ref[0]**o+(-0.389229756136)*x_ref[0]+(0.739548522805)*x_ref[1]**o+(0.131026218939)*x_ref[1]+(0.743271925877)*x_ref[2]**o+(-0.910220233981)*x_ref[2]
        ref[0,3,1,0]=(-0.259647259175)*x_ref[0]**o+(-0.208585364012)*x_ref[0]+(-0.323867955964)*x_ref[1]**o+(0.778380345151)*x_ref[1]+(-0.555780852614)*x_ref[2]**o+(-0.16366483426)*x_ref[2]
        ref[0,3,1,1]=(0.248968056109)*x_ref[0]**o+(0.84999990087)*x_ref[0]+(0.180937025847)*x_ref[1]**o+(0.332075828156)*x_ref[1]+(-0.00577042146131)*x_ref[2]**o+(0.708188905312)*x_ref[2]
        ref[0,3,2,0]=(-0.80327596428)*x_ref[0]**o+(-0.131004563471)*x_ref[0]+(0.443094741792)*x_ref[1]**o+(0.515867966266)*x_ref[1]+(-0.964003373109)*x_ref[2]**o+(-0.580721642935)*x_ref[2]
        ref[0,3,2,1]=(-0.107361389717)*x_ref[0]**o+(-0.543617395264)*x_ref[0]+(0.840867379108)*x_ref[1]**o+(-0.459941493027)*x_ref[1]+(-0.300033656162)*x_ref[2]**o+(0.440763554498)*x_ref[2]
        ref[0,4,0,0]=(0.272916362055)*x_ref[0]**o+(-0.554210888243)*x_ref[0]+(-0.458102512273)*x_ref[1]**o+(-0.54594406835)*x_ref[1]+(0.570670715852)*x_ref[2]**o+(0.792731049718)*x_ref[2]
        ref[0,4,0,1]=(0.896330380602)*x_ref[0]**o+(-0.929757950797)*x_ref[0]+(0.229049814758)*x_ref[1]**o+(0.1245974562)*x_ref[1]+(0.853250033203)*x_ref[2]**o+(-0.877052530974)*x_ref[2]
        ref[0,4,1,0]=(0.76027467594)*x_ref[0]**o+(-0.510095378376)*x_ref[0]+(0.697258966361)*x_ref[1]**o+(-0.44013714543)*x_ref[1]+(0.617675705023)*x_ref[2]**o+(-0.00118327085347)*x_ref[2]
        ref[0,4,1,1]=(0.92472377319)*x_ref[0]**o+(0.27783405618)*x_ref[0]+(-0.24317632451)*x_ref[1]**o+(-0.0264727232824)*x_ref[1]+(0.428751701408)*x_ref[2]**o+(0.184142273303)*x_ref[2]
        ref[0,4,2,0]=(0.619110983031)*x_ref[0]**o+(-0.375037512361)*x_ref[0]+(0.298576843317)*x_ref[1]**o+(-0.218802335776)*x_ref[1]+(-0.394889850336)*x_ref[2]**o+(-0.842359750924)*x_ref[2]
        ref[0,4,2,1]=(0.899714346356)*x_ref[0]**o+(0.0630133184109)*x_ref[0]+(0.0796796488932)*x_ref[1]**o+(-0.693371400327)*x_ref[1]+(-0.412090402419)*x_ref[2]**o+(0.681871832039)*x_ref[2]
        ref[1,0,0,0]=(-0.0367677770358)*x_ref[0]**o+(-0.75352301678)*x_ref[0]+(0.0705016755404)*x_ref[1]**o+(0.870115630516)*x_ref[1]+(0.136757694474)*x_ref[2]**o+(-0.557722157484)*x_ref[2]
        ref[1,0,0,1]=(0.550371937413)*x_ref[0]**o+(-0.590052751632)*x_ref[0]+(0.38665744963)*x_ref[1]**o+(-0.896537565188)*x_ref[1]+(0.550869514185)*x_ref[2]**o+(0.693828237691)*x_ref[2]
        ref[1,0,1,0]=(-0.150321579562)*x_ref[0]**o+(0.111135028893)*x_ref[0]+(0.582595079816)*x_ref[1]**o+(-0.250920804447)*x_ref[1]+(0.392998664266)*x_ref[2]**o+(-0.226735735878)*x_ref[2]
        ref[1,0,1,1]=(0.435196573865)*x_ref[0]**o+(-0.31161203165)*x_ref[0]+(-0.117580582163)*x_ref[1]**o+(0.99583282896)*x_ref[1]+(-0.164872214322)*x_ref[2]**o+(0.243837126961)*x_ref[2]
        ref[1,0,2,0]=(-0.320135099791)*x_ref[0]**o+(0.339871290643)*x_ref[0]+(0.593784119675)*x_ref[1]**o+(0.695166367768)*x_ref[1]+(0.407551427099)*x_ref[2]**o+(-0.768445418933)*x_ref[2]
        ref[1,0,2,1]=(-0.390718323619)*x_ref[0]**o+(-0.0515766643887)*x_ref[0]+(0.394112545837)*x_ref[1]**o+(-0.379592335881)*x_ref[1]+(-0.29680237927)*x_ref[2]**o+(0.500583589142)*x_ref[2]
        ref[1,1,0,0]=(-0.958930666062)*x_ref[0]**o+(0.516874100468)*x_ref[0]+(0.0859034666349)*x_ref[1]**o+(-0.370561856015)*x_ref[1]+(-0.175293417132)*x_ref[2]**o+(0.672554615126)*x_ref[2]
        ref[1,1,0,1]=(-0.38129381629)*x_ref[0]**o+(-0.615636530444)*x_ref[0]+(0.447284140424)*x_ref[1]**o+(0.507583882952)*x_ref[1]+(-0.942413761354)*x_ref[2]**o+(-0.607928807162)*x_ref[2]
        ref[1,1,1,0]=(-0.456462200294)*x_ref[0]**o+(0.778557108965)*x_ref[0]+(-0.939624730616)*x_ref[1]**o+(0.965857713341)*x_ref[1]+(-0.913772635427)*x_ref[2]**o+(-0.459963514384)*x_ref[2]
        ref[1,1,1,1]=(0.138896120228)*x_ref[0]**o+(0.941491395046)*x_ref[0]+(0.546538931051)*x_ref[1]**o+(-0.526098363992)*x_ref[1]+(0.32139556996)*x_ref[2]**o+(0.899346403792)*x_ref[2]
        ref[1,1,2,0]=(0.591597103411)*x_ref[0]**o+(0.926278363416)*x_ref[0]+(0.565926211046)*x_ref[1]**o+(0.930516925784)*x_ref[1]+(-0.498131779353)*x_ref[2]**o+(0.358157310681)*x_ref[2]
        ref[1,1,2,1]=(0.758395199797)*x_ref[0]**o+(0.769578766255)*x_ref[0]+(0.86819976051)*x_ref[1]**o+(-0.380436310842)*x_ref[1]+(-0.513012037048)*x_ref[2]**o+(-0.984361866072)*x_ref[2]
        ref[1,2,0,0]=(0.0845846076213)*x_ref[0]**o+(0.987145343601)*x_ref[0]+(-0.447680898266)*x_ref[1]**o+(-0.30278197108)*x_ref[1]+(0.271667503636)*x_ref[2]**o+(0.698967822656)*x_ref[2]
        ref[1,2,0,1]=(-0.314076111079)*x_ref[0]**o+(0.507407943673)*x_ref[0]+(-0.38467148586)*x_ref[1]**o+(0.802764957621)*x_ref[1]+(0.692949890402)*x_ref[2]**o+(-0.965766183849)*x_ref[2]
        ref[1,2,1,0]=(-0.855806343042)*x_ref[0]**o+(0.753362713203)*x_ref[0]+(-0.9877924012)*x_ref[1]**o+(-0.54047626663)*x_ref[1]+(-0.488859780266)*x_ref[2]**o+(0.444279382381)*x_ref[2]
        ref[1,2,1,1]=(-0.313257382113)*x_ref[0]**o+(-0.926058887186)*x_ref[0]+(0.509510780366)*x_ref[1]**o+(0.738049597453)*x_ref[1]+(0.191676625722)*x_ref[2]**o+(-0.354148482258)*x_ref[2]
        ref[1,2,2,0]=(-0.955204517346)*x_ref[0]**o+(0.234220317948)*x_ref[0]+(0.44435792198)*x_ref[1]**o+(0.96099571541)*x_ref[1]+(-0.422462745872)*x_ref[2]**o+(0.808817733192)*x_ref[2]
        ref[1,2,2,1]=(0.932454071169)*x_ref[0]**o+(-0.449841123911)*x_ref[0]+(-0.371587228288)*x_ref[1]**o+(-0.552185751938)*x_ref[1]+(-0.222070148528)*x_ref[2]**o+(-0.911768306425)*x_ref[2]
        ref[1,3,0,0]=(-0.355162862133)*x_ref[0]**o+(-0.00562439573347)*x_ref[0]+(0.982216168139)*x_ref[1]**o+(0.0493986220989)*x_ref[1]+(0.0242698649494)*x_ref[2]**o+(0.593411693569)*x_ref[2]
        ref[1,3,0,1]=(-0.982164819772)*x_ref[0]**o+(-0.417475530445)*x_ref[0]+(-0.421752719264)*x_ref[1]**o+(-0.102522772133)*x_ref[1]+(-0.384947583437)*x_ref[2]**o+(-0.675090951882)*x_ref[2]
        ref[1,3,1,0]=(0.862354670281)*x_ref[0]**o+(0.0332806244503)*x_ref[0]+(0.248152892705)*x_ref[1]**o+(-0.380941068103)*x_ref[1]+(0.499668747376)*x_ref[2]**o+(-0.732344330916)*x_ref[2]
        ref[1,3,1,1]=(-0.367669055003)*x_ref[0]**o+(-0.28761578782)*x_ref[0]+(0.10838982068)*x_ref[1]**o+(-0.495999837684)*x_ref[1]+(0.599492939122)*x_ref[2]**o+(0.886609670648)*x_ref[2]
        ref[1,3,2,0]=(0.707026122148)*x_ref[0]**o+(-0.625314188271)*x_ref[0]+(-0.734283031657)*x_ref[1]**o+(0.779848908302)*x_ref[1]+(-0.965798269237)*x_ref[2]**o+(0.35160103487)*x_ref[2]
        ref[1,3,2,1]=(0.198008431503)*x_ref[0]**o+(0.145470568903)*x_ref[0]+(-0.342584113223)*x_ref[1]**o+(-0.976390968602)*x_ref[1]+(-0.371678081685)*x_ref[2]**o+(0.569699640624)*x_ref[2]
        ref[1,4,0,0]=(-0.816642540092)*x_ref[0]**o+(0.466127552066)*x_ref[0]+(-0.491109738021)*x_ref[1]**o+(-0.177222784312)*x_ref[1]+(0.723624314914)*x_ref[2]**o+(-0.103929106752)*x_ref[2]
        ref[1,4,0,1]=(-0.461770336596)*x_ref[0]**o+(-0.416473447722)*x_ref[0]+(-0.844639456093)*x_ref[1]**o+(-0.021509624606)*x_ref[1]+(-0.507857805076)*x_ref[2]**o+(-0.0132005113079)*x_ref[2]
        ref[1,4,1,0]=(0.279614653607)*x_ref[0]**o+(-0.547332204747)*x_ref[0]+(0.329751635637)*x_ref[1]**o+(0.813149533168)*x_ref[1]+(0.0361758823053)*x_ref[2]**o+(0.0635753923527)*x_ref[2]
        ref[1,4,1,1]=(0.124940246141)*x_ref[0]**o+(-0.470715329674)*x_ref[0]+(0.401922174627)*x_ref[1]**o+(-0.292371975658)*x_ref[1]+(0.302438467505)*x_ref[2]**o+(0.131045792461)*x_ref[2]
        ref[1,4,2,0]=(0.817170444398)*x_ref[0]**o+(0.153182033447)*x_ref[0]+(-0.98908973731)*x_ref[1]**o+(0.359609127275)*x_ref[1]+(-0.891501663456)*x_ref[2]**o+(0.34700998982)*x_ref[2]
        ref[1,4,2,1]=(0.619050545861)*x_ref[0]**o+(-0.784466574079)*x_ref[0]+(-0.15316511355)*x_ref[1]**o+(0.587255937366)*x_ref[1]+(0.740644692907)*x_ref[2]**o+(0.0651236072508)*x_ref[2]
        ref[2,0,0,0]=(0.67904774175)*x_ref[0]**o+(-0.24998331377)*x_ref[0]+(-0.263635906274)*x_ref[1]**o+(0.157008949549)*x_ref[1]+(-0.905158896842)*x_ref[2]**o+(-0.975062080366)*x_ref[2]
        ref[2,0,0,1]=(0.724305963876)*x_ref[0]**o+(0.666891161317)*x_ref[0]+(0.769784501206)*x_ref[1]**o+(-0.84353242369)*x_ref[1]+(0.631055556571)*x_ref[2]**o+(-0.433528873211)*x_ref[2]
        ref[2,0,1,0]=(-0.545702871321)*x_ref[0]**o+(-0.153233364062)*x_ref[0]+(-0.892642075879)*x_ref[1]**o+(-0.303965445232)*x_ref[1]+(0.675938211305)*x_ref[2]**o+(0.134016007164)*x_ref[2]
        ref[2,0,1,1]=(0.612528479764)*x_ref[0]**o+(0.444039154606)*x_ref[0]+(-0.185281638413)*x_ref[1]**o+(0.643315773689)*x_ref[1]+(-0.5562387833)*x_ref[2]**o+(-0.960249468718)*x_ref[2]
        ref[2,0,2,0]=(-0.330135152971)*x_ref[0]**o+(0.359951176686)*x_ref[0]+(-0.797088515842)*x_ref[1]**o+(-0.257909258598)*x_ref[1]+(-0.927156080256)*x_ref[2]**o+(-0.843405252125)*x_ref[2]
        ref[2,0,2,1]=(0.447041226497)*x_ref[0]**o+(0.691225329527)*x_ref[0]+(0.462798475523)*x_ref[1]**o+(0.157497098383)*x_ref[1]+(-0.77785224695)*x_ref[2]**o+(0.886430058902)*x_ref[2]
        ref[2,1,0,0]=(0.397605326971)*x_ref[0]**o+(0.66227843262)*x_ref[0]+(-0.634922556623)*x_ref[1]**o+(-0.263342020187)*x_ref[1]+(-0.81486179124)*x_ref[2]**o+(-0.216716259698)*x_ref[2]
        ref[2,1,0,1]=(-0.868246547394)*x_ref[0]**o+(-0.194983059368)*x_ref[0]+(-0.667277750808)*x_ref[1]**o+(0.284011270958)*x_ref[1]+(-0.276295620425)*x_ref[2]**o+(0.215159516398)*x_ref[2]
        ref[2,1,1,0]=(0.990906320323)*x_ref[0]**o+(0.526754062323)*x_ref[0]+(-0.963398315893)*x_ref[1]**o+(-0.72561275544)*x_ref[1]+(0.539779065056)*x_ref[2]**o+(0.37738462043)*x_ref[2]
        ref[2,1,1,1]=(-0.668520028258)*x_ref[0]**o+(-0.292883094355)*x_ref[0]+(0.260949538172)*x_ref[1]**o+(-0.952706524522)*x_ref[1]+(-0.715386804875)*x_ref[2]**o+(-0.683320136224)*x_ref[2]
        ref[2,1,2,0]=(-0.646186301281)*x_ref[0]**o+(-0.62747987465)*x_ref[0]+(-0.858714171368)*x_ref[1]**o+(0.46910094021)*x_ref[1]+(0.468975652807)*x_ref[2]**o+(0.242115907129)*x_ref[2]
        ref[2,1,2,1]=(0.262205767991)*x_ref[0]**o+(-0.460376423099)*x_ref[0]+(0.88572453638)*x_ref[1]**o+(0.611761462581)*x_ref[1]+(0.763860971141)*x_ref[2]**o+(0.108470198525)*x_ref[2]
        ref[2,2,0,0]=(-0.0218691308052)*x_ref[0]**o+(0.942271392012)*x_ref[0]+(-0.778469966054)*x_ref[1]**o+(0.0873305333765)*x_ref[1]+(-0.300479183111)*x_ref[2]**o+(-0.862547713455)*x_ref[2]
        ref[2,2,0,1]=(-0.736808990311)*x_ref[0]**o+(-0.200478457878)*x_ref[0]+(0.272608362607)*x_ref[1]**o+(-0.0651056524306)*x_ref[1]+(0.616628495568)*x_ref[2]**o+(-0.8623490351)*x_ref[2]
        ref[2,2,1,0]=(0.931530668297)*x_ref[0]**o+(0.39701649536)*x_ref[0]+(0.552135074711)*x_ref[1]**o+(-0.489731326344)*x_ref[1]+(-0.864724191354)*x_ref[2]**o+(0.369557062583)*x_ref[2]
        ref[2,2,1,1]=(0.418780453395)*x_ref[0]**o+(-0.4935990418)*x_ref[0]+(0.34625989303)*x_ref[1]**o+(0.979829617988)*x_ref[1]+(0.853169987441)*x_ref[2]**o+(-0.945958722805)*x_ref[2]
        ref[2,2,2,0]=(0.973689940887)*x_ref[0]**o+(-0.0585784323994)*x_ref[0]+(-0.472419744635)*x_ref[1]**o+(0.851814193674)*x_ref[1]+(-0.393393340405)*x_ref[2]**o+(-0.755376551633)*x_ref[2]
        ref[2,2,2,1]=(-0.217049913648)*x_ref[0]**o+(-0.262150796255)*x_ref[0]+(-0.407805273745)*x_ref[1]**o+(0.819646964669)*x_ref[1]+(-0.0130316627236)*x_ref[2]**o+(0.179034130621)*x_ref[2]
        ref[2,3,0,0]=(-0.377708814803)*x_ref[0]**o+(-0.943893058411)*x_ref[0]+(-0.354341838264)*x_ref[1]**o+(0.0696491295689)*x_ref[1]+(-0.192838410156)*x_ref[2]**o+(-0.22208932914)*x_ref[2]
        ref[2,3,0,1]=(0.252429489766)*x_ref[0]**o+(0.808052283089)*x_ref[0]+(0.282941493712)*x_ref[1]**o+(0.89816007357)*x_ref[1]+(0.000979423783521)*x_ref[2]**o+(0.645599553602)*x_ref[2]
        ref[2,3,1,0]=(0.788089987784)*x_ref[0]**o+(0.414254044423)*x_ref[0]+(0.105204158639)*x_ref[1]**o+(0.97527041289)*x_ref[1]+(0.447463569594)*x_ref[2]**o+(0.91465488394)*x_ref[2]
        ref[2,3,1,1]=(0.882323493958)*x_ref[0]**o+(-0.98135461385)*x_ref[0]+(0.641881834427)*x_ref[1]**o+(-0.888068999894)*x_ref[1]+(0.705728558352)*x_ref[2]**o+(0.295815435834)*x_ref[2]
        ref[2,3,2,0]=(-0.570666876652)*x_ref[0]**o+(0.52308582534)*x_ref[0]+(-0.604213776925)*x_ref[1]**o+(-0.568684610916)*x_ref[1]+(-0.0378463784885)*x_ref[2]**o+(0.557994858276)*x_ref[2]
        ref[2,3,2,1]=(-0.497489149519)*x_ref[0]**o+(0.739341493998)*x_ref[0]+(-0.375869065986)*x_ref[1]**o+(-0.895553158302)*x_ref[1]+(0.916829505813)*x_ref[2]**o+(-0.650522069672)*x_ref[2]
        ref[2,4,0,0]=(0.627540843427)*x_ref[0]**o+(-0.572125230014)*x_ref[0]+(-0.404781893475)*x_ref[1]**o+(-0.114896790942)*x_ref[1]+(-0.108914764977)*x_ref[2]**o+(0.739758413931)*x_ref[2]
        ref[2,4,0,1]=(0.243205825969)*x_ref[0]**o+(-0.989602234642)*x_ref[0]+(-0.925730664677)*x_ref[1]**o+(0.454801040316)*x_ref[1]+(0.51758036389)*x_ref[2]**o+(0.325160660047)*x_ref[2]
        ref[2,4,1,0]=(0.559764013478)*x_ref[0]**o+(2.69034311577e-05)*x_ref[0]+(-0.691119937581)*x_ref[1]**o+(-0.347457112767)*x_ref[1]+(0.714683489268)*x_ref[2]**o+(-0.589270137254)*x_ref[2]
        ref[2,4,1,1]=(-0.657837524953)*x_ref[0]**o+(0.071158671662)*x_ref[0]+(-0.998677178035)*x_ref[1]**o+(-0.544136374784)*x_ref[1]+(0.6200901771)*x_ref[2]**o+(-0.280453395396)*x_ref[2]
        ref[2,4,2,0]=(0.692403743003)*x_ref[0]**o+(-0.570086669428)*x_ref[0]+(0.455036498939)*x_ref[1]**o+(0.291433179453)*x_ref[1]+(-0.203004457401)*x_ref[2]**o+(0.0192489593981)*x_ref[2]
        ref[2,4,2,1]=(-0.656783033826)*x_ref[0]**o+(0.865748949755)*x_ref[0]+(-0.297731124456)*x_ref[1]**o+(-0.826786841513)*x_ref[1]+(0.891374127144)*x_ref[2]**o+(-0.736789456498)*x_ref[2]
        ref[3,0,0,0]=(-0.963120586854)*x_ref[0]**o+(0.466089565408)*x_ref[0]+(-0.785137906849)*x_ref[1]**o+(0.194880437113)*x_ref[1]+(0.652057839639)*x_ref[2]**o+(-0.63674893761)*x_ref[2]
        ref[3,0,0,1]=(-0.923234625819)*x_ref[0]**o+(-0.303847080864)*x_ref[0]+(-0.374245465515)*x_ref[1]**o+(0.131093877726)*x_ref[1]+(-0.13469395748)*x_ref[2]**o+(0.42721676585)*x_ref[2]
        ref[3,0,1,0]=(-0.613510153481)*x_ref[0]**o+(0.986696685318)*x_ref[0]+(0.188583200151)*x_ref[1]**o+(-0.895175969251)*x_ref[1]+(-0.699510846377)*x_ref[2]**o+(0.953157679279)*x_ref[2]
        ref[3,0,1,1]=(-0.805809890231)*x_ref[0]**o+(-0.786467529592)*x_ref[0]+(-0.287240865489)*x_ref[1]**o+(-0.78714682888)*x_ref[1]+(-0.301034554724)*x_ref[2]**o+(0.698100092246)*x_ref[2]
        ref[3,0,2,0]=(-0.29320618172)*x_ref[0]**o+(0.482436772935)*x_ref[0]+(-0.126052929261)*x_ref[1]**o+(-0.0638862140753)*x_ref[1]+(0.42769411972)*x_ref[2]**o+(0.741118745829)*x_ref[2]
        ref[3,0,2,1]=(-0.380041427783)*x_ref[0]**o+(-0.0347904596008)*x_ref[0]+(-0.321599184546)*x_ref[1]**o+(-0.061333518544)*x_ref[1]+(-0.885193271364)*x_ref[2]**o+(0.216506075886)*x_ref[2]
        ref[3,1,0,0]=(-0.872457576682)*x_ref[0]**o+(0.1368749584)*x_ref[0]+(0.883209741333)*x_ref[1]**o+(0.946644762296)*x_ref[1]+(-0.941665219846)*x_ref[2]**o+(0.683787986795)*x_ref[2]
        ref[3,1,0,1]=(0.564843030077)*x_ref[0]**o+(0.987198898299)*x_ref[0]+(0.808116156661)*x_ref[1]**o+(-0.712172259479)*x_ref[1]+(0.942066890171)*x_ref[2]**o+(-0.265957158917)*x_ref[2]
        ref[3,1,1,0]=(-0.171018264631)*x_ref[0]**o+(0.277202848962)*x_ref[0]+(0.0726306773528)*x_ref[1]**o+(-0.0265640725409)*x_ref[1]+(-0.871118922929)*x_ref[2]**o+(0.0562314512261)*x_ref[2]
        ref[3,1,1,1]=(0.0861984079085)*x_ref[0]**o+(-0.954818837051)*x_ref[0]+(0.817362109646)*x_ref[1]**o+(-0.513086270222)*x_ref[1]+(-0.487472170741)*x_ref[2]**o+(0.702514192773)*x_ref[2]
        ref[3,1,2,0]=(0.272419057202)*x_ref[0]**o+(-0.0454292509397)*x_ref[0]+(-0.117471126603)*x_ref[1]**o+(0.638237510553)*x_ref[1]+(0.156858948691)*x_ref[2]**o+(0.499459702141)*x_ref[2]
        ref[3,1,2,1]=(-0.384525883334)*x_ref[0]**o+(-0.795638828759)*x_ref[0]+(-0.105795059118)*x_ref[1]**o+(0.539873566967)*x_ref[1]+(-0.402558868714)*x_ref[2]**o+(-0.690578475243)*x_ref[2]
        ref[3,2,0,0]=(0.776889517136)*x_ref[0]**o+(0.710931840891)*x_ref[0]+(0.335992677388)*x_ref[1]**o+(-0.689182863932)*x_ref[1]+(0.294149805572)*x_ref[2]**o+(-0.144223257292)*x_ref[2]
        ref[3,2,0,1]=(0.738703695733)*x_ref[0]**o+(-0.523207061995)*x_ref[0]+(0.707078352596)*x_ref[1]**o+(-0.594373978096)*x_ref[1]+(-0.833468457639)*x_ref[2]**o+(-0.60181674446)*x_ref[2]
        ref[3,2,1,0]=(-0.70243682985)*x_ref[0]**o+(-0.557914174271)*x_ref[0]+(-0.97153723053)*x_ref[1]**o+(0.791692594024)*x_ref[1]+(-0.442273808305)*x_ref[2]**o+(-0.74069605321)*x_ref[2]
        ref[3,2,1,1]=(0.136307039789)*x_ref[0]**o+(0.820661597623)*x_ref[0]+(0.379752060994)*x_ref[1]**o+(-0.739083116742)*x_ref[1]+(-0.964857055764)*x_ref[2]**o+(0.523159928671)*x_ref[2]
        ref[3,2,2,0]=(0.530883295074)*x_ref[0]**o+(-0.859538509888)*x_ref[0]+(-0.33917158254)*x_ref[1]**o+(0.476510741055)*x_ref[1]+(0.851975683024)*x_ref[2]**o+(0.146277015139)*x_ref[2]
        ref[3,2,2,1]=(-0.532984574212)*x_ref[0]**o+(-0.413169979754)*x_ref[0]+(0.0113284200501)*x_ref[1]**o+(0.406768750567)*x_ref[1]+(0.730968378807)*x_ref[2]**o+(-0.74004422909)*x_ref[2]
        ref[3,3,0,0]=(0.784793271465)*x_ref[0]**o+(0.879569037436)*x_ref[0]+(0.687804194006)*x_ref[1]**o+(0.153617608524)*x_ref[1]+(-0.410247334526)*x_ref[2]**o+(0.147941373268)*x_ref[2]
        ref[3,3,0,1]=(-0.896748874838)*x_ref[0]**o+(-0.408404228864)*x_ref[0]+(0.831967697989)*x_ref[1]**o+(-0.290875629476)*x_ref[1]+(-0.244101719785)*x_ref[2]**o+(0.240453425519)*x_ref[2]
        ref[3,3,1,0]=(-0.352293918285)*x_ref[0]**o+(0.0341233590345)*x_ref[0]+(-0.0165872699681)*x_ref[1]**o+(-0.0668507705942)*x_ref[1]+(-0.133733710061)*x_ref[2]**o+(-0.00620237772745)*x_ref[2]
        ref[3,3,1,1]=(0.173638606531)*x_ref[0]**o+(0.619705705811)*x_ref[0]+(0.279972531576)*x_ref[1]**o+(-0.0811181576037)*x_ref[1]+(-0.645673968637)*x_ref[2]**o+(0.913753411758)*x_ref[2]
        ref[3,3,2,0]=(0.0623921286142)*x_ref[0]**o+(0.393414889204)*x_ref[0]+(-0.300034851075)*x_ref[1]**o+(-0.519515206457)*x_ref[1]+(-0.0074981731879)*x_ref[2]**o+(0.202996242895)*x_ref[2]
        ref[3,3,2,1]=(0.583279597505)*x_ref[0]**o+(-0.940603487209)*x_ref[0]+(0.027966157338)*x_ref[1]**o+(0.251466344692)*x_ref[1]+(-0.97011033154)*x_ref[2]**o+(0.46527056203)*x_ref[2]
        ref[3,4,0,0]=(0.865170740218)*x_ref[0]**o+(-0.466853211289)*x_ref[0]+(0.0935492537278)*x_ref[1]**o+(-0.632374632657)*x_ref[1]+(0.673466346463)*x_ref[2]**o+(-0.47934739873)*x_ref[2]
        ref[3,4,0,1]=(0.0417451136027)*x_ref[0]**o+(-0.781570733286)*x_ref[0]+(0.547452389454)*x_ref[1]**o+(-0.0286368211528)*x_ref[1]+(-0.98907358808)*x_ref[2]**o+(0.62749495057)*x_ref[2]
        ref[3,4,1,0]=(-0.632067479664)*x_ref[0]**o+(-0.872835386336)*x_ref[0]+(-0.228785631078)*x_ref[1]**o+(-0.224206127306)*x_ref[1]+(0.769729051217)*x_ref[2]**o+(-0.406750034352)*x_ref[2]
        ref[3,4,1,1]=(0.216883071299)*x_ref[0]**o+(-0.225632200121)*x_ref[0]+(0.827186402743)*x_ref[1]**o+(0.161488636886)*x_ref[1]+(0.773801288836)*x_ref[2]**o+(0.977006832166)*x_ref[2]
        ref[3,4,2,0]=(-0.867098845884)*x_ref[0]**o+(0.14989544838)*x_ref[0]+(0.756720043262)*x_ref[1]**o+(-0.496074302626)*x_ref[1]+(-0.662787072451)*x_ref[2]**o+(-0.142473973368)*x_ref[2]
        ref[3,4,2,1]=(-0.294128396829)*x_ref[0]**o+(-0.931379392091)*x_ref[0]+(0.906552131161)*x_ref[1]**o+(-0.422008485494)*x_ref[1]+(0.806742455679)*x_ref[2]**o+(0.940488803006)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.387919945961)*x[0]**o+(-0.65988832446)*x[0]+(-0.271357984194)*x[1]**o+(0.911983947389)*x[1]
        ref=(-0.387919945961)*x_ref[0]**o+(-0.65988832446)*x_ref[0]+(-0.271357984194)*x_ref[1]**o+(0.911983947389)*x_ref[1]
      else:
        arg=(-0.0447075351093)*x[0]**o+(0.253461091432)*x[0]+(-0.218567784216)*x[1]**o+(-0.79030708754)*x[1]+(0.977690873872)*x[2]**o+(0.832406975763)*x[2]
        ref=(-0.0447075351093)*x_ref[0]**o+(0.253461091432)*x_ref[0]+(-0.218567784216)*x_ref[1]**o+(-0.79030708754)*x_ref[1]+(0.977690873872)*x_ref[2]**o+(0.832406975763)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.0351141106851)*x[0]**o+(-0.330587401)*x[0]+(0.244307738971)*x[1]**o+(0.623126601064)*x[1]
        arg[1]=(-0.828515624351)*x[0]**o+(-0.128313696693)*x[0]+(-0.419830633587)*x[1]**o+(0.207694344977)*x[1]
        ref[0]=(-0.0351141106851)*x_ref[0]**o+(-0.330587401)*x_ref[0]+(0.244307738971)*x_ref[1]**o+(0.623126601064)*x_ref[1]
        ref[1]=(-0.828515624351)*x_ref[0]**o+(-0.128313696693)*x_ref[0]+(-0.419830633587)*x_ref[1]**o+(0.207694344977)*x_ref[1]
      else:
        arg[0]=(-0.704794455532)*x[0]**o+(0.913653448977)*x[0]+(-0.159574360706)*x[1]**o+(-0.766199806615)*x[1]+(-0.571643737767)*x[2]**o+(-0.818256320426)*x[2]
        arg[1]=(-0.951104850067)*x[0]**o+(-0.661009825365)*x[0]+(0.415252265672)*x[1]**o+(-0.373834229809)*x[1]+(-0.727526942776)*x[2]**o+(-0.832051613933)*x[2]
        ref[0]=(-0.704794455532)*x_ref[0]**o+(0.913653448977)*x_ref[0]+(-0.159574360706)*x_ref[1]**o+(-0.766199806615)*x_ref[1]+(-0.571643737767)*x_ref[2]**o+(-0.818256320426)*x_ref[2]
        ref[1]=(-0.951104850067)*x_ref[0]**o+(-0.661009825365)*x_ref[0]+(0.415252265672)*x_ref[1]**o+(-0.373834229809)*x_ref[1]+(-0.727526942776)*x_ref[2]**o+(-0.832051613933)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.597205522265)*x[0]**o+(-0.990224489001)*x[0]+(0.409554155812)*x[1]**o+(0.0483989008223)*x[1]
        arg[0,1]=(0.397936023817)*x[0]**o+(-0.930167709289)*x[0]+(0.92611965278)*x[1]**o+(-0.964519543905)*x[1]
        arg[0,2]=(-0.311484417828)*x[0]**o+(-0.602916971577)*x[0]+(-0.420016358622)*x[1]**o+(0.837176635137)*x[1]
        arg[0,3]=(0.819115285277)*x[0]**o+(0.34838887804)*x[0]+(-0.690550904634)*x[1]**o+(0.373665882427)*x[1]
        arg[0,4]=(-0.7631912026)*x[0]**o+(-0.945320401033)*x[0]+(0.694292397122)*x[1]**o+(-0.420347487671)*x[1]
        arg[1,0]=(-0.572828652444)*x[0]**o+(0.354646387926)*x[0]+(-0.0507519017004)*x[1]**o+(-0.613889773716)*x[1]
        arg[1,1]=(-0.816957267715)*x[0]**o+(0.733537081373)*x[0]+(-0.660041553847)*x[1]**o+(0.441096986506)*x[1]
        arg[1,2]=(-0.803370878245)*x[0]**o+(-0.876568267674)*x[0]+(0.530143828975)*x[1]**o+(0.208130191749)*x[1]
        arg[1,3]=(-0.471900498331)*x[0]**o+(0.537812384385)*x[0]+(-0.882977296775)*x[1]**o+(-0.0668303027948)*x[1]
        arg[1,4]=(0.145525825808)*x[0]**o+(-0.743588928172)*x[0]+(-0.571360483951)*x[1]**o+(-0.779639689697)*x[1]
        arg[2,0]=(0.664624516473)*x[0]**o+(-0.903029393285)*x[0]+(-0.602274374813)*x[1]**o+(-0.0349025848298)*x[1]
        arg[2,1]=(-0.826235832601)*x[0]**o+(-0.56956830806)*x[0]+(-0.720529973236)*x[1]**o+(0.844549160654)*x[1]
        arg[2,2]=(0.865808902895)*x[0]**o+(0.806173922192)*x[0]+(-0.430500586315)*x[1]**o+(-0.903014287464)*x[1]
        arg[2,3]=(0.567909856768)*x[0]**o+(0.702374819746)*x[0]+(-0.931635013223)*x[1]**o+(0.35484582218)*x[1]
        arg[2,4]=(0.98750760384)*x[0]**o+(0.759340896388)*x[0]+(0.54913653446)*x[1]**o+(-0.620836882788)*x[1]
        arg[3,0]=(-0.886288358587)*x[0]**o+(0.393932009441)*x[0]+(0.370738890674)*x[1]**o+(0.540588748949)*x[1]
        arg[3,1]=(0.281435147449)*x[0]**o+(0.455838903759)*x[0]+(0.115176811183)*x[1]**o+(0.214335969407)*x[1]
        arg[3,2]=(-0.939824922583)*x[0]**o+(-0.725457358605)*x[0]+(0.0326844083908)*x[1]**o+(-0.743922869733)*x[1]
        arg[3,3]=(-0.189824771744)*x[0]**o+(-0.26508334021)*x[0]+(-0.323185410564)*x[1]**o+(0.242480018197)*x[1]
        arg[3,4]=(-0.675784178633)*x[0]**o+(0.0995857467734)*x[0]+(0.564599922701)*x[1]**o+(-0.716819470891)*x[1]
        ref[0,0]=(0.597205522265)*x_ref[0]**o+(-0.990224489001)*x_ref[0]+(0.409554155812)*x_ref[1]**o+(0.0483989008223)*x_ref[1]
        ref[0,1]=(0.397936023817)*x_ref[0]**o+(-0.930167709289)*x_ref[0]+(0.92611965278)*x_ref[1]**o+(-0.964519543905)*x_ref[1]
        ref[0,2]=(-0.311484417828)*x_ref[0]**o+(-0.602916971577)*x_ref[0]+(-0.420016358622)*x_ref[1]**o+(0.837176635137)*x_ref[1]
        ref[0,3]=(0.819115285277)*x_ref[0]**o+(0.34838887804)*x_ref[0]+(-0.690550904634)*x_ref[1]**o+(0.373665882427)*x_ref[1]
        ref[0,4]=(-0.7631912026)*x_ref[0]**o+(-0.945320401033)*x_ref[0]+(0.694292397122)*x_ref[1]**o+(-0.420347487671)*x_ref[1]
        ref[1,0]=(-0.572828652444)*x_ref[0]**o+(0.354646387926)*x_ref[0]+(-0.0507519017004)*x_ref[1]**o+(-0.613889773716)*x_ref[1]
        ref[1,1]=(-0.816957267715)*x_ref[0]**o+(0.733537081373)*x_ref[0]+(-0.660041553847)*x_ref[1]**o+(0.441096986506)*x_ref[1]
        ref[1,2]=(-0.803370878245)*x_ref[0]**o+(-0.876568267674)*x_ref[0]+(0.530143828975)*x_ref[1]**o+(0.208130191749)*x_ref[1]
        ref[1,3]=(-0.471900498331)*x_ref[0]**o+(0.537812384385)*x_ref[0]+(-0.882977296775)*x_ref[1]**o+(-0.0668303027948)*x_ref[1]
        ref[1,4]=(0.145525825808)*x_ref[0]**o+(-0.743588928172)*x_ref[0]+(-0.571360483951)*x_ref[1]**o+(-0.779639689697)*x_ref[1]
        ref[2,0]=(0.664624516473)*x_ref[0]**o+(-0.903029393285)*x_ref[0]+(-0.602274374813)*x_ref[1]**o+(-0.0349025848298)*x_ref[1]
        ref[2,1]=(-0.826235832601)*x_ref[0]**o+(-0.56956830806)*x_ref[0]+(-0.720529973236)*x_ref[1]**o+(0.844549160654)*x_ref[1]
        ref[2,2]=(0.865808902895)*x_ref[0]**o+(0.806173922192)*x_ref[0]+(-0.430500586315)*x_ref[1]**o+(-0.903014287464)*x_ref[1]
        ref[2,3]=(0.567909856768)*x_ref[0]**o+(0.702374819746)*x_ref[0]+(-0.931635013223)*x_ref[1]**o+(0.35484582218)*x_ref[1]
        ref[2,4]=(0.98750760384)*x_ref[0]**o+(0.759340896388)*x_ref[0]+(0.54913653446)*x_ref[1]**o+(-0.620836882788)*x_ref[1]
        ref[3,0]=(-0.886288358587)*x_ref[0]**o+(0.393932009441)*x_ref[0]+(0.370738890674)*x_ref[1]**o+(0.540588748949)*x_ref[1]
        ref[3,1]=(0.281435147449)*x_ref[0]**o+(0.455838903759)*x_ref[0]+(0.115176811183)*x_ref[1]**o+(0.214335969407)*x_ref[1]
        ref[3,2]=(-0.939824922583)*x_ref[0]**o+(-0.725457358605)*x_ref[0]+(0.0326844083908)*x_ref[1]**o+(-0.743922869733)*x_ref[1]
        ref[3,3]=(-0.189824771744)*x_ref[0]**o+(-0.26508334021)*x_ref[0]+(-0.323185410564)*x_ref[1]**o+(0.242480018197)*x_ref[1]
        ref[3,4]=(-0.675784178633)*x_ref[0]**o+(0.0995857467734)*x_ref[0]+(0.564599922701)*x_ref[1]**o+(-0.716819470891)*x_ref[1]
      else:
        arg[0,0]=(0.0429272016915)*x[0]**o+(-0.277438145968)*x[0]+(-0.663736923945)*x[1]**o+(0.754568662274)*x[1]+(-0.702929906882)*x[2]**o+(-0.744050838266)*x[2]
        arg[0,1]=(-0.245995343753)*x[0]**o+(-0.838699721329)*x[0]+(-0.266398695342)*x[1]**o+(0.116104359849)*x[1]+(-0.62406648751)*x[2]**o+(-0.85767565383)*x[2]
        arg[0,2]=(-0.969332932419)*x[0]**o+(-0.815839911463)*x[0]+(0.0532903209734)*x[1]**o+(-0.00108699727615)*x[1]+(0.142541999046)*x[2]**o+(0.971482543817)*x[2]
        arg[0,3]=(-0.73941517118)*x[0]**o+(0.728994326071)*x[0]+(0.962711932241)*x[1]**o+(-0.418607572448)*x[1]+(-0.374703265781)*x[2]**o+(-0.278661032739)*x[2]
        arg[0,4]=(-0.258020436163)*x[0]**o+(-0.0252121036831)*x[0]+(-0.756480314139)*x[1]**o+(-0.705045476776)*x[1]+(0.222309499093)*x[2]**o+(0.999237507123)*x[2]
        arg[1,0]=(-0.00753673411562)*x[0]**o+(-0.712947643092)*x[0]+(-0.946860154348)*x[1]**o+(-0.887293091702)*x[1]+(-0.57071975934)*x[2]**o+(-0.000995212472675)*x[2]
        arg[1,1]=(-0.547724214274)*x[0]**o+(-0.338492998124)*x[0]+(-0.191652366102)*x[1]**o+(-0.260315919511)*x[1]+(-0.103992070026)*x[2]**o+(-0.836638261621)*x[2]
        arg[1,2]=(-0.354711259362)*x[0]**o+(-0.916549411392)*x[0]+(0.744081321252)*x[1]**o+(-0.375805257126)*x[1]+(-0.78900115815)*x[2]**o+(-0.431116699853)*x[2]
        arg[1,3]=(0.543469030875)*x[0]**o+(-0.154316090927)*x[0]+(0.794234357287)*x[1]**o+(-0.894714087173)*x[1]+(-0.523490310245)*x[2]**o+(-0.734864830975)*x[2]
        arg[1,4]=(0.856478460898)*x[0]**o+(0.550846305146)*x[0]+(0.883099588119)*x[1]**o+(-0.928432890741)*x[1]+(0.953422014409)*x[2]**o+(-0.0295234113856)*x[2]
        arg[2,0]=(0.391420547389)*x[0]**o+(-0.0618665973255)*x[0]+(-0.536627285311)*x[1]**o+(-0.701891441965)*x[1]+(0.433098401229)*x[2]**o+(-0.821674981019)*x[2]
        arg[2,1]=(0.102455732687)*x[0]**o+(0.168078832413)*x[0]+(0.911099070324)*x[1]**o+(-0.776659572448)*x[1]+(0.942192938907)*x[2]**o+(-0.96284944292)*x[2]
        arg[2,2]=(0.677666801373)*x[0]**o+(0.579429978893)*x[0]+(-0.642666485576)*x[1]**o+(-0.0709929161913)*x[1]+(-0.686526389023)*x[2]**o+(0.0941806754766)*x[2]
        arg[2,3]=(-0.942137271102)*x[0]**o+(-0.448762007046)*x[0]+(0.403668316046)*x[1]**o+(-0.885790833279)*x[1]+(0.0448846125948)*x[2]**o+(0.123128346219)*x[2]
        arg[2,4]=(0.64615774988)*x[0]**o+(0.641971975809)*x[0]+(-0.0486200548159)*x[1]**o+(0.106284669825)*x[1]+(0.968867785284)*x[2]**o+(-0.693546581787)*x[2]
        arg[3,0]=(-0.0926306498682)*x[0]**o+(0.759541313455)*x[0]+(-0.569638025121)*x[1]**o+(0.236957551756)*x[1]+(-0.137429816873)*x[2]**o+(0.285836556676)*x[2]
        arg[3,1]=(0.0382484871512)*x[0]**o+(-0.481489628305)*x[0]+(0.809404670659)*x[1]**o+(0.124155422674)*x[1]+(-0.239216545446)*x[2]**o+(0.436110807769)*x[2]
        arg[3,2]=(-0.154044250435)*x[0]**o+(0.730285603192)*x[0]+(-0.986531989452)*x[1]**o+(0.0424454222102)*x[1]+(0.9651845632)*x[2]**o+(-0.173872544124)*x[2]
        arg[3,3]=(-0.528075843446)*x[0]**o+(0.226347788534)*x[0]+(-0.314941509835)*x[1]**o+(0.508180139398)*x[1]+(0.681251000904)*x[2]**o+(-0.442663397606)*x[2]
        arg[3,4]=(-0.335444251365)*x[0]**o+(-0.7788088839)*x[0]+(0.602538296871)*x[1]**o+(-0.440564659745)*x[1]+(-0.204809090144)*x[2]**o+(-0.449409088071)*x[2]
        ref[0,0]=(0.0429272016915)*x_ref[0]**o+(-0.277438145968)*x_ref[0]+(-0.663736923945)*x_ref[1]**o+(0.754568662274)*x_ref[1]+(-0.702929906882)*x_ref[2]**o+(-0.744050838266)*x_ref[2]
        ref[0,1]=(-0.245995343753)*x_ref[0]**o+(-0.838699721329)*x_ref[0]+(-0.266398695342)*x_ref[1]**o+(0.116104359849)*x_ref[1]+(-0.62406648751)*x_ref[2]**o+(-0.85767565383)*x_ref[2]
        ref[0,2]=(-0.969332932419)*x_ref[0]**o+(-0.815839911463)*x_ref[0]+(0.0532903209734)*x_ref[1]**o+(-0.00108699727615)*x_ref[1]+(0.142541999046)*x_ref[2]**o+(0.971482543817)*x_ref[2]
        ref[0,3]=(-0.73941517118)*x_ref[0]**o+(0.728994326071)*x_ref[0]+(0.962711932241)*x_ref[1]**o+(-0.418607572448)*x_ref[1]+(-0.374703265781)*x_ref[2]**o+(-0.278661032739)*x_ref[2]
        ref[0,4]=(-0.258020436163)*x_ref[0]**o+(-0.0252121036831)*x_ref[0]+(-0.756480314139)*x_ref[1]**o+(-0.705045476776)*x_ref[1]+(0.222309499093)*x_ref[2]**o+(0.999237507123)*x_ref[2]
        ref[1,0]=(-0.00753673411562)*x_ref[0]**o+(-0.712947643092)*x_ref[0]+(-0.946860154348)*x_ref[1]**o+(-0.887293091702)*x_ref[1]+(-0.57071975934)*x_ref[2]**o+(-0.000995212472675)*x_ref[2]
        ref[1,1]=(-0.547724214274)*x_ref[0]**o+(-0.338492998124)*x_ref[0]+(-0.191652366102)*x_ref[1]**o+(-0.260315919511)*x_ref[1]+(-0.103992070026)*x_ref[2]**o+(-0.836638261621)*x_ref[2]
        ref[1,2]=(-0.354711259362)*x_ref[0]**o+(-0.916549411392)*x_ref[0]+(0.744081321252)*x_ref[1]**o+(-0.375805257126)*x_ref[1]+(-0.78900115815)*x_ref[2]**o+(-0.431116699853)*x_ref[2]
        ref[1,3]=(0.543469030875)*x_ref[0]**o+(-0.154316090927)*x_ref[0]+(0.794234357287)*x_ref[1]**o+(-0.894714087173)*x_ref[1]+(-0.523490310245)*x_ref[2]**o+(-0.734864830975)*x_ref[2]
        ref[1,4]=(0.856478460898)*x_ref[0]**o+(0.550846305146)*x_ref[0]+(0.883099588119)*x_ref[1]**o+(-0.928432890741)*x_ref[1]+(0.953422014409)*x_ref[2]**o+(-0.0295234113856)*x_ref[2]
        ref[2,0]=(0.391420547389)*x_ref[0]**o+(-0.0618665973255)*x_ref[0]+(-0.536627285311)*x_ref[1]**o+(-0.701891441965)*x_ref[1]+(0.433098401229)*x_ref[2]**o+(-0.821674981019)*x_ref[2]
        ref[2,1]=(0.102455732687)*x_ref[0]**o+(0.168078832413)*x_ref[0]+(0.911099070324)*x_ref[1]**o+(-0.776659572448)*x_ref[1]+(0.942192938907)*x_ref[2]**o+(-0.96284944292)*x_ref[2]
        ref[2,2]=(0.677666801373)*x_ref[0]**o+(0.579429978893)*x_ref[0]+(-0.642666485576)*x_ref[1]**o+(-0.0709929161913)*x_ref[1]+(-0.686526389023)*x_ref[2]**o+(0.0941806754766)*x_ref[2]
        ref[2,3]=(-0.942137271102)*x_ref[0]**o+(-0.448762007046)*x_ref[0]+(0.403668316046)*x_ref[1]**o+(-0.885790833279)*x_ref[1]+(0.0448846125948)*x_ref[2]**o+(0.123128346219)*x_ref[2]
        ref[2,4]=(0.64615774988)*x_ref[0]**o+(0.641971975809)*x_ref[0]+(-0.0486200548159)*x_ref[1]**o+(0.106284669825)*x_ref[1]+(0.968867785284)*x_ref[2]**o+(-0.693546581787)*x_ref[2]
        ref[3,0]=(-0.0926306498682)*x_ref[0]**o+(0.759541313455)*x_ref[0]+(-0.569638025121)*x_ref[1]**o+(0.236957551756)*x_ref[1]+(-0.137429816873)*x_ref[2]**o+(0.285836556676)*x_ref[2]
        ref[3,1]=(0.0382484871512)*x_ref[0]**o+(-0.481489628305)*x_ref[0]+(0.809404670659)*x_ref[1]**o+(0.124155422674)*x_ref[1]+(-0.239216545446)*x_ref[2]**o+(0.436110807769)*x_ref[2]
        ref[3,2]=(-0.154044250435)*x_ref[0]**o+(0.730285603192)*x_ref[0]+(-0.986531989452)*x_ref[1]**o+(0.0424454222102)*x_ref[1]+(0.9651845632)*x_ref[2]**o+(-0.173872544124)*x_ref[2]
        ref[3,3]=(-0.528075843446)*x_ref[0]**o+(0.226347788534)*x_ref[0]+(-0.314941509835)*x_ref[1]**o+(0.508180139398)*x_ref[1]+(0.681251000904)*x_ref[2]**o+(-0.442663397606)*x_ref[2]
        ref[3,4]=(-0.335444251365)*x_ref[0]**o+(-0.7788088839)*x_ref[0]+(0.602538296871)*x_ref[1]**o+(-0.440564659745)*x_ref[1]+(-0.204809090144)*x_ref[2]**o+(-0.449409088071)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.88752640419)*x[0]**o+(-0.432514405615)*x[0]+(0.714823891809)*x[1]**o+(0.697307667986)*x[1]
        arg[0,0,1]=(0.738741506554)*x[0]**o+(0.188618518993)*x[0]+(-0.482898810269)*x[1]**o+(-0.914123176195)*x[1]
        arg[0,1,0]=(0.863872888691)*x[0]**o+(0.74507667202)*x[0]+(0.363690250942)*x[1]**o+(0.578470064401)*x[1]
        arg[0,1,1]=(-0.764594250812)*x[0]**o+(0.692105995894)*x[0]+(-0.950074761486)*x[1]**o+(-0.229643756622)*x[1]
        arg[1,0,0]=(-0.514123807377)*x[0]**o+(-0.770729480038)*x[0]+(-0.0503641588682)*x[1]**o+(-0.334340604627)*x[1]
        arg[1,0,1]=(-0.753427043851)*x[0]**o+(0.482153344254)*x[0]+(0.00162169870171)*x[1]**o+(-0.877830413948)*x[1]
        arg[1,1,0]=(0.603963041861)*x[0]**o+(-0.545048115282)*x[0]+(0.464500827027)*x[1]**o+(-0.527556750287)*x[1]
        arg[1,1,1]=(0.753201247064)*x[0]**o+(-0.99039225613)*x[0]+(-0.787959315932)*x[1]**o+(-0.918878072132)*x[1]
        arg[2,0,0]=(0.254002467345)*x[0]**o+(-0.918196533244)*x[0]+(-0.178408279918)*x[1]**o+(0.14954396814)*x[1]
        arg[2,0,1]=(0.449025666157)*x[0]**o+(-0.854430018688)*x[0]+(0.0393762923755)*x[1]**o+(0.576680446192)*x[1]
        arg[2,1,0]=(-0.252429553053)*x[0]**o+(-0.300398117312)*x[0]+(-0.0197102583755)*x[1]**o+(-0.189780116752)*x[1]
        arg[2,1,1]=(0.406291235871)*x[0]**o+(-0.763190166909)*x[0]+(-0.219957648446)*x[1]**o+(-0.928990138166)*x[1]
        arg[3,0,0]=(0.947321725555)*x[0]**o+(0.35310915009)*x[0]+(0.436067723051)*x[1]**o+(-0.86339979781)*x[1]
        arg[3,0,1]=(0.349458398624)*x[0]**o+(-0.139435138128)*x[0]+(-0.239809689531)*x[1]**o+(-0.0600998666403)*x[1]
        arg[3,1,0]=(-0.630539588121)*x[0]**o+(-0.700765237169)*x[0]+(-0.333988786061)*x[1]**o+(-0.125401172016)*x[1]
        arg[3,1,1]=(0.726389530806)*x[0]**o+(-0.691681620304)*x[0]+(0.241060086313)*x[1]**o+(-0.0970053499082)*x[1]
        arg[4,0,0]=(-0.798548958727)*x[0]**o+(0.429037250139)*x[0]+(0.598115431903)*x[1]**o+(-0.688464532796)*x[1]
        arg[4,0,1]=(0.698302852699)*x[0]**o+(0.36852807703)*x[0]+(-0.570086427187)*x[1]**o+(0.110387807246)*x[1]
        arg[4,1,0]=(0.295729624355)*x[0]**o+(0.247060092718)*x[0]+(-0.874774487091)*x[1]**o+(-0.40111840264)*x[1]
        arg[4,1,1]=(-0.773206505432)*x[0]**o+(-0.238827976714)*x[0]+(0.994288595399)*x[1]**o+(-0.6429945796)*x[1]
        arg[5,0,0]=(-0.398445578383)*x[0]**o+(-0.893518494161)*x[0]+(-0.855543638814)*x[1]**o+(0.93304149836)*x[1]
        arg[5,0,1]=(-0.811797150327)*x[0]**o+(0.0702614194127)*x[0]+(-0.0906132870911)*x[1]**o+(-0.944409144857)*x[1]
        arg[5,1,0]=(0.710407361984)*x[0]**o+(-0.854493576437)*x[0]+(-0.354683789897)*x[1]**o+(-0.896605671415)*x[1]
        arg[5,1,1]=(0.0668604398168)*x[0]**o+(0.464597275769)*x[0]+(-0.683756009627)*x[1]**o+(-0.368806656344)*x[1]
        ref[0,0,0]=(0.88752640419)*x_ref[0]**o+(-0.432514405615)*x_ref[0]+(0.714823891809)*x_ref[1]**o+(0.697307667986)*x_ref[1]
        ref[0,0,1]=(0.738741506554)*x_ref[0]**o+(0.188618518993)*x_ref[0]+(-0.482898810269)*x_ref[1]**o+(-0.914123176195)*x_ref[1]
        ref[0,1,0]=(0.863872888691)*x_ref[0]**o+(0.74507667202)*x_ref[0]+(0.363690250942)*x_ref[1]**o+(0.578470064401)*x_ref[1]
        ref[0,1,1]=(-0.764594250812)*x_ref[0]**o+(0.692105995894)*x_ref[0]+(-0.950074761486)*x_ref[1]**o+(-0.229643756622)*x_ref[1]
        ref[1,0,0]=(-0.514123807377)*x_ref[0]**o+(-0.770729480038)*x_ref[0]+(-0.0503641588682)*x_ref[1]**o+(-0.334340604627)*x_ref[1]
        ref[1,0,1]=(-0.753427043851)*x_ref[0]**o+(0.482153344254)*x_ref[0]+(0.00162169870171)*x_ref[1]**o+(-0.877830413948)*x_ref[1]
        ref[1,1,0]=(0.603963041861)*x_ref[0]**o+(-0.545048115282)*x_ref[0]+(0.464500827027)*x_ref[1]**o+(-0.527556750287)*x_ref[1]
        ref[1,1,1]=(0.753201247064)*x_ref[0]**o+(-0.99039225613)*x_ref[0]+(-0.787959315932)*x_ref[1]**o+(-0.918878072132)*x_ref[1]
        ref[2,0,0]=(0.254002467345)*x_ref[0]**o+(-0.918196533244)*x_ref[0]+(-0.178408279918)*x_ref[1]**o+(0.14954396814)*x_ref[1]
        ref[2,0,1]=(0.449025666157)*x_ref[0]**o+(-0.854430018688)*x_ref[0]+(0.0393762923755)*x_ref[1]**o+(0.576680446192)*x_ref[1]
        ref[2,1,0]=(-0.252429553053)*x_ref[0]**o+(-0.300398117312)*x_ref[0]+(-0.0197102583755)*x_ref[1]**o+(-0.189780116752)*x_ref[1]
        ref[2,1,1]=(0.406291235871)*x_ref[0]**o+(-0.763190166909)*x_ref[0]+(-0.219957648446)*x_ref[1]**o+(-0.928990138166)*x_ref[1]
        ref[3,0,0]=(0.947321725555)*x_ref[0]**o+(0.35310915009)*x_ref[0]+(0.436067723051)*x_ref[1]**o+(-0.86339979781)*x_ref[1]
        ref[3,0,1]=(0.349458398624)*x_ref[0]**o+(-0.139435138128)*x_ref[0]+(-0.239809689531)*x_ref[1]**o+(-0.0600998666403)*x_ref[1]
        ref[3,1,0]=(-0.630539588121)*x_ref[0]**o+(-0.700765237169)*x_ref[0]+(-0.333988786061)*x_ref[1]**o+(-0.125401172016)*x_ref[1]
        ref[3,1,1]=(0.726389530806)*x_ref[0]**o+(-0.691681620304)*x_ref[0]+(0.241060086313)*x_ref[1]**o+(-0.0970053499082)*x_ref[1]
        ref[4,0,0]=(-0.798548958727)*x_ref[0]**o+(0.429037250139)*x_ref[0]+(0.598115431903)*x_ref[1]**o+(-0.688464532796)*x_ref[1]
        ref[4,0,1]=(0.698302852699)*x_ref[0]**o+(0.36852807703)*x_ref[0]+(-0.570086427187)*x_ref[1]**o+(0.110387807246)*x_ref[1]
        ref[4,1,0]=(0.295729624355)*x_ref[0]**o+(0.247060092718)*x_ref[0]+(-0.874774487091)*x_ref[1]**o+(-0.40111840264)*x_ref[1]
        ref[4,1,1]=(-0.773206505432)*x_ref[0]**o+(-0.238827976714)*x_ref[0]+(0.994288595399)*x_ref[1]**o+(-0.6429945796)*x_ref[1]
        ref[5,0,0]=(-0.398445578383)*x_ref[0]**o+(-0.893518494161)*x_ref[0]+(-0.855543638814)*x_ref[1]**o+(0.93304149836)*x_ref[1]
        ref[5,0,1]=(-0.811797150327)*x_ref[0]**o+(0.0702614194127)*x_ref[0]+(-0.0906132870911)*x_ref[1]**o+(-0.944409144857)*x_ref[1]
        ref[5,1,0]=(0.710407361984)*x_ref[0]**o+(-0.854493576437)*x_ref[0]+(-0.354683789897)*x_ref[1]**o+(-0.896605671415)*x_ref[1]
        ref[5,1,1]=(0.0668604398168)*x_ref[0]**o+(0.464597275769)*x_ref[0]+(-0.683756009627)*x_ref[1]**o+(-0.368806656344)*x_ref[1]
      else:
        arg[0,0,0]=(-0.0873708254613)*x[0]**o+(-0.843989447866)*x[0]+(0.44751871924)*x[1]**o+(0.56511771844)*x[1]+(0.367414341304)*x[2]**o+(-0.147729223167)*x[2]
        arg[0,0,1]=(0.00752707306425)*x[0]**o+(0.819213438452)*x[0]+(0.14826826627)*x[1]**o+(-0.169362794728)*x[1]+(0.51736237082)*x[2]**o+(0.832746819884)*x[2]
        arg[0,1,0]=(0.607746431545)*x[0]**o+(0.253170100132)*x[0]+(-0.992132084299)*x[1]**o+(0.353865695431)*x[1]+(-0.787229685473)*x[2]**o+(0.86154695401)*x[2]
        arg[0,1,1]=(0.563636829468)*x[0]**o+(0.459627185677)*x[0]+(-0.511957702356)*x[1]**o+(0.309573033639)*x[1]+(0.806861185555)*x[2]**o+(0.182177897377)*x[2]
        arg[1,0,0]=(0.600850271814)*x[0]**o+(0.440922206975)*x[0]+(0.0307377466119)*x[1]**o+(-0.419096719454)*x[1]+(0.979284456332)*x[2]**o+(-0.00240732351601)*x[2]
        arg[1,0,1]=(0.51273517393)*x[0]**o+(-0.947643736857)*x[0]+(0.272403256871)*x[1]**o+(0.0119584891405)*x[1]+(-0.290604573286)*x[2]**o+(0.822291254588)*x[2]
        arg[1,1,0]=(-0.201909330605)*x[0]**o+(-0.560221198146)*x[0]+(0.564306911695)*x[1]**o+(-0.936933580593)*x[1]+(-0.41391834592)*x[2]**o+(0.133244793725)*x[2]
        arg[1,1,1]=(0.13318221571)*x[0]**o+(-0.565278280388)*x[0]+(0.677613822728)*x[1]**o+(-0.386612668638)*x[1]+(0.207570026134)*x[2]**o+(0.798362457061)*x[2]
        arg[2,0,0]=(-0.3777906845)*x[0]**o+(-0.285455928514)*x[0]+(-0.0901885959293)*x[1]**o+(-0.385922214051)*x[1]+(0.322566508003)*x[2]**o+(0.171191301702)*x[2]
        arg[2,0,1]=(-0.113954650916)*x[0]**o+(-0.373563461907)*x[0]+(0.73321196284)*x[1]**o+(0.535150793249)*x[1]+(-0.736944901672)*x[2]**o+(0.86133553525)*x[2]
        arg[2,1,0]=(-0.724228484035)*x[0]**o+(0.310971627294)*x[0]+(-0.497468717028)*x[1]**o+(0.467315341957)*x[1]+(-0.598941984951)*x[2]**o+(0.990367400534)*x[2]
        arg[2,1,1]=(0.594502107684)*x[0]**o+(0.261477996581)*x[0]+(0.738413631986)*x[1]**o+(0.363468523272)*x[1]+(-0.945845846052)*x[2]**o+(0.281144164925)*x[2]
        arg[3,0,0]=(0.804174935673)*x[0]**o+(-0.633954348085)*x[0]+(-0.953193492156)*x[1]**o+(0.632681376482)*x[1]+(-0.409900341362)*x[2]**o+(0.951988363964)*x[2]
        arg[3,0,1]=(-0.469895421121)*x[0]**o+(0.666115304582)*x[0]+(0.676351738247)*x[1]**o+(-0.0738256360284)*x[1]+(0.34976675523)*x[2]**o+(-0.561058490613)*x[2]
        arg[3,1,0]=(-0.999916478236)*x[0]**o+(-0.362871459365)*x[0]+(0.645604661917)*x[1]**o+(0.215878330855)*x[1]+(-0.0561173996823)*x[2]**o+(0.272143302204)*x[2]
        arg[3,1,1]=(0.233461828376)*x[0]**o+(-0.110673705352)*x[0]+(0.699710673197)*x[1]**o+(-0.628805264645)*x[1]+(0.677467640352)*x[2]**o+(-0.156224522726)*x[2]
        arg[4,0,0]=(0.180468931719)*x[0]**o+(0.299451524372)*x[0]+(-0.677131137371)*x[1]**o+(-0.422802649252)*x[1]+(0.237349200367)*x[2]**o+(0.739758623757)*x[2]
        arg[4,0,1]=(0.22209981155)*x[0]**o+(-0.226751058136)*x[0]+(-0.0345697170216)*x[1]**o+(0.265795819906)*x[1]+(0.542547943322)*x[2]**o+(0.785610668157)*x[2]
        arg[4,1,0]=(-0.520898604151)*x[0]**o+(-0.84307105084)*x[0]+(0.818061645146)*x[1]**o+(-0.173710357379)*x[1]+(0.344023171063)*x[2]**o+(0.194662166094)*x[2]
        arg[4,1,1]=(-0.288466404824)*x[0]**o+(-0.891856587908)*x[0]+(0.378329571068)*x[1]**o+(0.437919935526)*x[1]+(-0.630733422962)*x[2]**o+(0.518444630882)*x[2]
        arg[5,0,0]=(0.00978685202942)*x[0]**o+(-0.474443494209)*x[0]+(-0.509964215112)*x[1]**o+(0.341199671453)*x[1]+(-0.459333887251)*x[2]**o+(0.491961590667)*x[2]
        arg[5,0,1]=(-0.899850325159)*x[0]**o+(-0.191369137667)*x[0]+(0.18882430985)*x[1]**o+(-0.185214597442)*x[1]+(-0.967187797553)*x[2]**o+(0.366127839986)*x[2]
        arg[5,1,0]=(0.428007441668)*x[0]**o+(0.920828656579)*x[0]+(-0.298863178441)*x[1]**o+(-0.226020983335)*x[1]+(-0.01907141259)*x[2]**o+(-0.359698040557)*x[2]
        arg[5,1,1]=(0.980673633441)*x[0]**o+(0.362357489872)*x[0]+(0.825233899658)*x[1]**o+(0.846433827949)*x[1]+(-0.870197174533)*x[2]**o+(0.621884204716)*x[2]
        ref[0,0,0]=(-0.0873708254613)*x_ref[0]**o+(-0.843989447866)*x_ref[0]+(0.44751871924)*x_ref[1]**o+(0.56511771844)*x_ref[1]+(0.367414341304)*x_ref[2]**o+(-0.147729223167)*x_ref[2]
        ref[0,0,1]=(0.00752707306425)*x_ref[0]**o+(0.819213438452)*x_ref[0]+(0.14826826627)*x_ref[1]**o+(-0.169362794728)*x_ref[1]+(0.51736237082)*x_ref[2]**o+(0.832746819884)*x_ref[2]
        ref[0,1,0]=(0.607746431545)*x_ref[0]**o+(0.253170100132)*x_ref[0]+(-0.992132084299)*x_ref[1]**o+(0.353865695431)*x_ref[1]+(-0.787229685473)*x_ref[2]**o+(0.86154695401)*x_ref[2]
        ref[0,1,1]=(0.563636829468)*x_ref[0]**o+(0.459627185677)*x_ref[0]+(-0.511957702356)*x_ref[1]**o+(0.309573033639)*x_ref[1]+(0.806861185555)*x_ref[2]**o+(0.182177897377)*x_ref[2]
        ref[1,0,0]=(0.600850271814)*x_ref[0]**o+(0.440922206975)*x_ref[0]+(0.0307377466119)*x_ref[1]**o+(-0.419096719454)*x_ref[1]+(0.979284456332)*x_ref[2]**o+(-0.00240732351601)*x_ref[2]
        ref[1,0,1]=(0.51273517393)*x_ref[0]**o+(-0.947643736857)*x_ref[0]+(0.272403256871)*x_ref[1]**o+(0.0119584891405)*x_ref[1]+(-0.290604573286)*x_ref[2]**o+(0.822291254588)*x_ref[2]
        ref[1,1,0]=(-0.201909330605)*x_ref[0]**o+(-0.560221198146)*x_ref[0]+(0.564306911695)*x_ref[1]**o+(-0.936933580593)*x_ref[1]+(-0.41391834592)*x_ref[2]**o+(0.133244793725)*x_ref[2]
        ref[1,1,1]=(0.13318221571)*x_ref[0]**o+(-0.565278280388)*x_ref[0]+(0.677613822728)*x_ref[1]**o+(-0.386612668638)*x_ref[1]+(0.207570026134)*x_ref[2]**o+(0.798362457061)*x_ref[2]
        ref[2,0,0]=(-0.3777906845)*x_ref[0]**o+(-0.285455928514)*x_ref[0]+(-0.0901885959293)*x_ref[1]**o+(-0.385922214051)*x_ref[1]+(0.322566508003)*x_ref[2]**o+(0.171191301702)*x_ref[2]
        ref[2,0,1]=(-0.113954650916)*x_ref[0]**o+(-0.373563461907)*x_ref[0]+(0.73321196284)*x_ref[1]**o+(0.535150793249)*x_ref[1]+(-0.736944901672)*x_ref[2]**o+(0.86133553525)*x_ref[2]
        ref[2,1,0]=(-0.724228484035)*x_ref[0]**o+(0.310971627294)*x_ref[0]+(-0.497468717028)*x_ref[1]**o+(0.467315341957)*x_ref[1]+(-0.598941984951)*x_ref[2]**o+(0.990367400534)*x_ref[2]
        ref[2,1,1]=(0.594502107684)*x_ref[0]**o+(0.261477996581)*x_ref[0]+(0.738413631986)*x_ref[1]**o+(0.363468523272)*x_ref[1]+(-0.945845846052)*x_ref[2]**o+(0.281144164925)*x_ref[2]
        ref[3,0,0]=(0.804174935673)*x_ref[0]**o+(-0.633954348085)*x_ref[0]+(-0.953193492156)*x_ref[1]**o+(0.632681376482)*x_ref[1]+(-0.409900341362)*x_ref[2]**o+(0.951988363964)*x_ref[2]
        ref[3,0,1]=(-0.469895421121)*x_ref[0]**o+(0.666115304582)*x_ref[0]+(0.676351738247)*x_ref[1]**o+(-0.0738256360284)*x_ref[1]+(0.34976675523)*x_ref[2]**o+(-0.561058490613)*x_ref[2]
        ref[3,1,0]=(-0.999916478236)*x_ref[0]**o+(-0.362871459365)*x_ref[0]+(0.645604661917)*x_ref[1]**o+(0.215878330855)*x_ref[1]+(-0.0561173996823)*x_ref[2]**o+(0.272143302204)*x_ref[2]
        ref[3,1,1]=(0.233461828376)*x_ref[0]**o+(-0.110673705352)*x_ref[0]+(0.699710673197)*x_ref[1]**o+(-0.628805264645)*x_ref[1]+(0.677467640352)*x_ref[2]**o+(-0.156224522726)*x_ref[2]
        ref[4,0,0]=(0.180468931719)*x_ref[0]**o+(0.299451524372)*x_ref[0]+(-0.677131137371)*x_ref[1]**o+(-0.422802649252)*x_ref[1]+(0.237349200367)*x_ref[2]**o+(0.739758623757)*x_ref[2]
        ref[4,0,1]=(0.22209981155)*x_ref[0]**o+(-0.226751058136)*x_ref[0]+(-0.0345697170216)*x_ref[1]**o+(0.265795819906)*x_ref[1]+(0.542547943322)*x_ref[2]**o+(0.785610668157)*x_ref[2]
        ref[4,1,0]=(-0.520898604151)*x_ref[0]**o+(-0.84307105084)*x_ref[0]+(0.818061645146)*x_ref[1]**o+(-0.173710357379)*x_ref[1]+(0.344023171063)*x_ref[2]**o+(0.194662166094)*x_ref[2]
        ref[4,1,1]=(-0.288466404824)*x_ref[0]**o+(-0.891856587908)*x_ref[0]+(0.378329571068)*x_ref[1]**o+(0.437919935526)*x_ref[1]+(-0.630733422962)*x_ref[2]**o+(0.518444630882)*x_ref[2]
        ref[5,0,0]=(0.00978685202942)*x_ref[0]**o+(-0.474443494209)*x_ref[0]+(-0.509964215112)*x_ref[1]**o+(0.341199671453)*x_ref[1]+(-0.459333887251)*x_ref[2]**o+(0.491961590667)*x_ref[2]
        ref[5,0,1]=(-0.899850325159)*x_ref[0]**o+(-0.191369137667)*x_ref[0]+(0.18882430985)*x_ref[1]**o+(-0.185214597442)*x_ref[1]+(-0.967187797553)*x_ref[2]**o+(0.366127839986)*x_ref[2]
        ref[5,1,0]=(0.428007441668)*x_ref[0]**o+(0.920828656579)*x_ref[0]+(-0.298863178441)*x_ref[1]**o+(-0.226020983335)*x_ref[1]+(-0.01907141259)*x_ref[2]**o+(-0.359698040557)*x_ref[2]
        ref[5,1,1]=(0.980673633441)*x_ref[0]**o+(0.362357489872)*x_ref[0]+(0.825233899658)*x_ref[1]**o+(0.846433827949)*x_ref[1]+(-0.870197174533)*x_ref[2]**o+(0.621884204716)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.93572957134)*x[0]**o+(-0.878081515023)*x[0]+(0.821839395118)*x[1]**o+(-0.088026044081)*x[1]
        arg[0,0,0,1]=(-0.00393438000839)*x[0]**o+(-0.180693225971)*x[0]+(0.192781914487)*x[1]**o+(-0.561366977374)*x[1]
        arg[0,0,1,0]=(0.116028558326)*x[0]**o+(-0.47739282219)*x[0]+(0.97806148353)*x[1]**o+(0.660972051388)*x[1]
        arg[0,0,1,1]=(0.092683552469)*x[0]**o+(0.22083599745)*x[0]+(-0.423073290439)*x[1]**o+(-0.0789522599045)*x[1]
        arg[0,0,2,0]=(0.140023403084)*x[0]**o+(-0.927034796519)*x[0]+(0.0987957525564)*x[1]**o+(0.0703695028562)*x[1]
        arg[0,0,2,1]=(-0.00485147064392)*x[0]**o+(-0.649679595932)*x[0]+(0.503435626816)*x[1]**o+(0.393293792881)*x[1]
        arg[0,1,0,0]=(-0.753383897772)*x[0]**o+(-0.988499439357)*x[0]+(-0.651037417175)*x[1]**o+(0.45184892046)*x[1]
        arg[0,1,0,1]=(0.0120892569605)*x[0]**o+(0.52287896209)*x[0]+(0.103256695817)*x[1]**o+(-0.0244835681092)*x[1]
        arg[0,1,1,0]=(-0.0949239075135)*x[0]**o+(0.835062318231)*x[0]+(-0.0522981103769)*x[1]**o+(0.60020994838)*x[1]
        arg[0,1,1,1]=(0.139896941482)*x[0]**o+(0.409042945999)*x[0]+(0.10019215394)*x[1]**o+(-0.312224123017)*x[1]
        arg[0,1,2,0]=(-0.892437964342)*x[0]**o+(0.0828807279746)*x[0]+(0.777811646235)*x[1]**o+(-0.96509471481)*x[1]
        arg[0,1,2,1]=(-0.397341226148)*x[0]**o+(0.102987581336)*x[0]+(-0.69165111598)*x[1]**o+(0.743298211061)*x[1]
        arg[0,2,0,0]=(-0.300129864031)*x[0]**o+(-0.418981381232)*x[0]+(0.624854922555)*x[1]**o+(0.892651746561)*x[1]
        arg[0,2,0,1]=(-0.511463603881)*x[0]**o+(-0.808468332238)*x[0]+(-0.342755248919)*x[1]**o+(0.357985875808)*x[1]
        arg[0,2,1,0]=(-0.269520452932)*x[0]**o+(0.869553937081)*x[0]+(-0.183222122614)*x[1]**o+(-0.135280721586)*x[1]
        arg[0,2,1,1]=(0.0396712123969)*x[0]**o+(0.344314072793)*x[0]+(0.644676507168)*x[1]**o+(-0.0799551435053)*x[1]
        arg[0,2,2,0]=(0.302628743334)*x[0]**o+(-0.59557047376)*x[0]+(-0.277654790212)*x[1]**o+(0.0299035042863)*x[1]
        arg[0,2,2,1]=(0.805296568248)*x[0]**o+(0.726404711154)*x[0]+(0.0728356962577)*x[1]**o+(0.199997237962)*x[1]
        arg[0,3,0,0]=(-0.292658800952)*x[0]**o+(-0.40080106651)*x[0]+(0.844109590553)*x[1]**o+(-0.339364790362)*x[1]
        arg[0,3,0,1]=(-0.621465058894)*x[0]**o+(-0.376804350545)*x[0]+(0.745038406103)*x[1]**o+(-0.281482724757)*x[1]
        arg[0,3,1,0]=(0.00625816347702)*x[0]**o+(0.392750189511)*x[0]+(-0.232258708605)*x[1]**o+(-0.634552067635)*x[1]
        arg[0,3,1,1]=(-0.0893109545391)*x[0]**o+(-0.735678058768)*x[0]+(-0.603248350212)*x[1]**o+(-0.664366286468)*x[1]
        arg[0,3,2,0]=(-0.0539114688775)*x[0]**o+(-0.110100592496)*x[0]+(-0.700496078261)*x[1]**o+(0.203311460549)*x[1]
        arg[0,3,2,1]=(-0.554691350426)*x[0]**o+(-0.255001039659)*x[0]+(0.567977217289)*x[1]**o+(0.61993007447)*x[1]
        arg[0,4,0,0]=(-0.12193490862)*x[0]**o+(0.731139340661)*x[0]+(-0.321971724057)*x[1]**o+(0.741215588897)*x[1]
        arg[0,4,0,1]=(-0.394743164913)*x[0]**o+(-0.602249917929)*x[0]+(0.27675734517)*x[1]**o+(-0.968806169573)*x[1]
        arg[0,4,1,0]=(-0.629331232169)*x[0]**o+(0.687483125805)*x[0]+(0.633698693186)*x[1]**o+(0.837564332901)*x[1]
        arg[0,4,1,1]=(0.0470601891829)*x[0]**o+(-0.365490813497)*x[0]+(-0.162532786979)*x[1]**o+(0.504146804122)*x[1]
        arg[0,4,2,0]=(0.193790444091)*x[0]**o+(0.742132840869)*x[0]+(-0.871677496589)*x[1]**o+(0.492817750307)*x[1]
        arg[0,4,2,1]=(0.741218732478)*x[0]**o+(-0.746973604492)*x[0]+(-0.548073707926)*x[1]**o+(-0.42258185128)*x[1]
        arg[1,0,0,0]=(0.580989627459)*x[0]**o+(-0.498423525574)*x[0]+(-0.349651580594)*x[1]**o+(0.0193060779482)*x[1]
        arg[1,0,0,1]=(0.398917011903)*x[0]**o+(-0.122064497238)*x[0]+(0.924654557299)*x[1]**o+(0.485284066068)*x[1]
        arg[1,0,1,0]=(-0.985949382826)*x[0]**o+(-0.0188384874977)*x[0]+(0.177995429697)*x[1]**o+(0.967727833143)*x[1]
        arg[1,0,1,1]=(0.920799980503)*x[0]**o+(-0.749281119074)*x[0]+(-0.628411322607)*x[1]**o+(-0.446396361734)*x[1]
        arg[1,0,2,0]=(0.752863123667)*x[0]**o+(0.112427175329)*x[0]+(0.975568251269)*x[1]**o+(0.316938392669)*x[1]
        arg[1,0,2,1]=(0.798409861933)*x[0]**o+(-0.145715624761)*x[0]+(-0.847813831553)*x[1]**o+(0.302999756348)*x[1]
        arg[1,1,0,0]=(0.897029838502)*x[0]**o+(-0.540852276529)*x[0]+(-0.541450963121)*x[1]**o+(-0.118819043917)*x[1]
        arg[1,1,0,1]=(-0.616797542123)*x[0]**o+(-0.139394144428)*x[0]+(-0.480348413917)*x[1]**o+(0.500244829288)*x[1]
        arg[1,1,1,0]=(-0.960388730531)*x[0]**o+(0.87391147728)*x[0]+(0.0128673975991)*x[1]**o+(0.444341296582)*x[1]
        arg[1,1,1,1]=(0.136810869097)*x[0]**o+(-0.544756866447)*x[0]+(-0.198052789907)*x[1]**o+(-0.536732583929)*x[1]
        arg[1,1,2,0]=(-0.808981135979)*x[0]**o+(-0.122088197438)*x[0]+(-0.363958821187)*x[1]**o+(-0.521614066151)*x[1]
        arg[1,1,2,1]=(0.299382605072)*x[0]**o+(0.143397060894)*x[0]+(-0.223044874711)*x[1]**o+(-0.04723866032)*x[1]
        arg[1,2,0,0]=(-0.568967551455)*x[0]**o+(0.750777908696)*x[0]+(0.203117458853)*x[1]**o+(-0.0797168995572)*x[1]
        arg[1,2,0,1]=(0.0439147552635)*x[0]**o+(-0.0746749889653)*x[0]+(0.192885940929)*x[1]**o+(0.794090802345)*x[1]
        arg[1,2,1,0]=(0.137177876023)*x[0]**o+(0.625847668003)*x[0]+(0.252761858157)*x[1]**o+(0.496512978117)*x[1]
        arg[1,2,1,1]=(0.475218888846)*x[0]**o+(-0.286926184073)*x[0]+(0.551427803487)*x[1]**o+(0.782373661286)*x[1]
        arg[1,2,2,0]=(0.903004502545)*x[0]**o+(-0.262952251654)*x[0]+(0.110351860812)*x[1]**o+(-0.685355934696)*x[1]
        arg[1,2,2,1]=(-0.157457317404)*x[0]**o+(-0.243850623177)*x[0]+(0.28853283432)*x[1]**o+(0.612508908614)*x[1]
        arg[1,3,0,0]=(-0.741890002973)*x[0]**o+(-0.118262424825)*x[0]+(0.567100734646)*x[1]**o+(0.301557125811)*x[1]
        arg[1,3,0,1]=(-0.00745396583928)*x[0]**o+(-0.00558491344276)*x[0]+(-0.967132370197)*x[1]**o+(-0.61737594402)*x[1]
        arg[1,3,1,0]=(0.0767961523966)*x[0]**o+(0.341952000996)*x[0]+(0.762617832019)*x[1]**o+(-0.38337284713)*x[1]
        arg[1,3,1,1]=(0.32265018902)*x[0]**o+(-0.352521736206)*x[0]+(-0.0119143186529)*x[1]**o+(-0.64727561829)*x[1]
        arg[1,3,2,0]=(0.627690584587)*x[0]**o+(0.0993819822193)*x[0]+(-0.189363448841)*x[1]**o+(0.49004001194)*x[1]
        arg[1,3,2,1]=(-0.657836476156)*x[0]**o+(0.708970470136)*x[0]+(0.252190020576)*x[1]**o+(-0.481214196298)*x[1]
        arg[1,4,0,0]=(0.0837563164179)*x[0]**o+(0.741536525079)*x[0]+(0.645543094286)*x[1]**o+(0.62358090815)*x[1]
        arg[1,4,0,1]=(0.262043398371)*x[0]**o+(-0.47794750936)*x[0]+(0.93086244965)*x[1]**o+(-0.670399062677)*x[1]
        arg[1,4,1,0]=(-0.76634902923)*x[0]**o+(-0.809846143928)*x[0]+(0.639099549305)*x[1]**o+(0.793524639236)*x[1]
        arg[1,4,1,1]=(0.814737503957)*x[0]**o+(0.775593374042)*x[0]+(0.240976904111)*x[1]**o+(0.2352421356)*x[1]
        arg[1,4,2,0]=(0.519210584047)*x[0]**o+(-0.702579589691)*x[0]+(-0.450156317853)*x[1]**o+(0.588864692581)*x[1]
        arg[1,4,2,1]=(-0.927751546037)*x[0]**o+(-0.239737576019)*x[0]+(-0.527658440447)*x[1]**o+(0.79623440559)*x[1]
        arg[2,0,0,0]=(0.719144336419)*x[0]**o+(0.916780231277)*x[0]+(0.279073450277)*x[1]**o+(-0.0130671321167)*x[1]
        arg[2,0,0,1]=(-0.556604998962)*x[0]**o+(-0.661580370137)*x[0]+(0.980704664791)*x[1]**o+(0.935843251131)*x[1]
        arg[2,0,1,0]=(0.85501679753)*x[0]**o+(-0.908342467889)*x[0]+(0.969030151294)*x[1]**o+(0.534192961922)*x[1]
        arg[2,0,1,1]=(0.527672385529)*x[0]**o+(0.997699955667)*x[0]+(-0.388249629327)*x[1]**o+(-0.714956937512)*x[1]
        arg[2,0,2,0]=(-0.629149810667)*x[0]**o+(-0.0515018376692)*x[0]+(-0.80355991702)*x[1]**o+(0.235999988886)*x[1]
        arg[2,0,2,1]=(0.960715934156)*x[0]**o+(-0.782787305497)*x[0]+(-0.309348231959)*x[1]**o+(0.854911294717)*x[1]
        arg[2,1,0,0]=(-0.560749125074)*x[0]**o+(-0.120515615757)*x[0]+(0.055090634431)*x[1]**o+(-0.539205949358)*x[1]
        arg[2,1,0,1]=(-0.762113417295)*x[0]**o+(-0.207196980376)*x[0]+(0.904541245494)*x[1]**o+(0.320359459939)*x[1]
        arg[2,1,1,0]=(-0.272627586486)*x[0]**o+(-0.885987982522)*x[0]+(0.641737808216)*x[1]**o+(0.358390367373)*x[1]
        arg[2,1,1,1]=(-0.847502563761)*x[0]**o+(-0.430298277492)*x[0]+(-0.869043300185)*x[1]**o+(0.584700145915)*x[1]
        arg[2,1,2,0]=(0.497688368946)*x[0]**o+(-0.14413770704)*x[0]+(-0.82002941888)*x[1]**o+(0.573569419279)*x[1]
        arg[2,1,2,1]=(0.959549623603)*x[0]**o+(0.881196513588)*x[0]+(-0.202951096266)*x[1]**o+(-9.87983182765e-05)*x[1]
        arg[2,2,0,0]=(0.0811393810501)*x[0]**o+(-0.853555165896)*x[0]+(0.0259257246335)*x[1]**o+(0.638575306959)*x[1]
        arg[2,2,0,1]=(-0.525990031379)*x[0]**o+(0.11235160765)*x[0]+(-0.502853879455)*x[1]**o+(0.414066683507)*x[1]
        arg[2,2,1,0]=(-0.533919716834)*x[0]**o+(-0.773411100656)*x[0]+(0.311424028491)*x[1]**o+(0.332733301377)*x[1]
        arg[2,2,1,1]=(0.280548593462)*x[0]**o+(-0.0747640762104)*x[0]+(-0.7110232877)*x[1]**o+(-0.178672817441)*x[1]
        arg[2,2,2,0]=(0.445025926846)*x[0]**o+(-0.669782903005)*x[0]+(0.236610411793)*x[1]**o+(0.437836936476)*x[1]
        arg[2,2,2,1]=(-0.503761305098)*x[0]**o+(-0.889922659865)*x[0]+(-0.0767256016402)*x[1]**o+(-0.614874217835)*x[1]
        arg[2,3,0,0]=(-0.041960855672)*x[0]**o+(-0.593486242764)*x[0]+(0.344751892525)*x[1]**o+(-0.131268873786)*x[1]
        arg[2,3,0,1]=(-0.192704843083)*x[0]**o+(0.325356381696)*x[0]+(-0.235690651937)*x[1]**o+(0.59113535617)*x[1]
        arg[2,3,1,0]=(-0.0794745628024)*x[0]**o+(0.7996737586)*x[0]+(-0.324970505375)*x[1]**o+(-0.14920403468)*x[1]
        arg[2,3,1,1]=(0.31997432673)*x[0]**o+(-0.571739807799)*x[0]+(0.517683480263)*x[1]**o+(-0.250680741704)*x[1]
        arg[2,3,2,0]=(-0.379114277883)*x[0]**o+(0.953285017352)*x[0]+(-0.701111847548)*x[1]**o+(-0.354199942528)*x[1]
        arg[2,3,2,1]=(0.492245629486)*x[0]**o+(-0.299380102293)*x[0]+(-0.126371826833)*x[1]**o+(0.933664060647)*x[1]
        arg[2,4,0,0]=(0.714822159533)*x[0]**o+(0.359430380478)*x[0]+(-0.691610768506)*x[1]**o+(0.414061853203)*x[1]
        arg[2,4,0,1]=(0.412799253275)*x[0]**o+(0.862905533148)*x[0]+(-0.0507440946561)*x[1]**o+(0.779012369976)*x[1]
        arg[2,4,1,0]=(-0.981179053034)*x[0]**o+(0.984777005573)*x[0]+(0.715584788452)*x[1]**o+(0.920666888724)*x[1]
        arg[2,4,1,1]=(-0.0287775685032)*x[0]**o+(0.617592345344)*x[0]+(-0.767451297794)*x[1]**o+(0.95828976091)*x[1]
        arg[2,4,2,0]=(0.905037473705)*x[0]**o+(-0.706987158021)*x[0]+(0.535436652493)*x[1]**o+(-0.891712664249)*x[1]
        arg[2,4,2,1]=(-0.818709474925)*x[0]**o+(-0.0223479273553)*x[0]+(-0.370644568884)*x[1]**o+(0.286468739105)*x[1]
        arg[3,0,0,0]=(0.310461949876)*x[0]**o+(0.770193446516)*x[0]+(-0.0858335164699)*x[1]**o+(0.12572608122)*x[1]
        arg[3,0,0,1]=(-0.781332422249)*x[0]**o+(-0.238368527872)*x[0]+(-0.449272256582)*x[1]**o+(-0.645638738212)*x[1]
        arg[3,0,1,0]=(-0.91143447678)*x[0]**o+(-0.885131081008)*x[0]+(-0.253833463581)*x[1]**o+(0.051439649459)*x[1]
        arg[3,0,1,1]=(0.447730169203)*x[0]**o+(0.980354452325)*x[0]+(-0.0484566141047)*x[1]**o+(0.443407604362)*x[1]
        arg[3,0,2,0]=(-0.223975544426)*x[0]**o+(0.0141796274312)*x[0]+(0.664080206442)*x[1]**o+(0.193318634886)*x[1]
        arg[3,0,2,1]=(-0.388901068755)*x[0]**o+(-0.632210208163)*x[0]+(0.567264328931)*x[1]**o+(-0.308710733281)*x[1]
        arg[3,1,0,0]=(0.259562910069)*x[0]**o+(-0.315409720689)*x[0]+(0.999281841483)*x[1]**o+(0.695981779357)*x[1]
        arg[3,1,0,1]=(0.627508253716)*x[0]**o+(-0.540950954632)*x[0]+(0.626209667445)*x[1]**o+(-0.920403595109)*x[1]
        arg[3,1,1,0]=(0.750166528891)*x[0]**o+(-0.145494783124)*x[0]+(0.71048043416)*x[1]**o+(-0.0920505950401)*x[1]
        arg[3,1,1,1]=(-0.755570630755)*x[0]**o+(-0.323456877629)*x[0]+(-0.0318418288689)*x[1]**o+(0.551823313307)*x[1]
        arg[3,1,2,0]=(-0.301380276426)*x[0]**o+(-0.102899117196)*x[0]+(0.816857028902)*x[1]**o+(-0.571555621895)*x[1]
        arg[3,1,2,1]=(-0.791148729922)*x[0]**o+(0.400286696027)*x[0]+(-0.17362002139)*x[1]**o+(0.431696064424)*x[1]
        arg[3,2,0,0]=(0.241944451237)*x[0]**o+(-0.629401496412)*x[0]+(-0.00333636607647)*x[1]**o+(0.167334133677)*x[1]
        arg[3,2,0,1]=(-0.300944831932)*x[0]**o+(0.358282773504)*x[0]+(0.759495661081)*x[1]**o+(-0.225250488484)*x[1]
        arg[3,2,1,0]=(0.755950588768)*x[0]**o+(0.718423180013)*x[0]+(0.705150216505)*x[1]**o+(0.269283969196)*x[1]
        arg[3,2,1,1]=(0.811064150458)*x[0]**o+(-0.67847842696)*x[0]+(-0.478091805711)*x[1]**o+(0.292774429011)*x[1]
        arg[3,2,2,0]=(-0.307877253758)*x[0]**o+(-0.563961673764)*x[0]+(0.428322493356)*x[1]**o+(0.564230126489)*x[1]
        arg[3,2,2,1]=(-0.623877056979)*x[0]**o+(-0.148337578412)*x[0]+(-0.910662605581)*x[1]**o+(-0.38603493572)*x[1]
        arg[3,3,0,0]=(-0.826474270729)*x[0]**o+(0.977614689932)*x[0]+(-0.369824837717)*x[1]**o+(-0.985478011812)*x[1]
        arg[3,3,0,1]=(-0.251252769561)*x[0]**o+(-0.296732765057)*x[0]+(-0.888899418995)*x[1]**o+(0.640313055734)*x[1]
        arg[3,3,1,0]=(-0.189106218841)*x[0]**o+(-0.0441656339777)*x[0]+(0.368716219801)*x[1]**o+(-0.924144919367)*x[1]
        arg[3,3,1,1]=(-0.525205981948)*x[0]**o+(-0.148352446593)*x[0]+(0.869212821652)*x[1]**o+(-0.575463559793)*x[1]
        arg[3,3,2,0]=(0.668772663439)*x[0]**o+(0.4330630456)*x[0]+(-0.321158907274)*x[1]**o+(0.570853139693)*x[1]
        arg[3,3,2,1]=(0.0236674031998)*x[0]**o+(-0.390106024651)*x[0]+(0.541717893494)*x[1]**o+(-0.144655705849)*x[1]
        arg[3,4,0,0]=(0.971861188602)*x[0]**o+(-0.834463063474)*x[0]+(-0.419545662218)*x[1]**o+(-0.0387341522049)*x[1]
        arg[3,4,0,1]=(0.226357589896)*x[0]**o+(0.305148895256)*x[0]+(0.00306527642431)*x[1]**o+(0.4715619746)*x[1]
        arg[3,4,1,0]=(0.529602340964)*x[0]**o+(-0.482050910795)*x[0]+(0.964983911769)*x[1]**o+(0.524018220532)*x[1]
        arg[3,4,1,1]=(0.940676158872)*x[0]**o+(0.547078925647)*x[0]+(-0.636907424453)*x[1]**o+(0.938262396281)*x[1]
        arg[3,4,2,0]=(0.282512784255)*x[0]**o+(-0.556345478931)*x[0]+(0.651866614458)*x[1]**o+(0.8895132214)*x[1]
        arg[3,4,2,1]=(0.905077434472)*x[0]**o+(0.595878819874)*x[0]+(0.416350660905)*x[1]**o+(0.100262997545)*x[1]
        ref[0,0,0,0]=(-0.93572957134)*x_ref[0]**o+(-0.878081515023)*x_ref[0]+(0.821839395118)*x_ref[1]**o+(-0.088026044081)*x_ref[1]
        ref[0,0,0,1]=(-0.00393438000839)*x_ref[0]**o+(-0.180693225971)*x_ref[0]+(0.192781914487)*x_ref[1]**o+(-0.561366977374)*x_ref[1]
        ref[0,0,1,0]=(0.116028558326)*x_ref[0]**o+(-0.47739282219)*x_ref[0]+(0.97806148353)*x_ref[1]**o+(0.660972051388)*x_ref[1]
        ref[0,0,1,1]=(0.092683552469)*x_ref[0]**o+(0.22083599745)*x_ref[0]+(-0.423073290439)*x_ref[1]**o+(-0.0789522599045)*x_ref[1]
        ref[0,0,2,0]=(0.140023403084)*x_ref[0]**o+(-0.927034796519)*x_ref[0]+(0.0987957525564)*x_ref[1]**o+(0.0703695028562)*x_ref[1]
        ref[0,0,2,1]=(-0.00485147064392)*x_ref[0]**o+(-0.649679595932)*x_ref[0]+(0.503435626816)*x_ref[1]**o+(0.393293792881)*x_ref[1]
        ref[0,1,0,0]=(-0.753383897772)*x_ref[0]**o+(-0.988499439357)*x_ref[0]+(-0.651037417175)*x_ref[1]**o+(0.45184892046)*x_ref[1]
        ref[0,1,0,1]=(0.0120892569605)*x_ref[0]**o+(0.52287896209)*x_ref[0]+(0.103256695817)*x_ref[1]**o+(-0.0244835681092)*x_ref[1]
        ref[0,1,1,0]=(-0.0949239075135)*x_ref[0]**o+(0.835062318231)*x_ref[0]+(-0.0522981103769)*x_ref[1]**o+(0.60020994838)*x_ref[1]
        ref[0,1,1,1]=(0.139896941482)*x_ref[0]**o+(0.409042945999)*x_ref[0]+(0.10019215394)*x_ref[1]**o+(-0.312224123017)*x_ref[1]
        ref[0,1,2,0]=(-0.892437964342)*x_ref[0]**o+(0.0828807279746)*x_ref[0]+(0.777811646235)*x_ref[1]**o+(-0.96509471481)*x_ref[1]
        ref[0,1,2,1]=(-0.397341226148)*x_ref[0]**o+(0.102987581336)*x_ref[0]+(-0.69165111598)*x_ref[1]**o+(0.743298211061)*x_ref[1]
        ref[0,2,0,0]=(-0.300129864031)*x_ref[0]**o+(-0.418981381232)*x_ref[0]+(0.624854922555)*x_ref[1]**o+(0.892651746561)*x_ref[1]
        ref[0,2,0,1]=(-0.511463603881)*x_ref[0]**o+(-0.808468332238)*x_ref[0]+(-0.342755248919)*x_ref[1]**o+(0.357985875808)*x_ref[1]
        ref[0,2,1,0]=(-0.269520452932)*x_ref[0]**o+(0.869553937081)*x_ref[0]+(-0.183222122614)*x_ref[1]**o+(-0.135280721586)*x_ref[1]
        ref[0,2,1,1]=(0.0396712123969)*x_ref[0]**o+(0.344314072793)*x_ref[0]+(0.644676507168)*x_ref[1]**o+(-0.0799551435053)*x_ref[1]
        ref[0,2,2,0]=(0.302628743334)*x_ref[0]**o+(-0.59557047376)*x_ref[0]+(-0.277654790212)*x_ref[1]**o+(0.0299035042863)*x_ref[1]
        ref[0,2,2,1]=(0.805296568248)*x_ref[0]**o+(0.726404711154)*x_ref[0]+(0.0728356962577)*x_ref[1]**o+(0.199997237962)*x_ref[1]
        ref[0,3,0,0]=(-0.292658800952)*x_ref[0]**o+(-0.40080106651)*x_ref[0]+(0.844109590553)*x_ref[1]**o+(-0.339364790362)*x_ref[1]
        ref[0,3,0,1]=(-0.621465058894)*x_ref[0]**o+(-0.376804350545)*x_ref[0]+(0.745038406103)*x_ref[1]**o+(-0.281482724757)*x_ref[1]
        ref[0,3,1,0]=(0.00625816347702)*x_ref[0]**o+(0.392750189511)*x_ref[0]+(-0.232258708605)*x_ref[1]**o+(-0.634552067635)*x_ref[1]
        ref[0,3,1,1]=(-0.0893109545391)*x_ref[0]**o+(-0.735678058768)*x_ref[0]+(-0.603248350212)*x_ref[1]**o+(-0.664366286468)*x_ref[1]
        ref[0,3,2,0]=(-0.0539114688775)*x_ref[0]**o+(-0.110100592496)*x_ref[0]+(-0.700496078261)*x_ref[1]**o+(0.203311460549)*x_ref[1]
        ref[0,3,2,1]=(-0.554691350426)*x_ref[0]**o+(-0.255001039659)*x_ref[0]+(0.567977217289)*x_ref[1]**o+(0.61993007447)*x_ref[1]
        ref[0,4,0,0]=(-0.12193490862)*x_ref[0]**o+(0.731139340661)*x_ref[0]+(-0.321971724057)*x_ref[1]**o+(0.741215588897)*x_ref[1]
        ref[0,4,0,1]=(-0.394743164913)*x_ref[0]**o+(-0.602249917929)*x_ref[0]+(0.27675734517)*x_ref[1]**o+(-0.968806169573)*x_ref[1]
        ref[0,4,1,0]=(-0.629331232169)*x_ref[0]**o+(0.687483125805)*x_ref[0]+(0.633698693186)*x_ref[1]**o+(0.837564332901)*x_ref[1]
        ref[0,4,1,1]=(0.0470601891829)*x_ref[0]**o+(-0.365490813497)*x_ref[0]+(-0.162532786979)*x_ref[1]**o+(0.504146804122)*x_ref[1]
        ref[0,4,2,0]=(0.193790444091)*x_ref[0]**o+(0.742132840869)*x_ref[0]+(-0.871677496589)*x_ref[1]**o+(0.492817750307)*x_ref[1]
        ref[0,4,2,1]=(0.741218732478)*x_ref[0]**o+(-0.746973604492)*x_ref[0]+(-0.548073707926)*x_ref[1]**o+(-0.42258185128)*x_ref[1]
        ref[1,0,0,0]=(0.580989627459)*x_ref[0]**o+(-0.498423525574)*x_ref[0]+(-0.349651580594)*x_ref[1]**o+(0.0193060779482)*x_ref[1]
        ref[1,0,0,1]=(0.398917011903)*x_ref[0]**o+(-0.122064497238)*x_ref[0]+(0.924654557299)*x_ref[1]**o+(0.485284066068)*x_ref[1]
        ref[1,0,1,0]=(-0.985949382826)*x_ref[0]**o+(-0.0188384874977)*x_ref[0]+(0.177995429697)*x_ref[1]**o+(0.967727833143)*x_ref[1]
        ref[1,0,1,1]=(0.920799980503)*x_ref[0]**o+(-0.749281119074)*x_ref[0]+(-0.628411322607)*x_ref[1]**o+(-0.446396361734)*x_ref[1]
        ref[1,0,2,0]=(0.752863123667)*x_ref[0]**o+(0.112427175329)*x_ref[0]+(0.975568251269)*x_ref[1]**o+(0.316938392669)*x_ref[1]
        ref[1,0,2,1]=(0.798409861933)*x_ref[0]**o+(-0.145715624761)*x_ref[0]+(-0.847813831553)*x_ref[1]**o+(0.302999756348)*x_ref[1]
        ref[1,1,0,0]=(0.897029838502)*x_ref[0]**o+(-0.540852276529)*x_ref[0]+(-0.541450963121)*x_ref[1]**o+(-0.118819043917)*x_ref[1]
        ref[1,1,0,1]=(-0.616797542123)*x_ref[0]**o+(-0.139394144428)*x_ref[0]+(-0.480348413917)*x_ref[1]**o+(0.500244829288)*x_ref[1]
        ref[1,1,1,0]=(-0.960388730531)*x_ref[0]**o+(0.87391147728)*x_ref[0]+(0.0128673975991)*x_ref[1]**o+(0.444341296582)*x_ref[1]
        ref[1,1,1,1]=(0.136810869097)*x_ref[0]**o+(-0.544756866447)*x_ref[0]+(-0.198052789907)*x_ref[1]**o+(-0.536732583929)*x_ref[1]
        ref[1,1,2,0]=(-0.808981135979)*x_ref[0]**o+(-0.122088197438)*x_ref[0]+(-0.363958821187)*x_ref[1]**o+(-0.521614066151)*x_ref[1]
        ref[1,1,2,1]=(0.299382605072)*x_ref[0]**o+(0.143397060894)*x_ref[0]+(-0.223044874711)*x_ref[1]**o+(-0.04723866032)*x_ref[1]
        ref[1,2,0,0]=(-0.568967551455)*x_ref[0]**o+(0.750777908696)*x_ref[0]+(0.203117458853)*x_ref[1]**o+(-0.0797168995572)*x_ref[1]
        ref[1,2,0,1]=(0.0439147552635)*x_ref[0]**o+(-0.0746749889653)*x_ref[0]+(0.192885940929)*x_ref[1]**o+(0.794090802345)*x_ref[1]
        ref[1,2,1,0]=(0.137177876023)*x_ref[0]**o+(0.625847668003)*x_ref[0]+(0.252761858157)*x_ref[1]**o+(0.496512978117)*x_ref[1]
        ref[1,2,1,1]=(0.475218888846)*x_ref[0]**o+(-0.286926184073)*x_ref[0]+(0.551427803487)*x_ref[1]**o+(0.782373661286)*x_ref[1]
        ref[1,2,2,0]=(0.903004502545)*x_ref[0]**o+(-0.262952251654)*x_ref[0]+(0.110351860812)*x_ref[1]**o+(-0.685355934696)*x_ref[1]
        ref[1,2,2,1]=(-0.157457317404)*x_ref[0]**o+(-0.243850623177)*x_ref[0]+(0.28853283432)*x_ref[1]**o+(0.612508908614)*x_ref[1]
        ref[1,3,0,0]=(-0.741890002973)*x_ref[0]**o+(-0.118262424825)*x_ref[0]+(0.567100734646)*x_ref[1]**o+(0.301557125811)*x_ref[1]
        ref[1,3,0,1]=(-0.00745396583928)*x_ref[0]**o+(-0.00558491344276)*x_ref[0]+(-0.967132370197)*x_ref[1]**o+(-0.61737594402)*x_ref[1]
        ref[1,3,1,0]=(0.0767961523966)*x_ref[0]**o+(0.341952000996)*x_ref[0]+(0.762617832019)*x_ref[1]**o+(-0.38337284713)*x_ref[1]
        ref[1,3,1,1]=(0.32265018902)*x_ref[0]**o+(-0.352521736206)*x_ref[0]+(-0.0119143186529)*x_ref[1]**o+(-0.64727561829)*x_ref[1]
        ref[1,3,2,0]=(0.627690584587)*x_ref[0]**o+(0.0993819822193)*x_ref[0]+(-0.189363448841)*x_ref[1]**o+(0.49004001194)*x_ref[1]
        ref[1,3,2,1]=(-0.657836476156)*x_ref[0]**o+(0.708970470136)*x_ref[0]+(0.252190020576)*x_ref[1]**o+(-0.481214196298)*x_ref[1]
        ref[1,4,0,0]=(0.0837563164179)*x_ref[0]**o+(0.741536525079)*x_ref[0]+(0.645543094286)*x_ref[1]**o+(0.62358090815)*x_ref[1]
        ref[1,4,0,1]=(0.262043398371)*x_ref[0]**o+(-0.47794750936)*x_ref[0]+(0.93086244965)*x_ref[1]**o+(-0.670399062677)*x_ref[1]
        ref[1,4,1,0]=(-0.76634902923)*x_ref[0]**o+(-0.809846143928)*x_ref[0]+(0.639099549305)*x_ref[1]**o+(0.793524639236)*x_ref[1]
        ref[1,4,1,1]=(0.814737503957)*x_ref[0]**o+(0.775593374042)*x_ref[0]+(0.240976904111)*x_ref[1]**o+(0.2352421356)*x_ref[1]
        ref[1,4,2,0]=(0.519210584047)*x_ref[0]**o+(-0.702579589691)*x_ref[0]+(-0.450156317853)*x_ref[1]**o+(0.588864692581)*x_ref[1]
        ref[1,4,2,1]=(-0.927751546037)*x_ref[0]**o+(-0.239737576019)*x_ref[0]+(-0.527658440447)*x_ref[1]**o+(0.79623440559)*x_ref[1]
        ref[2,0,0,0]=(0.719144336419)*x_ref[0]**o+(0.916780231277)*x_ref[0]+(0.279073450277)*x_ref[1]**o+(-0.0130671321167)*x_ref[1]
        ref[2,0,0,1]=(-0.556604998962)*x_ref[0]**o+(-0.661580370137)*x_ref[0]+(0.980704664791)*x_ref[1]**o+(0.935843251131)*x_ref[1]
        ref[2,0,1,0]=(0.85501679753)*x_ref[0]**o+(-0.908342467889)*x_ref[0]+(0.969030151294)*x_ref[1]**o+(0.534192961922)*x_ref[1]
        ref[2,0,1,1]=(0.527672385529)*x_ref[0]**o+(0.997699955667)*x_ref[0]+(-0.388249629327)*x_ref[1]**o+(-0.714956937512)*x_ref[1]
        ref[2,0,2,0]=(-0.629149810667)*x_ref[0]**o+(-0.0515018376692)*x_ref[0]+(-0.80355991702)*x_ref[1]**o+(0.235999988886)*x_ref[1]
        ref[2,0,2,1]=(0.960715934156)*x_ref[0]**o+(-0.782787305497)*x_ref[0]+(-0.309348231959)*x_ref[1]**o+(0.854911294717)*x_ref[1]
        ref[2,1,0,0]=(-0.560749125074)*x_ref[0]**o+(-0.120515615757)*x_ref[0]+(0.055090634431)*x_ref[1]**o+(-0.539205949358)*x_ref[1]
        ref[2,1,0,1]=(-0.762113417295)*x_ref[0]**o+(-0.207196980376)*x_ref[0]+(0.904541245494)*x_ref[1]**o+(0.320359459939)*x_ref[1]
        ref[2,1,1,0]=(-0.272627586486)*x_ref[0]**o+(-0.885987982522)*x_ref[0]+(0.641737808216)*x_ref[1]**o+(0.358390367373)*x_ref[1]
        ref[2,1,1,1]=(-0.847502563761)*x_ref[0]**o+(-0.430298277492)*x_ref[0]+(-0.869043300185)*x_ref[1]**o+(0.584700145915)*x_ref[1]
        ref[2,1,2,0]=(0.497688368946)*x_ref[0]**o+(-0.14413770704)*x_ref[0]+(-0.82002941888)*x_ref[1]**o+(0.573569419279)*x_ref[1]
        ref[2,1,2,1]=(0.959549623603)*x_ref[0]**o+(0.881196513588)*x_ref[0]+(-0.202951096266)*x_ref[1]**o+(-9.87983182765e-05)*x_ref[1]
        ref[2,2,0,0]=(0.0811393810501)*x_ref[0]**o+(-0.853555165896)*x_ref[0]+(0.0259257246335)*x_ref[1]**o+(0.638575306959)*x_ref[1]
        ref[2,2,0,1]=(-0.525990031379)*x_ref[0]**o+(0.11235160765)*x_ref[0]+(-0.502853879455)*x_ref[1]**o+(0.414066683507)*x_ref[1]
        ref[2,2,1,0]=(-0.533919716834)*x_ref[0]**o+(-0.773411100656)*x_ref[0]+(0.311424028491)*x_ref[1]**o+(0.332733301377)*x_ref[1]
        ref[2,2,1,1]=(0.280548593462)*x_ref[0]**o+(-0.0747640762104)*x_ref[0]+(-0.7110232877)*x_ref[1]**o+(-0.178672817441)*x_ref[1]
        ref[2,2,2,0]=(0.445025926846)*x_ref[0]**o+(-0.669782903005)*x_ref[0]+(0.236610411793)*x_ref[1]**o+(0.437836936476)*x_ref[1]
        ref[2,2,2,1]=(-0.503761305098)*x_ref[0]**o+(-0.889922659865)*x_ref[0]+(-0.0767256016402)*x_ref[1]**o+(-0.614874217835)*x_ref[1]
        ref[2,3,0,0]=(-0.041960855672)*x_ref[0]**o+(-0.593486242764)*x_ref[0]+(0.344751892525)*x_ref[1]**o+(-0.131268873786)*x_ref[1]
        ref[2,3,0,1]=(-0.192704843083)*x_ref[0]**o+(0.325356381696)*x_ref[0]+(-0.235690651937)*x_ref[1]**o+(0.59113535617)*x_ref[1]
        ref[2,3,1,0]=(-0.0794745628024)*x_ref[0]**o+(0.7996737586)*x_ref[0]+(-0.324970505375)*x_ref[1]**o+(-0.14920403468)*x_ref[1]
        ref[2,3,1,1]=(0.31997432673)*x_ref[0]**o+(-0.571739807799)*x_ref[0]+(0.517683480263)*x_ref[1]**o+(-0.250680741704)*x_ref[1]
        ref[2,3,2,0]=(-0.379114277883)*x_ref[0]**o+(0.953285017352)*x_ref[0]+(-0.701111847548)*x_ref[1]**o+(-0.354199942528)*x_ref[1]
        ref[2,3,2,1]=(0.492245629486)*x_ref[0]**o+(-0.299380102293)*x_ref[0]+(-0.126371826833)*x_ref[1]**o+(0.933664060647)*x_ref[1]
        ref[2,4,0,0]=(0.714822159533)*x_ref[0]**o+(0.359430380478)*x_ref[0]+(-0.691610768506)*x_ref[1]**o+(0.414061853203)*x_ref[1]
        ref[2,4,0,1]=(0.412799253275)*x_ref[0]**o+(0.862905533148)*x_ref[0]+(-0.0507440946561)*x_ref[1]**o+(0.779012369976)*x_ref[1]
        ref[2,4,1,0]=(-0.981179053034)*x_ref[0]**o+(0.984777005573)*x_ref[0]+(0.715584788452)*x_ref[1]**o+(0.920666888724)*x_ref[1]
        ref[2,4,1,1]=(-0.0287775685032)*x_ref[0]**o+(0.617592345344)*x_ref[0]+(-0.767451297794)*x_ref[1]**o+(0.95828976091)*x_ref[1]
        ref[2,4,2,0]=(0.905037473705)*x_ref[0]**o+(-0.706987158021)*x_ref[0]+(0.535436652493)*x_ref[1]**o+(-0.891712664249)*x_ref[1]
        ref[2,4,2,1]=(-0.818709474925)*x_ref[0]**o+(-0.0223479273553)*x_ref[0]+(-0.370644568884)*x_ref[1]**o+(0.286468739105)*x_ref[1]
        ref[3,0,0,0]=(0.310461949876)*x_ref[0]**o+(0.770193446516)*x_ref[0]+(-0.0858335164699)*x_ref[1]**o+(0.12572608122)*x_ref[1]
        ref[3,0,0,1]=(-0.781332422249)*x_ref[0]**o+(-0.238368527872)*x_ref[0]+(-0.449272256582)*x_ref[1]**o+(-0.645638738212)*x_ref[1]
        ref[3,0,1,0]=(-0.91143447678)*x_ref[0]**o+(-0.885131081008)*x_ref[0]+(-0.253833463581)*x_ref[1]**o+(0.051439649459)*x_ref[1]
        ref[3,0,1,1]=(0.447730169203)*x_ref[0]**o+(0.980354452325)*x_ref[0]+(-0.0484566141047)*x_ref[1]**o+(0.443407604362)*x_ref[1]
        ref[3,0,2,0]=(-0.223975544426)*x_ref[0]**o+(0.0141796274312)*x_ref[0]+(0.664080206442)*x_ref[1]**o+(0.193318634886)*x_ref[1]
        ref[3,0,2,1]=(-0.388901068755)*x_ref[0]**o+(-0.632210208163)*x_ref[0]+(0.567264328931)*x_ref[1]**o+(-0.308710733281)*x_ref[1]
        ref[3,1,0,0]=(0.259562910069)*x_ref[0]**o+(-0.315409720689)*x_ref[0]+(0.999281841483)*x_ref[1]**o+(0.695981779357)*x_ref[1]
        ref[3,1,0,1]=(0.627508253716)*x_ref[0]**o+(-0.540950954632)*x_ref[0]+(0.626209667445)*x_ref[1]**o+(-0.920403595109)*x_ref[1]
        ref[3,1,1,0]=(0.750166528891)*x_ref[0]**o+(-0.145494783124)*x_ref[0]+(0.71048043416)*x_ref[1]**o+(-0.0920505950401)*x_ref[1]
        ref[3,1,1,1]=(-0.755570630755)*x_ref[0]**o+(-0.323456877629)*x_ref[0]+(-0.0318418288689)*x_ref[1]**o+(0.551823313307)*x_ref[1]
        ref[3,1,2,0]=(-0.301380276426)*x_ref[0]**o+(-0.102899117196)*x_ref[0]+(0.816857028902)*x_ref[1]**o+(-0.571555621895)*x_ref[1]
        ref[3,1,2,1]=(-0.791148729922)*x_ref[0]**o+(0.400286696027)*x_ref[0]+(-0.17362002139)*x_ref[1]**o+(0.431696064424)*x_ref[1]
        ref[3,2,0,0]=(0.241944451237)*x_ref[0]**o+(-0.629401496412)*x_ref[0]+(-0.00333636607647)*x_ref[1]**o+(0.167334133677)*x_ref[1]
        ref[3,2,0,1]=(-0.300944831932)*x_ref[0]**o+(0.358282773504)*x_ref[0]+(0.759495661081)*x_ref[1]**o+(-0.225250488484)*x_ref[1]
        ref[3,2,1,0]=(0.755950588768)*x_ref[0]**o+(0.718423180013)*x_ref[0]+(0.705150216505)*x_ref[1]**o+(0.269283969196)*x_ref[1]
        ref[3,2,1,1]=(0.811064150458)*x_ref[0]**o+(-0.67847842696)*x_ref[0]+(-0.478091805711)*x_ref[1]**o+(0.292774429011)*x_ref[1]
        ref[3,2,2,0]=(-0.307877253758)*x_ref[0]**o+(-0.563961673764)*x_ref[0]+(0.428322493356)*x_ref[1]**o+(0.564230126489)*x_ref[1]
        ref[3,2,2,1]=(-0.623877056979)*x_ref[0]**o+(-0.148337578412)*x_ref[0]+(-0.910662605581)*x_ref[1]**o+(-0.38603493572)*x_ref[1]
        ref[3,3,0,0]=(-0.826474270729)*x_ref[0]**o+(0.977614689932)*x_ref[0]+(-0.369824837717)*x_ref[1]**o+(-0.985478011812)*x_ref[1]
        ref[3,3,0,1]=(-0.251252769561)*x_ref[0]**o+(-0.296732765057)*x_ref[0]+(-0.888899418995)*x_ref[1]**o+(0.640313055734)*x_ref[1]
        ref[3,3,1,0]=(-0.189106218841)*x_ref[0]**o+(-0.0441656339777)*x_ref[0]+(0.368716219801)*x_ref[1]**o+(-0.924144919367)*x_ref[1]
        ref[3,3,1,1]=(-0.525205981948)*x_ref[0]**o+(-0.148352446593)*x_ref[0]+(0.869212821652)*x_ref[1]**o+(-0.575463559793)*x_ref[1]
        ref[3,3,2,0]=(0.668772663439)*x_ref[0]**o+(0.4330630456)*x_ref[0]+(-0.321158907274)*x_ref[1]**o+(0.570853139693)*x_ref[1]
        ref[3,3,2,1]=(0.0236674031998)*x_ref[0]**o+(-0.390106024651)*x_ref[0]+(0.541717893494)*x_ref[1]**o+(-0.144655705849)*x_ref[1]
        ref[3,4,0,0]=(0.971861188602)*x_ref[0]**o+(-0.834463063474)*x_ref[0]+(-0.419545662218)*x_ref[1]**o+(-0.0387341522049)*x_ref[1]
        ref[3,4,0,1]=(0.226357589896)*x_ref[0]**o+(0.305148895256)*x_ref[0]+(0.00306527642431)*x_ref[1]**o+(0.4715619746)*x_ref[1]
        ref[3,4,1,0]=(0.529602340964)*x_ref[0]**o+(-0.482050910795)*x_ref[0]+(0.964983911769)*x_ref[1]**o+(0.524018220532)*x_ref[1]
        ref[3,4,1,1]=(0.940676158872)*x_ref[0]**o+(0.547078925647)*x_ref[0]+(-0.636907424453)*x_ref[1]**o+(0.938262396281)*x_ref[1]
        ref[3,4,2,0]=(0.282512784255)*x_ref[0]**o+(-0.556345478931)*x_ref[0]+(0.651866614458)*x_ref[1]**o+(0.8895132214)*x_ref[1]
        ref[3,4,2,1]=(0.905077434472)*x_ref[0]**o+(0.595878819874)*x_ref[0]+(0.416350660905)*x_ref[1]**o+(0.100262997545)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.836302395465)*x[0]**o+(-0.842568675604)*x[0]+(-0.780133441207)*x[1]**o+(0.54483947113)*x[1]+(-0.130253019696)*x[2]**o+(0.358303389828)*x[2]
        arg[0,0,0,1]=(-0.722847500652)*x[0]**o+(0.42887368878)*x[0]+(0.979187056933)*x[1]**o+(0.624875714982)*x[1]+(-0.672529765247)*x[2]**o+(0.822363583702)*x[2]
        arg[0,0,1,0]=(-0.28989650691)*x[0]**o+(-0.847647954773)*x[0]+(-0.11119862113)*x[1]**o+(0.0809009676818)*x[1]+(-0.672626533587)*x[2]**o+(0.580561232267)*x[2]
        arg[0,0,1,1]=(-0.64783363248)*x[0]**o+(0.212828552873)*x[0]+(-0.845297180936)*x[1]**o+(-0.714760411186)*x[1]+(-0.385173605122)*x[2]**o+(-0.760750317556)*x[2]
        arg[0,0,2,0]=(-0.328757006046)*x[0]**o+(-0.0935907722247)*x[0]+(-0.895866135795)*x[1]**o+(-0.701241976745)*x[1]+(0.893684139941)*x[2]**o+(-0.265339318222)*x[2]
        arg[0,0,2,1]=(0.134066662346)*x[0]**o+(0.0383231508856)*x[0]+(-0.347744261995)*x[1]**o+(-0.86897907336)*x[1]+(-0.287792154521)*x[2]**o+(0.859351846891)*x[2]
        arg[0,1,0,0]=(-0.635128124936)*x[0]**o+(-0.659462189733)*x[0]+(-0.0964365747936)*x[1]**o+(-0.385702168747)*x[1]+(0.179539914057)*x[2]**o+(-0.0176420847859)*x[2]
        arg[0,1,0,1]=(-0.580750807831)*x[0]**o+(0.756493147583)*x[0]+(-0.436842251289)*x[1]**o+(-0.705624646095)*x[1]+(0.983064096262)*x[2]**o+(-0.0197153825423)*x[2]
        arg[0,1,1,0]=(-0.567748582365)*x[0]**o+(0.579324420356)*x[0]+(0.0434168493419)*x[1]**o+(-0.839692920334)*x[1]+(-0.119239847745)*x[2]**o+(0.527047942834)*x[2]
        arg[0,1,1,1]=(0.52870403655)*x[0]**o+(0.700090892471)*x[0]+(0.730224313921)*x[1]**o+(-0.815451751207)*x[1]+(-0.943945382413)*x[2]**o+(-0.571714141916)*x[2]
        arg[0,1,2,0]=(-0.803841094887)*x[0]**o+(0.0595126006237)*x[0]+(-0.612341797645)*x[1]**o+(0.944552805211)*x[1]+(-0.00253009731786)*x[2]**o+(0.0174867439912)*x[2]
        arg[0,1,2,1]=(-0.145623212549)*x[0]**o+(0.546039957008)*x[0]+(-0.221425237579)*x[1]**o+(0.229367671159)*x[1]+(-0.784903869759)*x[2]**o+(0.898045208871)*x[2]
        arg[0,2,0,0]=(0.417924494005)*x[0]**o+(-0.192820051987)*x[0]+(0.878916271492)*x[1]**o+(-0.35056320984)*x[1]+(0.982897575604)*x[2]**o+(-0.225793730616)*x[2]
        arg[0,2,0,1]=(-0.493279498849)*x[0]**o+(-0.154177411087)*x[0]+(-0.923592860492)*x[1]**o+(0.872010949252)*x[1]+(-0.936252147064)*x[2]**o+(0.947136819308)*x[2]
        arg[0,2,1,0]=(0.109357630607)*x[0]**o+(-0.518825477709)*x[0]+(-0.375510667498)*x[1]**o+(0.476537999644)*x[1]+(0.676017421029)*x[2]**o+(-0.633301158984)*x[2]
        arg[0,2,1,1]=(-0.0662265225539)*x[0]**o+(-0.276634748263)*x[0]+(-0.0655431981102)*x[1]**o+(0.0578482920991)*x[1]+(-0.264370611185)*x[2]**o+(-0.674612985594)*x[2]
        arg[0,2,2,0]=(-0.924186592743)*x[0]**o+(0.197603043149)*x[0]+(-0.676882805046)*x[1]**o+(0.564142845865)*x[1]+(0.590369345306)*x[2]**o+(-0.872019490978)*x[2]
        arg[0,2,2,1]=(0.480683776968)*x[0]**o+(-0.412055284173)*x[0]+(0.829996313473)*x[1]**o+(-0.235665473048)*x[1]+(-0.429191305284)*x[2]**o+(-0.0867804318457)*x[2]
        arg[0,3,0,0]=(-0.0903660006043)*x[0]**o+(-0.507571901766)*x[0]+(-0.0305305861809)*x[1]**o+(-0.915027969989)*x[1]+(0.955644774347)*x[2]**o+(0.739244618775)*x[2]
        arg[0,3,0,1]=(0.600619122998)*x[0]**o+(0.0128699484527)*x[0]+(-0.360431921281)*x[1]**o+(0.304773619811)*x[1]+(-0.302423314054)*x[2]**o+(0.680149401572)*x[2]
        arg[0,3,1,0]=(-0.638540513323)*x[0]**o+(0.766655754067)*x[0]+(0.276879854223)*x[1]**o+(-0.628688275247)*x[1]+(0.724584107114)*x[2]**o+(0.682193551967)*x[2]
        arg[0,3,1,1]=(-0.516656801349)*x[0]**o+(-0.513901319383)*x[0]+(-0.227957010777)*x[1]**o+(-0.164363207703)*x[1]+(-0.919936042047)*x[2]**o+(0.682400184997)*x[2]
        arg[0,3,2,0]=(-0.819629653095)*x[0]**o+(0.183937049274)*x[0]+(0.0411075800579)*x[1]**o+(0.0813088270699)*x[1]+(0.762416334575)*x[2]**o+(0.236553726705)*x[2]
        arg[0,3,2,1]=(0.513141183759)*x[0]**o+(-0.452475512734)*x[0]+(0.0461424221984)*x[1]**o+(-0.673452844636)*x[1]+(-0.194493328878)*x[2]**o+(0.812459956422)*x[2]
        arg[0,4,0,0]=(0.672167415195)*x[0]**o+(-0.641004730405)*x[0]+(-0.383031420851)*x[1]**o+(0.392366431186)*x[1]+(0.66479124224)*x[2]**o+(0.845503225104)*x[2]
        arg[0,4,0,1]=(0.908241966695)*x[0]**o+(-0.816989438726)*x[0]+(-0.144902267583)*x[1]**o+(-0.654901540598)*x[1]+(-0.9840470269)*x[2]**o+(0.510970695154)*x[2]
        arg[0,4,1,0]=(0.788858704366)*x[0]**o+(-0.50266306504)*x[0]+(-0.681248796723)*x[1]**o+(0.0317663112509)*x[1]+(0.662428035619)*x[2]**o+(0.377116420912)*x[2]
        arg[0,4,1,1]=(-0.0734522829037)*x[0]**o+(0.0822741484655)*x[0]+(-0.509279578815)*x[1]**o+(-0.95372420098)*x[1]+(0.682868362508)*x[2]**o+(-0.947156883638)*x[2]
        arg[0,4,2,0]=(-0.843236607337)*x[0]**o+(-0.925326238513)*x[0]+(-0.153688519983)*x[1]**o+(-0.68456413487)*x[1]+(0.920442617368)*x[2]**o+(0.852945557893)*x[2]
        arg[0,4,2,1]=(-0.283772512528)*x[0]**o+(0.202730738521)*x[0]+(-0.287375172356)*x[1]**o+(-0.176699132816)*x[1]+(0.446391113264)*x[2]**o+(-0.783266847348)*x[2]
        arg[1,0,0,0]=(0.81284364562)*x[0]**o+(-0.454432332505)*x[0]+(-0.684692596262)*x[1]**o+(-0.345923888237)*x[1]+(0.1201728631)*x[2]**o+(0.808246135028)*x[2]
        arg[1,0,0,1]=(0.941104328706)*x[0]**o+(-0.802008654964)*x[0]+(0.902047483677)*x[1]**o+(0.777049224612)*x[1]+(0.120184332854)*x[2]**o+(0.983216899102)*x[2]
        arg[1,0,1,0]=(-0.420996925767)*x[0]**o+(0.0941750615502)*x[0]+(-0.935342933578)*x[1]**o+(0.925891822058)*x[1]+(-0.153652985582)*x[2]**o+(-0.336548986153)*x[2]
        arg[1,0,1,1]=(-0.692886221795)*x[0]**o+(-0.920419410139)*x[0]+(-0.586575583263)*x[1]**o+(-0.269347362009)*x[1]+(-0.872169736045)*x[2]**o+(-0.724727186336)*x[2]
        arg[1,0,2,0]=(-0.0171222282058)*x[0]**o+(0.338180602976)*x[0]+(-0.0345958505395)*x[1]**o+(0.631728421505)*x[1]+(0.90596443453)*x[2]**o+(-0.144785193015)*x[2]
        arg[1,0,2,1]=(-0.91636583885)*x[0]**o+(-0.807496789574)*x[0]+(0.63472495957)*x[1]**o+(0.966464343993)*x[1]+(-0.471311384621)*x[2]**o+(-0.449420138238)*x[2]
        arg[1,1,0,0]=(-0.239930503474)*x[0]**o+(-0.476439715994)*x[0]+(0.421600511888)*x[1]**o+(0.477823144792)*x[1]+(0.251409914807)*x[2]**o+(0.636818001419)*x[2]
        arg[1,1,0,1]=(-0.427510410263)*x[0]**o+(0.494810857516)*x[0]+(-0.129415181352)*x[1]**o+(0.754147770071)*x[1]+(0.0218776365985)*x[2]**o+(0.706183216616)*x[2]
        arg[1,1,1,0]=(-0.469150375165)*x[0]**o+(-0.0651101638073)*x[0]+(-0.697853003861)*x[1]**o+(0.0389330102026)*x[1]+(-0.256557942045)*x[2]**o+(0.409565340425)*x[2]
        arg[1,1,1,1]=(0.14957005608)*x[0]**o+(-0.209482968126)*x[0]+(0.368899650309)*x[1]**o+(-0.0899321000137)*x[1]+(0.282389903357)*x[2]**o+(-0.166633883916)*x[2]
        arg[1,1,2,0]=(0.0374938231714)*x[0]**o+(-0.687972899893)*x[0]+(-0.0686478107779)*x[1]**o+(0.318977937888)*x[1]+(-0.601269126788)*x[2]**o+(0.453188855786)*x[2]
        arg[1,1,2,1]=(-0.853894223222)*x[0]**o+(0.441830151478)*x[0]+(0.973114898368)*x[1]**o+(-0.264454889071)*x[1]+(-0.828332687531)*x[2]**o+(0.302493748565)*x[2]
        arg[1,2,0,0]=(0.210365824111)*x[0]**o+(0.4674117857)*x[0]+(-0.129546969044)*x[1]**o+(0.566357125352)*x[1]+(-0.721733448514)*x[2]**o+(0.154253207106)*x[2]
        arg[1,2,0,1]=(0.805325508181)*x[0]**o+(-0.929113563452)*x[0]+(-0.487301761321)*x[1]**o+(-0.822405526519)*x[1]+(0.254754490326)*x[2]**o+(-0.196480710709)*x[2]
        arg[1,2,1,0]=(-0.520777275475)*x[0]**o+(-0.912569745644)*x[0]+(-0.784504791087)*x[1]**o+(-0.903473217653)*x[1]+(0.583297107374)*x[2]**o+(0.664811898195)*x[2]
        arg[1,2,1,1]=(0.129152152958)*x[0]**o+(0.497832200867)*x[0]+(0.121217248234)*x[1]**o+(-0.183368066714)*x[1]+(-0.776145485536)*x[2]**o+(-0.223165724715)*x[2]
        arg[1,2,2,0]=(-0.698981324488)*x[0]**o+(0.00951277777924)*x[0]+(0.997410976507)*x[1]**o+(0.534670303453)*x[1]+(0.251309223218)*x[2]**o+(-0.921148646979)*x[2]
        arg[1,2,2,1]=(-0.731509934487)*x[0]**o+(0.87731926834)*x[0]+(0.0865507713913)*x[1]**o+(0.708835987019)*x[1]+(0.294182269171)*x[2]**o+(-0.316660750729)*x[2]
        arg[1,3,0,0]=(-0.282903962918)*x[0]**o+(-0.793457978844)*x[0]+(0.287425807914)*x[1]**o+(-0.0252722364835)*x[1]+(-0.956202022515)*x[2]**o+(0.650271579135)*x[2]
        arg[1,3,0,1]=(-0.946723111202)*x[0]**o+(0.669431653693)*x[0]+(-0.791125976188)*x[1]**o+(-0.466484914794)*x[1]+(-0.217225801872)*x[2]**o+(0.224719473996)*x[2]
        arg[1,3,1,0]=(0.516479301043)*x[0]**o+(-0.42420332366)*x[0]+(-0.0649227247176)*x[1]**o+(0.208849752675)*x[1]+(-0.939001099693)*x[2]**o+(-0.0537827140932)*x[2]
        arg[1,3,1,1]=(-0.292077069459)*x[0]**o+(-0.444848119744)*x[0]+(0.268592719199)*x[1]**o+(-0.349401060323)*x[1]+(-0.540803784489)*x[2]**o+(-0.487170123526)*x[2]
        arg[1,3,2,0]=(-0.626222042661)*x[0]**o+(0.245937409624)*x[0]+(-0.24692162006)*x[1]**o+(0.108230029208)*x[1]+(-0.760463241928)*x[2]**o+(-0.587078263201)*x[2]
        arg[1,3,2,1]=(-0.923891147756)*x[0]**o+(0.0961285842162)*x[0]+(0.0146853466585)*x[1]**o+(0.00951339122359)*x[1]+(-0.261383704723)*x[2]**o+(-0.0262515691324)*x[2]
        arg[1,4,0,0]=(-0.0370003524292)*x[0]**o+(-0.0463848489957)*x[0]+(0.940061458694)*x[1]**o+(0.592787058925)*x[1]+(-0.217284379638)*x[2]**o+(-0.11746834109)*x[2]
        arg[1,4,0,1]=(0.830404283093)*x[0]**o+(0.268135490403)*x[0]+(-0.894500398195)*x[1]**o+(0.751042648917)*x[1]+(0.687583292754)*x[2]**o+(-0.884596509742)*x[2]
        arg[1,4,1,0]=(0.414764592469)*x[0]**o+(-0.529613347919)*x[0]+(-0.727435115594)*x[1]**o+(-0.546579777725)*x[1]+(0.378011893971)*x[2]**o+(-0.578081183238)*x[2]
        arg[1,4,1,1]=(0.406513376747)*x[0]**o+(-0.350132869579)*x[0]+(-0.557954631299)*x[1]**o+(0.868176175754)*x[1]+(0.793021832632)*x[2]**o+(0.280529771851)*x[2]
        arg[1,4,2,0]=(0.56741953575)*x[0]**o+(0.0115947684302)*x[0]+(0.872812725363)*x[1]**o+(-0.215183541745)*x[1]+(0.64853692203)*x[2]**o+(-0.139561471922)*x[2]
        arg[1,4,2,1]=(0.650916904666)*x[0]**o+(0.878765271839)*x[0]+(0.0141264348052)*x[1]**o+(-0.56314426774)*x[1]+(-0.267854377704)*x[2]**o+(0.852271380667)*x[2]
        arg[2,0,0,0]=(0.948359141577)*x[0]**o+(-0.0881747280974)*x[0]+(0.804871563368)*x[1]**o+(-0.0298585763446)*x[1]+(0.506923333569)*x[2]**o+(-0.920947093648)*x[2]
        arg[2,0,0,1]=(-0.0144221258584)*x[0]**o+(-0.679723072824)*x[0]+(0.00829336956687)*x[1]**o+(0.675680809303)*x[1]+(-0.803195231536)*x[2]**o+(0.70240358571)*x[2]
        arg[2,0,1,0]=(-0.159174987706)*x[0]**o+(0.922309271483)*x[0]+(0.344354343035)*x[1]**o+(-0.00954038311323)*x[1]+(-0.50225452804)*x[2]**o+(0.355899734308)*x[2]
        arg[2,0,1,1]=(0.0857611780404)*x[0]**o+(0.876479431983)*x[0]+(0.105049158502)*x[1]**o+(-0.530308461426)*x[1]+(-0.553136649343)*x[2]**o+(0.385458516481)*x[2]
        arg[2,0,2,0]=(-0.260086070842)*x[0]**o+(0.417766241499)*x[0]+(-0.292370383673)*x[1]**o+(-0.321315349553)*x[1]+(-0.437301444082)*x[2]**o+(0.730845669715)*x[2]
        arg[2,0,2,1]=(0.132547674276)*x[0]**o+(0.783602886333)*x[0]+(0.822361450413)*x[1]**o+(-0.810402783)*x[1]+(0.759701191023)*x[2]**o+(0.644375788608)*x[2]
        arg[2,1,0,0]=(-0.841458111227)*x[0]**o+(0.779917641736)*x[0]+(-0.0745587217256)*x[1]**o+(0.938545280143)*x[1]+(0.454600163373)*x[2]**o+(-0.490348377697)*x[2]
        arg[2,1,0,1]=(-0.876477136174)*x[0]**o+(-0.248200369939)*x[0]+(-0.973654243816)*x[1]**o+(0.482402123794)*x[1]+(-0.776970490827)*x[2]**o+(-0.571349467703)*x[2]
        arg[2,1,1,0]=(-0.433528193102)*x[0]**o+(-0.57712915767)*x[0]+(-0.328671198058)*x[1]**o+(-0.926873019075)*x[1]+(0.314232739356)*x[2]**o+(-0.199625466412)*x[2]
        arg[2,1,1,1]=(0.892707770392)*x[0]**o+(-0.680081898234)*x[0]+(-0.754181058483)*x[1]**o+(-0.253728945575)*x[1]+(-0.906373406684)*x[2]**o+(-0.0266435080833)*x[2]
        arg[2,1,2,0]=(-0.121605378472)*x[0]**o+(-0.503955015634)*x[0]+(-0.0960767196114)*x[1]**o+(-0.937914890361)*x[1]+(0.212042931853)*x[2]**o+(0.930703155408)*x[2]
        arg[2,1,2,1]=(0.0564159441047)*x[0]**o+(0.842332721105)*x[0]+(0.954741657608)*x[1]**o+(0.73503176)*x[1]+(-0.390679938582)*x[2]**o+(-0.440014747741)*x[2]
        arg[2,2,0,0]=(0.0100637970821)*x[0]**o+(-0.0283353154145)*x[0]+(0.865360758262)*x[1]**o+(0.780975027988)*x[1]+(0.16115842919)*x[2]**o+(0.247580155656)*x[2]
        arg[2,2,0,1]=(0.61102059426)*x[0]**o+(-0.198005744101)*x[0]+(0.503568797262)*x[1]**o+(0.0288809726497)*x[1]+(-0.765127098183)*x[2]**o+(0.409803857723)*x[2]
        arg[2,2,1,0]=(-0.712272018295)*x[0]**o+(-0.62562142063)*x[0]+(-0.579501949719)*x[1]**o+(0.628539275236)*x[1]+(0.0200945124964)*x[2]**o+(0.175966024769)*x[2]
        arg[2,2,1,1]=(-0.798244445519)*x[0]**o+(-0.376440962535)*x[0]+(-0.134467401497)*x[1]**o+(0.877538294477)*x[1]+(-0.484870741192)*x[2]**o+(-0.224064469935)*x[2]
        arg[2,2,2,0]=(0.200196926368)*x[0]**o+(-0.528000867086)*x[0]+(-0.708942342624)*x[1]**o+(-0.591994969141)*x[1]+(-0.990641592656)*x[2]**o+(-0.875897347303)*x[2]
        arg[2,2,2,1]=(0.949113190797)*x[0]**o+(0.867855496786)*x[0]+(-0.0314953438065)*x[1]**o+(-0.371928428005)*x[1]+(-0.526403191063)*x[2]**o+(0.653049122326)*x[2]
        arg[2,3,0,0]=(-0.353418327127)*x[0]**o+(-0.418588306429)*x[0]+(0.400872043269)*x[1]**o+(0.141907924042)*x[1]+(-0.999869579462)*x[2]**o+(-0.410423954563)*x[2]
        arg[2,3,0,1]=(-0.240390877307)*x[0]**o+(0.0901879916115)*x[0]+(-0.0734683209657)*x[1]**o+(0.274836005886)*x[1]+(-0.336375498181)*x[2]**o+(-0.754776455221)*x[2]
        arg[2,3,1,0]=(0.259928460822)*x[0]**o+(-0.227630454708)*x[0]+(0.725968066278)*x[1]**o+(-0.545440695763)*x[1]+(-0.0354926607815)*x[2]**o+(0.365010532181)*x[2]
        arg[2,3,1,1]=(-0.969799373398)*x[0]**o+(0.514498233214)*x[0]+(0.907450925318)*x[1]**o+(-0.234993706406)*x[1]+(0.206323910806)*x[2]**o+(0.209364250369)*x[2]
        arg[2,3,2,0]=(0.712905565941)*x[0]**o+(-0.353322758263)*x[0]+(0.816861181494)*x[1]**o+(-0.362388243619)*x[1]+(0.302543517607)*x[2]**o+(-0.831280418852)*x[2]
        arg[2,3,2,1]=(-0.386373711182)*x[0]**o+(0.946914567792)*x[0]+(-0.52795085887)*x[1]**o+(0.86635245964)*x[1]+(0.654261054178)*x[2]**o+(0.452334882498)*x[2]
        arg[2,4,0,0]=(0.374747033547)*x[0]**o+(-0.3471578669)*x[0]+(0.322217002639)*x[1]**o+(0.92988068982)*x[1]+(-0.642423472369)*x[2]**o+(-0.759623225653)*x[2]
        arg[2,4,0,1]=(-0.322255157841)*x[0]**o+(-0.253806844569)*x[0]+(-0.0688746199637)*x[1]**o+(0.232840317719)*x[1]+(-0.815623721026)*x[2]**o+(0.43055485562)*x[2]
        arg[2,4,1,0]=(-0.452320341493)*x[0]**o+(0.755240168228)*x[0]+(-0.892552611066)*x[1]**o+(0.438275161458)*x[1]+(0.441378369388)*x[2]**o+(-0.870501114753)*x[2]
        arg[2,4,1,1]=(0.496731258286)*x[0]**o+(-0.499918335322)*x[0]+(0.000476061942027)*x[1]**o+(-0.857762621856)*x[1]+(0.957433798133)*x[2]**o+(-0.766450629361)*x[2]
        arg[2,4,2,0]=(0.860379598562)*x[0]**o+(-0.0816478389485)*x[0]+(-0.438181864676)*x[1]**o+(0.78937427676)*x[1]+(-0.97940010041)*x[2]**o+(0.583914605259)*x[2]
        arg[2,4,2,1]=(0.23729037277)*x[0]**o+(-0.365380141474)*x[0]+(-0.471505632575)*x[1]**o+(-0.168838288)*x[1]+(-0.519847065791)*x[2]**o+(-0.816100518668)*x[2]
        arg[3,0,0,0]=(0.729921808661)*x[0]**o+(0.732868255388)*x[0]+(0.305434623795)*x[1]**o+(-0.45402804883)*x[1]+(0.604368334171)*x[2]**o+(-0.179233711008)*x[2]
        arg[3,0,0,1]=(-0.208916687176)*x[0]**o+(-0.297916049539)*x[0]+(0.854941564423)*x[1]**o+(0.775355543915)*x[1]+(-0.373382357688)*x[2]**o+(0.792557314074)*x[2]
        arg[3,0,1,0]=(0.606041025674)*x[0]**o+(0.115814644681)*x[0]+(-0.414470914985)*x[1]**o+(-0.136577390174)*x[1]+(0.672303216396)*x[2]**o+(-0.479680752765)*x[2]
        arg[3,0,1,1]=(-0.535633649471)*x[0]**o+(0.133761944775)*x[0]+(-0.980107507665)*x[1]**o+(0.211561747114)*x[1]+(-0.530648770004)*x[2]**o+(0.417977629512)*x[2]
        arg[3,0,2,0]=(-0.242797826282)*x[0]**o+(-0.828691638823)*x[0]+(-0.187401612749)*x[1]**o+(-0.453821620107)*x[1]+(-0.475718408605)*x[2]**o+(0.683502702497)*x[2]
        arg[3,0,2,1]=(-0.66640812951)*x[0]**o+(-0.259184719546)*x[0]+(0.463743732654)*x[1]**o+(0.6746349156)*x[1]+(-0.0532918236783)*x[2]**o+(-0.714197225876)*x[2]
        arg[3,1,0,0]=(-0.97693616823)*x[0]**o+(-0.825560555757)*x[0]+(0.118452690664)*x[1]**o+(-0.125279874388)*x[1]+(0.733803136335)*x[2]**o+(-0.884902209956)*x[2]
        arg[3,1,0,1]=(-0.652466190334)*x[0]**o+(0.370106640243)*x[0]+(0.0850861818306)*x[1]**o+(-0.49744492012)*x[1]+(-0.843498619618)*x[2]**o+(-0.0480070808698)*x[2]
        arg[3,1,1,0]=(-0.450526704276)*x[0]**o+(0.430774951776)*x[0]+(-0.956799540696)*x[1]**o+(0.820090546747)*x[1]+(0.831582198997)*x[2]**o+(0.491042906394)*x[2]
        arg[3,1,1,1]=(0.419717807665)*x[0]**o+(0.320904534666)*x[0]+(-0.483247938999)*x[1]**o+(0.82174341564)*x[1]+(-0.441719998974)*x[2]**o+(0.691713322239)*x[2]
        arg[3,1,2,0]=(0.255464093123)*x[0]**o+(-0.256334953799)*x[0]+(0.677105284128)*x[1]**o+(-0.584438025212)*x[1]+(0.505820591597)*x[2]**o+(0.247943562937)*x[2]
        arg[3,1,2,1]=(-0.495204107745)*x[0]**o+(0.234426349601)*x[0]+(0.736969108189)*x[1]**o+(-0.369684409765)*x[1]+(0.164303379282)*x[2]**o+(-0.595757590519)*x[2]
        arg[3,2,0,0]=(0.238639684591)*x[0]**o+(0.568638143194)*x[0]+(0.972646679042)*x[1]**o+(-0.652648194795)*x[1]+(-0.0530713046414)*x[2]**o+(0.356043665396)*x[2]
        arg[3,2,0,1]=(0.458539145106)*x[0]**o+(0.111417037398)*x[0]+(-0.135374868193)*x[1]**o+(-0.535451685997)*x[1]+(-0.820797762877)*x[2]**o+(0.122239787933)*x[2]
        arg[3,2,1,0]=(-0.323251425039)*x[0]**o+(0.106998797206)*x[0]+(-0.0671655760766)*x[1]**o+(0.836342794315)*x[1]+(0.34628715493)*x[2]**o+(-0.11058327806)*x[2]
        arg[3,2,1,1]=(0.181994203911)*x[0]**o+(-0.710678453639)*x[0]+(0.625606409786)*x[1]**o+(-0.363170328543)*x[1]+(-0.932998745422)*x[2]**o+(-0.0411581284314)*x[2]
        arg[3,2,2,0]=(0.710847104903)*x[0]**o+(0.483971915285)*x[0]+(0.42996482533)*x[1]**o+(0.456817348253)*x[1]+(0.844562058604)*x[2]**o+(-0.469586014721)*x[2]
        arg[3,2,2,1]=(-0.845771280837)*x[0]**o+(-0.081384696494)*x[0]+(0.71823377714)*x[1]**o+(-0.846672143168)*x[1]+(-0.139540918461)*x[2]**o+(0.837646131987)*x[2]
        arg[3,3,0,0]=(-0.958726792109)*x[0]**o+(0.53196455268)*x[0]+(-0.621810423658)*x[1]**o+(-0.0352106197658)*x[1]+(-0.636059207474)*x[2]**o+(-0.0491118207063)*x[2]
        arg[3,3,0,1]=(0.987836799279)*x[0]**o+(0.41633576224)*x[0]+(0.312021799176)*x[1]**o+(-0.168836582938)*x[1]+(-0.391725003735)*x[2]**o+(-0.158000182269)*x[2]
        arg[3,3,1,0]=(-0.280868116529)*x[0]**o+(-0.687924000092)*x[0]+(0.865840229722)*x[1]**o+(-0.0823203022874)*x[1]+(-0.857309899784)*x[2]**o+(-0.0657473493059)*x[2]
        arg[3,3,1,1]=(0.680496030567)*x[0]**o+(0.159575720724)*x[0]+(0.0531895122046)*x[1]**o+(0.359273896084)*x[1]+(0.35084774502)*x[2]**o+(0.0982220259531)*x[2]
        arg[3,3,2,0]=(0.140872311111)*x[0]**o+(-0.00845309046907)*x[0]+(-0.845527142503)*x[1]**o+(-0.67713103521)*x[1]+(0.972054187828)*x[2]**o+(-0.800664963509)*x[2]
        arg[3,3,2,1]=(-0.490866666782)*x[0]**o+(-0.0988939523952)*x[0]+(0.236210969508)*x[1]**o+(-0.672617589992)*x[1]+(-0.640558724403)*x[2]**o+(-0.203125267713)*x[2]
        arg[3,4,0,0]=(0.318438839683)*x[0]**o+(0.0864322377812)*x[0]+(-0.371419701638)*x[1]**o+(-0.311583139142)*x[1]+(-0.0533111244011)*x[2]**o+(0.841240973423)*x[2]
        arg[3,4,0,1]=(-0.573110771334)*x[0]**o+(0.703852077914)*x[0]+(0.613309217217)*x[1]**o+(0.247223193846)*x[1]+(0.618952485205)*x[2]**o+(-0.395998308184)*x[2]
        arg[3,4,1,0]=(0.512896938934)*x[0]**o+(0.848653171505)*x[0]+(-0.944903687062)*x[1]**o+(0.179782177607)*x[1]+(-0.0287288996538)*x[2]**o+(-0.385519118293)*x[2]
        arg[3,4,1,1]=(0.234190647718)*x[0]**o+(0.445478221419)*x[0]+(0.588931676434)*x[1]**o+(0.140788139998)*x[1]+(-0.761952359657)*x[2]**o+(0.92377426359)*x[2]
        arg[3,4,2,0]=(-0.836038966171)*x[0]**o+(-0.772990045933)*x[0]+(0.0622128209877)*x[1]**o+(-0.73733496141)*x[1]+(0.0234446925161)*x[2]**o+(-0.682234213315)*x[2]
        arg[3,4,2,1]=(0.250498084214)*x[0]**o+(0.539088356159)*x[0]+(-0.691842680176)*x[1]**o+(0.710040802953)*x[1]+(-0.43691551501)*x[2]**o+(0.710668600989)*x[2]
        ref[0,0,0,0]=(-0.836302395465)*x_ref[0]**o+(-0.842568675604)*x_ref[0]+(-0.780133441207)*x_ref[1]**o+(0.54483947113)*x_ref[1]+(-0.130253019696)*x_ref[2]**o+(0.358303389828)*x_ref[2]
        ref[0,0,0,1]=(-0.722847500652)*x_ref[0]**o+(0.42887368878)*x_ref[0]+(0.979187056933)*x_ref[1]**o+(0.624875714982)*x_ref[1]+(-0.672529765247)*x_ref[2]**o+(0.822363583702)*x_ref[2]
        ref[0,0,1,0]=(-0.28989650691)*x_ref[0]**o+(-0.847647954773)*x_ref[0]+(-0.11119862113)*x_ref[1]**o+(0.0809009676818)*x_ref[1]+(-0.672626533587)*x_ref[2]**o+(0.580561232267)*x_ref[2]
        ref[0,0,1,1]=(-0.64783363248)*x_ref[0]**o+(0.212828552873)*x_ref[0]+(-0.845297180936)*x_ref[1]**o+(-0.714760411186)*x_ref[1]+(-0.385173605122)*x_ref[2]**o+(-0.760750317556)*x_ref[2]
        ref[0,0,2,0]=(-0.328757006046)*x_ref[0]**o+(-0.0935907722247)*x_ref[0]+(-0.895866135795)*x_ref[1]**o+(-0.701241976745)*x_ref[1]+(0.893684139941)*x_ref[2]**o+(-0.265339318222)*x_ref[2]
        ref[0,0,2,1]=(0.134066662346)*x_ref[0]**o+(0.0383231508856)*x_ref[0]+(-0.347744261995)*x_ref[1]**o+(-0.86897907336)*x_ref[1]+(-0.287792154521)*x_ref[2]**o+(0.859351846891)*x_ref[2]
        ref[0,1,0,0]=(-0.635128124936)*x_ref[0]**o+(-0.659462189733)*x_ref[0]+(-0.0964365747936)*x_ref[1]**o+(-0.385702168747)*x_ref[1]+(0.179539914057)*x_ref[2]**o+(-0.0176420847859)*x_ref[2]
        ref[0,1,0,1]=(-0.580750807831)*x_ref[0]**o+(0.756493147583)*x_ref[0]+(-0.436842251289)*x_ref[1]**o+(-0.705624646095)*x_ref[1]+(0.983064096262)*x_ref[2]**o+(-0.0197153825423)*x_ref[2]
        ref[0,1,1,0]=(-0.567748582365)*x_ref[0]**o+(0.579324420356)*x_ref[0]+(0.0434168493419)*x_ref[1]**o+(-0.839692920334)*x_ref[1]+(-0.119239847745)*x_ref[2]**o+(0.527047942834)*x_ref[2]
        ref[0,1,1,1]=(0.52870403655)*x_ref[0]**o+(0.700090892471)*x_ref[0]+(0.730224313921)*x_ref[1]**o+(-0.815451751207)*x_ref[1]+(-0.943945382413)*x_ref[2]**o+(-0.571714141916)*x_ref[2]
        ref[0,1,2,0]=(-0.803841094887)*x_ref[0]**o+(0.0595126006237)*x_ref[0]+(-0.612341797645)*x_ref[1]**o+(0.944552805211)*x_ref[1]+(-0.00253009731786)*x_ref[2]**o+(0.0174867439912)*x_ref[2]
        ref[0,1,2,1]=(-0.145623212549)*x_ref[0]**o+(0.546039957008)*x_ref[0]+(-0.221425237579)*x_ref[1]**o+(0.229367671159)*x_ref[1]+(-0.784903869759)*x_ref[2]**o+(0.898045208871)*x_ref[2]
        ref[0,2,0,0]=(0.417924494005)*x_ref[0]**o+(-0.192820051987)*x_ref[0]+(0.878916271492)*x_ref[1]**o+(-0.35056320984)*x_ref[1]+(0.982897575604)*x_ref[2]**o+(-0.225793730616)*x_ref[2]
        ref[0,2,0,1]=(-0.493279498849)*x_ref[0]**o+(-0.154177411087)*x_ref[0]+(-0.923592860492)*x_ref[1]**o+(0.872010949252)*x_ref[1]+(-0.936252147064)*x_ref[2]**o+(0.947136819308)*x_ref[2]
        ref[0,2,1,0]=(0.109357630607)*x_ref[0]**o+(-0.518825477709)*x_ref[0]+(-0.375510667498)*x_ref[1]**o+(0.476537999644)*x_ref[1]+(0.676017421029)*x_ref[2]**o+(-0.633301158984)*x_ref[2]
        ref[0,2,1,1]=(-0.0662265225539)*x_ref[0]**o+(-0.276634748263)*x_ref[0]+(-0.0655431981102)*x_ref[1]**o+(0.0578482920991)*x_ref[1]+(-0.264370611185)*x_ref[2]**o+(-0.674612985594)*x_ref[2]
        ref[0,2,2,0]=(-0.924186592743)*x_ref[0]**o+(0.197603043149)*x_ref[0]+(-0.676882805046)*x_ref[1]**o+(0.564142845865)*x_ref[1]+(0.590369345306)*x_ref[2]**o+(-0.872019490978)*x_ref[2]
        ref[0,2,2,1]=(0.480683776968)*x_ref[0]**o+(-0.412055284173)*x_ref[0]+(0.829996313473)*x_ref[1]**o+(-0.235665473048)*x_ref[1]+(-0.429191305284)*x_ref[2]**o+(-0.0867804318457)*x_ref[2]
        ref[0,3,0,0]=(-0.0903660006043)*x_ref[0]**o+(-0.507571901766)*x_ref[0]+(-0.0305305861809)*x_ref[1]**o+(-0.915027969989)*x_ref[1]+(0.955644774347)*x_ref[2]**o+(0.739244618775)*x_ref[2]
        ref[0,3,0,1]=(0.600619122998)*x_ref[0]**o+(0.0128699484527)*x_ref[0]+(-0.360431921281)*x_ref[1]**o+(0.304773619811)*x_ref[1]+(-0.302423314054)*x_ref[2]**o+(0.680149401572)*x_ref[2]
        ref[0,3,1,0]=(-0.638540513323)*x_ref[0]**o+(0.766655754067)*x_ref[0]+(0.276879854223)*x_ref[1]**o+(-0.628688275247)*x_ref[1]+(0.724584107114)*x_ref[2]**o+(0.682193551967)*x_ref[2]
        ref[0,3,1,1]=(-0.516656801349)*x_ref[0]**o+(-0.513901319383)*x_ref[0]+(-0.227957010777)*x_ref[1]**o+(-0.164363207703)*x_ref[1]+(-0.919936042047)*x_ref[2]**o+(0.682400184997)*x_ref[2]
        ref[0,3,2,0]=(-0.819629653095)*x_ref[0]**o+(0.183937049274)*x_ref[0]+(0.0411075800579)*x_ref[1]**o+(0.0813088270699)*x_ref[1]+(0.762416334575)*x_ref[2]**o+(0.236553726705)*x_ref[2]
        ref[0,3,2,1]=(0.513141183759)*x_ref[0]**o+(-0.452475512734)*x_ref[0]+(0.0461424221984)*x_ref[1]**o+(-0.673452844636)*x_ref[1]+(-0.194493328878)*x_ref[2]**o+(0.812459956422)*x_ref[2]
        ref[0,4,0,0]=(0.672167415195)*x_ref[0]**o+(-0.641004730405)*x_ref[0]+(-0.383031420851)*x_ref[1]**o+(0.392366431186)*x_ref[1]+(0.66479124224)*x_ref[2]**o+(0.845503225104)*x_ref[2]
        ref[0,4,0,1]=(0.908241966695)*x_ref[0]**o+(-0.816989438726)*x_ref[0]+(-0.144902267583)*x_ref[1]**o+(-0.654901540598)*x_ref[1]+(-0.9840470269)*x_ref[2]**o+(0.510970695154)*x_ref[2]
        ref[0,4,1,0]=(0.788858704366)*x_ref[0]**o+(-0.50266306504)*x_ref[0]+(-0.681248796723)*x_ref[1]**o+(0.0317663112509)*x_ref[1]+(0.662428035619)*x_ref[2]**o+(0.377116420912)*x_ref[2]
        ref[0,4,1,1]=(-0.0734522829037)*x_ref[0]**o+(0.0822741484655)*x_ref[0]+(-0.509279578815)*x_ref[1]**o+(-0.95372420098)*x_ref[1]+(0.682868362508)*x_ref[2]**o+(-0.947156883638)*x_ref[2]
        ref[0,4,2,0]=(-0.843236607337)*x_ref[0]**o+(-0.925326238513)*x_ref[0]+(-0.153688519983)*x_ref[1]**o+(-0.68456413487)*x_ref[1]+(0.920442617368)*x_ref[2]**o+(0.852945557893)*x_ref[2]
        ref[0,4,2,1]=(-0.283772512528)*x_ref[0]**o+(0.202730738521)*x_ref[0]+(-0.287375172356)*x_ref[1]**o+(-0.176699132816)*x_ref[1]+(0.446391113264)*x_ref[2]**o+(-0.783266847348)*x_ref[2]
        ref[1,0,0,0]=(0.81284364562)*x_ref[0]**o+(-0.454432332505)*x_ref[0]+(-0.684692596262)*x_ref[1]**o+(-0.345923888237)*x_ref[1]+(0.1201728631)*x_ref[2]**o+(0.808246135028)*x_ref[2]
        ref[1,0,0,1]=(0.941104328706)*x_ref[0]**o+(-0.802008654964)*x_ref[0]+(0.902047483677)*x_ref[1]**o+(0.777049224612)*x_ref[1]+(0.120184332854)*x_ref[2]**o+(0.983216899102)*x_ref[2]
        ref[1,0,1,0]=(-0.420996925767)*x_ref[0]**o+(0.0941750615502)*x_ref[0]+(-0.935342933578)*x_ref[1]**o+(0.925891822058)*x_ref[1]+(-0.153652985582)*x_ref[2]**o+(-0.336548986153)*x_ref[2]
        ref[1,0,1,1]=(-0.692886221795)*x_ref[0]**o+(-0.920419410139)*x_ref[0]+(-0.586575583263)*x_ref[1]**o+(-0.269347362009)*x_ref[1]+(-0.872169736045)*x_ref[2]**o+(-0.724727186336)*x_ref[2]
        ref[1,0,2,0]=(-0.0171222282058)*x_ref[0]**o+(0.338180602976)*x_ref[0]+(-0.0345958505395)*x_ref[1]**o+(0.631728421505)*x_ref[1]+(0.90596443453)*x_ref[2]**o+(-0.144785193015)*x_ref[2]
        ref[1,0,2,1]=(-0.91636583885)*x_ref[0]**o+(-0.807496789574)*x_ref[0]+(0.63472495957)*x_ref[1]**o+(0.966464343993)*x_ref[1]+(-0.471311384621)*x_ref[2]**o+(-0.449420138238)*x_ref[2]
        ref[1,1,0,0]=(-0.239930503474)*x_ref[0]**o+(-0.476439715994)*x_ref[0]+(0.421600511888)*x_ref[1]**o+(0.477823144792)*x_ref[1]+(0.251409914807)*x_ref[2]**o+(0.636818001419)*x_ref[2]
        ref[1,1,0,1]=(-0.427510410263)*x_ref[0]**o+(0.494810857516)*x_ref[0]+(-0.129415181352)*x_ref[1]**o+(0.754147770071)*x_ref[1]+(0.0218776365985)*x_ref[2]**o+(0.706183216616)*x_ref[2]
        ref[1,1,1,0]=(-0.469150375165)*x_ref[0]**o+(-0.0651101638073)*x_ref[0]+(-0.697853003861)*x_ref[1]**o+(0.0389330102026)*x_ref[1]+(-0.256557942045)*x_ref[2]**o+(0.409565340425)*x_ref[2]
        ref[1,1,1,1]=(0.14957005608)*x_ref[0]**o+(-0.209482968126)*x_ref[0]+(0.368899650309)*x_ref[1]**o+(-0.0899321000137)*x_ref[1]+(0.282389903357)*x_ref[2]**o+(-0.166633883916)*x_ref[2]
        ref[1,1,2,0]=(0.0374938231714)*x_ref[0]**o+(-0.687972899893)*x_ref[0]+(-0.0686478107779)*x_ref[1]**o+(0.318977937888)*x_ref[1]+(-0.601269126788)*x_ref[2]**o+(0.453188855786)*x_ref[2]
        ref[1,1,2,1]=(-0.853894223222)*x_ref[0]**o+(0.441830151478)*x_ref[0]+(0.973114898368)*x_ref[1]**o+(-0.264454889071)*x_ref[1]+(-0.828332687531)*x_ref[2]**o+(0.302493748565)*x_ref[2]
        ref[1,2,0,0]=(0.210365824111)*x_ref[0]**o+(0.4674117857)*x_ref[0]+(-0.129546969044)*x_ref[1]**o+(0.566357125352)*x_ref[1]+(-0.721733448514)*x_ref[2]**o+(0.154253207106)*x_ref[2]
        ref[1,2,0,1]=(0.805325508181)*x_ref[0]**o+(-0.929113563452)*x_ref[0]+(-0.487301761321)*x_ref[1]**o+(-0.822405526519)*x_ref[1]+(0.254754490326)*x_ref[2]**o+(-0.196480710709)*x_ref[2]
        ref[1,2,1,0]=(-0.520777275475)*x_ref[0]**o+(-0.912569745644)*x_ref[0]+(-0.784504791087)*x_ref[1]**o+(-0.903473217653)*x_ref[1]+(0.583297107374)*x_ref[2]**o+(0.664811898195)*x_ref[2]
        ref[1,2,1,1]=(0.129152152958)*x_ref[0]**o+(0.497832200867)*x_ref[0]+(0.121217248234)*x_ref[1]**o+(-0.183368066714)*x_ref[1]+(-0.776145485536)*x_ref[2]**o+(-0.223165724715)*x_ref[2]
        ref[1,2,2,0]=(-0.698981324488)*x_ref[0]**o+(0.00951277777924)*x_ref[0]+(0.997410976507)*x_ref[1]**o+(0.534670303453)*x_ref[1]+(0.251309223218)*x_ref[2]**o+(-0.921148646979)*x_ref[2]
        ref[1,2,2,1]=(-0.731509934487)*x_ref[0]**o+(0.87731926834)*x_ref[0]+(0.0865507713913)*x_ref[1]**o+(0.708835987019)*x_ref[1]+(0.294182269171)*x_ref[2]**o+(-0.316660750729)*x_ref[2]
        ref[1,3,0,0]=(-0.282903962918)*x_ref[0]**o+(-0.793457978844)*x_ref[0]+(0.287425807914)*x_ref[1]**o+(-0.0252722364835)*x_ref[1]+(-0.956202022515)*x_ref[2]**o+(0.650271579135)*x_ref[2]
        ref[1,3,0,1]=(-0.946723111202)*x_ref[0]**o+(0.669431653693)*x_ref[0]+(-0.791125976188)*x_ref[1]**o+(-0.466484914794)*x_ref[1]+(-0.217225801872)*x_ref[2]**o+(0.224719473996)*x_ref[2]
        ref[1,3,1,0]=(0.516479301043)*x_ref[0]**o+(-0.42420332366)*x_ref[0]+(-0.0649227247176)*x_ref[1]**o+(0.208849752675)*x_ref[1]+(-0.939001099693)*x_ref[2]**o+(-0.0537827140932)*x_ref[2]
        ref[1,3,1,1]=(-0.292077069459)*x_ref[0]**o+(-0.444848119744)*x_ref[0]+(0.268592719199)*x_ref[1]**o+(-0.349401060323)*x_ref[1]+(-0.540803784489)*x_ref[2]**o+(-0.487170123526)*x_ref[2]
        ref[1,3,2,0]=(-0.626222042661)*x_ref[0]**o+(0.245937409624)*x_ref[0]+(-0.24692162006)*x_ref[1]**o+(0.108230029208)*x_ref[1]+(-0.760463241928)*x_ref[2]**o+(-0.587078263201)*x_ref[2]
        ref[1,3,2,1]=(-0.923891147756)*x_ref[0]**o+(0.0961285842162)*x_ref[0]+(0.0146853466585)*x_ref[1]**o+(0.00951339122359)*x_ref[1]+(-0.261383704723)*x_ref[2]**o+(-0.0262515691324)*x_ref[2]
        ref[1,4,0,0]=(-0.0370003524292)*x_ref[0]**o+(-0.0463848489957)*x_ref[0]+(0.940061458694)*x_ref[1]**o+(0.592787058925)*x_ref[1]+(-0.217284379638)*x_ref[2]**o+(-0.11746834109)*x_ref[2]
        ref[1,4,0,1]=(0.830404283093)*x_ref[0]**o+(0.268135490403)*x_ref[0]+(-0.894500398195)*x_ref[1]**o+(0.751042648917)*x_ref[1]+(0.687583292754)*x_ref[2]**o+(-0.884596509742)*x_ref[2]
        ref[1,4,1,0]=(0.414764592469)*x_ref[0]**o+(-0.529613347919)*x_ref[0]+(-0.727435115594)*x_ref[1]**o+(-0.546579777725)*x_ref[1]+(0.378011893971)*x_ref[2]**o+(-0.578081183238)*x_ref[2]
        ref[1,4,1,1]=(0.406513376747)*x_ref[0]**o+(-0.350132869579)*x_ref[0]+(-0.557954631299)*x_ref[1]**o+(0.868176175754)*x_ref[1]+(0.793021832632)*x_ref[2]**o+(0.280529771851)*x_ref[2]
        ref[1,4,2,0]=(0.56741953575)*x_ref[0]**o+(0.0115947684302)*x_ref[0]+(0.872812725363)*x_ref[1]**o+(-0.215183541745)*x_ref[1]+(0.64853692203)*x_ref[2]**o+(-0.139561471922)*x_ref[2]
        ref[1,4,2,1]=(0.650916904666)*x_ref[0]**o+(0.878765271839)*x_ref[0]+(0.0141264348052)*x_ref[1]**o+(-0.56314426774)*x_ref[1]+(-0.267854377704)*x_ref[2]**o+(0.852271380667)*x_ref[2]
        ref[2,0,0,0]=(0.948359141577)*x_ref[0]**o+(-0.0881747280974)*x_ref[0]+(0.804871563368)*x_ref[1]**o+(-0.0298585763446)*x_ref[1]+(0.506923333569)*x_ref[2]**o+(-0.920947093648)*x_ref[2]
        ref[2,0,0,1]=(-0.0144221258584)*x_ref[0]**o+(-0.679723072824)*x_ref[0]+(0.00829336956687)*x_ref[1]**o+(0.675680809303)*x_ref[1]+(-0.803195231536)*x_ref[2]**o+(0.70240358571)*x_ref[2]
        ref[2,0,1,0]=(-0.159174987706)*x_ref[0]**o+(0.922309271483)*x_ref[0]+(0.344354343035)*x_ref[1]**o+(-0.00954038311323)*x_ref[1]+(-0.50225452804)*x_ref[2]**o+(0.355899734308)*x_ref[2]
        ref[2,0,1,1]=(0.0857611780404)*x_ref[0]**o+(0.876479431983)*x_ref[0]+(0.105049158502)*x_ref[1]**o+(-0.530308461426)*x_ref[1]+(-0.553136649343)*x_ref[2]**o+(0.385458516481)*x_ref[2]
        ref[2,0,2,0]=(-0.260086070842)*x_ref[0]**o+(0.417766241499)*x_ref[0]+(-0.292370383673)*x_ref[1]**o+(-0.321315349553)*x_ref[1]+(-0.437301444082)*x_ref[2]**o+(0.730845669715)*x_ref[2]
        ref[2,0,2,1]=(0.132547674276)*x_ref[0]**o+(0.783602886333)*x_ref[0]+(0.822361450413)*x_ref[1]**o+(-0.810402783)*x_ref[1]+(0.759701191023)*x_ref[2]**o+(0.644375788608)*x_ref[2]
        ref[2,1,0,0]=(-0.841458111227)*x_ref[0]**o+(0.779917641736)*x_ref[0]+(-0.0745587217256)*x_ref[1]**o+(0.938545280143)*x_ref[1]+(0.454600163373)*x_ref[2]**o+(-0.490348377697)*x_ref[2]
        ref[2,1,0,1]=(-0.876477136174)*x_ref[0]**o+(-0.248200369939)*x_ref[0]+(-0.973654243816)*x_ref[1]**o+(0.482402123794)*x_ref[1]+(-0.776970490827)*x_ref[2]**o+(-0.571349467703)*x_ref[2]
        ref[2,1,1,0]=(-0.433528193102)*x_ref[0]**o+(-0.57712915767)*x_ref[0]+(-0.328671198058)*x_ref[1]**o+(-0.926873019075)*x_ref[1]+(0.314232739356)*x_ref[2]**o+(-0.199625466412)*x_ref[2]
        ref[2,1,1,1]=(0.892707770392)*x_ref[0]**o+(-0.680081898234)*x_ref[0]+(-0.754181058483)*x_ref[1]**o+(-0.253728945575)*x_ref[1]+(-0.906373406684)*x_ref[2]**o+(-0.0266435080833)*x_ref[2]
        ref[2,1,2,0]=(-0.121605378472)*x_ref[0]**o+(-0.503955015634)*x_ref[0]+(-0.0960767196114)*x_ref[1]**o+(-0.937914890361)*x_ref[1]+(0.212042931853)*x_ref[2]**o+(0.930703155408)*x_ref[2]
        ref[2,1,2,1]=(0.0564159441047)*x_ref[0]**o+(0.842332721105)*x_ref[0]+(0.954741657608)*x_ref[1]**o+(0.73503176)*x_ref[1]+(-0.390679938582)*x_ref[2]**o+(-0.440014747741)*x_ref[2]
        ref[2,2,0,0]=(0.0100637970821)*x_ref[0]**o+(-0.0283353154145)*x_ref[0]+(0.865360758262)*x_ref[1]**o+(0.780975027988)*x_ref[1]+(0.16115842919)*x_ref[2]**o+(0.247580155656)*x_ref[2]
        ref[2,2,0,1]=(0.61102059426)*x_ref[0]**o+(-0.198005744101)*x_ref[0]+(0.503568797262)*x_ref[1]**o+(0.0288809726497)*x_ref[1]+(-0.765127098183)*x_ref[2]**o+(0.409803857723)*x_ref[2]
        ref[2,2,1,0]=(-0.712272018295)*x_ref[0]**o+(-0.62562142063)*x_ref[0]+(-0.579501949719)*x_ref[1]**o+(0.628539275236)*x_ref[1]+(0.0200945124964)*x_ref[2]**o+(0.175966024769)*x_ref[2]
        ref[2,2,1,1]=(-0.798244445519)*x_ref[0]**o+(-0.376440962535)*x_ref[0]+(-0.134467401497)*x_ref[1]**o+(0.877538294477)*x_ref[1]+(-0.484870741192)*x_ref[2]**o+(-0.224064469935)*x_ref[2]
        ref[2,2,2,0]=(0.200196926368)*x_ref[0]**o+(-0.528000867086)*x_ref[0]+(-0.708942342624)*x_ref[1]**o+(-0.591994969141)*x_ref[1]+(-0.990641592656)*x_ref[2]**o+(-0.875897347303)*x_ref[2]
        ref[2,2,2,1]=(0.949113190797)*x_ref[0]**o+(0.867855496786)*x_ref[0]+(-0.0314953438065)*x_ref[1]**o+(-0.371928428005)*x_ref[1]+(-0.526403191063)*x_ref[2]**o+(0.653049122326)*x_ref[2]
        ref[2,3,0,0]=(-0.353418327127)*x_ref[0]**o+(-0.418588306429)*x_ref[0]+(0.400872043269)*x_ref[1]**o+(0.141907924042)*x_ref[1]+(-0.999869579462)*x_ref[2]**o+(-0.410423954563)*x_ref[2]
        ref[2,3,0,1]=(-0.240390877307)*x_ref[0]**o+(0.0901879916115)*x_ref[0]+(-0.0734683209657)*x_ref[1]**o+(0.274836005886)*x_ref[1]+(-0.336375498181)*x_ref[2]**o+(-0.754776455221)*x_ref[2]
        ref[2,3,1,0]=(0.259928460822)*x_ref[0]**o+(-0.227630454708)*x_ref[0]+(0.725968066278)*x_ref[1]**o+(-0.545440695763)*x_ref[1]+(-0.0354926607815)*x_ref[2]**o+(0.365010532181)*x_ref[2]
        ref[2,3,1,1]=(-0.969799373398)*x_ref[0]**o+(0.514498233214)*x_ref[0]+(0.907450925318)*x_ref[1]**o+(-0.234993706406)*x_ref[1]+(0.206323910806)*x_ref[2]**o+(0.209364250369)*x_ref[2]
        ref[2,3,2,0]=(0.712905565941)*x_ref[0]**o+(-0.353322758263)*x_ref[0]+(0.816861181494)*x_ref[1]**o+(-0.362388243619)*x_ref[1]+(0.302543517607)*x_ref[2]**o+(-0.831280418852)*x_ref[2]
        ref[2,3,2,1]=(-0.386373711182)*x_ref[0]**o+(0.946914567792)*x_ref[0]+(-0.52795085887)*x_ref[1]**o+(0.86635245964)*x_ref[1]+(0.654261054178)*x_ref[2]**o+(0.452334882498)*x_ref[2]
        ref[2,4,0,0]=(0.374747033547)*x_ref[0]**o+(-0.3471578669)*x_ref[0]+(0.322217002639)*x_ref[1]**o+(0.92988068982)*x_ref[1]+(-0.642423472369)*x_ref[2]**o+(-0.759623225653)*x_ref[2]
        ref[2,4,0,1]=(-0.322255157841)*x_ref[0]**o+(-0.253806844569)*x_ref[0]+(-0.0688746199637)*x_ref[1]**o+(0.232840317719)*x_ref[1]+(-0.815623721026)*x_ref[2]**o+(0.43055485562)*x_ref[2]
        ref[2,4,1,0]=(-0.452320341493)*x_ref[0]**o+(0.755240168228)*x_ref[0]+(-0.892552611066)*x_ref[1]**o+(0.438275161458)*x_ref[1]+(0.441378369388)*x_ref[2]**o+(-0.870501114753)*x_ref[2]
        ref[2,4,1,1]=(0.496731258286)*x_ref[0]**o+(-0.499918335322)*x_ref[0]+(0.000476061942027)*x_ref[1]**o+(-0.857762621856)*x_ref[1]+(0.957433798133)*x_ref[2]**o+(-0.766450629361)*x_ref[2]
        ref[2,4,2,0]=(0.860379598562)*x_ref[0]**o+(-0.0816478389485)*x_ref[0]+(-0.438181864676)*x_ref[1]**o+(0.78937427676)*x_ref[1]+(-0.97940010041)*x_ref[2]**o+(0.583914605259)*x_ref[2]
        ref[2,4,2,1]=(0.23729037277)*x_ref[0]**o+(-0.365380141474)*x_ref[0]+(-0.471505632575)*x_ref[1]**o+(-0.168838288)*x_ref[1]+(-0.519847065791)*x_ref[2]**o+(-0.816100518668)*x_ref[2]
        ref[3,0,0,0]=(0.729921808661)*x_ref[0]**o+(0.732868255388)*x_ref[0]+(0.305434623795)*x_ref[1]**o+(-0.45402804883)*x_ref[1]+(0.604368334171)*x_ref[2]**o+(-0.179233711008)*x_ref[2]
        ref[3,0,0,1]=(-0.208916687176)*x_ref[0]**o+(-0.297916049539)*x_ref[0]+(0.854941564423)*x_ref[1]**o+(0.775355543915)*x_ref[1]+(-0.373382357688)*x_ref[2]**o+(0.792557314074)*x_ref[2]
        ref[3,0,1,0]=(0.606041025674)*x_ref[0]**o+(0.115814644681)*x_ref[0]+(-0.414470914985)*x_ref[1]**o+(-0.136577390174)*x_ref[1]+(0.672303216396)*x_ref[2]**o+(-0.479680752765)*x_ref[2]
        ref[3,0,1,1]=(-0.535633649471)*x_ref[0]**o+(0.133761944775)*x_ref[0]+(-0.980107507665)*x_ref[1]**o+(0.211561747114)*x_ref[1]+(-0.530648770004)*x_ref[2]**o+(0.417977629512)*x_ref[2]
        ref[3,0,2,0]=(-0.242797826282)*x_ref[0]**o+(-0.828691638823)*x_ref[0]+(-0.187401612749)*x_ref[1]**o+(-0.453821620107)*x_ref[1]+(-0.475718408605)*x_ref[2]**o+(0.683502702497)*x_ref[2]
        ref[3,0,2,1]=(-0.66640812951)*x_ref[0]**o+(-0.259184719546)*x_ref[0]+(0.463743732654)*x_ref[1]**o+(0.6746349156)*x_ref[1]+(-0.0532918236783)*x_ref[2]**o+(-0.714197225876)*x_ref[2]
        ref[3,1,0,0]=(-0.97693616823)*x_ref[0]**o+(-0.825560555757)*x_ref[0]+(0.118452690664)*x_ref[1]**o+(-0.125279874388)*x_ref[1]+(0.733803136335)*x_ref[2]**o+(-0.884902209956)*x_ref[2]
        ref[3,1,0,1]=(-0.652466190334)*x_ref[0]**o+(0.370106640243)*x_ref[0]+(0.0850861818306)*x_ref[1]**o+(-0.49744492012)*x_ref[1]+(-0.843498619618)*x_ref[2]**o+(-0.0480070808698)*x_ref[2]
        ref[3,1,1,0]=(-0.450526704276)*x_ref[0]**o+(0.430774951776)*x_ref[0]+(-0.956799540696)*x_ref[1]**o+(0.820090546747)*x_ref[1]+(0.831582198997)*x_ref[2]**o+(0.491042906394)*x_ref[2]
        ref[3,1,1,1]=(0.419717807665)*x_ref[0]**o+(0.320904534666)*x_ref[0]+(-0.483247938999)*x_ref[1]**o+(0.82174341564)*x_ref[1]+(-0.441719998974)*x_ref[2]**o+(0.691713322239)*x_ref[2]
        ref[3,1,2,0]=(0.255464093123)*x_ref[0]**o+(-0.256334953799)*x_ref[0]+(0.677105284128)*x_ref[1]**o+(-0.584438025212)*x_ref[1]+(0.505820591597)*x_ref[2]**o+(0.247943562937)*x_ref[2]
        ref[3,1,2,1]=(-0.495204107745)*x_ref[0]**o+(0.234426349601)*x_ref[0]+(0.736969108189)*x_ref[1]**o+(-0.369684409765)*x_ref[1]+(0.164303379282)*x_ref[2]**o+(-0.595757590519)*x_ref[2]
        ref[3,2,0,0]=(0.238639684591)*x_ref[0]**o+(0.568638143194)*x_ref[0]+(0.972646679042)*x_ref[1]**o+(-0.652648194795)*x_ref[1]+(-0.0530713046414)*x_ref[2]**o+(0.356043665396)*x_ref[2]
        ref[3,2,0,1]=(0.458539145106)*x_ref[0]**o+(0.111417037398)*x_ref[0]+(-0.135374868193)*x_ref[1]**o+(-0.535451685997)*x_ref[1]+(-0.820797762877)*x_ref[2]**o+(0.122239787933)*x_ref[2]
        ref[3,2,1,0]=(-0.323251425039)*x_ref[0]**o+(0.106998797206)*x_ref[0]+(-0.0671655760766)*x_ref[1]**o+(0.836342794315)*x_ref[1]+(0.34628715493)*x_ref[2]**o+(-0.11058327806)*x_ref[2]
        ref[3,2,1,1]=(0.181994203911)*x_ref[0]**o+(-0.710678453639)*x_ref[0]+(0.625606409786)*x_ref[1]**o+(-0.363170328543)*x_ref[1]+(-0.932998745422)*x_ref[2]**o+(-0.0411581284314)*x_ref[2]
        ref[3,2,2,0]=(0.710847104903)*x_ref[0]**o+(0.483971915285)*x_ref[0]+(0.42996482533)*x_ref[1]**o+(0.456817348253)*x_ref[1]+(0.844562058604)*x_ref[2]**o+(-0.469586014721)*x_ref[2]
        ref[3,2,2,1]=(-0.845771280837)*x_ref[0]**o+(-0.081384696494)*x_ref[0]+(0.71823377714)*x_ref[1]**o+(-0.846672143168)*x_ref[1]+(-0.139540918461)*x_ref[2]**o+(0.837646131987)*x_ref[2]
        ref[3,3,0,0]=(-0.958726792109)*x_ref[0]**o+(0.53196455268)*x_ref[0]+(-0.621810423658)*x_ref[1]**o+(-0.0352106197658)*x_ref[1]+(-0.636059207474)*x_ref[2]**o+(-0.0491118207063)*x_ref[2]
        ref[3,3,0,1]=(0.987836799279)*x_ref[0]**o+(0.41633576224)*x_ref[0]+(0.312021799176)*x_ref[1]**o+(-0.168836582938)*x_ref[1]+(-0.391725003735)*x_ref[2]**o+(-0.158000182269)*x_ref[2]
        ref[3,3,1,0]=(-0.280868116529)*x_ref[0]**o+(-0.687924000092)*x_ref[0]+(0.865840229722)*x_ref[1]**o+(-0.0823203022874)*x_ref[1]+(-0.857309899784)*x_ref[2]**o+(-0.0657473493059)*x_ref[2]
        ref[3,3,1,1]=(0.680496030567)*x_ref[0]**o+(0.159575720724)*x_ref[0]+(0.0531895122046)*x_ref[1]**o+(0.359273896084)*x_ref[1]+(0.35084774502)*x_ref[2]**o+(0.0982220259531)*x_ref[2]
        ref[3,3,2,0]=(0.140872311111)*x_ref[0]**o+(-0.00845309046907)*x_ref[0]+(-0.845527142503)*x_ref[1]**o+(-0.67713103521)*x_ref[1]+(0.972054187828)*x_ref[2]**o+(-0.800664963509)*x_ref[2]
        ref[3,3,2,1]=(-0.490866666782)*x_ref[0]**o+(-0.0988939523952)*x_ref[0]+(0.236210969508)*x_ref[1]**o+(-0.672617589992)*x_ref[1]+(-0.640558724403)*x_ref[2]**o+(-0.203125267713)*x_ref[2]
        ref[3,4,0,0]=(0.318438839683)*x_ref[0]**o+(0.0864322377812)*x_ref[0]+(-0.371419701638)*x_ref[1]**o+(-0.311583139142)*x_ref[1]+(-0.0533111244011)*x_ref[2]**o+(0.841240973423)*x_ref[2]
        ref[3,4,0,1]=(-0.573110771334)*x_ref[0]**o+(0.703852077914)*x_ref[0]+(0.613309217217)*x_ref[1]**o+(0.247223193846)*x_ref[1]+(0.618952485205)*x_ref[2]**o+(-0.395998308184)*x_ref[2]
        ref[3,4,1,0]=(0.512896938934)*x_ref[0]**o+(0.848653171505)*x_ref[0]+(-0.944903687062)*x_ref[1]**o+(0.179782177607)*x_ref[1]+(-0.0287288996538)*x_ref[2]**o+(-0.385519118293)*x_ref[2]
        ref[3,4,1,1]=(0.234190647718)*x_ref[0]**o+(0.445478221419)*x_ref[0]+(0.588931676434)*x_ref[1]**o+(0.140788139998)*x_ref[1]+(-0.761952359657)*x_ref[2]**o+(0.92377426359)*x_ref[2]
        ref[3,4,2,0]=(-0.836038966171)*x_ref[0]**o+(-0.772990045933)*x_ref[0]+(0.0622128209877)*x_ref[1]**o+(-0.73733496141)*x_ref[1]+(0.0234446925161)*x_ref[2]**o+(-0.682234213315)*x_ref[2]
        ref[3,4,2,1]=(0.250498084214)*x_ref[0]**o+(0.539088356159)*x_ref[0]+(-0.691842680176)*x_ref[1]**o+(0.710040802953)*x_ref[1]+(-0.43691551501)*x_ref[2]**o+(0.710668600989)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank0(self):
      """
      tests interpolate for rank 0 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.70941034293)*x[0]**o+(-0.988234004949)*x[0]+(0.440342659022)*x[1]**o+(0.221444369839)*x[1]
        ref=(0.70941034293)*x_ref[0]**o+(-0.988234004949)*x_ref[0]+(0.440342659022)*x_ref[1]**o+(0.221444369839)*x_ref[1]
      else:
        arg=(0.720798141776)*x[0]**o+(0.841475017012)*x[0]+(-0.76285313232)*x[1]**o+(-0.214354625268)*x[1]+(-0.887218533751)*x[2]**o+(-0.066368425118)*x[2]
        ref=(0.720798141776)*x_ref[0]**o+(0.841475017012)*x_ref[0]+(-0.76285313232)*x_ref[1]**o+(-0.214354625268)*x_ref[1]+(-0.887218533751)*x_ref[2]**o+(-0.066368425118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank1(self):
      """
      tests interpolate for rank 1 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.940260123916)*x[0]**o+(-0.977192570293)*x[0]+(-0.0910686020625)*x[1]**o+(0.464205016172)*x[1]
        arg[1]=(-0.634146043992)*x[0]**o+(0.00438663094072)*x[0]+(0.905996907036)*x[1]**o+(0.175291641923)*x[1]
        ref[0]=(-0.940260123916)*x_ref[0]**o+(-0.977192570293)*x_ref[0]+(-0.0910686020625)*x_ref[1]**o+(0.464205016172)*x_ref[1]
        ref[1]=(-0.634146043992)*x_ref[0]**o+(0.00438663094072)*x_ref[0]+(0.905996907036)*x_ref[1]**o+(0.175291641923)*x_ref[1]
      else:
        arg[0]=(0.0746302635373)*x[0]**o+(0.611568121627)*x[0]+(0.0392401043636)*x[1]**o+(0.846339953866)*x[1]+(0.775935477701)*x[2]**o+(0.309530355508)*x[2]
        arg[1]=(0.790143430408)*x[0]**o+(0.987665779965)*x[0]+(0.569313447766)*x[1]**o+(0.170690891282)*x[1]+(0.224015411231)*x[2]**o+(-0.670946134012)*x[2]
        ref[0]=(0.0746302635373)*x_ref[0]**o+(0.611568121627)*x_ref[0]+(0.0392401043636)*x_ref[1]**o+(0.846339953866)*x_ref[1]+(0.775935477701)*x_ref[2]**o+(0.309530355508)*x_ref[2]
        ref[1]=(0.790143430408)*x_ref[0]**o+(0.987665779965)*x_ref[0]+(0.569313447766)*x_ref[1]**o+(0.170690891282)*x_ref[1]+(0.224015411231)*x_ref[2]**o+(-0.670946134012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank2(self):
      """
      tests interpolate for rank 2 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.339464191741)*x[0]**o+(0.898634031554)*x[0]+(-0.67217412133)*x[1]**o+(-0.732882671617)*x[1]
        arg[0,1]=(0.475031035618)*x[0]**o+(-0.712271710437)*x[0]+(-0.346485473024)*x[1]**o+(0.936873073538)*x[1]
        arg[0,2]=(0.139217050193)*x[0]**o+(0.532762047045)*x[0]+(-0.411674842781)*x[1]**o+(-0.423096144938)*x[1]
        arg[0,3]=(-0.146193322436)*x[0]**o+(-0.227693594412)*x[0]+(0.74715898621)*x[1]**o+(0.538402714773)*x[1]
        arg[0,4]=(0.233316268084)*x[0]**o+(-0.0155557244918)*x[0]+(0.588017790807)*x[1]**o+(-0.58604454457)*x[1]
        arg[1,0]=(0.30473687856)*x[0]**o+(0.933154007362)*x[0]+(-0.0797692178609)*x[1]**o+(0.270021035922)*x[1]
        arg[1,1]=(-0.495613786111)*x[0]**o+(0.0857849462525)*x[0]+(0.330048664047)*x[1]**o+(0.0442768259753)*x[1]
        arg[1,2]=(0.203830377049)*x[0]**o+(-0.891094217974)*x[0]+(-0.643947141174)*x[1]**o+(-0.564315919126)*x[1]
        arg[1,3]=(-0.783514920199)*x[0]**o+(0.356882300706)*x[0]+(-0.307634481801)*x[1]**o+(0.521370775315)*x[1]
        arg[1,4]=(0.418545793821)*x[0]**o+(0.574742804708)*x[0]+(-0.340992300811)*x[1]**o+(-0.892118088815)*x[1]
        arg[2,0]=(-0.361915184269)*x[0]**o+(-0.834492406999)*x[0]+(0.750344593844)*x[1]**o+(-0.516981733186)*x[1]
        arg[2,1]=(-0.0123811923638)*x[0]**o+(-0.9808055138)*x[0]+(-0.176978785352)*x[1]**o+(0.519804799089)*x[1]
        arg[2,2]=(-0.901764228726)*x[0]**o+(-0.610547400097)*x[0]+(0.384870544403)*x[1]**o+(0.0308096944571)*x[1]
        arg[2,3]=(-0.163626872732)*x[0]**o+(0.278523098981)*x[0]+(-0.0544994735633)*x[1]**o+(-0.319354318824)*x[1]
        arg[2,4]=(0.558959525812)*x[0]**o+(0.435157723929)*x[0]+(0.758922109245)*x[1]**o+(-0.702427033542)*x[1]
        arg[3,0]=(0.832800430277)*x[0]**o+(-0.709285644484)*x[0]+(-0.201572915312)*x[1]**o+(0.398294738552)*x[1]
        arg[3,1]=(-0.836258924261)*x[0]**o+(-0.461452427039)*x[0]+(0.314485542425)*x[1]**o+(0.860807479051)*x[1]
        arg[3,2]=(-0.569661610862)*x[0]**o+(-0.46125190469)*x[0]+(0.807311492846)*x[1]**o+(0.72312107176)*x[1]
        arg[3,3]=(0.842126813179)*x[0]**o+(-0.714991954637)*x[0]+(-0.853100178283)*x[1]**o+(-0.644363753291)*x[1]
        arg[3,4]=(-0.0480369646571)*x[0]**o+(-0.330580215997)*x[0]+(0.609757645513)*x[1]**o+(0.915764017541)*x[1]
        ref[0,0]=(0.339464191741)*x_ref[0]**o+(0.898634031554)*x_ref[0]+(-0.67217412133)*x_ref[1]**o+(-0.732882671617)*x_ref[1]
        ref[0,1]=(0.475031035618)*x_ref[0]**o+(-0.712271710437)*x_ref[0]+(-0.346485473024)*x_ref[1]**o+(0.936873073538)*x_ref[1]
        ref[0,2]=(0.139217050193)*x_ref[0]**o+(0.532762047045)*x_ref[0]+(-0.411674842781)*x_ref[1]**o+(-0.423096144938)*x_ref[1]
        ref[0,3]=(-0.146193322436)*x_ref[0]**o+(-0.227693594412)*x_ref[0]+(0.74715898621)*x_ref[1]**o+(0.538402714773)*x_ref[1]
        ref[0,4]=(0.233316268084)*x_ref[0]**o+(-0.0155557244918)*x_ref[0]+(0.588017790807)*x_ref[1]**o+(-0.58604454457)*x_ref[1]
        ref[1,0]=(0.30473687856)*x_ref[0]**o+(0.933154007362)*x_ref[0]+(-0.0797692178609)*x_ref[1]**o+(0.270021035922)*x_ref[1]
        ref[1,1]=(-0.495613786111)*x_ref[0]**o+(0.0857849462525)*x_ref[0]+(0.330048664047)*x_ref[1]**o+(0.0442768259753)*x_ref[1]
        ref[1,2]=(0.203830377049)*x_ref[0]**o+(-0.891094217974)*x_ref[0]+(-0.643947141174)*x_ref[1]**o+(-0.564315919126)*x_ref[1]
        ref[1,3]=(-0.783514920199)*x_ref[0]**o+(0.356882300706)*x_ref[0]+(-0.307634481801)*x_ref[1]**o+(0.521370775315)*x_ref[1]
        ref[1,4]=(0.418545793821)*x_ref[0]**o+(0.574742804708)*x_ref[0]+(-0.340992300811)*x_ref[1]**o+(-0.892118088815)*x_ref[1]
        ref[2,0]=(-0.361915184269)*x_ref[0]**o+(-0.834492406999)*x_ref[0]+(0.750344593844)*x_ref[1]**o+(-0.516981733186)*x_ref[1]
        ref[2,1]=(-0.0123811923638)*x_ref[0]**o+(-0.9808055138)*x_ref[0]+(-0.176978785352)*x_ref[1]**o+(0.519804799089)*x_ref[1]
        ref[2,2]=(-0.901764228726)*x_ref[0]**o+(-0.610547400097)*x_ref[0]+(0.384870544403)*x_ref[1]**o+(0.0308096944571)*x_ref[1]
        ref[2,3]=(-0.163626872732)*x_ref[0]**o+(0.278523098981)*x_ref[0]+(-0.0544994735633)*x_ref[1]**o+(-0.319354318824)*x_ref[1]
        ref[2,4]=(0.558959525812)*x_ref[0]**o+(0.435157723929)*x_ref[0]+(0.758922109245)*x_ref[1]**o+(-0.702427033542)*x_ref[1]
        ref[3,0]=(0.832800430277)*x_ref[0]**o+(-0.709285644484)*x_ref[0]+(-0.201572915312)*x_ref[1]**o+(0.398294738552)*x_ref[1]
        ref[3,1]=(-0.836258924261)*x_ref[0]**o+(-0.461452427039)*x_ref[0]+(0.314485542425)*x_ref[1]**o+(0.860807479051)*x_ref[1]
        ref[3,2]=(-0.569661610862)*x_ref[0]**o+(-0.46125190469)*x_ref[0]+(0.807311492846)*x_ref[1]**o+(0.72312107176)*x_ref[1]
        ref[3,3]=(0.842126813179)*x_ref[0]**o+(-0.714991954637)*x_ref[0]+(-0.853100178283)*x_ref[1]**o+(-0.644363753291)*x_ref[1]
        ref[3,4]=(-0.0480369646571)*x_ref[0]**o+(-0.330580215997)*x_ref[0]+(0.609757645513)*x_ref[1]**o+(0.915764017541)*x_ref[1]
      else:
        arg[0,0]=(-0.256088293513)*x[0]**o+(0.980138068946)*x[0]+(0.375373123738)*x[1]**o+(0.118268751779)*x[1]+(-0.848800006681)*x[2]**o+(0.861808952535)*x[2]
        arg[0,1]=(-0.0658163161725)*x[0]**o+(-0.21582867866)*x[0]+(0.99787757796)*x[1]**o+(0.0175357022646)*x[1]+(-0.305438242469)*x[2]**o+(0.114385992332)*x[2]
        arg[0,2]=(-0.296948302927)*x[0]**o+(0.146852900649)*x[0]+(0.386910961544)*x[1]**o+(0.353651457394)*x[1]+(0.420964895802)*x[2]**o+(0.71145244268)*x[2]
        arg[0,3]=(0.61102494979)*x[0]**o+(-0.316055853885)*x[0]+(-0.117200390837)*x[1]**o+(-0.808630441349)*x[1]+(0.0914828526025)*x[2]**o+(0.305837943839)*x[2]
        arg[0,4]=(-0.22863340097)*x[0]**o+(0.398083421617)*x[0]+(-0.39971917463)*x[1]**o+(-0.0307312416932)*x[1]+(0.510443863383)*x[2]**o+(-0.129495846726)*x[2]
        arg[1,0]=(0.423968874298)*x[0]**o+(0.371480448801)*x[0]+(-0.934565508125)*x[1]**o+(0.890972208341)*x[1]+(-0.169067233411)*x[2]**o+(-0.757264272484)*x[2]
        arg[1,1]=(0.999152573666)*x[0]**o+(-0.659676649566)*x[0]+(0.993153643567)*x[1]**o+(0.350141892848)*x[1]+(-0.493784056536)*x[2]**o+(-0.966494205893)*x[2]
        arg[1,2]=(-0.141337889298)*x[0]**o+(0.689935513838)*x[0]+(0.335444607712)*x[1]**o+(-0.733351283662)*x[1]+(0.323775887172)*x[2]**o+(0.314099156544)*x[2]
        arg[1,3]=(-0.157132412118)*x[0]**o+(0.602468153313)*x[0]+(-0.618134886591)*x[1]**o+(-0.404425578444)*x[1]+(-0.5069585924)*x[2]**o+(0.616829810596)*x[2]
        arg[1,4]=(-0.107902610028)*x[0]**o+(0.664787826515)*x[0]+(-0.19940280262)*x[1]**o+(0.461794954077)*x[1]+(-0.0516671817578)*x[2]**o+(0.426165289622)*x[2]
        arg[2,0]=(-0.568061237616)*x[0]**o+(-0.744340542576)*x[0]+(-0.0125924903194)*x[1]**o+(0.710153866172)*x[1]+(0.979871553596)*x[2]**o+(0.945862455036)*x[2]
        arg[2,1]=(0.201599120538)*x[0]**o+(-0.11057289344)*x[0]+(0.522031147141)*x[1]**o+(0.194694594101)*x[1]+(-0.771997994079)*x[2]**o+(-0.513574957413)*x[2]
        arg[2,2]=(0.401979025382)*x[0]**o+(-0.743641669014)*x[0]+(-0.796071557874)*x[1]**o+(-0.352651095657)*x[1]+(-0.564117536877)*x[2]**o+(0.952729715875)*x[2]
        arg[2,3]=(-0.526162893019)*x[0]**o+(0.691019471759)*x[0]+(-0.68646084088)*x[1]**o+(0.839630438544)*x[1]+(-0.491997512287)*x[2]**o+(-0.405069600537)*x[2]
        arg[2,4]=(0.117664099703)*x[0]**o+(-0.807496546596)*x[0]+(0.607810703702)*x[1]**o+(-0.696118154009)*x[1]+(-0.156011683803)*x[2]**o+(-0.26100420077)*x[2]
        arg[3,0]=(0.280066174527)*x[0]**o+(-0.225940135983)*x[0]+(-0.338716163489)*x[1]**o+(0.628405669096)*x[1]+(-0.890543405425)*x[2]**o+(-0.142082524497)*x[2]
        arg[3,1]=(-0.659241351226)*x[0]**o+(0.275483386126)*x[0]+(-0.318352239058)*x[1]**o+(0.168536648793)*x[1]+(0.539555610218)*x[2]**o+(-0.95280560435)*x[2]
        arg[3,2]=(0.673861565841)*x[0]**o+(-0.0303873544723)*x[0]+(-0.638241094445)*x[1]**o+(0.664488908579)*x[1]+(-0.789484082135)*x[2]**o+(-0.982437408359)*x[2]
        arg[3,3]=(-0.902346309293)*x[0]**o+(0.818106160371)*x[0]+(-0.410074078135)*x[1]**o+(0.977554592344)*x[1]+(-0.00195907247124)*x[2]**o+(0.94762151647)*x[2]
        arg[3,4]=(0.27272813538)*x[0]**o+(0.0563423990113)*x[0]+(0.254612551508)*x[1]**o+(-0.901948547799)*x[1]+(0.263613075865)*x[2]**o+(0.879691314116)*x[2]
        ref[0,0]=(-0.256088293513)*x_ref[0]**o+(0.980138068946)*x_ref[0]+(0.375373123738)*x_ref[1]**o+(0.118268751779)*x_ref[1]+(-0.848800006681)*x_ref[2]**o+(0.861808952535)*x_ref[2]
        ref[0,1]=(-0.0658163161725)*x_ref[0]**o+(-0.21582867866)*x_ref[0]+(0.99787757796)*x_ref[1]**o+(0.0175357022646)*x_ref[1]+(-0.305438242469)*x_ref[2]**o+(0.114385992332)*x_ref[2]
        ref[0,2]=(-0.296948302927)*x_ref[0]**o+(0.146852900649)*x_ref[0]+(0.386910961544)*x_ref[1]**o+(0.353651457394)*x_ref[1]+(0.420964895802)*x_ref[2]**o+(0.71145244268)*x_ref[2]
        ref[0,3]=(0.61102494979)*x_ref[0]**o+(-0.316055853885)*x_ref[0]+(-0.117200390837)*x_ref[1]**o+(-0.808630441349)*x_ref[1]+(0.0914828526025)*x_ref[2]**o+(0.305837943839)*x_ref[2]
        ref[0,4]=(-0.22863340097)*x_ref[0]**o+(0.398083421617)*x_ref[0]+(-0.39971917463)*x_ref[1]**o+(-0.0307312416932)*x_ref[1]+(0.510443863383)*x_ref[2]**o+(-0.129495846726)*x_ref[2]
        ref[1,0]=(0.423968874298)*x_ref[0]**o+(0.371480448801)*x_ref[0]+(-0.934565508125)*x_ref[1]**o+(0.890972208341)*x_ref[1]+(-0.169067233411)*x_ref[2]**o+(-0.757264272484)*x_ref[2]
        ref[1,1]=(0.999152573666)*x_ref[0]**o+(-0.659676649566)*x_ref[0]+(0.993153643567)*x_ref[1]**o+(0.350141892848)*x_ref[1]+(-0.493784056536)*x_ref[2]**o+(-0.966494205893)*x_ref[2]
        ref[1,2]=(-0.141337889298)*x_ref[0]**o+(0.689935513838)*x_ref[0]+(0.335444607712)*x_ref[1]**o+(-0.733351283662)*x_ref[1]+(0.323775887172)*x_ref[2]**o+(0.314099156544)*x_ref[2]
        ref[1,3]=(-0.157132412118)*x_ref[0]**o+(0.602468153313)*x_ref[0]+(-0.618134886591)*x_ref[1]**o+(-0.404425578444)*x_ref[1]+(-0.5069585924)*x_ref[2]**o+(0.616829810596)*x_ref[2]
        ref[1,4]=(-0.107902610028)*x_ref[0]**o+(0.664787826515)*x_ref[0]+(-0.19940280262)*x_ref[1]**o+(0.461794954077)*x_ref[1]+(-0.0516671817578)*x_ref[2]**o+(0.426165289622)*x_ref[2]
        ref[2,0]=(-0.568061237616)*x_ref[0]**o+(-0.744340542576)*x_ref[0]+(-0.0125924903194)*x_ref[1]**o+(0.710153866172)*x_ref[1]+(0.979871553596)*x_ref[2]**o+(0.945862455036)*x_ref[2]
        ref[2,1]=(0.201599120538)*x_ref[0]**o+(-0.11057289344)*x_ref[0]+(0.522031147141)*x_ref[1]**o+(0.194694594101)*x_ref[1]+(-0.771997994079)*x_ref[2]**o+(-0.513574957413)*x_ref[2]
        ref[2,2]=(0.401979025382)*x_ref[0]**o+(-0.743641669014)*x_ref[0]+(-0.796071557874)*x_ref[1]**o+(-0.352651095657)*x_ref[1]+(-0.564117536877)*x_ref[2]**o+(0.952729715875)*x_ref[2]
        ref[2,3]=(-0.526162893019)*x_ref[0]**o+(0.691019471759)*x_ref[0]+(-0.68646084088)*x_ref[1]**o+(0.839630438544)*x_ref[1]+(-0.491997512287)*x_ref[2]**o+(-0.405069600537)*x_ref[2]
        ref[2,4]=(0.117664099703)*x_ref[0]**o+(-0.807496546596)*x_ref[0]+(0.607810703702)*x_ref[1]**o+(-0.696118154009)*x_ref[1]+(-0.156011683803)*x_ref[2]**o+(-0.26100420077)*x_ref[2]
        ref[3,0]=(0.280066174527)*x_ref[0]**o+(-0.225940135983)*x_ref[0]+(-0.338716163489)*x_ref[1]**o+(0.628405669096)*x_ref[1]+(-0.890543405425)*x_ref[2]**o+(-0.142082524497)*x_ref[2]
        ref[3,1]=(-0.659241351226)*x_ref[0]**o+(0.275483386126)*x_ref[0]+(-0.318352239058)*x_ref[1]**o+(0.168536648793)*x_ref[1]+(0.539555610218)*x_ref[2]**o+(-0.95280560435)*x_ref[2]
        ref[3,2]=(0.673861565841)*x_ref[0]**o+(-0.0303873544723)*x_ref[0]+(-0.638241094445)*x_ref[1]**o+(0.664488908579)*x_ref[1]+(-0.789484082135)*x_ref[2]**o+(-0.982437408359)*x_ref[2]
        ref[3,3]=(-0.902346309293)*x_ref[0]**o+(0.818106160371)*x_ref[0]+(-0.410074078135)*x_ref[1]**o+(0.977554592344)*x_ref[1]+(-0.00195907247124)*x_ref[2]**o+(0.94762151647)*x_ref[2]
        ref[3,4]=(0.27272813538)*x_ref[0]**o+(0.0563423990113)*x_ref[0]+(0.254612551508)*x_ref[1]**o+(-0.901948547799)*x_ref[1]+(0.263613075865)*x_ref[2]**o+(0.879691314116)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank3(self):
      """
      tests interpolate for rank 3 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.023497804886)*x[0]**o+(0.934780781287)*x[0]+(0.488935915319)*x[1]**o+(-0.843152487959)*x[1]
        arg[0,0,1]=(0.074870516034)*x[0]**o+(-0.515115482624)*x[0]+(-0.582499912837)*x[1]**o+(-0.148911376041)*x[1]
        arg[0,1,0]=(0.0364775990427)*x[0]**o+(-0.258379408121)*x[0]+(0.109704222415)*x[1]**o+(-0.0149203380717)*x[1]
        arg[0,1,1]=(0.749804547117)*x[0]**o+(-0.65162531688)*x[0]+(-0.529952602569)*x[1]**o+(0.901735743956)*x[1]
        arg[1,0,0]=(0.74941277085)*x[0]**o+(-0.878903916612)*x[0]+(0.866605680273)*x[1]**o+(0.182572622995)*x[1]
        arg[1,0,1]=(-0.787476915117)*x[0]**o+(0.945314823283)*x[0]+(0.165599597226)*x[1]**o+(0.733712603726)*x[1]
        arg[1,1,0]=(0.696240085313)*x[0]**o+(0.775501454572)*x[0]+(0.225131553571)*x[1]**o+(-0.75288780057)*x[1]
        arg[1,1,1]=(-0.352416785284)*x[0]**o+(-0.681447081047)*x[0]+(0.164273429957)*x[1]**o+(-0.0977235074729)*x[1]
        arg[2,0,0]=(-0.484651384894)*x[0]**o+(0.162048679786)*x[0]+(0.433731200543)*x[1]**o+(-0.2255104779)*x[1]
        arg[2,0,1]=(0.230113349848)*x[0]**o+(0.674218493869)*x[0]+(0.124852942645)*x[1]**o+(0.361278270173)*x[1]
        arg[2,1,0]=(-0.604246197352)*x[0]**o+(0.26763755318)*x[0]+(-0.821482774631)*x[1]**o+(-0.821813801109)*x[1]
        arg[2,1,1]=(0.71669276429)*x[0]**o+(0.861782230871)*x[0]+(-0.837503788825)*x[1]**o+(0.0437253153979)*x[1]
        arg[3,0,0]=(-0.0941343724431)*x[0]**o+(0.376175369776)*x[0]+(-0.455752787585)*x[1]**o+(0.777480094583)*x[1]
        arg[3,0,1]=(-0.730897141543)*x[0]**o+(0.750046665085)*x[0]+(0.855742434657)*x[1]**o+(0.941826664412)*x[1]
        arg[3,1,0]=(-0.788757907886)*x[0]**o+(-0.0255067018415)*x[0]+(-0.875292088318)*x[1]**o+(0.401677848256)*x[1]
        arg[3,1,1]=(-0.980570910972)*x[0]**o+(0.936108649845)*x[0]+(-0.0463362409126)*x[1]**o+(0.630408734656)*x[1]
        arg[4,0,0]=(-0.72642075391)*x[0]**o+(0.283788406364)*x[0]+(-0.119645741265)*x[1]**o+(0.66948892422)*x[1]
        arg[4,0,1]=(0.178235706017)*x[0]**o+(0.693466107843)*x[0]+(-0.585602683437)*x[1]**o+(-0.493386713436)*x[1]
        arg[4,1,0]=(0.348039277129)*x[0]**o+(-0.806776084421)*x[0]+(-0.73591780841)*x[1]**o+(-0.310246367505)*x[1]
        arg[4,1,1]=(0.993294877418)*x[0]**o+(-0.683114302069)*x[0]+(0.0679022755308)*x[1]**o+(0.341981988936)*x[1]
        arg[5,0,0]=(-0.807380701959)*x[0]**o+(-0.78936779476)*x[0]+(0.142089012614)*x[1]**o+(-0.129335524031)*x[1]
        arg[5,0,1]=(-0.855714388031)*x[0]**o+(0.00825196217877)*x[0]+(0.310039590415)*x[1]**o+(-0.242861862397)*x[1]
        arg[5,1,0]=(-0.0908183346277)*x[0]**o+(-0.941876543235)*x[0]+(0.981459557493)*x[1]**o+(0.559563798858)*x[1]
        arg[5,1,1]=(-0.852223567371)*x[0]**o+(-0.272541886199)*x[0]+(-0.873908247032)*x[1]**o+(-0.387571769068)*x[1]
        ref[0,0,0]=(-0.023497804886)*x_ref[0]**o+(0.934780781287)*x_ref[0]+(0.488935915319)*x_ref[1]**o+(-0.843152487959)*x_ref[1]
        ref[0,0,1]=(0.074870516034)*x_ref[0]**o+(-0.515115482624)*x_ref[0]+(-0.582499912837)*x_ref[1]**o+(-0.148911376041)*x_ref[1]
        ref[0,1,0]=(0.0364775990427)*x_ref[0]**o+(-0.258379408121)*x_ref[0]+(0.109704222415)*x_ref[1]**o+(-0.0149203380717)*x_ref[1]
        ref[0,1,1]=(0.749804547117)*x_ref[0]**o+(-0.65162531688)*x_ref[0]+(-0.529952602569)*x_ref[1]**o+(0.901735743956)*x_ref[1]
        ref[1,0,0]=(0.74941277085)*x_ref[0]**o+(-0.878903916612)*x_ref[0]+(0.866605680273)*x_ref[1]**o+(0.182572622995)*x_ref[1]
        ref[1,0,1]=(-0.787476915117)*x_ref[0]**o+(0.945314823283)*x_ref[0]+(0.165599597226)*x_ref[1]**o+(0.733712603726)*x_ref[1]
        ref[1,1,0]=(0.696240085313)*x_ref[0]**o+(0.775501454572)*x_ref[0]+(0.225131553571)*x_ref[1]**o+(-0.75288780057)*x_ref[1]
        ref[1,1,1]=(-0.352416785284)*x_ref[0]**o+(-0.681447081047)*x_ref[0]+(0.164273429957)*x_ref[1]**o+(-0.0977235074729)*x_ref[1]
        ref[2,0,0]=(-0.484651384894)*x_ref[0]**o+(0.162048679786)*x_ref[0]+(0.433731200543)*x_ref[1]**o+(-0.2255104779)*x_ref[1]
        ref[2,0,1]=(0.230113349848)*x_ref[0]**o+(0.674218493869)*x_ref[0]+(0.124852942645)*x_ref[1]**o+(0.361278270173)*x_ref[1]
        ref[2,1,0]=(-0.604246197352)*x_ref[0]**o+(0.26763755318)*x_ref[0]+(-0.821482774631)*x_ref[1]**o+(-0.821813801109)*x_ref[1]
        ref[2,1,1]=(0.71669276429)*x_ref[0]**o+(0.861782230871)*x_ref[0]+(-0.837503788825)*x_ref[1]**o+(0.0437253153979)*x_ref[1]
        ref[3,0,0]=(-0.0941343724431)*x_ref[0]**o+(0.376175369776)*x_ref[0]+(-0.455752787585)*x_ref[1]**o+(0.777480094583)*x_ref[1]
        ref[3,0,1]=(-0.730897141543)*x_ref[0]**o+(0.750046665085)*x_ref[0]+(0.855742434657)*x_ref[1]**o+(0.941826664412)*x_ref[1]
        ref[3,1,0]=(-0.788757907886)*x_ref[0]**o+(-0.0255067018415)*x_ref[0]+(-0.875292088318)*x_ref[1]**o+(0.401677848256)*x_ref[1]
        ref[3,1,1]=(-0.980570910972)*x_ref[0]**o+(0.936108649845)*x_ref[0]+(-0.0463362409126)*x_ref[1]**o+(0.630408734656)*x_ref[1]
        ref[4,0,0]=(-0.72642075391)*x_ref[0]**o+(0.283788406364)*x_ref[0]+(-0.119645741265)*x_ref[1]**o+(0.66948892422)*x_ref[1]
        ref[4,0,1]=(0.178235706017)*x_ref[0]**o+(0.693466107843)*x_ref[0]+(-0.585602683437)*x_ref[1]**o+(-0.493386713436)*x_ref[1]
        ref[4,1,0]=(0.348039277129)*x_ref[0]**o+(-0.806776084421)*x_ref[0]+(-0.73591780841)*x_ref[1]**o+(-0.310246367505)*x_ref[1]
        ref[4,1,1]=(0.993294877418)*x_ref[0]**o+(-0.683114302069)*x_ref[0]+(0.0679022755308)*x_ref[1]**o+(0.341981988936)*x_ref[1]
        ref[5,0,0]=(-0.807380701959)*x_ref[0]**o+(-0.78936779476)*x_ref[0]+(0.142089012614)*x_ref[1]**o+(-0.129335524031)*x_ref[1]
        ref[5,0,1]=(-0.855714388031)*x_ref[0]**o+(0.00825196217877)*x_ref[0]+(0.310039590415)*x_ref[1]**o+(-0.242861862397)*x_ref[1]
        ref[5,1,0]=(-0.0908183346277)*x_ref[0]**o+(-0.941876543235)*x_ref[0]+(0.981459557493)*x_ref[1]**o+(0.559563798858)*x_ref[1]
        ref[5,1,1]=(-0.852223567371)*x_ref[0]**o+(-0.272541886199)*x_ref[0]+(-0.873908247032)*x_ref[1]**o+(-0.387571769068)*x_ref[1]
      else:
        arg[0,0,0]=(0.418981395334)*x[0]**o+(-0.720305875042)*x[0]+(0.876880439526)*x[1]**o+(-0.17785803746)*x[1]+(0.779797986429)*x[2]**o+(0.363827072746)*x[2]
        arg[0,0,1]=(0.159567966427)*x[0]**o+(-0.614151008569)*x[0]+(0.249982224851)*x[1]**o+(-0.119598774468)*x[1]+(-0.66603578755)*x[2]**o+(0.202190479552)*x[2]
        arg[0,1,0]=(0.337868020939)*x[0]**o+(-0.0828148711399)*x[0]+(-0.0631659329293)*x[1]**o+(-0.725100695213)*x[1]+(0.852068398846)*x[2]**o+(-0.951116521211)*x[2]
        arg[0,1,1]=(0.710866728186)*x[0]**o+(-0.517133891599)*x[0]+(0.320028321875)*x[1]**o+(0.0331233805016)*x[1]+(-0.0945382588101)*x[2]**o+(-0.795725457058)*x[2]
        arg[1,0,0]=(0.261503329866)*x[0]**o+(0.697062723662)*x[0]+(0.926550326146)*x[1]**o+(-0.318704885103)*x[1]+(-0.76904956799)*x[2]**o+(0.0750521033639)*x[2]
        arg[1,0,1]=(-0.594460213676)*x[0]**o+(0.356692399969)*x[0]+(-0.611558953982)*x[1]**o+(0.289469152887)*x[1]+(-0.73222413782)*x[2]**o+(-0.576883793051)*x[2]
        arg[1,1,0]=(-0.399324648161)*x[0]**o+(0.975428000926)*x[0]+(-0.889762441468)*x[1]**o+(-0.137573068244)*x[1]+(0.252276361744)*x[2]**o+(-0.554435968553)*x[2]
        arg[1,1,1]=(0.677933864695)*x[0]**o+(-0.19825134391)*x[0]+(0.556783296976)*x[1]**o+(0.0349255641106)*x[1]+(-0.387378575139)*x[2]**o+(0.889628545207)*x[2]
        arg[2,0,0]=(-0.839356850436)*x[0]**o+(0.0186947034157)*x[0]+(-0.264793716814)*x[1]**o+(-0.767468006197)*x[1]+(0.91661695419)*x[2]**o+(-0.374036622497)*x[2]
        arg[2,0,1]=(-0.030357536408)*x[0]**o+(0.57424762933)*x[0]+(-0.590115256932)*x[1]**o+(-0.176169776757)*x[1]+(-0.502845936823)*x[2]**o+(0.0399243413515)*x[2]
        arg[2,1,0]=(0.257324735469)*x[0]**o+(-0.686083724325)*x[0]+(-0.166609564696)*x[1]**o+(-0.132181292726)*x[1]+(0.10419146703)*x[2]**o+(-0.827571838114)*x[2]
        arg[2,1,1]=(0.339465982319)*x[0]**o+(0.766586466841)*x[0]+(0.277538823819)*x[1]**o+(0.222607512697)*x[1]+(-0.909105282607)*x[2]**o+(0.463729842225)*x[2]
        arg[3,0,0]=(-0.00132043337846)*x[0]**o+(0.670451981959)*x[0]+(-0.0122245380346)*x[1]**o+(-0.922535904926)*x[1]+(0.49486742)*x[2]**o+(0.0785274524546)*x[2]
        arg[3,0,1]=(0.0360876610851)*x[0]**o+(0.69319604661)*x[0]+(0.980826095079)*x[1]**o+(0.167689443923)*x[1]+(0.561634492527)*x[2]**o+(0.996948790413)*x[2]
        arg[3,1,0]=(-0.238705628819)*x[0]**o+(-0.915859436874)*x[0]+(0.942122040692)*x[1]**o+(0.971025407162)*x[1]+(-0.965945363654)*x[2]**o+(0.136537412741)*x[2]
        arg[3,1,1]=(-0.672314439537)*x[0]**o+(0.792647645097)*x[0]+(-0.432226224113)*x[1]**o+(0.424246857841)*x[1]+(-0.639842531517)*x[2]**o+(0.0570057507731)*x[2]
        arg[4,0,0]=(0.450831267081)*x[0]**o+(0.75992870254)*x[0]+(0.768779936214)*x[1]**o+(-0.0135947286596)*x[1]+(0.126345191973)*x[2]**o+(-0.619763308655)*x[2]
        arg[4,0,1]=(-0.325005948888)*x[0]**o+(-0.0655421631802)*x[0]+(0.607182269636)*x[1]**o+(-0.631044141835)*x[1]+(-0.356580309429)*x[2]**o+(-0.593198006905)*x[2]
        arg[4,1,0]=(0.304932527518)*x[0]**o+(-0.763995484499)*x[0]+(0.884855995183)*x[1]**o+(-0.42180322102)*x[1]+(0.605742059302)*x[2]**o+(0.448553789891)*x[2]
        arg[4,1,1]=(-0.787145437063)*x[0]**o+(-0.465220934558)*x[0]+(-0.370336760606)*x[1]**o+(-0.854850692759)*x[1]+(0.482896516171)*x[2]**o+(-0.551567519291)*x[2]
        arg[5,0,0]=(-0.537738873639)*x[0]**o+(-0.213268842545)*x[0]+(0.492208051415)*x[1]**o+(-0.609304634017)*x[1]+(-0.487753207718)*x[2]**o+(-0.70650655894)*x[2]
        arg[5,0,1]=(0.600359837292)*x[0]**o+(-0.604421599924)*x[0]+(-0.051356466466)*x[1]**o+(-0.168592544661)*x[1]+(0.789340570036)*x[2]**o+(0.910249026599)*x[2]
        arg[5,1,0]=(0.264506197816)*x[0]**o+(-0.711775430363)*x[0]+(0.384666624979)*x[1]**o+(0.0926714847986)*x[1]+(-0.551232557648)*x[2]**o+(0.758927448717)*x[2]
        arg[5,1,1]=(0.894998835449)*x[0]**o+(-0.466338210072)*x[0]+(-0.522475632898)*x[1]**o+(0.766949173671)*x[1]+(0.915063619607)*x[2]**o+(-0.270696846692)*x[2]
        ref[0,0,0]=(0.418981395334)*x_ref[0]**o+(-0.720305875042)*x_ref[0]+(0.876880439526)*x_ref[1]**o+(-0.17785803746)*x_ref[1]+(0.779797986429)*x_ref[2]**o+(0.363827072746)*x_ref[2]
        ref[0,0,1]=(0.159567966427)*x_ref[0]**o+(-0.614151008569)*x_ref[0]+(0.249982224851)*x_ref[1]**o+(-0.119598774468)*x_ref[1]+(-0.66603578755)*x_ref[2]**o+(0.202190479552)*x_ref[2]
        ref[0,1,0]=(0.337868020939)*x_ref[0]**o+(-0.0828148711399)*x_ref[0]+(-0.0631659329293)*x_ref[1]**o+(-0.725100695213)*x_ref[1]+(0.852068398846)*x_ref[2]**o+(-0.951116521211)*x_ref[2]
        ref[0,1,1]=(0.710866728186)*x_ref[0]**o+(-0.517133891599)*x_ref[0]+(0.320028321875)*x_ref[1]**o+(0.0331233805016)*x_ref[1]+(-0.0945382588101)*x_ref[2]**o+(-0.795725457058)*x_ref[2]
        ref[1,0,0]=(0.261503329866)*x_ref[0]**o+(0.697062723662)*x_ref[0]+(0.926550326146)*x_ref[1]**o+(-0.318704885103)*x_ref[1]+(-0.76904956799)*x_ref[2]**o+(0.0750521033639)*x_ref[2]
        ref[1,0,1]=(-0.594460213676)*x_ref[0]**o+(0.356692399969)*x_ref[0]+(-0.611558953982)*x_ref[1]**o+(0.289469152887)*x_ref[1]+(-0.73222413782)*x_ref[2]**o+(-0.576883793051)*x_ref[2]
        ref[1,1,0]=(-0.399324648161)*x_ref[0]**o+(0.975428000926)*x_ref[0]+(-0.889762441468)*x_ref[1]**o+(-0.137573068244)*x_ref[1]+(0.252276361744)*x_ref[2]**o+(-0.554435968553)*x_ref[2]
        ref[1,1,1]=(0.677933864695)*x_ref[0]**o+(-0.19825134391)*x_ref[0]+(0.556783296976)*x_ref[1]**o+(0.0349255641106)*x_ref[1]+(-0.387378575139)*x_ref[2]**o+(0.889628545207)*x_ref[2]
        ref[2,0,0]=(-0.839356850436)*x_ref[0]**o+(0.0186947034157)*x_ref[0]+(-0.264793716814)*x_ref[1]**o+(-0.767468006197)*x_ref[1]+(0.91661695419)*x_ref[2]**o+(-0.374036622497)*x_ref[2]
        ref[2,0,1]=(-0.030357536408)*x_ref[0]**o+(0.57424762933)*x_ref[0]+(-0.590115256932)*x_ref[1]**o+(-0.176169776757)*x_ref[1]+(-0.502845936823)*x_ref[2]**o+(0.0399243413515)*x_ref[2]
        ref[2,1,0]=(0.257324735469)*x_ref[0]**o+(-0.686083724325)*x_ref[0]+(-0.166609564696)*x_ref[1]**o+(-0.132181292726)*x_ref[1]+(0.10419146703)*x_ref[2]**o+(-0.827571838114)*x_ref[2]
        ref[2,1,1]=(0.339465982319)*x_ref[0]**o+(0.766586466841)*x_ref[0]+(0.277538823819)*x_ref[1]**o+(0.222607512697)*x_ref[1]+(-0.909105282607)*x_ref[2]**o+(0.463729842225)*x_ref[2]
        ref[3,0,0]=(-0.00132043337846)*x_ref[0]**o+(0.670451981959)*x_ref[0]+(-0.0122245380346)*x_ref[1]**o+(-0.922535904926)*x_ref[1]+(0.49486742)*x_ref[2]**o+(0.0785274524546)*x_ref[2]
        ref[3,0,1]=(0.0360876610851)*x_ref[0]**o+(0.69319604661)*x_ref[0]+(0.980826095079)*x_ref[1]**o+(0.167689443923)*x_ref[1]+(0.561634492527)*x_ref[2]**o+(0.996948790413)*x_ref[2]
        ref[3,1,0]=(-0.238705628819)*x_ref[0]**o+(-0.915859436874)*x_ref[0]+(0.942122040692)*x_ref[1]**o+(0.971025407162)*x_ref[1]+(-0.965945363654)*x_ref[2]**o+(0.136537412741)*x_ref[2]
        ref[3,1,1]=(-0.672314439537)*x_ref[0]**o+(0.792647645097)*x_ref[0]+(-0.432226224113)*x_ref[1]**o+(0.424246857841)*x_ref[1]+(-0.639842531517)*x_ref[2]**o+(0.0570057507731)*x_ref[2]
        ref[4,0,0]=(0.450831267081)*x_ref[0]**o+(0.75992870254)*x_ref[0]+(0.768779936214)*x_ref[1]**o+(-0.0135947286596)*x_ref[1]+(0.126345191973)*x_ref[2]**o+(-0.619763308655)*x_ref[2]
        ref[4,0,1]=(-0.325005948888)*x_ref[0]**o+(-0.0655421631802)*x_ref[0]+(0.607182269636)*x_ref[1]**o+(-0.631044141835)*x_ref[1]+(-0.356580309429)*x_ref[2]**o+(-0.593198006905)*x_ref[2]
        ref[4,1,0]=(0.304932527518)*x_ref[0]**o+(-0.763995484499)*x_ref[0]+(0.884855995183)*x_ref[1]**o+(-0.42180322102)*x_ref[1]+(0.605742059302)*x_ref[2]**o+(0.448553789891)*x_ref[2]
        ref[4,1,1]=(-0.787145437063)*x_ref[0]**o+(-0.465220934558)*x_ref[0]+(-0.370336760606)*x_ref[1]**o+(-0.854850692759)*x_ref[1]+(0.482896516171)*x_ref[2]**o+(-0.551567519291)*x_ref[2]
        ref[5,0,0]=(-0.537738873639)*x_ref[0]**o+(-0.213268842545)*x_ref[0]+(0.492208051415)*x_ref[1]**o+(-0.609304634017)*x_ref[1]+(-0.487753207718)*x_ref[2]**o+(-0.70650655894)*x_ref[2]
        ref[5,0,1]=(0.600359837292)*x_ref[0]**o+(-0.604421599924)*x_ref[0]+(-0.051356466466)*x_ref[1]**o+(-0.168592544661)*x_ref[1]+(0.789340570036)*x_ref[2]**o+(0.910249026599)*x_ref[2]
        ref[5,1,0]=(0.264506197816)*x_ref[0]**o+(-0.711775430363)*x_ref[0]+(0.384666624979)*x_ref[1]**o+(0.0926714847986)*x_ref[1]+(-0.551232557648)*x_ref[2]**o+(0.758927448717)*x_ref[2]
        ref[5,1,1]=(0.894998835449)*x_ref[0]**o+(-0.466338210072)*x_ref[0]+(-0.522475632898)*x_ref[1]**o+(0.766949173671)*x_ref[1]+(0.915063619607)*x_ref[2]**o+(-0.270696846692)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromData_FunctionOnBoundary_rank4(self):
      """
      tests interpolate for rank 4 Data onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.240887319278)*x[0]**o+(-0.192166587545)*x[0]+(0.0176974287409)*x[1]**o+(0.100761748824)*x[1]
        arg[0,0,0,1]=(-0.549649845775)*x[0]**o+(-0.293969245646)*x[0]+(-0.605885325133)*x[1]**o+(0.799558451173)*x[1]
        arg[0,0,1,0]=(-0.899658769582)*x[0]**o+(0.484524193987)*x[0]+(0.437485870351)*x[1]**o+(-0.220549034927)*x[1]
        arg[0,0,1,1]=(-0.146539961793)*x[0]**o+(-0.278343486929)*x[0]+(0.0913912456082)*x[1]**o+(-0.751028975123)*x[1]
        arg[0,0,2,0]=(0.806344701194)*x[0]**o+(0.0537872824561)*x[0]+(-0.70407828379)*x[1]**o+(-0.922963486601)*x[1]
        arg[0,0,2,1]=(0.307202902495)*x[0]**o+(0.301603506731)*x[0]+(0.232983743616)*x[1]**o+(-0.744267542759)*x[1]
        arg[0,1,0,0]=(-0.550616060824)*x[0]**o+(-0.661367750254)*x[0]+(-0.695905895864)*x[1]**o+(-0.788045922398)*x[1]
        arg[0,1,0,1]=(-0.934169134035)*x[0]**o+(-0.843861105521)*x[0]+(-0.237877254439)*x[1]**o+(-0.364956821348)*x[1]
        arg[0,1,1,0]=(0.0950675750026)*x[0]**o+(-0.815510640594)*x[0]+(0.72526035968)*x[1]**o+(0.898675190106)*x[1]
        arg[0,1,1,1]=(0.715609066776)*x[0]**o+(0.371314845206)*x[0]+(0.326878867244)*x[1]**o+(-0.468798950199)*x[1]
        arg[0,1,2,0]=(0.392175939547)*x[0]**o+(0.105332682674)*x[0]+(0.997967455477)*x[1]**o+(0.0322491822496)*x[1]
        arg[0,1,2,1]=(-0.804498399187)*x[0]**o+(0.819653744739)*x[0]+(0.609610573703)*x[1]**o+(-0.623299047794)*x[1]
        arg[0,2,0,0]=(0.993877917386)*x[0]**o+(0.815952897899)*x[0]+(-0.258432174486)*x[1]**o+(0.209324387308)*x[1]
        arg[0,2,0,1]=(0.395946889364)*x[0]**o+(-0.859500539908)*x[0]+(-0.68495734785)*x[1]**o+(0.791813946302)*x[1]
        arg[0,2,1,0]=(0.154347661108)*x[0]**o+(0.0873140744322)*x[0]+(-0.423935109956)*x[1]**o+(-0.754297239226)*x[1]
        arg[0,2,1,1]=(-0.623998419484)*x[0]**o+(0.487300130941)*x[0]+(-0.323613219784)*x[1]**o+(0.744877983729)*x[1]
        arg[0,2,2,0]=(0.515686374371)*x[0]**o+(0.512572153346)*x[0]+(-0.641245783272)*x[1]**o+(-0.390660158759)*x[1]
        arg[0,2,2,1]=(0.24398238139)*x[0]**o+(-0.172020271678)*x[0]+(0.980197812701)*x[1]**o+(0.336941934503)*x[1]
        arg[0,3,0,0]=(0.0154584447615)*x[0]**o+(-0.547048191238)*x[0]+(-0.555319223005)*x[1]**o+(0.0972794598106)*x[1]
        arg[0,3,0,1]=(0.44744389335)*x[0]**o+(0.475046337874)*x[0]+(0.876228982066)*x[1]**o+(-0.0636405241175)*x[1]
        arg[0,3,1,0]=(0.392948465298)*x[0]**o+(-0.341342654182)*x[0]+(0.747185683968)*x[1]**o+(-0.99981172166)*x[1]
        arg[0,3,1,1]=(-0.172355321621)*x[0]**o+(-0.531391484795)*x[0]+(0.451654274019)*x[1]**o+(0.0795298667915)*x[1]
        arg[0,3,2,0]=(0.0841642474674)*x[0]**o+(0.308494435164)*x[0]+(0.53181918181)*x[1]**o+(0.480258067959)*x[1]
        arg[0,3,2,1]=(-0.222903054864)*x[0]**o+(-0.0330655869754)*x[0]+(0.227482525015)*x[1]**o+(0.536136212347)*x[1]
        arg[0,4,0,0]=(0.574798381284)*x[0]**o+(0.181354466484)*x[0]+(0.103049514171)*x[1]**o+(0.156282818835)*x[1]
        arg[0,4,0,1]=(0.478314255289)*x[0]**o+(0.224797222592)*x[0]+(0.307668606272)*x[1]**o+(-0.785262182937)*x[1]
        arg[0,4,1,0]=(0.519356961182)*x[0]**o+(0.646200458344)*x[0]+(0.0522618783181)*x[1]**o+(0.702335307324)*x[1]
        arg[0,4,1,1]=(-0.168756029541)*x[0]**o+(-0.00618364750666)*x[0]+(0.472704616213)*x[1]**o+(0.974257451871)*x[1]
        arg[0,4,2,0]=(-0.279196869573)*x[0]**o+(-0.333083174091)*x[0]+(-0.980089574428)*x[1]**o+(0.405273709405)*x[1]
        arg[0,4,2,1]=(-0.199431380673)*x[0]**o+(0.823294899866)*x[0]+(-0.174125117685)*x[1]**o+(-0.730500493887)*x[1]
        arg[1,0,0,0]=(-0.577281489635)*x[0]**o+(0.231906139898)*x[0]+(0.537128708963)*x[1]**o+(0.363444540898)*x[1]
        arg[1,0,0,1]=(0.204108541882)*x[0]**o+(-0.217661714888)*x[0]+(0.963384293759)*x[1]**o+(-0.0817519801351)*x[1]
        arg[1,0,1,0]=(0.540681325439)*x[0]**o+(-0.496036960808)*x[0]+(0.421037689661)*x[1]**o+(0.19399086996)*x[1]
        arg[1,0,1,1]=(0.482938542126)*x[0]**o+(-0.0725027675209)*x[0]+(-0.441517962054)*x[1]**o+(0.787445860714)*x[1]
        arg[1,0,2,0]=(0.0378928111257)*x[0]**o+(0.451368848019)*x[0]+(-0.228668284485)*x[1]**o+(0.505927229432)*x[1]
        arg[1,0,2,1]=(-0.59833518872)*x[0]**o+(0.956068253354)*x[0]+(0.667905995469)*x[1]**o+(0.963273292214)*x[1]
        arg[1,1,0,0]=(0.901637865236)*x[0]**o+(-0.642678971748)*x[0]+(-0.603915019524)*x[1]**o+(-0.208004993919)*x[1]
        arg[1,1,0,1]=(-0.227162661052)*x[0]**o+(-0.5816434303)*x[0]+(0.274961293801)*x[1]**o+(-0.614001328621)*x[1]
        arg[1,1,1,0]=(0.078438078548)*x[0]**o+(0.570191242883)*x[0]+(0.0952878212484)*x[1]**o+(0.844460205497)*x[1]
        arg[1,1,1,1]=(-0.901445811786)*x[0]**o+(-0.996376775344)*x[0]+(0.707470471613)*x[1]**o+(-0.187517936674)*x[1]
        arg[1,1,2,0]=(-0.66891443067)*x[0]**o+(0.0462057239324)*x[0]+(-0.825820539128)*x[1]**o+(-0.766932836843)*x[1]
        arg[1,1,2,1]=(-0.851428828811)*x[0]**o+(0.504484422346)*x[0]+(0.121378315545)*x[1]**o+(0.903947012757)*x[1]
        arg[1,2,0,0]=(-0.393827449768)*x[0]**o+(-0.251827904608)*x[0]+(0.8320040882)*x[1]**o+(0.57595149855)*x[1]
        arg[1,2,0,1]=(0.638738110712)*x[0]**o+(-0.439962152539)*x[0]+(-0.977211078995)*x[1]**o+(0.0528475212451)*x[1]
        arg[1,2,1,0]=(0.824313588941)*x[0]**o+(-0.589968996823)*x[0]+(-0.512156794152)*x[1]**o+(0.142809794443)*x[1]
        arg[1,2,1,1]=(-0.31789333147)*x[0]**o+(0.598839896949)*x[0]+(0.638319398651)*x[1]**o+(0.353625999317)*x[1]
        arg[1,2,2,0]=(-0.720472365012)*x[0]**o+(0.896538030051)*x[0]+(-0.224639169972)*x[1]**o+(-0.898299360738)*x[1]
        arg[1,2,2,1]=(0.123077707167)*x[0]**o+(0.15922851038)*x[0]+(0.678984783843)*x[1]**o+(-0.778188269852)*x[1]
        arg[1,3,0,0]=(0.344256301209)*x[0]**o+(0.884367738289)*x[0]+(0.00556636219289)*x[1]**o+(0.684361331195)*x[1]
        arg[1,3,0,1]=(-0.812245131065)*x[0]**o+(0.971365671228)*x[0]+(0.0696053239569)*x[1]**o+(-0.300266235138)*x[1]
        arg[1,3,1,0]=(0.470254588058)*x[0]**o+(0.0946825509913)*x[0]+(-0.415619255025)*x[1]**o+(-0.27381652858)*x[1]
        arg[1,3,1,1]=(-0.84555138241)*x[0]**o+(0.873510832091)*x[0]+(-0.138010338152)*x[1]**o+(-0.982274515316)*x[1]
        arg[1,3,2,0]=(0.695614136148)*x[0]**o+(-0.192588082771)*x[0]+(-0.21579127791)*x[1]**o+(-0.0335939736388)*x[1]
        arg[1,3,2,1]=(-0.633363463896)*x[0]**o+(-0.958320549993)*x[0]+(0.197331067481)*x[1]**o+(0.315089888409)*x[1]
        arg[1,4,0,0]=(-0.211531219945)*x[0]**o+(-0.399700721799)*x[0]+(0.578927638839)*x[1]**o+(-0.389034976195)*x[1]
        arg[1,4,0,1]=(-0.846399571893)*x[0]**o+(-0.952904755903)*x[0]+(-0.858399802693)*x[1]**o+(-0.23264354907)*x[1]
        arg[1,4,1,0]=(0.255822441586)*x[0]**o+(0.896237710393)*x[0]+(-0.90248008271)*x[1]**o+(0.654804848244)*x[1]
        arg[1,4,1,1]=(-0.112475508163)*x[0]**o+(-0.861697204453)*x[0]+(-0.782581931187)*x[1]**o+(-0.129345341581)*x[1]
        arg[1,4,2,0]=(0.690095687033)*x[0]**o+(0.729317334019)*x[0]+(-0.00961483612288)*x[1]**o+(-0.288320551243)*x[1]
        arg[1,4,2,1]=(-0.648146433517)*x[0]**o+(-0.301246601263)*x[0]+(-0.490119612994)*x[1]**o+(0.231186845615)*x[1]
        arg[2,0,0,0]=(-0.645560891595)*x[0]**o+(-0.889067544988)*x[0]+(-0.193353414221)*x[1]**o+(-0.726988479344)*x[1]
        arg[2,0,0,1]=(-0.083614738484)*x[0]**o+(-0.822530927241)*x[0]+(0.660338934241)*x[1]**o+(0.0383821036508)*x[1]
        arg[2,0,1,0]=(-0.132250736664)*x[0]**o+(-0.571474370931)*x[0]+(-0.755931100342)*x[1]**o+(-0.145196865115)*x[1]
        arg[2,0,1,1]=(0.545219196727)*x[0]**o+(-0.846368787913)*x[0]+(-0.845706160842)*x[1]**o+(0.550703610989)*x[1]
        arg[2,0,2,0]=(-0.766303732195)*x[0]**o+(0.156822893641)*x[0]+(0.00954334254282)*x[1]**o+(0.201506144561)*x[1]
        arg[2,0,2,1]=(-0.274476449632)*x[0]**o+(-0.361690468185)*x[0]+(-0.816535747268)*x[1]**o+(-0.473913811701)*x[1]
        arg[2,1,0,0]=(-0.927592771482)*x[0]**o+(0.534615269814)*x[0]+(-0.571217133452)*x[1]**o+(0.57448299216)*x[1]
        arg[2,1,0,1]=(-0.298153953211)*x[0]**o+(-0.143636909462)*x[0]+(0.990660288864)*x[1]**o+(-0.495864284763)*x[1]
        arg[2,1,1,0]=(-0.110625727208)*x[0]**o+(-0.221057774317)*x[0]+(0.557195078358)*x[1]**o+(0.803162691808)*x[1]
        arg[2,1,1,1]=(-0.115589176775)*x[0]**o+(0.08253351782)*x[0]+(-0.595996958977)*x[1]**o+(0.501435717501)*x[1]
        arg[2,1,2,0]=(0.78764976658)*x[0]**o+(0.966729831711)*x[0]+(-0.585879575667)*x[1]**o+(0.341148176793)*x[1]
        arg[2,1,2,1]=(-0.551697945691)*x[0]**o+(0.545857249323)*x[0]+(-0.919321724495)*x[1]**o+(-0.286043327415)*x[1]
        arg[2,2,0,0]=(-0.669933890655)*x[0]**o+(-0.344862118495)*x[0]+(0.878331282597)*x[1]**o+(0.881893760718)*x[1]
        arg[2,2,0,1]=(-0.217876618414)*x[0]**o+(-0.289118007199)*x[0]+(0.513464171525)*x[1]**o+(0.310055899102)*x[1]
        arg[2,2,1,0]=(0.0472051379154)*x[0]**o+(0.42455388451)*x[0]+(-0.920571834891)*x[1]**o+(-0.779095076537)*x[1]
        arg[2,2,1,1]=(-0.299619304537)*x[0]**o+(-0.289777287294)*x[0]+(0.920505062988)*x[1]**o+(-0.243924179095)*x[1]
        arg[2,2,2,0]=(-0.819343530877)*x[0]**o+(0.975953690875)*x[0]+(-0.787239459241)*x[1]**o+(-0.199900734691)*x[1]
        arg[2,2,2,1]=(0.0980085780878)*x[0]**o+(0.642946515451)*x[0]+(0.951790535152)*x[1]**o+(-0.253914817366)*x[1]
        arg[2,3,0,0]=(-0.882075522824)*x[0]**o+(0.200934502682)*x[0]+(-0.83200464496)*x[1]**o+(0.941479345375)*x[1]
        arg[2,3,0,1]=(-0.615941782209)*x[0]**o+(-0.693242388092)*x[0]+(-0.640417278871)*x[1]**o+(-0.346459094803)*x[1]
        arg[2,3,1,0]=(-0.447919427952)*x[0]**o+(0.205055689141)*x[0]+(-0.453290032411)*x[1]**o+(0.879062836521)*x[1]
        arg[2,3,1,1]=(0.721133924016)*x[0]**o+(0.71294558459)*x[0]+(-0.250087045805)*x[1]**o+(0.605146721129)*x[1]
        arg[2,3,2,0]=(0.247782322921)*x[0]**o+(-0.956367425119)*x[0]+(-0.054045082005)*x[1]**o+(-0.128151832132)*x[1]
        arg[2,3,2,1]=(-0.427039259088)*x[0]**o+(-0.359849242839)*x[0]+(-0.244990934481)*x[1]**o+(-0.038020035629)*x[1]
        arg[2,4,0,0]=(-0.216409150834)*x[0]**o+(0.492547989921)*x[0]+(0.878317438469)*x[1]**o+(-0.112718719119)*x[1]
        arg[2,4,0,1]=(0.487403443102)*x[0]**o+(0.228155519468)*x[0]+(0.923150527516)*x[1]**o+(0.642654867908)*x[1]
        arg[2,4,1,0]=(0.916418483417)*x[0]**o+(-0.177198935173)*x[0]+(-0.790548940412)*x[1]**o+(0.87540167052)*x[1]
        arg[2,4,1,1]=(0.718502075538)*x[0]**o+(0.691254010569)*x[0]+(-0.0134430122729)*x[1]**o+(-0.739888770749)*x[1]
        arg[2,4,2,0]=(0.287825086861)*x[0]**o+(0.596621809347)*x[0]+(0.729784205391)*x[1]**o+(-0.25984257955)*x[1]
        arg[2,4,2,1]=(-0.515102480025)*x[0]**o+(-0.609951840015)*x[0]+(0.229834340075)*x[1]**o+(-0.727157816241)*x[1]
        arg[3,0,0,0]=(-0.349406091363)*x[0]**o+(0.802348451103)*x[0]+(-0.174421346276)*x[1]**o+(0.120711198036)*x[1]
        arg[3,0,0,1]=(0.367085169423)*x[0]**o+(-0.319683236434)*x[0]+(-0.590820542566)*x[1]**o+(0.607625953303)*x[1]
        arg[3,0,1,0]=(0.231259976433)*x[0]**o+(-0.493297408825)*x[0]+(0.919363640504)*x[1]**o+(-0.839841270429)*x[1]
        arg[3,0,1,1]=(0.57672135684)*x[0]**o+(0.882062879704)*x[0]+(-0.600979706444)*x[1]**o+(-0.704008344509)*x[1]
        arg[3,0,2,0]=(-0.208162114527)*x[0]**o+(0.593682243052)*x[0]+(-0.428436832487)*x[1]**o+(0.51007021604)*x[1]
        arg[3,0,2,1]=(0.315264672653)*x[0]**o+(-0.0835897376534)*x[0]+(-0.135315173106)*x[1]**o+(-0.646372309498)*x[1]
        arg[3,1,0,0]=(0.815505953752)*x[0]**o+(0.0625493643292)*x[0]+(0.383118433815)*x[1]**o+(0.17372720435)*x[1]
        arg[3,1,0,1]=(0.557295786969)*x[0]**o+(-0.568823508165)*x[0]+(-0.00874733267675)*x[1]**o+(-0.27454140598)*x[1]
        arg[3,1,1,0]=(0.649021026591)*x[0]**o+(0.232255030448)*x[0]+(0.845939296425)*x[1]**o+(0.116387378294)*x[1]
        arg[3,1,1,1]=(0.431124940343)*x[0]**o+(-0.0865937271969)*x[0]+(0.937911741288)*x[1]**o+(0.362850039234)*x[1]
        arg[3,1,2,0]=(0.155965414439)*x[0]**o+(0.110850421841)*x[0]+(0.244563080191)*x[1]**o+(0.266325498006)*x[1]
        arg[3,1,2,1]=(0.766813300935)*x[0]**o+(0.77492422889)*x[0]+(-0.0141436554714)*x[1]**o+(0.780444449009)*x[1]
        arg[3,2,0,0]=(0.619086019019)*x[0]**o+(-0.574130567601)*x[0]+(-0.417708159593)*x[1]**o+(0.707727284254)*x[1]
        arg[3,2,0,1]=(-0.845657409353)*x[0]**o+(0.705282507748)*x[0]+(0.124557418112)*x[1]**o+(-0.337493494733)*x[1]
        arg[3,2,1,0]=(0.114875146211)*x[0]**o+(-0.724480426892)*x[0]+(-0.0763476904405)*x[1]**o+(0.12064907114)*x[1]
        arg[3,2,1,1]=(-0.74994413376)*x[0]**o+(-0.916476035217)*x[0]+(-0.360412756599)*x[1]**o+(-0.917457363098)*x[1]
        arg[3,2,2,0]=(-0.446839470984)*x[0]**o+(-0.097704838316)*x[0]+(-0.653584364909)*x[1]**o+(0.561014757758)*x[1]
        arg[3,2,2,1]=(-0.178975301038)*x[0]**o+(-0.843669982667)*x[0]+(-0.566610094242)*x[1]**o+(0.213047649003)*x[1]
        arg[3,3,0,0]=(0.144115764355)*x[0]**o+(0.199435411954)*x[0]+(0.173334385199)*x[1]**o+(-0.942957035808)*x[1]
        arg[3,3,0,1]=(-0.417589926342)*x[0]**o+(-0.819188022852)*x[0]+(-0.68299160003)*x[1]**o+(-0.314851447894)*x[1]
        arg[3,3,1,0]=(0.737898219164)*x[0]**o+(0.612906277393)*x[0]+(-0.768349396729)*x[1]**o+(-0.192998755095)*x[1]
        arg[3,3,1,1]=(0.35433362007)*x[0]**o+(0.894174225208)*x[0]+(-0.450333030837)*x[1]**o+(0.819450589483)*x[1]
        arg[3,3,2,0]=(-0.455377131302)*x[0]**o+(-0.748183144687)*x[0]+(-0.482739450231)*x[1]**o+(0.300625929022)*x[1]
        arg[3,3,2,1]=(-0.602766693698)*x[0]**o+(-0.416288436985)*x[0]+(0.837501983536)*x[1]**o+(-0.745935308188)*x[1]
        arg[3,4,0,0]=(-0.706774010405)*x[0]**o+(0.407766818701)*x[0]+(-0.248964401443)*x[1]**o+(0.675466487229)*x[1]
        arg[3,4,0,1]=(0.512302883876)*x[0]**o+(0.0809544867105)*x[0]+(-0.204555203536)*x[1]**o+(-0.628574073749)*x[1]
        arg[3,4,1,0]=(0.126376796421)*x[0]**o+(0.174607814129)*x[0]+(0.68867194136)*x[1]**o+(0.940025934568)*x[1]
        arg[3,4,1,1]=(0.583042444739)*x[0]**o+(-0.786981763836)*x[0]+(0.319397032405)*x[1]**o+(-0.324993351432)*x[1]
        arg[3,4,2,0]=(-0.941997068836)*x[0]**o+(-0.892825520378)*x[0]+(0.275995419027)*x[1]**o+(0.841260061022)*x[1]
        arg[3,4,2,1]=(0.520110030788)*x[0]**o+(0.519262786324)*x[0]+(0.861221788948)*x[1]**o+(-0.376714934937)*x[1]
        ref[0,0,0,0]=(0.240887319278)*x_ref[0]**o+(-0.192166587545)*x_ref[0]+(0.0176974287409)*x_ref[1]**o+(0.100761748824)*x_ref[1]
        ref[0,0,0,1]=(-0.549649845775)*x_ref[0]**o+(-0.293969245646)*x_ref[0]+(-0.605885325133)*x_ref[1]**o+(0.799558451173)*x_ref[1]
        ref[0,0,1,0]=(-0.899658769582)*x_ref[0]**o+(0.484524193987)*x_ref[0]+(0.437485870351)*x_ref[1]**o+(-0.220549034927)*x_ref[1]
        ref[0,0,1,1]=(-0.146539961793)*x_ref[0]**o+(-0.278343486929)*x_ref[0]+(0.0913912456082)*x_ref[1]**o+(-0.751028975123)*x_ref[1]
        ref[0,0,2,0]=(0.806344701194)*x_ref[0]**o+(0.0537872824561)*x_ref[0]+(-0.70407828379)*x_ref[1]**o+(-0.922963486601)*x_ref[1]
        ref[0,0,2,1]=(0.307202902495)*x_ref[0]**o+(0.301603506731)*x_ref[0]+(0.232983743616)*x_ref[1]**o+(-0.744267542759)*x_ref[1]
        ref[0,1,0,0]=(-0.550616060824)*x_ref[0]**o+(-0.661367750254)*x_ref[0]+(-0.695905895864)*x_ref[1]**o+(-0.788045922398)*x_ref[1]
        ref[0,1,0,1]=(-0.934169134035)*x_ref[0]**o+(-0.843861105521)*x_ref[0]+(-0.237877254439)*x_ref[1]**o+(-0.364956821348)*x_ref[1]
        ref[0,1,1,0]=(0.0950675750026)*x_ref[0]**o+(-0.815510640594)*x_ref[0]+(0.72526035968)*x_ref[1]**o+(0.898675190106)*x_ref[1]
        ref[0,1,1,1]=(0.715609066776)*x_ref[0]**o+(0.371314845206)*x_ref[0]+(0.326878867244)*x_ref[1]**o+(-0.468798950199)*x_ref[1]
        ref[0,1,2,0]=(0.392175939547)*x_ref[0]**o+(0.105332682674)*x_ref[0]+(0.997967455477)*x_ref[1]**o+(0.0322491822496)*x_ref[1]
        ref[0,1,2,1]=(-0.804498399187)*x_ref[0]**o+(0.819653744739)*x_ref[0]+(0.609610573703)*x_ref[1]**o+(-0.623299047794)*x_ref[1]
        ref[0,2,0,0]=(0.993877917386)*x_ref[0]**o+(0.815952897899)*x_ref[0]+(-0.258432174486)*x_ref[1]**o+(0.209324387308)*x_ref[1]
        ref[0,2,0,1]=(0.395946889364)*x_ref[0]**o+(-0.859500539908)*x_ref[0]+(-0.68495734785)*x_ref[1]**o+(0.791813946302)*x_ref[1]
        ref[0,2,1,0]=(0.154347661108)*x_ref[0]**o+(0.0873140744322)*x_ref[0]+(-0.423935109956)*x_ref[1]**o+(-0.754297239226)*x_ref[1]
        ref[0,2,1,1]=(-0.623998419484)*x_ref[0]**o+(0.487300130941)*x_ref[0]+(-0.323613219784)*x_ref[1]**o+(0.744877983729)*x_ref[1]
        ref[0,2,2,0]=(0.515686374371)*x_ref[0]**o+(0.512572153346)*x_ref[0]+(-0.641245783272)*x_ref[1]**o+(-0.390660158759)*x_ref[1]
        ref[0,2,2,1]=(0.24398238139)*x_ref[0]**o+(-0.172020271678)*x_ref[0]+(0.980197812701)*x_ref[1]**o+(0.336941934503)*x_ref[1]
        ref[0,3,0,0]=(0.0154584447615)*x_ref[0]**o+(-0.547048191238)*x_ref[0]+(-0.555319223005)*x_ref[1]**o+(0.0972794598106)*x_ref[1]
        ref[0,3,0,1]=(0.44744389335)*x_ref[0]**o+(0.475046337874)*x_ref[0]+(0.876228982066)*x_ref[1]**o+(-0.0636405241175)*x_ref[1]
        ref[0,3,1,0]=(0.392948465298)*x_ref[0]**o+(-0.341342654182)*x_ref[0]+(0.747185683968)*x_ref[1]**o+(-0.99981172166)*x_ref[1]
        ref[0,3,1,1]=(-0.172355321621)*x_ref[0]**o+(-0.531391484795)*x_ref[0]+(0.451654274019)*x_ref[1]**o+(0.0795298667915)*x_ref[1]
        ref[0,3,2,0]=(0.0841642474674)*x_ref[0]**o+(0.308494435164)*x_ref[0]+(0.53181918181)*x_ref[1]**o+(0.480258067959)*x_ref[1]
        ref[0,3,2,1]=(-0.222903054864)*x_ref[0]**o+(-0.0330655869754)*x_ref[0]+(0.227482525015)*x_ref[1]**o+(0.536136212347)*x_ref[1]
        ref[0,4,0,0]=(0.574798381284)*x_ref[0]**o+(0.181354466484)*x_ref[0]+(0.103049514171)*x_ref[1]**o+(0.156282818835)*x_ref[1]
        ref[0,4,0,1]=(0.478314255289)*x_ref[0]**o+(0.224797222592)*x_ref[0]+(0.307668606272)*x_ref[1]**o+(-0.785262182937)*x_ref[1]
        ref[0,4,1,0]=(0.519356961182)*x_ref[0]**o+(0.646200458344)*x_ref[0]+(0.0522618783181)*x_ref[1]**o+(0.702335307324)*x_ref[1]
        ref[0,4,1,1]=(-0.168756029541)*x_ref[0]**o+(-0.00618364750666)*x_ref[0]+(0.472704616213)*x_ref[1]**o+(0.974257451871)*x_ref[1]
        ref[0,4,2,0]=(-0.279196869573)*x_ref[0]**o+(-0.333083174091)*x_ref[0]+(-0.980089574428)*x_ref[1]**o+(0.405273709405)*x_ref[1]
        ref[0,4,2,1]=(-0.199431380673)*x_ref[0]**o+(0.823294899866)*x_ref[0]+(-0.174125117685)*x_ref[1]**o+(-0.730500493887)*x_ref[1]
        ref[1,0,0,0]=(-0.577281489635)*x_ref[0]**o+(0.231906139898)*x_ref[0]+(0.537128708963)*x_ref[1]**o+(0.363444540898)*x_ref[1]
        ref[1,0,0,1]=(0.204108541882)*x_ref[0]**o+(-0.217661714888)*x_ref[0]+(0.963384293759)*x_ref[1]**o+(-0.0817519801351)*x_ref[1]
        ref[1,0,1,0]=(0.540681325439)*x_ref[0]**o+(-0.496036960808)*x_ref[0]+(0.421037689661)*x_ref[1]**o+(0.19399086996)*x_ref[1]
        ref[1,0,1,1]=(0.482938542126)*x_ref[0]**o+(-0.0725027675209)*x_ref[0]+(-0.441517962054)*x_ref[1]**o+(0.787445860714)*x_ref[1]
        ref[1,0,2,0]=(0.0378928111257)*x_ref[0]**o+(0.451368848019)*x_ref[0]+(-0.228668284485)*x_ref[1]**o+(0.505927229432)*x_ref[1]
        ref[1,0,2,1]=(-0.59833518872)*x_ref[0]**o+(0.956068253354)*x_ref[0]+(0.667905995469)*x_ref[1]**o+(0.963273292214)*x_ref[1]
        ref[1,1,0,0]=(0.901637865236)*x_ref[0]**o+(-0.642678971748)*x_ref[0]+(-0.603915019524)*x_ref[1]**o+(-0.208004993919)*x_ref[1]
        ref[1,1,0,1]=(-0.227162661052)*x_ref[0]**o+(-0.5816434303)*x_ref[0]+(0.274961293801)*x_ref[1]**o+(-0.614001328621)*x_ref[1]
        ref[1,1,1,0]=(0.078438078548)*x_ref[0]**o+(0.570191242883)*x_ref[0]+(0.0952878212484)*x_ref[1]**o+(0.844460205497)*x_ref[1]
        ref[1,1,1,1]=(-0.901445811786)*x_ref[0]**o+(-0.996376775344)*x_ref[0]+(0.707470471613)*x_ref[1]**o+(-0.187517936674)*x_ref[1]
        ref[1,1,2,0]=(-0.66891443067)*x_ref[0]**o+(0.0462057239324)*x_ref[0]+(-0.825820539128)*x_ref[1]**o+(-0.766932836843)*x_ref[1]
        ref[1,1,2,1]=(-0.851428828811)*x_ref[0]**o+(0.504484422346)*x_ref[0]+(0.121378315545)*x_ref[1]**o+(0.903947012757)*x_ref[1]
        ref[1,2,0,0]=(-0.393827449768)*x_ref[0]**o+(-0.251827904608)*x_ref[0]+(0.8320040882)*x_ref[1]**o+(0.57595149855)*x_ref[1]
        ref[1,2,0,1]=(0.638738110712)*x_ref[0]**o+(-0.439962152539)*x_ref[0]+(-0.977211078995)*x_ref[1]**o+(0.0528475212451)*x_ref[1]
        ref[1,2,1,0]=(0.824313588941)*x_ref[0]**o+(-0.589968996823)*x_ref[0]+(-0.512156794152)*x_ref[1]**o+(0.142809794443)*x_ref[1]
        ref[1,2,1,1]=(-0.31789333147)*x_ref[0]**o+(0.598839896949)*x_ref[0]+(0.638319398651)*x_ref[1]**o+(0.353625999317)*x_ref[1]
        ref[1,2,2,0]=(-0.720472365012)*x_ref[0]**o+(0.896538030051)*x_ref[0]+(-0.224639169972)*x_ref[1]**o+(-0.898299360738)*x_ref[1]
        ref[1,2,2,1]=(0.123077707167)*x_ref[0]**o+(0.15922851038)*x_ref[0]+(0.678984783843)*x_ref[1]**o+(-0.778188269852)*x_ref[1]
        ref[1,3,0,0]=(0.344256301209)*x_ref[0]**o+(0.884367738289)*x_ref[0]+(0.00556636219289)*x_ref[1]**o+(0.684361331195)*x_ref[1]
        ref[1,3,0,1]=(-0.812245131065)*x_ref[0]**o+(0.971365671228)*x_ref[0]+(0.0696053239569)*x_ref[1]**o+(-0.300266235138)*x_ref[1]
        ref[1,3,1,0]=(0.470254588058)*x_ref[0]**o+(0.0946825509913)*x_ref[0]+(-0.415619255025)*x_ref[1]**o+(-0.27381652858)*x_ref[1]
        ref[1,3,1,1]=(-0.84555138241)*x_ref[0]**o+(0.873510832091)*x_ref[0]+(-0.138010338152)*x_ref[1]**o+(-0.982274515316)*x_ref[1]
        ref[1,3,2,0]=(0.695614136148)*x_ref[0]**o+(-0.192588082771)*x_ref[0]+(-0.21579127791)*x_ref[1]**o+(-0.0335939736388)*x_ref[1]
        ref[1,3,2,1]=(-0.633363463896)*x_ref[0]**o+(-0.958320549993)*x_ref[0]+(0.197331067481)*x_ref[1]**o+(0.315089888409)*x_ref[1]
        ref[1,4,0,0]=(-0.211531219945)*x_ref[0]**o+(-0.399700721799)*x_ref[0]+(0.578927638839)*x_ref[1]**o+(-0.389034976195)*x_ref[1]
        ref[1,4,0,1]=(-0.846399571893)*x_ref[0]**o+(-0.952904755903)*x_ref[0]+(-0.858399802693)*x_ref[1]**o+(-0.23264354907)*x_ref[1]
        ref[1,4,1,0]=(0.255822441586)*x_ref[0]**o+(0.896237710393)*x_ref[0]+(-0.90248008271)*x_ref[1]**o+(0.654804848244)*x_ref[1]
        ref[1,4,1,1]=(-0.112475508163)*x_ref[0]**o+(-0.861697204453)*x_ref[0]+(-0.782581931187)*x_ref[1]**o+(-0.129345341581)*x_ref[1]
        ref[1,4,2,0]=(0.690095687033)*x_ref[0]**o+(0.729317334019)*x_ref[0]+(-0.00961483612288)*x_ref[1]**o+(-0.288320551243)*x_ref[1]
        ref[1,4,2,1]=(-0.648146433517)*x_ref[0]**o+(-0.301246601263)*x_ref[0]+(-0.490119612994)*x_ref[1]**o+(0.231186845615)*x_ref[1]
        ref[2,0,0,0]=(-0.645560891595)*x_ref[0]**o+(-0.889067544988)*x_ref[0]+(-0.193353414221)*x_ref[1]**o+(-0.726988479344)*x_ref[1]
        ref[2,0,0,1]=(-0.083614738484)*x_ref[0]**o+(-0.822530927241)*x_ref[0]+(0.660338934241)*x_ref[1]**o+(0.0383821036508)*x_ref[1]
        ref[2,0,1,0]=(-0.132250736664)*x_ref[0]**o+(-0.571474370931)*x_ref[0]+(-0.755931100342)*x_ref[1]**o+(-0.145196865115)*x_ref[1]
        ref[2,0,1,1]=(0.545219196727)*x_ref[0]**o+(-0.846368787913)*x_ref[0]+(-0.845706160842)*x_ref[1]**o+(0.550703610989)*x_ref[1]
        ref[2,0,2,0]=(-0.766303732195)*x_ref[0]**o+(0.156822893641)*x_ref[0]+(0.00954334254282)*x_ref[1]**o+(0.201506144561)*x_ref[1]
        ref[2,0,2,1]=(-0.274476449632)*x_ref[0]**o+(-0.361690468185)*x_ref[0]+(-0.816535747268)*x_ref[1]**o+(-0.473913811701)*x_ref[1]
        ref[2,1,0,0]=(-0.927592771482)*x_ref[0]**o+(0.534615269814)*x_ref[0]+(-0.571217133452)*x_ref[1]**o+(0.57448299216)*x_ref[1]
        ref[2,1,0,1]=(-0.298153953211)*x_ref[0]**o+(-0.143636909462)*x_ref[0]+(0.990660288864)*x_ref[1]**o+(-0.495864284763)*x_ref[1]
        ref[2,1,1,0]=(-0.110625727208)*x_ref[0]**o+(-0.221057774317)*x_ref[0]+(0.557195078358)*x_ref[1]**o+(0.803162691808)*x_ref[1]
        ref[2,1,1,1]=(-0.115589176775)*x_ref[0]**o+(0.08253351782)*x_ref[0]+(-0.595996958977)*x_ref[1]**o+(0.501435717501)*x_ref[1]
        ref[2,1,2,0]=(0.78764976658)*x_ref[0]**o+(0.966729831711)*x_ref[0]+(-0.585879575667)*x_ref[1]**o+(0.341148176793)*x_ref[1]
        ref[2,1,2,1]=(-0.551697945691)*x_ref[0]**o+(0.545857249323)*x_ref[0]+(-0.919321724495)*x_ref[1]**o+(-0.286043327415)*x_ref[1]
        ref[2,2,0,0]=(-0.669933890655)*x_ref[0]**o+(-0.344862118495)*x_ref[0]+(0.878331282597)*x_ref[1]**o+(0.881893760718)*x_ref[1]
        ref[2,2,0,1]=(-0.217876618414)*x_ref[0]**o+(-0.289118007199)*x_ref[0]+(0.513464171525)*x_ref[1]**o+(0.310055899102)*x_ref[1]
        ref[2,2,1,0]=(0.0472051379154)*x_ref[0]**o+(0.42455388451)*x_ref[0]+(-0.920571834891)*x_ref[1]**o+(-0.779095076537)*x_ref[1]
        ref[2,2,1,1]=(-0.299619304537)*x_ref[0]**o+(-0.289777287294)*x_ref[0]+(0.920505062988)*x_ref[1]**o+(-0.243924179095)*x_ref[1]
        ref[2,2,2,0]=(-0.819343530877)*x_ref[0]**o+(0.975953690875)*x_ref[0]+(-0.787239459241)*x_ref[1]**o+(-0.199900734691)*x_ref[1]
        ref[2,2,2,1]=(0.0980085780878)*x_ref[0]**o+(0.642946515451)*x_ref[0]+(0.951790535152)*x_ref[1]**o+(-0.253914817366)*x_ref[1]
        ref[2,3,0,0]=(-0.882075522824)*x_ref[0]**o+(0.200934502682)*x_ref[0]+(-0.83200464496)*x_ref[1]**o+(0.941479345375)*x_ref[1]
        ref[2,3,0,1]=(-0.615941782209)*x_ref[0]**o+(-0.693242388092)*x_ref[0]+(-0.640417278871)*x_ref[1]**o+(-0.346459094803)*x_ref[1]
        ref[2,3,1,0]=(-0.447919427952)*x_ref[0]**o+(0.205055689141)*x_ref[0]+(-0.453290032411)*x_ref[1]**o+(0.879062836521)*x_ref[1]
        ref[2,3,1,1]=(0.721133924016)*x_ref[0]**o+(0.71294558459)*x_ref[0]+(-0.250087045805)*x_ref[1]**o+(0.605146721129)*x_ref[1]
        ref[2,3,2,0]=(0.247782322921)*x_ref[0]**o+(-0.956367425119)*x_ref[0]+(-0.054045082005)*x_ref[1]**o+(-0.128151832132)*x_ref[1]
        ref[2,3,2,1]=(-0.427039259088)*x_ref[0]**o+(-0.359849242839)*x_ref[0]+(-0.244990934481)*x_ref[1]**o+(-0.038020035629)*x_ref[1]
        ref[2,4,0,0]=(-0.216409150834)*x_ref[0]**o+(0.492547989921)*x_ref[0]+(0.878317438469)*x_ref[1]**o+(-0.112718719119)*x_ref[1]
        ref[2,4,0,1]=(0.487403443102)*x_ref[0]**o+(0.228155519468)*x_ref[0]+(0.923150527516)*x_ref[1]**o+(0.642654867908)*x_ref[1]
        ref[2,4,1,0]=(0.916418483417)*x_ref[0]**o+(-0.177198935173)*x_ref[0]+(-0.790548940412)*x_ref[1]**o+(0.87540167052)*x_ref[1]
        ref[2,4,1,1]=(0.718502075538)*x_ref[0]**o+(0.691254010569)*x_ref[0]+(-0.0134430122729)*x_ref[1]**o+(-0.739888770749)*x_ref[1]
        ref[2,4,2,0]=(0.287825086861)*x_ref[0]**o+(0.596621809347)*x_ref[0]+(0.729784205391)*x_ref[1]**o+(-0.25984257955)*x_ref[1]
        ref[2,4,2,1]=(-0.515102480025)*x_ref[0]**o+(-0.609951840015)*x_ref[0]+(0.229834340075)*x_ref[1]**o+(-0.727157816241)*x_ref[1]
        ref[3,0,0,0]=(-0.349406091363)*x_ref[0]**o+(0.802348451103)*x_ref[0]+(-0.174421346276)*x_ref[1]**o+(0.120711198036)*x_ref[1]
        ref[3,0,0,1]=(0.367085169423)*x_ref[0]**o+(-0.319683236434)*x_ref[0]+(-0.590820542566)*x_ref[1]**o+(0.607625953303)*x_ref[1]
        ref[3,0,1,0]=(0.231259976433)*x_ref[0]**o+(-0.493297408825)*x_ref[0]+(0.919363640504)*x_ref[1]**o+(-0.839841270429)*x_ref[1]
        ref[3,0,1,1]=(0.57672135684)*x_ref[0]**o+(0.882062879704)*x_ref[0]+(-0.600979706444)*x_ref[1]**o+(-0.704008344509)*x_ref[1]
        ref[3,0,2,0]=(-0.208162114527)*x_ref[0]**o+(0.593682243052)*x_ref[0]+(-0.428436832487)*x_ref[1]**o+(0.51007021604)*x_ref[1]
        ref[3,0,2,1]=(0.315264672653)*x_ref[0]**o+(-0.0835897376534)*x_ref[0]+(-0.135315173106)*x_ref[1]**o+(-0.646372309498)*x_ref[1]
        ref[3,1,0,0]=(0.815505953752)*x_ref[0]**o+(0.0625493643292)*x_ref[0]+(0.383118433815)*x_ref[1]**o+(0.17372720435)*x_ref[1]
        ref[3,1,0,1]=(0.557295786969)*x_ref[0]**o+(-0.568823508165)*x_ref[0]+(-0.00874733267675)*x_ref[1]**o+(-0.27454140598)*x_ref[1]
        ref[3,1,1,0]=(0.649021026591)*x_ref[0]**o+(0.232255030448)*x_ref[0]+(0.845939296425)*x_ref[1]**o+(0.116387378294)*x_ref[1]
        ref[3,1,1,1]=(0.431124940343)*x_ref[0]**o+(-0.0865937271969)*x_ref[0]+(0.937911741288)*x_ref[1]**o+(0.362850039234)*x_ref[1]
        ref[3,1,2,0]=(0.155965414439)*x_ref[0]**o+(0.110850421841)*x_ref[0]+(0.244563080191)*x_ref[1]**o+(0.266325498006)*x_ref[1]
        ref[3,1,2,1]=(0.766813300935)*x_ref[0]**o+(0.77492422889)*x_ref[0]+(-0.0141436554714)*x_ref[1]**o+(0.780444449009)*x_ref[1]
        ref[3,2,0,0]=(0.619086019019)*x_ref[0]**o+(-0.574130567601)*x_ref[0]+(-0.417708159593)*x_ref[1]**o+(0.707727284254)*x_ref[1]
        ref[3,2,0,1]=(-0.845657409353)*x_ref[0]**o+(0.705282507748)*x_ref[0]+(0.124557418112)*x_ref[1]**o+(-0.337493494733)*x_ref[1]
        ref[3,2,1,0]=(0.114875146211)*x_ref[0]**o+(-0.724480426892)*x_ref[0]+(-0.0763476904405)*x_ref[1]**o+(0.12064907114)*x_ref[1]
        ref[3,2,1,1]=(-0.74994413376)*x_ref[0]**o+(-0.916476035217)*x_ref[0]+(-0.360412756599)*x_ref[1]**o+(-0.917457363098)*x_ref[1]
        ref[3,2,2,0]=(-0.446839470984)*x_ref[0]**o+(-0.097704838316)*x_ref[0]+(-0.653584364909)*x_ref[1]**o+(0.561014757758)*x_ref[1]
        ref[3,2,2,1]=(-0.178975301038)*x_ref[0]**o+(-0.843669982667)*x_ref[0]+(-0.566610094242)*x_ref[1]**o+(0.213047649003)*x_ref[1]
        ref[3,3,0,0]=(0.144115764355)*x_ref[0]**o+(0.199435411954)*x_ref[0]+(0.173334385199)*x_ref[1]**o+(-0.942957035808)*x_ref[1]
        ref[3,3,0,1]=(-0.417589926342)*x_ref[0]**o+(-0.819188022852)*x_ref[0]+(-0.68299160003)*x_ref[1]**o+(-0.314851447894)*x_ref[1]
        ref[3,3,1,0]=(0.737898219164)*x_ref[0]**o+(0.612906277393)*x_ref[0]+(-0.768349396729)*x_ref[1]**o+(-0.192998755095)*x_ref[1]
        ref[3,3,1,1]=(0.35433362007)*x_ref[0]**o+(0.894174225208)*x_ref[0]+(-0.450333030837)*x_ref[1]**o+(0.819450589483)*x_ref[1]
        ref[3,3,2,0]=(-0.455377131302)*x_ref[0]**o+(-0.748183144687)*x_ref[0]+(-0.482739450231)*x_ref[1]**o+(0.300625929022)*x_ref[1]
        ref[3,3,2,1]=(-0.602766693698)*x_ref[0]**o+(-0.416288436985)*x_ref[0]+(0.837501983536)*x_ref[1]**o+(-0.745935308188)*x_ref[1]
        ref[3,4,0,0]=(-0.706774010405)*x_ref[0]**o+(0.407766818701)*x_ref[0]+(-0.248964401443)*x_ref[1]**o+(0.675466487229)*x_ref[1]
        ref[3,4,0,1]=(0.512302883876)*x_ref[0]**o+(0.0809544867105)*x_ref[0]+(-0.204555203536)*x_ref[1]**o+(-0.628574073749)*x_ref[1]
        ref[3,4,1,0]=(0.126376796421)*x_ref[0]**o+(0.174607814129)*x_ref[0]+(0.68867194136)*x_ref[1]**o+(0.940025934568)*x_ref[1]
        ref[3,4,1,1]=(0.583042444739)*x_ref[0]**o+(-0.786981763836)*x_ref[0]+(0.319397032405)*x_ref[1]**o+(-0.324993351432)*x_ref[1]
        ref[3,4,2,0]=(-0.941997068836)*x_ref[0]**o+(-0.892825520378)*x_ref[0]+(0.275995419027)*x_ref[1]**o+(0.841260061022)*x_ref[1]
        ref[3,4,2,1]=(0.520110030788)*x_ref[0]**o+(0.519262786324)*x_ref[0]+(0.861221788948)*x_ref[1]**o+(-0.376714934937)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.615899511086)*x[0]**o+(0.823637401824)*x[0]+(-0.986967740198)*x[1]**o+(-0.782088977641)*x[1]+(-0.576799974679)*x[2]**o+(-0.910942885711)*x[2]
        arg[0,0,0,1]=(-0.790830593449)*x[0]**o+(-0.897724279499)*x[0]+(-0.881691805807)*x[1]**o+(-0.133793238901)*x[1]+(-0.0903391684489)*x[2]**o+(0.659893469877)*x[2]
        arg[0,0,1,0]=(-0.0567117900114)*x[0]**o+(-0.858031985686)*x[0]+(-0.397443445702)*x[1]**o+(0.549982838823)*x[1]+(-0.636905143813)*x[2]**o+(0.8488142632)*x[2]
        arg[0,0,1,1]=(-0.644557373082)*x[0]**o+(-0.323654898522)*x[0]+(0.210458705923)*x[1]**o+(0.990436827296)*x[1]+(-0.597872757151)*x[2]**o+(-0.83865548618)*x[2]
        arg[0,0,2,0]=(-0.29830667084)*x[0]**o+(0.327035546909)*x[0]+(0.132690049791)*x[1]**o+(-0.55625443282)*x[1]+(0.0972781385283)*x[2]**o+(0.151885837608)*x[2]
        arg[0,0,2,1]=(-0.435983465578)*x[0]**o+(-0.0852522953446)*x[0]+(0.257654807094)*x[1]**o+(-0.336603197583)*x[1]+(0.637913421592)*x[2]**o+(0.8151042055)*x[2]
        arg[0,1,0,0]=(-0.0373980191125)*x[0]**o+(-0.45937285505)*x[0]+(-0.857933684889)*x[1]**o+(-0.843826686256)*x[1]+(-0.727077320504)*x[2]**o+(-0.65222863395)*x[2]
        arg[0,1,0,1]=(0.63426017384)*x[0]**o+(0.845708641601)*x[0]+(0.246160302486)*x[1]**o+(-0.660855572094)*x[1]+(0.67420981328)*x[2]**o+(-0.280803968781)*x[2]
        arg[0,1,1,0]=(0.881193264597)*x[0]**o+(-0.881936969515)*x[0]+(-0.883246837152)*x[1]**o+(0.182222471927)*x[1]+(0.405380496272)*x[2]**o+(0.175721746883)*x[2]
        arg[0,1,1,1]=(-0.44447667237)*x[0]**o+(0.956169274606)*x[0]+(0.0370532940652)*x[1]**o+(-0.181239635583)*x[1]+(-0.678412161285)*x[2]**o+(-0.653172157401)*x[2]
        arg[0,1,2,0]=(-0.50038634718)*x[0]**o+(0.00278321765167)*x[0]+(-0.831539460218)*x[1]**o+(-0.555243320385)*x[1]+(-0.626789452003)*x[2]**o+(-0.495085243933)*x[2]
        arg[0,1,2,1]=(-0.32886359569)*x[0]**o+(0.406099965475)*x[0]+(0.764756177266)*x[1]**o+(-0.272070344181)*x[1]+(-0.41812596516)*x[2]**o+(0.349063696376)*x[2]
        arg[0,2,0,0]=(-0.580544978365)*x[0]**o+(0.53441241334)*x[0]+(-0.808749795876)*x[1]**o+(0.172110801632)*x[1]+(-0.939262606098)*x[2]**o+(-0.724859325104)*x[2]
        arg[0,2,0,1]=(0.256335221944)*x[0]**o+(0.463522371018)*x[0]+(0.990516478508)*x[1]**o+(0.423931177054)*x[1]+(0.184277693683)*x[2]**o+(-0.272967447689)*x[2]
        arg[0,2,1,0]=(-0.126943811008)*x[0]**o+(0.645821387926)*x[0]+(0.150370174987)*x[1]**o+(-0.150954538301)*x[1]+(-0.378720078559)*x[2]**o+(0.0975963807109)*x[2]
        arg[0,2,1,1]=(-0.451434544339)*x[0]**o+(0.557314109088)*x[0]+(-0.0402970749558)*x[1]**o+(0.306984557206)*x[1]+(-0.297018139186)*x[2]**o+(0.593800698785)*x[2]
        arg[0,2,2,0]=(-0.376199144911)*x[0]**o+(-0.230454116373)*x[0]+(0.992738297651)*x[1]**o+(0.183510534927)*x[1]+(0.409381711992)*x[2]**o+(0.171945265727)*x[2]
        arg[0,2,2,1]=(0.363233095971)*x[0]**o+(-0.130589148249)*x[0]+(0.809611042018)*x[1]**o+(-0.354546307961)*x[1]+(0.0631410169203)*x[2]**o+(0.630306161781)*x[2]
        arg[0,3,0,0]=(0.49655016161)*x[0]**o+(0.878957258614)*x[0]+(-0.397408696784)*x[1]**o+(-0.199445321288)*x[1]+(0.929857086663)*x[2]**o+(-0.648770887161)*x[2]
        arg[0,3,0,1]=(-0.85578334594)*x[0]**o+(-0.294617109229)*x[0]+(-0.0790295769768)*x[1]**o+(0.82151942311)*x[1]+(0.541628117444)*x[2]**o+(0.805418174827)*x[2]
        arg[0,3,1,0]=(0.800775454804)*x[0]**o+(0.599550462983)*x[0]+(-0.384788634082)*x[1]**o+(-0.356933972363)*x[1]+(-0.423116265473)*x[2]**o+(-0.125492982003)*x[2]
        arg[0,3,1,1]=(0.929652994661)*x[0]**o+(-0.57396682406)*x[0]+(0.313584116036)*x[1]**o+(-0.190575043496)*x[1]+(-0.29110884935)*x[2]**o+(0.967530639925)*x[2]
        arg[0,3,2,0]=(0.761136478868)*x[0]**o+(0.565721170014)*x[0]+(0.239619122131)*x[1]**o+(-0.0821651979956)*x[1]+(-0.699218775545)*x[2]**o+(0.19973415585)*x[2]
        arg[0,3,2,1]=(0.209543244648)*x[0]**o+(0.93957796874)*x[0]+(0.354405887426)*x[1]**o+(0.504177262864)*x[1]+(0.446866674894)*x[2]**o+(-0.522936485464)*x[2]
        arg[0,4,0,0]=(0.427541399652)*x[0]**o+(-0.114806067206)*x[0]+(0.236780877643)*x[1]**o+(0.422226381272)*x[1]+(-0.506450961214)*x[2]**o+(0.616091720212)*x[2]
        arg[0,4,0,1]=(0.698773571323)*x[0]**o+(-0.357878713085)*x[0]+(-0.453730359129)*x[1]**o+(-0.328701041355)*x[1]+(-0.28646346679)*x[2]**o+(-0.0449784265203)*x[2]
        arg[0,4,1,0]=(-0.419416306509)*x[0]**o+(-0.1970394955)*x[0]+(-0.814008932379)*x[1]**o+(-0.805127589929)*x[1]+(0.244059789875)*x[2]**o+(-0.983027843847)*x[2]
        arg[0,4,1,1]=(0.84837346004)*x[0]**o+(0.501808510624)*x[0]+(-0.105800351358)*x[1]**o+(0.903999231718)*x[1]+(-0.0473996978124)*x[2]**o+(0.683647253381)*x[2]
        arg[0,4,2,0]=(0.658722899682)*x[0]**o+(-0.401846635016)*x[0]+(-0.967322202581)*x[1]**o+(-0.948306922217)*x[1]+(0.496409308219)*x[2]**o+(-0.184768777052)*x[2]
        arg[0,4,2,1]=(-0.454468390888)*x[0]**o+(-0.220640468215)*x[0]+(0.995269925943)*x[1]**o+(0.778951329211)*x[1]+(0.0439534466302)*x[2]**o+(0.806197038383)*x[2]
        arg[1,0,0,0]=(-0.547461802236)*x[0]**o+(0.370768827789)*x[0]+(-0.619991757552)*x[1]**o+(0.391727195645)*x[1]+(-0.649274922953)*x[2]**o+(-0.478788248429)*x[2]
        arg[1,0,0,1]=(-0.899230505882)*x[0]**o+(0.603064321735)*x[0]+(-0.737475386349)*x[1]**o+(0.594197953001)*x[1]+(0.341324049064)*x[2]**o+(0.285202633917)*x[2]
        arg[1,0,1,0]=(-0.107063589618)*x[0]**o+(-0.260491562358)*x[0]+(-0.667059815618)*x[1]**o+(0.473227354537)*x[1]+(-0.587170510558)*x[2]**o+(-0.740299218393)*x[2]
        arg[1,0,1,1]=(0.87705715098)*x[0]**o+(0.312753471827)*x[0]+(-0.307530283315)*x[1]**o+(0.195564644822)*x[1]+(-0.64384485166)*x[2]**o+(-0.776362951729)*x[2]
        arg[1,0,2,0]=(0.0949408564739)*x[0]**o+(0.765922179855)*x[0]+(0.441215568253)*x[1]**o+(0.339112969437)*x[1]+(-0.0774615292449)*x[2]**o+(0.673258055783)*x[2]
        arg[1,0,2,1]=(0.114828212302)*x[0]**o+(0.0831258766277)*x[0]+(0.674540402248)*x[1]**o+(-0.304921039094)*x[1]+(-0.725514770578)*x[2]**o+(-0.627584773746)*x[2]
        arg[1,1,0,0]=(0.291561714675)*x[0]**o+(0.898587624783)*x[0]+(0.476696264545)*x[1]**o+(-0.454648604978)*x[1]+(0.677949150743)*x[2]**o+(-0.59438289851)*x[2]
        arg[1,1,0,1]=(0.356687258009)*x[0]**o+(0.806236251629)*x[0]+(0.843636690354)*x[1]**o+(-0.649776947181)*x[1]+(0.208542122388)*x[2]**o+(-0.0726045573458)*x[2]
        arg[1,1,1,0]=(0.712800450558)*x[0]**o+(0.748271875755)*x[0]+(0.403880649356)*x[1]**o+(0.956549336754)*x[1]+(0.0420128495043)*x[2]**o+(-0.83177870499)*x[2]
        arg[1,1,1,1]=(-0.886862904174)*x[0]**o+(-0.03822986594)*x[0]+(0.587151517496)*x[1]**o+(-0.120272060898)*x[1]+(-0.573366391933)*x[2]**o+(-0.59772763241)*x[2]
        arg[1,1,2,0]=(0.311960197804)*x[0]**o+(0.486854949444)*x[0]+(-0.647995651954)*x[1]**o+(0.30104277964)*x[1]+(-0.174718703748)*x[2]**o+(-0.506993216295)*x[2]
        arg[1,1,2,1]=(0.979056206991)*x[0]**o+(0.892104464812)*x[0]+(0.0507432703245)*x[1]**o+(0.103630858454)*x[1]+(-0.0254965952266)*x[2]**o+(0.13034027239)*x[2]
        arg[1,2,0,0]=(0.259797966548)*x[0]**o+(0.53018999371)*x[0]+(0.533745722992)*x[1]**o+(0.53029084173)*x[1]+(-0.595126791061)*x[2]**o+(-0.51668295558)*x[2]
        arg[1,2,0,1]=(0.926395590808)*x[0]**o+(-0.0700141207179)*x[0]+(-0.973392454012)*x[1]**o+(-0.46204067097)*x[1]+(0.814964497207)*x[2]**o+(-0.183546091376)*x[2]
        arg[1,2,1,0]=(-0.0872635441658)*x[0]**o+(-0.0307630839035)*x[0]+(0.539798653095)*x[1]**o+(0.966367734848)*x[1]+(-0.817495053612)*x[2]**o+(-0.268397867848)*x[2]
        arg[1,2,1,1]=(-0.933074613921)*x[0]**o+(0.48896533938)*x[0]+(0.453696133488)*x[1]**o+(-0.238730406749)*x[1]+(-0.696785951086)*x[2]**o+(-0.359858385642)*x[2]
        arg[1,2,2,0]=(-0.844406286809)*x[0]**o+(-0.506332243946)*x[0]+(0.976436731658)*x[1]**o+(-0.201489161857)*x[1]+(0.851236173166)*x[2]**o+(-0.998803612947)*x[2]
        arg[1,2,2,1]=(0.259707321449)*x[0]**o+(-0.888149478655)*x[0]+(-0.273502842869)*x[1]**o+(0.291193582158)*x[1]+(-0.897542178027)*x[2]**o+(-0.345994824429)*x[2]
        arg[1,3,0,0]=(0.961298984697)*x[0]**o+(0.248141115092)*x[0]+(-0.623053025178)*x[1]**o+(-0.775095452078)*x[1]+(0.955362707049)*x[2]**o+(-0.517547333705)*x[2]
        arg[1,3,0,1]=(0.858098596227)*x[0]**o+(0.309307483873)*x[0]+(-0.209217034755)*x[1]**o+(0.783803593379)*x[1]+(0.138164128092)*x[2]**o+(-0.922854003461)*x[2]
        arg[1,3,1,0]=(-0.139017942141)*x[0]**o+(-0.338223039119)*x[0]+(0.737882941498)*x[1]**o+(0.948854903087)*x[1]+(-0.0792461633266)*x[2]**o+(0.848875660203)*x[2]
        arg[1,3,1,1]=(0.693284332212)*x[0]**o+(-0.392272875348)*x[0]+(-0.475709045479)*x[1]**o+(0.182948077236)*x[1]+(0.0277569761538)*x[2]**o+(-0.550294947596)*x[2]
        arg[1,3,2,0]=(-0.105381649505)*x[0]**o+(-0.403051690238)*x[0]+(-0.722219735818)*x[1]**o+(-0.380114388869)*x[1]+(0.321452572491)*x[2]**o+(0.858583245253)*x[2]
        arg[1,3,2,1]=(-0.405370062024)*x[0]**o+(0.0743704840047)*x[0]+(0.998947017815)*x[1]**o+(0.983611079522)*x[1]+(0.854202982674)*x[2]**o+(0.288478746498)*x[2]
        arg[1,4,0,0]=(0.0584782489998)*x[0]**o+(0.999286927604)*x[0]+(-0.458166020555)*x[1]**o+(0.529008844128)*x[1]+(0.424592641922)*x[2]**o+(-0.489692340187)*x[2]
        arg[1,4,0,1]=(-0.150555966138)*x[0]**o+(0.362589320838)*x[0]+(0.515820915391)*x[1]**o+(0.742403661312)*x[1]+(0.230278843183)*x[2]**o+(-0.213977724366)*x[2]
        arg[1,4,1,0]=(-0.361161802362)*x[0]**o+(-0.49718332222)*x[0]+(0.143956046197)*x[1]**o+(-0.044550912758)*x[1]+(-0.299892874708)*x[2]**o+(0.370248709792)*x[2]
        arg[1,4,1,1]=(0.259365238173)*x[0]**o+(-0.474694548557)*x[0]+(0.732983396669)*x[1]**o+(-0.295151546774)*x[1]+(-0.337144628526)*x[2]**o+(0.664201775714)*x[2]
        arg[1,4,2,0]=(-0.318016754284)*x[0]**o+(0.555291443852)*x[0]+(0.538150727998)*x[1]**o+(-0.800118353132)*x[1]+(0.652262386596)*x[2]**o+(0.303156645857)*x[2]
        arg[1,4,2,1]=(0.0827142921796)*x[0]**o+(-0.88498648372)*x[0]+(-0.941249844349)*x[1]**o+(-0.162576659039)*x[1]+(0.867621252192)*x[2]**o+(0.0834583795227)*x[2]
        arg[2,0,0,0]=(0.0670088199264)*x[0]**o+(-0.0425961894346)*x[0]+(0.329605789302)*x[1]**o+(0.860453156363)*x[1]+(-0.37410469727)*x[2]**o+(0.55476432384)*x[2]
        arg[2,0,0,1]=(0.801738221879)*x[0]**o+(-0.836563530923)*x[0]+(-0.449651673142)*x[1]**o+(-0.19423133918)*x[1]+(0.947648200044)*x[2]**o+(0.0854067601423)*x[2]
        arg[2,0,1,0]=(0.986595046855)*x[0]**o+(-0.358319352175)*x[0]+(0.493917713513)*x[1]**o+(-0.92052282445)*x[1]+(-0.692726467259)*x[2]**o+(-0.0636943737807)*x[2]
        arg[2,0,1,1]=(0.38491150492)*x[0]**o+(0.766893245938)*x[0]+(0.420188917945)*x[1]**o+(0.43760538854)*x[1]+(-0.770749451981)*x[2]**o+(0.124882164378)*x[2]
        arg[2,0,2,0]=(0.775715718797)*x[0]**o+(0.89516703208)*x[0]+(-0.829021524898)*x[1]**o+(-0.650262454311)*x[1]+(-0.812556336598)*x[2]**o+(-0.997075743997)*x[2]
        arg[2,0,2,1]=(0.3503761008)*x[0]**o+(-0.822009897455)*x[0]+(-0.657657365354)*x[1]**o+(0.476295612156)*x[1]+(0.315554008711)*x[2]**o+(0.691631078049)*x[2]
        arg[2,1,0,0]=(0.680177375433)*x[0]**o+(-0.922726813311)*x[0]+(-0.581116746321)*x[1]**o+(0.628138099343)*x[1]+(-0.568262800293)*x[2]**o+(-0.228201382395)*x[2]
        arg[2,1,0,1]=(0.210094833324)*x[0]**o+(0.281549673876)*x[0]+(0.23785231073)*x[1]**o+(-0.103812455993)*x[1]+(0.530445971191)*x[2]**o+(-0.0830520449939)*x[2]
        arg[2,1,1,0]=(0.367133297563)*x[0]**o+(0.168289954585)*x[0]+(0.614117392787)*x[1]**o+(0.790467766841)*x[1]+(-0.958002150805)*x[2]**o+(0.328916248824)*x[2]
        arg[2,1,1,1]=(0.089535347231)*x[0]**o+(-0.93155993701)*x[0]+(-0.674670760794)*x[1]**o+(-0.274148261773)*x[1]+(-0.807020126056)*x[2]**o+(-0.323334769496)*x[2]
        arg[2,1,2,0]=(-0.948679835219)*x[0]**o+(-0.577235742425)*x[0]+(-0.732584834146)*x[1]**o+(-0.475242504288)*x[1]+(0.213663223262)*x[2]**o+(-0.151401354381)*x[2]
        arg[2,1,2,1]=(0.264513246206)*x[0]**o+(-0.605806437391)*x[0]+(-0.058730926693)*x[1]**o+(-0.661696304412)*x[1]+(-0.361277352457)*x[2]**o+(-0.804408258253)*x[2]
        arg[2,2,0,0]=(-0.144562358054)*x[0]**o+(-0.0960550555725)*x[0]+(-0.905096867803)*x[1]**o+(0.188861784103)*x[1]+(-0.616751005568)*x[2]**o+(0.604711974248)*x[2]
        arg[2,2,0,1]=(-0.0582524307858)*x[0]**o+(-0.503648879785)*x[0]+(-0.682755209956)*x[1]**o+(-0.808213164425)*x[1]+(0.539987276327)*x[2]**o+(0.650557703689)*x[2]
        arg[2,2,1,0]=(0.0607729567733)*x[0]**o+(-0.498819638428)*x[0]+(0.434442595334)*x[1]**o+(-0.491063977603)*x[1]+(0.499436212229)*x[2]**o+(-0.932071991206)*x[2]
        arg[2,2,1,1]=(-0.461295295294)*x[0]**o+(-0.153600767916)*x[0]+(-0.86452392231)*x[1]**o+(0.136760034153)*x[1]+(-0.471642416082)*x[2]**o+(0.453820403438)*x[2]
        arg[2,2,2,0]=(-0.970051888584)*x[0]**o+(0.415728825057)*x[0]+(0.971643909569)*x[1]**o+(0.923719420524)*x[1]+(-0.837313513804)*x[2]**o+(-0.738057676093)*x[2]
        arg[2,2,2,1]=(-0.941355918089)*x[0]**o+(-0.9629784524)*x[0]+(-0.931184348072)*x[1]**o+(0.720006097658)*x[1]+(-0.140750567978)*x[2]**o+(0.55381939119)*x[2]
        arg[2,3,0,0]=(0.753455933669)*x[0]**o+(0.352384922037)*x[0]+(0.737644348002)*x[1]**o+(-0.565144188248)*x[1]+(0.257417624014)*x[2]**o+(-0.660006641683)*x[2]
        arg[2,3,0,1]=(0.886996770925)*x[0]**o+(-0.876595418252)*x[0]+(-0.189778073871)*x[1]**o+(0.560867442336)*x[1]+(-0.262338156346)*x[2]**o+(0.614749098901)*x[2]
        arg[2,3,1,0]=(-0.722912374016)*x[0]**o+(0.447451310273)*x[0]+(0.397510140568)*x[1]**o+(0.932405917932)*x[1]+(-0.206969252336)*x[2]**o+(-0.897366208888)*x[2]
        arg[2,3,1,1]=(0.106324074547)*x[0]**o+(0.822109425884)*x[0]+(0.62699223277)*x[1]**o+(0.238371557839)*x[1]+(-0.422051841243)*x[2]**o+(0.868932130329)*x[2]
        arg[2,3,2,0]=(0.0155790371581)*x[0]**o+(-0.696029577918)*x[0]+(0.259909585217)*x[1]**o+(0.260209121862)*x[1]+(0.780977081459)*x[2]**o+(-0.819957534538)*x[2]
        arg[2,3,2,1]=(0.736223721904)*x[0]**o+(-0.433132598174)*x[0]+(-0.880913649812)*x[1]**o+(-0.375988777461)*x[1]+(-0.351763406307)*x[2]**o+(0.948468252467)*x[2]
        arg[2,4,0,0]=(0.943349180556)*x[0]**o+(-0.720044514103)*x[0]+(0.671900161807)*x[1]**o+(0.358571142942)*x[1]+(-0.132755998648)*x[2]**o+(0.0749815044365)*x[2]
        arg[2,4,0,1]=(0.639070752347)*x[0]**o+(0.551441363406)*x[0]+(0.507547844432)*x[1]**o+(-0.800043772307)*x[1]+(-0.506359503394)*x[2]**o+(0.952191432651)*x[2]
        arg[2,4,1,0]=(-0.655864732254)*x[0]**o+(0.142198390642)*x[0]+(0.075292486702)*x[1]**o+(0.191013189029)*x[1]+(0.106768453247)*x[2]**o+(0.974526515449)*x[2]
        arg[2,4,1,1]=(-0.354750205648)*x[0]**o+(0.141506849368)*x[0]+(-0.170074898141)*x[1]**o+(-0.0882776529335)*x[1]+(0.845254802876)*x[2]**o+(-0.435414692125)*x[2]
        arg[2,4,2,0]=(-0.412836027163)*x[0]**o+(0.830777132582)*x[0]+(-0.87554952236)*x[1]**o+(0.970647516313)*x[1]+(0.989864827286)*x[2]**o+(-0.67790755421)*x[2]
        arg[2,4,2,1]=(0.762907045031)*x[0]**o+(0.140771931118)*x[0]+(0.716859940142)*x[1]**o+(0.151769314189)*x[1]+(-0.812408853675)*x[2]**o+(0.755727078666)*x[2]
        arg[3,0,0,0]=(-0.13358237438)*x[0]**o+(-0.594735856789)*x[0]+(-0.367418402298)*x[1]**o+(0.5349162872)*x[1]+(0.238285320118)*x[2]**o+(0.999627906459)*x[2]
        arg[3,0,0,1]=(-0.537433369218)*x[0]**o+(0.133034652667)*x[0]+(-0.190302464596)*x[1]**o+(0.869919012762)*x[1]+(0.320294306008)*x[2]**o+(0.865908856629)*x[2]
        arg[3,0,1,0]=(0.480851043043)*x[0]**o+(0.833358259022)*x[0]+(0.893051476621)*x[1]**o+(-0.505255581199)*x[1]+(-0.15316145317)*x[2]**o+(-0.937103563776)*x[2]
        arg[3,0,1,1]=(0.765360137605)*x[0]**o+(0.502399552842)*x[0]+(-0.215707512276)*x[1]**o+(-0.402720816967)*x[1]+(0.231360098578)*x[2]**o+(0.732903327067)*x[2]
        arg[3,0,2,0]=(0.120836669045)*x[0]**o+(0.198346794548)*x[0]+(-0.28708447536)*x[1]**o+(0.289672075353)*x[1]+(-0.0305777738322)*x[2]**o+(-0.0686651569373)*x[2]
        arg[3,0,2,1]=(0.0574598844614)*x[0]**o+(0.031092492399)*x[0]+(-0.406805507341)*x[1]**o+(-0.623370509947)*x[1]+(-0.476094160966)*x[2]**o+(0.0840271758875)*x[2]
        arg[3,1,0,0]=(-0.569415249126)*x[0]**o+(-0.339675540399)*x[0]+(0.182536980707)*x[1]**o+(0.727480200453)*x[1]+(0.851672551504)*x[2]**o+(0.817312313075)*x[2]
        arg[3,1,0,1]=(0.921422642904)*x[0]**o+(0.697656143906)*x[0]+(0.969442847752)*x[1]**o+(-0.00957994710064)*x[1]+(-0.315230556399)*x[2]**o+(-0.870759158384)*x[2]
        arg[3,1,1,0]=(-0.813310465463)*x[0]**o+(-0.809362802903)*x[0]+(-0.0685101056428)*x[1]**o+(0.708195651847)*x[1]+(-0.876449456312)*x[2]**o+(0.882503536479)*x[2]
        arg[3,1,1,1]=(-0.277504922905)*x[0]**o+(-0.819428706347)*x[0]+(0.463203258731)*x[1]**o+(0.597493096401)*x[1]+(0.630522187524)*x[2]**o+(0.92480079466)*x[2]
        arg[3,1,2,0]=(-0.703167803764)*x[0]**o+(0.730459205382)*x[0]+(-0.507036921123)*x[1]**o+(0.419358545765)*x[1]+(-0.542758923679)*x[2]**o+(0.831844312917)*x[2]
        arg[3,1,2,1]=(0.135200299704)*x[0]**o+(-0.160734675188)*x[0]+(-0.893851689192)*x[1]**o+(0.748734906185)*x[1]+(0.970464447167)*x[2]**o+(0.277911376757)*x[2]
        arg[3,2,0,0]=(0.750104328784)*x[0]**o+(-0.818350593893)*x[0]+(-0.749900854281)*x[1]**o+(0.289117653816)*x[1]+(-0.219904822162)*x[2]**o+(0.805208050452)*x[2]
        arg[3,2,0,1]=(-0.129614389749)*x[0]**o+(-0.64067443816)*x[0]+(0.513013650233)*x[1]**o+(0.868869035341)*x[1]+(-0.040145121639)*x[2]**o+(-0.282350136119)*x[2]
        arg[3,2,1,0]=(-0.172814929104)*x[0]**o+(-0.296930471915)*x[0]+(-0.796114111585)*x[1]**o+(-0.210181080249)*x[1]+(-0.606107841173)*x[2]**o+(-0.954329084175)*x[2]
        arg[3,2,1,1]=(0.136149261959)*x[0]**o+(-0.613980696755)*x[0]+(0.968430714778)*x[1]**o+(0.683553952099)*x[1]+(-0.321383789121)*x[2]**o+(-0.139570984234)*x[2]
        arg[3,2,2,0]=(-0.273321720654)*x[0]**o+(0.963651110162)*x[0]+(-0.847402611499)*x[1]**o+(0.237131262905)*x[1]+(-0.617296821558)*x[2]**o+(-0.681856394629)*x[2]
        arg[3,2,2,1]=(-0.760071928624)*x[0]**o+(0.0345053779755)*x[0]+(-0.583680069732)*x[1]**o+(0.150497658077)*x[1]+(0.0257096621429)*x[2]**o+(-0.159406203385)*x[2]
        arg[3,3,0,0]=(-0.0668501746947)*x[0]**o+(-0.183645899611)*x[0]+(0.123667845765)*x[1]**o+(0.0928142420352)*x[1]+(-0.933250766317)*x[2]**o+(-0.629986949639)*x[2]
        arg[3,3,0,1]=(-0.0212790928206)*x[0]**o+(0.444036360325)*x[0]+(0.09182358127)*x[1]**o+(0.473088563672)*x[1]+(0.0852925382953)*x[2]**o+(-0.238517929786)*x[2]
        arg[3,3,1,0]=(0.0647235093728)*x[0]**o+(-0.266285898716)*x[0]+(0.944999560866)*x[1]**o+(-0.757696716486)*x[1]+(-0.439165278732)*x[2]**o+(0.960091479774)*x[2]
        arg[3,3,1,1]=(-0.792249238894)*x[0]**o+(-0.568585439587)*x[0]+(-0.0426655335607)*x[1]**o+(0.980004032206)*x[1]+(0.635227812889)*x[2]**o+(0.6687165384)*x[2]
        arg[3,3,2,0]=(0.340468442769)*x[0]**o+(0.791071154735)*x[0]+(-0.817919975547)*x[1]**o+(0.0328383066982)*x[1]+(-0.323093904705)*x[2]**o+(-0.302641620002)*x[2]
        arg[3,3,2,1]=(-0.476501733213)*x[0]**o+(-0.429845943537)*x[0]+(0.518551215268)*x[1]**o+(-0.551690143654)*x[1]+(0.591983049997)*x[2]**o+(0.586292683497)*x[2]
        arg[3,4,0,0]=(0.119036526248)*x[0]**o+(-0.156404044048)*x[0]+(0.158504048681)*x[1]**o+(-0.3779600709)*x[1]+(-0.362107671534)*x[2]**o+(-0.314508628804)*x[2]
        arg[3,4,0,1]=(0.382011103106)*x[0]**o+(-0.123711676964)*x[0]+(-0.939014457264)*x[1]**o+(-0.5986155807)*x[1]+(-0.852092491566)*x[2]**o+(-0.410438924788)*x[2]
        arg[3,4,1,0]=(0.83612586535)*x[0]**o+(-0.606098645092)*x[0]+(0.643469717148)*x[1]**o+(0.845106525654)*x[1]+(0.347455649691)*x[2]**o+(-0.166796580458)*x[2]
        arg[3,4,1,1]=(0.693431758156)*x[0]**o+(-0.313983401147)*x[0]+(-0.862690834893)*x[1]**o+(0.417233988055)*x[1]+(-0.361769997589)*x[2]**o+(-0.22749888397)*x[2]
        arg[3,4,2,0]=(0.680358405387)*x[0]**o+(-0.786041842215)*x[0]+(0.288823579415)*x[1]**o+(-0.961099851196)*x[1]+(0.669431785695)*x[2]**o+(0.956564160268)*x[2]
        arg[3,4,2,1]=(-0.21416500005)*x[0]**o+(0.48495191486)*x[0]+(-0.173465452645)*x[1]**o+(0.662530321557)*x[1]+(0.984573283824)*x[2]**o+(-0.568211383785)*x[2]
        ref[0,0,0,0]=(-0.615899511086)*x_ref[0]**o+(0.823637401824)*x_ref[0]+(-0.986967740198)*x_ref[1]**o+(-0.782088977641)*x_ref[1]+(-0.576799974679)*x_ref[2]**o+(-0.910942885711)*x_ref[2]
        ref[0,0,0,1]=(-0.790830593449)*x_ref[0]**o+(-0.897724279499)*x_ref[0]+(-0.881691805807)*x_ref[1]**o+(-0.133793238901)*x_ref[1]+(-0.0903391684489)*x_ref[2]**o+(0.659893469877)*x_ref[2]
        ref[0,0,1,0]=(-0.0567117900114)*x_ref[0]**o+(-0.858031985686)*x_ref[0]+(-0.397443445702)*x_ref[1]**o+(0.549982838823)*x_ref[1]+(-0.636905143813)*x_ref[2]**o+(0.8488142632)*x_ref[2]
        ref[0,0,1,1]=(-0.644557373082)*x_ref[0]**o+(-0.323654898522)*x_ref[0]+(0.210458705923)*x_ref[1]**o+(0.990436827296)*x_ref[1]+(-0.597872757151)*x_ref[2]**o+(-0.83865548618)*x_ref[2]
        ref[0,0,2,0]=(-0.29830667084)*x_ref[0]**o+(0.327035546909)*x_ref[0]+(0.132690049791)*x_ref[1]**o+(-0.55625443282)*x_ref[1]+(0.0972781385283)*x_ref[2]**o+(0.151885837608)*x_ref[2]
        ref[0,0,2,1]=(-0.435983465578)*x_ref[0]**o+(-0.0852522953446)*x_ref[0]+(0.257654807094)*x_ref[1]**o+(-0.336603197583)*x_ref[1]+(0.637913421592)*x_ref[2]**o+(0.8151042055)*x_ref[2]
        ref[0,1,0,0]=(-0.0373980191125)*x_ref[0]**o+(-0.45937285505)*x_ref[0]+(-0.857933684889)*x_ref[1]**o+(-0.843826686256)*x_ref[1]+(-0.727077320504)*x_ref[2]**o+(-0.65222863395)*x_ref[2]
        ref[0,1,0,1]=(0.63426017384)*x_ref[0]**o+(0.845708641601)*x_ref[0]+(0.246160302486)*x_ref[1]**o+(-0.660855572094)*x_ref[1]+(0.67420981328)*x_ref[2]**o+(-0.280803968781)*x_ref[2]
        ref[0,1,1,0]=(0.881193264597)*x_ref[0]**o+(-0.881936969515)*x_ref[0]+(-0.883246837152)*x_ref[1]**o+(0.182222471927)*x_ref[1]+(0.405380496272)*x_ref[2]**o+(0.175721746883)*x_ref[2]
        ref[0,1,1,1]=(-0.44447667237)*x_ref[0]**o+(0.956169274606)*x_ref[0]+(0.0370532940652)*x_ref[1]**o+(-0.181239635583)*x_ref[1]+(-0.678412161285)*x_ref[2]**o+(-0.653172157401)*x_ref[2]
        ref[0,1,2,0]=(-0.50038634718)*x_ref[0]**o+(0.00278321765167)*x_ref[0]+(-0.831539460218)*x_ref[1]**o+(-0.555243320385)*x_ref[1]+(-0.626789452003)*x_ref[2]**o+(-0.495085243933)*x_ref[2]
        ref[0,1,2,1]=(-0.32886359569)*x_ref[0]**o+(0.406099965475)*x_ref[0]+(0.764756177266)*x_ref[1]**o+(-0.272070344181)*x_ref[1]+(-0.41812596516)*x_ref[2]**o+(0.349063696376)*x_ref[2]
        ref[0,2,0,0]=(-0.580544978365)*x_ref[0]**o+(0.53441241334)*x_ref[0]+(-0.808749795876)*x_ref[1]**o+(0.172110801632)*x_ref[1]+(-0.939262606098)*x_ref[2]**o+(-0.724859325104)*x_ref[2]
        ref[0,2,0,1]=(0.256335221944)*x_ref[0]**o+(0.463522371018)*x_ref[0]+(0.990516478508)*x_ref[1]**o+(0.423931177054)*x_ref[1]+(0.184277693683)*x_ref[2]**o+(-0.272967447689)*x_ref[2]
        ref[0,2,1,0]=(-0.126943811008)*x_ref[0]**o+(0.645821387926)*x_ref[0]+(0.150370174987)*x_ref[1]**o+(-0.150954538301)*x_ref[1]+(-0.378720078559)*x_ref[2]**o+(0.0975963807109)*x_ref[2]
        ref[0,2,1,1]=(-0.451434544339)*x_ref[0]**o+(0.557314109088)*x_ref[0]+(-0.0402970749558)*x_ref[1]**o+(0.306984557206)*x_ref[1]+(-0.297018139186)*x_ref[2]**o+(0.593800698785)*x_ref[2]
        ref[0,2,2,0]=(-0.376199144911)*x_ref[0]**o+(-0.230454116373)*x_ref[0]+(0.992738297651)*x_ref[1]**o+(0.183510534927)*x_ref[1]+(0.409381711992)*x_ref[2]**o+(0.171945265727)*x_ref[2]
        ref[0,2,2,1]=(0.363233095971)*x_ref[0]**o+(-0.130589148249)*x_ref[0]+(0.809611042018)*x_ref[1]**o+(-0.354546307961)*x_ref[1]+(0.0631410169203)*x_ref[2]**o+(0.630306161781)*x_ref[2]
        ref[0,3,0,0]=(0.49655016161)*x_ref[0]**o+(0.878957258614)*x_ref[0]+(-0.397408696784)*x_ref[1]**o+(-0.199445321288)*x_ref[1]+(0.929857086663)*x_ref[2]**o+(-0.648770887161)*x_ref[2]
        ref[0,3,0,1]=(-0.85578334594)*x_ref[0]**o+(-0.294617109229)*x_ref[0]+(-0.0790295769768)*x_ref[1]**o+(0.82151942311)*x_ref[1]+(0.541628117444)*x_ref[2]**o+(0.805418174827)*x_ref[2]
        ref[0,3,1,0]=(0.800775454804)*x_ref[0]**o+(0.599550462983)*x_ref[0]+(-0.384788634082)*x_ref[1]**o+(-0.356933972363)*x_ref[1]+(-0.423116265473)*x_ref[2]**o+(-0.125492982003)*x_ref[2]
        ref[0,3,1,1]=(0.929652994661)*x_ref[0]**o+(-0.57396682406)*x_ref[0]+(0.313584116036)*x_ref[1]**o+(-0.190575043496)*x_ref[1]+(-0.29110884935)*x_ref[2]**o+(0.967530639925)*x_ref[2]
        ref[0,3,2,0]=(0.761136478868)*x_ref[0]**o+(0.565721170014)*x_ref[0]+(0.239619122131)*x_ref[1]**o+(-0.0821651979956)*x_ref[1]+(-0.699218775545)*x_ref[2]**o+(0.19973415585)*x_ref[2]
        ref[0,3,2,1]=(0.209543244648)*x_ref[0]**o+(0.93957796874)*x_ref[0]+(0.354405887426)*x_ref[1]**o+(0.504177262864)*x_ref[1]+(0.446866674894)*x_ref[2]**o+(-0.522936485464)*x_ref[2]
        ref[0,4,0,0]=(0.427541399652)*x_ref[0]**o+(-0.114806067206)*x_ref[0]+(0.236780877643)*x_ref[1]**o+(0.422226381272)*x_ref[1]+(-0.506450961214)*x_ref[2]**o+(0.616091720212)*x_ref[2]
        ref[0,4,0,1]=(0.698773571323)*x_ref[0]**o+(-0.357878713085)*x_ref[0]+(-0.453730359129)*x_ref[1]**o+(-0.328701041355)*x_ref[1]+(-0.28646346679)*x_ref[2]**o+(-0.0449784265203)*x_ref[2]
        ref[0,4,1,0]=(-0.419416306509)*x_ref[0]**o+(-0.1970394955)*x_ref[0]+(-0.814008932379)*x_ref[1]**o+(-0.805127589929)*x_ref[1]+(0.244059789875)*x_ref[2]**o+(-0.983027843847)*x_ref[2]
        ref[0,4,1,1]=(0.84837346004)*x_ref[0]**o+(0.501808510624)*x_ref[0]+(-0.105800351358)*x_ref[1]**o+(0.903999231718)*x_ref[1]+(-0.0473996978124)*x_ref[2]**o+(0.683647253381)*x_ref[2]
        ref[0,4,2,0]=(0.658722899682)*x_ref[0]**o+(-0.401846635016)*x_ref[0]+(-0.967322202581)*x_ref[1]**o+(-0.948306922217)*x_ref[1]+(0.496409308219)*x_ref[2]**o+(-0.184768777052)*x_ref[2]
        ref[0,4,2,1]=(-0.454468390888)*x_ref[0]**o+(-0.220640468215)*x_ref[0]+(0.995269925943)*x_ref[1]**o+(0.778951329211)*x_ref[1]+(0.0439534466302)*x_ref[2]**o+(0.806197038383)*x_ref[2]
        ref[1,0,0,0]=(-0.547461802236)*x_ref[0]**o+(0.370768827789)*x_ref[0]+(-0.619991757552)*x_ref[1]**o+(0.391727195645)*x_ref[1]+(-0.649274922953)*x_ref[2]**o+(-0.478788248429)*x_ref[2]
        ref[1,0,0,1]=(-0.899230505882)*x_ref[0]**o+(0.603064321735)*x_ref[0]+(-0.737475386349)*x_ref[1]**o+(0.594197953001)*x_ref[1]+(0.341324049064)*x_ref[2]**o+(0.285202633917)*x_ref[2]
        ref[1,0,1,0]=(-0.107063589618)*x_ref[0]**o+(-0.260491562358)*x_ref[0]+(-0.667059815618)*x_ref[1]**o+(0.473227354537)*x_ref[1]+(-0.587170510558)*x_ref[2]**o+(-0.740299218393)*x_ref[2]
        ref[1,0,1,1]=(0.87705715098)*x_ref[0]**o+(0.312753471827)*x_ref[0]+(-0.307530283315)*x_ref[1]**o+(0.195564644822)*x_ref[1]+(-0.64384485166)*x_ref[2]**o+(-0.776362951729)*x_ref[2]
        ref[1,0,2,0]=(0.0949408564739)*x_ref[0]**o+(0.765922179855)*x_ref[0]+(0.441215568253)*x_ref[1]**o+(0.339112969437)*x_ref[1]+(-0.0774615292449)*x_ref[2]**o+(0.673258055783)*x_ref[2]
        ref[1,0,2,1]=(0.114828212302)*x_ref[0]**o+(0.0831258766277)*x_ref[0]+(0.674540402248)*x_ref[1]**o+(-0.304921039094)*x_ref[1]+(-0.725514770578)*x_ref[2]**o+(-0.627584773746)*x_ref[2]
        ref[1,1,0,0]=(0.291561714675)*x_ref[0]**o+(0.898587624783)*x_ref[0]+(0.476696264545)*x_ref[1]**o+(-0.454648604978)*x_ref[1]+(0.677949150743)*x_ref[2]**o+(-0.59438289851)*x_ref[2]
        ref[1,1,0,1]=(0.356687258009)*x_ref[0]**o+(0.806236251629)*x_ref[0]+(0.843636690354)*x_ref[1]**o+(-0.649776947181)*x_ref[1]+(0.208542122388)*x_ref[2]**o+(-0.0726045573458)*x_ref[2]
        ref[1,1,1,0]=(0.712800450558)*x_ref[0]**o+(0.748271875755)*x_ref[0]+(0.403880649356)*x_ref[1]**o+(0.956549336754)*x_ref[1]+(0.0420128495043)*x_ref[2]**o+(-0.83177870499)*x_ref[2]
        ref[1,1,1,1]=(-0.886862904174)*x_ref[0]**o+(-0.03822986594)*x_ref[0]+(0.587151517496)*x_ref[1]**o+(-0.120272060898)*x_ref[1]+(-0.573366391933)*x_ref[2]**o+(-0.59772763241)*x_ref[2]
        ref[1,1,2,0]=(0.311960197804)*x_ref[0]**o+(0.486854949444)*x_ref[0]+(-0.647995651954)*x_ref[1]**o+(0.30104277964)*x_ref[1]+(-0.174718703748)*x_ref[2]**o+(-0.506993216295)*x_ref[2]
        ref[1,1,2,1]=(0.979056206991)*x_ref[0]**o+(0.892104464812)*x_ref[0]+(0.0507432703245)*x_ref[1]**o+(0.103630858454)*x_ref[1]+(-0.0254965952266)*x_ref[2]**o+(0.13034027239)*x_ref[2]
        ref[1,2,0,0]=(0.259797966548)*x_ref[0]**o+(0.53018999371)*x_ref[0]+(0.533745722992)*x_ref[1]**o+(0.53029084173)*x_ref[1]+(-0.595126791061)*x_ref[2]**o+(-0.51668295558)*x_ref[2]
        ref[1,2,0,1]=(0.926395590808)*x_ref[0]**o+(-0.0700141207179)*x_ref[0]+(-0.973392454012)*x_ref[1]**o+(-0.46204067097)*x_ref[1]+(0.814964497207)*x_ref[2]**o+(-0.183546091376)*x_ref[2]
        ref[1,2,1,0]=(-0.0872635441658)*x_ref[0]**o+(-0.0307630839035)*x_ref[0]+(0.539798653095)*x_ref[1]**o+(0.966367734848)*x_ref[1]+(-0.817495053612)*x_ref[2]**o+(-0.268397867848)*x_ref[2]
        ref[1,2,1,1]=(-0.933074613921)*x_ref[0]**o+(0.48896533938)*x_ref[0]+(0.453696133488)*x_ref[1]**o+(-0.238730406749)*x_ref[1]+(-0.696785951086)*x_ref[2]**o+(-0.359858385642)*x_ref[2]
        ref[1,2,2,0]=(-0.844406286809)*x_ref[0]**o+(-0.506332243946)*x_ref[0]+(0.976436731658)*x_ref[1]**o+(-0.201489161857)*x_ref[1]+(0.851236173166)*x_ref[2]**o+(-0.998803612947)*x_ref[2]
        ref[1,2,2,1]=(0.259707321449)*x_ref[0]**o+(-0.888149478655)*x_ref[0]+(-0.273502842869)*x_ref[1]**o+(0.291193582158)*x_ref[1]+(-0.897542178027)*x_ref[2]**o+(-0.345994824429)*x_ref[2]
        ref[1,3,0,0]=(0.961298984697)*x_ref[0]**o+(0.248141115092)*x_ref[0]+(-0.623053025178)*x_ref[1]**o+(-0.775095452078)*x_ref[1]+(0.955362707049)*x_ref[2]**o+(-0.517547333705)*x_ref[2]
        ref[1,3,0,1]=(0.858098596227)*x_ref[0]**o+(0.309307483873)*x_ref[0]+(-0.209217034755)*x_ref[1]**o+(0.783803593379)*x_ref[1]+(0.138164128092)*x_ref[2]**o+(-0.922854003461)*x_ref[2]
        ref[1,3,1,0]=(-0.139017942141)*x_ref[0]**o+(-0.338223039119)*x_ref[0]+(0.737882941498)*x_ref[1]**o+(0.948854903087)*x_ref[1]+(-0.0792461633266)*x_ref[2]**o+(0.848875660203)*x_ref[2]
        ref[1,3,1,1]=(0.693284332212)*x_ref[0]**o+(-0.392272875348)*x_ref[0]+(-0.475709045479)*x_ref[1]**o+(0.182948077236)*x_ref[1]+(0.0277569761538)*x_ref[2]**o+(-0.550294947596)*x_ref[2]
        ref[1,3,2,0]=(-0.105381649505)*x_ref[0]**o+(-0.403051690238)*x_ref[0]+(-0.722219735818)*x_ref[1]**o+(-0.380114388869)*x_ref[1]+(0.321452572491)*x_ref[2]**o+(0.858583245253)*x_ref[2]
        ref[1,3,2,1]=(-0.405370062024)*x_ref[0]**o+(0.0743704840047)*x_ref[0]+(0.998947017815)*x_ref[1]**o+(0.983611079522)*x_ref[1]+(0.854202982674)*x_ref[2]**o+(0.288478746498)*x_ref[2]
        ref[1,4,0,0]=(0.0584782489998)*x_ref[0]**o+(0.999286927604)*x_ref[0]+(-0.458166020555)*x_ref[1]**o+(0.529008844128)*x_ref[1]+(0.424592641922)*x_ref[2]**o+(-0.489692340187)*x_ref[2]
        ref[1,4,0,1]=(-0.150555966138)*x_ref[0]**o+(0.362589320838)*x_ref[0]+(0.515820915391)*x_ref[1]**o+(0.742403661312)*x_ref[1]+(0.230278843183)*x_ref[2]**o+(-0.213977724366)*x_ref[2]
        ref[1,4,1,0]=(-0.361161802362)*x_ref[0]**o+(-0.49718332222)*x_ref[0]+(0.143956046197)*x_ref[1]**o+(-0.044550912758)*x_ref[1]+(-0.299892874708)*x_ref[2]**o+(0.370248709792)*x_ref[2]
        ref[1,4,1,1]=(0.259365238173)*x_ref[0]**o+(-0.474694548557)*x_ref[0]+(0.732983396669)*x_ref[1]**o+(-0.295151546774)*x_ref[1]+(-0.337144628526)*x_ref[2]**o+(0.664201775714)*x_ref[2]
        ref[1,4,2,0]=(-0.318016754284)*x_ref[0]**o+(0.555291443852)*x_ref[0]+(0.538150727998)*x_ref[1]**o+(-0.800118353132)*x_ref[1]+(0.652262386596)*x_ref[2]**o+(0.303156645857)*x_ref[2]
        ref[1,4,2,1]=(0.0827142921796)*x_ref[0]**o+(-0.88498648372)*x_ref[0]+(-0.941249844349)*x_ref[1]**o+(-0.162576659039)*x_ref[1]+(0.867621252192)*x_ref[2]**o+(0.0834583795227)*x_ref[2]
        ref[2,0,0,0]=(0.0670088199264)*x_ref[0]**o+(-0.0425961894346)*x_ref[0]+(0.329605789302)*x_ref[1]**o+(0.860453156363)*x_ref[1]+(-0.37410469727)*x_ref[2]**o+(0.55476432384)*x_ref[2]
        ref[2,0,0,1]=(0.801738221879)*x_ref[0]**o+(-0.836563530923)*x_ref[0]+(-0.449651673142)*x_ref[1]**o+(-0.19423133918)*x_ref[1]+(0.947648200044)*x_ref[2]**o+(0.0854067601423)*x_ref[2]
        ref[2,0,1,0]=(0.986595046855)*x_ref[0]**o+(-0.358319352175)*x_ref[0]+(0.493917713513)*x_ref[1]**o+(-0.92052282445)*x_ref[1]+(-0.692726467259)*x_ref[2]**o+(-0.0636943737807)*x_ref[2]
        ref[2,0,1,1]=(0.38491150492)*x_ref[0]**o+(0.766893245938)*x_ref[0]+(0.420188917945)*x_ref[1]**o+(0.43760538854)*x_ref[1]+(-0.770749451981)*x_ref[2]**o+(0.124882164378)*x_ref[2]
        ref[2,0,2,0]=(0.775715718797)*x_ref[0]**o+(0.89516703208)*x_ref[0]+(-0.829021524898)*x_ref[1]**o+(-0.650262454311)*x_ref[1]+(-0.812556336598)*x_ref[2]**o+(-0.997075743997)*x_ref[2]
        ref[2,0,2,1]=(0.3503761008)*x_ref[0]**o+(-0.822009897455)*x_ref[0]+(-0.657657365354)*x_ref[1]**o+(0.476295612156)*x_ref[1]+(0.315554008711)*x_ref[2]**o+(0.691631078049)*x_ref[2]
        ref[2,1,0,0]=(0.680177375433)*x_ref[0]**o+(-0.922726813311)*x_ref[0]+(-0.581116746321)*x_ref[1]**o+(0.628138099343)*x_ref[1]+(-0.568262800293)*x_ref[2]**o+(-0.228201382395)*x_ref[2]
        ref[2,1,0,1]=(0.210094833324)*x_ref[0]**o+(0.281549673876)*x_ref[0]+(0.23785231073)*x_ref[1]**o+(-0.103812455993)*x_ref[1]+(0.530445971191)*x_ref[2]**o+(-0.0830520449939)*x_ref[2]
        ref[2,1,1,0]=(0.367133297563)*x_ref[0]**o+(0.168289954585)*x_ref[0]+(0.614117392787)*x_ref[1]**o+(0.790467766841)*x_ref[1]+(-0.958002150805)*x_ref[2]**o+(0.328916248824)*x_ref[2]
        ref[2,1,1,1]=(0.089535347231)*x_ref[0]**o+(-0.93155993701)*x_ref[0]+(-0.674670760794)*x_ref[1]**o+(-0.274148261773)*x_ref[1]+(-0.807020126056)*x_ref[2]**o+(-0.323334769496)*x_ref[2]
        ref[2,1,2,0]=(-0.948679835219)*x_ref[0]**o+(-0.577235742425)*x_ref[0]+(-0.732584834146)*x_ref[1]**o+(-0.475242504288)*x_ref[1]+(0.213663223262)*x_ref[2]**o+(-0.151401354381)*x_ref[2]
        ref[2,1,2,1]=(0.264513246206)*x_ref[0]**o+(-0.605806437391)*x_ref[0]+(-0.058730926693)*x_ref[1]**o+(-0.661696304412)*x_ref[1]+(-0.361277352457)*x_ref[2]**o+(-0.804408258253)*x_ref[2]
        ref[2,2,0,0]=(-0.144562358054)*x_ref[0]**o+(-0.0960550555725)*x_ref[0]+(-0.905096867803)*x_ref[1]**o+(0.188861784103)*x_ref[1]+(-0.616751005568)*x_ref[2]**o+(0.604711974248)*x_ref[2]
        ref[2,2,0,1]=(-0.0582524307858)*x_ref[0]**o+(-0.503648879785)*x_ref[0]+(-0.682755209956)*x_ref[1]**o+(-0.808213164425)*x_ref[1]+(0.539987276327)*x_ref[2]**o+(0.650557703689)*x_ref[2]
        ref[2,2,1,0]=(0.0607729567733)*x_ref[0]**o+(-0.498819638428)*x_ref[0]+(0.434442595334)*x_ref[1]**o+(-0.491063977603)*x_ref[1]+(0.499436212229)*x_ref[2]**o+(-0.932071991206)*x_ref[2]
        ref[2,2,1,1]=(-0.461295295294)*x_ref[0]**o+(-0.153600767916)*x_ref[0]+(-0.86452392231)*x_ref[1]**o+(0.136760034153)*x_ref[1]+(-0.471642416082)*x_ref[2]**o+(0.453820403438)*x_ref[2]
        ref[2,2,2,0]=(-0.970051888584)*x_ref[0]**o+(0.415728825057)*x_ref[0]+(0.971643909569)*x_ref[1]**o+(0.923719420524)*x_ref[1]+(-0.837313513804)*x_ref[2]**o+(-0.738057676093)*x_ref[2]
        ref[2,2,2,1]=(-0.941355918089)*x_ref[0]**o+(-0.9629784524)*x_ref[0]+(-0.931184348072)*x_ref[1]**o+(0.720006097658)*x_ref[1]+(-0.140750567978)*x_ref[2]**o+(0.55381939119)*x_ref[2]
        ref[2,3,0,0]=(0.753455933669)*x_ref[0]**o+(0.352384922037)*x_ref[0]+(0.737644348002)*x_ref[1]**o+(-0.565144188248)*x_ref[1]+(0.257417624014)*x_ref[2]**o+(-0.660006641683)*x_ref[2]
        ref[2,3,0,1]=(0.886996770925)*x_ref[0]**o+(-0.876595418252)*x_ref[0]+(-0.189778073871)*x_ref[1]**o+(0.560867442336)*x_ref[1]+(-0.262338156346)*x_ref[2]**o+(0.614749098901)*x_ref[2]
        ref[2,3,1,0]=(-0.722912374016)*x_ref[0]**o+(0.447451310273)*x_ref[0]+(0.397510140568)*x_ref[1]**o+(0.932405917932)*x_ref[1]+(-0.206969252336)*x_ref[2]**o+(-0.897366208888)*x_ref[2]
        ref[2,3,1,1]=(0.106324074547)*x_ref[0]**o+(0.822109425884)*x_ref[0]+(0.62699223277)*x_ref[1]**o+(0.238371557839)*x_ref[1]+(-0.422051841243)*x_ref[2]**o+(0.868932130329)*x_ref[2]
        ref[2,3,2,0]=(0.0155790371581)*x_ref[0]**o+(-0.696029577918)*x_ref[0]+(0.259909585217)*x_ref[1]**o+(0.260209121862)*x_ref[1]+(0.780977081459)*x_ref[2]**o+(-0.819957534538)*x_ref[2]
        ref[2,3,2,1]=(0.736223721904)*x_ref[0]**o+(-0.433132598174)*x_ref[0]+(-0.880913649812)*x_ref[1]**o+(-0.375988777461)*x_ref[1]+(-0.351763406307)*x_ref[2]**o+(0.948468252467)*x_ref[2]
        ref[2,4,0,0]=(0.943349180556)*x_ref[0]**o+(-0.720044514103)*x_ref[0]+(0.671900161807)*x_ref[1]**o+(0.358571142942)*x_ref[1]+(-0.132755998648)*x_ref[2]**o+(0.0749815044365)*x_ref[2]
        ref[2,4,0,1]=(0.639070752347)*x_ref[0]**o+(0.551441363406)*x_ref[0]+(0.507547844432)*x_ref[1]**o+(-0.800043772307)*x_ref[1]+(-0.506359503394)*x_ref[2]**o+(0.952191432651)*x_ref[2]
        ref[2,4,1,0]=(-0.655864732254)*x_ref[0]**o+(0.142198390642)*x_ref[0]+(0.075292486702)*x_ref[1]**o+(0.191013189029)*x_ref[1]+(0.106768453247)*x_ref[2]**o+(0.974526515449)*x_ref[2]
        ref[2,4,1,1]=(-0.354750205648)*x_ref[0]**o+(0.141506849368)*x_ref[0]+(-0.170074898141)*x_ref[1]**o+(-0.0882776529335)*x_ref[1]+(0.845254802876)*x_ref[2]**o+(-0.435414692125)*x_ref[2]
        ref[2,4,2,0]=(-0.412836027163)*x_ref[0]**o+(0.830777132582)*x_ref[0]+(-0.87554952236)*x_ref[1]**o+(0.970647516313)*x_ref[1]+(0.989864827286)*x_ref[2]**o+(-0.67790755421)*x_ref[2]
        ref[2,4,2,1]=(0.762907045031)*x_ref[0]**o+(0.140771931118)*x_ref[0]+(0.716859940142)*x_ref[1]**o+(0.151769314189)*x_ref[1]+(-0.812408853675)*x_ref[2]**o+(0.755727078666)*x_ref[2]
        ref[3,0,0,0]=(-0.13358237438)*x_ref[0]**o+(-0.594735856789)*x_ref[0]+(-0.367418402298)*x_ref[1]**o+(0.5349162872)*x_ref[1]+(0.238285320118)*x_ref[2]**o+(0.999627906459)*x_ref[2]
        ref[3,0,0,1]=(-0.537433369218)*x_ref[0]**o+(0.133034652667)*x_ref[0]+(-0.190302464596)*x_ref[1]**o+(0.869919012762)*x_ref[1]+(0.320294306008)*x_ref[2]**o+(0.865908856629)*x_ref[2]
        ref[3,0,1,0]=(0.480851043043)*x_ref[0]**o+(0.833358259022)*x_ref[0]+(0.893051476621)*x_ref[1]**o+(-0.505255581199)*x_ref[1]+(-0.15316145317)*x_ref[2]**o+(-0.937103563776)*x_ref[2]
        ref[3,0,1,1]=(0.765360137605)*x_ref[0]**o+(0.502399552842)*x_ref[0]+(-0.215707512276)*x_ref[1]**o+(-0.402720816967)*x_ref[1]+(0.231360098578)*x_ref[2]**o+(0.732903327067)*x_ref[2]
        ref[3,0,2,0]=(0.120836669045)*x_ref[0]**o+(0.198346794548)*x_ref[0]+(-0.28708447536)*x_ref[1]**o+(0.289672075353)*x_ref[1]+(-0.0305777738322)*x_ref[2]**o+(-0.0686651569373)*x_ref[2]
        ref[3,0,2,1]=(0.0574598844614)*x_ref[0]**o+(0.031092492399)*x_ref[0]+(-0.406805507341)*x_ref[1]**o+(-0.623370509947)*x_ref[1]+(-0.476094160966)*x_ref[2]**o+(0.0840271758875)*x_ref[2]
        ref[3,1,0,0]=(-0.569415249126)*x_ref[0]**o+(-0.339675540399)*x_ref[0]+(0.182536980707)*x_ref[1]**o+(0.727480200453)*x_ref[1]+(0.851672551504)*x_ref[2]**o+(0.817312313075)*x_ref[2]
        ref[3,1,0,1]=(0.921422642904)*x_ref[0]**o+(0.697656143906)*x_ref[0]+(0.969442847752)*x_ref[1]**o+(-0.00957994710064)*x_ref[1]+(-0.315230556399)*x_ref[2]**o+(-0.870759158384)*x_ref[2]
        ref[3,1,1,0]=(-0.813310465463)*x_ref[0]**o+(-0.809362802903)*x_ref[0]+(-0.0685101056428)*x_ref[1]**o+(0.708195651847)*x_ref[1]+(-0.876449456312)*x_ref[2]**o+(0.882503536479)*x_ref[2]
        ref[3,1,1,1]=(-0.277504922905)*x_ref[0]**o+(-0.819428706347)*x_ref[0]+(0.463203258731)*x_ref[1]**o+(0.597493096401)*x_ref[1]+(0.630522187524)*x_ref[2]**o+(0.92480079466)*x_ref[2]
        ref[3,1,2,0]=(-0.703167803764)*x_ref[0]**o+(0.730459205382)*x_ref[0]+(-0.507036921123)*x_ref[1]**o+(0.419358545765)*x_ref[1]+(-0.542758923679)*x_ref[2]**o+(0.831844312917)*x_ref[2]
        ref[3,1,2,1]=(0.135200299704)*x_ref[0]**o+(-0.160734675188)*x_ref[0]+(-0.893851689192)*x_ref[1]**o+(0.748734906185)*x_ref[1]+(0.970464447167)*x_ref[2]**o+(0.277911376757)*x_ref[2]
        ref[3,2,0,0]=(0.750104328784)*x_ref[0]**o+(-0.818350593893)*x_ref[0]+(-0.749900854281)*x_ref[1]**o+(0.289117653816)*x_ref[1]+(-0.219904822162)*x_ref[2]**o+(0.805208050452)*x_ref[2]
        ref[3,2,0,1]=(-0.129614389749)*x_ref[0]**o+(-0.64067443816)*x_ref[0]+(0.513013650233)*x_ref[1]**o+(0.868869035341)*x_ref[1]+(-0.040145121639)*x_ref[2]**o+(-0.282350136119)*x_ref[2]
        ref[3,2,1,0]=(-0.172814929104)*x_ref[0]**o+(-0.296930471915)*x_ref[0]+(-0.796114111585)*x_ref[1]**o+(-0.210181080249)*x_ref[1]+(-0.606107841173)*x_ref[2]**o+(-0.954329084175)*x_ref[2]
        ref[3,2,1,1]=(0.136149261959)*x_ref[0]**o+(-0.613980696755)*x_ref[0]+(0.968430714778)*x_ref[1]**o+(0.683553952099)*x_ref[1]+(-0.321383789121)*x_ref[2]**o+(-0.139570984234)*x_ref[2]
        ref[3,2,2,0]=(-0.273321720654)*x_ref[0]**o+(0.963651110162)*x_ref[0]+(-0.847402611499)*x_ref[1]**o+(0.237131262905)*x_ref[1]+(-0.617296821558)*x_ref[2]**o+(-0.681856394629)*x_ref[2]
        ref[3,2,2,1]=(-0.760071928624)*x_ref[0]**o+(0.0345053779755)*x_ref[0]+(-0.583680069732)*x_ref[1]**o+(0.150497658077)*x_ref[1]+(0.0257096621429)*x_ref[2]**o+(-0.159406203385)*x_ref[2]
        ref[3,3,0,0]=(-0.0668501746947)*x_ref[0]**o+(-0.183645899611)*x_ref[0]+(0.123667845765)*x_ref[1]**o+(0.0928142420352)*x_ref[1]+(-0.933250766317)*x_ref[2]**o+(-0.629986949639)*x_ref[2]
        ref[3,3,0,1]=(-0.0212790928206)*x_ref[0]**o+(0.444036360325)*x_ref[0]+(0.09182358127)*x_ref[1]**o+(0.473088563672)*x_ref[1]+(0.0852925382953)*x_ref[2]**o+(-0.238517929786)*x_ref[2]
        ref[3,3,1,0]=(0.0647235093728)*x_ref[0]**o+(-0.266285898716)*x_ref[0]+(0.944999560866)*x_ref[1]**o+(-0.757696716486)*x_ref[1]+(-0.439165278732)*x_ref[2]**o+(0.960091479774)*x_ref[2]
        ref[3,3,1,1]=(-0.792249238894)*x_ref[0]**o+(-0.568585439587)*x_ref[0]+(-0.0426655335607)*x_ref[1]**o+(0.980004032206)*x_ref[1]+(0.635227812889)*x_ref[2]**o+(0.6687165384)*x_ref[2]
        ref[3,3,2,0]=(0.340468442769)*x_ref[0]**o+(0.791071154735)*x_ref[0]+(-0.817919975547)*x_ref[1]**o+(0.0328383066982)*x_ref[1]+(-0.323093904705)*x_ref[2]**o+(-0.302641620002)*x_ref[2]
        ref[3,3,2,1]=(-0.476501733213)*x_ref[0]**o+(-0.429845943537)*x_ref[0]+(0.518551215268)*x_ref[1]**o+(-0.551690143654)*x_ref[1]+(0.591983049997)*x_ref[2]**o+(0.586292683497)*x_ref[2]
        ref[3,4,0,0]=(0.119036526248)*x_ref[0]**o+(-0.156404044048)*x_ref[0]+(0.158504048681)*x_ref[1]**o+(-0.3779600709)*x_ref[1]+(-0.362107671534)*x_ref[2]**o+(-0.314508628804)*x_ref[2]
        ref[3,4,0,1]=(0.382011103106)*x_ref[0]**o+(-0.123711676964)*x_ref[0]+(-0.939014457264)*x_ref[1]**o+(-0.5986155807)*x_ref[1]+(-0.852092491566)*x_ref[2]**o+(-0.410438924788)*x_ref[2]
        ref[3,4,1,0]=(0.83612586535)*x_ref[0]**o+(-0.606098645092)*x_ref[0]+(0.643469717148)*x_ref[1]**o+(0.845106525654)*x_ref[1]+(0.347455649691)*x_ref[2]**o+(-0.166796580458)*x_ref[2]
        ref[3,4,1,1]=(0.693431758156)*x_ref[0]**o+(-0.313983401147)*x_ref[0]+(-0.862690834893)*x_ref[1]**o+(0.417233988055)*x_ref[1]+(-0.361769997589)*x_ref[2]**o+(-0.22749888397)*x_ref[2]
        ref[3,4,2,0]=(0.680358405387)*x_ref[0]**o+(-0.786041842215)*x_ref[0]+(0.288823579415)*x_ref[1]**o+(-0.961099851196)*x_ref[1]+(0.669431785695)*x_ref[2]**o+(0.956564160268)*x_ref[2]
        ref[3,4,2,1]=(-0.21416500005)*x_ref[0]**o+(0.48495191486)*x_ref[0]+(-0.173465452645)*x_ref[1]**o+(0.662530321557)*x_ref[1]+(0.984573283824)*x_ref[2]**o+(-0.568211383785)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.852985368636)*x[0]**o+(-0.691601416209)*x[0]+(-0.689994415069)*x[1]**o+(-0.471926256627)*x[1]
        ref=(-0.852985368636)*x_ref[0]**o+(-0.691601416209)*x_ref[0]+(-0.689994415069)*x_ref[1]**o+(-0.471926256627)*x_ref[1]
      else:
        s=(-0.796357734502)*x[0]**o+(0.158918179797)*x[0]+(0.580745911056)*x[1]**o+(0.786294909139)*x[1]+(0.436014958708)*x[2]**o+(-0.600115428014)*x[2]
        ref=(-0.796357734502)*x_ref[0]**o+(0.158918179797)*x_ref[0]+(0.580745911056)*x_ref[1]**o+(0.786294909139)*x_ref[1]+(0.436014958708)*x_ref[2]**o+(-0.600115428014)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.997766043376)*x[0]**o+(0.495776426131)*x[0]+(-0.387839677144)*x[1]**o+(0.774520715668)*x[1]
        s[1]=(0.947961152483)*x[0]**o+(0.987829314406)*x[0]+(0.32828485757)*x[1]**o+(-0.617666605292)*x[1]
        ref[0]=(0.997766043376)*x_ref[0]**o+(0.495776426131)*x_ref[0]+(-0.387839677144)*x_ref[1]**o+(0.774520715668)*x_ref[1]
        ref[1]=(0.947961152483)*x_ref[0]**o+(0.987829314406)*x_ref[0]+(0.32828485757)*x_ref[1]**o+(-0.617666605292)*x_ref[1]
      else:
        s[0]=(-0.13474557089)*x[0]**o+(-0.588190696155)*x[0]+(0.128472574732)*x[1]**o+(0.690322787372)*x[1]+(0.884541159956)*x[2]**o+(-0.886834154713)*x[2]
        s[1]=(-0.751296359458)*x[0]**o+(-0.659908205442)*x[0]+(-0.415219823654)*x[1]**o+(0.550397764781)*x[1]+(-0.728215097593)*x[2]**o+(0.582854885508)*x[2]
        ref[0]=(-0.13474557089)*x_ref[0]**o+(-0.588190696155)*x_ref[0]+(0.128472574732)*x_ref[1]**o+(0.690322787372)*x_ref[1]+(0.884541159956)*x_ref[2]**o+(-0.886834154713)*x_ref[2]
        ref[1]=(-0.751296359458)*x_ref[0]**o+(-0.659908205442)*x_ref[0]+(-0.415219823654)*x_ref[1]**o+(0.550397764781)*x_ref[1]+(-0.728215097593)*x_ref[2]**o+(0.582854885508)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.540707708777)*x[0]**o+(-0.666948327254)*x[0]+(-0.471649914321)*x[1]**o+(0.65162222166)*x[1]
        s[0,1]=(0.39265499863)*x[0]**o+(0.927768486766)*x[0]+(-0.625791400147)*x[1]**o+(-0.406599346038)*x[1]
        s[0,2]=(0.721211493492)*x[0]**o+(0.0169359532239)*x[0]+(-0.450701248736)*x[1]**o+(0.792695848136)*x[1]
        s[0,3]=(0.349994038242)*x[0]**o+(0.124108838306)*x[0]+(0.221156059551)*x[1]**o+(-0.891451468693)*x[1]
        s[0,4]=(0.493264497246)*x[0]**o+(0.234167381112)*x[0]+(0.543699100028)*x[1]**o+(-0.838472895241)*x[1]
        s[1,0]=(0.359533817281)*x[0]**o+(-0.062563306784)*x[0]+(0.467858739349)*x[1]**o+(0.692754015398)*x[1]
        s[1,1]=(0.888396152909)*x[0]**o+(0.184738712864)*x[0]+(-0.0379184183599)*x[1]**o+(0.991506780667)*x[1]
        s[1,2]=(-0.68045741795)*x[0]**o+(0.832740353977)*x[0]+(0.491010493882)*x[1]**o+(0.454304365237)*x[1]
        s[1,3]=(-0.179660488547)*x[0]**o+(0.628405133401)*x[0]+(-0.69684323253)*x[1]**o+(0.889439582492)*x[1]
        s[1,4]=(-0.379088893433)*x[0]**o+(0.497309923662)*x[0]+(0.153969715423)*x[1]**o+(-0.386632220307)*x[1]
        s[2,0]=(-0.186297693534)*x[0]**o+(0.489487382751)*x[0]+(0.644068223202)*x[1]**o+(0.0743289639696)*x[1]
        s[2,1]=(0.593905543479)*x[0]**o+(-0.966302120537)*x[0]+(-0.861407416257)*x[1]**o+(0.2431998414)*x[1]
        s[2,2]=(0.683336278471)*x[0]**o+(0.337655287244)*x[0]+(0.876310641583)*x[1]**o+(-0.850438003892)*x[1]
        s[2,3]=(-0.109203794389)*x[0]**o+(0.245455391347)*x[0]+(-0.30542844788)*x[1]**o+(0.846645969509)*x[1]
        s[2,4]=(-0.478221810644)*x[0]**o+(0.555082091582)*x[0]+(-0.331626471767)*x[1]**o+(-0.363786378267)*x[1]
        s[3,0]=(-0.0373544633657)*x[0]**o+(-0.821373667974)*x[0]+(-0.569127067519)*x[1]**o+(0.178159969962)*x[1]
        s[3,1]=(0.574056120213)*x[0]**o+(-0.557444624837)*x[0]+(-0.858492289239)*x[1]**o+(0.82616627128)*x[1]
        s[3,2]=(0.0934423420331)*x[0]**o+(-0.64321213739)*x[0]+(0.0832655456405)*x[1]**o+(0.245929781754)*x[1]
        s[3,3]=(0.0254040950683)*x[0]**o+(0.169079366685)*x[0]+(0.150029664562)*x[1]**o+(0.987260372009)*x[1]
        s[3,4]=(0.201458676701)*x[0]**o+(0.840874176438)*x[0]+(-0.533157142378)*x[1]**o+(-0.216306366879)*x[1]
        ref[0,0]=(0.540707708777)*x_ref[0]**o+(-0.666948327254)*x_ref[0]+(-0.471649914321)*x_ref[1]**o+(0.65162222166)*x_ref[1]
        ref[0,1]=(0.39265499863)*x_ref[0]**o+(0.927768486766)*x_ref[0]+(-0.625791400147)*x_ref[1]**o+(-0.406599346038)*x_ref[1]
        ref[0,2]=(0.721211493492)*x_ref[0]**o+(0.0169359532239)*x_ref[0]+(-0.450701248736)*x_ref[1]**o+(0.792695848136)*x_ref[1]
        ref[0,3]=(0.349994038242)*x_ref[0]**o+(0.124108838306)*x_ref[0]+(0.221156059551)*x_ref[1]**o+(-0.891451468693)*x_ref[1]
        ref[0,4]=(0.493264497246)*x_ref[0]**o+(0.234167381112)*x_ref[0]+(0.543699100028)*x_ref[1]**o+(-0.838472895241)*x_ref[1]
        ref[1,0]=(0.359533817281)*x_ref[0]**o+(-0.062563306784)*x_ref[0]+(0.467858739349)*x_ref[1]**o+(0.692754015398)*x_ref[1]
        ref[1,1]=(0.888396152909)*x_ref[0]**o+(0.184738712864)*x_ref[0]+(-0.0379184183599)*x_ref[1]**o+(0.991506780667)*x_ref[1]
        ref[1,2]=(-0.68045741795)*x_ref[0]**o+(0.832740353977)*x_ref[0]+(0.491010493882)*x_ref[1]**o+(0.454304365237)*x_ref[1]
        ref[1,3]=(-0.179660488547)*x_ref[0]**o+(0.628405133401)*x_ref[0]+(-0.69684323253)*x_ref[1]**o+(0.889439582492)*x_ref[1]
        ref[1,4]=(-0.379088893433)*x_ref[0]**o+(0.497309923662)*x_ref[0]+(0.153969715423)*x_ref[1]**o+(-0.386632220307)*x_ref[1]
        ref[2,0]=(-0.186297693534)*x_ref[0]**o+(0.489487382751)*x_ref[0]+(0.644068223202)*x_ref[1]**o+(0.0743289639696)*x_ref[1]
        ref[2,1]=(0.593905543479)*x_ref[0]**o+(-0.966302120537)*x_ref[0]+(-0.861407416257)*x_ref[1]**o+(0.2431998414)*x_ref[1]
        ref[2,2]=(0.683336278471)*x_ref[0]**o+(0.337655287244)*x_ref[0]+(0.876310641583)*x_ref[1]**o+(-0.850438003892)*x_ref[1]
        ref[2,3]=(-0.109203794389)*x_ref[0]**o+(0.245455391347)*x_ref[0]+(-0.30542844788)*x_ref[1]**o+(0.846645969509)*x_ref[1]
        ref[2,4]=(-0.478221810644)*x_ref[0]**o+(0.555082091582)*x_ref[0]+(-0.331626471767)*x_ref[1]**o+(-0.363786378267)*x_ref[1]
        ref[3,0]=(-0.0373544633657)*x_ref[0]**o+(-0.821373667974)*x_ref[0]+(-0.569127067519)*x_ref[1]**o+(0.178159969962)*x_ref[1]
        ref[3,1]=(0.574056120213)*x_ref[0]**o+(-0.557444624837)*x_ref[0]+(-0.858492289239)*x_ref[1]**o+(0.82616627128)*x_ref[1]
        ref[3,2]=(0.0934423420331)*x_ref[0]**o+(-0.64321213739)*x_ref[0]+(0.0832655456405)*x_ref[1]**o+(0.245929781754)*x_ref[1]
        ref[3,3]=(0.0254040950683)*x_ref[0]**o+(0.169079366685)*x_ref[0]+(0.150029664562)*x_ref[1]**o+(0.987260372009)*x_ref[1]
        ref[3,4]=(0.201458676701)*x_ref[0]**o+(0.840874176438)*x_ref[0]+(-0.533157142378)*x_ref[1]**o+(-0.216306366879)*x_ref[1]
      else:
        s[0,0]=(0.872668762115)*x[0]**o+(-0.816027656579)*x[0]+(-0.867841046004)*x[1]**o+(-0.349526762676)*x[1]+(0.938408466875)*x[2]**o+(-0.306416040856)*x[2]
        s[0,1]=(0.909287177866)*x[0]**o+(-0.253902016034)*x[0]+(-0.546477610869)*x[1]**o+(-0.793452692383)*x[1]+(0.776576227123)*x[2]**o+(-0.0967411331785)*x[2]
        s[0,2]=(-0.398898197311)*x[0]**o+(-0.897448112384)*x[0]+(-0.455423203326)*x[1]**o+(-0.367585262313)*x[1]+(-0.0545607718974)*x[2]**o+(-0.282047253613)*x[2]
        s[0,3]=(0.482124916132)*x[0]**o+(-0.631619166977)*x[0]+(-0.0399181373809)*x[1]**o+(0.763797539726)*x[1]+(0.0291052931267)*x[2]**o+(-0.617858834155)*x[2]
        s[0,4]=(-0.0297695890339)*x[0]**o+(0.854962979284)*x[0]+(-0.781784598569)*x[1]**o+(0.299243628895)*x[1]+(-0.310475890278)*x[2]**o+(-0.505378646661)*x[2]
        s[1,0]=(0.484812356262)*x[0]**o+(-0.623958486741)*x[0]+(-0.470006063838)*x[1]**o+(0.660345336288)*x[1]+(-0.878879608367)*x[2]**o+(0.390207006706)*x[2]
        s[1,1]=(0.0351313080262)*x[0]**o+(-0.976289440937)*x[0]+(-0.775189613854)*x[1]**o+(0.765051272995)*x[1]+(-0.675689153966)*x[2]**o+(-0.783405673294)*x[2]
        s[1,2]=(0.662391389122)*x[0]**o+(0.323408772964)*x[0]+(-0.363746897244)*x[1]**o+(0.606811951514)*x[1]+(0.47946488956)*x[2]**o+(-0.966120072388)*x[2]
        s[1,3]=(-0.875339728461)*x[0]**o+(-0.705421436172)*x[0]+(0.291443522131)*x[1]**o+(-0.754961478752)*x[1]+(-0.321825067143)*x[2]**o+(-0.574459688222)*x[2]
        s[1,4]=(-0.255183687846)*x[0]**o+(0.154020081567)*x[0]+(0.84391736219)*x[1]**o+(0.0431273811842)*x[1]+(-0.667796573312)*x[2]**o+(-0.964117112042)*x[2]
        s[2,0]=(-0.530764723315)*x[0]**o+(-0.149048311546)*x[0]+(-0.451189249073)*x[1]**o+(0.999397419444)*x[1]+(0.799956886737)*x[2]**o+(-0.313688076191)*x[2]
        s[2,1]=(-0.749917464595)*x[0]**o+(0.292630460079)*x[0]+(0.506753622803)*x[1]**o+(-0.343689419436)*x[1]+(0.128258231841)*x[2]**o+(0.296077111022)*x[2]
        s[2,2]=(0.394787103006)*x[0]**o+(0.249722955553)*x[0]+(-0.839807889834)*x[1]**o+(-0.667642818246)*x[1]+(0.692437679503)*x[2]**o+(0.0127711080992)*x[2]
        s[2,3]=(-0.0220395920977)*x[0]**o+(-0.720750207444)*x[0]+(0.358116668188)*x[1]**o+(-0.204183323047)*x[1]+(-0.410539982758)*x[2]**o+(-0.671312870493)*x[2]
        s[2,4]=(-0.285147258365)*x[0]**o+(-0.813590616716)*x[0]+(-0.999369996254)*x[1]**o+(-0.348357331572)*x[1]+(-0.714185462982)*x[2]**o+(0.0724943925182)*x[2]
        s[3,0]=(-0.935258847991)*x[0]**o+(-0.78218618139)*x[0]+(0.611019605302)*x[1]**o+(0.518312268136)*x[1]+(0.518366032626)*x[2]**o+(-0.61673250713)*x[2]
        s[3,1]=(0.334978122301)*x[0]**o+(0.144777589204)*x[0]+(-0.654113643417)*x[1]**o+(0.874224972036)*x[1]+(0.988416020574)*x[2]**o+(-0.617165353999)*x[2]
        s[3,2]=(-0.614946557606)*x[0]**o+(-0.77713622969)*x[0]+(-0.0601180324016)*x[1]**o+(-0.493272161621)*x[1]+(-0.118228561563)*x[2]**o+(0.956097356096)*x[2]
        s[3,3]=(0.123381793762)*x[0]**o+(-0.842700232136)*x[0]+(0.142566215799)*x[1]**o+(-0.272131336716)*x[1]+(-0.160411232329)*x[2]**o+(0.344527936371)*x[2]
        s[3,4]=(0.195186658307)*x[0]**o+(-0.774658826133)*x[0]+(0.22329451736)*x[1]**o+(-0.934913350667)*x[1]+(0.883414583429)*x[2]**o+(0.209142336403)*x[2]
        ref[0,0]=(0.872668762115)*x_ref[0]**o+(-0.816027656579)*x_ref[0]+(-0.867841046004)*x_ref[1]**o+(-0.349526762676)*x_ref[1]+(0.938408466875)*x_ref[2]**o+(-0.306416040856)*x_ref[2]
        ref[0,1]=(0.909287177866)*x_ref[0]**o+(-0.253902016034)*x_ref[0]+(-0.546477610869)*x_ref[1]**o+(-0.793452692383)*x_ref[1]+(0.776576227123)*x_ref[2]**o+(-0.0967411331785)*x_ref[2]
        ref[0,2]=(-0.398898197311)*x_ref[0]**o+(-0.897448112384)*x_ref[0]+(-0.455423203326)*x_ref[1]**o+(-0.367585262313)*x_ref[1]+(-0.0545607718974)*x_ref[2]**o+(-0.282047253613)*x_ref[2]
        ref[0,3]=(0.482124916132)*x_ref[0]**o+(-0.631619166977)*x_ref[0]+(-0.0399181373809)*x_ref[1]**o+(0.763797539726)*x_ref[1]+(0.0291052931267)*x_ref[2]**o+(-0.617858834155)*x_ref[2]
        ref[0,4]=(-0.0297695890339)*x_ref[0]**o+(0.854962979284)*x_ref[0]+(-0.781784598569)*x_ref[1]**o+(0.299243628895)*x_ref[1]+(-0.310475890278)*x_ref[2]**o+(-0.505378646661)*x_ref[2]
        ref[1,0]=(0.484812356262)*x_ref[0]**o+(-0.623958486741)*x_ref[0]+(-0.470006063838)*x_ref[1]**o+(0.660345336288)*x_ref[1]+(-0.878879608367)*x_ref[2]**o+(0.390207006706)*x_ref[2]
        ref[1,1]=(0.0351313080262)*x_ref[0]**o+(-0.976289440937)*x_ref[0]+(-0.775189613854)*x_ref[1]**o+(0.765051272995)*x_ref[1]+(-0.675689153966)*x_ref[2]**o+(-0.783405673294)*x_ref[2]
        ref[1,2]=(0.662391389122)*x_ref[0]**o+(0.323408772964)*x_ref[0]+(-0.363746897244)*x_ref[1]**o+(0.606811951514)*x_ref[1]+(0.47946488956)*x_ref[2]**o+(-0.966120072388)*x_ref[2]
        ref[1,3]=(-0.875339728461)*x_ref[0]**o+(-0.705421436172)*x_ref[0]+(0.291443522131)*x_ref[1]**o+(-0.754961478752)*x_ref[1]+(-0.321825067143)*x_ref[2]**o+(-0.574459688222)*x_ref[2]
        ref[1,4]=(-0.255183687846)*x_ref[0]**o+(0.154020081567)*x_ref[0]+(0.84391736219)*x_ref[1]**o+(0.0431273811842)*x_ref[1]+(-0.667796573312)*x_ref[2]**o+(-0.964117112042)*x_ref[2]
        ref[2,0]=(-0.530764723315)*x_ref[0]**o+(-0.149048311546)*x_ref[0]+(-0.451189249073)*x_ref[1]**o+(0.999397419444)*x_ref[1]+(0.799956886737)*x_ref[2]**o+(-0.313688076191)*x_ref[2]
        ref[2,1]=(-0.749917464595)*x_ref[0]**o+(0.292630460079)*x_ref[0]+(0.506753622803)*x_ref[1]**o+(-0.343689419436)*x_ref[1]+(0.128258231841)*x_ref[2]**o+(0.296077111022)*x_ref[2]
        ref[2,2]=(0.394787103006)*x_ref[0]**o+(0.249722955553)*x_ref[0]+(-0.839807889834)*x_ref[1]**o+(-0.667642818246)*x_ref[1]+(0.692437679503)*x_ref[2]**o+(0.0127711080992)*x_ref[2]
        ref[2,3]=(-0.0220395920977)*x_ref[0]**o+(-0.720750207444)*x_ref[0]+(0.358116668188)*x_ref[1]**o+(-0.204183323047)*x_ref[1]+(-0.410539982758)*x_ref[2]**o+(-0.671312870493)*x_ref[2]
        ref[2,4]=(-0.285147258365)*x_ref[0]**o+(-0.813590616716)*x_ref[0]+(-0.999369996254)*x_ref[1]**o+(-0.348357331572)*x_ref[1]+(-0.714185462982)*x_ref[2]**o+(0.0724943925182)*x_ref[2]
        ref[3,0]=(-0.935258847991)*x_ref[0]**o+(-0.78218618139)*x_ref[0]+(0.611019605302)*x_ref[1]**o+(0.518312268136)*x_ref[1]+(0.518366032626)*x_ref[2]**o+(-0.61673250713)*x_ref[2]
        ref[3,1]=(0.334978122301)*x_ref[0]**o+(0.144777589204)*x_ref[0]+(-0.654113643417)*x_ref[1]**o+(0.874224972036)*x_ref[1]+(0.988416020574)*x_ref[2]**o+(-0.617165353999)*x_ref[2]
        ref[3,2]=(-0.614946557606)*x_ref[0]**o+(-0.77713622969)*x_ref[0]+(-0.0601180324016)*x_ref[1]**o+(-0.493272161621)*x_ref[1]+(-0.118228561563)*x_ref[2]**o+(0.956097356096)*x_ref[2]
        ref[3,3]=(0.123381793762)*x_ref[0]**o+(-0.842700232136)*x_ref[0]+(0.142566215799)*x_ref[1]**o+(-0.272131336716)*x_ref[1]+(-0.160411232329)*x_ref[2]**o+(0.344527936371)*x_ref[2]
        ref[3,4]=(0.195186658307)*x_ref[0]**o+(-0.774658826133)*x_ref[0]+(0.22329451736)*x_ref[1]**o+(-0.934913350667)*x_ref[1]+(0.883414583429)*x_ref[2]**o+(0.209142336403)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.156299768554)*x[0]**o+(0.459827039123)*x[0]+(-0.0197265970261)*x[1]**o+(-0.185466218355)*x[1]
        s[0,0,1]=(-0.332652682606)*x[0]**o+(-0.727191024608)*x[0]+(0.855023316467)*x[1]**o+(0.989258272735)*x[1]
        s[0,1,0]=(-0.684749431529)*x[0]**o+(-0.330976098391)*x[0]+(0.951823033761)*x[1]**o+(0.828372926831)*x[1]
        s[0,1,1]=(-0.947934025637)*x[0]**o+(-0.703882235645)*x[0]+(-0.382498726707)*x[1]**o+(-0.416712534592)*x[1]
        s[1,0,0]=(0.268886837209)*x[0]**o+(-0.00424369485341)*x[0]+(-0.901073289396)*x[1]**o+(0.652663823918)*x[1]
        s[1,0,1]=(-0.944532566222)*x[0]**o+(0.605970402437)*x[0]+(-0.872983074127)*x[1]**o+(-0.605342924929)*x[1]
        s[1,1,0]=(0.899833782923)*x[0]**o+(-0.423117683495)*x[0]+(0.360329277459)*x[1]**o+(-0.378555269956)*x[1]
        s[1,1,1]=(0.0804731081801)*x[0]**o+(0.28457153585)*x[0]+(-0.412672511383)*x[1]**o+(-0.446119799977)*x[1]
        s[2,0,0]=(-0.981685713427)*x[0]**o+(0.750030890662)*x[0]+(-0.942777059146)*x[1]**o+(0.736449759398)*x[1]
        s[2,0,1]=(0.586282353308)*x[0]**o+(-0.886945155404)*x[0]+(0.35336036468)*x[1]**o+(0.831440884627)*x[1]
        s[2,1,0]=(0.971383399562)*x[0]**o+(0.921512198603)*x[0]+(-0.983955960592)*x[1]**o+(-0.170180363255)*x[1]
        s[2,1,1]=(-0.956023304113)*x[0]**o+(0.197946744121)*x[0]+(-0.406938561509)*x[1]**o+(-0.50378748984)*x[1]
        s[3,0,0]=(-0.135318256749)*x[0]**o+(-0.845428466148)*x[0]+(0.19729474016)*x[1]**o+(-0.320895148489)*x[1]
        s[3,0,1]=(-0.145937639147)*x[0]**o+(-0.511536148587)*x[0]+(-0.22015558536)*x[1]**o+(0.120116967099)*x[1]
        s[3,1,0]=(-0.311099128945)*x[0]**o+(-0.981649008084)*x[0]+(0.214584333563)*x[1]**o+(0.403963738711)*x[1]
        s[3,1,1]=(-0.0672377429337)*x[0]**o+(0.447752491728)*x[0]+(0.865672486599)*x[1]**o+(0.758623643658)*x[1]
        s[4,0,0]=(0.125909493643)*x[0]**o+(0.77920124826)*x[0]+(-0.491111189228)*x[1]**o+(-0.982624198164)*x[1]
        s[4,0,1]=(0.963994569131)*x[0]**o+(0.590025563951)*x[0]+(-0.432193464931)*x[1]**o+(0.0609843850333)*x[1]
        s[4,1,0]=(-0.810094882894)*x[0]**o+(0.478510251999)*x[0]+(0.909645604317)*x[1]**o+(-0.162920820685)*x[1]
        s[4,1,1]=(-0.849640544361)*x[0]**o+(-0.16981520619)*x[0]+(0.548383377264)*x[1]**o+(0.70666908595)*x[1]
        s[5,0,0]=(0.639591398598)*x[0]**o+(-0.052037663461)*x[0]+(0.80293352341)*x[1]**o+(0.373765944073)*x[1]
        s[5,0,1]=(0.278782801517)*x[0]**o+(0.360704948502)*x[0]+(0.518835994538)*x[1]**o+(-0.748058641453)*x[1]
        s[5,1,0]=(0.376125218461)*x[0]**o+(0.587328226457)*x[0]+(-0.408741695392)*x[1]**o+(-0.922372017858)*x[1]
        s[5,1,1]=(0.921904823293)*x[0]**o+(-0.997993840905)*x[0]+(-0.569469660555)*x[1]**o+(0.246392170853)*x[1]
        ref[0,0,0]=(-0.156299768554)*x_ref[0]**o+(0.459827039123)*x_ref[0]+(-0.0197265970261)*x_ref[1]**o+(-0.185466218355)*x_ref[1]
        ref[0,0,1]=(-0.332652682606)*x_ref[0]**o+(-0.727191024608)*x_ref[0]+(0.855023316467)*x_ref[1]**o+(0.989258272735)*x_ref[1]
        ref[0,1,0]=(-0.684749431529)*x_ref[0]**o+(-0.330976098391)*x_ref[0]+(0.951823033761)*x_ref[1]**o+(0.828372926831)*x_ref[1]
        ref[0,1,1]=(-0.947934025637)*x_ref[0]**o+(-0.703882235645)*x_ref[0]+(-0.382498726707)*x_ref[1]**o+(-0.416712534592)*x_ref[1]
        ref[1,0,0]=(0.268886837209)*x_ref[0]**o+(-0.00424369485341)*x_ref[0]+(-0.901073289396)*x_ref[1]**o+(0.652663823918)*x_ref[1]
        ref[1,0,1]=(-0.944532566222)*x_ref[0]**o+(0.605970402437)*x_ref[0]+(-0.872983074127)*x_ref[1]**o+(-0.605342924929)*x_ref[1]
        ref[1,1,0]=(0.899833782923)*x_ref[0]**o+(-0.423117683495)*x_ref[0]+(0.360329277459)*x_ref[1]**o+(-0.378555269956)*x_ref[1]
        ref[1,1,1]=(0.0804731081801)*x_ref[0]**o+(0.28457153585)*x_ref[0]+(-0.412672511383)*x_ref[1]**o+(-0.446119799977)*x_ref[1]
        ref[2,0,0]=(-0.981685713427)*x_ref[0]**o+(0.750030890662)*x_ref[0]+(-0.942777059146)*x_ref[1]**o+(0.736449759398)*x_ref[1]
        ref[2,0,1]=(0.586282353308)*x_ref[0]**o+(-0.886945155404)*x_ref[0]+(0.35336036468)*x_ref[1]**o+(0.831440884627)*x_ref[1]
        ref[2,1,0]=(0.971383399562)*x_ref[0]**o+(0.921512198603)*x_ref[0]+(-0.983955960592)*x_ref[1]**o+(-0.170180363255)*x_ref[1]
        ref[2,1,1]=(-0.956023304113)*x_ref[0]**o+(0.197946744121)*x_ref[0]+(-0.406938561509)*x_ref[1]**o+(-0.50378748984)*x_ref[1]
        ref[3,0,0]=(-0.135318256749)*x_ref[0]**o+(-0.845428466148)*x_ref[0]+(0.19729474016)*x_ref[1]**o+(-0.320895148489)*x_ref[1]
        ref[3,0,1]=(-0.145937639147)*x_ref[0]**o+(-0.511536148587)*x_ref[0]+(-0.22015558536)*x_ref[1]**o+(0.120116967099)*x_ref[1]
        ref[3,1,0]=(-0.311099128945)*x_ref[0]**o+(-0.981649008084)*x_ref[0]+(0.214584333563)*x_ref[1]**o+(0.403963738711)*x_ref[1]
        ref[3,1,1]=(-0.0672377429337)*x_ref[0]**o+(0.447752491728)*x_ref[0]+(0.865672486599)*x_ref[1]**o+(0.758623643658)*x_ref[1]
        ref[4,0,0]=(0.125909493643)*x_ref[0]**o+(0.77920124826)*x_ref[0]+(-0.491111189228)*x_ref[1]**o+(-0.982624198164)*x_ref[1]
        ref[4,0,1]=(0.963994569131)*x_ref[0]**o+(0.590025563951)*x_ref[0]+(-0.432193464931)*x_ref[1]**o+(0.0609843850333)*x_ref[1]
        ref[4,1,0]=(-0.810094882894)*x_ref[0]**o+(0.478510251999)*x_ref[0]+(0.909645604317)*x_ref[1]**o+(-0.162920820685)*x_ref[1]
        ref[4,1,1]=(-0.849640544361)*x_ref[0]**o+(-0.16981520619)*x_ref[0]+(0.548383377264)*x_ref[1]**o+(0.70666908595)*x_ref[1]
        ref[5,0,0]=(0.639591398598)*x_ref[0]**o+(-0.052037663461)*x_ref[0]+(0.80293352341)*x_ref[1]**o+(0.373765944073)*x_ref[1]
        ref[5,0,1]=(0.278782801517)*x_ref[0]**o+(0.360704948502)*x_ref[0]+(0.518835994538)*x_ref[1]**o+(-0.748058641453)*x_ref[1]
        ref[5,1,0]=(0.376125218461)*x_ref[0]**o+(0.587328226457)*x_ref[0]+(-0.408741695392)*x_ref[1]**o+(-0.922372017858)*x_ref[1]
        ref[5,1,1]=(0.921904823293)*x_ref[0]**o+(-0.997993840905)*x_ref[0]+(-0.569469660555)*x_ref[1]**o+(0.246392170853)*x_ref[1]
      else:
        s[0,0,0]=(-0.215837732064)*x[0]**o+(0.708293902667)*x[0]+(-0.327646650956)*x[1]**o+(0.414511934051)*x[1]+(-0.163886757205)*x[2]**o+(-0.204189031155)*x[2]
        s[0,0,1]=(-0.731839397519)*x[0]**o+(-0.639212394687)*x[0]+(0.795699599563)*x[1]**o+(-0.487257713171)*x[1]+(0.390417080366)*x[2]**o+(0.0996253411809)*x[2]
        s[0,1,0]=(0.81024079229)*x[0]**o+(-0.508591074719)*x[0]+(0.591387161708)*x[1]**o+(-0.106008048523)*x[1]+(-0.865341838628)*x[2]**o+(0.636398034627)*x[2]
        s[0,1,1]=(0.107194327861)*x[0]**o+(-0.843547507026)*x[0]+(-0.797015379948)*x[1]**o+(0.92022128539)*x[1]+(-0.724414329973)*x[2]**o+(0.666266236918)*x[2]
        s[1,0,0]=(0.643940188921)*x[0]**o+(-0.365702296362)*x[0]+(0.808084016481)*x[1]**o+(-0.811845121877)*x[1]+(-0.948680432857)*x[2]**o+(-0.751105519587)*x[2]
        s[1,0,1]=(-0.365910713276)*x[0]**o+(-0.258101370414)*x[0]+(-0.376112129731)*x[1]**o+(0.969279125776)*x[1]+(-0.574916289815)*x[2]**o+(0.192443729433)*x[2]
        s[1,1,0]=(0.246180930634)*x[0]**o+(-0.65550133767)*x[0]+(-0.631929004343)*x[1]**o+(-0.270752141343)*x[1]+(-0.55380695748)*x[2]**o+(-0.0650898973585)*x[2]
        s[1,1,1]=(-0.624153597329)*x[0]**o+(0.471012703457)*x[0]+(-0.421626259377)*x[1]**o+(0.907052626802)*x[1]+(-0.943845344146)*x[2]**o+(0.847485825383)*x[2]
        s[2,0,0]=(-0.114129908207)*x[0]**o+(0.0887282187742)*x[0]+(-0.263100777717)*x[1]**o+(-0.813318704825)*x[1]+(-0.921371769744)*x[2]**o+(0.315384876801)*x[2]
        s[2,0,1]=(0.00919199747403)*x[0]**o+(-0.839076410274)*x[0]+(0.674300968816)*x[1]**o+(-0.266159980117)*x[1]+(-0.754781416929)*x[2]**o+(-0.330466108009)*x[2]
        s[2,1,0]=(0.269339357806)*x[0]**o+(0.978317771215)*x[0]+(0.483559805886)*x[1]**o+(-0.34789105417)*x[1]+(-0.301294669211)*x[2]**o+(-0.67971772971)*x[2]
        s[2,1,1]=(0.384387354744)*x[0]**o+(-0.127357593581)*x[0]+(0.567748162535)*x[1]**o+(0.312028506128)*x[1]+(0.906909215159)*x[2]**o+(-0.157017635377)*x[2]
        s[3,0,0]=(0.364487214036)*x[0]**o+(-0.0694000264348)*x[0]+(0.752067987858)*x[1]**o+(-0.0777040114767)*x[1]+(-0.536354586822)*x[2]**o+(0.846950905582)*x[2]
        s[3,0,1]=(0.215378827608)*x[0]**o+(-0.495485213939)*x[0]+(0.804796373941)*x[1]**o+(-0.948743952727)*x[1]+(0.562922488996)*x[2]**o+(0.611879748265)*x[2]
        s[3,1,0]=(-0.357559056788)*x[0]**o+(-0.966706464642)*x[0]+(-0.279364080486)*x[1]**o+(0.151836363465)*x[1]+(-0.0666607800869)*x[2]**o+(-0.155985265266)*x[2]
        s[3,1,1]=(-0.620991172464)*x[0]**o+(0.838549798787)*x[0]+(-0.69596391819)*x[1]**o+(0.345156777898)*x[1]+(-0.807545685144)*x[2]**o+(-0.343627808568)*x[2]
        s[4,0,0]=(0.752865578568)*x[0]**o+(0.0181463586995)*x[0]+(0.289423717925)*x[1]**o+(0.458330445591)*x[1]+(0.642952438313)*x[2]**o+(-0.302592553525)*x[2]
        s[4,0,1]=(0.38497342932)*x[0]**o+(0.197638554033)*x[0]+(-0.19259192208)*x[1]**o+(0.730648627642)*x[1]+(-0.684340024218)*x[2]**o+(-0.316428219221)*x[2]
        s[4,1,0]=(0.723339041107)*x[0]**o+(0.625053549512)*x[0]+(0.82493053308)*x[1]**o+(-0.222874319966)*x[1]+(0.0283928465938)*x[2]**o+(-0.262980725333)*x[2]
        s[4,1,1]=(0.527134347243)*x[0]**o+(-0.83920533899)*x[0]+(0.623982181406)*x[1]**o+(0.140699355979)*x[1]+(-0.189351111714)*x[2]**o+(0.133204166349)*x[2]
        s[5,0,0]=(0.155705059814)*x[0]**o+(-0.672186066824)*x[0]+(-0.279849819358)*x[1]**o+(-0.95943658469)*x[1]+(-0.750042065446)*x[2]**o+(-0.677006630121)*x[2]
        s[5,0,1]=(0.955099324881)*x[0]**o+(-0.609207748852)*x[0]+(-0.480726753052)*x[1]**o+(-0.110097202271)*x[1]+(-0.293113386852)*x[2]**o+(0.0810841023923)*x[2]
        s[5,1,0]=(-0.0858192742373)*x[0]**o+(0.456041657037)*x[0]+(-0.560905002533)*x[1]**o+(0.211057417206)*x[1]+(-0.296433850136)*x[2]**o+(-0.153349582946)*x[2]
        s[5,1,1]=(-0.914667297428)*x[0]**o+(0.856653497665)*x[0]+(0.994731884351)*x[1]**o+(-0.221838704365)*x[1]+(0.0905489452636)*x[2]**o+(0.937047976959)*x[2]
        ref[0,0,0]=(-0.215837732064)*x_ref[0]**o+(0.708293902667)*x_ref[0]+(-0.327646650956)*x_ref[1]**o+(0.414511934051)*x_ref[1]+(-0.163886757205)*x_ref[2]**o+(-0.204189031155)*x_ref[2]
        ref[0,0,1]=(-0.731839397519)*x_ref[0]**o+(-0.639212394687)*x_ref[0]+(0.795699599563)*x_ref[1]**o+(-0.487257713171)*x_ref[1]+(0.390417080366)*x_ref[2]**o+(0.0996253411809)*x_ref[2]
        ref[0,1,0]=(0.81024079229)*x_ref[0]**o+(-0.508591074719)*x_ref[0]+(0.591387161708)*x_ref[1]**o+(-0.106008048523)*x_ref[1]+(-0.865341838628)*x_ref[2]**o+(0.636398034627)*x_ref[2]
        ref[0,1,1]=(0.107194327861)*x_ref[0]**o+(-0.843547507026)*x_ref[0]+(-0.797015379948)*x_ref[1]**o+(0.92022128539)*x_ref[1]+(-0.724414329973)*x_ref[2]**o+(0.666266236918)*x_ref[2]
        ref[1,0,0]=(0.643940188921)*x_ref[0]**o+(-0.365702296362)*x_ref[0]+(0.808084016481)*x_ref[1]**o+(-0.811845121877)*x_ref[1]+(-0.948680432857)*x_ref[2]**o+(-0.751105519587)*x_ref[2]
        ref[1,0,1]=(-0.365910713276)*x_ref[0]**o+(-0.258101370414)*x_ref[0]+(-0.376112129731)*x_ref[1]**o+(0.969279125776)*x_ref[1]+(-0.574916289815)*x_ref[2]**o+(0.192443729433)*x_ref[2]
        ref[1,1,0]=(0.246180930634)*x_ref[0]**o+(-0.65550133767)*x_ref[0]+(-0.631929004343)*x_ref[1]**o+(-0.270752141343)*x_ref[1]+(-0.55380695748)*x_ref[2]**o+(-0.0650898973585)*x_ref[2]
        ref[1,1,1]=(-0.624153597329)*x_ref[0]**o+(0.471012703457)*x_ref[0]+(-0.421626259377)*x_ref[1]**o+(0.907052626802)*x_ref[1]+(-0.943845344146)*x_ref[2]**o+(0.847485825383)*x_ref[2]
        ref[2,0,0]=(-0.114129908207)*x_ref[0]**o+(0.0887282187742)*x_ref[0]+(-0.263100777717)*x_ref[1]**o+(-0.813318704825)*x_ref[1]+(-0.921371769744)*x_ref[2]**o+(0.315384876801)*x_ref[2]
        ref[2,0,1]=(0.00919199747403)*x_ref[0]**o+(-0.839076410274)*x_ref[0]+(0.674300968816)*x_ref[1]**o+(-0.266159980117)*x_ref[1]+(-0.754781416929)*x_ref[2]**o+(-0.330466108009)*x_ref[2]
        ref[2,1,0]=(0.269339357806)*x_ref[0]**o+(0.978317771215)*x_ref[0]+(0.483559805886)*x_ref[1]**o+(-0.34789105417)*x_ref[1]+(-0.301294669211)*x_ref[2]**o+(-0.67971772971)*x_ref[2]
        ref[2,1,1]=(0.384387354744)*x_ref[0]**o+(-0.127357593581)*x_ref[0]+(0.567748162535)*x_ref[1]**o+(0.312028506128)*x_ref[1]+(0.906909215159)*x_ref[2]**o+(-0.157017635377)*x_ref[2]
        ref[3,0,0]=(0.364487214036)*x_ref[0]**o+(-0.0694000264348)*x_ref[0]+(0.752067987858)*x_ref[1]**o+(-0.0777040114767)*x_ref[1]+(-0.536354586822)*x_ref[2]**o+(0.846950905582)*x_ref[2]
        ref[3,0,1]=(0.215378827608)*x_ref[0]**o+(-0.495485213939)*x_ref[0]+(0.804796373941)*x_ref[1]**o+(-0.948743952727)*x_ref[1]+(0.562922488996)*x_ref[2]**o+(0.611879748265)*x_ref[2]
        ref[3,1,0]=(-0.357559056788)*x_ref[0]**o+(-0.966706464642)*x_ref[0]+(-0.279364080486)*x_ref[1]**o+(0.151836363465)*x_ref[1]+(-0.0666607800869)*x_ref[2]**o+(-0.155985265266)*x_ref[2]
        ref[3,1,1]=(-0.620991172464)*x_ref[0]**o+(0.838549798787)*x_ref[0]+(-0.69596391819)*x_ref[1]**o+(0.345156777898)*x_ref[1]+(-0.807545685144)*x_ref[2]**o+(-0.343627808568)*x_ref[2]
        ref[4,0,0]=(0.752865578568)*x_ref[0]**o+(0.0181463586995)*x_ref[0]+(0.289423717925)*x_ref[1]**o+(0.458330445591)*x_ref[1]+(0.642952438313)*x_ref[2]**o+(-0.302592553525)*x_ref[2]
        ref[4,0,1]=(0.38497342932)*x_ref[0]**o+(0.197638554033)*x_ref[0]+(-0.19259192208)*x_ref[1]**o+(0.730648627642)*x_ref[1]+(-0.684340024218)*x_ref[2]**o+(-0.316428219221)*x_ref[2]
        ref[4,1,0]=(0.723339041107)*x_ref[0]**o+(0.625053549512)*x_ref[0]+(0.82493053308)*x_ref[1]**o+(-0.222874319966)*x_ref[1]+(0.0283928465938)*x_ref[2]**o+(-0.262980725333)*x_ref[2]
        ref[4,1,1]=(0.527134347243)*x_ref[0]**o+(-0.83920533899)*x_ref[0]+(0.623982181406)*x_ref[1]**o+(0.140699355979)*x_ref[1]+(-0.189351111714)*x_ref[2]**o+(0.133204166349)*x_ref[2]
        ref[5,0,0]=(0.155705059814)*x_ref[0]**o+(-0.672186066824)*x_ref[0]+(-0.279849819358)*x_ref[1]**o+(-0.95943658469)*x_ref[1]+(-0.750042065446)*x_ref[2]**o+(-0.677006630121)*x_ref[2]
        ref[5,0,1]=(0.955099324881)*x_ref[0]**o+(-0.609207748852)*x_ref[0]+(-0.480726753052)*x_ref[1]**o+(-0.110097202271)*x_ref[1]+(-0.293113386852)*x_ref[2]**o+(0.0810841023923)*x_ref[2]
        ref[5,1,0]=(-0.0858192742373)*x_ref[0]**o+(0.456041657037)*x_ref[0]+(-0.560905002533)*x_ref[1]**o+(0.211057417206)*x_ref[1]+(-0.296433850136)*x_ref[2]**o+(-0.153349582946)*x_ref[2]
        ref[5,1,1]=(-0.914667297428)*x_ref[0]**o+(0.856653497665)*x_ref[0]+(0.994731884351)*x_ref[1]**o+(-0.221838704365)*x_ref[1]+(0.0905489452636)*x_ref[2]**o+(0.937047976959)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.0777636614651)*x[0]**o+(-0.0667428739907)*x[0]+(-0.977685288599)*x[1]**o+(0.520832422148)*x[1]
        s[0,0,0,1]=(0.0801946653417)*x[0]**o+(0.391444443109)*x[0]+(0.60513760227)*x[1]**o+(-0.202405127189)*x[1]
        s[0,0,1,0]=(-0.107845479947)*x[0]**o+(-0.603283099882)*x[0]+(-0.648914480134)*x[1]**o+(0.134055978332)*x[1]
        s[0,0,1,1]=(-0.307399325623)*x[0]**o+(0.740265076637)*x[0]+(-0.426634593638)*x[1]**o+(0.524782346481)*x[1]
        s[0,0,2,0]=(0.209544366051)*x[0]**o+(0.456746428209)*x[0]+(0.746428003703)*x[1]**o+(0.891861715667)*x[1]
        s[0,0,2,1]=(0.67352776723)*x[0]**o+(0.087047388831)*x[0]+(0.983599793566)*x[1]**o+(-0.711447053356)*x[1]
        s[0,1,0,0]=(-0.155608608254)*x[0]**o+(-0.563919705574)*x[0]+(-0.819109501425)*x[1]**o+(-0.307153392665)*x[1]
        s[0,1,0,1]=(-0.837016003308)*x[0]**o+(0.759786384968)*x[0]+(-0.286876836113)*x[1]**o+(-0.450054346913)*x[1]
        s[0,1,1,0]=(0.317132680973)*x[0]**o+(-0.541058603922)*x[0]+(-0.438996048591)*x[1]**o+(-0.0804910701489)*x[1]
        s[0,1,1,1]=(-0.383941172325)*x[0]**o+(0.14081095151)*x[0]+(-0.964196329254)*x[1]**o+(0.137424835923)*x[1]
        s[0,1,2,0]=(0.78404512518)*x[0]**o+(-0.618654770204)*x[0]+(-0.120723400095)*x[1]**o+(0.337084654664)*x[1]
        s[0,1,2,1]=(0.370706702266)*x[0]**o+(0.362674269383)*x[0]+(0.129330112286)*x[1]**o+(0.27594329828)*x[1]
        s[0,2,0,0]=(-0.0618534907888)*x[0]**o+(-0.36997399231)*x[0]+(-0.701320112334)*x[1]**o+(0.587057599723)*x[1]
        s[0,2,0,1]=(-0.00083395539837)*x[0]**o+(-0.316061641051)*x[0]+(0.888164277705)*x[1]**o+(-0.708426734155)*x[1]
        s[0,2,1,0]=(-0.882600666018)*x[0]**o+(-0.734482083444)*x[0]+(-0.15689139088)*x[1]**o+(0.885520053714)*x[1]
        s[0,2,1,1]=(0.307710272644)*x[0]**o+(-0.343736402048)*x[0]+(0.00525340811259)*x[1]**o+(-0.365110058652)*x[1]
        s[0,2,2,0]=(-0.477696915031)*x[0]**o+(-0.636820200526)*x[0]+(-0.187124878764)*x[1]**o+(0.353033803175)*x[1]
        s[0,2,2,1]=(0.255079299946)*x[0]**o+(-0.132216578075)*x[0]+(0.540401170028)*x[1]**o+(-0.502440213664)*x[1]
        s[0,3,0,0]=(-0.0567446607807)*x[0]**o+(-0.859056478176)*x[0]+(0.258047569261)*x[1]**o+(-0.197460307626)*x[1]
        s[0,3,0,1]=(-0.412568633139)*x[0]**o+(0.143081394877)*x[0]+(-0.958152901527)*x[1]**o+(-0.946724885421)*x[1]
        s[0,3,1,0]=(0.746006182263)*x[0]**o+(0.415697572299)*x[0]+(0.633094060917)*x[1]**o+(-0.598666742083)*x[1]
        s[0,3,1,1]=(0.820559555653)*x[0]**o+(0.0393656665181)*x[0]+(-0.477283875434)*x[1]**o+(-0.629531033295)*x[1]
        s[0,3,2,0]=(-0.79478196151)*x[0]**o+(-0.682204178506)*x[0]+(0.920770217929)*x[1]**o+(-0.51779073656)*x[1]
        s[0,3,2,1]=(0.098810436334)*x[0]**o+(0.13601558071)*x[0]+(-0.900622849009)*x[1]**o+(0.358263179912)*x[1]
        s[0,4,0,0]=(0.636302027774)*x[0]**o+(-0.888743699041)*x[0]+(0.615869973466)*x[1]**o+(0.810532610229)*x[1]
        s[0,4,0,1]=(-0.350888533752)*x[0]**o+(0.430002320082)*x[0]+(0.184767543248)*x[1]**o+(-0.741038087278)*x[1]
        s[0,4,1,0]=(0.966040628072)*x[0]**o+(0.235593098824)*x[0]+(-0.217971115972)*x[1]**o+(0.211888609783)*x[1]
        s[0,4,1,1]=(-0.735728370129)*x[0]**o+(-0.890691876549)*x[0]+(0.0050361984863)*x[1]**o+(0.401300502845)*x[1]
        s[0,4,2,0]=(0.477438928895)*x[0]**o+(-0.0687120914109)*x[0]+(-0.822568493235)*x[1]**o+(0.692426496641)*x[1]
        s[0,4,2,1]=(-0.620965107536)*x[0]**o+(0.589395692553)*x[0]+(-0.330365701829)*x[1]**o+(0.752996781921)*x[1]
        s[1,0,0,0]=(0.983952631887)*x[0]**o+(-0.595120675382)*x[0]+(-0.834106811619)*x[1]**o+(-0.636013378457)*x[1]
        s[1,0,0,1]=(0.126740296939)*x[0]**o+(-0.520119640058)*x[0]+(-0.345481995205)*x[1]**o+(0.0433678023223)*x[1]
        s[1,0,1,0]=(0.565690778246)*x[0]**o+(0.867963746632)*x[0]+(-0.725334931245)*x[1]**o+(-0.620431577357)*x[1]
        s[1,0,1,1]=(0.191774597085)*x[0]**o+(-0.59036914479)*x[0]+(-0.277730803126)*x[1]**o+(0.909240521676)*x[1]
        s[1,0,2,0]=(-0.479427633319)*x[0]**o+(0.758085256512)*x[0]+(0.86321953036)*x[1]**o+(0.789915199597)*x[1]
        s[1,0,2,1]=(0.478767279402)*x[0]**o+(0.061686054009)*x[0]+(0.403654858589)*x[1]**o+(0.413582841972)*x[1]
        s[1,1,0,0]=(0.839949350991)*x[0]**o+(0.083434136022)*x[0]+(-0.237548283201)*x[1]**o+(0.340524677978)*x[1]
        s[1,1,0,1]=(-0.44615314004)*x[0]**o+(-0.0748079090653)*x[0]+(0.669071324293)*x[1]**o+(-0.00926065712951)*x[1]
        s[1,1,1,0]=(0.634663494412)*x[0]**o+(0.414921856585)*x[0]+(0.656468620085)*x[1]**o+(0.154428434525)*x[1]
        s[1,1,1,1]=(0.120264291322)*x[0]**o+(0.274387126898)*x[0]+(-0.914217101263)*x[1]**o+(0.509056318161)*x[1]
        s[1,1,2,0]=(-0.361610733139)*x[0]**o+(-0.419244707352)*x[0]+(-0.0936805656392)*x[1]**o+(0.658845522463)*x[1]
        s[1,1,2,1]=(-0.0152227030107)*x[0]**o+(0.716681107595)*x[0]+(0.339720150512)*x[1]**o+(0.623368924873)*x[1]
        s[1,2,0,0]=(-0.208625207667)*x[0]**o+(0.0892371419356)*x[0]+(-0.357406324393)*x[1]**o+(0.0545982207211)*x[1]
        s[1,2,0,1]=(-0.133416102554)*x[0]**o+(0.536329273122)*x[0]+(-0.600990988318)*x[1]**o+(-0.937230468764)*x[1]
        s[1,2,1,0]=(-0.549763100528)*x[0]**o+(-0.334287316265)*x[0]+(-0.889824442167)*x[1]**o+(0.164963321178)*x[1]
        s[1,2,1,1]=(-0.55705799726)*x[0]**o+(0.125817791326)*x[0]+(0.719976054813)*x[1]**o+(0.718873789349)*x[1]
        s[1,2,2,0]=(-0.73171705212)*x[0]**o+(0.734055655405)*x[0]+(0.202137285455)*x[1]**o+(0.684220988395)*x[1]
        s[1,2,2,1]=(-0.0397765024932)*x[0]**o+(-0.633490619367)*x[0]+(-0.195588405339)*x[1]**o+(-0.957674729264)*x[1]
        s[1,3,0,0]=(0.993464844097)*x[0]**o+(0.234619248397)*x[0]+(-0.51343997173)*x[1]**o+(0.856105886668)*x[1]
        s[1,3,0,1]=(0.0811328827072)*x[0]**o+(0.12838793682)*x[0]+(-0.522572607464)*x[1]**o+(-0.375599526602)*x[1]
        s[1,3,1,0]=(0.826184052821)*x[0]**o+(-0.050690194008)*x[0]+(0.13981816603)*x[1]**o+(0.00952178435275)*x[1]
        s[1,3,1,1]=(0.217029350446)*x[0]**o+(0.714803047883)*x[0]+(-0.957500687556)*x[1]**o+(-0.181988303734)*x[1]
        s[1,3,2,0]=(-0.380490770961)*x[0]**o+(0.279064641297)*x[0]+(0.587953927724)*x[1]**o+(0.0967128120583)*x[1]
        s[1,3,2,1]=(-0.563049792849)*x[0]**o+(0.179929757095)*x[0]+(0.256239332721)*x[1]**o+(0.719366227917)*x[1]
        s[1,4,0,0]=(-0.935374185099)*x[0]**o+(0.686212475195)*x[0]+(-0.696243048462)*x[1]**o+(-0.320962125289)*x[1]
        s[1,4,0,1]=(0.877754242018)*x[0]**o+(-0.131209004599)*x[0]+(-0.665712708561)*x[1]**o+(0.746518141138)*x[1]
        s[1,4,1,0]=(0.734524693133)*x[0]**o+(-0.379930095743)*x[0]+(0.934573403823)*x[1]**o+(-0.00810739306485)*x[1]
        s[1,4,1,1]=(0.0822815241388)*x[0]**o+(0.662617104334)*x[0]+(-0.0794803324896)*x[1]**o+(0.331645391345)*x[1]
        s[1,4,2,0]=(0.232252551171)*x[0]**o+(0.814128846678)*x[0]+(-0.628231676896)*x[1]**o+(-0.702203701604)*x[1]
        s[1,4,2,1]=(-0.700532752151)*x[0]**o+(0.277528305962)*x[0]+(-0.0521465863885)*x[1]**o+(-0.31154867965)*x[1]
        s[2,0,0,0]=(0.472871337554)*x[0]**o+(0.413549495857)*x[0]+(0.311158506504)*x[1]**o+(0.198806166981)*x[1]
        s[2,0,0,1]=(0.0340811960202)*x[0]**o+(-0.00583425318175)*x[0]+(-0.750117294155)*x[1]**o+(-0.905524986631)*x[1]
        s[2,0,1,0]=(0.570790284556)*x[0]**o+(0.673577393533)*x[0]+(-0.325056654336)*x[1]**o+(0.653038510272)*x[1]
        s[2,0,1,1]=(-0.297592197799)*x[0]**o+(0.871440030306)*x[0]+(-0.533637236239)*x[1]**o+(-0.4099903042)*x[1]
        s[2,0,2,0]=(0.293978916599)*x[0]**o+(-0.0182866541826)*x[0]+(0.843245387706)*x[1]**o+(-0.139447965404)*x[1]
        s[2,0,2,1]=(-0.10091027426)*x[0]**o+(-0.100495744095)*x[0]+(-0.0141975035033)*x[1]**o+(-0.375812318845)*x[1]
        s[2,1,0,0]=(-0.830824351182)*x[0]**o+(-0.972181712173)*x[0]+(-0.60270773635)*x[1]**o+(-0.449692597572)*x[1]
        s[2,1,0,1]=(0.580664484955)*x[0]**o+(0.453307511725)*x[0]+(-0.69006079487)*x[1]**o+(-0.282709994647)*x[1]
        s[2,1,1,0]=(0.0350693872862)*x[0]**o+(0.551895773929)*x[0]+(0.946054026898)*x[1]**o+(0.0805073878157)*x[1]
        s[2,1,1,1]=(-0.996166074405)*x[0]**o+(-0.0291388611313)*x[0]+(0.107628876337)*x[1]**o+(0.181623761662)*x[1]
        s[2,1,2,0]=(0.878542357796)*x[0]**o+(0.0263960313825)*x[0]+(-0.470099210068)*x[1]**o+(0.775356436208)*x[1]
        s[2,1,2,1]=(0.771458787313)*x[0]**o+(-0.238120631661)*x[0]+(0.0545258501919)*x[1]**o+(0.178539047763)*x[1]
        s[2,2,0,0]=(0.398263199258)*x[0]**o+(0.858180962314)*x[0]+(-0.00348131786996)*x[1]**o+(-0.958044163072)*x[1]
        s[2,2,0,1]=(0.776760912484)*x[0]**o+(0.364111724681)*x[0]+(0.26458908016)*x[1]**o+(-0.0590754024201)*x[1]
        s[2,2,1,0]=(0.704662231933)*x[0]**o+(-0.954239844678)*x[0]+(-0.212398080843)*x[1]**o+(-0.547628870359)*x[1]
        s[2,2,1,1]=(0.741611255723)*x[0]**o+(0.963847765909)*x[0]+(-0.943663876716)*x[1]**o+(0.745169950034)*x[1]
        s[2,2,2,0]=(-0.818935776346)*x[0]**o+(-0.2967013158)*x[0]+(0.179993824489)*x[1]**o+(0.061135531911)*x[1]
        s[2,2,2,1]=(0.704463632969)*x[0]**o+(0.49682411039)*x[0]+(0.522725375403)*x[1]**o+(-0.0991411582378)*x[1]
        s[2,3,0,0]=(-0.394678165334)*x[0]**o+(-0.490081142088)*x[0]+(-0.437269252729)*x[1]**o+(0.437548658318)*x[1]
        s[2,3,0,1]=(-0.8935046735)*x[0]**o+(-0.0514148814426)*x[0]+(-0.758657504339)*x[1]**o+(-0.91790254494)*x[1]
        s[2,3,1,0]=(0.678673237528)*x[0]**o+(-0.820565721442)*x[0]+(0.758797168017)*x[1]**o+(-0.915258841164)*x[1]
        s[2,3,1,1]=(-0.70927157191)*x[0]**o+(-0.742658350028)*x[0]+(0.111130223627)*x[1]**o+(0.179307074639)*x[1]
        s[2,3,2,0]=(0.228074760628)*x[0]**o+(0.533399652471)*x[0]+(0.560064736493)*x[1]**o+(-0.588589967203)*x[1]
        s[2,3,2,1]=(0.994518467999)*x[0]**o+(-0.201788029147)*x[0]+(-0.983374807613)*x[1]**o+(-0.386854409534)*x[1]
        s[2,4,0,0]=(0.763484049607)*x[0]**o+(-0.262913676303)*x[0]+(-0.498088896053)*x[1]**o+(-0.672931052896)*x[1]
        s[2,4,0,1]=(-0.46331695102)*x[0]**o+(0.91157906649)*x[0]+(-0.647323300295)*x[1]**o+(-0.309269630073)*x[1]
        s[2,4,1,0]=(-0.0526757673514)*x[0]**o+(-0.578131288345)*x[0]+(0.39711966866)*x[1]**o+(0.804245669408)*x[1]
        s[2,4,1,1]=(0.223417129057)*x[0]**o+(-0.923622687767)*x[0]+(0.618075111678)*x[1]**o+(0.381994832856)*x[1]
        s[2,4,2,0]=(0.540391658848)*x[0]**o+(0.531453650116)*x[0]+(-0.146608482269)*x[1]**o+(0.940324004341)*x[1]
        s[2,4,2,1]=(-0.986820975671)*x[0]**o+(0.00121909081031)*x[0]+(-0.697255554926)*x[1]**o+(0.967740580108)*x[1]
        s[3,0,0,0]=(-0.0658669634781)*x[0]**o+(0.70728861958)*x[0]+(0.966340481481)*x[1]**o+(0.986894244456)*x[1]
        s[3,0,0,1]=(0.721157686077)*x[0]**o+(-0.0629283817117)*x[0]+(-0.965971306999)*x[1]**o+(0.291794228985)*x[1]
        s[3,0,1,0]=(-0.360437990219)*x[0]**o+(-0.570216212399)*x[0]+(-0.607512137662)*x[1]**o+(-0.515098731307)*x[1]
        s[3,0,1,1]=(0.642086516897)*x[0]**o+(0.52547404535)*x[0]+(0.016876402075)*x[1]**o+(-0.862152788234)*x[1]
        s[3,0,2,0]=(-0.181658178552)*x[0]**o+(-0.203447330476)*x[0]+(-0.547194819622)*x[1]**o+(-0.0593895310488)*x[1]
        s[3,0,2,1]=(-0.582071301036)*x[0]**o+(-0.557616629544)*x[0]+(-0.60723171731)*x[1]**o+(0.204081805298)*x[1]
        s[3,1,0,0]=(0.272716289788)*x[0]**o+(-0.947135640157)*x[0]+(-0.25901560402)*x[1]**o+(-0.0247786104764)*x[1]
        s[3,1,0,1]=(-0.891714371313)*x[0]**o+(0.755281444278)*x[0]+(0.351999986261)*x[1]**o+(0.661376231103)*x[1]
        s[3,1,1,0]=(0.432312310444)*x[0]**o+(-0.0375106988813)*x[0]+(0.843778132955)*x[1]**o+(-0.787134076357)*x[1]
        s[3,1,1,1]=(-0.57131496233)*x[0]**o+(-0.614200739152)*x[0]+(-0.0817103577524)*x[1]**o+(-0.0809087761527)*x[1]
        s[3,1,2,0]=(0.775899774689)*x[0]**o+(0.650087297096)*x[0]+(-0.363483124762)*x[1]**o+(0.0406785055243)*x[1]
        s[3,1,2,1]=(-0.8228824793)*x[0]**o+(0.0223028194838)*x[0]+(-0.246990198359)*x[1]**o+(-0.447554242783)*x[1]
        s[3,2,0,0]=(0.250699655536)*x[0]**o+(0.425847139574)*x[0]+(0.185272955249)*x[1]**o+(0.52788179093)*x[1]
        s[3,2,0,1]=(-0.6916019182)*x[0]**o+(0.468697041897)*x[0]+(0.0929416146358)*x[1]**o+(0.0632517742213)*x[1]
        s[3,2,1,0]=(-0.0137715023834)*x[0]**o+(-0.876273608156)*x[0]+(0.333020658958)*x[1]**o+(-0.546002404224)*x[1]
        s[3,2,1,1]=(-0.0260174553873)*x[0]**o+(0.668170105508)*x[0]+(-0.515918906011)*x[1]**o+(0.427775146345)*x[1]
        s[3,2,2,0]=(0.451288901693)*x[0]**o+(-0.307365533804)*x[0]+(-0.902054304556)*x[1]**o+(0.561501391872)*x[1]
        s[3,2,2,1]=(-0.67166870387)*x[0]**o+(-0.851949794629)*x[0]+(-0.198839514379)*x[1]**o+(-0.0539885252629)*x[1]
        s[3,3,0,0]=(0.319371411392)*x[0]**o+(0.521157698767)*x[0]+(0.0619379731457)*x[1]**o+(0.85473464537)*x[1]
        s[3,3,0,1]=(0.605988887111)*x[0]**o+(-0.75556314419)*x[0]+(-0.551855748356)*x[1]**o+(0.152400883628)*x[1]
        s[3,3,1,0]=(-0.963972268356)*x[0]**o+(-0.221723218793)*x[0]+(-0.940231983732)*x[1]**o+(0.509102233847)*x[1]
        s[3,3,1,1]=(0.934837503535)*x[0]**o+(-0.762674616346)*x[0]+(0.196730758209)*x[1]**o+(0.93895486771)*x[1]
        s[3,3,2,0]=(0.683137246803)*x[0]**o+(0.638846953858)*x[0]+(0.508007741206)*x[1]**o+(0.625453420423)*x[1]
        s[3,3,2,1]=(-0.0951254112087)*x[0]**o+(0.507174561327)*x[0]+(0.98495014583)*x[1]**o+(0.922324997316)*x[1]
        s[3,4,0,0]=(-0.39889705668)*x[0]**o+(-0.00933499671216)*x[0]+(-0.308567417711)*x[1]**o+(0.0365845447897)*x[1]
        s[3,4,0,1]=(0.699792659761)*x[0]**o+(-0.339911160323)*x[0]+(-0.0380630017731)*x[1]**o+(0.891498223182)*x[1]
        s[3,4,1,0]=(-0.486198638614)*x[0]**o+(0.910921882689)*x[0]+(0.544226196572)*x[1]**o+(0.221206282494)*x[1]
        s[3,4,1,1]=(0.0656938900284)*x[0]**o+(-0.146193696097)*x[0]+(0.580985806286)*x[1]**o+(0.913369274555)*x[1]
        s[3,4,2,0]=(-0.31430372984)*x[0]**o+(0.132101392811)*x[0]+(0.823927100539)*x[1]**o+(-0.0349106467441)*x[1]
        s[3,4,2,1]=(0.454900950832)*x[0]**o+(0.875474839303)*x[0]+(-0.496859632101)*x[1]**o+(0.143908936177)*x[1]
        ref[0,0,0,0]=(-0.0777636614651)*x_ref[0]**o+(-0.0667428739907)*x_ref[0]+(-0.977685288599)*x_ref[1]**o+(0.520832422148)*x_ref[1]
        ref[0,0,0,1]=(0.0801946653417)*x_ref[0]**o+(0.391444443109)*x_ref[0]+(0.60513760227)*x_ref[1]**o+(-0.202405127189)*x_ref[1]
        ref[0,0,1,0]=(-0.107845479947)*x_ref[0]**o+(-0.603283099882)*x_ref[0]+(-0.648914480134)*x_ref[1]**o+(0.134055978332)*x_ref[1]
        ref[0,0,1,1]=(-0.307399325623)*x_ref[0]**o+(0.740265076637)*x_ref[0]+(-0.426634593638)*x_ref[1]**o+(0.524782346481)*x_ref[1]
        ref[0,0,2,0]=(0.209544366051)*x_ref[0]**o+(0.456746428209)*x_ref[0]+(0.746428003703)*x_ref[1]**o+(0.891861715667)*x_ref[1]
        ref[0,0,2,1]=(0.67352776723)*x_ref[0]**o+(0.087047388831)*x_ref[0]+(0.983599793566)*x_ref[1]**o+(-0.711447053356)*x_ref[1]
        ref[0,1,0,0]=(-0.155608608254)*x_ref[0]**o+(-0.563919705574)*x_ref[0]+(-0.819109501425)*x_ref[1]**o+(-0.307153392665)*x_ref[1]
        ref[0,1,0,1]=(-0.837016003308)*x_ref[0]**o+(0.759786384968)*x_ref[0]+(-0.286876836113)*x_ref[1]**o+(-0.450054346913)*x_ref[1]
        ref[0,1,1,0]=(0.317132680973)*x_ref[0]**o+(-0.541058603922)*x_ref[0]+(-0.438996048591)*x_ref[1]**o+(-0.0804910701489)*x_ref[1]
        ref[0,1,1,1]=(-0.383941172325)*x_ref[0]**o+(0.14081095151)*x_ref[0]+(-0.964196329254)*x_ref[1]**o+(0.137424835923)*x_ref[1]
        ref[0,1,2,0]=(0.78404512518)*x_ref[0]**o+(-0.618654770204)*x_ref[0]+(-0.120723400095)*x_ref[1]**o+(0.337084654664)*x_ref[1]
        ref[0,1,2,1]=(0.370706702266)*x_ref[0]**o+(0.362674269383)*x_ref[0]+(0.129330112286)*x_ref[1]**o+(0.27594329828)*x_ref[1]
        ref[0,2,0,0]=(-0.0618534907888)*x_ref[0]**o+(-0.36997399231)*x_ref[0]+(-0.701320112334)*x_ref[1]**o+(0.587057599723)*x_ref[1]
        ref[0,2,0,1]=(-0.00083395539837)*x_ref[0]**o+(-0.316061641051)*x_ref[0]+(0.888164277705)*x_ref[1]**o+(-0.708426734155)*x_ref[1]
        ref[0,2,1,0]=(-0.882600666018)*x_ref[0]**o+(-0.734482083444)*x_ref[0]+(-0.15689139088)*x_ref[1]**o+(0.885520053714)*x_ref[1]
        ref[0,2,1,1]=(0.307710272644)*x_ref[0]**o+(-0.343736402048)*x_ref[0]+(0.00525340811259)*x_ref[1]**o+(-0.365110058652)*x_ref[1]
        ref[0,2,2,0]=(-0.477696915031)*x_ref[0]**o+(-0.636820200526)*x_ref[0]+(-0.187124878764)*x_ref[1]**o+(0.353033803175)*x_ref[1]
        ref[0,2,2,1]=(0.255079299946)*x_ref[0]**o+(-0.132216578075)*x_ref[0]+(0.540401170028)*x_ref[1]**o+(-0.502440213664)*x_ref[1]
        ref[0,3,0,0]=(-0.0567446607807)*x_ref[0]**o+(-0.859056478176)*x_ref[0]+(0.258047569261)*x_ref[1]**o+(-0.197460307626)*x_ref[1]
        ref[0,3,0,1]=(-0.412568633139)*x_ref[0]**o+(0.143081394877)*x_ref[0]+(-0.958152901527)*x_ref[1]**o+(-0.946724885421)*x_ref[1]
        ref[0,3,1,0]=(0.746006182263)*x_ref[0]**o+(0.415697572299)*x_ref[0]+(0.633094060917)*x_ref[1]**o+(-0.598666742083)*x_ref[1]
        ref[0,3,1,1]=(0.820559555653)*x_ref[0]**o+(0.0393656665181)*x_ref[0]+(-0.477283875434)*x_ref[1]**o+(-0.629531033295)*x_ref[1]
        ref[0,3,2,0]=(-0.79478196151)*x_ref[0]**o+(-0.682204178506)*x_ref[0]+(0.920770217929)*x_ref[1]**o+(-0.51779073656)*x_ref[1]
        ref[0,3,2,1]=(0.098810436334)*x_ref[0]**o+(0.13601558071)*x_ref[0]+(-0.900622849009)*x_ref[1]**o+(0.358263179912)*x_ref[1]
        ref[0,4,0,0]=(0.636302027774)*x_ref[0]**o+(-0.888743699041)*x_ref[0]+(0.615869973466)*x_ref[1]**o+(0.810532610229)*x_ref[1]
        ref[0,4,0,1]=(-0.350888533752)*x_ref[0]**o+(0.430002320082)*x_ref[0]+(0.184767543248)*x_ref[1]**o+(-0.741038087278)*x_ref[1]
        ref[0,4,1,0]=(0.966040628072)*x_ref[0]**o+(0.235593098824)*x_ref[0]+(-0.217971115972)*x_ref[1]**o+(0.211888609783)*x_ref[1]
        ref[0,4,1,1]=(-0.735728370129)*x_ref[0]**o+(-0.890691876549)*x_ref[0]+(0.0050361984863)*x_ref[1]**o+(0.401300502845)*x_ref[1]
        ref[0,4,2,0]=(0.477438928895)*x_ref[0]**o+(-0.0687120914109)*x_ref[0]+(-0.822568493235)*x_ref[1]**o+(0.692426496641)*x_ref[1]
        ref[0,4,2,1]=(-0.620965107536)*x_ref[0]**o+(0.589395692553)*x_ref[0]+(-0.330365701829)*x_ref[1]**o+(0.752996781921)*x_ref[1]
        ref[1,0,0,0]=(0.983952631887)*x_ref[0]**o+(-0.595120675382)*x_ref[0]+(-0.834106811619)*x_ref[1]**o+(-0.636013378457)*x_ref[1]
        ref[1,0,0,1]=(0.126740296939)*x_ref[0]**o+(-0.520119640058)*x_ref[0]+(-0.345481995205)*x_ref[1]**o+(0.0433678023223)*x_ref[1]
        ref[1,0,1,0]=(0.565690778246)*x_ref[0]**o+(0.867963746632)*x_ref[0]+(-0.725334931245)*x_ref[1]**o+(-0.620431577357)*x_ref[1]
        ref[1,0,1,1]=(0.191774597085)*x_ref[0]**o+(-0.59036914479)*x_ref[0]+(-0.277730803126)*x_ref[1]**o+(0.909240521676)*x_ref[1]
        ref[1,0,2,0]=(-0.479427633319)*x_ref[0]**o+(0.758085256512)*x_ref[0]+(0.86321953036)*x_ref[1]**o+(0.789915199597)*x_ref[1]
        ref[1,0,2,1]=(0.478767279402)*x_ref[0]**o+(0.061686054009)*x_ref[0]+(0.403654858589)*x_ref[1]**o+(0.413582841972)*x_ref[1]
        ref[1,1,0,0]=(0.839949350991)*x_ref[0]**o+(0.083434136022)*x_ref[0]+(-0.237548283201)*x_ref[1]**o+(0.340524677978)*x_ref[1]
        ref[1,1,0,1]=(-0.44615314004)*x_ref[0]**o+(-0.0748079090653)*x_ref[0]+(0.669071324293)*x_ref[1]**o+(-0.00926065712951)*x_ref[1]
        ref[1,1,1,0]=(0.634663494412)*x_ref[0]**o+(0.414921856585)*x_ref[0]+(0.656468620085)*x_ref[1]**o+(0.154428434525)*x_ref[1]
        ref[1,1,1,1]=(0.120264291322)*x_ref[0]**o+(0.274387126898)*x_ref[0]+(-0.914217101263)*x_ref[1]**o+(0.509056318161)*x_ref[1]
        ref[1,1,2,0]=(-0.361610733139)*x_ref[0]**o+(-0.419244707352)*x_ref[0]+(-0.0936805656392)*x_ref[1]**o+(0.658845522463)*x_ref[1]
        ref[1,1,2,1]=(-0.0152227030107)*x_ref[0]**o+(0.716681107595)*x_ref[0]+(0.339720150512)*x_ref[1]**o+(0.623368924873)*x_ref[1]
        ref[1,2,0,0]=(-0.208625207667)*x_ref[0]**o+(0.0892371419356)*x_ref[0]+(-0.357406324393)*x_ref[1]**o+(0.0545982207211)*x_ref[1]
        ref[1,2,0,1]=(-0.133416102554)*x_ref[0]**o+(0.536329273122)*x_ref[0]+(-0.600990988318)*x_ref[1]**o+(-0.937230468764)*x_ref[1]
        ref[1,2,1,0]=(-0.549763100528)*x_ref[0]**o+(-0.334287316265)*x_ref[0]+(-0.889824442167)*x_ref[1]**o+(0.164963321178)*x_ref[1]
        ref[1,2,1,1]=(-0.55705799726)*x_ref[0]**o+(0.125817791326)*x_ref[0]+(0.719976054813)*x_ref[1]**o+(0.718873789349)*x_ref[1]
        ref[1,2,2,0]=(-0.73171705212)*x_ref[0]**o+(0.734055655405)*x_ref[0]+(0.202137285455)*x_ref[1]**o+(0.684220988395)*x_ref[1]
        ref[1,2,2,1]=(-0.0397765024932)*x_ref[0]**o+(-0.633490619367)*x_ref[0]+(-0.195588405339)*x_ref[1]**o+(-0.957674729264)*x_ref[1]
        ref[1,3,0,0]=(0.993464844097)*x_ref[0]**o+(0.234619248397)*x_ref[0]+(-0.51343997173)*x_ref[1]**o+(0.856105886668)*x_ref[1]
        ref[1,3,0,1]=(0.0811328827072)*x_ref[0]**o+(0.12838793682)*x_ref[0]+(-0.522572607464)*x_ref[1]**o+(-0.375599526602)*x_ref[1]
        ref[1,3,1,0]=(0.826184052821)*x_ref[0]**o+(-0.050690194008)*x_ref[0]+(0.13981816603)*x_ref[1]**o+(0.00952178435275)*x_ref[1]
        ref[1,3,1,1]=(0.217029350446)*x_ref[0]**o+(0.714803047883)*x_ref[0]+(-0.957500687556)*x_ref[1]**o+(-0.181988303734)*x_ref[1]
        ref[1,3,2,0]=(-0.380490770961)*x_ref[0]**o+(0.279064641297)*x_ref[0]+(0.587953927724)*x_ref[1]**o+(0.0967128120583)*x_ref[1]
        ref[1,3,2,1]=(-0.563049792849)*x_ref[0]**o+(0.179929757095)*x_ref[0]+(0.256239332721)*x_ref[1]**o+(0.719366227917)*x_ref[1]
        ref[1,4,0,0]=(-0.935374185099)*x_ref[0]**o+(0.686212475195)*x_ref[0]+(-0.696243048462)*x_ref[1]**o+(-0.320962125289)*x_ref[1]
        ref[1,4,0,1]=(0.877754242018)*x_ref[0]**o+(-0.131209004599)*x_ref[0]+(-0.665712708561)*x_ref[1]**o+(0.746518141138)*x_ref[1]
        ref[1,4,1,0]=(0.734524693133)*x_ref[0]**o+(-0.379930095743)*x_ref[0]+(0.934573403823)*x_ref[1]**o+(-0.00810739306485)*x_ref[1]
        ref[1,4,1,1]=(0.0822815241388)*x_ref[0]**o+(0.662617104334)*x_ref[0]+(-0.0794803324896)*x_ref[1]**o+(0.331645391345)*x_ref[1]
        ref[1,4,2,0]=(0.232252551171)*x_ref[0]**o+(0.814128846678)*x_ref[0]+(-0.628231676896)*x_ref[1]**o+(-0.702203701604)*x_ref[1]
        ref[1,4,2,1]=(-0.700532752151)*x_ref[0]**o+(0.277528305962)*x_ref[0]+(-0.0521465863885)*x_ref[1]**o+(-0.31154867965)*x_ref[1]
        ref[2,0,0,0]=(0.472871337554)*x_ref[0]**o+(0.413549495857)*x_ref[0]+(0.311158506504)*x_ref[1]**o+(0.198806166981)*x_ref[1]
        ref[2,0,0,1]=(0.0340811960202)*x_ref[0]**o+(-0.00583425318175)*x_ref[0]+(-0.750117294155)*x_ref[1]**o+(-0.905524986631)*x_ref[1]
        ref[2,0,1,0]=(0.570790284556)*x_ref[0]**o+(0.673577393533)*x_ref[0]+(-0.325056654336)*x_ref[1]**o+(0.653038510272)*x_ref[1]
        ref[2,0,1,1]=(-0.297592197799)*x_ref[0]**o+(0.871440030306)*x_ref[0]+(-0.533637236239)*x_ref[1]**o+(-0.4099903042)*x_ref[1]
        ref[2,0,2,0]=(0.293978916599)*x_ref[0]**o+(-0.0182866541826)*x_ref[0]+(0.843245387706)*x_ref[1]**o+(-0.139447965404)*x_ref[1]
        ref[2,0,2,1]=(-0.10091027426)*x_ref[0]**o+(-0.100495744095)*x_ref[0]+(-0.0141975035033)*x_ref[1]**o+(-0.375812318845)*x_ref[1]
        ref[2,1,0,0]=(-0.830824351182)*x_ref[0]**o+(-0.972181712173)*x_ref[0]+(-0.60270773635)*x_ref[1]**o+(-0.449692597572)*x_ref[1]
        ref[2,1,0,1]=(0.580664484955)*x_ref[0]**o+(0.453307511725)*x_ref[0]+(-0.69006079487)*x_ref[1]**o+(-0.282709994647)*x_ref[1]
        ref[2,1,1,0]=(0.0350693872862)*x_ref[0]**o+(0.551895773929)*x_ref[0]+(0.946054026898)*x_ref[1]**o+(0.0805073878157)*x_ref[1]
        ref[2,1,1,1]=(-0.996166074405)*x_ref[0]**o+(-0.0291388611313)*x_ref[0]+(0.107628876337)*x_ref[1]**o+(0.181623761662)*x_ref[1]
        ref[2,1,2,0]=(0.878542357796)*x_ref[0]**o+(0.0263960313825)*x_ref[0]+(-0.470099210068)*x_ref[1]**o+(0.775356436208)*x_ref[1]
        ref[2,1,2,1]=(0.771458787313)*x_ref[0]**o+(-0.238120631661)*x_ref[0]+(0.0545258501919)*x_ref[1]**o+(0.178539047763)*x_ref[1]
        ref[2,2,0,0]=(0.398263199258)*x_ref[0]**o+(0.858180962314)*x_ref[0]+(-0.00348131786996)*x_ref[1]**o+(-0.958044163072)*x_ref[1]
        ref[2,2,0,1]=(0.776760912484)*x_ref[0]**o+(0.364111724681)*x_ref[0]+(0.26458908016)*x_ref[1]**o+(-0.0590754024201)*x_ref[1]
        ref[2,2,1,0]=(0.704662231933)*x_ref[0]**o+(-0.954239844678)*x_ref[0]+(-0.212398080843)*x_ref[1]**o+(-0.547628870359)*x_ref[1]
        ref[2,2,1,1]=(0.741611255723)*x_ref[0]**o+(0.963847765909)*x_ref[0]+(-0.943663876716)*x_ref[1]**o+(0.745169950034)*x_ref[1]
        ref[2,2,2,0]=(-0.818935776346)*x_ref[0]**o+(-0.2967013158)*x_ref[0]+(0.179993824489)*x_ref[1]**o+(0.061135531911)*x_ref[1]
        ref[2,2,2,1]=(0.704463632969)*x_ref[0]**o+(0.49682411039)*x_ref[0]+(0.522725375403)*x_ref[1]**o+(-0.0991411582378)*x_ref[1]
        ref[2,3,0,0]=(-0.394678165334)*x_ref[0]**o+(-0.490081142088)*x_ref[0]+(-0.437269252729)*x_ref[1]**o+(0.437548658318)*x_ref[1]
        ref[2,3,0,1]=(-0.8935046735)*x_ref[0]**o+(-0.0514148814426)*x_ref[0]+(-0.758657504339)*x_ref[1]**o+(-0.91790254494)*x_ref[1]
        ref[2,3,1,0]=(0.678673237528)*x_ref[0]**o+(-0.820565721442)*x_ref[0]+(0.758797168017)*x_ref[1]**o+(-0.915258841164)*x_ref[1]
        ref[2,3,1,1]=(-0.70927157191)*x_ref[0]**o+(-0.742658350028)*x_ref[0]+(0.111130223627)*x_ref[1]**o+(0.179307074639)*x_ref[1]
        ref[2,3,2,0]=(0.228074760628)*x_ref[0]**o+(0.533399652471)*x_ref[0]+(0.560064736493)*x_ref[1]**o+(-0.588589967203)*x_ref[1]
        ref[2,3,2,1]=(0.994518467999)*x_ref[0]**o+(-0.201788029147)*x_ref[0]+(-0.983374807613)*x_ref[1]**o+(-0.386854409534)*x_ref[1]
        ref[2,4,0,0]=(0.763484049607)*x_ref[0]**o+(-0.262913676303)*x_ref[0]+(-0.498088896053)*x_ref[1]**o+(-0.672931052896)*x_ref[1]
        ref[2,4,0,1]=(-0.46331695102)*x_ref[0]**o+(0.91157906649)*x_ref[0]+(-0.647323300295)*x_ref[1]**o+(-0.309269630073)*x_ref[1]
        ref[2,4,1,0]=(-0.0526757673514)*x_ref[0]**o+(-0.578131288345)*x_ref[0]+(0.39711966866)*x_ref[1]**o+(0.804245669408)*x_ref[1]
        ref[2,4,1,1]=(0.223417129057)*x_ref[0]**o+(-0.923622687767)*x_ref[0]+(0.618075111678)*x_ref[1]**o+(0.381994832856)*x_ref[1]
        ref[2,4,2,0]=(0.540391658848)*x_ref[0]**o+(0.531453650116)*x_ref[0]+(-0.146608482269)*x_ref[1]**o+(0.940324004341)*x_ref[1]
        ref[2,4,2,1]=(-0.986820975671)*x_ref[0]**o+(0.00121909081031)*x_ref[0]+(-0.697255554926)*x_ref[1]**o+(0.967740580108)*x_ref[1]
        ref[3,0,0,0]=(-0.0658669634781)*x_ref[0]**o+(0.70728861958)*x_ref[0]+(0.966340481481)*x_ref[1]**o+(0.986894244456)*x_ref[1]
        ref[3,0,0,1]=(0.721157686077)*x_ref[0]**o+(-0.0629283817117)*x_ref[0]+(-0.965971306999)*x_ref[1]**o+(0.291794228985)*x_ref[1]
        ref[3,0,1,0]=(-0.360437990219)*x_ref[0]**o+(-0.570216212399)*x_ref[0]+(-0.607512137662)*x_ref[1]**o+(-0.515098731307)*x_ref[1]
        ref[3,0,1,1]=(0.642086516897)*x_ref[0]**o+(0.52547404535)*x_ref[0]+(0.016876402075)*x_ref[1]**o+(-0.862152788234)*x_ref[1]
        ref[3,0,2,0]=(-0.181658178552)*x_ref[0]**o+(-0.203447330476)*x_ref[0]+(-0.547194819622)*x_ref[1]**o+(-0.0593895310488)*x_ref[1]
        ref[3,0,2,1]=(-0.582071301036)*x_ref[0]**o+(-0.557616629544)*x_ref[0]+(-0.60723171731)*x_ref[1]**o+(0.204081805298)*x_ref[1]
        ref[3,1,0,0]=(0.272716289788)*x_ref[0]**o+(-0.947135640157)*x_ref[0]+(-0.25901560402)*x_ref[1]**o+(-0.0247786104764)*x_ref[1]
        ref[3,1,0,1]=(-0.891714371313)*x_ref[0]**o+(0.755281444278)*x_ref[0]+(0.351999986261)*x_ref[1]**o+(0.661376231103)*x_ref[1]
        ref[3,1,1,0]=(0.432312310444)*x_ref[0]**o+(-0.0375106988813)*x_ref[0]+(0.843778132955)*x_ref[1]**o+(-0.787134076357)*x_ref[1]
        ref[3,1,1,1]=(-0.57131496233)*x_ref[0]**o+(-0.614200739152)*x_ref[0]+(-0.0817103577524)*x_ref[1]**o+(-0.0809087761527)*x_ref[1]
        ref[3,1,2,0]=(0.775899774689)*x_ref[0]**o+(0.650087297096)*x_ref[0]+(-0.363483124762)*x_ref[1]**o+(0.0406785055243)*x_ref[1]
        ref[3,1,2,1]=(-0.8228824793)*x_ref[0]**o+(0.0223028194838)*x_ref[0]+(-0.246990198359)*x_ref[1]**o+(-0.447554242783)*x_ref[1]
        ref[3,2,0,0]=(0.250699655536)*x_ref[0]**o+(0.425847139574)*x_ref[0]+(0.185272955249)*x_ref[1]**o+(0.52788179093)*x_ref[1]
        ref[3,2,0,1]=(-0.6916019182)*x_ref[0]**o+(0.468697041897)*x_ref[0]+(0.0929416146358)*x_ref[1]**o+(0.0632517742213)*x_ref[1]
        ref[3,2,1,0]=(-0.0137715023834)*x_ref[0]**o+(-0.876273608156)*x_ref[0]+(0.333020658958)*x_ref[1]**o+(-0.546002404224)*x_ref[1]
        ref[3,2,1,1]=(-0.0260174553873)*x_ref[0]**o+(0.668170105508)*x_ref[0]+(-0.515918906011)*x_ref[1]**o+(0.427775146345)*x_ref[1]
        ref[3,2,2,0]=(0.451288901693)*x_ref[0]**o+(-0.307365533804)*x_ref[0]+(-0.902054304556)*x_ref[1]**o+(0.561501391872)*x_ref[1]
        ref[3,2,2,1]=(-0.67166870387)*x_ref[0]**o+(-0.851949794629)*x_ref[0]+(-0.198839514379)*x_ref[1]**o+(-0.0539885252629)*x_ref[1]
        ref[3,3,0,0]=(0.319371411392)*x_ref[0]**o+(0.521157698767)*x_ref[0]+(0.0619379731457)*x_ref[1]**o+(0.85473464537)*x_ref[1]
        ref[3,3,0,1]=(0.605988887111)*x_ref[0]**o+(-0.75556314419)*x_ref[0]+(-0.551855748356)*x_ref[1]**o+(0.152400883628)*x_ref[1]
        ref[3,3,1,0]=(-0.963972268356)*x_ref[0]**o+(-0.221723218793)*x_ref[0]+(-0.940231983732)*x_ref[1]**o+(0.509102233847)*x_ref[1]
        ref[3,3,1,1]=(0.934837503535)*x_ref[0]**o+(-0.762674616346)*x_ref[0]+(0.196730758209)*x_ref[1]**o+(0.93895486771)*x_ref[1]
        ref[3,3,2,0]=(0.683137246803)*x_ref[0]**o+(0.638846953858)*x_ref[0]+(0.508007741206)*x_ref[1]**o+(0.625453420423)*x_ref[1]
        ref[3,3,2,1]=(-0.0951254112087)*x_ref[0]**o+(0.507174561327)*x_ref[0]+(0.98495014583)*x_ref[1]**o+(0.922324997316)*x_ref[1]
        ref[3,4,0,0]=(-0.39889705668)*x_ref[0]**o+(-0.00933499671216)*x_ref[0]+(-0.308567417711)*x_ref[1]**o+(0.0365845447897)*x_ref[1]
        ref[3,4,0,1]=(0.699792659761)*x_ref[0]**o+(-0.339911160323)*x_ref[0]+(-0.0380630017731)*x_ref[1]**o+(0.891498223182)*x_ref[1]
        ref[3,4,1,0]=(-0.486198638614)*x_ref[0]**o+(0.910921882689)*x_ref[0]+(0.544226196572)*x_ref[1]**o+(0.221206282494)*x_ref[1]
        ref[3,4,1,1]=(0.0656938900284)*x_ref[0]**o+(-0.146193696097)*x_ref[0]+(0.580985806286)*x_ref[1]**o+(0.913369274555)*x_ref[1]
        ref[3,4,2,0]=(-0.31430372984)*x_ref[0]**o+(0.132101392811)*x_ref[0]+(0.823927100539)*x_ref[1]**o+(-0.0349106467441)*x_ref[1]
        ref[3,4,2,1]=(0.454900950832)*x_ref[0]**o+(0.875474839303)*x_ref[0]+(-0.496859632101)*x_ref[1]**o+(0.143908936177)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.893553650353)*x[0]**o+(-0.351842521516)*x[0]+(-0.776239256603)*x[1]**o+(-0.402276687439)*x[1]+(-0.822402864919)*x[2]**o+(0.103529960909)*x[2]
        s[0,0,0,1]=(-0.15411940588)*x[0]**o+(-0.719214069052)*x[0]+(-0.492951656336)*x[1]**o+(0.809209951326)*x[1]+(-0.593114280214)*x[2]**o+(0.647536354987)*x[2]
        s[0,0,1,0]=(-0.4726578657)*x[0]**o+(-0.900529164403)*x[0]+(-0.497658085089)*x[1]**o+(0.441307048821)*x[1]+(0.627699814902)*x[2]**o+(-0.255473726512)*x[2]
        s[0,0,1,1]=(0.937189233751)*x[0]**o+(-0.394633254149)*x[0]+(-0.83447573166)*x[1]**o+(-0.189038884898)*x[1]+(0.852530614411)*x[2]**o+(0.967144561098)*x[2]
        s[0,0,2,0]=(-0.530674910941)*x[0]**o+(-0.349001272718)*x[0]+(0.587518089749)*x[1]**o+(-0.892322978725)*x[1]+(-0.504626322472)*x[2]**o+(-0.867528772077)*x[2]
        s[0,0,2,1]=(-0.850967403091)*x[0]**o+(-0.830253463355)*x[0]+(-0.748829089276)*x[1]**o+(0.171444793758)*x[1]+(-0.793998192991)*x[2]**o+(-0.661765935989)*x[2]
        s[0,1,0,0]=(-0.785128579692)*x[0]**o+(-0.359210459481)*x[0]+(-0.509642957058)*x[1]**o+(0.318074588083)*x[1]+(-0.15398452232)*x[2]**o+(-0.423529526133)*x[2]
        s[0,1,0,1]=(0.198797074339)*x[0]**o+(0.754388829431)*x[0]+(0.884953371408)*x[1]**o+(0.661707862973)*x[1]+(0.663813806214)*x[2]**o+(-0.237609669098)*x[2]
        s[0,1,1,0]=(0.296403271945)*x[0]**o+(0.888911865921)*x[0]+(-0.114728439653)*x[1]**o+(-0.190296446591)*x[1]+(0.699471155167)*x[2]**o+(0.257430171529)*x[2]
        s[0,1,1,1]=(-0.937866711277)*x[0]**o+(0.199715798396)*x[0]+(0.682429346583)*x[1]**o+(0.0300499830473)*x[1]+(0.796837425072)*x[2]**o+(0.618110327057)*x[2]
        s[0,1,2,0]=(0.83267626043)*x[0]**o+(0.389527167021)*x[0]+(0.696622541573)*x[1]**o+(0.323025867562)*x[1]+(-0.335192370542)*x[2]**o+(0.598268793919)*x[2]
        s[0,1,2,1]=(-0.860895701689)*x[0]**o+(0.571893463085)*x[0]+(-0.837208302676)*x[1]**o+(0.582037641186)*x[1]+(0.0204844436813)*x[2]**o+(0.84175443928)*x[2]
        s[0,2,0,0]=(0.995641912188)*x[0]**o+(-0.683043378142)*x[0]+(0.856582569766)*x[1]**o+(0.609700650036)*x[1]+(-0.216498152749)*x[2]**o+(0.404344185816)*x[2]
        s[0,2,0,1]=(-0.0336188475062)*x[0]**o+(0.844294087444)*x[0]+(-0.229218254217)*x[1]**o+(0.115102437743)*x[1]+(0.571093445813)*x[2]**o+(0.266786250904)*x[2]
        s[0,2,1,0]=(-0.480605609044)*x[0]**o+(0.895506339528)*x[0]+(0.497786538492)*x[1]**o+(0.826429641685)*x[1]+(-0.0404930804723)*x[2]**o+(-0.549699397313)*x[2]
        s[0,2,1,1]=(-0.635472022806)*x[0]**o+(0.426611422786)*x[0]+(-0.589809517041)*x[1]**o+(-0.510141409468)*x[1]+(0.861957013957)*x[2]**o+(0.547568522693)*x[2]
        s[0,2,2,0]=(0.194664120443)*x[0]**o+(-0.824830613291)*x[0]+(0.885340662561)*x[1]**o+(0.174047647392)*x[1]+(0.0378551578969)*x[2]**o+(-0.485832036974)*x[2]
        s[0,2,2,1]=(-0.269452123964)*x[0]**o+(-0.0487386347542)*x[0]+(-0.496833515693)*x[1]**o+(0.501015798563)*x[1]+(-0.561879691824)*x[2]**o+(-0.231151995284)*x[2]
        s[0,3,0,0]=(-0.657879229166)*x[0]**o+(0.437034424448)*x[0]+(-0.279440364362)*x[1]**o+(0.306164743472)*x[1]+(-0.976575119833)*x[2]**o+(0.422887219126)*x[2]
        s[0,3,0,1]=(0.463706573378)*x[0]**o+(-0.0560780321685)*x[0]+(0.160348057871)*x[1]**o+(-0.394155833611)*x[1]+(0.354926830678)*x[2]**o+(-0.836986282947)*x[2]
        s[0,3,1,0]=(-0.741940744824)*x[0]**o+(0.185216337632)*x[0]+(0.345258688006)*x[1]**o+(-0.967694578778)*x[1]+(-0.824610904123)*x[2]**o+(-0.735528949759)*x[2]
        s[0,3,1,1]=(-0.0871457875019)*x[0]**o+(0.658300473204)*x[0]+(-0.131866896)*x[1]**o+(0.0651584246177)*x[1]+(0.480033790713)*x[2]**o+(0.240844483345)*x[2]
        s[0,3,2,0]=(-0.921746479348)*x[0]**o+(0.822395759575)*x[0]+(0.0527896299189)*x[1]**o+(0.0119091366531)*x[1]+(-0.659801021057)*x[2]**o+(0.813802745814)*x[2]
        s[0,3,2,1]=(0.668738391756)*x[0]**o+(0.836664681044)*x[0]+(-0.130670296341)*x[1]**o+(0.542822650083)*x[1]+(-0.556597236729)*x[2]**o+(-0.208288038549)*x[2]
        s[0,4,0,0]=(0.086889831216)*x[0]**o+(0.934721202553)*x[0]+(-0.323285065677)*x[1]**o+(0.0162227173407)*x[1]+(0.303014495731)*x[2]**o+(0.569121998148)*x[2]
        s[0,4,0,1]=(-0.498353567584)*x[0]**o+(-0.936339865441)*x[0]+(0.187200383841)*x[1]**o+(0.0572760624614)*x[1]+(-0.176971935188)*x[2]**o+(0.888193107076)*x[2]
        s[0,4,1,0]=(0.748849248567)*x[0]**o+(-0.597794365534)*x[0]+(-0.74698027669)*x[1]**o+(0.461712848089)*x[1]+(0.984587526203)*x[2]**o+(0.210336004682)*x[2]
        s[0,4,1,1]=(0.644641868836)*x[0]**o+(0.569657613081)*x[0]+(0.569132875942)*x[1]**o+(0.229575134126)*x[1]+(0.0821684810315)*x[2]**o+(0.23719873578)*x[2]
        s[0,4,2,0]=(-0.583799917268)*x[0]**o+(0.87257955676)*x[0]+(-0.262198279859)*x[1]**o+(0.638709803441)*x[1]+(0.693720376703)*x[2]**o+(0.0626506206935)*x[2]
        s[0,4,2,1]=(-0.555140469776)*x[0]**o+(-0.699667339844)*x[0]+(-0.175579595413)*x[1]**o+(0.330869902579)*x[1]+(-0.43443643788)*x[2]**o+(0.431247661025)*x[2]
        s[1,0,0,0]=(-0.616105274011)*x[0]**o+(0.766286637713)*x[0]+(-0.852345396201)*x[1]**o+(0.53484561268)*x[1]+(0.568276493883)*x[2]**o+(0.856054828485)*x[2]
        s[1,0,0,1]=(0.908849409393)*x[0]**o+(-0.745153969179)*x[0]+(-0.279959776499)*x[1]**o+(-0.854744604289)*x[1]+(-0.936424853683)*x[2]**o+(0.894518923387)*x[2]
        s[1,0,1,0]=(0.84860289881)*x[0]**o+(-0.914008514988)*x[0]+(-0.262407548228)*x[1]**o+(-0.711226151818)*x[1]+(0.639319422513)*x[2]**o+(-0.27853905681)*x[2]
        s[1,0,1,1]=(0.163323251559)*x[0]**o+(-0.93845875669)*x[0]+(-0.443114805807)*x[1]**o+(-0.533303207459)*x[1]+(0.431839047632)*x[2]**o+(-0.569656544847)*x[2]
        s[1,0,2,0]=(0.949714964792)*x[0]**o+(-0.976181981164)*x[0]+(-0.446526176314)*x[1]**o+(0.383526413445)*x[1]+(-0.182860699476)*x[2]**o+(0.959339975866)*x[2]
        s[1,0,2,1]=(0.867191335472)*x[0]**o+(0.751916463079)*x[0]+(-0.378585476419)*x[1]**o+(-0.500587362177)*x[1]+(-0.27137832483)*x[2]**o+(0.551490270001)*x[2]
        s[1,1,0,0]=(-0.583361795968)*x[0]**o+(0.944694170692)*x[0]+(-0.779315277681)*x[1]**o+(-0.113881659539)*x[1]+(0.931018446727)*x[2]**o+(0.480738722055)*x[2]
        s[1,1,0,1]=(0.229826570157)*x[0]**o+(0.714777306362)*x[0]+(0.442416481157)*x[1]**o+(0.605660985582)*x[1]+(-0.525954520894)*x[2]**o+(0.807486338335)*x[2]
        s[1,1,1,0]=(0.316589515903)*x[0]**o+(-0.686516469152)*x[0]+(0.840774801669)*x[1]**o+(0.447175951045)*x[1]+(-0.716670116885)*x[2]**o+(0.790809180069)*x[2]
        s[1,1,1,1]=(-0.508708038573)*x[0]**o+(-0.934322667237)*x[0]+(0.241647787833)*x[1]**o+(0.538683722428)*x[1]+(0.560593542667)*x[2]**o+(-0.943298870861)*x[2]
        s[1,1,2,0]=(0.21474264864)*x[0]**o+(0.914869403498)*x[0]+(-0.154022728013)*x[1]**o+(0.161351894986)*x[1]+(0.412310479319)*x[2]**o+(-0.664814506861)*x[2]
        s[1,1,2,1]=(-0.143696386498)*x[0]**o+(0.883320802442)*x[0]+(-0.907984528188)*x[1]**o+(-0.386745629926)*x[1]+(0.15492968318)*x[2]**o+(0.21067446716)*x[2]
        s[1,2,0,0]=(0.660203100234)*x[0]**o+(0.284397546215)*x[0]+(0.367929049046)*x[1]**o+(0.344269738617)*x[1]+(-0.511307639941)*x[2]**o+(-0.392647111269)*x[2]
        s[1,2,0,1]=(0.867950390849)*x[0]**o+(0.816012056188)*x[0]+(-0.0418899900381)*x[1]**o+(0.240447024929)*x[1]+(0.717832398567)*x[2]**o+(0.945994010423)*x[2]
        s[1,2,1,0]=(-0.242292281013)*x[0]**o+(0.986187733215)*x[0]+(-0.0762418474615)*x[1]**o+(0.545636818354)*x[1]+(-0.401003911045)*x[2]**o+(-0.0737632171919)*x[2]
        s[1,2,1,1]=(0.233293099685)*x[0]**o+(0.973431444625)*x[0]+(0.721124631798)*x[1]**o+(-0.172302126377)*x[1]+(0.116437603679)*x[2]**o+(0.735419551773)*x[2]
        s[1,2,2,0]=(0.971110632456)*x[0]**o+(0.185002876302)*x[0]+(-0.790484468882)*x[1]**o+(0.616520494387)*x[1]+(-0.818053831626)*x[2]**o+(0.705809191087)*x[2]
        s[1,2,2,1]=(-0.937151426491)*x[0]**o+(0.0344588436018)*x[0]+(0.178600320405)*x[1]**o+(0.774945208314)*x[1]+(0.00671749768773)*x[2]**o+(0.600596280965)*x[2]
        s[1,3,0,0]=(-0.510629068288)*x[0]**o+(-0.9867441798)*x[0]+(0.608512062668)*x[1]**o+(0.0882661466679)*x[1]+(0.517357685392)*x[2]**o+(0.921906359759)*x[2]
        s[1,3,0,1]=(0.857935816338)*x[0]**o+(-0.568613593339)*x[0]+(-0.311233160541)*x[1]**o+(0.456373110337)*x[1]+(0.294859133869)*x[2]**o+(0.972992715921)*x[2]
        s[1,3,1,0]=(-0.906870982033)*x[0]**o+(-0.123891310654)*x[0]+(-0.926024571289)*x[1]**o+(-0.971883326233)*x[1]+(-0.270701126081)*x[2]**o+(-0.314054166211)*x[2]
        s[1,3,1,1]=(0.932643998906)*x[0]**o+(0.591312871747)*x[0]+(-0.297447853121)*x[1]**o+(-0.326446485245)*x[1]+(0.355711757012)*x[2]**o+(0.941580099148)*x[2]
        s[1,3,2,0]=(-0.791232221354)*x[0]**o+(-0.545934762256)*x[0]+(-0.527723158311)*x[1]**o+(0.964298416277)*x[1]+(-0.821314347444)*x[2]**o+(-0.889147503887)*x[2]
        s[1,3,2,1]=(0.468387240601)*x[0]**o+(0.831139810807)*x[0]+(-0.780171451212)*x[1]**o+(-0.154078607613)*x[1]+(0.473061938421)*x[2]**o+(-0.237542463566)*x[2]
        s[1,4,0,0]=(0.23783592874)*x[0]**o+(0.922830638473)*x[0]+(-0.53102227341)*x[1]**o+(0.196635165848)*x[1]+(-0.983681936618)*x[2]**o+(-0.946261418219)*x[2]
        s[1,4,0,1]=(0.0626656945751)*x[0]**o+(0.0915443649303)*x[0]+(-0.306766147008)*x[1]**o+(-0.220472003734)*x[1]+(-0.819326989073)*x[2]**o+(0.584600281341)*x[2]
        s[1,4,1,0]=(-0.144427843844)*x[0]**o+(0.854025358607)*x[0]+(0.694868129931)*x[1]**o+(-0.641915817761)*x[1]+(-0.580940477231)*x[2]**o+(-0.252220617111)*x[2]
        s[1,4,1,1]=(0.632987045152)*x[0]**o+(-0.222995494274)*x[0]+(0.218892976575)*x[1]**o+(-0.870083564081)*x[1]+(-0.315656003785)*x[2]**o+(-0.806577859045)*x[2]
        s[1,4,2,0]=(-0.686677179384)*x[0]**o+(0.975041194178)*x[0]+(0.840117893335)*x[1]**o+(-0.357121725686)*x[1]+(-0.113947327576)*x[2]**o+(0.722741254436)*x[2]
        s[1,4,2,1]=(-0.238773557516)*x[0]**o+(0.589422174847)*x[0]+(-0.39358233128)*x[1]**o+(0.159348401427)*x[1]+(0.887216043103)*x[2]**o+(-0.924304870822)*x[2]
        s[2,0,0,0]=(-0.725901161347)*x[0]**o+(0.144957603906)*x[0]+(0.231954835231)*x[1]**o+(-0.760212283123)*x[1]+(-0.725424437779)*x[2]**o+(-0.19688509792)*x[2]
        s[2,0,0,1]=(0.418995010234)*x[0]**o+(-0.942078773861)*x[0]+(-0.818049256222)*x[1]**o+(-0.594681673779)*x[1]+(-0.63178803463)*x[2]**o+(-0.989664715484)*x[2]
        s[2,0,1,0]=(-0.524535447707)*x[0]**o+(-0.746654195315)*x[0]+(-0.721012736821)*x[1]**o+(-0.0345481441746)*x[1]+(0.0537586563756)*x[2]**o+(-0.445033317006)*x[2]
        s[2,0,1,1]=(-0.411001804179)*x[0]**o+(0.00970860318518)*x[0]+(0.783986695315)*x[1]**o+(-0.270151475512)*x[1]+(-0.652403140687)*x[2]**o+(-0.261945554182)*x[2]
        s[2,0,2,0]=(-0.497529491593)*x[0]**o+(0.00339011573669)*x[0]+(0.696815843509)*x[1]**o+(-0.277295752437)*x[1]+(0.204803051766)*x[2]**o+(-0.631981280563)*x[2]
        s[2,0,2,1]=(0.321760872287)*x[0]**o+(0.54137067299)*x[0]+(-0.971971260117)*x[1]**o+(0.767307092712)*x[1]+(0.471809380038)*x[2]**o+(0.178504669856)*x[2]
        s[2,1,0,0]=(-0.627353418617)*x[0]**o+(-0.610846214602)*x[0]+(-0.610318029871)*x[1]**o+(0.884848156201)*x[1]+(-0.227603047617)*x[2]**o+(0.728596663375)*x[2]
        s[2,1,0,1]=(0.205703168727)*x[0]**o+(-0.869108006912)*x[0]+(0.195195654062)*x[1]**o+(0.586300414612)*x[1]+(0.787224810167)*x[2]**o+(-0.0278209500568)*x[2]
        s[2,1,1,0]=(-0.978822062351)*x[0]**o+(0.929683906062)*x[0]+(-0.865432742145)*x[1]**o+(-0.083101823301)*x[1]+(0.264956210564)*x[2]**o+(0.452513746821)*x[2]
        s[2,1,1,1]=(-0.607905485886)*x[0]**o+(-0.683168972026)*x[0]+(-0.973085006343)*x[1]**o+(0.905969981853)*x[1]+(0.336105649928)*x[2]**o+(0.976694664618)*x[2]
        s[2,1,2,0]=(-0.542965644364)*x[0]**o+(-0.740434309738)*x[0]+(0.756162713574)*x[1]**o+(-0.543216421535)*x[1]+(0.0942001276823)*x[2]**o+(0.0482815123591)*x[2]
        s[2,1,2,1]=(-0.226868852637)*x[0]**o+(0.863234399903)*x[0]+(-0.318469617308)*x[1]**o+(0.0192280454449)*x[1]+(-0.23611990172)*x[2]**o+(0.232713876707)*x[2]
        s[2,2,0,0]=(0.726586748484)*x[0]**o+(-0.711554904751)*x[0]+(-0.206168377552)*x[1]**o+(0.84673411536)*x[1]+(-0.594942018777)*x[2]**o+(-0.199098137058)*x[2]
        s[2,2,0,1]=(0.797788406588)*x[0]**o+(0.326417288103)*x[0]+(0.954847835729)*x[1]**o+(-0.823541118309)*x[1]+(-0.183282866634)*x[2]**o+(-0.73879700622)*x[2]
        s[2,2,1,0]=(-0.542135645496)*x[0]**o+(0.617419480945)*x[0]+(-0.858155129941)*x[1]**o+(0.622909241986)*x[1]+(-0.817913616615)*x[2]**o+(0.551023741548)*x[2]
        s[2,2,1,1]=(-0.033791510266)*x[0]**o+(0.960755607224)*x[0]+(0.992551377675)*x[1]**o+(-0.40568563699)*x[1]+(0.635141614881)*x[2]**o+(-0.28322398176)*x[2]
        s[2,2,2,0]=(0.200301024931)*x[0]**o+(-0.526020099906)*x[0]+(0.296360677546)*x[1]**o+(-0.346356255978)*x[1]+(-0.751246948633)*x[2]**o+(0.936356516645)*x[2]
        s[2,2,2,1]=(0.851662248368)*x[0]**o+(0.306264287809)*x[0]+(0.413913851675)*x[1]**o+(0.796748819208)*x[1]+(0.686640823111)*x[2]**o+(0.270797124789)*x[2]
        s[2,3,0,0]=(0.873612484608)*x[0]**o+(-0.635443314156)*x[0]+(0.0336582752521)*x[1]**o+(0.397952617963)*x[1]+(0.708037547846)*x[2]**o+(-0.835497214327)*x[2]
        s[2,3,0,1]=(0.266263219836)*x[0]**o+(0.718604372519)*x[0]+(-0.706375167711)*x[1]**o+(0.267962253067)*x[1]+(-0.0369901462112)*x[2]**o+(0.541887114998)*x[2]
        s[2,3,1,0]=(-0.0883529797185)*x[0]**o+(-0.695144807666)*x[0]+(-0.332082923487)*x[1]**o+(0.0386905137005)*x[1]+(-0.0402663506352)*x[2]**o+(0.897417224025)*x[2]
        s[2,3,1,1]=(-0.275313948401)*x[0]**o+(0.66334859913)*x[0]+(-0.139887516171)*x[1]**o+(0.874104988552)*x[1]+(-0.474834712531)*x[2]**o+(-0.350478879758)*x[2]
        s[2,3,2,0]=(-0.370570829115)*x[0]**o+(0.448685344991)*x[0]+(-0.663934134305)*x[1]**o+(0.270424636731)*x[1]+(0.257023713378)*x[2]**o+(-0.558834446028)*x[2]
        s[2,3,2,1]=(0.853600423341)*x[0]**o+(-0.34026373254)*x[0]+(-0.880960615323)*x[1]**o+(0.720247030686)*x[1]+(-0.179396842577)*x[2]**o+(-0.581055218136)*x[2]
        s[2,4,0,0]=(0.661293829706)*x[0]**o+(-0.278404566365)*x[0]+(-0.668007328124)*x[1]**o+(-0.0762718064116)*x[1]+(-0.559881233142)*x[2]**o+(0.38863355198)*x[2]
        s[2,4,0,1]=(0.841925292626)*x[0]**o+(-0.907683026635)*x[0]+(0.606133080471)*x[1]**o+(-0.743094760572)*x[1]+(-0.963492675546)*x[2]**o+(0.551853881446)*x[2]
        s[2,4,1,0]=(0.106157048406)*x[0]**o+(-0.839019337739)*x[0]+(-0.646225647224)*x[1]**o+(0.596433542879)*x[1]+(-0.440653099815)*x[2]**o+(0.612390706411)*x[2]
        s[2,4,1,1]=(0.0894808258178)*x[0]**o+(0.601107281081)*x[0]+(0.764980611835)*x[1]**o+(-0.028637689046)*x[1]+(-0.163992620355)*x[2]**o+(0.811740898974)*x[2]
        s[2,4,2,0]=(0.159301482106)*x[0]**o+(-0.2732663929)*x[0]+(0.642848902873)*x[1]**o+(-0.486181762406)*x[1]+(-0.0353954685288)*x[2]**o+(-0.17658875721)*x[2]
        s[2,4,2,1]=(-0.190069716212)*x[0]**o+(0.911197492671)*x[0]+(-0.404646749357)*x[1]**o+(0.496457219159)*x[1]+(0.331503854252)*x[2]**o+(-0.0824993997725)*x[2]
        s[3,0,0,0]=(-0.00960152574795)*x[0]**o+(-0.588053191696)*x[0]+(0.134548796202)*x[1]**o+(-0.734698264954)*x[1]+(0.369880572143)*x[2]**o+(-0.184585884928)*x[2]
        s[3,0,0,1]=(-0.894375930982)*x[0]**o+(-0.757908178588)*x[0]+(-0.862478750821)*x[1]**o+(-0.860344696497)*x[1]+(-0.606693340193)*x[2]**o+(-0.211110504173)*x[2]
        s[3,0,1,0]=(0.29208602662)*x[0]**o+(0.428795740188)*x[0]+(0.730773643572)*x[1]**o+(0.473531891026)*x[1]+(-0.475140560425)*x[2]**o+(0.973125507458)*x[2]
        s[3,0,1,1]=(-0.15741631987)*x[0]**o+(-0.34595691825)*x[0]+(-0.291584586864)*x[1]**o+(0.202892853712)*x[1]+(-0.48541451272)*x[2]**o+(-0.0611059721993)*x[2]
        s[3,0,2,0]=(-0.309937469651)*x[0]**o+(-0.515728964606)*x[0]+(-0.212592556423)*x[1]**o+(-0.419645818415)*x[1]+(-0.825715279525)*x[2]**o+(0.701775440231)*x[2]
        s[3,0,2,1]=(-0.527254609361)*x[0]**o+(-0.921224942687)*x[0]+(-0.219965563549)*x[1]**o+(-0.520137812628)*x[1]+(0.228621345698)*x[2]**o+(0.607942457443)*x[2]
        s[3,1,0,0]=(-0.0191513045468)*x[0]**o+(0.0593277991822)*x[0]+(0.679065928035)*x[1]**o+(-0.824466221695)*x[1]+(-0.0207465360942)*x[2]**o+(-0.462448546311)*x[2]
        s[3,1,0,1]=(-0.0226385405157)*x[0]**o+(-0.990729100489)*x[0]+(0.928646343096)*x[1]**o+(0.877879386627)*x[1]+(-0.774330622719)*x[2]**o+(0.45794763385)*x[2]
        s[3,1,1,0]=(0.991939307411)*x[0]**o+(0.424144367324)*x[0]+(0.287966132208)*x[1]**o+(0.560134521129)*x[1]+(-0.366498917899)*x[2]**o+(-0.5499574933)*x[2]
        s[3,1,1,1]=(0.131575237703)*x[0]**o+(0.350065088754)*x[0]+(0.845775797719)*x[1]**o+(-0.951498449401)*x[1]+(-0.749988925873)*x[2]**o+(0.437681907383)*x[2]
        s[3,1,2,0]=(-0.177690055962)*x[0]**o+(0.400513514288)*x[0]+(0.654580691817)*x[1]**o+(0.781954834622)*x[1]+(0.498508301282)*x[2]**o+(0.489109718057)*x[2]
        s[3,1,2,1]=(0.141266612215)*x[0]**o+(0.578894900597)*x[0]+(-0.0990011451546)*x[1]**o+(0.455191961)*x[1]+(0.446639012332)*x[2]**o+(0.398374036483)*x[2]
        s[3,2,0,0]=(-0.88698978852)*x[0]**o+(0.782997043916)*x[0]+(-0.0203503423342)*x[1]**o+(-0.729317082717)*x[1]+(-0.916473605189)*x[2]**o+(0.706748074855)*x[2]
        s[3,2,0,1]=(0.955314840013)*x[0]**o+(0.784566529939)*x[0]+(-0.607468453224)*x[1]**o+(-0.960125382144)*x[1]+(-0.62579570096)*x[2]**o+(-0.0943302040401)*x[2]
        s[3,2,1,0]=(-0.255287575915)*x[0]**o+(0.418535282029)*x[0]+(0.606559849688)*x[1]**o+(0.940877031124)*x[1]+(0.714131644166)*x[2]**o+(-0.77112172198)*x[2]
        s[3,2,1,1]=(0.566383366127)*x[0]**o+(0.301059560952)*x[0]+(0.181921162673)*x[1]**o+(-0.972651860466)*x[1]+(-0.435658160785)*x[2]**o+(-0.449044377115)*x[2]
        s[3,2,2,0]=(0.730443884106)*x[0]**o+(0.535861749024)*x[0]+(-0.257910068997)*x[1]**o+(-0.879923650189)*x[1]+(0.24309905047)*x[2]**o+(-0.407314031684)*x[2]
        s[3,2,2,1]=(-0.734460242423)*x[0]**o+(0.694756195588)*x[0]+(-0.0110855577415)*x[1]**o+(0.213770377606)*x[1]+(0.404975692507)*x[2]**o+(0.235122345113)*x[2]
        s[3,3,0,0]=(-0.0854800701504)*x[0]**o+(-0.998984290225)*x[0]+(-0.781970276212)*x[1]**o+(0.866788214566)*x[1]+(-0.21747695978)*x[2]**o+(-0.615256484308)*x[2]
        s[3,3,0,1]=(-0.509246205985)*x[0]**o+(-0.870342470261)*x[0]+(0.699615910083)*x[1]**o+(0.81056073388)*x[1]+(0.137830446242)*x[2]**o+(-0.355111214574)*x[2]
        s[3,3,1,0]=(-0.423472624091)*x[0]**o+(0.812256460885)*x[0]+(-0.675497703361)*x[1]**o+(0.296366723996)*x[1]+(-0.218492550353)*x[2]**o+(-0.852823171625)*x[2]
        s[3,3,1,1]=(-0.0718862672517)*x[0]**o+(-0.122792076285)*x[0]+(-0.794191931729)*x[1]**o+(0.569212982974)*x[1]+(-0.646851040773)*x[2]**o+(-0.3073262807)*x[2]
        s[3,3,2,0]=(-0.994778576692)*x[0]**o+(-0.250054159815)*x[0]+(0.726266056262)*x[1]**o+(-0.827929598129)*x[1]+(0.038573450577)*x[2]**o+(-0.402624412259)*x[2]
        s[3,3,2,1]=(-0.158546722118)*x[0]**o+(0.438480217845)*x[0]+(-0.757053187223)*x[1]**o+(0.96669102229)*x[1]+(0.585680018048)*x[2]**o+(0.917795585404)*x[2]
        s[3,4,0,0]=(0.964942257873)*x[0]**o+(0.222693858545)*x[0]+(0.708449965849)*x[1]**o+(-0.786172527336)*x[1]+(0.813715873004)*x[2]**o+(-0.0971222939825)*x[2]
        s[3,4,0,1]=(-0.14236975971)*x[0]**o+(-0.748307677379)*x[0]+(-0.269260913515)*x[1]**o+(-0.315981317137)*x[1]+(-0.375598072402)*x[2]**o+(0.0797466318359)*x[2]
        s[3,4,1,0]=(0.52711029533)*x[0]**o+(0.492047263035)*x[0]+(0.678479511876)*x[1]**o+(0.619219421066)*x[1]+(-0.528779830025)*x[2]**o+(-0.237989972194)*x[2]
        s[3,4,1,1]=(0.711881464696)*x[0]**o+(-0.398127815977)*x[0]+(-0.943946294128)*x[1]**o+(0.584800438848)*x[1]+(-0.683342581136)*x[2]**o+(0.895377849601)*x[2]
        s[3,4,2,0]=(-0.698456333887)*x[0]**o+(-0.36871456039)*x[0]+(0.216620614245)*x[1]**o+(-0.242036817594)*x[1]+(-0.845502631302)*x[2]**o+(-0.705164123503)*x[2]
        s[3,4,2,1]=(-0.000297852657095)*x[0]**o+(-0.847929455729)*x[0]+(0.496739572695)*x[1]**o+(0.574135132795)*x[1]+(0.585537276618)*x[2]**o+(-0.934539532574)*x[2]
        ref[0,0,0,0]=(-0.893553650353)*x_ref[0]**o+(-0.351842521516)*x_ref[0]+(-0.776239256603)*x_ref[1]**o+(-0.402276687439)*x_ref[1]+(-0.822402864919)*x_ref[2]**o+(0.103529960909)*x_ref[2]
        ref[0,0,0,1]=(-0.15411940588)*x_ref[0]**o+(-0.719214069052)*x_ref[0]+(-0.492951656336)*x_ref[1]**o+(0.809209951326)*x_ref[1]+(-0.593114280214)*x_ref[2]**o+(0.647536354987)*x_ref[2]
        ref[0,0,1,0]=(-0.4726578657)*x_ref[0]**o+(-0.900529164403)*x_ref[0]+(-0.497658085089)*x_ref[1]**o+(0.441307048821)*x_ref[1]+(0.627699814902)*x_ref[2]**o+(-0.255473726512)*x_ref[2]
        ref[0,0,1,1]=(0.937189233751)*x_ref[0]**o+(-0.394633254149)*x_ref[0]+(-0.83447573166)*x_ref[1]**o+(-0.189038884898)*x_ref[1]+(0.852530614411)*x_ref[2]**o+(0.967144561098)*x_ref[2]
        ref[0,0,2,0]=(-0.530674910941)*x_ref[0]**o+(-0.349001272718)*x_ref[0]+(0.587518089749)*x_ref[1]**o+(-0.892322978725)*x_ref[1]+(-0.504626322472)*x_ref[2]**o+(-0.867528772077)*x_ref[2]
        ref[0,0,2,1]=(-0.850967403091)*x_ref[0]**o+(-0.830253463355)*x_ref[0]+(-0.748829089276)*x_ref[1]**o+(0.171444793758)*x_ref[1]+(-0.793998192991)*x_ref[2]**o+(-0.661765935989)*x_ref[2]
        ref[0,1,0,0]=(-0.785128579692)*x_ref[0]**o+(-0.359210459481)*x_ref[0]+(-0.509642957058)*x_ref[1]**o+(0.318074588083)*x_ref[1]+(-0.15398452232)*x_ref[2]**o+(-0.423529526133)*x_ref[2]
        ref[0,1,0,1]=(0.198797074339)*x_ref[0]**o+(0.754388829431)*x_ref[0]+(0.884953371408)*x_ref[1]**o+(0.661707862973)*x_ref[1]+(0.663813806214)*x_ref[2]**o+(-0.237609669098)*x_ref[2]
        ref[0,1,1,0]=(0.296403271945)*x_ref[0]**o+(0.888911865921)*x_ref[0]+(-0.114728439653)*x_ref[1]**o+(-0.190296446591)*x_ref[1]+(0.699471155167)*x_ref[2]**o+(0.257430171529)*x_ref[2]
        ref[0,1,1,1]=(-0.937866711277)*x_ref[0]**o+(0.199715798396)*x_ref[0]+(0.682429346583)*x_ref[1]**o+(0.0300499830473)*x_ref[1]+(0.796837425072)*x_ref[2]**o+(0.618110327057)*x_ref[2]
        ref[0,1,2,0]=(0.83267626043)*x_ref[0]**o+(0.389527167021)*x_ref[0]+(0.696622541573)*x_ref[1]**o+(0.323025867562)*x_ref[1]+(-0.335192370542)*x_ref[2]**o+(0.598268793919)*x_ref[2]
        ref[0,1,2,1]=(-0.860895701689)*x_ref[0]**o+(0.571893463085)*x_ref[0]+(-0.837208302676)*x_ref[1]**o+(0.582037641186)*x_ref[1]+(0.0204844436813)*x_ref[2]**o+(0.84175443928)*x_ref[2]
        ref[0,2,0,0]=(0.995641912188)*x_ref[0]**o+(-0.683043378142)*x_ref[0]+(0.856582569766)*x_ref[1]**o+(0.609700650036)*x_ref[1]+(-0.216498152749)*x_ref[2]**o+(0.404344185816)*x_ref[2]
        ref[0,2,0,1]=(-0.0336188475062)*x_ref[0]**o+(0.844294087444)*x_ref[0]+(-0.229218254217)*x_ref[1]**o+(0.115102437743)*x_ref[1]+(0.571093445813)*x_ref[2]**o+(0.266786250904)*x_ref[2]
        ref[0,2,1,0]=(-0.480605609044)*x_ref[0]**o+(0.895506339528)*x_ref[0]+(0.497786538492)*x_ref[1]**o+(0.826429641685)*x_ref[1]+(-0.0404930804723)*x_ref[2]**o+(-0.549699397313)*x_ref[2]
        ref[0,2,1,1]=(-0.635472022806)*x_ref[0]**o+(0.426611422786)*x_ref[0]+(-0.589809517041)*x_ref[1]**o+(-0.510141409468)*x_ref[1]+(0.861957013957)*x_ref[2]**o+(0.547568522693)*x_ref[2]
        ref[0,2,2,0]=(0.194664120443)*x_ref[0]**o+(-0.824830613291)*x_ref[0]+(0.885340662561)*x_ref[1]**o+(0.174047647392)*x_ref[1]+(0.0378551578969)*x_ref[2]**o+(-0.485832036974)*x_ref[2]
        ref[0,2,2,1]=(-0.269452123964)*x_ref[0]**o+(-0.0487386347542)*x_ref[0]+(-0.496833515693)*x_ref[1]**o+(0.501015798563)*x_ref[1]+(-0.561879691824)*x_ref[2]**o+(-0.231151995284)*x_ref[2]
        ref[0,3,0,0]=(-0.657879229166)*x_ref[0]**o+(0.437034424448)*x_ref[0]+(-0.279440364362)*x_ref[1]**o+(0.306164743472)*x_ref[1]+(-0.976575119833)*x_ref[2]**o+(0.422887219126)*x_ref[2]
        ref[0,3,0,1]=(0.463706573378)*x_ref[0]**o+(-0.0560780321685)*x_ref[0]+(0.160348057871)*x_ref[1]**o+(-0.394155833611)*x_ref[1]+(0.354926830678)*x_ref[2]**o+(-0.836986282947)*x_ref[2]
        ref[0,3,1,0]=(-0.741940744824)*x_ref[0]**o+(0.185216337632)*x_ref[0]+(0.345258688006)*x_ref[1]**o+(-0.967694578778)*x_ref[1]+(-0.824610904123)*x_ref[2]**o+(-0.735528949759)*x_ref[2]
        ref[0,3,1,1]=(-0.0871457875019)*x_ref[0]**o+(0.658300473204)*x_ref[0]+(-0.131866896)*x_ref[1]**o+(0.0651584246177)*x_ref[1]+(0.480033790713)*x_ref[2]**o+(0.240844483345)*x_ref[2]
        ref[0,3,2,0]=(-0.921746479348)*x_ref[0]**o+(0.822395759575)*x_ref[0]+(0.0527896299189)*x_ref[1]**o+(0.0119091366531)*x_ref[1]+(-0.659801021057)*x_ref[2]**o+(0.813802745814)*x_ref[2]
        ref[0,3,2,1]=(0.668738391756)*x_ref[0]**o+(0.836664681044)*x_ref[0]+(-0.130670296341)*x_ref[1]**o+(0.542822650083)*x_ref[1]+(-0.556597236729)*x_ref[2]**o+(-0.208288038549)*x_ref[2]
        ref[0,4,0,0]=(0.086889831216)*x_ref[0]**o+(0.934721202553)*x_ref[0]+(-0.323285065677)*x_ref[1]**o+(0.0162227173407)*x_ref[1]+(0.303014495731)*x_ref[2]**o+(0.569121998148)*x_ref[2]
        ref[0,4,0,1]=(-0.498353567584)*x_ref[0]**o+(-0.936339865441)*x_ref[0]+(0.187200383841)*x_ref[1]**o+(0.0572760624614)*x_ref[1]+(-0.176971935188)*x_ref[2]**o+(0.888193107076)*x_ref[2]
        ref[0,4,1,0]=(0.748849248567)*x_ref[0]**o+(-0.597794365534)*x_ref[0]+(-0.74698027669)*x_ref[1]**o+(0.461712848089)*x_ref[1]+(0.984587526203)*x_ref[2]**o+(0.210336004682)*x_ref[2]
        ref[0,4,1,1]=(0.644641868836)*x_ref[0]**o+(0.569657613081)*x_ref[0]+(0.569132875942)*x_ref[1]**o+(0.229575134126)*x_ref[1]+(0.0821684810315)*x_ref[2]**o+(0.23719873578)*x_ref[2]
        ref[0,4,2,0]=(-0.583799917268)*x_ref[0]**o+(0.87257955676)*x_ref[0]+(-0.262198279859)*x_ref[1]**o+(0.638709803441)*x_ref[1]+(0.693720376703)*x_ref[2]**o+(0.0626506206935)*x_ref[2]
        ref[0,4,2,1]=(-0.555140469776)*x_ref[0]**o+(-0.699667339844)*x_ref[0]+(-0.175579595413)*x_ref[1]**o+(0.330869902579)*x_ref[1]+(-0.43443643788)*x_ref[2]**o+(0.431247661025)*x_ref[2]
        ref[1,0,0,0]=(-0.616105274011)*x_ref[0]**o+(0.766286637713)*x_ref[0]+(-0.852345396201)*x_ref[1]**o+(0.53484561268)*x_ref[1]+(0.568276493883)*x_ref[2]**o+(0.856054828485)*x_ref[2]
        ref[1,0,0,1]=(0.908849409393)*x_ref[0]**o+(-0.745153969179)*x_ref[0]+(-0.279959776499)*x_ref[1]**o+(-0.854744604289)*x_ref[1]+(-0.936424853683)*x_ref[2]**o+(0.894518923387)*x_ref[2]
        ref[1,0,1,0]=(0.84860289881)*x_ref[0]**o+(-0.914008514988)*x_ref[0]+(-0.262407548228)*x_ref[1]**o+(-0.711226151818)*x_ref[1]+(0.639319422513)*x_ref[2]**o+(-0.27853905681)*x_ref[2]
        ref[1,0,1,1]=(0.163323251559)*x_ref[0]**o+(-0.93845875669)*x_ref[0]+(-0.443114805807)*x_ref[1]**o+(-0.533303207459)*x_ref[1]+(0.431839047632)*x_ref[2]**o+(-0.569656544847)*x_ref[2]
        ref[1,0,2,0]=(0.949714964792)*x_ref[0]**o+(-0.976181981164)*x_ref[0]+(-0.446526176314)*x_ref[1]**o+(0.383526413445)*x_ref[1]+(-0.182860699476)*x_ref[2]**o+(0.959339975866)*x_ref[2]
        ref[1,0,2,1]=(0.867191335472)*x_ref[0]**o+(0.751916463079)*x_ref[0]+(-0.378585476419)*x_ref[1]**o+(-0.500587362177)*x_ref[1]+(-0.27137832483)*x_ref[2]**o+(0.551490270001)*x_ref[2]
        ref[1,1,0,0]=(-0.583361795968)*x_ref[0]**o+(0.944694170692)*x_ref[0]+(-0.779315277681)*x_ref[1]**o+(-0.113881659539)*x_ref[1]+(0.931018446727)*x_ref[2]**o+(0.480738722055)*x_ref[2]
        ref[1,1,0,1]=(0.229826570157)*x_ref[0]**o+(0.714777306362)*x_ref[0]+(0.442416481157)*x_ref[1]**o+(0.605660985582)*x_ref[1]+(-0.525954520894)*x_ref[2]**o+(0.807486338335)*x_ref[2]
        ref[1,1,1,0]=(0.316589515903)*x_ref[0]**o+(-0.686516469152)*x_ref[0]+(0.840774801669)*x_ref[1]**o+(0.447175951045)*x_ref[1]+(-0.716670116885)*x_ref[2]**o+(0.790809180069)*x_ref[2]
        ref[1,1,1,1]=(-0.508708038573)*x_ref[0]**o+(-0.934322667237)*x_ref[0]+(0.241647787833)*x_ref[1]**o+(0.538683722428)*x_ref[1]+(0.560593542667)*x_ref[2]**o+(-0.943298870861)*x_ref[2]
        ref[1,1,2,0]=(0.21474264864)*x_ref[0]**o+(0.914869403498)*x_ref[0]+(-0.154022728013)*x_ref[1]**o+(0.161351894986)*x_ref[1]+(0.412310479319)*x_ref[2]**o+(-0.664814506861)*x_ref[2]
        ref[1,1,2,1]=(-0.143696386498)*x_ref[0]**o+(0.883320802442)*x_ref[0]+(-0.907984528188)*x_ref[1]**o+(-0.386745629926)*x_ref[1]+(0.15492968318)*x_ref[2]**o+(0.21067446716)*x_ref[2]
        ref[1,2,0,0]=(0.660203100234)*x_ref[0]**o+(0.284397546215)*x_ref[0]+(0.367929049046)*x_ref[1]**o+(0.344269738617)*x_ref[1]+(-0.511307639941)*x_ref[2]**o+(-0.392647111269)*x_ref[2]
        ref[1,2,0,1]=(0.867950390849)*x_ref[0]**o+(0.816012056188)*x_ref[0]+(-0.0418899900381)*x_ref[1]**o+(0.240447024929)*x_ref[1]+(0.717832398567)*x_ref[2]**o+(0.945994010423)*x_ref[2]
        ref[1,2,1,0]=(-0.242292281013)*x_ref[0]**o+(0.986187733215)*x_ref[0]+(-0.0762418474615)*x_ref[1]**o+(0.545636818354)*x_ref[1]+(-0.401003911045)*x_ref[2]**o+(-0.0737632171919)*x_ref[2]
        ref[1,2,1,1]=(0.233293099685)*x_ref[0]**o+(0.973431444625)*x_ref[0]+(0.721124631798)*x_ref[1]**o+(-0.172302126377)*x_ref[1]+(0.116437603679)*x_ref[2]**o+(0.735419551773)*x_ref[2]
        ref[1,2,2,0]=(0.971110632456)*x_ref[0]**o+(0.185002876302)*x_ref[0]+(-0.790484468882)*x_ref[1]**o+(0.616520494387)*x_ref[1]+(-0.818053831626)*x_ref[2]**o+(0.705809191087)*x_ref[2]
        ref[1,2,2,1]=(-0.937151426491)*x_ref[0]**o+(0.0344588436018)*x_ref[0]+(0.178600320405)*x_ref[1]**o+(0.774945208314)*x_ref[1]+(0.00671749768773)*x_ref[2]**o+(0.600596280965)*x_ref[2]
        ref[1,3,0,0]=(-0.510629068288)*x_ref[0]**o+(-0.9867441798)*x_ref[0]+(0.608512062668)*x_ref[1]**o+(0.0882661466679)*x_ref[1]+(0.517357685392)*x_ref[2]**o+(0.921906359759)*x_ref[2]
        ref[1,3,0,1]=(0.857935816338)*x_ref[0]**o+(-0.568613593339)*x_ref[0]+(-0.311233160541)*x_ref[1]**o+(0.456373110337)*x_ref[1]+(0.294859133869)*x_ref[2]**o+(0.972992715921)*x_ref[2]
        ref[1,3,1,0]=(-0.906870982033)*x_ref[0]**o+(-0.123891310654)*x_ref[0]+(-0.926024571289)*x_ref[1]**o+(-0.971883326233)*x_ref[1]+(-0.270701126081)*x_ref[2]**o+(-0.314054166211)*x_ref[2]
        ref[1,3,1,1]=(0.932643998906)*x_ref[0]**o+(0.591312871747)*x_ref[0]+(-0.297447853121)*x_ref[1]**o+(-0.326446485245)*x_ref[1]+(0.355711757012)*x_ref[2]**o+(0.941580099148)*x_ref[2]
        ref[1,3,2,0]=(-0.791232221354)*x_ref[0]**o+(-0.545934762256)*x_ref[0]+(-0.527723158311)*x_ref[1]**o+(0.964298416277)*x_ref[1]+(-0.821314347444)*x_ref[2]**o+(-0.889147503887)*x_ref[2]
        ref[1,3,2,1]=(0.468387240601)*x_ref[0]**o+(0.831139810807)*x_ref[0]+(-0.780171451212)*x_ref[1]**o+(-0.154078607613)*x_ref[1]+(0.473061938421)*x_ref[2]**o+(-0.237542463566)*x_ref[2]
        ref[1,4,0,0]=(0.23783592874)*x_ref[0]**o+(0.922830638473)*x_ref[0]+(-0.53102227341)*x_ref[1]**o+(0.196635165848)*x_ref[1]+(-0.983681936618)*x_ref[2]**o+(-0.946261418219)*x_ref[2]
        ref[1,4,0,1]=(0.0626656945751)*x_ref[0]**o+(0.0915443649303)*x_ref[0]+(-0.306766147008)*x_ref[1]**o+(-0.220472003734)*x_ref[1]+(-0.819326989073)*x_ref[2]**o+(0.584600281341)*x_ref[2]
        ref[1,4,1,0]=(-0.144427843844)*x_ref[0]**o+(0.854025358607)*x_ref[0]+(0.694868129931)*x_ref[1]**o+(-0.641915817761)*x_ref[1]+(-0.580940477231)*x_ref[2]**o+(-0.252220617111)*x_ref[2]
        ref[1,4,1,1]=(0.632987045152)*x_ref[0]**o+(-0.222995494274)*x_ref[0]+(0.218892976575)*x_ref[1]**o+(-0.870083564081)*x_ref[1]+(-0.315656003785)*x_ref[2]**o+(-0.806577859045)*x_ref[2]
        ref[1,4,2,0]=(-0.686677179384)*x_ref[0]**o+(0.975041194178)*x_ref[0]+(0.840117893335)*x_ref[1]**o+(-0.357121725686)*x_ref[1]+(-0.113947327576)*x_ref[2]**o+(0.722741254436)*x_ref[2]
        ref[1,4,2,1]=(-0.238773557516)*x_ref[0]**o+(0.589422174847)*x_ref[0]+(-0.39358233128)*x_ref[1]**o+(0.159348401427)*x_ref[1]+(0.887216043103)*x_ref[2]**o+(-0.924304870822)*x_ref[2]
        ref[2,0,0,0]=(-0.725901161347)*x_ref[0]**o+(0.144957603906)*x_ref[0]+(0.231954835231)*x_ref[1]**o+(-0.760212283123)*x_ref[1]+(-0.725424437779)*x_ref[2]**o+(-0.19688509792)*x_ref[2]
        ref[2,0,0,1]=(0.418995010234)*x_ref[0]**o+(-0.942078773861)*x_ref[0]+(-0.818049256222)*x_ref[1]**o+(-0.594681673779)*x_ref[1]+(-0.63178803463)*x_ref[2]**o+(-0.989664715484)*x_ref[2]
        ref[2,0,1,0]=(-0.524535447707)*x_ref[0]**o+(-0.746654195315)*x_ref[0]+(-0.721012736821)*x_ref[1]**o+(-0.0345481441746)*x_ref[1]+(0.0537586563756)*x_ref[2]**o+(-0.445033317006)*x_ref[2]
        ref[2,0,1,1]=(-0.411001804179)*x_ref[0]**o+(0.00970860318518)*x_ref[0]+(0.783986695315)*x_ref[1]**o+(-0.270151475512)*x_ref[1]+(-0.652403140687)*x_ref[2]**o+(-0.261945554182)*x_ref[2]
        ref[2,0,2,0]=(-0.497529491593)*x_ref[0]**o+(0.00339011573669)*x_ref[0]+(0.696815843509)*x_ref[1]**o+(-0.277295752437)*x_ref[1]+(0.204803051766)*x_ref[2]**o+(-0.631981280563)*x_ref[2]
        ref[2,0,2,1]=(0.321760872287)*x_ref[0]**o+(0.54137067299)*x_ref[0]+(-0.971971260117)*x_ref[1]**o+(0.767307092712)*x_ref[1]+(0.471809380038)*x_ref[2]**o+(0.178504669856)*x_ref[2]
        ref[2,1,0,0]=(-0.627353418617)*x_ref[0]**o+(-0.610846214602)*x_ref[0]+(-0.610318029871)*x_ref[1]**o+(0.884848156201)*x_ref[1]+(-0.227603047617)*x_ref[2]**o+(0.728596663375)*x_ref[2]
        ref[2,1,0,1]=(0.205703168727)*x_ref[0]**o+(-0.869108006912)*x_ref[0]+(0.195195654062)*x_ref[1]**o+(0.586300414612)*x_ref[1]+(0.787224810167)*x_ref[2]**o+(-0.0278209500568)*x_ref[2]
        ref[2,1,1,0]=(-0.978822062351)*x_ref[0]**o+(0.929683906062)*x_ref[0]+(-0.865432742145)*x_ref[1]**o+(-0.083101823301)*x_ref[1]+(0.264956210564)*x_ref[2]**o+(0.452513746821)*x_ref[2]
        ref[2,1,1,1]=(-0.607905485886)*x_ref[0]**o+(-0.683168972026)*x_ref[0]+(-0.973085006343)*x_ref[1]**o+(0.905969981853)*x_ref[1]+(0.336105649928)*x_ref[2]**o+(0.976694664618)*x_ref[2]
        ref[2,1,2,0]=(-0.542965644364)*x_ref[0]**o+(-0.740434309738)*x_ref[0]+(0.756162713574)*x_ref[1]**o+(-0.543216421535)*x_ref[1]+(0.0942001276823)*x_ref[2]**o+(0.0482815123591)*x_ref[2]
        ref[2,1,2,1]=(-0.226868852637)*x_ref[0]**o+(0.863234399903)*x_ref[0]+(-0.318469617308)*x_ref[1]**o+(0.0192280454449)*x_ref[1]+(-0.23611990172)*x_ref[2]**o+(0.232713876707)*x_ref[2]
        ref[2,2,0,0]=(0.726586748484)*x_ref[0]**o+(-0.711554904751)*x_ref[0]+(-0.206168377552)*x_ref[1]**o+(0.84673411536)*x_ref[1]+(-0.594942018777)*x_ref[2]**o+(-0.199098137058)*x_ref[2]
        ref[2,2,0,1]=(0.797788406588)*x_ref[0]**o+(0.326417288103)*x_ref[0]+(0.954847835729)*x_ref[1]**o+(-0.823541118309)*x_ref[1]+(-0.183282866634)*x_ref[2]**o+(-0.73879700622)*x_ref[2]
        ref[2,2,1,0]=(-0.542135645496)*x_ref[0]**o+(0.617419480945)*x_ref[0]+(-0.858155129941)*x_ref[1]**o+(0.622909241986)*x_ref[1]+(-0.817913616615)*x_ref[2]**o+(0.551023741548)*x_ref[2]
        ref[2,2,1,1]=(-0.033791510266)*x_ref[0]**o+(0.960755607224)*x_ref[0]+(0.992551377675)*x_ref[1]**o+(-0.40568563699)*x_ref[1]+(0.635141614881)*x_ref[2]**o+(-0.28322398176)*x_ref[2]
        ref[2,2,2,0]=(0.200301024931)*x_ref[0]**o+(-0.526020099906)*x_ref[0]+(0.296360677546)*x_ref[1]**o+(-0.346356255978)*x_ref[1]+(-0.751246948633)*x_ref[2]**o+(0.936356516645)*x_ref[2]
        ref[2,2,2,1]=(0.851662248368)*x_ref[0]**o+(0.306264287809)*x_ref[0]+(0.413913851675)*x_ref[1]**o+(0.796748819208)*x_ref[1]+(0.686640823111)*x_ref[2]**o+(0.270797124789)*x_ref[2]
        ref[2,3,0,0]=(0.873612484608)*x_ref[0]**o+(-0.635443314156)*x_ref[0]+(0.0336582752521)*x_ref[1]**o+(0.397952617963)*x_ref[1]+(0.708037547846)*x_ref[2]**o+(-0.835497214327)*x_ref[2]
        ref[2,3,0,1]=(0.266263219836)*x_ref[0]**o+(0.718604372519)*x_ref[0]+(-0.706375167711)*x_ref[1]**o+(0.267962253067)*x_ref[1]+(-0.0369901462112)*x_ref[2]**o+(0.541887114998)*x_ref[2]
        ref[2,3,1,0]=(-0.0883529797185)*x_ref[0]**o+(-0.695144807666)*x_ref[0]+(-0.332082923487)*x_ref[1]**o+(0.0386905137005)*x_ref[1]+(-0.0402663506352)*x_ref[2]**o+(0.897417224025)*x_ref[2]
        ref[2,3,1,1]=(-0.275313948401)*x_ref[0]**o+(0.66334859913)*x_ref[0]+(-0.139887516171)*x_ref[1]**o+(0.874104988552)*x_ref[1]+(-0.474834712531)*x_ref[2]**o+(-0.350478879758)*x_ref[2]
        ref[2,3,2,0]=(-0.370570829115)*x_ref[0]**o+(0.448685344991)*x_ref[0]+(-0.663934134305)*x_ref[1]**o+(0.270424636731)*x_ref[1]+(0.257023713378)*x_ref[2]**o+(-0.558834446028)*x_ref[2]
        ref[2,3,2,1]=(0.853600423341)*x_ref[0]**o+(-0.34026373254)*x_ref[0]+(-0.880960615323)*x_ref[1]**o+(0.720247030686)*x_ref[1]+(-0.179396842577)*x_ref[2]**o+(-0.581055218136)*x_ref[2]
        ref[2,4,0,0]=(0.661293829706)*x_ref[0]**o+(-0.278404566365)*x_ref[0]+(-0.668007328124)*x_ref[1]**o+(-0.0762718064116)*x_ref[1]+(-0.559881233142)*x_ref[2]**o+(0.38863355198)*x_ref[2]
        ref[2,4,0,1]=(0.841925292626)*x_ref[0]**o+(-0.907683026635)*x_ref[0]+(0.606133080471)*x_ref[1]**o+(-0.743094760572)*x_ref[1]+(-0.963492675546)*x_ref[2]**o+(0.551853881446)*x_ref[2]
        ref[2,4,1,0]=(0.106157048406)*x_ref[0]**o+(-0.839019337739)*x_ref[0]+(-0.646225647224)*x_ref[1]**o+(0.596433542879)*x_ref[1]+(-0.440653099815)*x_ref[2]**o+(0.612390706411)*x_ref[2]
        ref[2,4,1,1]=(0.0894808258178)*x_ref[0]**o+(0.601107281081)*x_ref[0]+(0.764980611835)*x_ref[1]**o+(-0.028637689046)*x_ref[1]+(-0.163992620355)*x_ref[2]**o+(0.811740898974)*x_ref[2]
        ref[2,4,2,0]=(0.159301482106)*x_ref[0]**o+(-0.2732663929)*x_ref[0]+(0.642848902873)*x_ref[1]**o+(-0.486181762406)*x_ref[1]+(-0.0353954685288)*x_ref[2]**o+(-0.17658875721)*x_ref[2]
        ref[2,4,2,1]=(-0.190069716212)*x_ref[0]**o+(0.911197492671)*x_ref[0]+(-0.404646749357)*x_ref[1]**o+(0.496457219159)*x_ref[1]+(0.331503854252)*x_ref[2]**o+(-0.0824993997725)*x_ref[2]
        ref[3,0,0,0]=(-0.00960152574795)*x_ref[0]**o+(-0.588053191696)*x_ref[0]+(0.134548796202)*x_ref[1]**o+(-0.734698264954)*x_ref[1]+(0.369880572143)*x_ref[2]**o+(-0.184585884928)*x_ref[2]
        ref[3,0,0,1]=(-0.894375930982)*x_ref[0]**o+(-0.757908178588)*x_ref[0]+(-0.862478750821)*x_ref[1]**o+(-0.860344696497)*x_ref[1]+(-0.606693340193)*x_ref[2]**o+(-0.211110504173)*x_ref[2]
        ref[3,0,1,0]=(0.29208602662)*x_ref[0]**o+(0.428795740188)*x_ref[0]+(0.730773643572)*x_ref[1]**o+(0.473531891026)*x_ref[1]+(-0.475140560425)*x_ref[2]**o+(0.973125507458)*x_ref[2]
        ref[3,0,1,1]=(-0.15741631987)*x_ref[0]**o+(-0.34595691825)*x_ref[0]+(-0.291584586864)*x_ref[1]**o+(0.202892853712)*x_ref[1]+(-0.48541451272)*x_ref[2]**o+(-0.0611059721993)*x_ref[2]
        ref[3,0,2,0]=(-0.309937469651)*x_ref[0]**o+(-0.515728964606)*x_ref[0]+(-0.212592556423)*x_ref[1]**o+(-0.419645818415)*x_ref[1]+(-0.825715279525)*x_ref[2]**o+(0.701775440231)*x_ref[2]
        ref[3,0,2,1]=(-0.527254609361)*x_ref[0]**o+(-0.921224942687)*x_ref[0]+(-0.219965563549)*x_ref[1]**o+(-0.520137812628)*x_ref[1]+(0.228621345698)*x_ref[2]**o+(0.607942457443)*x_ref[2]
        ref[3,1,0,0]=(-0.0191513045468)*x_ref[0]**o+(0.0593277991822)*x_ref[0]+(0.679065928035)*x_ref[1]**o+(-0.824466221695)*x_ref[1]+(-0.0207465360942)*x_ref[2]**o+(-0.462448546311)*x_ref[2]
        ref[3,1,0,1]=(-0.0226385405157)*x_ref[0]**o+(-0.990729100489)*x_ref[0]+(0.928646343096)*x_ref[1]**o+(0.877879386627)*x_ref[1]+(-0.774330622719)*x_ref[2]**o+(0.45794763385)*x_ref[2]
        ref[3,1,1,0]=(0.991939307411)*x_ref[0]**o+(0.424144367324)*x_ref[0]+(0.287966132208)*x_ref[1]**o+(0.560134521129)*x_ref[1]+(-0.366498917899)*x_ref[2]**o+(-0.5499574933)*x_ref[2]
        ref[3,1,1,1]=(0.131575237703)*x_ref[0]**o+(0.350065088754)*x_ref[0]+(0.845775797719)*x_ref[1]**o+(-0.951498449401)*x_ref[1]+(-0.749988925873)*x_ref[2]**o+(0.437681907383)*x_ref[2]
        ref[3,1,2,0]=(-0.177690055962)*x_ref[0]**o+(0.400513514288)*x_ref[0]+(0.654580691817)*x_ref[1]**o+(0.781954834622)*x_ref[1]+(0.498508301282)*x_ref[2]**o+(0.489109718057)*x_ref[2]
        ref[3,1,2,1]=(0.141266612215)*x_ref[0]**o+(0.578894900597)*x_ref[0]+(-0.0990011451546)*x_ref[1]**o+(0.455191961)*x_ref[1]+(0.446639012332)*x_ref[2]**o+(0.398374036483)*x_ref[2]
        ref[3,2,0,0]=(-0.88698978852)*x_ref[0]**o+(0.782997043916)*x_ref[0]+(-0.0203503423342)*x_ref[1]**o+(-0.729317082717)*x_ref[1]+(-0.916473605189)*x_ref[2]**o+(0.706748074855)*x_ref[2]
        ref[3,2,0,1]=(0.955314840013)*x_ref[0]**o+(0.784566529939)*x_ref[0]+(-0.607468453224)*x_ref[1]**o+(-0.960125382144)*x_ref[1]+(-0.62579570096)*x_ref[2]**o+(-0.0943302040401)*x_ref[2]
        ref[3,2,1,0]=(-0.255287575915)*x_ref[0]**o+(0.418535282029)*x_ref[0]+(0.606559849688)*x_ref[1]**o+(0.940877031124)*x_ref[1]+(0.714131644166)*x_ref[2]**o+(-0.77112172198)*x_ref[2]
        ref[3,2,1,1]=(0.566383366127)*x_ref[0]**o+(0.301059560952)*x_ref[0]+(0.181921162673)*x_ref[1]**o+(-0.972651860466)*x_ref[1]+(-0.435658160785)*x_ref[2]**o+(-0.449044377115)*x_ref[2]
        ref[3,2,2,0]=(0.730443884106)*x_ref[0]**o+(0.535861749024)*x_ref[0]+(-0.257910068997)*x_ref[1]**o+(-0.879923650189)*x_ref[1]+(0.24309905047)*x_ref[2]**o+(-0.407314031684)*x_ref[2]
        ref[3,2,2,1]=(-0.734460242423)*x_ref[0]**o+(0.694756195588)*x_ref[0]+(-0.0110855577415)*x_ref[1]**o+(0.213770377606)*x_ref[1]+(0.404975692507)*x_ref[2]**o+(0.235122345113)*x_ref[2]
        ref[3,3,0,0]=(-0.0854800701504)*x_ref[0]**o+(-0.998984290225)*x_ref[0]+(-0.781970276212)*x_ref[1]**o+(0.866788214566)*x_ref[1]+(-0.21747695978)*x_ref[2]**o+(-0.615256484308)*x_ref[2]
        ref[3,3,0,1]=(-0.509246205985)*x_ref[0]**o+(-0.870342470261)*x_ref[0]+(0.699615910083)*x_ref[1]**o+(0.81056073388)*x_ref[1]+(0.137830446242)*x_ref[2]**o+(-0.355111214574)*x_ref[2]
        ref[3,3,1,0]=(-0.423472624091)*x_ref[0]**o+(0.812256460885)*x_ref[0]+(-0.675497703361)*x_ref[1]**o+(0.296366723996)*x_ref[1]+(-0.218492550353)*x_ref[2]**o+(-0.852823171625)*x_ref[2]
        ref[3,3,1,1]=(-0.0718862672517)*x_ref[0]**o+(-0.122792076285)*x_ref[0]+(-0.794191931729)*x_ref[1]**o+(0.569212982974)*x_ref[1]+(-0.646851040773)*x_ref[2]**o+(-0.3073262807)*x_ref[2]
        ref[3,3,2,0]=(-0.994778576692)*x_ref[0]**o+(-0.250054159815)*x_ref[0]+(0.726266056262)*x_ref[1]**o+(-0.827929598129)*x_ref[1]+(0.038573450577)*x_ref[2]**o+(-0.402624412259)*x_ref[2]
        ref[3,3,2,1]=(-0.158546722118)*x_ref[0]**o+(0.438480217845)*x_ref[0]+(-0.757053187223)*x_ref[1]**o+(0.96669102229)*x_ref[1]+(0.585680018048)*x_ref[2]**o+(0.917795585404)*x_ref[2]
        ref[3,4,0,0]=(0.964942257873)*x_ref[0]**o+(0.222693858545)*x_ref[0]+(0.708449965849)*x_ref[1]**o+(-0.786172527336)*x_ref[1]+(0.813715873004)*x_ref[2]**o+(-0.0971222939825)*x_ref[2]
        ref[3,4,0,1]=(-0.14236975971)*x_ref[0]**o+(-0.748307677379)*x_ref[0]+(-0.269260913515)*x_ref[1]**o+(-0.315981317137)*x_ref[1]+(-0.375598072402)*x_ref[2]**o+(0.0797466318359)*x_ref[2]
        ref[3,4,1,0]=(0.52711029533)*x_ref[0]**o+(0.492047263035)*x_ref[0]+(0.678479511876)*x_ref[1]**o+(0.619219421066)*x_ref[1]+(-0.528779830025)*x_ref[2]**o+(-0.237989972194)*x_ref[2]
        ref[3,4,1,1]=(0.711881464696)*x_ref[0]**o+(-0.398127815977)*x_ref[0]+(-0.943946294128)*x_ref[1]**o+(0.584800438848)*x_ref[1]+(-0.683342581136)*x_ref[2]**o+(0.895377849601)*x_ref[2]
        ref[3,4,2,0]=(-0.698456333887)*x_ref[0]**o+(-0.36871456039)*x_ref[0]+(0.216620614245)*x_ref[1]**o+(-0.242036817594)*x_ref[1]+(-0.845502631302)*x_ref[2]**o+(-0.705164123503)*x_ref[2]
        ref[3,4,2,1]=(-0.000297852657095)*x_ref[0]**o+(-0.847929455729)*x_ref[0]+(0.496739572695)*x_ref[1]**o+(0.574135132795)*x_ref[1]+(0.585537276618)*x_ref[2]**o+(-0.934539532574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.713276041917)*x[0]**o+(-0.0688503309598)*x[0]+(-0.0478916615918)*x[1]**o+(-0.868054043492)*x[1]
        ref=(-0.713276041917)*x_ref[0]**o+(-0.0688503309598)*x_ref[0]+(-0.0478916615918)*x_ref[1]**o+(-0.868054043492)*x_ref[1]
      else:
        s=(-0.931008947411)*x[0]**o+(0.772025637695)*x[0]+(-0.155082242593)*x[1]**o+(0.808407227202)*x[1]+(0.860761693977)*x[2]**o+(-0.965019813018)*x[2]
        ref=(-0.931008947411)*x_ref[0]**o+(0.772025637695)*x_ref[0]+(-0.155082242593)*x_ref[1]**o+(0.808407227202)*x_ref[1]+(0.860761693977)*x_ref[2]**o+(-0.965019813018)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.752443452146)*x[0]**o+(0.0736601797185)*x[0]+(-0.735031045379)*x[1]**o+(0.538489905213)*x[1]
        s[1]=(0.799601080205)*x[0]**o+(0.214497318519)*x[0]+(0.168613900079)*x[1]**o+(0.0388630346028)*x[1]
        ref[0]=(0.752443452146)*x_ref[0]**o+(0.0736601797185)*x_ref[0]+(-0.735031045379)*x_ref[1]**o+(0.538489905213)*x_ref[1]
        ref[1]=(0.799601080205)*x_ref[0]**o+(0.214497318519)*x_ref[0]+(0.168613900079)*x_ref[1]**o+(0.0388630346028)*x_ref[1]
      else:
        s[0]=(0.175265620206)*x[0]**o+(-0.333727702895)*x[0]+(-0.932525223484)*x[1]**o+(0.925315017116)*x[1]+(0.928659911973)*x[2]**o+(0.445185886277)*x[2]
        s[1]=(0.311109115611)*x[0]**o+(0.0231809158727)*x[0]+(0.393168814415)*x[1]**o+(0.452573797902)*x[1]+(0.896927701022)*x[2]**o+(0.944685279345)*x[2]
        ref[0]=(0.175265620206)*x_ref[0]**o+(-0.333727702895)*x_ref[0]+(-0.932525223484)*x_ref[1]**o+(0.925315017116)*x_ref[1]+(0.928659911973)*x_ref[2]**o+(0.445185886277)*x_ref[2]
        ref[1]=(0.311109115611)*x_ref[0]**o+(0.0231809158727)*x_ref[0]+(0.393168814415)*x_ref[1]**o+(0.452573797902)*x_ref[1]+(0.896927701022)*x_ref[2]**o+(0.944685279345)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.708553013661)*x[0]**o+(0.114066657094)*x[0]+(-0.84986272259)*x[1]**o+(-0.0539087301019)*x[1]
        s[0,1]=(-0.266588917197)*x[0]**o+(0.0504848378446)*x[0]+(-0.635596388347)*x[1]**o+(-0.261063433899)*x[1]
        s[0,2]=(-0.0894634722171)*x[0]**o+(0.573385597336)*x[0]+(-0.842136337596)*x[1]**o+(0.238667481456)*x[1]
        s[0,3]=(0.87882171546)*x[0]**o+(0.417532202681)*x[0]+(0.936866438423)*x[1]**o+(-0.724174392442)*x[1]
        s[0,4]=(0.0145872637015)*x[0]**o+(0.421850663727)*x[0]+(0.90053153032)*x[1]**o+(0.955508744656)*x[1]
        s[1,0]=(0.140196302181)*x[0]**o+(-0.638550845113)*x[0]+(0.468359848634)*x[1]**o+(0.893083141798)*x[1]
        s[1,1]=(0.984992789447)*x[0]**o+(-0.843621073237)*x[0]+(0.798478310102)*x[1]**o+(-0.363619894589)*x[1]
        s[1,2]=(0.895636184991)*x[0]**o+(-0.229789622077)*x[0]+(0.098171182568)*x[1]**o+(0.579163598552)*x[1]
        s[1,3]=(0.927820509391)*x[0]**o+(-0.947766499615)*x[0]+(0.183243198595)*x[1]**o+(-0.31472874822)*x[1]
        s[1,4]=(-0.966327205994)*x[0]**o+(-0.289895049587)*x[0]+(0.514876200606)*x[1]**o+(0.653407964622)*x[1]
        s[2,0]=(-0.903904444429)*x[0]**o+(-0.263217684631)*x[0]+(0.291271291348)*x[1]**o+(-0.646562462633)*x[1]
        s[2,1]=(-0.528540658615)*x[0]**o+(0.246812336359)*x[0]+(0.658538113503)*x[1]**o+(-0.407851907836)*x[1]
        s[2,2]=(-0.715562167768)*x[0]**o+(-0.855788728416)*x[0]+(-0.993665928364)*x[1]**o+(-0.608513139044)*x[1]
        s[2,3]=(0.363176542145)*x[0]**o+(0.588619594218)*x[0]+(-0.260592717735)*x[1]**o+(-0.558179141071)*x[1]
        s[2,4]=(0.532120055093)*x[0]**o+(0.090598503196)*x[0]+(0.80389069649)*x[1]**o+(-0.988747865176)*x[1]
        s[3,0]=(0.855196983003)*x[0]**o+(0.348709023449)*x[0]+(-0.592402303364)*x[1]**o+(-0.535439951031)*x[1]
        s[3,1]=(-0.132346044705)*x[0]**o+(-0.849374461726)*x[0]+(-0.812790438958)*x[1]**o+(0.291967653653)*x[1]
        s[3,2]=(-0.39273694335)*x[0]**o+(-0.871595348254)*x[0]+(-0.744651099213)*x[1]**o+(-0.231171036464)*x[1]
        s[3,3]=(-0.616438342441)*x[0]**o+(0.0131646132899)*x[0]+(-0.265184331359)*x[1]**o+(0.447239576276)*x[1]
        s[3,4]=(0.870983993837)*x[0]**o+(-0.734468653473)*x[0]+(0.888611194222)*x[1]**o+(0.713808943447)*x[1]
        ref[0,0]=(-0.708553013661)*x_ref[0]**o+(0.114066657094)*x_ref[0]+(-0.84986272259)*x_ref[1]**o+(-0.0539087301019)*x_ref[1]
        ref[0,1]=(-0.266588917197)*x_ref[0]**o+(0.0504848378446)*x_ref[0]+(-0.635596388347)*x_ref[1]**o+(-0.261063433899)*x_ref[1]
        ref[0,2]=(-0.0894634722171)*x_ref[0]**o+(0.573385597336)*x_ref[0]+(-0.842136337596)*x_ref[1]**o+(0.238667481456)*x_ref[1]
        ref[0,3]=(0.87882171546)*x_ref[0]**o+(0.417532202681)*x_ref[0]+(0.936866438423)*x_ref[1]**o+(-0.724174392442)*x_ref[1]
        ref[0,4]=(0.0145872637015)*x_ref[0]**o+(0.421850663727)*x_ref[0]+(0.90053153032)*x_ref[1]**o+(0.955508744656)*x_ref[1]
        ref[1,0]=(0.140196302181)*x_ref[0]**o+(-0.638550845113)*x_ref[0]+(0.468359848634)*x_ref[1]**o+(0.893083141798)*x_ref[1]
        ref[1,1]=(0.984992789447)*x_ref[0]**o+(-0.843621073237)*x_ref[0]+(0.798478310102)*x_ref[1]**o+(-0.363619894589)*x_ref[1]
        ref[1,2]=(0.895636184991)*x_ref[0]**o+(-0.229789622077)*x_ref[0]+(0.098171182568)*x_ref[1]**o+(0.579163598552)*x_ref[1]
        ref[1,3]=(0.927820509391)*x_ref[0]**o+(-0.947766499615)*x_ref[0]+(0.183243198595)*x_ref[1]**o+(-0.31472874822)*x_ref[1]
        ref[1,4]=(-0.966327205994)*x_ref[0]**o+(-0.289895049587)*x_ref[0]+(0.514876200606)*x_ref[1]**o+(0.653407964622)*x_ref[1]
        ref[2,0]=(-0.903904444429)*x_ref[0]**o+(-0.263217684631)*x_ref[0]+(0.291271291348)*x_ref[1]**o+(-0.646562462633)*x_ref[1]
        ref[2,1]=(-0.528540658615)*x_ref[0]**o+(0.246812336359)*x_ref[0]+(0.658538113503)*x_ref[1]**o+(-0.407851907836)*x_ref[1]
        ref[2,2]=(-0.715562167768)*x_ref[0]**o+(-0.855788728416)*x_ref[0]+(-0.993665928364)*x_ref[1]**o+(-0.608513139044)*x_ref[1]
        ref[2,3]=(0.363176542145)*x_ref[0]**o+(0.588619594218)*x_ref[0]+(-0.260592717735)*x_ref[1]**o+(-0.558179141071)*x_ref[1]
        ref[2,4]=(0.532120055093)*x_ref[0]**o+(0.090598503196)*x_ref[0]+(0.80389069649)*x_ref[1]**o+(-0.988747865176)*x_ref[1]
        ref[3,0]=(0.855196983003)*x_ref[0]**o+(0.348709023449)*x_ref[0]+(-0.592402303364)*x_ref[1]**o+(-0.535439951031)*x_ref[1]
        ref[3,1]=(-0.132346044705)*x_ref[0]**o+(-0.849374461726)*x_ref[0]+(-0.812790438958)*x_ref[1]**o+(0.291967653653)*x_ref[1]
        ref[3,2]=(-0.39273694335)*x_ref[0]**o+(-0.871595348254)*x_ref[0]+(-0.744651099213)*x_ref[1]**o+(-0.231171036464)*x_ref[1]
        ref[3,3]=(-0.616438342441)*x_ref[0]**o+(0.0131646132899)*x_ref[0]+(-0.265184331359)*x_ref[1]**o+(0.447239576276)*x_ref[1]
        ref[3,4]=(0.870983993837)*x_ref[0]**o+(-0.734468653473)*x_ref[0]+(0.888611194222)*x_ref[1]**o+(0.713808943447)*x_ref[1]
      else:
        s[0,0]=(-0.810387756876)*x[0]**o+(-0.562586750225)*x[0]+(0.896156420696)*x[1]**o+(-0.996536575244)*x[1]+(0.427600721566)*x[2]**o+(0.837763430808)*x[2]
        s[0,1]=(-0.63001573449)*x[0]**o+(0.654477767706)*x[0]+(-0.795127551271)*x[1]**o+(0.536595939418)*x[1]+(-0.727127268215)*x[2]**o+(-0.358649556898)*x[2]
        s[0,2]=(0.584792270415)*x[0]**o+(-0.13886821113)*x[0]+(-0.526668048387)*x[1]**o+(-0.0830908739307)*x[1]+(0.647801230167)*x[2]**o+(-0.299674557564)*x[2]
        s[0,3]=(0.237272327463)*x[0]**o+(-0.665360840533)*x[0]+(0.642053734612)*x[1]**o+(0.0162314692731)*x[1]+(0.927935045044)*x[2]**o+(0.164570226784)*x[2]
        s[0,4]=(0.219203335717)*x[0]**o+(0.453183788553)*x[0]+(0.124112536481)*x[1]**o+(-0.0284133490265)*x[1]+(-0.919674987737)*x[2]**o+(-0.0149293124196)*x[2]
        s[1,0]=(-0.0768501669381)*x[0]**o+(0.60547593141)*x[0]+(0.715752118361)*x[1]**o+(0.886572828266)*x[1]+(0.00765847416371)*x[2]**o+(0.861052492746)*x[2]
        s[1,1]=(0.255734836806)*x[0]**o+(-0.246883769311)*x[0]+(0.0800805057057)*x[1]**o+(-0.790568208751)*x[1]+(-0.748973939025)*x[2]**o+(-0.182812700448)*x[2]
        s[1,2]=(0.199422259597)*x[0]**o+(-0.627479494962)*x[0]+(0.0565633628195)*x[1]**o+(0.350443086044)*x[1]+(-0.380622372229)*x[2]**o+(0.00708476531876)*x[2]
        s[1,3]=(0.40010280022)*x[0]**o+(-0.767881252757)*x[0]+(-0.56339937337)*x[1]**o+(-0.352542604473)*x[1]+(0.801372900062)*x[2]**o+(-0.126263266304)*x[2]
        s[1,4]=(-0.128959410592)*x[0]**o+(-0.59236859859)*x[0]+(0.0980131449714)*x[1]**o+(0.853907069528)*x[1]+(-0.691379624)*x[2]**o+(0.386547512092)*x[2]
        s[2,0]=(-0.360626473774)*x[0]**o+(-0.76083965008)*x[0]+(0.470179818787)*x[1]**o+(-0.824599812059)*x[1]+(0.420014625648)*x[2]**o+(-0.479269079111)*x[2]
        s[2,1]=(-0.81467966055)*x[0]**o+(-0.329346577295)*x[0]+(0.748928864266)*x[1]**o+(0.989795060424)*x[1]+(0.840479614557)*x[2]**o+(0.361335074457)*x[2]
        s[2,2]=(0.0627744010829)*x[0]**o+(-0.155252499175)*x[0]+(0.195454946586)*x[1]**o+(0.628864956458)*x[1]+(0.664724662178)*x[2]**o+(0.0471499062877)*x[2]
        s[2,3]=(-0.150860729652)*x[0]**o+(-0.18636002517)*x[0]+(0.964787336251)*x[1]**o+(0.871287020262)*x[1]+(-0.557208138103)*x[2]**o+(0.734747013137)*x[2]
        s[2,4]=(0.146220522342)*x[0]**o+(0.828100181189)*x[0]+(0.275317867253)*x[1]**o+(-0.533950851352)*x[1]+(0.058219246545)*x[2]**o+(-0.264489071187)*x[2]
        s[3,0]=(0.759487099327)*x[0]**o+(-0.943901668987)*x[0]+(0.917271393015)*x[1]**o+(0.991514325689)*x[1]+(-0.0693941475922)*x[2]**o+(0.971629779968)*x[2]
        s[3,1]=(-0.26255428914)*x[0]**o+(-0.126823739856)*x[0]+(0.769392620344)*x[1]**o+(0.124852282039)*x[1]+(0.719253662224)*x[2]**o+(0.189838410592)*x[2]
        s[3,2]=(0.513890040965)*x[0]**o+(0.849300040037)*x[0]+(0.375236082211)*x[1]**o+(-0.271681479664)*x[1]+(0.580679441212)*x[2]**o+(-0.0117174804836)*x[2]
        s[3,3]=(0.886420007368)*x[0]**o+(-0.140667253897)*x[0]+(0.85708792282)*x[1]**o+(0.867659131336)*x[1]+(0.418020814822)*x[2]**o+(0.817953483616)*x[2]
        s[3,4]=(0.449966058878)*x[0]**o+(0.424180609442)*x[0]+(0.42140819426)*x[1]**o+(0.180925977263)*x[1]+(-0.571592032058)*x[2]**o+(0.257682237168)*x[2]
        ref[0,0]=(-0.810387756876)*x_ref[0]**o+(-0.562586750225)*x_ref[0]+(0.896156420696)*x_ref[1]**o+(-0.996536575244)*x_ref[1]+(0.427600721566)*x_ref[2]**o+(0.837763430808)*x_ref[2]
        ref[0,1]=(-0.63001573449)*x_ref[0]**o+(0.654477767706)*x_ref[0]+(-0.795127551271)*x_ref[1]**o+(0.536595939418)*x_ref[1]+(-0.727127268215)*x_ref[2]**o+(-0.358649556898)*x_ref[2]
        ref[0,2]=(0.584792270415)*x_ref[0]**o+(-0.13886821113)*x_ref[0]+(-0.526668048387)*x_ref[1]**o+(-0.0830908739307)*x_ref[1]+(0.647801230167)*x_ref[2]**o+(-0.299674557564)*x_ref[2]
        ref[0,3]=(0.237272327463)*x_ref[0]**o+(-0.665360840533)*x_ref[0]+(0.642053734612)*x_ref[1]**o+(0.0162314692731)*x_ref[1]+(0.927935045044)*x_ref[2]**o+(0.164570226784)*x_ref[2]
        ref[0,4]=(0.219203335717)*x_ref[0]**o+(0.453183788553)*x_ref[0]+(0.124112536481)*x_ref[1]**o+(-0.0284133490265)*x_ref[1]+(-0.919674987737)*x_ref[2]**o+(-0.0149293124196)*x_ref[2]
        ref[1,0]=(-0.0768501669381)*x_ref[0]**o+(0.60547593141)*x_ref[0]+(0.715752118361)*x_ref[1]**o+(0.886572828266)*x_ref[1]+(0.00765847416371)*x_ref[2]**o+(0.861052492746)*x_ref[2]
        ref[1,1]=(0.255734836806)*x_ref[0]**o+(-0.246883769311)*x_ref[0]+(0.0800805057057)*x_ref[1]**o+(-0.790568208751)*x_ref[1]+(-0.748973939025)*x_ref[2]**o+(-0.182812700448)*x_ref[2]
        ref[1,2]=(0.199422259597)*x_ref[0]**o+(-0.627479494962)*x_ref[0]+(0.0565633628195)*x_ref[1]**o+(0.350443086044)*x_ref[1]+(-0.380622372229)*x_ref[2]**o+(0.00708476531876)*x_ref[2]
        ref[1,3]=(0.40010280022)*x_ref[0]**o+(-0.767881252757)*x_ref[0]+(-0.56339937337)*x_ref[1]**o+(-0.352542604473)*x_ref[1]+(0.801372900062)*x_ref[2]**o+(-0.126263266304)*x_ref[2]
        ref[1,4]=(-0.128959410592)*x_ref[0]**o+(-0.59236859859)*x_ref[0]+(0.0980131449714)*x_ref[1]**o+(0.853907069528)*x_ref[1]+(-0.691379624)*x_ref[2]**o+(0.386547512092)*x_ref[2]
        ref[2,0]=(-0.360626473774)*x_ref[0]**o+(-0.76083965008)*x_ref[0]+(0.470179818787)*x_ref[1]**o+(-0.824599812059)*x_ref[1]+(0.420014625648)*x_ref[2]**o+(-0.479269079111)*x_ref[2]
        ref[2,1]=(-0.81467966055)*x_ref[0]**o+(-0.329346577295)*x_ref[0]+(0.748928864266)*x_ref[1]**o+(0.989795060424)*x_ref[1]+(0.840479614557)*x_ref[2]**o+(0.361335074457)*x_ref[2]
        ref[2,2]=(0.0627744010829)*x_ref[0]**o+(-0.155252499175)*x_ref[0]+(0.195454946586)*x_ref[1]**o+(0.628864956458)*x_ref[1]+(0.664724662178)*x_ref[2]**o+(0.0471499062877)*x_ref[2]
        ref[2,3]=(-0.150860729652)*x_ref[0]**o+(-0.18636002517)*x_ref[0]+(0.964787336251)*x_ref[1]**o+(0.871287020262)*x_ref[1]+(-0.557208138103)*x_ref[2]**o+(0.734747013137)*x_ref[2]
        ref[2,4]=(0.146220522342)*x_ref[0]**o+(0.828100181189)*x_ref[0]+(0.275317867253)*x_ref[1]**o+(-0.533950851352)*x_ref[1]+(0.058219246545)*x_ref[2]**o+(-0.264489071187)*x_ref[2]
        ref[3,0]=(0.759487099327)*x_ref[0]**o+(-0.943901668987)*x_ref[0]+(0.917271393015)*x_ref[1]**o+(0.991514325689)*x_ref[1]+(-0.0693941475922)*x_ref[2]**o+(0.971629779968)*x_ref[2]
        ref[3,1]=(-0.26255428914)*x_ref[0]**o+(-0.126823739856)*x_ref[0]+(0.769392620344)*x_ref[1]**o+(0.124852282039)*x_ref[1]+(0.719253662224)*x_ref[2]**o+(0.189838410592)*x_ref[2]
        ref[3,2]=(0.513890040965)*x_ref[0]**o+(0.849300040037)*x_ref[0]+(0.375236082211)*x_ref[1]**o+(-0.271681479664)*x_ref[1]+(0.580679441212)*x_ref[2]**o+(-0.0117174804836)*x_ref[2]
        ref[3,3]=(0.886420007368)*x_ref[0]**o+(-0.140667253897)*x_ref[0]+(0.85708792282)*x_ref[1]**o+(0.867659131336)*x_ref[1]+(0.418020814822)*x_ref[2]**o+(0.817953483616)*x_ref[2]
        ref[3,4]=(0.449966058878)*x_ref[0]**o+(0.424180609442)*x_ref[0]+(0.42140819426)*x_ref[1]**o+(0.180925977263)*x_ref[1]+(-0.571592032058)*x_ref[2]**o+(0.257682237168)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.104698071101)*x[0]**o+(0.85225441166)*x[0]+(0.643738758183)*x[1]**o+(0.8443629101)*x[1]
        s[0,0,1]=(-0.293745943532)*x[0]**o+(0.677723913417)*x[0]+(-0.12767285172)*x[1]**o+(-0.228090688839)*x[1]
        s[0,1,0]=(0.114974377468)*x[0]**o+(-0.769098874811)*x[0]+(-0.173276425412)*x[1]**o+(0.0507852449622)*x[1]
        s[0,1,1]=(0.779779857754)*x[0]**o+(0.0482150028893)*x[0]+(0.208388483635)*x[1]**o+(-0.542545383735)*x[1]
        s[1,0,0]=(0.857554163662)*x[0]**o+(0.890423092097)*x[0]+(0.603776777208)*x[1]**o+(-0.35588830751)*x[1]
        s[1,0,1]=(0.277165242151)*x[0]**o+(0.805484036517)*x[0]+(-0.577456699494)*x[1]**o+(-0.160490000703)*x[1]
        s[1,1,0]=(0.928438839711)*x[0]**o+(-0.716888927671)*x[0]+(-0.138675664609)*x[1]**o+(0.146500337023)*x[1]
        s[1,1,1]=(-0.3626191689)*x[0]**o+(-0.442937599469)*x[0]+(-0.407066971319)*x[1]**o+(0.623768042338)*x[1]
        s[2,0,0]=(-0.347034739676)*x[0]**o+(0.715507004381)*x[0]+(0.856404818205)*x[1]**o+(-0.04028097386)*x[1]
        s[2,0,1]=(-0.0606946296735)*x[0]**o+(0.128860413259)*x[0]+(-0.142401804193)*x[1]**o+(-0.0530515258875)*x[1]
        s[2,1,0]=(-0.418284924355)*x[0]**o+(0.608535815348)*x[0]+(-0.753014535299)*x[1]**o+(-0.678095181481)*x[1]
        s[2,1,1]=(0.15035667769)*x[0]**o+(0.0492019049213)*x[0]+(-0.677433635501)*x[1]**o+(0.352404951558)*x[1]
        s[3,0,0]=(-0.0290046263356)*x[0]**o+(-0.804064667193)*x[0]+(-0.804537375276)*x[1]**o+(0.475053057415)*x[1]
        s[3,0,1]=(-0.340794340714)*x[0]**o+(-0.155934699109)*x[0]+(0.91002644066)*x[1]**o+(0.784480548529)*x[1]
        s[3,1,0]=(-0.484483021942)*x[0]**o+(0.52465323919)*x[0]+(0.751560457535)*x[1]**o+(0.28163378712)*x[1]
        s[3,1,1]=(-0.309783414062)*x[0]**o+(0.066240252028)*x[0]+(0.183403362957)*x[1]**o+(-0.523037580611)*x[1]
        s[4,0,0]=(-0.568631885289)*x[0]**o+(-0.654263636024)*x[0]+(-0.605614348372)*x[1]**o+(-0.894236319906)*x[1]
        s[4,0,1]=(0.325194519014)*x[0]**o+(0.585375626733)*x[0]+(0.20845589395)*x[1]**o+(0.193035797172)*x[1]
        s[4,1,0]=(-0.142515818097)*x[0]**o+(-0.654569638237)*x[0]+(-0.928863338211)*x[1]**o+(-0.963259968871)*x[1]
        s[4,1,1]=(0.313307585897)*x[0]**o+(0.185118633115)*x[0]+(-0.439344812316)*x[1]**o+(-0.760250886639)*x[1]
        s[5,0,0]=(-0.592313178206)*x[0]**o+(-0.264499099416)*x[0]+(-0.620889224638)*x[1]**o+(0.457484719025)*x[1]
        s[5,0,1]=(-0.083928120416)*x[0]**o+(0.25893121162)*x[0]+(0.403250910043)*x[1]**o+(0.0822798058585)*x[1]
        s[5,1,0]=(0.617807311)*x[0]**o+(0.0627862768802)*x[0]+(-0.595563604156)*x[1]**o+(-0.157413618531)*x[1]
        s[5,1,1]=(0.919933775557)*x[0]**o+(-0.736703943919)*x[0]+(-0.935265088031)*x[1]**o+(-0.953284151506)*x[1]
        ref[0,0,0]=(0.104698071101)*x_ref[0]**o+(0.85225441166)*x_ref[0]+(0.643738758183)*x_ref[1]**o+(0.8443629101)*x_ref[1]
        ref[0,0,1]=(-0.293745943532)*x_ref[0]**o+(0.677723913417)*x_ref[0]+(-0.12767285172)*x_ref[1]**o+(-0.228090688839)*x_ref[1]
        ref[0,1,0]=(0.114974377468)*x_ref[0]**o+(-0.769098874811)*x_ref[0]+(-0.173276425412)*x_ref[1]**o+(0.0507852449622)*x_ref[1]
        ref[0,1,1]=(0.779779857754)*x_ref[0]**o+(0.0482150028893)*x_ref[0]+(0.208388483635)*x_ref[1]**o+(-0.542545383735)*x_ref[1]
        ref[1,0,0]=(0.857554163662)*x_ref[0]**o+(0.890423092097)*x_ref[0]+(0.603776777208)*x_ref[1]**o+(-0.35588830751)*x_ref[1]
        ref[1,0,1]=(0.277165242151)*x_ref[0]**o+(0.805484036517)*x_ref[0]+(-0.577456699494)*x_ref[1]**o+(-0.160490000703)*x_ref[1]
        ref[1,1,0]=(0.928438839711)*x_ref[0]**o+(-0.716888927671)*x_ref[0]+(-0.138675664609)*x_ref[1]**o+(0.146500337023)*x_ref[1]
        ref[1,1,1]=(-0.3626191689)*x_ref[0]**o+(-0.442937599469)*x_ref[0]+(-0.407066971319)*x_ref[1]**o+(0.623768042338)*x_ref[1]
        ref[2,0,0]=(-0.347034739676)*x_ref[0]**o+(0.715507004381)*x_ref[0]+(0.856404818205)*x_ref[1]**o+(-0.04028097386)*x_ref[1]
        ref[2,0,1]=(-0.0606946296735)*x_ref[0]**o+(0.128860413259)*x_ref[0]+(-0.142401804193)*x_ref[1]**o+(-0.0530515258875)*x_ref[1]
        ref[2,1,0]=(-0.418284924355)*x_ref[0]**o+(0.608535815348)*x_ref[0]+(-0.753014535299)*x_ref[1]**o+(-0.678095181481)*x_ref[1]
        ref[2,1,1]=(0.15035667769)*x_ref[0]**o+(0.0492019049213)*x_ref[0]+(-0.677433635501)*x_ref[1]**o+(0.352404951558)*x_ref[1]
        ref[3,0,0]=(-0.0290046263356)*x_ref[0]**o+(-0.804064667193)*x_ref[0]+(-0.804537375276)*x_ref[1]**o+(0.475053057415)*x_ref[1]
        ref[3,0,1]=(-0.340794340714)*x_ref[0]**o+(-0.155934699109)*x_ref[0]+(0.91002644066)*x_ref[1]**o+(0.784480548529)*x_ref[1]
        ref[3,1,0]=(-0.484483021942)*x_ref[0]**o+(0.52465323919)*x_ref[0]+(0.751560457535)*x_ref[1]**o+(0.28163378712)*x_ref[1]
        ref[3,1,1]=(-0.309783414062)*x_ref[0]**o+(0.066240252028)*x_ref[0]+(0.183403362957)*x_ref[1]**o+(-0.523037580611)*x_ref[1]
        ref[4,0,0]=(-0.568631885289)*x_ref[0]**o+(-0.654263636024)*x_ref[0]+(-0.605614348372)*x_ref[1]**o+(-0.894236319906)*x_ref[1]
        ref[4,0,1]=(0.325194519014)*x_ref[0]**o+(0.585375626733)*x_ref[0]+(0.20845589395)*x_ref[1]**o+(0.193035797172)*x_ref[1]
        ref[4,1,0]=(-0.142515818097)*x_ref[0]**o+(-0.654569638237)*x_ref[0]+(-0.928863338211)*x_ref[1]**o+(-0.963259968871)*x_ref[1]
        ref[4,1,1]=(0.313307585897)*x_ref[0]**o+(0.185118633115)*x_ref[0]+(-0.439344812316)*x_ref[1]**o+(-0.760250886639)*x_ref[1]
        ref[5,0,0]=(-0.592313178206)*x_ref[0]**o+(-0.264499099416)*x_ref[0]+(-0.620889224638)*x_ref[1]**o+(0.457484719025)*x_ref[1]
        ref[5,0,1]=(-0.083928120416)*x_ref[0]**o+(0.25893121162)*x_ref[0]+(0.403250910043)*x_ref[1]**o+(0.0822798058585)*x_ref[1]
        ref[5,1,0]=(0.617807311)*x_ref[0]**o+(0.0627862768802)*x_ref[0]+(-0.595563604156)*x_ref[1]**o+(-0.157413618531)*x_ref[1]
        ref[5,1,1]=(0.919933775557)*x_ref[0]**o+(-0.736703943919)*x_ref[0]+(-0.935265088031)*x_ref[1]**o+(-0.953284151506)*x_ref[1]
      else:
        s[0,0,0]=(-0.542725715353)*x[0]**o+(-0.388718702529)*x[0]+(-0.260726707092)*x[1]**o+(0.00868281860494)*x[1]+(-0.422209705753)*x[2]**o+(-0.482952338663)*x[2]
        s[0,0,1]=(0.517312862567)*x[0]**o+(-0.213607124389)*x[0]+(-0.717854605214)*x[1]**o+(0.102559996175)*x[1]+(-0.248146688044)*x[2]**o+(-0.472262287207)*x[2]
        s[0,1,0]=(-0.561062159702)*x[0]**o+(0.625784702743)*x[0]+(0.127217512593)*x[1]**o+(0.431860689921)*x[1]+(-0.743622124184)*x[2]**o+(-0.549196149668)*x[2]
        s[0,1,1]=(-0.47244041365)*x[0]**o+(-0.905145373286)*x[0]+(0.0327050909564)*x[1]**o+(-0.0753932245132)*x[1]+(-0.669475816436)*x[2]**o+(-0.913342589953)*x[2]
        s[1,0,0]=(-0.107737055645)*x[0]**o+(0.890187876057)*x[0]+(0.619389534096)*x[1]**o+(0.505636695416)*x[1]+(-0.773316774174)*x[2]**o+(0.57289098512)*x[2]
        s[1,0,1]=(0.998166227402)*x[0]**o+(0.426101053356)*x[0]+(-0.275256494275)*x[1]**o+(-0.642996762298)*x[1]+(0.80180314192)*x[2]**o+(-0.777630700737)*x[2]
        s[1,1,0]=(0.0567929745565)*x[0]**o+(-0.634357283891)*x[0]+(0.756220569552)*x[1]**o+(0.225394362055)*x[1]+(-0.380906558144)*x[2]**o+(0.656365329349)*x[2]
        s[1,1,1]=(-0.118345871482)*x[0]**o+(-0.601934055233)*x[0]+(-0.466776830591)*x[1]**o+(0.83729192062)*x[1]+(-0.760321613956)*x[2]**o+(-0.0364679505353)*x[2]
        s[2,0,0]=(-0.0254736115914)*x[0]**o+(0.219281883056)*x[0]+(0.616921568912)*x[1]**o+(0.955364895459)*x[1]+(0.468886733279)*x[2]**o+(-0.665678021975)*x[2]
        s[2,0,1]=(0.452552793802)*x[0]**o+(0.418962651335)*x[0]+(-0.493607930291)*x[1]**o+(0.446553418335)*x[1]+(0.297233373077)*x[2]**o+(0.140710729735)*x[2]
        s[2,1,0]=(-0.636104671214)*x[0]**o+(0.770136919493)*x[0]+(0.559068996696)*x[1]**o+(0.1863482658)*x[1]+(-0.486115384192)*x[2]**o+(0.766345710304)*x[2]
        s[2,1,1]=(0.365405358671)*x[0]**o+(-0.349630033342)*x[0]+(0.134851622297)*x[1]**o+(0.274415828646)*x[1]+(0.736462073823)*x[2]**o+(0.113366567452)*x[2]
        s[3,0,0]=(0.103210866706)*x[0]**o+(0.781535459867)*x[0]+(-0.278787888162)*x[1]**o+(-0.931650808802)*x[1]+(0.640371737202)*x[2]**o+(-0.0105202640396)*x[2]
        s[3,0,1]=(-0.249043848359)*x[0]**o+(0.524828750016)*x[0]+(0.533986967075)*x[1]**o+(-0.267136752764)*x[1]+(0.451666436163)*x[2]**o+(-0.205612541545)*x[2]
        s[3,1,0]=(-0.604476259822)*x[0]**o+(-0.0957734520284)*x[0]+(-0.970823460222)*x[1]**o+(0.980588655716)*x[1]+(-0.124494413992)*x[2]**o+(0.252216095396)*x[2]
        s[3,1,1]=(0.0212563582378)*x[0]**o+(-0.0106688543408)*x[0]+(-0.362460274705)*x[1]**o+(-0.210672777607)*x[1]+(0.459825480445)*x[2]**o+(0.317321978801)*x[2]
        s[4,0,0]=(0.587000133255)*x[0]**o+(0.362106866018)*x[0]+(0.839110518219)*x[1]**o+(-0.183918723198)*x[1]+(-0.357506983794)*x[2]**o+(0.964414772317)*x[2]
        s[4,0,1]=(-0.914924387748)*x[0]**o+(-0.712924404308)*x[0]+(-0.81536262433)*x[1]**o+(-0.246055912409)*x[1]+(0.550257872298)*x[2]**o+(0.308874758093)*x[2]
        s[4,1,0]=(-0.451464088139)*x[0]**o+(0.113750801605)*x[0]+(-0.204501128464)*x[1]**o+(0.694001185736)*x[1]+(0.80368669429)*x[2]**o+(0.564414553079)*x[2]
        s[4,1,1]=(-0.354511540467)*x[0]**o+(0.295005960713)*x[0]+(-0.581487363676)*x[1]**o+(-0.862369285262)*x[1]+(0.268927694418)*x[2]**o+(0.91534230742)*x[2]
        s[5,0,0]=(-0.361145485775)*x[0]**o+(-0.136825333824)*x[0]+(-0.917163652695)*x[1]**o+(-0.92225469793)*x[1]+(0.142972538406)*x[2]**o+(0.625895771061)*x[2]
        s[5,0,1]=(0.976797635598)*x[0]**o+(-0.30046259343)*x[0]+(-0.831191492174)*x[1]**o+(-0.422489136999)*x[1]+(0.963327796572)*x[2]**o+(0.216709868453)*x[2]
        s[5,1,0]=(-0.396424567827)*x[0]**o+(-0.037861045243)*x[0]+(0.461627073863)*x[1]**o+(0.299397593712)*x[1]+(0.607619683163)*x[2]**o+(-0.260194783511)*x[2]
        s[5,1,1]=(0.419140017883)*x[0]**o+(0.383165340851)*x[0]+(0.561345759207)*x[1]**o+(-0.406706766251)*x[1]+(-0.730360027768)*x[2]**o+(-0.0907334422202)*x[2]
        ref[0,0,0]=(-0.542725715353)*x_ref[0]**o+(-0.388718702529)*x_ref[0]+(-0.260726707092)*x_ref[1]**o+(0.00868281860494)*x_ref[1]+(-0.422209705753)*x_ref[2]**o+(-0.482952338663)*x_ref[2]
        ref[0,0,1]=(0.517312862567)*x_ref[0]**o+(-0.213607124389)*x_ref[0]+(-0.717854605214)*x_ref[1]**o+(0.102559996175)*x_ref[1]+(-0.248146688044)*x_ref[2]**o+(-0.472262287207)*x_ref[2]
        ref[0,1,0]=(-0.561062159702)*x_ref[0]**o+(0.625784702743)*x_ref[0]+(0.127217512593)*x_ref[1]**o+(0.431860689921)*x_ref[1]+(-0.743622124184)*x_ref[2]**o+(-0.549196149668)*x_ref[2]
        ref[0,1,1]=(-0.47244041365)*x_ref[0]**o+(-0.905145373286)*x_ref[0]+(0.0327050909564)*x_ref[1]**o+(-0.0753932245132)*x_ref[1]+(-0.669475816436)*x_ref[2]**o+(-0.913342589953)*x_ref[2]
        ref[1,0,0]=(-0.107737055645)*x_ref[0]**o+(0.890187876057)*x_ref[0]+(0.619389534096)*x_ref[1]**o+(0.505636695416)*x_ref[1]+(-0.773316774174)*x_ref[2]**o+(0.57289098512)*x_ref[2]
        ref[1,0,1]=(0.998166227402)*x_ref[0]**o+(0.426101053356)*x_ref[0]+(-0.275256494275)*x_ref[1]**o+(-0.642996762298)*x_ref[1]+(0.80180314192)*x_ref[2]**o+(-0.777630700737)*x_ref[2]
        ref[1,1,0]=(0.0567929745565)*x_ref[0]**o+(-0.634357283891)*x_ref[0]+(0.756220569552)*x_ref[1]**o+(0.225394362055)*x_ref[1]+(-0.380906558144)*x_ref[2]**o+(0.656365329349)*x_ref[2]
        ref[1,1,1]=(-0.118345871482)*x_ref[0]**o+(-0.601934055233)*x_ref[0]+(-0.466776830591)*x_ref[1]**o+(0.83729192062)*x_ref[1]+(-0.760321613956)*x_ref[2]**o+(-0.0364679505353)*x_ref[2]
        ref[2,0,0]=(-0.0254736115914)*x_ref[0]**o+(0.219281883056)*x_ref[0]+(0.616921568912)*x_ref[1]**o+(0.955364895459)*x_ref[1]+(0.468886733279)*x_ref[2]**o+(-0.665678021975)*x_ref[2]
        ref[2,0,1]=(0.452552793802)*x_ref[0]**o+(0.418962651335)*x_ref[0]+(-0.493607930291)*x_ref[1]**o+(0.446553418335)*x_ref[1]+(0.297233373077)*x_ref[2]**o+(0.140710729735)*x_ref[2]
        ref[2,1,0]=(-0.636104671214)*x_ref[0]**o+(0.770136919493)*x_ref[0]+(0.559068996696)*x_ref[1]**o+(0.1863482658)*x_ref[1]+(-0.486115384192)*x_ref[2]**o+(0.766345710304)*x_ref[2]
        ref[2,1,1]=(0.365405358671)*x_ref[0]**o+(-0.349630033342)*x_ref[0]+(0.134851622297)*x_ref[1]**o+(0.274415828646)*x_ref[1]+(0.736462073823)*x_ref[2]**o+(0.113366567452)*x_ref[2]
        ref[3,0,0]=(0.103210866706)*x_ref[0]**o+(0.781535459867)*x_ref[0]+(-0.278787888162)*x_ref[1]**o+(-0.931650808802)*x_ref[1]+(0.640371737202)*x_ref[2]**o+(-0.0105202640396)*x_ref[2]
        ref[3,0,1]=(-0.249043848359)*x_ref[0]**o+(0.524828750016)*x_ref[0]+(0.533986967075)*x_ref[1]**o+(-0.267136752764)*x_ref[1]+(0.451666436163)*x_ref[2]**o+(-0.205612541545)*x_ref[2]
        ref[3,1,0]=(-0.604476259822)*x_ref[0]**o+(-0.0957734520284)*x_ref[0]+(-0.970823460222)*x_ref[1]**o+(0.980588655716)*x_ref[1]+(-0.124494413992)*x_ref[2]**o+(0.252216095396)*x_ref[2]
        ref[3,1,1]=(0.0212563582378)*x_ref[0]**o+(-0.0106688543408)*x_ref[0]+(-0.362460274705)*x_ref[1]**o+(-0.210672777607)*x_ref[1]+(0.459825480445)*x_ref[2]**o+(0.317321978801)*x_ref[2]
        ref[4,0,0]=(0.587000133255)*x_ref[0]**o+(0.362106866018)*x_ref[0]+(0.839110518219)*x_ref[1]**o+(-0.183918723198)*x_ref[1]+(-0.357506983794)*x_ref[2]**o+(0.964414772317)*x_ref[2]
        ref[4,0,1]=(-0.914924387748)*x_ref[0]**o+(-0.712924404308)*x_ref[0]+(-0.81536262433)*x_ref[1]**o+(-0.246055912409)*x_ref[1]+(0.550257872298)*x_ref[2]**o+(0.308874758093)*x_ref[2]
        ref[4,1,0]=(-0.451464088139)*x_ref[0]**o+(0.113750801605)*x_ref[0]+(-0.204501128464)*x_ref[1]**o+(0.694001185736)*x_ref[1]+(0.80368669429)*x_ref[2]**o+(0.564414553079)*x_ref[2]
        ref[4,1,1]=(-0.354511540467)*x_ref[0]**o+(0.295005960713)*x_ref[0]+(-0.581487363676)*x_ref[1]**o+(-0.862369285262)*x_ref[1]+(0.268927694418)*x_ref[2]**o+(0.91534230742)*x_ref[2]
        ref[5,0,0]=(-0.361145485775)*x_ref[0]**o+(-0.136825333824)*x_ref[0]+(-0.917163652695)*x_ref[1]**o+(-0.92225469793)*x_ref[1]+(0.142972538406)*x_ref[2]**o+(0.625895771061)*x_ref[2]
        ref[5,0,1]=(0.976797635598)*x_ref[0]**o+(-0.30046259343)*x_ref[0]+(-0.831191492174)*x_ref[1]**o+(-0.422489136999)*x_ref[1]+(0.963327796572)*x_ref[2]**o+(0.216709868453)*x_ref[2]
        ref[5,1,0]=(-0.396424567827)*x_ref[0]**o+(-0.037861045243)*x_ref[0]+(0.461627073863)*x_ref[1]**o+(0.299397593712)*x_ref[1]+(0.607619683163)*x_ref[2]**o+(-0.260194783511)*x_ref[2]
        ref[5,1,1]=(0.419140017883)*x_ref[0]**o+(0.383165340851)*x_ref[0]+(0.561345759207)*x_ref[1]**o+(-0.406706766251)*x_ref[1]+(-0.730360027768)*x_ref[2]**o+(-0.0907334422202)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.367629818255)*x[0]**o+(-0.682992553528)*x[0]+(-0.958578674565)*x[1]**o+(0.994004366609)*x[1]
        s[0,0,0,1]=(-0.933871216086)*x[0]**o+(0.089303727901)*x[0]+(0.876179784981)*x[1]**o+(-0.995418350289)*x[1]
        s[0,0,1,0]=(-0.165833461957)*x[0]**o+(0.765518912392)*x[0]+(0.435852402075)*x[1]**o+(0.508949151899)*x[1]
        s[0,0,1,1]=(0.936657616616)*x[0]**o+(-0.502803686461)*x[0]+(0.46372571939)*x[1]**o+(0.49848921833)*x[1]
        s[0,0,2,0]=(-0.637976080243)*x[0]**o+(-0.543248936311)*x[0]+(0.123903782748)*x[1]**o+(-0.775999535262)*x[1]
        s[0,0,2,1]=(-0.644671364311)*x[0]**o+(-0.843862083932)*x[0]+(-0.235872470082)*x[1]**o+(-0.976009591762)*x[1]
        s[0,1,0,0]=(-0.0549568778158)*x[0]**o+(0.585187582572)*x[0]+(0.511901547931)*x[1]**o+(0.214452316763)*x[1]
        s[0,1,0,1]=(0.904417527552)*x[0]**o+(0.120032953669)*x[0]+(-0.946416892695)*x[1]**o+(-0.283712005862)*x[1]
        s[0,1,1,0]=(0.847745098439)*x[0]**o+(-0.275082804508)*x[0]+(0.120958807786)*x[1]**o+(0.118387324002)*x[1]
        s[0,1,1,1]=(-0.504242295046)*x[0]**o+(-0.422074691957)*x[0]+(-0.1482990171)*x[1]**o+(0.0615566868579)*x[1]
        s[0,1,2,0]=(0.874047261815)*x[0]**o+(-0.13709741107)*x[0]+(-0.785816772906)*x[1]**o+(-0.677819998582)*x[1]
        s[0,1,2,1]=(-0.427567619451)*x[0]**o+(-0.27495790934)*x[0]+(0.240860007422)*x[1]**o+(-0.652639707096)*x[1]
        s[0,2,0,0]=(-0.0122530611615)*x[0]**o+(0.933497855191)*x[0]+(0.576340026858)*x[1]**o+(0.36528784459)*x[1]
        s[0,2,0,1]=(-0.536014689424)*x[0]**o+(-0.748239369266)*x[0]+(0.932275411848)*x[1]**o+(-0.610630368247)*x[1]
        s[0,2,1,0]=(-0.949632388534)*x[0]**o+(0.426052596169)*x[0]+(0.388027927797)*x[1]**o+(-0.825568754704)*x[1]
        s[0,2,1,1]=(0.647266576966)*x[0]**o+(0.558728080881)*x[0]+(-0.424360645973)*x[1]**o+(-0.946182937899)*x[1]
        s[0,2,2,0]=(0.00223655708367)*x[0]**o+(0.824278028798)*x[0]+(0.353003022728)*x[1]**o+(0.731718952112)*x[1]
        s[0,2,2,1]=(0.123272031365)*x[0]**o+(-0.51433738104)*x[0]+(-0.773965295951)*x[1]**o+(0.53173719763)*x[1]
        s[0,3,0,0]=(-0.867031991936)*x[0]**o+(-0.242455168132)*x[0]+(-0.0277817911953)*x[1]**o+(0.778617461199)*x[1]
        s[0,3,0,1]=(0.832013516177)*x[0]**o+(0.0919740140778)*x[0]+(-0.285306764452)*x[1]**o+(-0.398556244817)*x[1]
        s[0,3,1,0]=(-0.16075720255)*x[0]**o+(0.321403077687)*x[0]+(0.918586498854)*x[1]**o+(-0.0840017219508)*x[1]
        s[0,3,1,1]=(0.266578383859)*x[0]**o+(-0.255315447308)*x[0]+(-0.418644958222)*x[1]**o+(-0.552092333928)*x[1]
        s[0,3,2,0]=(-0.664300381841)*x[0]**o+(-0.903669098099)*x[0]+(0.734688383764)*x[1]**o+(-0.864082112475)*x[1]
        s[0,3,2,1]=(0.714042909464)*x[0]**o+(0.353570781531)*x[0]+(0.231167603054)*x[1]**o+(-0.478826190976)*x[1]
        s[0,4,0,0]=(-0.406664445927)*x[0]**o+(0.664845255753)*x[0]+(-0.0240782311961)*x[1]**o+(0.0609339070635)*x[1]
        s[0,4,0,1]=(0.579749599828)*x[0]**o+(0.868795370092)*x[0]+(-0.0139046093829)*x[1]**o+(-0.940717402493)*x[1]
        s[0,4,1,0]=(-0.0741351360441)*x[0]**o+(-0.063360022137)*x[0]+(0.964335931753)*x[1]**o+(-0.207493867266)*x[1]
        s[0,4,1,1]=(0.631911437012)*x[0]**o+(-0.554978524864)*x[0]+(0.790289062983)*x[1]**o+(0.432183124276)*x[1]
        s[0,4,2,0]=(0.389406102187)*x[0]**o+(0.249535881142)*x[0]+(0.0410391462706)*x[1]**o+(0.197332868001)*x[1]
        s[0,4,2,1]=(0.857250789566)*x[0]**o+(0.940933403021)*x[0]+(-0.292525450517)*x[1]**o+(-0.954295367343)*x[1]
        s[1,0,0,0]=(-0.928911526693)*x[0]**o+(0.950310526617)*x[0]+(-0.591433972944)*x[1]**o+(-0.353614169594)*x[1]
        s[1,0,0,1]=(-0.014540197205)*x[0]**o+(-0.750613108365)*x[0]+(-0.442388474744)*x[1]**o+(0.544170338222)*x[1]
        s[1,0,1,0]=(-0.313691536279)*x[0]**o+(-0.780985041598)*x[0]+(-0.323526696855)*x[1]**o+(0.96420998889)*x[1]
        s[1,0,1,1]=(0.806169224432)*x[0]**o+(-0.575250530787)*x[0]+(-0.926680195674)*x[1]**o+(0.979234114904)*x[1]
        s[1,0,2,0]=(0.316089233223)*x[0]**o+(-0.751194409767)*x[0]+(-0.224471299082)*x[1]**o+(-0.201937113732)*x[1]
        s[1,0,2,1]=(0.169596743188)*x[0]**o+(0.72246732303)*x[0]+(-0.850004427207)*x[1]**o+(0.178328119542)*x[1]
        s[1,1,0,0]=(0.878699774988)*x[0]**o+(0.665096951747)*x[0]+(0.81121004378)*x[1]**o+(-0.693898989577)*x[1]
        s[1,1,0,1]=(-0.149911779278)*x[0]**o+(-0.0473525862361)*x[0]+(-0.817244390143)*x[1]**o+(0.134983631028)*x[1]
        s[1,1,1,0]=(0.632694242262)*x[0]**o+(0.371991474102)*x[0]+(0.143801607881)*x[1]**o+(-0.539347762682)*x[1]
        s[1,1,1,1]=(-0.098791767857)*x[0]**o+(0.367846917358)*x[0]+(0.307623783461)*x[1]**o+(-0.634129908922)*x[1]
        s[1,1,2,0]=(0.790142711691)*x[0]**o+(0.384060824946)*x[0]+(-0.816682559718)*x[1]**o+(0.215076058109)*x[1]
        s[1,1,2,1]=(-0.0309558649524)*x[0]**o+(0.177277209384)*x[0]+(-0.450608986592)*x[1]**o+(-0.722601031838)*x[1]
        s[1,2,0,0]=(0.678612759518)*x[0]**o+(-0.388005661064)*x[0]+(-0.704573701308)*x[1]**o+(0.228580862805)*x[1]
        s[1,2,0,1]=(-0.982279249068)*x[0]**o+(-0.772496362129)*x[0]+(-0.871863675039)*x[1]**o+(0.924707417628)*x[1]
        s[1,2,1,0]=(-0.60771962999)*x[0]**o+(0.328917128624)*x[0]+(-0.366890766298)*x[1]**o+(0.7822593789)*x[1]
        s[1,2,1,1]=(-0.370452908148)*x[0]**o+(0.60883142941)*x[0]+(0.737450910439)*x[1]**o+(-0.250024874152)*x[1]
        s[1,2,2,0]=(-0.496193128154)*x[0]**o+(0.658353278368)*x[0]+(0.250685386258)*x[1]**o+(-0.0420890429033)*x[1]
        s[1,2,2,1]=(-0.396634397946)*x[0]**o+(0.995854445028)*x[0]+(-0.875299234489)*x[1]**o+(-0.794293605194)*x[1]
        s[1,3,0,0]=(-0.927281481375)*x[0]**o+(-0.747949026933)*x[0]+(-0.338650975226)*x[1]**o+(0.8631277432)*x[1]
        s[1,3,0,1]=(0.0643929398073)*x[0]**o+(-0.795784057297)*x[0]+(-0.77568400328)*x[1]**o+(-0.225054290337)*x[1]
        s[1,3,1,0]=(-0.636490397115)*x[0]**o+(-0.445636556924)*x[0]+(-0.906474178436)*x[1]**o+(0.380552377428)*x[1]
        s[1,3,1,1]=(0.215117725278)*x[0]**o+(-0.291445381983)*x[0]+(-0.789290110829)*x[1]**o+(0.23201632636)*x[1]
        s[1,3,2,0]=(0.26375154986)*x[0]**o+(-0.983679100192)*x[0]+(-0.210171301872)*x[1]**o+(0.773227656215)*x[1]
        s[1,3,2,1]=(0.0839267909242)*x[0]**o+(-0.805367088799)*x[0]+(0.155312230897)*x[1]**o+(-0.0808934605086)*x[1]
        s[1,4,0,0]=(0.199593117795)*x[0]**o+(0.965501409685)*x[0]+(0.323286570062)*x[1]**o+(-0.429707844705)*x[1]
        s[1,4,0,1]=(0.805178613493)*x[0]**o+(-0.853858046784)*x[0]+(-0.353638415278)*x[1]**o+(0.171890721243)*x[1]
        s[1,4,1,0]=(0.644814955839)*x[0]**o+(0.299698778855)*x[0]+(-0.88203132112)*x[1]**o+(-0.26100971325)*x[1]
        s[1,4,1,1]=(0.355086073535)*x[0]**o+(0.141140017392)*x[0]+(-0.676113152929)*x[1]**o+(0.954388216319)*x[1]
        s[1,4,2,0]=(0.766116840139)*x[0]**o+(0.00368283914615)*x[0]+(-0.889711965249)*x[1]**o+(-0.00156916153733)*x[1]
        s[1,4,2,1]=(0.501384592515)*x[0]**o+(-0.863973658641)*x[0]+(0.367327889997)*x[1]**o+(0.913819612899)*x[1]
        s[2,0,0,0]=(0.747990722914)*x[0]**o+(-0.64012379706)*x[0]+(-0.565267688923)*x[1]**o+(-0.462605252331)*x[1]
        s[2,0,0,1]=(-0.68417559629)*x[0]**o+(-0.896081416916)*x[0]+(-0.639839159011)*x[1]**o+(-0.721969231804)*x[1]
        s[2,0,1,0]=(-0.690688995555)*x[0]**o+(0.0759300115056)*x[0]+(-0.27942910767)*x[1]**o+(-0.677278037631)*x[1]
        s[2,0,1,1]=(-0.196814931401)*x[0]**o+(-0.848391602254)*x[0]+(-0.95899520009)*x[1]**o+(-0.321410974451)*x[1]
        s[2,0,2,0]=(0.360610787396)*x[0]**o+(-0.659652161288)*x[0]+(0.769472827641)*x[1]**o+(-0.272229772557)*x[1]
        s[2,0,2,1]=(-0.856112621744)*x[0]**o+(0.818082587577)*x[0]+(0.870122740873)*x[1]**o+(0.26234458071)*x[1]
        s[2,1,0,0]=(-0.820858353117)*x[0]**o+(0.815272881968)*x[0]+(0.479239358759)*x[1]**o+(-0.397781217456)*x[1]
        s[2,1,0,1]=(0.450155189682)*x[0]**o+(0.877090117209)*x[0]+(-0.704715374887)*x[1]**o+(0.839517808158)*x[1]
        s[2,1,1,0]=(0.713505879711)*x[0]**o+(0.610444859873)*x[0]+(0.36123692406)*x[1]**o+(0.0356966035771)*x[1]
        s[2,1,1,1]=(0.585841995568)*x[0]**o+(0.487958000149)*x[0]+(0.0536737460141)*x[1]**o+(0.89251806915)*x[1]
        s[2,1,2,0]=(0.122419410022)*x[0]**o+(-0.454371392861)*x[0]+(-0.508915915608)*x[1]**o+(0.182258261833)*x[1]
        s[2,1,2,1]=(-0.0541029930504)*x[0]**o+(-0.864251132735)*x[0]+(-0.726628983103)*x[1]**o+(-0.783989882996)*x[1]
        s[2,2,0,0]=(0.366784417091)*x[0]**o+(0.650637898135)*x[0]+(0.010811456998)*x[1]**o+(-0.373261079417)*x[1]
        s[2,2,0,1]=(0.147500724262)*x[0]**o+(-0.48265020035)*x[0]+(0.156030753671)*x[1]**o+(0.520269666679)*x[1]
        s[2,2,1,0]=(-0.851491519172)*x[0]**o+(-0.221010369594)*x[0]+(-0.367173559702)*x[1]**o+(0.700206278544)*x[1]
        s[2,2,1,1]=(-0.803178424841)*x[0]**o+(0.914630691651)*x[0]+(0.945631327847)*x[1]**o+(0.838368867894)*x[1]
        s[2,2,2,0]=(0.205807876192)*x[0]**o+(-0.109928074685)*x[0]+(-0.0126997666433)*x[1]**o+(0.0977593300095)*x[1]
        s[2,2,2,1]=(0.1106357035)*x[0]**o+(-0.848494903572)*x[0]+(-0.726652770048)*x[1]**o+(0.286175831783)*x[1]
        s[2,3,0,0]=(0.257704015918)*x[0]**o+(-0.890384711683)*x[0]+(-0.778573777872)*x[1]**o+(0.992701220719)*x[1]
        s[2,3,0,1]=(0.0752025499766)*x[0]**o+(0.0430345988982)*x[0]+(-0.136802110279)*x[1]**o+(-0.172056621279)*x[1]
        s[2,3,1,0]=(-0.0611919320989)*x[0]**o+(-0.785758488707)*x[0]+(0.0422134422585)*x[1]**o+(-0.225292684738)*x[1]
        s[2,3,1,1]=(0.693351228594)*x[0]**o+(0.0471475457897)*x[0]+(0.998048196978)*x[1]**o+(-0.46372201856)*x[1]
        s[2,3,2,0]=(-0.261098659889)*x[0]**o+(-0.531925602534)*x[0]+(-0.0557602486972)*x[1]**o+(-0.727790427141)*x[1]
        s[2,3,2,1]=(0.367102083636)*x[0]**o+(0.753294773683)*x[0]+(-0.408790787069)*x[1]**o+(0.860250168871)*x[1]
        s[2,4,0,0]=(-0.408802974965)*x[0]**o+(0.112012515842)*x[0]+(0.785440176371)*x[1]**o+(-0.437347103411)*x[1]
        s[2,4,0,1]=(-0.129910738114)*x[0]**o+(0.0517376480938)*x[0]+(-0.162147517056)*x[1]**o+(-0.679977272409)*x[1]
        s[2,4,1,0]=(-0.0432480639967)*x[0]**o+(0.952982408035)*x[0]+(0.14096382614)*x[1]**o+(0.155577971783)*x[1]
        s[2,4,1,1]=(-0.64723868539)*x[0]**o+(0.120062216545)*x[0]+(0.66577389931)*x[1]**o+(-0.479348695737)*x[1]
        s[2,4,2,0]=(0.971347133416)*x[0]**o+(-0.985122892501)*x[0]+(-0.646064679447)*x[1]**o+(-0.01550762144)*x[1]
        s[2,4,2,1]=(-0.644236125094)*x[0]**o+(0.735842344335)*x[0]+(0.213176548419)*x[1]**o+(-0.30985200546)*x[1]
        s[3,0,0,0]=(-0.984551843203)*x[0]**o+(0.168671646169)*x[0]+(0.349224064129)*x[1]**o+(-0.58160612166)*x[1]
        s[3,0,0,1]=(-0.812761903179)*x[0]**o+(-0.361937090398)*x[0]+(-0.24686279456)*x[1]**o+(0.710593506458)*x[1]
        s[3,0,1,0]=(-0.49188081867)*x[0]**o+(0.124030127513)*x[0]+(0.939873021882)*x[1]**o+(-0.548040221422)*x[1]
        s[3,0,1,1]=(0.872847662987)*x[0]**o+(-0.510872116798)*x[0]+(0.281463732211)*x[1]**o+(0.334088464398)*x[1]
        s[3,0,2,0]=(0.578448528377)*x[0]**o+(-0.576452752848)*x[0]+(-0.768113122135)*x[1]**o+(0.493525403618)*x[1]
        s[3,0,2,1]=(0.770913406943)*x[0]**o+(-0.420032731108)*x[0]+(-0.72069168838)*x[1]**o+(0.901782864345)*x[1]
        s[3,1,0,0]=(0.691817282198)*x[0]**o+(0.0450167655711)*x[0]+(0.40008707832)*x[1]**o+(0.326190694348)*x[1]
        s[3,1,0,1]=(-0.609859249467)*x[0]**o+(0.410229838383)*x[0]+(-0.267983779193)*x[1]**o+(0.775650294164)*x[1]
        s[3,1,1,0]=(-0.159109069553)*x[0]**o+(0.907250351987)*x[0]+(0.207780057825)*x[1]**o+(-0.342636180806)*x[1]
        s[3,1,1,1]=(-0.617610724309)*x[0]**o+(-0.78996204048)*x[0]+(-0.208417118399)*x[1]**o+(-0.423010342209)*x[1]
        s[3,1,2,0]=(-0.654973001955)*x[0]**o+(-0.731180791646)*x[0]+(-0.204094989969)*x[1]**o+(0.0168285385179)*x[1]
        s[3,1,2,1]=(0.928948644167)*x[0]**o+(0.734089325039)*x[0]+(-0.231953650601)*x[1]**o+(-0.249238293473)*x[1]
        s[3,2,0,0]=(0.157179675881)*x[0]**o+(0.410129310591)*x[0]+(0.298525358008)*x[1]**o+(-0.35693765329)*x[1]
        s[3,2,0,1]=(-0.269476743663)*x[0]**o+(-0.412585477307)*x[0]+(-0.9954766662)*x[1]**o+(0.273715511284)*x[1]
        s[3,2,1,0]=(-0.408532610441)*x[0]**o+(0.0259540763142)*x[0]+(-0.23989441562)*x[1]**o+(0.295094463976)*x[1]
        s[3,2,1,1]=(-0.110244223386)*x[0]**o+(-0.67945468919)*x[0]+(0.923264133425)*x[1]**o+(0.215905691981)*x[1]
        s[3,2,2,0]=(0.834448012095)*x[0]**o+(-0.940738696056)*x[0]+(-0.876206011085)*x[1]**o+(-0.298849473071)*x[1]
        s[3,2,2,1]=(0.475805958188)*x[0]**o+(0.621609329804)*x[0]+(-0.00780130203733)*x[1]**o+(0.888801855773)*x[1]
        s[3,3,0,0]=(-0.57606903705)*x[0]**o+(-0.226190138868)*x[0]+(0.40330410251)*x[1]**o+(-0.609371823463)*x[1]
        s[3,3,0,1]=(-0.0205092983359)*x[0]**o+(0.521468339789)*x[0]+(0.853034535096)*x[1]**o+(0.964338973709)*x[1]
        s[3,3,1,0]=(-0.272825676838)*x[0]**o+(-0.843511724287)*x[0]+(0.188117720498)*x[1]**o+(0.664721656786)*x[1]
        s[3,3,1,1]=(0.504997029501)*x[0]**o+(0.657034903786)*x[0]+(-0.198981636802)*x[1]**o+(-0.610769272074)*x[1]
        s[3,3,2,0]=(-0.118598480477)*x[0]**o+(-0.390266333689)*x[0]+(0.0363960976193)*x[1]**o+(0.0333161022002)*x[1]
        s[3,3,2,1]=(0.0560887527733)*x[0]**o+(-0.0757497077315)*x[0]+(-0.638185285374)*x[1]**o+(-0.0772348567365)*x[1]
        s[3,4,0,0]=(-0.77162241293)*x[0]**o+(-0.638997804963)*x[0]+(-0.103979693129)*x[1]**o+(-0.403273375409)*x[1]
        s[3,4,0,1]=(-0.835975178768)*x[0]**o+(-0.776970415353)*x[0]+(-0.123032313988)*x[1]**o+(-0.402259255094)*x[1]
        s[3,4,1,0]=(0.391513520306)*x[0]**o+(0.751880253565)*x[0]+(-0.59575974702)*x[1]**o+(-0.478354879315)*x[1]
        s[3,4,1,1]=(-0.890188758697)*x[0]**o+(0.635319326122)*x[0]+(0.796942175081)*x[1]**o+(-0.325193620429)*x[1]
        s[3,4,2,0]=(-0.906513454158)*x[0]**o+(0.599411990631)*x[0]+(-0.0273060032343)*x[1]**o+(-0.0720151635364)*x[1]
        s[3,4,2,1]=(-0.00512989795909)*x[0]**o+(-0.622272404387)*x[0]+(0.453748642854)*x[1]**o+(-0.261040385951)*x[1]
        ref[0,0,0,0]=(0.367629818255)*x_ref[0]**o+(-0.682992553528)*x_ref[0]+(-0.958578674565)*x_ref[1]**o+(0.994004366609)*x_ref[1]
        ref[0,0,0,1]=(-0.933871216086)*x_ref[0]**o+(0.089303727901)*x_ref[0]+(0.876179784981)*x_ref[1]**o+(-0.995418350289)*x_ref[1]
        ref[0,0,1,0]=(-0.165833461957)*x_ref[0]**o+(0.765518912392)*x_ref[0]+(0.435852402075)*x_ref[1]**o+(0.508949151899)*x_ref[1]
        ref[0,0,1,1]=(0.936657616616)*x_ref[0]**o+(-0.502803686461)*x_ref[0]+(0.46372571939)*x_ref[1]**o+(0.49848921833)*x_ref[1]
        ref[0,0,2,0]=(-0.637976080243)*x_ref[0]**o+(-0.543248936311)*x_ref[0]+(0.123903782748)*x_ref[1]**o+(-0.775999535262)*x_ref[1]
        ref[0,0,2,1]=(-0.644671364311)*x_ref[0]**o+(-0.843862083932)*x_ref[0]+(-0.235872470082)*x_ref[1]**o+(-0.976009591762)*x_ref[1]
        ref[0,1,0,0]=(-0.0549568778158)*x_ref[0]**o+(0.585187582572)*x_ref[0]+(0.511901547931)*x_ref[1]**o+(0.214452316763)*x_ref[1]
        ref[0,1,0,1]=(0.904417527552)*x_ref[0]**o+(0.120032953669)*x_ref[0]+(-0.946416892695)*x_ref[1]**o+(-0.283712005862)*x_ref[1]
        ref[0,1,1,0]=(0.847745098439)*x_ref[0]**o+(-0.275082804508)*x_ref[0]+(0.120958807786)*x_ref[1]**o+(0.118387324002)*x_ref[1]
        ref[0,1,1,1]=(-0.504242295046)*x_ref[0]**o+(-0.422074691957)*x_ref[0]+(-0.1482990171)*x_ref[1]**o+(0.0615566868579)*x_ref[1]
        ref[0,1,2,0]=(0.874047261815)*x_ref[0]**o+(-0.13709741107)*x_ref[0]+(-0.785816772906)*x_ref[1]**o+(-0.677819998582)*x_ref[1]
        ref[0,1,2,1]=(-0.427567619451)*x_ref[0]**o+(-0.27495790934)*x_ref[0]+(0.240860007422)*x_ref[1]**o+(-0.652639707096)*x_ref[1]
        ref[0,2,0,0]=(-0.0122530611615)*x_ref[0]**o+(0.933497855191)*x_ref[0]+(0.576340026858)*x_ref[1]**o+(0.36528784459)*x_ref[1]
        ref[0,2,0,1]=(-0.536014689424)*x_ref[0]**o+(-0.748239369266)*x_ref[0]+(0.932275411848)*x_ref[1]**o+(-0.610630368247)*x_ref[1]
        ref[0,2,1,0]=(-0.949632388534)*x_ref[0]**o+(0.426052596169)*x_ref[0]+(0.388027927797)*x_ref[1]**o+(-0.825568754704)*x_ref[1]
        ref[0,2,1,1]=(0.647266576966)*x_ref[0]**o+(0.558728080881)*x_ref[0]+(-0.424360645973)*x_ref[1]**o+(-0.946182937899)*x_ref[1]
        ref[0,2,2,0]=(0.00223655708367)*x_ref[0]**o+(0.824278028798)*x_ref[0]+(0.353003022728)*x_ref[1]**o+(0.731718952112)*x_ref[1]
        ref[0,2,2,1]=(0.123272031365)*x_ref[0]**o+(-0.51433738104)*x_ref[0]+(-0.773965295951)*x_ref[1]**o+(0.53173719763)*x_ref[1]
        ref[0,3,0,0]=(-0.867031991936)*x_ref[0]**o+(-0.242455168132)*x_ref[0]+(-0.0277817911953)*x_ref[1]**o+(0.778617461199)*x_ref[1]
        ref[0,3,0,1]=(0.832013516177)*x_ref[0]**o+(0.0919740140778)*x_ref[0]+(-0.285306764452)*x_ref[1]**o+(-0.398556244817)*x_ref[1]
        ref[0,3,1,0]=(-0.16075720255)*x_ref[0]**o+(0.321403077687)*x_ref[0]+(0.918586498854)*x_ref[1]**o+(-0.0840017219508)*x_ref[1]
        ref[0,3,1,1]=(0.266578383859)*x_ref[0]**o+(-0.255315447308)*x_ref[0]+(-0.418644958222)*x_ref[1]**o+(-0.552092333928)*x_ref[1]
        ref[0,3,2,0]=(-0.664300381841)*x_ref[0]**o+(-0.903669098099)*x_ref[0]+(0.734688383764)*x_ref[1]**o+(-0.864082112475)*x_ref[1]
        ref[0,3,2,1]=(0.714042909464)*x_ref[0]**o+(0.353570781531)*x_ref[0]+(0.231167603054)*x_ref[1]**o+(-0.478826190976)*x_ref[1]
        ref[0,4,0,0]=(-0.406664445927)*x_ref[0]**o+(0.664845255753)*x_ref[0]+(-0.0240782311961)*x_ref[1]**o+(0.0609339070635)*x_ref[1]
        ref[0,4,0,1]=(0.579749599828)*x_ref[0]**o+(0.868795370092)*x_ref[0]+(-0.0139046093829)*x_ref[1]**o+(-0.940717402493)*x_ref[1]
        ref[0,4,1,0]=(-0.0741351360441)*x_ref[0]**o+(-0.063360022137)*x_ref[0]+(0.964335931753)*x_ref[1]**o+(-0.207493867266)*x_ref[1]
        ref[0,4,1,1]=(0.631911437012)*x_ref[0]**o+(-0.554978524864)*x_ref[0]+(0.790289062983)*x_ref[1]**o+(0.432183124276)*x_ref[1]
        ref[0,4,2,0]=(0.389406102187)*x_ref[0]**o+(0.249535881142)*x_ref[0]+(0.0410391462706)*x_ref[1]**o+(0.197332868001)*x_ref[1]
        ref[0,4,2,1]=(0.857250789566)*x_ref[0]**o+(0.940933403021)*x_ref[0]+(-0.292525450517)*x_ref[1]**o+(-0.954295367343)*x_ref[1]
        ref[1,0,0,0]=(-0.928911526693)*x_ref[0]**o+(0.950310526617)*x_ref[0]+(-0.591433972944)*x_ref[1]**o+(-0.353614169594)*x_ref[1]
        ref[1,0,0,1]=(-0.014540197205)*x_ref[0]**o+(-0.750613108365)*x_ref[0]+(-0.442388474744)*x_ref[1]**o+(0.544170338222)*x_ref[1]
        ref[1,0,1,0]=(-0.313691536279)*x_ref[0]**o+(-0.780985041598)*x_ref[0]+(-0.323526696855)*x_ref[1]**o+(0.96420998889)*x_ref[1]
        ref[1,0,1,1]=(0.806169224432)*x_ref[0]**o+(-0.575250530787)*x_ref[0]+(-0.926680195674)*x_ref[1]**o+(0.979234114904)*x_ref[1]
        ref[1,0,2,0]=(0.316089233223)*x_ref[0]**o+(-0.751194409767)*x_ref[0]+(-0.224471299082)*x_ref[1]**o+(-0.201937113732)*x_ref[1]
        ref[1,0,2,1]=(0.169596743188)*x_ref[0]**o+(0.72246732303)*x_ref[0]+(-0.850004427207)*x_ref[1]**o+(0.178328119542)*x_ref[1]
        ref[1,1,0,0]=(0.878699774988)*x_ref[0]**o+(0.665096951747)*x_ref[0]+(0.81121004378)*x_ref[1]**o+(-0.693898989577)*x_ref[1]
        ref[1,1,0,1]=(-0.149911779278)*x_ref[0]**o+(-0.0473525862361)*x_ref[0]+(-0.817244390143)*x_ref[1]**o+(0.134983631028)*x_ref[1]
        ref[1,1,1,0]=(0.632694242262)*x_ref[0]**o+(0.371991474102)*x_ref[0]+(0.143801607881)*x_ref[1]**o+(-0.539347762682)*x_ref[1]
        ref[1,1,1,1]=(-0.098791767857)*x_ref[0]**o+(0.367846917358)*x_ref[0]+(0.307623783461)*x_ref[1]**o+(-0.634129908922)*x_ref[1]
        ref[1,1,2,0]=(0.790142711691)*x_ref[0]**o+(0.384060824946)*x_ref[0]+(-0.816682559718)*x_ref[1]**o+(0.215076058109)*x_ref[1]
        ref[1,1,2,1]=(-0.0309558649524)*x_ref[0]**o+(0.177277209384)*x_ref[0]+(-0.450608986592)*x_ref[1]**o+(-0.722601031838)*x_ref[1]
        ref[1,2,0,0]=(0.678612759518)*x_ref[0]**o+(-0.388005661064)*x_ref[0]+(-0.704573701308)*x_ref[1]**o+(0.228580862805)*x_ref[1]
        ref[1,2,0,1]=(-0.982279249068)*x_ref[0]**o+(-0.772496362129)*x_ref[0]+(-0.871863675039)*x_ref[1]**o+(0.924707417628)*x_ref[1]
        ref[1,2,1,0]=(-0.60771962999)*x_ref[0]**o+(0.328917128624)*x_ref[0]+(-0.366890766298)*x_ref[1]**o+(0.7822593789)*x_ref[1]
        ref[1,2,1,1]=(-0.370452908148)*x_ref[0]**o+(0.60883142941)*x_ref[0]+(0.737450910439)*x_ref[1]**o+(-0.250024874152)*x_ref[1]
        ref[1,2,2,0]=(-0.496193128154)*x_ref[0]**o+(0.658353278368)*x_ref[0]+(0.250685386258)*x_ref[1]**o+(-0.0420890429033)*x_ref[1]
        ref[1,2,2,1]=(-0.396634397946)*x_ref[0]**o+(0.995854445028)*x_ref[0]+(-0.875299234489)*x_ref[1]**o+(-0.794293605194)*x_ref[1]
        ref[1,3,0,0]=(-0.927281481375)*x_ref[0]**o+(-0.747949026933)*x_ref[0]+(-0.338650975226)*x_ref[1]**o+(0.8631277432)*x_ref[1]
        ref[1,3,0,1]=(0.0643929398073)*x_ref[0]**o+(-0.795784057297)*x_ref[0]+(-0.77568400328)*x_ref[1]**o+(-0.225054290337)*x_ref[1]
        ref[1,3,1,0]=(-0.636490397115)*x_ref[0]**o+(-0.445636556924)*x_ref[0]+(-0.906474178436)*x_ref[1]**o+(0.380552377428)*x_ref[1]
        ref[1,3,1,1]=(0.215117725278)*x_ref[0]**o+(-0.291445381983)*x_ref[0]+(-0.789290110829)*x_ref[1]**o+(0.23201632636)*x_ref[1]
        ref[1,3,2,0]=(0.26375154986)*x_ref[0]**o+(-0.983679100192)*x_ref[0]+(-0.210171301872)*x_ref[1]**o+(0.773227656215)*x_ref[1]
        ref[1,3,2,1]=(0.0839267909242)*x_ref[0]**o+(-0.805367088799)*x_ref[0]+(0.155312230897)*x_ref[1]**o+(-0.0808934605086)*x_ref[1]
        ref[1,4,0,0]=(0.199593117795)*x_ref[0]**o+(0.965501409685)*x_ref[0]+(0.323286570062)*x_ref[1]**o+(-0.429707844705)*x_ref[1]
        ref[1,4,0,1]=(0.805178613493)*x_ref[0]**o+(-0.853858046784)*x_ref[0]+(-0.353638415278)*x_ref[1]**o+(0.171890721243)*x_ref[1]
        ref[1,4,1,0]=(0.644814955839)*x_ref[0]**o+(0.299698778855)*x_ref[0]+(-0.88203132112)*x_ref[1]**o+(-0.26100971325)*x_ref[1]
        ref[1,4,1,1]=(0.355086073535)*x_ref[0]**o+(0.141140017392)*x_ref[0]+(-0.676113152929)*x_ref[1]**o+(0.954388216319)*x_ref[1]
        ref[1,4,2,0]=(0.766116840139)*x_ref[0]**o+(0.00368283914615)*x_ref[0]+(-0.889711965249)*x_ref[1]**o+(-0.00156916153733)*x_ref[1]
        ref[1,4,2,1]=(0.501384592515)*x_ref[0]**o+(-0.863973658641)*x_ref[0]+(0.367327889997)*x_ref[1]**o+(0.913819612899)*x_ref[1]
        ref[2,0,0,0]=(0.747990722914)*x_ref[0]**o+(-0.64012379706)*x_ref[0]+(-0.565267688923)*x_ref[1]**o+(-0.462605252331)*x_ref[1]
        ref[2,0,0,1]=(-0.68417559629)*x_ref[0]**o+(-0.896081416916)*x_ref[0]+(-0.639839159011)*x_ref[1]**o+(-0.721969231804)*x_ref[1]
        ref[2,0,1,0]=(-0.690688995555)*x_ref[0]**o+(0.0759300115056)*x_ref[0]+(-0.27942910767)*x_ref[1]**o+(-0.677278037631)*x_ref[1]
        ref[2,0,1,1]=(-0.196814931401)*x_ref[0]**o+(-0.848391602254)*x_ref[0]+(-0.95899520009)*x_ref[1]**o+(-0.321410974451)*x_ref[1]
        ref[2,0,2,0]=(0.360610787396)*x_ref[0]**o+(-0.659652161288)*x_ref[0]+(0.769472827641)*x_ref[1]**o+(-0.272229772557)*x_ref[1]
        ref[2,0,2,1]=(-0.856112621744)*x_ref[0]**o+(0.818082587577)*x_ref[0]+(0.870122740873)*x_ref[1]**o+(0.26234458071)*x_ref[1]
        ref[2,1,0,0]=(-0.820858353117)*x_ref[0]**o+(0.815272881968)*x_ref[0]+(0.479239358759)*x_ref[1]**o+(-0.397781217456)*x_ref[1]
        ref[2,1,0,1]=(0.450155189682)*x_ref[0]**o+(0.877090117209)*x_ref[0]+(-0.704715374887)*x_ref[1]**o+(0.839517808158)*x_ref[1]
        ref[2,1,1,0]=(0.713505879711)*x_ref[0]**o+(0.610444859873)*x_ref[0]+(0.36123692406)*x_ref[1]**o+(0.0356966035771)*x_ref[1]
        ref[2,1,1,1]=(0.585841995568)*x_ref[0]**o+(0.487958000149)*x_ref[0]+(0.0536737460141)*x_ref[1]**o+(0.89251806915)*x_ref[1]
        ref[2,1,2,0]=(0.122419410022)*x_ref[0]**o+(-0.454371392861)*x_ref[0]+(-0.508915915608)*x_ref[1]**o+(0.182258261833)*x_ref[1]
        ref[2,1,2,1]=(-0.0541029930504)*x_ref[0]**o+(-0.864251132735)*x_ref[0]+(-0.726628983103)*x_ref[1]**o+(-0.783989882996)*x_ref[1]
        ref[2,2,0,0]=(0.366784417091)*x_ref[0]**o+(0.650637898135)*x_ref[0]+(0.010811456998)*x_ref[1]**o+(-0.373261079417)*x_ref[1]
        ref[2,2,0,1]=(0.147500724262)*x_ref[0]**o+(-0.48265020035)*x_ref[0]+(0.156030753671)*x_ref[1]**o+(0.520269666679)*x_ref[1]
        ref[2,2,1,0]=(-0.851491519172)*x_ref[0]**o+(-0.221010369594)*x_ref[0]+(-0.367173559702)*x_ref[1]**o+(0.700206278544)*x_ref[1]
        ref[2,2,1,1]=(-0.803178424841)*x_ref[0]**o+(0.914630691651)*x_ref[0]+(0.945631327847)*x_ref[1]**o+(0.838368867894)*x_ref[1]
        ref[2,2,2,0]=(0.205807876192)*x_ref[0]**o+(-0.109928074685)*x_ref[0]+(-0.0126997666433)*x_ref[1]**o+(0.0977593300095)*x_ref[1]
        ref[2,2,2,1]=(0.1106357035)*x_ref[0]**o+(-0.848494903572)*x_ref[0]+(-0.726652770048)*x_ref[1]**o+(0.286175831783)*x_ref[1]
        ref[2,3,0,0]=(0.257704015918)*x_ref[0]**o+(-0.890384711683)*x_ref[0]+(-0.778573777872)*x_ref[1]**o+(0.992701220719)*x_ref[1]
        ref[2,3,0,1]=(0.0752025499766)*x_ref[0]**o+(0.0430345988982)*x_ref[0]+(-0.136802110279)*x_ref[1]**o+(-0.172056621279)*x_ref[1]
        ref[2,3,1,0]=(-0.0611919320989)*x_ref[0]**o+(-0.785758488707)*x_ref[0]+(0.0422134422585)*x_ref[1]**o+(-0.225292684738)*x_ref[1]
        ref[2,3,1,1]=(0.693351228594)*x_ref[0]**o+(0.0471475457897)*x_ref[0]+(0.998048196978)*x_ref[1]**o+(-0.46372201856)*x_ref[1]
        ref[2,3,2,0]=(-0.261098659889)*x_ref[0]**o+(-0.531925602534)*x_ref[0]+(-0.0557602486972)*x_ref[1]**o+(-0.727790427141)*x_ref[1]
        ref[2,3,2,1]=(0.367102083636)*x_ref[0]**o+(0.753294773683)*x_ref[0]+(-0.408790787069)*x_ref[1]**o+(0.860250168871)*x_ref[1]
        ref[2,4,0,0]=(-0.408802974965)*x_ref[0]**o+(0.112012515842)*x_ref[0]+(0.785440176371)*x_ref[1]**o+(-0.437347103411)*x_ref[1]
        ref[2,4,0,1]=(-0.129910738114)*x_ref[0]**o+(0.0517376480938)*x_ref[0]+(-0.162147517056)*x_ref[1]**o+(-0.679977272409)*x_ref[1]
        ref[2,4,1,0]=(-0.0432480639967)*x_ref[0]**o+(0.952982408035)*x_ref[0]+(0.14096382614)*x_ref[1]**o+(0.155577971783)*x_ref[1]
        ref[2,4,1,1]=(-0.64723868539)*x_ref[0]**o+(0.120062216545)*x_ref[0]+(0.66577389931)*x_ref[1]**o+(-0.479348695737)*x_ref[1]
        ref[2,4,2,0]=(0.971347133416)*x_ref[0]**o+(-0.985122892501)*x_ref[0]+(-0.646064679447)*x_ref[1]**o+(-0.01550762144)*x_ref[1]
        ref[2,4,2,1]=(-0.644236125094)*x_ref[0]**o+(0.735842344335)*x_ref[0]+(0.213176548419)*x_ref[1]**o+(-0.30985200546)*x_ref[1]
        ref[3,0,0,0]=(-0.984551843203)*x_ref[0]**o+(0.168671646169)*x_ref[0]+(0.349224064129)*x_ref[1]**o+(-0.58160612166)*x_ref[1]
        ref[3,0,0,1]=(-0.812761903179)*x_ref[0]**o+(-0.361937090398)*x_ref[0]+(-0.24686279456)*x_ref[1]**o+(0.710593506458)*x_ref[1]
        ref[3,0,1,0]=(-0.49188081867)*x_ref[0]**o+(0.124030127513)*x_ref[0]+(0.939873021882)*x_ref[1]**o+(-0.548040221422)*x_ref[1]
        ref[3,0,1,1]=(0.872847662987)*x_ref[0]**o+(-0.510872116798)*x_ref[0]+(0.281463732211)*x_ref[1]**o+(0.334088464398)*x_ref[1]
        ref[3,0,2,0]=(0.578448528377)*x_ref[0]**o+(-0.576452752848)*x_ref[0]+(-0.768113122135)*x_ref[1]**o+(0.493525403618)*x_ref[1]
        ref[3,0,2,1]=(0.770913406943)*x_ref[0]**o+(-0.420032731108)*x_ref[0]+(-0.72069168838)*x_ref[1]**o+(0.901782864345)*x_ref[1]
        ref[3,1,0,0]=(0.691817282198)*x_ref[0]**o+(0.0450167655711)*x_ref[0]+(0.40008707832)*x_ref[1]**o+(0.326190694348)*x_ref[1]
        ref[3,1,0,1]=(-0.609859249467)*x_ref[0]**o+(0.410229838383)*x_ref[0]+(-0.267983779193)*x_ref[1]**o+(0.775650294164)*x_ref[1]
        ref[3,1,1,0]=(-0.159109069553)*x_ref[0]**o+(0.907250351987)*x_ref[0]+(0.207780057825)*x_ref[1]**o+(-0.342636180806)*x_ref[1]
        ref[3,1,1,1]=(-0.617610724309)*x_ref[0]**o+(-0.78996204048)*x_ref[0]+(-0.208417118399)*x_ref[1]**o+(-0.423010342209)*x_ref[1]
        ref[3,1,2,0]=(-0.654973001955)*x_ref[0]**o+(-0.731180791646)*x_ref[0]+(-0.204094989969)*x_ref[1]**o+(0.0168285385179)*x_ref[1]
        ref[3,1,2,1]=(0.928948644167)*x_ref[0]**o+(0.734089325039)*x_ref[0]+(-0.231953650601)*x_ref[1]**o+(-0.249238293473)*x_ref[1]
        ref[3,2,0,0]=(0.157179675881)*x_ref[0]**o+(0.410129310591)*x_ref[0]+(0.298525358008)*x_ref[1]**o+(-0.35693765329)*x_ref[1]
        ref[3,2,0,1]=(-0.269476743663)*x_ref[0]**o+(-0.412585477307)*x_ref[0]+(-0.9954766662)*x_ref[1]**o+(0.273715511284)*x_ref[1]
        ref[3,2,1,0]=(-0.408532610441)*x_ref[0]**o+(0.0259540763142)*x_ref[0]+(-0.23989441562)*x_ref[1]**o+(0.295094463976)*x_ref[1]
        ref[3,2,1,1]=(-0.110244223386)*x_ref[0]**o+(-0.67945468919)*x_ref[0]+(0.923264133425)*x_ref[1]**o+(0.215905691981)*x_ref[1]
        ref[3,2,2,0]=(0.834448012095)*x_ref[0]**o+(-0.940738696056)*x_ref[0]+(-0.876206011085)*x_ref[1]**o+(-0.298849473071)*x_ref[1]
        ref[3,2,2,1]=(0.475805958188)*x_ref[0]**o+(0.621609329804)*x_ref[0]+(-0.00780130203733)*x_ref[1]**o+(0.888801855773)*x_ref[1]
        ref[3,3,0,0]=(-0.57606903705)*x_ref[0]**o+(-0.226190138868)*x_ref[0]+(0.40330410251)*x_ref[1]**o+(-0.609371823463)*x_ref[1]
        ref[3,3,0,1]=(-0.0205092983359)*x_ref[0]**o+(0.521468339789)*x_ref[0]+(0.853034535096)*x_ref[1]**o+(0.964338973709)*x_ref[1]
        ref[3,3,1,0]=(-0.272825676838)*x_ref[0]**o+(-0.843511724287)*x_ref[0]+(0.188117720498)*x_ref[1]**o+(0.664721656786)*x_ref[1]
        ref[3,3,1,1]=(0.504997029501)*x_ref[0]**o+(0.657034903786)*x_ref[0]+(-0.198981636802)*x_ref[1]**o+(-0.610769272074)*x_ref[1]
        ref[3,3,2,0]=(-0.118598480477)*x_ref[0]**o+(-0.390266333689)*x_ref[0]+(0.0363960976193)*x_ref[1]**o+(0.0333161022002)*x_ref[1]
        ref[3,3,2,1]=(0.0560887527733)*x_ref[0]**o+(-0.0757497077315)*x_ref[0]+(-0.638185285374)*x_ref[1]**o+(-0.0772348567365)*x_ref[1]
        ref[3,4,0,0]=(-0.77162241293)*x_ref[0]**o+(-0.638997804963)*x_ref[0]+(-0.103979693129)*x_ref[1]**o+(-0.403273375409)*x_ref[1]
        ref[3,4,0,1]=(-0.835975178768)*x_ref[0]**o+(-0.776970415353)*x_ref[0]+(-0.123032313988)*x_ref[1]**o+(-0.402259255094)*x_ref[1]
        ref[3,4,1,0]=(0.391513520306)*x_ref[0]**o+(0.751880253565)*x_ref[0]+(-0.59575974702)*x_ref[1]**o+(-0.478354879315)*x_ref[1]
        ref[3,4,1,1]=(-0.890188758697)*x_ref[0]**o+(0.635319326122)*x_ref[0]+(0.796942175081)*x_ref[1]**o+(-0.325193620429)*x_ref[1]
        ref[3,4,2,0]=(-0.906513454158)*x_ref[0]**o+(0.599411990631)*x_ref[0]+(-0.0273060032343)*x_ref[1]**o+(-0.0720151635364)*x_ref[1]
        ref[3,4,2,1]=(-0.00512989795909)*x_ref[0]**o+(-0.622272404387)*x_ref[0]+(0.453748642854)*x_ref[1]**o+(-0.261040385951)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.173295097983)*x[0]**o+(0.773632210562)*x[0]+(-0.521496418821)*x[1]**o+(-0.947907473579)*x[1]+(0.847721381317)*x[2]**o+(-0.887851963083)*x[2]
        s[0,0,0,1]=(0.0654213598408)*x[0]**o+(0.763489939538)*x[0]+(-0.176800713154)*x[1]**o+(-0.602428137248)*x[1]+(-0.851802251531)*x[2]**o+(-0.763084376785)*x[2]
        s[0,0,1,0]=(0.30505329426)*x[0]**o+(-0.378878190786)*x[0]+(-0.615554910302)*x[1]**o+(0.0555314463878)*x[1]+(-0.063211109707)*x[2]**o+(-0.623580981446)*x[2]
        s[0,0,1,1]=(-0.526220589068)*x[0]**o+(-0.551655771219)*x[0]+(0.163011887243)*x[1]**o+(-0.0863220304472)*x[1]+(-0.638708196977)*x[2]**o+(-0.298073090554)*x[2]
        s[0,0,2,0]=(0.0698028632972)*x[0]**o+(0.508381168266)*x[0]+(-0.120587433211)*x[1]**o+(-0.419454430666)*x[1]+(0.389702566269)*x[2]**o+(-0.169655820731)*x[2]
        s[0,0,2,1]=(-0.80612077313)*x[0]**o+(0.640104753466)*x[0]+(0.518635274506)*x[1]**o+(-0.545279092505)*x[1]+(0.195002020672)*x[2]**o+(0.563537057084)*x[2]
        s[0,1,0,0]=(-0.613471743331)*x[0]**o+(-0.274063857866)*x[0]+(-0.259665265722)*x[1]**o+(0.0830136895696)*x[1]+(0.825880322476)*x[2]**o+(0.133039077085)*x[2]
        s[0,1,0,1]=(0.727956094813)*x[0]**o+(0.605313602734)*x[0]+(0.708549587523)*x[1]**o+(-0.964527272447)*x[1]+(-0.740107990066)*x[2]**o+(0.629263961477)*x[2]
        s[0,1,1,0]=(-0.70107143852)*x[0]**o+(-0.055344456929)*x[0]+(-0.954082414306)*x[1]**o+(-0.702136485683)*x[1]+(-0.801391066649)*x[2]**o+(-0.905767030094)*x[2]
        s[0,1,1,1]=(0.67172040238)*x[0]**o+(0.740486014051)*x[0]+(-0.335682691861)*x[1]**o+(0.0793015412254)*x[1]+(-0.868582520395)*x[2]**o+(0.0848453101016)*x[2]
        s[0,1,2,0]=(0.224190760234)*x[0]**o+(0.139996595496)*x[0]+(-0.519770806543)*x[1]**o+(-0.683808752006)*x[1]+(-0.837082787693)*x[2]**o+(-0.613900014511)*x[2]
        s[0,1,2,1]=(0.213979011641)*x[0]**o+(-0.309663935547)*x[0]+(0.0211525822739)*x[1]**o+(0.844506291028)*x[1]+(-0.688933639701)*x[2]**o+(-0.336614773412)*x[2]
        s[0,2,0,0]=(-0.0188626152353)*x[0]**o+(-0.694756450245)*x[0]+(0.960579615684)*x[1]**o+(0.57325819325)*x[1]+(0.917037597322)*x[2]**o+(-0.0584252009816)*x[2]
        s[0,2,0,1]=(0.803966493511)*x[0]**o+(-0.169885689634)*x[0]+(-0.556021959109)*x[1]**o+(0.273901944163)*x[1]+(-0.737708052821)*x[2]**o+(0.627870713104)*x[2]
        s[0,2,1,0]=(-0.511561329117)*x[0]**o+(-0.0854308669951)*x[0]+(0.464246331007)*x[1]**o+(-0.97461601243)*x[1]+(-0.223895747162)*x[2]**o+(-0.385389479659)*x[2]
        s[0,2,1,1]=(-0.723508721341)*x[0]**o+(-0.201904055697)*x[0]+(0.0789759340837)*x[1]**o+(-0.00248269457662)*x[1]+(-0.0626643900787)*x[2]**o+(-0.128428965973)*x[2]
        s[0,2,2,0]=(0.934647998503)*x[0]**o+(-0.0922965345918)*x[0]+(0.717876178289)*x[1]**o+(0.344875905273)*x[1]+(0.0410512922166)*x[2]**o+(0.164242310844)*x[2]
        s[0,2,2,1]=(-0.699306080605)*x[0]**o+(0.209729161919)*x[0]+(0.820435958574)*x[1]**o+(-0.173302359677)*x[1]+(0.659952656349)*x[2]**o+(-0.78057378087)*x[2]
        s[0,3,0,0]=(-0.495298694962)*x[0]**o+(0.571367877972)*x[0]+(0.931242429977)*x[1]**o+(0.888499081573)*x[1]+(-0.408432694783)*x[2]**o+(-0.898594385103)*x[2]
        s[0,3,0,1]=(-0.96889210815)*x[0]**o+(0.404624542424)*x[0]+(0.380170350345)*x[1]**o+(-0.557523650891)*x[1]+(-0.566134006)*x[2]**o+(0.74677094551)*x[2]
        s[0,3,1,0]=(-0.224311775759)*x[0]**o+(-0.660099304736)*x[0]+(-0.513720022847)*x[1]**o+(0.534948126324)*x[1]+(0.162072633001)*x[2]**o+(-0.402416011787)*x[2]
        s[0,3,1,1]=(0.224350718771)*x[0]**o+(0.283343383933)*x[0]+(0.138291871569)*x[1]**o+(0.543289554343)*x[1]+(-0.844288550977)*x[2]**o+(0.463476015453)*x[2]
        s[0,3,2,0]=(-0.685488143712)*x[0]**o+(0.300413146898)*x[0]+(0.748708379841)*x[1]**o+(-0.598212294807)*x[1]+(0.479982541942)*x[2]**o+(0.23866536682)*x[2]
        s[0,3,2,1]=(0.940674112182)*x[0]**o+(-0.35846015272)*x[0]+(-0.770304297672)*x[1]**o+(-0.73897352885)*x[1]+(-0.682102723348)*x[2]**o+(0.742528100693)*x[2]
        s[0,4,0,0]=(0.0162066911802)*x[0]**o+(0.160336749821)*x[0]+(0.121948909833)*x[1]**o+(-0.102093937457)*x[1]+(-0.768661339258)*x[2]**o+(0.98805683085)*x[2]
        s[0,4,0,1]=(-0.420990735691)*x[0]**o+(-0.500244152083)*x[0]+(0.910641832988)*x[1]**o+(0.979658885966)*x[1]+(-0.946590371476)*x[2]**o+(-0.190997339007)*x[2]
        s[0,4,1,0]=(-0.445779943656)*x[0]**o+(-0.378857349604)*x[0]+(0.125534672956)*x[1]**o+(0.124248627713)*x[1]+(-0.685767125932)*x[2]**o+(0.643604950542)*x[2]
        s[0,4,1,1]=(0.518152834346)*x[0]**o+(-0.891873054984)*x[0]+(-0.648744206994)*x[1]**o+(0.207049340927)*x[1]+(0.198529322067)*x[2]**o+(-0.814832030848)*x[2]
        s[0,4,2,0]=(0.477174707789)*x[0]**o+(-0.724452116357)*x[0]+(0.209008883429)*x[1]**o+(0.79968814697)*x[1]+(0.129144596117)*x[2]**o+(0.214098340906)*x[2]
        s[0,4,2,1]=(0.117050329845)*x[0]**o+(-0.721574040514)*x[0]+(0.382424906743)*x[1]**o+(-0.135392966957)*x[1]+(0.24455410794)*x[2]**o+(-0.679165459966)*x[2]
        s[1,0,0,0]=(-0.180925772343)*x[0]**o+(0.664151970586)*x[0]+(0.0496840383663)*x[1]**o+(-0.899468244962)*x[1]+(0.942251196146)*x[2]**o+(0.228837829842)*x[2]
        s[1,0,0,1]=(0.15346023848)*x[0]**o+(-0.642882884964)*x[0]+(-0.462925865669)*x[1]**o+(0.738240095188)*x[1]+(-0.57213723904)*x[2]**o+(-0.900335371347)*x[2]
        s[1,0,1,0]=(-0.366392792085)*x[0]**o+(-0.227326930302)*x[0]+(-0.323111857481)*x[1]**o+(0.220826082736)*x[1]+(0.237722021225)*x[2]**o+(0.835402283453)*x[2]
        s[1,0,1,1]=(-0.293188444394)*x[0]**o+(0.338812067406)*x[0]+(0.679884979413)*x[1]**o+(-0.614317813911)*x[1]+(0.936906665544)*x[2]**o+(-0.652679743405)*x[2]
        s[1,0,2,0]=(0.617621972453)*x[0]**o+(-0.936326816619)*x[0]+(0.760134028716)*x[1]**o+(-0.809380984591)*x[1]+(0.562612935213)*x[2]**o+(0.0794516002575)*x[2]
        s[1,0,2,1]=(0.448514601948)*x[0]**o+(0.327970604314)*x[0]+(-0.936492917037)*x[1]**o+(-0.425126227791)*x[1]+(-0.372116400796)*x[2]**o+(-0.958054327266)*x[2]
        s[1,1,0,0]=(-0.850336943937)*x[0]**o+(0.379951917933)*x[0]+(0.693148856845)*x[1]**o+(-0.691233930311)*x[1]+(0.548624260329)*x[2]**o+(-0.193125714718)*x[2]
        s[1,1,0,1]=(-0.823007972929)*x[0]**o+(0.408191529362)*x[0]+(-0.023921849403)*x[1]**o+(0.79830304115)*x[1]+(0.506756293069)*x[2]**o+(0.434230444707)*x[2]
        s[1,1,1,0]=(0.263932235445)*x[0]**o+(-0.600679211293)*x[0]+(0.480835197573)*x[1]**o+(-0.976282735826)*x[1]+(0.714583200236)*x[2]**o+(0.942610909813)*x[2]
        s[1,1,1,1]=(0.481454842664)*x[0]**o+(-0.508309985427)*x[0]+(-0.739763965139)*x[1]**o+(0.211871147037)*x[1]+(-0.848812488744)*x[2]**o+(0.885973013107)*x[2]
        s[1,1,2,0]=(-0.796437949787)*x[0]**o+(-0.211960906694)*x[0]+(0.823389343923)*x[1]**o+(-0.0162189137972)*x[1]+(0.77124757744)*x[2]**o+(0.877807214589)*x[2]
        s[1,1,2,1]=(0.899446895743)*x[0]**o+(-0.303223847939)*x[0]+(0.409199240049)*x[1]**o+(0.430938116282)*x[1]+(0.189267561328)*x[2]**o+(0.582418236586)*x[2]
        s[1,2,0,0]=(0.687732048304)*x[0]**o+(0.338825649717)*x[0]+(-0.535870442333)*x[1]**o+(0.507773000137)*x[1]+(-0.892023180429)*x[2]**o+(0.35147003979)*x[2]
        s[1,2,0,1]=(-0.315183497372)*x[0]**o+(-0.0302851860468)*x[0]+(-0.966617816351)*x[1]**o+(-0.876528463134)*x[1]+(-0.682249112679)*x[2]**o+(-0.830991774314)*x[2]
        s[1,2,1,0]=(0.90817845095)*x[0]**o+(-0.629542958154)*x[0]+(0.663791954021)*x[1]**o+(0.584246811118)*x[1]+(0.421582877787)*x[2]**o+(0.316731433829)*x[2]
        s[1,2,1,1]=(-0.948554425084)*x[0]**o+(-0.365749706899)*x[0]+(0.282031925475)*x[1]**o+(0.467581708634)*x[1]+(0.158633292573)*x[2]**o+(-0.0990441468359)*x[2]
        s[1,2,2,0]=(0.113170413925)*x[0]**o+(0.481722903666)*x[0]+(0.0127374268921)*x[1]**o+(-0.778057043706)*x[1]+(0.399544836507)*x[2]**o+(-0.157225042768)*x[2]
        s[1,2,2,1]=(0.320987735687)*x[0]**o+(-0.29324177158)*x[0]+(0.02800800296)*x[1]**o+(-0.277948207921)*x[1]+(-0.271808717111)*x[2]**o+(-0.0379931737995)*x[2]
        s[1,3,0,0]=(0.09942034907)*x[0]**o+(0.939629344394)*x[0]+(-0.810686196185)*x[1]**o+(-0.538230260705)*x[1]+(0.85351760208)*x[2]**o+(-0.737196688288)*x[2]
        s[1,3,0,1]=(0.48690360509)*x[0]**o+(-0.065370723503)*x[0]+(-0.0566373715183)*x[1]**o+(0.871878739279)*x[1]+(0.272054032769)*x[2]**o+(0.540626473697)*x[2]
        s[1,3,1,0]=(0.472815976006)*x[0]**o+(-0.775416335945)*x[0]+(0.213646821631)*x[1]**o+(0.415546582713)*x[1]+(-0.196674759494)*x[2]**o+(0.235475790794)*x[2]
        s[1,3,1,1]=(-0.699232847611)*x[0]**o+(0.504961844211)*x[0]+(-0.0115303749449)*x[1]**o+(-0.756286148171)*x[1]+(-0.0922447811444)*x[2]**o+(-0.330281180023)*x[2]
        s[1,3,2,0]=(0.149323550569)*x[0]**o+(0.977735452052)*x[0]+(0.255972870292)*x[1]**o+(-0.855635543118)*x[1]+(0.519602177962)*x[2]**o+(0.922189470945)*x[2]
        s[1,3,2,1]=(-0.190749790276)*x[0]**o+(-0.351506874986)*x[0]+(-0.0672783552454)*x[1]**o+(-0.223773986529)*x[1]+(-0.593750649495)*x[2]**o+(0.158607153203)*x[2]
        s[1,4,0,0]=(0.245306614552)*x[0]**o+(-0.191799968697)*x[0]+(0.85455961157)*x[1]**o+(0.597426986918)*x[1]+(-0.824363631294)*x[2]**o+(-0.0161601323913)*x[2]
        s[1,4,0,1]=(0.925974209919)*x[0]**o+(-0.606855935989)*x[0]+(0.182619376779)*x[1]**o+(0.82202480925)*x[1]+(0.0511805106413)*x[2]**o+(0.0866309784619)*x[2]
        s[1,4,1,0]=(0.272918227857)*x[0]**o+(-0.406306617661)*x[0]+(-0.0535134980185)*x[1]**o+(-0.186200126868)*x[1]+(0.92546875085)*x[2]**o+(-0.474811198949)*x[2]
        s[1,4,1,1]=(0.803708551531)*x[0]**o+(-0.715626841623)*x[0]+(-0.0822999615139)*x[1]**o+(0.612334309398)*x[1]+(-0.157577434512)*x[2]**o+(-0.795469427996)*x[2]
        s[1,4,2,0]=(-0.837159054411)*x[0]**o+(-0.234523360065)*x[0]+(-0.198337862205)*x[1]**o+(0.897958649268)*x[1]+(-0.519907680436)*x[2]**o+(0.742526415023)*x[2]
        s[1,4,2,1]=(-0.792448458791)*x[0]**o+(0.456487739729)*x[0]+(0.0128609741891)*x[1]**o+(-0.344964414689)*x[1]+(-0.205584911868)*x[2]**o+(0.429183240889)*x[2]
        s[2,0,0,0]=(0.401501024951)*x[0]**o+(-0.497039379761)*x[0]+(-0.228675343279)*x[1]**o+(0.133323657822)*x[1]+(-0.203006510606)*x[2]**o+(0.407718693225)*x[2]
        s[2,0,0,1]=(-0.547448204328)*x[0]**o+(-0.109789709109)*x[0]+(-0.626633454424)*x[1]**o+(-0.117871748454)*x[1]+(0.961790647477)*x[2]**o+(0.652883936373)*x[2]
        s[2,0,1,0]=(-0.85536471584)*x[0]**o+(0.489291116489)*x[0]+(-0.929596786116)*x[1]**o+(0.0110017367519)*x[1]+(-0.703617460934)*x[2]**o+(0.0496869388223)*x[2]
        s[2,0,1,1]=(-0.501093953991)*x[0]**o+(-0.722768363488)*x[0]+(0.84017017736)*x[1]**o+(-0.22900603012)*x[1]+(0.760492299454)*x[2]**o+(0.767346260882)*x[2]
        s[2,0,2,0]=(-0.328498451765)*x[0]**o+(-0.387534616444)*x[0]+(-0.0973382392242)*x[1]**o+(0.418582689709)*x[1]+(0.206439339821)*x[2]**o+(-0.706353598324)*x[2]
        s[2,0,2,1]=(0.179977568909)*x[0]**o+(0.642598753874)*x[0]+(0.711020603016)*x[1]**o+(-0.287193672421)*x[1]+(-0.362612184622)*x[2]**o+(-0.18146063868)*x[2]
        s[2,1,0,0]=(0.0470678392348)*x[0]**o+(0.00649068845495)*x[0]+(-0.239912084684)*x[1]**o+(0.278051515796)*x[1]+(-0.98924625944)*x[2]**o+(0.847957009386)*x[2]
        s[2,1,0,1]=(-0.287954694237)*x[0]**o+(0.127947128577)*x[0]+(0.443869903193)*x[1]**o+(-0.344694028625)*x[1]+(-0.14884506191)*x[2]**o+(0.748140862905)*x[2]
        s[2,1,1,0]=(-0.597041891422)*x[0]**o+(0.914716550666)*x[0]+(0.68496200279)*x[1]**o+(0.0420126536598)*x[1]+(-0.539966353171)*x[2]**o+(-0.886373035297)*x[2]
        s[2,1,1,1]=(0.462776858955)*x[0]**o+(-0.989367011955)*x[0]+(0.421691081662)*x[1]**o+(0.264858346124)*x[1]+(0.513954688709)*x[2]**o+(0.00863898872886)*x[2]
        s[2,1,2,0]=(0.503601863243)*x[0]**o+(0.505973913611)*x[0]+(0.0952160099415)*x[1]**o+(-0.77103310327)*x[1]+(-0.316507531754)*x[2]**o+(0.506969348695)*x[2]
        s[2,1,2,1]=(0.521634307558)*x[0]**o+(-0.910583030486)*x[0]+(0.977891014192)*x[1]**o+(0.985145579284)*x[1]+(0.0465919834269)*x[2]**o+(0.113946013883)*x[2]
        s[2,2,0,0]=(0.178722112454)*x[0]**o+(0.582800628485)*x[0]+(0.141950810923)*x[1]**o+(0.226899624421)*x[1]+(0.445681888363)*x[2]**o+(-0.127689044751)*x[2]
        s[2,2,0,1]=(-0.433214095841)*x[0]**o+(-0.61199374178)*x[0]+(0.655706389409)*x[1]**o+(-0.0160036542096)*x[1]+(0.570017878158)*x[2]**o+(-0.00718475090704)*x[2]
        s[2,2,1,0]=(-0.205396523704)*x[0]**o+(0.87306641582)*x[0]+(-0.664912090364)*x[1]**o+(-0.324577426693)*x[1]+(-0.37269275568)*x[2]**o+(-0.0753237226649)*x[2]
        s[2,2,1,1]=(0.716795494369)*x[0]**o+(-0.725543593985)*x[0]+(0.751101352157)*x[1]**o+(0.0629757438809)*x[1]+(-0.388859722245)*x[2]**o+(0.633579629088)*x[2]
        s[2,2,2,0]=(-0.785119558914)*x[0]**o+(-0.289212761766)*x[0]+(-0.947517483982)*x[1]**o+(-0.898908873087)*x[1]+(-0.758807720992)*x[2]**o+(-0.747003629416)*x[2]
        s[2,2,2,1]=(0.384610894358)*x[0]**o+(0.0170687884025)*x[0]+(0.776211872471)*x[1]**o+(-0.313821704113)*x[1]+(-0.433021717398)*x[2]**o+(-0.972385486304)*x[2]
        s[2,3,0,0]=(0.59307540612)*x[0]**o+(0.93544260405)*x[0]+(0.519016668994)*x[1]**o+(-0.175290607112)*x[1]+(-0.269720511811)*x[2]**o+(0.117237249214)*x[2]
        s[2,3,0,1]=(-0.0752366310333)*x[0]**o+(-0.168792429308)*x[0]+(0.316236130047)*x[1]**o+(0.021991183167)*x[1]+(-0.2024837663)*x[2]**o+(-0.676832860181)*x[2]
        s[2,3,1,0]=(-0.2682564416)*x[0]**o+(0.276503108458)*x[0]+(0.789470125022)*x[1]**o+(-0.339044117352)*x[1]+(-0.408507243588)*x[2]**o+(0.0666267661149)*x[2]
        s[2,3,1,1]=(-0.053395293244)*x[0]**o+(0.425880490109)*x[0]+(-0.068128758144)*x[1]**o+(0.413635486374)*x[1]+(0.965050652205)*x[2]**o+(0.126502830836)*x[2]
        s[2,3,2,0]=(0.601928913545)*x[0]**o+(-0.0443247453956)*x[0]+(-0.606811664164)*x[1]**o+(0.0206686600296)*x[1]+(-0.804806810474)*x[2]**o+(-0.019745599298)*x[2]
        s[2,3,2,1]=(0.911324514835)*x[0]**o+(-0.397270365997)*x[0]+(0.511154806396)*x[1]**o+(0.722003975461)*x[1]+(-0.332922355553)*x[2]**o+(-0.132309188597)*x[2]
        s[2,4,0,0]=(0.194698324972)*x[0]**o+(-0.496022175033)*x[0]+(0.804006600349)*x[1]**o+(-0.643493478792)*x[1]+(0.148395455552)*x[2]**o+(-0.408862596804)*x[2]
        s[2,4,0,1]=(0.734891277828)*x[0]**o+(-0.101225311176)*x[0]+(-0.813960759076)*x[1]**o+(0.862888786761)*x[1]+(0.830408706284)*x[2]**o+(0.523449046413)*x[2]
        s[2,4,1,0]=(0.286629494089)*x[0]**o+(0.599214589464)*x[0]+(-0.208652251419)*x[1]**o+(-0.830028802843)*x[1]+(0.805360345419)*x[2]**o+(-0.865349771687)*x[2]
        s[2,4,1,1]=(0.0367955096711)*x[0]**o+(-0.0261723558157)*x[0]+(0.200900558315)*x[1]**o+(-0.532346409176)*x[1]+(-0.183663881543)*x[2]**o+(0.93191706611)*x[2]
        s[2,4,2,0]=(-0.932223437809)*x[0]**o+(-0.155278731606)*x[0]+(-0.932519381895)*x[1]**o+(0.0589900357358)*x[1]+(-0.187551917707)*x[2]**o+(-0.81649531847)*x[2]
        s[2,4,2,1]=(0.555481188864)*x[0]**o+(0.946358167786)*x[0]+(-0.231784522)*x[1]**o+(0.592776954307)*x[1]+(-0.0373202061188)*x[2]**o+(0.0302770826388)*x[2]
        s[3,0,0,0]=(0.603747717227)*x[0]**o+(-0.969329597018)*x[0]+(0.311876345528)*x[1]**o+(-0.589486433983)*x[1]+(-0.294507657675)*x[2]**o+(-0.743511183116)*x[2]
        s[3,0,0,1]=(-0.846568198319)*x[0]**o+(0.93521569126)*x[0]+(-0.874618479052)*x[1]**o+(0.502042421273)*x[1]+(-0.945100352471)*x[2]**o+(0.765854746295)*x[2]
        s[3,0,1,0]=(-0.908353314852)*x[0]**o+(0.701990826648)*x[0]+(0.798549198275)*x[1]**o+(-0.616631523859)*x[1]+(-0.999800863738)*x[2]**o+(-0.156585259768)*x[2]
        s[3,0,1,1]=(-0.433258077523)*x[0]**o+(-0.641049552618)*x[0]+(0.427903347848)*x[1]**o+(-0.175290793023)*x[1]+(-0.557154490598)*x[2]**o+(-0.00495777825998)*x[2]
        s[3,0,2,0]=(0.998253378112)*x[0]**o+(0.2484283851)*x[0]+(0.955897183936)*x[1]**o+(-0.0351483480831)*x[1]+(0.690211608431)*x[2]**o+(0.7420822406)*x[2]
        s[3,0,2,1]=(0.273153627728)*x[0]**o+(0.931572979903)*x[0]+(-0.279174117029)*x[1]**o+(-0.741239387871)*x[1]+(0.148372377955)*x[2]**o+(0.781476364189)*x[2]
        s[3,1,0,0]=(0.231049869286)*x[0]**o+(-0.871461717509)*x[0]+(0.553545032585)*x[1]**o+(0.979019482224)*x[1]+(0.819288655429)*x[2]**o+(-0.01489981549)*x[2]
        s[3,1,0,1]=(-0.891391000148)*x[0]**o+(-0.360048424965)*x[0]+(-0.390043258734)*x[1]**o+(0.547316173383)*x[1]+(-0.972094018922)*x[2]**o+(0.7798960782)*x[2]
        s[3,1,1,0]=(0.675405837097)*x[0]**o+(0.976689341648)*x[0]+(-0.416251051342)*x[1]**o+(0.696304358663)*x[1]+(-0.395743544878)*x[2]**o+(-0.0340340734736)*x[2]
        s[3,1,1,1]=(-0.531871504573)*x[0]**o+(-0.926770430235)*x[0]+(0.752552585571)*x[1]**o+(-0.403402334727)*x[1]+(0.07593559269)*x[2]**o+(0.126742096076)*x[2]
        s[3,1,2,0]=(-0.033528429617)*x[0]**o+(0.187766565831)*x[0]+(0.192128207394)*x[1]**o+(-0.0969373324914)*x[1]+(-0.889231964389)*x[2]**o+(0.493198117275)*x[2]
        s[3,1,2,1]=(-0.182267751454)*x[0]**o+(-0.115192609603)*x[0]+(-0.252701809463)*x[1]**o+(0.355949096242)*x[1]+(0.859478772116)*x[2]**o+(0.539090466518)*x[2]
        s[3,2,0,0]=(0.25232058799)*x[0]**o+(-0.246680227462)*x[0]+(0.316105270194)*x[1]**o+(-0.426505186439)*x[1]+(0.980259506048)*x[2]**o+(0.144514677615)*x[2]
        s[3,2,0,1]=(-0.583071409136)*x[0]**o+(0.969922333041)*x[0]+(-0.0251030152687)*x[1]**o+(0.121967009007)*x[1]+(0.391834007838)*x[2]**o+(0.0339488845613)*x[2]
        s[3,2,1,0]=(0.706904708517)*x[0]**o+(0.266141083673)*x[0]+(-0.0339160692688)*x[1]**o+(-0.152115678673)*x[1]+(0.122213919626)*x[2]**o+(-0.576478442144)*x[2]
        s[3,2,1,1]=(-0.917141077151)*x[0]**o+(0.569284998999)*x[0]+(-0.0979622311673)*x[1]**o+(0.554915758725)*x[1]+(0.824303662531)*x[2]**o+(-0.449201813108)*x[2]
        s[3,2,2,0]=(-0.284000203352)*x[0]**o+(0.0535558865246)*x[0]+(-0.233670301273)*x[1]**o+(0.383324533244)*x[1]+(-0.483402264684)*x[2]**o+(-0.482290332534)*x[2]
        s[3,2,2,1]=(-0.557106835416)*x[0]**o+(-0.0263124485334)*x[0]+(-0.132921973706)*x[1]**o+(0.0351808996763)*x[1]+(0.979819304163)*x[2]**o+(-0.908131098335)*x[2]
        s[3,3,0,0]=(0.982856633407)*x[0]**o+(-0.180985287571)*x[0]+(0.364546381679)*x[1]**o+(0.746225949823)*x[1]+(0.847682125532)*x[2]**o+(-0.782701591901)*x[2]
        s[3,3,0,1]=(-0.33362442802)*x[0]**o+(-0.933319602818)*x[0]+(-0.830312620251)*x[1]**o+(-0.938028393959)*x[1]+(0.859281853158)*x[2]**o+(0.430732343672)*x[2]
        s[3,3,1,0]=(0.381562874557)*x[0]**o+(0.0939156358041)*x[0]+(0.207361800369)*x[1]**o+(0.115172135617)*x[1]+(0.747494745689)*x[2]**o+(-0.45448955026)*x[2]
        s[3,3,1,1]=(0.095679833076)*x[0]**o+(-0.936886715289)*x[0]+(0.887146346873)*x[1]**o+(0.159246683343)*x[1]+(0.817874529811)*x[2]**o+(0.516091968965)*x[2]
        s[3,3,2,0]=(-0.476953971316)*x[0]**o+(0.176188538145)*x[0]+(0.568224935975)*x[1]**o+(0.56192429612)*x[1]+(-0.700911152919)*x[2]**o+(-0.0114437684935)*x[2]
        s[3,3,2,1]=(0.29541245686)*x[0]**o+(0.546841142855)*x[0]+(-0.23081431627)*x[1]**o+(-0.633697713725)*x[1]+(0.33519295746)*x[2]**o+(-0.951229146931)*x[2]
        s[3,4,0,0]=(-0.879148775346)*x[0]**o+(-0.703893190615)*x[0]+(0.409831208498)*x[1]**o+(0.754374287841)*x[1]+(0.688380721157)*x[2]**o+(-0.395460537774)*x[2]
        s[3,4,0,1]=(0.924489816348)*x[0]**o+(0.999127631985)*x[0]+(-0.892638544934)*x[1]**o+(-0.923913790342)*x[1]+(0.736487445972)*x[2]**o+(0.250131270052)*x[2]
        s[3,4,1,0]=(0.264892665874)*x[0]**o+(-0.776154012834)*x[0]+(-0.379853719079)*x[1]**o+(-0.748063746312)*x[1]+(-0.0108974578949)*x[2]**o+(-0.803449911818)*x[2]
        s[3,4,1,1]=(-0.699932530879)*x[0]**o+(-0.901128866521)*x[0]+(0.628561941772)*x[1]**o+(0.507625196603)*x[1]+(0.147922829864)*x[2]**o+(0.475783034469)*x[2]
        s[3,4,2,0]=(0.482139761317)*x[0]**o+(-0.627593411974)*x[0]+(-0.84668128017)*x[1]**o+(-0.324236153907)*x[1]+(-0.688406716909)*x[2]**o+(-0.0669177294519)*x[2]
        s[3,4,2,1]=(-0.531896625188)*x[0]**o+(-0.574459691455)*x[0]+(-0.824941329118)*x[1]**o+(0.0812437772861)*x[1]+(-0.257954477449)*x[2]**o+(-0.219142537545)*x[2]
        ref[0,0,0,0]=(-0.173295097983)*x_ref[0]**o+(0.773632210562)*x_ref[0]+(-0.521496418821)*x_ref[1]**o+(-0.947907473579)*x_ref[1]+(0.847721381317)*x_ref[2]**o+(-0.887851963083)*x_ref[2]
        ref[0,0,0,1]=(0.0654213598408)*x_ref[0]**o+(0.763489939538)*x_ref[0]+(-0.176800713154)*x_ref[1]**o+(-0.602428137248)*x_ref[1]+(-0.851802251531)*x_ref[2]**o+(-0.763084376785)*x_ref[2]
        ref[0,0,1,0]=(0.30505329426)*x_ref[0]**o+(-0.378878190786)*x_ref[0]+(-0.615554910302)*x_ref[1]**o+(0.0555314463878)*x_ref[1]+(-0.063211109707)*x_ref[2]**o+(-0.623580981446)*x_ref[2]
        ref[0,0,1,1]=(-0.526220589068)*x_ref[0]**o+(-0.551655771219)*x_ref[0]+(0.163011887243)*x_ref[1]**o+(-0.0863220304472)*x_ref[1]+(-0.638708196977)*x_ref[2]**o+(-0.298073090554)*x_ref[2]
        ref[0,0,2,0]=(0.0698028632972)*x_ref[0]**o+(0.508381168266)*x_ref[0]+(-0.120587433211)*x_ref[1]**o+(-0.419454430666)*x_ref[1]+(0.389702566269)*x_ref[2]**o+(-0.169655820731)*x_ref[2]
        ref[0,0,2,1]=(-0.80612077313)*x_ref[0]**o+(0.640104753466)*x_ref[0]+(0.518635274506)*x_ref[1]**o+(-0.545279092505)*x_ref[1]+(0.195002020672)*x_ref[2]**o+(0.563537057084)*x_ref[2]
        ref[0,1,0,0]=(-0.613471743331)*x_ref[0]**o+(-0.274063857866)*x_ref[0]+(-0.259665265722)*x_ref[1]**o+(0.0830136895696)*x_ref[1]+(0.825880322476)*x_ref[2]**o+(0.133039077085)*x_ref[2]
        ref[0,1,0,1]=(0.727956094813)*x_ref[0]**o+(0.605313602734)*x_ref[0]+(0.708549587523)*x_ref[1]**o+(-0.964527272447)*x_ref[1]+(-0.740107990066)*x_ref[2]**o+(0.629263961477)*x_ref[2]
        ref[0,1,1,0]=(-0.70107143852)*x_ref[0]**o+(-0.055344456929)*x_ref[0]+(-0.954082414306)*x_ref[1]**o+(-0.702136485683)*x_ref[1]+(-0.801391066649)*x_ref[2]**o+(-0.905767030094)*x_ref[2]
        ref[0,1,1,1]=(0.67172040238)*x_ref[0]**o+(0.740486014051)*x_ref[0]+(-0.335682691861)*x_ref[1]**o+(0.0793015412254)*x_ref[1]+(-0.868582520395)*x_ref[2]**o+(0.0848453101016)*x_ref[2]
        ref[0,1,2,0]=(0.224190760234)*x_ref[0]**o+(0.139996595496)*x_ref[0]+(-0.519770806543)*x_ref[1]**o+(-0.683808752006)*x_ref[1]+(-0.837082787693)*x_ref[2]**o+(-0.613900014511)*x_ref[2]
        ref[0,1,2,1]=(0.213979011641)*x_ref[0]**o+(-0.309663935547)*x_ref[0]+(0.0211525822739)*x_ref[1]**o+(0.844506291028)*x_ref[1]+(-0.688933639701)*x_ref[2]**o+(-0.336614773412)*x_ref[2]
        ref[0,2,0,0]=(-0.0188626152353)*x_ref[0]**o+(-0.694756450245)*x_ref[0]+(0.960579615684)*x_ref[1]**o+(0.57325819325)*x_ref[1]+(0.917037597322)*x_ref[2]**o+(-0.0584252009816)*x_ref[2]
        ref[0,2,0,1]=(0.803966493511)*x_ref[0]**o+(-0.169885689634)*x_ref[0]+(-0.556021959109)*x_ref[1]**o+(0.273901944163)*x_ref[1]+(-0.737708052821)*x_ref[2]**o+(0.627870713104)*x_ref[2]
        ref[0,2,1,0]=(-0.511561329117)*x_ref[0]**o+(-0.0854308669951)*x_ref[0]+(0.464246331007)*x_ref[1]**o+(-0.97461601243)*x_ref[1]+(-0.223895747162)*x_ref[2]**o+(-0.385389479659)*x_ref[2]
        ref[0,2,1,1]=(-0.723508721341)*x_ref[0]**o+(-0.201904055697)*x_ref[0]+(0.0789759340837)*x_ref[1]**o+(-0.00248269457662)*x_ref[1]+(-0.0626643900787)*x_ref[2]**o+(-0.128428965973)*x_ref[2]
        ref[0,2,2,0]=(0.934647998503)*x_ref[0]**o+(-0.0922965345918)*x_ref[0]+(0.717876178289)*x_ref[1]**o+(0.344875905273)*x_ref[1]+(0.0410512922166)*x_ref[2]**o+(0.164242310844)*x_ref[2]
        ref[0,2,2,1]=(-0.699306080605)*x_ref[0]**o+(0.209729161919)*x_ref[0]+(0.820435958574)*x_ref[1]**o+(-0.173302359677)*x_ref[1]+(0.659952656349)*x_ref[2]**o+(-0.78057378087)*x_ref[2]
        ref[0,3,0,0]=(-0.495298694962)*x_ref[0]**o+(0.571367877972)*x_ref[0]+(0.931242429977)*x_ref[1]**o+(0.888499081573)*x_ref[1]+(-0.408432694783)*x_ref[2]**o+(-0.898594385103)*x_ref[2]
        ref[0,3,0,1]=(-0.96889210815)*x_ref[0]**o+(0.404624542424)*x_ref[0]+(0.380170350345)*x_ref[1]**o+(-0.557523650891)*x_ref[1]+(-0.566134006)*x_ref[2]**o+(0.74677094551)*x_ref[2]
        ref[0,3,1,0]=(-0.224311775759)*x_ref[0]**o+(-0.660099304736)*x_ref[0]+(-0.513720022847)*x_ref[1]**o+(0.534948126324)*x_ref[1]+(0.162072633001)*x_ref[2]**o+(-0.402416011787)*x_ref[2]
        ref[0,3,1,1]=(0.224350718771)*x_ref[0]**o+(0.283343383933)*x_ref[0]+(0.138291871569)*x_ref[1]**o+(0.543289554343)*x_ref[1]+(-0.844288550977)*x_ref[2]**o+(0.463476015453)*x_ref[2]
        ref[0,3,2,0]=(-0.685488143712)*x_ref[0]**o+(0.300413146898)*x_ref[0]+(0.748708379841)*x_ref[1]**o+(-0.598212294807)*x_ref[1]+(0.479982541942)*x_ref[2]**o+(0.23866536682)*x_ref[2]
        ref[0,3,2,1]=(0.940674112182)*x_ref[0]**o+(-0.35846015272)*x_ref[0]+(-0.770304297672)*x_ref[1]**o+(-0.73897352885)*x_ref[1]+(-0.682102723348)*x_ref[2]**o+(0.742528100693)*x_ref[2]
        ref[0,4,0,0]=(0.0162066911802)*x_ref[0]**o+(0.160336749821)*x_ref[0]+(0.121948909833)*x_ref[1]**o+(-0.102093937457)*x_ref[1]+(-0.768661339258)*x_ref[2]**o+(0.98805683085)*x_ref[2]
        ref[0,4,0,1]=(-0.420990735691)*x_ref[0]**o+(-0.500244152083)*x_ref[0]+(0.910641832988)*x_ref[1]**o+(0.979658885966)*x_ref[1]+(-0.946590371476)*x_ref[2]**o+(-0.190997339007)*x_ref[2]
        ref[0,4,1,0]=(-0.445779943656)*x_ref[0]**o+(-0.378857349604)*x_ref[0]+(0.125534672956)*x_ref[1]**o+(0.124248627713)*x_ref[1]+(-0.685767125932)*x_ref[2]**o+(0.643604950542)*x_ref[2]
        ref[0,4,1,1]=(0.518152834346)*x_ref[0]**o+(-0.891873054984)*x_ref[0]+(-0.648744206994)*x_ref[1]**o+(0.207049340927)*x_ref[1]+(0.198529322067)*x_ref[2]**o+(-0.814832030848)*x_ref[2]
        ref[0,4,2,0]=(0.477174707789)*x_ref[0]**o+(-0.724452116357)*x_ref[0]+(0.209008883429)*x_ref[1]**o+(0.79968814697)*x_ref[1]+(0.129144596117)*x_ref[2]**o+(0.214098340906)*x_ref[2]
        ref[0,4,2,1]=(0.117050329845)*x_ref[0]**o+(-0.721574040514)*x_ref[0]+(0.382424906743)*x_ref[1]**o+(-0.135392966957)*x_ref[1]+(0.24455410794)*x_ref[2]**o+(-0.679165459966)*x_ref[2]
        ref[1,0,0,0]=(-0.180925772343)*x_ref[0]**o+(0.664151970586)*x_ref[0]+(0.0496840383663)*x_ref[1]**o+(-0.899468244962)*x_ref[1]+(0.942251196146)*x_ref[2]**o+(0.228837829842)*x_ref[2]
        ref[1,0,0,1]=(0.15346023848)*x_ref[0]**o+(-0.642882884964)*x_ref[0]+(-0.462925865669)*x_ref[1]**o+(0.738240095188)*x_ref[1]+(-0.57213723904)*x_ref[2]**o+(-0.900335371347)*x_ref[2]
        ref[1,0,1,0]=(-0.366392792085)*x_ref[0]**o+(-0.227326930302)*x_ref[0]+(-0.323111857481)*x_ref[1]**o+(0.220826082736)*x_ref[1]+(0.237722021225)*x_ref[2]**o+(0.835402283453)*x_ref[2]
        ref[1,0,1,1]=(-0.293188444394)*x_ref[0]**o+(0.338812067406)*x_ref[0]+(0.679884979413)*x_ref[1]**o+(-0.614317813911)*x_ref[1]+(0.936906665544)*x_ref[2]**o+(-0.652679743405)*x_ref[2]
        ref[1,0,2,0]=(0.617621972453)*x_ref[0]**o+(-0.936326816619)*x_ref[0]+(0.760134028716)*x_ref[1]**o+(-0.809380984591)*x_ref[1]+(0.562612935213)*x_ref[2]**o+(0.0794516002575)*x_ref[2]
        ref[1,0,2,1]=(0.448514601948)*x_ref[0]**o+(0.327970604314)*x_ref[0]+(-0.936492917037)*x_ref[1]**o+(-0.425126227791)*x_ref[1]+(-0.372116400796)*x_ref[2]**o+(-0.958054327266)*x_ref[2]
        ref[1,1,0,0]=(-0.850336943937)*x_ref[0]**o+(0.379951917933)*x_ref[0]+(0.693148856845)*x_ref[1]**o+(-0.691233930311)*x_ref[1]+(0.548624260329)*x_ref[2]**o+(-0.193125714718)*x_ref[2]
        ref[1,1,0,1]=(-0.823007972929)*x_ref[0]**o+(0.408191529362)*x_ref[0]+(-0.023921849403)*x_ref[1]**o+(0.79830304115)*x_ref[1]+(0.506756293069)*x_ref[2]**o+(0.434230444707)*x_ref[2]
        ref[1,1,1,0]=(0.263932235445)*x_ref[0]**o+(-0.600679211293)*x_ref[0]+(0.480835197573)*x_ref[1]**o+(-0.976282735826)*x_ref[1]+(0.714583200236)*x_ref[2]**o+(0.942610909813)*x_ref[2]
        ref[1,1,1,1]=(0.481454842664)*x_ref[0]**o+(-0.508309985427)*x_ref[0]+(-0.739763965139)*x_ref[1]**o+(0.211871147037)*x_ref[1]+(-0.848812488744)*x_ref[2]**o+(0.885973013107)*x_ref[2]
        ref[1,1,2,0]=(-0.796437949787)*x_ref[0]**o+(-0.211960906694)*x_ref[0]+(0.823389343923)*x_ref[1]**o+(-0.0162189137972)*x_ref[1]+(0.77124757744)*x_ref[2]**o+(0.877807214589)*x_ref[2]
        ref[1,1,2,1]=(0.899446895743)*x_ref[0]**o+(-0.303223847939)*x_ref[0]+(0.409199240049)*x_ref[1]**o+(0.430938116282)*x_ref[1]+(0.189267561328)*x_ref[2]**o+(0.582418236586)*x_ref[2]
        ref[1,2,0,0]=(0.687732048304)*x_ref[0]**o+(0.338825649717)*x_ref[0]+(-0.535870442333)*x_ref[1]**o+(0.507773000137)*x_ref[1]+(-0.892023180429)*x_ref[2]**o+(0.35147003979)*x_ref[2]
        ref[1,2,0,1]=(-0.315183497372)*x_ref[0]**o+(-0.0302851860468)*x_ref[0]+(-0.966617816351)*x_ref[1]**o+(-0.876528463134)*x_ref[1]+(-0.682249112679)*x_ref[2]**o+(-0.830991774314)*x_ref[2]
        ref[1,2,1,0]=(0.90817845095)*x_ref[0]**o+(-0.629542958154)*x_ref[0]+(0.663791954021)*x_ref[1]**o+(0.584246811118)*x_ref[1]+(0.421582877787)*x_ref[2]**o+(0.316731433829)*x_ref[2]
        ref[1,2,1,1]=(-0.948554425084)*x_ref[0]**o+(-0.365749706899)*x_ref[0]+(0.282031925475)*x_ref[1]**o+(0.467581708634)*x_ref[1]+(0.158633292573)*x_ref[2]**o+(-0.0990441468359)*x_ref[2]
        ref[1,2,2,0]=(0.113170413925)*x_ref[0]**o+(0.481722903666)*x_ref[0]+(0.0127374268921)*x_ref[1]**o+(-0.778057043706)*x_ref[1]+(0.399544836507)*x_ref[2]**o+(-0.157225042768)*x_ref[2]
        ref[1,2,2,1]=(0.320987735687)*x_ref[0]**o+(-0.29324177158)*x_ref[0]+(0.02800800296)*x_ref[1]**o+(-0.277948207921)*x_ref[1]+(-0.271808717111)*x_ref[2]**o+(-0.0379931737995)*x_ref[2]
        ref[1,3,0,0]=(0.09942034907)*x_ref[0]**o+(0.939629344394)*x_ref[0]+(-0.810686196185)*x_ref[1]**o+(-0.538230260705)*x_ref[1]+(0.85351760208)*x_ref[2]**o+(-0.737196688288)*x_ref[2]
        ref[1,3,0,1]=(0.48690360509)*x_ref[0]**o+(-0.065370723503)*x_ref[0]+(-0.0566373715183)*x_ref[1]**o+(0.871878739279)*x_ref[1]+(0.272054032769)*x_ref[2]**o+(0.540626473697)*x_ref[2]
        ref[1,3,1,0]=(0.472815976006)*x_ref[0]**o+(-0.775416335945)*x_ref[0]+(0.213646821631)*x_ref[1]**o+(0.415546582713)*x_ref[1]+(-0.196674759494)*x_ref[2]**o+(0.235475790794)*x_ref[2]
        ref[1,3,1,1]=(-0.699232847611)*x_ref[0]**o+(0.504961844211)*x_ref[0]+(-0.0115303749449)*x_ref[1]**o+(-0.756286148171)*x_ref[1]+(-0.0922447811444)*x_ref[2]**o+(-0.330281180023)*x_ref[2]
        ref[1,3,2,0]=(0.149323550569)*x_ref[0]**o+(0.977735452052)*x_ref[0]+(0.255972870292)*x_ref[1]**o+(-0.855635543118)*x_ref[1]+(0.519602177962)*x_ref[2]**o+(0.922189470945)*x_ref[2]
        ref[1,3,2,1]=(-0.190749790276)*x_ref[0]**o+(-0.351506874986)*x_ref[0]+(-0.0672783552454)*x_ref[1]**o+(-0.223773986529)*x_ref[1]+(-0.593750649495)*x_ref[2]**o+(0.158607153203)*x_ref[2]
        ref[1,4,0,0]=(0.245306614552)*x_ref[0]**o+(-0.191799968697)*x_ref[0]+(0.85455961157)*x_ref[1]**o+(0.597426986918)*x_ref[1]+(-0.824363631294)*x_ref[2]**o+(-0.0161601323913)*x_ref[2]
        ref[1,4,0,1]=(0.925974209919)*x_ref[0]**o+(-0.606855935989)*x_ref[0]+(0.182619376779)*x_ref[1]**o+(0.82202480925)*x_ref[1]+(0.0511805106413)*x_ref[2]**o+(0.0866309784619)*x_ref[2]
        ref[1,4,1,0]=(0.272918227857)*x_ref[0]**o+(-0.406306617661)*x_ref[0]+(-0.0535134980185)*x_ref[1]**o+(-0.186200126868)*x_ref[1]+(0.92546875085)*x_ref[2]**o+(-0.474811198949)*x_ref[2]
        ref[1,4,1,1]=(0.803708551531)*x_ref[0]**o+(-0.715626841623)*x_ref[0]+(-0.0822999615139)*x_ref[1]**o+(0.612334309398)*x_ref[1]+(-0.157577434512)*x_ref[2]**o+(-0.795469427996)*x_ref[2]
        ref[1,4,2,0]=(-0.837159054411)*x_ref[0]**o+(-0.234523360065)*x_ref[0]+(-0.198337862205)*x_ref[1]**o+(0.897958649268)*x_ref[1]+(-0.519907680436)*x_ref[2]**o+(0.742526415023)*x_ref[2]
        ref[1,4,2,1]=(-0.792448458791)*x_ref[0]**o+(0.456487739729)*x_ref[0]+(0.0128609741891)*x_ref[1]**o+(-0.344964414689)*x_ref[1]+(-0.205584911868)*x_ref[2]**o+(0.429183240889)*x_ref[2]
        ref[2,0,0,0]=(0.401501024951)*x_ref[0]**o+(-0.497039379761)*x_ref[0]+(-0.228675343279)*x_ref[1]**o+(0.133323657822)*x_ref[1]+(-0.203006510606)*x_ref[2]**o+(0.407718693225)*x_ref[2]
        ref[2,0,0,1]=(-0.547448204328)*x_ref[0]**o+(-0.109789709109)*x_ref[0]+(-0.626633454424)*x_ref[1]**o+(-0.117871748454)*x_ref[1]+(0.961790647477)*x_ref[2]**o+(0.652883936373)*x_ref[2]
        ref[2,0,1,0]=(-0.85536471584)*x_ref[0]**o+(0.489291116489)*x_ref[0]+(-0.929596786116)*x_ref[1]**o+(0.0110017367519)*x_ref[1]+(-0.703617460934)*x_ref[2]**o+(0.0496869388223)*x_ref[2]
        ref[2,0,1,1]=(-0.501093953991)*x_ref[0]**o+(-0.722768363488)*x_ref[0]+(0.84017017736)*x_ref[1]**o+(-0.22900603012)*x_ref[1]+(0.760492299454)*x_ref[2]**o+(0.767346260882)*x_ref[2]
        ref[2,0,2,0]=(-0.328498451765)*x_ref[0]**o+(-0.387534616444)*x_ref[0]+(-0.0973382392242)*x_ref[1]**o+(0.418582689709)*x_ref[1]+(0.206439339821)*x_ref[2]**o+(-0.706353598324)*x_ref[2]
        ref[2,0,2,1]=(0.179977568909)*x_ref[0]**o+(0.642598753874)*x_ref[0]+(0.711020603016)*x_ref[1]**o+(-0.287193672421)*x_ref[1]+(-0.362612184622)*x_ref[2]**o+(-0.18146063868)*x_ref[2]
        ref[2,1,0,0]=(0.0470678392348)*x_ref[0]**o+(0.00649068845495)*x_ref[0]+(-0.239912084684)*x_ref[1]**o+(0.278051515796)*x_ref[1]+(-0.98924625944)*x_ref[2]**o+(0.847957009386)*x_ref[2]
        ref[2,1,0,1]=(-0.287954694237)*x_ref[0]**o+(0.127947128577)*x_ref[0]+(0.443869903193)*x_ref[1]**o+(-0.344694028625)*x_ref[1]+(-0.14884506191)*x_ref[2]**o+(0.748140862905)*x_ref[2]
        ref[2,1,1,0]=(-0.597041891422)*x_ref[0]**o+(0.914716550666)*x_ref[0]+(0.68496200279)*x_ref[1]**o+(0.0420126536598)*x_ref[1]+(-0.539966353171)*x_ref[2]**o+(-0.886373035297)*x_ref[2]
        ref[2,1,1,1]=(0.462776858955)*x_ref[0]**o+(-0.989367011955)*x_ref[0]+(0.421691081662)*x_ref[1]**o+(0.264858346124)*x_ref[1]+(0.513954688709)*x_ref[2]**o+(0.00863898872886)*x_ref[2]
        ref[2,1,2,0]=(0.503601863243)*x_ref[0]**o+(0.505973913611)*x_ref[0]+(0.0952160099415)*x_ref[1]**o+(-0.77103310327)*x_ref[1]+(-0.316507531754)*x_ref[2]**o+(0.506969348695)*x_ref[2]
        ref[2,1,2,1]=(0.521634307558)*x_ref[0]**o+(-0.910583030486)*x_ref[0]+(0.977891014192)*x_ref[1]**o+(0.985145579284)*x_ref[1]+(0.0465919834269)*x_ref[2]**o+(0.113946013883)*x_ref[2]
        ref[2,2,0,0]=(0.178722112454)*x_ref[0]**o+(0.582800628485)*x_ref[0]+(0.141950810923)*x_ref[1]**o+(0.226899624421)*x_ref[1]+(0.445681888363)*x_ref[2]**o+(-0.127689044751)*x_ref[2]
        ref[2,2,0,1]=(-0.433214095841)*x_ref[0]**o+(-0.61199374178)*x_ref[0]+(0.655706389409)*x_ref[1]**o+(-0.0160036542096)*x_ref[1]+(0.570017878158)*x_ref[2]**o+(-0.00718475090704)*x_ref[2]
        ref[2,2,1,0]=(-0.205396523704)*x_ref[0]**o+(0.87306641582)*x_ref[0]+(-0.664912090364)*x_ref[1]**o+(-0.324577426693)*x_ref[1]+(-0.37269275568)*x_ref[2]**o+(-0.0753237226649)*x_ref[2]
        ref[2,2,1,1]=(0.716795494369)*x_ref[0]**o+(-0.725543593985)*x_ref[0]+(0.751101352157)*x_ref[1]**o+(0.0629757438809)*x_ref[1]+(-0.388859722245)*x_ref[2]**o+(0.633579629088)*x_ref[2]
        ref[2,2,2,0]=(-0.785119558914)*x_ref[0]**o+(-0.289212761766)*x_ref[0]+(-0.947517483982)*x_ref[1]**o+(-0.898908873087)*x_ref[1]+(-0.758807720992)*x_ref[2]**o+(-0.747003629416)*x_ref[2]
        ref[2,2,2,1]=(0.384610894358)*x_ref[0]**o+(0.0170687884025)*x_ref[0]+(0.776211872471)*x_ref[1]**o+(-0.313821704113)*x_ref[1]+(-0.433021717398)*x_ref[2]**o+(-0.972385486304)*x_ref[2]
        ref[2,3,0,0]=(0.59307540612)*x_ref[0]**o+(0.93544260405)*x_ref[0]+(0.519016668994)*x_ref[1]**o+(-0.175290607112)*x_ref[1]+(-0.269720511811)*x_ref[2]**o+(0.117237249214)*x_ref[2]
        ref[2,3,0,1]=(-0.0752366310333)*x_ref[0]**o+(-0.168792429308)*x_ref[0]+(0.316236130047)*x_ref[1]**o+(0.021991183167)*x_ref[1]+(-0.2024837663)*x_ref[2]**o+(-0.676832860181)*x_ref[2]
        ref[2,3,1,0]=(-0.2682564416)*x_ref[0]**o+(0.276503108458)*x_ref[0]+(0.789470125022)*x_ref[1]**o+(-0.339044117352)*x_ref[1]+(-0.408507243588)*x_ref[2]**o+(0.0666267661149)*x_ref[2]
        ref[2,3,1,1]=(-0.053395293244)*x_ref[0]**o+(0.425880490109)*x_ref[0]+(-0.068128758144)*x_ref[1]**o+(0.413635486374)*x_ref[1]+(0.965050652205)*x_ref[2]**o+(0.126502830836)*x_ref[2]
        ref[2,3,2,0]=(0.601928913545)*x_ref[0]**o+(-0.0443247453956)*x_ref[0]+(-0.606811664164)*x_ref[1]**o+(0.0206686600296)*x_ref[1]+(-0.804806810474)*x_ref[2]**o+(-0.019745599298)*x_ref[2]
        ref[2,3,2,1]=(0.911324514835)*x_ref[0]**o+(-0.397270365997)*x_ref[0]+(0.511154806396)*x_ref[1]**o+(0.722003975461)*x_ref[1]+(-0.332922355553)*x_ref[2]**o+(-0.132309188597)*x_ref[2]
        ref[2,4,0,0]=(0.194698324972)*x_ref[0]**o+(-0.496022175033)*x_ref[0]+(0.804006600349)*x_ref[1]**o+(-0.643493478792)*x_ref[1]+(0.148395455552)*x_ref[2]**o+(-0.408862596804)*x_ref[2]
        ref[2,4,0,1]=(0.734891277828)*x_ref[0]**o+(-0.101225311176)*x_ref[0]+(-0.813960759076)*x_ref[1]**o+(0.862888786761)*x_ref[1]+(0.830408706284)*x_ref[2]**o+(0.523449046413)*x_ref[2]
        ref[2,4,1,0]=(0.286629494089)*x_ref[0]**o+(0.599214589464)*x_ref[0]+(-0.208652251419)*x_ref[1]**o+(-0.830028802843)*x_ref[1]+(0.805360345419)*x_ref[2]**o+(-0.865349771687)*x_ref[2]
        ref[2,4,1,1]=(0.0367955096711)*x_ref[0]**o+(-0.0261723558157)*x_ref[0]+(0.200900558315)*x_ref[1]**o+(-0.532346409176)*x_ref[1]+(-0.183663881543)*x_ref[2]**o+(0.93191706611)*x_ref[2]
        ref[2,4,2,0]=(-0.932223437809)*x_ref[0]**o+(-0.155278731606)*x_ref[0]+(-0.932519381895)*x_ref[1]**o+(0.0589900357358)*x_ref[1]+(-0.187551917707)*x_ref[2]**o+(-0.81649531847)*x_ref[2]
        ref[2,4,2,1]=(0.555481188864)*x_ref[0]**o+(0.946358167786)*x_ref[0]+(-0.231784522)*x_ref[1]**o+(0.592776954307)*x_ref[1]+(-0.0373202061188)*x_ref[2]**o+(0.0302770826388)*x_ref[2]
        ref[3,0,0,0]=(0.603747717227)*x_ref[0]**o+(-0.969329597018)*x_ref[0]+(0.311876345528)*x_ref[1]**o+(-0.589486433983)*x_ref[1]+(-0.294507657675)*x_ref[2]**o+(-0.743511183116)*x_ref[2]
        ref[3,0,0,1]=(-0.846568198319)*x_ref[0]**o+(0.93521569126)*x_ref[0]+(-0.874618479052)*x_ref[1]**o+(0.502042421273)*x_ref[1]+(-0.945100352471)*x_ref[2]**o+(0.765854746295)*x_ref[2]
        ref[3,0,1,0]=(-0.908353314852)*x_ref[0]**o+(0.701990826648)*x_ref[0]+(0.798549198275)*x_ref[1]**o+(-0.616631523859)*x_ref[1]+(-0.999800863738)*x_ref[2]**o+(-0.156585259768)*x_ref[2]
        ref[3,0,1,1]=(-0.433258077523)*x_ref[0]**o+(-0.641049552618)*x_ref[0]+(0.427903347848)*x_ref[1]**o+(-0.175290793023)*x_ref[1]+(-0.557154490598)*x_ref[2]**o+(-0.00495777825998)*x_ref[2]
        ref[3,0,2,0]=(0.998253378112)*x_ref[0]**o+(0.2484283851)*x_ref[0]+(0.955897183936)*x_ref[1]**o+(-0.0351483480831)*x_ref[1]+(0.690211608431)*x_ref[2]**o+(0.7420822406)*x_ref[2]
        ref[3,0,2,1]=(0.273153627728)*x_ref[0]**o+(0.931572979903)*x_ref[0]+(-0.279174117029)*x_ref[1]**o+(-0.741239387871)*x_ref[1]+(0.148372377955)*x_ref[2]**o+(0.781476364189)*x_ref[2]
        ref[3,1,0,0]=(0.231049869286)*x_ref[0]**o+(-0.871461717509)*x_ref[0]+(0.553545032585)*x_ref[1]**o+(0.979019482224)*x_ref[1]+(0.819288655429)*x_ref[2]**o+(-0.01489981549)*x_ref[2]
        ref[3,1,0,1]=(-0.891391000148)*x_ref[0]**o+(-0.360048424965)*x_ref[0]+(-0.390043258734)*x_ref[1]**o+(0.547316173383)*x_ref[1]+(-0.972094018922)*x_ref[2]**o+(0.7798960782)*x_ref[2]
        ref[3,1,1,0]=(0.675405837097)*x_ref[0]**o+(0.976689341648)*x_ref[0]+(-0.416251051342)*x_ref[1]**o+(0.696304358663)*x_ref[1]+(-0.395743544878)*x_ref[2]**o+(-0.0340340734736)*x_ref[2]
        ref[3,1,1,1]=(-0.531871504573)*x_ref[0]**o+(-0.926770430235)*x_ref[0]+(0.752552585571)*x_ref[1]**o+(-0.403402334727)*x_ref[1]+(0.07593559269)*x_ref[2]**o+(0.126742096076)*x_ref[2]
        ref[3,1,2,0]=(-0.033528429617)*x_ref[0]**o+(0.187766565831)*x_ref[0]+(0.192128207394)*x_ref[1]**o+(-0.0969373324914)*x_ref[1]+(-0.889231964389)*x_ref[2]**o+(0.493198117275)*x_ref[2]
        ref[3,1,2,1]=(-0.182267751454)*x_ref[0]**o+(-0.115192609603)*x_ref[0]+(-0.252701809463)*x_ref[1]**o+(0.355949096242)*x_ref[1]+(0.859478772116)*x_ref[2]**o+(0.539090466518)*x_ref[2]
        ref[3,2,0,0]=(0.25232058799)*x_ref[0]**o+(-0.246680227462)*x_ref[0]+(0.316105270194)*x_ref[1]**o+(-0.426505186439)*x_ref[1]+(0.980259506048)*x_ref[2]**o+(0.144514677615)*x_ref[2]
        ref[3,2,0,1]=(-0.583071409136)*x_ref[0]**o+(0.969922333041)*x_ref[0]+(-0.0251030152687)*x_ref[1]**o+(0.121967009007)*x_ref[1]+(0.391834007838)*x_ref[2]**o+(0.0339488845613)*x_ref[2]
        ref[3,2,1,0]=(0.706904708517)*x_ref[0]**o+(0.266141083673)*x_ref[0]+(-0.0339160692688)*x_ref[1]**o+(-0.152115678673)*x_ref[1]+(0.122213919626)*x_ref[2]**o+(-0.576478442144)*x_ref[2]
        ref[3,2,1,1]=(-0.917141077151)*x_ref[0]**o+(0.569284998999)*x_ref[0]+(-0.0979622311673)*x_ref[1]**o+(0.554915758725)*x_ref[1]+(0.824303662531)*x_ref[2]**o+(-0.449201813108)*x_ref[2]
        ref[3,2,2,0]=(-0.284000203352)*x_ref[0]**o+(0.0535558865246)*x_ref[0]+(-0.233670301273)*x_ref[1]**o+(0.383324533244)*x_ref[1]+(-0.483402264684)*x_ref[2]**o+(-0.482290332534)*x_ref[2]
        ref[3,2,2,1]=(-0.557106835416)*x_ref[0]**o+(-0.0263124485334)*x_ref[0]+(-0.132921973706)*x_ref[1]**o+(0.0351808996763)*x_ref[1]+(0.979819304163)*x_ref[2]**o+(-0.908131098335)*x_ref[2]
        ref[3,3,0,0]=(0.982856633407)*x_ref[0]**o+(-0.180985287571)*x_ref[0]+(0.364546381679)*x_ref[1]**o+(0.746225949823)*x_ref[1]+(0.847682125532)*x_ref[2]**o+(-0.782701591901)*x_ref[2]
        ref[3,3,0,1]=(-0.33362442802)*x_ref[0]**o+(-0.933319602818)*x_ref[0]+(-0.830312620251)*x_ref[1]**o+(-0.938028393959)*x_ref[1]+(0.859281853158)*x_ref[2]**o+(0.430732343672)*x_ref[2]
        ref[3,3,1,0]=(0.381562874557)*x_ref[0]**o+(0.0939156358041)*x_ref[0]+(0.207361800369)*x_ref[1]**o+(0.115172135617)*x_ref[1]+(0.747494745689)*x_ref[2]**o+(-0.45448955026)*x_ref[2]
        ref[3,3,1,1]=(0.095679833076)*x_ref[0]**o+(-0.936886715289)*x_ref[0]+(0.887146346873)*x_ref[1]**o+(0.159246683343)*x_ref[1]+(0.817874529811)*x_ref[2]**o+(0.516091968965)*x_ref[2]
        ref[3,3,2,0]=(-0.476953971316)*x_ref[0]**o+(0.176188538145)*x_ref[0]+(0.568224935975)*x_ref[1]**o+(0.56192429612)*x_ref[1]+(-0.700911152919)*x_ref[2]**o+(-0.0114437684935)*x_ref[2]
        ref[3,3,2,1]=(0.29541245686)*x_ref[0]**o+(0.546841142855)*x_ref[0]+(-0.23081431627)*x_ref[1]**o+(-0.633697713725)*x_ref[1]+(0.33519295746)*x_ref[2]**o+(-0.951229146931)*x_ref[2]
        ref[3,4,0,0]=(-0.879148775346)*x_ref[0]**o+(-0.703893190615)*x_ref[0]+(0.409831208498)*x_ref[1]**o+(0.754374287841)*x_ref[1]+(0.688380721157)*x_ref[2]**o+(-0.395460537774)*x_ref[2]
        ref[3,4,0,1]=(0.924489816348)*x_ref[0]**o+(0.999127631985)*x_ref[0]+(-0.892638544934)*x_ref[1]**o+(-0.923913790342)*x_ref[1]+(0.736487445972)*x_ref[2]**o+(0.250131270052)*x_ref[2]
        ref[3,4,1,0]=(0.264892665874)*x_ref[0]**o+(-0.776154012834)*x_ref[0]+(-0.379853719079)*x_ref[1]**o+(-0.748063746312)*x_ref[1]+(-0.0108974578949)*x_ref[2]**o+(-0.803449911818)*x_ref[2]
        ref[3,4,1,1]=(-0.699932530879)*x_ref[0]**o+(-0.901128866521)*x_ref[0]+(0.628561941772)*x_ref[1]**o+(0.507625196603)*x_ref[1]+(0.147922829864)*x_ref[2]**o+(0.475783034469)*x_ref[2]
        ref[3,4,2,0]=(0.482139761317)*x_ref[0]**o+(-0.627593411974)*x_ref[0]+(-0.84668128017)*x_ref[1]**o+(-0.324236153907)*x_ref[1]+(-0.688406716909)*x_ref[2]**o+(-0.0669177294519)*x_ref[2]
        ref[3,4,2,1]=(-0.531896625188)*x_ref[0]**o+(-0.574459691455)*x_ref[0]+(-0.824941329118)*x_ref[1]**o+(0.0812437772861)*x_ref[1]+(-0.257954477449)*x_ref[2]**o+(-0.219142537545)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.318228469869)*x[0]+(-0.774132349214)*x[1]
        ref=(0.318228469869)*x_ref[0]+(-0.774132349214)*x_ref[1]
      else:
        s=(0.377557485345)*x[0]+(1.02901239394)*x[1]+(0.038224819288)*x[2]
        ref=(0.377557485345)*x_ref[0]+(1.02901239394)*x_ref[1]+(0.038224819288)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        s[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
        ref[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        ref[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
      else:
        s[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        s[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
        ref[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        ref[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0621856629646)*x[0]+(0.783319354938)*x[1]
        s[0,1]=(-0.205526239577)*x[0]+(-0.401770179168)*x[1]
        s[0,2]=(0.543085761147)*x[0]+(0.915124776348)*x[1]
        s[0,3]=(1.3114416307)*x[0]+(-0.339977887579)*x[1]
        s[0,4]=(0.19180167645)*x[0]+(0.0744591496582)*x[1]
        s[1,0]=(0.833949474687)*x[0]+(-0.513960535503)*x[1]
        s[1,1]=(1.63749958183)*x[0]+(1.49589084809)*x[1]
        s[1,2]=(0.52690970122)*x[0]+(-0.497651760738)*x[1]
        s[1,3]=(0.157987079101)*x[0]+(0.244658238724)*x[1]
        s[1,4]=(-0.557780706846)*x[0]+(-1.49932552405)*x[1]
        s[2,0]=(-0.00409488968736)*x[0]+(0.366430715843)*x[1]
        s[2,1]=(-0.867654712773)*x[0]+(-0.129712066272)*x[1]
        s[2,2]=(1.33589989715)*x[0]+(-0.135732989085)*x[1]
        s[2,3]=(0.208531325369)*x[0]+(1.3627828974)*x[1]
        s[2,4]=(-0.667299295418)*x[0]+(0.981775718726)*x[1]
        s[3,0]=(1.40151192578)*x[0]+(-0.28272210702)*x[1]
        s[3,1]=(-1.13395485034)*x[0]+(-0.148770918012)*x[1]
        s[3,2]=(0.315288948224)*x[0]+(-0.140365593027)*x[1]
        s[3,3]=(-0.0421809531999)*x[0]+(0.522687748574)*x[1]
        s[3,4]=(0.135043757063)*x[0]+(-1.52489039301)*x[1]
        ref[0,0]=(0.0621856629646)*x_ref[0]+(0.783319354938)*x_ref[1]
        ref[0,1]=(-0.205526239577)*x_ref[0]+(-0.401770179168)*x_ref[1]
        ref[0,2]=(0.543085761147)*x_ref[0]+(0.915124776348)*x_ref[1]
        ref[0,3]=(1.3114416307)*x_ref[0]+(-0.339977887579)*x_ref[1]
        ref[0,4]=(0.19180167645)*x_ref[0]+(0.0744591496582)*x_ref[1]
        ref[1,0]=(0.833949474687)*x_ref[0]+(-0.513960535503)*x_ref[1]
        ref[1,1]=(1.63749958183)*x_ref[0]+(1.49589084809)*x_ref[1]
        ref[1,2]=(0.52690970122)*x_ref[0]+(-0.497651760738)*x_ref[1]
        ref[1,3]=(0.157987079101)*x_ref[0]+(0.244658238724)*x_ref[1]
        ref[1,4]=(-0.557780706846)*x_ref[0]+(-1.49932552405)*x_ref[1]
        ref[2,0]=(-0.00409488968736)*x_ref[0]+(0.366430715843)*x_ref[1]
        ref[2,1]=(-0.867654712773)*x_ref[0]+(-0.129712066272)*x_ref[1]
        ref[2,2]=(1.33589989715)*x_ref[0]+(-0.135732989085)*x_ref[1]
        ref[2,3]=(0.208531325369)*x_ref[0]+(1.3627828974)*x_ref[1]
        ref[2,4]=(-0.667299295418)*x_ref[0]+(0.981775718726)*x_ref[1]
        ref[3,0]=(1.40151192578)*x_ref[0]+(-0.28272210702)*x_ref[1]
        ref[3,1]=(-1.13395485034)*x_ref[0]+(-0.148770918012)*x_ref[1]
        ref[3,2]=(0.315288948224)*x_ref[0]+(-0.140365593027)*x_ref[1]
        ref[3,3]=(-0.0421809531999)*x_ref[0]+(0.522687748574)*x_ref[1]
        ref[3,4]=(0.135043757063)*x_ref[0]+(-1.52489039301)*x_ref[1]
      else:
        s[0,0]=(-1.07097471698)*x[0]+(0.511673191366)*x[1]+(0.986422233856)*x[2]
        s[0,1]=(1.35021866887)*x[0]+(-1.18036897489)*x[1]+(0.143096009704)*x[2]
        s[0,2]=(1.92136475346)*x[0]+(-1.52041779666)*x[1]+(1.50412123174)*x[2]
        s[0,3]=(-1.26349432492)*x[0]+(-1.69090463821)*x[1]+(0.0519848717435)*x[2]
        s[0,4]=(-0.450643290235)*x[0]+(-1.0918592714)*x[1]+(-1.78531176031)*x[2]
        s[1,0]=(1.24575021839)*x[0]+(0.378183524368)*x[1]+(0.164647643767)*x[2]
        s[1,1]=(-0.0164625547855)*x[0]+(-1.06539959746)*x[1]+(-1.5644531354)*x[2]
        s[1,2]=(0.126087223544)*x[0]+(0.876303692726)*x[1]+(1.3521510598)*x[2]
        s[1,3]=(-1.15749467575)*x[0]+(-0.850991452556)*x[1]+(-0.903705777185)*x[2]
        s[1,4]=(-0.680094762766)*x[0]+(0.280093157366)*x[1]+(0.460572306025)*x[2]
        s[2,0]=(-0.201094411285)*x[0]+(0.610240266355)*x[1]+(0.741298609653)*x[2]
        s[2,1]=(0.754412798233)*x[0]+(0.765169641301)*x[1]+(0.522349231028)*x[2]
        s[2,2]=(-0.330239429991)*x[0]+(1.04468899978)*x[1]+(-0.493706328863)*x[2]
        s[2,3]=(-0.107189727407)*x[0]+(-1.57053989855)*x[1]+(1.5864416691)*x[2]
        s[2,4]=(1.78349630817)*x[0]+(0.415138686617)*x[1]+(-0.567656058268)*x[2]
        s[3,0]=(-0.242864411179)*x[0]+(0.342401319249)*x[1]+(-0.963749227433)*x[2]
        s[3,1]=(-0.677695348369)*x[0]+(-1.52126003219)*x[1]+(-0.197904354713)*x[2]
        s[3,2]=(-0.549674242841)*x[0]+(1.23313354594)*x[1]+(0.701202865744)*x[2]
        s[3,3]=(0.278097672739)*x[0]+(-1.31336623233)*x[1]+(1.69771904282)*x[2]
        s[3,4]=(0.0982668788821)*x[0]+(-1.11857597802)*x[1]+(-0.0505715204344)*x[2]
        ref[0,0]=(-1.07097471698)*x_ref[0]+(0.511673191366)*x_ref[1]+(0.986422233856)*x_ref[2]
        ref[0,1]=(1.35021866887)*x_ref[0]+(-1.18036897489)*x_ref[1]+(0.143096009704)*x_ref[2]
        ref[0,2]=(1.92136475346)*x_ref[0]+(-1.52041779666)*x_ref[1]+(1.50412123174)*x_ref[2]
        ref[0,3]=(-1.26349432492)*x_ref[0]+(-1.69090463821)*x_ref[1]+(0.0519848717435)*x_ref[2]
        ref[0,4]=(-0.450643290235)*x_ref[0]+(-1.0918592714)*x_ref[1]+(-1.78531176031)*x_ref[2]
        ref[1,0]=(1.24575021839)*x_ref[0]+(0.378183524368)*x_ref[1]+(0.164647643767)*x_ref[2]
        ref[1,1]=(-0.0164625547855)*x_ref[0]+(-1.06539959746)*x_ref[1]+(-1.5644531354)*x_ref[2]
        ref[1,2]=(0.126087223544)*x_ref[0]+(0.876303692726)*x_ref[1]+(1.3521510598)*x_ref[2]
        ref[1,3]=(-1.15749467575)*x_ref[0]+(-0.850991452556)*x_ref[1]+(-0.903705777185)*x_ref[2]
        ref[1,4]=(-0.680094762766)*x_ref[0]+(0.280093157366)*x_ref[1]+(0.460572306025)*x_ref[2]
        ref[2,0]=(-0.201094411285)*x_ref[0]+(0.610240266355)*x_ref[1]+(0.741298609653)*x_ref[2]
        ref[2,1]=(0.754412798233)*x_ref[0]+(0.765169641301)*x_ref[1]+(0.522349231028)*x_ref[2]
        ref[2,2]=(-0.330239429991)*x_ref[0]+(1.04468899978)*x_ref[1]+(-0.493706328863)*x_ref[2]
        ref[2,3]=(-0.107189727407)*x_ref[0]+(-1.57053989855)*x_ref[1]+(1.5864416691)*x_ref[2]
        ref[2,4]=(1.78349630817)*x_ref[0]+(0.415138686617)*x_ref[1]+(-0.567656058268)*x_ref[2]
        ref[3,0]=(-0.242864411179)*x_ref[0]+(0.342401319249)*x_ref[1]+(-0.963749227433)*x_ref[2]
        ref[3,1]=(-0.677695348369)*x_ref[0]+(-1.52126003219)*x_ref[1]+(-0.197904354713)*x_ref[2]
        ref[3,2]=(-0.549674242841)*x_ref[0]+(1.23313354594)*x_ref[1]+(0.701202865744)*x_ref[2]
        ref[3,3]=(0.278097672739)*x_ref[0]+(-1.31336623233)*x_ref[1]+(1.69771904282)*x_ref[2]
        ref[3,4]=(0.0982668788821)*x_ref[0]+(-1.11857597802)*x_ref[1]+(-0.0505715204344)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.121576835783)*x[0]+(-0.941342629086)*x[1]
        s[0,0,1]=(0.113281061637)*x[0]+(1.53625452894)*x[1]
        s[0,1,0]=(0.0674434661275)*x[0]+(-0.633539894942)*x[1]
        s[0,1,1]=(0.22586221125)*x[0]+(-0.756374425353)*x[1]
        s[1,0,0]=(0.0826651782713)*x[0]+(-0.78928571118)*x[1]
        s[1,0,1]=(-0.844609848337)*x[0]+(1.15779995507)*x[1]
        s[1,1,0]=(0.137225056583)*x[0]+(-1.03725215253)*x[1]
        s[1,1,1]=(1.66612561218)*x[0]+(-0.243423071065)*x[1]
        s[2,0,0]=(0.0833021990187)*x[0]+(-1.2712750293)*x[1]
        s[2,0,1]=(0.13762583458)*x[0]+(-0.385843401588)*x[1]
        s[2,1,0]=(-0.926228562488)*x[0]+(-0.305689405746)*x[1]
        s[2,1,1]=(0.725733502304)*x[0]+(1.40832913183)*x[1]
        s[3,0,0]=(1.09943065211)*x[0]+(1.13514402138)*x[1]
        s[3,0,1]=(-1.0354232693)*x[0]+(-0.678629246255)*x[1]
        s[3,1,0]=(-0.44364521298)*x[0]+(0.857496257444)*x[1]
        s[3,1,1]=(1.53635223417)*x[0]+(-0.0887540544341)*x[1]
        s[4,0,0]=(0.553400675937)*x[0]+(-0.680780568865)*x[1]
        s[4,0,1]=(-0.00341234496916)*x[0]+(0.115104542523)*x[1]
        s[4,1,0]=(1.07927709685)*x[0]+(-0.270560535373)*x[1]
        s[4,1,1]=(-0.681458204636)*x[0]+(-0.607605729645)*x[1]
        s[5,0,0]=(0.10322587433)*x[0]+(0.982393144196)*x[1]
        s[5,0,1]=(-0.229957227498)*x[0]+(0.940017423086)*x[1]
        s[5,1,0]=(-0.465473033025)*x[0]+(-0.61380542705)*x[1]
        s[5,1,1]=(1.43248132924)*x[0]+(-1.71390853514)*x[1]
        ref[0,0,0]=(0.121576835783)*x_ref[0]+(-0.941342629086)*x_ref[1]
        ref[0,0,1]=(0.113281061637)*x_ref[0]+(1.53625452894)*x_ref[1]
        ref[0,1,0]=(0.0674434661275)*x_ref[0]+(-0.633539894942)*x_ref[1]
        ref[0,1,1]=(0.22586221125)*x_ref[0]+(-0.756374425353)*x_ref[1]
        ref[1,0,0]=(0.0826651782713)*x_ref[0]+(-0.78928571118)*x_ref[1]
        ref[1,0,1]=(-0.844609848337)*x_ref[0]+(1.15779995507)*x_ref[1]
        ref[1,1,0]=(0.137225056583)*x_ref[0]+(-1.03725215253)*x_ref[1]
        ref[1,1,1]=(1.66612561218)*x_ref[0]+(-0.243423071065)*x_ref[1]
        ref[2,0,0]=(0.0833021990187)*x_ref[0]+(-1.2712750293)*x_ref[1]
        ref[2,0,1]=(0.13762583458)*x_ref[0]+(-0.385843401588)*x_ref[1]
        ref[2,1,0]=(-0.926228562488)*x_ref[0]+(-0.305689405746)*x_ref[1]
        ref[2,1,1]=(0.725733502304)*x_ref[0]+(1.40832913183)*x_ref[1]
        ref[3,0,0]=(1.09943065211)*x_ref[0]+(1.13514402138)*x_ref[1]
        ref[3,0,1]=(-1.0354232693)*x_ref[0]+(-0.678629246255)*x_ref[1]
        ref[3,1,0]=(-0.44364521298)*x_ref[0]+(0.857496257444)*x_ref[1]
        ref[3,1,1]=(1.53635223417)*x_ref[0]+(-0.0887540544341)*x_ref[1]
        ref[4,0,0]=(0.553400675937)*x_ref[0]+(-0.680780568865)*x_ref[1]
        ref[4,0,1]=(-0.00341234496916)*x_ref[0]+(0.115104542523)*x_ref[1]
        ref[4,1,0]=(1.07927709685)*x_ref[0]+(-0.270560535373)*x_ref[1]
        ref[4,1,1]=(-0.681458204636)*x_ref[0]+(-0.607605729645)*x_ref[1]
        ref[5,0,0]=(0.10322587433)*x_ref[0]+(0.982393144196)*x_ref[1]
        ref[5,0,1]=(-0.229957227498)*x_ref[0]+(0.940017423086)*x_ref[1]
        ref[5,1,0]=(-0.465473033025)*x_ref[0]+(-0.61380542705)*x_ref[1]
        ref[5,1,1]=(1.43248132924)*x_ref[0]+(-1.71390853514)*x_ref[1]
      else:
        s[0,0,0]=(0.663545175027)*x[0]+(-1.05528281336)*x[1]+(-0.25024128376)*x[2]
        s[0,0,1]=(1.10817148343)*x[0]+(0.224734467846)*x[1]+(0.712872841532)*x[2]
        s[0,1,0]=(-0.345072989586)*x[0]+(-1.244678278)*x[1]+(0.103446646879)*x[2]
        s[0,1,1]=(-0.106279741301)*x[0]+(-0.525721333511)*x[1]+(0.401917882446)*x[2]
        s[1,0,0]=(0.838242796285)*x[0]+(-0.706020082817)*x[1]+(0.443650945014)*x[2]
        s[1,0,1]=(-1.16557336842)*x[0]+(0.215324659268)*x[1]+(-0.285210566147)*x[2]
        s[1,1,0]=(-0.141574656707)*x[0]+(1.66054358217)*x[1]+(-0.554616610711)*x[2]
        s[1,1,1]=(0.523682935472)*x[0]+(0.687178896331)*x[1]+(0.875639049678)*x[2]
        s[2,0,0]=(0.585144782395)*x[0]+(0.172063481865)*x[1]+(0.920891875835)*x[2]
        s[2,0,1]=(0.660815435485)*x[0]+(-1.63324849021)*x[1]+(0.0810858621718)*x[2]
        s[2,1,0]=(-0.238166824253)*x[0]+(0.0442487879497)*x[1]+(1.05748755823)*x[2]
        s[2,1,1]=(1.14314890535)*x[0]+(0.63161746042)*x[1]+(1.5024770616)*x[2]
        s[3,0,0]=(-1.33327391952)*x[0]+(1.58312319401)*x[1]+(-1.24202894696)*x[2]
        s[3,0,1]=(-0.0242608819668)*x[0]+(-1.12901018411)*x[1]+(-0.443840381999)*x[2]
        s[3,1,0]=(1.12400540809)*x[0]+(-0.25104247006)*x[1]+(0.55935664575)*x[2]
        s[3,1,1]=(0.976977053811)*x[0]+(1.46134206601)*x[1]+(-0.883829686094)*x[2]
        s[4,0,0]=(0.68585022695)*x[0]+(0.320211070942)*x[1]+(-0.715034754238)*x[2]
        s[4,0,1]=(0.745090118519)*x[0]+(-0.56687124464)*x[1]+(-1.41483630619)*x[2]
        s[4,1,0]=(-0.864645032271)*x[0]+(-0.0852453841156)*x[1]+(-1.02272230917)*x[2]
        s[4,1,1]=(0.243107783262)*x[0]+(-1.76263127813)*x[1]+(-0.0396122398093)*x[2]
        s[5,0,0]=(-1.33890039547)*x[0]+(0.95916460762)*x[1]+(-0.118457193655)*x[2]
        s[5,0,1]=(-1.03526146458)*x[0]+(-0.279038868252)*x[1]+(-0.702378064321)*x[2]
        s[5,1,0]=(-0.434705839506)*x[0]+(0.388520593856)*x[1]+(0.213085443323)*x[2]
        s[5,1,1]=(-0.156276503308)*x[0]+(-0.463824282648)*x[1]+(-0.801704395726)*x[2]
        ref[0,0,0]=(0.663545175027)*x_ref[0]+(-1.05528281336)*x_ref[1]+(-0.25024128376)*x_ref[2]
        ref[0,0,1]=(1.10817148343)*x_ref[0]+(0.224734467846)*x_ref[1]+(0.712872841532)*x_ref[2]
        ref[0,1,0]=(-0.345072989586)*x_ref[0]+(-1.244678278)*x_ref[1]+(0.103446646879)*x_ref[2]
        ref[0,1,1]=(-0.106279741301)*x_ref[0]+(-0.525721333511)*x_ref[1]+(0.401917882446)*x_ref[2]
        ref[1,0,0]=(0.838242796285)*x_ref[0]+(-0.706020082817)*x_ref[1]+(0.443650945014)*x_ref[2]
        ref[1,0,1]=(-1.16557336842)*x_ref[0]+(0.215324659268)*x_ref[1]+(-0.285210566147)*x_ref[2]
        ref[1,1,0]=(-0.141574656707)*x_ref[0]+(1.66054358217)*x_ref[1]+(-0.554616610711)*x_ref[2]
        ref[1,1,1]=(0.523682935472)*x_ref[0]+(0.687178896331)*x_ref[1]+(0.875639049678)*x_ref[2]
        ref[2,0,0]=(0.585144782395)*x_ref[0]+(0.172063481865)*x_ref[1]+(0.920891875835)*x_ref[2]
        ref[2,0,1]=(0.660815435485)*x_ref[0]+(-1.63324849021)*x_ref[1]+(0.0810858621718)*x_ref[2]
        ref[2,1,0]=(-0.238166824253)*x_ref[0]+(0.0442487879497)*x_ref[1]+(1.05748755823)*x_ref[2]
        ref[2,1,1]=(1.14314890535)*x_ref[0]+(0.63161746042)*x_ref[1]+(1.5024770616)*x_ref[2]
        ref[3,0,0]=(-1.33327391952)*x_ref[0]+(1.58312319401)*x_ref[1]+(-1.24202894696)*x_ref[2]
        ref[3,0,1]=(-0.0242608819668)*x_ref[0]+(-1.12901018411)*x_ref[1]+(-0.443840381999)*x_ref[2]
        ref[3,1,0]=(1.12400540809)*x_ref[0]+(-0.25104247006)*x_ref[1]+(0.55935664575)*x_ref[2]
        ref[3,1,1]=(0.976977053811)*x_ref[0]+(1.46134206601)*x_ref[1]+(-0.883829686094)*x_ref[2]
        ref[4,0,0]=(0.68585022695)*x_ref[0]+(0.320211070942)*x_ref[1]+(-0.715034754238)*x_ref[2]
        ref[4,0,1]=(0.745090118519)*x_ref[0]+(-0.56687124464)*x_ref[1]+(-1.41483630619)*x_ref[2]
        ref[4,1,0]=(-0.864645032271)*x_ref[0]+(-0.0852453841156)*x_ref[1]+(-1.02272230917)*x_ref[2]
        ref[4,1,1]=(0.243107783262)*x_ref[0]+(-1.76263127813)*x_ref[1]+(-0.0396122398093)*x_ref[2]
        ref[5,0,0]=(-1.33890039547)*x_ref[0]+(0.95916460762)*x_ref[1]+(-0.118457193655)*x_ref[2]
        ref[5,0,1]=(-1.03526146458)*x_ref[0]+(-0.279038868252)*x_ref[1]+(-0.702378064321)*x_ref[2]
        ref[5,1,0]=(-0.434705839506)*x_ref[0]+(0.388520593856)*x_ref[1]+(0.213085443323)*x_ref[2]
        ref[5,1,1]=(-0.156276503308)*x_ref[0]+(-0.463824282648)*x_ref[1]+(-0.801704395726)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.239181398192)*x[0]+(-0.39363876765)*x[1]
        s[0,0,0,1]=(0.391461093031)*x[0]+(-0.825968486042)*x[1]
        s[0,0,1,0]=(0.327281495381)*x[0]+(-0.201869630768)*x[1]
        s[0,0,1,1]=(-0.0266300089051)*x[0]+(0.0315885163331)*x[1]
        s[0,0,2,0]=(-1.29808531891)*x[0]+(0.908442635353)*x[1]
        s[0,0,2,1]=(-0.0719703714734)*x[0]+(-1.2949593192)*x[1]
        s[0,1,0,0]=(1.56337010697)*x[0]+(0.554963325139)*x[1]
        s[0,1,0,1]=(-0.590177179718)*x[0]+(-0.577417888525)*x[1]
        s[0,1,1,0]=(1.12019940888)*x[0]+(0.515714775221)*x[1]
        s[0,1,1,1]=(1.62187733115)*x[0]+(-0.0496003559467)*x[1]
        s[0,1,2,0]=(1.33693246276)*x[0]+(0.18028194694)*x[1]
        s[0,1,2,1]=(1.197520649)*x[0]+(-0.189794758648)*x[1]
        s[0,2,0,0]=(0.458741675295)*x[0]+(-1.23374520557)*x[1]
        s[0,2,0,1]=(0.122618584159)*x[0]+(1.68218857683)*x[1]
        s[0,2,1,0]=(-0.629416301315)*x[0]+(0.00821950501686)*x[1]
        s[0,2,1,1]=(0.372444097376)*x[0]+(-0.389807082811)*x[1]
        s[0,2,2,0]=(0.453580881811)*x[0]+(0.234675328844)*x[1]
        s[0,2,2,1]=(-0.488447987119)*x[0]+(-0.795290862711)*x[1]
        s[0,3,0,0]=(-0.0558838669873)*x[0]+(-0.123245413784)*x[1]
        s[0,3,0,1]=(1.37779240976)*x[0]+(0.0843495250295)*x[1]
        s[0,3,1,0]=(1.17819744943)*x[0]+(-0.859044697828)*x[1]
        s[0,3,1,1]=(1.25175416429)*x[0]+(-0.727964075982)*x[1]
        s[0,3,2,0]=(0.331480991751)*x[0]+(0.115364502392)*x[1]
        s[0,3,2,1]=(-0.791037073807)*x[0]+(-0.51631676808)*x[1]
        s[0,4,0,0]=(0.884389418081)*x[0]+(1.15006122432)*x[1]
        s[0,4,0,1]=(-0.914023114312)*x[0]+(-0.211608734049)*x[1]
        s[0,4,1,0]=(0.55983523554)*x[0]+(-0.792920109589)*x[1]
        s[0,4,1,1]=(0.70901081021)*x[0]+(0.205352000801)*x[1]
        s[0,4,2,0]=(0.140341122546)*x[0]+(0.633534699565)*x[1]
        s[0,4,2,1]=(-0.399843857722)*x[0]+(0.0360346234753)*x[1]
        s[1,0,0,0]=(0.649005423866)*x[0]+(-0.72958629181)*x[1]
        s[1,0,0,1]=(-1.23313007881)*x[0]+(0.240005909545)*x[1]
        s[1,0,1,0]=(0.817440794961)*x[0]+(0.219684770228)*x[1]
        s[1,0,1,1]=(0.300714219818)*x[0]+(0.339948691074)*x[1]
        s[1,0,2,0]=(0.692412749315)*x[0]+(0.847435101682)*x[1]
        s[1,0,2,1]=(0.650288735017)*x[0]+(1.25808519806)*x[1]
        s[1,1,0,0]=(-0.723761201921)*x[0]+(-0.337875653383)*x[1]
        s[1,1,0,1]=(1.10005713123)*x[0]+(0.465822601087)*x[1]
        s[1,1,1,0]=(0.140014505481)*x[0]+(0.289986937348)*x[1]
        s[1,1,1,1]=(0.0941270067899)*x[0]+(-1.45892231406)*x[1]
        s[1,1,2,0]=(1.04493752023)*x[0]+(-0.474063991258)*x[1]
        s[1,1,2,1]=(0.703284346556)*x[0]+(0.482116021536)*x[1]
        s[1,2,0,0]=(-1.66498217207)*x[0]+(-0.144077385929)*x[1]
        s[1,2,0,1]=(1.50511312044)*x[0]+(1.02587270487)*x[1]
        s[1,2,1,0]=(-1.15095468052)*x[0]+(-0.472480042698)*x[1]
        s[1,2,1,1]=(-0.453604215856)*x[0]+(-1.58856703352)*x[1]
        s[1,2,2,0]=(0.659481843854)*x[0]+(0.649931585882)*x[1]
        s[1,2,2,1]=(0.642785065122)*x[0]+(0.0614989297125)*x[1]
        s[1,3,0,0]=(1.65661008057)*x[0]+(-1.08472986336)*x[1]
        s[1,3,0,1]=(-0.64853562643)*x[0]+(-0.912673933935)*x[1]
        s[1,3,1,0]=(-0.108999021371)*x[0]+(-1.44843200164)*x[1]
        s[1,3,1,1]=(0.568395064875)*x[0]+(-0.471819846301)*x[1]
        s[1,3,2,0]=(-0.801680619008)*x[0]+(0.197249605781)*x[1]
        s[1,3,2,1]=(-0.512763932006)*x[0]+(-1.03221569621)*x[1]
        s[1,4,0,0]=(0.54142249246)*x[0]+(-1.28623467967)*x[1]
        s[1,4,0,1]=(1.19094828259)*x[0]+(0.58814933754)*x[1]
        s[1,4,1,0]=(-0.203699090085)*x[0]+(0.839194924785)*x[1]
        s[1,4,1,1]=(-1.73647778175)*x[0]+(1.26399102178)*x[1]
        s[1,4,2,0]=(0.688685884338)*x[0]+(-0.0441772317601)*x[1]
        s[1,4,2,1]=(-0.811158497264)*x[0]+(-0.663923769116)*x[1]
        s[2,0,0,0]=(0.237794148299)*x[0]+(0.125329413285)*x[1]
        s[2,0,0,1]=(-1.40518934715)*x[0]+(1.22320828318)*x[1]
        s[2,0,1,0]=(0.223901464538)*x[0]+(-0.00989463351767)*x[1]
        s[2,0,1,1]=(-0.912246132725)*x[0]+(0.686091520129)*x[1]
        s[2,0,2,0]=(0.181752601691)*x[0]+(-1.12779553611)*x[1]
        s[2,0,2,1]=(-0.061707870206)*x[0]+(-1.3429678748)*x[1]
        s[2,1,0,0]=(1.56396877322)*x[0]+(-1.05119596867)*x[1]
        s[2,1,0,1]=(0.929556514862)*x[0]+(0.406040622531)*x[1]
        s[2,1,1,0]=(-0.509355797822)*x[0]+(0.0605066527971)*x[1]
        s[2,1,1,1]=(-0.528244794646)*x[0]+(0.0388070048758)*x[1]
        s[2,1,2,0]=(-0.200007488719)*x[0]+(-0.711333050321)*x[1]
        s[2,1,2,1]=(0.175160525472)*x[0]+(-1.24343736284)*x[1]
        s[2,2,0,0]=(0.894960105109)*x[0]+(-0.747881639384)*x[1]
        s[2,2,0,1]=(-0.0894204005642)*x[0]+(-0.694792125757)*x[1]
        s[2,2,1,0]=(0.172037435959)*x[0]+(1.29233768559)*x[1]
        s[2,2,1,1]=(-0.312372601854)*x[0]+(-1.35937909779)*x[1]
        s[2,2,2,0]=(0.323370754787)*x[0]+(-0.0601280403525)*x[1]
        s[2,2,2,1]=(0.648154329469)*x[0]+(-0.976168040714)*x[1]
        s[2,3,0,0]=(0.178127534181)*x[0]+(-0.451322025537)*x[1]
        s[2,3,0,1]=(-1.88695440231)*x[0]+(0.164327427196)*x[1]
        s[2,3,1,0]=(0.543479878129)*x[0]+(-0.553094838979)*x[1]
        s[2,3,1,1]=(-0.745599057313)*x[0]+(0.710208052579)*x[1]
        s[2,3,2,0]=(-0.262128106404)*x[0]+(0.839008375922)*x[1]
        s[2,3,2,1]=(-1.20847299902)*x[0]+(-0.321234640326)*x[1]
        s[2,4,0,0]=(-1.24374530208)*x[0]+(-0.961997778937)*x[1]
        s[2,4,0,1]=(-0.0306817051594)*x[0]+(-0.127120502282)*x[1]
        s[2,4,1,0]=(1.25878772522)*x[0]+(1.04432409904)*x[1]
        s[2,4,1,1]=(0.271036209294)*x[0]+(-1.19597311792)*x[1]
        s[2,4,2,0]=(-0.073359480179)*x[0]+(-0.31523273696)*x[1]
        s[2,4,2,1]=(0.738924603079)*x[0]+(0.251467051457)*x[1]
        s[3,0,0,0]=(-0.257569738752)*x[0]+(-1.51373140433)*x[1]
        s[3,0,0,1]=(-0.0113459170081)*x[0]+(-0.356375170129)*x[1]
        s[3,0,1,0]=(-0.290333097834)*x[0]+(-1.56956309356)*x[1]
        s[3,0,1,1]=(-1.0806247672)*x[0]+(-0.376429365315)*x[1]
        s[3,0,2,0]=(-1.32161176104)*x[0]+(0.784442685433)*x[1]
        s[3,0,2,1]=(-1.00382464524)*x[0]+(-0.571979697031)*x[1]
        s[3,1,0,0]=(0.140827544385)*x[0]+(1.3690229529)*x[1]
        s[3,1,0,1]=(0.315011101794)*x[0]+(-1.0263379129)*x[1]
        s[3,1,1,0]=(-0.478940977535)*x[0]+(0.718456942414)*x[1]
        s[3,1,1,1]=(-0.655873236005)*x[0]+(-0.148701297422)*x[1]
        s[3,1,2,0]=(0.56126503141)*x[0]+(1.68526215251)*x[1]
        s[3,1,2,1]=(-0.337284606646)*x[0]+(-0.197295432114)*x[1]
        s[3,2,0,0]=(-0.5979320732)*x[0]+(-0.0600053477444)*x[1]
        s[3,2,0,1]=(-1.20531807175)*x[0]+(-0.148853538473)*x[1]
        s[3,2,1,0]=(0.313468114426)*x[0]+(-0.436912848514)*x[1]
        s[3,2,1,1]=(-0.544279507407)*x[0]+(0.189063470736)*x[1]
        s[3,2,2,0]=(-0.551741475397)*x[0]+(0.196370248113)*x[1]
        s[3,2,2,1]=(-0.0216545656433)*x[0]+(0.23669839429)*x[1]
        s[3,3,0,0]=(-0.406652896064)*x[0]+(0.00397676360081)*x[1]
        s[3,3,0,1]=(-0.305847961815)*x[0]+(-1.2563276235)*x[1]
        s[3,3,1,0]=(-0.450621831697)*x[0]+(-0.213476321469)*x[1]
        s[3,3,1,1]=(1.20548652879)*x[0]+(-0.286503833918)*x[1]
        s[3,3,2,0]=(0.516344981347)*x[0]+(-0.343787067175)*x[1]
        s[3,3,2,1]=(-0.01938029713)*x[0]+(-0.779285558404)*x[1]
        s[3,4,0,0]=(-0.406233598349)*x[0]+(-0.498045431333)*x[1]
        s[3,4,0,1]=(-0.790577593976)*x[0]+(1.07251305331)*x[1]
        s[3,4,1,0]=(0.510569891955)*x[0]+(0.526099613153)*x[1]
        s[3,4,1,1]=(0.150065560826)*x[0]+(0.966145292988)*x[1]
        s[3,4,2,0]=(0.718402792955)*x[0]+(-0.430416841145)*x[1]
        s[3,4,2,1]=(-0.430240857917)*x[0]+(-0.0679614650591)*x[1]
        ref[0,0,0,0]=(0.239181398192)*x_ref[0]+(-0.39363876765)*x_ref[1]
        ref[0,0,0,1]=(0.391461093031)*x_ref[0]+(-0.825968486042)*x_ref[1]
        ref[0,0,1,0]=(0.327281495381)*x_ref[0]+(-0.201869630768)*x_ref[1]
        ref[0,0,1,1]=(-0.0266300089051)*x_ref[0]+(0.0315885163331)*x_ref[1]
        ref[0,0,2,0]=(-1.29808531891)*x_ref[0]+(0.908442635353)*x_ref[1]
        ref[0,0,2,1]=(-0.0719703714734)*x_ref[0]+(-1.2949593192)*x_ref[1]
        ref[0,1,0,0]=(1.56337010697)*x_ref[0]+(0.554963325139)*x_ref[1]
        ref[0,1,0,1]=(-0.590177179718)*x_ref[0]+(-0.577417888525)*x_ref[1]
        ref[0,1,1,0]=(1.12019940888)*x_ref[0]+(0.515714775221)*x_ref[1]
        ref[0,1,1,1]=(1.62187733115)*x_ref[0]+(-0.0496003559467)*x_ref[1]
        ref[0,1,2,0]=(1.33693246276)*x_ref[0]+(0.18028194694)*x_ref[1]
        ref[0,1,2,1]=(1.197520649)*x_ref[0]+(-0.189794758648)*x_ref[1]
        ref[0,2,0,0]=(0.458741675295)*x_ref[0]+(-1.23374520557)*x_ref[1]
        ref[0,2,0,1]=(0.122618584159)*x_ref[0]+(1.68218857683)*x_ref[1]
        ref[0,2,1,0]=(-0.629416301315)*x_ref[0]+(0.00821950501686)*x_ref[1]
        ref[0,2,1,1]=(0.372444097376)*x_ref[0]+(-0.389807082811)*x_ref[1]
        ref[0,2,2,0]=(0.453580881811)*x_ref[0]+(0.234675328844)*x_ref[1]
        ref[0,2,2,1]=(-0.488447987119)*x_ref[0]+(-0.795290862711)*x_ref[1]
        ref[0,3,0,0]=(-0.0558838669873)*x_ref[0]+(-0.123245413784)*x_ref[1]
        ref[0,3,0,1]=(1.37779240976)*x_ref[0]+(0.0843495250295)*x_ref[1]
        ref[0,3,1,0]=(1.17819744943)*x_ref[0]+(-0.859044697828)*x_ref[1]
        ref[0,3,1,1]=(1.25175416429)*x_ref[0]+(-0.727964075982)*x_ref[1]
        ref[0,3,2,0]=(0.331480991751)*x_ref[0]+(0.115364502392)*x_ref[1]
        ref[0,3,2,1]=(-0.791037073807)*x_ref[0]+(-0.51631676808)*x_ref[1]
        ref[0,4,0,0]=(0.884389418081)*x_ref[0]+(1.15006122432)*x_ref[1]
        ref[0,4,0,1]=(-0.914023114312)*x_ref[0]+(-0.211608734049)*x_ref[1]
        ref[0,4,1,0]=(0.55983523554)*x_ref[0]+(-0.792920109589)*x_ref[1]
        ref[0,4,1,1]=(0.70901081021)*x_ref[0]+(0.205352000801)*x_ref[1]
        ref[0,4,2,0]=(0.140341122546)*x_ref[0]+(0.633534699565)*x_ref[1]
        ref[0,4,2,1]=(-0.399843857722)*x_ref[0]+(0.0360346234753)*x_ref[1]
        ref[1,0,0,0]=(0.649005423866)*x_ref[0]+(-0.72958629181)*x_ref[1]
        ref[1,0,0,1]=(-1.23313007881)*x_ref[0]+(0.240005909545)*x_ref[1]
        ref[1,0,1,0]=(0.817440794961)*x_ref[0]+(0.219684770228)*x_ref[1]
        ref[1,0,1,1]=(0.300714219818)*x_ref[0]+(0.339948691074)*x_ref[1]
        ref[1,0,2,0]=(0.692412749315)*x_ref[0]+(0.847435101682)*x_ref[1]
        ref[1,0,2,1]=(0.650288735017)*x_ref[0]+(1.25808519806)*x_ref[1]
        ref[1,1,0,0]=(-0.723761201921)*x_ref[0]+(-0.337875653383)*x_ref[1]
        ref[1,1,0,1]=(1.10005713123)*x_ref[0]+(0.465822601087)*x_ref[1]
        ref[1,1,1,0]=(0.140014505481)*x_ref[0]+(0.289986937348)*x_ref[1]
        ref[1,1,1,1]=(0.0941270067899)*x_ref[0]+(-1.45892231406)*x_ref[1]
        ref[1,1,2,0]=(1.04493752023)*x_ref[0]+(-0.474063991258)*x_ref[1]
        ref[1,1,2,1]=(0.703284346556)*x_ref[0]+(0.482116021536)*x_ref[1]
        ref[1,2,0,0]=(-1.66498217207)*x_ref[0]+(-0.144077385929)*x_ref[1]
        ref[1,2,0,1]=(1.50511312044)*x_ref[0]+(1.02587270487)*x_ref[1]
        ref[1,2,1,0]=(-1.15095468052)*x_ref[0]+(-0.472480042698)*x_ref[1]
        ref[1,2,1,1]=(-0.453604215856)*x_ref[0]+(-1.58856703352)*x_ref[1]
        ref[1,2,2,0]=(0.659481843854)*x_ref[0]+(0.649931585882)*x_ref[1]
        ref[1,2,2,1]=(0.642785065122)*x_ref[0]+(0.0614989297125)*x_ref[1]
        ref[1,3,0,0]=(1.65661008057)*x_ref[0]+(-1.08472986336)*x_ref[1]
        ref[1,3,0,1]=(-0.64853562643)*x_ref[0]+(-0.912673933935)*x_ref[1]
        ref[1,3,1,0]=(-0.108999021371)*x_ref[0]+(-1.44843200164)*x_ref[1]
        ref[1,3,1,1]=(0.568395064875)*x_ref[0]+(-0.471819846301)*x_ref[1]
        ref[1,3,2,0]=(-0.801680619008)*x_ref[0]+(0.197249605781)*x_ref[1]
        ref[1,3,2,1]=(-0.512763932006)*x_ref[0]+(-1.03221569621)*x_ref[1]
        ref[1,4,0,0]=(0.54142249246)*x_ref[0]+(-1.28623467967)*x_ref[1]
        ref[1,4,0,1]=(1.19094828259)*x_ref[0]+(0.58814933754)*x_ref[1]
        ref[1,4,1,0]=(-0.203699090085)*x_ref[0]+(0.839194924785)*x_ref[1]
        ref[1,4,1,1]=(-1.73647778175)*x_ref[0]+(1.26399102178)*x_ref[1]
        ref[1,4,2,0]=(0.688685884338)*x_ref[0]+(-0.0441772317601)*x_ref[1]
        ref[1,4,2,1]=(-0.811158497264)*x_ref[0]+(-0.663923769116)*x_ref[1]
        ref[2,0,0,0]=(0.237794148299)*x_ref[0]+(0.125329413285)*x_ref[1]
        ref[2,0,0,1]=(-1.40518934715)*x_ref[0]+(1.22320828318)*x_ref[1]
        ref[2,0,1,0]=(0.223901464538)*x_ref[0]+(-0.00989463351767)*x_ref[1]
        ref[2,0,1,1]=(-0.912246132725)*x_ref[0]+(0.686091520129)*x_ref[1]
        ref[2,0,2,0]=(0.181752601691)*x_ref[0]+(-1.12779553611)*x_ref[1]
        ref[2,0,2,1]=(-0.061707870206)*x_ref[0]+(-1.3429678748)*x_ref[1]
        ref[2,1,0,0]=(1.56396877322)*x_ref[0]+(-1.05119596867)*x_ref[1]
        ref[2,1,0,1]=(0.929556514862)*x_ref[0]+(0.406040622531)*x_ref[1]
        ref[2,1,1,0]=(-0.509355797822)*x_ref[0]+(0.0605066527971)*x_ref[1]
        ref[2,1,1,1]=(-0.528244794646)*x_ref[0]+(0.0388070048758)*x_ref[1]
        ref[2,1,2,0]=(-0.200007488719)*x_ref[0]+(-0.711333050321)*x_ref[1]
        ref[2,1,2,1]=(0.175160525472)*x_ref[0]+(-1.24343736284)*x_ref[1]
        ref[2,2,0,0]=(0.894960105109)*x_ref[0]+(-0.747881639384)*x_ref[1]
        ref[2,2,0,1]=(-0.0894204005642)*x_ref[0]+(-0.694792125757)*x_ref[1]
        ref[2,2,1,0]=(0.172037435959)*x_ref[0]+(1.29233768559)*x_ref[1]
        ref[2,2,1,1]=(-0.312372601854)*x_ref[0]+(-1.35937909779)*x_ref[1]
        ref[2,2,2,0]=(0.323370754787)*x_ref[0]+(-0.0601280403525)*x_ref[1]
        ref[2,2,2,1]=(0.648154329469)*x_ref[0]+(-0.976168040714)*x_ref[1]
        ref[2,3,0,0]=(0.178127534181)*x_ref[0]+(-0.451322025537)*x_ref[1]
        ref[2,3,0,1]=(-1.88695440231)*x_ref[0]+(0.164327427196)*x_ref[1]
        ref[2,3,1,0]=(0.543479878129)*x_ref[0]+(-0.553094838979)*x_ref[1]
        ref[2,3,1,1]=(-0.745599057313)*x_ref[0]+(0.710208052579)*x_ref[1]
        ref[2,3,2,0]=(-0.262128106404)*x_ref[0]+(0.839008375922)*x_ref[1]
        ref[2,3,2,1]=(-1.20847299902)*x_ref[0]+(-0.321234640326)*x_ref[1]
        ref[2,4,0,0]=(-1.24374530208)*x_ref[0]+(-0.961997778937)*x_ref[1]
        ref[2,4,0,1]=(-0.0306817051594)*x_ref[0]+(-0.127120502282)*x_ref[1]
        ref[2,4,1,0]=(1.25878772522)*x_ref[0]+(1.04432409904)*x_ref[1]
        ref[2,4,1,1]=(0.271036209294)*x_ref[0]+(-1.19597311792)*x_ref[1]
        ref[2,4,2,0]=(-0.073359480179)*x_ref[0]+(-0.31523273696)*x_ref[1]
        ref[2,4,2,1]=(0.738924603079)*x_ref[0]+(0.251467051457)*x_ref[1]
        ref[3,0,0,0]=(-0.257569738752)*x_ref[0]+(-1.51373140433)*x_ref[1]
        ref[3,0,0,1]=(-0.0113459170081)*x_ref[0]+(-0.356375170129)*x_ref[1]
        ref[3,0,1,0]=(-0.290333097834)*x_ref[0]+(-1.56956309356)*x_ref[1]
        ref[3,0,1,1]=(-1.0806247672)*x_ref[0]+(-0.376429365315)*x_ref[1]
        ref[3,0,2,0]=(-1.32161176104)*x_ref[0]+(0.784442685433)*x_ref[1]
        ref[3,0,2,1]=(-1.00382464524)*x_ref[0]+(-0.571979697031)*x_ref[1]
        ref[3,1,0,0]=(0.140827544385)*x_ref[0]+(1.3690229529)*x_ref[1]
        ref[3,1,0,1]=(0.315011101794)*x_ref[0]+(-1.0263379129)*x_ref[1]
        ref[3,1,1,0]=(-0.478940977535)*x_ref[0]+(0.718456942414)*x_ref[1]
        ref[3,1,1,1]=(-0.655873236005)*x_ref[0]+(-0.148701297422)*x_ref[1]
        ref[3,1,2,0]=(0.56126503141)*x_ref[0]+(1.68526215251)*x_ref[1]
        ref[3,1,2,1]=(-0.337284606646)*x_ref[0]+(-0.197295432114)*x_ref[1]
        ref[3,2,0,0]=(-0.5979320732)*x_ref[0]+(-0.0600053477444)*x_ref[1]
        ref[3,2,0,1]=(-1.20531807175)*x_ref[0]+(-0.148853538473)*x_ref[1]
        ref[3,2,1,0]=(0.313468114426)*x_ref[0]+(-0.436912848514)*x_ref[1]
        ref[3,2,1,1]=(-0.544279507407)*x_ref[0]+(0.189063470736)*x_ref[1]
        ref[3,2,2,0]=(-0.551741475397)*x_ref[0]+(0.196370248113)*x_ref[1]
        ref[3,2,2,1]=(-0.0216545656433)*x_ref[0]+(0.23669839429)*x_ref[1]
        ref[3,3,0,0]=(-0.406652896064)*x_ref[0]+(0.00397676360081)*x_ref[1]
        ref[3,3,0,1]=(-0.305847961815)*x_ref[0]+(-1.2563276235)*x_ref[1]
        ref[3,3,1,0]=(-0.450621831697)*x_ref[0]+(-0.213476321469)*x_ref[1]
        ref[3,3,1,1]=(1.20548652879)*x_ref[0]+(-0.286503833918)*x_ref[1]
        ref[3,3,2,0]=(0.516344981347)*x_ref[0]+(-0.343787067175)*x_ref[1]
        ref[3,3,2,1]=(-0.01938029713)*x_ref[0]+(-0.779285558404)*x_ref[1]
        ref[3,4,0,0]=(-0.406233598349)*x_ref[0]+(-0.498045431333)*x_ref[1]
        ref[3,4,0,1]=(-0.790577593976)*x_ref[0]+(1.07251305331)*x_ref[1]
        ref[3,4,1,0]=(0.510569891955)*x_ref[0]+(0.526099613153)*x_ref[1]
        ref[3,4,1,1]=(0.150065560826)*x_ref[0]+(0.966145292988)*x_ref[1]
        ref[3,4,2,0]=(0.718402792955)*x_ref[0]+(-0.430416841145)*x_ref[1]
        ref[3,4,2,1]=(-0.430240857917)*x_ref[0]+(-0.0679614650591)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.512691010939)*x[0]+(-1.31295640543)*x[1]+(1.45426060088)*x[2]
        s[0,0,0,1]=(-0.57748372698)*x[0]+(0.215297337003)*x[1]+(0.0853439946124)*x[2]
        s[0,0,1,0]=(-0.785306554604)*x[0]+(-0.0274911695421)*x[1]+(0.693440767035)*x[2]
        s[0,0,1,1]=(-0.56932407873)*x[0]+(-0.981593188827)*x[1]+(0.87532777027)*x[2]
        s[0,0,2,0]=(0.145925104392)*x[0]+(1.45139125753)*x[1]+(1.25041746183)*x[2]
        s[0,0,2,1]=(-1.70110072582)*x[0]+(-0.0335136921878)*x[1]+(0.276204889346)*x[2]
        s[0,1,0,0]=(-0.415026748605)*x[0]+(-0.322339100074)*x[1]+(1.1050553524)*x[2]
        s[0,1,0,1]=(0.267336702938)*x[0]+(-1.44498697586)*x[1]+(0.600455793139)*x[2]
        s[0,1,1,0]=(-0.46618843993)*x[0]+(0.318230555457)*x[1]+(0.281788961443)*x[2]
        s[0,1,1,1]=(-1.12963168407)*x[0]+(-0.661100272253)*x[1]+(0.293642402995)*x[2]
        s[0,1,2,0]=(-0.0041407279814)*x[0]+(0.330916770241)*x[1]+(1.30265962609)*x[2]
        s[0,1,2,1]=(1.37513192104)*x[0]+(-1.39837371617)*x[1]+(-1.23490743489)*x[2]
        s[0,2,0,0]=(0.240920374522)*x[0]+(1.30092830627)*x[1]+(0.189720240056)*x[2]
        s[0,2,0,1]=(0.741337711467)*x[0]+(0.00893414596869)*x[1]+(1.22438764071)*x[2]
        s[0,2,1,0]=(1.0524903095)*x[0]+(-1.00088524111)*x[1]+(0.985885896119)*x[2]
        s[0,2,1,1]=(-0.252168964788)*x[0]+(1.2685968377)*x[1]+(-0.025373437998)*x[2]
        s[0,2,2,0]=(0.728523707179)*x[0]+(0.00656996105623)*x[1]+(1.85098622635)*x[2]
        s[0,2,2,1]=(0.193123464463)*x[0]+(-0.426452202023)*x[1]+(-0.144471734372)*x[2]
        s[0,3,0,0]=(0.446506355449)*x[0]+(0.379921574537)*x[1]+(-0.303150079281)*x[2]
        s[0,3,0,1]=(0.455373734156)*x[0]+(0.134600165673)*x[1]+(-0.604957024603)*x[2]
        s[0,3,1,0]=(-1.63808806962)*x[0]+(1.02897840042)*x[1]+(1.18123170106)*x[2]
        s[0,3,1,1]=(0.587448315418)*x[0]+(0.0991725296547)*x[1]+(1.33682286695)*x[2]
        s[0,3,2,0]=(-1.12085613449)*x[0]+(0.454771039944)*x[1]+(-0.131089499156)*x[2]
        s[0,3,2,1]=(-0.351021639018)*x[0]+(0.47905316732)*x[1]+(-1.07449842534)*x[2]
        s[0,4,0,0]=(0.232526828642)*x[0]+(1.33907066655)*x[1]+(-0.699701401092)*x[2]
        s[0,4,0,1]=(0.244905848535)*x[0]+(-0.799313917773)*x[1]+(-0.16513296799)*x[2]
        s[0,4,1,0]=(-1.10152273815)*x[0]+(0.898176161732)*x[1]+(-0.101734265033)*x[2]
        s[0,4,1,1]=(-0.0804598444811)*x[0]+(-0.77819195606)*x[1]+(0.277776761782)*x[2]
        s[0,4,2,0]=(1.05979692744)*x[0]+(-0.195285573862)*x[1]+(-0.0920717991265)*x[2]
        s[0,4,2,1]=(0.44924595646)*x[0]+(0.409652246508)*x[1]+(-0.0705754685165)*x[2]
        s[1,0,0,0]=(1.14991675937)*x[0]+(-0.435947647215)*x[1]+(-0.45614944131)*x[2]
        s[1,0,0,1]=(0.306890350036)*x[0]+(-1.4400848018)*x[1]+(-0.744474442385)*x[2]
        s[1,0,1,0]=(1.07930565176)*x[0]+(-0.273335650157)*x[1]+(-0.230960378838)*x[2]
        s[1,0,1,1]=(-0.97797563017)*x[0]+(-1.3159717465)*x[1]+(-0.357902208089)*x[2]
        s[1,0,2,0]=(-1.02063328249)*x[0]+(-1.18907185687)*x[1]+(-0.509296457902)*x[2]
        s[1,0,2,1]=(-0.508446666232)*x[0]+(-0.69279963667)*x[1]+(-0.497990637708)*x[2]
        s[1,1,0,0]=(-0.476141863054)*x[0]+(-0.180458121777)*x[1]+(-0.318855084826)*x[2]
        s[1,1,0,1]=(0.480566371651)*x[0]+(-1.74196768367)*x[1]+(1.23057574863)*x[2]
        s[1,1,1,0]=(-0.0213138624722)*x[0]+(-0.0846532154764)*x[1]+(-0.61169792023)*x[2]
        s[1,1,1,1]=(-0.389381507111)*x[0]+(0.436088515729)*x[1]+(-0.603019053837)*x[2]
        s[1,1,2,0]=(-1.24478786129)*x[0]+(0.908984432701)*x[1]+(-1.49483946407)*x[2]
        s[1,1,2,1]=(0.866266264162)*x[0]+(-0.190579076131)*x[1]+(-1.63524384216)*x[2]
        s[1,2,0,0]=(0.0333832222755)*x[0]+(-0.551591529565)*x[1]+(1.22170194609)*x[2]
        s[1,2,0,1]=(-0.551682390802)*x[0]+(0.223940652339)*x[1]+(0.141063634211)*x[2]
        s[1,2,1,0]=(1.17072817252)*x[0]+(0.28318356571)*x[1]+(-0.788096581969)*x[2]
        s[1,2,1,1]=(-1.01576849129)*x[0]+(-0.502669485722)*x[1]+(-0.966677100703)*x[2]
        s[1,2,2,0]=(-1.59522074083)*x[0]+(-1.16016984107)*x[1]+(-0.117030183187)*x[2]
        s[1,2,2,1]=(-0.979165259497)*x[0]+(-0.707460176459)*x[1]+(-0.275081507376)*x[2]
        s[1,3,0,0]=(0.692493088771)*x[0]+(-1.04755874336)*x[1]+(-0.373632326826)*x[2]
        s[1,3,0,1]=(0.717001163418)*x[0]+(-0.782379062328)*x[1]+(0.479776988529)*x[2]
        s[1,3,1,0]=(-0.733062748373)*x[0]+(-0.217972411497)*x[1]+(1.43785098293)*x[2]
        s[1,3,1,1]=(0.186430904908)*x[0]+(0.375218175232)*x[1]+(-0.461668650588)*x[2]
        s[1,3,2,0]=(-1.12951451301)*x[0]+(-1.07034955937)*x[1]+(-0.59888614247)*x[2]
        s[1,3,2,1]=(1.34023062833)*x[0]+(-0.520544300692)*x[1]+(-0.363001059538)*x[2]
        s[1,4,0,0]=(0.42295611834)*x[0]+(-0.718852108989)*x[1]+(-1.61995684533)*x[2]
        s[1,4,0,1]=(-0.422467947319)*x[0]+(-0.209334480945)*x[1]+(1.59579413592)*x[2]
        s[1,4,1,0]=(0.387543472208)*x[0]+(-0.0579909254613)*x[1]+(-1.03104850867)*x[2]
        s[1,4,1,1]=(-1.15893685519)*x[0]+(-0.971412946627)*x[1]+(-0.583440884462)*x[2]
        s[1,4,2,0]=(0.554042723162)*x[0]+(0.082633597735)*x[1]+(0.990177730288)*x[2]
        s[1,4,2,1]=(0.150114027051)*x[0]+(0.479664757576)*x[1]+(-1.5769513289)*x[2]
        s[2,0,0,0]=(1.33963588901)*x[0]+(0.146971184264)*x[1]+(-1.29921529821)*x[2]
        s[2,0,0,1]=(0.0115342792238)*x[0]+(0.0451643711892)*x[1]+(-0.656407384758)*x[2]
        s[2,0,1,0]=(0.175368903926)*x[0]+(-0.40573949846)*x[1]+(0.295387861218)*x[2]
        s[2,0,1,1]=(0.531254864472)*x[0]+(-0.536193761259)*x[1]+(-1.21752823876)*x[2]
        s[2,0,2,0]=(-0.362296220089)*x[0]+(0.378651497918)*x[1]+(-0.535206713131)*x[2]
        s[2,0,2,1]=(-0.85728549969)*x[0]+(0.233179293609)*x[1]+(0.345412963315)*x[2]
        s[2,1,0,0]=(1.01059063177)*x[0]+(0.808116187589)*x[1]+(0.468588054081)*x[2]
        s[2,1,0,1]=(-1.17212735529)*x[0]+(-0.871560017197)*x[1]+(1.10553749428)*x[2]
        s[2,1,1,0]=(1.00177793873)*x[0]+(-0.573447532545)*x[1]+(-0.80780539944)*x[2]
        s[2,1,1,1]=(-0.101912168689)*x[0]+(0.0054724970873)*x[1]+(0.873799338046)*x[2]
        s[2,1,2,0]=(0.162601365201)*x[0]+(-0.508620489728)*x[1]+(0.114861348434)*x[2]
        s[2,1,2,1]=(-1.48812556245)*x[0]+(0.29841054692)*x[1]+(0.773432180321)*x[2]
        s[2,2,0,0]=(0.0177469158963)*x[0]+(0.140208281265)*x[1]+(1.12393579682)*x[2]
        s[2,2,0,1]=(0.127835907966)*x[0]+(0.150072336346)*x[1]+(1.59437274191)*x[2]
        s[2,2,1,0]=(-0.549906307264)*x[0]+(-0.572101026775)*x[1]+(-0.747028158715)*x[2]
        s[2,2,1,1]=(0.134272548587)*x[0]+(0.866777056589)*x[1]+(-0.0658173600046)*x[2]
        s[2,2,2,0]=(0.160006947347)*x[0]+(-0.940281662941)*x[1]+(0.671280858545)*x[2]
        s[2,2,2,1]=(-1.55550646869)*x[0]+(0.0485277897059)*x[1]+(1.02027216264)*x[2]
        s[2,3,0,0]=(0.0931028968111)*x[0]+(0.194744643733)*x[1]+(1.7368246452)*x[2]
        s[2,3,0,1]=(1.66077510679)*x[0]+(-0.342785893413)*x[1]+(-0.278302501974)*x[2]
        s[2,3,1,0]=(-0.0719600278272)*x[0]+(0.898388137955)*x[1]+(-0.110811429033)*x[2]
        s[2,3,1,1]=(-0.781997247811)*x[0]+(-1.32497050043)*x[1]+(0.544850665961)*x[2]
        s[2,3,2,0]=(0.0290847482287)*x[0]+(0.895295560038)*x[1]+(-1.68555124085)*x[2]
        s[2,3,2,1]=(0.200540179332)*x[0]+(-0.843842814948)*x[1]+(-0.943779465271)*x[2]
        s[2,4,0,0]=(1.08999954494)*x[0]+(-1.32324120166)*x[1]+(-0.0152368215856)*x[2]
        s[2,4,0,1]=(1.02219584162)*x[0]+(0.208263165731)*x[1]+(0.0703408478849)*x[2]
        s[2,4,1,0]=(0.433866479442)*x[0]+(0.327590037731)*x[1]+(0.752498033029)*x[2]
        s[2,4,1,1]=(1.16322078616)*x[0]+(-1.38234503561)*x[1]+(-0.129845447078)*x[2]
        s[2,4,2,0]=(-0.430572860596)*x[0]+(-1.48079933792)*x[1]+(0.310197458762)*x[2]
        s[2,4,2,1]=(-0.0901138166505)*x[0]+(0.130679443905)*x[1]+(-0.482309483824)*x[2]
        s[3,0,0,0]=(0.239787389644)*x[0]+(0.690754646072)*x[1]+(0.745418481522)*x[2]
        s[3,0,0,1]=(0.346448639801)*x[0]+(-0.287823450604)*x[1]+(0.287003737791)*x[2]
        s[3,0,1,0]=(-1.50502949668)*x[0]+(-1.67521500964)*x[1]+(0.31630953667)*x[2]
        s[3,0,1,1]=(-0.594355763268)*x[0]+(-0.262023313233)*x[1]+(-1.41203969641)*x[2]
        s[3,0,2,0]=(-0.0275545216089)*x[0]+(0.367433407785)*x[1]+(-1.420568196)*x[2]
        s[3,0,2,1]=(-1.18797738587)*x[0]+(-0.496775264183)*x[1]+(0.513942381739)*x[2]
        s[3,1,0,0]=(0.475014043678)*x[0]+(-1.2205702601)*x[1]+(1.06353608529)*x[2]
        s[3,1,0,1]=(-0.0685583494755)*x[0]+(-1.49448219897)*x[1]+(-1.23320402959)*x[2]
        s[3,1,1,0]=(-0.216184790554)*x[0]+(0.554046932828)*x[1]+(0.697487187904)*x[2]
        s[3,1,1,1]=(-0.495370181694)*x[0]+(-1.49207400185)*x[1]+(-0.233763280927)*x[2]
        s[3,1,2,0]=(-0.0462888472218)*x[0]+(0.371127867531)*x[1]+(0.534977833221)*x[2]
        s[3,1,2,1]=(-1.27057460029)*x[0]+(-0.572725702684)*x[1]+(-0.871217452212)*x[2]
        s[3,2,0,0]=(-0.208549808264)*x[0]+(0.847294134183)*x[1]+(0.643419233684)*x[2]
        s[3,2,0,1]=(-0.850445146291)*x[0]+(1.19605447923)*x[1]+(-1.52958469348)*x[2]
        s[3,2,1,0]=(0.181672364304)*x[0]+(-1.60804541136)*x[1]+(0.068065187694)*x[2]
        s[3,2,1,1]=(-0.195284925405)*x[0]+(0.330399376496)*x[1]+(-0.453658972198)*x[2]
        s[3,2,2,0]=(1.44390910788)*x[0]+(0.718618445973)*x[1]+(-0.800501828022)*x[2]
        s[3,2,2,1]=(0.491997052311)*x[0]+(-0.202966322046)*x[1]+(1.39185177314)*x[2]
        s[3,3,0,0]=(-0.670818169452)*x[0]+(0.842747476683)*x[1]+(-0.705022077785)*x[2]
        s[3,3,0,1]=(0.729442429399)*x[0]+(-1.47998423588)*x[1]+(1.72626243143)*x[2]
        s[3,3,1,0]=(-1.08179915812)*x[0]+(-0.117953758788)*x[1]+(1.60080937509)*x[2]
        s[3,3,1,1]=(0.110049458343)*x[0]+(0.571461685559)*x[1]+(-0.483581817598)*x[2]
        s[3,3,2,0]=(-0.386163706678)*x[0]+(-0.146336916781)*x[1]+(0.0360311139076)*x[2]
        s[3,3,2,1]=(-0.779748009644)*x[0]+(-0.780523381896)*x[1]+(-1.20157768536)*x[2]
        s[3,4,0,0]=(1.74580569218)*x[0]+(-0.318257555715)*x[1]+(-0.33745202709)*x[2]
        s[3,4,0,1]=(1.1463597084)*x[0]+(0.516488947224)*x[1]+(-0.205035082746)*x[2]
        s[3,4,1,0]=(0.647937211136)*x[0]+(0.574941676843)*x[1]+(-1.00930234083)*x[2]
        s[3,4,1,1]=(-0.79188909394)*x[0]+(0.770198112387)*x[1]+(-0.0146013355132)*x[2]
        s[3,4,2,0]=(-1.70734926495)*x[0]+(-1.01352599249)*x[1]+(0.878522778655)*x[2]
        s[3,4,2,1]=(0.0249166011531)*x[0]+(-0.0714570534569)*x[1]+(1.10212723792)*x[2]
        ref[0,0,0,0]=(-0.512691010939)*x_ref[0]+(-1.31295640543)*x_ref[1]+(1.45426060088)*x_ref[2]
        ref[0,0,0,1]=(-0.57748372698)*x_ref[0]+(0.215297337003)*x_ref[1]+(0.0853439946124)*x_ref[2]
        ref[0,0,1,0]=(-0.785306554604)*x_ref[0]+(-0.0274911695421)*x_ref[1]+(0.693440767035)*x_ref[2]
        ref[0,0,1,1]=(-0.56932407873)*x_ref[0]+(-0.981593188827)*x_ref[1]+(0.87532777027)*x_ref[2]
        ref[0,0,2,0]=(0.145925104392)*x_ref[0]+(1.45139125753)*x_ref[1]+(1.25041746183)*x_ref[2]
        ref[0,0,2,1]=(-1.70110072582)*x_ref[0]+(-0.0335136921878)*x_ref[1]+(0.276204889346)*x_ref[2]
        ref[0,1,0,0]=(-0.415026748605)*x_ref[0]+(-0.322339100074)*x_ref[1]+(1.1050553524)*x_ref[2]
        ref[0,1,0,1]=(0.267336702938)*x_ref[0]+(-1.44498697586)*x_ref[1]+(0.600455793139)*x_ref[2]
        ref[0,1,1,0]=(-0.46618843993)*x_ref[0]+(0.318230555457)*x_ref[1]+(0.281788961443)*x_ref[2]
        ref[0,1,1,1]=(-1.12963168407)*x_ref[0]+(-0.661100272253)*x_ref[1]+(0.293642402995)*x_ref[2]
        ref[0,1,2,0]=(-0.0041407279814)*x_ref[0]+(0.330916770241)*x_ref[1]+(1.30265962609)*x_ref[2]
        ref[0,1,2,1]=(1.37513192104)*x_ref[0]+(-1.39837371617)*x_ref[1]+(-1.23490743489)*x_ref[2]
        ref[0,2,0,0]=(0.240920374522)*x_ref[0]+(1.30092830627)*x_ref[1]+(0.189720240056)*x_ref[2]
        ref[0,2,0,1]=(0.741337711467)*x_ref[0]+(0.00893414596869)*x_ref[1]+(1.22438764071)*x_ref[2]
        ref[0,2,1,0]=(1.0524903095)*x_ref[0]+(-1.00088524111)*x_ref[1]+(0.985885896119)*x_ref[2]
        ref[0,2,1,1]=(-0.252168964788)*x_ref[0]+(1.2685968377)*x_ref[1]+(-0.025373437998)*x_ref[2]
        ref[0,2,2,0]=(0.728523707179)*x_ref[0]+(0.00656996105623)*x_ref[1]+(1.85098622635)*x_ref[2]
        ref[0,2,2,1]=(0.193123464463)*x_ref[0]+(-0.426452202023)*x_ref[1]+(-0.144471734372)*x_ref[2]
        ref[0,3,0,0]=(0.446506355449)*x_ref[0]+(0.379921574537)*x_ref[1]+(-0.303150079281)*x_ref[2]
        ref[0,3,0,1]=(0.455373734156)*x_ref[0]+(0.134600165673)*x_ref[1]+(-0.604957024603)*x_ref[2]
        ref[0,3,1,0]=(-1.63808806962)*x_ref[0]+(1.02897840042)*x_ref[1]+(1.18123170106)*x_ref[2]
        ref[0,3,1,1]=(0.587448315418)*x_ref[0]+(0.0991725296547)*x_ref[1]+(1.33682286695)*x_ref[2]
        ref[0,3,2,0]=(-1.12085613449)*x_ref[0]+(0.454771039944)*x_ref[1]+(-0.131089499156)*x_ref[2]
        ref[0,3,2,1]=(-0.351021639018)*x_ref[0]+(0.47905316732)*x_ref[1]+(-1.07449842534)*x_ref[2]
        ref[0,4,0,0]=(0.232526828642)*x_ref[0]+(1.33907066655)*x_ref[1]+(-0.699701401092)*x_ref[2]
        ref[0,4,0,1]=(0.244905848535)*x_ref[0]+(-0.799313917773)*x_ref[1]+(-0.16513296799)*x_ref[2]
        ref[0,4,1,0]=(-1.10152273815)*x_ref[0]+(0.898176161732)*x_ref[1]+(-0.101734265033)*x_ref[2]
        ref[0,4,1,1]=(-0.0804598444811)*x_ref[0]+(-0.77819195606)*x_ref[1]+(0.277776761782)*x_ref[2]
        ref[0,4,2,0]=(1.05979692744)*x_ref[0]+(-0.195285573862)*x_ref[1]+(-0.0920717991265)*x_ref[2]
        ref[0,4,2,1]=(0.44924595646)*x_ref[0]+(0.409652246508)*x_ref[1]+(-0.0705754685165)*x_ref[2]
        ref[1,0,0,0]=(1.14991675937)*x_ref[0]+(-0.435947647215)*x_ref[1]+(-0.45614944131)*x_ref[2]
        ref[1,0,0,1]=(0.306890350036)*x_ref[0]+(-1.4400848018)*x_ref[1]+(-0.744474442385)*x_ref[2]
        ref[1,0,1,0]=(1.07930565176)*x_ref[0]+(-0.273335650157)*x_ref[1]+(-0.230960378838)*x_ref[2]
        ref[1,0,1,1]=(-0.97797563017)*x_ref[0]+(-1.3159717465)*x_ref[1]+(-0.357902208089)*x_ref[2]
        ref[1,0,2,0]=(-1.02063328249)*x_ref[0]+(-1.18907185687)*x_ref[1]+(-0.509296457902)*x_ref[2]
        ref[1,0,2,1]=(-0.508446666232)*x_ref[0]+(-0.69279963667)*x_ref[1]+(-0.497990637708)*x_ref[2]
        ref[1,1,0,0]=(-0.476141863054)*x_ref[0]+(-0.180458121777)*x_ref[1]+(-0.318855084826)*x_ref[2]
        ref[1,1,0,1]=(0.480566371651)*x_ref[0]+(-1.74196768367)*x_ref[1]+(1.23057574863)*x_ref[2]
        ref[1,1,1,0]=(-0.0213138624722)*x_ref[0]+(-0.0846532154764)*x_ref[1]+(-0.61169792023)*x_ref[2]
        ref[1,1,1,1]=(-0.389381507111)*x_ref[0]+(0.436088515729)*x_ref[1]+(-0.603019053837)*x_ref[2]
        ref[1,1,2,0]=(-1.24478786129)*x_ref[0]+(0.908984432701)*x_ref[1]+(-1.49483946407)*x_ref[2]
        ref[1,1,2,1]=(0.866266264162)*x_ref[0]+(-0.190579076131)*x_ref[1]+(-1.63524384216)*x_ref[2]
        ref[1,2,0,0]=(0.0333832222755)*x_ref[0]+(-0.551591529565)*x_ref[1]+(1.22170194609)*x_ref[2]
        ref[1,2,0,1]=(-0.551682390802)*x_ref[0]+(0.223940652339)*x_ref[1]+(0.141063634211)*x_ref[2]
        ref[1,2,1,0]=(1.17072817252)*x_ref[0]+(0.28318356571)*x_ref[1]+(-0.788096581969)*x_ref[2]
        ref[1,2,1,1]=(-1.01576849129)*x_ref[0]+(-0.502669485722)*x_ref[1]+(-0.966677100703)*x_ref[2]
        ref[1,2,2,0]=(-1.59522074083)*x_ref[0]+(-1.16016984107)*x_ref[1]+(-0.117030183187)*x_ref[2]
        ref[1,2,2,1]=(-0.979165259497)*x_ref[0]+(-0.707460176459)*x_ref[1]+(-0.275081507376)*x_ref[2]
        ref[1,3,0,0]=(0.692493088771)*x_ref[0]+(-1.04755874336)*x_ref[1]+(-0.373632326826)*x_ref[2]
        ref[1,3,0,1]=(0.717001163418)*x_ref[0]+(-0.782379062328)*x_ref[1]+(0.479776988529)*x_ref[2]
        ref[1,3,1,0]=(-0.733062748373)*x_ref[0]+(-0.217972411497)*x_ref[1]+(1.43785098293)*x_ref[2]
        ref[1,3,1,1]=(0.186430904908)*x_ref[0]+(0.375218175232)*x_ref[1]+(-0.461668650588)*x_ref[2]
        ref[1,3,2,0]=(-1.12951451301)*x_ref[0]+(-1.07034955937)*x_ref[1]+(-0.59888614247)*x_ref[2]
        ref[1,3,2,1]=(1.34023062833)*x_ref[0]+(-0.520544300692)*x_ref[1]+(-0.363001059538)*x_ref[2]
        ref[1,4,0,0]=(0.42295611834)*x_ref[0]+(-0.718852108989)*x_ref[1]+(-1.61995684533)*x_ref[2]
        ref[1,4,0,1]=(-0.422467947319)*x_ref[0]+(-0.209334480945)*x_ref[1]+(1.59579413592)*x_ref[2]
        ref[1,4,1,0]=(0.387543472208)*x_ref[0]+(-0.0579909254613)*x_ref[1]+(-1.03104850867)*x_ref[2]
        ref[1,4,1,1]=(-1.15893685519)*x_ref[0]+(-0.971412946627)*x_ref[1]+(-0.583440884462)*x_ref[2]
        ref[1,4,2,0]=(0.554042723162)*x_ref[0]+(0.082633597735)*x_ref[1]+(0.990177730288)*x_ref[2]
        ref[1,4,2,1]=(0.150114027051)*x_ref[0]+(0.479664757576)*x_ref[1]+(-1.5769513289)*x_ref[2]
        ref[2,0,0,0]=(1.33963588901)*x_ref[0]+(0.146971184264)*x_ref[1]+(-1.29921529821)*x_ref[2]
        ref[2,0,0,1]=(0.0115342792238)*x_ref[0]+(0.0451643711892)*x_ref[1]+(-0.656407384758)*x_ref[2]
        ref[2,0,1,0]=(0.175368903926)*x_ref[0]+(-0.40573949846)*x_ref[1]+(0.295387861218)*x_ref[2]
        ref[2,0,1,1]=(0.531254864472)*x_ref[0]+(-0.536193761259)*x_ref[1]+(-1.21752823876)*x_ref[2]
        ref[2,0,2,0]=(-0.362296220089)*x_ref[0]+(0.378651497918)*x_ref[1]+(-0.535206713131)*x_ref[2]
        ref[2,0,2,1]=(-0.85728549969)*x_ref[0]+(0.233179293609)*x_ref[1]+(0.345412963315)*x_ref[2]
        ref[2,1,0,0]=(1.01059063177)*x_ref[0]+(0.808116187589)*x_ref[1]+(0.468588054081)*x_ref[2]
        ref[2,1,0,1]=(-1.17212735529)*x_ref[0]+(-0.871560017197)*x_ref[1]+(1.10553749428)*x_ref[2]
        ref[2,1,1,0]=(1.00177793873)*x_ref[0]+(-0.573447532545)*x_ref[1]+(-0.80780539944)*x_ref[2]
        ref[2,1,1,1]=(-0.101912168689)*x_ref[0]+(0.0054724970873)*x_ref[1]+(0.873799338046)*x_ref[2]
        ref[2,1,2,0]=(0.162601365201)*x_ref[0]+(-0.508620489728)*x_ref[1]+(0.114861348434)*x_ref[2]
        ref[2,1,2,1]=(-1.48812556245)*x_ref[0]+(0.29841054692)*x_ref[1]+(0.773432180321)*x_ref[2]
        ref[2,2,0,0]=(0.0177469158963)*x_ref[0]+(0.140208281265)*x_ref[1]+(1.12393579682)*x_ref[2]
        ref[2,2,0,1]=(0.127835907966)*x_ref[0]+(0.150072336346)*x_ref[1]+(1.59437274191)*x_ref[2]
        ref[2,2,1,0]=(-0.549906307264)*x_ref[0]+(-0.572101026775)*x_ref[1]+(-0.747028158715)*x_ref[2]
        ref[2,2,1,1]=(0.134272548587)*x_ref[0]+(0.866777056589)*x_ref[1]+(-0.0658173600046)*x_ref[2]
        ref[2,2,2,0]=(0.160006947347)*x_ref[0]+(-0.940281662941)*x_ref[1]+(0.671280858545)*x_ref[2]
        ref[2,2,2,1]=(-1.55550646869)*x_ref[0]+(0.0485277897059)*x_ref[1]+(1.02027216264)*x_ref[2]
        ref[2,3,0,0]=(0.0931028968111)*x_ref[0]+(0.194744643733)*x_ref[1]+(1.7368246452)*x_ref[2]
        ref[2,3,0,1]=(1.66077510679)*x_ref[0]+(-0.342785893413)*x_ref[1]+(-0.278302501974)*x_ref[2]
        ref[2,3,1,0]=(-0.0719600278272)*x_ref[0]+(0.898388137955)*x_ref[1]+(-0.110811429033)*x_ref[2]
        ref[2,3,1,1]=(-0.781997247811)*x_ref[0]+(-1.32497050043)*x_ref[1]+(0.544850665961)*x_ref[2]
        ref[2,3,2,0]=(0.0290847482287)*x_ref[0]+(0.895295560038)*x_ref[1]+(-1.68555124085)*x_ref[2]
        ref[2,3,2,1]=(0.200540179332)*x_ref[0]+(-0.843842814948)*x_ref[1]+(-0.943779465271)*x_ref[2]
        ref[2,4,0,0]=(1.08999954494)*x_ref[0]+(-1.32324120166)*x_ref[1]+(-0.0152368215856)*x_ref[2]
        ref[2,4,0,1]=(1.02219584162)*x_ref[0]+(0.208263165731)*x_ref[1]+(0.0703408478849)*x_ref[2]
        ref[2,4,1,0]=(0.433866479442)*x_ref[0]+(0.327590037731)*x_ref[1]+(0.752498033029)*x_ref[2]
        ref[2,4,1,1]=(1.16322078616)*x_ref[0]+(-1.38234503561)*x_ref[1]+(-0.129845447078)*x_ref[2]
        ref[2,4,2,0]=(-0.430572860596)*x_ref[0]+(-1.48079933792)*x_ref[1]+(0.310197458762)*x_ref[2]
        ref[2,4,2,1]=(-0.0901138166505)*x_ref[0]+(0.130679443905)*x_ref[1]+(-0.482309483824)*x_ref[2]
        ref[3,0,0,0]=(0.239787389644)*x_ref[0]+(0.690754646072)*x_ref[1]+(0.745418481522)*x_ref[2]
        ref[3,0,0,1]=(0.346448639801)*x_ref[0]+(-0.287823450604)*x_ref[1]+(0.287003737791)*x_ref[2]
        ref[3,0,1,0]=(-1.50502949668)*x_ref[0]+(-1.67521500964)*x_ref[1]+(0.31630953667)*x_ref[2]
        ref[3,0,1,1]=(-0.594355763268)*x_ref[0]+(-0.262023313233)*x_ref[1]+(-1.41203969641)*x_ref[2]
        ref[3,0,2,0]=(-0.0275545216089)*x_ref[0]+(0.367433407785)*x_ref[1]+(-1.420568196)*x_ref[2]
        ref[3,0,2,1]=(-1.18797738587)*x_ref[0]+(-0.496775264183)*x_ref[1]+(0.513942381739)*x_ref[2]
        ref[3,1,0,0]=(0.475014043678)*x_ref[0]+(-1.2205702601)*x_ref[1]+(1.06353608529)*x_ref[2]
        ref[3,1,0,1]=(-0.0685583494755)*x_ref[0]+(-1.49448219897)*x_ref[1]+(-1.23320402959)*x_ref[2]
        ref[3,1,1,0]=(-0.216184790554)*x_ref[0]+(0.554046932828)*x_ref[1]+(0.697487187904)*x_ref[2]
        ref[3,1,1,1]=(-0.495370181694)*x_ref[0]+(-1.49207400185)*x_ref[1]+(-0.233763280927)*x_ref[2]
        ref[3,1,2,0]=(-0.0462888472218)*x_ref[0]+(0.371127867531)*x_ref[1]+(0.534977833221)*x_ref[2]
        ref[3,1,2,1]=(-1.27057460029)*x_ref[0]+(-0.572725702684)*x_ref[1]+(-0.871217452212)*x_ref[2]
        ref[3,2,0,0]=(-0.208549808264)*x_ref[0]+(0.847294134183)*x_ref[1]+(0.643419233684)*x_ref[2]
        ref[3,2,0,1]=(-0.850445146291)*x_ref[0]+(1.19605447923)*x_ref[1]+(-1.52958469348)*x_ref[2]
        ref[3,2,1,0]=(0.181672364304)*x_ref[0]+(-1.60804541136)*x_ref[1]+(0.068065187694)*x_ref[2]
        ref[3,2,1,1]=(-0.195284925405)*x_ref[0]+(0.330399376496)*x_ref[1]+(-0.453658972198)*x_ref[2]
        ref[3,2,2,0]=(1.44390910788)*x_ref[0]+(0.718618445973)*x_ref[1]+(-0.800501828022)*x_ref[2]
        ref[3,2,2,1]=(0.491997052311)*x_ref[0]+(-0.202966322046)*x_ref[1]+(1.39185177314)*x_ref[2]
        ref[3,3,0,0]=(-0.670818169452)*x_ref[0]+(0.842747476683)*x_ref[1]+(-0.705022077785)*x_ref[2]
        ref[3,3,0,1]=(0.729442429399)*x_ref[0]+(-1.47998423588)*x_ref[1]+(1.72626243143)*x_ref[2]
        ref[3,3,1,0]=(-1.08179915812)*x_ref[0]+(-0.117953758788)*x_ref[1]+(1.60080937509)*x_ref[2]
        ref[3,3,1,1]=(0.110049458343)*x_ref[0]+(0.571461685559)*x_ref[1]+(-0.483581817598)*x_ref[2]
        ref[3,3,2,0]=(-0.386163706678)*x_ref[0]+(-0.146336916781)*x_ref[1]+(0.0360311139076)*x_ref[2]
        ref[3,3,2,1]=(-0.779748009644)*x_ref[0]+(-0.780523381896)*x_ref[1]+(-1.20157768536)*x_ref[2]
        ref[3,4,0,0]=(1.74580569218)*x_ref[0]+(-0.318257555715)*x_ref[1]+(-0.33745202709)*x_ref[2]
        ref[3,4,0,1]=(1.1463597084)*x_ref[0]+(0.516488947224)*x_ref[1]+(-0.205035082746)*x_ref[2]
        ref[3,4,1,0]=(0.647937211136)*x_ref[0]+(0.574941676843)*x_ref[1]+(-1.00930234083)*x_ref[2]
        ref[3,4,1,1]=(-0.79188909394)*x_ref[0]+(0.770198112387)*x_ref[1]+(-0.0146013355132)*x_ref[2]
        ref[3,4,2,0]=(-1.70734926495)*x_ref[0]+(-1.01352599249)*x_ref[1]+(0.878522778655)*x_ref[2]
        ref[3,4,2,1]=(0.0249166011531)*x_ref[0]+(-0.0714570534569)*x_ref[1]+(1.10212723792)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.318228469869)*x[0]+(-0.774132349214)*x[1]
        ref=(0.318228469869)*x_ref[0]+(-0.774132349214)*x_ref[1]
      else:
        s=(0.377557485345)*x[0]+(1.02901239394)*x[1]+(0.038224819288)*x[2]
        ref=(0.377557485345)*x_ref[0]+(1.02901239394)*x_ref[1]+(0.038224819288)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        s[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
        ref[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        ref[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
      else:
        s[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        s[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
        ref[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        ref[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0621856629646)*x[0]+(0.783319354938)*x[1]
        s[0,1]=(-0.205526239577)*x[0]+(-0.401770179168)*x[1]
        s[0,2]=(0.543085761147)*x[0]+(0.915124776348)*x[1]
        s[0,3]=(1.3114416307)*x[0]+(-0.339977887579)*x[1]
        s[0,4]=(0.19180167645)*x[0]+(0.0744591496582)*x[1]
        s[1,0]=(0.833949474687)*x[0]+(-0.513960535503)*x[1]
        s[1,1]=(1.63749958183)*x[0]+(1.49589084809)*x[1]
        s[1,2]=(0.52690970122)*x[0]+(-0.497651760738)*x[1]
        s[1,3]=(0.157987079101)*x[0]+(0.244658238724)*x[1]
        s[1,4]=(-0.557780706846)*x[0]+(-1.49932552405)*x[1]
        s[2,0]=(-0.00409488968736)*x[0]+(0.366430715843)*x[1]
        s[2,1]=(-0.867654712773)*x[0]+(-0.129712066272)*x[1]
        s[2,2]=(1.33589989715)*x[0]+(-0.135732989085)*x[1]
        s[2,3]=(0.208531325369)*x[0]+(1.3627828974)*x[1]
        s[2,4]=(-0.667299295418)*x[0]+(0.981775718726)*x[1]
        s[3,0]=(1.40151192578)*x[0]+(-0.28272210702)*x[1]
        s[3,1]=(-1.13395485034)*x[0]+(-0.148770918012)*x[1]
        s[3,2]=(0.315288948224)*x[0]+(-0.140365593027)*x[1]
        s[3,3]=(-0.0421809531999)*x[0]+(0.522687748574)*x[1]
        s[3,4]=(0.135043757063)*x[0]+(-1.52489039301)*x[1]
        ref[0,0]=(0.0621856629646)*x_ref[0]+(0.783319354938)*x_ref[1]
        ref[0,1]=(-0.205526239577)*x_ref[0]+(-0.401770179168)*x_ref[1]
        ref[0,2]=(0.543085761147)*x_ref[0]+(0.915124776348)*x_ref[1]
        ref[0,3]=(1.3114416307)*x_ref[0]+(-0.339977887579)*x_ref[1]
        ref[0,4]=(0.19180167645)*x_ref[0]+(0.0744591496582)*x_ref[1]
        ref[1,0]=(0.833949474687)*x_ref[0]+(-0.513960535503)*x_ref[1]
        ref[1,1]=(1.63749958183)*x_ref[0]+(1.49589084809)*x_ref[1]
        ref[1,2]=(0.52690970122)*x_ref[0]+(-0.497651760738)*x_ref[1]
        ref[1,3]=(0.157987079101)*x_ref[0]+(0.244658238724)*x_ref[1]
        ref[1,4]=(-0.557780706846)*x_ref[0]+(-1.49932552405)*x_ref[1]
        ref[2,0]=(-0.00409488968736)*x_ref[0]+(0.366430715843)*x_ref[1]
        ref[2,1]=(-0.867654712773)*x_ref[0]+(-0.129712066272)*x_ref[1]
        ref[2,2]=(1.33589989715)*x_ref[0]+(-0.135732989085)*x_ref[1]
        ref[2,3]=(0.208531325369)*x_ref[0]+(1.3627828974)*x_ref[1]
        ref[2,4]=(-0.667299295418)*x_ref[0]+(0.981775718726)*x_ref[1]
        ref[3,0]=(1.40151192578)*x_ref[0]+(-0.28272210702)*x_ref[1]
        ref[3,1]=(-1.13395485034)*x_ref[0]+(-0.148770918012)*x_ref[1]
        ref[3,2]=(0.315288948224)*x_ref[0]+(-0.140365593027)*x_ref[1]
        ref[3,3]=(-0.0421809531999)*x_ref[0]+(0.522687748574)*x_ref[1]
        ref[3,4]=(0.135043757063)*x_ref[0]+(-1.52489039301)*x_ref[1]
      else:
        s[0,0]=(-1.07097471698)*x[0]+(0.511673191366)*x[1]+(0.986422233856)*x[2]
        s[0,1]=(1.35021866887)*x[0]+(-1.18036897489)*x[1]+(0.143096009704)*x[2]
        s[0,2]=(1.92136475346)*x[0]+(-1.52041779666)*x[1]+(1.50412123174)*x[2]
        s[0,3]=(-1.26349432492)*x[0]+(-1.69090463821)*x[1]+(0.0519848717435)*x[2]
        s[0,4]=(-0.450643290235)*x[0]+(-1.0918592714)*x[1]+(-1.78531176031)*x[2]
        s[1,0]=(1.24575021839)*x[0]+(0.378183524368)*x[1]+(0.164647643767)*x[2]
        s[1,1]=(-0.0164625547855)*x[0]+(-1.06539959746)*x[1]+(-1.5644531354)*x[2]
        s[1,2]=(0.126087223544)*x[0]+(0.876303692726)*x[1]+(1.3521510598)*x[2]
        s[1,3]=(-1.15749467575)*x[0]+(-0.850991452556)*x[1]+(-0.903705777185)*x[2]
        s[1,4]=(-0.680094762766)*x[0]+(0.280093157366)*x[1]+(0.460572306025)*x[2]
        s[2,0]=(-0.201094411285)*x[0]+(0.610240266355)*x[1]+(0.741298609653)*x[2]
        s[2,1]=(0.754412798233)*x[0]+(0.765169641301)*x[1]+(0.522349231028)*x[2]
        s[2,2]=(-0.330239429991)*x[0]+(1.04468899978)*x[1]+(-0.493706328863)*x[2]
        s[2,3]=(-0.107189727407)*x[0]+(-1.57053989855)*x[1]+(1.5864416691)*x[2]
        s[2,4]=(1.78349630817)*x[0]+(0.415138686617)*x[1]+(-0.567656058268)*x[2]
        s[3,0]=(-0.242864411179)*x[0]+(0.342401319249)*x[1]+(-0.963749227433)*x[2]
        s[3,1]=(-0.677695348369)*x[0]+(-1.52126003219)*x[1]+(-0.197904354713)*x[2]
        s[3,2]=(-0.549674242841)*x[0]+(1.23313354594)*x[1]+(0.701202865744)*x[2]
        s[3,3]=(0.278097672739)*x[0]+(-1.31336623233)*x[1]+(1.69771904282)*x[2]
        s[3,4]=(0.0982668788821)*x[0]+(-1.11857597802)*x[1]+(-0.0505715204344)*x[2]
        ref[0,0]=(-1.07097471698)*x_ref[0]+(0.511673191366)*x_ref[1]+(0.986422233856)*x_ref[2]
        ref[0,1]=(1.35021866887)*x_ref[0]+(-1.18036897489)*x_ref[1]+(0.143096009704)*x_ref[2]
        ref[0,2]=(1.92136475346)*x_ref[0]+(-1.52041779666)*x_ref[1]+(1.50412123174)*x_ref[2]
        ref[0,3]=(-1.26349432492)*x_ref[0]+(-1.69090463821)*x_ref[1]+(0.0519848717435)*x_ref[2]
        ref[0,4]=(-0.450643290235)*x_ref[0]+(-1.0918592714)*x_ref[1]+(-1.78531176031)*x_ref[2]
        ref[1,0]=(1.24575021839)*x_ref[0]+(0.378183524368)*x_ref[1]+(0.164647643767)*x_ref[2]
        ref[1,1]=(-0.0164625547855)*x_ref[0]+(-1.06539959746)*x_ref[1]+(-1.5644531354)*x_ref[2]
        ref[1,2]=(0.126087223544)*x_ref[0]+(0.876303692726)*x_ref[1]+(1.3521510598)*x_ref[2]
        ref[1,3]=(-1.15749467575)*x_ref[0]+(-0.850991452556)*x_ref[1]+(-0.903705777185)*x_ref[2]
        ref[1,4]=(-0.680094762766)*x_ref[0]+(0.280093157366)*x_ref[1]+(0.460572306025)*x_ref[2]
        ref[2,0]=(-0.201094411285)*x_ref[0]+(0.610240266355)*x_ref[1]+(0.741298609653)*x_ref[2]
        ref[2,1]=(0.754412798233)*x_ref[0]+(0.765169641301)*x_ref[1]+(0.522349231028)*x_ref[2]
        ref[2,2]=(-0.330239429991)*x_ref[0]+(1.04468899978)*x_ref[1]+(-0.493706328863)*x_ref[2]
        ref[2,3]=(-0.107189727407)*x_ref[0]+(-1.57053989855)*x_ref[1]+(1.5864416691)*x_ref[2]
        ref[2,4]=(1.78349630817)*x_ref[0]+(0.415138686617)*x_ref[1]+(-0.567656058268)*x_ref[2]
        ref[3,0]=(-0.242864411179)*x_ref[0]+(0.342401319249)*x_ref[1]+(-0.963749227433)*x_ref[2]
        ref[3,1]=(-0.677695348369)*x_ref[0]+(-1.52126003219)*x_ref[1]+(-0.197904354713)*x_ref[2]
        ref[3,2]=(-0.549674242841)*x_ref[0]+(1.23313354594)*x_ref[1]+(0.701202865744)*x_ref[2]
        ref[3,3]=(0.278097672739)*x_ref[0]+(-1.31336623233)*x_ref[1]+(1.69771904282)*x_ref[2]
        ref[3,4]=(0.0982668788821)*x_ref[0]+(-1.11857597802)*x_ref[1]+(-0.0505715204344)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.121576835783)*x[0]+(-0.941342629086)*x[1]
        s[0,0,1]=(0.113281061637)*x[0]+(1.53625452894)*x[1]
        s[0,1,0]=(0.0674434661275)*x[0]+(-0.633539894942)*x[1]
        s[0,1,1]=(0.22586221125)*x[0]+(-0.756374425353)*x[1]
        s[1,0,0]=(0.0826651782713)*x[0]+(-0.78928571118)*x[1]
        s[1,0,1]=(-0.844609848337)*x[0]+(1.15779995507)*x[1]
        s[1,1,0]=(0.137225056583)*x[0]+(-1.03725215253)*x[1]
        s[1,1,1]=(1.66612561218)*x[0]+(-0.243423071065)*x[1]
        s[2,0,0]=(0.0833021990187)*x[0]+(-1.2712750293)*x[1]
        s[2,0,1]=(0.13762583458)*x[0]+(-0.385843401588)*x[1]
        s[2,1,0]=(-0.926228562488)*x[0]+(-0.305689405746)*x[1]
        s[2,1,1]=(0.725733502304)*x[0]+(1.40832913183)*x[1]
        s[3,0,0]=(1.09943065211)*x[0]+(1.13514402138)*x[1]
        s[3,0,1]=(-1.0354232693)*x[0]+(-0.678629246255)*x[1]
        s[3,1,0]=(-0.44364521298)*x[0]+(0.857496257444)*x[1]
        s[3,1,1]=(1.53635223417)*x[0]+(-0.0887540544341)*x[1]
        s[4,0,0]=(0.553400675937)*x[0]+(-0.680780568865)*x[1]
        s[4,0,1]=(-0.00341234496916)*x[0]+(0.115104542523)*x[1]
        s[4,1,0]=(1.07927709685)*x[0]+(-0.270560535373)*x[1]
        s[4,1,1]=(-0.681458204636)*x[0]+(-0.607605729645)*x[1]
        s[5,0,0]=(0.10322587433)*x[0]+(0.982393144196)*x[1]
        s[5,0,1]=(-0.229957227498)*x[0]+(0.940017423086)*x[1]
        s[5,1,0]=(-0.465473033025)*x[0]+(-0.61380542705)*x[1]
        s[5,1,1]=(1.43248132924)*x[0]+(-1.71390853514)*x[1]
        ref[0,0,0]=(0.121576835783)*x_ref[0]+(-0.941342629086)*x_ref[1]
        ref[0,0,1]=(0.113281061637)*x_ref[0]+(1.53625452894)*x_ref[1]
        ref[0,1,0]=(0.0674434661275)*x_ref[0]+(-0.633539894942)*x_ref[1]
        ref[0,1,1]=(0.22586221125)*x_ref[0]+(-0.756374425353)*x_ref[1]
        ref[1,0,0]=(0.0826651782713)*x_ref[0]+(-0.78928571118)*x_ref[1]
        ref[1,0,1]=(-0.844609848337)*x_ref[0]+(1.15779995507)*x_ref[1]
        ref[1,1,0]=(0.137225056583)*x_ref[0]+(-1.03725215253)*x_ref[1]
        ref[1,1,1]=(1.66612561218)*x_ref[0]+(-0.243423071065)*x_ref[1]
        ref[2,0,0]=(0.0833021990187)*x_ref[0]+(-1.2712750293)*x_ref[1]
        ref[2,0,1]=(0.13762583458)*x_ref[0]+(-0.385843401588)*x_ref[1]
        ref[2,1,0]=(-0.926228562488)*x_ref[0]+(-0.305689405746)*x_ref[1]
        ref[2,1,1]=(0.725733502304)*x_ref[0]+(1.40832913183)*x_ref[1]
        ref[3,0,0]=(1.09943065211)*x_ref[0]+(1.13514402138)*x_ref[1]
        ref[3,0,1]=(-1.0354232693)*x_ref[0]+(-0.678629246255)*x_ref[1]
        ref[3,1,0]=(-0.44364521298)*x_ref[0]+(0.857496257444)*x_ref[1]
        ref[3,1,1]=(1.53635223417)*x_ref[0]+(-0.0887540544341)*x_ref[1]
        ref[4,0,0]=(0.553400675937)*x_ref[0]+(-0.680780568865)*x_ref[1]
        ref[4,0,1]=(-0.00341234496916)*x_ref[0]+(0.115104542523)*x_ref[1]
        ref[4,1,0]=(1.07927709685)*x_ref[0]+(-0.270560535373)*x_ref[1]
        ref[4,1,1]=(-0.681458204636)*x_ref[0]+(-0.607605729645)*x_ref[1]
        ref[5,0,0]=(0.10322587433)*x_ref[0]+(0.982393144196)*x_ref[1]
        ref[5,0,1]=(-0.229957227498)*x_ref[0]+(0.940017423086)*x_ref[1]
        ref[5,1,0]=(-0.465473033025)*x_ref[0]+(-0.61380542705)*x_ref[1]
        ref[5,1,1]=(1.43248132924)*x_ref[0]+(-1.71390853514)*x_ref[1]
      else:
        s[0,0,0]=(0.663545175027)*x[0]+(-1.05528281336)*x[1]+(-0.25024128376)*x[2]
        s[0,0,1]=(1.10817148343)*x[0]+(0.224734467846)*x[1]+(0.712872841532)*x[2]
        s[0,1,0]=(-0.345072989586)*x[0]+(-1.244678278)*x[1]+(0.103446646879)*x[2]
        s[0,1,1]=(-0.106279741301)*x[0]+(-0.525721333511)*x[1]+(0.401917882446)*x[2]
        s[1,0,0]=(0.838242796285)*x[0]+(-0.706020082817)*x[1]+(0.443650945014)*x[2]
        s[1,0,1]=(-1.16557336842)*x[0]+(0.215324659268)*x[1]+(-0.285210566147)*x[2]
        s[1,1,0]=(-0.141574656707)*x[0]+(1.66054358217)*x[1]+(-0.554616610711)*x[2]
        s[1,1,1]=(0.523682935472)*x[0]+(0.687178896331)*x[1]+(0.875639049678)*x[2]
        s[2,0,0]=(0.585144782395)*x[0]+(0.172063481865)*x[1]+(0.920891875835)*x[2]
        s[2,0,1]=(0.660815435485)*x[0]+(-1.63324849021)*x[1]+(0.0810858621718)*x[2]
        s[2,1,0]=(-0.238166824253)*x[0]+(0.0442487879497)*x[1]+(1.05748755823)*x[2]
        s[2,1,1]=(1.14314890535)*x[0]+(0.63161746042)*x[1]+(1.5024770616)*x[2]
        s[3,0,0]=(-1.33327391952)*x[0]+(1.58312319401)*x[1]+(-1.24202894696)*x[2]
        s[3,0,1]=(-0.0242608819668)*x[0]+(-1.12901018411)*x[1]+(-0.443840381999)*x[2]
        s[3,1,0]=(1.12400540809)*x[0]+(-0.25104247006)*x[1]+(0.55935664575)*x[2]
        s[3,1,1]=(0.976977053811)*x[0]+(1.46134206601)*x[1]+(-0.883829686094)*x[2]
        s[4,0,0]=(0.68585022695)*x[0]+(0.320211070942)*x[1]+(-0.715034754238)*x[2]
        s[4,0,1]=(0.745090118519)*x[0]+(-0.56687124464)*x[1]+(-1.41483630619)*x[2]
        s[4,1,0]=(-0.864645032271)*x[0]+(-0.0852453841156)*x[1]+(-1.02272230917)*x[2]
        s[4,1,1]=(0.243107783262)*x[0]+(-1.76263127813)*x[1]+(-0.0396122398093)*x[2]
        s[5,0,0]=(-1.33890039547)*x[0]+(0.95916460762)*x[1]+(-0.118457193655)*x[2]
        s[5,0,1]=(-1.03526146458)*x[0]+(-0.279038868252)*x[1]+(-0.702378064321)*x[2]
        s[5,1,0]=(-0.434705839506)*x[0]+(0.388520593856)*x[1]+(0.213085443323)*x[2]
        s[5,1,1]=(-0.156276503308)*x[0]+(-0.463824282648)*x[1]+(-0.801704395726)*x[2]
        ref[0,0,0]=(0.663545175027)*x_ref[0]+(-1.05528281336)*x_ref[1]+(-0.25024128376)*x_ref[2]
        ref[0,0,1]=(1.10817148343)*x_ref[0]+(0.224734467846)*x_ref[1]+(0.712872841532)*x_ref[2]
        ref[0,1,0]=(-0.345072989586)*x_ref[0]+(-1.244678278)*x_ref[1]+(0.103446646879)*x_ref[2]
        ref[0,1,1]=(-0.106279741301)*x_ref[0]+(-0.525721333511)*x_ref[1]+(0.401917882446)*x_ref[2]
        ref[1,0,0]=(0.838242796285)*x_ref[0]+(-0.706020082817)*x_ref[1]+(0.443650945014)*x_ref[2]
        ref[1,0,1]=(-1.16557336842)*x_ref[0]+(0.215324659268)*x_ref[1]+(-0.285210566147)*x_ref[2]
        ref[1,1,0]=(-0.141574656707)*x_ref[0]+(1.66054358217)*x_ref[1]+(-0.554616610711)*x_ref[2]
        ref[1,1,1]=(0.523682935472)*x_ref[0]+(0.687178896331)*x_ref[1]+(0.875639049678)*x_ref[2]
        ref[2,0,0]=(0.585144782395)*x_ref[0]+(0.172063481865)*x_ref[1]+(0.920891875835)*x_ref[2]
        ref[2,0,1]=(0.660815435485)*x_ref[0]+(-1.63324849021)*x_ref[1]+(0.0810858621718)*x_ref[2]
        ref[2,1,0]=(-0.238166824253)*x_ref[0]+(0.0442487879497)*x_ref[1]+(1.05748755823)*x_ref[2]
        ref[2,1,1]=(1.14314890535)*x_ref[0]+(0.63161746042)*x_ref[1]+(1.5024770616)*x_ref[2]
        ref[3,0,0]=(-1.33327391952)*x_ref[0]+(1.58312319401)*x_ref[1]+(-1.24202894696)*x_ref[2]
        ref[3,0,1]=(-0.0242608819668)*x_ref[0]+(-1.12901018411)*x_ref[1]+(-0.443840381999)*x_ref[2]
        ref[3,1,0]=(1.12400540809)*x_ref[0]+(-0.25104247006)*x_ref[1]+(0.55935664575)*x_ref[2]
        ref[3,1,1]=(0.976977053811)*x_ref[0]+(1.46134206601)*x_ref[1]+(-0.883829686094)*x_ref[2]
        ref[4,0,0]=(0.68585022695)*x_ref[0]+(0.320211070942)*x_ref[1]+(-0.715034754238)*x_ref[2]
        ref[4,0,1]=(0.745090118519)*x_ref[0]+(-0.56687124464)*x_ref[1]+(-1.41483630619)*x_ref[2]
        ref[4,1,0]=(-0.864645032271)*x_ref[0]+(-0.0852453841156)*x_ref[1]+(-1.02272230917)*x_ref[2]
        ref[4,1,1]=(0.243107783262)*x_ref[0]+(-1.76263127813)*x_ref[1]+(-0.0396122398093)*x_ref[2]
        ref[5,0,0]=(-1.33890039547)*x_ref[0]+(0.95916460762)*x_ref[1]+(-0.118457193655)*x_ref[2]
        ref[5,0,1]=(-1.03526146458)*x_ref[0]+(-0.279038868252)*x_ref[1]+(-0.702378064321)*x_ref[2]
        ref[5,1,0]=(-0.434705839506)*x_ref[0]+(0.388520593856)*x_ref[1]+(0.213085443323)*x_ref[2]
        ref[5,1,1]=(-0.156276503308)*x_ref[0]+(-0.463824282648)*x_ref[1]+(-0.801704395726)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto FunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.239181398192)*x[0]+(-0.39363876765)*x[1]
        s[0,0,0,1]=(0.391461093031)*x[0]+(-0.825968486042)*x[1]
        s[0,0,1,0]=(0.327281495381)*x[0]+(-0.201869630768)*x[1]
        s[0,0,1,1]=(-0.0266300089051)*x[0]+(0.0315885163331)*x[1]
        s[0,0,2,0]=(-1.29808531891)*x[0]+(0.908442635353)*x[1]
        s[0,0,2,1]=(-0.0719703714734)*x[0]+(-1.2949593192)*x[1]
        s[0,1,0,0]=(1.56337010697)*x[0]+(0.554963325139)*x[1]
        s[0,1,0,1]=(-0.590177179718)*x[0]+(-0.577417888525)*x[1]
        s[0,1,1,0]=(1.12019940888)*x[0]+(0.515714775221)*x[1]
        s[0,1,1,1]=(1.62187733115)*x[0]+(-0.0496003559467)*x[1]
        s[0,1,2,0]=(1.33693246276)*x[0]+(0.18028194694)*x[1]
        s[0,1,2,1]=(1.197520649)*x[0]+(-0.189794758648)*x[1]
        s[0,2,0,0]=(0.458741675295)*x[0]+(-1.23374520557)*x[1]
        s[0,2,0,1]=(0.122618584159)*x[0]+(1.68218857683)*x[1]
        s[0,2,1,0]=(-0.629416301315)*x[0]+(0.00821950501686)*x[1]
        s[0,2,1,1]=(0.372444097376)*x[0]+(-0.389807082811)*x[1]
        s[0,2,2,0]=(0.453580881811)*x[0]+(0.234675328844)*x[1]
        s[0,2,2,1]=(-0.488447987119)*x[0]+(-0.795290862711)*x[1]
        s[0,3,0,0]=(-0.0558838669873)*x[0]+(-0.123245413784)*x[1]
        s[0,3,0,1]=(1.37779240976)*x[0]+(0.0843495250295)*x[1]
        s[0,3,1,0]=(1.17819744943)*x[0]+(-0.859044697828)*x[1]
        s[0,3,1,1]=(1.25175416429)*x[0]+(-0.727964075982)*x[1]
        s[0,3,2,0]=(0.331480991751)*x[0]+(0.115364502392)*x[1]
        s[0,3,2,1]=(-0.791037073807)*x[0]+(-0.51631676808)*x[1]
        s[0,4,0,0]=(0.884389418081)*x[0]+(1.15006122432)*x[1]
        s[0,4,0,1]=(-0.914023114312)*x[0]+(-0.211608734049)*x[1]
        s[0,4,1,0]=(0.55983523554)*x[0]+(-0.792920109589)*x[1]
        s[0,4,1,1]=(0.70901081021)*x[0]+(0.205352000801)*x[1]
        s[0,4,2,0]=(0.140341122546)*x[0]+(0.633534699565)*x[1]
        s[0,4,2,1]=(-0.399843857722)*x[0]+(0.0360346234753)*x[1]
        s[1,0,0,0]=(0.649005423866)*x[0]+(-0.72958629181)*x[1]
        s[1,0,0,1]=(-1.23313007881)*x[0]+(0.240005909545)*x[1]
        s[1,0,1,0]=(0.817440794961)*x[0]+(0.219684770228)*x[1]
        s[1,0,1,1]=(0.300714219818)*x[0]+(0.339948691074)*x[1]
        s[1,0,2,0]=(0.692412749315)*x[0]+(0.847435101682)*x[1]
        s[1,0,2,1]=(0.650288735017)*x[0]+(1.25808519806)*x[1]
        s[1,1,0,0]=(-0.723761201921)*x[0]+(-0.337875653383)*x[1]
        s[1,1,0,1]=(1.10005713123)*x[0]+(0.465822601087)*x[1]
        s[1,1,1,0]=(0.140014505481)*x[0]+(0.289986937348)*x[1]
        s[1,1,1,1]=(0.0941270067899)*x[0]+(-1.45892231406)*x[1]
        s[1,1,2,0]=(1.04493752023)*x[0]+(-0.474063991258)*x[1]
        s[1,1,2,1]=(0.703284346556)*x[0]+(0.482116021536)*x[1]
        s[1,2,0,0]=(-1.66498217207)*x[0]+(-0.144077385929)*x[1]
        s[1,2,0,1]=(1.50511312044)*x[0]+(1.02587270487)*x[1]
        s[1,2,1,0]=(-1.15095468052)*x[0]+(-0.472480042698)*x[1]
        s[1,2,1,1]=(-0.453604215856)*x[0]+(-1.58856703352)*x[1]
        s[1,2,2,0]=(0.659481843854)*x[0]+(0.649931585882)*x[1]
        s[1,2,2,1]=(0.642785065122)*x[0]+(0.0614989297125)*x[1]
        s[1,3,0,0]=(1.65661008057)*x[0]+(-1.08472986336)*x[1]
        s[1,3,0,1]=(-0.64853562643)*x[0]+(-0.912673933935)*x[1]
        s[1,3,1,0]=(-0.108999021371)*x[0]+(-1.44843200164)*x[1]
        s[1,3,1,1]=(0.568395064875)*x[0]+(-0.471819846301)*x[1]
        s[1,3,2,0]=(-0.801680619008)*x[0]+(0.197249605781)*x[1]
        s[1,3,2,1]=(-0.512763932006)*x[0]+(-1.03221569621)*x[1]
        s[1,4,0,0]=(0.54142249246)*x[0]+(-1.28623467967)*x[1]
        s[1,4,0,1]=(1.19094828259)*x[0]+(0.58814933754)*x[1]
        s[1,4,1,0]=(-0.203699090085)*x[0]+(0.839194924785)*x[1]
        s[1,4,1,1]=(-1.73647778175)*x[0]+(1.26399102178)*x[1]
        s[1,4,2,0]=(0.688685884338)*x[0]+(-0.0441772317601)*x[1]
        s[1,4,2,1]=(-0.811158497264)*x[0]+(-0.663923769116)*x[1]
        s[2,0,0,0]=(0.237794148299)*x[0]+(0.125329413285)*x[1]
        s[2,0,0,1]=(-1.40518934715)*x[0]+(1.22320828318)*x[1]
        s[2,0,1,0]=(0.223901464538)*x[0]+(-0.00989463351767)*x[1]
        s[2,0,1,1]=(-0.912246132725)*x[0]+(0.686091520129)*x[1]
        s[2,0,2,0]=(0.181752601691)*x[0]+(-1.12779553611)*x[1]
        s[2,0,2,1]=(-0.061707870206)*x[0]+(-1.3429678748)*x[1]
        s[2,1,0,0]=(1.56396877322)*x[0]+(-1.05119596867)*x[1]
        s[2,1,0,1]=(0.929556514862)*x[0]+(0.406040622531)*x[1]
        s[2,1,1,0]=(-0.509355797822)*x[0]+(0.0605066527971)*x[1]
        s[2,1,1,1]=(-0.528244794646)*x[0]+(0.0388070048758)*x[1]
        s[2,1,2,0]=(-0.200007488719)*x[0]+(-0.711333050321)*x[1]
        s[2,1,2,1]=(0.175160525472)*x[0]+(-1.24343736284)*x[1]
        s[2,2,0,0]=(0.894960105109)*x[0]+(-0.747881639384)*x[1]
        s[2,2,0,1]=(-0.0894204005642)*x[0]+(-0.694792125757)*x[1]
        s[2,2,1,0]=(0.172037435959)*x[0]+(1.29233768559)*x[1]
        s[2,2,1,1]=(-0.312372601854)*x[0]+(-1.35937909779)*x[1]
        s[2,2,2,0]=(0.323370754787)*x[0]+(-0.0601280403525)*x[1]
        s[2,2,2,1]=(0.648154329469)*x[0]+(-0.976168040714)*x[1]
        s[2,3,0,0]=(0.178127534181)*x[0]+(-0.451322025537)*x[1]
        s[2,3,0,1]=(-1.88695440231)*x[0]+(0.164327427196)*x[1]
        s[2,3,1,0]=(0.543479878129)*x[0]+(-0.553094838979)*x[1]
        s[2,3,1,1]=(-0.745599057313)*x[0]+(0.710208052579)*x[1]
        s[2,3,2,0]=(-0.262128106404)*x[0]+(0.839008375922)*x[1]
        s[2,3,2,1]=(-1.20847299902)*x[0]+(-0.321234640326)*x[1]
        s[2,4,0,0]=(-1.24374530208)*x[0]+(-0.961997778937)*x[1]
        s[2,4,0,1]=(-0.0306817051594)*x[0]+(-0.127120502282)*x[1]
        s[2,4,1,0]=(1.25878772522)*x[0]+(1.04432409904)*x[1]
        s[2,4,1,1]=(0.271036209294)*x[0]+(-1.19597311792)*x[1]
        s[2,4,2,0]=(-0.073359480179)*x[0]+(-0.31523273696)*x[1]
        s[2,4,2,1]=(0.738924603079)*x[0]+(0.251467051457)*x[1]
        s[3,0,0,0]=(-0.257569738752)*x[0]+(-1.51373140433)*x[1]
        s[3,0,0,1]=(-0.0113459170081)*x[0]+(-0.356375170129)*x[1]
        s[3,0,1,0]=(-0.290333097834)*x[0]+(-1.56956309356)*x[1]
        s[3,0,1,1]=(-1.0806247672)*x[0]+(-0.376429365315)*x[1]
        s[3,0,2,0]=(-1.32161176104)*x[0]+(0.784442685433)*x[1]
        s[3,0,2,1]=(-1.00382464524)*x[0]+(-0.571979697031)*x[1]
        s[3,1,0,0]=(0.140827544385)*x[0]+(1.3690229529)*x[1]
        s[3,1,0,1]=(0.315011101794)*x[0]+(-1.0263379129)*x[1]
        s[3,1,1,0]=(-0.478940977535)*x[0]+(0.718456942414)*x[1]
        s[3,1,1,1]=(-0.655873236005)*x[0]+(-0.148701297422)*x[1]
        s[3,1,2,0]=(0.56126503141)*x[0]+(1.68526215251)*x[1]
        s[3,1,2,1]=(-0.337284606646)*x[0]+(-0.197295432114)*x[1]
        s[3,2,0,0]=(-0.5979320732)*x[0]+(-0.0600053477444)*x[1]
        s[3,2,0,1]=(-1.20531807175)*x[0]+(-0.148853538473)*x[1]
        s[3,2,1,0]=(0.313468114426)*x[0]+(-0.436912848514)*x[1]
        s[3,2,1,1]=(-0.544279507407)*x[0]+(0.189063470736)*x[1]
        s[3,2,2,0]=(-0.551741475397)*x[0]+(0.196370248113)*x[1]
        s[3,2,2,1]=(-0.0216545656433)*x[0]+(0.23669839429)*x[1]
        s[3,3,0,0]=(-0.406652896064)*x[0]+(0.00397676360081)*x[1]
        s[3,3,0,1]=(-0.305847961815)*x[0]+(-1.2563276235)*x[1]
        s[3,3,1,0]=(-0.450621831697)*x[0]+(-0.213476321469)*x[1]
        s[3,3,1,1]=(1.20548652879)*x[0]+(-0.286503833918)*x[1]
        s[3,3,2,0]=(0.516344981347)*x[0]+(-0.343787067175)*x[1]
        s[3,3,2,1]=(-0.01938029713)*x[0]+(-0.779285558404)*x[1]
        s[3,4,0,0]=(-0.406233598349)*x[0]+(-0.498045431333)*x[1]
        s[3,4,0,1]=(-0.790577593976)*x[0]+(1.07251305331)*x[1]
        s[3,4,1,0]=(0.510569891955)*x[0]+(0.526099613153)*x[1]
        s[3,4,1,1]=(0.150065560826)*x[0]+(0.966145292988)*x[1]
        s[3,4,2,0]=(0.718402792955)*x[0]+(-0.430416841145)*x[1]
        s[3,4,2,1]=(-0.430240857917)*x[0]+(-0.0679614650591)*x[1]
        ref[0,0,0,0]=(0.239181398192)*x_ref[0]+(-0.39363876765)*x_ref[1]
        ref[0,0,0,1]=(0.391461093031)*x_ref[0]+(-0.825968486042)*x_ref[1]
        ref[0,0,1,0]=(0.327281495381)*x_ref[0]+(-0.201869630768)*x_ref[1]
        ref[0,0,1,1]=(-0.0266300089051)*x_ref[0]+(0.0315885163331)*x_ref[1]
        ref[0,0,2,0]=(-1.29808531891)*x_ref[0]+(0.908442635353)*x_ref[1]
        ref[0,0,2,1]=(-0.0719703714734)*x_ref[0]+(-1.2949593192)*x_ref[1]
        ref[0,1,0,0]=(1.56337010697)*x_ref[0]+(0.554963325139)*x_ref[1]
        ref[0,1,0,1]=(-0.590177179718)*x_ref[0]+(-0.577417888525)*x_ref[1]
        ref[0,1,1,0]=(1.12019940888)*x_ref[0]+(0.515714775221)*x_ref[1]
        ref[0,1,1,1]=(1.62187733115)*x_ref[0]+(-0.0496003559467)*x_ref[1]
        ref[0,1,2,0]=(1.33693246276)*x_ref[0]+(0.18028194694)*x_ref[1]
        ref[0,1,2,1]=(1.197520649)*x_ref[0]+(-0.189794758648)*x_ref[1]
        ref[0,2,0,0]=(0.458741675295)*x_ref[0]+(-1.23374520557)*x_ref[1]
        ref[0,2,0,1]=(0.122618584159)*x_ref[0]+(1.68218857683)*x_ref[1]
        ref[0,2,1,0]=(-0.629416301315)*x_ref[0]+(0.00821950501686)*x_ref[1]
        ref[0,2,1,1]=(0.372444097376)*x_ref[0]+(-0.389807082811)*x_ref[1]
        ref[0,2,2,0]=(0.453580881811)*x_ref[0]+(0.234675328844)*x_ref[1]
        ref[0,2,2,1]=(-0.488447987119)*x_ref[0]+(-0.795290862711)*x_ref[1]
        ref[0,3,0,0]=(-0.0558838669873)*x_ref[0]+(-0.123245413784)*x_ref[1]
        ref[0,3,0,1]=(1.37779240976)*x_ref[0]+(0.0843495250295)*x_ref[1]
        ref[0,3,1,0]=(1.17819744943)*x_ref[0]+(-0.859044697828)*x_ref[1]
        ref[0,3,1,1]=(1.25175416429)*x_ref[0]+(-0.727964075982)*x_ref[1]
        ref[0,3,2,0]=(0.331480991751)*x_ref[0]+(0.115364502392)*x_ref[1]
        ref[0,3,2,1]=(-0.791037073807)*x_ref[0]+(-0.51631676808)*x_ref[1]
        ref[0,4,0,0]=(0.884389418081)*x_ref[0]+(1.15006122432)*x_ref[1]
        ref[0,4,0,1]=(-0.914023114312)*x_ref[0]+(-0.211608734049)*x_ref[1]
        ref[0,4,1,0]=(0.55983523554)*x_ref[0]+(-0.792920109589)*x_ref[1]
        ref[0,4,1,1]=(0.70901081021)*x_ref[0]+(0.205352000801)*x_ref[1]
        ref[0,4,2,0]=(0.140341122546)*x_ref[0]+(0.633534699565)*x_ref[1]
        ref[0,4,2,1]=(-0.399843857722)*x_ref[0]+(0.0360346234753)*x_ref[1]
        ref[1,0,0,0]=(0.649005423866)*x_ref[0]+(-0.72958629181)*x_ref[1]
        ref[1,0,0,1]=(-1.23313007881)*x_ref[0]+(0.240005909545)*x_ref[1]
        ref[1,0,1,0]=(0.817440794961)*x_ref[0]+(0.219684770228)*x_ref[1]
        ref[1,0,1,1]=(0.300714219818)*x_ref[0]+(0.339948691074)*x_ref[1]
        ref[1,0,2,0]=(0.692412749315)*x_ref[0]+(0.847435101682)*x_ref[1]
        ref[1,0,2,1]=(0.650288735017)*x_ref[0]+(1.25808519806)*x_ref[1]
        ref[1,1,0,0]=(-0.723761201921)*x_ref[0]+(-0.337875653383)*x_ref[1]
        ref[1,1,0,1]=(1.10005713123)*x_ref[0]+(0.465822601087)*x_ref[1]
        ref[1,1,1,0]=(0.140014505481)*x_ref[0]+(0.289986937348)*x_ref[1]
        ref[1,1,1,1]=(0.0941270067899)*x_ref[0]+(-1.45892231406)*x_ref[1]
        ref[1,1,2,0]=(1.04493752023)*x_ref[0]+(-0.474063991258)*x_ref[1]
        ref[1,1,2,1]=(0.703284346556)*x_ref[0]+(0.482116021536)*x_ref[1]
        ref[1,2,0,0]=(-1.66498217207)*x_ref[0]+(-0.144077385929)*x_ref[1]
        ref[1,2,0,1]=(1.50511312044)*x_ref[0]+(1.02587270487)*x_ref[1]
        ref[1,2,1,0]=(-1.15095468052)*x_ref[0]+(-0.472480042698)*x_ref[1]
        ref[1,2,1,1]=(-0.453604215856)*x_ref[0]+(-1.58856703352)*x_ref[1]
        ref[1,2,2,0]=(0.659481843854)*x_ref[0]+(0.649931585882)*x_ref[1]
        ref[1,2,2,1]=(0.642785065122)*x_ref[0]+(0.0614989297125)*x_ref[1]
        ref[1,3,0,0]=(1.65661008057)*x_ref[0]+(-1.08472986336)*x_ref[1]
        ref[1,3,0,1]=(-0.64853562643)*x_ref[0]+(-0.912673933935)*x_ref[1]
        ref[1,3,1,0]=(-0.108999021371)*x_ref[0]+(-1.44843200164)*x_ref[1]
        ref[1,3,1,1]=(0.568395064875)*x_ref[0]+(-0.471819846301)*x_ref[1]
        ref[1,3,2,0]=(-0.801680619008)*x_ref[0]+(0.197249605781)*x_ref[1]
        ref[1,3,2,1]=(-0.512763932006)*x_ref[0]+(-1.03221569621)*x_ref[1]
        ref[1,4,0,0]=(0.54142249246)*x_ref[0]+(-1.28623467967)*x_ref[1]
        ref[1,4,0,1]=(1.19094828259)*x_ref[0]+(0.58814933754)*x_ref[1]
        ref[1,4,1,0]=(-0.203699090085)*x_ref[0]+(0.839194924785)*x_ref[1]
        ref[1,4,1,1]=(-1.73647778175)*x_ref[0]+(1.26399102178)*x_ref[1]
        ref[1,4,2,0]=(0.688685884338)*x_ref[0]+(-0.0441772317601)*x_ref[1]
        ref[1,4,2,1]=(-0.811158497264)*x_ref[0]+(-0.663923769116)*x_ref[1]
        ref[2,0,0,0]=(0.237794148299)*x_ref[0]+(0.125329413285)*x_ref[1]
        ref[2,0,0,1]=(-1.40518934715)*x_ref[0]+(1.22320828318)*x_ref[1]
        ref[2,0,1,0]=(0.223901464538)*x_ref[0]+(-0.00989463351767)*x_ref[1]
        ref[2,0,1,1]=(-0.912246132725)*x_ref[0]+(0.686091520129)*x_ref[1]
        ref[2,0,2,0]=(0.181752601691)*x_ref[0]+(-1.12779553611)*x_ref[1]
        ref[2,0,2,1]=(-0.061707870206)*x_ref[0]+(-1.3429678748)*x_ref[1]
        ref[2,1,0,0]=(1.56396877322)*x_ref[0]+(-1.05119596867)*x_ref[1]
        ref[2,1,0,1]=(0.929556514862)*x_ref[0]+(0.406040622531)*x_ref[1]
        ref[2,1,1,0]=(-0.509355797822)*x_ref[0]+(0.0605066527971)*x_ref[1]
        ref[2,1,1,1]=(-0.528244794646)*x_ref[0]+(0.0388070048758)*x_ref[1]
        ref[2,1,2,0]=(-0.200007488719)*x_ref[0]+(-0.711333050321)*x_ref[1]
        ref[2,1,2,1]=(0.175160525472)*x_ref[0]+(-1.24343736284)*x_ref[1]
        ref[2,2,0,0]=(0.894960105109)*x_ref[0]+(-0.747881639384)*x_ref[1]
        ref[2,2,0,1]=(-0.0894204005642)*x_ref[0]+(-0.694792125757)*x_ref[1]
        ref[2,2,1,0]=(0.172037435959)*x_ref[0]+(1.29233768559)*x_ref[1]
        ref[2,2,1,1]=(-0.312372601854)*x_ref[0]+(-1.35937909779)*x_ref[1]
        ref[2,2,2,0]=(0.323370754787)*x_ref[0]+(-0.0601280403525)*x_ref[1]
        ref[2,2,2,1]=(0.648154329469)*x_ref[0]+(-0.976168040714)*x_ref[1]
        ref[2,3,0,0]=(0.178127534181)*x_ref[0]+(-0.451322025537)*x_ref[1]
        ref[2,3,0,1]=(-1.88695440231)*x_ref[0]+(0.164327427196)*x_ref[1]
        ref[2,3,1,0]=(0.543479878129)*x_ref[0]+(-0.553094838979)*x_ref[1]
        ref[2,3,1,1]=(-0.745599057313)*x_ref[0]+(0.710208052579)*x_ref[1]
        ref[2,3,2,0]=(-0.262128106404)*x_ref[0]+(0.839008375922)*x_ref[1]
        ref[2,3,2,1]=(-1.20847299902)*x_ref[0]+(-0.321234640326)*x_ref[1]
        ref[2,4,0,0]=(-1.24374530208)*x_ref[0]+(-0.961997778937)*x_ref[1]
        ref[2,4,0,1]=(-0.0306817051594)*x_ref[0]+(-0.127120502282)*x_ref[1]
        ref[2,4,1,0]=(1.25878772522)*x_ref[0]+(1.04432409904)*x_ref[1]
        ref[2,4,1,1]=(0.271036209294)*x_ref[0]+(-1.19597311792)*x_ref[1]
        ref[2,4,2,0]=(-0.073359480179)*x_ref[0]+(-0.31523273696)*x_ref[1]
        ref[2,4,2,1]=(0.738924603079)*x_ref[0]+(0.251467051457)*x_ref[1]
        ref[3,0,0,0]=(-0.257569738752)*x_ref[0]+(-1.51373140433)*x_ref[1]
        ref[3,0,0,1]=(-0.0113459170081)*x_ref[0]+(-0.356375170129)*x_ref[1]
        ref[3,0,1,0]=(-0.290333097834)*x_ref[0]+(-1.56956309356)*x_ref[1]
        ref[3,0,1,1]=(-1.0806247672)*x_ref[0]+(-0.376429365315)*x_ref[1]
        ref[3,0,2,0]=(-1.32161176104)*x_ref[0]+(0.784442685433)*x_ref[1]
        ref[3,0,2,1]=(-1.00382464524)*x_ref[0]+(-0.571979697031)*x_ref[1]
        ref[3,1,0,0]=(0.140827544385)*x_ref[0]+(1.3690229529)*x_ref[1]
        ref[3,1,0,1]=(0.315011101794)*x_ref[0]+(-1.0263379129)*x_ref[1]
        ref[3,1,1,0]=(-0.478940977535)*x_ref[0]+(0.718456942414)*x_ref[1]
        ref[3,1,1,1]=(-0.655873236005)*x_ref[0]+(-0.148701297422)*x_ref[1]
        ref[3,1,2,0]=(0.56126503141)*x_ref[0]+(1.68526215251)*x_ref[1]
        ref[3,1,2,1]=(-0.337284606646)*x_ref[0]+(-0.197295432114)*x_ref[1]
        ref[3,2,0,0]=(-0.5979320732)*x_ref[0]+(-0.0600053477444)*x_ref[1]
        ref[3,2,0,1]=(-1.20531807175)*x_ref[0]+(-0.148853538473)*x_ref[1]
        ref[3,2,1,0]=(0.313468114426)*x_ref[0]+(-0.436912848514)*x_ref[1]
        ref[3,2,1,1]=(-0.544279507407)*x_ref[0]+(0.189063470736)*x_ref[1]
        ref[3,2,2,0]=(-0.551741475397)*x_ref[0]+(0.196370248113)*x_ref[1]
        ref[3,2,2,1]=(-0.0216545656433)*x_ref[0]+(0.23669839429)*x_ref[1]
        ref[3,3,0,0]=(-0.406652896064)*x_ref[0]+(0.00397676360081)*x_ref[1]
        ref[3,3,0,1]=(-0.305847961815)*x_ref[0]+(-1.2563276235)*x_ref[1]
        ref[3,3,1,0]=(-0.450621831697)*x_ref[0]+(-0.213476321469)*x_ref[1]
        ref[3,3,1,1]=(1.20548652879)*x_ref[0]+(-0.286503833918)*x_ref[1]
        ref[3,3,2,0]=(0.516344981347)*x_ref[0]+(-0.343787067175)*x_ref[1]
        ref[3,3,2,1]=(-0.01938029713)*x_ref[0]+(-0.779285558404)*x_ref[1]
        ref[3,4,0,0]=(-0.406233598349)*x_ref[0]+(-0.498045431333)*x_ref[1]
        ref[3,4,0,1]=(-0.790577593976)*x_ref[0]+(1.07251305331)*x_ref[1]
        ref[3,4,1,0]=(0.510569891955)*x_ref[0]+(0.526099613153)*x_ref[1]
        ref[3,4,1,1]=(0.150065560826)*x_ref[0]+(0.966145292988)*x_ref[1]
        ref[3,4,2,0]=(0.718402792955)*x_ref[0]+(-0.430416841145)*x_ref[1]
        ref[3,4,2,1]=(-0.430240857917)*x_ref[0]+(-0.0679614650591)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.512691010939)*x[0]+(-1.31295640543)*x[1]+(1.45426060088)*x[2]
        s[0,0,0,1]=(-0.57748372698)*x[0]+(0.215297337003)*x[1]+(0.0853439946124)*x[2]
        s[0,0,1,0]=(-0.785306554604)*x[0]+(-0.0274911695421)*x[1]+(0.693440767035)*x[2]
        s[0,0,1,1]=(-0.56932407873)*x[0]+(-0.981593188827)*x[1]+(0.87532777027)*x[2]
        s[0,0,2,0]=(0.145925104392)*x[0]+(1.45139125753)*x[1]+(1.25041746183)*x[2]
        s[0,0,2,1]=(-1.70110072582)*x[0]+(-0.0335136921878)*x[1]+(0.276204889346)*x[2]
        s[0,1,0,0]=(-0.415026748605)*x[0]+(-0.322339100074)*x[1]+(1.1050553524)*x[2]
        s[0,1,0,1]=(0.267336702938)*x[0]+(-1.44498697586)*x[1]+(0.600455793139)*x[2]
        s[0,1,1,0]=(-0.46618843993)*x[0]+(0.318230555457)*x[1]+(0.281788961443)*x[2]
        s[0,1,1,1]=(-1.12963168407)*x[0]+(-0.661100272253)*x[1]+(0.293642402995)*x[2]
        s[0,1,2,0]=(-0.0041407279814)*x[0]+(0.330916770241)*x[1]+(1.30265962609)*x[2]
        s[0,1,2,1]=(1.37513192104)*x[0]+(-1.39837371617)*x[1]+(-1.23490743489)*x[2]
        s[0,2,0,0]=(0.240920374522)*x[0]+(1.30092830627)*x[1]+(0.189720240056)*x[2]
        s[0,2,0,1]=(0.741337711467)*x[0]+(0.00893414596869)*x[1]+(1.22438764071)*x[2]
        s[0,2,1,0]=(1.0524903095)*x[0]+(-1.00088524111)*x[1]+(0.985885896119)*x[2]
        s[0,2,1,1]=(-0.252168964788)*x[0]+(1.2685968377)*x[1]+(-0.025373437998)*x[2]
        s[0,2,2,0]=(0.728523707179)*x[0]+(0.00656996105623)*x[1]+(1.85098622635)*x[2]
        s[0,2,2,1]=(0.193123464463)*x[0]+(-0.426452202023)*x[1]+(-0.144471734372)*x[2]
        s[0,3,0,0]=(0.446506355449)*x[0]+(0.379921574537)*x[1]+(-0.303150079281)*x[2]
        s[0,3,0,1]=(0.455373734156)*x[0]+(0.134600165673)*x[1]+(-0.604957024603)*x[2]
        s[0,3,1,0]=(-1.63808806962)*x[0]+(1.02897840042)*x[1]+(1.18123170106)*x[2]
        s[0,3,1,1]=(0.587448315418)*x[0]+(0.0991725296547)*x[1]+(1.33682286695)*x[2]
        s[0,3,2,0]=(-1.12085613449)*x[0]+(0.454771039944)*x[1]+(-0.131089499156)*x[2]
        s[0,3,2,1]=(-0.351021639018)*x[0]+(0.47905316732)*x[1]+(-1.07449842534)*x[2]
        s[0,4,0,0]=(0.232526828642)*x[0]+(1.33907066655)*x[1]+(-0.699701401092)*x[2]
        s[0,4,0,1]=(0.244905848535)*x[0]+(-0.799313917773)*x[1]+(-0.16513296799)*x[2]
        s[0,4,1,0]=(-1.10152273815)*x[0]+(0.898176161732)*x[1]+(-0.101734265033)*x[2]
        s[0,4,1,1]=(-0.0804598444811)*x[0]+(-0.77819195606)*x[1]+(0.277776761782)*x[2]
        s[0,4,2,0]=(1.05979692744)*x[0]+(-0.195285573862)*x[1]+(-0.0920717991265)*x[2]
        s[0,4,2,1]=(0.44924595646)*x[0]+(0.409652246508)*x[1]+(-0.0705754685165)*x[2]
        s[1,0,0,0]=(1.14991675937)*x[0]+(-0.435947647215)*x[1]+(-0.45614944131)*x[2]
        s[1,0,0,1]=(0.306890350036)*x[0]+(-1.4400848018)*x[1]+(-0.744474442385)*x[2]
        s[1,0,1,0]=(1.07930565176)*x[0]+(-0.273335650157)*x[1]+(-0.230960378838)*x[2]
        s[1,0,1,1]=(-0.97797563017)*x[0]+(-1.3159717465)*x[1]+(-0.357902208089)*x[2]
        s[1,0,2,0]=(-1.02063328249)*x[0]+(-1.18907185687)*x[1]+(-0.509296457902)*x[2]
        s[1,0,2,1]=(-0.508446666232)*x[0]+(-0.69279963667)*x[1]+(-0.497990637708)*x[2]
        s[1,1,0,0]=(-0.476141863054)*x[0]+(-0.180458121777)*x[1]+(-0.318855084826)*x[2]
        s[1,1,0,1]=(0.480566371651)*x[0]+(-1.74196768367)*x[1]+(1.23057574863)*x[2]
        s[1,1,1,0]=(-0.0213138624722)*x[0]+(-0.0846532154764)*x[1]+(-0.61169792023)*x[2]
        s[1,1,1,1]=(-0.389381507111)*x[0]+(0.436088515729)*x[1]+(-0.603019053837)*x[2]
        s[1,1,2,0]=(-1.24478786129)*x[0]+(0.908984432701)*x[1]+(-1.49483946407)*x[2]
        s[1,1,2,1]=(0.866266264162)*x[0]+(-0.190579076131)*x[1]+(-1.63524384216)*x[2]
        s[1,2,0,0]=(0.0333832222755)*x[0]+(-0.551591529565)*x[1]+(1.22170194609)*x[2]
        s[1,2,0,1]=(-0.551682390802)*x[0]+(0.223940652339)*x[1]+(0.141063634211)*x[2]
        s[1,2,1,0]=(1.17072817252)*x[0]+(0.28318356571)*x[1]+(-0.788096581969)*x[2]
        s[1,2,1,1]=(-1.01576849129)*x[0]+(-0.502669485722)*x[1]+(-0.966677100703)*x[2]
        s[1,2,2,0]=(-1.59522074083)*x[0]+(-1.16016984107)*x[1]+(-0.117030183187)*x[2]
        s[1,2,2,1]=(-0.979165259497)*x[0]+(-0.707460176459)*x[1]+(-0.275081507376)*x[2]
        s[1,3,0,0]=(0.692493088771)*x[0]+(-1.04755874336)*x[1]+(-0.373632326826)*x[2]
        s[1,3,0,1]=(0.717001163418)*x[0]+(-0.782379062328)*x[1]+(0.479776988529)*x[2]
        s[1,3,1,0]=(-0.733062748373)*x[0]+(-0.217972411497)*x[1]+(1.43785098293)*x[2]
        s[1,3,1,1]=(0.186430904908)*x[0]+(0.375218175232)*x[1]+(-0.461668650588)*x[2]
        s[1,3,2,0]=(-1.12951451301)*x[0]+(-1.07034955937)*x[1]+(-0.59888614247)*x[2]
        s[1,3,2,1]=(1.34023062833)*x[0]+(-0.520544300692)*x[1]+(-0.363001059538)*x[2]
        s[1,4,0,0]=(0.42295611834)*x[0]+(-0.718852108989)*x[1]+(-1.61995684533)*x[2]
        s[1,4,0,1]=(-0.422467947319)*x[0]+(-0.209334480945)*x[1]+(1.59579413592)*x[2]
        s[1,4,1,0]=(0.387543472208)*x[0]+(-0.0579909254613)*x[1]+(-1.03104850867)*x[2]
        s[1,4,1,1]=(-1.15893685519)*x[0]+(-0.971412946627)*x[1]+(-0.583440884462)*x[2]
        s[1,4,2,0]=(0.554042723162)*x[0]+(0.082633597735)*x[1]+(0.990177730288)*x[2]
        s[1,4,2,1]=(0.150114027051)*x[0]+(0.479664757576)*x[1]+(-1.5769513289)*x[2]
        s[2,0,0,0]=(1.33963588901)*x[0]+(0.146971184264)*x[1]+(-1.29921529821)*x[2]
        s[2,0,0,1]=(0.0115342792238)*x[0]+(0.0451643711892)*x[1]+(-0.656407384758)*x[2]
        s[2,0,1,0]=(0.175368903926)*x[0]+(-0.40573949846)*x[1]+(0.295387861218)*x[2]
        s[2,0,1,1]=(0.531254864472)*x[0]+(-0.536193761259)*x[1]+(-1.21752823876)*x[2]
        s[2,0,2,0]=(-0.362296220089)*x[0]+(0.378651497918)*x[1]+(-0.535206713131)*x[2]
        s[2,0,2,1]=(-0.85728549969)*x[0]+(0.233179293609)*x[1]+(0.345412963315)*x[2]
        s[2,1,0,0]=(1.01059063177)*x[0]+(0.808116187589)*x[1]+(0.468588054081)*x[2]
        s[2,1,0,1]=(-1.17212735529)*x[0]+(-0.871560017197)*x[1]+(1.10553749428)*x[2]
        s[2,1,1,0]=(1.00177793873)*x[0]+(-0.573447532545)*x[1]+(-0.80780539944)*x[2]
        s[2,1,1,1]=(-0.101912168689)*x[0]+(0.0054724970873)*x[1]+(0.873799338046)*x[2]
        s[2,1,2,0]=(0.162601365201)*x[0]+(-0.508620489728)*x[1]+(0.114861348434)*x[2]
        s[2,1,2,1]=(-1.48812556245)*x[0]+(0.29841054692)*x[1]+(0.773432180321)*x[2]
        s[2,2,0,0]=(0.0177469158963)*x[0]+(0.140208281265)*x[1]+(1.12393579682)*x[2]
        s[2,2,0,1]=(0.127835907966)*x[0]+(0.150072336346)*x[1]+(1.59437274191)*x[2]
        s[2,2,1,0]=(-0.549906307264)*x[0]+(-0.572101026775)*x[1]+(-0.747028158715)*x[2]
        s[2,2,1,1]=(0.134272548587)*x[0]+(0.866777056589)*x[1]+(-0.0658173600046)*x[2]
        s[2,2,2,0]=(0.160006947347)*x[0]+(-0.940281662941)*x[1]+(0.671280858545)*x[2]
        s[2,2,2,1]=(-1.55550646869)*x[0]+(0.0485277897059)*x[1]+(1.02027216264)*x[2]
        s[2,3,0,0]=(0.0931028968111)*x[0]+(0.194744643733)*x[1]+(1.7368246452)*x[2]
        s[2,3,0,1]=(1.66077510679)*x[0]+(-0.342785893413)*x[1]+(-0.278302501974)*x[2]
        s[2,3,1,0]=(-0.0719600278272)*x[0]+(0.898388137955)*x[1]+(-0.110811429033)*x[2]
        s[2,3,1,1]=(-0.781997247811)*x[0]+(-1.32497050043)*x[1]+(0.544850665961)*x[2]
        s[2,3,2,0]=(0.0290847482287)*x[0]+(0.895295560038)*x[1]+(-1.68555124085)*x[2]
        s[2,3,2,1]=(0.200540179332)*x[0]+(-0.843842814948)*x[1]+(-0.943779465271)*x[2]
        s[2,4,0,0]=(1.08999954494)*x[0]+(-1.32324120166)*x[1]+(-0.0152368215856)*x[2]
        s[2,4,0,1]=(1.02219584162)*x[0]+(0.208263165731)*x[1]+(0.0703408478849)*x[2]
        s[2,4,1,0]=(0.433866479442)*x[0]+(0.327590037731)*x[1]+(0.752498033029)*x[2]
        s[2,4,1,1]=(1.16322078616)*x[0]+(-1.38234503561)*x[1]+(-0.129845447078)*x[2]
        s[2,4,2,0]=(-0.430572860596)*x[0]+(-1.48079933792)*x[1]+(0.310197458762)*x[2]
        s[2,4,2,1]=(-0.0901138166505)*x[0]+(0.130679443905)*x[1]+(-0.482309483824)*x[2]
        s[3,0,0,0]=(0.239787389644)*x[0]+(0.690754646072)*x[1]+(0.745418481522)*x[2]
        s[3,0,0,1]=(0.346448639801)*x[0]+(-0.287823450604)*x[1]+(0.287003737791)*x[2]
        s[3,0,1,0]=(-1.50502949668)*x[0]+(-1.67521500964)*x[1]+(0.31630953667)*x[2]
        s[3,0,1,1]=(-0.594355763268)*x[0]+(-0.262023313233)*x[1]+(-1.41203969641)*x[2]
        s[3,0,2,0]=(-0.0275545216089)*x[0]+(0.367433407785)*x[1]+(-1.420568196)*x[2]
        s[3,0,2,1]=(-1.18797738587)*x[0]+(-0.496775264183)*x[1]+(0.513942381739)*x[2]
        s[3,1,0,0]=(0.475014043678)*x[0]+(-1.2205702601)*x[1]+(1.06353608529)*x[2]
        s[3,1,0,1]=(-0.0685583494755)*x[0]+(-1.49448219897)*x[1]+(-1.23320402959)*x[2]
        s[3,1,1,0]=(-0.216184790554)*x[0]+(0.554046932828)*x[1]+(0.697487187904)*x[2]
        s[3,1,1,1]=(-0.495370181694)*x[0]+(-1.49207400185)*x[1]+(-0.233763280927)*x[2]
        s[3,1,2,0]=(-0.0462888472218)*x[0]+(0.371127867531)*x[1]+(0.534977833221)*x[2]
        s[3,1,2,1]=(-1.27057460029)*x[0]+(-0.572725702684)*x[1]+(-0.871217452212)*x[2]
        s[3,2,0,0]=(-0.208549808264)*x[0]+(0.847294134183)*x[1]+(0.643419233684)*x[2]
        s[3,2,0,1]=(-0.850445146291)*x[0]+(1.19605447923)*x[1]+(-1.52958469348)*x[2]
        s[3,2,1,0]=(0.181672364304)*x[0]+(-1.60804541136)*x[1]+(0.068065187694)*x[2]
        s[3,2,1,1]=(-0.195284925405)*x[0]+(0.330399376496)*x[1]+(-0.453658972198)*x[2]
        s[3,2,2,0]=(1.44390910788)*x[0]+(0.718618445973)*x[1]+(-0.800501828022)*x[2]
        s[3,2,2,1]=(0.491997052311)*x[0]+(-0.202966322046)*x[1]+(1.39185177314)*x[2]
        s[3,3,0,0]=(-0.670818169452)*x[0]+(0.842747476683)*x[1]+(-0.705022077785)*x[2]
        s[3,3,0,1]=(0.729442429399)*x[0]+(-1.47998423588)*x[1]+(1.72626243143)*x[2]
        s[3,3,1,0]=(-1.08179915812)*x[0]+(-0.117953758788)*x[1]+(1.60080937509)*x[2]
        s[3,3,1,1]=(0.110049458343)*x[0]+(0.571461685559)*x[1]+(-0.483581817598)*x[2]
        s[3,3,2,0]=(-0.386163706678)*x[0]+(-0.146336916781)*x[1]+(0.0360311139076)*x[2]
        s[3,3,2,1]=(-0.779748009644)*x[0]+(-0.780523381896)*x[1]+(-1.20157768536)*x[2]
        s[3,4,0,0]=(1.74580569218)*x[0]+(-0.318257555715)*x[1]+(-0.33745202709)*x[2]
        s[3,4,0,1]=(1.1463597084)*x[0]+(0.516488947224)*x[1]+(-0.205035082746)*x[2]
        s[3,4,1,0]=(0.647937211136)*x[0]+(0.574941676843)*x[1]+(-1.00930234083)*x[2]
        s[3,4,1,1]=(-0.79188909394)*x[0]+(0.770198112387)*x[1]+(-0.0146013355132)*x[2]
        s[3,4,2,0]=(-1.70734926495)*x[0]+(-1.01352599249)*x[1]+(0.878522778655)*x[2]
        s[3,4,2,1]=(0.0249166011531)*x[0]+(-0.0714570534569)*x[1]+(1.10212723792)*x[2]
        ref[0,0,0,0]=(-0.512691010939)*x_ref[0]+(-1.31295640543)*x_ref[1]+(1.45426060088)*x_ref[2]
        ref[0,0,0,1]=(-0.57748372698)*x_ref[0]+(0.215297337003)*x_ref[1]+(0.0853439946124)*x_ref[2]
        ref[0,0,1,0]=(-0.785306554604)*x_ref[0]+(-0.0274911695421)*x_ref[1]+(0.693440767035)*x_ref[2]
        ref[0,0,1,1]=(-0.56932407873)*x_ref[0]+(-0.981593188827)*x_ref[1]+(0.87532777027)*x_ref[2]
        ref[0,0,2,0]=(0.145925104392)*x_ref[0]+(1.45139125753)*x_ref[1]+(1.25041746183)*x_ref[2]
        ref[0,0,2,1]=(-1.70110072582)*x_ref[0]+(-0.0335136921878)*x_ref[1]+(0.276204889346)*x_ref[2]
        ref[0,1,0,0]=(-0.415026748605)*x_ref[0]+(-0.322339100074)*x_ref[1]+(1.1050553524)*x_ref[2]
        ref[0,1,0,1]=(0.267336702938)*x_ref[0]+(-1.44498697586)*x_ref[1]+(0.600455793139)*x_ref[2]
        ref[0,1,1,0]=(-0.46618843993)*x_ref[0]+(0.318230555457)*x_ref[1]+(0.281788961443)*x_ref[2]
        ref[0,1,1,1]=(-1.12963168407)*x_ref[0]+(-0.661100272253)*x_ref[1]+(0.293642402995)*x_ref[2]
        ref[0,1,2,0]=(-0.0041407279814)*x_ref[0]+(0.330916770241)*x_ref[1]+(1.30265962609)*x_ref[2]
        ref[0,1,2,1]=(1.37513192104)*x_ref[0]+(-1.39837371617)*x_ref[1]+(-1.23490743489)*x_ref[2]
        ref[0,2,0,0]=(0.240920374522)*x_ref[0]+(1.30092830627)*x_ref[1]+(0.189720240056)*x_ref[2]
        ref[0,2,0,1]=(0.741337711467)*x_ref[0]+(0.00893414596869)*x_ref[1]+(1.22438764071)*x_ref[2]
        ref[0,2,1,0]=(1.0524903095)*x_ref[0]+(-1.00088524111)*x_ref[1]+(0.985885896119)*x_ref[2]
        ref[0,2,1,1]=(-0.252168964788)*x_ref[0]+(1.2685968377)*x_ref[1]+(-0.025373437998)*x_ref[2]
        ref[0,2,2,0]=(0.728523707179)*x_ref[0]+(0.00656996105623)*x_ref[1]+(1.85098622635)*x_ref[2]
        ref[0,2,2,1]=(0.193123464463)*x_ref[0]+(-0.426452202023)*x_ref[1]+(-0.144471734372)*x_ref[2]
        ref[0,3,0,0]=(0.446506355449)*x_ref[0]+(0.379921574537)*x_ref[1]+(-0.303150079281)*x_ref[2]
        ref[0,3,0,1]=(0.455373734156)*x_ref[0]+(0.134600165673)*x_ref[1]+(-0.604957024603)*x_ref[2]
        ref[0,3,1,0]=(-1.63808806962)*x_ref[0]+(1.02897840042)*x_ref[1]+(1.18123170106)*x_ref[2]
        ref[0,3,1,1]=(0.587448315418)*x_ref[0]+(0.0991725296547)*x_ref[1]+(1.33682286695)*x_ref[2]
        ref[0,3,2,0]=(-1.12085613449)*x_ref[0]+(0.454771039944)*x_ref[1]+(-0.131089499156)*x_ref[2]
        ref[0,3,2,1]=(-0.351021639018)*x_ref[0]+(0.47905316732)*x_ref[1]+(-1.07449842534)*x_ref[2]
        ref[0,4,0,0]=(0.232526828642)*x_ref[0]+(1.33907066655)*x_ref[1]+(-0.699701401092)*x_ref[2]
        ref[0,4,0,1]=(0.244905848535)*x_ref[0]+(-0.799313917773)*x_ref[1]+(-0.16513296799)*x_ref[2]
        ref[0,4,1,0]=(-1.10152273815)*x_ref[0]+(0.898176161732)*x_ref[1]+(-0.101734265033)*x_ref[2]
        ref[0,4,1,1]=(-0.0804598444811)*x_ref[0]+(-0.77819195606)*x_ref[1]+(0.277776761782)*x_ref[2]
        ref[0,4,2,0]=(1.05979692744)*x_ref[0]+(-0.195285573862)*x_ref[1]+(-0.0920717991265)*x_ref[2]
        ref[0,4,2,1]=(0.44924595646)*x_ref[0]+(0.409652246508)*x_ref[1]+(-0.0705754685165)*x_ref[2]
        ref[1,0,0,0]=(1.14991675937)*x_ref[0]+(-0.435947647215)*x_ref[1]+(-0.45614944131)*x_ref[2]
        ref[1,0,0,1]=(0.306890350036)*x_ref[0]+(-1.4400848018)*x_ref[1]+(-0.744474442385)*x_ref[2]
        ref[1,0,1,0]=(1.07930565176)*x_ref[0]+(-0.273335650157)*x_ref[1]+(-0.230960378838)*x_ref[2]
        ref[1,0,1,1]=(-0.97797563017)*x_ref[0]+(-1.3159717465)*x_ref[1]+(-0.357902208089)*x_ref[2]
        ref[1,0,2,0]=(-1.02063328249)*x_ref[0]+(-1.18907185687)*x_ref[1]+(-0.509296457902)*x_ref[2]
        ref[1,0,2,1]=(-0.508446666232)*x_ref[0]+(-0.69279963667)*x_ref[1]+(-0.497990637708)*x_ref[2]
        ref[1,1,0,0]=(-0.476141863054)*x_ref[0]+(-0.180458121777)*x_ref[1]+(-0.318855084826)*x_ref[2]
        ref[1,1,0,1]=(0.480566371651)*x_ref[0]+(-1.74196768367)*x_ref[1]+(1.23057574863)*x_ref[2]
        ref[1,1,1,0]=(-0.0213138624722)*x_ref[0]+(-0.0846532154764)*x_ref[1]+(-0.61169792023)*x_ref[2]
        ref[1,1,1,1]=(-0.389381507111)*x_ref[0]+(0.436088515729)*x_ref[1]+(-0.603019053837)*x_ref[2]
        ref[1,1,2,0]=(-1.24478786129)*x_ref[0]+(0.908984432701)*x_ref[1]+(-1.49483946407)*x_ref[2]
        ref[1,1,2,1]=(0.866266264162)*x_ref[0]+(-0.190579076131)*x_ref[1]+(-1.63524384216)*x_ref[2]
        ref[1,2,0,0]=(0.0333832222755)*x_ref[0]+(-0.551591529565)*x_ref[1]+(1.22170194609)*x_ref[2]
        ref[1,2,0,1]=(-0.551682390802)*x_ref[0]+(0.223940652339)*x_ref[1]+(0.141063634211)*x_ref[2]
        ref[1,2,1,0]=(1.17072817252)*x_ref[0]+(0.28318356571)*x_ref[1]+(-0.788096581969)*x_ref[2]
        ref[1,2,1,1]=(-1.01576849129)*x_ref[0]+(-0.502669485722)*x_ref[1]+(-0.966677100703)*x_ref[2]
        ref[1,2,2,0]=(-1.59522074083)*x_ref[0]+(-1.16016984107)*x_ref[1]+(-0.117030183187)*x_ref[2]
        ref[1,2,2,1]=(-0.979165259497)*x_ref[0]+(-0.707460176459)*x_ref[1]+(-0.275081507376)*x_ref[2]
        ref[1,3,0,0]=(0.692493088771)*x_ref[0]+(-1.04755874336)*x_ref[1]+(-0.373632326826)*x_ref[2]
        ref[1,3,0,1]=(0.717001163418)*x_ref[0]+(-0.782379062328)*x_ref[1]+(0.479776988529)*x_ref[2]
        ref[1,3,1,0]=(-0.733062748373)*x_ref[0]+(-0.217972411497)*x_ref[1]+(1.43785098293)*x_ref[2]
        ref[1,3,1,1]=(0.186430904908)*x_ref[0]+(0.375218175232)*x_ref[1]+(-0.461668650588)*x_ref[2]
        ref[1,3,2,0]=(-1.12951451301)*x_ref[0]+(-1.07034955937)*x_ref[1]+(-0.59888614247)*x_ref[2]
        ref[1,3,2,1]=(1.34023062833)*x_ref[0]+(-0.520544300692)*x_ref[1]+(-0.363001059538)*x_ref[2]
        ref[1,4,0,0]=(0.42295611834)*x_ref[0]+(-0.718852108989)*x_ref[1]+(-1.61995684533)*x_ref[2]
        ref[1,4,0,1]=(-0.422467947319)*x_ref[0]+(-0.209334480945)*x_ref[1]+(1.59579413592)*x_ref[2]
        ref[1,4,1,0]=(0.387543472208)*x_ref[0]+(-0.0579909254613)*x_ref[1]+(-1.03104850867)*x_ref[2]
        ref[1,4,1,1]=(-1.15893685519)*x_ref[0]+(-0.971412946627)*x_ref[1]+(-0.583440884462)*x_ref[2]
        ref[1,4,2,0]=(0.554042723162)*x_ref[0]+(0.082633597735)*x_ref[1]+(0.990177730288)*x_ref[2]
        ref[1,4,2,1]=(0.150114027051)*x_ref[0]+(0.479664757576)*x_ref[1]+(-1.5769513289)*x_ref[2]
        ref[2,0,0,0]=(1.33963588901)*x_ref[0]+(0.146971184264)*x_ref[1]+(-1.29921529821)*x_ref[2]
        ref[2,0,0,1]=(0.0115342792238)*x_ref[0]+(0.0451643711892)*x_ref[1]+(-0.656407384758)*x_ref[2]
        ref[2,0,1,0]=(0.175368903926)*x_ref[0]+(-0.40573949846)*x_ref[1]+(0.295387861218)*x_ref[2]
        ref[2,0,1,1]=(0.531254864472)*x_ref[0]+(-0.536193761259)*x_ref[1]+(-1.21752823876)*x_ref[2]
        ref[2,0,2,0]=(-0.362296220089)*x_ref[0]+(0.378651497918)*x_ref[1]+(-0.535206713131)*x_ref[2]
        ref[2,0,2,1]=(-0.85728549969)*x_ref[0]+(0.233179293609)*x_ref[1]+(0.345412963315)*x_ref[2]
        ref[2,1,0,0]=(1.01059063177)*x_ref[0]+(0.808116187589)*x_ref[1]+(0.468588054081)*x_ref[2]
        ref[2,1,0,1]=(-1.17212735529)*x_ref[0]+(-0.871560017197)*x_ref[1]+(1.10553749428)*x_ref[2]
        ref[2,1,1,0]=(1.00177793873)*x_ref[0]+(-0.573447532545)*x_ref[1]+(-0.80780539944)*x_ref[2]
        ref[2,1,1,1]=(-0.101912168689)*x_ref[0]+(0.0054724970873)*x_ref[1]+(0.873799338046)*x_ref[2]
        ref[2,1,2,0]=(0.162601365201)*x_ref[0]+(-0.508620489728)*x_ref[1]+(0.114861348434)*x_ref[2]
        ref[2,1,2,1]=(-1.48812556245)*x_ref[0]+(0.29841054692)*x_ref[1]+(0.773432180321)*x_ref[2]
        ref[2,2,0,0]=(0.0177469158963)*x_ref[0]+(0.140208281265)*x_ref[1]+(1.12393579682)*x_ref[2]
        ref[2,2,0,1]=(0.127835907966)*x_ref[0]+(0.150072336346)*x_ref[1]+(1.59437274191)*x_ref[2]
        ref[2,2,1,0]=(-0.549906307264)*x_ref[0]+(-0.572101026775)*x_ref[1]+(-0.747028158715)*x_ref[2]
        ref[2,2,1,1]=(0.134272548587)*x_ref[0]+(0.866777056589)*x_ref[1]+(-0.0658173600046)*x_ref[2]
        ref[2,2,2,0]=(0.160006947347)*x_ref[0]+(-0.940281662941)*x_ref[1]+(0.671280858545)*x_ref[2]
        ref[2,2,2,1]=(-1.55550646869)*x_ref[0]+(0.0485277897059)*x_ref[1]+(1.02027216264)*x_ref[2]
        ref[2,3,0,0]=(0.0931028968111)*x_ref[0]+(0.194744643733)*x_ref[1]+(1.7368246452)*x_ref[2]
        ref[2,3,0,1]=(1.66077510679)*x_ref[0]+(-0.342785893413)*x_ref[1]+(-0.278302501974)*x_ref[2]
        ref[2,3,1,0]=(-0.0719600278272)*x_ref[0]+(0.898388137955)*x_ref[1]+(-0.110811429033)*x_ref[2]
        ref[2,3,1,1]=(-0.781997247811)*x_ref[0]+(-1.32497050043)*x_ref[1]+(0.544850665961)*x_ref[2]
        ref[2,3,2,0]=(0.0290847482287)*x_ref[0]+(0.895295560038)*x_ref[1]+(-1.68555124085)*x_ref[2]
        ref[2,3,2,1]=(0.200540179332)*x_ref[0]+(-0.843842814948)*x_ref[1]+(-0.943779465271)*x_ref[2]
        ref[2,4,0,0]=(1.08999954494)*x_ref[0]+(-1.32324120166)*x_ref[1]+(-0.0152368215856)*x_ref[2]
        ref[2,4,0,1]=(1.02219584162)*x_ref[0]+(0.208263165731)*x_ref[1]+(0.0703408478849)*x_ref[2]
        ref[2,4,1,0]=(0.433866479442)*x_ref[0]+(0.327590037731)*x_ref[1]+(0.752498033029)*x_ref[2]
        ref[2,4,1,1]=(1.16322078616)*x_ref[0]+(-1.38234503561)*x_ref[1]+(-0.129845447078)*x_ref[2]
        ref[2,4,2,0]=(-0.430572860596)*x_ref[0]+(-1.48079933792)*x_ref[1]+(0.310197458762)*x_ref[2]
        ref[2,4,2,1]=(-0.0901138166505)*x_ref[0]+(0.130679443905)*x_ref[1]+(-0.482309483824)*x_ref[2]
        ref[3,0,0,0]=(0.239787389644)*x_ref[0]+(0.690754646072)*x_ref[1]+(0.745418481522)*x_ref[2]
        ref[3,0,0,1]=(0.346448639801)*x_ref[0]+(-0.287823450604)*x_ref[1]+(0.287003737791)*x_ref[2]
        ref[3,0,1,0]=(-1.50502949668)*x_ref[0]+(-1.67521500964)*x_ref[1]+(0.31630953667)*x_ref[2]
        ref[3,0,1,1]=(-0.594355763268)*x_ref[0]+(-0.262023313233)*x_ref[1]+(-1.41203969641)*x_ref[2]
        ref[3,0,2,0]=(-0.0275545216089)*x_ref[0]+(0.367433407785)*x_ref[1]+(-1.420568196)*x_ref[2]
        ref[3,0,2,1]=(-1.18797738587)*x_ref[0]+(-0.496775264183)*x_ref[1]+(0.513942381739)*x_ref[2]
        ref[3,1,0,0]=(0.475014043678)*x_ref[0]+(-1.2205702601)*x_ref[1]+(1.06353608529)*x_ref[2]
        ref[3,1,0,1]=(-0.0685583494755)*x_ref[0]+(-1.49448219897)*x_ref[1]+(-1.23320402959)*x_ref[2]
        ref[3,1,1,0]=(-0.216184790554)*x_ref[0]+(0.554046932828)*x_ref[1]+(0.697487187904)*x_ref[2]
        ref[3,1,1,1]=(-0.495370181694)*x_ref[0]+(-1.49207400185)*x_ref[1]+(-0.233763280927)*x_ref[2]
        ref[3,1,2,0]=(-0.0462888472218)*x_ref[0]+(0.371127867531)*x_ref[1]+(0.534977833221)*x_ref[2]
        ref[3,1,2,1]=(-1.27057460029)*x_ref[0]+(-0.572725702684)*x_ref[1]+(-0.871217452212)*x_ref[2]
        ref[3,2,0,0]=(-0.208549808264)*x_ref[0]+(0.847294134183)*x_ref[1]+(0.643419233684)*x_ref[2]
        ref[3,2,0,1]=(-0.850445146291)*x_ref[0]+(1.19605447923)*x_ref[1]+(-1.52958469348)*x_ref[2]
        ref[3,2,1,0]=(0.181672364304)*x_ref[0]+(-1.60804541136)*x_ref[1]+(0.068065187694)*x_ref[2]
        ref[3,2,1,1]=(-0.195284925405)*x_ref[0]+(0.330399376496)*x_ref[1]+(-0.453658972198)*x_ref[2]
        ref[3,2,2,0]=(1.44390910788)*x_ref[0]+(0.718618445973)*x_ref[1]+(-0.800501828022)*x_ref[2]
        ref[3,2,2,1]=(0.491997052311)*x_ref[0]+(-0.202966322046)*x_ref[1]+(1.39185177314)*x_ref[2]
        ref[3,3,0,0]=(-0.670818169452)*x_ref[0]+(0.842747476683)*x_ref[1]+(-0.705022077785)*x_ref[2]
        ref[3,3,0,1]=(0.729442429399)*x_ref[0]+(-1.47998423588)*x_ref[1]+(1.72626243143)*x_ref[2]
        ref[3,3,1,0]=(-1.08179915812)*x_ref[0]+(-0.117953758788)*x_ref[1]+(1.60080937509)*x_ref[2]
        ref[3,3,1,1]=(0.110049458343)*x_ref[0]+(0.571461685559)*x_ref[1]+(-0.483581817598)*x_ref[2]
        ref[3,3,2,0]=(-0.386163706678)*x_ref[0]+(-0.146336916781)*x_ref[1]+(0.0360311139076)*x_ref[2]
        ref[3,3,2,1]=(-0.779748009644)*x_ref[0]+(-0.780523381896)*x_ref[1]+(-1.20157768536)*x_ref[2]
        ref[3,4,0,0]=(1.74580569218)*x_ref[0]+(-0.318257555715)*x_ref[1]+(-0.33745202709)*x_ref[2]
        ref[3,4,0,1]=(1.1463597084)*x_ref[0]+(0.516488947224)*x_ref[1]+(-0.205035082746)*x_ref[2]
        ref[3,4,1,0]=(0.647937211136)*x_ref[0]+(0.574941676843)*x_ref[1]+(-1.00930234083)*x_ref[2]
        ref[3,4,1,1]=(-0.79188909394)*x_ref[0]+(0.770198112387)*x_ref[1]+(-0.0146013355132)*x_ref[2]
        ref[3,4,2,0]=(-1.70734926495)*x_ref[0]+(-1.01352599249)*x_ref[1]+(0.878522778655)*x_ref[2]
        ref[3,4,2,1]=(0.0249166011531)*x_ref[0]+(-0.0714570534569)*x_ref[1]+(1.10212723792)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.550361277741)*x[0]**o+(-0.431343639093)*x[0]+(-0.822566478678)*x[1]**o+(0.04348108437)*x[1]
        ref=(-0.550361277741)*x_ref[0]**o+(-0.431343639093)*x_ref[0]+(-0.822566478678)*x_ref[1]**o+(0.04348108437)*x_ref[1]
      else:
        s=(0.127530018259)*x[0]**o+(0.898197612625)*x[0]+(0.91995886196)*x[1]**o+(0.209489509472)*x[1]+(0.254553892458)*x[2]**o+(-0.00541573953033)*x[2]
        ref=(0.127530018259)*x_ref[0]**o+(0.898197612625)*x_ref[0]+(0.91995886196)*x_ref[1]**o+(0.209489509472)*x_ref[1]+(0.254553892458)*x_ref[2]**o+(-0.00541573953033)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.84171096789)*x[0]**o+(0.528614089652)*x[0]+(-0.909660047031)*x[1]**o+(0.0218639321632)*x[1]
        s[1]=(-0.196234616294)*x[0]**o+(-0.430220669717)*x[0]+(0.403594243834)*x[1]**o+(-0.555455789431)*x[1]
        ref[0]=(-0.84171096789)*x_ref[0]**o+(0.528614089652)*x_ref[0]+(-0.909660047031)*x_ref[1]**o+(0.0218639321632)*x_ref[1]
        ref[1]=(-0.196234616294)*x_ref[0]**o+(-0.430220669717)*x_ref[0]+(0.403594243834)*x_ref[1]**o+(-0.555455789431)*x_ref[1]
      else:
        s[0]=(0.236033772163)*x[0]**o+(-0.0384017745802)*x[0]+(0.594646313286)*x[1]**o+(-0.806924545742)*x[1]+(0.932734743639)*x[2]**o+(0.847787230656)*x[2]
        s[1]=(-0.472954470826)*x[0]**o+(0.874442648071)*x[0]+(0.645990600767)*x[1]**o+(-0.702715273742)*x[1]+(-0.738405297118)*x[2]**o+(0.313975158833)*x[2]
        ref[0]=(0.236033772163)*x_ref[0]**o+(-0.0384017745802)*x_ref[0]+(0.594646313286)*x_ref[1]**o+(-0.806924545742)*x_ref[1]+(0.932734743639)*x_ref[2]**o+(0.847787230656)*x_ref[2]
        ref[1]=(-0.472954470826)*x_ref[0]**o+(0.874442648071)*x_ref[0]+(0.645990600767)*x_ref[1]**o+(-0.702715273742)*x_ref[1]+(-0.738405297118)*x_ref[2]**o+(0.313975158833)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.247732656593)*x[0]**o+(-0.964555728772)*x[0]+(-0.356955683726)*x[1]**o+(0.411681255483)*x[1]
        s[0,1]=(-0.123553741129)*x[0]**o+(-0.511682932244)*x[0]+(-0.472291564447)*x[1]**o+(0.520295617334)*x[1]
        s[0,2]=(0.901714918359)*x[0]**o+(0.145027792176)*x[0]+(-0.860009935335)*x[1]**o+(-0.940605686559)*x[1]
        s[0,3]=(0.54699645799)*x[0]**o+(-0.64710426401)*x[0]+(-0.34117383247)*x[1]**o+(0.54121464127)*x[1]
        s[0,4]=(0.468931826985)*x[0]**o+(0.924799717102)*x[0]+(0.941672574697)*x[1]**o+(-0.190797551241)*x[1]
        s[1,0]=(0.0383673807592)*x[0]**o+(0.613252630082)*x[0]+(0.964352561462)*x[1]**o+(0.785919255777)*x[1]
        s[1,1]=(-0.355086324569)*x[0]**o+(0.558878226045)*x[0]+(-0.628061734673)*x[1]**o+(-0.520982697177)*x[1]
        s[1,2]=(0.723496483187)*x[0]**o+(-0.232875759963)*x[0]+(-0.17448761332)*x[1]**o+(-0.257229542886)*x[1]
        s[1,3]=(0.611356020012)*x[0]**o+(-0.28713845432)*x[0]+(-0.316681374895)*x[1]**o+(0.83320730326)*x[1]
        s[1,4]=(-0.984450469255)*x[0]**o+(0.832904475807)*x[0]+(0.150921155673)*x[1]**o+(0.158270289725)*x[1]
        s[2,0]=(0.281985706195)*x[0]**o+(0.292760063689)*x[0]+(-0.595863049075)*x[1]**o+(0.64338477328)*x[1]
        s[2,1]=(0.674281261333)*x[0]**o+(-0.282392899278)*x[0]+(0.323252947475)*x[1]**o+(-0.692156142053)*x[1]
        s[2,2]=(-0.483800717142)*x[0]**o+(-0.426306421122)*x[0]+(-0.778645643106)*x[1]**o+(0.243301515533)*x[1]
        s[2,3]=(-0.00888255736078)*x[0]**o+(-0.438384491475)*x[0]+(-0.143871169727)*x[1]**o+(0.896491696215)*x[1]
        s[2,4]=(-0.810632478514)*x[0]**o+(-0.802769421087)*x[0]+(-0.666796590858)*x[1]**o+(0.0311747062735)*x[1]
        s[3,0]=(0.919881133239)*x[0]**o+(-0.2869105463)*x[0]+(-0.33204138861)*x[1]**o+(0.790928481885)*x[1]
        s[3,1]=(-0.698545028974)*x[0]**o+(-0.526842225737)*x[0]+(-0.61919152938)*x[1]**o+(0.942452744855)*x[1]
        s[3,2]=(-0.204090966647)*x[0]**o+(-0.425679526188)*x[0]+(-0.194975545878)*x[1]**o+(0.559023937073)*x[1]
        s[3,3]=(0.852288716375)*x[0]**o+(-0.949732191466)*x[0]+(-0.0115651638362)*x[1]**o+(0.271548550263)*x[1]
        s[3,4]=(-0.684283367871)*x[0]**o+(0.887910737342)*x[0]+(-0.171186272812)*x[1]**o+(-0.069360036001)*x[1]
        ref[0,0]=(-0.247732656593)*x_ref[0]**o+(-0.964555728772)*x_ref[0]+(-0.356955683726)*x_ref[1]**o+(0.411681255483)*x_ref[1]
        ref[0,1]=(-0.123553741129)*x_ref[0]**o+(-0.511682932244)*x_ref[0]+(-0.472291564447)*x_ref[1]**o+(0.520295617334)*x_ref[1]
        ref[0,2]=(0.901714918359)*x_ref[0]**o+(0.145027792176)*x_ref[0]+(-0.860009935335)*x_ref[1]**o+(-0.940605686559)*x_ref[1]
        ref[0,3]=(0.54699645799)*x_ref[0]**o+(-0.64710426401)*x_ref[0]+(-0.34117383247)*x_ref[1]**o+(0.54121464127)*x_ref[1]
        ref[0,4]=(0.468931826985)*x_ref[0]**o+(0.924799717102)*x_ref[0]+(0.941672574697)*x_ref[1]**o+(-0.190797551241)*x_ref[1]
        ref[1,0]=(0.0383673807592)*x_ref[0]**o+(0.613252630082)*x_ref[0]+(0.964352561462)*x_ref[1]**o+(0.785919255777)*x_ref[1]
        ref[1,1]=(-0.355086324569)*x_ref[0]**o+(0.558878226045)*x_ref[0]+(-0.628061734673)*x_ref[1]**o+(-0.520982697177)*x_ref[1]
        ref[1,2]=(0.723496483187)*x_ref[0]**o+(-0.232875759963)*x_ref[0]+(-0.17448761332)*x_ref[1]**o+(-0.257229542886)*x_ref[1]
        ref[1,3]=(0.611356020012)*x_ref[0]**o+(-0.28713845432)*x_ref[0]+(-0.316681374895)*x_ref[1]**o+(0.83320730326)*x_ref[1]
        ref[1,4]=(-0.984450469255)*x_ref[0]**o+(0.832904475807)*x_ref[0]+(0.150921155673)*x_ref[1]**o+(0.158270289725)*x_ref[1]
        ref[2,0]=(0.281985706195)*x_ref[0]**o+(0.292760063689)*x_ref[0]+(-0.595863049075)*x_ref[1]**o+(0.64338477328)*x_ref[1]
        ref[2,1]=(0.674281261333)*x_ref[0]**o+(-0.282392899278)*x_ref[0]+(0.323252947475)*x_ref[1]**o+(-0.692156142053)*x_ref[1]
        ref[2,2]=(-0.483800717142)*x_ref[0]**o+(-0.426306421122)*x_ref[0]+(-0.778645643106)*x_ref[1]**o+(0.243301515533)*x_ref[1]
        ref[2,3]=(-0.00888255736078)*x_ref[0]**o+(-0.438384491475)*x_ref[0]+(-0.143871169727)*x_ref[1]**o+(0.896491696215)*x_ref[1]
        ref[2,4]=(-0.810632478514)*x_ref[0]**o+(-0.802769421087)*x_ref[0]+(-0.666796590858)*x_ref[1]**o+(0.0311747062735)*x_ref[1]
        ref[3,0]=(0.919881133239)*x_ref[0]**o+(-0.2869105463)*x_ref[0]+(-0.33204138861)*x_ref[1]**o+(0.790928481885)*x_ref[1]
        ref[3,1]=(-0.698545028974)*x_ref[0]**o+(-0.526842225737)*x_ref[0]+(-0.61919152938)*x_ref[1]**o+(0.942452744855)*x_ref[1]
        ref[3,2]=(-0.204090966647)*x_ref[0]**o+(-0.425679526188)*x_ref[0]+(-0.194975545878)*x_ref[1]**o+(0.559023937073)*x_ref[1]
        ref[3,3]=(0.852288716375)*x_ref[0]**o+(-0.949732191466)*x_ref[0]+(-0.0115651638362)*x_ref[1]**o+(0.271548550263)*x_ref[1]
        ref[3,4]=(-0.684283367871)*x_ref[0]**o+(0.887910737342)*x_ref[0]+(-0.171186272812)*x_ref[1]**o+(-0.069360036001)*x_ref[1]
      else:
        s[0,0]=(-0.146549436504)*x[0]**o+(-0.929982499842)*x[0]+(0.0884445727171)*x[1]**o+(-0.396974650474)*x[1]+(-0.0647429699084)*x[2]**o+(0.109485258811)*x[2]
        s[0,1]=(0.542977833506)*x[0]**o+(0.251279104804)*x[0]+(-0.590300769519)*x[1]**o+(-0.0490256576926)*x[1]+(-0.706162640719)*x[2]**o+(0.307510689521)*x[2]
        s[0,2]=(0.239778231777)*x[0]**o+(-0.790527320914)*x[0]+(0.344144886282)*x[1]**o+(-0.143939808236)*x[1]+(-0.799065328837)*x[2]**o+(0.297556608867)*x[2]
        s[0,3]=(-0.657196676576)*x[0]**o+(0.750906099544)*x[0]+(0.982936179835)*x[1]**o+(-0.503213227605)*x[1]+(0.733421910241)*x[2]**o+(0.484345689534)*x[2]
        s[0,4]=(-0.10428912667)*x[0]**o+(-0.955639074345)*x[0]+(-0.120263699632)*x[1]**o+(0.358019739226)*x[1]+(-0.967358029478)*x[2]**o+(0.515480564718)*x[2]
        s[1,0]=(0.762873752492)*x[0]**o+(-0.865993029835)*x[0]+(-0.227215328566)*x[1]**o+(0.628614878944)*x[1]+(-0.481609778611)*x[2]**o+(0.680214047083)*x[2]
        s[1,1]=(0.942225721603)*x[0]**o+(0.53391180863)*x[0]+(0.631906892953)*x[1]**o+(-0.907916670051)*x[1]+(-0.621285107992)*x[2]**o+(-0.613776303306)*x[2]
        s[1,2]=(-0.911879095454)*x[0]**o+(0.79055949266)*x[0]+(0.165379072622)*x[1]**o+(0.350589081179)*x[1]+(0.980050071498)*x[2]**o+(-0.728078468886)*x[2]
        s[1,3]=(0.927875284375)*x[0]**o+(0.267789172844)*x[0]+(-0.801387363889)*x[1]**o+(0.445894170128)*x[1]+(0.978344523663)*x[2]**o+(0.290141654981)*x[2]
        s[1,4]=(0.700012915504)*x[0]**o+(-0.247394585438)*x[0]+(-0.889582786398)*x[1]**o+(0.65907227973)*x[1]+(-0.0559926046073)*x[2]**o+(-0.954804094449)*x[2]
        s[2,0]=(0.303693881754)*x[0]**o+(0.409350273091)*x[0]+(0.611015645589)*x[1]**o+(-0.0919333788167)*x[1]+(0.161448267028)*x[2]**o+(0.200336560957)*x[2]
        s[2,1]=(0.010588069689)*x[0]**o+(-0.193147278154)*x[0]+(-0.846857030144)*x[1]**o+(-0.167769695014)*x[1]+(0.518409682456)*x[2]**o+(-0.882303521363)*x[2]
        s[2,2]=(0.82323572893)*x[0]**o+(0.712556931497)*x[0]+(0.677658458739)*x[1]**o+(0.582443300339)*x[1]+(0.291648194217)*x[2]**o+(-0.485160007834)*x[2]
        s[2,3]=(0.782471869781)*x[0]**o+(-0.841455423835)*x[0]+(0.546825262189)*x[1]**o+(0.529412907865)*x[1]+(0.0226844946614)*x[2]**o+(0.712355610287)*x[2]
        s[2,4]=(-0.0954280871852)*x[0]**o+(0.535095858201)*x[0]+(-0.107028351709)*x[1]**o+(0.446785157036)*x[1]+(-0.31582995285)*x[2]**o+(-0.00891179040655)*x[2]
        s[3,0]=(0.0804857346776)*x[0]**o+(-0.281583457975)*x[0]+(0.220634050401)*x[1]**o+(0.876350726718)*x[1]+(0.790789468637)*x[2]**o+(0.202028059276)*x[2]
        s[3,1]=(0.545828636902)*x[0]**o+(0.166260836796)*x[0]+(-0.357145146842)*x[1]**o+(-0.747029491567)*x[1]+(-0.965612583637)*x[2]**o+(0.863127371545)*x[2]
        s[3,2]=(-0.829350796721)*x[0]**o+(0.691296313876)*x[0]+(-0.170515248987)*x[1]**o+(-0.739412695686)*x[1]+(-0.244741653636)*x[2]**o+(-0.414333914881)*x[2]
        s[3,3]=(-0.636245569881)*x[0]**o+(-0.374567359716)*x[0]+(-0.80804727564)*x[1]**o+(0.551213025284)*x[1]+(-0.210200137584)*x[2]**o+(-0.975486200974)*x[2]
        s[3,4]=(0.195033054442)*x[0]**o+(-0.5792439888)*x[0]+(-0.366295355782)*x[1]**o+(-0.88151827146)*x[1]+(-0.438942980834)*x[2]**o+(0.378273994019)*x[2]
        ref[0,0]=(-0.146549436504)*x_ref[0]**o+(-0.929982499842)*x_ref[0]+(0.0884445727171)*x_ref[1]**o+(-0.396974650474)*x_ref[1]+(-0.0647429699084)*x_ref[2]**o+(0.109485258811)*x_ref[2]
        ref[0,1]=(0.542977833506)*x_ref[0]**o+(0.251279104804)*x_ref[0]+(-0.590300769519)*x_ref[1]**o+(-0.0490256576926)*x_ref[1]+(-0.706162640719)*x_ref[2]**o+(0.307510689521)*x_ref[2]
        ref[0,2]=(0.239778231777)*x_ref[0]**o+(-0.790527320914)*x_ref[0]+(0.344144886282)*x_ref[1]**o+(-0.143939808236)*x_ref[1]+(-0.799065328837)*x_ref[2]**o+(0.297556608867)*x_ref[2]
        ref[0,3]=(-0.657196676576)*x_ref[0]**o+(0.750906099544)*x_ref[0]+(0.982936179835)*x_ref[1]**o+(-0.503213227605)*x_ref[1]+(0.733421910241)*x_ref[2]**o+(0.484345689534)*x_ref[2]
        ref[0,4]=(-0.10428912667)*x_ref[0]**o+(-0.955639074345)*x_ref[0]+(-0.120263699632)*x_ref[1]**o+(0.358019739226)*x_ref[1]+(-0.967358029478)*x_ref[2]**o+(0.515480564718)*x_ref[2]
        ref[1,0]=(0.762873752492)*x_ref[0]**o+(-0.865993029835)*x_ref[0]+(-0.227215328566)*x_ref[1]**o+(0.628614878944)*x_ref[1]+(-0.481609778611)*x_ref[2]**o+(0.680214047083)*x_ref[2]
        ref[1,1]=(0.942225721603)*x_ref[0]**o+(0.53391180863)*x_ref[0]+(0.631906892953)*x_ref[1]**o+(-0.907916670051)*x_ref[1]+(-0.621285107992)*x_ref[2]**o+(-0.613776303306)*x_ref[2]
        ref[1,2]=(-0.911879095454)*x_ref[0]**o+(0.79055949266)*x_ref[0]+(0.165379072622)*x_ref[1]**o+(0.350589081179)*x_ref[1]+(0.980050071498)*x_ref[2]**o+(-0.728078468886)*x_ref[2]
        ref[1,3]=(0.927875284375)*x_ref[0]**o+(0.267789172844)*x_ref[0]+(-0.801387363889)*x_ref[1]**o+(0.445894170128)*x_ref[1]+(0.978344523663)*x_ref[2]**o+(0.290141654981)*x_ref[2]
        ref[1,4]=(0.700012915504)*x_ref[0]**o+(-0.247394585438)*x_ref[0]+(-0.889582786398)*x_ref[1]**o+(0.65907227973)*x_ref[1]+(-0.0559926046073)*x_ref[2]**o+(-0.954804094449)*x_ref[2]
        ref[2,0]=(0.303693881754)*x_ref[0]**o+(0.409350273091)*x_ref[0]+(0.611015645589)*x_ref[1]**o+(-0.0919333788167)*x_ref[1]+(0.161448267028)*x_ref[2]**o+(0.200336560957)*x_ref[2]
        ref[2,1]=(0.010588069689)*x_ref[0]**o+(-0.193147278154)*x_ref[0]+(-0.846857030144)*x_ref[1]**o+(-0.167769695014)*x_ref[1]+(0.518409682456)*x_ref[2]**o+(-0.882303521363)*x_ref[2]
        ref[2,2]=(0.82323572893)*x_ref[0]**o+(0.712556931497)*x_ref[0]+(0.677658458739)*x_ref[1]**o+(0.582443300339)*x_ref[1]+(0.291648194217)*x_ref[2]**o+(-0.485160007834)*x_ref[2]
        ref[2,3]=(0.782471869781)*x_ref[0]**o+(-0.841455423835)*x_ref[0]+(0.546825262189)*x_ref[1]**o+(0.529412907865)*x_ref[1]+(0.0226844946614)*x_ref[2]**o+(0.712355610287)*x_ref[2]
        ref[2,4]=(-0.0954280871852)*x_ref[0]**o+(0.535095858201)*x_ref[0]+(-0.107028351709)*x_ref[1]**o+(0.446785157036)*x_ref[1]+(-0.31582995285)*x_ref[2]**o+(-0.00891179040655)*x_ref[2]
        ref[3,0]=(0.0804857346776)*x_ref[0]**o+(-0.281583457975)*x_ref[0]+(0.220634050401)*x_ref[1]**o+(0.876350726718)*x_ref[1]+(0.790789468637)*x_ref[2]**o+(0.202028059276)*x_ref[2]
        ref[3,1]=(0.545828636902)*x_ref[0]**o+(0.166260836796)*x_ref[0]+(-0.357145146842)*x_ref[1]**o+(-0.747029491567)*x_ref[1]+(-0.965612583637)*x_ref[2]**o+(0.863127371545)*x_ref[2]
        ref[3,2]=(-0.829350796721)*x_ref[0]**o+(0.691296313876)*x_ref[0]+(-0.170515248987)*x_ref[1]**o+(-0.739412695686)*x_ref[1]+(-0.244741653636)*x_ref[2]**o+(-0.414333914881)*x_ref[2]
        ref[3,3]=(-0.636245569881)*x_ref[0]**o+(-0.374567359716)*x_ref[0]+(-0.80804727564)*x_ref[1]**o+(0.551213025284)*x_ref[1]+(-0.210200137584)*x_ref[2]**o+(-0.975486200974)*x_ref[2]
        ref[3,4]=(0.195033054442)*x_ref[0]**o+(-0.5792439888)*x_ref[0]+(-0.366295355782)*x_ref[1]**o+(-0.88151827146)*x_ref[1]+(-0.438942980834)*x_ref[2]**o+(0.378273994019)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.0313098361481)*x[0]**o+(-0.943408799922)*x[0]+(0.695838509745)*x[1]**o+(-0.332495099387)*x[1]
        s[0,0,1]=(0.431627451331)*x[0]**o+(-0.628725729291)*x[0]+(-0.35286078386)*x[1]**o+(-0.955736740817)*x[1]
        s[0,1,0]=(0.968131554454)*x[0]**o+(0.0281368471267)*x[0]+(0.0459790991737)*x[1]**o+(0.393556307708)*x[1]
        s[0,1,1]=(-0.204304807192)*x[0]**o+(-0.475856856629)*x[0]+(-0.480008001652)*x[1]**o+(-0.550159532343)*x[1]
        s[1,0,0]=(-0.367566613332)*x[0]**o+(-0.538384769558)*x[0]+(0.96041325378)*x[1]**o+(0.435390124318)*x[1]
        s[1,0,1]=(0.204676073638)*x[0]**o+(0.250019350651)*x[0]+(0.103416708864)*x[1]**o+(-0.524130759283)*x[1]
        s[1,1,0]=(0.855261863743)*x[0]**o+(-0.675190292289)*x[0]+(-0.514777393648)*x[1]**o+(0.942267330164)*x[1]
        s[1,1,1]=(0.508909921316)*x[0]**o+(0.499552817754)*x[0]+(0.303561948422)*x[1]**o+(0.248598631154)*x[1]
        s[2,0,0]=(-0.642212923403)*x[0]**o+(0.956027472963)*x[0]+(0.056374728577)*x[1]**o+(0.486941288345)*x[1]
        s[2,0,1]=(-0.860437111201)*x[0]**o+(0.922745184749)*x[0]+(0.448750379137)*x[1]**o+(0.733106647733)*x[1]
        s[2,1,0]=(-0.399141476478)*x[0]**o+(-0.623902788197)*x[0]+(0.185646936042)*x[1]**o+(0.158930593879)*x[1]
        s[2,1,1]=(-0.25915892886)*x[0]**o+(-0.534454787558)*x[0]+(-0.0196197895513)*x[1]**o+(-0.921288033314)*x[1]
        s[3,0,0]=(0.718044900278)*x[0]**o+(-0.881912685669)*x[0]+(0.328338604205)*x[1]**o+(-0.364500357873)*x[1]
        s[3,0,1]=(-0.316477236945)*x[0]**o+(-0.0594343871907)*x[0]+(0.564350847567)*x[1]**o+(0.612426270173)*x[1]
        s[3,1,0]=(0.862764583147)*x[0]**o+(0.0895137700398)*x[0]+(0.376837454709)*x[1]**o+(-0.276425178685)*x[1]
        s[3,1,1]=(0.911550499093)*x[0]**o+(0.47964604046)*x[0]+(-0.0936664482923)*x[1]**o+(-0.355680673199)*x[1]
        s[4,0,0]=(-0.126679727917)*x[0]**o+(0.0230703577025)*x[0]+(-0.749904232032)*x[1]**o+(0.120202683819)*x[1]
        s[4,0,1]=(0.859805437446)*x[0]**o+(-0.0547285079407)*x[0]+(-0.0209396655171)*x[1]**o+(-0.835017485682)*x[1]
        s[4,1,0]=(0.223141900386)*x[0]**o+(0.570612780326)*x[0]+(0.147581871479)*x[1]**o+(-0.0724129492652)*x[1]
        s[4,1,1]=(0.479285654861)*x[0]**o+(-0.368842797927)*x[0]+(0.516904484685)*x[1]**o+(-0.74560082895)*x[1]
        s[5,0,0]=(0.895865547536)*x[0]**o+(0.354087196226)*x[0]+(-0.596919891679)*x[1]**o+(0.512426451597)*x[1]
        s[5,0,1]=(0.467255278736)*x[0]**o+(-0.724626510449)*x[0]+(0.0624000180202)*x[1]**o+(0.860859099558)*x[1]
        s[5,1,0]=(-0.632133782251)*x[0]**o+(0.639067592984)*x[0]+(0.974027711622)*x[1]**o+(0.835256883107)*x[1]
        s[5,1,1]=(0.0618918945789)*x[0]**o+(0.541323411311)*x[0]+(-0.229029906018)*x[1]**o+(0.489605390069)*x[1]
        ref[0,0,0]=(0.0313098361481)*x_ref[0]**o+(-0.943408799922)*x_ref[0]+(0.695838509745)*x_ref[1]**o+(-0.332495099387)*x_ref[1]
        ref[0,0,1]=(0.431627451331)*x_ref[0]**o+(-0.628725729291)*x_ref[0]+(-0.35286078386)*x_ref[1]**o+(-0.955736740817)*x_ref[1]
        ref[0,1,0]=(0.968131554454)*x_ref[0]**o+(0.0281368471267)*x_ref[0]+(0.0459790991737)*x_ref[1]**o+(0.393556307708)*x_ref[1]
        ref[0,1,1]=(-0.204304807192)*x_ref[0]**o+(-0.475856856629)*x_ref[0]+(-0.480008001652)*x_ref[1]**o+(-0.550159532343)*x_ref[1]
        ref[1,0,0]=(-0.367566613332)*x_ref[0]**o+(-0.538384769558)*x_ref[0]+(0.96041325378)*x_ref[1]**o+(0.435390124318)*x_ref[1]
        ref[1,0,1]=(0.204676073638)*x_ref[0]**o+(0.250019350651)*x_ref[0]+(0.103416708864)*x_ref[1]**o+(-0.524130759283)*x_ref[1]
        ref[1,1,0]=(0.855261863743)*x_ref[0]**o+(-0.675190292289)*x_ref[0]+(-0.514777393648)*x_ref[1]**o+(0.942267330164)*x_ref[1]
        ref[1,1,1]=(0.508909921316)*x_ref[0]**o+(0.499552817754)*x_ref[0]+(0.303561948422)*x_ref[1]**o+(0.248598631154)*x_ref[1]
        ref[2,0,0]=(-0.642212923403)*x_ref[0]**o+(0.956027472963)*x_ref[0]+(0.056374728577)*x_ref[1]**o+(0.486941288345)*x_ref[1]
        ref[2,0,1]=(-0.860437111201)*x_ref[0]**o+(0.922745184749)*x_ref[0]+(0.448750379137)*x_ref[1]**o+(0.733106647733)*x_ref[1]
        ref[2,1,0]=(-0.399141476478)*x_ref[0]**o+(-0.623902788197)*x_ref[0]+(0.185646936042)*x_ref[1]**o+(0.158930593879)*x_ref[1]
        ref[2,1,1]=(-0.25915892886)*x_ref[0]**o+(-0.534454787558)*x_ref[0]+(-0.0196197895513)*x_ref[1]**o+(-0.921288033314)*x_ref[1]
        ref[3,0,0]=(0.718044900278)*x_ref[0]**o+(-0.881912685669)*x_ref[0]+(0.328338604205)*x_ref[1]**o+(-0.364500357873)*x_ref[1]
        ref[3,0,1]=(-0.316477236945)*x_ref[0]**o+(-0.0594343871907)*x_ref[0]+(0.564350847567)*x_ref[1]**o+(0.612426270173)*x_ref[1]
        ref[3,1,0]=(0.862764583147)*x_ref[0]**o+(0.0895137700398)*x_ref[0]+(0.376837454709)*x_ref[1]**o+(-0.276425178685)*x_ref[1]
        ref[3,1,1]=(0.911550499093)*x_ref[0]**o+(0.47964604046)*x_ref[0]+(-0.0936664482923)*x_ref[1]**o+(-0.355680673199)*x_ref[1]
        ref[4,0,0]=(-0.126679727917)*x_ref[0]**o+(0.0230703577025)*x_ref[0]+(-0.749904232032)*x_ref[1]**o+(0.120202683819)*x_ref[1]
        ref[4,0,1]=(0.859805437446)*x_ref[0]**o+(-0.0547285079407)*x_ref[0]+(-0.0209396655171)*x_ref[1]**o+(-0.835017485682)*x_ref[1]
        ref[4,1,0]=(0.223141900386)*x_ref[0]**o+(0.570612780326)*x_ref[0]+(0.147581871479)*x_ref[1]**o+(-0.0724129492652)*x_ref[1]
        ref[4,1,1]=(0.479285654861)*x_ref[0]**o+(-0.368842797927)*x_ref[0]+(0.516904484685)*x_ref[1]**o+(-0.74560082895)*x_ref[1]
        ref[5,0,0]=(0.895865547536)*x_ref[0]**o+(0.354087196226)*x_ref[0]+(-0.596919891679)*x_ref[1]**o+(0.512426451597)*x_ref[1]
        ref[5,0,1]=(0.467255278736)*x_ref[0]**o+(-0.724626510449)*x_ref[0]+(0.0624000180202)*x_ref[1]**o+(0.860859099558)*x_ref[1]
        ref[5,1,0]=(-0.632133782251)*x_ref[0]**o+(0.639067592984)*x_ref[0]+(0.974027711622)*x_ref[1]**o+(0.835256883107)*x_ref[1]
        ref[5,1,1]=(0.0618918945789)*x_ref[0]**o+(0.541323411311)*x_ref[0]+(-0.229029906018)*x_ref[1]**o+(0.489605390069)*x_ref[1]
      else:
        s[0,0,0]=(-0.110630395456)*x[0]**o+(0.409103918414)*x[0]+(0.574294002864)*x[1]**o+(0.187245897037)*x[1]+(-0.577953556131)*x[2]**o+(-0.0999500699254)*x[2]
        s[0,0,1]=(0.60081436578)*x[0]**o+(0.0576573739024)*x[0]+(-0.151144644304)*x[1]**o+(-0.38821542392)*x[1]+(-0.698001301019)*x[2]**o+(0.959658889512)*x[2]
        s[0,1,0]=(-0.0612229861582)*x[0]**o+(-0.221080853704)*x[0]+(-0.806155040024)*x[1]**o+(0.810149964095)*x[1]+(-0.970105552614)*x[2]**o+(0.901396266753)*x[2]
        s[0,1,1]=(-0.1815051419)*x[0]**o+(0.205375880458)*x[0]+(-0.865264408549)*x[1]**o+(0.76028929015)*x[1]+(0.520419877863)*x[2]**o+(0.593892129073)*x[2]
        s[1,0,0]=(0.585747762904)*x[0]**o+(0.483216361168)*x[0]+(0.5552914794)*x[1]**o+(0.297595584239)*x[1]+(-0.459345030181)*x[2]**o+(0.740748940805)*x[2]
        s[1,0,1]=(-0.482750725051)*x[0]**o+(0.588601093972)*x[0]+(0.440244419924)*x[1]**o+(0.222485858829)*x[1]+(0.797934157822)*x[2]**o+(0.15404361699)*x[2]
        s[1,1,0]=(0.498237939377)*x[0]**o+(0.0115319665364)*x[0]+(-0.295972109504)*x[1]**o+(0.256948190994)*x[1]+(-0.711317031697)*x[2]**o+(-0.00740724985641)*x[2]
        s[1,1,1]=(-0.861126786241)*x[0]**o+(0.371079744214)*x[0]+(-0.660762067211)*x[1]**o+(0.081182574056)*x[1]+(-0.77101451732)*x[2]**o+(0.899135035361)*x[2]
        s[2,0,0]=(0.666936736186)*x[0]**o+(0.951940175322)*x[0]+(-0.0863391086362)*x[1]**o+(0.057346020976)*x[1]+(0.442105193705)*x[2]**o+(-0.07460449107)*x[2]
        s[2,0,1]=(0.980045171639)*x[0]**o+(0.445573999083)*x[0]+(0.442430445618)*x[1]**o+(0.551873411845)*x[1]+(0.0674935335431)*x[2]**o+(0.772168296609)*x[2]
        s[2,1,0]=(-0.271924080302)*x[0]**o+(-0.081769641884)*x[0]+(-0.902170977359)*x[1]**o+(-0.37134444501)*x[1]+(0.490234732275)*x[2]**o+(-0.972142319324)*x[2]
        s[2,1,1]=(-0.931544234095)*x[0]**o+(-0.996346369309)*x[0]+(0.313123243249)*x[1]**o+(-0.377115311908)*x[1]+(-0.840592475346)*x[2]**o+(-0.147076866517)*x[2]
        s[3,0,0]=(0.920849350915)*x[0]**o+(-0.268166526715)*x[0]+(-0.639196917804)*x[1]**o+(0.847896119248)*x[1]+(-0.488855455712)*x[2]**o+(-0.893202374665)*x[2]
        s[3,0,1]=(-0.827357725384)*x[0]**o+(-0.675036050349)*x[0]+(-0.0259058469974)*x[1]**o+(-0.143686855851)*x[1]+(0.431815308934)*x[2]**o+(-0.123586502935)*x[2]
        s[3,1,0]=(-0.482077477865)*x[0]**o+(0.256319753359)*x[0]+(0.688028798818)*x[1]**o+(-0.0557357547278)*x[1]+(0.025118048506)*x[2]**o+(-0.53002694196)*x[2]
        s[3,1,1]=(-0.00203098763538)*x[0]**o+(0.922732878998)*x[0]+(0.470339945576)*x[1]**o+(-0.288881331631)*x[1]+(0.716132166051)*x[2]**o+(0.376878681481)*x[2]
        s[4,0,0]=(0.308528023012)*x[0]**o+(-0.68058295085)*x[0]+(-0.633458164572)*x[1]**o+(-0.236218088529)*x[1]+(-0.853026908063)*x[2]**o+(-0.560991626371)*x[2]
        s[4,0,1]=(-0.276542785164)*x[0]**o+(-0.858464870187)*x[0]+(-0.339731559043)*x[1]**o+(0.625570996156)*x[1]+(-0.971283564072)*x[2]**o+(-0.893431722273)*x[2]
        s[4,1,0]=(-0.132528753997)*x[0]**o+(0.223249232076)*x[0]+(0.841435681826)*x[1]**o+(-0.0253915085492)*x[1]+(-0.518508375076)*x[2]**o+(-0.698041771818)*x[2]
        s[4,1,1]=(-0.978796591572)*x[0]**o+(0.780188335993)*x[0]+(-0.812394692896)*x[1]**o+(0.518054410279)*x[1]+(0.571803768515)*x[2]**o+(-0.0269003440822)*x[2]
        s[5,0,0]=(0.562021340848)*x[0]**o+(0.68880638377)*x[0]+(0.85211165801)*x[1]**o+(0.0451146367354)*x[1]+(0.241282824032)*x[2]**o+(0.661805360853)*x[2]
        s[5,0,1]=(0.500871152579)*x[0]**o+(0.21948004589)*x[0]+(0.731267125182)*x[1]**o+(-0.0641860674019)*x[1]+(0.109187270896)*x[2]**o+(-0.843149143759)*x[2]
        s[5,1,0]=(0.478541279317)*x[0]**o+(-0.34649289488)*x[0]+(0.551551005174)*x[1]**o+(-0.750137447886)*x[1]+(-0.112460330983)*x[2]**o+(-0.731771671002)*x[2]
        s[5,1,1]=(-0.93838335953)*x[0]**o+(0.499456439132)*x[0]+(0.655920157621)*x[1]**o+(0.233198833152)*x[1]+(0.444839351692)*x[2]**o+(0.666750348313)*x[2]
        ref[0,0,0]=(-0.110630395456)*x_ref[0]**o+(0.409103918414)*x_ref[0]+(0.574294002864)*x_ref[1]**o+(0.187245897037)*x_ref[1]+(-0.577953556131)*x_ref[2]**o+(-0.0999500699254)*x_ref[2]
        ref[0,0,1]=(0.60081436578)*x_ref[0]**o+(0.0576573739024)*x_ref[0]+(-0.151144644304)*x_ref[1]**o+(-0.38821542392)*x_ref[1]+(-0.698001301019)*x_ref[2]**o+(0.959658889512)*x_ref[2]
        ref[0,1,0]=(-0.0612229861582)*x_ref[0]**o+(-0.221080853704)*x_ref[0]+(-0.806155040024)*x_ref[1]**o+(0.810149964095)*x_ref[1]+(-0.970105552614)*x_ref[2]**o+(0.901396266753)*x_ref[2]
        ref[0,1,1]=(-0.1815051419)*x_ref[0]**o+(0.205375880458)*x_ref[0]+(-0.865264408549)*x_ref[1]**o+(0.76028929015)*x_ref[1]+(0.520419877863)*x_ref[2]**o+(0.593892129073)*x_ref[2]
        ref[1,0,0]=(0.585747762904)*x_ref[0]**o+(0.483216361168)*x_ref[0]+(0.5552914794)*x_ref[1]**o+(0.297595584239)*x_ref[1]+(-0.459345030181)*x_ref[2]**o+(0.740748940805)*x_ref[2]
        ref[1,0,1]=(-0.482750725051)*x_ref[0]**o+(0.588601093972)*x_ref[0]+(0.440244419924)*x_ref[1]**o+(0.222485858829)*x_ref[1]+(0.797934157822)*x_ref[2]**o+(0.15404361699)*x_ref[2]
        ref[1,1,0]=(0.498237939377)*x_ref[0]**o+(0.0115319665364)*x_ref[0]+(-0.295972109504)*x_ref[1]**o+(0.256948190994)*x_ref[1]+(-0.711317031697)*x_ref[2]**o+(-0.00740724985641)*x_ref[2]
        ref[1,1,1]=(-0.861126786241)*x_ref[0]**o+(0.371079744214)*x_ref[0]+(-0.660762067211)*x_ref[1]**o+(0.081182574056)*x_ref[1]+(-0.77101451732)*x_ref[2]**o+(0.899135035361)*x_ref[2]
        ref[2,0,0]=(0.666936736186)*x_ref[0]**o+(0.951940175322)*x_ref[0]+(-0.0863391086362)*x_ref[1]**o+(0.057346020976)*x_ref[1]+(0.442105193705)*x_ref[2]**o+(-0.07460449107)*x_ref[2]
        ref[2,0,1]=(0.980045171639)*x_ref[0]**o+(0.445573999083)*x_ref[0]+(0.442430445618)*x_ref[1]**o+(0.551873411845)*x_ref[1]+(0.0674935335431)*x_ref[2]**o+(0.772168296609)*x_ref[2]
        ref[2,1,0]=(-0.271924080302)*x_ref[0]**o+(-0.081769641884)*x_ref[0]+(-0.902170977359)*x_ref[1]**o+(-0.37134444501)*x_ref[1]+(0.490234732275)*x_ref[2]**o+(-0.972142319324)*x_ref[2]
        ref[2,1,1]=(-0.931544234095)*x_ref[0]**o+(-0.996346369309)*x_ref[0]+(0.313123243249)*x_ref[1]**o+(-0.377115311908)*x_ref[1]+(-0.840592475346)*x_ref[2]**o+(-0.147076866517)*x_ref[2]
        ref[3,0,0]=(0.920849350915)*x_ref[0]**o+(-0.268166526715)*x_ref[0]+(-0.639196917804)*x_ref[1]**o+(0.847896119248)*x_ref[1]+(-0.488855455712)*x_ref[2]**o+(-0.893202374665)*x_ref[2]
        ref[3,0,1]=(-0.827357725384)*x_ref[0]**o+(-0.675036050349)*x_ref[0]+(-0.0259058469974)*x_ref[1]**o+(-0.143686855851)*x_ref[1]+(0.431815308934)*x_ref[2]**o+(-0.123586502935)*x_ref[2]
        ref[3,1,0]=(-0.482077477865)*x_ref[0]**o+(0.256319753359)*x_ref[0]+(0.688028798818)*x_ref[1]**o+(-0.0557357547278)*x_ref[1]+(0.025118048506)*x_ref[2]**o+(-0.53002694196)*x_ref[2]
        ref[3,1,1]=(-0.00203098763538)*x_ref[0]**o+(0.922732878998)*x_ref[0]+(0.470339945576)*x_ref[1]**o+(-0.288881331631)*x_ref[1]+(0.716132166051)*x_ref[2]**o+(0.376878681481)*x_ref[2]
        ref[4,0,0]=(0.308528023012)*x_ref[0]**o+(-0.68058295085)*x_ref[0]+(-0.633458164572)*x_ref[1]**o+(-0.236218088529)*x_ref[1]+(-0.853026908063)*x_ref[2]**o+(-0.560991626371)*x_ref[2]
        ref[4,0,1]=(-0.276542785164)*x_ref[0]**o+(-0.858464870187)*x_ref[0]+(-0.339731559043)*x_ref[1]**o+(0.625570996156)*x_ref[1]+(-0.971283564072)*x_ref[2]**o+(-0.893431722273)*x_ref[2]
        ref[4,1,0]=(-0.132528753997)*x_ref[0]**o+(0.223249232076)*x_ref[0]+(0.841435681826)*x_ref[1]**o+(-0.0253915085492)*x_ref[1]+(-0.518508375076)*x_ref[2]**o+(-0.698041771818)*x_ref[2]
        ref[4,1,1]=(-0.978796591572)*x_ref[0]**o+(0.780188335993)*x_ref[0]+(-0.812394692896)*x_ref[1]**o+(0.518054410279)*x_ref[1]+(0.571803768515)*x_ref[2]**o+(-0.0269003440822)*x_ref[2]
        ref[5,0,0]=(0.562021340848)*x_ref[0]**o+(0.68880638377)*x_ref[0]+(0.85211165801)*x_ref[1]**o+(0.0451146367354)*x_ref[1]+(0.241282824032)*x_ref[2]**o+(0.661805360853)*x_ref[2]
        ref[5,0,1]=(0.500871152579)*x_ref[0]**o+(0.21948004589)*x_ref[0]+(0.731267125182)*x_ref[1]**o+(-0.0641860674019)*x_ref[1]+(0.109187270896)*x_ref[2]**o+(-0.843149143759)*x_ref[2]
        ref[5,1,0]=(0.478541279317)*x_ref[0]**o+(-0.34649289488)*x_ref[0]+(0.551551005174)*x_ref[1]**o+(-0.750137447886)*x_ref[1]+(-0.112460330983)*x_ref[2]**o+(-0.731771671002)*x_ref[2]
        ref[5,1,1]=(-0.93838335953)*x_ref[0]**o+(0.499456439132)*x_ref[0]+(0.655920157621)*x_ref[1]**o+(0.233198833152)*x_ref[1]+(0.444839351692)*x_ref[2]**o+(0.666750348313)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onFunctionOnBoundary_fromSymbol_FunctionOnBoundary_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the FunctionOnBoundary

      assumptions: self.domain supports inpterpolation from FunctionOnBoundary onto FunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=FunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.771805529053)*x[0]**o+(-0.00794071414241)*x[0]+(-0.546566317942)*x[1]**o+(0.209469633398)*x[1]
        s[0,0,0,1]=(-0.0138300734518)*x[0]**o+(0.783507392816)*x[0]+(0.637453082951)*x[1]**o+(-0.0262186183303)*x[1]
        s[0,0,1,0]=(-0.622230122991)*x[0]**o+(0.760867623401)*x[0]+(0.340084769283)*x[1]**o+(0.493207077521)*x[1]
        s[0,0,1,1]=(0.772195055793)*x[0]**o+(-0.481066683803)*x[0]+(0.198371969404)*x[1]**o+(0.841664205482)*x[1]
        s[0,0,2,0]=(0.858213005013)*x[0]**o+(0.819391260667)*x[0]+(0.100404710733)*x[1]**o+(0.0757471444158)*x[1]
        s[0,0,2,1]=(0.865434306062)*x[0]**o+(0.411158028281)*x[0]+(0.717843255665)*x[1]**o+(-0.742991318999)*x[1]
        s[0,1,0,0]=(-0.851618984895)*x[0]**o+(0.251966806193)*x[0]+(-0.803642825762)*x[1]**o+(-0.0631169142808)*x[1]
        s[0,1,0,1]=(-0.758981775218)*x[0]**o+(-0.561508941004)*x[0]+(0.577901652618)*x[1]**o+(-0.725243471023)*x[1]
        s[0,1,1,0]=(0.742032672941)*x[0]**o+(-0.0526002854749)*x[0]+(0.252153928183)*x[1]**o+(-0.388978984633)*x[1]
        s[0,1,1,1]=(-0.980858344737)*x[0]**o+(-0.871658904716)*x[0]+(0.164033322064)*x[1]**o+(0.678658472838)*x[1]
        s[0,1,2,0]=(0.823816388003)*x[0]**o+(0.274859643706)*x[0]+(0.322475967163)*x[1]**o+(-0.598198426633)*x[1]
        s[0,1,2,1]=(0.481980285537)*x[0]**o+(-0.806234350243)*x[0]+(0.564463264623)*x[1]**o+(-0.548427948656)*x[1]
        s[0,2,0,0]=(0.582989836924)*x[0]**o+(0.293733803309)*x[0]+(0.0589622495939)*x[1]**o+(-0.661119752655)*x[1]
        s[0,2,0,1]=(0.675133991144)*x[0]**o+(-0.964990891523)*x[0]+(-0.488115773474)*x[1]**o+(-0.601204526426)*x[1]
        s[0,2,1,0]=(-0.504646764871)*x[0]**o+(-0.496630864273)*x[0]+(-0.533328001228)*x[1]**o+(0.666398746532)*x[1]
        s[0,2,1,1]=(-0.947971259712)*x[0]**o+(0.134598721868)*x[0]+(0.217693108013)*x[1]**o+(0.740476931936)*x[1]
        s[0,2,2,0]=(-0.117189923757)*x[0]**o+(0.433626806763)*x[0]+(-0.746094327712)*x[1]**o+(0.179831846248)*x[1]
        s[0,2,2,1]=(-0.579577944086)*x[0]**o+(0.852457625123)*x[0]+(0.518004334819)*x[1]**o+(0.479873938319)*x[1]
        s[0,3,0,0]=(-0.289116748576)*x[0]**o+(-0.828870395045)*x[0]+(-0.315720408924)*x[1]**o+(0.472146367463)*x[1]
        s[0,3,0,1]=(0.0109829201758)*x[0]**o+(-0.750333402059)*x[0]+(-0.0844671059504)*x[1]**o+(-0.166689650524)*x[1]
        s[0,3,1,0]=(-0.746333456886)*x[0]**o+(0.974025974133)*x[0]+(-0.670772415415)*x[1]**o+(-0.288183829047)*x[1]
        s[0,3,1,1]=(0.999847690339)*x[0]**o+(0.927687112542)*x[0]+(-0.682088385211)*x[1]**o+(-0.185190802895)*x[1]
        s[0,3,2,0]=(-0.122143422759)*x[0]**o+(-0.483572799559)*x[0]+(-0.421650699123)*x[1]**o+(0.751112515854)*x[1]
        s[0,3,2,1]=(0.347046010599)*x[0]**o+(0.86478919879)*x[0]+(-0.376342341601)*x[1]**o+(0.254005796044)*x[1]
        s[0,4,0,0]=(-0.94308296157)*x[0]**o+(-0.278466613273)*x[0]+(0.429411202974)*x[1]**o+(-0.686514518522)*x[1]
        s[0,4,0,1]=(-0.591206969855)*x[0]**o+(0.625779617413)*x[0]+(-0.359728224291)*x[1]**o+(-0.468411179387)*x[1]
        s[0,4,1,0]=(0.871007238866)*x[0]**o+(0.0640301246531)*x[0]+(0.241526675204)*x[1]**o+(-0.198284928754)*x[1]
        s[0,4,1,1]=(0.807190539194)*x[0]**o+(-0.394442569228)*x[0]+(0.137420197381)*x[1]**o+(-0.840168205779)*x[1]
        s[0,4,2,0]=(-0.895780376203)*x[0]**o+(0.489426109167)*x[0]+(0.104706061393)*x[1]**o+(0.580327421236)*x[1]
        s[0,4,2,1]=(0.866870983947)*x[0]**o+(-0.634848129356)*x[0]+(-0.324429185788)*x[1]**o+(-0.902930131617)*x[1]
        s[1,0,0,0]=(-0.435460582678)*x[0]**o+(-0.513879950465)*x[0]+(-0.12284906436)*x[1]**o+(-0.189956950806)*x[1]
        s[1,0,0,1]=(-0.10596403879)*x[0]**o+(0.584299361023)*x[0]+(0.643290036141)*x[1]**o+(0.410255914402)*x[1]
        s[1,0,1,0]=(0.596321097839)*x[0]**o+(0.404038905356)*x[0]+(-0.887663971354)*x[1]**o+(-0.602426219327)*x[1]
        s[1,0,1,1]=(-0.749976485032)*x[0]**o+(0.624337802476)*x[0]+(0.230489334804)*x[1]**o+(-0.169509371147)*x[1]
        s[1,0,2,0]=(-0.338056360533)*x[0]**o+(-0.334004127698)*x[0]+(-0.378429755542)*x[1]**o+(0.656144700111)*x[1]
        s[1,0,2,1]=(0.360646534328)*x[0]**o+(-0.823927735389)*x[0]+(0.758795958018)*x[1]**o+(0.99470110659)*x[1]
        s[1,1,0,0]=(0.831774733955)*x[0]**o+(-0.161890693722)*x[0]+(-0.211705245129)*x[1]**o+(-0.231627795548)*x[1]
        s[1,1,0,1]=(0.994381914656)*x[0]**o+(0.483843191342)*x[0]+(0.703906230692)*x[1]**o+(0.516762394585)*x[1]
        s[1,1,1,0]=(-0.269625525789)*x[0]**o+(0.804559850277)*x[0]+(0.645074410485)*x[1]**o+(0.93472757775)*x[1]
        s[1,1,1,1]=(0.887430533737)*x[0]**o+(0.248852072717)*x[0]+(0.902457877398)*x[1]**o+(-0.749542407535)*x[1]
        s[1,1,2,0]=(0.999384101785)*x[0]**o+(-0.0730419780583)*x[0]+(0.0514349007487)*x[1]**o+(0.151139683749)*x[1]
        s[1,1,2,1]=(0.707403402546)*x[0]**o+(0.421383069726)*x[0]+(-0.713007416618)*x[1]**o+(0.268783526448)*x[1]
        s[1,2,0,0]=(-0.411586906507)*x[0]**o+(0.941961720922)*x[0]+(0.913012821339)*x[1]**o+(-0.296923947382)*x[1]
        s[1,2,0,1]=(-0.520356480672)*x[0]**o+(0.530942059766)*x[0]+(0.0664413255121)*x[1]**o+(0.910541121569)*x[1]
        s[1,2,1,0]=(0.546406246315)*x[0]**o+(0.236311366799)*x[0]+(0.876031584809)*x[1]**o+(0.68767015821)*x[1]
        s[1,2,1,1]=(0.408868402043)*x[0]**o+(0.0910602281951)*x[0]+(0.717385383734)*x[1]**o+(0.799924371281)*x[1]
        s[1,2,2,0]=(-0.00645488169841)*x[0]**o+(-0.193747155577)*x[0]+(0.0532024922569)*x[1]**o+(0.0346603274986)*x[1]
        s[1,2,2,1]=(-0.469056034545)*x[0]**o+(0.221139250978)*x[0]+(0.694972877007)*x[1]**o+(-0.390789722006)*x[1]
        s[1,3,0,0]=(-0.592518388291)*x[0]**o+(0.68133290065)*x[0]+(0.256304394552)*x[1]**o+(0.432327297684)*x[1]
        s[1,3,0,1]=(-0.00024413313477)*x[0]**o+(-0.233897782087)*x[0]+(-0.48099566426)*x[1]**o+(-0.0780681896879)*x[1]
        s[1,3,1,0]=(-0.748402672608)*x[0]**o+(0.663174776498)*x[0]+(0.0516063977957)*x[1]**o+(0.707459316309)*x[1]
        s[1,3,1,1]=(0.856685363283)*x[0]**o+(-0.975581237163)*x[0]+(-0.808077402325)*x[1]**o+(-0.792869908311)*x[1]
        s[1,3,2,0]=(-0.608434761452)*x[0]**o+(-0.538084150506)*x[0]+(0.963534643788)*x[1]**o+(-0.542477024165)*x[1]
        s[1,3,2,1]=(0.0321829705121)*x[0]**o+(-0.646103011762)*x[0]+(-0.0198446078784)*x[1]**o+(0.919550816734)*x[1]
        s[1,4,0,0]=(0.222383667675)*x[0]**o+(-0.687847703153)*x[0]+(-0.59651951219)*x[1]**o+(0.81129360922)*x[1]
        s[1,4,0,1]=(-0.808175325395)*x[0]**o+(0.217063469977)*x[0]+(-0.26488805688)*x[1]**o+(0.0575808940543)*x[1]
        s[1,4,1,0]=(-0.839796412974)*x[0]**o+(-0.42930678659)*x[0]+(0.856238206181)*x[1]**o+(0.0618344272746)*x[1]
        s[1,4,1,1]=(-0.701180815436)*x[0]**o+(-0.87848285733)*x[0]+(0.809946317726)*x[1]**o+(0.733322831203)*x[1]
        s[1,4,2,0]=(-0.00706224969808)*x[0]**o+(0.164085065344)*x[0]+(0.651640705138)*x[1]**o+(-0.947949213151)*x[1]
        s[1,4,2,1]=(-0.55255682366)*x[0]**o+(-3.61985980193e-05)*x[0]+(-0.998780219628)*x[1]**o+(0.68880003343)*x[1]
        s[2,0,0,0]=(0.923223655511)*x[0]**o+(0.320103985146)*x[0]+(0.999706876829)*x[1]**o+(-0.920680964711)*x[1]
        s[2,0,0,1]=(-0.145894916434)*x[0]**o+(0.509339634387)*x[0]+(-0.877941500009)*x[1]**o+(-0.792912551717)*x[1]
        s[2,0,1,0]=(0.736573167219)*x[0]**o+(0.744999602293)*x[0]+(-0.446569219347)*x[1]**o+(0.700985615215)*x[1]
        s[2,0,1,1]=(-0.390389830532)*x[0]**o+(-0.6656221359)*x[0]+(-0.27554460662)*x[1]**o+(-0.458879579705)*x[1]
        s[2,0,2,0]=(-0.673016930466)*x[0]**o+(0.610425320745)*x[0]+(-0.948708604918)*x[1]**o+(0.44912430221)*x[1]
        s[2,0,2,1]=(0.596963521799)*x[0]**o+(-0.990485203537)*x[0]+(0.164522035463)*x[1]**o+(0.539120075165)*x[1]
        s[2,1,0,0]=(-0.932518355596)*x[0]**o+(-0.590131952474)*x[0]+(-0.533511934929)*x[1]**o+(0.0907025905348)*x[1]
        s[2,1,0,1]=(0.865266539976)*x[0]**o+(0.876881424586)*x[0]+(-0.0288484839401)*x[1]**o+(-0.807305428358)*x[1]
        s[2,1,1,0]=(-0.763528720307)*x[0]**o+(-0.0232723097938)*x[0]+(-0.525533773303)*x[1]**o+(0.806025194949)*x[1]
        s[2,1,1,1]=(-0.696217606839)*x[0]**o+(-0.14983123685)*x[0]+(-0.783578700293)*x[1]**o+(0.447836291404)*x[1]
        s[2,1,2,0]=(0.0948683273761)*x[0]**o+(-0.721171113714)*x[0]+(0.701343762426)*x[1]**o+(0.0101281640664)*x[1]
        s[2,1,2,1]=(0.562629645787)*x[0]**o+(-0.763724767765)*x[0]+(-0.449544469039)*x[1]**o+(0.532094652638)*x[1]
        s[2,2,0,0]=(0.841490428822)*x[0]**o+(0.681330810792)*x[0]+(0.668346714262)*x[1]**o+(-0.0336199322961)*x[1]
        s[2,2,0,1]=(0.771997063856)*x[0]**o+(-0.914459974236)*x[0]+(0.129697241086)*x[1]**o+(0.183872706861)*x[1]
        s[2,2,1,0]=(0.447126530052)*x[0]**o+(0.195523018871)*x[0]+(-0.607523037429)*x[1]**o+(0.0880216010007)*x[1]
        s[2,2,1,1]=(0.76876849616)*x[0]**o+(-0.918037921936)*x[0]+(0.65247942976)*x[1]**o+(-0.462180182464)*x[1]
        s[2,2,2,0]=(0.728080043749)*x[0]**o+(-0.716254868732)*x[0]+(0.374917205524)*x[1]**o+(0.311620466651)*x[1]
        s[2,2,2,1]=(0.574841312542)*x[0]**o+(-0.734971890822)*x[0]+(-0.761677234234)*x[1]**o+(0.723259862313)*x[1]
        s[2,3,0,0]=(0.270456229657)*x[0]**o+(-0.58716739336)*x[0]+(0.232577840441)*x[1]**o+(0.747330716699)*x[1]
        s[2,3,0,1]=(-0.820314477652)*x[0]**o+(-0.719999661485)*x[0]+(0.79996006468)*x[1]**o+(-0.831712763153)*x[1]
        s[2,3,1,0]=(0.847137634307)*x[0]**o+(-0.357186776539)*x[0]+(0.09775554965)*x[1]**o+(-0.327915828517)*x[1]
        s[2,3,1,1]=(0.67586077388)*x[0]**o+(0.0335299900089)*x[0]+(-0.179177480197)*x[1]**o+(0.378749816102)*x[1]
        s[2,3,2,0]=(-0.223679858716)*x[0]**o+(-0.844559556039)*x[0]+(0.255235296117)*x[1]**o+(-0.80166933466)*x[1]
        s[2,3,2,1]=(0.668162512403)*x[0]**o+(0.496871152853)*x[0]+(0.745978534888)*x[1]**o+(0.799256534355)*x[1]
        s[2,4,0,0]=(0.363336239392)*x[0]**o+(0.22196304418)*x[0]+(0.875013792472)*x[1]**o+(0.299160735097)*x[1]
        s[2,4,0,1]=(-0.77295625153)*x[0]**o+(-0.109539742009)*x[0]+(0.42425251466)*x[1]**o+(0.857114441732)*x[1]
        s[2,4,1,0]=(-0.453314408352)*x[0]**o+(0.263814392877)*x[0]+(-0.725268476809)*x[1]**o+(-0.347648385435)*x[1]
        s[2,4,1,1]=(-0.35138702248)*x[0]**o+(0.0172444410003)*x[0]+(0.705499375267)*x[1]**o+(0.989996399519)*x[1]
        s[2,4,2,0]=(-0.151100098883)*x[0]**o+(0.0561722886965)*x[0]+(0.701976335813)*x[1]**o+(-0.318092303985)*x[1]
        s[2,4,2,1]=(0.936223946358)*x[0]**o+(0.525004572217)*x[0]+(-0.112487209398)*x[1]**o+(0.196069142522)*x[1]
        s[3,0,0,0]=(-0.296095919553)*x[0]**o+(-0.607841982284)*x[0]+(0.654302153755)*x[1]**o+(0.0410915588113)*x[1]
        s[3,0,0,1]=(0.778444828767)*x[0]**o+(-0.148613776811)*x[0]+(0.911906751967)*x[1]**o+(-0.72527304075)*x[1]
        s[3,0,1,0]=(-0.964041638986)*x[0]**o+(-0.444730606402)*x[0]+(0.856465666005)*x[1]**o+(0.497912010281)*x[1]
        s[3,0,1,1]=(0.0278725007089)*x[0]**o+(-0.102825566476)*x[0]+(-0.925031761314)*x[1]**o+(-0.99399390758)*x[1]
        s[3,0,2,0]=(-0.750580200302)*x[0]**o+(0.792361919091)*x[0]+(0.283376094631)*x[1]**o+(0.210963254645)*x[1]
        s[3,0,2,1]=(-0.202368637069)*x[0]**o+(0.405267270651)*x[0]+(-0.199978746176)*x[1]**o+(0.57738788439)*x[1]
        s[3,1,0,0]=(0.82614638217)*x[0]**o+(-0.342667586691)*x[0]+(0.608532783812)*x[1]**o+(0.768159873725)*x[1]
        s[3,1,0,1]=(0.753566806412)*x[0]**o+(-0.251587300566)*x[0]+(-0.340721126292)*x[1]**o+(-0.757484954136)*x[1]
        s[3,1,1,0]=(0.218156803727)*x[0]**o+(0.527606199926)*x[0]+(-0.212766545043)*x[1]**o+(0.363494347038)*x[1]
        s[3,1,1,1]=(-0.419579821731)*x[0]**o+(-0.848758805521)*x[0]+(0.127749661772)*x[1]**o+(-0.674368477333)*x[1]
        s[3,1,2,0]=(-0.0378627520537)*x[0]**o+(-0.373536670854)*x[0]+(-0.473995670566)*x[1]**o+(0.817069259381)*x[1]
        s[3,1,2,1]=(0.0637470079874)*x[0]**o+(0.902092510251)*x[0]+(0.613777956237)*x[1]**o+(0.984903229699)*x[1]
        s[3,2,0,0]=(-0.550317501871)*x[0]**o+(0.387530710046)*x[0]+(0.51892851937)*x[1]**o+(0.830001465659)*x[1]
        s[3,2,0,1]=(0.159293029872)*x[0]**o+(0.0752213137273)*x[0]+(0.828173132448)*x[1]**o+(0.526955740589)*x[1]
        s[3,2,1,0]=(-0.439770798894)*x[0]**o+(-0.0441574714373)*x[0]+(0.674623413183)*x[1]**o+(-0.958611493787)*x[1]
        s[3,2,1,1]=(0.739682304539)*x[0]**o+(0.734753273429)*x[0]+(-0.20502427226)*x[1]**o+(-0.158955519702)*x[1]
        s[3,2,2,0]=(-0.940998669744)*x[0]**o+(-0.41965535276)*x[0]+(0.893147081307)*x[1]**o+(-0.225209541903)*x[1]
        s[3,2,2,1]=(0.389223021003)*x[0]**o+(-0.483125135422)*x[0]+(0.777672638794)*x[1]**o+(0.839086315551)*x[1]
        s[3,3,0,0]=(-0.320672758354)*x[0]**o+(0.0620118134094)*x[0]+(-0.680049325665)*x[1]**o+(0.411421232123)*x[1]
        s[3,3,0,1]=(-0.0893153008416)*x[0]**o+(0.213720188183)*x[0]+(-0.34570418095)*x[1]**o+(0.625048919615)*x[1]
        s[3,3,1,0]=(0.0158017598762)*x[0]**o+(0.605723583993)*x[0]+(0.739309396699)*x[1]**o+(-0.636482587738)*x[1]
        s[3,3,1,1]=(0.910657933249)*x[0]**o+(-0.441319152646)*x[0]+(0.23268716261)*x[1]**o+(0.24329792568)*x[1]
        s[3,3,2,0]=(0.902717620152)*x[0]**o+(-0.549313425275)*x[0]+(0.464933158282)*x[1]**o+(-0.972615897723)*x[1]
        s[3,3,2,1]=(0.0282461432227)*x[0]**o+(0.761575417375)*x[0]+(0.569262249337)*x[1]**o+(0.828621514356)*x[1]
        s[3,4,0,0]=(0.399367935866)*x[0]**o+(-0.82801744208)*x[0]+(-0.0726407406113)*x[1]**o+(0.530343378961)*x[1]
        s[3,4,0,1]=(0.281064774416)*x[0]**o+(0.338573820292)*x[0]+(0.138863275341)*x[1]**o+(0.148411768993)*x[1]
        s[3,4,1,0]=(-0.686579931373)*x[0]**o+(0.639562092961)*x[0]+(-0.0380362333106)*x[1]**o+(-0.467126934241)*x[1]
        s[3,4,1,1]=(0.611069057705)*x[0]**o+(0.448509606798)*x[0]+(0.663215146298)*x[1]**o+(-0.571323998876)*x[1]
        s[3,4,2,0]=(0.21998109831)*x[0]**o+(-0.737645914394)*x[0]+(0.669008118713)*x[1]**o+(-0.291237893231)*x[1]
        s[3,4,2,1]=(0.595080778165)*x[0]**o+(0.988680777579)*x[0]+(-0.361694479664)*x[1]**o+(0.210229549739)*x[1]
        ref[0,0,0,0]=(0.771805529053)*x_ref[0]**o+(-0.00794071414241)*x_ref[0]+(-0.546566317942)*x_ref[1]**o+(0.209469633398)*x_ref[1]
        ref[0,0,0,1]=(-0.0138300734518)*x_ref[0]**o+(0.783507392816)*x_ref[0]+(0.637453082951)*x_ref[1]**o+(-0.0262186183303)*x_ref[1]
        ref[0,0,1,0]=(-0.622230122991)*x_ref[0]**o+(0.760867623401)*x_ref[0]+(0.340084769283)*x_ref[1]**o+(0.493207077521)*x_ref[1]
        ref[0,0,1,1]=(0.772195055793)*x_ref[0]**o+(-0.481066683803)*x_ref[0]+(0.198371969404)*x_ref[1]**o+(0.841664205482)*x_ref[1]
        ref[0,0,2,0]=(0.858213005013)*x_ref[0]**o+(0.819391260667)*x_ref[0]+(0.100404710733)*x_ref[1]**o+(0.0757471444158)*x_ref[1]
        ref[0,0,2,1]=(0.865434306062)*x_ref[0]**o+(0.411158028281)*x_ref[0]+(0.717843255665)*x_ref[1]**o+(-0.742991318999)*x_ref[1]
        ref[0,1,0,0]=(-0.851618984895)*x_ref[0]**o+(0.251966806193)*x_ref[0]+(-0.803642825762)*x_ref[1]**o+(-0.0631169142808)*x_ref[1]
        ref[0,1,0,1]=(-0.758981775218)*x_ref[0]**o+(-0.561508941004)*x_ref[0]+(0.577901652618)*x_ref[1]**o+(-0.725243471023)*x_ref[1]
        ref[0,1,1,0]=(0.742032672941)*x_ref[0]**o+(-0.0526002854749)*x_ref[0]+(0.252153928183)*x_ref[1]**o+(-0.388978984633)*x_ref[1]
        ref[0,1,1,1]=(-0.980858344737)*x_ref[0]**o+(-0.871658904716)*x_ref[0]+(0.164033322064)*x_ref[1]**o+(0.678658472838)*x_ref[1]
        ref[0,1,2,0]=(0.823816388003)*x_ref[0]**o+(0.274859643706)*x_ref[0]+(0.322475967163)*x_ref[1]**o+(-0.598198426633)*x_ref[1]
        ref[0,1,2,1]=(0.481980285537)*x_ref[0]**o+(-0.806234350243)*x_ref[0]+(0.564463264623)*x_ref[1]**o+(-0.548427948656)*x_ref[1]
        ref[0,2,0,0]=(0.582989836924)*x_ref[0]**o+(0.293733803309)*x_ref[0]+(0.0589622495939)*x_ref[1]**o+(-0.661119752655)*x_ref[1]
        ref[0,2,0,1]=(0.675133991144)*x_ref[0]**o+(-0.964990891523)*x_ref[0]+(-0.488115773474)*x_ref[1]**o+(-0.601204526426)*x_ref[1]
        ref[0,2,1,0]=(-0.504646764871)*x_ref[0]**o+(-0.496630864273)*x_ref[0]+(-0.533328001228)*x_ref[1]**o+(0.666398746532)*x_ref[1]
        ref[0,2,1,1]=(-0.947971259712)*x_ref[0]**o+(0.134598721868)*x_ref[0]+(0.217693108013)*x_ref[1]**o+(0.740476931936)*x_ref[1]
        ref[0,2,2,0]=(-0.117189923757)*x_ref[0]**o+(0.433626806763)*x_ref[0]+(-0.746094327712)*x_ref[1]**o+(0.179831846248)*x_ref[1]
        ref[0,2,2,1]=(-0.579577944086)*x_ref[0]**o+(0.852457625123)*x_ref[0]+(0.518004334819)*x_ref[1]**o+(0.479873938319)*x_ref[1]
        ref[0,3,0,0]=(-0.289116748576)*x_ref[0]**o+(-0.828870395045)*x_ref[0]+(-0.315720408924)*x_ref[1]**o+(0.472146367463)*x_ref[1]
        ref[0,3,0,1]=(0.0109829201758)*x_ref[0]**o+(-0.750333402059)*x_ref[0]+(-0.0844671059504)*x_ref[1]**o+(-0.166689650524)*x_ref[1]
        ref[0,3,1,0]=(-0.746333456886)*x_ref[0]**o+(0.974025974133)*x_ref[0]+(-0.670772415415)*x_ref[1]**o+(-0.288183829047)*x_ref[1]
        ref[0,3,1,1]=(0.999847690339)*x_ref[0]**o+(0.927687112542)*x_ref[0]+(-0.682088385211)*x_ref[1]**o+(-0.185190802895)*x_ref[1]
        ref[0,3,2,0]=(-0.122143422759)*x_ref[0]**o+(-0.483572799559)*x_ref[0]+(-0.421650699123)*x_ref[1]**o+(0.751112515854)*x_ref[1]
        ref[0,3,2,1]=(0.347046010599)*x_ref[0]**o+(0.86478919879)*x_ref[0]+(-0.376342341601)*x_ref[1]**o+(0.254005796044)*x_ref[1]
        ref[0,4,0,0]=(-0.94308296157)*x_ref[0]**o+(-0.278466613273)*x_ref[0]+(0.429411202974)*x_ref[1]**o+(-0.686514518522)*x_ref[1]
        ref[0,4,0,1]=(-0.591206969855)*x_ref[0]**o+(0.625779617413)*x_ref[0]+(-0.359728224291)*x_ref[1]**o+(-0.468411179387)*x_ref[1]
        ref[0,4,1,0]=(0.871007238866)*x_ref[0]**o+(0.0640301246531)*x_ref[0]+(0.241526675204)*x_ref[1]**o+(-0.198284928754)*x_ref[1]
        ref[0,4,1,1]=(0.807190539194)*x_ref[0]**o+(-0.394442569228)*x_ref[0]+(0.137420197381)*x_ref[1]**o+(-0.840168205779)*x_ref[1]
        ref[0,4,2,0]=(-0.895780376203)*x_ref[0]**o+(0.489426109167)*x_ref[0]+(0.104706061393)*x_ref[1]**o+(0.580327421236)*x_ref[1]
        ref[0,4,2,1]=(0.866870983947)*x_ref[0]**o+(-0.634848129356)*x_ref[0]+(-0.324429185788)*x_ref[1]**o+(-0.902930131617)*x_ref[1]
        ref[1,0,0,0]=(-0.435460582678)*x_ref[0]**o+(-0.513879950465)*x_ref[0]+(-0.12284906436)*x_ref[1]**o+(-0.189956950806)*x_ref[1]
        ref[1,0,0,1]=(-0.10596403879)*x_ref[0]**o+(0.584299361023)*x_ref[0]+(0.643290036141)*x_ref[1]**o+(0.410255914402)*x_ref[1]
        ref[1,0,1,0]=(0.596321097839)*x_ref[0]**o+(0.404038905356)*x_ref[0]+(-0.887663971354)*x_ref[1]**o+(-0.602426219327)*x_ref[1]
        ref[1,0,1,1]=(-0.749976485032)*x_ref[0]**o+(0.624337802476)*x_ref[0]+(0.230489334804)*x_ref[1]**o+(-0.169509371147)*x_ref[1]
        ref[1,0,2,0]=(-0.338056360533)*x_ref[0]**o+(-0.334004127698)*x_ref[0]+(-0.378429755542)*x_ref[1]**o+(0.656144700111)*x_ref[1]
        ref[1,0,2,1]=(0.360646534328)*x_ref[0]**o+(-0.823927735389)*x_ref[0]+(0.758795958018)*x_ref[1]**o+(0.99470110659)*x_ref[1]
        ref[1,1,0,0]=(0.831774733955)*x_ref[0]**o+(-0.161890693722)*x_ref[0]+(-0.211705245129)*x_ref[1]**o+(-0.231627795548)*x_ref[1]
        ref[1,1,0,1]=(0.994381914656)*x_ref[0]**o+(0.483843191342)*x_ref[0]+(0.703906230692)*x_ref[1]**o+(0.516762394585)*x_ref[1]
        ref[1,1,1,0]=(-0.269625525789)*x_ref[0]**o+(0.804559850277)*x_ref[0]+(0.645074410485)*x_ref[1]**o+(0.93472757775)*x_ref[1]
        ref[1,1,1,1]=(0.887430533737)*x_ref[0]**o+(0.248852072717)*x_ref[0]+(0.902457877398)*x_ref[1]**o+(-0.749542407535)*x_ref[1]
        ref[1,1,2,0]=(0.999384101785)*x_ref[0]**o+(-0.0730419780583)*x_ref[0]+(0.0514349007487)*x_ref[1]**o+(0.151139683749)*x_ref[1]
        ref[1,1,2,1]=(0.707403402546)*x_ref[0]**o+(0.421383069726)*x_ref[0]+(-0.713007416618)*x_ref[1]**o+(0.268783526448)*x_ref[1]
        ref[1,2,0,0]=(-0.411586906507)*x_ref[0]**o+(0.941961720922)*x_ref[0]+(0.913012821339)*x_ref[1]**o+(-0.296923947382)*x_ref[1]
        ref[1,2,0,1]=(-0.520356480672)*x_ref[0]**o+(0.530942059766)*x_ref[0]+(0.0664413255121)*x_ref[1]**o+(0.910541121569)*x_ref[1]
        ref[1,2,1,0]=(0.546406246315)*x_ref[0]**o+(0.236311366799)*x_ref[0]+(0.876031584809)*x_ref[1]**o+(0.68767015821)*x_ref[1]
        ref[1,2,1,1]=(0.408868402043)*x_ref[0]**o+(0.0910602281951)*x_ref[0]+(0.717385383734)*x_ref[1]**o+(0.799924371281)*x_ref[1]
        ref[1,2,2,0]=(-0.00645488169841)*x_ref[0]**o+(-0.193747155577)*x_ref[0]+(0.0532024922569)*x_ref[1]**o+(0.0346603274986)*x_ref[1]
        ref[1,2,2,1]=(-0.469056034545)*x_ref[0]**o+(0.221139250978)*x_ref[0]+(0.694972877007)*x_ref[1]**o+(-0.390789722006)*x_ref[1]
        ref[1,3,0,0]=(-0.592518388291)*x_ref[0]**o+(0.68133290065)*x_ref[0]+(0.256304394552)*x_ref[1]**o+(0.432327297684)*x_ref[1]
        ref[1,3,0,1]=(-0.00024413313477)*x_ref[0]**o+(-0.233897782087)*x_ref[0]+(-0.48099566426)*x_ref[1]**o+(-0.0780681896879)*x_ref[1]
        ref[1,3,1,0]=(-0.748402672608)*x_ref[0]**o+(0.663174776498)*x_ref[0]+(0.0516063977957)*x_ref[1]**o+(0.707459316309)*x_ref[1]
        ref[1,3,1,1]=(0.856685363283)*x_ref[0]**o+(-0.975581237163)*x_ref[0]+(-0.808077402325)*x_ref[1]**o+(-0.792869908311)*x_ref[1]
        ref[1,3,2,0]=(-0.608434761452)*x_ref[0]**o+(-0.538084150506)*x_ref[0]+(0.963534643788)*x_ref[1]**o+(-0.542477024165)*x_ref[1]
        ref[1,3,2,1]=(0.0321829705121)*x_ref[0]**o+(-0.646103011762)*x_ref[0]+(-0.0198446078784)*x_ref[1]**o+(0.919550816734)*x_ref[1]
        ref[1,4,0,0]=(0.222383667675)*x_ref[0]**o+(-0.687847703153)*x_ref[0]+(-0.59651951219)*x_ref[1]**o+(0.81129360922)*x_ref[1]
        ref[1,4,0,1]=(-0.808175325395)*x_ref[0]**o+(0.217063469977)*x_ref[0]+(-0.26488805688)*x_ref[1]**o+(0.0575808940543)*x_ref[1]
        ref[1,4,1,0]=(-0.839796412974)*x_ref[0]**o+(-0.42930678659)*x_ref[0]+(0.856238206181)*x_ref[1]**o+(0.0618344272746)*x_ref[1]
        ref[1,4,1,1]=(-0.701180815436)*x_ref[0]**o+(-0.87848285733)*x_ref[0]+(0.809946317726)*x_ref[1]**o+(0.733322831203)*x_ref[1]
        ref[1,4,2,0]=(-0.00706224969808)*x_ref[0]**o+(0.164085065344)*x_ref[0]+(0.651640705138)*x_ref[1]**o+(-0.947949213151)*x_ref[1]
        ref[1,4,2,1]=(-0.55255682366)*x_ref[0]**o+(-3.61985980193e-05)*x_ref[0]+(-0.998780219628)*x_ref[1]**o+(0.68880003343)*x_ref[1]
        ref[2,0,0,0]=(0.923223655511)*x_ref[0]**o+(0.320103985146)*x_ref[0]+(0.999706876829)*x_ref[1]**o+(-0.920680964711)*x_ref[1]
        ref[2,0,0,1]=(-0.145894916434)*x_ref[0]**o+(0.509339634387)*x_ref[0]+(-0.877941500009)*x_ref[1]**o+(-0.792912551717)*x_ref[1]
        ref[2,0,1,0]=(0.736573167219)*x_ref[0]**o+(0.744999602293)*x_ref[0]+(-0.446569219347)*x_ref[1]**o+(0.700985615215)*x_ref[1]
        ref[2,0,1,1]=(-0.390389830532)*x_ref[0]**o+(-0.6656221359)*x_ref[0]+(-0.27554460662)*x_ref[1]**o+(-0.458879579705)*x_ref[1]
        ref[2,0,2,0]=(-0.673016930466)*x_ref[0]**o+(0.610425320745)*x_ref[0]+(-0.948708604918)*x_ref[1]**o+(0.44912430221)*x_ref[1]
        ref[2,0,2,1]=(0.596963521799)*x_ref[0]**o+(-0.990485203537)*x_ref[0]+(0.164522035463)*x_ref[1]**o+(0.539120075165)*x_ref[1]
        ref[2,1,0,0]=(-0.932518355596)*x_ref[0]**o+(-0.590131952474)*x_ref[0]+(-0.533511934929)*x_ref[1]**o+(0.0907025905348)*x_ref[1]
        ref[2,1,0,1]=(0.865266539976)*x_ref[0]**o+(0.876881424586)*x_ref[0]+(-0.0288484839401)*x_ref[1]**o+(-0.807305428358)*x_ref[1]
        ref[2,1,1,0]=(-0.763528720307)*x_ref[0]**o+(-0.0232723097938)*x_ref[0]+(-0.525533773303)*x_ref[1]**o+(0.806025194949)*x_ref[1]
        ref[2,1,1,1]=(-0.696217606839)*x_ref[0]**o+(-0.14983123685)*x_ref[0]+(-0.783578700293)*x_ref[1]**o+(0.447836291404)*x_ref[1]
        ref[2,1,2,0]=(0.0948683273761)*x_ref[0]**o+(-0.721171113714)*x_ref[0]+(0.701343762426)*x_ref[1]**o+(0.0101281640664)*x_ref[1]
        ref[2,1,2,1]=(0.562629645787)*x_ref[0]**o+(-0.763724767765)*x_ref[0]+(-0.449544469039)*x_ref[1]**o+(0.532094652638)*x_ref[1]
        ref[2,2,0,0]=(0.841490428822)*x_ref[0]**o+(0.681330810792)*x_ref[0]+(0.668346714262)*x_ref[1]**o+(-0.0336199322961)*x_ref[1]
        ref[2,2,0,1]=(0.771997063856)*x_ref[0]**o+(-0.914459974236)*x_ref[0]+(0.129697241086)*x_ref[1]**o+(0.183872706861)*x_ref[1]
        ref[2,2,1,0]=(0.447126530052)*x_ref[0]**o+(0.195523018871)*x_ref[0]+(-0.607523037429)*x_ref[1]**o+(0.0880216010007)*x_ref[1]
        ref[2,2,1,1]=(0.76876849616)*x_ref[0]**o+(-0.918037921936)*x_ref[0]+(0.65247942976)*x_ref[1]**o+(-0.462180182464)*x_ref[1]
        ref[2,2,2,0]=(0.728080043749)*x_ref[0]**o+(-0.716254868732)*x_ref[0]+(0.374917205524)*x_ref[1]**o+(0.311620466651)*x_ref[1]
        ref[2,2,2,1]=(0.574841312542)*x_ref[0]**o+(-0.734971890822)*x_ref[0]+(-0.761677234234)*x_ref[1]**o+(0.723259862313)*x_ref[1]
        ref[2,3,0,0]=(0.270456229657)*x_ref[0]**o+(-0.58716739336)*x_ref[0]+(0.232577840441)*x_ref[1]**o+(0.747330716699)*x_ref[1]
        ref[2,3,0,1]=(-0.820314477652)*x_ref[0]**o+(-0.719999661485)*x_ref[0]+(0.79996006468)*x_ref[1]**o+(-0.831712763153)*x_ref[1]
        ref[2,3,1,0]=(0.847137634307)*x_ref[0]**o+(-0.357186776539)*x_ref[0]+(0.09775554965)*x_ref[1]**o+(-0.327915828517)*x_ref[1]
        ref[2,3,1,1]=(0.67586077388)*x_ref[0]**o+(0.0335299900089)*x_ref[0]+(-0.179177480197)*x_ref[1]**o+(0.378749816102)*x_ref[1]
        ref[2,3,2,0]=(-0.223679858716)*x_ref[0]**o+(-0.844559556039)*x_ref[0]+(0.255235296117)*x_ref[1]**o+(-0.80166933466)*x_ref[1]
        ref[2,3,2,1]=(0.668162512403)*x_ref[0]**o+(0.496871152853)*x_ref[0]+(0.745978534888)*x_ref[1]**o+(0.799256534355)*x_ref[1]
        ref[2,4,0,0]=(0.363336239392)*x_ref[0]**o+(0.22196304418)*x_ref[0]+(0.875013792472)*x_ref[1]**o+(0.299160735097)*x_ref[1]
        ref[2,4,0,1]=(-0.77295625153)*x_ref[0]**o+(-0.109539742009)*x_ref[0]+(0.42425251466)*x_ref[1]**o+(0.857114441732)*x_ref[1]
        ref[2,4,1,0]=(-0.453314408352)*x_ref[0]**o+(0.263814392877)*x_ref[0]+(-0.725268476809)*x_ref[1]**o+(-0.347648385435)*x_ref[1]
        ref[2,4,1,1]=(-0.35138702248)*x_ref[0]**o+(0.0172444410003)*x_ref[0]+(0.705499375267)*x_ref[1]**o+(0.989996399519)*x_ref[1]
        ref[2,4,2,0]=(-0.151100098883)*x_ref[0]**o+(0.0561722886965)*x_ref[0]+(0.701976335813)*x_ref[1]**o+(-0.318092303985)*x_ref[1]
        ref[2,4,2,1]=(0.936223946358)*x_ref[0]**o+(0.525004572217)*x_ref[0]+(-0.112487209398)*x_ref[1]**o+(0.196069142522)*x_ref[1]
        ref[3,0,0,0]=(-0.296095919553)*x_ref[0]**o+(-0.607841982284)*x_ref[0]+(0.654302153755)*x_ref[1]**o+(0.0410915588113)*x_ref[1]
        ref[3,0,0,1]=(0.778444828767)*x_ref[0]**o+(-0.148613776811)*x_ref[0]+(0.911906751967)*x_ref[1]**o+(-0.72527304075)*x_ref[1]
        ref[3,0,1,0]=(-0.964041638986)*x_ref[0]**o+(-0.444730606402)*x_ref[0]+(0.856465666005)*x_ref[1]**o+(0.497912010281)*x_ref[1]
        ref[3,0,1,1]=(0.0278725007089)*x_ref[0]**o+(-0.102825566476)*x_ref[0]+(-0.925031761314)*x_ref[1]**o+(-0.99399390758)*x_ref[1]
        ref[3,0,2,0]=(-0.750580200302)*x_ref[0]**o+(0.792361919091)*x_ref[0]+(0.283376094631)*x_ref[1]**o+(0.210963254645)*x_ref[1]
        ref[3,0,2,1]=(-0.202368637069)*x_ref[0]**o+(0.405267270651)*x_ref[0]+(-0.199978746176)*x_ref[1]**o+(0.57738788439)*x_ref[1]
        ref[3,1,0,0]=(0.82614638217)*x_ref[0]**o+(-0.342667586691)*x_ref[0]+(0.608532783812)*x_ref[1]**o+(0.768159873725)*x_ref[1]
        ref[3,1,0,1]=(0.753566806412)*x_ref[0]**o+(-0.251587300566)*x_ref[0]+(-0.340721126292)*x_ref[1]**o+(-0.757484954136)*x_ref[1]
        ref[3,1,1,0]=(0.218156803727)*x_ref[0]**o+(0.527606199926)*x_ref[0]+(-0.212766545043)*x_ref[1]**o+(0.363494347038)*x_ref[1]
        ref[3,1,1,1]=(-0.419579821731)*x_ref[0]**o+(-0.848758805521)*x_ref[0]+(0.127749661772)*x_ref[1]**o+(-0.674368477333)*x_ref[1]
        ref[3,1,2,0]=(-0.0378627520537)*x_ref[0]**o+(-0.373536670854)*x_ref[0]+(-0.473995670566)*x_ref[1]**o+(0.817069259381)*x_ref[1]
        ref[3,1,2,1]=(0.0637470079874)*x_ref[0]**o+(0.902092510251)*x_ref[0]+(0.613777956237)*x_ref[1]**o+(0.984903229699)*x_ref[1]
        ref[3,2,0,0]=(-0.550317501871)*x_ref[0]**o+(0.387530710046)*x_ref[0]+(0.51892851937)*x_ref[1]**o+(0.830001465659)*x_ref[1]
        ref[3,2,0,1]=(0.159293029872)*x_ref[0]**o+(0.0752213137273)*x_ref[0]+(0.828173132448)*x_ref[1]**o+(0.526955740589)*x_ref[1]
        ref[3,2,1,0]=(-0.439770798894)*x_ref[0]**o+(-0.0441574714373)*x_ref[0]+(0.674623413183)*x_ref[1]**o+(-0.958611493787)*x_ref[1]
        ref[3,2,1,1]=(0.739682304539)*x_ref[0]**o+(0.734753273429)*x_ref[0]+(-0.20502427226)*x_ref[1]**o+(-0.158955519702)*x_ref[1]
        ref[3,2,2,0]=(-0.940998669744)*x_ref[0]**o+(-0.41965535276)*x_ref[0]+(0.893147081307)*x_ref[1]**o+(-0.225209541903)*x_ref[1]
        ref[3,2,2,1]=(0.389223021003)*x_ref[0]**o+(-0.483125135422)*x_ref[0]+(0.777672638794)*x_ref[1]**o+(0.839086315551)*x_ref[1]
        ref[3,3,0,0]=(-0.320672758354)*x_ref[0]**o+(0.0620118134094)*x_ref[0]+(-0.680049325665)*x_ref[1]**o+(0.411421232123)*x_ref[1]
        ref[3,3,0,1]=(-0.0893153008416)*x_ref[0]**o+(0.213720188183)*x_ref[0]+(-0.34570418095)*x_ref[1]**o+(0.625048919615)*x_ref[1]
        ref[3,3,1,0]=(0.0158017598762)*x_ref[0]**o+(0.605723583993)*x_ref[0]+(0.739309396699)*x_ref[1]**o+(-0.636482587738)*x_ref[1]
        ref[3,3,1,1]=(0.910657933249)*x_ref[0]**o+(-0.441319152646)*x_ref[0]+(0.23268716261)*x_ref[1]**o+(0.24329792568)*x_ref[1]
        ref[3,3,2,0]=(0.902717620152)*x_ref[0]**o+(-0.549313425275)*x_ref[0]+(0.464933158282)*x_ref[1]**o+(-0.972615897723)*x_ref[1]
        ref[3,3,2,1]=(0.0282461432227)*x_ref[0]**o+(0.761575417375)*x_ref[0]+(0.569262249337)*x_ref[1]**o+(0.828621514356)*x_ref[1]
        ref[3,4,0,0]=(0.399367935866)*x_ref[0]**o+(-0.82801744208)*x_ref[0]+(-0.0726407406113)*x_ref[1]**o+(0.530343378961)*x_ref[1]
        ref[3,4,0,1]=(0.281064774416)*x_ref[0]**o+(0.338573820292)*x_ref[0]+(0.138863275341)*x_ref[1]**o+(0.148411768993)*x_ref[1]
        ref[3,4,1,0]=(-0.686579931373)*x_ref[0]**o+(0.639562092961)*x_ref[0]+(-0.0380362333106)*x_ref[1]**o+(-0.467126934241)*x_ref[1]
        ref[3,4,1,1]=(0.611069057705)*x_ref[0]**o+(0.448509606798)*x_ref[0]+(0.663215146298)*x_ref[1]**o+(-0.571323998876)*x_ref[1]
        ref[3,4,2,0]=(0.21998109831)*x_ref[0]**o+(-0.737645914394)*x_ref[0]+(0.669008118713)*x_ref[1]**o+(-0.291237893231)*x_ref[1]
        ref[3,4,2,1]=(0.595080778165)*x_ref[0]**o+(0.988680777579)*x_ref[0]+(-0.361694479664)*x_ref[1]**o+(0.210229549739)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.209698190814)*x[0]**o+(0.0202198726379)*x[0]+(-0.776762051242)*x[1]**o+(0.0103026495025)*x[1]+(0.0277886960026)*x[2]**o+(-0.255065901162)*x[2]
        s[0,0,0,1]=(-0.663085958734)*x[0]**o+(0.698525770087)*x[0]+(0.650288968593)*x[1]**o+(0.6069536467)*x[1]+(0.789898121185)*x[2]**o+(-0.13315288412)*x[2]
        s[0,0,1,0]=(-0.0998227591334)*x[0]**o+(0.75314681829)*x[0]+(-0.43575224825)*x[1]**o+(0.649609416117)*x[1]+(0.271394201051)*x[2]**o+(-0.353712050582)*x[2]
        s[0,0,1,1]=(0.205441533003)*x[0]**o+(-0.61122881962)*x[0]+(-0.431370221976)*x[1]**o+(-0.183655447483)*x[1]+(0.116437364187)*x[2]**o+(0.741511658665)*x[2]
        s[0,0,2,0]=(0.516133026558)*x[0]**o+(-0.904105438162)*x[0]+(-0.579514088677)*x[1]**o+(-0.0581332954324)*x[1]+(0.557937742112)*x[2]**o+(0.138639028149)*x[2]
        s[0,0,2,1]=(0.8991763477)*x[0]**o+(-0.792213819942)*x[0]+(0.674755445696)*x[1]**o+(0.483315869458)*x[1]+(-0.0722431463476)*x[2]**o+(0.0270814412191)*x[2]
        s[0,1,0,0]=(0.366564587764)*x[0]**o+(0.928578857146)*x[0]+(0.785092228323)*x[1]**o+(0.860812095832)*x[1]+(-0.603464315283)*x[2]**o+(-0.556636685093)*x[2]
        s[0,1,0,1]=(0.980630839809)*x[0]**o+(-0.255324039441)*x[0]+(0.173630938787)*x[1]**o+(0.110433932158)*x[1]+(-0.392851296251)*x[2]**o+(-0.272438563385)*x[2]
        s[0,1,1,0]=(-0.63489459343)*x[0]**o+(0.0328950965715)*x[0]+(-0.199264976942)*x[1]**o+(0.704589249988)*x[1]+(0.442306934744)*x[2]**o+(-0.531996394873)*x[2]
        s[0,1,1,1]=(-0.110263035707)*x[0]**o+(0.904756528197)*x[0]+(0.876124769497)*x[1]**o+(-0.55469630081)*x[1]+(-0.857675499295)*x[2]**o+(0.662352375455)*x[2]
        s[0,1,2,0]=(-0.383336639984)*x[0]**o+(0.863605840878)*x[0]+(0.742039996033)*x[1]**o+(0.649587811572)*x[1]+(0.0749104131981)*x[2]**o+(-0.943341654469)*x[2]
        s[0,1,2,1]=(0.200285377439)*x[0]**o+(0.148329648537)*x[0]+(-0.0994444073677)*x[1]**o+(-0.535438035359)*x[1]+(0.662913271697)*x[2]**o+(0.112409762773)*x[2]
        s[0,2,0,0]=(0.351533608017)*x[0]**o+(0.72576521785)*x[0]+(0.305528610204)*x[1]**o+(-0.139400610385)*x[1]+(-0.455936488645)*x[2]**o+(-0.992626031371)*x[2]
        s[0,2,0,1]=(0.351745471157)*x[0]**o+(0.749900692382)*x[0]+(0.212677388818)*x[1]**o+(0.449543777767)*x[1]+(0.9265756875)*x[2]**o+(0.0973575097882)*x[2]
        s[0,2,1,0]=(0.308471390044)*x[0]**o+(-0.236492160276)*x[0]+(-0.0132386386917)*x[1]**o+(-0.657824382959)*x[1]+(0.567169125655)*x[2]**o+(-0.0239580661983)*x[2]
        s[0,2,1,1]=(-0.71982992429)*x[0]**o+(-0.518285193827)*x[0]+(0.806477863481)*x[1]**o+(-0.949670246654)*x[1]+(-0.559408139171)*x[2]**o+(-0.740059454923)*x[2]
        s[0,2,2,0]=(-0.795819227823)*x[0]**o+(0.0382042059208)*x[0]+(-0.993434888742)*x[1]**o+(-0.714572217225)*x[1]+(0.826057540197)*x[2]**o+(-0.351385211283)*x[2]
        s[0,2,2,1]=(-0.691428669925)*x[0]**o+(0.795109370028)*x[0]+(-0.845580170561)*x[1]**o+(0.66720897925)*x[1]+(-0.604013104606)*x[2]**o+(-0.56003576527)*x[2]
        s[0,3,0,0]=(-0.864934911539)*x[0]**o+(-0.871210034972)*x[0]+(-0.633807262008)*x[1]**o+(-0.0448147244258)*x[1]+(0.601871519726)*x[2]**o+(-0.423895847468)*x[2]
        s[0,3,0,1]=(0.542426383922)*x[0]**o+(0.648956460184)*x[0]+(0.975476287059)*x[1]**o+(-0.110876177567)*x[1]+(-0.334216196036)*x[2]**o+(-0.696393938466)*x[2]
        s[0,3,1,0]=(-0.897499573749)*x[0]**o+(0.504413212966)*x[0]+(-0.114527615291)*x[1]**o+(0.415575654212)*x[1]+(-0.271870634022)*x[2]**o+(0.698810118326)*x[2]
        s[0,3,1,1]=(0.231758813026)*x[0]**o+(-0.207508138983)*x[0]+(-0.744927387959)*x[1]**o+(0.00489240105965)*x[1]+(0.821414365164)*x[2]**o+(0.701679220215)*x[2]
        s[0,3,2,0]=(-0.135996215059)*x[0]**o+(-0.513005242192)*x[0]+(-0.6068585358)*x[1]**o+(-0.852285533431)*x[1]+(0.115701663464)*x[2]**o+(0.243886826101)*x[2]
        s[0,3,2,1]=(-0.80781983055)*x[0]**o+(0.142781343516)*x[0]+(0.764546896757)*x[1]**o+(0.478359930452)*x[1]+(-0.0458636815218)*x[2]**o+(0.582425549793)*x[2]
        s[0,4,0,0]=(-0.382194450205)*x[0]**o+(-0.209573752081)*x[0]+(0.959742867105)*x[1]**o+(0.962658893536)*x[1]+(0.126101921123)*x[2]**o+(0.980806872196)*x[2]
        s[0,4,0,1]=(0.764116933095)*x[0]**o+(-0.648257603162)*x[0]+(-0.839456477431)*x[1]**o+(-0.972616251638)*x[1]+(-0.815159939908)*x[2]**o+(-0.262639210819)*x[2]
        s[0,4,1,0]=(0.453438243261)*x[0]**o+(0.303833796601)*x[0]+(0.100367120622)*x[1]**o+(-0.0716774769052)*x[1]+(-0.659569961883)*x[2]**o+(-0.290044782372)*x[2]
        s[0,4,1,1]=(-0.0286107263502)*x[0]**o+(-0.596438980382)*x[0]+(0.545044960979)*x[1]**o+(-0.796415090612)*x[1]+(0.206907297358)*x[2]**o+(0.306889564609)*x[2]
        s[0,4,2,0]=(-0.201162762403)*x[0]**o+(-0.981302954607)*x[0]+(-0.201138026976)*x[1]**o+(-0.93628245382)*x[1]+(-0.758945968148)*x[2]**o+(-0.860206802477)*x[2]
        s[0,4,2,1]=(0.251350117686)*x[0]**o+(-0.29118495379)*x[0]+(-0.569363617929)*x[1]**o+(0.465078119323)*x[1]+(0.426184283109)*x[2]**o+(-0.873643118829)*x[2]
        s[1,0,0,0]=(-0.0185687413154)*x[0]**o+(0.411847218239)*x[0]+(0.428984420237)*x[1]**o+(-0.547264849731)*x[1]+(0.486727670471)*x[2]**o+(0.179455122763)*x[2]
        s[1,0,0,1]=(0.518433022846)*x[0]**o+(0.025823449242)*x[0]+(-0.0540511749223)*x[1]**o+(-0.85848829682)*x[1]+(0.158385940315)*x[2]**o+(-0.178171294535)*x[2]
        s[1,0,1,0]=(0.0367258516392)*x[0]**o+(0.545049981245)*x[0]+(-0.579044650548)*x[1]**o+(-0.77725873354)*x[1]+(-0.034299394836)*x[2]**o+(-0.179489216766)*x[2]
        s[1,0,1,1]=(0.274129185725)*x[0]**o+(0.166006734196)*x[0]+(-0.773979014851)*x[1]**o+(-0.519501835507)*x[1]+(-0.63912427991)*x[2]**o+(0.466724690101)*x[2]
        s[1,0,2,0]=(0.482955790279)*x[0]**o+(0.471838320459)*x[0]+(0.617771038211)*x[1]**o+(0.048311205466)*x[1]+(0.162393942754)*x[2]**o+(-0.694439775285)*x[2]
        s[1,0,2,1]=(-0.517879311536)*x[0]**o+(-0.730313900842)*x[0]+(-0.991744866252)*x[1]**o+(-0.849707440389)*x[1]+(-0.83668608481)*x[2]**o+(0.581082133088)*x[2]
        s[1,1,0,0]=(0.217328148299)*x[0]**o+(0.187729052039)*x[0]+(-0.68877270235)*x[1]**o+(-0.86757435657)*x[1]+(-0.427061492994)*x[2]**o+(-0.84402543773)*x[2]
        s[1,1,0,1]=(0.766541346426)*x[0]**o+(0.649768542608)*x[0]+(-0.113463877205)*x[1]**o+(-0.637749133894)*x[1]+(0.809878689222)*x[2]**o+(0.315784984005)*x[2]
        s[1,1,1,0]=(0.94173972573)*x[0]**o+(0.668276294204)*x[0]+(0.427204723683)*x[1]**o+(-0.607750747607)*x[1]+(-0.0376503804716)*x[2]**o+(0.960260009026)*x[2]
        s[1,1,1,1]=(-0.400233578622)*x[0]**o+(-0.923564602772)*x[0]+(0.47175931621)*x[1]**o+(-0.612010762622)*x[1]+(0.609225052578)*x[2]**o+(0.526546928045)*x[2]
        s[1,1,2,0]=(0.377678503521)*x[0]**o+(0.892169209135)*x[0]+(0.39772744849)*x[1]**o+(-0.166966626427)*x[1]+(0.874470437184)*x[2]**o+(0.314522738087)*x[2]
        s[1,1,2,1]=(0.520217518976)*x[0]**o+(0.536152467541)*x[0]+(0.444145502111)*x[1]**o+(-0.489860603506)*x[1]+(-0.596335702925)*x[2]**o+(0.646045841797)*x[2]
        s[1,2,0,0]=(-0.976922812773)*x[0]**o+(0.907038408665)*x[0]+(-0.268706703732)*x[1]**o+(-0.820850017822)*x[1]+(-0.164634286435)*x[2]**o+(0.931655051043)*x[2]
        s[1,2,0,1]=(-0.659226400612)*x[0]**o+(0.693884857459)*x[0]+(0.935680326964)*x[1]**o+(0.0529754617738)*x[1]+(-0.975704228522)*x[2]**o+(-0.0103982230046)*x[2]
        s[1,2,1,0]=(-0.728054292372)*x[0]**o+(0.845592631054)*x[0]+(-0.535424426181)*x[1]**o+(-0.489415765636)*x[1]+(0.177904920871)*x[2]**o+(-0.501151392201)*x[2]
        s[1,2,1,1]=(-0.367806306107)*x[0]**o+(-0.979526890288)*x[0]+(-0.851751678512)*x[1]**o+(0.0610833858163)*x[1]+(-0.865092215813)*x[2]**o+(-0.982729156385)*x[2]
        s[1,2,2,0]=(0.890317441495)*x[0]**o+(-0.613493546336)*x[0]+(-0.832732187365)*x[1]**o+(-0.942517758759)*x[1]+(-0.678373862946)*x[2]**o+(-0.760490954483)*x[2]
        s[1,2,2,1]=(0.134821253565)*x[0]**o+(-0.357154926295)*x[0]+(0.984345806097)*x[1]**o+(0.16201908882)*x[1]+(0.673893972474)*x[2]**o+(0.757465706767)*x[2]
        s[1,3,0,0]=(-0.178752771739)*x[0]**o+(0.247240990536)*x[0]+(0.250254101973)*x[1]**o+(-0.290469953005)*x[1]+(-0.179849656966)*x[2]**o+(0.0488351177508)*x[2]
        s[1,3,0,1]=(0.0918959399079)*x[0]**o+(0.929499443933)*x[0]+(-0.142486990007)*x[1]**o+(-0.208154992937)*x[1]+(0.144490842938)*x[2]**o+(0.878566553587)*x[2]
        s[1,3,1,0]=(-0.0951485338271)*x[0]**o+(0.723847998698)*x[0]+(-0.654103310693)*x[1]**o+(-0.00133788653433)*x[1]+(0.0639363181565)*x[2]**o+(-0.240419269084)*x[2]
        s[1,3,1,1]=(-0.217090356907)*x[0]**o+(0.467191197701)*x[0]+(0.0568877529529)*x[1]**o+(0.839967116927)*x[1]+(-0.161819635829)*x[2]**o+(-0.380255238471)*x[2]
        s[1,3,2,0]=(0.139965969921)*x[0]**o+(-0.344141746066)*x[0]+(-0.343097609928)*x[1]**o+(0.323082611123)*x[1]+(0.62470424365)*x[2]**o+(0.747310443443)*x[2]
        s[1,3,2,1]=(-0.3718984783)*x[0]**o+(0.71977639778)*x[0]+(0.191430375709)*x[1]**o+(-0.873015242874)*x[1]+(0.276339628333)*x[2]**o+(0.164476716927)*x[2]
        s[1,4,0,0]=(-0.704023480703)*x[0]**o+(-0.795084252465)*x[0]+(0.637714051748)*x[1]**o+(0.808663623755)*x[1]+(0.718066739763)*x[2]**o+(-0.645424428426)*x[2]
        s[1,4,0,1]=(-0.733930327044)*x[0]**o+(-0.676332066451)*x[0]+(-0.672853173064)*x[1]**o+(0.3429113994)*x[1]+(-0.841264501476)*x[2]**o+(0.79006662583)*x[2]
        s[1,4,1,0]=(-0.528487338584)*x[0]**o+(-0.917646520301)*x[0]+(0.0304777432533)*x[1]**o+(-0.0369253092314)*x[1]+(-0.0256615497575)*x[2]**o+(0.697398576346)*x[2]
        s[1,4,1,1]=(0.326729309985)*x[0]**o+(-0.894823167111)*x[0]+(0.514358304744)*x[1]**o+(-0.173763492858)*x[1]+(-0.802294460015)*x[2]**o+(-0.602884798454)*x[2]
        s[1,4,2,0]=(-0.415051572423)*x[0]**o+(0.22496337443)*x[0]+(0.873058773308)*x[1]**o+(0.655519869148)*x[1]+(-0.480155248885)*x[2]**o+(0.590182974905)*x[2]
        s[1,4,2,1]=(-0.35633147422)*x[0]**o+(-0.717999757228)*x[0]+(0.187573310475)*x[1]**o+(-0.338108475286)*x[1]+(-0.0186934354872)*x[2]**o+(-0.560831597369)*x[2]
        s[2,0,0,0]=(0.0722896121356)*x[0]**o+(0.520666448864)*x[0]+(0.643415007545)*x[1]**o+(0.476128126603)*x[1]+(-0.0519193210173)*x[2]**o+(-0.610576319559)*x[2]
        s[2,0,0,1]=(-0.139391066523)*x[0]**o+(-0.977991803321)*x[0]+(0.222182962215)*x[1]**o+(0.537563638082)*x[1]+(0.61114193397)*x[2]**o+(0.9477815486)*x[2]
        s[2,0,1,0]=(0.177891356821)*x[0]**o+(0.108505736038)*x[0]+(0.251720839868)*x[1]**o+(-0.653300262204)*x[1]+(0.617079986603)*x[2]**o+(0.532007572217)*x[2]
        s[2,0,1,1]=(-0.0963478035027)*x[0]**o+(0.484266295486)*x[0]+(0.575754465131)*x[1]**o+(0.180169918461)*x[1]+(-0.357729254421)*x[2]**o+(-0.198225410436)*x[2]
        s[2,0,2,0]=(-0.389382670145)*x[0]**o+(-0.757173416032)*x[0]+(0.737917980599)*x[1]**o+(0.631381284716)*x[1]+(-0.540034322357)*x[2]**o+(0.0935289005824)*x[2]
        s[2,0,2,1]=(0.18046408091)*x[0]**o+(0.381098419076)*x[0]+(0.375985371309)*x[1]**o+(-0.770665651856)*x[1]+(0.746864092282)*x[2]**o+(0.196258206324)*x[2]
        s[2,1,0,0]=(0.0948044547956)*x[0]**o+(0.530048465174)*x[0]+(0.339203738259)*x[1]**o+(0.889491980197)*x[1]+(-0.701680171971)*x[2]**o+(-0.591236689057)*x[2]
        s[2,1,0,1]=(-0.0504750115604)*x[0]**o+(0.568637246489)*x[0]+(-0.782879070299)*x[1]**o+(-0.398995842249)*x[1]+(0.0900392439579)*x[2]**o+(0.330991009712)*x[2]
        s[2,1,1,0]=(-0.370473438507)*x[0]**o+(-0.894136144949)*x[0]+(-0.724710648209)*x[1]**o+(-0.889034264448)*x[1]+(-0.967606285686)*x[2]**o+(0.229302756578)*x[2]
        s[2,1,1,1]=(-0.557548738746)*x[0]**o+(0.736473161672)*x[0]+(0.132516051242)*x[1]**o+(-0.982095944769)*x[1]+(0.0862581974686)*x[2]**o+(0.648294874083)*x[2]
        s[2,1,2,0]=(0.715221366738)*x[0]**o+(0.180947754635)*x[0]+(0.774447016462)*x[1]**o+(0.2737069937)*x[1]+(-0.308017003693)*x[2]**o+(0.311365430602)*x[2]
        s[2,1,2,1]=(0.057020023918)*x[0]**o+(0.674489665955)*x[0]+(0.907045289702)*x[1]**o+(0.526829006995)*x[1]+(-0.0319147619289)*x[2]**o+(0.561637490349)*x[2]
        s[2,2,0,0]=(0.73848705426)*x[0]**o+(0.578730401802)*x[0]+(-0.877511944183)*x[1]**o+(-0.141821605426)*x[1]+(-0.610728551353)*x[2]**o+(0.921196540913)*x[2]
        s[2,2,0,1]=(-0.518273725024)*x[0]**o+(-0.293070039849)*x[0]+(-0.032874432475)*x[1]**o+(-0.876002064755)*x[1]+(-0.00232979051139)*x[2]**o+(-0.731518329401)*x[2]
        s[2,2,1,0]=(-0.179932544586)*x[0]**o+(-0.298299616601)*x[0]+(-0.507158321537)*x[1]**o+(0.477603386439)*x[1]+(-0.535441161992)*x[2]**o+(-0.114782555164)*x[2]
        s[2,2,1,1]=(-0.357820499429)*x[0]**o+(0.916561554529)*x[0]+(0.918007613979)*x[1]**o+(-0.0497200703463)*x[1]+(-0.196248955784)*x[2]**o+(0.28070766346)*x[2]
        s[2,2,2,0]=(0.686174077835)*x[0]**o+(0.227206066309)*x[0]+(0.0930909399009)*x[1]**o+(-0.772599295714)*x[1]+(0.944100518066)*x[2]**o+(0.515206215749)*x[2]
        s[2,2,2,1]=(0.996149526552)*x[0]**o+(0.828382876873)*x[0]+(0.834473760119)*x[1]**o+(-0.38781643114)*x[1]+(0.946045294719)*x[2]**o+(-0.488214962746)*x[2]
        s[2,3,0,0]=(-0.736854216588)*x[0]**o+(0.837912169829)*x[0]+(0.349541907359)*x[1]**o+(-0.821602065524)*x[1]+(-0.315670105831)*x[2]**o+(0.589034152687)*x[2]
        s[2,3,0,1]=(0.576280744681)*x[0]**o+(0.0780631366472)*x[0]+(0.0280696961332)*x[1]**o+(0.756258137578)*x[1]+(-0.778783563455)*x[2]**o+(-0.549597098669)*x[2]
        s[2,3,1,0]=(-0.986203055922)*x[0]**o+(-0.707899260726)*x[0]+(-0.739480759255)*x[1]**o+(0.438644391188)*x[1]+(-0.66655978156)*x[2]**o+(-0.96252852924)*x[2]
        s[2,3,1,1]=(0.847297144034)*x[0]**o+(-0.107063915702)*x[0]+(-0.737175500381)*x[1]**o+(-0.218515978982)*x[1]+(-0.271489760295)*x[2]**o+(0.178627729308)*x[2]
        s[2,3,2,0]=(-0.375942207511)*x[0]**o+(0.229896842267)*x[0]+(0.0831801744855)*x[1]**o+(0.869789896537)*x[1]+(0.758135854706)*x[2]**o+(-0.452003165838)*x[2]
        s[2,3,2,1]=(-0.612379338977)*x[0]**o+(0.285316223911)*x[0]+(0.571346124001)*x[1]**o+(-0.0206855235392)*x[1]+(-0.229102686518)*x[2]**o+(-0.563723359965)*x[2]
        s[2,4,0,0]=(0.119706352983)*x[0]**o+(-0.32065970677)*x[0]+(0.767856818538)*x[1]**o+(0.188328533895)*x[1]+(0.490938813855)*x[2]**o+(-0.487888173624)*x[2]
        s[2,4,0,1]=(-0.62082438463)*x[0]**o+(0.410195190757)*x[0]+(-0.801936437468)*x[1]**o+(-0.77449746445)*x[1]+(-0.943352435188)*x[2]**o+(-0.0965070084637)*x[2]
        s[2,4,1,0]=(-0.333278752646)*x[0]**o+(0.544632414515)*x[0]+(0.713044447385)*x[1]**o+(-0.134469991847)*x[1]+(-0.477562490558)*x[2]**o+(-0.717272888426)*x[2]
        s[2,4,1,1]=(0.328298581074)*x[0]**o+(0.52907257362)*x[0]+(-0.208998823722)*x[1]**o+(-0.528945934534)*x[1]+(0.230853332264)*x[2]**o+(-0.209517472451)*x[2]
        s[2,4,2,0]=(0.867173594744)*x[0]**o+(0.899212165085)*x[0]+(-0.601859471791)*x[1]**o+(-0.680256055533)*x[1]+(-0.194552761602)*x[2]**o+(0.147401121852)*x[2]
        s[2,4,2,1]=(0.453813057013)*x[0]**o+(0.698426853807)*x[0]+(-0.6460619929)*x[1]**o+(0.033010287169)*x[1]+(0.667588694113)*x[2]**o+(0.356223824338)*x[2]
        s[3,0,0,0]=(-0.292671113925)*x[0]**o+(0.616458693028)*x[0]+(-0.0103186095744)*x[1]**o+(-0.0728030520767)*x[1]+(-0.31974364911)*x[2]**o+(0.33247399209)*x[2]
        s[3,0,0,1]=(-0.992666531346)*x[0]**o+(0.0614528865103)*x[0]+(-0.534772052349)*x[1]**o+(-0.65137500858)*x[1]+(-0.192753007212)*x[2]**o+(0.320918136483)*x[2]
        s[3,0,1,0]=(-0.195020954989)*x[0]**o+(0.391152265328)*x[0]+(-0.0854218011043)*x[1]**o+(-0.925896280434)*x[1]+(0.794208026902)*x[2]**o+(-0.664891436596)*x[2]
        s[3,0,1,1]=(0.76700624636)*x[0]**o+(-0.711830311063)*x[0]+(0.415964892543)*x[1]**o+(-0.266148608643)*x[1]+(0.14347598218)*x[2]**o+(-0.444318428953)*x[2]
        s[3,0,2,0]=(-0.432044237254)*x[0]**o+(0.50230711918)*x[0]+(-0.7887127186)*x[1]**o+(-0.00439564134841)*x[1]+(0.198952339193)*x[2]**o+(-0.365097069005)*x[2]
        s[3,0,2,1]=(0.0731802841742)*x[0]**o+(-0.722529140979)*x[0]+(-0.880840005335)*x[1]**o+(-0.818607956602)*x[1]+(0.893103760434)*x[2]**o+(-0.078910408537)*x[2]
        s[3,1,0,0]=(0.424632189118)*x[0]**o+(-0.9575902588)*x[0]+(-0.669839365109)*x[1]**o+(-0.751737160991)*x[1]+(0.42069725738)*x[2]**o+(0.551903660979)*x[2]
        s[3,1,0,1]=(0.780084669674)*x[0]**o+(0.307501685577)*x[0]+(0.0810136913684)*x[1]**o+(0.702710121149)*x[1]+(-0.123438662606)*x[2]**o+(0.6469150298)*x[2]
        s[3,1,1,0]=(-0.207209742507)*x[0]**o+(0.382918019966)*x[0]+(-0.541859357863)*x[1]**o+(-0.261161834606)*x[1]+(-0.71224108949)*x[2]**o+(0.344568549136)*x[2]
        s[3,1,1,1]=(0.294545201987)*x[0]**o+(-0.310315049878)*x[0]+(0.138438017172)*x[1]**o+(0.343984655033)*x[1]+(0.156101701639)*x[2]**o+(-0.776032313589)*x[2]
        s[3,1,2,0]=(0.670052338135)*x[0]**o+(0.205560268763)*x[0]+(0.104283662799)*x[1]**o+(-0.765280448137)*x[1]+(0.562416102577)*x[2]**o+(-0.876415742019)*x[2]
        s[3,1,2,1]=(0.664650295655)*x[0]**o+(0.795254957531)*x[0]+(-0.552860322817)*x[1]**o+(-0.427967184203)*x[1]+(0.952252946717)*x[2]**o+(0.865393511228)*x[2]
        s[3,2,0,0]=(-0.708432407412)*x[0]**o+(0.308974057108)*x[0]+(-0.749146897323)*x[1]**o+(-0.00624264988729)*x[1]+(0.694675829788)*x[2]**o+(-0.831205537426)*x[2]
        s[3,2,0,1]=(-0.512843864369)*x[0]**o+(0.750200436937)*x[0]+(0.219611724574)*x[1]**o+(0.538914229585)*x[1]+(-0.750949138305)*x[2]**o+(-0.730109185869)*x[2]
        s[3,2,1,0]=(0.307377290744)*x[0]**o+(0.422564975595)*x[0]+(-0.575777057405)*x[1]**o+(0.0843754344109)*x[1]+(0.563608883617)*x[2]**o+(0.0203552581819)*x[2]
        s[3,2,1,1]=(0.844926629988)*x[0]**o+(0.00463764382667)*x[0]+(-0.801305450023)*x[1]**o+(-0.297827538391)*x[1]+(-0.334832523936)*x[2]**o+(-0.145250163264)*x[2]
        s[3,2,2,0]=(-0.571129120261)*x[0]**o+(0.794459190551)*x[0]+(-0.0558676868113)*x[1]**o+(-0.12434806027)*x[1]+(0.666910606164)*x[2]**o+(0.0904375266689)*x[2]
        s[3,2,2,1]=(0.537887532259)*x[0]**o+(0.271597496543)*x[0]+(-0.345121311853)*x[1]**o+(-0.0110951188557)*x[1]+(-0.691735728081)*x[2]**o+(0.227334216755)*x[2]
        s[3,3,0,0]=(-0.617039404022)*x[0]**o+(-0.0600488492522)*x[0]+(0.567478593699)*x[1]**o+(0.625697236766)*x[1]+(0.226173040972)*x[2]**o+(0.592063112338)*x[2]
        s[3,3,0,1]=(0.569700282793)*x[0]**o+(-0.00215167041353)*x[0]+(0.60591609073)*x[1]**o+(0.362583373854)*x[1]+(0.654506836835)*x[2]**o+(-0.144610108143)*x[2]
        s[3,3,1,0]=(0.246640361949)*x[0]**o+(-0.86358850448)*x[0]+(0.157520564836)*x[1]**o+(-0.059233078017)*x[1]+(0.86570653333)*x[2]**o+(0.685576804797)*x[2]
        s[3,3,1,1]=(-0.958130500315)*x[0]**o+(-0.259028395499)*x[0]+(0.120316829457)*x[1]**o+(-0.758878478978)*x[1]+(-0.43066120723)*x[2]**o+(0.800327000625)*x[2]
        s[3,3,2,0]=(0.920887970019)*x[0]**o+(-0.503403737241)*x[0]+(0.0282527472159)*x[1]**o+(-0.167650834072)*x[1]+(0.416811246461)*x[2]**o+(-0.103817262329)*x[2]
        s[3,3,2,1]=(-0.58133852513)*x[0]**o+(-0.176946132911)*x[0]+(0.0347620468937)*x[1]**o+(-0.973916175156)*x[1]+(0.71011981519)*x[2]**o+(-0.339266963019)*x[2]
        s[3,4,0,0]=(-0.0120671975843)*x[0]**o+(0.109594354705)*x[0]+(0.151599950487)*x[1]**o+(0.881150957088)*x[1]+(0.959435811252)*x[2]**o+(-0.86774776724)*x[2]
        s[3,4,0,1]=(-0.0545164901018)*x[0]**o+(-0.987079399801)*x[0]+(0.0168167160052)*x[1]**o+(0.323974666845)*x[1]+(-0.169334724946)*x[2]**o+(-0.95604743937)*x[2]
        s[3,4,1,0]=(0.497905965869)*x[0]**o+(-0.215618743574)*x[0]+(0.333131793569)*x[1]**o+(0.465071861103)*x[1]+(0.472276780484)*x[2]**o+(-0.587116523215)*x[2]
        s[3,4,1,1]=(-0.798850078462)*x[0]**o+(-0.385483486673)*x[0]+(-0.954899272011)*x[1]**o+(0.985953307837)*x[1]+(0.5064455397)*x[2]**o+(0.201146433376)*x[2]
        s[3,4,2,0]=(0.736565587072)*x[0]**o+(-0.383544786329)*x[0]+(0.345205635474)*x[1]**o+(0.118507188601)*x[1]+(0.933209008402)*x[2]**o+(0.00295145110263)*x[2]
        s[3,4,2,1]=(-0.0892194210246)*x[0]**o+(0.152917048)*x[0]+(0.291092526754)*x[1]**o+(-0.13749152394)*x[1]+(-0.820302461745)*x[2]**o+(-0.971364751692)*x[2]
        ref[0,0,0,0]=(-0.209698190814)*x_ref[0]**o+(0.0202198726379)*x_ref[0]+(-0.776762051242)*x_ref[1]**o+(0.0103026495025)*x_ref[1]+(0.0277886960026)*x_ref[2]**o+(-0.255065901162)*x_ref[2]
        ref[0,0,0,1]=(-0.663085958734)*x_ref[0]**o+(0.698525770087)*x_ref[0]+(0.650288968593)*x_ref[1]**o+(0.6069536467)*x_ref[1]+(0.789898121185)*x_ref[2]**o+(-0.13315288412)*x_ref[2]
        ref[0,0,1,0]=(-0.0998227591334)*x_ref[0]**o+(0.75314681829)*x_ref[0]+(-0.43575224825)*x_ref[1]**o+(0.649609416117)*x_ref[1]+(0.271394201051)*x_ref[2]**o+(-0.353712050582)*x_ref[2]
        ref[0,0,1,1]=(0.205441533003)*x_ref[0]**o+(-0.61122881962)*x_ref[0]+(-0.431370221976)*x_ref[1]**o+(-0.183655447483)*x_ref[1]+(0.116437364187)*x_ref[2]**o+(0.741511658665)*x_ref[2]
        ref[0,0,2,0]=(0.516133026558)*x_ref[0]**o+(-0.904105438162)*x_ref[0]+(-0.579514088677)*x_ref[1]**o+(-0.0581332954324)*x_ref[1]+(0.557937742112)*x_ref[2]**o+(0.138639028149)*x_ref[2]
        ref[0,0,2,1]=(0.8991763477)*x_ref[0]**o+(-0.792213819942)*x_ref[0]+(0.674755445696)*x_ref[1]**o+(0.483315869458)*x_ref[1]+(-0.0722431463476)*x_ref[2]**o+(0.0270814412191)*x_ref[2]
        ref[0,1,0,0]=(0.366564587764)*x_ref[0]**o+(0.928578857146)*x_ref[0]+(0.785092228323)*x_ref[1]**o+(0.860812095832)*x_ref[1]+(-0.603464315283)*x_ref[2]**o+(-0.556636685093)*x_ref[2]
        ref[0,1,0,1]=(0.980630839809)*x_ref[0]**o+(-0.255324039441)*x_ref[0]+(0.173630938787)*x_ref[1]**o+(0.110433932158)*x_ref[1]+(-0.392851296251)*x_ref[2]**o+(-0.272438563385)*x_ref[2]
        ref[0,1,1,0]=(-0.63489459343)*x_ref[0]**o+(0.0328950965715)*x_ref[0]+(-0.199264976942)*x_ref[1]**o+(0.704589249988)*x_ref[1]+(0.442306934744)*x_ref[2]**o+(-0.531996394873)*x_ref[2]
        ref[0,1,1,1]=(-0.110263035707)*x_ref[0]**o+(0.904756528197)*x_ref[0]+(0.876124769497)*x_ref[1]**o+(-0.55469630081)*x_ref[1]+(-0.857675499295)*x_ref[2]**o+(0.662352375455)*x_ref[2]
        ref[0,1,2,0]=(-0.383336639984)*x_ref[0]**o+(0.863605840878)*x_ref[0]+(0.742039996033)*x_ref[1]**o+(0.649587811572)*x_ref[1]+(0.0749104131981)*x_ref[2]**o+(-0.943341654469)*x_ref[2]
        ref[0,1,2,1]=(0.200285377439)*x_ref[0]**o+(0.148329648537)*x_ref[0]+(-0.0994444073677)*x_ref[1]**o+(-0.535438035359)*x_ref[1]+(0.662913271697)*x_ref[2]**o+(0.112409762773)*x_ref[2]
        ref[0,2,0,0]=(0.351533608017)*x_ref[0]**o+(0.72576521785)*x_ref[0]+(0.305528610204)*x_ref[1]**o+(-0.139400610385)*x_ref[1]+(-0.455936488645)*x_ref[2]**o+(-0.992626031371)*x_ref[2]
        ref[0,2,0,1]=(0.351745471157)*x_ref[0]**o+(0.749900692382)*x_ref[0]+(0.212677388818)*x_ref[1]**o+(0.449543777767)*x_ref[1]+(0.9265756875)*x_ref[2]**o+(0.0973575097882)*x_ref[2]
        ref[0,2,1,0]=(0.308471390044)*x_ref[0]**o+(-0.236492160276)*x_ref[0]+(-0.0132386386917)*x_ref[1]**o+(-0.657824382959)*x_ref[1]+(0.567169125655)*x_ref[2]**o+(-0.0239580661983)*x_ref[2]
        ref[0,2,1,1]=(-0.71982992429)*x_ref[0]**o+(-0.518285193827)*x_ref[0]+(0.806477863481)*x_ref[1]**o+(-0.949670246654)*x_ref[1]+(-0.559408139171)*x_ref[2]**o+(-0.740059454923)*x_ref[2]
        ref[0,2,2,0]=(-0.795819227823)*x_ref[0]**o+(0.0382042059208)*x_ref[0]+(-0.993434888742)*x_ref[1]**o+(-0.714572217225)*x_ref[1]+(0.826057540197)*x_ref[2]**o+(-0.351385211283)*x_ref[2]
        ref[0,2,2,1]=(-0.691428669925)*x_ref[0]**o+(0.795109370028)*x_ref[0]+(-0.845580170561)*x_ref[1]**o+(0.66720897925)*x_ref[1]+(-0.604013104606)*x_ref[2]**o+(-0.56003576527)*x_ref[2]
        ref[0,3,0,0]=(-0.864934911539)*x_ref[0]**o+(-0.871210034972)*x_ref[0]+(-0.633807262008)*x_ref[1]**o+(-0.0448147244258)*x_ref[1]+(0.601871519726)*x_ref[2]**o+(-0.423895847468)*x_ref[2]
        ref[0,3,0,1]=(0.542426383922)*x_ref[0]**o+(0.648956460184)*x_ref[0]+(0.975476287059)*x_ref[1]**o+(-0.110876177567)*x_ref[1]+(-0.334216196036)*x_ref[2]**o+(-0.696393938466)*x_ref[2]
        ref[0,3,1,0]=(-0.897499573749)*x_ref[0]**o+(0.504413212966)*x_ref[0]+(-0.114527615291)*x_ref[1]**o+(0.415575654212)*x_ref[1]+(-0.271870634022)*x_ref[2]**o+(0.698810118326)*x_ref[2]
        ref[0,3,1,1]=(0.231758813026)*x_ref[0]**o+(-0.207508138983)*x_ref[0]+(-0.744927387959)*x_ref[1]**o+(0.00489240105965)*x_ref[1]+(0.821414365164)*x_ref[2]**o+(0.701679220215)*x_ref[2]
        ref[0,3,2,0]=(-0.135996215059)*x_ref[0]**o+(-0.513005242192)*x_ref[0]+(-0.6068585358)*x_ref[1]**o+(-0.852285533431)*x_ref[1]+(0.115701663464)*x_ref[2]**o+(0.243886826101)*x_ref[2]
        ref[0,3,2,1]=(-0.80781983055)*x_ref[0]**o+(0.142781343516)*x_ref[0]+(0.764546896757)*x_ref[1]**o+(0.478359930452)*x_ref[1]+(-0.0458636815218)*x_ref[2]**o+(0.582425549793)*x_ref[2]
        ref[0,4,0,0]=(-0.382194450205)*x_ref[0]**o+(-0.209573752081)*x_ref[0]+(0.959742867105)*x_ref[1]**o+(0.962658893536)*x_ref[1]+(0.126101921123)*x_ref[2]**o+(0.980806872196)*x_ref[2]
        ref[0,4,0,1]=(0.764116933095)*x_ref[0]**o+(-0.648257603162)*x_ref[0]+(-0.839456477431)*x_ref[1]**o+(-0.972616251638)*x_ref[1]+(-0.815159939908)*x_ref[2]**o+(-0.262639210819)*x_ref[2]
        ref[0,4,1,0]=(0.453438243261)*x_ref[0]**o+(0.303833796601)*x_ref[0]+(0.100367120622)*x_ref[1]**o+(-0.0716774769052)*x_ref[1]+(-0.659569961883)*x_ref[2]**o+(-0.290044782372)*x_ref[2]
        ref[0,4,1,1]=(-0.0286107263502)*x_ref[0]**o+(-0.596438980382)*x_ref[0]+(0.545044960979)*x_ref[1]**o+(-0.796415090612)*x_ref[1]+(0.206907297358)*x_ref[2]**o+(0.306889564609)*x_ref[2]
        ref[0,4,2,0]=(-0.201162762403)*x_ref[0]**o+(-0.981302954607)*x_ref[0]+(-0.201138026976)*x_ref[1]**o+(-0.93628245382)*x_ref[1]+(-0.758945968148)*x_ref[2]**o+(-0.860206802477)*x_ref[2]
        ref[0,4,2,1]=(0.251350117686)*x_ref[0]**o+(-0.29118495379)*x_ref[0]+(-0.569363617929)*x_ref[1]**o+(0.465078119323)*x_ref[1]+(0.426184283109)*x_ref[2]**o+(-0.873643118829)*x_ref[2]
        ref[1,0,0,0]=(-0.0185687413154)*x_ref[0]**o+(0.411847218239)*x_ref[0]+(0.428984420237)*x_ref[1]**o+(-0.547264849731)*x_ref[1]+(0.486727670471)*x_ref[2]**o+(0.179455122763)*x_ref[2]
        ref[1,0,0,1]=(0.518433022846)*x_ref[0]**o+(0.025823449242)*x_ref[0]+(-0.0540511749223)*x_ref[1]**o+(-0.85848829682)*x_ref[1]+(0.158385940315)*x_ref[2]**o+(-0.178171294535)*x_ref[2]
        ref[1,0,1,0]=(0.0367258516392)*x_ref[0]**o+(0.545049981245)*x_ref[0]+(-0.579044650548)*x_ref[1]**o+(-0.77725873354)*x_ref[1]+(-0.034299394836)*x_ref[2]**o+(-0.179489216766)*x_ref[2]
        ref[1,0,1,1]=(0.274129185725)*x_ref[0]**o+(0.166006734196)*x_ref[0]+(-0.773979014851)*x_ref[1]**o+(-0.519501835507)*x_ref[1]+(-0.63912427991)*x_ref[2]**o+(0.466724690101)*x_ref[2]
        ref[1,0,2,0]=(0.482955790279)*x_ref[0]**o+(0.471838320459)*x_ref[0]+(0.617771038211)*x_ref[1]**o+(0.048311205466)*x_ref[1]+(0.162393942754)*x_ref[2]**o+(-0.694439775285)*x_ref[2]
        ref[1,0,2,1]=(-0.517879311536)*x_ref[0]**o+(-0.730313900842)*x_ref[0]+(-0.991744866252)*x_ref[1]**o+(-0.849707440389)*x_ref[1]+(-0.83668608481)*x_ref[2]**o+(0.581082133088)*x_ref[2]
        ref[1,1,0,0]=(0.217328148299)*x_ref[0]**o+(0.187729052039)*x_ref[0]+(-0.68877270235)*x_ref[1]**o+(-0.86757435657)*x_ref[1]+(-0.427061492994)*x_ref[2]**o+(-0.84402543773)*x_ref[2]
        ref[1,1,0,1]=(0.766541346426)*x_ref[0]**o+(0.649768542608)*x_ref[0]+(-0.113463877205)*x_ref[1]**o+(-0.637749133894)*x_ref[1]+(0.809878689222)*x_ref[2]**o+(0.315784984005)*x_ref[2]
        ref[1,1,1,0]=(0.94173972573)*x_ref[0]**o+(0.668276294204)*x_ref[0]+(0.427204723683)*x_ref[1]**o+(-0.607750747607)*x_ref[1]+(-0.0376503804716)*x_ref[2]**o+(0.960260009026)*x_ref[2]
        ref[1,1,1,1]=(-0.400233578622)*x_ref[0]**o+(-0.923564602772)*x_ref[0]+(0.47175931621)*x_ref[1]**o+(-0.612010762622)*x_ref[1]+(0.609225052578)*x_ref[2]**o+(0.526546928045)*x_ref[2]
        ref[1,1,2,0]=(0.377678503521)*x_ref[0]**o+(0.892169209135)*x_ref[0]+(0.39772744849)*x_ref[1]**o+(-0.166966626427)*x_ref[1]+(0.874470437184)*x_ref[2]**o+(0.314522738087)*x_ref[2]
        ref[1,1,2,1]=(0.520217518976)*x_ref[0]**o+(0.536152467541)*x_ref[0]+(0.444145502111)*x_ref[1]**o+(-0.489860603506)*x_ref[1]+(-0.596335702925)*x_ref[2]**o+(0.646045841797)*x_ref[2]
        ref[1,2,0,0]=(-0.976922812773)*x_ref[0]**o+(0.907038408665)*x_ref[0]+(-0.268706703732)*x_ref[1]**o+(-0.820850017822)*x_ref[1]+(-0.164634286435)*x_ref[2]**o+(0.931655051043)*x_ref[2]
        ref[1,2,0,1]=(-0.659226400612)*x_ref[0]**o+(0.693884857459)*x_ref[0]+(0.935680326964)*x_ref[1]**o+(0.0529754617738)*x_ref[1]+(-0.975704228522)*x_ref[2]**o+(-0.0103982230046)*x_ref[2]
        ref[1,2,1,0]=(-0.728054292372)*x_ref[0]**o+(0.845592631054)*x_ref[0]+(-0.535424426181)*x_ref[1]**o+(-0.489415765636)*x_ref[1]+(0.177904920871)*x_ref[2]**o+(-0.501151392201)*x_ref[2]
        ref[1,2,1,1]=(-0.367806306107)*x_ref[0]**o+(-0.979526890288)*x_ref[0]+(-0.851751678512)*x_ref[1]**o+(0.0610833858163)*x_ref[1]+(-0.865092215813)*x_ref[2]**o+(-0.982729156385)*x_ref[2]
        ref[1,2,2,0]=(0.890317441495)*x_ref[0]**o+(-0.613493546336)*x_ref[0]+(-0.832732187365)*x_ref[1]**o+(-0.942517758759)*x_ref[1]+(-0.678373862946)*x_ref[2]**o+(-0.760490954483)*x_ref[2]
        ref[1,2,2,1]=(0.134821253565)*x_ref[0]**o+(-0.357154926295)*x_ref[0]+(0.984345806097)*x_ref[1]**o+(0.16201908882)*x_ref[1]+(0.673893972474)*x_ref[2]**o+(0.757465706767)*x_ref[2]
        ref[1,3,0,0]=(-0.178752771739)*x_ref[0]**o+(0.247240990536)*x_ref[0]+(0.250254101973)*x_ref[1]**o+(-0.290469953005)*x_ref[1]+(-0.179849656966)*x_ref[2]**o+(0.0488351177508)*x_ref[2]
        ref[1,3,0,1]=(0.0918959399079)*x_ref[0]**o+(0.929499443933)*x_ref[0]+(-0.142486990007)*x_ref[1]**o+(-0.208154992937)*x_ref[1]+(0.144490842938)*x_ref[2]**o+(0.878566553587)*x_ref[2]
        ref[1,3,1,0]=(-0.0951485338271)*x_ref[0]**o+(0.723847998698)*x_ref[0]+(-0.654103310693)*x_ref[1]**o+(-0.00133788653433)*x_ref[1]+(0.0639363181565)*x_ref[2]**o+(-0.240419269084)*x_ref[2]
        ref[1,3,1,1]=(-0.217090356907)*x_ref[0]**o+(0.467191197701)*x_ref[0]+(0.0568877529529)*x_ref[1]**o+(0.839967116927)*x_ref[1]+(-0.161819635829)*x_ref[2]**o+(-0.380255238471)*x_ref[2]
        ref[1,3,2,0]=(0.139965969921)*x_ref[0]**o+(-0.344141746066)*x_ref[0]+(-0.343097609928)*x_ref[1]**o+(0.323082611123)*x_ref[1]+(0.62470424365)*x_ref[2]**o+(0.747310443443)*x_ref[2]
        ref[1,3,2,1]=(-0.3718984783)*x_ref[0]**o+(0.71977639778)*x_ref[0]+(0.191430375709)*x_ref[1]**o+(-0.873015242874)*x_ref[1]+(0.276339628333)*x_ref[2]**o+(0.164476716927)*x_ref[2]
        ref[1,4,0,0]=(-0.704023480703)*x_ref[0]**o+(-0.795084252465)*x_ref[0]+(0.637714051748)*x_ref[1]**o+(0.808663623755)*x_ref[1]+(0.718066739763)*x_ref[2]**o+(-0.645424428426)*x_ref[2]
        ref[1,4,0,1]=(-0.733930327044)*x_ref[0]**o+(-0.676332066451)*x_ref[0]+(-0.672853173064)*x_ref[1]**o+(0.3429113994)*x_ref[1]+(-0.841264501476)*x_ref[2]**o+(0.79006662583)*x_ref[2]
        ref[1,4,1,0]=(-0.528487338584)*x_ref[0]**o+(-0.917646520301)*x_ref[0]+(0.0304777432533)*x_ref[1]**o+(-0.0369253092314)*x_ref[1]+(-0.0256615497575)*x_ref[2]**o+(0.697398576346)*x_ref[2]
        ref[1,4,1,1]=(0.326729309985)*x_ref[0]**o+(-0.894823167111)*x_ref[0]+(0.514358304744)*x_ref[1]**o+(-0.173763492858)*x_ref[1]+(-0.802294460015)*x_ref[2]**o+(-0.602884798454)*x_ref[2]
        ref[1,4,2,0]=(-0.415051572423)*x_ref[0]**o+(0.22496337443)*x_ref[0]+(0.873058773308)*x_ref[1]**o+(0.655519869148)*x_ref[1]+(-0.480155248885)*x_ref[2]**o+(0.590182974905)*x_ref[2]
        ref[1,4,2,1]=(-0.35633147422)*x_ref[0]**o+(-0.717999757228)*x_ref[0]+(0.187573310475)*x_ref[1]**o+(-0.338108475286)*x_ref[1]+(-0.0186934354872)*x_ref[2]**o+(-0.560831597369)*x_ref[2]
        ref[2,0,0,0]=(0.0722896121356)*x_ref[0]**o+(0.520666448864)*x_ref[0]+(0.643415007545)*x_ref[1]**o+(0.476128126603)*x_ref[1]+(-0.0519193210173)*x_ref[2]**o+(-0.610576319559)*x_ref[2]
        ref[2,0,0,1]=(-0.139391066523)*x_ref[0]**o+(-0.977991803321)*x_ref[0]+(0.222182962215)*x_ref[1]**o+(0.537563638082)*x_ref[1]+(0.61114193397)*x_ref[2]**o+(0.9477815486)*x_ref[2]
        ref[2,0,1,0]=(0.177891356821)*x_ref[0]**o+(0.108505736038)*x_ref[0]+(0.251720839868)*x_ref[1]**o+(-0.653300262204)*x_ref[1]+(0.617079986603)*x_ref[2]**o+(0.532007572217)*x_ref[2]
        ref[2,0,1,1]=(-0.0963478035027)*x_ref[0]**o+(0.484266295486)*x_ref[0]+(0.575754465131)*x_ref[1]**o+(0.180169918461)*x_ref[1]+(-0.357729254421)*x_ref[2]**o+(-0.198225410436)*x_ref[2]
        ref[2,0,2,0]=(-0.389382670145)*x_ref[0]**o+(-0.757173416032)*x_ref[0]+(0.737917980599)*x_ref[1]**o+(0.631381284716)*x_ref[1]+(-0.540034322357)*x_ref[2]**o+(0.0935289005824)*x_ref[2]
        ref[2,0,2,1]=(0.18046408091)*x_ref[0]**o+(0.381098419076)*x_ref[0]+(0.375985371309)*x_ref[1]**o+(-0.770665651856)*x_ref[1]+(0.746864092282)*x_ref[2]**o+(0.196258206324)*x_ref[2]
        ref[2,1,0,0]=(0.0948044547956)*x_ref[0]**o+(0.530048465174)*x_ref[0]+(0.339203738259)*x_ref[1]**o+(0.889491980197)*x_ref[1]+(-0.701680171971)*x_ref[2]**o+(-0.591236689057)*x_ref[2]
        ref[2,1,0,1]=(-0.0504750115604)*x_ref[0]**o+(0.568637246489)*x_ref[0]+(-0.782879070299)*x_ref[1]**o+(-0.398995842249)*x_ref[1]+(0.0900392439579)*x_ref[2]**o+(0.330991009712)*x_ref[2]
        ref[2,1,1,0]=(-0.370473438507)*x_ref[0]**o+(-0.894136144949)*x_ref[0]+(-0.724710648209)*x_ref[1]**o+(-0.889034264448)*x_ref[1]+(-0.967606285686)*x_ref[2]**o+(0.229302756578)*x_ref[2]
        ref[2,1,1,1]=(-0.557548738746)*x_ref[0]**o+(0.736473161672)*x_ref[0]+(0.132516051242)*x_ref[1]**o+(-0.982095944769)*x_ref[1]+(0.0862581974686)*x_ref[2]**o+(0.648294874083)*x_ref[2]
        ref[2,1,2,0]=(0.715221366738)*x_ref[0]**o+(0.180947754635)*x_ref[0]+(0.774447016462)*x_ref[1]**o+(0.2737069937)*x_ref[1]+(-0.308017003693)*x_ref[2]**o+(0.311365430602)*x_ref[2]
        ref[2,1,2,1]=(0.057020023918)*x_ref[0]**o+(0.674489665955)*x_ref[0]+(0.907045289702)*x_ref[1]**o+(0.526829006995)*x_ref[1]+(-0.0319147619289)*x_ref[2]**o+(0.561637490349)*x_ref[2]
        ref[2,2,0,0]=(0.73848705426)*x_ref[0]**o+(0.578730401802)*x_ref[0]+(-0.877511944183)*x_ref[1]**o+(-0.141821605426)*x_ref[1]+(-0.610728551353)*x_ref[2]**o+(0.921196540913)*x_ref[2]
        ref[2,2,0,1]=(-0.518273725024)*x_ref[0]**o+(-0.293070039849)*x_ref[0]+(-0.032874432475)*x_ref[1]**o+(-0.876002064755)*x_ref[1]+(-0.00232979051139)*x_ref[2]**o+(-0.731518329401)*x_ref[2]
        ref[2,2,1,0]=(-0.179932544586)*x_ref[0]**o+(-0.298299616601)*x_ref[0]+(-0.507158321537)*x_ref[1]**o+(0.477603386439)*x_ref[1]+(-0.535441161992)*x_ref[2]**o+(-0.114782555164)*x_ref[2]
        ref[2,2,1,1]=(-0.357820499429)*x_ref[0]**o+(0.916561554529)*x_ref[0]+(0.918007613979)*x_ref[1]**o+(-0.0497200703463)*x_ref[1]+(-0.196248955784)*x_ref[2]**o+(0.28070766346)*x_ref[2]
        ref[2,2,2,0]=(0.686174077835)*x_ref[0]**o+(0.227206066309)*x_ref[0]+(0.0930909399009)*x_ref[1]**o+(-0.772599295714)*x_ref[1]+(0.944100518066)*x_ref[2]**o+(0.515206215749)*x_ref[2]
        ref[2,2,2,1]=(0.996149526552)*x_ref[0]**o+(0.828382876873)*x_ref[0]+(0.834473760119)*x_ref[1]**o+(-0.38781643114)*x_ref[1]+(0.946045294719)*x_ref[2]**o+(-0.488214962746)*x_ref[2]
        ref[2,3,0,0]=(-0.736854216588)*x_ref[0]**o+(0.837912169829)*x_ref[0]+(0.349541907359)*x_ref[1]**o+(-0.821602065524)*x_ref[1]+(-0.315670105831)*x_ref[2]**o+(0.589034152687)*x_ref[2]
        ref[2,3,0,1]=(0.576280744681)*x_ref[0]**o+(0.0780631366472)*x_ref[0]+(0.0280696961332)*x_ref[1]**o+(0.756258137578)*x_ref[1]+(-0.778783563455)*x_ref[2]**o+(-0.549597098669)*x_ref[2]
        ref[2,3,1,0]=(-0.986203055922)*x_ref[0]**o+(-0.707899260726)*x_ref[0]+(-0.739480759255)*x_ref[1]**o+(0.438644391188)*x_ref[1]+(-0.66655978156)*x_ref[2]**o+(-0.96252852924)*x_ref[2]
        ref[2,3,1,1]=(0.847297144034)*x_ref[0]**o+(-0.107063915702)*x_ref[0]+(-0.737175500381)*x_ref[1]**o+(-0.218515978982)*x_ref[1]+(-0.271489760295)*x_ref[2]**o+(0.178627729308)*x_ref[2]
        ref[2,3,2,0]=(-0.375942207511)*x_ref[0]**o+(0.229896842267)*x_ref[0]+(0.0831801744855)*x_ref[1]**o+(0.869789896537)*x_ref[1]+(0.758135854706)*x_ref[2]**o+(-0.452003165838)*x_ref[2]
        ref[2,3,2,1]=(-0.612379338977)*x_ref[0]**o+(0.285316223911)*x_ref[0]+(0.571346124001)*x_ref[1]**o+(-0.0206855235392)*x_ref[1]+(-0.229102686518)*x_ref[2]**o+(-0.563723359965)*x_ref[2]
        ref[2,4,0,0]=(0.119706352983)*x_ref[0]**o+(-0.32065970677)*x_ref[0]+(0.767856818538)*x_ref[1]**o+(0.188328533895)*x_ref[1]+(0.490938813855)*x_ref[2]**o+(-0.487888173624)*x_ref[2]
        ref[2,4,0,1]=(-0.62082438463)*x_ref[0]**o+(0.410195190757)*x_ref[0]+(-0.801936437468)*x_ref[1]**o+(-0.77449746445)*x_ref[1]+(-0.943352435188)*x_ref[2]**o+(-0.0965070084637)*x_ref[2]
        ref[2,4,1,0]=(-0.333278752646)*x_ref[0]**o+(0.544632414515)*x_ref[0]+(0.713044447385)*x_ref[1]**o+(-0.134469991847)*x_ref[1]+(-0.477562490558)*x_ref[2]**o+(-0.717272888426)*x_ref[2]
        ref[2,4,1,1]=(0.328298581074)*x_ref[0]**o+(0.52907257362)*x_ref[0]+(-0.208998823722)*x_ref[1]**o+(-0.528945934534)*x_ref[1]+(0.230853332264)*x_ref[2]**o+(-0.209517472451)*x_ref[2]
        ref[2,4,2,0]=(0.867173594744)*x_ref[0]**o+(0.899212165085)*x_ref[0]+(-0.601859471791)*x_ref[1]**o+(-0.680256055533)*x_ref[1]+(-0.194552761602)*x_ref[2]**o+(0.147401121852)*x_ref[2]
        ref[2,4,2,1]=(0.453813057013)*x_ref[0]**o+(0.698426853807)*x_ref[0]+(-0.6460619929)*x_ref[1]**o+(0.033010287169)*x_ref[1]+(0.667588694113)*x_ref[2]**o+(0.356223824338)*x_ref[2]
        ref[3,0,0,0]=(-0.292671113925)*x_ref[0]**o+(0.616458693028)*x_ref[0]+(-0.0103186095744)*x_ref[1]**o+(-0.0728030520767)*x_ref[1]+(-0.31974364911)*x_ref[2]**o+(0.33247399209)*x_ref[2]
        ref[3,0,0,1]=(-0.992666531346)*x_ref[0]**o+(0.0614528865103)*x_ref[0]+(-0.534772052349)*x_ref[1]**o+(-0.65137500858)*x_ref[1]+(-0.192753007212)*x_ref[2]**o+(0.320918136483)*x_ref[2]
        ref[3,0,1,0]=(-0.195020954989)*x_ref[0]**o+(0.391152265328)*x_ref[0]+(-0.0854218011043)*x_ref[1]**o+(-0.925896280434)*x_ref[1]+(0.794208026902)*x_ref[2]**o+(-0.664891436596)*x_ref[2]
        ref[3,0,1,1]=(0.76700624636)*x_ref[0]**o+(-0.711830311063)*x_ref[0]+(0.415964892543)*x_ref[1]**o+(-0.266148608643)*x_ref[1]+(0.14347598218)*x_ref[2]**o+(-0.444318428953)*x_ref[2]
        ref[3,0,2,0]=(-0.432044237254)*x_ref[0]**o+(0.50230711918)*x_ref[0]+(-0.7887127186)*x_ref[1]**o+(-0.00439564134841)*x_ref[1]+(0.198952339193)*x_ref[2]**o+(-0.365097069005)*x_ref[2]
        ref[3,0,2,1]=(0.0731802841742)*x_ref[0]**o+(-0.722529140979)*x_ref[0]+(-0.880840005335)*x_ref[1]**o+(-0.818607956602)*x_ref[1]+(0.893103760434)*x_ref[2]**o+(-0.078910408537)*x_ref[2]
        ref[3,1,0,0]=(0.424632189118)*x_ref[0]**o+(-0.9575902588)*x_ref[0]+(-0.669839365109)*x_ref[1]**o+(-0.751737160991)*x_ref[1]+(0.42069725738)*x_ref[2]**o+(0.551903660979)*x_ref[2]
        ref[3,1,0,1]=(0.780084669674)*x_ref[0]**o+(0.307501685577)*x_ref[0]+(0.0810136913684)*x_ref[1]**o+(0.702710121149)*x_ref[1]+(-0.123438662606)*x_ref[2]**o+(0.6469150298)*x_ref[2]
        ref[3,1,1,0]=(-0.207209742507)*x_ref[0]**o+(0.382918019966)*x_ref[0]+(-0.541859357863)*x_ref[1]**o+(-0.261161834606)*x_ref[1]+(-0.71224108949)*x_ref[2]**o+(0.344568549136)*x_ref[2]
        ref[3,1,1,1]=(0.294545201987)*x_ref[0]**o+(-0.310315049878)*x_ref[0]+(0.138438017172)*x_ref[1]**o+(0.343984655033)*x_ref[1]+(0.156101701639)*x_ref[2]**o+(-0.776032313589)*x_ref[2]
        ref[3,1,2,0]=(0.670052338135)*x_ref[0]**o+(0.205560268763)*x_ref[0]+(0.104283662799)*x_ref[1]**o+(-0.765280448137)*x_ref[1]+(0.562416102577)*x_ref[2]**o+(-0.876415742019)*x_ref[2]
        ref[3,1,2,1]=(0.664650295655)*x_ref[0]**o+(0.795254957531)*x_ref[0]+(-0.552860322817)*x_ref[1]**o+(-0.427967184203)*x_ref[1]+(0.952252946717)*x_ref[2]**o+(0.865393511228)*x_ref[2]
        ref[3,2,0,0]=(-0.708432407412)*x_ref[0]**o+(0.308974057108)*x_ref[0]+(-0.749146897323)*x_ref[1]**o+(-0.00624264988729)*x_ref[1]+(0.694675829788)*x_ref[2]**o+(-0.831205537426)*x_ref[2]
        ref[3,2,0,1]=(-0.512843864369)*x_ref[0]**o+(0.750200436937)*x_ref[0]+(0.219611724574)*x_ref[1]**o+(0.538914229585)*x_ref[1]+(-0.750949138305)*x_ref[2]**o+(-0.730109185869)*x_ref[2]
        ref[3,2,1,0]=(0.307377290744)*x_ref[0]**o+(0.422564975595)*x_ref[0]+(-0.575777057405)*x_ref[1]**o+(0.0843754344109)*x_ref[1]+(0.563608883617)*x_ref[2]**o+(0.0203552581819)*x_ref[2]
        ref[3,2,1,1]=(0.844926629988)*x_ref[0]**o+(0.00463764382667)*x_ref[0]+(-0.801305450023)*x_ref[1]**o+(-0.297827538391)*x_ref[1]+(-0.334832523936)*x_ref[2]**o+(-0.145250163264)*x_ref[2]
        ref[3,2,2,0]=(-0.571129120261)*x_ref[0]**o+(0.794459190551)*x_ref[0]+(-0.0558676868113)*x_ref[1]**o+(-0.12434806027)*x_ref[1]+(0.666910606164)*x_ref[2]**o+(0.0904375266689)*x_ref[2]
        ref[3,2,2,1]=(0.537887532259)*x_ref[0]**o+(0.271597496543)*x_ref[0]+(-0.345121311853)*x_ref[1]**o+(-0.0110951188557)*x_ref[1]+(-0.691735728081)*x_ref[2]**o+(0.227334216755)*x_ref[2]
        ref[3,3,0,0]=(-0.617039404022)*x_ref[0]**o+(-0.0600488492522)*x_ref[0]+(0.567478593699)*x_ref[1]**o+(0.625697236766)*x_ref[1]+(0.226173040972)*x_ref[2]**o+(0.592063112338)*x_ref[2]
        ref[3,3,0,1]=(0.569700282793)*x_ref[0]**o+(-0.00215167041353)*x_ref[0]+(0.60591609073)*x_ref[1]**o+(0.362583373854)*x_ref[1]+(0.654506836835)*x_ref[2]**o+(-0.144610108143)*x_ref[2]
        ref[3,3,1,0]=(0.246640361949)*x_ref[0]**o+(-0.86358850448)*x_ref[0]+(0.157520564836)*x_ref[1]**o+(-0.059233078017)*x_ref[1]+(0.86570653333)*x_ref[2]**o+(0.685576804797)*x_ref[2]
        ref[3,3,1,1]=(-0.958130500315)*x_ref[0]**o+(-0.259028395499)*x_ref[0]+(0.120316829457)*x_ref[1]**o+(-0.758878478978)*x_ref[1]+(-0.43066120723)*x_ref[2]**o+(0.800327000625)*x_ref[2]
        ref[3,3,2,0]=(0.920887970019)*x_ref[0]**o+(-0.503403737241)*x_ref[0]+(0.0282527472159)*x_ref[1]**o+(-0.167650834072)*x_ref[1]+(0.416811246461)*x_ref[2]**o+(-0.103817262329)*x_ref[2]
        ref[3,3,2,1]=(-0.58133852513)*x_ref[0]**o+(-0.176946132911)*x_ref[0]+(0.0347620468937)*x_ref[1]**o+(-0.973916175156)*x_ref[1]+(0.71011981519)*x_ref[2]**o+(-0.339266963019)*x_ref[2]
        ref[3,4,0,0]=(-0.0120671975843)*x_ref[0]**o+(0.109594354705)*x_ref[0]+(0.151599950487)*x_ref[1]**o+(0.881150957088)*x_ref[1]+(0.959435811252)*x_ref[2]**o+(-0.86774776724)*x_ref[2]
        ref[3,4,0,1]=(-0.0545164901018)*x_ref[0]**o+(-0.987079399801)*x_ref[0]+(0.0168167160052)*x_ref[1]**o+(0.323974666845)*x_ref[1]+(-0.169334724946)*x_ref[2]**o+(-0.95604743937)*x_ref[2]
        ref[3,4,1,0]=(0.497905965869)*x_ref[0]**o+(-0.215618743574)*x_ref[0]+(0.333131793569)*x_ref[1]**o+(0.465071861103)*x_ref[1]+(0.472276780484)*x_ref[2]**o+(-0.587116523215)*x_ref[2]
        ref[3,4,1,1]=(-0.798850078462)*x_ref[0]**o+(-0.385483486673)*x_ref[0]+(-0.954899272011)*x_ref[1]**o+(0.985953307837)*x_ref[1]+(0.5064455397)*x_ref[2]**o+(0.201146433376)*x_ref[2]
        ref[3,4,2,0]=(0.736565587072)*x_ref[0]**o+(-0.383544786329)*x_ref[0]+(0.345205635474)*x_ref[1]**o+(0.118507188601)*x_ref[1]+(0.933209008402)*x_ref[2]**o+(0.00295145110263)*x_ref[2]
        ref[3,4,2,1]=(-0.0892194210246)*x_ref[0]**o+(0.152917048)*x_ref[0]+(0.291092526754)*x_ref[1]**o+(-0.13749152394)*x_ref[1]+(-0.820302461745)*x_ref[2]**o+(-0.971364751692)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.264333591663)*x[0]**o+(-0.884594789709)*x[0]+(0.809768917911)*x[1]**o+(-0.999064121889)*x[1]
        arg[1]=(-0.476663257532)*x[0]**o+(0.703121978294)*x[0]+(0.768222344055)*x[1]**o+(0.0126615637065)*x[1]
        ref[0]=(0.264333591663)*x_ref[0]**o+(-0.884594789709)*x_ref[0]+(0.809768917911)*x_ref[1]**o+(-0.999064121889)*x_ref[1]
        ref[1]=(-0.476663257532)*x_ref[0]**o+(0.703121978294)*x_ref[0]+(0.768222344055)*x_ref[1]**o+(0.0126615637065)*x_ref[1]
      else:
        arg[0]=(-0.814801897766)*x[0]**o+(0.417782597744)*x[0]+(0.354691012333)*x[1]**o+(-0.0529647248706)*x[1]+(-0.455085903639)*x[2]**o+(-0.501751102496)*x[2]
        arg[1]=(0.165354246971)*x[0]**o+(-0.238754573504)*x[0]+(-0.293225479626)*x[1]**o+(-0.42192384305)*x[1]+(-0.109509840118)*x[2]**o+(0.511666964461)*x[2]
        ref[0]=(-0.814801897766)*x_ref[0]**o+(0.417782597744)*x_ref[0]+(0.354691012333)*x_ref[1]**o+(-0.0529647248706)*x_ref[1]+(-0.455085903639)*x_ref[2]**o+(-0.501751102496)*x_ref[2]
        ref[1]=(0.165354246971)*x_ref[0]**o+(-0.238754573504)*x_ref[0]+(-0.293225479626)*x_ref[1]**o+(-0.42192384305)*x_ref[1]+(-0.109509840118)*x_ref[2]**o+(0.511666964461)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.469225177227)*x[0]**o+(-0.312308556752)*x[0]+(0.0207638302484)*x[1]**o+(-0.42312789395)*x[1]
        arg[0,1]=(0.435465800442)*x[0]**o+(0.66981866069)*x[0]+(-0.563664236008)*x[1]**o+(-0.979640187021)*x[1]
        arg[0,2]=(0.761422556363)*x[0]**o+(-0.374824039734)*x[0]+(-0.258990622087)*x[1]**o+(-0.553555723801)*x[1]
        arg[0,3]=(-0.346737956865)*x[0]**o+(-0.509820084012)*x[0]+(-0.270279814718)*x[1]**o+(-0.77787682055)*x[1]
        arg[0,4]=(-0.607151504575)*x[0]**o+(0.0370061792286)*x[0]+(-0.597194755504)*x[1]**o+(0.463748874183)*x[1]
        arg[1,0]=(-0.614871053235)*x[0]**o+(0.984151090349)*x[0]+(0.176558255935)*x[1]**o+(-0.00143697546002)*x[1]
        arg[1,1]=(-0.678550535699)*x[0]**o+(-0.40068020768)*x[0]+(-0.73609233692)*x[1]**o+(0.204641737633)*x[1]
        arg[1,2]=(0.644566611818)*x[0]**o+(0.676838663571)*x[0]+(0.592803079261)*x[1]**o+(-0.257851556005)*x[1]
        arg[1,3]=(-0.811008507599)*x[0]**o+(-0.810627035248)*x[0]+(0.470993697001)*x[1]**o+(0.524157551045)*x[1]
        arg[1,4]=(0.507989473752)*x[0]**o+(0.474629400564)*x[0]+(0.972372960952)*x[1]**o+(-0.713136826767)*x[1]
        arg[2,0]=(-0.18215096122)*x[0]**o+(0.0926167766546)*x[0]+(-0.526818906453)*x[1]**o+(-0.169071065789)*x[1]
        arg[2,1]=(-0.402265903759)*x[0]**o+(-0.460219787922)*x[0]+(0.160493468523)*x[1]**o+(-0.816631415356)*x[1]
        arg[2,2]=(-0.826968582935)*x[0]**o+(-0.477060281375)*x[0]+(-0.753204398647)*x[1]**o+(0.340944419869)*x[1]
        arg[2,3]=(-0.289758403392)*x[0]**o+(-0.216496476621)*x[0]+(-0.0729120551729)*x[1]**o+(0.0206397202194)*x[1]
        arg[2,4]=(0.900190015504)*x[0]**o+(0.554334407718)*x[0]+(-0.188091500011)*x[1]**o+(-0.239344021421)*x[1]
        arg[3,0]=(-0.846253042969)*x[0]**o+(0.212019924467)*x[0]+(-0.209309311993)*x[1]**o+(0.47889463943)*x[1]
        arg[3,1]=(-0.796067196362)*x[0]**o+(0.570882191233)*x[0]+(0.789930258)*x[1]**o+(-0.687939656508)*x[1]
        arg[3,2]=(0.0603011963958)*x[0]**o+(-0.291794860225)*x[0]+(-0.572598612637)*x[1]**o+(0.308585554728)*x[1]
        arg[3,3]=(0.883976313141)*x[0]**o+(0.857027876956)*x[0]+(-0.991113135222)*x[1]**o+(-0.0649848727938)*x[1]
        arg[3,4]=(-0.420604320052)*x[0]**o+(-0.147041915772)*x[0]+(0.696294525095)*x[1]**o+(-0.45791291995)*x[1]
        ref[0,0]=(-0.469225177227)*x_ref[0]**o+(-0.312308556752)*x_ref[0]+(0.0207638302484)*x_ref[1]**o+(-0.42312789395)*x_ref[1]
        ref[0,1]=(0.435465800442)*x_ref[0]**o+(0.66981866069)*x_ref[0]+(-0.563664236008)*x_ref[1]**o+(-0.979640187021)*x_ref[1]
        ref[0,2]=(0.761422556363)*x_ref[0]**o+(-0.374824039734)*x_ref[0]+(-0.258990622087)*x_ref[1]**o+(-0.553555723801)*x_ref[1]
        ref[0,3]=(-0.346737956865)*x_ref[0]**o+(-0.509820084012)*x_ref[0]+(-0.270279814718)*x_ref[1]**o+(-0.77787682055)*x_ref[1]
        ref[0,4]=(-0.607151504575)*x_ref[0]**o+(0.0370061792286)*x_ref[0]+(-0.597194755504)*x_ref[1]**o+(0.463748874183)*x_ref[1]
        ref[1,0]=(-0.614871053235)*x_ref[0]**o+(0.984151090349)*x_ref[0]+(0.176558255935)*x_ref[1]**o+(-0.00143697546002)*x_ref[1]
        ref[1,1]=(-0.678550535699)*x_ref[0]**o+(-0.40068020768)*x_ref[0]+(-0.73609233692)*x_ref[1]**o+(0.204641737633)*x_ref[1]
        ref[1,2]=(0.644566611818)*x_ref[0]**o+(0.676838663571)*x_ref[0]+(0.592803079261)*x_ref[1]**o+(-0.257851556005)*x_ref[1]
        ref[1,3]=(-0.811008507599)*x_ref[0]**o+(-0.810627035248)*x_ref[0]+(0.470993697001)*x_ref[1]**o+(0.524157551045)*x_ref[1]
        ref[1,4]=(0.507989473752)*x_ref[0]**o+(0.474629400564)*x_ref[0]+(0.972372960952)*x_ref[1]**o+(-0.713136826767)*x_ref[1]
        ref[2,0]=(-0.18215096122)*x_ref[0]**o+(0.0926167766546)*x_ref[0]+(-0.526818906453)*x_ref[1]**o+(-0.169071065789)*x_ref[1]
        ref[2,1]=(-0.402265903759)*x_ref[0]**o+(-0.460219787922)*x_ref[0]+(0.160493468523)*x_ref[1]**o+(-0.816631415356)*x_ref[1]
        ref[2,2]=(-0.826968582935)*x_ref[0]**o+(-0.477060281375)*x_ref[0]+(-0.753204398647)*x_ref[1]**o+(0.340944419869)*x_ref[1]
        ref[2,3]=(-0.289758403392)*x_ref[0]**o+(-0.216496476621)*x_ref[0]+(-0.0729120551729)*x_ref[1]**o+(0.0206397202194)*x_ref[1]
        ref[2,4]=(0.900190015504)*x_ref[0]**o+(0.554334407718)*x_ref[0]+(-0.188091500011)*x_ref[1]**o+(-0.239344021421)*x_ref[1]
        ref[3,0]=(-0.846253042969)*x_ref[0]**o+(0.212019924467)*x_ref[0]+(-0.209309311993)*x_ref[1]**o+(0.47889463943)*x_ref[1]
        ref[3,1]=(-0.796067196362)*x_ref[0]**o+(0.570882191233)*x_ref[0]+(0.789930258)*x_ref[1]**o+(-0.687939656508)*x_ref[1]
        ref[3,2]=(0.0603011963958)*x_ref[0]**o+(-0.291794860225)*x_ref[0]+(-0.572598612637)*x_ref[1]**o+(0.308585554728)*x_ref[1]
        ref[3,3]=(0.883976313141)*x_ref[0]**o+(0.857027876956)*x_ref[0]+(-0.991113135222)*x_ref[1]**o+(-0.0649848727938)*x_ref[1]
        ref[3,4]=(-0.420604320052)*x_ref[0]**o+(-0.147041915772)*x_ref[0]+(0.696294525095)*x_ref[1]**o+(-0.45791291995)*x_ref[1]
      else:
        arg[0,0]=(0.650989082369)*x[0]**o+(-0.291482513935)*x[0]+(-0.791203000903)*x[1]**o+(0.297385821221)*x[1]+(-0.169814832171)*x[2]**o+(-0.125360282254)*x[2]
        arg[0,1]=(0.166614460758)*x[0]**o+(0.0840046062434)*x[0]+(-0.114646748006)*x[1]**o+(-0.872705089326)*x[1]+(-0.687707782796)*x[2]**o+(-0.230067474846)*x[2]
        arg[0,2]=(-0.219927466535)*x[0]**o+(-0.466007112543)*x[0]+(-0.255341401047)*x[1]**o+(0.615864283673)*x[1]+(-0.983984041665)*x[2]**o+(0.305409617185)*x[2]
        arg[0,3]=(-0.0879672823913)*x[0]**o+(-0.956382814278)*x[0]+(-0.3091322614)*x[1]**o+(0.147177613425)*x[1]+(-0.98911137256)*x[2]**o+(-0.750791857003)*x[2]
        arg[0,4]=(0.715277956634)*x[0]**o+(-0.73358535985)*x[0]+(0.14587372226)*x[1]**o+(0.411467948058)*x[1]+(0.925673970492)*x[2]**o+(-0.0678728528684)*x[2]
        arg[1,0]=(0.150067062674)*x[0]**o+(-0.522799232667)*x[0]+(-0.256261929148)*x[1]**o+(-0.316346866901)*x[1]+(-0.849523865324)*x[2]**o+(-0.585421363115)*x[2]
        arg[1,1]=(-0.49684130662)*x[0]**o+(0.868986326423)*x[0]+(0.650641361203)*x[1]**o+(0.558696373611)*x[1]+(0.945790183046)*x[2]**o+(0.379250336542)*x[2]
        arg[1,2]=(-0.67349806443)*x[0]**o+(-0.337586770393)*x[0]+(0.719579305241)*x[1]**o+(-0.300711356473)*x[1]+(-0.350130081791)*x[2]**o+(0.604354519926)*x[2]
        arg[1,3]=(-0.242995155153)*x[0]**o+(0.668842006698)*x[0]+(-0.981902121891)*x[1]**o+(-0.358779020262)*x[1]+(0.298801927955)*x[2]**o+(-0.859807798836)*x[2]
        arg[1,4]=(-0.248474519407)*x[0]**o+(-0.309793754496)*x[0]+(-0.967889559474)*x[1]**o+(0.541761608865)*x[1]+(0.573726565934)*x[2]**o+(-0.876387334383)*x[2]
        arg[2,0]=(0.879662395476)*x[0]**o+(0.172656747739)*x[0]+(0.74129137412)*x[1]**o+(0.236748254611)*x[1]+(0.764805785803)*x[2]**o+(-0.257470519211)*x[2]
        arg[2,1]=(0.44326363254)*x[0]**o+(0.121548286191)*x[0]+(0.605969252858)*x[1]**o+(0.261621973968)*x[1]+(-0.694403080782)*x[2]**o+(0.15039482013)*x[2]
        arg[2,2]=(-0.372758098923)*x[0]**o+(0.577481130716)*x[0]+(-0.120284392247)*x[1]**o+(-0.630815000386)*x[1]+(-0.601595664522)*x[2]**o+(0.806894830983)*x[2]
        arg[2,3]=(0.548413470095)*x[0]**o+(0.804535150265)*x[0]+(0.469766192947)*x[1]**o+(-0.796431483578)*x[1]+(0.434732585165)*x[2]**o+(-0.20040808975)*x[2]
        arg[2,4]=(-0.119065000638)*x[0]**o+(-0.651948510575)*x[0]+(0.868418571224)*x[1]**o+(-0.987758488584)*x[1]+(0.495264099949)*x[2]**o+(-0.928684424484)*x[2]
        arg[3,0]=(-0.712239839816)*x[0]**o+(0.90734601523)*x[0]+(0.901855037856)*x[1]**o+(0.773630983477)*x[1]+(0.599873877074)*x[2]**o+(-0.511864534352)*x[2]
        arg[3,1]=(0.0499812109271)*x[0]**o+(0.959791265887)*x[0]+(0.861696669737)*x[1]**o+(-0.887520952646)*x[1]+(-0.609317719697)*x[2]**o+(-0.314327480839)*x[2]
        arg[3,2]=(0.717930038261)*x[0]**o+(-0.480998377737)*x[0]+(0.890930499387)*x[1]**o+(-0.880493180277)*x[1]+(0.540792831261)*x[2]**o+(0.343985055993)*x[2]
        arg[3,3]=(0.20531032325)*x[0]**o+(0.0400856366105)*x[0]+(0.829583965232)*x[1]**o+(0.701555952938)*x[1]+(0.676177356543)*x[2]**o+(0.934250384088)*x[2]
        arg[3,4]=(-0.426083844591)*x[0]**o+(0.0732912754379)*x[0]+(0.555586413705)*x[1]**o+(0.379137165516)*x[1]+(-0.481638055369)*x[2]**o+(-0.156787271652)*x[2]
        ref[0,0]=(0.650989082369)*x_ref[0]**o+(-0.291482513935)*x_ref[0]+(-0.791203000903)*x_ref[1]**o+(0.297385821221)*x_ref[1]+(-0.169814832171)*x_ref[2]**o+(-0.125360282254)*x_ref[2]
        ref[0,1]=(0.166614460758)*x_ref[0]**o+(0.0840046062434)*x_ref[0]+(-0.114646748006)*x_ref[1]**o+(-0.872705089326)*x_ref[1]+(-0.687707782796)*x_ref[2]**o+(-0.230067474846)*x_ref[2]
        ref[0,2]=(-0.219927466535)*x_ref[0]**o+(-0.466007112543)*x_ref[0]+(-0.255341401047)*x_ref[1]**o+(0.615864283673)*x_ref[1]+(-0.983984041665)*x_ref[2]**o+(0.305409617185)*x_ref[2]
        ref[0,3]=(-0.0879672823913)*x_ref[0]**o+(-0.956382814278)*x_ref[0]+(-0.3091322614)*x_ref[1]**o+(0.147177613425)*x_ref[1]+(-0.98911137256)*x_ref[2]**o+(-0.750791857003)*x_ref[2]
        ref[0,4]=(0.715277956634)*x_ref[0]**o+(-0.73358535985)*x_ref[0]+(0.14587372226)*x_ref[1]**o+(0.411467948058)*x_ref[1]+(0.925673970492)*x_ref[2]**o+(-0.0678728528684)*x_ref[2]
        ref[1,0]=(0.150067062674)*x_ref[0]**o+(-0.522799232667)*x_ref[0]+(-0.256261929148)*x_ref[1]**o+(-0.316346866901)*x_ref[1]+(-0.849523865324)*x_ref[2]**o+(-0.585421363115)*x_ref[2]
        ref[1,1]=(-0.49684130662)*x_ref[0]**o+(0.868986326423)*x_ref[0]+(0.650641361203)*x_ref[1]**o+(0.558696373611)*x_ref[1]+(0.945790183046)*x_ref[2]**o+(0.379250336542)*x_ref[2]
        ref[1,2]=(-0.67349806443)*x_ref[0]**o+(-0.337586770393)*x_ref[0]+(0.719579305241)*x_ref[1]**o+(-0.300711356473)*x_ref[1]+(-0.350130081791)*x_ref[2]**o+(0.604354519926)*x_ref[2]
        ref[1,3]=(-0.242995155153)*x_ref[0]**o+(0.668842006698)*x_ref[0]+(-0.981902121891)*x_ref[1]**o+(-0.358779020262)*x_ref[1]+(0.298801927955)*x_ref[2]**o+(-0.859807798836)*x_ref[2]
        ref[1,4]=(-0.248474519407)*x_ref[0]**o+(-0.309793754496)*x_ref[0]+(-0.967889559474)*x_ref[1]**o+(0.541761608865)*x_ref[1]+(0.573726565934)*x_ref[2]**o+(-0.876387334383)*x_ref[2]
        ref[2,0]=(0.879662395476)*x_ref[0]**o+(0.172656747739)*x_ref[0]+(0.74129137412)*x_ref[1]**o+(0.236748254611)*x_ref[1]+(0.764805785803)*x_ref[2]**o+(-0.257470519211)*x_ref[2]
        ref[2,1]=(0.44326363254)*x_ref[0]**o+(0.121548286191)*x_ref[0]+(0.605969252858)*x_ref[1]**o+(0.261621973968)*x_ref[1]+(-0.694403080782)*x_ref[2]**o+(0.15039482013)*x_ref[2]
        ref[2,2]=(-0.372758098923)*x_ref[0]**o+(0.577481130716)*x_ref[0]+(-0.120284392247)*x_ref[1]**o+(-0.630815000386)*x_ref[1]+(-0.601595664522)*x_ref[2]**o+(0.806894830983)*x_ref[2]
        ref[2,3]=(0.548413470095)*x_ref[0]**o+(0.804535150265)*x_ref[0]+(0.469766192947)*x_ref[1]**o+(-0.796431483578)*x_ref[1]+(0.434732585165)*x_ref[2]**o+(-0.20040808975)*x_ref[2]
        ref[2,4]=(-0.119065000638)*x_ref[0]**o+(-0.651948510575)*x_ref[0]+(0.868418571224)*x_ref[1]**o+(-0.987758488584)*x_ref[1]+(0.495264099949)*x_ref[2]**o+(-0.928684424484)*x_ref[2]
        ref[3,0]=(-0.712239839816)*x_ref[0]**o+(0.90734601523)*x_ref[0]+(0.901855037856)*x_ref[1]**o+(0.773630983477)*x_ref[1]+(0.599873877074)*x_ref[2]**o+(-0.511864534352)*x_ref[2]
        ref[3,1]=(0.0499812109271)*x_ref[0]**o+(0.959791265887)*x_ref[0]+(0.861696669737)*x_ref[1]**o+(-0.887520952646)*x_ref[1]+(-0.609317719697)*x_ref[2]**o+(-0.314327480839)*x_ref[2]
        ref[3,2]=(0.717930038261)*x_ref[0]**o+(-0.480998377737)*x_ref[0]+(0.890930499387)*x_ref[1]**o+(-0.880493180277)*x_ref[1]+(0.540792831261)*x_ref[2]**o+(0.343985055993)*x_ref[2]
        ref[3,3]=(0.20531032325)*x_ref[0]**o+(0.0400856366105)*x_ref[0]+(0.829583965232)*x_ref[1]**o+(0.701555952938)*x_ref[1]+(0.676177356543)*x_ref[2]**o+(0.934250384088)*x_ref[2]
        ref[3,4]=(-0.426083844591)*x_ref[0]**o+(0.0732912754379)*x_ref[0]+(0.555586413705)*x_ref[1]**o+(0.379137165516)*x_ref[1]+(-0.481638055369)*x_ref[2]**o+(-0.156787271652)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.0451039988637)*x[0]**o+(0.159426346984)*x[0]+(0.693271264689)*x[1]**o+(-0.129057826038)*x[1]
        arg[0,0,1]=(0.165096887432)*x[0]**o+(0.421267288474)*x[0]+(0.259248346979)*x[1]**o+(-0.216110688389)*x[1]
        arg[0,1,0]=(-0.419500329289)*x[0]**o+(-0.516326782996)*x[0]+(-0.726209291778)*x[1]**o+(0.80434579811)*x[1]
        arg[0,1,1]=(-0.680429758627)*x[0]**o+(-0.539365634112)*x[0]+(-0.20379891956)*x[1]**o+(-0.760436888151)*x[1]
        arg[1,0,0]=(-0.472638888322)*x[0]**o+(0.982469790589)*x[0]+(-0.383590346736)*x[1]**o+(-0.69130786846)*x[1]
        arg[1,0,1]=(0.334498522873)*x[0]**o+(0.190404359107)*x[0]+(0.210624993429)*x[1]**o+(-0.744921567562)*x[1]
        arg[1,1,0]=(0.632591775382)*x[0]**o+(-0.405215541727)*x[0]+(-0.880471123936)*x[1]**o+(0.0942729925938)*x[1]
        arg[1,1,1]=(-0.837885933046)*x[0]**o+(-0.798882152889)*x[0]+(0.648322104814)*x[1]**o+(-0.357620132856)*x[1]
        arg[2,0,0]=(-0.490592533882)*x[0]**o+(0.540431305386)*x[0]+(-0.777299904284)*x[1]**o+(-0.308682146808)*x[1]
        arg[2,0,1]=(0.971265898941)*x[0]**o+(0.546875456819)*x[0]+(0.627171799906)*x[1]**o+(0.0732789392684)*x[1]
        arg[2,1,0]=(0.178076638358)*x[0]**o+(0.587800697151)*x[0]+(0.748336208169)*x[1]**o+(-0.0341258196613)*x[1]
        arg[2,1,1]=(0.36345594982)*x[0]**o+(-0.899402201831)*x[0]+(0.0206286665096)*x[1]**o+(0.303557722727)*x[1]
        arg[3,0,0]=(-0.544119125777)*x[0]**o+(-0.805891689123)*x[0]+(-0.687902925271)*x[1]**o+(0.346236636736)*x[1]
        arg[3,0,1]=(0.725902084271)*x[0]**o+(0.605461404125)*x[0]+(0.452767036006)*x[1]**o+(0.266584718883)*x[1]
        arg[3,1,0]=(0.322052654392)*x[0]**o+(-0.91949636775)*x[0]+(0.948097272559)*x[1]**o+(-0.463715765101)*x[1]
        arg[3,1,1]=(-0.369192142232)*x[0]**o+(0.616001201865)*x[0]+(-0.257989673651)*x[1]**o+(0.507550625249)*x[1]
        arg[4,0,0]=(0.270343836597)*x[0]**o+(0.851686984957)*x[0]+(0.838756000051)*x[1]**o+(-0.329625536762)*x[1]
        arg[4,0,1]=(0.00026238459417)*x[0]**o+(0.0274581707442)*x[0]+(0.770120998708)*x[1]**o+(0.574619799143)*x[1]
        arg[4,1,0]=(-0.709788768284)*x[0]**o+(0.0856634157603)*x[0]+(-0.294662391976)*x[1]**o+(-0.90639308515)*x[1]
        arg[4,1,1]=(0.223381811001)*x[0]**o+(0.941451590223)*x[0]+(0.917865807171)*x[1]**o+(0.912972271206)*x[1]
        arg[5,0,0]=(-0.731163054411)*x[0]**o+(0.185548632151)*x[0]+(0.907527980776)*x[1]**o+(0.816017248525)*x[1]
        arg[5,0,1]=(0.719177781335)*x[0]**o+(-0.788006096642)*x[0]+(-0.434802402615)*x[1]**o+(0.0249465251663)*x[1]
        arg[5,1,0]=(0.982421863977)*x[0]**o+(0.127223012805)*x[0]+(-0.858491508817)*x[1]**o+(-0.152253965468)*x[1]
        arg[5,1,1]=(-0.779596616421)*x[0]**o+(0.049815217296)*x[0]+(0.892121656728)*x[1]**o+(0.258587664554)*x[1]
        ref[0,0,0]=(0.0451039988637)*x_ref[0]**o+(0.159426346984)*x_ref[0]+(0.693271264689)*x_ref[1]**o+(-0.129057826038)*x_ref[1]
        ref[0,0,1]=(0.165096887432)*x_ref[0]**o+(0.421267288474)*x_ref[0]+(0.259248346979)*x_ref[1]**o+(-0.216110688389)*x_ref[1]
        ref[0,1,0]=(-0.419500329289)*x_ref[0]**o+(-0.516326782996)*x_ref[0]+(-0.726209291778)*x_ref[1]**o+(0.80434579811)*x_ref[1]
        ref[0,1,1]=(-0.680429758627)*x_ref[0]**o+(-0.539365634112)*x_ref[0]+(-0.20379891956)*x_ref[1]**o+(-0.760436888151)*x_ref[1]
        ref[1,0,0]=(-0.472638888322)*x_ref[0]**o+(0.982469790589)*x_ref[0]+(-0.383590346736)*x_ref[1]**o+(-0.69130786846)*x_ref[1]
        ref[1,0,1]=(0.334498522873)*x_ref[0]**o+(0.190404359107)*x_ref[0]+(0.210624993429)*x_ref[1]**o+(-0.744921567562)*x_ref[1]
        ref[1,1,0]=(0.632591775382)*x_ref[0]**o+(-0.405215541727)*x_ref[0]+(-0.880471123936)*x_ref[1]**o+(0.0942729925938)*x_ref[1]
        ref[1,1,1]=(-0.837885933046)*x_ref[0]**o+(-0.798882152889)*x_ref[0]+(0.648322104814)*x_ref[1]**o+(-0.357620132856)*x_ref[1]
        ref[2,0,0]=(-0.490592533882)*x_ref[0]**o+(0.540431305386)*x_ref[0]+(-0.777299904284)*x_ref[1]**o+(-0.308682146808)*x_ref[1]
        ref[2,0,1]=(0.971265898941)*x_ref[0]**o+(0.546875456819)*x_ref[0]+(0.627171799906)*x_ref[1]**o+(0.0732789392684)*x_ref[1]
        ref[2,1,0]=(0.178076638358)*x_ref[0]**o+(0.587800697151)*x_ref[0]+(0.748336208169)*x_ref[1]**o+(-0.0341258196613)*x_ref[1]
        ref[2,1,1]=(0.36345594982)*x_ref[0]**o+(-0.899402201831)*x_ref[0]+(0.0206286665096)*x_ref[1]**o+(0.303557722727)*x_ref[1]
        ref[3,0,0]=(-0.544119125777)*x_ref[0]**o+(-0.805891689123)*x_ref[0]+(-0.687902925271)*x_ref[1]**o+(0.346236636736)*x_ref[1]
        ref[3,0,1]=(0.725902084271)*x_ref[0]**o+(0.605461404125)*x_ref[0]+(0.452767036006)*x_ref[1]**o+(0.266584718883)*x_ref[1]
        ref[3,1,0]=(0.322052654392)*x_ref[0]**o+(-0.91949636775)*x_ref[0]+(0.948097272559)*x_ref[1]**o+(-0.463715765101)*x_ref[1]
        ref[3,1,1]=(-0.369192142232)*x_ref[0]**o+(0.616001201865)*x_ref[0]+(-0.257989673651)*x_ref[1]**o+(0.507550625249)*x_ref[1]
        ref[4,0,0]=(0.270343836597)*x_ref[0]**o+(0.851686984957)*x_ref[0]+(0.838756000051)*x_ref[1]**o+(-0.329625536762)*x_ref[1]
        ref[4,0,1]=(0.00026238459417)*x_ref[0]**o+(0.0274581707442)*x_ref[0]+(0.770120998708)*x_ref[1]**o+(0.574619799143)*x_ref[1]
        ref[4,1,0]=(-0.709788768284)*x_ref[0]**o+(0.0856634157603)*x_ref[0]+(-0.294662391976)*x_ref[1]**o+(-0.90639308515)*x_ref[1]
        ref[4,1,1]=(0.223381811001)*x_ref[0]**o+(0.941451590223)*x_ref[0]+(0.917865807171)*x_ref[1]**o+(0.912972271206)*x_ref[1]
        ref[5,0,0]=(-0.731163054411)*x_ref[0]**o+(0.185548632151)*x_ref[0]+(0.907527980776)*x_ref[1]**o+(0.816017248525)*x_ref[1]
        ref[5,0,1]=(0.719177781335)*x_ref[0]**o+(-0.788006096642)*x_ref[0]+(-0.434802402615)*x_ref[1]**o+(0.0249465251663)*x_ref[1]
        ref[5,1,0]=(0.982421863977)*x_ref[0]**o+(0.127223012805)*x_ref[0]+(-0.858491508817)*x_ref[1]**o+(-0.152253965468)*x_ref[1]
        ref[5,1,1]=(-0.779596616421)*x_ref[0]**o+(0.049815217296)*x_ref[0]+(0.892121656728)*x_ref[1]**o+(0.258587664554)*x_ref[1]
      else:
        arg[0,0,0]=(0.804406156662)*x[0]**o+(0.304298949556)*x[0]+(-9.23949223481e-05)*x[1]**o+(0.521245437597)*x[1]+(0.822885620595)*x[2]**o+(0.44472437921)*x[2]
        arg[0,0,1]=(0.208119461011)*x[0]**o+(-0.967847500694)*x[0]+(-0.182961335156)*x[1]**o+(-0.0510538508635)*x[1]+(0.932285255322)*x[2]**o+(-0.463525963095)*x[2]
        arg[0,1,0]=(-0.0514207812926)*x[0]**o+(-0.730723672364)*x[0]+(-0.471144684451)*x[1]**o+(0.0359308791998)*x[1]+(-0.935079157697)*x[2]**o+(0.841965819816)*x[2]
        arg[0,1,1]=(0.607427052267)*x[0]**o+(-0.344216102016)*x[0]+(0.967271353509)*x[1]**o+(-0.412288382331)*x[1]+(-0.0716617724983)*x[2]**o+(0.56895080823)*x[2]
        arg[1,0,0]=(-0.574314769591)*x[0]**o+(-0.819434106932)*x[0]+(0.143229095273)*x[1]**o+(0.505343958082)*x[1]+(0.497764227469)*x[2]**o+(0.592609757968)*x[2]
        arg[1,0,1]=(0.414984900558)*x[0]**o+(0.624574053857)*x[0]+(-0.367330776954)*x[1]**o+(-0.30093924896)*x[1]+(-0.99416380749)*x[2]**o+(0.492443599924)*x[2]
        arg[1,1,0]=(0.171673119954)*x[0]**o+(0.612986140068)*x[0]+(-0.671355907363)*x[1]**o+(-0.292080539973)*x[1]+(-0.450890819125)*x[2]**o+(-0.64742047997)*x[2]
        arg[1,1,1]=(0.452494887295)*x[0]**o+(0.187308285528)*x[0]+(-0.995316825981)*x[1]**o+(-0.825194018202)*x[1]+(-0.00651224511881)*x[2]**o+(-0.715747547803)*x[2]
        arg[2,0,0]=(0.546297169466)*x[0]**o+(-0.424626884041)*x[0]+(0.327647062351)*x[1]**o+(-0.050651105527)*x[1]+(-0.667117310429)*x[2]**o+(-0.37659304263)*x[2]
        arg[2,0,1]=(-0.323595958622)*x[0]**o+(0.541718593289)*x[0]+(0.6285998056)*x[1]**o+(0.488539567593)*x[1]+(0.983086232412)*x[2]**o+(0.692821558469)*x[2]
        arg[2,1,0]=(0.16009856161)*x[0]**o+(-0.776069988482)*x[0]+(0.750717144319)*x[1]**o+(-0.244287895989)*x[1]+(0.345242766297)*x[2]**o+(0.572063651895)*x[2]
        arg[2,1,1]=(-0.34626843638)*x[0]**o+(0.197125033681)*x[0]+(-0.570015603866)*x[1]**o+(-0.633161422872)*x[1]+(-0.602998606808)*x[2]**o+(-0.674829268895)*x[2]
        arg[3,0,0]=(-0.659696747189)*x[0]**o+(-0.571128578512)*x[0]+(-0.704414117225)*x[1]**o+(-0.334924171135)*x[1]+(-0.0200665796175)*x[2]**o+(-0.660688341854)*x[2]
        arg[3,0,1]=(-0.497019087713)*x[0]**o+(0.25850813993)*x[0]+(-0.713426904693)*x[1]**o+(-0.994061006282)*x[1]+(-0.669422282701)*x[2]**o+(-0.553981025574)*x[2]
        arg[3,1,0]=(-0.151249618276)*x[0]**o+(0.29432574729)*x[0]+(0.173047919293)*x[1]**o+(0.844456199843)*x[1]+(0.845402272715)*x[2]**o+(-0.115698924822)*x[2]
        arg[3,1,1]=(0.840440627264)*x[0]**o+(0.593400475213)*x[0]+(0.974677794562)*x[1]**o+(0.938845603167)*x[1]+(-0.944730436891)*x[2]**o+(-0.657543633218)*x[2]
        arg[4,0,0]=(0.968290161629)*x[0]**o+(0.578471458995)*x[0]+(0.946558149029)*x[1]**o+(-0.232692317459)*x[1]+(-0.97762469848)*x[2]**o+(0.660519364613)*x[2]
        arg[4,0,1]=(-0.187317070761)*x[0]**o+(-0.0363170179961)*x[0]+(0.278499613269)*x[1]**o+(-0.0685831047696)*x[1]+(-0.540282212359)*x[2]**o+(0.564656560066)*x[2]
        arg[4,1,0]=(-0.211825530749)*x[0]**o+(0.225546568724)*x[0]+(-0.83241974026)*x[1]**o+(-0.503828393954)*x[1]+(-0.635937384108)*x[2]**o+(-0.513183062177)*x[2]
        arg[4,1,1]=(-0.257123983215)*x[0]**o+(-0.0700924567638)*x[0]+(0.550859393713)*x[1]**o+(0.132070212001)*x[1]+(-0.478549332853)*x[2]**o+(-0.560087260102)*x[2]
        arg[5,0,0]=(0.677259763344)*x[0]**o+(-0.213422444013)*x[0]+(-0.294112138735)*x[1]**o+(0.083135232348)*x[1]+(-0.899953896237)*x[2]**o+(-0.117143433064)*x[2]
        arg[5,0,1]=(-0.489822592306)*x[0]**o+(0.368217759904)*x[0]+(-0.927837261915)*x[1]**o+(-0.708243327427)*x[1]+(-0.132256551163)*x[2]**o+(0.316204052014)*x[2]
        arg[5,1,0]=(0.314897334392)*x[0]**o+(-0.848168680639)*x[0]+(0.92421318346)*x[1]**o+(0.995505186786)*x[1]+(0.372381949158)*x[2]**o+(-0.563415970977)*x[2]
        arg[5,1,1]=(0.745011483885)*x[0]**o+(-0.6423283108)*x[0]+(-0.119592965986)*x[1]**o+(0.241068250891)*x[1]+(0.477172198315)*x[2]**o+(-0.468580220406)*x[2]
        ref[0,0,0]=(0.804406156662)*x_ref[0]**o+(0.304298949556)*x_ref[0]+(-9.23949223481e-05)*x_ref[1]**o+(0.521245437597)*x_ref[1]+(0.822885620595)*x_ref[2]**o+(0.44472437921)*x_ref[2]
        ref[0,0,1]=(0.208119461011)*x_ref[0]**o+(-0.967847500694)*x_ref[0]+(-0.182961335156)*x_ref[1]**o+(-0.0510538508635)*x_ref[1]+(0.932285255322)*x_ref[2]**o+(-0.463525963095)*x_ref[2]
        ref[0,1,0]=(-0.0514207812926)*x_ref[0]**o+(-0.730723672364)*x_ref[0]+(-0.471144684451)*x_ref[1]**o+(0.0359308791998)*x_ref[1]+(-0.935079157697)*x_ref[2]**o+(0.841965819816)*x_ref[2]
        ref[0,1,1]=(0.607427052267)*x_ref[0]**o+(-0.344216102016)*x_ref[0]+(0.967271353509)*x_ref[1]**o+(-0.412288382331)*x_ref[1]+(-0.0716617724983)*x_ref[2]**o+(0.56895080823)*x_ref[2]
        ref[1,0,0]=(-0.574314769591)*x_ref[0]**o+(-0.819434106932)*x_ref[0]+(0.143229095273)*x_ref[1]**o+(0.505343958082)*x_ref[1]+(0.497764227469)*x_ref[2]**o+(0.592609757968)*x_ref[2]
        ref[1,0,1]=(0.414984900558)*x_ref[0]**o+(0.624574053857)*x_ref[0]+(-0.367330776954)*x_ref[1]**o+(-0.30093924896)*x_ref[1]+(-0.99416380749)*x_ref[2]**o+(0.492443599924)*x_ref[2]
        ref[1,1,0]=(0.171673119954)*x_ref[0]**o+(0.612986140068)*x_ref[0]+(-0.671355907363)*x_ref[1]**o+(-0.292080539973)*x_ref[1]+(-0.450890819125)*x_ref[2]**o+(-0.64742047997)*x_ref[2]
        ref[1,1,1]=(0.452494887295)*x_ref[0]**o+(0.187308285528)*x_ref[0]+(-0.995316825981)*x_ref[1]**o+(-0.825194018202)*x_ref[1]+(-0.00651224511881)*x_ref[2]**o+(-0.715747547803)*x_ref[2]
        ref[2,0,0]=(0.546297169466)*x_ref[0]**o+(-0.424626884041)*x_ref[0]+(0.327647062351)*x_ref[1]**o+(-0.050651105527)*x_ref[1]+(-0.667117310429)*x_ref[2]**o+(-0.37659304263)*x_ref[2]
        ref[2,0,1]=(-0.323595958622)*x_ref[0]**o+(0.541718593289)*x_ref[0]+(0.6285998056)*x_ref[1]**o+(0.488539567593)*x_ref[1]+(0.983086232412)*x_ref[2]**o+(0.692821558469)*x_ref[2]
        ref[2,1,0]=(0.16009856161)*x_ref[0]**o+(-0.776069988482)*x_ref[0]+(0.750717144319)*x_ref[1]**o+(-0.244287895989)*x_ref[1]+(0.345242766297)*x_ref[2]**o+(0.572063651895)*x_ref[2]
        ref[2,1,1]=(-0.34626843638)*x_ref[0]**o+(0.197125033681)*x_ref[0]+(-0.570015603866)*x_ref[1]**o+(-0.633161422872)*x_ref[1]+(-0.602998606808)*x_ref[2]**o+(-0.674829268895)*x_ref[2]
        ref[3,0,0]=(-0.659696747189)*x_ref[0]**o+(-0.571128578512)*x_ref[0]+(-0.704414117225)*x_ref[1]**o+(-0.334924171135)*x_ref[1]+(-0.0200665796175)*x_ref[2]**o+(-0.660688341854)*x_ref[2]
        ref[3,0,1]=(-0.497019087713)*x_ref[0]**o+(0.25850813993)*x_ref[0]+(-0.713426904693)*x_ref[1]**o+(-0.994061006282)*x_ref[1]+(-0.669422282701)*x_ref[2]**o+(-0.553981025574)*x_ref[2]
        ref[3,1,0]=(-0.151249618276)*x_ref[0]**o+(0.29432574729)*x_ref[0]+(0.173047919293)*x_ref[1]**o+(0.844456199843)*x_ref[1]+(0.845402272715)*x_ref[2]**o+(-0.115698924822)*x_ref[2]
        ref[3,1,1]=(0.840440627264)*x_ref[0]**o+(0.593400475213)*x_ref[0]+(0.974677794562)*x_ref[1]**o+(0.938845603167)*x_ref[1]+(-0.944730436891)*x_ref[2]**o+(-0.657543633218)*x_ref[2]
        ref[4,0,0]=(0.968290161629)*x_ref[0]**o+(0.578471458995)*x_ref[0]+(0.946558149029)*x_ref[1]**o+(-0.232692317459)*x_ref[1]+(-0.97762469848)*x_ref[2]**o+(0.660519364613)*x_ref[2]
        ref[4,0,1]=(-0.187317070761)*x_ref[0]**o+(-0.0363170179961)*x_ref[0]+(0.278499613269)*x_ref[1]**o+(-0.0685831047696)*x_ref[1]+(-0.540282212359)*x_ref[2]**o+(0.564656560066)*x_ref[2]
        ref[4,1,0]=(-0.211825530749)*x_ref[0]**o+(0.225546568724)*x_ref[0]+(-0.83241974026)*x_ref[1]**o+(-0.503828393954)*x_ref[1]+(-0.635937384108)*x_ref[2]**o+(-0.513183062177)*x_ref[2]
        ref[4,1,1]=(-0.257123983215)*x_ref[0]**o+(-0.0700924567638)*x_ref[0]+(0.550859393713)*x_ref[1]**o+(0.132070212001)*x_ref[1]+(-0.478549332853)*x_ref[2]**o+(-0.560087260102)*x_ref[2]
        ref[5,0,0]=(0.677259763344)*x_ref[0]**o+(-0.213422444013)*x_ref[0]+(-0.294112138735)*x_ref[1]**o+(0.083135232348)*x_ref[1]+(-0.899953896237)*x_ref[2]**o+(-0.117143433064)*x_ref[2]
        ref[5,0,1]=(-0.489822592306)*x_ref[0]**o+(0.368217759904)*x_ref[0]+(-0.927837261915)*x_ref[1]**o+(-0.708243327427)*x_ref[1]+(-0.132256551163)*x_ref[2]**o+(0.316204052014)*x_ref[2]
        ref[5,1,0]=(0.314897334392)*x_ref[0]**o+(-0.848168680639)*x_ref[0]+(0.92421318346)*x_ref[1]**o+(0.995505186786)*x_ref[1]+(0.372381949158)*x_ref[2]**o+(-0.563415970977)*x_ref[2]
        ref[5,1,1]=(0.745011483885)*x_ref[0]**o+(-0.6423283108)*x_ref[0]+(-0.119592965986)*x_ref[1]**o+(0.241068250891)*x_ref[1]+(0.477172198315)*x_ref[2]**o+(-0.468580220406)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.633576516648)*x[0]**o+(0.77399256211)*x[0]+(-0.396696417195)*x[1]**o+(-0.604572540725)*x[1]
        arg[0,0,0,1]=(0.612672991978)*x[0]**o+(-0.331012302611)*x[0]+(0.885533578235)*x[1]**o+(0.334686818058)*x[1]
        arg[0,0,1,0]=(-0.982701069025)*x[0]**o+(0.467849033642)*x[0]+(0.525392773375)*x[1]**o+(0.355622324425)*x[1]
        arg[0,0,1,1]=(-0.0243671469666)*x[0]**o+(0.790044842383)*x[0]+(0.413384172033)*x[1]**o+(0.740805135968)*x[1]
        arg[0,0,2,0]=(0.0384789957691)*x[0]**o+(-0.317009899417)*x[0]+(0.0389710369615)*x[1]**o+(0.784549852598)*x[1]
        arg[0,0,2,1]=(0.361832259252)*x[0]**o+(0.410815792944)*x[0]+(0.39496145831)*x[1]**o+(-0.820784143964)*x[1]
        arg[0,1,0,0]=(0.224261088976)*x[0]**o+(-0.824694527256)*x[0]+(0.318674472135)*x[1]**o+(0.827696354338)*x[1]
        arg[0,1,0,1]=(-0.442154901359)*x[0]**o+(0.141775320862)*x[0]+(0.688711841268)*x[1]**o+(-0.27047542755)*x[1]
        arg[0,1,1,0]=(-0.570125802019)*x[0]**o+(0.050303762895)*x[0]+(-0.972329837417)*x[1]**o+(0.143700303765)*x[1]
        arg[0,1,1,1]=(0.852888550499)*x[0]**o+(-0.51216090329)*x[0]+(-0.895663922335)*x[1]**o+(-0.183466225547)*x[1]
        arg[0,1,2,0]=(0.863840920739)*x[0]**o+(-0.247789353524)*x[0]+(0.63056436157)*x[1]**o+(0.932522589942)*x[1]
        arg[0,1,2,1]=(-0.403832121794)*x[0]**o+(-0.769625259789)*x[0]+(0.507724303191)*x[1]**o+(0.314178775526)*x[1]
        arg[0,2,0,0]=(0.543867153515)*x[0]**o+(-0.414107817028)*x[0]+(0.423922720842)*x[1]**o+(0.166632407916)*x[1]
        arg[0,2,0,1]=(0.731968209929)*x[0]**o+(-0.114742736683)*x[0]+(0.112562415021)*x[1]**o+(0.166924480325)*x[1]
        arg[0,2,1,0]=(0.919472066242)*x[0]**o+(0.560565315205)*x[0]+(0.4967650028)*x[1]**o+(0.708376671625)*x[1]
        arg[0,2,1,1]=(0.147500239739)*x[0]**o+(-0.632670389919)*x[0]+(-0.992479063437)*x[1]**o+(0.258941122962)*x[1]
        arg[0,2,2,0]=(0.474481785329)*x[0]**o+(0.592902604094)*x[0]+(0.64058685211)*x[1]**o+(-0.477843307901)*x[1]
        arg[0,2,2,1]=(0.769898355146)*x[0]**o+(0.563458279487)*x[0]+(0.876276425449)*x[1]**o+(-0.0625727749869)*x[1]
        arg[0,3,0,0]=(-0.330055037059)*x[0]**o+(-0.727656105163)*x[0]+(-0.559901684392)*x[1]**o+(-0.433646878263)*x[1]
        arg[0,3,0,1]=(-0.725314328666)*x[0]**o+(0.638326609105)*x[0]+(-0.126619368417)*x[1]**o+(-0.232194758676)*x[1]
        arg[0,3,1,0]=(0.318096005545)*x[0]**o+(0.325247494289)*x[0]+(-0.514305253643)*x[1]**o+(0.108523350086)*x[1]
        arg[0,3,1,1]=(0.298034894956)*x[0]**o+(0.825714393651)*x[0]+(-0.67162850079)*x[1]**o+(0.681809083004)*x[1]
        arg[0,3,2,0]=(-0.00715890582648)*x[0]**o+(0.893900731227)*x[0]+(-0.773710274545)*x[1]**o+(0.415913728569)*x[1]
        arg[0,3,2,1]=(0.0853551940958)*x[0]**o+(-0.228876011127)*x[0]+(0.87699456762)*x[1]**o+(-0.744605229253)*x[1]
        arg[0,4,0,0]=(0.942646937302)*x[0]**o+(0.92717236759)*x[0]+(0.794806821569)*x[1]**o+(0.848766303526)*x[1]
        arg[0,4,0,1]=(0.20140672561)*x[0]**o+(0.974490372414)*x[0]+(-0.450332419587)*x[1]**o+(-0.859102184782)*x[1]
        arg[0,4,1,0]=(0.632153053193)*x[0]**o+(0.702336756277)*x[0]+(-0.571496076236)*x[1]**o+(-0.882472852219)*x[1]
        arg[0,4,1,1]=(0.366452829499)*x[0]**o+(-0.514660164741)*x[0]+(-0.152794832253)*x[1]**o+(-0.658688117796)*x[1]
        arg[0,4,2,0]=(0.406218394702)*x[0]**o+(0.782359698198)*x[0]+(-0.0582499587562)*x[1]**o+(0.664506852595)*x[1]
        arg[0,4,2,1]=(0.768240870343)*x[0]**o+(0.188470592904)*x[0]+(0.855933509171)*x[1]**o+(0.639174087541)*x[1]
        arg[1,0,0,0]=(-0.946841398046)*x[0]**o+(0.892729256183)*x[0]+(-0.15644094115)*x[1]**o+(-0.859371368233)*x[1]
        arg[1,0,0,1]=(0.238113647666)*x[0]**o+(0.173695470467)*x[0]+(-0.753529252313)*x[1]**o+(-0.243203902463)*x[1]
        arg[1,0,1,0]=(0.267937523443)*x[0]**o+(0.600200492374)*x[0]+(0.442653529257)*x[1]**o+(-0.394984210541)*x[1]
        arg[1,0,1,1]=(-0.495022907622)*x[0]**o+(-0.593273808093)*x[0]+(0.888764746608)*x[1]**o+(-0.745127825763)*x[1]
        arg[1,0,2,0]=(0.352827238496)*x[0]**o+(0.293200928839)*x[0]+(-0.00369812109034)*x[1]**o+(0.478875774193)*x[1]
        arg[1,0,2,1]=(-0.403358878981)*x[0]**o+(0.756947364667)*x[0]+(0.649335795039)*x[1]**o+(-0.339962796859)*x[1]
        arg[1,1,0,0]=(-0.607505349015)*x[0]**o+(-0.392535399582)*x[0]+(0.799782389634)*x[1]**o+(0.186339165863)*x[1]
        arg[1,1,0,1]=(0.997492304627)*x[0]**o+(-0.281456622581)*x[0]+(-0.959294693613)*x[1]**o+(-0.454812606674)*x[1]
        arg[1,1,1,0]=(0.454682037549)*x[0]**o+(0.797129935647)*x[0]+(-0.949539573053)*x[1]**o+(-0.918499382199)*x[1]
        arg[1,1,1,1]=(0.903016449758)*x[0]**o+(-0.575039046836)*x[0]+(0.957158277074)*x[1]**o+(-0.715797460906)*x[1]
        arg[1,1,2,0]=(0.92005547899)*x[0]**o+(-0.716975919302)*x[0]+(-0.900904311708)*x[1]**o+(0.989458809643)*x[1]
        arg[1,1,2,1]=(-0.633214991087)*x[0]**o+(0.680658578627)*x[0]+(0.552194054556)*x[1]**o+(0.343557072769)*x[1]
        arg[1,2,0,0]=(-0.0182903530463)*x[0]**o+(-0.0445569497267)*x[0]+(-0.259920069365)*x[1]**o+(0.118765787279)*x[1]
        arg[1,2,0,1]=(-0.667382174733)*x[0]**o+(-0.79898800024)*x[0]+(-0.217638556567)*x[1]**o+(-0.0288504127936)*x[1]
        arg[1,2,1,0]=(-0.437982445862)*x[0]**o+(-0.104054454808)*x[0]+(0.711429718088)*x[1]**o+(-0.607226296329)*x[1]
        arg[1,2,1,1]=(0.252861218037)*x[0]**o+(-0.635701456504)*x[0]+(-0.555222304592)*x[1]**o+(0.723744010092)*x[1]
        arg[1,2,2,0]=(-0.73532710885)*x[0]**o+(0.600599078741)*x[0]+(0.0465795173194)*x[1]**o+(0.524165544645)*x[1]
        arg[1,2,2,1]=(-0.025246019744)*x[0]**o+(-0.376526675619)*x[0]+(0.213011788698)*x[1]**o+(-0.477018225777)*x[1]
        arg[1,3,0,0]=(-0.52682690957)*x[0]**o+(0.0567472858499)*x[0]+(0.574116256102)*x[1]**o+(-0.130012791898)*x[1]
        arg[1,3,0,1]=(0.113502182936)*x[0]**o+(0.7643861313)*x[0]+(0.48414948847)*x[1]**o+(0.863168250318)*x[1]
        arg[1,3,1,0]=(-0.208102162378)*x[0]**o+(0.327770728981)*x[0]+(0.843778895278)*x[1]**o+(0.474843504177)*x[1]
        arg[1,3,1,1]=(-0.549818192184)*x[0]**o+(-0.863333816685)*x[0]+(-0.365631583957)*x[1]**o+(-0.491320441182)*x[1]
        arg[1,3,2,0]=(-0.265833845759)*x[0]**o+(-0.94416746081)*x[0]+(0.48465442593)*x[1]**o+(0.258284621408)*x[1]
        arg[1,3,2,1]=(-0.625716082268)*x[0]**o+(-0.234697130359)*x[0]+(-0.973704406131)*x[1]**o+(0.00141746117847)*x[1]
        arg[1,4,0,0]=(0.565044361964)*x[0]**o+(-0.0672474991748)*x[0]+(-0.125901930524)*x[1]**o+(-0.753103899732)*x[1]
        arg[1,4,0,1]=(-0.979743467289)*x[0]**o+(-0.25225085214)*x[0]+(-0.141635164635)*x[1]**o+(0.378153168667)*x[1]
        arg[1,4,1,0]=(0.97728399206)*x[0]**o+(-0.202043160532)*x[0]+(0.997445424954)*x[1]**o+(-0.100049830154)*x[1]
        arg[1,4,1,1]=(0.364840105486)*x[0]**o+(-0.682234327437)*x[0]+(0.252584592259)*x[1]**o+(-0.0785430242425)*x[1]
        arg[1,4,2,0]=(-0.0168764722616)*x[0]**o+(-0.997460139026)*x[0]+(0.560548548912)*x[1]**o+(0.220532224022)*x[1]
        arg[1,4,2,1]=(0.307927943715)*x[0]**o+(0.496390872708)*x[0]+(0.982750813274)*x[1]**o+(0.392968327778)*x[1]
        arg[2,0,0,0]=(0.645991188948)*x[0]**o+(0.104035822886)*x[0]+(0.18285829097)*x[1]**o+(0.51186286294)*x[1]
        arg[2,0,0,1]=(-0.996971717318)*x[0]**o+(0.126488047141)*x[0]+(0.123585312562)*x[1]**o+(-0.775854123529)*x[1]
        arg[2,0,1,0]=(-0.0258590124655)*x[0]**o+(0.0079454735671)*x[0]+(-0.42819024422)*x[1]**o+(0.12377360967)*x[1]
        arg[2,0,1,1]=(0.990441351225)*x[0]**o+(-0.453806376472)*x[0]+(0.134316618645)*x[1]**o+(-0.902666953282)*x[1]
        arg[2,0,2,0]=(0.149304385108)*x[0]**o+(-0.226132046999)*x[0]+(0.647186106203)*x[1]**o+(-0.269247029225)*x[1]
        arg[2,0,2,1]=(-0.159068164422)*x[0]**o+(-0.392275577366)*x[0]+(0.131423093404)*x[1]**o+(-0.353311282658)*x[1]
        arg[2,1,0,0]=(-0.2862918681)*x[0]**o+(-0.392264098939)*x[0]+(-0.344033237285)*x[1]**o+(0.644261713963)*x[1]
        arg[2,1,0,1]=(0.398022911658)*x[0]**o+(-0.84552546032)*x[0]+(-0.823861260943)*x[1]**o+(0.739994988755)*x[1]
        arg[2,1,1,0]=(-0.696854341234)*x[0]**o+(-0.810851183298)*x[0]+(-0.712182042896)*x[1]**o+(0.594887691783)*x[1]
        arg[2,1,1,1]=(-0.182583840117)*x[0]**o+(0.962594377088)*x[0]+(-0.0157575661453)*x[1]**o+(-0.432167995007)*x[1]
        arg[2,1,2,0]=(0.293029196216)*x[0]**o+(0.46839918212)*x[0]+(0.262375931577)*x[1]**o+(0.0348536956306)*x[1]
        arg[2,1,2,1]=(0.775632135095)*x[0]**o+(-0.874598534107)*x[0]+(0.711602976736)*x[1]**o+(0.37065543293)*x[1]
        arg[2,2,0,0]=(-0.103625214158)*x[0]**o+(0.553558117877)*x[0]+(-0.0522465477589)*x[1]**o+(-0.747139195529)*x[1]
        arg[2,2,0,1]=(0.507427663615)*x[0]**o+(-0.536361186134)*x[0]+(0.809054726193)*x[1]**o+(-0.979546347794)*x[1]
        arg[2,2,1,0]=(0.786210623833)*x[0]**o+(0.244510127091)*x[0]+(0.159799334111)*x[1]**o+(0.0499263839273)*x[1]
        arg[2,2,1,1]=(0.454205244371)*x[0]**o+(-0.650927822092)*x[0]+(-0.939252656087)*x[1]**o+(-0.894014755314)*x[1]
        arg[2,2,2,0]=(-0.829617920904)*x[0]**o+(0.764070900994)*x[0]+(0.972419443556)*x[1]**o+(-0.824913543157)*x[1]
        arg[2,2,2,1]=(0.694310748478)*x[0]**o+(-0.191413364091)*x[0]+(0.588258464472)*x[1]**o+(0.743973223755)*x[1]
        arg[2,3,0,0]=(-0.588397365184)*x[0]**o+(-0.634207034861)*x[0]+(0.0853498555546)*x[1]**o+(0.559933556865)*x[1]
        arg[2,3,0,1]=(0.626829165689)*x[0]**o+(0.942185296616)*x[0]+(0.0360770405565)*x[1]**o+(0.635966218547)*x[1]
        arg[2,3,1,0]=(-0.0103526873852)*x[0]**o+(0.546330663496)*x[0]+(0.799224103454)*x[1]**o+(0.791922324897)*x[1]
        arg[2,3,1,1]=(0.921564474495)*x[0]**o+(-0.420410835144)*x[0]+(-0.20482503843)*x[1]**o+(-0.483331643037)*x[1]
        arg[2,3,2,0]=(-0.903600870314)*x[0]**o+(0.901956716274)*x[0]+(-0.113926342023)*x[1]**o+(-0.061596184768)*x[1]
        arg[2,3,2,1]=(0.653250267262)*x[0]**o+(-0.13770933726)*x[0]+(-0.0791563636079)*x[1]**o+(-0.578313150893)*x[1]
        arg[2,4,0,0]=(0.993903405725)*x[0]**o+(0.694324226249)*x[0]+(0.203872883541)*x[1]**o+(-0.526880767264)*x[1]
        arg[2,4,0,1]=(0.0120058513647)*x[0]**o+(0.00907603337269)*x[0]+(0.0247958601677)*x[1]**o+(-0.996916120875)*x[1]
        arg[2,4,1,0]=(0.281518049478)*x[0]**o+(0.0773799323777)*x[0]+(-0.968565098773)*x[1]**o+(-0.275558076229)*x[1]
        arg[2,4,1,1]=(0.565275744757)*x[0]**o+(0.978636675869)*x[0]+(-0.352951657751)*x[1]**o+(-0.54170487914)*x[1]
        arg[2,4,2,0]=(-0.571207550678)*x[0]**o+(0.196754283318)*x[0]+(0.665558917127)*x[1]**o+(0.361915269355)*x[1]
        arg[2,4,2,1]=(0.54579191885)*x[0]**o+(0.798319371798)*x[0]+(0.961669996133)*x[1]**o+(-0.549305529342)*x[1]
        arg[3,0,0,0]=(0.264064946121)*x[0]**o+(-0.314396185593)*x[0]+(-0.017492206135)*x[1]**o+(0.190513393091)*x[1]
        arg[3,0,0,1]=(0.812213666919)*x[0]**o+(-0.757919357489)*x[0]+(0.294829087285)*x[1]**o+(0.916181101963)*x[1]
        arg[3,0,1,0]=(-0.529906091574)*x[0]**o+(0.553682716765)*x[0]+(-0.693177625995)*x[1]**o+(0.377145561683)*x[1]
        arg[3,0,1,1]=(-0.298942248169)*x[0]**o+(0.876878593705)*x[0]+(-0.609807982832)*x[1]**o+(-0.574954919539)*x[1]
        arg[3,0,2,0]=(-0.729324018245)*x[0]**o+(-0.495194565254)*x[0]+(-0.616162751356)*x[1]**o+(0.878549207376)*x[1]
        arg[3,0,2,1]=(0.625373988248)*x[0]**o+(0.106736108396)*x[0]+(-0.871225833944)*x[1]**o+(0.840614493429)*x[1]
        arg[3,1,0,0]=(-0.90880839589)*x[0]**o+(0.390298288126)*x[0]+(0.446872042362)*x[1]**o+(-0.0703294219415)*x[1]
        arg[3,1,0,1]=(-0.0706115592617)*x[0]**o+(-0.158386152093)*x[0]+(-0.687459181052)*x[1]**o+(-0.251170026598)*x[1]
        arg[3,1,1,0]=(-0.286907297081)*x[0]**o+(0.185480985579)*x[0]+(-0.49763263834)*x[1]**o+(-0.773056048642)*x[1]
        arg[3,1,1,1]=(-0.849572703562)*x[0]**o+(0.481739532516)*x[0]+(-0.227573063955)*x[1]**o+(-0.282367611381)*x[1]
        arg[3,1,2,0]=(-0.524975224603)*x[0]**o+(-0.795450756917)*x[0]+(-0.656865716441)*x[1]**o+(-0.0166902340102)*x[1]
        arg[3,1,2,1]=(-0.297441818145)*x[0]**o+(-0.726759852457)*x[0]+(0.146859888138)*x[1]**o+(-0.73690394554)*x[1]
        arg[3,2,0,0]=(-0.278794538596)*x[0]**o+(-0.463428732496)*x[0]+(-0.478500378607)*x[1]**o+(0.806643392698)*x[1]
        arg[3,2,0,1]=(0.573853299578)*x[0]**o+(-0.324830388013)*x[0]+(-0.185528736416)*x[1]**o+(0.444228024864)*x[1]
        arg[3,2,1,0]=(0.686054084195)*x[0]**o+(-0.914380567743)*x[0]+(0.751959744445)*x[1]**o+(0.531759675659)*x[1]
        arg[3,2,1,1]=(-0.876330263589)*x[0]**o+(0.566155144894)*x[0]+(-0.0233988901429)*x[1]**o+(0.42480184987)*x[1]
        arg[3,2,2,0]=(0.957053098936)*x[0]**o+(0.152292007141)*x[0]+(0.332521504492)*x[1]**o+(-0.289068173241)*x[1]
        arg[3,2,2,1]=(0.923170924493)*x[0]**o+(0.191154526504)*x[0]+(0.647848811717)*x[1]**o+(0.72509033544)*x[1]
        arg[3,3,0,0]=(0.388829993841)*x[0]**o+(0.243404004988)*x[0]+(-0.0313000323506)*x[1]**o+(0.558135056894)*x[1]
        arg[3,3,0,1]=(-0.784568790167)*x[0]**o+(0.742074534256)*x[0]+(-0.624869261724)*x[1]**o+(-0.412100194843)*x[1]
        arg[3,3,1,0]=(0.924173486785)*x[0]**o+(-0.966635363007)*x[0]+(0.395550623492)*x[1]**o+(0.469204722173)*x[1]
        arg[3,3,1,1]=(-0.842963105645)*x[0]**o+(0.685039360106)*x[0]+(0.279605123662)*x[1]**o+(-0.953092961091)*x[1]
        arg[3,3,2,0]=(0.598860382403)*x[0]**o+(-0.226413712435)*x[0]+(0.722250410625)*x[1]**o+(-0.944980469407)*x[1]
        arg[3,3,2,1]=(0.787289357927)*x[0]**o+(-0.270737247579)*x[0]+(-0.134351386793)*x[1]**o+(-0.0437387143766)*x[1]
        arg[3,4,0,0]=(-0.0601939617429)*x[0]**o+(-0.212346121247)*x[0]+(0.0137510326612)*x[1]**o+(0.749376479477)*x[1]
        arg[3,4,0,1]=(-0.515391847891)*x[0]**o+(-0.191867157458)*x[0]+(0.0883438399597)*x[1]**o+(-0.155095894625)*x[1]
        arg[3,4,1,0]=(0.172678225476)*x[0]**o+(-0.917390756059)*x[0]+(-0.498696112089)*x[1]**o+(-0.267655707135)*x[1]
        arg[3,4,1,1]=(-0.498294994476)*x[0]**o+(-0.301078252047)*x[0]+(0.100215523329)*x[1]**o+(-0.424039854141)*x[1]
        arg[3,4,2,0]=(0.904127003142)*x[0]**o+(-0.70704194874)*x[0]+(0.309638449466)*x[1]**o+(-0.290894515899)*x[1]
        arg[3,4,2,1]=(-0.924994398191)*x[0]**o+(0.624996423224)*x[0]+(0.461586866277)*x[1]**o+(-0.499709859121)*x[1]
        ref[0,0,0,0]=(-0.633576516648)*x_ref[0]**o+(0.77399256211)*x_ref[0]+(-0.396696417195)*x_ref[1]**o+(-0.604572540725)*x_ref[1]
        ref[0,0,0,1]=(0.612672991978)*x_ref[0]**o+(-0.331012302611)*x_ref[0]+(0.885533578235)*x_ref[1]**o+(0.334686818058)*x_ref[1]
        ref[0,0,1,0]=(-0.982701069025)*x_ref[0]**o+(0.467849033642)*x_ref[0]+(0.525392773375)*x_ref[1]**o+(0.355622324425)*x_ref[1]
        ref[0,0,1,1]=(-0.0243671469666)*x_ref[0]**o+(0.790044842383)*x_ref[0]+(0.413384172033)*x_ref[1]**o+(0.740805135968)*x_ref[1]
        ref[0,0,2,0]=(0.0384789957691)*x_ref[0]**o+(-0.317009899417)*x_ref[0]+(0.0389710369615)*x_ref[1]**o+(0.784549852598)*x_ref[1]
        ref[0,0,2,1]=(0.361832259252)*x_ref[0]**o+(0.410815792944)*x_ref[0]+(0.39496145831)*x_ref[1]**o+(-0.820784143964)*x_ref[1]
        ref[0,1,0,0]=(0.224261088976)*x_ref[0]**o+(-0.824694527256)*x_ref[0]+(0.318674472135)*x_ref[1]**o+(0.827696354338)*x_ref[1]
        ref[0,1,0,1]=(-0.442154901359)*x_ref[0]**o+(0.141775320862)*x_ref[0]+(0.688711841268)*x_ref[1]**o+(-0.27047542755)*x_ref[1]
        ref[0,1,1,0]=(-0.570125802019)*x_ref[0]**o+(0.050303762895)*x_ref[0]+(-0.972329837417)*x_ref[1]**o+(0.143700303765)*x_ref[1]
        ref[0,1,1,1]=(0.852888550499)*x_ref[0]**o+(-0.51216090329)*x_ref[0]+(-0.895663922335)*x_ref[1]**o+(-0.183466225547)*x_ref[1]
        ref[0,1,2,0]=(0.863840920739)*x_ref[0]**o+(-0.247789353524)*x_ref[0]+(0.63056436157)*x_ref[1]**o+(0.932522589942)*x_ref[1]
        ref[0,1,2,1]=(-0.403832121794)*x_ref[0]**o+(-0.769625259789)*x_ref[0]+(0.507724303191)*x_ref[1]**o+(0.314178775526)*x_ref[1]
        ref[0,2,0,0]=(0.543867153515)*x_ref[0]**o+(-0.414107817028)*x_ref[0]+(0.423922720842)*x_ref[1]**o+(0.166632407916)*x_ref[1]
        ref[0,2,0,1]=(0.731968209929)*x_ref[0]**o+(-0.114742736683)*x_ref[0]+(0.112562415021)*x_ref[1]**o+(0.166924480325)*x_ref[1]
        ref[0,2,1,0]=(0.919472066242)*x_ref[0]**o+(0.560565315205)*x_ref[0]+(0.4967650028)*x_ref[1]**o+(0.708376671625)*x_ref[1]
        ref[0,2,1,1]=(0.147500239739)*x_ref[0]**o+(-0.632670389919)*x_ref[0]+(-0.992479063437)*x_ref[1]**o+(0.258941122962)*x_ref[1]
        ref[0,2,2,0]=(0.474481785329)*x_ref[0]**o+(0.592902604094)*x_ref[0]+(0.64058685211)*x_ref[1]**o+(-0.477843307901)*x_ref[1]
        ref[0,2,2,1]=(0.769898355146)*x_ref[0]**o+(0.563458279487)*x_ref[0]+(0.876276425449)*x_ref[1]**o+(-0.0625727749869)*x_ref[1]
        ref[0,3,0,0]=(-0.330055037059)*x_ref[0]**o+(-0.727656105163)*x_ref[0]+(-0.559901684392)*x_ref[1]**o+(-0.433646878263)*x_ref[1]
        ref[0,3,0,1]=(-0.725314328666)*x_ref[0]**o+(0.638326609105)*x_ref[0]+(-0.126619368417)*x_ref[1]**o+(-0.232194758676)*x_ref[1]
        ref[0,3,1,0]=(0.318096005545)*x_ref[0]**o+(0.325247494289)*x_ref[0]+(-0.514305253643)*x_ref[1]**o+(0.108523350086)*x_ref[1]
        ref[0,3,1,1]=(0.298034894956)*x_ref[0]**o+(0.825714393651)*x_ref[0]+(-0.67162850079)*x_ref[1]**o+(0.681809083004)*x_ref[1]
        ref[0,3,2,0]=(-0.00715890582648)*x_ref[0]**o+(0.893900731227)*x_ref[0]+(-0.773710274545)*x_ref[1]**o+(0.415913728569)*x_ref[1]
        ref[0,3,2,1]=(0.0853551940958)*x_ref[0]**o+(-0.228876011127)*x_ref[0]+(0.87699456762)*x_ref[1]**o+(-0.744605229253)*x_ref[1]
        ref[0,4,0,0]=(0.942646937302)*x_ref[0]**o+(0.92717236759)*x_ref[0]+(0.794806821569)*x_ref[1]**o+(0.848766303526)*x_ref[1]
        ref[0,4,0,1]=(0.20140672561)*x_ref[0]**o+(0.974490372414)*x_ref[0]+(-0.450332419587)*x_ref[1]**o+(-0.859102184782)*x_ref[1]
        ref[0,4,1,0]=(0.632153053193)*x_ref[0]**o+(0.702336756277)*x_ref[0]+(-0.571496076236)*x_ref[1]**o+(-0.882472852219)*x_ref[1]
        ref[0,4,1,1]=(0.366452829499)*x_ref[0]**o+(-0.514660164741)*x_ref[0]+(-0.152794832253)*x_ref[1]**o+(-0.658688117796)*x_ref[1]
        ref[0,4,2,0]=(0.406218394702)*x_ref[0]**o+(0.782359698198)*x_ref[0]+(-0.0582499587562)*x_ref[1]**o+(0.664506852595)*x_ref[1]
        ref[0,4,2,1]=(0.768240870343)*x_ref[0]**o+(0.188470592904)*x_ref[0]+(0.855933509171)*x_ref[1]**o+(0.639174087541)*x_ref[1]
        ref[1,0,0,0]=(-0.946841398046)*x_ref[0]**o+(0.892729256183)*x_ref[0]+(-0.15644094115)*x_ref[1]**o+(-0.859371368233)*x_ref[1]
        ref[1,0,0,1]=(0.238113647666)*x_ref[0]**o+(0.173695470467)*x_ref[0]+(-0.753529252313)*x_ref[1]**o+(-0.243203902463)*x_ref[1]
        ref[1,0,1,0]=(0.267937523443)*x_ref[0]**o+(0.600200492374)*x_ref[0]+(0.442653529257)*x_ref[1]**o+(-0.394984210541)*x_ref[1]
        ref[1,0,1,1]=(-0.495022907622)*x_ref[0]**o+(-0.593273808093)*x_ref[0]+(0.888764746608)*x_ref[1]**o+(-0.745127825763)*x_ref[1]
        ref[1,0,2,0]=(0.352827238496)*x_ref[0]**o+(0.293200928839)*x_ref[0]+(-0.00369812109034)*x_ref[1]**o+(0.478875774193)*x_ref[1]
        ref[1,0,2,1]=(-0.403358878981)*x_ref[0]**o+(0.756947364667)*x_ref[0]+(0.649335795039)*x_ref[1]**o+(-0.339962796859)*x_ref[1]
        ref[1,1,0,0]=(-0.607505349015)*x_ref[0]**o+(-0.392535399582)*x_ref[0]+(0.799782389634)*x_ref[1]**o+(0.186339165863)*x_ref[1]
        ref[1,1,0,1]=(0.997492304627)*x_ref[0]**o+(-0.281456622581)*x_ref[0]+(-0.959294693613)*x_ref[1]**o+(-0.454812606674)*x_ref[1]
        ref[1,1,1,0]=(0.454682037549)*x_ref[0]**o+(0.797129935647)*x_ref[0]+(-0.949539573053)*x_ref[1]**o+(-0.918499382199)*x_ref[1]
        ref[1,1,1,1]=(0.903016449758)*x_ref[0]**o+(-0.575039046836)*x_ref[0]+(0.957158277074)*x_ref[1]**o+(-0.715797460906)*x_ref[1]
        ref[1,1,2,0]=(0.92005547899)*x_ref[0]**o+(-0.716975919302)*x_ref[0]+(-0.900904311708)*x_ref[1]**o+(0.989458809643)*x_ref[1]
        ref[1,1,2,1]=(-0.633214991087)*x_ref[0]**o+(0.680658578627)*x_ref[0]+(0.552194054556)*x_ref[1]**o+(0.343557072769)*x_ref[1]
        ref[1,2,0,0]=(-0.0182903530463)*x_ref[0]**o+(-0.0445569497267)*x_ref[0]+(-0.259920069365)*x_ref[1]**o+(0.118765787279)*x_ref[1]
        ref[1,2,0,1]=(-0.667382174733)*x_ref[0]**o+(-0.79898800024)*x_ref[0]+(-0.217638556567)*x_ref[1]**o+(-0.0288504127936)*x_ref[1]
        ref[1,2,1,0]=(-0.437982445862)*x_ref[0]**o+(-0.104054454808)*x_ref[0]+(0.711429718088)*x_ref[1]**o+(-0.607226296329)*x_ref[1]
        ref[1,2,1,1]=(0.252861218037)*x_ref[0]**o+(-0.635701456504)*x_ref[0]+(-0.555222304592)*x_ref[1]**o+(0.723744010092)*x_ref[1]
        ref[1,2,2,0]=(-0.73532710885)*x_ref[0]**o+(0.600599078741)*x_ref[0]+(0.0465795173194)*x_ref[1]**o+(0.524165544645)*x_ref[1]
        ref[1,2,2,1]=(-0.025246019744)*x_ref[0]**o+(-0.376526675619)*x_ref[0]+(0.213011788698)*x_ref[1]**o+(-0.477018225777)*x_ref[1]
        ref[1,3,0,0]=(-0.52682690957)*x_ref[0]**o+(0.0567472858499)*x_ref[0]+(0.574116256102)*x_ref[1]**o+(-0.130012791898)*x_ref[1]
        ref[1,3,0,1]=(0.113502182936)*x_ref[0]**o+(0.7643861313)*x_ref[0]+(0.48414948847)*x_ref[1]**o+(0.863168250318)*x_ref[1]
        ref[1,3,1,0]=(-0.208102162378)*x_ref[0]**o+(0.327770728981)*x_ref[0]+(0.843778895278)*x_ref[1]**o+(0.474843504177)*x_ref[1]
        ref[1,3,1,1]=(-0.549818192184)*x_ref[0]**o+(-0.863333816685)*x_ref[0]+(-0.365631583957)*x_ref[1]**o+(-0.491320441182)*x_ref[1]
        ref[1,3,2,0]=(-0.265833845759)*x_ref[0]**o+(-0.94416746081)*x_ref[0]+(0.48465442593)*x_ref[1]**o+(0.258284621408)*x_ref[1]
        ref[1,3,2,1]=(-0.625716082268)*x_ref[0]**o+(-0.234697130359)*x_ref[0]+(-0.973704406131)*x_ref[1]**o+(0.00141746117847)*x_ref[1]
        ref[1,4,0,0]=(0.565044361964)*x_ref[0]**o+(-0.0672474991748)*x_ref[0]+(-0.125901930524)*x_ref[1]**o+(-0.753103899732)*x_ref[1]
        ref[1,4,0,1]=(-0.979743467289)*x_ref[0]**o+(-0.25225085214)*x_ref[0]+(-0.141635164635)*x_ref[1]**o+(0.378153168667)*x_ref[1]
        ref[1,4,1,0]=(0.97728399206)*x_ref[0]**o+(-0.202043160532)*x_ref[0]+(0.997445424954)*x_ref[1]**o+(-0.100049830154)*x_ref[1]
        ref[1,4,1,1]=(0.364840105486)*x_ref[0]**o+(-0.682234327437)*x_ref[0]+(0.252584592259)*x_ref[1]**o+(-0.0785430242425)*x_ref[1]
        ref[1,4,2,0]=(-0.0168764722616)*x_ref[0]**o+(-0.997460139026)*x_ref[0]+(0.560548548912)*x_ref[1]**o+(0.220532224022)*x_ref[1]
        ref[1,4,2,1]=(0.307927943715)*x_ref[0]**o+(0.496390872708)*x_ref[0]+(0.982750813274)*x_ref[1]**o+(0.392968327778)*x_ref[1]
        ref[2,0,0,0]=(0.645991188948)*x_ref[0]**o+(0.104035822886)*x_ref[0]+(0.18285829097)*x_ref[1]**o+(0.51186286294)*x_ref[1]
        ref[2,0,0,1]=(-0.996971717318)*x_ref[0]**o+(0.126488047141)*x_ref[0]+(0.123585312562)*x_ref[1]**o+(-0.775854123529)*x_ref[1]
        ref[2,0,1,0]=(-0.0258590124655)*x_ref[0]**o+(0.0079454735671)*x_ref[0]+(-0.42819024422)*x_ref[1]**o+(0.12377360967)*x_ref[1]
        ref[2,0,1,1]=(0.990441351225)*x_ref[0]**o+(-0.453806376472)*x_ref[0]+(0.134316618645)*x_ref[1]**o+(-0.902666953282)*x_ref[1]
        ref[2,0,2,0]=(0.149304385108)*x_ref[0]**o+(-0.226132046999)*x_ref[0]+(0.647186106203)*x_ref[1]**o+(-0.269247029225)*x_ref[1]
        ref[2,0,2,1]=(-0.159068164422)*x_ref[0]**o+(-0.392275577366)*x_ref[0]+(0.131423093404)*x_ref[1]**o+(-0.353311282658)*x_ref[1]
        ref[2,1,0,0]=(-0.2862918681)*x_ref[0]**o+(-0.392264098939)*x_ref[0]+(-0.344033237285)*x_ref[1]**o+(0.644261713963)*x_ref[1]
        ref[2,1,0,1]=(0.398022911658)*x_ref[0]**o+(-0.84552546032)*x_ref[0]+(-0.823861260943)*x_ref[1]**o+(0.739994988755)*x_ref[1]
        ref[2,1,1,0]=(-0.696854341234)*x_ref[0]**o+(-0.810851183298)*x_ref[0]+(-0.712182042896)*x_ref[1]**o+(0.594887691783)*x_ref[1]
        ref[2,1,1,1]=(-0.182583840117)*x_ref[0]**o+(0.962594377088)*x_ref[0]+(-0.0157575661453)*x_ref[1]**o+(-0.432167995007)*x_ref[1]
        ref[2,1,2,0]=(0.293029196216)*x_ref[0]**o+(0.46839918212)*x_ref[0]+(0.262375931577)*x_ref[1]**o+(0.0348536956306)*x_ref[1]
        ref[2,1,2,1]=(0.775632135095)*x_ref[0]**o+(-0.874598534107)*x_ref[0]+(0.711602976736)*x_ref[1]**o+(0.37065543293)*x_ref[1]
        ref[2,2,0,0]=(-0.103625214158)*x_ref[0]**o+(0.553558117877)*x_ref[0]+(-0.0522465477589)*x_ref[1]**o+(-0.747139195529)*x_ref[1]
        ref[2,2,0,1]=(0.507427663615)*x_ref[0]**o+(-0.536361186134)*x_ref[0]+(0.809054726193)*x_ref[1]**o+(-0.979546347794)*x_ref[1]
        ref[2,2,1,0]=(0.786210623833)*x_ref[0]**o+(0.244510127091)*x_ref[0]+(0.159799334111)*x_ref[1]**o+(0.0499263839273)*x_ref[1]
        ref[2,2,1,1]=(0.454205244371)*x_ref[0]**o+(-0.650927822092)*x_ref[0]+(-0.939252656087)*x_ref[1]**o+(-0.894014755314)*x_ref[1]
        ref[2,2,2,0]=(-0.829617920904)*x_ref[0]**o+(0.764070900994)*x_ref[0]+(0.972419443556)*x_ref[1]**o+(-0.824913543157)*x_ref[1]
        ref[2,2,2,1]=(0.694310748478)*x_ref[0]**o+(-0.191413364091)*x_ref[0]+(0.588258464472)*x_ref[1]**o+(0.743973223755)*x_ref[1]
        ref[2,3,0,0]=(-0.588397365184)*x_ref[0]**o+(-0.634207034861)*x_ref[0]+(0.0853498555546)*x_ref[1]**o+(0.559933556865)*x_ref[1]
        ref[2,3,0,1]=(0.626829165689)*x_ref[0]**o+(0.942185296616)*x_ref[0]+(0.0360770405565)*x_ref[1]**o+(0.635966218547)*x_ref[1]
        ref[2,3,1,0]=(-0.0103526873852)*x_ref[0]**o+(0.546330663496)*x_ref[0]+(0.799224103454)*x_ref[1]**o+(0.791922324897)*x_ref[1]
        ref[2,3,1,1]=(0.921564474495)*x_ref[0]**o+(-0.420410835144)*x_ref[0]+(-0.20482503843)*x_ref[1]**o+(-0.483331643037)*x_ref[1]
        ref[2,3,2,0]=(-0.903600870314)*x_ref[0]**o+(0.901956716274)*x_ref[0]+(-0.113926342023)*x_ref[1]**o+(-0.061596184768)*x_ref[1]
        ref[2,3,2,1]=(0.653250267262)*x_ref[0]**o+(-0.13770933726)*x_ref[0]+(-0.0791563636079)*x_ref[1]**o+(-0.578313150893)*x_ref[1]
        ref[2,4,0,0]=(0.993903405725)*x_ref[0]**o+(0.694324226249)*x_ref[0]+(0.203872883541)*x_ref[1]**o+(-0.526880767264)*x_ref[1]
        ref[2,4,0,1]=(0.0120058513647)*x_ref[0]**o+(0.00907603337269)*x_ref[0]+(0.0247958601677)*x_ref[1]**o+(-0.996916120875)*x_ref[1]
        ref[2,4,1,0]=(0.281518049478)*x_ref[0]**o+(0.0773799323777)*x_ref[0]+(-0.968565098773)*x_ref[1]**o+(-0.275558076229)*x_ref[1]
        ref[2,4,1,1]=(0.565275744757)*x_ref[0]**o+(0.978636675869)*x_ref[0]+(-0.352951657751)*x_ref[1]**o+(-0.54170487914)*x_ref[1]
        ref[2,4,2,0]=(-0.571207550678)*x_ref[0]**o+(0.196754283318)*x_ref[0]+(0.665558917127)*x_ref[1]**o+(0.361915269355)*x_ref[1]
        ref[2,4,2,1]=(0.54579191885)*x_ref[0]**o+(0.798319371798)*x_ref[0]+(0.961669996133)*x_ref[1]**o+(-0.549305529342)*x_ref[1]
        ref[3,0,0,0]=(0.264064946121)*x_ref[0]**o+(-0.314396185593)*x_ref[0]+(-0.017492206135)*x_ref[1]**o+(0.190513393091)*x_ref[1]
        ref[3,0,0,1]=(0.812213666919)*x_ref[0]**o+(-0.757919357489)*x_ref[0]+(0.294829087285)*x_ref[1]**o+(0.916181101963)*x_ref[1]
        ref[3,0,1,0]=(-0.529906091574)*x_ref[0]**o+(0.553682716765)*x_ref[0]+(-0.693177625995)*x_ref[1]**o+(0.377145561683)*x_ref[1]
        ref[3,0,1,1]=(-0.298942248169)*x_ref[0]**o+(0.876878593705)*x_ref[0]+(-0.609807982832)*x_ref[1]**o+(-0.574954919539)*x_ref[1]
        ref[3,0,2,0]=(-0.729324018245)*x_ref[0]**o+(-0.495194565254)*x_ref[0]+(-0.616162751356)*x_ref[1]**o+(0.878549207376)*x_ref[1]
        ref[3,0,2,1]=(0.625373988248)*x_ref[0]**o+(0.106736108396)*x_ref[0]+(-0.871225833944)*x_ref[1]**o+(0.840614493429)*x_ref[1]
        ref[3,1,0,0]=(-0.90880839589)*x_ref[0]**o+(0.390298288126)*x_ref[0]+(0.446872042362)*x_ref[1]**o+(-0.0703294219415)*x_ref[1]
        ref[3,1,0,1]=(-0.0706115592617)*x_ref[0]**o+(-0.158386152093)*x_ref[0]+(-0.687459181052)*x_ref[1]**o+(-0.251170026598)*x_ref[1]
        ref[3,1,1,0]=(-0.286907297081)*x_ref[0]**o+(0.185480985579)*x_ref[0]+(-0.49763263834)*x_ref[1]**o+(-0.773056048642)*x_ref[1]
        ref[3,1,1,1]=(-0.849572703562)*x_ref[0]**o+(0.481739532516)*x_ref[0]+(-0.227573063955)*x_ref[1]**o+(-0.282367611381)*x_ref[1]
        ref[3,1,2,0]=(-0.524975224603)*x_ref[0]**o+(-0.795450756917)*x_ref[0]+(-0.656865716441)*x_ref[1]**o+(-0.0166902340102)*x_ref[1]
        ref[3,1,2,1]=(-0.297441818145)*x_ref[0]**o+(-0.726759852457)*x_ref[0]+(0.146859888138)*x_ref[1]**o+(-0.73690394554)*x_ref[1]
        ref[3,2,0,0]=(-0.278794538596)*x_ref[0]**o+(-0.463428732496)*x_ref[0]+(-0.478500378607)*x_ref[1]**o+(0.806643392698)*x_ref[1]
        ref[3,2,0,1]=(0.573853299578)*x_ref[0]**o+(-0.324830388013)*x_ref[0]+(-0.185528736416)*x_ref[1]**o+(0.444228024864)*x_ref[1]
        ref[3,2,1,0]=(0.686054084195)*x_ref[0]**o+(-0.914380567743)*x_ref[0]+(0.751959744445)*x_ref[1]**o+(0.531759675659)*x_ref[1]
        ref[3,2,1,1]=(-0.876330263589)*x_ref[0]**o+(0.566155144894)*x_ref[0]+(-0.0233988901429)*x_ref[1]**o+(0.42480184987)*x_ref[1]
        ref[3,2,2,0]=(0.957053098936)*x_ref[0]**o+(0.152292007141)*x_ref[0]+(0.332521504492)*x_ref[1]**o+(-0.289068173241)*x_ref[1]
        ref[3,2,2,1]=(0.923170924493)*x_ref[0]**o+(0.191154526504)*x_ref[0]+(0.647848811717)*x_ref[1]**o+(0.72509033544)*x_ref[1]
        ref[3,3,0,0]=(0.388829993841)*x_ref[0]**o+(0.243404004988)*x_ref[0]+(-0.0313000323506)*x_ref[1]**o+(0.558135056894)*x_ref[1]
        ref[3,3,0,1]=(-0.784568790167)*x_ref[0]**o+(0.742074534256)*x_ref[0]+(-0.624869261724)*x_ref[1]**o+(-0.412100194843)*x_ref[1]
        ref[3,3,1,0]=(0.924173486785)*x_ref[0]**o+(-0.966635363007)*x_ref[0]+(0.395550623492)*x_ref[1]**o+(0.469204722173)*x_ref[1]
        ref[3,3,1,1]=(-0.842963105645)*x_ref[0]**o+(0.685039360106)*x_ref[0]+(0.279605123662)*x_ref[1]**o+(-0.953092961091)*x_ref[1]
        ref[3,3,2,0]=(0.598860382403)*x_ref[0]**o+(-0.226413712435)*x_ref[0]+(0.722250410625)*x_ref[1]**o+(-0.944980469407)*x_ref[1]
        ref[3,3,2,1]=(0.787289357927)*x_ref[0]**o+(-0.270737247579)*x_ref[0]+(-0.134351386793)*x_ref[1]**o+(-0.0437387143766)*x_ref[1]
        ref[3,4,0,0]=(-0.0601939617429)*x_ref[0]**o+(-0.212346121247)*x_ref[0]+(0.0137510326612)*x_ref[1]**o+(0.749376479477)*x_ref[1]
        ref[3,4,0,1]=(-0.515391847891)*x_ref[0]**o+(-0.191867157458)*x_ref[0]+(0.0883438399597)*x_ref[1]**o+(-0.155095894625)*x_ref[1]
        ref[3,4,1,0]=(0.172678225476)*x_ref[0]**o+(-0.917390756059)*x_ref[0]+(-0.498696112089)*x_ref[1]**o+(-0.267655707135)*x_ref[1]
        ref[3,4,1,1]=(-0.498294994476)*x_ref[0]**o+(-0.301078252047)*x_ref[0]+(0.100215523329)*x_ref[1]**o+(-0.424039854141)*x_ref[1]
        ref[3,4,2,0]=(0.904127003142)*x_ref[0]**o+(-0.70704194874)*x_ref[0]+(0.309638449466)*x_ref[1]**o+(-0.290894515899)*x_ref[1]
        ref[3,4,2,1]=(-0.924994398191)*x_ref[0]**o+(0.624996423224)*x_ref[0]+(0.461586866277)*x_ref[1]**o+(-0.499709859121)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.0998996137438)*x[0]**o+(-0.648495171588)*x[0]+(-0.836177837758)*x[1]**o+(-0.9597381047)*x[1]+(-0.958069967907)*x[2]**o+(0.399097251753)*x[2]
        arg[0,0,0,1]=(0.625409406207)*x[0]**o+(-0.522476054218)*x[0]+(0.794724584797)*x[1]**o+(-0.450875661781)*x[1]+(-0.189505729613)*x[2]**o+(0.0910808218379)*x[2]
        arg[0,0,1,0]=(-0.36073198549)*x[0]**o+(-0.0286144078107)*x[0]+(-0.53827851785)*x[1]**o+(-0.563600303848)*x[1]+(0.970211345168)*x[2]**o+(-0.405253675856)*x[2]
        arg[0,0,1,1]=(-0.0204302558323)*x[0]**o+(0.927570053009)*x[0]+(-0.358571532503)*x[1]**o+(0.282513525404)*x[1]+(0.105530612734)*x[2]**o+(-0.156753729378)*x[2]
        arg[0,0,2,0]=(0.25049329858)*x[0]**o+(0.412299654624)*x[0]+(0.827683104949)*x[1]**o+(0.0195885473957)*x[1]+(0.849521623993)*x[2]**o+(-0.107922097347)*x[2]
        arg[0,0,2,1]=(0.824054630326)*x[0]**o+(0.315612619612)*x[0]+(-0.33528975625)*x[1]**o+(-0.848543364468)*x[1]+(0.67462279419)*x[2]**o+(-0.0387606720463)*x[2]
        arg[0,1,0,0]=(0.251613993994)*x[0]**o+(-0.871217704935)*x[0]+(-0.648182240752)*x[1]**o+(0.324536900202)*x[1]+(-0.884871901147)*x[2]**o+(0.353160185772)*x[2]
        arg[0,1,0,1]=(-0.629894976165)*x[0]**o+(0.99092287348)*x[0]+(0.87471936536)*x[1]**o+(0.732884844801)*x[1]+(0.423531105096)*x[2]**o+(0.648959408663)*x[2]
        arg[0,1,1,0]=(0.123366820062)*x[0]**o+(-0.856900246151)*x[0]+(-0.887819649211)*x[1]**o+(-0.901367472665)*x[1]+(0.222535500196)*x[2]**o+(0.42830671462)*x[2]
        arg[0,1,1,1]=(0.343789367141)*x[0]**o+(0.614284333526)*x[0]+(0.976225499577)*x[1]**o+(0.955115562801)*x[1]+(0.0700409126681)*x[2]**o+(-0.709602275548)*x[2]
        arg[0,1,2,0]=(-0.863502900367)*x[0]**o+(-0.452731447667)*x[0]+(0.495519161334)*x[1]**o+(0.0254261045314)*x[1]+(-0.484479110985)*x[2]**o+(0.615421138564)*x[2]
        arg[0,1,2,1]=(-0.861966549825)*x[0]**o+(0.957385177208)*x[0]+(0.710449966969)*x[1]**o+(-0.367149885665)*x[1]+(-0.100800220614)*x[2]**o+(-0.266781849874)*x[2]
        arg[0,2,0,0]=(-0.781309947495)*x[0]**o+(-0.709036446117)*x[0]+(0.86136976691)*x[1]**o+(-0.745743057084)*x[1]+(0.519795761984)*x[2]**o+(0.298838092751)*x[2]
        arg[0,2,0,1]=(-0.514037154361)*x[0]**o+(-0.155867632386)*x[0]+(-0.125008826668)*x[1]**o+(-0.138197902584)*x[1]+(0.360698647284)*x[2]**o+(0.432385729464)*x[2]
        arg[0,2,1,0]=(-0.19580514344)*x[0]**o+(-0.430510395896)*x[0]+(0.676498502854)*x[1]**o+(-0.472411864377)*x[1]+(-0.323468590594)*x[2]**o+(0.837971311859)*x[2]
        arg[0,2,1,1]=(-0.358714577548)*x[0]**o+(-0.583080123482)*x[0]+(-0.225887451598)*x[1]**o+(0.744101609163)*x[1]+(-0.0300539574094)*x[2]**o+(-0.985181847505)*x[2]
        arg[0,2,2,0]=(0.820610118447)*x[0]**o+(-0.560334915671)*x[0]+(0.478672329569)*x[1]**o+(-0.956892480643)*x[1]+(0.302100338619)*x[2]**o+(0.597584594886)*x[2]
        arg[0,2,2,1]=(-0.0658925658941)*x[0]**o+(-0.136738027019)*x[0]+(0.617259524472)*x[1]**o+(-0.486181324584)*x[1]+(0.225924222893)*x[2]**o+(-0.304893108343)*x[2]
        arg[0,3,0,0]=(0.0824288958004)*x[0]**o+(-0.67546530428)*x[0]+(-0.726843419073)*x[1]**o+(0.150753236521)*x[1]+(0.200621841908)*x[2]**o+(0.869623957921)*x[2]
        arg[0,3,0,1]=(0.66227806631)*x[0]**o+(-0.389229756136)*x[0]+(0.739548522805)*x[1]**o+(0.131026218939)*x[1]+(0.743271925877)*x[2]**o+(-0.910220233981)*x[2]
        arg[0,3,1,0]=(-0.259647259175)*x[0]**o+(-0.208585364012)*x[0]+(-0.323867955964)*x[1]**o+(0.778380345151)*x[1]+(-0.555780852614)*x[2]**o+(-0.16366483426)*x[2]
        arg[0,3,1,1]=(0.248968056109)*x[0]**o+(0.84999990087)*x[0]+(0.180937025847)*x[1]**o+(0.332075828156)*x[1]+(-0.00577042146131)*x[2]**o+(0.708188905312)*x[2]
        arg[0,3,2,0]=(-0.80327596428)*x[0]**o+(-0.131004563471)*x[0]+(0.443094741792)*x[1]**o+(0.515867966266)*x[1]+(-0.964003373109)*x[2]**o+(-0.580721642935)*x[2]
        arg[0,3,2,1]=(-0.107361389717)*x[0]**o+(-0.543617395264)*x[0]+(0.840867379108)*x[1]**o+(-0.459941493027)*x[1]+(-0.300033656162)*x[2]**o+(0.440763554498)*x[2]
        arg[0,4,0,0]=(0.272916362055)*x[0]**o+(-0.554210888243)*x[0]+(-0.458102512273)*x[1]**o+(-0.54594406835)*x[1]+(0.570670715852)*x[2]**o+(0.792731049718)*x[2]
        arg[0,4,0,1]=(0.896330380602)*x[0]**o+(-0.929757950797)*x[0]+(0.229049814758)*x[1]**o+(0.1245974562)*x[1]+(0.853250033203)*x[2]**o+(-0.877052530974)*x[2]
        arg[0,4,1,0]=(0.76027467594)*x[0]**o+(-0.510095378376)*x[0]+(0.697258966361)*x[1]**o+(-0.44013714543)*x[1]+(0.617675705023)*x[2]**o+(-0.00118327085347)*x[2]
        arg[0,4,1,1]=(0.92472377319)*x[0]**o+(0.27783405618)*x[0]+(-0.24317632451)*x[1]**o+(-0.0264727232824)*x[1]+(0.428751701408)*x[2]**o+(0.184142273303)*x[2]
        arg[0,4,2,0]=(0.619110983031)*x[0]**o+(-0.375037512361)*x[0]+(0.298576843317)*x[1]**o+(-0.218802335776)*x[1]+(-0.394889850336)*x[2]**o+(-0.842359750924)*x[2]
        arg[0,4,2,1]=(0.899714346356)*x[0]**o+(0.0630133184109)*x[0]+(0.0796796488932)*x[1]**o+(-0.693371400327)*x[1]+(-0.412090402419)*x[2]**o+(0.681871832039)*x[2]
        arg[1,0,0,0]=(-0.0367677770358)*x[0]**o+(-0.75352301678)*x[0]+(0.0705016755404)*x[1]**o+(0.870115630516)*x[1]+(0.136757694474)*x[2]**o+(-0.557722157484)*x[2]
        arg[1,0,0,1]=(0.550371937413)*x[0]**o+(-0.590052751632)*x[0]+(0.38665744963)*x[1]**o+(-0.896537565188)*x[1]+(0.550869514185)*x[2]**o+(0.693828237691)*x[2]
        arg[1,0,1,0]=(-0.150321579562)*x[0]**o+(0.111135028893)*x[0]+(0.582595079816)*x[1]**o+(-0.250920804447)*x[1]+(0.392998664266)*x[2]**o+(-0.226735735878)*x[2]
        arg[1,0,1,1]=(0.435196573865)*x[0]**o+(-0.31161203165)*x[0]+(-0.117580582163)*x[1]**o+(0.99583282896)*x[1]+(-0.164872214322)*x[2]**o+(0.243837126961)*x[2]
        arg[1,0,2,0]=(-0.320135099791)*x[0]**o+(0.339871290643)*x[0]+(0.593784119675)*x[1]**o+(0.695166367768)*x[1]+(0.407551427099)*x[2]**o+(-0.768445418933)*x[2]
        arg[1,0,2,1]=(-0.390718323619)*x[0]**o+(-0.0515766643887)*x[0]+(0.394112545837)*x[1]**o+(-0.379592335881)*x[1]+(-0.29680237927)*x[2]**o+(0.500583589142)*x[2]
        arg[1,1,0,0]=(-0.958930666062)*x[0]**o+(0.516874100468)*x[0]+(0.0859034666349)*x[1]**o+(-0.370561856015)*x[1]+(-0.175293417132)*x[2]**o+(0.672554615126)*x[2]
        arg[1,1,0,1]=(-0.38129381629)*x[0]**o+(-0.615636530444)*x[0]+(0.447284140424)*x[1]**o+(0.507583882952)*x[1]+(-0.942413761354)*x[2]**o+(-0.607928807162)*x[2]
        arg[1,1,1,0]=(-0.456462200294)*x[0]**o+(0.778557108965)*x[0]+(-0.939624730616)*x[1]**o+(0.965857713341)*x[1]+(-0.913772635427)*x[2]**o+(-0.459963514384)*x[2]
        arg[1,1,1,1]=(0.138896120228)*x[0]**o+(0.941491395046)*x[0]+(0.546538931051)*x[1]**o+(-0.526098363992)*x[1]+(0.32139556996)*x[2]**o+(0.899346403792)*x[2]
        arg[1,1,2,0]=(0.591597103411)*x[0]**o+(0.926278363416)*x[0]+(0.565926211046)*x[1]**o+(0.930516925784)*x[1]+(-0.498131779353)*x[2]**o+(0.358157310681)*x[2]
        arg[1,1,2,1]=(0.758395199797)*x[0]**o+(0.769578766255)*x[0]+(0.86819976051)*x[1]**o+(-0.380436310842)*x[1]+(-0.513012037048)*x[2]**o+(-0.984361866072)*x[2]
        arg[1,2,0,0]=(0.0845846076213)*x[0]**o+(0.987145343601)*x[0]+(-0.447680898266)*x[1]**o+(-0.30278197108)*x[1]+(0.271667503636)*x[2]**o+(0.698967822656)*x[2]
        arg[1,2,0,1]=(-0.314076111079)*x[0]**o+(0.507407943673)*x[0]+(-0.38467148586)*x[1]**o+(0.802764957621)*x[1]+(0.692949890402)*x[2]**o+(-0.965766183849)*x[2]
        arg[1,2,1,0]=(-0.855806343042)*x[0]**o+(0.753362713203)*x[0]+(-0.9877924012)*x[1]**o+(-0.54047626663)*x[1]+(-0.488859780266)*x[2]**o+(0.444279382381)*x[2]
        arg[1,2,1,1]=(-0.313257382113)*x[0]**o+(-0.926058887186)*x[0]+(0.509510780366)*x[1]**o+(0.738049597453)*x[1]+(0.191676625722)*x[2]**o+(-0.354148482258)*x[2]
        arg[1,2,2,0]=(-0.955204517346)*x[0]**o+(0.234220317948)*x[0]+(0.44435792198)*x[1]**o+(0.96099571541)*x[1]+(-0.422462745872)*x[2]**o+(0.808817733192)*x[2]
        arg[1,2,2,1]=(0.932454071169)*x[0]**o+(-0.449841123911)*x[0]+(-0.371587228288)*x[1]**o+(-0.552185751938)*x[1]+(-0.222070148528)*x[2]**o+(-0.911768306425)*x[2]
        arg[1,3,0,0]=(-0.355162862133)*x[0]**o+(-0.00562439573347)*x[0]+(0.982216168139)*x[1]**o+(0.0493986220989)*x[1]+(0.0242698649494)*x[2]**o+(0.593411693569)*x[2]
        arg[1,3,0,1]=(-0.982164819772)*x[0]**o+(-0.417475530445)*x[0]+(-0.421752719264)*x[1]**o+(-0.102522772133)*x[1]+(-0.384947583437)*x[2]**o+(-0.675090951882)*x[2]
        arg[1,3,1,0]=(0.862354670281)*x[0]**o+(0.0332806244503)*x[0]+(0.248152892705)*x[1]**o+(-0.380941068103)*x[1]+(0.499668747376)*x[2]**o+(-0.732344330916)*x[2]
        arg[1,3,1,1]=(-0.367669055003)*x[0]**o+(-0.28761578782)*x[0]+(0.10838982068)*x[1]**o+(-0.495999837684)*x[1]+(0.599492939122)*x[2]**o+(0.886609670648)*x[2]
        arg[1,3,2,0]=(0.707026122148)*x[0]**o+(-0.625314188271)*x[0]+(-0.734283031657)*x[1]**o+(0.779848908302)*x[1]+(-0.965798269237)*x[2]**o+(0.35160103487)*x[2]
        arg[1,3,2,1]=(0.198008431503)*x[0]**o+(0.145470568903)*x[0]+(-0.342584113223)*x[1]**o+(-0.976390968602)*x[1]+(-0.371678081685)*x[2]**o+(0.569699640624)*x[2]
        arg[1,4,0,0]=(-0.816642540092)*x[0]**o+(0.466127552066)*x[0]+(-0.491109738021)*x[1]**o+(-0.177222784312)*x[1]+(0.723624314914)*x[2]**o+(-0.103929106752)*x[2]
        arg[1,4,0,1]=(-0.461770336596)*x[0]**o+(-0.416473447722)*x[0]+(-0.844639456093)*x[1]**o+(-0.021509624606)*x[1]+(-0.507857805076)*x[2]**o+(-0.0132005113079)*x[2]
        arg[1,4,1,0]=(0.279614653607)*x[0]**o+(-0.547332204747)*x[0]+(0.329751635637)*x[1]**o+(0.813149533168)*x[1]+(0.0361758823053)*x[2]**o+(0.0635753923527)*x[2]
        arg[1,4,1,1]=(0.124940246141)*x[0]**o+(-0.470715329674)*x[0]+(0.401922174627)*x[1]**o+(-0.292371975658)*x[1]+(0.302438467505)*x[2]**o+(0.131045792461)*x[2]
        arg[1,4,2,0]=(0.817170444398)*x[0]**o+(0.153182033447)*x[0]+(-0.98908973731)*x[1]**o+(0.359609127275)*x[1]+(-0.891501663456)*x[2]**o+(0.34700998982)*x[2]
        arg[1,4,2,1]=(0.619050545861)*x[0]**o+(-0.784466574079)*x[0]+(-0.15316511355)*x[1]**o+(0.587255937366)*x[1]+(0.740644692907)*x[2]**o+(0.0651236072508)*x[2]
        arg[2,0,0,0]=(0.67904774175)*x[0]**o+(-0.24998331377)*x[0]+(-0.263635906274)*x[1]**o+(0.157008949549)*x[1]+(-0.905158896842)*x[2]**o+(-0.975062080366)*x[2]
        arg[2,0,0,1]=(0.724305963876)*x[0]**o+(0.666891161317)*x[0]+(0.769784501206)*x[1]**o+(-0.84353242369)*x[1]+(0.631055556571)*x[2]**o+(-0.433528873211)*x[2]
        arg[2,0,1,0]=(-0.545702871321)*x[0]**o+(-0.153233364062)*x[0]+(-0.892642075879)*x[1]**o+(-0.303965445232)*x[1]+(0.675938211305)*x[2]**o+(0.134016007164)*x[2]
        arg[2,0,1,1]=(0.612528479764)*x[0]**o+(0.444039154606)*x[0]+(-0.185281638413)*x[1]**o+(0.643315773689)*x[1]+(-0.5562387833)*x[2]**o+(-0.960249468718)*x[2]
        arg[2,0,2,0]=(-0.330135152971)*x[0]**o+(0.359951176686)*x[0]+(-0.797088515842)*x[1]**o+(-0.257909258598)*x[1]+(-0.927156080256)*x[2]**o+(-0.843405252125)*x[2]
        arg[2,0,2,1]=(0.447041226497)*x[0]**o+(0.691225329527)*x[0]+(0.462798475523)*x[1]**o+(0.157497098383)*x[1]+(-0.77785224695)*x[2]**o+(0.886430058902)*x[2]
        arg[2,1,0,0]=(0.397605326971)*x[0]**o+(0.66227843262)*x[0]+(-0.634922556623)*x[1]**o+(-0.263342020187)*x[1]+(-0.81486179124)*x[2]**o+(-0.216716259698)*x[2]
        arg[2,1,0,1]=(-0.868246547394)*x[0]**o+(-0.194983059368)*x[0]+(-0.667277750808)*x[1]**o+(0.284011270958)*x[1]+(-0.276295620425)*x[2]**o+(0.215159516398)*x[2]
        arg[2,1,1,0]=(0.990906320323)*x[0]**o+(0.526754062323)*x[0]+(-0.963398315893)*x[1]**o+(-0.72561275544)*x[1]+(0.539779065056)*x[2]**o+(0.37738462043)*x[2]
        arg[2,1,1,1]=(-0.668520028258)*x[0]**o+(-0.292883094355)*x[0]+(0.260949538172)*x[1]**o+(-0.952706524522)*x[1]+(-0.715386804875)*x[2]**o+(-0.683320136224)*x[2]
        arg[2,1,2,0]=(-0.646186301281)*x[0]**o+(-0.62747987465)*x[0]+(-0.858714171368)*x[1]**o+(0.46910094021)*x[1]+(0.468975652807)*x[2]**o+(0.242115907129)*x[2]
        arg[2,1,2,1]=(0.262205767991)*x[0]**o+(-0.460376423099)*x[0]+(0.88572453638)*x[1]**o+(0.611761462581)*x[1]+(0.763860971141)*x[2]**o+(0.108470198525)*x[2]
        arg[2,2,0,0]=(-0.0218691308052)*x[0]**o+(0.942271392012)*x[0]+(-0.778469966054)*x[1]**o+(0.0873305333765)*x[1]+(-0.300479183111)*x[2]**o+(-0.862547713455)*x[2]
        arg[2,2,0,1]=(-0.736808990311)*x[0]**o+(-0.200478457878)*x[0]+(0.272608362607)*x[1]**o+(-0.0651056524306)*x[1]+(0.616628495568)*x[2]**o+(-0.8623490351)*x[2]
        arg[2,2,1,0]=(0.931530668297)*x[0]**o+(0.39701649536)*x[0]+(0.552135074711)*x[1]**o+(-0.489731326344)*x[1]+(-0.864724191354)*x[2]**o+(0.369557062583)*x[2]
        arg[2,2,1,1]=(0.418780453395)*x[0]**o+(-0.4935990418)*x[0]+(0.34625989303)*x[1]**o+(0.979829617988)*x[1]+(0.853169987441)*x[2]**o+(-0.945958722805)*x[2]
        arg[2,2,2,0]=(0.973689940887)*x[0]**o+(-0.0585784323994)*x[0]+(-0.472419744635)*x[1]**o+(0.851814193674)*x[1]+(-0.393393340405)*x[2]**o+(-0.755376551633)*x[2]
        arg[2,2,2,1]=(-0.217049913648)*x[0]**o+(-0.262150796255)*x[0]+(-0.407805273745)*x[1]**o+(0.819646964669)*x[1]+(-0.0130316627236)*x[2]**o+(0.179034130621)*x[2]
        arg[2,3,0,0]=(-0.377708814803)*x[0]**o+(-0.943893058411)*x[0]+(-0.354341838264)*x[1]**o+(0.0696491295689)*x[1]+(-0.192838410156)*x[2]**o+(-0.22208932914)*x[2]
        arg[2,3,0,1]=(0.252429489766)*x[0]**o+(0.808052283089)*x[0]+(0.282941493712)*x[1]**o+(0.89816007357)*x[1]+(0.000979423783521)*x[2]**o+(0.645599553602)*x[2]
        arg[2,3,1,0]=(0.788089987784)*x[0]**o+(0.414254044423)*x[0]+(0.105204158639)*x[1]**o+(0.97527041289)*x[1]+(0.447463569594)*x[2]**o+(0.91465488394)*x[2]
        arg[2,3,1,1]=(0.882323493958)*x[0]**o+(-0.98135461385)*x[0]+(0.641881834427)*x[1]**o+(-0.888068999894)*x[1]+(0.705728558352)*x[2]**o+(0.295815435834)*x[2]
        arg[2,3,2,0]=(-0.570666876652)*x[0]**o+(0.52308582534)*x[0]+(-0.604213776925)*x[1]**o+(-0.568684610916)*x[1]+(-0.0378463784885)*x[2]**o+(0.557994858276)*x[2]
        arg[2,3,2,1]=(-0.497489149519)*x[0]**o+(0.739341493998)*x[0]+(-0.375869065986)*x[1]**o+(-0.895553158302)*x[1]+(0.916829505813)*x[2]**o+(-0.650522069672)*x[2]
        arg[2,4,0,0]=(0.627540843427)*x[0]**o+(-0.572125230014)*x[0]+(-0.404781893475)*x[1]**o+(-0.114896790942)*x[1]+(-0.108914764977)*x[2]**o+(0.739758413931)*x[2]
        arg[2,4,0,1]=(0.243205825969)*x[0]**o+(-0.989602234642)*x[0]+(-0.925730664677)*x[1]**o+(0.454801040316)*x[1]+(0.51758036389)*x[2]**o+(0.325160660047)*x[2]
        arg[2,4,1,0]=(0.559764013478)*x[0]**o+(2.69034311577e-05)*x[0]+(-0.691119937581)*x[1]**o+(-0.347457112767)*x[1]+(0.714683489268)*x[2]**o+(-0.589270137254)*x[2]
        arg[2,4,1,1]=(-0.657837524953)*x[0]**o+(0.071158671662)*x[0]+(-0.998677178035)*x[1]**o+(-0.544136374784)*x[1]+(0.6200901771)*x[2]**o+(-0.280453395396)*x[2]
        arg[2,4,2,0]=(0.692403743003)*x[0]**o+(-0.570086669428)*x[0]+(0.455036498939)*x[1]**o+(0.291433179453)*x[1]+(-0.203004457401)*x[2]**o+(0.0192489593981)*x[2]
        arg[2,4,2,1]=(-0.656783033826)*x[0]**o+(0.865748949755)*x[0]+(-0.297731124456)*x[1]**o+(-0.826786841513)*x[1]+(0.891374127144)*x[2]**o+(-0.736789456498)*x[2]
        arg[3,0,0,0]=(-0.963120586854)*x[0]**o+(0.466089565408)*x[0]+(-0.785137906849)*x[1]**o+(0.194880437113)*x[1]+(0.652057839639)*x[2]**o+(-0.63674893761)*x[2]
        arg[3,0,0,1]=(-0.923234625819)*x[0]**o+(-0.303847080864)*x[0]+(-0.374245465515)*x[1]**o+(0.131093877726)*x[1]+(-0.13469395748)*x[2]**o+(0.42721676585)*x[2]
        arg[3,0,1,0]=(-0.613510153481)*x[0]**o+(0.986696685318)*x[0]+(0.188583200151)*x[1]**o+(-0.895175969251)*x[1]+(-0.699510846377)*x[2]**o+(0.953157679279)*x[2]
        arg[3,0,1,1]=(-0.805809890231)*x[0]**o+(-0.786467529592)*x[0]+(-0.287240865489)*x[1]**o+(-0.78714682888)*x[1]+(-0.301034554724)*x[2]**o+(0.698100092246)*x[2]
        arg[3,0,2,0]=(-0.29320618172)*x[0]**o+(0.482436772935)*x[0]+(-0.126052929261)*x[1]**o+(-0.0638862140753)*x[1]+(0.42769411972)*x[2]**o+(0.741118745829)*x[2]
        arg[3,0,2,1]=(-0.380041427783)*x[0]**o+(-0.0347904596008)*x[0]+(-0.321599184546)*x[1]**o+(-0.061333518544)*x[1]+(-0.885193271364)*x[2]**o+(0.216506075886)*x[2]
        arg[3,1,0,0]=(-0.872457576682)*x[0]**o+(0.1368749584)*x[0]+(0.883209741333)*x[1]**o+(0.946644762296)*x[1]+(-0.941665219846)*x[2]**o+(0.683787986795)*x[2]
        arg[3,1,0,1]=(0.564843030077)*x[0]**o+(0.987198898299)*x[0]+(0.808116156661)*x[1]**o+(-0.712172259479)*x[1]+(0.942066890171)*x[2]**o+(-0.265957158917)*x[2]
        arg[3,1,1,0]=(-0.171018264631)*x[0]**o+(0.277202848962)*x[0]+(0.0726306773528)*x[1]**o+(-0.0265640725409)*x[1]+(-0.871118922929)*x[2]**o+(0.0562314512261)*x[2]
        arg[3,1,1,1]=(0.0861984079085)*x[0]**o+(-0.954818837051)*x[0]+(0.817362109646)*x[1]**o+(-0.513086270222)*x[1]+(-0.487472170741)*x[2]**o+(0.702514192773)*x[2]
        arg[3,1,2,0]=(0.272419057202)*x[0]**o+(-0.0454292509397)*x[0]+(-0.117471126603)*x[1]**o+(0.638237510553)*x[1]+(0.156858948691)*x[2]**o+(0.499459702141)*x[2]
        arg[3,1,2,1]=(-0.384525883334)*x[0]**o+(-0.795638828759)*x[0]+(-0.105795059118)*x[1]**o+(0.539873566967)*x[1]+(-0.402558868714)*x[2]**o+(-0.690578475243)*x[2]
        arg[3,2,0,0]=(0.776889517136)*x[0]**o+(0.710931840891)*x[0]+(0.335992677388)*x[1]**o+(-0.689182863932)*x[1]+(0.294149805572)*x[2]**o+(-0.144223257292)*x[2]
        arg[3,2,0,1]=(0.738703695733)*x[0]**o+(-0.523207061995)*x[0]+(0.707078352596)*x[1]**o+(-0.594373978096)*x[1]+(-0.833468457639)*x[2]**o+(-0.60181674446)*x[2]
        arg[3,2,1,0]=(-0.70243682985)*x[0]**o+(-0.557914174271)*x[0]+(-0.97153723053)*x[1]**o+(0.791692594024)*x[1]+(-0.442273808305)*x[2]**o+(-0.74069605321)*x[2]
        arg[3,2,1,1]=(0.136307039789)*x[0]**o+(0.820661597623)*x[0]+(0.379752060994)*x[1]**o+(-0.739083116742)*x[1]+(-0.964857055764)*x[2]**o+(0.523159928671)*x[2]
        arg[3,2,2,0]=(0.530883295074)*x[0]**o+(-0.859538509888)*x[0]+(-0.33917158254)*x[1]**o+(0.476510741055)*x[1]+(0.851975683024)*x[2]**o+(0.146277015139)*x[2]
        arg[3,2,2,1]=(-0.532984574212)*x[0]**o+(-0.413169979754)*x[0]+(0.0113284200501)*x[1]**o+(0.406768750567)*x[1]+(0.730968378807)*x[2]**o+(-0.74004422909)*x[2]
        arg[3,3,0,0]=(0.784793271465)*x[0]**o+(0.879569037436)*x[0]+(0.687804194006)*x[1]**o+(0.153617608524)*x[1]+(-0.410247334526)*x[2]**o+(0.147941373268)*x[2]
        arg[3,3,0,1]=(-0.896748874838)*x[0]**o+(-0.408404228864)*x[0]+(0.831967697989)*x[1]**o+(-0.290875629476)*x[1]+(-0.244101719785)*x[2]**o+(0.240453425519)*x[2]
        arg[3,3,1,0]=(-0.352293918285)*x[0]**o+(0.0341233590345)*x[0]+(-0.0165872699681)*x[1]**o+(-0.0668507705942)*x[1]+(-0.133733710061)*x[2]**o+(-0.00620237772745)*x[2]
        arg[3,3,1,1]=(0.173638606531)*x[0]**o+(0.619705705811)*x[0]+(0.279972531576)*x[1]**o+(-0.0811181576037)*x[1]+(-0.645673968637)*x[2]**o+(0.913753411758)*x[2]
        arg[3,3,2,0]=(0.0623921286142)*x[0]**o+(0.393414889204)*x[0]+(-0.300034851075)*x[1]**o+(-0.519515206457)*x[1]+(-0.0074981731879)*x[2]**o+(0.202996242895)*x[2]
        arg[3,3,2,1]=(0.583279597505)*x[0]**o+(-0.940603487209)*x[0]+(0.027966157338)*x[1]**o+(0.251466344692)*x[1]+(-0.97011033154)*x[2]**o+(0.46527056203)*x[2]
        arg[3,4,0,0]=(0.865170740218)*x[0]**o+(-0.466853211289)*x[0]+(0.0935492537278)*x[1]**o+(-0.632374632657)*x[1]+(0.673466346463)*x[2]**o+(-0.47934739873)*x[2]
        arg[3,4,0,1]=(0.0417451136027)*x[0]**o+(-0.781570733286)*x[0]+(0.547452389454)*x[1]**o+(-0.0286368211528)*x[1]+(-0.98907358808)*x[2]**o+(0.62749495057)*x[2]
        arg[3,4,1,0]=(-0.632067479664)*x[0]**o+(-0.872835386336)*x[0]+(-0.228785631078)*x[1]**o+(-0.224206127306)*x[1]+(0.769729051217)*x[2]**o+(-0.406750034352)*x[2]
        arg[3,4,1,1]=(0.216883071299)*x[0]**o+(-0.225632200121)*x[0]+(0.827186402743)*x[1]**o+(0.161488636886)*x[1]+(0.773801288836)*x[2]**o+(0.977006832166)*x[2]
        arg[3,4,2,0]=(-0.867098845884)*x[0]**o+(0.14989544838)*x[0]+(0.756720043262)*x[1]**o+(-0.496074302626)*x[1]+(-0.662787072451)*x[2]**o+(-0.142473973368)*x[2]
        arg[3,4,2,1]=(-0.294128396829)*x[0]**o+(-0.931379392091)*x[0]+(0.906552131161)*x[1]**o+(-0.422008485494)*x[1]+(0.806742455679)*x[2]**o+(0.940488803006)*x[2]
        ref[0,0,0,0]=(-0.0998996137438)*x_ref[0]**o+(-0.648495171588)*x_ref[0]+(-0.836177837758)*x_ref[1]**o+(-0.9597381047)*x_ref[1]+(-0.958069967907)*x_ref[2]**o+(0.399097251753)*x_ref[2]
        ref[0,0,0,1]=(0.625409406207)*x_ref[0]**o+(-0.522476054218)*x_ref[0]+(0.794724584797)*x_ref[1]**o+(-0.450875661781)*x_ref[1]+(-0.189505729613)*x_ref[2]**o+(0.0910808218379)*x_ref[2]
        ref[0,0,1,0]=(-0.36073198549)*x_ref[0]**o+(-0.0286144078107)*x_ref[0]+(-0.53827851785)*x_ref[1]**o+(-0.563600303848)*x_ref[1]+(0.970211345168)*x_ref[2]**o+(-0.405253675856)*x_ref[2]
        ref[0,0,1,1]=(-0.0204302558323)*x_ref[0]**o+(0.927570053009)*x_ref[0]+(-0.358571532503)*x_ref[1]**o+(0.282513525404)*x_ref[1]+(0.105530612734)*x_ref[2]**o+(-0.156753729378)*x_ref[2]
        ref[0,0,2,0]=(0.25049329858)*x_ref[0]**o+(0.412299654624)*x_ref[0]+(0.827683104949)*x_ref[1]**o+(0.0195885473957)*x_ref[1]+(0.849521623993)*x_ref[2]**o+(-0.107922097347)*x_ref[2]
        ref[0,0,2,1]=(0.824054630326)*x_ref[0]**o+(0.315612619612)*x_ref[0]+(-0.33528975625)*x_ref[1]**o+(-0.848543364468)*x_ref[1]+(0.67462279419)*x_ref[2]**o+(-0.0387606720463)*x_ref[2]
        ref[0,1,0,0]=(0.251613993994)*x_ref[0]**o+(-0.871217704935)*x_ref[0]+(-0.648182240752)*x_ref[1]**o+(0.324536900202)*x_ref[1]+(-0.884871901147)*x_ref[2]**o+(0.353160185772)*x_ref[2]
        ref[0,1,0,1]=(-0.629894976165)*x_ref[0]**o+(0.99092287348)*x_ref[0]+(0.87471936536)*x_ref[1]**o+(0.732884844801)*x_ref[1]+(0.423531105096)*x_ref[2]**o+(0.648959408663)*x_ref[2]
        ref[0,1,1,0]=(0.123366820062)*x_ref[0]**o+(-0.856900246151)*x_ref[0]+(-0.887819649211)*x_ref[1]**o+(-0.901367472665)*x_ref[1]+(0.222535500196)*x_ref[2]**o+(0.42830671462)*x_ref[2]
        ref[0,1,1,1]=(0.343789367141)*x_ref[0]**o+(0.614284333526)*x_ref[0]+(0.976225499577)*x_ref[1]**o+(0.955115562801)*x_ref[1]+(0.0700409126681)*x_ref[2]**o+(-0.709602275548)*x_ref[2]
        ref[0,1,2,0]=(-0.863502900367)*x_ref[0]**o+(-0.452731447667)*x_ref[0]+(0.495519161334)*x_ref[1]**o+(0.0254261045314)*x_ref[1]+(-0.484479110985)*x_ref[2]**o+(0.615421138564)*x_ref[2]
        ref[0,1,2,1]=(-0.861966549825)*x_ref[0]**o+(0.957385177208)*x_ref[0]+(0.710449966969)*x_ref[1]**o+(-0.367149885665)*x_ref[1]+(-0.100800220614)*x_ref[2]**o+(-0.266781849874)*x_ref[2]
        ref[0,2,0,0]=(-0.781309947495)*x_ref[0]**o+(-0.709036446117)*x_ref[0]+(0.86136976691)*x_ref[1]**o+(-0.745743057084)*x_ref[1]+(0.519795761984)*x_ref[2]**o+(0.298838092751)*x_ref[2]
        ref[0,2,0,1]=(-0.514037154361)*x_ref[0]**o+(-0.155867632386)*x_ref[0]+(-0.125008826668)*x_ref[1]**o+(-0.138197902584)*x_ref[1]+(0.360698647284)*x_ref[2]**o+(0.432385729464)*x_ref[2]
        ref[0,2,1,0]=(-0.19580514344)*x_ref[0]**o+(-0.430510395896)*x_ref[0]+(0.676498502854)*x_ref[1]**o+(-0.472411864377)*x_ref[1]+(-0.323468590594)*x_ref[2]**o+(0.837971311859)*x_ref[2]
        ref[0,2,1,1]=(-0.358714577548)*x_ref[0]**o+(-0.583080123482)*x_ref[0]+(-0.225887451598)*x_ref[1]**o+(0.744101609163)*x_ref[1]+(-0.0300539574094)*x_ref[2]**o+(-0.985181847505)*x_ref[2]
        ref[0,2,2,0]=(0.820610118447)*x_ref[0]**o+(-0.560334915671)*x_ref[0]+(0.478672329569)*x_ref[1]**o+(-0.956892480643)*x_ref[1]+(0.302100338619)*x_ref[2]**o+(0.597584594886)*x_ref[2]
        ref[0,2,2,1]=(-0.0658925658941)*x_ref[0]**o+(-0.136738027019)*x_ref[0]+(0.617259524472)*x_ref[1]**o+(-0.486181324584)*x_ref[1]+(0.225924222893)*x_ref[2]**o+(-0.304893108343)*x_ref[2]
        ref[0,3,0,0]=(0.0824288958004)*x_ref[0]**o+(-0.67546530428)*x_ref[0]+(-0.726843419073)*x_ref[1]**o+(0.150753236521)*x_ref[1]+(0.200621841908)*x_ref[2]**o+(0.869623957921)*x_ref[2]
        ref[0,3,0,1]=(0.66227806631)*x_ref[0]**o+(-0.389229756136)*x_ref[0]+(0.739548522805)*x_ref[1]**o+(0.131026218939)*x_ref[1]+(0.743271925877)*x_ref[2]**o+(-0.910220233981)*x_ref[2]
        ref[0,3,1,0]=(-0.259647259175)*x_ref[0]**o+(-0.208585364012)*x_ref[0]+(-0.323867955964)*x_ref[1]**o+(0.778380345151)*x_ref[1]+(-0.555780852614)*x_ref[2]**o+(-0.16366483426)*x_ref[2]
        ref[0,3,1,1]=(0.248968056109)*x_ref[0]**o+(0.84999990087)*x_ref[0]+(0.180937025847)*x_ref[1]**o+(0.332075828156)*x_ref[1]+(-0.00577042146131)*x_ref[2]**o+(0.708188905312)*x_ref[2]
        ref[0,3,2,0]=(-0.80327596428)*x_ref[0]**o+(-0.131004563471)*x_ref[0]+(0.443094741792)*x_ref[1]**o+(0.515867966266)*x_ref[1]+(-0.964003373109)*x_ref[2]**o+(-0.580721642935)*x_ref[2]
        ref[0,3,2,1]=(-0.107361389717)*x_ref[0]**o+(-0.543617395264)*x_ref[0]+(0.840867379108)*x_ref[1]**o+(-0.459941493027)*x_ref[1]+(-0.300033656162)*x_ref[2]**o+(0.440763554498)*x_ref[2]
        ref[0,4,0,0]=(0.272916362055)*x_ref[0]**o+(-0.554210888243)*x_ref[0]+(-0.458102512273)*x_ref[1]**o+(-0.54594406835)*x_ref[1]+(0.570670715852)*x_ref[2]**o+(0.792731049718)*x_ref[2]
        ref[0,4,0,1]=(0.896330380602)*x_ref[0]**o+(-0.929757950797)*x_ref[0]+(0.229049814758)*x_ref[1]**o+(0.1245974562)*x_ref[1]+(0.853250033203)*x_ref[2]**o+(-0.877052530974)*x_ref[2]
        ref[0,4,1,0]=(0.76027467594)*x_ref[0]**o+(-0.510095378376)*x_ref[0]+(0.697258966361)*x_ref[1]**o+(-0.44013714543)*x_ref[1]+(0.617675705023)*x_ref[2]**o+(-0.00118327085347)*x_ref[2]
        ref[0,4,1,1]=(0.92472377319)*x_ref[0]**o+(0.27783405618)*x_ref[0]+(-0.24317632451)*x_ref[1]**o+(-0.0264727232824)*x_ref[1]+(0.428751701408)*x_ref[2]**o+(0.184142273303)*x_ref[2]
        ref[0,4,2,0]=(0.619110983031)*x_ref[0]**o+(-0.375037512361)*x_ref[0]+(0.298576843317)*x_ref[1]**o+(-0.218802335776)*x_ref[1]+(-0.394889850336)*x_ref[2]**o+(-0.842359750924)*x_ref[2]
        ref[0,4,2,1]=(0.899714346356)*x_ref[0]**o+(0.0630133184109)*x_ref[0]+(0.0796796488932)*x_ref[1]**o+(-0.693371400327)*x_ref[1]+(-0.412090402419)*x_ref[2]**o+(0.681871832039)*x_ref[2]
        ref[1,0,0,0]=(-0.0367677770358)*x_ref[0]**o+(-0.75352301678)*x_ref[0]+(0.0705016755404)*x_ref[1]**o+(0.870115630516)*x_ref[1]+(0.136757694474)*x_ref[2]**o+(-0.557722157484)*x_ref[2]
        ref[1,0,0,1]=(0.550371937413)*x_ref[0]**o+(-0.590052751632)*x_ref[0]+(0.38665744963)*x_ref[1]**o+(-0.896537565188)*x_ref[1]+(0.550869514185)*x_ref[2]**o+(0.693828237691)*x_ref[2]
        ref[1,0,1,0]=(-0.150321579562)*x_ref[0]**o+(0.111135028893)*x_ref[0]+(0.582595079816)*x_ref[1]**o+(-0.250920804447)*x_ref[1]+(0.392998664266)*x_ref[2]**o+(-0.226735735878)*x_ref[2]
        ref[1,0,1,1]=(0.435196573865)*x_ref[0]**o+(-0.31161203165)*x_ref[0]+(-0.117580582163)*x_ref[1]**o+(0.99583282896)*x_ref[1]+(-0.164872214322)*x_ref[2]**o+(0.243837126961)*x_ref[2]
        ref[1,0,2,0]=(-0.320135099791)*x_ref[0]**o+(0.339871290643)*x_ref[0]+(0.593784119675)*x_ref[1]**o+(0.695166367768)*x_ref[1]+(0.407551427099)*x_ref[2]**o+(-0.768445418933)*x_ref[2]
        ref[1,0,2,1]=(-0.390718323619)*x_ref[0]**o+(-0.0515766643887)*x_ref[0]+(0.394112545837)*x_ref[1]**o+(-0.379592335881)*x_ref[1]+(-0.29680237927)*x_ref[2]**o+(0.500583589142)*x_ref[2]
        ref[1,1,0,0]=(-0.958930666062)*x_ref[0]**o+(0.516874100468)*x_ref[0]+(0.0859034666349)*x_ref[1]**o+(-0.370561856015)*x_ref[1]+(-0.175293417132)*x_ref[2]**o+(0.672554615126)*x_ref[2]
        ref[1,1,0,1]=(-0.38129381629)*x_ref[0]**o+(-0.615636530444)*x_ref[0]+(0.447284140424)*x_ref[1]**o+(0.507583882952)*x_ref[1]+(-0.942413761354)*x_ref[2]**o+(-0.607928807162)*x_ref[2]
        ref[1,1,1,0]=(-0.456462200294)*x_ref[0]**o+(0.778557108965)*x_ref[0]+(-0.939624730616)*x_ref[1]**o+(0.965857713341)*x_ref[1]+(-0.913772635427)*x_ref[2]**o+(-0.459963514384)*x_ref[2]
        ref[1,1,1,1]=(0.138896120228)*x_ref[0]**o+(0.941491395046)*x_ref[0]+(0.546538931051)*x_ref[1]**o+(-0.526098363992)*x_ref[1]+(0.32139556996)*x_ref[2]**o+(0.899346403792)*x_ref[2]
        ref[1,1,2,0]=(0.591597103411)*x_ref[0]**o+(0.926278363416)*x_ref[0]+(0.565926211046)*x_ref[1]**o+(0.930516925784)*x_ref[1]+(-0.498131779353)*x_ref[2]**o+(0.358157310681)*x_ref[2]
        ref[1,1,2,1]=(0.758395199797)*x_ref[0]**o+(0.769578766255)*x_ref[0]+(0.86819976051)*x_ref[1]**o+(-0.380436310842)*x_ref[1]+(-0.513012037048)*x_ref[2]**o+(-0.984361866072)*x_ref[2]
        ref[1,2,0,0]=(0.0845846076213)*x_ref[0]**o+(0.987145343601)*x_ref[0]+(-0.447680898266)*x_ref[1]**o+(-0.30278197108)*x_ref[1]+(0.271667503636)*x_ref[2]**o+(0.698967822656)*x_ref[2]
        ref[1,2,0,1]=(-0.314076111079)*x_ref[0]**o+(0.507407943673)*x_ref[0]+(-0.38467148586)*x_ref[1]**o+(0.802764957621)*x_ref[1]+(0.692949890402)*x_ref[2]**o+(-0.965766183849)*x_ref[2]
        ref[1,2,1,0]=(-0.855806343042)*x_ref[0]**o+(0.753362713203)*x_ref[0]+(-0.9877924012)*x_ref[1]**o+(-0.54047626663)*x_ref[1]+(-0.488859780266)*x_ref[2]**o+(0.444279382381)*x_ref[2]
        ref[1,2,1,1]=(-0.313257382113)*x_ref[0]**o+(-0.926058887186)*x_ref[0]+(0.509510780366)*x_ref[1]**o+(0.738049597453)*x_ref[1]+(0.191676625722)*x_ref[2]**o+(-0.354148482258)*x_ref[2]
        ref[1,2,2,0]=(-0.955204517346)*x_ref[0]**o+(0.234220317948)*x_ref[0]+(0.44435792198)*x_ref[1]**o+(0.96099571541)*x_ref[1]+(-0.422462745872)*x_ref[2]**o+(0.808817733192)*x_ref[2]
        ref[1,2,2,1]=(0.932454071169)*x_ref[0]**o+(-0.449841123911)*x_ref[0]+(-0.371587228288)*x_ref[1]**o+(-0.552185751938)*x_ref[1]+(-0.222070148528)*x_ref[2]**o+(-0.911768306425)*x_ref[2]
        ref[1,3,0,0]=(-0.355162862133)*x_ref[0]**o+(-0.00562439573347)*x_ref[0]+(0.982216168139)*x_ref[1]**o+(0.0493986220989)*x_ref[1]+(0.0242698649494)*x_ref[2]**o+(0.593411693569)*x_ref[2]
        ref[1,3,0,1]=(-0.982164819772)*x_ref[0]**o+(-0.417475530445)*x_ref[0]+(-0.421752719264)*x_ref[1]**o+(-0.102522772133)*x_ref[1]+(-0.384947583437)*x_ref[2]**o+(-0.675090951882)*x_ref[2]
        ref[1,3,1,0]=(0.862354670281)*x_ref[0]**o+(0.0332806244503)*x_ref[0]+(0.248152892705)*x_ref[1]**o+(-0.380941068103)*x_ref[1]+(0.499668747376)*x_ref[2]**o+(-0.732344330916)*x_ref[2]
        ref[1,3,1,1]=(-0.367669055003)*x_ref[0]**o+(-0.28761578782)*x_ref[0]+(0.10838982068)*x_ref[1]**o+(-0.495999837684)*x_ref[1]+(0.599492939122)*x_ref[2]**o+(0.886609670648)*x_ref[2]
        ref[1,3,2,0]=(0.707026122148)*x_ref[0]**o+(-0.625314188271)*x_ref[0]+(-0.734283031657)*x_ref[1]**o+(0.779848908302)*x_ref[1]+(-0.965798269237)*x_ref[2]**o+(0.35160103487)*x_ref[2]
        ref[1,3,2,1]=(0.198008431503)*x_ref[0]**o+(0.145470568903)*x_ref[0]+(-0.342584113223)*x_ref[1]**o+(-0.976390968602)*x_ref[1]+(-0.371678081685)*x_ref[2]**o+(0.569699640624)*x_ref[2]
        ref[1,4,0,0]=(-0.816642540092)*x_ref[0]**o+(0.466127552066)*x_ref[0]+(-0.491109738021)*x_ref[1]**o+(-0.177222784312)*x_ref[1]+(0.723624314914)*x_ref[2]**o+(-0.103929106752)*x_ref[2]
        ref[1,4,0,1]=(-0.461770336596)*x_ref[0]**o+(-0.416473447722)*x_ref[0]+(-0.844639456093)*x_ref[1]**o+(-0.021509624606)*x_ref[1]+(-0.507857805076)*x_ref[2]**o+(-0.0132005113079)*x_ref[2]
        ref[1,4,1,0]=(0.279614653607)*x_ref[0]**o+(-0.547332204747)*x_ref[0]+(0.329751635637)*x_ref[1]**o+(0.813149533168)*x_ref[1]+(0.0361758823053)*x_ref[2]**o+(0.0635753923527)*x_ref[2]
        ref[1,4,1,1]=(0.124940246141)*x_ref[0]**o+(-0.470715329674)*x_ref[0]+(0.401922174627)*x_ref[1]**o+(-0.292371975658)*x_ref[1]+(0.302438467505)*x_ref[2]**o+(0.131045792461)*x_ref[2]
        ref[1,4,2,0]=(0.817170444398)*x_ref[0]**o+(0.153182033447)*x_ref[0]+(-0.98908973731)*x_ref[1]**o+(0.359609127275)*x_ref[1]+(-0.891501663456)*x_ref[2]**o+(0.34700998982)*x_ref[2]
        ref[1,4,2,1]=(0.619050545861)*x_ref[0]**o+(-0.784466574079)*x_ref[0]+(-0.15316511355)*x_ref[1]**o+(0.587255937366)*x_ref[1]+(0.740644692907)*x_ref[2]**o+(0.0651236072508)*x_ref[2]
        ref[2,0,0,0]=(0.67904774175)*x_ref[0]**o+(-0.24998331377)*x_ref[0]+(-0.263635906274)*x_ref[1]**o+(0.157008949549)*x_ref[1]+(-0.905158896842)*x_ref[2]**o+(-0.975062080366)*x_ref[2]
        ref[2,0,0,1]=(0.724305963876)*x_ref[0]**o+(0.666891161317)*x_ref[0]+(0.769784501206)*x_ref[1]**o+(-0.84353242369)*x_ref[1]+(0.631055556571)*x_ref[2]**o+(-0.433528873211)*x_ref[2]
        ref[2,0,1,0]=(-0.545702871321)*x_ref[0]**o+(-0.153233364062)*x_ref[0]+(-0.892642075879)*x_ref[1]**o+(-0.303965445232)*x_ref[1]+(0.675938211305)*x_ref[2]**o+(0.134016007164)*x_ref[2]
        ref[2,0,1,1]=(0.612528479764)*x_ref[0]**o+(0.444039154606)*x_ref[0]+(-0.185281638413)*x_ref[1]**o+(0.643315773689)*x_ref[1]+(-0.5562387833)*x_ref[2]**o+(-0.960249468718)*x_ref[2]
        ref[2,0,2,0]=(-0.330135152971)*x_ref[0]**o+(0.359951176686)*x_ref[0]+(-0.797088515842)*x_ref[1]**o+(-0.257909258598)*x_ref[1]+(-0.927156080256)*x_ref[2]**o+(-0.843405252125)*x_ref[2]
        ref[2,0,2,1]=(0.447041226497)*x_ref[0]**o+(0.691225329527)*x_ref[0]+(0.462798475523)*x_ref[1]**o+(0.157497098383)*x_ref[1]+(-0.77785224695)*x_ref[2]**o+(0.886430058902)*x_ref[2]
        ref[2,1,0,0]=(0.397605326971)*x_ref[0]**o+(0.66227843262)*x_ref[0]+(-0.634922556623)*x_ref[1]**o+(-0.263342020187)*x_ref[1]+(-0.81486179124)*x_ref[2]**o+(-0.216716259698)*x_ref[2]
        ref[2,1,0,1]=(-0.868246547394)*x_ref[0]**o+(-0.194983059368)*x_ref[0]+(-0.667277750808)*x_ref[1]**o+(0.284011270958)*x_ref[1]+(-0.276295620425)*x_ref[2]**o+(0.215159516398)*x_ref[2]
        ref[2,1,1,0]=(0.990906320323)*x_ref[0]**o+(0.526754062323)*x_ref[0]+(-0.963398315893)*x_ref[1]**o+(-0.72561275544)*x_ref[1]+(0.539779065056)*x_ref[2]**o+(0.37738462043)*x_ref[2]
        ref[2,1,1,1]=(-0.668520028258)*x_ref[0]**o+(-0.292883094355)*x_ref[0]+(0.260949538172)*x_ref[1]**o+(-0.952706524522)*x_ref[1]+(-0.715386804875)*x_ref[2]**o+(-0.683320136224)*x_ref[2]
        ref[2,1,2,0]=(-0.646186301281)*x_ref[0]**o+(-0.62747987465)*x_ref[0]+(-0.858714171368)*x_ref[1]**o+(0.46910094021)*x_ref[1]+(0.468975652807)*x_ref[2]**o+(0.242115907129)*x_ref[2]
        ref[2,1,2,1]=(0.262205767991)*x_ref[0]**o+(-0.460376423099)*x_ref[0]+(0.88572453638)*x_ref[1]**o+(0.611761462581)*x_ref[1]+(0.763860971141)*x_ref[2]**o+(0.108470198525)*x_ref[2]
        ref[2,2,0,0]=(-0.0218691308052)*x_ref[0]**o+(0.942271392012)*x_ref[0]+(-0.778469966054)*x_ref[1]**o+(0.0873305333765)*x_ref[1]+(-0.300479183111)*x_ref[2]**o+(-0.862547713455)*x_ref[2]
        ref[2,2,0,1]=(-0.736808990311)*x_ref[0]**o+(-0.200478457878)*x_ref[0]+(0.272608362607)*x_ref[1]**o+(-0.0651056524306)*x_ref[1]+(0.616628495568)*x_ref[2]**o+(-0.8623490351)*x_ref[2]
        ref[2,2,1,0]=(0.931530668297)*x_ref[0]**o+(0.39701649536)*x_ref[0]+(0.552135074711)*x_ref[1]**o+(-0.489731326344)*x_ref[1]+(-0.864724191354)*x_ref[2]**o+(0.369557062583)*x_ref[2]
        ref[2,2,1,1]=(0.418780453395)*x_ref[0]**o+(-0.4935990418)*x_ref[0]+(0.34625989303)*x_ref[1]**o+(0.979829617988)*x_ref[1]+(0.853169987441)*x_ref[2]**o+(-0.945958722805)*x_ref[2]
        ref[2,2,2,0]=(0.973689940887)*x_ref[0]**o+(-0.0585784323994)*x_ref[0]+(-0.472419744635)*x_ref[1]**o+(0.851814193674)*x_ref[1]+(-0.393393340405)*x_ref[2]**o+(-0.755376551633)*x_ref[2]
        ref[2,2,2,1]=(-0.217049913648)*x_ref[0]**o+(-0.262150796255)*x_ref[0]+(-0.407805273745)*x_ref[1]**o+(0.819646964669)*x_ref[1]+(-0.0130316627236)*x_ref[2]**o+(0.179034130621)*x_ref[2]
        ref[2,3,0,0]=(-0.377708814803)*x_ref[0]**o+(-0.943893058411)*x_ref[0]+(-0.354341838264)*x_ref[1]**o+(0.0696491295689)*x_ref[1]+(-0.192838410156)*x_ref[2]**o+(-0.22208932914)*x_ref[2]
        ref[2,3,0,1]=(0.252429489766)*x_ref[0]**o+(0.808052283089)*x_ref[0]+(0.282941493712)*x_ref[1]**o+(0.89816007357)*x_ref[1]+(0.000979423783521)*x_ref[2]**o+(0.645599553602)*x_ref[2]
        ref[2,3,1,0]=(0.788089987784)*x_ref[0]**o+(0.414254044423)*x_ref[0]+(0.105204158639)*x_ref[1]**o+(0.97527041289)*x_ref[1]+(0.447463569594)*x_ref[2]**o+(0.91465488394)*x_ref[2]
        ref[2,3,1,1]=(0.882323493958)*x_ref[0]**o+(-0.98135461385)*x_ref[0]+(0.641881834427)*x_ref[1]**o+(-0.888068999894)*x_ref[1]+(0.705728558352)*x_ref[2]**o+(0.295815435834)*x_ref[2]
        ref[2,3,2,0]=(-0.570666876652)*x_ref[0]**o+(0.52308582534)*x_ref[0]+(-0.604213776925)*x_ref[1]**o+(-0.568684610916)*x_ref[1]+(-0.0378463784885)*x_ref[2]**o+(0.557994858276)*x_ref[2]
        ref[2,3,2,1]=(-0.497489149519)*x_ref[0]**o+(0.739341493998)*x_ref[0]+(-0.375869065986)*x_ref[1]**o+(-0.895553158302)*x_ref[1]+(0.916829505813)*x_ref[2]**o+(-0.650522069672)*x_ref[2]
        ref[2,4,0,0]=(0.627540843427)*x_ref[0]**o+(-0.572125230014)*x_ref[0]+(-0.404781893475)*x_ref[1]**o+(-0.114896790942)*x_ref[1]+(-0.108914764977)*x_ref[2]**o+(0.739758413931)*x_ref[2]
        ref[2,4,0,1]=(0.243205825969)*x_ref[0]**o+(-0.989602234642)*x_ref[0]+(-0.925730664677)*x_ref[1]**o+(0.454801040316)*x_ref[1]+(0.51758036389)*x_ref[2]**o+(0.325160660047)*x_ref[2]
        ref[2,4,1,0]=(0.559764013478)*x_ref[0]**o+(2.69034311577e-05)*x_ref[0]+(-0.691119937581)*x_ref[1]**o+(-0.347457112767)*x_ref[1]+(0.714683489268)*x_ref[2]**o+(-0.589270137254)*x_ref[2]
        ref[2,4,1,1]=(-0.657837524953)*x_ref[0]**o+(0.071158671662)*x_ref[0]+(-0.998677178035)*x_ref[1]**o+(-0.544136374784)*x_ref[1]+(0.6200901771)*x_ref[2]**o+(-0.280453395396)*x_ref[2]
        ref[2,4,2,0]=(0.692403743003)*x_ref[0]**o+(-0.570086669428)*x_ref[0]+(0.455036498939)*x_ref[1]**o+(0.291433179453)*x_ref[1]+(-0.203004457401)*x_ref[2]**o+(0.0192489593981)*x_ref[2]
        ref[2,4,2,1]=(-0.656783033826)*x_ref[0]**o+(0.865748949755)*x_ref[0]+(-0.297731124456)*x_ref[1]**o+(-0.826786841513)*x_ref[1]+(0.891374127144)*x_ref[2]**o+(-0.736789456498)*x_ref[2]
        ref[3,0,0,0]=(-0.963120586854)*x_ref[0]**o+(0.466089565408)*x_ref[0]+(-0.785137906849)*x_ref[1]**o+(0.194880437113)*x_ref[1]+(0.652057839639)*x_ref[2]**o+(-0.63674893761)*x_ref[2]
        ref[3,0,0,1]=(-0.923234625819)*x_ref[0]**o+(-0.303847080864)*x_ref[0]+(-0.374245465515)*x_ref[1]**o+(0.131093877726)*x_ref[1]+(-0.13469395748)*x_ref[2]**o+(0.42721676585)*x_ref[2]
        ref[3,0,1,0]=(-0.613510153481)*x_ref[0]**o+(0.986696685318)*x_ref[0]+(0.188583200151)*x_ref[1]**o+(-0.895175969251)*x_ref[1]+(-0.699510846377)*x_ref[2]**o+(0.953157679279)*x_ref[2]
        ref[3,0,1,1]=(-0.805809890231)*x_ref[0]**o+(-0.786467529592)*x_ref[0]+(-0.287240865489)*x_ref[1]**o+(-0.78714682888)*x_ref[1]+(-0.301034554724)*x_ref[2]**o+(0.698100092246)*x_ref[2]
        ref[3,0,2,0]=(-0.29320618172)*x_ref[0]**o+(0.482436772935)*x_ref[0]+(-0.126052929261)*x_ref[1]**o+(-0.0638862140753)*x_ref[1]+(0.42769411972)*x_ref[2]**o+(0.741118745829)*x_ref[2]
        ref[3,0,2,1]=(-0.380041427783)*x_ref[0]**o+(-0.0347904596008)*x_ref[0]+(-0.321599184546)*x_ref[1]**o+(-0.061333518544)*x_ref[1]+(-0.885193271364)*x_ref[2]**o+(0.216506075886)*x_ref[2]
        ref[3,1,0,0]=(-0.872457576682)*x_ref[0]**o+(0.1368749584)*x_ref[0]+(0.883209741333)*x_ref[1]**o+(0.946644762296)*x_ref[1]+(-0.941665219846)*x_ref[2]**o+(0.683787986795)*x_ref[2]
        ref[3,1,0,1]=(0.564843030077)*x_ref[0]**o+(0.987198898299)*x_ref[0]+(0.808116156661)*x_ref[1]**o+(-0.712172259479)*x_ref[1]+(0.942066890171)*x_ref[2]**o+(-0.265957158917)*x_ref[2]
        ref[3,1,1,0]=(-0.171018264631)*x_ref[0]**o+(0.277202848962)*x_ref[0]+(0.0726306773528)*x_ref[1]**o+(-0.0265640725409)*x_ref[1]+(-0.871118922929)*x_ref[2]**o+(0.0562314512261)*x_ref[2]
        ref[3,1,1,1]=(0.0861984079085)*x_ref[0]**o+(-0.954818837051)*x_ref[0]+(0.817362109646)*x_ref[1]**o+(-0.513086270222)*x_ref[1]+(-0.487472170741)*x_ref[2]**o+(0.702514192773)*x_ref[2]
        ref[3,1,2,0]=(0.272419057202)*x_ref[0]**o+(-0.0454292509397)*x_ref[0]+(-0.117471126603)*x_ref[1]**o+(0.638237510553)*x_ref[1]+(0.156858948691)*x_ref[2]**o+(0.499459702141)*x_ref[2]
        ref[3,1,2,1]=(-0.384525883334)*x_ref[0]**o+(-0.795638828759)*x_ref[0]+(-0.105795059118)*x_ref[1]**o+(0.539873566967)*x_ref[1]+(-0.402558868714)*x_ref[2]**o+(-0.690578475243)*x_ref[2]
        ref[3,2,0,0]=(0.776889517136)*x_ref[0]**o+(0.710931840891)*x_ref[0]+(0.335992677388)*x_ref[1]**o+(-0.689182863932)*x_ref[1]+(0.294149805572)*x_ref[2]**o+(-0.144223257292)*x_ref[2]
        ref[3,2,0,1]=(0.738703695733)*x_ref[0]**o+(-0.523207061995)*x_ref[0]+(0.707078352596)*x_ref[1]**o+(-0.594373978096)*x_ref[1]+(-0.833468457639)*x_ref[2]**o+(-0.60181674446)*x_ref[2]
        ref[3,2,1,0]=(-0.70243682985)*x_ref[0]**o+(-0.557914174271)*x_ref[0]+(-0.97153723053)*x_ref[1]**o+(0.791692594024)*x_ref[1]+(-0.442273808305)*x_ref[2]**o+(-0.74069605321)*x_ref[2]
        ref[3,2,1,1]=(0.136307039789)*x_ref[0]**o+(0.820661597623)*x_ref[0]+(0.379752060994)*x_ref[1]**o+(-0.739083116742)*x_ref[1]+(-0.964857055764)*x_ref[2]**o+(0.523159928671)*x_ref[2]
        ref[3,2,2,0]=(0.530883295074)*x_ref[0]**o+(-0.859538509888)*x_ref[0]+(-0.33917158254)*x_ref[1]**o+(0.476510741055)*x_ref[1]+(0.851975683024)*x_ref[2]**o+(0.146277015139)*x_ref[2]
        ref[3,2,2,1]=(-0.532984574212)*x_ref[0]**o+(-0.413169979754)*x_ref[0]+(0.0113284200501)*x_ref[1]**o+(0.406768750567)*x_ref[1]+(0.730968378807)*x_ref[2]**o+(-0.74004422909)*x_ref[2]
        ref[3,3,0,0]=(0.784793271465)*x_ref[0]**o+(0.879569037436)*x_ref[0]+(0.687804194006)*x_ref[1]**o+(0.153617608524)*x_ref[1]+(-0.410247334526)*x_ref[2]**o+(0.147941373268)*x_ref[2]
        ref[3,3,0,1]=(-0.896748874838)*x_ref[0]**o+(-0.408404228864)*x_ref[0]+(0.831967697989)*x_ref[1]**o+(-0.290875629476)*x_ref[1]+(-0.244101719785)*x_ref[2]**o+(0.240453425519)*x_ref[2]
        ref[3,3,1,0]=(-0.352293918285)*x_ref[0]**o+(0.0341233590345)*x_ref[0]+(-0.0165872699681)*x_ref[1]**o+(-0.0668507705942)*x_ref[1]+(-0.133733710061)*x_ref[2]**o+(-0.00620237772745)*x_ref[2]
        ref[3,3,1,1]=(0.173638606531)*x_ref[0]**o+(0.619705705811)*x_ref[0]+(0.279972531576)*x_ref[1]**o+(-0.0811181576037)*x_ref[1]+(-0.645673968637)*x_ref[2]**o+(0.913753411758)*x_ref[2]
        ref[3,3,2,0]=(0.0623921286142)*x_ref[0]**o+(0.393414889204)*x_ref[0]+(-0.300034851075)*x_ref[1]**o+(-0.519515206457)*x_ref[1]+(-0.0074981731879)*x_ref[2]**o+(0.202996242895)*x_ref[2]
        ref[3,3,2,1]=(0.583279597505)*x_ref[0]**o+(-0.940603487209)*x_ref[0]+(0.027966157338)*x_ref[1]**o+(0.251466344692)*x_ref[1]+(-0.97011033154)*x_ref[2]**o+(0.46527056203)*x_ref[2]
        ref[3,4,0,0]=(0.865170740218)*x_ref[0]**o+(-0.466853211289)*x_ref[0]+(0.0935492537278)*x_ref[1]**o+(-0.632374632657)*x_ref[1]+(0.673466346463)*x_ref[2]**o+(-0.47934739873)*x_ref[2]
        ref[3,4,0,1]=(0.0417451136027)*x_ref[0]**o+(-0.781570733286)*x_ref[0]+(0.547452389454)*x_ref[1]**o+(-0.0286368211528)*x_ref[1]+(-0.98907358808)*x_ref[2]**o+(0.62749495057)*x_ref[2]
        ref[3,4,1,0]=(-0.632067479664)*x_ref[0]**o+(-0.872835386336)*x_ref[0]+(-0.228785631078)*x_ref[1]**o+(-0.224206127306)*x_ref[1]+(0.769729051217)*x_ref[2]**o+(-0.406750034352)*x_ref[2]
        ref[3,4,1,1]=(0.216883071299)*x_ref[0]**o+(-0.225632200121)*x_ref[0]+(0.827186402743)*x_ref[1]**o+(0.161488636886)*x_ref[1]+(0.773801288836)*x_ref[2]**o+(0.977006832166)*x_ref[2]
        ref[3,4,2,0]=(-0.867098845884)*x_ref[0]**o+(0.14989544838)*x_ref[0]+(0.756720043262)*x_ref[1]**o+(-0.496074302626)*x_ref[1]+(-0.662787072451)*x_ref[2]**o+(-0.142473973368)*x_ref[2]
        ref[3,4,2,1]=(-0.294128396829)*x_ref[0]**o+(-0.931379392091)*x_ref[0]+(0.906552131161)*x_ref[1]**o+(-0.422008485494)*x_ref[1]+(0.806742455679)*x_ref[2]**o+(0.940488803006)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.387919945961)*x[0]**o+(-0.65988832446)*x[0]+(-0.271357984194)*x[1]**o+(0.911983947389)*x[1]
        ref=(-0.387919945961)*x_ref[0]**o+(-0.65988832446)*x_ref[0]+(-0.271357984194)*x_ref[1]**o+(0.911983947389)*x_ref[1]
      else:
        arg=(-0.0447075351093)*x[0]**o+(0.253461091432)*x[0]+(-0.218567784216)*x[1]**o+(-0.79030708754)*x[1]+(0.977690873872)*x[2]**o+(0.832406975763)*x[2]
        ref=(-0.0447075351093)*x_ref[0]**o+(0.253461091432)*x_ref[0]+(-0.218567784216)*x_ref[1]**o+(-0.79030708754)*x_ref[1]+(0.977690873872)*x_ref[2]**o+(0.832406975763)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.0351141106851)*x[0]**o+(-0.330587401)*x[0]+(0.244307738971)*x[1]**o+(0.623126601064)*x[1]
        arg[1]=(-0.828515624351)*x[0]**o+(-0.128313696693)*x[0]+(-0.419830633587)*x[1]**o+(0.207694344977)*x[1]
        ref[0]=(-0.0351141106851)*x_ref[0]**o+(-0.330587401)*x_ref[0]+(0.244307738971)*x_ref[1]**o+(0.623126601064)*x_ref[1]
        ref[1]=(-0.828515624351)*x_ref[0]**o+(-0.128313696693)*x_ref[0]+(-0.419830633587)*x_ref[1]**o+(0.207694344977)*x_ref[1]
      else:
        arg[0]=(-0.704794455532)*x[0]**o+(0.913653448977)*x[0]+(-0.159574360706)*x[1]**o+(-0.766199806615)*x[1]+(-0.571643737767)*x[2]**o+(-0.818256320426)*x[2]
        arg[1]=(-0.951104850067)*x[0]**o+(-0.661009825365)*x[0]+(0.415252265672)*x[1]**o+(-0.373834229809)*x[1]+(-0.727526942776)*x[2]**o+(-0.832051613933)*x[2]
        ref[0]=(-0.704794455532)*x_ref[0]**o+(0.913653448977)*x_ref[0]+(-0.159574360706)*x_ref[1]**o+(-0.766199806615)*x_ref[1]+(-0.571643737767)*x_ref[2]**o+(-0.818256320426)*x_ref[2]
        ref[1]=(-0.951104850067)*x_ref[0]**o+(-0.661009825365)*x_ref[0]+(0.415252265672)*x_ref[1]**o+(-0.373834229809)*x_ref[1]+(-0.727526942776)*x_ref[2]**o+(-0.832051613933)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.597205522265)*x[0]**o+(-0.990224489001)*x[0]+(0.409554155812)*x[1]**o+(0.0483989008223)*x[1]
        arg[0,1]=(0.397936023817)*x[0]**o+(-0.930167709289)*x[0]+(0.92611965278)*x[1]**o+(-0.964519543905)*x[1]
        arg[0,2]=(-0.311484417828)*x[0]**o+(-0.602916971577)*x[0]+(-0.420016358622)*x[1]**o+(0.837176635137)*x[1]
        arg[0,3]=(0.819115285277)*x[0]**o+(0.34838887804)*x[0]+(-0.690550904634)*x[1]**o+(0.373665882427)*x[1]
        arg[0,4]=(-0.7631912026)*x[0]**o+(-0.945320401033)*x[0]+(0.694292397122)*x[1]**o+(-0.420347487671)*x[1]
        arg[1,0]=(-0.572828652444)*x[0]**o+(0.354646387926)*x[0]+(-0.0507519017004)*x[1]**o+(-0.613889773716)*x[1]
        arg[1,1]=(-0.816957267715)*x[0]**o+(0.733537081373)*x[0]+(-0.660041553847)*x[1]**o+(0.441096986506)*x[1]
        arg[1,2]=(-0.803370878245)*x[0]**o+(-0.876568267674)*x[0]+(0.530143828975)*x[1]**o+(0.208130191749)*x[1]
        arg[1,3]=(-0.471900498331)*x[0]**o+(0.537812384385)*x[0]+(-0.882977296775)*x[1]**o+(-0.0668303027948)*x[1]
        arg[1,4]=(0.145525825808)*x[0]**o+(-0.743588928172)*x[0]+(-0.571360483951)*x[1]**o+(-0.779639689697)*x[1]
        arg[2,0]=(0.664624516473)*x[0]**o+(-0.903029393285)*x[0]+(-0.602274374813)*x[1]**o+(-0.0349025848298)*x[1]
        arg[2,1]=(-0.826235832601)*x[0]**o+(-0.56956830806)*x[0]+(-0.720529973236)*x[1]**o+(0.844549160654)*x[1]
        arg[2,2]=(0.865808902895)*x[0]**o+(0.806173922192)*x[0]+(-0.430500586315)*x[1]**o+(-0.903014287464)*x[1]
        arg[2,3]=(0.567909856768)*x[0]**o+(0.702374819746)*x[0]+(-0.931635013223)*x[1]**o+(0.35484582218)*x[1]
        arg[2,4]=(0.98750760384)*x[0]**o+(0.759340896388)*x[0]+(0.54913653446)*x[1]**o+(-0.620836882788)*x[1]
        arg[3,0]=(-0.886288358587)*x[0]**o+(0.393932009441)*x[0]+(0.370738890674)*x[1]**o+(0.540588748949)*x[1]
        arg[3,1]=(0.281435147449)*x[0]**o+(0.455838903759)*x[0]+(0.115176811183)*x[1]**o+(0.214335969407)*x[1]
        arg[3,2]=(-0.939824922583)*x[0]**o+(-0.725457358605)*x[0]+(0.0326844083908)*x[1]**o+(-0.743922869733)*x[1]
        arg[3,3]=(-0.189824771744)*x[0]**o+(-0.26508334021)*x[0]+(-0.323185410564)*x[1]**o+(0.242480018197)*x[1]
        arg[3,4]=(-0.675784178633)*x[0]**o+(0.0995857467734)*x[0]+(0.564599922701)*x[1]**o+(-0.716819470891)*x[1]
        ref[0,0]=(0.597205522265)*x_ref[0]**o+(-0.990224489001)*x_ref[0]+(0.409554155812)*x_ref[1]**o+(0.0483989008223)*x_ref[1]
        ref[0,1]=(0.397936023817)*x_ref[0]**o+(-0.930167709289)*x_ref[0]+(0.92611965278)*x_ref[1]**o+(-0.964519543905)*x_ref[1]
        ref[0,2]=(-0.311484417828)*x_ref[0]**o+(-0.602916971577)*x_ref[0]+(-0.420016358622)*x_ref[1]**o+(0.837176635137)*x_ref[1]
        ref[0,3]=(0.819115285277)*x_ref[0]**o+(0.34838887804)*x_ref[0]+(-0.690550904634)*x_ref[1]**o+(0.373665882427)*x_ref[1]
        ref[0,4]=(-0.7631912026)*x_ref[0]**o+(-0.945320401033)*x_ref[0]+(0.694292397122)*x_ref[1]**o+(-0.420347487671)*x_ref[1]
        ref[1,0]=(-0.572828652444)*x_ref[0]**o+(0.354646387926)*x_ref[0]+(-0.0507519017004)*x_ref[1]**o+(-0.613889773716)*x_ref[1]
        ref[1,1]=(-0.816957267715)*x_ref[0]**o+(0.733537081373)*x_ref[0]+(-0.660041553847)*x_ref[1]**o+(0.441096986506)*x_ref[1]
        ref[1,2]=(-0.803370878245)*x_ref[0]**o+(-0.876568267674)*x_ref[0]+(0.530143828975)*x_ref[1]**o+(0.208130191749)*x_ref[1]
        ref[1,3]=(-0.471900498331)*x_ref[0]**o+(0.537812384385)*x_ref[0]+(-0.882977296775)*x_ref[1]**o+(-0.0668303027948)*x_ref[1]
        ref[1,4]=(0.145525825808)*x_ref[0]**o+(-0.743588928172)*x_ref[0]+(-0.571360483951)*x_ref[1]**o+(-0.779639689697)*x_ref[1]
        ref[2,0]=(0.664624516473)*x_ref[0]**o+(-0.903029393285)*x_ref[0]+(-0.602274374813)*x_ref[1]**o+(-0.0349025848298)*x_ref[1]
        ref[2,1]=(-0.826235832601)*x_ref[0]**o+(-0.56956830806)*x_ref[0]+(-0.720529973236)*x_ref[1]**o+(0.844549160654)*x_ref[1]
        ref[2,2]=(0.865808902895)*x_ref[0]**o+(0.806173922192)*x_ref[0]+(-0.430500586315)*x_ref[1]**o+(-0.903014287464)*x_ref[1]
        ref[2,3]=(0.567909856768)*x_ref[0]**o+(0.702374819746)*x_ref[0]+(-0.931635013223)*x_ref[1]**o+(0.35484582218)*x_ref[1]
        ref[2,4]=(0.98750760384)*x_ref[0]**o+(0.759340896388)*x_ref[0]+(0.54913653446)*x_ref[1]**o+(-0.620836882788)*x_ref[1]
        ref[3,0]=(-0.886288358587)*x_ref[0]**o+(0.393932009441)*x_ref[0]+(0.370738890674)*x_ref[1]**o+(0.540588748949)*x_ref[1]
        ref[3,1]=(0.281435147449)*x_ref[0]**o+(0.455838903759)*x_ref[0]+(0.115176811183)*x_ref[1]**o+(0.214335969407)*x_ref[1]
        ref[3,2]=(-0.939824922583)*x_ref[0]**o+(-0.725457358605)*x_ref[0]+(0.0326844083908)*x_ref[1]**o+(-0.743922869733)*x_ref[1]
        ref[3,3]=(-0.189824771744)*x_ref[0]**o+(-0.26508334021)*x_ref[0]+(-0.323185410564)*x_ref[1]**o+(0.242480018197)*x_ref[1]
        ref[3,4]=(-0.675784178633)*x_ref[0]**o+(0.0995857467734)*x_ref[0]+(0.564599922701)*x_ref[1]**o+(-0.716819470891)*x_ref[1]
      else:
        arg[0,0]=(0.0429272016915)*x[0]**o+(-0.277438145968)*x[0]+(-0.663736923945)*x[1]**o+(0.754568662274)*x[1]+(-0.702929906882)*x[2]**o+(-0.744050838266)*x[2]
        arg[0,1]=(-0.245995343753)*x[0]**o+(-0.838699721329)*x[0]+(-0.266398695342)*x[1]**o+(0.116104359849)*x[1]+(-0.62406648751)*x[2]**o+(-0.85767565383)*x[2]
        arg[0,2]=(-0.969332932419)*x[0]**o+(-0.815839911463)*x[0]+(0.0532903209734)*x[1]**o+(-0.00108699727615)*x[1]+(0.142541999046)*x[2]**o+(0.971482543817)*x[2]
        arg[0,3]=(-0.73941517118)*x[0]**o+(0.728994326071)*x[0]+(0.962711932241)*x[1]**o+(-0.418607572448)*x[1]+(-0.374703265781)*x[2]**o+(-0.278661032739)*x[2]
        arg[0,4]=(-0.258020436163)*x[0]**o+(-0.0252121036831)*x[0]+(-0.756480314139)*x[1]**o+(-0.705045476776)*x[1]+(0.222309499093)*x[2]**o+(0.999237507123)*x[2]
        arg[1,0]=(-0.00753673411562)*x[0]**o+(-0.712947643092)*x[0]+(-0.946860154348)*x[1]**o+(-0.887293091702)*x[1]+(-0.57071975934)*x[2]**o+(-0.000995212472675)*x[2]
        arg[1,1]=(-0.547724214274)*x[0]**o+(-0.338492998124)*x[0]+(-0.191652366102)*x[1]**o+(-0.260315919511)*x[1]+(-0.103992070026)*x[2]**o+(-0.836638261621)*x[2]
        arg[1,2]=(-0.354711259362)*x[0]**o+(-0.916549411392)*x[0]+(0.744081321252)*x[1]**o+(-0.375805257126)*x[1]+(-0.78900115815)*x[2]**o+(-0.431116699853)*x[2]
        arg[1,3]=(0.543469030875)*x[0]**o+(-0.154316090927)*x[0]+(0.794234357287)*x[1]**o+(-0.894714087173)*x[1]+(-0.523490310245)*x[2]**o+(-0.734864830975)*x[2]
        arg[1,4]=(0.856478460898)*x[0]**o+(0.550846305146)*x[0]+(0.883099588119)*x[1]**o+(-0.928432890741)*x[1]+(0.953422014409)*x[2]**o+(-0.0295234113856)*x[2]
        arg[2,0]=(0.391420547389)*x[0]**o+(-0.0618665973255)*x[0]+(-0.536627285311)*x[1]**o+(-0.701891441965)*x[1]+(0.433098401229)*x[2]**o+(-0.821674981019)*x[2]
        arg[2,1]=(0.102455732687)*x[0]**o+(0.168078832413)*x[0]+(0.911099070324)*x[1]**o+(-0.776659572448)*x[1]+(0.942192938907)*x[2]**o+(-0.96284944292)*x[2]
        arg[2,2]=(0.677666801373)*x[0]**o+(0.579429978893)*x[0]+(-0.642666485576)*x[1]**o+(-0.0709929161913)*x[1]+(-0.686526389023)*x[2]**o+(0.0941806754766)*x[2]
        arg[2,3]=(-0.942137271102)*x[0]**o+(-0.448762007046)*x[0]+(0.403668316046)*x[1]**o+(-0.885790833279)*x[1]+(0.0448846125948)*x[2]**o+(0.123128346219)*x[2]
        arg[2,4]=(0.64615774988)*x[0]**o+(0.641971975809)*x[0]+(-0.0486200548159)*x[1]**o+(0.106284669825)*x[1]+(0.968867785284)*x[2]**o+(-0.693546581787)*x[2]
        arg[3,0]=(-0.0926306498682)*x[0]**o+(0.759541313455)*x[0]+(-0.569638025121)*x[1]**o+(0.236957551756)*x[1]+(-0.137429816873)*x[2]**o+(0.285836556676)*x[2]
        arg[3,1]=(0.0382484871512)*x[0]**o+(-0.481489628305)*x[0]+(0.809404670659)*x[1]**o+(0.124155422674)*x[1]+(-0.239216545446)*x[2]**o+(0.436110807769)*x[2]
        arg[3,2]=(-0.154044250435)*x[0]**o+(0.730285603192)*x[0]+(-0.986531989452)*x[1]**o+(0.0424454222102)*x[1]+(0.9651845632)*x[2]**o+(-0.173872544124)*x[2]
        arg[3,3]=(-0.528075843446)*x[0]**o+(0.226347788534)*x[0]+(-0.314941509835)*x[1]**o+(0.508180139398)*x[1]+(0.681251000904)*x[2]**o+(-0.442663397606)*x[2]
        arg[3,4]=(-0.335444251365)*x[0]**o+(-0.7788088839)*x[0]+(0.602538296871)*x[1]**o+(-0.440564659745)*x[1]+(-0.204809090144)*x[2]**o+(-0.449409088071)*x[2]
        ref[0,0]=(0.0429272016915)*x_ref[0]**o+(-0.277438145968)*x_ref[0]+(-0.663736923945)*x_ref[1]**o+(0.754568662274)*x_ref[1]+(-0.702929906882)*x_ref[2]**o+(-0.744050838266)*x_ref[2]
        ref[0,1]=(-0.245995343753)*x_ref[0]**o+(-0.838699721329)*x_ref[0]+(-0.266398695342)*x_ref[1]**o+(0.116104359849)*x_ref[1]+(-0.62406648751)*x_ref[2]**o+(-0.85767565383)*x_ref[2]
        ref[0,2]=(-0.969332932419)*x_ref[0]**o+(-0.815839911463)*x_ref[0]+(0.0532903209734)*x_ref[1]**o+(-0.00108699727615)*x_ref[1]+(0.142541999046)*x_ref[2]**o+(0.971482543817)*x_ref[2]
        ref[0,3]=(-0.73941517118)*x_ref[0]**o+(0.728994326071)*x_ref[0]+(0.962711932241)*x_ref[1]**o+(-0.418607572448)*x_ref[1]+(-0.374703265781)*x_ref[2]**o+(-0.278661032739)*x_ref[2]
        ref[0,4]=(-0.258020436163)*x_ref[0]**o+(-0.0252121036831)*x_ref[0]+(-0.756480314139)*x_ref[1]**o+(-0.705045476776)*x_ref[1]+(0.222309499093)*x_ref[2]**o+(0.999237507123)*x_ref[2]
        ref[1,0]=(-0.00753673411562)*x_ref[0]**o+(-0.712947643092)*x_ref[0]+(-0.946860154348)*x_ref[1]**o+(-0.887293091702)*x_ref[1]+(-0.57071975934)*x_ref[2]**o+(-0.000995212472675)*x_ref[2]
        ref[1,1]=(-0.547724214274)*x_ref[0]**o+(-0.338492998124)*x_ref[0]+(-0.191652366102)*x_ref[1]**o+(-0.260315919511)*x_ref[1]+(-0.103992070026)*x_ref[2]**o+(-0.836638261621)*x_ref[2]
        ref[1,2]=(-0.354711259362)*x_ref[0]**o+(-0.916549411392)*x_ref[0]+(0.744081321252)*x_ref[1]**o+(-0.375805257126)*x_ref[1]+(-0.78900115815)*x_ref[2]**o+(-0.431116699853)*x_ref[2]
        ref[1,3]=(0.543469030875)*x_ref[0]**o+(-0.154316090927)*x_ref[0]+(0.794234357287)*x_ref[1]**o+(-0.894714087173)*x_ref[1]+(-0.523490310245)*x_ref[2]**o+(-0.734864830975)*x_ref[2]
        ref[1,4]=(0.856478460898)*x_ref[0]**o+(0.550846305146)*x_ref[0]+(0.883099588119)*x_ref[1]**o+(-0.928432890741)*x_ref[1]+(0.953422014409)*x_ref[2]**o+(-0.0295234113856)*x_ref[2]
        ref[2,0]=(0.391420547389)*x_ref[0]**o+(-0.0618665973255)*x_ref[0]+(-0.536627285311)*x_ref[1]**o+(-0.701891441965)*x_ref[1]+(0.433098401229)*x_ref[2]**o+(-0.821674981019)*x_ref[2]
        ref[2,1]=(0.102455732687)*x_ref[0]**o+(0.168078832413)*x_ref[0]+(0.911099070324)*x_ref[1]**o+(-0.776659572448)*x_ref[1]+(0.942192938907)*x_ref[2]**o+(-0.96284944292)*x_ref[2]
        ref[2,2]=(0.677666801373)*x_ref[0]**o+(0.579429978893)*x_ref[0]+(-0.642666485576)*x_ref[1]**o+(-0.0709929161913)*x_ref[1]+(-0.686526389023)*x_ref[2]**o+(0.0941806754766)*x_ref[2]
        ref[2,3]=(-0.942137271102)*x_ref[0]**o+(-0.448762007046)*x_ref[0]+(0.403668316046)*x_ref[1]**o+(-0.885790833279)*x_ref[1]+(0.0448846125948)*x_ref[2]**o+(0.123128346219)*x_ref[2]
        ref[2,4]=(0.64615774988)*x_ref[0]**o+(0.641971975809)*x_ref[0]+(-0.0486200548159)*x_ref[1]**o+(0.106284669825)*x_ref[1]+(0.968867785284)*x_ref[2]**o+(-0.693546581787)*x_ref[2]
        ref[3,0]=(-0.0926306498682)*x_ref[0]**o+(0.759541313455)*x_ref[0]+(-0.569638025121)*x_ref[1]**o+(0.236957551756)*x_ref[1]+(-0.137429816873)*x_ref[2]**o+(0.285836556676)*x_ref[2]
        ref[3,1]=(0.0382484871512)*x_ref[0]**o+(-0.481489628305)*x_ref[0]+(0.809404670659)*x_ref[1]**o+(0.124155422674)*x_ref[1]+(-0.239216545446)*x_ref[2]**o+(0.436110807769)*x_ref[2]
        ref[3,2]=(-0.154044250435)*x_ref[0]**o+(0.730285603192)*x_ref[0]+(-0.986531989452)*x_ref[1]**o+(0.0424454222102)*x_ref[1]+(0.9651845632)*x_ref[2]**o+(-0.173872544124)*x_ref[2]
        ref[3,3]=(-0.528075843446)*x_ref[0]**o+(0.226347788534)*x_ref[0]+(-0.314941509835)*x_ref[1]**o+(0.508180139398)*x_ref[1]+(0.681251000904)*x_ref[2]**o+(-0.442663397606)*x_ref[2]
        ref[3,4]=(-0.335444251365)*x_ref[0]**o+(-0.7788088839)*x_ref[0]+(0.602538296871)*x_ref[1]**o+(-0.440564659745)*x_ref[1]+(-0.204809090144)*x_ref[2]**o+(-0.449409088071)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.88752640419)*x[0]**o+(-0.432514405615)*x[0]+(0.714823891809)*x[1]**o+(0.697307667986)*x[1]
        arg[0,0,1]=(0.738741506554)*x[0]**o+(0.188618518993)*x[0]+(-0.482898810269)*x[1]**o+(-0.914123176195)*x[1]
        arg[0,1,0]=(0.863872888691)*x[0]**o+(0.74507667202)*x[0]+(0.363690250942)*x[1]**o+(0.578470064401)*x[1]
        arg[0,1,1]=(-0.764594250812)*x[0]**o+(0.692105995894)*x[0]+(-0.950074761486)*x[1]**o+(-0.229643756622)*x[1]
        arg[1,0,0]=(-0.514123807377)*x[0]**o+(-0.770729480038)*x[0]+(-0.0503641588682)*x[1]**o+(-0.334340604627)*x[1]
        arg[1,0,1]=(-0.753427043851)*x[0]**o+(0.482153344254)*x[0]+(0.00162169870171)*x[1]**o+(-0.877830413948)*x[1]
        arg[1,1,0]=(0.603963041861)*x[0]**o+(-0.545048115282)*x[0]+(0.464500827027)*x[1]**o+(-0.527556750287)*x[1]
        arg[1,1,1]=(0.753201247064)*x[0]**o+(-0.99039225613)*x[0]+(-0.787959315932)*x[1]**o+(-0.918878072132)*x[1]
        arg[2,0,0]=(0.254002467345)*x[0]**o+(-0.918196533244)*x[0]+(-0.178408279918)*x[1]**o+(0.14954396814)*x[1]
        arg[2,0,1]=(0.449025666157)*x[0]**o+(-0.854430018688)*x[0]+(0.0393762923755)*x[1]**o+(0.576680446192)*x[1]
        arg[2,1,0]=(-0.252429553053)*x[0]**o+(-0.300398117312)*x[0]+(-0.0197102583755)*x[1]**o+(-0.189780116752)*x[1]
        arg[2,1,1]=(0.406291235871)*x[0]**o+(-0.763190166909)*x[0]+(-0.219957648446)*x[1]**o+(-0.928990138166)*x[1]
        arg[3,0,0]=(0.947321725555)*x[0]**o+(0.35310915009)*x[0]+(0.436067723051)*x[1]**o+(-0.86339979781)*x[1]
        arg[3,0,1]=(0.349458398624)*x[0]**o+(-0.139435138128)*x[0]+(-0.239809689531)*x[1]**o+(-0.0600998666403)*x[1]
        arg[3,1,0]=(-0.630539588121)*x[0]**o+(-0.700765237169)*x[0]+(-0.333988786061)*x[1]**o+(-0.125401172016)*x[1]
        arg[3,1,1]=(0.726389530806)*x[0]**o+(-0.691681620304)*x[0]+(0.241060086313)*x[1]**o+(-0.0970053499082)*x[1]
        arg[4,0,0]=(-0.798548958727)*x[0]**o+(0.429037250139)*x[0]+(0.598115431903)*x[1]**o+(-0.688464532796)*x[1]
        arg[4,0,1]=(0.698302852699)*x[0]**o+(0.36852807703)*x[0]+(-0.570086427187)*x[1]**o+(0.110387807246)*x[1]
        arg[4,1,0]=(0.295729624355)*x[0]**o+(0.247060092718)*x[0]+(-0.874774487091)*x[1]**o+(-0.40111840264)*x[1]
        arg[4,1,1]=(-0.773206505432)*x[0]**o+(-0.238827976714)*x[0]+(0.994288595399)*x[1]**o+(-0.6429945796)*x[1]
        arg[5,0,0]=(-0.398445578383)*x[0]**o+(-0.893518494161)*x[0]+(-0.855543638814)*x[1]**o+(0.93304149836)*x[1]
        arg[5,0,1]=(-0.811797150327)*x[0]**o+(0.0702614194127)*x[0]+(-0.0906132870911)*x[1]**o+(-0.944409144857)*x[1]
        arg[5,1,0]=(0.710407361984)*x[0]**o+(-0.854493576437)*x[0]+(-0.354683789897)*x[1]**o+(-0.896605671415)*x[1]
        arg[5,1,1]=(0.0668604398168)*x[0]**o+(0.464597275769)*x[0]+(-0.683756009627)*x[1]**o+(-0.368806656344)*x[1]
        ref[0,0,0]=(0.88752640419)*x_ref[0]**o+(-0.432514405615)*x_ref[0]+(0.714823891809)*x_ref[1]**o+(0.697307667986)*x_ref[1]
        ref[0,0,1]=(0.738741506554)*x_ref[0]**o+(0.188618518993)*x_ref[0]+(-0.482898810269)*x_ref[1]**o+(-0.914123176195)*x_ref[1]
        ref[0,1,0]=(0.863872888691)*x_ref[0]**o+(0.74507667202)*x_ref[0]+(0.363690250942)*x_ref[1]**o+(0.578470064401)*x_ref[1]
        ref[0,1,1]=(-0.764594250812)*x_ref[0]**o+(0.692105995894)*x_ref[0]+(-0.950074761486)*x_ref[1]**o+(-0.229643756622)*x_ref[1]
        ref[1,0,0]=(-0.514123807377)*x_ref[0]**o+(-0.770729480038)*x_ref[0]+(-0.0503641588682)*x_ref[1]**o+(-0.334340604627)*x_ref[1]
        ref[1,0,1]=(-0.753427043851)*x_ref[0]**o+(0.482153344254)*x_ref[0]+(0.00162169870171)*x_ref[1]**o+(-0.877830413948)*x_ref[1]
        ref[1,1,0]=(0.603963041861)*x_ref[0]**o+(-0.545048115282)*x_ref[0]+(0.464500827027)*x_ref[1]**o+(-0.527556750287)*x_ref[1]
        ref[1,1,1]=(0.753201247064)*x_ref[0]**o+(-0.99039225613)*x_ref[0]+(-0.787959315932)*x_ref[1]**o+(-0.918878072132)*x_ref[1]
        ref[2,0,0]=(0.254002467345)*x_ref[0]**o+(-0.918196533244)*x_ref[0]+(-0.178408279918)*x_ref[1]**o+(0.14954396814)*x_ref[1]
        ref[2,0,1]=(0.449025666157)*x_ref[0]**o+(-0.854430018688)*x_ref[0]+(0.0393762923755)*x_ref[1]**o+(0.576680446192)*x_ref[1]
        ref[2,1,0]=(-0.252429553053)*x_ref[0]**o+(-0.300398117312)*x_ref[0]+(-0.0197102583755)*x_ref[1]**o+(-0.189780116752)*x_ref[1]
        ref[2,1,1]=(0.406291235871)*x_ref[0]**o+(-0.763190166909)*x_ref[0]+(-0.219957648446)*x_ref[1]**o+(-0.928990138166)*x_ref[1]
        ref[3,0,0]=(0.947321725555)*x_ref[0]**o+(0.35310915009)*x_ref[0]+(0.436067723051)*x_ref[1]**o+(-0.86339979781)*x_ref[1]
        ref[3,0,1]=(0.349458398624)*x_ref[0]**o+(-0.139435138128)*x_ref[0]+(-0.239809689531)*x_ref[1]**o+(-0.0600998666403)*x_ref[1]
        ref[3,1,0]=(-0.630539588121)*x_ref[0]**o+(-0.700765237169)*x_ref[0]+(-0.333988786061)*x_ref[1]**o+(-0.125401172016)*x_ref[1]
        ref[3,1,1]=(0.726389530806)*x_ref[0]**o+(-0.691681620304)*x_ref[0]+(0.241060086313)*x_ref[1]**o+(-0.0970053499082)*x_ref[1]
        ref[4,0,0]=(-0.798548958727)*x_ref[0]**o+(0.429037250139)*x_ref[0]+(0.598115431903)*x_ref[1]**o+(-0.688464532796)*x_ref[1]
        ref[4,0,1]=(0.698302852699)*x_ref[0]**o+(0.36852807703)*x_ref[0]+(-0.570086427187)*x_ref[1]**o+(0.110387807246)*x_ref[1]
        ref[4,1,0]=(0.295729624355)*x_ref[0]**o+(0.247060092718)*x_ref[0]+(-0.874774487091)*x_ref[1]**o+(-0.40111840264)*x_ref[1]
        ref[4,1,1]=(-0.773206505432)*x_ref[0]**o+(-0.238827976714)*x_ref[0]+(0.994288595399)*x_ref[1]**o+(-0.6429945796)*x_ref[1]
        ref[5,0,0]=(-0.398445578383)*x_ref[0]**o+(-0.893518494161)*x_ref[0]+(-0.855543638814)*x_ref[1]**o+(0.93304149836)*x_ref[1]
        ref[5,0,1]=(-0.811797150327)*x_ref[0]**o+(0.0702614194127)*x_ref[0]+(-0.0906132870911)*x_ref[1]**o+(-0.944409144857)*x_ref[1]
        ref[5,1,0]=(0.710407361984)*x_ref[0]**o+(-0.854493576437)*x_ref[0]+(-0.354683789897)*x_ref[1]**o+(-0.896605671415)*x_ref[1]
        ref[5,1,1]=(0.0668604398168)*x_ref[0]**o+(0.464597275769)*x_ref[0]+(-0.683756009627)*x_ref[1]**o+(-0.368806656344)*x_ref[1]
      else:
        arg[0,0,0]=(-0.0873708254613)*x[0]**o+(-0.843989447866)*x[0]+(0.44751871924)*x[1]**o+(0.56511771844)*x[1]+(0.367414341304)*x[2]**o+(-0.147729223167)*x[2]
        arg[0,0,1]=(0.00752707306425)*x[0]**o+(0.819213438452)*x[0]+(0.14826826627)*x[1]**o+(-0.169362794728)*x[1]+(0.51736237082)*x[2]**o+(0.832746819884)*x[2]
        arg[0,1,0]=(0.607746431545)*x[0]**o+(0.253170100132)*x[0]+(-0.992132084299)*x[1]**o+(0.353865695431)*x[1]+(-0.787229685473)*x[2]**o+(0.86154695401)*x[2]
        arg[0,1,1]=(0.563636829468)*x[0]**o+(0.459627185677)*x[0]+(-0.511957702356)*x[1]**o+(0.309573033639)*x[1]+(0.806861185555)*x[2]**o+(0.182177897377)*x[2]
        arg[1,0,0]=(0.600850271814)*x[0]**o+(0.440922206975)*x[0]+(0.0307377466119)*x[1]**o+(-0.419096719454)*x[1]+(0.979284456332)*x[2]**o+(-0.00240732351601)*x[2]
        arg[1,0,1]=(0.51273517393)*x[0]**o+(-0.947643736857)*x[0]+(0.272403256871)*x[1]**o+(0.0119584891405)*x[1]+(-0.290604573286)*x[2]**o+(0.822291254588)*x[2]
        arg[1,1,0]=(-0.201909330605)*x[0]**o+(-0.560221198146)*x[0]+(0.564306911695)*x[1]**o+(-0.936933580593)*x[1]+(-0.41391834592)*x[2]**o+(0.133244793725)*x[2]
        arg[1,1,1]=(0.13318221571)*x[0]**o+(-0.565278280388)*x[0]+(0.677613822728)*x[1]**o+(-0.386612668638)*x[1]+(0.207570026134)*x[2]**o+(0.798362457061)*x[2]
        arg[2,0,0]=(-0.3777906845)*x[0]**o+(-0.285455928514)*x[0]+(-0.0901885959293)*x[1]**o+(-0.385922214051)*x[1]+(0.322566508003)*x[2]**o+(0.171191301702)*x[2]
        arg[2,0,1]=(-0.113954650916)*x[0]**o+(-0.373563461907)*x[0]+(0.73321196284)*x[1]**o+(0.535150793249)*x[1]+(-0.736944901672)*x[2]**o+(0.86133553525)*x[2]
        arg[2,1,0]=(-0.724228484035)*x[0]**o+(0.310971627294)*x[0]+(-0.497468717028)*x[1]**o+(0.467315341957)*x[1]+(-0.598941984951)*x[2]**o+(0.990367400534)*x[2]
        arg[2,1,1]=(0.594502107684)*x[0]**o+(0.261477996581)*x[0]+(0.738413631986)*x[1]**o+(0.363468523272)*x[1]+(-0.945845846052)*x[2]**o+(0.281144164925)*x[2]
        arg[3,0,0]=(0.804174935673)*x[0]**o+(-0.633954348085)*x[0]+(-0.953193492156)*x[1]**o+(0.632681376482)*x[1]+(-0.409900341362)*x[2]**o+(0.951988363964)*x[2]
        arg[3,0,1]=(-0.469895421121)*x[0]**o+(0.666115304582)*x[0]+(0.676351738247)*x[1]**o+(-0.0738256360284)*x[1]+(0.34976675523)*x[2]**o+(-0.561058490613)*x[2]
        arg[3,1,0]=(-0.999916478236)*x[0]**o+(-0.362871459365)*x[0]+(0.645604661917)*x[1]**o+(0.215878330855)*x[1]+(-0.0561173996823)*x[2]**o+(0.272143302204)*x[2]
        arg[3,1,1]=(0.233461828376)*x[0]**o+(-0.110673705352)*x[0]+(0.699710673197)*x[1]**o+(-0.628805264645)*x[1]+(0.677467640352)*x[2]**o+(-0.156224522726)*x[2]
        arg[4,0,0]=(0.180468931719)*x[0]**o+(0.299451524372)*x[0]+(-0.677131137371)*x[1]**o+(-0.422802649252)*x[1]+(0.237349200367)*x[2]**o+(0.739758623757)*x[2]
        arg[4,0,1]=(0.22209981155)*x[0]**o+(-0.226751058136)*x[0]+(-0.0345697170216)*x[1]**o+(0.265795819906)*x[1]+(0.542547943322)*x[2]**o+(0.785610668157)*x[2]
        arg[4,1,0]=(-0.520898604151)*x[0]**o+(-0.84307105084)*x[0]+(0.818061645146)*x[1]**o+(-0.173710357379)*x[1]+(0.344023171063)*x[2]**o+(0.194662166094)*x[2]
        arg[4,1,1]=(-0.288466404824)*x[0]**o+(-0.891856587908)*x[0]+(0.378329571068)*x[1]**o+(0.437919935526)*x[1]+(-0.630733422962)*x[2]**o+(0.518444630882)*x[2]
        arg[5,0,0]=(0.00978685202942)*x[0]**o+(-0.474443494209)*x[0]+(-0.509964215112)*x[1]**o+(0.341199671453)*x[1]+(-0.459333887251)*x[2]**o+(0.491961590667)*x[2]
        arg[5,0,1]=(-0.899850325159)*x[0]**o+(-0.191369137667)*x[0]+(0.18882430985)*x[1]**o+(-0.185214597442)*x[1]+(-0.967187797553)*x[2]**o+(0.366127839986)*x[2]
        arg[5,1,0]=(0.428007441668)*x[0]**o+(0.920828656579)*x[0]+(-0.298863178441)*x[1]**o+(-0.226020983335)*x[1]+(-0.01907141259)*x[2]**o+(-0.359698040557)*x[2]
        arg[5,1,1]=(0.980673633441)*x[0]**o+(0.362357489872)*x[0]+(0.825233899658)*x[1]**o+(0.846433827949)*x[1]+(-0.870197174533)*x[2]**o+(0.621884204716)*x[2]
        ref[0,0,0]=(-0.0873708254613)*x_ref[0]**o+(-0.843989447866)*x_ref[0]+(0.44751871924)*x_ref[1]**o+(0.56511771844)*x_ref[1]+(0.367414341304)*x_ref[2]**o+(-0.147729223167)*x_ref[2]
        ref[0,0,1]=(0.00752707306425)*x_ref[0]**o+(0.819213438452)*x_ref[0]+(0.14826826627)*x_ref[1]**o+(-0.169362794728)*x_ref[1]+(0.51736237082)*x_ref[2]**o+(0.832746819884)*x_ref[2]
        ref[0,1,0]=(0.607746431545)*x_ref[0]**o+(0.253170100132)*x_ref[0]+(-0.992132084299)*x_ref[1]**o+(0.353865695431)*x_ref[1]+(-0.787229685473)*x_ref[2]**o+(0.86154695401)*x_ref[2]
        ref[0,1,1]=(0.563636829468)*x_ref[0]**o+(0.459627185677)*x_ref[0]+(-0.511957702356)*x_ref[1]**o+(0.309573033639)*x_ref[1]+(0.806861185555)*x_ref[2]**o+(0.182177897377)*x_ref[2]
        ref[1,0,0]=(0.600850271814)*x_ref[0]**o+(0.440922206975)*x_ref[0]+(0.0307377466119)*x_ref[1]**o+(-0.419096719454)*x_ref[1]+(0.979284456332)*x_ref[2]**o+(-0.00240732351601)*x_ref[2]
        ref[1,0,1]=(0.51273517393)*x_ref[0]**o+(-0.947643736857)*x_ref[0]+(0.272403256871)*x_ref[1]**o+(0.0119584891405)*x_ref[1]+(-0.290604573286)*x_ref[2]**o+(0.822291254588)*x_ref[2]
        ref[1,1,0]=(-0.201909330605)*x_ref[0]**o+(-0.560221198146)*x_ref[0]+(0.564306911695)*x_ref[1]**o+(-0.936933580593)*x_ref[1]+(-0.41391834592)*x_ref[2]**o+(0.133244793725)*x_ref[2]
        ref[1,1,1]=(0.13318221571)*x_ref[0]**o+(-0.565278280388)*x_ref[0]+(0.677613822728)*x_ref[1]**o+(-0.386612668638)*x_ref[1]+(0.207570026134)*x_ref[2]**o+(0.798362457061)*x_ref[2]
        ref[2,0,0]=(-0.3777906845)*x_ref[0]**o+(-0.285455928514)*x_ref[0]+(-0.0901885959293)*x_ref[1]**o+(-0.385922214051)*x_ref[1]+(0.322566508003)*x_ref[2]**o+(0.171191301702)*x_ref[2]
        ref[2,0,1]=(-0.113954650916)*x_ref[0]**o+(-0.373563461907)*x_ref[0]+(0.73321196284)*x_ref[1]**o+(0.535150793249)*x_ref[1]+(-0.736944901672)*x_ref[2]**o+(0.86133553525)*x_ref[2]
        ref[2,1,0]=(-0.724228484035)*x_ref[0]**o+(0.310971627294)*x_ref[0]+(-0.497468717028)*x_ref[1]**o+(0.467315341957)*x_ref[1]+(-0.598941984951)*x_ref[2]**o+(0.990367400534)*x_ref[2]
        ref[2,1,1]=(0.594502107684)*x_ref[0]**o+(0.261477996581)*x_ref[0]+(0.738413631986)*x_ref[1]**o+(0.363468523272)*x_ref[1]+(-0.945845846052)*x_ref[2]**o+(0.281144164925)*x_ref[2]
        ref[3,0,0]=(0.804174935673)*x_ref[0]**o+(-0.633954348085)*x_ref[0]+(-0.953193492156)*x_ref[1]**o+(0.632681376482)*x_ref[1]+(-0.409900341362)*x_ref[2]**o+(0.951988363964)*x_ref[2]
        ref[3,0,1]=(-0.469895421121)*x_ref[0]**o+(0.666115304582)*x_ref[0]+(0.676351738247)*x_ref[1]**o+(-0.0738256360284)*x_ref[1]+(0.34976675523)*x_ref[2]**o+(-0.561058490613)*x_ref[2]
        ref[3,1,0]=(-0.999916478236)*x_ref[0]**o+(-0.362871459365)*x_ref[0]+(0.645604661917)*x_ref[1]**o+(0.215878330855)*x_ref[1]+(-0.0561173996823)*x_ref[2]**o+(0.272143302204)*x_ref[2]
        ref[3,1,1]=(0.233461828376)*x_ref[0]**o+(-0.110673705352)*x_ref[0]+(0.699710673197)*x_ref[1]**o+(-0.628805264645)*x_ref[1]+(0.677467640352)*x_ref[2]**o+(-0.156224522726)*x_ref[2]
        ref[4,0,0]=(0.180468931719)*x_ref[0]**o+(0.299451524372)*x_ref[0]+(-0.677131137371)*x_ref[1]**o+(-0.422802649252)*x_ref[1]+(0.237349200367)*x_ref[2]**o+(0.739758623757)*x_ref[2]
        ref[4,0,1]=(0.22209981155)*x_ref[0]**o+(-0.226751058136)*x_ref[0]+(-0.0345697170216)*x_ref[1]**o+(0.265795819906)*x_ref[1]+(0.542547943322)*x_ref[2]**o+(0.785610668157)*x_ref[2]
        ref[4,1,0]=(-0.520898604151)*x_ref[0]**o+(-0.84307105084)*x_ref[0]+(0.818061645146)*x_ref[1]**o+(-0.173710357379)*x_ref[1]+(0.344023171063)*x_ref[2]**o+(0.194662166094)*x_ref[2]
        ref[4,1,1]=(-0.288466404824)*x_ref[0]**o+(-0.891856587908)*x_ref[0]+(0.378329571068)*x_ref[1]**o+(0.437919935526)*x_ref[1]+(-0.630733422962)*x_ref[2]**o+(0.518444630882)*x_ref[2]
        ref[5,0,0]=(0.00978685202942)*x_ref[0]**o+(-0.474443494209)*x_ref[0]+(-0.509964215112)*x_ref[1]**o+(0.341199671453)*x_ref[1]+(-0.459333887251)*x_ref[2]**o+(0.491961590667)*x_ref[2]
        ref[5,0,1]=(-0.899850325159)*x_ref[0]**o+(-0.191369137667)*x_ref[0]+(0.18882430985)*x_ref[1]**o+(-0.185214597442)*x_ref[1]+(-0.967187797553)*x_ref[2]**o+(0.366127839986)*x_ref[2]
        ref[5,1,0]=(0.428007441668)*x_ref[0]**o+(0.920828656579)*x_ref[0]+(-0.298863178441)*x_ref[1]**o+(-0.226020983335)*x_ref[1]+(-0.01907141259)*x_ref[2]**o+(-0.359698040557)*x_ref[2]
        ref[5,1,1]=(0.980673633441)*x_ref[0]**o+(0.362357489872)*x_ref[0]+(0.825233899658)*x_ref[1]**o+(0.846433827949)*x_ref[1]+(-0.870197174533)*x_ref[2]**o+(0.621884204716)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.93572957134)*x[0]**o+(-0.878081515023)*x[0]+(0.821839395118)*x[1]**o+(-0.088026044081)*x[1]
        arg[0,0,0,1]=(-0.00393438000839)*x[0]**o+(-0.180693225971)*x[0]+(0.192781914487)*x[1]**o+(-0.561366977374)*x[1]
        arg[0,0,1,0]=(0.116028558326)*x[0]**o+(-0.47739282219)*x[0]+(0.97806148353)*x[1]**o+(0.660972051388)*x[1]
        arg[0,0,1,1]=(0.092683552469)*x[0]**o+(0.22083599745)*x[0]+(-0.423073290439)*x[1]**o+(-0.0789522599045)*x[1]
        arg[0,0,2,0]=(0.140023403084)*x[0]**o+(-0.927034796519)*x[0]+(0.0987957525564)*x[1]**o+(0.0703695028562)*x[1]
        arg[0,0,2,1]=(-0.00485147064392)*x[0]**o+(-0.649679595932)*x[0]+(0.503435626816)*x[1]**o+(0.393293792881)*x[1]
        arg[0,1,0,0]=(-0.753383897772)*x[0]**o+(-0.988499439357)*x[0]+(-0.651037417175)*x[1]**o+(0.45184892046)*x[1]
        arg[0,1,0,1]=(0.0120892569605)*x[0]**o+(0.52287896209)*x[0]+(0.103256695817)*x[1]**o+(-0.0244835681092)*x[1]
        arg[0,1,1,0]=(-0.0949239075135)*x[0]**o+(0.835062318231)*x[0]+(-0.0522981103769)*x[1]**o+(0.60020994838)*x[1]
        arg[0,1,1,1]=(0.139896941482)*x[0]**o+(0.409042945999)*x[0]+(0.10019215394)*x[1]**o+(-0.312224123017)*x[1]
        arg[0,1,2,0]=(-0.892437964342)*x[0]**o+(0.0828807279746)*x[0]+(0.777811646235)*x[1]**o+(-0.96509471481)*x[1]
        arg[0,1,2,1]=(-0.397341226148)*x[0]**o+(0.102987581336)*x[0]+(-0.69165111598)*x[1]**o+(0.743298211061)*x[1]
        arg[0,2,0,0]=(-0.300129864031)*x[0]**o+(-0.418981381232)*x[0]+(0.624854922555)*x[1]**o+(0.892651746561)*x[1]
        arg[0,2,0,1]=(-0.511463603881)*x[0]**o+(-0.808468332238)*x[0]+(-0.342755248919)*x[1]**o+(0.357985875808)*x[1]
        arg[0,2,1,0]=(-0.269520452932)*x[0]**o+(0.869553937081)*x[0]+(-0.183222122614)*x[1]**o+(-0.135280721586)*x[1]
        arg[0,2,1,1]=(0.0396712123969)*x[0]**o+(0.344314072793)*x[0]+(0.644676507168)*x[1]**o+(-0.0799551435053)*x[1]
        arg[0,2,2,0]=(0.302628743334)*x[0]**o+(-0.59557047376)*x[0]+(-0.277654790212)*x[1]**o+(0.0299035042863)*x[1]
        arg[0,2,2,1]=(0.805296568248)*x[0]**o+(0.726404711154)*x[0]+(0.0728356962577)*x[1]**o+(0.199997237962)*x[1]
        arg[0,3,0,0]=(-0.292658800952)*x[0]**o+(-0.40080106651)*x[0]+(0.844109590553)*x[1]**o+(-0.339364790362)*x[1]
        arg[0,3,0,1]=(-0.621465058894)*x[0]**o+(-0.376804350545)*x[0]+(0.745038406103)*x[1]**o+(-0.281482724757)*x[1]
        arg[0,3,1,0]=(0.00625816347702)*x[0]**o+(0.392750189511)*x[0]+(-0.232258708605)*x[1]**o+(-0.634552067635)*x[1]
        arg[0,3,1,1]=(-0.0893109545391)*x[0]**o+(-0.735678058768)*x[0]+(-0.603248350212)*x[1]**o+(-0.664366286468)*x[1]
        arg[0,3,2,0]=(-0.0539114688775)*x[0]**o+(-0.110100592496)*x[0]+(-0.700496078261)*x[1]**o+(0.203311460549)*x[1]
        arg[0,3,2,1]=(-0.554691350426)*x[0]**o+(-0.255001039659)*x[0]+(0.567977217289)*x[1]**o+(0.61993007447)*x[1]
        arg[0,4,0,0]=(-0.12193490862)*x[0]**o+(0.731139340661)*x[0]+(-0.321971724057)*x[1]**o+(0.741215588897)*x[1]
        arg[0,4,0,1]=(-0.394743164913)*x[0]**o+(-0.602249917929)*x[0]+(0.27675734517)*x[1]**o+(-0.968806169573)*x[1]
        arg[0,4,1,0]=(-0.629331232169)*x[0]**o+(0.687483125805)*x[0]+(0.633698693186)*x[1]**o+(0.837564332901)*x[1]
        arg[0,4,1,1]=(0.0470601891829)*x[0]**o+(-0.365490813497)*x[0]+(-0.162532786979)*x[1]**o+(0.504146804122)*x[1]
        arg[0,4,2,0]=(0.193790444091)*x[0]**o+(0.742132840869)*x[0]+(-0.871677496589)*x[1]**o+(0.492817750307)*x[1]
        arg[0,4,2,1]=(0.741218732478)*x[0]**o+(-0.746973604492)*x[0]+(-0.548073707926)*x[1]**o+(-0.42258185128)*x[1]
        arg[1,0,0,0]=(0.580989627459)*x[0]**o+(-0.498423525574)*x[0]+(-0.349651580594)*x[1]**o+(0.0193060779482)*x[1]
        arg[1,0,0,1]=(0.398917011903)*x[0]**o+(-0.122064497238)*x[0]+(0.924654557299)*x[1]**o+(0.485284066068)*x[1]
        arg[1,0,1,0]=(-0.985949382826)*x[0]**o+(-0.0188384874977)*x[0]+(0.177995429697)*x[1]**o+(0.967727833143)*x[1]
        arg[1,0,1,1]=(0.920799980503)*x[0]**o+(-0.749281119074)*x[0]+(-0.628411322607)*x[1]**o+(-0.446396361734)*x[1]
        arg[1,0,2,0]=(0.752863123667)*x[0]**o+(0.112427175329)*x[0]+(0.975568251269)*x[1]**o+(0.316938392669)*x[1]
        arg[1,0,2,1]=(0.798409861933)*x[0]**o+(-0.145715624761)*x[0]+(-0.847813831553)*x[1]**o+(0.302999756348)*x[1]
        arg[1,1,0,0]=(0.897029838502)*x[0]**o+(-0.540852276529)*x[0]+(-0.541450963121)*x[1]**o+(-0.118819043917)*x[1]
        arg[1,1,0,1]=(-0.616797542123)*x[0]**o+(-0.139394144428)*x[0]+(-0.480348413917)*x[1]**o+(0.500244829288)*x[1]
        arg[1,1,1,0]=(-0.960388730531)*x[0]**o+(0.87391147728)*x[0]+(0.0128673975991)*x[1]**o+(0.444341296582)*x[1]
        arg[1,1,1,1]=(0.136810869097)*x[0]**o+(-0.544756866447)*x[0]+(-0.198052789907)*x[1]**o+(-0.536732583929)*x[1]
        arg[1,1,2,0]=(-0.808981135979)*x[0]**o+(-0.122088197438)*x[0]+(-0.363958821187)*x[1]**o+(-0.521614066151)*x[1]
        arg[1,1,2,1]=(0.299382605072)*x[0]**o+(0.143397060894)*x[0]+(-0.223044874711)*x[1]**o+(-0.04723866032)*x[1]
        arg[1,2,0,0]=(-0.568967551455)*x[0]**o+(0.750777908696)*x[0]+(0.203117458853)*x[1]**o+(-0.0797168995572)*x[1]
        arg[1,2,0,1]=(0.0439147552635)*x[0]**o+(-0.0746749889653)*x[0]+(0.192885940929)*x[1]**o+(0.794090802345)*x[1]
        arg[1,2,1,0]=(0.137177876023)*x[0]**o+(0.625847668003)*x[0]+(0.252761858157)*x[1]**o+(0.496512978117)*x[1]
        arg[1,2,1,1]=(0.475218888846)*x[0]**o+(-0.286926184073)*x[0]+(0.551427803487)*x[1]**o+(0.782373661286)*x[1]
        arg[1,2,2,0]=(0.903004502545)*x[0]**o+(-0.262952251654)*x[0]+(0.110351860812)*x[1]**o+(-0.685355934696)*x[1]
        arg[1,2,2,1]=(-0.157457317404)*x[0]**o+(-0.243850623177)*x[0]+(0.28853283432)*x[1]**o+(0.612508908614)*x[1]
        arg[1,3,0,0]=(-0.741890002973)*x[0]**o+(-0.118262424825)*x[0]+(0.567100734646)*x[1]**o+(0.301557125811)*x[1]
        arg[1,3,0,1]=(-0.00745396583928)*x[0]**o+(-0.00558491344276)*x[0]+(-0.967132370197)*x[1]**o+(-0.61737594402)*x[1]
        arg[1,3,1,0]=(0.0767961523966)*x[0]**o+(0.341952000996)*x[0]+(0.762617832019)*x[1]**o+(-0.38337284713)*x[1]
        arg[1,3,1,1]=(0.32265018902)*x[0]**o+(-0.352521736206)*x[0]+(-0.0119143186529)*x[1]**o+(-0.64727561829)*x[1]
        arg[1,3,2,0]=(0.627690584587)*x[0]**o+(0.0993819822193)*x[0]+(-0.189363448841)*x[1]**o+(0.49004001194)*x[1]
        arg[1,3,2,1]=(-0.657836476156)*x[0]**o+(0.708970470136)*x[0]+(0.252190020576)*x[1]**o+(-0.481214196298)*x[1]
        arg[1,4,0,0]=(0.0837563164179)*x[0]**o+(0.741536525079)*x[0]+(0.645543094286)*x[1]**o+(0.62358090815)*x[1]
        arg[1,4,0,1]=(0.262043398371)*x[0]**o+(-0.47794750936)*x[0]+(0.93086244965)*x[1]**o+(-0.670399062677)*x[1]
        arg[1,4,1,0]=(-0.76634902923)*x[0]**o+(-0.809846143928)*x[0]+(0.639099549305)*x[1]**o+(0.793524639236)*x[1]
        arg[1,4,1,1]=(0.814737503957)*x[0]**o+(0.775593374042)*x[0]+(0.240976904111)*x[1]**o+(0.2352421356)*x[1]
        arg[1,4,2,0]=(0.519210584047)*x[0]**o+(-0.702579589691)*x[0]+(-0.450156317853)*x[1]**o+(0.588864692581)*x[1]
        arg[1,4,2,1]=(-0.927751546037)*x[0]**o+(-0.239737576019)*x[0]+(-0.527658440447)*x[1]**o+(0.79623440559)*x[1]
        arg[2,0,0,0]=(0.719144336419)*x[0]**o+(0.916780231277)*x[0]+(0.279073450277)*x[1]**o+(-0.0130671321167)*x[1]
        arg[2,0,0,1]=(-0.556604998962)*x[0]**o+(-0.661580370137)*x[0]+(0.980704664791)*x[1]**o+(0.935843251131)*x[1]
        arg[2,0,1,0]=(0.85501679753)*x[0]**o+(-0.908342467889)*x[0]+(0.969030151294)*x[1]**o+(0.534192961922)*x[1]
        arg[2,0,1,1]=(0.527672385529)*x[0]**o+(0.997699955667)*x[0]+(-0.388249629327)*x[1]**o+(-0.714956937512)*x[1]
        arg[2,0,2,0]=(-0.629149810667)*x[0]**o+(-0.0515018376692)*x[0]+(-0.80355991702)*x[1]**o+(0.235999988886)*x[1]
        arg[2,0,2,1]=(0.960715934156)*x[0]**o+(-0.782787305497)*x[0]+(-0.309348231959)*x[1]**o+(0.854911294717)*x[1]
        arg[2,1,0,0]=(-0.560749125074)*x[0]**o+(-0.120515615757)*x[0]+(0.055090634431)*x[1]**o+(-0.539205949358)*x[1]
        arg[2,1,0,1]=(-0.762113417295)*x[0]**o+(-0.207196980376)*x[0]+(0.904541245494)*x[1]**o+(0.320359459939)*x[1]
        arg[2,1,1,0]=(-0.272627586486)*x[0]**o+(-0.885987982522)*x[0]+(0.641737808216)*x[1]**o+(0.358390367373)*x[1]
        arg[2,1,1,1]=(-0.847502563761)*x[0]**o+(-0.430298277492)*x[0]+(-0.869043300185)*x[1]**o+(0.584700145915)*x[1]
        arg[2,1,2,0]=(0.497688368946)*x[0]**o+(-0.14413770704)*x[0]+(-0.82002941888)*x[1]**o+(0.573569419279)*x[1]
        arg[2,1,2,1]=(0.959549623603)*x[0]**o+(0.881196513588)*x[0]+(-0.202951096266)*x[1]**o+(-9.87983182765e-05)*x[1]
        arg[2,2,0,0]=(0.0811393810501)*x[0]**o+(-0.853555165896)*x[0]+(0.0259257246335)*x[1]**o+(0.638575306959)*x[1]
        arg[2,2,0,1]=(-0.525990031379)*x[0]**o+(0.11235160765)*x[0]+(-0.502853879455)*x[1]**o+(0.414066683507)*x[1]
        arg[2,2,1,0]=(-0.533919716834)*x[0]**o+(-0.773411100656)*x[0]+(0.311424028491)*x[1]**o+(0.332733301377)*x[1]
        arg[2,2,1,1]=(0.280548593462)*x[0]**o+(-0.0747640762104)*x[0]+(-0.7110232877)*x[1]**o+(-0.178672817441)*x[1]
        arg[2,2,2,0]=(0.445025926846)*x[0]**o+(-0.669782903005)*x[0]+(0.236610411793)*x[1]**o+(0.437836936476)*x[1]
        arg[2,2,2,1]=(-0.503761305098)*x[0]**o+(-0.889922659865)*x[0]+(-0.0767256016402)*x[1]**o+(-0.614874217835)*x[1]
        arg[2,3,0,0]=(-0.041960855672)*x[0]**o+(-0.593486242764)*x[0]+(0.344751892525)*x[1]**o+(-0.131268873786)*x[1]
        arg[2,3,0,1]=(-0.192704843083)*x[0]**o+(0.325356381696)*x[0]+(-0.235690651937)*x[1]**o+(0.59113535617)*x[1]
        arg[2,3,1,0]=(-0.0794745628024)*x[0]**o+(0.7996737586)*x[0]+(-0.324970505375)*x[1]**o+(-0.14920403468)*x[1]
        arg[2,3,1,1]=(0.31997432673)*x[0]**o+(-0.571739807799)*x[0]+(0.517683480263)*x[1]**o+(-0.250680741704)*x[1]
        arg[2,3,2,0]=(-0.379114277883)*x[0]**o+(0.953285017352)*x[0]+(-0.701111847548)*x[1]**o+(-0.354199942528)*x[1]
        arg[2,3,2,1]=(0.492245629486)*x[0]**o+(-0.299380102293)*x[0]+(-0.126371826833)*x[1]**o+(0.933664060647)*x[1]
        arg[2,4,0,0]=(0.714822159533)*x[0]**o+(0.359430380478)*x[0]+(-0.691610768506)*x[1]**o+(0.414061853203)*x[1]
        arg[2,4,0,1]=(0.412799253275)*x[0]**o+(0.862905533148)*x[0]+(-0.0507440946561)*x[1]**o+(0.779012369976)*x[1]
        arg[2,4,1,0]=(-0.981179053034)*x[0]**o+(0.984777005573)*x[0]+(0.715584788452)*x[1]**o+(0.920666888724)*x[1]
        arg[2,4,1,1]=(-0.0287775685032)*x[0]**o+(0.617592345344)*x[0]+(-0.767451297794)*x[1]**o+(0.95828976091)*x[1]
        arg[2,4,2,0]=(0.905037473705)*x[0]**o+(-0.706987158021)*x[0]+(0.535436652493)*x[1]**o+(-0.891712664249)*x[1]
        arg[2,4,2,1]=(-0.818709474925)*x[0]**o+(-0.0223479273553)*x[0]+(-0.370644568884)*x[1]**o+(0.286468739105)*x[1]
        arg[3,0,0,0]=(0.310461949876)*x[0]**o+(0.770193446516)*x[0]+(-0.0858335164699)*x[1]**o+(0.12572608122)*x[1]
        arg[3,0,0,1]=(-0.781332422249)*x[0]**o+(-0.238368527872)*x[0]+(-0.449272256582)*x[1]**o+(-0.645638738212)*x[1]
        arg[3,0,1,0]=(-0.91143447678)*x[0]**o+(-0.885131081008)*x[0]+(-0.253833463581)*x[1]**o+(0.051439649459)*x[1]
        arg[3,0,1,1]=(0.447730169203)*x[0]**o+(0.980354452325)*x[0]+(-0.0484566141047)*x[1]**o+(0.443407604362)*x[1]
        arg[3,0,2,0]=(-0.223975544426)*x[0]**o+(0.0141796274312)*x[0]+(0.664080206442)*x[1]**o+(0.193318634886)*x[1]
        arg[3,0,2,1]=(-0.388901068755)*x[0]**o+(-0.632210208163)*x[0]+(0.567264328931)*x[1]**o+(-0.308710733281)*x[1]
        arg[3,1,0,0]=(0.259562910069)*x[0]**o+(-0.315409720689)*x[0]+(0.999281841483)*x[1]**o+(0.695981779357)*x[1]
        arg[3,1,0,1]=(0.627508253716)*x[0]**o+(-0.540950954632)*x[0]+(0.626209667445)*x[1]**o+(-0.920403595109)*x[1]
        arg[3,1,1,0]=(0.750166528891)*x[0]**o+(-0.145494783124)*x[0]+(0.71048043416)*x[1]**o+(-0.0920505950401)*x[1]
        arg[3,1,1,1]=(-0.755570630755)*x[0]**o+(-0.323456877629)*x[0]+(-0.0318418288689)*x[1]**o+(0.551823313307)*x[1]
        arg[3,1,2,0]=(-0.301380276426)*x[0]**o+(-0.102899117196)*x[0]+(0.816857028902)*x[1]**o+(-0.571555621895)*x[1]
        arg[3,1,2,1]=(-0.791148729922)*x[0]**o+(0.400286696027)*x[0]+(-0.17362002139)*x[1]**o+(0.431696064424)*x[1]
        arg[3,2,0,0]=(0.241944451237)*x[0]**o+(-0.629401496412)*x[0]+(-0.00333636607647)*x[1]**o+(0.167334133677)*x[1]
        arg[3,2,0,1]=(-0.300944831932)*x[0]**o+(0.358282773504)*x[0]+(0.759495661081)*x[1]**o+(-0.225250488484)*x[1]
        arg[3,2,1,0]=(0.755950588768)*x[0]**o+(0.718423180013)*x[0]+(0.705150216505)*x[1]**o+(0.269283969196)*x[1]
        arg[3,2,1,1]=(0.811064150458)*x[0]**o+(-0.67847842696)*x[0]+(-0.478091805711)*x[1]**o+(0.292774429011)*x[1]
        arg[3,2,2,0]=(-0.307877253758)*x[0]**o+(-0.563961673764)*x[0]+(0.428322493356)*x[1]**o+(0.564230126489)*x[1]
        arg[3,2,2,1]=(-0.623877056979)*x[0]**o+(-0.148337578412)*x[0]+(-0.910662605581)*x[1]**o+(-0.38603493572)*x[1]
        arg[3,3,0,0]=(-0.826474270729)*x[0]**o+(0.977614689932)*x[0]+(-0.369824837717)*x[1]**o+(-0.985478011812)*x[1]
        arg[3,3,0,1]=(-0.251252769561)*x[0]**o+(-0.296732765057)*x[0]+(-0.888899418995)*x[1]**o+(0.640313055734)*x[1]
        arg[3,3,1,0]=(-0.189106218841)*x[0]**o+(-0.0441656339777)*x[0]+(0.368716219801)*x[1]**o+(-0.924144919367)*x[1]
        arg[3,3,1,1]=(-0.525205981948)*x[0]**o+(-0.148352446593)*x[0]+(0.869212821652)*x[1]**o+(-0.575463559793)*x[1]
        arg[3,3,2,0]=(0.668772663439)*x[0]**o+(0.4330630456)*x[0]+(-0.321158907274)*x[1]**o+(0.570853139693)*x[1]
        arg[3,3,2,1]=(0.0236674031998)*x[0]**o+(-0.390106024651)*x[0]+(0.541717893494)*x[1]**o+(-0.144655705849)*x[1]
        arg[3,4,0,0]=(0.971861188602)*x[0]**o+(-0.834463063474)*x[0]+(-0.419545662218)*x[1]**o+(-0.0387341522049)*x[1]
        arg[3,4,0,1]=(0.226357589896)*x[0]**o+(0.305148895256)*x[0]+(0.00306527642431)*x[1]**o+(0.4715619746)*x[1]
        arg[3,4,1,0]=(0.529602340964)*x[0]**o+(-0.482050910795)*x[0]+(0.964983911769)*x[1]**o+(0.524018220532)*x[1]
        arg[3,4,1,1]=(0.940676158872)*x[0]**o+(0.547078925647)*x[0]+(-0.636907424453)*x[1]**o+(0.938262396281)*x[1]
        arg[3,4,2,0]=(0.282512784255)*x[0]**o+(-0.556345478931)*x[0]+(0.651866614458)*x[1]**o+(0.8895132214)*x[1]
        arg[3,4,2,1]=(0.905077434472)*x[0]**o+(0.595878819874)*x[0]+(0.416350660905)*x[1]**o+(0.100262997545)*x[1]
        ref[0,0,0,0]=(-0.93572957134)*x_ref[0]**o+(-0.878081515023)*x_ref[0]+(0.821839395118)*x_ref[1]**o+(-0.088026044081)*x_ref[1]
        ref[0,0,0,1]=(-0.00393438000839)*x_ref[0]**o+(-0.180693225971)*x_ref[0]+(0.192781914487)*x_ref[1]**o+(-0.561366977374)*x_ref[1]
        ref[0,0,1,0]=(0.116028558326)*x_ref[0]**o+(-0.47739282219)*x_ref[0]+(0.97806148353)*x_ref[1]**o+(0.660972051388)*x_ref[1]
        ref[0,0,1,1]=(0.092683552469)*x_ref[0]**o+(0.22083599745)*x_ref[0]+(-0.423073290439)*x_ref[1]**o+(-0.0789522599045)*x_ref[1]
        ref[0,0,2,0]=(0.140023403084)*x_ref[0]**o+(-0.927034796519)*x_ref[0]+(0.0987957525564)*x_ref[1]**o+(0.0703695028562)*x_ref[1]
        ref[0,0,2,1]=(-0.00485147064392)*x_ref[0]**o+(-0.649679595932)*x_ref[0]+(0.503435626816)*x_ref[1]**o+(0.393293792881)*x_ref[1]
        ref[0,1,0,0]=(-0.753383897772)*x_ref[0]**o+(-0.988499439357)*x_ref[0]+(-0.651037417175)*x_ref[1]**o+(0.45184892046)*x_ref[1]
        ref[0,1,0,1]=(0.0120892569605)*x_ref[0]**o+(0.52287896209)*x_ref[0]+(0.103256695817)*x_ref[1]**o+(-0.0244835681092)*x_ref[1]
        ref[0,1,1,0]=(-0.0949239075135)*x_ref[0]**o+(0.835062318231)*x_ref[0]+(-0.0522981103769)*x_ref[1]**o+(0.60020994838)*x_ref[1]
        ref[0,1,1,1]=(0.139896941482)*x_ref[0]**o+(0.409042945999)*x_ref[0]+(0.10019215394)*x_ref[1]**o+(-0.312224123017)*x_ref[1]
        ref[0,1,2,0]=(-0.892437964342)*x_ref[0]**o+(0.0828807279746)*x_ref[0]+(0.777811646235)*x_ref[1]**o+(-0.96509471481)*x_ref[1]
        ref[0,1,2,1]=(-0.397341226148)*x_ref[0]**o+(0.102987581336)*x_ref[0]+(-0.69165111598)*x_ref[1]**o+(0.743298211061)*x_ref[1]
        ref[0,2,0,0]=(-0.300129864031)*x_ref[0]**o+(-0.418981381232)*x_ref[0]+(0.624854922555)*x_ref[1]**o+(0.892651746561)*x_ref[1]
        ref[0,2,0,1]=(-0.511463603881)*x_ref[0]**o+(-0.808468332238)*x_ref[0]+(-0.342755248919)*x_ref[1]**o+(0.357985875808)*x_ref[1]
        ref[0,2,1,0]=(-0.269520452932)*x_ref[0]**o+(0.869553937081)*x_ref[0]+(-0.183222122614)*x_ref[1]**o+(-0.135280721586)*x_ref[1]
        ref[0,2,1,1]=(0.0396712123969)*x_ref[0]**o+(0.344314072793)*x_ref[0]+(0.644676507168)*x_ref[1]**o+(-0.0799551435053)*x_ref[1]
        ref[0,2,2,0]=(0.302628743334)*x_ref[0]**o+(-0.59557047376)*x_ref[0]+(-0.277654790212)*x_ref[1]**o+(0.0299035042863)*x_ref[1]
        ref[0,2,2,1]=(0.805296568248)*x_ref[0]**o+(0.726404711154)*x_ref[0]+(0.0728356962577)*x_ref[1]**o+(0.199997237962)*x_ref[1]
        ref[0,3,0,0]=(-0.292658800952)*x_ref[0]**o+(-0.40080106651)*x_ref[0]+(0.844109590553)*x_ref[1]**o+(-0.339364790362)*x_ref[1]
        ref[0,3,0,1]=(-0.621465058894)*x_ref[0]**o+(-0.376804350545)*x_ref[0]+(0.745038406103)*x_ref[1]**o+(-0.281482724757)*x_ref[1]
        ref[0,3,1,0]=(0.00625816347702)*x_ref[0]**o+(0.392750189511)*x_ref[0]+(-0.232258708605)*x_ref[1]**o+(-0.634552067635)*x_ref[1]
        ref[0,3,1,1]=(-0.0893109545391)*x_ref[0]**o+(-0.735678058768)*x_ref[0]+(-0.603248350212)*x_ref[1]**o+(-0.664366286468)*x_ref[1]
        ref[0,3,2,0]=(-0.0539114688775)*x_ref[0]**o+(-0.110100592496)*x_ref[0]+(-0.700496078261)*x_ref[1]**o+(0.203311460549)*x_ref[1]
        ref[0,3,2,1]=(-0.554691350426)*x_ref[0]**o+(-0.255001039659)*x_ref[0]+(0.567977217289)*x_ref[1]**o+(0.61993007447)*x_ref[1]
        ref[0,4,0,0]=(-0.12193490862)*x_ref[0]**o+(0.731139340661)*x_ref[0]+(-0.321971724057)*x_ref[1]**o+(0.741215588897)*x_ref[1]
        ref[0,4,0,1]=(-0.394743164913)*x_ref[0]**o+(-0.602249917929)*x_ref[0]+(0.27675734517)*x_ref[1]**o+(-0.968806169573)*x_ref[1]
        ref[0,4,1,0]=(-0.629331232169)*x_ref[0]**o+(0.687483125805)*x_ref[0]+(0.633698693186)*x_ref[1]**o+(0.837564332901)*x_ref[1]
        ref[0,4,1,1]=(0.0470601891829)*x_ref[0]**o+(-0.365490813497)*x_ref[0]+(-0.162532786979)*x_ref[1]**o+(0.504146804122)*x_ref[1]
        ref[0,4,2,0]=(0.193790444091)*x_ref[0]**o+(0.742132840869)*x_ref[0]+(-0.871677496589)*x_ref[1]**o+(0.492817750307)*x_ref[1]
        ref[0,4,2,1]=(0.741218732478)*x_ref[0]**o+(-0.746973604492)*x_ref[0]+(-0.548073707926)*x_ref[1]**o+(-0.42258185128)*x_ref[1]
        ref[1,0,0,0]=(0.580989627459)*x_ref[0]**o+(-0.498423525574)*x_ref[0]+(-0.349651580594)*x_ref[1]**o+(0.0193060779482)*x_ref[1]
        ref[1,0,0,1]=(0.398917011903)*x_ref[0]**o+(-0.122064497238)*x_ref[0]+(0.924654557299)*x_ref[1]**o+(0.485284066068)*x_ref[1]
        ref[1,0,1,0]=(-0.985949382826)*x_ref[0]**o+(-0.0188384874977)*x_ref[0]+(0.177995429697)*x_ref[1]**o+(0.967727833143)*x_ref[1]
        ref[1,0,1,1]=(0.920799980503)*x_ref[0]**o+(-0.749281119074)*x_ref[0]+(-0.628411322607)*x_ref[1]**o+(-0.446396361734)*x_ref[1]
        ref[1,0,2,0]=(0.752863123667)*x_ref[0]**o+(0.112427175329)*x_ref[0]+(0.975568251269)*x_ref[1]**o+(0.316938392669)*x_ref[1]
        ref[1,0,2,1]=(0.798409861933)*x_ref[0]**o+(-0.145715624761)*x_ref[0]+(-0.847813831553)*x_ref[1]**o+(0.302999756348)*x_ref[1]
        ref[1,1,0,0]=(0.897029838502)*x_ref[0]**o+(-0.540852276529)*x_ref[0]+(-0.541450963121)*x_ref[1]**o+(-0.118819043917)*x_ref[1]
        ref[1,1,0,1]=(-0.616797542123)*x_ref[0]**o+(-0.139394144428)*x_ref[0]+(-0.480348413917)*x_ref[1]**o+(0.500244829288)*x_ref[1]
        ref[1,1,1,0]=(-0.960388730531)*x_ref[0]**o+(0.87391147728)*x_ref[0]+(0.0128673975991)*x_ref[1]**o+(0.444341296582)*x_ref[1]
        ref[1,1,1,1]=(0.136810869097)*x_ref[0]**o+(-0.544756866447)*x_ref[0]+(-0.198052789907)*x_ref[1]**o+(-0.536732583929)*x_ref[1]
        ref[1,1,2,0]=(-0.808981135979)*x_ref[0]**o+(-0.122088197438)*x_ref[0]+(-0.363958821187)*x_ref[1]**o+(-0.521614066151)*x_ref[1]
        ref[1,1,2,1]=(0.299382605072)*x_ref[0]**o+(0.143397060894)*x_ref[0]+(-0.223044874711)*x_ref[1]**o+(-0.04723866032)*x_ref[1]
        ref[1,2,0,0]=(-0.568967551455)*x_ref[0]**o+(0.750777908696)*x_ref[0]+(0.203117458853)*x_ref[1]**o+(-0.0797168995572)*x_ref[1]
        ref[1,2,0,1]=(0.0439147552635)*x_ref[0]**o+(-0.0746749889653)*x_ref[0]+(0.192885940929)*x_ref[1]**o+(0.794090802345)*x_ref[1]
        ref[1,2,1,0]=(0.137177876023)*x_ref[0]**o+(0.625847668003)*x_ref[0]+(0.252761858157)*x_ref[1]**o+(0.496512978117)*x_ref[1]
        ref[1,2,1,1]=(0.475218888846)*x_ref[0]**o+(-0.286926184073)*x_ref[0]+(0.551427803487)*x_ref[1]**o+(0.782373661286)*x_ref[1]
        ref[1,2,2,0]=(0.903004502545)*x_ref[0]**o+(-0.262952251654)*x_ref[0]+(0.110351860812)*x_ref[1]**o+(-0.685355934696)*x_ref[1]
        ref[1,2,2,1]=(-0.157457317404)*x_ref[0]**o+(-0.243850623177)*x_ref[0]+(0.28853283432)*x_ref[1]**o+(0.612508908614)*x_ref[1]
        ref[1,3,0,0]=(-0.741890002973)*x_ref[0]**o+(-0.118262424825)*x_ref[0]+(0.567100734646)*x_ref[1]**o+(0.301557125811)*x_ref[1]
        ref[1,3,0,1]=(-0.00745396583928)*x_ref[0]**o+(-0.00558491344276)*x_ref[0]+(-0.967132370197)*x_ref[1]**o+(-0.61737594402)*x_ref[1]
        ref[1,3,1,0]=(0.0767961523966)*x_ref[0]**o+(0.341952000996)*x_ref[0]+(0.762617832019)*x_ref[1]**o+(-0.38337284713)*x_ref[1]
        ref[1,3,1,1]=(0.32265018902)*x_ref[0]**o+(-0.352521736206)*x_ref[0]+(-0.0119143186529)*x_ref[1]**o+(-0.64727561829)*x_ref[1]
        ref[1,3,2,0]=(0.627690584587)*x_ref[0]**o+(0.0993819822193)*x_ref[0]+(-0.189363448841)*x_ref[1]**o+(0.49004001194)*x_ref[1]
        ref[1,3,2,1]=(-0.657836476156)*x_ref[0]**o+(0.708970470136)*x_ref[0]+(0.252190020576)*x_ref[1]**o+(-0.481214196298)*x_ref[1]
        ref[1,4,0,0]=(0.0837563164179)*x_ref[0]**o+(0.741536525079)*x_ref[0]+(0.645543094286)*x_ref[1]**o+(0.62358090815)*x_ref[1]
        ref[1,4,0,1]=(0.262043398371)*x_ref[0]**o+(-0.47794750936)*x_ref[0]+(0.93086244965)*x_ref[1]**o+(-0.670399062677)*x_ref[1]
        ref[1,4,1,0]=(-0.76634902923)*x_ref[0]**o+(-0.809846143928)*x_ref[0]+(0.639099549305)*x_ref[1]**o+(0.793524639236)*x_ref[1]
        ref[1,4,1,1]=(0.814737503957)*x_ref[0]**o+(0.775593374042)*x_ref[0]+(0.240976904111)*x_ref[1]**o+(0.2352421356)*x_ref[1]
        ref[1,4,2,0]=(0.519210584047)*x_ref[0]**o+(-0.702579589691)*x_ref[0]+(-0.450156317853)*x_ref[1]**o+(0.588864692581)*x_ref[1]
        ref[1,4,2,1]=(-0.927751546037)*x_ref[0]**o+(-0.239737576019)*x_ref[0]+(-0.527658440447)*x_ref[1]**o+(0.79623440559)*x_ref[1]
        ref[2,0,0,0]=(0.719144336419)*x_ref[0]**o+(0.916780231277)*x_ref[0]+(0.279073450277)*x_ref[1]**o+(-0.0130671321167)*x_ref[1]
        ref[2,0,0,1]=(-0.556604998962)*x_ref[0]**o+(-0.661580370137)*x_ref[0]+(0.980704664791)*x_ref[1]**o+(0.935843251131)*x_ref[1]
        ref[2,0,1,0]=(0.85501679753)*x_ref[0]**o+(-0.908342467889)*x_ref[0]+(0.969030151294)*x_ref[1]**o+(0.534192961922)*x_ref[1]
        ref[2,0,1,1]=(0.527672385529)*x_ref[0]**o+(0.997699955667)*x_ref[0]+(-0.388249629327)*x_ref[1]**o+(-0.714956937512)*x_ref[1]
        ref[2,0,2,0]=(-0.629149810667)*x_ref[0]**o+(-0.0515018376692)*x_ref[0]+(-0.80355991702)*x_ref[1]**o+(0.235999988886)*x_ref[1]
        ref[2,0,2,1]=(0.960715934156)*x_ref[0]**o+(-0.782787305497)*x_ref[0]+(-0.309348231959)*x_ref[1]**o+(0.854911294717)*x_ref[1]
        ref[2,1,0,0]=(-0.560749125074)*x_ref[0]**o+(-0.120515615757)*x_ref[0]+(0.055090634431)*x_ref[1]**o+(-0.539205949358)*x_ref[1]
        ref[2,1,0,1]=(-0.762113417295)*x_ref[0]**o+(-0.207196980376)*x_ref[0]+(0.904541245494)*x_ref[1]**o+(0.320359459939)*x_ref[1]
        ref[2,1,1,0]=(-0.272627586486)*x_ref[0]**o+(-0.885987982522)*x_ref[0]+(0.641737808216)*x_ref[1]**o+(0.358390367373)*x_ref[1]
        ref[2,1,1,1]=(-0.847502563761)*x_ref[0]**o+(-0.430298277492)*x_ref[0]+(-0.869043300185)*x_ref[1]**o+(0.584700145915)*x_ref[1]
        ref[2,1,2,0]=(0.497688368946)*x_ref[0]**o+(-0.14413770704)*x_ref[0]+(-0.82002941888)*x_ref[1]**o+(0.573569419279)*x_ref[1]
        ref[2,1,2,1]=(0.959549623603)*x_ref[0]**o+(0.881196513588)*x_ref[0]+(-0.202951096266)*x_ref[1]**o+(-9.87983182765e-05)*x_ref[1]
        ref[2,2,0,0]=(0.0811393810501)*x_ref[0]**o+(-0.853555165896)*x_ref[0]+(0.0259257246335)*x_ref[1]**o+(0.638575306959)*x_ref[1]
        ref[2,2,0,1]=(-0.525990031379)*x_ref[0]**o+(0.11235160765)*x_ref[0]+(-0.502853879455)*x_ref[1]**o+(0.414066683507)*x_ref[1]
        ref[2,2,1,0]=(-0.533919716834)*x_ref[0]**o+(-0.773411100656)*x_ref[0]+(0.311424028491)*x_ref[1]**o+(0.332733301377)*x_ref[1]
        ref[2,2,1,1]=(0.280548593462)*x_ref[0]**o+(-0.0747640762104)*x_ref[0]+(-0.7110232877)*x_ref[1]**o+(-0.178672817441)*x_ref[1]
        ref[2,2,2,0]=(0.445025926846)*x_ref[0]**o+(-0.669782903005)*x_ref[0]+(0.236610411793)*x_ref[1]**o+(0.437836936476)*x_ref[1]
        ref[2,2,2,1]=(-0.503761305098)*x_ref[0]**o+(-0.889922659865)*x_ref[0]+(-0.0767256016402)*x_ref[1]**o+(-0.614874217835)*x_ref[1]
        ref[2,3,0,0]=(-0.041960855672)*x_ref[0]**o+(-0.593486242764)*x_ref[0]+(0.344751892525)*x_ref[1]**o+(-0.131268873786)*x_ref[1]
        ref[2,3,0,1]=(-0.192704843083)*x_ref[0]**o+(0.325356381696)*x_ref[0]+(-0.235690651937)*x_ref[1]**o+(0.59113535617)*x_ref[1]
        ref[2,3,1,0]=(-0.0794745628024)*x_ref[0]**o+(0.7996737586)*x_ref[0]+(-0.324970505375)*x_ref[1]**o+(-0.14920403468)*x_ref[1]
        ref[2,3,1,1]=(0.31997432673)*x_ref[0]**o+(-0.571739807799)*x_ref[0]+(0.517683480263)*x_ref[1]**o+(-0.250680741704)*x_ref[1]
        ref[2,3,2,0]=(-0.379114277883)*x_ref[0]**o+(0.953285017352)*x_ref[0]+(-0.701111847548)*x_ref[1]**o+(-0.354199942528)*x_ref[1]
        ref[2,3,2,1]=(0.492245629486)*x_ref[0]**o+(-0.299380102293)*x_ref[0]+(-0.126371826833)*x_ref[1]**o+(0.933664060647)*x_ref[1]
        ref[2,4,0,0]=(0.714822159533)*x_ref[0]**o+(0.359430380478)*x_ref[0]+(-0.691610768506)*x_ref[1]**o+(0.414061853203)*x_ref[1]
        ref[2,4,0,1]=(0.412799253275)*x_ref[0]**o+(0.862905533148)*x_ref[0]+(-0.0507440946561)*x_ref[1]**o+(0.779012369976)*x_ref[1]
        ref[2,4,1,0]=(-0.981179053034)*x_ref[0]**o+(0.984777005573)*x_ref[0]+(0.715584788452)*x_ref[1]**o+(0.920666888724)*x_ref[1]
        ref[2,4,1,1]=(-0.0287775685032)*x_ref[0]**o+(0.617592345344)*x_ref[0]+(-0.767451297794)*x_ref[1]**o+(0.95828976091)*x_ref[1]
        ref[2,4,2,0]=(0.905037473705)*x_ref[0]**o+(-0.706987158021)*x_ref[0]+(0.535436652493)*x_ref[1]**o+(-0.891712664249)*x_ref[1]
        ref[2,4,2,1]=(-0.818709474925)*x_ref[0]**o+(-0.0223479273553)*x_ref[0]+(-0.370644568884)*x_ref[1]**o+(0.286468739105)*x_ref[1]
        ref[3,0,0,0]=(0.310461949876)*x_ref[0]**o+(0.770193446516)*x_ref[0]+(-0.0858335164699)*x_ref[1]**o+(0.12572608122)*x_ref[1]
        ref[3,0,0,1]=(-0.781332422249)*x_ref[0]**o+(-0.238368527872)*x_ref[0]+(-0.449272256582)*x_ref[1]**o+(-0.645638738212)*x_ref[1]
        ref[3,0,1,0]=(-0.91143447678)*x_ref[0]**o+(-0.885131081008)*x_ref[0]+(-0.253833463581)*x_ref[1]**o+(0.051439649459)*x_ref[1]
        ref[3,0,1,1]=(0.447730169203)*x_ref[0]**o+(0.980354452325)*x_ref[0]+(-0.0484566141047)*x_ref[1]**o+(0.443407604362)*x_ref[1]
        ref[3,0,2,0]=(-0.223975544426)*x_ref[0]**o+(0.0141796274312)*x_ref[0]+(0.664080206442)*x_ref[1]**o+(0.193318634886)*x_ref[1]
        ref[3,0,2,1]=(-0.388901068755)*x_ref[0]**o+(-0.632210208163)*x_ref[0]+(0.567264328931)*x_ref[1]**o+(-0.308710733281)*x_ref[1]
        ref[3,1,0,0]=(0.259562910069)*x_ref[0]**o+(-0.315409720689)*x_ref[0]+(0.999281841483)*x_ref[1]**o+(0.695981779357)*x_ref[1]
        ref[3,1,0,1]=(0.627508253716)*x_ref[0]**o+(-0.540950954632)*x_ref[0]+(0.626209667445)*x_ref[1]**o+(-0.920403595109)*x_ref[1]
        ref[3,1,1,0]=(0.750166528891)*x_ref[0]**o+(-0.145494783124)*x_ref[0]+(0.71048043416)*x_ref[1]**o+(-0.0920505950401)*x_ref[1]
        ref[3,1,1,1]=(-0.755570630755)*x_ref[0]**o+(-0.323456877629)*x_ref[0]+(-0.0318418288689)*x_ref[1]**o+(0.551823313307)*x_ref[1]
        ref[3,1,2,0]=(-0.301380276426)*x_ref[0]**o+(-0.102899117196)*x_ref[0]+(0.816857028902)*x_ref[1]**o+(-0.571555621895)*x_ref[1]
        ref[3,1,2,1]=(-0.791148729922)*x_ref[0]**o+(0.400286696027)*x_ref[0]+(-0.17362002139)*x_ref[1]**o+(0.431696064424)*x_ref[1]
        ref[3,2,0,0]=(0.241944451237)*x_ref[0]**o+(-0.629401496412)*x_ref[0]+(-0.00333636607647)*x_ref[1]**o+(0.167334133677)*x_ref[1]
        ref[3,2,0,1]=(-0.300944831932)*x_ref[0]**o+(0.358282773504)*x_ref[0]+(0.759495661081)*x_ref[1]**o+(-0.225250488484)*x_ref[1]
        ref[3,2,1,0]=(0.755950588768)*x_ref[0]**o+(0.718423180013)*x_ref[0]+(0.705150216505)*x_ref[1]**o+(0.269283969196)*x_ref[1]
        ref[3,2,1,1]=(0.811064150458)*x_ref[0]**o+(-0.67847842696)*x_ref[0]+(-0.478091805711)*x_ref[1]**o+(0.292774429011)*x_ref[1]
        ref[3,2,2,0]=(-0.307877253758)*x_ref[0]**o+(-0.563961673764)*x_ref[0]+(0.428322493356)*x_ref[1]**o+(0.564230126489)*x_ref[1]
        ref[3,2,2,1]=(-0.623877056979)*x_ref[0]**o+(-0.148337578412)*x_ref[0]+(-0.910662605581)*x_ref[1]**o+(-0.38603493572)*x_ref[1]
        ref[3,3,0,0]=(-0.826474270729)*x_ref[0]**o+(0.977614689932)*x_ref[0]+(-0.369824837717)*x_ref[1]**o+(-0.985478011812)*x_ref[1]
        ref[3,3,0,1]=(-0.251252769561)*x_ref[0]**o+(-0.296732765057)*x_ref[0]+(-0.888899418995)*x_ref[1]**o+(0.640313055734)*x_ref[1]
        ref[3,3,1,0]=(-0.189106218841)*x_ref[0]**o+(-0.0441656339777)*x_ref[0]+(0.368716219801)*x_ref[1]**o+(-0.924144919367)*x_ref[1]
        ref[3,3,1,1]=(-0.525205981948)*x_ref[0]**o+(-0.148352446593)*x_ref[0]+(0.869212821652)*x_ref[1]**o+(-0.575463559793)*x_ref[1]
        ref[3,3,2,0]=(0.668772663439)*x_ref[0]**o+(0.4330630456)*x_ref[0]+(-0.321158907274)*x_ref[1]**o+(0.570853139693)*x_ref[1]
        ref[3,3,2,1]=(0.0236674031998)*x_ref[0]**o+(-0.390106024651)*x_ref[0]+(0.541717893494)*x_ref[1]**o+(-0.144655705849)*x_ref[1]
        ref[3,4,0,0]=(0.971861188602)*x_ref[0]**o+(-0.834463063474)*x_ref[0]+(-0.419545662218)*x_ref[1]**o+(-0.0387341522049)*x_ref[1]
        ref[3,4,0,1]=(0.226357589896)*x_ref[0]**o+(0.305148895256)*x_ref[0]+(0.00306527642431)*x_ref[1]**o+(0.4715619746)*x_ref[1]
        ref[3,4,1,0]=(0.529602340964)*x_ref[0]**o+(-0.482050910795)*x_ref[0]+(0.964983911769)*x_ref[1]**o+(0.524018220532)*x_ref[1]
        ref[3,4,1,1]=(0.940676158872)*x_ref[0]**o+(0.547078925647)*x_ref[0]+(-0.636907424453)*x_ref[1]**o+(0.938262396281)*x_ref[1]
        ref[3,4,2,0]=(0.282512784255)*x_ref[0]**o+(-0.556345478931)*x_ref[0]+(0.651866614458)*x_ref[1]**o+(0.8895132214)*x_ref[1]
        ref[3,4,2,1]=(0.905077434472)*x_ref[0]**o+(0.595878819874)*x_ref[0]+(0.416350660905)*x_ref[1]**o+(0.100262997545)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.836302395465)*x[0]**o+(-0.842568675604)*x[0]+(-0.780133441207)*x[1]**o+(0.54483947113)*x[1]+(-0.130253019696)*x[2]**o+(0.358303389828)*x[2]
        arg[0,0,0,1]=(-0.722847500652)*x[0]**o+(0.42887368878)*x[0]+(0.979187056933)*x[1]**o+(0.624875714982)*x[1]+(-0.672529765247)*x[2]**o+(0.822363583702)*x[2]
        arg[0,0,1,0]=(-0.28989650691)*x[0]**o+(-0.847647954773)*x[0]+(-0.11119862113)*x[1]**o+(0.0809009676818)*x[1]+(-0.672626533587)*x[2]**o+(0.580561232267)*x[2]
        arg[0,0,1,1]=(-0.64783363248)*x[0]**o+(0.212828552873)*x[0]+(-0.845297180936)*x[1]**o+(-0.714760411186)*x[1]+(-0.385173605122)*x[2]**o+(-0.760750317556)*x[2]
        arg[0,0,2,0]=(-0.328757006046)*x[0]**o+(-0.0935907722247)*x[0]+(-0.895866135795)*x[1]**o+(-0.701241976745)*x[1]+(0.893684139941)*x[2]**o+(-0.265339318222)*x[2]
        arg[0,0,2,1]=(0.134066662346)*x[0]**o+(0.0383231508856)*x[0]+(-0.347744261995)*x[1]**o+(-0.86897907336)*x[1]+(-0.287792154521)*x[2]**o+(0.859351846891)*x[2]
        arg[0,1,0,0]=(-0.635128124936)*x[0]**o+(-0.659462189733)*x[0]+(-0.0964365747936)*x[1]**o+(-0.385702168747)*x[1]+(0.179539914057)*x[2]**o+(-0.0176420847859)*x[2]
        arg[0,1,0,1]=(-0.580750807831)*x[0]**o+(0.756493147583)*x[0]+(-0.436842251289)*x[1]**o+(-0.705624646095)*x[1]+(0.983064096262)*x[2]**o+(-0.0197153825423)*x[2]
        arg[0,1,1,0]=(-0.567748582365)*x[0]**o+(0.579324420356)*x[0]+(0.0434168493419)*x[1]**o+(-0.839692920334)*x[1]+(-0.119239847745)*x[2]**o+(0.527047942834)*x[2]
        arg[0,1,1,1]=(0.52870403655)*x[0]**o+(0.700090892471)*x[0]+(0.730224313921)*x[1]**o+(-0.815451751207)*x[1]+(-0.943945382413)*x[2]**o+(-0.571714141916)*x[2]
        arg[0,1,2,0]=(-0.803841094887)*x[0]**o+(0.0595126006237)*x[0]+(-0.612341797645)*x[1]**o+(0.944552805211)*x[1]+(-0.00253009731786)*x[2]**o+(0.0174867439912)*x[2]
        arg[0,1,2,1]=(-0.145623212549)*x[0]**o+(0.546039957008)*x[0]+(-0.221425237579)*x[1]**o+(0.229367671159)*x[1]+(-0.784903869759)*x[2]**o+(0.898045208871)*x[2]
        arg[0,2,0,0]=(0.417924494005)*x[0]**o+(-0.192820051987)*x[0]+(0.878916271492)*x[1]**o+(-0.35056320984)*x[1]+(0.982897575604)*x[2]**o+(-0.225793730616)*x[2]
        arg[0,2,0,1]=(-0.493279498849)*x[0]**o+(-0.154177411087)*x[0]+(-0.923592860492)*x[1]**o+(0.872010949252)*x[1]+(-0.936252147064)*x[2]**o+(0.947136819308)*x[2]
        arg[0,2,1,0]=(0.109357630607)*x[0]**o+(-0.518825477709)*x[0]+(-0.375510667498)*x[1]**o+(0.476537999644)*x[1]+(0.676017421029)*x[2]**o+(-0.633301158984)*x[2]
        arg[0,2,1,1]=(-0.0662265225539)*x[0]**o+(-0.276634748263)*x[0]+(-0.0655431981102)*x[1]**o+(0.0578482920991)*x[1]+(-0.264370611185)*x[2]**o+(-0.674612985594)*x[2]
        arg[0,2,2,0]=(-0.924186592743)*x[0]**o+(0.197603043149)*x[0]+(-0.676882805046)*x[1]**o+(0.564142845865)*x[1]+(0.590369345306)*x[2]**o+(-0.872019490978)*x[2]
        arg[0,2,2,1]=(0.480683776968)*x[0]**o+(-0.412055284173)*x[0]+(0.829996313473)*x[1]**o+(-0.235665473048)*x[1]+(-0.429191305284)*x[2]**o+(-0.0867804318457)*x[2]
        arg[0,3,0,0]=(-0.0903660006043)*x[0]**o+(-0.507571901766)*x[0]+(-0.0305305861809)*x[1]**o+(-0.915027969989)*x[1]+(0.955644774347)*x[2]**o+(0.739244618775)*x[2]
        arg[0,3,0,1]=(0.600619122998)*x[0]**o+(0.0128699484527)*x[0]+(-0.360431921281)*x[1]**o+(0.304773619811)*x[1]+(-0.302423314054)*x[2]**o+(0.680149401572)*x[2]
        arg[0,3,1,0]=(-0.638540513323)*x[0]**o+(0.766655754067)*x[0]+(0.276879854223)*x[1]**o+(-0.628688275247)*x[1]+(0.724584107114)*x[2]**o+(0.682193551967)*x[2]
        arg[0,3,1,1]=(-0.516656801349)*x[0]**o+(-0.513901319383)*x[0]+(-0.227957010777)*x[1]**o+(-0.164363207703)*x[1]+(-0.919936042047)*x[2]**o+(0.682400184997)*x[2]
        arg[0,3,2,0]=(-0.819629653095)*x[0]**o+(0.183937049274)*x[0]+(0.0411075800579)*x[1]**o+(0.0813088270699)*x[1]+(0.762416334575)*x[2]**o+(0.236553726705)*x[2]
        arg[0,3,2,1]=(0.513141183759)*x[0]**o+(-0.452475512734)*x[0]+(0.0461424221984)*x[1]**o+(-0.673452844636)*x[1]+(-0.194493328878)*x[2]**o+(0.812459956422)*x[2]
        arg[0,4,0,0]=(0.672167415195)*x[0]**o+(-0.641004730405)*x[0]+(-0.383031420851)*x[1]**o+(0.392366431186)*x[1]+(0.66479124224)*x[2]**o+(0.845503225104)*x[2]
        arg[0,4,0,1]=(0.908241966695)*x[0]**o+(-0.816989438726)*x[0]+(-0.144902267583)*x[1]**o+(-0.654901540598)*x[1]+(-0.9840470269)*x[2]**o+(0.510970695154)*x[2]
        arg[0,4,1,0]=(0.788858704366)*x[0]**o+(-0.50266306504)*x[0]+(-0.681248796723)*x[1]**o+(0.0317663112509)*x[1]+(0.662428035619)*x[2]**o+(0.377116420912)*x[2]
        arg[0,4,1,1]=(-0.0734522829037)*x[0]**o+(0.0822741484655)*x[0]+(-0.509279578815)*x[1]**o+(-0.95372420098)*x[1]+(0.682868362508)*x[2]**o+(-0.947156883638)*x[2]
        arg[0,4,2,0]=(-0.843236607337)*x[0]**o+(-0.925326238513)*x[0]+(-0.153688519983)*x[1]**o+(-0.68456413487)*x[1]+(0.920442617368)*x[2]**o+(0.852945557893)*x[2]
        arg[0,4,2,1]=(-0.283772512528)*x[0]**o+(0.202730738521)*x[0]+(-0.287375172356)*x[1]**o+(-0.176699132816)*x[1]+(0.446391113264)*x[2]**o+(-0.783266847348)*x[2]
        arg[1,0,0,0]=(0.81284364562)*x[0]**o+(-0.454432332505)*x[0]+(-0.684692596262)*x[1]**o+(-0.345923888237)*x[1]+(0.1201728631)*x[2]**o+(0.808246135028)*x[2]
        arg[1,0,0,1]=(0.941104328706)*x[0]**o+(-0.802008654964)*x[0]+(0.902047483677)*x[1]**o+(0.777049224612)*x[1]+(0.120184332854)*x[2]**o+(0.983216899102)*x[2]
        arg[1,0,1,0]=(-0.420996925767)*x[0]**o+(0.0941750615502)*x[0]+(-0.935342933578)*x[1]**o+(0.925891822058)*x[1]+(-0.153652985582)*x[2]**o+(-0.336548986153)*x[2]
        arg[1,0,1,1]=(-0.692886221795)*x[0]**o+(-0.920419410139)*x[0]+(-0.586575583263)*x[1]**o+(-0.269347362009)*x[1]+(-0.872169736045)*x[2]**o+(-0.724727186336)*x[2]
        arg[1,0,2,0]=(-0.0171222282058)*x[0]**o+(0.338180602976)*x[0]+(-0.0345958505395)*x[1]**o+(0.631728421505)*x[1]+(0.90596443453)*x[2]**o+(-0.144785193015)*x[2]
        arg[1,0,2,1]=(-0.91636583885)*x[0]**o+(-0.807496789574)*x[0]+(0.63472495957)*x[1]**o+(0.966464343993)*x[1]+(-0.471311384621)*x[2]**o+(-0.449420138238)*x[2]
        arg[1,1,0,0]=(-0.239930503474)*x[0]**o+(-0.476439715994)*x[0]+(0.421600511888)*x[1]**o+(0.477823144792)*x[1]+(0.251409914807)*x[2]**o+(0.636818001419)*x[2]
        arg[1,1,0,1]=(-0.427510410263)*x[0]**o+(0.494810857516)*x[0]+(-0.129415181352)*x[1]**o+(0.754147770071)*x[1]+(0.0218776365985)*x[2]**o+(0.706183216616)*x[2]
        arg[1,1,1,0]=(-0.469150375165)*x[0]**o+(-0.0651101638073)*x[0]+(-0.697853003861)*x[1]**o+(0.0389330102026)*x[1]+(-0.256557942045)*x[2]**o+(0.409565340425)*x[2]
        arg[1,1,1,1]=(0.14957005608)*x[0]**o+(-0.209482968126)*x[0]+(0.368899650309)*x[1]**o+(-0.0899321000137)*x[1]+(0.282389903357)*x[2]**o+(-0.166633883916)*x[2]
        arg[1,1,2,0]=(0.0374938231714)*x[0]**o+(-0.687972899893)*x[0]+(-0.0686478107779)*x[1]**o+(0.318977937888)*x[1]+(-0.601269126788)*x[2]**o+(0.453188855786)*x[2]
        arg[1,1,2,1]=(-0.853894223222)*x[0]**o+(0.441830151478)*x[0]+(0.973114898368)*x[1]**o+(-0.264454889071)*x[1]+(-0.828332687531)*x[2]**o+(0.302493748565)*x[2]
        arg[1,2,0,0]=(0.210365824111)*x[0]**o+(0.4674117857)*x[0]+(-0.129546969044)*x[1]**o+(0.566357125352)*x[1]+(-0.721733448514)*x[2]**o+(0.154253207106)*x[2]
        arg[1,2,0,1]=(0.805325508181)*x[0]**o+(-0.929113563452)*x[0]+(-0.487301761321)*x[1]**o+(-0.822405526519)*x[1]+(0.254754490326)*x[2]**o+(-0.196480710709)*x[2]
        arg[1,2,1,0]=(-0.520777275475)*x[0]**o+(-0.912569745644)*x[0]+(-0.784504791087)*x[1]**o+(-0.903473217653)*x[1]+(0.583297107374)*x[2]**o+(0.664811898195)*x[2]
        arg[1,2,1,1]=(0.129152152958)*x[0]**o+(0.497832200867)*x[0]+(0.121217248234)*x[1]**o+(-0.183368066714)*x[1]+(-0.776145485536)*x[2]**o+(-0.223165724715)*x[2]
        arg[1,2,2,0]=(-0.698981324488)*x[0]**o+(0.00951277777924)*x[0]+(0.997410976507)*x[1]**o+(0.534670303453)*x[1]+(0.251309223218)*x[2]**o+(-0.921148646979)*x[2]
        arg[1,2,2,1]=(-0.731509934487)*x[0]**o+(0.87731926834)*x[0]+(0.0865507713913)*x[1]**o+(0.708835987019)*x[1]+(0.294182269171)*x[2]**o+(-0.316660750729)*x[2]
        arg[1,3,0,0]=(-0.282903962918)*x[0]**o+(-0.793457978844)*x[0]+(0.287425807914)*x[1]**o+(-0.0252722364835)*x[1]+(-0.956202022515)*x[2]**o+(0.650271579135)*x[2]
        arg[1,3,0,1]=(-0.946723111202)*x[0]**o+(0.669431653693)*x[0]+(-0.791125976188)*x[1]**o+(-0.466484914794)*x[1]+(-0.217225801872)*x[2]**o+(0.224719473996)*x[2]
        arg[1,3,1,0]=(0.516479301043)*x[0]**o+(-0.42420332366)*x[0]+(-0.0649227247176)*x[1]**o+(0.208849752675)*x[1]+(-0.939001099693)*x[2]**o+(-0.0537827140932)*x[2]
        arg[1,3,1,1]=(-0.292077069459)*x[0]**o+(-0.444848119744)*x[0]+(0.268592719199)*x[1]**o+(-0.349401060323)*x[1]+(-0.540803784489)*x[2]**o+(-0.487170123526)*x[2]
        arg[1,3,2,0]=(-0.626222042661)*x[0]**o+(0.245937409624)*x[0]+(-0.24692162006)*x[1]**o+(0.108230029208)*x[1]+(-0.760463241928)*x[2]**o+(-0.587078263201)*x[2]
        arg[1,3,2,1]=(-0.923891147756)*x[0]**o+(0.0961285842162)*x[0]+(0.0146853466585)*x[1]**o+(0.00951339122359)*x[1]+(-0.261383704723)*x[2]**o+(-0.0262515691324)*x[2]
        arg[1,4,0,0]=(-0.0370003524292)*x[0]**o+(-0.0463848489957)*x[0]+(0.940061458694)*x[1]**o+(0.592787058925)*x[1]+(-0.217284379638)*x[2]**o+(-0.11746834109)*x[2]
        arg[1,4,0,1]=(0.830404283093)*x[0]**o+(0.268135490403)*x[0]+(-0.894500398195)*x[1]**o+(0.751042648917)*x[1]+(0.687583292754)*x[2]**o+(-0.884596509742)*x[2]
        arg[1,4,1,0]=(0.414764592469)*x[0]**o+(-0.529613347919)*x[0]+(-0.727435115594)*x[1]**o+(-0.546579777725)*x[1]+(0.378011893971)*x[2]**o+(-0.578081183238)*x[2]
        arg[1,4,1,1]=(0.406513376747)*x[0]**o+(-0.350132869579)*x[0]+(-0.557954631299)*x[1]**o+(0.868176175754)*x[1]+(0.793021832632)*x[2]**o+(0.280529771851)*x[2]
        arg[1,4,2,0]=(0.56741953575)*x[0]**o+(0.0115947684302)*x[0]+(0.872812725363)*x[1]**o+(-0.215183541745)*x[1]+(0.64853692203)*x[2]**o+(-0.139561471922)*x[2]
        arg[1,4,2,1]=(0.650916904666)*x[0]**o+(0.878765271839)*x[0]+(0.0141264348052)*x[1]**o+(-0.56314426774)*x[1]+(-0.267854377704)*x[2]**o+(0.852271380667)*x[2]
        arg[2,0,0,0]=(0.948359141577)*x[0]**o+(-0.0881747280974)*x[0]+(0.804871563368)*x[1]**o+(-0.0298585763446)*x[1]+(0.506923333569)*x[2]**o+(-0.920947093648)*x[2]
        arg[2,0,0,1]=(-0.0144221258584)*x[0]**o+(-0.679723072824)*x[0]+(0.00829336956687)*x[1]**o+(0.675680809303)*x[1]+(-0.803195231536)*x[2]**o+(0.70240358571)*x[2]
        arg[2,0,1,0]=(-0.159174987706)*x[0]**o+(0.922309271483)*x[0]+(0.344354343035)*x[1]**o+(-0.00954038311323)*x[1]+(-0.50225452804)*x[2]**o+(0.355899734308)*x[2]
        arg[2,0,1,1]=(0.0857611780404)*x[0]**o+(0.876479431983)*x[0]+(0.105049158502)*x[1]**o+(-0.530308461426)*x[1]+(-0.553136649343)*x[2]**o+(0.385458516481)*x[2]
        arg[2,0,2,0]=(-0.260086070842)*x[0]**o+(0.417766241499)*x[0]+(-0.292370383673)*x[1]**o+(-0.321315349553)*x[1]+(-0.437301444082)*x[2]**o+(0.730845669715)*x[2]
        arg[2,0,2,1]=(0.132547674276)*x[0]**o+(0.783602886333)*x[0]+(0.822361450413)*x[1]**o+(-0.810402783)*x[1]+(0.759701191023)*x[2]**o+(0.644375788608)*x[2]
        arg[2,1,0,0]=(-0.841458111227)*x[0]**o+(0.779917641736)*x[0]+(-0.0745587217256)*x[1]**o+(0.938545280143)*x[1]+(0.454600163373)*x[2]**o+(-0.490348377697)*x[2]
        arg[2,1,0,1]=(-0.876477136174)*x[0]**o+(-0.248200369939)*x[0]+(-0.973654243816)*x[1]**o+(0.482402123794)*x[1]+(-0.776970490827)*x[2]**o+(-0.571349467703)*x[2]
        arg[2,1,1,0]=(-0.433528193102)*x[0]**o+(-0.57712915767)*x[0]+(-0.328671198058)*x[1]**o+(-0.926873019075)*x[1]+(0.314232739356)*x[2]**o+(-0.199625466412)*x[2]
        arg[2,1,1,1]=(0.892707770392)*x[0]**o+(-0.680081898234)*x[0]+(-0.754181058483)*x[1]**o+(-0.253728945575)*x[1]+(-0.906373406684)*x[2]**o+(-0.0266435080833)*x[2]
        arg[2,1,2,0]=(-0.121605378472)*x[0]**o+(-0.503955015634)*x[0]+(-0.0960767196114)*x[1]**o+(-0.937914890361)*x[1]+(0.212042931853)*x[2]**o+(0.930703155408)*x[2]
        arg[2,1,2,1]=(0.0564159441047)*x[0]**o+(0.842332721105)*x[0]+(0.954741657608)*x[1]**o+(0.73503176)*x[1]+(-0.390679938582)*x[2]**o+(-0.440014747741)*x[2]
        arg[2,2,0,0]=(0.0100637970821)*x[0]**o+(-0.0283353154145)*x[0]+(0.865360758262)*x[1]**o+(0.780975027988)*x[1]+(0.16115842919)*x[2]**o+(0.247580155656)*x[2]
        arg[2,2,0,1]=(0.61102059426)*x[0]**o+(-0.198005744101)*x[0]+(0.503568797262)*x[1]**o+(0.0288809726497)*x[1]+(-0.765127098183)*x[2]**o+(0.409803857723)*x[2]
        arg[2,2,1,0]=(-0.712272018295)*x[0]**o+(-0.62562142063)*x[0]+(-0.579501949719)*x[1]**o+(0.628539275236)*x[1]+(0.0200945124964)*x[2]**o+(0.175966024769)*x[2]
        arg[2,2,1,1]=(-0.798244445519)*x[0]**o+(-0.376440962535)*x[0]+(-0.134467401497)*x[1]**o+(0.877538294477)*x[1]+(-0.484870741192)*x[2]**o+(-0.224064469935)*x[2]
        arg[2,2,2,0]=(0.200196926368)*x[0]**o+(-0.528000867086)*x[0]+(-0.708942342624)*x[1]**o+(-0.591994969141)*x[1]+(-0.990641592656)*x[2]**o+(-0.875897347303)*x[2]
        arg[2,2,2,1]=(0.949113190797)*x[0]**o+(0.867855496786)*x[0]+(-0.0314953438065)*x[1]**o+(-0.371928428005)*x[1]+(-0.526403191063)*x[2]**o+(0.653049122326)*x[2]
        arg[2,3,0,0]=(-0.353418327127)*x[0]**o+(-0.418588306429)*x[0]+(0.400872043269)*x[1]**o+(0.141907924042)*x[1]+(-0.999869579462)*x[2]**o+(-0.410423954563)*x[2]
        arg[2,3,0,1]=(-0.240390877307)*x[0]**o+(0.0901879916115)*x[0]+(-0.0734683209657)*x[1]**o+(0.274836005886)*x[1]+(-0.336375498181)*x[2]**o+(-0.754776455221)*x[2]
        arg[2,3,1,0]=(0.259928460822)*x[0]**o+(-0.227630454708)*x[0]+(0.725968066278)*x[1]**o+(-0.545440695763)*x[1]+(-0.0354926607815)*x[2]**o+(0.365010532181)*x[2]
        arg[2,3,1,1]=(-0.969799373398)*x[0]**o+(0.514498233214)*x[0]+(0.907450925318)*x[1]**o+(-0.234993706406)*x[1]+(0.206323910806)*x[2]**o+(0.209364250369)*x[2]
        arg[2,3,2,0]=(0.712905565941)*x[0]**o+(-0.353322758263)*x[0]+(0.816861181494)*x[1]**o+(-0.362388243619)*x[1]+(0.302543517607)*x[2]**o+(-0.831280418852)*x[2]
        arg[2,3,2,1]=(-0.386373711182)*x[0]**o+(0.946914567792)*x[0]+(-0.52795085887)*x[1]**o+(0.86635245964)*x[1]+(0.654261054178)*x[2]**o+(0.452334882498)*x[2]
        arg[2,4,0,0]=(0.374747033547)*x[0]**o+(-0.3471578669)*x[0]+(0.322217002639)*x[1]**o+(0.92988068982)*x[1]+(-0.642423472369)*x[2]**o+(-0.759623225653)*x[2]
        arg[2,4,0,1]=(-0.322255157841)*x[0]**o+(-0.253806844569)*x[0]+(-0.0688746199637)*x[1]**o+(0.232840317719)*x[1]+(-0.815623721026)*x[2]**o+(0.43055485562)*x[2]
        arg[2,4,1,0]=(-0.452320341493)*x[0]**o+(0.755240168228)*x[0]+(-0.892552611066)*x[1]**o+(0.438275161458)*x[1]+(0.441378369388)*x[2]**o+(-0.870501114753)*x[2]
        arg[2,4,1,1]=(0.496731258286)*x[0]**o+(-0.499918335322)*x[0]+(0.000476061942027)*x[1]**o+(-0.857762621856)*x[1]+(0.957433798133)*x[2]**o+(-0.766450629361)*x[2]
        arg[2,4,2,0]=(0.860379598562)*x[0]**o+(-0.0816478389485)*x[0]+(-0.438181864676)*x[1]**o+(0.78937427676)*x[1]+(-0.97940010041)*x[2]**o+(0.583914605259)*x[2]
        arg[2,4,2,1]=(0.23729037277)*x[0]**o+(-0.365380141474)*x[0]+(-0.471505632575)*x[1]**o+(-0.168838288)*x[1]+(-0.519847065791)*x[2]**o+(-0.816100518668)*x[2]
        arg[3,0,0,0]=(0.729921808661)*x[0]**o+(0.732868255388)*x[0]+(0.305434623795)*x[1]**o+(-0.45402804883)*x[1]+(0.604368334171)*x[2]**o+(-0.179233711008)*x[2]
        arg[3,0,0,1]=(-0.208916687176)*x[0]**o+(-0.297916049539)*x[0]+(0.854941564423)*x[1]**o+(0.775355543915)*x[1]+(-0.373382357688)*x[2]**o+(0.792557314074)*x[2]
        arg[3,0,1,0]=(0.606041025674)*x[0]**o+(0.115814644681)*x[0]+(-0.414470914985)*x[1]**o+(-0.136577390174)*x[1]+(0.672303216396)*x[2]**o+(-0.479680752765)*x[2]
        arg[3,0,1,1]=(-0.535633649471)*x[0]**o+(0.133761944775)*x[0]+(-0.980107507665)*x[1]**o+(0.211561747114)*x[1]+(-0.530648770004)*x[2]**o+(0.417977629512)*x[2]
        arg[3,0,2,0]=(-0.242797826282)*x[0]**o+(-0.828691638823)*x[0]+(-0.187401612749)*x[1]**o+(-0.453821620107)*x[1]+(-0.475718408605)*x[2]**o+(0.683502702497)*x[2]
        arg[3,0,2,1]=(-0.66640812951)*x[0]**o+(-0.259184719546)*x[0]+(0.463743732654)*x[1]**o+(0.6746349156)*x[1]+(-0.0532918236783)*x[2]**o+(-0.714197225876)*x[2]
        arg[3,1,0,0]=(-0.97693616823)*x[0]**o+(-0.825560555757)*x[0]+(0.118452690664)*x[1]**o+(-0.125279874388)*x[1]+(0.733803136335)*x[2]**o+(-0.884902209956)*x[2]
        arg[3,1,0,1]=(-0.652466190334)*x[0]**o+(0.370106640243)*x[0]+(0.0850861818306)*x[1]**o+(-0.49744492012)*x[1]+(-0.843498619618)*x[2]**o+(-0.0480070808698)*x[2]
        arg[3,1,1,0]=(-0.450526704276)*x[0]**o+(0.430774951776)*x[0]+(-0.956799540696)*x[1]**o+(0.820090546747)*x[1]+(0.831582198997)*x[2]**o+(0.491042906394)*x[2]
        arg[3,1,1,1]=(0.419717807665)*x[0]**o+(0.320904534666)*x[0]+(-0.483247938999)*x[1]**o+(0.82174341564)*x[1]+(-0.441719998974)*x[2]**o+(0.691713322239)*x[2]
        arg[3,1,2,0]=(0.255464093123)*x[0]**o+(-0.256334953799)*x[0]+(0.677105284128)*x[1]**o+(-0.584438025212)*x[1]+(0.505820591597)*x[2]**o+(0.247943562937)*x[2]
        arg[3,1,2,1]=(-0.495204107745)*x[0]**o+(0.234426349601)*x[0]+(0.736969108189)*x[1]**o+(-0.369684409765)*x[1]+(0.164303379282)*x[2]**o+(-0.595757590519)*x[2]
        arg[3,2,0,0]=(0.238639684591)*x[0]**o+(0.568638143194)*x[0]+(0.972646679042)*x[1]**o+(-0.652648194795)*x[1]+(-0.0530713046414)*x[2]**o+(0.356043665396)*x[2]
        arg[3,2,0,1]=(0.458539145106)*x[0]**o+(0.111417037398)*x[0]+(-0.135374868193)*x[1]**o+(-0.535451685997)*x[1]+(-0.820797762877)*x[2]**o+(0.122239787933)*x[2]
        arg[3,2,1,0]=(-0.323251425039)*x[0]**o+(0.106998797206)*x[0]+(-0.0671655760766)*x[1]**o+(0.836342794315)*x[1]+(0.34628715493)*x[2]**o+(-0.11058327806)*x[2]
        arg[3,2,1,1]=(0.181994203911)*x[0]**o+(-0.710678453639)*x[0]+(0.625606409786)*x[1]**o+(-0.363170328543)*x[1]+(-0.932998745422)*x[2]**o+(-0.0411581284314)*x[2]
        arg[3,2,2,0]=(0.710847104903)*x[0]**o+(0.483971915285)*x[0]+(0.42996482533)*x[1]**o+(0.456817348253)*x[1]+(0.844562058604)*x[2]**o+(-0.469586014721)*x[2]
        arg[3,2,2,1]=(-0.845771280837)*x[0]**o+(-0.081384696494)*x[0]+(0.71823377714)*x[1]**o+(-0.846672143168)*x[1]+(-0.139540918461)*x[2]**o+(0.837646131987)*x[2]
        arg[3,3,0,0]=(-0.958726792109)*x[0]**o+(0.53196455268)*x[0]+(-0.621810423658)*x[1]**o+(-0.0352106197658)*x[1]+(-0.636059207474)*x[2]**o+(-0.0491118207063)*x[2]
        arg[3,3,0,1]=(0.987836799279)*x[0]**o+(0.41633576224)*x[0]+(0.312021799176)*x[1]**o+(-0.168836582938)*x[1]+(-0.391725003735)*x[2]**o+(-0.158000182269)*x[2]
        arg[3,3,1,0]=(-0.280868116529)*x[0]**o+(-0.687924000092)*x[0]+(0.865840229722)*x[1]**o+(-0.0823203022874)*x[1]+(-0.857309899784)*x[2]**o+(-0.0657473493059)*x[2]
        arg[3,3,1,1]=(0.680496030567)*x[0]**o+(0.159575720724)*x[0]+(0.0531895122046)*x[1]**o+(0.359273896084)*x[1]+(0.35084774502)*x[2]**o+(0.0982220259531)*x[2]
        arg[3,3,2,0]=(0.140872311111)*x[0]**o+(-0.00845309046907)*x[0]+(-0.845527142503)*x[1]**o+(-0.67713103521)*x[1]+(0.972054187828)*x[2]**o+(-0.800664963509)*x[2]
        arg[3,3,2,1]=(-0.490866666782)*x[0]**o+(-0.0988939523952)*x[0]+(0.236210969508)*x[1]**o+(-0.672617589992)*x[1]+(-0.640558724403)*x[2]**o+(-0.203125267713)*x[2]
        arg[3,4,0,0]=(0.318438839683)*x[0]**o+(0.0864322377812)*x[0]+(-0.371419701638)*x[1]**o+(-0.311583139142)*x[1]+(-0.0533111244011)*x[2]**o+(0.841240973423)*x[2]
        arg[3,4,0,1]=(-0.573110771334)*x[0]**o+(0.703852077914)*x[0]+(0.613309217217)*x[1]**o+(0.247223193846)*x[1]+(0.618952485205)*x[2]**o+(-0.395998308184)*x[2]
        arg[3,4,1,0]=(0.512896938934)*x[0]**o+(0.848653171505)*x[0]+(-0.944903687062)*x[1]**o+(0.179782177607)*x[1]+(-0.0287288996538)*x[2]**o+(-0.385519118293)*x[2]
        arg[3,4,1,1]=(0.234190647718)*x[0]**o+(0.445478221419)*x[0]+(0.588931676434)*x[1]**o+(0.140788139998)*x[1]+(-0.761952359657)*x[2]**o+(0.92377426359)*x[2]
        arg[3,4,2,0]=(-0.836038966171)*x[0]**o+(-0.772990045933)*x[0]+(0.0622128209877)*x[1]**o+(-0.73733496141)*x[1]+(0.0234446925161)*x[2]**o+(-0.682234213315)*x[2]
        arg[3,4,2,1]=(0.250498084214)*x[0]**o+(0.539088356159)*x[0]+(-0.691842680176)*x[1]**o+(0.710040802953)*x[1]+(-0.43691551501)*x[2]**o+(0.710668600989)*x[2]
        ref[0,0,0,0]=(-0.836302395465)*x_ref[0]**o+(-0.842568675604)*x_ref[0]+(-0.780133441207)*x_ref[1]**o+(0.54483947113)*x_ref[1]+(-0.130253019696)*x_ref[2]**o+(0.358303389828)*x_ref[2]
        ref[0,0,0,1]=(-0.722847500652)*x_ref[0]**o+(0.42887368878)*x_ref[0]+(0.979187056933)*x_ref[1]**o+(0.624875714982)*x_ref[1]+(-0.672529765247)*x_ref[2]**o+(0.822363583702)*x_ref[2]
        ref[0,0,1,0]=(-0.28989650691)*x_ref[0]**o+(-0.847647954773)*x_ref[0]+(-0.11119862113)*x_ref[1]**o+(0.0809009676818)*x_ref[1]+(-0.672626533587)*x_ref[2]**o+(0.580561232267)*x_ref[2]
        ref[0,0,1,1]=(-0.64783363248)*x_ref[0]**o+(0.212828552873)*x_ref[0]+(-0.845297180936)*x_ref[1]**o+(-0.714760411186)*x_ref[1]+(-0.385173605122)*x_ref[2]**o+(-0.760750317556)*x_ref[2]
        ref[0,0,2,0]=(-0.328757006046)*x_ref[0]**o+(-0.0935907722247)*x_ref[0]+(-0.895866135795)*x_ref[1]**o+(-0.701241976745)*x_ref[1]+(0.893684139941)*x_ref[2]**o+(-0.265339318222)*x_ref[2]
        ref[0,0,2,1]=(0.134066662346)*x_ref[0]**o+(0.0383231508856)*x_ref[0]+(-0.347744261995)*x_ref[1]**o+(-0.86897907336)*x_ref[1]+(-0.287792154521)*x_ref[2]**o+(0.859351846891)*x_ref[2]
        ref[0,1,0,0]=(-0.635128124936)*x_ref[0]**o+(-0.659462189733)*x_ref[0]+(-0.0964365747936)*x_ref[1]**o+(-0.385702168747)*x_ref[1]+(0.179539914057)*x_ref[2]**o+(-0.0176420847859)*x_ref[2]
        ref[0,1,0,1]=(-0.580750807831)*x_ref[0]**o+(0.756493147583)*x_ref[0]+(-0.436842251289)*x_ref[1]**o+(-0.705624646095)*x_ref[1]+(0.983064096262)*x_ref[2]**o+(-0.0197153825423)*x_ref[2]
        ref[0,1,1,0]=(-0.567748582365)*x_ref[0]**o+(0.579324420356)*x_ref[0]+(0.0434168493419)*x_ref[1]**o+(-0.839692920334)*x_ref[1]+(-0.119239847745)*x_ref[2]**o+(0.527047942834)*x_ref[2]
        ref[0,1,1,1]=(0.52870403655)*x_ref[0]**o+(0.700090892471)*x_ref[0]+(0.730224313921)*x_ref[1]**o+(-0.815451751207)*x_ref[1]+(-0.943945382413)*x_ref[2]**o+(-0.571714141916)*x_ref[2]
        ref[0,1,2,0]=(-0.803841094887)*x_ref[0]**o+(0.0595126006237)*x_ref[0]+(-0.612341797645)*x_ref[1]**o+(0.944552805211)*x_ref[1]+(-0.00253009731786)*x_ref[2]**o+(0.0174867439912)*x_ref[2]
        ref[0,1,2,1]=(-0.145623212549)*x_ref[0]**o+(0.546039957008)*x_ref[0]+(-0.221425237579)*x_ref[1]**o+(0.229367671159)*x_ref[1]+(-0.784903869759)*x_ref[2]**o+(0.898045208871)*x_ref[2]
        ref[0,2,0,0]=(0.417924494005)*x_ref[0]**o+(-0.192820051987)*x_ref[0]+(0.878916271492)*x_ref[1]**o+(-0.35056320984)*x_ref[1]+(0.982897575604)*x_ref[2]**o+(-0.225793730616)*x_ref[2]
        ref[0,2,0,1]=(-0.493279498849)*x_ref[0]**o+(-0.154177411087)*x_ref[0]+(-0.923592860492)*x_ref[1]**o+(0.872010949252)*x_ref[1]+(-0.936252147064)*x_ref[2]**o+(0.947136819308)*x_ref[2]
        ref[0,2,1,0]=(0.109357630607)*x_ref[0]**o+(-0.518825477709)*x_ref[0]+(-0.375510667498)*x_ref[1]**o+(0.476537999644)*x_ref[1]+(0.676017421029)*x_ref[2]**o+(-0.633301158984)*x_ref[2]
        ref[0,2,1,1]=(-0.0662265225539)*x_ref[0]**o+(-0.276634748263)*x_ref[0]+(-0.0655431981102)*x_ref[1]**o+(0.0578482920991)*x_ref[1]+(-0.264370611185)*x_ref[2]**o+(-0.674612985594)*x_ref[2]
        ref[0,2,2,0]=(-0.924186592743)*x_ref[0]**o+(0.197603043149)*x_ref[0]+(-0.676882805046)*x_ref[1]**o+(0.564142845865)*x_ref[1]+(0.590369345306)*x_ref[2]**o+(-0.872019490978)*x_ref[2]
        ref[0,2,2,1]=(0.480683776968)*x_ref[0]**o+(-0.412055284173)*x_ref[0]+(0.829996313473)*x_ref[1]**o+(-0.235665473048)*x_ref[1]+(-0.429191305284)*x_ref[2]**o+(-0.0867804318457)*x_ref[2]
        ref[0,3,0,0]=(-0.0903660006043)*x_ref[0]**o+(-0.507571901766)*x_ref[0]+(-0.0305305861809)*x_ref[1]**o+(-0.915027969989)*x_ref[1]+(0.955644774347)*x_ref[2]**o+(0.739244618775)*x_ref[2]
        ref[0,3,0,1]=(0.600619122998)*x_ref[0]**o+(0.0128699484527)*x_ref[0]+(-0.360431921281)*x_ref[1]**o+(0.304773619811)*x_ref[1]+(-0.302423314054)*x_ref[2]**o+(0.680149401572)*x_ref[2]
        ref[0,3,1,0]=(-0.638540513323)*x_ref[0]**o+(0.766655754067)*x_ref[0]+(0.276879854223)*x_ref[1]**o+(-0.628688275247)*x_ref[1]+(0.724584107114)*x_ref[2]**o+(0.682193551967)*x_ref[2]
        ref[0,3,1,1]=(-0.516656801349)*x_ref[0]**o+(-0.513901319383)*x_ref[0]+(-0.227957010777)*x_ref[1]**o+(-0.164363207703)*x_ref[1]+(-0.919936042047)*x_ref[2]**o+(0.682400184997)*x_ref[2]
        ref[0,3,2,0]=(-0.819629653095)*x_ref[0]**o+(0.183937049274)*x_ref[0]+(0.0411075800579)*x_ref[1]**o+(0.0813088270699)*x_ref[1]+(0.762416334575)*x_ref[2]**o+(0.236553726705)*x_ref[2]
        ref[0,3,2,1]=(0.513141183759)*x_ref[0]**o+(-0.452475512734)*x_ref[0]+(0.0461424221984)*x_ref[1]**o+(-0.673452844636)*x_ref[1]+(-0.194493328878)*x_ref[2]**o+(0.812459956422)*x_ref[2]
        ref[0,4,0,0]=(0.672167415195)*x_ref[0]**o+(-0.641004730405)*x_ref[0]+(-0.383031420851)*x_ref[1]**o+(0.392366431186)*x_ref[1]+(0.66479124224)*x_ref[2]**o+(0.845503225104)*x_ref[2]
        ref[0,4,0,1]=(0.908241966695)*x_ref[0]**o+(-0.816989438726)*x_ref[0]+(-0.144902267583)*x_ref[1]**o+(-0.654901540598)*x_ref[1]+(-0.9840470269)*x_ref[2]**o+(0.510970695154)*x_ref[2]
        ref[0,4,1,0]=(0.788858704366)*x_ref[0]**o+(-0.50266306504)*x_ref[0]+(-0.681248796723)*x_ref[1]**o+(0.0317663112509)*x_ref[1]+(0.662428035619)*x_ref[2]**o+(0.377116420912)*x_ref[2]
        ref[0,4,1,1]=(-0.0734522829037)*x_ref[0]**o+(0.0822741484655)*x_ref[0]+(-0.509279578815)*x_ref[1]**o+(-0.95372420098)*x_ref[1]+(0.682868362508)*x_ref[2]**o+(-0.947156883638)*x_ref[2]
        ref[0,4,2,0]=(-0.843236607337)*x_ref[0]**o+(-0.925326238513)*x_ref[0]+(-0.153688519983)*x_ref[1]**o+(-0.68456413487)*x_ref[1]+(0.920442617368)*x_ref[2]**o+(0.852945557893)*x_ref[2]
        ref[0,4,2,1]=(-0.283772512528)*x_ref[0]**o+(0.202730738521)*x_ref[0]+(-0.287375172356)*x_ref[1]**o+(-0.176699132816)*x_ref[1]+(0.446391113264)*x_ref[2]**o+(-0.783266847348)*x_ref[2]
        ref[1,0,0,0]=(0.81284364562)*x_ref[0]**o+(-0.454432332505)*x_ref[0]+(-0.684692596262)*x_ref[1]**o+(-0.345923888237)*x_ref[1]+(0.1201728631)*x_ref[2]**o+(0.808246135028)*x_ref[2]
        ref[1,0,0,1]=(0.941104328706)*x_ref[0]**o+(-0.802008654964)*x_ref[0]+(0.902047483677)*x_ref[1]**o+(0.777049224612)*x_ref[1]+(0.120184332854)*x_ref[2]**o+(0.983216899102)*x_ref[2]
        ref[1,0,1,0]=(-0.420996925767)*x_ref[0]**o+(0.0941750615502)*x_ref[0]+(-0.935342933578)*x_ref[1]**o+(0.925891822058)*x_ref[1]+(-0.153652985582)*x_ref[2]**o+(-0.336548986153)*x_ref[2]
        ref[1,0,1,1]=(-0.692886221795)*x_ref[0]**o+(-0.920419410139)*x_ref[0]+(-0.586575583263)*x_ref[1]**o+(-0.269347362009)*x_ref[1]+(-0.872169736045)*x_ref[2]**o+(-0.724727186336)*x_ref[2]
        ref[1,0,2,0]=(-0.0171222282058)*x_ref[0]**o+(0.338180602976)*x_ref[0]+(-0.0345958505395)*x_ref[1]**o+(0.631728421505)*x_ref[1]+(0.90596443453)*x_ref[2]**o+(-0.144785193015)*x_ref[2]
        ref[1,0,2,1]=(-0.91636583885)*x_ref[0]**o+(-0.807496789574)*x_ref[0]+(0.63472495957)*x_ref[1]**o+(0.966464343993)*x_ref[1]+(-0.471311384621)*x_ref[2]**o+(-0.449420138238)*x_ref[2]
        ref[1,1,0,0]=(-0.239930503474)*x_ref[0]**o+(-0.476439715994)*x_ref[0]+(0.421600511888)*x_ref[1]**o+(0.477823144792)*x_ref[1]+(0.251409914807)*x_ref[2]**o+(0.636818001419)*x_ref[2]
        ref[1,1,0,1]=(-0.427510410263)*x_ref[0]**o+(0.494810857516)*x_ref[0]+(-0.129415181352)*x_ref[1]**o+(0.754147770071)*x_ref[1]+(0.0218776365985)*x_ref[2]**o+(0.706183216616)*x_ref[2]
        ref[1,1,1,0]=(-0.469150375165)*x_ref[0]**o+(-0.0651101638073)*x_ref[0]+(-0.697853003861)*x_ref[1]**o+(0.0389330102026)*x_ref[1]+(-0.256557942045)*x_ref[2]**o+(0.409565340425)*x_ref[2]
        ref[1,1,1,1]=(0.14957005608)*x_ref[0]**o+(-0.209482968126)*x_ref[0]+(0.368899650309)*x_ref[1]**o+(-0.0899321000137)*x_ref[1]+(0.282389903357)*x_ref[2]**o+(-0.166633883916)*x_ref[2]
        ref[1,1,2,0]=(0.0374938231714)*x_ref[0]**o+(-0.687972899893)*x_ref[0]+(-0.0686478107779)*x_ref[1]**o+(0.318977937888)*x_ref[1]+(-0.601269126788)*x_ref[2]**o+(0.453188855786)*x_ref[2]
        ref[1,1,2,1]=(-0.853894223222)*x_ref[0]**o+(0.441830151478)*x_ref[0]+(0.973114898368)*x_ref[1]**o+(-0.264454889071)*x_ref[1]+(-0.828332687531)*x_ref[2]**o+(0.302493748565)*x_ref[2]
        ref[1,2,0,0]=(0.210365824111)*x_ref[0]**o+(0.4674117857)*x_ref[0]+(-0.129546969044)*x_ref[1]**o+(0.566357125352)*x_ref[1]+(-0.721733448514)*x_ref[2]**o+(0.154253207106)*x_ref[2]
        ref[1,2,0,1]=(0.805325508181)*x_ref[0]**o+(-0.929113563452)*x_ref[0]+(-0.487301761321)*x_ref[1]**o+(-0.822405526519)*x_ref[1]+(0.254754490326)*x_ref[2]**o+(-0.196480710709)*x_ref[2]
        ref[1,2,1,0]=(-0.520777275475)*x_ref[0]**o+(-0.912569745644)*x_ref[0]+(-0.784504791087)*x_ref[1]**o+(-0.903473217653)*x_ref[1]+(0.583297107374)*x_ref[2]**o+(0.664811898195)*x_ref[2]
        ref[1,2,1,1]=(0.129152152958)*x_ref[0]**o+(0.497832200867)*x_ref[0]+(0.121217248234)*x_ref[1]**o+(-0.183368066714)*x_ref[1]+(-0.776145485536)*x_ref[2]**o+(-0.223165724715)*x_ref[2]
        ref[1,2,2,0]=(-0.698981324488)*x_ref[0]**o+(0.00951277777924)*x_ref[0]+(0.997410976507)*x_ref[1]**o+(0.534670303453)*x_ref[1]+(0.251309223218)*x_ref[2]**o+(-0.921148646979)*x_ref[2]
        ref[1,2,2,1]=(-0.731509934487)*x_ref[0]**o+(0.87731926834)*x_ref[0]+(0.0865507713913)*x_ref[1]**o+(0.708835987019)*x_ref[1]+(0.294182269171)*x_ref[2]**o+(-0.316660750729)*x_ref[2]
        ref[1,3,0,0]=(-0.282903962918)*x_ref[0]**o+(-0.793457978844)*x_ref[0]+(0.287425807914)*x_ref[1]**o+(-0.0252722364835)*x_ref[1]+(-0.956202022515)*x_ref[2]**o+(0.650271579135)*x_ref[2]
        ref[1,3,0,1]=(-0.946723111202)*x_ref[0]**o+(0.669431653693)*x_ref[0]+(-0.791125976188)*x_ref[1]**o+(-0.466484914794)*x_ref[1]+(-0.217225801872)*x_ref[2]**o+(0.224719473996)*x_ref[2]
        ref[1,3,1,0]=(0.516479301043)*x_ref[0]**o+(-0.42420332366)*x_ref[0]+(-0.0649227247176)*x_ref[1]**o+(0.208849752675)*x_ref[1]+(-0.939001099693)*x_ref[2]**o+(-0.0537827140932)*x_ref[2]
        ref[1,3,1,1]=(-0.292077069459)*x_ref[0]**o+(-0.444848119744)*x_ref[0]+(0.268592719199)*x_ref[1]**o+(-0.349401060323)*x_ref[1]+(-0.540803784489)*x_ref[2]**o+(-0.487170123526)*x_ref[2]
        ref[1,3,2,0]=(-0.626222042661)*x_ref[0]**o+(0.245937409624)*x_ref[0]+(-0.24692162006)*x_ref[1]**o+(0.108230029208)*x_ref[1]+(-0.760463241928)*x_ref[2]**o+(-0.587078263201)*x_ref[2]
        ref[1,3,2,1]=(-0.923891147756)*x_ref[0]**o+(0.0961285842162)*x_ref[0]+(0.0146853466585)*x_ref[1]**o+(0.00951339122359)*x_ref[1]+(-0.261383704723)*x_ref[2]**o+(-0.0262515691324)*x_ref[2]
        ref[1,4,0,0]=(-0.0370003524292)*x_ref[0]**o+(-0.0463848489957)*x_ref[0]+(0.940061458694)*x_ref[1]**o+(0.592787058925)*x_ref[1]+(-0.217284379638)*x_ref[2]**o+(-0.11746834109)*x_ref[2]
        ref[1,4,0,1]=(0.830404283093)*x_ref[0]**o+(0.268135490403)*x_ref[0]+(-0.894500398195)*x_ref[1]**o+(0.751042648917)*x_ref[1]+(0.687583292754)*x_ref[2]**o+(-0.884596509742)*x_ref[2]
        ref[1,4,1,0]=(0.414764592469)*x_ref[0]**o+(-0.529613347919)*x_ref[0]+(-0.727435115594)*x_ref[1]**o+(-0.546579777725)*x_ref[1]+(0.378011893971)*x_ref[2]**o+(-0.578081183238)*x_ref[2]
        ref[1,4,1,1]=(0.406513376747)*x_ref[0]**o+(-0.350132869579)*x_ref[0]+(-0.557954631299)*x_ref[1]**o+(0.868176175754)*x_ref[1]+(0.793021832632)*x_ref[2]**o+(0.280529771851)*x_ref[2]
        ref[1,4,2,0]=(0.56741953575)*x_ref[0]**o+(0.0115947684302)*x_ref[0]+(0.872812725363)*x_ref[1]**o+(-0.215183541745)*x_ref[1]+(0.64853692203)*x_ref[2]**o+(-0.139561471922)*x_ref[2]
        ref[1,4,2,1]=(0.650916904666)*x_ref[0]**o+(0.878765271839)*x_ref[0]+(0.0141264348052)*x_ref[1]**o+(-0.56314426774)*x_ref[1]+(-0.267854377704)*x_ref[2]**o+(0.852271380667)*x_ref[2]
        ref[2,0,0,0]=(0.948359141577)*x_ref[0]**o+(-0.0881747280974)*x_ref[0]+(0.804871563368)*x_ref[1]**o+(-0.0298585763446)*x_ref[1]+(0.506923333569)*x_ref[2]**o+(-0.920947093648)*x_ref[2]
        ref[2,0,0,1]=(-0.0144221258584)*x_ref[0]**o+(-0.679723072824)*x_ref[0]+(0.00829336956687)*x_ref[1]**o+(0.675680809303)*x_ref[1]+(-0.803195231536)*x_ref[2]**o+(0.70240358571)*x_ref[2]
        ref[2,0,1,0]=(-0.159174987706)*x_ref[0]**o+(0.922309271483)*x_ref[0]+(0.344354343035)*x_ref[1]**o+(-0.00954038311323)*x_ref[1]+(-0.50225452804)*x_ref[2]**o+(0.355899734308)*x_ref[2]
        ref[2,0,1,1]=(0.0857611780404)*x_ref[0]**o+(0.876479431983)*x_ref[0]+(0.105049158502)*x_ref[1]**o+(-0.530308461426)*x_ref[1]+(-0.553136649343)*x_ref[2]**o+(0.385458516481)*x_ref[2]
        ref[2,0,2,0]=(-0.260086070842)*x_ref[0]**o+(0.417766241499)*x_ref[0]+(-0.292370383673)*x_ref[1]**o+(-0.321315349553)*x_ref[1]+(-0.437301444082)*x_ref[2]**o+(0.730845669715)*x_ref[2]
        ref[2,0,2,1]=(0.132547674276)*x_ref[0]**o+(0.783602886333)*x_ref[0]+(0.822361450413)*x_ref[1]**o+(-0.810402783)*x_ref[1]+(0.759701191023)*x_ref[2]**o+(0.644375788608)*x_ref[2]
        ref[2,1,0,0]=(-0.841458111227)*x_ref[0]**o+(0.779917641736)*x_ref[0]+(-0.0745587217256)*x_ref[1]**o+(0.938545280143)*x_ref[1]+(0.454600163373)*x_ref[2]**o+(-0.490348377697)*x_ref[2]
        ref[2,1,0,1]=(-0.876477136174)*x_ref[0]**o+(-0.248200369939)*x_ref[0]+(-0.973654243816)*x_ref[1]**o+(0.482402123794)*x_ref[1]+(-0.776970490827)*x_ref[2]**o+(-0.571349467703)*x_ref[2]
        ref[2,1,1,0]=(-0.433528193102)*x_ref[0]**o+(-0.57712915767)*x_ref[0]+(-0.328671198058)*x_ref[1]**o+(-0.926873019075)*x_ref[1]+(0.314232739356)*x_ref[2]**o+(-0.199625466412)*x_ref[2]
        ref[2,1,1,1]=(0.892707770392)*x_ref[0]**o+(-0.680081898234)*x_ref[0]+(-0.754181058483)*x_ref[1]**o+(-0.253728945575)*x_ref[1]+(-0.906373406684)*x_ref[2]**o+(-0.0266435080833)*x_ref[2]
        ref[2,1,2,0]=(-0.121605378472)*x_ref[0]**o+(-0.503955015634)*x_ref[0]+(-0.0960767196114)*x_ref[1]**o+(-0.937914890361)*x_ref[1]+(0.212042931853)*x_ref[2]**o+(0.930703155408)*x_ref[2]
        ref[2,1,2,1]=(0.0564159441047)*x_ref[0]**o+(0.842332721105)*x_ref[0]+(0.954741657608)*x_ref[1]**o+(0.73503176)*x_ref[1]+(-0.390679938582)*x_ref[2]**o+(-0.440014747741)*x_ref[2]
        ref[2,2,0,0]=(0.0100637970821)*x_ref[0]**o+(-0.0283353154145)*x_ref[0]+(0.865360758262)*x_ref[1]**o+(0.780975027988)*x_ref[1]+(0.16115842919)*x_ref[2]**o+(0.247580155656)*x_ref[2]
        ref[2,2,0,1]=(0.61102059426)*x_ref[0]**o+(-0.198005744101)*x_ref[0]+(0.503568797262)*x_ref[1]**o+(0.0288809726497)*x_ref[1]+(-0.765127098183)*x_ref[2]**o+(0.409803857723)*x_ref[2]
        ref[2,2,1,0]=(-0.712272018295)*x_ref[0]**o+(-0.62562142063)*x_ref[0]+(-0.579501949719)*x_ref[1]**o+(0.628539275236)*x_ref[1]+(0.0200945124964)*x_ref[2]**o+(0.175966024769)*x_ref[2]
        ref[2,2,1,1]=(-0.798244445519)*x_ref[0]**o+(-0.376440962535)*x_ref[0]+(-0.134467401497)*x_ref[1]**o+(0.877538294477)*x_ref[1]+(-0.484870741192)*x_ref[2]**o+(-0.224064469935)*x_ref[2]
        ref[2,2,2,0]=(0.200196926368)*x_ref[0]**o+(-0.528000867086)*x_ref[0]+(-0.708942342624)*x_ref[1]**o+(-0.591994969141)*x_ref[1]+(-0.990641592656)*x_ref[2]**o+(-0.875897347303)*x_ref[2]
        ref[2,2,2,1]=(0.949113190797)*x_ref[0]**o+(0.867855496786)*x_ref[0]+(-0.0314953438065)*x_ref[1]**o+(-0.371928428005)*x_ref[1]+(-0.526403191063)*x_ref[2]**o+(0.653049122326)*x_ref[2]
        ref[2,3,0,0]=(-0.353418327127)*x_ref[0]**o+(-0.418588306429)*x_ref[0]+(0.400872043269)*x_ref[1]**o+(0.141907924042)*x_ref[1]+(-0.999869579462)*x_ref[2]**o+(-0.410423954563)*x_ref[2]
        ref[2,3,0,1]=(-0.240390877307)*x_ref[0]**o+(0.0901879916115)*x_ref[0]+(-0.0734683209657)*x_ref[1]**o+(0.274836005886)*x_ref[1]+(-0.336375498181)*x_ref[2]**o+(-0.754776455221)*x_ref[2]
        ref[2,3,1,0]=(0.259928460822)*x_ref[0]**o+(-0.227630454708)*x_ref[0]+(0.725968066278)*x_ref[1]**o+(-0.545440695763)*x_ref[1]+(-0.0354926607815)*x_ref[2]**o+(0.365010532181)*x_ref[2]
        ref[2,3,1,1]=(-0.969799373398)*x_ref[0]**o+(0.514498233214)*x_ref[0]+(0.907450925318)*x_ref[1]**o+(-0.234993706406)*x_ref[1]+(0.206323910806)*x_ref[2]**o+(0.209364250369)*x_ref[2]
        ref[2,3,2,0]=(0.712905565941)*x_ref[0]**o+(-0.353322758263)*x_ref[0]+(0.816861181494)*x_ref[1]**o+(-0.362388243619)*x_ref[1]+(0.302543517607)*x_ref[2]**o+(-0.831280418852)*x_ref[2]
        ref[2,3,2,1]=(-0.386373711182)*x_ref[0]**o+(0.946914567792)*x_ref[0]+(-0.52795085887)*x_ref[1]**o+(0.86635245964)*x_ref[1]+(0.654261054178)*x_ref[2]**o+(0.452334882498)*x_ref[2]
        ref[2,4,0,0]=(0.374747033547)*x_ref[0]**o+(-0.3471578669)*x_ref[0]+(0.322217002639)*x_ref[1]**o+(0.92988068982)*x_ref[1]+(-0.642423472369)*x_ref[2]**o+(-0.759623225653)*x_ref[2]
        ref[2,4,0,1]=(-0.322255157841)*x_ref[0]**o+(-0.253806844569)*x_ref[0]+(-0.0688746199637)*x_ref[1]**o+(0.232840317719)*x_ref[1]+(-0.815623721026)*x_ref[2]**o+(0.43055485562)*x_ref[2]
        ref[2,4,1,0]=(-0.452320341493)*x_ref[0]**o+(0.755240168228)*x_ref[0]+(-0.892552611066)*x_ref[1]**o+(0.438275161458)*x_ref[1]+(0.441378369388)*x_ref[2]**o+(-0.870501114753)*x_ref[2]
        ref[2,4,1,1]=(0.496731258286)*x_ref[0]**o+(-0.499918335322)*x_ref[0]+(0.000476061942027)*x_ref[1]**o+(-0.857762621856)*x_ref[1]+(0.957433798133)*x_ref[2]**o+(-0.766450629361)*x_ref[2]
        ref[2,4,2,0]=(0.860379598562)*x_ref[0]**o+(-0.0816478389485)*x_ref[0]+(-0.438181864676)*x_ref[1]**o+(0.78937427676)*x_ref[1]+(-0.97940010041)*x_ref[2]**o+(0.583914605259)*x_ref[2]
        ref[2,4,2,1]=(0.23729037277)*x_ref[0]**o+(-0.365380141474)*x_ref[0]+(-0.471505632575)*x_ref[1]**o+(-0.168838288)*x_ref[1]+(-0.519847065791)*x_ref[2]**o+(-0.816100518668)*x_ref[2]
        ref[3,0,0,0]=(0.729921808661)*x_ref[0]**o+(0.732868255388)*x_ref[0]+(0.305434623795)*x_ref[1]**o+(-0.45402804883)*x_ref[1]+(0.604368334171)*x_ref[2]**o+(-0.179233711008)*x_ref[2]
        ref[3,0,0,1]=(-0.208916687176)*x_ref[0]**o+(-0.297916049539)*x_ref[0]+(0.854941564423)*x_ref[1]**o+(0.775355543915)*x_ref[1]+(-0.373382357688)*x_ref[2]**o+(0.792557314074)*x_ref[2]
        ref[3,0,1,0]=(0.606041025674)*x_ref[0]**o+(0.115814644681)*x_ref[0]+(-0.414470914985)*x_ref[1]**o+(-0.136577390174)*x_ref[1]+(0.672303216396)*x_ref[2]**o+(-0.479680752765)*x_ref[2]
        ref[3,0,1,1]=(-0.535633649471)*x_ref[0]**o+(0.133761944775)*x_ref[0]+(-0.980107507665)*x_ref[1]**o+(0.211561747114)*x_ref[1]+(-0.530648770004)*x_ref[2]**o+(0.417977629512)*x_ref[2]
        ref[3,0,2,0]=(-0.242797826282)*x_ref[0]**o+(-0.828691638823)*x_ref[0]+(-0.187401612749)*x_ref[1]**o+(-0.453821620107)*x_ref[1]+(-0.475718408605)*x_ref[2]**o+(0.683502702497)*x_ref[2]
        ref[3,0,2,1]=(-0.66640812951)*x_ref[0]**o+(-0.259184719546)*x_ref[0]+(0.463743732654)*x_ref[1]**o+(0.6746349156)*x_ref[1]+(-0.0532918236783)*x_ref[2]**o+(-0.714197225876)*x_ref[2]
        ref[3,1,0,0]=(-0.97693616823)*x_ref[0]**o+(-0.825560555757)*x_ref[0]+(0.118452690664)*x_ref[1]**o+(-0.125279874388)*x_ref[1]+(0.733803136335)*x_ref[2]**o+(-0.884902209956)*x_ref[2]
        ref[3,1,0,1]=(-0.652466190334)*x_ref[0]**o+(0.370106640243)*x_ref[0]+(0.0850861818306)*x_ref[1]**o+(-0.49744492012)*x_ref[1]+(-0.843498619618)*x_ref[2]**o+(-0.0480070808698)*x_ref[2]
        ref[3,1,1,0]=(-0.450526704276)*x_ref[0]**o+(0.430774951776)*x_ref[0]+(-0.956799540696)*x_ref[1]**o+(0.820090546747)*x_ref[1]+(0.831582198997)*x_ref[2]**o+(0.491042906394)*x_ref[2]
        ref[3,1,1,1]=(0.419717807665)*x_ref[0]**o+(0.320904534666)*x_ref[0]+(-0.483247938999)*x_ref[1]**o+(0.82174341564)*x_ref[1]+(-0.441719998974)*x_ref[2]**o+(0.691713322239)*x_ref[2]
        ref[3,1,2,0]=(0.255464093123)*x_ref[0]**o+(-0.256334953799)*x_ref[0]+(0.677105284128)*x_ref[1]**o+(-0.584438025212)*x_ref[1]+(0.505820591597)*x_ref[2]**o+(0.247943562937)*x_ref[2]
        ref[3,1,2,1]=(-0.495204107745)*x_ref[0]**o+(0.234426349601)*x_ref[0]+(0.736969108189)*x_ref[1]**o+(-0.369684409765)*x_ref[1]+(0.164303379282)*x_ref[2]**o+(-0.595757590519)*x_ref[2]
        ref[3,2,0,0]=(0.238639684591)*x_ref[0]**o+(0.568638143194)*x_ref[0]+(0.972646679042)*x_ref[1]**o+(-0.652648194795)*x_ref[1]+(-0.0530713046414)*x_ref[2]**o+(0.356043665396)*x_ref[2]
        ref[3,2,0,1]=(0.458539145106)*x_ref[0]**o+(0.111417037398)*x_ref[0]+(-0.135374868193)*x_ref[1]**o+(-0.535451685997)*x_ref[1]+(-0.820797762877)*x_ref[2]**o+(0.122239787933)*x_ref[2]
        ref[3,2,1,0]=(-0.323251425039)*x_ref[0]**o+(0.106998797206)*x_ref[0]+(-0.0671655760766)*x_ref[1]**o+(0.836342794315)*x_ref[1]+(0.34628715493)*x_ref[2]**o+(-0.11058327806)*x_ref[2]
        ref[3,2,1,1]=(0.181994203911)*x_ref[0]**o+(-0.710678453639)*x_ref[0]+(0.625606409786)*x_ref[1]**o+(-0.363170328543)*x_ref[1]+(-0.932998745422)*x_ref[2]**o+(-0.0411581284314)*x_ref[2]
        ref[3,2,2,0]=(0.710847104903)*x_ref[0]**o+(0.483971915285)*x_ref[0]+(0.42996482533)*x_ref[1]**o+(0.456817348253)*x_ref[1]+(0.844562058604)*x_ref[2]**o+(-0.469586014721)*x_ref[2]
        ref[3,2,2,1]=(-0.845771280837)*x_ref[0]**o+(-0.081384696494)*x_ref[0]+(0.71823377714)*x_ref[1]**o+(-0.846672143168)*x_ref[1]+(-0.139540918461)*x_ref[2]**o+(0.837646131987)*x_ref[2]
        ref[3,3,0,0]=(-0.958726792109)*x_ref[0]**o+(0.53196455268)*x_ref[0]+(-0.621810423658)*x_ref[1]**o+(-0.0352106197658)*x_ref[1]+(-0.636059207474)*x_ref[2]**o+(-0.0491118207063)*x_ref[2]
        ref[3,3,0,1]=(0.987836799279)*x_ref[0]**o+(0.41633576224)*x_ref[0]+(0.312021799176)*x_ref[1]**o+(-0.168836582938)*x_ref[1]+(-0.391725003735)*x_ref[2]**o+(-0.158000182269)*x_ref[2]
        ref[3,3,1,0]=(-0.280868116529)*x_ref[0]**o+(-0.687924000092)*x_ref[0]+(0.865840229722)*x_ref[1]**o+(-0.0823203022874)*x_ref[1]+(-0.857309899784)*x_ref[2]**o+(-0.0657473493059)*x_ref[2]
        ref[3,3,1,1]=(0.680496030567)*x_ref[0]**o+(0.159575720724)*x_ref[0]+(0.0531895122046)*x_ref[1]**o+(0.359273896084)*x_ref[1]+(0.35084774502)*x_ref[2]**o+(0.0982220259531)*x_ref[2]
        ref[3,3,2,0]=(0.140872311111)*x_ref[0]**o+(-0.00845309046907)*x_ref[0]+(-0.845527142503)*x_ref[1]**o+(-0.67713103521)*x_ref[1]+(0.972054187828)*x_ref[2]**o+(-0.800664963509)*x_ref[2]
        ref[3,3,2,1]=(-0.490866666782)*x_ref[0]**o+(-0.0988939523952)*x_ref[0]+(0.236210969508)*x_ref[1]**o+(-0.672617589992)*x_ref[1]+(-0.640558724403)*x_ref[2]**o+(-0.203125267713)*x_ref[2]
        ref[3,4,0,0]=(0.318438839683)*x_ref[0]**o+(0.0864322377812)*x_ref[0]+(-0.371419701638)*x_ref[1]**o+(-0.311583139142)*x_ref[1]+(-0.0533111244011)*x_ref[2]**o+(0.841240973423)*x_ref[2]
        ref[3,4,0,1]=(-0.573110771334)*x_ref[0]**o+(0.703852077914)*x_ref[0]+(0.613309217217)*x_ref[1]**o+(0.247223193846)*x_ref[1]+(0.618952485205)*x_ref[2]**o+(-0.395998308184)*x_ref[2]
        ref[3,4,1,0]=(0.512896938934)*x_ref[0]**o+(0.848653171505)*x_ref[0]+(-0.944903687062)*x_ref[1]**o+(0.179782177607)*x_ref[1]+(-0.0287288996538)*x_ref[2]**o+(-0.385519118293)*x_ref[2]
        ref[3,4,1,1]=(0.234190647718)*x_ref[0]**o+(0.445478221419)*x_ref[0]+(0.588931676434)*x_ref[1]**o+(0.140788139998)*x_ref[1]+(-0.761952359657)*x_ref[2]**o+(0.92377426359)*x_ref[2]
        ref[3,4,2,0]=(-0.836038966171)*x_ref[0]**o+(-0.772990045933)*x_ref[0]+(0.0622128209877)*x_ref[1]**o+(-0.73733496141)*x_ref[1]+(0.0234446925161)*x_ref[2]**o+(-0.682234213315)*x_ref[2]
        ref[3,4,2,1]=(0.250498084214)*x_ref[0]**o+(0.539088356159)*x_ref[0]+(-0.691842680176)*x_ref[1]**o+(0.710040802953)*x_ref[1]+(-0.43691551501)*x_ref[2]**o+(0.710668600989)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.742484780403)*x[0]+(0.680847966059)*x[1]
        ref=(0.742484780403)*x_ref[0]+(0.680847966059)*x_ref[1]
      else:
        arg=(-0.928733928018)*x[0]+(-0.0797736636663)*x[1]+(0.517280695569)*x[2]
        ref=(-0.928733928018)*x_ref[0]+(-0.0797736636663)*x_ref[1]+(0.517280695569)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        arg[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
        ref[0]=(0.697998850914)*x[0]+(0.0279368494078)*x[1]
        ref[1]=(1.03722470341)*x[0]+(-1.46776547946)*x[1]
      else:
        arg[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        arg[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
        ref[0]=(-0.541575573414)*x[0]+(-0.200155171505)*x[1]+(-0.750342574153)*x[2]
        ref[1]=(0.47794945122)*x[0]+(0.151104203771)*x[1]+(-0.522836359633)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0998962983997)*x[0]+(0.52180906409)*x[1]
        arg[0,1]=(-0.0315655223925)*x[0]+(-0.0162720186305)*x[1]
        arg[0,2]=(0.361840834336)*x[0]+(-1.08346005817)*x[1]
        arg[0,3]=(0.673393991338)*x[0]+(0.571192080367)*x[1]
        arg[0,4]=(-0.725648335559)*x[0]+(0.866678056281)*x[1]
        arg[1,0]=(0.7094760871)*x[0]+(0.666899441355)*x[1]
        arg[1,1]=(-0.0152490017553)*x[0]+(-1.1169188656)*x[1]
        arg[1,2]=(-1.01408330279)*x[0]+(-1.10348930367)*x[1]
        arg[1,3]=(0.690904468829)*x[0]+(1.12602486331)*x[1]
        arg[1,4]=(-0.0391507405365)*x[0]+(-0.338120635003)*x[1]
        arg[2,0]=(0.175234757619)*x[0]+(1.74258329734)*x[1]
        arg[2,1]=(0.281950269687)*x[0]+(0.0489231995634)*x[1]
        arg[2,2]=(1.24244714043)*x[0]+(0.0149577033777)*x[1]
        arg[2,3]=(-0.706844294482)*x[0]+(-1.85050242116)*x[1]
        arg[2,4]=(-0.494011860105)*x[0]+(0.0809275159202)*x[1]
        arg[3,0]=(-0.380877016583)*x[0]+(0.0345520619524)*x[1]
        arg[3,1]=(1.10464018251)*x[0]+(0.0246274384536)*x[1]
        arg[3,2]=(-0.00138129982685)*x[0]+(0.776306480938)*x[1]
        arg[3,3]=(-0.0830044393648)*x[0]+(-0.173691604089)*x[1]
        arg[3,4]=(0.0909675137463)*x[0]+(0.208083539568)*x[1]
        ref[0,0]=(0.0998962983997)*x_ref[0]+(0.52180906409)*x_ref[1]
        ref[0,1]=(-0.0315655223925)*x_ref[0]+(-0.0162720186305)*x_ref[1]
        ref[0,2]=(0.361840834336)*x_ref[0]+(-1.08346005817)*x_ref[1]
        ref[0,3]=(0.673393991338)*x_ref[0]+(0.571192080367)*x_ref[1]
        ref[0,4]=(-0.725648335559)*x_ref[0]+(0.866678056281)*x_ref[1]
        ref[1,0]=(0.7094760871)*x_ref[0]+(0.666899441355)*x_ref[1]
        ref[1,1]=(-0.0152490017553)*x_ref[0]+(-1.1169188656)*x_ref[1]
        ref[1,2]=(-1.01408330279)*x_ref[0]+(-1.10348930367)*x_ref[1]
        ref[1,3]=(0.690904468829)*x_ref[0]+(1.12602486331)*x_ref[1]
        ref[1,4]=(-0.0391507405365)*x_ref[0]+(-0.338120635003)*x_ref[1]
        ref[2,0]=(0.175234757619)*x_ref[0]+(1.74258329734)*x_ref[1]
        ref[2,1]=(0.281950269687)*x_ref[0]+(0.0489231995634)*x_ref[1]
        ref[2,2]=(1.24244714043)*x_ref[0]+(0.0149577033777)*x_ref[1]
        ref[2,3]=(-0.706844294482)*x_ref[0]+(-1.85050242116)*x_ref[1]
        ref[2,4]=(-0.494011860105)*x_ref[0]+(0.0809275159202)*x_ref[1]
        ref[3,0]=(-0.380877016583)*x_ref[0]+(0.0345520619524)*x_ref[1]
        ref[3,1]=(1.10464018251)*x_ref[0]+(0.0246274384536)*x_ref[1]
        ref[3,2]=(-0.00138129982685)*x_ref[0]+(0.776306480938)*x_ref[1]
        ref[3,3]=(-0.0830044393648)*x_ref[0]+(-0.173691604089)*x_ref[1]
        ref[3,4]=(0.0909675137463)*x_ref[0]+(0.208083539568)*x_ref[1]
      else:
        arg[0,0]=(0.0539484604772)*x[0]+(-0.220817236218)*x[1]+(-0.853759265774)*x[2]
        arg[0,1]=(0.336027623724)*x[0]+(-0.111775563754)*x[1]+(0.646672879096)*x[2]
        arg[0,2]=(0.650722528854)*x[0]+(0.0924672090652)*x[1]+(-0.881798566157)*x[2]
        arg[0,3]=(0.276745374112)*x[0]+(-1.34491685419)*x[1]+(-0.833420124328)*x[2]
        arg[0,4]=(1.50686931219)*x[0]+(0.489530477585)*x[1]+(1.74836249628)*x[2]
        arg[1,0]=(-0.240001512118)*x[0]+(0.350823736593)*x[1]+(-0.227106949276)*x[2]
        arg[1,1]=(0.354636302436)*x[0]+(-0.0893490250257)*x[1]+(-0.0650366762176)*x[2]
        arg[1,2]=(-0.319714829694)*x[0]+(0.43056121683)*x[1]+(0.109856817034)*x[2]
        arg[1,3]=(-1.17581255544)*x[0]+(-0.491431894533)*x[1]+(0.220857799733)*x[2]
        arg[1,4]=(1.5680027633)*x[0]+(-0.249723385268)*x[1]+(-1.72056978299)*x[2]
        arg[2,0]=(0.197753675106)*x[0]+(0.314723848519)*x[1]+(0.718038926838)*x[2]
        arg[2,1]=(0.13916867566)*x[0]+(-0.0345156812431)*x[1]+(0.34208613933)*x[2]
        arg[2,2]=(-0.275775747114)*x[0]+(0.299200648073)*x[1]+(0.37950303615)*x[2]
        arg[2,3]=(0.685209641306)*x[0]+(-0.347743700679)*x[1]+(0.356023331014)*x[2]
        arg[2,4]=(-0.232082602587)*x[0]+(-0.942200667607)*x[1]+(0.961251099323)*x[2]
        arg[3,0]=(0.59399111477)*x[0]+(1.06884076218)*x[1]+(-0.194044225109)*x[2]
        arg[3,1]=(0.522006012644)*x[0]+(0.0804272287843)*x[1]+(0.263440832249)*x[2]
        arg[3,2]=(1.25330346872)*x[0]+(-0.757485536992)*x[1]+(0.398906137626)*x[2]
        arg[3,3]=(0.395712858673)*x[0]+(0.423226476819)*x[1]+(0.186209217392)*x[2]
        arg[3,4]=(0.153657073245)*x[0]+(0.536681875397)*x[1]+(0.773972080743)*x[2]
        ref[0,0]=(0.0539484604772)*x_ref[0]+(-0.220817236218)*x_ref[1]+(-0.853759265774)*x_ref[2]
        ref[0,1]=(0.336027623724)*x_ref[0]+(-0.111775563754)*x_ref[1]+(0.646672879096)*x_ref[2]
        ref[0,2]=(0.650722528854)*x_ref[0]+(0.0924672090652)*x_ref[1]+(-0.881798566157)*x_ref[2]
        ref[0,3]=(0.276745374112)*x_ref[0]+(-1.34491685419)*x_ref[1]+(-0.833420124328)*x_ref[2]
        ref[0,4]=(1.50686931219)*x_ref[0]+(0.489530477585)*x_ref[1]+(1.74836249628)*x_ref[2]
        ref[1,0]=(-0.240001512118)*x_ref[0]+(0.350823736593)*x_ref[1]+(-0.227106949276)*x_ref[2]
        ref[1,1]=(0.354636302436)*x_ref[0]+(-0.0893490250257)*x_ref[1]+(-0.0650366762176)*x_ref[2]
        ref[1,2]=(-0.319714829694)*x_ref[0]+(0.43056121683)*x_ref[1]+(0.109856817034)*x_ref[2]
        ref[1,3]=(-1.17581255544)*x_ref[0]+(-0.491431894533)*x_ref[1]+(0.220857799733)*x_ref[2]
        ref[1,4]=(1.5680027633)*x_ref[0]+(-0.249723385268)*x_ref[1]+(-1.72056978299)*x_ref[2]
        ref[2,0]=(0.197753675106)*x_ref[0]+(0.314723848519)*x_ref[1]+(0.718038926838)*x_ref[2]
        ref[2,1]=(0.13916867566)*x_ref[0]+(-0.0345156812431)*x_ref[1]+(0.34208613933)*x_ref[2]
        ref[2,2]=(-0.275775747114)*x_ref[0]+(0.299200648073)*x_ref[1]+(0.37950303615)*x_ref[2]
        ref[2,3]=(0.685209641306)*x_ref[0]+(-0.347743700679)*x_ref[1]+(0.356023331014)*x_ref[2]
        ref[2,4]=(-0.232082602587)*x_ref[0]+(-0.942200667607)*x_ref[1]+(0.961251099323)*x_ref[2]
        ref[3,0]=(0.59399111477)*x_ref[0]+(1.06884076218)*x_ref[1]+(-0.194044225109)*x_ref[2]
        ref[3,1]=(0.522006012644)*x_ref[0]+(0.0804272287843)*x_ref[1]+(0.263440832249)*x_ref[2]
        ref[3,2]=(1.25330346872)*x_ref[0]+(-0.757485536992)*x_ref[1]+(0.398906137626)*x_ref[2]
        ref[3,3]=(0.395712858673)*x_ref[0]+(0.423226476819)*x_ref[1]+(0.186209217392)*x_ref[2]
        ref[3,4]=(0.153657073245)*x_ref[0]+(0.536681875397)*x_ref[1]+(0.773972080743)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.223040346102)*x[0]+(-0.40941514813)*x[1]
        arg[0,0,1]=(-0.580051757755)*x[0]+(0.0793034026033)*x[1]
        arg[0,1,0]=(-0.0576028239384)*x[0]+(1.058218319)*x[1]
        arg[0,1,1]=(0.0914458475861)*x[0]+(1.48556548656)*x[1]
        arg[1,0,0]=(0.211508891263)*x[0]+(-1.44088884228)*x[1]
        arg[1,0,1]=(-0.753661543994)*x[0]+(0.675867607628)*x[1]
        arg[1,1,0]=(-0.289500075744)*x[0]+(0.573476965486)*x[1]
        arg[1,1,1]=(0.182150073163)*x[0]+(0.937376164291)*x[1]
        arg[2,0,0]=(1.18815471968)*x[0]+(-0.860110330094)*x[1]
        arg[2,0,1]=(-0.238305806178)*x[0]+(-1.27916609495)*x[1]
        arg[2,1,0]=(-0.955511742125)*x[0]+(-0.605135532767)*x[1]
        arg[2,1,1]=(0.342919038059)*x[0]+(-0.0980330687702)*x[1]
        arg[3,0,0]=(0.700429242018)*x[0]+(0.742017035289)*x[1]
        arg[3,0,1]=(0.344011429773)*x[0]+(-1.64251335376)*x[1]
        arg[3,1,0]=(-0.177856966174)*x[0]+(0.180846653076)*x[1]
        arg[3,1,1]=(-1.59212766197)*x[0]+(0.523595606775)*x[1]
        arg[4,0,0]=(-0.264023123775)*x[0]+(0.545951813489)*x[1]
        arg[4,0,1]=(-1.31580310435)*x[0]+(-0.105998549363)*x[1]
        arg[4,1,0]=(0.878775616997)*x[0]+(0.0436540350692)*x[1]
        arg[4,1,1]=(-1.50886945727)*x[0]+(-0.156133758872)*x[1]
        arg[5,0,0]=(-0.196603805975)*x[0]+(1.3924198011)*x[1]
        arg[5,0,1]=(0.688423025668)*x[0]+(-1.1304740704)*x[1]
        arg[5,1,0]=(1.05132745721)*x[0]+(-0.0190306576753)*x[1]
        arg[5,1,1]=(-0.336157880135)*x[0]+(0.948775404759)*x[1]
        ref[0,0,0]=(0.223040346102)*x_ref[0]+(-0.40941514813)*x_ref[1]
        ref[0,0,1]=(-0.580051757755)*x_ref[0]+(0.0793034026033)*x_ref[1]
        ref[0,1,0]=(-0.0576028239384)*x_ref[0]+(1.058218319)*x_ref[1]
        ref[0,1,1]=(0.0914458475861)*x_ref[0]+(1.48556548656)*x_ref[1]
        ref[1,0,0]=(0.211508891263)*x_ref[0]+(-1.44088884228)*x_ref[1]
        ref[1,0,1]=(-0.753661543994)*x_ref[0]+(0.675867607628)*x_ref[1]
        ref[1,1,0]=(-0.289500075744)*x_ref[0]+(0.573476965486)*x_ref[1]
        ref[1,1,1]=(0.182150073163)*x_ref[0]+(0.937376164291)*x_ref[1]
        ref[2,0,0]=(1.18815471968)*x_ref[0]+(-0.860110330094)*x_ref[1]
        ref[2,0,1]=(-0.238305806178)*x_ref[0]+(-1.27916609495)*x_ref[1]
        ref[2,1,0]=(-0.955511742125)*x_ref[0]+(-0.605135532767)*x_ref[1]
        ref[2,1,1]=(0.342919038059)*x_ref[0]+(-0.0980330687702)*x_ref[1]
        ref[3,0,0]=(0.700429242018)*x_ref[0]+(0.742017035289)*x_ref[1]
        ref[3,0,1]=(0.344011429773)*x_ref[0]+(-1.64251335376)*x_ref[1]
        ref[3,1,0]=(-0.177856966174)*x_ref[0]+(0.180846653076)*x_ref[1]
        ref[3,1,1]=(-1.59212766197)*x_ref[0]+(0.523595606775)*x_ref[1]
        ref[4,0,0]=(-0.264023123775)*x_ref[0]+(0.545951813489)*x_ref[1]
        ref[4,0,1]=(-1.31580310435)*x_ref[0]+(-0.105998549363)*x_ref[1]
        ref[4,1,0]=(0.878775616997)*x_ref[0]+(0.0436540350692)*x_ref[1]
        ref[4,1,1]=(-1.50886945727)*x_ref[0]+(-0.156133758872)*x_ref[1]
        ref[5,0,0]=(-0.196603805975)*x_ref[0]+(1.3924198011)*x_ref[1]
        ref[5,0,1]=(0.688423025668)*x_ref[0]+(-1.1304740704)*x_ref[1]
        ref[5,1,0]=(1.05132745721)*x_ref[0]+(-0.0190306576753)*x_ref[1]
        ref[5,1,1]=(-0.336157880135)*x_ref[0]+(0.948775404759)*x_ref[1]
      else:
        arg[0,0,0]=(-0.884050876189)*x[0]+(-0.394515582819)*x[1]+(1.3097094442)*x[2]
        arg[0,0,1]=(-0.816123587792)*x[0]+(0.355197775988)*x[1]+(0.276746708309)*x[2]
        arg[0,1,0]=(-0.124049709168)*x[0]+(1.18541954468)*x[1]+(0.209504045468)*x[2]
        arg[0,1,1]=(0.0683100149522)*x[0]+(0.362926075694)*x[1]+(-0.0669691928299)*x[2]
        arg[1,0,0]=(-1.1421316954)*x[0]+(0.390740725692)*x[1]+(-0.726193980086)*x[2]
        arg[1,0,1]=(-0.843848580922)*x[0]+(-0.525850215762)*x[1]+(-0.00216507276212)*x[2]
        arg[1,1,0]=(1.48370169071)*x[0]+(1.0755925183)*x[1]+(0.581274760413)*x[2]
        arg[1,1,1]=(-1.32186510041)*x[0]+(0.415895688227)*x[1]+(0.0211839725586)*x[2]
        arg[2,0,0]=(0.064488762235)*x[0]+(1.49328658794)*x[1]+(1.58091496737)*x[2]
        arg[2,0,1]=(0.0366527940907)*x[0]+(0.777734357144)*x[1]+(-1.42974160614)*x[2]
        arg[2,1,0]=(1.55170579579)*x[0]+(-1.93288667897)*x[1]+(-0.293286140809)*x[2]
        arg[2,1,1]=(1.20599946796)*x[0]+(0.0227207800543)*x[1]+(0.754694156356)*x[2]
        arg[3,0,0]=(-0.096426023819)*x[0]+(0.254350817295)*x[1]+(-0.845563802639)*x[2]
        arg[3,0,1]=(0.259201676184)*x[0]+(-0.10777841018)*x[1]+(0.198814801338)*x[2]
        arg[3,1,0]=(-0.375693547297)*x[0]+(-0.566070306507)*x[1]+(0.842662788128)*x[2]
        arg[3,1,1]=(-0.254690046974)*x[0]+(1.68423179439)*x[1]+(-0.514749238215)*x[2]
        arg[4,0,0]=(-0.283580142327)*x[0]+(-0.906119047897)*x[1]+(1.16264752395)*x[2]
        arg[4,0,1]=(0.559319137591)*x[0]+(-0.135216628913)*x[1]+(0.0241520443893)*x[2]
        arg[4,1,0]=(1.51344686911)*x[0]+(0.595733533885)*x[1]+(-0.49014841511)*x[2]
        arg[4,1,1]=(0.261868556117)*x[0]+(-0.51266470705)*x[1]+(0.617731786544)*x[2]
        arg[5,0,0]=(-0.947197547436)*x[0]+(1.18970238499)*x[1]+(-0.0491748623501)*x[2]
        arg[5,0,1]=(0.0541604147191)*x[0]+(-0.844601205099)*x[1]+(1.67419390002)*x[2]
        arg[5,1,0]=(0.370982789151)*x[0]+(-0.344762868862)*x[1]+(-0.48077628478)*x[2]
        arg[5,1,1]=(-0.0845487748754)*x[0]+(1.43148171186)*x[1]+(-0.519880423555)*x[2]
        ref[0,0,0]=(-0.884050876189)*x_ref[0]+(-0.394515582819)*x_ref[1]+(1.3097094442)*x_ref[2]
        ref[0,0,1]=(-0.816123587792)*x_ref[0]+(0.355197775988)*x_ref[1]+(0.276746708309)*x_ref[2]
        ref[0,1,0]=(-0.124049709168)*x_ref[0]+(1.18541954468)*x_ref[1]+(0.209504045468)*x_ref[2]
        ref[0,1,1]=(0.0683100149522)*x_ref[0]+(0.362926075694)*x_ref[1]+(-0.0669691928299)*x_ref[2]
        ref[1,0,0]=(-1.1421316954)*x_ref[0]+(0.390740725692)*x_ref[1]+(-0.726193980086)*x_ref[2]
        ref[1,0,1]=(-0.843848580922)*x_ref[0]+(-0.525850215762)*x_ref[1]+(-0.00216507276212)*x_ref[2]
        ref[1,1,0]=(1.48370169071)*x_ref[0]+(1.0755925183)*x_ref[1]+(0.581274760413)*x_ref[2]
        ref[1,1,1]=(-1.32186510041)*x_ref[0]+(0.415895688227)*x_ref[1]+(0.0211839725586)*x_ref[2]
        ref[2,0,0]=(0.064488762235)*x_ref[0]+(1.49328658794)*x_ref[1]+(1.58091496737)*x_ref[2]
        ref[2,0,1]=(0.0366527940907)*x_ref[0]+(0.777734357144)*x_ref[1]+(-1.42974160614)*x_ref[2]
        ref[2,1,0]=(1.55170579579)*x_ref[0]+(-1.93288667897)*x_ref[1]+(-0.293286140809)*x_ref[2]
        ref[2,1,1]=(1.20599946796)*x_ref[0]+(0.0227207800543)*x_ref[1]+(0.754694156356)*x_ref[2]
        ref[3,0,0]=(-0.096426023819)*x_ref[0]+(0.254350817295)*x_ref[1]+(-0.845563802639)*x_ref[2]
        ref[3,0,1]=(0.259201676184)*x_ref[0]+(-0.10777841018)*x_ref[1]+(0.198814801338)*x_ref[2]
        ref[3,1,0]=(-0.375693547297)*x_ref[0]+(-0.566070306507)*x_ref[1]+(0.842662788128)*x_ref[2]
        ref[3,1,1]=(-0.254690046974)*x_ref[0]+(1.68423179439)*x_ref[1]+(-0.514749238215)*x_ref[2]
        ref[4,0,0]=(-0.283580142327)*x_ref[0]+(-0.906119047897)*x_ref[1]+(1.16264752395)*x_ref[2]
        ref[4,0,1]=(0.559319137591)*x_ref[0]+(-0.135216628913)*x_ref[1]+(0.0241520443893)*x_ref[2]
        ref[4,1,0]=(1.51344686911)*x_ref[0]+(0.595733533885)*x_ref[1]+(-0.49014841511)*x_ref[2]
        ref[4,1,1]=(0.261868556117)*x_ref[0]+(-0.51266470705)*x_ref[1]+(0.617731786544)*x_ref[2]
        ref[5,0,0]=(-0.947197547436)*x_ref[0]+(1.18970238499)*x_ref[1]+(-0.0491748623501)*x_ref[2]
        ref[5,0,1]=(0.0541604147191)*x_ref[0]+(-0.844601205099)*x_ref[1]+(1.67419390002)*x_ref[2]
        ref[5,1,0]=(0.370982789151)*x_ref[0]+(-0.344762868862)*x_ref[1]+(-0.48077628478)*x_ref[2]
        ref[5,1,1]=(-0.0845487748754)*x_ref[0]+(1.43148171186)*x_ref[1]+(-0.519880423555)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.748732264044)*x[0]+(0.326649183672)*x[1]
        arg[0,0,0,1]=(-0.118874184707)*x[0]+(0.512027378857)*x[1]
        arg[0,0,1,0]=(0.0907534655112)*x[0]+(0.145249508199)*x[1]
        arg[0,0,1,1]=(0.879689742122)*x[0]+(-0.238488796672)*x[1]
        arg[0,0,2,0]=(-0.285226080478)*x[0]+(-0.560477167124)*x[1]
        arg[0,0,2,1]=(1.3620160918)*x[0]+(-0.97877396699)*x[1]
        arg[0,1,0,0]=(-1.52439734042)*x[0]+(0.981913225695)*x[1]
        arg[0,1,0,1]=(-1.19963597544)*x[0]+(1.1571861154)*x[1]
        arg[0,1,1,0]=(0.822934040682)*x[0]+(-1.75039812758)*x[1]
        arg[0,1,1,1]=(-0.690374545295)*x[0]+(1.10465673793)*x[1]
        arg[0,1,2,0]=(-0.222743457309)*x[0]+(-0.499809039611)*x[1]
        arg[0,1,2,1]=(0.774614333867)*x[0]+(-0.140097281276)*x[1]
        arg[0,2,0,0]=(0.92399157301)*x[0]+(-1.53184881088)*x[1]
        arg[0,2,0,1]=(0.14612723475)*x[0]+(0.892066346169)*x[1]
        arg[0,2,1,0]=(-0.136826339137)*x[0]+(0.0588253058296)*x[1]
        arg[0,2,1,1]=(-0.73687808603)*x[0]+(0.763659413232)*x[1]
        arg[0,2,2,0]=(0.669031805455)*x[0]+(0.452748434512)*x[1]
        arg[0,2,2,1]=(0.969822865927)*x[0]+(0.84449053089)*x[1]
        arg[0,3,0,0]=(1.05482233425)*x[0]+(0.592437824455)*x[1]
        arg[0,3,0,1]=(0.632482870556)*x[0]+(0.272691402585)*x[1]
        arg[0,3,1,0]=(-0.530755710624)*x[0]+(0.798233454742)*x[1]
        arg[0,3,1,1]=(-0.326644901487)*x[0]+(0.912945575753)*x[1]
        arg[0,3,2,0]=(0.313672017285)*x[0]+(0.948624232644)*x[1]
        arg[0,3,2,1]=(-0.985713086674)*x[0]+(0.557682757171)*x[1]
        arg[0,4,0,0]=(-1.05604811856)*x[0]+(0.953109137724)*x[1]
        arg[0,4,0,1]=(0.443301072651)*x[0]+(0.063796068208)*x[1]
        arg[0,4,1,0]=(-0.0166520333953)*x[0]+(0.808947969176)*x[1]
        arg[0,4,1,1]=(0.447406784488)*x[0]+(0.511727243983)*x[1]
        arg[0,4,2,0]=(0.966474340588)*x[0]+(0.789208895767)*x[1]
        arg[0,4,2,1]=(0.5007322363)*x[0]+(0.1211006989)*x[1]
        arg[1,0,0,0]=(0.328606613916)*x[0]+(-0.874029860851)*x[1]
        arg[1,0,0,1]=(-0.906174034813)*x[0]+(-0.705820652424)*x[1]
        arg[1,0,1,0]=(-0.543917624184)*x[0]+(-0.963384956871)*x[1]
        arg[1,0,1,1]=(-0.516192972362)*x[0]+(0.644782734001)*x[1]
        arg[1,0,2,0]=(-1.39622140542)*x[0]+(0.292606264905)*x[1]
        arg[1,0,2,1]=(1.39306984816)*x[0]+(-0.544853887253)*x[1]
        arg[1,1,0,0]=(-0.0499160669149)*x[0]+(-0.783142164006)*x[1]
        arg[1,1,0,1]=(0.968157339631)*x[0]+(-1.68767943891)*x[1]
        arg[1,1,1,0]=(-1.56616531333)*x[0]+(1.057002431)*x[1]
        arg[1,1,1,1]=(0.759774810848)*x[0]+(-0.217500595483)*x[1]
        arg[1,1,2,0]=(-0.558666811623)*x[0]+(-0.56110405702)*x[1]
        arg[1,1,2,1]=(-1.51823291284)*x[0]+(0.968598028275)*x[1]
        arg[1,2,0,0]=(0.459934430553)*x[0]+(-0.425611928234)*x[1]
        arg[1,2,0,1]=(0.497077152958)*x[0]+(-0.217468584657)*x[1]
        arg[1,2,1,0]=(-0.84050521972)*x[0]+(-0.639967173826)*x[1]
        arg[1,2,1,1]=(1.23982979981)*x[0]+(-0.340962636339)*x[1]
        arg[1,2,2,0]=(-0.0796290414672)*x[0]+(-0.418766465742)*x[1]
        arg[1,2,2,1]=(-0.00357466869904)*x[0]+(0.38318423828)*x[1]
        arg[1,3,0,0]=(-1.06978632139)*x[0]+(-0.293480140024)*x[1]
        arg[1,3,0,1]=(0.315898405933)*x[0]+(-0.63338147672)*x[1]
        arg[1,3,1,0]=(-0.570294815011)*x[0]+(0.0492498556666)*x[1]
        arg[1,3,1,1]=(0.0535633163606)*x[0]+(1.03597167921)*x[1]
        arg[1,3,2,0]=(0.740363487843)*x[0]+(-0.945398692608)*x[1]
        arg[1,3,2,1]=(0.0118916598378)*x[0]+(1.13783979963)*x[1]
        arg[1,4,0,0]=(-0.90818285841)*x[0]+(0.0630620135139)*x[1]
        arg[1,4,0,1]=(1.28854967547)*x[0]+(0.564058166595)*x[1]
        arg[1,4,1,0]=(0.0536837223463)*x[0]+(-0.754361313431)*x[1]
        arg[1,4,1,1]=(-1.4274155179)*x[0]+(-0.0330452395142)*x[1]
        arg[1,4,2,0]=(-0.0853100557203)*x[0]+(-0.0698244726734)*x[1]
        arg[1,4,2,1]=(-0.969990013612)*x[0]+(-0.0928692035314)*x[1]
        arg[2,0,0,0]=(1.402108752)*x[0]+(-0.0601730744903)*x[1]
        arg[2,0,0,1]=(-0.544125618042)*x[0]+(1.02669845397)*x[1]
        arg[2,0,1,0]=(-0.259299705101)*x[0]+(-0.449507095724)*x[1]
        arg[2,0,1,1]=(0.572470083182)*x[0]+(-0.140083185208)*x[1]
        arg[2,0,2,0]=(1.7708938802)*x[0]+(-0.0924975206473)*x[1]
        arg[2,0,2,1]=(-0.262417029616)*x[0]+(-0.169562773872)*x[1]
        arg[2,1,0,0]=(-1.30668657807)*x[0]+(-1.01393009909)*x[1]
        arg[2,1,0,1]=(0.799588711881)*x[0]+(-1.17126698679)*x[1]
        arg[2,1,1,0]=(-1.17423919188)*x[0]+(0.321090068168)*x[1]
        arg[2,1,1,1]=(-0.778912666585)*x[0]+(0.00447436139975)*x[1]
        arg[2,1,2,0]=(0.40622401744)*x[0]+(0.61089755242)*x[1]
        arg[2,1,2,1]=(0.550326994175)*x[0]+(1.61303737224)*x[1]
        arg[2,2,0,0]=(0.931647812671)*x[0]+(-0.93554163705)*x[1]
        arg[2,2,0,1]=(0.281543117633)*x[0]+(-0.520873020458)*x[1]
        arg[2,2,1,0]=(1.46731959732)*x[0]+(0.518798296946)*x[1]
        arg[2,2,1,1]=(-1.18123650343)*x[0]+(-0.145310254643)*x[1]
        arg[2,2,2,0]=(-0.560745819237)*x[0]+(-1.47844975303)*x[1]
        arg[2,2,2,1]=(-1.11775616574)*x[0]+(-1.21514252434)*x[1]
        arg[2,3,0,0]=(0.315369902788)*x[0]+(0.964595236889)*x[1]
        arg[2,3,0,1]=(-0.907229008161)*x[0]+(1.20677953168)*x[1]
        arg[2,3,1,0]=(0.0552545457484)*x[0]+(-0.202468129673)*x[1]
        arg[2,3,1,1]=(-0.832968133648)*x[0]+(0.798887606761)*x[1]
        arg[2,3,2,0]=(1.0330803155)*x[0]+(0.546834262777)*x[1]
        arg[2,3,2,1]=(-1.78895672921)*x[0]+(-1.43034307109)*x[1]
        arg[2,4,0,0]=(0.333456542345)*x[0]+(0.620879982969)*x[1]
        arg[2,4,0,1]=(-0.141328336229)*x[0]+(-0.568586127044)*x[1]
        arg[2,4,1,0]=(0.936638826549)*x[0]+(0.262046172423)*x[1]
        arg[2,4,1,1]=(0.594781145732)*x[0]+(-0.795944581484)*x[1]
        arg[2,4,2,0]=(-0.0594467711107)*x[0]+(1.65280673317)*x[1]
        arg[2,4,2,1]=(-0.295492856013)*x[0]+(0.244222123606)*x[1]
        arg[3,0,0,0]=(0.907207795877)*x[0]+(0.0238682693857)*x[1]
        arg[3,0,0,1]=(0.767877824626)*x[0]+(0.781729746211)*x[1]
        arg[3,0,1,0]=(-0.266760960719)*x[0]+(0.109690121766)*x[1]
        arg[3,0,1,1]=(0.885111787835)*x[0]+(1.83341408518)*x[1]
        arg[3,0,2,0]=(0.482836215622)*x[0]+(0.202385773036)*x[1]
        arg[3,0,2,1]=(-0.722931824523)*x[0]+(0.469906208805)*x[1]
        arg[3,1,0,0]=(0.34246223362)*x[0]+(0.652551099051)*x[1]
        arg[3,1,0,1]=(-1.24987233614)*x[0]+(-0.208386614512)*x[1]
        arg[3,1,1,0]=(-0.426021351048)*x[0]+(-0.481440089744)*x[1]
        arg[3,1,1,1]=(0.402917477117)*x[0]+(0.280066256773)*x[1]
        arg[3,1,2,0]=(0.846917449307)*x[0]+(-0.890613362071)*x[1]
        arg[3,1,2,1]=(0.848971364583)*x[0]+(1.28188639634)*x[1]
        arg[3,2,0,0]=(0.244851870449)*x[0]+(-1.29335472533)*x[1]
        arg[3,2,0,1]=(-0.679420773687)*x[0]+(0.0533049166584)*x[1]
        arg[3,2,1,0]=(0.209683196995)*x[0]+(-1.18725405512)*x[1]
        arg[3,2,1,1]=(1.14374677567)*x[0]+(-0.0360927695675)*x[1]
        arg[3,2,2,0]=(-0.675139110144)*x[0]+(0.203993983732)*x[1]
        arg[3,2,2,1]=(-0.237314875828)*x[0]+(-0.581326039718)*x[1]
        arg[3,3,0,0]=(-0.429825943466)*x[0]+(-0.256971982048)*x[1]
        arg[3,3,0,1]=(0.520863156649)*x[0]+(-0.149770317657)*x[1]
        arg[3,3,1,0]=(0.534493615852)*x[0]+(-0.486279758704)*x[1]
        arg[3,3,1,1]=(1.40465616102)*x[0]+(-0.131521707731)*x[1]
        arg[3,3,2,0]=(0.171604624229)*x[0]+(0.810160978932)*x[1]
        arg[3,3,2,1]=(1.28980934106)*x[0]+(-1.0563749584)*x[1]
        arg[3,4,0,0]=(-0.631680784482)*x[0]+(-1.10885428097)*x[1]
        arg[3,4,0,1]=(-1.02297901709)*x[0]+(1.30007360034)*x[1]
        arg[3,4,1,0]=(0.728467970192)*x[0]+(-0.872220126316)*x[1]
        arg[3,4,1,1]=(1.18596842635)*x[0]+(-0.246158517864)*x[1]
        arg[3,4,2,0]=(0.158818463065)*x[0]+(-0.687328798914)*x[1]
        arg[3,4,2,1]=(-0.267029459109)*x[0]+(0.161790218602)*x[1]
        ref[0,0,0,0]=(0.748732264044)*x_ref[0]+(0.326649183672)*x_ref[1]
        ref[0,0,0,1]=(-0.118874184707)*x_ref[0]+(0.512027378857)*x_ref[1]
        ref[0,0,1,0]=(0.0907534655112)*x_ref[0]+(0.145249508199)*x_ref[1]
        ref[0,0,1,1]=(0.879689742122)*x_ref[0]+(-0.238488796672)*x_ref[1]
        ref[0,0,2,0]=(-0.285226080478)*x_ref[0]+(-0.560477167124)*x_ref[1]
        ref[0,0,2,1]=(1.3620160918)*x_ref[0]+(-0.97877396699)*x_ref[1]
        ref[0,1,0,0]=(-1.52439734042)*x_ref[0]+(0.981913225695)*x_ref[1]
        ref[0,1,0,1]=(-1.19963597544)*x_ref[0]+(1.1571861154)*x_ref[1]
        ref[0,1,1,0]=(0.822934040682)*x_ref[0]+(-1.75039812758)*x_ref[1]
        ref[0,1,1,1]=(-0.690374545295)*x_ref[0]+(1.10465673793)*x_ref[1]
        ref[0,1,2,0]=(-0.222743457309)*x_ref[0]+(-0.499809039611)*x_ref[1]
        ref[0,1,2,1]=(0.774614333867)*x_ref[0]+(-0.140097281276)*x_ref[1]
        ref[0,2,0,0]=(0.92399157301)*x_ref[0]+(-1.53184881088)*x_ref[1]
        ref[0,2,0,1]=(0.14612723475)*x_ref[0]+(0.892066346169)*x_ref[1]
        ref[0,2,1,0]=(-0.136826339137)*x_ref[0]+(0.0588253058296)*x_ref[1]
        ref[0,2,1,1]=(-0.73687808603)*x_ref[0]+(0.763659413232)*x_ref[1]
        ref[0,2,2,0]=(0.669031805455)*x_ref[0]+(0.452748434512)*x_ref[1]
        ref[0,2,2,1]=(0.969822865927)*x_ref[0]+(0.84449053089)*x_ref[1]
        ref[0,3,0,0]=(1.05482233425)*x_ref[0]+(0.592437824455)*x_ref[1]
        ref[0,3,0,1]=(0.632482870556)*x_ref[0]+(0.272691402585)*x_ref[1]
        ref[0,3,1,0]=(-0.530755710624)*x_ref[0]+(0.798233454742)*x_ref[1]
        ref[0,3,1,1]=(-0.326644901487)*x_ref[0]+(0.912945575753)*x_ref[1]
        ref[0,3,2,0]=(0.313672017285)*x_ref[0]+(0.948624232644)*x_ref[1]
        ref[0,3,2,1]=(-0.985713086674)*x_ref[0]+(0.557682757171)*x_ref[1]
        ref[0,4,0,0]=(-1.05604811856)*x_ref[0]+(0.953109137724)*x_ref[1]
        ref[0,4,0,1]=(0.443301072651)*x_ref[0]+(0.063796068208)*x_ref[1]
        ref[0,4,1,0]=(-0.0166520333953)*x_ref[0]+(0.808947969176)*x_ref[1]
        ref[0,4,1,1]=(0.447406784488)*x_ref[0]+(0.511727243983)*x_ref[1]
        ref[0,4,2,0]=(0.966474340588)*x_ref[0]+(0.789208895767)*x_ref[1]
        ref[0,4,2,1]=(0.5007322363)*x_ref[0]+(0.1211006989)*x_ref[1]
        ref[1,0,0,0]=(0.328606613916)*x_ref[0]+(-0.874029860851)*x_ref[1]
        ref[1,0,0,1]=(-0.906174034813)*x_ref[0]+(-0.705820652424)*x_ref[1]
        ref[1,0,1,0]=(-0.543917624184)*x_ref[0]+(-0.963384956871)*x_ref[1]
        ref[1,0,1,1]=(-0.516192972362)*x_ref[0]+(0.644782734001)*x_ref[1]
        ref[1,0,2,0]=(-1.39622140542)*x_ref[0]+(0.292606264905)*x_ref[1]
        ref[1,0,2,1]=(1.39306984816)*x_ref[0]+(-0.544853887253)*x_ref[1]
        ref[1,1,0,0]=(-0.0499160669149)*x_ref[0]+(-0.783142164006)*x_ref[1]
        ref[1,1,0,1]=(0.968157339631)*x_ref[0]+(-1.68767943891)*x_ref[1]
        ref[1,1,1,0]=(-1.56616531333)*x_ref[0]+(1.057002431)*x_ref[1]
        ref[1,1,1,1]=(0.759774810848)*x_ref[0]+(-0.217500595483)*x_ref[1]
        ref[1,1,2,0]=(-0.558666811623)*x_ref[0]+(-0.56110405702)*x_ref[1]
        ref[1,1,2,1]=(-1.51823291284)*x_ref[0]+(0.968598028275)*x_ref[1]
        ref[1,2,0,0]=(0.459934430553)*x_ref[0]+(-0.425611928234)*x_ref[1]
        ref[1,2,0,1]=(0.497077152958)*x_ref[0]+(-0.217468584657)*x_ref[1]
        ref[1,2,1,0]=(-0.84050521972)*x_ref[0]+(-0.639967173826)*x_ref[1]
        ref[1,2,1,1]=(1.23982979981)*x_ref[0]+(-0.340962636339)*x_ref[1]
        ref[1,2,2,0]=(-0.0796290414672)*x_ref[0]+(-0.418766465742)*x_ref[1]
        ref[1,2,2,1]=(-0.00357466869904)*x_ref[0]+(0.38318423828)*x_ref[1]
        ref[1,3,0,0]=(-1.06978632139)*x_ref[0]+(-0.293480140024)*x_ref[1]
        ref[1,3,0,1]=(0.315898405933)*x_ref[0]+(-0.63338147672)*x_ref[1]
        ref[1,3,1,0]=(-0.570294815011)*x_ref[0]+(0.0492498556666)*x_ref[1]
        ref[1,3,1,1]=(0.0535633163606)*x_ref[0]+(1.03597167921)*x_ref[1]
        ref[1,3,2,0]=(0.740363487843)*x_ref[0]+(-0.945398692608)*x_ref[1]
        ref[1,3,2,1]=(0.0118916598378)*x_ref[0]+(1.13783979963)*x_ref[1]
        ref[1,4,0,0]=(-0.90818285841)*x_ref[0]+(0.0630620135139)*x_ref[1]
        ref[1,4,0,1]=(1.28854967547)*x_ref[0]+(0.564058166595)*x_ref[1]
        ref[1,4,1,0]=(0.0536837223463)*x_ref[0]+(-0.754361313431)*x_ref[1]
        ref[1,4,1,1]=(-1.4274155179)*x_ref[0]+(-0.0330452395142)*x_ref[1]
        ref[1,4,2,0]=(-0.0853100557203)*x_ref[0]+(-0.0698244726734)*x_ref[1]
        ref[1,4,2,1]=(-0.969990013612)*x_ref[0]+(-0.0928692035314)*x_ref[1]
        ref[2,0,0,0]=(1.402108752)*x_ref[0]+(-0.0601730744903)*x_ref[1]
        ref[2,0,0,1]=(-0.544125618042)*x_ref[0]+(1.02669845397)*x_ref[1]
        ref[2,0,1,0]=(-0.259299705101)*x_ref[0]+(-0.449507095724)*x_ref[1]
        ref[2,0,1,1]=(0.572470083182)*x_ref[0]+(-0.140083185208)*x_ref[1]
        ref[2,0,2,0]=(1.7708938802)*x_ref[0]+(-0.0924975206473)*x_ref[1]
        ref[2,0,2,1]=(-0.262417029616)*x_ref[0]+(-0.169562773872)*x_ref[1]
        ref[2,1,0,0]=(-1.30668657807)*x_ref[0]+(-1.01393009909)*x_ref[1]
        ref[2,1,0,1]=(0.799588711881)*x_ref[0]+(-1.17126698679)*x_ref[1]
        ref[2,1,1,0]=(-1.17423919188)*x_ref[0]+(0.321090068168)*x_ref[1]
        ref[2,1,1,1]=(-0.778912666585)*x_ref[0]+(0.00447436139975)*x_ref[1]
        ref[2,1,2,0]=(0.40622401744)*x_ref[0]+(0.61089755242)*x_ref[1]
        ref[2,1,2,1]=(0.550326994175)*x_ref[0]+(1.61303737224)*x_ref[1]
        ref[2,2,0,0]=(0.931647812671)*x_ref[0]+(-0.93554163705)*x_ref[1]
        ref[2,2,0,1]=(0.281543117633)*x_ref[0]+(-0.520873020458)*x_ref[1]
        ref[2,2,1,0]=(1.46731959732)*x_ref[0]+(0.518798296946)*x_ref[1]
        ref[2,2,1,1]=(-1.18123650343)*x_ref[0]+(-0.145310254643)*x_ref[1]
        ref[2,2,2,0]=(-0.560745819237)*x_ref[0]+(-1.47844975303)*x_ref[1]
        ref[2,2,2,1]=(-1.11775616574)*x_ref[0]+(-1.21514252434)*x_ref[1]
        ref[2,3,0,0]=(0.315369902788)*x_ref[0]+(0.964595236889)*x_ref[1]
        ref[2,3,0,1]=(-0.907229008161)*x_ref[0]+(1.20677953168)*x_ref[1]
        ref[2,3,1,0]=(0.0552545457484)*x_ref[0]+(-0.202468129673)*x_ref[1]
        ref[2,3,1,1]=(-0.832968133648)*x_ref[0]+(0.798887606761)*x_ref[1]
        ref[2,3,2,0]=(1.0330803155)*x_ref[0]+(0.546834262777)*x_ref[1]
        ref[2,3,2,1]=(-1.78895672921)*x_ref[0]+(-1.43034307109)*x_ref[1]
        ref[2,4,0,0]=(0.333456542345)*x_ref[0]+(0.620879982969)*x_ref[1]
        ref[2,4,0,1]=(-0.141328336229)*x_ref[0]+(-0.568586127044)*x_ref[1]
        ref[2,4,1,0]=(0.936638826549)*x_ref[0]+(0.262046172423)*x_ref[1]
        ref[2,4,1,1]=(0.594781145732)*x_ref[0]+(-0.795944581484)*x_ref[1]
        ref[2,4,2,0]=(-0.0594467711107)*x_ref[0]+(1.65280673317)*x_ref[1]
        ref[2,4,2,1]=(-0.295492856013)*x_ref[0]+(0.244222123606)*x_ref[1]
        ref[3,0,0,0]=(0.907207795877)*x_ref[0]+(0.0238682693857)*x_ref[1]
        ref[3,0,0,1]=(0.767877824626)*x_ref[0]+(0.781729746211)*x_ref[1]
        ref[3,0,1,0]=(-0.266760960719)*x_ref[0]+(0.109690121766)*x_ref[1]
        ref[3,0,1,1]=(0.885111787835)*x_ref[0]+(1.83341408518)*x_ref[1]
        ref[3,0,2,0]=(0.482836215622)*x_ref[0]+(0.202385773036)*x_ref[1]
        ref[3,0,2,1]=(-0.722931824523)*x_ref[0]+(0.469906208805)*x_ref[1]
        ref[3,1,0,0]=(0.34246223362)*x_ref[0]+(0.652551099051)*x_ref[1]
        ref[3,1,0,1]=(-1.24987233614)*x_ref[0]+(-0.208386614512)*x_ref[1]
        ref[3,1,1,0]=(-0.426021351048)*x_ref[0]+(-0.481440089744)*x_ref[1]
        ref[3,1,1,1]=(0.402917477117)*x_ref[0]+(0.280066256773)*x_ref[1]
        ref[3,1,2,0]=(0.846917449307)*x_ref[0]+(-0.890613362071)*x_ref[1]
        ref[3,1,2,1]=(0.848971364583)*x_ref[0]+(1.28188639634)*x_ref[1]
        ref[3,2,0,0]=(0.244851870449)*x_ref[0]+(-1.29335472533)*x_ref[1]
        ref[3,2,0,1]=(-0.679420773687)*x_ref[0]+(0.0533049166584)*x_ref[1]
        ref[3,2,1,0]=(0.209683196995)*x_ref[0]+(-1.18725405512)*x_ref[1]
        ref[3,2,1,1]=(1.14374677567)*x_ref[0]+(-0.0360927695675)*x_ref[1]
        ref[3,2,2,0]=(-0.675139110144)*x_ref[0]+(0.203993983732)*x_ref[1]
        ref[3,2,2,1]=(-0.237314875828)*x_ref[0]+(-0.581326039718)*x_ref[1]
        ref[3,3,0,0]=(-0.429825943466)*x_ref[0]+(-0.256971982048)*x_ref[1]
        ref[3,3,0,1]=(0.520863156649)*x_ref[0]+(-0.149770317657)*x_ref[1]
        ref[3,3,1,0]=(0.534493615852)*x_ref[0]+(-0.486279758704)*x_ref[1]
        ref[3,3,1,1]=(1.40465616102)*x_ref[0]+(-0.131521707731)*x_ref[1]
        ref[3,3,2,0]=(0.171604624229)*x_ref[0]+(0.810160978932)*x_ref[1]
        ref[3,3,2,1]=(1.28980934106)*x_ref[0]+(-1.0563749584)*x_ref[1]
        ref[3,4,0,0]=(-0.631680784482)*x_ref[0]+(-1.10885428097)*x_ref[1]
        ref[3,4,0,1]=(-1.02297901709)*x_ref[0]+(1.30007360034)*x_ref[1]
        ref[3,4,1,0]=(0.728467970192)*x_ref[0]+(-0.872220126316)*x_ref[1]
        ref[3,4,1,1]=(1.18596842635)*x_ref[0]+(-0.246158517864)*x_ref[1]
        ref[3,4,2,0]=(0.158818463065)*x_ref[0]+(-0.687328798914)*x_ref[1]
        ref[3,4,2,1]=(-0.267029459109)*x_ref[0]+(0.161790218602)*x_ref[1]
      else:
        arg[0,0,0,0]=(-1.41028779384)*x[0]+(-1.69085638593)*x[1]+(-0.626478106748)*x[2]
        arg[0,0,0,1]=(-0.322729407761)*x[0]+(-0.311218478587)*x[1]+(-0.329375635778)*x[2]
        arg[0,0,1,0]=(-0.16111144722)*x[0]+(0.864377600354)*x[1]+(-1.01349863575)*x[2]
        arg[0,0,1,1]=(0.0720111273067)*x[0]+(1.40365232221)*x[1]+(0.292610528868)*x[2]
        arg[0,0,2,0]=(-1.21827641959)*x[0]+(0.168559044212)*x[1]+(0.841389848275)*x[2]
        arg[0,0,2,1]=(0.631258821218)*x[0]+(-1.22185752256)*x[1]+(0.0213133635991)*x[2]
        arg[0,1,0,0]=(-0.55959466492)*x[0]+(0.567105256147)*x[1]+(0.185414627961)*x[2]
        arg[0,1,0,1]=(1.42576865692)*x[0]+(-1.38354595053)*x[1]+(1.08902163491)*x[2]
        arg[0,1,1,0]=(-0.441288277712)*x[0]+(-0.412579680073)*x[1]+(0.85969923606)*x[2]
        arg[0,1,1,1]=(1.20924692667)*x[0]+(-0.263790869016)*x[1]+(-0.117388627799)*x[2]
        arg[0,1,2,0]=(-0.849546239235)*x[0]+(-0.41970601628)*x[1]+(0.712637692553)*x[2]
        arg[0,1,2,1]=(0.665825646772)*x[0]+(-0.49641548077)*x[1]+(0.0243383084422)*x[2]
        arg[0,2,0,0]=(0.158298703766)*x[0]+(-1.38559416875)*x[1]+(-0.549526684962)*x[2]
        arg[0,2,0,1]=(0.290664973202)*x[0]+(-0.725089032773)*x[1]+(-0.420805410829)*x[2]
        arg[0,2,1,0]=(-0.371388631358)*x[0]+(0.712788591622)*x[1]+(0.708499722908)*x[2]
        arg[0,2,1,1]=(-0.209234718871)*x[0]+(-0.559384946846)*x[1]+(-0.18126082565)*x[2]
        arg[0,2,2,0]=(0.295605948328)*x[0]+(0.0787080259554)*x[1]+(-0.413453606631)*x[2]
        arg[0,2,2,1]=(-0.951652076608)*x[0]+(0.412030442608)*x[1]+(-1.0098591628)*x[2]
        arg[0,3,0,0]=(0.38805130758)*x[0]+(0.725062113278)*x[1]+(-0.0789722367715)*x[2]
        arg[0,3,0,1]=(-0.856689544368)*x[0]+(-0.378806165724)*x[1]+(-1.56132477796)*x[2]
        arg[0,3,1,0]=(0.948414643284)*x[0]+(-0.223099011255)*x[1]+(-0.164540510924)*x[2]
        arg[0,3,1,1]=(-0.290272025223)*x[0]+(0.0110236571203)*x[1]+(0.0344804634468)*x[2]
        arg[0,3,2,0]=(-1.06973937948)*x[0]+(-0.19210617162)*x[1]+(-0.0319179806513)*x[2]
        arg[0,3,2,1]=(0.873398122756)*x[0]+(0.87424619659)*x[1]+(0.130681723853)*x[2]
        arg[0,4,0,0]=(-0.231238303513)*x[0]+(1.13248606226)*x[1]+(1.08638890671)*x[2]
        arg[0,4,0,1]=(0.0325659323288)*x[0]+(1.53278397594)*x[1]+(-0.0823455709949)*x[2]
        arg[0,4,1,0]=(-0.415558245155)*x[0]+(-0.468794623975)*x[1]+(-0.4219946425)*x[2]
        arg[0,4,1,1]=(0.687857094934)*x[0]+(-0.316748377318)*x[1]+(1.51841645514)*x[2]
        arg[0,4,2,0]=(-0.842163178579)*x[0]+(-0.459357139258)*x[1]+(-0.774999285967)*x[2]
        arg[0,4,2,1]=(0.361059525968)*x[0]+(0.302800510186)*x[1]+(0.164988681972)*x[2]
        arg[1,0,0,0]=(-0.549884089801)*x[0]+(0.18615759395)*x[1]+(-0.150676638905)*x[2]
        arg[1,0,0,1]=(-0.908800158159)*x[0]+(1.74574577127)*x[1]+(0.347894281217)*x[2]
        arg[1,0,1,0]=(1.21798491919)*x[0]+(-1.04989223901)*x[1]+(-0.706457284537)*x[2]
        arg[1,0,1,1]=(0.441083030576)*x[0]+(0.458319482203)*x[1]+(0.704887998977)*x[2]
        arg[1,0,2,0]=(-0.386479556415)*x[0]+(-1.28347087396)*x[1]+(-0.876058219438)*x[2]
        arg[1,0,2,1]=(0.743100114846)*x[0]+(1.02919050439)*x[1]+(-0.357203265224)*x[2]
        arg[1,1,0,0]=(-0.309112244319)*x[0]+(0.469639650979)*x[1]+(0.371325594811)*x[2]
        arg[1,1,0,1]=(1.53067095144)*x[0]+(1.18235592763)*x[1]+(-0.959436428335)*x[2]
        arg[1,1,1,0]=(0.291735282406)*x[0]+(0.449515270669)*x[1]+(0.612383052683)*x[2]
        arg[1,1,1,1]=(-0.102313390769)*x[0]+(1.25745082506)*x[1]+(-1.03742917135)*x[2]
        arg[1,1,2,0]=(-0.951576197839)*x[0]+(1.03094297086)*x[1]+(-0.10195348802)*x[2]
        arg[1,1,2,1]=(1.54751905175)*x[0]+(1.47958820289)*x[1]+(-0.357466311206)*x[2]
        arg[1,2,0,0]=(-0.452360614767)*x[0]+(-0.865741367113)*x[1]+(1.12034992845)*x[2]
        arg[1,2,0,1]=(0.00940323370511)*x[0]+(-1.03701972012)*x[1]+(0.0102243182131)*x[2]
        arg[1,2,1,0]=(-1.50670950911)*x[0]+(-0.219993644831)*x[1]+(-0.1612757943)*x[2]
        arg[1,2,1,1]=(-0.792871502871)*x[0]+(-0.0163752900095)*x[1]+(1.54206281352)*x[2]
        arg[1,2,2,0]=(1.02928968974)*x[0]+(-0.171516999194)*x[1]+(-0.446145675054)*x[2]
        arg[1,2,2,1]=(-1.51230609746)*x[0]+(0.2034193081)*x[1]+(-0.291776924482)*x[2]
        arg[1,3,0,0]=(-0.0312379262641)*x[0]+(0.722239260943)*x[1]+(-0.279504141505)*x[2]
        arg[1,3,0,1]=(-1.12038082645)*x[0]+(-0.328148046828)*x[1]+(0.825288681731)*x[2]
        arg[1,3,1,0]=(0.547173116241)*x[0]+(0.136609628569)*x[1]+(-0.476006212278)*x[2]
        arg[1,3,1,1]=(0.000854188874487)*x[0]+(0.473989971126)*x[1]+(1.58153857042)*x[2]
        arg[1,3,2,0]=(-0.746628247481)*x[0]+(0.859384716891)*x[1]+(0.618348305951)*x[2]
        arg[1,3,2,1]=(0.919601890657)*x[0]+(0.422457734769)*x[1]+(-1.11235109226)*x[2]
        arg[1,4,0,0]=(0.962587287633)*x[0]+(-1.2526632022)*x[1]+(0.210254126936)*x[2]
        arg[1,4,0,1]=(-0.330257182624)*x[0]+(-0.251562095795)*x[1]+(1.29427728094)*x[2]
        arg[1,4,1,0]=(-0.577555922072)*x[0]+(0.844462157498)*x[1]+(-0.120809700805)*x[2]
        arg[1,4,1,1]=(-0.91963436975)*x[0]+(-0.415146584044)*x[1]+(-0.105600826517)*x[2]
        arg[1,4,2,0]=(-1.28940482584)*x[0]+(-1.55548284063)*x[1]+(-0.835108666511)*x[2]
        arg[1,4,2,1]=(-0.0299379665752)*x[0]+(-0.992000458707)*x[1]+(-0.469600133762)*x[2]
        arg[2,0,0,0]=(0.0107575962552)*x[0]+(-1.06407817799)*x[1]+(-1.84217380223)*x[2]
        arg[2,0,0,1]=(-0.0596138757136)*x[0]+(1.75177421542)*x[1]+(0.131511355216)*x[2]
        arg[2,0,1,0]=(-0.29485191211)*x[0]+(1.55396303732)*x[1]+(0.137822438484)*x[2]
        arg[2,0,1,1]=(0.968093798582)*x[0]+(-0.348515863297)*x[1]+(-0.885253223355)*x[2]
        arg[2,0,2,0]=(0.304703601471)*x[0]+(0.350016020435)*x[1]+(0.0348615188696)*x[2]
        arg[2,0,2,1]=(0.242706426227)*x[0]+(0.00660859017533)*x[1]+(-0.122732284886)*x[2]
        arg[2,1,0,0]=(0.540596184477)*x[0]+(-0.0272014403068)*x[1]+(0.559631760626)*x[2]
        arg[2,1,0,1]=(0.537160103765)*x[0]+(-0.974891403717)*x[1]+(-1.2571032682)*x[2]
        arg[2,1,1,0]=(-0.134244705433)*x[0]+(1.63419015928)*x[1]+(0.0879689992598)*x[2]
        arg[2,1,1,1]=(0.885004305439)*x[0]+(1.63473219215)*x[1]+(-0.714475475618)*x[2]
        arg[2,1,2,0]=(-0.220742006931)*x[0]+(0.298113077571)*x[1]+(0.129338822854)*x[2]
        arg[2,1,2,1]=(0.0696132558998)*x[0]+(1.28044127087)*x[1]+(-0.182025504014)*x[2]
        arg[2,2,0,0]=(-0.108760249345)*x[0]+(-0.727726654762)*x[1]+(1.1810583418)*x[2]
        arg[2,2,0,1]=(1.69464976816)*x[0]+(0.771779692301)*x[1]+(-0.746303209271)*x[2]
        arg[2,2,1,0]=(1.6897653643)*x[0]+(-1.21078732523)*x[1]+(-0.000978666646647)*x[2]
        arg[2,2,1,1]=(1.60773073603)*x[0]+(-0.583968900051)*x[1]+(-0.0511730994465)*x[2]
        arg[2,2,2,0]=(0.506948201854)*x[0]+(-0.00990858555453)*x[1]+(1.06433022033)*x[2]
        arg[2,2,2,1]=(0.449523090508)*x[0]+(-1.61710851818)*x[1]+(0.440720932423)*x[2]
        arg[2,3,0,0]=(0.0167109660325)*x[0]+(-1.08812386645)*x[1]+(-1.22089632364)*x[2]
        arg[2,3,0,1]=(-0.841668282066)*x[0]+(0.636645352356)*x[1]+(-0.537753623856)*x[2]
        arg[2,3,1,0]=(1.79552636749)*x[0]+(0.481662215698)*x[1]+(-0.371782307951)*x[2]
        arg[2,3,1,1]=(-0.747945987059)*x[0]+(0.818378132535)*x[1]+(1.66032057212)*x[2]
        arg[2,3,2,0]=(-0.126253738036)*x[0]+(-0.681619876818)*x[1]+(0.117269467345)*x[2]
        arg[2,3,2,1]=(-0.338521414346)*x[0]+(0.381458529772)*x[1]+(1.00819568542)*x[2]
        arg[2,4,0,0]=(0.0110334310038)*x[0]+(0.229750132795)*x[1]+(-0.900772274961)*x[2]
        arg[2,4,0,1]=(-0.798344038508)*x[0]+(0.0410740025116)*x[1]+(1.46856652632)*x[2]
        arg[2,4,1,0]=(0.193355735165)*x[0]+(-0.539307787984)*x[1]+(-0.375854987378)*x[2]
        arg[2,4,1,1]=(0.568143022174)*x[0]+(-1.61770216281)*x[1]+(0.208204629987)*x[2]
        arg[2,4,2,0]=(1.37290817619)*x[0]+(-0.174324684678)*x[1]+(-0.217538755825)*x[2]
        arg[2,4,2,1]=(-0.483601819019)*x[0]+(1.45019519818)*x[1]+(-0.623519730492)*x[2]
        arg[3,0,0,0]=(-0.793731389518)*x[0]+(-0.154763125819)*x[1]+(1.10863435441)*x[2]
        arg[3,0,0,1]=(0.507036346178)*x[0]+(-0.309380836403)*x[1]+(-0.418720111636)*x[2]
        arg[3,0,1,0]=(-0.777608365831)*x[0]+(0.982848447238)*x[1]+(-0.762479658423)*x[2]
        arg[3,0,1,1]=(-0.73325130954)*x[0]+(-0.253614119609)*x[1]+(-0.366314075481)*x[2]
        arg[3,0,2,0]=(0.0989317940836)*x[0]+(0.0772067160489)*x[1]+(-1.84682997853)*x[2]
        arg[3,0,2,1]=(0.256206660712)*x[0]+(1.2868655538)*x[1]+(-0.522346638322)*x[2]
        arg[3,1,0,0]=(-1.41604178327)*x[0]+(0.662242482319)*x[1]+(1.7526698965)*x[2]
        arg[3,1,0,1]=(-0.442315980785)*x[0]+(0.423309436413)*x[1]+(0.175362864372)*x[2]
        arg[3,1,1,0]=(0.353602081392)*x[0]+(1.55530884146)*x[1]+(-0.115267400964)*x[2]
        arg[3,1,1,1]=(0.1445651493)*x[0]+(0.652413545545)*x[1]+(-1.47341714223)*x[2]
        arg[3,1,2,0]=(0.288044123145)*x[0]+(1.45339107563)*x[1]+(0.0253961147274)*x[2]
        arg[3,1,2,1]=(0.738725097174)*x[0]+(0.556287291563)*x[1]+(0.685592801206)*x[2]
        arg[3,2,0,0]=(-1.6952458849)*x[0]+(-0.450982750203)*x[1]+(-0.176030737746)*x[2]
        arg[3,2,0,1]=(-0.267377959403)*x[0]+(-0.329354216237)*x[1]+(0.368496833805)*x[2]
        arg[3,2,1,0]=(1.00419353409)*x[0]+(1.50507467849)*x[1]+(-0.529113363067)*x[2]
        arg[3,2,1,1]=(-1.69175338104)*x[0]+(-0.101436360248)*x[1]+(0.963918243087)*x[2]
        arg[3,2,2,0]=(0.093465616883)*x[0]+(-0.390155190153)*x[1]+(-0.354824565592)*x[2]
        arg[3,2,2,1]=(-0.414577385531)*x[0]+(-0.665726090904)*x[1]+(-1.39857731593)*x[2]
        arg[3,3,0,0]=(-0.528594178602)*x[0]+(-0.435472295413)*x[1]+(0.397972761993)*x[2]
        arg[3,3,0,1]=(1.24885378376)*x[0]+(1.63537745582)*x[1]+(-0.00703582234262)*x[2]
        arg[3,3,1,0]=(0.63369530476)*x[0]+(0.296615915965)*x[1]+(0.593277434193)*x[2]
        arg[3,3,1,1]=(-0.197615624091)*x[0]+(1.45678045242)*x[1]+(0.571039120854)*x[2]
        arg[3,3,2,0]=(-1.01304527292)*x[0]+(0.00892977300805)*x[1]+(0.284285861002)*x[2]
        arg[3,3,2,1]=(0.647872454128)*x[0]+(-0.373279273659)*x[1]+(1.27785668874)*x[2]
        arg[3,4,0,0]=(0.708994891373)*x[0]+(-0.439675903635)*x[1]+(0.612884939946)*x[2]
        arg[3,4,0,1]=(0.535549893212)*x[0]+(-0.940830832131)*x[1]+(-0.30796510692)*x[2]
        arg[3,4,1,0]=(-1.24560696134)*x[0]+(0.125328014324)*x[1]+(0.168757304331)*x[2]
        arg[3,4,1,1]=(0.604969099581)*x[0]+(0.212563821283)*x[1]+(0.146243255184)*x[2]
        arg[3,4,2,0]=(-0.0402174498351)*x[0]+(0.967028407195)*x[1]+(-0.991740640214)*x[2]
        arg[3,4,2,1]=(0.0572331664331)*x[0]+(-0.836713731055)*x[1]+(0.78195371901)*x[2]
        ref[0,0,0,0]=(-1.41028779384)*x_ref[0]+(-1.69085638593)*x_ref[1]+(-0.626478106748)*x_ref[2]
        ref[0,0,0,1]=(-0.322729407761)*x_ref[0]+(-0.311218478587)*x_ref[1]+(-0.329375635778)*x_ref[2]
        ref[0,0,1,0]=(-0.16111144722)*x_ref[0]+(0.864377600354)*x_ref[1]+(-1.01349863575)*x_ref[2]
        ref[0,0,1,1]=(0.0720111273067)*x_ref[0]+(1.40365232221)*x_ref[1]+(0.292610528868)*x_ref[2]
        ref[0,0,2,0]=(-1.21827641959)*x_ref[0]+(0.168559044212)*x_ref[1]+(0.841389848275)*x_ref[2]
        ref[0,0,2,1]=(0.631258821218)*x_ref[0]+(-1.22185752256)*x_ref[1]+(0.0213133635991)*x_ref[2]
        ref[0,1,0,0]=(-0.55959466492)*x_ref[0]+(0.567105256147)*x_ref[1]+(0.185414627961)*x_ref[2]
        ref[0,1,0,1]=(1.42576865692)*x_ref[0]+(-1.38354595053)*x_ref[1]+(1.08902163491)*x_ref[2]
        ref[0,1,1,0]=(-0.441288277712)*x_ref[0]+(-0.412579680073)*x_ref[1]+(0.85969923606)*x_ref[2]
        ref[0,1,1,1]=(1.20924692667)*x_ref[0]+(-0.263790869016)*x_ref[1]+(-0.117388627799)*x_ref[2]
        ref[0,1,2,0]=(-0.849546239235)*x_ref[0]+(-0.41970601628)*x_ref[1]+(0.712637692553)*x_ref[2]
        ref[0,1,2,1]=(0.665825646772)*x_ref[0]+(-0.49641548077)*x_ref[1]+(0.0243383084422)*x_ref[2]
        ref[0,2,0,0]=(0.158298703766)*x_ref[0]+(-1.38559416875)*x_ref[1]+(-0.549526684962)*x_ref[2]
        ref[0,2,0,1]=(0.290664973202)*x_ref[0]+(-0.725089032773)*x_ref[1]+(-0.420805410829)*x_ref[2]
        ref[0,2,1,0]=(-0.371388631358)*x_ref[0]+(0.712788591622)*x_ref[1]+(0.708499722908)*x_ref[2]
        ref[0,2,1,1]=(-0.209234718871)*x_ref[0]+(-0.559384946846)*x_ref[1]+(-0.18126082565)*x_ref[2]
        ref[0,2,2,0]=(0.295605948328)*x_ref[0]+(0.0787080259554)*x_ref[1]+(-0.413453606631)*x_ref[2]
        ref[0,2,2,1]=(-0.951652076608)*x_ref[0]+(0.412030442608)*x_ref[1]+(-1.0098591628)*x_ref[2]
        ref[0,3,0,0]=(0.38805130758)*x_ref[0]+(0.725062113278)*x_ref[1]+(-0.0789722367715)*x_ref[2]
        ref[0,3,0,1]=(-0.856689544368)*x_ref[0]+(-0.378806165724)*x_ref[1]+(-1.56132477796)*x_ref[2]
        ref[0,3,1,0]=(0.948414643284)*x_ref[0]+(-0.223099011255)*x_ref[1]+(-0.164540510924)*x_ref[2]
        ref[0,3,1,1]=(-0.290272025223)*x_ref[0]+(0.0110236571203)*x_ref[1]+(0.0344804634468)*x_ref[2]
        ref[0,3,2,0]=(-1.06973937948)*x_ref[0]+(-0.19210617162)*x_ref[1]+(-0.0319179806513)*x_ref[2]
        ref[0,3,2,1]=(0.873398122756)*x_ref[0]+(0.87424619659)*x_ref[1]+(0.130681723853)*x_ref[2]
        ref[0,4,0,0]=(-0.231238303513)*x_ref[0]+(1.13248606226)*x_ref[1]+(1.08638890671)*x_ref[2]
        ref[0,4,0,1]=(0.0325659323288)*x_ref[0]+(1.53278397594)*x_ref[1]+(-0.0823455709949)*x_ref[2]
        ref[0,4,1,0]=(-0.415558245155)*x_ref[0]+(-0.468794623975)*x_ref[1]+(-0.4219946425)*x_ref[2]
        ref[0,4,1,1]=(0.687857094934)*x_ref[0]+(-0.316748377318)*x_ref[1]+(1.51841645514)*x_ref[2]
        ref[0,4,2,0]=(-0.842163178579)*x_ref[0]+(-0.459357139258)*x_ref[1]+(-0.774999285967)*x_ref[2]
        ref[0,4,2,1]=(0.361059525968)*x_ref[0]+(0.302800510186)*x_ref[1]+(0.164988681972)*x_ref[2]
        ref[1,0,0,0]=(-0.549884089801)*x_ref[0]+(0.18615759395)*x_ref[1]+(-0.150676638905)*x_ref[2]
        ref[1,0,0,1]=(-0.908800158159)*x_ref[0]+(1.74574577127)*x_ref[1]+(0.347894281217)*x_ref[2]
        ref[1,0,1,0]=(1.21798491919)*x_ref[0]+(-1.04989223901)*x_ref[1]+(-0.706457284537)*x_ref[2]
        ref[1,0,1,1]=(0.441083030576)*x_ref[0]+(0.458319482203)*x_ref[1]+(0.704887998977)*x_ref[2]
        ref[1,0,2,0]=(-0.386479556415)*x_ref[0]+(-1.28347087396)*x_ref[1]+(-0.876058219438)*x_ref[2]
        ref[1,0,2,1]=(0.743100114846)*x_ref[0]+(1.02919050439)*x_ref[1]+(-0.357203265224)*x_ref[2]
        ref[1,1,0,0]=(-0.309112244319)*x_ref[0]+(0.469639650979)*x_ref[1]+(0.371325594811)*x_ref[2]
        ref[1,1,0,1]=(1.53067095144)*x_ref[0]+(1.18235592763)*x_ref[1]+(-0.959436428335)*x_ref[2]
        ref[1,1,1,0]=(0.291735282406)*x_ref[0]+(0.449515270669)*x_ref[1]+(0.612383052683)*x_ref[2]
        ref[1,1,1,1]=(-0.102313390769)*x_ref[0]+(1.25745082506)*x_ref[1]+(-1.03742917135)*x_ref[2]
        ref[1,1,2,0]=(-0.951576197839)*x_ref[0]+(1.03094297086)*x_ref[1]+(-0.10195348802)*x_ref[2]
        ref[1,1,2,1]=(1.54751905175)*x_ref[0]+(1.47958820289)*x_ref[1]+(-0.357466311206)*x_ref[2]
        ref[1,2,0,0]=(-0.452360614767)*x_ref[0]+(-0.865741367113)*x_ref[1]+(1.12034992845)*x_ref[2]
        ref[1,2,0,1]=(0.00940323370511)*x_ref[0]+(-1.03701972012)*x_ref[1]+(0.0102243182131)*x_ref[2]
        ref[1,2,1,0]=(-1.50670950911)*x_ref[0]+(-0.219993644831)*x_ref[1]+(-0.1612757943)*x_ref[2]
        ref[1,2,1,1]=(-0.792871502871)*x_ref[0]+(-0.0163752900095)*x_ref[1]+(1.54206281352)*x_ref[2]
        ref[1,2,2,0]=(1.02928968974)*x_ref[0]+(-0.171516999194)*x_ref[1]+(-0.446145675054)*x_ref[2]
        ref[1,2,2,1]=(-1.51230609746)*x_ref[0]+(0.2034193081)*x_ref[1]+(-0.291776924482)*x_ref[2]
        ref[1,3,0,0]=(-0.0312379262641)*x_ref[0]+(0.722239260943)*x_ref[1]+(-0.279504141505)*x_ref[2]
        ref[1,3,0,1]=(-1.12038082645)*x_ref[0]+(-0.328148046828)*x_ref[1]+(0.825288681731)*x_ref[2]
        ref[1,3,1,0]=(0.547173116241)*x_ref[0]+(0.136609628569)*x_ref[1]+(-0.476006212278)*x_ref[2]
        ref[1,3,1,1]=(0.000854188874487)*x_ref[0]+(0.473989971126)*x_ref[1]+(1.58153857042)*x_ref[2]
        ref[1,3,2,0]=(-0.746628247481)*x_ref[0]+(0.859384716891)*x_ref[1]+(0.618348305951)*x_ref[2]
        ref[1,3,2,1]=(0.919601890657)*x_ref[0]+(0.422457734769)*x_ref[1]+(-1.11235109226)*x_ref[2]
        ref[1,4,0,0]=(0.962587287633)*x_ref[0]+(-1.2526632022)*x_ref[1]+(0.210254126936)*x_ref[2]
        ref[1,4,0,1]=(-0.330257182624)*x_ref[0]+(-0.251562095795)*x_ref[1]+(1.29427728094)*x_ref[2]
        ref[1,4,1,0]=(-0.577555922072)*x_ref[0]+(0.844462157498)*x_ref[1]+(-0.120809700805)*x_ref[2]
        ref[1,4,1,1]=(-0.91963436975)*x_ref[0]+(-0.415146584044)*x_ref[1]+(-0.105600826517)*x_ref[2]
        ref[1,4,2,0]=(-1.28940482584)*x_ref[0]+(-1.55548284063)*x_ref[1]+(-0.835108666511)*x_ref[2]
        ref[1,4,2,1]=(-0.0299379665752)*x_ref[0]+(-0.992000458707)*x_ref[1]+(-0.469600133762)*x_ref[2]
        ref[2,0,0,0]=(0.0107575962552)*x_ref[0]+(-1.06407817799)*x_ref[1]+(-1.84217380223)*x_ref[2]
        ref[2,0,0,1]=(-0.0596138757136)*x_ref[0]+(1.75177421542)*x_ref[1]+(0.131511355216)*x_ref[2]
        ref[2,0,1,0]=(-0.29485191211)*x_ref[0]+(1.55396303732)*x_ref[1]+(0.137822438484)*x_ref[2]
        ref[2,0,1,1]=(0.968093798582)*x_ref[0]+(-0.348515863297)*x_ref[1]+(-0.885253223355)*x_ref[2]
        ref[2,0,2,0]=(0.304703601471)*x_ref[0]+(0.350016020435)*x_ref[1]+(0.0348615188696)*x_ref[2]
        ref[2,0,2,1]=(0.242706426227)*x_ref[0]+(0.00660859017533)*x_ref[1]+(-0.122732284886)*x_ref[2]
        ref[2,1,0,0]=(0.540596184477)*x_ref[0]+(-0.0272014403068)*x_ref[1]+(0.559631760626)*x_ref[2]
        ref[2,1,0,1]=(0.537160103765)*x_ref[0]+(-0.974891403717)*x_ref[1]+(-1.2571032682)*x_ref[2]
        ref[2,1,1,0]=(-0.134244705433)*x_ref[0]+(1.63419015928)*x_ref[1]+(0.0879689992598)*x_ref[2]
        ref[2,1,1,1]=(0.885004305439)*x_ref[0]+(1.63473219215)*x_ref[1]+(-0.714475475618)*x_ref[2]
        ref[2,1,2,0]=(-0.220742006931)*x_ref[0]+(0.298113077571)*x_ref[1]+(0.129338822854)*x_ref[2]
        ref[2,1,2,1]=(0.0696132558998)*x_ref[0]+(1.28044127087)*x_ref[1]+(-0.182025504014)*x_ref[2]
        ref[2,2,0,0]=(-0.108760249345)*x_ref[0]+(-0.727726654762)*x_ref[1]+(1.1810583418)*x_ref[2]
        ref[2,2,0,1]=(1.69464976816)*x_ref[0]+(0.771779692301)*x_ref[1]+(-0.746303209271)*x_ref[2]
        ref[2,2,1,0]=(1.6897653643)*x_ref[0]+(-1.21078732523)*x_ref[1]+(-0.000978666646647)*x_ref[2]
        ref[2,2,1,1]=(1.60773073603)*x_ref[0]+(-0.583968900051)*x_ref[1]+(-0.0511730994465)*x_ref[2]
        ref[2,2,2,0]=(0.506948201854)*x_ref[0]+(-0.00990858555453)*x_ref[1]+(1.06433022033)*x_ref[2]
        ref[2,2,2,1]=(0.449523090508)*x_ref[0]+(-1.61710851818)*x_ref[1]+(0.440720932423)*x_ref[2]
        ref[2,3,0,0]=(0.0167109660325)*x_ref[0]+(-1.08812386645)*x_ref[1]+(-1.22089632364)*x_ref[2]
        ref[2,3,0,1]=(-0.841668282066)*x_ref[0]+(0.636645352356)*x_ref[1]+(-0.537753623856)*x_ref[2]
        ref[2,3,1,0]=(1.79552636749)*x_ref[0]+(0.481662215698)*x_ref[1]+(-0.371782307951)*x_ref[2]
        ref[2,3,1,1]=(-0.747945987059)*x_ref[0]+(0.818378132535)*x_ref[1]+(1.66032057212)*x_ref[2]
        ref[2,3,2,0]=(-0.126253738036)*x_ref[0]+(-0.681619876818)*x_ref[1]+(0.117269467345)*x_ref[2]
        ref[2,3,2,1]=(-0.338521414346)*x_ref[0]+(0.381458529772)*x_ref[1]+(1.00819568542)*x_ref[2]
        ref[2,4,0,0]=(0.0110334310038)*x_ref[0]+(0.229750132795)*x_ref[1]+(-0.900772274961)*x_ref[2]
        ref[2,4,0,1]=(-0.798344038508)*x_ref[0]+(0.0410740025116)*x_ref[1]+(1.46856652632)*x_ref[2]
        ref[2,4,1,0]=(0.193355735165)*x_ref[0]+(-0.539307787984)*x_ref[1]+(-0.375854987378)*x_ref[2]
        ref[2,4,1,1]=(0.568143022174)*x_ref[0]+(-1.61770216281)*x_ref[1]+(0.208204629987)*x_ref[2]
        ref[2,4,2,0]=(1.37290817619)*x_ref[0]+(-0.174324684678)*x_ref[1]+(-0.217538755825)*x_ref[2]
        ref[2,4,2,1]=(-0.483601819019)*x_ref[0]+(1.45019519818)*x_ref[1]+(-0.623519730492)*x_ref[2]
        ref[3,0,0,0]=(-0.793731389518)*x_ref[0]+(-0.154763125819)*x_ref[1]+(1.10863435441)*x_ref[2]
        ref[3,0,0,1]=(0.507036346178)*x_ref[0]+(-0.309380836403)*x_ref[1]+(-0.418720111636)*x_ref[2]
        ref[3,0,1,0]=(-0.777608365831)*x_ref[0]+(0.982848447238)*x_ref[1]+(-0.762479658423)*x_ref[2]
        ref[3,0,1,1]=(-0.73325130954)*x_ref[0]+(-0.253614119609)*x_ref[1]+(-0.366314075481)*x_ref[2]
        ref[3,0,2,0]=(0.0989317940836)*x_ref[0]+(0.0772067160489)*x_ref[1]+(-1.84682997853)*x_ref[2]
        ref[3,0,2,1]=(0.256206660712)*x_ref[0]+(1.2868655538)*x_ref[1]+(-0.522346638322)*x_ref[2]
        ref[3,1,0,0]=(-1.41604178327)*x_ref[0]+(0.662242482319)*x_ref[1]+(1.7526698965)*x_ref[2]
        ref[3,1,0,1]=(-0.442315980785)*x_ref[0]+(0.423309436413)*x_ref[1]+(0.175362864372)*x_ref[2]
        ref[3,1,1,0]=(0.353602081392)*x_ref[0]+(1.55530884146)*x_ref[1]+(-0.115267400964)*x_ref[2]
        ref[3,1,1,1]=(0.1445651493)*x_ref[0]+(0.652413545545)*x_ref[1]+(-1.47341714223)*x_ref[2]
        ref[3,1,2,0]=(0.288044123145)*x_ref[0]+(1.45339107563)*x_ref[1]+(0.0253961147274)*x_ref[2]
        ref[3,1,2,1]=(0.738725097174)*x_ref[0]+(0.556287291563)*x_ref[1]+(0.685592801206)*x_ref[2]
        ref[3,2,0,0]=(-1.6952458849)*x_ref[0]+(-0.450982750203)*x_ref[1]+(-0.176030737746)*x_ref[2]
        ref[3,2,0,1]=(-0.267377959403)*x_ref[0]+(-0.329354216237)*x_ref[1]+(0.368496833805)*x_ref[2]
        ref[3,2,1,0]=(1.00419353409)*x_ref[0]+(1.50507467849)*x_ref[1]+(-0.529113363067)*x_ref[2]
        ref[3,2,1,1]=(-1.69175338104)*x_ref[0]+(-0.101436360248)*x_ref[1]+(0.963918243087)*x_ref[2]
        ref[3,2,2,0]=(0.093465616883)*x_ref[0]+(-0.390155190153)*x_ref[1]+(-0.354824565592)*x_ref[2]
        ref[3,2,2,1]=(-0.414577385531)*x_ref[0]+(-0.665726090904)*x_ref[1]+(-1.39857731593)*x_ref[2]
        ref[3,3,0,0]=(-0.528594178602)*x_ref[0]+(-0.435472295413)*x_ref[1]+(0.397972761993)*x_ref[2]
        ref[3,3,0,1]=(1.24885378376)*x_ref[0]+(1.63537745582)*x_ref[1]+(-0.00703582234262)*x_ref[2]
        ref[3,3,1,0]=(0.63369530476)*x_ref[0]+(0.296615915965)*x_ref[1]+(0.593277434193)*x_ref[2]
        ref[3,3,1,1]=(-0.197615624091)*x_ref[0]+(1.45678045242)*x_ref[1]+(0.571039120854)*x_ref[2]
        ref[3,3,2,0]=(-1.01304527292)*x_ref[0]+(0.00892977300805)*x_ref[1]+(0.284285861002)*x_ref[2]
        ref[3,3,2,1]=(0.647872454128)*x_ref[0]+(-0.373279273659)*x_ref[1]+(1.27785668874)*x_ref[2]
        ref[3,4,0,0]=(0.708994891373)*x_ref[0]+(-0.439675903635)*x_ref[1]+(0.612884939946)*x_ref[2]
        ref[3,4,0,1]=(0.535549893212)*x_ref[0]+(-0.940830832131)*x_ref[1]+(-0.30796510692)*x_ref[2]
        ref[3,4,1,0]=(-1.24560696134)*x_ref[0]+(0.125328014324)*x_ref[1]+(0.168757304331)*x_ref[2]
        ref[3,4,1,1]=(0.604969099581)*x_ref[0]+(0.212563821283)*x_ref[1]+(0.146243255184)*x_ref[2]
        ref[3,4,2,0]=(-0.0402174498351)*x_ref[0]+(0.967028407195)*x_ref[1]+(-0.991740640214)*x_ref[2]
        ref[3,4,2,1]=(0.0572331664331)*x_ref[0]+(-0.836713731055)*x_ref[1]+(0.78195371901)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.70941034293)*x[0]**o+(-0.988234004949)*x[0]+(0.440342659022)*x[1]**o+(0.221444369839)*x[1]
        ref=(0.70941034293)*x_ref[0]**o+(-0.988234004949)*x_ref[0]+(0.440342659022)*x_ref[1]**o+(0.221444369839)*x_ref[1]
      else:
        arg=(0.720798141776)*x[0]**o+(0.841475017012)*x[0]+(-0.76285313232)*x[1]**o+(-0.214354625268)*x[1]+(-0.887218533751)*x[2]**o+(-0.066368425118)*x[2]
        ref=(0.720798141776)*x_ref[0]**o+(0.841475017012)*x_ref[0]+(-0.76285313232)*x_ref[1]**o+(-0.214354625268)*x_ref[1]+(-0.887218533751)*x_ref[2]**o+(-0.066368425118)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.940260123916)*x[0]**o+(-0.977192570293)*x[0]+(-0.0910686020625)*x[1]**o+(0.464205016172)*x[1]
        arg[1]=(-0.634146043992)*x[0]**o+(0.00438663094072)*x[0]+(0.905996907036)*x[1]**o+(0.175291641923)*x[1]
        ref[0]=(-0.940260123916)*x_ref[0]**o+(-0.977192570293)*x_ref[0]+(-0.0910686020625)*x_ref[1]**o+(0.464205016172)*x_ref[1]
        ref[1]=(-0.634146043992)*x_ref[0]**o+(0.00438663094072)*x_ref[0]+(0.905996907036)*x_ref[1]**o+(0.175291641923)*x_ref[1]
      else:
        arg[0]=(0.0746302635373)*x[0]**o+(0.611568121627)*x[0]+(0.0392401043636)*x[1]**o+(0.846339953866)*x[1]+(0.775935477701)*x[2]**o+(0.309530355508)*x[2]
        arg[1]=(0.790143430408)*x[0]**o+(0.987665779965)*x[0]+(0.569313447766)*x[1]**o+(0.170690891282)*x[1]+(0.224015411231)*x[2]**o+(-0.670946134012)*x[2]
        ref[0]=(0.0746302635373)*x_ref[0]**o+(0.611568121627)*x_ref[0]+(0.0392401043636)*x_ref[1]**o+(0.846339953866)*x_ref[1]+(0.775935477701)*x_ref[2]**o+(0.309530355508)*x_ref[2]
        ref[1]=(0.790143430408)*x_ref[0]**o+(0.987665779965)*x_ref[0]+(0.569313447766)*x_ref[1]**o+(0.170690891282)*x_ref[1]+(0.224015411231)*x_ref[2]**o+(-0.670946134012)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.339464191741)*x[0]**o+(0.898634031554)*x[0]+(-0.67217412133)*x[1]**o+(-0.732882671617)*x[1]
        arg[0,1]=(0.475031035618)*x[0]**o+(-0.712271710437)*x[0]+(-0.346485473024)*x[1]**o+(0.936873073538)*x[1]
        arg[0,2]=(0.139217050193)*x[0]**o+(0.532762047045)*x[0]+(-0.411674842781)*x[1]**o+(-0.423096144938)*x[1]
        arg[0,3]=(-0.146193322436)*x[0]**o+(-0.227693594412)*x[0]+(0.74715898621)*x[1]**o+(0.538402714773)*x[1]
        arg[0,4]=(0.233316268084)*x[0]**o+(-0.0155557244918)*x[0]+(0.588017790807)*x[1]**o+(-0.58604454457)*x[1]
        arg[1,0]=(0.30473687856)*x[0]**o+(0.933154007362)*x[0]+(-0.0797692178609)*x[1]**o+(0.270021035922)*x[1]
        arg[1,1]=(-0.495613786111)*x[0]**o+(0.0857849462525)*x[0]+(0.330048664047)*x[1]**o+(0.0442768259753)*x[1]
        arg[1,2]=(0.203830377049)*x[0]**o+(-0.891094217974)*x[0]+(-0.643947141174)*x[1]**o+(-0.564315919126)*x[1]
        arg[1,3]=(-0.783514920199)*x[0]**o+(0.356882300706)*x[0]+(-0.307634481801)*x[1]**o+(0.521370775315)*x[1]
        arg[1,4]=(0.418545793821)*x[0]**o+(0.574742804708)*x[0]+(-0.340992300811)*x[1]**o+(-0.892118088815)*x[1]
        arg[2,0]=(-0.361915184269)*x[0]**o+(-0.834492406999)*x[0]+(0.750344593844)*x[1]**o+(-0.516981733186)*x[1]
        arg[2,1]=(-0.0123811923638)*x[0]**o+(-0.9808055138)*x[0]+(-0.176978785352)*x[1]**o+(0.519804799089)*x[1]
        arg[2,2]=(-0.901764228726)*x[0]**o+(-0.610547400097)*x[0]+(0.384870544403)*x[1]**o+(0.0308096944571)*x[1]
        arg[2,3]=(-0.163626872732)*x[0]**o+(0.278523098981)*x[0]+(-0.0544994735633)*x[1]**o+(-0.319354318824)*x[1]
        arg[2,4]=(0.558959525812)*x[0]**o+(0.435157723929)*x[0]+(0.758922109245)*x[1]**o+(-0.702427033542)*x[1]
        arg[3,0]=(0.832800430277)*x[0]**o+(-0.709285644484)*x[0]+(-0.201572915312)*x[1]**o+(0.398294738552)*x[1]
        arg[3,1]=(-0.836258924261)*x[0]**o+(-0.461452427039)*x[0]+(0.314485542425)*x[1]**o+(0.860807479051)*x[1]
        arg[3,2]=(-0.569661610862)*x[0]**o+(-0.46125190469)*x[0]+(0.807311492846)*x[1]**o+(0.72312107176)*x[1]
        arg[3,3]=(0.842126813179)*x[0]**o+(-0.714991954637)*x[0]+(-0.853100178283)*x[1]**o+(-0.644363753291)*x[1]
        arg[3,4]=(-0.0480369646571)*x[0]**o+(-0.330580215997)*x[0]+(0.609757645513)*x[1]**o+(0.915764017541)*x[1]
        ref[0,0]=(0.339464191741)*x_ref[0]**o+(0.898634031554)*x_ref[0]+(-0.67217412133)*x_ref[1]**o+(-0.732882671617)*x_ref[1]
        ref[0,1]=(0.475031035618)*x_ref[0]**o+(-0.712271710437)*x_ref[0]+(-0.346485473024)*x_ref[1]**o+(0.936873073538)*x_ref[1]
        ref[0,2]=(0.139217050193)*x_ref[0]**o+(0.532762047045)*x_ref[0]+(-0.411674842781)*x_ref[1]**o+(-0.423096144938)*x_ref[1]
        ref[0,3]=(-0.146193322436)*x_ref[0]**o+(-0.227693594412)*x_ref[0]+(0.74715898621)*x_ref[1]**o+(0.538402714773)*x_ref[1]
        ref[0,4]=(0.233316268084)*x_ref[0]**o+(-0.0155557244918)*x_ref[0]+(0.588017790807)*x_ref[1]**o+(-0.58604454457)*x_ref[1]
        ref[1,0]=(0.30473687856)*x_ref[0]**o+(0.933154007362)*x_ref[0]+(-0.0797692178609)*x_ref[1]**o+(0.270021035922)*x_ref[1]
        ref[1,1]=(-0.495613786111)*x_ref[0]**o+(0.0857849462525)*x_ref[0]+(0.330048664047)*x_ref[1]**o+(0.0442768259753)*x_ref[1]
        ref[1,2]=(0.203830377049)*x_ref[0]**o+(-0.891094217974)*x_ref[0]+(-0.643947141174)*x_ref[1]**o+(-0.564315919126)*x_ref[1]
        ref[1,3]=(-0.783514920199)*x_ref[0]**o+(0.356882300706)*x_ref[0]+(-0.307634481801)*x_ref[1]**o+(0.521370775315)*x_ref[1]
        ref[1,4]=(0.418545793821)*x_ref[0]**o+(0.574742804708)*x_ref[0]+(-0.340992300811)*x_ref[1]**o+(-0.892118088815)*x_ref[1]
        ref[2,0]=(-0.361915184269)*x_ref[0]**o+(-0.834492406999)*x_ref[0]+(0.750344593844)*x_ref[1]**o+(-0.516981733186)*x_ref[1]
        ref[2,1]=(-0.0123811923638)*x_ref[0]**o+(-0.9808055138)*x_ref[0]+(-0.176978785352)*x_ref[1]**o+(0.519804799089)*x_ref[1]
        ref[2,2]=(-0.901764228726)*x_ref[0]**o+(-0.610547400097)*x_ref[0]+(0.384870544403)*x_ref[1]**o+(0.0308096944571)*x_ref[1]
        ref[2,3]=(-0.163626872732)*x_ref[0]**o+(0.278523098981)*x_ref[0]+(-0.0544994735633)*x_ref[1]**o+(-0.319354318824)*x_ref[1]
        ref[2,4]=(0.558959525812)*x_ref[0]**o+(0.435157723929)*x_ref[0]+(0.758922109245)*x_ref[1]**o+(-0.702427033542)*x_ref[1]
        ref[3,0]=(0.832800430277)*x_ref[0]**o+(-0.709285644484)*x_ref[0]+(-0.201572915312)*x_ref[1]**o+(0.398294738552)*x_ref[1]
        ref[3,1]=(-0.836258924261)*x_ref[0]**o+(-0.461452427039)*x_ref[0]+(0.314485542425)*x_ref[1]**o+(0.860807479051)*x_ref[1]
        ref[3,2]=(-0.569661610862)*x_ref[0]**o+(-0.46125190469)*x_ref[0]+(0.807311492846)*x_ref[1]**o+(0.72312107176)*x_ref[1]
        ref[3,3]=(0.842126813179)*x_ref[0]**o+(-0.714991954637)*x_ref[0]+(-0.853100178283)*x_ref[1]**o+(-0.644363753291)*x_ref[1]
        ref[3,4]=(-0.0480369646571)*x_ref[0]**o+(-0.330580215997)*x_ref[0]+(0.609757645513)*x_ref[1]**o+(0.915764017541)*x_ref[1]
      else:
        arg[0,0]=(-0.256088293513)*x[0]**o+(0.980138068946)*x[0]+(0.375373123738)*x[1]**o+(0.118268751779)*x[1]+(-0.848800006681)*x[2]**o+(0.861808952535)*x[2]
        arg[0,1]=(-0.0658163161725)*x[0]**o+(-0.21582867866)*x[0]+(0.99787757796)*x[1]**o+(0.0175357022646)*x[1]+(-0.305438242469)*x[2]**o+(0.114385992332)*x[2]
        arg[0,2]=(-0.296948302927)*x[0]**o+(0.146852900649)*x[0]+(0.386910961544)*x[1]**o+(0.353651457394)*x[1]+(0.420964895802)*x[2]**o+(0.71145244268)*x[2]
        arg[0,3]=(0.61102494979)*x[0]**o+(-0.316055853885)*x[0]+(-0.117200390837)*x[1]**o+(-0.808630441349)*x[1]+(0.0914828526025)*x[2]**o+(0.305837943839)*x[2]
        arg[0,4]=(-0.22863340097)*x[0]**o+(0.398083421617)*x[0]+(-0.39971917463)*x[1]**o+(-0.0307312416932)*x[1]+(0.510443863383)*x[2]**o+(-0.129495846726)*x[2]
        arg[1,0]=(0.423968874298)*x[0]**o+(0.371480448801)*x[0]+(-0.934565508125)*x[1]**o+(0.890972208341)*x[1]+(-0.169067233411)*x[2]**o+(-0.757264272484)*x[2]
        arg[1,1]=(0.999152573666)*x[0]**o+(-0.659676649566)*x[0]+(0.993153643567)*x[1]**o+(0.350141892848)*x[1]+(-0.493784056536)*x[2]**o+(-0.966494205893)*x[2]
        arg[1,2]=(-0.141337889298)*x[0]**o+(0.689935513838)*x[0]+(0.335444607712)*x[1]**o+(-0.733351283662)*x[1]+(0.323775887172)*x[2]**o+(0.314099156544)*x[2]
        arg[1,3]=(-0.157132412118)*x[0]**o+(0.602468153313)*x[0]+(-0.618134886591)*x[1]**o+(-0.404425578444)*x[1]+(-0.5069585924)*x[2]**o+(0.616829810596)*x[2]
        arg[1,4]=(-0.107902610028)*x[0]**o+(0.664787826515)*x[0]+(-0.19940280262)*x[1]**o+(0.461794954077)*x[1]+(-0.0516671817578)*x[2]**o+(0.426165289622)*x[2]
        arg[2,0]=(-0.568061237616)*x[0]**o+(-0.744340542576)*x[0]+(-0.0125924903194)*x[1]**o+(0.710153866172)*x[1]+(0.979871553596)*x[2]**o+(0.945862455036)*x[2]
        arg[2,1]=(0.201599120538)*x[0]**o+(-0.11057289344)*x[0]+(0.522031147141)*x[1]**o+(0.194694594101)*x[1]+(-0.771997994079)*x[2]**o+(-0.513574957413)*x[2]
        arg[2,2]=(0.401979025382)*x[0]**o+(-0.743641669014)*x[0]+(-0.796071557874)*x[1]**o+(-0.352651095657)*x[1]+(-0.564117536877)*x[2]**o+(0.952729715875)*x[2]
        arg[2,3]=(-0.526162893019)*x[0]**o+(0.691019471759)*x[0]+(-0.68646084088)*x[1]**o+(0.839630438544)*x[1]+(-0.491997512287)*x[2]**o+(-0.405069600537)*x[2]
        arg[2,4]=(0.117664099703)*x[0]**o+(-0.807496546596)*x[0]+(0.607810703702)*x[1]**o+(-0.696118154009)*x[1]+(-0.156011683803)*x[2]**o+(-0.26100420077)*x[2]
        arg[3,0]=(0.280066174527)*x[0]**o+(-0.225940135983)*x[0]+(-0.338716163489)*x[1]**o+(0.628405669096)*x[1]+(-0.890543405425)*x[2]**o+(-0.142082524497)*x[2]
        arg[3,1]=(-0.659241351226)*x[0]**o+(0.275483386126)*x[0]+(-0.318352239058)*x[1]**o+(0.168536648793)*x[1]+(0.539555610218)*x[2]**o+(-0.95280560435)*x[2]
        arg[3,2]=(0.673861565841)*x[0]**o+(-0.0303873544723)*x[0]+(-0.638241094445)*x[1]**o+(0.664488908579)*x[1]+(-0.789484082135)*x[2]**o+(-0.982437408359)*x[2]
        arg[3,3]=(-0.902346309293)*x[0]**o+(0.818106160371)*x[0]+(-0.410074078135)*x[1]**o+(0.977554592344)*x[1]+(-0.00195907247124)*x[2]**o+(0.94762151647)*x[2]
        arg[3,4]=(0.27272813538)*x[0]**o+(0.0563423990113)*x[0]+(0.254612551508)*x[1]**o+(-0.901948547799)*x[1]+(0.263613075865)*x[2]**o+(0.879691314116)*x[2]
        ref[0,0]=(-0.256088293513)*x_ref[0]**o+(0.980138068946)*x_ref[0]+(0.375373123738)*x_ref[1]**o+(0.118268751779)*x_ref[1]+(-0.848800006681)*x_ref[2]**o+(0.861808952535)*x_ref[2]
        ref[0,1]=(-0.0658163161725)*x_ref[0]**o+(-0.21582867866)*x_ref[0]+(0.99787757796)*x_ref[1]**o+(0.0175357022646)*x_ref[1]+(-0.305438242469)*x_ref[2]**o+(0.114385992332)*x_ref[2]
        ref[0,2]=(-0.296948302927)*x_ref[0]**o+(0.146852900649)*x_ref[0]+(0.386910961544)*x_ref[1]**o+(0.353651457394)*x_ref[1]+(0.420964895802)*x_ref[2]**o+(0.71145244268)*x_ref[2]
        ref[0,3]=(0.61102494979)*x_ref[0]**o+(-0.316055853885)*x_ref[0]+(-0.117200390837)*x_ref[1]**o+(-0.808630441349)*x_ref[1]+(0.0914828526025)*x_ref[2]**o+(0.305837943839)*x_ref[2]
        ref[0,4]=(-0.22863340097)*x_ref[0]**o+(0.398083421617)*x_ref[0]+(-0.39971917463)*x_ref[1]**o+(-0.0307312416932)*x_ref[1]+(0.510443863383)*x_ref[2]**o+(-0.129495846726)*x_ref[2]
        ref[1,0]=(0.423968874298)*x_ref[0]**o+(0.371480448801)*x_ref[0]+(-0.934565508125)*x_ref[1]**o+(0.890972208341)*x_ref[1]+(-0.169067233411)*x_ref[2]**o+(-0.757264272484)*x_ref[2]
        ref[1,1]=(0.999152573666)*x_ref[0]**o+(-0.659676649566)*x_ref[0]+(0.993153643567)*x_ref[1]**o+(0.350141892848)*x_ref[1]+(-0.493784056536)*x_ref[2]**o+(-0.966494205893)*x_ref[2]
        ref[1,2]=(-0.141337889298)*x_ref[0]**o+(0.689935513838)*x_ref[0]+(0.335444607712)*x_ref[1]**o+(-0.733351283662)*x_ref[1]+(0.323775887172)*x_ref[2]**o+(0.314099156544)*x_ref[2]
        ref[1,3]=(-0.157132412118)*x_ref[0]**o+(0.602468153313)*x_ref[0]+(-0.618134886591)*x_ref[1]**o+(-0.404425578444)*x_ref[1]+(-0.5069585924)*x_ref[2]**o+(0.616829810596)*x_ref[2]
        ref[1,4]=(-0.107902610028)*x_ref[0]**o+(0.664787826515)*x_ref[0]+(-0.19940280262)*x_ref[1]**o+(0.461794954077)*x_ref[1]+(-0.0516671817578)*x_ref[2]**o+(0.426165289622)*x_ref[2]
        ref[2,0]=(-0.568061237616)*x_ref[0]**o+(-0.744340542576)*x_ref[0]+(-0.0125924903194)*x_ref[1]**o+(0.710153866172)*x_ref[1]+(0.979871553596)*x_ref[2]**o+(0.945862455036)*x_ref[2]
        ref[2,1]=(0.201599120538)*x_ref[0]**o+(-0.11057289344)*x_ref[0]+(0.522031147141)*x_ref[1]**o+(0.194694594101)*x_ref[1]+(-0.771997994079)*x_ref[2]**o+(-0.513574957413)*x_ref[2]
        ref[2,2]=(0.401979025382)*x_ref[0]**o+(-0.743641669014)*x_ref[0]+(-0.796071557874)*x_ref[1]**o+(-0.352651095657)*x_ref[1]+(-0.564117536877)*x_ref[2]**o+(0.952729715875)*x_ref[2]
        ref[2,3]=(-0.526162893019)*x_ref[0]**o+(0.691019471759)*x_ref[0]+(-0.68646084088)*x_ref[1]**o+(0.839630438544)*x_ref[1]+(-0.491997512287)*x_ref[2]**o+(-0.405069600537)*x_ref[2]
        ref[2,4]=(0.117664099703)*x_ref[0]**o+(-0.807496546596)*x_ref[0]+(0.607810703702)*x_ref[1]**o+(-0.696118154009)*x_ref[1]+(-0.156011683803)*x_ref[2]**o+(-0.26100420077)*x_ref[2]
        ref[3,0]=(0.280066174527)*x_ref[0]**o+(-0.225940135983)*x_ref[0]+(-0.338716163489)*x_ref[1]**o+(0.628405669096)*x_ref[1]+(-0.890543405425)*x_ref[2]**o+(-0.142082524497)*x_ref[2]
        ref[3,1]=(-0.659241351226)*x_ref[0]**o+(0.275483386126)*x_ref[0]+(-0.318352239058)*x_ref[1]**o+(0.168536648793)*x_ref[1]+(0.539555610218)*x_ref[2]**o+(-0.95280560435)*x_ref[2]
        ref[3,2]=(0.673861565841)*x_ref[0]**o+(-0.0303873544723)*x_ref[0]+(-0.638241094445)*x_ref[1]**o+(0.664488908579)*x_ref[1]+(-0.789484082135)*x_ref[2]**o+(-0.982437408359)*x_ref[2]
        ref[3,3]=(-0.902346309293)*x_ref[0]**o+(0.818106160371)*x_ref[0]+(-0.410074078135)*x_ref[1]**o+(0.977554592344)*x_ref[1]+(-0.00195907247124)*x_ref[2]**o+(0.94762151647)*x_ref[2]
        ref[3,4]=(0.27272813538)*x_ref[0]**o+(0.0563423990113)*x_ref[0]+(0.254612551508)*x_ref[1]**o+(-0.901948547799)*x_ref[1]+(0.263613075865)*x_ref[2]**o+(0.879691314116)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.023497804886)*x[0]**o+(0.934780781287)*x[0]+(0.488935915319)*x[1]**o+(-0.843152487959)*x[1]
        arg[0,0,1]=(0.074870516034)*x[0]**o+(-0.515115482624)*x[0]+(-0.582499912837)*x[1]**o+(-0.148911376041)*x[1]
        arg[0,1,0]=(0.0364775990427)*x[0]**o+(-0.258379408121)*x[0]+(0.109704222415)*x[1]**o+(-0.0149203380717)*x[1]
        arg[0,1,1]=(0.749804547117)*x[0]**o+(-0.65162531688)*x[0]+(-0.529952602569)*x[1]**o+(0.901735743956)*x[1]
        arg[1,0,0]=(0.74941277085)*x[0]**o+(-0.878903916612)*x[0]+(0.866605680273)*x[1]**o+(0.182572622995)*x[1]
        arg[1,0,1]=(-0.787476915117)*x[0]**o+(0.945314823283)*x[0]+(0.165599597226)*x[1]**o+(0.733712603726)*x[1]
        arg[1,1,0]=(0.696240085313)*x[0]**o+(0.775501454572)*x[0]+(0.225131553571)*x[1]**o+(-0.75288780057)*x[1]
        arg[1,1,1]=(-0.352416785284)*x[0]**o+(-0.681447081047)*x[0]+(0.164273429957)*x[1]**o+(-0.0977235074729)*x[1]
        arg[2,0,0]=(-0.484651384894)*x[0]**o+(0.162048679786)*x[0]+(0.433731200543)*x[1]**o+(-0.2255104779)*x[1]
        arg[2,0,1]=(0.230113349848)*x[0]**o+(0.674218493869)*x[0]+(0.124852942645)*x[1]**o+(0.361278270173)*x[1]
        arg[2,1,0]=(-0.604246197352)*x[0]**o+(0.26763755318)*x[0]+(-0.821482774631)*x[1]**o+(-0.821813801109)*x[1]
        arg[2,1,1]=(0.71669276429)*x[0]**o+(0.861782230871)*x[0]+(-0.837503788825)*x[1]**o+(0.0437253153979)*x[1]
        arg[3,0,0]=(-0.0941343724431)*x[0]**o+(0.376175369776)*x[0]+(-0.455752787585)*x[1]**o+(0.777480094583)*x[1]
        arg[3,0,1]=(-0.730897141543)*x[0]**o+(0.750046665085)*x[0]+(0.855742434657)*x[1]**o+(0.941826664412)*x[1]
        arg[3,1,0]=(-0.788757907886)*x[0]**o+(-0.0255067018415)*x[0]+(-0.875292088318)*x[1]**o+(0.401677848256)*x[1]
        arg[3,1,1]=(-0.980570910972)*x[0]**o+(0.936108649845)*x[0]+(-0.0463362409126)*x[1]**o+(0.630408734656)*x[1]
        arg[4,0,0]=(-0.72642075391)*x[0]**o+(0.283788406364)*x[0]+(-0.119645741265)*x[1]**o+(0.66948892422)*x[1]
        arg[4,0,1]=(0.178235706017)*x[0]**o+(0.693466107843)*x[0]+(-0.585602683437)*x[1]**o+(-0.493386713436)*x[1]
        arg[4,1,0]=(0.348039277129)*x[0]**o+(-0.806776084421)*x[0]+(-0.73591780841)*x[1]**o+(-0.310246367505)*x[1]
        arg[4,1,1]=(0.993294877418)*x[0]**o+(-0.683114302069)*x[0]+(0.0679022755308)*x[1]**o+(0.341981988936)*x[1]
        arg[5,0,0]=(-0.807380701959)*x[0]**o+(-0.78936779476)*x[0]+(0.142089012614)*x[1]**o+(-0.129335524031)*x[1]
        arg[5,0,1]=(-0.855714388031)*x[0]**o+(0.00825196217877)*x[0]+(0.310039590415)*x[1]**o+(-0.242861862397)*x[1]
        arg[5,1,0]=(-0.0908183346277)*x[0]**o+(-0.941876543235)*x[0]+(0.981459557493)*x[1]**o+(0.559563798858)*x[1]
        arg[5,1,1]=(-0.852223567371)*x[0]**o+(-0.272541886199)*x[0]+(-0.873908247032)*x[1]**o+(-0.387571769068)*x[1]
        ref[0,0,0]=(-0.023497804886)*x_ref[0]**o+(0.934780781287)*x_ref[0]+(0.488935915319)*x_ref[1]**o+(-0.843152487959)*x_ref[1]
        ref[0,0,1]=(0.074870516034)*x_ref[0]**o+(-0.515115482624)*x_ref[0]+(-0.582499912837)*x_ref[1]**o+(-0.148911376041)*x_ref[1]
        ref[0,1,0]=(0.0364775990427)*x_ref[0]**o+(-0.258379408121)*x_ref[0]+(0.109704222415)*x_ref[1]**o+(-0.0149203380717)*x_ref[1]
        ref[0,1,1]=(0.749804547117)*x_ref[0]**o+(-0.65162531688)*x_ref[0]+(-0.529952602569)*x_ref[1]**o+(0.901735743956)*x_ref[1]
        ref[1,0,0]=(0.74941277085)*x_ref[0]**o+(-0.878903916612)*x_ref[0]+(0.866605680273)*x_ref[1]**o+(0.182572622995)*x_ref[1]
        ref[1,0,1]=(-0.787476915117)*x_ref[0]**o+(0.945314823283)*x_ref[0]+(0.165599597226)*x_ref[1]**o+(0.733712603726)*x_ref[1]
        ref[1,1,0]=(0.696240085313)*x_ref[0]**o+(0.775501454572)*x_ref[0]+(0.225131553571)*x_ref[1]**o+(-0.75288780057)*x_ref[1]
        ref[1,1,1]=(-0.352416785284)*x_ref[0]**o+(-0.681447081047)*x_ref[0]+(0.164273429957)*x_ref[1]**o+(-0.0977235074729)*x_ref[1]
        ref[2,0,0]=(-0.484651384894)*x_ref[0]**o+(0.162048679786)*x_ref[0]+(0.433731200543)*x_ref[1]**o+(-0.2255104779)*x_ref[1]
        ref[2,0,1]=(0.230113349848)*x_ref[0]**o+(0.674218493869)*x_ref[0]+(0.124852942645)*x_ref[1]**o+(0.361278270173)*x_ref[1]
        ref[2,1,0]=(-0.604246197352)*x_ref[0]**o+(0.26763755318)*x_ref[0]+(-0.821482774631)*x_ref[1]**o+(-0.821813801109)*x_ref[1]
        ref[2,1,1]=(0.71669276429)*x_ref[0]**o+(0.861782230871)*x_ref[0]+(-0.837503788825)*x_ref[1]**o+(0.0437253153979)*x_ref[1]
        ref[3,0,0]=(-0.0941343724431)*x_ref[0]**o+(0.376175369776)*x_ref[0]+(-0.455752787585)*x_ref[1]**o+(0.777480094583)*x_ref[1]
        ref[3,0,1]=(-0.730897141543)*x_ref[0]**o+(0.750046665085)*x_ref[0]+(0.855742434657)*x_ref[1]**o+(0.941826664412)*x_ref[1]
        ref[3,1,0]=(-0.788757907886)*x_ref[0]**o+(-0.0255067018415)*x_ref[0]+(-0.875292088318)*x_ref[1]**o+(0.401677848256)*x_ref[1]
        ref[3,1,1]=(-0.980570910972)*x_ref[0]**o+(0.936108649845)*x_ref[0]+(-0.0463362409126)*x_ref[1]**o+(0.630408734656)*x_ref[1]
        ref[4,0,0]=(-0.72642075391)*x_ref[0]**o+(0.283788406364)*x_ref[0]+(-0.119645741265)*x_ref[1]**o+(0.66948892422)*x_ref[1]
        ref[4,0,1]=(0.178235706017)*x_ref[0]**o+(0.693466107843)*x_ref[0]+(-0.585602683437)*x_ref[1]**o+(-0.493386713436)*x_ref[1]
        ref[4,1,0]=(0.348039277129)*x_ref[0]**o+(-0.806776084421)*x_ref[0]+(-0.73591780841)*x_ref[1]**o+(-0.310246367505)*x_ref[1]
        ref[4,1,1]=(0.993294877418)*x_ref[0]**o+(-0.683114302069)*x_ref[0]+(0.0679022755308)*x_ref[1]**o+(0.341981988936)*x_ref[1]
        ref[5,0,0]=(-0.807380701959)*x_ref[0]**o+(-0.78936779476)*x_ref[0]+(0.142089012614)*x_ref[1]**o+(-0.129335524031)*x_ref[1]
        ref[5,0,1]=(-0.855714388031)*x_ref[0]**o+(0.00825196217877)*x_ref[0]+(0.310039590415)*x_ref[1]**o+(-0.242861862397)*x_ref[1]
        ref[5,1,0]=(-0.0908183346277)*x_ref[0]**o+(-0.941876543235)*x_ref[0]+(0.981459557493)*x_ref[1]**o+(0.559563798858)*x_ref[1]
        ref[5,1,1]=(-0.852223567371)*x_ref[0]**o+(-0.272541886199)*x_ref[0]+(-0.873908247032)*x_ref[1]**o+(-0.387571769068)*x_ref[1]
      else:
        arg[0,0,0]=(0.418981395334)*x[0]**o+(-0.720305875042)*x[0]+(0.876880439526)*x[1]**o+(-0.17785803746)*x[1]+(0.779797986429)*x[2]**o+(0.363827072746)*x[2]
        arg[0,0,1]=(0.159567966427)*x[0]**o+(-0.614151008569)*x[0]+(0.249982224851)*x[1]**o+(-0.119598774468)*x[1]+(-0.66603578755)*x[2]**o+(0.202190479552)*x[2]
        arg[0,1,0]=(0.337868020939)*x[0]**o+(-0.0828148711399)*x[0]+(-0.0631659329293)*x[1]**o+(-0.725100695213)*x[1]+(0.852068398846)*x[2]**o+(-0.951116521211)*x[2]
        arg[0,1,1]=(0.710866728186)*x[0]**o+(-0.517133891599)*x[0]+(0.320028321875)*x[1]**o+(0.0331233805016)*x[1]+(-0.0945382588101)*x[2]**o+(-0.795725457058)*x[2]
        arg[1,0,0]=(0.261503329866)*x[0]**o+(0.697062723662)*x[0]+(0.926550326146)*x[1]**o+(-0.318704885103)*x[1]+(-0.76904956799)*x[2]**o+(0.0750521033639)*x[2]
        arg[1,0,1]=(-0.594460213676)*x[0]**o+(0.356692399969)*x[0]+(-0.611558953982)*x[1]**o+(0.289469152887)*x[1]+(-0.73222413782)*x[2]**o+(-0.576883793051)*x[2]
        arg[1,1,0]=(-0.399324648161)*x[0]**o+(0.975428000926)*x[0]+(-0.889762441468)*x[1]**o+(-0.137573068244)*x[1]+(0.252276361744)*x[2]**o+(-0.554435968553)*x[2]
        arg[1,1,1]=(0.677933864695)*x[0]**o+(-0.19825134391)*x[0]+(0.556783296976)*x[1]**o+(0.0349255641106)*x[1]+(-0.387378575139)*x[2]**o+(0.889628545207)*x[2]
        arg[2,0,0]=(-0.839356850436)*x[0]**o+(0.0186947034157)*x[0]+(-0.264793716814)*x[1]**o+(-0.767468006197)*x[1]+(0.91661695419)*x[2]**o+(-0.374036622497)*x[2]
        arg[2,0,1]=(-0.030357536408)*x[0]**o+(0.57424762933)*x[0]+(-0.590115256932)*x[1]**o+(-0.176169776757)*x[1]+(-0.502845936823)*x[2]**o+(0.0399243413515)*x[2]
        arg[2,1,0]=(0.257324735469)*x[0]**o+(-0.686083724325)*x[0]+(-0.166609564696)*x[1]**o+(-0.132181292726)*x[1]+(0.10419146703)*x[2]**o+(-0.827571838114)*x[2]
        arg[2,1,1]=(0.339465982319)*x[0]**o+(0.766586466841)*x[0]+(0.277538823819)*x[1]**o+(0.222607512697)*x[1]+(-0.909105282607)*x[2]**o+(0.463729842225)*x[2]
        arg[3,0,0]=(-0.00132043337846)*x[0]**o+(0.670451981959)*x[0]+(-0.0122245380346)*x[1]**o+(-0.922535904926)*x[1]+(0.49486742)*x[2]**o+(0.0785274524546)*x[2]
        arg[3,0,1]=(0.0360876610851)*x[0]**o+(0.69319604661)*x[0]+(0.980826095079)*x[1]**o+(0.167689443923)*x[1]+(0.561634492527)*x[2]**o+(0.996948790413)*x[2]
        arg[3,1,0]=(-0.238705628819)*x[0]**o+(-0.915859436874)*x[0]+(0.942122040692)*x[1]**o+(0.971025407162)*x[1]+(-0.965945363654)*x[2]**o+(0.136537412741)*x[2]
        arg[3,1,1]=(-0.672314439537)*x[0]**o+(0.792647645097)*x[0]+(-0.432226224113)*x[1]**o+(0.424246857841)*x[1]+(-0.639842531517)*x[2]**o+(0.0570057507731)*x[2]
        arg[4,0,0]=(0.450831267081)*x[0]**o+(0.75992870254)*x[0]+(0.768779936214)*x[1]**o+(-0.0135947286596)*x[1]+(0.126345191973)*x[2]**o+(-0.619763308655)*x[2]
        arg[4,0,1]=(-0.325005948888)*x[0]**o+(-0.0655421631802)*x[0]+(0.607182269636)*x[1]**o+(-0.631044141835)*x[1]+(-0.356580309429)*x[2]**o+(-0.593198006905)*x[2]
        arg[4,1,0]=(0.304932527518)*x[0]**o+(-0.763995484499)*x[0]+(0.884855995183)*x[1]**o+(-0.42180322102)*x[1]+(0.605742059302)*x[2]**o+(0.448553789891)*x[2]
        arg[4,1,1]=(-0.787145437063)*x[0]**o+(-0.465220934558)*x[0]+(-0.370336760606)*x[1]**o+(-0.854850692759)*x[1]+(0.482896516171)*x[2]**o+(-0.551567519291)*x[2]
        arg[5,0,0]=(-0.537738873639)*x[0]**o+(-0.213268842545)*x[0]+(0.492208051415)*x[1]**o+(-0.609304634017)*x[1]+(-0.487753207718)*x[2]**o+(-0.70650655894)*x[2]
        arg[5,0,1]=(0.600359837292)*x[0]**o+(-0.604421599924)*x[0]+(-0.051356466466)*x[1]**o+(-0.168592544661)*x[1]+(0.789340570036)*x[2]**o+(0.910249026599)*x[2]
        arg[5,1,0]=(0.264506197816)*x[0]**o+(-0.711775430363)*x[0]+(0.384666624979)*x[1]**o+(0.0926714847986)*x[1]+(-0.551232557648)*x[2]**o+(0.758927448717)*x[2]
        arg[5,1,1]=(0.894998835449)*x[0]**o+(-0.466338210072)*x[0]+(-0.522475632898)*x[1]**o+(0.766949173671)*x[1]+(0.915063619607)*x[2]**o+(-0.270696846692)*x[2]
        ref[0,0,0]=(0.418981395334)*x_ref[0]**o+(-0.720305875042)*x_ref[0]+(0.876880439526)*x_ref[1]**o+(-0.17785803746)*x_ref[1]+(0.779797986429)*x_ref[2]**o+(0.363827072746)*x_ref[2]
        ref[0,0,1]=(0.159567966427)*x_ref[0]**o+(-0.614151008569)*x_ref[0]+(0.249982224851)*x_ref[1]**o+(-0.119598774468)*x_ref[1]+(-0.66603578755)*x_ref[2]**o+(0.202190479552)*x_ref[2]
        ref[0,1,0]=(0.337868020939)*x_ref[0]**o+(-0.0828148711399)*x_ref[0]+(-0.0631659329293)*x_ref[1]**o+(-0.725100695213)*x_ref[1]+(0.852068398846)*x_ref[2]**o+(-0.951116521211)*x_ref[2]
        ref[0,1,1]=(0.710866728186)*x_ref[0]**o+(-0.517133891599)*x_ref[0]+(0.320028321875)*x_ref[1]**o+(0.0331233805016)*x_ref[1]+(-0.0945382588101)*x_ref[2]**o+(-0.795725457058)*x_ref[2]
        ref[1,0,0]=(0.261503329866)*x_ref[0]**o+(0.697062723662)*x_ref[0]+(0.926550326146)*x_ref[1]**o+(-0.318704885103)*x_ref[1]+(-0.76904956799)*x_ref[2]**o+(0.0750521033639)*x_ref[2]
        ref[1,0,1]=(-0.594460213676)*x_ref[0]**o+(0.356692399969)*x_ref[0]+(-0.611558953982)*x_ref[1]**o+(0.289469152887)*x_ref[1]+(-0.73222413782)*x_ref[2]**o+(-0.576883793051)*x_ref[2]
        ref[1,1,0]=(-0.399324648161)*x_ref[0]**o+(0.975428000926)*x_ref[0]+(-0.889762441468)*x_ref[1]**o+(-0.137573068244)*x_ref[1]+(0.252276361744)*x_ref[2]**o+(-0.554435968553)*x_ref[2]
        ref[1,1,1]=(0.677933864695)*x_ref[0]**o+(-0.19825134391)*x_ref[0]+(0.556783296976)*x_ref[1]**o+(0.0349255641106)*x_ref[1]+(-0.387378575139)*x_ref[2]**o+(0.889628545207)*x_ref[2]
        ref[2,0,0]=(-0.839356850436)*x_ref[0]**o+(0.0186947034157)*x_ref[0]+(-0.264793716814)*x_ref[1]**o+(-0.767468006197)*x_ref[1]+(0.91661695419)*x_ref[2]**o+(-0.374036622497)*x_ref[2]
        ref[2,0,1]=(-0.030357536408)*x_ref[0]**o+(0.57424762933)*x_ref[0]+(-0.590115256932)*x_ref[1]**o+(-0.176169776757)*x_ref[1]+(-0.502845936823)*x_ref[2]**o+(0.0399243413515)*x_ref[2]
        ref[2,1,0]=(0.257324735469)*x_ref[0]**o+(-0.686083724325)*x_ref[0]+(-0.166609564696)*x_ref[1]**o+(-0.132181292726)*x_ref[1]+(0.10419146703)*x_ref[2]**o+(-0.827571838114)*x_ref[2]
        ref[2,1,1]=(0.339465982319)*x_ref[0]**o+(0.766586466841)*x_ref[0]+(0.277538823819)*x_ref[1]**o+(0.222607512697)*x_ref[1]+(-0.909105282607)*x_ref[2]**o+(0.463729842225)*x_ref[2]
        ref[3,0,0]=(-0.00132043337846)*x_ref[0]**o+(0.670451981959)*x_ref[0]+(-0.0122245380346)*x_ref[1]**o+(-0.922535904926)*x_ref[1]+(0.49486742)*x_ref[2]**o+(0.0785274524546)*x_ref[2]
        ref[3,0,1]=(0.0360876610851)*x_ref[0]**o+(0.69319604661)*x_ref[0]+(0.980826095079)*x_ref[1]**o+(0.167689443923)*x_ref[1]+(0.561634492527)*x_ref[2]**o+(0.996948790413)*x_ref[2]
        ref[3,1,0]=(-0.238705628819)*x_ref[0]**o+(-0.915859436874)*x_ref[0]+(0.942122040692)*x_ref[1]**o+(0.971025407162)*x_ref[1]+(-0.965945363654)*x_ref[2]**o+(0.136537412741)*x_ref[2]
        ref[3,1,1]=(-0.672314439537)*x_ref[0]**o+(0.792647645097)*x_ref[0]+(-0.432226224113)*x_ref[1]**o+(0.424246857841)*x_ref[1]+(-0.639842531517)*x_ref[2]**o+(0.0570057507731)*x_ref[2]
        ref[4,0,0]=(0.450831267081)*x_ref[0]**o+(0.75992870254)*x_ref[0]+(0.768779936214)*x_ref[1]**o+(-0.0135947286596)*x_ref[1]+(0.126345191973)*x_ref[2]**o+(-0.619763308655)*x_ref[2]
        ref[4,0,1]=(-0.325005948888)*x_ref[0]**o+(-0.0655421631802)*x_ref[0]+(0.607182269636)*x_ref[1]**o+(-0.631044141835)*x_ref[1]+(-0.356580309429)*x_ref[2]**o+(-0.593198006905)*x_ref[2]
        ref[4,1,0]=(0.304932527518)*x_ref[0]**o+(-0.763995484499)*x_ref[0]+(0.884855995183)*x_ref[1]**o+(-0.42180322102)*x_ref[1]+(0.605742059302)*x_ref[2]**o+(0.448553789891)*x_ref[2]
        ref[4,1,1]=(-0.787145437063)*x_ref[0]**o+(-0.465220934558)*x_ref[0]+(-0.370336760606)*x_ref[1]**o+(-0.854850692759)*x_ref[1]+(0.482896516171)*x_ref[2]**o+(-0.551567519291)*x_ref[2]
        ref[5,0,0]=(-0.537738873639)*x_ref[0]**o+(-0.213268842545)*x_ref[0]+(0.492208051415)*x_ref[1]**o+(-0.609304634017)*x_ref[1]+(-0.487753207718)*x_ref[2]**o+(-0.70650655894)*x_ref[2]
        ref[5,0,1]=(0.600359837292)*x_ref[0]**o+(-0.604421599924)*x_ref[0]+(-0.051356466466)*x_ref[1]**o+(-0.168592544661)*x_ref[1]+(0.789340570036)*x_ref[2]**o+(0.910249026599)*x_ref[2]
        ref[5,1,0]=(0.264506197816)*x_ref[0]**o+(-0.711775430363)*x_ref[0]+(0.384666624979)*x_ref[1]**o+(0.0926714847986)*x_ref[1]+(-0.551232557648)*x_ref[2]**o+(0.758927448717)*x_ref[2]
        ref[5,1,1]=(0.894998835449)*x_ref[0]**o+(-0.466338210072)*x_ref[0]+(-0.522475632898)*x_ref[1]**o+(0.766949173671)*x_ref[1]+(0.915063619607)*x_ref[2]**o+(-0.270696846692)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromData_ReducedFunctionOnBoundary_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.240887319278)*x[0]**o+(-0.192166587545)*x[0]+(0.0176974287409)*x[1]**o+(0.100761748824)*x[1]
        arg[0,0,0,1]=(-0.549649845775)*x[0]**o+(-0.293969245646)*x[0]+(-0.605885325133)*x[1]**o+(0.799558451173)*x[1]
        arg[0,0,1,0]=(-0.899658769582)*x[0]**o+(0.484524193987)*x[0]+(0.437485870351)*x[1]**o+(-0.220549034927)*x[1]
        arg[0,0,1,1]=(-0.146539961793)*x[0]**o+(-0.278343486929)*x[0]+(0.0913912456082)*x[1]**o+(-0.751028975123)*x[1]
        arg[0,0,2,0]=(0.806344701194)*x[0]**o+(0.0537872824561)*x[0]+(-0.70407828379)*x[1]**o+(-0.922963486601)*x[1]
        arg[0,0,2,1]=(0.307202902495)*x[0]**o+(0.301603506731)*x[0]+(0.232983743616)*x[1]**o+(-0.744267542759)*x[1]
        arg[0,1,0,0]=(-0.550616060824)*x[0]**o+(-0.661367750254)*x[0]+(-0.695905895864)*x[1]**o+(-0.788045922398)*x[1]
        arg[0,1,0,1]=(-0.934169134035)*x[0]**o+(-0.843861105521)*x[0]+(-0.237877254439)*x[1]**o+(-0.364956821348)*x[1]
        arg[0,1,1,0]=(0.0950675750026)*x[0]**o+(-0.815510640594)*x[0]+(0.72526035968)*x[1]**o+(0.898675190106)*x[1]
        arg[0,1,1,1]=(0.715609066776)*x[0]**o+(0.371314845206)*x[0]+(0.326878867244)*x[1]**o+(-0.468798950199)*x[1]
        arg[0,1,2,0]=(0.392175939547)*x[0]**o+(0.105332682674)*x[0]+(0.997967455477)*x[1]**o+(0.0322491822496)*x[1]
        arg[0,1,2,1]=(-0.804498399187)*x[0]**o+(0.819653744739)*x[0]+(0.609610573703)*x[1]**o+(-0.623299047794)*x[1]
        arg[0,2,0,0]=(0.993877917386)*x[0]**o+(0.815952897899)*x[0]+(-0.258432174486)*x[1]**o+(0.209324387308)*x[1]
        arg[0,2,0,1]=(0.395946889364)*x[0]**o+(-0.859500539908)*x[0]+(-0.68495734785)*x[1]**o+(0.791813946302)*x[1]
        arg[0,2,1,0]=(0.154347661108)*x[0]**o+(0.0873140744322)*x[0]+(-0.423935109956)*x[1]**o+(-0.754297239226)*x[1]
        arg[0,2,1,1]=(-0.623998419484)*x[0]**o+(0.487300130941)*x[0]+(-0.323613219784)*x[1]**o+(0.744877983729)*x[1]
        arg[0,2,2,0]=(0.515686374371)*x[0]**o+(0.512572153346)*x[0]+(-0.641245783272)*x[1]**o+(-0.390660158759)*x[1]
        arg[0,2,2,1]=(0.24398238139)*x[0]**o+(-0.172020271678)*x[0]+(0.980197812701)*x[1]**o+(0.336941934503)*x[1]
        arg[0,3,0,0]=(0.0154584447615)*x[0]**o+(-0.547048191238)*x[0]+(-0.555319223005)*x[1]**o+(0.0972794598106)*x[1]
        arg[0,3,0,1]=(0.44744389335)*x[0]**o+(0.475046337874)*x[0]+(0.876228982066)*x[1]**o+(-0.0636405241175)*x[1]
        arg[0,3,1,0]=(0.392948465298)*x[0]**o+(-0.341342654182)*x[0]+(0.747185683968)*x[1]**o+(-0.99981172166)*x[1]
        arg[0,3,1,1]=(-0.172355321621)*x[0]**o+(-0.531391484795)*x[0]+(0.451654274019)*x[1]**o+(0.0795298667915)*x[1]
        arg[0,3,2,0]=(0.0841642474674)*x[0]**o+(0.308494435164)*x[0]+(0.53181918181)*x[1]**o+(0.480258067959)*x[1]
        arg[0,3,2,1]=(-0.222903054864)*x[0]**o+(-0.0330655869754)*x[0]+(0.227482525015)*x[1]**o+(0.536136212347)*x[1]
        arg[0,4,0,0]=(0.574798381284)*x[0]**o+(0.181354466484)*x[0]+(0.103049514171)*x[1]**o+(0.156282818835)*x[1]
        arg[0,4,0,1]=(0.478314255289)*x[0]**o+(0.224797222592)*x[0]+(0.307668606272)*x[1]**o+(-0.785262182937)*x[1]
        arg[0,4,1,0]=(0.519356961182)*x[0]**o+(0.646200458344)*x[0]+(0.0522618783181)*x[1]**o+(0.702335307324)*x[1]
        arg[0,4,1,1]=(-0.168756029541)*x[0]**o+(-0.00618364750666)*x[0]+(0.472704616213)*x[1]**o+(0.974257451871)*x[1]
        arg[0,4,2,0]=(-0.279196869573)*x[0]**o+(-0.333083174091)*x[0]+(-0.980089574428)*x[1]**o+(0.405273709405)*x[1]
        arg[0,4,2,1]=(-0.199431380673)*x[0]**o+(0.823294899866)*x[0]+(-0.174125117685)*x[1]**o+(-0.730500493887)*x[1]
        arg[1,0,0,0]=(-0.577281489635)*x[0]**o+(0.231906139898)*x[0]+(0.537128708963)*x[1]**o+(0.363444540898)*x[1]
        arg[1,0,0,1]=(0.204108541882)*x[0]**o+(-0.217661714888)*x[0]+(0.963384293759)*x[1]**o+(-0.0817519801351)*x[1]
        arg[1,0,1,0]=(0.540681325439)*x[0]**o+(-0.496036960808)*x[0]+(0.421037689661)*x[1]**o+(0.19399086996)*x[1]
        arg[1,0,1,1]=(0.482938542126)*x[0]**o+(-0.0725027675209)*x[0]+(-0.441517962054)*x[1]**o+(0.787445860714)*x[1]
        arg[1,0,2,0]=(0.0378928111257)*x[0]**o+(0.451368848019)*x[0]+(-0.228668284485)*x[1]**o+(0.505927229432)*x[1]
        arg[1,0,2,1]=(-0.59833518872)*x[0]**o+(0.956068253354)*x[0]+(0.667905995469)*x[1]**o+(0.963273292214)*x[1]
        arg[1,1,0,0]=(0.901637865236)*x[0]**o+(-0.642678971748)*x[0]+(-0.603915019524)*x[1]**o+(-0.208004993919)*x[1]
        arg[1,1,0,1]=(-0.227162661052)*x[0]**o+(-0.5816434303)*x[0]+(0.274961293801)*x[1]**o+(-0.614001328621)*x[1]
        arg[1,1,1,0]=(0.078438078548)*x[0]**o+(0.570191242883)*x[0]+(0.0952878212484)*x[1]**o+(0.844460205497)*x[1]
        arg[1,1,1,1]=(-0.901445811786)*x[0]**o+(-0.996376775344)*x[0]+(0.707470471613)*x[1]**o+(-0.187517936674)*x[1]
        arg[1,1,2,0]=(-0.66891443067)*x[0]**o+(0.0462057239324)*x[0]+(-0.825820539128)*x[1]**o+(-0.766932836843)*x[1]
        arg[1,1,2,1]=(-0.851428828811)*x[0]**o+(0.504484422346)*x[0]+(0.121378315545)*x[1]**o+(0.903947012757)*x[1]
        arg[1,2,0,0]=(-0.393827449768)*x[0]**o+(-0.251827904608)*x[0]+(0.8320040882)*x[1]**o+(0.57595149855)*x[1]
        arg[1,2,0,1]=(0.638738110712)*x[0]**o+(-0.439962152539)*x[0]+(-0.977211078995)*x[1]**o+(0.0528475212451)*x[1]
        arg[1,2,1,0]=(0.824313588941)*x[0]**o+(-0.589968996823)*x[0]+(-0.512156794152)*x[1]**o+(0.142809794443)*x[1]
        arg[1,2,1,1]=(-0.31789333147)*x[0]**o+(0.598839896949)*x[0]+(0.638319398651)*x[1]**o+(0.353625999317)*x[1]
        arg[1,2,2,0]=(-0.720472365012)*x[0]**o+(0.896538030051)*x[0]+(-0.224639169972)*x[1]**o+(-0.898299360738)*x[1]
        arg[1,2,2,1]=(0.123077707167)*x[0]**o+(0.15922851038)*x[0]+(0.678984783843)*x[1]**o+(-0.778188269852)*x[1]
        arg[1,3,0,0]=(0.344256301209)*x[0]**o+(0.884367738289)*x[0]+(0.00556636219289)*x[1]**o+(0.684361331195)*x[1]
        arg[1,3,0,1]=(-0.812245131065)*x[0]**o+(0.971365671228)*x[0]+(0.0696053239569)*x[1]**o+(-0.300266235138)*x[1]
        arg[1,3,1,0]=(0.470254588058)*x[0]**o+(0.0946825509913)*x[0]+(-0.415619255025)*x[1]**o+(-0.27381652858)*x[1]
        arg[1,3,1,1]=(-0.84555138241)*x[0]**o+(0.873510832091)*x[0]+(-0.138010338152)*x[1]**o+(-0.982274515316)*x[1]
        arg[1,3,2,0]=(0.695614136148)*x[0]**o+(-0.192588082771)*x[0]+(-0.21579127791)*x[1]**o+(-0.0335939736388)*x[1]
        arg[1,3,2,1]=(-0.633363463896)*x[0]**o+(-0.958320549993)*x[0]+(0.197331067481)*x[1]**o+(0.315089888409)*x[1]
        arg[1,4,0,0]=(-0.211531219945)*x[0]**o+(-0.399700721799)*x[0]+(0.578927638839)*x[1]**o+(-0.389034976195)*x[1]
        arg[1,4,0,1]=(-0.846399571893)*x[0]**o+(-0.952904755903)*x[0]+(-0.858399802693)*x[1]**o+(-0.23264354907)*x[1]
        arg[1,4,1,0]=(0.255822441586)*x[0]**o+(0.896237710393)*x[0]+(-0.90248008271)*x[1]**o+(0.654804848244)*x[1]
        arg[1,4,1,1]=(-0.112475508163)*x[0]**o+(-0.861697204453)*x[0]+(-0.782581931187)*x[1]**o+(-0.129345341581)*x[1]
        arg[1,4,2,0]=(0.690095687033)*x[0]**o+(0.729317334019)*x[0]+(-0.00961483612288)*x[1]**o+(-0.288320551243)*x[1]
        arg[1,4,2,1]=(-0.648146433517)*x[0]**o+(-0.301246601263)*x[0]+(-0.490119612994)*x[1]**o+(0.231186845615)*x[1]
        arg[2,0,0,0]=(-0.645560891595)*x[0]**o+(-0.889067544988)*x[0]+(-0.193353414221)*x[1]**o+(-0.726988479344)*x[1]
        arg[2,0,0,1]=(-0.083614738484)*x[0]**o+(-0.822530927241)*x[0]+(0.660338934241)*x[1]**o+(0.0383821036508)*x[1]
        arg[2,0,1,0]=(-0.132250736664)*x[0]**o+(-0.571474370931)*x[0]+(-0.755931100342)*x[1]**o+(-0.145196865115)*x[1]
        arg[2,0,1,1]=(0.545219196727)*x[0]**o+(-0.846368787913)*x[0]+(-0.845706160842)*x[1]**o+(0.550703610989)*x[1]
        arg[2,0,2,0]=(-0.766303732195)*x[0]**o+(0.156822893641)*x[0]+(0.00954334254282)*x[1]**o+(0.201506144561)*x[1]
        arg[2,0,2,1]=(-0.274476449632)*x[0]**o+(-0.361690468185)*x[0]+(-0.816535747268)*x[1]**o+(-0.473913811701)*x[1]
        arg[2,1,0,0]=(-0.927592771482)*x[0]**o+(0.534615269814)*x[0]+(-0.571217133452)*x[1]**o+(0.57448299216)*x[1]
        arg[2,1,0,1]=(-0.298153953211)*x[0]**o+(-0.143636909462)*x[0]+(0.990660288864)*x[1]**o+(-0.495864284763)*x[1]
        arg[2,1,1,0]=(-0.110625727208)*x[0]**o+(-0.221057774317)*x[0]+(0.557195078358)*x[1]**o+(0.803162691808)*x[1]
        arg[2,1,1,1]=(-0.115589176775)*x[0]**o+(0.08253351782)*x[0]+(-0.595996958977)*x[1]**o+(0.501435717501)*x[1]
        arg[2,1,2,0]=(0.78764976658)*x[0]**o+(0.966729831711)*x[0]+(-0.585879575667)*x[1]**o+(0.341148176793)*x[1]
        arg[2,1,2,1]=(-0.551697945691)*x[0]**o+(0.545857249323)*x[0]+(-0.919321724495)*x[1]**o+(-0.286043327415)*x[1]
        arg[2,2,0,0]=(-0.669933890655)*x[0]**o+(-0.344862118495)*x[0]+(0.878331282597)*x[1]**o+(0.881893760718)*x[1]
        arg[2,2,0,1]=(-0.217876618414)*x[0]**o+(-0.289118007199)*x[0]+(0.513464171525)*x[1]**o+(0.310055899102)*x[1]
        arg[2,2,1,0]=(0.0472051379154)*x[0]**o+(0.42455388451)*x[0]+(-0.920571834891)*x[1]**o+(-0.779095076537)*x[1]
        arg[2,2,1,1]=(-0.299619304537)*x[0]**o+(-0.289777287294)*x[0]+(0.920505062988)*x[1]**o+(-0.243924179095)*x[1]
        arg[2,2,2,0]=(-0.819343530877)*x[0]**o+(0.975953690875)*x[0]+(-0.787239459241)*x[1]**o+(-0.199900734691)*x[1]
        arg[2,2,2,1]=(0.0980085780878)*x[0]**o+(0.642946515451)*x[0]+(0.951790535152)*x[1]**o+(-0.253914817366)*x[1]
        arg[2,3,0,0]=(-0.882075522824)*x[0]**o+(0.200934502682)*x[0]+(-0.83200464496)*x[1]**o+(0.941479345375)*x[1]
        arg[2,3,0,1]=(-0.615941782209)*x[0]**o+(-0.693242388092)*x[0]+(-0.640417278871)*x[1]**o+(-0.346459094803)*x[1]
        arg[2,3,1,0]=(-0.447919427952)*x[0]**o+(0.205055689141)*x[0]+(-0.453290032411)*x[1]**o+(0.879062836521)*x[1]
        arg[2,3,1,1]=(0.721133924016)*x[0]**o+(0.71294558459)*x[0]+(-0.250087045805)*x[1]**o+(0.605146721129)*x[1]
        arg[2,3,2,0]=(0.247782322921)*x[0]**o+(-0.956367425119)*x[0]+(-0.054045082005)*x[1]**o+(-0.128151832132)*x[1]
        arg[2,3,2,1]=(-0.427039259088)*x[0]**o+(-0.359849242839)*x[0]+(-0.244990934481)*x[1]**o+(-0.038020035629)*x[1]
        arg[2,4,0,0]=(-0.216409150834)*x[0]**o+(0.492547989921)*x[0]+(0.878317438469)*x[1]**o+(-0.112718719119)*x[1]
        arg[2,4,0,1]=(0.487403443102)*x[0]**o+(0.228155519468)*x[0]+(0.923150527516)*x[1]**o+(0.642654867908)*x[1]
        arg[2,4,1,0]=(0.916418483417)*x[0]**o+(-0.177198935173)*x[0]+(-0.790548940412)*x[1]**o+(0.87540167052)*x[1]
        arg[2,4,1,1]=(0.718502075538)*x[0]**o+(0.691254010569)*x[0]+(-0.0134430122729)*x[1]**o+(-0.739888770749)*x[1]
        arg[2,4,2,0]=(0.287825086861)*x[0]**o+(0.596621809347)*x[0]+(0.729784205391)*x[1]**o+(-0.25984257955)*x[1]
        arg[2,4,2,1]=(-0.515102480025)*x[0]**o+(-0.609951840015)*x[0]+(0.229834340075)*x[1]**o+(-0.727157816241)*x[1]
        arg[3,0,0,0]=(-0.349406091363)*x[0]**o+(0.802348451103)*x[0]+(-0.174421346276)*x[1]**o+(0.120711198036)*x[1]
        arg[3,0,0,1]=(0.367085169423)*x[0]**o+(-0.319683236434)*x[0]+(-0.590820542566)*x[1]**o+(0.607625953303)*x[1]
        arg[3,0,1,0]=(0.231259976433)*x[0]**o+(-0.493297408825)*x[0]+(0.919363640504)*x[1]**o+(-0.839841270429)*x[1]
        arg[3,0,1,1]=(0.57672135684)*x[0]**o+(0.882062879704)*x[0]+(-0.600979706444)*x[1]**o+(-0.704008344509)*x[1]
        arg[3,0,2,0]=(-0.208162114527)*x[0]**o+(0.593682243052)*x[0]+(-0.428436832487)*x[1]**o+(0.51007021604)*x[1]
        arg[3,0,2,1]=(0.315264672653)*x[0]**o+(-0.0835897376534)*x[0]+(-0.135315173106)*x[1]**o+(-0.646372309498)*x[1]
        arg[3,1,0,0]=(0.815505953752)*x[0]**o+(0.0625493643292)*x[0]+(0.383118433815)*x[1]**o+(0.17372720435)*x[1]
        arg[3,1,0,1]=(0.557295786969)*x[0]**o+(-0.568823508165)*x[0]+(-0.00874733267675)*x[1]**o+(-0.27454140598)*x[1]
        arg[3,1,1,0]=(0.649021026591)*x[0]**o+(0.232255030448)*x[0]+(0.845939296425)*x[1]**o+(0.116387378294)*x[1]
        arg[3,1,1,1]=(0.431124940343)*x[0]**o+(-0.0865937271969)*x[0]+(0.937911741288)*x[1]**o+(0.362850039234)*x[1]
        arg[3,1,2,0]=(0.155965414439)*x[0]**o+(0.110850421841)*x[0]+(0.244563080191)*x[1]**o+(0.266325498006)*x[1]
        arg[3,1,2,1]=(0.766813300935)*x[0]**o+(0.77492422889)*x[0]+(-0.0141436554714)*x[1]**o+(0.780444449009)*x[1]
        arg[3,2,0,0]=(0.619086019019)*x[0]**o+(-0.574130567601)*x[0]+(-0.417708159593)*x[1]**o+(0.707727284254)*x[1]
        arg[3,2,0,1]=(-0.845657409353)*x[0]**o+(0.705282507748)*x[0]+(0.124557418112)*x[1]**o+(-0.337493494733)*x[1]
        arg[3,2,1,0]=(0.114875146211)*x[0]**o+(-0.724480426892)*x[0]+(-0.0763476904405)*x[1]**o+(0.12064907114)*x[1]
        arg[3,2,1,1]=(-0.74994413376)*x[0]**o+(-0.916476035217)*x[0]+(-0.360412756599)*x[1]**o+(-0.917457363098)*x[1]
        arg[3,2,2,0]=(-0.446839470984)*x[0]**o+(-0.097704838316)*x[0]+(-0.653584364909)*x[1]**o+(0.561014757758)*x[1]
        arg[3,2,2,1]=(-0.178975301038)*x[0]**o+(-0.843669982667)*x[0]+(-0.566610094242)*x[1]**o+(0.213047649003)*x[1]
        arg[3,3,0,0]=(0.144115764355)*x[0]**o+(0.199435411954)*x[0]+(0.173334385199)*x[1]**o+(-0.942957035808)*x[1]
        arg[3,3,0,1]=(-0.417589926342)*x[0]**o+(-0.819188022852)*x[0]+(-0.68299160003)*x[1]**o+(-0.314851447894)*x[1]
        arg[3,3,1,0]=(0.737898219164)*x[0]**o+(0.612906277393)*x[0]+(-0.768349396729)*x[1]**o+(-0.192998755095)*x[1]
        arg[3,3,1,1]=(0.35433362007)*x[0]**o+(0.894174225208)*x[0]+(-0.450333030837)*x[1]**o+(0.819450589483)*x[1]
        arg[3,3,2,0]=(-0.455377131302)*x[0]**o+(-0.748183144687)*x[0]+(-0.482739450231)*x[1]**o+(0.300625929022)*x[1]
        arg[3,3,2,1]=(-0.602766693698)*x[0]**o+(-0.416288436985)*x[0]+(0.837501983536)*x[1]**o+(-0.745935308188)*x[1]
        arg[3,4,0,0]=(-0.706774010405)*x[0]**o+(0.407766818701)*x[0]+(-0.248964401443)*x[1]**o+(0.675466487229)*x[1]
        arg[3,4,0,1]=(0.512302883876)*x[0]**o+(0.0809544867105)*x[0]+(-0.204555203536)*x[1]**o+(-0.628574073749)*x[1]
        arg[3,4,1,0]=(0.126376796421)*x[0]**o+(0.174607814129)*x[0]+(0.68867194136)*x[1]**o+(0.940025934568)*x[1]
        arg[3,4,1,1]=(0.583042444739)*x[0]**o+(-0.786981763836)*x[0]+(0.319397032405)*x[1]**o+(-0.324993351432)*x[1]
        arg[3,4,2,0]=(-0.941997068836)*x[0]**o+(-0.892825520378)*x[0]+(0.275995419027)*x[1]**o+(0.841260061022)*x[1]
        arg[3,4,2,1]=(0.520110030788)*x[0]**o+(0.519262786324)*x[0]+(0.861221788948)*x[1]**o+(-0.376714934937)*x[1]
        ref[0,0,0,0]=(0.240887319278)*x_ref[0]**o+(-0.192166587545)*x_ref[0]+(0.0176974287409)*x_ref[1]**o+(0.100761748824)*x_ref[1]
        ref[0,0,0,1]=(-0.549649845775)*x_ref[0]**o+(-0.293969245646)*x_ref[0]+(-0.605885325133)*x_ref[1]**o+(0.799558451173)*x_ref[1]
        ref[0,0,1,0]=(-0.899658769582)*x_ref[0]**o+(0.484524193987)*x_ref[0]+(0.437485870351)*x_ref[1]**o+(-0.220549034927)*x_ref[1]
        ref[0,0,1,1]=(-0.146539961793)*x_ref[0]**o+(-0.278343486929)*x_ref[0]+(0.0913912456082)*x_ref[1]**o+(-0.751028975123)*x_ref[1]
        ref[0,0,2,0]=(0.806344701194)*x_ref[0]**o+(0.0537872824561)*x_ref[0]+(-0.70407828379)*x_ref[1]**o+(-0.922963486601)*x_ref[1]
        ref[0,0,2,1]=(0.307202902495)*x_ref[0]**o+(0.301603506731)*x_ref[0]+(0.232983743616)*x_ref[1]**o+(-0.744267542759)*x_ref[1]
        ref[0,1,0,0]=(-0.550616060824)*x_ref[0]**o+(-0.661367750254)*x_ref[0]+(-0.695905895864)*x_ref[1]**o+(-0.788045922398)*x_ref[1]
        ref[0,1,0,1]=(-0.934169134035)*x_ref[0]**o+(-0.843861105521)*x_ref[0]+(-0.237877254439)*x_ref[1]**o+(-0.364956821348)*x_ref[1]
        ref[0,1,1,0]=(0.0950675750026)*x_ref[0]**o+(-0.815510640594)*x_ref[0]+(0.72526035968)*x_ref[1]**o+(0.898675190106)*x_ref[1]
        ref[0,1,1,1]=(0.715609066776)*x_ref[0]**o+(0.371314845206)*x_ref[0]+(0.326878867244)*x_ref[1]**o+(-0.468798950199)*x_ref[1]
        ref[0,1,2,0]=(0.392175939547)*x_ref[0]**o+(0.105332682674)*x_ref[0]+(0.997967455477)*x_ref[1]**o+(0.0322491822496)*x_ref[1]
        ref[0,1,2,1]=(-0.804498399187)*x_ref[0]**o+(0.819653744739)*x_ref[0]+(0.609610573703)*x_ref[1]**o+(-0.623299047794)*x_ref[1]
        ref[0,2,0,0]=(0.993877917386)*x_ref[0]**o+(0.815952897899)*x_ref[0]+(-0.258432174486)*x_ref[1]**o+(0.209324387308)*x_ref[1]
        ref[0,2,0,1]=(0.395946889364)*x_ref[0]**o+(-0.859500539908)*x_ref[0]+(-0.68495734785)*x_ref[1]**o+(0.791813946302)*x_ref[1]
        ref[0,2,1,0]=(0.154347661108)*x_ref[0]**o+(0.0873140744322)*x_ref[0]+(-0.423935109956)*x_ref[1]**o+(-0.754297239226)*x_ref[1]
        ref[0,2,1,1]=(-0.623998419484)*x_ref[0]**o+(0.487300130941)*x_ref[0]+(-0.323613219784)*x_ref[1]**o+(0.744877983729)*x_ref[1]
        ref[0,2,2,0]=(0.515686374371)*x_ref[0]**o+(0.512572153346)*x_ref[0]+(-0.641245783272)*x_ref[1]**o+(-0.390660158759)*x_ref[1]
        ref[0,2,2,1]=(0.24398238139)*x_ref[0]**o+(-0.172020271678)*x_ref[0]+(0.980197812701)*x_ref[1]**o+(0.336941934503)*x_ref[1]
        ref[0,3,0,0]=(0.0154584447615)*x_ref[0]**o+(-0.547048191238)*x_ref[0]+(-0.555319223005)*x_ref[1]**o+(0.0972794598106)*x_ref[1]
        ref[0,3,0,1]=(0.44744389335)*x_ref[0]**o+(0.475046337874)*x_ref[0]+(0.876228982066)*x_ref[1]**o+(-0.0636405241175)*x_ref[1]
        ref[0,3,1,0]=(0.392948465298)*x_ref[0]**o+(-0.341342654182)*x_ref[0]+(0.747185683968)*x_ref[1]**o+(-0.99981172166)*x_ref[1]
        ref[0,3,1,1]=(-0.172355321621)*x_ref[0]**o+(-0.531391484795)*x_ref[0]+(0.451654274019)*x_ref[1]**o+(0.0795298667915)*x_ref[1]
        ref[0,3,2,0]=(0.0841642474674)*x_ref[0]**o+(0.308494435164)*x_ref[0]+(0.53181918181)*x_ref[1]**o+(0.480258067959)*x_ref[1]
        ref[0,3,2,1]=(-0.222903054864)*x_ref[0]**o+(-0.0330655869754)*x_ref[0]+(0.227482525015)*x_ref[1]**o+(0.536136212347)*x_ref[1]
        ref[0,4,0,0]=(0.574798381284)*x_ref[0]**o+(0.181354466484)*x_ref[0]+(0.103049514171)*x_ref[1]**o+(0.156282818835)*x_ref[1]
        ref[0,4,0,1]=(0.478314255289)*x_ref[0]**o+(0.224797222592)*x_ref[0]+(0.307668606272)*x_ref[1]**o+(-0.785262182937)*x_ref[1]
        ref[0,4,1,0]=(0.519356961182)*x_ref[0]**o+(0.646200458344)*x_ref[0]+(0.0522618783181)*x_ref[1]**o+(0.702335307324)*x_ref[1]
        ref[0,4,1,1]=(-0.168756029541)*x_ref[0]**o+(-0.00618364750666)*x_ref[0]+(0.472704616213)*x_ref[1]**o+(0.974257451871)*x_ref[1]
        ref[0,4,2,0]=(-0.279196869573)*x_ref[0]**o+(-0.333083174091)*x_ref[0]+(-0.980089574428)*x_ref[1]**o+(0.405273709405)*x_ref[1]
        ref[0,4,2,1]=(-0.199431380673)*x_ref[0]**o+(0.823294899866)*x_ref[0]+(-0.174125117685)*x_ref[1]**o+(-0.730500493887)*x_ref[1]
        ref[1,0,0,0]=(-0.577281489635)*x_ref[0]**o+(0.231906139898)*x_ref[0]+(0.537128708963)*x_ref[1]**o+(0.363444540898)*x_ref[1]
        ref[1,0,0,1]=(0.204108541882)*x_ref[0]**o+(-0.217661714888)*x_ref[0]+(0.963384293759)*x_ref[1]**o+(-0.0817519801351)*x_ref[1]
        ref[1,0,1,0]=(0.540681325439)*x_ref[0]**o+(-0.496036960808)*x_ref[0]+(0.421037689661)*x_ref[1]**o+(0.19399086996)*x_ref[1]
        ref[1,0,1,1]=(0.482938542126)*x_ref[0]**o+(-0.0725027675209)*x_ref[0]+(-0.441517962054)*x_ref[1]**o+(0.787445860714)*x_ref[1]
        ref[1,0,2,0]=(0.0378928111257)*x_ref[0]**o+(0.451368848019)*x_ref[0]+(-0.228668284485)*x_ref[1]**o+(0.505927229432)*x_ref[1]
        ref[1,0,2,1]=(-0.59833518872)*x_ref[0]**o+(0.956068253354)*x_ref[0]+(0.667905995469)*x_ref[1]**o+(0.963273292214)*x_ref[1]
        ref[1,1,0,0]=(0.901637865236)*x_ref[0]**o+(-0.642678971748)*x_ref[0]+(-0.603915019524)*x_ref[1]**o+(-0.208004993919)*x_ref[1]
        ref[1,1,0,1]=(-0.227162661052)*x_ref[0]**o+(-0.5816434303)*x_ref[0]+(0.274961293801)*x_ref[1]**o+(-0.614001328621)*x_ref[1]
        ref[1,1,1,0]=(0.078438078548)*x_ref[0]**o+(0.570191242883)*x_ref[0]+(0.0952878212484)*x_ref[1]**o+(0.844460205497)*x_ref[1]
        ref[1,1,1,1]=(-0.901445811786)*x_ref[0]**o+(-0.996376775344)*x_ref[0]+(0.707470471613)*x_ref[1]**o+(-0.187517936674)*x_ref[1]
        ref[1,1,2,0]=(-0.66891443067)*x_ref[0]**o+(0.0462057239324)*x_ref[0]+(-0.825820539128)*x_ref[1]**o+(-0.766932836843)*x_ref[1]
        ref[1,1,2,1]=(-0.851428828811)*x_ref[0]**o+(0.504484422346)*x_ref[0]+(0.121378315545)*x_ref[1]**o+(0.903947012757)*x_ref[1]
        ref[1,2,0,0]=(-0.393827449768)*x_ref[0]**o+(-0.251827904608)*x_ref[0]+(0.8320040882)*x_ref[1]**o+(0.57595149855)*x_ref[1]
        ref[1,2,0,1]=(0.638738110712)*x_ref[0]**o+(-0.439962152539)*x_ref[0]+(-0.977211078995)*x_ref[1]**o+(0.0528475212451)*x_ref[1]
        ref[1,2,1,0]=(0.824313588941)*x_ref[0]**o+(-0.589968996823)*x_ref[0]+(-0.512156794152)*x_ref[1]**o+(0.142809794443)*x_ref[1]
        ref[1,2,1,1]=(-0.31789333147)*x_ref[0]**o+(0.598839896949)*x_ref[0]+(0.638319398651)*x_ref[1]**o+(0.353625999317)*x_ref[1]
        ref[1,2,2,0]=(-0.720472365012)*x_ref[0]**o+(0.896538030051)*x_ref[0]+(-0.224639169972)*x_ref[1]**o+(-0.898299360738)*x_ref[1]
        ref[1,2,2,1]=(0.123077707167)*x_ref[0]**o+(0.15922851038)*x_ref[0]+(0.678984783843)*x_ref[1]**o+(-0.778188269852)*x_ref[1]
        ref[1,3,0,0]=(0.344256301209)*x_ref[0]**o+(0.884367738289)*x_ref[0]+(0.00556636219289)*x_ref[1]**o+(0.684361331195)*x_ref[1]
        ref[1,3,0,1]=(-0.812245131065)*x_ref[0]**o+(0.971365671228)*x_ref[0]+(0.0696053239569)*x_ref[1]**o+(-0.300266235138)*x_ref[1]
        ref[1,3,1,0]=(0.470254588058)*x_ref[0]**o+(0.0946825509913)*x_ref[0]+(-0.415619255025)*x_ref[1]**o+(-0.27381652858)*x_ref[1]
        ref[1,3,1,1]=(-0.84555138241)*x_ref[0]**o+(0.873510832091)*x_ref[0]+(-0.138010338152)*x_ref[1]**o+(-0.982274515316)*x_ref[1]
        ref[1,3,2,0]=(0.695614136148)*x_ref[0]**o+(-0.192588082771)*x_ref[0]+(-0.21579127791)*x_ref[1]**o+(-0.0335939736388)*x_ref[1]
        ref[1,3,2,1]=(-0.633363463896)*x_ref[0]**o+(-0.958320549993)*x_ref[0]+(0.197331067481)*x_ref[1]**o+(0.315089888409)*x_ref[1]
        ref[1,4,0,0]=(-0.211531219945)*x_ref[0]**o+(-0.399700721799)*x_ref[0]+(0.578927638839)*x_ref[1]**o+(-0.389034976195)*x_ref[1]
        ref[1,4,0,1]=(-0.846399571893)*x_ref[0]**o+(-0.952904755903)*x_ref[0]+(-0.858399802693)*x_ref[1]**o+(-0.23264354907)*x_ref[1]
        ref[1,4,1,0]=(0.255822441586)*x_ref[0]**o+(0.896237710393)*x_ref[0]+(-0.90248008271)*x_ref[1]**o+(0.654804848244)*x_ref[1]
        ref[1,4,1,1]=(-0.112475508163)*x_ref[0]**o+(-0.861697204453)*x_ref[0]+(-0.782581931187)*x_ref[1]**o+(-0.129345341581)*x_ref[1]
        ref[1,4,2,0]=(0.690095687033)*x_ref[0]**o+(0.729317334019)*x_ref[0]+(-0.00961483612288)*x_ref[1]**o+(-0.288320551243)*x_ref[1]
        ref[1,4,2,1]=(-0.648146433517)*x_ref[0]**o+(-0.301246601263)*x_ref[0]+(-0.490119612994)*x_ref[1]**o+(0.231186845615)*x_ref[1]
        ref[2,0,0,0]=(-0.645560891595)*x_ref[0]**o+(-0.889067544988)*x_ref[0]+(-0.193353414221)*x_ref[1]**o+(-0.726988479344)*x_ref[1]
        ref[2,0,0,1]=(-0.083614738484)*x_ref[0]**o+(-0.822530927241)*x_ref[0]+(0.660338934241)*x_ref[1]**o+(0.0383821036508)*x_ref[1]
        ref[2,0,1,0]=(-0.132250736664)*x_ref[0]**o+(-0.571474370931)*x_ref[0]+(-0.755931100342)*x_ref[1]**o+(-0.145196865115)*x_ref[1]
        ref[2,0,1,1]=(0.545219196727)*x_ref[0]**o+(-0.846368787913)*x_ref[0]+(-0.845706160842)*x_ref[1]**o+(0.550703610989)*x_ref[1]
        ref[2,0,2,0]=(-0.766303732195)*x_ref[0]**o+(0.156822893641)*x_ref[0]+(0.00954334254282)*x_ref[1]**o+(0.201506144561)*x_ref[1]
        ref[2,0,2,1]=(-0.274476449632)*x_ref[0]**o+(-0.361690468185)*x_ref[0]+(-0.816535747268)*x_ref[1]**o+(-0.473913811701)*x_ref[1]
        ref[2,1,0,0]=(-0.927592771482)*x_ref[0]**o+(0.534615269814)*x_ref[0]+(-0.571217133452)*x_ref[1]**o+(0.57448299216)*x_ref[1]
        ref[2,1,0,1]=(-0.298153953211)*x_ref[0]**o+(-0.143636909462)*x_ref[0]+(0.990660288864)*x_ref[1]**o+(-0.495864284763)*x_ref[1]
        ref[2,1,1,0]=(-0.110625727208)*x_ref[0]**o+(-0.221057774317)*x_ref[0]+(0.557195078358)*x_ref[1]**o+(0.803162691808)*x_ref[1]
        ref[2,1,1,1]=(-0.115589176775)*x_ref[0]**o+(0.08253351782)*x_ref[0]+(-0.595996958977)*x_ref[1]**o+(0.501435717501)*x_ref[1]
        ref[2,1,2,0]=(0.78764976658)*x_ref[0]**o+(0.966729831711)*x_ref[0]+(-0.585879575667)*x_ref[1]**o+(0.341148176793)*x_ref[1]
        ref[2,1,2,1]=(-0.551697945691)*x_ref[0]**o+(0.545857249323)*x_ref[0]+(-0.919321724495)*x_ref[1]**o+(-0.286043327415)*x_ref[1]
        ref[2,2,0,0]=(-0.669933890655)*x_ref[0]**o+(-0.344862118495)*x_ref[0]+(0.878331282597)*x_ref[1]**o+(0.881893760718)*x_ref[1]
        ref[2,2,0,1]=(-0.217876618414)*x_ref[0]**o+(-0.289118007199)*x_ref[0]+(0.513464171525)*x_ref[1]**o+(0.310055899102)*x_ref[1]
        ref[2,2,1,0]=(0.0472051379154)*x_ref[0]**o+(0.42455388451)*x_ref[0]+(-0.920571834891)*x_ref[1]**o+(-0.779095076537)*x_ref[1]
        ref[2,2,1,1]=(-0.299619304537)*x_ref[0]**o+(-0.289777287294)*x_ref[0]+(0.920505062988)*x_ref[1]**o+(-0.243924179095)*x_ref[1]
        ref[2,2,2,0]=(-0.819343530877)*x_ref[0]**o+(0.975953690875)*x_ref[0]+(-0.787239459241)*x_ref[1]**o+(-0.199900734691)*x_ref[1]
        ref[2,2,2,1]=(0.0980085780878)*x_ref[0]**o+(0.642946515451)*x_ref[0]+(0.951790535152)*x_ref[1]**o+(-0.253914817366)*x_ref[1]
        ref[2,3,0,0]=(-0.882075522824)*x_ref[0]**o+(0.200934502682)*x_ref[0]+(-0.83200464496)*x_ref[1]**o+(0.941479345375)*x_ref[1]
        ref[2,3,0,1]=(-0.615941782209)*x_ref[0]**o+(-0.693242388092)*x_ref[0]+(-0.640417278871)*x_ref[1]**o+(-0.346459094803)*x_ref[1]
        ref[2,3,1,0]=(-0.447919427952)*x_ref[0]**o+(0.205055689141)*x_ref[0]+(-0.453290032411)*x_ref[1]**o+(0.879062836521)*x_ref[1]
        ref[2,3,1,1]=(0.721133924016)*x_ref[0]**o+(0.71294558459)*x_ref[0]+(-0.250087045805)*x_ref[1]**o+(0.605146721129)*x_ref[1]
        ref[2,3,2,0]=(0.247782322921)*x_ref[0]**o+(-0.956367425119)*x_ref[0]+(-0.054045082005)*x_ref[1]**o+(-0.128151832132)*x_ref[1]
        ref[2,3,2,1]=(-0.427039259088)*x_ref[0]**o+(-0.359849242839)*x_ref[0]+(-0.244990934481)*x_ref[1]**o+(-0.038020035629)*x_ref[1]
        ref[2,4,0,0]=(-0.216409150834)*x_ref[0]**o+(0.492547989921)*x_ref[0]+(0.878317438469)*x_ref[1]**o+(-0.112718719119)*x_ref[1]
        ref[2,4,0,1]=(0.487403443102)*x_ref[0]**o+(0.228155519468)*x_ref[0]+(0.923150527516)*x_ref[1]**o+(0.642654867908)*x_ref[1]
        ref[2,4,1,0]=(0.916418483417)*x_ref[0]**o+(-0.177198935173)*x_ref[0]+(-0.790548940412)*x_ref[1]**o+(0.87540167052)*x_ref[1]
        ref[2,4,1,1]=(0.718502075538)*x_ref[0]**o+(0.691254010569)*x_ref[0]+(-0.0134430122729)*x_ref[1]**o+(-0.739888770749)*x_ref[1]
        ref[2,4,2,0]=(0.287825086861)*x_ref[0]**o+(0.596621809347)*x_ref[0]+(0.729784205391)*x_ref[1]**o+(-0.25984257955)*x_ref[1]
        ref[2,4,2,1]=(-0.515102480025)*x_ref[0]**o+(-0.609951840015)*x_ref[0]+(0.229834340075)*x_ref[1]**o+(-0.727157816241)*x_ref[1]
        ref[3,0,0,0]=(-0.349406091363)*x_ref[0]**o+(0.802348451103)*x_ref[0]+(-0.174421346276)*x_ref[1]**o+(0.120711198036)*x_ref[1]
        ref[3,0,0,1]=(0.367085169423)*x_ref[0]**o+(-0.319683236434)*x_ref[0]+(-0.590820542566)*x_ref[1]**o+(0.607625953303)*x_ref[1]
        ref[3,0,1,0]=(0.231259976433)*x_ref[0]**o+(-0.493297408825)*x_ref[0]+(0.919363640504)*x_ref[1]**o+(-0.839841270429)*x_ref[1]
        ref[3,0,1,1]=(0.57672135684)*x_ref[0]**o+(0.882062879704)*x_ref[0]+(-0.600979706444)*x_ref[1]**o+(-0.704008344509)*x_ref[1]
        ref[3,0,2,0]=(-0.208162114527)*x_ref[0]**o+(0.593682243052)*x_ref[0]+(-0.428436832487)*x_ref[1]**o+(0.51007021604)*x_ref[1]
        ref[3,0,2,1]=(0.315264672653)*x_ref[0]**o+(-0.0835897376534)*x_ref[0]+(-0.135315173106)*x_ref[1]**o+(-0.646372309498)*x_ref[1]
        ref[3,1,0,0]=(0.815505953752)*x_ref[0]**o+(0.0625493643292)*x_ref[0]+(0.383118433815)*x_ref[1]**o+(0.17372720435)*x_ref[1]
        ref[3,1,0,1]=(0.557295786969)*x_ref[0]**o+(-0.568823508165)*x_ref[0]+(-0.00874733267675)*x_ref[1]**o+(-0.27454140598)*x_ref[1]
        ref[3,1,1,0]=(0.649021026591)*x_ref[0]**o+(0.232255030448)*x_ref[0]+(0.845939296425)*x_ref[1]**o+(0.116387378294)*x_ref[1]
        ref[3,1,1,1]=(0.431124940343)*x_ref[0]**o+(-0.0865937271969)*x_ref[0]+(0.937911741288)*x_ref[1]**o+(0.362850039234)*x_ref[1]
        ref[3,1,2,0]=(0.155965414439)*x_ref[0]**o+(0.110850421841)*x_ref[0]+(0.244563080191)*x_ref[1]**o+(0.266325498006)*x_ref[1]
        ref[3,1,2,1]=(0.766813300935)*x_ref[0]**o+(0.77492422889)*x_ref[0]+(-0.0141436554714)*x_ref[1]**o+(0.780444449009)*x_ref[1]
        ref[3,2,0,0]=(0.619086019019)*x_ref[0]**o+(-0.574130567601)*x_ref[0]+(-0.417708159593)*x_ref[1]**o+(0.707727284254)*x_ref[1]
        ref[3,2,0,1]=(-0.845657409353)*x_ref[0]**o+(0.705282507748)*x_ref[0]+(0.124557418112)*x_ref[1]**o+(-0.337493494733)*x_ref[1]
        ref[3,2,1,0]=(0.114875146211)*x_ref[0]**o+(-0.724480426892)*x_ref[0]+(-0.0763476904405)*x_ref[1]**o+(0.12064907114)*x_ref[1]
        ref[3,2,1,1]=(-0.74994413376)*x_ref[0]**o+(-0.916476035217)*x_ref[0]+(-0.360412756599)*x_ref[1]**o+(-0.917457363098)*x_ref[1]
        ref[3,2,2,0]=(-0.446839470984)*x_ref[0]**o+(-0.097704838316)*x_ref[0]+(-0.653584364909)*x_ref[1]**o+(0.561014757758)*x_ref[1]
        ref[3,2,2,1]=(-0.178975301038)*x_ref[0]**o+(-0.843669982667)*x_ref[0]+(-0.566610094242)*x_ref[1]**o+(0.213047649003)*x_ref[1]
        ref[3,3,0,0]=(0.144115764355)*x_ref[0]**o+(0.199435411954)*x_ref[0]+(0.173334385199)*x_ref[1]**o+(-0.942957035808)*x_ref[1]
        ref[3,3,0,1]=(-0.417589926342)*x_ref[0]**o+(-0.819188022852)*x_ref[0]+(-0.68299160003)*x_ref[1]**o+(-0.314851447894)*x_ref[1]
        ref[3,3,1,0]=(0.737898219164)*x_ref[0]**o+(0.612906277393)*x_ref[0]+(-0.768349396729)*x_ref[1]**o+(-0.192998755095)*x_ref[1]
        ref[3,3,1,1]=(0.35433362007)*x_ref[0]**o+(0.894174225208)*x_ref[0]+(-0.450333030837)*x_ref[1]**o+(0.819450589483)*x_ref[1]
        ref[3,3,2,0]=(-0.455377131302)*x_ref[0]**o+(-0.748183144687)*x_ref[0]+(-0.482739450231)*x_ref[1]**o+(0.300625929022)*x_ref[1]
        ref[3,3,2,1]=(-0.602766693698)*x_ref[0]**o+(-0.416288436985)*x_ref[0]+(0.837501983536)*x_ref[1]**o+(-0.745935308188)*x_ref[1]
        ref[3,4,0,0]=(-0.706774010405)*x_ref[0]**o+(0.407766818701)*x_ref[0]+(-0.248964401443)*x_ref[1]**o+(0.675466487229)*x_ref[1]
        ref[3,4,0,1]=(0.512302883876)*x_ref[0]**o+(0.0809544867105)*x_ref[0]+(-0.204555203536)*x_ref[1]**o+(-0.628574073749)*x_ref[1]
        ref[3,4,1,0]=(0.126376796421)*x_ref[0]**o+(0.174607814129)*x_ref[0]+(0.68867194136)*x_ref[1]**o+(0.940025934568)*x_ref[1]
        ref[3,4,1,1]=(0.583042444739)*x_ref[0]**o+(-0.786981763836)*x_ref[0]+(0.319397032405)*x_ref[1]**o+(-0.324993351432)*x_ref[1]
        ref[3,4,2,0]=(-0.941997068836)*x_ref[0]**o+(-0.892825520378)*x_ref[0]+(0.275995419027)*x_ref[1]**o+(0.841260061022)*x_ref[1]
        ref[3,4,2,1]=(0.520110030788)*x_ref[0]**o+(0.519262786324)*x_ref[0]+(0.861221788948)*x_ref[1]**o+(-0.376714934937)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.615899511086)*x[0]**o+(0.823637401824)*x[0]+(-0.986967740198)*x[1]**o+(-0.782088977641)*x[1]+(-0.576799974679)*x[2]**o+(-0.910942885711)*x[2]
        arg[0,0,0,1]=(-0.790830593449)*x[0]**o+(-0.897724279499)*x[0]+(-0.881691805807)*x[1]**o+(-0.133793238901)*x[1]+(-0.0903391684489)*x[2]**o+(0.659893469877)*x[2]
        arg[0,0,1,0]=(-0.0567117900114)*x[0]**o+(-0.858031985686)*x[0]+(-0.397443445702)*x[1]**o+(0.549982838823)*x[1]+(-0.636905143813)*x[2]**o+(0.8488142632)*x[2]
        arg[0,0,1,1]=(-0.644557373082)*x[0]**o+(-0.323654898522)*x[0]+(0.210458705923)*x[1]**o+(0.990436827296)*x[1]+(-0.597872757151)*x[2]**o+(-0.83865548618)*x[2]
        arg[0,0,2,0]=(-0.29830667084)*x[0]**o+(0.327035546909)*x[0]+(0.132690049791)*x[1]**o+(-0.55625443282)*x[1]+(0.0972781385283)*x[2]**o+(0.151885837608)*x[2]
        arg[0,0,2,1]=(-0.435983465578)*x[0]**o+(-0.0852522953446)*x[0]+(0.257654807094)*x[1]**o+(-0.336603197583)*x[1]+(0.637913421592)*x[2]**o+(0.8151042055)*x[2]
        arg[0,1,0,0]=(-0.0373980191125)*x[0]**o+(-0.45937285505)*x[0]+(-0.857933684889)*x[1]**o+(-0.843826686256)*x[1]+(-0.727077320504)*x[2]**o+(-0.65222863395)*x[2]
        arg[0,1,0,1]=(0.63426017384)*x[0]**o+(0.845708641601)*x[0]+(0.246160302486)*x[1]**o+(-0.660855572094)*x[1]+(0.67420981328)*x[2]**o+(-0.280803968781)*x[2]
        arg[0,1,1,0]=(0.881193264597)*x[0]**o+(-0.881936969515)*x[0]+(-0.883246837152)*x[1]**o+(0.182222471927)*x[1]+(0.405380496272)*x[2]**o+(0.175721746883)*x[2]
        arg[0,1,1,1]=(-0.44447667237)*x[0]**o+(0.956169274606)*x[0]+(0.0370532940652)*x[1]**o+(-0.181239635583)*x[1]+(-0.678412161285)*x[2]**o+(-0.653172157401)*x[2]
        arg[0,1,2,0]=(-0.50038634718)*x[0]**o+(0.00278321765167)*x[0]+(-0.831539460218)*x[1]**o+(-0.555243320385)*x[1]+(-0.626789452003)*x[2]**o+(-0.495085243933)*x[2]
        arg[0,1,2,1]=(-0.32886359569)*x[0]**o+(0.406099965475)*x[0]+(0.764756177266)*x[1]**o+(-0.272070344181)*x[1]+(-0.41812596516)*x[2]**o+(0.349063696376)*x[2]
        arg[0,2,0,0]=(-0.580544978365)*x[0]**o+(0.53441241334)*x[0]+(-0.808749795876)*x[1]**o+(0.172110801632)*x[1]+(-0.939262606098)*x[2]**o+(-0.724859325104)*x[2]
        arg[0,2,0,1]=(0.256335221944)*x[0]**o+(0.463522371018)*x[0]+(0.990516478508)*x[1]**o+(0.423931177054)*x[1]+(0.184277693683)*x[2]**o+(-0.272967447689)*x[2]
        arg[0,2,1,0]=(-0.126943811008)*x[0]**o+(0.645821387926)*x[0]+(0.150370174987)*x[1]**o+(-0.150954538301)*x[1]+(-0.378720078559)*x[2]**o+(0.0975963807109)*x[2]
        arg[0,2,1,1]=(-0.451434544339)*x[0]**o+(0.557314109088)*x[0]+(-0.0402970749558)*x[1]**o+(0.306984557206)*x[1]+(-0.297018139186)*x[2]**o+(0.593800698785)*x[2]
        arg[0,2,2,0]=(-0.376199144911)*x[0]**o+(-0.230454116373)*x[0]+(0.992738297651)*x[1]**o+(0.183510534927)*x[1]+(0.409381711992)*x[2]**o+(0.171945265727)*x[2]
        arg[0,2,2,1]=(0.363233095971)*x[0]**o+(-0.130589148249)*x[0]+(0.809611042018)*x[1]**o+(-0.354546307961)*x[1]+(0.0631410169203)*x[2]**o+(0.630306161781)*x[2]
        arg[0,3,0,0]=(0.49655016161)*x[0]**o+(0.878957258614)*x[0]+(-0.397408696784)*x[1]**o+(-0.199445321288)*x[1]+(0.929857086663)*x[2]**o+(-0.648770887161)*x[2]
        arg[0,3,0,1]=(-0.85578334594)*x[0]**o+(-0.294617109229)*x[0]+(-0.0790295769768)*x[1]**o+(0.82151942311)*x[1]+(0.541628117444)*x[2]**o+(0.805418174827)*x[2]
        arg[0,3,1,0]=(0.800775454804)*x[0]**o+(0.599550462983)*x[0]+(-0.384788634082)*x[1]**o+(-0.356933972363)*x[1]+(-0.423116265473)*x[2]**o+(-0.125492982003)*x[2]
        arg[0,3,1,1]=(0.929652994661)*x[0]**o+(-0.57396682406)*x[0]+(0.313584116036)*x[1]**o+(-0.190575043496)*x[1]+(-0.29110884935)*x[2]**o+(0.967530639925)*x[2]
        arg[0,3,2,0]=(0.761136478868)*x[0]**o+(0.565721170014)*x[0]+(0.239619122131)*x[1]**o+(-0.0821651979956)*x[1]+(-0.699218775545)*x[2]**o+(0.19973415585)*x[2]
        arg[0,3,2,1]=(0.209543244648)*x[0]**o+(0.93957796874)*x[0]+(0.354405887426)*x[1]**o+(0.504177262864)*x[1]+(0.446866674894)*x[2]**o+(-0.522936485464)*x[2]
        arg[0,4,0,0]=(0.427541399652)*x[0]**o+(-0.114806067206)*x[0]+(0.236780877643)*x[1]**o+(0.422226381272)*x[1]+(-0.506450961214)*x[2]**o+(0.616091720212)*x[2]
        arg[0,4,0,1]=(0.698773571323)*x[0]**o+(-0.357878713085)*x[0]+(-0.453730359129)*x[1]**o+(-0.328701041355)*x[1]+(-0.28646346679)*x[2]**o+(-0.0449784265203)*x[2]
        arg[0,4,1,0]=(-0.419416306509)*x[0]**o+(-0.1970394955)*x[0]+(-0.814008932379)*x[1]**o+(-0.805127589929)*x[1]+(0.244059789875)*x[2]**o+(-0.983027843847)*x[2]
        arg[0,4,1,1]=(0.84837346004)*x[0]**o+(0.501808510624)*x[0]+(-0.105800351358)*x[1]**o+(0.903999231718)*x[1]+(-0.0473996978124)*x[2]**o+(0.683647253381)*x[2]
        arg[0,4,2,0]=(0.658722899682)*x[0]**o+(-0.401846635016)*x[0]+(-0.967322202581)*x[1]**o+(-0.948306922217)*x[1]+(0.496409308219)*x[2]**o+(-0.184768777052)*x[2]
        arg[0,4,2,1]=(-0.454468390888)*x[0]**o+(-0.220640468215)*x[0]+(0.995269925943)*x[1]**o+(0.778951329211)*x[1]+(0.0439534466302)*x[2]**o+(0.806197038383)*x[2]
        arg[1,0,0,0]=(-0.547461802236)*x[0]**o+(0.370768827789)*x[0]+(-0.619991757552)*x[1]**o+(0.391727195645)*x[1]+(-0.649274922953)*x[2]**o+(-0.478788248429)*x[2]
        arg[1,0,0,1]=(-0.899230505882)*x[0]**o+(0.603064321735)*x[0]+(-0.737475386349)*x[1]**o+(0.594197953001)*x[1]+(0.341324049064)*x[2]**o+(0.285202633917)*x[2]
        arg[1,0,1,0]=(-0.107063589618)*x[0]**o+(-0.260491562358)*x[0]+(-0.667059815618)*x[1]**o+(0.473227354537)*x[1]+(-0.587170510558)*x[2]**o+(-0.740299218393)*x[2]
        arg[1,0,1,1]=(0.87705715098)*x[0]**o+(0.312753471827)*x[0]+(-0.307530283315)*x[1]**o+(0.195564644822)*x[1]+(-0.64384485166)*x[2]**o+(-0.776362951729)*x[2]
        arg[1,0,2,0]=(0.0949408564739)*x[0]**o+(0.765922179855)*x[0]+(0.441215568253)*x[1]**o+(0.339112969437)*x[1]+(-0.0774615292449)*x[2]**o+(0.673258055783)*x[2]
        arg[1,0,2,1]=(0.114828212302)*x[0]**o+(0.0831258766277)*x[0]+(0.674540402248)*x[1]**o+(-0.304921039094)*x[1]+(-0.725514770578)*x[2]**o+(-0.627584773746)*x[2]
        arg[1,1,0,0]=(0.291561714675)*x[0]**o+(0.898587624783)*x[0]+(0.476696264545)*x[1]**o+(-0.454648604978)*x[1]+(0.677949150743)*x[2]**o+(-0.59438289851)*x[2]
        arg[1,1,0,1]=(0.356687258009)*x[0]**o+(0.806236251629)*x[0]+(0.843636690354)*x[1]**o+(-0.649776947181)*x[1]+(0.208542122388)*x[2]**o+(-0.0726045573458)*x[2]
        arg[1,1,1,0]=(0.712800450558)*x[0]**o+(0.748271875755)*x[0]+(0.403880649356)*x[1]**o+(0.956549336754)*x[1]+(0.0420128495043)*x[2]**o+(-0.83177870499)*x[2]
        arg[1,1,1,1]=(-0.886862904174)*x[0]**o+(-0.03822986594)*x[0]+(0.587151517496)*x[1]**o+(-0.120272060898)*x[1]+(-0.573366391933)*x[2]**o+(-0.59772763241)*x[2]
        arg[1,1,2,0]=(0.311960197804)*x[0]**o+(0.486854949444)*x[0]+(-0.647995651954)*x[1]**o+(0.30104277964)*x[1]+(-0.174718703748)*x[2]**o+(-0.506993216295)*x[2]
        arg[1,1,2,1]=(0.979056206991)*x[0]**o+(0.892104464812)*x[0]+(0.0507432703245)*x[1]**o+(0.103630858454)*x[1]+(-0.0254965952266)*x[2]**o+(0.13034027239)*x[2]
        arg[1,2,0,0]=(0.259797966548)*x[0]**o+(0.53018999371)*x[0]+(0.533745722992)*x[1]**o+(0.53029084173)*x[1]+(-0.595126791061)*x[2]**o+(-0.51668295558)*x[2]
        arg[1,2,0,1]=(0.926395590808)*x[0]**o+(-0.0700141207179)*x[0]+(-0.973392454012)*x[1]**o+(-0.46204067097)*x[1]+(0.814964497207)*x[2]**o+(-0.183546091376)*x[2]
        arg[1,2,1,0]=(-0.0872635441658)*x[0]**o+(-0.0307630839035)*x[0]+(0.539798653095)*x[1]**o+(0.966367734848)*x[1]+(-0.817495053612)*x[2]**o+(-0.268397867848)*x[2]
        arg[1,2,1,1]=(-0.933074613921)*x[0]**o+(0.48896533938)*x[0]+(0.453696133488)*x[1]**o+(-0.238730406749)*x[1]+(-0.696785951086)*x[2]**o+(-0.359858385642)*x[2]
        arg[1,2,2,0]=(-0.844406286809)*x[0]**o+(-0.506332243946)*x[0]+(0.976436731658)*x[1]**o+(-0.201489161857)*x[1]+(0.851236173166)*x[2]**o+(-0.998803612947)*x[2]
        arg[1,2,2,1]=(0.259707321449)*x[0]**o+(-0.888149478655)*x[0]+(-0.273502842869)*x[1]**o+(0.291193582158)*x[1]+(-0.897542178027)*x[2]**o+(-0.345994824429)*x[2]
        arg[1,3,0,0]=(0.961298984697)*x[0]**o+(0.248141115092)*x[0]+(-0.623053025178)*x[1]**o+(-0.775095452078)*x[1]+(0.955362707049)*x[2]**o+(-0.517547333705)*x[2]
        arg[1,3,0,1]=(0.858098596227)*x[0]**o+(0.309307483873)*x[0]+(-0.209217034755)*x[1]**o+(0.783803593379)*x[1]+(0.138164128092)*x[2]**o+(-0.922854003461)*x[2]
        arg[1,3,1,0]=(-0.139017942141)*x[0]**o+(-0.338223039119)*x[0]+(0.737882941498)*x[1]**o+(0.948854903087)*x[1]+(-0.0792461633266)*x[2]**o+(0.848875660203)*x[2]
        arg[1,3,1,1]=(0.693284332212)*x[0]**o+(-0.392272875348)*x[0]+(-0.475709045479)*x[1]**o+(0.182948077236)*x[1]+(0.0277569761538)*x[2]**o+(-0.550294947596)*x[2]
        arg[1,3,2,0]=(-0.105381649505)*x[0]**o+(-0.403051690238)*x[0]+(-0.722219735818)*x[1]**o+(-0.380114388869)*x[1]+(0.321452572491)*x[2]**o+(0.858583245253)*x[2]
        arg[1,3,2,1]=(-0.405370062024)*x[0]**o+(0.0743704840047)*x[0]+(0.998947017815)*x[1]**o+(0.983611079522)*x[1]+(0.854202982674)*x[2]**o+(0.288478746498)*x[2]
        arg[1,4,0,0]=(0.0584782489998)*x[0]**o+(0.999286927604)*x[0]+(-0.458166020555)*x[1]**o+(0.529008844128)*x[1]+(0.424592641922)*x[2]**o+(-0.489692340187)*x[2]
        arg[1,4,0,1]=(-0.150555966138)*x[0]**o+(0.362589320838)*x[0]+(0.515820915391)*x[1]**o+(0.742403661312)*x[1]+(0.230278843183)*x[2]**o+(-0.213977724366)*x[2]
        arg[1,4,1,0]=(-0.361161802362)*x[0]**o+(-0.49718332222)*x[0]+(0.143956046197)*x[1]**o+(-0.044550912758)*x[1]+(-0.299892874708)*x[2]**o+(0.370248709792)*x[2]
        arg[1,4,1,1]=(0.259365238173)*x[0]**o+(-0.474694548557)*x[0]+(0.732983396669)*x[1]**o+(-0.295151546774)*x[1]+(-0.337144628526)*x[2]**o+(0.664201775714)*x[2]
        arg[1,4,2,0]=(-0.318016754284)*x[0]**o+(0.555291443852)*x[0]+(0.538150727998)*x[1]**o+(-0.800118353132)*x[1]+(0.652262386596)*x[2]**o+(0.303156645857)*x[2]
        arg[1,4,2,1]=(0.0827142921796)*x[0]**o+(-0.88498648372)*x[0]+(-0.941249844349)*x[1]**o+(-0.162576659039)*x[1]+(0.867621252192)*x[2]**o+(0.0834583795227)*x[2]
        arg[2,0,0,0]=(0.0670088199264)*x[0]**o+(-0.0425961894346)*x[0]+(0.329605789302)*x[1]**o+(0.860453156363)*x[1]+(-0.37410469727)*x[2]**o+(0.55476432384)*x[2]
        arg[2,0,0,1]=(0.801738221879)*x[0]**o+(-0.836563530923)*x[0]+(-0.449651673142)*x[1]**o+(-0.19423133918)*x[1]+(0.947648200044)*x[2]**o+(0.0854067601423)*x[2]
        arg[2,0,1,0]=(0.986595046855)*x[0]**o+(-0.358319352175)*x[0]+(0.493917713513)*x[1]**o+(-0.92052282445)*x[1]+(-0.692726467259)*x[2]**o+(-0.0636943737807)*x[2]
        arg[2,0,1,1]=(0.38491150492)*x[0]**o+(0.766893245938)*x[0]+(0.420188917945)*x[1]**o+(0.43760538854)*x[1]+(-0.770749451981)*x[2]**o+(0.124882164378)*x[2]
        arg[2,0,2,0]=(0.775715718797)*x[0]**o+(0.89516703208)*x[0]+(-0.829021524898)*x[1]**o+(-0.650262454311)*x[1]+(-0.812556336598)*x[2]**o+(-0.997075743997)*x[2]
        arg[2,0,2,1]=(0.3503761008)*x[0]**o+(-0.822009897455)*x[0]+(-0.657657365354)*x[1]**o+(0.476295612156)*x[1]+(0.315554008711)*x[2]**o+(0.691631078049)*x[2]
        arg[2,1,0,0]=(0.680177375433)*x[0]**o+(-0.922726813311)*x[0]+(-0.581116746321)*x[1]**o+(0.628138099343)*x[1]+(-0.568262800293)*x[2]**o+(-0.228201382395)*x[2]
        arg[2,1,0,1]=(0.210094833324)*x[0]**o+(0.281549673876)*x[0]+(0.23785231073)*x[1]**o+(-0.103812455993)*x[1]+(0.530445971191)*x[2]**o+(-0.0830520449939)*x[2]
        arg[2,1,1,0]=(0.367133297563)*x[0]**o+(0.168289954585)*x[0]+(0.614117392787)*x[1]**o+(0.790467766841)*x[1]+(-0.958002150805)*x[2]**o+(0.328916248824)*x[2]
        arg[2,1,1,1]=(0.089535347231)*x[0]**o+(-0.93155993701)*x[0]+(-0.674670760794)*x[1]**o+(-0.274148261773)*x[1]+(-0.807020126056)*x[2]**o+(-0.323334769496)*x[2]
        arg[2,1,2,0]=(-0.948679835219)*x[0]**o+(-0.577235742425)*x[0]+(-0.732584834146)*x[1]**o+(-0.475242504288)*x[1]+(0.213663223262)*x[2]**o+(-0.151401354381)*x[2]
        arg[2,1,2,1]=(0.264513246206)*x[0]**o+(-0.605806437391)*x[0]+(-0.058730926693)*x[1]**o+(-0.661696304412)*x[1]+(-0.361277352457)*x[2]**o+(-0.804408258253)*x[2]
        arg[2,2,0,0]=(-0.144562358054)*x[0]**o+(-0.0960550555725)*x[0]+(-0.905096867803)*x[1]**o+(0.188861784103)*x[1]+(-0.616751005568)*x[2]**o+(0.604711974248)*x[2]
        arg[2,2,0,1]=(-0.0582524307858)*x[0]**o+(-0.503648879785)*x[0]+(-0.682755209956)*x[1]**o+(-0.808213164425)*x[1]+(0.539987276327)*x[2]**o+(0.650557703689)*x[2]
        arg[2,2,1,0]=(0.0607729567733)*x[0]**o+(-0.498819638428)*x[0]+(0.434442595334)*x[1]**o+(-0.491063977603)*x[1]+(0.499436212229)*x[2]**o+(-0.932071991206)*x[2]
        arg[2,2,1,1]=(-0.461295295294)*x[0]**o+(-0.153600767916)*x[0]+(-0.86452392231)*x[1]**o+(0.136760034153)*x[1]+(-0.471642416082)*x[2]**o+(0.453820403438)*x[2]
        arg[2,2,2,0]=(-0.970051888584)*x[0]**o+(0.415728825057)*x[0]+(0.971643909569)*x[1]**o+(0.923719420524)*x[1]+(-0.837313513804)*x[2]**o+(-0.738057676093)*x[2]
        arg[2,2,2,1]=(-0.941355918089)*x[0]**o+(-0.9629784524)*x[0]+(-0.931184348072)*x[1]**o+(0.720006097658)*x[1]+(-0.140750567978)*x[2]**o+(0.55381939119)*x[2]
        arg[2,3,0,0]=(0.753455933669)*x[0]**o+(0.352384922037)*x[0]+(0.737644348002)*x[1]**o+(-0.565144188248)*x[1]+(0.257417624014)*x[2]**o+(-0.660006641683)*x[2]
        arg[2,3,0,1]=(0.886996770925)*x[0]**o+(-0.876595418252)*x[0]+(-0.189778073871)*x[1]**o+(0.560867442336)*x[1]+(-0.262338156346)*x[2]**o+(0.614749098901)*x[2]
        arg[2,3,1,0]=(-0.722912374016)*x[0]**o+(0.447451310273)*x[0]+(0.397510140568)*x[1]**o+(0.932405917932)*x[1]+(-0.206969252336)*x[2]**o+(-0.897366208888)*x[2]
        arg[2,3,1,1]=(0.106324074547)*x[0]**o+(0.822109425884)*x[0]+(0.62699223277)*x[1]**o+(0.238371557839)*x[1]+(-0.422051841243)*x[2]**o+(0.868932130329)*x[2]
        arg[2,3,2,0]=(0.0155790371581)*x[0]**o+(-0.696029577918)*x[0]+(0.259909585217)*x[1]**o+(0.260209121862)*x[1]+(0.780977081459)*x[2]**o+(-0.819957534538)*x[2]
        arg[2,3,2,1]=(0.736223721904)*x[0]**o+(-0.433132598174)*x[0]+(-0.880913649812)*x[1]**o+(-0.375988777461)*x[1]+(-0.351763406307)*x[2]**o+(0.948468252467)*x[2]
        arg[2,4,0,0]=(0.943349180556)*x[0]**o+(-0.720044514103)*x[0]+(0.671900161807)*x[1]**o+(0.358571142942)*x[1]+(-0.132755998648)*x[2]**o+(0.0749815044365)*x[2]
        arg[2,4,0,1]=(0.639070752347)*x[0]**o+(0.551441363406)*x[0]+(0.507547844432)*x[1]**o+(-0.800043772307)*x[1]+(-0.506359503394)*x[2]**o+(0.952191432651)*x[2]
        arg[2,4,1,0]=(-0.655864732254)*x[0]**o+(0.142198390642)*x[0]+(0.075292486702)*x[1]**o+(0.191013189029)*x[1]+(0.106768453247)*x[2]**o+(0.974526515449)*x[2]
        arg[2,4,1,1]=(-0.354750205648)*x[0]**o+(0.141506849368)*x[0]+(-0.170074898141)*x[1]**o+(-0.0882776529335)*x[1]+(0.845254802876)*x[2]**o+(-0.435414692125)*x[2]
        arg[2,4,2,0]=(-0.412836027163)*x[0]**o+(0.830777132582)*x[0]+(-0.87554952236)*x[1]**o+(0.970647516313)*x[1]+(0.989864827286)*x[2]**o+(-0.67790755421)*x[2]
        arg[2,4,2,1]=(0.762907045031)*x[0]**o+(0.140771931118)*x[0]+(0.716859940142)*x[1]**o+(0.151769314189)*x[1]+(-0.812408853675)*x[2]**o+(0.755727078666)*x[2]
        arg[3,0,0,0]=(-0.13358237438)*x[0]**o+(-0.594735856789)*x[0]+(-0.367418402298)*x[1]**o+(0.5349162872)*x[1]+(0.238285320118)*x[2]**o+(0.999627906459)*x[2]
        arg[3,0,0,1]=(-0.537433369218)*x[0]**o+(0.133034652667)*x[0]+(-0.190302464596)*x[1]**o+(0.869919012762)*x[1]+(0.320294306008)*x[2]**o+(0.865908856629)*x[2]
        arg[3,0,1,0]=(0.480851043043)*x[0]**o+(0.833358259022)*x[0]+(0.893051476621)*x[1]**o+(-0.505255581199)*x[1]+(-0.15316145317)*x[2]**o+(-0.937103563776)*x[2]
        arg[3,0,1,1]=(0.765360137605)*x[0]**o+(0.502399552842)*x[0]+(-0.215707512276)*x[1]**o+(-0.402720816967)*x[1]+(0.231360098578)*x[2]**o+(0.732903327067)*x[2]
        arg[3,0,2,0]=(0.120836669045)*x[0]**o+(0.198346794548)*x[0]+(-0.28708447536)*x[1]**o+(0.289672075353)*x[1]+(-0.0305777738322)*x[2]**o+(-0.0686651569373)*x[2]
        arg[3,0,2,1]=(0.0574598844614)*x[0]**o+(0.031092492399)*x[0]+(-0.406805507341)*x[1]**o+(-0.623370509947)*x[1]+(-0.476094160966)*x[2]**o+(0.0840271758875)*x[2]
        arg[3,1,0,0]=(-0.569415249126)*x[0]**o+(-0.339675540399)*x[0]+(0.182536980707)*x[1]**o+(0.727480200453)*x[1]+(0.851672551504)*x[2]**o+(0.817312313075)*x[2]
        arg[3,1,0,1]=(0.921422642904)*x[0]**o+(0.697656143906)*x[0]+(0.969442847752)*x[1]**o+(-0.00957994710064)*x[1]+(-0.315230556399)*x[2]**o+(-0.870759158384)*x[2]
        arg[3,1,1,0]=(-0.813310465463)*x[0]**o+(-0.809362802903)*x[0]+(-0.0685101056428)*x[1]**o+(0.708195651847)*x[1]+(-0.876449456312)*x[2]**o+(0.882503536479)*x[2]
        arg[3,1,1,1]=(-0.277504922905)*x[0]**o+(-0.819428706347)*x[0]+(0.463203258731)*x[1]**o+(0.597493096401)*x[1]+(0.630522187524)*x[2]**o+(0.92480079466)*x[2]
        arg[3,1,2,0]=(-0.703167803764)*x[0]**o+(0.730459205382)*x[0]+(-0.507036921123)*x[1]**o+(0.419358545765)*x[1]+(-0.542758923679)*x[2]**o+(0.831844312917)*x[2]
        arg[3,1,2,1]=(0.135200299704)*x[0]**o+(-0.160734675188)*x[0]+(-0.893851689192)*x[1]**o+(0.748734906185)*x[1]+(0.970464447167)*x[2]**o+(0.277911376757)*x[2]
        arg[3,2,0,0]=(0.750104328784)*x[0]**o+(-0.818350593893)*x[0]+(-0.749900854281)*x[1]**o+(0.289117653816)*x[1]+(-0.219904822162)*x[2]**o+(0.805208050452)*x[2]
        arg[3,2,0,1]=(-0.129614389749)*x[0]**o+(-0.64067443816)*x[0]+(0.513013650233)*x[1]**o+(0.868869035341)*x[1]+(-0.040145121639)*x[2]**o+(-0.282350136119)*x[2]
        arg[3,2,1,0]=(-0.172814929104)*x[0]**o+(-0.296930471915)*x[0]+(-0.796114111585)*x[1]**o+(-0.210181080249)*x[1]+(-0.606107841173)*x[2]**o+(-0.954329084175)*x[2]
        arg[3,2,1,1]=(0.136149261959)*x[0]**o+(-0.613980696755)*x[0]+(0.968430714778)*x[1]**o+(0.683553952099)*x[1]+(-0.321383789121)*x[2]**o+(-0.139570984234)*x[2]
        arg[3,2,2,0]=(-0.273321720654)*x[0]**o+(0.963651110162)*x[0]+(-0.847402611499)*x[1]**o+(0.237131262905)*x[1]+(-0.617296821558)*x[2]**o+(-0.681856394629)*x[2]
        arg[3,2,2,1]=(-0.760071928624)*x[0]**o+(0.0345053779755)*x[0]+(-0.583680069732)*x[1]**o+(0.150497658077)*x[1]+(0.0257096621429)*x[2]**o+(-0.159406203385)*x[2]
        arg[3,3,0,0]=(-0.0668501746947)*x[0]**o+(-0.183645899611)*x[0]+(0.123667845765)*x[1]**o+(0.0928142420352)*x[1]+(-0.933250766317)*x[2]**o+(-0.629986949639)*x[2]
        arg[3,3,0,1]=(-0.0212790928206)*x[0]**o+(0.444036360325)*x[0]+(0.09182358127)*x[1]**o+(0.473088563672)*x[1]+(0.0852925382953)*x[2]**o+(-0.238517929786)*x[2]
        arg[3,3,1,0]=(0.0647235093728)*x[0]**o+(-0.266285898716)*x[0]+(0.944999560866)*x[1]**o+(-0.757696716486)*x[1]+(-0.439165278732)*x[2]**o+(0.960091479774)*x[2]
        arg[3,3,1,1]=(-0.792249238894)*x[0]**o+(-0.568585439587)*x[0]+(-0.0426655335607)*x[1]**o+(0.980004032206)*x[1]+(0.635227812889)*x[2]**o+(0.6687165384)*x[2]
        arg[3,3,2,0]=(0.340468442769)*x[0]**o+(0.791071154735)*x[0]+(-0.817919975547)*x[1]**o+(0.0328383066982)*x[1]+(-0.323093904705)*x[2]**o+(-0.302641620002)*x[2]
        arg[3,3,2,1]=(-0.476501733213)*x[0]**o+(-0.429845943537)*x[0]+(0.518551215268)*x[1]**o+(-0.551690143654)*x[1]+(0.591983049997)*x[2]**o+(0.586292683497)*x[2]
        arg[3,4,0,0]=(0.119036526248)*x[0]**o+(-0.156404044048)*x[0]+(0.158504048681)*x[1]**o+(-0.3779600709)*x[1]+(-0.362107671534)*x[2]**o+(-0.314508628804)*x[2]
        arg[3,4,0,1]=(0.382011103106)*x[0]**o+(-0.123711676964)*x[0]+(-0.939014457264)*x[1]**o+(-0.5986155807)*x[1]+(-0.852092491566)*x[2]**o+(-0.410438924788)*x[2]
        arg[3,4,1,0]=(0.83612586535)*x[0]**o+(-0.606098645092)*x[0]+(0.643469717148)*x[1]**o+(0.845106525654)*x[1]+(0.347455649691)*x[2]**o+(-0.166796580458)*x[2]
        arg[3,4,1,1]=(0.693431758156)*x[0]**o+(-0.313983401147)*x[0]+(-0.862690834893)*x[1]**o+(0.417233988055)*x[1]+(-0.361769997589)*x[2]**o+(-0.22749888397)*x[2]
        arg[3,4,2,0]=(0.680358405387)*x[0]**o+(-0.786041842215)*x[0]+(0.288823579415)*x[1]**o+(-0.961099851196)*x[1]+(0.669431785695)*x[2]**o+(0.956564160268)*x[2]
        arg[3,4,2,1]=(-0.21416500005)*x[0]**o+(0.48495191486)*x[0]+(-0.173465452645)*x[1]**o+(0.662530321557)*x[1]+(0.984573283824)*x[2]**o+(-0.568211383785)*x[2]
        ref[0,0,0,0]=(-0.615899511086)*x_ref[0]**o+(0.823637401824)*x_ref[0]+(-0.986967740198)*x_ref[1]**o+(-0.782088977641)*x_ref[1]+(-0.576799974679)*x_ref[2]**o+(-0.910942885711)*x_ref[2]
        ref[0,0,0,1]=(-0.790830593449)*x_ref[0]**o+(-0.897724279499)*x_ref[0]+(-0.881691805807)*x_ref[1]**o+(-0.133793238901)*x_ref[1]+(-0.0903391684489)*x_ref[2]**o+(0.659893469877)*x_ref[2]
        ref[0,0,1,0]=(-0.0567117900114)*x_ref[0]**o+(-0.858031985686)*x_ref[0]+(-0.397443445702)*x_ref[1]**o+(0.549982838823)*x_ref[1]+(-0.636905143813)*x_ref[2]**o+(0.8488142632)*x_ref[2]
        ref[0,0,1,1]=(-0.644557373082)*x_ref[0]**o+(-0.323654898522)*x_ref[0]+(0.210458705923)*x_ref[1]**o+(0.990436827296)*x_ref[1]+(-0.597872757151)*x_ref[2]**o+(-0.83865548618)*x_ref[2]
        ref[0,0,2,0]=(-0.29830667084)*x_ref[0]**o+(0.327035546909)*x_ref[0]+(0.132690049791)*x_ref[1]**o+(-0.55625443282)*x_ref[1]+(0.0972781385283)*x_ref[2]**o+(0.151885837608)*x_ref[2]
        ref[0,0,2,1]=(-0.435983465578)*x_ref[0]**o+(-0.0852522953446)*x_ref[0]+(0.257654807094)*x_ref[1]**o+(-0.336603197583)*x_ref[1]+(0.637913421592)*x_ref[2]**o+(0.8151042055)*x_ref[2]
        ref[0,1,0,0]=(-0.0373980191125)*x_ref[0]**o+(-0.45937285505)*x_ref[0]+(-0.857933684889)*x_ref[1]**o+(-0.843826686256)*x_ref[1]+(-0.727077320504)*x_ref[2]**o+(-0.65222863395)*x_ref[2]
        ref[0,1,0,1]=(0.63426017384)*x_ref[0]**o+(0.845708641601)*x_ref[0]+(0.246160302486)*x_ref[1]**o+(-0.660855572094)*x_ref[1]+(0.67420981328)*x_ref[2]**o+(-0.280803968781)*x_ref[2]
        ref[0,1,1,0]=(0.881193264597)*x_ref[0]**o+(-0.881936969515)*x_ref[0]+(-0.883246837152)*x_ref[1]**o+(0.182222471927)*x_ref[1]+(0.405380496272)*x_ref[2]**o+(0.175721746883)*x_ref[2]
        ref[0,1,1,1]=(-0.44447667237)*x_ref[0]**o+(0.956169274606)*x_ref[0]+(0.0370532940652)*x_ref[1]**o+(-0.181239635583)*x_ref[1]+(-0.678412161285)*x_ref[2]**o+(-0.653172157401)*x_ref[2]
        ref[0,1,2,0]=(-0.50038634718)*x_ref[0]**o+(0.00278321765167)*x_ref[0]+(-0.831539460218)*x_ref[1]**o+(-0.555243320385)*x_ref[1]+(-0.626789452003)*x_ref[2]**o+(-0.495085243933)*x_ref[2]
        ref[0,1,2,1]=(-0.32886359569)*x_ref[0]**o+(0.406099965475)*x_ref[0]+(0.764756177266)*x_ref[1]**o+(-0.272070344181)*x_ref[1]+(-0.41812596516)*x_ref[2]**o+(0.349063696376)*x_ref[2]
        ref[0,2,0,0]=(-0.580544978365)*x_ref[0]**o+(0.53441241334)*x_ref[0]+(-0.808749795876)*x_ref[1]**o+(0.172110801632)*x_ref[1]+(-0.939262606098)*x_ref[2]**o+(-0.724859325104)*x_ref[2]
        ref[0,2,0,1]=(0.256335221944)*x_ref[0]**o+(0.463522371018)*x_ref[0]+(0.990516478508)*x_ref[1]**o+(0.423931177054)*x_ref[1]+(0.184277693683)*x_ref[2]**o+(-0.272967447689)*x_ref[2]
        ref[0,2,1,0]=(-0.126943811008)*x_ref[0]**o+(0.645821387926)*x_ref[0]+(0.150370174987)*x_ref[1]**o+(-0.150954538301)*x_ref[1]+(-0.378720078559)*x_ref[2]**o+(0.0975963807109)*x_ref[2]
        ref[0,2,1,1]=(-0.451434544339)*x_ref[0]**o+(0.557314109088)*x_ref[0]+(-0.0402970749558)*x_ref[1]**o+(0.306984557206)*x_ref[1]+(-0.297018139186)*x_ref[2]**o+(0.593800698785)*x_ref[2]
        ref[0,2,2,0]=(-0.376199144911)*x_ref[0]**o+(-0.230454116373)*x_ref[0]+(0.992738297651)*x_ref[1]**o+(0.183510534927)*x_ref[1]+(0.409381711992)*x_ref[2]**o+(0.171945265727)*x_ref[2]
        ref[0,2,2,1]=(0.363233095971)*x_ref[0]**o+(-0.130589148249)*x_ref[0]+(0.809611042018)*x_ref[1]**o+(-0.354546307961)*x_ref[1]+(0.0631410169203)*x_ref[2]**o+(0.630306161781)*x_ref[2]
        ref[0,3,0,0]=(0.49655016161)*x_ref[0]**o+(0.878957258614)*x_ref[0]+(-0.397408696784)*x_ref[1]**o+(-0.199445321288)*x_ref[1]+(0.929857086663)*x_ref[2]**o+(-0.648770887161)*x_ref[2]
        ref[0,3,0,1]=(-0.85578334594)*x_ref[0]**o+(-0.294617109229)*x_ref[0]+(-0.0790295769768)*x_ref[1]**o+(0.82151942311)*x_ref[1]+(0.541628117444)*x_ref[2]**o+(0.805418174827)*x_ref[2]
        ref[0,3,1,0]=(0.800775454804)*x_ref[0]**o+(0.599550462983)*x_ref[0]+(-0.384788634082)*x_ref[1]**o+(-0.356933972363)*x_ref[1]+(-0.423116265473)*x_ref[2]**o+(-0.125492982003)*x_ref[2]
        ref[0,3,1,1]=(0.929652994661)*x_ref[0]**o+(-0.57396682406)*x_ref[0]+(0.313584116036)*x_ref[1]**o+(-0.190575043496)*x_ref[1]+(-0.29110884935)*x_ref[2]**o+(0.967530639925)*x_ref[2]
        ref[0,3,2,0]=(0.761136478868)*x_ref[0]**o+(0.565721170014)*x_ref[0]+(0.239619122131)*x_ref[1]**o+(-0.0821651979956)*x_ref[1]+(-0.699218775545)*x_ref[2]**o+(0.19973415585)*x_ref[2]
        ref[0,3,2,1]=(0.209543244648)*x_ref[0]**o+(0.93957796874)*x_ref[0]+(0.354405887426)*x_ref[1]**o+(0.504177262864)*x_ref[1]+(0.446866674894)*x_ref[2]**o+(-0.522936485464)*x_ref[2]
        ref[0,4,0,0]=(0.427541399652)*x_ref[0]**o+(-0.114806067206)*x_ref[0]+(0.236780877643)*x_ref[1]**o+(0.422226381272)*x_ref[1]+(-0.506450961214)*x_ref[2]**o+(0.616091720212)*x_ref[2]
        ref[0,4,0,1]=(0.698773571323)*x_ref[0]**o+(-0.357878713085)*x_ref[0]+(-0.453730359129)*x_ref[1]**o+(-0.328701041355)*x_ref[1]+(-0.28646346679)*x_ref[2]**o+(-0.0449784265203)*x_ref[2]
        ref[0,4,1,0]=(-0.419416306509)*x_ref[0]**o+(-0.1970394955)*x_ref[0]+(-0.814008932379)*x_ref[1]**o+(-0.805127589929)*x_ref[1]+(0.244059789875)*x_ref[2]**o+(-0.983027843847)*x_ref[2]
        ref[0,4,1,1]=(0.84837346004)*x_ref[0]**o+(0.501808510624)*x_ref[0]+(-0.105800351358)*x_ref[1]**o+(0.903999231718)*x_ref[1]+(-0.0473996978124)*x_ref[2]**o+(0.683647253381)*x_ref[2]
        ref[0,4,2,0]=(0.658722899682)*x_ref[0]**o+(-0.401846635016)*x_ref[0]+(-0.967322202581)*x_ref[1]**o+(-0.948306922217)*x_ref[1]+(0.496409308219)*x_ref[2]**o+(-0.184768777052)*x_ref[2]
        ref[0,4,2,1]=(-0.454468390888)*x_ref[0]**o+(-0.220640468215)*x_ref[0]+(0.995269925943)*x_ref[1]**o+(0.778951329211)*x_ref[1]+(0.0439534466302)*x_ref[2]**o+(0.806197038383)*x_ref[2]
        ref[1,0,0,0]=(-0.547461802236)*x_ref[0]**o+(0.370768827789)*x_ref[0]+(-0.619991757552)*x_ref[1]**o+(0.391727195645)*x_ref[1]+(-0.649274922953)*x_ref[2]**o+(-0.478788248429)*x_ref[2]
        ref[1,0,0,1]=(-0.899230505882)*x_ref[0]**o+(0.603064321735)*x_ref[0]+(-0.737475386349)*x_ref[1]**o+(0.594197953001)*x_ref[1]+(0.341324049064)*x_ref[2]**o+(0.285202633917)*x_ref[2]
        ref[1,0,1,0]=(-0.107063589618)*x_ref[0]**o+(-0.260491562358)*x_ref[0]+(-0.667059815618)*x_ref[1]**o+(0.473227354537)*x_ref[1]+(-0.587170510558)*x_ref[2]**o+(-0.740299218393)*x_ref[2]
        ref[1,0,1,1]=(0.87705715098)*x_ref[0]**o+(0.312753471827)*x_ref[0]+(-0.307530283315)*x_ref[1]**o+(0.195564644822)*x_ref[1]+(-0.64384485166)*x_ref[2]**o+(-0.776362951729)*x_ref[2]
        ref[1,0,2,0]=(0.0949408564739)*x_ref[0]**o+(0.765922179855)*x_ref[0]+(0.441215568253)*x_ref[1]**o+(0.339112969437)*x_ref[1]+(-0.0774615292449)*x_ref[2]**o+(0.673258055783)*x_ref[2]
        ref[1,0,2,1]=(0.114828212302)*x_ref[0]**o+(0.0831258766277)*x_ref[0]+(0.674540402248)*x_ref[1]**o+(-0.304921039094)*x_ref[1]+(-0.725514770578)*x_ref[2]**o+(-0.627584773746)*x_ref[2]
        ref[1,1,0,0]=(0.291561714675)*x_ref[0]**o+(0.898587624783)*x_ref[0]+(0.476696264545)*x_ref[1]**o+(-0.454648604978)*x_ref[1]+(0.677949150743)*x_ref[2]**o+(-0.59438289851)*x_ref[2]
        ref[1,1,0,1]=(0.356687258009)*x_ref[0]**o+(0.806236251629)*x_ref[0]+(0.843636690354)*x_ref[1]**o+(-0.649776947181)*x_ref[1]+(0.208542122388)*x_ref[2]**o+(-0.0726045573458)*x_ref[2]
        ref[1,1,1,0]=(0.712800450558)*x_ref[0]**o+(0.748271875755)*x_ref[0]+(0.403880649356)*x_ref[1]**o+(0.956549336754)*x_ref[1]+(0.0420128495043)*x_ref[2]**o+(-0.83177870499)*x_ref[2]
        ref[1,1,1,1]=(-0.886862904174)*x_ref[0]**o+(-0.03822986594)*x_ref[0]+(0.587151517496)*x_ref[1]**o+(-0.120272060898)*x_ref[1]+(-0.573366391933)*x_ref[2]**o+(-0.59772763241)*x_ref[2]
        ref[1,1,2,0]=(0.311960197804)*x_ref[0]**o+(0.486854949444)*x_ref[0]+(-0.647995651954)*x_ref[1]**o+(0.30104277964)*x_ref[1]+(-0.174718703748)*x_ref[2]**o+(-0.506993216295)*x_ref[2]
        ref[1,1,2,1]=(0.979056206991)*x_ref[0]**o+(0.892104464812)*x_ref[0]+(0.0507432703245)*x_ref[1]**o+(0.103630858454)*x_ref[1]+(-0.0254965952266)*x_ref[2]**o+(0.13034027239)*x_ref[2]
        ref[1,2,0,0]=(0.259797966548)*x_ref[0]**o+(0.53018999371)*x_ref[0]+(0.533745722992)*x_ref[1]**o+(0.53029084173)*x_ref[1]+(-0.595126791061)*x_ref[2]**o+(-0.51668295558)*x_ref[2]
        ref[1,2,0,1]=(0.926395590808)*x_ref[0]**o+(-0.0700141207179)*x_ref[0]+(-0.973392454012)*x_ref[1]**o+(-0.46204067097)*x_ref[1]+(0.814964497207)*x_ref[2]**o+(-0.183546091376)*x_ref[2]
        ref[1,2,1,0]=(-0.0872635441658)*x_ref[0]**o+(-0.0307630839035)*x_ref[0]+(0.539798653095)*x_ref[1]**o+(0.966367734848)*x_ref[1]+(-0.817495053612)*x_ref[2]**o+(-0.268397867848)*x_ref[2]
        ref[1,2,1,1]=(-0.933074613921)*x_ref[0]**o+(0.48896533938)*x_ref[0]+(0.453696133488)*x_ref[1]**o+(-0.238730406749)*x_ref[1]+(-0.696785951086)*x_ref[2]**o+(-0.359858385642)*x_ref[2]
        ref[1,2,2,0]=(-0.844406286809)*x_ref[0]**o+(-0.506332243946)*x_ref[0]+(0.976436731658)*x_ref[1]**o+(-0.201489161857)*x_ref[1]+(0.851236173166)*x_ref[2]**o+(-0.998803612947)*x_ref[2]
        ref[1,2,2,1]=(0.259707321449)*x_ref[0]**o+(-0.888149478655)*x_ref[0]+(-0.273502842869)*x_ref[1]**o+(0.291193582158)*x_ref[1]+(-0.897542178027)*x_ref[2]**o+(-0.345994824429)*x_ref[2]
        ref[1,3,0,0]=(0.961298984697)*x_ref[0]**o+(0.248141115092)*x_ref[0]+(-0.623053025178)*x_ref[1]**o+(-0.775095452078)*x_ref[1]+(0.955362707049)*x_ref[2]**o+(-0.517547333705)*x_ref[2]
        ref[1,3,0,1]=(0.858098596227)*x_ref[0]**o+(0.309307483873)*x_ref[0]+(-0.209217034755)*x_ref[1]**o+(0.783803593379)*x_ref[1]+(0.138164128092)*x_ref[2]**o+(-0.922854003461)*x_ref[2]
        ref[1,3,1,0]=(-0.139017942141)*x_ref[0]**o+(-0.338223039119)*x_ref[0]+(0.737882941498)*x_ref[1]**o+(0.948854903087)*x_ref[1]+(-0.0792461633266)*x_ref[2]**o+(0.848875660203)*x_ref[2]
        ref[1,3,1,1]=(0.693284332212)*x_ref[0]**o+(-0.392272875348)*x_ref[0]+(-0.475709045479)*x_ref[1]**o+(0.182948077236)*x_ref[1]+(0.0277569761538)*x_ref[2]**o+(-0.550294947596)*x_ref[2]
        ref[1,3,2,0]=(-0.105381649505)*x_ref[0]**o+(-0.403051690238)*x_ref[0]+(-0.722219735818)*x_ref[1]**o+(-0.380114388869)*x_ref[1]+(0.321452572491)*x_ref[2]**o+(0.858583245253)*x_ref[2]
        ref[1,3,2,1]=(-0.405370062024)*x_ref[0]**o+(0.0743704840047)*x_ref[0]+(0.998947017815)*x_ref[1]**o+(0.983611079522)*x_ref[1]+(0.854202982674)*x_ref[2]**o+(0.288478746498)*x_ref[2]
        ref[1,4,0,0]=(0.0584782489998)*x_ref[0]**o+(0.999286927604)*x_ref[0]+(-0.458166020555)*x_ref[1]**o+(0.529008844128)*x_ref[1]+(0.424592641922)*x_ref[2]**o+(-0.489692340187)*x_ref[2]
        ref[1,4,0,1]=(-0.150555966138)*x_ref[0]**o+(0.362589320838)*x_ref[0]+(0.515820915391)*x_ref[1]**o+(0.742403661312)*x_ref[1]+(0.230278843183)*x_ref[2]**o+(-0.213977724366)*x_ref[2]
        ref[1,4,1,0]=(-0.361161802362)*x_ref[0]**o+(-0.49718332222)*x_ref[0]+(0.143956046197)*x_ref[1]**o+(-0.044550912758)*x_ref[1]+(-0.299892874708)*x_ref[2]**o+(0.370248709792)*x_ref[2]
        ref[1,4,1,1]=(0.259365238173)*x_ref[0]**o+(-0.474694548557)*x_ref[0]+(0.732983396669)*x_ref[1]**o+(-0.295151546774)*x_ref[1]+(-0.337144628526)*x_ref[2]**o+(0.664201775714)*x_ref[2]
        ref[1,4,2,0]=(-0.318016754284)*x_ref[0]**o+(0.555291443852)*x_ref[0]+(0.538150727998)*x_ref[1]**o+(-0.800118353132)*x_ref[1]+(0.652262386596)*x_ref[2]**o+(0.303156645857)*x_ref[2]
        ref[1,4,2,1]=(0.0827142921796)*x_ref[0]**o+(-0.88498648372)*x_ref[0]+(-0.941249844349)*x_ref[1]**o+(-0.162576659039)*x_ref[1]+(0.867621252192)*x_ref[2]**o+(0.0834583795227)*x_ref[2]
        ref[2,0,0,0]=(0.0670088199264)*x_ref[0]**o+(-0.0425961894346)*x_ref[0]+(0.329605789302)*x_ref[1]**o+(0.860453156363)*x_ref[1]+(-0.37410469727)*x_ref[2]**o+(0.55476432384)*x_ref[2]
        ref[2,0,0,1]=(0.801738221879)*x_ref[0]**o+(-0.836563530923)*x_ref[0]+(-0.449651673142)*x_ref[1]**o+(-0.19423133918)*x_ref[1]+(0.947648200044)*x_ref[2]**o+(0.0854067601423)*x_ref[2]
        ref[2,0,1,0]=(0.986595046855)*x_ref[0]**o+(-0.358319352175)*x_ref[0]+(0.493917713513)*x_ref[1]**o+(-0.92052282445)*x_ref[1]+(-0.692726467259)*x_ref[2]**o+(-0.0636943737807)*x_ref[2]
        ref[2,0,1,1]=(0.38491150492)*x_ref[0]**o+(0.766893245938)*x_ref[0]+(0.420188917945)*x_ref[1]**o+(0.43760538854)*x_ref[1]+(-0.770749451981)*x_ref[2]**o+(0.124882164378)*x_ref[2]
        ref[2,0,2,0]=(0.775715718797)*x_ref[0]**o+(0.89516703208)*x_ref[0]+(-0.829021524898)*x_ref[1]**o+(-0.650262454311)*x_ref[1]+(-0.812556336598)*x_ref[2]**o+(-0.997075743997)*x_ref[2]
        ref[2,0,2,1]=(0.3503761008)*x_ref[0]**o+(-0.822009897455)*x_ref[0]+(-0.657657365354)*x_ref[1]**o+(0.476295612156)*x_ref[1]+(0.315554008711)*x_ref[2]**o+(0.691631078049)*x_ref[2]
        ref[2,1,0,0]=(0.680177375433)*x_ref[0]**o+(-0.922726813311)*x_ref[0]+(-0.581116746321)*x_ref[1]**o+(0.628138099343)*x_ref[1]+(-0.568262800293)*x_ref[2]**o+(-0.228201382395)*x_ref[2]
        ref[2,1,0,1]=(0.210094833324)*x_ref[0]**o+(0.281549673876)*x_ref[0]+(0.23785231073)*x_ref[1]**o+(-0.103812455993)*x_ref[1]+(0.530445971191)*x_ref[2]**o+(-0.0830520449939)*x_ref[2]
        ref[2,1,1,0]=(0.367133297563)*x_ref[0]**o+(0.168289954585)*x_ref[0]+(0.614117392787)*x_ref[1]**o+(0.790467766841)*x_ref[1]+(-0.958002150805)*x_ref[2]**o+(0.328916248824)*x_ref[2]
        ref[2,1,1,1]=(0.089535347231)*x_ref[0]**o+(-0.93155993701)*x_ref[0]+(-0.674670760794)*x_ref[1]**o+(-0.274148261773)*x_ref[1]+(-0.807020126056)*x_ref[2]**o+(-0.323334769496)*x_ref[2]
        ref[2,1,2,0]=(-0.948679835219)*x_ref[0]**o+(-0.577235742425)*x_ref[0]+(-0.732584834146)*x_ref[1]**o+(-0.475242504288)*x_ref[1]+(0.213663223262)*x_ref[2]**o+(-0.151401354381)*x_ref[2]
        ref[2,1,2,1]=(0.264513246206)*x_ref[0]**o+(-0.605806437391)*x_ref[0]+(-0.058730926693)*x_ref[1]**o+(-0.661696304412)*x_ref[1]+(-0.361277352457)*x_ref[2]**o+(-0.804408258253)*x_ref[2]
        ref[2,2,0,0]=(-0.144562358054)*x_ref[0]**o+(-0.0960550555725)*x_ref[0]+(-0.905096867803)*x_ref[1]**o+(0.188861784103)*x_ref[1]+(-0.616751005568)*x_ref[2]**o+(0.604711974248)*x_ref[2]
        ref[2,2,0,1]=(-0.0582524307858)*x_ref[0]**o+(-0.503648879785)*x_ref[0]+(-0.682755209956)*x_ref[1]**o+(-0.808213164425)*x_ref[1]+(0.539987276327)*x_ref[2]**o+(0.650557703689)*x_ref[2]
        ref[2,2,1,0]=(0.0607729567733)*x_ref[0]**o+(-0.498819638428)*x_ref[0]+(0.434442595334)*x_ref[1]**o+(-0.491063977603)*x_ref[1]+(0.499436212229)*x_ref[2]**o+(-0.932071991206)*x_ref[2]
        ref[2,2,1,1]=(-0.461295295294)*x_ref[0]**o+(-0.153600767916)*x_ref[0]+(-0.86452392231)*x_ref[1]**o+(0.136760034153)*x_ref[1]+(-0.471642416082)*x_ref[2]**o+(0.453820403438)*x_ref[2]
        ref[2,2,2,0]=(-0.970051888584)*x_ref[0]**o+(0.415728825057)*x_ref[0]+(0.971643909569)*x_ref[1]**o+(0.923719420524)*x_ref[1]+(-0.837313513804)*x_ref[2]**o+(-0.738057676093)*x_ref[2]
        ref[2,2,2,1]=(-0.941355918089)*x_ref[0]**o+(-0.9629784524)*x_ref[0]+(-0.931184348072)*x_ref[1]**o+(0.720006097658)*x_ref[1]+(-0.140750567978)*x_ref[2]**o+(0.55381939119)*x_ref[2]
        ref[2,3,0,0]=(0.753455933669)*x_ref[0]**o+(0.352384922037)*x_ref[0]+(0.737644348002)*x_ref[1]**o+(-0.565144188248)*x_ref[1]+(0.257417624014)*x_ref[2]**o+(-0.660006641683)*x_ref[2]
        ref[2,3,0,1]=(0.886996770925)*x_ref[0]**o+(-0.876595418252)*x_ref[0]+(-0.189778073871)*x_ref[1]**o+(0.560867442336)*x_ref[1]+(-0.262338156346)*x_ref[2]**o+(0.614749098901)*x_ref[2]
        ref[2,3,1,0]=(-0.722912374016)*x_ref[0]**o+(0.447451310273)*x_ref[0]+(0.397510140568)*x_ref[1]**o+(0.932405917932)*x_ref[1]+(-0.206969252336)*x_ref[2]**o+(-0.897366208888)*x_ref[2]
        ref[2,3,1,1]=(0.106324074547)*x_ref[0]**o+(0.822109425884)*x_ref[0]+(0.62699223277)*x_ref[1]**o+(0.238371557839)*x_ref[1]+(-0.422051841243)*x_ref[2]**o+(0.868932130329)*x_ref[2]
        ref[2,3,2,0]=(0.0155790371581)*x_ref[0]**o+(-0.696029577918)*x_ref[0]+(0.259909585217)*x_ref[1]**o+(0.260209121862)*x_ref[1]+(0.780977081459)*x_ref[2]**o+(-0.819957534538)*x_ref[2]
        ref[2,3,2,1]=(0.736223721904)*x_ref[0]**o+(-0.433132598174)*x_ref[0]+(-0.880913649812)*x_ref[1]**o+(-0.375988777461)*x_ref[1]+(-0.351763406307)*x_ref[2]**o+(0.948468252467)*x_ref[2]
        ref[2,4,0,0]=(0.943349180556)*x_ref[0]**o+(-0.720044514103)*x_ref[0]+(0.671900161807)*x_ref[1]**o+(0.358571142942)*x_ref[1]+(-0.132755998648)*x_ref[2]**o+(0.0749815044365)*x_ref[2]
        ref[2,4,0,1]=(0.639070752347)*x_ref[0]**o+(0.551441363406)*x_ref[0]+(0.507547844432)*x_ref[1]**o+(-0.800043772307)*x_ref[1]+(-0.506359503394)*x_ref[2]**o+(0.952191432651)*x_ref[2]
        ref[2,4,1,0]=(-0.655864732254)*x_ref[0]**o+(0.142198390642)*x_ref[0]+(0.075292486702)*x_ref[1]**o+(0.191013189029)*x_ref[1]+(0.106768453247)*x_ref[2]**o+(0.974526515449)*x_ref[2]
        ref[2,4,1,1]=(-0.354750205648)*x_ref[0]**o+(0.141506849368)*x_ref[0]+(-0.170074898141)*x_ref[1]**o+(-0.0882776529335)*x_ref[1]+(0.845254802876)*x_ref[2]**o+(-0.435414692125)*x_ref[2]
        ref[2,4,2,0]=(-0.412836027163)*x_ref[0]**o+(0.830777132582)*x_ref[0]+(-0.87554952236)*x_ref[1]**o+(0.970647516313)*x_ref[1]+(0.989864827286)*x_ref[2]**o+(-0.67790755421)*x_ref[2]
        ref[2,4,2,1]=(0.762907045031)*x_ref[0]**o+(0.140771931118)*x_ref[0]+(0.716859940142)*x_ref[1]**o+(0.151769314189)*x_ref[1]+(-0.812408853675)*x_ref[2]**o+(0.755727078666)*x_ref[2]
        ref[3,0,0,0]=(-0.13358237438)*x_ref[0]**o+(-0.594735856789)*x_ref[0]+(-0.367418402298)*x_ref[1]**o+(0.5349162872)*x_ref[1]+(0.238285320118)*x_ref[2]**o+(0.999627906459)*x_ref[2]
        ref[3,0,0,1]=(-0.537433369218)*x_ref[0]**o+(0.133034652667)*x_ref[0]+(-0.190302464596)*x_ref[1]**o+(0.869919012762)*x_ref[1]+(0.320294306008)*x_ref[2]**o+(0.865908856629)*x_ref[2]
        ref[3,0,1,0]=(0.480851043043)*x_ref[0]**o+(0.833358259022)*x_ref[0]+(0.893051476621)*x_ref[1]**o+(-0.505255581199)*x_ref[1]+(-0.15316145317)*x_ref[2]**o+(-0.937103563776)*x_ref[2]
        ref[3,0,1,1]=(0.765360137605)*x_ref[0]**o+(0.502399552842)*x_ref[0]+(-0.215707512276)*x_ref[1]**o+(-0.402720816967)*x_ref[1]+(0.231360098578)*x_ref[2]**o+(0.732903327067)*x_ref[2]
        ref[3,0,2,0]=(0.120836669045)*x_ref[0]**o+(0.198346794548)*x_ref[0]+(-0.28708447536)*x_ref[1]**o+(0.289672075353)*x_ref[1]+(-0.0305777738322)*x_ref[2]**o+(-0.0686651569373)*x_ref[2]
        ref[3,0,2,1]=(0.0574598844614)*x_ref[0]**o+(0.031092492399)*x_ref[0]+(-0.406805507341)*x_ref[1]**o+(-0.623370509947)*x_ref[1]+(-0.476094160966)*x_ref[2]**o+(0.0840271758875)*x_ref[2]
        ref[3,1,0,0]=(-0.569415249126)*x_ref[0]**o+(-0.339675540399)*x_ref[0]+(0.182536980707)*x_ref[1]**o+(0.727480200453)*x_ref[1]+(0.851672551504)*x_ref[2]**o+(0.817312313075)*x_ref[2]
        ref[3,1,0,1]=(0.921422642904)*x_ref[0]**o+(0.697656143906)*x_ref[0]+(0.969442847752)*x_ref[1]**o+(-0.00957994710064)*x_ref[1]+(-0.315230556399)*x_ref[2]**o+(-0.870759158384)*x_ref[2]
        ref[3,1,1,0]=(-0.813310465463)*x_ref[0]**o+(-0.809362802903)*x_ref[0]+(-0.0685101056428)*x_ref[1]**o+(0.708195651847)*x_ref[1]+(-0.876449456312)*x_ref[2]**o+(0.882503536479)*x_ref[2]
        ref[3,1,1,1]=(-0.277504922905)*x_ref[0]**o+(-0.819428706347)*x_ref[0]+(0.463203258731)*x_ref[1]**o+(0.597493096401)*x_ref[1]+(0.630522187524)*x_ref[2]**o+(0.92480079466)*x_ref[2]
        ref[3,1,2,0]=(-0.703167803764)*x_ref[0]**o+(0.730459205382)*x_ref[0]+(-0.507036921123)*x_ref[1]**o+(0.419358545765)*x_ref[1]+(-0.542758923679)*x_ref[2]**o+(0.831844312917)*x_ref[2]
        ref[3,1,2,1]=(0.135200299704)*x_ref[0]**o+(-0.160734675188)*x_ref[0]+(-0.893851689192)*x_ref[1]**o+(0.748734906185)*x_ref[1]+(0.970464447167)*x_ref[2]**o+(0.277911376757)*x_ref[2]
        ref[3,2,0,0]=(0.750104328784)*x_ref[0]**o+(-0.818350593893)*x_ref[0]+(-0.749900854281)*x_ref[1]**o+(0.289117653816)*x_ref[1]+(-0.219904822162)*x_ref[2]**o+(0.805208050452)*x_ref[2]
        ref[3,2,0,1]=(-0.129614389749)*x_ref[0]**o+(-0.64067443816)*x_ref[0]+(0.513013650233)*x_ref[1]**o+(0.868869035341)*x_ref[1]+(-0.040145121639)*x_ref[2]**o+(-0.282350136119)*x_ref[2]
        ref[3,2,1,0]=(-0.172814929104)*x_ref[0]**o+(-0.296930471915)*x_ref[0]+(-0.796114111585)*x_ref[1]**o+(-0.210181080249)*x_ref[1]+(-0.606107841173)*x_ref[2]**o+(-0.954329084175)*x_ref[2]
        ref[3,2,1,1]=(0.136149261959)*x_ref[0]**o+(-0.613980696755)*x_ref[0]+(0.968430714778)*x_ref[1]**o+(0.683553952099)*x_ref[1]+(-0.321383789121)*x_ref[2]**o+(-0.139570984234)*x_ref[2]
        ref[3,2,2,0]=(-0.273321720654)*x_ref[0]**o+(0.963651110162)*x_ref[0]+(-0.847402611499)*x_ref[1]**o+(0.237131262905)*x_ref[1]+(-0.617296821558)*x_ref[2]**o+(-0.681856394629)*x_ref[2]
        ref[3,2,2,1]=(-0.760071928624)*x_ref[0]**o+(0.0345053779755)*x_ref[0]+(-0.583680069732)*x_ref[1]**o+(0.150497658077)*x_ref[1]+(0.0257096621429)*x_ref[2]**o+(-0.159406203385)*x_ref[2]
        ref[3,3,0,0]=(-0.0668501746947)*x_ref[0]**o+(-0.183645899611)*x_ref[0]+(0.123667845765)*x_ref[1]**o+(0.0928142420352)*x_ref[1]+(-0.933250766317)*x_ref[2]**o+(-0.629986949639)*x_ref[2]
        ref[3,3,0,1]=(-0.0212790928206)*x_ref[0]**o+(0.444036360325)*x_ref[0]+(0.09182358127)*x_ref[1]**o+(0.473088563672)*x_ref[1]+(0.0852925382953)*x_ref[2]**o+(-0.238517929786)*x_ref[2]
        ref[3,3,1,0]=(0.0647235093728)*x_ref[0]**o+(-0.266285898716)*x_ref[0]+(0.944999560866)*x_ref[1]**o+(-0.757696716486)*x_ref[1]+(-0.439165278732)*x_ref[2]**o+(0.960091479774)*x_ref[2]
        ref[3,3,1,1]=(-0.792249238894)*x_ref[0]**o+(-0.568585439587)*x_ref[0]+(-0.0426655335607)*x_ref[1]**o+(0.980004032206)*x_ref[1]+(0.635227812889)*x_ref[2]**o+(0.6687165384)*x_ref[2]
        ref[3,3,2,0]=(0.340468442769)*x_ref[0]**o+(0.791071154735)*x_ref[0]+(-0.817919975547)*x_ref[1]**o+(0.0328383066982)*x_ref[1]+(-0.323093904705)*x_ref[2]**o+(-0.302641620002)*x_ref[2]
        ref[3,3,2,1]=(-0.476501733213)*x_ref[0]**o+(-0.429845943537)*x_ref[0]+(0.518551215268)*x_ref[1]**o+(-0.551690143654)*x_ref[1]+(0.591983049997)*x_ref[2]**o+(0.586292683497)*x_ref[2]
        ref[3,4,0,0]=(0.119036526248)*x_ref[0]**o+(-0.156404044048)*x_ref[0]+(0.158504048681)*x_ref[1]**o+(-0.3779600709)*x_ref[1]+(-0.362107671534)*x_ref[2]**o+(-0.314508628804)*x_ref[2]
        ref[3,4,0,1]=(0.382011103106)*x_ref[0]**o+(-0.123711676964)*x_ref[0]+(-0.939014457264)*x_ref[1]**o+(-0.5986155807)*x_ref[1]+(-0.852092491566)*x_ref[2]**o+(-0.410438924788)*x_ref[2]
        ref[3,4,1,0]=(0.83612586535)*x_ref[0]**o+(-0.606098645092)*x_ref[0]+(0.643469717148)*x_ref[1]**o+(0.845106525654)*x_ref[1]+(0.347455649691)*x_ref[2]**o+(-0.166796580458)*x_ref[2]
        ref[3,4,1,1]=(0.693431758156)*x_ref[0]**o+(-0.313983401147)*x_ref[0]+(-0.862690834893)*x_ref[1]**o+(0.417233988055)*x_ref[1]+(-0.361769997589)*x_ref[2]**o+(-0.22749888397)*x_ref[2]
        ref[3,4,2,0]=(0.680358405387)*x_ref[0]**o+(-0.786041842215)*x_ref[0]+(0.288823579415)*x_ref[1]**o+(-0.961099851196)*x_ref[1]+(0.669431785695)*x_ref[2]**o+(0.956564160268)*x_ref[2]
        ref[3,4,2,1]=(-0.21416500005)*x_ref[0]**o+(0.48495191486)*x_ref[0]+(-0.173465452645)*x_ref[1]**o+(0.662530321557)*x_ref[1]+(0.984573283824)*x_ref[2]**o+(-0.568211383785)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.852985368636)*x[0]**o+(-0.691601416209)*x[0]+(-0.689994415069)*x[1]**o+(-0.471926256627)*x[1]
        ref=(-0.852985368636)*x_ref[0]**o+(-0.691601416209)*x_ref[0]+(-0.689994415069)*x_ref[1]**o+(-0.471926256627)*x_ref[1]
      else:
        s=(-0.796357734502)*x[0]**o+(0.158918179797)*x[0]+(0.580745911056)*x[1]**o+(0.786294909139)*x[1]+(0.436014958708)*x[2]**o+(-0.600115428014)*x[2]
        ref=(-0.796357734502)*x_ref[0]**o+(0.158918179797)*x_ref[0]+(0.580745911056)*x_ref[1]**o+(0.786294909139)*x_ref[1]+(0.436014958708)*x_ref[2]**o+(-0.600115428014)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.997766043376)*x[0]**o+(0.495776426131)*x[0]+(-0.387839677144)*x[1]**o+(0.774520715668)*x[1]
        s[1]=(0.947961152483)*x[0]**o+(0.987829314406)*x[0]+(0.32828485757)*x[1]**o+(-0.617666605292)*x[1]
        ref[0]=(0.997766043376)*x_ref[0]**o+(0.495776426131)*x_ref[0]+(-0.387839677144)*x_ref[1]**o+(0.774520715668)*x_ref[1]
        ref[1]=(0.947961152483)*x_ref[0]**o+(0.987829314406)*x_ref[0]+(0.32828485757)*x_ref[1]**o+(-0.617666605292)*x_ref[1]
      else:
        s[0]=(-0.13474557089)*x[0]**o+(-0.588190696155)*x[0]+(0.128472574732)*x[1]**o+(0.690322787372)*x[1]+(0.884541159956)*x[2]**o+(-0.886834154713)*x[2]
        s[1]=(-0.751296359458)*x[0]**o+(-0.659908205442)*x[0]+(-0.415219823654)*x[1]**o+(0.550397764781)*x[1]+(-0.728215097593)*x[2]**o+(0.582854885508)*x[2]
        ref[0]=(-0.13474557089)*x_ref[0]**o+(-0.588190696155)*x_ref[0]+(0.128472574732)*x_ref[1]**o+(0.690322787372)*x_ref[1]+(0.884541159956)*x_ref[2]**o+(-0.886834154713)*x_ref[2]
        ref[1]=(-0.751296359458)*x_ref[0]**o+(-0.659908205442)*x_ref[0]+(-0.415219823654)*x_ref[1]**o+(0.550397764781)*x_ref[1]+(-0.728215097593)*x_ref[2]**o+(0.582854885508)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.540707708777)*x[0]**o+(-0.666948327254)*x[0]+(-0.471649914321)*x[1]**o+(0.65162222166)*x[1]
        s[0,1]=(0.39265499863)*x[0]**o+(0.927768486766)*x[0]+(-0.625791400147)*x[1]**o+(-0.406599346038)*x[1]
        s[0,2]=(0.721211493492)*x[0]**o+(0.0169359532239)*x[0]+(-0.450701248736)*x[1]**o+(0.792695848136)*x[1]
        s[0,3]=(0.349994038242)*x[0]**o+(0.124108838306)*x[0]+(0.221156059551)*x[1]**o+(-0.891451468693)*x[1]
        s[0,4]=(0.493264497246)*x[0]**o+(0.234167381112)*x[0]+(0.543699100028)*x[1]**o+(-0.838472895241)*x[1]
        s[1,0]=(0.359533817281)*x[0]**o+(-0.062563306784)*x[0]+(0.467858739349)*x[1]**o+(0.692754015398)*x[1]
        s[1,1]=(0.888396152909)*x[0]**o+(0.184738712864)*x[0]+(-0.0379184183599)*x[1]**o+(0.991506780667)*x[1]
        s[1,2]=(-0.68045741795)*x[0]**o+(0.832740353977)*x[0]+(0.491010493882)*x[1]**o+(0.454304365237)*x[1]
        s[1,3]=(-0.179660488547)*x[0]**o+(0.628405133401)*x[0]+(-0.69684323253)*x[1]**o+(0.889439582492)*x[1]
        s[1,4]=(-0.379088893433)*x[0]**o+(0.497309923662)*x[0]+(0.153969715423)*x[1]**o+(-0.386632220307)*x[1]
        s[2,0]=(-0.186297693534)*x[0]**o+(0.489487382751)*x[0]+(0.644068223202)*x[1]**o+(0.0743289639696)*x[1]
        s[2,1]=(0.593905543479)*x[0]**o+(-0.966302120537)*x[0]+(-0.861407416257)*x[1]**o+(0.2431998414)*x[1]
        s[2,2]=(0.683336278471)*x[0]**o+(0.337655287244)*x[0]+(0.876310641583)*x[1]**o+(-0.850438003892)*x[1]
        s[2,3]=(-0.109203794389)*x[0]**o+(0.245455391347)*x[0]+(-0.30542844788)*x[1]**o+(0.846645969509)*x[1]
        s[2,4]=(-0.478221810644)*x[0]**o+(0.555082091582)*x[0]+(-0.331626471767)*x[1]**o+(-0.363786378267)*x[1]
        s[3,0]=(-0.0373544633657)*x[0]**o+(-0.821373667974)*x[0]+(-0.569127067519)*x[1]**o+(0.178159969962)*x[1]
        s[3,1]=(0.574056120213)*x[0]**o+(-0.557444624837)*x[0]+(-0.858492289239)*x[1]**o+(0.82616627128)*x[1]
        s[3,2]=(0.0934423420331)*x[0]**o+(-0.64321213739)*x[0]+(0.0832655456405)*x[1]**o+(0.245929781754)*x[1]
        s[3,3]=(0.0254040950683)*x[0]**o+(0.169079366685)*x[0]+(0.150029664562)*x[1]**o+(0.987260372009)*x[1]
        s[3,4]=(0.201458676701)*x[0]**o+(0.840874176438)*x[0]+(-0.533157142378)*x[1]**o+(-0.216306366879)*x[1]
        ref[0,0]=(0.540707708777)*x_ref[0]**o+(-0.666948327254)*x_ref[0]+(-0.471649914321)*x_ref[1]**o+(0.65162222166)*x_ref[1]
        ref[0,1]=(0.39265499863)*x_ref[0]**o+(0.927768486766)*x_ref[0]+(-0.625791400147)*x_ref[1]**o+(-0.406599346038)*x_ref[1]
        ref[0,2]=(0.721211493492)*x_ref[0]**o+(0.0169359532239)*x_ref[0]+(-0.450701248736)*x_ref[1]**o+(0.792695848136)*x_ref[1]
        ref[0,3]=(0.349994038242)*x_ref[0]**o+(0.124108838306)*x_ref[0]+(0.221156059551)*x_ref[1]**o+(-0.891451468693)*x_ref[1]
        ref[0,4]=(0.493264497246)*x_ref[0]**o+(0.234167381112)*x_ref[0]+(0.543699100028)*x_ref[1]**o+(-0.838472895241)*x_ref[1]
        ref[1,0]=(0.359533817281)*x_ref[0]**o+(-0.062563306784)*x_ref[0]+(0.467858739349)*x_ref[1]**o+(0.692754015398)*x_ref[1]
        ref[1,1]=(0.888396152909)*x_ref[0]**o+(0.184738712864)*x_ref[0]+(-0.0379184183599)*x_ref[1]**o+(0.991506780667)*x_ref[1]
        ref[1,2]=(-0.68045741795)*x_ref[0]**o+(0.832740353977)*x_ref[0]+(0.491010493882)*x_ref[1]**o+(0.454304365237)*x_ref[1]
        ref[1,3]=(-0.179660488547)*x_ref[0]**o+(0.628405133401)*x_ref[0]+(-0.69684323253)*x_ref[1]**o+(0.889439582492)*x_ref[1]
        ref[1,4]=(-0.379088893433)*x_ref[0]**o+(0.497309923662)*x_ref[0]+(0.153969715423)*x_ref[1]**o+(-0.386632220307)*x_ref[1]
        ref[2,0]=(-0.186297693534)*x_ref[0]**o+(0.489487382751)*x_ref[0]+(0.644068223202)*x_ref[1]**o+(0.0743289639696)*x_ref[1]
        ref[2,1]=(0.593905543479)*x_ref[0]**o+(-0.966302120537)*x_ref[0]+(-0.861407416257)*x_ref[1]**o+(0.2431998414)*x_ref[1]
        ref[2,2]=(0.683336278471)*x_ref[0]**o+(0.337655287244)*x_ref[0]+(0.876310641583)*x_ref[1]**o+(-0.850438003892)*x_ref[1]
        ref[2,3]=(-0.109203794389)*x_ref[0]**o+(0.245455391347)*x_ref[0]+(-0.30542844788)*x_ref[1]**o+(0.846645969509)*x_ref[1]
        ref[2,4]=(-0.478221810644)*x_ref[0]**o+(0.555082091582)*x_ref[0]+(-0.331626471767)*x_ref[1]**o+(-0.363786378267)*x_ref[1]
        ref[3,0]=(-0.0373544633657)*x_ref[0]**o+(-0.821373667974)*x_ref[0]+(-0.569127067519)*x_ref[1]**o+(0.178159969962)*x_ref[1]
        ref[3,1]=(0.574056120213)*x_ref[0]**o+(-0.557444624837)*x_ref[0]+(-0.858492289239)*x_ref[1]**o+(0.82616627128)*x_ref[1]
        ref[3,2]=(0.0934423420331)*x_ref[0]**o+(-0.64321213739)*x_ref[0]+(0.0832655456405)*x_ref[1]**o+(0.245929781754)*x_ref[1]
        ref[3,3]=(0.0254040950683)*x_ref[0]**o+(0.169079366685)*x_ref[0]+(0.150029664562)*x_ref[1]**o+(0.987260372009)*x_ref[1]
        ref[3,4]=(0.201458676701)*x_ref[0]**o+(0.840874176438)*x_ref[0]+(-0.533157142378)*x_ref[1]**o+(-0.216306366879)*x_ref[1]
      else:
        s[0,0]=(0.872668762115)*x[0]**o+(-0.816027656579)*x[0]+(-0.867841046004)*x[1]**o+(-0.349526762676)*x[1]+(0.938408466875)*x[2]**o+(-0.306416040856)*x[2]
        s[0,1]=(0.909287177866)*x[0]**o+(-0.253902016034)*x[0]+(-0.546477610869)*x[1]**o+(-0.793452692383)*x[1]+(0.776576227123)*x[2]**o+(-0.0967411331785)*x[2]
        s[0,2]=(-0.398898197311)*x[0]**o+(-0.897448112384)*x[0]+(-0.455423203326)*x[1]**o+(-0.367585262313)*x[1]+(-0.0545607718974)*x[2]**o+(-0.282047253613)*x[2]
        s[0,3]=(0.482124916132)*x[0]**o+(-0.631619166977)*x[0]+(-0.0399181373809)*x[1]**o+(0.763797539726)*x[1]+(0.0291052931267)*x[2]**o+(-0.617858834155)*x[2]
        s[0,4]=(-0.0297695890339)*x[0]**o+(0.854962979284)*x[0]+(-0.781784598569)*x[1]**o+(0.299243628895)*x[1]+(-0.310475890278)*x[2]**o+(-0.505378646661)*x[2]
        s[1,0]=(0.484812356262)*x[0]**o+(-0.623958486741)*x[0]+(-0.470006063838)*x[1]**o+(0.660345336288)*x[1]+(-0.878879608367)*x[2]**o+(0.390207006706)*x[2]
        s[1,1]=(0.0351313080262)*x[0]**o+(-0.976289440937)*x[0]+(-0.775189613854)*x[1]**o+(0.765051272995)*x[1]+(-0.675689153966)*x[2]**o+(-0.783405673294)*x[2]
        s[1,2]=(0.662391389122)*x[0]**o+(0.323408772964)*x[0]+(-0.363746897244)*x[1]**o+(0.606811951514)*x[1]+(0.47946488956)*x[2]**o+(-0.966120072388)*x[2]
        s[1,3]=(-0.875339728461)*x[0]**o+(-0.705421436172)*x[0]+(0.291443522131)*x[1]**o+(-0.754961478752)*x[1]+(-0.321825067143)*x[2]**o+(-0.574459688222)*x[2]
        s[1,4]=(-0.255183687846)*x[0]**o+(0.154020081567)*x[0]+(0.84391736219)*x[1]**o+(0.0431273811842)*x[1]+(-0.667796573312)*x[2]**o+(-0.964117112042)*x[2]
        s[2,0]=(-0.530764723315)*x[0]**o+(-0.149048311546)*x[0]+(-0.451189249073)*x[1]**o+(0.999397419444)*x[1]+(0.799956886737)*x[2]**o+(-0.313688076191)*x[2]
        s[2,1]=(-0.749917464595)*x[0]**o+(0.292630460079)*x[0]+(0.506753622803)*x[1]**o+(-0.343689419436)*x[1]+(0.128258231841)*x[2]**o+(0.296077111022)*x[2]
        s[2,2]=(0.394787103006)*x[0]**o+(0.249722955553)*x[0]+(-0.839807889834)*x[1]**o+(-0.667642818246)*x[1]+(0.692437679503)*x[2]**o+(0.0127711080992)*x[2]
        s[2,3]=(-0.0220395920977)*x[0]**o+(-0.720750207444)*x[0]+(0.358116668188)*x[1]**o+(-0.204183323047)*x[1]+(-0.410539982758)*x[2]**o+(-0.671312870493)*x[2]
        s[2,4]=(-0.285147258365)*x[0]**o+(-0.813590616716)*x[0]+(-0.999369996254)*x[1]**o+(-0.348357331572)*x[1]+(-0.714185462982)*x[2]**o+(0.0724943925182)*x[2]
        s[3,0]=(-0.935258847991)*x[0]**o+(-0.78218618139)*x[0]+(0.611019605302)*x[1]**o+(0.518312268136)*x[1]+(0.518366032626)*x[2]**o+(-0.61673250713)*x[2]
        s[3,1]=(0.334978122301)*x[0]**o+(0.144777589204)*x[0]+(-0.654113643417)*x[1]**o+(0.874224972036)*x[1]+(0.988416020574)*x[2]**o+(-0.617165353999)*x[2]
        s[3,2]=(-0.614946557606)*x[0]**o+(-0.77713622969)*x[0]+(-0.0601180324016)*x[1]**o+(-0.493272161621)*x[1]+(-0.118228561563)*x[2]**o+(0.956097356096)*x[2]
        s[3,3]=(0.123381793762)*x[0]**o+(-0.842700232136)*x[0]+(0.142566215799)*x[1]**o+(-0.272131336716)*x[1]+(-0.160411232329)*x[2]**o+(0.344527936371)*x[2]
        s[3,4]=(0.195186658307)*x[0]**o+(-0.774658826133)*x[0]+(0.22329451736)*x[1]**o+(-0.934913350667)*x[1]+(0.883414583429)*x[2]**o+(0.209142336403)*x[2]
        ref[0,0]=(0.872668762115)*x_ref[0]**o+(-0.816027656579)*x_ref[0]+(-0.867841046004)*x_ref[1]**o+(-0.349526762676)*x_ref[1]+(0.938408466875)*x_ref[2]**o+(-0.306416040856)*x_ref[2]
        ref[0,1]=(0.909287177866)*x_ref[0]**o+(-0.253902016034)*x_ref[0]+(-0.546477610869)*x_ref[1]**o+(-0.793452692383)*x_ref[1]+(0.776576227123)*x_ref[2]**o+(-0.0967411331785)*x_ref[2]
        ref[0,2]=(-0.398898197311)*x_ref[0]**o+(-0.897448112384)*x_ref[0]+(-0.455423203326)*x_ref[1]**o+(-0.367585262313)*x_ref[1]+(-0.0545607718974)*x_ref[2]**o+(-0.282047253613)*x_ref[2]
        ref[0,3]=(0.482124916132)*x_ref[0]**o+(-0.631619166977)*x_ref[0]+(-0.0399181373809)*x_ref[1]**o+(0.763797539726)*x_ref[1]+(0.0291052931267)*x_ref[2]**o+(-0.617858834155)*x_ref[2]
        ref[0,4]=(-0.0297695890339)*x_ref[0]**o+(0.854962979284)*x_ref[0]+(-0.781784598569)*x_ref[1]**o+(0.299243628895)*x_ref[1]+(-0.310475890278)*x_ref[2]**o+(-0.505378646661)*x_ref[2]
        ref[1,0]=(0.484812356262)*x_ref[0]**o+(-0.623958486741)*x_ref[0]+(-0.470006063838)*x_ref[1]**o+(0.660345336288)*x_ref[1]+(-0.878879608367)*x_ref[2]**o+(0.390207006706)*x_ref[2]
        ref[1,1]=(0.0351313080262)*x_ref[0]**o+(-0.976289440937)*x_ref[0]+(-0.775189613854)*x_ref[1]**o+(0.765051272995)*x_ref[1]+(-0.675689153966)*x_ref[2]**o+(-0.783405673294)*x_ref[2]
        ref[1,2]=(0.662391389122)*x_ref[0]**o+(0.323408772964)*x_ref[0]+(-0.363746897244)*x_ref[1]**o+(0.606811951514)*x_ref[1]+(0.47946488956)*x_ref[2]**o+(-0.966120072388)*x_ref[2]
        ref[1,3]=(-0.875339728461)*x_ref[0]**o+(-0.705421436172)*x_ref[0]+(0.291443522131)*x_ref[1]**o+(-0.754961478752)*x_ref[1]+(-0.321825067143)*x_ref[2]**o+(-0.574459688222)*x_ref[2]
        ref[1,4]=(-0.255183687846)*x_ref[0]**o+(0.154020081567)*x_ref[0]+(0.84391736219)*x_ref[1]**o+(0.0431273811842)*x_ref[1]+(-0.667796573312)*x_ref[2]**o+(-0.964117112042)*x_ref[2]
        ref[2,0]=(-0.530764723315)*x_ref[0]**o+(-0.149048311546)*x_ref[0]+(-0.451189249073)*x_ref[1]**o+(0.999397419444)*x_ref[1]+(0.799956886737)*x_ref[2]**o+(-0.313688076191)*x_ref[2]
        ref[2,1]=(-0.749917464595)*x_ref[0]**o+(0.292630460079)*x_ref[0]+(0.506753622803)*x_ref[1]**o+(-0.343689419436)*x_ref[1]+(0.128258231841)*x_ref[2]**o+(0.296077111022)*x_ref[2]
        ref[2,2]=(0.394787103006)*x_ref[0]**o+(0.249722955553)*x_ref[0]+(-0.839807889834)*x_ref[1]**o+(-0.667642818246)*x_ref[1]+(0.692437679503)*x_ref[2]**o+(0.0127711080992)*x_ref[2]
        ref[2,3]=(-0.0220395920977)*x_ref[0]**o+(-0.720750207444)*x_ref[0]+(0.358116668188)*x_ref[1]**o+(-0.204183323047)*x_ref[1]+(-0.410539982758)*x_ref[2]**o+(-0.671312870493)*x_ref[2]
        ref[2,4]=(-0.285147258365)*x_ref[0]**o+(-0.813590616716)*x_ref[0]+(-0.999369996254)*x_ref[1]**o+(-0.348357331572)*x_ref[1]+(-0.714185462982)*x_ref[2]**o+(0.0724943925182)*x_ref[2]
        ref[3,0]=(-0.935258847991)*x_ref[0]**o+(-0.78218618139)*x_ref[0]+(0.611019605302)*x_ref[1]**o+(0.518312268136)*x_ref[1]+(0.518366032626)*x_ref[2]**o+(-0.61673250713)*x_ref[2]
        ref[3,1]=(0.334978122301)*x_ref[0]**o+(0.144777589204)*x_ref[0]+(-0.654113643417)*x_ref[1]**o+(0.874224972036)*x_ref[1]+(0.988416020574)*x_ref[2]**o+(-0.617165353999)*x_ref[2]
        ref[3,2]=(-0.614946557606)*x_ref[0]**o+(-0.77713622969)*x_ref[0]+(-0.0601180324016)*x_ref[1]**o+(-0.493272161621)*x_ref[1]+(-0.118228561563)*x_ref[2]**o+(0.956097356096)*x_ref[2]
        ref[3,3]=(0.123381793762)*x_ref[0]**o+(-0.842700232136)*x_ref[0]+(0.142566215799)*x_ref[1]**o+(-0.272131336716)*x_ref[1]+(-0.160411232329)*x_ref[2]**o+(0.344527936371)*x_ref[2]
        ref[3,4]=(0.195186658307)*x_ref[0]**o+(-0.774658826133)*x_ref[0]+(0.22329451736)*x_ref[1]**o+(-0.934913350667)*x_ref[1]+(0.883414583429)*x_ref[2]**o+(0.209142336403)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.156299768554)*x[0]**o+(0.459827039123)*x[0]+(-0.0197265970261)*x[1]**o+(-0.185466218355)*x[1]
        s[0,0,1]=(-0.332652682606)*x[0]**o+(-0.727191024608)*x[0]+(0.855023316467)*x[1]**o+(0.989258272735)*x[1]
        s[0,1,0]=(-0.684749431529)*x[0]**o+(-0.330976098391)*x[0]+(0.951823033761)*x[1]**o+(0.828372926831)*x[1]
        s[0,1,1]=(-0.947934025637)*x[0]**o+(-0.703882235645)*x[0]+(-0.382498726707)*x[1]**o+(-0.416712534592)*x[1]
        s[1,0,0]=(0.268886837209)*x[0]**o+(-0.00424369485341)*x[0]+(-0.901073289396)*x[1]**o+(0.652663823918)*x[1]
        s[1,0,1]=(-0.944532566222)*x[0]**o+(0.605970402437)*x[0]+(-0.872983074127)*x[1]**o+(-0.605342924929)*x[1]
        s[1,1,0]=(0.899833782923)*x[0]**o+(-0.423117683495)*x[0]+(0.360329277459)*x[1]**o+(-0.378555269956)*x[1]
        s[1,1,1]=(0.0804731081801)*x[0]**o+(0.28457153585)*x[0]+(-0.412672511383)*x[1]**o+(-0.446119799977)*x[1]
        s[2,0,0]=(-0.981685713427)*x[0]**o+(0.750030890662)*x[0]+(-0.942777059146)*x[1]**o+(0.736449759398)*x[1]
        s[2,0,1]=(0.586282353308)*x[0]**o+(-0.886945155404)*x[0]+(0.35336036468)*x[1]**o+(0.831440884627)*x[1]
        s[2,1,0]=(0.971383399562)*x[0]**o+(0.921512198603)*x[0]+(-0.983955960592)*x[1]**o+(-0.170180363255)*x[1]
        s[2,1,1]=(-0.956023304113)*x[0]**o+(0.197946744121)*x[0]+(-0.406938561509)*x[1]**o+(-0.50378748984)*x[1]
        s[3,0,0]=(-0.135318256749)*x[0]**o+(-0.845428466148)*x[0]+(0.19729474016)*x[1]**o+(-0.320895148489)*x[1]
        s[3,0,1]=(-0.145937639147)*x[0]**o+(-0.511536148587)*x[0]+(-0.22015558536)*x[1]**o+(0.120116967099)*x[1]
        s[3,1,0]=(-0.311099128945)*x[0]**o+(-0.981649008084)*x[0]+(0.214584333563)*x[1]**o+(0.403963738711)*x[1]
        s[3,1,1]=(-0.0672377429337)*x[0]**o+(0.447752491728)*x[0]+(0.865672486599)*x[1]**o+(0.758623643658)*x[1]
        s[4,0,0]=(0.125909493643)*x[0]**o+(0.77920124826)*x[0]+(-0.491111189228)*x[1]**o+(-0.982624198164)*x[1]
        s[4,0,1]=(0.963994569131)*x[0]**o+(0.590025563951)*x[0]+(-0.432193464931)*x[1]**o+(0.0609843850333)*x[1]
        s[4,1,0]=(-0.810094882894)*x[0]**o+(0.478510251999)*x[0]+(0.909645604317)*x[1]**o+(-0.162920820685)*x[1]
        s[4,1,1]=(-0.849640544361)*x[0]**o+(-0.16981520619)*x[0]+(0.548383377264)*x[1]**o+(0.70666908595)*x[1]
        s[5,0,0]=(0.639591398598)*x[0]**o+(-0.052037663461)*x[0]+(0.80293352341)*x[1]**o+(0.373765944073)*x[1]
        s[5,0,1]=(0.278782801517)*x[0]**o+(0.360704948502)*x[0]+(0.518835994538)*x[1]**o+(-0.748058641453)*x[1]
        s[5,1,0]=(0.376125218461)*x[0]**o+(0.587328226457)*x[0]+(-0.408741695392)*x[1]**o+(-0.922372017858)*x[1]
        s[5,1,1]=(0.921904823293)*x[0]**o+(-0.997993840905)*x[0]+(-0.569469660555)*x[1]**o+(0.246392170853)*x[1]
        ref[0,0,0]=(-0.156299768554)*x_ref[0]**o+(0.459827039123)*x_ref[0]+(-0.0197265970261)*x_ref[1]**o+(-0.185466218355)*x_ref[1]
        ref[0,0,1]=(-0.332652682606)*x_ref[0]**o+(-0.727191024608)*x_ref[0]+(0.855023316467)*x_ref[1]**o+(0.989258272735)*x_ref[1]
        ref[0,1,0]=(-0.684749431529)*x_ref[0]**o+(-0.330976098391)*x_ref[0]+(0.951823033761)*x_ref[1]**o+(0.828372926831)*x_ref[1]
        ref[0,1,1]=(-0.947934025637)*x_ref[0]**o+(-0.703882235645)*x_ref[0]+(-0.382498726707)*x_ref[1]**o+(-0.416712534592)*x_ref[1]
        ref[1,0,0]=(0.268886837209)*x_ref[0]**o+(-0.00424369485341)*x_ref[0]+(-0.901073289396)*x_ref[1]**o+(0.652663823918)*x_ref[1]
        ref[1,0,1]=(-0.944532566222)*x_ref[0]**o+(0.605970402437)*x_ref[0]+(-0.872983074127)*x_ref[1]**o+(-0.605342924929)*x_ref[1]
        ref[1,1,0]=(0.899833782923)*x_ref[0]**o+(-0.423117683495)*x_ref[0]+(0.360329277459)*x_ref[1]**o+(-0.378555269956)*x_ref[1]
        ref[1,1,1]=(0.0804731081801)*x_ref[0]**o+(0.28457153585)*x_ref[0]+(-0.412672511383)*x_ref[1]**o+(-0.446119799977)*x_ref[1]
        ref[2,0,0]=(-0.981685713427)*x_ref[0]**o+(0.750030890662)*x_ref[0]+(-0.942777059146)*x_ref[1]**o+(0.736449759398)*x_ref[1]
        ref[2,0,1]=(0.586282353308)*x_ref[0]**o+(-0.886945155404)*x_ref[0]+(0.35336036468)*x_ref[1]**o+(0.831440884627)*x_ref[1]
        ref[2,1,0]=(0.971383399562)*x_ref[0]**o+(0.921512198603)*x_ref[0]+(-0.983955960592)*x_ref[1]**o+(-0.170180363255)*x_ref[1]
        ref[2,1,1]=(-0.956023304113)*x_ref[0]**o+(0.197946744121)*x_ref[0]+(-0.406938561509)*x_ref[1]**o+(-0.50378748984)*x_ref[1]
        ref[3,0,0]=(-0.135318256749)*x_ref[0]**o+(-0.845428466148)*x_ref[0]+(0.19729474016)*x_ref[1]**o+(-0.320895148489)*x_ref[1]
        ref[3,0,1]=(-0.145937639147)*x_ref[0]**o+(-0.511536148587)*x_ref[0]+(-0.22015558536)*x_ref[1]**o+(0.120116967099)*x_ref[1]
        ref[3,1,0]=(-0.311099128945)*x_ref[0]**o+(-0.981649008084)*x_ref[0]+(0.214584333563)*x_ref[1]**o+(0.403963738711)*x_ref[1]
        ref[3,1,1]=(-0.0672377429337)*x_ref[0]**o+(0.447752491728)*x_ref[0]+(0.865672486599)*x_ref[1]**o+(0.758623643658)*x_ref[1]
        ref[4,0,0]=(0.125909493643)*x_ref[0]**o+(0.77920124826)*x_ref[0]+(-0.491111189228)*x_ref[1]**o+(-0.982624198164)*x_ref[1]
        ref[4,0,1]=(0.963994569131)*x_ref[0]**o+(0.590025563951)*x_ref[0]+(-0.432193464931)*x_ref[1]**o+(0.0609843850333)*x_ref[1]
        ref[4,1,0]=(-0.810094882894)*x_ref[0]**o+(0.478510251999)*x_ref[0]+(0.909645604317)*x_ref[1]**o+(-0.162920820685)*x_ref[1]
        ref[4,1,1]=(-0.849640544361)*x_ref[0]**o+(-0.16981520619)*x_ref[0]+(0.548383377264)*x_ref[1]**o+(0.70666908595)*x_ref[1]
        ref[5,0,0]=(0.639591398598)*x_ref[0]**o+(-0.052037663461)*x_ref[0]+(0.80293352341)*x_ref[1]**o+(0.373765944073)*x_ref[1]
        ref[5,0,1]=(0.278782801517)*x_ref[0]**o+(0.360704948502)*x_ref[0]+(0.518835994538)*x_ref[1]**o+(-0.748058641453)*x_ref[1]
        ref[5,1,0]=(0.376125218461)*x_ref[0]**o+(0.587328226457)*x_ref[0]+(-0.408741695392)*x_ref[1]**o+(-0.922372017858)*x_ref[1]
        ref[5,1,1]=(0.921904823293)*x_ref[0]**o+(-0.997993840905)*x_ref[0]+(-0.569469660555)*x_ref[1]**o+(0.246392170853)*x_ref[1]
      else:
        s[0,0,0]=(-0.215837732064)*x[0]**o+(0.708293902667)*x[0]+(-0.327646650956)*x[1]**o+(0.414511934051)*x[1]+(-0.163886757205)*x[2]**o+(-0.204189031155)*x[2]
        s[0,0,1]=(-0.731839397519)*x[0]**o+(-0.639212394687)*x[0]+(0.795699599563)*x[1]**o+(-0.487257713171)*x[1]+(0.390417080366)*x[2]**o+(0.0996253411809)*x[2]
        s[0,1,0]=(0.81024079229)*x[0]**o+(-0.508591074719)*x[0]+(0.591387161708)*x[1]**o+(-0.106008048523)*x[1]+(-0.865341838628)*x[2]**o+(0.636398034627)*x[2]
        s[0,1,1]=(0.107194327861)*x[0]**o+(-0.843547507026)*x[0]+(-0.797015379948)*x[1]**o+(0.92022128539)*x[1]+(-0.724414329973)*x[2]**o+(0.666266236918)*x[2]
        s[1,0,0]=(0.643940188921)*x[0]**o+(-0.365702296362)*x[0]+(0.808084016481)*x[1]**o+(-0.811845121877)*x[1]+(-0.948680432857)*x[2]**o+(-0.751105519587)*x[2]
        s[1,0,1]=(-0.365910713276)*x[0]**o+(-0.258101370414)*x[0]+(-0.376112129731)*x[1]**o+(0.969279125776)*x[1]+(-0.574916289815)*x[2]**o+(0.192443729433)*x[2]
        s[1,1,0]=(0.246180930634)*x[0]**o+(-0.65550133767)*x[0]+(-0.631929004343)*x[1]**o+(-0.270752141343)*x[1]+(-0.55380695748)*x[2]**o+(-0.0650898973585)*x[2]
        s[1,1,1]=(-0.624153597329)*x[0]**o+(0.471012703457)*x[0]+(-0.421626259377)*x[1]**o+(0.907052626802)*x[1]+(-0.943845344146)*x[2]**o+(0.847485825383)*x[2]
        s[2,0,0]=(-0.114129908207)*x[0]**o+(0.0887282187742)*x[0]+(-0.263100777717)*x[1]**o+(-0.813318704825)*x[1]+(-0.921371769744)*x[2]**o+(0.315384876801)*x[2]
        s[2,0,1]=(0.00919199747403)*x[0]**o+(-0.839076410274)*x[0]+(0.674300968816)*x[1]**o+(-0.266159980117)*x[1]+(-0.754781416929)*x[2]**o+(-0.330466108009)*x[2]
        s[2,1,0]=(0.269339357806)*x[0]**o+(0.978317771215)*x[0]+(0.483559805886)*x[1]**o+(-0.34789105417)*x[1]+(-0.301294669211)*x[2]**o+(-0.67971772971)*x[2]
        s[2,1,1]=(0.384387354744)*x[0]**o+(-0.127357593581)*x[0]+(0.567748162535)*x[1]**o+(0.312028506128)*x[1]+(0.906909215159)*x[2]**o+(-0.157017635377)*x[2]
        s[3,0,0]=(0.364487214036)*x[0]**o+(-0.0694000264348)*x[0]+(0.752067987858)*x[1]**o+(-0.0777040114767)*x[1]+(-0.536354586822)*x[2]**o+(0.846950905582)*x[2]
        s[3,0,1]=(0.215378827608)*x[0]**o+(-0.495485213939)*x[0]+(0.804796373941)*x[1]**o+(-0.948743952727)*x[1]+(0.562922488996)*x[2]**o+(0.611879748265)*x[2]
        s[3,1,0]=(-0.357559056788)*x[0]**o+(-0.966706464642)*x[0]+(-0.279364080486)*x[1]**o+(0.151836363465)*x[1]+(-0.0666607800869)*x[2]**o+(-0.155985265266)*x[2]
        s[3,1,1]=(-0.620991172464)*x[0]**o+(0.838549798787)*x[0]+(-0.69596391819)*x[1]**o+(0.345156777898)*x[1]+(-0.807545685144)*x[2]**o+(-0.343627808568)*x[2]
        s[4,0,0]=(0.752865578568)*x[0]**o+(0.0181463586995)*x[0]+(0.289423717925)*x[1]**o+(0.458330445591)*x[1]+(0.642952438313)*x[2]**o+(-0.302592553525)*x[2]
        s[4,0,1]=(0.38497342932)*x[0]**o+(0.197638554033)*x[0]+(-0.19259192208)*x[1]**o+(0.730648627642)*x[1]+(-0.684340024218)*x[2]**o+(-0.316428219221)*x[2]
        s[4,1,0]=(0.723339041107)*x[0]**o+(0.625053549512)*x[0]+(0.82493053308)*x[1]**o+(-0.222874319966)*x[1]+(0.0283928465938)*x[2]**o+(-0.262980725333)*x[2]
        s[4,1,1]=(0.527134347243)*x[0]**o+(-0.83920533899)*x[0]+(0.623982181406)*x[1]**o+(0.140699355979)*x[1]+(-0.189351111714)*x[2]**o+(0.133204166349)*x[2]
        s[5,0,0]=(0.155705059814)*x[0]**o+(-0.672186066824)*x[0]+(-0.279849819358)*x[1]**o+(-0.95943658469)*x[1]+(-0.750042065446)*x[2]**o+(-0.677006630121)*x[2]
        s[5,0,1]=(0.955099324881)*x[0]**o+(-0.609207748852)*x[0]+(-0.480726753052)*x[1]**o+(-0.110097202271)*x[1]+(-0.293113386852)*x[2]**o+(0.0810841023923)*x[2]
        s[5,1,0]=(-0.0858192742373)*x[0]**o+(0.456041657037)*x[0]+(-0.560905002533)*x[1]**o+(0.211057417206)*x[1]+(-0.296433850136)*x[2]**o+(-0.153349582946)*x[2]
        s[5,1,1]=(-0.914667297428)*x[0]**o+(0.856653497665)*x[0]+(0.994731884351)*x[1]**o+(-0.221838704365)*x[1]+(0.0905489452636)*x[2]**o+(0.937047976959)*x[2]
        ref[0,0,0]=(-0.215837732064)*x_ref[0]**o+(0.708293902667)*x_ref[0]+(-0.327646650956)*x_ref[1]**o+(0.414511934051)*x_ref[1]+(-0.163886757205)*x_ref[2]**o+(-0.204189031155)*x_ref[2]
        ref[0,0,1]=(-0.731839397519)*x_ref[0]**o+(-0.639212394687)*x_ref[0]+(0.795699599563)*x_ref[1]**o+(-0.487257713171)*x_ref[1]+(0.390417080366)*x_ref[2]**o+(0.0996253411809)*x_ref[2]
        ref[0,1,0]=(0.81024079229)*x_ref[0]**o+(-0.508591074719)*x_ref[0]+(0.591387161708)*x_ref[1]**o+(-0.106008048523)*x_ref[1]+(-0.865341838628)*x_ref[2]**o+(0.636398034627)*x_ref[2]
        ref[0,1,1]=(0.107194327861)*x_ref[0]**o+(-0.843547507026)*x_ref[0]+(-0.797015379948)*x_ref[1]**o+(0.92022128539)*x_ref[1]+(-0.724414329973)*x_ref[2]**o+(0.666266236918)*x_ref[2]
        ref[1,0,0]=(0.643940188921)*x_ref[0]**o+(-0.365702296362)*x_ref[0]+(0.808084016481)*x_ref[1]**o+(-0.811845121877)*x_ref[1]+(-0.948680432857)*x_ref[2]**o+(-0.751105519587)*x_ref[2]
        ref[1,0,1]=(-0.365910713276)*x_ref[0]**o+(-0.258101370414)*x_ref[0]+(-0.376112129731)*x_ref[1]**o+(0.969279125776)*x_ref[1]+(-0.574916289815)*x_ref[2]**o+(0.192443729433)*x_ref[2]
        ref[1,1,0]=(0.246180930634)*x_ref[0]**o+(-0.65550133767)*x_ref[0]+(-0.631929004343)*x_ref[1]**o+(-0.270752141343)*x_ref[1]+(-0.55380695748)*x_ref[2]**o+(-0.0650898973585)*x_ref[2]
        ref[1,1,1]=(-0.624153597329)*x_ref[0]**o+(0.471012703457)*x_ref[0]+(-0.421626259377)*x_ref[1]**o+(0.907052626802)*x_ref[1]+(-0.943845344146)*x_ref[2]**o+(0.847485825383)*x_ref[2]
        ref[2,0,0]=(-0.114129908207)*x_ref[0]**o+(0.0887282187742)*x_ref[0]+(-0.263100777717)*x_ref[1]**o+(-0.813318704825)*x_ref[1]+(-0.921371769744)*x_ref[2]**o+(0.315384876801)*x_ref[2]
        ref[2,0,1]=(0.00919199747403)*x_ref[0]**o+(-0.839076410274)*x_ref[0]+(0.674300968816)*x_ref[1]**o+(-0.266159980117)*x_ref[1]+(-0.754781416929)*x_ref[2]**o+(-0.330466108009)*x_ref[2]
        ref[2,1,0]=(0.269339357806)*x_ref[0]**o+(0.978317771215)*x_ref[0]+(0.483559805886)*x_ref[1]**o+(-0.34789105417)*x_ref[1]+(-0.301294669211)*x_ref[2]**o+(-0.67971772971)*x_ref[2]
        ref[2,1,1]=(0.384387354744)*x_ref[0]**o+(-0.127357593581)*x_ref[0]+(0.567748162535)*x_ref[1]**o+(0.312028506128)*x_ref[1]+(0.906909215159)*x_ref[2]**o+(-0.157017635377)*x_ref[2]
        ref[3,0,0]=(0.364487214036)*x_ref[0]**o+(-0.0694000264348)*x_ref[0]+(0.752067987858)*x_ref[1]**o+(-0.0777040114767)*x_ref[1]+(-0.536354586822)*x_ref[2]**o+(0.846950905582)*x_ref[2]
        ref[3,0,1]=(0.215378827608)*x_ref[0]**o+(-0.495485213939)*x_ref[0]+(0.804796373941)*x_ref[1]**o+(-0.948743952727)*x_ref[1]+(0.562922488996)*x_ref[2]**o+(0.611879748265)*x_ref[2]
        ref[3,1,0]=(-0.357559056788)*x_ref[0]**o+(-0.966706464642)*x_ref[0]+(-0.279364080486)*x_ref[1]**o+(0.151836363465)*x_ref[1]+(-0.0666607800869)*x_ref[2]**o+(-0.155985265266)*x_ref[2]
        ref[3,1,1]=(-0.620991172464)*x_ref[0]**o+(0.838549798787)*x_ref[0]+(-0.69596391819)*x_ref[1]**o+(0.345156777898)*x_ref[1]+(-0.807545685144)*x_ref[2]**o+(-0.343627808568)*x_ref[2]
        ref[4,0,0]=(0.752865578568)*x_ref[0]**o+(0.0181463586995)*x_ref[0]+(0.289423717925)*x_ref[1]**o+(0.458330445591)*x_ref[1]+(0.642952438313)*x_ref[2]**o+(-0.302592553525)*x_ref[2]
        ref[4,0,1]=(0.38497342932)*x_ref[0]**o+(0.197638554033)*x_ref[0]+(-0.19259192208)*x_ref[1]**o+(0.730648627642)*x_ref[1]+(-0.684340024218)*x_ref[2]**o+(-0.316428219221)*x_ref[2]
        ref[4,1,0]=(0.723339041107)*x_ref[0]**o+(0.625053549512)*x_ref[0]+(0.82493053308)*x_ref[1]**o+(-0.222874319966)*x_ref[1]+(0.0283928465938)*x_ref[2]**o+(-0.262980725333)*x_ref[2]
        ref[4,1,1]=(0.527134347243)*x_ref[0]**o+(-0.83920533899)*x_ref[0]+(0.623982181406)*x_ref[1]**o+(0.140699355979)*x_ref[1]+(-0.189351111714)*x_ref[2]**o+(0.133204166349)*x_ref[2]
        ref[5,0,0]=(0.155705059814)*x_ref[0]**o+(-0.672186066824)*x_ref[0]+(-0.279849819358)*x_ref[1]**o+(-0.95943658469)*x_ref[1]+(-0.750042065446)*x_ref[2]**o+(-0.677006630121)*x_ref[2]
        ref[5,0,1]=(0.955099324881)*x_ref[0]**o+(-0.609207748852)*x_ref[0]+(-0.480726753052)*x_ref[1]**o+(-0.110097202271)*x_ref[1]+(-0.293113386852)*x_ref[2]**o+(0.0810841023923)*x_ref[2]
        ref[5,1,0]=(-0.0858192742373)*x_ref[0]**o+(0.456041657037)*x_ref[0]+(-0.560905002533)*x_ref[1]**o+(0.211057417206)*x_ref[1]+(-0.296433850136)*x_ref[2]**o+(-0.153349582946)*x_ref[2]
        ref[5,1,1]=(-0.914667297428)*x_ref[0]**o+(0.856653497665)*x_ref[0]+(0.994731884351)*x_ref[1]**o+(-0.221838704365)*x_ref[1]+(0.0905489452636)*x_ref[2]**o+(0.937047976959)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.0777636614651)*x[0]**o+(-0.0667428739907)*x[0]+(-0.977685288599)*x[1]**o+(0.520832422148)*x[1]
        s[0,0,0,1]=(0.0801946653417)*x[0]**o+(0.391444443109)*x[0]+(0.60513760227)*x[1]**o+(-0.202405127189)*x[1]
        s[0,0,1,0]=(-0.107845479947)*x[0]**o+(-0.603283099882)*x[0]+(-0.648914480134)*x[1]**o+(0.134055978332)*x[1]
        s[0,0,1,1]=(-0.307399325623)*x[0]**o+(0.740265076637)*x[0]+(-0.426634593638)*x[1]**o+(0.524782346481)*x[1]
        s[0,0,2,0]=(0.209544366051)*x[0]**o+(0.456746428209)*x[0]+(0.746428003703)*x[1]**o+(0.891861715667)*x[1]
        s[0,0,2,1]=(0.67352776723)*x[0]**o+(0.087047388831)*x[0]+(0.983599793566)*x[1]**o+(-0.711447053356)*x[1]
        s[0,1,0,0]=(-0.155608608254)*x[0]**o+(-0.563919705574)*x[0]+(-0.819109501425)*x[1]**o+(-0.307153392665)*x[1]
        s[0,1,0,1]=(-0.837016003308)*x[0]**o+(0.759786384968)*x[0]+(-0.286876836113)*x[1]**o+(-0.450054346913)*x[1]
        s[0,1,1,0]=(0.317132680973)*x[0]**o+(-0.541058603922)*x[0]+(-0.438996048591)*x[1]**o+(-0.0804910701489)*x[1]
        s[0,1,1,1]=(-0.383941172325)*x[0]**o+(0.14081095151)*x[0]+(-0.964196329254)*x[1]**o+(0.137424835923)*x[1]
        s[0,1,2,0]=(0.78404512518)*x[0]**o+(-0.618654770204)*x[0]+(-0.120723400095)*x[1]**o+(0.337084654664)*x[1]
        s[0,1,2,1]=(0.370706702266)*x[0]**o+(0.362674269383)*x[0]+(0.129330112286)*x[1]**o+(0.27594329828)*x[1]
        s[0,2,0,0]=(-0.0618534907888)*x[0]**o+(-0.36997399231)*x[0]+(-0.701320112334)*x[1]**o+(0.587057599723)*x[1]
        s[0,2,0,1]=(-0.00083395539837)*x[0]**o+(-0.316061641051)*x[0]+(0.888164277705)*x[1]**o+(-0.708426734155)*x[1]
        s[0,2,1,0]=(-0.882600666018)*x[0]**o+(-0.734482083444)*x[0]+(-0.15689139088)*x[1]**o+(0.885520053714)*x[1]
        s[0,2,1,1]=(0.307710272644)*x[0]**o+(-0.343736402048)*x[0]+(0.00525340811259)*x[1]**o+(-0.365110058652)*x[1]
        s[0,2,2,0]=(-0.477696915031)*x[0]**o+(-0.636820200526)*x[0]+(-0.187124878764)*x[1]**o+(0.353033803175)*x[1]
        s[0,2,2,1]=(0.255079299946)*x[0]**o+(-0.132216578075)*x[0]+(0.540401170028)*x[1]**o+(-0.502440213664)*x[1]
        s[0,3,0,0]=(-0.0567446607807)*x[0]**o+(-0.859056478176)*x[0]+(0.258047569261)*x[1]**o+(-0.197460307626)*x[1]
        s[0,3,0,1]=(-0.412568633139)*x[0]**o+(0.143081394877)*x[0]+(-0.958152901527)*x[1]**o+(-0.946724885421)*x[1]
        s[0,3,1,0]=(0.746006182263)*x[0]**o+(0.415697572299)*x[0]+(0.633094060917)*x[1]**o+(-0.598666742083)*x[1]
        s[0,3,1,1]=(0.820559555653)*x[0]**o+(0.0393656665181)*x[0]+(-0.477283875434)*x[1]**o+(-0.629531033295)*x[1]
        s[0,3,2,0]=(-0.79478196151)*x[0]**o+(-0.682204178506)*x[0]+(0.920770217929)*x[1]**o+(-0.51779073656)*x[1]
        s[0,3,2,1]=(0.098810436334)*x[0]**o+(0.13601558071)*x[0]+(-0.900622849009)*x[1]**o+(0.358263179912)*x[1]
        s[0,4,0,0]=(0.636302027774)*x[0]**o+(-0.888743699041)*x[0]+(0.615869973466)*x[1]**o+(0.810532610229)*x[1]
        s[0,4,0,1]=(-0.350888533752)*x[0]**o+(0.430002320082)*x[0]+(0.184767543248)*x[1]**o+(-0.741038087278)*x[1]
        s[0,4,1,0]=(0.966040628072)*x[0]**o+(0.235593098824)*x[0]+(-0.217971115972)*x[1]**o+(0.211888609783)*x[1]
        s[0,4,1,1]=(-0.735728370129)*x[0]**o+(-0.890691876549)*x[0]+(0.0050361984863)*x[1]**o+(0.401300502845)*x[1]
        s[0,4,2,0]=(0.477438928895)*x[0]**o+(-0.0687120914109)*x[0]+(-0.822568493235)*x[1]**o+(0.692426496641)*x[1]
        s[0,4,2,1]=(-0.620965107536)*x[0]**o+(0.589395692553)*x[0]+(-0.330365701829)*x[1]**o+(0.752996781921)*x[1]
        s[1,0,0,0]=(0.983952631887)*x[0]**o+(-0.595120675382)*x[0]+(-0.834106811619)*x[1]**o+(-0.636013378457)*x[1]
        s[1,0,0,1]=(0.126740296939)*x[0]**o+(-0.520119640058)*x[0]+(-0.345481995205)*x[1]**o+(0.0433678023223)*x[1]
        s[1,0,1,0]=(0.565690778246)*x[0]**o+(0.867963746632)*x[0]+(-0.725334931245)*x[1]**o+(-0.620431577357)*x[1]
        s[1,0,1,1]=(0.191774597085)*x[0]**o+(-0.59036914479)*x[0]+(-0.277730803126)*x[1]**o+(0.909240521676)*x[1]
        s[1,0,2,0]=(-0.479427633319)*x[0]**o+(0.758085256512)*x[0]+(0.86321953036)*x[1]**o+(0.789915199597)*x[1]
        s[1,0,2,1]=(0.478767279402)*x[0]**o+(0.061686054009)*x[0]+(0.403654858589)*x[1]**o+(0.413582841972)*x[1]
        s[1,1,0,0]=(0.839949350991)*x[0]**o+(0.083434136022)*x[0]+(-0.237548283201)*x[1]**o+(0.340524677978)*x[1]
        s[1,1,0,1]=(-0.44615314004)*x[0]**o+(-0.0748079090653)*x[0]+(0.669071324293)*x[1]**o+(-0.00926065712951)*x[1]
        s[1,1,1,0]=(0.634663494412)*x[0]**o+(0.414921856585)*x[0]+(0.656468620085)*x[1]**o+(0.154428434525)*x[1]
        s[1,1,1,1]=(0.120264291322)*x[0]**o+(0.274387126898)*x[0]+(-0.914217101263)*x[1]**o+(0.509056318161)*x[1]
        s[1,1,2,0]=(-0.361610733139)*x[0]**o+(-0.419244707352)*x[0]+(-0.0936805656392)*x[1]**o+(0.658845522463)*x[1]
        s[1,1,2,1]=(-0.0152227030107)*x[0]**o+(0.716681107595)*x[0]+(0.339720150512)*x[1]**o+(0.623368924873)*x[1]
        s[1,2,0,0]=(-0.208625207667)*x[0]**o+(0.0892371419356)*x[0]+(-0.357406324393)*x[1]**o+(0.0545982207211)*x[1]
        s[1,2,0,1]=(-0.133416102554)*x[0]**o+(0.536329273122)*x[0]+(-0.600990988318)*x[1]**o+(-0.937230468764)*x[1]
        s[1,2,1,0]=(-0.549763100528)*x[0]**o+(-0.334287316265)*x[0]+(-0.889824442167)*x[1]**o+(0.164963321178)*x[1]
        s[1,2,1,1]=(-0.55705799726)*x[0]**o+(0.125817791326)*x[0]+(0.719976054813)*x[1]**o+(0.718873789349)*x[1]
        s[1,2,2,0]=(-0.73171705212)*x[0]**o+(0.734055655405)*x[0]+(0.202137285455)*x[1]**o+(0.684220988395)*x[1]
        s[1,2,2,1]=(-0.0397765024932)*x[0]**o+(-0.633490619367)*x[0]+(-0.195588405339)*x[1]**o+(-0.957674729264)*x[1]
        s[1,3,0,0]=(0.993464844097)*x[0]**o+(0.234619248397)*x[0]+(-0.51343997173)*x[1]**o+(0.856105886668)*x[1]
        s[1,3,0,1]=(0.0811328827072)*x[0]**o+(0.12838793682)*x[0]+(-0.522572607464)*x[1]**o+(-0.375599526602)*x[1]
        s[1,3,1,0]=(0.826184052821)*x[0]**o+(-0.050690194008)*x[0]+(0.13981816603)*x[1]**o+(0.00952178435275)*x[1]
        s[1,3,1,1]=(0.217029350446)*x[0]**o+(0.714803047883)*x[0]+(-0.957500687556)*x[1]**o+(-0.181988303734)*x[1]
        s[1,3,2,0]=(-0.380490770961)*x[0]**o+(0.279064641297)*x[0]+(0.587953927724)*x[1]**o+(0.0967128120583)*x[1]
        s[1,3,2,1]=(-0.563049792849)*x[0]**o+(0.179929757095)*x[0]+(0.256239332721)*x[1]**o+(0.719366227917)*x[1]
        s[1,4,0,0]=(-0.935374185099)*x[0]**o+(0.686212475195)*x[0]+(-0.696243048462)*x[1]**o+(-0.320962125289)*x[1]
        s[1,4,0,1]=(0.877754242018)*x[0]**o+(-0.131209004599)*x[0]+(-0.665712708561)*x[1]**o+(0.746518141138)*x[1]
        s[1,4,1,0]=(0.734524693133)*x[0]**o+(-0.379930095743)*x[0]+(0.934573403823)*x[1]**o+(-0.00810739306485)*x[1]
        s[1,4,1,1]=(0.0822815241388)*x[0]**o+(0.662617104334)*x[0]+(-0.0794803324896)*x[1]**o+(0.331645391345)*x[1]
        s[1,4,2,0]=(0.232252551171)*x[0]**o+(0.814128846678)*x[0]+(-0.628231676896)*x[1]**o+(-0.702203701604)*x[1]
        s[1,4,2,1]=(-0.700532752151)*x[0]**o+(0.277528305962)*x[0]+(-0.0521465863885)*x[1]**o+(-0.31154867965)*x[1]
        s[2,0,0,0]=(0.472871337554)*x[0]**o+(0.413549495857)*x[0]+(0.311158506504)*x[1]**o+(0.198806166981)*x[1]
        s[2,0,0,1]=(0.0340811960202)*x[0]**o+(-0.00583425318175)*x[0]+(-0.750117294155)*x[1]**o+(-0.905524986631)*x[1]
        s[2,0,1,0]=(0.570790284556)*x[0]**o+(0.673577393533)*x[0]+(-0.325056654336)*x[1]**o+(0.653038510272)*x[1]
        s[2,0,1,1]=(-0.297592197799)*x[0]**o+(0.871440030306)*x[0]+(-0.533637236239)*x[1]**o+(-0.4099903042)*x[1]
        s[2,0,2,0]=(0.293978916599)*x[0]**o+(-0.0182866541826)*x[0]+(0.843245387706)*x[1]**o+(-0.139447965404)*x[1]
        s[2,0,2,1]=(-0.10091027426)*x[0]**o+(-0.100495744095)*x[0]+(-0.0141975035033)*x[1]**o+(-0.375812318845)*x[1]
        s[2,1,0,0]=(-0.830824351182)*x[0]**o+(-0.972181712173)*x[0]+(-0.60270773635)*x[1]**o+(-0.449692597572)*x[1]
        s[2,1,0,1]=(0.580664484955)*x[0]**o+(0.453307511725)*x[0]+(-0.69006079487)*x[1]**o+(-0.282709994647)*x[1]
        s[2,1,1,0]=(0.0350693872862)*x[0]**o+(0.551895773929)*x[0]+(0.946054026898)*x[1]**o+(0.0805073878157)*x[1]
        s[2,1,1,1]=(-0.996166074405)*x[0]**o+(-0.0291388611313)*x[0]+(0.107628876337)*x[1]**o+(0.181623761662)*x[1]
        s[2,1,2,0]=(0.878542357796)*x[0]**o+(0.0263960313825)*x[0]+(-0.470099210068)*x[1]**o+(0.775356436208)*x[1]
        s[2,1,2,1]=(0.771458787313)*x[0]**o+(-0.238120631661)*x[0]+(0.0545258501919)*x[1]**o+(0.178539047763)*x[1]
        s[2,2,0,0]=(0.398263199258)*x[0]**o+(0.858180962314)*x[0]+(-0.00348131786996)*x[1]**o+(-0.958044163072)*x[1]
        s[2,2,0,1]=(0.776760912484)*x[0]**o+(0.364111724681)*x[0]+(0.26458908016)*x[1]**o+(-0.0590754024201)*x[1]
        s[2,2,1,0]=(0.704662231933)*x[0]**o+(-0.954239844678)*x[0]+(-0.212398080843)*x[1]**o+(-0.547628870359)*x[1]
        s[2,2,1,1]=(0.741611255723)*x[0]**o+(0.963847765909)*x[0]+(-0.943663876716)*x[1]**o+(0.745169950034)*x[1]
        s[2,2,2,0]=(-0.818935776346)*x[0]**o+(-0.2967013158)*x[0]+(0.179993824489)*x[1]**o+(0.061135531911)*x[1]
        s[2,2,2,1]=(0.704463632969)*x[0]**o+(0.49682411039)*x[0]+(0.522725375403)*x[1]**o+(-0.0991411582378)*x[1]
        s[2,3,0,0]=(-0.394678165334)*x[0]**o+(-0.490081142088)*x[0]+(-0.437269252729)*x[1]**o+(0.437548658318)*x[1]
        s[2,3,0,1]=(-0.8935046735)*x[0]**o+(-0.0514148814426)*x[0]+(-0.758657504339)*x[1]**o+(-0.91790254494)*x[1]
        s[2,3,1,0]=(0.678673237528)*x[0]**o+(-0.820565721442)*x[0]+(0.758797168017)*x[1]**o+(-0.915258841164)*x[1]
        s[2,3,1,1]=(-0.70927157191)*x[0]**o+(-0.742658350028)*x[0]+(0.111130223627)*x[1]**o+(0.179307074639)*x[1]
        s[2,3,2,0]=(0.228074760628)*x[0]**o+(0.533399652471)*x[0]+(0.560064736493)*x[1]**o+(-0.588589967203)*x[1]
        s[2,3,2,1]=(0.994518467999)*x[0]**o+(-0.201788029147)*x[0]+(-0.983374807613)*x[1]**o+(-0.386854409534)*x[1]
        s[2,4,0,0]=(0.763484049607)*x[0]**o+(-0.262913676303)*x[0]+(-0.498088896053)*x[1]**o+(-0.672931052896)*x[1]
        s[2,4,0,1]=(-0.46331695102)*x[0]**o+(0.91157906649)*x[0]+(-0.647323300295)*x[1]**o+(-0.309269630073)*x[1]
        s[2,4,1,0]=(-0.0526757673514)*x[0]**o+(-0.578131288345)*x[0]+(0.39711966866)*x[1]**o+(0.804245669408)*x[1]
        s[2,4,1,1]=(0.223417129057)*x[0]**o+(-0.923622687767)*x[0]+(0.618075111678)*x[1]**o+(0.381994832856)*x[1]
        s[2,4,2,0]=(0.540391658848)*x[0]**o+(0.531453650116)*x[0]+(-0.146608482269)*x[1]**o+(0.940324004341)*x[1]
        s[2,4,2,1]=(-0.986820975671)*x[0]**o+(0.00121909081031)*x[0]+(-0.697255554926)*x[1]**o+(0.967740580108)*x[1]
        s[3,0,0,0]=(-0.0658669634781)*x[0]**o+(0.70728861958)*x[0]+(0.966340481481)*x[1]**o+(0.986894244456)*x[1]
        s[3,0,0,1]=(0.721157686077)*x[0]**o+(-0.0629283817117)*x[0]+(-0.965971306999)*x[1]**o+(0.291794228985)*x[1]
        s[3,0,1,0]=(-0.360437990219)*x[0]**o+(-0.570216212399)*x[0]+(-0.607512137662)*x[1]**o+(-0.515098731307)*x[1]
        s[3,0,1,1]=(0.642086516897)*x[0]**o+(0.52547404535)*x[0]+(0.016876402075)*x[1]**o+(-0.862152788234)*x[1]
        s[3,0,2,0]=(-0.181658178552)*x[0]**o+(-0.203447330476)*x[0]+(-0.547194819622)*x[1]**o+(-0.0593895310488)*x[1]
        s[3,0,2,1]=(-0.582071301036)*x[0]**o+(-0.557616629544)*x[0]+(-0.60723171731)*x[1]**o+(0.204081805298)*x[1]
        s[3,1,0,0]=(0.272716289788)*x[0]**o+(-0.947135640157)*x[0]+(-0.25901560402)*x[1]**o+(-0.0247786104764)*x[1]
        s[3,1,0,1]=(-0.891714371313)*x[0]**o+(0.755281444278)*x[0]+(0.351999986261)*x[1]**o+(0.661376231103)*x[1]
        s[3,1,1,0]=(0.432312310444)*x[0]**o+(-0.0375106988813)*x[0]+(0.843778132955)*x[1]**o+(-0.787134076357)*x[1]
        s[3,1,1,1]=(-0.57131496233)*x[0]**o+(-0.614200739152)*x[0]+(-0.0817103577524)*x[1]**o+(-0.0809087761527)*x[1]
        s[3,1,2,0]=(0.775899774689)*x[0]**o+(0.650087297096)*x[0]+(-0.363483124762)*x[1]**o+(0.0406785055243)*x[1]
        s[3,1,2,1]=(-0.8228824793)*x[0]**o+(0.0223028194838)*x[0]+(-0.246990198359)*x[1]**o+(-0.447554242783)*x[1]
        s[3,2,0,0]=(0.250699655536)*x[0]**o+(0.425847139574)*x[0]+(0.185272955249)*x[1]**o+(0.52788179093)*x[1]
        s[3,2,0,1]=(-0.6916019182)*x[0]**o+(0.468697041897)*x[0]+(0.0929416146358)*x[1]**o+(0.0632517742213)*x[1]
        s[3,2,1,0]=(-0.0137715023834)*x[0]**o+(-0.876273608156)*x[0]+(0.333020658958)*x[1]**o+(-0.546002404224)*x[1]
        s[3,2,1,1]=(-0.0260174553873)*x[0]**o+(0.668170105508)*x[0]+(-0.515918906011)*x[1]**o+(0.427775146345)*x[1]
        s[3,2,2,0]=(0.451288901693)*x[0]**o+(-0.307365533804)*x[0]+(-0.902054304556)*x[1]**o+(0.561501391872)*x[1]
        s[3,2,2,1]=(-0.67166870387)*x[0]**o+(-0.851949794629)*x[0]+(-0.198839514379)*x[1]**o+(-0.0539885252629)*x[1]
        s[3,3,0,0]=(0.319371411392)*x[0]**o+(0.521157698767)*x[0]+(0.0619379731457)*x[1]**o+(0.85473464537)*x[1]
        s[3,3,0,1]=(0.605988887111)*x[0]**o+(-0.75556314419)*x[0]+(-0.551855748356)*x[1]**o+(0.152400883628)*x[1]
        s[3,3,1,0]=(-0.963972268356)*x[0]**o+(-0.221723218793)*x[0]+(-0.940231983732)*x[1]**o+(0.509102233847)*x[1]
        s[3,3,1,1]=(0.934837503535)*x[0]**o+(-0.762674616346)*x[0]+(0.196730758209)*x[1]**o+(0.93895486771)*x[1]
        s[3,3,2,0]=(0.683137246803)*x[0]**o+(0.638846953858)*x[0]+(0.508007741206)*x[1]**o+(0.625453420423)*x[1]
        s[3,3,2,1]=(-0.0951254112087)*x[0]**o+(0.507174561327)*x[0]+(0.98495014583)*x[1]**o+(0.922324997316)*x[1]
        s[3,4,0,0]=(-0.39889705668)*x[0]**o+(-0.00933499671216)*x[0]+(-0.308567417711)*x[1]**o+(0.0365845447897)*x[1]
        s[3,4,0,1]=(0.699792659761)*x[0]**o+(-0.339911160323)*x[0]+(-0.0380630017731)*x[1]**o+(0.891498223182)*x[1]
        s[3,4,1,0]=(-0.486198638614)*x[0]**o+(0.910921882689)*x[0]+(0.544226196572)*x[1]**o+(0.221206282494)*x[1]
        s[3,4,1,1]=(0.0656938900284)*x[0]**o+(-0.146193696097)*x[0]+(0.580985806286)*x[1]**o+(0.913369274555)*x[1]
        s[3,4,2,0]=(-0.31430372984)*x[0]**o+(0.132101392811)*x[0]+(0.823927100539)*x[1]**o+(-0.0349106467441)*x[1]
        s[3,4,2,1]=(0.454900950832)*x[0]**o+(0.875474839303)*x[0]+(-0.496859632101)*x[1]**o+(0.143908936177)*x[1]
        ref[0,0,0,0]=(-0.0777636614651)*x_ref[0]**o+(-0.0667428739907)*x_ref[0]+(-0.977685288599)*x_ref[1]**o+(0.520832422148)*x_ref[1]
        ref[0,0,0,1]=(0.0801946653417)*x_ref[0]**o+(0.391444443109)*x_ref[0]+(0.60513760227)*x_ref[1]**o+(-0.202405127189)*x_ref[1]
        ref[0,0,1,0]=(-0.107845479947)*x_ref[0]**o+(-0.603283099882)*x_ref[0]+(-0.648914480134)*x_ref[1]**o+(0.134055978332)*x_ref[1]
        ref[0,0,1,1]=(-0.307399325623)*x_ref[0]**o+(0.740265076637)*x_ref[0]+(-0.426634593638)*x_ref[1]**o+(0.524782346481)*x_ref[1]
        ref[0,0,2,0]=(0.209544366051)*x_ref[0]**o+(0.456746428209)*x_ref[0]+(0.746428003703)*x_ref[1]**o+(0.891861715667)*x_ref[1]
        ref[0,0,2,1]=(0.67352776723)*x_ref[0]**o+(0.087047388831)*x_ref[0]+(0.983599793566)*x_ref[1]**o+(-0.711447053356)*x_ref[1]
        ref[0,1,0,0]=(-0.155608608254)*x_ref[0]**o+(-0.563919705574)*x_ref[0]+(-0.819109501425)*x_ref[1]**o+(-0.307153392665)*x_ref[1]
        ref[0,1,0,1]=(-0.837016003308)*x_ref[0]**o+(0.759786384968)*x_ref[0]+(-0.286876836113)*x_ref[1]**o+(-0.450054346913)*x_ref[1]
        ref[0,1,1,0]=(0.317132680973)*x_ref[0]**o+(-0.541058603922)*x_ref[0]+(-0.438996048591)*x_ref[1]**o+(-0.0804910701489)*x_ref[1]
        ref[0,1,1,1]=(-0.383941172325)*x_ref[0]**o+(0.14081095151)*x_ref[0]+(-0.964196329254)*x_ref[1]**o+(0.137424835923)*x_ref[1]
        ref[0,1,2,0]=(0.78404512518)*x_ref[0]**o+(-0.618654770204)*x_ref[0]+(-0.120723400095)*x_ref[1]**o+(0.337084654664)*x_ref[1]
        ref[0,1,2,1]=(0.370706702266)*x_ref[0]**o+(0.362674269383)*x_ref[0]+(0.129330112286)*x_ref[1]**o+(0.27594329828)*x_ref[1]
        ref[0,2,0,0]=(-0.0618534907888)*x_ref[0]**o+(-0.36997399231)*x_ref[0]+(-0.701320112334)*x_ref[1]**o+(0.587057599723)*x_ref[1]
        ref[0,2,0,1]=(-0.00083395539837)*x_ref[0]**o+(-0.316061641051)*x_ref[0]+(0.888164277705)*x_ref[1]**o+(-0.708426734155)*x_ref[1]
        ref[0,2,1,0]=(-0.882600666018)*x_ref[0]**o+(-0.734482083444)*x_ref[0]+(-0.15689139088)*x_ref[1]**o+(0.885520053714)*x_ref[1]
        ref[0,2,1,1]=(0.307710272644)*x_ref[0]**o+(-0.343736402048)*x_ref[0]+(0.00525340811259)*x_ref[1]**o+(-0.365110058652)*x_ref[1]
        ref[0,2,2,0]=(-0.477696915031)*x_ref[0]**o+(-0.636820200526)*x_ref[0]+(-0.187124878764)*x_ref[1]**o+(0.353033803175)*x_ref[1]
        ref[0,2,2,1]=(0.255079299946)*x_ref[0]**o+(-0.132216578075)*x_ref[0]+(0.540401170028)*x_ref[1]**o+(-0.502440213664)*x_ref[1]
        ref[0,3,0,0]=(-0.0567446607807)*x_ref[0]**o+(-0.859056478176)*x_ref[0]+(0.258047569261)*x_ref[1]**o+(-0.197460307626)*x_ref[1]
        ref[0,3,0,1]=(-0.412568633139)*x_ref[0]**o+(0.143081394877)*x_ref[0]+(-0.958152901527)*x_ref[1]**o+(-0.946724885421)*x_ref[1]
        ref[0,3,1,0]=(0.746006182263)*x_ref[0]**o+(0.415697572299)*x_ref[0]+(0.633094060917)*x_ref[1]**o+(-0.598666742083)*x_ref[1]
        ref[0,3,1,1]=(0.820559555653)*x_ref[0]**o+(0.0393656665181)*x_ref[0]+(-0.477283875434)*x_ref[1]**o+(-0.629531033295)*x_ref[1]
        ref[0,3,2,0]=(-0.79478196151)*x_ref[0]**o+(-0.682204178506)*x_ref[0]+(0.920770217929)*x_ref[1]**o+(-0.51779073656)*x_ref[1]
        ref[0,3,2,1]=(0.098810436334)*x_ref[0]**o+(0.13601558071)*x_ref[0]+(-0.900622849009)*x_ref[1]**o+(0.358263179912)*x_ref[1]
        ref[0,4,0,0]=(0.636302027774)*x_ref[0]**o+(-0.888743699041)*x_ref[0]+(0.615869973466)*x_ref[1]**o+(0.810532610229)*x_ref[1]
        ref[0,4,0,1]=(-0.350888533752)*x_ref[0]**o+(0.430002320082)*x_ref[0]+(0.184767543248)*x_ref[1]**o+(-0.741038087278)*x_ref[1]
        ref[0,4,1,0]=(0.966040628072)*x_ref[0]**o+(0.235593098824)*x_ref[0]+(-0.217971115972)*x_ref[1]**o+(0.211888609783)*x_ref[1]
        ref[0,4,1,1]=(-0.735728370129)*x_ref[0]**o+(-0.890691876549)*x_ref[0]+(0.0050361984863)*x_ref[1]**o+(0.401300502845)*x_ref[1]
        ref[0,4,2,0]=(0.477438928895)*x_ref[0]**o+(-0.0687120914109)*x_ref[0]+(-0.822568493235)*x_ref[1]**o+(0.692426496641)*x_ref[1]
        ref[0,4,2,1]=(-0.620965107536)*x_ref[0]**o+(0.589395692553)*x_ref[0]+(-0.330365701829)*x_ref[1]**o+(0.752996781921)*x_ref[1]
        ref[1,0,0,0]=(0.983952631887)*x_ref[0]**o+(-0.595120675382)*x_ref[0]+(-0.834106811619)*x_ref[1]**o+(-0.636013378457)*x_ref[1]
        ref[1,0,0,1]=(0.126740296939)*x_ref[0]**o+(-0.520119640058)*x_ref[0]+(-0.345481995205)*x_ref[1]**o+(0.0433678023223)*x_ref[1]
        ref[1,0,1,0]=(0.565690778246)*x_ref[0]**o+(0.867963746632)*x_ref[0]+(-0.725334931245)*x_ref[1]**o+(-0.620431577357)*x_ref[1]
        ref[1,0,1,1]=(0.191774597085)*x_ref[0]**o+(-0.59036914479)*x_ref[0]+(-0.277730803126)*x_ref[1]**o+(0.909240521676)*x_ref[1]
        ref[1,0,2,0]=(-0.479427633319)*x_ref[0]**o+(0.758085256512)*x_ref[0]+(0.86321953036)*x_ref[1]**o+(0.789915199597)*x_ref[1]
        ref[1,0,2,1]=(0.478767279402)*x_ref[0]**o+(0.061686054009)*x_ref[0]+(0.403654858589)*x_ref[1]**o+(0.413582841972)*x_ref[1]
        ref[1,1,0,0]=(0.839949350991)*x_ref[0]**o+(0.083434136022)*x_ref[0]+(-0.237548283201)*x_ref[1]**o+(0.340524677978)*x_ref[1]
        ref[1,1,0,1]=(-0.44615314004)*x_ref[0]**o+(-0.0748079090653)*x_ref[0]+(0.669071324293)*x_ref[1]**o+(-0.00926065712951)*x_ref[1]
        ref[1,1,1,0]=(0.634663494412)*x_ref[0]**o+(0.414921856585)*x_ref[0]+(0.656468620085)*x_ref[1]**o+(0.154428434525)*x_ref[1]
        ref[1,1,1,1]=(0.120264291322)*x_ref[0]**o+(0.274387126898)*x_ref[0]+(-0.914217101263)*x_ref[1]**o+(0.509056318161)*x_ref[1]
        ref[1,1,2,0]=(-0.361610733139)*x_ref[0]**o+(-0.419244707352)*x_ref[0]+(-0.0936805656392)*x_ref[1]**o+(0.658845522463)*x_ref[1]
        ref[1,1,2,1]=(-0.0152227030107)*x_ref[0]**o+(0.716681107595)*x_ref[0]+(0.339720150512)*x_ref[1]**o+(0.623368924873)*x_ref[1]
        ref[1,2,0,0]=(-0.208625207667)*x_ref[0]**o+(0.0892371419356)*x_ref[0]+(-0.357406324393)*x_ref[1]**o+(0.0545982207211)*x_ref[1]
        ref[1,2,0,1]=(-0.133416102554)*x_ref[0]**o+(0.536329273122)*x_ref[0]+(-0.600990988318)*x_ref[1]**o+(-0.937230468764)*x_ref[1]
        ref[1,2,1,0]=(-0.549763100528)*x_ref[0]**o+(-0.334287316265)*x_ref[0]+(-0.889824442167)*x_ref[1]**o+(0.164963321178)*x_ref[1]
        ref[1,2,1,1]=(-0.55705799726)*x_ref[0]**o+(0.125817791326)*x_ref[0]+(0.719976054813)*x_ref[1]**o+(0.718873789349)*x_ref[1]
        ref[1,2,2,0]=(-0.73171705212)*x_ref[0]**o+(0.734055655405)*x_ref[0]+(0.202137285455)*x_ref[1]**o+(0.684220988395)*x_ref[1]
        ref[1,2,2,1]=(-0.0397765024932)*x_ref[0]**o+(-0.633490619367)*x_ref[0]+(-0.195588405339)*x_ref[1]**o+(-0.957674729264)*x_ref[1]
        ref[1,3,0,0]=(0.993464844097)*x_ref[0]**o+(0.234619248397)*x_ref[0]+(-0.51343997173)*x_ref[1]**o+(0.856105886668)*x_ref[1]
        ref[1,3,0,1]=(0.0811328827072)*x_ref[0]**o+(0.12838793682)*x_ref[0]+(-0.522572607464)*x_ref[1]**o+(-0.375599526602)*x_ref[1]
        ref[1,3,1,0]=(0.826184052821)*x_ref[0]**o+(-0.050690194008)*x_ref[0]+(0.13981816603)*x_ref[1]**o+(0.00952178435275)*x_ref[1]
        ref[1,3,1,1]=(0.217029350446)*x_ref[0]**o+(0.714803047883)*x_ref[0]+(-0.957500687556)*x_ref[1]**o+(-0.181988303734)*x_ref[1]
        ref[1,3,2,0]=(-0.380490770961)*x_ref[0]**o+(0.279064641297)*x_ref[0]+(0.587953927724)*x_ref[1]**o+(0.0967128120583)*x_ref[1]
        ref[1,3,2,1]=(-0.563049792849)*x_ref[0]**o+(0.179929757095)*x_ref[0]+(0.256239332721)*x_ref[1]**o+(0.719366227917)*x_ref[1]
        ref[1,4,0,0]=(-0.935374185099)*x_ref[0]**o+(0.686212475195)*x_ref[0]+(-0.696243048462)*x_ref[1]**o+(-0.320962125289)*x_ref[1]
        ref[1,4,0,1]=(0.877754242018)*x_ref[0]**o+(-0.131209004599)*x_ref[0]+(-0.665712708561)*x_ref[1]**o+(0.746518141138)*x_ref[1]
        ref[1,4,1,0]=(0.734524693133)*x_ref[0]**o+(-0.379930095743)*x_ref[0]+(0.934573403823)*x_ref[1]**o+(-0.00810739306485)*x_ref[1]
        ref[1,4,1,1]=(0.0822815241388)*x_ref[0]**o+(0.662617104334)*x_ref[0]+(-0.0794803324896)*x_ref[1]**o+(0.331645391345)*x_ref[1]
        ref[1,4,2,0]=(0.232252551171)*x_ref[0]**o+(0.814128846678)*x_ref[0]+(-0.628231676896)*x_ref[1]**o+(-0.702203701604)*x_ref[1]
        ref[1,4,2,1]=(-0.700532752151)*x_ref[0]**o+(0.277528305962)*x_ref[0]+(-0.0521465863885)*x_ref[1]**o+(-0.31154867965)*x_ref[1]
        ref[2,0,0,0]=(0.472871337554)*x_ref[0]**o+(0.413549495857)*x_ref[0]+(0.311158506504)*x_ref[1]**o+(0.198806166981)*x_ref[1]
        ref[2,0,0,1]=(0.0340811960202)*x_ref[0]**o+(-0.00583425318175)*x_ref[0]+(-0.750117294155)*x_ref[1]**o+(-0.905524986631)*x_ref[1]
        ref[2,0,1,0]=(0.570790284556)*x_ref[0]**o+(0.673577393533)*x_ref[0]+(-0.325056654336)*x_ref[1]**o+(0.653038510272)*x_ref[1]
        ref[2,0,1,1]=(-0.297592197799)*x_ref[0]**o+(0.871440030306)*x_ref[0]+(-0.533637236239)*x_ref[1]**o+(-0.4099903042)*x_ref[1]
        ref[2,0,2,0]=(0.293978916599)*x_ref[0]**o+(-0.0182866541826)*x_ref[0]+(0.843245387706)*x_ref[1]**o+(-0.139447965404)*x_ref[1]
        ref[2,0,2,1]=(-0.10091027426)*x_ref[0]**o+(-0.100495744095)*x_ref[0]+(-0.0141975035033)*x_ref[1]**o+(-0.375812318845)*x_ref[1]
        ref[2,1,0,0]=(-0.830824351182)*x_ref[0]**o+(-0.972181712173)*x_ref[0]+(-0.60270773635)*x_ref[1]**o+(-0.449692597572)*x_ref[1]
        ref[2,1,0,1]=(0.580664484955)*x_ref[0]**o+(0.453307511725)*x_ref[0]+(-0.69006079487)*x_ref[1]**o+(-0.282709994647)*x_ref[1]
        ref[2,1,1,0]=(0.0350693872862)*x_ref[0]**o+(0.551895773929)*x_ref[0]+(0.946054026898)*x_ref[1]**o+(0.0805073878157)*x_ref[1]
        ref[2,1,1,1]=(-0.996166074405)*x_ref[0]**o+(-0.0291388611313)*x_ref[0]+(0.107628876337)*x_ref[1]**o+(0.181623761662)*x_ref[1]
        ref[2,1,2,0]=(0.878542357796)*x_ref[0]**o+(0.0263960313825)*x_ref[0]+(-0.470099210068)*x_ref[1]**o+(0.775356436208)*x_ref[1]
        ref[2,1,2,1]=(0.771458787313)*x_ref[0]**o+(-0.238120631661)*x_ref[0]+(0.0545258501919)*x_ref[1]**o+(0.178539047763)*x_ref[1]
        ref[2,2,0,0]=(0.398263199258)*x_ref[0]**o+(0.858180962314)*x_ref[0]+(-0.00348131786996)*x_ref[1]**o+(-0.958044163072)*x_ref[1]
        ref[2,2,0,1]=(0.776760912484)*x_ref[0]**o+(0.364111724681)*x_ref[0]+(0.26458908016)*x_ref[1]**o+(-0.0590754024201)*x_ref[1]
        ref[2,2,1,0]=(0.704662231933)*x_ref[0]**o+(-0.954239844678)*x_ref[0]+(-0.212398080843)*x_ref[1]**o+(-0.547628870359)*x_ref[1]
        ref[2,2,1,1]=(0.741611255723)*x_ref[0]**o+(0.963847765909)*x_ref[0]+(-0.943663876716)*x_ref[1]**o+(0.745169950034)*x_ref[1]
        ref[2,2,2,0]=(-0.818935776346)*x_ref[0]**o+(-0.2967013158)*x_ref[0]+(0.179993824489)*x_ref[1]**o+(0.061135531911)*x_ref[1]
        ref[2,2,2,1]=(0.704463632969)*x_ref[0]**o+(0.49682411039)*x_ref[0]+(0.522725375403)*x_ref[1]**o+(-0.0991411582378)*x_ref[1]
        ref[2,3,0,0]=(-0.394678165334)*x_ref[0]**o+(-0.490081142088)*x_ref[0]+(-0.437269252729)*x_ref[1]**o+(0.437548658318)*x_ref[1]
        ref[2,3,0,1]=(-0.8935046735)*x_ref[0]**o+(-0.0514148814426)*x_ref[0]+(-0.758657504339)*x_ref[1]**o+(-0.91790254494)*x_ref[1]
        ref[2,3,1,0]=(0.678673237528)*x_ref[0]**o+(-0.820565721442)*x_ref[0]+(0.758797168017)*x_ref[1]**o+(-0.915258841164)*x_ref[1]
        ref[2,3,1,1]=(-0.70927157191)*x_ref[0]**o+(-0.742658350028)*x_ref[0]+(0.111130223627)*x_ref[1]**o+(0.179307074639)*x_ref[1]
        ref[2,3,2,0]=(0.228074760628)*x_ref[0]**o+(0.533399652471)*x_ref[0]+(0.560064736493)*x_ref[1]**o+(-0.588589967203)*x_ref[1]
        ref[2,3,2,1]=(0.994518467999)*x_ref[0]**o+(-0.201788029147)*x_ref[0]+(-0.983374807613)*x_ref[1]**o+(-0.386854409534)*x_ref[1]
        ref[2,4,0,0]=(0.763484049607)*x_ref[0]**o+(-0.262913676303)*x_ref[0]+(-0.498088896053)*x_ref[1]**o+(-0.672931052896)*x_ref[1]
        ref[2,4,0,1]=(-0.46331695102)*x_ref[0]**o+(0.91157906649)*x_ref[0]+(-0.647323300295)*x_ref[1]**o+(-0.309269630073)*x_ref[1]
        ref[2,4,1,0]=(-0.0526757673514)*x_ref[0]**o+(-0.578131288345)*x_ref[0]+(0.39711966866)*x_ref[1]**o+(0.804245669408)*x_ref[1]
        ref[2,4,1,1]=(0.223417129057)*x_ref[0]**o+(-0.923622687767)*x_ref[0]+(0.618075111678)*x_ref[1]**o+(0.381994832856)*x_ref[1]
        ref[2,4,2,0]=(0.540391658848)*x_ref[0]**o+(0.531453650116)*x_ref[0]+(-0.146608482269)*x_ref[1]**o+(0.940324004341)*x_ref[1]
        ref[2,4,2,1]=(-0.986820975671)*x_ref[0]**o+(0.00121909081031)*x_ref[0]+(-0.697255554926)*x_ref[1]**o+(0.967740580108)*x_ref[1]
        ref[3,0,0,0]=(-0.0658669634781)*x_ref[0]**o+(0.70728861958)*x_ref[0]+(0.966340481481)*x_ref[1]**o+(0.986894244456)*x_ref[1]
        ref[3,0,0,1]=(0.721157686077)*x_ref[0]**o+(-0.0629283817117)*x_ref[0]+(-0.965971306999)*x_ref[1]**o+(0.291794228985)*x_ref[1]
        ref[3,0,1,0]=(-0.360437990219)*x_ref[0]**o+(-0.570216212399)*x_ref[0]+(-0.607512137662)*x_ref[1]**o+(-0.515098731307)*x_ref[1]
        ref[3,0,1,1]=(0.642086516897)*x_ref[0]**o+(0.52547404535)*x_ref[0]+(0.016876402075)*x_ref[1]**o+(-0.862152788234)*x_ref[1]
        ref[3,0,2,0]=(-0.181658178552)*x_ref[0]**o+(-0.203447330476)*x_ref[0]+(-0.547194819622)*x_ref[1]**o+(-0.0593895310488)*x_ref[1]
        ref[3,0,2,1]=(-0.582071301036)*x_ref[0]**o+(-0.557616629544)*x_ref[0]+(-0.60723171731)*x_ref[1]**o+(0.204081805298)*x_ref[1]
        ref[3,1,0,0]=(0.272716289788)*x_ref[0]**o+(-0.947135640157)*x_ref[0]+(-0.25901560402)*x_ref[1]**o+(-0.0247786104764)*x_ref[1]
        ref[3,1,0,1]=(-0.891714371313)*x_ref[0]**o+(0.755281444278)*x_ref[0]+(0.351999986261)*x_ref[1]**o+(0.661376231103)*x_ref[1]
        ref[3,1,1,0]=(0.432312310444)*x_ref[0]**o+(-0.0375106988813)*x_ref[0]+(0.843778132955)*x_ref[1]**o+(-0.787134076357)*x_ref[1]
        ref[3,1,1,1]=(-0.57131496233)*x_ref[0]**o+(-0.614200739152)*x_ref[0]+(-0.0817103577524)*x_ref[1]**o+(-0.0809087761527)*x_ref[1]
        ref[3,1,2,0]=(0.775899774689)*x_ref[0]**o+(0.650087297096)*x_ref[0]+(-0.363483124762)*x_ref[1]**o+(0.0406785055243)*x_ref[1]
        ref[3,1,2,1]=(-0.8228824793)*x_ref[0]**o+(0.0223028194838)*x_ref[0]+(-0.246990198359)*x_ref[1]**o+(-0.447554242783)*x_ref[1]
        ref[3,2,0,0]=(0.250699655536)*x_ref[0]**o+(0.425847139574)*x_ref[0]+(0.185272955249)*x_ref[1]**o+(0.52788179093)*x_ref[1]
        ref[3,2,0,1]=(-0.6916019182)*x_ref[0]**o+(0.468697041897)*x_ref[0]+(0.0929416146358)*x_ref[1]**o+(0.0632517742213)*x_ref[1]
        ref[3,2,1,0]=(-0.0137715023834)*x_ref[0]**o+(-0.876273608156)*x_ref[0]+(0.333020658958)*x_ref[1]**o+(-0.546002404224)*x_ref[1]
        ref[3,2,1,1]=(-0.0260174553873)*x_ref[0]**o+(0.668170105508)*x_ref[0]+(-0.515918906011)*x_ref[1]**o+(0.427775146345)*x_ref[1]
        ref[3,2,2,0]=(0.451288901693)*x_ref[0]**o+(-0.307365533804)*x_ref[0]+(-0.902054304556)*x_ref[1]**o+(0.561501391872)*x_ref[1]
        ref[3,2,2,1]=(-0.67166870387)*x_ref[0]**o+(-0.851949794629)*x_ref[0]+(-0.198839514379)*x_ref[1]**o+(-0.0539885252629)*x_ref[1]
        ref[3,3,0,0]=(0.319371411392)*x_ref[0]**o+(0.521157698767)*x_ref[0]+(0.0619379731457)*x_ref[1]**o+(0.85473464537)*x_ref[1]
        ref[3,3,0,1]=(0.605988887111)*x_ref[0]**o+(-0.75556314419)*x_ref[0]+(-0.551855748356)*x_ref[1]**o+(0.152400883628)*x_ref[1]
        ref[3,3,1,0]=(-0.963972268356)*x_ref[0]**o+(-0.221723218793)*x_ref[0]+(-0.940231983732)*x_ref[1]**o+(0.509102233847)*x_ref[1]
        ref[3,3,1,1]=(0.934837503535)*x_ref[0]**o+(-0.762674616346)*x_ref[0]+(0.196730758209)*x_ref[1]**o+(0.93895486771)*x_ref[1]
        ref[3,3,2,0]=(0.683137246803)*x_ref[0]**o+(0.638846953858)*x_ref[0]+(0.508007741206)*x_ref[1]**o+(0.625453420423)*x_ref[1]
        ref[3,3,2,1]=(-0.0951254112087)*x_ref[0]**o+(0.507174561327)*x_ref[0]+(0.98495014583)*x_ref[1]**o+(0.922324997316)*x_ref[1]
        ref[3,4,0,0]=(-0.39889705668)*x_ref[0]**o+(-0.00933499671216)*x_ref[0]+(-0.308567417711)*x_ref[1]**o+(0.0365845447897)*x_ref[1]
        ref[3,4,0,1]=(0.699792659761)*x_ref[0]**o+(-0.339911160323)*x_ref[0]+(-0.0380630017731)*x_ref[1]**o+(0.891498223182)*x_ref[1]
        ref[3,4,1,0]=(-0.486198638614)*x_ref[0]**o+(0.910921882689)*x_ref[0]+(0.544226196572)*x_ref[1]**o+(0.221206282494)*x_ref[1]
        ref[3,4,1,1]=(0.0656938900284)*x_ref[0]**o+(-0.146193696097)*x_ref[0]+(0.580985806286)*x_ref[1]**o+(0.913369274555)*x_ref[1]
        ref[3,4,2,0]=(-0.31430372984)*x_ref[0]**o+(0.132101392811)*x_ref[0]+(0.823927100539)*x_ref[1]**o+(-0.0349106467441)*x_ref[1]
        ref[3,4,2,1]=(0.454900950832)*x_ref[0]**o+(0.875474839303)*x_ref[0]+(-0.496859632101)*x_ref[1]**o+(0.143908936177)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.893553650353)*x[0]**o+(-0.351842521516)*x[0]+(-0.776239256603)*x[1]**o+(-0.402276687439)*x[1]+(-0.822402864919)*x[2]**o+(0.103529960909)*x[2]
        s[0,0,0,1]=(-0.15411940588)*x[0]**o+(-0.719214069052)*x[0]+(-0.492951656336)*x[1]**o+(0.809209951326)*x[1]+(-0.593114280214)*x[2]**o+(0.647536354987)*x[2]
        s[0,0,1,0]=(-0.4726578657)*x[0]**o+(-0.900529164403)*x[0]+(-0.497658085089)*x[1]**o+(0.441307048821)*x[1]+(0.627699814902)*x[2]**o+(-0.255473726512)*x[2]
        s[0,0,1,1]=(0.937189233751)*x[0]**o+(-0.394633254149)*x[0]+(-0.83447573166)*x[1]**o+(-0.189038884898)*x[1]+(0.852530614411)*x[2]**o+(0.967144561098)*x[2]
        s[0,0,2,0]=(-0.530674910941)*x[0]**o+(-0.349001272718)*x[0]+(0.587518089749)*x[1]**o+(-0.892322978725)*x[1]+(-0.504626322472)*x[2]**o+(-0.867528772077)*x[2]
        s[0,0,2,1]=(-0.850967403091)*x[0]**o+(-0.830253463355)*x[0]+(-0.748829089276)*x[1]**o+(0.171444793758)*x[1]+(-0.793998192991)*x[2]**o+(-0.661765935989)*x[2]
        s[0,1,0,0]=(-0.785128579692)*x[0]**o+(-0.359210459481)*x[0]+(-0.509642957058)*x[1]**o+(0.318074588083)*x[1]+(-0.15398452232)*x[2]**o+(-0.423529526133)*x[2]
        s[0,1,0,1]=(0.198797074339)*x[0]**o+(0.754388829431)*x[0]+(0.884953371408)*x[1]**o+(0.661707862973)*x[1]+(0.663813806214)*x[2]**o+(-0.237609669098)*x[2]
        s[0,1,1,0]=(0.296403271945)*x[0]**o+(0.888911865921)*x[0]+(-0.114728439653)*x[1]**o+(-0.190296446591)*x[1]+(0.699471155167)*x[2]**o+(0.257430171529)*x[2]
        s[0,1,1,1]=(-0.937866711277)*x[0]**o+(0.199715798396)*x[0]+(0.682429346583)*x[1]**o+(0.0300499830473)*x[1]+(0.796837425072)*x[2]**o+(0.618110327057)*x[2]
        s[0,1,2,0]=(0.83267626043)*x[0]**o+(0.389527167021)*x[0]+(0.696622541573)*x[1]**o+(0.323025867562)*x[1]+(-0.335192370542)*x[2]**o+(0.598268793919)*x[2]
        s[0,1,2,1]=(-0.860895701689)*x[0]**o+(0.571893463085)*x[0]+(-0.837208302676)*x[1]**o+(0.582037641186)*x[1]+(0.0204844436813)*x[2]**o+(0.84175443928)*x[2]
        s[0,2,0,0]=(0.995641912188)*x[0]**o+(-0.683043378142)*x[0]+(0.856582569766)*x[1]**o+(0.609700650036)*x[1]+(-0.216498152749)*x[2]**o+(0.404344185816)*x[2]
        s[0,2,0,1]=(-0.0336188475062)*x[0]**o+(0.844294087444)*x[0]+(-0.229218254217)*x[1]**o+(0.115102437743)*x[1]+(0.571093445813)*x[2]**o+(0.266786250904)*x[2]
        s[0,2,1,0]=(-0.480605609044)*x[0]**o+(0.895506339528)*x[0]+(0.497786538492)*x[1]**o+(0.826429641685)*x[1]+(-0.0404930804723)*x[2]**o+(-0.549699397313)*x[2]
        s[0,2,1,1]=(-0.635472022806)*x[0]**o+(0.426611422786)*x[0]+(-0.589809517041)*x[1]**o+(-0.510141409468)*x[1]+(0.861957013957)*x[2]**o+(0.547568522693)*x[2]
        s[0,2,2,0]=(0.194664120443)*x[0]**o+(-0.824830613291)*x[0]+(0.885340662561)*x[1]**o+(0.174047647392)*x[1]+(0.0378551578969)*x[2]**o+(-0.485832036974)*x[2]
        s[0,2,2,1]=(-0.269452123964)*x[0]**o+(-0.0487386347542)*x[0]+(-0.496833515693)*x[1]**o+(0.501015798563)*x[1]+(-0.561879691824)*x[2]**o+(-0.231151995284)*x[2]
        s[0,3,0,0]=(-0.657879229166)*x[0]**o+(0.437034424448)*x[0]+(-0.279440364362)*x[1]**o+(0.306164743472)*x[1]+(-0.976575119833)*x[2]**o+(0.422887219126)*x[2]
        s[0,3,0,1]=(0.463706573378)*x[0]**o+(-0.0560780321685)*x[0]+(0.160348057871)*x[1]**o+(-0.394155833611)*x[1]+(0.354926830678)*x[2]**o+(-0.836986282947)*x[2]
        s[0,3,1,0]=(-0.741940744824)*x[0]**o+(0.185216337632)*x[0]+(0.345258688006)*x[1]**o+(-0.967694578778)*x[1]+(-0.824610904123)*x[2]**o+(-0.735528949759)*x[2]
        s[0,3,1,1]=(-0.0871457875019)*x[0]**o+(0.658300473204)*x[0]+(-0.131866896)*x[1]**o+(0.0651584246177)*x[1]+(0.480033790713)*x[2]**o+(0.240844483345)*x[2]
        s[0,3,2,0]=(-0.921746479348)*x[0]**o+(0.822395759575)*x[0]+(0.0527896299189)*x[1]**o+(0.0119091366531)*x[1]+(-0.659801021057)*x[2]**o+(0.813802745814)*x[2]
        s[0,3,2,1]=(0.668738391756)*x[0]**o+(0.836664681044)*x[0]+(-0.130670296341)*x[1]**o+(0.542822650083)*x[1]+(-0.556597236729)*x[2]**o+(-0.208288038549)*x[2]
        s[0,4,0,0]=(0.086889831216)*x[0]**o+(0.934721202553)*x[0]+(-0.323285065677)*x[1]**o+(0.0162227173407)*x[1]+(0.303014495731)*x[2]**o+(0.569121998148)*x[2]
        s[0,4,0,1]=(-0.498353567584)*x[0]**o+(-0.936339865441)*x[0]+(0.187200383841)*x[1]**o+(0.0572760624614)*x[1]+(-0.176971935188)*x[2]**o+(0.888193107076)*x[2]
        s[0,4,1,0]=(0.748849248567)*x[0]**o+(-0.597794365534)*x[0]+(-0.74698027669)*x[1]**o+(0.461712848089)*x[1]+(0.984587526203)*x[2]**o+(0.210336004682)*x[2]
        s[0,4,1,1]=(0.644641868836)*x[0]**o+(0.569657613081)*x[0]+(0.569132875942)*x[1]**o+(0.229575134126)*x[1]+(0.0821684810315)*x[2]**o+(0.23719873578)*x[2]
        s[0,4,2,0]=(-0.583799917268)*x[0]**o+(0.87257955676)*x[0]+(-0.262198279859)*x[1]**o+(0.638709803441)*x[1]+(0.693720376703)*x[2]**o+(0.0626506206935)*x[2]
        s[0,4,2,1]=(-0.555140469776)*x[0]**o+(-0.699667339844)*x[0]+(-0.175579595413)*x[1]**o+(0.330869902579)*x[1]+(-0.43443643788)*x[2]**o+(0.431247661025)*x[2]
        s[1,0,0,0]=(-0.616105274011)*x[0]**o+(0.766286637713)*x[0]+(-0.852345396201)*x[1]**o+(0.53484561268)*x[1]+(0.568276493883)*x[2]**o+(0.856054828485)*x[2]
        s[1,0,0,1]=(0.908849409393)*x[0]**o+(-0.745153969179)*x[0]+(-0.279959776499)*x[1]**o+(-0.854744604289)*x[1]+(-0.936424853683)*x[2]**o+(0.894518923387)*x[2]
        s[1,0,1,0]=(0.84860289881)*x[0]**o+(-0.914008514988)*x[0]+(-0.262407548228)*x[1]**o+(-0.711226151818)*x[1]+(0.639319422513)*x[2]**o+(-0.27853905681)*x[2]
        s[1,0,1,1]=(0.163323251559)*x[0]**o+(-0.93845875669)*x[0]+(-0.443114805807)*x[1]**o+(-0.533303207459)*x[1]+(0.431839047632)*x[2]**o+(-0.569656544847)*x[2]
        s[1,0,2,0]=(0.949714964792)*x[0]**o+(-0.976181981164)*x[0]+(-0.446526176314)*x[1]**o+(0.383526413445)*x[1]+(-0.182860699476)*x[2]**o+(0.959339975866)*x[2]
        s[1,0,2,1]=(0.867191335472)*x[0]**o+(0.751916463079)*x[0]+(-0.378585476419)*x[1]**o+(-0.500587362177)*x[1]+(-0.27137832483)*x[2]**o+(0.551490270001)*x[2]
        s[1,1,0,0]=(-0.583361795968)*x[0]**o+(0.944694170692)*x[0]+(-0.779315277681)*x[1]**o+(-0.113881659539)*x[1]+(0.931018446727)*x[2]**o+(0.480738722055)*x[2]
        s[1,1,0,1]=(0.229826570157)*x[0]**o+(0.714777306362)*x[0]+(0.442416481157)*x[1]**o+(0.605660985582)*x[1]+(-0.525954520894)*x[2]**o+(0.807486338335)*x[2]
        s[1,1,1,0]=(0.316589515903)*x[0]**o+(-0.686516469152)*x[0]+(0.840774801669)*x[1]**o+(0.447175951045)*x[1]+(-0.716670116885)*x[2]**o+(0.790809180069)*x[2]
        s[1,1,1,1]=(-0.508708038573)*x[0]**o+(-0.934322667237)*x[0]+(0.241647787833)*x[1]**o+(0.538683722428)*x[1]+(0.560593542667)*x[2]**o+(-0.943298870861)*x[2]
        s[1,1,2,0]=(0.21474264864)*x[0]**o+(0.914869403498)*x[0]+(-0.154022728013)*x[1]**o+(0.161351894986)*x[1]+(0.412310479319)*x[2]**o+(-0.664814506861)*x[2]
        s[1,1,2,1]=(-0.143696386498)*x[0]**o+(0.883320802442)*x[0]+(-0.907984528188)*x[1]**o+(-0.386745629926)*x[1]+(0.15492968318)*x[2]**o+(0.21067446716)*x[2]
        s[1,2,0,0]=(0.660203100234)*x[0]**o+(0.284397546215)*x[0]+(0.367929049046)*x[1]**o+(0.344269738617)*x[1]+(-0.511307639941)*x[2]**o+(-0.392647111269)*x[2]
        s[1,2,0,1]=(0.867950390849)*x[0]**o+(0.816012056188)*x[0]+(-0.0418899900381)*x[1]**o+(0.240447024929)*x[1]+(0.717832398567)*x[2]**o+(0.945994010423)*x[2]
        s[1,2,1,0]=(-0.242292281013)*x[0]**o+(0.986187733215)*x[0]+(-0.0762418474615)*x[1]**o+(0.545636818354)*x[1]+(-0.401003911045)*x[2]**o+(-0.0737632171919)*x[2]
        s[1,2,1,1]=(0.233293099685)*x[0]**o+(0.973431444625)*x[0]+(0.721124631798)*x[1]**o+(-0.172302126377)*x[1]+(0.116437603679)*x[2]**o+(0.735419551773)*x[2]
        s[1,2,2,0]=(0.971110632456)*x[0]**o+(0.185002876302)*x[0]+(-0.790484468882)*x[1]**o+(0.616520494387)*x[1]+(-0.818053831626)*x[2]**o+(0.705809191087)*x[2]
        s[1,2,2,1]=(-0.937151426491)*x[0]**o+(0.0344588436018)*x[0]+(0.178600320405)*x[1]**o+(0.774945208314)*x[1]+(0.00671749768773)*x[2]**o+(0.600596280965)*x[2]
        s[1,3,0,0]=(-0.510629068288)*x[0]**o+(-0.9867441798)*x[0]+(0.608512062668)*x[1]**o+(0.0882661466679)*x[1]+(0.517357685392)*x[2]**o+(0.921906359759)*x[2]
        s[1,3,0,1]=(0.857935816338)*x[0]**o+(-0.568613593339)*x[0]+(-0.311233160541)*x[1]**o+(0.456373110337)*x[1]+(0.294859133869)*x[2]**o+(0.972992715921)*x[2]
        s[1,3,1,0]=(-0.906870982033)*x[0]**o+(-0.123891310654)*x[0]+(-0.926024571289)*x[1]**o+(-0.971883326233)*x[1]+(-0.270701126081)*x[2]**o+(-0.314054166211)*x[2]
        s[1,3,1,1]=(0.932643998906)*x[0]**o+(0.591312871747)*x[0]+(-0.297447853121)*x[1]**o+(-0.326446485245)*x[1]+(0.355711757012)*x[2]**o+(0.941580099148)*x[2]
        s[1,3,2,0]=(-0.791232221354)*x[0]**o+(-0.545934762256)*x[0]+(-0.527723158311)*x[1]**o+(0.964298416277)*x[1]+(-0.821314347444)*x[2]**o+(-0.889147503887)*x[2]
        s[1,3,2,1]=(0.468387240601)*x[0]**o+(0.831139810807)*x[0]+(-0.780171451212)*x[1]**o+(-0.154078607613)*x[1]+(0.473061938421)*x[2]**o+(-0.237542463566)*x[2]
        s[1,4,0,0]=(0.23783592874)*x[0]**o+(0.922830638473)*x[0]+(-0.53102227341)*x[1]**o+(0.196635165848)*x[1]+(-0.983681936618)*x[2]**o+(-0.946261418219)*x[2]
        s[1,4,0,1]=(0.0626656945751)*x[0]**o+(0.0915443649303)*x[0]+(-0.306766147008)*x[1]**o+(-0.220472003734)*x[1]+(-0.819326989073)*x[2]**o+(0.584600281341)*x[2]
        s[1,4,1,0]=(-0.144427843844)*x[0]**o+(0.854025358607)*x[0]+(0.694868129931)*x[1]**o+(-0.641915817761)*x[1]+(-0.580940477231)*x[2]**o+(-0.252220617111)*x[2]
        s[1,4,1,1]=(0.632987045152)*x[0]**o+(-0.222995494274)*x[0]+(0.218892976575)*x[1]**o+(-0.870083564081)*x[1]+(-0.315656003785)*x[2]**o+(-0.806577859045)*x[2]
        s[1,4,2,0]=(-0.686677179384)*x[0]**o+(0.975041194178)*x[0]+(0.840117893335)*x[1]**o+(-0.357121725686)*x[1]+(-0.113947327576)*x[2]**o+(0.722741254436)*x[2]
        s[1,4,2,1]=(-0.238773557516)*x[0]**o+(0.589422174847)*x[0]+(-0.39358233128)*x[1]**o+(0.159348401427)*x[1]+(0.887216043103)*x[2]**o+(-0.924304870822)*x[2]
        s[2,0,0,0]=(-0.725901161347)*x[0]**o+(0.144957603906)*x[0]+(0.231954835231)*x[1]**o+(-0.760212283123)*x[1]+(-0.725424437779)*x[2]**o+(-0.19688509792)*x[2]
        s[2,0,0,1]=(0.418995010234)*x[0]**o+(-0.942078773861)*x[0]+(-0.818049256222)*x[1]**o+(-0.594681673779)*x[1]+(-0.63178803463)*x[2]**o+(-0.989664715484)*x[2]
        s[2,0,1,0]=(-0.524535447707)*x[0]**o+(-0.746654195315)*x[0]+(-0.721012736821)*x[1]**o+(-0.0345481441746)*x[1]+(0.0537586563756)*x[2]**o+(-0.445033317006)*x[2]
        s[2,0,1,1]=(-0.411001804179)*x[0]**o+(0.00970860318518)*x[0]+(0.783986695315)*x[1]**o+(-0.270151475512)*x[1]+(-0.652403140687)*x[2]**o+(-0.261945554182)*x[2]
        s[2,0,2,0]=(-0.497529491593)*x[0]**o+(0.00339011573669)*x[0]+(0.696815843509)*x[1]**o+(-0.277295752437)*x[1]+(0.204803051766)*x[2]**o+(-0.631981280563)*x[2]
        s[2,0,2,1]=(0.321760872287)*x[0]**o+(0.54137067299)*x[0]+(-0.971971260117)*x[1]**o+(0.767307092712)*x[1]+(0.471809380038)*x[2]**o+(0.178504669856)*x[2]
        s[2,1,0,0]=(-0.627353418617)*x[0]**o+(-0.610846214602)*x[0]+(-0.610318029871)*x[1]**o+(0.884848156201)*x[1]+(-0.227603047617)*x[2]**o+(0.728596663375)*x[2]
        s[2,1,0,1]=(0.205703168727)*x[0]**o+(-0.869108006912)*x[0]+(0.195195654062)*x[1]**o+(0.586300414612)*x[1]+(0.787224810167)*x[2]**o+(-0.0278209500568)*x[2]
        s[2,1,1,0]=(-0.978822062351)*x[0]**o+(0.929683906062)*x[0]+(-0.865432742145)*x[1]**o+(-0.083101823301)*x[1]+(0.264956210564)*x[2]**o+(0.452513746821)*x[2]
        s[2,1,1,1]=(-0.607905485886)*x[0]**o+(-0.683168972026)*x[0]+(-0.973085006343)*x[1]**o+(0.905969981853)*x[1]+(0.336105649928)*x[2]**o+(0.976694664618)*x[2]
        s[2,1,2,0]=(-0.542965644364)*x[0]**o+(-0.740434309738)*x[0]+(0.756162713574)*x[1]**o+(-0.543216421535)*x[1]+(0.0942001276823)*x[2]**o+(0.0482815123591)*x[2]
        s[2,1,2,1]=(-0.226868852637)*x[0]**o+(0.863234399903)*x[0]+(-0.318469617308)*x[1]**o+(0.0192280454449)*x[1]+(-0.23611990172)*x[2]**o+(0.232713876707)*x[2]
        s[2,2,0,0]=(0.726586748484)*x[0]**o+(-0.711554904751)*x[0]+(-0.206168377552)*x[1]**o+(0.84673411536)*x[1]+(-0.594942018777)*x[2]**o+(-0.199098137058)*x[2]
        s[2,2,0,1]=(0.797788406588)*x[0]**o+(0.326417288103)*x[0]+(0.954847835729)*x[1]**o+(-0.823541118309)*x[1]+(-0.183282866634)*x[2]**o+(-0.73879700622)*x[2]
        s[2,2,1,0]=(-0.542135645496)*x[0]**o+(0.617419480945)*x[0]+(-0.858155129941)*x[1]**o+(0.622909241986)*x[1]+(-0.817913616615)*x[2]**o+(0.551023741548)*x[2]
        s[2,2,1,1]=(-0.033791510266)*x[0]**o+(0.960755607224)*x[0]+(0.992551377675)*x[1]**o+(-0.40568563699)*x[1]+(0.635141614881)*x[2]**o+(-0.28322398176)*x[2]
        s[2,2,2,0]=(0.200301024931)*x[0]**o+(-0.526020099906)*x[0]+(0.296360677546)*x[1]**o+(-0.346356255978)*x[1]+(-0.751246948633)*x[2]**o+(0.936356516645)*x[2]
        s[2,2,2,1]=(0.851662248368)*x[0]**o+(0.306264287809)*x[0]+(0.413913851675)*x[1]**o+(0.796748819208)*x[1]+(0.686640823111)*x[2]**o+(0.270797124789)*x[2]
        s[2,3,0,0]=(0.873612484608)*x[0]**o+(-0.635443314156)*x[0]+(0.0336582752521)*x[1]**o+(0.397952617963)*x[1]+(0.708037547846)*x[2]**o+(-0.835497214327)*x[2]
        s[2,3,0,1]=(0.266263219836)*x[0]**o+(0.718604372519)*x[0]+(-0.706375167711)*x[1]**o+(0.267962253067)*x[1]+(-0.0369901462112)*x[2]**o+(0.541887114998)*x[2]
        s[2,3,1,0]=(-0.0883529797185)*x[0]**o+(-0.695144807666)*x[0]+(-0.332082923487)*x[1]**o+(0.0386905137005)*x[1]+(-0.0402663506352)*x[2]**o+(0.897417224025)*x[2]
        s[2,3,1,1]=(-0.275313948401)*x[0]**o+(0.66334859913)*x[0]+(-0.139887516171)*x[1]**o+(0.874104988552)*x[1]+(-0.474834712531)*x[2]**o+(-0.350478879758)*x[2]
        s[2,3,2,0]=(-0.370570829115)*x[0]**o+(0.448685344991)*x[0]+(-0.663934134305)*x[1]**o+(0.270424636731)*x[1]+(0.257023713378)*x[2]**o+(-0.558834446028)*x[2]
        s[2,3,2,1]=(0.853600423341)*x[0]**o+(-0.34026373254)*x[0]+(-0.880960615323)*x[1]**o+(0.720247030686)*x[1]+(-0.179396842577)*x[2]**o+(-0.581055218136)*x[2]
        s[2,4,0,0]=(0.661293829706)*x[0]**o+(-0.278404566365)*x[0]+(-0.668007328124)*x[1]**o+(-0.0762718064116)*x[1]+(-0.559881233142)*x[2]**o+(0.38863355198)*x[2]
        s[2,4,0,1]=(0.841925292626)*x[0]**o+(-0.907683026635)*x[0]+(0.606133080471)*x[1]**o+(-0.743094760572)*x[1]+(-0.963492675546)*x[2]**o+(0.551853881446)*x[2]
        s[2,4,1,0]=(0.106157048406)*x[0]**o+(-0.839019337739)*x[0]+(-0.646225647224)*x[1]**o+(0.596433542879)*x[1]+(-0.440653099815)*x[2]**o+(0.612390706411)*x[2]
        s[2,4,1,1]=(0.0894808258178)*x[0]**o+(0.601107281081)*x[0]+(0.764980611835)*x[1]**o+(-0.028637689046)*x[1]+(-0.163992620355)*x[2]**o+(0.811740898974)*x[2]
        s[2,4,2,0]=(0.159301482106)*x[0]**o+(-0.2732663929)*x[0]+(0.642848902873)*x[1]**o+(-0.486181762406)*x[1]+(-0.0353954685288)*x[2]**o+(-0.17658875721)*x[2]
        s[2,4,2,1]=(-0.190069716212)*x[0]**o+(0.911197492671)*x[0]+(-0.404646749357)*x[1]**o+(0.496457219159)*x[1]+(0.331503854252)*x[2]**o+(-0.0824993997725)*x[2]
        s[3,0,0,0]=(-0.00960152574795)*x[0]**o+(-0.588053191696)*x[0]+(0.134548796202)*x[1]**o+(-0.734698264954)*x[1]+(0.369880572143)*x[2]**o+(-0.184585884928)*x[2]
        s[3,0,0,1]=(-0.894375930982)*x[0]**o+(-0.757908178588)*x[0]+(-0.862478750821)*x[1]**o+(-0.860344696497)*x[1]+(-0.606693340193)*x[2]**o+(-0.211110504173)*x[2]
        s[3,0,1,0]=(0.29208602662)*x[0]**o+(0.428795740188)*x[0]+(0.730773643572)*x[1]**o+(0.473531891026)*x[1]+(-0.475140560425)*x[2]**o+(0.973125507458)*x[2]
        s[3,0,1,1]=(-0.15741631987)*x[0]**o+(-0.34595691825)*x[0]+(-0.291584586864)*x[1]**o+(0.202892853712)*x[1]+(-0.48541451272)*x[2]**o+(-0.0611059721993)*x[2]
        s[3,0,2,0]=(-0.309937469651)*x[0]**o+(-0.515728964606)*x[0]+(-0.212592556423)*x[1]**o+(-0.419645818415)*x[1]+(-0.825715279525)*x[2]**o+(0.701775440231)*x[2]
        s[3,0,2,1]=(-0.527254609361)*x[0]**o+(-0.921224942687)*x[0]+(-0.219965563549)*x[1]**o+(-0.520137812628)*x[1]+(0.228621345698)*x[2]**o+(0.607942457443)*x[2]
        s[3,1,0,0]=(-0.0191513045468)*x[0]**o+(0.0593277991822)*x[0]+(0.679065928035)*x[1]**o+(-0.824466221695)*x[1]+(-0.0207465360942)*x[2]**o+(-0.462448546311)*x[2]
        s[3,1,0,1]=(-0.0226385405157)*x[0]**o+(-0.990729100489)*x[0]+(0.928646343096)*x[1]**o+(0.877879386627)*x[1]+(-0.774330622719)*x[2]**o+(0.45794763385)*x[2]
        s[3,1,1,0]=(0.991939307411)*x[0]**o+(0.424144367324)*x[0]+(0.287966132208)*x[1]**o+(0.560134521129)*x[1]+(-0.366498917899)*x[2]**o+(-0.5499574933)*x[2]
        s[3,1,1,1]=(0.131575237703)*x[0]**o+(0.350065088754)*x[0]+(0.845775797719)*x[1]**o+(-0.951498449401)*x[1]+(-0.749988925873)*x[2]**o+(0.437681907383)*x[2]
        s[3,1,2,0]=(-0.177690055962)*x[0]**o+(0.400513514288)*x[0]+(0.654580691817)*x[1]**o+(0.781954834622)*x[1]+(0.498508301282)*x[2]**o+(0.489109718057)*x[2]
        s[3,1,2,1]=(0.141266612215)*x[0]**o+(0.578894900597)*x[0]+(-0.0990011451546)*x[1]**o+(0.455191961)*x[1]+(0.446639012332)*x[2]**o+(0.398374036483)*x[2]
        s[3,2,0,0]=(-0.88698978852)*x[0]**o+(0.782997043916)*x[0]+(-0.0203503423342)*x[1]**o+(-0.729317082717)*x[1]+(-0.916473605189)*x[2]**o+(0.706748074855)*x[2]
        s[3,2,0,1]=(0.955314840013)*x[0]**o+(0.784566529939)*x[0]+(-0.607468453224)*x[1]**o+(-0.960125382144)*x[1]+(-0.62579570096)*x[2]**o+(-0.0943302040401)*x[2]
        s[3,2,1,0]=(-0.255287575915)*x[0]**o+(0.418535282029)*x[0]+(0.606559849688)*x[1]**o+(0.940877031124)*x[1]+(0.714131644166)*x[2]**o+(-0.77112172198)*x[2]
        s[3,2,1,1]=(0.566383366127)*x[0]**o+(0.301059560952)*x[0]+(0.181921162673)*x[1]**o+(-0.972651860466)*x[1]+(-0.435658160785)*x[2]**o+(-0.449044377115)*x[2]
        s[3,2,2,0]=(0.730443884106)*x[0]**o+(0.535861749024)*x[0]+(-0.257910068997)*x[1]**o+(-0.879923650189)*x[1]+(0.24309905047)*x[2]**o+(-0.407314031684)*x[2]
        s[3,2,2,1]=(-0.734460242423)*x[0]**o+(0.694756195588)*x[0]+(-0.0110855577415)*x[1]**o+(0.213770377606)*x[1]+(0.404975692507)*x[2]**o+(0.235122345113)*x[2]
        s[3,3,0,0]=(-0.0854800701504)*x[0]**o+(-0.998984290225)*x[0]+(-0.781970276212)*x[1]**o+(0.866788214566)*x[1]+(-0.21747695978)*x[2]**o+(-0.615256484308)*x[2]
        s[3,3,0,1]=(-0.509246205985)*x[0]**o+(-0.870342470261)*x[0]+(0.699615910083)*x[1]**o+(0.81056073388)*x[1]+(0.137830446242)*x[2]**o+(-0.355111214574)*x[2]
        s[3,3,1,0]=(-0.423472624091)*x[0]**o+(0.812256460885)*x[0]+(-0.675497703361)*x[1]**o+(0.296366723996)*x[1]+(-0.218492550353)*x[2]**o+(-0.852823171625)*x[2]
        s[3,3,1,1]=(-0.0718862672517)*x[0]**o+(-0.122792076285)*x[0]+(-0.794191931729)*x[1]**o+(0.569212982974)*x[1]+(-0.646851040773)*x[2]**o+(-0.3073262807)*x[2]
        s[3,3,2,0]=(-0.994778576692)*x[0]**o+(-0.250054159815)*x[0]+(0.726266056262)*x[1]**o+(-0.827929598129)*x[1]+(0.038573450577)*x[2]**o+(-0.402624412259)*x[2]
        s[3,3,2,1]=(-0.158546722118)*x[0]**o+(0.438480217845)*x[0]+(-0.757053187223)*x[1]**o+(0.96669102229)*x[1]+(0.585680018048)*x[2]**o+(0.917795585404)*x[2]
        s[3,4,0,0]=(0.964942257873)*x[0]**o+(0.222693858545)*x[0]+(0.708449965849)*x[1]**o+(-0.786172527336)*x[1]+(0.813715873004)*x[2]**o+(-0.0971222939825)*x[2]
        s[3,4,0,1]=(-0.14236975971)*x[0]**o+(-0.748307677379)*x[0]+(-0.269260913515)*x[1]**o+(-0.315981317137)*x[1]+(-0.375598072402)*x[2]**o+(0.0797466318359)*x[2]
        s[3,4,1,0]=(0.52711029533)*x[0]**o+(0.492047263035)*x[0]+(0.678479511876)*x[1]**o+(0.619219421066)*x[1]+(-0.528779830025)*x[2]**o+(-0.237989972194)*x[2]
        s[3,4,1,1]=(0.711881464696)*x[0]**o+(-0.398127815977)*x[0]+(-0.943946294128)*x[1]**o+(0.584800438848)*x[1]+(-0.683342581136)*x[2]**o+(0.895377849601)*x[2]
        s[3,4,2,0]=(-0.698456333887)*x[0]**o+(-0.36871456039)*x[0]+(0.216620614245)*x[1]**o+(-0.242036817594)*x[1]+(-0.845502631302)*x[2]**o+(-0.705164123503)*x[2]
        s[3,4,2,1]=(-0.000297852657095)*x[0]**o+(-0.847929455729)*x[0]+(0.496739572695)*x[1]**o+(0.574135132795)*x[1]+(0.585537276618)*x[2]**o+(-0.934539532574)*x[2]
        ref[0,0,0,0]=(-0.893553650353)*x_ref[0]**o+(-0.351842521516)*x_ref[0]+(-0.776239256603)*x_ref[1]**o+(-0.402276687439)*x_ref[1]+(-0.822402864919)*x_ref[2]**o+(0.103529960909)*x_ref[2]
        ref[0,0,0,1]=(-0.15411940588)*x_ref[0]**o+(-0.719214069052)*x_ref[0]+(-0.492951656336)*x_ref[1]**o+(0.809209951326)*x_ref[1]+(-0.593114280214)*x_ref[2]**o+(0.647536354987)*x_ref[2]
        ref[0,0,1,0]=(-0.4726578657)*x_ref[0]**o+(-0.900529164403)*x_ref[0]+(-0.497658085089)*x_ref[1]**o+(0.441307048821)*x_ref[1]+(0.627699814902)*x_ref[2]**o+(-0.255473726512)*x_ref[2]
        ref[0,0,1,1]=(0.937189233751)*x_ref[0]**o+(-0.394633254149)*x_ref[0]+(-0.83447573166)*x_ref[1]**o+(-0.189038884898)*x_ref[1]+(0.852530614411)*x_ref[2]**o+(0.967144561098)*x_ref[2]
        ref[0,0,2,0]=(-0.530674910941)*x_ref[0]**o+(-0.349001272718)*x_ref[0]+(0.587518089749)*x_ref[1]**o+(-0.892322978725)*x_ref[1]+(-0.504626322472)*x_ref[2]**o+(-0.867528772077)*x_ref[2]
        ref[0,0,2,1]=(-0.850967403091)*x_ref[0]**o+(-0.830253463355)*x_ref[0]+(-0.748829089276)*x_ref[1]**o+(0.171444793758)*x_ref[1]+(-0.793998192991)*x_ref[2]**o+(-0.661765935989)*x_ref[2]
        ref[0,1,0,0]=(-0.785128579692)*x_ref[0]**o+(-0.359210459481)*x_ref[0]+(-0.509642957058)*x_ref[1]**o+(0.318074588083)*x_ref[1]+(-0.15398452232)*x_ref[2]**o+(-0.423529526133)*x_ref[2]
        ref[0,1,0,1]=(0.198797074339)*x_ref[0]**o+(0.754388829431)*x_ref[0]+(0.884953371408)*x_ref[1]**o+(0.661707862973)*x_ref[1]+(0.663813806214)*x_ref[2]**o+(-0.237609669098)*x_ref[2]
        ref[0,1,1,0]=(0.296403271945)*x_ref[0]**o+(0.888911865921)*x_ref[0]+(-0.114728439653)*x_ref[1]**o+(-0.190296446591)*x_ref[1]+(0.699471155167)*x_ref[2]**o+(0.257430171529)*x_ref[2]
        ref[0,1,1,1]=(-0.937866711277)*x_ref[0]**o+(0.199715798396)*x_ref[0]+(0.682429346583)*x_ref[1]**o+(0.0300499830473)*x_ref[1]+(0.796837425072)*x_ref[2]**o+(0.618110327057)*x_ref[2]
        ref[0,1,2,0]=(0.83267626043)*x_ref[0]**o+(0.389527167021)*x_ref[0]+(0.696622541573)*x_ref[1]**o+(0.323025867562)*x_ref[1]+(-0.335192370542)*x_ref[2]**o+(0.598268793919)*x_ref[2]
        ref[0,1,2,1]=(-0.860895701689)*x_ref[0]**o+(0.571893463085)*x_ref[0]+(-0.837208302676)*x_ref[1]**o+(0.582037641186)*x_ref[1]+(0.0204844436813)*x_ref[2]**o+(0.84175443928)*x_ref[2]
        ref[0,2,0,0]=(0.995641912188)*x_ref[0]**o+(-0.683043378142)*x_ref[0]+(0.856582569766)*x_ref[1]**o+(0.609700650036)*x_ref[1]+(-0.216498152749)*x_ref[2]**o+(0.404344185816)*x_ref[2]
        ref[0,2,0,1]=(-0.0336188475062)*x_ref[0]**o+(0.844294087444)*x_ref[0]+(-0.229218254217)*x_ref[1]**o+(0.115102437743)*x_ref[1]+(0.571093445813)*x_ref[2]**o+(0.266786250904)*x_ref[2]
        ref[0,2,1,0]=(-0.480605609044)*x_ref[0]**o+(0.895506339528)*x_ref[0]+(0.497786538492)*x_ref[1]**o+(0.826429641685)*x_ref[1]+(-0.0404930804723)*x_ref[2]**o+(-0.549699397313)*x_ref[2]
        ref[0,2,1,1]=(-0.635472022806)*x_ref[0]**o+(0.426611422786)*x_ref[0]+(-0.589809517041)*x_ref[1]**o+(-0.510141409468)*x_ref[1]+(0.861957013957)*x_ref[2]**o+(0.547568522693)*x_ref[2]
        ref[0,2,2,0]=(0.194664120443)*x_ref[0]**o+(-0.824830613291)*x_ref[0]+(0.885340662561)*x_ref[1]**o+(0.174047647392)*x_ref[1]+(0.0378551578969)*x_ref[2]**o+(-0.485832036974)*x_ref[2]
        ref[0,2,2,1]=(-0.269452123964)*x_ref[0]**o+(-0.0487386347542)*x_ref[0]+(-0.496833515693)*x_ref[1]**o+(0.501015798563)*x_ref[1]+(-0.561879691824)*x_ref[2]**o+(-0.231151995284)*x_ref[2]
        ref[0,3,0,0]=(-0.657879229166)*x_ref[0]**o+(0.437034424448)*x_ref[0]+(-0.279440364362)*x_ref[1]**o+(0.306164743472)*x_ref[1]+(-0.976575119833)*x_ref[2]**o+(0.422887219126)*x_ref[2]
        ref[0,3,0,1]=(0.463706573378)*x_ref[0]**o+(-0.0560780321685)*x_ref[0]+(0.160348057871)*x_ref[1]**o+(-0.394155833611)*x_ref[1]+(0.354926830678)*x_ref[2]**o+(-0.836986282947)*x_ref[2]
        ref[0,3,1,0]=(-0.741940744824)*x_ref[0]**o+(0.185216337632)*x_ref[0]+(0.345258688006)*x_ref[1]**o+(-0.967694578778)*x_ref[1]+(-0.824610904123)*x_ref[2]**o+(-0.735528949759)*x_ref[2]
        ref[0,3,1,1]=(-0.0871457875019)*x_ref[0]**o+(0.658300473204)*x_ref[0]+(-0.131866896)*x_ref[1]**o+(0.0651584246177)*x_ref[1]+(0.480033790713)*x_ref[2]**o+(0.240844483345)*x_ref[2]
        ref[0,3,2,0]=(-0.921746479348)*x_ref[0]**o+(0.822395759575)*x_ref[0]+(0.0527896299189)*x_ref[1]**o+(0.0119091366531)*x_ref[1]+(-0.659801021057)*x_ref[2]**o+(0.813802745814)*x_ref[2]
        ref[0,3,2,1]=(0.668738391756)*x_ref[0]**o+(0.836664681044)*x_ref[0]+(-0.130670296341)*x_ref[1]**o+(0.542822650083)*x_ref[1]+(-0.556597236729)*x_ref[2]**o+(-0.208288038549)*x_ref[2]
        ref[0,4,0,0]=(0.086889831216)*x_ref[0]**o+(0.934721202553)*x_ref[0]+(-0.323285065677)*x_ref[1]**o+(0.0162227173407)*x_ref[1]+(0.303014495731)*x_ref[2]**o+(0.569121998148)*x_ref[2]
        ref[0,4,0,1]=(-0.498353567584)*x_ref[0]**o+(-0.936339865441)*x_ref[0]+(0.187200383841)*x_ref[1]**o+(0.0572760624614)*x_ref[1]+(-0.176971935188)*x_ref[2]**o+(0.888193107076)*x_ref[2]
        ref[0,4,1,0]=(0.748849248567)*x_ref[0]**o+(-0.597794365534)*x_ref[0]+(-0.74698027669)*x_ref[1]**o+(0.461712848089)*x_ref[1]+(0.984587526203)*x_ref[2]**o+(0.210336004682)*x_ref[2]
        ref[0,4,1,1]=(0.644641868836)*x_ref[0]**o+(0.569657613081)*x_ref[0]+(0.569132875942)*x_ref[1]**o+(0.229575134126)*x_ref[1]+(0.0821684810315)*x_ref[2]**o+(0.23719873578)*x_ref[2]
        ref[0,4,2,0]=(-0.583799917268)*x_ref[0]**o+(0.87257955676)*x_ref[0]+(-0.262198279859)*x_ref[1]**o+(0.638709803441)*x_ref[1]+(0.693720376703)*x_ref[2]**o+(0.0626506206935)*x_ref[2]
        ref[0,4,2,1]=(-0.555140469776)*x_ref[0]**o+(-0.699667339844)*x_ref[0]+(-0.175579595413)*x_ref[1]**o+(0.330869902579)*x_ref[1]+(-0.43443643788)*x_ref[2]**o+(0.431247661025)*x_ref[2]
        ref[1,0,0,0]=(-0.616105274011)*x_ref[0]**o+(0.766286637713)*x_ref[0]+(-0.852345396201)*x_ref[1]**o+(0.53484561268)*x_ref[1]+(0.568276493883)*x_ref[2]**o+(0.856054828485)*x_ref[2]
        ref[1,0,0,1]=(0.908849409393)*x_ref[0]**o+(-0.745153969179)*x_ref[0]+(-0.279959776499)*x_ref[1]**o+(-0.854744604289)*x_ref[1]+(-0.936424853683)*x_ref[2]**o+(0.894518923387)*x_ref[2]
        ref[1,0,1,0]=(0.84860289881)*x_ref[0]**o+(-0.914008514988)*x_ref[0]+(-0.262407548228)*x_ref[1]**o+(-0.711226151818)*x_ref[1]+(0.639319422513)*x_ref[2]**o+(-0.27853905681)*x_ref[2]
        ref[1,0,1,1]=(0.163323251559)*x_ref[0]**o+(-0.93845875669)*x_ref[0]+(-0.443114805807)*x_ref[1]**o+(-0.533303207459)*x_ref[1]+(0.431839047632)*x_ref[2]**o+(-0.569656544847)*x_ref[2]
        ref[1,0,2,0]=(0.949714964792)*x_ref[0]**o+(-0.976181981164)*x_ref[0]+(-0.446526176314)*x_ref[1]**o+(0.383526413445)*x_ref[1]+(-0.182860699476)*x_ref[2]**o+(0.959339975866)*x_ref[2]
        ref[1,0,2,1]=(0.867191335472)*x_ref[0]**o+(0.751916463079)*x_ref[0]+(-0.378585476419)*x_ref[1]**o+(-0.500587362177)*x_ref[1]+(-0.27137832483)*x_ref[2]**o+(0.551490270001)*x_ref[2]
        ref[1,1,0,0]=(-0.583361795968)*x_ref[0]**o+(0.944694170692)*x_ref[0]+(-0.779315277681)*x_ref[1]**o+(-0.113881659539)*x_ref[1]+(0.931018446727)*x_ref[2]**o+(0.480738722055)*x_ref[2]
        ref[1,1,0,1]=(0.229826570157)*x_ref[0]**o+(0.714777306362)*x_ref[0]+(0.442416481157)*x_ref[1]**o+(0.605660985582)*x_ref[1]+(-0.525954520894)*x_ref[2]**o+(0.807486338335)*x_ref[2]
        ref[1,1,1,0]=(0.316589515903)*x_ref[0]**o+(-0.686516469152)*x_ref[0]+(0.840774801669)*x_ref[1]**o+(0.447175951045)*x_ref[1]+(-0.716670116885)*x_ref[2]**o+(0.790809180069)*x_ref[2]
        ref[1,1,1,1]=(-0.508708038573)*x_ref[0]**o+(-0.934322667237)*x_ref[0]+(0.241647787833)*x_ref[1]**o+(0.538683722428)*x_ref[1]+(0.560593542667)*x_ref[2]**o+(-0.943298870861)*x_ref[2]
        ref[1,1,2,0]=(0.21474264864)*x_ref[0]**o+(0.914869403498)*x_ref[0]+(-0.154022728013)*x_ref[1]**o+(0.161351894986)*x_ref[1]+(0.412310479319)*x_ref[2]**o+(-0.664814506861)*x_ref[2]
        ref[1,1,2,1]=(-0.143696386498)*x_ref[0]**o+(0.883320802442)*x_ref[0]+(-0.907984528188)*x_ref[1]**o+(-0.386745629926)*x_ref[1]+(0.15492968318)*x_ref[2]**o+(0.21067446716)*x_ref[2]
        ref[1,2,0,0]=(0.660203100234)*x_ref[0]**o+(0.284397546215)*x_ref[0]+(0.367929049046)*x_ref[1]**o+(0.344269738617)*x_ref[1]+(-0.511307639941)*x_ref[2]**o+(-0.392647111269)*x_ref[2]
        ref[1,2,0,1]=(0.867950390849)*x_ref[0]**o+(0.816012056188)*x_ref[0]+(-0.0418899900381)*x_ref[1]**o+(0.240447024929)*x_ref[1]+(0.717832398567)*x_ref[2]**o+(0.945994010423)*x_ref[2]
        ref[1,2,1,0]=(-0.242292281013)*x_ref[0]**o+(0.986187733215)*x_ref[0]+(-0.0762418474615)*x_ref[1]**o+(0.545636818354)*x_ref[1]+(-0.401003911045)*x_ref[2]**o+(-0.0737632171919)*x_ref[2]
        ref[1,2,1,1]=(0.233293099685)*x_ref[0]**o+(0.973431444625)*x_ref[0]+(0.721124631798)*x_ref[1]**o+(-0.172302126377)*x_ref[1]+(0.116437603679)*x_ref[2]**o+(0.735419551773)*x_ref[2]
        ref[1,2,2,0]=(0.971110632456)*x_ref[0]**o+(0.185002876302)*x_ref[0]+(-0.790484468882)*x_ref[1]**o+(0.616520494387)*x_ref[1]+(-0.818053831626)*x_ref[2]**o+(0.705809191087)*x_ref[2]
        ref[1,2,2,1]=(-0.937151426491)*x_ref[0]**o+(0.0344588436018)*x_ref[0]+(0.178600320405)*x_ref[1]**o+(0.774945208314)*x_ref[1]+(0.00671749768773)*x_ref[2]**o+(0.600596280965)*x_ref[2]
        ref[1,3,0,0]=(-0.510629068288)*x_ref[0]**o+(-0.9867441798)*x_ref[0]+(0.608512062668)*x_ref[1]**o+(0.0882661466679)*x_ref[1]+(0.517357685392)*x_ref[2]**o+(0.921906359759)*x_ref[2]
        ref[1,3,0,1]=(0.857935816338)*x_ref[0]**o+(-0.568613593339)*x_ref[0]+(-0.311233160541)*x_ref[1]**o+(0.456373110337)*x_ref[1]+(0.294859133869)*x_ref[2]**o+(0.972992715921)*x_ref[2]
        ref[1,3,1,0]=(-0.906870982033)*x_ref[0]**o+(-0.123891310654)*x_ref[0]+(-0.926024571289)*x_ref[1]**o+(-0.971883326233)*x_ref[1]+(-0.270701126081)*x_ref[2]**o+(-0.314054166211)*x_ref[2]
        ref[1,3,1,1]=(0.932643998906)*x_ref[0]**o+(0.591312871747)*x_ref[0]+(-0.297447853121)*x_ref[1]**o+(-0.326446485245)*x_ref[1]+(0.355711757012)*x_ref[2]**o+(0.941580099148)*x_ref[2]
        ref[1,3,2,0]=(-0.791232221354)*x_ref[0]**o+(-0.545934762256)*x_ref[0]+(-0.527723158311)*x_ref[1]**o+(0.964298416277)*x_ref[1]+(-0.821314347444)*x_ref[2]**o+(-0.889147503887)*x_ref[2]
        ref[1,3,2,1]=(0.468387240601)*x_ref[0]**o+(0.831139810807)*x_ref[0]+(-0.780171451212)*x_ref[1]**o+(-0.154078607613)*x_ref[1]+(0.473061938421)*x_ref[2]**o+(-0.237542463566)*x_ref[2]
        ref[1,4,0,0]=(0.23783592874)*x_ref[0]**o+(0.922830638473)*x_ref[0]+(-0.53102227341)*x_ref[1]**o+(0.196635165848)*x_ref[1]+(-0.983681936618)*x_ref[2]**o+(-0.946261418219)*x_ref[2]
        ref[1,4,0,1]=(0.0626656945751)*x_ref[0]**o+(0.0915443649303)*x_ref[0]+(-0.306766147008)*x_ref[1]**o+(-0.220472003734)*x_ref[1]+(-0.819326989073)*x_ref[2]**o+(0.584600281341)*x_ref[2]
        ref[1,4,1,0]=(-0.144427843844)*x_ref[0]**o+(0.854025358607)*x_ref[0]+(0.694868129931)*x_ref[1]**o+(-0.641915817761)*x_ref[1]+(-0.580940477231)*x_ref[2]**o+(-0.252220617111)*x_ref[2]
        ref[1,4,1,1]=(0.632987045152)*x_ref[0]**o+(-0.222995494274)*x_ref[0]+(0.218892976575)*x_ref[1]**o+(-0.870083564081)*x_ref[1]+(-0.315656003785)*x_ref[2]**o+(-0.806577859045)*x_ref[2]
        ref[1,4,2,0]=(-0.686677179384)*x_ref[0]**o+(0.975041194178)*x_ref[0]+(0.840117893335)*x_ref[1]**o+(-0.357121725686)*x_ref[1]+(-0.113947327576)*x_ref[2]**o+(0.722741254436)*x_ref[2]
        ref[1,4,2,1]=(-0.238773557516)*x_ref[0]**o+(0.589422174847)*x_ref[0]+(-0.39358233128)*x_ref[1]**o+(0.159348401427)*x_ref[1]+(0.887216043103)*x_ref[2]**o+(-0.924304870822)*x_ref[2]
        ref[2,0,0,0]=(-0.725901161347)*x_ref[0]**o+(0.144957603906)*x_ref[0]+(0.231954835231)*x_ref[1]**o+(-0.760212283123)*x_ref[1]+(-0.725424437779)*x_ref[2]**o+(-0.19688509792)*x_ref[2]
        ref[2,0,0,1]=(0.418995010234)*x_ref[0]**o+(-0.942078773861)*x_ref[0]+(-0.818049256222)*x_ref[1]**o+(-0.594681673779)*x_ref[1]+(-0.63178803463)*x_ref[2]**o+(-0.989664715484)*x_ref[2]
        ref[2,0,1,0]=(-0.524535447707)*x_ref[0]**o+(-0.746654195315)*x_ref[0]+(-0.721012736821)*x_ref[1]**o+(-0.0345481441746)*x_ref[1]+(0.0537586563756)*x_ref[2]**o+(-0.445033317006)*x_ref[2]
        ref[2,0,1,1]=(-0.411001804179)*x_ref[0]**o+(0.00970860318518)*x_ref[0]+(0.783986695315)*x_ref[1]**o+(-0.270151475512)*x_ref[1]+(-0.652403140687)*x_ref[2]**o+(-0.261945554182)*x_ref[2]
        ref[2,0,2,0]=(-0.497529491593)*x_ref[0]**o+(0.00339011573669)*x_ref[0]+(0.696815843509)*x_ref[1]**o+(-0.277295752437)*x_ref[1]+(0.204803051766)*x_ref[2]**o+(-0.631981280563)*x_ref[2]
        ref[2,0,2,1]=(0.321760872287)*x_ref[0]**o+(0.54137067299)*x_ref[0]+(-0.971971260117)*x_ref[1]**o+(0.767307092712)*x_ref[1]+(0.471809380038)*x_ref[2]**o+(0.178504669856)*x_ref[2]
        ref[2,1,0,0]=(-0.627353418617)*x_ref[0]**o+(-0.610846214602)*x_ref[0]+(-0.610318029871)*x_ref[1]**o+(0.884848156201)*x_ref[1]+(-0.227603047617)*x_ref[2]**o+(0.728596663375)*x_ref[2]
        ref[2,1,0,1]=(0.205703168727)*x_ref[0]**o+(-0.869108006912)*x_ref[0]+(0.195195654062)*x_ref[1]**o+(0.586300414612)*x_ref[1]+(0.787224810167)*x_ref[2]**o+(-0.0278209500568)*x_ref[2]
        ref[2,1,1,0]=(-0.978822062351)*x_ref[0]**o+(0.929683906062)*x_ref[0]+(-0.865432742145)*x_ref[1]**o+(-0.083101823301)*x_ref[1]+(0.264956210564)*x_ref[2]**o+(0.452513746821)*x_ref[2]
        ref[2,1,1,1]=(-0.607905485886)*x_ref[0]**o+(-0.683168972026)*x_ref[0]+(-0.973085006343)*x_ref[1]**o+(0.905969981853)*x_ref[1]+(0.336105649928)*x_ref[2]**o+(0.976694664618)*x_ref[2]
        ref[2,1,2,0]=(-0.542965644364)*x_ref[0]**o+(-0.740434309738)*x_ref[0]+(0.756162713574)*x_ref[1]**o+(-0.543216421535)*x_ref[1]+(0.0942001276823)*x_ref[2]**o+(0.0482815123591)*x_ref[2]
        ref[2,1,2,1]=(-0.226868852637)*x_ref[0]**o+(0.863234399903)*x_ref[0]+(-0.318469617308)*x_ref[1]**o+(0.0192280454449)*x_ref[1]+(-0.23611990172)*x_ref[2]**o+(0.232713876707)*x_ref[2]
        ref[2,2,0,0]=(0.726586748484)*x_ref[0]**o+(-0.711554904751)*x_ref[0]+(-0.206168377552)*x_ref[1]**o+(0.84673411536)*x_ref[1]+(-0.594942018777)*x_ref[2]**o+(-0.199098137058)*x_ref[2]
        ref[2,2,0,1]=(0.797788406588)*x_ref[0]**o+(0.326417288103)*x_ref[0]+(0.954847835729)*x_ref[1]**o+(-0.823541118309)*x_ref[1]+(-0.183282866634)*x_ref[2]**o+(-0.73879700622)*x_ref[2]
        ref[2,2,1,0]=(-0.542135645496)*x_ref[0]**o+(0.617419480945)*x_ref[0]+(-0.858155129941)*x_ref[1]**o+(0.622909241986)*x_ref[1]+(-0.817913616615)*x_ref[2]**o+(0.551023741548)*x_ref[2]
        ref[2,2,1,1]=(-0.033791510266)*x_ref[0]**o+(0.960755607224)*x_ref[0]+(0.992551377675)*x_ref[1]**o+(-0.40568563699)*x_ref[1]+(0.635141614881)*x_ref[2]**o+(-0.28322398176)*x_ref[2]
        ref[2,2,2,0]=(0.200301024931)*x_ref[0]**o+(-0.526020099906)*x_ref[0]+(0.296360677546)*x_ref[1]**o+(-0.346356255978)*x_ref[1]+(-0.751246948633)*x_ref[2]**o+(0.936356516645)*x_ref[2]
        ref[2,2,2,1]=(0.851662248368)*x_ref[0]**o+(0.306264287809)*x_ref[0]+(0.413913851675)*x_ref[1]**o+(0.796748819208)*x_ref[1]+(0.686640823111)*x_ref[2]**o+(0.270797124789)*x_ref[2]
        ref[2,3,0,0]=(0.873612484608)*x_ref[0]**o+(-0.635443314156)*x_ref[0]+(0.0336582752521)*x_ref[1]**o+(0.397952617963)*x_ref[1]+(0.708037547846)*x_ref[2]**o+(-0.835497214327)*x_ref[2]
        ref[2,3,0,1]=(0.266263219836)*x_ref[0]**o+(0.718604372519)*x_ref[0]+(-0.706375167711)*x_ref[1]**o+(0.267962253067)*x_ref[1]+(-0.0369901462112)*x_ref[2]**o+(0.541887114998)*x_ref[2]
        ref[2,3,1,0]=(-0.0883529797185)*x_ref[0]**o+(-0.695144807666)*x_ref[0]+(-0.332082923487)*x_ref[1]**o+(0.0386905137005)*x_ref[1]+(-0.0402663506352)*x_ref[2]**o+(0.897417224025)*x_ref[2]
        ref[2,3,1,1]=(-0.275313948401)*x_ref[0]**o+(0.66334859913)*x_ref[0]+(-0.139887516171)*x_ref[1]**o+(0.874104988552)*x_ref[1]+(-0.474834712531)*x_ref[2]**o+(-0.350478879758)*x_ref[2]
        ref[2,3,2,0]=(-0.370570829115)*x_ref[0]**o+(0.448685344991)*x_ref[0]+(-0.663934134305)*x_ref[1]**o+(0.270424636731)*x_ref[1]+(0.257023713378)*x_ref[2]**o+(-0.558834446028)*x_ref[2]
        ref[2,3,2,1]=(0.853600423341)*x_ref[0]**o+(-0.34026373254)*x_ref[0]+(-0.880960615323)*x_ref[1]**o+(0.720247030686)*x_ref[1]+(-0.179396842577)*x_ref[2]**o+(-0.581055218136)*x_ref[2]
        ref[2,4,0,0]=(0.661293829706)*x_ref[0]**o+(-0.278404566365)*x_ref[0]+(-0.668007328124)*x_ref[1]**o+(-0.0762718064116)*x_ref[1]+(-0.559881233142)*x_ref[2]**o+(0.38863355198)*x_ref[2]
        ref[2,4,0,1]=(0.841925292626)*x_ref[0]**o+(-0.907683026635)*x_ref[0]+(0.606133080471)*x_ref[1]**o+(-0.743094760572)*x_ref[1]+(-0.963492675546)*x_ref[2]**o+(0.551853881446)*x_ref[2]
        ref[2,4,1,0]=(0.106157048406)*x_ref[0]**o+(-0.839019337739)*x_ref[0]+(-0.646225647224)*x_ref[1]**o+(0.596433542879)*x_ref[1]+(-0.440653099815)*x_ref[2]**o+(0.612390706411)*x_ref[2]
        ref[2,4,1,1]=(0.0894808258178)*x_ref[0]**o+(0.601107281081)*x_ref[0]+(0.764980611835)*x_ref[1]**o+(-0.028637689046)*x_ref[1]+(-0.163992620355)*x_ref[2]**o+(0.811740898974)*x_ref[2]
        ref[2,4,2,0]=(0.159301482106)*x_ref[0]**o+(-0.2732663929)*x_ref[0]+(0.642848902873)*x_ref[1]**o+(-0.486181762406)*x_ref[1]+(-0.0353954685288)*x_ref[2]**o+(-0.17658875721)*x_ref[2]
        ref[2,4,2,1]=(-0.190069716212)*x_ref[0]**o+(0.911197492671)*x_ref[0]+(-0.404646749357)*x_ref[1]**o+(0.496457219159)*x_ref[1]+(0.331503854252)*x_ref[2]**o+(-0.0824993997725)*x_ref[2]
        ref[3,0,0,0]=(-0.00960152574795)*x_ref[0]**o+(-0.588053191696)*x_ref[0]+(0.134548796202)*x_ref[1]**o+(-0.734698264954)*x_ref[1]+(0.369880572143)*x_ref[2]**o+(-0.184585884928)*x_ref[2]
        ref[3,0,0,1]=(-0.894375930982)*x_ref[0]**o+(-0.757908178588)*x_ref[0]+(-0.862478750821)*x_ref[1]**o+(-0.860344696497)*x_ref[1]+(-0.606693340193)*x_ref[2]**o+(-0.211110504173)*x_ref[2]
        ref[3,0,1,0]=(0.29208602662)*x_ref[0]**o+(0.428795740188)*x_ref[0]+(0.730773643572)*x_ref[1]**o+(0.473531891026)*x_ref[1]+(-0.475140560425)*x_ref[2]**o+(0.973125507458)*x_ref[2]
        ref[3,0,1,1]=(-0.15741631987)*x_ref[0]**o+(-0.34595691825)*x_ref[0]+(-0.291584586864)*x_ref[1]**o+(0.202892853712)*x_ref[1]+(-0.48541451272)*x_ref[2]**o+(-0.0611059721993)*x_ref[2]
        ref[3,0,2,0]=(-0.309937469651)*x_ref[0]**o+(-0.515728964606)*x_ref[0]+(-0.212592556423)*x_ref[1]**o+(-0.419645818415)*x_ref[1]+(-0.825715279525)*x_ref[2]**o+(0.701775440231)*x_ref[2]
        ref[3,0,2,1]=(-0.527254609361)*x_ref[0]**o+(-0.921224942687)*x_ref[0]+(-0.219965563549)*x_ref[1]**o+(-0.520137812628)*x_ref[1]+(0.228621345698)*x_ref[2]**o+(0.607942457443)*x_ref[2]
        ref[3,1,0,0]=(-0.0191513045468)*x_ref[0]**o+(0.0593277991822)*x_ref[0]+(0.679065928035)*x_ref[1]**o+(-0.824466221695)*x_ref[1]+(-0.0207465360942)*x_ref[2]**o+(-0.462448546311)*x_ref[2]
        ref[3,1,0,1]=(-0.0226385405157)*x_ref[0]**o+(-0.990729100489)*x_ref[0]+(0.928646343096)*x_ref[1]**o+(0.877879386627)*x_ref[1]+(-0.774330622719)*x_ref[2]**o+(0.45794763385)*x_ref[2]
        ref[3,1,1,0]=(0.991939307411)*x_ref[0]**o+(0.424144367324)*x_ref[0]+(0.287966132208)*x_ref[1]**o+(0.560134521129)*x_ref[1]+(-0.366498917899)*x_ref[2]**o+(-0.5499574933)*x_ref[2]
        ref[3,1,1,1]=(0.131575237703)*x_ref[0]**o+(0.350065088754)*x_ref[0]+(0.845775797719)*x_ref[1]**o+(-0.951498449401)*x_ref[1]+(-0.749988925873)*x_ref[2]**o+(0.437681907383)*x_ref[2]
        ref[3,1,2,0]=(-0.177690055962)*x_ref[0]**o+(0.400513514288)*x_ref[0]+(0.654580691817)*x_ref[1]**o+(0.781954834622)*x_ref[1]+(0.498508301282)*x_ref[2]**o+(0.489109718057)*x_ref[2]
        ref[3,1,2,1]=(0.141266612215)*x_ref[0]**o+(0.578894900597)*x_ref[0]+(-0.0990011451546)*x_ref[1]**o+(0.455191961)*x_ref[1]+(0.446639012332)*x_ref[2]**o+(0.398374036483)*x_ref[2]
        ref[3,2,0,0]=(-0.88698978852)*x_ref[0]**o+(0.782997043916)*x_ref[0]+(-0.0203503423342)*x_ref[1]**o+(-0.729317082717)*x_ref[1]+(-0.916473605189)*x_ref[2]**o+(0.706748074855)*x_ref[2]
        ref[3,2,0,1]=(0.955314840013)*x_ref[0]**o+(0.784566529939)*x_ref[0]+(-0.607468453224)*x_ref[1]**o+(-0.960125382144)*x_ref[1]+(-0.62579570096)*x_ref[2]**o+(-0.0943302040401)*x_ref[2]
        ref[3,2,1,0]=(-0.255287575915)*x_ref[0]**o+(0.418535282029)*x_ref[0]+(0.606559849688)*x_ref[1]**o+(0.940877031124)*x_ref[1]+(0.714131644166)*x_ref[2]**o+(-0.77112172198)*x_ref[2]
        ref[3,2,1,1]=(0.566383366127)*x_ref[0]**o+(0.301059560952)*x_ref[0]+(0.181921162673)*x_ref[1]**o+(-0.972651860466)*x_ref[1]+(-0.435658160785)*x_ref[2]**o+(-0.449044377115)*x_ref[2]
        ref[3,2,2,0]=(0.730443884106)*x_ref[0]**o+(0.535861749024)*x_ref[0]+(-0.257910068997)*x_ref[1]**o+(-0.879923650189)*x_ref[1]+(0.24309905047)*x_ref[2]**o+(-0.407314031684)*x_ref[2]
        ref[3,2,2,1]=(-0.734460242423)*x_ref[0]**o+(0.694756195588)*x_ref[0]+(-0.0110855577415)*x_ref[1]**o+(0.213770377606)*x_ref[1]+(0.404975692507)*x_ref[2]**o+(0.235122345113)*x_ref[2]
        ref[3,3,0,0]=(-0.0854800701504)*x_ref[0]**o+(-0.998984290225)*x_ref[0]+(-0.781970276212)*x_ref[1]**o+(0.866788214566)*x_ref[1]+(-0.21747695978)*x_ref[2]**o+(-0.615256484308)*x_ref[2]
        ref[3,3,0,1]=(-0.509246205985)*x_ref[0]**o+(-0.870342470261)*x_ref[0]+(0.699615910083)*x_ref[1]**o+(0.81056073388)*x_ref[1]+(0.137830446242)*x_ref[2]**o+(-0.355111214574)*x_ref[2]
        ref[3,3,1,0]=(-0.423472624091)*x_ref[0]**o+(0.812256460885)*x_ref[0]+(-0.675497703361)*x_ref[1]**o+(0.296366723996)*x_ref[1]+(-0.218492550353)*x_ref[2]**o+(-0.852823171625)*x_ref[2]
        ref[3,3,1,1]=(-0.0718862672517)*x_ref[0]**o+(-0.122792076285)*x_ref[0]+(-0.794191931729)*x_ref[1]**o+(0.569212982974)*x_ref[1]+(-0.646851040773)*x_ref[2]**o+(-0.3073262807)*x_ref[2]
        ref[3,3,2,0]=(-0.994778576692)*x_ref[0]**o+(-0.250054159815)*x_ref[0]+(0.726266056262)*x_ref[1]**o+(-0.827929598129)*x_ref[1]+(0.038573450577)*x_ref[2]**o+(-0.402624412259)*x_ref[2]
        ref[3,3,2,1]=(-0.158546722118)*x_ref[0]**o+(0.438480217845)*x_ref[0]+(-0.757053187223)*x_ref[1]**o+(0.96669102229)*x_ref[1]+(0.585680018048)*x_ref[2]**o+(0.917795585404)*x_ref[2]
        ref[3,4,0,0]=(0.964942257873)*x_ref[0]**o+(0.222693858545)*x_ref[0]+(0.708449965849)*x_ref[1]**o+(-0.786172527336)*x_ref[1]+(0.813715873004)*x_ref[2]**o+(-0.0971222939825)*x_ref[2]
        ref[3,4,0,1]=(-0.14236975971)*x_ref[0]**o+(-0.748307677379)*x_ref[0]+(-0.269260913515)*x_ref[1]**o+(-0.315981317137)*x_ref[1]+(-0.375598072402)*x_ref[2]**o+(0.0797466318359)*x_ref[2]
        ref[3,4,1,0]=(0.52711029533)*x_ref[0]**o+(0.492047263035)*x_ref[0]+(0.678479511876)*x_ref[1]**o+(0.619219421066)*x_ref[1]+(-0.528779830025)*x_ref[2]**o+(-0.237989972194)*x_ref[2]
        ref[3,4,1,1]=(0.711881464696)*x_ref[0]**o+(-0.398127815977)*x_ref[0]+(-0.943946294128)*x_ref[1]**o+(0.584800438848)*x_ref[1]+(-0.683342581136)*x_ref[2]**o+(0.895377849601)*x_ref[2]
        ref[3,4,2,0]=(-0.698456333887)*x_ref[0]**o+(-0.36871456039)*x_ref[0]+(0.216620614245)*x_ref[1]**o+(-0.242036817594)*x_ref[1]+(-0.845502631302)*x_ref[2]**o+(-0.705164123503)*x_ref[2]
        ref[3,4,2,1]=(-0.000297852657095)*x_ref[0]**o+(-0.847929455729)*x_ref[0]+(0.496739572695)*x_ref[1]**o+(0.574135132795)*x_ref[1]+(0.585537276618)*x_ref[2]**o+(-0.934539532574)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.713276041917)*x[0]**o+(-0.0688503309598)*x[0]+(-0.0478916615918)*x[1]**o+(-0.868054043492)*x[1]
        ref=(-0.713276041917)*x_ref[0]**o+(-0.0688503309598)*x_ref[0]+(-0.0478916615918)*x_ref[1]**o+(-0.868054043492)*x_ref[1]
      else:
        s=(-0.931008947411)*x[0]**o+(0.772025637695)*x[0]+(-0.155082242593)*x[1]**o+(0.808407227202)*x[1]+(0.860761693977)*x[2]**o+(-0.965019813018)*x[2]
        ref=(-0.931008947411)*x_ref[0]**o+(0.772025637695)*x_ref[0]+(-0.155082242593)*x_ref[1]**o+(0.808407227202)*x_ref[1]+(0.860761693977)*x_ref[2]**o+(-0.965019813018)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.752443452146)*x[0]**o+(0.0736601797185)*x[0]+(-0.735031045379)*x[1]**o+(0.538489905213)*x[1]
        s[1]=(0.799601080205)*x[0]**o+(0.214497318519)*x[0]+(0.168613900079)*x[1]**o+(0.0388630346028)*x[1]
        ref[0]=(0.752443452146)*x_ref[0]**o+(0.0736601797185)*x_ref[0]+(-0.735031045379)*x_ref[1]**o+(0.538489905213)*x_ref[1]
        ref[1]=(0.799601080205)*x_ref[0]**o+(0.214497318519)*x_ref[0]+(0.168613900079)*x_ref[1]**o+(0.0388630346028)*x_ref[1]
      else:
        s[0]=(0.175265620206)*x[0]**o+(-0.333727702895)*x[0]+(-0.932525223484)*x[1]**o+(0.925315017116)*x[1]+(0.928659911973)*x[2]**o+(0.445185886277)*x[2]
        s[1]=(0.311109115611)*x[0]**o+(0.0231809158727)*x[0]+(0.393168814415)*x[1]**o+(0.452573797902)*x[1]+(0.896927701022)*x[2]**o+(0.944685279345)*x[2]
        ref[0]=(0.175265620206)*x_ref[0]**o+(-0.333727702895)*x_ref[0]+(-0.932525223484)*x_ref[1]**o+(0.925315017116)*x_ref[1]+(0.928659911973)*x_ref[2]**o+(0.445185886277)*x_ref[2]
        ref[1]=(0.311109115611)*x_ref[0]**o+(0.0231809158727)*x_ref[0]+(0.393168814415)*x_ref[1]**o+(0.452573797902)*x_ref[1]+(0.896927701022)*x_ref[2]**o+(0.944685279345)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.708553013661)*x[0]**o+(0.114066657094)*x[0]+(-0.84986272259)*x[1]**o+(-0.0539087301019)*x[1]
        s[0,1]=(-0.266588917197)*x[0]**o+(0.0504848378446)*x[0]+(-0.635596388347)*x[1]**o+(-0.261063433899)*x[1]
        s[0,2]=(-0.0894634722171)*x[0]**o+(0.573385597336)*x[0]+(-0.842136337596)*x[1]**o+(0.238667481456)*x[1]
        s[0,3]=(0.87882171546)*x[0]**o+(0.417532202681)*x[0]+(0.936866438423)*x[1]**o+(-0.724174392442)*x[1]
        s[0,4]=(0.0145872637015)*x[0]**o+(0.421850663727)*x[0]+(0.90053153032)*x[1]**o+(0.955508744656)*x[1]
        s[1,0]=(0.140196302181)*x[0]**o+(-0.638550845113)*x[0]+(0.468359848634)*x[1]**o+(0.893083141798)*x[1]
        s[1,1]=(0.984992789447)*x[0]**o+(-0.843621073237)*x[0]+(0.798478310102)*x[1]**o+(-0.363619894589)*x[1]
        s[1,2]=(0.895636184991)*x[0]**o+(-0.229789622077)*x[0]+(0.098171182568)*x[1]**o+(0.579163598552)*x[1]
        s[1,3]=(0.927820509391)*x[0]**o+(-0.947766499615)*x[0]+(0.183243198595)*x[1]**o+(-0.31472874822)*x[1]
        s[1,4]=(-0.966327205994)*x[0]**o+(-0.289895049587)*x[0]+(0.514876200606)*x[1]**o+(0.653407964622)*x[1]
        s[2,0]=(-0.903904444429)*x[0]**o+(-0.263217684631)*x[0]+(0.291271291348)*x[1]**o+(-0.646562462633)*x[1]
        s[2,1]=(-0.528540658615)*x[0]**o+(0.246812336359)*x[0]+(0.658538113503)*x[1]**o+(-0.407851907836)*x[1]
        s[2,2]=(-0.715562167768)*x[0]**o+(-0.855788728416)*x[0]+(-0.993665928364)*x[1]**o+(-0.608513139044)*x[1]
        s[2,3]=(0.363176542145)*x[0]**o+(0.588619594218)*x[0]+(-0.260592717735)*x[1]**o+(-0.558179141071)*x[1]
        s[2,4]=(0.532120055093)*x[0]**o+(0.090598503196)*x[0]+(0.80389069649)*x[1]**o+(-0.988747865176)*x[1]
        s[3,0]=(0.855196983003)*x[0]**o+(0.348709023449)*x[0]+(-0.592402303364)*x[1]**o+(-0.535439951031)*x[1]
        s[3,1]=(-0.132346044705)*x[0]**o+(-0.849374461726)*x[0]+(-0.812790438958)*x[1]**o+(0.291967653653)*x[1]
        s[3,2]=(-0.39273694335)*x[0]**o+(-0.871595348254)*x[0]+(-0.744651099213)*x[1]**o+(-0.231171036464)*x[1]
        s[3,3]=(-0.616438342441)*x[0]**o+(0.0131646132899)*x[0]+(-0.265184331359)*x[1]**o+(0.447239576276)*x[1]
        s[3,4]=(0.870983993837)*x[0]**o+(-0.734468653473)*x[0]+(0.888611194222)*x[1]**o+(0.713808943447)*x[1]
        ref[0,0]=(-0.708553013661)*x_ref[0]**o+(0.114066657094)*x_ref[0]+(-0.84986272259)*x_ref[1]**o+(-0.0539087301019)*x_ref[1]
        ref[0,1]=(-0.266588917197)*x_ref[0]**o+(0.0504848378446)*x_ref[0]+(-0.635596388347)*x_ref[1]**o+(-0.261063433899)*x_ref[1]
        ref[0,2]=(-0.0894634722171)*x_ref[0]**o+(0.573385597336)*x_ref[0]+(-0.842136337596)*x_ref[1]**o+(0.238667481456)*x_ref[1]
        ref[0,3]=(0.87882171546)*x_ref[0]**o+(0.417532202681)*x_ref[0]+(0.936866438423)*x_ref[1]**o+(-0.724174392442)*x_ref[1]
        ref[0,4]=(0.0145872637015)*x_ref[0]**o+(0.421850663727)*x_ref[0]+(0.90053153032)*x_ref[1]**o+(0.955508744656)*x_ref[1]
        ref[1,0]=(0.140196302181)*x_ref[0]**o+(-0.638550845113)*x_ref[0]+(0.468359848634)*x_ref[1]**o+(0.893083141798)*x_ref[1]
        ref[1,1]=(0.984992789447)*x_ref[0]**o+(-0.843621073237)*x_ref[0]+(0.798478310102)*x_ref[1]**o+(-0.363619894589)*x_ref[1]
        ref[1,2]=(0.895636184991)*x_ref[0]**o+(-0.229789622077)*x_ref[0]+(0.098171182568)*x_ref[1]**o+(0.579163598552)*x_ref[1]
        ref[1,3]=(0.927820509391)*x_ref[0]**o+(-0.947766499615)*x_ref[0]+(0.183243198595)*x_ref[1]**o+(-0.31472874822)*x_ref[1]
        ref[1,4]=(-0.966327205994)*x_ref[0]**o+(-0.289895049587)*x_ref[0]+(0.514876200606)*x_ref[1]**o+(0.653407964622)*x_ref[1]
        ref[2,0]=(-0.903904444429)*x_ref[0]**o+(-0.263217684631)*x_ref[0]+(0.291271291348)*x_ref[1]**o+(-0.646562462633)*x_ref[1]
        ref[2,1]=(-0.528540658615)*x_ref[0]**o+(0.246812336359)*x_ref[0]+(0.658538113503)*x_ref[1]**o+(-0.407851907836)*x_ref[1]
        ref[2,2]=(-0.715562167768)*x_ref[0]**o+(-0.855788728416)*x_ref[0]+(-0.993665928364)*x_ref[1]**o+(-0.608513139044)*x_ref[1]
        ref[2,3]=(0.363176542145)*x_ref[0]**o+(0.588619594218)*x_ref[0]+(-0.260592717735)*x_ref[1]**o+(-0.558179141071)*x_ref[1]
        ref[2,4]=(0.532120055093)*x_ref[0]**o+(0.090598503196)*x_ref[0]+(0.80389069649)*x_ref[1]**o+(-0.988747865176)*x_ref[1]
        ref[3,0]=(0.855196983003)*x_ref[0]**o+(0.348709023449)*x_ref[0]+(-0.592402303364)*x_ref[1]**o+(-0.535439951031)*x_ref[1]
        ref[3,1]=(-0.132346044705)*x_ref[0]**o+(-0.849374461726)*x_ref[0]+(-0.812790438958)*x_ref[1]**o+(0.291967653653)*x_ref[1]
        ref[3,2]=(-0.39273694335)*x_ref[0]**o+(-0.871595348254)*x_ref[0]+(-0.744651099213)*x_ref[1]**o+(-0.231171036464)*x_ref[1]
        ref[3,3]=(-0.616438342441)*x_ref[0]**o+(0.0131646132899)*x_ref[0]+(-0.265184331359)*x_ref[1]**o+(0.447239576276)*x_ref[1]
        ref[3,4]=(0.870983993837)*x_ref[0]**o+(-0.734468653473)*x_ref[0]+(0.888611194222)*x_ref[1]**o+(0.713808943447)*x_ref[1]
      else:
        s[0,0]=(-0.810387756876)*x[0]**o+(-0.562586750225)*x[0]+(0.896156420696)*x[1]**o+(-0.996536575244)*x[1]+(0.427600721566)*x[2]**o+(0.837763430808)*x[2]
        s[0,1]=(-0.63001573449)*x[0]**o+(0.654477767706)*x[0]+(-0.795127551271)*x[1]**o+(0.536595939418)*x[1]+(-0.727127268215)*x[2]**o+(-0.358649556898)*x[2]
        s[0,2]=(0.584792270415)*x[0]**o+(-0.13886821113)*x[0]+(-0.526668048387)*x[1]**o+(-0.0830908739307)*x[1]+(0.647801230167)*x[2]**o+(-0.299674557564)*x[2]
        s[0,3]=(0.237272327463)*x[0]**o+(-0.665360840533)*x[0]+(0.642053734612)*x[1]**o+(0.0162314692731)*x[1]+(0.927935045044)*x[2]**o+(0.164570226784)*x[2]
        s[0,4]=(0.219203335717)*x[0]**o+(0.453183788553)*x[0]+(0.124112536481)*x[1]**o+(-0.0284133490265)*x[1]+(-0.919674987737)*x[2]**o+(-0.0149293124196)*x[2]
        s[1,0]=(-0.0768501669381)*x[0]**o+(0.60547593141)*x[0]+(0.715752118361)*x[1]**o+(0.886572828266)*x[1]+(0.00765847416371)*x[2]**o+(0.861052492746)*x[2]
        s[1,1]=(0.255734836806)*x[0]**o+(-0.246883769311)*x[0]+(0.0800805057057)*x[1]**o+(-0.790568208751)*x[1]+(-0.748973939025)*x[2]**o+(-0.182812700448)*x[2]
        s[1,2]=(0.199422259597)*x[0]**o+(-0.627479494962)*x[0]+(0.0565633628195)*x[1]**o+(0.350443086044)*x[1]+(-0.380622372229)*x[2]**o+(0.00708476531876)*x[2]
        s[1,3]=(0.40010280022)*x[0]**o+(-0.767881252757)*x[0]+(-0.56339937337)*x[1]**o+(-0.352542604473)*x[1]+(0.801372900062)*x[2]**o+(-0.126263266304)*x[2]
        s[1,4]=(-0.128959410592)*x[0]**o+(-0.59236859859)*x[0]+(0.0980131449714)*x[1]**o+(0.853907069528)*x[1]+(-0.691379624)*x[2]**o+(0.386547512092)*x[2]
        s[2,0]=(-0.360626473774)*x[0]**o+(-0.76083965008)*x[0]+(0.470179818787)*x[1]**o+(-0.824599812059)*x[1]+(0.420014625648)*x[2]**o+(-0.479269079111)*x[2]
        s[2,1]=(-0.81467966055)*x[0]**o+(-0.329346577295)*x[0]+(0.748928864266)*x[1]**o+(0.989795060424)*x[1]+(0.840479614557)*x[2]**o+(0.361335074457)*x[2]
        s[2,2]=(0.0627744010829)*x[0]**o+(-0.155252499175)*x[0]+(0.195454946586)*x[1]**o+(0.628864956458)*x[1]+(0.664724662178)*x[2]**o+(0.0471499062877)*x[2]
        s[2,3]=(-0.150860729652)*x[0]**o+(-0.18636002517)*x[0]+(0.964787336251)*x[1]**o+(0.871287020262)*x[1]+(-0.557208138103)*x[2]**o+(0.734747013137)*x[2]
        s[2,4]=(0.146220522342)*x[0]**o+(0.828100181189)*x[0]+(0.275317867253)*x[1]**o+(-0.533950851352)*x[1]+(0.058219246545)*x[2]**o+(-0.264489071187)*x[2]
        s[3,0]=(0.759487099327)*x[0]**o+(-0.943901668987)*x[0]+(0.917271393015)*x[1]**o+(0.991514325689)*x[1]+(-0.0693941475922)*x[2]**o+(0.971629779968)*x[2]
        s[3,1]=(-0.26255428914)*x[0]**o+(-0.126823739856)*x[0]+(0.769392620344)*x[1]**o+(0.124852282039)*x[1]+(0.719253662224)*x[2]**o+(0.189838410592)*x[2]
        s[3,2]=(0.513890040965)*x[0]**o+(0.849300040037)*x[0]+(0.375236082211)*x[1]**o+(-0.271681479664)*x[1]+(0.580679441212)*x[2]**o+(-0.0117174804836)*x[2]
        s[3,3]=(0.886420007368)*x[0]**o+(-0.140667253897)*x[0]+(0.85708792282)*x[1]**o+(0.867659131336)*x[1]+(0.418020814822)*x[2]**o+(0.817953483616)*x[2]
        s[3,4]=(0.449966058878)*x[0]**o+(0.424180609442)*x[0]+(0.42140819426)*x[1]**o+(0.180925977263)*x[1]+(-0.571592032058)*x[2]**o+(0.257682237168)*x[2]
        ref[0,0]=(-0.810387756876)*x_ref[0]**o+(-0.562586750225)*x_ref[0]+(0.896156420696)*x_ref[1]**o+(-0.996536575244)*x_ref[1]+(0.427600721566)*x_ref[2]**o+(0.837763430808)*x_ref[2]
        ref[0,1]=(-0.63001573449)*x_ref[0]**o+(0.654477767706)*x_ref[0]+(-0.795127551271)*x_ref[1]**o+(0.536595939418)*x_ref[1]+(-0.727127268215)*x_ref[2]**o+(-0.358649556898)*x_ref[2]
        ref[0,2]=(0.584792270415)*x_ref[0]**o+(-0.13886821113)*x_ref[0]+(-0.526668048387)*x_ref[1]**o+(-0.0830908739307)*x_ref[1]+(0.647801230167)*x_ref[2]**o+(-0.299674557564)*x_ref[2]
        ref[0,3]=(0.237272327463)*x_ref[0]**o+(-0.665360840533)*x_ref[0]+(0.642053734612)*x_ref[1]**o+(0.0162314692731)*x_ref[1]+(0.927935045044)*x_ref[2]**o+(0.164570226784)*x_ref[2]
        ref[0,4]=(0.219203335717)*x_ref[0]**o+(0.453183788553)*x_ref[0]+(0.124112536481)*x_ref[1]**o+(-0.0284133490265)*x_ref[1]+(-0.919674987737)*x_ref[2]**o+(-0.0149293124196)*x_ref[2]
        ref[1,0]=(-0.0768501669381)*x_ref[0]**o+(0.60547593141)*x_ref[0]+(0.715752118361)*x_ref[1]**o+(0.886572828266)*x_ref[1]+(0.00765847416371)*x_ref[2]**o+(0.861052492746)*x_ref[2]
        ref[1,1]=(0.255734836806)*x_ref[0]**o+(-0.246883769311)*x_ref[0]+(0.0800805057057)*x_ref[1]**o+(-0.790568208751)*x_ref[1]+(-0.748973939025)*x_ref[2]**o+(-0.182812700448)*x_ref[2]
        ref[1,2]=(0.199422259597)*x_ref[0]**o+(-0.627479494962)*x_ref[0]+(0.0565633628195)*x_ref[1]**o+(0.350443086044)*x_ref[1]+(-0.380622372229)*x_ref[2]**o+(0.00708476531876)*x_ref[2]
        ref[1,3]=(0.40010280022)*x_ref[0]**o+(-0.767881252757)*x_ref[0]+(-0.56339937337)*x_ref[1]**o+(-0.352542604473)*x_ref[1]+(0.801372900062)*x_ref[2]**o+(-0.126263266304)*x_ref[2]
        ref[1,4]=(-0.128959410592)*x_ref[0]**o+(-0.59236859859)*x_ref[0]+(0.0980131449714)*x_ref[1]**o+(0.853907069528)*x_ref[1]+(-0.691379624)*x_ref[2]**o+(0.386547512092)*x_ref[2]
        ref[2,0]=(-0.360626473774)*x_ref[0]**o+(-0.76083965008)*x_ref[0]+(0.470179818787)*x_ref[1]**o+(-0.824599812059)*x_ref[1]+(0.420014625648)*x_ref[2]**o+(-0.479269079111)*x_ref[2]
        ref[2,1]=(-0.81467966055)*x_ref[0]**o+(-0.329346577295)*x_ref[0]+(0.748928864266)*x_ref[1]**o+(0.989795060424)*x_ref[1]+(0.840479614557)*x_ref[2]**o+(0.361335074457)*x_ref[2]
        ref[2,2]=(0.0627744010829)*x_ref[0]**o+(-0.155252499175)*x_ref[0]+(0.195454946586)*x_ref[1]**o+(0.628864956458)*x_ref[1]+(0.664724662178)*x_ref[2]**o+(0.0471499062877)*x_ref[2]
        ref[2,3]=(-0.150860729652)*x_ref[0]**o+(-0.18636002517)*x_ref[0]+(0.964787336251)*x_ref[1]**o+(0.871287020262)*x_ref[1]+(-0.557208138103)*x_ref[2]**o+(0.734747013137)*x_ref[2]
        ref[2,4]=(0.146220522342)*x_ref[0]**o+(0.828100181189)*x_ref[0]+(0.275317867253)*x_ref[1]**o+(-0.533950851352)*x_ref[1]+(0.058219246545)*x_ref[2]**o+(-0.264489071187)*x_ref[2]
        ref[3,0]=(0.759487099327)*x_ref[0]**o+(-0.943901668987)*x_ref[0]+(0.917271393015)*x_ref[1]**o+(0.991514325689)*x_ref[1]+(-0.0693941475922)*x_ref[2]**o+(0.971629779968)*x_ref[2]
        ref[3,1]=(-0.26255428914)*x_ref[0]**o+(-0.126823739856)*x_ref[0]+(0.769392620344)*x_ref[1]**o+(0.124852282039)*x_ref[1]+(0.719253662224)*x_ref[2]**o+(0.189838410592)*x_ref[2]
        ref[3,2]=(0.513890040965)*x_ref[0]**o+(0.849300040037)*x_ref[0]+(0.375236082211)*x_ref[1]**o+(-0.271681479664)*x_ref[1]+(0.580679441212)*x_ref[2]**o+(-0.0117174804836)*x_ref[2]
        ref[3,3]=(0.886420007368)*x_ref[0]**o+(-0.140667253897)*x_ref[0]+(0.85708792282)*x_ref[1]**o+(0.867659131336)*x_ref[1]+(0.418020814822)*x_ref[2]**o+(0.817953483616)*x_ref[2]
        ref[3,4]=(0.449966058878)*x_ref[0]**o+(0.424180609442)*x_ref[0]+(0.42140819426)*x_ref[1]**o+(0.180925977263)*x_ref[1]+(-0.571592032058)*x_ref[2]**o+(0.257682237168)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.104698071101)*x[0]**o+(0.85225441166)*x[0]+(0.643738758183)*x[1]**o+(0.8443629101)*x[1]
        s[0,0,1]=(-0.293745943532)*x[0]**o+(0.677723913417)*x[0]+(-0.12767285172)*x[1]**o+(-0.228090688839)*x[1]
        s[0,1,0]=(0.114974377468)*x[0]**o+(-0.769098874811)*x[0]+(-0.173276425412)*x[1]**o+(0.0507852449622)*x[1]
        s[0,1,1]=(0.779779857754)*x[0]**o+(0.0482150028893)*x[0]+(0.208388483635)*x[1]**o+(-0.542545383735)*x[1]
        s[1,0,0]=(0.857554163662)*x[0]**o+(0.890423092097)*x[0]+(0.603776777208)*x[1]**o+(-0.35588830751)*x[1]
        s[1,0,1]=(0.277165242151)*x[0]**o+(0.805484036517)*x[0]+(-0.577456699494)*x[1]**o+(-0.160490000703)*x[1]
        s[1,1,0]=(0.928438839711)*x[0]**o+(-0.716888927671)*x[0]+(-0.138675664609)*x[1]**o+(0.146500337023)*x[1]
        s[1,1,1]=(-0.3626191689)*x[0]**o+(-0.442937599469)*x[0]+(-0.407066971319)*x[1]**o+(0.623768042338)*x[1]
        s[2,0,0]=(-0.347034739676)*x[0]**o+(0.715507004381)*x[0]+(0.856404818205)*x[1]**o+(-0.04028097386)*x[1]
        s[2,0,1]=(-0.0606946296735)*x[0]**o+(0.128860413259)*x[0]+(-0.142401804193)*x[1]**o+(-0.0530515258875)*x[1]
        s[2,1,0]=(-0.418284924355)*x[0]**o+(0.608535815348)*x[0]+(-0.753014535299)*x[1]**o+(-0.678095181481)*x[1]
        s[2,1,1]=(0.15035667769)*x[0]**o+(0.0492019049213)*x[0]+(-0.677433635501)*x[1]**o+(0.352404951558)*x[1]
        s[3,0,0]=(-0.0290046263356)*x[0]**o+(-0.804064667193)*x[0]+(-0.804537375276)*x[1]**o+(0.475053057415)*x[1]
        s[3,0,1]=(-0.340794340714)*x[0]**o+(-0.155934699109)*x[0]+(0.91002644066)*x[1]**o+(0.784480548529)*x[1]
        s[3,1,0]=(-0.484483021942)*x[0]**o+(0.52465323919)*x[0]+(0.751560457535)*x[1]**o+(0.28163378712)*x[1]
        s[3,1,1]=(-0.309783414062)*x[0]**o+(0.066240252028)*x[0]+(0.183403362957)*x[1]**o+(-0.523037580611)*x[1]
        s[4,0,0]=(-0.568631885289)*x[0]**o+(-0.654263636024)*x[0]+(-0.605614348372)*x[1]**o+(-0.894236319906)*x[1]
        s[4,0,1]=(0.325194519014)*x[0]**o+(0.585375626733)*x[0]+(0.20845589395)*x[1]**o+(0.193035797172)*x[1]
        s[4,1,0]=(-0.142515818097)*x[0]**o+(-0.654569638237)*x[0]+(-0.928863338211)*x[1]**o+(-0.963259968871)*x[1]
        s[4,1,1]=(0.313307585897)*x[0]**o+(0.185118633115)*x[0]+(-0.439344812316)*x[1]**o+(-0.760250886639)*x[1]
        s[5,0,0]=(-0.592313178206)*x[0]**o+(-0.264499099416)*x[0]+(-0.620889224638)*x[1]**o+(0.457484719025)*x[1]
        s[5,0,1]=(-0.083928120416)*x[0]**o+(0.25893121162)*x[0]+(0.403250910043)*x[1]**o+(0.0822798058585)*x[1]
        s[5,1,0]=(0.617807311)*x[0]**o+(0.0627862768802)*x[0]+(-0.595563604156)*x[1]**o+(-0.157413618531)*x[1]
        s[5,1,1]=(0.919933775557)*x[0]**o+(-0.736703943919)*x[0]+(-0.935265088031)*x[1]**o+(-0.953284151506)*x[1]
        ref[0,0,0]=(0.104698071101)*x_ref[0]**o+(0.85225441166)*x_ref[0]+(0.643738758183)*x_ref[1]**o+(0.8443629101)*x_ref[1]
        ref[0,0,1]=(-0.293745943532)*x_ref[0]**o+(0.677723913417)*x_ref[0]+(-0.12767285172)*x_ref[1]**o+(-0.228090688839)*x_ref[1]
        ref[0,1,0]=(0.114974377468)*x_ref[0]**o+(-0.769098874811)*x_ref[0]+(-0.173276425412)*x_ref[1]**o+(0.0507852449622)*x_ref[1]
        ref[0,1,1]=(0.779779857754)*x_ref[0]**o+(0.0482150028893)*x_ref[0]+(0.208388483635)*x_ref[1]**o+(-0.542545383735)*x_ref[1]
        ref[1,0,0]=(0.857554163662)*x_ref[0]**o+(0.890423092097)*x_ref[0]+(0.603776777208)*x_ref[1]**o+(-0.35588830751)*x_ref[1]
        ref[1,0,1]=(0.277165242151)*x_ref[0]**o+(0.805484036517)*x_ref[0]+(-0.577456699494)*x_ref[1]**o+(-0.160490000703)*x_ref[1]
        ref[1,1,0]=(0.928438839711)*x_ref[0]**o+(-0.716888927671)*x_ref[0]+(-0.138675664609)*x_ref[1]**o+(0.146500337023)*x_ref[1]
        ref[1,1,1]=(-0.3626191689)*x_ref[0]**o+(-0.442937599469)*x_ref[0]+(-0.407066971319)*x_ref[1]**o+(0.623768042338)*x_ref[1]
        ref[2,0,0]=(-0.347034739676)*x_ref[0]**o+(0.715507004381)*x_ref[0]+(0.856404818205)*x_ref[1]**o+(-0.04028097386)*x_ref[1]
        ref[2,0,1]=(-0.0606946296735)*x_ref[0]**o+(0.128860413259)*x_ref[0]+(-0.142401804193)*x_ref[1]**o+(-0.0530515258875)*x_ref[1]
        ref[2,1,0]=(-0.418284924355)*x_ref[0]**o+(0.608535815348)*x_ref[0]+(-0.753014535299)*x_ref[1]**o+(-0.678095181481)*x_ref[1]
        ref[2,1,1]=(0.15035667769)*x_ref[0]**o+(0.0492019049213)*x_ref[0]+(-0.677433635501)*x_ref[1]**o+(0.352404951558)*x_ref[1]
        ref[3,0,0]=(-0.0290046263356)*x_ref[0]**o+(-0.804064667193)*x_ref[0]+(-0.804537375276)*x_ref[1]**o+(0.475053057415)*x_ref[1]
        ref[3,0,1]=(-0.340794340714)*x_ref[0]**o+(-0.155934699109)*x_ref[0]+(0.91002644066)*x_ref[1]**o+(0.784480548529)*x_ref[1]
        ref[3,1,0]=(-0.484483021942)*x_ref[0]**o+(0.52465323919)*x_ref[0]+(0.751560457535)*x_ref[1]**o+(0.28163378712)*x_ref[1]
        ref[3,1,1]=(-0.309783414062)*x_ref[0]**o+(0.066240252028)*x_ref[0]+(0.183403362957)*x_ref[1]**o+(-0.523037580611)*x_ref[1]
        ref[4,0,0]=(-0.568631885289)*x_ref[0]**o+(-0.654263636024)*x_ref[0]+(-0.605614348372)*x_ref[1]**o+(-0.894236319906)*x_ref[1]
        ref[4,0,1]=(0.325194519014)*x_ref[0]**o+(0.585375626733)*x_ref[0]+(0.20845589395)*x_ref[1]**o+(0.193035797172)*x_ref[1]
        ref[4,1,0]=(-0.142515818097)*x_ref[0]**o+(-0.654569638237)*x_ref[0]+(-0.928863338211)*x_ref[1]**o+(-0.963259968871)*x_ref[1]
        ref[4,1,1]=(0.313307585897)*x_ref[0]**o+(0.185118633115)*x_ref[0]+(-0.439344812316)*x_ref[1]**o+(-0.760250886639)*x_ref[1]
        ref[5,0,0]=(-0.592313178206)*x_ref[0]**o+(-0.264499099416)*x_ref[0]+(-0.620889224638)*x_ref[1]**o+(0.457484719025)*x_ref[1]
        ref[5,0,1]=(-0.083928120416)*x_ref[0]**o+(0.25893121162)*x_ref[0]+(0.403250910043)*x_ref[1]**o+(0.0822798058585)*x_ref[1]
        ref[5,1,0]=(0.617807311)*x_ref[0]**o+(0.0627862768802)*x_ref[0]+(-0.595563604156)*x_ref[1]**o+(-0.157413618531)*x_ref[1]
        ref[5,1,1]=(0.919933775557)*x_ref[0]**o+(-0.736703943919)*x_ref[0]+(-0.935265088031)*x_ref[1]**o+(-0.953284151506)*x_ref[1]
      else:
        s[0,0,0]=(-0.542725715353)*x[0]**o+(-0.388718702529)*x[0]+(-0.260726707092)*x[1]**o+(0.00868281860494)*x[1]+(-0.422209705753)*x[2]**o+(-0.482952338663)*x[2]
        s[0,0,1]=(0.517312862567)*x[0]**o+(-0.213607124389)*x[0]+(-0.717854605214)*x[1]**o+(0.102559996175)*x[1]+(-0.248146688044)*x[2]**o+(-0.472262287207)*x[2]
        s[0,1,0]=(-0.561062159702)*x[0]**o+(0.625784702743)*x[0]+(0.127217512593)*x[1]**o+(0.431860689921)*x[1]+(-0.743622124184)*x[2]**o+(-0.549196149668)*x[2]
        s[0,1,1]=(-0.47244041365)*x[0]**o+(-0.905145373286)*x[0]+(0.0327050909564)*x[1]**o+(-0.0753932245132)*x[1]+(-0.669475816436)*x[2]**o+(-0.913342589953)*x[2]
        s[1,0,0]=(-0.107737055645)*x[0]**o+(0.890187876057)*x[0]+(0.619389534096)*x[1]**o+(0.505636695416)*x[1]+(-0.773316774174)*x[2]**o+(0.57289098512)*x[2]
        s[1,0,1]=(0.998166227402)*x[0]**o+(0.426101053356)*x[0]+(-0.275256494275)*x[1]**o+(-0.642996762298)*x[1]+(0.80180314192)*x[2]**o+(-0.777630700737)*x[2]
        s[1,1,0]=(0.0567929745565)*x[0]**o+(-0.634357283891)*x[0]+(0.756220569552)*x[1]**o+(0.225394362055)*x[1]+(-0.380906558144)*x[2]**o+(0.656365329349)*x[2]
        s[1,1,1]=(-0.118345871482)*x[0]**o+(-0.601934055233)*x[0]+(-0.466776830591)*x[1]**o+(0.83729192062)*x[1]+(-0.760321613956)*x[2]**o+(-0.0364679505353)*x[2]
        s[2,0,0]=(-0.0254736115914)*x[0]**o+(0.219281883056)*x[0]+(0.616921568912)*x[1]**o+(0.955364895459)*x[1]+(0.468886733279)*x[2]**o+(-0.665678021975)*x[2]
        s[2,0,1]=(0.452552793802)*x[0]**o+(0.418962651335)*x[0]+(-0.493607930291)*x[1]**o+(0.446553418335)*x[1]+(0.297233373077)*x[2]**o+(0.140710729735)*x[2]
        s[2,1,0]=(-0.636104671214)*x[0]**o+(0.770136919493)*x[0]+(0.559068996696)*x[1]**o+(0.1863482658)*x[1]+(-0.486115384192)*x[2]**o+(0.766345710304)*x[2]
        s[2,1,1]=(0.365405358671)*x[0]**o+(-0.349630033342)*x[0]+(0.134851622297)*x[1]**o+(0.274415828646)*x[1]+(0.736462073823)*x[2]**o+(0.113366567452)*x[2]
        s[3,0,0]=(0.103210866706)*x[0]**o+(0.781535459867)*x[0]+(-0.278787888162)*x[1]**o+(-0.931650808802)*x[1]+(0.640371737202)*x[2]**o+(-0.0105202640396)*x[2]
        s[3,0,1]=(-0.249043848359)*x[0]**o+(0.524828750016)*x[0]+(0.533986967075)*x[1]**o+(-0.267136752764)*x[1]+(0.451666436163)*x[2]**o+(-0.205612541545)*x[2]
        s[3,1,0]=(-0.604476259822)*x[0]**o+(-0.0957734520284)*x[0]+(-0.970823460222)*x[1]**o+(0.980588655716)*x[1]+(-0.124494413992)*x[2]**o+(0.252216095396)*x[2]
        s[3,1,1]=(0.0212563582378)*x[0]**o+(-0.0106688543408)*x[0]+(-0.362460274705)*x[1]**o+(-0.210672777607)*x[1]+(0.459825480445)*x[2]**o+(0.317321978801)*x[2]
        s[4,0,0]=(0.587000133255)*x[0]**o+(0.362106866018)*x[0]+(0.839110518219)*x[1]**o+(-0.183918723198)*x[1]+(-0.357506983794)*x[2]**o+(0.964414772317)*x[2]
        s[4,0,1]=(-0.914924387748)*x[0]**o+(-0.712924404308)*x[0]+(-0.81536262433)*x[1]**o+(-0.246055912409)*x[1]+(0.550257872298)*x[2]**o+(0.308874758093)*x[2]
        s[4,1,0]=(-0.451464088139)*x[0]**o+(0.113750801605)*x[0]+(-0.204501128464)*x[1]**o+(0.694001185736)*x[1]+(0.80368669429)*x[2]**o+(0.564414553079)*x[2]
        s[4,1,1]=(-0.354511540467)*x[0]**o+(0.295005960713)*x[0]+(-0.581487363676)*x[1]**o+(-0.862369285262)*x[1]+(0.268927694418)*x[2]**o+(0.91534230742)*x[2]
        s[5,0,0]=(-0.361145485775)*x[0]**o+(-0.136825333824)*x[0]+(-0.917163652695)*x[1]**o+(-0.92225469793)*x[1]+(0.142972538406)*x[2]**o+(0.625895771061)*x[2]
        s[5,0,1]=(0.976797635598)*x[0]**o+(-0.30046259343)*x[0]+(-0.831191492174)*x[1]**o+(-0.422489136999)*x[1]+(0.963327796572)*x[2]**o+(0.216709868453)*x[2]
        s[5,1,0]=(-0.396424567827)*x[0]**o+(-0.037861045243)*x[0]+(0.461627073863)*x[1]**o+(0.299397593712)*x[1]+(0.607619683163)*x[2]**o+(-0.260194783511)*x[2]
        s[5,1,1]=(0.419140017883)*x[0]**o+(0.383165340851)*x[0]+(0.561345759207)*x[1]**o+(-0.406706766251)*x[1]+(-0.730360027768)*x[2]**o+(-0.0907334422202)*x[2]
        ref[0,0,0]=(-0.542725715353)*x_ref[0]**o+(-0.388718702529)*x_ref[0]+(-0.260726707092)*x_ref[1]**o+(0.00868281860494)*x_ref[1]+(-0.422209705753)*x_ref[2]**o+(-0.482952338663)*x_ref[2]
        ref[0,0,1]=(0.517312862567)*x_ref[0]**o+(-0.213607124389)*x_ref[0]+(-0.717854605214)*x_ref[1]**o+(0.102559996175)*x_ref[1]+(-0.248146688044)*x_ref[2]**o+(-0.472262287207)*x_ref[2]
        ref[0,1,0]=(-0.561062159702)*x_ref[0]**o+(0.625784702743)*x_ref[0]+(0.127217512593)*x_ref[1]**o+(0.431860689921)*x_ref[1]+(-0.743622124184)*x_ref[2]**o+(-0.549196149668)*x_ref[2]
        ref[0,1,1]=(-0.47244041365)*x_ref[0]**o+(-0.905145373286)*x_ref[0]+(0.0327050909564)*x_ref[1]**o+(-0.0753932245132)*x_ref[1]+(-0.669475816436)*x_ref[2]**o+(-0.913342589953)*x_ref[2]
        ref[1,0,0]=(-0.107737055645)*x_ref[0]**o+(0.890187876057)*x_ref[0]+(0.619389534096)*x_ref[1]**o+(0.505636695416)*x_ref[1]+(-0.773316774174)*x_ref[2]**o+(0.57289098512)*x_ref[2]
        ref[1,0,1]=(0.998166227402)*x_ref[0]**o+(0.426101053356)*x_ref[0]+(-0.275256494275)*x_ref[1]**o+(-0.642996762298)*x_ref[1]+(0.80180314192)*x_ref[2]**o+(-0.777630700737)*x_ref[2]
        ref[1,1,0]=(0.0567929745565)*x_ref[0]**o+(-0.634357283891)*x_ref[0]+(0.756220569552)*x_ref[1]**o+(0.225394362055)*x_ref[1]+(-0.380906558144)*x_ref[2]**o+(0.656365329349)*x_ref[2]
        ref[1,1,1]=(-0.118345871482)*x_ref[0]**o+(-0.601934055233)*x_ref[0]+(-0.466776830591)*x_ref[1]**o+(0.83729192062)*x_ref[1]+(-0.760321613956)*x_ref[2]**o+(-0.0364679505353)*x_ref[2]
        ref[2,0,0]=(-0.0254736115914)*x_ref[0]**o+(0.219281883056)*x_ref[0]+(0.616921568912)*x_ref[1]**o+(0.955364895459)*x_ref[1]+(0.468886733279)*x_ref[2]**o+(-0.665678021975)*x_ref[2]
        ref[2,0,1]=(0.452552793802)*x_ref[0]**o+(0.418962651335)*x_ref[0]+(-0.493607930291)*x_ref[1]**o+(0.446553418335)*x_ref[1]+(0.297233373077)*x_ref[2]**o+(0.140710729735)*x_ref[2]
        ref[2,1,0]=(-0.636104671214)*x_ref[0]**o+(0.770136919493)*x_ref[0]+(0.559068996696)*x_ref[1]**o+(0.1863482658)*x_ref[1]+(-0.486115384192)*x_ref[2]**o+(0.766345710304)*x_ref[2]
        ref[2,1,1]=(0.365405358671)*x_ref[0]**o+(-0.349630033342)*x_ref[0]+(0.134851622297)*x_ref[1]**o+(0.274415828646)*x_ref[1]+(0.736462073823)*x_ref[2]**o+(0.113366567452)*x_ref[2]
        ref[3,0,0]=(0.103210866706)*x_ref[0]**o+(0.781535459867)*x_ref[0]+(-0.278787888162)*x_ref[1]**o+(-0.931650808802)*x_ref[1]+(0.640371737202)*x_ref[2]**o+(-0.0105202640396)*x_ref[2]
        ref[3,0,1]=(-0.249043848359)*x_ref[0]**o+(0.524828750016)*x_ref[0]+(0.533986967075)*x_ref[1]**o+(-0.267136752764)*x_ref[1]+(0.451666436163)*x_ref[2]**o+(-0.205612541545)*x_ref[2]
        ref[3,1,0]=(-0.604476259822)*x_ref[0]**o+(-0.0957734520284)*x_ref[0]+(-0.970823460222)*x_ref[1]**o+(0.980588655716)*x_ref[1]+(-0.124494413992)*x_ref[2]**o+(0.252216095396)*x_ref[2]
        ref[3,1,1]=(0.0212563582378)*x_ref[0]**o+(-0.0106688543408)*x_ref[0]+(-0.362460274705)*x_ref[1]**o+(-0.210672777607)*x_ref[1]+(0.459825480445)*x_ref[2]**o+(0.317321978801)*x_ref[2]
        ref[4,0,0]=(0.587000133255)*x_ref[0]**o+(0.362106866018)*x_ref[0]+(0.839110518219)*x_ref[1]**o+(-0.183918723198)*x_ref[1]+(-0.357506983794)*x_ref[2]**o+(0.964414772317)*x_ref[2]
        ref[4,0,1]=(-0.914924387748)*x_ref[0]**o+(-0.712924404308)*x_ref[0]+(-0.81536262433)*x_ref[1]**o+(-0.246055912409)*x_ref[1]+(0.550257872298)*x_ref[2]**o+(0.308874758093)*x_ref[2]
        ref[4,1,0]=(-0.451464088139)*x_ref[0]**o+(0.113750801605)*x_ref[0]+(-0.204501128464)*x_ref[1]**o+(0.694001185736)*x_ref[1]+(0.80368669429)*x_ref[2]**o+(0.564414553079)*x_ref[2]
        ref[4,1,1]=(-0.354511540467)*x_ref[0]**o+(0.295005960713)*x_ref[0]+(-0.581487363676)*x_ref[1]**o+(-0.862369285262)*x_ref[1]+(0.268927694418)*x_ref[2]**o+(0.91534230742)*x_ref[2]
        ref[5,0,0]=(-0.361145485775)*x_ref[0]**o+(-0.136825333824)*x_ref[0]+(-0.917163652695)*x_ref[1]**o+(-0.92225469793)*x_ref[1]+(0.142972538406)*x_ref[2]**o+(0.625895771061)*x_ref[2]
        ref[5,0,1]=(0.976797635598)*x_ref[0]**o+(-0.30046259343)*x_ref[0]+(-0.831191492174)*x_ref[1]**o+(-0.422489136999)*x_ref[1]+(0.963327796572)*x_ref[2]**o+(0.216709868453)*x_ref[2]
        ref[5,1,0]=(-0.396424567827)*x_ref[0]**o+(-0.037861045243)*x_ref[0]+(0.461627073863)*x_ref[1]**o+(0.299397593712)*x_ref[1]+(0.607619683163)*x_ref[2]**o+(-0.260194783511)*x_ref[2]
        ref[5,1,1]=(0.419140017883)*x_ref[0]**o+(0.383165340851)*x_ref[0]+(0.561345759207)*x_ref[1]**o+(-0.406706766251)*x_ref[1]+(-0.730360027768)*x_ref[2]**o+(-0.0907334422202)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from Solution onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.367629818255)*x[0]**o+(-0.682992553528)*x[0]+(-0.958578674565)*x[1]**o+(0.994004366609)*x[1]
        s[0,0,0,1]=(-0.933871216086)*x[0]**o+(0.089303727901)*x[0]+(0.876179784981)*x[1]**o+(-0.995418350289)*x[1]
        s[0,0,1,0]=(-0.165833461957)*x[0]**o+(0.765518912392)*x[0]+(0.435852402075)*x[1]**o+(0.508949151899)*x[1]
        s[0,0,1,1]=(0.936657616616)*x[0]**o+(-0.502803686461)*x[0]+(0.46372571939)*x[1]**o+(0.49848921833)*x[1]
        s[0,0,2,0]=(-0.637976080243)*x[0]**o+(-0.543248936311)*x[0]+(0.123903782748)*x[1]**o+(-0.775999535262)*x[1]
        s[0,0,2,1]=(-0.644671364311)*x[0]**o+(-0.843862083932)*x[0]+(-0.235872470082)*x[1]**o+(-0.976009591762)*x[1]
        s[0,1,0,0]=(-0.0549568778158)*x[0]**o+(0.585187582572)*x[0]+(0.511901547931)*x[1]**o+(0.214452316763)*x[1]
        s[0,1,0,1]=(0.904417527552)*x[0]**o+(0.120032953669)*x[0]+(-0.946416892695)*x[1]**o+(-0.283712005862)*x[1]
        s[0,1,1,0]=(0.847745098439)*x[0]**o+(-0.275082804508)*x[0]+(0.120958807786)*x[1]**o+(0.118387324002)*x[1]
        s[0,1,1,1]=(-0.504242295046)*x[0]**o+(-0.422074691957)*x[0]+(-0.1482990171)*x[1]**o+(0.0615566868579)*x[1]
        s[0,1,2,0]=(0.874047261815)*x[0]**o+(-0.13709741107)*x[0]+(-0.785816772906)*x[1]**o+(-0.677819998582)*x[1]
        s[0,1,2,1]=(-0.427567619451)*x[0]**o+(-0.27495790934)*x[0]+(0.240860007422)*x[1]**o+(-0.652639707096)*x[1]
        s[0,2,0,0]=(-0.0122530611615)*x[0]**o+(0.933497855191)*x[0]+(0.576340026858)*x[1]**o+(0.36528784459)*x[1]
        s[0,2,0,1]=(-0.536014689424)*x[0]**o+(-0.748239369266)*x[0]+(0.932275411848)*x[1]**o+(-0.610630368247)*x[1]
        s[0,2,1,0]=(-0.949632388534)*x[0]**o+(0.426052596169)*x[0]+(0.388027927797)*x[1]**o+(-0.825568754704)*x[1]
        s[0,2,1,1]=(0.647266576966)*x[0]**o+(0.558728080881)*x[0]+(-0.424360645973)*x[1]**o+(-0.946182937899)*x[1]
        s[0,2,2,0]=(0.00223655708367)*x[0]**o+(0.824278028798)*x[0]+(0.353003022728)*x[1]**o+(0.731718952112)*x[1]
        s[0,2,2,1]=(0.123272031365)*x[0]**o+(-0.51433738104)*x[0]+(-0.773965295951)*x[1]**o+(0.53173719763)*x[1]
        s[0,3,0,0]=(-0.867031991936)*x[0]**o+(-0.242455168132)*x[0]+(-0.0277817911953)*x[1]**o+(0.778617461199)*x[1]
        s[0,3,0,1]=(0.832013516177)*x[0]**o+(0.0919740140778)*x[0]+(-0.285306764452)*x[1]**o+(-0.398556244817)*x[1]
        s[0,3,1,0]=(-0.16075720255)*x[0]**o+(0.321403077687)*x[0]+(0.918586498854)*x[1]**o+(-0.0840017219508)*x[1]
        s[0,3,1,1]=(0.266578383859)*x[0]**o+(-0.255315447308)*x[0]+(-0.418644958222)*x[1]**o+(-0.552092333928)*x[1]
        s[0,3,2,0]=(-0.664300381841)*x[0]**o+(-0.903669098099)*x[0]+(0.734688383764)*x[1]**o+(-0.864082112475)*x[1]
        s[0,3,2,1]=(0.714042909464)*x[0]**o+(0.353570781531)*x[0]+(0.231167603054)*x[1]**o+(-0.478826190976)*x[1]
        s[0,4,0,0]=(-0.406664445927)*x[0]**o+(0.664845255753)*x[0]+(-0.0240782311961)*x[1]**o+(0.0609339070635)*x[1]
        s[0,4,0,1]=(0.579749599828)*x[0]**o+(0.868795370092)*x[0]+(-0.0139046093829)*x[1]**o+(-0.940717402493)*x[1]
        s[0,4,1,0]=(-0.0741351360441)*x[0]**o+(-0.063360022137)*x[0]+(0.964335931753)*x[1]**o+(-0.207493867266)*x[1]
        s[0,4,1,1]=(0.631911437012)*x[0]**o+(-0.554978524864)*x[0]+(0.790289062983)*x[1]**o+(0.432183124276)*x[1]
        s[0,4,2,0]=(0.389406102187)*x[0]**o+(0.249535881142)*x[0]+(0.0410391462706)*x[1]**o+(0.197332868001)*x[1]
        s[0,4,2,1]=(0.857250789566)*x[0]**o+(0.940933403021)*x[0]+(-0.292525450517)*x[1]**o+(-0.954295367343)*x[1]
        s[1,0,0,0]=(-0.928911526693)*x[0]**o+(0.950310526617)*x[0]+(-0.591433972944)*x[1]**o+(-0.353614169594)*x[1]
        s[1,0,0,1]=(-0.014540197205)*x[0]**o+(-0.750613108365)*x[0]+(-0.442388474744)*x[1]**o+(0.544170338222)*x[1]
        s[1,0,1,0]=(-0.313691536279)*x[0]**o+(-0.780985041598)*x[0]+(-0.323526696855)*x[1]**o+(0.96420998889)*x[1]
        s[1,0,1,1]=(0.806169224432)*x[0]**o+(-0.575250530787)*x[0]+(-0.926680195674)*x[1]**o+(0.979234114904)*x[1]
        s[1,0,2,0]=(0.316089233223)*x[0]**o+(-0.751194409767)*x[0]+(-0.224471299082)*x[1]**o+(-0.201937113732)*x[1]
        s[1,0,2,1]=(0.169596743188)*x[0]**o+(0.72246732303)*x[0]+(-0.850004427207)*x[1]**o+(0.178328119542)*x[1]
        s[1,1,0,0]=(0.878699774988)*x[0]**o+(0.665096951747)*x[0]+(0.81121004378)*x[1]**o+(-0.693898989577)*x[1]
        s[1,1,0,1]=(-0.149911779278)*x[0]**o+(-0.0473525862361)*x[0]+(-0.817244390143)*x[1]**o+(0.134983631028)*x[1]
        s[1,1,1,0]=(0.632694242262)*x[0]**o+(0.371991474102)*x[0]+(0.143801607881)*x[1]**o+(-0.539347762682)*x[1]
        s[1,1,1,1]=(-0.098791767857)*x[0]**o+(0.367846917358)*x[0]+(0.307623783461)*x[1]**o+(-0.634129908922)*x[1]
        s[1,1,2,0]=(0.790142711691)*x[0]**o+(0.384060824946)*x[0]+(-0.816682559718)*x[1]**o+(0.215076058109)*x[1]
        s[1,1,2,1]=(-0.0309558649524)*x[0]**o+(0.177277209384)*x[0]+(-0.450608986592)*x[1]**o+(-0.722601031838)*x[1]
        s[1,2,0,0]=(0.678612759518)*x[0]**o+(-0.388005661064)*x[0]+(-0.704573701308)*x[1]**o+(0.228580862805)*x[1]
        s[1,2,0,1]=(-0.982279249068)*x[0]**o+(-0.772496362129)*x[0]+(-0.871863675039)*x[1]**o+(0.924707417628)*x[1]
        s[1,2,1,0]=(-0.60771962999)*x[0]**o+(0.328917128624)*x[0]+(-0.366890766298)*x[1]**o+(0.7822593789)*x[1]
        s[1,2,1,1]=(-0.370452908148)*x[0]**o+(0.60883142941)*x[0]+(0.737450910439)*x[1]**o+(-0.250024874152)*x[1]
        s[1,2,2,0]=(-0.496193128154)*x[0]**o+(0.658353278368)*x[0]+(0.250685386258)*x[1]**o+(-0.0420890429033)*x[1]
        s[1,2,2,1]=(-0.396634397946)*x[0]**o+(0.995854445028)*x[0]+(-0.875299234489)*x[1]**o+(-0.794293605194)*x[1]
        s[1,3,0,0]=(-0.927281481375)*x[0]**o+(-0.747949026933)*x[0]+(-0.338650975226)*x[1]**o+(0.8631277432)*x[1]
        s[1,3,0,1]=(0.0643929398073)*x[0]**o+(-0.795784057297)*x[0]+(-0.77568400328)*x[1]**o+(-0.225054290337)*x[1]
        s[1,3,1,0]=(-0.636490397115)*x[0]**o+(-0.445636556924)*x[0]+(-0.906474178436)*x[1]**o+(0.380552377428)*x[1]
        s[1,3,1,1]=(0.215117725278)*x[0]**o+(-0.291445381983)*x[0]+(-0.789290110829)*x[1]**o+(0.23201632636)*x[1]
        s[1,3,2,0]=(0.26375154986)*x[0]**o+(-0.983679100192)*x[0]+(-0.210171301872)*x[1]**o+(0.773227656215)*x[1]
        s[1,3,2,1]=(0.0839267909242)*x[0]**o+(-0.805367088799)*x[0]+(0.155312230897)*x[1]**o+(-0.0808934605086)*x[1]
        s[1,4,0,0]=(0.199593117795)*x[0]**o+(0.965501409685)*x[0]+(0.323286570062)*x[1]**o+(-0.429707844705)*x[1]
        s[1,4,0,1]=(0.805178613493)*x[0]**o+(-0.853858046784)*x[0]+(-0.353638415278)*x[1]**o+(0.171890721243)*x[1]
        s[1,4,1,0]=(0.644814955839)*x[0]**o+(0.299698778855)*x[0]+(-0.88203132112)*x[1]**o+(-0.26100971325)*x[1]
        s[1,4,1,1]=(0.355086073535)*x[0]**o+(0.141140017392)*x[0]+(-0.676113152929)*x[1]**o+(0.954388216319)*x[1]
        s[1,4,2,0]=(0.766116840139)*x[0]**o+(0.00368283914615)*x[0]+(-0.889711965249)*x[1]**o+(-0.00156916153733)*x[1]
        s[1,4,2,1]=(0.501384592515)*x[0]**o+(-0.863973658641)*x[0]+(0.367327889997)*x[1]**o+(0.913819612899)*x[1]
        s[2,0,0,0]=(0.747990722914)*x[0]**o+(-0.64012379706)*x[0]+(-0.565267688923)*x[1]**o+(-0.462605252331)*x[1]
        s[2,0,0,1]=(-0.68417559629)*x[0]**o+(-0.896081416916)*x[0]+(-0.639839159011)*x[1]**o+(-0.721969231804)*x[1]
        s[2,0,1,0]=(-0.690688995555)*x[0]**o+(0.0759300115056)*x[0]+(-0.27942910767)*x[1]**o+(-0.677278037631)*x[1]
        s[2,0,1,1]=(-0.196814931401)*x[0]**o+(-0.848391602254)*x[0]+(-0.95899520009)*x[1]**o+(-0.321410974451)*x[1]
        s[2,0,2,0]=(0.360610787396)*x[0]**o+(-0.659652161288)*x[0]+(0.769472827641)*x[1]**o+(-0.272229772557)*x[1]
        s[2,0,2,1]=(-0.856112621744)*x[0]**o+(0.818082587577)*x[0]+(0.870122740873)*x[1]**o+(0.26234458071)*x[1]
        s[2,1,0,0]=(-0.820858353117)*x[0]**o+(0.815272881968)*x[0]+(0.479239358759)*x[1]**o+(-0.397781217456)*x[1]
        s[2,1,0,1]=(0.450155189682)*x[0]**o+(0.877090117209)*x[0]+(-0.704715374887)*x[1]**o+(0.839517808158)*x[1]
        s[2,1,1,0]=(0.713505879711)*x[0]**o+(0.610444859873)*x[0]+(0.36123692406)*x[1]**o+(0.0356966035771)*x[1]
        s[2,1,1,1]=(0.585841995568)*x[0]**o+(0.487958000149)*x[0]+(0.0536737460141)*x[1]**o+(0.89251806915)*x[1]
        s[2,1,2,0]=(0.122419410022)*x[0]**o+(-0.454371392861)*x[0]+(-0.508915915608)*x[1]**o+(0.182258261833)*x[1]
        s[2,1,2,1]=(-0.0541029930504)*x[0]**o+(-0.864251132735)*x[0]+(-0.726628983103)*x[1]**o+(-0.783989882996)*x[1]
        s[2,2,0,0]=(0.366784417091)*x[0]**o+(0.650637898135)*x[0]+(0.010811456998)*x[1]**o+(-0.373261079417)*x[1]
        s[2,2,0,1]=(0.147500724262)*x[0]**o+(-0.48265020035)*x[0]+(0.156030753671)*x[1]**o+(0.520269666679)*x[1]
        s[2,2,1,0]=(-0.851491519172)*x[0]**o+(-0.221010369594)*x[0]+(-0.367173559702)*x[1]**o+(0.700206278544)*x[1]
        s[2,2,1,1]=(-0.803178424841)*x[0]**o+(0.914630691651)*x[0]+(0.945631327847)*x[1]**o+(0.838368867894)*x[1]
        s[2,2,2,0]=(0.205807876192)*x[0]**o+(-0.109928074685)*x[0]+(-0.0126997666433)*x[1]**o+(0.0977593300095)*x[1]
        s[2,2,2,1]=(0.1106357035)*x[0]**o+(-0.848494903572)*x[0]+(-0.726652770048)*x[1]**o+(0.286175831783)*x[1]
        s[2,3,0,0]=(0.257704015918)*x[0]**o+(-0.890384711683)*x[0]+(-0.778573777872)*x[1]**o+(0.992701220719)*x[1]
        s[2,3,0,1]=(0.0752025499766)*x[0]**o+(0.0430345988982)*x[0]+(-0.136802110279)*x[1]**o+(-0.172056621279)*x[1]
        s[2,3,1,0]=(-0.0611919320989)*x[0]**o+(-0.785758488707)*x[0]+(0.0422134422585)*x[1]**o+(-0.225292684738)*x[1]
        s[2,3,1,1]=(0.693351228594)*x[0]**o+(0.0471475457897)*x[0]+(0.998048196978)*x[1]**o+(-0.46372201856)*x[1]
        s[2,3,2,0]=(-0.261098659889)*x[0]**o+(-0.531925602534)*x[0]+(-0.0557602486972)*x[1]**o+(-0.727790427141)*x[1]
        s[2,3,2,1]=(0.367102083636)*x[0]**o+(0.753294773683)*x[0]+(-0.408790787069)*x[1]**o+(0.860250168871)*x[1]
        s[2,4,0,0]=(-0.408802974965)*x[0]**o+(0.112012515842)*x[0]+(0.785440176371)*x[1]**o+(-0.437347103411)*x[1]
        s[2,4,0,1]=(-0.129910738114)*x[0]**o+(0.0517376480938)*x[0]+(-0.162147517056)*x[1]**o+(-0.679977272409)*x[1]
        s[2,4,1,0]=(-0.0432480639967)*x[0]**o+(0.952982408035)*x[0]+(0.14096382614)*x[1]**o+(0.155577971783)*x[1]
        s[2,4,1,1]=(-0.64723868539)*x[0]**o+(0.120062216545)*x[0]+(0.66577389931)*x[1]**o+(-0.479348695737)*x[1]
        s[2,4,2,0]=(0.971347133416)*x[0]**o+(-0.985122892501)*x[0]+(-0.646064679447)*x[1]**o+(-0.01550762144)*x[1]
        s[2,4,2,1]=(-0.644236125094)*x[0]**o+(0.735842344335)*x[0]+(0.213176548419)*x[1]**o+(-0.30985200546)*x[1]
        s[3,0,0,0]=(-0.984551843203)*x[0]**o+(0.168671646169)*x[0]+(0.349224064129)*x[1]**o+(-0.58160612166)*x[1]
        s[3,0,0,1]=(-0.812761903179)*x[0]**o+(-0.361937090398)*x[0]+(-0.24686279456)*x[1]**o+(0.710593506458)*x[1]
        s[3,0,1,0]=(-0.49188081867)*x[0]**o+(0.124030127513)*x[0]+(0.939873021882)*x[1]**o+(-0.548040221422)*x[1]
        s[3,0,1,1]=(0.872847662987)*x[0]**o+(-0.510872116798)*x[0]+(0.281463732211)*x[1]**o+(0.334088464398)*x[1]
        s[3,0,2,0]=(0.578448528377)*x[0]**o+(-0.576452752848)*x[0]+(-0.768113122135)*x[1]**o+(0.493525403618)*x[1]
        s[3,0,2,1]=(0.770913406943)*x[0]**o+(-0.420032731108)*x[0]+(-0.72069168838)*x[1]**o+(0.901782864345)*x[1]
        s[3,1,0,0]=(0.691817282198)*x[0]**o+(0.0450167655711)*x[0]+(0.40008707832)*x[1]**o+(0.326190694348)*x[1]
        s[3,1,0,1]=(-0.609859249467)*x[0]**o+(0.410229838383)*x[0]+(-0.267983779193)*x[1]**o+(0.775650294164)*x[1]
        s[3,1,1,0]=(-0.159109069553)*x[0]**o+(0.907250351987)*x[0]+(0.207780057825)*x[1]**o+(-0.342636180806)*x[1]
        s[3,1,1,1]=(-0.617610724309)*x[0]**o+(-0.78996204048)*x[0]+(-0.208417118399)*x[1]**o+(-0.423010342209)*x[1]
        s[3,1,2,0]=(-0.654973001955)*x[0]**o+(-0.731180791646)*x[0]+(-0.204094989969)*x[1]**o+(0.0168285385179)*x[1]
        s[3,1,2,1]=(0.928948644167)*x[0]**o+(0.734089325039)*x[0]+(-0.231953650601)*x[1]**o+(-0.249238293473)*x[1]
        s[3,2,0,0]=(0.157179675881)*x[0]**o+(0.410129310591)*x[0]+(0.298525358008)*x[1]**o+(-0.35693765329)*x[1]
        s[3,2,0,1]=(-0.269476743663)*x[0]**o+(-0.412585477307)*x[0]+(-0.9954766662)*x[1]**o+(0.273715511284)*x[1]
        s[3,2,1,0]=(-0.408532610441)*x[0]**o+(0.0259540763142)*x[0]+(-0.23989441562)*x[1]**o+(0.295094463976)*x[1]
        s[3,2,1,1]=(-0.110244223386)*x[0]**o+(-0.67945468919)*x[0]+(0.923264133425)*x[1]**o+(0.215905691981)*x[1]
        s[3,2,2,0]=(0.834448012095)*x[0]**o+(-0.940738696056)*x[0]+(-0.876206011085)*x[1]**o+(-0.298849473071)*x[1]
        s[3,2,2,1]=(0.475805958188)*x[0]**o+(0.621609329804)*x[0]+(-0.00780130203733)*x[1]**o+(0.888801855773)*x[1]
        s[3,3,0,0]=(-0.57606903705)*x[0]**o+(-0.226190138868)*x[0]+(0.40330410251)*x[1]**o+(-0.609371823463)*x[1]
        s[3,3,0,1]=(-0.0205092983359)*x[0]**o+(0.521468339789)*x[0]+(0.853034535096)*x[1]**o+(0.964338973709)*x[1]
        s[3,3,1,0]=(-0.272825676838)*x[0]**o+(-0.843511724287)*x[0]+(0.188117720498)*x[1]**o+(0.664721656786)*x[1]
        s[3,3,1,1]=(0.504997029501)*x[0]**o+(0.657034903786)*x[0]+(-0.198981636802)*x[1]**o+(-0.610769272074)*x[1]
        s[3,3,2,0]=(-0.118598480477)*x[0]**o+(-0.390266333689)*x[0]+(0.0363960976193)*x[1]**o+(0.0333161022002)*x[1]
        s[3,3,2,1]=(0.0560887527733)*x[0]**o+(-0.0757497077315)*x[0]+(-0.638185285374)*x[1]**o+(-0.0772348567365)*x[1]
        s[3,4,0,0]=(-0.77162241293)*x[0]**o+(-0.638997804963)*x[0]+(-0.103979693129)*x[1]**o+(-0.403273375409)*x[1]
        s[3,4,0,1]=(-0.835975178768)*x[0]**o+(-0.776970415353)*x[0]+(-0.123032313988)*x[1]**o+(-0.402259255094)*x[1]
        s[3,4,1,0]=(0.391513520306)*x[0]**o+(0.751880253565)*x[0]+(-0.59575974702)*x[1]**o+(-0.478354879315)*x[1]
        s[3,4,1,1]=(-0.890188758697)*x[0]**o+(0.635319326122)*x[0]+(0.796942175081)*x[1]**o+(-0.325193620429)*x[1]
        s[3,4,2,0]=(-0.906513454158)*x[0]**o+(0.599411990631)*x[0]+(-0.0273060032343)*x[1]**o+(-0.0720151635364)*x[1]
        s[3,4,2,1]=(-0.00512989795909)*x[0]**o+(-0.622272404387)*x[0]+(0.453748642854)*x[1]**o+(-0.261040385951)*x[1]
        ref[0,0,0,0]=(0.367629818255)*x_ref[0]**o+(-0.682992553528)*x_ref[0]+(-0.958578674565)*x_ref[1]**o+(0.994004366609)*x_ref[1]
        ref[0,0,0,1]=(-0.933871216086)*x_ref[0]**o+(0.089303727901)*x_ref[0]+(0.876179784981)*x_ref[1]**o+(-0.995418350289)*x_ref[1]
        ref[0,0,1,0]=(-0.165833461957)*x_ref[0]**o+(0.765518912392)*x_ref[0]+(0.435852402075)*x_ref[1]**o+(0.508949151899)*x_ref[1]
        ref[0,0,1,1]=(0.936657616616)*x_ref[0]**o+(-0.502803686461)*x_ref[0]+(0.46372571939)*x_ref[1]**o+(0.49848921833)*x_ref[1]
        ref[0,0,2,0]=(-0.637976080243)*x_ref[0]**o+(-0.543248936311)*x_ref[0]+(0.123903782748)*x_ref[1]**o+(-0.775999535262)*x_ref[1]
        ref[0,0,2,1]=(-0.644671364311)*x_ref[0]**o+(-0.843862083932)*x_ref[0]+(-0.235872470082)*x_ref[1]**o+(-0.976009591762)*x_ref[1]
        ref[0,1,0,0]=(-0.0549568778158)*x_ref[0]**o+(0.585187582572)*x_ref[0]+(0.511901547931)*x_ref[1]**o+(0.214452316763)*x_ref[1]
        ref[0,1,0,1]=(0.904417527552)*x_ref[0]**o+(0.120032953669)*x_ref[0]+(-0.946416892695)*x_ref[1]**o+(-0.283712005862)*x_ref[1]
        ref[0,1,1,0]=(0.847745098439)*x_ref[0]**o+(-0.275082804508)*x_ref[0]+(0.120958807786)*x_ref[1]**o+(0.118387324002)*x_ref[1]
        ref[0,1,1,1]=(-0.504242295046)*x_ref[0]**o+(-0.422074691957)*x_ref[0]+(-0.1482990171)*x_ref[1]**o+(0.0615566868579)*x_ref[1]
        ref[0,1,2,0]=(0.874047261815)*x_ref[0]**o+(-0.13709741107)*x_ref[0]+(-0.785816772906)*x_ref[1]**o+(-0.677819998582)*x_ref[1]
        ref[0,1,2,1]=(-0.427567619451)*x_ref[0]**o+(-0.27495790934)*x_ref[0]+(0.240860007422)*x_ref[1]**o+(-0.652639707096)*x_ref[1]
        ref[0,2,0,0]=(-0.0122530611615)*x_ref[0]**o+(0.933497855191)*x_ref[0]+(0.576340026858)*x_ref[1]**o+(0.36528784459)*x_ref[1]
        ref[0,2,0,1]=(-0.536014689424)*x_ref[0]**o+(-0.748239369266)*x_ref[0]+(0.932275411848)*x_ref[1]**o+(-0.610630368247)*x_ref[1]
        ref[0,2,1,0]=(-0.949632388534)*x_ref[0]**o+(0.426052596169)*x_ref[0]+(0.388027927797)*x_ref[1]**o+(-0.825568754704)*x_ref[1]
        ref[0,2,1,1]=(0.647266576966)*x_ref[0]**o+(0.558728080881)*x_ref[0]+(-0.424360645973)*x_ref[1]**o+(-0.946182937899)*x_ref[1]
        ref[0,2,2,0]=(0.00223655708367)*x_ref[0]**o+(0.824278028798)*x_ref[0]+(0.353003022728)*x_ref[1]**o+(0.731718952112)*x_ref[1]
        ref[0,2,2,1]=(0.123272031365)*x_ref[0]**o+(-0.51433738104)*x_ref[0]+(-0.773965295951)*x_ref[1]**o+(0.53173719763)*x_ref[1]
        ref[0,3,0,0]=(-0.867031991936)*x_ref[0]**o+(-0.242455168132)*x_ref[0]+(-0.0277817911953)*x_ref[1]**o+(0.778617461199)*x_ref[1]
        ref[0,3,0,1]=(0.832013516177)*x_ref[0]**o+(0.0919740140778)*x_ref[0]+(-0.285306764452)*x_ref[1]**o+(-0.398556244817)*x_ref[1]
        ref[0,3,1,0]=(-0.16075720255)*x_ref[0]**o+(0.321403077687)*x_ref[0]+(0.918586498854)*x_ref[1]**o+(-0.0840017219508)*x_ref[1]
        ref[0,3,1,1]=(0.266578383859)*x_ref[0]**o+(-0.255315447308)*x_ref[0]+(-0.418644958222)*x_ref[1]**o+(-0.552092333928)*x_ref[1]
        ref[0,3,2,0]=(-0.664300381841)*x_ref[0]**o+(-0.903669098099)*x_ref[0]+(0.734688383764)*x_ref[1]**o+(-0.864082112475)*x_ref[1]
        ref[0,3,2,1]=(0.714042909464)*x_ref[0]**o+(0.353570781531)*x_ref[0]+(0.231167603054)*x_ref[1]**o+(-0.478826190976)*x_ref[1]
        ref[0,4,0,0]=(-0.406664445927)*x_ref[0]**o+(0.664845255753)*x_ref[0]+(-0.0240782311961)*x_ref[1]**o+(0.0609339070635)*x_ref[1]
        ref[0,4,0,1]=(0.579749599828)*x_ref[0]**o+(0.868795370092)*x_ref[0]+(-0.0139046093829)*x_ref[1]**o+(-0.940717402493)*x_ref[1]
        ref[0,4,1,0]=(-0.0741351360441)*x_ref[0]**o+(-0.063360022137)*x_ref[0]+(0.964335931753)*x_ref[1]**o+(-0.207493867266)*x_ref[1]
        ref[0,4,1,1]=(0.631911437012)*x_ref[0]**o+(-0.554978524864)*x_ref[0]+(0.790289062983)*x_ref[1]**o+(0.432183124276)*x_ref[1]
        ref[0,4,2,0]=(0.389406102187)*x_ref[0]**o+(0.249535881142)*x_ref[0]+(0.0410391462706)*x_ref[1]**o+(0.197332868001)*x_ref[1]
        ref[0,4,2,1]=(0.857250789566)*x_ref[0]**o+(0.940933403021)*x_ref[0]+(-0.292525450517)*x_ref[1]**o+(-0.954295367343)*x_ref[1]
        ref[1,0,0,0]=(-0.928911526693)*x_ref[0]**o+(0.950310526617)*x_ref[0]+(-0.591433972944)*x_ref[1]**o+(-0.353614169594)*x_ref[1]
        ref[1,0,0,1]=(-0.014540197205)*x_ref[0]**o+(-0.750613108365)*x_ref[0]+(-0.442388474744)*x_ref[1]**o+(0.544170338222)*x_ref[1]
        ref[1,0,1,0]=(-0.313691536279)*x_ref[0]**o+(-0.780985041598)*x_ref[0]+(-0.323526696855)*x_ref[1]**o+(0.96420998889)*x_ref[1]
        ref[1,0,1,1]=(0.806169224432)*x_ref[0]**o+(-0.575250530787)*x_ref[0]+(-0.926680195674)*x_ref[1]**o+(0.979234114904)*x_ref[1]
        ref[1,0,2,0]=(0.316089233223)*x_ref[0]**o+(-0.751194409767)*x_ref[0]+(-0.224471299082)*x_ref[1]**o+(-0.201937113732)*x_ref[1]
        ref[1,0,2,1]=(0.169596743188)*x_ref[0]**o+(0.72246732303)*x_ref[0]+(-0.850004427207)*x_ref[1]**o+(0.178328119542)*x_ref[1]
        ref[1,1,0,0]=(0.878699774988)*x_ref[0]**o+(0.665096951747)*x_ref[0]+(0.81121004378)*x_ref[1]**o+(-0.693898989577)*x_ref[1]
        ref[1,1,0,1]=(-0.149911779278)*x_ref[0]**o+(-0.0473525862361)*x_ref[0]+(-0.817244390143)*x_ref[1]**o+(0.134983631028)*x_ref[1]
        ref[1,1,1,0]=(0.632694242262)*x_ref[0]**o+(0.371991474102)*x_ref[0]+(0.143801607881)*x_ref[1]**o+(-0.539347762682)*x_ref[1]
        ref[1,1,1,1]=(-0.098791767857)*x_ref[0]**o+(0.367846917358)*x_ref[0]+(0.307623783461)*x_ref[1]**o+(-0.634129908922)*x_ref[1]
        ref[1,1,2,0]=(0.790142711691)*x_ref[0]**o+(0.384060824946)*x_ref[0]+(-0.816682559718)*x_ref[1]**o+(0.215076058109)*x_ref[1]
        ref[1,1,2,1]=(-0.0309558649524)*x_ref[0]**o+(0.177277209384)*x_ref[0]+(-0.450608986592)*x_ref[1]**o+(-0.722601031838)*x_ref[1]
        ref[1,2,0,0]=(0.678612759518)*x_ref[0]**o+(-0.388005661064)*x_ref[0]+(-0.704573701308)*x_ref[1]**o+(0.228580862805)*x_ref[1]
        ref[1,2,0,1]=(-0.982279249068)*x_ref[0]**o+(-0.772496362129)*x_ref[0]+(-0.871863675039)*x_ref[1]**o+(0.924707417628)*x_ref[1]
        ref[1,2,1,0]=(-0.60771962999)*x_ref[0]**o+(0.328917128624)*x_ref[0]+(-0.366890766298)*x_ref[1]**o+(0.7822593789)*x_ref[1]
        ref[1,2,1,1]=(-0.370452908148)*x_ref[0]**o+(0.60883142941)*x_ref[0]+(0.737450910439)*x_ref[1]**o+(-0.250024874152)*x_ref[1]
        ref[1,2,2,0]=(-0.496193128154)*x_ref[0]**o+(0.658353278368)*x_ref[0]+(0.250685386258)*x_ref[1]**o+(-0.0420890429033)*x_ref[1]
        ref[1,2,2,1]=(-0.396634397946)*x_ref[0]**o+(0.995854445028)*x_ref[0]+(-0.875299234489)*x_ref[1]**o+(-0.794293605194)*x_ref[1]
        ref[1,3,0,0]=(-0.927281481375)*x_ref[0]**o+(-0.747949026933)*x_ref[0]+(-0.338650975226)*x_ref[1]**o+(0.8631277432)*x_ref[1]
        ref[1,3,0,1]=(0.0643929398073)*x_ref[0]**o+(-0.795784057297)*x_ref[0]+(-0.77568400328)*x_ref[1]**o+(-0.225054290337)*x_ref[1]
        ref[1,3,1,0]=(-0.636490397115)*x_ref[0]**o+(-0.445636556924)*x_ref[0]+(-0.906474178436)*x_ref[1]**o+(0.380552377428)*x_ref[1]
        ref[1,3,1,1]=(0.215117725278)*x_ref[0]**o+(-0.291445381983)*x_ref[0]+(-0.789290110829)*x_ref[1]**o+(0.23201632636)*x_ref[1]
        ref[1,3,2,0]=(0.26375154986)*x_ref[0]**o+(-0.983679100192)*x_ref[0]+(-0.210171301872)*x_ref[1]**o+(0.773227656215)*x_ref[1]
        ref[1,3,2,1]=(0.0839267909242)*x_ref[0]**o+(-0.805367088799)*x_ref[0]+(0.155312230897)*x_ref[1]**o+(-0.0808934605086)*x_ref[1]
        ref[1,4,0,0]=(0.199593117795)*x_ref[0]**o+(0.965501409685)*x_ref[0]+(0.323286570062)*x_ref[1]**o+(-0.429707844705)*x_ref[1]
        ref[1,4,0,1]=(0.805178613493)*x_ref[0]**o+(-0.853858046784)*x_ref[0]+(-0.353638415278)*x_ref[1]**o+(0.171890721243)*x_ref[1]
        ref[1,4,1,0]=(0.644814955839)*x_ref[0]**o+(0.299698778855)*x_ref[0]+(-0.88203132112)*x_ref[1]**o+(-0.26100971325)*x_ref[1]
        ref[1,4,1,1]=(0.355086073535)*x_ref[0]**o+(0.141140017392)*x_ref[0]+(-0.676113152929)*x_ref[1]**o+(0.954388216319)*x_ref[1]
        ref[1,4,2,0]=(0.766116840139)*x_ref[0]**o+(0.00368283914615)*x_ref[0]+(-0.889711965249)*x_ref[1]**o+(-0.00156916153733)*x_ref[1]
        ref[1,4,2,1]=(0.501384592515)*x_ref[0]**o+(-0.863973658641)*x_ref[0]+(0.367327889997)*x_ref[1]**o+(0.913819612899)*x_ref[1]
        ref[2,0,0,0]=(0.747990722914)*x_ref[0]**o+(-0.64012379706)*x_ref[0]+(-0.565267688923)*x_ref[1]**o+(-0.462605252331)*x_ref[1]
        ref[2,0,0,1]=(-0.68417559629)*x_ref[0]**o+(-0.896081416916)*x_ref[0]+(-0.639839159011)*x_ref[1]**o+(-0.721969231804)*x_ref[1]
        ref[2,0,1,0]=(-0.690688995555)*x_ref[0]**o+(0.0759300115056)*x_ref[0]+(-0.27942910767)*x_ref[1]**o+(-0.677278037631)*x_ref[1]
        ref[2,0,1,1]=(-0.196814931401)*x_ref[0]**o+(-0.848391602254)*x_ref[0]+(-0.95899520009)*x_ref[1]**o+(-0.321410974451)*x_ref[1]
        ref[2,0,2,0]=(0.360610787396)*x_ref[0]**o+(-0.659652161288)*x_ref[0]+(0.769472827641)*x_ref[1]**o+(-0.272229772557)*x_ref[1]
        ref[2,0,2,1]=(-0.856112621744)*x_ref[0]**o+(0.818082587577)*x_ref[0]+(0.870122740873)*x_ref[1]**o+(0.26234458071)*x_ref[1]
        ref[2,1,0,0]=(-0.820858353117)*x_ref[0]**o+(0.815272881968)*x_ref[0]+(0.479239358759)*x_ref[1]**o+(-0.397781217456)*x_ref[1]
        ref[2,1,0,1]=(0.450155189682)*x_ref[0]**o+(0.877090117209)*x_ref[0]+(-0.704715374887)*x_ref[1]**o+(0.839517808158)*x_ref[1]
        ref[2,1,1,0]=(0.713505879711)*x_ref[0]**o+(0.610444859873)*x_ref[0]+(0.36123692406)*x_ref[1]**o+(0.0356966035771)*x_ref[1]
        ref[2,1,1,1]=(0.585841995568)*x_ref[0]**o+(0.487958000149)*x_ref[0]+(0.0536737460141)*x_ref[1]**o+(0.89251806915)*x_ref[1]
        ref[2,1,2,0]=(0.122419410022)*x_ref[0]**o+(-0.454371392861)*x_ref[0]+(-0.508915915608)*x_ref[1]**o+(0.182258261833)*x_ref[1]
        ref[2,1,2,1]=(-0.0541029930504)*x_ref[0]**o+(-0.864251132735)*x_ref[0]+(-0.726628983103)*x_ref[1]**o+(-0.783989882996)*x_ref[1]
        ref[2,2,0,0]=(0.366784417091)*x_ref[0]**o+(0.650637898135)*x_ref[0]+(0.010811456998)*x_ref[1]**o+(-0.373261079417)*x_ref[1]
        ref[2,2,0,1]=(0.147500724262)*x_ref[0]**o+(-0.48265020035)*x_ref[0]+(0.156030753671)*x_ref[1]**o+(0.520269666679)*x_ref[1]
        ref[2,2,1,0]=(-0.851491519172)*x_ref[0]**o+(-0.221010369594)*x_ref[0]+(-0.367173559702)*x_ref[1]**o+(0.700206278544)*x_ref[1]
        ref[2,2,1,1]=(-0.803178424841)*x_ref[0]**o+(0.914630691651)*x_ref[0]+(0.945631327847)*x_ref[1]**o+(0.838368867894)*x_ref[1]
        ref[2,2,2,0]=(0.205807876192)*x_ref[0]**o+(-0.109928074685)*x_ref[0]+(-0.0126997666433)*x_ref[1]**o+(0.0977593300095)*x_ref[1]
        ref[2,2,2,1]=(0.1106357035)*x_ref[0]**o+(-0.848494903572)*x_ref[0]+(-0.726652770048)*x_ref[1]**o+(0.286175831783)*x_ref[1]
        ref[2,3,0,0]=(0.257704015918)*x_ref[0]**o+(-0.890384711683)*x_ref[0]+(-0.778573777872)*x_ref[1]**o+(0.992701220719)*x_ref[1]
        ref[2,3,0,1]=(0.0752025499766)*x_ref[0]**o+(0.0430345988982)*x_ref[0]+(-0.136802110279)*x_ref[1]**o+(-0.172056621279)*x_ref[1]
        ref[2,3,1,0]=(-0.0611919320989)*x_ref[0]**o+(-0.785758488707)*x_ref[0]+(0.0422134422585)*x_ref[1]**o+(-0.225292684738)*x_ref[1]
        ref[2,3,1,1]=(0.693351228594)*x_ref[0]**o+(0.0471475457897)*x_ref[0]+(0.998048196978)*x_ref[1]**o+(-0.46372201856)*x_ref[1]
        ref[2,3,2,0]=(-0.261098659889)*x_ref[0]**o+(-0.531925602534)*x_ref[0]+(-0.0557602486972)*x_ref[1]**o+(-0.727790427141)*x_ref[1]
        ref[2,3,2,1]=(0.367102083636)*x_ref[0]**o+(0.753294773683)*x_ref[0]+(-0.408790787069)*x_ref[1]**o+(0.860250168871)*x_ref[1]
        ref[2,4,0,0]=(-0.408802974965)*x_ref[0]**o+(0.112012515842)*x_ref[0]+(0.785440176371)*x_ref[1]**o+(-0.437347103411)*x_ref[1]
        ref[2,4,0,1]=(-0.129910738114)*x_ref[0]**o+(0.0517376480938)*x_ref[0]+(-0.162147517056)*x_ref[1]**o+(-0.679977272409)*x_ref[1]
        ref[2,4,1,0]=(-0.0432480639967)*x_ref[0]**o+(0.952982408035)*x_ref[0]+(0.14096382614)*x_ref[1]**o+(0.155577971783)*x_ref[1]
        ref[2,4,1,1]=(-0.64723868539)*x_ref[0]**o+(0.120062216545)*x_ref[0]+(0.66577389931)*x_ref[1]**o+(-0.479348695737)*x_ref[1]
        ref[2,4,2,0]=(0.971347133416)*x_ref[0]**o+(-0.985122892501)*x_ref[0]+(-0.646064679447)*x_ref[1]**o+(-0.01550762144)*x_ref[1]
        ref[2,4,2,1]=(-0.644236125094)*x_ref[0]**o+(0.735842344335)*x_ref[0]+(0.213176548419)*x_ref[1]**o+(-0.30985200546)*x_ref[1]
        ref[3,0,0,0]=(-0.984551843203)*x_ref[0]**o+(0.168671646169)*x_ref[0]+(0.349224064129)*x_ref[1]**o+(-0.58160612166)*x_ref[1]
        ref[3,0,0,1]=(-0.812761903179)*x_ref[0]**o+(-0.361937090398)*x_ref[0]+(-0.24686279456)*x_ref[1]**o+(0.710593506458)*x_ref[1]
        ref[3,0,1,0]=(-0.49188081867)*x_ref[0]**o+(0.124030127513)*x_ref[0]+(0.939873021882)*x_ref[1]**o+(-0.548040221422)*x_ref[1]
        ref[3,0,1,1]=(0.872847662987)*x_ref[0]**o+(-0.510872116798)*x_ref[0]+(0.281463732211)*x_ref[1]**o+(0.334088464398)*x_ref[1]
        ref[3,0,2,0]=(0.578448528377)*x_ref[0]**o+(-0.576452752848)*x_ref[0]+(-0.768113122135)*x_ref[1]**o+(0.493525403618)*x_ref[1]
        ref[3,0,2,1]=(0.770913406943)*x_ref[0]**o+(-0.420032731108)*x_ref[0]+(-0.72069168838)*x_ref[1]**o+(0.901782864345)*x_ref[1]
        ref[3,1,0,0]=(0.691817282198)*x_ref[0]**o+(0.0450167655711)*x_ref[0]+(0.40008707832)*x_ref[1]**o+(0.326190694348)*x_ref[1]
        ref[3,1,0,1]=(-0.609859249467)*x_ref[0]**o+(0.410229838383)*x_ref[0]+(-0.267983779193)*x_ref[1]**o+(0.775650294164)*x_ref[1]
        ref[3,1,1,0]=(-0.159109069553)*x_ref[0]**o+(0.907250351987)*x_ref[0]+(0.207780057825)*x_ref[1]**o+(-0.342636180806)*x_ref[1]
        ref[3,1,1,1]=(-0.617610724309)*x_ref[0]**o+(-0.78996204048)*x_ref[0]+(-0.208417118399)*x_ref[1]**o+(-0.423010342209)*x_ref[1]
        ref[3,1,2,0]=(-0.654973001955)*x_ref[0]**o+(-0.731180791646)*x_ref[0]+(-0.204094989969)*x_ref[1]**o+(0.0168285385179)*x_ref[1]
        ref[3,1,2,1]=(0.928948644167)*x_ref[0]**o+(0.734089325039)*x_ref[0]+(-0.231953650601)*x_ref[1]**o+(-0.249238293473)*x_ref[1]
        ref[3,2,0,0]=(0.157179675881)*x_ref[0]**o+(0.410129310591)*x_ref[0]+(0.298525358008)*x_ref[1]**o+(-0.35693765329)*x_ref[1]
        ref[3,2,0,1]=(-0.269476743663)*x_ref[0]**o+(-0.412585477307)*x_ref[0]+(-0.9954766662)*x_ref[1]**o+(0.273715511284)*x_ref[1]
        ref[3,2,1,0]=(-0.408532610441)*x_ref[0]**o+(0.0259540763142)*x_ref[0]+(-0.23989441562)*x_ref[1]**o+(0.295094463976)*x_ref[1]
        ref[3,2,1,1]=(-0.110244223386)*x_ref[0]**o+(-0.67945468919)*x_ref[0]+(0.923264133425)*x_ref[1]**o+(0.215905691981)*x_ref[1]
        ref[3,2,2,0]=(0.834448012095)*x_ref[0]**o+(-0.940738696056)*x_ref[0]+(-0.876206011085)*x_ref[1]**o+(-0.298849473071)*x_ref[1]
        ref[3,2,2,1]=(0.475805958188)*x_ref[0]**o+(0.621609329804)*x_ref[0]+(-0.00780130203733)*x_ref[1]**o+(0.888801855773)*x_ref[1]
        ref[3,3,0,0]=(-0.57606903705)*x_ref[0]**o+(-0.226190138868)*x_ref[0]+(0.40330410251)*x_ref[1]**o+(-0.609371823463)*x_ref[1]
        ref[3,3,0,1]=(-0.0205092983359)*x_ref[0]**o+(0.521468339789)*x_ref[0]+(0.853034535096)*x_ref[1]**o+(0.964338973709)*x_ref[1]
        ref[3,3,1,0]=(-0.272825676838)*x_ref[0]**o+(-0.843511724287)*x_ref[0]+(0.188117720498)*x_ref[1]**o+(0.664721656786)*x_ref[1]
        ref[3,3,1,1]=(0.504997029501)*x_ref[0]**o+(0.657034903786)*x_ref[0]+(-0.198981636802)*x_ref[1]**o+(-0.610769272074)*x_ref[1]
        ref[3,3,2,0]=(-0.118598480477)*x_ref[0]**o+(-0.390266333689)*x_ref[0]+(0.0363960976193)*x_ref[1]**o+(0.0333161022002)*x_ref[1]
        ref[3,3,2,1]=(0.0560887527733)*x_ref[0]**o+(-0.0757497077315)*x_ref[0]+(-0.638185285374)*x_ref[1]**o+(-0.0772348567365)*x_ref[1]
        ref[3,4,0,0]=(-0.77162241293)*x_ref[0]**o+(-0.638997804963)*x_ref[0]+(-0.103979693129)*x_ref[1]**o+(-0.403273375409)*x_ref[1]
        ref[3,4,0,1]=(-0.835975178768)*x_ref[0]**o+(-0.776970415353)*x_ref[0]+(-0.123032313988)*x_ref[1]**o+(-0.402259255094)*x_ref[1]
        ref[3,4,1,0]=(0.391513520306)*x_ref[0]**o+(0.751880253565)*x_ref[0]+(-0.59575974702)*x_ref[1]**o+(-0.478354879315)*x_ref[1]
        ref[3,4,1,1]=(-0.890188758697)*x_ref[0]**o+(0.635319326122)*x_ref[0]+(0.796942175081)*x_ref[1]**o+(-0.325193620429)*x_ref[1]
        ref[3,4,2,0]=(-0.906513454158)*x_ref[0]**o+(0.599411990631)*x_ref[0]+(-0.0273060032343)*x_ref[1]**o+(-0.0720151635364)*x_ref[1]
        ref[3,4,2,1]=(-0.00512989795909)*x_ref[0]**o+(-0.622272404387)*x_ref[0]+(0.453748642854)*x_ref[1]**o+(-0.261040385951)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.173295097983)*x[0]**o+(0.773632210562)*x[0]+(-0.521496418821)*x[1]**o+(-0.947907473579)*x[1]+(0.847721381317)*x[2]**o+(-0.887851963083)*x[2]
        s[0,0,0,1]=(0.0654213598408)*x[0]**o+(0.763489939538)*x[0]+(-0.176800713154)*x[1]**o+(-0.602428137248)*x[1]+(-0.851802251531)*x[2]**o+(-0.763084376785)*x[2]
        s[0,0,1,0]=(0.30505329426)*x[0]**o+(-0.378878190786)*x[0]+(-0.615554910302)*x[1]**o+(0.0555314463878)*x[1]+(-0.063211109707)*x[2]**o+(-0.623580981446)*x[2]
        s[0,0,1,1]=(-0.526220589068)*x[0]**o+(-0.551655771219)*x[0]+(0.163011887243)*x[1]**o+(-0.0863220304472)*x[1]+(-0.638708196977)*x[2]**o+(-0.298073090554)*x[2]
        s[0,0,2,0]=(0.0698028632972)*x[0]**o+(0.508381168266)*x[0]+(-0.120587433211)*x[1]**o+(-0.419454430666)*x[1]+(0.389702566269)*x[2]**o+(-0.169655820731)*x[2]
        s[0,0,2,1]=(-0.80612077313)*x[0]**o+(0.640104753466)*x[0]+(0.518635274506)*x[1]**o+(-0.545279092505)*x[1]+(0.195002020672)*x[2]**o+(0.563537057084)*x[2]
        s[0,1,0,0]=(-0.613471743331)*x[0]**o+(-0.274063857866)*x[0]+(-0.259665265722)*x[1]**o+(0.0830136895696)*x[1]+(0.825880322476)*x[2]**o+(0.133039077085)*x[2]
        s[0,1,0,1]=(0.727956094813)*x[0]**o+(0.605313602734)*x[0]+(0.708549587523)*x[1]**o+(-0.964527272447)*x[1]+(-0.740107990066)*x[2]**o+(0.629263961477)*x[2]
        s[0,1,1,0]=(-0.70107143852)*x[0]**o+(-0.055344456929)*x[0]+(-0.954082414306)*x[1]**o+(-0.702136485683)*x[1]+(-0.801391066649)*x[2]**o+(-0.905767030094)*x[2]
        s[0,1,1,1]=(0.67172040238)*x[0]**o+(0.740486014051)*x[0]+(-0.335682691861)*x[1]**o+(0.0793015412254)*x[1]+(-0.868582520395)*x[2]**o+(0.0848453101016)*x[2]
        s[0,1,2,0]=(0.224190760234)*x[0]**o+(0.139996595496)*x[0]+(-0.519770806543)*x[1]**o+(-0.683808752006)*x[1]+(-0.837082787693)*x[2]**o+(-0.613900014511)*x[2]
        s[0,1,2,1]=(0.213979011641)*x[0]**o+(-0.309663935547)*x[0]+(0.0211525822739)*x[1]**o+(0.844506291028)*x[1]+(-0.688933639701)*x[2]**o+(-0.336614773412)*x[2]
        s[0,2,0,0]=(-0.0188626152353)*x[0]**o+(-0.694756450245)*x[0]+(0.960579615684)*x[1]**o+(0.57325819325)*x[1]+(0.917037597322)*x[2]**o+(-0.0584252009816)*x[2]
        s[0,2,0,1]=(0.803966493511)*x[0]**o+(-0.169885689634)*x[0]+(-0.556021959109)*x[1]**o+(0.273901944163)*x[1]+(-0.737708052821)*x[2]**o+(0.627870713104)*x[2]
        s[0,2,1,0]=(-0.511561329117)*x[0]**o+(-0.0854308669951)*x[0]+(0.464246331007)*x[1]**o+(-0.97461601243)*x[1]+(-0.223895747162)*x[2]**o+(-0.385389479659)*x[2]
        s[0,2,1,1]=(-0.723508721341)*x[0]**o+(-0.201904055697)*x[0]+(0.0789759340837)*x[1]**o+(-0.00248269457662)*x[1]+(-0.0626643900787)*x[2]**o+(-0.128428965973)*x[2]
        s[0,2,2,0]=(0.934647998503)*x[0]**o+(-0.0922965345918)*x[0]+(0.717876178289)*x[1]**o+(0.344875905273)*x[1]+(0.0410512922166)*x[2]**o+(0.164242310844)*x[2]
        s[0,2,2,1]=(-0.699306080605)*x[0]**o+(0.209729161919)*x[0]+(0.820435958574)*x[1]**o+(-0.173302359677)*x[1]+(0.659952656349)*x[2]**o+(-0.78057378087)*x[2]
        s[0,3,0,0]=(-0.495298694962)*x[0]**o+(0.571367877972)*x[0]+(0.931242429977)*x[1]**o+(0.888499081573)*x[1]+(-0.408432694783)*x[2]**o+(-0.898594385103)*x[2]
        s[0,3,0,1]=(-0.96889210815)*x[0]**o+(0.404624542424)*x[0]+(0.380170350345)*x[1]**o+(-0.557523650891)*x[1]+(-0.566134006)*x[2]**o+(0.74677094551)*x[2]
        s[0,3,1,0]=(-0.224311775759)*x[0]**o+(-0.660099304736)*x[0]+(-0.513720022847)*x[1]**o+(0.534948126324)*x[1]+(0.162072633001)*x[2]**o+(-0.402416011787)*x[2]
        s[0,3,1,1]=(0.224350718771)*x[0]**o+(0.283343383933)*x[0]+(0.138291871569)*x[1]**o+(0.543289554343)*x[1]+(-0.844288550977)*x[2]**o+(0.463476015453)*x[2]
        s[0,3,2,0]=(-0.685488143712)*x[0]**o+(0.300413146898)*x[0]+(0.748708379841)*x[1]**o+(-0.598212294807)*x[1]+(0.479982541942)*x[2]**o+(0.23866536682)*x[2]
        s[0,3,2,1]=(0.940674112182)*x[0]**o+(-0.35846015272)*x[0]+(-0.770304297672)*x[1]**o+(-0.73897352885)*x[1]+(-0.682102723348)*x[2]**o+(0.742528100693)*x[2]
        s[0,4,0,0]=(0.0162066911802)*x[0]**o+(0.160336749821)*x[0]+(0.121948909833)*x[1]**o+(-0.102093937457)*x[1]+(-0.768661339258)*x[2]**o+(0.98805683085)*x[2]
        s[0,4,0,1]=(-0.420990735691)*x[0]**o+(-0.500244152083)*x[0]+(0.910641832988)*x[1]**o+(0.979658885966)*x[1]+(-0.946590371476)*x[2]**o+(-0.190997339007)*x[2]
        s[0,4,1,0]=(-0.445779943656)*x[0]**o+(-0.378857349604)*x[0]+(0.125534672956)*x[1]**o+(0.124248627713)*x[1]+(-0.685767125932)*x[2]**o+(0.643604950542)*x[2]
        s[0,4,1,1]=(0.518152834346)*x[0]**o+(-0.891873054984)*x[0]+(-0.648744206994)*x[1]**o+(0.207049340927)*x[1]+(0.198529322067)*x[2]**o+(-0.814832030848)*x[2]
        s[0,4,2,0]=(0.477174707789)*x[0]**o+(-0.724452116357)*x[0]+(0.209008883429)*x[1]**o+(0.79968814697)*x[1]+(0.129144596117)*x[2]**o+(0.214098340906)*x[2]
        s[0,4,2,1]=(0.117050329845)*x[0]**o+(-0.721574040514)*x[0]+(0.382424906743)*x[1]**o+(-0.135392966957)*x[1]+(0.24455410794)*x[2]**o+(-0.679165459966)*x[2]
        s[1,0,0,0]=(-0.180925772343)*x[0]**o+(0.664151970586)*x[0]+(0.0496840383663)*x[1]**o+(-0.899468244962)*x[1]+(0.942251196146)*x[2]**o+(0.228837829842)*x[2]
        s[1,0,0,1]=(0.15346023848)*x[0]**o+(-0.642882884964)*x[0]+(-0.462925865669)*x[1]**o+(0.738240095188)*x[1]+(-0.57213723904)*x[2]**o+(-0.900335371347)*x[2]
        s[1,0,1,0]=(-0.366392792085)*x[0]**o+(-0.227326930302)*x[0]+(-0.323111857481)*x[1]**o+(0.220826082736)*x[1]+(0.237722021225)*x[2]**o+(0.835402283453)*x[2]
        s[1,0,1,1]=(-0.293188444394)*x[0]**o+(0.338812067406)*x[0]+(0.679884979413)*x[1]**o+(-0.614317813911)*x[1]+(0.936906665544)*x[2]**o+(-0.652679743405)*x[2]
        s[1,0,2,0]=(0.617621972453)*x[0]**o+(-0.936326816619)*x[0]+(0.760134028716)*x[1]**o+(-0.809380984591)*x[1]+(0.562612935213)*x[2]**o+(0.0794516002575)*x[2]
        s[1,0,2,1]=(0.448514601948)*x[0]**o+(0.327970604314)*x[0]+(-0.936492917037)*x[1]**o+(-0.425126227791)*x[1]+(-0.372116400796)*x[2]**o+(-0.958054327266)*x[2]
        s[1,1,0,0]=(-0.850336943937)*x[0]**o+(0.379951917933)*x[0]+(0.693148856845)*x[1]**o+(-0.691233930311)*x[1]+(0.548624260329)*x[2]**o+(-0.193125714718)*x[2]
        s[1,1,0,1]=(-0.823007972929)*x[0]**o+(0.408191529362)*x[0]+(-0.023921849403)*x[1]**o+(0.79830304115)*x[1]+(0.506756293069)*x[2]**o+(0.434230444707)*x[2]
        s[1,1,1,0]=(0.263932235445)*x[0]**o+(-0.600679211293)*x[0]+(0.480835197573)*x[1]**o+(-0.976282735826)*x[1]+(0.714583200236)*x[2]**o+(0.942610909813)*x[2]
        s[1,1,1,1]=(0.481454842664)*x[0]**o+(-0.508309985427)*x[0]+(-0.739763965139)*x[1]**o+(0.211871147037)*x[1]+(-0.848812488744)*x[2]**o+(0.885973013107)*x[2]
        s[1,1,2,0]=(-0.796437949787)*x[0]**o+(-0.211960906694)*x[0]+(0.823389343923)*x[1]**o+(-0.0162189137972)*x[1]+(0.77124757744)*x[2]**o+(0.877807214589)*x[2]
        s[1,1,2,1]=(0.899446895743)*x[0]**o+(-0.303223847939)*x[0]+(0.409199240049)*x[1]**o+(0.430938116282)*x[1]+(0.189267561328)*x[2]**o+(0.582418236586)*x[2]
        s[1,2,0,0]=(0.687732048304)*x[0]**o+(0.338825649717)*x[0]+(-0.535870442333)*x[1]**o+(0.507773000137)*x[1]+(-0.892023180429)*x[2]**o+(0.35147003979)*x[2]
        s[1,2,0,1]=(-0.315183497372)*x[0]**o+(-0.0302851860468)*x[0]+(-0.966617816351)*x[1]**o+(-0.876528463134)*x[1]+(-0.682249112679)*x[2]**o+(-0.830991774314)*x[2]
        s[1,2,1,0]=(0.90817845095)*x[0]**o+(-0.629542958154)*x[0]+(0.663791954021)*x[1]**o+(0.584246811118)*x[1]+(0.421582877787)*x[2]**o+(0.316731433829)*x[2]
        s[1,2,1,1]=(-0.948554425084)*x[0]**o+(-0.365749706899)*x[0]+(0.282031925475)*x[1]**o+(0.467581708634)*x[1]+(0.158633292573)*x[2]**o+(-0.0990441468359)*x[2]
        s[1,2,2,0]=(0.113170413925)*x[0]**o+(0.481722903666)*x[0]+(0.0127374268921)*x[1]**o+(-0.778057043706)*x[1]+(0.399544836507)*x[2]**o+(-0.157225042768)*x[2]
        s[1,2,2,1]=(0.320987735687)*x[0]**o+(-0.29324177158)*x[0]+(0.02800800296)*x[1]**o+(-0.277948207921)*x[1]+(-0.271808717111)*x[2]**o+(-0.0379931737995)*x[2]
        s[1,3,0,0]=(0.09942034907)*x[0]**o+(0.939629344394)*x[0]+(-0.810686196185)*x[1]**o+(-0.538230260705)*x[1]+(0.85351760208)*x[2]**o+(-0.737196688288)*x[2]
        s[1,3,0,1]=(0.48690360509)*x[0]**o+(-0.065370723503)*x[0]+(-0.0566373715183)*x[1]**o+(0.871878739279)*x[1]+(0.272054032769)*x[2]**o+(0.540626473697)*x[2]
        s[1,3,1,0]=(0.472815976006)*x[0]**o+(-0.775416335945)*x[0]+(0.213646821631)*x[1]**o+(0.415546582713)*x[1]+(-0.196674759494)*x[2]**o+(0.235475790794)*x[2]
        s[1,3,1,1]=(-0.699232847611)*x[0]**o+(0.504961844211)*x[0]+(-0.0115303749449)*x[1]**o+(-0.756286148171)*x[1]+(-0.0922447811444)*x[2]**o+(-0.330281180023)*x[2]
        s[1,3,2,0]=(0.149323550569)*x[0]**o+(0.977735452052)*x[0]+(0.255972870292)*x[1]**o+(-0.855635543118)*x[1]+(0.519602177962)*x[2]**o+(0.922189470945)*x[2]
        s[1,3,2,1]=(-0.190749790276)*x[0]**o+(-0.351506874986)*x[0]+(-0.0672783552454)*x[1]**o+(-0.223773986529)*x[1]+(-0.593750649495)*x[2]**o+(0.158607153203)*x[2]
        s[1,4,0,0]=(0.245306614552)*x[0]**o+(-0.191799968697)*x[0]+(0.85455961157)*x[1]**o+(0.597426986918)*x[1]+(-0.824363631294)*x[2]**o+(-0.0161601323913)*x[2]
        s[1,4,0,1]=(0.925974209919)*x[0]**o+(-0.606855935989)*x[0]+(0.182619376779)*x[1]**o+(0.82202480925)*x[1]+(0.0511805106413)*x[2]**o+(0.0866309784619)*x[2]
        s[1,4,1,0]=(0.272918227857)*x[0]**o+(-0.406306617661)*x[0]+(-0.0535134980185)*x[1]**o+(-0.186200126868)*x[1]+(0.92546875085)*x[2]**o+(-0.474811198949)*x[2]
        s[1,4,1,1]=(0.803708551531)*x[0]**o+(-0.715626841623)*x[0]+(-0.0822999615139)*x[1]**o+(0.612334309398)*x[1]+(-0.157577434512)*x[2]**o+(-0.795469427996)*x[2]
        s[1,4,2,0]=(-0.837159054411)*x[0]**o+(-0.234523360065)*x[0]+(-0.198337862205)*x[1]**o+(0.897958649268)*x[1]+(-0.519907680436)*x[2]**o+(0.742526415023)*x[2]
        s[1,4,2,1]=(-0.792448458791)*x[0]**o+(0.456487739729)*x[0]+(0.0128609741891)*x[1]**o+(-0.344964414689)*x[1]+(-0.205584911868)*x[2]**o+(0.429183240889)*x[2]
        s[2,0,0,0]=(0.401501024951)*x[0]**o+(-0.497039379761)*x[0]+(-0.228675343279)*x[1]**o+(0.133323657822)*x[1]+(-0.203006510606)*x[2]**o+(0.407718693225)*x[2]
        s[2,0,0,1]=(-0.547448204328)*x[0]**o+(-0.109789709109)*x[0]+(-0.626633454424)*x[1]**o+(-0.117871748454)*x[1]+(0.961790647477)*x[2]**o+(0.652883936373)*x[2]
        s[2,0,1,0]=(-0.85536471584)*x[0]**o+(0.489291116489)*x[0]+(-0.929596786116)*x[1]**o+(0.0110017367519)*x[1]+(-0.703617460934)*x[2]**o+(0.0496869388223)*x[2]
        s[2,0,1,1]=(-0.501093953991)*x[0]**o+(-0.722768363488)*x[0]+(0.84017017736)*x[1]**o+(-0.22900603012)*x[1]+(0.760492299454)*x[2]**o+(0.767346260882)*x[2]
        s[2,0,2,0]=(-0.328498451765)*x[0]**o+(-0.387534616444)*x[0]+(-0.0973382392242)*x[1]**o+(0.418582689709)*x[1]+(0.206439339821)*x[2]**o+(-0.706353598324)*x[2]
        s[2,0,2,1]=(0.179977568909)*x[0]**o+(0.642598753874)*x[0]+(0.711020603016)*x[1]**o+(-0.287193672421)*x[1]+(-0.362612184622)*x[2]**o+(-0.18146063868)*x[2]
        s[2,1,0,0]=(0.0470678392348)*x[0]**o+(0.00649068845495)*x[0]+(-0.239912084684)*x[1]**o+(0.278051515796)*x[1]+(-0.98924625944)*x[2]**o+(0.847957009386)*x[2]
        s[2,1,0,1]=(-0.287954694237)*x[0]**o+(0.127947128577)*x[0]+(0.443869903193)*x[1]**o+(-0.344694028625)*x[1]+(-0.14884506191)*x[2]**o+(0.748140862905)*x[2]
        s[2,1,1,0]=(-0.597041891422)*x[0]**o+(0.914716550666)*x[0]+(0.68496200279)*x[1]**o+(0.0420126536598)*x[1]+(-0.539966353171)*x[2]**o+(-0.886373035297)*x[2]
        s[2,1,1,1]=(0.462776858955)*x[0]**o+(-0.989367011955)*x[0]+(0.421691081662)*x[1]**o+(0.264858346124)*x[1]+(0.513954688709)*x[2]**o+(0.00863898872886)*x[2]
        s[2,1,2,0]=(0.503601863243)*x[0]**o+(0.505973913611)*x[0]+(0.0952160099415)*x[1]**o+(-0.77103310327)*x[1]+(-0.316507531754)*x[2]**o+(0.506969348695)*x[2]
        s[2,1,2,1]=(0.521634307558)*x[0]**o+(-0.910583030486)*x[0]+(0.977891014192)*x[1]**o+(0.985145579284)*x[1]+(0.0465919834269)*x[2]**o+(0.113946013883)*x[2]
        s[2,2,0,0]=(0.178722112454)*x[0]**o+(0.582800628485)*x[0]+(0.141950810923)*x[1]**o+(0.226899624421)*x[1]+(0.445681888363)*x[2]**o+(-0.127689044751)*x[2]
        s[2,2,0,1]=(-0.433214095841)*x[0]**o+(-0.61199374178)*x[0]+(0.655706389409)*x[1]**o+(-0.0160036542096)*x[1]+(0.570017878158)*x[2]**o+(-0.00718475090704)*x[2]
        s[2,2,1,0]=(-0.205396523704)*x[0]**o+(0.87306641582)*x[0]+(-0.664912090364)*x[1]**o+(-0.324577426693)*x[1]+(-0.37269275568)*x[2]**o+(-0.0753237226649)*x[2]
        s[2,2,1,1]=(0.716795494369)*x[0]**o+(-0.725543593985)*x[0]+(0.751101352157)*x[1]**o+(0.0629757438809)*x[1]+(-0.388859722245)*x[2]**o+(0.633579629088)*x[2]
        s[2,2,2,0]=(-0.785119558914)*x[0]**o+(-0.289212761766)*x[0]+(-0.947517483982)*x[1]**o+(-0.898908873087)*x[1]+(-0.758807720992)*x[2]**o+(-0.747003629416)*x[2]
        s[2,2,2,1]=(0.384610894358)*x[0]**o+(0.0170687884025)*x[0]+(0.776211872471)*x[1]**o+(-0.313821704113)*x[1]+(-0.433021717398)*x[2]**o+(-0.972385486304)*x[2]
        s[2,3,0,0]=(0.59307540612)*x[0]**o+(0.93544260405)*x[0]+(0.519016668994)*x[1]**o+(-0.175290607112)*x[1]+(-0.269720511811)*x[2]**o+(0.117237249214)*x[2]
        s[2,3,0,1]=(-0.0752366310333)*x[0]**o+(-0.168792429308)*x[0]+(0.316236130047)*x[1]**o+(0.021991183167)*x[1]+(-0.2024837663)*x[2]**o+(-0.676832860181)*x[2]
        s[2,3,1,0]=(-0.2682564416)*x[0]**o+(0.276503108458)*x[0]+(0.789470125022)*x[1]**o+(-0.339044117352)*x[1]+(-0.408507243588)*x[2]**o+(0.0666267661149)*x[2]
        s[2,3,1,1]=(-0.053395293244)*x[0]**o+(0.425880490109)*x[0]+(-0.068128758144)*x[1]**o+(0.413635486374)*x[1]+(0.965050652205)*x[2]**o+(0.126502830836)*x[2]
        s[2,3,2,0]=(0.601928913545)*x[0]**o+(-0.0443247453956)*x[0]+(-0.606811664164)*x[1]**o+(0.0206686600296)*x[1]+(-0.804806810474)*x[2]**o+(-0.019745599298)*x[2]
        s[2,3,2,1]=(0.911324514835)*x[0]**o+(-0.397270365997)*x[0]+(0.511154806396)*x[1]**o+(0.722003975461)*x[1]+(-0.332922355553)*x[2]**o+(-0.132309188597)*x[2]
        s[2,4,0,0]=(0.194698324972)*x[0]**o+(-0.496022175033)*x[0]+(0.804006600349)*x[1]**o+(-0.643493478792)*x[1]+(0.148395455552)*x[2]**o+(-0.408862596804)*x[2]
        s[2,4,0,1]=(0.734891277828)*x[0]**o+(-0.101225311176)*x[0]+(-0.813960759076)*x[1]**o+(0.862888786761)*x[1]+(0.830408706284)*x[2]**o+(0.523449046413)*x[2]
        s[2,4,1,0]=(0.286629494089)*x[0]**o+(0.599214589464)*x[0]+(-0.208652251419)*x[1]**o+(-0.830028802843)*x[1]+(0.805360345419)*x[2]**o+(-0.865349771687)*x[2]
        s[2,4,1,1]=(0.0367955096711)*x[0]**o+(-0.0261723558157)*x[0]+(0.200900558315)*x[1]**o+(-0.532346409176)*x[1]+(-0.183663881543)*x[2]**o+(0.93191706611)*x[2]
        s[2,4,2,0]=(-0.932223437809)*x[0]**o+(-0.155278731606)*x[0]+(-0.932519381895)*x[1]**o+(0.0589900357358)*x[1]+(-0.187551917707)*x[2]**o+(-0.81649531847)*x[2]
        s[2,4,2,1]=(0.555481188864)*x[0]**o+(0.946358167786)*x[0]+(-0.231784522)*x[1]**o+(0.592776954307)*x[1]+(-0.0373202061188)*x[2]**o+(0.0302770826388)*x[2]
        s[3,0,0,0]=(0.603747717227)*x[0]**o+(-0.969329597018)*x[0]+(0.311876345528)*x[1]**o+(-0.589486433983)*x[1]+(-0.294507657675)*x[2]**o+(-0.743511183116)*x[2]
        s[3,0,0,1]=(-0.846568198319)*x[0]**o+(0.93521569126)*x[0]+(-0.874618479052)*x[1]**o+(0.502042421273)*x[1]+(-0.945100352471)*x[2]**o+(0.765854746295)*x[2]
        s[3,0,1,0]=(-0.908353314852)*x[0]**o+(0.701990826648)*x[0]+(0.798549198275)*x[1]**o+(-0.616631523859)*x[1]+(-0.999800863738)*x[2]**o+(-0.156585259768)*x[2]
        s[3,0,1,1]=(-0.433258077523)*x[0]**o+(-0.641049552618)*x[0]+(0.427903347848)*x[1]**o+(-0.175290793023)*x[1]+(-0.557154490598)*x[2]**o+(-0.00495777825998)*x[2]
        s[3,0,2,0]=(0.998253378112)*x[0]**o+(0.2484283851)*x[0]+(0.955897183936)*x[1]**o+(-0.0351483480831)*x[1]+(0.690211608431)*x[2]**o+(0.7420822406)*x[2]
        s[3,0,2,1]=(0.273153627728)*x[0]**o+(0.931572979903)*x[0]+(-0.279174117029)*x[1]**o+(-0.741239387871)*x[1]+(0.148372377955)*x[2]**o+(0.781476364189)*x[2]
        s[3,1,0,0]=(0.231049869286)*x[0]**o+(-0.871461717509)*x[0]+(0.553545032585)*x[1]**o+(0.979019482224)*x[1]+(0.819288655429)*x[2]**o+(-0.01489981549)*x[2]
        s[3,1,0,1]=(-0.891391000148)*x[0]**o+(-0.360048424965)*x[0]+(-0.390043258734)*x[1]**o+(0.547316173383)*x[1]+(-0.972094018922)*x[2]**o+(0.7798960782)*x[2]
        s[3,1,1,0]=(0.675405837097)*x[0]**o+(0.976689341648)*x[0]+(-0.416251051342)*x[1]**o+(0.696304358663)*x[1]+(-0.395743544878)*x[2]**o+(-0.0340340734736)*x[2]
        s[3,1,1,1]=(-0.531871504573)*x[0]**o+(-0.926770430235)*x[0]+(0.752552585571)*x[1]**o+(-0.403402334727)*x[1]+(0.07593559269)*x[2]**o+(0.126742096076)*x[2]
        s[3,1,2,0]=(-0.033528429617)*x[0]**o+(0.187766565831)*x[0]+(0.192128207394)*x[1]**o+(-0.0969373324914)*x[1]+(-0.889231964389)*x[2]**o+(0.493198117275)*x[2]
        s[3,1,2,1]=(-0.182267751454)*x[0]**o+(-0.115192609603)*x[0]+(-0.252701809463)*x[1]**o+(0.355949096242)*x[1]+(0.859478772116)*x[2]**o+(0.539090466518)*x[2]
        s[3,2,0,0]=(0.25232058799)*x[0]**o+(-0.246680227462)*x[0]+(0.316105270194)*x[1]**o+(-0.426505186439)*x[1]+(0.980259506048)*x[2]**o+(0.144514677615)*x[2]
        s[3,2,0,1]=(-0.583071409136)*x[0]**o+(0.969922333041)*x[0]+(-0.0251030152687)*x[1]**o+(0.121967009007)*x[1]+(0.391834007838)*x[2]**o+(0.0339488845613)*x[2]
        s[3,2,1,0]=(0.706904708517)*x[0]**o+(0.266141083673)*x[0]+(-0.0339160692688)*x[1]**o+(-0.152115678673)*x[1]+(0.122213919626)*x[2]**o+(-0.576478442144)*x[2]
        s[3,2,1,1]=(-0.917141077151)*x[0]**o+(0.569284998999)*x[0]+(-0.0979622311673)*x[1]**o+(0.554915758725)*x[1]+(0.824303662531)*x[2]**o+(-0.449201813108)*x[2]
        s[3,2,2,0]=(-0.284000203352)*x[0]**o+(0.0535558865246)*x[0]+(-0.233670301273)*x[1]**o+(0.383324533244)*x[1]+(-0.483402264684)*x[2]**o+(-0.482290332534)*x[2]
        s[3,2,2,1]=(-0.557106835416)*x[0]**o+(-0.0263124485334)*x[0]+(-0.132921973706)*x[1]**o+(0.0351808996763)*x[1]+(0.979819304163)*x[2]**o+(-0.908131098335)*x[2]
        s[3,3,0,0]=(0.982856633407)*x[0]**o+(-0.180985287571)*x[0]+(0.364546381679)*x[1]**o+(0.746225949823)*x[1]+(0.847682125532)*x[2]**o+(-0.782701591901)*x[2]
        s[3,3,0,1]=(-0.33362442802)*x[0]**o+(-0.933319602818)*x[0]+(-0.830312620251)*x[1]**o+(-0.938028393959)*x[1]+(0.859281853158)*x[2]**o+(0.430732343672)*x[2]
        s[3,3,1,0]=(0.381562874557)*x[0]**o+(0.0939156358041)*x[0]+(0.207361800369)*x[1]**o+(0.115172135617)*x[1]+(0.747494745689)*x[2]**o+(-0.45448955026)*x[2]
        s[3,3,1,1]=(0.095679833076)*x[0]**o+(-0.936886715289)*x[0]+(0.887146346873)*x[1]**o+(0.159246683343)*x[1]+(0.817874529811)*x[2]**o+(0.516091968965)*x[2]
        s[3,3,2,0]=(-0.476953971316)*x[0]**o+(0.176188538145)*x[0]+(0.568224935975)*x[1]**o+(0.56192429612)*x[1]+(-0.700911152919)*x[2]**o+(-0.0114437684935)*x[2]
        s[3,3,2,1]=(0.29541245686)*x[0]**o+(0.546841142855)*x[0]+(-0.23081431627)*x[1]**o+(-0.633697713725)*x[1]+(0.33519295746)*x[2]**o+(-0.951229146931)*x[2]
        s[3,4,0,0]=(-0.879148775346)*x[0]**o+(-0.703893190615)*x[0]+(0.409831208498)*x[1]**o+(0.754374287841)*x[1]+(0.688380721157)*x[2]**o+(-0.395460537774)*x[2]
        s[3,4,0,1]=(0.924489816348)*x[0]**o+(0.999127631985)*x[0]+(-0.892638544934)*x[1]**o+(-0.923913790342)*x[1]+(0.736487445972)*x[2]**o+(0.250131270052)*x[2]
        s[3,4,1,0]=(0.264892665874)*x[0]**o+(-0.776154012834)*x[0]+(-0.379853719079)*x[1]**o+(-0.748063746312)*x[1]+(-0.0108974578949)*x[2]**o+(-0.803449911818)*x[2]
        s[3,4,1,1]=(-0.699932530879)*x[0]**o+(-0.901128866521)*x[0]+(0.628561941772)*x[1]**o+(0.507625196603)*x[1]+(0.147922829864)*x[2]**o+(0.475783034469)*x[2]
        s[3,4,2,0]=(0.482139761317)*x[0]**o+(-0.627593411974)*x[0]+(-0.84668128017)*x[1]**o+(-0.324236153907)*x[1]+(-0.688406716909)*x[2]**o+(-0.0669177294519)*x[2]
        s[3,4,2,1]=(-0.531896625188)*x[0]**o+(-0.574459691455)*x[0]+(-0.824941329118)*x[1]**o+(0.0812437772861)*x[1]+(-0.257954477449)*x[2]**o+(-0.219142537545)*x[2]
        ref[0,0,0,0]=(-0.173295097983)*x_ref[0]**o+(0.773632210562)*x_ref[0]+(-0.521496418821)*x_ref[1]**o+(-0.947907473579)*x_ref[1]+(0.847721381317)*x_ref[2]**o+(-0.887851963083)*x_ref[2]
        ref[0,0,0,1]=(0.0654213598408)*x_ref[0]**o+(0.763489939538)*x_ref[0]+(-0.176800713154)*x_ref[1]**o+(-0.602428137248)*x_ref[1]+(-0.851802251531)*x_ref[2]**o+(-0.763084376785)*x_ref[2]
        ref[0,0,1,0]=(0.30505329426)*x_ref[0]**o+(-0.378878190786)*x_ref[0]+(-0.615554910302)*x_ref[1]**o+(0.0555314463878)*x_ref[1]+(-0.063211109707)*x_ref[2]**o+(-0.623580981446)*x_ref[2]
        ref[0,0,1,1]=(-0.526220589068)*x_ref[0]**o+(-0.551655771219)*x_ref[0]+(0.163011887243)*x_ref[1]**o+(-0.0863220304472)*x_ref[1]+(-0.638708196977)*x_ref[2]**o+(-0.298073090554)*x_ref[2]
        ref[0,0,2,0]=(0.0698028632972)*x_ref[0]**o+(0.508381168266)*x_ref[0]+(-0.120587433211)*x_ref[1]**o+(-0.419454430666)*x_ref[1]+(0.389702566269)*x_ref[2]**o+(-0.169655820731)*x_ref[2]
        ref[0,0,2,1]=(-0.80612077313)*x_ref[0]**o+(0.640104753466)*x_ref[0]+(0.518635274506)*x_ref[1]**o+(-0.545279092505)*x_ref[1]+(0.195002020672)*x_ref[2]**o+(0.563537057084)*x_ref[2]
        ref[0,1,0,0]=(-0.613471743331)*x_ref[0]**o+(-0.274063857866)*x_ref[0]+(-0.259665265722)*x_ref[1]**o+(0.0830136895696)*x_ref[1]+(0.825880322476)*x_ref[2]**o+(0.133039077085)*x_ref[2]
        ref[0,1,0,1]=(0.727956094813)*x_ref[0]**o+(0.605313602734)*x_ref[0]+(0.708549587523)*x_ref[1]**o+(-0.964527272447)*x_ref[1]+(-0.740107990066)*x_ref[2]**o+(0.629263961477)*x_ref[2]
        ref[0,1,1,0]=(-0.70107143852)*x_ref[0]**o+(-0.055344456929)*x_ref[0]+(-0.954082414306)*x_ref[1]**o+(-0.702136485683)*x_ref[1]+(-0.801391066649)*x_ref[2]**o+(-0.905767030094)*x_ref[2]
        ref[0,1,1,1]=(0.67172040238)*x_ref[0]**o+(0.740486014051)*x_ref[0]+(-0.335682691861)*x_ref[1]**o+(0.0793015412254)*x_ref[1]+(-0.868582520395)*x_ref[2]**o+(0.0848453101016)*x_ref[2]
        ref[0,1,2,0]=(0.224190760234)*x_ref[0]**o+(0.139996595496)*x_ref[0]+(-0.519770806543)*x_ref[1]**o+(-0.683808752006)*x_ref[1]+(-0.837082787693)*x_ref[2]**o+(-0.613900014511)*x_ref[2]
        ref[0,1,2,1]=(0.213979011641)*x_ref[0]**o+(-0.309663935547)*x_ref[0]+(0.0211525822739)*x_ref[1]**o+(0.844506291028)*x_ref[1]+(-0.688933639701)*x_ref[2]**o+(-0.336614773412)*x_ref[2]
        ref[0,2,0,0]=(-0.0188626152353)*x_ref[0]**o+(-0.694756450245)*x_ref[0]+(0.960579615684)*x_ref[1]**o+(0.57325819325)*x_ref[1]+(0.917037597322)*x_ref[2]**o+(-0.0584252009816)*x_ref[2]
        ref[0,2,0,1]=(0.803966493511)*x_ref[0]**o+(-0.169885689634)*x_ref[0]+(-0.556021959109)*x_ref[1]**o+(0.273901944163)*x_ref[1]+(-0.737708052821)*x_ref[2]**o+(0.627870713104)*x_ref[2]
        ref[0,2,1,0]=(-0.511561329117)*x_ref[0]**o+(-0.0854308669951)*x_ref[0]+(0.464246331007)*x_ref[1]**o+(-0.97461601243)*x_ref[1]+(-0.223895747162)*x_ref[2]**o+(-0.385389479659)*x_ref[2]
        ref[0,2,1,1]=(-0.723508721341)*x_ref[0]**o+(-0.201904055697)*x_ref[0]+(0.0789759340837)*x_ref[1]**o+(-0.00248269457662)*x_ref[1]+(-0.0626643900787)*x_ref[2]**o+(-0.128428965973)*x_ref[2]
        ref[0,2,2,0]=(0.934647998503)*x_ref[0]**o+(-0.0922965345918)*x_ref[0]+(0.717876178289)*x_ref[1]**o+(0.344875905273)*x_ref[1]+(0.0410512922166)*x_ref[2]**o+(0.164242310844)*x_ref[2]
        ref[0,2,2,1]=(-0.699306080605)*x_ref[0]**o+(0.209729161919)*x_ref[0]+(0.820435958574)*x_ref[1]**o+(-0.173302359677)*x_ref[1]+(0.659952656349)*x_ref[2]**o+(-0.78057378087)*x_ref[2]
        ref[0,3,0,0]=(-0.495298694962)*x_ref[0]**o+(0.571367877972)*x_ref[0]+(0.931242429977)*x_ref[1]**o+(0.888499081573)*x_ref[1]+(-0.408432694783)*x_ref[2]**o+(-0.898594385103)*x_ref[2]
        ref[0,3,0,1]=(-0.96889210815)*x_ref[0]**o+(0.404624542424)*x_ref[0]+(0.380170350345)*x_ref[1]**o+(-0.557523650891)*x_ref[1]+(-0.566134006)*x_ref[2]**o+(0.74677094551)*x_ref[2]
        ref[0,3,1,0]=(-0.224311775759)*x_ref[0]**o+(-0.660099304736)*x_ref[0]+(-0.513720022847)*x_ref[1]**o+(0.534948126324)*x_ref[1]+(0.162072633001)*x_ref[2]**o+(-0.402416011787)*x_ref[2]
        ref[0,3,1,1]=(0.224350718771)*x_ref[0]**o+(0.283343383933)*x_ref[0]+(0.138291871569)*x_ref[1]**o+(0.543289554343)*x_ref[1]+(-0.844288550977)*x_ref[2]**o+(0.463476015453)*x_ref[2]
        ref[0,3,2,0]=(-0.685488143712)*x_ref[0]**o+(0.300413146898)*x_ref[0]+(0.748708379841)*x_ref[1]**o+(-0.598212294807)*x_ref[1]+(0.479982541942)*x_ref[2]**o+(0.23866536682)*x_ref[2]
        ref[0,3,2,1]=(0.940674112182)*x_ref[0]**o+(-0.35846015272)*x_ref[0]+(-0.770304297672)*x_ref[1]**o+(-0.73897352885)*x_ref[1]+(-0.682102723348)*x_ref[2]**o+(0.742528100693)*x_ref[2]
        ref[0,4,0,0]=(0.0162066911802)*x_ref[0]**o+(0.160336749821)*x_ref[0]+(0.121948909833)*x_ref[1]**o+(-0.102093937457)*x_ref[1]+(-0.768661339258)*x_ref[2]**o+(0.98805683085)*x_ref[2]
        ref[0,4,0,1]=(-0.420990735691)*x_ref[0]**o+(-0.500244152083)*x_ref[0]+(0.910641832988)*x_ref[1]**o+(0.979658885966)*x_ref[1]+(-0.946590371476)*x_ref[2]**o+(-0.190997339007)*x_ref[2]
        ref[0,4,1,0]=(-0.445779943656)*x_ref[0]**o+(-0.378857349604)*x_ref[0]+(0.125534672956)*x_ref[1]**o+(0.124248627713)*x_ref[1]+(-0.685767125932)*x_ref[2]**o+(0.643604950542)*x_ref[2]
        ref[0,4,1,1]=(0.518152834346)*x_ref[0]**o+(-0.891873054984)*x_ref[0]+(-0.648744206994)*x_ref[1]**o+(0.207049340927)*x_ref[1]+(0.198529322067)*x_ref[2]**o+(-0.814832030848)*x_ref[2]
        ref[0,4,2,0]=(0.477174707789)*x_ref[0]**o+(-0.724452116357)*x_ref[0]+(0.209008883429)*x_ref[1]**o+(0.79968814697)*x_ref[1]+(0.129144596117)*x_ref[2]**o+(0.214098340906)*x_ref[2]
        ref[0,4,2,1]=(0.117050329845)*x_ref[0]**o+(-0.721574040514)*x_ref[0]+(0.382424906743)*x_ref[1]**o+(-0.135392966957)*x_ref[1]+(0.24455410794)*x_ref[2]**o+(-0.679165459966)*x_ref[2]
        ref[1,0,0,0]=(-0.180925772343)*x_ref[0]**o+(0.664151970586)*x_ref[0]+(0.0496840383663)*x_ref[1]**o+(-0.899468244962)*x_ref[1]+(0.942251196146)*x_ref[2]**o+(0.228837829842)*x_ref[2]
        ref[1,0,0,1]=(0.15346023848)*x_ref[0]**o+(-0.642882884964)*x_ref[0]+(-0.462925865669)*x_ref[1]**o+(0.738240095188)*x_ref[1]+(-0.57213723904)*x_ref[2]**o+(-0.900335371347)*x_ref[2]
        ref[1,0,1,0]=(-0.366392792085)*x_ref[0]**o+(-0.227326930302)*x_ref[0]+(-0.323111857481)*x_ref[1]**o+(0.220826082736)*x_ref[1]+(0.237722021225)*x_ref[2]**o+(0.835402283453)*x_ref[2]
        ref[1,0,1,1]=(-0.293188444394)*x_ref[0]**o+(0.338812067406)*x_ref[0]+(0.679884979413)*x_ref[1]**o+(-0.614317813911)*x_ref[1]+(0.936906665544)*x_ref[2]**o+(-0.652679743405)*x_ref[2]
        ref[1,0,2,0]=(0.617621972453)*x_ref[0]**o+(-0.936326816619)*x_ref[0]+(0.760134028716)*x_ref[1]**o+(-0.809380984591)*x_ref[1]+(0.562612935213)*x_ref[2]**o+(0.0794516002575)*x_ref[2]
        ref[1,0,2,1]=(0.448514601948)*x_ref[0]**o+(0.327970604314)*x_ref[0]+(-0.936492917037)*x_ref[1]**o+(-0.425126227791)*x_ref[1]+(-0.372116400796)*x_ref[2]**o+(-0.958054327266)*x_ref[2]
        ref[1,1,0,0]=(-0.850336943937)*x_ref[0]**o+(0.379951917933)*x_ref[0]+(0.693148856845)*x_ref[1]**o+(-0.691233930311)*x_ref[1]+(0.548624260329)*x_ref[2]**o+(-0.193125714718)*x_ref[2]
        ref[1,1,0,1]=(-0.823007972929)*x_ref[0]**o+(0.408191529362)*x_ref[0]+(-0.023921849403)*x_ref[1]**o+(0.79830304115)*x_ref[1]+(0.506756293069)*x_ref[2]**o+(0.434230444707)*x_ref[2]
        ref[1,1,1,0]=(0.263932235445)*x_ref[0]**o+(-0.600679211293)*x_ref[0]+(0.480835197573)*x_ref[1]**o+(-0.976282735826)*x_ref[1]+(0.714583200236)*x_ref[2]**o+(0.942610909813)*x_ref[2]
        ref[1,1,1,1]=(0.481454842664)*x_ref[0]**o+(-0.508309985427)*x_ref[0]+(-0.739763965139)*x_ref[1]**o+(0.211871147037)*x_ref[1]+(-0.848812488744)*x_ref[2]**o+(0.885973013107)*x_ref[2]
        ref[1,1,2,0]=(-0.796437949787)*x_ref[0]**o+(-0.211960906694)*x_ref[0]+(0.823389343923)*x_ref[1]**o+(-0.0162189137972)*x_ref[1]+(0.77124757744)*x_ref[2]**o+(0.877807214589)*x_ref[2]
        ref[1,1,2,1]=(0.899446895743)*x_ref[0]**o+(-0.303223847939)*x_ref[0]+(0.409199240049)*x_ref[1]**o+(0.430938116282)*x_ref[1]+(0.189267561328)*x_ref[2]**o+(0.582418236586)*x_ref[2]
        ref[1,2,0,0]=(0.687732048304)*x_ref[0]**o+(0.338825649717)*x_ref[0]+(-0.535870442333)*x_ref[1]**o+(0.507773000137)*x_ref[1]+(-0.892023180429)*x_ref[2]**o+(0.35147003979)*x_ref[2]
        ref[1,2,0,1]=(-0.315183497372)*x_ref[0]**o+(-0.0302851860468)*x_ref[0]+(-0.966617816351)*x_ref[1]**o+(-0.876528463134)*x_ref[1]+(-0.682249112679)*x_ref[2]**o+(-0.830991774314)*x_ref[2]
        ref[1,2,1,0]=(0.90817845095)*x_ref[0]**o+(-0.629542958154)*x_ref[0]+(0.663791954021)*x_ref[1]**o+(0.584246811118)*x_ref[1]+(0.421582877787)*x_ref[2]**o+(0.316731433829)*x_ref[2]
        ref[1,2,1,1]=(-0.948554425084)*x_ref[0]**o+(-0.365749706899)*x_ref[0]+(0.282031925475)*x_ref[1]**o+(0.467581708634)*x_ref[1]+(0.158633292573)*x_ref[2]**o+(-0.0990441468359)*x_ref[2]
        ref[1,2,2,0]=(0.113170413925)*x_ref[0]**o+(0.481722903666)*x_ref[0]+(0.0127374268921)*x_ref[1]**o+(-0.778057043706)*x_ref[1]+(0.399544836507)*x_ref[2]**o+(-0.157225042768)*x_ref[2]
        ref[1,2,2,1]=(0.320987735687)*x_ref[0]**o+(-0.29324177158)*x_ref[0]+(0.02800800296)*x_ref[1]**o+(-0.277948207921)*x_ref[1]+(-0.271808717111)*x_ref[2]**o+(-0.0379931737995)*x_ref[2]
        ref[1,3,0,0]=(0.09942034907)*x_ref[0]**o+(0.939629344394)*x_ref[0]+(-0.810686196185)*x_ref[1]**o+(-0.538230260705)*x_ref[1]+(0.85351760208)*x_ref[2]**o+(-0.737196688288)*x_ref[2]
        ref[1,3,0,1]=(0.48690360509)*x_ref[0]**o+(-0.065370723503)*x_ref[0]+(-0.0566373715183)*x_ref[1]**o+(0.871878739279)*x_ref[1]+(0.272054032769)*x_ref[2]**o+(0.540626473697)*x_ref[2]
        ref[1,3,1,0]=(0.472815976006)*x_ref[0]**o+(-0.775416335945)*x_ref[0]+(0.213646821631)*x_ref[1]**o+(0.415546582713)*x_ref[1]+(-0.196674759494)*x_ref[2]**o+(0.235475790794)*x_ref[2]
        ref[1,3,1,1]=(-0.699232847611)*x_ref[0]**o+(0.504961844211)*x_ref[0]+(-0.0115303749449)*x_ref[1]**o+(-0.756286148171)*x_ref[1]+(-0.0922447811444)*x_ref[2]**o+(-0.330281180023)*x_ref[2]
        ref[1,3,2,0]=(0.149323550569)*x_ref[0]**o+(0.977735452052)*x_ref[0]+(0.255972870292)*x_ref[1]**o+(-0.855635543118)*x_ref[1]+(0.519602177962)*x_ref[2]**o+(0.922189470945)*x_ref[2]
        ref[1,3,2,1]=(-0.190749790276)*x_ref[0]**o+(-0.351506874986)*x_ref[0]+(-0.0672783552454)*x_ref[1]**o+(-0.223773986529)*x_ref[1]+(-0.593750649495)*x_ref[2]**o+(0.158607153203)*x_ref[2]
        ref[1,4,0,0]=(0.245306614552)*x_ref[0]**o+(-0.191799968697)*x_ref[0]+(0.85455961157)*x_ref[1]**o+(0.597426986918)*x_ref[1]+(-0.824363631294)*x_ref[2]**o+(-0.0161601323913)*x_ref[2]
        ref[1,4,0,1]=(0.925974209919)*x_ref[0]**o+(-0.606855935989)*x_ref[0]+(0.182619376779)*x_ref[1]**o+(0.82202480925)*x_ref[1]+(0.0511805106413)*x_ref[2]**o+(0.0866309784619)*x_ref[2]
        ref[1,4,1,0]=(0.272918227857)*x_ref[0]**o+(-0.406306617661)*x_ref[0]+(-0.0535134980185)*x_ref[1]**o+(-0.186200126868)*x_ref[1]+(0.92546875085)*x_ref[2]**o+(-0.474811198949)*x_ref[2]
        ref[1,4,1,1]=(0.803708551531)*x_ref[0]**o+(-0.715626841623)*x_ref[0]+(-0.0822999615139)*x_ref[1]**o+(0.612334309398)*x_ref[1]+(-0.157577434512)*x_ref[2]**o+(-0.795469427996)*x_ref[2]
        ref[1,4,2,0]=(-0.837159054411)*x_ref[0]**o+(-0.234523360065)*x_ref[0]+(-0.198337862205)*x_ref[1]**o+(0.897958649268)*x_ref[1]+(-0.519907680436)*x_ref[2]**o+(0.742526415023)*x_ref[2]
        ref[1,4,2,1]=(-0.792448458791)*x_ref[0]**o+(0.456487739729)*x_ref[0]+(0.0128609741891)*x_ref[1]**o+(-0.344964414689)*x_ref[1]+(-0.205584911868)*x_ref[2]**o+(0.429183240889)*x_ref[2]
        ref[2,0,0,0]=(0.401501024951)*x_ref[0]**o+(-0.497039379761)*x_ref[0]+(-0.228675343279)*x_ref[1]**o+(0.133323657822)*x_ref[1]+(-0.203006510606)*x_ref[2]**o+(0.407718693225)*x_ref[2]
        ref[2,0,0,1]=(-0.547448204328)*x_ref[0]**o+(-0.109789709109)*x_ref[0]+(-0.626633454424)*x_ref[1]**o+(-0.117871748454)*x_ref[1]+(0.961790647477)*x_ref[2]**o+(0.652883936373)*x_ref[2]
        ref[2,0,1,0]=(-0.85536471584)*x_ref[0]**o+(0.489291116489)*x_ref[0]+(-0.929596786116)*x_ref[1]**o+(0.0110017367519)*x_ref[1]+(-0.703617460934)*x_ref[2]**o+(0.0496869388223)*x_ref[2]
        ref[2,0,1,1]=(-0.501093953991)*x_ref[0]**o+(-0.722768363488)*x_ref[0]+(0.84017017736)*x_ref[1]**o+(-0.22900603012)*x_ref[1]+(0.760492299454)*x_ref[2]**o+(0.767346260882)*x_ref[2]
        ref[2,0,2,0]=(-0.328498451765)*x_ref[0]**o+(-0.387534616444)*x_ref[0]+(-0.0973382392242)*x_ref[1]**o+(0.418582689709)*x_ref[1]+(0.206439339821)*x_ref[2]**o+(-0.706353598324)*x_ref[2]
        ref[2,0,2,1]=(0.179977568909)*x_ref[0]**o+(0.642598753874)*x_ref[0]+(0.711020603016)*x_ref[1]**o+(-0.287193672421)*x_ref[1]+(-0.362612184622)*x_ref[2]**o+(-0.18146063868)*x_ref[2]
        ref[2,1,0,0]=(0.0470678392348)*x_ref[0]**o+(0.00649068845495)*x_ref[0]+(-0.239912084684)*x_ref[1]**o+(0.278051515796)*x_ref[1]+(-0.98924625944)*x_ref[2]**o+(0.847957009386)*x_ref[2]
        ref[2,1,0,1]=(-0.287954694237)*x_ref[0]**o+(0.127947128577)*x_ref[0]+(0.443869903193)*x_ref[1]**o+(-0.344694028625)*x_ref[1]+(-0.14884506191)*x_ref[2]**o+(0.748140862905)*x_ref[2]
        ref[2,1,1,0]=(-0.597041891422)*x_ref[0]**o+(0.914716550666)*x_ref[0]+(0.68496200279)*x_ref[1]**o+(0.0420126536598)*x_ref[1]+(-0.539966353171)*x_ref[2]**o+(-0.886373035297)*x_ref[2]
        ref[2,1,1,1]=(0.462776858955)*x_ref[0]**o+(-0.989367011955)*x_ref[0]+(0.421691081662)*x_ref[1]**o+(0.264858346124)*x_ref[1]+(0.513954688709)*x_ref[2]**o+(0.00863898872886)*x_ref[2]
        ref[2,1,2,0]=(0.503601863243)*x_ref[0]**o+(0.505973913611)*x_ref[0]+(0.0952160099415)*x_ref[1]**o+(-0.77103310327)*x_ref[1]+(-0.316507531754)*x_ref[2]**o+(0.506969348695)*x_ref[2]
        ref[2,1,2,1]=(0.521634307558)*x_ref[0]**o+(-0.910583030486)*x_ref[0]+(0.977891014192)*x_ref[1]**o+(0.985145579284)*x_ref[1]+(0.0465919834269)*x_ref[2]**o+(0.113946013883)*x_ref[2]
        ref[2,2,0,0]=(0.178722112454)*x_ref[0]**o+(0.582800628485)*x_ref[0]+(0.141950810923)*x_ref[1]**o+(0.226899624421)*x_ref[1]+(0.445681888363)*x_ref[2]**o+(-0.127689044751)*x_ref[2]
        ref[2,2,0,1]=(-0.433214095841)*x_ref[0]**o+(-0.61199374178)*x_ref[0]+(0.655706389409)*x_ref[1]**o+(-0.0160036542096)*x_ref[1]+(0.570017878158)*x_ref[2]**o+(-0.00718475090704)*x_ref[2]
        ref[2,2,1,0]=(-0.205396523704)*x_ref[0]**o+(0.87306641582)*x_ref[0]+(-0.664912090364)*x_ref[1]**o+(-0.324577426693)*x_ref[1]+(-0.37269275568)*x_ref[2]**o+(-0.0753237226649)*x_ref[2]
        ref[2,2,1,1]=(0.716795494369)*x_ref[0]**o+(-0.725543593985)*x_ref[0]+(0.751101352157)*x_ref[1]**o+(0.0629757438809)*x_ref[1]+(-0.388859722245)*x_ref[2]**o+(0.633579629088)*x_ref[2]
        ref[2,2,2,0]=(-0.785119558914)*x_ref[0]**o+(-0.289212761766)*x_ref[0]+(-0.947517483982)*x_ref[1]**o+(-0.898908873087)*x_ref[1]+(-0.758807720992)*x_ref[2]**o+(-0.747003629416)*x_ref[2]
        ref[2,2,2,1]=(0.384610894358)*x_ref[0]**o+(0.0170687884025)*x_ref[0]+(0.776211872471)*x_ref[1]**o+(-0.313821704113)*x_ref[1]+(-0.433021717398)*x_ref[2]**o+(-0.972385486304)*x_ref[2]
        ref[2,3,0,0]=(0.59307540612)*x_ref[0]**o+(0.93544260405)*x_ref[0]+(0.519016668994)*x_ref[1]**o+(-0.175290607112)*x_ref[1]+(-0.269720511811)*x_ref[2]**o+(0.117237249214)*x_ref[2]
        ref[2,3,0,1]=(-0.0752366310333)*x_ref[0]**o+(-0.168792429308)*x_ref[0]+(0.316236130047)*x_ref[1]**o+(0.021991183167)*x_ref[1]+(-0.2024837663)*x_ref[2]**o+(-0.676832860181)*x_ref[2]
        ref[2,3,1,0]=(-0.2682564416)*x_ref[0]**o+(0.276503108458)*x_ref[0]+(0.789470125022)*x_ref[1]**o+(-0.339044117352)*x_ref[1]+(-0.408507243588)*x_ref[2]**o+(0.0666267661149)*x_ref[2]
        ref[2,3,1,1]=(-0.053395293244)*x_ref[0]**o+(0.425880490109)*x_ref[0]+(-0.068128758144)*x_ref[1]**o+(0.413635486374)*x_ref[1]+(0.965050652205)*x_ref[2]**o+(0.126502830836)*x_ref[2]
        ref[2,3,2,0]=(0.601928913545)*x_ref[0]**o+(-0.0443247453956)*x_ref[0]+(-0.606811664164)*x_ref[1]**o+(0.0206686600296)*x_ref[1]+(-0.804806810474)*x_ref[2]**o+(-0.019745599298)*x_ref[2]
        ref[2,3,2,1]=(0.911324514835)*x_ref[0]**o+(-0.397270365997)*x_ref[0]+(0.511154806396)*x_ref[1]**o+(0.722003975461)*x_ref[1]+(-0.332922355553)*x_ref[2]**o+(-0.132309188597)*x_ref[2]
        ref[2,4,0,0]=(0.194698324972)*x_ref[0]**o+(-0.496022175033)*x_ref[0]+(0.804006600349)*x_ref[1]**o+(-0.643493478792)*x_ref[1]+(0.148395455552)*x_ref[2]**o+(-0.408862596804)*x_ref[2]
        ref[2,4,0,1]=(0.734891277828)*x_ref[0]**o+(-0.101225311176)*x_ref[0]+(-0.813960759076)*x_ref[1]**o+(0.862888786761)*x_ref[1]+(0.830408706284)*x_ref[2]**o+(0.523449046413)*x_ref[2]
        ref[2,4,1,0]=(0.286629494089)*x_ref[0]**o+(0.599214589464)*x_ref[0]+(-0.208652251419)*x_ref[1]**o+(-0.830028802843)*x_ref[1]+(0.805360345419)*x_ref[2]**o+(-0.865349771687)*x_ref[2]
        ref[2,4,1,1]=(0.0367955096711)*x_ref[0]**o+(-0.0261723558157)*x_ref[0]+(0.200900558315)*x_ref[1]**o+(-0.532346409176)*x_ref[1]+(-0.183663881543)*x_ref[2]**o+(0.93191706611)*x_ref[2]
        ref[2,4,2,0]=(-0.932223437809)*x_ref[0]**o+(-0.155278731606)*x_ref[0]+(-0.932519381895)*x_ref[1]**o+(0.0589900357358)*x_ref[1]+(-0.187551917707)*x_ref[2]**o+(-0.81649531847)*x_ref[2]
        ref[2,4,2,1]=(0.555481188864)*x_ref[0]**o+(0.946358167786)*x_ref[0]+(-0.231784522)*x_ref[1]**o+(0.592776954307)*x_ref[1]+(-0.0373202061188)*x_ref[2]**o+(0.0302770826388)*x_ref[2]
        ref[3,0,0,0]=(0.603747717227)*x_ref[0]**o+(-0.969329597018)*x_ref[0]+(0.311876345528)*x_ref[1]**o+(-0.589486433983)*x_ref[1]+(-0.294507657675)*x_ref[2]**o+(-0.743511183116)*x_ref[2]
        ref[3,0,0,1]=(-0.846568198319)*x_ref[0]**o+(0.93521569126)*x_ref[0]+(-0.874618479052)*x_ref[1]**o+(0.502042421273)*x_ref[1]+(-0.945100352471)*x_ref[2]**o+(0.765854746295)*x_ref[2]
        ref[3,0,1,0]=(-0.908353314852)*x_ref[0]**o+(0.701990826648)*x_ref[0]+(0.798549198275)*x_ref[1]**o+(-0.616631523859)*x_ref[1]+(-0.999800863738)*x_ref[2]**o+(-0.156585259768)*x_ref[2]
        ref[3,0,1,1]=(-0.433258077523)*x_ref[0]**o+(-0.641049552618)*x_ref[0]+(0.427903347848)*x_ref[1]**o+(-0.175290793023)*x_ref[1]+(-0.557154490598)*x_ref[2]**o+(-0.00495777825998)*x_ref[2]
        ref[3,0,2,0]=(0.998253378112)*x_ref[0]**o+(0.2484283851)*x_ref[0]+(0.955897183936)*x_ref[1]**o+(-0.0351483480831)*x_ref[1]+(0.690211608431)*x_ref[2]**o+(0.7420822406)*x_ref[2]
        ref[3,0,2,1]=(0.273153627728)*x_ref[0]**o+(0.931572979903)*x_ref[0]+(-0.279174117029)*x_ref[1]**o+(-0.741239387871)*x_ref[1]+(0.148372377955)*x_ref[2]**o+(0.781476364189)*x_ref[2]
        ref[3,1,0,0]=(0.231049869286)*x_ref[0]**o+(-0.871461717509)*x_ref[0]+(0.553545032585)*x_ref[1]**o+(0.979019482224)*x_ref[1]+(0.819288655429)*x_ref[2]**o+(-0.01489981549)*x_ref[2]
        ref[3,1,0,1]=(-0.891391000148)*x_ref[0]**o+(-0.360048424965)*x_ref[0]+(-0.390043258734)*x_ref[1]**o+(0.547316173383)*x_ref[1]+(-0.972094018922)*x_ref[2]**o+(0.7798960782)*x_ref[2]
        ref[3,1,1,0]=(0.675405837097)*x_ref[0]**o+(0.976689341648)*x_ref[0]+(-0.416251051342)*x_ref[1]**o+(0.696304358663)*x_ref[1]+(-0.395743544878)*x_ref[2]**o+(-0.0340340734736)*x_ref[2]
        ref[3,1,1,1]=(-0.531871504573)*x_ref[0]**o+(-0.926770430235)*x_ref[0]+(0.752552585571)*x_ref[1]**o+(-0.403402334727)*x_ref[1]+(0.07593559269)*x_ref[2]**o+(0.126742096076)*x_ref[2]
        ref[3,1,2,0]=(-0.033528429617)*x_ref[0]**o+(0.187766565831)*x_ref[0]+(0.192128207394)*x_ref[1]**o+(-0.0969373324914)*x_ref[1]+(-0.889231964389)*x_ref[2]**o+(0.493198117275)*x_ref[2]
        ref[3,1,2,1]=(-0.182267751454)*x_ref[0]**o+(-0.115192609603)*x_ref[0]+(-0.252701809463)*x_ref[1]**o+(0.355949096242)*x_ref[1]+(0.859478772116)*x_ref[2]**o+(0.539090466518)*x_ref[2]
        ref[3,2,0,0]=(0.25232058799)*x_ref[0]**o+(-0.246680227462)*x_ref[0]+(0.316105270194)*x_ref[1]**o+(-0.426505186439)*x_ref[1]+(0.980259506048)*x_ref[2]**o+(0.144514677615)*x_ref[2]
        ref[3,2,0,1]=(-0.583071409136)*x_ref[0]**o+(0.969922333041)*x_ref[0]+(-0.0251030152687)*x_ref[1]**o+(0.121967009007)*x_ref[1]+(0.391834007838)*x_ref[2]**o+(0.0339488845613)*x_ref[2]
        ref[3,2,1,0]=(0.706904708517)*x_ref[0]**o+(0.266141083673)*x_ref[0]+(-0.0339160692688)*x_ref[1]**o+(-0.152115678673)*x_ref[1]+(0.122213919626)*x_ref[2]**o+(-0.576478442144)*x_ref[2]
        ref[3,2,1,1]=(-0.917141077151)*x_ref[0]**o+(0.569284998999)*x_ref[0]+(-0.0979622311673)*x_ref[1]**o+(0.554915758725)*x_ref[1]+(0.824303662531)*x_ref[2]**o+(-0.449201813108)*x_ref[2]
        ref[3,2,2,0]=(-0.284000203352)*x_ref[0]**o+(0.0535558865246)*x_ref[0]+(-0.233670301273)*x_ref[1]**o+(0.383324533244)*x_ref[1]+(-0.483402264684)*x_ref[2]**o+(-0.482290332534)*x_ref[2]
        ref[3,2,2,1]=(-0.557106835416)*x_ref[0]**o+(-0.0263124485334)*x_ref[0]+(-0.132921973706)*x_ref[1]**o+(0.0351808996763)*x_ref[1]+(0.979819304163)*x_ref[2]**o+(-0.908131098335)*x_ref[2]
        ref[3,3,0,0]=(0.982856633407)*x_ref[0]**o+(-0.180985287571)*x_ref[0]+(0.364546381679)*x_ref[1]**o+(0.746225949823)*x_ref[1]+(0.847682125532)*x_ref[2]**o+(-0.782701591901)*x_ref[2]
        ref[3,3,0,1]=(-0.33362442802)*x_ref[0]**o+(-0.933319602818)*x_ref[0]+(-0.830312620251)*x_ref[1]**o+(-0.938028393959)*x_ref[1]+(0.859281853158)*x_ref[2]**o+(0.430732343672)*x_ref[2]
        ref[3,3,1,0]=(0.381562874557)*x_ref[0]**o+(0.0939156358041)*x_ref[0]+(0.207361800369)*x_ref[1]**o+(0.115172135617)*x_ref[1]+(0.747494745689)*x_ref[2]**o+(-0.45448955026)*x_ref[2]
        ref[3,3,1,1]=(0.095679833076)*x_ref[0]**o+(-0.936886715289)*x_ref[0]+(0.887146346873)*x_ref[1]**o+(0.159246683343)*x_ref[1]+(0.817874529811)*x_ref[2]**o+(0.516091968965)*x_ref[2]
        ref[3,3,2,0]=(-0.476953971316)*x_ref[0]**o+(0.176188538145)*x_ref[0]+(0.568224935975)*x_ref[1]**o+(0.56192429612)*x_ref[1]+(-0.700911152919)*x_ref[2]**o+(-0.0114437684935)*x_ref[2]
        ref[3,3,2,1]=(0.29541245686)*x_ref[0]**o+(0.546841142855)*x_ref[0]+(-0.23081431627)*x_ref[1]**o+(-0.633697713725)*x_ref[1]+(0.33519295746)*x_ref[2]**o+(-0.951229146931)*x_ref[2]
        ref[3,4,0,0]=(-0.879148775346)*x_ref[0]**o+(-0.703893190615)*x_ref[0]+(0.409831208498)*x_ref[1]**o+(0.754374287841)*x_ref[1]+(0.688380721157)*x_ref[2]**o+(-0.395460537774)*x_ref[2]
        ref[3,4,0,1]=(0.924489816348)*x_ref[0]**o+(0.999127631985)*x_ref[0]+(-0.892638544934)*x_ref[1]**o+(-0.923913790342)*x_ref[1]+(0.736487445972)*x_ref[2]**o+(0.250131270052)*x_ref[2]
        ref[3,4,1,0]=(0.264892665874)*x_ref[0]**o+(-0.776154012834)*x_ref[0]+(-0.379853719079)*x_ref[1]**o+(-0.748063746312)*x_ref[1]+(-0.0108974578949)*x_ref[2]**o+(-0.803449911818)*x_ref[2]
        ref[3,4,1,1]=(-0.699932530879)*x_ref[0]**o+(-0.901128866521)*x_ref[0]+(0.628561941772)*x_ref[1]**o+(0.507625196603)*x_ref[1]+(0.147922829864)*x_ref[2]**o+(0.475783034469)*x_ref[2]
        ref[3,4,2,0]=(0.482139761317)*x_ref[0]**o+(-0.627593411974)*x_ref[0]+(-0.84668128017)*x_ref[1]**o+(-0.324236153907)*x_ref[1]+(-0.688406716909)*x_ref[2]**o+(-0.0669177294519)*x_ref[2]
        ref[3,4,2,1]=(-0.531896625188)*x_ref[0]**o+(-0.574459691455)*x_ref[0]+(-0.824941329118)*x_ref[1]**o+(0.0812437772861)*x_ref[1]+(-0.257954477449)*x_ref[2]**o+(-0.219142537545)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.318228469869)*x[0]+(-0.774132349214)*x[1]
        ref=(0.318228469869)*x_ref[0]+(-0.774132349214)*x_ref[1]
      else:
        s=(0.377557485345)*x[0]+(1.02901239394)*x[1]+(0.038224819288)*x[2]
        ref=(0.377557485345)*x_ref[0]+(1.02901239394)*x_ref[1]+(0.038224819288)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        s[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
        ref[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        ref[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
      else:
        s[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        s[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
        ref[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        ref[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0621856629646)*x[0]+(0.783319354938)*x[1]
        s[0,1]=(-0.205526239577)*x[0]+(-0.401770179168)*x[1]
        s[0,2]=(0.543085761147)*x[0]+(0.915124776348)*x[1]
        s[0,3]=(1.3114416307)*x[0]+(-0.339977887579)*x[1]
        s[0,4]=(0.19180167645)*x[0]+(0.0744591496582)*x[1]
        s[1,0]=(0.833949474687)*x[0]+(-0.513960535503)*x[1]
        s[1,1]=(1.63749958183)*x[0]+(1.49589084809)*x[1]
        s[1,2]=(0.52690970122)*x[0]+(-0.497651760738)*x[1]
        s[1,3]=(0.157987079101)*x[0]+(0.244658238724)*x[1]
        s[1,4]=(-0.557780706846)*x[0]+(-1.49932552405)*x[1]
        s[2,0]=(-0.00409488968736)*x[0]+(0.366430715843)*x[1]
        s[2,1]=(-0.867654712773)*x[0]+(-0.129712066272)*x[1]
        s[2,2]=(1.33589989715)*x[0]+(-0.135732989085)*x[1]
        s[2,3]=(0.208531325369)*x[0]+(1.3627828974)*x[1]
        s[2,4]=(-0.667299295418)*x[0]+(0.981775718726)*x[1]
        s[3,0]=(1.40151192578)*x[0]+(-0.28272210702)*x[1]
        s[3,1]=(-1.13395485034)*x[0]+(-0.148770918012)*x[1]
        s[3,2]=(0.315288948224)*x[0]+(-0.140365593027)*x[1]
        s[3,3]=(-0.0421809531999)*x[0]+(0.522687748574)*x[1]
        s[3,4]=(0.135043757063)*x[0]+(-1.52489039301)*x[1]
        ref[0,0]=(0.0621856629646)*x_ref[0]+(0.783319354938)*x_ref[1]
        ref[0,1]=(-0.205526239577)*x_ref[0]+(-0.401770179168)*x_ref[1]
        ref[0,2]=(0.543085761147)*x_ref[0]+(0.915124776348)*x_ref[1]
        ref[0,3]=(1.3114416307)*x_ref[0]+(-0.339977887579)*x_ref[1]
        ref[0,4]=(0.19180167645)*x_ref[0]+(0.0744591496582)*x_ref[1]
        ref[1,0]=(0.833949474687)*x_ref[0]+(-0.513960535503)*x_ref[1]
        ref[1,1]=(1.63749958183)*x_ref[0]+(1.49589084809)*x_ref[1]
        ref[1,2]=(0.52690970122)*x_ref[0]+(-0.497651760738)*x_ref[1]
        ref[1,3]=(0.157987079101)*x_ref[0]+(0.244658238724)*x_ref[1]
        ref[1,4]=(-0.557780706846)*x_ref[0]+(-1.49932552405)*x_ref[1]
        ref[2,0]=(-0.00409488968736)*x_ref[0]+(0.366430715843)*x_ref[1]
        ref[2,1]=(-0.867654712773)*x_ref[0]+(-0.129712066272)*x_ref[1]
        ref[2,2]=(1.33589989715)*x_ref[0]+(-0.135732989085)*x_ref[1]
        ref[2,3]=(0.208531325369)*x_ref[0]+(1.3627828974)*x_ref[1]
        ref[2,4]=(-0.667299295418)*x_ref[0]+(0.981775718726)*x_ref[1]
        ref[3,0]=(1.40151192578)*x_ref[0]+(-0.28272210702)*x_ref[1]
        ref[3,1]=(-1.13395485034)*x_ref[0]+(-0.148770918012)*x_ref[1]
        ref[3,2]=(0.315288948224)*x_ref[0]+(-0.140365593027)*x_ref[1]
        ref[3,3]=(-0.0421809531999)*x_ref[0]+(0.522687748574)*x_ref[1]
        ref[3,4]=(0.135043757063)*x_ref[0]+(-1.52489039301)*x_ref[1]
      else:
        s[0,0]=(-1.07097471698)*x[0]+(0.511673191366)*x[1]+(0.986422233856)*x[2]
        s[0,1]=(1.35021866887)*x[0]+(-1.18036897489)*x[1]+(0.143096009704)*x[2]
        s[0,2]=(1.92136475346)*x[0]+(-1.52041779666)*x[1]+(1.50412123174)*x[2]
        s[0,3]=(-1.26349432492)*x[0]+(-1.69090463821)*x[1]+(0.0519848717435)*x[2]
        s[0,4]=(-0.450643290235)*x[0]+(-1.0918592714)*x[1]+(-1.78531176031)*x[2]
        s[1,0]=(1.24575021839)*x[0]+(0.378183524368)*x[1]+(0.164647643767)*x[2]
        s[1,1]=(-0.0164625547855)*x[0]+(-1.06539959746)*x[1]+(-1.5644531354)*x[2]
        s[1,2]=(0.126087223544)*x[0]+(0.876303692726)*x[1]+(1.3521510598)*x[2]
        s[1,3]=(-1.15749467575)*x[0]+(-0.850991452556)*x[1]+(-0.903705777185)*x[2]
        s[1,4]=(-0.680094762766)*x[0]+(0.280093157366)*x[1]+(0.460572306025)*x[2]
        s[2,0]=(-0.201094411285)*x[0]+(0.610240266355)*x[1]+(0.741298609653)*x[2]
        s[2,1]=(0.754412798233)*x[0]+(0.765169641301)*x[1]+(0.522349231028)*x[2]
        s[2,2]=(-0.330239429991)*x[0]+(1.04468899978)*x[1]+(-0.493706328863)*x[2]
        s[2,3]=(-0.107189727407)*x[0]+(-1.57053989855)*x[1]+(1.5864416691)*x[2]
        s[2,4]=(1.78349630817)*x[0]+(0.415138686617)*x[1]+(-0.567656058268)*x[2]
        s[3,0]=(-0.242864411179)*x[0]+(0.342401319249)*x[1]+(-0.963749227433)*x[2]
        s[3,1]=(-0.677695348369)*x[0]+(-1.52126003219)*x[1]+(-0.197904354713)*x[2]
        s[3,2]=(-0.549674242841)*x[0]+(1.23313354594)*x[1]+(0.701202865744)*x[2]
        s[3,3]=(0.278097672739)*x[0]+(-1.31336623233)*x[1]+(1.69771904282)*x[2]
        s[3,4]=(0.0982668788821)*x[0]+(-1.11857597802)*x[1]+(-0.0505715204344)*x[2]
        ref[0,0]=(-1.07097471698)*x_ref[0]+(0.511673191366)*x_ref[1]+(0.986422233856)*x_ref[2]
        ref[0,1]=(1.35021866887)*x_ref[0]+(-1.18036897489)*x_ref[1]+(0.143096009704)*x_ref[2]
        ref[0,2]=(1.92136475346)*x_ref[0]+(-1.52041779666)*x_ref[1]+(1.50412123174)*x_ref[2]
        ref[0,3]=(-1.26349432492)*x_ref[0]+(-1.69090463821)*x_ref[1]+(0.0519848717435)*x_ref[2]
        ref[0,4]=(-0.450643290235)*x_ref[0]+(-1.0918592714)*x_ref[1]+(-1.78531176031)*x_ref[2]
        ref[1,0]=(1.24575021839)*x_ref[0]+(0.378183524368)*x_ref[1]+(0.164647643767)*x_ref[2]
        ref[1,1]=(-0.0164625547855)*x_ref[0]+(-1.06539959746)*x_ref[1]+(-1.5644531354)*x_ref[2]
        ref[1,2]=(0.126087223544)*x_ref[0]+(0.876303692726)*x_ref[1]+(1.3521510598)*x_ref[2]
        ref[1,3]=(-1.15749467575)*x_ref[0]+(-0.850991452556)*x_ref[1]+(-0.903705777185)*x_ref[2]
        ref[1,4]=(-0.680094762766)*x_ref[0]+(0.280093157366)*x_ref[1]+(0.460572306025)*x_ref[2]
        ref[2,0]=(-0.201094411285)*x_ref[0]+(0.610240266355)*x_ref[1]+(0.741298609653)*x_ref[2]
        ref[2,1]=(0.754412798233)*x_ref[0]+(0.765169641301)*x_ref[1]+(0.522349231028)*x_ref[2]
        ref[2,2]=(-0.330239429991)*x_ref[0]+(1.04468899978)*x_ref[1]+(-0.493706328863)*x_ref[2]
        ref[2,3]=(-0.107189727407)*x_ref[0]+(-1.57053989855)*x_ref[1]+(1.5864416691)*x_ref[2]
        ref[2,4]=(1.78349630817)*x_ref[0]+(0.415138686617)*x_ref[1]+(-0.567656058268)*x_ref[2]
        ref[3,0]=(-0.242864411179)*x_ref[0]+(0.342401319249)*x_ref[1]+(-0.963749227433)*x_ref[2]
        ref[3,1]=(-0.677695348369)*x_ref[0]+(-1.52126003219)*x_ref[1]+(-0.197904354713)*x_ref[2]
        ref[3,2]=(-0.549674242841)*x_ref[0]+(1.23313354594)*x_ref[1]+(0.701202865744)*x_ref[2]
        ref[3,3]=(0.278097672739)*x_ref[0]+(-1.31336623233)*x_ref[1]+(1.69771904282)*x_ref[2]
        ref[3,4]=(0.0982668788821)*x_ref[0]+(-1.11857597802)*x_ref[1]+(-0.0505715204344)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.121576835783)*x[0]+(-0.941342629086)*x[1]
        s[0,0,1]=(0.113281061637)*x[0]+(1.53625452894)*x[1]
        s[0,1,0]=(0.0674434661275)*x[0]+(-0.633539894942)*x[1]
        s[0,1,1]=(0.22586221125)*x[0]+(-0.756374425353)*x[1]
        s[1,0,0]=(0.0826651782713)*x[0]+(-0.78928571118)*x[1]
        s[1,0,1]=(-0.844609848337)*x[0]+(1.15779995507)*x[1]
        s[1,1,0]=(0.137225056583)*x[0]+(-1.03725215253)*x[1]
        s[1,1,1]=(1.66612561218)*x[0]+(-0.243423071065)*x[1]
        s[2,0,0]=(0.0833021990187)*x[0]+(-1.2712750293)*x[1]
        s[2,0,1]=(0.13762583458)*x[0]+(-0.385843401588)*x[1]
        s[2,1,0]=(-0.926228562488)*x[0]+(-0.305689405746)*x[1]
        s[2,1,1]=(0.725733502304)*x[0]+(1.40832913183)*x[1]
        s[3,0,0]=(1.09943065211)*x[0]+(1.13514402138)*x[1]
        s[3,0,1]=(-1.0354232693)*x[0]+(-0.678629246255)*x[1]
        s[3,1,0]=(-0.44364521298)*x[0]+(0.857496257444)*x[1]
        s[3,1,1]=(1.53635223417)*x[0]+(-0.0887540544341)*x[1]
        s[4,0,0]=(0.553400675937)*x[0]+(-0.680780568865)*x[1]
        s[4,0,1]=(-0.00341234496916)*x[0]+(0.115104542523)*x[1]
        s[4,1,0]=(1.07927709685)*x[0]+(-0.270560535373)*x[1]
        s[4,1,1]=(-0.681458204636)*x[0]+(-0.607605729645)*x[1]
        s[5,0,0]=(0.10322587433)*x[0]+(0.982393144196)*x[1]
        s[5,0,1]=(-0.229957227498)*x[0]+(0.940017423086)*x[1]
        s[5,1,0]=(-0.465473033025)*x[0]+(-0.61380542705)*x[1]
        s[5,1,1]=(1.43248132924)*x[0]+(-1.71390853514)*x[1]
        ref[0,0,0]=(0.121576835783)*x_ref[0]+(-0.941342629086)*x_ref[1]
        ref[0,0,1]=(0.113281061637)*x_ref[0]+(1.53625452894)*x_ref[1]
        ref[0,1,0]=(0.0674434661275)*x_ref[0]+(-0.633539894942)*x_ref[1]
        ref[0,1,1]=(0.22586221125)*x_ref[0]+(-0.756374425353)*x_ref[1]
        ref[1,0,0]=(0.0826651782713)*x_ref[0]+(-0.78928571118)*x_ref[1]
        ref[1,0,1]=(-0.844609848337)*x_ref[0]+(1.15779995507)*x_ref[1]
        ref[1,1,0]=(0.137225056583)*x_ref[0]+(-1.03725215253)*x_ref[1]
        ref[1,1,1]=(1.66612561218)*x_ref[0]+(-0.243423071065)*x_ref[1]
        ref[2,0,0]=(0.0833021990187)*x_ref[0]+(-1.2712750293)*x_ref[1]
        ref[2,0,1]=(0.13762583458)*x_ref[0]+(-0.385843401588)*x_ref[1]
        ref[2,1,0]=(-0.926228562488)*x_ref[0]+(-0.305689405746)*x_ref[1]
        ref[2,1,1]=(0.725733502304)*x_ref[0]+(1.40832913183)*x_ref[1]
        ref[3,0,0]=(1.09943065211)*x_ref[0]+(1.13514402138)*x_ref[1]
        ref[3,0,1]=(-1.0354232693)*x_ref[0]+(-0.678629246255)*x_ref[1]
        ref[3,1,0]=(-0.44364521298)*x_ref[0]+(0.857496257444)*x_ref[1]
        ref[3,1,1]=(1.53635223417)*x_ref[0]+(-0.0887540544341)*x_ref[1]
        ref[4,0,0]=(0.553400675937)*x_ref[0]+(-0.680780568865)*x_ref[1]
        ref[4,0,1]=(-0.00341234496916)*x_ref[0]+(0.115104542523)*x_ref[1]
        ref[4,1,0]=(1.07927709685)*x_ref[0]+(-0.270560535373)*x_ref[1]
        ref[4,1,1]=(-0.681458204636)*x_ref[0]+(-0.607605729645)*x_ref[1]
        ref[5,0,0]=(0.10322587433)*x_ref[0]+(0.982393144196)*x_ref[1]
        ref[5,0,1]=(-0.229957227498)*x_ref[0]+(0.940017423086)*x_ref[1]
        ref[5,1,0]=(-0.465473033025)*x_ref[0]+(-0.61380542705)*x_ref[1]
        ref[5,1,1]=(1.43248132924)*x_ref[0]+(-1.71390853514)*x_ref[1]
      else:
        s[0,0,0]=(0.663545175027)*x[0]+(-1.05528281336)*x[1]+(-0.25024128376)*x[2]
        s[0,0,1]=(1.10817148343)*x[0]+(0.224734467846)*x[1]+(0.712872841532)*x[2]
        s[0,1,0]=(-0.345072989586)*x[0]+(-1.244678278)*x[1]+(0.103446646879)*x[2]
        s[0,1,1]=(-0.106279741301)*x[0]+(-0.525721333511)*x[1]+(0.401917882446)*x[2]
        s[1,0,0]=(0.838242796285)*x[0]+(-0.706020082817)*x[1]+(0.443650945014)*x[2]
        s[1,0,1]=(-1.16557336842)*x[0]+(0.215324659268)*x[1]+(-0.285210566147)*x[2]
        s[1,1,0]=(-0.141574656707)*x[0]+(1.66054358217)*x[1]+(-0.554616610711)*x[2]
        s[1,1,1]=(0.523682935472)*x[0]+(0.687178896331)*x[1]+(0.875639049678)*x[2]
        s[2,0,0]=(0.585144782395)*x[0]+(0.172063481865)*x[1]+(0.920891875835)*x[2]
        s[2,0,1]=(0.660815435485)*x[0]+(-1.63324849021)*x[1]+(0.0810858621718)*x[2]
        s[2,1,0]=(-0.238166824253)*x[0]+(0.0442487879497)*x[1]+(1.05748755823)*x[2]
        s[2,1,1]=(1.14314890535)*x[0]+(0.63161746042)*x[1]+(1.5024770616)*x[2]
        s[3,0,0]=(-1.33327391952)*x[0]+(1.58312319401)*x[1]+(-1.24202894696)*x[2]
        s[3,0,1]=(-0.0242608819668)*x[0]+(-1.12901018411)*x[1]+(-0.443840381999)*x[2]
        s[3,1,0]=(1.12400540809)*x[0]+(-0.25104247006)*x[1]+(0.55935664575)*x[2]
        s[3,1,1]=(0.976977053811)*x[0]+(1.46134206601)*x[1]+(-0.883829686094)*x[2]
        s[4,0,0]=(0.68585022695)*x[0]+(0.320211070942)*x[1]+(-0.715034754238)*x[2]
        s[4,0,1]=(0.745090118519)*x[0]+(-0.56687124464)*x[1]+(-1.41483630619)*x[2]
        s[4,1,0]=(-0.864645032271)*x[0]+(-0.0852453841156)*x[1]+(-1.02272230917)*x[2]
        s[4,1,1]=(0.243107783262)*x[0]+(-1.76263127813)*x[1]+(-0.0396122398093)*x[2]
        s[5,0,0]=(-1.33890039547)*x[0]+(0.95916460762)*x[1]+(-0.118457193655)*x[2]
        s[5,0,1]=(-1.03526146458)*x[0]+(-0.279038868252)*x[1]+(-0.702378064321)*x[2]
        s[5,1,0]=(-0.434705839506)*x[0]+(0.388520593856)*x[1]+(0.213085443323)*x[2]
        s[5,1,1]=(-0.156276503308)*x[0]+(-0.463824282648)*x[1]+(-0.801704395726)*x[2]
        ref[0,0,0]=(0.663545175027)*x_ref[0]+(-1.05528281336)*x_ref[1]+(-0.25024128376)*x_ref[2]
        ref[0,0,1]=(1.10817148343)*x_ref[0]+(0.224734467846)*x_ref[1]+(0.712872841532)*x_ref[2]
        ref[0,1,0]=(-0.345072989586)*x_ref[0]+(-1.244678278)*x_ref[1]+(0.103446646879)*x_ref[2]
        ref[0,1,1]=(-0.106279741301)*x_ref[0]+(-0.525721333511)*x_ref[1]+(0.401917882446)*x_ref[2]
        ref[1,0,0]=(0.838242796285)*x_ref[0]+(-0.706020082817)*x_ref[1]+(0.443650945014)*x_ref[2]
        ref[1,0,1]=(-1.16557336842)*x_ref[0]+(0.215324659268)*x_ref[1]+(-0.285210566147)*x_ref[2]
        ref[1,1,0]=(-0.141574656707)*x_ref[0]+(1.66054358217)*x_ref[1]+(-0.554616610711)*x_ref[2]
        ref[1,1,1]=(0.523682935472)*x_ref[0]+(0.687178896331)*x_ref[1]+(0.875639049678)*x_ref[2]
        ref[2,0,0]=(0.585144782395)*x_ref[0]+(0.172063481865)*x_ref[1]+(0.920891875835)*x_ref[2]
        ref[2,0,1]=(0.660815435485)*x_ref[0]+(-1.63324849021)*x_ref[1]+(0.0810858621718)*x_ref[2]
        ref[2,1,0]=(-0.238166824253)*x_ref[0]+(0.0442487879497)*x_ref[1]+(1.05748755823)*x_ref[2]
        ref[2,1,1]=(1.14314890535)*x_ref[0]+(0.63161746042)*x_ref[1]+(1.5024770616)*x_ref[2]
        ref[3,0,0]=(-1.33327391952)*x_ref[0]+(1.58312319401)*x_ref[1]+(-1.24202894696)*x_ref[2]
        ref[3,0,1]=(-0.0242608819668)*x_ref[0]+(-1.12901018411)*x_ref[1]+(-0.443840381999)*x_ref[2]
        ref[3,1,0]=(1.12400540809)*x_ref[0]+(-0.25104247006)*x_ref[1]+(0.55935664575)*x_ref[2]
        ref[3,1,1]=(0.976977053811)*x_ref[0]+(1.46134206601)*x_ref[1]+(-0.883829686094)*x_ref[2]
        ref[4,0,0]=(0.68585022695)*x_ref[0]+(0.320211070942)*x_ref[1]+(-0.715034754238)*x_ref[2]
        ref[4,0,1]=(0.745090118519)*x_ref[0]+(-0.56687124464)*x_ref[1]+(-1.41483630619)*x_ref[2]
        ref[4,1,0]=(-0.864645032271)*x_ref[0]+(-0.0852453841156)*x_ref[1]+(-1.02272230917)*x_ref[2]
        ref[4,1,1]=(0.243107783262)*x_ref[0]+(-1.76263127813)*x_ref[1]+(-0.0396122398093)*x_ref[2]
        ref[5,0,0]=(-1.33890039547)*x_ref[0]+(0.95916460762)*x_ref[1]+(-0.118457193655)*x_ref[2]
        ref[5,0,1]=(-1.03526146458)*x_ref[0]+(-0.279038868252)*x_ref[1]+(-0.702378064321)*x_ref[2]
        ref[5,1,0]=(-0.434705839506)*x_ref[0]+(0.388520593856)*x_ref[1]+(0.213085443323)*x_ref[2]
        ref[5,1,1]=(-0.156276503308)*x_ref[0]+(-0.463824282648)*x_ref[1]+(-0.801704395726)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.239181398192)*x[0]+(-0.39363876765)*x[1]
        s[0,0,0,1]=(0.391461093031)*x[0]+(-0.825968486042)*x[1]
        s[0,0,1,0]=(0.327281495381)*x[0]+(-0.201869630768)*x[1]
        s[0,0,1,1]=(-0.0266300089051)*x[0]+(0.0315885163331)*x[1]
        s[0,0,2,0]=(-1.29808531891)*x[0]+(0.908442635353)*x[1]
        s[0,0,2,1]=(-0.0719703714734)*x[0]+(-1.2949593192)*x[1]
        s[0,1,0,0]=(1.56337010697)*x[0]+(0.554963325139)*x[1]
        s[0,1,0,1]=(-0.590177179718)*x[0]+(-0.577417888525)*x[1]
        s[0,1,1,0]=(1.12019940888)*x[0]+(0.515714775221)*x[1]
        s[0,1,1,1]=(1.62187733115)*x[0]+(-0.0496003559467)*x[1]
        s[0,1,2,0]=(1.33693246276)*x[0]+(0.18028194694)*x[1]
        s[0,1,2,1]=(1.197520649)*x[0]+(-0.189794758648)*x[1]
        s[0,2,0,0]=(0.458741675295)*x[0]+(-1.23374520557)*x[1]
        s[0,2,0,1]=(0.122618584159)*x[0]+(1.68218857683)*x[1]
        s[0,2,1,0]=(-0.629416301315)*x[0]+(0.00821950501686)*x[1]
        s[0,2,1,1]=(0.372444097376)*x[0]+(-0.389807082811)*x[1]
        s[0,2,2,0]=(0.453580881811)*x[0]+(0.234675328844)*x[1]
        s[0,2,2,1]=(-0.488447987119)*x[0]+(-0.795290862711)*x[1]
        s[0,3,0,0]=(-0.0558838669873)*x[0]+(-0.123245413784)*x[1]
        s[0,3,0,1]=(1.37779240976)*x[0]+(0.0843495250295)*x[1]
        s[0,3,1,0]=(1.17819744943)*x[0]+(-0.859044697828)*x[1]
        s[0,3,1,1]=(1.25175416429)*x[0]+(-0.727964075982)*x[1]
        s[0,3,2,0]=(0.331480991751)*x[0]+(0.115364502392)*x[1]
        s[0,3,2,1]=(-0.791037073807)*x[0]+(-0.51631676808)*x[1]
        s[0,4,0,0]=(0.884389418081)*x[0]+(1.15006122432)*x[1]
        s[0,4,0,1]=(-0.914023114312)*x[0]+(-0.211608734049)*x[1]
        s[0,4,1,0]=(0.55983523554)*x[0]+(-0.792920109589)*x[1]
        s[0,4,1,1]=(0.70901081021)*x[0]+(0.205352000801)*x[1]
        s[0,4,2,0]=(0.140341122546)*x[0]+(0.633534699565)*x[1]
        s[0,4,2,1]=(-0.399843857722)*x[0]+(0.0360346234753)*x[1]
        s[1,0,0,0]=(0.649005423866)*x[0]+(-0.72958629181)*x[1]
        s[1,0,0,1]=(-1.23313007881)*x[0]+(0.240005909545)*x[1]
        s[1,0,1,0]=(0.817440794961)*x[0]+(0.219684770228)*x[1]
        s[1,0,1,1]=(0.300714219818)*x[0]+(0.339948691074)*x[1]
        s[1,0,2,0]=(0.692412749315)*x[0]+(0.847435101682)*x[1]
        s[1,0,2,1]=(0.650288735017)*x[0]+(1.25808519806)*x[1]
        s[1,1,0,0]=(-0.723761201921)*x[0]+(-0.337875653383)*x[1]
        s[1,1,0,1]=(1.10005713123)*x[0]+(0.465822601087)*x[1]
        s[1,1,1,0]=(0.140014505481)*x[0]+(0.289986937348)*x[1]
        s[1,1,1,1]=(0.0941270067899)*x[0]+(-1.45892231406)*x[1]
        s[1,1,2,0]=(1.04493752023)*x[0]+(-0.474063991258)*x[1]
        s[1,1,2,1]=(0.703284346556)*x[0]+(0.482116021536)*x[1]
        s[1,2,0,0]=(-1.66498217207)*x[0]+(-0.144077385929)*x[1]
        s[1,2,0,1]=(1.50511312044)*x[0]+(1.02587270487)*x[1]
        s[1,2,1,0]=(-1.15095468052)*x[0]+(-0.472480042698)*x[1]
        s[1,2,1,1]=(-0.453604215856)*x[0]+(-1.58856703352)*x[1]
        s[1,2,2,0]=(0.659481843854)*x[0]+(0.649931585882)*x[1]
        s[1,2,2,1]=(0.642785065122)*x[0]+(0.0614989297125)*x[1]
        s[1,3,0,0]=(1.65661008057)*x[0]+(-1.08472986336)*x[1]
        s[1,3,0,1]=(-0.64853562643)*x[0]+(-0.912673933935)*x[1]
        s[1,3,1,0]=(-0.108999021371)*x[0]+(-1.44843200164)*x[1]
        s[1,3,1,1]=(0.568395064875)*x[0]+(-0.471819846301)*x[1]
        s[1,3,2,0]=(-0.801680619008)*x[0]+(0.197249605781)*x[1]
        s[1,3,2,1]=(-0.512763932006)*x[0]+(-1.03221569621)*x[1]
        s[1,4,0,0]=(0.54142249246)*x[0]+(-1.28623467967)*x[1]
        s[1,4,0,1]=(1.19094828259)*x[0]+(0.58814933754)*x[1]
        s[1,4,1,0]=(-0.203699090085)*x[0]+(0.839194924785)*x[1]
        s[1,4,1,1]=(-1.73647778175)*x[0]+(1.26399102178)*x[1]
        s[1,4,2,0]=(0.688685884338)*x[0]+(-0.0441772317601)*x[1]
        s[1,4,2,1]=(-0.811158497264)*x[0]+(-0.663923769116)*x[1]
        s[2,0,0,0]=(0.237794148299)*x[0]+(0.125329413285)*x[1]
        s[2,0,0,1]=(-1.40518934715)*x[0]+(1.22320828318)*x[1]
        s[2,0,1,0]=(0.223901464538)*x[0]+(-0.00989463351767)*x[1]
        s[2,0,1,1]=(-0.912246132725)*x[0]+(0.686091520129)*x[1]
        s[2,0,2,0]=(0.181752601691)*x[0]+(-1.12779553611)*x[1]
        s[2,0,2,1]=(-0.061707870206)*x[0]+(-1.3429678748)*x[1]
        s[2,1,0,0]=(1.56396877322)*x[0]+(-1.05119596867)*x[1]
        s[2,1,0,1]=(0.929556514862)*x[0]+(0.406040622531)*x[1]
        s[2,1,1,0]=(-0.509355797822)*x[0]+(0.0605066527971)*x[1]
        s[2,1,1,1]=(-0.528244794646)*x[0]+(0.0388070048758)*x[1]
        s[2,1,2,0]=(-0.200007488719)*x[0]+(-0.711333050321)*x[1]
        s[2,1,2,1]=(0.175160525472)*x[0]+(-1.24343736284)*x[1]
        s[2,2,0,0]=(0.894960105109)*x[0]+(-0.747881639384)*x[1]
        s[2,2,0,1]=(-0.0894204005642)*x[0]+(-0.694792125757)*x[1]
        s[2,2,1,0]=(0.172037435959)*x[0]+(1.29233768559)*x[1]
        s[2,2,1,1]=(-0.312372601854)*x[0]+(-1.35937909779)*x[1]
        s[2,2,2,0]=(0.323370754787)*x[0]+(-0.0601280403525)*x[1]
        s[2,2,2,1]=(0.648154329469)*x[0]+(-0.976168040714)*x[1]
        s[2,3,0,0]=(0.178127534181)*x[0]+(-0.451322025537)*x[1]
        s[2,3,0,1]=(-1.88695440231)*x[0]+(0.164327427196)*x[1]
        s[2,3,1,0]=(0.543479878129)*x[0]+(-0.553094838979)*x[1]
        s[2,3,1,1]=(-0.745599057313)*x[0]+(0.710208052579)*x[1]
        s[2,3,2,0]=(-0.262128106404)*x[0]+(0.839008375922)*x[1]
        s[2,3,2,1]=(-1.20847299902)*x[0]+(-0.321234640326)*x[1]
        s[2,4,0,0]=(-1.24374530208)*x[0]+(-0.961997778937)*x[1]
        s[2,4,0,1]=(-0.0306817051594)*x[0]+(-0.127120502282)*x[1]
        s[2,4,1,0]=(1.25878772522)*x[0]+(1.04432409904)*x[1]
        s[2,4,1,1]=(0.271036209294)*x[0]+(-1.19597311792)*x[1]
        s[2,4,2,0]=(-0.073359480179)*x[0]+(-0.31523273696)*x[1]
        s[2,4,2,1]=(0.738924603079)*x[0]+(0.251467051457)*x[1]
        s[3,0,0,0]=(-0.257569738752)*x[0]+(-1.51373140433)*x[1]
        s[3,0,0,1]=(-0.0113459170081)*x[0]+(-0.356375170129)*x[1]
        s[3,0,1,0]=(-0.290333097834)*x[0]+(-1.56956309356)*x[1]
        s[3,0,1,1]=(-1.0806247672)*x[0]+(-0.376429365315)*x[1]
        s[3,0,2,0]=(-1.32161176104)*x[0]+(0.784442685433)*x[1]
        s[3,0,2,1]=(-1.00382464524)*x[0]+(-0.571979697031)*x[1]
        s[3,1,0,0]=(0.140827544385)*x[0]+(1.3690229529)*x[1]
        s[3,1,0,1]=(0.315011101794)*x[0]+(-1.0263379129)*x[1]
        s[3,1,1,0]=(-0.478940977535)*x[0]+(0.718456942414)*x[1]
        s[3,1,1,1]=(-0.655873236005)*x[0]+(-0.148701297422)*x[1]
        s[3,1,2,0]=(0.56126503141)*x[0]+(1.68526215251)*x[1]
        s[3,1,2,1]=(-0.337284606646)*x[0]+(-0.197295432114)*x[1]
        s[3,2,0,0]=(-0.5979320732)*x[0]+(-0.0600053477444)*x[1]
        s[3,2,0,1]=(-1.20531807175)*x[0]+(-0.148853538473)*x[1]
        s[3,2,1,0]=(0.313468114426)*x[0]+(-0.436912848514)*x[1]
        s[3,2,1,1]=(-0.544279507407)*x[0]+(0.189063470736)*x[1]
        s[3,2,2,0]=(-0.551741475397)*x[0]+(0.196370248113)*x[1]
        s[3,2,2,1]=(-0.0216545656433)*x[0]+(0.23669839429)*x[1]
        s[3,3,0,0]=(-0.406652896064)*x[0]+(0.00397676360081)*x[1]
        s[3,3,0,1]=(-0.305847961815)*x[0]+(-1.2563276235)*x[1]
        s[3,3,1,0]=(-0.450621831697)*x[0]+(-0.213476321469)*x[1]
        s[3,3,1,1]=(1.20548652879)*x[0]+(-0.286503833918)*x[1]
        s[3,3,2,0]=(0.516344981347)*x[0]+(-0.343787067175)*x[1]
        s[3,3,2,1]=(-0.01938029713)*x[0]+(-0.779285558404)*x[1]
        s[3,4,0,0]=(-0.406233598349)*x[0]+(-0.498045431333)*x[1]
        s[3,4,0,1]=(-0.790577593976)*x[0]+(1.07251305331)*x[1]
        s[3,4,1,0]=(0.510569891955)*x[0]+(0.526099613153)*x[1]
        s[3,4,1,1]=(0.150065560826)*x[0]+(0.966145292988)*x[1]
        s[3,4,2,0]=(0.718402792955)*x[0]+(-0.430416841145)*x[1]
        s[3,4,2,1]=(-0.430240857917)*x[0]+(-0.0679614650591)*x[1]
        ref[0,0,0,0]=(0.239181398192)*x_ref[0]+(-0.39363876765)*x_ref[1]
        ref[0,0,0,1]=(0.391461093031)*x_ref[0]+(-0.825968486042)*x_ref[1]
        ref[0,0,1,0]=(0.327281495381)*x_ref[0]+(-0.201869630768)*x_ref[1]
        ref[0,0,1,1]=(-0.0266300089051)*x_ref[0]+(0.0315885163331)*x_ref[1]
        ref[0,0,2,0]=(-1.29808531891)*x_ref[0]+(0.908442635353)*x_ref[1]
        ref[0,0,2,1]=(-0.0719703714734)*x_ref[0]+(-1.2949593192)*x_ref[1]
        ref[0,1,0,0]=(1.56337010697)*x_ref[0]+(0.554963325139)*x_ref[1]
        ref[0,1,0,1]=(-0.590177179718)*x_ref[0]+(-0.577417888525)*x_ref[1]
        ref[0,1,1,0]=(1.12019940888)*x_ref[0]+(0.515714775221)*x_ref[1]
        ref[0,1,1,1]=(1.62187733115)*x_ref[0]+(-0.0496003559467)*x_ref[1]
        ref[0,1,2,0]=(1.33693246276)*x_ref[0]+(0.18028194694)*x_ref[1]
        ref[0,1,2,1]=(1.197520649)*x_ref[0]+(-0.189794758648)*x_ref[1]
        ref[0,2,0,0]=(0.458741675295)*x_ref[0]+(-1.23374520557)*x_ref[1]
        ref[0,2,0,1]=(0.122618584159)*x_ref[0]+(1.68218857683)*x_ref[1]
        ref[0,2,1,0]=(-0.629416301315)*x_ref[0]+(0.00821950501686)*x_ref[1]
        ref[0,2,1,1]=(0.372444097376)*x_ref[0]+(-0.389807082811)*x_ref[1]
        ref[0,2,2,0]=(0.453580881811)*x_ref[0]+(0.234675328844)*x_ref[1]
        ref[0,2,2,1]=(-0.488447987119)*x_ref[0]+(-0.795290862711)*x_ref[1]
        ref[0,3,0,0]=(-0.0558838669873)*x_ref[0]+(-0.123245413784)*x_ref[1]
        ref[0,3,0,1]=(1.37779240976)*x_ref[0]+(0.0843495250295)*x_ref[1]
        ref[0,3,1,0]=(1.17819744943)*x_ref[0]+(-0.859044697828)*x_ref[1]
        ref[0,3,1,1]=(1.25175416429)*x_ref[0]+(-0.727964075982)*x_ref[1]
        ref[0,3,2,0]=(0.331480991751)*x_ref[0]+(0.115364502392)*x_ref[1]
        ref[0,3,2,1]=(-0.791037073807)*x_ref[0]+(-0.51631676808)*x_ref[1]
        ref[0,4,0,0]=(0.884389418081)*x_ref[0]+(1.15006122432)*x_ref[1]
        ref[0,4,0,1]=(-0.914023114312)*x_ref[0]+(-0.211608734049)*x_ref[1]
        ref[0,4,1,0]=(0.55983523554)*x_ref[0]+(-0.792920109589)*x_ref[1]
        ref[0,4,1,1]=(0.70901081021)*x_ref[0]+(0.205352000801)*x_ref[1]
        ref[0,4,2,0]=(0.140341122546)*x_ref[0]+(0.633534699565)*x_ref[1]
        ref[0,4,2,1]=(-0.399843857722)*x_ref[0]+(0.0360346234753)*x_ref[1]
        ref[1,0,0,0]=(0.649005423866)*x_ref[0]+(-0.72958629181)*x_ref[1]
        ref[1,0,0,1]=(-1.23313007881)*x_ref[0]+(0.240005909545)*x_ref[1]
        ref[1,0,1,0]=(0.817440794961)*x_ref[0]+(0.219684770228)*x_ref[1]
        ref[1,0,1,1]=(0.300714219818)*x_ref[0]+(0.339948691074)*x_ref[1]
        ref[1,0,2,0]=(0.692412749315)*x_ref[0]+(0.847435101682)*x_ref[1]
        ref[1,0,2,1]=(0.650288735017)*x_ref[0]+(1.25808519806)*x_ref[1]
        ref[1,1,0,0]=(-0.723761201921)*x_ref[0]+(-0.337875653383)*x_ref[1]
        ref[1,1,0,1]=(1.10005713123)*x_ref[0]+(0.465822601087)*x_ref[1]
        ref[1,1,1,0]=(0.140014505481)*x_ref[0]+(0.289986937348)*x_ref[1]
        ref[1,1,1,1]=(0.0941270067899)*x_ref[0]+(-1.45892231406)*x_ref[1]
        ref[1,1,2,0]=(1.04493752023)*x_ref[0]+(-0.474063991258)*x_ref[1]
        ref[1,1,2,1]=(0.703284346556)*x_ref[0]+(0.482116021536)*x_ref[1]
        ref[1,2,0,0]=(-1.66498217207)*x_ref[0]+(-0.144077385929)*x_ref[1]
        ref[1,2,0,1]=(1.50511312044)*x_ref[0]+(1.02587270487)*x_ref[1]
        ref[1,2,1,0]=(-1.15095468052)*x_ref[0]+(-0.472480042698)*x_ref[1]
        ref[1,2,1,1]=(-0.453604215856)*x_ref[0]+(-1.58856703352)*x_ref[1]
        ref[1,2,2,0]=(0.659481843854)*x_ref[0]+(0.649931585882)*x_ref[1]
        ref[1,2,2,1]=(0.642785065122)*x_ref[0]+(0.0614989297125)*x_ref[1]
        ref[1,3,0,0]=(1.65661008057)*x_ref[0]+(-1.08472986336)*x_ref[1]
        ref[1,3,0,1]=(-0.64853562643)*x_ref[0]+(-0.912673933935)*x_ref[1]
        ref[1,3,1,0]=(-0.108999021371)*x_ref[0]+(-1.44843200164)*x_ref[1]
        ref[1,3,1,1]=(0.568395064875)*x_ref[0]+(-0.471819846301)*x_ref[1]
        ref[1,3,2,0]=(-0.801680619008)*x_ref[0]+(0.197249605781)*x_ref[1]
        ref[1,3,2,1]=(-0.512763932006)*x_ref[0]+(-1.03221569621)*x_ref[1]
        ref[1,4,0,0]=(0.54142249246)*x_ref[0]+(-1.28623467967)*x_ref[1]
        ref[1,4,0,1]=(1.19094828259)*x_ref[0]+(0.58814933754)*x_ref[1]
        ref[1,4,1,0]=(-0.203699090085)*x_ref[0]+(0.839194924785)*x_ref[1]
        ref[1,4,1,1]=(-1.73647778175)*x_ref[0]+(1.26399102178)*x_ref[1]
        ref[1,4,2,0]=(0.688685884338)*x_ref[0]+(-0.0441772317601)*x_ref[1]
        ref[1,4,2,1]=(-0.811158497264)*x_ref[0]+(-0.663923769116)*x_ref[1]
        ref[2,0,0,0]=(0.237794148299)*x_ref[0]+(0.125329413285)*x_ref[1]
        ref[2,0,0,1]=(-1.40518934715)*x_ref[0]+(1.22320828318)*x_ref[1]
        ref[2,0,1,0]=(0.223901464538)*x_ref[0]+(-0.00989463351767)*x_ref[1]
        ref[2,0,1,1]=(-0.912246132725)*x_ref[0]+(0.686091520129)*x_ref[1]
        ref[2,0,2,0]=(0.181752601691)*x_ref[0]+(-1.12779553611)*x_ref[1]
        ref[2,0,2,1]=(-0.061707870206)*x_ref[0]+(-1.3429678748)*x_ref[1]
        ref[2,1,0,0]=(1.56396877322)*x_ref[0]+(-1.05119596867)*x_ref[1]
        ref[2,1,0,1]=(0.929556514862)*x_ref[0]+(0.406040622531)*x_ref[1]
        ref[2,1,1,0]=(-0.509355797822)*x_ref[0]+(0.0605066527971)*x_ref[1]
        ref[2,1,1,1]=(-0.528244794646)*x_ref[0]+(0.0388070048758)*x_ref[1]
        ref[2,1,2,0]=(-0.200007488719)*x_ref[0]+(-0.711333050321)*x_ref[1]
        ref[2,1,2,1]=(0.175160525472)*x_ref[0]+(-1.24343736284)*x_ref[1]
        ref[2,2,0,0]=(0.894960105109)*x_ref[0]+(-0.747881639384)*x_ref[1]
        ref[2,2,0,1]=(-0.0894204005642)*x_ref[0]+(-0.694792125757)*x_ref[1]
        ref[2,2,1,0]=(0.172037435959)*x_ref[0]+(1.29233768559)*x_ref[1]
        ref[2,2,1,1]=(-0.312372601854)*x_ref[0]+(-1.35937909779)*x_ref[1]
        ref[2,2,2,0]=(0.323370754787)*x_ref[0]+(-0.0601280403525)*x_ref[1]
        ref[2,2,2,1]=(0.648154329469)*x_ref[0]+(-0.976168040714)*x_ref[1]
        ref[2,3,0,0]=(0.178127534181)*x_ref[0]+(-0.451322025537)*x_ref[1]
        ref[2,3,0,1]=(-1.88695440231)*x_ref[0]+(0.164327427196)*x_ref[1]
        ref[2,3,1,0]=(0.543479878129)*x_ref[0]+(-0.553094838979)*x_ref[1]
        ref[2,3,1,1]=(-0.745599057313)*x_ref[0]+(0.710208052579)*x_ref[1]
        ref[2,3,2,0]=(-0.262128106404)*x_ref[0]+(0.839008375922)*x_ref[1]
        ref[2,3,2,1]=(-1.20847299902)*x_ref[0]+(-0.321234640326)*x_ref[1]
        ref[2,4,0,0]=(-1.24374530208)*x_ref[0]+(-0.961997778937)*x_ref[1]
        ref[2,4,0,1]=(-0.0306817051594)*x_ref[0]+(-0.127120502282)*x_ref[1]
        ref[2,4,1,0]=(1.25878772522)*x_ref[0]+(1.04432409904)*x_ref[1]
        ref[2,4,1,1]=(0.271036209294)*x_ref[0]+(-1.19597311792)*x_ref[1]
        ref[2,4,2,0]=(-0.073359480179)*x_ref[0]+(-0.31523273696)*x_ref[1]
        ref[2,4,2,1]=(0.738924603079)*x_ref[0]+(0.251467051457)*x_ref[1]
        ref[3,0,0,0]=(-0.257569738752)*x_ref[0]+(-1.51373140433)*x_ref[1]
        ref[3,0,0,1]=(-0.0113459170081)*x_ref[0]+(-0.356375170129)*x_ref[1]
        ref[3,0,1,0]=(-0.290333097834)*x_ref[0]+(-1.56956309356)*x_ref[1]
        ref[3,0,1,1]=(-1.0806247672)*x_ref[0]+(-0.376429365315)*x_ref[1]
        ref[3,0,2,0]=(-1.32161176104)*x_ref[0]+(0.784442685433)*x_ref[1]
        ref[3,0,2,1]=(-1.00382464524)*x_ref[0]+(-0.571979697031)*x_ref[1]
        ref[3,1,0,0]=(0.140827544385)*x_ref[0]+(1.3690229529)*x_ref[1]
        ref[3,1,0,1]=(0.315011101794)*x_ref[0]+(-1.0263379129)*x_ref[1]
        ref[3,1,1,0]=(-0.478940977535)*x_ref[0]+(0.718456942414)*x_ref[1]
        ref[3,1,1,1]=(-0.655873236005)*x_ref[0]+(-0.148701297422)*x_ref[1]
        ref[3,1,2,0]=(0.56126503141)*x_ref[0]+(1.68526215251)*x_ref[1]
        ref[3,1,2,1]=(-0.337284606646)*x_ref[0]+(-0.197295432114)*x_ref[1]
        ref[3,2,0,0]=(-0.5979320732)*x_ref[0]+(-0.0600053477444)*x_ref[1]
        ref[3,2,0,1]=(-1.20531807175)*x_ref[0]+(-0.148853538473)*x_ref[1]
        ref[3,2,1,0]=(0.313468114426)*x_ref[0]+(-0.436912848514)*x_ref[1]
        ref[3,2,1,1]=(-0.544279507407)*x_ref[0]+(0.189063470736)*x_ref[1]
        ref[3,2,2,0]=(-0.551741475397)*x_ref[0]+(0.196370248113)*x_ref[1]
        ref[3,2,2,1]=(-0.0216545656433)*x_ref[0]+(0.23669839429)*x_ref[1]
        ref[3,3,0,0]=(-0.406652896064)*x_ref[0]+(0.00397676360081)*x_ref[1]
        ref[3,3,0,1]=(-0.305847961815)*x_ref[0]+(-1.2563276235)*x_ref[1]
        ref[3,3,1,0]=(-0.450621831697)*x_ref[0]+(-0.213476321469)*x_ref[1]
        ref[3,3,1,1]=(1.20548652879)*x_ref[0]+(-0.286503833918)*x_ref[1]
        ref[3,3,2,0]=(0.516344981347)*x_ref[0]+(-0.343787067175)*x_ref[1]
        ref[3,3,2,1]=(-0.01938029713)*x_ref[0]+(-0.779285558404)*x_ref[1]
        ref[3,4,0,0]=(-0.406233598349)*x_ref[0]+(-0.498045431333)*x_ref[1]
        ref[3,4,0,1]=(-0.790577593976)*x_ref[0]+(1.07251305331)*x_ref[1]
        ref[3,4,1,0]=(0.510569891955)*x_ref[0]+(0.526099613153)*x_ref[1]
        ref[3,4,1,1]=(0.150065560826)*x_ref[0]+(0.966145292988)*x_ref[1]
        ref[3,4,2,0]=(0.718402792955)*x_ref[0]+(-0.430416841145)*x_ref[1]
        ref[3,4,2,1]=(-0.430240857917)*x_ref[0]+(-0.0679614650591)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.512691010939)*x[0]+(-1.31295640543)*x[1]+(1.45426060088)*x[2]
        s[0,0,0,1]=(-0.57748372698)*x[0]+(0.215297337003)*x[1]+(0.0853439946124)*x[2]
        s[0,0,1,0]=(-0.785306554604)*x[0]+(-0.0274911695421)*x[1]+(0.693440767035)*x[2]
        s[0,0,1,1]=(-0.56932407873)*x[0]+(-0.981593188827)*x[1]+(0.87532777027)*x[2]
        s[0,0,2,0]=(0.145925104392)*x[0]+(1.45139125753)*x[1]+(1.25041746183)*x[2]
        s[0,0,2,1]=(-1.70110072582)*x[0]+(-0.0335136921878)*x[1]+(0.276204889346)*x[2]
        s[0,1,0,0]=(-0.415026748605)*x[0]+(-0.322339100074)*x[1]+(1.1050553524)*x[2]
        s[0,1,0,1]=(0.267336702938)*x[0]+(-1.44498697586)*x[1]+(0.600455793139)*x[2]
        s[0,1,1,0]=(-0.46618843993)*x[0]+(0.318230555457)*x[1]+(0.281788961443)*x[2]
        s[0,1,1,1]=(-1.12963168407)*x[0]+(-0.661100272253)*x[1]+(0.293642402995)*x[2]
        s[0,1,2,0]=(-0.0041407279814)*x[0]+(0.330916770241)*x[1]+(1.30265962609)*x[2]
        s[0,1,2,1]=(1.37513192104)*x[0]+(-1.39837371617)*x[1]+(-1.23490743489)*x[2]
        s[0,2,0,0]=(0.240920374522)*x[0]+(1.30092830627)*x[1]+(0.189720240056)*x[2]
        s[0,2,0,1]=(0.741337711467)*x[0]+(0.00893414596869)*x[1]+(1.22438764071)*x[2]
        s[0,2,1,0]=(1.0524903095)*x[0]+(-1.00088524111)*x[1]+(0.985885896119)*x[2]
        s[0,2,1,1]=(-0.252168964788)*x[0]+(1.2685968377)*x[1]+(-0.025373437998)*x[2]
        s[0,2,2,0]=(0.728523707179)*x[0]+(0.00656996105623)*x[1]+(1.85098622635)*x[2]
        s[0,2,2,1]=(0.193123464463)*x[0]+(-0.426452202023)*x[1]+(-0.144471734372)*x[2]
        s[0,3,0,0]=(0.446506355449)*x[0]+(0.379921574537)*x[1]+(-0.303150079281)*x[2]
        s[0,3,0,1]=(0.455373734156)*x[0]+(0.134600165673)*x[1]+(-0.604957024603)*x[2]
        s[0,3,1,0]=(-1.63808806962)*x[0]+(1.02897840042)*x[1]+(1.18123170106)*x[2]
        s[0,3,1,1]=(0.587448315418)*x[0]+(0.0991725296547)*x[1]+(1.33682286695)*x[2]
        s[0,3,2,0]=(-1.12085613449)*x[0]+(0.454771039944)*x[1]+(-0.131089499156)*x[2]
        s[0,3,2,1]=(-0.351021639018)*x[0]+(0.47905316732)*x[1]+(-1.07449842534)*x[2]
        s[0,4,0,0]=(0.232526828642)*x[0]+(1.33907066655)*x[1]+(-0.699701401092)*x[2]
        s[0,4,0,1]=(0.244905848535)*x[0]+(-0.799313917773)*x[1]+(-0.16513296799)*x[2]
        s[0,4,1,0]=(-1.10152273815)*x[0]+(0.898176161732)*x[1]+(-0.101734265033)*x[2]
        s[0,4,1,1]=(-0.0804598444811)*x[0]+(-0.77819195606)*x[1]+(0.277776761782)*x[2]
        s[0,4,2,0]=(1.05979692744)*x[0]+(-0.195285573862)*x[1]+(-0.0920717991265)*x[2]
        s[0,4,2,1]=(0.44924595646)*x[0]+(0.409652246508)*x[1]+(-0.0705754685165)*x[2]
        s[1,0,0,0]=(1.14991675937)*x[0]+(-0.435947647215)*x[1]+(-0.45614944131)*x[2]
        s[1,0,0,1]=(0.306890350036)*x[0]+(-1.4400848018)*x[1]+(-0.744474442385)*x[2]
        s[1,0,1,0]=(1.07930565176)*x[0]+(-0.273335650157)*x[1]+(-0.230960378838)*x[2]
        s[1,0,1,1]=(-0.97797563017)*x[0]+(-1.3159717465)*x[1]+(-0.357902208089)*x[2]
        s[1,0,2,0]=(-1.02063328249)*x[0]+(-1.18907185687)*x[1]+(-0.509296457902)*x[2]
        s[1,0,2,1]=(-0.508446666232)*x[0]+(-0.69279963667)*x[1]+(-0.497990637708)*x[2]
        s[1,1,0,0]=(-0.476141863054)*x[0]+(-0.180458121777)*x[1]+(-0.318855084826)*x[2]
        s[1,1,0,1]=(0.480566371651)*x[0]+(-1.74196768367)*x[1]+(1.23057574863)*x[2]
        s[1,1,1,0]=(-0.0213138624722)*x[0]+(-0.0846532154764)*x[1]+(-0.61169792023)*x[2]
        s[1,1,1,1]=(-0.389381507111)*x[0]+(0.436088515729)*x[1]+(-0.603019053837)*x[2]
        s[1,1,2,0]=(-1.24478786129)*x[0]+(0.908984432701)*x[1]+(-1.49483946407)*x[2]
        s[1,1,2,1]=(0.866266264162)*x[0]+(-0.190579076131)*x[1]+(-1.63524384216)*x[2]
        s[1,2,0,0]=(0.0333832222755)*x[0]+(-0.551591529565)*x[1]+(1.22170194609)*x[2]
        s[1,2,0,1]=(-0.551682390802)*x[0]+(0.223940652339)*x[1]+(0.141063634211)*x[2]
        s[1,2,1,0]=(1.17072817252)*x[0]+(0.28318356571)*x[1]+(-0.788096581969)*x[2]
        s[1,2,1,1]=(-1.01576849129)*x[0]+(-0.502669485722)*x[1]+(-0.966677100703)*x[2]
        s[1,2,2,0]=(-1.59522074083)*x[0]+(-1.16016984107)*x[1]+(-0.117030183187)*x[2]
        s[1,2,2,1]=(-0.979165259497)*x[0]+(-0.707460176459)*x[1]+(-0.275081507376)*x[2]
        s[1,3,0,0]=(0.692493088771)*x[0]+(-1.04755874336)*x[1]+(-0.373632326826)*x[2]
        s[1,3,0,1]=(0.717001163418)*x[0]+(-0.782379062328)*x[1]+(0.479776988529)*x[2]
        s[1,3,1,0]=(-0.733062748373)*x[0]+(-0.217972411497)*x[1]+(1.43785098293)*x[2]
        s[1,3,1,1]=(0.186430904908)*x[0]+(0.375218175232)*x[1]+(-0.461668650588)*x[2]
        s[1,3,2,0]=(-1.12951451301)*x[0]+(-1.07034955937)*x[1]+(-0.59888614247)*x[2]
        s[1,3,2,1]=(1.34023062833)*x[0]+(-0.520544300692)*x[1]+(-0.363001059538)*x[2]
        s[1,4,0,0]=(0.42295611834)*x[0]+(-0.718852108989)*x[1]+(-1.61995684533)*x[2]
        s[1,4,0,1]=(-0.422467947319)*x[0]+(-0.209334480945)*x[1]+(1.59579413592)*x[2]
        s[1,4,1,0]=(0.387543472208)*x[0]+(-0.0579909254613)*x[1]+(-1.03104850867)*x[2]
        s[1,4,1,1]=(-1.15893685519)*x[0]+(-0.971412946627)*x[1]+(-0.583440884462)*x[2]
        s[1,4,2,0]=(0.554042723162)*x[0]+(0.082633597735)*x[1]+(0.990177730288)*x[2]
        s[1,4,2,1]=(0.150114027051)*x[0]+(0.479664757576)*x[1]+(-1.5769513289)*x[2]
        s[2,0,0,0]=(1.33963588901)*x[0]+(0.146971184264)*x[1]+(-1.29921529821)*x[2]
        s[2,0,0,1]=(0.0115342792238)*x[0]+(0.0451643711892)*x[1]+(-0.656407384758)*x[2]
        s[2,0,1,0]=(0.175368903926)*x[0]+(-0.40573949846)*x[1]+(0.295387861218)*x[2]
        s[2,0,1,1]=(0.531254864472)*x[0]+(-0.536193761259)*x[1]+(-1.21752823876)*x[2]
        s[2,0,2,0]=(-0.362296220089)*x[0]+(0.378651497918)*x[1]+(-0.535206713131)*x[2]
        s[2,0,2,1]=(-0.85728549969)*x[0]+(0.233179293609)*x[1]+(0.345412963315)*x[2]
        s[2,1,0,0]=(1.01059063177)*x[0]+(0.808116187589)*x[1]+(0.468588054081)*x[2]
        s[2,1,0,1]=(-1.17212735529)*x[0]+(-0.871560017197)*x[1]+(1.10553749428)*x[2]
        s[2,1,1,0]=(1.00177793873)*x[0]+(-0.573447532545)*x[1]+(-0.80780539944)*x[2]
        s[2,1,1,1]=(-0.101912168689)*x[0]+(0.0054724970873)*x[1]+(0.873799338046)*x[2]
        s[2,1,2,0]=(0.162601365201)*x[0]+(-0.508620489728)*x[1]+(0.114861348434)*x[2]
        s[2,1,2,1]=(-1.48812556245)*x[0]+(0.29841054692)*x[1]+(0.773432180321)*x[2]
        s[2,2,0,0]=(0.0177469158963)*x[0]+(0.140208281265)*x[1]+(1.12393579682)*x[2]
        s[2,2,0,1]=(0.127835907966)*x[0]+(0.150072336346)*x[1]+(1.59437274191)*x[2]
        s[2,2,1,0]=(-0.549906307264)*x[0]+(-0.572101026775)*x[1]+(-0.747028158715)*x[2]
        s[2,2,1,1]=(0.134272548587)*x[0]+(0.866777056589)*x[1]+(-0.0658173600046)*x[2]
        s[2,2,2,0]=(0.160006947347)*x[0]+(-0.940281662941)*x[1]+(0.671280858545)*x[2]
        s[2,2,2,1]=(-1.55550646869)*x[0]+(0.0485277897059)*x[1]+(1.02027216264)*x[2]
        s[2,3,0,0]=(0.0931028968111)*x[0]+(0.194744643733)*x[1]+(1.7368246452)*x[2]
        s[2,3,0,1]=(1.66077510679)*x[0]+(-0.342785893413)*x[1]+(-0.278302501974)*x[2]
        s[2,3,1,0]=(-0.0719600278272)*x[0]+(0.898388137955)*x[1]+(-0.110811429033)*x[2]
        s[2,3,1,1]=(-0.781997247811)*x[0]+(-1.32497050043)*x[1]+(0.544850665961)*x[2]
        s[2,3,2,0]=(0.0290847482287)*x[0]+(0.895295560038)*x[1]+(-1.68555124085)*x[2]
        s[2,3,2,1]=(0.200540179332)*x[0]+(-0.843842814948)*x[1]+(-0.943779465271)*x[2]
        s[2,4,0,0]=(1.08999954494)*x[0]+(-1.32324120166)*x[1]+(-0.0152368215856)*x[2]
        s[2,4,0,1]=(1.02219584162)*x[0]+(0.208263165731)*x[1]+(0.0703408478849)*x[2]
        s[2,4,1,0]=(0.433866479442)*x[0]+(0.327590037731)*x[1]+(0.752498033029)*x[2]
        s[2,4,1,1]=(1.16322078616)*x[0]+(-1.38234503561)*x[1]+(-0.129845447078)*x[2]
        s[2,4,2,0]=(-0.430572860596)*x[0]+(-1.48079933792)*x[1]+(0.310197458762)*x[2]
        s[2,4,2,1]=(-0.0901138166505)*x[0]+(0.130679443905)*x[1]+(-0.482309483824)*x[2]
        s[3,0,0,0]=(0.239787389644)*x[0]+(0.690754646072)*x[1]+(0.745418481522)*x[2]
        s[3,0,0,1]=(0.346448639801)*x[0]+(-0.287823450604)*x[1]+(0.287003737791)*x[2]
        s[3,0,1,0]=(-1.50502949668)*x[0]+(-1.67521500964)*x[1]+(0.31630953667)*x[2]
        s[3,0,1,1]=(-0.594355763268)*x[0]+(-0.262023313233)*x[1]+(-1.41203969641)*x[2]
        s[3,0,2,0]=(-0.0275545216089)*x[0]+(0.367433407785)*x[1]+(-1.420568196)*x[2]
        s[3,0,2,1]=(-1.18797738587)*x[0]+(-0.496775264183)*x[1]+(0.513942381739)*x[2]
        s[3,1,0,0]=(0.475014043678)*x[0]+(-1.2205702601)*x[1]+(1.06353608529)*x[2]
        s[3,1,0,1]=(-0.0685583494755)*x[0]+(-1.49448219897)*x[1]+(-1.23320402959)*x[2]
        s[3,1,1,0]=(-0.216184790554)*x[0]+(0.554046932828)*x[1]+(0.697487187904)*x[2]
        s[3,1,1,1]=(-0.495370181694)*x[0]+(-1.49207400185)*x[1]+(-0.233763280927)*x[2]
        s[3,1,2,0]=(-0.0462888472218)*x[0]+(0.371127867531)*x[1]+(0.534977833221)*x[2]
        s[3,1,2,1]=(-1.27057460029)*x[0]+(-0.572725702684)*x[1]+(-0.871217452212)*x[2]
        s[3,2,0,0]=(-0.208549808264)*x[0]+(0.847294134183)*x[1]+(0.643419233684)*x[2]
        s[3,2,0,1]=(-0.850445146291)*x[0]+(1.19605447923)*x[1]+(-1.52958469348)*x[2]
        s[3,2,1,0]=(0.181672364304)*x[0]+(-1.60804541136)*x[1]+(0.068065187694)*x[2]
        s[3,2,1,1]=(-0.195284925405)*x[0]+(0.330399376496)*x[1]+(-0.453658972198)*x[2]
        s[3,2,2,0]=(1.44390910788)*x[0]+(0.718618445973)*x[1]+(-0.800501828022)*x[2]
        s[3,2,2,1]=(0.491997052311)*x[0]+(-0.202966322046)*x[1]+(1.39185177314)*x[2]
        s[3,3,0,0]=(-0.670818169452)*x[0]+(0.842747476683)*x[1]+(-0.705022077785)*x[2]
        s[3,3,0,1]=(0.729442429399)*x[0]+(-1.47998423588)*x[1]+(1.72626243143)*x[2]
        s[3,3,1,0]=(-1.08179915812)*x[0]+(-0.117953758788)*x[1]+(1.60080937509)*x[2]
        s[3,3,1,1]=(0.110049458343)*x[0]+(0.571461685559)*x[1]+(-0.483581817598)*x[2]
        s[3,3,2,0]=(-0.386163706678)*x[0]+(-0.146336916781)*x[1]+(0.0360311139076)*x[2]
        s[3,3,2,1]=(-0.779748009644)*x[0]+(-0.780523381896)*x[1]+(-1.20157768536)*x[2]
        s[3,4,0,0]=(1.74580569218)*x[0]+(-0.318257555715)*x[1]+(-0.33745202709)*x[2]
        s[3,4,0,1]=(1.1463597084)*x[0]+(0.516488947224)*x[1]+(-0.205035082746)*x[2]
        s[3,4,1,0]=(0.647937211136)*x[0]+(0.574941676843)*x[1]+(-1.00930234083)*x[2]
        s[3,4,1,1]=(-0.79188909394)*x[0]+(0.770198112387)*x[1]+(-0.0146013355132)*x[2]
        s[3,4,2,0]=(-1.70734926495)*x[0]+(-1.01352599249)*x[1]+(0.878522778655)*x[2]
        s[3,4,2,1]=(0.0249166011531)*x[0]+(-0.0714570534569)*x[1]+(1.10212723792)*x[2]
        ref[0,0,0,0]=(-0.512691010939)*x_ref[0]+(-1.31295640543)*x_ref[1]+(1.45426060088)*x_ref[2]
        ref[0,0,0,1]=(-0.57748372698)*x_ref[0]+(0.215297337003)*x_ref[1]+(0.0853439946124)*x_ref[2]
        ref[0,0,1,0]=(-0.785306554604)*x_ref[0]+(-0.0274911695421)*x_ref[1]+(0.693440767035)*x_ref[2]
        ref[0,0,1,1]=(-0.56932407873)*x_ref[0]+(-0.981593188827)*x_ref[1]+(0.87532777027)*x_ref[2]
        ref[0,0,2,0]=(0.145925104392)*x_ref[0]+(1.45139125753)*x_ref[1]+(1.25041746183)*x_ref[2]
        ref[0,0,2,1]=(-1.70110072582)*x_ref[0]+(-0.0335136921878)*x_ref[1]+(0.276204889346)*x_ref[2]
        ref[0,1,0,0]=(-0.415026748605)*x_ref[0]+(-0.322339100074)*x_ref[1]+(1.1050553524)*x_ref[2]
        ref[0,1,0,1]=(0.267336702938)*x_ref[0]+(-1.44498697586)*x_ref[1]+(0.600455793139)*x_ref[2]
        ref[0,1,1,0]=(-0.46618843993)*x_ref[0]+(0.318230555457)*x_ref[1]+(0.281788961443)*x_ref[2]
        ref[0,1,1,1]=(-1.12963168407)*x_ref[0]+(-0.661100272253)*x_ref[1]+(0.293642402995)*x_ref[2]
        ref[0,1,2,0]=(-0.0041407279814)*x_ref[0]+(0.330916770241)*x_ref[1]+(1.30265962609)*x_ref[2]
        ref[0,1,2,1]=(1.37513192104)*x_ref[0]+(-1.39837371617)*x_ref[1]+(-1.23490743489)*x_ref[2]
        ref[0,2,0,0]=(0.240920374522)*x_ref[0]+(1.30092830627)*x_ref[1]+(0.189720240056)*x_ref[2]
        ref[0,2,0,1]=(0.741337711467)*x_ref[0]+(0.00893414596869)*x_ref[1]+(1.22438764071)*x_ref[2]
        ref[0,2,1,0]=(1.0524903095)*x_ref[0]+(-1.00088524111)*x_ref[1]+(0.985885896119)*x_ref[2]
        ref[0,2,1,1]=(-0.252168964788)*x_ref[0]+(1.2685968377)*x_ref[1]+(-0.025373437998)*x_ref[2]
        ref[0,2,2,0]=(0.728523707179)*x_ref[0]+(0.00656996105623)*x_ref[1]+(1.85098622635)*x_ref[2]
        ref[0,2,2,1]=(0.193123464463)*x_ref[0]+(-0.426452202023)*x_ref[1]+(-0.144471734372)*x_ref[2]
        ref[0,3,0,0]=(0.446506355449)*x_ref[0]+(0.379921574537)*x_ref[1]+(-0.303150079281)*x_ref[2]
        ref[0,3,0,1]=(0.455373734156)*x_ref[0]+(0.134600165673)*x_ref[1]+(-0.604957024603)*x_ref[2]
        ref[0,3,1,0]=(-1.63808806962)*x_ref[0]+(1.02897840042)*x_ref[1]+(1.18123170106)*x_ref[2]
        ref[0,3,1,1]=(0.587448315418)*x_ref[0]+(0.0991725296547)*x_ref[1]+(1.33682286695)*x_ref[2]
        ref[0,3,2,0]=(-1.12085613449)*x_ref[0]+(0.454771039944)*x_ref[1]+(-0.131089499156)*x_ref[2]
        ref[0,3,2,1]=(-0.351021639018)*x_ref[0]+(0.47905316732)*x_ref[1]+(-1.07449842534)*x_ref[2]
        ref[0,4,0,0]=(0.232526828642)*x_ref[0]+(1.33907066655)*x_ref[1]+(-0.699701401092)*x_ref[2]
        ref[0,4,0,1]=(0.244905848535)*x_ref[0]+(-0.799313917773)*x_ref[1]+(-0.16513296799)*x_ref[2]
        ref[0,4,1,0]=(-1.10152273815)*x_ref[0]+(0.898176161732)*x_ref[1]+(-0.101734265033)*x_ref[2]
        ref[0,4,1,1]=(-0.0804598444811)*x_ref[0]+(-0.77819195606)*x_ref[1]+(0.277776761782)*x_ref[2]
        ref[0,4,2,0]=(1.05979692744)*x_ref[0]+(-0.195285573862)*x_ref[1]+(-0.0920717991265)*x_ref[2]
        ref[0,4,2,1]=(0.44924595646)*x_ref[0]+(0.409652246508)*x_ref[1]+(-0.0705754685165)*x_ref[2]
        ref[1,0,0,0]=(1.14991675937)*x_ref[0]+(-0.435947647215)*x_ref[1]+(-0.45614944131)*x_ref[2]
        ref[1,0,0,1]=(0.306890350036)*x_ref[0]+(-1.4400848018)*x_ref[1]+(-0.744474442385)*x_ref[2]
        ref[1,0,1,0]=(1.07930565176)*x_ref[0]+(-0.273335650157)*x_ref[1]+(-0.230960378838)*x_ref[2]
        ref[1,0,1,1]=(-0.97797563017)*x_ref[0]+(-1.3159717465)*x_ref[1]+(-0.357902208089)*x_ref[2]
        ref[1,0,2,0]=(-1.02063328249)*x_ref[0]+(-1.18907185687)*x_ref[1]+(-0.509296457902)*x_ref[2]
        ref[1,0,2,1]=(-0.508446666232)*x_ref[0]+(-0.69279963667)*x_ref[1]+(-0.497990637708)*x_ref[2]
        ref[1,1,0,0]=(-0.476141863054)*x_ref[0]+(-0.180458121777)*x_ref[1]+(-0.318855084826)*x_ref[2]
        ref[1,1,0,1]=(0.480566371651)*x_ref[0]+(-1.74196768367)*x_ref[1]+(1.23057574863)*x_ref[2]
        ref[1,1,1,0]=(-0.0213138624722)*x_ref[0]+(-0.0846532154764)*x_ref[1]+(-0.61169792023)*x_ref[2]
        ref[1,1,1,1]=(-0.389381507111)*x_ref[0]+(0.436088515729)*x_ref[1]+(-0.603019053837)*x_ref[2]
        ref[1,1,2,0]=(-1.24478786129)*x_ref[0]+(0.908984432701)*x_ref[1]+(-1.49483946407)*x_ref[2]
        ref[1,1,2,1]=(0.866266264162)*x_ref[0]+(-0.190579076131)*x_ref[1]+(-1.63524384216)*x_ref[2]
        ref[1,2,0,0]=(0.0333832222755)*x_ref[0]+(-0.551591529565)*x_ref[1]+(1.22170194609)*x_ref[2]
        ref[1,2,0,1]=(-0.551682390802)*x_ref[0]+(0.223940652339)*x_ref[1]+(0.141063634211)*x_ref[2]
        ref[1,2,1,0]=(1.17072817252)*x_ref[0]+(0.28318356571)*x_ref[1]+(-0.788096581969)*x_ref[2]
        ref[1,2,1,1]=(-1.01576849129)*x_ref[0]+(-0.502669485722)*x_ref[1]+(-0.966677100703)*x_ref[2]
        ref[1,2,2,0]=(-1.59522074083)*x_ref[0]+(-1.16016984107)*x_ref[1]+(-0.117030183187)*x_ref[2]
        ref[1,2,2,1]=(-0.979165259497)*x_ref[0]+(-0.707460176459)*x_ref[1]+(-0.275081507376)*x_ref[2]
        ref[1,3,0,0]=(0.692493088771)*x_ref[0]+(-1.04755874336)*x_ref[1]+(-0.373632326826)*x_ref[2]
        ref[1,3,0,1]=(0.717001163418)*x_ref[0]+(-0.782379062328)*x_ref[1]+(0.479776988529)*x_ref[2]
        ref[1,3,1,0]=(-0.733062748373)*x_ref[0]+(-0.217972411497)*x_ref[1]+(1.43785098293)*x_ref[2]
        ref[1,3,1,1]=(0.186430904908)*x_ref[0]+(0.375218175232)*x_ref[1]+(-0.461668650588)*x_ref[2]
        ref[1,3,2,0]=(-1.12951451301)*x_ref[0]+(-1.07034955937)*x_ref[1]+(-0.59888614247)*x_ref[2]
        ref[1,3,2,1]=(1.34023062833)*x_ref[0]+(-0.520544300692)*x_ref[1]+(-0.363001059538)*x_ref[2]
        ref[1,4,0,0]=(0.42295611834)*x_ref[0]+(-0.718852108989)*x_ref[1]+(-1.61995684533)*x_ref[2]
        ref[1,4,0,1]=(-0.422467947319)*x_ref[0]+(-0.209334480945)*x_ref[1]+(1.59579413592)*x_ref[2]
        ref[1,4,1,0]=(0.387543472208)*x_ref[0]+(-0.0579909254613)*x_ref[1]+(-1.03104850867)*x_ref[2]
        ref[1,4,1,1]=(-1.15893685519)*x_ref[0]+(-0.971412946627)*x_ref[1]+(-0.583440884462)*x_ref[2]
        ref[1,4,2,0]=(0.554042723162)*x_ref[0]+(0.082633597735)*x_ref[1]+(0.990177730288)*x_ref[2]
        ref[1,4,2,1]=(0.150114027051)*x_ref[0]+(0.479664757576)*x_ref[1]+(-1.5769513289)*x_ref[2]
        ref[2,0,0,0]=(1.33963588901)*x_ref[0]+(0.146971184264)*x_ref[1]+(-1.29921529821)*x_ref[2]
        ref[2,0,0,1]=(0.0115342792238)*x_ref[0]+(0.0451643711892)*x_ref[1]+(-0.656407384758)*x_ref[2]
        ref[2,0,1,0]=(0.175368903926)*x_ref[0]+(-0.40573949846)*x_ref[1]+(0.295387861218)*x_ref[2]
        ref[2,0,1,1]=(0.531254864472)*x_ref[0]+(-0.536193761259)*x_ref[1]+(-1.21752823876)*x_ref[2]
        ref[2,0,2,0]=(-0.362296220089)*x_ref[0]+(0.378651497918)*x_ref[1]+(-0.535206713131)*x_ref[2]
        ref[2,0,2,1]=(-0.85728549969)*x_ref[0]+(0.233179293609)*x_ref[1]+(0.345412963315)*x_ref[2]
        ref[2,1,0,0]=(1.01059063177)*x_ref[0]+(0.808116187589)*x_ref[1]+(0.468588054081)*x_ref[2]
        ref[2,1,0,1]=(-1.17212735529)*x_ref[0]+(-0.871560017197)*x_ref[1]+(1.10553749428)*x_ref[2]
        ref[2,1,1,0]=(1.00177793873)*x_ref[0]+(-0.573447532545)*x_ref[1]+(-0.80780539944)*x_ref[2]
        ref[2,1,1,1]=(-0.101912168689)*x_ref[0]+(0.0054724970873)*x_ref[1]+(0.873799338046)*x_ref[2]
        ref[2,1,2,0]=(0.162601365201)*x_ref[0]+(-0.508620489728)*x_ref[1]+(0.114861348434)*x_ref[2]
        ref[2,1,2,1]=(-1.48812556245)*x_ref[0]+(0.29841054692)*x_ref[1]+(0.773432180321)*x_ref[2]
        ref[2,2,0,0]=(0.0177469158963)*x_ref[0]+(0.140208281265)*x_ref[1]+(1.12393579682)*x_ref[2]
        ref[2,2,0,1]=(0.127835907966)*x_ref[0]+(0.150072336346)*x_ref[1]+(1.59437274191)*x_ref[2]
        ref[2,2,1,0]=(-0.549906307264)*x_ref[0]+(-0.572101026775)*x_ref[1]+(-0.747028158715)*x_ref[2]
        ref[2,2,1,1]=(0.134272548587)*x_ref[0]+(0.866777056589)*x_ref[1]+(-0.0658173600046)*x_ref[2]
        ref[2,2,2,0]=(0.160006947347)*x_ref[0]+(-0.940281662941)*x_ref[1]+(0.671280858545)*x_ref[2]
        ref[2,2,2,1]=(-1.55550646869)*x_ref[0]+(0.0485277897059)*x_ref[1]+(1.02027216264)*x_ref[2]
        ref[2,3,0,0]=(0.0931028968111)*x_ref[0]+(0.194744643733)*x_ref[1]+(1.7368246452)*x_ref[2]
        ref[2,3,0,1]=(1.66077510679)*x_ref[0]+(-0.342785893413)*x_ref[1]+(-0.278302501974)*x_ref[2]
        ref[2,3,1,0]=(-0.0719600278272)*x_ref[0]+(0.898388137955)*x_ref[1]+(-0.110811429033)*x_ref[2]
        ref[2,3,1,1]=(-0.781997247811)*x_ref[0]+(-1.32497050043)*x_ref[1]+(0.544850665961)*x_ref[2]
        ref[2,3,2,0]=(0.0290847482287)*x_ref[0]+(0.895295560038)*x_ref[1]+(-1.68555124085)*x_ref[2]
        ref[2,3,2,1]=(0.200540179332)*x_ref[0]+(-0.843842814948)*x_ref[1]+(-0.943779465271)*x_ref[2]
        ref[2,4,0,0]=(1.08999954494)*x_ref[0]+(-1.32324120166)*x_ref[1]+(-0.0152368215856)*x_ref[2]
        ref[2,4,0,1]=(1.02219584162)*x_ref[0]+(0.208263165731)*x_ref[1]+(0.0703408478849)*x_ref[2]
        ref[2,4,1,0]=(0.433866479442)*x_ref[0]+(0.327590037731)*x_ref[1]+(0.752498033029)*x_ref[2]
        ref[2,4,1,1]=(1.16322078616)*x_ref[0]+(-1.38234503561)*x_ref[1]+(-0.129845447078)*x_ref[2]
        ref[2,4,2,0]=(-0.430572860596)*x_ref[0]+(-1.48079933792)*x_ref[1]+(0.310197458762)*x_ref[2]
        ref[2,4,2,1]=(-0.0901138166505)*x_ref[0]+(0.130679443905)*x_ref[1]+(-0.482309483824)*x_ref[2]
        ref[3,0,0,0]=(0.239787389644)*x_ref[0]+(0.690754646072)*x_ref[1]+(0.745418481522)*x_ref[2]
        ref[3,0,0,1]=(0.346448639801)*x_ref[0]+(-0.287823450604)*x_ref[1]+(0.287003737791)*x_ref[2]
        ref[3,0,1,0]=(-1.50502949668)*x_ref[0]+(-1.67521500964)*x_ref[1]+(0.31630953667)*x_ref[2]
        ref[3,0,1,1]=(-0.594355763268)*x_ref[0]+(-0.262023313233)*x_ref[1]+(-1.41203969641)*x_ref[2]
        ref[3,0,2,0]=(-0.0275545216089)*x_ref[0]+(0.367433407785)*x_ref[1]+(-1.420568196)*x_ref[2]
        ref[3,0,2,1]=(-1.18797738587)*x_ref[0]+(-0.496775264183)*x_ref[1]+(0.513942381739)*x_ref[2]
        ref[3,1,0,0]=(0.475014043678)*x_ref[0]+(-1.2205702601)*x_ref[1]+(1.06353608529)*x_ref[2]
        ref[3,1,0,1]=(-0.0685583494755)*x_ref[0]+(-1.49448219897)*x_ref[1]+(-1.23320402959)*x_ref[2]
        ref[3,1,1,0]=(-0.216184790554)*x_ref[0]+(0.554046932828)*x_ref[1]+(0.697487187904)*x_ref[2]
        ref[3,1,1,1]=(-0.495370181694)*x_ref[0]+(-1.49207400185)*x_ref[1]+(-0.233763280927)*x_ref[2]
        ref[3,1,2,0]=(-0.0462888472218)*x_ref[0]+(0.371127867531)*x_ref[1]+(0.534977833221)*x_ref[2]
        ref[3,1,2,1]=(-1.27057460029)*x_ref[0]+(-0.572725702684)*x_ref[1]+(-0.871217452212)*x_ref[2]
        ref[3,2,0,0]=(-0.208549808264)*x_ref[0]+(0.847294134183)*x_ref[1]+(0.643419233684)*x_ref[2]
        ref[3,2,0,1]=(-0.850445146291)*x_ref[0]+(1.19605447923)*x_ref[1]+(-1.52958469348)*x_ref[2]
        ref[3,2,1,0]=(0.181672364304)*x_ref[0]+(-1.60804541136)*x_ref[1]+(0.068065187694)*x_ref[2]
        ref[3,2,1,1]=(-0.195284925405)*x_ref[0]+(0.330399376496)*x_ref[1]+(-0.453658972198)*x_ref[2]
        ref[3,2,2,0]=(1.44390910788)*x_ref[0]+(0.718618445973)*x_ref[1]+(-0.800501828022)*x_ref[2]
        ref[3,2,2,1]=(0.491997052311)*x_ref[0]+(-0.202966322046)*x_ref[1]+(1.39185177314)*x_ref[2]
        ref[3,3,0,0]=(-0.670818169452)*x_ref[0]+(0.842747476683)*x_ref[1]+(-0.705022077785)*x_ref[2]
        ref[3,3,0,1]=(0.729442429399)*x_ref[0]+(-1.47998423588)*x_ref[1]+(1.72626243143)*x_ref[2]
        ref[3,3,1,0]=(-1.08179915812)*x_ref[0]+(-0.117953758788)*x_ref[1]+(1.60080937509)*x_ref[2]
        ref[3,3,1,1]=(0.110049458343)*x_ref[0]+(0.571461685559)*x_ref[1]+(-0.483581817598)*x_ref[2]
        ref[3,3,2,0]=(-0.386163706678)*x_ref[0]+(-0.146336916781)*x_ref[1]+(0.0360311139076)*x_ref[2]
        ref[3,3,2,1]=(-0.779748009644)*x_ref[0]+(-0.780523381896)*x_ref[1]+(-1.20157768536)*x_ref[2]
        ref[3,4,0,0]=(1.74580569218)*x_ref[0]+(-0.318257555715)*x_ref[1]+(-0.33745202709)*x_ref[2]
        ref[3,4,0,1]=(1.1463597084)*x_ref[0]+(0.516488947224)*x_ref[1]+(-0.205035082746)*x_ref[2]
        ref[3,4,1,0]=(0.647937211136)*x_ref[0]+(0.574941676843)*x_ref[1]+(-1.00930234083)*x_ref[2]
        ref[3,4,1,1]=(-0.79188909394)*x_ref[0]+(0.770198112387)*x_ref[1]+(-0.0146013355132)*x_ref[2]
        ref[3,4,2,0]=(-1.70734926495)*x_ref[0]+(-1.01352599249)*x_ref[1]+(0.878522778655)*x_ref[2]
        ref[3,4,2,1]=(0.0249166011531)*x_ref[0]+(-0.0714570534569)*x_ref[1]+(1.10212723792)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.318228469869)*x[0]+(-0.774132349214)*x[1]
        ref=(0.318228469869)*x_ref[0]+(-0.774132349214)*x_ref[1]
      else:
        s=(0.377557485345)*x[0]+(1.02901239394)*x[1]+(0.038224819288)*x[2]
        ref=(0.377557485345)*x_ref[0]+(1.02901239394)*x_ref[1]+(0.038224819288)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        s[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
        ref[0]=(0.636316116182)*x[0]+(0.162852458484)*x[1]
        ref[1]=(0.566639517171)*x[0]+(-1.09414196398)*x[1]
      else:
        s[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        s[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
        ref[0]=(-0.32246676666)*x[0]+(-1.05372238193)*x[1]+(-0.578494885194)*x[2]
        ref[1]=(1.19735628563)*x[0]+(-0.68060068022)*x[1]+(-0.55344749243)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0621856629646)*x[0]+(0.783319354938)*x[1]
        s[0,1]=(-0.205526239577)*x[0]+(-0.401770179168)*x[1]
        s[0,2]=(0.543085761147)*x[0]+(0.915124776348)*x[1]
        s[0,3]=(1.3114416307)*x[0]+(-0.339977887579)*x[1]
        s[0,4]=(0.19180167645)*x[0]+(0.0744591496582)*x[1]
        s[1,0]=(0.833949474687)*x[0]+(-0.513960535503)*x[1]
        s[1,1]=(1.63749958183)*x[0]+(1.49589084809)*x[1]
        s[1,2]=(0.52690970122)*x[0]+(-0.497651760738)*x[1]
        s[1,3]=(0.157987079101)*x[0]+(0.244658238724)*x[1]
        s[1,4]=(-0.557780706846)*x[0]+(-1.49932552405)*x[1]
        s[2,0]=(-0.00409488968736)*x[0]+(0.366430715843)*x[1]
        s[2,1]=(-0.867654712773)*x[0]+(-0.129712066272)*x[1]
        s[2,2]=(1.33589989715)*x[0]+(-0.135732989085)*x[1]
        s[2,3]=(0.208531325369)*x[0]+(1.3627828974)*x[1]
        s[2,4]=(-0.667299295418)*x[0]+(0.981775718726)*x[1]
        s[3,0]=(1.40151192578)*x[0]+(-0.28272210702)*x[1]
        s[3,1]=(-1.13395485034)*x[0]+(-0.148770918012)*x[1]
        s[3,2]=(0.315288948224)*x[0]+(-0.140365593027)*x[1]
        s[3,3]=(-0.0421809531999)*x[0]+(0.522687748574)*x[1]
        s[3,4]=(0.135043757063)*x[0]+(-1.52489039301)*x[1]
        ref[0,0]=(0.0621856629646)*x_ref[0]+(0.783319354938)*x_ref[1]
        ref[0,1]=(-0.205526239577)*x_ref[0]+(-0.401770179168)*x_ref[1]
        ref[0,2]=(0.543085761147)*x_ref[0]+(0.915124776348)*x_ref[1]
        ref[0,3]=(1.3114416307)*x_ref[0]+(-0.339977887579)*x_ref[1]
        ref[0,4]=(0.19180167645)*x_ref[0]+(0.0744591496582)*x_ref[1]
        ref[1,0]=(0.833949474687)*x_ref[0]+(-0.513960535503)*x_ref[1]
        ref[1,1]=(1.63749958183)*x_ref[0]+(1.49589084809)*x_ref[1]
        ref[1,2]=(0.52690970122)*x_ref[0]+(-0.497651760738)*x_ref[1]
        ref[1,3]=(0.157987079101)*x_ref[0]+(0.244658238724)*x_ref[1]
        ref[1,4]=(-0.557780706846)*x_ref[0]+(-1.49932552405)*x_ref[1]
        ref[2,0]=(-0.00409488968736)*x_ref[0]+(0.366430715843)*x_ref[1]
        ref[2,1]=(-0.867654712773)*x_ref[0]+(-0.129712066272)*x_ref[1]
        ref[2,2]=(1.33589989715)*x_ref[0]+(-0.135732989085)*x_ref[1]
        ref[2,3]=(0.208531325369)*x_ref[0]+(1.3627828974)*x_ref[1]
        ref[2,4]=(-0.667299295418)*x_ref[0]+(0.981775718726)*x_ref[1]
        ref[3,0]=(1.40151192578)*x_ref[0]+(-0.28272210702)*x_ref[1]
        ref[3,1]=(-1.13395485034)*x_ref[0]+(-0.148770918012)*x_ref[1]
        ref[3,2]=(0.315288948224)*x_ref[0]+(-0.140365593027)*x_ref[1]
        ref[3,3]=(-0.0421809531999)*x_ref[0]+(0.522687748574)*x_ref[1]
        ref[3,4]=(0.135043757063)*x_ref[0]+(-1.52489039301)*x_ref[1]
      else:
        s[0,0]=(-1.07097471698)*x[0]+(0.511673191366)*x[1]+(0.986422233856)*x[2]
        s[0,1]=(1.35021866887)*x[0]+(-1.18036897489)*x[1]+(0.143096009704)*x[2]
        s[0,2]=(1.92136475346)*x[0]+(-1.52041779666)*x[1]+(1.50412123174)*x[2]
        s[0,3]=(-1.26349432492)*x[0]+(-1.69090463821)*x[1]+(0.0519848717435)*x[2]
        s[0,4]=(-0.450643290235)*x[0]+(-1.0918592714)*x[1]+(-1.78531176031)*x[2]
        s[1,0]=(1.24575021839)*x[0]+(0.378183524368)*x[1]+(0.164647643767)*x[2]
        s[1,1]=(-0.0164625547855)*x[0]+(-1.06539959746)*x[1]+(-1.5644531354)*x[2]
        s[1,2]=(0.126087223544)*x[0]+(0.876303692726)*x[1]+(1.3521510598)*x[2]
        s[1,3]=(-1.15749467575)*x[0]+(-0.850991452556)*x[1]+(-0.903705777185)*x[2]
        s[1,4]=(-0.680094762766)*x[0]+(0.280093157366)*x[1]+(0.460572306025)*x[2]
        s[2,0]=(-0.201094411285)*x[0]+(0.610240266355)*x[1]+(0.741298609653)*x[2]
        s[2,1]=(0.754412798233)*x[0]+(0.765169641301)*x[1]+(0.522349231028)*x[2]
        s[2,2]=(-0.330239429991)*x[0]+(1.04468899978)*x[1]+(-0.493706328863)*x[2]
        s[2,3]=(-0.107189727407)*x[0]+(-1.57053989855)*x[1]+(1.5864416691)*x[2]
        s[2,4]=(1.78349630817)*x[0]+(0.415138686617)*x[1]+(-0.567656058268)*x[2]
        s[3,0]=(-0.242864411179)*x[0]+(0.342401319249)*x[1]+(-0.963749227433)*x[2]
        s[3,1]=(-0.677695348369)*x[0]+(-1.52126003219)*x[1]+(-0.197904354713)*x[2]
        s[3,2]=(-0.549674242841)*x[0]+(1.23313354594)*x[1]+(0.701202865744)*x[2]
        s[3,3]=(0.278097672739)*x[0]+(-1.31336623233)*x[1]+(1.69771904282)*x[2]
        s[3,4]=(0.0982668788821)*x[0]+(-1.11857597802)*x[1]+(-0.0505715204344)*x[2]
        ref[0,0]=(-1.07097471698)*x_ref[0]+(0.511673191366)*x_ref[1]+(0.986422233856)*x_ref[2]
        ref[0,1]=(1.35021866887)*x_ref[0]+(-1.18036897489)*x_ref[1]+(0.143096009704)*x_ref[2]
        ref[0,2]=(1.92136475346)*x_ref[0]+(-1.52041779666)*x_ref[1]+(1.50412123174)*x_ref[2]
        ref[0,3]=(-1.26349432492)*x_ref[0]+(-1.69090463821)*x_ref[1]+(0.0519848717435)*x_ref[2]
        ref[0,4]=(-0.450643290235)*x_ref[0]+(-1.0918592714)*x_ref[1]+(-1.78531176031)*x_ref[2]
        ref[1,0]=(1.24575021839)*x_ref[0]+(0.378183524368)*x_ref[1]+(0.164647643767)*x_ref[2]
        ref[1,1]=(-0.0164625547855)*x_ref[0]+(-1.06539959746)*x_ref[1]+(-1.5644531354)*x_ref[2]
        ref[1,2]=(0.126087223544)*x_ref[0]+(0.876303692726)*x_ref[1]+(1.3521510598)*x_ref[2]
        ref[1,3]=(-1.15749467575)*x_ref[0]+(-0.850991452556)*x_ref[1]+(-0.903705777185)*x_ref[2]
        ref[1,4]=(-0.680094762766)*x_ref[0]+(0.280093157366)*x_ref[1]+(0.460572306025)*x_ref[2]
        ref[2,0]=(-0.201094411285)*x_ref[0]+(0.610240266355)*x_ref[1]+(0.741298609653)*x_ref[2]
        ref[2,1]=(0.754412798233)*x_ref[0]+(0.765169641301)*x_ref[1]+(0.522349231028)*x_ref[2]
        ref[2,2]=(-0.330239429991)*x_ref[0]+(1.04468899978)*x_ref[1]+(-0.493706328863)*x_ref[2]
        ref[2,3]=(-0.107189727407)*x_ref[0]+(-1.57053989855)*x_ref[1]+(1.5864416691)*x_ref[2]
        ref[2,4]=(1.78349630817)*x_ref[0]+(0.415138686617)*x_ref[1]+(-0.567656058268)*x_ref[2]
        ref[3,0]=(-0.242864411179)*x_ref[0]+(0.342401319249)*x_ref[1]+(-0.963749227433)*x_ref[2]
        ref[3,1]=(-0.677695348369)*x_ref[0]+(-1.52126003219)*x_ref[1]+(-0.197904354713)*x_ref[2]
        ref[3,2]=(-0.549674242841)*x_ref[0]+(1.23313354594)*x_ref[1]+(0.701202865744)*x_ref[2]
        ref[3,3]=(0.278097672739)*x_ref[0]+(-1.31336623233)*x_ref[1]+(1.69771904282)*x_ref[2]
        ref[3,4]=(0.0982668788821)*x_ref[0]+(-1.11857597802)*x_ref[1]+(-0.0505715204344)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.121576835783)*x[0]+(-0.941342629086)*x[1]
        s[0,0,1]=(0.113281061637)*x[0]+(1.53625452894)*x[1]
        s[0,1,0]=(0.0674434661275)*x[0]+(-0.633539894942)*x[1]
        s[0,1,1]=(0.22586221125)*x[0]+(-0.756374425353)*x[1]
        s[1,0,0]=(0.0826651782713)*x[0]+(-0.78928571118)*x[1]
        s[1,0,1]=(-0.844609848337)*x[0]+(1.15779995507)*x[1]
        s[1,1,0]=(0.137225056583)*x[0]+(-1.03725215253)*x[1]
        s[1,1,1]=(1.66612561218)*x[0]+(-0.243423071065)*x[1]
        s[2,0,0]=(0.0833021990187)*x[0]+(-1.2712750293)*x[1]
        s[2,0,1]=(0.13762583458)*x[0]+(-0.385843401588)*x[1]
        s[2,1,0]=(-0.926228562488)*x[0]+(-0.305689405746)*x[1]
        s[2,1,1]=(0.725733502304)*x[0]+(1.40832913183)*x[1]
        s[3,0,0]=(1.09943065211)*x[0]+(1.13514402138)*x[1]
        s[3,0,1]=(-1.0354232693)*x[0]+(-0.678629246255)*x[1]
        s[3,1,0]=(-0.44364521298)*x[0]+(0.857496257444)*x[1]
        s[3,1,1]=(1.53635223417)*x[0]+(-0.0887540544341)*x[1]
        s[4,0,0]=(0.553400675937)*x[0]+(-0.680780568865)*x[1]
        s[4,0,1]=(-0.00341234496916)*x[0]+(0.115104542523)*x[1]
        s[4,1,0]=(1.07927709685)*x[0]+(-0.270560535373)*x[1]
        s[4,1,1]=(-0.681458204636)*x[0]+(-0.607605729645)*x[1]
        s[5,0,0]=(0.10322587433)*x[0]+(0.982393144196)*x[1]
        s[5,0,1]=(-0.229957227498)*x[0]+(0.940017423086)*x[1]
        s[5,1,0]=(-0.465473033025)*x[0]+(-0.61380542705)*x[1]
        s[5,1,1]=(1.43248132924)*x[0]+(-1.71390853514)*x[1]
        ref[0,0,0]=(0.121576835783)*x_ref[0]+(-0.941342629086)*x_ref[1]
        ref[0,0,1]=(0.113281061637)*x_ref[0]+(1.53625452894)*x_ref[1]
        ref[0,1,0]=(0.0674434661275)*x_ref[0]+(-0.633539894942)*x_ref[1]
        ref[0,1,1]=(0.22586221125)*x_ref[0]+(-0.756374425353)*x_ref[1]
        ref[1,0,0]=(0.0826651782713)*x_ref[0]+(-0.78928571118)*x_ref[1]
        ref[1,0,1]=(-0.844609848337)*x_ref[0]+(1.15779995507)*x_ref[1]
        ref[1,1,0]=(0.137225056583)*x_ref[0]+(-1.03725215253)*x_ref[1]
        ref[1,1,1]=(1.66612561218)*x_ref[0]+(-0.243423071065)*x_ref[1]
        ref[2,0,0]=(0.0833021990187)*x_ref[0]+(-1.2712750293)*x_ref[1]
        ref[2,0,1]=(0.13762583458)*x_ref[0]+(-0.385843401588)*x_ref[1]
        ref[2,1,0]=(-0.926228562488)*x_ref[0]+(-0.305689405746)*x_ref[1]
        ref[2,1,1]=(0.725733502304)*x_ref[0]+(1.40832913183)*x_ref[1]
        ref[3,0,0]=(1.09943065211)*x_ref[0]+(1.13514402138)*x_ref[1]
        ref[3,0,1]=(-1.0354232693)*x_ref[0]+(-0.678629246255)*x_ref[1]
        ref[3,1,0]=(-0.44364521298)*x_ref[0]+(0.857496257444)*x_ref[1]
        ref[3,1,1]=(1.53635223417)*x_ref[0]+(-0.0887540544341)*x_ref[1]
        ref[4,0,0]=(0.553400675937)*x_ref[0]+(-0.680780568865)*x_ref[1]
        ref[4,0,1]=(-0.00341234496916)*x_ref[0]+(0.115104542523)*x_ref[1]
        ref[4,1,0]=(1.07927709685)*x_ref[0]+(-0.270560535373)*x_ref[1]
        ref[4,1,1]=(-0.681458204636)*x_ref[0]+(-0.607605729645)*x_ref[1]
        ref[5,0,0]=(0.10322587433)*x_ref[0]+(0.982393144196)*x_ref[1]
        ref[5,0,1]=(-0.229957227498)*x_ref[0]+(0.940017423086)*x_ref[1]
        ref[5,1,0]=(-0.465473033025)*x_ref[0]+(-0.61380542705)*x_ref[1]
        ref[5,1,1]=(1.43248132924)*x_ref[0]+(-1.71390853514)*x_ref[1]
      else:
        s[0,0,0]=(0.663545175027)*x[0]+(-1.05528281336)*x[1]+(-0.25024128376)*x[2]
        s[0,0,1]=(1.10817148343)*x[0]+(0.224734467846)*x[1]+(0.712872841532)*x[2]
        s[0,1,0]=(-0.345072989586)*x[0]+(-1.244678278)*x[1]+(0.103446646879)*x[2]
        s[0,1,1]=(-0.106279741301)*x[0]+(-0.525721333511)*x[1]+(0.401917882446)*x[2]
        s[1,0,0]=(0.838242796285)*x[0]+(-0.706020082817)*x[1]+(0.443650945014)*x[2]
        s[1,0,1]=(-1.16557336842)*x[0]+(0.215324659268)*x[1]+(-0.285210566147)*x[2]
        s[1,1,0]=(-0.141574656707)*x[0]+(1.66054358217)*x[1]+(-0.554616610711)*x[2]
        s[1,1,1]=(0.523682935472)*x[0]+(0.687178896331)*x[1]+(0.875639049678)*x[2]
        s[2,0,0]=(0.585144782395)*x[0]+(0.172063481865)*x[1]+(0.920891875835)*x[2]
        s[2,0,1]=(0.660815435485)*x[0]+(-1.63324849021)*x[1]+(0.0810858621718)*x[2]
        s[2,1,0]=(-0.238166824253)*x[0]+(0.0442487879497)*x[1]+(1.05748755823)*x[2]
        s[2,1,1]=(1.14314890535)*x[0]+(0.63161746042)*x[1]+(1.5024770616)*x[2]
        s[3,0,0]=(-1.33327391952)*x[0]+(1.58312319401)*x[1]+(-1.24202894696)*x[2]
        s[3,0,1]=(-0.0242608819668)*x[0]+(-1.12901018411)*x[1]+(-0.443840381999)*x[2]
        s[3,1,0]=(1.12400540809)*x[0]+(-0.25104247006)*x[1]+(0.55935664575)*x[2]
        s[3,1,1]=(0.976977053811)*x[0]+(1.46134206601)*x[1]+(-0.883829686094)*x[2]
        s[4,0,0]=(0.68585022695)*x[0]+(0.320211070942)*x[1]+(-0.715034754238)*x[2]
        s[4,0,1]=(0.745090118519)*x[0]+(-0.56687124464)*x[1]+(-1.41483630619)*x[2]
        s[4,1,0]=(-0.864645032271)*x[0]+(-0.0852453841156)*x[1]+(-1.02272230917)*x[2]
        s[4,1,1]=(0.243107783262)*x[0]+(-1.76263127813)*x[1]+(-0.0396122398093)*x[2]
        s[5,0,0]=(-1.33890039547)*x[0]+(0.95916460762)*x[1]+(-0.118457193655)*x[2]
        s[5,0,1]=(-1.03526146458)*x[0]+(-0.279038868252)*x[1]+(-0.702378064321)*x[2]
        s[5,1,0]=(-0.434705839506)*x[0]+(0.388520593856)*x[1]+(0.213085443323)*x[2]
        s[5,1,1]=(-0.156276503308)*x[0]+(-0.463824282648)*x[1]+(-0.801704395726)*x[2]
        ref[0,0,0]=(0.663545175027)*x_ref[0]+(-1.05528281336)*x_ref[1]+(-0.25024128376)*x_ref[2]
        ref[0,0,1]=(1.10817148343)*x_ref[0]+(0.224734467846)*x_ref[1]+(0.712872841532)*x_ref[2]
        ref[0,1,0]=(-0.345072989586)*x_ref[0]+(-1.244678278)*x_ref[1]+(0.103446646879)*x_ref[2]
        ref[0,1,1]=(-0.106279741301)*x_ref[0]+(-0.525721333511)*x_ref[1]+(0.401917882446)*x_ref[2]
        ref[1,0,0]=(0.838242796285)*x_ref[0]+(-0.706020082817)*x_ref[1]+(0.443650945014)*x_ref[2]
        ref[1,0,1]=(-1.16557336842)*x_ref[0]+(0.215324659268)*x_ref[1]+(-0.285210566147)*x_ref[2]
        ref[1,1,0]=(-0.141574656707)*x_ref[0]+(1.66054358217)*x_ref[1]+(-0.554616610711)*x_ref[2]
        ref[1,1,1]=(0.523682935472)*x_ref[0]+(0.687178896331)*x_ref[1]+(0.875639049678)*x_ref[2]
        ref[2,0,0]=(0.585144782395)*x_ref[0]+(0.172063481865)*x_ref[1]+(0.920891875835)*x_ref[2]
        ref[2,0,1]=(0.660815435485)*x_ref[0]+(-1.63324849021)*x_ref[1]+(0.0810858621718)*x_ref[2]
        ref[2,1,0]=(-0.238166824253)*x_ref[0]+(0.0442487879497)*x_ref[1]+(1.05748755823)*x_ref[2]
        ref[2,1,1]=(1.14314890535)*x_ref[0]+(0.63161746042)*x_ref[1]+(1.5024770616)*x_ref[2]
        ref[3,0,0]=(-1.33327391952)*x_ref[0]+(1.58312319401)*x_ref[1]+(-1.24202894696)*x_ref[2]
        ref[3,0,1]=(-0.0242608819668)*x_ref[0]+(-1.12901018411)*x_ref[1]+(-0.443840381999)*x_ref[2]
        ref[3,1,0]=(1.12400540809)*x_ref[0]+(-0.25104247006)*x_ref[1]+(0.55935664575)*x_ref[2]
        ref[3,1,1]=(0.976977053811)*x_ref[0]+(1.46134206601)*x_ref[1]+(-0.883829686094)*x_ref[2]
        ref[4,0,0]=(0.68585022695)*x_ref[0]+(0.320211070942)*x_ref[1]+(-0.715034754238)*x_ref[2]
        ref[4,0,1]=(0.745090118519)*x_ref[0]+(-0.56687124464)*x_ref[1]+(-1.41483630619)*x_ref[2]
        ref[4,1,0]=(-0.864645032271)*x_ref[0]+(-0.0852453841156)*x_ref[1]+(-1.02272230917)*x_ref[2]
        ref[4,1,1]=(0.243107783262)*x_ref[0]+(-1.76263127813)*x_ref[1]+(-0.0396122398093)*x_ref[2]
        ref[5,0,0]=(-1.33890039547)*x_ref[0]+(0.95916460762)*x_ref[1]+(-0.118457193655)*x_ref[2]
        ref[5,0,1]=(-1.03526146458)*x_ref[0]+(-0.279038868252)*x_ref[1]+(-0.702378064321)*x_ref[2]
        ref[5,1,0]=(-0.434705839506)*x_ref[0]+(0.388520593856)*x_ref[1]+(0.213085443323)*x_ref[2]
        ref[5,1,1]=(-0.156276503308)*x_ref[0]+(-0.463824282648)*x_ref[1]+(-0.801704395726)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedFunctionOnBoundary
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.239181398192)*x[0]+(-0.39363876765)*x[1]
        s[0,0,0,1]=(0.391461093031)*x[0]+(-0.825968486042)*x[1]
        s[0,0,1,0]=(0.327281495381)*x[0]+(-0.201869630768)*x[1]
        s[0,0,1,1]=(-0.0266300089051)*x[0]+(0.0315885163331)*x[1]
        s[0,0,2,0]=(-1.29808531891)*x[0]+(0.908442635353)*x[1]
        s[0,0,2,1]=(-0.0719703714734)*x[0]+(-1.2949593192)*x[1]
        s[0,1,0,0]=(1.56337010697)*x[0]+(0.554963325139)*x[1]
        s[0,1,0,1]=(-0.590177179718)*x[0]+(-0.577417888525)*x[1]
        s[0,1,1,0]=(1.12019940888)*x[0]+(0.515714775221)*x[1]
        s[0,1,1,1]=(1.62187733115)*x[0]+(-0.0496003559467)*x[1]
        s[0,1,2,0]=(1.33693246276)*x[0]+(0.18028194694)*x[1]
        s[0,1,2,1]=(1.197520649)*x[0]+(-0.189794758648)*x[1]
        s[0,2,0,0]=(0.458741675295)*x[0]+(-1.23374520557)*x[1]
        s[0,2,0,1]=(0.122618584159)*x[0]+(1.68218857683)*x[1]
        s[0,2,1,0]=(-0.629416301315)*x[0]+(0.00821950501686)*x[1]
        s[0,2,1,1]=(0.372444097376)*x[0]+(-0.389807082811)*x[1]
        s[0,2,2,0]=(0.453580881811)*x[0]+(0.234675328844)*x[1]
        s[0,2,2,1]=(-0.488447987119)*x[0]+(-0.795290862711)*x[1]
        s[0,3,0,0]=(-0.0558838669873)*x[0]+(-0.123245413784)*x[1]
        s[0,3,0,1]=(1.37779240976)*x[0]+(0.0843495250295)*x[1]
        s[0,3,1,0]=(1.17819744943)*x[0]+(-0.859044697828)*x[1]
        s[0,3,1,1]=(1.25175416429)*x[0]+(-0.727964075982)*x[1]
        s[0,3,2,0]=(0.331480991751)*x[0]+(0.115364502392)*x[1]
        s[0,3,2,1]=(-0.791037073807)*x[0]+(-0.51631676808)*x[1]
        s[0,4,0,0]=(0.884389418081)*x[0]+(1.15006122432)*x[1]
        s[0,4,0,1]=(-0.914023114312)*x[0]+(-0.211608734049)*x[1]
        s[0,4,1,0]=(0.55983523554)*x[0]+(-0.792920109589)*x[1]
        s[0,4,1,1]=(0.70901081021)*x[0]+(0.205352000801)*x[1]
        s[0,4,2,0]=(0.140341122546)*x[0]+(0.633534699565)*x[1]
        s[0,4,2,1]=(-0.399843857722)*x[0]+(0.0360346234753)*x[1]
        s[1,0,0,0]=(0.649005423866)*x[0]+(-0.72958629181)*x[1]
        s[1,0,0,1]=(-1.23313007881)*x[0]+(0.240005909545)*x[1]
        s[1,0,1,0]=(0.817440794961)*x[0]+(0.219684770228)*x[1]
        s[1,0,1,1]=(0.300714219818)*x[0]+(0.339948691074)*x[1]
        s[1,0,2,0]=(0.692412749315)*x[0]+(0.847435101682)*x[1]
        s[1,0,2,1]=(0.650288735017)*x[0]+(1.25808519806)*x[1]
        s[1,1,0,0]=(-0.723761201921)*x[0]+(-0.337875653383)*x[1]
        s[1,1,0,1]=(1.10005713123)*x[0]+(0.465822601087)*x[1]
        s[1,1,1,0]=(0.140014505481)*x[0]+(0.289986937348)*x[1]
        s[1,1,1,1]=(0.0941270067899)*x[0]+(-1.45892231406)*x[1]
        s[1,1,2,0]=(1.04493752023)*x[0]+(-0.474063991258)*x[1]
        s[1,1,2,1]=(0.703284346556)*x[0]+(0.482116021536)*x[1]
        s[1,2,0,0]=(-1.66498217207)*x[0]+(-0.144077385929)*x[1]
        s[1,2,0,1]=(1.50511312044)*x[0]+(1.02587270487)*x[1]
        s[1,2,1,0]=(-1.15095468052)*x[0]+(-0.472480042698)*x[1]
        s[1,2,1,1]=(-0.453604215856)*x[0]+(-1.58856703352)*x[1]
        s[1,2,2,0]=(0.659481843854)*x[0]+(0.649931585882)*x[1]
        s[1,2,2,1]=(0.642785065122)*x[0]+(0.0614989297125)*x[1]
        s[1,3,0,0]=(1.65661008057)*x[0]+(-1.08472986336)*x[1]
        s[1,3,0,1]=(-0.64853562643)*x[0]+(-0.912673933935)*x[1]
        s[1,3,1,0]=(-0.108999021371)*x[0]+(-1.44843200164)*x[1]
        s[1,3,1,1]=(0.568395064875)*x[0]+(-0.471819846301)*x[1]
        s[1,3,2,0]=(-0.801680619008)*x[0]+(0.197249605781)*x[1]
        s[1,3,2,1]=(-0.512763932006)*x[0]+(-1.03221569621)*x[1]
        s[1,4,0,0]=(0.54142249246)*x[0]+(-1.28623467967)*x[1]
        s[1,4,0,1]=(1.19094828259)*x[0]+(0.58814933754)*x[1]
        s[1,4,1,0]=(-0.203699090085)*x[0]+(0.839194924785)*x[1]
        s[1,4,1,1]=(-1.73647778175)*x[0]+(1.26399102178)*x[1]
        s[1,4,2,0]=(0.688685884338)*x[0]+(-0.0441772317601)*x[1]
        s[1,4,2,1]=(-0.811158497264)*x[0]+(-0.663923769116)*x[1]
        s[2,0,0,0]=(0.237794148299)*x[0]+(0.125329413285)*x[1]
        s[2,0,0,1]=(-1.40518934715)*x[0]+(1.22320828318)*x[1]
        s[2,0,1,0]=(0.223901464538)*x[0]+(-0.00989463351767)*x[1]
        s[2,0,1,1]=(-0.912246132725)*x[0]+(0.686091520129)*x[1]
        s[2,0,2,0]=(0.181752601691)*x[0]+(-1.12779553611)*x[1]
        s[2,0,2,1]=(-0.061707870206)*x[0]+(-1.3429678748)*x[1]
        s[2,1,0,0]=(1.56396877322)*x[0]+(-1.05119596867)*x[1]
        s[2,1,0,1]=(0.929556514862)*x[0]+(0.406040622531)*x[1]
        s[2,1,1,0]=(-0.509355797822)*x[0]+(0.0605066527971)*x[1]
        s[2,1,1,1]=(-0.528244794646)*x[0]+(0.0388070048758)*x[1]
        s[2,1,2,0]=(-0.200007488719)*x[0]+(-0.711333050321)*x[1]
        s[2,1,2,1]=(0.175160525472)*x[0]+(-1.24343736284)*x[1]
        s[2,2,0,0]=(0.894960105109)*x[0]+(-0.747881639384)*x[1]
        s[2,2,0,1]=(-0.0894204005642)*x[0]+(-0.694792125757)*x[1]
        s[2,2,1,0]=(0.172037435959)*x[0]+(1.29233768559)*x[1]
        s[2,2,1,1]=(-0.312372601854)*x[0]+(-1.35937909779)*x[1]
        s[2,2,2,0]=(0.323370754787)*x[0]+(-0.0601280403525)*x[1]
        s[2,2,2,1]=(0.648154329469)*x[0]+(-0.976168040714)*x[1]
        s[2,3,0,0]=(0.178127534181)*x[0]+(-0.451322025537)*x[1]
        s[2,3,0,1]=(-1.88695440231)*x[0]+(0.164327427196)*x[1]
        s[2,3,1,0]=(0.543479878129)*x[0]+(-0.553094838979)*x[1]
        s[2,3,1,1]=(-0.745599057313)*x[0]+(0.710208052579)*x[1]
        s[2,3,2,0]=(-0.262128106404)*x[0]+(0.839008375922)*x[1]
        s[2,3,2,1]=(-1.20847299902)*x[0]+(-0.321234640326)*x[1]
        s[2,4,0,0]=(-1.24374530208)*x[0]+(-0.961997778937)*x[1]
        s[2,4,0,1]=(-0.0306817051594)*x[0]+(-0.127120502282)*x[1]
        s[2,4,1,0]=(1.25878772522)*x[0]+(1.04432409904)*x[1]
        s[2,4,1,1]=(0.271036209294)*x[0]+(-1.19597311792)*x[1]
        s[2,4,2,0]=(-0.073359480179)*x[0]+(-0.31523273696)*x[1]
        s[2,4,2,1]=(0.738924603079)*x[0]+(0.251467051457)*x[1]
        s[3,0,0,0]=(-0.257569738752)*x[0]+(-1.51373140433)*x[1]
        s[3,0,0,1]=(-0.0113459170081)*x[0]+(-0.356375170129)*x[1]
        s[3,0,1,0]=(-0.290333097834)*x[0]+(-1.56956309356)*x[1]
        s[3,0,1,1]=(-1.0806247672)*x[0]+(-0.376429365315)*x[1]
        s[3,0,2,0]=(-1.32161176104)*x[0]+(0.784442685433)*x[1]
        s[3,0,2,1]=(-1.00382464524)*x[0]+(-0.571979697031)*x[1]
        s[3,1,0,0]=(0.140827544385)*x[0]+(1.3690229529)*x[1]
        s[3,1,0,1]=(0.315011101794)*x[0]+(-1.0263379129)*x[1]
        s[3,1,1,0]=(-0.478940977535)*x[0]+(0.718456942414)*x[1]
        s[3,1,1,1]=(-0.655873236005)*x[0]+(-0.148701297422)*x[1]
        s[3,1,2,0]=(0.56126503141)*x[0]+(1.68526215251)*x[1]
        s[3,1,2,1]=(-0.337284606646)*x[0]+(-0.197295432114)*x[1]
        s[3,2,0,0]=(-0.5979320732)*x[0]+(-0.0600053477444)*x[1]
        s[3,2,0,1]=(-1.20531807175)*x[0]+(-0.148853538473)*x[1]
        s[3,2,1,0]=(0.313468114426)*x[0]+(-0.436912848514)*x[1]
        s[3,2,1,1]=(-0.544279507407)*x[0]+(0.189063470736)*x[1]
        s[3,2,2,0]=(-0.551741475397)*x[0]+(0.196370248113)*x[1]
        s[3,2,2,1]=(-0.0216545656433)*x[0]+(0.23669839429)*x[1]
        s[3,3,0,0]=(-0.406652896064)*x[0]+(0.00397676360081)*x[1]
        s[3,3,0,1]=(-0.305847961815)*x[0]+(-1.2563276235)*x[1]
        s[3,3,1,0]=(-0.450621831697)*x[0]+(-0.213476321469)*x[1]
        s[3,3,1,1]=(1.20548652879)*x[0]+(-0.286503833918)*x[1]
        s[3,3,2,0]=(0.516344981347)*x[0]+(-0.343787067175)*x[1]
        s[3,3,2,1]=(-0.01938029713)*x[0]+(-0.779285558404)*x[1]
        s[3,4,0,0]=(-0.406233598349)*x[0]+(-0.498045431333)*x[1]
        s[3,4,0,1]=(-0.790577593976)*x[0]+(1.07251305331)*x[1]
        s[3,4,1,0]=(0.510569891955)*x[0]+(0.526099613153)*x[1]
        s[3,4,1,1]=(0.150065560826)*x[0]+(0.966145292988)*x[1]
        s[3,4,2,0]=(0.718402792955)*x[0]+(-0.430416841145)*x[1]
        s[3,4,2,1]=(-0.430240857917)*x[0]+(-0.0679614650591)*x[1]
        ref[0,0,0,0]=(0.239181398192)*x_ref[0]+(-0.39363876765)*x_ref[1]
        ref[0,0,0,1]=(0.391461093031)*x_ref[0]+(-0.825968486042)*x_ref[1]
        ref[0,0,1,0]=(0.327281495381)*x_ref[0]+(-0.201869630768)*x_ref[1]
        ref[0,0,1,1]=(-0.0266300089051)*x_ref[0]+(0.0315885163331)*x_ref[1]
        ref[0,0,2,0]=(-1.29808531891)*x_ref[0]+(0.908442635353)*x_ref[1]
        ref[0,0,2,1]=(-0.0719703714734)*x_ref[0]+(-1.2949593192)*x_ref[1]
        ref[0,1,0,0]=(1.56337010697)*x_ref[0]+(0.554963325139)*x_ref[1]
        ref[0,1,0,1]=(-0.590177179718)*x_ref[0]+(-0.577417888525)*x_ref[1]
        ref[0,1,1,0]=(1.12019940888)*x_ref[0]+(0.515714775221)*x_ref[1]
        ref[0,1,1,1]=(1.62187733115)*x_ref[0]+(-0.0496003559467)*x_ref[1]
        ref[0,1,2,0]=(1.33693246276)*x_ref[0]+(0.18028194694)*x_ref[1]
        ref[0,1,2,1]=(1.197520649)*x_ref[0]+(-0.189794758648)*x_ref[1]
        ref[0,2,0,0]=(0.458741675295)*x_ref[0]+(-1.23374520557)*x_ref[1]
        ref[0,2,0,1]=(0.122618584159)*x_ref[0]+(1.68218857683)*x_ref[1]
        ref[0,2,1,0]=(-0.629416301315)*x_ref[0]+(0.00821950501686)*x_ref[1]
        ref[0,2,1,1]=(0.372444097376)*x_ref[0]+(-0.389807082811)*x_ref[1]
        ref[0,2,2,0]=(0.453580881811)*x_ref[0]+(0.234675328844)*x_ref[1]
        ref[0,2,2,1]=(-0.488447987119)*x_ref[0]+(-0.795290862711)*x_ref[1]
        ref[0,3,0,0]=(-0.0558838669873)*x_ref[0]+(-0.123245413784)*x_ref[1]
        ref[0,3,0,1]=(1.37779240976)*x_ref[0]+(0.0843495250295)*x_ref[1]
        ref[0,3,1,0]=(1.17819744943)*x_ref[0]+(-0.859044697828)*x_ref[1]
        ref[0,3,1,1]=(1.25175416429)*x_ref[0]+(-0.727964075982)*x_ref[1]
        ref[0,3,2,0]=(0.331480991751)*x_ref[0]+(0.115364502392)*x_ref[1]
        ref[0,3,2,1]=(-0.791037073807)*x_ref[0]+(-0.51631676808)*x_ref[1]
        ref[0,4,0,0]=(0.884389418081)*x_ref[0]+(1.15006122432)*x_ref[1]
        ref[0,4,0,1]=(-0.914023114312)*x_ref[0]+(-0.211608734049)*x_ref[1]
        ref[0,4,1,0]=(0.55983523554)*x_ref[0]+(-0.792920109589)*x_ref[1]
        ref[0,4,1,1]=(0.70901081021)*x_ref[0]+(0.205352000801)*x_ref[1]
        ref[0,4,2,0]=(0.140341122546)*x_ref[0]+(0.633534699565)*x_ref[1]
        ref[0,4,2,1]=(-0.399843857722)*x_ref[0]+(0.0360346234753)*x_ref[1]
        ref[1,0,0,0]=(0.649005423866)*x_ref[0]+(-0.72958629181)*x_ref[1]
        ref[1,0,0,1]=(-1.23313007881)*x_ref[0]+(0.240005909545)*x_ref[1]
        ref[1,0,1,0]=(0.817440794961)*x_ref[0]+(0.219684770228)*x_ref[1]
        ref[1,0,1,1]=(0.300714219818)*x_ref[0]+(0.339948691074)*x_ref[1]
        ref[1,0,2,0]=(0.692412749315)*x_ref[0]+(0.847435101682)*x_ref[1]
        ref[1,0,2,1]=(0.650288735017)*x_ref[0]+(1.25808519806)*x_ref[1]
        ref[1,1,0,0]=(-0.723761201921)*x_ref[0]+(-0.337875653383)*x_ref[1]
        ref[1,1,0,1]=(1.10005713123)*x_ref[0]+(0.465822601087)*x_ref[1]
        ref[1,1,1,0]=(0.140014505481)*x_ref[0]+(0.289986937348)*x_ref[1]
        ref[1,1,1,1]=(0.0941270067899)*x_ref[0]+(-1.45892231406)*x_ref[1]
        ref[1,1,2,0]=(1.04493752023)*x_ref[0]+(-0.474063991258)*x_ref[1]
        ref[1,1,2,1]=(0.703284346556)*x_ref[0]+(0.482116021536)*x_ref[1]
        ref[1,2,0,0]=(-1.66498217207)*x_ref[0]+(-0.144077385929)*x_ref[1]
        ref[1,2,0,1]=(1.50511312044)*x_ref[0]+(1.02587270487)*x_ref[1]
        ref[1,2,1,0]=(-1.15095468052)*x_ref[0]+(-0.472480042698)*x_ref[1]
        ref[1,2,1,1]=(-0.453604215856)*x_ref[0]+(-1.58856703352)*x_ref[1]
        ref[1,2,2,0]=(0.659481843854)*x_ref[0]+(0.649931585882)*x_ref[1]
        ref[1,2,2,1]=(0.642785065122)*x_ref[0]+(0.0614989297125)*x_ref[1]
        ref[1,3,0,0]=(1.65661008057)*x_ref[0]+(-1.08472986336)*x_ref[1]
        ref[1,3,0,1]=(-0.64853562643)*x_ref[0]+(-0.912673933935)*x_ref[1]
        ref[1,3,1,0]=(-0.108999021371)*x_ref[0]+(-1.44843200164)*x_ref[1]
        ref[1,3,1,1]=(0.568395064875)*x_ref[0]+(-0.471819846301)*x_ref[1]
        ref[1,3,2,0]=(-0.801680619008)*x_ref[0]+(0.197249605781)*x_ref[1]
        ref[1,3,2,1]=(-0.512763932006)*x_ref[0]+(-1.03221569621)*x_ref[1]
        ref[1,4,0,0]=(0.54142249246)*x_ref[0]+(-1.28623467967)*x_ref[1]
        ref[1,4,0,1]=(1.19094828259)*x_ref[0]+(0.58814933754)*x_ref[1]
        ref[1,4,1,0]=(-0.203699090085)*x_ref[0]+(0.839194924785)*x_ref[1]
        ref[1,4,1,1]=(-1.73647778175)*x_ref[0]+(1.26399102178)*x_ref[1]
        ref[1,4,2,0]=(0.688685884338)*x_ref[0]+(-0.0441772317601)*x_ref[1]
        ref[1,4,2,1]=(-0.811158497264)*x_ref[0]+(-0.663923769116)*x_ref[1]
        ref[2,0,0,0]=(0.237794148299)*x_ref[0]+(0.125329413285)*x_ref[1]
        ref[2,0,0,1]=(-1.40518934715)*x_ref[0]+(1.22320828318)*x_ref[1]
        ref[2,0,1,0]=(0.223901464538)*x_ref[0]+(-0.00989463351767)*x_ref[1]
        ref[2,0,1,1]=(-0.912246132725)*x_ref[0]+(0.686091520129)*x_ref[1]
        ref[2,0,2,0]=(0.181752601691)*x_ref[0]+(-1.12779553611)*x_ref[1]
        ref[2,0,2,1]=(-0.061707870206)*x_ref[0]+(-1.3429678748)*x_ref[1]
        ref[2,1,0,0]=(1.56396877322)*x_ref[0]+(-1.05119596867)*x_ref[1]
        ref[2,1,0,1]=(0.929556514862)*x_ref[0]+(0.406040622531)*x_ref[1]
        ref[2,1,1,0]=(-0.509355797822)*x_ref[0]+(0.0605066527971)*x_ref[1]
        ref[2,1,1,1]=(-0.528244794646)*x_ref[0]+(0.0388070048758)*x_ref[1]
        ref[2,1,2,0]=(-0.200007488719)*x_ref[0]+(-0.711333050321)*x_ref[1]
        ref[2,1,2,1]=(0.175160525472)*x_ref[0]+(-1.24343736284)*x_ref[1]
        ref[2,2,0,0]=(0.894960105109)*x_ref[0]+(-0.747881639384)*x_ref[1]
        ref[2,2,0,1]=(-0.0894204005642)*x_ref[0]+(-0.694792125757)*x_ref[1]
        ref[2,2,1,0]=(0.172037435959)*x_ref[0]+(1.29233768559)*x_ref[1]
        ref[2,2,1,1]=(-0.312372601854)*x_ref[0]+(-1.35937909779)*x_ref[1]
        ref[2,2,2,0]=(0.323370754787)*x_ref[0]+(-0.0601280403525)*x_ref[1]
        ref[2,2,2,1]=(0.648154329469)*x_ref[0]+(-0.976168040714)*x_ref[1]
        ref[2,3,0,0]=(0.178127534181)*x_ref[0]+(-0.451322025537)*x_ref[1]
        ref[2,3,0,1]=(-1.88695440231)*x_ref[0]+(0.164327427196)*x_ref[1]
        ref[2,3,1,0]=(0.543479878129)*x_ref[0]+(-0.553094838979)*x_ref[1]
        ref[2,3,1,1]=(-0.745599057313)*x_ref[0]+(0.710208052579)*x_ref[1]
        ref[2,3,2,0]=(-0.262128106404)*x_ref[0]+(0.839008375922)*x_ref[1]
        ref[2,3,2,1]=(-1.20847299902)*x_ref[0]+(-0.321234640326)*x_ref[1]
        ref[2,4,0,0]=(-1.24374530208)*x_ref[0]+(-0.961997778937)*x_ref[1]
        ref[2,4,0,1]=(-0.0306817051594)*x_ref[0]+(-0.127120502282)*x_ref[1]
        ref[2,4,1,0]=(1.25878772522)*x_ref[0]+(1.04432409904)*x_ref[1]
        ref[2,4,1,1]=(0.271036209294)*x_ref[0]+(-1.19597311792)*x_ref[1]
        ref[2,4,2,0]=(-0.073359480179)*x_ref[0]+(-0.31523273696)*x_ref[1]
        ref[2,4,2,1]=(0.738924603079)*x_ref[0]+(0.251467051457)*x_ref[1]
        ref[3,0,0,0]=(-0.257569738752)*x_ref[0]+(-1.51373140433)*x_ref[1]
        ref[3,0,0,1]=(-0.0113459170081)*x_ref[0]+(-0.356375170129)*x_ref[1]
        ref[3,0,1,0]=(-0.290333097834)*x_ref[0]+(-1.56956309356)*x_ref[1]
        ref[3,0,1,1]=(-1.0806247672)*x_ref[0]+(-0.376429365315)*x_ref[1]
        ref[3,0,2,0]=(-1.32161176104)*x_ref[0]+(0.784442685433)*x_ref[1]
        ref[3,0,2,1]=(-1.00382464524)*x_ref[0]+(-0.571979697031)*x_ref[1]
        ref[3,1,0,0]=(0.140827544385)*x_ref[0]+(1.3690229529)*x_ref[1]
        ref[3,1,0,1]=(0.315011101794)*x_ref[0]+(-1.0263379129)*x_ref[1]
        ref[3,1,1,0]=(-0.478940977535)*x_ref[0]+(0.718456942414)*x_ref[1]
        ref[3,1,1,1]=(-0.655873236005)*x_ref[0]+(-0.148701297422)*x_ref[1]
        ref[3,1,2,0]=(0.56126503141)*x_ref[0]+(1.68526215251)*x_ref[1]
        ref[3,1,2,1]=(-0.337284606646)*x_ref[0]+(-0.197295432114)*x_ref[1]
        ref[3,2,0,0]=(-0.5979320732)*x_ref[0]+(-0.0600053477444)*x_ref[1]
        ref[3,2,0,1]=(-1.20531807175)*x_ref[0]+(-0.148853538473)*x_ref[1]
        ref[3,2,1,0]=(0.313468114426)*x_ref[0]+(-0.436912848514)*x_ref[1]
        ref[3,2,1,1]=(-0.544279507407)*x_ref[0]+(0.189063470736)*x_ref[1]
        ref[3,2,2,0]=(-0.551741475397)*x_ref[0]+(0.196370248113)*x_ref[1]
        ref[3,2,2,1]=(-0.0216545656433)*x_ref[0]+(0.23669839429)*x_ref[1]
        ref[3,3,0,0]=(-0.406652896064)*x_ref[0]+(0.00397676360081)*x_ref[1]
        ref[3,3,0,1]=(-0.305847961815)*x_ref[0]+(-1.2563276235)*x_ref[1]
        ref[3,3,1,0]=(-0.450621831697)*x_ref[0]+(-0.213476321469)*x_ref[1]
        ref[3,3,1,1]=(1.20548652879)*x_ref[0]+(-0.286503833918)*x_ref[1]
        ref[3,3,2,0]=(0.516344981347)*x_ref[0]+(-0.343787067175)*x_ref[1]
        ref[3,3,2,1]=(-0.01938029713)*x_ref[0]+(-0.779285558404)*x_ref[1]
        ref[3,4,0,0]=(-0.406233598349)*x_ref[0]+(-0.498045431333)*x_ref[1]
        ref[3,4,0,1]=(-0.790577593976)*x_ref[0]+(1.07251305331)*x_ref[1]
        ref[3,4,1,0]=(0.510569891955)*x_ref[0]+(0.526099613153)*x_ref[1]
        ref[3,4,1,1]=(0.150065560826)*x_ref[0]+(0.966145292988)*x_ref[1]
        ref[3,4,2,0]=(0.718402792955)*x_ref[0]+(-0.430416841145)*x_ref[1]
        ref[3,4,2,1]=(-0.430240857917)*x_ref[0]+(-0.0679614650591)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.512691010939)*x[0]+(-1.31295640543)*x[1]+(1.45426060088)*x[2]
        s[0,0,0,1]=(-0.57748372698)*x[0]+(0.215297337003)*x[1]+(0.0853439946124)*x[2]
        s[0,0,1,0]=(-0.785306554604)*x[0]+(-0.0274911695421)*x[1]+(0.693440767035)*x[2]
        s[0,0,1,1]=(-0.56932407873)*x[0]+(-0.981593188827)*x[1]+(0.87532777027)*x[2]
        s[0,0,2,0]=(0.145925104392)*x[0]+(1.45139125753)*x[1]+(1.25041746183)*x[2]
        s[0,0,2,1]=(-1.70110072582)*x[0]+(-0.0335136921878)*x[1]+(0.276204889346)*x[2]
        s[0,1,0,0]=(-0.415026748605)*x[0]+(-0.322339100074)*x[1]+(1.1050553524)*x[2]
        s[0,1,0,1]=(0.267336702938)*x[0]+(-1.44498697586)*x[1]+(0.600455793139)*x[2]
        s[0,1,1,0]=(-0.46618843993)*x[0]+(0.318230555457)*x[1]+(0.281788961443)*x[2]
        s[0,1,1,1]=(-1.12963168407)*x[0]+(-0.661100272253)*x[1]+(0.293642402995)*x[2]
        s[0,1,2,0]=(-0.0041407279814)*x[0]+(0.330916770241)*x[1]+(1.30265962609)*x[2]
        s[0,1,2,1]=(1.37513192104)*x[0]+(-1.39837371617)*x[1]+(-1.23490743489)*x[2]
        s[0,2,0,0]=(0.240920374522)*x[0]+(1.30092830627)*x[1]+(0.189720240056)*x[2]
        s[0,2,0,1]=(0.741337711467)*x[0]+(0.00893414596869)*x[1]+(1.22438764071)*x[2]
        s[0,2,1,0]=(1.0524903095)*x[0]+(-1.00088524111)*x[1]+(0.985885896119)*x[2]
        s[0,2,1,1]=(-0.252168964788)*x[0]+(1.2685968377)*x[1]+(-0.025373437998)*x[2]
        s[0,2,2,0]=(0.728523707179)*x[0]+(0.00656996105623)*x[1]+(1.85098622635)*x[2]
        s[0,2,2,1]=(0.193123464463)*x[0]+(-0.426452202023)*x[1]+(-0.144471734372)*x[2]
        s[0,3,0,0]=(0.446506355449)*x[0]+(0.379921574537)*x[1]+(-0.303150079281)*x[2]
        s[0,3,0,1]=(0.455373734156)*x[0]+(0.134600165673)*x[1]+(-0.604957024603)*x[2]
        s[0,3,1,0]=(-1.63808806962)*x[0]+(1.02897840042)*x[1]+(1.18123170106)*x[2]
        s[0,3,1,1]=(0.587448315418)*x[0]+(0.0991725296547)*x[1]+(1.33682286695)*x[2]
        s[0,3,2,0]=(-1.12085613449)*x[0]+(0.454771039944)*x[1]+(-0.131089499156)*x[2]
        s[0,3,2,1]=(-0.351021639018)*x[0]+(0.47905316732)*x[1]+(-1.07449842534)*x[2]
        s[0,4,0,0]=(0.232526828642)*x[0]+(1.33907066655)*x[1]+(-0.699701401092)*x[2]
        s[0,4,0,1]=(0.244905848535)*x[0]+(-0.799313917773)*x[1]+(-0.16513296799)*x[2]
        s[0,4,1,0]=(-1.10152273815)*x[0]+(0.898176161732)*x[1]+(-0.101734265033)*x[2]
        s[0,4,1,1]=(-0.0804598444811)*x[0]+(-0.77819195606)*x[1]+(0.277776761782)*x[2]
        s[0,4,2,0]=(1.05979692744)*x[0]+(-0.195285573862)*x[1]+(-0.0920717991265)*x[2]
        s[0,4,2,1]=(0.44924595646)*x[0]+(0.409652246508)*x[1]+(-0.0705754685165)*x[2]
        s[1,0,0,0]=(1.14991675937)*x[0]+(-0.435947647215)*x[1]+(-0.45614944131)*x[2]
        s[1,0,0,1]=(0.306890350036)*x[0]+(-1.4400848018)*x[1]+(-0.744474442385)*x[2]
        s[1,0,1,0]=(1.07930565176)*x[0]+(-0.273335650157)*x[1]+(-0.230960378838)*x[2]
        s[1,0,1,1]=(-0.97797563017)*x[0]+(-1.3159717465)*x[1]+(-0.357902208089)*x[2]
        s[1,0,2,0]=(-1.02063328249)*x[0]+(-1.18907185687)*x[1]+(-0.509296457902)*x[2]
        s[1,0,2,1]=(-0.508446666232)*x[0]+(-0.69279963667)*x[1]+(-0.497990637708)*x[2]
        s[1,1,0,0]=(-0.476141863054)*x[0]+(-0.180458121777)*x[1]+(-0.318855084826)*x[2]
        s[1,1,0,1]=(0.480566371651)*x[0]+(-1.74196768367)*x[1]+(1.23057574863)*x[2]
        s[1,1,1,0]=(-0.0213138624722)*x[0]+(-0.0846532154764)*x[1]+(-0.61169792023)*x[2]
        s[1,1,1,1]=(-0.389381507111)*x[0]+(0.436088515729)*x[1]+(-0.603019053837)*x[2]
        s[1,1,2,0]=(-1.24478786129)*x[0]+(0.908984432701)*x[1]+(-1.49483946407)*x[2]
        s[1,1,2,1]=(0.866266264162)*x[0]+(-0.190579076131)*x[1]+(-1.63524384216)*x[2]
        s[1,2,0,0]=(0.0333832222755)*x[0]+(-0.551591529565)*x[1]+(1.22170194609)*x[2]
        s[1,2,0,1]=(-0.551682390802)*x[0]+(0.223940652339)*x[1]+(0.141063634211)*x[2]
        s[1,2,1,0]=(1.17072817252)*x[0]+(0.28318356571)*x[1]+(-0.788096581969)*x[2]
        s[1,2,1,1]=(-1.01576849129)*x[0]+(-0.502669485722)*x[1]+(-0.966677100703)*x[2]
        s[1,2,2,0]=(-1.59522074083)*x[0]+(-1.16016984107)*x[1]+(-0.117030183187)*x[2]
        s[1,2,2,1]=(-0.979165259497)*x[0]+(-0.707460176459)*x[1]+(-0.275081507376)*x[2]
        s[1,3,0,0]=(0.692493088771)*x[0]+(-1.04755874336)*x[1]+(-0.373632326826)*x[2]
        s[1,3,0,1]=(0.717001163418)*x[0]+(-0.782379062328)*x[1]+(0.479776988529)*x[2]
        s[1,3,1,0]=(-0.733062748373)*x[0]+(-0.217972411497)*x[1]+(1.43785098293)*x[2]
        s[1,3,1,1]=(0.186430904908)*x[0]+(0.375218175232)*x[1]+(-0.461668650588)*x[2]
        s[1,3,2,0]=(-1.12951451301)*x[0]+(-1.07034955937)*x[1]+(-0.59888614247)*x[2]
        s[1,3,2,1]=(1.34023062833)*x[0]+(-0.520544300692)*x[1]+(-0.363001059538)*x[2]
        s[1,4,0,0]=(0.42295611834)*x[0]+(-0.718852108989)*x[1]+(-1.61995684533)*x[2]
        s[1,4,0,1]=(-0.422467947319)*x[0]+(-0.209334480945)*x[1]+(1.59579413592)*x[2]
        s[1,4,1,0]=(0.387543472208)*x[0]+(-0.0579909254613)*x[1]+(-1.03104850867)*x[2]
        s[1,4,1,1]=(-1.15893685519)*x[0]+(-0.971412946627)*x[1]+(-0.583440884462)*x[2]
        s[1,4,2,0]=(0.554042723162)*x[0]+(0.082633597735)*x[1]+(0.990177730288)*x[2]
        s[1,4,2,1]=(0.150114027051)*x[0]+(0.479664757576)*x[1]+(-1.5769513289)*x[2]
        s[2,0,0,0]=(1.33963588901)*x[0]+(0.146971184264)*x[1]+(-1.29921529821)*x[2]
        s[2,0,0,1]=(0.0115342792238)*x[0]+(0.0451643711892)*x[1]+(-0.656407384758)*x[2]
        s[2,0,1,0]=(0.175368903926)*x[0]+(-0.40573949846)*x[1]+(0.295387861218)*x[2]
        s[2,0,1,1]=(0.531254864472)*x[0]+(-0.536193761259)*x[1]+(-1.21752823876)*x[2]
        s[2,0,2,0]=(-0.362296220089)*x[0]+(0.378651497918)*x[1]+(-0.535206713131)*x[2]
        s[2,0,2,1]=(-0.85728549969)*x[0]+(0.233179293609)*x[1]+(0.345412963315)*x[2]
        s[2,1,0,0]=(1.01059063177)*x[0]+(0.808116187589)*x[1]+(0.468588054081)*x[2]
        s[2,1,0,1]=(-1.17212735529)*x[0]+(-0.871560017197)*x[1]+(1.10553749428)*x[2]
        s[2,1,1,0]=(1.00177793873)*x[0]+(-0.573447532545)*x[1]+(-0.80780539944)*x[2]
        s[2,1,1,1]=(-0.101912168689)*x[0]+(0.0054724970873)*x[1]+(0.873799338046)*x[2]
        s[2,1,2,0]=(0.162601365201)*x[0]+(-0.508620489728)*x[1]+(0.114861348434)*x[2]
        s[2,1,2,1]=(-1.48812556245)*x[0]+(0.29841054692)*x[1]+(0.773432180321)*x[2]
        s[2,2,0,0]=(0.0177469158963)*x[0]+(0.140208281265)*x[1]+(1.12393579682)*x[2]
        s[2,2,0,1]=(0.127835907966)*x[0]+(0.150072336346)*x[1]+(1.59437274191)*x[2]
        s[2,2,1,0]=(-0.549906307264)*x[0]+(-0.572101026775)*x[1]+(-0.747028158715)*x[2]
        s[2,2,1,1]=(0.134272548587)*x[0]+(0.866777056589)*x[1]+(-0.0658173600046)*x[2]
        s[2,2,2,0]=(0.160006947347)*x[0]+(-0.940281662941)*x[1]+(0.671280858545)*x[2]
        s[2,2,2,1]=(-1.55550646869)*x[0]+(0.0485277897059)*x[1]+(1.02027216264)*x[2]
        s[2,3,0,0]=(0.0931028968111)*x[0]+(0.194744643733)*x[1]+(1.7368246452)*x[2]
        s[2,3,0,1]=(1.66077510679)*x[0]+(-0.342785893413)*x[1]+(-0.278302501974)*x[2]
        s[2,3,1,0]=(-0.0719600278272)*x[0]+(0.898388137955)*x[1]+(-0.110811429033)*x[2]
        s[2,3,1,1]=(-0.781997247811)*x[0]+(-1.32497050043)*x[1]+(0.544850665961)*x[2]
        s[2,3,2,0]=(0.0290847482287)*x[0]+(0.895295560038)*x[1]+(-1.68555124085)*x[2]
        s[2,3,2,1]=(0.200540179332)*x[0]+(-0.843842814948)*x[1]+(-0.943779465271)*x[2]
        s[2,4,0,0]=(1.08999954494)*x[0]+(-1.32324120166)*x[1]+(-0.0152368215856)*x[2]
        s[2,4,0,1]=(1.02219584162)*x[0]+(0.208263165731)*x[1]+(0.0703408478849)*x[2]
        s[2,4,1,0]=(0.433866479442)*x[0]+(0.327590037731)*x[1]+(0.752498033029)*x[2]
        s[2,4,1,1]=(1.16322078616)*x[0]+(-1.38234503561)*x[1]+(-0.129845447078)*x[2]
        s[2,4,2,0]=(-0.430572860596)*x[0]+(-1.48079933792)*x[1]+(0.310197458762)*x[2]
        s[2,4,2,1]=(-0.0901138166505)*x[0]+(0.130679443905)*x[1]+(-0.482309483824)*x[2]
        s[3,0,0,0]=(0.239787389644)*x[0]+(0.690754646072)*x[1]+(0.745418481522)*x[2]
        s[3,0,0,1]=(0.346448639801)*x[0]+(-0.287823450604)*x[1]+(0.287003737791)*x[2]
        s[3,0,1,0]=(-1.50502949668)*x[0]+(-1.67521500964)*x[1]+(0.31630953667)*x[2]
        s[3,0,1,1]=(-0.594355763268)*x[0]+(-0.262023313233)*x[1]+(-1.41203969641)*x[2]
        s[3,0,2,0]=(-0.0275545216089)*x[0]+(0.367433407785)*x[1]+(-1.420568196)*x[2]
        s[3,0,2,1]=(-1.18797738587)*x[0]+(-0.496775264183)*x[1]+(0.513942381739)*x[2]
        s[3,1,0,0]=(0.475014043678)*x[0]+(-1.2205702601)*x[1]+(1.06353608529)*x[2]
        s[3,1,0,1]=(-0.0685583494755)*x[0]+(-1.49448219897)*x[1]+(-1.23320402959)*x[2]
        s[3,1,1,0]=(-0.216184790554)*x[0]+(0.554046932828)*x[1]+(0.697487187904)*x[2]
        s[3,1,1,1]=(-0.495370181694)*x[0]+(-1.49207400185)*x[1]+(-0.233763280927)*x[2]
        s[3,1,2,0]=(-0.0462888472218)*x[0]+(0.371127867531)*x[1]+(0.534977833221)*x[2]
        s[3,1,2,1]=(-1.27057460029)*x[0]+(-0.572725702684)*x[1]+(-0.871217452212)*x[2]
        s[3,2,0,0]=(-0.208549808264)*x[0]+(0.847294134183)*x[1]+(0.643419233684)*x[2]
        s[3,2,0,1]=(-0.850445146291)*x[0]+(1.19605447923)*x[1]+(-1.52958469348)*x[2]
        s[3,2,1,0]=(0.181672364304)*x[0]+(-1.60804541136)*x[1]+(0.068065187694)*x[2]
        s[3,2,1,1]=(-0.195284925405)*x[0]+(0.330399376496)*x[1]+(-0.453658972198)*x[2]
        s[3,2,2,0]=(1.44390910788)*x[0]+(0.718618445973)*x[1]+(-0.800501828022)*x[2]
        s[3,2,2,1]=(0.491997052311)*x[0]+(-0.202966322046)*x[1]+(1.39185177314)*x[2]
        s[3,3,0,0]=(-0.670818169452)*x[0]+(0.842747476683)*x[1]+(-0.705022077785)*x[2]
        s[3,3,0,1]=(0.729442429399)*x[0]+(-1.47998423588)*x[1]+(1.72626243143)*x[2]
        s[3,3,1,0]=(-1.08179915812)*x[0]+(-0.117953758788)*x[1]+(1.60080937509)*x[2]
        s[3,3,1,1]=(0.110049458343)*x[0]+(0.571461685559)*x[1]+(-0.483581817598)*x[2]
        s[3,3,2,0]=(-0.386163706678)*x[0]+(-0.146336916781)*x[1]+(0.0360311139076)*x[2]
        s[3,3,2,1]=(-0.779748009644)*x[0]+(-0.780523381896)*x[1]+(-1.20157768536)*x[2]
        s[3,4,0,0]=(1.74580569218)*x[0]+(-0.318257555715)*x[1]+(-0.33745202709)*x[2]
        s[3,4,0,1]=(1.1463597084)*x[0]+(0.516488947224)*x[1]+(-0.205035082746)*x[2]
        s[3,4,1,0]=(0.647937211136)*x[0]+(0.574941676843)*x[1]+(-1.00930234083)*x[2]
        s[3,4,1,1]=(-0.79188909394)*x[0]+(0.770198112387)*x[1]+(-0.0146013355132)*x[2]
        s[3,4,2,0]=(-1.70734926495)*x[0]+(-1.01352599249)*x[1]+(0.878522778655)*x[2]
        s[3,4,2,1]=(0.0249166011531)*x[0]+(-0.0714570534569)*x[1]+(1.10212723792)*x[2]
        ref[0,0,0,0]=(-0.512691010939)*x_ref[0]+(-1.31295640543)*x_ref[1]+(1.45426060088)*x_ref[2]
        ref[0,0,0,1]=(-0.57748372698)*x_ref[0]+(0.215297337003)*x_ref[1]+(0.0853439946124)*x_ref[2]
        ref[0,0,1,0]=(-0.785306554604)*x_ref[0]+(-0.0274911695421)*x_ref[1]+(0.693440767035)*x_ref[2]
        ref[0,0,1,1]=(-0.56932407873)*x_ref[0]+(-0.981593188827)*x_ref[1]+(0.87532777027)*x_ref[2]
        ref[0,0,2,0]=(0.145925104392)*x_ref[0]+(1.45139125753)*x_ref[1]+(1.25041746183)*x_ref[2]
        ref[0,0,2,1]=(-1.70110072582)*x_ref[0]+(-0.0335136921878)*x_ref[1]+(0.276204889346)*x_ref[2]
        ref[0,1,0,0]=(-0.415026748605)*x_ref[0]+(-0.322339100074)*x_ref[1]+(1.1050553524)*x_ref[2]
        ref[0,1,0,1]=(0.267336702938)*x_ref[0]+(-1.44498697586)*x_ref[1]+(0.600455793139)*x_ref[2]
        ref[0,1,1,0]=(-0.46618843993)*x_ref[0]+(0.318230555457)*x_ref[1]+(0.281788961443)*x_ref[2]
        ref[0,1,1,1]=(-1.12963168407)*x_ref[0]+(-0.661100272253)*x_ref[1]+(0.293642402995)*x_ref[2]
        ref[0,1,2,0]=(-0.0041407279814)*x_ref[0]+(0.330916770241)*x_ref[1]+(1.30265962609)*x_ref[2]
        ref[0,1,2,1]=(1.37513192104)*x_ref[0]+(-1.39837371617)*x_ref[1]+(-1.23490743489)*x_ref[2]
        ref[0,2,0,0]=(0.240920374522)*x_ref[0]+(1.30092830627)*x_ref[1]+(0.189720240056)*x_ref[2]
        ref[0,2,0,1]=(0.741337711467)*x_ref[0]+(0.00893414596869)*x_ref[1]+(1.22438764071)*x_ref[2]
        ref[0,2,1,0]=(1.0524903095)*x_ref[0]+(-1.00088524111)*x_ref[1]+(0.985885896119)*x_ref[2]
        ref[0,2,1,1]=(-0.252168964788)*x_ref[0]+(1.2685968377)*x_ref[1]+(-0.025373437998)*x_ref[2]
        ref[0,2,2,0]=(0.728523707179)*x_ref[0]+(0.00656996105623)*x_ref[1]+(1.85098622635)*x_ref[2]
        ref[0,2,2,1]=(0.193123464463)*x_ref[0]+(-0.426452202023)*x_ref[1]+(-0.144471734372)*x_ref[2]
        ref[0,3,0,0]=(0.446506355449)*x_ref[0]+(0.379921574537)*x_ref[1]+(-0.303150079281)*x_ref[2]
        ref[0,3,0,1]=(0.455373734156)*x_ref[0]+(0.134600165673)*x_ref[1]+(-0.604957024603)*x_ref[2]
        ref[0,3,1,0]=(-1.63808806962)*x_ref[0]+(1.02897840042)*x_ref[1]+(1.18123170106)*x_ref[2]
        ref[0,3,1,1]=(0.587448315418)*x_ref[0]+(0.0991725296547)*x_ref[1]+(1.33682286695)*x_ref[2]
        ref[0,3,2,0]=(-1.12085613449)*x_ref[0]+(0.454771039944)*x_ref[1]+(-0.131089499156)*x_ref[2]
        ref[0,3,2,1]=(-0.351021639018)*x_ref[0]+(0.47905316732)*x_ref[1]+(-1.07449842534)*x_ref[2]
        ref[0,4,0,0]=(0.232526828642)*x_ref[0]+(1.33907066655)*x_ref[1]+(-0.699701401092)*x_ref[2]
        ref[0,4,0,1]=(0.244905848535)*x_ref[0]+(-0.799313917773)*x_ref[1]+(-0.16513296799)*x_ref[2]
        ref[0,4,1,0]=(-1.10152273815)*x_ref[0]+(0.898176161732)*x_ref[1]+(-0.101734265033)*x_ref[2]
        ref[0,4,1,1]=(-0.0804598444811)*x_ref[0]+(-0.77819195606)*x_ref[1]+(0.277776761782)*x_ref[2]
        ref[0,4,2,0]=(1.05979692744)*x_ref[0]+(-0.195285573862)*x_ref[1]+(-0.0920717991265)*x_ref[2]
        ref[0,4,2,1]=(0.44924595646)*x_ref[0]+(0.409652246508)*x_ref[1]+(-0.0705754685165)*x_ref[2]
        ref[1,0,0,0]=(1.14991675937)*x_ref[0]+(-0.435947647215)*x_ref[1]+(-0.45614944131)*x_ref[2]
        ref[1,0,0,1]=(0.306890350036)*x_ref[0]+(-1.4400848018)*x_ref[1]+(-0.744474442385)*x_ref[2]
        ref[1,0,1,0]=(1.07930565176)*x_ref[0]+(-0.273335650157)*x_ref[1]+(-0.230960378838)*x_ref[2]
        ref[1,0,1,1]=(-0.97797563017)*x_ref[0]+(-1.3159717465)*x_ref[1]+(-0.357902208089)*x_ref[2]
        ref[1,0,2,0]=(-1.02063328249)*x_ref[0]+(-1.18907185687)*x_ref[1]+(-0.509296457902)*x_ref[2]
        ref[1,0,2,1]=(-0.508446666232)*x_ref[0]+(-0.69279963667)*x_ref[1]+(-0.497990637708)*x_ref[2]
        ref[1,1,0,0]=(-0.476141863054)*x_ref[0]+(-0.180458121777)*x_ref[1]+(-0.318855084826)*x_ref[2]
        ref[1,1,0,1]=(0.480566371651)*x_ref[0]+(-1.74196768367)*x_ref[1]+(1.23057574863)*x_ref[2]
        ref[1,1,1,0]=(-0.0213138624722)*x_ref[0]+(-0.0846532154764)*x_ref[1]+(-0.61169792023)*x_ref[2]
        ref[1,1,1,1]=(-0.389381507111)*x_ref[0]+(0.436088515729)*x_ref[1]+(-0.603019053837)*x_ref[2]
        ref[1,1,2,0]=(-1.24478786129)*x_ref[0]+(0.908984432701)*x_ref[1]+(-1.49483946407)*x_ref[2]
        ref[1,1,2,1]=(0.866266264162)*x_ref[0]+(-0.190579076131)*x_ref[1]+(-1.63524384216)*x_ref[2]
        ref[1,2,0,0]=(0.0333832222755)*x_ref[0]+(-0.551591529565)*x_ref[1]+(1.22170194609)*x_ref[2]
        ref[1,2,0,1]=(-0.551682390802)*x_ref[0]+(0.223940652339)*x_ref[1]+(0.141063634211)*x_ref[2]
        ref[1,2,1,0]=(1.17072817252)*x_ref[0]+(0.28318356571)*x_ref[1]+(-0.788096581969)*x_ref[2]
        ref[1,2,1,1]=(-1.01576849129)*x_ref[0]+(-0.502669485722)*x_ref[1]+(-0.966677100703)*x_ref[2]
        ref[1,2,2,0]=(-1.59522074083)*x_ref[0]+(-1.16016984107)*x_ref[1]+(-0.117030183187)*x_ref[2]
        ref[1,2,2,1]=(-0.979165259497)*x_ref[0]+(-0.707460176459)*x_ref[1]+(-0.275081507376)*x_ref[2]
        ref[1,3,0,0]=(0.692493088771)*x_ref[0]+(-1.04755874336)*x_ref[1]+(-0.373632326826)*x_ref[2]
        ref[1,3,0,1]=(0.717001163418)*x_ref[0]+(-0.782379062328)*x_ref[1]+(0.479776988529)*x_ref[2]
        ref[1,3,1,0]=(-0.733062748373)*x_ref[0]+(-0.217972411497)*x_ref[1]+(1.43785098293)*x_ref[2]
        ref[1,3,1,1]=(0.186430904908)*x_ref[0]+(0.375218175232)*x_ref[1]+(-0.461668650588)*x_ref[2]
        ref[1,3,2,0]=(-1.12951451301)*x_ref[0]+(-1.07034955937)*x_ref[1]+(-0.59888614247)*x_ref[2]
        ref[1,3,2,1]=(1.34023062833)*x_ref[0]+(-0.520544300692)*x_ref[1]+(-0.363001059538)*x_ref[2]
        ref[1,4,0,0]=(0.42295611834)*x_ref[0]+(-0.718852108989)*x_ref[1]+(-1.61995684533)*x_ref[2]
        ref[1,4,0,1]=(-0.422467947319)*x_ref[0]+(-0.209334480945)*x_ref[1]+(1.59579413592)*x_ref[2]
        ref[1,4,1,0]=(0.387543472208)*x_ref[0]+(-0.0579909254613)*x_ref[1]+(-1.03104850867)*x_ref[2]
        ref[1,4,1,1]=(-1.15893685519)*x_ref[0]+(-0.971412946627)*x_ref[1]+(-0.583440884462)*x_ref[2]
        ref[1,4,2,0]=(0.554042723162)*x_ref[0]+(0.082633597735)*x_ref[1]+(0.990177730288)*x_ref[2]
        ref[1,4,2,1]=(0.150114027051)*x_ref[0]+(0.479664757576)*x_ref[1]+(-1.5769513289)*x_ref[2]
        ref[2,0,0,0]=(1.33963588901)*x_ref[0]+(0.146971184264)*x_ref[1]+(-1.29921529821)*x_ref[2]
        ref[2,0,0,1]=(0.0115342792238)*x_ref[0]+(0.0451643711892)*x_ref[1]+(-0.656407384758)*x_ref[2]
        ref[2,0,1,0]=(0.175368903926)*x_ref[0]+(-0.40573949846)*x_ref[1]+(0.295387861218)*x_ref[2]
        ref[2,0,1,1]=(0.531254864472)*x_ref[0]+(-0.536193761259)*x_ref[1]+(-1.21752823876)*x_ref[2]
        ref[2,0,2,0]=(-0.362296220089)*x_ref[0]+(0.378651497918)*x_ref[1]+(-0.535206713131)*x_ref[2]
        ref[2,0,2,1]=(-0.85728549969)*x_ref[0]+(0.233179293609)*x_ref[1]+(0.345412963315)*x_ref[2]
        ref[2,1,0,0]=(1.01059063177)*x_ref[0]+(0.808116187589)*x_ref[1]+(0.468588054081)*x_ref[2]
        ref[2,1,0,1]=(-1.17212735529)*x_ref[0]+(-0.871560017197)*x_ref[1]+(1.10553749428)*x_ref[2]
        ref[2,1,1,0]=(1.00177793873)*x_ref[0]+(-0.573447532545)*x_ref[1]+(-0.80780539944)*x_ref[2]
        ref[2,1,1,1]=(-0.101912168689)*x_ref[0]+(0.0054724970873)*x_ref[1]+(0.873799338046)*x_ref[2]
        ref[2,1,2,0]=(0.162601365201)*x_ref[0]+(-0.508620489728)*x_ref[1]+(0.114861348434)*x_ref[2]
        ref[2,1,2,1]=(-1.48812556245)*x_ref[0]+(0.29841054692)*x_ref[1]+(0.773432180321)*x_ref[2]
        ref[2,2,0,0]=(0.0177469158963)*x_ref[0]+(0.140208281265)*x_ref[1]+(1.12393579682)*x_ref[2]
        ref[2,2,0,1]=(0.127835907966)*x_ref[0]+(0.150072336346)*x_ref[1]+(1.59437274191)*x_ref[2]
        ref[2,2,1,0]=(-0.549906307264)*x_ref[0]+(-0.572101026775)*x_ref[1]+(-0.747028158715)*x_ref[2]
        ref[2,2,1,1]=(0.134272548587)*x_ref[0]+(0.866777056589)*x_ref[1]+(-0.0658173600046)*x_ref[2]
        ref[2,2,2,0]=(0.160006947347)*x_ref[0]+(-0.940281662941)*x_ref[1]+(0.671280858545)*x_ref[2]
        ref[2,2,2,1]=(-1.55550646869)*x_ref[0]+(0.0485277897059)*x_ref[1]+(1.02027216264)*x_ref[2]
        ref[2,3,0,0]=(0.0931028968111)*x_ref[0]+(0.194744643733)*x_ref[1]+(1.7368246452)*x_ref[2]
        ref[2,3,0,1]=(1.66077510679)*x_ref[0]+(-0.342785893413)*x_ref[1]+(-0.278302501974)*x_ref[2]
        ref[2,3,1,0]=(-0.0719600278272)*x_ref[0]+(0.898388137955)*x_ref[1]+(-0.110811429033)*x_ref[2]
        ref[2,3,1,1]=(-0.781997247811)*x_ref[0]+(-1.32497050043)*x_ref[1]+(0.544850665961)*x_ref[2]
        ref[2,3,2,0]=(0.0290847482287)*x_ref[0]+(0.895295560038)*x_ref[1]+(-1.68555124085)*x_ref[2]
        ref[2,3,2,1]=(0.200540179332)*x_ref[0]+(-0.843842814948)*x_ref[1]+(-0.943779465271)*x_ref[2]
        ref[2,4,0,0]=(1.08999954494)*x_ref[0]+(-1.32324120166)*x_ref[1]+(-0.0152368215856)*x_ref[2]
        ref[2,4,0,1]=(1.02219584162)*x_ref[0]+(0.208263165731)*x_ref[1]+(0.0703408478849)*x_ref[2]
        ref[2,4,1,0]=(0.433866479442)*x_ref[0]+(0.327590037731)*x_ref[1]+(0.752498033029)*x_ref[2]
        ref[2,4,1,1]=(1.16322078616)*x_ref[0]+(-1.38234503561)*x_ref[1]+(-0.129845447078)*x_ref[2]
        ref[2,4,2,0]=(-0.430572860596)*x_ref[0]+(-1.48079933792)*x_ref[1]+(0.310197458762)*x_ref[2]
        ref[2,4,2,1]=(-0.0901138166505)*x_ref[0]+(0.130679443905)*x_ref[1]+(-0.482309483824)*x_ref[2]
        ref[3,0,0,0]=(0.239787389644)*x_ref[0]+(0.690754646072)*x_ref[1]+(0.745418481522)*x_ref[2]
        ref[3,0,0,1]=(0.346448639801)*x_ref[0]+(-0.287823450604)*x_ref[1]+(0.287003737791)*x_ref[2]
        ref[3,0,1,0]=(-1.50502949668)*x_ref[0]+(-1.67521500964)*x_ref[1]+(0.31630953667)*x_ref[2]
        ref[3,0,1,1]=(-0.594355763268)*x_ref[0]+(-0.262023313233)*x_ref[1]+(-1.41203969641)*x_ref[2]
        ref[3,0,2,0]=(-0.0275545216089)*x_ref[0]+(0.367433407785)*x_ref[1]+(-1.420568196)*x_ref[2]
        ref[3,0,2,1]=(-1.18797738587)*x_ref[0]+(-0.496775264183)*x_ref[1]+(0.513942381739)*x_ref[2]
        ref[3,1,0,0]=(0.475014043678)*x_ref[0]+(-1.2205702601)*x_ref[1]+(1.06353608529)*x_ref[2]
        ref[3,1,0,1]=(-0.0685583494755)*x_ref[0]+(-1.49448219897)*x_ref[1]+(-1.23320402959)*x_ref[2]
        ref[3,1,1,0]=(-0.216184790554)*x_ref[0]+(0.554046932828)*x_ref[1]+(0.697487187904)*x_ref[2]
        ref[3,1,1,1]=(-0.495370181694)*x_ref[0]+(-1.49207400185)*x_ref[1]+(-0.233763280927)*x_ref[2]
        ref[3,1,2,0]=(-0.0462888472218)*x_ref[0]+(0.371127867531)*x_ref[1]+(0.534977833221)*x_ref[2]
        ref[3,1,2,1]=(-1.27057460029)*x_ref[0]+(-0.572725702684)*x_ref[1]+(-0.871217452212)*x_ref[2]
        ref[3,2,0,0]=(-0.208549808264)*x_ref[0]+(0.847294134183)*x_ref[1]+(0.643419233684)*x_ref[2]
        ref[3,2,0,1]=(-0.850445146291)*x_ref[0]+(1.19605447923)*x_ref[1]+(-1.52958469348)*x_ref[2]
        ref[3,2,1,0]=(0.181672364304)*x_ref[0]+(-1.60804541136)*x_ref[1]+(0.068065187694)*x_ref[2]
        ref[3,2,1,1]=(-0.195284925405)*x_ref[0]+(0.330399376496)*x_ref[1]+(-0.453658972198)*x_ref[2]
        ref[3,2,2,0]=(1.44390910788)*x_ref[0]+(0.718618445973)*x_ref[1]+(-0.800501828022)*x_ref[2]
        ref[3,2,2,1]=(0.491997052311)*x_ref[0]+(-0.202966322046)*x_ref[1]+(1.39185177314)*x_ref[2]
        ref[3,3,0,0]=(-0.670818169452)*x_ref[0]+(0.842747476683)*x_ref[1]+(-0.705022077785)*x_ref[2]
        ref[3,3,0,1]=(0.729442429399)*x_ref[0]+(-1.47998423588)*x_ref[1]+(1.72626243143)*x_ref[2]
        ref[3,3,1,0]=(-1.08179915812)*x_ref[0]+(-0.117953758788)*x_ref[1]+(1.60080937509)*x_ref[2]
        ref[3,3,1,1]=(0.110049458343)*x_ref[0]+(0.571461685559)*x_ref[1]+(-0.483581817598)*x_ref[2]
        ref[3,3,2,0]=(-0.386163706678)*x_ref[0]+(-0.146336916781)*x_ref[1]+(0.0360311139076)*x_ref[2]
        ref[3,3,2,1]=(-0.779748009644)*x_ref[0]+(-0.780523381896)*x_ref[1]+(-1.20157768536)*x_ref[2]
        ref[3,4,0,0]=(1.74580569218)*x_ref[0]+(-0.318257555715)*x_ref[1]+(-0.33745202709)*x_ref[2]
        ref[3,4,0,1]=(1.1463597084)*x_ref[0]+(0.516488947224)*x_ref[1]+(-0.205035082746)*x_ref[2]
        ref[3,4,1,0]=(0.647937211136)*x_ref[0]+(0.574941676843)*x_ref[1]+(-1.00930234083)*x_ref[2]
        ref[3,4,1,1]=(-0.79188909394)*x_ref[0]+(0.770198112387)*x_ref[1]+(-0.0146013355132)*x_ref[2]
        ref[3,4,2,0]=(-1.70734926495)*x_ref[0]+(-1.01352599249)*x_ref[1]+(0.878522778655)*x_ref[2]
        ref[3,4,2,1]=(0.0249166011531)*x_ref[0]+(-0.0714570534569)*x_ref[1]+(1.10212723792)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.550361277741)*x[0]**o+(-0.431343639093)*x[0]+(-0.822566478678)*x[1]**o+(0.04348108437)*x[1]
        ref=(-0.550361277741)*x_ref[0]**o+(-0.431343639093)*x_ref[0]+(-0.822566478678)*x_ref[1]**o+(0.04348108437)*x_ref[1]
      else:
        s=(0.127530018259)*x[0]**o+(0.898197612625)*x[0]+(0.91995886196)*x[1]**o+(0.209489509472)*x[1]+(0.254553892458)*x[2]**o+(-0.00541573953033)*x[2]
        ref=(0.127530018259)*x_ref[0]**o+(0.898197612625)*x_ref[0]+(0.91995886196)*x_ref[1]**o+(0.209489509472)*x_ref[1]+(0.254553892458)*x_ref[2]**o+(-0.00541573953033)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.84171096789)*x[0]**o+(0.528614089652)*x[0]+(-0.909660047031)*x[1]**o+(0.0218639321632)*x[1]
        s[1]=(-0.196234616294)*x[0]**o+(-0.430220669717)*x[0]+(0.403594243834)*x[1]**o+(-0.555455789431)*x[1]
        ref[0]=(-0.84171096789)*x_ref[0]**o+(0.528614089652)*x_ref[0]+(-0.909660047031)*x_ref[1]**o+(0.0218639321632)*x_ref[1]
        ref[1]=(-0.196234616294)*x_ref[0]**o+(-0.430220669717)*x_ref[0]+(0.403594243834)*x_ref[1]**o+(-0.555455789431)*x_ref[1]
      else:
        s[0]=(0.236033772163)*x[0]**o+(-0.0384017745802)*x[0]+(0.594646313286)*x[1]**o+(-0.806924545742)*x[1]+(0.932734743639)*x[2]**o+(0.847787230656)*x[2]
        s[1]=(-0.472954470826)*x[0]**o+(0.874442648071)*x[0]+(0.645990600767)*x[1]**o+(-0.702715273742)*x[1]+(-0.738405297118)*x[2]**o+(0.313975158833)*x[2]
        ref[0]=(0.236033772163)*x_ref[0]**o+(-0.0384017745802)*x_ref[0]+(0.594646313286)*x_ref[1]**o+(-0.806924545742)*x_ref[1]+(0.932734743639)*x_ref[2]**o+(0.847787230656)*x_ref[2]
        ref[1]=(-0.472954470826)*x_ref[0]**o+(0.874442648071)*x_ref[0]+(0.645990600767)*x_ref[1]**o+(-0.702715273742)*x_ref[1]+(-0.738405297118)*x_ref[2]**o+(0.313975158833)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.247732656593)*x[0]**o+(-0.964555728772)*x[0]+(-0.356955683726)*x[1]**o+(0.411681255483)*x[1]
        s[0,1]=(-0.123553741129)*x[0]**o+(-0.511682932244)*x[0]+(-0.472291564447)*x[1]**o+(0.520295617334)*x[1]
        s[0,2]=(0.901714918359)*x[0]**o+(0.145027792176)*x[0]+(-0.860009935335)*x[1]**o+(-0.940605686559)*x[1]
        s[0,3]=(0.54699645799)*x[0]**o+(-0.64710426401)*x[0]+(-0.34117383247)*x[1]**o+(0.54121464127)*x[1]
        s[0,4]=(0.468931826985)*x[0]**o+(0.924799717102)*x[0]+(0.941672574697)*x[1]**o+(-0.190797551241)*x[1]
        s[1,0]=(0.0383673807592)*x[0]**o+(0.613252630082)*x[0]+(0.964352561462)*x[1]**o+(0.785919255777)*x[1]
        s[1,1]=(-0.355086324569)*x[0]**o+(0.558878226045)*x[0]+(-0.628061734673)*x[1]**o+(-0.520982697177)*x[1]
        s[1,2]=(0.723496483187)*x[0]**o+(-0.232875759963)*x[0]+(-0.17448761332)*x[1]**o+(-0.257229542886)*x[1]
        s[1,3]=(0.611356020012)*x[0]**o+(-0.28713845432)*x[0]+(-0.316681374895)*x[1]**o+(0.83320730326)*x[1]
        s[1,4]=(-0.984450469255)*x[0]**o+(0.832904475807)*x[0]+(0.150921155673)*x[1]**o+(0.158270289725)*x[1]
        s[2,0]=(0.281985706195)*x[0]**o+(0.292760063689)*x[0]+(-0.595863049075)*x[1]**o+(0.64338477328)*x[1]
        s[2,1]=(0.674281261333)*x[0]**o+(-0.282392899278)*x[0]+(0.323252947475)*x[1]**o+(-0.692156142053)*x[1]
        s[2,2]=(-0.483800717142)*x[0]**o+(-0.426306421122)*x[0]+(-0.778645643106)*x[1]**o+(0.243301515533)*x[1]
        s[2,3]=(-0.00888255736078)*x[0]**o+(-0.438384491475)*x[0]+(-0.143871169727)*x[1]**o+(0.896491696215)*x[1]
        s[2,4]=(-0.810632478514)*x[0]**o+(-0.802769421087)*x[0]+(-0.666796590858)*x[1]**o+(0.0311747062735)*x[1]
        s[3,0]=(0.919881133239)*x[0]**o+(-0.2869105463)*x[0]+(-0.33204138861)*x[1]**o+(0.790928481885)*x[1]
        s[3,1]=(-0.698545028974)*x[0]**o+(-0.526842225737)*x[0]+(-0.61919152938)*x[1]**o+(0.942452744855)*x[1]
        s[3,2]=(-0.204090966647)*x[0]**o+(-0.425679526188)*x[0]+(-0.194975545878)*x[1]**o+(0.559023937073)*x[1]
        s[3,3]=(0.852288716375)*x[0]**o+(-0.949732191466)*x[0]+(-0.0115651638362)*x[1]**o+(0.271548550263)*x[1]
        s[3,4]=(-0.684283367871)*x[0]**o+(0.887910737342)*x[0]+(-0.171186272812)*x[1]**o+(-0.069360036001)*x[1]
        ref[0,0]=(-0.247732656593)*x_ref[0]**o+(-0.964555728772)*x_ref[0]+(-0.356955683726)*x_ref[1]**o+(0.411681255483)*x_ref[1]
        ref[0,1]=(-0.123553741129)*x_ref[0]**o+(-0.511682932244)*x_ref[0]+(-0.472291564447)*x_ref[1]**o+(0.520295617334)*x_ref[1]
        ref[0,2]=(0.901714918359)*x_ref[0]**o+(0.145027792176)*x_ref[0]+(-0.860009935335)*x_ref[1]**o+(-0.940605686559)*x_ref[1]
        ref[0,3]=(0.54699645799)*x_ref[0]**o+(-0.64710426401)*x_ref[0]+(-0.34117383247)*x_ref[1]**o+(0.54121464127)*x_ref[1]
        ref[0,4]=(0.468931826985)*x_ref[0]**o+(0.924799717102)*x_ref[0]+(0.941672574697)*x_ref[1]**o+(-0.190797551241)*x_ref[1]
        ref[1,0]=(0.0383673807592)*x_ref[0]**o+(0.613252630082)*x_ref[0]+(0.964352561462)*x_ref[1]**o+(0.785919255777)*x_ref[1]
        ref[1,1]=(-0.355086324569)*x_ref[0]**o+(0.558878226045)*x_ref[0]+(-0.628061734673)*x_ref[1]**o+(-0.520982697177)*x_ref[1]
        ref[1,2]=(0.723496483187)*x_ref[0]**o+(-0.232875759963)*x_ref[0]+(-0.17448761332)*x_ref[1]**o+(-0.257229542886)*x_ref[1]
        ref[1,3]=(0.611356020012)*x_ref[0]**o+(-0.28713845432)*x_ref[0]+(-0.316681374895)*x_ref[1]**o+(0.83320730326)*x_ref[1]
        ref[1,4]=(-0.984450469255)*x_ref[0]**o+(0.832904475807)*x_ref[0]+(0.150921155673)*x_ref[1]**o+(0.158270289725)*x_ref[1]
        ref[2,0]=(0.281985706195)*x_ref[0]**o+(0.292760063689)*x_ref[0]+(-0.595863049075)*x_ref[1]**o+(0.64338477328)*x_ref[1]
        ref[2,1]=(0.674281261333)*x_ref[0]**o+(-0.282392899278)*x_ref[0]+(0.323252947475)*x_ref[1]**o+(-0.692156142053)*x_ref[1]
        ref[2,2]=(-0.483800717142)*x_ref[0]**o+(-0.426306421122)*x_ref[0]+(-0.778645643106)*x_ref[1]**o+(0.243301515533)*x_ref[1]
        ref[2,3]=(-0.00888255736078)*x_ref[0]**o+(-0.438384491475)*x_ref[0]+(-0.143871169727)*x_ref[1]**o+(0.896491696215)*x_ref[1]
        ref[2,4]=(-0.810632478514)*x_ref[0]**o+(-0.802769421087)*x_ref[0]+(-0.666796590858)*x_ref[1]**o+(0.0311747062735)*x_ref[1]
        ref[3,0]=(0.919881133239)*x_ref[0]**o+(-0.2869105463)*x_ref[0]+(-0.33204138861)*x_ref[1]**o+(0.790928481885)*x_ref[1]
        ref[3,1]=(-0.698545028974)*x_ref[0]**o+(-0.526842225737)*x_ref[0]+(-0.61919152938)*x_ref[1]**o+(0.942452744855)*x_ref[1]
        ref[3,2]=(-0.204090966647)*x_ref[0]**o+(-0.425679526188)*x_ref[0]+(-0.194975545878)*x_ref[1]**o+(0.559023937073)*x_ref[1]
        ref[3,3]=(0.852288716375)*x_ref[0]**o+(-0.949732191466)*x_ref[0]+(-0.0115651638362)*x_ref[1]**o+(0.271548550263)*x_ref[1]
        ref[3,4]=(-0.684283367871)*x_ref[0]**o+(0.887910737342)*x_ref[0]+(-0.171186272812)*x_ref[1]**o+(-0.069360036001)*x_ref[1]
      else:
        s[0,0]=(-0.146549436504)*x[0]**o+(-0.929982499842)*x[0]+(0.0884445727171)*x[1]**o+(-0.396974650474)*x[1]+(-0.0647429699084)*x[2]**o+(0.109485258811)*x[2]
        s[0,1]=(0.542977833506)*x[0]**o+(0.251279104804)*x[0]+(-0.590300769519)*x[1]**o+(-0.0490256576926)*x[1]+(-0.706162640719)*x[2]**o+(0.307510689521)*x[2]
        s[0,2]=(0.239778231777)*x[0]**o+(-0.790527320914)*x[0]+(0.344144886282)*x[1]**o+(-0.143939808236)*x[1]+(-0.799065328837)*x[2]**o+(0.297556608867)*x[2]
        s[0,3]=(-0.657196676576)*x[0]**o+(0.750906099544)*x[0]+(0.982936179835)*x[1]**o+(-0.503213227605)*x[1]+(0.733421910241)*x[2]**o+(0.484345689534)*x[2]
        s[0,4]=(-0.10428912667)*x[0]**o+(-0.955639074345)*x[0]+(-0.120263699632)*x[1]**o+(0.358019739226)*x[1]+(-0.967358029478)*x[2]**o+(0.515480564718)*x[2]
        s[1,0]=(0.762873752492)*x[0]**o+(-0.865993029835)*x[0]+(-0.227215328566)*x[1]**o+(0.628614878944)*x[1]+(-0.481609778611)*x[2]**o+(0.680214047083)*x[2]
        s[1,1]=(0.942225721603)*x[0]**o+(0.53391180863)*x[0]+(0.631906892953)*x[1]**o+(-0.907916670051)*x[1]+(-0.621285107992)*x[2]**o+(-0.613776303306)*x[2]
        s[1,2]=(-0.911879095454)*x[0]**o+(0.79055949266)*x[0]+(0.165379072622)*x[1]**o+(0.350589081179)*x[1]+(0.980050071498)*x[2]**o+(-0.728078468886)*x[2]
        s[1,3]=(0.927875284375)*x[0]**o+(0.267789172844)*x[0]+(-0.801387363889)*x[1]**o+(0.445894170128)*x[1]+(0.978344523663)*x[2]**o+(0.290141654981)*x[2]
        s[1,4]=(0.700012915504)*x[0]**o+(-0.247394585438)*x[0]+(-0.889582786398)*x[1]**o+(0.65907227973)*x[1]+(-0.0559926046073)*x[2]**o+(-0.954804094449)*x[2]
        s[2,0]=(0.303693881754)*x[0]**o+(0.409350273091)*x[0]+(0.611015645589)*x[1]**o+(-0.0919333788167)*x[1]+(0.161448267028)*x[2]**o+(0.200336560957)*x[2]
        s[2,1]=(0.010588069689)*x[0]**o+(-0.193147278154)*x[0]+(-0.846857030144)*x[1]**o+(-0.167769695014)*x[1]+(0.518409682456)*x[2]**o+(-0.882303521363)*x[2]
        s[2,2]=(0.82323572893)*x[0]**o+(0.712556931497)*x[0]+(0.677658458739)*x[1]**o+(0.582443300339)*x[1]+(0.291648194217)*x[2]**o+(-0.485160007834)*x[2]
        s[2,3]=(0.782471869781)*x[0]**o+(-0.841455423835)*x[0]+(0.546825262189)*x[1]**o+(0.529412907865)*x[1]+(0.0226844946614)*x[2]**o+(0.712355610287)*x[2]
        s[2,4]=(-0.0954280871852)*x[0]**o+(0.535095858201)*x[0]+(-0.107028351709)*x[1]**o+(0.446785157036)*x[1]+(-0.31582995285)*x[2]**o+(-0.00891179040655)*x[2]
        s[3,0]=(0.0804857346776)*x[0]**o+(-0.281583457975)*x[0]+(0.220634050401)*x[1]**o+(0.876350726718)*x[1]+(0.790789468637)*x[2]**o+(0.202028059276)*x[2]
        s[3,1]=(0.545828636902)*x[0]**o+(0.166260836796)*x[0]+(-0.357145146842)*x[1]**o+(-0.747029491567)*x[1]+(-0.965612583637)*x[2]**o+(0.863127371545)*x[2]
        s[3,2]=(-0.829350796721)*x[0]**o+(0.691296313876)*x[0]+(-0.170515248987)*x[1]**o+(-0.739412695686)*x[1]+(-0.244741653636)*x[2]**o+(-0.414333914881)*x[2]
        s[3,3]=(-0.636245569881)*x[0]**o+(-0.374567359716)*x[0]+(-0.80804727564)*x[1]**o+(0.551213025284)*x[1]+(-0.210200137584)*x[2]**o+(-0.975486200974)*x[2]
        s[3,4]=(0.195033054442)*x[0]**o+(-0.5792439888)*x[0]+(-0.366295355782)*x[1]**o+(-0.88151827146)*x[1]+(-0.438942980834)*x[2]**o+(0.378273994019)*x[2]
        ref[0,0]=(-0.146549436504)*x_ref[0]**o+(-0.929982499842)*x_ref[0]+(0.0884445727171)*x_ref[1]**o+(-0.396974650474)*x_ref[1]+(-0.0647429699084)*x_ref[2]**o+(0.109485258811)*x_ref[2]
        ref[0,1]=(0.542977833506)*x_ref[0]**o+(0.251279104804)*x_ref[0]+(-0.590300769519)*x_ref[1]**o+(-0.0490256576926)*x_ref[1]+(-0.706162640719)*x_ref[2]**o+(0.307510689521)*x_ref[2]
        ref[0,2]=(0.239778231777)*x_ref[0]**o+(-0.790527320914)*x_ref[0]+(0.344144886282)*x_ref[1]**o+(-0.143939808236)*x_ref[1]+(-0.799065328837)*x_ref[2]**o+(0.297556608867)*x_ref[2]
        ref[0,3]=(-0.657196676576)*x_ref[0]**o+(0.750906099544)*x_ref[0]+(0.982936179835)*x_ref[1]**o+(-0.503213227605)*x_ref[1]+(0.733421910241)*x_ref[2]**o+(0.484345689534)*x_ref[2]
        ref[0,4]=(-0.10428912667)*x_ref[0]**o+(-0.955639074345)*x_ref[0]+(-0.120263699632)*x_ref[1]**o+(0.358019739226)*x_ref[1]+(-0.967358029478)*x_ref[2]**o+(0.515480564718)*x_ref[2]
        ref[1,0]=(0.762873752492)*x_ref[0]**o+(-0.865993029835)*x_ref[0]+(-0.227215328566)*x_ref[1]**o+(0.628614878944)*x_ref[1]+(-0.481609778611)*x_ref[2]**o+(0.680214047083)*x_ref[2]
        ref[1,1]=(0.942225721603)*x_ref[0]**o+(0.53391180863)*x_ref[0]+(0.631906892953)*x_ref[1]**o+(-0.907916670051)*x_ref[1]+(-0.621285107992)*x_ref[2]**o+(-0.613776303306)*x_ref[2]
        ref[1,2]=(-0.911879095454)*x_ref[0]**o+(0.79055949266)*x_ref[0]+(0.165379072622)*x_ref[1]**o+(0.350589081179)*x_ref[1]+(0.980050071498)*x_ref[2]**o+(-0.728078468886)*x_ref[2]
        ref[1,3]=(0.927875284375)*x_ref[0]**o+(0.267789172844)*x_ref[0]+(-0.801387363889)*x_ref[1]**o+(0.445894170128)*x_ref[1]+(0.978344523663)*x_ref[2]**o+(0.290141654981)*x_ref[2]
        ref[1,4]=(0.700012915504)*x_ref[0]**o+(-0.247394585438)*x_ref[0]+(-0.889582786398)*x_ref[1]**o+(0.65907227973)*x_ref[1]+(-0.0559926046073)*x_ref[2]**o+(-0.954804094449)*x_ref[2]
        ref[2,0]=(0.303693881754)*x_ref[0]**o+(0.409350273091)*x_ref[0]+(0.611015645589)*x_ref[1]**o+(-0.0919333788167)*x_ref[1]+(0.161448267028)*x_ref[2]**o+(0.200336560957)*x_ref[2]
        ref[2,1]=(0.010588069689)*x_ref[0]**o+(-0.193147278154)*x_ref[0]+(-0.846857030144)*x_ref[1]**o+(-0.167769695014)*x_ref[1]+(0.518409682456)*x_ref[2]**o+(-0.882303521363)*x_ref[2]
        ref[2,2]=(0.82323572893)*x_ref[0]**o+(0.712556931497)*x_ref[0]+(0.677658458739)*x_ref[1]**o+(0.582443300339)*x_ref[1]+(0.291648194217)*x_ref[2]**o+(-0.485160007834)*x_ref[2]
        ref[2,3]=(0.782471869781)*x_ref[0]**o+(-0.841455423835)*x_ref[0]+(0.546825262189)*x_ref[1]**o+(0.529412907865)*x_ref[1]+(0.0226844946614)*x_ref[2]**o+(0.712355610287)*x_ref[2]
        ref[2,4]=(-0.0954280871852)*x_ref[0]**o+(0.535095858201)*x_ref[0]+(-0.107028351709)*x_ref[1]**o+(0.446785157036)*x_ref[1]+(-0.31582995285)*x_ref[2]**o+(-0.00891179040655)*x_ref[2]
        ref[3,0]=(0.0804857346776)*x_ref[0]**o+(-0.281583457975)*x_ref[0]+(0.220634050401)*x_ref[1]**o+(0.876350726718)*x_ref[1]+(0.790789468637)*x_ref[2]**o+(0.202028059276)*x_ref[2]
        ref[3,1]=(0.545828636902)*x_ref[0]**o+(0.166260836796)*x_ref[0]+(-0.357145146842)*x_ref[1]**o+(-0.747029491567)*x_ref[1]+(-0.965612583637)*x_ref[2]**o+(0.863127371545)*x_ref[2]
        ref[3,2]=(-0.829350796721)*x_ref[0]**o+(0.691296313876)*x_ref[0]+(-0.170515248987)*x_ref[1]**o+(-0.739412695686)*x_ref[1]+(-0.244741653636)*x_ref[2]**o+(-0.414333914881)*x_ref[2]
        ref[3,3]=(-0.636245569881)*x_ref[0]**o+(-0.374567359716)*x_ref[0]+(-0.80804727564)*x_ref[1]**o+(0.551213025284)*x_ref[1]+(-0.210200137584)*x_ref[2]**o+(-0.975486200974)*x_ref[2]
        ref[3,4]=(0.195033054442)*x_ref[0]**o+(-0.5792439888)*x_ref[0]+(-0.366295355782)*x_ref[1]**o+(-0.88151827146)*x_ref[1]+(-0.438942980834)*x_ref[2]**o+(0.378273994019)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.0313098361481)*x[0]**o+(-0.943408799922)*x[0]+(0.695838509745)*x[1]**o+(-0.332495099387)*x[1]
        s[0,0,1]=(0.431627451331)*x[0]**o+(-0.628725729291)*x[0]+(-0.35286078386)*x[1]**o+(-0.955736740817)*x[1]
        s[0,1,0]=(0.968131554454)*x[0]**o+(0.0281368471267)*x[0]+(0.0459790991737)*x[1]**o+(0.393556307708)*x[1]
        s[0,1,1]=(-0.204304807192)*x[0]**o+(-0.475856856629)*x[0]+(-0.480008001652)*x[1]**o+(-0.550159532343)*x[1]
        s[1,0,0]=(-0.367566613332)*x[0]**o+(-0.538384769558)*x[0]+(0.96041325378)*x[1]**o+(0.435390124318)*x[1]
        s[1,0,1]=(0.204676073638)*x[0]**o+(0.250019350651)*x[0]+(0.103416708864)*x[1]**o+(-0.524130759283)*x[1]
        s[1,1,0]=(0.855261863743)*x[0]**o+(-0.675190292289)*x[0]+(-0.514777393648)*x[1]**o+(0.942267330164)*x[1]
        s[1,1,1]=(0.508909921316)*x[0]**o+(0.499552817754)*x[0]+(0.303561948422)*x[1]**o+(0.248598631154)*x[1]
        s[2,0,0]=(-0.642212923403)*x[0]**o+(0.956027472963)*x[0]+(0.056374728577)*x[1]**o+(0.486941288345)*x[1]
        s[2,0,1]=(-0.860437111201)*x[0]**o+(0.922745184749)*x[0]+(0.448750379137)*x[1]**o+(0.733106647733)*x[1]
        s[2,1,0]=(-0.399141476478)*x[0]**o+(-0.623902788197)*x[0]+(0.185646936042)*x[1]**o+(0.158930593879)*x[1]
        s[2,1,1]=(-0.25915892886)*x[0]**o+(-0.534454787558)*x[0]+(-0.0196197895513)*x[1]**o+(-0.921288033314)*x[1]
        s[3,0,0]=(0.718044900278)*x[0]**o+(-0.881912685669)*x[0]+(0.328338604205)*x[1]**o+(-0.364500357873)*x[1]
        s[3,0,1]=(-0.316477236945)*x[0]**o+(-0.0594343871907)*x[0]+(0.564350847567)*x[1]**o+(0.612426270173)*x[1]
        s[3,1,0]=(0.862764583147)*x[0]**o+(0.0895137700398)*x[0]+(0.376837454709)*x[1]**o+(-0.276425178685)*x[1]
        s[3,1,1]=(0.911550499093)*x[0]**o+(0.47964604046)*x[0]+(-0.0936664482923)*x[1]**o+(-0.355680673199)*x[1]
        s[4,0,0]=(-0.126679727917)*x[0]**o+(0.0230703577025)*x[0]+(-0.749904232032)*x[1]**o+(0.120202683819)*x[1]
        s[4,0,1]=(0.859805437446)*x[0]**o+(-0.0547285079407)*x[0]+(-0.0209396655171)*x[1]**o+(-0.835017485682)*x[1]
        s[4,1,0]=(0.223141900386)*x[0]**o+(0.570612780326)*x[0]+(0.147581871479)*x[1]**o+(-0.0724129492652)*x[1]
        s[4,1,1]=(0.479285654861)*x[0]**o+(-0.368842797927)*x[0]+(0.516904484685)*x[1]**o+(-0.74560082895)*x[1]
        s[5,0,0]=(0.895865547536)*x[0]**o+(0.354087196226)*x[0]+(-0.596919891679)*x[1]**o+(0.512426451597)*x[1]
        s[5,0,1]=(0.467255278736)*x[0]**o+(-0.724626510449)*x[0]+(0.0624000180202)*x[1]**o+(0.860859099558)*x[1]
        s[5,1,0]=(-0.632133782251)*x[0]**o+(0.639067592984)*x[0]+(0.974027711622)*x[1]**o+(0.835256883107)*x[1]
        s[5,1,1]=(0.0618918945789)*x[0]**o+(0.541323411311)*x[0]+(-0.229029906018)*x[1]**o+(0.489605390069)*x[1]
        ref[0,0,0]=(0.0313098361481)*x_ref[0]**o+(-0.943408799922)*x_ref[0]+(0.695838509745)*x_ref[1]**o+(-0.332495099387)*x_ref[1]
        ref[0,0,1]=(0.431627451331)*x_ref[0]**o+(-0.628725729291)*x_ref[0]+(-0.35286078386)*x_ref[1]**o+(-0.955736740817)*x_ref[1]
        ref[0,1,0]=(0.968131554454)*x_ref[0]**o+(0.0281368471267)*x_ref[0]+(0.0459790991737)*x_ref[1]**o+(0.393556307708)*x_ref[1]
        ref[0,1,1]=(-0.204304807192)*x_ref[0]**o+(-0.475856856629)*x_ref[0]+(-0.480008001652)*x_ref[1]**o+(-0.550159532343)*x_ref[1]
        ref[1,0,0]=(-0.367566613332)*x_ref[0]**o+(-0.538384769558)*x_ref[0]+(0.96041325378)*x_ref[1]**o+(0.435390124318)*x_ref[1]
        ref[1,0,1]=(0.204676073638)*x_ref[0]**o+(0.250019350651)*x_ref[0]+(0.103416708864)*x_ref[1]**o+(-0.524130759283)*x_ref[1]
        ref[1,1,0]=(0.855261863743)*x_ref[0]**o+(-0.675190292289)*x_ref[0]+(-0.514777393648)*x_ref[1]**o+(0.942267330164)*x_ref[1]
        ref[1,1,1]=(0.508909921316)*x_ref[0]**o+(0.499552817754)*x_ref[0]+(0.303561948422)*x_ref[1]**o+(0.248598631154)*x_ref[1]
        ref[2,0,0]=(-0.642212923403)*x_ref[0]**o+(0.956027472963)*x_ref[0]+(0.056374728577)*x_ref[1]**o+(0.486941288345)*x_ref[1]
        ref[2,0,1]=(-0.860437111201)*x_ref[0]**o+(0.922745184749)*x_ref[0]+(0.448750379137)*x_ref[1]**o+(0.733106647733)*x_ref[1]
        ref[2,1,0]=(-0.399141476478)*x_ref[0]**o+(-0.623902788197)*x_ref[0]+(0.185646936042)*x_ref[1]**o+(0.158930593879)*x_ref[1]
        ref[2,1,1]=(-0.25915892886)*x_ref[0]**o+(-0.534454787558)*x_ref[0]+(-0.0196197895513)*x_ref[1]**o+(-0.921288033314)*x_ref[1]
        ref[3,0,0]=(0.718044900278)*x_ref[0]**o+(-0.881912685669)*x_ref[0]+(0.328338604205)*x_ref[1]**o+(-0.364500357873)*x_ref[1]
        ref[3,0,1]=(-0.316477236945)*x_ref[0]**o+(-0.0594343871907)*x_ref[0]+(0.564350847567)*x_ref[1]**o+(0.612426270173)*x_ref[1]
        ref[3,1,0]=(0.862764583147)*x_ref[0]**o+(0.0895137700398)*x_ref[0]+(0.376837454709)*x_ref[1]**o+(-0.276425178685)*x_ref[1]
        ref[3,1,1]=(0.911550499093)*x_ref[0]**o+(0.47964604046)*x_ref[0]+(-0.0936664482923)*x_ref[1]**o+(-0.355680673199)*x_ref[1]
        ref[4,0,0]=(-0.126679727917)*x_ref[0]**o+(0.0230703577025)*x_ref[0]+(-0.749904232032)*x_ref[1]**o+(0.120202683819)*x_ref[1]
        ref[4,0,1]=(0.859805437446)*x_ref[0]**o+(-0.0547285079407)*x_ref[0]+(-0.0209396655171)*x_ref[1]**o+(-0.835017485682)*x_ref[1]
        ref[4,1,0]=(0.223141900386)*x_ref[0]**o+(0.570612780326)*x_ref[0]+(0.147581871479)*x_ref[1]**o+(-0.0724129492652)*x_ref[1]
        ref[4,1,1]=(0.479285654861)*x_ref[0]**o+(-0.368842797927)*x_ref[0]+(0.516904484685)*x_ref[1]**o+(-0.74560082895)*x_ref[1]
        ref[5,0,0]=(0.895865547536)*x_ref[0]**o+(0.354087196226)*x_ref[0]+(-0.596919891679)*x_ref[1]**o+(0.512426451597)*x_ref[1]
        ref[5,0,1]=(0.467255278736)*x_ref[0]**o+(-0.724626510449)*x_ref[0]+(0.0624000180202)*x_ref[1]**o+(0.860859099558)*x_ref[1]
        ref[5,1,0]=(-0.632133782251)*x_ref[0]**o+(0.639067592984)*x_ref[0]+(0.974027711622)*x_ref[1]**o+(0.835256883107)*x_ref[1]
        ref[5,1,1]=(0.0618918945789)*x_ref[0]**o+(0.541323411311)*x_ref[0]+(-0.229029906018)*x_ref[1]**o+(0.489605390069)*x_ref[1]
      else:
        s[0,0,0]=(-0.110630395456)*x[0]**o+(0.409103918414)*x[0]+(0.574294002864)*x[1]**o+(0.187245897037)*x[1]+(-0.577953556131)*x[2]**o+(-0.0999500699254)*x[2]
        s[0,0,1]=(0.60081436578)*x[0]**o+(0.0576573739024)*x[0]+(-0.151144644304)*x[1]**o+(-0.38821542392)*x[1]+(-0.698001301019)*x[2]**o+(0.959658889512)*x[2]
        s[0,1,0]=(-0.0612229861582)*x[0]**o+(-0.221080853704)*x[0]+(-0.806155040024)*x[1]**o+(0.810149964095)*x[1]+(-0.970105552614)*x[2]**o+(0.901396266753)*x[2]
        s[0,1,1]=(-0.1815051419)*x[0]**o+(0.205375880458)*x[0]+(-0.865264408549)*x[1]**o+(0.76028929015)*x[1]+(0.520419877863)*x[2]**o+(0.593892129073)*x[2]
        s[1,0,0]=(0.585747762904)*x[0]**o+(0.483216361168)*x[0]+(0.5552914794)*x[1]**o+(0.297595584239)*x[1]+(-0.459345030181)*x[2]**o+(0.740748940805)*x[2]
        s[1,0,1]=(-0.482750725051)*x[0]**o+(0.588601093972)*x[0]+(0.440244419924)*x[1]**o+(0.222485858829)*x[1]+(0.797934157822)*x[2]**o+(0.15404361699)*x[2]
        s[1,1,0]=(0.498237939377)*x[0]**o+(0.0115319665364)*x[0]+(-0.295972109504)*x[1]**o+(0.256948190994)*x[1]+(-0.711317031697)*x[2]**o+(-0.00740724985641)*x[2]
        s[1,1,1]=(-0.861126786241)*x[0]**o+(0.371079744214)*x[0]+(-0.660762067211)*x[1]**o+(0.081182574056)*x[1]+(-0.77101451732)*x[2]**o+(0.899135035361)*x[2]
        s[2,0,0]=(0.666936736186)*x[0]**o+(0.951940175322)*x[0]+(-0.0863391086362)*x[1]**o+(0.057346020976)*x[1]+(0.442105193705)*x[2]**o+(-0.07460449107)*x[2]
        s[2,0,1]=(0.980045171639)*x[0]**o+(0.445573999083)*x[0]+(0.442430445618)*x[1]**o+(0.551873411845)*x[1]+(0.0674935335431)*x[2]**o+(0.772168296609)*x[2]
        s[2,1,0]=(-0.271924080302)*x[0]**o+(-0.081769641884)*x[0]+(-0.902170977359)*x[1]**o+(-0.37134444501)*x[1]+(0.490234732275)*x[2]**o+(-0.972142319324)*x[2]
        s[2,1,1]=(-0.931544234095)*x[0]**o+(-0.996346369309)*x[0]+(0.313123243249)*x[1]**o+(-0.377115311908)*x[1]+(-0.840592475346)*x[2]**o+(-0.147076866517)*x[2]
        s[3,0,0]=(0.920849350915)*x[0]**o+(-0.268166526715)*x[0]+(-0.639196917804)*x[1]**o+(0.847896119248)*x[1]+(-0.488855455712)*x[2]**o+(-0.893202374665)*x[2]
        s[3,0,1]=(-0.827357725384)*x[0]**o+(-0.675036050349)*x[0]+(-0.0259058469974)*x[1]**o+(-0.143686855851)*x[1]+(0.431815308934)*x[2]**o+(-0.123586502935)*x[2]
        s[3,1,0]=(-0.482077477865)*x[0]**o+(0.256319753359)*x[0]+(0.688028798818)*x[1]**o+(-0.0557357547278)*x[1]+(0.025118048506)*x[2]**o+(-0.53002694196)*x[2]
        s[3,1,1]=(-0.00203098763538)*x[0]**o+(0.922732878998)*x[0]+(0.470339945576)*x[1]**o+(-0.288881331631)*x[1]+(0.716132166051)*x[2]**o+(0.376878681481)*x[2]
        s[4,0,0]=(0.308528023012)*x[0]**o+(-0.68058295085)*x[0]+(-0.633458164572)*x[1]**o+(-0.236218088529)*x[1]+(-0.853026908063)*x[2]**o+(-0.560991626371)*x[2]
        s[4,0,1]=(-0.276542785164)*x[0]**o+(-0.858464870187)*x[0]+(-0.339731559043)*x[1]**o+(0.625570996156)*x[1]+(-0.971283564072)*x[2]**o+(-0.893431722273)*x[2]
        s[4,1,0]=(-0.132528753997)*x[0]**o+(0.223249232076)*x[0]+(0.841435681826)*x[1]**o+(-0.0253915085492)*x[1]+(-0.518508375076)*x[2]**o+(-0.698041771818)*x[2]
        s[4,1,1]=(-0.978796591572)*x[0]**o+(0.780188335993)*x[0]+(-0.812394692896)*x[1]**o+(0.518054410279)*x[1]+(0.571803768515)*x[2]**o+(-0.0269003440822)*x[2]
        s[5,0,0]=(0.562021340848)*x[0]**o+(0.68880638377)*x[0]+(0.85211165801)*x[1]**o+(0.0451146367354)*x[1]+(0.241282824032)*x[2]**o+(0.661805360853)*x[2]
        s[5,0,1]=(0.500871152579)*x[0]**o+(0.21948004589)*x[0]+(0.731267125182)*x[1]**o+(-0.0641860674019)*x[1]+(0.109187270896)*x[2]**o+(-0.843149143759)*x[2]
        s[5,1,0]=(0.478541279317)*x[0]**o+(-0.34649289488)*x[0]+(0.551551005174)*x[1]**o+(-0.750137447886)*x[1]+(-0.112460330983)*x[2]**o+(-0.731771671002)*x[2]
        s[5,1,1]=(-0.93838335953)*x[0]**o+(0.499456439132)*x[0]+(0.655920157621)*x[1]**o+(0.233198833152)*x[1]+(0.444839351692)*x[2]**o+(0.666750348313)*x[2]
        ref[0,0,0]=(-0.110630395456)*x_ref[0]**o+(0.409103918414)*x_ref[0]+(0.574294002864)*x_ref[1]**o+(0.187245897037)*x_ref[1]+(-0.577953556131)*x_ref[2]**o+(-0.0999500699254)*x_ref[2]
        ref[0,0,1]=(0.60081436578)*x_ref[0]**o+(0.0576573739024)*x_ref[0]+(-0.151144644304)*x_ref[1]**o+(-0.38821542392)*x_ref[1]+(-0.698001301019)*x_ref[2]**o+(0.959658889512)*x_ref[2]
        ref[0,1,0]=(-0.0612229861582)*x_ref[0]**o+(-0.221080853704)*x_ref[0]+(-0.806155040024)*x_ref[1]**o+(0.810149964095)*x_ref[1]+(-0.970105552614)*x_ref[2]**o+(0.901396266753)*x_ref[2]
        ref[0,1,1]=(-0.1815051419)*x_ref[0]**o+(0.205375880458)*x_ref[0]+(-0.865264408549)*x_ref[1]**o+(0.76028929015)*x_ref[1]+(0.520419877863)*x_ref[2]**o+(0.593892129073)*x_ref[2]
        ref[1,0,0]=(0.585747762904)*x_ref[0]**o+(0.483216361168)*x_ref[0]+(0.5552914794)*x_ref[1]**o+(0.297595584239)*x_ref[1]+(-0.459345030181)*x_ref[2]**o+(0.740748940805)*x_ref[2]
        ref[1,0,1]=(-0.482750725051)*x_ref[0]**o+(0.588601093972)*x_ref[0]+(0.440244419924)*x_ref[1]**o+(0.222485858829)*x_ref[1]+(0.797934157822)*x_ref[2]**o+(0.15404361699)*x_ref[2]
        ref[1,1,0]=(0.498237939377)*x_ref[0]**o+(0.0115319665364)*x_ref[0]+(-0.295972109504)*x_ref[1]**o+(0.256948190994)*x_ref[1]+(-0.711317031697)*x_ref[2]**o+(-0.00740724985641)*x_ref[2]
        ref[1,1,1]=(-0.861126786241)*x_ref[0]**o+(0.371079744214)*x_ref[0]+(-0.660762067211)*x_ref[1]**o+(0.081182574056)*x_ref[1]+(-0.77101451732)*x_ref[2]**o+(0.899135035361)*x_ref[2]
        ref[2,0,0]=(0.666936736186)*x_ref[0]**o+(0.951940175322)*x_ref[0]+(-0.0863391086362)*x_ref[1]**o+(0.057346020976)*x_ref[1]+(0.442105193705)*x_ref[2]**o+(-0.07460449107)*x_ref[2]
        ref[2,0,1]=(0.980045171639)*x_ref[0]**o+(0.445573999083)*x_ref[0]+(0.442430445618)*x_ref[1]**o+(0.551873411845)*x_ref[1]+(0.0674935335431)*x_ref[2]**o+(0.772168296609)*x_ref[2]
        ref[2,1,0]=(-0.271924080302)*x_ref[0]**o+(-0.081769641884)*x_ref[0]+(-0.902170977359)*x_ref[1]**o+(-0.37134444501)*x_ref[1]+(0.490234732275)*x_ref[2]**o+(-0.972142319324)*x_ref[2]
        ref[2,1,1]=(-0.931544234095)*x_ref[0]**o+(-0.996346369309)*x_ref[0]+(0.313123243249)*x_ref[1]**o+(-0.377115311908)*x_ref[1]+(-0.840592475346)*x_ref[2]**o+(-0.147076866517)*x_ref[2]
        ref[3,0,0]=(0.920849350915)*x_ref[0]**o+(-0.268166526715)*x_ref[0]+(-0.639196917804)*x_ref[1]**o+(0.847896119248)*x_ref[1]+(-0.488855455712)*x_ref[2]**o+(-0.893202374665)*x_ref[2]
        ref[3,0,1]=(-0.827357725384)*x_ref[0]**o+(-0.675036050349)*x_ref[0]+(-0.0259058469974)*x_ref[1]**o+(-0.143686855851)*x_ref[1]+(0.431815308934)*x_ref[2]**o+(-0.123586502935)*x_ref[2]
        ref[3,1,0]=(-0.482077477865)*x_ref[0]**o+(0.256319753359)*x_ref[0]+(0.688028798818)*x_ref[1]**o+(-0.0557357547278)*x_ref[1]+(0.025118048506)*x_ref[2]**o+(-0.53002694196)*x_ref[2]
        ref[3,1,1]=(-0.00203098763538)*x_ref[0]**o+(0.922732878998)*x_ref[0]+(0.470339945576)*x_ref[1]**o+(-0.288881331631)*x_ref[1]+(0.716132166051)*x_ref[2]**o+(0.376878681481)*x_ref[2]
        ref[4,0,0]=(0.308528023012)*x_ref[0]**o+(-0.68058295085)*x_ref[0]+(-0.633458164572)*x_ref[1]**o+(-0.236218088529)*x_ref[1]+(-0.853026908063)*x_ref[2]**o+(-0.560991626371)*x_ref[2]
        ref[4,0,1]=(-0.276542785164)*x_ref[0]**o+(-0.858464870187)*x_ref[0]+(-0.339731559043)*x_ref[1]**o+(0.625570996156)*x_ref[1]+(-0.971283564072)*x_ref[2]**o+(-0.893431722273)*x_ref[2]
        ref[4,1,0]=(-0.132528753997)*x_ref[0]**o+(0.223249232076)*x_ref[0]+(0.841435681826)*x_ref[1]**o+(-0.0253915085492)*x_ref[1]+(-0.518508375076)*x_ref[2]**o+(-0.698041771818)*x_ref[2]
        ref[4,1,1]=(-0.978796591572)*x_ref[0]**o+(0.780188335993)*x_ref[0]+(-0.812394692896)*x_ref[1]**o+(0.518054410279)*x_ref[1]+(0.571803768515)*x_ref[2]**o+(-0.0269003440822)*x_ref[2]
        ref[5,0,0]=(0.562021340848)*x_ref[0]**o+(0.68880638377)*x_ref[0]+(0.85211165801)*x_ref[1]**o+(0.0451146367354)*x_ref[1]+(0.241282824032)*x_ref[2]**o+(0.661805360853)*x_ref[2]
        ref[5,0,1]=(0.500871152579)*x_ref[0]**o+(0.21948004589)*x_ref[0]+(0.731267125182)*x_ref[1]**o+(-0.0641860674019)*x_ref[1]+(0.109187270896)*x_ref[2]**o+(-0.843149143759)*x_ref[2]
        ref[5,1,0]=(0.478541279317)*x_ref[0]**o+(-0.34649289488)*x_ref[0]+(0.551551005174)*x_ref[1]**o+(-0.750137447886)*x_ref[1]+(-0.112460330983)*x_ref[2]**o+(-0.731771671002)*x_ref[2]
        ref[5,1,1]=(-0.93838335953)*x_ref[0]**o+(0.499456439132)*x_ref[0]+(0.655920157621)*x_ref[1]**o+(0.233198833152)*x_ref[1]+(0.444839351692)*x_ref[2]**o+(0.666750348313)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedFunctionOnBoundary_fromSymbol_ReducedFunctionOnBoundary_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedFunctionOnBoundary

      assumptions: self.domain supports inpterpolation from ReducedFunctionOnBoundary onto ReducedFunctionOnBoundary
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunctionOnBoundary(self.domain)
      x_ref=w_ref.getX()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.771805529053)*x[0]**o+(-0.00794071414241)*x[0]+(-0.546566317942)*x[1]**o+(0.209469633398)*x[1]
        s[0,0,0,1]=(-0.0138300734518)*x[0]**o+(0.783507392816)*x[0]+(0.637453082951)*x[1]**o+(-0.0262186183303)*x[1]
        s[0,0,1,0]=(-0.622230122991)*x[0]**o+(0.760867623401)*x[0]+(0.340084769283)*x[1]**o+(0.493207077521)*x[1]
        s[0,0,1,1]=(0.772195055793)*x[0]**o+(-0.481066683803)*x[0]+(0.198371969404)*x[1]**o+(0.841664205482)*x[1]
        s[0,0,2,0]=(0.858213005013)*x[0]**o+(0.819391260667)*x[0]+(0.100404710733)*x[1]**o+(0.0757471444158)*x[1]
        s[0,0,2,1]=(0.865434306062)*x[0]**o+(0.411158028281)*x[0]+(0.717843255665)*x[1]**o+(-0.742991318999)*x[1]
        s[0,1,0,0]=(-0.851618984895)*x[0]**o+(0.251966806193)*x[0]+(-0.803642825762)*x[1]**o+(-0.0631169142808)*x[1]
        s[0,1,0,1]=(-0.758981775218)*x[0]**o+(-0.561508941004)*x[0]+(0.577901652618)*x[1]**o+(-0.725243471023)*x[1]
        s[0,1,1,0]=(0.742032672941)*x[0]**o+(-0.0526002854749)*x[0]+(0.252153928183)*x[1]**o+(-0.388978984633)*x[1]
        s[0,1,1,1]=(-0.980858344737)*x[0]**o+(-0.871658904716)*x[0]+(0.164033322064)*x[1]**o+(0.678658472838)*x[1]
        s[0,1,2,0]=(0.823816388003)*x[0]**o+(0.274859643706)*x[0]+(0.322475967163)*x[1]**o+(-0.598198426633)*x[1]
        s[0,1,2,1]=(0.481980285537)*x[0]**o+(-0.806234350243)*x[0]+(0.564463264623)*x[1]**o+(-0.548427948656)*x[1]
        s[0,2,0,0]=(0.582989836924)*x[0]**o+(0.293733803309)*x[0]+(0.0589622495939)*x[1]**o+(-0.661119752655)*x[1]
        s[0,2,0,1]=(0.675133991144)*x[0]**o+(-0.964990891523)*x[0]+(-0.488115773474)*x[1]**o+(-0.601204526426)*x[1]
        s[0,2,1,0]=(-0.504646764871)*x[0]**o+(-0.496630864273)*x[0]+(-0.533328001228)*x[1]**o+(0.666398746532)*x[1]
        s[0,2,1,1]=(-0.947971259712)*x[0]**o+(0.134598721868)*x[0]+(0.217693108013)*x[1]**o+(0.740476931936)*x[1]
        s[0,2,2,0]=(-0.117189923757)*x[0]**o+(0.433626806763)*x[0]+(-0.746094327712)*x[1]**o+(0.179831846248)*x[1]
        s[0,2,2,1]=(-0.579577944086)*x[0]**o+(0.852457625123)*x[0]+(0.518004334819)*x[1]**o+(0.479873938319)*x[1]
        s[0,3,0,0]=(-0.289116748576)*x[0]**o+(-0.828870395045)*x[0]+(-0.315720408924)*x[1]**o+(0.472146367463)*x[1]
        s[0,3,0,1]=(0.0109829201758)*x[0]**o+(-0.750333402059)*x[0]+(-0.0844671059504)*x[1]**o+(-0.166689650524)*x[1]
        s[0,3,1,0]=(-0.746333456886)*x[0]**o+(0.974025974133)*x[0]+(-0.670772415415)*x[1]**o+(-0.288183829047)*x[1]
        s[0,3,1,1]=(0.999847690339)*x[0]**o+(0.927687112542)*x[0]+(-0.682088385211)*x[1]**o+(-0.185190802895)*x[1]
        s[0,3,2,0]=(-0.122143422759)*x[0]**o+(-0.483572799559)*x[0]+(-0.421650699123)*x[1]**o+(0.751112515854)*x[1]
        s[0,3,2,1]=(0.347046010599)*x[0]**o+(0.86478919879)*x[0]+(-0.376342341601)*x[1]**o+(0.254005796044)*x[1]
        s[0,4,0,0]=(-0.94308296157)*x[0]**o+(-0.278466613273)*x[0]+(0.429411202974)*x[1]**o+(-0.686514518522)*x[1]
        s[0,4,0,1]=(-0.591206969855)*x[0]**o+(0.625779617413)*x[0]+(-0.359728224291)*x[1]**o+(-0.468411179387)*x[1]
        s[0,4,1,0]=(0.871007238866)*x[0]**o+(0.0640301246531)*x[0]+(0.241526675204)*x[1]**o+(-0.198284928754)*x[1]
        s[0,4,1,1]=(0.807190539194)*x[0]**o+(-0.394442569228)*x[0]+(0.137420197381)*x[1]**o+(-0.840168205779)*x[1]
        s[0,4,2,0]=(-0.895780376203)*x[0]**o+(0.489426109167)*x[0]+(0.104706061393)*x[1]**o+(0.580327421236)*x[1]
        s[0,4,2,1]=(0.866870983947)*x[0]**o+(-0.634848129356)*x[0]+(-0.324429185788)*x[1]**o+(-0.902930131617)*x[1]
        s[1,0,0,0]=(-0.435460582678)*x[0]**o+(-0.513879950465)*x[0]+(-0.12284906436)*x[1]**o+(-0.189956950806)*x[1]
        s[1,0,0,1]=(-0.10596403879)*x[0]**o+(0.584299361023)*x[0]+(0.643290036141)*x[1]**o+(0.410255914402)*x[1]
        s[1,0,1,0]=(0.596321097839)*x[0]**o+(0.404038905356)*x[0]+(-0.887663971354)*x[1]**o+(-0.602426219327)*x[1]
        s[1,0,1,1]=(-0.749976485032)*x[0]**o+(0.624337802476)*x[0]+(0.230489334804)*x[1]**o+(-0.169509371147)*x[1]
        s[1,0,2,0]=(-0.338056360533)*x[0]**o+(-0.334004127698)*x[0]+(-0.378429755542)*x[1]**o+(0.656144700111)*x[1]
        s[1,0,2,1]=(0.360646534328)*x[0]**o+(-0.823927735389)*x[0]+(0.758795958018)*x[1]**o+(0.99470110659)*x[1]
        s[1,1,0,0]=(0.831774733955)*x[0]**o+(-0.161890693722)*x[0]+(-0.211705245129)*x[1]**o+(-0.231627795548)*x[1]
        s[1,1,0,1]=(0.994381914656)*x[0]**o+(0.483843191342)*x[0]+(0.703906230692)*x[1]**o+(0.516762394585)*x[1]
        s[1,1,1,0]=(-0.269625525789)*x[0]**o+(0.804559850277)*x[0]+(0.645074410485)*x[1]**o+(0.93472757775)*x[1]
        s[1,1,1,1]=(0.887430533737)*x[0]**o+(0.248852072717)*x[0]+(0.902457877398)*x[1]**o+(-0.749542407535)*x[1]
        s[1,1,2,0]=(0.999384101785)*x[0]**o+(-0.0730419780583)*x[0]+(0.0514349007487)*x[1]**o+(0.151139683749)*x[1]
        s[1,1,2,1]=(0.707403402546)*x[0]**o+(0.421383069726)*x[0]+(-0.713007416618)*x[1]**o+(0.268783526448)*x[1]
        s[1,2,0,0]=(-0.411586906507)*x[0]**o+(0.941961720922)*x[0]+(0.913012821339)*x[1]**o+(-0.296923947382)*x[1]
        s[1,2,0,1]=(-0.520356480672)*x[0]**o+(0.530942059766)*x[0]+(0.0664413255121)*x[1]**o+(0.910541121569)*x[1]
        s[1,2,1,0]=(0.546406246315)*x[0]**o+(0.236311366799)*x[0]+(0.876031584809)*x[1]**o+(0.68767015821)*x[1]
        s[1,2,1,1]=(0.408868402043)*x[0]**o+(0.0910602281951)*x[0]+(0.717385383734)*x[1]**o+(0.799924371281)*x[1]
        s[1,2,2,0]=(-0.00645488169841)*x[0]**o+(-0.193747155577)*x[0]+(0.0532024922569)*x[1]**o+(0.0346603274986)*x[1]
        s[1,2,2,1]=(-0.469056034545)*x[0]**o+(0.221139250978)*x[0]+(0.694972877007)*x[1]**o+(-0.390789722006)*x[1]
        s[1,3,0,0]=(-0.592518388291)*x[0]**o+(0.68133290065)*x[0]+(0.256304394552)*x[1]**o+(0.432327297684)*x[1]
        s[1,3,0,1]=(-0.00024413313477)*x[0]**o+(-0.233897782087)*x[0]+(-0.48099566426)*x[1]**o+(-0.0780681896879)*x[1]
        s[1,3,1,0]=(-0.748402672608)*x[0]**o+(0.663174776498)*x[0]+(0.0516063977957)*x[1]**o+(0.707459316309)*x[1]
        s[1,3,1,1]=(0.856685363283)*x[0]**o+(-0.975581237163)*x[0]+(-0.808077402325)*x[1]**o+(-0.792869908311)*x[1]
        s[1,3,2,0]=(-0.608434761452)*x[0]**o+(-0.538084150506)*x[0]+(0.963534643788)*x[1]**o+(-0.542477024165)*x[1]
        s[1,3,2,1]=(0.0321829705121)*x[0]**o+(-0.646103011762)*x[0]+(-0.0198446078784)*x[1]**o+(0.919550816734)*x[1]
        s[1,4,0,0]=(0.222383667675)*x[0]**o+(-0.687847703153)*x[0]+(-0.59651951219)*x[1]**o+(0.81129360922)*x[1]
        s[1,4,0,1]=(-0.808175325395)*x[0]**o+(0.217063469977)*x[0]+(-0.26488805688)*x[1]**o+(0.0575808940543)*x[1]
        s[1,4,1,0]=(-0.839796412974)*x[0]**o+(-0.42930678659)*x[0]+(0.856238206181)*x[1]**o+(0.0618344272746)*x[1]
        s[1,4,1,1]=(-0.701180815436)*x[0]**o+(-0.87848285733)*x[0]+(0.809946317726)*x[1]**o+(0.733322831203)*x[1]
        s[1,4,2,0]=(-0.00706224969808)*x[0]**o+(0.164085065344)*x[0]+(0.651640705138)*x[1]**o+(-0.947949213151)*x[1]
        s[1,4,2,1]=(-0.55255682366)*x[0]**o+(-3.61985980193e-05)*x[0]+(-0.998780219628)*x[1]**o+(0.68880003343)*x[1]
        s[2,0,0,0]=(0.923223655511)*x[0]**o+(0.320103985146)*x[0]+(0.999706876829)*x[1]**o+(-0.920680964711)*x[1]
        s[2,0,0,1]=(-0.145894916434)*x[0]**o+(0.509339634387)*x[0]+(-0.877941500009)*x[1]**o+(-0.792912551717)*x[1]
        s[2,0,1,0]=(0.736573167219)*x[0]**o+(0.744999602293)*x[0]+(-0.446569219347)*x[1]**o+(0.700985615215)*x[1]
        s[2,0,1,1]=(-0.390389830532)*x[0]**o+(-0.6656221359)*x[0]+(-0.27554460662)*x[1]**o+(-0.458879579705)*x[1]
        s[2,0,2,0]=(-0.673016930466)*x[0]**o+(0.610425320745)*x[0]+(-0.948708604918)*x[1]**o+(0.44912430221)*x[1]
        s[2,0,2,1]=(0.596963521799)*x[0]**o+(-0.990485203537)*x[0]+(0.164522035463)*x[1]**o+(0.539120075165)*x[1]
        s[2,1,0,0]=(-0.932518355596)*x[0]**o+(-0.590131952474)*x[0]+(-0.533511934929)*x[1]**o+(0.0907025905348)*x[1]
        s[2,1,0,1]=(0.865266539976)*x[0]**o+(0.876881424586)*x[0]+(-0.0288484839401)*x[1]**o+(-0.807305428358)*x[1]
        s[2,1,1,0]=(-0.763528720307)*x[0]**o+(-0.0232723097938)*x[0]+(-0.525533773303)*x[1]**o+(0.806025194949)*x[1]
        s[2,1,1,1]=(-0.696217606839)*x[0]**o+(-0.14983123685)*x[0]+(-0.783578700293)*x[1]**o+(0.447836291404)*x[1]
        s[2,1,2,0]=(0.0948683273761)*x[0]**o+(-0.721171113714)*x[0]+(0.701343762426)*x[1]**o+(0.0101281640664)*x[1]
        s[2,1,2,1]=(0.562629645787)*x[0]**o+(-0.763724767765)*x[0]+(-0.449544469039)*x[1]**o+(0.532094652638)*x[1]
        s[2,2,0,0]=(0.841490428822)*x[0]**o+(0.681330810792)*x[0]+(0.668346714262)*x[1]**o+(-0.0336199322961)*x[1]
        s[2,2,0,1]=(0.771997063856)*x[0]**o+(-0.914459974236)*x[0]+(0.129697241086)*x[1]**o+(0.183872706861)*x[1]
        s[2,2,1,0]=(0.447126530052)*x[0]**o+(0.195523018871)*x[0]+(-0.607523037429)*x[1]**o+(0.0880216010007)*x[1]
        s[2,2,1,1]=(0.76876849616)*x[0]**o+(-0.918037921936)*x[0]+(0.65247942976)*x[1]**o+(-0.462180182464)*x[1]
        s[2,2,2,0]=(0.728080043749)*x[0]**o+(-0.716254868732)*x[0]+(0.374917205524)*x[1]**o+(0.311620466651)*x[1]
        s[2,2,2,1]=(0.574841312542)*x[0]**o+(-0.734971890822)*x[0]+(-0.761677234234)*x[1]**o+(0.723259862313)*x[1]
        s[2,3,0,0]=(0.270456229657)*x[0]**o+(-0.58716739336)*x[0]+(0.232577840441)*x[1]**o+(0.747330716699)*x[1]
        s[2,3,0,1]=(-0.820314477652)*x[0]**o+(-0.719999661485)*x[0]+(0.79996006468)*x[1]**o+(-0.831712763153)*x[1]
        s[2,3,1,0]=(0.847137634307)*x[0]**o+(-0.357186776539)*x[0]+(0.09775554965)*x[1]**o+(-0.327915828517)*x[1]
        s[2,3,1,1]=(0.67586077388)*x[0]**o+(0.0335299900089)*x[0]+(-0.179177480197)*x[1]**o+(0.378749816102)*x[1]
        s[2,3,2,0]=(-0.223679858716)*x[0]**o+(-0.844559556039)*x[0]+(0.255235296117)*x[1]**o+(-0.80166933466)*x[1]
        s[2,3,2,1]=(0.668162512403)*x[0]**o+(0.496871152853)*x[0]+(0.745978534888)*x[1]**o+(0.799256534355)*x[1]
        s[2,4,0,0]=(0.363336239392)*x[0]**o+(0.22196304418)*x[0]+(0.875013792472)*x[1]**o+(0.299160735097)*x[1]
        s[2,4,0,1]=(-0.77295625153)*x[0]**o+(-0.109539742009)*x[0]+(0.42425251466)*x[1]**o+(0.857114441732)*x[1]
        s[2,4,1,0]=(-0.453314408352)*x[0]**o+(0.263814392877)*x[0]+(-0.725268476809)*x[1]**o+(-0.347648385435)*x[1]
        s[2,4,1,1]=(-0.35138702248)*x[0]**o+(0.0172444410003)*x[0]+(0.705499375267)*x[1]**o+(0.989996399519)*x[1]
        s[2,4,2,0]=(-0.151100098883)*x[0]**o+(0.0561722886965)*x[0]+(0.701976335813)*x[1]**o+(-0.318092303985)*x[1]
        s[2,4,2,1]=(0.936223946358)*x[0]**o+(0.525004572217)*x[0]+(-0.112487209398)*x[1]**o+(0.196069142522)*x[1]
        s[3,0,0,0]=(-0.296095919553)*x[0]**o+(-0.607841982284)*x[0]+(0.654302153755)*x[1]**o+(0.0410915588113)*x[1]
        s[3,0,0,1]=(0.778444828767)*x[0]**o+(-0.148613776811)*x[0]+(0.911906751967)*x[1]**o+(-0.72527304075)*x[1]
        s[3,0,1,0]=(-0.964041638986)*x[0]**o+(-0.444730606402)*x[0]+(0.856465666005)*x[1]**o+(0.497912010281)*x[1]
        s[3,0,1,1]=(0.0278725007089)*x[0]**o+(-0.102825566476)*x[0]+(-0.925031761314)*x[1]**o+(-0.99399390758)*x[1]
        s[3,0,2,0]=(-0.750580200302)*x[0]**o+(0.792361919091)*x[0]+(0.283376094631)*x[1]**o+(0.210963254645)*x[1]
        s[3,0,2,1]=(-0.202368637069)*x[0]**o+(0.405267270651)*x[0]+(-0.199978746176)*x[1]**o+(0.57738788439)*x[1]
        s[3,1,0,0]=(0.82614638217)*x[0]**o+(-0.342667586691)*x[0]+(0.608532783812)*x[1]**o+(0.768159873725)*x[1]
        s[3,1,0,1]=(0.753566806412)*x[0]**o+(-0.251587300566)*x[0]+(-0.340721126292)*x[1]**o+(-0.757484954136)*x[1]
        s[3,1,1,0]=(0.218156803727)*x[0]**o+(0.527606199926)*x[0]+(-0.212766545043)*x[1]**o+(0.363494347038)*x[1]
        s[3,1,1,1]=(-0.419579821731)*x[0]**o+(-0.848758805521)*x[0]+(0.127749661772)*x[1]**o+(-0.674368477333)*x[1]
        s[3,1,2,0]=(-0.0378627520537)*x[0]**o+(-0.373536670854)*x[0]+(-0.473995670566)*x[1]**o+(0.817069259381)*x[1]
        s[3,1,2,1]=(0.0637470079874)*x[0]**o+(0.902092510251)*x[0]+(0.613777956237)*x[1]**o+(0.984903229699)*x[1]
        s[3,2,0,0]=(-0.550317501871)*x[0]**o+(0.387530710046)*x[0]+(0.51892851937)*x[1]**o+(0.830001465659)*x[1]
        s[3,2,0,1]=(0.159293029872)*x[0]**o+(0.0752213137273)*x[0]+(0.828173132448)*x[1]**o+(0.526955740589)*x[1]
        s[3,2,1,0]=(-0.439770798894)*x[0]**o+(-0.0441574714373)*x[0]+(0.674623413183)*x[1]**o+(-0.958611493787)*x[1]
        s[3,2,1,1]=(0.739682304539)*x[0]**o+(0.734753273429)*x[0]+(-0.20502427226)*x[1]**o+(-0.158955519702)*x[1]
        s[3,2,2,0]=(-0.940998669744)*x[0]**o+(-0.41965535276)*x[0]+(0.893147081307)*x[1]**o+(-0.225209541903)*x[1]
        s[3,2,2,1]=(0.389223021003)*x[0]**o+(-0.483125135422)*x[0]+(0.777672638794)*x[1]**o+(0.839086315551)*x[1]
        s[3,3,0,0]=(-0.320672758354)*x[0]**o+(0.0620118134094)*x[0]+(-0.680049325665)*x[1]**o+(0.411421232123)*x[1]
        s[3,3,0,1]=(-0.0893153008416)*x[0]**o+(0.213720188183)*x[0]+(-0.34570418095)*x[1]**o+(0.625048919615)*x[1]
        s[3,3,1,0]=(0.0158017598762)*x[0]**o+(0.605723583993)*x[0]+(0.739309396699)*x[1]**o+(-0.636482587738)*x[1]
        s[3,3,1,1]=(0.910657933249)*x[0]**o+(-0.441319152646)*x[0]+(0.23268716261)*x[1]**o+(0.24329792568)*x[1]
        s[3,3,2,0]=(0.902717620152)*x[0]**o+(-0.549313425275)*x[0]+(0.464933158282)*x[1]**o+(-0.972615897723)*x[1]
        s[3,3,2,1]=(0.0282461432227)*x[0]**o+(0.761575417375)*x[0]+(0.569262249337)*x[1]**o+(0.828621514356)*x[1]
        s[3,4,0,0]=(0.399367935866)*x[0]**o+(-0.82801744208)*x[0]+(-0.0726407406113)*x[1]**o+(0.530343378961)*x[1]
        s[3,4,0,1]=(0.281064774416)*x[0]**o+(0.338573820292)*x[0]+(0.138863275341)*x[1]**o+(0.148411768993)*x[1]
        s[3,4,1,0]=(-0.686579931373)*x[0]**o+(0.639562092961)*x[0]+(-0.0380362333106)*x[1]**o+(-0.467126934241)*x[1]
        s[3,4,1,1]=(0.611069057705)*x[0]**o+(0.448509606798)*x[0]+(0.663215146298)*x[1]**o+(-0.571323998876)*x[1]
        s[3,4,2,0]=(0.21998109831)*x[0]**o+(-0.737645914394)*x[0]+(0.669008118713)*x[1]**o+(-0.291237893231)*x[1]
        s[3,4,2,1]=(0.595080778165)*x[0]**o+(0.988680777579)*x[0]+(-0.361694479664)*x[1]**o+(0.210229549739)*x[1]
        ref[0,0,0,0]=(0.771805529053)*x_ref[0]**o+(-0.00794071414241)*x_ref[0]+(-0.546566317942)*x_ref[1]**o+(0.209469633398)*x_ref[1]
        ref[0,0,0,1]=(-0.0138300734518)*x_ref[0]**o+(0.783507392816)*x_ref[0]+(0.637453082951)*x_ref[1]**o+(-0.0262186183303)*x_ref[1]
        ref[0,0,1,0]=(-0.622230122991)*x_ref[0]**o+(0.760867623401)*x_ref[0]+(0.340084769283)*x_ref[1]**o+(0.493207077521)*x_ref[1]
        ref[0,0,1,1]=(0.772195055793)*x_ref[0]**o+(-0.481066683803)*x_ref[0]+(0.198371969404)*x_ref[1]**o+(0.841664205482)*x_ref[1]
        ref[0,0,2,0]=(0.858213005013)*x_ref[0]**o+(0.819391260667)*x_ref[0]+(0.100404710733)*x_ref[1]**o+(0.0757471444158)*x_ref[1]
        ref[0,0,2,1]=(0.865434306062)*x_ref[0]**o+(0.411158028281)*x_ref[0]+(0.717843255665)*x_ref[1]**o+(-0.742991318999)*x_ref[1]
        ref[0,1,0,0]=(-0.851618984895)*x_ref[0]**o+(0.251966806193)*x_ref[0]+(-0.803642825762)*x_ref[1]**o+(-0.0631169142808)*x_ref[1]
        ref[0,1,0,1]=(-0.758981775218)*x_ref[0]**o+(-0.561508941004)*x_ref[0]+(0.577901652618)*x_ref[1]**o+(-0.725243471023)*x_ref[1]
        ref[0,1,1,0]=(0.742032672941)*x_ref[0]**o+(-0.0526002854749)*x_ref[0]+(0.252153928183)*x_ref[1]**o+(-0.388978984633)*x_ref[1]
        ref[0,1,1,1]=(-0.980858344737)*x_ref[0]**o+(-0.871658904716)*x_ref[0]+(0.164033322064)*x_ref[1]**o+(0.678658472838)*x_ref[1]
        ref[0,1,2,0]=(0.823816388003)*x_ref[0]**o+(0.274859643706)*x_ref[0]+(0.322475967163)*x_ref[1]**o+(-0.598198426633)*x_ref[1]
        ref[0,1,2,1]=(0.481980285537)*x_ref[0]**o+(-0.806234350243)*x_ref[0]+(0.564463264623)*x_ref[1]**o+(-0.548427948656)*x_ref[1]
        ref[0,2,0,0]=(0.582989836924)*x_ref[0]**o+(0.293733803309)*x_ref[0]+(0.0589622495939)*x_ref[1]**o+(-0.661119752655)*x_ref[1]
        ref[0,2,0,1]=(0.675133991144)*x_ref[0]**o+(-0.964990891523)*x_ref[0]+(-0.488115773474)*x_ref[1]**o+(-0.601204526426)*x_ref[1]
        ref[0,2,1,0]=(-0.504646764871)*x_ref[0]**o+(-0.496630864273)*x_ref[0]+(-0.533328001228)*x_ref[1]**o+(0.666398746532)*x_ref[1]
        ref[0,2,1,1]=(-0.947971259712)*x_ref[0]**o+(0.134598721868)*x_ref[0]+(0.217693108013)*x_ref[1]**o+(0.740476931936)*x_ref[1]
        ref[0,2,2,0]=(-0.117189923757)*x_ref[0]**o+(0.433626806763)*x_ref[0]+(-0.746094327712)*x_ref[1]**o+(0.179831846248)*x_ref[1]
        ref[0,2,2,1]=(-0.579577944086)*x_ref[0]**o+(0.852457625123)*x_ref[0]+(0.518004334819)*x_ref[1]**o+(0.479873938319)*x_ref[1]
        ref[0,3,0,0]=(-0.289116748576)*x_ref[0]**o+(-0.828870395045)*x_ref[0]+(-0.315720408924)*x_ref[1]**o+(0.472146367463)*x_ref[1]
        ref[0,3,0,1]=(0.0109829201758)*x_ref[0]**o+(-0.750333402059)*x_ref[0]+(-0.0844671059504)*x_ref[1]**o+(-0.166689650524)*x_ref[1]
        ref[0,3,1,0]=(-0.746333456886)*x_ref[0]**o+(0.974025974133)*x_ref[0]+(-0.670772415415)*x_ref[1]**o+(-0.288183829047)*x_ref[1]
        ref[0,3,1,1]=(0.999847690339)*x_ref[0]**o+(0.927687112542)*x_ref[0]+(-0.682088385211)*x_ref[1]**o+(-0.185190802895)*x_ref[1]
        ref[0,3,2,0]=(-0.122143422759)*x_ref[0]**o+(-0.483572799559)*x_ref[0]+(-0.421650699123)*x_ref[1]**o+(0.751112515854)*x_ref[1]
        ref[0,3,2,1]=(0.347046010599)*x_ref[0]**o+(0.86478919879)*x_ref[0]+(-0.376342341601)*x_ref[1]**o+(0.254005796044)*x_ref[1]
        ref[0,4,0,0]=(-0.94308296157)*x_ref[0]**o+(-0.278466613273)*x_ref[0]+(0.429411202974)*x_ref[1]**o+(-0.686514518522)*x_ref[1]
        ref[0,4,0,1]=(-0.591206969855)*x_ref[0]**o+(0.625779617413)*x_ref[0]+(-0.359728224291)*x_ref[1]**o+(-0.468411179387)*x_ref[1]
        ref[0,4,1,0]=(0.871007238866)*x_ref[0]**o+(0.0640301246531)*x_ref[0]+(0.241526675204)*x_ref[1]**o+(-0.198284928754)*x_ref[1]
        ref[0,4,1,1]=(0.807190539194)*x_ref[0]**o+(-0.394442569228)*x_ref[0]+(0.137420197381)*x_ref[1]**o+(-0.840168205779)*x_ref[1]
        ref[0,4,2,0]=(-0.895780376203)*x_ref[0]**o+(0.489426109167)*x_ref[0]+(0.104706061393)*x_ref[1]**o+(0.580327421236)*x_ref[1]
        ref[0,4,2,1]=(0.866870983947)*x_ref[0]**o+(-0.634848129356)*x_ref[0]+(-0.324429185788)*x_ref[1]**o+(-0.902930131617)*x_ref[1]
        ref[1,0,0,0]=(-0.435460582678)*x_ref[0]**o+(-0.513879950465)*x_ref[0]+(-0.12284906436)*x_ref[1]**o+(-0.189956950806)*x_ref[1]
        ref[1,0,0,1]=(-0.10596403879)*x_ref[0]**o+(0.584299361023)*x_ref[0]+(0.643290036141)*x_ref[1]**o+(0.410255914402)*x_ref[1]
        ref[1,0,1,0]=(0.596321097839)*x_ref[0]**o+(0.404038905356)*x_ref[0]+(-0.887663971354)*x_ref[1]**o+(-0.602426219327)*x_ref[1]
        ref[1,0,1,1]=(-0.749976485032)*x_ref[0]**o+(0.624337802476)*x_ref[0]+(0.230489334804)*x_ref[1]**o+(-0.169509371147)*x_ref[1]
        ref[1,0,2,0]=(-0.338056360533)*x_ref[0]**o+(-0.334004127698)*x_ref[0]+(-0.378429755542)*x_ref[1]**o+(0.656144700111)*x_ref[1]
        ref[1,0,2,1]=(0.360646534328)*x_ref[0]**o+(-0.823927735389)*x_ref[0]+(0.758795958018)*x_ref[1]**o+(0.99470110659)*x_ref[1]
        ref[1,1,0,0]=(0.831774733955)*x_ref[0]**o+(-0.161890693722)*x_ref[0]+(-0.211705245129)*x_ref[1]**o+(-0.231627795548)*x_ref[1]
        ref[1,1,0,1]=(0.994381914656)*x_ref[0]**o+(0.483843191342)*x_ref[0]+(0.703906230692)*x_ref[1]**o+(0.516762394585)*x_ref[1]
        ref[1,1,1,0]=(-0.269625525789)*x_ref[0]**o+(0.804559850277)*x_ref[0]+(0.645074410485)*x_ref[1]**o+(0.93472757775)*x_ref[1]
        ref[1,1,1,1]=(0.887430533737)*x_ref[0]**o+(0.248852072717)*x_ref[0]+(0.902457877398)*x_ref[1]**o+(-0.749542407535)*x_ref[1]
        ref[1,1,2,0]=(0.999384101785)*x_ref[0]**o+(-0.0730419780583)*x_ref[0]+(0.0514349007487)*x_ref[1]**o+(0.151139683749)*x_ref[1]
        ref[1,1,2,1]=(0.707403402546)*x_ref[0]**o+(0.421383069726)*x_ref[0]+(-0.713007416618)*x_ref[1]**o+(0.268783526448)*x_ref[1]
        ref[1,2,0,0]=(-0.411586906507)*x_ref[0]**o+(0.941961720922)*x_ref[0]+(0.913012821339)*x_ref[1]**o+(-0.296923947382)*x_ref[1]
        ref[1,2,0,1]=(-0.520356480672)*x_ref[0]**o+(0.530942059766)*x_ref[0]+(0.0664413255121)*x_ref[1]**o+(0.910541121569)*x_ref[1]
        ref[1,2,1,0]=(0.546406246315)*x_ref[0]**o+(0.236311366799)*x_ref[0]+(0.876031584809)*x_ref[1]**o+(0.68767015821)*x_ref[1]
        ref[1,2,1,1]=(0.408868402043)*x_ref[0]**o+(0.0910602281951)*x_ref[0]+(0.717385383734)*x_ref[1]**o+(0.799924371281)*x_ref[1]
        ref[1,2,2,0]=(-0.00645488169841)*x_ref[0]**o+(-0.193747155577)*x_ref[0]+(0.0532024922569)*x_ref[1]**o+(0.0346603274986)*x_ref[1]
        ref[1,2,2,1]=(-0.469056034545)*x_ref[0]**o+(0.221139250978)*x_ref[0]+(0.694972877007)*x_ref[1]**o+(-0.390789722006)*x_ref[1]
        ref[1,3,0,0]=(-0.592518388291)*x_ref[0]**o+(0.68133290065)*x_ref[0]+(0.256304394552)*x_ref[1]**o+(0.432327297684)*x_ref[1]
        ref[1,3,0,1]=(-0.00024413313477)*x_ref[0]**o+(-0.233897782087)*x_ref[0]+(-0.48099566426)*x_ref[1]**o+(-0.0780681896879)*x_ref[1]
        ref[1,3,1,0]=(-0.748402672608)*x_ref[0]**o+(0.663174776498)*x_ref[0]+(0.0516063977957)*x_ref[1]**o+(0.707459316309)*x_ref[1]
        ref[1,3,1,1]=(0.856685363283)*x_ref[0]**o+(-0.975581237163)*x_ref[0]+(-0.808077402325)*x_ref[1]**o+(-0.792869908311)*x_ref[1]
        ref[1,3,2,0]=(-0.608434761452)*x_ref[0]**o+(-0.538084150506)*x_ref[0]+(0.963534643788)*x_ref[1]**o+(-0.542477024165)*x_ref[1]
        ref[1,3,2,1]=(0.0321829705121)*x_ref[0]**o+(-0.646103011762)*x_ref[0]+(-0.0198446078784)*x_ref[1]**o+(0.919550816734)*x_ref[1]
        ref[1,4,0,0]=(0.222383667675)*x_ref[0]**o+(-0.687847703153)*x_ref[0]+(-0.59651951219)*x_ref[1]**o+(0.81129360922)*x_ref[1]
        ref[1,4,0,1]=(-0.808175325395)*x_ref[0]**o+(0.217063469977)*x_ref[0]+(-0.26488805688)*x_ref[1]**o+(0.0575808940543)*x_ref[1]
        ref[1,4,1,0]=(-0.839796412974)*x_ref[0]**o+(-0.42930678659)*x_ref[0]+(0.856238206181)*x_ref[1]**o+(0.0618344272746)*x_ref[1]
        ref[1,4,1,1]=(-0.701180815436)*x_ref[0]**o+(-0.87848285733)*x_ref[0]+(0.809946317726)*x_ref[1]**o+(0.733322831203)*x_ref[1]
        ref[1,4,2,0]=(-0.00706224969808)*x_ref[0]**o+(0.164085065344)*x_ref[0]+(0.651640705138)*x_ref[1]**o+(-0.947949213151)*x_ref[1]
        ref[1,4,2,1]=(-0.55255682366)*x_ref[0]**o+(-3.61985980193e-05)*x_ref[0]+(-0.998780219628)*x_ref[1]**o+(0.68880003343)*x_ref[1]
        ref[2,0,0,0]=(0.923223655511)*x_ref[0]**o+(0.320103985146)*x_ref[0]+(0.999706876829)*x_ref[1]**o+(-0.920680964711)*x_ref[1]
        ref[2,0,0,1]=(-0.145894916434)*x_ref[0]**o+(0.509339634387)*x_ref[0]+(-0.877941500009)*x_ref[1]**o+(-0.792912551717)*x_ref[1]
        ref[2,0,1,0]=(0.736573167219)*x_ref[0]**o+(0.744999602293)*x_ref[0]+(-0.446569219347)*x_ref[1]**o+(0.700985615215)*x_ref[1]
        ref[2,0,1,1]=(-0.390389830532)*x_ref[0]**o+(-0.6656221359)*x_ref[0]+(-0.27554460662)*x_ref[1]**o+(-0.458879579705)*x_ref[1]
        ref[2,0,2,0]=(-0.673016930466)*x_ref[0]**o+(0.610425320745)*x_ref[0]+(-0.948708604918)*x_ref[1]**o+(0.44912430221)*x_ref[1]
        ref[2,0,2,1]=(0.596963521799)*x_ref[0]**o+(-0.990485203537)*x_ref[0]+(0.164522035463)*x_ref[1]**o+(0.539120075165)*x_ref[1]
        ref[2,1,0,0]=(-0.932518355596)*x_ref[0]**o+(-0.590131952474)*x_ref[0]+(-0.533511934929)*x_ref[1]**o+(0.0907025905348)*x_ref[1]
        ref[2,1,0,1]=(0.865266539976)*x_ref[0]**o+(0.876881424586)*x_ref[0]+(-0.0288484839401)*x_ref[1]**o+(-0.807305428358)*x_ref[1]
        ref[2,1,1,0]=(-0.763528720307)*x_ref[0]**o+(-0.0232723097938)*x_ref[0]+(-0.525533773303)*x_ref[1]**o+(0.806025194949)*x_ref[1]
        ref[2,1,1,1]=(-0.696217606839)*x_ref[0]**o+(-0.14983123685)*x_ref[0]+(-0.783578700293)*x_ref[1]**o+(0.447836291404)*x_ref[1]
        ref[2,1,2,0]=(0.0948683273761)*x_ref[0]**o+(-0.721171113714)*x_ref[0]+(0.701343762426)*x_ref[1]**o+(0.0101281640664)*x_ref[1]
        ref[2,1,2,1]=(0.562629645787)*x_ref[0]**o+(-0.763724767765)*x_ref[0]+(-0.449544469039)*x_ref[1]**o+(0.532094652638)*x_ref[1]
        ref[2,2,0,0]=(0.841490428822)*x_ref[0]**o+(0.681330810792)*x_ref[0]+(0.668346714262)*x_ref[1]**o+(-0.0336199322961)*x_ref[1]
        ref[2,2,0,1]=(0.771997063856)*x_ref[0]**o+(-0.914459974236)*x_ref[0]+(0.129697241086)*x_ref[1]**o+(0.183872706861)*x_ref[1]
        ref[2,2,1,0]=(0.447126530052)*x_ref[0]**o+(0.195523018871)*x_ref[0]+(-0.607523037429)*x_ref[1]**o+(0.0880216010007)*x_ref[1]
        ref[2,2,1,1]=(0.76876849616)*x_ref[0]**o+(-0.918037921936)*x_ref[0]+(0.65247942976)*x_ref[1]**o+(-0.462180182464)*x_ref[1]
        ref[2,2,2,0]=(0.728080043749)*x_ref[0]**o+(-0.716254868732)*x_ref[0]+(0.374917205524)*x_ref[1]**o+(0.311620466651)*x_ref[1]
        ref[2,2,2,1]=(0.574841312542)*x_ref[0]**o+(-0.734971890822)*x_ref[0]+(-0.761677234234)*x_ref[1]**o+(0.723259862313)*x_ref[1]
        ref[2,3,0,0]=(0.270456229657)*x_ref[0]**o+(-0.58716739336)*x_ref[0]+(0.232577840441)*x_ref[1]**o+(0.747330716699)*x_ref[1]
        ref[2,3,0,1]=(-0.820314477652)*x_ref[0]**o+(-0.719999661485)*x_ref[0]+(0.79996006468)*x_ref[1]**o+(-0.831712763153)*x_ref[1]
        ref[2,3,1,0]=(0.847137634307)*x_ref[0]**o+(-0.357186776539)*x_ref[0]+(0.09775554965)*x_ref[1]**o+(-0.327915828517)*x_ref[1]
        ref[2,3,1,1]=(0.67586077388)*x_ref[0]**o+(0.0335299900089)*x_ref[0]+(-0.179177480197)*x_ref[1]**o+(0.378749816102)*x_ref[1]
        ref[2,3,2,0]=(-0.223679858716)*x_ref[0]**o+(-0.844559556039)*x_ref[0]+(0.255235296117)*x_ref[1]**o+(-0.80166933466)*x_ref[1]
        ref[2,3,2,1]=(0.668162512403)*x_ref[0]**o+(0.496871152853)*x_ref[0]+(0.745978534888)*x_ref[1]**o+(0.799256534355)*x_ref[1]
        ref[2,4,0,0]=(0.363336239392)*x_ref[0]**o+(0.22196304418)*x_ref[0]+(0.875013792472)*x_ref[1]**o+(0.299160735097)*x_ref[1]
        ref[2,4,0,1]=(-0.77295625153)*x_ref[0]**o+(-0.109539742009)*x_ref[0]+(0.42425251466)*x_ref[1]**o+(0.857114441732)*x_ref[1]
        ref[2,4,1,0]=(-0.453314408352)*x_ref[0]**o+(0.263814392877)*x_ref[0]+(-0.725268476809)*x_ref[1]**o+(-0.347648385435)*x_ref[1]
        ref[2,4,1,1]=(-0.35138702248)*x_ref[0]**o+(0.0172444410003)*x_ref[0]+(0.705499375267)*x_ref[1]**o+(0.989996399519)*x_ref[1]
        ref[2,4,2,0]=(-0.151100098883)*x_ref[0]**o+(0.0561722886965)*x_ref[0]+(0.701976335813)*x_ref[1]**o+(-0.318092303985)*x_ref[1]
        ref[2,4,2,1]=(0.936223946358)*x_ref[0]**o+(0.525004572217)*x_ref[0]+(-0.112487209398)*x_ref[1]**o+(0.196069142522)*x_ref[1]
        ref[3,0,0,0]=(-0.296095919553)*x_ref[0]**o+(-0.607841982284)*x_ref[0]+(0.654302153755)*x_ref[1]**o+(0.0410915588113)*x_ref[1]
        ref[3,0,0,1]=(0.778444828767)*x_ref[0]**o+(-0.148613776811)*x_ref[0]+(0.911906751967)*x_ref[1]**o+(-0.72527304075)*x_ref[1]
        ref[3,0,1,0]=(-0.964041638986)*x_ref[0]**o+(-0.444730606402)*x_ref[0]+(0.856465666005)*x_ref[1]**o+(0.497912010281)*x_ref[1]
        ref[3,0,1,1]=(0.0278725007089)*x_ref[0]**o+(-0.102825566476)*x_ref[0]+(-0.925031761314)*x_ref[1]**o+(-0.99399390758)*x_ref[1]
        ref[3,0,2,0]=(-0.750580200302)*x_ref[0]**o+(0.792361919091)*x_ref[0]+(0.283376094631)*x_ref[1]**o+(0.210963254645)*x_ref[1]
        ref[3,0,2,1]=(-0.202368637069)*x_ref[0]**o+(0.405267270651)*x_ref[0]+(-0.199978746176)*x_ref[1]**o+(0.57738788439)*x_ref[1]
        ref[3,1,0,0]=(0.82614638217)*x_ref[0]**o+(-0.342667586691)*x_ref[0]+(0.608532783812)*x_ref[1]**o+(0.768159873725)*x_ref[1]
        ref[3,1,0,1]=(0.753566806412)*x_ref[0]**o+(-0.251587300566)*x_ref[0]+(-0.340721126292)*x_ref[1]**o+(-0.757484954136)*x_ref[1]
        ref[3,1,1,0]=(0.218156803727)*x_ref[0]**o+(0.527606199926)*x_ref[0]+(-0.212766545043)*x_ref[1]**o+(0.363494347038)*x_ref[1]
        ref[3,1,1,1]=(-0.419579821731)*x_ref[0]**o+(-0.848758805521)*x_ref[0]+(0.127749661772)*x_ref[1]**o+(-0.674368477333)*x_ref[1]
        ref[3,1,2,0]=(-0.0378627520537)*x_ref[0]**o+(-0.373536670854)*x_ref[0]+(-0.473995670566)*x_ref[1]**o+(0.817069259381)*x_ref[1]
        ref[3,1,2,1]=(0.0637470079874)*x_ref[0]**o+(0.902092510251)*x_ref[0]+(0.613777956237)*x_ref[1]**o+(0.984903229699)*x_ref[1]
        ref[3,2,0,0]=(-0.550317501871)*x_ref[0]**o+(0.387530710046)*x_ref[0]+(0.51892851937)*x_ref[1]**o+(0.830001465659)*x_ref[1]
        ref[3,2,0,1]=(0.159293029872)*x_ref[0]**o+(0.0752213137273)*x_ref[0]+(0.828173132448)*x_ref[1]**o+(0.526955740589)*x_ref[1]
        ref[3,2,1,0]=(-0.439770798894)*x_ref[0]**o+(-0.0441574714373)*x_ref[0]+(0.674623413183)*x_ref[1]**o+(-0.958611493787)*x_ref[1]
        ref[3,2,1,1]=(0.739682304539)*x_ref[0]**o+(0.734753273429)*x_ref[0]+(-0.20502427226)*x_ref[1]**o+(-0.158955519702)*x_ref[1]
        ref[3,2,2,0]=(-0.940998669744)*x_ref[0]**o+(-0.41965535276)*x_ref[0]+(0.893147081307)*x_ref[1]**o+(-0.225209541903)*x_ref[1]
        ref[3,2,2,1]=(0.389223021003)*x_ref[0]**o+(-0.483125135422)*x_ref[0]+(0.777672638794)*x_ref[1]**o+(0.839086315551)*x_ref[1]
        ref[3,3,0,0]=(-0.320672758354)*x_ref[0]**o+(0.0620118134094)*x_ref[0]+(-0.680049325665)*x_ref[1]**o+(0.411421232123)*x_ref[1]
        ref[3,3,0,1]=(-0.0893153008416)*x_ref[0]**o+(0.213720188183)*x_ref[0]+(-0.34570418095)*x_ref[1]**o+(0.625048919615)*x_ref[1]
        ref[3,3,1,0]=(0.0158017598762)*x_ref[0]**o+(0.605723583993)*x_ref[0]+(0.739309396699)*x_ref[1]**o+(-0.636482587738)*x_ref[1]
        ref[3,3,1,1]=(0.910657933249)*x_ref[0]**o+(-0.441319152646)*x_ref[0]+(0.23268716261)*x_ref[1]**o+(0.24329792568)*x_ref[1]
        ref[3,3,2,0]=(0.902717620152)*x_ref[0]**o+(-0.549313425275)*x_ref[0]+(0.464933158282)*x_ref[1]**o+(-0.972615897723)*x_ref[1]
        ref[3,3,2,1]=(0.0282461432227)*x_ref[0]**o+(0.761575417375)*x_ref[0]+(0.569262249337)*x_ref[1]**o+(0.828621514356)*x_ref[1]
        ref[3,4,0,0]=(0.399367935866)*x_ref[0]**o+(-0.82801744208)*x_ref[0]+(-0.0726407406113)*x_ref[1]**o+(0.530343378961)*x_ref[1]
        ref[3,4,0,1]=(0.281064774416)*x_ref[0]**o+(0.338573820292)*x_ref[0]+(0.138863275341)*x_ref[1]**o+(0.148411768993)*x_ref[1]
        ref[3,4,1,0]=(-0.686579931373)*x_ref[0]**o+(0.639562092961)*x_ref[0]+(-0.0380362333106)*x_ref[1]**o+(-0.467126934241)*x_ref[1]
        ref[3,4,1,1]=(0.611069057705)*x_ref[0]**o+(0.448509606798)*x_ref[0]+(0.663215146298)*x_ref[1]**o+(-0.571323998876)*x_ref[1]
        ref[3,4,2,0]=(0.21998109831)*x_ref[0]**o+(-0.737645914394)*x_ref[0]+(0.669008118713)*x_ref[1]**o+(-0.291237893231)*x_ref[1]
        ref[3,4,2,1]=(0.595080778165)*x_ref[0]**o+(0.988680777579)*x_ref[0]+(-0.361694479664)*x_ref[1]**o+(0.210229549739)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.209698190814)*x[0]**o+(0.0202198726379)*x[0]+(-0.776762051242)*x[1]**o+(0.0103026495025)*x[1]+(0.0277886960026)*x[2]**o+(-0.255065901162)*x[2]
        s[0,0,0,1]=(-0.663085958734)*x[0]**o+(0.698525770087)*x[0]+(0.650288968593)*x[1]**o+(0.6069536467)*x[1]+(0.789898121185)*x[2]**o+(-0.13315288412)*x[2]
        s[0,0,1,0]=(-0.0998227591334)*x[0]**o+(0.75314681829)*x[0]+(-0.43575224825)*x[1]**o+(0.649609416117)*x[1]+(0.271394201051)*x[2]**o+(-0.353712050582)*x[2]
        s[0,0,1,1]=(0.205441533003)*x[0]**o+(-0.61122881962)*x[0]+(-0.431370221976)*x[1]**o+(-0.183655447483)*x[1]+(0.116437364187)*x[2]**o+(0.741511658665)*x[2]
        s[0,0,2,0]=(0.516133026558)*x[0]**o+(-0.904105438162)*x[0]+(-0.579514088677)*x[1]**o+(-0.0581332954324)*x[1]+(0.557937742112)*x[2]**o+(0.138639028149)*x[2]
        s[0,0,2,1]=(0.8991763477)*x[0]**o+(-0.792213819942)*x[0]+(0.674755445696)*x[1]**o+(0.483315869458)*x[1]+(-0.0722431463476)*x[2]**o+(0.0270814412191)*x[2]
        s[0,1,0,0]=(0.366564587764)*x[0]**o+(0.928578857146)*x[0]+(0.785092228323)*x[1]**o+(0.860812095832)*x[1]+(-0.603464315283)*x[2]**o+(-0.556636685093)*x[2]
        s[0,1,0,1]=(0.980630839809)*x[0]**o+(-0.255324039441)*x[0]+(0.173630938787)*x[1]**o+(0.110433932158)*x[1]+(-0.392851296251)*x[2]**o+(-0.272438563385)*x[2]
        s[0,1,1,0]=(-0.63489459343)*x[0]**o+(0.0328950965715)*x[0]+(-0.199264976942)*x[1]**o+(0.704589249988)*x[1]+(0.442306934744)*x[2]**o+(-0.531996394873)*x[2]
        s[0,1,1,1]=(-0.110263035707)*x[0]**o+(0.904756528197)*x[0]+(0.876124769497)*x[1]**o+(-0.55469630081)*x[1]+(-0.857675499295)*x[2]**o+(0.662352375455)*x[2]
        s[0,1,2,0]=(-0.383336639984)*x[0]**o+(0.863605840878)*x[0]+(0.742039996033)*x[1]**o+(0.649587811572)*x[1]+(0.0749104131981)*x[2]**o+(-0.943341654469)*x[2]
        s[0,1,2,1]=(0.200285377439)*x[0]**o+(0.148329648537)*x[0]+(-0.0994444073677)*x[1]**o+(-0.535438035359)*x[1]+(0.662913271697)*x[2]**o+(0.112409762773)*x[2]
        s[0,2,0,0]=(0.351533608017)*x[0]**o+(0.72576521785)*x[0]+(0.305528610204)*x[1]**o+(-0.139400610385)*x[1]+(-0.455936488645)*x[2]**o+(-0.992626031371)*x[2]
        s[0,2,0,1]=(0.351745471157)*x[0]**o+(0.749900692382)*x[0]+(0.212677388818)*x[1]**o+(0.449543777767)*x[1]+(0.9265756875)*x[2]**o+(0.0973575097882)*x[2]
        s[0,2,1,0]=(0.308471390044)*x[0]**o+(-0.236492160276)*x[0]+(-0.0132386386917)*x[1]**o+(-0.657824382959)*x[1]+(0.567169125655)*x[2]**o+(-0.0239580661983)*x[2]
        s[0,2,1,1]=(-0.71982992429)*x[0]**o+(-0.518285193827)*x[0]+(0.806477863481)*x[1]**o+(-0.949670246654)*x[1]+(-0.559408139171)*x[2]**o+(-0.740059454923)*x[2]
        s[0,2,2,0]=(-0.795819227823)*x[0]**o+(0.0382042059208)*x[0]+(-0.993434888742)*x[1]**o+(-0.714572217225)*x[1]+(0.826057540197)*x[2]**o+(-0.351385211283)*x[2]
        s[0,2,2,1]=(-0.691428669925)*x[0]**o+(0.795109370028)*x[0]+(-0.845580170561)*x[1]**o+(0.66720897925)*x[1]+(-0.604013104606)*x[2]**o+(-0.56003576527)*x[2]
        s[0,3,0,0]=(-0.864934911539)*x[0]**o+(-0.871210034972)*x[0]+(-0.633807262008)*x[1]**o+(-0.0448147244258)*x[1]+(0.601871519726)*x[2]**o+(-0.423895847468)*x[2]
        s[0,3,0,1]=(0.542426383922)*x[0]**o+(0.648956460184)*x[0]+(0.975476287059)*x[1]**o+(-0.110876177567)*x[1]+(-0.334216196036)*x[2]**o+(-0.696393938466)*x[2]
        s[0,3,1,0]=(-0.897499573749)*x[0]**o+(0.504413212966)*x[0]+(-0.114527615291)*x[1]**o+(0.415575654212)*x[1]+(-0.271870634022)*x[2]**o+(0.698810118326)*x[2]
        s[0,3,1,1]=(0.231758813026)*x[0]**o+(-0.207508138983)*x[0]+(-0.744927387959)*x[1]**o+(0.00489240105965)*x[1]+(0.821414365164)*x[2]**o+(0.701679220215)*x[2]
        s[0,3,2,0]=(-0.135996215059)*x[0]**o+(-0.513005242192)*x[0]+(-0.6068585358)*x[1]**o+(-0.852285533431)*x[1]+(0.115701663464)*x[2]**o+(0.243886826101)*x[2]
        s[0,3,2,1]=(-0.80781983055)*x[0]**o+(0.142781343516)*x[0]+(0.764546896757)*x[1]**o+(0.478359930452)*x[1]+(-0.0458636815218)*x[2]**o+(0.582425549793)*x[2]
        s[0,4,0,0]=(-0.382194450205)*x[0]**o+(-0.209573752081)*x[0]+(0.959742867105)*x[1]**o+(0.962658893536)*x[1]+(0.126101921123)*x[2]**o+(0.980806872196)*x[2]
        s[0,4,0,1]=(0.764116933095)*x[0]**o+(-0.648257603162)*x[0]+(-0.839456477431)*x[1]**o+(-0.972616251638)*x[1]+(-0.815159939908)*x[2]**o+(-0.262639210819)*x[2]
        s[0,4,1,0]=(0.453438243261)*x[0]**o+(0.303833796601)*x[0]+(0.100367120622)*x[1]**o+(-0.0716774769052)*x[1]+(-0.659569961883)*x[2]**o+(-0.290044782372)*x[2]
        s[0,4,1,1]=(-0.0286107263502)*x[0]**o+(-0.596438980382)*x[0]+(0.545044960979)*x[1]**o+(-0.796415090612)*x[1]+(0.206907297358)*x[2]**o+(0.306889564609)*x[2]
        s[0,4,2,0]=(-0.201162762403)*x[0]**o+(-0.981302954607)*x[0]+(-0.201138026976)*x[1]**o+(-0.93628245382)*x[1]+(-0.758945968148)*x[2]**o+(-0.860206802477)*x[2]
        s[0,4,2,1]=(0.251350117686)*x[0]**o+(-0.29118495379)*x[0]+(-0.569363617929)*x[1]**o+(0.465078119323)*x[1]+(0.426184283109)*x[2]**o+(-0.873643118829)*x[2]
        s[1,0,0,0]=(-0.0185687413154)*x[0]**o+(0.411847218239)*x[0]+(0.428984420237)*x[1]**o+(-0.547264849731)*x[1]+(0.486727670471)*x[2]**o+(0.179455122763)*x[2]
        s[1,0,0,1]=(0.518433022846)*x[0]**o+(0.025823449242)*x[0]+(-0.0540511749223)*x[1]**o+(-0.85848829682)*x[1]+(0.158385940315)*x[2]**o+(-0.178171294535)*x[2]
        s[1,0,1,0]=(0.0367258516392)*x[0]**o+(0.545049981245)*x[0]+(-0.579044650548)*x[1]**o+(-0.77725873354)*x[1]+(-0.034299394836)*x[2]**o+(-0.179489216766)*x[2]
        s[1,0,1,1]=(0.274129185725)*x[0]**o+(0.166006734196)*x[0]+(-0.773979014851)*x[1]**o+(-0.519501835507)*x[1]+(-0.63912427991)*x[2]**o+(0.466724690101)*x[2]
        s[1,0,2,0]=(0.482955790279)*x[0]**o+(0.471838320459)*x[0]+(0.617771038211)*x[1]**o+(0.048311205466)*x[1]+(0.162393942754)*x[2]**o+(-0.694439775285)*x[2]
        s[1,0,2,1]=(-0.517879311536)*x[0]**o+(-0.730313900842)*x[0]+(-0.991744866252)*x[1]**o+(-0.849707440389)*x[1]+(-0.83668608481)*x[2]**o+(0.581082133088)*x[2]
        s[1,1,0,0]=(0.217328148299)*x[0]**o+(0.187729052039)*x[0]+(-0.68877270235)*x[1]**o+(-0.86757435657)*x[1]+(-0.427061492994)*x[2]**o+(-0.84402543773)*x[2]
        s[1,1,0,1]=(0.766541346426)*x[0]**o+(0.649768542608)*x[0]+(-0.113463877205)*x[1]**o+(-0.637749133894)*x[1]+(0.809878689222)*x[2]**o+(0.315784984005)*x[2]
        s[1,1,1,0]=(0.94173972573)*x[0]**o+(0.668276294204)*x[0]+(0.427204723683)*x[1]**o+(-0.607750747607)*x[1]+(-0.0376503804716)*x[2]**o+(0.960260009026)*x[2]
        s[1,1,1,1]=(-0.400233578622)*x[0]**o+(-0.923564602772)*x[0]+(0.47175931621)*x[1]**o+(-0.612010762622)*x[1]+(0.609225052578)*x[2]**o+(0.526546928045)*x[2]
        s[1,1,2,0]=(0.377678503521)*x[0]**o+(0.892169209135)*x[0]+(0.39772744849)*x[1]**o+(-0.166966626427)*x[1]+(0.874470437184)*x[2]**o+(0.314522738087)*x[2]
        s[1,1,2,1]=(0.520217518976)*x[0]**o+(0.536152467541)*x[0]+(0.444145502111)*x[1]**o+(-0.489860603506)*x[1]+(-0.596335702925)*x[2]**o+(0.646045841797)*x[2]
        s[1,2,0,0]=(-0.976922812773)*x[0]**o+(0.907038408665)*x[0]+(-0.268706703732)*x[1]**o+(-0.820850017822)*x[1]+(-0.164634286435)*x[2]**o+(0.931655051043)*x[2]
        s[1,2,0,1]=(-0.659226400612)*x[0]**o+(0.693884857459)*x[0]+(0.935680326964)*x[1]**o+(0.0529754617738)*x[1]+(-0.975704228522)*x[2]**o+(-0.0103982230046)*x[2]
        s[1,2,1,0]=(-0.728054292372)*x[0]**o+(0.845592631054)*x[0]+(-0.535424426181)*x[1]**o+(-0.489415765636)*x[1]+(0.177904920871)*x[2]**o+(-0.501151392201)*x[2]
        s[1,2,1,1]=(-0.367806306107)*x[0]**o+(-0.979526890288)*x[0]+(-0.851751678512)*x[1]**o+(0.0610833858163)*x[1]+(-0.865092215813)*x[2]**o+(-0.982729156385)*x[2]
        s[1,2,2,0]=(0.890317441495)*x[0]**o+(-0.613493546336)*x[0]+(-0.832732187365)*x[1]**o+(-0.942517758759)*x[1]+(-0.678373862946)*x[2]**o+(-0.760490954483)*x[2]
        s[1,2,2,1]=(0.134821253565)*x[0]**o+(-0.357154926295)*x[0]+(0.984345806097)*x[1]**o+(0.16201908882)*x[1]+(0.673893972474)*x[2]**o+(0.757465706767)*x[2]
        s[1,3,0,0]=(-0.178752771739)*x[0]**o+(0.247240990536)*x[0]+(0.250254101973)*x[1]**o+(-0.290469953005)*x[1]+(-0.179849656966)*x[2]**o+(0.0488351177508)*x[2]
        s[1,3,0,1]=(0.0918959399079)*x[0]**o+(0.929499443933)*x[0]+(-0.142486990007)*x[1]**o+(-0.208154992937)*x[1]+(0.144490842938)*x[2]**o+(0.878566553587)*x[2]
        s[1,3,1,0]=(-0.0951485338271)*x[0]**o+(0.723847998698)*x[0]+(-0.654103310693)*x[1]**o+(-0.00133788653433)*x[1]+(0.0639363181565)*x[2]**o+(-0.240419269084)*x[2]
        s[1,3,1,1]=(-0.217090356907)*x[0]**o+(0.467191197701)*x[0]+(0.0568877529529)*x[1]**o+(0.839967116927)*x[1]+(-0.161819635829)*x[2]**o+(-0.380255238471)*x[2]
        s[1,3,2,0]=(0.139965969921)*x[0]**o+(-0.344141746066)*x[0]+(-0.343097609928)*x[1]**o+(0.323082611123)*x[1]+(0.62470424365)*x[2]**o+(0.747310443443)*x[2]
        s[1,3,2,1]=(-0.3718984783)*x[0]**o+(0.71977639778)*x[0]+(0.191430375709)*x[1]**o+(-0.873015242874)*x[1]+(0.276339628333)*x[2]**o+(0.164476716927)*x[2]
        s[1,4,0,0]=(-0.704023480703)*x[0]**o+(-0.795084252465)*x[0]+(0.637714051748)*x[1]**o+(0.808663623755)*x[1]+(0.718066739763)*x[2]**o+(-0.645424428426)*x[2]
        s[1,4,0,1]=(-0.733930327044)*x[0]**o+(-0.676332066451)*x[0]+(-0.672853173064)*x[1]**o+(0.3429113994)*x[1]+(-0.841264501476)*x[2]**o+(0.79006662583)*x[2]
        s[1,4,1,0]=(-0.528487338584)*x[0]**o+(-0.917646520301)*x[0]+(0.0304777432533)*x[1]**o+(-0.0369253092314)*x[1]+(-0.0256615497575)*x[2]**o+(0.697398576346)*x[2]
        s[1,4,1,1]=(0.326729309985)*x[0]**o+(-0.894823167111)*x[0]+(0.514358304744)*x[1]**o+(-0.173763492858)*x[1]+(-0.802294460015)*x[2]**o+(-0.602884798454)*x[2]
        s[1,4,2,0]=(-0.415051572423)*x[0]**o+(0.22496337443)*x[0]+(0.873058773308)*x[1]**o+(0.655519869148)*x[1]+(-0.480155248885)*x[2]**o+(0.590182974905)*x[2]
        s[1,4,2,1]=(-0.35633147422)*x[0]**o+(-0.717999757228)*x[0]+(0.187573310475)*x[1]**o+(-0.338108475286)*x[1]+(-0.0186934354872)*x[2]**o+(-0.560831597369)*x[2]
        s[2,0,0,0]=(0.0722896121356)*x[0]**o+(0.520666448864)*x[0]+(0.643415007545)*x[1]**o+(0.476128126603)*x[1]+(-0.0519193210173)*x[2]**o+(-0.610576319559)*x[2]
        s[2,0,0,1]=(-0.139391066523)*x[0]**o+(-0.977991803321)*x[0]+(0.222182962215)*x[1]**o+(0.537563638082)*x[1]+(0.61114193397)*x[2]**o+(0.9477815486)*x[2]
        s[2,0,1,0]=(0.177891356821)*x[0]**o+(0.108505736038)*x[0]+(0.251720839868)*x[1]**o+(-0.653300262204)*x[1]+(0.617079986603)*x[2]**o+(0.532007572217)*x[2]
        s[2,0,1,1]=(-0.0963478035027)*x[0]**o+(0.484266295486)*x[0]+(0.575754465131)*x[1]**o+(0.180169918461)*x[1]+(-0.357729254421)*x[2]**o+(-0.198225410436)*x[2]
        s[2,0,2,0]=(-0.389382670145)*x[0]**o+(-0.757173416032)*x[0]+(0.737917980599)*x[1]**o+(0.631381284716)*x[1]+(-0.540034322357)*x[2]**o+(0.0935289005824)*x[2]
        s[2,0,2,1]=(0.18046408091)*x[0]**o+(0.381098419076)*x[0]+(0.375985371309)*x[1]**o+(-0.770665651856)*x[1]+(0.746864092282)*x[2]**o+(0.196258206324)*x[2]
        s[2,1,0,0]=(0.0948044547956)*x[0]**o+(0.530048465174)*x[0]+(0.339203738259)*x[1]**o+(0.889491980197)*x[1]+(-0.701680171971)*x[2]**o+(-0.591236689057)*x[2]
        s[2,1,0,1]=(-0.0504750115604)*x[0]**o+(0.568637246489)*x[0]+(-0.782879070299)*x[1]**o+(-0.398995842249)*x[1]+(0.0900392439579)*x[2]**o+(0.330991009712)*x[2]
        s[2,1,1,0]=(-0.370473438507)*x[0]**o+(-0.894136144949)*x[0]+(-0.724710648209)*x[1]**o+(-0.889034264448)*x[1]+(-0.967606285686)*x[2]**o+(0.229302756578)*x[2]
        s[2,1,1,1]=(-0.557548738746)*x[0]**o+(0.736473161672)*x[0]+(0.132516051242)*x[1]**o+(-0.982095944769)*x[1]+(0.0862581974686)*x[2]**o+(0.648294874083)*x[2]
        s[2,1,2,0]=(0.715221366738)*x[0]**o+(0.180947754635)*x[0]+(0.774447016462)*x[1]**o+(0.2737069937)*x[1]+(-0.308017003693)*x[2]**o+(0.311365430602)*x[2]
        s[2,1,2,1]=(0.057020023918)*x[0]**o+(0.674489665955)*x[0]+(0.907045289702)*x[1]**o+(0.526829006995)*x[1]+(-0.0319147619289)*x[2]**o+(0.561637490349)*x[2]
        s[2,2,0,0]=(0.73848705426)*x[0]**o+(0.578730401802)*x[0]+(-0.877511944183)*x[1]**o+(-0.141821605426)*x[1]+(-0.610728551353)*x[2]**o+(0.921196540913)*x[2]
        s[2,2,0,1]=(-0.518273725024)*x[0]**o+(-0.293070039849)*x[0]+(-0.032874432475)*x[1]**o+(-0.876002064755)*x[1]+(-0.00232979051139)*x[2]**o+(-0.731518329401)*x[2]
        s[2,2,1,0]=(-0.179932544586)*x[0]**o+(-0.298299616601)*x[0]+(-0.507158321537)*x[1]**o+(0.477603386439)*x[1]+(-0.535441161992)*x[2]**o+(-0.114782555164)*x[2]
        s[2,2,1,1]=(-0.357820499429)*x[0]**o+(0.916561554529)*x[0]+(0.918007613979)*x[1]**o+(-0.0497200703463)*x[1]+(-0.196248955784)*x[2]**o+(0.28070766346)*x[2]
        s[2,2,2,0]=(0.686174077835)*x[0]**o+(0.227206066309)*x[0]+(0.0930909399009)*x[1]**o+(-0.772599295714)*x[1]+(0.944100518066)*x[2]**o+(0.515206215749)*x[2]
        s[2,2,2,1]=(0.996149526552)*x[0]**o+(0.828382876873)*x[0]+(0.834473760119)*x[1]**o+(-0.38781643114)*x[1]+(0.946045294719)*x[2]**o+(-0.488214962746)*x[2]
        s[2,3,0,0]=(-0.736854216588)*x[0]**o+(0.837912169829)*x[0]+(0.349541907359)*x[1]**o+(-0.821602065524)*x[1]+(-0.315670105831)*x[2]**o+(0.589034152687)*x[2]
        s[2,3,0,1]=(0.576280744681)*x[0]**o+(0.0780631366472)*x[0]+(0.0280696961332)*x[1]**o+(0.756258137578)*x[1]+(-0.778783563455)*x[2]**o+(-0.549597098669)*x[2]
        s[2,3,1,0]=(-0.986203055922)*x[0]**o+(-0.707899260726)*x[0]+(-0.739480759255)*x[1]**o+(0.438644391188)*x[1]+(-0.66655978156)*x[2]**o+(-0.96252852924)*x[2]
        s[2,3,1,1]=(0.847297144034)*x[0]**o+(-0.107063915702)*x[0]+(-0.737175500381)*x[1]**o+(-0.218515978982)*x[1]+(-0.271489760295)*x[2]**o+(0.178627729308)*x[2]
        s[2,3,2,0]=(-0.375942207511)*x[0]**o+(0.229896842267)*x[0]+(0.0831801744855)*x[1]**o+(0.869789896537)*x[1]+(0.758135854706)*x[2]**o+(-0.452003165838)*x[2]
        s[2,3,2,1]=(-0.612379338977)*x[0]**o+(0.285316223911)*x[0]+(0.571346124001)*x[1]**o+(-0.0206855235392)*x[1]+(-0.229102686518)*x[2]**o+(-0.563723359965)*x[2]
        s[2,4,0,0]=(0.119706352983)*x[0]**o+(-0.32065970677)*x[0]+(0.767856818538)*x[1]**o+(0.188328533895)*x[1]+(0.490938813855)*x[2]**o+(-0.487888173624)*x[2]
        s[2,4,0,1]=(-0.62082438463)*x[0]**o+(0.410195190757)*x[0]+(-0.801936437468)*x[1]**o+(-0.77449746445)*x[1]+(-0.943352435188)*x[2]**o+(-0.0965070084637)*x[2]
        s[2,4,1,0]=(-0.333278752646)*x[0]**o+(0.544632414515)*x[0]+(0.713044447385)*x[1]**o+(-0.134469991847)*x[1]+(-0.477562490558)*x[2]**o+(-0.717272888426)*x[2]
        s[2,4,1,1]=(0.328298581074)*x[0]**o+(0.52907257362)*x[0]+(-0.208998823722)*x[1]**o+(-0.528945934534)*x[1]+(0.230853332264)*x[2]**o+(-0.209517472451)*x[2]
        s[2,4,2,0]=(0.867173594744)*x[0]**o+(0.899212165085)*x[0]+(-0.601859471791)*x[1]**o+(-0.680256055533)*x[1]+(-0.194552761602)*x[2]**o+(0.147401121852)*x[2]
        s[2,4,2,1]=(0.453813057013)*x[0]**o+(0.698426853807)*x[0]+(-0.6460619929)*x[1]**o+(0.033010287169)*x[1]+(0.667588694113)*x[2]**o+(0.356223824338)*x[2]
        s[3,0,0,0]=(-0.292671113925)*x[0]**o+(0.616458693028)*x[0]+(-0.0103186095744)*x[1]**o+(-0.0728030520767)*x[1]+(-0.31974364911)*x[2]**o+(0.33247399209)*x[2]
        s[3,0,0,1]=(-0.992666531346)*x[0]**o+(0.0614528865103)*x[0]+(-0.534772052349)*x[1]**o+(-0.65137500858)*x[1]+(-0.192753007212)*x[2]**o+(0.320918136483)*x[2]
        s[3,0,1,0]=(-0.195020954989)*x[0]**o+(0.391152265328)*x[0]+(-0.0854218011043)*x[1]**o+(-0.925896280434)*x[1]+(0.794208026902)*x[2]**o+(-0.664891436596)*x[2]
        s[3,0,1,1]=(0.76700624636)*x[0]**o+(-0.711830311063)*x[0]+(0.415964892543)*x[1]**o+(-0.266148608643)*x[1]+(0.14347598218)*x[2]**o+(-0.444318428953)*x[2]
        s[3,0,2,0]=(-0.432044237254)*x[0]**o+(0.50230711918)*x[0]+(-0.7887127186)*x[1]**o+(-0.00439564134841)*x[1]+(0.198952339193)*x[2]**o+(-0.365097069005)*x[2]
        s[3,0,2,1]=(0.0731802841742)*x[0]**o+(-0.722529140979)*x[0]+(-0.880840005335)*x[1]**o+(-0.818607956602)*x[1]+(0.893103760434)*x[2]**o+(-0.078910408537)*x[2]
        s[3,1,0,0]=(0.424632189118)*x[0]**o+(-0.9575902588)*x[0]+(-0.669839365109)*x[1]**o+(-0.751737160991)*x[1]+(0.42069725738)*x[2]**o+(0.551903660979)*x[2]
        s[3,1,0,1]=(0.780084669674)*x[0]**o+(0.307501685577)*x[0]+(0.0810136913684)*x[1]**o+(0.702710121149)*x[1]+(-0.123438662606)*x[2]**o+(0.6469150298)*x[2]
        s[3,1,1,0]=(-0.207209742507)*x[0]**o+(0.382918019966)*x[0]+(-0.541859357863)*x[1]**o+(-0.261161834606)*x[1]+(-0.71224108949)*x[2]**o+(0.344568549136)*x[2]
        s[3,1,1,1]=(0.294545201987)*x[0]**o+(-0.310315049878)*x[0]+(0.138438017172)*x[1]**o+(0.343984655033)*x[1]+(0.156101701639)*x[2]**o+(-0.776032313589)*x[2]
        s[3,1,2,0]=(0.670052338135)*x[0]**o+(0.205560268763)*x[0]+(0.104283662799)*x[1]**o+(-0.765280448137)*x[1]+(0.562416102577)*x[2]**o+(-0.876415742019)*x[2]
        s[3,1,2,1]=(0.664650295655)*x[0]**o+(0.795254957531)*x[0]+(-0.552860322817)*x[1]**o+(-0.427967184203)*x[1]+(0.952252946717)*x[2]**o+(0.865393511228)*x[2]
        s[3,2,0,0]=(-0.708432407412)*x[0]**o+(0.308974057108)*x[0]+(-0.749146897323)*x[1]**o+(-0.00624264988729)*x[1]+(0.694675829788)*x[2]**o+(-0.831205537426)*x[2]
        s[3,2,0,1]=(-0.512843864369)*x[0]**o+(0.750200436937)*x[0]+(0.219611724574)*x[1]**o+(0.538914229585)*x[1]+(-0.750949138305)*x[2]**o+(-0.730109185869)*x[2]
        s[3,2,1,0]=(0.307377290744)*x[0]**o+(0.422564975595)*x[0]+(-0.575777057405)*x[1]**o+(0.0843754344109)*x[1]+(0.563608883617)*x[2]**o+(0.0203552581819)*x[2]
        s[3,2,1,1]=(0.844926629988)*x[0]**o+(0.00463764382667)*x[0]+(-0.801305450023)*x[1]**o+(-0.297827538391)*x[1]+(-0.334832523936)*x[2]**o+(-0.145250163264)*x[2]
        s[3,2,2,0]=(-0.571129120261)*x[0]**o+(0.794459190551)*x[0]+(-0.0558676868113)*x[1]**o+(-0.12434806027)*x[1]+(0.666910606164)*x[2]**o+(0.0904375266689)*x[2]
        s[3,2,2,1]=(0.537887532259)*x[0]**o+(0.271597496543)*x[0]+(-0.345121311853)*x[1]**o+(-0.0110951188557)*x[1]+(-0.691735728081)*x[2]**o+(0.227334216755)*x[2]
        s[3,3,0,0]=(-0.617039404022)*x[0]**o+(-0.0600488492522)*x[0]+(0.567478593699)*x[1]**o+(0.625697236766)*x[1]+(0.226173040972)*x[2]**o+(0.592063112338)*x[2]
        s[3,3,0,1]=(0.569700282793)*x[0]**o+(-0.00215167041353)*x[0]+(0.60591609073)*x[1]**o+(0.362583373854)*x[1]+(0.654506836835)*x[2]**o+(-0.144610108143)*x[2]
        s[3,3,1,0]=(0.246640361949)*x[0]**o+(-0.86358850448)*x[0]+(0.157520564836)*x[1]**o+(-0.059233078017)*x[1]+(0.86570653333)*x[2]**o+(0.685576804797)*x[2]
        s[3,3,1,1]=(-0.958130500315)*x[0]**o+(-0.259028395499)*x[0]+(0.120316829457)*x[1]**o+(-0.758878478978)*x[1]+(-0.43066120723)*x[2]**o+(0.800327000625)*x[2]
        s[3,3,2,0]=(0.920887970019)*x[0]**o+(-0.503403737241)*x[0]+(0.0282527472159)*x[1]**o+(-0.167650834072)*x[1]+(0.416811246461)*x[2]**o+(-0.103817262329)*x[2]
        s[3,3,2,1]=(-0.58133852513)*x[0]**o+(-0.176946132911)*x[0]+(0.0347620468937)*x[1]**o+(-0.973916175156)*x[1]+(0.71011981519)*x[2]**o+(-0.339266963019)*x[2]
        s[3,4,0,0]=(-0.0120671975843)*x[0]**o+(0.109594354705)*x[0]+(0.151599950487)*x[1]**o+(0.881150957088)*x[1]+(0.959435811252)*x[2]**o+(-0.86774776724)*x[2]
        s[3,4,0,1]=(-0.0545164901018)*x[0]**o+(-0.987079399801)*x[0]+(0.0168167160052)*x[1]**o+(0.323974666845)*x[1]+(-0.169334724946)*x[2]**o+(-0.95604743937)*x[2]
        s[3,4,1,0]=(0.497905965869)*x[0]**o+(-0.215618743574)*x[0]+(0.333131793569)*x[1]**o+(0.465071861103)*x[1]+(0.472276780484)*x[2]**o+(-0.587116523215)*x[2]
        s[3,4,1,1]=(-0.798850078462)*x[0]**o+(-0.385483486673)*x[0]+(-0.954899272011)*x[1]**o+(0.985953307837)*x[1]+(0.5064455397)*x[2]**o+(0.201146433376)*x[2]
        s[3,4,2,0]=(0.736565587072)*x[0]**o+(-0.383544786329)*x[0]+(0.345205635474)*x[1]**o+(0.118507188601)*x[1]+(0.933209008402)*x[2]**o+(0.00295145110263)*x[2]
        s[3,4,2,1]=(-0.0892194210246)*x[0]**o+(0.152917048)*x[0]+(0.291092526754)*x[1]**o+(-0.13749152394)*x[1]+(-0.820302461745)*x[2]**o+(-0.971364751692)*x[2]
        ref[0,0,0,0]=(-0.209698190814)*x_ref[0]**o+(0.0202198726379)*x_ref[0]+(-0.776762051242)*x_ref[1]**o+(0.0103026495025)*x_ref[1]+(0.0277886960026)*x_ref[2]**o+(-0.255065901162)*x_ref[2]
        ref[0,0,0,1]=(-0.663085958734)*x_ref[0]**o+(0.698525770087)*x_ref[0]+(0.650288968593)*x_ref[1]**o+(0.6069536467)*x_ref[1]+(0.789898121185)*x_ref[2]**o+(-0.13315288412)*x_ref[2]
        ref[0,0,1,0]=(-0.0998227591334)*x_ref[0]**o+(0.75314681829)*x_ref[0]+(-0.43575224825)*x_ref[1]**o+(0.649609416117)*x_ref[1]+(0.271394201051)*x_ref[2]**o+(-0.353712050582)*x_ref[2]
        ref[0,0,1,1]=(0.205441533003)*x_ref[0]**o+(-0.61122881962)*x_ref[0]+(-0.431370221976)*x_ref[1]**o+(-0.183655447483)*x_ref[1]+(0.116437364187)*x_ref[2]**o+(0.741511658665)*x_ref[2]
        ref[0,0,2,0]=(0.516133026558)*x_ref[0]**o+(-0.904105438162)*x_ref[0]+(-0.579514088677)*x_ref[1]**o+(-0.0581332954324)*x_ref[1]+(0.557937742112)*x_ref[2]**o+(0.138639028149)*x_ref[2]
        ref[0,0,2,1]=(0.8991763477)*x_ref[0]**o+(-0.792213819942)*x_ref[0]+(0.674755445696)*x_ref[1]**o+(0.483315869458)*x_ref[1]+(-0.0722431463476)*x_ref[2]**o+(0.0270814412191)*x_ref[2]
        ref[0,1,0,0]=(0.366564587764)*x_ref[0]**o+(0.928578857146)*x_ref[0]+(0.785092228323)*x_ref[1]**o+(0.860812095832)*x_ref[1]+(-0.603464315283)*x_ref[2]**o+(-0.556636685093)*x_ref[2]
        ref[0,1,0,1]=(0.980630839809)*x_ref[0]**o+(-0.255324039441)*x_ref[0]+(0.173630938787)*x_ref[1]**o+(0.110433932158)*x_ref[1]+(-0.392851296251)*x_ref[2]**o+(-0.272438563385)*x_ref[2]
        ref[0,1,1,0]=(-0.63489459343)*x_ref[0]**o+(0.0328950965715)*x_ref[0]+(-0.199264976942)*x_ref[1]**o+(0.704589249988)*x_ref[1]+(0.442306934744)*x_ref[2]**o+(-0.531996394873)*x_ref[2]
        ref[0,1,1,1]=(-0.110263035707)*x_ref[0]**o+(0.904756528197)*x_ref[0]+(0.876124769497)*x_ref[1]**o+(-0.55469630081)*x_ref[1]+(-0.857675499295)*x_ref[2]**o+(0.662352375455)*x_ref[2]
        ref[0,1,2,0]=(-0.383336639984)*x_ref[0]**o+(0.863605840878)*x_ref[0]+(0.742039996033)*x_ref[1]**o+(0.649587811572)*x_ref[1]+(0.0749104131981)*x_ref[2]**o+(-0.943341654469)*x_ref[2]
        ref[0,1,2,1]=(0.200285377439)*x_ref[0]**o+(0.148329648537)*x_ref[0]+(-0.0994444073677)*x_ref[1]**o+(-0.535438035359)*x_ref[1]+(0.662913271697)*x_ref[2]**o+(0.112409762773)*x_ref[2]
        ref[0,2,0,0]=(0.351533608017)*x_ref[0]**o+(0.72576521785)*x_ref[0]+(0.305528610204)*x_ref[1]**o+(-0.139400610385)*x_ref[1]+(-0.455936488645)*x_ref[2]**o+(-0.992626031371)*x_ref[2]
        ref[0,2,0,1]=(0.351745471157)*x_ref[0]**o+(0.749900692382)*x_ref[0]+(0.212677388818)*x_ref[1]**o+(0.449543777767)*x_ref[1]+(0.9265756875)*x_ref[2]**o+(0.0973575097882)*x_ref[2]
        ref[0,2,1,0]=(0.308471390044)*x_ref[0]**o+(-0.236492160276)*x_ref[0]+(-0.0132386386917)*x_ref[1]**o+(-0.657824382959)*x_ref[1]+(0.567169125655)*x_ref[2]**o+(-0.0239580661983)*x_ref[2]
        ref[0,2,1,1]=(-0.71982992429)*x_ref[0]**o+(-0.518285193827)*x_ref[0]+(0.806477863481)*x_ref[1]**o+(-0.949670246654)*x_ref[1]+(-0.559408139171)*x_ref[2]**o+(-0.740059454923)*x_ref[2]
        ref[0,2,2,0]=(-0.795819227823)*x_ref[0]**o+(0.0382042059208)*x_ref[0]+(-0.993434888742)*x_ref[1]**o+(-0.714572217225)*x_ref[1]+(0.826057540197)*x_ref[2]**o+(-0.351385211283)*x_ref[2]
        ref[0,2,2,1]=(-0.691428669925)*x_ref[0]**o+(0.795109370028)*x_ref[0]+(-0.845580170561)*x_ref[1]**o+(0.66720897925)*x_ref[1]+(-0.604013104606)*x_ref[2]**o+(-0.56003576527)*x_ref[2]
        ref[0,3,0,0]=(-0.864934911539)*x_ref[0]**o+(-0.871210034972)*x_ref[0]+(-0.633807262008)*x_ref[1]**o+(-0.0448147244258)*x_ref[1]+(0.601871519726)*x_ref[2]**o+(-0.423895847468)*x_ref[2]
        ref[0,3,0,1]=(0.542426383922)*x_ref[0]**o+(0.648956460184)*x_ref[0]+(0.975476287059)*x_ref[1]**o+(-0.110876177567)*x_ref[1]+(-0.334216196036)*x_ref[2]**o+(-0.696393938466)*x_ref[2]
        ref[0,3,1,0]=(-0.897499573749)*x_ref[0]**o+(0.504413212966)*x_ref[0]+(-0.114527615291)*x_ref[1]**o+(0.415575654212)*x_ref[1]+(-0.271870634022)*x_ref[2]**o+(0.698810118326)*x_ref[2]
        ref[0,3,1,1]=(0.231758813026)*x_ref[0]**o+(-0.207508138983)*x_ref[0]+(-0.744927387959)*x_ref[1]**o+(0.00489240105965)*x_ref[1]+(0.821414365164)*x_ref[2]**o+(0.701679220215)*x_ref[2]
        ref[0,3,2,0]=(-0.135996215059)*x_ref[0]**o+(-0.513005242192)*x_ref[0]+(-0.6068585358)*x_ref[1]**o+(-0.852285533431)*x_ref[1]+(0.115701663464)*x_ref[2]**o+(0.243886826101)*x_ref[2]
        ref[0,3,2,1]=(-0.80781983055)*x_ref[0]**o+(0.142781343516)*x_ref[0]+(0.764546896757)*x_ref[1]**o+(0.478359930452)*x_ref[1]+(-0.0458636815218)*x_ref[2]**o+(0.582425549793)*x_ref[2]
        ref[0,4,0,0]=(-0.382194450205)*x_ref[0]**o+(-0.209573752081)*x_ref[0]+(0.959742867105)*x_ref[1]**o+(0.962658893536)*x_ref[1]+(0.126101921123)*x_ref[2]**o+(0.980806872196)*x_ref[2]
        ref[0,4,0,1]=(0.764116933095)*x_ref[0]**o+(-0.648257603162)*x_ref[0]+(-0.839456477431)*x_ref[1]**o+(-0.972616251638)*x_ref[1]+(-0.815159939908)*x_ref[2]**o+(-0.262639210819)*x_ref[2]
        ref[0,4,1,0]=(0.453438243261)*x_ref[0]**o+(0.303833796601)*x_ref[0]+(0.100367120622)*x_ref[1]**o+(-0.0716774769052)*x_ref[1]+(-0.659569961883)*x_ref[2]**o+(-0.290044782372)*x_ref[2]
        ref[0,4,1,1]=(-0.0286107263502)*x_ref[0]**o+(-0.596438980382)*x_ref[0]+(0.545044960979)*x_ref[1]**o+(-0.796415090612)*x_ref[1]+(0.206907297358)*x_ref[2]**o+(0.306889564609)*x_ref[2]
        ref[0,4,2,0]=(-0.201162762403)*x_ref[0]**o+(-0.981302954607)*x_ref[0]+(-0.201138026976)*x_ref[1]**o+(-0.93628245382)*x_ref[1]+(-0.758945968148)*x_ref[2]**o+(-0.860206802477)*x_ref[2]
        ref[0,4,2,1]=(0.251350117686)*x_ref[0]**o+(-0.29118495379)*x_ref[0]+(-0.569363617929)*x_ref[1]**o+(0.465078119323)*x_ref[1]+(0.426184283109)*x_ref[2]**o+(-0.873643118829)*x_ref[2]
        ref[1,0,0,0]=(-0.0185687413154)*x_ref[0]**o+(0.411847218239)*x_ref[0]+(0.428984420237)*x_ref[1]**o+(-0.547264849731)*x_ref[1]+(0.486727670471)*x_ref[2]**o+(0.179455122763)*x_ref[2]
        ref[1,0,0,1]=(0.518433022846)*x_ref[0]**o+(0.025823449242)*x_ref[0]+(-0.0540511749223)*x_ref[1]**o+(-0.85848829682)*x_ref[1]+(0.158385940315)*x_ref[2]**o+(-0.178171294535)*x_ref[2]
        ref[1,0,1,0]=(0.0367258516392)*x_ref[0]**o+(0.545049981245)*x_ref[0]+(-0.579044650548)*x_ref[1]**o+(-0.77725873354)*x_ref[1]+(-0.034299394836)*x_ref[2]**o+(-0.179489216766)*x_ref[2]
        ref[1,0,1,1]=(0.274129185725)*x_ref[0]**o+(0.166006734196)*x_ref[0]+(-0.773979014851)*x_ref[1]**o+(-0.519501835507)*x_ref[1]+(-0.63912427991)*x_ref[2]**o+(0.466724690101)*x_ref[2]
        ref[1,0,2,0]=(0.482955790279)*x_ref[0]**o+(0.471838320459)*x_ref[0]+(0.617771038211)*x_ref[1]**o+(0.048311205466)*x_ref[1]+(0.162393942754)*x_ref[2]**o+(-0.694439775285)*x_ref[2]
        ref[1,0,2,1]=(-0.517879311536)*x_ref[0]**o+(-0.730313900842)*x_ref[0]+(-0.991744866252)*x_ref[1]**o+(-0.849707440389)*x_ref[1]+(-0.83668608481)*x_ref[2]**o+(0.581082133088)*x_ref[2]
        ref[1,1,0,0]=(0.217328148299)*x_ref[0]**o+(0.187729052039)*x_ref[0]+(-0.68877270235)*x_ref[1]**o+(-0.86757435657)*x_ref[1]+(-0.427061492994)*x_ref[2]**o+(-0.84402543773)*x_ref[2]
        ref[1,1,0,1]=(0.766541346426)*x_ref[0]**o+(0.649768542608)*x_ref[0]+(-0.113463877205)*x_ref[1]**o+(-0.637749133894)*x_ref[1]+(0.809878689222)*x_ref[2]**o+(0.315784984005)*x_ref[2]
        ref[1,1,1,0]=(0.94173972573)*x_ref[0]**o+(0.668276294204)*x_ref[0]+(0.427204723683)*x_ref[1]**o+(-0.607750747607)*x_ref[1]+(-0.0376503804716)*x_ref[2]**o+(0.960260009026)*x_ref[2]
        ref[1,1,1,1]=(-0.400233578622)*x_ref[0]**o+(-0.923564602772)*x_ref[0]+(0.47175931621)*x_ref[1]**o+(-0.612010762622)*x_ref[1]+(0.609225052578)*x_ref[2]**o+(0.526546928045)*x_ref[2]
        ref[1,1,2,0]=(0.377678503521)*x_ref[0]**o+(0.892169209135)*x_ref[0]+(0.39772744849)*x_ref[1]**o+(-0.166966626427)*x_ref[1]+(0.874470437184)*x_ref[2]**o+(0.314522738087)*x_ref[2]
        ref[1,1,2,1]=(0.520217518976)*x_ref[0]**o+(0.536152467541)*x_ref[0]+(0.444145502111)*x_ref[1]**o+(-0.489860603506)*x_ref[1]+(-0.596335702925)*x_ref[2]**o+(0.646045841797)*x_ref[2]
        ref[1,2,0,0]=(-0.976922812773)*x_ref[0]**o+(0.907038408665)*x_ref[0]+(-0.268706703732)*x_ref[1]**o+(-0.820850017822)*x_ref[1]+(-0.164634286435)*x_ref[2]**o+(0.931655051043)*x_ref[2]
        ref[1,2,0,1]=(-0.659226400612)*x_ref[0]**o+(0.693884857459)*x_ref[0]+(0.935680326964)*x_ref[1]**o+(0.0529754617738)*x_ref[1]+(-0.975704228522)*x_ref[2]**o+(-0.0103982230046)*x_ref[2]
        ref[1,2,1,0]=(-0.728054292372)*x_ref[0]**o+(0.845592631054)*x_ref[0]+(-0.535424426181)*x_ref[1]**o+(-0.489415765636)*x_ref[1]+(0.177904920871)*x_ref[2]**o+(-0.501151392201)*x_ref[2]
        ref[1,2,1,1]=(-0.367806306107)*x_ref[0]**o+(-0.979526890288)*x_ref[0]+(-0.851751678512)*x_ref[1]**o+(0.0610833858163)*x_ref[1]+(-0.865092215813)*x_ref[2]**o+(-0.982729156385)*x_ref[2]
        ref[1,2,2,0]=(0.890317441495)*x_ref[0]**o+(-0.613493546336)*x_ref[0]+(-0.832732187365)*x_ref[1]**o+(-0.942517758759)*x_ref[1]+(-0.678373862946)*x_ref[2]**o+(-0.760490954483)*x_ref[2]
        ref[1,2,2,1]=(0.134821253565)*x_ref[0]**o+(-0.357154926295)*x_ref[0]+(0.984345806097)*x_ref[1]**o+(0.16201908882)*x_ref[1]+(0.673893972474)*x_ref[2]**o+(0.757465706767)*x_ref[2]
        ref[1,3,0,0]=(-0.178752771739)*x_ref[0]**o+(0.247240990536)*x_ref[0]+(0.250254101973)*x_ref[1]**o+(-0.290469953005)*x_ref[1]+(-0.179849656966)*x_ref[2]**o+(0.0488351177508)*x_ref[2]
        ref[1,3,0,1]=(0.0918959399079)*x_ref[0]**o+(0.929499443933)*x_ref[0]+(-0.142486990007)*x_ref[1]**o+(-0.208154992937)*x_ref[1]+(0.144490842938)*x_ref[2]**o+(0.878566553587)*x_ref[2]
        ref[1,3,1,0]=(-0.0951485338271)*x_ref[0]**o+(0.723847998698)*x_ref[0]+(-0.654103310693)*x_ref[1]**o+(-0.00133788653433)*x_ref[1]+(0.0639363181565)*x_ref[2]**o+(-0.240419269084)*x_ref[2]
        ref[1,3,1,1]=(-0.217090356907)*x_ref[0]**o+(0.467191197701)*x_ref[0]+(0.0568877529529)*x_ref[1]**o+(0.839967116927)*x_ref[1]+(-0.161819635829)*x_ref[2]**o+(-0.380255238471)*x_ref[2]
        ref[1,3,2,0]=(0.139965969921)*x_ref[0]**o+(-0.344141746066)*x_ref[0]+(-0.343097609928)*x_ref[1]**o+(0.323082611123)*x_ref[1]+(0.62470424365)*x_ref[2]**o+(0.747310443443)*x_ref[2]
        ref[1,3,2,1]=(-0.3718984783)*x_ref[0]**o+(0.71977639778)*x_ref[0]+(0.191430375709)*x_ref[1]**o+(-0.873015242874)*x_ref[1]+(0.276339628333)*x_ref[2]**o+(0.164476716927)*x_ref[2]
        ref[1,4,0,0]=(-0.704023480703)*x_ref[0]**o+(-0.795084252465)*x_ref[0]+(0.637714051748)*x_ref[1]**o+(0.808663623755)*x_ref[1]+(0.718066739763)*x_ref[2]**o+(-0.645424428426)*x_ref[2]
        ref[1,4,0,1]=(-0.733930327044)*x_ref[0]**o+(-0.676332066451)*x_ref[0]+(-0.672853173064)*x_ref[1]**o+(0.3429113994)*x_ref[1]+(-0.841264501476)*x_ref[2]**o+(0.79006662583)*x_ref[2]
        ref[1,4,1,0]=(-0.528487338584)*x_ref[0]**o+(-0.917646520301)*x_ref[0]+(0.0304777432533)*x_ref[1]**o+(-0.0369253092314)*x_ref[1]+(-0.0256615497575)*x_ref[2]**o+(0.697398576346)*x_ref[2]
        ref[1,4,1,1]=(0.326729309985)*x_ref[0]**o+(-0.894823167111)*x_ref[0]+(0.514358304744)*x_ref[1]**o+(-0.173763492858)*x_ref[1]+(-0.802294460015)*x_ref[2]**o+(-0.602884798454)*x_ref[2]
        ref[1,4,2,0]=(-0.415051572423)*x_ref[0]**o+(0.22496337443)*x_ref[0]+(0.873058773308)*x_ref[1]**o+(0.655519869148)*x_ref[1]+(-0.480155248885)*x_ref[2]**o+(0.590182974905)*x_ref[2]
        ref[1,4,2,1]=(-0.35633147422)*x_ref[0]**o+(-0.717999757228)*x_ref[0]+(0.187573310475)*x_ref[1]**o+(-0.338108475286)*x_ref[1]+(-0.0186934354872)*x_ref[2]**o+(-0.560831597369)*x_ref[2]
        ref[2,0,0,0]=(0.0722896121356)*x_ref[0]**o+(0.520666448864)*x_ref[0]+(0.643415007545)*x_ref[1]**o+(0.476128126603)*x_ref[1]+(-0.0519193210173)*x_ref[2]**o+(-0.610576319559)*x_ref[2]
        ref[2,0,0,1]=(-0.139391066523)*x_ref[0]**o+(-0.977991803321)*x_ref[0]+(0.222182962215)*x_ref[1]**o+(0.537563638082)*x_ref[1]+(0.61114193397)*x_ref[2]**o+(0.9477815486)*x_ref[2]
        ref[2,0,1,0]=(0.177891356821)*x_ref[0]**o+(0.108505736038)*x_ref[0]+(0.251720839868)*x_ref[1]**o+(-0.653300262204)*x_ref[1]+(0.617079986603)*x_ref[2]**o+(0.532007572217)*x_ref[2]
        ref[2,0,1,1]=(-0.0963478035027)*x_ref[0]**o+(0.484266295486)*x_ref[0]+(0.575754465131)*x_ref[1]**o+(0.180169918461)*x_ref[1]+(-0.357729254421)*x_ref[2]**o+(-0.198225410436)*x_ref[2]
        ref[2,0,2,0]=(-0.389382670145)*x_ref[0]**o+(-0.757173416032)*x_ref[0]+(0.737917980599)*x_ref[1]**o+(0.631381284716)*x_ref[1]+(-0.540034322357)*x_ref[2]**o+(0.0935289005824)*x_ref[2]
        ref[2,0,2,1]=(0.18046408091)*x_ref[0]**o+(0.381098419076)*x_ref[0]+(0.375985371309)*x_ref[1]**o+(-0.770665651856)*x_ref[1]+(0.746864092282)*x_ref[2]**o+(0.196258206324)*x_ref[2]
        ref[2,1,0,0]=(0.0948044547956)*x_ref[0]**o+(0.530048465174)*x_ref[0]+(0.339203738259)*x_ref[1]**o+(0.889491980197)*x_ref[1]+(-0.701680171971)*x_ref[2]**o+(-0.591236689057)*x_ref[2]
        ref[2,1,0,1]=(-0.0504750115604)*x_ref[0]**o+(0.568637246489)*x_ref[0]+(-0.782879070299)*x_ref[1]**o+(-0.398995842249)*x_ref[1]+(0.0900392439579)*x_ref[2]**o+(0.330991009712)*x_ref[2]
        ref[2,1,1,0]=(-0.370473438507)*x_ref[0]**o+(-0.894136144949)*x_ref[0]+(-0.724710648209)*x_ref[1]**o+(-0.889034264448)*x_ref[1]+(-0.967606285686)*x_ref[2]**o+(0.229302756578)*x_ref[2]
        ref[2,1,1,1]=(-0.557548738746)*x_ref[0]**o+(0.736473161672)*x_ref[0]+(0.132516051242)*x_ref[1]**o+(-0.982095944769)*x_ref[1]+(0.0862581974686)*x_ref[2]**o+(0.648294874083)*x_ref[2]
        ref[2,1,2,0]=(0.715221366738)*x_ref[0]**o+(0.180947754635)*x_ref[0]+(0.774447016462)*x_ref[1]**o+(0.2737069937)*x_ref[1]+(-0.308017003693)*x_ref[2]**o+(0.311365430602)*x_ref[2]
        ref[2,1,2,1]=(0.057020023918)*x_ref[0]**o+(0.674489665955)*x_ref[0]+(0.907045289702)*x_ref[1]**o+(0.526829006995)*x_ref[1]+(-0.0319147619289)*x_ref[2]**o+(0.561637490349)*x_ref[2]
        ref[2,2,0,0]=(0.73848705426)*x_ref[0]**o+(0.578730401802)*x_ref[0]+(-0.877511944183)*x_ref[1]**o+(-0.141821605426)*x_ref[1]+(-0.610728551353)*x_ref[2]**o+(0.921196540913)*x_ref[2]
        ref[2,2,0,1]=(-0.518273725024)*x_ref[0]**o+(-0.293070039849)*x_ref[0]+(-0.032874432475)*x_ref[1]**o+(-0.876002064755)*x_ref[1]+(-0.00232979051139)*x_ref[2]**o+(-0.731518329401)*x_ref[2]
        ref[2,2,1,0]=(-0.179932544586)*x_ref[0]**o+(-0.298299616601)*x_ref[0]+(-0.507158321537)*x_ref[1]**o+(0.477603386439)*x_ref[1]+(-0.535441161992)*x_ref[2]**o+(-0.114782555164)*x_ref[2]
        ref[2,2,1,1]=(-0.357820499429)*x_ref[0]**o+(0.916561554529)*x_ref[0]+(0.918007613979)*x_ref[1]**o+(-0.0497200703463)*x_ref[1]+(-0.196248955784)*x_ref[2]**o+(0.28070766346)*x_ref[2]
        ref[2,2,2,0]=(0.686174077835)*x_ref[0]**o+(0.227206066309)*x_ref[0]+(0.0930909399009)*x_ref[1]**o+(-0.772599295714)*x_ref[1]+(0.944100518066)*x_ref[2]**o+(0.515206215749)*x_ref[2]
        ref[2,2,2,1]=(0.996149526552)*x_ref[0]**o+(0.828382876873)*x_ref[0]+(0.834473760119)*x_ref[1]**o+(-0.38781643114)*x_ref[1]+(0.946045294719)*x_ref[2]**o+(-0.488214962746)*x_ref[2]
        ref[2,3,0,0]=(-0.736854216588)*x_ref[0]**o+(0.837912169829)*x_ref[0]+(0.349541907359)*x_ref[1]**o+(-0.821602065524)*x_ref[1]+(-0.315670105831)*x_ref[2]**o+(0.589034152687)*x_ref[2]
        ref[2,3,0,1]=(0.576280744681)*x_ref[0]**o+(0.0780631366472)*x_ref[0]+(0.0280696961332)*x_ref[1]**o+(0.756258137578)*x_ref[1]+(-0.778783563455)*x_ref[2]**o+(-0.549597098669)*x_ref[2]
        ref[2,3,1,0]=(-0.986203055922)*x_ref[0]**o+(-0.707899260726)*x_ref[0]+(-0.739480759255)*x_ref[1]**o+(0.438644391188)*x_ref[1]+(-0.66655978156)*x_ref[2]**o+(-0.96252852924)*x_ref[2]
        ref[2,3,1,1]=(0.847297144034)*x_ref[0]**o+(-0.107063915702)*x_ref[0]+(-0.737175500381)*x_ref[1]**o+(-0.218515978982)*x_ref[1]+(-0.271489760295)*x_ref[2]**o+(0.178627729308)*x_ref[2]
        ref[2,3,2,0]=(-0.375942207511)*x_ref[0]**o+(0.229896842267)*x_ref[0]+(0.0831801744855)*x_ref[1]**o+(0.869789896537)*x_ref[1]+(0.758135854706)*x_ref[2]**o+(-0.452003165838)*x_ref[2]
        ref[2,3,2,1]=(-0.612379338977)*x_ref[0]**o+(0.285316223911)*x_ref[0]+(0.571346124001)*x_ref[1]**o+(-0.0206855235392)*x_ref[1]+(-0.229102686518)*x_ref[2]**o+(-0.563723359965)*x_ref[2]
        ref[2,4,0,0]=(0.119706352983)*x_ref[0]**o+(-0.32065970677)*x_ref[0]+(0.767856818538)*x_ref[1]**o+(0.188328533895)*x_ref[1]+(0.490938813855)*x_ref[2]**o+(-0.487888173624)*x_ref[2]
        ref[2,4,0,1]=(-0.62082438463)*x_ref[0]**o+(0.410195190757)*x_ref[0]+(-0.801936437468)*x_ref[1]**o+(-0.77449746445)*x_ref[1]+(-0.943352435188)*x_ref[2]**o+(-0.0965070084637)*x_ref[2]
        ref[2,4,1,0]=(-0.333278752646)*x_ref[0]**o+(0.544632414515)*x_ref[0]+(0.713044447385)*x_ref[1]**o+(-0.134469991847)*x_ref[1]+(-0.477562490558)*x_ref[2]**o+(-0.717272888426)*x_ref[2]
        ref[2,4,1,1]=(0.328298581074)*x_ref[0]**o+(0.52907257362)*x_ref[0]+(-0.208998823722)*x_ref[1]**o+(-0.528945934534)*x_ref[1]+(0.230853332264)*x_ref[2]**o+(-0.209517472451)*x_ref[2]
        ref[2,4,2,0]=(0.867173594744)*x_ref[0]**o+(0.899212165085)*x_ref[0]+(-0.601859471791)*x_ref[1]**o+(-0.680256055533)*x_ref[1]+(-0.194552761602)*x_ref[2]**o+(0.147401121852)*x_ref[2]
        ref[2,4,2,1]=(0.453813057013)*x_ref[0]**o+(0.698426853807)*x_ref[0]+(-0.6460619929)*x_ref[1]**o+(0.033010287169)*x_ref[1]+(0.667588694113)*x_ref[2]**o+(0.356223824338)*x_ref[2]
        ref[3,0,0,0]=(-0.292671113925)*x_ref[0]**o+(0.616458693028)*x_ref[0]+(-0.0103186095744)*x_ref[1]**o+(-0.0728030520767)*x_ref[1]+(-0.31974364911)*x_ref[2]**o+(0.33247399209)*x_ref[2]
        ref[3,0,0,1]=(-0.992666531346)*x_ref[0]**o+(0.0614528865103)*x_ref[0]+(-0.534772052349)*x_ref[1]**o+(-0.65137500858)*x_ref[1]+(-0.192753007212)*x_ref[2]**o+(0.320918136483)*x_ref[2]
        ref[3,0,1,0]=(-0.195020954989)*x_ref[0]**o+(0.391152265328)*x_ref[0]+(-0.0854218011043)*x_ref[1]**o+(-0.925896280434)*x_ref[1]+(0.794208026902)*x_ref[2]**o+(-0.664891436596)*x_ref[2]
        ref[3,0,1,1]=(0.76700624636)*x_ref[0]**o+(-0.711830311063)*x_ref[0]+(0.415964892543)*x_ref[1]**o+(-0.266148608643)*x_ref[1]+(0.14347598218)*x_ref[2]**o+(-0.444318428953)*x_ref[2]
        ref[3,0,2,0]=(-0.432044237254)*x_ref[0]**o+(0.50230711918)*x_ref[0]+(-0.7887127186)*x_ref[1]**o+(-0.00439564134841)*x_ref[1]+(0.198952339193)*x_ref[2]**o+(-0.365097069005)*x_ref[2]
        ref[3,0,2,1]=(0.0731802841742)*x_ref[0]**o+(-0.722529140979)*x_ref[0]+(-0.880840005335)*x_ref[1]**o+(-0.818607956602)*x_ref[1]+(0.893103760434)*x_ref[2]**o+(-0.078910408537)*x_ref[2]
        ref[3,1,0,0]=(0.424632189118)*x_ref[0]**o+(-0.9575902588)*x_ref[0]+(-0.669839365109)*x_ref[1]**o+(-0.751737160991)*x_ref[1]+(0.42069725738)*x_ref[2]**o+(0.551903660979)*x_ref[2]
        ref[3,1,0,1]=(0.780084669674)*x_ref[0]**o+(0.307501685577)*x_ref[0]+(0.0810136913684)*x_ref[1]**o+(0.702710121149)*x_ref[1]+(-0.123438662606)*x_ref[2]**o+(0.6469150298)*x_ref[2]
        ref[3,1,1,0]=(-0.207209742507)*x_ref[0]**o+(0.382918019966)*x_ref[0]+(-0.541859357863)*x_ref[1]**o+(-0.261161834606)*x_ref[1]+(-0.71224108949)*x_ref[2]**o+(0.344568549136)*x_ref[2]
        ref[3,1,1,1]=(0.294545201987)*x_ref[0]**o+(-0.310315049878)*x_ref[0]+(0.138438017172)*x_ref[1]**o+(0.343984655033)*x_ref[1]+(0.156101701639)*x_ref[2]**o+(-0.776032313589)*x_ref[2]
        ref[3,1,2,0]=(0.670052338135)*x_ref[0]**o+(0.205560268763)*x_ref[0]+(0.104283662799)*x_ref[1]**o+(-0.765280448137)*x_ref[1]+(0.562416102577)*x_ref[2]**o+(-0.876415742019)*x_ref[2]
        ref[3,1,2,1]=(0.664650295655)*x_ref[0]**o+(0.795254957531)*x_ref[0]+(-0.552860322817)*x_ref[1]**o+(-0.427967184203)*x_ref[1]+(0.952252946717)*x_ref[2]**o+(0.865393511228)*x_ref[2]
        ref[3,2,0,0]=(-0.708432407412)*x_ref[0]**o+(0.308974057108)*x_ref[0]+(-0.749146897323)*x_ref[1]**o+(-0.00624264988729)*x_ref[1]+(0.694675829788)*x_ref[2]**o+(-0.831205537426)*x_ref[2]
        ref[3,2,0,1]=(-0.512843864369)*x_ref[0]**o+(0.750200436937)*x_ref[0]+(0.219611724574)*x_ref[1]**o+(0.538914229585)*x_ref[1]+(-0.750949138305)*x_ref[2]**o+(-0.730109185869)*x_ref[2]
        ref[3,2,1,0]=(0.307377290744)*x_ref[0]**o+(0.422564975595)*x_ref[0]+(-0.575777057405)*x_ref[1]**o+(0.0843754344109)*x_ref[1]+(0.563608883617)*x_ref[2]**o+(0.0203552581819)*x_ref[2]
        ref[3,2,1,1]=(0.844926629988)*x_ref[0]**o+(0.00463764382667)*x_ref[0]+(-0.801305450023)*x_ref[1]**o+(-0.297827538391)*x_ref[1]+(-0.334832523936)*x_ref[2]**o+(-0.145250163264)*x_ref[2]
        ref[3,2,2,0]=(-0.571129120261)*x_ref[0]**o+(0.794459190551)*x_ref[0]+(-0.0558676868113)*x_ref[1]**o+(-0.12434806027)*x_ref[1]+(0.666910606164)*x_ref[2]**o+(0.0904375266689)*x_ref[2]
        ref[3,2,2,1]=(0.537887532259)*x_ref[0]**o+(0.271597496543)*x_ref[0]+(-0.345121311853)*x_ref[1]**o+(-0.0110951188557)*x_ref[1]+(-0.691735728081)*x_ref[2]**o+(0.227334216755)*x_ref[2]
        ref[3,3,0,0]=(-0.617039404022)*x_ref[0]**o+(-0.0600488492522)*x_ref[0]+(0.567478593699)*x_ref[1]**o+(0.625697236766)*x_ref[1]+(0.226173040972)*x_ref[2]**o+(0.592063112338)*x_ref[2]
        ref[3,3,0,1]=(0.569700282793)*x_ref[0]**o+(-0.00215167041353)*x_ref[0]+(0.60591609073)*x_ref[1]**o+(0.362583373854)*x_ref[1]+(0.654506836835)*x_ref[2]**o+(-0.144610108143)*x_ref[2]
        ref[3,3,1,0]=(0.246640361949)*x_ref[0]**o+(-0.86358850448)*x_ref[0]+(0.157520564836)*x_ref[1]**o+(-0.059233078017)*x_ref[1]+(0.86570653333)*x_ref[2]**o+(0.685576804797)*x_ref[2]
        ref[3,3,1,1]=(-0.958130500315)*x_ref[0]**o+(-0.259028395499)*x_ref[0]+(0.120316829457)*x_ref[1]**o+(-0.758878478978)*x_ref[1]+(-0.43066120723)*x_ref[2]**o+(0.800327000625)*x_ref[2]
        ref[3,3,2,0]=(0.920887970019)*x_ref[0]**o+(-0.503403737241)*x_ref[0]+(0.0282527472159)*x_ref[1]**o+(-0.167650834072)*x_ref[1]+(0.416811246461)*x_ref[2]**o+(-0.103817262329)*x_ref[2]
        ref[3,3,2,1]=(-0.58133852513)*x_ref[0]**o+(-0.176946132911)*x_ref[0]+(0.0347620468937)*x_ref[1]**o+(-0.973916175156)*x_ref[1]+(0.71011981519)*x_ref[2]**o+(-0.339266963019)*x_ref[2]
        ref[3,4,0,0]=(-0.0120671975843)*x_ref[0]**o+(0.109594354705)*x_ref[0]+(0.151599950487)*x_ref[1]**o+(0.881150957088)*x_ref[1]+(0.959435811252)*x_ref[2]**o+(-0.86774776724)*x_ref[2]
        ref[3,4,0,1]=(-0.0545164901018)*x_ref[0]**o+(-0.987079399801)*x_ref[0]+(0.0168167160052)*x_ref[1]**o+(0.323974666845)*x_ref[1]+(-0.169334724946)*x_ref[2]**o+(-0.95604743937)*x_ref[2]
        ref[3,4,1,0]=(0.497905965869)*x_ref[0]**o+(-0.215618743574)*x_ref[0]+(0.333131793569)*x_ref[1]**o+(0.465071861103)*x_ref[1]+(0.472276780484)*x_ref[2]**o+(-0.587116523215)*x_ref[2]
        ref[3,4,1,1]=(-0.798850078462)*x_ref[0]**o+(-0.385483486673)*x_ref[0]+(-0.954899272011)*x_ref[1]**o+(0.985953307837)*x_ref[1]+(0.5064455397)*x_ref[2]**o+(0.201146433376)*x_ref[2]
        ref[3,4,2,0]=(0.736565587072)*x_ref[0]**o+(-0.383544786329)*x_ref[0]+(0.345205635474)*x_ref[1]**o+(0.118507188601)*x_ref[1]+(0.933209008402)*x_ref[2]**o+(0.00295145110263)*x_ref[2]
        ref[3,4,2,1]=(-0.0892194210246)*x_ref[0]**o+(0.152917048)*x_ref[0]+(0.291092526754)*x_ref[1]**o+(-0.13749152394)*x_ref[1]+(-0.820302461745)*x_ref[2]**o+(-0.971364751692)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-0.901681414005)*x[0]**o+(-0.888451321413)*x[0]+(-0.22985947361)*x[1]**o+(0.492641717896)*x[1]
        ref=(-0.901681414005)*x_ref[0]**o+(-0.888451321413)*x_ref[0]+(-0.22985947361)*x_ref[1]**o+(0.492641717896)*x_ref[1]
      else:
        arg=(0.428268592739)*x[0]**o+(-0.996603298562)*x[0]+(0.677084172419)*x[1]**o+(0.0826120162779)*x[1]+(0.0206562071525)*x[2]**o+(-0.79500921274)*x[2]
        ref=(0.428268592739)*x_ref[0]**o+(-0.996603298562)*x_ref[0]+(0.677084172419)*x_ref[1]**o+(0.0826120162779)*x_ref[1]+(0.0206562071525)*x_ref[2]**o+(-0.79500921274)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.756853446555)*x[0]**o+(-0.767585641318)*x[0]+(-0.426982757541)*x[1]**o+(0.399806070854)*x[1]
        arg[1]=(-0.821677396516)*x[0]**o+(-0.591723513375)*x[0]+(-0.355824690337)*x[1]**o+(0.1636370609)*x[1]
        ref[0]=(0.756853446555)*x_ref[0]**o+(-0.767585641318)*x_ref[0]+(-0.426982757541)*x_ref[1]**o+(0.399806070854)*x_ref[1]
        ref[1]=(-0.821677396516)*x_ref[0]**o+(-0.591723513375)*x_ref[0]+(-0.355824690337)*x_ref[1]**o+(0.1636370609)*x_ref[1]
      else:
        arg[0]=(0.938054074343)*x[0]**o+(0.539056499967)*x[0]+(-0.617107808198)*x[1]**o+(-0.00223894362929)*x[1]+(-0.789674052883)*x[2]**o+(-0.61592268025)*x[2]
        arg[1]=(-0.820479870378)*x[0]**o+(0.34390095545)*x[0]+(-0.501478928282)*x[1]**o+(0.174440566415)*x[1]+(-0.964557050796)*x[2]**o+(0.116014110585)*x[2]
        ref[0]=(0.938054074343)*x_ref[0]**o+(0.539056499967)*x_ref[0]+(-0.617107808198)*x_ref[1]**o+(-0.00223894362929)*x_ref[1]+(-0.789674052883)*x_ref[2]**o+(-0.61592268025)*x_ref[2]
        ref[1]=(-0.820479870378)*x_ref[0]**o+(0.34390095545)*x_ref[0]+(-0.501478928282)*x_ref[1]**o+(0.174440566415)*x_ref[1]+(-0.964557050796)*x_ref[2]**o+(0.116014110585)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.717924477933)*x[0]**o+(-0.629682284247)*x[0]+(0.0982832832089)*x[1]**o+(-0.762193509689)*x[1]
        arg[0,1]=(-0.997822794773)*x[0]**o+(-0.561696934643)*x[0]+(-0.145516013467)*x[1]**o+(0.630378080139)*x[1]
        arg[0,2]=(-0.751323417918)*x[0]**o+(0.479440101564)*x[0]+(-0.00432956740012)*x[1]**o+(0.767912131968)*x[1]
        arg[0,3]=(0.507272228985)*x[0]**o+(-0.389291668683)*x[0]+(-0.883016581216)*x[1]**o+(0.0437908939974)*x[1]
        arg[0,4]=(-0.137343257739)*x[0]**o+(0.743817161998)*x[0]+(0.507769921129)*x[1]**o+(-0.300283389883)*x[1]
        arg[1,0]=(0.257422912353)*x[0]**o+(0.681717676022)*x[0]+(0.79106312381)*x[1]**o+(-0.671634660385)*x[1]
        arg[1,1]=(0.136314663047)*x[0]**o+(-0.782099147193)*x[0]+(0.885974280892)*x[1]**o+(-0.798425202013)*x[1]
        arg[1,2]=(0.247486687708)*x[0]**o+(0.795328036132)*x[0]+(-0.875686910387)*x[1]**o+(0.0568336903007)*x[1]
        arg[1,3]=(0.080829127843)*x[0]**o+(-0.386433809573)*x[0]+(0.314034401629)*x[1]**o+(-0.872662795541)*x[1]
        arg[1,4]=(-0.807283287271)*x[0]**o+(0.725492503192)*x[0]+(-0.584629907108)*x[1]**o+(-0.268021081298)*x[1]
        arg[2,0]=(0.581563251557)*x[0]**o+(-0.886151193174)*x[0]+(0.193339118534)*x[1]**o+(0.978747305418)*x[1]
        arg[2,1]=(0.141736956675)*x[0]**o+(0.668515974495)*x[0]+(-0.0791701432021)*x[1]**o+(0.410197138729)*x[1]
        arg[2,2]=(-0.287456411619)*x[0]**o+(-0.772391150719)*x[0]+(-0.748682403636)*x[1]**o+(0.538840924967)*x[1]
        arg[2,3]=(0.637529128236)*x[0]**o+(0.471131728149)*x[0]+(-0.433070242121)*x[1]**o+(-0.583240138198)*x[1]
        arg[2,4]=(-0.624702895751)*x[0]**o+(-0.118770841979)*x[0]+(0.587084874695)*x[1]**o+(0.46959556491)*x[1]
        arg[3,0]=(0.209166082888)*x[0]**o+(0.870738524234)*x[0]+(0.00990651539088)*x[1]**o+(-0.423308444734)*x[1]
        arg[3,1]=(0.00446369533847)*x[0]**o+(-0.445436215897)*x[0]+(-0.939826442104)*x[1]**o+(0.486172187596)*x[1]
        arg[3,2]=(-0.697054010365)*x[0]**o+(0.543343700538)*x[0]+(0.491284777488)*x[1]**o+(-0.021410495956)*x[1]
        arg[3,3]=(-0.00457600194633)*x[0]**o+(-0.234648384385)*x[0]+(0.200988754362)*x[1]**o+(0.74772788001)*x[1]
        arg[3,4]=(0.721800958056)*x[0]**o+(0.930824787753)*x[0]+(0.374340208263)*x[1]**o+(0.474293362128)*x[1]
        ref[0,0]=(-0.717924477933)*x_ref[0]**o+(-0.629682284247)*x_ref[0]+(0.0982832832089)*x_ref[1]**o+(-0.762193509689)*x_ref[1]
        ref[0,1]=(-0.997822794773)*x_ref[0]**o+(-0.561696934643)*x_ref[0]+(-0.145516013467)*x_ref[1]**o+(0.630378080139)*x_ref[1]
        ref[0,2]=(-0.751323417918)*x_ref[0]**o+(0.479440101564)*x_ref[0]+(-0.00432956740012)*x_ref[1]**o+(0.767912131968)*x_ref[1]
        ref[0,3]=(0.507272228985)*x_ref[0]**o+(-0.389291668683)*x_ref[0]+(-0.883016581216)*x_ref[1]**o+(0.0437908939974)*x_ref[1]
        ref[0,4]=(-0.137343257739)*x_ref[0]**o+(0.743817161998)*x_ref[0]+(0.507769921129)*x_ref[1]**o+(-0.300283389883)*x_ref[1]
        ref[1,0]=(0.257422912353)*x_ref[0]**o+(0.681717676022)*x_ref[0]+(0.79106312381)*x_ref[1]**o+(-0.671634660385)*x_ref[1]
        ref[1,1]=(0.136314663047)*x_ref[0]**o+(-0.782099147193)*x_ref[0]+(0.885974280892)*x_ref[1]**o+(-0.798425202013)*x_ref[1]
        ref[1,2]=(0.247486687708)*x_ref[0]**o+(0.795328036132)*x_ref[0]+(-0.875686910387)*x_ref[1]**o+(0.0568336903007)*x_ref[1]
        ref[1,3]=(0.080829127843)*x_ref[0]**o+(-0.386433809573)*x_ref[0]+(0.314034401629)*x_ref[1]**o+(-0.872662795541)*x_ref[1]
        ref[1,4]=(-0.807283287271)*x_ref[0]**o+(0.725492503192)*x_ref[0]+(-0.584629907108)*x_ref[1]**o+(-0.268021081298)*x_ref[1]
        ref[2,0]=(0.581563251557)*x_ref[0]**o+(-0.886151193174)*x_ref[0]+(0.193339118534)*x_ref[1]**o+(0.978747305418)*x_ref[1]
        ref[2,1]=(0.141736956675)*x_ref[0]**o+(0.668515974495)*x_ref[0]+(-0.0791701432021)*x_ref[1]**o+(0.410197138729)*x_ref[1]
        ref[2,2]=(-0.287456411619)*x_ref[0]**o+(-0.772391150719)*x_ref[0]+(-0.748682403636)*x_ref[1]**o+(0.538840924967)*x_ref[1]
        ref[2,3]=(0.637529128236)*x_ref[0]**o+(0.471131728149)*x_ref[0]+(-0.433070242121)*x_ref[1]**o+(-0.583240138198)*x_ref[1]
        ref[2,4]=(-0.624702895751)*x_ref[0]**o+(-0.118770841979)*x_ref[0]+(0.587084874695)*x_ref[1]**o+(0.46959556491)*x_ref[1]
        ref[3,0]=(0.209166082888)*x_ref[0]**o+(0.870738524234)*x_ref[0]+(0.00990651539088)*x_ref[1]**o+(-0.423308444734)*x_ref[1]
        ref[3,1]=(0.00446369533847)*x_ref[0]**o+(-0.445436215897)*x_ref[0]+(-0.939826442104)*x_ref[1]**o+(0.486172187596)*x_ref[1]
        ref[3,2]=(-0.697054010365)*x_ref[0]**o+(0.543343700538)*x_ref[0]+(0.491284777488)*x_ref[1]**o+(-0.021410495956)*x_ref[1]
        ref[3,3]=(-0.00457600194633)*x_ref[0]**o+(-0.234648384385)*x_ref[0]+(0.200988754362)*x_ref[1]**o+(0.74772788001)*x_ref[1]
        ref[3,4]=(0.721800958056)*x_ref[0]**o+(0.930824787753)*x_ref[0]+(0.374340208263)*x_ref[1]**o+(0.474293362128)*x_ref[1]
      else:
        arg[0,0]=(-0.740386815771)*x[0]**o+(-0.0523021332821)*x[0]+(0.0398159725211)*x[1]**o+(-0.426888207497)*x[1]+(0.0567170299031)*x[2]**o+(-0.183938925645)*x[2]
        arg[0,1]=(-0.201789590886)*x[0]**o+(-0.367737630124)*x[0]+(0.109364462083)*x[1]**o+(-0.28601743506)*x[1]+(-0.399129253776)*x[2]**o+(0.583952239438)*x[2]
        arg[0,2]=(0.542532512804)*x[0]**o+(0.494398493909)*x[0]+(0.292582889596)*x[1]**o+(-0.954001260258)*x[1]+(0.943184269693)*x[2]**o+(-0.52768765864)*x[2]
        arg[0,3]=(0.804183352494)*x[0]**o+(-0.975349076994)*x[0]+(-0.827375030117)*x[1]**o+(-0.669898333686)*x[1]+(0.748094239981)*x[2]**o+(0.979045988454)*x[2]
        arg[0,4]=(0.425028347472)*x[0]**o+(0.875274586351)*x[0]+(0.854098942537)*x[1]**o+(0.579588487828)*x[1]+(0.082708589378)*x[2]**o+(0.476235952573)*x[2]
        arg[1,0]=(-0.624794007494)*x[0]**o+(-0.138866900862)*x[0]+(0.970509187706)*x[1]**o+(0.0253783907447)*x[1]+(0.468865716232)*x[2]**o+(0.676697194907)*x[2]
        arg[1,1]=(-0.153766978847)*x[0]**o+(-0.426626120688)*x[0]+(-0.773847907631)*x[1]**o+(0.245927709658)*x[1]+(0.255639913539)*x[2]**o+(0.808676567134)*x[2]
        arg[1,2]=(0.232094367521)*x[0]**o+(0.297706438396)*x[0]+(0.446407126908)*x[1]**o+(0.740223236781)*x[1]+(0.39141114872)*x[2]**o+(0.773847600596)*x[2]
        arg[1,3]=(0.255568261799)*x[0]**o+(-0.978846043266)*x[0]+(0.354385282456)*x[1]**o+(0.105066752677)*x[1]+(0.942760820578)*x[2]**o+(0.780092891693)*x[2]
        arg[1,4]=(0.529084170195)*x[0]**o+(0.747311333653)*x[0]+(-0.545359901284)*x[1]**o+(-0.281724707651)*x[1]+(0.313967713269)*x[2]**o+(-0.215058831724)*x[2]
        arg[2,0]=(-0.831081561199)*x[0]**o+(-0.773735958748)*x[0]+(0.368966594678)*x[1]**o+(0.666937704315)*x[1]+(-0.974922863204)*x[2]**o+(0.715578049273)*x[2]
        arg[2,1]=(0.578362294564)*x[0]**o+(0.889684843543)*x[0]+(0.588806147025)*x[1]**o+(0.0617034215889)*x[1]+(-0.16631520681)*x[2]**o+(0.730086241349)*x[2]
        arg[2,2]=(0.448838826514)*x[0]**o+(-0.0709459785553)*x[0]+(-0.876698569979)*x[1]**o+(-0.506314680523)*x[1]+(0.688738500291)*x[2]**o+(0.618955116168)*x[2]
        arg[2,3]=(0.290663008974)*x[0]**o+(0.73644965206)*x[0]+(-0.14313225421)*x[1]**o+(0.83074119153)*x[1]+(-0.247893391581)*x[2]**o+(-0.541433435359)*x[2]
        arg[2,4]=(-0.215556795169)*x[0]**o+(-0.340257823449)*x[0]+(0.326604718415)*x[1]**o+(-0.939889251927)*x[1]+(0.425604617369)*x[2]**o+(0.430639358886)*x[2]
        arg[3,0]=(0.0317232868586)*x[0]**o+(0.820693319131)*x[0]+(0.973940331646)*x[1]**o+(0.626976945683)*x[1]+(0.788447937745)*x[2]**o+(-0.718491682844)*x[2]
        arg[3,1]=(0.292924833121)*x[0]**o+(-0.648128515891)*x[0]+(-0.300791515646)*x[1]**o+(-0.26123312127)*x[1]+(-0.263541248274)*x[2]**o+(0.890699912174)*x[2]
        arg[3,2]=(-0.741612826841)*x[0]**o+(-0.095466430566)*x[0]+(-0.987682602605)*x[1]**o+(-0.0258256251685)*x[1]+(0.544387724417)*x[2]**o+(0.917676395583)*x[2]
        arg[3,3]=(0.148313573908)*x[0]**o+(-0.438648673746)*x[0]+(-0.151998255103)*x[1]**o+(-0.500903665338)*x[1]+(0.61275976182)*x[2]**o+(-0.0509716866391)*x[2]
        arg[3,4]=(0.292189003653)*x[0]**o+(0.564355242514)*x[0]+(-0.585718580333)*x[1]**o+(0.635365614994)*x[1]+(-0.00185257696349)*x[2]**o+(0.813357014726)*x[2]
        ref[0,0]=(-0.740386815771)*x_ref[0]**o+(-0.0523021332821)*x_ref[0]+(0.0398159725211)*x_ref[1]**o+(-0.426888207497)*x_ref[1]+(0.0567170299031)*x_ref[2]**o+(-0.183938925645)*x_ref[2]
        ref[0,1]=(-0.201789590886)*x_ref[0]**o+(-0.367737630124)*x_ref[0]+(0.109364462083)*x_ref[1]**o+(-0.28601743506)*x_ref[1]+(-0.399129253776)*x_ref[2]**o+(0.583952239438)*x_ref[2]
        ref[0,2]=(0.542532512804)*x_ref[0]**o+(0.494398493909)*x_ref[0]+(0.292582889596)*x_ref[1]**o+(-0.954001260258)*x_ref[1]+(0.943184269693)*x_ref[2]**o+(-0.52768765864)*x_ref[2]
        ref[0,3]=(0.804183352494)*x_ref[0]**o+(-0.975349076994)*x_ref[0]+(-0.827375030117)*x_ref[1]**o+(-0.669898333686)*x_ref[1]+(0.748094239981)*x_ref[2]**o+(0.979045988454)*x_ref[2]
        ref[0,4]=(0.425028347472)*x_ref[0]**o+(0.875274586351)*x_ref[0]+(0.854098942537)*x_ref[1]**o+(0.579588487828)*x_ref[1]+(0.082708589378)*x_ref[2]**o+(0.476235952573)*x_ref[2]
        ref[1,0]=(-0.624794007494)*x_ref[0]**o+(-0.138866900862)*x_ref[0]+(0.970509187706)*x_ref[1]**o+(0.0253783907447)*x_ref[1]+(0.468865716232)*x_ref[2]**o+(0.676697194907)*x_ref[2]
        ref[1,1]=(-0.153766978847)*x_ref[0]**o+(-0.426626120688)*x_ref[0]+(-0.773847907631)*x_ref[1]**o+(0.245927709658)*x_ref[1]+(0.255639913539)*x_ref[2]**o+(0.808676567134)*x_ref[2]
        ref[1,2]=(0.232094367521)*x_ref[0]**o+(0.297706438396)*x_ref[0]+(0.446407126908)*x_ref[1]**o+(0.740223236781)*x_ref[1]+(0.39141114872)*x_ref[2]**o+(0.773847600596)*x_ref[2]
        ref[1,3]=(0.255568261799)*x_ref[0]**o+(-0.978846043266)*x_ref[0]+(0.354385282456)*x_ref[1]**o+(0.105066752677)*x_ref[1]+(0.942760820578)*x_ref[2]**o+(0.780092891693)*x_ref[2]
        ref[1,4]=(0.529084170195)*x_ref[0]**o+(0.747311333653)*x_ref[0]+(-0.545359901284)*x_ref[1]**o+(-0.281724707651)*x_ref[1]+(0.313967713269)*x_ref[2]**o+(-0.215058831724)*x_ref[2]
        ref[2,0]=(-0.831081561199)*x_ref[0]**o+(-0.773735958748)*x_ref[0]+(0.368966594678)*x_ref[1]**o+(0.666937704315)*x_ref[1]+(-0.974922863204)*x_ref[2]**o+(0.715578049273)*x_ref[2]
        ref[2,1]=(0.578362294564)*x_ref[0]**o+(0.889684843543)*x_ref[0]+(0.588806147025)*x_ref[1]**o+(0.0617034215889)*x_ref[1]+(-0.16631520681)*x_ref[2]**o+(0.730086241349)*x_ref[2]
        ref[2,2]=(0.448838826514)*x_ref[0]**o+(-0.0709459785553)*x_ref[0]+(-0.876698569979)*x_ref[1]**o+(-0.506314680523)*x_ref[1]+(0.688738500291)*x_ref[2]**o+(0.618955116168)*x_ref[2]
        ref[2,3]=(0.290663008974)*x_ref[0]**o+(0.73644965206)*x_ref[0]+(-0.14313225421)*x_ref[1]**o+(0.83074119153)*x_ref[1]+(-0.247893391581)*x_ref[2]**o+(-0.541433435359)*x_ref[2]
        ref[2,4]=(-0.215556795169)*x_ref[0]**o+(-0.340257823449)*x_ref[0]+(0.326604718415)*x_ref[1]**o+(-0.939889251927)*x_ref[1]+(0.425604617369)*x_ref[2]**o+(0.430639358886)*x_ref[2]
        ref[3,0]=(0.0317232868586)*x_ref[0]**o+(0.820693319131)*x_ref[0]+(0.973940331646)*x_ref[1]**o+(0.626976945683)*x_ref[1]+(0.788447937745)*x_ref[2]**o+(-0.718491682844)*x_ref[2]
        ref[3,1]=(0.292924833121)*x_ref[0]**o+(-0.648128515891)*x_ref[0]+(-0.300791515646)*x_ref[1]**o+(-0.26123312127)*x_ref[1]+(-0.263541248274)*x_ref[2]**o+(0.890699912174)*x_ref[2]
        ref[3,2]=(-0.741612826841)*x_ref[0]**o+(-0.095466430566)*x_ref[0]+(-0.987682602605)*x_ref[1]**o+(-0.0258256251685)*x_ref[1]+(0.544387724417)*x_ref[2]**o+(0.917676395583)*x_ref[2]
        ref[3,3]=(0.148313573908)*x_ref[0]**o+(-0.438648673746)*x_ref[0]+(-0.151998255103)*x_ref[1]**o+(-0.500903665338)*x_ref[1]+(0.61275976182)*x_ref[2]**o+(-0.0509716866391)*x_ref[2]
        ref[3,4]=(0.292189003653)*x_ref[0]**o+(0.564355242514)*x_ref[0]+(-0.585718580333)*x_ref[1]**o+(0.635365614994)*x_ref[1]+(-0.00185257696349)*x_ref[2]**o+(0.813357014726)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.0128414848839)*x[0]**o+(0.465050006132)*x[0]+(0.965707055318)*x[1]**o+(-0.0245081370356)*x[1]
        arg[0,0,1]=(-0.200472047017)*x[0]**o+(-0.663028900371)*x[0]+(-0.943040910702)*x[1]**o+(-0.120563561139)*x[1]
        arg[0,1,0]=(-0.582794137412)*x[0]**o+(-0.17277488426)*x[0]+(-0.328856513193)*x[1]**o+(-0.995542154323)*x[1]
        arg[0,1,1]=(0.628900216244)*x[0]**o+(0.42314911968)*x[0]+(0.720137185717)*x[1]**o+(-0.360847007396)*x[1]
        arg[1,0,0]=(-0.320599577005)*x[0]**o+(0.551919914795)*x[0]+(-0.0459652088094)*x[1]**o+(-0.559395421304)*x[1]
        arg[1,0,1]=(0.148179561705)*x[0]**o+(0.434674232208)*x[0]+(-0.475375603732)*x[1]**o+(-0.58939903081)*x[1]
        arg[1,1,0]=(0.617416045561)*x[0]**o+(0.937293925453)*x[0]+(0.348092333305)*x[1]**o+(0.503125307597)*x[1]
        arg[1,1,1]=(-0.569221008948)*x[0]**o+(-0.00405502766105)*x[0]+(-0.389443082681)*x[1]**o+(-0.149118473431)*x[1]
        arg[2,0,0]=(0.118769588477)*x[0]**o+(-0.96708351582)*x[0]+(-0.0145855500917)*x[1]**o+(0.370997674141)*x[1]
        arg[2,0,1]=(-0.3129615573)*x[0]**o+(0.206299148764)*x[0]+(-0.721288027396)*x[1]**o+(-0.884411444055)*x[1]
        arg[2,1,0]=(-0.780163860911)*x[0]**o+(-0.568714964563)*x[0]+(-0.0381936992532)*x[1]**o+(0.830243364454)*x[1]
        arg[2,1,1]=(0.440674933834)*x[0]**o+(0.659074990645)*x[0]+(0.373520241513)*x[1]**o+(0.609301500641)*x[1]
        arg[3,0,0]=(0.377299608317)*x[0]**o+(0.258769326825)*x[0]+(0.108755954544)*x[1]**o+(0.416736701504)*x[1]
        arg[3,0,1]=(-0.368375671218)*x[0]**o+(-0.194408664196)*x[0]+(0.22066484944)*x[1]**o+(-0.105513517239)*x[1]
        arg[3,1,0]=(-0.45311889531)*x[0]**o+(0.758137223649)*x[0]+(0.642250313677)*x[1]**o+(0.67780041241)*x[1]
        arg[3,1,1]=(-0.584467518038)*x[0]**o+(0.873450785862)*x[0]+(0.949121750467)*x[1]**o+(-0.774601385823)*x[1]
        arg[4,0,0]=(0.36378291068)*x[0]**o+(-0.489477884746)*x[0]+(-0.665910069342)*x[1]**o+(0.9839391739)*x[1]
        arg[4,0,1]=(-0.539599763616)*x[0]**o+(0.651501126417)*x[0]+(-0.291030902108)*x[1]**o+(0.719794225744)*x[1]
        arg[4,1,0]=(0.144776367586)*x[0]**o+(-0.154981280527)*x[0]+(0.170526162601)*x[1]**o+(-0.821596149241)*x[1]
        arg[4,1,1]=(-0.460621351209)*x[0]**o+(-0.75486897537)*x[0]+(0.615982502995)*x[1]**o+(-0.756893047476)*x[1]
        arg[5,0,0]=(0.209926220994)*x[0]**o+(0.412132938536)*x[0]+(0.28119780747)*x[1]**o+(-0.829072062922)*x[1]
        arg[5,0,1]=(0.347615632334)*x[0]**o+(-0.101954769092)*x[0]+(-0.535842225724)*x[1]**o+(-0.0790159065772)*x[1]
        arg[5,1,0]=(-0.951105825038)*x[0]**o+(-0.167113372047)*x[0]+(-0.368722457931)*x[1]**o+(0.00398630826759)*x[1]
        arg[5,1,1]=(-0.280966220945)*x[0]**o+(0.749467953656)*x[0]+(-0.862358852162)*x[1]**o+(-0.947469365289)*x[1]
        ref[0,0,0]=(-0.0128414848839)*x_ref[0]**o+(0.465050006132)*x_ref[0]+(0.965707055318)*x_ref[1]**o+(-0.0245081370356)*x_ref[1]
        ref[0,0,1]=(-0.200472047017)*x_ref[0]**o+(-0.663028900371)*x_ref[0]+(-0.943040910702)*x_ref[1]**o+(-0.120563561139)*x_ref[1]
        ref[0,1,0]=(-0.582794137412)*x_ref[0]**o+(-0.17277488426)*x_ref[0]+(-0.328856513193)*x_ref[1]**o+(-0.995542154323)*x_ref[1]
        ref[0,1,1]=(0.628900216244)*x_ref[0]**o+(0.42314911968)*x_ref[0]+(0.720137185717)*x_ref[1]**o+(-0.360847007396)*x_ref[1]
        ref[1,0,0]=(-0.320599577005)*x_ref[0]**o+(0.551919914795)*x_ref[0]+(-0.0459652088094)*x_ref[1]**o+(-0.559395421304)*x_ref[1]
        ref[1,0,1]=(0.148179561705)*x_ref[0]**o+(0.434674232208)*x_ref[0]+(-0.475375603732)*x_ref[1]**o+(-0.58939903081)*x_ref[1]
        ref[1,1,0]=(0.617416045561)*x_ref[0]**o+(0.937293925453)*x_ref[0]+(0.348092333305)*x_ref[1]**o+(0.503125307597)*x_ref[1]
        ref[1,1,1]=(-0.569221008948)*x_ref[0]**o+(-0.00405502766105)*x_ref[0]+(-0.389443082681)*x_ref[1]**o+(-0.149118473431)*x_ref[1]
        ref[2,0,0]=(0.118769588477)*x_ref[0]**o+(-0.96708351582)*x_ref[0]+(-0.0145855500917)*x_ref[1]**o+(0.370997674141)*x_ref[1]
        ref[2,0,1]=(-0.3129615573)*x_ref[0]**o+(0.206299148764)*x_ref[0]+(-0.721288027396)*x_ref[1]**o+(-0.884411444055)*x_ref[1]
        ref[2,1,0]=(-0.780163860911)*x_ref[0]**o+(-0.568714964563)*x_ref[0]+(-0.0381936992532)*x_ref[1]**o+(0.830243364454)*x_ref[1]
        ref[2,1,1]=(0.440674933834)*x_ref[0]**o+(0.659074990645)*x_ref[0]+(0.373520241513)*x_ref[1]**o+(0.609301500641)*x_ref[1]
        ref[3,0,0]=(0.377299608317)*x_ref[0]**o+(0.258769326825)*x_ref[0]+(0.108755954544)*x_ref[1]**o+(0.416736701504)*x_ref[1]
        ref[3,0,1]=(-0.368375671218)*x_ref[0]**o+(-0.194408664196)*x_ref[0]+(0.22066484944)*x_ref[1]**o+(-0.105513517239)*x_ref[1]
        ref[3,1,0]=(-0.45311889531)*x_ref[0]**o+(0.758137223649)*x_ref[0]+(0.642250313677)*x_ref[1]**o+(0.67780041241)*x_ref[1]
        ref[3,1,1]=(-0.584467518038)*x_ref[0]**o+(0.873450785862)*x_ref[0]+(0.949121750467)*x_ref[1]**o+(-0.774601385823)*x_ref[1]
        ref[4,0,0]=(0.36378291068)*x_ref[0]**o+(-0.489477884746)*x_ref[0]+(-0.665910069342)*x_ref[1]**o+(0.9839391739)*x_ref[1]
        ref[4,0,1]=(-0.539599763616)*x_ref[0]**o+(0.651501126417)*x_ref[0]+(-0.291030902108)*x_ref[1]**o+(0.719794225744)*x_ref[1]
        ref[4,1,0]=(0.144776367586)*x_ref[0]**o+(-0.154981280527)*x_ref[0]+(0.170526162601)*x_ref[1]**o+(-0.821596149241)*x_ref[1]
        ref[4,1,1]=(-0.460621351209)*x_ref[0]**o+(-0.75486897537)*x_ref[0]+(0.615982502995)*x_ref[1]**o+(-0.756893047476)*x_ref[1]
        ref[5,0,0]=(0.209926220994)*x_ref[0]**o+(0.412132938536)*x_ref[0]+(0.28119780747)*x_ref[1]**o+(-0.829072062922)*x_ref[1]
        ref[5,0,1]=(0.347615632334)*x_ref[0]**o+(-0.101954769092)*x_ref[0]+(-0.535842225724)*x_ref[1]**o+(-0.0790159065772)*x_ref[1]
        ref[5,1,0]=(-0.951105825038)*x_ref[0]**o+(-0.167113372047)*x_ref[0]+(-0.368722457931)*x_ref[1]**o+(0.00398630826759)*x_ref[1]
        ref[5,1,1]=(-0.280966220945)*x_ref[0]**o+(0.749467953656)*x_ref[0]+(-0.862358852162)*x_ref[1]**o+(-0.947469365289)*x_ref[1]
      else:
        arg[0,0,0]=(0.0415833284333)*x[0]**o+(0.194911622005)*x[0]+(0.50312866507)*x[1]**o+(0.0264724613449)*x[1]+(0.8791018499)*x[2]**o+(0.55217332601)*x[2]
        arg[0,0,1]=(0.272403975118)*x[0]**o+(0.747717699355)*x[0]+(0.630267108269)*x[1]**o+(-0.532392610953)*x[1]+(-0.335800539303)*x[2]**o+(-0.45202877276)*x[2]
        arg[0,1,0]=(-0.345230079718)*x[0]**o+(-0.463428154207)*x[0]+(-0.235612061263)*x[1]**o+(0.25205454387)*x[1]+(0.00383801658594)*x[2]**o+(0.94886179389)*x[2]
        arg[0,1,1]=(0.83425074828)*x[0]**o+(0.0678634336979)*x[0]+(0.58303694875)*x[1]**o+(-0.390423615317)*x[1]+(-0.198506665254)*x[2]**o+(-0.153926424072)*x[2]
        arg[1,0,0]=(0.748583887807)*x[0]**o+(-0.619758023753)*x[0]+(-0.637192600624)*x[1]**o+(-0.77944796466)*x[1]+(-0.745635449844)*x[2]**o+(0.385052490848)*x[2]
        arg[1,0,1]=(0.178249452349)*x[0]**o+(-0.758000510749)*x[0]+(0.173656449057)*x[1]**o+(-0.496282271341)*x[1]+(-0.75929155861)*x[2]**o+(-0.0916648747906)*x[2]
        arg[1,1,0]=(-0.106375867266)*x[0]**o+(-0.985973054508)*x[0]+(-0.828346807601)*x[1]**o+(-0.780869332747)*x[1]+(0.956166673219)*x[2]**o+(0.0115486522377)*x[2]
        arg[1,1,1]=(-0.347056651581)*x[0]**o+(0.159310474468)*x[0]+(0.169356329598)*x[1]**o+(0.0379131764296)*x[1]+(0.547302672178)*x[2]**o+(-0.192168337756)*x[2]
        arg[2,0,0]=(-0.264646719961)*x[0]**o+(-0.32990817159)*x[0]+(0.360599653268)*x[1]**o+(0.53317997944)*x[1]+(-0.288609084526)*x[2]**o+(0.117445043254)*x[2]
        arg[2,0,1]=(-0.717210086649)*x[0]**o+(-0.913859607432)*x[0]+(0.494198228468)*x[1]**o+(-0.52463221583)*x[1]+(0.711597567219)*x[2]**o+(0.585196869172)*x[2]
        arg[2,1,0]=(-0.398182449924)*x[0]**o+(-0.224237207359)*x[0]+(0.245111044519)*x[1]**o+(0.556757042489)*x[1]+(-0.747490064606)*x[2]**o+(0.974848330189)*x[2]
        arg[2,1,1]=(-0.525098204339)*x[0]**o+(-0.787485492675)*x[0]+(-0.233307380294)*x[1]**o+(-0.609570071623)*x[1]+(-0.709676638973)*x[2]**o+(0.428630406545)*x[2]
        arg[3,0,0]=(-0.604801204556)*x[0]**o+(0.483355489609)*x[0]+(0.602251424119)*x[1]**o+(-0.512483059948)*x[1]+(0.530298952371)*x[2]**o+(0.460582484929)*x[2]
        arg[3,0,1]=(-0.223026657674)*x[0]**o+(-0.843908334265)*x[0]+(0.630773219249)*x[1]**o+(-0.962463994312)*x[1]+(-0.086366368162)*x[2]**o+(0.441208496119)*x[2]
        arg[3,1,0]=(-0.861051836152)*x[0]**o+(0.282995621585)*x[0]+(-0.971513512619)*x[1]**o+(0.691494798209)*x[1]+(-0.264269824543)*x[2]**o+(0.430075879498)*x[2]
        arg[3,1,1]=(0.195995685886)*x[0]**o+(-0.441184708365)*x[0]+(-0.768193293663)*x[1]**o+(-0.308275906926)*x[1]+(0.849067046286)*x[2]**o+(-0.822000111049)*x[2]
        arg[4,0,0]=(-0.776936350216)*x[0]**o+(0.987846537986)*x[0]+(0.027119101665)*x[1]**o+(-0.0179353982234)*x[1]+(-0.453287099644)*x[2]**o+(0.970207770741)*x[2]
        arg[4,0,1]=(0.781708265017)*x[0]**o+(-0.625176653149)*x[0]+(0.0637837953361)*x[1]**o+(0.381401415932)*x[1]+(0.575142751522)*x[2]**o+(0.482009069423)*x[2]
        arg[4,1,0]=(-0.864838932985)*x[0]**o+(0.116211445658)*x[0]+(-0.103985339842)*x[1]**o+(0.616435777123)*x[1]+(0.0343826165658)*x[2]**o+(0.67732900142)*x[2]
        arg[4,1,1]=(0.637545268689)*x[0]**o+(0.495935489884)*x[0]+(-0.949001441035)*x[1]**o+(-0.600568665738)*x[1]+(0.339278287549)*x[2]**o+(0.979389021599)*x[2]
        arg[5,0,0]=(-0.486412877541)*x[0]**o+(-0.220311090368)*x[0]+(-0.941323254025)*x[1]**o+(0.122065018067)*x[1]+(0.537799827381)*x[2]**o+(0.942570757444)*x[2]
        arg[5,0,1]=(0.713697533198)*x[0]**o+(-0.382404769437)*x[0]+(0.629644547983)*x[1]**o+(-0.618351494049)*x[1]+(0.257178470315)*x[2]**o+(-0.488029666124)*x[2]
        arg[5,1,0]=(-0.531207567018)*x[0]**o+(-0.873678479376)*x[0]+(0.344856295667)*x[1]**o+(-0.401836367886)*x[1]+(0.397967958128)*x[2]**o+(-0.518169984782)*x[2]
        arg[5,1,1]=(-0.0270647126884)*x[0]**o+(-0.493698732948)*x[0]+(-0.523416157841)*x[1]**o+(-0.710897753488)*x[1]+(-0.597569566681)*x[2]**o+(0.240935331308)*x[2]
        ref[0,0,0]=(0.0415833284333)*x_ref[0]**o+(0.194911622005)*x_ref[0]+(0.50312866507)*x_ref[1]**o+(0.0264724613449)*x_ref[1]+(0.8791018499)*x_ref[2]**o+(0.55217332601)*x_ref[2]
        ref[0,0,1]=(0.272403975118)*x_ref[0]**o+(0.747717699355)*x_ref[0]+(0.630267108269)*x_ref[1]**o+(-0.532392610953)*x_ref[1]+(-0.335800539303)*x_ref[2]**o+(-0.45202877276)*x_ref[2]
        ref[0,1,0]=(-0.345230079718)*x_ref[0]**o+(-0.463428154207)*x_ref[0]+(-0.235612061263)*x_ref[1]**o+(0.25205454387)*x_ref[1]+(0.00383801658594)*x_ref[2]**o+(0.94886179389)*x_ref[2]
        ref[0,1,1]=(0.83425074828)*x_ref[0]**o+(0.0678634336979)*x_ref[0]+(0.58303694875)*x_ref[1]**o+(-0.390423615317)*x_ref[1]+(-0.198506665254)*x_ref[2]**o+(-0.153926424072)*x_ref[2]
        ref[1,0,0]=(0.748583887807)*x_ref[0]**o+(-0.619758023753)*x_ref[0]+(-0.637192600624)*x_ref[1]**o+(-0.77944796466)*x_ref[1]+(-0.745635449844)*x_ref[2]**o+(0.385052490848)*x_ref[2]
        ref[1,0,1]=(0.178249452349)*x_ref[0]**o+(-0.758000510749)*x_ref[0]+(0.173656449057)*x_ref[1]**o+(-0.496282271341)*x_ref[1]+(-0.75929155861)*x_ref[2]**o+(-0.0916648747906)*x_ref[2]
        ref[1,1,0]=(-0.106375867266)*x_ref[0]**o+(-0.985973054508)*x_ref[0]+(-0.828346807601)*x_ref[1]**o+(-0.780869332747)*x_ref[1]+(0.956166673219)*x_ref[2]**o+(0.0115486522377)*x_ref[2]
        ref[1,1,1]=(-0.347056651581)*x_ref[0]**o+(0.159310474468)*x_ref[0]+(0.169356329598)*x_ref[1]**o+(0.0379131764296)*x_ref[1]+(0.547302672178)*x_ref[2]**o+(-0.192168337756)*x_ref[2]
        ref[2,0,0]=(-0.264646719961)*x_ref[0]**o+(-0.32990817159)*x_ref[0]+(0.360599653268)*x_ref[1]**o+(0.53317997944)*x_ref[1]+(-0.288609084526)*x_ref[2]**o+(0.117445043254)*x_ref[2]
        ref[2,0,1]=(-0.717210086649)*x_ref[0]**o+(-0.913859607432)*x_ref[0]+(0.494198228468)*x_ref[1]**o+(-0.52463221583)*x_ref[1]+(0.711597567219)*x_ref[2]**o+(0.585196869172)*x_ref[2]
        ref[2,1,0]=(-0.398182449924)*x_ref[0]**o+(-0.224237207359)*x_ref[0]+(0.245111044519)*x_ref[1]**o+(0.556757042489)*x_ref[1]+(-0.747490064606)*x_ref[2]**o+(0.974848330189)*x_ref[2]
        ref[2,1,1]=(-0.525098204339)*x_ref[0]**o+(-0.787485492675)*x_ref[0]+(-0.233307380294)*x_ref[1]**o+(-0.609570071623)*x_ref[1]+(-0.709676638973)*x_ref[2]**o+(0.428630406545)*x_ref[2]
        ref[3,0,0]=(-0.604801204556)*x_ref[0]**o+(0.483355489609)*x_ref[0]+(0.602251424119)*x_ref[1]**o+(-0.512483059948)*x_ref[1]+(0.530298952371)*x_ref[2]**o+(0.460582484929)*x_ref[2]
        ref[3,0,1]=(-0.223026657674)*x_ref[0]**o+(-0.843908334265)*x_ref[0]+(0.630773219249)*x_ref[1]**o+(-0.962463994312)*x_ref[1]+(-0.086366368162)*x_ref[2]**o+(0.441208496119)*x_ref[2]
        ref[3,1,0]=(-0.861051836152)*x_ref[0]**o+(0.282995621585)*x_ref[0]+(-0.971513512619)*x_ref[1]**o+(0.691494798209)*x_ref[1]+(-0.264269824543)*x_ref[2]**o+(0.430075879498)*x_ref[2]
        ref[3,1,1]=(0.195995685886)*x_ref[0]**o+(-0.441184708365)*x_ref[0]+(-0.768193293663)*x_ref[1]**o+(-0.308275906926)*x_ref[1]+(0.849067046286)*x_ref[2]**o+(-0.822000111049)*x_ref[2]
        ref[4,0,0]=(-0.776936350216)*x_ref[0]**o+(0.987846537986)*x_ref[0]+(0.027119101665)*x_ref[1]**o+(-0.0179353982234)*x_ref[1]+(-0.453287099644)*x_ref[2]**o+(0.970207770741)*x_ref[2]
        ref[4,0,1]=(0.781708265017)*x_ref[0]**o+(-0.625176653149)*x_ref[0]+(0.0637837953361)*x_ref[1]**o+(0.381401415932)*x_ref[1]+(0.575142751522)*x_ref[2]**o+(0.482009069423)*x_ref[2]
        ref[4,1,0]=(-0.864838932985)*x_ref[0]**o+(0.116211445658)*x_ref[0]+(-0.103985339842)*x_ref[1]**o+(0.616435777123)*x_ref[1]+(0.0343826165658)*x_ref[2]**o+(0.67732900142)*x_ref[2]
        ref[4,1,1]=(0.637545268689)*x_ref[0]**o+(0.495935489884)*x_ref[0]+(-0.949001441035)*x_ref[1]**o+(-0.600568665738)*x_ref[1]+(0.339278287549)*x_ref[2]**o+(0.979389021599)*x_ref[2]
        ref[5,0,0]=(-0.486412877541)*x_ref[0]**o+(-0.220311090368)*x_ref[0]+(-0.941323254025)*x_ref[1]**o+(0.122065018067)*x_ref[1]+(0.537799827381)*x_ref[2]**o+(0.942570757444)*x_ref[2]
        ref[5,0,1]=(0.713697533198)*x_ref[0]**o+(-0.382404769437)*x_ref[0]+(0.629644547983)*x_ref[1]**o+(-0.618351494049)*x_ref[1]+(0.257178470315)*x_ref[2]**o+(-0.488029666124)*x_ref[2]
        ref[5,1,0]=(-0.531207567018)*x_ref[0]**o+(-0.873678479376)*x_ref[0]+(0.344856295667)*x_ref[1]**o+(-0.401836367886)*x_ref[1]+(0.397967958128)*x_ref[2]**o+(-0.518169984782)*x_ref[2]
        ref[5,1,1]=(-0.0270647126884)*x_ref[0]**o+(-0.493698732948)*x_ref[0]+(-0.523416157841)*x_ref[1]**o+(-0.710897753488)*x_ref[1]+(-0.597569566681)*x_ref[2]**o+(0.240935331308)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.70041600073)*x[0]**o+(0.0368444393254)*x[0]+(-0.939011570256)*x[1]**o+(0.762796835248)*x[1]
        arg[0,0,0,1]=(-0.383188465228)*x[0]**o+(-0.956202930503)*x[0]+(-0.0433250191873)*x[1]**o+(-0.263494367743)*x[1]
        arg[0,0,1,0]=(0.494889208666)*x[0]**o+(0.721800653649)*x[0]+(0.275748820606)*x[1]**o+(0.190701062508)*x[1]
        arg[0,0,1,1]=(0.317345288325)*x[0]**o+(-0.0713887739659)*x[0]+(-0.0901699882517)*x[1]**o+(0.350861486991)*x[1]
        arg[0,0,2,0]=(-0.0399664417282)*x[0]**o+(-0.73007860251)*x[0]+(-0.123301660907)*x[1]**o+(0.935774110316)*x[1]
        arg[0,0,2,1]=(-0.774484178038)*x[0]**o+(-0.427643344733)*x[0]+(0.00466934570081)*x[1]**o+(0.644433150981)*x[1]
        arg[0,1,0,0]=(0.532951038821)*x[0]**o+(-0.267404179599)*x[0]+(-0.488651484135)*x[1]**o+(-0.354416641555)*x[1]
        arg[0,1,0,1]=(0.816219473407)*x[0]**o+(0.095041670769)*x[0]+(0.130824542151)*x[1]**o+(-0.91799466789)*x[1]
        arg[0,1,1,0]=(-0.926659688215)*x[0]**o+(-0.99257560771)*x[0]+(0.76472846278)*x[1]**o+(-0.77301872817)*x[1]
        arg[0,1,1,1]=(-0.291848070507)*x[0]**o+(0.988869135358)*x[0]+(-0.355431590669)*x[1]**o+(-0.397109808154)*x[1]
        arg[0,1,2,0]=(-0.630380994644)*x[0]**o+(0.121188015244)*x[0]+(-0.527521198931)*x[1]**o+(0.743015163843)*x[1]
        arg[0,1,2,1]=(0.624350853711)*x[0]**o+(-0.269065524076)*x[0]+(0.381181149379)*x[1]**o+(0.964176747301)*x[1]
        arg[0,2,0,0]=(0.386078831223)*x[0]**o+(-0.143893625755)*x[0]+(-0.957649154953)*x[1]**o+(-0.728625629809)*x[1]
        arg[0,2,0,1]=(-0.752542988554)*x[0]**o+(-0.334338454849)*x[0]+(-0.888583790871)*x[1]**o+(-0.49515709924)*x[1]
        arg[0,2,1,0]=(0.67213418036)*x[0]**o+(0.63778853358)*x[0]+(-0.305139731169)*x[1]**o+(-0.349307973811)*x[1]
        arg[0,2,1,1]=(0.226718551492)*x[0]**o+(0.0927018122394)*x[0]+(-0.742387419573)*x[1]**o+(0.528087866126)*x[1]
        arg[0,2,2,0]=(-0.724349079038)*x[0]**o+(-0.758167121924)*x[0]+(-0.687742086176)*x[1]**o+(-0.415119729664)*x[1]
        arg[0,2,2,1]=(0.0399247837941)*x[0]**o+(-0.728504254588)*x[0]+(0.382011768693)*x[1]**o+(-0.743726886782)*x[1]
        arg[0,3,0,0]=(-0.994002993417)*x[0]**o+(0.293311671699)*x[0]+(-0.9731492726)*x[1]**o+(0.0673759518597)*x[1]
        arg[0,3,0,1]=(-0.638426251977)*x[0]**o+(-0.537134730931)*x[0]+(0.993741859056)*x[1]**o+(-0.909049905282)*x[1]
        arg[0,3,1,0]=(-0.994595072064)*x[0]**o+(0.0252993150106)*x[0]+(-0.739129165128)*x[1]**o+(-0.132376880599)*x[1]
        arg[0,3,1,1]=(-0.539219874104)*x[0]**o+(0.796849273908)*x[0]+(0.682514446887)*x[1]**o+(-0.704901251482)*x[1]
        arg[0,3,2,0]=(0.0401802421987)*x[0]**o+(-0.0354713202277)*x[0]+(-0.660064980444)*x[1]**o+(0.0477264910037)*x[1]
        arg[0,3,2,1]=(0.832824897985)*x[0]**o+(0.877654166375)*x[0]+(-0.909219867154)*x[1]**o+(-0.796258795798)*x[1]
        arg[0,4,0,0]=(0.0310731814412)*x[0]**o+(0.39714436047)*x[0]+(0.61669106719)*x[1]**o+(0.789287424452)*x[1]
        arg[0,4,0,1]=(-0.8919943646)*x[0]**o+(0.55041132884)*x[0]+(-0.0397475694211)*x[1]**o+(0.993245641481)*x[1]
        arg[0,4,1,0]=(0.550134389567)*x[0]**o+(-0.657732937381)*x[0]+(0.204770242863)*x[1]**o+(-0.0296391840031)*x[1]
        arg[0,4,1,1]=(0.753185285169)*x[0]**o+(0.74420982803)*x[0]+(0.108834676972)*x[1]**o+(0.183891847557)*x[1]
        arg[0,4,2,0]=(-0.259653246466)*x[0]**o+(-0.120308818888)*x[0]+(-0.827119024697)*x[1]**o+(-0.672357922311)*x[1]
        arg[0,4,2,1]=(-0.807666107771)*x[0]**o+(0.873748263895)*x[0]+(-0.160494957388)*x[1]**o+(0.7551018226)*x[1]
        arg[1,0,0,0]=(0.780978036182)*x[0]**o+(-0.260409851047)*x[0]+(-0.597455360543)*x[1]**o+(0.0689482125215)*x[1]
        arg[1,0,0,1]=(0.811121169417)*x[0]**o+(-0.614791318516)*x[0]+(0.663679164632)*x[1]**o+(-0.853931508608)*x[1]
        arg[1,0,1,0]=(-0.601473156596)*x[0]**o+(-0.1232863608)*x[0]+(0.200905517492)*x[1]**o+(0.337784236126)*x[1]
        arg[1,0,1,1]=(-0.924015099396)*x[0]**o+(0.197096316191)*x[0]+(-0.580606258161)*x[1]**o+(0.517410299916)*x[1]
        arg[1,0,2,0]=(0.32384939376)*x[0]**o+(0.871622229492)*x[0]+(0.403013167399)*x[1]**o+(-0.50285362219)*x[1]
        arg[1,0,2,1]=(-0.875266612786)*x[0]**o+(0.773236652169)*x[0]+(0.93949088522)*x[1]**o+(-0.607791659163)*x[1]
        arg[1,1,0,0]=(-0.972124688294)*x[0]**o+(0.895545224323)*x[0]+(-0.904953764575)*x[1]**o+(0.388629839442)*x[1]
        arg[1,1,0,1]=(0.792386948636)*x[0]**o+(-0.467370445542)*x[0]+(0.16664753332)*x[1]**o+(-0.202002832911)*x[1]
        arg[1,1,1,0]=(0.748979883957)*x[0]**o+(-0.272594191544)*x[0]+(-0.724533609431)*x[1]**o+(0.0670080576995)*x[1]
        arg[1,1,1,1]=(0.155216275716)*x[0]**o+(-0.394137469825)*x[0]+(-0.527098097796)*x[1]**o+(0.504014563188)*x[1]
        arg[1,1,2,0]=(-0.806704090439)*x[0]**o+(-0.791435275759)*x[0]+(0.866962482)*x[1]**o+(0.60302217776)*x[1]
        arg[1,1,2,1]=(-0.728053309956)*x[0]**o+(0.0559583022356)*x[0]+(0.0926294355801)*x[1]**o+(-0.835984838669)*x[1]
        arg[1,2,0,0]=(-0.0661747052735)*x[0]**o+(0.0537098973346)*x[0]+(0.310702398279)*x[1]**o+(-0.489393478975)*x[1]
        arg[1,2,0,1]=(-0.0910654471721)*x[0]**o+(0.327801831182)*x[0]+(-0.42833439187)*x[1]**o+(0.267067714743)*x[1]
        arg[1,2,1,0]=(-0.693916583922)*x[0]**o+(-0.100272651463)*x[0]+(-0.195971225883)*x[1]**o+(0.660851591786)*x[1]
        arg[1,2,1,1]=(-0.158662866194)*x[0]**o+(0.756412009601)*x[0]+(-0.923974707134)*x[1]**o+(0.61664700863)*x[1]
        arg[1,2,2,0]=(-0.449981632451)*x[0]**o+(0.272083805383)*x[0]+(-0.207156704608)*x[1]**o+(0.413596665304)*x[1]
        arg[1,2,2,1]=(0.67039491127)*x[0]**o+(-0.606081942118)*x[0]+(-0.358862791857)*x[1]**o+(-0.191412169743)*x[1]
        arg[1,3,0,0]=(0.214562206436)*x[0]**o+(0.475358432768)*x[0]+(0.639237285541)*x[1]**o+(-0.221799470417)*x[1]
        arg[1,3,0,1]=(0.468332920849)*x[0]**o+(0.359173095186)*x[0]+(0.513170435198)*x[1]**o+(0.472621770345)*x[1]
        arg[1,3,1,0]=(-0.722500106774)*x[0]**o+(0.781087784307)*x[0]+(0.633141730145)*x[1]**o+(0.958777327006)*x[1]
        arg[1,3,1,1]=(0.488269631028)*x[0]**o+(0.148575659631)*x[0]+(-0.229858507326)*x[1]**o+(-0.857118507195)*x[1]
        arg[1,3,2,0]=(0.99463505025)*x[0]**o+(-0.970275234551)*x[0]+(-0.431529655117)*x[1]**o+(-0.208112330684)*x[1]
        arg[1,3,2,1]=(-0.715411408332)*x[0]**o+(0.554401881121)*x[0]+(-0.980883542633)*x[1]**o+(0.316356494975)*x[1]
        arg[1,4,0,0]=(0.747620638111)*x[0]**o+(-0.505093883319)*x[0]+(-0.217289877652)*x[1]**o+(-0.122961101484)*x[1]
        arg[1,4,0,1]=(-0.329916740819)*x[0]**o+(0.778840575985)*x[0]+(-0.696382782242)*x[1]**o+(-0.096185679235)*x[1]
        arg[1,4,1,0]=(0.574218122656)*x[0]**o+(0.0949666147475)*x[0]+(-0.789015428301)*x[1]**o+(-0.400376747438)*x[1]
        arg[1,4,1,1]=(0.569036824328)*x[0]**o+(0.132679527923)*x[0]+(-0.229208784312)*x[1]**o+(0.405134795549)*x[1]
        arg[1,4,2,0]=(0.0123943827354)*x[0]**o+(-0.0446024014977)*x[0]+(-0.221944819571)*x[1]**o+(-0.596890233794)*x[1]
        arg[1,4,2,1]=(0.499247385143)*x[0]**o+(0.891843495388)*x[0]+(-0.348380026198)*x[1]**o+(0.146344723761)*x[1]
        arg[2,0,0,0]=(-0.929032869237)*x[0]**o+(-0.438035650633)*x[0]+(0.117030848904)*x[1]**o+(-0.489094945626)*x[1]
        arg[2,0,0,1]=(0.661960656348)*x[0]**o+(0.437346307001)*x[0]+(0.15713464744)*x[1]**o+(0.578309634731)*x[1]
        arg[2,0,1,0]=(-0.368410149772)*x[0]**o+(-0.135484081321)*x[0]+(-0.51575875197)*x[1]**o+(0.26997748887)*x[1]
        arg[2,0,1,1]=(0.784159013964)*x[0]**o+(-0.558040177257)*x[0]+(-0.100963880851)*x[1]**o+(-0.319503517522)*x[1]
        arg[2,0,2,0]=(-0.0260076717983)*x[0]**o+(-0.252140406297)*x[0]+(-0.335458650519)*x[1]**o+(-0.0739605444814)*x[1]
        arg[2,0,2,1]=(0.484818876559)*x[0]**o+(-0.59157296366)*x[0]+(-0.83491407929)*x[1]**o+(0.143217129439)*x[1]
        arg[2,1,0,0]=(0.672216209481)*x[0]**o+(0.225722770162)*x[0]+(-0.46608814621)*x[1]**o+(-0.779035303765)*x[1]
        arg[2,1,0,1]=(-0.738780643364)*x[0]**o+(0.0741487570855)*x[0]+(-0.362480267529)*x[1]**o+(0.474314990314)*x[1]
        arg[2,1,1,0]=(0.942380641613)*x[0]**o+(0.834468091419)*x[0]+(-0.413376836135)*x[1]**o+(0.996367350352)*x[1]
        arg[2,1,1,1]=(0.544360198404)*x[0]**o+(-0.327857763859)*x[0]+(-0.00306338907092)*x[1]**o+(0.448394961269)*x[1]
        arg[2,1,2,0]=(-0.345632293594)*x[0]**o+(0.505451030738)*x[0]+(-0.644095213138)*x[1]**o+(0.603588130789)*x[1]
        arg[2,1,2,1]=(-0.300809177049)*x[0]**o+(-0.531168939571)*x[0]+(0.513243419239)*x[1]**o+(0.62918148667)*x[1]
        arg[2,2,0,0]=(0.610906338615)*x[0]**o+(-0.656455478681)*x[0]+(-0.386484840223)*x[1]**o+(0.720074853711)*x[1]
        arg[2,2,0,1]=(-0.403600090425)*x[0]**o+(0.454789490943)*x[0]+(-0.399073455918)*x[1]**o+(-0.0552326255369)*x[1]
        arg[2,2,1,0]=(-0.308592576779)*x[0]**o+(0.253298307559)*x[0]+(-0.862040156595)*x[1]**o+(0.195088197599)*x[1]
        arg[2,2,1,1]=(0.143353860538)*x[0]**o+(-0.590315683091)*x[0]+(-0.955465510619)*x[1]**o+(0.372352027984)*x[1]
        arg[2,2,2,0]=(-0.13314652456)*x[0]**o+(0.580744198064)*x[0]+(-0.903035808104)*x[1]**o+(-0.193570555816)*x[1]
        arg[2,2,2,1]=(-0.647218355992)*x[0]**o+(0.00521075379725)*x[0]+(0.877022602564)*x[1]**o+(-0.104569127278)*x[1]
        arg[2,3,0,0]=(-0.532184390508)*x[0]**o+(0.739226859366)*x[0]+(0.595122128361)*x[1]**o+(0.0135063212516)*x[1]
        arg[2,3,0,1]=(-0.121073574394)*x[0]**o+(-0.175990877109)*x[0]+(-0.00319744367579)*x[1]**o+(0.140264888668)*x[1]
        arg[2,3,1,0]=(-0.0778296806357)*x[0]**o+(0.760344573573)*x[0]+(-0.250209333031)*x[1]**o+(-0.413073766011)*x[1]
        arg[2,3,1,1]=(-0.809655592364)*x[0]**o+(-0.147000481704)*x[0]+(-0.760732613787)*x[1]**o+(0.399494900513)*x[1]
        arg[2,3,2,0]=(-0.131633255327)*x[0]**o+(-0.0268295912955)*x[0]+(-0.934414490057)*x[1]**o+(-0.339746795066)*x[1]
        arg[2,3,2,1]=(0.503687469605)*x[0]**o+(-0.415050204706)*x[0]+(0.152421197486)*x[1]**o+(0.529931932615)*x[1]
        arg[2,4,0,0]=(0.930580833077)*x[0]**o+(-0.377748298816)*x[0]+(-0.159931528324)*x[1]**o+(0.374143651635)*x[1]
        arg[2,4,0,1]=(0.254339441506)*x[0]**o+(0.695672173367)*x[0]+(-0.499989877641)*x[1]**o+(-0.745148857716)*x[1]
        arg[2,4,1,0]=(0.205325187322)*x[0]**o+(0.317887475767)*x[0]+(0.755338840391)*x[1]**o+(-0.106253370416)*x[1]
        arg[2,4,1,1]=(0.175711808412)*x[0]**o+(-0.437006733188)*x[0]+(-0.449737979076)*x[1]**o+(-0.0421697213234)*x[1]
        arg[2,4,2,0]=(-0.543184540283)*x[0]**o+(0.616024774255)*x[0]+(-0.429403317932)*x[1]**o+(-0.557076956301)*x[1]
        arg[2,4,2,1]=(0.224704163386)*x[0]**o+(-0.945537746026)*x[0]+(0.947303520688)*x[1]**o+(0.909695468251)*x[1]
        arg[3,0,0,0]=(-0.671167109353)*x[0]**o+(-0.513526771737)*x[0]+(-0.511672331777)*x[1]**o+(-0.301926771817)*x[1]
        arg[3,0,0,1]=(0.39343937992)*x[0]**o+(-0.194833946828)*x[0]+(0.882702377998)*x[1]**o+(0.191544038874)*x[1]
        arg[3,0,1,0]=(0.521271797889)*x[0]**o+(0.924897421226)*x[0]+(-0.597268916449)*x[1]**o+(-0.897621830518)*x[1]
        arg[3,0,1,1]=(-0.446311941735)*x[0]**o+(-0.223997368297)*x[0]+(-0.348064954666)*x[1]**o+(0.661704488202)*x[1]
        arg[3,0,2,0]=(-0.785706283015)*x[0]**o+(0.434487336824)*x[0]+(-0.0447968030358)*x[1]**o+(0.0231287166506)*x[1]
        arg[3,0,2,1]=(0.347155629717)*x[0]**o+(-0.836660082409)*x[0]+(0.391492038747)*x[1]**o+(0.883578619998)*x[1]
        arg[3,1,0,0]=(0.571669585809)*x[0]**o+(0.966740526379)*x[0]+(-0.657144215536)*x[1]**o+(0.357025591929)*x[1]
        arg[3,1,0,1]=(-0.269900462792)*x[0]**o+(0.770680682031)*x[0]+(0.744696440247)*x[1]**o+(0.755227887237)*x[1]
        arg[3,1,1,0]=(-0.667568521409)*x[0]**o+(-0.483641971546)*x[0]+(0.687266187401)*x[1]**o+(-0.597241584922)*x[1]
        arg[3,1,1,1]=(-0.947316636771)*x[0]**o+(0.0797873425176)*x[0]+(0.254921442278)*x[1]**o+(-0.0500536165406)*x[1]
        arg[3,1,2,0]=(0.966192971983)*x[0]**o+(-0.597929938773)*x[0]+(-0.174360538)*x[1]**o+(0.46405478391)*x[1]
        arg[3,1,2,1]=(-0.675528960994)*x[0]**o+(0.51539766317)*x[0]+(-0.99673299807)*x[1]**o+(-0.488746102106)*x[1]
        arg[3,2,0,0]=(-0.604138916979)*x[0]**o+(-0.0431165687312)*x[0]+(-0.346697341122)*x[1]**o+(-0.966811623167)*x[1]
        arg[3,2,0,1]=(0.963965750281)*x[0]**o+(-0.059166688536)*x[0]+(0.166515828506)*x[1]**o+(0.832637512663)*x[1]
        arg[3,2,1,0]=(-0.808155600589)*x[0]**o+(-0.178526483687)*x[0]+(-0.012382108719)*x[1]**o+(0.194810458281)*x[1]
        arg[3,2,1,1]=(0.905922872414)*x[0]**o+(0.580755461959)*x[0]+(0.84253193847)*x[1]**o+(0.300676716124)*x[1]
        arg[3,2,2,0]=(-0.560928128044)*x[0]**o+(0.215498997249)*x[0]+(0.630359499164)*x[1]**o+(0.0955049343157)*x[1]
        arg[3,2,2,1]=(-0.977307154392)*x[0]**o+(0.543314643158)*x[0]+(-0.617962918616)*x[1]**o+(0.222318165123)*x[1]
        arg[3,3,0,0]=(0.910351367774)*x[0]**o+(-0.349015106312)*x[0]+(-0.367900620181)*x[1]**o+(-0.371131872755)*x[1]
        arg[3,3,0,1]=(0.663510744888)*x[0]**o+(-0.16380815063)*x[0]+(0.770895156906)*x[1]**o+(0.84395768389)*x[1]
        arg[3,3,1,0]=(0.553097521137)*x[0]**o+(-0.139012828014)*x[0]+(-0.986291992888)*x[1]**o+(-0.780024041157)*x[1]
        arg[3,3,1,1]=(-0.894441791323)*x[0]**o+(0.91199469535)*x[0]+(-0.349154843905)*x[1]**o+(0.538946349908)*x[1]
        arg[3,3,2,0]=(-0.464541735353)*x[0]**o+(-0.44894927329)*x[0]+(-0.488621463959)*x[1]**o+(-0.987756549629)*x[1]
        arg[3,3,2,1]=(0.136069699682)*x[0]**o+(-0.794904844715)*x[0]+(0.300038435268)*x[1]**o+(-0.599562792523)*x[1]
        arg[3,4,0,0]=(0.391227355996)*x[0]**o+(-0.634937772688)*x[0]+(0.0618326566602)*x[1]**o+(0.396379987618)*x[1]
        arg[3,4,0,1]=(0.485984857628)*x[0]**o+(-0.884694203334)*x[0]+(-0.791392877516)*x[1]**o+(0.832758707832)*x[1]
        arg[3,4,1,0]=(0.983115009543)*x[0]**o+(0.830522788331)*x[0]+(-0.947942131374)*x[1]**o+(0.890681434885)*x[1]
        arg[3,4,1,1]=(-0.0596041988514)*x[0]**o+(0.203878482428)*x[0]+(-0.739288417295)*x[1]**o+(0.150351531263)*x[1]
        arg[3,4,2,0]=(-0.821412631737)*x[0]**o+(0.366097632474)*x[0]+(-0.978999319755)*x[1]**o+(-0.556584079904)*x[1]
        arg[3,4,2,1]=(0.677346101991)*x[0]**o+(0.617303832579)*x[0]+(0.963322024959)*x[1]**o+(0.0765656809305)*x[1]
        ref[0,0,0,0]=(0.70041600073)*x_ref[0]**o+(0.0368444393254)*x_ref[0]+(-0.939011570256)*x_ref[1]**o+(0.762796835248)*x_ref[1]
        ref[0,0,0,1]=(-0.383188465228)*x_ref[0]**o+(-0.956202930503)*x_ref[0]+(-0.0433250191873)*x_ref[1]**o+(-0.263494367743)*x_ref[1]
        ref[0,0,1,0]=(0.494889208666)*x_ref[0]**o+(0.721800653649)*x_ref[0]+(0.275748820606)*x_ref[1]**o+(0.190701062508)*x_ref[1]
        ref[0,0,1,1]=(0.317345288325)*x_ref[0]**o+(-0.0713887739659)*x_ref[0]+(-0.0901699882517)*x_ref[1]**o+(0.350861486991)*x_ref[1]
        ref[0,0,2,0]=(-0.0399664417282)*x_ref[0]**o+(-0.73007860251)*x_ref[0]+(-0.123301660907)*x_ref[1]**o+(0.935774110316)*x_ref[1]
        ref[0,0,2,1]=(-0.774484178038)*x_ref[0]**o+(-0.427643344733)*x_ref[0]+(0.00466934570081)*x_ref[1]**o+(0.644433150981)*x_ref[1]
        ref[0,1,0,0]=(0.532951038821)*x_ref[0]**o+(-0.267404179599)*x_ref[0]+(-0.488651484135)*x_ref[1]**o+(-0.354416641555)*x_ref[1]
        ref[0,1,0,1]=(0.816219473407)*x_ref[0]**o+(0.095041670769)*x_ref[0]+(0.130824542151)*x_ref[1]**o+(-0.91799466789)*x_ref[1]
        ref[0,1,1,0]=(-0.926659688215)*x_ref[0]**o+(-0.99257560771)*x_ref[0]+(0.76472846278)*x_ref[1]**o+(-0.77301872817)*x_ref[1]
        ref[0,1,1,1]=(-0.291848070507)*x_ref[0]**o+(0.988869135358)*x_ref[0]+(-0.355431590669)*x_ref[1]**o+(-0.397109808154)*x_ref[1]
        ref[0,1,2,0]=(-0.630380994644)*x_ref[0]**o+(0.121188015244)*x_ref[0]+(-0.527521198931)*x_ref[1]**o+(0.743015163843)*x_ref[1]
        ref[0,1,2,1]=(0.624350853711)*x_ref[0]**o+(-0.269065524076)*x_ref[0]+(0.381181149379)*x_ref[1]**o+(0.964176747301)*x_ref[1]
        ref[0,2,0,0]=(0.386078831223)*x_ref[0]**o+(-0.143893625755)*x_ref[0]+(-0.957649154953)*x_ref[1]**o+(-0.728625629809)*x_ref[1]
        ref[0,2,0,1]=(-0.752542988554)*x_ref[0]**o+(-0.334338454849)*x_ref[0]+(-0.888583790871)*x_ref[1]**o+(-0.49515709924)*x_ref[1]
        ref[0,2,1,0]=(0.67213418036)*x_ref[0]**o+(0.63778853358)*x_ref[0]+(-0.305139731169)*x_ref[1]**o+(-0.349307973811)*x_ref[1]
        ref[0,2,1,1]=(0.226718551492)*x_ref[0]**o+(0.0927018122394)*x_ref[0]+(-0.742387419573)*x_ref[1]**o+(0.528087866126)*x_ref[1]
        ref[0,2,2,0]=(-0.724349079038)*x_ref[0]**o+(-0.758167121924)*x_ref[0]+(-0.687742086176)*x_ref[1]**o+(-0.415119729664)*x_ref[1]
        ref[0,2,2,1]=(0.0399247837941)*x_ref[0]**o+(-0.728504254588)*x_ref[0]+(0.382011768693)*x_ref[1]**o+(-0.743726886782)*x_ref[1]
        ref[0,3,0,0]=(-0.994002993417)*x_ref[0]**o+(0.293311671699)*x_ref[0]+(-0.9731492726)*x_ref[1]**o+(0.0673759518597)*x_ref[1]
        ref[0,3,0,1]=(-0.638426251977)*x_ref[0]**o+(-0.537134730931)*x_ref[0]+(0.993741859056)*x_ref[1]**o+(-0.909049905282)*x_ref[1]
        ref[0,3,1,0]=(-0.994595072064)*x_ref[0]**o+(0.0252993150106)*x_ref[0]+(-0.739129165128)*x_ref[1]**o+(-0.132376880599)*x_ref[1]
        ref[0,3,1,1]=(-0.539219874104)*x_ref[0]**o+(0.796849273908)*x_ref[0]+(0.682514446887)*x_ref[1]**o+(-0.704901251482)*x_ref[1]
        ref[0,3,2,0]=(0.0401802421987)*x_ref[0]**o+(-0.0354713202277)*x_ref[0]+(-0.660064980444)*x_ref[1]**o+(0.0477264910037)*x_ref[1]
        ref[0,3,2,1]=(0.832824897985)*x_ref[0]**o+(0.877654166375)*x_ref[0]+(-0.909219867154)*x_ref[1]**o+(-0.796258795798)*x_ref[1]
        ref[0,4,0,0]=(0.0310731814412)*x_ref[0]**o+(0.39714436047)*x_ref[0]+(0.61669106719)*x_ref[1]**o+(0.789287424452)*x_ref[1]
        ref[0,4,0,1]=(-0.8919943646)*x_ref[0]**o+(0.55041132884)*x_ref[0]+(-0.0397475694211)*x_ref[1]**o+(0.993245641481)*x_ref[1]
        ref[0,4,1,0]=(0.550134389567)*x_ref[0]**o+(-0.657732937381)*x_ref[0]+(0.204770242863)*x_ref[1]**o+(-0.0296391840031)*x_ref[1]
        ref[0,4,1,1]=(0.753185285169)*x_ref[0]**o+(0.74420982803)*x_ref[0]+(0.108834676972)*x_ref[1]**o+(0.183891847557)*x_ref[1]
        ref[0,4,2,0]=(-0.259653246466)*x_ref[0]**o+(-0.120308818888)*x_ref[0]+(-0.827119024697)*x_ref[1]**o+(-0.672357922311)*x_ref[1]
        ref[0,4,2,1]=(-0.807666107771)*x_ref[0]**o+(0.873748263895)*x_ref[0]+(-0.160494957388)*x_ref[1]**o+(0.7551018226)*x_ref[1]
        ref[1,0,0,0]=(0.780978036182)*x_ref[0]**o+(-0.260409851047)*x_ref[0]+(-0.597455360543)*x_ref[1]**o+(0.0689482125215)*x_ref[1]
        ref[1,0,0,1]=(0.811121169417)*x_ref[0]**o+(-0.614791318516)*x_ref[0]+(0.663679164632)*x_ref[1]**o+(-0.853931508608)*x_ref[1]
        ref[1,0,1,0]=(-0.601473156596)*x_ref[0]**o+(-0.1232863608)*x_ref[0]+(0.200905517492)*x_ref[1]**o+(0.337784236126)*x_ref[1]
        ref[1,0,1,1]=(-0.924015099396)*x_ref[0]**o+(0.197096316191)*x_ref[0]+(-0.580606258161)*x_ref[1]**o+(0.517410299916)*x_ref[1]
        ref[1,0,2,0]=(0.32384939376)*x_ref[0]**o+(0.871622229492)*x_ref[0]+(0.403013167399)*x_ref[1]**o+(-0.50285362219)*x_ref[1]
        ref[1,0,2,1]=(-0.875266612786)*x_ref[0]**o+(0.773236652169)*x_ref[0]+(0.93949088522)*x_ref[1]**o+(-0.607791659163)*x_ref[1]
        ref[1,1,0,0]=(-0.972124688294)*x_ref[0]**o+(0.895545224323)*x_ref[0]+(-0.904953764575)*x_ref[1]**o+(0.388629839442)*x_ref[1]
        ref[1,1,0,1]=(0.792386948636)*x_ref[0]**o+(-0.467370445542)*x_ref[0]+(0.16664753332)*x_ref[1]**o+(-0.202002832911)*x_ref[1]
        ref[1,1,1,0]=(0.748979883957)*x_ref[0]**o+(-0.272594191544)*x_ref[0]+(-0.724533609431)*x_ref[1]**o+(0.0670080576995)*x_ref[1]
        ref[1,1,1,1]=(0.155216275716)*x_ref[0]**o+(-0.394137469825)*x_ref[0]+(-0.527098097796)*x_ref[1]**o+(0.504014563188)*x_ref[1]
        ref[1,1,2,0]=(-0.806704090439)*x_ref[0]**o+(-0.791435275759)*x_ref[0]+(0.866962482)*x_ref[1]**o+(0.60302217776)*x_ref[1]
        ref[1,1,2,1]=(-0.728053309956)*x_ref[0]**o+(0.0559583022356)*x_ref[0]+(0.0926294355801)*x_ref[1]**o+(-0.835984838669)*x_ref[1]
        ref[1,2,0,0]=(-0.0661747052735)*x_ref[0]**o+(0.0537098973346)*x_ref[0]+(0.310702398279)*x_ref[1]**o+(-0.489393478975)*x_ref[1]
        ref[1,2,0,1]=(-0.0910654471721)*x_ref[0]**o+(0.327801831182)*x_ref[0]+(-0.42833439187)*x_ref[1]**o+(0.267067714743)*x_ref[1]
        ref[1,2,1,0]=(-0.693916583922)*x_ref[0]**o+(-0.100272651463)*x_ref[0]+(-0.195971225883)*x_ref[1]**o+(0.660851591786)*x_ref[1]
        ref[1,2,1,1]=(-0.158662866194)*x_ref[0]**o+(0.756412009601)*x_ref[0]+(-0.923974707134)*x_ref[1]**o+(0.61664700863)*x_ref[1]
        ref[1,2,2,0]=(-0.449981632451)*x_ref[0]**o+(0.272083805383)*x_ref[0]+(-0.207156704608)*x_ref[1]**o+(0.413596665304)*x_ref[1]
        ref[1,2,2,1]=(0.67039491127)*x_ref[0]**o+(-0.606081942118)*x_ref[0]+(-0.358862791857)*x_ref[1]**o+(-0.191412169743)*x_ref[1]
        ref[1,3,0,0]=(0.214562206436)*x_ref[0]**o+(0.475358432768)*x_ref[0]+(0.639237285541)*x_ref[1]**o+(-0.221799470417)*x_ref[1]
        ref[1,3,0,1]=(0.468332920849)*x_ref[0]**o+(0.359173095186)*x_ref[0]+(0.513170435198)*x_ref[1]**o+(0.472621770345)*x_ref[1]
        ref[1,3,1,0]=(-0.722500106774)*x_ref[0]**o+(0.781087784307)*x_ref[0]+(0.633141730145)*x_ref[1]**o+(0.958777327006)*x_ref[1]
        ref[1,3,1,1]=(0.488269631028)*x_ref[0]**o+(0.148575659631)*x_ref[0]+(-0.229858507326)*x_ref[1]**o+(-0.857118507195)*x_ref[1]
        ref[1,3,2,0]=(0.99463505025)*x_ref[0]**o+(-0.970275234551)*x_ref[0]+(-0.431529655117)*x_ref[1]**o+(-0.208112330684)*x_ref[1]
        ref[1,3,2,1]=(-0.715411408332)*x_ref[0]**o+(0.554401881121)*x_ref[0]+(-0.980883542633)*x_ref[1]**o+(0.316356494975)*x_ref[1]
        ref[1,4,0,0]=(0.747620638111)*x_ref[0]**o+(-0.505093883319)*x_ref[0]+(-0.217289877652)*x_ref[1]**o+(-0.122961101484)*x_ref[1]
        ref[1,4,0,1]=(-0.329916740819)*x_ref[0]**o+(0.778840575985)*x_ref[0]+(-0.696382782242)*x_ref[1]**o+(-0.096185679235)*x_ref[1]
        ref[1,4,1,0]=(0.574218122656)*x_ref[0]**o+(0.0949666147475)*x_ref[0]+(-0.789015428301)*x_ref[1]**o+(-0.400376747438)*x_ref[1]
        ref[1,4,1,1]=(0.569036824328)*x_ref[0]**o+(0.132679527923)*x_ref[0]+(-0.229208784312)*x_ref[1]**o+(0.405134795549)*x_ref[1]
        ref[1,4,2,0]=(0.0123943827354)*x_ref[0]**o+(-0.0446024014977)*x_ref[0]+(-0.221944819571)*x_ref[1]**o+(-0.596890233794)*x_ref[1]
        ref[1,4,2,1]=(0.499247385143)*x_ref[0]**o+(0.891843495388)*x_ref[0]+(-0.348380026198)*x_ref[1]**o+(0.146344723761)*x_ref[1]
        ref[2,0,0,0]=(-0.929032869237)*x_ref[0]**o+(-0.438035650633)*x_ref[0]+(0.117030848904)*x_ref[1]**o+(-0.489094945626)*x_ref[1]
        ref[2,0,0,1]=(0.661960656348)*x_ref[0]**o+(0.437346307001)*x_ref[0]+(0.15713464744)*x_ref[1]**o+(0.578309634731)*x_ref[1]
        ref[2,0,1,0]=(-0.368410149772)*x_ref[0]**o+(-0.135484081321)*x_ref[0]+(-0.51575875197)*x_ref[1]**o+(0.26997748887)*x_ref[1]
        ref[2,0,1,1]=(0.784159013964)*x_ref[0]**o+(-0.558040177257)*x_ref[0]+(-0.100963880851)*x_ref[1]**o+(-0.319503517522)*x_ref[1]
        ref[2,0,2,0]=(-0.0260076717983)*x_ref[0]**o+(-0.252140406297)*x_ref[0]+(-0.335458650519)*x_ref[1]**o+(-0.0739605444814)*x_ref[1]
        ref[2,0,2,1]=(0.484818876559)*x_ref[0]**o+(-0.59157296366)*x_ref[0]+(-0.83491407929)*x_ref[1]**o+(0.143217129439)*x_ref[1]
        ref[2,1,0,0]=(0.672216209481)*x_ref[0]**o+(0.225722770162)*x_ref[0]+(-0.46608814621)*x_ref[1]**o+(-0.779035303765)*x_ref[1]
        ref[2,1,0,1]=(-0.738780643364)*x_ref[0]**o+(0.0741487570855)*x_ref[0]+(-0.362480267529)*x_ref[1]**o+(0.474314990314)*x_ref[1]
        ref[2,1,1,0]=(0.942380641613)*x_ref[0]**o+(0.834468091419)*x_ref[0]+(-0.413376836135)*x_ref[1]**o+(0.996367350352)*x_ref[1]
        ref[2,1,1,1]=(0.544360198404)*x_ref[0]**o+(-0.327857763859)*x_ref[0]+(-0.00306338907092)*x_ref[1]**o+(0.448394961269)*x_ref[1]
        ref[2,1,2,0]=(-0.345632293594)*x_ref[0]**o+(0.505451030738)*x_ref[0]+(-0.644095213138)*x_ref[1]**o+(0.603588130789)*x_ref[1]
        ref[2,1,2,1]=(-0.300809177049)*x_ref[0]**o+(-0.531168939571)*x_ref[0]+(0.513243419239)*x_ref[1]**o+(0.62918148667)*x_ref[1]
        ref[2,2,0,0]=(0.610906338615)*x_ref[0]**o+(-0.656455478681)*x_ref[0]+(-0.386484840223)*x_ref[1]**o+(0.720074853711)*x_ref[1]
        ref[2,2,0,1]=(-0.403600090425)*x_ref[0]**o+(0.454789490943)*x_ref[0]+(-0.399073455918)*x_ref[1]**o+(-0.0552326255369)*x_ref[1]
        ref[2,2,1,0]=(-0.308592576779)*x_ref[0]**o+(0.253298307559)*x_ref[0]+(-0.862040156595)*x_ref[1]**o+(0.195088197599)*x_ref[1]
        ref[2,2,1,1]=(0.143353860538)*x_ref[0]**o+(-0.590315683091)*x_ref[0]+(-0.955465510619)*x_ref[1]**o+(0.372352027984)*x_ref[1]
        ref[2,2,2,0]=(-0.13314652456)*x_ref[0]**o+(0.580744198064)*x_ref[0]+(-0.903035808104)*x_ref[1]**o+(-0.193570555816)*x_ref[1]
        ref[2,2,2,1]=(-0.647218355992)*x_ref[0]**o+(0.00521075379725)*x_ref[0]+(0.877022602564)*x_ref[1]**o+(-0.104569127278)*x_ref[1]
        ref[2,3,0,0]=(-0.532184390508)*x_ref[0]**o+(0.739226859366)*x_ref[0]+(0.595122128361)*x_ref[1]**o+(0.0135063212516)*x_ref[1]
        ref[2,3,0,1]=(-0.121073574394)*x_ref[0]**o+(-0.175990877109)*x_ref[0]+(-0.00319744367579)*x_ref[1]**o+(0.140264888668)*x_ref[1]
        ref[2,3,1,0]=(-0.0778296806357)*x_ref[0]**o+(0.760344573573)*x_ref[0]+(-0.250209333031)*x_ref[1]**o+(-0.413073766011)*x_ref[1]
        ref[2,3,1,1]=(-0.809655592364)*x_ref[0]**o+(-0.147000481704)*x_ref[0]+(-0.760732613787)*x_ref[1]**o+(0.399494900513)*x_ref[1]
        ref[2,3,2,0]=(-0.131633255327)*x_ref[0]**o+(-0.0268295912955)*x_ref[0]+(-0.934414490057)*x_ref[1]**o+(-0.339746795066)*x_ref[1]
        ref[2,3,2,1]=(0.503687469605)*x_ref[0]**o+(-0.415050204706)*x_ref[0]+(0.152421197486)*x_ref[1]**o+(0.529931932615)*x_ref[1]
        ref[2,4,0,0]=(0.930580833077)*x_ref[0]**o+(-0.377748298816)*x_ref[0]+(-0.159931528324)*x_ref[1]**o+(0.374143651635)*x_ref[1]
        ref[2,4,0,1]=(0.254339441506)*x_ref[0]**o+(0.695672173367)*x_ref[0]+(-0.499989877641)*x_ref[1]**o+(-0.745148857716)*x_ref[1]
        ref[2,4,1,0]=(0.205325187322)*x_ref[0]**o+(0.317887475767)*x_ref[0]+(0.755338840391)*x_ref[1]**o+(-0.106253370416)*x_ref[1]
        ref[2,4,1,1]=(0.175711808412)*x_ref[0]**o+(-0.437006733188)*x_ref[0]+(-0.449737979076)*x_ref[1]**o+(-0.0421697213234)*x_ref[1]
        ref[2,4,2,0]=(-0.543184540283)*x_ref[0]**o+(0.616024774255)*x_ref[0]+(-0.429403317932)*x_ref[1]**o+(-0.557076956301)*x_ref[1]
        ref[2,4,2,1]=(0.224704163386)*x_ref[0]**o+(-0.945537746026)*x_ref[0]+(0.947303520688)*x_ref[1]**o+(0.909695468251)*x_ref[1]
        ref[3,0,0,0]=(-0.671167109353)*x_ref[0]**o+(-0.513526771737)*x_ref[0]+(-0.511672331777)*x_ref[1]**o+(-0.301926771817)*x_ref[1]
        ref[3,0,0,1]=(0.39343937992)*x_ref[0]**o+(-0.194833946828)*x_ref[0]+(0.882702377998)*x_ref[1]**o+(0.191544038874)*x_ref[1]
        ref[3,0,1,0]=(0.521271797889)*x_ref[0]**o+(0.924897421226)*x_ref[0]+(-0.597268916449)*x_ref[1]**o+(-0.897621830518)*x_ref[1]
        ref[3,0,1,1]=(-0.446311941735)*x_ref[0]**o+(-0.223997368297)*x_ref[0]+(-0.348064954666)*x_ref[1]**o+(0.661704488202)*x_ref[1]
        ref[3,0,2,0]=(-0.785706283015)*x_ref[0]**o+(0.434487336824)*x_ref[0]+(-0.0447968030358)*x_ref[1]**o+(0.0231287166506)*x_ref[1]
        ref[3,0,2,1]=(0.347155629717)*x_ref[0]**o+(-0.836660082409)*x_ref[0]+(0.391492038747)*x_ref[1]**o+(0.883578619998)*x_ref[1]
        ref[3,1,0,0]=(0.571669585809)*x_ref[0]**o+(0.966740526379)*x_ref[0]+(-0.657144215536)*x_ref[1]**o+(0.357025591929)*x_ref[1]
        ref[3,1,0,1]=(-0.269900462792)*x_ref[0]**o+(0.770680682031)*x_ref[0]+(0.744696440247)*x_ref[1]**o+(0.755227887237)*x_ref[1]
        ref[3,1,1,0]=(-0.667568521409)*x_ref[0]**o+(-0.483641971546)*x_ref[0]+(0.687266187401)*x_ref[1]**o+(-0.597241584922)*x_ref[1]
        ref[3,1,1,1]=(-0.947316636771)*x_ref[0]**o+(0.0797873425176)*x_ref[0]+(0.254921442278)*x_ref[1]**o+(-0.0500536165406)*x_ref[1]
        ref[3,1,2,0]=(0.966192971983)*x_ref[0]**o+(-0.597929938773)*x_ref[0]+(-0.174360538)*x_ref[1]**o+(0.46405478391)*x_ref[1]
        ref[3,1,2,1]=(-0.675528960994)*x_ref[0]**o+(0.51539766317)*x_ref[0]+(-0.99673299807)*x_ref[1]**o+(-0.488746102106)*x_ref[1]
        ref[3,2,0,0]=(-0.604138916979)*x_ref[0]**o+(-0.0431165687312)*x_ref[0]+(-0.346697341122)*x_ref[1]**o+(-0.966811623167)*x_ref[1]
        ref[3,2,0,1]=(0.963965750281)*x_ref[0]**o+(-0.059166688536)*x_ref[0]+(0.166515828506)*x_ref[1]**o+(0.832637512663)*x_ref[1]
        ref[3,2,1,0]=(-0.808155600589)*x_ref[0]**o+(-0.178526483687)*x_ref[0]+(-0.012382108719)*x_ref[1]**o+(0.194810458281)*x_ref[1]
        ref[3,2,1,1]=(0.905922872414)*x_ref[0]**o+(0.580755461959)*x_ref[0]+(0.84253193847)*x_ref[1]**o+(0.300676716124)*x_ref[1]
        ref[3,2,2,0]=(-0.560928128044)*x_ref[0]**o+(0.215498997249)*x_ref[0]+(0.630359499164)*x_ref[1]**o+(0.0955049343157)*x_ref[1]
        ref[3,2,2,1]=(-0.977307154392)*x_ref[0]**o+(0.543314643158)*x_ref[0]+(-0.617962918616)*x_ref[1]**o+(0.222318165123)*x_ref[1]
        ref[3,3,0,0]=(0.910351367774)*x_ref[0]**o+(-0.349015106312)*x_ref[0]+(-0.367900620181)*x_ref[1]**o+(-0.371131872755)*x_ref[1]
        ref[3,3,0,1]=(0.663510744888)*x_ref[0]**o+(-0.16380815063)*x_ref[0]+(0.770895156906)*x_ref[1]**o+(0.84395768389)*x_ref[1]
        ref[3,3,1,0]=(0.553097521137)*x_ref[0]**o+(-0.139012828014)*x_ref[0]+(-0.986291992888)*x_ref[1]**o+(-0.780024041157)*x_ref[1]
        ref[3,3,1,1]=(-0.894441791323)*x_ref[0]**o+(0.91199469535)*x_ref[0]+(-0.349154843905)*x_ref[1]**o+(0.538946349908)*x_ref[1]
        ref[3,3,2,0]=(-0.464541735353)*x_ref[0]**o+(-0.44894927329)*x_ref[0]+(-0.488621463959)*x_ref[1]**o+(-0.987756549629)*x_ref[1]
        ref[3,3,2,1]=(0.136069699682)*x_ref[0]**o+(-0.794904844715)*x_ref[0]+(0.300038435268)*x_ref[1]**o+(-0.599562792523)*x_ref[1]
        ref[3,4,0,0]=(0.391227355996)*x_ref[0]**o+(-0.634937772688)*x_ref[0]+(0.0618326566602)*x_ref[1]**o+(0.396379987618)*x_ref[1]
        ref[3,4,0,1]=(0.485984857628)*x_ref[0]**o+(-0.884694203334)*x_ref[0]+(-0.791392877516)*x_ref[1]**o+(0.832758707832)*x_ref[1]
        ref[3,4,1,0]=(0.983115009543)*x_ref[0]**o+(0.830522788331)*x_ref[0]+(-0.947942131374)*x_ref[1]**o+(0.890681434885)*x_ref[1]
        ref[3,4,1,1]=(-0.0596041988514)*x_ref[0]**o+(0.203878482428)*x_ref[0]+(-0.739288417295)*x_ref[1]**o+(0.150351531263)*x_ref[1]
        ref[3,4,2,0]=(-0.821412631737)*x_ref[0]**o+(0.366097632474)*x_ref[0]+(-0.978999319755)*x_ref[1]**o+(-0.556584079904)*x_ref[1]
        ref[3,4,2,1]=(0.677346101991)*x_ref[0]**o+(0.617303832579)*x_ref[0]+(0.963322024959)*x_ref[1]**o+(0.0765656809305)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.412268765993)*x[0]**o+(-0.31214585911)*x[0]+(0.153520487818)*x[1]**o+(-0.3104168266)*x[1]+(0.909798069186)*x[2]**o+(0.742916044571)*x[2]
        arg[0,0,0,1]=(0.0735781063807)*x[0]**o+(0.725527279895)*x[0]+(-0.884760504783)*x[1]**o+(0.751678022946)*x[1]+(-0.169364499242)*x[2]**o+(-0.946699946704)*x[2]
        arg[0,0,1,0]=(-0.308928255723)*x[0]**o+(-0.601764425359)*x[0]+(-0.176466074749)*x[1]**o+(-0.14838705699)*x[1]+(0.897444985593)*x[2]**o+(-0.0552696055317)*x[2]
        arg[0,0,1,1]=(0.278269305614)*x[0]**o+(-0.419731200355)*x[0]+(0.450955076279)*x[1]**o+(-0.899203207333)*x[1]+(-0.458812838787)*x[2]**o+(0.507871397895)*x[2]
        arg[0,0,2,0]=(0.887873653495)*x[0]**o+(0.130110213077)*x[0]+(0.203761425496)*x[1]**o+(-0.0563898018944)*x[1]+(0.0125548052261)*x[2]**o+(-0.594578723877)*x[2]
        arg[0,0,2,1]=(-0.104210554142)*x[0]**o+(-0.792596213906)*x[0]+(0.570869430862)*x[1]**o+(-0.798160727294)*x[1]+(0.241269856254)*x[2]**o+(0.194885978057)*x[2]
        arg[0,1,0,0]=(0.819550431682)*x[0]**o+(-0.239970258367)*x[0]+(-0.960362264523)*x[1]**o+(0.872093886273)*x[1]+(-0.857325000718)*x[2]**o+(-0.907620072404)*x[2]
        arg[0,1,0,1]=(-0.989011816788)*x[0]**o+(0.853577413178)*x[0]+(-0.160335863166)*x[1]**o+(-0.751266166108)*x[1]+(0.631146799273)*x[2]**o+(0.739741622363)*x[2]
        arg[0,1,1,0]=(0.162828987854)*x[0]**o+(-0.200652438446)*x[0]+(-0.890837153528)*x[1]**o+(-0.972029967593)*x[1]+(0.850395155009)*x[2]**o+(-0.0868272213807)*x[2]
        arg[0,1,1,1]=(-0.770711232994)*x[0]**o+(0.986518762896)*x[0]+(0.0970713645745)*x[1]**o+(0.996843418865)*x[1]+(0.208134397469)*x[2]**o+(0.946600905711)*x[2]
        arg[0,1,2,0]=(-0.543213476288)*x[0]**o+(0.0335009416664)*x[0]+(0.165751788147)*x[1]**o+(0.660276877927)*x[1]+(-0.485925597349)*x[2]**o+(-0.632302900376)*x[2]
        arg[0,1,2,1]=(0.175798920586)*x[0]**o+(-0.42873239224)*x[0]+(0.164896129395)*x[1]**o+(0.662142729145)*x[1]+(-0.425944275652)*x[2]**o+(0.306978282183)*x[2]
        arg[0,2,0,0]=(0.790303828628)*x[0]**o+(0.169346034559)*x[0]+(-0.901026918205)*x[1]**o+(0.661856377517)*x[1]+(0.194850846261)*x[2]**o+(0.155282302183)*x[2]
        arg[0,2,0,1]=(-0.43947970094)*x[0]**o+(-0.535569903832)*x[0]+(-0.534035397281)*x[1]**o+(-0.169724793112)*x[1]+(-0.717226228021)*x[2]**o+(0.771292292609)*x[2]
        arg[0,2,1,0]=(-0.424991151717)*x[0]**o+(-0.0837800409811)*x[0]+(-0.747139428246)*x[1]**o+(0.739025612454)*x[1]+(0.464486611081)*x[2]**o+(0.824926522179)*x[2]
        arg[0,2,1,1]=(-0.4666399509)*x[0]**o+(-0.479130555457)*x[0]+(-0.676718441278)*x[1]**o+(0.352685580535)*x[1]+(0.537789487962)*x[2]**o+(-0.774383681009)*x[2]
        arg[0,2,2,0]=(-0.30983510311)*x[0]**o+(0.00141603592636)*x[0]+(0.546192108985)*x[1]**o+(0.229628329192)*x[1]+(0.788466929747)*x[2]**o+(-0.169587087096)*x[2]
        arg[0,2,2,1]=(0.810282347322)*x[0]**o+(0.961056346191)*x[0]+(0.204622339898)*x[1]**o+(0.288786527052)*x[1]+(0.87096784221)*x[2]**o+(0.967565754848)*x[2]
        arg[0,3,0,0]=(-0.154669358868)*x[0]**o+(-0.449349763143)*x[0]+(-0.692997905123)*x[1]**o+(0.666308436198)*x[1]+(0.0638963436229)*x[2]**o+(-0.318233405188)*x[2]
        arg[0,3,0,1]=(0.951324381413)*x[0]**o+(0.10400502165)*x[0]+(-0.910416795399)*x[1]**o+(-0.931501370526)*x[1]+(0.455893141186)*x[2]**o+(0.536556088937)*x[2]
        arg[0,3,1,0]=(-0.315779909544)*x[0]**o+(0.189718508813)*x[0]+(0.0995068185023)*x[1]**o+(-0.863775500494)*x[1]+(0.99696081153)*x[2]**o+(0.0941784151393)*x[2]
        arg[0,3,1,1]=(0.122766381329)*x[0]**o+(-0.953220464103)*x[0]+(-0.853780873931)*x[1]**o+(-0.719623724002)*x[1]+(-0.58783153111)*x[2]**o+(-0.919075056141)*x[2]
        arg[0,3,2,0]=(-0.740349380942)*x[0]**o+(-0.0161629946027)*x[0]+(0.790593892223)*x[1]**o+(0.81563688519)*x[1]+(0.515002236312)*x[2]**o+(0.906399338047)*x[2]
        arg[0,3,2,1]=(-0.21876846821)*x[0]**o+(0.099226679818)*x[0]+(-0.501220667305)*x[1]**o+(-0.645539150063)*x[1]+(0.705501523136)*x[2]**o+(-0.640426266054)*x[2]
        arg[0,4,0,0]=(0.382430876525)*x[0]**o+(0.728752519135)*x[0]+(-0.696198853186)*x[1]**o+(0.181187001577)*x[1]+(0.818931936956)*x[2]**o+(0.267607197602)*x[2]
        arg[0,4,0,1]=(0.667658893159)*x[0]**o+(0.129176508255)*x[0]+(0.579266506883)*x[1]**o+(-0.27290178784)*x[1]+(0.804307104985)*x[2]**o+(-0.384179099369)*x[2]
        arg[0,4,1,0]=(0.472611036284)*x[0]**o+(0.532026319531)*x[0]+(-0.11900311657)*x[1]**o+(0.818635083182)*x[1]+(-0.127618053257)*x[2]**o+(-0.8464024968)*x[2]
        arg[0,4,1,1]=(-0.52572498077)*x[0]**o+(-0.222527963121)*x[0]+(-0.733344395469)*x[1]**o+(-0.519230573211)*x[1]+(0.550298161034)*x[2]**o+(-0.091253041563)*x[2]
        arg[0,4,2,0]=(-0.995556194807)*x[0]**o+(-0.157265619525)*x[0]+(0.646987111139)*x[1]**o+(0.114269194694)*x[1]+(0.787559342851)*x[2]**o+(0.655874189666)*x[2]
        arg[0,4,2,1]=(-0.534706379451)*x[0]**o+(0.89989539183)*x[0]+(0.584227019406)*x[1]**o+(-0.261650143902)*x[1]+(-0.0741247974759)*x[2]**o+(-0.646603374886)*x[2]
        arg[1,0,0,0]=(-0.16346294005)*x[0]**o+(0.515625993676)*x[0]+(-0.0294710658489)*x[1]**o+(-0.886489878353)*x[1]+(0.184376222466)*x[2]**o+(0.872455440027)*x[2]
        arg[1,0,0,1]=(-0.239569565644)*x[0]**o+(-0.0196989227301)*x[0]+(-0.333999804595)*x[1]**o+(0.79703425655)*x[1]+(0.332925813867)*x[2]**o+(0.384766118619)*x[2]
        arg[1,0,1,0]=(0.337775521215)*x[0]**o+(-0.214105347244)*x[0]+(-0.055476365922)*x[1]**o+(-0.511469291523)*x[1]+(0.390455425321)*x[2]**o+(0.346749928211)*x[2]
        arg[1,0,1,1]=(0.255034684548)*x[0]**o+(-0.975969854422)*x[0]+(-0.588154151576)*x[1]**o+(0.981825996023)*x[1]+(-0.16323013592)*x[2]**o+(-0.165779157778)*x[2]
        arg[1,0,2,0]=(-0.421647673573)*x[0]**o+(0.0892056009306)*x[0]+(-0.659566191892)*x[1]**o+(-0.640769435519)*x[1]+(-0.5474352085)*x[2]**o+(0.619855556292)*x[2]
        arg[1,0,2,1]=(0.678408254014)*x[0]**o+(0.114013125353)*x[0]+(0.222310135576)*x[1]**o+(-0.297909583371)*x[1]+(0.885716237809)*x[2]**o+(0.151576219626)*x[2]
        arg[1,1,0,0]=(-0.724825546043)*x[0]**o+(-0.796280833482)*x[0]+(0.894034526494)*x[1]**o+(-0.496334573913)*x[1]+(0.161130154754)*x[2]**o+(0.0137999166704)*x[2]
        arg[1,1,0,1]=(-0.959049520871)*x[0]**o+(0.0727868983195)*x[0]+(-0.0475524824482)*x[1]**o+(0.8098016511)*x[1]+(0.154268238454)*x[2]**o+(-0.0583213692933)*x[2]
        arg[1,1,1,0]=(-0.684320939672)*x[0]**o+(-0.841165491324)*x[0]+(0.307147508965)*x[1]**o+(0.0204906256209)*x[1]+(0.658432728833)*x[2]**o+(0.398570233826)*x[2]
        arg[1,1,1,1]=(0.921894102193)*x[0]**o+(0.33017900913)*x[0]+(0.566856960125)*x[1]**o+(0.826695387569)*x[1]+(-0.6253390282)*x[2]**o+(-0.527610353599)*x[2]
        arg[1,1,2,0]=(0.0640251272456)*x[0]**o+(0.789847887603)*x[0]+(-0.00457885872972)*x[1]**o+(-0.24681292381)*x[1]+(-0.0832674042894)*x[2]**o+(-0.254280656434)*x[2]
        arg[1,1,2,1]=(0.786208197486)*x[0]**o+(0.519575725184)*x[0]+(-0.452657209679)*x[1]**o+(0.36570806994)*x[1]+(0.875049922923)*x[2]**o+(0.881026792786)*x[2]
        arg[1,2,0,0]=(-0.415919514711)*x[0]**o+(-0.933890170641)*x[0]+(0.717990882823)*x[1]**o+(0.355555237378)*x[1]+(0.458971778535)*x[2]**o+(0.888647676837)*x[2]
        arg[1,2,0,1]=(-0.453925437723)*x[0]**o+(0.153501471971)*x[0]+(0.815830410181)*x[1]**o+(-0.519500430324)*x[1]+(-0.791530261324)*x[2]**o+(0.849179201427)*x[2]
        arg[1,2,1,0]=(0.299736697301)*x[0]**o+(0.645276033981)*x[0]+(-0.731636401343)*x[1]**o+(0.127189134157)*x[1]+(0.891427235216)*x[2]**o+(0.750052326597)*x[2]
        arg[1,2,1,1]=(-0.987392548561)*x[0]**o+(0.129053618821)*x[0]+(0.220541501123)*x[1]**o+(0.276873448998)*x[1]+(0.645231056419)*x[2]**o+(0.776634462468)*x[2]
        arg[1,2,2,0]=(-0.879738240565)*x[0]**o+(-0.0352288276658)*x[0]+(-0.247550105519)*x[1]**o+(0.782133317388)*x[1]+(-0.948666864992)*x[2]**o+(0.310357478071)*x[2]
        arg[1,2,2,1]=(-0.606778611022)*x[0]**o+(0.964356172073)*x[0]+(-0.520291905065)*x[1]**o+(-0.219411824153)*x[1]+(-0.405356968701)*x[2]**o+(0.969492630653)*x[2]
        arg[1,3,0,0]=(-0.991624055852)*x[0]**o+(-0.762240034313)*x[0]+(0.423262307551)*x[1]**o+(-0.05789904696)*x[1]+(-0.696181866276)*x[2]**o+(-0.327406198937)*x[2]
        arg[1,3,0,1]=(-0.614181964006)*x[0]**o+(0.927675019363)*x[0]+(-0.750120185372)*x[1]**o+(0.440548313639)*x[1]+(-0.116776982636)*x[2]**o+(-0.799181582827)*x[2]
        arg[1,3,1,0]=(0.286032474963)*x[0]**o+(0.391407042942)*x[0]+(-0.268363949002)*x[1]**o+(-0.421319501252)*x[1]+(-0.318595882607)*x[2]**o+(0.273120761382)*x[2]
        arg[1,3,1,1]=(0.918425098597)*x[0]**o+(-0.211260796786)*x[0]+(-0.793780530123)*x[1]**o+(-0.839335876894)*x[1]+(-0.595786792228)*x[2]**o+(-0.592690829684)*x[2]
        arg[1,3,2,0]=(-0.959108168319)*x[0]**o+(-0.94492160223)*x[0]+(0.0778030192626)*x[1]**o+(-0.169988420265)*x[1]+(0.042653831433)*x[2]**o+(-0.595426371148)*x[2]
        arg[1,3,2,1]=(0.897478414413)*x[0]**o+(-0.222864711011)*x[0]+(0.900965727203)*x[1]**o+(0.501243022513)*x[1]+(0.75496626209)*x[2]**o+(0.312225549338)*x[2]
        arg[1,4,0,0]=(0.168321984115)*x[0]**o+(-0.495995847038)*x[0]+(0.595773350289)*x[1]**o+(-0.914906345927)*x[1]+(0.224831726299)*x[2]**o+(-0.0743642212453)*x[2]
        arg[1,4,0,1]=(0.352717252696)*x[0]**o+(-0.873077111264)*x[0]+(-0.296366541486)*x[1]**o+(-0.00338438754009)*x[1]+(0.0294923296914)*x[2]**o+(-0.067479335313)*x[2]
        arg[1,4,1,0]=(-0.518757232583)*x[0]**o+(0.463948954744)*x[0]+(0.712607983792)*x[1]**o+(0.717184663437)*x[1]+(0.323391383016)*x[2]**o+(-0.225868711026)*x[2]
        arg[1,4,1,1]=(-0.24381531137)*x[0]**o+(0.630251169076)*x[0]+(0.70462966259)*x[1]**o+(-0.125357204688)*x[1]+(-0.258633435955)*x[2]**o+(0.902171387163)*x[2]
        arg[1,4,2,0]=(-0.632179111447)*x[0]**o+(-0.59290738268)*x[0]+(-0.825795231706)*x[1]**o+(0.399660693305)*x[1]+(0.0620015248102)*x[2]**o+(-0.451556558781)*x[2]
        arg[1,4,2,1]=(-0.858145688916)*x[0]**o+(-0.787993091462)*x[0]+(0.803723603768)*x[1]**o+(-0.835043986384)*x[1]+(0.467577922972)*x[2]**o+(0.819139387663)*x[2]
        arg[2,0,0,0]=(0.274886550898)*x[0]**o+(-0.0742978734335)*x[0]+(-0.773015479116)*x[1]**o+(-0.138971598539)*x[1]+(-0.702175483988)*x[2]**o+(0.9147072714)*x[2]
        arg[2,0,0,1]=(-0.598714805644)*x[0]**o+(0.11878664701)*x[0]+(-0.554608391808)*x[1]**o+(-0.734151257491)*x[1]+(0.234629053469)*x[2]**o+(0.911072319646)*x[2]
        arg[2,0,1,0]=(-0.83850906403)*x[0]**o+(0.799369711531)*x[0]+(0.436211860763)*x[1]**o+(-0.855173818525)*x[1]+(0.256241482884)*x[2]**o+(0.657883842259)*x[2]
        arg[2,0,1,1]=(-0.724328143383)*x[0]**o+(-0.781110608395)*x[0]+(-0.987464957106)*x[1]**o+(0.936448332957)*x[1]+(0.711870454531)*x[2]**o+(0.64958966746)*x[2]
        arg[2,0,2,0]=(0.485490298364)*x[0]**o+(-0.648402414524)*x[0]+(0.0493575279485)*x[1]**o+(-0.168695660031)*x[1]+(0.236500977054)*x[2]**o+(-0.699035972752)*x[2]
        arg[2,0,2,1]=(-0.508561382238)*x[0]**o+(0.0587152225749)*x[0]+(-0.92072872998)*x[1]**o+(-0.210536816562)*x[1]+(0.735465423941)*x[2]**o+(0.557002016885)*x[2]
        arg[2,1,0,0]=(-0.301683530613)*x[0]**o+(-0.501004347082)*x[0]+(-0.415125116853)*x[1]**o+(-0.975209043974)*x[1]+(0.0952958379219)*x[2]**o+(-0.882009529251)*x[2]
        arg[2,1,0,1]=(0.569047454657)*x[0]**o+(0.87107862461)*x[0]+(0.977655371485)*x[1]**o+(-0.870823933133)*x[1]+(-0.12817645442)*x[2]**o+(-0.412450404053)*x[2]
        arg[2,1,1,0]=(0.532638832559)*x[0]**o+(-0.669778224526)*x[0]+(0.516969768896)*x[1]**o+(0.725495018612)*x[1]+(0.604820640277)*x[2]**o+(-0.537483520132)*x[2]
        arg[2,1,1,1]=(-0.413776169533)*x[0]**o+(0.19037961067)*x[0]+(-0.109551737415)*x[1]**o+(-0.195599236834)*x[1]+(-0.183237365211)*x[2]**o+(0.0279625528407)*x[2]
        arg[2,1,2,0]=(-0.46089173487)*x[0]**o+(-0.151158992117)*x[0]+(0.359508089079)*x[1]**o+(-0.0808350778984)*x[1]+(0.845464278946)*x[2]**o+(0.177891463715)*x[2]
        arg[2,1,2,1]=(0.416828200435)*x[0]**o+(0.763110094817)*x[0]+(0.855595717948)*x[1]**o+(0.379358815658)*x[1]+(0.954688280021)*x[2]**o+(-0.808098938468)*x[2]
        arg[2,2,0,0]=(-0.697342667428)*x[0]**o+(-0.855608259014)*x[0]+(0.866084171318)*x[1]**o+(0.632470977708)*x[1]+(-0.399703926338)*x[2]**o+(0.211173996927)*x[2]
        arg[2,2,0,1]=(-0.940742486357)*x[0]**o+(0.249267367828)*x[0]+(0.197336418576)*x[1]**o+(-0.0573784352507)*x[1]+(-0.328108062127)*x[2]**o+(0.570204614968)*x[2]
        arg[2,2,1,0]=(0.134960606335)*x[0]**o+(0.490123643194)*x[0]+(-0.102294210304)*x[1]**o+(0.427727810449)*x[1]+(-0.158572569775)*x[2]**o+(-0.165504764483)*x[2]
        arg[2,2,1,1]=(-0.427912549906)*x[0]**o+(-0.311524081823)*x[0]+(0.220211421031)*x[1]**o+(0.123679381711)*x[1]+(0.593149361283)*x[2]**o+(-0.283233437917)*x[2]
        arg[2,2,2,0]=(0.901179489164)*x[0]**o+(0.292780095159)*x[0]+(-0.295149620179)*x[1]**o+(0.860483795735)*x[1]+(0.848872993178)*x[2]**o+(-0.612494025994)*x[2]
        arg[2,2,2,1]=(0.904689835334)*x[0]**o+(0.33689603833)*x[0]+(0.0765595244364)*x[1]**o+(0.215174063849)*x[1]+(0.973122153293)*x[2]**o+(-0.49840689696)*x[2]
        arg[2,3,0,0]=(0.973713543725)*x[0]**o+(0.0727815483094)*x[0]+(-0.0723336244169)*x[1]**o+(-0.102547958784)*x[1]+(0.573980704603)*x[2]**o+(0.784756916348)*x[2]
        arg[2,3,0,1]=(0.871042616423)*x[0]**o+(-0.966021391258)*x[0]+(0.9402258232)*x[1]**o+(0.163522806825)*x[1]+(-0.260871238896)*x[2]**o+(0.68201099906)*x[2]
        arg[2,3,1,0]=(0.916349358422)*x[0]**o+(-0.923720463388)*x[0]+(0.814116732807)*x[1]**o+(-0.521079005717)*x[1]+(-0.996963157941)*x[2]**o+(-0.251248775427)*x[2]
        arg[2,3,1,1]=(0.752568485119)*x[0]**o+(-0.134517175046)*x[0]+(0.691477362575)*x[1]**o+(-0.864047765263)*x[1]+(0.0632525427291)*x[2]**o+(0.27617797274)*x[2]
        arg[2,3,2,0]=(0.028335700298)*x[0]**o+(-0.853592964385)*x[0]+(0.554336703119)*x[1]**o+(0.0895176258833)*x[1]+(-0.932756504547)*x[2]**o+(-0.18860048352)*x[2]
        arg[2,3,2,1]=(-0.959044064582)*x[0]**o+(-0.287170878547)*x[0]+(-0.897642111161)*x[1]**o+(0.831990727144)*x[1]+(-0.92281664542)*x[2]**o+(0.96636082362)*x[2]
        arg[2,4,0,0]=(-0.626564609232)*x[0]**o+(0.20773355223)*x[0]+(0.356017212402)*x[1]**o+(0.308760707002)*x[1]+(0.206460527236)*x[2]**o+(-0.343529244774)*x[2]
        arg[2,4,0,1]=(0.176611466931)*x[0]**o+(-0.838329948167)*x[0]+(0.354707816246)*x[1]**o+(0.649522955877)*x[1]+(0.545930184768)*x[2]**o+(-0.859928910201)*x[2]
        arg[2,4,1,0]=(-0.487262434448)*x[0]**o+(0.162363232036)*x[0]+(0.0818357859487)*x[1]**o+(-0.0114240923675)*x[1]+(0.391159686245)*x[2]**o+(0.591883722488)*x[2]
        arg[2,4,1,1]=(0.89780703396)*x[0]**o+(-0.672932240376)*x[0]+(0.166200721661)*x[1]**o+(0.556829138872)*x[1]+(0.420925668076)*x[2]**o+(0.543645319787)*x[2]
        arg[2,4,2,0]=(-0.25384993317)*x[0]**o+(0.347964473446)*x[0]+(-0.579025655422)*x[1]**o+(-0.790871375452)*x[1]+(0.937422382916)*x[2]**o+(0.34952117072)*x[2]
        arg[2,4,2,1]=(-0.979549805807)*x[0]**o+(-0.514338851166)*x[0]+(0.590481348069)*x[1]**o+(0.70090898505)*x[1]+(0.0211126831373)*x[2]**o+(0.112132700392)*x[2]
        arg[3,0,0,0]=(0.712615071708)*x[0]**o+(-0.395279000567)*x[0]+(0.572303465332)*x[1]**o+(-0.18990757808)*x[1]+(0.552173988053)*x[2]**o+(0.66384838566)*x[2]
        arg[3,0,0,1]=(-0.144778596155)*x[0]**o+(0.587515130999)*x[0]+(-0.329480229216)*x[1]**o+(-0.67629797205)*x[1]+(0.132217152221)*x[2]**o+(0.688653172908)*x[2]
        arg[3,0,1,0]=(0.540370436223)*x[0]**o+(0.34677963828)*x[0]+(0.554331354626)*x[1]**o+(0.936508271818)*x[1]+(-0.375652281911)*x[2]**o+(-0.599300575121)*x[2]
        arg[3,0,1,1]=(0.534919542204)*x[0]**o+(-0.153133704085)*x[0]+(-0.121018513777)*x[1]**o+(0.850253887541)*x[1]+(-0.356227750213)*x[2]**o+(0.610895662256)*x[2]
        arg[3,0,2,0]=(-0.29877753893)*x[0]**o+(0.0284310686839)*x[0]+(0.647432584397)*x[1]**o+(-0.215364318325)*x[1]+(-0.580839810547)*x[2]**o+(0.811531717828)*x[2]
        arg[3,0,2,1]=(-0.166467908391)*x[0]**o+(0.75238635157)*x[0]+(0.387810278118)*x[1]**o+(0.516762295584)*x[1]+(-0.451298232627)*x[2]**o+(-0.877792995315)*x[2]
        arg[3,1,0,0]=(0.732332840481)*x[0]**o+(0.0567319375815)*x[0]+(-0.405681445532)*x[1]**o+(0.866920327583)*x[1]+(0.744087567632)*x[2]**o+(-0.257414527262)*x[2]
        arg[3,1,0,1]=(0.641165454863)*x[0]**o+(-0.880320701499)*x[0]+(-0.488201727677)*x[1]**o+(-0.604286576258)*x[1]+(-0.396469224601)*x[2]**o+(0.919924897798)*x[2]
        arg[3,1,1,0]=(0.276953637864)*x[0]**o+(0.825294595252)*x[0]+(-0.00681547733671)*x[1]**o+(0.610898921858)*x[1]+(-0.113640919359)*x[2]**o+(-0.41745293894)*x[2]
        arg[3,1,1,1]=(-0.0556423502541)*x[0]**o+(0.0992203569857)*x[0]+(0.26999259247)*x[1]**o+(0.687729401581)*x[1]+(-0.42626708563)*x[2]**o+(0.137838345424)*x[2]
        arg[3,1,2,0]=(-0.82862652423)*x[0]**o+(0.384688571211)*x[0]+(-0.485754978842)*x[1]**o+(0.289363174139)*x[1]+(0.809823491656)*x[2]**o+(-0.468263092294)*x[2]
        arg[3,1,2,1]=(0.330774389497)*x[0]**o+(0.29069658199)*x[0]+(-0.0852835960508)*x[1]**o+(-0.496327335153)*x[1]+(-0.701004642133)*x[2]**o+(-0.850968582466)*x[2]
        arg[3,2,0,0]=(-0.211872121822)*x[0]**o+(-0.544218825487)*x[0]+(-0.554174609604)*x[1]**o+(0.28186606841)*x[1]+(0.272511831479)*x[2]**o+(0.130059249976)*x[2]
        arg[3,2,0,1]=(0.496592232013)*x[0]**o+(-0.541037353713)*x[0]+(0.848137836475)*x[1]**o+(-0.508526127555)*x[1]+(0.468412010792)*x[2]**o+(0.0139501620456)*x[2]
        arg[3,2,1,0]=(0.987830151765)*x[0]**o+(0.356732413851)*x[0]+(0.958334312356)*x[1]**o+(0.244871130977)*x[1]+(0.200379593754)*x[2]**o+(0.105621740392)*x[2]
        arg[3,2,1,1]=(-0.751728177957)*x[0]**o+(-0.71132733669)*x[0]+(-0.792801223542)*x[1]**o+(0.932243495719)*x[1]+(0.278721571387)*x[2]**o+(0.16204818826)*x[2]
        arg[3,2,2,0]=(-0.844470031769)*x[0]**o+(0.233237514806)*x[0]+(0.567935434819)*x[1]**o+(0.315504620349)*x[1]+(-0.886427966592)*x[2]**o+(-0.514415287917)*x[2]
        arg[3,2,2,1]=(-0.113390105251)*x[0]**o+(-0.227293501546)*x[0]+(-0.501842049116)*x[1]**o+(-0.2393770551)*x[1]+(-0.636288744693)*x[2]**o+(-0.364637404491)*x[2]
        arg[3,3,0,0]=(-0.454817772137)*x[0]**o+(0.0991199341563)*x[0]+(-0.891035204309)*x[1]**o+(-0.894694305011)*x[1]+(0.476722687543)*x[2]**o+(-0.616323189052)*x[2]
        arg[3,3,0,1]=(-0.641204554575)*x[0]**o+(0.418692269871)*x[0]+(-0.497353963751)*x[1]**o+(0.443236835952)*x[1]+(0.771380772745)*x[2]**o+(0.913143060887)*x[2]
        arg[3,3,1,0]=(0.340731926682)*x[0]**o+(0.231807962221)*x[0]+(-0.35677786751)*x[1]**o+(-0.171036937234)*x[1]+(-0.697055158669)*x[2]**o+(0.58599029377)*x[2]
        arg[3,3,1,1]=(0.754473421244)*x[0]**o+(0.658054374737)*x[0]+(0.696211053347)*x[1]**o+(0.401778988971)*x[1]+(0.121272999311)*x[2]**o+(0.509498775106)*x[2]
        arg[3,3,2,0]=(-0.0930185763819)*x[0]**o+(-0.0798587560353)*x[0]+(-0.0370412597984)*x[1]**o+(-0.440515510516)*x[1]+(-0.855771461636)*x[2]**o+(0.216680759655)*x[2]
        arg[3,3,2,1]=(0.763831414202)*x[0]**o+(0.207864762366)*x[0]+(-0.815789136019)*x[1]**o+(-0.808073090924)*x[1]+(0.97581837589)*x[2]**o+(-0.560842077761)*x[2]
        arg[3,4,0,0]=(-0.241036031728)*x[0]**o+(-0.634789191932)*x[0]+(-0.321678486205)*x[1]**o+(-0.754677817943)*x[1]+(-0.078216578971)*x[2]**o+(-0.924028965218)*x[2]
        arg[3,4,0,1]=(-0.851189221011)*x[0]**o+(0.425561499671)*x[0]+(-0.293305262525)*x[1]**o+(-0.191023566093)*x[1]+(-0.810379587547)*x[2]**o+(0.593883389908)*x[2]
        arg[3,4,1,0]=(0.905139594202)*x[0]**o+(-0.904771548418)*x[0]+(0.228410726617)*x[1]**o+(0.327253374618)*x[1]+(-0.214649399974)*x[2]**o+(-0.766320611635)*x[2]
        arg[3,4,1,1]=(0.494131404603)*x[0]**o+(-0.902996362946)*x[0]+(-0.88445944326)*x[1]**o+(-0.458372750995)*x[1]+(-0.306694339897)*x[2]**o+(0.0486088169447)*x[2]
        arg[3,4,2,0]=(0.856321787222)*x[0]**o+(0.314315880734)*x[0]+(-0.996526923819)*x[1]**o+(-0.905235331678)*x[1]+(-0.775589521164)*x[2]**o+(0.812013030327)*x[2]
        arg[3,4,2,1]=(0.271281347275)*x[0]**o+(-0.600074638655)*x[0]+(-0.16332213744)*x[1]**o+(-0.483539594381)*x[1]+(0.820910011338)*x[2]**o+(0.14947223917)*x[2]
        ref[0,0,0,0]=(0.412268765993)*x_ref[0]**o+(-0.31214585911)*x_ref[0]+(0.153520487818)*x_ref[1]**o+(-0.3104168266)*x_ref[1]+(0.909798069186)*x_ref[2]**o+(0.742916044571)*x_ref[2]
        ref[0,0,0,1]=(0.0735781063807)*x_ref[0]**o+(0.725527279895)*x_ref[0]+(-0.884760504783)*x_ref[1]**o+(0.751678022946)*x_ref[1]+(-0.169364499242)*x_ref[2]**o+(-0.946699946704)*x_ref[2]
        ref[0,0,1,0]=(-0.308928255723)*x_ref[0]**o+(-0.601764425359)*x_ref[0]+(-0.176466074749)*x_ref[1]**o+(-0.14838705699)*x_ref[1]+(0.897444985593)*x_ref[2]**o+(-0.0552696055317)*x_ref[2]
        ref[0,0,1,1]=(0.278269305614)*x_ref[0]**o+(-0.419731200355)*x_ref[0]+(0.450955076279)*x_ref[1]**o+(-0.899203207333)*x_ref[1]+(-0.458812838787)*x_ref[2]**o+(0.507871397895)*x_ref[2]
        ref[0,0,2,0]=(0.887873653495)*x_ref[0]**o+(0.130110213077)*x_ref[0]+(0.203761425496)*x_ref[1]**o+(-0.0563898018944)*x_ref[1]+(0.0125548052261)*x_ref[2]**o+(-0.594578723877)*x_ref[2]
        ref[0,0,2,1]=(-0.104210554142)*x_ref[0]**o+(-0.792596213906)*x_ref[0]+(0.570869430862)*x_ref[1]**o+(-0.798160727294)*x_ref[1]+(0.241269856254)*x_ref[2]**o+(0.194885978057)*x_ref[2]
        ref[0,1,0,0]=(0.819550431682)*x_ref[0]**o+(-0.239970258367)*x_ref[0]+(-0.960362264523)*x_ref[1]**o+(0.872093886273)*x_ref[1]+(-0.857325000718)*x_ref[2]**o+(-0.907620072404)*x_ref[2]
        ref[0,1,0,1]=(-0.989011816788)*x_ref[0]**o+(0.853577413178)*x_ref[0]+(-0.160335863166)*x_ref[1]**o+(-0.751266166108)*x_ref[1]+(0.631146799273)*x_ref[2]**o+(0.739741622363)*x_ref[2]
        ref[0,1,1,0]=(0.162828987854)*x_ref[0]**o+(-0.200652438446)*x_ref[0]+(-0.890837153528)*x_ref[1]**o+(-0.972029967593)*x_ref[1]+(0.850395155009)*x_ref[2]**o+(-0.0868272213807)*x_ref[2]
        ref[0,1,1,1]=(-0.770711232994)*x_ref[0]**o+(0.986518762896)*x_ref[0]+(0.0970713645745)*x_ref[1]**o+(0.996843418865)*x_ref[1]+(0.208134397469)*x_ref[2]**o+(0.946600905711)*x_ref[2]
        ref[0,1,2,0]=(-0.543213476288)*x_ref[0]**o+(0.0335009416664)*x_ref[0]+(0.165751788147)*x_ref[1]**o+(0.660276877927)*x_ref[1]+(-0.485925597349)*x_ref[2]**o+(-0.632302900376)*x_ref[2]
        ref[0,1,2,1]=(0.175798920586)*x_ref[0]**o+(-0.42873239224)*x_ref[0]+(0.164896129395)*x_ref[1]**o+(0.662142729145)*x_ref[1]+(-0.425944275652)*x_ref[2]**o+(0.306978282183)*x_ref[2]
        ref[0,2,0,0]=(0.790303828628)*x_ref[0]**o+(0.169346034559)*x_ref[0]+(-0.901026918205)*x_ref[1]**o+(0.661856377517)*x_ref[1]+(0.194850846261)*x_ref[2]**o+(0.155282302183)*x_ref[2]
        ref[0,2,0,1]=(-0.43947970094)*x_ref[0]**o+(-0.535569903832)*x_ref[0]+(-0.534035397281)*x_ref[1]**o+(-0.169724793112)*x_ref[1]+(-0.717226228021)*x_ref[2]**o+(0.771292292609)*x_ref[2]
        ref[0,2,1,0]=(-0.424991151717)*x_ref[0]**o+(-0.0837800409811)*x_ref[0]+(-0.747139428246)*x_ref[1]**o+(0.739025612454)*x_ref[1]+(0.464486611081)*x_ref[2]**o+(0.824926522179)*x_ref[2]
        ref[0,2,1,1]=(-0.4666399509)*x_ref[0]**o+(-0.479130555457)*x_ref[0]+(-0.676718441278)*x_ref[1]**o+(0.352685580535)*x_ref[1]+(0.537789487962)*x_ref[2]**o+(-0.774383681009)*x_ref[2]
        ref[0,2,2,0]=(-0.30983510311)*x_ref[0]**o+(0.00141603592636)*x_ref[0]+(0.546192108985)*x_ref[1]**o+(0.229628329192)*x_ref[1]+(0.788466929747)*x_ref[2]**o+(-0.169587087096)*x_ref[2]
        ref[0,2,2,1]=(0.810282347322)*x_ref[0]**o+(0.961056346191)*x_ref[0]+(0.204622339898)*x_ref[1]**o+(0.288786527052)*x_ref[1]+(0.87096784221)*x_ref[2]**o+(0.967565754848)*x_ref[2]
        ref[0,3,0,0]=(-0.154669358868)*x_ref[0]**o+(-0.449349763143)*x_ref[0]+(-0.692997905123)*x_ref[1]**o+(0.666308436198)*x_ref[1]+(0.0638963436229)*x_ref[2]**o+(-0.318233405188)*x_ref[2]
        ref[0,3,0,1]=(0.951324381413)*x_ref[0]**o+(0.10400502165)*x_ref[0]+(-0.910416795399)*x_ref[1]**o+(-0.931501370526)*x_ref[1]+(0.455893141186)*x_ref[2]**o+(0.536556088937)*x_ref[2]
        ref[0,3,1,0]=(-0.315779909544)*x_ref[0]**o+(0.189718508813)*x_ref[0]+(0.0995068185023)*x_ref[1]**o+(-0.863775500494)*x_ref[1]+(0.99696081153)*x_ref[2]**o+(0.0941784151393)*x_ref[2]
        ref[0,3,1,1]=(0.122766381329)*x_ref[0]**o+(-0.953220464103)*x_ref[0]+(-0.853780873931)*x_ref[1]**o+(-0.719623724002)*x_ref[1]+(-0.58783153111)*x_ref[2]**o+(-0.919075056141)*x_ref[2]
        ref[0,3,2,0]=(-0.740349380942)*x_ref[0]**o+(-0.0161629946027)*x_ref[0]+(0.790593892223)*x_ref[1]**o+(0.81563688519)*x_ref[1]+(0.515002236312)*x_ref[2]**o+(0.906399338047)*x_ref[2]
        ref[0,3,2,1]=(-0.21876846821)*x_ref[0]**o+(0.099226679818)*x_ref[0]+(-0.501220667305)*x_ref[1]**o+(-0.645539150063)*x_ref[1]+(0.705501523136)*x_ref[2]**o+(-0.640426266054)*x_ref[2]
        ref[0,4,0,0]=(0.382430876525)*x_ref[0]**o+(0.728752519135)*x_ref[0]+(-0.696198853186)*x_ref[1]**o+(0.181187001577)*x_ref[1]+(0.818931936956)*x_ref[2]**o+(0.267607197602)*x_ref[2]
        ref[0,4,0,1]=(0.667658893159)*x_ref[0]**o+(0.129176508255)*x_ref[0]+(0.579266506883)*x_ref[1]**o+(-0.27290178784)*x_ref[1]+(0.804307104985)*x_ref[2]**o+(-0.384179099369)*x_ref[2]
        ref[0,4,1,0]=(0.472611036284)*x_ref[0]**o+(0.532026319531)*x_ref[0]+(-0.11900311657)*x_ref[1]**o+(0.818635083182)*x_ref[1]+(-0.127618053257)*x_ref[2]**o+(-0.8464024968)*x_ref[2]
        ref[0,4,1,1]=(-0.52572498077)*x_ref[0]**o+(-0.222527963121)*x_ref[0]+(-0.733344395469)*x_ref[1]**o+(-0.519230573211)*x_ref[1]+(0.550298161034)*x_ref[2]**o+(-0.091253041563)*x_ref[2]
        ref[0,4,2,0]=(-0.995556194807)*x_ref[0]**o+(-0.157265619525)*x_ref[0]+(0.646987111139)*x_ref[1]**o+(0.114269194694)*x_ref[1]+(0.787559342851)*x_ref[2]**o+(0.655874189666)*x_ref[2]
        ref[0,4,2,1]=(-0.534706379451)*x_ref[0]**o+(0.89989539183)*x_ref[0]+(0.584227019406)*x_ref[1]**o+(-0.261650143902)*x_ref[1]+(-0.0741247974759)*x_ref[2]**o+(-0.646603374886)*x_ref[2]
        ref[1,0,0,0]=(-0.16346294005)*x_ref[0]**o+(0.515625993676)*x_ref[0]+(-0.0294710658489)*x_ref[1]**o+(-0.886489878353)*x_ref[1]+(0.184376222466)*x_ref[2]**o+(0.872455440027)*x_ref[2]
        ref[1,0,0,1]=(-0.239569565644)*x_ref[0]**o+(-0.0196989227301)*x_ref[0]+(-0.333999804595)*x_ref[1]**o+(0.79703425655)*x_ref[1]+(0.332925813867)*x_ref[2]**o+(0.384766118619)*x_ref[2]
        ref[1,0,1,0]=(0.337775521215)*x_ref[0]**o+(-0.214105347244)*x_ref[0]+(-0.055476365922)*x_ref[1]**o+(-0.511469291523)*x_ref[1]+(0.390455425321)*x_ref[2]**o+(0.346749928211)*x_ref[2]
        ref[1,0,1,1]=(0.255034684548)*x_ref[0]**o+(-0.975969854422)*x_ref[0]+(-0.588154151576)*x_ref[1]**o+(0.981825996023)*x_ref[1]+(-0.16323013592)*x_ref[2]**o+(-0.165779157778)*x_ref[2]
        ref[1,0,2,0]=(-0.421647673573)*x_ref[0]**o+(0.0892056009306)*x_ref[0]+(-0.659566191892)*x_ref[1]**o+(-0.640769435519)*x_ref[1]+(-0.5474352085)*x_ref[2]**o+(0.619855556292)*x_ref[2]
        ref[1,0,2,1]=(0.678408254014)*x_ref[0]**o+(0.114013125353)*x_ref[0]+(0.222310135576)*x_ref[1]**o+(-0.297909583371)*x_ref[1]+(0.885716237809)*x_ref[2]**o+(0.151576219626)*x_ref[2]
        ref[1,1,0,0]=(-0.724825546043)*x_ref[0]**o+(-0.796280833482)*x_ref[0]+(0.894034526494)*x_ref[1]**o+(-0.496334573913)*x_ref[1]+(0.161130154754)*x_ref[2]**o+(0.0137999166704)*x_ref[2]
        ref[1,1,0,1]=(-0.959049520871)*x_ref[0]**o+(0.0727868983195)*x_ref[0]+(-0.0475524824482)*x_ref[1]**o+(0.8098016511)*x_ref[1]+(0.154268238454)*x_ref[2]**o+(-0.0583213692933)*x_ref[2]
        ref[1,1,1,0]=(-0.684320939672)*x_ref[0]**o+(-0.841165491324)*x_ref[0]+(0.307147508965)*x_ref[1]**o+(0.0204906256209)*x_ref[1]+(0.658432728833)*x_ref[2]**o+(0.398570233826)*x_ref[2]
        ref[1,1,1,1]=(0.921894102193)*x_ref[0]**o+(0.33017900913)*x_ref[0]+(0.566856960125)*x_ref[1]**o+(0.826695387569)*x_ref[1]+(-0.6253390282)*x_ref[2]**o+(-0.527610353599)*x_ref[2]
        ref[1,1,2,0]=(0.0640251272456)*x_ref[0]**o+(0.789847887603)*x_ref[0]+(-0.00457885872972)*x_ref[1]**o+(-0.24681292381)*x_ref[1]+(-0.0832674042894)*x_ref[2]**o+(-0.254280656434)*x_ref[2]
        ref[1,1,2,1]=(0.786208197486)*x_ref[0]**o+(0.519575725184)*x_ref[0]+(-0.452657209679)*x_ref[1]**o+(0.36570806994)*x_ref[1]+(0.875049922923)*x_ref[2]**o+(0.881026792786)*x_ref[2]
        ref[1,2,0,0]=(-0.415919514711)*x_ref[0]**o+(-0.933890170641)*x_ref[0]+(0.717990882823)*x_ref[1]**o+(0.355555237378)*x_ref[1]+(0.458971778535)*x_ref[2]**o+(0.888647676837)*x_ref[2]
        ref[1,2,0,1]=(-0.453925437723)*x_ref[0]**o+(0.153501471971)*x_ref[0]+(0.815830410181)*x_ref[1]**o+(-0.519500430324)*x_ref[1]+(-0.791530261324)*x_ref[2]**o+(0.849179201427)*x_ref[2]
        ref[1,2,1,0]=(0.299736697301)*x_ref[0]**o+(0.645276033981)*x_ref[0]+(-0.731636401343)*x_ref[1]**o+(0.127189134157)*x_ref[1]+(0.891427235216)*x_ref[2]**o+(0.750052326597)*x_ref[2]
        ref[1,2,1,1]=(-0.987392548561)*x_ref[0]**o+(0.129053618821)*x_ref[0]+(0.220541501123)*x_ref[1]**o+(0.276873448998)*x_ref[1]+(0.645231056419)*x_ref[2]**o+(0.776634462468)*x_ref[2]
        ref[1,2,2,0]=(-0.879738240565)*x_ref[0]**o+(-0.0352288276658)*x_ref[0]+(-0.247550105519)*x_ref[1]**o+(0.782133317388)*x_ref[1]+(-0.948666864992)*x_ref[2]**o+(0.310357478071)*x_ref[2]
        ref[1,2,2,1]=(-0.606778611022)*x_ref[0]**o+(0.964356172073)*x_ref[0]+(-0.520291905065)*x_ref[1]**o+(-0.219411824153)*x_ref[1]+(-0.405356968701)*x_ref[2]**o+(0.969492630653)*x_ref[2]
        ref[1,3,0,0]=(-0.991624055852)*x_ref[0]**o+(-0.762240034313)*x_ref[0]+(0.423262307551)*x_ref[1]**o+(-0.05789904696)*x_ref[1]+(-0.696181866276)*x_ref[2]**o+(-0.327406198937)*x_ref[2]
        ref[1,3,0,1]=(-0.614181964006)*x_ref[0]**o+(0.927675019363)*x_ref[0]+(-0.750120185372)*x_ref[1]**o+(0.440548313639)*x_ref[1]+(-0.116776982636)*x_ref[2]**o+(-0.799181582827)*x_ref[2]
        ref[1,3,1,0]=(0.286032474963)*x_ref[0]**o+(0.391407042942)*x_ref[0]+(-0.268363949002)*x_ref[1]**o+(-0.421319501252)*x_ref[1]+(-0.318595882607)*x_ref[2]**o+(0.273120761382)*x_ref[2]
        ref[1,3,1,1]=(0.918425098597)*x_ref[0]**o+(-0.211260796786)*x_ref[0]+(-0.793780530123)*x_ref[1]**o+(-0.839335876894)*x_ref[1]+(-0.595786792228)*x_ref[2]**o+(-0.592690829684)*x_ref[2]
        ref[1,3,2,0]=(-0.959108168319)*x_ref[0]**o+(-0.94492160223)*x_ref[0]+(0.0778030192626)*x_ref[1]**o+(-0.169988420265)*x_ref[1]+(0.042653831433)*x_ref[2]**o+(-0.595426371148)*x_ref[2]
        ref[1,3,2,1]=(0.897478414413)*x_ref[0]**o+(-0.222864711011)*x_ref[0]+(0.900965727203)*x_ref[1]**o+(0.501243022513)*x_ref[1]+(0.75496626209)*x_ref[2]**o+(0.312225549338)*x_ref[2]
        ref[1,4,0,0]=(0.168321984115)*x_ref[0]**o+(-0.495995847038)*x_ref[0]+(0.595773350289)*x_ref[1]**o+(-0.914906345927)*x_ref[1]+(0.224831726299)*x_ref[2]**o+(-0.0743642212453)*x_ref[2]
        ref[1,4,0,1]=(0.352717252696)*x_ref[0]**o+(-0.873077111264)*x_ref[0]+(-0.296366541486)*x_ref[1]**o+(-0.00338438754009)*x_ref[1]+(0.0294923296914)*x_ref[2]**o+(-0.067479335313)*x_ref[2]
        ref[1,4,1,0]=(-0.518757232583)*x_ref[0]**o+(0.463948954744)*x_ref[0]+(0.712607983792)*x_ref[1]**o+(0.717184663437)*x_ref[1]+(0.323391383016)*x_ref[2]**o+(-0.225868711026)*x_ref[2]
        ref[1,4,1,1]=(-0.24381531137)*x_ref[0]**o+(0.630251169076)*x_ref[0]+(0.70462966259)*x_ref[1]**o+(-0.125357204688)*x_ref[1]+(-0.258633435955)*x_ref[2]**o+(0.902171387163)*x_ref[2]
        ref[1,4,2,0]=(-0.632179111447)*x_ref[0]**o+(-0.59290738268)*x_ref[0]+(-0.825795231706)*x_ref[1]**o+(0.399660693305)*x_ref[1]+(0.0620015248102)*x_ref[2]**o+(-0.451556558781)*x_ref[2]
        ref[1,4,2,1]=(-0.858145688916)*x_ref[0]**o+(-0.787993091462)*x_ref[0]+(0.803723603768)*x_ref[1]**o+(-0.835043986384)*x_ref[1]+(0.467577922972)*x_ref[2]**o+(0.819139387663)*x_ref[2]
        ref[2,0,0,0]=(0.274886550898)*x_ref[0]**o+(-0.0742978734335)*x_ref[0]+(-0.773015479116)*x_ref[1]**o+(-0.138971598539)*x_ref[1]+(-0.702175483988)*x_ref[2]**o+(0.9147072714)*x_ref[2]
        ref[2,0,0,1]=(-0.598714805644)*x_ref[0]**o+(0.11878664701)*x_ref[0]+(-0.554608391808)*x_ref[1]**o+(-0.734151257491)*x_ref[1]+(0.234629053469)*x_ref[2]**o+(0.911072319646)*x_ref[2]
        ref[2,0,1,0]=(-0.83850906403)*x_ref[0]**o+(0.799369711531)*x_ref[0]+(0.436211860763)*x_ref[1]**o+(-0.855173818525)*x_ref[1]+(0.256241482884)*x_ref[2]**o+(0.657883842259)*x_ref[2]
        ref[2,0,1,1]=(-0.724328143383)*x_ref[0]**o+(-0.781110608395)*x_ref[0]+(-0.987464957106)*x_ref[1]**o+(0.936448332957)*x_ref[1]+(0.711870454531)*x_ref[2]**o+(0.64958966746)*x_ref[2]
        ref[2,0,2,0]=(0.485490298364)*x_ref[0]**o+(-0.648402414524)*x_ref[0]+(0.0493575279485)*x_ref[1]**o+(-0.168695660031)*x_ref[1]+(0.236500977054)*x_ref[2]**o+(-0.699035972752)*x_ref[2]
        ref[2,0,2,1]=(-0.508561382238)*x_ref[0]**o+(0.0587152225749)*x_ref[0]+(-0.92072872998)*x_ref[1]**o+(-0.210536816562)*x_ref[1]+(0.735465423941)*x_ref[2]**o+(0.557002016885)*x_ref[2]
        ref[2,1,0,0]=(-0.301683530613)*x_ref[0]**o+(-0.501004347082)*x_ref[0]+(-0.415125116853)*x_ref[1]**o+(-0.975209043974)*x_ref[1]+(0.0952958379219)*x_ref[2]**o+(-0.882009529251)*x_ref[2]
        ref[2,1,0,1]=(0.569047454657)*x_ref[0]**o+(0.87107862461)*x_ref[0]+(0.977655371485)*x_ref[1]**o+(-0.870823933133)*x_ref[1]+(-0.12817645442)*x_ref[2]**o+(-0.412450404053)*x_ref[2]
        ref[2,1,1,0]=(0.532638832559)*x_ref[0]**o+(-0.669778224526)*x_ref[0]+(0.516969768896)*x_ref[1]**o+(0.725495018612)*x_ref[1]+(0.604820640277)*x_ref[2]**o+(-0.537483520132)*x_ref[2]
        ref[2,1,1,1]=(-0.413776169533)*x_ref[0]**o+(0.19037961067)*x_ref[0]+(-0.109551737415)*x_ref[1]**o+(-0.195599236834)*x_ref[1]+(-0.183237365211)*x_ref[2]**o+(0.0279625528407)*x_ref[2]
        ref[2,1,2,0]=(-0.46089173487)*x_ref[0]**o+(-0.151158992117)*x_ref[0]+(0.359508089079)*x_ref[1]**o+(-0.0808350778984)*x_ref[1]+(0.845464278946)*x_ref[2]**o+(0.177891463715)*x_ref[2]
        ref[2,1,2,1]=(0.416828200435)*x_ref[0]**o+(0.763110094817)*x_ref[0]+(0.855595717948)*x_ref[1]**o+(0.379358815658)*x_ref[1]+(0.954688280021)*x_ref[2]**o+(-0.808098938468)*x_ref[2]
        ref[2,2,0,0]=(-0.697342667428)*x_ref[0]**o+(-0.855608259014)*x_ref[0]+(0.866084171318)*x_ref[1]**o+(0.632470977708)*x_ref[1]+(-0.399703926338)*x_ref[2]**o+(0.211173996927)*x_ref[2]
        ref[2,2,0,1]=(-0.940742486357)*x_ref[0]**o+(0.249267367828)*x_ref[0]+(0.197336418576)*x_ref[1]**o+(-0.0573784352507)*x_ref[1]+(-0.328108062127)*x_ref[2]**o+(0.570204614968)*x_ref[2]
        ref[2,2,1,0]=(0.134960606335)*x_ref[0]**o+(0.490123643194)*x_ref[0]+(-0.102294210304)*x_ref[1]**o+(0.427727810449)*x_ref[1]+(-0.158572569775)*x_ref[2]**o+(-0.165504764483)*x_ref[2]
        ref[2,2,1,1]=(-0.427912549906)*x_ref[0]**o+(-0.311524081823)*x_ref[0]+(0.220211421031)*x_ref[1]**o+(0.123679381711)*x_ref[1]+(0.593149361283)*x_ref[2]**o+(-0.283233437917)*x_ref[2]
        ref[2,2,2,0]=(0.901179489164)*x_ref[0]**o+(0.292780095159)*x_ref[0]+(-0.295149620179)*x_ref[1]**o+(0.860483795735)*x_ref[1]+(0.848872993178)*x_ref[2]**o+(-0.612494025994)*x_ref[2]
        ref[2,2,2,1]=(0.904689835334)*x_ref[0]**o+(0.33689603833)*x_ref[0]+(0.0765595244364)*x_ref[1]**o+(0.215174063849)*x_ref[1]+(0.973122153293)*x_ref[2]**o+(-0.49840689696)*x_ref[2]
        ref[2,3,0,0]=(0.973713543725)*x_ref[0]**o+(0.0727815483094)*x_ref[0]+(-0.0723336244169)*x_ref[1]**o+(-0.102547958784)*x_ref[1]+(0.573980704603)*x_ref[2]**o+(0.784756916348)*x_ref[2]
        ref[2,3,0,1]=(0.871042616423)*x_ref[0]**o+(-0.966021391258)*x_ref[0]+(0.9402258232)*x_ref[1]**o+(0.163522806825)*x_ref[1]+(-0.260871238896)*x_ref[2]**o+(0.68201099906)*x_ref[2]
        ref[2,3,1,0]=(0.916349358422)*x_ref[0]**o+(-0.923720463388)*x_ref[0]+(0.814116732807)*x_ref[1]**o+(-0.521079005717)*x_ref[1]+(-0.996963157941)*x_ref[2]**o+(-0.251248775427)*x_ref[2]
        ref[2,3,1,1]=(0.752568485119)*x_ref[0]**o+(-0.134517175046)*x_ref[0]+(0.691477362575)*x_ref[1]**o+(-0.864047765263)*x_ref[1]+(0.0632525427291)*x_ref[2]**o+(0.27617797274)*x_ref[2]
        ref[2,3,2,0]=(0.028335700298)*x_ref[0]**o+(-0.853592964385)*x_ref[0]+(0.554336703119)*x_ref[1]**o+(0.0895176258833)*x_ref[1]+(-0.932756504547)*x_ref[2]**o+(-0.18860048352)*x_ref[2]
        ref[2,3,2,1]=(-0.959044064582)*x_ref[0]**o+(-0.287170878547)*x_ref[0]+(-0.897642111161)*x_ref[1]**o+(0.831990727144)*x_ref[1]+(-0.92281664542)*x_ref[2]**o+(0.96636082362)*x_ref[2]
        ref[2,4,0,0]=(-0.626564609232)*x_ref[0]**o+(0.20773355223)*x_ref[0]+(0.356017212402)*x_ref[1]**o+(0.308760707002)*x_ref[1]+(0.206460527236)*x_ref[2]**o+(-0.343529244774)*x_ref[2]
        ref[2,4,0,1]=(0.176611466931)*x_ref[0]**o+(-0.838329948167)*x_ref[0]+(0.354707816246)*x_ref[1]**o+(0.649522955877)*x_ref[1]+(0.545930184768)*x_ref[2]**o+(-0.859928910201)*x_ref[2]
        ref[2,4,1,0]=(-0.487262434448)*x_ref[0]**o+(0.162363232036)*x_ref[0]+(0.0818357859487)*x_ref[1]**o+(-0.0114240923675)*x_ref[1]+(0.391159686245)*x_ref[2]**o+(0.591883722488)*x_ref[2]
        ref[2,4,1,1]=(0.89780703396)*x_ref[0]**o+(-0.672932240376)*x_ref[0]+(0.166200721661)*x_ref[1]**o+(0.556829138872)*x_ref[1]+(0.420925668076)*x_ref[2]**o+(0.543645319787)*x_ref[2]
        ref[2,4,2,0]=(-0.25384993317)*x_ref[0]**o+(0.347964473446)*x_ref[0]+(-0.579025655422)*x_ref[1]**o+(-0.790871375452)*x_ref[1]+(0.937422382916)*x_ref[2]**o+(0.34952117072)*x_ref[2]
        ref[2,4,2,1]=(-0.979549805807)*x_ref[0]**o+(-0.514338851166)*x_ref[0]+(0.590481348069)*x_ref[1]**o+(0.70090898505)*x_ref[1]+(0.0211126831373)*x_ref[2]**o+(0.112132700392)*x_ref[2]
        ref[3,0,0,0]=(0.712615071708)*x_ref[0]**o+(-0.395279000567)*x_ref[0]+(0.572303465332)*x_ref[1]**o+(-0.18990757808)*x_ref[1]+(0.552173988053)*x_ref[2]**o+(0.66384838566)*x_ref[2]
        ref[3,0,0,1]=(-0.144778596155)*x_ref[0]**o+(0.587515130999)*x_ref[0]+(-0.329480229216)*x_ref[1]**o+(-0.67629797205)*x_ref[1]+(0.132217152221)*x_ref[2]**o+(0.688653172908)*x_ref[2]
        ref[3,0,1,0]=(0.540370436223)*x_ref[0]**o+(0.34677963828)*x_ref[0]+(0.554331354626)*x_ref[1]**o+(0.936508271818)*x_ref[1]+(-0.375652281911)*x_ref[2]**o+(-0.599300575121)*x_ref[2]
        ref[3,0,1,1]=(0.534919542204)*x_ref[0]**o+(-0.153133704085)*x_ref[0]+(-0.121018513777)*x_ref[1]**o+(0.850253887541)*x_ref[1]+(-0.356227750213)*x_ref[2]**o+(0.610895662256)*x_ref[2]
        ref[3,0,2,0]=(-0.29877753893)*x_ref[0]**o+(0.0284310686839)*x_ref[0]+(0.647432584397)*x_ref[1]**o+(-0.215364318325)*x_ref[1]+(-0.580839810547)*x_ref[2]**o+(0.811531717828)*x_ref[2]
        ref[3,0,2,1]=(-0.166467908391)*x_ref[0]**o+(0.75238635157)*x_ref[0]+(0.387810278118)*x_ref[1]**o+(0.516762295584)*x_ref[1]+(-0.451298232627)*x_ref[2]**o+(-0.877792995315)*x_ref[2]
        ref[3,1,0,0]=(0.732332840481)*x_ref[0]**o+(0.0567319375815)*x_ref[0]+(-0.405681445532)*x_ref[1]**o+(0.866920327583)*x_ref[1]+(0.744087567632)*x_ref[2]**o+(-0.257414527262)*x_ref[2]
        ref[3,1,0,1]=(0.641165454863)*x_ref[0]**o+(-0.880320701499)*x_ref[0]+(-0.488201727677)*x_ref[1]**o+(-0.604286576258)*x_ref[1]+(-0.396469224601)*x_ref[2]**o+(0.919924897798)*x_ref[2]
        ref[3,1,1,0]=(0.276953637864)*x_ref[0]**o+(0.825294595252)*x_ref[0]+(-0.00681547733671)*x_ref[1]**o+(0.610898921858)*x_ref[1]+(-0.113640919359)*x_ref[2]**o+(-0.41745293894)*x_ref[2]
        ref[3,1,1,1]=(-0.0556423502541)*x_ref[0]**o+(0.0992203569857)*x_ref[0]+(0.26999259247)*x_ref[1]**o+(0.687729401581)*x_ref[1]+(-0.42626708563)*x_ref[2]**o+(0.137838345424)*x_ref[2]
        ref[3,1,2,0]=(-0.82862652423)*x_ref[0]**o+(0.384688571211)*x_ref[0]+(-0.485754978842)*x_ref[1]**o+(0.289363174139)*x_ref[1]+(0.809823491656)*x_ref[2]**o+(-0.468263092294)*x_ref[2]
        ref[3,1,2,1]=(0.330774389497)*x_ref[0]**o+(0.29069658199)*x_ref[0]+(-0.0852835960508)*x_ref[1]**o+(-0.496327335153)*x_ref[1]+(-0.701004642133)*x_ref[2]**o+(-0.850968582466)*x_ref[2]
        ref[3,2,0,0]=(-0.211872121822)*x_ref[0]**o+(-0.544218825487)*x_ref[0]+(-0.554174609604)*x_ref[1]**o+(0.28186606841)*x_ref[1]+(0.272511831479)*x_ref[2]**o+(0.130059249976)*x_ref[2]
        ref[3,2,0,1]=(0.496592232013)*x_ref[0]**o+(-0.541037353713)*x_ref[0]+(0.848137836475)*x_ref[1]**o+(-0.508526127555)*x_ref[1]+(0.468412010792)*x_ref[2]**o+(0.0139501620456)*x_ref[2]
        ref[3,2,1,0]=(0.987830151765)*x_ref[0]**o+(0.356732413851)*x_ref[0]+(0.958334312356)*x_ref[1]**o+(0.244871130977)*x_ref[1]+(0.200379593754)*x_ref[2]**o+(0.105621740392)*x_ref[2]
        ref[3,2,1,1]=(-0.751728177957)*x_ref[0]**o+(-0.71132733669)*x_ref[0]+(-0.792801223542)*x_ref[1]**o+(0.932243495719)*x_ref[1]+(0.278721571387)*x_ref[2]**o+(0.16204818826)*x_ref[2]
        ref[3,2,2,0]=(-0.844470031769)*x_ref[0]**o+(0.233237514806)*x_ref[0]+(0.567935434819)*x_ref[1]**o+(0.315504620349)*x_ref[1]+(-0.886427966592)*x_ref[2]**o+(-0.514415287917)*x_ref[2]
        ref[3,2,2,1]=(-0.113390105251)*x_ref[0]**o+(-0.227293501546)*x_ref[0]+(-0.501842049116)*x_ref[1]**o+(-0.2393770551)*x_ref[1]+(-0.636288744693)*x_ref[2]**o+(-0.364637404491)*x_ref[2]
        ref[3,3,0,0]=(-0.454817772137)*x_ref[0]**o+(0.0991199341563)*x_ref[0]+(-0.891035204309)*x_ref[1]**o+(-0.894694305011)*x_ref[1]+(0.476722687543)*x_ref[2]**o+(-0.616323189052)*x_ref[2]
        ref[3,3,0,1]=(-0.641204554575)*x_ref[0]**o+(0.418692269871)*x_ref[0]+(-0.497353963751)*x_ref[1]**o+(0.443236835952)*x_ref[1]+(0.771380772745)*x_ref[2]**o+(0.913143060887)*x_ref[2]
        ref[3,3,1,0]=(0.340731926682)*x_ref[0]**o+(0.231807962221)*x_ref[0]+(-0.35677786751)*x_ref[1]**o+(-0.171036937234)*x_ref[1]+(-0.697055158669)*x_ref[2]**o+(0.58599029377)*x_ref[2]
        ref[3,3,1,1]=(0.754473421244)*x_ref[0]**o+(0.658054374737)*x_ref[0]+(0.696211053347)*x_ref[1]**o+(0.401778988971)*x_ref[1]+(0.121272999311)*x_ref[2]**o+(0.509498775106)*x_ref[2]
        ref[3,3,2,0]=(-0.0930185763819)*x_ref[0]**o+(-0.0798587560353)*x_ref[0]+(-0.0370412597984)*x_ref[1]**o+(-0.440515510516)*x_ref[1]+(-0.855771461636)*x_ref[2]**o+(0.216680759655)*x_ref[2]
        ref[3,3,2,1]=(0.763831414202)*x_ref[0]**o+(0.207864762366)*x_ref[0]+(-0.815789136019)*x_ref[1]**o+(-0.808073090924)*x_ref[1]+(0.97581837589)*x_ref[2]**o+(-0.560842077761)*x_ref[2]
        ref[3,4,0,0]=(-0.241036031728)*x_ref[0]**o+(-0.634789191932)*x_ref[0]+(-0.321678486205)*x_ref[1]**o+(-0.754677817943)*x_ref[1]+(-0.078216578971)*x_ref[2]**o+(-0.924028965218)*x_ref[2]
        ref[3,4,0,1]=(-0.851189221011)*x_ref[0]**o+(0.425561499671)*x_ref[0]+(-0.293305262525)*x_ref[1]**o+(-0.191023566093)*x_ref[1]+(-0.810379587547)*x_ref[2]**o+(0.593883389908)*x_ref[2]
        ref[3,4,1,0]=(0.905139594202)*x_ref[0]**o+(-0.904771548418)*x_ref[0]+(0.228410726617)*x_ref[1]**o+(0.327253374618)*x_ref[1]+(-0.214649399974)*x_ref[2]**o+(-0.766320611635)*x_ref[2]
        ref[3,4,1,1]=(0.494131404603)*x_ref[0]**o+(-0.902996362946)*x_ref[0]+(-0.88445944326)*x_ref[1]**o+(-0.458372750995)*x_ref[1]+(-0.306694339897)*x_ref[2]**o+(0.0486088169447)*x_ref[2]
        ref[3,4,2,0]=(0.856321787222)*x_ref[0]**o+(0.314315880734)*x_ref[0]+(-0.996526923819)*x_ref[1]**o+(-0.905235331678)*x_ref[1]+(-0.775589521164)*x_ref[2]**o+(0.812013030327)*x_ref[2]
        ref[3,4,2,1]=(0.271281347275)*x_ref[0]**o+(-0.600074638655)*x_ref[0]+(-0.16332213744)*x_ref[1]**o+(-0.483539594381)*x_ref[1]+(0.820910011338)*x_ref[2]**o+(0.14947223917)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0368926203259)*x[0]**o+(-0.256519664931)*x[0]+(0.168576661932)*x[1]**o+(-0.231158401549)*x[1]
        ref=(0.0368926203259)*x_ref[0]**o+(-0.256519664931)*x_ref[0]+(0.168576661932)*x_ref[1]**o+(-0.231158401549)*x_ref[1]
      else:
        arg=(0.597124147524)*x[0]**o+(-0.404660180764)*x[0]+(-0.923138194509)*x[1]**o+(-0.263963516316)*x[1]+(-0.897643246613)*x[2]**o+(0.0180371552704)*x[2]
        ref=(0.597124147524)*x_ref[0]**o+(-0.404660180764)*x_ref[0]+(-0.923138194509)*x_ref[1]**o+(-0.263963516316)*x_ref[1]+(-0.897643246613)*x_ref[2]**o+(0.0180371552704)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(0.639330723902)*x[0]**o+(0.404929658475)*x[0]+(0.000241871115531)*x[1]**o+(0.770530713516)*x[1]
        arg[1]=(0.245348516629)*x[0]**o+(-0.00142757994203)*x[0]+(0.297525985266)*x[1]**o+(0.683033847851)*x[1]
        ref[0]=(0.639330723902)*x_ref[0]**o+(0.404929658475)*x_ref[0]+(0.000241871115531)*x_ref[1]**o+(0.770530713516)*x_ref[1]
        ref[1]=(0.245348516629)*x_ref[0]**o+(-0.00142757994203)*x_ref[0]+(0.297525985266)*x_ref[1]**o+(0.683033847851)*x_ref[1]
      else:
        arg[0]=(-0.45360978548)*x[0]**o+(-0.069702953017)*x[0]+(0.745603275442)*x[1]**o+(0.591869498191)*x[1]+(-0.90835287387)*x[2]**o+(-0.0405228474208)*x[2]
        arg[1]=(-0.659979134044)*x[0]**o+(0.393634998529)*x[0]+(-0.685193662459)*x[1]**o+(0.773497371593)*x[1]+(0.139726737595)*x[2]**o+(-0.549174223126)*x[2]
        ref[0]=(-0.45360978548)*x_ref[0]**o+(-0.069702953017)*x_ref[0]+(0.745603275442)*x_ref[1]**o+(0.591869498191)*x_ref[1]+(-0.90835287387)*x_ref[2]**o+(-0.0405228474208)*x_ref[2]
        ref[1]=(-0.659979134044)*x_ref[0]**o+(0.393634998529)*x_ref[0]+(-0.685193662459)*x_ref[1]**o+(0.773497371593)*x_ref[1]+(0.139726737595)*x_ref[2]**o+(-0.549174223126)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(0.0734861043324)*x[0]**o+(0.698132844974)*x[0]+(-0.0711338889844)*x[1]**o+(0.667313854486)*x[1]
        arg[0,1]=(0.846132724611)*x[0]**o+(-0.215521619475)*x[0]+(-0.920637540655)*x[1]**o+(0.666739063184)*x[1]
        arg[0,2]=(-0.963337851184)*x[0]**o+(-0.961935972833)*x[0]+(-0.0689798144115)*x[1]**o+(0.0106557770119)*x[1]
        arg[0,3]=(0.875063015269)*x[0]**o+(-0.735727431445)*x[0]+(-0.599251007104)*x[1]**o+(-0.274659381946)*x[1]
        arg[0,4]=(-0.378214565543)*x[0]**o+(-0.476230056857)*x[0]+(0.617722084788)*x[1]**o+(-0.808732634287)*x[1]
        arg[1,0]=(-0.324652264865)*x[0]**o+(-0.314957074449)*x[0]+(0.467828791079)*x[1]**o+(-0.105499299461)*x[1]
        arg[1,1]=(-0.0815291997441)*x[0]**o+(0.1473072227)*x[0]+(-0.866265670368)*x[1]**o+(-0.66256092454)*x[1]
        arg[1,2]=(-0.83962092333)*x[0]**o+(0.681365396039)*x[0]+(-0.607267339395)*x[1]**o+(0.278637239291)*x[1]
        arg[1,3]=(-0.300965517299)*x[0]**o+(-0.0884159654351)*x[0]+(-0.547465656655)*x[1]**o+(0.708740949324)*x[1]
        arg[1,4]=(-0.0250728833518)*x[0]**o+(0.306828709412)*x[0]+(-0.972269802286)*x[1]**o+(0.619649412235)*x[1]
        arg[2,0]=(-0.873228641325)*x[0]**o+(-0.184974115411)*x[0]+(-0.500142861985)*x[1]**o+(0.708256240395)*x[1]
        arg[2,1]=(-0.588540228523)*x[0]**o+(-0.747378078994)*x[0]+(0.560791728487)*x[1]**o+(-0.619937635414)*x[1]
        arg[2,2]=(-0.738023386417)*x[0]**o+(-0.35311314517)*x[0]+(0.0350349516849)*x[1]**o+(-0.256133328724)*x[1]
        arg[2,3]=(0.836816159205)*x[0]**o+(-0.15869161496)*x[0]+(-0.115472743499)*x[1]**o+(-0.664972307042)*x[1]
        arg[2,4]=(0.257379090396)*x[0]**o+(0.970107472837)*x[0]+(0.323300957751)*x[1]**o+(0.028118478034)*x[1]
        arg[3,0]=(0.198639991638)*x[0]**o+(-0.386442010103)*x[0]+(0.566123916917)*x[1]**o+(0.442943885392)*x[1]
        arg[3,1]=(-0.108796148814)*x[0]**o+(0.896004033429)*x[0]+(-0.860364745295)*x[1]**o+(-0.362632925024)*x[1]
        arg[3,2]=(0.247204027814)*x[0]**o+(0.469777186671)*x[0]+(0.545991233155)*x[1]**o+(-0.320776951704)*x[1]
        arg[3,3]=(-0.98116086999)*x[0]**o+(-0.860611730622)*x[0]+(-0.333595175081)*x[1]**o+(0.437722032729)*x[1]
        arg[3,4]=(0.678428238286)*x[0]**o+(-0.827472935935)*x[0]+(0.355143303916)*x[1]**o+(-0.613985799175)*x[1]
        ref[0,0]=(0.0734861043324)*x_ref[0]**o+(0.698132844974)*x_ref[0]+(-0.0711338889844)*x_ref[1]**o+(0.667313854486)*x_ref[1]
        ref[0,1]=(0.846132724611)*x_ref[0]**o+(-0.215521619475)*x_ref[0]+(-0.920637540655)*x_ref[1]**o+(0.666739063184)*x_ref[1]
        ref[0,2]=(-0.963337851184)*x_ref[0]**o+(-0.961935972833)*x_ref[0]+(-0.0689798144115)*x_ref[1]**o+(0.0106557770119)*x_ref[1]
        ref[0,3]=(0.875063015269)*x_ref[0]**o+(-0.735727431445)*x_ref[0]+(-0.599251007104)*x_ref[1]**o+(-0.274659381946)*x_ref[1]
        ref[0,4]=(-0.378214565543)*x_ref[0]**o+(-0.476230056857)*x_ref[0]+(0.617722084788)*x_ref[1]**o+(-0.808732634287)*x_ref[1]
        ref[1,0]=(-0.324652264865)*x_ref[0]**o+(-0.314957074449)*x_ref[0]+(0.467828791079)*x_ref[1]**o+(-0.105499299461)*x_ref[1]
        ref[1,1]=(-0.0815291997441)*x_ref[0]**o+(0.1473072227)*x_ref[0]+(-0.866265670368)*x_ref[1]**o+(-0.66256092454)*x_ref[1]
        ref[1,2]=(-0.83962092333)*x_ref[0]**o+(0.681365396039)*x_ref[0]+(-0.607267339395)*x_ref[1]**o+(0.278637239291)*x_ref[1]
        ref[1,3]=(-0.300965517299)*x_ref[0]**o+(-0.0884159654351)*x_ref[0]+(-0.547465656655)*x_ref[1]**o+(0.708740949324)*x_ref[1]
        ref[1,4]=(-0.0250728833518)*x_ref[0]**o+(0.306828709412)*x_ref[0]+(-0.972269802286)*x_ref[1]**o+(0.619649412235)*x_ref[1]
        ref[2,0]=(-0.873228641325)*x_ref[0]**o+(-0.184974115411)*x_ref[0]+(-0.500142861985)*x_ref[1]**o+(0.708256240395)*x_ref[1]
        ref[2,1]=(-0.588540228523)*x_ref[0]**o+(-0.747378078994)*x_ref[0]+(0.560791728487)*x_ref[1]**o+(-0.619937635414)*x_ref[1]
        ref[2,2]=(-0.738023386417)*x_ref[0]**o+(-0.35311314517)*x_ref[0]+(0.0350349516849)*x_ref[1]**o+(-0.256133328724)*x_ref[1]
        ref[2,3]=(0.836816159205)*x_ref[0]**o+(-0.15869161496)*x_ref[0]+(-0.115472743499)*x_ref[1]**o+(-0.664972307042)*x_ref[1]
        ref[2,4]=(0.257379090396)*x_ref[0]**o+(0.970107472837)*x_ref[0]+(0.323300957751)*x_ref[1]**o+(0.028118478034)*x_ref[1]
        ref[3,0]=(0.198639991638)*x_ref[0]**o+(-0.386442010103)*x_ref[0]+(0.566123916917)*x_ref[1]**o+(0.442943885392)*x_ref[1]
        ref[3,1]=(-0.108796148814)*x_ref[0]**o+(0.896004033429)*x_ref[0]+(-0.860364745295)*x_ref[1]**o+(-0.362632925024)*x_ref[1]
        ref[3,2]=(0.247204027814)*x_ref[0]**o+(0.469777186671)*x_ref[0]+(0.545991233155)*x_ref[1]**o+(-0.320776951704)*x_ref[1]
        ref[3,3]=(-0.98116086999)*x_ref[0]**o+(-0.860611730622)*x_ref[0]+(-0.333595175081)*x_ref[1]**o+(0.437722032729)*x_ref[1]
        ref[3,4]=(0.678428238286)*x_ref[0]**o+(-0.827472935935)*x_ref[0]+(0.355143303916)*x_ref[1]**o+(-0.613985799175)*x_ref[1]
      else:
        arg[0,0]=(-0.34607777608)*x[0]**o+(0.932253964366)*x[0]+(0.698472742413)*x[1]**o+(-0.661983421645)*x[1]+(-0.965529107787)*x[2]**o+(-0.478528818669)*x[2]
        arg[0,1]=(0.664589900347)*x[0]**o+(-0.0719631882089)*x[0]+(-0.273316024389)*x[1]**o+(-0.887470310768)*x[1]+(-0.362253378888)*x[2]**o+(-0.604337291362)*x[2]
        arg[0,2]=(-0.0505574996951)*x[0]**o+(0.249062378634)*x[0]+(-0.760304066926)*x[1]**o+(-0.314569115107)*x[1]+(-0.679403446488)*x[2]**o+(-0.788567095894)*x[2]
        arg[0,3]=(0.737488025877)*x[0]**o+(0.489265711974)*x[0]+(0.192836157119)*x[1]**o+(-0.599394457037)*x[1]+(-0.972415976641)*x[2]**o+(0.620281802505)*x[2]
        arg[0,4]=(0.830347930734)*x[0]**o+(0.65690057858)*x[0]+(0.756058142792)*x[1]**o+(-0.886073273277)*x[1]+(-0.817193788749)*x[2]**o+(-0.906989660372)*x[2]
        arg[1,0]=(-0.961741874676)*x[0]**o+(0.231117463615)*x[0]+(-0.911422849525)*x[1]**o+(0.616612023918)*x[1]+(-0.563171272809)*x[2]**o+(0.00995177904006)*x[2]
        arg[1,1]=(-0.431814251982)*x[0]**o+(-0.244589543993)*x[0]+(-0.169319798745)*x[1]**o+(0.818064351799)*x[1]+(-0.899145214151)*x[2]**o+(-0.164644770082)*x[2]
        arg[1,2]=(0.149443250669)*x[0]**o+(-0.574248417945)*x[0]+(0.660724984377)*x[1]**o+(0.180990753976)*x[1]+(-0.36059400132)*x[2]**o+(-0.864734906872)*x[2]
        arg[1,3]=(-0.450956708576)*x[0]**o+(0.80934912947)*x[0]+(-0.921267395954)*x[1]**o+(0.449214650071)*x[1]+(-0.833237094175)*x[2]**o+(0.579827382459)*x[2]
        arg[1,4]=(-0.078106995473)*x[0]**o+(0.379085563974)*x[0]+(0.227204387882)*x[1]**o+(0.968199115202)*x[1]+(-0.272353745148)*x[2]**o+(0.380201872298)*x[2]
        arg[2,0]=(-0.852858598591)*x[0]**o+(0.720596990849)*x[0]+(0.281332272808)*x[1]**o+(-0.692404010367)*x[1]+(0.00869602393889)*x[2]**o+(0.0870674708905)*x[2]
        arg[2,1]=(0.466694162196)*x[0]**o+(-0.985083301887)*x[0]+(-0.208640915086)*x[1]**o+(-0.515440367775)*x[1]+(-0.76882243815)*x[2]**o+(0.345968814999)*x[2]
        arg[2,2]=(-0.0459664460618)*x[0]**o+(0.0919391841846)*x[0]+(0.687996833826)*x[1]**o+(0.94270298674)*x[1]+(0.0757885229342)*x[2]**o+(0.60760259541)*x[2]
        arg[2,3]=(0.968401710017)*x[0]**o+(0.0562744847066)*x[0]+(-0.345586665712)*x[1]**o+(-0.190986464506)*x[1]+(-0.202797239715)*x[2]**o+(-0.817657264098)*x[2]
        arg[2,4]=(-0.0288083775852)*x[0]**o+(0.164133034209)*x[0]+(-0.597572578553)*x[1]**o+(-0.669747923644)*x[1]+(-0.77713249687)*x[2]**o+(-0.0955335853718)*x[2]
        arg[3,0]=(-0.173723012835)*x[0]**o+(-0.306216351741)*x[0]+(-0.253810127374)*x[1]**o+(0.145737581929)*x[1]+(-0.950430852252)*x[2]**o+(-0.425608679334)*x[2]
        arg[3,1]=(0.0107577437502)*x[0]**o+(0.0213302772169)*x[0]+(0.403278323078)*x[1]**o+(0.390227936687)*x[1]+(-0.398743418731)*x[2]**o+(0.39090803029)*x[2]
        arg[3,2]=(0.151330768449)*x[0]**o+(0.290347114058)*x[0]+(-0.709672288966)*x[1]**o+(0.379791524303)*x[1]+(-0.403890944944)*x[2]**o+(0.327620813352)*x[2]
        arg[3,3]=(0.377649942493)*x[0]**o+(-0.140974638681)*x[0]+(-0.150965798704)*x[1]**o+(0.203654011715)*x[1]+(-0.70422609343)*x[2]**o+(-0.238707376032)*x[2]
        arg[3,4]=(0.454156920234)*x[0]**o+(-0.0803177888347)*x[0]+(0.325504272926)*x[1]**o+(0.285029937842)*x[1]+(-0.0684685710649)*x[2]**o+(0.471348026862)*x[2]
        ref[0,0]=(-0.34607777608)*x_ref[0]**o+(0.932253964366)*x_ref[0]+(0.698472742413)*x_ref[1]**o+(-0.661983421645)*x_ref[1]+(-0.965529107787)*x_ref[2]**o+(-0.478528818669)*x_ref[2]
        ref[0,1]=(0.664589900347)*x_ref[0]**o+(-0.0719631882089)*x_ref[0]+(-0.273316024389)*x_ref[1]**o+(-0.887470310768)*x_ref[1]+(-0.362253378888)*x_ref[2]**o+(-0.604337291362)*x_ref[2]
        ref[0,2]=(-0.0505574996951)*x_ref[0]**o+(0.249062378634)*x_ref[0]+(-0.760304066926)*x_ref[1]**o+(-0.314569115107)*x_ref[1]+(-0.679403446488)*x_ref[2]**o+(-0.788567095894)*x_ref[2]
        ref[0,3]=(0.737488025877)*x_ref[0]**o+(0.489265711974)*x_ref[0]+(0.192836157119)*x_ref[1]**o+(-0.599394457037)*x_ref[1]+(-0.972415976641)*x_ref[2]**o+(0.620281802505)*x_ref[2]
        ref[0,4]=(0.830347930734)*x_ref[0]**o+(0.65690057858)*x_ref[0]+(0.756058142792)*x_ref[1]**o+(-0.886073273277)*x_ref[1]+(-0.817193788749)*x_ref[2]**o+(-0.906989660372)*x_ref[2]
        ref[1,0]=(-0.961741874676)*x_ref[0]**o+(0.231117463615)*x_ref[0]+(-0.911422849525)*x_ref[1]**o+(0.616612023918)*x_ref[1]+(-0.563171272809)*x_ref[2]**o+(0.00995177904006)*x_ref[2]
        ref[1,1]=(-0.431814251982)*x_ref[0]**o+(-0.244589543993)*x_ref[0]+(-0.169319798745)*x_ref[1]**o+(0.818064351799)*x_ref[1]+(-0.899145214151)*x_ref[2]**o+(-0.164644770082)*x_ref[2]
        ref[1,2]=(0.149443250669)*x_ref[0]**o+(-0.574248417945)*x_ref[0]+(0.660724984377)*x_ref[1]**o+(0.180990753976)*x_ref[1]+(-0.36059400132)*x_ref[2]**o+(-0.864734906872)*x_ref[2]
        ref[1,3]=(-0.450956708576)*x_ref[0]**o+(0.80934912947)*x_ref[0]+(-0.921267395954)*x_ref[1]**o+(0.449214650071)*x_ref[1]+(-0.833237094175)*x_ref[2]**o+(0.579827382459)*x_ref[2]
        ref[1,4]=(-0.078106995473)*x_ref[0]**o+(0.379085563974)*x_ref[0]+(0.227204387882)*x_ref[1]**o+(0.968199115202)*x_ref[1]+(-0.272353745148)*x_ref[2]**o+(0.380201872298)*x_ref[2]
        ref[2,0]=(-0.852858598591)*x_ref[0]**o+(0.720596990849)*x_ref[0]+(0.281332272808)*x_ref[1]**o+(-0.692404010367)*x_ref[1]+(0.00869602393889)*x_ref[2]**o+(0.0870674708905)*x_ref[2]
        ref[2,1]=(0.466694162196)*x_ref[0]**o+(-0.985083301887)*x_ref[0]+(-0.208640915086)*x_ref[1]**o+(-0.515440367775)*x_ref[1]+(-0.76882243815)*x_ref[2]**o+(0.345968814999)*x_ref[2]
        ref[2,2]=(-0.0459664460618)*x_ref[0]**o+(0.0919391841846)*x_ref[0]+(0.687996833826)*x_ref[1]**o+(0.94270298674)*x_ref[1]+(0.0757885229342)*x_ref[2]**o+(0.60760259541)*x_ref[2]
        ref[2,3]=(0.968401710017)*x_ref[0]**o+(0.0562744847066)*x_ref[0]+(-0.345586665712)*x_ref[1]**o+(-0.190986464506)*x_ref[1]+(-0.202797239715)*x_ref[2]**o+(-0.817657264098)*x_ref[2]
        ref[2,4]=(-0.0288083775852)*x_ref[0]**o+(0.164133034209)*x_ref[0]+(-0.597572578553)*x_ref[1]**o+(-0.669747923644)*x_ref[1]+(-0.77713249687)*x_ref[2]**o+(-0.0955335853718)*x_ref[2]
        ref[3,0]=(-0.173723012835)*x_ref[0]**o+(-0.306216351741)*x_ref[0]+(-0.253810127374)*x_ref[1]**o+(0.145737581929)*x_ref[1]+(-0.950430852252)*x_ref[2]**o+(-0.425608679334)*x_ref[2]
        ref[3,1]=(0.0107577437502)*x_ref[0]**o+(0.0213302772169)*x_ref[0]+(0.403278323078)*x_ref[1]**o+(0.390227936687)*x_ref[1]+(-0.398743418731)*x_ref[2]**o+(0.39090803029)*x_ref[2]
        ref[3,2]=(0.151330768449)*x_ref[0]**o+(0.290347114058)*x_ref[0]+(-0.709672288966)*x_ref[1]**o+(0.379791524303)*x_ref[1]+(-0.403890944944)*x_ref[2]**o+(0.327620813352)*x_ref[2]
        ref[3,3]=(0.377649942493)*x_ref[0]**o+(-0.140974638681)*x_ref[0]+(-0.150965798704)*x_ref[1]**o+(0.203654011715)*x_ref[1]+(-0.70422609343)*x_ref[2]**o+(-0.238707376032)*x_ref[2]
        ref[3,4]=(0.454156920234)*x_ref[0]**o+(-0.0803177888347)*x_ref[0]+(0.325504272926)*x_ref[1]**o+(0.285029937842)*x_ref[1]+(-0.0684685710649)*x_ref[2]**o+(0.471348026862)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.754456420402)*x[0]**o+(0.379017389471)*x[0]+(-0.131681017079)*x[1]**o+(0.0844844232338)*x[1]
        arg[0,0,1]=(0.358583374641)*x[0]**o+(-0.237291325422)*x[0]+(-0.636434065455)*x[1]**o+(0.797375166528)*x[1]
        arg[0,1,0]=(0.715400241041)*x[0]**o+(-0.953885532065)*x[0]+(0.505634627788)*x[1]**o+(-0.0943694996143)*x[1]
        arg[0,1,1]=(-0.748388212428)*x[0]**o+(0.645715603392)*x[0]+(0.225587797337)*x[1]**o+(-0.274312832296)*x[1]
        arg[1,0,0]=(-0.696048542115)*x[0]**o+(0.728824156388)*x[0]+(0.475125287255)*x[1]**o+(0.231272864783)*x[1]
        arg[1,0,1]=(-0.171961266224)*x[0]**o+(0.751228523936)*x[0]+(0.0204128165955)*x[1]**o+(0.391751936939)*x[1]
        arg[1,1,0]=(0.702487008399)*x[0]**o+(0.408391530323)*x[0]+(-0.523719800121)*x[1]**o+(-0.142757224099)*x[1]
        arg[1,1,1]=(0.356477993939)*x[0]**o+(-0.552570052325)*x[0]+(-0.728412167408)*x[1]**o+(-0.138741983568)*x[1]
        arg[2,0,0]=(-0.972746877268)*x[0]**o+(-0.563197198314)*x[0]+(0.660003706244)*x[1]**o+(0.0312816821703)*x[1]
        arg[2,0,1]=(0.181440001149)*x[0]**o+(-0.551943857642)*x[0]+(0.550354945981)*x[1]**o+(0.222234118435)*x[1]
        arg[2,1,0]=(-0.843578289392)*x[0]**o+(-0.926317212728)*x[0]+(0.473372460648)*x[1]**o+(0.271910634427)*x[1]
        arg[2,1,1]=(-0.388589229676)*x[0]**o+(-0.117517308339)*x[0]+(0.0211660415594)*x[1]**o+(0.620489966427)*x[1]
        arg[3,0,0]=(-0.259455891555)*x[0]**o+(0.602520729775)*x[0]+(0.755086640963)*x[1]**o+(0.420352546305)*x[1]
        arg[3,0,1]=(0.901413641143)*x[0]**o+(-0.229939015002)*x[0]+(-0.0225626235937)*x[1]**o+(-0.319692416473)*x[1]
        arg[3,1,0]=(-0.398546041269)*x[0]**o+(-0.715449072888)*x[0]+(-0.079576973955)*x[1]**o+(0.132797262272)*x[1]
        arg[3,1,1]=(-0.0790389300192)*x[0]**o+(0.681411537598)*x[0]+(0.311941366364)*x[1]**o+(-0.350457682161)*x[1]
        arg[4,0,0]=(0.046674838542)*x[0]**o+(-0.765876197237)*x[0]+(-0.791075696221)*x[1]**o+(0.0338652019786)*x[1]
        arg[4,0,1]=(-0.00225965167336)*x[0]**o+(-0.773773082204)*x[0]+(0.696918916543)*x[1]**o+(-0.215060949558)*x[1]
        arg[4,1,0]=(-0.38907085938)*x[0]**o+(-0.907941787301)*x[0]+(-0.447045949233)*x[1]**o+(0.780078047474)*x[1]
        arg[4,1,1]=(0.198269028442)*x[0]**o+(0.991680341644)*x[0]+(-0.117854134367)*x[1]**o+(0.818191822875)*x[1]
        arg[5,0,0]=(-0.205861859192)*x[0]**o+(-0.230700784928)*x[0]+(-0.91205826553)*x[1]**o+(-0.563882902438)*x[1]
        arg[5,0,1]=(-0.667447962115)*x[0]**o+(0.0539650266455)*x[0]+(0.917333817791)*x[1]**o+(0.0571269304606)*x[1]
        arg[5,1,0]=(0.69461644672)*x[0]**o+(-0.901397746901)*x[0]+(0.388360725574)*x[1]**o+(-0.833240932275)*x[1]
        arg[5,1,1]=(0.236741248684)*x[0]**o+(0.637532818313)*x[0]+(-0.774742465882)*x[1]**o+(0.895299029069)*x[1]
        ref[0,0,0]=(0.754456420402)*x_ref[0]**o+(0.379017389471)*x_ref[0]+(-0.131681017079)*x_ref[1]**o+(0.0844844232338)*x_ref[1]
        ref[0,0,1]=(0.358583374641)*x_ref[0]**o+(-0.237291325422)*x_ref[0]+(-0.636434065455)*x_ref[1]**o+(0.797375166528)*x_ref[1]
        ref[0,1,0]=(0.715400241041)*x_ref[0]**o+(-0.953885532065)*x_ref[0]+(0.505634627788)*x_ref[1]**o+(-0.0943694996143)*x_ref[1]
        ref[0,1,1]=(-0.748388212428)*x_ref[0]**o+(0.645715603392)*x_ref[0]+(0.225587797337)*x_ref[1]**o+(-0.274312832296)*x_ref[1]
        ref[1,0,0]=(-0.696048542115)*x_ref[0]**o+(0.728824156388)*x_ref[0]+(0.475125287255)*x_ref[1]**o+(0.231272864783)*x_ref[1]
        ref[1,0,1]=(-0.171961266224)*x_ref[0]**o+(0.751228523936)*x_ref[0]+(0.0204128165955)*x_ref[1]**o+(0.391751936939)*x_ref[1]
        ref[1,1,0]=(0.702487008399)*x_ref[0]**o+(0.408391530323)*x_ref[0]+(-0.523719800121)*x_ref[1]**o+(-0.142757224099)*x_ref[1]
        ref[1,1,1]=(0.356477993939)*x_ref[0]**o+(-0.552570052325)*x_ref[0]+(-0.728412167408)*x_ref[1]**o+(-0.138741983568)*x_ref[1]
        ref[2,0,0]=(-0.972746877268)*x_ref[0]**o+(-0.563197198314)*x_ref[0]+(0.660003706244)*x_ref[1]**o+(0.0312816821703)*x_ref[1]
        ref[2,0,1]=(0.181440001149)*x_ref[0]**o+(-0.551943857642)*x_ref[0]+(0.550354945981)*x_ref[1]**o+(0.222234118435)*x_ref[1]
        ref[2,1,0]=(-0.843578289392)*x_ref[0]**o+(-0.926317212728)*x_ref[0]+(0.473372460648)*x_ref[1]**o+(0.271910634427)*x_ref[1]
        ref[2,1,1]=(-0.388589229676)*x_ref[0]**o+(-0.117517308339)*x_ref[0]+(0.0211660415594)*x_ref[1]**o+(0.620489966427)*x_ref[1]
        ref[3,0,0]=(-0.259455891555)*x_ref[0]**o+(0.602520729775)*x_ref[0]+(0.755086640963)*x_ref[1]**o+(0.420352546305)*x_ref[1]
        ref[3,0,1]=(0.901413641143)*x_ref[0]**o+(-0.229939015002)*x_ref[0]+(-0.0225626235937)*x_ref[1]**o+(-0.319692416473)*x_ref[1]
        ref[3,1,0]=(-0.398546041269)*x_ref[0]**o+(-0.715449072888)*x_ref[0]+(-0.079576973955)*x_ref[1]**o+(0.132797262272)*x_ref[1]
        ref[3,1,1]=(-0.0790389300192)*x_ref[0]**o+(0.681411537598)*x_ref[0]+(0.311941366364)*x_ref[1]**o+(-0.350457682161)*x_ref[1]
        ref[4,0,0]=(0.046674838542)*x_ref[0]**o+(-0.765876197237)*x_ref[0]+(-0.791075696221)*x_ref[1]**o+(0.0338652019786)*x_ref[1]
        ref[4,0,1]=(-0.00225965167336)*x_ref[0]**o+(-0.773773082204)*x_ref[0]+(0.696918916543)*x_ref[1]**o+(-0.215060949558)*x_ref[1]
        ref[4,1,0]=(-0.38907085938)*x_ref[0]**o+(-0.907941787301)*x_ref[0]+(-0.447045949233)*x_ref[1]**o+(0.780078047474)*x_ref[1]
        ref[4,1,1]=(0.198269028442)*x_ref[0]**o+(0.991680341644)*x_ref[0]+(-0.117854134367)*x_ref[1]**o+(0.818191822875)*x_ref[1]
        ref[5,0,0]=(-0.205861859192)*x_ref[0]**o+(-0.230700784928)*x_ref[0]+(-0.91205826553)*x_ref[1]**o+(-0.563882902438)*x_ref[1]
        ref[5,0,1]=(-0.667447962115)*x_ref[0]**o+(0.0539650266455)*x_ref[0]+(0.917333817791)*x_ref[1]**o+(0.0571269304606)*x_ref[1]
        ref[5,1,0]=(0.69461644672)*x_ref[0]**o+(-0.901397746901)*x_ref[0]+(0.388360725574)*x_ref[1]**o+(-0.833240932275)*x_ref[1]
        ref[5,1,1]=(0.236741248684)*x_ref[0]**o+(0.637532818313)*x_ref[0]+(-0.774742465882)*x_ref[1]**o+(0.895299029069)*x_ref[1]
      else:
        arg[0,0,0]=(-0.349847065801)*x[0]**o+(-0.373197072565)*x[0]+(-0.744002096082)*x[1]**o+(-0.32885837005)*x[1]+(-0.123733527788)*x[2]**o+(0.0357511439931)*x[2]
        arg[0,0,1]=(0.92390791115)*x[0]**o+(0.359314162684)*x[0]+(-0.344639305985)*x[1]**o+(0.87224607823)*x[1]+(0.603450217818)*x[2]**o+(0.767736049618)*x[2]
        arg[0,1,0]=(0.619626616187)*x[0]**o+(0.0777214591672)*x[0]+(-0.454938621339)*x[1]**o+(0.397372303072)*x[1]+(0.64005891461)*x[2]**o+(0.384837098241)*x[2]
        arg[0,1,1]=(0.694681223586)*x[0]**o+(0.746056110128)*x[0]+(-0.299128956618)*x[1]**o+(-0.411577781871)*x[1]+(0.0769629748738)*x[2]**o+(0.185345614784)*x[2]
        arg[1,0,0]=(-0.686338455095)*x[0]**o+(0.516443141718)*x[0]+(0.0963836215517)*x[1]**o+(-0.142756276631)*x[1]+(0.783984002173)*x[2]**o+(0.453983016774)*x[2]
        arg[1,0,1]=(-0.261515294422)*x[0]**o+(0.193353051852)*x[0]+(0.536103258837)*x[1]**o+(0.739075865062)*x[1]+(0.0854181253451)*x[2]**o+(0.993960851485)*x[2]
        arg[1,1,0]=(-0.041143202263)*x[0]**o+(-0.0483966496614)*x[0]+(0.100238616242)*x[1]**o+(0.229736237164)*x[1]+(-0.427145221147)*x[2]**o+(-0.143755555255)*x[2]
        arg[1,1,1]=(-0.489208229053)*x[0]**o+(-0.339638394047)*x[0]+(-0.0197108849498)*x[1]**o+(-0.64396285961)*x[1]+(0.644185397988)*x[2]**o+(0.952646865856)*x[2]
        arg[2,0,0]=(0.524059731144)*x[0]**o+(0.184311280754)*x[0]+(0.119889599722)*x[1]**o+(0.953938302354)*x[1]+(0.0331959462044)*x[2]**o+(-0.695477785189)*x[2]
        arg[2,0,1]=(0.925889695911)*x[0]**o+(0.612691423308)*x[0]+(0.492933561499)*x[1]**o+(-0.320060666384)*x[1]+(-0.308466393813)*x[2]**o+(0.907998151313)*x[2]
        arg[2,1,0]=(-0.874135284498)*x[0]**o+(0.0151718942733)*x[0]+(0.861045338983)*x[1]**o+(-0.228228955072)*x[1]+(-0.350972290538)*x[2]**o+(0.961075321834)*x[2]
        arg[2,1,1]=(-0.889805119627)*x[0]**o+(0.561790945688)*x[0]+(-0.500273719691)*x[1]**o+(-0.633623450372)*x[1]+(-0.0153487772927)*x[2]**o+(-0.0798904596833)*x[2]
        arg[3,0,0]=(-0.873621421496)*x[0]**o+(0.513411619207)*x[0]+(-0.961906901835)*x[1]**o+(0.321006445811)*x[1]+(-0.592020038506)*x[2]**o+(0.555697967812)*x[2]
        arg[3,0,1]=(0.850390230672)*x[0]**o+(-0.391435099771)*x[0]+(0.700909972778)*x[1]**o+(0.794987613709)*x[1]+(-0.775188800108)*x[2]**o+(-0.132226119354)*x[2]
        arg[3,1,0]=(0.597260092409)*x[0]**o+(0.393550227561)*x[0]+(-0.554683881247)*x[1]**o+(-0.371724214354)*x[1]+(-0.53562488411)*x[2]**o+(-0.555005739246)*x[2]
        arg[3,1,1]=(-0.14964970997)*x[0]**o+(-0.524377521321)*x[0]+(-0.738788452632)*x[1]**o+(-0.832575942733)*x[1]+(-0.766690459508)*x[2]**o+(-0.788273000486)*x[2]
        arg[4,0,0]=(-0.772922450679)*x[0]**o+(-0.0329959778533)*x[0]+(-0.504822565972)*x[1]**o+(-0.0328377918948)*x[1]+(0.598087910099)*x[2]**o+(-0.598910302754)*x[2]
        arg[4,0,1]=(-0.516823078367)*x[0]**o+(-0.783263163828)*x[0]+(0.793230730885)*x[1]**o+(-0.760393659268)*x[1]+(-0.39824557053)*x[2]**o+(0.448151384682)*x[2]
        arg[4,1,0]=(0.95067110965)*x[0]**o+(0.107038151555)*x[0]+(0.986100034409)*x[1]**o+(-0.603396251285)*x[1]+(0.463709896646)*x[2]**o+(-0.0752001209014)*x[2]
        arg[4,1,1]=(0.704778063889)*x[0]**o+(-0.00723361955023)*x[0]+(-0.229254019288)*x[1]**o+(0.972066929322)*x[1]+(-0.792377966898)*x[2]**o+(0.00519477641991)*x[2]
        arg[5,0,0]=(-0.37253642897)*x[0]**o+(-0.283784448234)*x[0]+(0.361753608877)*x[1]**o+(-0.119118907688)*x[1]+(0.403505165985)*x[2]**o+(0.223093000071)*x[2]
        arg[5,0,1]=(-0.78071834268)*x[0]**o+(0.0321159500156)*x[0]+(0.829040628045)*x[1]**o+(0.864378422992)*x[1]+(0.343294472729)*x[2]**o+(-0.502881977207)*x[2]
        arg[5,1,0]=(0.445553133352)*x[0]**o+(0.167144149279)*x[0]+(-0.773477294809)*x[1]**o+(0.400540684182)*x[1]+(-0.806904730971)*x[2]**o+(-0.789278588588)*x[2]
        arg[5,1,1]=(0.290545952984)*x[0]**o+(0.692814030915)*x[0]+(0.176313232933)*x[1]**o+(-0.88461128488)*x[1]+(-0.17903700494)*x[2]**o+(0.448842621603)*x[2]
        ref[0,0,0]=(-0.349847065801)*x_ref[0]**o+(-0.373197072565)*x_ref[0]+(-0.744002096082)*x_ref[1]**o+(-0.32885837005)*x_ref[1]+(-0.123733527788)*x_ref[2]**o+(0.0357511439931)*x_ref[2]
        ref[0,0,1]=(0.92390791115)*x_ref[0]**o+(0.359314162684)*x_ref[0]+(-0.344639305985)*x_ref[1]**o+(0.87224607823)*x_ref[1]+(0.603450217818)*x_ref[2]**o+(0.767736049618)*x_ref[2]
        ref[0,1,0]=(0.619626616187)*x_ref[0]**o+(0.0777214591672)*x_ref[0]+(-0.454938621339)*x_ref[1]**o+(0.397372303072)*x_ref[1]+(0.64005891461)*x_ref[2]**o+(0.384837098241)*x_ref[2]
        ref[0,1,1]=(0.694681223586)*x_ref[0]**o+(0.746056110128)*x_ref[0]+(-0.299128956618)*x_ref[1]**o+(-0.411577781871)*x_ref[1]+(0.0769629748738)*x_ref[2]**o+(0.185345614784)*x_ref[2]
        ref[1,0,0]=(-0.686338455095)*x_ref[0]**o+(0.516443141718)*x_ref[0]+(0.0963836215517)*x_ref[1]**o+(-0.142756276631)*x_ref[1]+(0.783984002173)*x_ref[2]**o+(0.453983016774)*x_ref[2]
        ref[1,0,1]=(-0.261515294422)*x_ref[0]**o+(0.193353051852)*x_ref[0]+(0.536103258837)*x_ref[1]**o+(0.739075865062)*x_ref[1]+(0.0854181253451)*x_ref[2]**o+(0.993960851485)*x_ref[2]
        ref[1,1,0]=(-0.041143202263)*x_ref[0]**o+(-0.0483966496614)*x_ref[0]+(0.100238616242)*x_ref[1]**o+(0.229736237164)*x_ref[1]+(-0.427145221147)*x_ref[2]**o+(-0.143755555255)*x_ref[2]
        ref[1,1,1]=(-0.489208229053)*x_ref[0]**o+(-0.339638394047)*x_ref[0]+(-0.0197108849498)*x_ref[1]**o+(-0.64396285961)*x_ref[1]+(0.644185397988)*x_ref[2]**o+(0.952646865856)*x_ref[2]
        ref[2,0,0]=(0.524059731144)*x_ref[0]**o+(0.184311280754)*x_ref[0]+(0.119889599722)*x_ref[1]**o+(0.953938302354)*x_ref[1]+(0.0331959462044)*x_ref[2]**o+(-0.695477785189)*x_ref[2]
        ref[2,0,1]=(0.925889695911)*x_ref[0]**o+(0.612691423308)*x_ref[0]+(0.492933561499)*x_ref[1]**o+(-0.320060666384)*x_ref[1]+(-0.308466393813)*x_ref[2]**o+(0.907998151313)*x_ref[2]
        ref[2,1,0]=(-0.874135284498)*x_ref[0]**o+(0.0151718942733)*x_ref[0]+(0.861045338983)*x_ref[1]**o+(-0.228228955072)*x_ref[1]+(-0.350972290538)*x_ref[2]**o+(0.961075321834)*x_ref[2]
        ref[2,1,1]=(-0.889805119627)*x_ref[0]**o+(0.561790945688)*x_ref[0]+(-0.500273719691)*x_ref[1]**o+(-0.633623450372)*x_ref[1]+(-0.0153487772927)*x_ref[2]**o+(-0.0798904596833)*x_ref[2]
        ref[3,0,0]=(-0.873621421496)*x_ref[0]**o+(0.513411619207)*x_ref[0]+(-0.961906901835)*x_ref[1]**o+(0.321006445811)*x_ref[1]+(-0.592020038506)*x_ref[2]**o+(0.555697967812)*x_ref[2]
        ref[3,0,1]=(0.850390230672)*x_ref[0]**o+(-0.391435099771)*x_ref[0]+(0.700909972778)*x_ref[1]**o+(0.794987613709)*x_ref[1]+(-0.775188800108)*x_ref[2]**o+(-0.132226119354)*x_ref[2]
        ref[3,1,0]=(0.597260092409)*x_ref[0]**o+(0.393550227561)*x_ref[0]+(-0.554683881247)*x_ref[1]**o+(-0.371724214354)*x_ref[1]+(-0.53562488411)*x_ref[2]**o+(-0.555005739246)*x_ref[2]
        ref[3,1,1]=(-0.14964970997)*x_ref[0]**o+(-0.524377521321)*x_ref[0]+(-0.738788452632)*x_ref[1]**o+(-0.832575942733)*x_ref[1]+(-0.766690459508)*x_ref[2]**o+(-0.788273000486)*x_ref[2]
        ref[4,0,0]=(-0.772922450679)*x_ref[0]**o+(-0.0329959778533)*x_ref[0]+(-0.504822565972)*x_ref[1]**o+(-0.0328377918948)*x_ref[1]+(0.598087910099)*x_ref[2]**o+(-0.598910302754)*x_ref[2]
        ref[4,0,1]=(-0.516823078367)*x_ref[0]**o+(-0.783263163828)*x_ref[0]+(0.793230730885)*x_ref[1]**o+(-0.760393659268)*x_ref[1]+(-0.39824557053)*x_ref[2]**o+(0.448151384682)*x_ref[2]
        ref[4,1,0]=(0.95067110965)*x_ref[0]**o+(0.107038151555)*x_ref[0]+(0.986100034409)*x_ref[1]**o+(-0.603396251285)*x_ref[1]+(0.463709896646)*x_ref[2]**o+(-0.0752001209014)*x_ref[2]
        ref[4,1,1]=(0.704778063889)*x_ref[0]**o+(-0.00723361955023)*x_ref[0]+(-0.229254019288)*x_ref[1]**o+(0.972066929322)*x_ref[1]+(-0.792377966898)*x_ref[2]**o+(0.00519477641991)*x_ref[2]
        ref[5,0,0]=(-0.37253642897)*x_ref[0]**o+(-0.283784448234)*x_ref[0]+(0.361753608877)*x_ref[1]**o+(-0.119118907688)*x_ref[1]+(0.403505165985)*x_ref[2]**o+(0.223093000071)*x_ref[2]
        ref[5,0,1]=(-0.78071834268)*x_ref[0]**o+(0.0321159500156)*x_ref[0]+(0.829040628045)*x_ref[1]**o+(0.864378422992)*x_ref[1]+(0.343294472729)*x_ref[2]**o+(-0.502881977207)*x_ref[2]
        ref[5,1,0]=(0.445553133352)*x_ref[0]**o+(0.167144149279)*x_ref[0]+(-0.773477294809)*x_ref[1]**o+(0.400540684182)*x_ref[1]+(-0.806904730971)*x_ref[2]**o+(-0.789278588588)*x_ref[2]
        ref[5,1,1]=(0.290545952984)*x_ref[0]**o+(0.692814030915)*x_ref[0]+(0.176313232933)*x_ref[1]**o+(-0.88461128488)*x_ref[1]+(-0.17903700494)*x_ref[2]**o+(0.448842621603)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.355090869915)*x[0]**o+(0.62333594265)*x[0]+(0.127393746399)*x[1]**o+(0.938815550559)*x[1]
        arg[0,0,0,1]=(0.361108442937)*x[0]**o+(0.848759016468)*x[0]+(-0.100173964534)*x[1]**o+(-0.798463736192)*x[1]
        arg[0,0,1,0]=(0.85312225866)*x[0]**o+(0.862709543579)*x[0]+(0.0794424365586)*x[1]**o+(-0.798377510326)*x[1]
        arg[0,0,1,1]=(-0.370853792692)*x[0]**o+(-0.967610789809)*x[0]+(0.277035447492)*x[1]**o+(-0.880157165176)*x[1]
        arg[0,0,2,0]=(-0.717702490059)*x[0]**o+(0.701614319774)*x[0]+(0.870301462626)*x[1]**o+(0.97842528383)*x[1]
        arg[0,0,2,1]=(0.398938846238)*x[0]**o+(0.460600049445)*x[0]+(-0.719014936029)*x[1]**o+(0.909351103488)*x[1]
        arg[0,1,0,0]=(-0.607340194261)*x[0]**o+(-0.851772216829)*x[0]+(0.7593824952)*x[1]**o+(-0.621669238017)*x[1]
        arg[0,1,0,1]=(-0.769916079359)*x[0]**o+(0.628665644279)*x[0]+(0.198700163204)*x[1]**o+(-0.878872706777)*x[1]
        arg[0,1,1,0]=(-0.133970652198)*x[0]**o+(0.817475343341)*x[0]+(-0.553628012005)*x[1]**o+(-0.780184324119)*x[1]
        arg[0,1,1,1]=(-0.91240774502)*x[0]**o+(-0.438683032351)*x[0]+(-0.651970251415)*x[1]**o+(-0.985718981791)*x[1]
        arg[0,1,2,0]=(-0.0894695484803)*x[0]**o+(0.230969119002)*x[0]+(-0.181318345335)*x[1]**o+(-0.684304266075)*x[1]
        arg[0,1,2,1]=(0.116582610294)*x[0]**o+(0.508506980325)*x[0]+(-0.227569256068)*x[1]**o+(-0.743396388621)*x[1]
        arg[0,2,0,0]=(0.163058551784)*x[0]**o+(-0.819992657695)*x[0]+(-0.0400315416665)*x[1]**o+(-0.368673546217)*x[1]
        arg[0,2,0,1]=(-0.760195093607)*x[0]**o+(-0.186788923915)*x[0]+(0.0439683490286)*x[1]**o+(-0.329021652328)*x[1]
        arg[0,2,1,0]=(-0.740086957456)*x[0]**o+(0.67648478989)*x[0]+(0.233575048863)*x[1]**o+(-0.353142296901)*x[1]
        arg[0,2,1,1]=(0.0374034623006)*x[0]**o+(-0.912304735553)*x[0]+(-0.982504180489)*x[1]**o+(0.388144019319)*x[1]
        arg[0,2,2,0]=(-0.812061018965)*x[0]**o+(0.641603736452)*x[0]+(0.260352725676)*x[1]**o+(-0.361534248663)*x[1]
        arg[0,2,2,1]=(0.0783398648132)*x[0]**o+(0.276408362801)*x[0]+(-0.107410956411)*x[1]**o+(-0.0733118712953)*x[1]
        arg[0,3,0,0]=(-0.085393595974)*x[0]**o+(0.545214990205)*x[0]+(0.451103178758)*x[1]**o+(0.741889771181)*x[1]
        arg[0,3,0,1]=(-0.377253694991)*x[0]**o+(0.243160465116)*x[0]+(-0.330892504638)*x[1]**o+(0.0688453403018)*x[1]
        arg[0,3,1,0]=(0.569018176673)*x[0]**o+(-0.519384243562)*x[0]+(0.807469891684)*x[1]**o+(-0.511601914558)*x[1]
        arg[0,3,1,1]=(-0.624467771068)*x[0]**o+(0.722555599188)*x[0]+(0.165243550597)*x[1]**o+(0.124887773794)*x[1]
        arg[0,3,2,0]=(0.846969233067)*x[0]**o+(0.143910732534)*x[0]+(-0.933748958347)*x[1]**o+(-0.468985899594)*x[1]
        arg[0,3,2,1]=(-0.00559860682228)*x[0]**o+(0.6265077144)*x[0]+(-0.597502391816)*x[1]**o+(-0.0269950883749)*x[1]
        arg[0,4,0,0]=(-0.929830228823)*x[0]**o+(0.533173796675)*x[0]+(-0.513355534671)*x[1]**o+(0.739452877653)*x[1]
        arg[0,4,0,1]=(0.573157694447)*x[0]**o+(0.533950659465)*x[0]+(-0.470668156002)*x[1]**o+(-0.267083337629)*x[1]
        arg[0,4,1,0]=(-0.786758281623)*x[0]**o+(-0.559211099907)*x[0]+(-0.981412883007)*x[1]**o+(-0.760247332612)*x[1]
        arg[0,4,1,1]=(0.811476615714)*x[0]**o+(-0.483217733654)*x[0]+(-0.473107949588)*x[1]**o+(-0.388655922158)*x[1]
        arg[0,4,2,0]=(0.743903538516)*x[0]**o+(-0.0914243557971)*x[0]+(0.523767499227)*x[1]**o+(0.105519926963)*x[1]
        arg[0,4,2,1]=(-0.425629408466)*x[0]**o+(0.294957108568)*x[0]+(-0.248527445843)*x[1]**o+(0.343947931379)*x[1]
        arg[1,0,0,0]=(0.413841637504)*x[0]**o+(0.609974192806)*x[0]+(-0.0785287734637)*x[1]**o+(0.538510227782)*x[1]
        arg[1,0,0,1]=(-0.905065001974)*x[0]**o+(-0.440318385611)*x[0]+(-0.0108293746138)*x[1]**o+(0.132125994317)*x[1]
        arg[1,0,1,0]=(0.212448400234)*x[0]**o+(0.117680077385)*x[0]+(0.501566152733)*x[1]**o+(-0.0914837162066)*x[1]
        arg[1,0,1,1]=(-0.763152636104)*x[0]**o+(0.318595673875)*x[0]+(-0.166759749724)*x[1]**o+(0.742420099653)*x[1]
        arg[1,0,2,0]=(-0.711420698498)*x[0]**o+(-0.0777910925778)*x[0]+(-0.885593755559)*x[1]**o+(0.302761673106)*x[1]
        arg[1,0,2,1]=(0.980964252534)*x[0]**o+(-0.890328597184)*x[0]+(-0.230421339981)*x[1]**o+(-0.656523995481)*x[1]
        arg[1,1,0,0]=(0.376070672474)*x[0]**o+(0.940362151504)*x[0]+(-0.297285488825)*x[1]**o+(0.451414529855)*x[1]
        arg[1,1,0,1]=(0.867987607318)*x[0]**o+(0.0477620815134)*x[0]+(0.257575148838)*x[1]**o+(-0.54187477723)*x[1]
        arg[1,1,1,0]=(-0.0643073021305)*x[0]**o+(0.20805724465)*x[0]+(-0.16501072904)*x[1]**o+(-0.341858346811)*x[1]
        arg[1,1,1,1]=(0.317830146671)*x[0]**o+(-0.307344984878)*x[0]+(0.170352713722)*x[1]**o+(-0.219561938951)*x[1]
        arg[1,1,2,0]=(0.89815928322)*x[0]**o+(0.51678942721)*x[0]+(0.897095493092)*x[1]**o+(-0.732217042123)*x[1]
        arg[1,1,2,1]=(0.0216578477569)*x[0]**o+(-0.942894641381)*x[0]+(-0.475731377321)*x[1]**o+(-0.753173025139)*x[1]
        arg[1,2,0,0]=(-0.897860515192)*x[0]**o+(0.833870034804)*x[0]+(0.554274053224)*x[1]**o+(-0.785991934149)*x[1]
        arg[1,2,0,1]=(-0.0648893111877)*x[0]**o+(0.173757424631)*x[0]+(0.860427552735)*x[1]**o+(0.139897107451)*x[1]
        arg[1,2,1,0]=(-0.146915585006)*x[0]**o+(-0.290881625114)*x[0]+(0.98541624122)*x[1]**o+(0.868617044417)*x[1]
        arg[1,2,1,1]=(0.17817541898)*x[0]**o+(-0.0112388444881)*x[0]+(0.646731986079)*x[1]**o+(0.653380181018)*x[1]
        arg[1,2,2,0]=(0.208250797377)*x[0]**o+(0.362989066599)*x[0]+(-0.384302232322)*x[1]**o+(0.376460044532)*x[1]
        arg[1,2,2,1]=(0.0437567097332)*x[0]**o+(0.739696851008)*x[0]+(0.0308848949312)*x[1]**o+(-0.445528203988)*x[1]
        arg[1,3,0,0]=(-0.257780793962)*x[0]**o+(0.840582446797)*x[0]+(0.599894892775)*x[1]**o+(-0.744691829118)*x[1]
        arg[1,3,0,1]=(0.913814269186)*x[0]**o+(0.295261304054)*x[0]+(0.0934778065196)*x[1]**o+(0.567115895371)*x[1]
        arg[1,3,1,0]=(-0.399785878199)*x[0]**o+(0.0336841328611)*x[0]+(0.236101490273)*x[1]**o+(0.915159045152)*x[1]
        arg[1,3,1,1]=(-0.955438743542)*x[0]**o+(-0.457060494715)*x[0]+(0.582140402321)*x[1]**o+(-0.999341752789)*x[1]
        arg[1,3,2,0]=(0.0206746776356)*x[0]**o+(-0.307327438556)*x[0]+(-0.646238381067)*x[1]**o+(0.248313182827)*x[1]
        arg[1,3,2,1]=(-0.10971097551)*x[0]**o+(-0.825303025298)*x[0]+(0.876884308015)*x[1]**o+(0.53621927318)*x[1]
        arg[1,4,0,0]=(0.833490742222)*x[0]**o+(0.865604127412)*x[0]+(0.627418353815)*x[1]**o+(0.523133814878)*x[1]
        arg[1,4,0,1]=(0.57616288545)*x[0]**o+(0.438552559891)*x[0]+(0.0256681709722)*x[1]**o+(-0.738314027898)*x[1]
        arg[1,4,1,0]=(0.319863593844)*x[0]**o+(0.177342228601)*x[0]+(0.0865784775196)*x[1]**o+(0.642550298518)*x[1]
        arg[1,4,1,1]=(-0.0141646689097)*x[0]**o+(0.317108687762)*x[0]+(0.188163493751)*x[1]**o+(0.990958509748)*x[1]
        arg[1,4,2,0]=(0.199211769035)*x[0]**o+(-0.0280132568124)*x[0]+(0.144960116505)*x[1]**o+(0.226821898255)*x[1]
        arg[1,4,2,1]=(-0.892464680507)*x[0]**o+(-0.60166056141)*x[0]+(-0.492226295363)*x[1]**o+(0.189924737916)*x[1]
        arg[2,0,0,0]=(0.329199148463)*x[0]**o+(-0.345343145094)*x[0]+(-0.22723822677)*x[1]**o+(-0.214232512114)*x[1]
        arg[2,0,0,1]=(-0.195525660216)*x[0]**o+(0.323518517888)*x[0]+(-0.116808991213)*x[1]**o+(0.22923091239)*x[1]
        arg[2,0,1,0]=(-0.826978585914)*x[0]**o+(0.641345887454)*x[0]+(-0.335583694811)*x[1]**o+(0.168685506673)*x[1]
        arg[2,0,1,1]=(-0.756460980453)*x[0]**o+(-0.9181609332)*x[0]+(0.0735111723828)*x[1]**o+(0.312524164436)*x[1]
        arg[2,0,2,0]=(0.297314747638)*x[0]**o+(0.522844740561)*x[0]+(0.890404335261)*x[1]**o+(-0.158861326888)*x[1]
        arg[2,0,2,1]=(-0.280311121771)*x[0]**o+(0.561223136791)*x[0]+(-0.952737235492)*x[1]**o+(0.581165512883)*x[1]
        arg[2,1,0,0]=(-0.202509823332)*x[0]**o+(0.298626146584)*x[0]+(0.533843705865)*x[1]**o+(0.652127174748)*x[1]
        arg[2,1,0,1]=(-0.349734740031)*x[0]**o+(-0.0589736478602)*x[0]+(0.741348559428)*x[1]**o+(0.993586735262)*x[1]
        arg[2,1,1,0]=(-0.350986159862)*x[0]**o+(-0.814566621733)*x[0]+(0.0556573702074)*x[1]**o+(0.363924401853)*x[1]
        arg[2,1,1,1]=(-0.562086543404)*x[0]**o+(0.126278013645)*x[0]+(-0.488684138283)*x[1]**o+(-0.990773306296)*x[1]
        arg[2,1,2,0]=(-0.873707667752)*x[0]**o+(-0.00610530100216)*x[0]+(-0.927847882023)*x[1]**o+(0.255850655097)*x[1]
        arg[2,1,2,1]=(0.683462220662)*x[0]**o+(-0.515973835798)*x[0]+(-0.676735030671)*x[1]**o+(0.704690967291)*x[1]
        arg[2,2,0,0]=(-0.0891987051347)*x[0]**o+(0.783509807114)*x[0]+(0.583865611337)*x[1]**o+(-0.373096806492)*x[1]
        arg[2,2,0,1]=(0.606443213957)*x[0]**o+(0.887055818748)*x[0]+(-0.281844491159)*x[1]**o+(0.420965776696)*x[1]
        arg[2,2,1,0]=(0.205051077925)*x[0]**o+(0.267690802839)*x[0]+(-0.821781634855)*x[1]**o+(-0.903592743179)*x[1]
        arg[2,2,1,1]=(-0.0438536252522)*x[0]**o+(-0.870635324748)*x[0]+(-0.238830174768)*x[1]**o+(-0.230782152037)*x[1]
        arg[2,2,2,0]=(-0.94437182489)*x[0]**o+(0.0666311523262)*x[0]+(0.481913247572)*x[1]**o+(-0.646415708448)*x[1]
        arg[2,2,2,1]=(0.443901323099)*x[0]**o+(0.349352493291)*x[0]+(-0.158789593075)*x[1]**o+(-0.603521202861)*x[1]
        arg[2,3,0,0]=(0.716829455981)*x[0]**o+(0.0186823101139)*x[0]+(-0.209053562063)*x[1]**o+(0.847906310296)*x[1]
        arg[2,3,0,1]=(0.305837642219)*x[0]**o+(0.138301301592)*x[0]+(0.0645236431399)*x[1]**o+(0.909641719567)*x[1]
        arg[2,3,1,0]=(-0.0177501904561)*x[0]**o+(0.177278703549)*x[0]+(-0.445915965222)*x[1]**o+(-0.894038892768)*x[1]
        arg[2,3,1,1]=(0.00187039478357)*x[0]**o+(-0.598690809424)*x[0]+(0.299307927538)*x[1]**o+(-0.882235917926)*x[1]
        arg[2,3,2,0]=(0.730056097248)*x[0]**o+(-0.307654523095)*x[0]+(0.246914489171)*x[1]**o+(-0.60035541725)*x[1]
        arg[2,3,2,1]=(-0.608344323894)*x[0]**o+(-0.081794637624)*x[0]+(0.099524567015)*x[1]**o+(-0.0654861171366)*x[1]
        arg[2,4,0,0]=(-0.225654564715)*x[0]**o+(-0.529959598356)*x[0]+(-0.890574962317)*x[1]**o+(0.994618025898)*x[1]
        arg[2,4,0,1]=(-0.231311865123)*x[0]**o+(0.479613722979)*x[0]+(-0.614977654518)*x[1]**o+(0.804794516306)*x[1]
        arg[2,4,1,0]=(0.365865384972)*x[0]**o+(-0.889202382492)*x[0]+(-0.300736557219)*x[1]**o+(0.838201123405)*x[1]
        arg[2,4,1,1]=(-0.809876332779)*x[0]**o+(0.1890391825)*x[0]+(-0.607869144071)*x[1]**o+(0.0448966585205)*x[1]
        arg[2,4,2,0]=(-0.83080984753)*x[0]**o+(0.305865817658)*x[0]+(-0.359639965091)*x[1]**o+(-0.288164510937)*x[1]
        arg[2,4,2,1]=(0.616022940224)*x[0]**o+(-0.636356653421)*x[0]+(0.693033289027)*x[1]**o+(-0.780662522342)*x[1]
        arg[3,0,0,0]=(0.985497135193)*x[0]**o+(-0.441259417159)*x[0]+(0.0550399586931)*x[1]**o+(-0.401551085867)*x[1]
        arg[3,0,0,1]=(0.66919872999)*x[0]**o+(-0.21372913425)*x[0]+(-0.904643614761)*x[1]**o+(0.254555219564)*x[1]
        arg[3,0,1,0]=(0.342630495616)*x[0]**o+(-0.688520932208)*x[0]+(-0.795149453533)*x[1]**o+(-0.532353008569)*x[1]
        arg[3,0,1,1]=(-0.157354065395)*x[0]**o+(-0.574302953359)*x[0]+(-0.0505170558232)*x[1]**o+(-0.916081229428)*x[1]
        arg[3,0,2,0]=(-0.276092076819)*x[0]**o+(0.755203166655)*x[0]+(0.71481609597)*x[1]**o+(-0.623929271337)*x[1]
        arg[3,0,2,1]=(0.901066640027)*x[0]**o+(0.615459930577)*x[0]+(-0.65213904597)*x[1]**o+(0.52496253783)*x[1]
        arg[3,1,0,0]=(-0.90847876416)*x[0]**o+(0.787519284245)*x[0]+(0.260252163627)*x[1]**o+(0.24849374459)*x[1]
        arg[3,1,0,1]=(-0.519950434741)*x[0]**o+(-0.146327164967)*x[0]+(0.199399508881)*x[1]**o+(0.583497630868)*x[1]
        arg[3,1,1,0]=(-0.956798390834)*x[0]**o+(0.145681684376)*x[0]+(0.308458821145)*x[1]**o+(-0.759589804755)*x[1]
        arg[3,1,1,1]=(-0.195569939954)*x[0]**o+(-0.544313649532)*x[0]+(0.387017502787)*x[1]**o+(0.174663629204)*x[1]
        arg[3,1,2,0]=(-0.3697024469)*x[0]**o+(0.356841678749)*x[0]+(0.922691790563)*x[1]**o+(-0.906755290703)*x[1]
        arg[3,1,2,1]=(-0.343599294558)*x[0]**o+(-0.213227630549)*x[0]+(-0.822516525353)*x[1]**o+(-0.0350701846979)*x[1]
        arg[3,2,0,0]=(-0.671120249859)*x[0]**o+(-0.409571854187)*x[0]+(-0.948171208395)*x[1]**o+(0.00128117529499)*x[1]
        arg[3,2,0,1]=(0.573086687235)*x[0]**o+(0.272624522478)*x[0]+(0.404610375833)*x[1]**o+(0.177742792198)*x[1]
        arg[3,2,1,0]=(-0.0481585168757)*x[0]**o+(0.108306965224)*x[0]+(0.962229835277)*x[1]**o+(0.457705229155)*x[1]
        arg[3,2,1,1]=(0.843866939471)*x[0]**o+(-0.209955643474)*x[0]+(0.861076954064)*x[1]**o+(-0.0920496448252)*x[1]
        arg[3,2,2,0]=(-0.482466555781)*x[0]**o+(0.905888895109)*x[0]+(0.969984046224)*x[1]**o+(0.779155706115)*x[1]
        arg[3,2,2,1]=(-0.119059605916)*x[0]**o+(-0.413165838565)*x[0]+(-0.458952085266)*x[1]**o+(0.667357851674)*x[1]
        arg[3,3,0,0]=(-0.139020214893)*x[0]**o+(-0.430716093157)*x[0]+(0.587151976237)*x[1]**o+(0.538779583614)*x[1]
        arg[3,3,0,1]=(-0.196698073225)*x[0]**o+(0.28357504894)*x[0]+(0.628834405238)*x[1]**o+(0.991210138395)*x[1]
        arg[3,3,1,0]=(-0.375644375731)*x[0]**o+(0.889016565145)*x[0]+(-0.495700828743)*x[1]**o+(-0.498859003938)*x[1]
        arg[3,3,1,1]=(0.783773071811)*x[0]**o+(-0.0875026023418)*x[0]+(-0.421213063239)*x[1]**o+(0.750220508147)*x[1]
        arg[3,3,2,0]=(0.211305046835)*x[0]**o+(-0.828436860631)*x[0]+(0.225275220049)*x[1]**o+(0.264723427485)*x[1]
        arg[3,3,2,1]=(0.582818095069)*x[0]**o+(0.497614150982)*x[0]+(-0.445084955033)*x[1]**o+(-0.600656630963)*x[1]
        arg[3,4,0,0]=(-0.886999530258)*x[0]**o+(0.386008089061)*x[0]+(-0.292729716725)*x[1]**o+(-0.159269788338)*x[1]
        arg[3,4,0,1]=(0.260932511148)*x[0]**o+(-0.533553803234)*x[0]+(0.174946866584)*x[1]**o+(-0.17701796222)*x[1]
        arg[3,4,1,0]=(0.37985010505)*x[0]**o+(0.934809225579)*x[0]+(-0.76271501884)*x[1]**o+(0.171072659)*x[1]
        arg[3,4,1,1]=(-0.369797459679)*x[0]**o+(-0.714539720717)*x[0]+(-0.122698805149)*x[1]**o+(0.874232715421)*x[1]
        arg[3,4,2,0]=(0.912569854856)*x[0]**o+(-0.773508960793)*x[0]+(-0.447748330632)*x[1]**o+(-0.0117517938995)*x[1]
        arg[3,4,2,1]=(-0.467442835658)*x[0]**o+(-0.792740521854)*x[0]+(-0.33441666815)*x[1]**o+(-0.821650769855)*x[1]
        ref[0,0,0,0]=(0.355090869915)*x_ref[0]**o+(0.62333594265)*x_ref[0]+(0.127393746399)*x_ref[1]**o+(0.938815550559)*x_ref[1]
        ref[0,0,0,1]=(0.361108442937)*x_ref[0]**o+(0.848759016468)*x_ref[0]+(-0.100173964534)*x_ref[1]**o+(-0.798463736192)*x_ref[1]
        ref[0,0,1,0]=(0.85312225866)*x_ref[0]**o+(0.862709543579)*x_ref[0]+(0.0794424365586)*x_ref[1]**o+(-0.798377510326)*x_ref[1]
        ref[0,0,1,1]=(-0.370853792692)*x_ref[0]**o+(-0.967610789809)*x_ref[0]+(0.277035447492)*x_ref[1]**o+(-0.880157165176)*x_ref[1]
        ref[0,0,2,0]=(-0.717702490059)*x_ref[0]**o+(0.701614319774)*x_ref[0]+(0.870301462626)*x_ref[1]**o+(0.97842528383)*x_ref[1]
        ref[0,0,2,1]=(0.398938846238)*x_ref[0]**o+(0.460600049445)*x_ref[0]+(-0.719014936029)*x_ref[1]**o+(0.909351103488)*x_ref[1]
        ref[0,1,0,0]=(-0.607340194261)*x_ref[0]**o+(-0.851772216829)*x_ref[0]+(0.7593824952)*x_ref[1]**o+(-0.621669238017)*x_ref[1]
        ref[0,1,0,1]=(-0.769916079359)*x_ref[0]**o+(0.628665644279)*x_ref[0]+(0.198700163204)*x_ref[1]**o+(-0.878872706777)*x_ref[1]
        ref[0,1,1,0]=(-0.133970652198)*x_ref[0]**o+(0.817475343341)*x_ref[0]+(-0.553628012005)*x_ref[1]**o+(-0.780184324119)*x_ref[1]
        ref[0,1,1,1]=(-0.91240774502)*x_ref[0]**o+(-0.438683032351)*x_ref[0]+(-0.651970251415)*x_ref[1]**o+(-0.985718981791)*x_ref[1]
        ref[0,1,2,0]=(-0.0894695484803)*x_ref[0]**o+(0.230969119002)*x_ref[0]+(-0.181318345335)*x_ref[1]**o+(-0.684304266075)*x_ref[1]
        ref[0,1,2,1]=(0.116582610294)*x_ref[0]**o+(0.508506980325)*x_ref[0]+(-0.227569256068)*x_ref[1]**o+(-0.743396388621)*x_ref[1]
        ref[0,2,0,0]=(0.163058551784)*x_ref[0]**o+(-0.819992657695)*x_ref[0]+(-0.0400315416665)*x_ref[1]**o+(-0.368673546217)*x_ref[1]
        ref[0,2,0,1]=(-0.760195093607)*x_ref[0]**o+(-0.186788923915)*x_ref[0]+(0.0439683490286)*x_ref[1]**o+(-0.329021652328)*x_ref[1]
        ref[0,2,1,0]=(-0.740086957456)*x_ref[0]**o+(0.67648478989)*x_ref[0]+(0.233575048863)*x_ref[1]**o+(-0.353142296901)*x_ref[1]
        ref[0,2,1,1]=(0.0374034623006)*x_ref[0]**o+(-0.912304735553)*x_ref[0]+(-0.982504180489)*x_ref[1]**o+(0.388144019319)*x_ref[1]
        ref[0,2,2,0]=(-0.812061018965)*x_ref[0]**o+(0.641603736452)*x_ref[0]+(0.260352725676)*x_ref[1]**o+(-0.361534248663)*x_ref[1]
        ref[0,2,2,1]=(0.0783398648132)*x_ref[0]**o+(0.276408362801)*x_ref[0]+(-0.107410956411)*x_ref[1]**o+(-0.0733118712953)*x_ref[1]
        ref[0,3,0,0]=(-0.085393595974)*x_ref[0]**o+(0.545214990205)*x_ref[0]+(0.451103178758)*x_ref[1]**o+(0.741889771181)*x_ref[1]
        ref[0,3,0,1]=(-0.377253694991)*x_ref[0]**o+(0.243160465116)*x_ref[0]+(-0.330892504638)*x_ref[1]**o+(0.0688453403018)*x_ref[1]
        ref[0,3,1,0]=(0.569018176673)*x_ref[0]**o+(-0.519384243562)*x_ref[0]+(0.807469891684)*x_ref[1]**o+(-0.511601914558)*x_ref[1]
        ref[0,3,1,1]=(-0.624467771068)*x_ref[0]**o+(0.722555599188)*x_ref[0]+(0.165243550597)*x_ref[1]**o+(0.124887773794)*x_ref[1]
        ref[0,3,2,0]=(0.846969233067)*x_ref[0]**o+(0.143910732534)*x_ref[0]+(-0.933748958347)*x_ref[1]**o+(-0.468985899594)*x_ref[1]
        ref[0,3,2,1]=(-0.00559860682228)*x_ref[0]**o+(0.6265077144)*x_ref[0]+(-0.597502391816)*x_ref[1]**o+(-0.0269950883749)*x_ref[1]
        ref[0,4,0,0]=(-0.929830228823)*x_ref[0]**o+(0.533173796675)*x_ref[0]+(-0.513355534671)*x_ref[1]**o+(0.739452877653)*x_ref[1]
        ref[0,4,0,1]=(0.573157694447)*x_ref[0]**o+(0.533950659465)*x_ref[0]+(-0.470668156002)*x_ref[1]**o+(-0.267083337629)*x_ref[1]
        ref[0,4,1,0]=(-0.786758281623)*x_ref[0]**o+(-0.559211099907)*x_ref[0]+(-0.981412883007)*x_ref[1]**o+(-0.760247332612)*x_ref[1]
        ref[0,4,1,1]=(0.811476615714)*x_ref[0]**o+(-0.483217733654)*x_ref[0]+(-0.473107949588)*x_ref[1]**o+(-0.388655922158)*x_ref[1]
        ref[0,4,2,0]=(0.743903538516)*x_ref[0]**o+(-0.0914243557971)*x_ref[0]+(0.523767499227)*x_ref[1]**o+(0.105519926963)*x_ref[1]
        ref[0,4,2,1]=(-0.425629408466)*x_ref[0]**o+(0.294957108568)*x_ref[0]+(-0.248527445843)*x_ref[1]**o+(0.343947931379)*x_ref[1]
        ref[1,0,0,0]=(0.413841637504)*x_ref[0]**o+(0.609974192806)*x_ref[0]+(-0.0785287734637)*x_ref[1]**o+(0.538510227782)*x_ref[1]
        ref[1,0,0,1]=(-0.905065001974)*x_ref[0]**o+(-0.440318385611)*x_ref[0]+(-0.0108293746138)*x_ref[1]**o+(0.132125994317)*x_ref[1]
        ref[1,0,1,0]=(0.212448400234)*x_ref[0]**o+(0.117680077385)*x_ref[0]+(0.501566152733)*x_ref[1]**o+(-0.0914837162066)*x_ref[1]
        ref[1,0,1,1]=(-0.763152636104)*x_ref[0]**o+(0.318595673875)*x_ref[0]+(-0.166759749724)*x_ref[1]**o+(0.742420099653)*x_ref[1]
        ref[1,0,2,0]=(-0.711420698498)*x_ref[0]**o+(-0.0777910925778)*x_ref[0]+(-0.885593755559)*x_ref[1]**o+(0.302761673106)*x_ref[1]
        ref[1,0,2,1]=(0.980964252534)*x_ref[0]**o+(-0.890328597184)*x_ref[0]+(-0.230421339981)*x_ref[1]**o+(-0.656523995481)*x_ref[1]
        ref[1,1,0,0]=(0.376070672474)*x_ref[0]**o+(0.940362151504)*x_ref[0]+(-0.297285488825)*x_ref[1]**o+(0.451414529855)*x_ref[1]
        ref[1,1,0,1]=(0.867987607318)*x_ref[0]**o+(0.0477620815134)*x_ref[0]+(0.257575148838)*x_ref[1]**o+(-0.54187477723)*x_ref[1]
        ref[1,1,1,0]=(-0.0643073021305)*x_ref[0]**o+(0.20805724465)*x_ref[0]+(-0.16501072904)*x_ref[1]**o+(-0.341858346811)*x_ref[1]
        ref[1,1,1,1]=(0.317830146671)*x_ref[0]**o+(-0.307344984878)*x_ref[0]+(0.170352713722)*x_ref[1]**o+(-0.219561938951)*x_ref[1]
        ref[1,1,2,0]=(0.89815928322)*x_ref[0]**o+(0.51678942721)*x_ref[0]+(0.897095493092)*x_ref[1]**o+(-0.732217042123)*x_ref[1]
        ref[1,1,2,1]=(0.0216578477569)*x_ref[0]**o+(-0.942894641381)*x_ref[0]+(-0.475731377321)*x_ref[1]**o+(-0.753173025139)*x_ref[1]
        ref[1,2,0,0]=(-0.897860515192)*x_ref[0]**o+(0.833870034804)*x_ref[0]+(0.554274053224)*x_ref[1]**o+(-0.785991934149)*x_ref[1]
        ref[1,2,0,1]=(-0.0648893111877)*x_ref[0]**o+(0.173757424631)*x_ref[0]+(0.860427552735)*x_ref[1]**o+(0.139897107451)*x_ref[1]
        ref[1,2,1,0]=(-0.146915585006)*x_ref[0]**o+(-0.290881625114)*x_ref[0]+(0.98541624122)*x_ref[1]**o+(0.868617044417)*x_ref[1]
        ref[1,2,1,1]=(0.17817541898)*x_ref[0]**o+(-0.0112388444881)*x_ref[0]+(0.646731986079)*x_ref[1]**o+(0.653380181018)*x_ref[1]
        ref[1,2,2,0]=(0.208250797377)*x_ref[0]**o+(0.362989066599)*x_ref[0]+(-0.384302232322)*x_ref[1]**o+(0.376460044532)*x_ref[1]
        ref[1,2,2,1]=(0.0437567097332)*x_ref[0]**o+(0.739696851008)*x_ref[0]+(0.0308848949312)*x_ref[1]**o+(-0.445528203988)*x_ref[1]
        ref[1,3,0,0]=(-0.257780793962)*x_ref[0]**o+(0.840582446797)*x_ref[0]+(0.599894892775)*x_ref[1]**o+(-0.744691829118)*x_ref[1]
        ref[1,3,0,1]=(0.913814269186)*x_ref[0]**o+(0.295261304054)*x_ref[0]+(0.0934778065196)*x_ref[1]**o+(0.567115895371)*x_ref[1]
        ref[1,3,1,0]=(-0.399785878199)*x_ref[0]**o+(0.0336841328611)*x_ref[0]+(0.236101490273)*x_ref[1]**o+(0.915159045152)*x_ref[1]
        ref[1,3,1,1]=(-0.955438743542)*x_ref[0]**o+(-0.457060494715)*x_ref[0]+(0.582140402321)*x_ref[1]**o+(-0.999341752789)*x_ref[1]
        ref[1,3,2,0]=(0.0206746776356)*x_ref[0]**o+(-0.307327438556)*x_ref[0]+(-0.646238381067)*x_ref[1]**o+(0.248313182827)*x_ref[1]
        ref[1,3,2,1]=(-0.10971097551)*x_ref[0]**o+(-0.825303025298)*x_ref[0]+(0.876884308015)*x_ref[1]**o+(0.53621927318)*x_ref[1]
        ref[1,4,0,0]=(0.833490742222)*x_ref[0]**o+(0.865604127412)*x_ref[0]+(0.627418353815)*x_ref[1]**o+(0.523133814878)*x_ref[1]
        ref[1,4,0,1]=(0.57616288545)*x_ref[0]**o+(0.438552559891)*x_ref[0]+(0.0256681709722)*x_ref[1]**o+(-0.738314027898)*x_ref[1]
        ref[1,4,1,0]=(0.319863593844)*x_ref[0]**o+(0.177342228601)*x_ref[0]+(0.0865784775196)*x_ref[1]**o+(0.642550298518)*x_ref[1]
        ref[1,4,1,1]=(-0.0141646689097)*x_ref[0]**o+(0.317108687762)*x_ref[0]+(0.188163493751)*x_ref[1]**o+(0.990958509748)*x_ref[1]
        ref[1,4,2,0]=(0.199211769035)*x_ref[0]**o+(-0.0280132568124)*x_ref[0]+(0.144960116505)*x_ref[1]**o+(0.226821898255)*x_ref[1]
        ref[1,4,2,1]=(-0.892464680507)*x_ref[0]**o+(-0.60166056141)*x_ref[0]+(-0.492226295363)*x_ref[1]**o+(0.189924737916)*x_ref[1]
        ref[2,0,0,0]=(0.329199148463)*x_ref[0]**o+(-0.345343145094)*x_ref[0]+(-0.22723822677)*x_ref[1]**o+(-0.214232512114)*x_ref[1]
        ref[2,0,0,1]=(-0.195525660216)*x_ref[0]**o+(0.323518517888)*x_ref[0]+(-0.116808991213)*x_ref[1]**o+(0.22923091239)*x_ref[1]
        ref[2,0,1,0]=(-0.826978585914)*x_ref[0]**o+(0.641345887454)*x_ref[0]+(-0.335583694811)*x_ref[1]**o+(0.168685506673)*x_ref[1]
        ref[2,0,1,1]=(-0.756460980453)*x_ref[0]**o+(-0.9181609332)*x_ref[0]+(0.0735111723828)*x_ref[1]**o+(0.312524164436)*x_ref[1]
        ref[2,0,2,0]=(0.297314747638)*x_ref[0]**o+(0.522844740561)*x_ref[0]+(0.890404335261)*x_ref[1]**o+(-0.158861326888)*x_ref[1]
        ref[2,0,2,1]=(-0.280311121771)*x_ref[0]**o+(0.561223136791)*x_ref[0]+(-0.952737235492)*x_ref[1]**o+(0.581165512883)*x_ref[1]
        ref[2,1,0,0]=(-0.202509823332)*x_ref[0]**o+(0.298626146584)*x_ref[0]+(0.533843705865)*x_ref[1]**o+(0.652127174748)*x_ref[1]
        ref[2,1,0,1]=(-0.349734740031)*x_ref[0]**o+(-0.0589736478602)*x_ref[0]+(0.741348559428)*x_ref[1]**o+(0.993586735262)*x_ref[1]
        ref[2,1,1,0]=(-0.350986159862)*x_ref[0]**o+(-0.814566621733)*x_ref[0]+(0.0556573702074)*x_ref[1]**o+(0.363924401853)*x_ref[1]
        ref[2,1,1,1]=(-0.562086543404)*x_ref[0]**o+(0.126278013645)*x_ref[0]+(-0.488684138283)*x_ref[1]**o+(-0.990773306296)*x_ref[1]
        ref[2,1,2,0]=(-0.873707667752)*x_ref[0]**o+(-0.00610530100216)*x_ref[0]+(-0.927847882023)*x_ref[1]**o+(0.255850655097)*x_ref[1]
        ref[2,1,2,1]=(0.683462220662)*x_ref[0]**o+(-0.515973835798)*x_ref[0]+(-0.676735030671)*x_ref[1]**o+(0.704690967291)*x_ref[1]
        ref[2,2,0,0]=(-0.0891987051347)*x_ref[0]**o+(0.783509807114)*x_ref[0]+(0.583865611337)*x_ref[1]**o+(-0.373096806492)*x_ref[1]
        ref[2,2,0,1]=(0.606443213957)*x_ref[0]**o+(0.887055818748)*x_ref[0]+(-0.281844491159)*x_ref[1]**o+(0.420965776696)*x_ref[1]
        ref[2,2,1,0]=(0.205051077925)*x_ref[0]**o+(0.267690802839)*x_ref[0]+(-0.821781634855)*x_ref[1]**o+(-0.903592743179)*x_ref[1]
        ref[2,2,1,1]=(-0.0438536252522)*x_ref[0]**o+(-0.870635324748)*x_ref[0]+(-0.238830174768)*x_ref[1]**o+(-0.230782152037)*x_ref[1]
        ref[2,2,2,0]=(-0.94437182489)*x_ref[0]**o+(0.0666311523262)*x_ref[0]+(0.481913247572)*x_ref[1]**o+(-0.646415708448)*x_ref[1]
        ref[2,2,2,1]=(0.443901323099)*x_ref[0]**o+(0.349352493291)*x_ref[0]+(-0.158789593075)*x_ref[1]**o+(-0.603521202861)*x_ref[1]
        ref[2,3,0,0]=(0.716829455981)*x_ref[0]**o+(0.0186823101139)*x_ref[0]+(-0.209053562063)*x_ref[1]**o+(0.847906310296)*x_ref[1]
        ref[2,3,0,1]=(0.305837642219)*x_ref[0]**o+(0.138301301592)*x_ref[0]+(0.0645236431399)*x_ref[1]**o+(0.909641719567)*x_ref[1]
        ref[2,3,1,0]=(-0.0177501904561)*x_ref[0]**o+(0.177278703549)*x_ref[0]+(-0.445915965222)*x_ref[1]**o+(-0.894038892768)*x_ref[1]
        ref[2,3,1,1]=(0.00187039478357)*x_ref[0]**o+(-0.598690809424)*x_ref[0]+(0.299307927538)*x_ref[1]**o+(-0.882235917926)*x_ref[1]
        ref[2,3,2,0]=(0.730056097248)*x_ref[0]**o+(-0.307654523095)*x_ref[0]+(0.246914489171)*x_ref[1]**o+(-0.60035541725)*x_ref[1]
        ref[2,3,2,1]=(-0.608344323894)*x_ref[0]**o+(-0.081794637624)*x_ref[0]+(0.099524567015)*x_ref[1]**o+(-0.0654861171366)*x_ref[1]
        ref[2,4,0,0]=(-0.225654564715)*x_ref[0]**o+(-0.529959598356)*x_ref[0]+(-0.890574962317)*x_ref[1]**o+(0.994618025898)*x_ref[1]
        ref[2,4,0,1]=(-0.231311865123)*x_ref[0]**o+(0.479613722979)*x_ref[0]+(-0.614977654518)*x_ref[1]**o+(0.804794516306)*x_ref[1]
        ref[2,4,1,0]=(0.365865384972)*x_ref[0]**o+(-0.889202382492)*x_ref[0]+(-0.300736557219)*x_ref[1]**o+(0.838201123405)*x_ref[1]
        ref[2,4,1,1]=(-0.809876332779)*x_ref[0]**o+(0.1890391825)*x_ref[0]+(-0.607869144071)*x_ref[1]**o+(0.0448966585205)*x_ref[1]
        ref[2,4,2,0]=(-0.83080984753)*x_ref[0]**o+(0.305865817658)*x_ref[0]+(-0.359639965091)*x_ref[1]**o+(-0.288164510937)*x_ref[1]
        ref[2,4,2,1]=(0.616022940224)*x_ref[0]**o+(-0.636356653421)*x_ref[0]+(0.693033289027)*x_ref[1]**o+(-0.780662522342)*x_ref[1]
        ref[3,0,0,0]=(0.985497135193)*x_ref[0]**o+(-0.441259417159)*x_ref[0]+(0.0550399586931)*x_ref[1]**o+(-0.401551085867)*x_ref[1]
        ref[3,0,0,1]=(0.66919872999)*x_ref[0]**o+(-0.21372913425)*x_ref[0]+(-0.904643614761)*x_ref[1]**o+(0.254555219564)*x_ref[1]
        ref[3,0,1,0]=(0.342630495616)*x_ref[0]**o+(-0.688520932208)*x_ref[0]+(-0.795149453533)*x_ref[1]**o+(-0.532353008569)*x_ref[1]
        ref[3,0,1,1]=(-0.157354065395)*x_ref[0]**o+(-0.574302953359)*x_ref[0]+(-0.0505170558232)*x_ref[1]**o+(-0.916081229428)*x_ref[1]
        ref[3,0,2,0]=(-0.276092076819)*x_ref[0]**o+(0.755203166655)*x_ref[0]+(0.71481609597)*x_ref[1]**o+(-0.623929271337)*x_ref[1]
        ref[3,0,2,1]=(0.901066640027)*x_ref[0]**o+(0.615459930577)*x_ref[0]+(-0.65213904597)*x_ref[1]**o+(0.52496253783)*x_ref[1]
        ref[3,1,0,0]=(-0.90847876416)*x_ref[0]**o+(0.787519284245)*x_ref[0]+(0.260252163627)*x_ref[1]**o+(0.24849374459)*x_ref[1]
        ref[3,1,0,1]=(-0.519950434741)*x_ref[0]**o+(-0.146327164967)*x_ref[0]+(0.199399508881)*x_ref[1]**o+(0.583497630868)*x_ref[1]
        ref[3,1,1,0]=(-0.956798390834)*x_ref[0]**o+(0.145681684376)*x_ref[0]+(0.308458821145)*x_ref[1]**o+(-0.759589804755)*x_ref[1]
        ref[3,1,1,1]=(-0.195569939954)*x_ref[0]**o+(-0.544313649532)*x_ref[0]+(0.387017502787)*x_ref[1]**o+(0.174663629204)*x_ref[1]
        ref[3,1,2,0]=(-0.3697024469)*x_ref[0]**o+(0.356841678749)*x_ref[0]+(0.922691790563)*x_ref[1]**o+(-0.906755290703)*x_ref[1]
        ref[3,1,2,1]=(-0.343599294558)*x_ref[0]**o+(-0.213227630549)*x_ref[0]+(-0.822516525353)*x_ref[1]**o+(-0.0350701846979)*x_ref[1]
        ref[3,2,0,0]=(-0.671120249859)*x_ref[0]**o+(-0.409571854187)*x_ref[0]+(-0.948171208395)*x_ref[1]**o+(0.00128117529499)*x_ref[1]
        ref[3,2,0,1]=(0.573086687235)*x_ref[0]**o+(0.272624522478)*x_ref[0]+(0.404610375833)*x_ref[1]**o+(0.177742792198)*x_ref[1]
        ref[3,2,1,0]=(-0.0481585168757)*x_ref[0]**o+(0.108306965224)*x_ref[0]+(0.962229835277)*x_ref[1]**o+(0.457705229155)*x_ref[1]
        ref[3,2,1,1]=(0.843866939471)*x_ref[0]**o+(-0.209955643474)*x_ref[0]+(0.861076954064)*x_ref[1]**o+(-0.0920496448252)*x_ref[1]
        ref[3,2,2,0]=(-0.482466555781)*x_ref[0]**o+(0.905888895109)*x_ref[0]+(0.969984046224)*x_ref[1]**o+(0.779155706115)*x_ref[1]
        ref[3,2,2,1]=(-0.119059605916)*x_ref[0]**o+(-0.413165838565)*x_ref[0]+(-0.458952085266)*x_ref[1]**o+(0.667357851674)*x_ref[1]
        ref[3,3,0,0]=(-0.139020214893)*x_ref[0]**o+(-0.430716093157)*x_ref[0]+(0.587151976237)*x_ref[1]**o+(0.538779583614)*x_ref[1]
        ref[3,3,0,1]=(-0.196698073225)*x_ref[0]**o+(0.28357504894)*x_ref[0]+(0.628834405238)*x_ref[1]**o+(0.991210138395)*x_ref[1]
        ref[3,3,1,0]=(-0.375644375731)*x_ref[0]**o+(0.889016565145)*x_ref[0]+(-0.495700828743)*x_ref[1]**o+(-0.498859003938)*x_ref[1]
        ref[3,3,1,1]=(0.783773071811)*x_ref[0]**o+(-0.0875026023418)*x_ref[0]+(-0.421213063239)*x_ref[1]**o+(0.750220508147)*x_ref[1]
        ref[3,3,2,0]=(0.211305046835)*x_ref[0]**o+(-0.828436860631)*x_ref[0]+(0.225275220049)*x_ref[1]**o+(0.264723427485)*x_ref[1]
        ref[3,3,2,1]=(0.582818095069)*x_ref[0]**o+(0.497614150982)*x_ref[0]+(-0.445084955033)*x_ref[1]**o+(-0.600656630963)*x_ref[1]
        ref[3,4,0,0]=(-0.886999530258)*x_ref[0]**o+(0.386008089061)*x_ref[0]+(-0.292729716725)*x_ref[1]**o+(-0.159269788338)*x_ref[1]
        ref[3,4,0,1]=(0.260932511148)*x_ref[0]**o+(-0.533553803234)*x_ref[0]+(0.174946866584)*x_ref[1]**o+(-0.17701796222)*x_ref[1]
        ref[3,4,1,0]=(0.37985010505)*x_ref[0]**o+(0.934809225579)*x_ref[0]+(-0.76271501884)*x_ref[1]**o+(0.171072659)*x_ref[1]
        ref[3,4,1,1]=(-0.369797459679)*x_ref[0]**o+(-0.714539720717)*x_ref[0]+(-0.122698805149)*x_ref[1]**o+(0.874232715421)*x_ref[1]
        ref[3,4,2,0]=(0.912569854856)*x_ref[0]**o+(-0.773508960793)*x_ref[0]+(-0.447748330632)*x_ref[1]**o+(-0.0117517938995)*x_ref[1]
        ref[3,4,2,1]=(-0.467442835658)*x_ref[0]**o+(-0.792740521854)*x_ref[0]+(-0.33441666815)*x_ref[1]**o+(-0.821650769855)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.171866645738)*x[0]**o+(0.561850203514)*x[0]+(0.416451445795)*x[1]**o+(-0.235582248693)*x[1]+(-0.196259994197)*x[2]**o+(0.940566952036)*x[2]
        arg[0,0,0,1]=(-0.00971758894531)*x[0]**o+(-0.590140440451)*x[0]+(-0.719992277571)*x[1]**o+(-0.0411344370809)*x[1]+(-0.301518484556)*x[2]**o+(-0.624210342075)*x[2]
        arg[0,0,1,0]=(0.142493885612)*x[0]**o+(0.473915143223)*x[0]+(0.236217877001)*x[1]**o+(-0.26535451945)*x[1]+(0.197580058644)*x[2]**o+(0.440093349171)*x[2]
        arg[0,0,1,1]=(0.532930185145)*x[0]**o+(0.206573108032)*x[0]+(-0.903235976279)*x[1]**o+(-0.192856647484)*x[1]+(0.247399954122)*x[2]**o+(-0.396060116726)*x[2]
        arg[0,0,2,0]=(0.106261668354)*x[0]**o+(-0.169092441923)*x[0]+(0.495324512568)*x[1]**o+(0.612873784609)*x[1]+(-0.181887252227)*x[2]**o+(-0.349265048779)*x[2]
        arg[0,0,2,1]=(0.0407741271369)*x[0]**o+(-0.286131832952)*x[0]+(0.00801790328545)*x[1]**o+(-0.930580202402)*x[1]+(-0.977595004998)*x[2]**o+(-0.176312831325)*x[2]
        arg[0,1,0,0]=(0.0660805950534)*x[0]**o+(0.27279659698)*x[0]+(0.383210617595)*x[1]**o+(-0.453071384847)*x[1]+(-0.497520020161)*x[2]**o+(-0.0622858051806)*x[2]
        arg[0,1,0,1]=(0.343928280381)*x[0]**o+(0.612026453223)*x[0]+(-0.278958038519)*x[1]**o+(-0.287367815759)*x[1]+(-0.975555454735)*x[2]**o+(-0.0381461641238)*x[2]
        arg[0,1,1,0]=(-0.651824952239)*x[0]**o+(0.504032353063)*x[0]+(-0.505340720597)*x[1]**o+(0.804430789885)*x[1]+(0.626321487603)*x[2]**o+(0.771921294247)*x[2]
        arg[0,1,1,1]=(-0.677075301947)*x[0]**o+(0.977678280199)*x[0]+(-0.0282503291518)*x[1]**o+(0.383805851778)*x[1]+(-0.942267069472)*x[2]**o+(-0.168439203764)*x[2]
        arg[0,1,2,0]=(-0.218809773899)*x[0]**o+(-0.923486201344)*x[0]+(-0.771580374186)*x[1]**o+(-0.802318678915)*x[1]+(-0.817711294844)*x[2]**o+(-0.524488450611)*x[2]
        arg[0,1,2,1]=(-0.258192021864)*x[0]**o+(-0.30201835948)*x[0]+(0.908539532951)*x[1]**o+(0.53218981056)*x[1]+(0.0337395431484)*x[2]**o+(0.358676387738)*x[2]
        arg[0,2,0,0]=(0.181349802806)*x[0]**o+(-0.67447696451)*x[0]+(-0.905389539679)*x[1]**o+(-0.0192162208351)*x[1]+(0.475994769738)*x[2]**o+(0.258880626391)*x[2]
        arg[0,2,0,1]=(0.683461009428)*x[0]**o+(0.101633874876)*x[0]+(-0.666120146454)*x[1]**o+(-0.79937847801)*x[1]+(-0.0519440794346)*x[2]**o+(0.412543335301)*x[2]
        arg[0,2,1,0]=(-0.705756906195)*x[0]**o+(-0.914133519297)*x[0]+(0.203329174425)*x[1]**o+(-0.125897560162)*x[1]+(-0.845555978817)*x[2]**o+(-0.376117448037)*x[2]
        arg[0,2,1,1]=(0.129675218633)*x[0]**o+(0.230278463142)*x[0]+(-0.0465924540599)*x[1]**o+(0.618850924426)*x[1]+(-0.981020601489)*x[2]**o+(-0.446786649897)*x[2]
        arg[0,2,2,0]=(0.709293140521)*x[0]**o+(-0.117595235921)*x[0]+(0.00527289231769)*x[1]**o+(0.407867521466)*x[1]+(-0.0513211295402)*x[2]**o+(-0.775396743388)*x[2]
        arg[0,2,2,1]=(0.456146533185)*x[0]**o+(0.357343068604)*x[0]+(-0.596648639307)*x[1]**o+(0.0400180634504)*x[1]+(-0.773508416513)*x[2]**o+(0.873275482135)*x[2]
        arg[0,3,0,0]=(-0.553923806627)*x[0]**o+(0.416008424454)*x[0]+(-0.18857445771)*x[1]**o+(0.668616517137)*x[1]+(0.850445273446)*x[2]**o+(0.0324670478739)*x[2]
        arg[0,3,0,1]=(-0.556129092409)*x[0]**o+(-0.207129568151)*x[0]+(-0.616200225149)*x[1]**o+(0.173694871558)*x[1]+(-0.389943310499)*x[2]**o+(-0.0171456312507)*x[2]
        arg[0,3,1,0]=(0.029375229156)*x[0]**o+(-0.710817311305)*x[0]+(0.487826213322)*x[1]**o+(-0.353110144226)*x[1]+(-0.326240277953)*x[2]**o+(0.435853126161)*x[2]
        arg[0,3,1,1]=(-0.973283235545)*x[0]**o+(0.119293099332)*x[0]+(-0.987231590871)*x[1]**o+(-0.778410371876)*x[1]+(-0.123668803712)*x[2]**o+(0.836409669329)*x[2]
        arg[0,3,2,0]=(-0.953711717094)*x[0]**o+(0.0211462821501)*x[0]+(0.989232030291)*x[1]**o+(0.684525652819)*x[1]+(-0.934499312995)*x[2]**o+(0.0759185573729)*x[2]
        arg[0,3,2,1]=(0.962555034668)*x[0]**o+(-0.344084097742)*x[0]+(0.627919365161)*x[1]**o+(-0.0520511084603)*x[1]+(0.548149991599)*x[2]**o+(0.885011282903)*x[2]
        arg[0,4,0,0]=(-0.604624034071)*x[0]**o+(0.194224357892)*x[0]+(-0.171628912052)*x[1]**o+(0.0361170699761)*x[1]+(0.68259161986)*x[2]**o+(-0.47807189026)*x[2]
        arg[0,4,0,1]=(-0.441358360042)*x[0]**o+(0.854885170771)*x[0]+(0.739245976992)*x[1]**o+(-0.712495821809)*x[1]+(0.786863605702)*x[2]**o+(0.244289741577)*x[2]
        arg[0,4,1,0]=(-0.93437768143)*x[0]**o+(0.909235101604)*x[0]+(0.398343776489)*x[1]**o+(-0.8015457882)*x[1]+(-0.045056420534)*x[2]**o+(-0.0696087139589)*x[2]
        arg[0,4,1,1]=(0.997282980634)*x[0]**o+(-0.523146965095)*x[0]+(0.888300277406)*x[1]**o+(0.0692584552317)*x[1]+(-0.361183524085)*x[2]**o+(0.298003855572)*x[2]
        arg[0,4,2,0]=(0.510704730088)*x[0]**o+(-0.288621287507)*x[0]+(0.694375966407)*x[1]**o+(0.438558764462)*x[1]+(0.999793245458)*x[2]**o+(0.146242565823)*x[2]
        arg[0,4,2,1]=(0.461717367889)*x[0]**o+(0.585419576536)*x[0]+(0.172514336678)*x[1]**o+(-0.372717466388)*x[1]+(0.354162503692)*x[2]**o+(0.465497372897)*x[2]
        arg[1,0,0,0]=(0.536758620313)*x[0]**o+(0.428277991055)*x[0]+(-0.60617935999)*x[1]**o+(-0.889897682367)*x[1]+(-0.326378737219)*x[2]**o+(0.735260248159)*x[2]
        arg[1,0,0,1]=(-0.629157417304)*x[0]**o+(-0.00171372949641)*x[0]+(0.110550936359)*x[1]**o+(-0.418314227619)*x[1]+(0.968147035552)*x[2]**o+(-0.58304546387)*x[2]
        arg[1,0,1,0]=(0.528767817229)*x[0]**o+(0.111986940482)*x[0]+(0.853287140505)*x[1]**o+(0.551761795603)*x[1]+(-0.848026685332)*x[2]**o+(0.799725876993)*x[2]
        arg[1,0,1,1]=(0.154974399407)*x[0]**o+(-0.662212037776)*x[0]+(0.628908352518)*x[1]**o+(-0.0175576929699)*x[1]+(-0.0501253203152)*x[2]**o+(0.147651837545)*x[2]
        arg[1,0,2,0]=(0.53576644491)*x[0]**o+(-0.935865707952)*x[0]+(-0.256921357737)*x[1]**o+(0.452311436169)*x[1]+(-0.449646854354)*x[2]**o+(0.0137632023964)*x[2]
        arg[1,0,2,1]=(-0.103323182353)*x[0]**o+(-0.0636296047221)*x[0]+(-0.681883331812)*x[1]**o+(-0.284589673307)*x[1]+(-0.525524911635)*x[2]**o+(0.525757772842)*x[2]
        arg[1,1,0,0]=(0.106186299307)*x[0]**o+(0.193833690397)*x[0]+(0.755686652252)*x[1]**o+(-0.954503081175)*x[1]+(-0.946058471992)*x[2]**o+(0.258850283735)*x[2]
        arg[1,1,0,1]=(-0.352425152661)*x[0]**o+(-0.955207432287)*x[0]+(0.946989898911)*x[1]**o+(0.426365715943)*x[1]+(-0.225898340088)*x[2]**o+(-0.108613105442)*x[2]
        arg[1,1,1,0]=(0.505488482593)*x[0]**o+(-0.0193114896733)*x[0]+(-0.809537000564)*x[1]**o+(0.39999431472)*x[1]+(-0.968315699948)*x[2]**o+(0.672744543525)*x[2]
        arg[1,1,1,1]=(0.0967454491182)*x[0]**o+(-0.608840080579)*x[0]+(0.287989672484)*x[1]**o+(-0.989435924681)*x[1]+(0.46833865882)*x[2]**o+(0.0211042567481)*x[2]
        arg[1,1,2,0]=(-0.461777513493)*x[0]**o+(-0.539619783995)*x[0]+(-0.719051252797)*x[1]**o+(-0.980491094424)*x[1]+(0.526872369892)*x[2]**o+(0.534649090672)*x[2]
        arg[1,1,2,1]=(-0.498068667635)*x[0]**o+(-0.977138256259)*x[0]+(0.388405877405)*x[1]**o+(-0.652220953523)*x[1]+(0.214733833456)*x[2]**o+(0.925064130776)*x[2]
        arg[1,2,0,0]=(-0.728915088302)*x[0]**o+(-0.867653377354)*x[0]+(-0.0240751909884)*x[1]**o+(-0.617932013511)*x[1]+(0.342417783703)*x[2]**o+(0.269029568448)*x[2]
        arg[1,2,0,1]=(0.659030580895)*x[0]**o+(0.460057036148)*x[0]+(0.0556749040721)*x[1]**o+(-0.0943860780082)*x[1]+(0.940068201493)*x[2]**o+(-0.156531042318)*x[2]
        arg[1,2,1,0]=(-0.0373627034239)*x[0]**o+(-0.631862915794)*x[0]+(0.384889489295)*x[1]**o+(-0.85645836005)*x[1]+(-0.448074134473)*x[2]**o+(-0.103825620602)*x[2]
        arg[1,2,1,1]=(-0.19142928199)*x[0]**o+(-0.993719455322)*x[0]+(0.204439624689)*x[1]**o+(-0.0616134769914)*x[1]+(0.581621603387)*x[2]**o+(-0.554871585864)*x[2]
        arg[1,2,2,0]=(0.684658549351)*x[0]**o+(-0.0520805658251)*x[0]+(-0.549405651115)*x[1]**o+(-0.0890223507953)*x[1]+(0.538847108562)*x[2]**o+(0.555575374976)*x[2]
        arg[1,2,2,1]=(-0.197138807828)*x[0]**o+(0.691528695453)*x[0]+(0.337770131451)*x[1]**o+(-0.0662330700506)*x[1]+(-0.556170962811)*x[2]**o+(-0.415321654558)*x[2]
        arg[1,3,0,0]=(-0.255542847866)*x[0]**o+(-0.731103056475)*x[0]+(-0.764119472108)*x[1]**o+(0.264775911336)*x[1]+(0.976604892324)*x[2]**o+(-0.929820558374)*x[2]
        arg[1,3,0,1]=(-0.506143192998)*x[0]**o+(-0.139848041055)*x[0]+(-0.074839899541)*x[1]**o+(-0.35100597507)*x[1]+(-0.127574366194)*x[2]**o+(-0.0084655916902)*x[2]
        arg[1,3,1,0]=(0.099997170805)*x[0]**o+(-0.149089017918)*x[0]+(0.419790611192)*x[1]**o+(-0.148698964676)*x[1]+(0.490423198386)*x[2]**o+(-0.210076971679)*x[2]
        arg[1,3,1,1]=(-0.139371350292)*x[0]**o+(-0.320648977772)*x[0]+(0.836558622998)*x[1]**o+(0.394491153689)*x[1]+(0.0857621395676)*x[2]**o+(-0.709138669156)*x[2]
        arg[1,3,2,0]=(0.743901534332)*x[0]**o+(-0.249486095838)*x[0]+(0.281087339807)*x[1]**o+(0.617669579771)*x[1]+(-0.383135240681)*x[2]**o+(-0.647267353645)*x[2]
        arg[1,3,2,1]=(-0.465587857279)*x[0]**o+(0.578094609537)*x[0]+(0.400668301655)*x[1]**o+(-0.999984469723)*x[1]+(-0.460584024789)*x[2]**o+(0.943466859654)*x[2]
        arg[1,4,0,0]=(-0.246915327144)*x[0]**o+(-0.203163427715)*x[0]+(0.138589009969)*x[1]**o+(-0.705709971036)*x[1]+(0.273085758612)*x[2]**o+(0.901084781824)*x[2]
        arg[1,4,0,1]=(-0.919682459969)*x[0]**o+(-0.476395399467)*x[0]+(0.00638460254264)*x[1]**o+(-0.474710529098)*x[1]+(-0.935681998683)*x[2]**o+(0.461344674976)*x[2]
        arg[1,4,1,0]=(0.8225038116)*x[0]**o+(0.0855444334027)*x[0]+(-0.685983943046)*x[1]**o+(-0.46700703117)*x[1]+(0.760260028729)*x[2]**o+(-0.190951535155)*x[2]
        arg[1,4,1,1]=(0.512940503439)*x[0]**o+(0.998803686398)*x[0]+(0.200914807938)*x[1]**o+(0.793533216819)*x[1]+(0.980475771536)*x[2]**o+(-0.00307515759413)*x[2]
        arg[1,4,2,0]=(0.278761841603)*x[0]**o+(-0.368232992504)*x[0]+(0.939116142167)*x[1]**o+(0.13661977072)*x[1]+(-0.490378973691)*x[2]**o+(-0.409117913136)*x[2]
        arg[1,4,2,1]=(-0.456090186141)*x[0]**o+(-0.881413614887)*x[0]+(-0.951099957833)*x[1]**o+(0.859690794224)*x[1]+(0.0723671154601)*x[2]**o+(0.468823687196)*x[2]
        arg[2,0,0,0]=(-0.415483671814)*x[0]**o+(0.415160471625)*x[0]+(-0.730798143253)*x[1]**o+(-0.90883538519)*x[1]+(-0.241870631235)*x[2]**o+(-0.549577536182)*x[2]
        arg[2,0,0,1]=(0.86985374761)*x[0]**o+(0.378073808491)*x[0]+(0.0923146623212)*x[1]**o+(-0.912604209727)*x[1]+(0.529162434789)*x[2]**o+(-0.135836425669)*x[2]
        arg[2,0,1,0]=(0.702609345242)*x[0]**o+(-0.63859466548)*x[0]+(-0.2668268067)*x[1]**o+(0.107377268995)*x[1]+(-0.575458650368)*x[2]**o+(-0.359972696269)*x[2]
        arg[2,0,1,1]=(-0.169295117134)*x[0]**o+(-0.0780580709275)*x[0]+(0.9803920116)*x[1]**o+(0.794090800797)*x[1]+(0.71271688197)*x[2]**o+(0.265355025589)*x[2]
        arg[2,0,2,0]=(0.700475435131)*x[0]**o+(0.317025249264)*x[0]+(0.885140001575)*x[1]**o+(-0.733992012528)*x[1]+(-0.937990577058)*x[2]**o+(-0.566197708973)*x[2]
        arg[2,0,2,1]=(0.369801749993)*x[0]**o+(0.292267254334)*x[0]+(0.24962184892)*x[1]**o+(-0.912670113872)*x[1]+(0.590235505871)*x[2]**o+(0.199279333126)*x[2]
        arg[2,1,0,0]=(-0.165657934182)*x[0]**o+(0.837030192613)*x[0]+(-0.615301249273)*x[1]**o+(-0.947633937776)*x[1]+(0.916677465451)*x[2]**o+(0.633727682324)*x[2]
        arg[2,1,0,1]=(0.260787013495)*x[0]**o+(0.367146934614)*x[0]+(0.0271903085878)*x[1]**o+(0.853398837558)*x[1]+(-0.381379969861)*x[2]**o+(0.140001581749)*x[2]
        arg[2,1,1,0]=(0.121628758983)*x[0]**o+(-0.212350487769)*x[0]+(-0.198031244861)*x[1]**o+(-0.695620829484)*x[1]+(-0.0533521134282)*x[2]**o+(0.956103910898)*x[2]
        arg[2,1,1,1]=(-0.321102504727)*x[0]**o+(-0.920842546991)*x[0]+(0.082643684589)*x[1]**o+(-0.0960346141926)*x[1]+(0.475269198456)*x[2]**o+(-0.184473007427)*x[2]
        arg[2,1,2,0]=(0.412066628472)*x[0]**o+(-0.667420990332)*x[0]+(0.754028295628)*x[1]**o+(0.967928426879)*x[1]+(-0.129320768473)*x[2]**o+(-0.496085464716)*x[2]
        arg[2,1,2,1]=(-0.379495871082)*x[0]**o+(0.768674450671)*x[0]+(0.786595353344)*x[1]**o+(-0.191866280074)*x[1]+(0.314572026852)*x[2]**o+(-0.0245751436484)*x[2]
        arg[2,2,0,0]=(-0.291171443165)*x[0]**o+(0.857660522727)*x[0]+(-0.0611031323308)*x[1]**o+(0.365691706254)*x[1]+(0.906199720505)*x[2]**o+(0.621747944156)*x[2]
        arg[2,2,0,1]=(-0.110402534173)*x[0]**o+(-0.465943589112)*x[0]+(0.896364485404)*x[1]**o+(0.16234913823)*x[1]+(0.443564946892)*x[2]**o+(0.259358795418)*x[2]
        arg[2,2,1,0]=(-0.480613805459)*x[0]**o+(-0.837751646963)*x[0]+(0.803669049568)*x[1]**o+(-0.859501917481)*x[1]+(0.633523536226)*x[2]**o+(0.351742190573)*x[2]
        arg[2,2,1,1]=(0.133665578883)*x[0]**o+(-0.283052337652)*x[0]+(-0.382761111917)*x[1]**o+(-0.197039443436)*x[1]+(0.081520912443)*x[2]**o+(-0.167181778234)*x[2]
        arg[2,2,2,0]=(0.207787204976)*x[0]**o+(0.245207436631)*x[0]+(0.340092449827)*x[1]**o+(0.541457158709)*x[1]+(-0.606955763027)*x[2]**o+(-0.197506280875)*x[2]
        arg[2,2,2,1]=(0.399727427867)*x[0]**o+(-0.0452717156729)*x[0]+(0.85643441)*x[1]**o+(-0.938966108133)*x[1]+(0.832098573693)*x[2]**o+(-0.375366657117)*x[2]
        arg[2,3,0,0]=(-0.661411409657)*x[0]**o+(0.702058724486)*x[0]+(0.851820224497)*x[1]**o+(-0.984572528809)*x[1]+(0.700617112775)*x[2]**o+(0.990437560702)*x[2]
        arg[2,3,0,1]=(-0.937637737824)*x[0]**o+(0.970774222765)*x[0]+(0.0263154279607)*x[1]**o+(-0.453145473796)*x[1]+(-0.740101764089)*x[2]**o+(0.248604522836)*x[2]
        arg[2,3,1,0]=(0.624217308865)*x[0]**o+(0.682749411369)*x[0]+(0.394444314015)*x[1]**o+(0.288021257523)*x[1]+(-0.943712429774)*x[2]**o+(-0.299272610192)*x[2]
        arg[2,3,1,1]=(0.00512058360895)*x[0]**o+(-0.0263977566592)*x[0]+(-0.769491595627)*x[1]**o+(-0.403091037334)*x[1]+(-0.940178247778)*x[2]**o+(-0.664590180141)*x[2]
        arg[2,3,2,0]=(-0.606361369179)*x[0]**o+(0.0252279171688)*x[0]+(0.0125424811264)*x[1]**o+(-0.249080159522)*x[1]+(0.969599361427)*x[2]**o+(0.369428527325)*x[2]
        arg[2,3,2,1]=(-0.899501409225)*x[0]**o+(0.834437734927)*x[0]+(-0.992484449733)*x[1]**o+(0.00567086116985)*x[1]+(0.112432151044)*x[2]**o+(0.725013775434)*x[2]
        arg[2,4,0,0]=(-0.864861209468)*x[0]**o+(0.475518351325)*x[0]+(-0.575172287508)*x[1]**o+(-0.695754751991)*x[1]+(0.935933655733)*x[2]**o+(-0.156489388279)*x[2]
        arg[2,4,0,1]=(-0.072905918035)*x[0]**o+(0.912617659416)*x[0]+(0.76767638187)*x[1]**o+(0.517738398356)*x[1]+(-0.116771137305)*x[2]**o+(-0.887027853796)*x[2]
        arg[2,4,1,0]=(-0.169351806151)*x[0]**o+(-0.667994142736)*x[0]+(-0.742400183746)*x[1]**o+(-0.345483488309)*x[1]+(0.534459208989)*x[2]**o+(-0.314632556106)*x[2]
        arg[2,4,1,1]=(0.571889164138)*x[0]**o+(-0.603799705213)*x[0]+(0.574246013631)*x[1]**o+(-0.156121300765)*x[1]+(-0.300131083217)*x[2]**o+(0.781527389701)*x[2]
        arg[2,4,2,0]=(-0.433340783774)*x[0]**o+(-0.337577532056)*x[0]+(-0.991017325155)*x[1]**o+(0.169866142612)*x[1]+(0.531407605197)*x[2]**o+(0.91525228665)*x[2]
        arg[2,4,2,1]=(0.837883720242)*x[0]**o+(-0.647682324772)*x[0]+(0.125057981769)*x[1]**o+(-0.734369688409)*x[1]+(0.5864622868)*x[2]**o+(-0.378030111621)*x[2]
        arg[3,0,0,0]=(-0.497406654549)*x[0]**o+(0.773396244829)*x[0]+(0.587080735986)*x[1]**o+(0.80571073543)*x[1]+(-0.621583700283)*x[2]**o+(0.214586226671)*x[2]
        arg[3,0,0,1]=(0.602557120764)*x[0]**o+(0.362884272628)*x[0]+(-0.469709604155)*x[1]**o+(0.570294839284)*x[1]+(-0.977082842755)*x[2]**o+(-0.871580274035)*x[2]
        arg[3,0,1,0]=(0.801845637209)*x[0]**o+(0.65646927893)*x[0]+(0.181490001743)*x[1]**o+(-0.579155840484)*x[1]+(-0.690604456924)*x[2]**o+(0.999798826875)*x[2]
        arg[3,0,1,1]=(-0.0835741146147)*x[0]**o+(0.251697832709)*x[0]+(0.915417127133)*x[1]**o+(0.859412289228)*x[1]+(0.529025246336)*x[2]**o+(0.0290945910046)*x[2]
        arg[3,0,2,0]=(0.43632860403)*x[0]**o+(-0.71378809102)*x[0]+(0.996769149715)*x[1]**o+(-0.694095684973)*x[1]+(0.24606095312)*x[2]**o+(0.0974139948716)*x[2]
        arg[3,0,2,1]=(-0.615800860737)*x[0]**o+(-0.0194774462371)*x[0]+(-0.323266780598)*x[1]**o+(0.70039319915)*x[1]+(0.0990606501268)*x[2]**o+(-0.265400117193)*x[2]
        arg[3,1,0,0]=(0.637788917841)*x[0]**o+(-0.898134292781)*x[0]+(0.397058421405)*x[1]**o+(0.424875003774)*x[1]+(0.493154896437)*x[2]**o+(0.426966521522)*x[2]
        arg[3,1,0,1]=(0.307016370261)*x[0]**o+(0.583037237501)*x[0]+(-0.715510433757)*x[1]**o+(-0.778084356162)*x[1]+(0.624410868665)*x[2]**o+(0.534413229876)*x[2]
        arg[3,1,1,0]=(0.53494197621)*x[0]**o+(0.0347747301727)*x[0]+(0.387683566308)*x[1]**o+(0.79111981305)*x[1]+(0.499346741486)*x[2]**o+(-0.89736311061)*x[2]
        arg[3,1,1,1]=(-0.414505914492)*x[0]**o+(0.738343334487)*x[0]+(0.0839982715753)*x[1]**o+(0.306059781071)*x[1]+(0.205292188103)*x[2]**o+(0.478479664712)*x[2]
        arg[3,1,2,0]=(0.649444916094)*x[0]**o+(-0.616363283285)*x[0]+(-0.890710011556)*x[1]**o+(-0.55638070962)*x[1]+(0.63397005463)*x[2]**o+(-0.296632593945)*x[2]
        arg[3,1,2,1]=(0.867500426375)*x[0]**o+(0.087056846585)*x[0]+(-0.347167872287)*x[1]**o+(-0.593410409382)*x[1]+(-0.0853161439992)*x[2]**o+(0.32855927156)*x[2]
        arg[3,2,0,0]=(0.176880322402)*x[0]**o+(-0.342939990014)*x[0]+(0.927975205318)*x[1]**o+(0.395166493409)*x[1]+(-0.114428155908)*x[2]**o+(0.518356015938)*x[2]
        arg[3,2,0,1]=(-0.560834784775)*x[0]**o+(0.989745819909)*x[0]+(-0.201732235802)*x[1]**o+(-0.953838293418)*x[1]+(-0.797223863765)*x[2]**o+(-0.44533752286)*x[2]
        arg[3,2,1,0]=(-0.83129861256)*x[0]**o+(-0.768067351258)*x[0]+(-0.564704574946)*x[1]**o+(0.419961478425)*x[1]+(0.561243126484)*x[2]**o+(0.536547545961)*x[2]
        arg[3,2,1,1]=(-0.245876224002)*x[0]**o+(-0.0399839529678)*x[0]+(-0.728071866725)*x[1]**o+(0.68478402008)*x[1]+(0.907895423543)*x[2]**o+(0.454698589619)*x[2]
        arg[3,2,2,0]=(-0.412012705012)*x[0]**o+(-0.575272496028)*x[0]+(0.877861906104)*x[1]**o+(-0.586319475679)*x[1]+(0.668997061563)*x[2]**o+(-0.418342969428)*x[2]
        arg[3,2,2,1]=(0.270232215934)*x[0]**o+(0.236642564222)*x[0]+(-0.0693779443521)*x[1]**o+(0.77329883812)*x[1]+(-0.743424502879)*x[2]**o+(0.296073856497)*x[2]
        arg[3,3,0,0]=(-0.0659865092333)*x[0]**o+(-0.645286369943)*x[0]+(0.746214220465)*x[1]**o+(-0.549592848453)*x[1]+(-0.194552740752)*x[2]**o+(0.136734604935)*x[2]
        arg[3,3,0,1]=(0.0407640382533)*x[0]**o+(-0.720002890774)*x[0]+(-0.902500650998)*x[1]**o+(-0.0763091975748)*x[1]+(-0.70279308164)*x[2]**o+(-0.412384977965)*x[2]
        arg[3,3,1,0]=(0.460226151062)*x[0]**o+(-0.359911871464)*x[0]+(0.654909639152)*x[1]**o+(-0.574154331236)*x[1]+(-0.226246538275)*x[2]**o+(-0.168145308987)*x[2]
        arg[3,3,1,1]=(-0.966169813904)*x[0]**o+(-0.784593122487)*x[0]+(-0.857455334787)*x[1]**o+(-0.116963343366)*x[1]+(0.231055675695)*x[2]**o+(-0.953091037377)*x[2]
        arg[3,3,2,0]=(-0.796170783237)*x[0]**o+(-0.605030579381)*x[0]+(0.388310055252)*x[1]**o+(-0.504349251868)*x[1]+(0.83840677955)*x[2]**o+(-0.586099851479)*x[2]
        arg[3,3,2,1]=(0.195092011781)*x[0]**o+(-0.358423372419)*x[0]+(-0.643669425281)*x[1]**o+(0.721564003575)*x[1]+(-0.588431193874)*x[2]**o+(-0.378625221371)*x[2]
        arg[3,4,0,0]=(0.585756187041)*x[0]**o+(-0.753400165107)*x[0]+(0.790439809615)*x[1]**o+(0.0473109698221)*x[1]+(0.951768026456)*x[2]**o+(0.423960993524)*x[2]
        arg[3,4,0,1]=(0.32857275647)*x[0]**o+(0.566198295813)*x[0]+(-0.372998380277)*x[1]**o+(0.671551057967)*x[1]+(-0.990512903698)*x[2]**o+(-0.598820986842)*x[2]
        arg[3,4,1,0]=(0.512362923169)*x[0]**o+(0.161976775973)*x[0]+(-0.823370116651)*x[1]**o+(-0.539694552746)*x[1]+(-0.252135888305)*x[2]**o+(-0.323536133539)*x[2]
        arg[3,4,1,1]=(0.342762199745)*x[0]**o+(0.569702637139)*x[0]+(0.514189130968)*x[1]**o+(-0.219203654565)*x[1]+(-0.839606678057)*x[2]**o+(0.374386070142)*x[2]
        arg[3,4,2,0]=(0.0568713695577)*x[0]**o+(-0.494349350234)*x[0]+(0.829657653367)*x[1]**o+(0.541489358189)*x[1]+(-0.9469529858)*x[2]**o+(0.230257255487)*x[2]
        arg[3,4,2,1]=(0.874227414064)*x[0]**o+(0.76572262681)*x[0]+(0.305716770957)*x[1]**o+(0.225904980409)*x[1]+(-0.10063057638)*x[2]**o+(-0.87920628616)*x[2]
        ref[0,0,0,0]=(-0.171866645738)*x_ref[0]**o+(0.561850203514)*x_ref[0]+(0.416451445795)*x_ref[1]**o+(-0.235582248693)*x_ref[1]+(-0.196259994197)*x_ref[2]**o+(0.940566952036)*x_ref[2]
        ref[0,0,0,1]=(-0.00971758894531)*x_ref[0]**o+(-0.590140440451)*x_ref[0]+(-0.719992277571)*x_ref[1]**o+(-0.0411344370809)*x_ref[1]+(-0.301518484556)*x_ref[2]**o+(-0.624210342075)*x_ref[2]
        ref[0,0,1,0]=(0.142493885612)*x_ref[0]**o+(0.473915143223)*x_ref[0]+(0.236217877001)*x_ref[1]**o+(-0.26535451945)*x_ref[1]+(0.197580058644)*x_ref[2]**o+(0.440093349171)*x_ref[2]
        ref[0,0,1,1]=(0.532930185145)*x_ref[0]**o+(0.206573108032)*x_ref[0]+(-0.903235976279)*x_ref[1]**o+(-0.192856647484)*x_ref[1]+(0.247399954122)*x_ref[2]**o+(-0.396060116726)*x_ref[2]
        ref[0,0,2,0]=(0.106261668354)*x_ref[0]**o+(-0.169092441923)*x_ref[0]+(0.495324512568)*x_ref[1]**o+(0.612873784609)*x_ref[1]+(-0.181887252227)*x_ref[2]**o+(-0.349265048779)*x_ref[2]
        ref[0,0,2,1]=(0.0407741271369)*x_ref[0]**o+(-0.286131832952)*x_ref[0]+(0.00801790328545)*x_ref[1]**o+(-0.930580202402)*x_ref[1]+(-0.977595004998)*x_ref[2]**o+(-0.176312831325)*x_ref[2]
        ref[0,1,0,0]=(0.0660805950534)*x_ref[0]**o+(0.27279659698)*x_ref[0]+(0.383210617595)*x_ref[1]**o+(-0.453071384847)*x_ref[1]+(-0.497520020161)*x_ref[2]**o+(-0.0622858051806)*x_ref[2]
        ref[0,1,0,1]=(0.343928280381)*x_ref[0]**o+(0.612026453223)*x_ref[0]+(-0.278958038519)*x_ref[1]**o+(-0.287367815759)*x_ref[1]+(-0.975555454735)*x_ref[2]**o+(-0.0381461641238)*x_ref[2]
        ref[0,1,1,0]=(-0.651824952239)*x_ref[0]**o+(0.504032353063)*x_ref[0]+(-0.505340720597)*x_ref[1]**o+(0.804430789885)*x_ref[1]+(0.626321487603)*x_ref[2]**o+(0.771921294247)*x_ref[2]
        ref[0,1,1,1]=(-0.677075301947)*x_ref[0]**o+(0.977678280199)*x_ref[0]+(-0.0282503291518)*x_ref[1]**o+(0.383805851778)*x_ref[1]+(-0.942267069472)*x_ref[2]**o+(-0.168439203764)*x_ref[2]
        ref[0,1,2,0]=(-0.218809773899)*x_ref[0]**o+(-0.923486201344)*x_ref[0]+(-0.771580374186)*x_ref[1]**o+(-0.802318678915)*x_ref[1]+(-0.817711294844)*x_ref[2]**o+(-0.524488450611)*x_ref[2]
        ref[0,1,2,1]=(-0.258192021864)*x_ref[0]**o+(-0.30201835948)*x_ref[0]+(0.908539532951)*x_ref[1]**o+(0.53218981056)*x_ref[1]+(0.0337395431484)*x_ref[2]**o+(0.358676387738)*x_ref[2]
        ref[0,2,0,0]=(0.181349802806)*x_ref[0]**o+(-0.67447696451)*x_ref[0]+(-0.905389539679)*x_ref[1]**o+(-0.0192162208351)*x_ref[1]+(0.475994769738)*x_ref[2]**o+(0.258880626391)*x_ref[2]
        ref[0,2,0,1]=(0.683461009428)*x_ref[0]**o+(0.101633874876)*x_ref[0]+(-0.666120146454)*x_ref[1]**o+(-0.79937847801)*x_ref[1]+(-0.0519440794346)*x_ref[2]**o+(0.412543335301)*x_ref[2]
        ref[0,2,1,0]=(-0.705756906195)*x_ref[0]**o+(-0.914133519297)*x_ref[0]+(0.203329174425)*x_ref[1]**o+(-0.125897560162)*x_ref[1]+(-0.845555978817)*x_ref[2]**o+(-0.376117448037)*x_ref[2]
        ref[0,2,1,1]=(0.129675218633)*x_ref[0]**o+(0.230278463142)*x_ref[0]+(-0.0465924540599)*x_ref[1]**o+(0.618850924426)*x_ref[1]+(-0.981020601489)*x_ref[2]**o+(-0.446786649897)*x_ref[2]
        ref[0,2,2,0]=(0.709293140521)*x_ref[0]**o+(-0.117595235921)*x_ref[0]+(0.00527289231769)*x_ref[1]**o+(0.407867521466)*x_ref[1]+(-0.0513211295402)*x_ref[2]**o+(-0.775396743388)*x_ref[2]
        ref[0,2,2,1]=(0.456146533185)*x_ref[0]**o+(0.357343068604)*x_ref[0]+(-0.596648639307)*x_ref[1]**o+(0.0400180634504)*x_ref[1]+(-0.773508416513)*x_ref[2]**o+(0.873275482135)*x_ref[2]
        ref[0,3,0,0]=(-0.553923806627)*x_ref[0]**o+(0.416008424454)*x_ref[0]+(-0.18857445771)*x_ref[1]**o+(0.668616517137)*x_ref[1]+(0.850445273446)*x_ref[2]**o+(0.0324670478739)*x_ref[2]
        ref[0,3,0,1]=(-0.556129092409)*x_ref[0]**o+(-0.207129568151)*x_ref[0]+(-0.616200225149)*x_ref[1]**o+(0.173694871558)*x_ref[1]+(-0.389943310499)*x_ref[2]**o+(-0.0171456312507)*x_ref[2]
        ref[0,3,1,0]=(0.029375229156)*x_ref[0]**o+(-0.710817311305)*x_ref[0]+(0.487826213322)*x_ref[1]**o+(-0.353110144226)*x_ref[1]+(-0.326240277953)*x_ref[2]**o+(0.435853126161)*x_ref[2]
        ref[0,3,1,1]=(-0.973283235545)*x_ref[0]**o+(0.119293099332)*x_ref[0]+(-0.987231590871)*x_ref[1]**o+(-0.778410371876)*x_ref[1]+(-0.123668803712)*x_ref[2]**o+(0.836409669329)*x_ref[2]
        ref[0,3,2,0]=(-0.953711717094)*x_ref[0]**o+(0.0211462821501)*x_ref[0]+(0.989232030291)*x_ref[1]**o+(0.684525652819)*x_ref[1]+(-0.934499312995)*x_ref[2]**o+(0.0759185573729)*x_ref[2]
        ref[0,3,2,1]=(0.962555034668)*x_ref[0]**o+(-0.344084097742)*x_ref[0]+(0.627919365161)*x_ref[1]**o+(-0.0520511084603)*x_ref[1]+(0.548149991599)*x_ref[2]**o+(0.885011282903)*x_ref[2]
        ref[0,4,0,0]=(-0.604624034071)*x_ref[0]**o+(0.194224357892)*x_ref[0]+(-0.171628912052)*x_ref[1]**o+(0.0361170699761)*x_ref[1]+(0.68259161986)*x_ref[2]**o+(-0.47807189026)*x_ref[2]
        ref[0,4,0,1]=(-0.441358360042)*x_ref[0]**o+(0.854885170771)*x_ref[0]+(0.739245976992)*x_ref[1]**o+(-0.712495821809)*x_ref[1]+(0.786863605702)*x_ref[2]**o+(0.244289741577)*x_ref[2]
        ref[0,4,1,0]=(-0.93437768143)*x_ref[0]**o+(0.909235101604)*x_ref[0]+(0.398343776489)*x_ref[1]**o+(-0.8015457882)*x_ref[1]+(-0.045056420534)*x_ref[2]**o+(-0.0696087139589)*x_ref[2]
        ref[0,4,1,1]=(0.997282980634)*x_ref[0]**o+(-0.523146965095)*x_ref[0]+(0.888300277406)*x_ref[1]**o+(0.0692584552317)*x_ref[1]+(-0.361183524085)*x_ref[2]**o+(0.298003855572)*x_ref[2]
        ref[0,4,2,0]=(0.510704730088)*x_ref[0]**o+(-0.288621287507)*x_ref[0]+(0.694375966407)*x_ref[1]**o+(0.438558764462)*x_ref[1]+(0.999793245458)*x_ref[2]**o+(0.146242565823)*x_ref[2]
        ref[0,4,2,1]=(0.461717367889)*x_ref[0]**o+(0.585419576536)*x_ref[0]+(0.172514336678)*x_ref[1]**o+(-0.372717466388)*x_ref[1]+(0.354162503692)*x_ref[2]**o+(0.465497372897)*x_ref[2]
        ref[1,0,0,0]=(0.536758620313)*x_ref[0]**o+(0.428277991055)*x_ref[0]+(-0.60617935999)*x_ref[1]**o+(-0.889897682367)*x_ref[1]+(-0.326378737219)*x_ref[2]**o+(0.735260248159)*x_ref[2]
        ref[1,0,0,1]=(-0.629157417304)*x_ref[0]**o+(-0.00171372949641)*x_ref[0]+(0.110550936359)*x_ref[1]**o+(-0.418314227619)*x_ref[1]+(0.968147035552)*x_ref[2]**o+(-0.58304546387)*x_ref[2]
        ref[1,0,1,0]=(0.528767817229)*x_ref[0]**o+(0.111986940482)*x_ref[0]+(0.853287140505)*x_ref[1]**o+(0.551761795603)*x_ref[1]+(-0.848026685332)*x_ref[2]**o+(0.799725876993)*x_ref[2]
        ref[1,0,1,1]=(0.154974399407)*x_ref[0]**o+(-0.662212037776)*x_ref[0]+(0.628908352518)*x_ref[1]**o+(-0.0175576929699)*x_ref[1]+(-0.0501253203152)*x_ref[2]**o+(0.147651837545)*x_ref[2]
        ref[1,0,2,0]=(0.53576644491)*x_ref[0]**o+(-0.935865707952)*x_ref[0]+(-0.256921357737)*x_ref[1]**o+(0.452311436169)*x_ref[1]+(-0.449646854354)*x_ref[2]**o+(0.0137632023964)*x_ref[2]
        ref[1,0,2,1]=(-0.103323182353)*x_ref[0]**o+(-0.0636296047221)*x_ref[0]+(-0.681883331812)*x_ref[1]**o+(-0.284589673307)*x_ref[1]+(-0.525524911635)*x_ref[2]**o+(0.525757772842)*x_ref[2]
        ref[1,1,0,0]=(0.106186299307)*x_ref[0]**o+(0.193833690397)*x_ref[0]+(0.755686652252)*x_ref[1]**o+(-0.954503081175)*x_ref[1]+(-0.946058471992)*x_ref[2]**o+(0.258850283735)*x_ref[2]
        ref[1,1,0,1]=(-0.352425152661)*x_ref[0]**o+(-0.955207432287)*x_ref[0]+(0.946989898911)*x_ref[1]**o+(0.426365715943)*x_ref[1]+(-0.225898340088)*x_ref[2]**o+(-0.108613105442)*x_ref[2]
        ref[1,1,1,0]=(0.505488482593)*x_ref[0]**o+(-0.0193114896733)*x_ref[0]+(-0.809537000564)*x_ref[1]**o+(0.39999431472)*x_ref[1]+(-0.968315699948)*x_ref[2]**o+(0.672744543525)*x_ref[2]
        ref[1,1,1,1]=(0.0967454491182)*x_ref[0]**o+(-0.608840080579)*x_ref[0]+(0.287989672484)*x_ref[1]**o+(-0.989435924681)*x_ref[1]+(0.46833865882)*x_ref[2]**o+(0.0211042567481)*x_ref[2]
        ref[1,1,2,0]=(-0.461777513493)*x_ref[0]**o+(-0.539619783995)*x_ref[0]+(-0.719051252797)*x_ref[1]**o+(-0.980491094424)*x_ref[1]+(0.526872369892)*x_ref[2]**o+(0.534649090672)*x_ref[2]
        ref[1,1,2,1]=(-0.498068667635)*x_ref[0]**o+(-0.977138256259)*x_ref[0]+(0.388405877405)*x_ref[1]**o+(-0.652220953523)*x_ref[1]+(0.214733833456)*x_ref[2]**o+(0.925064130776)*x_ref[2]
        ref[1,2,0,0]=(-0.728915088302)*x_ref[0]**o+(-0.867653377354)*x_ref[0]+(-0.0240751909884)*x_ref[1]**o+(-0.617932013511)*x_ref[1]+(0.342417783703)*x_ref[2]**o+(0.269029568448)*x_ref[2]
        ref[1,2,0,1]=(0.659030580895)*x_ref[0]**o+(0.460057036148)*x_ref[0]+(0.0556749040721)*x_ref[1]**o+(-0.0943860780082)*x_ref[1]+(0.940068201493)*x_ref[2]**o+(-0.156531042318)*x_ref[2]
        ref[1,2,1,0]=(-0.0373627034239)*x_ref[0]**o+(-0.631862915794)*x_ref[0]+(0.384889489295)*x_ref[1]**o+(-0.85645836005)*x_ref[1]+(-0.448074134473)*x_ref[2]**o+(-0.103825620602)*x_ref[2]
        ref[1,2,1,1]=(-0.19142928199)*x_ref[0]**o+(-0.993719455322)*x_ref[0]+(0.204439624689)*x_ref[1]**o+(-0.0616134769914)*x_ref[1]+(0.581621603387)*x_ref[2]**o+(-0.554871585864)*x_ref[2]
        ref[1,2,2,0]=(0.684658549351)*x_ref[0]**o+(-0.0520805658251)*x_ref[0]+(-0.549405651115)*x_ref[1]**o+(-0.0890223507953)*x_ref[1]+(0.538847108562)*x_ref[2]**o+(0.555575374976)*x_ref[2]
        ref[1,2,2,1]=(-0.197138807828)*x_ref[0]**o+(0.691528695453)*x_ref[0]+(0.337770131451)*x_ref[1]**o+(-0.0662330700506)*x_ref[1]+(-0.556170962811)*x_ref[2]**o+(-0.415321654558)*x_ref[2]
        ref[1,3,0,0]=(-0.255542847866)*x_ref[0]**o+(-0.731103056475)*x_ref[0]+(-0.764119472108)*x_ref[1]**o+(0.264775911336)*x_ref[1]+(0.976604892324)*x_ref[2]**o+(-0.929820558374)*x_ref[2]
        ref[1,3,0,1]=(-0.506143192998)*x_ref[0]**o+(-0.139848041055)*x_ref[0]+(-0.074839899541)*x_ref[1]**o+(-0.35100597507)*x_ref[1]+(-0.127574366194)*x_ref[2]**o+(-0.0084655916902)*x_ref[2]
        ref[1,3,1,0]=(0.099997170805)*x_ref[0]**o+(-0.149089017918)*x_ref[0]+(0.419790611192)*x_ref[1]**o+(-0.148698964676)*x_ref[1]+(0.490423198386)*x_ref[2]**o+(-0.210076971679)*x_ref[2]
        ref[1,3,1,1]=(-0.139371350292)*x_ref[0]**o+(-0.320648977772)*x_ref[0]+(0.836558622998)*x_ref[1]**o+(0.394491153689)*x_ref[1]+(0.0857621395676)*x_ref[2]**o+(-0.709138669156)*x_ref[2]
        ref[1,3,2,0]=(0.743901534332)*x_ref[0]**o+(-0.249486095838)*x_ref[0]+(0.281087339807)*x_ref[1]**o+(0.617669579771)*x_ref[1]+(-0.383135240681)*x_ref[2]**o+(-0.647267353645)*x_ref[2]
        ref[1,3,2,1]=(-0.465587857279)*x_ref[0]**o+(0.578094609537)*x_ref[0]+(0.400668301655)*x_ref[1]**o+(-0.999984469723)*x_ref[1]+(-0.460584024789)*x_ref[2]**o+(0.943466859654)*x_ref[2]
        ref[1,4,0,0]=(-0.246915327144)*x_ref[0]**o+(-0.203163427715)*x_ref[0]+(0.138589009969)*x_ref[1]**o+(-0.705709971036)*x_ref[1]+(0.273085758612)*x_ref[2]**o+(0.901084781824)*x_ref[2]
        ref[1,4,0,1]=(-0.919682459969)*x_ref[0]**o+(-0.476395399467)*x_ref[0]+(0.00638460254264)*x_ref[1]**o+(-0.474710529098)*x_ref[1]+(-0.935681998683)*x_ref[2]**o+(0.461344674976)*x_ref[2]
        ref[1,4,1,0]=(0.8225038116)*x_ref[0]**o+(0.0855444334027)*x_ref[0]+(-0.685983943046)*x_ref[1]**o+(-0.46700703117)*x_ref[1]+(0.760260028729)*x_ref[2]**o+(-0.190951535155)*x_ref[2]
        ref[1,4,1,1]=(0.512940503439)*x_ref[0]**o+(0.998803686398)*x_ref[0]+(0.200914807938)*x_ref[1]**o+(0.793533216819)*x_ref[1]+(0.980475771536)*x_ref[2]**o+(-0.00307515759413)*x_ref[2]
        ref[1,4,2,0]=(0.278761841603)*x_ref[0]**o+(-0.368232992504)*x_ref[0]+(0.939116142167)*x_ref[1]**o+(0.13661977072)*x_ref[1]+(-0.490378973691)*x_ref[2]**o+(-0.409117913136)*x_ref[2]
        ref[1,4,2,1]=(-0.456090186141)*x_ref[0]**o+(-0.881413614887)*x_ref[0]+(-0.951099957833)*x_ref[1]**o+(0.859690794224)*x_ref[1]+(0.0723671154601)*x_ref[2]**o+(0.468823687196)*x_ref[2]
        ref[2,0,0,0]=(-0.415483671814)*x_ref[0]**o+(0.415160471625)*x_ref[0]+(-0.730798143253)*x_ref[1]**o+(-0.90883538519)*x_ref[1]+(-0.241870631235)*x_ref[2]**o+(-0.549577536182)*x_ref[2]
        ref[2,0,0,1]=(0.86985374761)*x_ref[0]**o+(0.378073808491)*x_ref[0]+(0.0923146623212)*x_ref[1]**o+(-0.912604209727)*x_ref[1]+(0.529162434789)*x_ref[2]**o+(-0.135836425669)*x_ref[2]
        ref[2,0,1,0]=(0.702609345242)*x_ref[0]**o+(-0.63859466548)*x_ref[0]+(-0.2668268067)*x_ref[1]**o+(0.107377268995)*x_ref[1]+(-0.575458650368)*x_ref[2]**o+(-0.359972696269)*x_ref[2]
        ref[2,0,1,1]=(-0.169295117134)*x_ref[0]**o+(-0.0780580709275)*x_ref[0]+(0.9803920116)*x_ref[1]**o+(0.794090800797)*x_ref[1]+(0.71271688197)*x_ref[2]**o+(0.265355025589)*x_ref[2]
        ref[2,0,2,0]=(0.700475435131)*x_ref[0]**o+(0.317025249264)*x_ref[0]+(0.885140001575)*x_ref[1]**o+(-0.733992012528)*x_ref[1]+(-0.937990577058)*x_ref[2]**o+(-0.566197708973)*x_ref[2]
        ref[2,0,2,1]=(0.369801749993)*x_ref[0]**o+(0.292267254334)*x_ref[0]+(0.24962184892)*x_ref[1]**o+(-0.912670113872)*x_ref[1]+(0.590235505871)*x_ref[2]**o+(0.199279333126)*x_ref[2]
        ref[2,1,0,0]=(-0.165657934182)*x_ref[0]**o+(0.837030192613)*x_ref[0]+(-0.615301249273)*x_ref[1]**o+(-0.947633937776)*x_ref[1]+(0.916677465451)*x_ref[2]**o+(0.633727682324)*x_ref[2]
        ref[2,1,0,1]=(0.260787013495)*x_ref[0]**o+(0.367146934614)*x_ref[0]+(0.0271903085878)*x_ref[1]**o+(0.853398837558)*x_ref[1]+(-0.381379969861)*x_ref[2]**o+(0.140001581749)*x_ref[2]
        ref[2,1,1,0]=(0.121628758983)*x_ref[0]**o+(-0.212350487769)*x_ref[0]+(-0.198031244861)*x_ref[1]**o+(-0.695620829484)*x_ref[1]+(-0.0533521134282)*x_ref[2]**o+(0.956103910898)*x_ref[2]
        ref[2,1,1,1]=(-0.321102504727)*x_ref[0]**o+(-0.920842546991)*x_ref[0]+(0.082643684589)*x_ref[1]**o+(-0.0960346141926)*x_ref[1]+(0.475269198456)*x_ref[2]**o+(-0.184473007427)*x_ref[2]
        ref[2,1,2,0]=(0.412066628472)*x_ref[0]**o+(-0.667420990332)*x_ref[0]+(0.754028295628)*x_ref[1]**o+(0.967928426879)*x_ref[1]+(-0.129320768473)*x_ref[2]**o+(-0.496085464716)*x_ref[2]
        ref[2,1,2,1]=(-0.379495871082)*x_ref[0]**o+(0.768674450671)*x_ref[0]+(0.786595353344)*x_ref[1]**o+(-0.191866280074)*x_ref[1]+(0.314572026852)*x_ref[2]**o+(-0.0245751436484)*x_ref[2]
        ref[2,2,0,0]=(-0.291171443165)*x_ref[0]**o+(0.857660522727)*x_ref[0]+(-0.0611031323308)*x_ref[1]**o+(0.365691706254)*x_ref[1]+(0.906199720505)*x_ref[2]**o+(0.621747944156)*x_ref[2]
        ref[2,2,0,1]=(-0.110402534173)*x_ref[0]**o+(-0.465943589112)*x_ref[0]+(0.896364485404)*x_ref[1]**o+(0.16234913823)*x_ref[1]+(0.443564946892)*x_ref[2]**o+(0.259358795418)*x_ref[2]
        ref[2,2,1,0]=(-0.480613805459)*x_ref[0]**o+(-0.837751646963)*x_ref[0]+(0.803669049568)*x_ref[1]**o+(-0.859501917481)*x_ref[1]+(0.633523536226)*x_ref[2]**o+(0.351742190573)*x_ref[2]
        ref[2,2,1,1]=(0.133665578883)*x_ref[0]**o+(-0.283052337652)*x_ref[0]+(-0.382761111917)*x_ref[1]**o+(-0.197039443436)*x_ref[1]+(0.081520912443)*x_ref[2]**o+(-0.167181778234)*x_ref[2]
        ref[2,2,2,0]=(0.207787204976)*x_ref[0]**o+(0.245207436631)*x_ref[0]+(0.340092449827)*x_ref[1]**o+(0.541457158709)*x_ref[1]+(-0.606955763027)*x_ref[2]**o+(-0.197506280875)*x_ref[2]
        ref[2,2,2,1]=(0.399727427867)*x_ref[0]**o+(-0.0452717156729)*x_ref[0]+(0.85643441)*x_ref[1]**o+(-0.938966108133)*x_ref[1]+(0.832098573693)*x_ref[2]**o+(-0.375366657117)*x_ref[2]
        ref[2,3,0,0]=(-0.661411409657)*x_ref[0]**o+(0.702058724486)*x_ref[0]+(0.851820224497)*x_ref[1]**o+(-0.984572528809)*x_ref[1]+(0.700617112775)*x_ref[2]**o+(0.990437560702)*x_ref[2]
        ref[2,3,0,1]=(-0.937637737824)*x_ref[0]**o+(0.970774222765)*x_ref[0]+(0.0263154279607)*x_ref[1]**o+(-0.453145473796)*x_ref[1]+(-0.740101764089)*x_ref[2]**o+(0.248604522836)*x_ref[2]
        ref[2,3,1,0]=(0.624217308865)*x_ref[0]**o+(0.682749411369)*x_ref[0]+(0.394444314015)*x_ref[1]**o+(0.288021257523)*x_ref[1]+(-0.943712429774)*x_ref[2]**o+(-0.299272610192)*x_ref[2]
        ref[2,3,1,1]=(0.00512058360895)*x_ref[0]**o+(-0.0263977566592)*x_ref[0]+(-0.769491595627)*x_ref[1]**o+(-0.403091037334)*x_ref[1]+(-0.940178247778)*x_ref[2]**o+(-0.664590180141)*x_ref[2]
        ref[2,3,2,0]=(-0.606361369179)*x_ref[0]**o+(0.0252279171688)*x_ref[0]+(0.0125424811264)*x_ref[1]**o+(-0.249080159522)*x_ref[1]+(0.969599361427)*x_ref[2]**o+(0.369428527325)*x_ref[2]
        ref[2,3,2,1]=(-0.899501409225)*x_ref[0]**o+(0.834437734927)*x_ref[0]+(-0.992484449733)*x_ref[1]**o+(0.00567086116985)*x_ref[1]+(0.112432151044)*x_ref[2]**o+(0.725013775434)*x_ref[2]
        ref[2,4,0,0]=(-0.864861209468)*x_ref[0]**o+(0.475518351325)*x_ref[0]+(-0.575172287508)*x_ref[1]**o+(-0.695754751991)*x_ref[1]+(0.935933655733)*x_ref[2]**o+(-0.156489388279)*x_ref[2]
        ref[2,4,0,1]=(-0.072905918035)*x_ref[0]**o+(0.912617659416)*x_ref[0]+(0.76767638187)*x_ref[1]**o+(0.517738398356)*x_ref[1]+(-0.116771137305)*x_ref[2]**o+(-0.887027853796)*x_ref[2]
        ref[2,4,1,0]=(-0.169351806151)*x_ref[0]**o+(-0.667994142736)*x_ref[0]+(-0.742400183746)*x_ref[1]**o+(-0.345483488309)*x_ref[1]+(0.534459208989)*x_ref[2]**o+(-0.314632556106)*x_ref[2]
        ref[2,4,1,1]=(0.571889164138)*x_ref[0]**o+(-0.603799705213)*x_ref[0]+(0.574246013631)*x_ref[1]**o+(-0.156121300765)*x_ref[1]+(-0.300131083217)*x_ref[2]**o+(0.781527389701)*x_ref[2]
        ref[2,4,2,0]=(-0.433340783774)*x_ref[0]**o+(-0.337577532056)*x_ref[0]+(-0.991017325155)*x_ref[1]**o+(0.169866142612)*x_ref[1]+(0.531407605197)*x_ref[2]**o+(0.91525228665)*x_ref[2]
        ref[2,4,2,1]=(0.837883720242)*x_ref[0]**o+(-0.647682324772)*x_ref[0]+(0.125057981769)*x_ref[1]**o+(-0.734369688409)*x_ref[1]+(0.5864622868)*x_ref[2]**o+(-0.378030111621)*x_ref[2]
        ref[3,0,0,0]=(-0.497406654549)*x_ref[0]**o+(0.773396244829)*x_ref[0]+(0.587080735986)*x_ref[1]**o+(0.80571073543)*x_ref[1]+(-0.621583700283)*x_ref[2]**o+(0.214586226671)*x_ref[2]
        ref[3,0,0,1]=(0.602557120764)*x_ref[0]**o+(0.362884272628)*x_ref[0]+(-0.469709604155)*x_ref[1]**o+(0.570294839284)*x_ref[1]+(-0.977082842755)*x_ref[2]**o+(-0.871580274035)*x_ref[2]
        ref[3,0,1,0]=(0.801845637209)*x_ref[0]**o+(0.65646927893)*x_ref[0]+(0.181490001743)*x_ref[1]**o+(-0.579155840484)*x_ref[1]+(-0.690604456924)*x_ref[2]**o+(0.999798826875)*x_ref[2]
        ref[3,0,1,1]=(-0.0835741146147)*x_ref[0]**o+(0.251697832709)*x_ref[0]+(0.915417127133)*x_ref[1]**o+(0.859412289228)*x_ref[1]+(0.529025246336)*x_ref[2]**o+(0.0290945910046)*x_ref[2]
        ref[3,0,2,0]=(0.43632860403)*x_ref[0]**o+(-0.71378809102)*x_ref[0]+(0.996769149715)*x_ref[1]**o+(-0.694095684973)*x_ref[1]+(0.24606095312)*x_ref[2]**o+(0.0974139948716)*x_ref[2]
        ref[3,0,2,1]=(-0.615800860737)*x_ref[0]**o+(-0.0194774462371)*x_ref[0]+(-0.323266780598)*x_ref[1]**o+(0.70039319915)*x_ref[1]+(0.0990606501268)*x_ref[2]**o+(-0.265400117193)*x_ref[2]
        ref[3,1,0,0]=(0.637788917841)*x_ref[0]**o+(-0.898134292781)*x_ref[0]+(0.397058421405)*x_ref[1]**o+(0.424875003774)*x_ref[1]+(0.493154896437)*x_ref[2]**o+(0.426966521522)*x_ref[2]
        ref[3,1,0,1]=(0.307016370261)*x_ref[0]**o+(0.583037237501)*x_ref[0]+(-0.715510433757)*x_ref[1]**o+(-0.778084356162)*x_ref[1]+(0.624410868665)*x_ref[2]**o+(0.534413229876)*x_ref[2]
        ref[3,1,1,0]=(0.53494197621)*x_ref[0]**o+(0.0347747301727)*x_ref[0]+(0.387683566308)*x_ref[1]**o+(0.79111981305)*x_ref[1]+(0.499346741486)*x_ref[2]**o+(-0.89736311061)*x_ref[2]
        ref[3,1,1,1]=(-0.414505914492)*x_ref[0]**o+(0.738343334487)*x_ref[0]+(0.0839982715753)*x_ref[1]**o+(0.306059781071)*x_ref[1]+(0.205292188103)*x_ref[2]**o+(0.478479664712)*x_ref[2]
        ref[3,1,2,0]=(0.649444916094)*x_ref[0]**o+(-0.616363283285)*x_ref[0]+(-0.890710011556)*x_ref[1]**o+(-0.55638070962)*x_ref[1]+(0.63397005463)*x_ref[2]**o+(-0.296632593945)*x_ref[2]
        ref[3,1,2,1]=(0.867500426375)*x_ref[0]**o+(0.087056846585)*x_ref[0]+(-0.347167872287)*x_ref[1]**o+(-0.593410409382)*x_ref[1]+(-0.0853161439992)*x_ref[2]**o+(0.32855927156)*x_ref[2]
        ref[3,2,0,0]=(0.176880322402)*x_ref[0]**o+(-0.342939990014)*x_ref[0]+(0.927975205318)*x_ref[1]**o+(0.395166493409)*x_ref[1]+(-0.114428155908)*x_ref[2]**o+(0.518356015938)*x_ref[2]
        ref[3,2,0,1]=(-0.560834784775)*x_ref[0]**o+(0.989745819909)*x_ref[0]+(-0.201732235802)*x_ref[1]**o+(-0.953838293418)*x_ref[1]+(-0.797223863765)*x_ref[2]**o+(-0.44533752286)*x_ref[2]
        ref[3,2,1,0]=(-0.83129861256)*x_ref[0]**o+(-0.768067351258)*x_ref[0]+(-0.564704574946)*x_ref[1]**o+(0.419961478425)*x_ref[1]+(0.561243126484)*x_ref[2]**o+(0.536547545961)*x_ref[2]
        ref[3,2,1,1]=(-0.245876224002)*x_ref[0]**o+(-0.0399839529678)*x_ref[0]+(-0.728071866725)*x_ref[1]**o+(0.68478402008)*x_ref[1]+(0.907895423543)*x_ref[2]**o+(0.454698589619)*x_ref[2]
        ref[3,2,2,0]=(-0.412012705012)*x_ref[0]**o+(-0.575272496028)*x_ref[0]+(0.877861906104)*x_ref[1]**o+(-0.586319475679)*x_ref[1]+(0.668997061563)*x_ref[2]**o+(-0.418342969428)*x_ref[2]
        ref[3,2,2,1]=(0.270232215934)*x_ref[0]**o+(0.236642564222)*x_ref[0]+(-0.0693779443521)*x_ref[1]**o+(0.77329883812)*x_ref[1]+(-0.743424502879)*x_ref[2]**o+(0.296073856497)*x_ref[2]
        ref[3,3,0,0]=(-0.0659865092333)*x_ref[0]**o+(-0.645286369943)*x_ref[0]+(0.746214220465)*x_ref[1]**o+(-0.549592848453)*x_ref[1]+(-0.194552740752)*x_ref[2]**o+(0.136734604935)*x_ref[2]
        ref[3,3,0,1]=(0.0407640382533)*x_ref[0]**o+(-0.720002890774)*x_ref[0]+(-0.902500650998)*x_ref[1]**o+(-0.0763091975748)*x_ref[1]+(-0.70279308164)*x_ref[2]**o+(-0.412384977965)*x_ref[2]
        ref[3,3,1,0]=(0.460226151062)*x_ref[0]**o+(-0.359911871464)*x_ref[0]+(0.654909639152)*x_ref[1]**o+(-0.574154331236)*x_ref[1]+(-0.226246538275)*x_ref[2]**o+(-0.168145308987)*x_ref[2]
        ref[3,3,1,1]=(-0.966169813904)*x_ref[0]**o+(-0.784593122487)*x_ref[0]+(-0.857455334787)*x_ref[1]**o+(-0.116963343366)*x_ref[1]+(0.231055675695)*x_ref[2]**o+(-0.953091037377)*x_ref[2]
        ref[3,3,2,0]=(-0.796170783237)*x_ref[0]**o+(-0.605030579381)*x_ref[0]+(0.388310055252)*x_ref[1]**o+(-0.504349251868)*x_ref[1]+(0.83840677955)*x_ref[2]**o+(-0.586099851479)*x_ref[2]
        ref[3,3,2,1]=(0.195092011781)*x_ref[0]**o+(-0.358423372419)*x_ref[0]+(-0.643669425281)*x_ref[1]**o+(0.721564003575)*x_ref[1]+(-0.588431193874)*x_ref[2]**o+(-0.378625221371)*x_ref[2]
        ref[3,4,0,0]=(0.585756187041)*x_ref[0]**o+(-0.753400165107)*x_ref[0]+(0.790439809615)*x_ref[1]**o+(0.0473109698221)*x_ref[1]+(0.951768026456)*x_ref[2]**o+(0.423960993524)*x_ref[2]
        ref[3,4,0,1]=(0.32857275647)*x_ref[0]**o+(0.566198295813)*x_ref[0]+(-0.372998380277)*x_ref[1]**o+(0.671551057967)*x_ref[1]+(-0.990512903698)*x_ref[2]**o+(-0.598820986842)*x_ref[2]
        ref[3,4,1,0]=(0.512362923169)*x_ref[0]**o+(0.161976775973)*x_ref[0]+(-0.823370116651)*x_ref[1]**o+(-0.539694552746)*x_ref[1]+(-0.252135888305)*x_ref[2]**o+(-0.323536133539)*x_ref[2]
        ref[3,4,1,1]=(0.342762199745)*x_ref[0]**o+(0.569702637139)*x_ref[0]+(0.514189130968)*x_ref[1]**o+(-0.219203654565)*x_ref[1]+(-0.839606678057)*x_ref[2]**o+(0.374386070142)*x_ref[2]
        ref[3,4,2,0]=(0.0568713695577)*x_ref[0]**o+(-0.494349350234)*x_ref[0]+(0.829657653367)*x_ref[1]**o+(0.541489358189)*x_ref[1]+(-0.9469529858)*x_ref[2]**o+(0.230257255487)*x_ref[2]
        ref[3,4,2,1]=(0.874227414064)*x_ref[0]**o+(0.76572262681)*x_ref[0]+(0.305716770957)*x_ref[1]**o+(0.225904980409)*x_ref[1]+(-0.10063057638)*x_ref[2]**o+(-0.87920628616)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.736277927355)*x[0]**o+(-0.476414432477)*x[0]+(0.402171901957)*x[1]**o+(0.513013989524)*x[1]
        ref=(-0.736277927355)*x_ref[0]**o+(-0.476414432477)*x_ref[0]+(0.402171901957)*x_ref[1]**o+(0.513013989524)*x_ref[1]
      else:
        s=(-0.660090850382)*x[0]**o+(-0.0241444775117)*x[0]+(-0.931668488162)*x[1]**o+(0.16225007976)*x[1]+(-0.819642300119)*x[2]**o+(0.617450214622)*x[2]
        ref=(-0.660090850382)*x_ref[0]**o+(-0.0241444775117)*x_ref[0]+(-0.931668488162)*x_ref[1]**o+(0.16225007976)*x_ref[1]+(-0.819642300119)*x_ref[2]**o+(0.617450214622)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.86278283944)*x[0]**o+(-0.279486703584)*x[0]+(0.194198518517)*x[1]**o+(0.128254964922)*x[1]
        s[1]=(0.0136464795833)*x[0]**o+(-0.345339502844)*x[0]+(-0.728145864724)*x[1]**o+(-0.924329335672)*x[1]
        ref[0]=(0.86278283944)*x_ref[0]**o+(-0.279486703584)*x_ref[0]+(0.194198518517)*x_ref[1]**o+(0.128254964922)*x_ref[1]
        ref[1]=(0.0136464795833)*x_ref[0]**o+(-0.345339502844)*x_ref[0]+(-0.728145864724)*x_ref[1]**o+(-0.924329335672)*x_ref[1]
      else:
        s[0]=(0.929705498355)*x[0]**o+(0.306643423816)*x[0]+(0.759529927473)*x[1]**o+(-0.685403980862)*x[1]+(-0.111619761182)*x[2]**o+(0.17003034258)*x[2]
        s[1]=(0.245767218516)*x[0]**o+(-0.844768915901)*x[0]+(0.775524902568)*x[1]**o+(-0.787651539274)*x[1]+(-0.435474729264)*x[2]**o+(0.314357860053)*x[2]
        ref[0]=(0.929705498355)*x_ref[0]**o+(0.306643423816)*x_ref[0]+(0.759529927473)*x_ref[1]**o+(-0.685403980862)*x_ref[1]+(-0.111619761182)*x_ref[2]**o+(0.17003034258)*x_ref[2]
        ref[1]=(0.245767218516)*x_ref[0]**o+(-0.844768915901)*x_ref[0]+(0.775524902568)*x_ref[1]**o+(-0.787651539274)*x_ref[1]+(-0.435474729264)*x_ref[2]**o+(0.314357860053)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0337667400114)*x[0]**o+(-0.144567298896)*x[0]+(-0.729192037092)*x[1]**o+(-0.196957755821)*x[1]
        s[0,1]=(-0.548096676153)*x[0]**o+(0.227315441286)*x[0]+(-0.577631256099)*x[1]**o+(0.18994220721)*x[1]
        s[0,2]=(-0.186709083962)*x[0]**o+(-0.197285531944)*x[0]+(0.948284978854)*x[1]**o+(-0.160726434453)*x[1]
        s[0,3]=(-0.830570775205)*x[0]**o+(-0.253222916576)*x[0]+(-0.648472282053)*x[1]**o+(-0.529978129703)*x[1]
        s[0,4]=(-0.31128056765)*x[0]**o+(-0.270855637019)*x[0]+(-0.0949832365323)*x[1]**o+(-0.980079438215)*x[1]
        s[1,0]=(-0.480406963361)*x[0]**o+(-0.515002496872)*x[0]+(0.731527569308)*x[1]**o+(-0.939907032287)*x[1]
        s[1,1]=(0.485865444556)*x[0]**o+(-0.0284517097445)*x[0]+(-0.375718047775)*x[1]**o+(0.174507894197)*x[1]
        s[1,2]=(0.375072918194)*x[0]**o+(-0.181968235521)*x[0]+(0.54010168758)*x[1]**o+(-0.176274740835)*x[1]
        s[1,3]=(0.858373108753)*x[0]**o+(-0.893829472449)*x[0]+(0.0254515597876)*x[1]**o+(-0.182775473331)*x[1]
        s[1,4]=(0.231841055731)*x[0]**o+(0.0425526152491)*x[0]+(0.361969315465)*x[1]**o+(0.707541143115)*x[1]
        s[2,0]=(-0.239456728122)*x[0]**o+(0.61635858132)*x[0]+(-0.939143396859)*x[1]**o+(-0.787361035499)*x[1]
        s[2,1]=(0.244360860229)*x[0]**o+(0.158360526976)*x[0]+(0.245578056432)*x[1]**o+(0.81105351071)*x[1]
        s[2,2]=(-0.753508481655)*x[0]**o+(0.718380120127)*x[0]+(0.562449137438)*x[1]**o+(-0.0448650181766)*x[1]
        s[2,3]=(0.0901443513604)*x[0]**o+(-0.917555155063)*x[0]+(-0.092792062139)*x[1]**o+(-0.464257045436)*x[1]
        s[2,4]=(0.55127541805)*x[0]**o+(-0.638282785441)*x[0]+(-0.496842401395)*x[1]**o+(-0.159370133271)*x[1]
        s[3,0]=(-0.159892028098)*x[0]**o+(-0.409391269251)*x[0]+(-0.360297648355)*x[1]**o+(-0.868392608184)*x[1]
        s[3,1]=(-0.957845254495)*x[0]**o+(0.837876852047)*x[0]+(0.172913675004)*x[1]**o+(-0.66122565147)*x[1]
        s[3,2]=(0.440455002168)*x[0]**o+(-0.30893214485)*x[0]+(0.740852753456)*x[1]**o+(-0.512664946771)*x[1]
        s[3,3]=(0.697352443519)*x[0]**o+(-0.07285854251)*x[0]+(0.504340563232)*x[1]**o+(-0.0975858236356)*x[1]
        s[3,4]=(-0.721383935365)*x[0]**o+(-0.555062115517)*x[0]+(0.340560152933)*x[1]**o+(-0.594112991805)*x[1]
        ref[0,0]=(0.0337667400114)*x_ref[0]**o+(-0.144567298896)*x_ref[0]+(-0.729192037092)*x_ref[1]**o+(-0.196957755821)*x_ref[1]
        ref[0,1]=(-0.548096676153)*x_ref[0]**o+(0.227315441286)*x_ref[0]+(-0.577631256099)*x_ref[1]**o+(0.18994220721)*x_ref[1]
        ref[0,2]=(-0.186709083962)*x_ref[0]**o+(-0.197285531944)*x_ref[0]+(0.948284978854)*x_ref[1]**o+(-0.160726434453)*x_ref[1]
        ref[0,3]=(-0.830570775205)*x_ref[0]**o+(-0.253222916576)*x_ref[0]+(-0.648472282053)*x_ref[1]**o+(-0.529978129703)*x_ref[1]
        ref[0,4]=(-0.31128056765)*x_ref[0]**o+(-0.270855637019)*x_ref[0]+(-0.0949832365323)*x_ref[1]**o+(-0.980079438215)*x_ref[1]
        ref[1,0]=(-0.480406963361)*x_ref[0]**o+(-0.515002496872)*x_ref[0]+(0.731527569308)*x_ref[1]**o+(-0.939907032287)*x_ref[1]
        ref[1,1]=(0.485865444556)*x_ref[0]**o+(-0.0284517097445)*x_ref[0]+(-0.375718047775)*x_ref[1]**o+(0.174507894197)*x_ref[1]
        ref[1,2]=(0.375072918194)*x_ref[0]**o+(-0.181968235521)*x_ref[0]+(0.54010168758)*x_ref[1]**o+(-0.176274740835)*x_ref[1]
        ref[1,3]=(0.858373108753)*x_ref[0]**o+(-0.893829472449)*x_ref[0]+(0.0254515597876)*x_ref[1]**o+(-0.182775473331)*x_ref[1]
        ref[1,4]=(0.231841055731)*x_ref[0]**o+(0.0425526152491)*x_ref[0]+(0.361969315465)*x_ref[1]**o+(0.707541143115)*x_ref[1]
        ref[2,0]=(-0.239456728122)*x_ref[0]**o+(0.61635858132)*x_ref[0]+(-0.939143396859)*x_ref[1]**o+(-0.787361035499)*x_ref[1]
        ref[2,1]=(0.244360860229)*x_ref[0]**o+(0.158360526976)*x_ref[0]+(0.245578056432)*x_ref[1]**o+(0.81105351071)*x_ref[1]
        ref[2,2]=(-0.753508481655)*x_ref[0]**o+(0.718380120127)*x_ref[0]+(0.562449137438)*x_ref[1]**o+(-0.0448650181766)*x_ref[1]
        ref[2,3]=(0.0901443513604)*x_ref[0]**o+(-0.917555155063)*x_ref[0]+(-0.092792062139)*x_ref[1]**o+(-0.464257045436)*x_ref[1]
        ref[2,4]=(0.55127541805)*x_ref[0]**o+(-0.638282785441)*x_ref[0]+(-0.496842401395)*x_ref[1]**o+(-0.159370133271)*x_ref[1]
        ref[3,0]=(-0.159892028098)*x_ref[0]**o+(-0.409391269251)*x_ref[0]+(-0.360297648355)*x_ref[1]**o+(-0.868392608184)*x_ref[1]
        ref[3,1]=(-0.957845254495)*x_ref[0]**o+(0.837876852047)*x_ref[0]+(0.172913675004)*x_ref[1]**o+(-0.66122565147)*x_ref[1]
        ref[3,2]=(0.440455002168)*x_ref[0]**o+(-0.30893214485)*x_ref[0]+(0.740852753456)*x_ref[1]**o+(-0.512664946771)*x_ref[1]
        ref[3,3]=(0.697352443519)*x_ref[0]**o+(-0.07285854251)*x_ref[0]+(0.504340563232)*x_ref[1]**o+(-0.0975858236356)*x_ref[1]
        ref[3,4]=(-0.721383935365)*x_ref[0]**o+(-0.555062115517)*x_ref[0]+(0.340560152933)*x_ref[1]**o+(-0.594112991805)*x_ref[1]
      else:
        s[0,0]=(-0.967377639498)*x[0]**o+(-0.109920285309)*x[0]+(0.790023694689)*x[1]**o+(0.790135197999)*x[1]+(-0.376354957184)*x[2]**o+(0.315664191487)*x[2]
        s[0,1]=(0.650749725502)*x[0]**o+(0.674379444488)*x[0]+(-0.752025927149)*x[1]**o+(0.431506037287)*x[1]+(-0.084874783658)*x[2]**o+(-0.0888630081146)*x[2]
        s[0,2]=(-0.763286591921)*x[0]**o+(0.232714380191)*x[0]+(0.274698563825)*x[1]**o+(0.13990714623)*x[1]+(0.282390738049)*x[2]**o+(-0.184118626498)*x[2]
        s[0,3]=(0.34722730236)*x[0]**o+(0.418090664447)*x[0]+(-0.0838934385)*x[1]**o+(0.118081782538)*x[1]+(0.707359962969)*x[2]**o+(0.010498166132)*x[2]
        s[0,4]=(-0.820019724515)*x[0]**o+(-0.476361009274)*x[0]+(0.962907202308)*x[1]**o+(0.838318225627)*x[1]+(-0.905815700124)*x[2]**o+(-0.961550206244)*x[2]
        s[1,0]=(-0.0649909942694)*x[0]**o+(-0.328625970273)*x[0]+(0.610572667455)*x[1]**o+(-0.268664458718)*x[1]+(-0.313756386239)*x[2]**o+(0.720402364594)*x[2]
        s[1,1]=(0.0456136625907)*x[0]**o+(-0.622033665985)*x[0]+(-0.979924107437)*x[1]**o+(0.241507237423)*x[1]+(-0.370645584336)*x[2]**o+(-0.618155561342)*x[2]
        s[1,2]=(-0.852061303718)*x[0]**o+(0.561047462531)*x[0]+(-0.223593329251)*x[1]**o+(-0.692921684975)*x[1]+(0.80422655603)*x[2]**o+(0.743428212724)*x[2]
        s[1,3]=(0.955200963309)*x[0]**o+(-0.801150153584)*x[0]+(-0.727872659346)*x[1]**o+(0.64852846397)*x[1]+(-0.0405590568783)*x[2]**o+(-0.681671077115)*x[2]
        s[1,4]=(0.0993688158354)*x[0]**o+(0.727178434695)*x[0]+(-0.834406417429)*x[1]**o+(0.605517045886)*x[1]+(0.684003006822)*x[2]**o+(0.883465420474)*x[2]
        s[2,0]=(0.248900715036)*x[0]**o+(-0.759820597611)*x[0]+(-0.814124541845)*x[1]**o+(0.587467485061)*x[1]+(-0.330748439312)*x[2]**o+(-0.47257010003)*x[2]
        s[2,1]=(0.222439965082)*x[0]**o+(0.766620883901)*x[0]+(-0.650782200796)*x[1]**o+(-0.751282035166)*x[1]+(-0.782467147285)*x[2]**o+(-0.667654575856)*x[2]
        s[2,2]=(-0.851718757105)*x[0]**o+(-0.295924519468)*x[0]+(0.536983910701)*x[1]**o+(0.70777617491)*x[1]+(-0.741563944842)*x[2]**o+(0.212433737988)*x[2]
        s[2,3]=(0.697731035116)*x[0]**o+(-0.592559519318)*x[0]+(0.484117466251)*x[1]**o+(0.167951462995)*x[1]+(0.0533744104194)*x[2]**o+(0.110675471197)*x[2]
        s[2,4]=(0.282157423925)*x[0]**o+(0.439920183895)*x[0]+(0.185095912521)*x[1]**o+(0.370041462063)*x[1]+(0.450240823523)*x[2]**o+(-0.479623803278)*x[2]
        s[3,0]=(0.292902368083)*x[0]**o+(0.368230887824)*x[0]+(0.637303670164)*x[1]**o+(0.715357664832)*x[1]+(0.937283036957)*x[2]**o+(0.479775130282)*x[2]
        s[3,1]=(0.824148876813)*x[0]**o+(-0.926110112375)*x[0]+(0.0450323833338)*x[1]**o+(-0.191108374438)*x[1]+(-0.463357711258)*x[2]**o+(0.297966369624)*x[2]
        s[3,2]=(-0.901438562502)*x[0]**o+(-0.738382228581)*x[0]+(-0.1241453933)*x[1]**o+(-0.814572423284)*x[1]+(0.0656304038666)*x[2]**o+(0.306737038857)*x[2]
        s[3,3]=(0.00375128114386)*x[0]**o+(0.78799039913)*x[0]+(-0.72519128443)*x[1]**o+(-0.645654054935)*x[1]+(0.784379143536)*x[2]**o+(-0.652046298822)*x[2]
        s[3,4]=(0.399305042359)*x[0]**o+(-0.336232215448)*x[0]+(0.303432659398)*x[1]**o+(0.390226881075)*x[1]+(-0.291910473153)*x[2]**o+(-0.709277333147)*x[2]
        ref[0,0]=(-0.967377639498)*x_ref[0]**o+(-0.109920285309)*x_ref[0]+(0.790023694689)*x_ref[1]**o+(0.790135197999)*x_ref[1]+(-0.376354957184)*x_ref[2]**o+(0.315664191487)*x_ref[2]
        ref[0,1]=(0.650749725502)*x_ref[0]**o+(0.674379444488)*x_ref[0]+(-0.752025927149)*x_ref[1]**o+(0.431506037287)*x_ref[1]+(-0.084874783658)*x_ref[2]**o+(-0.0888630081146)*x_ref[2]
        ref[0,2]=(-0.763286591921)*x_ref[0]**o+(0.232714380191)*x_ref[0]+(0.274698563825)*x_ref[1]**o+(0.13990714623)*x_ref[1]+(0.282390738049)*x_ref[2]**o+(-0.184118626498)*x_ref[2]
        ref[0,3]=(0.34722730236)*x_ref[0]**o+(0.418090664447)*x_ref[0]+(-0.0838934385)*x_ref[1]**o+(0.118081782538)*x_ref[1]+(0.707359962969)*x_ref[2]**o+(0.010498166132)*x_ref[2]
        ref[0,4]=(-0.820019724515)*x_ref[0]**o+(-0.476361009274)*x_ref[0]+(0.962907202308)*x_ref[1]**o+(0.838318225627)*x_ref[1]+(-0.905815700124)*x_ref[2]**o+(-0.961550206244)*x_ref[2]
        ref[1,0]=(-0.0649909942694)*x_ref[0]**o+(-0.328625970273)*x_ref[0]+(0.610572667455)*x_ref[1]**o+(-0.268664458718)*x_ref[1]+(-0.313756386239)*x_ref[2]**o+(0.720402364594)*x_ref[2]
        ref[1,1]=(0.0456136625907)*x_ref[0]**o+(-0.622033665985)*x_ref[0]+(-0.979924107437)*x_ref[1]**o+(0.241507237423)*x_ref[1]+(-0.370645584336)*x_ref[2]**o+(-0.618155561342)*x_ref[2]
        ref[1,2]=(-0.852061303718)*x_ref[0]**o+(0.561047462531)*x_ref[0]+(-0.223593329251)*x_ref[1]**o+(-0.692921684975)*x_ref[1]+(0.80422655603)*x_ref[2]**o+(0.743428212724)*x_ref[2]
        ref[1,3]=(0.955200963309)*x_ref[0]**o+(-0.801150153584)*x_ref[0]+(-0.727872659346)*x_ref[1]**o+(0.64852846397)*x_ref[1]+(-0.0405590568783)*x_ref[2]**o+(-0.681671077115)*x_ref[2]
        ref[1,4]=(0.0993688158354)*x_ref[0]**o+(0.727178434695)*x_ref[0]+(-0.834406417429)*x_ref[1]**o+(0.605517045886)*x_ref[1]+(0.684003006822)*x_ref[2]**o+(0.883465420474)*x_ref[2]
        ref[2,0]=(0.248900715036)*x_ref[0]**o+(-0.759820597611)*x_ref[0]+(-0.814124541845)*x_ref[1]**o+(0.587467485061)*x_ref[1]+(-0.330748439312)*x_ref[2]**o+(-0.47257010003)*x_ref[2]
        ref[2,1]=(0.222439965082)*x_ref[0]**o+(0.766620883901)*x_ref[0]+(-0.650782200796)*x_ref[1]**o+(-0.751282035166)*x_ref[1]+(-0.782467147285)*x_ref[2]**o+(-0.667654575856)*x_ref[2]
        ref[2,2]=(-0.851718757105)*x_ref[0]**o+(-0.295924519468)*x_ref[0]+(0.536983910701)*x_ref[1]**o+(0.70777617491)*x_ref[1]+(-0.741563944842)*x_ref[2]**o+(0.212433737988)*x_ref[2]
        ref[2,3]=(0.697731035116)*x_ref[0]**o+(-0.592559519318)*x_ref[0]+(0.484117466251)*x_ref[1]**o+(0.167951462995)*x_ref[1]+(0.0533744104194)*x_ref[2]**o+(0.110675471197)*x_ref[2]
        ref[2,4]=(0.282157423925)*x_ref[0]**o+(0.439920183895)*x_ref[0]+(0.185095912521)*x_ref[1]**o+(0.370041462063)*x_ref[1]+(0.450240823523)*x_ref[2]**o+(-0.479623803278)*x_ref[2]
        ref[3,0]=(0.292902368083)*x_ref[0]**o+(0.368230887824)*x_ref[0]+(0.637303670164)*x_ref[1]**o+(0.715357664832)*x_ref[1]+(0.937283036957)*x_ref[2]**o+(0.479775130282)*x_ref[2]
        ref[3,1]=(0.824148876813)*x_ref[0]**o+(-0.926110112375)*x_ref[0]+(0.0450323833338)*x_ref[1]**o+(-0.191108374438)*x_ref[1]+(-0.463357711258)*x_ref[2]**o+(0.297966369624)*x_ref[2]
        ref[3,2]=(-0.901438562502)*x_ref[0]**o+(-0.738382228581)*x_ref[0]+(-0.1241453933)*x_ref[1]**o+(-0.814572423284)*x_ref[1]+(0.0656304038666)*x_ref[2]**o+(0.306737038857)*x_ref[2]
        ref[3,3]=(0.00375128114386)*x_ref[0]**o+(0.78799039913)*x_ref[0]+(-0.72519128443)*x_ref[1]**o+(-0.645654054935)*x_ref[1]+(0.784379143536)*x_ref[2]**o+(-0.652046298822)*x_ref[2]
        ref[3,4]=(0.399305042359)*x_ref[0]**o+(-0.336232215448)*x_ref[0]+(0.303432659398)*x_ref[1]**o+(0.390226881075)*x_ref[1]+(-0.291910473153)*x_ref[2]**o+(-0.709277333147)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.488879032434)*x[0]**o+(-0.419085910765)*x[0]+(-0.755718625911)*x[1]**o+(-0.254428186752)*x[1]
        s[0,0,1]=(-0.904839660242)*x[0]**o+(-0.618819153867)*x[0]+(-0.671293222257)*x[1]**o+(-0.366943993112)*x[1]
        s[0,1,0]=(-0.953474338603)*x[0]**o+(0.550673593041)*x[0]+(0.533702014747)*x[1]**o+(0.81060141963)*x[1]
        s[0,1,1]=(-0.0356898595559)*x[0]**o+(0.241340079721)*x[0]+(-0.240381862881)*x[1]**o+(0.635463326935)*x[1]
        s[1,0,0]=(0.35633722978)*x[0]**o+(0.943833120316)*x[0]+(0.452388847946)*x[1]**o+(-0.0998245755659)*x[1]
        s[1,0,1]=(-0.831534150513)*x[0]**o+(-0.680412614304)*x[0]+(-0.856553348266)*x[1]**o+(0.219998922098)*x[1]
        s[1,1,0]=(-0.200130387995)*x[0]**o+(0.387228507883)*x[0]+(0.780953622765)*x[1]**o+(-0.64089653682)*x[1]
        s[1,1,1]=(0.0932824107394)*x[0]**o+(-0.362539555401)*x[0]+(-0.29927067753)*x[1]**o+(0.224901647646)*x[1]
        s[2,0,0]=(0.863924771778)*x[0]**o+(0.909358579156)*x[0]+(-0.434836327319)*x[1]**o+(-0.631414294919)*x[1]
        s[2,0,1]=(0.31668035823)*x[0]**o+(0.180052554722)*x[0]+(-0.797251079846)*x[1]**o+(-0.477114935654)*x[1]
        s[2,1,0]=(-0.450532839951)*x[0]**o+(-0.438773315612)*x[0]+(0.530733342787)*x[1]**o+(0.227746538396)*x[1]
        s[2,1,1]=(-0.488584124758)*x[0]**o+(0.128512414903)*x[0]+(0.706652500801)*x[1]**o+(-0.829998769069)*x[1]
        s[3,0,0]=(0.27862017316)*x[0]**o+(0.369176262774)*x[0]+(-0.581486650387)*x[1]**o+(-0.175123673806)*x[1]
        s[3,0,1]=(-0.628352626986)*x[0]**o+(0.690194440108)*x[0]+(-0.687923341892)*x[1]**o+(0.948107923339)*x[1]
        s[3,1,0]=(-0.256476381198)*x[0]**o+(0.213358011656)*x[0]+(-0.840677898993)*x[1]**o+(0.670007577587)*x[1]
        s[3,1,1]=(-0.401373206415)*x[0]**o+(-0.812712823241)*x[0]+(-0.574574543144)*x[1]**o+(0.692960697563)*x[1]
        s[4,0,0]=(-0.971423526579)*x[0]**o+(0.337902086118)*x[0]+(-0.953952226509)*x[1]**o+(0.7328508807)*x[1]
        s[4,0,1]=(0.479940027288)*x[0]**o+(-0.708892375928)*x[0]+(0.54410706006)*x[1]**o+(0.373807311811)*x[1]
        s[4,1,0]=(0.583981464664)*x[0]**o+(-0.106566397504)*x[0]+(-0.891529762026)*x[1]**o+(0.0821843680382)*x[1]
        s[4,1,1]=(-0.223312577496)*x[0]**o+(0.221190368202)*x[0]+(-0.366639258124)*x[1]**o+(0.797102014958)*x[1]
        s[5,0,0]=(0.438772003872)*x[0]**o+(-0.134932042724)*x[0]+(-0.424371460472)*x[1]**o+(0.203989551975)*x[1]
        s[5,0,1]=(-0.479444924379)*x[0]**o+(0.0335454463879)*x[0]+(-0.345877770388)*x[1]**o+(0.887826142155)*x[1]
        s[5,1,0]=(-0.412406950127)*x[0]**o+(-0.174208136285)*x[0]+(-0.189962208005)*x[1]**o+(-0.733302596704)*x[1]
        s[5,1,1]=(0.690213898435)*x[0]**o+(-0.33706829523)*x[0]+(-0.663636888358)*x[1]**o+(0.902361917167)*x[1]
        ref[0,0,0]=(0.488879032434)*x_ref[0]**o+(-0.419085910765)*x_ref[0]+(-0.755718625911)*x_ref[1]**o+(-0.254428186752)*x_ref[1]
        ref[0,0,1]=(-0.904839660242)*x_ref[0]**o+(-0.618819153867)*x_ref[0]+(-0.671293222257)*x_ref[1]**o+(-0.366943993112)*x_ref[1]
        ref[0,1,0]=(-0.953474338603)*x_ref[0]**o+(0.550673593041)*x_ref[0]+(0.533702014747)*x_ref[1]**o+(0.81060141963)*x_ref[1]
        ref[0,1,1]=(-0.0356898595559)*x_ref[0]**o+(0.241340079721)*x_ref[0]+(-0.240381862881)*x_ref[1]**o+(0.635463326935)*x_ref[1]
        ref[1,0,0]=(0.35633722978)*x_ref[0]**o+(0.943833120316)*x_ref[0]+(0.452388847946)*x_ref[1]**o+(-0.0998245755659)*x_ref[1]
        ref[1,0,1]=(-0.831534150513)*x_ref[0]**o+(-0.680412614304)*x_ref[0]+(-0.856553348266)*x_ref[1]**o+(0.219998922098)*x_ref[1]
        ref[1,1,0]=(-0.200130387995)*x_ref[0]**o+(0.387228507883)*x_ref[0]+(0.780953622765)*x_ref[1]**o+(-0.64089653682)*x_ref[1]
        ref[1,1,1]=(0.0932824107394)*x_ref[0]**o+(-0.362539555401)*x_ref[0]+(-0.29927067753)*x_ref[1]**o+(0.224901647646)*x_ref[1]
        ref[2,0,0]=(0.863924771778)*x_ref[0]**o+(0.909358579156)*x_ref[0]+(-0.434836327319)*x_ref[1]**o+(-0.631414294919)*x_ref[1]
        ref[2,0,1]=(0.31668035823)*x_ref[0]**o+(0.180052554722)*x_ref[0]+(-0.797251079846)*x_ref[1]**o+(-0.477114935654)*x_ref[1]
        ref[2,1,0]=(-0.450532839951)*x_ref[0]**o+(-0.438773315612)*x_ref[0]+(0.530733342787)*x_ref[1]**o+(0.227746538396)*x_ref[1]
        ref[2,1,1]=(-0.488584124758)*x_ref[0]**o+(0.128512414903)*x_ref[0]+(0.706652500801)*x_ref[1]**o+(-0.829998769069)*x_ref[1]
        ref[3,0,0]=(0.27862017316)*x_ref[0]**o+(0.369176262774)*x_ref[0]+(-0.581486650387)*x_ref[1]**o+(-0.175123673806)*x_ref[1]
        ref[3,0,1]=(-0.628352626986)*x_ref[0]**o+(0.690194440108)*x_ref[0]+(-0.687923341892)*x_ref[1]**o+(0.948107923339)*x_ref[1]
        ref[3,1,0]=(-0.256476381198)*x_ref[0]**o+(0.213358011656)*x_ref[0]+(-0.840677898993)*x_ref[1]**o+(0.670007577587)*x_ref[1]
        ref[3,1,1]=(-0.401373206415)*x_ref[0]**o+(-0.812712823241)*x_ref[0]+(-0.574574543144)*x_ref[1]**o+(0.692960697563)*x_ref[1]
        ref[4,0,0]=(-0.971423526579)*x_ref[0]**o+(0.337902086118)*x_ref[0]+(-0.953952226509)*x_ref[1]**o+(0.7328508807)*x_ref[1]
        ref[4,0,1]=(0.479940027288)*x_ref[0]**o+(-0.708892375928)*x_ref[0]+(0.54410706006)*x_ref[1]**o+(0.373807311811)*x_ref[1]
        ref[4,1,0]=(0.583981464664)*x_ref[0]**o+(-0.106566397504)*x_ref[0]+(-0.891529762026)*x_ref[1]**o+(0.0821843680382)*x_ref[1]
        ref[4,1,1]=(-0.223312577496)*x_ref[0]**o+(0.221190368202)*x_ref[0]+(-0.366639258124)*x_ref[1]**o+(0.797102014958)*x_ref[1]
        ref[5,0,0]=(0.438772003872)*x_ref[0]**o+(-0.134932042724)*x_ref[0]+(-0.424371460472)*x_ref[1]**o+(0.203989551975)*x_ref[1]
        ref[5,0,1]=(-0.479444924379)*x_ref[0]**o+(0.0335454463879)*x_ref[0]+(-0.345877770388)*x_ref[1]**o+(0.887826142155)*x_ref[1]
        ref[5,1,0]=(-0.412406950127)*x_ref[0]**o+(-0.174208136285)*x_ref[0]+(-0.189962208005)*x_ref[1]**o+(-0.733302596704)*x_ref[1]
        ref[5,1,1]=(0.690213898435)*x_ref[0]**o+(-0.33706829523)*x_ref[0]+(-0.663636888358)*x_ref[1]**o+(0.902361917167)*x_ref[1]
      else:
        s[0,0,0]=(0.556139202578)*x[0]**o+(0.483108770828)*x[0]+(0.718434815012)*x[1]**o+(0.188244775369)*x[1]+(0.622708148322)*x[2]**o+(-0.630994467257)*x[2]
        s[0,0,1]=(0.451704497908)*x[0]**o+(-0.829972306153)*x[0]+(0.379360817088)*x[1]**o+(-0.238554130947)*x[1]+(-0.893955768332)*x[2]**o+(-0.652900915754)*x[2]
        s[0,1,0]=(-0.675514336491)*x[0]**o+(-0.52159773217)*x[0]+(-0.465109199128)*x[1]**o+(0.257294736754)*x[1]+(0.11567684103)*x[2]**o+(0.834925399585)*x[2]
        s[0,1,1]=(-0.033394605795)*x[0]**o+(0.429522473901)*x[0]+(-0.313816643963)*x[1]**o+(0.486911048289)*x[1]+(0.117894547392)*x[2]**o+(-0.644996938318)*x[2]
        s[1,0,0]=(0.493831313861)*x[0]**o+(0.902573197159)*x[0]+(0.187792903803)*x[1]**o+(-0.509959040152)*x[1]+(0.299926429238)*x[2]**o+(-0.381495320902)*x[2]
        s[1,0,1]=(0.811104276564)*x[0]**o+(0.14715524742)*x[0]+(-0.559655747476)*x[1]**o+(0.293583759106)*x[1]+(-0.715695502258)*x[2]**o+(-0.153041714729)*x[2]
        s[1,1,0]=(0.772329934175)*x[0]**o+(-0.650188321276)*x[0]+(0.0255873645356)*x[1]**o+(-0.900184876279)*x[1]+(-0.0811029635763)*x[2]**o+(-0.639863228931)*x[2]
        s[1,1,1]=(0.012849266133)*x[0]**o+(-0.534722528695)*x[0]+(-0.447904371394)*x[1]**o+(-0.800402363397)*x[1]+(0.630178460753)*x[2]**o+(0.997086157166)*x[2]
        s[2,0,0]=(-0.904892832858)*x[0]**o+(0.666114322253)*x[0]+(0.583997936075)*x[1]**o+(-0.269007848965)*x[1]+(-0.157884735811)*x[2]**o+(-0.0686213349147)*x[2]
        s[2,0,1]=(0.667977269603)*x[0]**o+(-0.640683848003)*x[0]+(-0.835083232597)*x[1]**o+(0.259208008388)*x[1]+(0.123418168733)*x[2]**o+(-0.742362322352)*x[2]
        s[2,1,0]=(-0.1478220156)*x[0]**o+(-0.640018988934)*x[0]+(-0.563977249742)*x[1]**o+(-0.311946814831)*x[1]+(0.91681445185)*x[2]**o+(-0.527960425061)*x[2]
        s[2,1,1]=(0.840717283475)*x[0]**o+(-0.776370401124)*x[0]+(-0.409078261075)*x[1]**o+(0.195327340291)*x[1]+(0.229432377346)*x[2]**o+(-0.25948787778)*x[2]
        s[3,0,0]=(-0.207904126371)*x[0]**o+(-0.214014439618)*x[0]+(0.0662583628322)*x[1]**o+(0.660136794478)*x[1]+(0.64474063125)*x[2]**o+(0.187827434354)*x[2]
        s[3,0,1]=(0.389334881352)*x[0]**o+(-0.0683701164906)*x[0]+(-0.830390000851)*x[1]**o+(-0.858716119574)*x[1]+(-0.101328021481)*x[2]**o+(0.85170211833)*x[2]
        s[3,1,0]=(0.0591220139583)*x[0]**o+(0.0662651765216)*x[0]+(-0.0324716950732)*x[1]**o+(0.124718755281)*x[1]+(0.159309216695)*x[2]**o+(-0.909374271062)*x[2]
        s[3,1,1]=(0.122716304679)*x[0]**o+(-0.573353577297)*x[0]+(-0.527652673105)*x[1]**o+(-0.726777764476)*x[1]+(0.807740799068)*x[2]**o+(-0.804635168227)*x[2]
        s[4,0,0]=(0.4769518359)*x[0]**o+(-0.739383338101)*x[0]+(-0.66504775515)*x[1]**o+(0.316930246851)*x[1]+(-0.364202789983)*x[2]**o+(-0.625328020633)*x[2]
        s[4,0,1]=(0.356791348967)*x[0]**o+(-0.786156538192)*x[0]+(0.806815720338)*x[1]**o+(-0.869965211663)*x[1]+(-0.0801679071555)*x[2]**o+(0.432551799627)*x[2]
        s[4,1,0]=(-0.874016752615)*x[0]**o+(0.697364820089)*x[0]+(0.584294471412)*x[1]**o+(0.545279752858)*x[1]+(-0.503589699758)*x[2]**o+(-0.821026187376)*x[2]
        s[4,1,1]=(0.411492383085)*x[0]**o+(-0.546368425944)*x[0]+(0.269146946026)*x[1]**o+(0.983029434649)*x[1]+(0.0288073822165)*x[2]**o+(0.737609217447)*x[2]
        s[5,0,0]=(-0.611031098314)*x[0]**o+(0.509946604728)*x[0]+(0.413409344768)*x[1]**o+(-0.569432793888)*x[1]+(-0.361402473522)*x[2]**o+(0.399423497542)*x[2]
        s[5,0,1]=(-0.690537233831)*x[0]**o+(0.731302717324)*x[0]+(0.241825928343)*x[1]**o+(-0.715863522048)*x[1]+(-0.991752691576)*x[2]**o+(-0.404815867325)*x[2]
        s[5,1,0]=(0.198486687168)*x[0]**o+(-0.990125426636)*x[0]+(0.187211838401)*x[1]**o+(0.746502352031)*x[1]+(0.590564572255)*x[2]**o+(0.339996821991)*x[2]
        s[5,1,1]=(-0.660239307252)*x[0]**o+(0.718370759986)*x[0]+(-0.261471198229)*x[1]**o+(0.84243447248)*x[1]+(-0.491636834274)*x[2]**o+(0.50442143487)*x[2]
        ref[0,0,0]=(0.556139202578)*x_ref[0]**o+(0.483108770828)*x_ref[0]+(0.718434815012)*x_ref[1]**o+(0.188244775369)*x_ref[1]+(0.622708148322)*x_ref[2]**o+(-0.630994467257)*x_ref[2]
        ref[0,0,1]=(0.451704497908)*x_ref[0]**o+(-0.829972306153)*x_ref[0]+(0.379360817088)*x_ref[1]**o+(-0.238554130947)*x_ref[1]+(-0.893955768332)*x_ref[2]**o+(-0.652900915754)*x_ref[2]
        ref[0,1,0]=(-0.675514336491)*x_ref[0]**o+(-0.52159773217)*x_ref[0]+(-0.465109199128)*x_ref[1]**o+(0.257294736754)*x_ref[1]+(0.11567684103)*x_ref[2]**o+(0.834925399585)*x_ref[2]
        ref[0,1,1]=(-0.033394605795)*x_ref[0]**o+(0.429522473901)*x_ref[0]+(-0.313816643963)*x_ref[1]**o+(0.486911048289)*x_ref[1]+(0.117894547392)*x_ref[2]**o+(-0.644996938318)*x_ref[2]
        ref[1,0,0]=(0.493831313861)*x_ref[0]**o+(0.902573197159)*x_ref[0]+(0.187792903803)*x_ref[1]**o+(-0.509959040152)*x_ref[1]+(0.299926429238)*x_ref[2]**o+(-0.381495320902)*x_ref[2]
        ref[1,0,1]=(0.811104276564)*x_ref[0]**o+(0.14715524742)*x_ref[0]+(-0.559655747476)*x_ref[1]**o+(0.293583759106)*x_ref[1]+(-0.715695502258)*x_ref[2]**o+(-0.153041714729)*x_ref[2]
        ref[1,1,0]=(0.772329934175)*x_ref[0]**o+(-0.650188321276)*x_ref[0]+(0.0255873645356)*x_ref[1]**o+(-0.900184876279)*x_ref[1]+(-0.0811029635763)*x_ref[2]**o+(-0.639863228931)*x_ref[2]
        ref[1,1,1]=(0.012849266133)*x_ref[0]**o+(-0.534722528695)*x_ref[0]+(-0.447904371394)*x_ref[1]**o+(-0.800402363397)*x_ref[1]+(0.630178460753)*x_ref[2]**o+(0.997086157166)*x_ref[2]
        ref[2,0,0]=(-0.904892832858)*x_ref[0]**o+(0.666114322253)*x_ref[0]+(0.583997936075)*x_ref[1]**o+(-0.269007848965)*x_ref[1]+(-0.157884735811)*x_ref[2]**o+(-0.0686213349147)*x_ref[2]
        ref[2,0,1]=(0.667977269603)*x_ref[0]**o+(-0.640683848003)*x_ref[0]+(-0.835083232597)*x_ref[1]**o+(0.259208008388)*x_ref[1]+(0.123418168733)*x_ref[2]**o+(-0.742362322352)*x_ref[2]
        ref[2,1,0]=(-0.1478220156)*x_ref[0]**o+(-0.640018988934)*x_ref[0]+(-0.563977249742)*x_ref[1]**o+(-0.311946814831)*x_ref[1]+(0.91681445185)*x_ref[2]**o+(-0.527960425061)*x_ref[2]
        ref[2,1,1]=(0.840717283475)*x_ref[0]**o+(-0.776370401124)*x_ref[0]+(-0.409078261075)*x_ref[1]**o+(0.195327340291)*x_ref[1]+(0.229432377346)*x_ref[2]**o+(-0.25948787778)*x_ref[2]
        ref[3,0,0]=(-0.207904126371)*x_ref[0]**o+(-0.214014439618)*x_ref[0]+(0.0662583628322)*x_ref[1]**o+(0.660136794478)*x_ref[1]+(0.64474063125)*x_ref[2]**o+(0.187827434354)*x_ref[2]
        ref[3,0,1]=(0.389334881352)*x_ref[0]**o+(-0.0683701164906)*x_ref[0]+(-0.830390000851)*x_ref[1]**o+(-0.858716119574)*x_ref[1]+(-0.101328021481)*x_ref[2]**o+(0.85170211833)*x_ref[2]
        ref[3,1,0]=(0.0591220139583)*x_ref[0]**o+(0.0662651765216)*x_ref[0]+(-0.0324716950732)*x_ref[1]**o+(0.124718755281)*x_ref[1]+(0.159309216695)*x_ref[2]**o+(-0.909374271062)*x_ref[2]
        ref[3,1,1]=(0.122716304679)*x_ref[0]**o+(-0.573353577297)*x_ref[0]+(-0.527652673105)*x_ref[1]**o+(-0.726777764476)*x_ref[1]+(0.807740799068)*x_ref[2]**o+(-0.804635168227)*x_ref[2]
        ref[4,0,0]=(0.4769518359)*x_ref[0]**o+(-0.739383338101)*x_ref[0]+(-0.66504775515)*x_ref[1]**o+(0.316930246851)*x_ref[1]+(-0.364202789983)*x_ref[2]**o+(-0.625328020633)*x_ref[2]
        ref[4,0,1]=(0.356791348967)*x_ref[0]**o+(-0.786156538192)*x_ref[0]+(0.806815720338)*x_ref[1]**o+(-0.869965211663)*x_ref[1]+(-0.0801679071555)*x_ref[2]**o+(0.432551799627)*x_ref[2]
        ref[4,1,0]=(-0.874016752615)*x_ref[0]**o+(0.697364820089)*x_ref[0]+(0.584294471412)*x_ref[1]**o+(0.545279752858)*x_ref[1]+(-0.503589699758)*x_ref[2]**o+(-0.821026187376)*x_ref[2]
        ref[4,1,1]=(0.411492383085)*x_ref[0]**o+(-0.546368425944)*x_ref[0]+(0.269146946026)*x_ref[1]**o+(0.983029434649)*x_ref[1]+(0.0288073822165)*x_ref[2]**o+(0.737609217447)*x_ref[2]
        ref[5,0,0]=(-0.611031098314)*x_ref[0]**o+(0.509946604728)*x_ref[0]+(0.413409344768)*x_ref[1]**o+(-0.569432793888)*x_ref[1]+(-0.361402473522)*x_ref[2]**o+(0.399423497542)*x_ref[2]
        ref[5,0,1]=(-0.690537233831)*x_ref[0]**o+(0.731302717324)*x_ref[0]+(0.241825928343)*x_ref[1]**o+(-0.715863522048)*x_ref[1]+(-0.991752691576)*x_ref[2]**o+(-0.404815867325)*x_ref[2]
        ref[5,1,0]=(0.198486687168)*x_ref[0]**o+(-0.990125426636)*x_ref[0]+(0.187211838401)*x_ref[1]**o+(0.746502352031)*x_ref[1]+(0.590564572255)*x_ref[2]**o+(0.339996821991)*x_ref[2]
        ref[5,1,1]=(-0.660239307252)*x_ref[0]**o+(0.718370759986)*x_ref[0]+(-0.261471198229)*x_ref[1]**o+(0.84243447248)*x_ref[1]+(-0.491636834274)*x_ref[2]**o+(0.50442143487)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.37450835707)*x[0]**o+(0.990288321092)*x[0]+(-0.480930380158)*x[1]**o+(0.202137804833)*x[1]
        s[0,0,0,1]=(-0.875710328537)*x[0]**o+(-0.0940825796501)*x[0]+(0.301742014656)*x[1]**o+(-0.993801115617)*x[1]
        s[0,0,1,0]=(0.97598680404)*x[0]**o+(-0.191639303877)*x[0]+(0.380758510953)*x[1]**o+(0.126468511314)*x[1]
        s[0,0,1,1]=(0.451498155122)*x[0]**o+(0.231131468091)*x[0]+(0.840405822252)*x[1]**o+(0.923623816557)*x[1]
        s[0,0,2,0]=(-0.367499287698)*x[0]**o+(0.0779189049191)*x[0]+(0.287387386382)*x[1]**o+(-0.116085504404)*x[1]
        s[0,0,2,1]=(0.461302744851)*x[0]**o+(0.0283200605131)*x[0]+(0.549004183429)*x[1]**o+(0.660772412474)*x[1]
        s[0,1,0,0]=(0.0755949067732)*x[0]**o+(-0.569435653938)*x[0]+(-0.798149061278)*x[1]**o+(-0.0686631203794)*x[1]
        s[0,1,0,1]=(-0.650624149564)*x[0]**o+(0.192828748367)*x[0]+(0.289825796533)*x[1]**o+(-0.824839147135)*x[1]
        s[0,1,1,0]=(-0.626431984016)*x[0]**o+(-0.00708622835329)*x[0]+(-0.931178901827)*x[1]**o+(-0.574044323706)*x[1]
        s[0,1,1,1]=(-0.569428232892)*x[0]**o+(0.905860534109)*x[0]+(0.0841296314741)*x[1]**o+(0.48761826338)*x[1]
        s[0,1,2,0]=(0.565817935202)*x[0]**o+(0.035779536976)*x[0]+(0.84284398667)*x[1]**o+(-0.73256735365)*x[1]
        s[0,1,2,1]=(-0.665377244222)*x[0]**o+(-0.636499912646)*x[0]+(-0.104610459211)*x[1]**o+(0.831651275587)*x[1]
        s[0,2,0,0]=(-0.800860947164)*x[0]**o+(0.533068766649)*x[0]+(-0.696783562627)*x[1]**o+(-0.321358794768)*x[1]
        s[0,2,0,1]=(-0.588063673229)*x[0]**o+(-0.154202767393)*x[0]+(0.239704744914)*x[1]**o+(0.163822658136)*x[1]
        s[0,2,1,0]=(-0.813659156515)*x[0]**o+(-0.992416486956)*x[0]+(-0.597008554078)*x[1]**o+(0.28361887764)*x[1]
        s[0,2,1,1]=(-0.166263594209)*x[0]**o+(0.791471099325)*x[0]+(-0.671352060229)*x[1]**o+(-0.967418331531)*x[1]
        s[0,2,2,0]=(0.464390496016)*x[0]**o+(0.833444533962)*x[0]+(-0.89271244386)*x[1]**o+(-0.031927536987)*x[1]
        s[0,2,2,1]=(0.824132325693)*x[0]**o+(-0.908594018853)*x[0]+(0.213579818878)*x[1]**o+(-0.69863890097)*x[1]
        s[0,3,0,0]=(0.0856609391471)*x[0]**o+(0.0313146992715)*x[0]+(0.79513792257)*x[1]**o+(-0.580426230893)*x[1]
        s[0,3,0,1]=(0.412719041576)*x[0]**o+(0.259960022707)*x[0]+(0.681604474836)*x[1]**o+(0.930447067081)*x[1]
        s[0,3,1,0]=(-0.81236335246)*x[0]**o+(0.582902992637)*x[0]+(-0.657674496381)*x[1]**o+(-0.809214530492)*x[1]
        s[0,3,1,1]=(0.0108714464824)*x[0]**o+(-0.652658835484)*x[0]+(0.686209117429)*x[1]**o+(-0.469397365861)*x[1]
        s[0,3,2,0]=(-0.893534385322)*x[0]**o+(0.409040375322)*x[0]+(-0.573934925145)*x[1]**o+(0.711005992322)*x[1]
        s[0,3,2,1]=(0.452805819519)*x[0]**o+(0.746144038764)*x[0]+(-0.958508690131)*x[1]**o+(-0.819230696284)*x[1]
        s[0,4,0,0]=(0.937258004003)*x[0]**o+(0.903436203617)*x[0]+(-0.26157800701)*x[1]**o+(-0.080528425668)*x[1]
        s[0,4,0,1]=(-0.687215548433)*x[0]**o+(0.230328062522)*x[0]+(-0.118256837327)*x[1]**o+(-0.0940014727183)*x[1]
        s[0,4,1,0]=(-0.920674911733)*x[0]**o+(0.427538345799)*x[0]+(-0.292211102042)*x[1]**o+(0.0882624369001)*x[1]
        s[0,4,1,1]=(0.606695638414)*x[0]**o+(-0.112342924086)*x[0]+(-0.65923299177)*x[1]**o+(0.780841761165)*x[1]
        s[0,4,2,0]=(-0.52833676208)*x[0]**o+(0.984933828765)*x[0]+(-0.406078172032)*x[1]**o+(-0.516684187549)*x[1]
        s[0,4,2,1]=(0.0756973398128)*x[0]**o+(-0.802937750129)*x[0]+(-0.490045708607)*x[1]**o+(0.443321999269)*x[1]
        s[1,0,0,0]=(-0.350925293536)*x[0]**o+(0.900522176127)*x[0]+(0.0854142525066)*x[1]**o+(0.863691073272)*x[1]
        s[1,0,0,1]=(0.577766188943)*x[0]**o+(-0.365440480716)*x[0]+(0.275785044039)*x[1]**o+(0.0208759284957)*x[1]
        s[1,0,1,0]=(0.93877913201)*x[0]**o+(0.194168118528)*x[0]+(0.789002345841)*x[1]**o+(0.711520807592)*x[1]
        s[1,0,1,1]=(-0.648231665095)*x[0]**o+(-0.179053835732)*x[0]+(-0.0417021488468)*x[1]**o+(0.168796173666)*x[1]
        s[1,0,2,0]=(0.291581059495)*x[0]**o+(-0.33783821993)*x[0]+(-0.116040941117)*x[1]**o+(0.72105551969)*x[1]
        s[1,0,2,1]=(-0.240636936525)*x[0]**o+(-0.175821913621)*x[0]+(0.42023871306)*x[1]**o+(0.140142985813)*x[1]
        s[1,1,0,0]=(0.660374713441)*x[0]**o+(-0.231261742968)*x[0]+(0.853169359766)*x[1]**o+(-0.875479447038)*x[1]
        s[1,1,0,1]=(-0.923871242736)*x[0]**o+(0.57473871192)*x[0]+(0.412950104582)*x[1]**o+(-0.780654847088)*x[1]
        s[1,1,1,0]=(0.841003131696)*x[0]**o+(-0.609509143946)*x[0]+(0.73793477459)*x[1]**o+(0.649241561589)*x[1]
        s[1,1,1,1]=(-0.0859437769074)*x[0]**o+(0.381398238915)*x[0]+(-0.350522885404)*x[1]**o+(0.39379147388)*x[1]
        s[1,1,2,0]=(-0.18572616124)*x[0]**o+(-0.137546839567)*x[0]+(0.561254983838)*x[1]**o+(0.38014910208)*x[1]
        s[1,1,2,1]=(0.348114345885)*x[0]**o+(0.140093390451)*x[0]+(-0.247534457143)*x[1]**o+(-0.729987769498)*x[1]
        s[1,2,0,0]=(0.429017157396)*x[0]**o+(0.235226440205)*x[0]+(0.3628422536)*x[1]**o+(0.861576892575)*x[1]
        s[1,2,0,1]=(-0.907510120093)*x[0]**o+(0.642935046604)*x[0]+(0.9261466131)*x[1]**o+(-0.761150506563)*x[1]
        s[1,2,1,0]=(-0.615325522968)*x[0]**o+(-0.652417143609)*x[0]+(0.17837108607)*x[1]**o+(-0.717146726328)*x[1]
        s[1,2,1,1]=(-0.685429665038)*x[0]**o+(0.588148805439)*x[0]+(-0.923180034037)*x[1]**o+(-0.291944173246)*x[1]
        s[1,2,2,0]=(-0.323454190048)*x[0]**o+(-0.262572099067)*x[0]+(0.301000040115)*x[1]**o+(-0.909720002648)*x[1]
        s[1,2,2,1]=(-0.0948111891739)*x[0]**o+(-0.276061546201)*x[0]+(-0.717727499179)*x[1]**o+(-0.224148452777)*x[1]
        s[1,3,0,0]=(0.205097514101)*x[0]**o+(0.327065321062)*x[0]+(-0.0686643116442)*x[1]**o+(-0.948549018608)*x[1]
        s[1,3,0,1]=(0.604036030601)*x[0]**o+(-0.844486815234)*x[0]+(0.969373006301)*x[1]**o+(-0.878058609434)*x[1]
        s[1,3,1,0]=(-0.418657943937)*x[0]**o+(0.340544958398)*x[0]+(-0.75063042552)*x[1]**o+(0.18344558735)*x[1]
        s[1,3,1,1]=(-0.0477476460454)*x[0]**o+(0.978134566095)*x[0]+(0.644406086468)*x[1]**o+(-0.798034419179)*x[1]
        s[1,3,2,0]=(0.107140685034)*x[0]**o+(0.272786704394)*x[0]+(0.840555334128)*x[1]**o+(-0.470334111563)*x[1]
        s[1,3,2,1]=(0.384533978539)*x[0]**o+(-0.353111544153)*x[0]+(-0.23153624799)*x[1]**o+(0.432173718401)*x[1]
        s[1,4,0,0]=(-0.87826109307)*x[0]**o+(0.221735990763)*x[0]+(-0.78202098232)*x[1]**o+(-0.313027782205)*x[1]
        s[1,4,0,1]=(0.984426230099)*x[0]**o+(0.126914836337)*x[0]+(-0.948893595244)*x[1]**o+(-0.505675618037)*x[1]
        s[1,4,1,0]=(-0.667410610403)*x[0]**o+(0.948305658037)*x[0]+(0.988050832262)*x[1]**o+(-0.248918791348)*x[1]
        s[1,4,1,1]=(-0.0263864210158)*x[0]**o+(-0.580327012392)*x[0]+(0.348487637888)*x[1]**o+(0.868892668155)*x[1]
        s[1,4,2,0]=(-0.733877961216)*x[0]**o+(-0.911527434212)*x[0]+(-0.28163005548)*x[1]**o+(0.883560016744)*x[1]
        s[1,4,2,1]=(-0.192785834067)*x[0]**o+(-0.847904650515)*x[0]+(0.580723979635)*x[1]**o+(-0.669335156232)*x[1]
        s[2,0,0,0]=(0.958168699931)*x[0]**o+(0.24076046389)*x[0]+(-0.958036542606)*x[1]**o+(-0.043918146754)*x[1]
        s[2,0,0,1]=(-0.278946581544)*x[0]**o+(-0.687190175633)*x[0]+(-0.777992189778)*x[1]**o+(-0.863531673141)*x[1]
        s[2,0,1,0]=(-0.898934260505)*x[0]**o+(0.485724452034)*x[0]+(0.603351033362)*x[1]**o+(0.159529868291)*x[1]
        s[2,0,1,1]=(-0.514839945077)*x[0]**o+(-0.81595136879)*x[0]+(-0.3477380492)*x[1]**o+(-0.688930104436)*x[1]
        s[2,0,2,0]=(-0.960670396576)*x[0]**o+(0.0985357322969)*x[0]+(0.629527055349)*x[1]**o+(0.926623281781)*x[1]
        s[2,0,2,1]=(-0.273203836864)*x[0]**o+(-0.399187137413)*x[0]+(0.832946536678)*x[1]**o+(0.0217006310206)*x[1]
        s[2,1,0,0]=(-0.197009304646)*x[0]**o+(0.848123926024)*x[0]+(-0.25900240264)*x[1]**o+(0.866656060994)*x[1]
        s[2,1,0,1]=(0.517502395224)*x[0]**o+(-0.643130570578)*x[0]+(-0.119776492143)*x[1]**o+(-0.783507146107)*x[1]
        s[2,1,1,0]=(0.301724997432)*x[0]**o+(-0.17817186107)*x[0]+(0.490458496534)*x[1]**o+(0.957511419444)*x[1]
        s[2,1,1,1]=(0.123336470212)*x[0]**o+(-0.788887324009)*x[0]+(-0.60272928302)*x[1]**o+(0.76421490053)*x[1]
        s[2,1,2,0]=(-0.490704865336)*x[0]**o+(-0.436595694855)*x[0]+(0.353922207557)*x[1]**o+(-0.600984342959)*x[1]
        s[2,1,2,1]=(-0.0132583349499)*x[0]**o+(0.633323660998)*x[0]+(0.871898397149)*x[1]**o+(0.633440389923)*x[1]
        s[2,2,0,0]=(-0.967690719992)*x[0]**o+(0.783767816205)*x[0]+(-0.761197116769)*x[1]**o+(-0.743782137719)*x[1]
        s[2,2,0,1]=(-0.00698718562172)*x[0]**o+(-0.0907825985417)*x[0]+(-0.347555456925)*x[1]**o+(-0.0470206888935)*x[1]
        s[2,2,1,0]=(-0.534291937297)*x[0]**o+(0.226504796701)*x[0]+(-0.949633959735)*x[1]**o+(0.876580619086)*x[1]
        s[2,2,1,1]=(-0.0854529204744)*x[0]**o+(0.64262215979)*x[0]+(-0.0139718386613)*x[1]**o+(-0.469446776434)*x[1]
        s[2,2,2,0]=(-0.363232080956)*x[0]**o+(-0.170730102979)*x[0]+(0.842559971049)*x[1]**o+(-0.575212952787)*x[1]
        s[2,2,2,1]=(0.863982917432)*x[0]**o+(0.499685449711)*x[0]+(-0.621817585753)*x[1]**o+(-0.129423923828)*x[1]
        s[2,3,0,0]=(-0.82412707087)*x[0]**o+(-0.179035796521)*x[0]+(0.879712642079)*x[1]**o+(-0.592863517326)*x[1]
        s[2,3,0,1]=(-0.243096812517)*x[0]**o+(0.48900653218)*x[0]+(0.583132840752)*x[1]**o+(0.863284042165)*x[1]
        s[2,3,1,0]=(0.708539467913)*x[0]**o+(-0.362846832908)*x[0]+(0.488885306443)*x[1]**o+(0.11504854653)*x[1]
        s[2,3,1,1]=(0.971205974333)*x[0]**o+(-0.379819843412)*x[0]+(-0.53477369508)*x[1]**o+(0.338103748243)*x[1]
        s[2,3,2,0]=(-0.241496648212)*x[0]**o+(-0.426087073582)*x[0]+(-0.956589859481)*x[1]**o+(-0.578733188064)*x[1]
        s[2,3,2,1]=(0.650880333072)*x[0]**o+(0.267027063987)*x[0]+(0.760093042282)*x[1]**o+(0.200523731534)*x[1]
        s[2,4,0,0]=(0.910170271345)*x[0]**o+(0.281643984948)*x[0]+(-0.747138557624)*x[1]**o+(-0.0550309118151)*x[1]
        s[2,4,0,1]=(-0.349010335129)*x[0]**o+(-0.00364309496926)*x[0]+(-0.807770436885)*x[1]**o+(-0.0616602161099)*x[1]
        s[2,4,1,0]=(0.512141616515)*x[0]**o+(-0.929334219558)*x[0]+(-0.963669071062)*x[1]**o+(0.659887358212)*x[1]
        s[2,4,1,1]=(-0.138590598738)*x[0]**o+(-0.155856707442)*x[0]+(-0.0772258372581)*x[1]**o+(0.679657019646)*x[1]
        s[2,4,2,0]=(-0.887443583313)*x[0]**o+(-0.426525176662)*x[0]+(0.376025010808)*x[1]**o+(-0.262326050538)*x[1]
        s[2,4,2,1]=(-0.458553824587)*x[0]**o+(0.116339198696)*x[0]+(0.123872771612)*x[1]**o+(0.515809839916)*x[1]
        s[3,0,0,0]=(-0.0729206411854)*x[0]**o+(-0.510467187501)*x[0]+(0.253610009615)*x[1]**o+(0.145910286704)*x[1]
        s[3,0,0,1]=(-0.550384973025)*x[0]**o+(0.0844842856395)*x[0]+(-0.520575587296)*x[1]**o+(-0.675884738141)*x[1]
        s[3,0,1,0]=(-0.308077215416)*x[0]**o+(0.996474784055)*x[0]+(0.0969573296699)*x[1]**o+(0.0284996811435)*x[1]
        s[3,0,1,1]=(-0.648258160349)*x[0]**o+(0.0318285490499)*x[0]+(0.0472274103746)*x[1]**o+(0.146150131328)*x[1]
        s[3,0,2,0]=(0.0181823241342)*x[0]**o+(0.26190851226)*x[0]+(-0.691813017812)*x[1]**o+(0.8009100087)*x[1]
        s[3,0,2,1]=(-0.0874957101881)*x[0]**o+(0.870556245598)*x[0]+(-0.105597330125)*x[1]**o+(0.342588727073)*x[1]
        s[3,1,0,0]=(-0.522688561902)*x[0]**o+(-0.982164303589)*x[0]+(0.584983804692)*x[1]**o+(-0.793876228528)*x[1]
        s[3,1,0,1]=(-0.389530331812)*x[0]**o+(-0.96456801768)*x[0]+(-0.252358189439)*x[1]**o+(-0.0028692565331)*x[1]
        s[3,1,1,0]=(0.433182446414)*x[0]**o+(0.499789289629)*x[0]+(-0.280867766666)*x[1]**o+(0.713293236302)*x[1]
        s[3,1,1,1]=(0.996696514585)*x[0]**o+(-0.985056111984)*x[0]+(-0.780836677971)*x[1]**o+(-0.672838124952)*x[1]
        s[3,1,2,0]=(0.350670380339)*x[0]**o+(0.461658777851)*x[0]+(0.173718158472)*x[1]**o+(-0.0361435953373)*x[1]
        s[3,1,2,1]=(0.349350935239)*x[0]**o+(0.420242310781)*x[0]+(-0.196833484121)*x[1]**o+(-0.126277066836)*x[1]
        s[3,2,0,0]=(0.893790646542)*x[0]**o+(-0.00903845037459)*x[0]+(0.178807098568)*x[1]**o+(0.233299971548)*x[1]
        s[3,2,0,1]=(-0.0943791780627)*x[0]**o+(0.0695327298132)*x[0]+(-0.604082655563)*x[1]**o+(0.996891558378)*x[1]
        s[3,2,1,0]=(-0.419557382554)*x[0]**o+(0.0237478836942)*x[0]+(0.890148225147)*x[1]**o+(0.239878155335)*x[1]
        s[3,2,1,1]=(-0.569572000668)*x[0]**o+(-0.161312095781)*x[0]+(-0.807036822805)*x[1]**o+(-0.0832452791017)*x[1]
        s[3,2,2,0]=(0.50144810802)*x[0]**o+(-0.187000342991)*x[0]+(0.109794995649)*x[1]**o+(-0.412756985657)*x[1]
        s[3,2,2,1]=(-0.806306419034)*x[0]**o+(0.642993688491)*x[0]+(-0.292257201966)*x[1]**o+(0.797375494056)*x[1]
        s[3,3,0,0]=(-0.255668439927)*x[0]**o+(-0.830046640529)*x[0]+(0.716931237073)*x[1]**o+(-0.602462806139)*x[1]
        s[3,3,0,1]=(-0.979196979205)*x[0]**o+(0.273450220647)*x[0]+(-0.843217096514)*x[1]**o+(0.659951414442)*x[1]
        s[3,3,1,0]=(0.478892696333)*x[0]**o+(0.872933917716)*x[0]+(-0.142146478546)*x[1]**o+(0.161337372681)*x[1]
        s[3,3,1,1]=(0.685834600178)*x[0]**o+(0.0858588717379)*x[0]+(-0.0236340928121)*x[1]**o+(0.019356809354)*x[1]
        s[3,3,2,0]=(-0.887370573844)*x[0]**o+(-0.472393104761)*x[0]+(-0.414262982854)*x[1]**o+(0.0737533482009)*x[1]
        s[3,3,2,1]=(-0.211975027701)*x[0]**o+(0.523851599435)*x[0]+(-0.813017539538)*x[1]**o+(-0.545932201009)*x[1]
        s[3,4,0,0]=(0.382214600916)*x[0]**o+(-0.0219464326647)*x[0]+(-0.400360845588)*x[1]**o+(-0.400386585379)*x[1]
        s[3,4,0,1]=(-0.2748424199)*x[0]**o+(-0.506425849157)*x[0]+(0.792560501603)*x[1]**o+(-0.507204525032)*x[1]
        s[3,4,1,0]=(0.288582520692)*x[0]**o+(-0.279954872894)*x[0]+(0.411298773822)*x[1]**o+(0.806904170953)*x[1]
        s[3,4,1,1]=(0.711803548476)*x[0]**o+(-0.384181310938)*x[0]+(-0.516640633187)*x[1]**o+(0.915840688176)*x[1]
        s[3,4,2,0]=(0.174265351332)*x[0]**o+(-0.895064256281)*x[0]+(-0.363843999908)*x[1]**o+(-0.355873468083)*x[1]
        s[3,4,2,1]=(-0.704343161813)*x[0]**o+(-0.284623504785)*x[0]+(-0.331121960038)*x[1]**o+(-0.648228215962)*x[1]
        ref[0,0,0,0]=(0.37450835707)*x_ref[0]**o+(0.990288321092)*x_ref[0]+(-0.480930380158)*x_ref[1]**o+(0.202137804833)*x_ref[1]
        ref[0,0,0,1]=(-0.875710328537)*x_ref[0]**o+(-0.0940825796501)*x_ref[0]+(0.301742014656)*x_ref[1]**o+(-0.993801115617)*x_ref[1]
        ref[0,0,1,0]=(0.97598680404)*x_ref[0]**o+(-0.191639303877)*x_ref[0]+(0.380758510953)*x_ref[1]**o+(0.126468511314)*x_ref[1]
        ref[0,0,1,1]=(0.451498155122)*x_ref[0]**o+(0.231131468091)*x_ref[0]+(0.840405822252)*x_ref[1]**o+(0.923623816557)*x_ref[1]
        ref[0,0,2,0]=(-0.367499287698)*x_ref[0]**o+(0.0779189049191)*x_ref[0]+(0.287387386382)*x_ref[1]**o+(-0.116085504404)*x_ref[1]
        ref[0,0,2,1]=(0.461302744851)*x_ref[0]**o+(0.0283200605131)*x_ref[0]+(0.549004183429)*x_ref[1]**o+(0.660772412474)*x_ref[1]
        ref[0,1,0,0]=(0.0755949067732)*x_ref[0]**o+(-0.569435653938)*x_ref[0]+(-0.798149061278)*x_ref[1]**o+(-0.0686631203794)*x_ref[1]
        ref[0,1,0,1]=(-0.650624149564)*x_ref[0]**o+(0.192828748367)*x_ref[0]+(0.289825796533)*x_ref[1]**o+(-0.824839147135)*x_ref[1]
        ref[0,1,1,0]=(-0.626431984016)*x_ref[0]**o+(-0.00708622835329)*x_ref[0]+(-0.931178901827)*x_ref[1]**o+(-0.574044323706)*x_ref[1]
        ref[0,1,1,1]=(-0.569428232892)*x_ref[0]**o+(0.905860534109)*x_ref[0]+(0.0841296314741)*x_ref[1]**o+(0.48761826338)*x_ref[1]
        ref[0,1,2,0]=(0.565817935202)*x_ref[0]**o+(0.035779536976)*x_ref[0]+(0.84284398667)*x_ref[1]**o+(-0.73256735365)*x_ref[1]
        ref[0,1,2,1]=(-0.665377244222)*x_ref[0]**o+(-0.636499912646)*x_ref[0]+(-0.104610459211)*x_ref[1]**o+(0.831651275587)*x_ref[1]
        ref[0,2,0,0]=(-0.800860947164)*x_ref[0]**o+(0.533068766649)*x_ref[0]+(-0.696783562627)*x_ref[1]**o+(-0.321358794768)*x_ref[1]
        ref[0,2,0,1]=(-0.588063673229)*x_ref[0]**o+(-0.154202767393)*x_ref[0]+(0.239704744914)*x_ref[1]**o+(0.163822658136)*x_ref[1]
        ref[0,2,1,0]=(-0.813659156515)*x_ref[0]**o+(-0.992416486956)*x_ref[0]+(-0.597008554078)*x_ref[1]**o+(0.28361887764)*x_ref[1]
        ref[0,2,1,1]=(-0.166263594209)*x_ref[0]**o+(0.791471099325)*x_ref[0]+(-0.671352060229)*x_ref[1]**o+(-0.967418331531)*x_ref[1]
        ref[0,2,2,0]=(0.464390496016)*x_ref[0]**o+(0.833444533962)*x_ref[0]+(-0.89271244386)*x_ref[1]**o+(-0.031927536987)*x_ref[1]
        ref[0,2,2,1]=(0.824132325693)*x_ref[0]**o+(-0.908594018853)*x_ref[0]+(0.213579818878)*x_ref[1]**o+(-0.69863890097)*x_ref[1]
        ref[0,3,0,0]=(0.0856609391471)*x_ref[0]**o+(0.0313146992715)*x_ref[0]+(0.79513792257)*x_ref[1]**o+(-0.580426230893)*x_ref[1]
        ref[0,3,0,1]=(0.412719041576)*x_ref[0]**o+(0.259960022707)*x_ref[0]+(0.681604474836)*x_ref[1]**o+(0.930447067081)*x_ref[1]
        ref[0,3,1,0]=(-0.81236335246)*x_ref[0]**o+(0.582902992637)*x_ref[0]+(-0.657674496381)*x_ref[1]**o+(-0.809214530492)*x_ref[1]
        ref[0,3,1,1]=(0.0108714464824)*x_ref[0]**o+(-0.652658835484)*x_ref[0]+(0.686209117429)*x_ref[1]**o+(-0.469397365861)*x_ref[1]
        ref[0,3,2,0]=(-0.893534385322)*x_ref[0]**o+(0.409040375322)*x_ref[0]+(-0.573934925145)*x_ref[1]**o+(0.711005992322)*x_ref[1]
        ref[0,3,2,1]=(0.452805819519)*x_ref[0]**o+(0.746144038764)*x_ref[0]+(-0.958508690131)*x_ref[1]**o+(-0.819230696284)*x_ref[1]
        ref[0,4,0,0]=(0.937258004003)*x_ref[0]**o+(0.903436203617)*x_ref[0]+(-0.26157800701)*x_ref[1]**o+(-0.080528425668)*x_ref[1]
        ref[0,4,0,1]=(-0.687215548433)*x_ref[0]**o+(0.230328062522)*x_ref[0]+(-0.118256837327)*x_ref[1]**o+(-0.0940014727183)*x_ref[1]
        ref[0,4,1,0]=(-0.920674911733)*x_ref[0]**o+(0.427538345799)*x_ref[0]+(-0.292211102042)*x_ref[1]**o+(0.0882624369001)*x_ref[1]
        ref[0,4,1,1]=(0.606695638414)*x_ref[0]**o+(-0.112342924086)*x_ref[0]+(-0.65923299177)*x_ref[1]**o+(0.780841761165)*x_ref[1]
        ref[0,4,2,0]=(-0.52833676208)*x_ref[0]**o+(0.984933828765)*x_ref[0]+(-0.406078172032)*x_ref[1]**o+(-0.516684187549)*x_ref[1]
        ref[0,4,2,1]=(0.0756973398128)*x_ref[0]**o+(-0.802937750129)*x_ref[0]+(-0.490045708607)*x_ref[1]**o+(0.443321999269)*x_ref[1]
        ref[1,0,0,0]=(-0.350925293536)*x_ref[0]**o+(0.900522176127)*x_ref[0]+(0.0854142525066)*x_ref[1]**o+(0.863691073272)*x_ref[1]
        ref[1,0,0,1]=(0.577766188943)*x_ref[0]**o+(-0.365440480716)*x_ref[0]+(0.275785044039)*x_ref[1]**o+(0.0208759284957)*x_ref[1]
        ref[1,0,1,0]=(0.93877913201)*x_ref[0]**o+(0.194168118528)*x_ref[0]+(0.789002345841)*x_ref[1]**o+(0.711520807592)*x_ref[1]
        ref[1,0,1,1]=(-0.648231665095)*x_ref[0]**o+(-0.179053835732)*x_ref[0]+(-0.0417021488468)*x_ref[1]**o+(0.168796173666)*x_ref[1]
        ref[1,0,2,0]=(0.291581059495)*x_ref[0]**o+(-0.33783821993)*x_ref[0]+(-0.116040941117)*x_ref[1]**o+(0.72105551969)*x_ref[1]
        ref[1,0,2,1]=(-0.240636936525)*x_ref[0]**o+(-0.175821913621)*x_ref[0]+(0.42023871306)*x_ref[1]**o+(0.140142985813)*x_ref[1]
        ref[1,1,0,0]=(0.660374713441)*x_ref[0]**o+(-0.231261742968)*x_ref[0]+(0.853169359766)*x_ref[1]**o+(-0.875479447038)*x_ref[1]
        ref[1,1,0,1]=(-0.923871242736)*x_ref[0]**o+(0.57473871192)*x_ref[0]+(0.412950104582)*x_ref[1]**o+(-0.780654847088)*x_ref[1]
        ref[1,1,1,0]=(0.841003131696)*x_ref[0]**o+(-0.609509143946)*x_ref[0]+(0.73793477459)*x_ref[1]**o+(0.649241561589)*x_ref[1]
        ref[1,1,1,1]=(-0.0859437769074)*x_ref[0]**o+(0.381398238915)*x_ref[0]+(-0.350522885404)*x_ref[1]**o+(0.39379147388)*x_ref[1]
        ref[1,1,2,0]=(-0.18572616124)*x_ref[0]**o+(-0.137546839567)*x_ref[0]+(0.561254983838)*x_ref[1]**o+(0.38014910208)*x_ref[1]
        ref[1,1,2,1]=(0.348114345885)*x_ref[0]**o+(0.140093390451)*x_ref[0]+(-0.247534457143)*x_ref[1]**o+(-0.729987769498)*x_ref[1]
        ref[1,2,0,0]=(0.429017157396)*x_ref[0]**o+(0.235226440205)*x_ref[0]+(0.3628422536)*x_ref[1]**o+(0.861576892575)*x_ref[1]
        ref[1,2,0,1]=(-0.907510120093)*x_ref[0]**o+(0.642935046604)*x_ref[0]+(0.9261466131)*x_ref[1]**o+(-0.761150506563)*x_ref[1]
        ref[1,2,1,0]=(-0.615325522968)*x_ref[0]**o+(-0.652417143609)*x_ref[0]+(0.17837108607)*x_ref[1]**o+(-0.717146726328)*x_ref[1]
        ref[1,2,1,1]=(-0.685429665038)*x_ref[0]**o+(0.588148805439)*x_ref[0]+(-0.923180034037)*x_ref[1]**o+(-0.291944173246)*x_ref[1]
        ref[1,2,2,0]=(-0.323454190048)*x_ref[0]**o+(-0.262572099067)*x_ref[0]+(0.301000040115)*x_ref[1]**o+(-0.909720002648)*x_ref[1]
        ref[1,2,2,1]=(-0.0948111891739)*x_ref[0]**o+(-0.276061546201)*x_ref[0]+(-0.717727499179)*x_ref[1]**o+(-0.224148452777)*x_ref[1]
        ref[1,3,0,0]=(0.205097514101)*x_ref[0]**o+(0.327065321062)*x_ref[0]+(-0.0686643116442)*x_ref[1]**o+(-0.948549018608)*x_ref[1]
        ref[1,3,0,1]=(0.604036030601)*x_ref[0]**o+(-0.844486815234)*x_ref[0]+(0.969373006301)*x_ref[1]**o+(-0.878058609434)*x_ref[1]
        ref[1,3,1,0]=(-0.418657943937)*x_ref[0]**o+(0.340544958398)*x_ref[0]+(-0.75063042552)*x_ref[1]**o+(0.18344558735)*x_ref[1]
        ref[1,3,1,1]=(-0.0477476460454)*x_ref[0]**o+(0.978134566095)*x_ref[0]+(0.644406086468)*x_ref[1]**o+(-0.798034419179)*x_ref[1]
        ref[1,3,2,0]=(0.107140685034)*x_ref[0]**o+(0.272786704394)*x_ref[0]+(0.840555334128)*x_ref[1]**o+(-0.470334111563)*x_ref[1]
        ref[1,3,2,1]=(0.384533978539)*x_ref[0]**o+(-0.353111544153)*x_ref[0]+(-0.23153624799)*x_ref[1]**o+(0.432173718401)*x_ref[1]
        ref[1,4,0,0]=(-0.87826109307)*x_ref[0]**o+(0.221735990763)*x_ref[0]+(-0.78202098232)*x_ref[1]**o+(-0.313027782205)*x_ref[1]
        ref[1,4,0,1]=(0.984426230099)*x_ref[0]**o+(0.126914836337)*x_ref[0]+(-0.948893595244)*x_ref[1]**o+(-0.505675618037)*x_ref[1]
        ref[1,4,1,0]=(-0.667410610403)*x_ref[0]**o+(0.948305658037)*x_ref[0]+(0.988050832262)*x_ref[1]**o+(-0.248918791348)*x_ref[1]
        ref[1,4,1,1]=(-0.0263864210158)*x_ref[0]**o+(-0.580327012392)*x_ref[0]+(0.348487637888)*x_ref[1]**o+(0.868892668155)*x_ref[1]
        ref[1,4,2,0]=(-0.733877961216)*x_ref[0]**o+(-0.911527434212)*x_ref[0]+(-0.28163005548)*x_ref[1]**o+(0.883560016744)*x_ref[1]
        ref[1,4,2,1]=(-0.192785834067)*x_ref[0]**o+(-0.847904650515)*x_ref[0]+(0.580723979635)*x_ref[1]**o+(-0.669335156232)*x_ref[1]
        ref[2,0,0,0]=(0.958168699931)*x_ref[0]**o+(0.24076046389)*x_ref[0]+(-0.958036542606)*x_ref[1]**o+(-0.043918146754)*x_ref[1]
        ref[2,0,0,1]=(-0.278946581544)*x_ref[0]**o+(-0.687190175633)*x_ref[0]+(-0.777992189778)*x_ref[1]**o+(-0.863531673141)*x_ref[1]
        ref[2,0,1,0]=(-0.898934260505)*x_ref[0]**o+(0.485724452034)*x_ref[0]+(0.603351033362)*x_ref[1]**o+(0.159529868291)*x_ref[1]
        ref[2,0,1,1]=(-0.514839945077)*x_ref[0]**o+(-0.81595136879)*x_ref[0]+(-0.3477380492)*x_ref[1]**o+(-0.688930104436)*x_ref[1]
        ref[2,0,2,0]=(-0.960670396576)*x_ref[0]**o+(0.0985357322969)*x_ref[0]+(0.629527055349)*x_ref[1]**o+(0.926623281781)*x_ref[1]
        ref[2,0,2,1]=(-0.273203836864)*x_ref[0]**o+(-0.399187137413)*x_ref[0]+(0.832946536678)*x_ref[1]**o+(0.0217006310206)*x_ref[1]
        ref[2,1,0,0]=(-0.197009304646)*x_ref[0]**o+(0.848123926024)*x_ref[0]+(-0.25900240264)*x_ref[1]**o+(0.866656060994)*x_ref[1]
        ref[2,1,0,1]=(0.517502395224)*x_ref[0]**o+(-0.643130570578)*x_ref[0]+(-0.119776492143)*x_ref[1]**o+(-0.783507146107)*x_ref[1]
        ref[2,1,1,0]=(0.301724997432)*x_ref[0]**o+(-0.17817186107)*x_ref[0]+(0.490458496534)*x_ref[1]**o+(0.957511419444)*x_ref[1]
        ref[2,1,1,1]=(0.123336470212)*x_ref[0]**o+(-0.788887324009)*x_ref[0]+(-0.60272928302)*x_ref[1]**o+(0.76421490053)*x_ref[1]
        ref[2,1,2,0]=(-0.490704865336)*x_ref[0]**o+(-0.436595694855)*x_ref[0]+(0.353922207557)*x_ref[1]**o+(-0.600984342959)*x_ref[1]
        ref[2,1,2,1]=(-0.0132583349499)*x_ref[0]**o+(0.633323660998)*x_ref[0]+(0.871898397149)*x_ref[1]**o+(0.633440389923)*x_ref[1]
        ref[2,2,0,0]=(-0.967690719992)*x_ref[0]**o+(0.783767816205)*x_ref[0]+(-0.761197116769)*x_ref[1]**o+(-0.743782137719)*x_ref[1]
        ref[2,2,0,1]=(-0.00698718562172)*x_ref[0]**o+(-0.0907825985417)*x_ref[0]+(-0.347555456925)*x_ref[1]**o+(-0.0470206888935)*x_ref[1]
        ref[2,2,1,0]=(-0.534291937297)*x_ref[0]**o+(0.226504796701)*x_ref[0]+(-0.949633959735)*x_ref[1]**o+(0.876580619086)*x_ref[1]
        ref[2,2,1,1]=(-0.0854529204744)*x_ref[0]**o+(0.64262215979)*x_ref[0]+(-0.0139718386613)*x_ref[1]**o+(-0.469446776434)*x_ref[1]
        ref[2,2,2,0]=(-0.363232080956)*x_ref[0]**o+(-0.170730102979)*x_ref[0]+(0.842559971049)*x_ref[1]**o+(-0.575212952787)*x_ref[1]
        ref[2,2,2,1]=(0.863982917432)*x_ref[0]**o+(0.499685449711)*x_ref[0]+(-0.621817585753)*x_ref[1]**o+(-0.129423923828)*x_ref[1]
        ref[2,3,0,0]=(-0.82412707087)*x_ref[0]**o+(-0.179035796521)*x_ref[0]+(0.879712642079)*x_ref[1]**o+(-0.592863517326)*x_ref[1]
        ref[2,3,0,1]=(-0.243096812517)*x_ref[0]**o+(0.48900653218)*x_ref[0]+(0.583132840752)*x_ref[1]**o+(0.863284042165)*x_ref[1]
        ref[2,3,1,0]=(0.708539467913)*x_ref[0]**o+(-0.362846832908)*x_ref[0]+(0.488885306443)*x_ref[1]**o+(0.11504854653)*x_ref[1]
        ref[2,3,1,1]=(0.971205974333)*x_ref[0]**o+(-0.379819843412)*x_ref[0]+(-0.53477369508)*x_ref[1]**o+(0.338103748243)*x_ref[1]
        ref[2,3,2,0]=(-0.241496648212)*x_ref[0]**o+(-0.426087073582)*x_ref[0]+(-0.956589859481)*x_ref[1]**o+(-0.578733188064)*x_ref[1]
        ref[2,3,2,1]=(0.650880333072)*x_ref[0]**o+(0.267027063987)*x_ref[0]+(0.760093042282)*x_ref[1]**o+(0.200523731534)*x_ref[1]
        ref[2,4,0,0]=(0.910170271345)*x_ref[0]**o+(0.281643984948)*x_ref[0]+(-0.747138557624)*x_ref[1]**o+(-0.0550309118151)*x_ref[1]
        ref[2,4,0,1]=(-0.349010335129)*x_ref[0]**o+(-0.00364309496926)*x_ref[0]+(-0.807770436885)*x_ref[1]**o+(-0.0616602161099)*x_ref[1]
        ref[2,4,1,0]=(0.512141616515)*x_ref[0]**o+(-0.929334219558)*x_ref[0]+(-0.963669071062)*x_ref[1]**o+(0.659887358212)*x_ref[1]
        ref[2,4,1,1]=(-0.138590598738)*x_ref[0]**o+(-0.155856707442)*x_ref[0]+(-0.0772258372581)*x_ref[1]**o+(0.679657019646)*x_ref[1]
        ref[2,4,2,0]=(-0.887443583313)*x_ref[0]**o+(-0.426525176662)*x_ref[0]+(0.376025010808)*x_ref[1]**o+(-0.262326050538)*x_ref[1]
        ref[2,4,2,1]=(-0.458553824587)*x_ref[0]**o+(0.116339198696)*x_ref[0]+(0.123872771612)*x_ref[1]**o+(0.515809839916)*x_ref[1]
        ref[3,0,0,0]=(-0.0729206411854)*x_ref[0]**o+(-0.510467187501)*x_ref[0]+(0.253610009615)*x_ref[1]**o+(0.145910286704)*x_ref[1]
        ref[3,0,0,1]=(-0.550384973025)*x_ref[0]**o+(0.0844842856395)*x_ref[0]+(-0.520575587296)*x_ref[1]**o+(-0.675884738141)*x_ref[1]
        ref[3,0,1,0]=(-0.308077215416)*x_ref[0]**o+(0.996474784055)*x_ref[0]+(0.0969573296699)*x_ref[1]**o+(0.0284996811435)*x_ref[1]
        ref[3,0,1,1]=(-0.648258160349)*x_ref[0]**o+(0.0318285490499)*x_ref[0]+(0.0472274103746)*x_ref[1]**o+(0.146150131328)*x_ref[1]
        ref[3,0,2,0]=(0.0181823241342)*x_ref[0]**o+(0.26190851226)*x_ref[0]+(-0.691813017812)*x_ref[1]**o+(0.8009100087)*x_ref[1]
        ref[3,0,2,1]=(-0.0874957101881)*x_ref[0]**o+(0.870556245598)*x_ref[0]+(-0.105597330125)*x_ref[1]**o+(0.342588727073)*x_ref[1]
        ref[3,1,0,0]=(-0.522688561902)*x_ref[0]**o+(-0.982164303589)*x_ref[0]+(0.584983804692)*x_ref[1]**o+(-0.793876228528)*x_ref[1]
        ref[3,1,0,1]=(-0.389530331812)*x_ref[0]**o+(-0.96456801768)*x_ref[0]+(-0.252358189439)*x_ref[1]**o+(-0.0028692565331)*x_ref[1]
        ref[3,1,1,0]=(0.433182446414)*x_ref[0]**o+(0.499789289629)*x_ref[0]+(-0.280867766666)*x_ref[1]**o+(0.713293236302)*x_ref[1]
        ref[3,1,1,1]=(0.996696514585)*x_ref[0]**o+(-0.985056111984)*x_ref[0]+(-0.780836677971)*x_ref[1]**o+(-0.672838124952)*x_ref[1]
        ref[3,1,2,0]=(0.350670380339)*x_ref[0]**o+(0.461658777851)*x_ref[0]+(0.173718158472)*x_ref[1]**o+(-0.0361435953373)*x_ref[1]
        ref[3,1,2,1]=(0.349350935239)*x_ref[0]**o+(0.420242310781)*x_ref[0]+(-0.196833484121)*x_ref[1]**o+(-0.126277066836)*x_ref[1]
        ref[3,2,0,0]=(0.893790646542)*x_ref[0]**o+(-0.00903845037459)*x_ref[0]+(0.178807098568)*x_ref[1]**o+(0.233299971548)*x_ref[1]
        ref[3,2,0,1]=(-0.0943791780627)*x_ref[0]**o+(0.0695327298132)*x_ref[0]+(-0.604082655563)*x_ref[1]**o+(0.996891558378)*x_ref[1]
        ref[3,2,1,0]=(-0.419557382554)*x_ref[0]**o+(0.0237478836942)*x_ref[0]+(0.890148225147)*x_ref[1]**o+(0.239878155335)*x_ref[1]
        ref[3,2,1,1]=(-0.569572000668)*x_ref[0]**o+(-0.161312095781)*x_ref[0]+(-0.807036822805)*x_ref[1]**o+(-0.0832452791017)*x_ref[1]
        ref[3,2,2,0]=(0.50144810802)*x_ref[0]**o+(-0.187000342991)*x_ref[0]+(0.109794995649)*x_ref[1]**o+(-0.412756985657)*x_ref[1]
        ref[3,2,2,1]=(-0.806306419034)*x_ref[0]**o+(0.642993688491)*x_ref[0]+(-0.292257201966)*x_ref[1]**o+(0.797375494056)*x_ref[1]
        ref[3,3,0,0]=(-0.255668439927)*x_ref[0]**o+(-0.830046640529)*x_ref[0]+(0.716931237073)*x_ref[1]**o+(-0.602462806139)*x_ref[1]
        ref[3,3,0,1]=(-0.979196979205)*x_ref[0]**o+(0.273450220647)*x_ref[0]+(-0.843217096514)*x_ref[1]**o+(0.659951414442)*x_ref[1]
        ref[3,3,1,0]=(0.478892696333)*x_ref[0]**o+(0.872933917716)*x_ref[0]+(-0.142146478546)*x_ref[1]**o+(0.161337372681)*x_ref[1]
        ref[3,3,1,1]=(0.685834600178)*x_ref[0]**o+(0.0858588717379)*x_ref[0]+(-0.0236340928121)*x_ref[1]**o+(0.019356809354)*x_ref[1]
        ref[3,3,2,0]=(-0.887370573844)*x_ref[0]**o+(-0.472393104761)*x_ref[0]+(-0.414262982854)*x_ref[1]**o+(0.0737533482009)*x_ref[1]
        ref[3,3,2,1]=(-0.211975027701)*x_ref[0]**o+(0.523851599435)*x_ref[0]+(-0.813017539538)*x_ref[1]**o+(-0.545932201009)*x_ref[1]
        ref[3,4,0,0]=(0.382214600916)*x_ref[0]**o+(-0.0219464326647)*x_ref[0]+(-0.400360845588)*x_ref[1]**o+(-0.400386585379)*x_ref[1]
        ref[3,4,0,1]=(-0.2748424199)*x_ref[0]**o+(-0.506425849157)*x_ref[0]+(0.792560501603)*x_ref[1]**o+(-0.507204525032)*x_ref[1]
        ref[3,4,1,0]=(0.288582520692)*x_ref[0]**o+(-0.279954872894)*x_ref[0]+(0.411298773822)*x_ref[1]**o+(0.806904170953)*x_ref[1]
        ref[3,4,1,1]=(0.711803548476)*x_ref[0]**o+(-0.384181310938)*x_ref[0]+(-0.516640633187)*x_ref[1]**o+(0.915840688176)*x_ref[1]
        ref[3,4,2,0]=(0.174265351332)*x_ref[0]**o+(-0.895064256281)*x_ref[0]+(-0.363843999908)*x_ref[1]**o+(-0.355873468083)*x_ref[1]
        ref[3,4,2,1]=(-0.704343161813)*x_ref[0]**o+(-0.284623504785)*x_ref[0]+(-0.331121960038)*x_ref[1]**o+(-0.648228215962)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.700519095834)*x[0]**o+(-0.646036033021)*x[0]+(-0.274029307292)*x[1]**o+(0.530317572518)*x[1]+(-0.792541640653)*x[2]**o+(-0.880682673379)*x[2]
        s[0,0,0,1]=(0.528832672893)*x[0]**o+(-0.760075289405)*x[0]+(-0.898495461865)*x[1]**o+(0.322670463092)*x[1]+(0.892711553073)*x[2]**o+(-0.433771853349)*x[2]
        s[0,0,1,0]=(0.718010497697)*x[0]**o+(0.618486176543)*x[0]+(-0.486576139267)*x[1]**o+(0.325448300794)*x[1]+(-0.412519101092)*x[2]**o+(-0.620470491359)*x[2]
        s[0,0,1,1]=(-0.0534804057273)*x[0]**o+(-0.00509205691994)*x[0]+(0.971332123478)*x[1]**o+(0.233826077571)*x[1]+(-0.868817891613)*x[2]**o+(-0.281435180207)*x[2]
        s[0,0,2,0]=(-0.337017966084)*x[0]**o+(-0.157687387867)*x[0]+(0.663362074434)*x[1]**o+(-0.770885866632)*x[1]+(0.0757657613445)*x[2]**o+(0.145704787826)*x[2]
        s[0,0,2,1]=(-0.622638143674)*x[0]**o+(0.692263425833)*x[0]+(-0.774201699147)*x[1]**o+(-0.623439001714)*x[1]+(0.372531552491)*x[2]**o+(0.119810463995)*x[2]
        s[0,1,0,0]=(0.033624490626)*x[0]**o+(0.613642855879)*x[0]+(-0.0316167902744)*x[1]**o+(-0.0291344390495)*x[1]+(0.959446527556)*x[2]**o+(0.0824096613993)*x[2]
        s[0,1,0,1]=(-0.330720880473)*x[0]**o+(0.679944413133)*x[0]+(0.931154199834)*x[1]**o+(0.323183383162)*x[1]+(0.207858701122)*x[2]**o+(-0.803638923634)*x[2]
        s[0,1,1,0]=(0.127521751083)*x[0]**o+(-0.70072489837)*x[0]+(0.979950174564)*x[1]**o+(0.749753439334)*x[1]+(0.323684566704)*x[2]**o+(-0.953658423079)*x[2]
        s[0,1,1,1]=(0.892584697191)*x[0]**o+(-0.507033571849)*x[0]+(-0.155876961943)*x[1]**o+(0.725120384958)*x[1]+(-0.626894392039)*x[2]**o+(-0.637598293592)*x[2]
        s[0,1,2,0]=(0.943780578566)*x[0]**o+(0.671891498451)*x[0]+(-0.700147134085)*x[1]**o+(-0.253018669626)*x[1]+(0.327866423384)*x[2]**o+(0.113748950444)*x[2]
        s[0,1,2,1]=(0.472913509903)*x[0]**o+(0.869176106119)*x[0]+(0.111972787166)*x[1]**o+(0.778171565687)*x[1]+(-0.412982015257)*x[2]**o+(-0.178755623341)*x[2]
        s[0,2,0,0]=(-0.674769287337)*x[0]**o+(0.713528167562)*x[0]+(0.993661629164)*x[1]**o+(0.186566847302)*x[1]+(-0.899704600828)*x[2]**o+(-0.990580436616)*x[2]
        s[0,2,0,1]=(0.440601783568)*x[0]**o+(0.135196515316)*x[0]+(-0.550174311014)*x[1]**o+(0.0656485395837)*x[1]+(0.634636307936)*x[2]**o+(0.766275319334)*x[2]
        s[0,2,1,0]=(-0.981322452576)*x[0]**o+(-0.111241714755)*x[0]+(0.371742919372)*x[1]**o+(-0.174195552868)*x[1]+(0.803053234515)*x[2]**o+(0.139585240156)*x[2]
        s[0,2,1,1]=(0.374155625507)*x[0]**o+(0.625627548529)*x[0]+(0.940841784856)*x[1]**o+(0.270225590359)*x[1]+(-0.9935977654)*x[2]**o+(-0.913522645063)*x[2]
        s[0,2,2,0]=(0.401782031796)*x[0]**o+(0.532848886474)*x[0]+(-0.735502524133)*x[1]**o+(-0.752194262754)*x[1]+(-0.48943693706)*x[2]**o+(-0.31404412358)*x[2]
        s[0,2,2,1]=(0.59477802484)*x[0]**o+(-0.609227477682)*x[0]+(0.434121643967)*x[1]**o+(0.367659592412)*x[1]+(0.150570677497)*x[2]**o+(-0.765818843062)*x[2]
        s[0,3,0,0]=(-0.693590253312)*x[0]**o+(-0.258076694041)*x[0]+(-0.547395998804)*x[1]**o+(0.257874232026)*x[1]+(0.459659864323)*x[2]**o+(-0.94127845744)*x[2]
        s[0,3,0,1]=(-0.309288966076)*x[0]**o+(-0.173191962489)*x[0]+(0.540628102572)*x[1]**o+(-0.0443083494784)*x[1]+(0.0435715587905)*x[2]**o+(-0.301785508858)*x[2]
        s[0,3,1,0]=(0.145159950132)*x[0]**o+(-0.287822257724)*x[0]+(0.759130284956)*x[1]**o+(0.0579397521075)*x[1]+(0.699550687352)*x[2]**o+(0.175478553574)*x[2]
        s[0,3,1,1]=(0.474727781805)*x[0]**o+(-0.123878818145)*x[0]+(-0.805368102867)*x[1]**o+(-0.650609144709)*x[1]+(0.656988695082)*x[2]**o+(-0.0136028573608)*x[2]
        s[0,3,2,0]=(-0.319935169115)*x[0]**o+(0.421488978156)*x[0]+(0.496417273253)*x[1]**o+(-0.129350388592)*x[1]+(0.172506181508)*x[2]**o+(0.762146639203)*x[2]
        s[0,3,2,1]=(0.269722271977)*x[0]**o+(-0.754539522531)*x[0]+(0.992009096198)*x[1]**o+(0.705597359762)*x[1]+(-0.334352203866)*x[2]**o+(0.143158060849)*x[2]
        s[0,4,0,0]=(-0.431531468752)*x[0]**o+(0.555318608886)*x[0]+(0.986613799736)*x[1]**o+(0.251035501881)*x[1]+(-0.406355028528)*x[2]**o+(-0.499620043752)*x[2]
        s[0,4,0,1]=(-0.0627649075987)*x[0]**o+(-0.972278432629)*x[0]+(-0.979219556729)*x[1]**o+(0.224664798935)*x[1]+(0.310699044854)*x[2]**o+(0.127604283478)*x[2]
        s[0,4,1,0]=(-0.267874066531)*x[0]**o+(0.991041579767)*x[0]+(0.790132020473)*x[1]**o+(0.228443092608)*x[1]+(-0.890727088684)*x[2]**o+(0.120869830054)*x[2]
        s[0,4,1,1]=(-0.268575786982)*x[0]**o+(0.78564765473)*x[0]+(0.208376098211)*x[1]**o+(0.348864057531)*x[1]+(0.627737586342)*x[2]**o+(-0.540625434546)*x[2]
        s[0,4,2,0]=(-0.628026871983)*x[0]**o+(-0.637731369616)*x[0]+(0.891005717981)*x[1]**o+(-0.281048323549)*x[1]+(0.373980259828)*x[2]**o+(0.676961901941)*x[2]
        s[0,4,2,1]=(0.12750050336)*x[0]**o+(-0.481299457156)*x[0]+(0.511901526031)*x[1]**o+(0.913595029063)*x[1]+(-0.263389445846)*x[2]**o+(-0.167030063489)*x[2]
        s[1,0,0,0]=(0.785350719519)*x[0]**o+(0.259099548564)*x[0]+(0.267304446557)*x[1]**o+(-0.864615565491)*x[1]+(-0.0824408467375)*x[2]**o+(0.724026653941)*x[2]
        s[1,0,0,1]=(0.741066935778)*x[0]**o+(0.444348447224)*x[0]+(0.950815368396)*x[1]**o+(0.630967241982)*x[1]+(-0.173602147175)*x[2]**o+(-0.758408767101)*x[2]
        s[1,0,1,0]=(0.615189483755)*x[0]**o+(0.113461303397)*x[0]+(-0.204008801629)*x[1]**o+(-0.821819803961)*x[1]+(0.996060487559)*x[2]**o+(0.501575420465)*x[2]
        s[1,0,1,1]=(-0.262723988452)*x[0]**o+(0.54464307861)*x[0]+(-0.776865346115)*x[1]**o+(0.62546224345)*x[1]+(-0.576679080633)*x[2]**o+(-0.641871487717)*x[2]
        s[1,0,2,0]=(0.00882312632838)*x[0]**o+(0.886325126754)*x[0]+(-0.290744552898)*x[1]**o+(-0.191666192698)*x[1]+(-0.58836559482)*x[2]**o+(-0.723263075029)*x[2]
        s[1,0,2,1]=(-0.8634554794)*x[0]**o+(-0.156143013973)*x[0]+(-0.954950131813)*x[1]**o+(0.698956259432)*x[1]+(-0.404041520631)*x[2]**o+(0.00457698889215)*x[2]
        s[1,1,0,0]=(-0.728882693177)*x[0]**o+(-0.900052268016)*x[0]+(-0.852025073682)*x[1]**o+(-0.0533567939337)*x[1]+(-0.209580555028)*x[2]**o+(-0.896328429332)*x[2]
        s[1,1,0,1]=(-0.629571937898)*x[0]**o+(0.165066159679)*x[0]+(-0.538661749525)*x[1]**o+(-0.211485405646)*x[1]+(-0.68390849131)*x[2]**o+(0.395749832239)*x[2]
        s[1,1,1,0]=(0.26332743592)*x[0]**o+(-0.0752845023801)*x[0]+(-0.0561271762911)*x[1]**o+(-0.211107669851)*x[1]+(0.678154574805)*x[2]**o+(-0.819108248966)*x[2]
        s[1,1,1,1]=(0.772808345449)*x[0]**o+(-0.262453525307)*x[0]+(-0.692191575144)*x[1]**o+(0.779456283419)*x[1]+(-0.331784968982)*x[2]**o+(0.137130165506)*x[2]
        s[1,1,2,0]=(-0.126130110627)*x[0]**o+(-0.450688959983)*x[0]+(0.246073262195)*x[1]**o+(0.452795873046)*x[1]+(-0.144810908676)*x[2]**o+(-0.100339618687)*x[2]
        s[1,1,2,1]=(-0.494674080271)*x[0]**o+(-0.92562703505)*x[0]+(-0.0365548469018)*x[1]**o+(0.138971357509)*x[1]+(-0.461030467649)*x[2]**o+(0.560886019716)*x[2]
        s[1,2,0,0]=(0.103018022262)*x[0]**o+(-0.95153827596)*x[0]+(0.075101671678)*x[1]**o+(0.385014541504)*x[1]+(0.430777100819)*x[2]**o+(-0.799382254202)*x[2]
        s[1,2,0,1]=(-0.69504642431)*x[0]**o+(-0.989343939843)*x[0]+(0.614616202255)*x[1]**o+(0.119187311793)*x[1]+(0.677533926016)*x[2]**o+(-0.940961001327)*x[2]
        s[1,2,1,0]=(0.45997486768)*x[0]**o+(0.131324255258)*x[0]+(0.51531436597)*x[1]**o+(0.570477569764)*x[1]+(0.670091635832)*x[2]**o+(0.991528689938)*x[2]
        s[1,2,1,1]=(0.693813612837)*x[0]**o+(0.278464811874)*x[0]+(-0.971105053619)*x[1]**o+(-0.290401167365)*x[1]+(0.563462235533)*x[2]**o+(0.715283195872)*x[2]
        s[1,2,2,0]=(0.245799944524)*x[0]**o+(0.72672691689)*x[0]+(0.807071969617)*x[1]**o+(0.825450000507)*x[1]+(-0.859649268645)*x[2]**o+(0.945957906767)*x[2]
        s[1,2,2,1]=(0.168765200995)*x[0]**o+(-0.789419024479)*x[0]+(-0.539642443181)*x[1]**o+(0.463771463373)*x[1]+(0.915893721757)*x[2]**o+(-0.992301458598)*x[2]
        s[1,3,0,0]=(0.578524691791)*x[0]**o+(-0.615099053887)*x[0]+(-0.307646534709)*x[1]**o+(-0.494406889327)*x[1]+(-0.0189087238376)*x[2]**o+(0.190080821114)*x[2]
        s[1,3,0,1]=(0.915034870134)*x[0]**o+(0.657276161564)*x[0]+(0.806274816666)*x[1]**o+(-0.341020582334)*x[1]+(0.729295103751)*x[2]**o+(0.649468360593)*x[2]
        s[1,3,1,0]=(0.9673609292)*x[0]**o+(-0.315653538456)*x[0]+(-0.0346648713203)*x[1]**o+(0.331870978455)*x[1]+(-0.341269715094)*x[2]**o+(-0.0249349656976)*x[2]
        s[1,3,1,1]=(-0.150933678366)*x[0]**o+(-0.389726747893)*x[0]+(-0.204342564941)*x[1]**o+(-0.711033538116)*x[1]+(-0.671896492614)*x[2]**o+(-0.490644811516)*x[2]
        s[1,3,2,0]=(-0.488823609516)*x[0]**o+(0.985637410007)*x[0]+(0.728491418657)*x[1]**o+(0.36207817836)*x[1]+(0.0810824824344)*x[2]**o+(-0.8152491822)*x[2]
        s[1,3,2,1]=(-0.517875264099)*x[0]**o+(-0.727465681243)*x[0]+(0.211140808921)*x[1]**o+(-0.433707511139)*x[1]+(0.344268020901)*x[2]**o+(-0.847540940817)*x[2]
        s[1,4,0,0]=(-0.121268264679)*x[0]**o+(-0.0117838122301)*x[0]+(-0.63745424162)*x[1]**o+(0.945563449658)*x[1]+(-0.923254628525)*x[2]**o+(0.838843523244)*x[2]
        s[1,4,0,1]=(-0.152712892877)*x[0]**o+(0.115026184086)*x[0]+(0.049691447791)*x[1]**o+(-0.766235899011)*x[1]+(-0.45149176462)*x[2]**o+(0.0476296119238)*x[2]
        s[1,4,1,0]=(-0.172747241975)*x[0]**o+(-0.751040839036)*x[0]+(0.785947613925)*x[1]**o+(0.369802888838)*x[1]+(-0.732967210812)*x[2]**o+(0.341006418057)*x[2]
        s[1,4,1,1]=(-0.280256623684)*x[0]**o+(-0.817596493324)*x[0]+(0.210362099145)*x[1]**o+(-0.690356660671)*x[1]+(0.564051447608)*x[2]**o+(-0.81476511073)*x[2]
        s[1,4,2,0]=(-0.492106835788)*x[0]**o+(0.1875588035)*x[0]+(0.408972075649)*x[1]**o+(-0.0979465471156)*x[1]+(-0.191951030868)*x[2]**o+(0.909540651582)*x[2]
        s[1,4,2,1]=(-0.419018988132)*x[0]**o+(-0.0863722256375)*x[0]+(0.621753648601)*x[1]**o+(-0.0865542477263)*x[1]+(0.44530474031)*x[2]**o+(0.314386126939)*x[2]
        s[2,0,0,0]=(0.65858233579)*x[0]**o+(-0.226735659791)*x[0]+(-0.154886066807)*x[1]**o+(0.456030801777)*x[1]+(-0.930548820846)*x[2]**o+(-0.97883928396)*x[2]
        s[2,0,0,1]=(-0.0589535588936)*x[0]**o+(-0.267387198117)*x[0]+(0.807737417612)*x[1]**o+(0.670756731749)*x[1]+(0.565928465377)*x[2]**o+(0.338838873783)*x[2]
        s[2,0,1,0]=(0.745290598767)*x[0]**o+(0.444308877108)*x[0]+(-0.0653699257348)*x[1]**o+(-0.446686429445)*x[1]+(0.431580278588)*x[2]**o+(-0.895205483299)*x[2]
        s[2,0,1,1]=(0.0515503491363)*x[0]**o+(0.325788398831)*x[0]+(-0.275580931208)*x[1]**o+(-0.591278303275)*x[1]+(-0.231550181927)*x[2]**o+(0.830932642648)*x[2]
        s[2,0,2,0]=(-0.782241049983)*x[0]**o+(0.420935438009)*x[0]+(0.765777512717)*x[1]**o+(0.139685303264)*x[1]+(0.84210391036)*x[2]**o+(-0.799020031388)*x[2]
        s[2,0,2,1]=(0.192115194447)*x[0]**o+(0.821179393437)*x[0]+(-0.29035052348)*x[1]**o+(0.459217831059)*x[1]+(0.24675222867)*x[2]**o+(-0.0295595098143)*x[2]
        s[2,1,0,0]=(0.606333105811)*x[0]**o+(-0.908981052972)*x[0]+(0.194455118385)*x[1]**o+(0.475347895918)*x[1]+(0.154123701129)*x[2]**o+(0.0656401634163)*x[2]
        s[2,1,0,1]=(0.924110494936)*x[0]**o+(-0.155517993703)*x[0]+(-0.105954511753)*x[1]**o+(0.427136658346)*x[1]+(-0.842319072137)*x[2]**o+(0.893854788048)*x[2]
        s[2,1,1,0]=(0.775844998945)*x[0]**o+(0.86288098975)*x[0]+(-0.877212097351)*x[1]**o+(-0.58811092774)*x[1]+(-0.456603143488)*x[2]**o+(0.781738385011)*x[2]
        s[2,1,1,1]=(0.745162233175)*x[0]**o+(0.576206343886)*x[0]+(0.897229190634)*x[1]**o+(-0.748603588172)*x[1]+(0.147340946009)*x[2]**o+(-0.97456869073)*x[2]
        s[2,1,2,0]=(-0.995679867038)*x[0]**o+(-0.512547860004)*x[0]+(-0.125757163987)*x[1]**o+(0.284748805033)*x[1]+(0.748279176563)*x[2]**o+(-0.370905992823)*x[2]
        s[2,1,2,1]=(-0.725652456641)*x[0]**o+(0.303144081957)*x[0]+(0.308580461688)*x[1]**o+(-0.398783760014)*x[1]+(0.256017075184)*x[2]**o+(0.244843828536)*x[2]
        s[2,2,0,0]=(0.262768301493)*x[0]**o+(0.994422492631)*x[0]+(0.215209633103)*x[1]**o+(-0.328208518526)*x[1]+(-0.791000179758)*x[2]**o+(-0.0931810905124)*x[2]
        s[2,2,0,1]=(0.430751983854)*x[0]**o+(0.697487380231)*x[0]+(-0.407533720752)*x[1]**o+(0.542395064178)*x[1]+(0.560493455122)*x[2]**o+(0.862130615461)*x[2]
        s[2,2,1,0]=(0.0427580536739)*x[0]**o+(0.755377456297)*x[0]+(-0.882269449771)*x[1]**o+(0.48654379074)*x[1]+(0.0361187725004)*x[2]**o+(0.1808373915)*x[2]
        s[2,2,1,1]=(-0.0473723357884)*x[0]**o+(0.480334631394)*x[0]+(-0.0913487472458)*x[1]**o+(-0.900524837935)*x[1]+(0.233994712545)*x[2]**o+(-0.0576679105272)*x[2]
        s[2,2,2,0]=(-0.501358552905)*x[0]**o+(-0.794253710665)*x[0]+(-0.368314153003)*x[1]**o+(-0.190680411227)*x[1]+(0.867826651831)*x[2]**o+(0.779026383345)*x[2]
        s[2,2,2,1]=(0.403937335227)*x[0]**o+(0.23341509868)*x[0]+(-0.867133942472)*x[1]**o+(0.0236839523692)*x[1]+(0.530643643468)*x[2]**o+(0.153666452771)*x[2]
        s[2,3,0,0]=(-0.911977982923)*x[0]**o+(0.0651402894633)*x[0]+(-0.150699952611)*x[1]**o+(-0.223159951427)*x[1]+(0.0985310057795)*x[2]**o+(-0.512754531906)*x[2]
        s[2,3,0,1]=(0.433292132106)*x[0]**o+(-0.741489420094)*x[0]+(-0.85661323724)*x[1]**o+(-0.470049843131)*x[1]+(-0.127893768764)*x[2]**o+(0.445865536866)*x[2]
        s[2,3,1,0]=(0.176638607161)*x[0]**o+(-0.804679828936)*x[0]+(0.362072277714)*x[1]**o+(-0.208356054989)*x[1]+(0.270297511604)*x[2]**o+(0.411240202674)*x[2]
        s[2,3,1,1]=(-0.731456178523)*x[0]**o+(-0.494011319472)*x[0]+(-0.656951979236)*x[1]**o+(-0.442233264154)*x[1]+(-0.506683809041)*x[2]**o+(0.212265538187)*x[2]
        s[2,3,2,0]=(-0.551032190098)*x[0]**o+(0.643467191611)*x[0]+(0.543872371582)*x[1]**o+(0.72584876243)*x[1]+(0.985681498699)*x[2]**o+(-0.60209773896)*x[2]
        s[2,3,2,1]=(0.0835903452585)*x[0]**o+(0.440086089408)*x[0]+(-0.578402114944)*x[1]**o+(0.487833997481)*x[1]+(0.25662705791)*x[2]**o+(-0.473020070147)*x[2]
        s[2,4,0,0]=(-0.891315579854)*x[0]**o+(0.952879579742)*x[0]+(-0.227350081544)*x[1]**o+(0.600033334324)*x[1]+(0.179737000053)*x[2]**o+(-0.504558083705)*x[2]
        s[2,4,0,1]=(0.225549297486)*x[0]**o+(-0.543285628363)*x[0]+(0.155686671765)*x[1]**o+(-0.718395371077)*x[1]+(0.674896476991)*x[2]**o+(0.213411998034)*x[2]
        s[2,4,1,0]=(-0.35980223796)*x[0]**o+(0.444715759834)*x[0]+(-0.543703848794)*x[1]**o+(0.183552103878)*x[1]+(0.586337472615)*x[2]**o+(0.213304038734)*x[2]
        s[2,4,1,1]=(-0.292792312207)*x[0]**o+(0.196332398934)*x[0]+(-0.798485726334)*x[1]**o+(0.782220141806)*x[1]+(0.0640127628843)*x[2]**o+(-0.77727238011)*x[2]
        s[2,4,2,0]=(0.932887369251)*x[0]**o+(0.0141383057976)*x[0]+(-0.272693549414)*x[1]**o+(-0.978856349954)*x[1]+(0.89717148371)*x[2]**o+(-0.424084272244)*x[2]
        s[2,4,2,1]=(-0.668326323236)*x[0]**o+(-0.973037477253)*x[0]+(-0.33722487276)*x[1]**o+(-0.867851333935)*x[1]+(0.216920541233)*x[2]**o+(-0.657272351842)*x[2]
        s[3,0,0,0]=(-0.502029864543)*x[0]**o+(-0.22816384379)*x[0]+(0.656486587866)*x[1]**o+(-0.364517454501)*x[1]+(0.166636753172)*x[2]**o+(-0.548031611216)*x[2]
        s[3,0,0,1]=(-0.129789963885)*x[0]**o+(-0.924955467733)*x[0]+(-0.318286603412)*x[1]**o+(0.597531687741)*x[1]+(0.664722349723)*x[2]**o+(0.880847610571)*x[2]
        s[3,0,1,0]=(0.138386711079)*x[0]**o+(-0.304181867173)*x[0]+(0.189482296687)*x[1]**o+(0.0576456570922)*x[1]+(-0.026472114191)*x[2]**o+(-0.509366463558)*x[2]
        s[3,0,1,1]=(-0.177846734744)*x[0]**o+(0.504670109456)*x[0]+(-0.802497403902)*x[1]**o+(-0.920870115368)*x[1]+(0.830468060312)*x[2]**o+(0.76112636422)*x[2]
        s[3,0,2,0]=(0.101147363593)*x[0]**o+(0.678670740322)*x[0]+(0.929981085383)*x[1]**o+(-0.78781173205)*x[1]+(0.728075424496)*x[2]**o+(0.597383557857)*x[2]
        s[3,0,2,1]=(-0.622642327441)*x[0]**o+(0.994549062442)*x[0]+(0.155580041238)*x[1]**o+(-0.359913413167)*x[1]+(-0.125600523598)*x[2]**o+(0.727866512034)*x[2]
        s[3,1,0,0]=(-0.339382309297)*x[0]**o+(0.269183464726)*x[0]+(-0.82410881539)*x[1]**o+(0.523860360233)*x[1]+(0.672374739783)*x[2]**o+(0.304612132802)*x[2]
        s[3,1,0,1]=(-0.398676828773)*x[0]**o+(0.178258323741)*x[0]+(-0.842177075615)*x[1]**o+(0.928781440686)*x[1]+(-0.698371966337)*x[2]**o+(-0.069158873645)*x[2]
        s[3,1,1,0]=(0.517076773904)*x[0]**o+(-0.722930310231)*x[0]+(-0.948481363727)*x[1]**o+(-0.313572917282)*x[1]+(-0.216936906887)*x[2]**o+(-0.617732400043)*x[2]
        s[3,1,1,1]=(-0.506541451394)*x[0]**o+(-0.672313212249)*x[0]+(0.884209908638)*x[1]**o+(0.164981753359)*x[1]+(0.38179585119)*x[2]**o+(0.409267628038)*x[2]
        s[3,1,2,0]=(0.772486931191)*x[0]**o+(0.550551917766)*x[0]+(0.579543984786)*x[1]**o+(0.457817667176)*x[1]+(0.332038328462)*x[2]**o+(-0.565647686923)*x[2]
        s[3,1,2,1]=(-0.934824056723)*x[0]**o+(0.0324683981232)*x[0]+(0.731737791013)*x[1]**o+(0.450277257345)*x[1]+(-0.0505877599703)*x[2]**o+(-0.357925404256)*x[2]
        s[3,2,0,0]=(-0.846960246959)*x[0]**o+(0.998323807325)*x[0]+(-0.854597485859)*x[1]**o+(0.752751841074)*x[1]+(0.371090045318)*x[2]**o+(-0.766611225212)*x[2]
        s[3,2,0,1]=(0.339046155557)*x[0]**o+(-0.302854568693)*x[0]+(-0.944840471752)*x[1]**o+(0.566745543225)*x[1]+(-0.829238416325)*x[2]**o+(0.407197535798)*x[2]
        s[3,2,1,0]=(-0.527237680707)*x[0]**o+(0.571797579499)*x[0]+(0.803890750428)*x[1]**o+(0.518904421497)*x[1]+(0.157105476854)*x[2]**o+(-0.430705894568)*x[2]
        s[3,2,1,1]=(0.0391938839127)*x[0]**o+(-0.151396805547)*x[0]+(0.22856560496)*x[1]**o+(-0.207972615421)*x[1]+(-0.0493211514337)*x[2]**o+(-0.804654404843)*x[2]
        s[3,2,2,0]=(-0.653544001068)*x[0]**o+(-0.414094661044)*x[0]+(0.494230663809)*x[1]**o+(-0.506949140354)*x[1]+(0.520476398187)*x[2]**o+(-0.969930682073)*x[2]
        s[3,2,2,1]=(-0.147585384358)*x[0]**o+(0.775604709346)*x[0]+(0.677576282676)*x[1]**o+(-0.505391680821)*x[1]+(0.704307147449)*x[2]**o+(0.558458914373)*x[2]
        s[3,3,0,0]=(-0.65879046599)*x[0]**o+(0.114409484281)*x[0]+(0.622072786527)*x[1]**o+(0.924427931612)*x[1]+(-0.115550639369)*x[2]**o+(-0.483381484729)*x[2]
        s[3,3,0,1]=(0.198285160028)*x[0]**o+(-0.603311408844)*x[0]+(-0.0471223647065)*x[1]**o+(-0.438133307138)*x[1]+(0.388251801711)*x[2]**o+(-0.0732531194363)*x[2]
        s[3,3,1,0]=(0.150939105573)*x[0]**o+(0.108946746296)*x[0]+(-0.807347867525)*x[1]**o+(-0.256128441651)*x[1]+(0.231286821393)*x[2]**o+(-0.703002240761)*x[2]
        s[3,3,1,1]=(-0.641529661237)*x[0]**o+(-0.120398231904)*x[0]+(0.351100779315)*x[1]**o+(-0.576531857716)*x[1]+(-0.765644574653)*x[2]**o+(0.664743279322)*x[2]
        s[3,3,2,0]=(-0.47897462164)*x[0]**o+(-0.569196375202)*x[0]+(-0.422261130469)*x[1]**o+(0.840196504879)*x[1]+(-0.267300663782)*x[2]**o+(-0.893226595716)*x[2]
        s[3,3,2,1]=(0.605260487329)*x[0]**o+(-0.20572150917)*x[0]+(-0.290112951218)*x[1]**o+(0.887680899465)*x[1]+(-0.240985548337)*x[2]**o+(0.180645577595)*x[2]
        s[3,4,0,0]=(0.165393596846)*x[0]**o+(0.134944494629)*x[0]+(0.515339273402)*x[1]**o+(-0.606138667469)*x[1]+(0.0150579936235)*x[2]**o+(0.261591191815)*x[2]
        s[3,4,0,1]=(0.541923776337)*x[0]**o+(0.905517150748)*x[0]+(-0.728025844328)*x[1]**o+(-0.643009143952)*x[1]+(-0.345013615196)*x[2]**o+(0.457263127345)*x[2]
        s[3,4,1,0]=(-0.423845417015)*x[0]**o+(0.882795189836)*x[0]+(0.99729841901)*x[1]**o+(0.106637707066)*x[1]+(-0.960096403487)*x[2]**o+(0.841258780132)*x[2]
        s[3,4,1,1]=(0.256176917841)*x[0]**o+(0.120931832818)*x[0]+(0.908290189359)*x[1]**o+(-0.870707945117)*x[1]+(0.726505555181)*x[2]**o+(-0.110841837509)*x[2]
        s[3,4,2,0]=(-0.0209556990671)*x[0]**o+(-0.0986852276803)*x[0]+(0.758132562139)*x[1]**o+(0.0827733457302)*x[1]+(-0.689039269813)*x[2]**o+(-0.120148663779)*x[2]
        s[3,4,2,1]=(-0.189903364845)*x[0]**o+(-0.145753796459)*x[0]+(0.608678649752)*x[1]**o+(-0.51842417051)*x[1]+(-0.212194895669)*x[2]**o+(0.321128974995)*x[2]
        ref[0,0,0,0]=(-0.700519095834)*x_ref[0]**o+(-0.646036033021)*x_ref[0]+(-0.274029307292)*x_ref[1]**o+(0.530317572518)*x_ref[1]+(-0.792541640653)*x_ref[2]**o+(-0.880682673379)*x_ref[2]
        ref[0,0,0,1]=(0.528832672893)*x_ref[0]**o+(-0.760075289405)*x_ref[0]+(-0.898495461865)*x_ref[1]**o+(0.322670463092)*x_ref[1]+(0.892711553073)*x_ref[2]**o+(-0.433771853349)*x_ref[2]
        ref[0,0,1,0]=(0.718010497697)*x_ref[0]**o+(0.618486176543)*x_ref[0]+(-0.486576139267)*x_ref[1]**o+(0.325448300794)*x_ref[1]+(-0.412519101092)*x_ref[2]**o+(-0.620470491359)*x_ref[2]
        ref[0,0,1,1]=(-0.0534804057273)*x_ref[0]**o+(-0.00509205691994)*x_ref[0]+(0.971332123478)*x_ref[1]**o+(0.233826077571)*x_ref[1]+(-0.868817891613)*x_ref[2]**o+(-0.281435180207)*x_ref[2]
        ref[0,0,2,0]=(-0.337017966084)*x_ref[0]**o+(-0.157687387867)*x_ref[0]+(0.663362074434)*x_ref[1]**o+(-0.770885866632)*x_ref[1]+(0.0757657613445)*x_ref[2]**o+(0.145704787826)*x_ref[2]
        ref[0,0,2,1]=(-0.622638143674)*x_ref[0]**o+(0.692263425833)*x_ref[0]+(-0.774201699147)*x_ref[1]**o+(-0.623439001714)*x_ref[1]+(0.372531552491)*x_ref[2]**o+(0.119810463995)*x_ref[2]
        ref[0,1,0,0]=(0.033624490626)*x_ref[0]**o+(0.613642855879)*x_ref[0]+(-0.0316167902744)*x_ref[1]**o+(-0.0291344390495)*x_ref[1]+(0.959446527556)*x_ref[2]**o+(0.0824096613993)*x_ref[2]
        ref[0,1,0,1]=(-0.330720880473)*x_ref[0]**o+(0.679944413133)*x_ref[0]+(0.931154199834)*x_ref[1]**o+(0.323183383162)*x_ref[1]+(0.207858701122)*x_ref[2]**o+(-0.803638923634)*x_ref[2]
        ref[0,1,1,0]=(0.127521751083)*x_ref[0]**o+(-0.70072489837)*x_ref[0]+(0.979950174564)*x_ref[1]**o+(0.749753439334)*x_ref[1]+(0.323684566704)*x_ref[2]**o+(-0.953658423079)*x_ref[2]
        ref[0,1,1,1]=(0.892584697191)*x_ref[0]**o+(-0.507033571849)*x_ref[0]+(-0.155876961943)*x_ref[1]**o+(0.725120384958)*x_ref[1]+(-0.626894392039)*x_ref[2]**o+(-0.637598293592)*x_ref[2]
        ref[0,1,2,0]=(0.943780578566)*x_ref[0]**o+(0.671891498451)*x_ref[0]+(-0.700147134085)*x_ref[1]**o+(-0.253018669626)*x_ref[1]+(0.327866423384)*x_ref[2]**o+(0.113748950444)*x_ref[2]
        ref[0,1,2,1]=(0.472913509903)*x_ref[0]**o+(0.869176106119)*x_ref[0]+(0.111972787166)*x_ref[1]**o+(0.778171565687)*x_ref[1]+(-0.412982015257)*x_ref[2]**o+(-0.178755623341)*x_ref[2]
        ref[0,2,0,0]=(-0.674769287337)*x_ref[0]**o+(0.713528167562)*x_ref[0]+(0.993661629164)*x_ref[1]**o+(0.186566847302)*x_ref[1]+(-0.899704600828)*x_ref[2]**o+(-0.990580436616)*x_ref[2]
        ref[0,2,0,1]=(0.440601783568)*x_ref[0]**o+(0.135196515316)*x_ref[0]+(-0.550174311014)*x_ref[1]**o+(0.0656485395837)*x_ref[1]+(0.634636307936)*x_ref[2]**o+(0.766275319334)*x_ref[2]
        ref[0,2,1,0]=(-0.981322452576)*x_ref[0]**o+(-0.111241714755)*x_ref[0]+(0.371742919372)*x_ref[1]**o+(-0.174195552868)*x_ref[1]+(0.803053234515)*x_ref[2]**o+(0.139585240156)*x_ref[2]
        ref[0,2,1,1]=(0.374155625507)*x_ref[0]**o+(0.625627548529)*x_ref[0]+(0.940841784856)*x_ref[1]**o+(0.270225590359)*x_ref[1]+(-0.9935977654)*x_ref[2]**o+(-0.913522645063)*x_ref[2]
        ref[0,2,2,0]=(0.401782031796)*x_ref[0]**o+(0.532848886474)*x_ref[0]+(-0.735502524133)*x_ref[1]**o+(-0.752194262754)*x_ref[1]+(-0.48943693706)*x_ref[2]**o+(-0.31404412358)*x_ref[2]
        ref[0,2,2,1]=(0.59477802484)*x_ref[0]**o+(-0.609227477682)*x_ref[0]+(0.434121643967)*x_ref[1]**o+(0.367659592412)*x_ref[1]+(0.150570677497)*x_ref[2]**o+(-0.765818843062)*x_ref[2]
        ref[0,3,0,0]=(-0.693590253312)*x_ref[0]**o+(-0.258076694041)*x_ref[0]+(-0.547395998804)*x_ref[1]**o+(0.257874232026)*x_ref[1]+(0.459659864323)*x_ref[2]**o+(-0.94127845744)*x_ref[2]
        ref[0,3,0,1]=(-0.309288966076)*x_ref[0]**o+(-0.173191962489)*x_ref[0]+(0.540628102572)*x_ref[1]**o+(-0.0443083494784)*x_ref[1]+(0.0435715587905)*x_ref[2]**o+(-0.301785508858)*x_ref[2]
        ref[0,3,1,0]=(0.145159950132)*x_ref[0]**o+(-0.287822257724)*x_ref[0]+(0.759130284956)*x_ref[1]**o+(0.0579397521075)*x_ref[1]+(0.699550687352)*x_ref[2]**o+(0.175478553574)*x_ref[2]
        ref[0,3,1,1]=(0.474727781805)*x_ref[0]**o+(-0.123878818145)*x_ref[0]+(-0.805368102867)*x_ref[1]**o+(-0.650609144709)*x_ref[1]+(0.656988695082)*x_ref[2]**o+(-0.0136028573608)*x_ref[2]
        ref[0,3,2,0]=(-0.319935169115)*x_ref[0]**o+(0.421488978156)*x_ref[0]+(0.496417273253)*x_ref[1]**o+(-0.129350388592)*x_ref[1]+(0.172506181508)*x_ref[2]**o+(0.762146639203)*x_ref[2]
        ref[0,3,2,1]=(0.269722271977)*x_ref[0]**o+(-0.754539522531)*x_ref[0]+(0.992009096198)*x_ref[1]**o+(0.705597359762)*x_ref[1]+(-0.334352203866)*x_ref[2]**o+(0.143158060849)*x_ref[2]
        ref[0,4,0,0]=(-0.431531468752)*x_ref[0]**o+(0.555318608886)*x_ref[0]+(0.986613799736)*x_ref[1]**o+(0.251035501881)*x_ref[1]+(-0.406355028528)*x_ref[2]**o+(-0.499620043752)*x_ref[2]
        ref[0,4,0,1]=(-0.0627649075987)*x_ref[0]**o+(-0.972278432629)*x_ref[0]+(-0.979219556729)*x_ref[1]**o+(0.224664798935)*x_ref[1]+(0.310699044854)*x_ref[2]**o+(0.127604283478)*x_ref[2]
        ref[0,4,1,0]=(-0.267874066531)*x_ref[0]**o+(0.991041579767)*x_ref[0]+(0.790132020473)*x_ref[1]**o+(0.228443092608)*x_ref[1]+(-0.890727088684)*x_ref[2]**o+(0.120869830054)*x_ref[2]
        ref[0,4,1,1]=(-0.268575786982)*x_ref[0]**o+(0.78564765473)*x_ref[0]+(0.208376098211)*x_ref[1]**o+(0.348864057531)*x_ref[1]+(0.627737586342)*x_ref[2]**o+(-0.540625434546)*x_ref[2]
        ref[0,4,2,0]=(-0.628026871983)*x_ref[0]**o+(-0.637731369616)*x_ref[0]+(0.891005717981)*x_ref[1]**o+(-0.281048323549)*x_ref[1]+(0.373980259828)*x_ref[2]**o+(0.676961901941)*x_ref[2]
        ref[0,4,2,1]=(0.12750050336)*x_ref[0]**o+(-0.481299457156)*x_ref[0]+(0.511901526031)*x_ref[1]**o+(0.913595029063)*x_ref[1]+(-0.263389445846)*x_ref[2]**o+(-0.167030063489)*x_ref[2]
        ref[1,0,0,0]=(0.785350719519)*x_ref[0]**o+(0.259099548564)*x_ref[0]+(0.267304446557)*x_ref[1]**o+(-0.864615565491)*x_ref[1]+(-0.0824408467375)*x_ref[2]**o+(0.724026653941)*x_ref[2]
        ref[1,0,0,1]=(0.741066935778)*x_ref[0]**o+(0.444348447224)*x_ref[0]+(0.950815368396)*x_ref[1]**o+(0.630967241982)*x_ref[1]+(-0.173602147175)*x_ref[2]**o+(-0.758408767101)*x_ref[2]
        ref[1,0,1,0]=(0.615189483755)*x_ref[0]**o+(0.113461303397)*x_ref[0]+(-0.204008801629)*x_ref[1]**o+(-0.821819803961)*x_ref[1]+(0.996060487559)*x_ref[2]**o+(0.501575420465)*x_ref[2]
        ref[1,0,1,1]=(-0.262723988452)*x_ref[0]**o+(0.54464307861)*x_ref[0]+(-0.776865346115)*x_ref[1]**o+(0.62546224345)*x_ref[1]+(-0.576679080633)*x_ref[2]**o+(-0.641871487717)*x_ref[2]
        ref[1,0,2,0]=(0.00882312632838)*x_ref[0]**o+(0.886325126754)*x_ref[0]+(-0.290744552898)*x_ref[1]**o+(-0.191666192698)*x_ref[1]+(-0.58836559482)*x_ref[2]**o+(-0.723263075029)*x_ref[2]
        ref[1,0,2,1]=(-0.8634554794)*x_ref[0]**o+(-0.156143013973)*x_ref[0]+(-0.954950131813)*x_ref[1]**o+(0.698956259432)*x_ref[1]+(-0.404041520631)*x_ref[2]**o+(0.00457698889215)*x_ref[2]
        ref[1,1,0,0]=(-0.728882693177)*x_ref[0]**o+(-0.900052268016)*x_ref[0]+(-0.852025073682)*x_ref[1]**o+(-0.0533567939337)*x_ref[1]+(-0.209580555028)*x_ref[2]**o+(-0.896328429332)*x_ref[2]
        ref[1,1,0,1]=(-0.629571937898)*x_ref[0]**o+(0.165066159679)*x_ref[0]+(-0.538661749525)*x_ref[1]**o+(-0.211485405646)*x_ref[1]+(-0.68390849131)*x_ref[2]**o+(0.395749832239)*x_ref[2]
        ref[1,1,1,0]=(0.26332743592)*x_ref[0]**o+(-0.0752845023801)*x_ref[0]+(-0.0561271762911)*x_ref[1]**o+(-0.211107669851)*x_ref[1]+(0.678154574805)*x_ref[2]**o+(-0.819108248966)*x_ref[2]
        ref[1,1,1,1]=(0.772808345449)*x_ref[0]**o+(-0.262453525307)*x_ref[0]+(-0.692191575144)*x_ref[1]**o+(0.779456283419)*x_ref[1]+(-0.331784968982)*x_ref[2]**o+(0.137130165506)*x_ref[2]
        ref[1,1,2,0]=(-0.126130110627)*x_ref[0]**o+(-0.450688959983)*x_ref[0]+(0.246073262195)*x_ref[1]**o+(0.452795873046)*x_ref[1]+(-0.144810908676)*x_ref[2]**o+(-0.100339618687)*x_ref[2]
        ref[1,1,2,1]=(-0.494674080271)*x_ref[0]**o+(-0.92562703505)*x_ref[0]+(-0.0365548469018)*x_ref[1]**o+(0.138971357509)*x_ref[1]+(-0.461030467649)*x_ref[2]**o+(0.560886019716)*x_ref[2]
        ref[1,2,0,0]=(0.103018022262)*x_ref[0]**o+(-0.95153827596)*x_ref[0]+(0.075101671678)*x_ref[1]**o+(0.385014541504)*x_ref[1]+(0.430777100819)*x_ref[2]**o+(-0.799382254202)*x_ref[2]
        ref[1,2,0,1]=(-0.69504642431)*x_ref[0]**o+(-0.989343939843)*x_ref[0]+(0.614616202255)*x_ref[1]**o+(0.119187311793)*x_ref[1]+(0.677533926016)*x_ref[2]**o+(-0.940961001327)*x_ref[2]
        ref[1,2,1,0]=(0.45997486768)*x_ref[0]**o+(0.131324255258)*x_ref[0]+(0.51531436597)*x_ref[1]**o+(0.570477569764)*x_ref[1]+(0.670091635832)*x_ref[2]**o+(0.991528689938)*x_ref[2]
        ref[1,2,1,1]=(0.693813612837)*x_ref[0]**o+(0.278464811874)*x_ref[0]+(-0.971105053619)*x_ref[1]**o+(-0.290401167365)*x_ref[1]+(0.563462235533)*x_ref[2]**o+(0.715283195872)*x_ref[2]
        ref[1,2,2,0]=(0.245799944524)*x_ref[0]**o+(0.72672691689)*x_ref[0]+(0.807071969617)*x_ref[1]**o+(0.825450000507)*x_ref[1]+(-0.859649268645)*x_ref[2]**o+(0.945957906767)*x_ref[2]
        ref[1,2,2,1]=(0.168765200995)*x_ref[0]**o+(-0.789419024479)*x_ref[0]+(-0.539642443181)*x_ref[1]**o+(0.463771463373)*x_ref[1]+(0.915893721757)*x_ref[2]**o+(-0.992301458598)*x_ref[2]
        ref[1,3,0,0]=(0.578524691791)*x_ref[0]**o+(-0.615099053887)*x_ref[0]+(-0.307646534709)*x_ref[1]**o+(-0.494406889327)*x_ref[1]+(-0.0189087238376)*x_ref[2]**o+(0.190080821114)*x_ref[2]
        ref[1,3,0,1]=(0.915034870134)*x_ref[0]**o+(0.657276161564)*x_ref[0]+(0.806274816666)*x_ref[1]**o+(-0.341020582334)*x_ref[1]+(0.729295103751)*x_ref[2]**o+(0.649468360593)*x_ref[2]
        ref[1,3,1,0]=(0.9673609292)*x_ref[0]**o+(-0.315653538456)*x_ref[0]+(-0.0346648713203)*x_ref[1]**o+(0.331870978455)*x_ref[1]+(-0.341269715094)*x_ref[2]**o+(-0.0249349656976)*x_ref[2]
        ref[1,3,1,1]=(-0.150933678366)*x_ref[0]**o+(-0.389726747893)*x_ref[0]+(-0.204342564941)*x_ref[1]**o+(-0.711033538116)*x_ref[1]+(-0.671896492614)*x_ref[2]**o+(-0.490644811516)*x_ref[2]
        ref[1,3,2,0]=(-0.488823609516)*x_ref[0]**o+(0.985637410007)*x_ref[0]+(0.728491418657)*x_ref[1]**o+(0.36207817836)*x_ref[1]+(0.0810824824344)*x_ref[2]**o+(-0.8152491822)*x_ref[2]
        ref[1,3,2,1]=(-0.517875264099)*x_ref[0]**o+(-0.727465681243)*x_ref[0]+(0.211140808921)*x_ref[1]**o+(-0.433707511139)*x_ref[1]+(0.344268020901)*x_ref[2]**o+(-0.847540940817)*x_ref[2]
        ref[1,4,0,0]=(-0.121268264679)*x_ref[0]**o+(-0.0117838122301)*x_ref[0]+(-0.63745424162)*x_ref[1]**o+(0.945563449658)*x_ref[1]+(-0.923254628525)*x_ref[2]**o+(0.838843523244)*x_ref[2]
        ref[1,4,0,1]=(-0.152712892877)*x_ref[0]**o+(0.115026184086)*x_ref[0]+(0.049691447791)*x_ref[1]**o+(-0.766235899011)*x_ref[1]+(-0.45149176462)*x_ref[2]**o+(0.0476296119238)*x_ref[2]
        ref[1,4,1,0]=(-0.172747241975)*x_ref[0]**o+(-0.751040839036)*x_ref[0]+(0.785947613925)*x_ref[1]**o+(0.369802888838)*x_ref[1]+(-0.732967210812)*x_ref[2]**o+(0.341006418057)*x_ref[2]
        ref[1,4,1,1]=(-0.280256623684)*x_ref[0]**o+(-0.817596493324)*x_ref[0]+(0.210362099145)*x_ref[1]**o+(-0.690356660671)*x_ref[1]+(0.564051447608)*x_ref[2]**o+(-0.81476511073)*x_ref[2]
        ref[1,4,2,0]=(-0.492106835788)*x_ref[0]**o+(0.1875588035)*x_ref[0]+(0.408972075649)*x_ref[1]**o+(-0.0979465471156)*x_ref[1]+(-0.191951030868)*x_ref[2]**o+(0.909540651582)*x_ref[2]
        ref[1,4,2,1]=(-0.419018988132)*x_ref[0]**o+(-0.0863722256375)*x_ref[0]+(0.621753648601)*x_ref[1]**o+(-0.0865542477263)*x_ref[1]+(0.44530474031)*x_ref[2]**o+(0.314386126939)*x_ref[2]
        ref[2,0,0,0]=(0.65858233579)*x_ref[0]**o+(-0.226735659791)*x_ref[0]+(-0.154886066807)*x_ref[1]**o+(0.456030801777)*x_ref[1]+(-0.930548820846)*x_ref[2]**o+(-0.97883928396)*x_ref[2]
        ref[2,0,0,1]=(-0.0589535588936)*x_ref[0]**o+(-0.267387198117)*x_ref[0]+(0.807737417612)*x_ref[1]**o+(0.670756731749)*x_ref[1]+(0.565928465377)*x_ref[2]**o+(0.338838873783)*x_ref[2]
        ref[2,0,1,0]=(0.745290598767)*x_ref[0]**o+(0.444308877108)*x_ref[0]+(-0.0653699257348)*x_ref[1]**o+(-0.446686429445)*x_ref[1]+(0.431580278588)*x_ref[2]**o+(-0.895205483299)*x_ref[2]
        ref[2,0,1,1]=(0.0515503491363)*x_ref[0]**o+(0.325788398831)*x_ref[0]+(-0.275580931208)*x_ref[1]**o+(-0.591278303275)*x_ref[1]+(-0.231550181927)*x_ref[2]**o+(0.830932642648)*x_ref[2]
        ref[2,0,2,0]=(-0.782241049983)*x_ref[0]**o+(0.420935438009)*x_ref[0]+(0.765777512717)*x_ref[1]**o+(0.139685303264)*x_ref[1]+(0.84210391036)*x_ref[2]**o+(-0.799020031388)*x_ref[2]
        ref[2,0,2,1]=(0.192115194447)*x_ref[0]**o+(0.821179393437)*x_ref[0]+(-0.29035052348)*x_ref[1]**o+(0.459217831059)*x_ref[1]+(0.24675222867)*x_ref[2]**o+(-0.0295595098143)*x_ref[2]
        ref[2,1,0,0]=(0.606333105811)*x_ref[0]**o+(-0.908981052972)*x_ref[0]+(0.194455118385)*x_ref[1]**o+(0.475347895918)*x_ref[1]+(0.154123701129)*x_ref[2]**o+(0.0656401634163)*x_ref[2]
        ref[2,1,0,1]=(0.924110494936)*x_ref[0]**o+(-0.155517993703)*x_ref[0]+(-0.105954511753)*x_ref[1]**o+(0.427136658346)*x_ref[1]+(-0.842319072137)*x_ref[2]**o+(0.893854788048)*x_ref[2]
        ref[2,1,1,0]=(0.775844998945)*x_ref[0]**o+(0.86288098975)*x_ref[0]+(-0.877212097351)*x_ref[1]**o+(-0.58811092774)*x_ref[1]+(-0.456603143488)*x_ref[2]**o+(0.781738385011)*x_ref[2]
        ref[2,1,1,1]=(0.745162233175)*x_ref[0]**o+(0.576206343886)*x_ref[0]+(0.897229190634)*x_ref[1]**o+(-0.748603588172)*x_ref[1]+(0.147340946009)*x_ref[2]**o+(-0.97456869073)*x_ref[2]
        ref[2,1,2,0]=(-0.995679867038)*x_ref[0]**o+(-0.512547860004)*x_ref[0]+(-0.125757163987)*x_ref[1]**o+(0.284748805033)*x_ref[1]+(0.748279176563)*x_ref[2]**o+(-0.370905992823)*x_ref[2]
        ref[2,1,2,1]=(-0.725652456641)*x_ref[0]**o+(0.303144081957)*x_ref[0]+(0.308580461688)*x_ref[1]**o+(-0.398783760014)*x_ref[1]+(0.256017075184)*x_ref[2]**o+(0.244843828536)*x_ref[2]
        ref[2,2,0,0]=(0.262768301493)*x_ref[0]**o+(0.994422492631)*x_ref[0]+(0.215209633103)*x_ref[1]**o+(-0.328208518526)*x_ref[1]+(-0.791000179758)*x_ref[2]**o+(-0.0931810905124)*x_ref[2]
        ref[2,2,0,1]=(0.430751983854)*x_ref[0]**o+(0.697487380231)*x_ref[0]+(-0.407533720752)*x_ref[1]**o+(0.542395064178)*x_ref[1]+(0.560493455122)*x_ref[2]**o+(0.862130615461)*x_ref[2]
        ref[2,2,1,0]=(0.0427580536739)*x_ref[0]**o+(0.755377456297)*x_ref[0]+(-0.882269449771)*x_ref[1]**o+(0.48654379074)*x_ref[1]+(0.0361187725004)*x_ref[2]**o+(0.1808373915)*x_ref[2]
        ref[2,2,1,1]=(-0.0473723357884)*x_ref[0]**o+(0.480334631394)*x_ref[0]+(-0.0913487472458)*x_ref[1]**o+(-0.900524837935)*x_ref[1]+(0.233994712545)*x_ref[2]**o+(-0.0576679105272)*x_ref[2]
        ref[2,2,2,0]=(-0.501358552905)*x_ref[0]**o+(-0.794253710665)*x_ref[0]+(-0.368314153003)*x_ref[1]**o+(-0.190680411227)*x_ref[1]+(0.867826651831)*x_ref[2]**o+(0.779026383345)*x_ref[2]
        ref[2,2,2,1]=(0.403937335227)*x_ref[0]**o+(0.23341509868)*x_ref[0]+(-0.867133942472)*x_ref[1]**o+(0.0236839523692)*x_ref[1]+(0.530643643468)*x_ref[2]**o+(0.153666452771)*x_ref[2]
        ref[2,3,0,0]=(-0.911977982923)*x_ref[0]**o+(0.0651402894633)*x_ref[0]+(-0.150699952611)*x_ref[1]**o+(-0.223159951427)*x_ref[1]+(0.0985310057795)*x_ref[2]**o+(-0.512754531906)*x_ref[2]
        ref[2,3,0,1]=(0.433292132106)*x_ref[0]**o+(-0.741489420094)*x_ref[0]+(-0.85661323724)*x_ref[1]**o+(-0.470049843131)*x_ref[1]+(-0.127893768764)*x_ref[2]**o+(0.445865536866)*x_ref[2]
        ref[2,3,1,0]=(0.176638607161)*x_ref[0]**o+(-0.804679828936)*x_ref[0]+(0.362072277714)*x_ref[1]**o+(-0.208356054989)*x_ref[1]+(0.270297511604)*x_ref[2]**o+(0.411240202674)*x_ref[2]
        ref[2,3,1,1]=(-0.731456178523)*x_ref[0]**o+(-0.494011319472)*x_ref[0]+(-0.656951979236)*x_ref[1]**o+(-0.442233264154)*x_ref[1]+(-0.506683809041)*x_ref[2]**o+(0.212265538187)*x_ref[2]
        ref[2,3,2,0]=(-0.551032190098)*x_ref[0]**o+(0.643467191611)*x_ref[0]+(0.543872371582)*x_ref[1]**o+(0.72584876243)*x_ref[1]+(0.985681498699)*x_ref[2]**o+(-0.60209773896)*x_ref[2]
        ref[2,3,2,1]=(0.0835903452585)*x_ref[0]**o+(0.440086089408)*x_ref[0]+(-0.578402114944)*x_ref[1]**o+(0.487833997481)*x_ref[1]+(0.25662705791)*x_ref[2]**o+(-0.473020070147)*x_ref[2]
        ref[2,4,0,0]=(-0.891315579854)*x_ref[0]**o+(0.952879579742)*x_ref[0]+(-0.227350081544)*x_ref[1]**o+(0.600033334324)*x_ref[1]+(0.179737000053)*x_ref[2]**o+(-0.504558083705)*x_ref[2]
        ref[2,4,0,1]=(0.225549297486)*x_ref[0]**o+(-0.543285628363)*x_ref[0]+(0.155686671765)*x_ref[1]**o+(-0.718395371077)*x_ref[1]+(0.674896476991)*x_ref[2]**o+(0.213411998034)*x_ref[2]
        ref[2,4,1,0]=(-0.35980223796)*x_ref[0]**o+(0.444715759834)*x_ref[0]+(-0.543703848794)*x_ref[1]**o+(0.183552103878)*x_ref[1]+(0.586337472615)*x_ref[2]**o+(0.213304038734)*x_ref[2]
        ref[2,4,1,1]=(-0.292792312207)*x_ref[0]**o+(0.196332398934)*x_ref[0]+(-0.798485726334)*x_ref[1]**o+(0.782220141806)*x_ref[1]+(0.0640127628843)*x_ref[2]**o+(-0.77727238011)*x_ref[2]
        ref[2,4,2,0]=(0.932887369251)*x_ref[0]**o+(0.0141383057976)*x_ref[0]+(-0.272693549414)*x_ref[1]**o+(-0.978856349954)*x_ref[1]+(0.89717148371)*x_ref[2]**o+(-0.424084272244)*x_ref[2]
        ref[2,4,2,1]=(-0.668326323236)*x_ref[0]**o+(-0.973037477253)*x_ref[0]+(-0.33722487276)*x_ref[1]**o+(-0.867851333935)*x_ref[1]+(0.216920541233)*x_ref[2]**o+(-0.657272351842)*x_ref[2]
        ref[3,0,0,0]=(-0.502029864543)*x_ref[0]**o+(-0.22816384379)*x_ref[0]+(0.656486587866)*x_ref[1]**o+(-0.364517454501)*x_ref[1]+(0.166636753172)*x_ref[2]**o+(-0.548031611216)*x_ref[2]
        ref[3,0,0,1]=(-0.129789963885)*x_ref[0]**o+(-0.924955467733)*x_ref[0]+(-0.318286603412)*x_ref[1]**o+(0.597531687741)*x_ref[1]+(0.664722349723)*x_ref[2]**o+(0.880847610571)*x_ref[2]
        ref[3,0,1,0]=(0.138386711079)*x_ref[0]**o+(-0.304181867173)*x_ref[0]+(0.189482296687)*x_ref[1]**o+(0.0576456570922)*x_ref[1]+(-0.026472114191)*x_ref[2]**o+(-0.509366463558)*x_ref[2]
        ref[3,0,1,1]=(-0.177846734744)*x_ref[0]**o+(0.504670109456)*x_ref[0]+(-0.802497403902)*x_ref[1]**o+(-0.920870115368)*x_ref[1]+(0.830468060312)*x_ref[2]**o+(0.76112636422)*x_ref[2]
        ref[3,0,2,0]=(0.101147363593)*x_ref[0]**o+(0.678670740322)*x_ref[0]+(0.929981085383)*x_ref[1]**o+(-0.78781173205)*x_ref[1]+(0.728075424496)*x_ref[2]**o+(0.597383557857)*x_ref[2]
        ref[3,0,2,1]=(-0.622642327441)*x_ref[0]**o+(0.994549062442)*x_ref[0]+(0.155580041238)*x_ref[1]**o+(-0.359913413167)*x_ref[1]+(-0.125600523598)*x_ref[2]**o+(0.727866512034)*x_ref[2]
        ref[3,1,0,0]=(-0.339382309297)*x_ref[0]**o+(0.269183464726)*x_ref[0]+(-0.82410881539)*x_ref[1]**o+(0.523860360233)*x_ref[1]+(0.672374739783)*x_ref[2]**o+(0.304612132802)*x_ref[2]
        ref[3,1,0,1]=(-0.398676828773)*x_ref[0]**o+(0.178258323741)*x_ref[0]+(-0.842177075615)*x_ref[1]**o+(0.928781440686)*x_ref[1]+(-0.698371966337)*x_ref[2]**o+(-0.069158873645)*x_ref[2]
        ref[3,1,1,0]=(0.517076773904)*x_ref[0]**o+(-0.722930310231)*x_ref[0]+(-0.948481363727)*x_ref[1]**o+(-0.313572917282)*x_ref[1]+(-0.216936906887)*x_ref[2]**o+(-0.617732400043)*x_ref[2]
        ref[3,1,1,1]=(-0.506541451394)*x_ref[0]**o+(-0.672313212249)*x_ref[0]+(0.884209908638)*x_ref[1]**o+(0.164981753359)*x_ref[1]+(0.38179585119)*x_ref[2]**o+(0.409267628038)*x_ref[2]
        ref[3,1,2,0]=(0.772486931191)*x_ref[0]**o+(0.550551917766)*x_ref[0]+(0.579543984786)*x_ref[1]**o+(0.457817667176)*x_ref[1]+(0.332038328462)*x_ref[2]**o+(-0.565647686923)*x_ref[2]
        ref[3,1,2,1]=(-0.934824056723)*x_ref[0]**o+(0.0324683981232)*x_ref[0]+(0.731737791013)*x_ref[1]**o+(0.450277257345)*x_ref[1]+(-0.0505877599703)*x_ref[2]**o+(-0.357925404256)*x_ref[2]
        ref[3,2,0,0]=(-0.846960246959)*x_ref[0]**o+(0.998323807325)*x_ref[0]+(-0.854597485859)*x_ref[1]**o+(0.752751841074)*x_ref[1]+(0.371090045318)*x_ref[2]**o+(-0.766611225212)*x_ref[2]
        ref[3,2,0,1]=(0.339046155557)*x_ref[0]**o+(-0.302854568693)*x_ref[0]+(-0.944840471752)*x_ref[1]**o+(0.566745543225)*x_ref[1]+(-0.829238416325)*x_ref[2]**o+(0.407197535798)*x_ref[2]
        ref[3,2,1,0]=(-0.527237680707)*x_ref[0]**o+(0.571797579499)*x_ref[0]+(0.803890750428)*x_ref[1]**o+(0.518904421497)*x_ref[1]+(0.157105476854)*x_ref[2]**o+(-0.430705894568)*x_ref[2]
        ref[3,2,1,1]=(0.0391938839127)*x_ref[0]**o+(-0.151396805547)*x_ref[0]+(0.22856560496)*x_ref[1]**o+(-0.207972615421)*x_ref[1]+(-0.0493211514337)*x_ref[2]**o+(-0.804654404843)*x_ref[2]
        ref[3,2,2,0]=(-0.653544001068)*x_ref[0]**o+(-0.414094661044)*x_ref[0]+(0.494230663809)*x_ref[1]**o+(-0.506949140354)*x_ref[1]+(0.520476398187)*x_ref[2]**o+(-0.969930682073)*x_ref[2]
        ref[3,2,2,1]=(-0.147585384358)*x_ref[0]**o+(0.775604709346)*x_ref[0]+(0.677576282676)*x_ref[1]**o+(-0.505391680821)*x_ref[1]+(0.704307147449)*x_ref[2]**o+(0.558458914373)*x_ref[2]
        ref[3,3,0,0]=(-0.65879046599)*x_ref[0]**o+(0.114409484281)*x_ref[0]+(0.622072786527)*x_ref[1]**o+(0.924427931612)*x_ref[1]+(-0.115550639369)*x_ref[2]**o+(-0.483381484729)*x_ref[2]
        ref[3,3,0,1]=(0.198285160028)*x_ref[0]**o+(-0.603311408844)*x_ref[0]+(-0.0471223647065)*x_ref[1]**o+(-0.438133307138)*x_ref[1]+(0.388251801711)*x_ref[2]**o+(-0.0732531194363)*x_ref[2]
        ref[3,3,1,0]=(0.150939105573)*x_ref[0]**o+(0.108946746296)*x_ref[0]+(-0.807347867525)*x_ref[1]**o+(-0.256128441651)*x_ref[1]+(0.231286821393)*x_ref[2]**o+(-0.703002240761)*x_ref[2]
        ref[3,3,1,1]=(-0.641529661237)*x_ref[0]**o+(-0.120398231904)*x_ref[0]+(0.351100779315)*x_ref[1]**o+(-0.576531857716)*x_ref[1]+(-0.765644574653)*x_ref[2]**o+(0.664743279322)*x_ref[2]
        ref[3,3,2,0]=(-0.47897462164)*x_ref[0]**o+(-0.569196375202)*x_ref[0]+(-0.422261130469)*x_ref[1]**o+(0.840196504879)*x_ref[1]+(-0.267300663782)*x_ref[2]**o+(-0.893226595716)*x_ref[2]
        ref[3,3,2,1]=(0.605260487329)*x_ref[0]**o+(-0.20572150917)*x_ref[0]+(-0.290112951218)*x_ref[1]**o+(0.887680899465)*x_ref[1]+(-0.240985548337)*x_ref[2]**o+(0.180645577595)*x_ref[2]
        ref[3,4,0,0]=(0.165393596846)*x_ref[0]**o+(0.134944494629)*x_ref[0]+(0.515339273402)*x_ref[1]**o+(-0.606138667469)*x_ref[1]+(0.0150579936235)*x_ref[2]**o+(0.261591191815)*x_ref[2]
        ref[3,4,0,1]=(0.541923776337)*x_ref[0]**o+(0.905517150748)*x_ref[0]+(-0.728025844328)*x_ref[1]**o+(-0.643009143952)*x_ref[1]+(-0.345013615196)*x_ref[2]**o+(0.457263127345)*x_ref[2]
        ref[3,4,1,0]=(-0.423845417015)*x_ref[0]**o+(0.882795189836)*x_ref[0]+(0.99729841901)*x_ref[1]**o+(0.106637707066)*x_ref[1]+(-0.960096403487)*x_ref[2]**o+(0.841258780132)*x_ref[2]
        ref[3,4,1,1]=(0.256176917841)*x_ref[0]**o+(0.120931832818)*x_ref[0]+(0.908290189359)*x_ref[1]**o+(-0.870707945117)*x_ref[1]+(0.726505555181)*x_ref[2]**o+(-0.110841837509)*x_ref[2]
        ref[3,4,2,0]=(-0.0209556990671)*x_ref[0]**o+(-0.0986852276803)*x_ref[0]+(0.758132562139)*x_ref[1]**o+(0.0827733457302)*x_ref[1]+(-0.689039269813)*x_ref[2]**o+(-0.120148663779)*x_ref[2]
        ref[3,4,2,1]=(-0.189903364845)*x_ref[0]**o+(-0.145753796459)*x_ref[0]+(0.608678649752)*x_ref[1]**o+(-0.51842417051)*x_ref[1]+(-0.212194895669)*x_ref[2]**o+(0.321128974995)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.517080828008)*x[0]**o+(0.816940025452)*x[0]+(-0.863917100487)*x[1]**o+(-0.733059827221)*x[1]
        ref=(-0.517080828008)*x_ref[0]**o+(0.816940025452)*x_ref[0]+(-0.863917100487)*x_ref[1]**o+(-0.733059827221)*x_ref[1]
      else:
        s=(-0.499547235343)*x[0]**o+(-0.374394256068)*x[0]+(0.487571133184)*x[1]**o+(0.0897352557867)*x[1]+(0.549399105787)*x[2]**o+(0.866160943214)*x[2]
        ref=(-0.499547235343)*x_ref[0]**o+(-0.374394256068)*x_ref[0]+(0.487571133184)*x_ref[1]**o+(0.0897352557867)*x_ref[1]+(0.549399105787)*x_ref[2]**o+(0.866160943214)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(0.118375355539)*x[0]**o+(0.606871963148)*x[0]+(-0.792889140772)*x[1]**o+(-0.132004959608)*x[1]
        s[1]=(-0.96232807347)*x[0]**o+(-0.246198113355)*x[0]+(0.979860762173)*x[1]**o+(0.0895137747321)*x[1]
        ref[0]=(0.118375355539)*x_ref[0]**o+(0.606871963148)*x_ref[0]+(-0.792889140772)*x_ref[1]**o+(-0.132004959608)*x_ref[1]
        ref[1]=(-0.96232807347)*x_ref[0]**o+(-0.246198113355)*x_ref[0]+(0.979860762173)*x_ref[1]**o+(0.0895137747321)*x_ref[1]
      else:
        s[0]=(0.871399819091)*x[0]**o+(0.953883292108)*x[0]+(0.132344511435)*x[1]**o+(-0.688489026098)*x[1]+(-0.975794578087)*x[2]**o+(0.220648789861)*x[2]
        s[1]=(0.524543460725)*x[0]**o+(0.821761406148)*x[0]+(0.176747887636)*x[1]**o+(0.301952231259)*x[1]+(-0.0934120381526)*x[2]**o+(0.49987105682)*x[2]
        ref[0]=(0.871399819091)*x_ref[0]**o+(0.953883292108)*x_ref[0]+(0.132344511435)*x_ref[1]**o+(-0.688489026098)*x_ref[1]+(-0.975794578087)*x_ref[2]**o+(0.220648789861)*x_ref[2]
        ref[1]=(0.524543460725)*x_ref[0]**o+(0.821761406148)*x_ref[0]+(0.176747887636)*x_ref[1]**o+(0.301952231259)*x_ref[1]+(-0.0934120381526)*x_ref[2]**o+(0.49987105682)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.024902457613)*x[0]**o+(0.203976235166)*x[0]+(0.628459615768)*x[1]**o+(0.477560188705)*x[1]
        s[0,1]=(0.879517679908)*x[0]**o+(0.256126580826)*x[0]+(-0.0765272150363)*x[1]**o+(0.156282237684)*x[1]
        s[0,2]=(0.365353751433)*x[0]**o+(-0.91109682061)*x[0]+(0.14737314357)*x[1]**o+(0.0548681994126)*x[1]
        s[0,3]=(0.864175430484)*x[0]**o+(0.69042358011)*x[0]+(0.557446526854)*x[1]**o+(-0.599892539367)*x[1]
        s[0,4]=(0.101599454608)*x[0]**o+(0.095267955818)*x[0]+(0.852444074131)*x[1]**o+(-0.191743684531)*x[1]
        s[1,0]=(0.864588717269)*x[0]**o+(0.785378570351)*x[0]+(-0.0610940916664)*x[1]**o+(-0.303076611091)*x[1]
        s[1,1]=(0.256297437402)*x[0]**o+(0.348120390412)*x[0]+(0.751284911835)*x[1]**o+(-0.729588029473)*x[1]
        s[1,2]=(-0.440719279416)*x[0]**o+(-0.927912511637)*x[0]+(-0.911660900179)*x[1]**o+(0.905919826079)*x[1]
        s[1,3]=(0.937342196512)*x[0]**o+(-0.37712882469)*x[0]+(0.13938594539)*x[1]**o+(-0.90915248708)*x[1]
        s[1,4]=(-0.470517530843)*x[0]**o+(0.923326317433)*x[0]+(-0.215615694693)*x[1]**o+(-0.00831704931824)*x[1]
        s[2,0]=(-0.0923157720676)*x[0]**o+(-0.621794935502)*x[0]+(-0.189179129378)*x[1]**o+(0.308295797468)*x[1]
        s[2,1]=(0.212834713743)*x[0]**o+(0.88507127286)*x[0]+(-0.930573386294)*x[1]**o+(0.382952758778)*x[1]
        s[2,2]=(0.795151348124)*x[0]**o+(-0.153979548679)*x[0]+(-0.478754272467)*x[1]**o+(-0.0342752176147)*x[1]
        s[2,3]=(-0.25462026179)*x[0]**o+(-0.597291379533)*x[0]+(-0.229990916585)*x[1]**o+(0.841384036796)*x[1]
        s[2,4]=(0.766213848178)*x[0]**o+(-0.228607539271)*x[0]+(-0.17581810411)*x[1]**o+(0.688293027334)*x[1]
        s[3,0]=(0.308623266249)*x[0]**o+(0.076711970906)*x[0]+(0.714429570549)*x[1]**o+(-0.631005244988)*x[1]
        s[3,1]=(-0.878500450629)*x[0]**o+(0.339475789334)*x[0]+(-0.328537299485)*x[1]**o+(-0.170179906482)*x[1]
        s[3,2]=(-0.508481058155)*x[0]**o+(-0.98805110722)*x[0]+(0.469532054898)*x[1]**o+(0.813014428064)*x[1]
        s[3,3]=(-0.605821608801)*x[0]**o+(-0.571897502616)*x[0]+(-0.155954537498)*x[1]**o+(0.894775861604)*x[1]
        s[3,4]=(0.344877430061)*x[0]**o+(-0.460182472259)*x[0]+(0.996885806248)*x[1]**o+(0.888994222115)*x[1]
        ref[0,0]=(0.024902457613)*x_ref[0]**o+(0.203976235166)*x_ref[0]+(0.628459615768)*x_ref[1]**o+(0.477560188705)*x_ref[1]
        ref[0,1]=(0.879517679908)*x_ref[0]**o+(0.256126580826)*x_ref[0]+(-0.0765272150363)*x_ref[1]**o+(0.156282237684)*x_ref[1]
        ref[0,2]=(0.365353751433)*x_ref[0]**o+(-0.91109682061)*x_ref[0]+(0.14737314357)*x_ref[1]**o+(0.0548681994126)*x_ref[1]
        ref[0,3]=(0.864175430484)*x_ref[0]**o+(0.69042358011)*x_ref[0]+(0.557446526854)*x_ref[1]**o+(-0.599892539367)*x_ref[1]
        ref[0,4]=(0.101599454608)*x_ref[0]**o+(0.095267955818)*x_ref[0]+(0.852444074131)*x_ref[1]**o+(-0.191743684531)*x_ref[1]
        ref[1,0]=(0.864588717269)*x_ref[0]**o+(0.785378570351)*x_ref[0]+(-0.0610940916664)*x_ref[1]**o+(-0.303076611091)*x_ref[1]
        ref[1,1]=(0.256297437402)*x_ref[0]**o+(0.348120390412)*x_ref[0]+(0.751284911835)*x_ref[1]**o+(-0.729588029473)*x_ref[1]
        ref[1,2]=(-0.440719279416)*x_ref[0]**o+(-0.927912511637)*x_ref[0]+(-0.911660900179)*x_ref[1]**o+(0.905919826079)*x_ref[1]
        ref[1,3]=(0.937342196512)*x_ref[0]**o+(-0.37712882469)*x_ref[0]+(0.13938594539)*x_ref[1]**o+(-0.90915248708)*x_ref[1]
        ref[1,4]=(-0.470517530843)*x_ref[0]**o+(0.923326317433)*x_ref[0]+(-0.215615694693)*x_ref[1]**o+(-0.00831704931824)*x_ref[1]
        ref[2,0]=(-0.0923157720676)*x_ref[0]**o+(-0.621794935502)*x_ref[0]+(-0.189179129378)*x_ref[1]**o+(0.308295797468)*x_ref[1]
        ref[2,1]=(0.212834713743)*x_ref[0]**o+(0.88507127286)*x_ref[0]+(-0.930573386294)*x_ref[1]**o+(0.382952758778)*x_ref[1]
        ref[2,2]=(0.795151348124)*x_ref[0]**o+(-0.153979548679)*x_ref[0]+(-0.478754272467)*x_ref[1]**o+(-0.0342752176147)*x_ref[1]
        ref[2,3]=(-0.25462026179)*x_ref[0]**o+(-0.597291379533)*x_ref[0]+(-0.229990916585)*x_ref[1]**o+(0.841384036796)*x_ref[1]
        ref[2,4]=(0.766213848178)*x_ref[0]**o+(-0.228607539271)*x_ref[0]+(-0.17581810411)*x_ref[1]**o+(0.688293027334)*x_ref[1]
        ref[3,0]=(0.308623266249)*x_ref[0]**o+(0.076711970906)*x_ref[0]+(0.714429570549)*x_ref[1]**o+(-0.631005244988)*x_ref[1]
        ref[3,1]=(-0.878500450629)*x_ref[0]**o+(0.339475789334)*x_ref[0]+(-0.328537299485)*x_ref[1]**o+(-0.170179906482)*x_ref[1]
        ref[3,2]=(-0.508481058155)*x_ref[0]**o+(-0.98805110722)*x_ref[0]+(0.469532054898)*x_ref[1]**o+(0.813014428064)*x_ref[1]
        ref[3,3]=(-0.605821608801)*x_ref[0]**o+(-0.571897502616)*x_ref[0]+(-0.155954537498)*x_ref[1]**o+(0.894775861604)*x_ref[1]
        ref[3,4]=(0.344877430061)*x_ref[0]**o+(-0.460182472259)*x_ref[0]+(0.996885806248)*x_ref[1]**o+(0.888994222115)*x_ref[1]
      else:
        s[0,0]=(0.735330979003)*x[0]**o+(-0.187442795075)*x[0]+(0.657167234776)*x[1]**o+(0.999186075661)*x[1]+(0.0575488282349)*x[2]**o+(0.129253824991)*x[2]
        s[0,1]=(-0.157944993021)*x[0]**o+(0.960809659389)*x[0]+(-0.525836839883)*x[1]**o+(0.307428937928)*x[1]+(0.974287689142)*x[2]**o+(-0.0168563458634)*x[2]
        s[0,2]=(-0.318430570819)*x[0]**o+(0.000898489675827)*x[0]+(0.522994887101)*x[1]**o+(0.437401922013)*x[1]+(0.706640354472)*x[2]**o+(0.935332272606)*x[2]
        s[0,3]=(0.987729193431)*x[0]**o+(-0.390976298631)*x[0]+(0.562564630336)*x[1]**o+(-0.112154175922)*x[1]+(0.957175624895)*x[2]**o+(-0.801618518576)*x[2]
        s[0,4]=(-0.483652813287)*x[0]**o+(-0.0380163380099)*x[0]+(-0.99759861693)*x[1]**o+(-0.775699732856)*x[1]+(0.23000582537)*x[2]**o+(0.802965034057)*x[2]
        s[1,0]=(-0.609648414613)*x[0]**o+(-0.260798859841)*x[0]+(0.122422869599)*x[1]**o+(0.977634041)*x[1]+(0.0753219966441)*x[2]**o+(-0.474039520094)*x[2]
        s[1,1]=(-0.170909164803)*x[0]**o+(-0.734651964907)*x[0]+(-0.265011031724)*x[1]**o+(-0.0744686063401)*x[1]+(0.694499004733)*x[2]**o+(0.447100902964)*x[2]
        s[1,2]=(0.868461811889)*x[0]**o+(-0.426276230695)*x[0]+(0.707166478071)*x[1]**o+(0.585046443775)*x[1]+(0.960329103202)*x[2]**o+(0.56523072631)*x[2]
        s[1,3]=(0.648879573312)*x[0]**o+(0.086549837765)*x[0]+(0.612958547216)*x[1]**o+(0.320922603122)*x[1]+(0.152475907413)*x[2]**o+(0.879361543019)*x[2]
        s[1,4]=(0.763029097021)*x[0]**o+(0.0102114956179)*x[0]+(0.819317226634)*x[1]**o+(0.563140883434)*x[1]+(0.388977386477)*x[2]**o+(0.195692043167)*x[2]
        s[2,0]=(0.0890570134033)*x[0]**o+(-0.858058515043)*x[0]+(0.992023551061)*x[1]**o+(0.511161321044)*x[1]+(0.608278635931)*x[2]**o+(-0.330006205612)*x[2]
        s[2,1]=(-0.56891254443)*x[0]**o+(-0.921450822386)*x[0]+(0.706442237836)*x[1]**o+(-0.539764468602)*x[1]+(-0.868826400334)*x[2]**o+(-0.589805045051)*x[2]
        s[2,2]=(-0.172056002838)*x[0]**o+(0.143735290505)*x[0]+(0.0829083028752)*x[1]**o+(-0.949316661365)*x[1]+(0.692853163325)*x[2]**o+(0.89963936808)*x[2]
        s[2,3]=(0.866953231148)*x[0]**o+(0.492820447945)*x[0]+(0.727790485253)*x[1]**o+(0.861063706133)*x[1]+(0.272069465745)*x[2]**o+(-0.787281640867)*x[2]
        s[2,4]=(0.336142920614)*x[0]**o+(0.0174574477667)*x[0]+(0.571740441965)*x[1]**o+(0.322772849346)*x[1]+(-0.392942231888)*x[2]**o+(-0.231930509062)*x[2]
        s[3,0]=(-0.0143597894395)*x[0]**o+(0.948757696673)*x[0]+(-0.21685145343)*x[1]**o+(-0.507179225542)*x[1]+(0.796629577476)*x[2]**o+(-0.727952562306)*x[2]
        s[3,1]=(-0.305551777318)*x[0]**o+(-0.497961303192)*x[0]+(-0.649685028436)*x[1]**o+(0.614209312174)*x[1]+(0.367109274906)*x[2]**o+(-0.658788490538)*x[2]
        s[3,2]=(0.356879630024)*x[0]**o+(0.840561547569)*x[0]+(0.621063943906)*x[1]**o+(-0.579283601074)*x[1]+(0.253307235203)*x[2]**o+(0.411968412753)*x[2]
        s[3,3]=(-0.854175748046)*x[0]**o+(-0.126751572242)*x[0]+(0.657808185846)*x[1]**o+(-0.885332612288)*x[1]+(0.696176279396)*x[2]**o+(-0.120045386984)*x[2]
        s[3,4]=(-0.509064446808)*x[0]**o+(0.625879823023)*x[0]+(0.318941374341)*x[1]**o+(-0.554947143366)*x[1]+(0.582980700295)*x[2]**o+(-0.385567234453)*x[2]
        ref[0,0]=(0.735330979003)*x_ref[0]**o+(-0.187442795075)*x_ref[0]+(0.657167234776)*x_ref[1]**o+(0.999186075661)*x_ref[1]+(0.0575488282349)*x_ref[2]**o+(0.129253824991)*x_ref[2]
        ref[0,1]=(-0.157944993021)*x_ref[0]**o+(0.960809659389)*x_ref[0]+(-0.525836839883)*x_ref[1]**o+(0.307428937928)*x_ref[1]+(0.974287689142)*x_ref[2]**o+(-0.0168563458634)*x_ref[2]
        ref[0,2]=(-0.318430570819)*x_ref[0]**o+(0.000898489675827)*x_ref[0]+(0.522994887101)*x_ref[1]**o+(0.437401922013)*x_ref[1]+(0.706640354472)*x_ref[2]**o+(0.935332272606)*x_ref[2]
        ref[0,3]=(0.987729193431)*x_ref[0]**o+(-0.390976298631)*x_ref[0]+(0.562564630336)*x_ref[1]**o+(-0.112154175922)*x_ref[1]+(0.957175624895)*x_ref[2]**o+(-0.801618518576)*x_ref[2]
        ref[0,4]=(-0.483652813287)*x_ref[0]**o+(-0.0380163380099)*x_ref[0]+(-0.99759861693)*x_ref[1]**o+(-0.775699732856)*x_ref[1]+(0.23000582537)*x_ref[2]**o+(0.802965034057)*x_ref[2]
        ref[1,0]=(-0.609648414613)*x_ref[0]**o+(-0.260798859841)*x_ref[0]+(0.122422869599)*x_ref[1]**o+(0.977634041)*x_ref[1]+(0.0753219966441)*x_ref[2]**o+(-0.474039520094)*x_ref[2]
        ref[1,1]=(-0.170909164803)*x_ref[0]**o+(-0.734651964907)*x_ref[0]+(-0.265011031724)*x_ref[1]**o+(-0.0744686063401)*x_ref[1]+(0.694499004733)*x_ref[2]**o+(0.447100902964)*x_ref[2]
        ref[1,2]=(0.868461811889)*x_ref[0]**o+(-0.426276230695)*x_ref[0]+(0.707166478071)*x_ref[1]**o+(0.585046443775)*x_ref[1]+(0.960329103202)*x_ref[2]**o+(0.56523072631)*x_ref[2]
        ref[1,3]=(0.648879573312)*x_ref[0]**o+(0.086549837765)*x_ref[0]+(0.612958547216)*x_ref[1]**o+(0.320922603122)*x_ref[1]+(0.152475907413)*x_ref[2]**o+(0.879361543019)*x_ref[2]
        ref[1,4]=(0.763029097021)*x_ref[0]**o+(0.0102114956179)*x_ref[0]+(0.819317226634)*x_ref[1]**o+(0.563140883434)*x_ref[1]+(0.388977386477)*x_ref[2]**o+(0.195692043167)*x_ref[2]
        ref[2,0]=(0.0890570134033)*x_ref[0]**o+(-0.858058515043)*x_ref[0]+(0.992023551061)*x_ref[1]**o+(0.511161321044)*x_ref[1]+(0.608278635931)*x_ref[2]**o+(-0.330006205612)*x_ref[2]
        ref[2,1]=(-0.56891254443)*x_ref[0]**o+(-0.921450822386)*x_ref[0]+(0.706442237836)*x_ref[1]**o+(-0.539764468602)*x_ref[1]+(-0.868826400334)*x_ref[2]**o+(-0.589805045051)*x_ref[2]
        ref[2,2]=(-0.172056002838)*x_ref[0]**o+(0.143735290505)*x_ref[0]+(0.0829083028752)*x_ref[1]**o+(-0.949316661365)*x_ref[1]+(0.692853163325)*x_ref[2]**o+(0.89963936808)*x_ref[2]
        ref[2,3]=(0.866953231148)*x_ref[0]**o+(0.492820447945)*x_ref[0]+(0.727790485253)*x_ref[1]**o+(0.861063706133)*x_ref[1]+(0.272069465745)*x_ref[2]**o+(-0.787281640867)*x_ref[2]
        ref[2,4]=(0.336142920614)*x_ref[0]**o+(0.0174574477667)*x_ref[0]+(0.571740441965)*x_ref[1]**o+(0.322772849346)*x_ref[1]+(-0.392942231888)*x_ref[2]**o+(-0.231930509062)*x_ref[2]
        ref[3,0]=(-0.0143597894395)*x_ref[0]**o+(0.948757696673)*x_ref[0]+(-0.21685145343)*x_ref[1]**o+(-0.507179225542)*x_ref[1]+(0.796629577476)*x_ref[2]**o+(-0.727952562306)*x_ref[2]
        ref[3,1]=(-0.305551777318)*x_ref[0]**o+(-0.497961303192)*x_ref[0]+(-0.649685028436)*x_ref[1]**o+(0.614209312174)*x_ref[1]+(0.367109274906)*x_ref[2]**o+(-0.658788490538)*x_ref[2]
        ref[3,2]=(0.356879630024)*x_ref[0]**o+(0.840561547569)*x_ref[0]+(0.621063943906)*x_ref[1]**o+(-0.579283601074)*x_ref[1]+(0.253307235203)*x_ref[2]**o+(0.411968412753)*x_ref[2]
        ref[3,3]=(-0.854175748046)*x_ref[0]**o+(-0.126751572242)*x_ref[0]+(0.657808185846)*x_ref[1]**o+(-0.885332612288)*x_ref[1]+(0.696176279396)*x_ref[2]**o+(-0.120045386984)*x_ref[2]
        ref[3,4]=(-0.509064446808)*x_ref[0]**o+(0.625879823023)*x_ref[0]+(0.318941374341)*x_ref[1]**o+(-0.554947143366)*x_ref[1]+(0.582980700295)*x_ref[2]**o+(-0.385567234453)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.815314462093)*x[0]**o+(0.873826225948)*x[0]+(-0.554612541206)*x[1]**o+(-0.226621541398)*x[1]
        s[0,0,1]=(0.0128681913107)*x[0]**o+(-0.970763827703)*x[0]+(0.561143556525)*x[1]**o+(-0.489929062303)*x[1]
        s[0,1,0]=(0.810443300177)*x[0]**o+(0.815179482793)*x[0]+(-0.533320816715)*x[1]**o+(-0.263660851191)*x[1]
        s[0,1,1]=(-0.700273022268)*x[0]**o+(0.7542047746)*x[0]+(-0.608062940355)*x[1]**o+(0.165446265888)*x[1]
        s[1,0,0]=(-0.371305351324)*x[0]**o+(-0.658300667938)*x[0]+(-0.279035065475)*x[1]**o+(-0.426302834141)*x[1]
        s[1,0,1]=(0.479540028576)*x[0]**o+(0.548137277675)*x[0]+(-0.634709103526)*x[1]**o+(0.805406305543)*x[1]
        s[1,1,0]=(0.878975308128)*x[0]**o+(0.866886677728)*x[0]+(-0.0860689734435)*x[1]**o+(0.459268475449)*x[1]
        s[1,1,1]=(0.257574704806)*x[0]**o+(0.23677011273)*x[0]+(-0.167616745684)*x[1]**o+(0.335064186727)*x[1]
        s[2,0,0]=(-0.496561086088)*x[0]**o+(0.183444541815)*x[0]+(0.938753552035)*x[1]**o+(-0.65016148117)*x[1]
        s[2,0,1]=(-0.450876555427)*x[0]**o+(-0.826542061721)*x[0]+(0.991779661336)*x[1]**o+(-0.270460893052)*x[1]
        s[2,1,0]=(0.720651539816)*x[0]**o+(0.0333740371474)*x[0]+(0.106378152652)*x[1]**o+(-0.963538633907)*x[1]
        s[2,1,1]=(0.0463090835205)*x[0]**o+(-0.580911121364)*x[0]+(-0.950760000822)*x[1]**o+(0.345076271482)*x[1]
        s[3,0,0]=(-0.490958747939)*x[0]**o+(-0.473219210997)*x[0]+(0.394347301955)*x[1]**o+(0.0904571384668)*x[1]
        s[3,0,1]=(-0.719645713153)*x[0]**o+(0.579721027819)*x[0]+(0.341754270769)*x[1]**o+(0.381462300425)*x[1]
        s[3,1,0]=(0.987038883509)*x[0]**o+(-0.0168274174136)*x[0]+(-0.131309177111)*x[1]**o+(0.852247989409)*x[1]
        s[3,1,1]=(0.409343859078)*x[0]**o+(-0.955989546105)*x[0]+(-0.0278518652665)*x[1]**o+(0.230387377652)*x[1]
        s[4,0,0]=(-0.656714290973)*x[0]**o+(0.325615526697)*x[0]+(0.283818524063)*x[1]**o+(-0.590774747741)*x[1]
        s[4,0,1]=(-0.69952841223)*x[0]**o+(-0.114368546727)*x[0]+(0.374117478005)*x[1]**o+(0.328081831358)*x[1]
        s[4,1,0]=(0.380984662834)*x[0]**o+(0.0508406245408)*x[0]+(0.820208081773)*x[1]**o+(0.704399837049)*x[1]
        s[4,1,1]=(0.519149340125)*x[0]**o+(0.284927220267)*x[0]+(-0.764126727298)*x[1]**o+(-0.560612778285)*x[1]
        s[5,0,0]=(-0.710078531728)*x[0]**o+(0.0972655660835)*x[0]+(-0.228336763317)*x[1]**o+(0.723480791499)*x[1]
        s[5,0,1]=(-0.119027330003)*x[0]**o+(0.597105680171)*x[0]+(-0.319507310449)*x[1]**o+(0.978114408819)*x[1]
        s[5,1,0]=(0.576593282976)*x[0]**o+(0.796125984002)*x[0]+(-0.998293249002)*x[1]**o+(-0.713436214295)*x[1]
        s[5,1,1]=(-0.291197494362)*x[0]**o+(0.319138611544)*x[0]+(-0.677230907828)*x[1]**o+(0.549425117355)*x[1]
        ref[0,0,0]=(0.815314462093)*x_ref[0]**o+(0.873826225948)*x_ref[0]+(-0.554612541206)*x_ref[1]**o+(-0.226621541398)*x_ref[1]
        ref[0,0,1]=(0.0128681913107)*x_ref[0]**o+(-0.970763827703)*x_ref[0]+(0.561143556525)*x_ref[1]**o+(-0.489929062303)*x_ref[1]
        ref[0,1,0]=(0.810443300177)*x_ref[0]**o+(0.815179482793)*x_ref[0]+(-0.533320816715)*x_ref[1]**o+(-0.263660851191)*x_ref[1]
        ref[0,1,1]=(-0.700273022268)*x_ref[0]**o+(0.7542047746)*x_ref[0]+(-0.608062940355)*x_ref[1]**o+(0.165446265888)*x_ref[1]
        ref[1,0,0]=(-0.371305351324)*x_ref[0]**o+(-0.658300667938)*x_ref[0]+(-0.279035065475)*x_ref[1]**o+(-0.426302834141)*x_ref[1]
        ref[1,0,1]=(0.479540028576)*x_ref[0]**o+(0.548137277675)*x_ref[0]+(-0.634709103526)*x_ref[1]**o+(0.805406305543)*x_ref[1]
        ref[1,1,0]=(0.878975308128)*x_ref[0]**o+(0.866886677728)*x_ref[0]+(-0.0860689734435)*x_ref[1]**o+(0.459268475449)*x_ref[1]
        ref[1,1,1]=(0.257574704806)*x_ref[0]**o+(0.23677011273)*x_ref[0]+(-0.167616745684)*x_ref[1]**o+(0.335064186727)*x_ref[1]
        ref[2,0,0]=(-0.496561086088)*x_ref[0]**o+(0.183444541815)*x_ref[0]+(0.938753552035)*x_ref[1]**o+(-0.65016148117)*x_ref[1]
        ref[2,0,1]=(-0.450876555427)*x_ref[0]**o+(-0.826542061721)*x_ref[0]+(0.991779661336)*x_ref[1]**o+(-0.270460893052)*x_ref[1]
        ref[2,1,0]=(0.720651539816)*x_ref[0]**o+(0.0333740371474)*x_ref[0]+(0.106378152652)*x_ref[1]**o+(-0.963538633907)*x_ref[1]
        ref[2,1,1]=(0.0463090835205)*x_ref[0]**o+(-0.580911121364)*x_ref[0]+(-0.950760000822)*x_ref[1]**o+(0.345076271482)*x_ref[1]
        ref[3,0,0]=(-0.490958747939)*x_ref[0]**o+(-0.473219210997)*x_ref[0]+(0.394347301955)*x_ref[1]**o+(0.0904571384668)*x_ref[1]
        ref[3,0,1]=(-0.719645713153)*x_ref[0]**o+(0.579721027819)*x_ref[0]+(0.341754270769)*x_ref[1]**o+(0.381462300425)*x_ref[1]
        ref[3,1,0]=(0.987038883509)*x_ref[0]**o+(-0.0168274174136)*x_ref[0]+(-0.131309177111)*x_ref[1]**o+(0.852247989409)*x_ref[1]
        ref[3,1,1]=(0.409343859078)*x_ref[0]**o+(-0.955989546105)*x_ref[0]+(-0.0278518652665)*x_ref[1]**o+(0.230387377652)*x_ref[1]
        ref[4,0,0]=(-0.656714290973)*x_ref[0]**o+(0.325615526697)*x_ref[0]+(0.283818524063)*x_ref[1]**o+(-0.590774747741)*x_ref[1]
        ref[4,0,1]=(-0.69952841223)*x_ref[0]**o+(-0.114368546727)*x_ref[0]+(0.374117478005)*x_ref[1]**o+(0.328081831358)*x_ref[1]
        ref[4,1,0]=(0.380984662834)*x_ref[0]**o+(0.0508406245408)*x_ref[0]+(0.820208081773)*x_ref[1]**o+(0.704399837049)*x_ref[1]
        ref[4,1,1]=(0.519149340125)*x_ref[0]**o+(0.284927220267)*x_ref[0]+(-0.764126727298)*x_ref[1]**o+(-0.560612778285)*x_ref[1]
        ref[5,0,0]=(-0.710078531728)*x_ref[0]**o+(0.0972655660835)*x_ref[0]+(-0.228336763317)*x_ref[1]**o+(0.723480791499)*x_ref[1]
        ref[5,0,1]=(-0.119027330003)*x_ref[0]**o+(0.597105680171)*x_ref[0]+(-0.319507310449)*x_ref[1]**o+(0.978114408819)*x_ref[1]
        ref[5,1,0]=(0.576593282976)*x_ref[0]**o+(0.796125984002)*x_ref[0]+(-0.998293249002)*x_ref[1]**o+(-0.713436214295)*x_ref[1]
        ref[5,1,1]=(-0.291197494362)*x_ref[0]**o+(0.319138611544)*x_ref[0]+(-0.677230907828)*x_ref[1]**o+(0.549425117355)*x_ref[1]
      else:
        s[0,0,0]=(-0.789735577827)*x[0]**o+(0.713461925327)*x[0]+(-0.446338384344)*x[1]**o+(0.383247040547)*x[1]+(0.145159260009)*x[2]**o+(0.329895283826)*x[2]
        s[0,0,1]=(-0.583456019572)*x[0]**o+(0.409798276223)*x[0]+(0.443677194056)*x[1]**o+(0.686196976063)*x[1]+(0.482541428129)*x[2]**o+(-0.025427695773)*x[2]
        s[0,1,0]=(0.982231667045)*x[0]**o+(-0.8272954182)*x[0]+(0.989187050601)*x[1]**o+(0.2851725407)*x[1]+(-0.188195134162)*x[2]**o+(-0.0748834526694)*x[2]
        s[0,1,1]=(-0.211241906114)*x[0]**o+(-0.213099374182)*x[0]+(-0.924671222931)*x[1]**o+(-0.498297943877)*x[1]+(-0.204668529404)*x[2]**o+(-0.14158975248)*x[2]
        s[1,0,0]=(0.76821312782)*x[0]**o+(0.0971791123021)*x[0]+(-0.875792256474)*x[1]**o+(0.105989140339)*x[1]+(0.00172615516766)*x[2]**o+(0.733396501357)*x[2]
        s[1,0,1]=(-0.586069681251)*x[0]**o+(0.682163598775)*x[0]+(-0.254310626636)*x[1]**o+(-0.697248981934)*x[1]+(-0.914707288038)*x[2]**o+(0.982945477807)*x[2]
        s[1,1,0]=(0.689420994393)*x[0]**o+(0.362698077408)*x[0]+(0.227899444247)*x[1]**o+(0.842841324088)*x[1]+(0.989161166639)*x[2]**o+(0.261683373582)*x[2]
        s[1,1,1]=(-0.0608713083269)*x[0]**o+(-0.83002053898)*x[0]+(0.162780967784)*x[1]**o+(0.272418033641)*x[1]+(0.640461216896)*x[2]**o+(0.936590957229)*x[2]
        s[2,0,0]=(-0.015751747006)*x[0]**o+(-0.141225961145)*x[0]+(-0.598793022943)*x[1]**o+(-0.632201213977)*x[1]+(-0.96827087047)*x[2]**o+(0.938221980204)*x[2]
        s[2,0,1]=(0.868654881221)*x[0]**o+(-0.39045636307)*x[0]+(0.835370633402)*x[1]**o+(0.153562267342)*x[1]+(-0.339728418539)*x[2]**o+(-0.989504590751)*x[2]
        s[2,1,0]=(-0.738324858976)*x[0]**o+(-0.883228316124)*x[0]+(-0.32742025202)*x[1]**o+(-0.538559327566)*x[1]+(-0.327285880839)*x[2]**o+(-0.995335367768)*x[2]
        s[2,1,1]=(-0.245409556297)*x[0]**o+(-0.803489329445)*x[0]+(0.686378206879)*x[1]**o+(0.598842577126)*x[1]+(0.0633854069699)*x[2]**o+(0.865256004644)*x[2]
        s[3,0,0]=(-0.55091240715)*x[0]**o+(0.779276975802)*x[0]+(0.508012919276)*x[1]**o+(0.957172723827)*x[1]+(-0.233595787463)*x[2]**o+(-0.602493955459)*x[2]
        s[3,0,1]=(-0.287963232329)*x[0]**o+(-0.650532878115)*x[0]+(-0.622338339839)*x[1]**o+(-0.350819079399)*x[1]+(-0.953816980758)*x[2]**o+(-0.591757304037)*x[2]
        s[3,1,0]=(-0.409729121837)*x[0]**o+(0.0126023915441)*x[0]+(0.84160935605)*x[1]**o+(-0.981309168724)*x[1]+(0.0258647879557)*x[2]**o+(-0.0289412452009)*x[2]
        s[3,1,1]=(0.133740025964)*x[0]**o+(-0.214954545711)*x[0]+(-0.316781035239)*x[1]**o+(0.827691234859)*x[1]+(-0.57318095307)*x[2]**o+(-0.152707657741)*x[2]
        s[4,0,0]=(0.284463153936)*x[0]**o+(-0.349583695688)*x[0]+(-0.860205809075)*x[1]**o+(-0.29678786146)*x[1]+(0.339925025962)*x[2]**o+(0.558151297865)*x[2]
        s[4,0,1]=(0.360946013094)*x[0]**o+(0.580839810853)*x[0]+(0.0607810485038)*x[1]**o+(0.967279470731)*x[1]+(0.113379424093)*x[2]**o+(0.0406112690846)*x[2]
        s[4,1,0]=(-0.511449739485)*x[0]**o+(-0.0560394148804)*x[0]+(-0.337209156144)*x[1]**o+(-0.347040361941)*x[1]+(0.782083562886)*x[2]**o+(-0.0780823015893)*x[2]
        s[4,1,1]=(0.977977937656)*x[0]**o+(0.189561852609)*x[0]+(0.784306735372)*x[1]**o+(-0.557382590912)*x[1]+(-0.0653931079759)*x[2]**o+(-0.922120116714)*x[2]
        s[5,0,0]=(-0.061008751556)*x[0]**o+(-0.204644394893)*x[0]+(0.308938436964)*x[1]**o+(0.164825513108)*x[1]+(-0.838362859851)*x[2]**o+(-0.43799871322)*x[2]
        s[5,0,1]=(-0.260122988999)*x[0]**o+(-0.931251055816)*x[0]+(-0.757721140693)*x[1]**o+(0.866624734941)*x[1]+(0.178157743412)*x[2]**o+(0.742437867927)*x[2]
        s[5,1,0]=(0.478571128726)*x[0]**o+(0.201442942613)*x[0]+(0.231154730699)*x[1]**o+(0.147666786055)*x[1]+(-0.313353636734)*x[2]**o+(-0.503116506135)*x[2]
        s[5,1,1]=(-0.641068613201)*x[0]**o+(-0.00611578385723)*x[0]+(0.113061698591)*x[1]**o+(-0.347626408051)*x[1]+(-0.612874795512)*x[2]**o+(-0.185420324283)*x[2]
        ref[0,0,0]=(-0.789735577827)*x_ref[0]**o+(0.713461925327)*x_ref[0]+(-0.446338384344)*x_ref[1]**o+(0.383247040547)*x_ref[1]+(0.145159260009)*x_ref[2]**o+(0.329895283826)*x_ref[2]
        ref[0,0,1]=(-0.583456019572)*x_ref[0]**o+(0.409798276223)*x_ref[0]+(0.443677194056)*x_ref[1]**o+(0.686196976063)*x_ref[1]+(0.482541428129)*x_ref[2]**o+(-0.025427695773)*x_ref[2]
        ref[0,1,0]=(0.982231667045)*x_ref[0]**o+(-0.8272954182)*x_ref[0]+(0.989187050601)*x_ref[1]**o+(0.2851725407)*x_ref[1]+(-0.188195134162)*x_ref[2]**o+(-0.0748834526694)*x_ref[2]
        ref[0,1,1]=(-0.211241906114)*x_ref[0]**o+(-0.213099374182)*x_ref[0]+(-0.924671222931)*x_ref[1]**o+(-0.498297943877)*x_ref[1]+(-0.204668529404)*x_ref[2]**o+(-0.14158975248)*x_ref[2]
        ref[1,0,0]=(0.76821312782)*x_ref[0]**o+(0.0971791123021)*x_ref[0]+(-0.875792256474)*x_ref[1]**o+(0.105989140339)*x_ref[1]+(0.00172615516766)*x_ref[2]**o+(0.733396501357)*x_ref[2]
        ref[1,0,1]=(-0.586069681251)*x_ref[0]**o+(0.682163598775)*x_ref[0]+(-0.254310626636)*x_ref[1]**o+(-0.697248981934)*x_ref[1]+(-0.914707288038)*x_ref[2]**o+(0.982945477807)*x_ref[2]
        ref[1,1,0]=(0.689420994393)*x_ref[0]**o+(0.362698077408)*x_ref[0]+(0.227899444247)*x_ref[1]**o+(0.842841324088)*x_ref[1]+(0.989161166639)*x_ref[2]**o+(0.261683373582)*x_ref[2]
        ref[1,1,1]=(-0.0608713083269)*x_ref[0]**o+(-0.83002053898)*x_ref[0]+(0.162780967784)*x_ref[1]**o+(0.272418033641)*x_ref[1]+(0.640461216896)*x_ref[2]**o+(0.936590957229)*x_ref[2]
        ref[2,0,0]=(-0.015751747006)*x_ref[0]**o+(-0.141225961145)*x_ref[0]+(-0.598793022943)*x_ref[1]**o+(-0.632201213977)*x_ref[1]+(-0.96827087047)*x_ref[2]**o+(0.938221980204)*x_ref[2]
        ref[2,0,1]=(0.868654881221)*x_ref[0]**o+(-0.39045636307)*x_ref[0]+(0.835370633402)*x_ref[1]**o+(0.153562267342)*x_ref[1]+(-0.339728418539)*x_ref[2]**o+(-0.989504590751)*x_ref[2]
        ref[2,1,0]=(-0.738324858976)*x_ref[0]**o+(-0.883228316124)*x_ref[0]+(-0.32742025202)*x_ref[1]**o+(-0.538559327566)*x_ref[1]+(-0.327285880839)*x_ref[2]**o+(-0.995335367768)*x_ref[2]
        ref[2,1,1]=(-0.245409556297)*x_ref[0]**o+(-0.803489329445)*x_ref[0]+(0.686378206879)*x_ref[1]**o+(0.598842577126)*x_ref[1]+(0.0633854069699)*x_ref[2]**o+(0.865256004644)*x_ref[2]
        ref[3,0,0]=(-0.55091240715)*x_ref[0]**o+(0.779276975802)*x_ref[0]+(0.508012919276)*x_ref[1]**o+(0.957172723827)*x_ref[1]+(-0.233595787463)*x_ref[2]**o+(-0.602493955459)*x_ref[2]
        ref[3,0,1]=(-0.287963232329)*x_ref[0]**o+(-0.650532878115)*x_ref[0]+(-0.622338339839)*x_ref[1]**o+(-0.350819079399)*x_ref[1]+(-0.953816980758)*x_ref[2]**o+(-0.591757304037)*x_ref[2]
        ref[3,1,0]=(-0.409729121837)*x_ref[0]**o+(0.0126023915441)*x_ref[0]+(0.84160935605)*x_ref[1]**o+(-0.981309168724)*x_ref[1]+(0.0258647879557)*x_ref[2]**o+(-0.0289412452009)*x_ref[2]
        ref[3,1,1]=(0.133740025964)*x_ref[0]**o+(-0.214954545711)*x_ref[0]+(-0.316781035239)*x_ref[1]**o+(0.827691234859)*x_ref[1]+(-0.57318095307)*x_ref[2]**o+(-0.152707657741)*x_ref[2]
        ref[4,0,0]=(0.284463153936)*x_ref[0]**o+(-0.349583695688)*x_ref[0]+(-0.860205809075)*x_ref[1]**o+(-0.29678786146)*x_ref[1]+(0.339925025962)*x_ref[2]**o+(0.558151297865)*x_ref[2]
        ref[4,0,1]=(0.360946013094)*x_ref[0]**o+(0.580839810853)*x_ref[0]+(0.0607810485038)*x_ref[1]**o+(0.967279470731)*x_ref[1]+(0.113379424093)*x_ref[2]**o+(0.0406112690846)*x_ref[2]
        ref[4,1,0]=(-0.511449739485)*x_ref[0]**o+(-0.0560394148804)*x_ref[0]+(-0.337209156144)*x_ref[1]**o+(-0.347040361941)*x_ref[1]+(0.782083562886)*x_ref[2]**o+(-0.0780823015893)*x_ref[2]
        ref[4,1,1]=(0.977977937656)*x_ref[0]**o+(0.189561852609)*x_ref[0]+(0.784306735372)*x_ref[1]**o+(-0.557382590912)*x_ref[1]+(-0.0653931079759)*x_ref[2]**o+(-0.922120116714)*x_ref[2]
        ref[5,0,0]=(-0.061008751556)*x_ref[0]**o+(-0.204644394893)*x_ref[0]+(0.308938436964)*x_ref[1]**o+(0.164825513108)*x_ref[1]+(-0.838362859851)*x_ref[2]**o+(-0.43799871322)*x_ref[2]
        ref[5,0,1]=(-0.260122988999)*x_ref[0]**o+(-0.931251055816)*x_ref[0]+(-0.757721140693)*x_ref[1]**o+(0.866624734941)*x_ref[1]+(0.178157743412)*x_ref[2]**o+(0.742437867927)*x_ref[2]
        ref[5,1,0]=(0.478571128726)*x_ref[0]**o+(0.201442942613)*x_ref[0]+(0.231154730699)*x_ref[1]**o+(0.147666786055)*x_ref[1]+(-0.313353636734)*x_ref[2]**o+(-0.503116506135)*x_ref[2]
        ref[5,1,1]=(-0.641068613201)*x_ref[0]**o+(-0.00611578385723)*x_ref[0]+(0.113061698591)*x_ref[1]**o+(-0.347626408051)*x_ref[1]+(-0.612874795512)*x_ref[2]**o+(-0.185420324283)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onSolution_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the Solution

      assumptions: self.domain supports inpterpolation from Solution onto Solution
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Solution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(0.660866943555)*x[0]**o+(-0.69695057937)*x[0]+(-0.250467442461)*x[1]**o+(-0.274418805949)*x[1]
        s[0,0,0,1]=(0.000608842108103)*x[0]**o+(0.696871568098)*x[0]+(0.0187896686535)*x[1]**o+(-0.0453529548001)*x[1]
        s[0,0,1,0]=(-0.240765142378)*x[0]**o+(0.846756487798)*x[0]+(0.891604103164)*x[1]**o+(0.721802825508)*x[1]
        s[0,0,1,1]=(-0.70530582062)*x[0]**o+(-0.334127022771)*x[0]+(-0.95574798294)*x[1]**o+(-0.00757492972355)*x[1]
        s[0,0,2,0]=(0.632035299678)*x[0]**o+(-0.699918396333)*x[0]+(-0.663604240838)*x[1]**o+(0.900905470496)*x[1]
        s[0,0,2,1]=(-0.426120671328)*x[0]**o+(-0.844531284589)*x[0]+(0.459314587139)*x[1]**o+(-0.870486877148)*x[1]
        s[0,1,0,0]=(-0.566504175673)*x[0]**o+(0.0116278724505)*x[0]+(0.897736919548)*x[1]**o+(-0.288446089721)*x[1]
        s[0,1,0,1]=(0.546502403758)*x[0]**o+(0.573073492121)*x[0]+(0.716218807816)*x[1]**o+(0.485057461309)*x[1]
        s[0,1,1,0]=(-0.74532715864)*x[0]**o+(-0.84923938709)*x[0]+(-0.534608284813)*x[1]**o+(0.7197051089)*x[1]
        s[0,1,1,1]=(-0.921328665174)*x[0]**o+(-0.901425527075)*x[0]+(-0.216486707202)*x[1]**o+(-0.67173056569)*x[1]
        s[0,1,2,0]=(-0.959516269092)*x[0]**o+(-0.465841970545)*x[0]+(-0.235774367828)*x[1]**o+(0.442773930914)*x[1]
        s[0,1,2,1]=(0.711637013768)*x[0]**o+(-0.418137384829)*x[0]+(0.486530293439)*x[1]**o+(-0.499106697555)*x[1]
        s[0,2,0,0]=(-0.333968500743)*x[0]**o+(0.221903342892)*x[0]+(-0.455272799456)*x[1]**o+(0.908238147055)*x[1]
        s[0,2,0,1]=(0.502278138368)*x[0]**o+(-0.457108256032)*x[0]+(0.948658345868)*x[1]**o+(0.298218610775)*x[1]
        s[0,2,1,0]=(-0.674400286676)*x[0]**o+(0.951670945597)*x[0]+(0.230561733409)*x[1]**o+(-0.370171317738)*x[1]
        s[0,2,1,1]=(0.870778786508)*x[0]**o+(-0.0330884780249)*x[0]+(0.906960508962)*x[1]**o+(0.530885972049)*x[1]
        s[0,2,2,0]=(0.778651165953)*x[0]**o+(-0.84832814955)*x[0]+(0.190161186455)*x[1]**o+(0.754714785008)*x[1]
        s[0,2,2,1]=(0.815802649658)*x[0]**o+(0.195022234096)*x[0]+(0.740446839482)*x[1]**o+(-0.636095294959)*x[1]
        s[0,3,0,0]=(-0.522361421038)*x[0]**o+(0.767027676789)*x[0]+(0.462273930511)*x[1]**o+(0.596970109634)*x[1]
        s[0,3,0,1]=(-0.453964613059)*x[0]**o+(-0.893497373251)*x[0]+(0.758041901161)*x[1]**o+(0.844733400115)*x[1]
        s[0,3,1,0]=(0.974979048326)*x[0]**o+(-0.768486788284)*x[0]+(0.219102891257)*x[1]**o+(0.0619004645998)*x[1]
        s[0,3,1,1]=(0.0623620754035)*x[0]**o+(0.565586249437)*x[0]+(0.400114536695)*x[1]**o+(0.522879378012)*x[1]
        s[0,3,2,0]=(0.70720180819)*x[0]**o+(-0.137294797593)*x[0]+(-0.524395446851)*x[1]**o+(0.366427763772)*x[1]
        s[0,3,2,1]=(0.652058057827)*x[0]**o+(0.649118057169)*x[0]+(-0.73680433522)*x[1]**o+(0.290439763876)*x[1]
        s[0,4,0,0]=(-0.751426506393)*x[0]**o+(-0.610061668119)*x[0]+(-0.991937875988)*x[1]**o+(0.506340630761)*x[1]
        s[0,4,0,1]=(0.350661537672)*x[0]**o+(-0.107364823572)*x[0]+(0.0434206596654)*x[1]**o+(-0.292547533181)*x[1]
        s[0,4,1,0]=(-0.700676601904)*x[0]**o+(-0.837588039311)*x[0]+(0.224725557518)*x[1]**o+(-0.460649043112)*x[1]
        s[0,4,1,1]=(0.0601174037971)*x[0]**o+(-0.511103838233)*x[0]+(-0.0457685911136)*x[1]**o+(0.844052286928)*x[1]
        s[0,4,2,0]=(0.868703323043)*x[0]**o+(0.332921136135)*x[0]+(-0.752420352376)*x[1]**o+(0.596422517404)*x[1]
        s[0,4,2,1]=(-0.502917500488)*x[0]**o+(-0.17936644769)*x[0]+(0.528529056308)*x[1]**o+(-0.310716591912)*x[1]
        s[1,0,0,0]=(0.306103545562)*x[0]**o+(-0.4622631399)*x[0]+(0.481409282179)*x[1]**o+(-0.664712050849)*x[1]
        s[1,0,0,1]=(-0.366801423523)*x[0]**o+(0.313906121036)*x[0]+(0.654132121812)*x[1]**o+(0.888013275901)*x[1]
        s[1,0,1,0]=(-0.133206578634)*x[0]**o+(0.959683755236)*x[0]+(0.854831306595)*x[1]**o+(-0.134068370471)*x[1]
        s[1,0,1,1]=(-0.724230066407)*x[0]**o+(0.294223474488)*x[0]+(-0.712156264035)*x[1]**o+(0.141030523454)*x[1]
        s[1,0,2,0]=(-0.759479778952)*x[0]**o+(0.350452958391)*x[0]+(-0.678584005775)*x[1]**o+(0.861034825483)*x[1]
        s[1,0,2,1]=(-0.787253725243)*x[0]**o+(-0.124038882324)*x[0]+(-0.853758990369)*x[1]**o+(0.316695465913)*x[1]
        s[1,1,0,0]=(-0.792535933434)*x[0]**o+(0.827296938961)*x[0]+(-0.781181526198)*x[1]**o+(-0.501946069669)*x[1]
        s[1,1,0,1]=(0.125764776751)*x[0]**o+(-0.22197448981)*x[0]+(-0.525720822467)*x[1]**o+(0.491502391671)*x[1]
        s[1,1,1,0]=(-0.443164997733)*x[0]**o+(0.903514594855)*x[0]+(-0.209034942956)*x[1]**o+(-0.944420946193)*x[1]
        s[1,1,1,1]=(-0.0733577966014)*x[0]**o+(0.64287074783)*x[0]+(0.976307308908)*x[1]**o+(-0.859994995094)*x[1]
        s[1,1,2,0]=(0.26008702136)*x[0]**o+(0.158663809624)*x[0]+(0.532980284319)*x[1]**o+(0.732083505032)*x[1]
        s[1,1,2,1]=(-0.735363642678)*x[0]**o+(-0.0449828239956)*x[0]+(0.225177892214)*x[1]**o+(0.872692876616)*x[1]
        s[1,2,0,0]=(-0.407076115111)*x[0]**o+(-0.258588469747)*x[0]+(-0.961556153809)*x[1]**o+(-0.121191173495)*x[1]
        s[1,2,0,1]=(-0.628930332498)*x[0]**o+(-0.865981936989)*x[0]+(-0.193969072453)*x[1]**o+(0.281837632563)*x[1]
        s[1,2,1,0]=(-0.528238487527)*x[0]**o+(-0.956371806399)*x[0]+(0.526934236806)*x[1]**o+(0.561448403552)*x[1]
        s[1,2,1,1]=(-0.785141153502)*x[0]**o+(0.897147964347)*x[0]+(0.283588200891)*x[1]**o+(0.0818048242882)*x[1]
        s[1,2,2,0]=(0.479988338985)*x[0]**o+(0.744353706546)*x[0]+(-0.0889121554735)*x[1]**o+(0.0663836364303)*x[1]
        s[1,2,2,1]=(-0.433422910155)*x[0]**o+(0.413382234167)*x[0]+(0.491055005038)*x[1]**o+(0.226589212433)*x[1]
        s[1,3,0,0]=(0.0534747766519)*x[0]**o+(0.842086781016)*x[0]+(-0.26010049679)*x[1]**o+(0.912642784579)*x[1]
        s[1,3,0,1]=(-0.0496511875565)*x[0]**o+(-0.831079045277)*x[0]+(-0.670769124412)*x[1]**o+(-0.193175103104)*x[1]
        s[1,3,1,0]=(-0.298941035368)*x[0]**o+(0.317545691104)*x[0]+(0.819208983648)*x[1]**o+(-0.811364813886)*x[1]
        s[1,3,1,1]=(-0.636689800754)*x[0]**o+(-0.271123157494)*x[0]+(0.971256227556)*x[1]**o+(0.551860375765)*x[1]
        s[1,3,2,0]=(-0.41793819747)*x[0]**o+(0.0455167282452)*x[0]+(-0.100744221021)*x[1]**o+(-0.0938585201979)*x[1]
        s[1,3,2,1]=(0.773358459961)*x[0]**o+(-0.238180123205)*x[0]+(-0.164546046241)*x[1]**o+(0.766225394329)*x[1]
        s[1,4,0,0]=(-0.574065286667)*x[0]**o+(0.439058157914)*x[0]+(0.67083169123)*x[1]**o+(0.130659800696)*x[1]
        s[1,4,0,1]=(0.162394241218)*x[0]**o+(-0.891220801239)*x[0]+(-0.955097003404)*x[1]**o+(0.13773131437)*x[1]
        s[1,4,1,0]=(-0.74162834841)*x[0]**o+(-0.703072461876)*x[0]+(-0.343008525881)*x[1]**o+(-0.151188176108)*x[1]
        s[1,4,1,1]=(0.930997317905)*x[0]**o+(-0.456365664196)*x[0]+(-0.980371151527)*x[1]**o+(0.850653673992)*x[1]
        s[1,4,2,0]=(0.500808288015)*x[0]**o+(0.288371444998)*x[0]+(-0.391595004735)*x[1]**o+(0.13882243776)*x[1]
        s[1,4,2,1]=(0.389530918637)*x[0]**o+(-0.67935582556)*x[0]+(-0.66508876501)*x[1]**o+(-0.638594177603)*x[1]
        s[2,0,0,0]=(-0.726607190425)*x[0]**o+(0.35987718022)*x[0]+(0.198286208235)*x[1]**o+(-0.10467945887)*x[1]
        s[2,0,0,1]=(0.553840848629)*x[0]**o+(-0.805320208727)*x[0]+(-0.386397083568)*x[1]**o+(-0.0376567352009)*x[1]
        s[2,0,1,0]=(-0.490743817014)*x[0]**o+(0.865832409075)*x[0]+(-0.808848699394)*x[1]**o+(0.986023649801)*x[1]
        s[2,0,1,1]=(-0.410719421383)*x[0]**o+(0.95618705464)*x[0]+(-0.365709247367)*x[1]**o+(-0.599794878906)*x[1]
        s[2,0,2,0]=(-0.288732164624)*x[0]**o+(-0.110131049055)*x[0]+(-0.790190148557)*x[1]**o+(-0.646184208603)*x[1]
        s[2,0,2,1]=(0.386423919129)*x[0]**o+(0.639090033283)*x[0]+(0.32377411877)*x[1]**o+(0.421258365826)*x[1]
        s[2,1,0,0]=(0.0852900668985)*x[0]**o+(-0.221116937927)*x[0]+(0.999573518383)*x[1]**o+(0.717665452429)*x[1]
        s[2,1,0,1]=(0.0585991419426)*x[0]**o+(-0.51528127078)*x[0]+(0.995668094637)*x[1]**o+(0.666963843202)*x[1]
        s[2,1,1,0]=(-0.452964121227)*x[0]**o+(0.326598939914)*x[0]+(0.832818926454)*x[1]**o+(0.274512132514)*x[1]
        s[2,1,1,1]=(0.596298524791)*x[0]**o+(0.307217908684)*x[0]+(0.517972533541)*x[1]**o+(-0.844413041951)*x[1]
        s[2,1,2,0]=(-0.364801163986)*x[0]**o+(-0.330550039809)*x[0]+(-0.633596555924)*x[1]**o+(0.670455930999)*x[1]
        s[2,1,2,1]=(-0.422877677484)*x[0]**o+(0.0552246586047)*x[0]+(-0.495189898204)*x[1]**o+(-0.583721895158)*x[1]
        s[2,2,0,0]=(-0.260229907075)*x[0]**o+(0.177649435874)*x[0]+(0.411800309789)*x[1]**o+(0.387048527641)*x[1]
        s[2,2,0,1]=(0.242333136871)*x[0]**o+(-0.588296408385)*x[0]+(0.467570088467)*x[1]**o+(0.298734168304)*x[1]
        s[2,2,1,0]=(0.259928708536)*x[0]**o+(0.653483739103)*x[0]+(-0.995039631674)*x[1]**o+(-0.413239600116)*x[1]
        s[2,2,1,1]=(0.0152698121722)*x[0]**o+(0.823258564878)*x[0]+(-0.138386765925)*x[1]**o+(-0.468138866813)*x[1]
        s[2,2,2,0]=(-0.965031041562)*x[0]**o+(0.737382561376)*x[0]+(-0.106502248694)*x[1]**o+(-0.717199422327)*x[1]
        s[2,2,2,1]=(-0.482938567193)*x[0]**o+(-0.491533829004)*x[0]+(0.0791168083042)*x[1]**o+(-0.63228810248)*x[1]
        s[2,3,0,0]=(-0.828344342476)*x[0]**o+(-0.781483429674)*x[0]+(0.867903846758)*x[1]**o+(0.569111225986)*x[1]
        s[2,3,0,1]=(0.659597328826)*x[0]**o+(0.532419181635)*x[0]+(-0.80620973473)*x[1]**o+(-0.46673471071)*x[1]
        s[2,3,1,0]=(-0.885134823658)*x[0]**o+(-0.546337219398)*x[0]+(0.166804034744)*x[1]**o+(0.299587086812)*x[1]
        s[2,3,1,1]=(0.112360533523)*x[0]**o+(0.0953930412409)*x[0]+(-0.97632552045)*x[1]**o+(0.945659336453)*x[1]
        s[2,3,2,0]=(0.678405267904)*x[0]**o+(0.342440777945)*x[0]+(-0.99053148169)*x[1]**o+(-0.969942468062)*x[1]
        s[2,3,2,1]=(-0.927867119592)*x[0]**o+(-0.690642273601)*x[0]+(-0.742887119506)*x[1]**o+(-0.271373802799)*x[1]
        s[2,4,0,0]=(0.147456822243)*x[0]**o+(-0.30678558608)*x[0]+(0.824457828303)*x[1]**o+(-0.236387246684)*x[1]
        s[2,4,0,1]=(-0.419906137988)*x[0]**o+(0.940583417457)*x[0]+(-0.252744709881)*x[1]**o+(-0.311929952137)*x[1]
        s[2,4,1,0]=(0.676956544501)*x[0]**o+(-0.0569121505607)*x[0]+(0.518218768651)*x[1]**o+(-0.14376234793)*x[1]
        s[2,4,1,1]=(-0.438909519823)*x[0]**o+(0.363536809935)*x[0]+(0.594959566494)*x[1]**o+(0.161508700798)*x[1]
        s[2,4,2,0]=(-0.398965010329)*x[0]**o+(-0.519311871529)*x[0]+(0.0996092594307)*x[1]**o+(0.299113884607)*x[1]
        s[2,4,2,1]=(0.438911683911)*x[0]**o+(-0.330751949884)*x[0]+(0.72396811226)*x[1]**o+(-0.378618495259)*x[1]
        s[3,0,0,0]=(0.889235578987)*x[0]**o+(0.301798885253)*x[0]+(-0.876751409061)*x[1]**o+(0.474669758163)*x[1]
        s[3,0,0,1]=(0.935815251238)*x[0]**o+(0.674255198571)*x[0]+(0.341766237248)*x[1]**o+(-0.72357698277)*x[1]
        s[3,0,1,0]=(0.419855380231)*x[0]**o+(0.755034058786)*x[0]+(-0.59112491561)*x[1]**o+(-0.288034233407)*x[1]
        s[3,0,1,1]=(-0.403840813699)*x[0]**o+(0.770567211689)*x[0]+(0.416460080616)*x[1]**o+(-0.632767925192)*x[1]
        s[3,0,2,0]=(-0.847006017592)*x[0]**o+(-0.447883355706)*x[0]+(-0.628290176502)*x[1]**o+(0.360720747081)*x[1]
        s[3,0,2,1]=(0.843924560726)*x[0]**o+(-0.887772077659)*x[0]+(-0.0951024444874)*x[1]**o+(0.218823836893)*x[1]
        s[3,1,0,0]=(0.574007571036)*x[0]**o+(-0.248963911474)*x[0]+(-0.899931219543)*x[1]**o+(0.710520579857)*x[1]
        s[3,1,0,1]=(-0.667806054096)*x[0]**o+(-0.241865813708)*x[0]+(0.588561916587)*x[1]**o+(0.788227756226)*x[1]
        s[3,1,1,0]=(-0.258519497867)*x[0]**o+(0.645611508516)*x[0]+(-0.69368336683)*x[1]**o+(0.217859823865)*x[1]
        s[3,1,1,1]=(-0.863926174373)*x[0]**o+(-0.0852128898867)*x[0]+(-0.648543494951)*x[1]**o+(-0.402985545677)*x[1]
        s[3,1,2,0]=(-0.499342603145)*x[0]**o+(0.10223769398)*x[0]+(-0.968301649305)*x[1]**o+(0.295206479399)*x[1]
        s[3,1,2,1]=(-0.190629994791)*x[0]**o+(-0.795765651657)*x[0]+(0.793726572154)*x[1]**o+(-0.367897693585)*x[1]
        s[3,2,0,0]=(0.315102787334)*x[0]**o+(0.21152066219)*x[0]+(0.227000389892)*x[1]**o+(-0.0796967599621)*x[1]
        s[3,2,0,1]=(0.826049810996)*x[0]**o+(0.281856067318)*x[0]+(0.433050065825)*x[1]**o+(-0.218614730929)*x[1]
        s[3,2,1,0]=(0.369148754088)*x[0]**o+(0.800521572309)*x[0]+(0.877288542563)*x[1]**o+(-0.135428392825)*x[1]
        s[3,2,1,1]=(0.808782390733)*x[0]**o+(-0.326877566881)*x[0]+(0.771288189422)*x[1]**o+(-0.906951778691)*x[1]
        s[3,2,2,0]=(-0.381896054964)*x[0]**o+(0.142176864318)*x[0]+(-0.636114766341)*x[1]**o+(-0.793416969698)*x[1]
        s[3,2,2,1]=(-0.291092484455)*x[0]**o+(0.679789632507)*x[0]+(0.882235235177)*x[1]**o+(-0.826738171565)*x[1]
        s[3,3,0,0]=(0.71811796826)*x[0]**o+(0.0627548627403)*x[0]+(0.376656804533)*x[1]**o+(-0.00934001961218)*x[1]
        s[3,3,0,1]=(-0.509178665563)*x[0]**o+(0.804608293206)*x[0]+(-0.752113912219)*x[1]**o+(-0.313105572748)*x[1]
        s[3,3,1,0]=(-0.896783097316)*x[0]**o+(-0.560162135134)*x[0]+(-0.830841561786)*x[1]**o+(-0.333506968769)*x[1]
        s[3,3,1,1]=(0.322497189516)*x[0]**o+(0.277292905042)*x[0]+(0.631980511231)*x[1]**o+(0.926483120043)*x[1]
        s[3,3,2,0]=(-0.995133214294)*x[0]**o+(-0.501556834675)*x[0]+(-0.279959032716)*x[1]**o+(-0.606903349874)*x[1]
        s[3,3,2,1]=(-0.52901549305)*x[0]**o+(-0.345583335877)*x[0]+(0.613327737258)*x[1]**o+(-0.235057519501)*x[1]
        s[3,4,0,0]=(0.129978557871)*x[0]**o+(-0.145655660564)*x[0]+(-0.217775687907)*x[1]**o+(0.528084545073)*x[1]
        s[3,4,0,1]=(0.94863938123)*x[0]**o+(-0.858645154335)*x[0]+(0.950889681616)*x[1]**o+(-0.560518438352)*x[1]
        s[3,4,1,0]=(0.775184423382)*x[0]**o+(0.783822082576)*x[0]+(-0.130919265341)*x[1]**o+(0.491447558856)*x[1]
        s[3,4,1,1]=(-0.383273309117)*x[0]**o+(0.268662318873)*x[0]+(0.0821612756798)*x[1]**o+(0.508481517445)*x[1]
        s[3,4,2,0]=(-0.151657688401)*x[0]**o+(0.661522859198)*x[0]+(0.082765340123)*x[1]**o+(-0.209987690619)*x[1]
        s[3,4,2,1]=(-0.137879400495)*x[0]**o+(-0.291576763731)*x[0]+(-0.342092940232)*x[1]**o+(-0.486411344299)*x[1]
        ref[0,0,0,0]=(0.660866943555)*x_ref[0]**o+(-0.69695057937)*x_ref[0]+(-0.250467442461)*x_ref[1]**o+(-0.274418805949)*x_ref[1]
        ref[0,0,0,1]=(0.000608842108103)*x_ref[0]**o+(0.696871568098)*x_ref[0]+(0.0187896686535)*x_ref[1]**o+(-0.0453529548001)*x_ref[1]
        ref[0,0,1,0]=(-0.240765142378)*x_ref[0]**o+(0.846756487798)*x_ref[0]+(0.891604103164)*x_ref[1]**o+(0.721802825508)*x_ref[1]
        ref[0,0,1,1]=(-0.70530582062)*x_ref[0]**o+(-0.334127022771)*x_ref[0]+(-0.95574798294)*x_ref[1]**o+(-0.00757492972355)*x_ref[1]
        ref[0,0,2,0]=(0.632035299678)*x_ref[0]**o+(-0.699918396333)*x_ref[0]+(-0.663604240838)*x_ref[1]**o+(0.900905470496)*x_ref[1]
        ref[0,0,2,1]=(-0.426120671328)*x_ref[0]**o+(-0.844531284589)*x_ref[0]+(0.459314587139)*x_ref[1]**o+(-0.870486877148)*x_ref[1]
        ref[0,1,0,0]=(-0.566504175673)*x_ref[0]**o+(0.0116278724505)*x_ref[0]+(0.897736919548)*x_ref[1]**o+(-0.288446089721)*x_ref[1]
        ref[0,1,0,1]=(0.546502403758)*x_ref[0]**o+(0.573073492121)*x_ref[0]+(0.716218807816)*x_ref[1]**o+(0.485057461309)*x_ref[1]
        ref[0,1,1,0]=(-0.74532715864)*x_ref[0]**o+(-0.84923938709)*x_ref[0]+(-0.534608284813)*x_ref[1]**o+(0.7197051089)*x_ref[1]
        ref[0,1,1,1]=(-0.921328665174)*x_ref[0]**o+(-0.901425527075)*x_ref[0]+(-0.216486707202)*x_ref[1]**o+(-0.67173056569)*x_ref[1]
        ref[0,1,2,0]=(-0.959516269092)*x_ref[0]**o+(-0.465841970545)*x_ref[0]+(-0.235774367828)*x_ref[1]**o+(0.442773930914)*x_ref[1]
        ref[0,1,2,1]=(0.711637013768)*x_ref[0]**o+(-0.418137384829)*x_ref[0]+(0.486530293439)*x_ref[1]**o+(-0.499106697555)*x_ref[1]
        ref[0,2,0,0]=(-0.333968500743)*x_ref[0]**o+(0.221903342892)*x_ref[0]+(-0.455272799456)*x_ref[1]**o+(0.908238147055)*x_ref[1]
        ref[0,2,0,1]=(0.502278138368)*x_ref[0]**o+(-0.457108256032)*x_ref[0]+(0.948658345868)*x_ref[1]**o+(0.298218610775)*x_ref[1]
        ref[0,2,1,0]=(-0.674400286676)*x_ref[0]**o+(0.951670945597)*x_ref[0]+(0.230561733409)*x_ref[1]**o+(-0.370171317738)*x_ref[1]
        ref[0,2,1,1]=(0.870778786508)*x_ref[0]**o+(-0.0330884780249)*x_ref[0]+(0.906960508962)*x_ref[1]**o+(0.530885972049)*x_ref[1]
        ref[0,2,2,0]=(0.778651165953)*x_ref[0]**o+(-0.84832814955)*x_ref[0]+(0.190161186455)*x_ref[1]**o+(0.754714785008)*x_ref[1]
        ref[0,2,2,1]=(0.815802649658)*x_ref[0]**o+(0.195022234096)*x_ref[0]+(0.740446839482)*x_ref[1]**o+(-0.636095294959)*x_ref[1]
        ref[0,3,0,0]=(-0.522361421038)*x_ref[0]**o+(0.767027676789)*x_ref[0]+(0.462273930511)*x_ref[1]**o+(0.596970109634)*x_ref[1]
        ref[0,3,0,1]=(-0.453964613059)*x_ref[0]**o+(-0.893497373251)*x_ref[0]+(0.758041901161)*x_ref[1]**o+(0.844733400115)*x_ref[1]
        ref[0,3,1,0]=(0.974979048326)*x_ref[0]**o+(-0.768486788284)*x_ref[0]+(0.219102891257)*x_ref[1]**o+(0.0619004645998)*x_ref[1]
        ref[0,3,1,1]=(0.0623620754035)*x_ref[0]**o+(0.565586249437)*x_ref[0]+(0.400114536695)*x_ref[1]**o+(0.522879378012)*x_ref[1]
        ref[0,3,2,0]=(0.70720180819)*x_ref[0]**o+(-0.137294797593)*x_ref[0]+(-0.524395446851)*x_ref[1]**o+(0.366427763772)*x_ref[1]
        ref[0,3,2,1]=(0.652058057827)*x_ref[0]**o+(0.649118057169)*x_ref[0]+(-0.73680433522)*x_ref[1]**o+(0.290439763876)*x_ref[1]
        ref[0,4,0,0]=(-0.751426506393)*x_ref[0]**o+(-0.610061668119)*x_ref[0]+(-0.991937875988)*x_ref[1]**o+(0.506340630761)*x_ref[1]
        ref[0,4,0,1]=(0.350661537672)*x_ref[0]**o+(-0.107364823572)*x_ref[0]+(0.0434206596654)*x_ref[1]**o+(-0.292547533181)*x_ref[1]
        ref[0,4,1,0]=(-0.700676601904)*x_ref[0]**o+(-0.837588039311)*x_ref[0]+(0.224725557518)*x_ref[1]**o+(-0.460649043112)*x_ref[1]
        ref[0,4,1,1]=(0.0601174037971)*x_ref[0]**o+(-0.511103838233)*x_ref[0]+(-0.0457685911136)*x_ref[1]**o+(0.844052286928)*x_ref[1]
        ref[0,4,2,0]=(0.868703323043)*x_ref[0]**o+(0.332921136135)*x_ref[0]+(-0.752420352376)*x_ref[1]**o+(0.596422517404)*x_ref[1]
        ref[0,4,2,1]=(-0.502917500488)*x_ref[0]**o+(-0.17936644769)*x_ref[0]+(0.528529056308)*x_ref[1]**o+(-0.310716591912)*x_ref[1]
        ref[1,0,0,0]=(0.306103545562)*x_ref[0]**o+(-0.4622631399)*x_ref[0]+(0.481409282179)*x_ref[1]**o+(-0.664712050849)*x_ref[1]
        ref[1,0,0,1]=(-0.366801423523)*x_ref[0]**o+(0.313906121036)*x_ref[0]+(0.654132121812)*x_ref[1]**o+(0.888013275901)*x_ref[1]
        ref[1,0,1,0]=(-0.133206578634)*x_ref[0]**o+(0.959683755236)*x_ref[0]+(0.854831306595)*x_ref[1]**o+(-0.134068370471)*x_ref[1]
        ref[1,0,1,1]=(-0.724230066407)*x_ref[0]**o+(0.294223474488)*x_ref[0]+(-0.712156264035)*x_ref[1]**o+(0.141030523454)*x_ref[1]
        ref[1,0,2,0]=(-0.759479778952)*x_ref[0]**o+(0.350452958391)*x_ref[0]+(-0.678584005775)*x_ref[1]**o+(0.861034825483)*x_ref[1]
        ref[1,0,2,1]=(-0.787253725243)*x_ref[0]**o+(-0.124038882324)*x_ref[0]+(-0.853758990369)*x_ref[1]**o+(0.316695465913)*x_ref[1]
        ref[1,1,0,0]=(-0.792535933434)*x_ref[0]**o+(0.827296938961)*x_ref[0]+(-0.781181526198)*x_ref[1]**o+(-0.501946069669)*x_ref[1]
        ref[1,1,0,1]=(0.125764776751)*x_ref[0]**o+(-0.22197448981)*x_ref[0]+(-0.525720822467)*x_ref[1]**o+(0.491502391671)*x_ref[1]
        ref[1,1,1,0]=(-0.443164997733)*x_ref[0]**o+(0.903514594855)*x_ref[0]+(-0.209034942956)*x_ref[1]**o+(-0.944420946193)*x_ref[1]
        ref[1,1,1,1]=(-0.0733577966014)*x_ref[0]**o+(0.64287074783)*x_ref[0]+(0.976307308908)*x_ref[1]**o+(-0.859994995094)*x_ref[1]
        ref[1,1,2,0]=(0.26008702136)*x_ref[0]**o+(0.158663809624)*x_ref[0]+(0.532980284319)*x_ref[1]**o+(0.732083505032)*x_ref[1]
        ref[1,1,2,1]=(-0.735363642678)*x_ref[0]**o+(-0.0449828239956)*x_ref[0]+(0.225177892214)*x_ref[1]**o+(0.872692876616)*x_ref[1]
        ref[1,2,0,0]=(-0.407076115111)*x_ref[0]**o+(-0.258588469747)*x_ref[0]+(-0.961556153809)*x_ref[1]**o+(-0.121191173495)*x_ref[1]
        ref[1,2,0,1]=(-0.628930332498)*x_ref[0]**o+(-0.865981936989)*x_ref[0]+(-0.193969072453)*x_ref[1]**o+(0.281837632563)*x_ref[1]
        ref[1,2,1,0]=(-0.528238487527)*x_ref[0]**o+(-0.956371806399)*x_ref[0]+(0.526934236806)*x_ref[1]**o+(0.561448403552)*x_ref[1]
        ref[1,2,1,1]=(-0.785141153502)*x_ref[0]**o+(0.897147964347)*x_ref[0]+(0.283588200891)*x_ref[1]**o+(0.0818048242882)*x_ref[1]
        ref[1,2,2,0]=(0.479988338985)*x_ref[0]**o+(0.744353706546)*x_ref[0]+(-0.0889121554735)*x_ref[1]**o+(0.0663836364303)*x_ref[1]
        ref[1,2,2,1]=(-0.433422910155)*x_ref[0]**o+(0.413382234167)*x_ref[0]+(0.491055005038)*x_ref[1]**o+(0.226589212433)*x_ref[1]
        ref[1,3,0,0]=(0.0534747766519)*x_ref[0]**o+(0.842086781016)*x_ref[0]+(-0.26010049679)*x_ref[1]**o+(0.912642784579)*x_ref[1]
        ref[1,3,0,1]=(-0.0496511875565)*x_ref[0]**o+(-0.831079045277)*x_ref[0]+(-0.670769124412)*x_ref[1]**o+(-0.193175103104)*x_ref[1]
        ref[1,3,1,0]=(-0.298941035368)*x_ref[0]**o+(0.317545691104)*x_ref[0]+(0.819208983648)*x_ref[1]**o+(-0.811364813886)*x_ref[1]
        ref[1,3,1,1]=(-0.636689800754)*x_ref[0]**o+(-0.271123157494)*x_ref[0]+(0.971256227556)*x_ref[1]**o+(0.551860375765)*x_ref[1]
        ref[1,3,2,0]=(-0.41793819747)*x_ref[0]**o+(0.0455167282452)*x_ref[0]+(-0.100744221021)*x_ref[1]**o+(-0.0938585201979)*x_ref[1]
        ref[1,3,2,1]=(0.773358459961)*x_ref[0]**o+(-0.238180123205)*x_ref[0]+(-0.164546046241)*x_ref[1]**o+(0.766225394329)*x_ref[1]
        ref[1,4,0,0]=(-0.574065286667)*x_ref[0]**o+(0.439058157914)*x_ref[0]+(0.67083169123)*x_ref[1]**o+(0.130659800696)*x_ref[1]
        ref[1,4,0,1]=(0.162394241218)*x_ref[0]**o+(-0.891220801239)*x_ref[0]+(-0.955097003404)*x_ref[1]**o+(0.13773131437)*x_ref[1]
        ref[1,4,1,0]=(-0.74162834841)*x_ref[0]**o+(-0.703072461876)*x_ref[0]+(-0.343008525881)*x_ref[1]**o+(-0.151188176108)*x_ref[1]
        ref[1,4,1,1]=(0.930997317905)*x_ref[0]**o+(-0.456365664196)*x_ref[0]+(-0.980371151527)*x_ref[1]**o+(0.850653673992)*x_ref[1]
        ref[1,4,2,0]=(0.500808288015)*x_ref[0]**o+(0.288371444998)*x_ref[0]+(-0.391595004735)*x_ref[1]**o+(0.13882243776)*x_ref[1]
        ref[1,4,2,1]=(0.389530918637)*x_ref[0]**o+(-0.67935582556)*x_ref[0]+(-0.66508876501)*x_ref[1]**o+(-0.638594177603)*x_ref[1]
        ref[2,0,0,0]=(-0.726607190425)*x_ref[0]**o+(0.35987718022)*x_ref[0]+(0.198286208235)*x_ref[1]**o+(-0.10467945887)*x_ref[1]
        ref[2,0,0,1]=(0.553840848629)*x_ref[0]**o+(-0.805320208727)*x_ref[0]+(-0.386397083568)*x_ref[1]**o+(-0.0376567352009)*x_ref[1]
        ref[2,0,1,0]=(-0.490743817014)*x_ref[0]**o+(0.865832409075)*x_ref[0]+(-0.808848699394)*x_ref[1]**o+(0.986023649801)*x_ref[1]
        ref[2,0,1,1]=(-0.410719421383)*x_ref[0]**o+(0.95618705464)*x_ref[0]+(-0.365709247367)*x_ref[1]**o+(-0.599794878906)*x_ref[1]
        ref[2,0,2,0]=(-0.288732164624)*x_ref[0]**o+(-0.110131049055)*x_ref[0]+(-0.790190148557)*x_ref[1]**o+(-0.646184208603)*x_ref[1]
        ref[2,0,2,1]=(0.386423919129)*x_ref[0]**o+(0.639090033283)*x_ref[0]+(0.32377411877)*x_ref[1]**o+(0.421258365826)*x_ref[1]
        ref[2,1,0,0]=(0.0852900668985)*x_ref[0]**o+(-0.221116937927)*x_ref[0]+(0.999573518383)*x_ref[1]**o+(0.717665452429)*x_ref[1]
        ref[2,1,0,1]=(0.0585991419426)*x_ref[0]**o+(-0.51528127078)*x_ref[0]+(0.995668094637)*x_ref[1]**o+(0.666963843202)*x_ref[1]
        ref[2,1,1,0]=(-0.452964121227)*x_ref[0]**o+(0.326598939914)*x_ref[0]+(0.832818926454)*x_ref[1]**o+(0.274512132514)*x_ref[1]
        ref[2,1,1,1]=(0.596298524791)*x_ref[0]**o+(0.307217908684)*x_ref[0]+(0.517972533541)*x_ref[1]**o+(-0.844413041951)*x_ref[1]
        ref[2,1,2,0]=(-0.364801163986)*x_ref[0]**o+(-0.330550039809)*x_ref[0]+(-0.633596555924)*x_ref[1]**o+(0.670455930999)*x_ref[1]
        ref[2,1,2,1]=(-0.422877677484)*x_ref[0]**o+(0.0552246586047)*x_ref[0]+(-0.495189898204)*x_ref[1]**o+(-0.583721895158)*x_ref[1]
        ref[2,2,0,0]=(-0.260229907075)*x_ref[0]**o+(0.177649435874)*x_ref[0]+(0.411800309789)*x_ref[1]**o+(0.387048527641)*x_ref[1]
        ref[2,2,0,1]=(0.242333136871)*x_ref[0]**o+(-0.588296408385)*x_ref[0]+(0.467570088467)*x_ref[1]**o+(0.298734168304)*x_ref[1]
        ref[2,2,1,0]=(0.259928708536)*x_ref[0]**o+(0.653483739103)*x_ref[0]+(-0.995039631674)*x_ref[1]**o+(-0.413239600116)*x_ref[1]
        ref[2,2,1,1]=(0.0152698121722)*x_ref[0]**o+(0.823258564878)*x_ref[0]+(-0.138386765925)*x_ref[1]**o+(-0.468138866813)*x_ref[1]
        ref[2,2,2,0]=(-0.965031041562)*x_ref[0]**o+(0.737382561376)*x_ref[0]+(-0.106502248694)*x_ref[1]**o+(-0.717199422327)*x_ref[1]
        ref[2,2,2,1]=(-0.482938567193)*x_ref[0]**o+(-0.491533829004)*x_ref[0]+(0.0791168083042)*x_ref[1]**o+(-0.63228810248)*x_ref[1]
        ref[2,3,0,0]=(-0.828344342476)*x_ref[0]**o+(-0.781483429674)*x_ref[0]+(0.867903846758)*x_ref[1]**o+(0.569111225986)*x_ref[1]
        ref[2,3,0,1]=(0.659597328826)*x_ref[0]**o+(0.532419181635)*x_ref[0]+(-0.80620973473)*x_ref[1]**o+(-0.46673471071)*x_ref[1]
        ref[2,3,1,0]=(-0.885134823658)*x_ref[0]**o+(-0.546337219398)*x_ref[0]+(0.166804034744)*x_ref[1]**o+(0.299587086812)*x_ref[1]
        ref[2,3,1,1]=(0.112360533523)*x_ref[0]**o+(0.0953930412409)*x_ref[0]+(-0.97632552045)*x_ref[1]**o+(0.945659336453)*x_ref[1]
        ref[2,3,2,0]=(0.678405267904)*x_ref[0]**o+(0.342440777945)*x_ref[0]+(-0.99053148169)*x_ref[1]**o+(-0.969942468062)*x_ref[1]
        ref[2,3,2,1]=(-0.927867119592)*x_ref[0]**o+(-0.690642273601)*x_ref[0]+(-0.742887119506)*x_ref[1]**o+(-0.271373802799)*x_ref[1]
        ref[2,4,0,0]=(0.147456822243)*x_ref[0]**o+(-0.30678558608)*x_ref[0]+(0.824457828303)*x_ref[1]**o+(-0.236387246684)*x_ref[1]
        ref[2,4,0,1]=(-0.419906137988)*x_ref[0]**o+(0.940583417457)*x_ref[0]+(-0.252744709881)*x_ref[1]**o+(-0.311929952137)*x_ref[1]
        ref[2,4,1,0]=(0.676956544501)*x_ref[0]**o+(-0.0569121505607)*x_ref[0]+(0.518218768651)*x_ref[1]**o+(-0.14376234793)*x_ref[1]
        ref[2,4,1,1]=(-0.438909519823)*x_ref[0]**o+(0.363536809935)*x_ref[0]+(0.594959566494)*x_ref[1]**o+(0.161508700798)*x_ref[1]
        ref[2,4,2,0]=(-0.398965010329)*x_ref[0]**o+(-0.519311871529)*x_ref[0]+(0.0996092594307)*x_ref[1]**o+(0.299113884607)*x_ref[1]
        ref[2,4,2,1]=(0.438911683911)*x_ref[0]**o+(-0.330751949884)*x_ref[0]+(0.72396811226)*x_ref[1]**o+(-0.378618495259)*x_ref[1]
        ref[3,0,0,0]=(0.889235578987)*x_ref[0]**o+(0.301798885253)*x_ref[0]+(-0.876751409061)*x_ref[1]**o+(0.474669758163)*x_ref[1]
        ref[3,0,0,1]=(0.935815251238)*x_ref[0]**o+(0.674255198571)*x_ref[0]+(0.341766237248)*x_ref[1]**o+(-0.72357698277)*x_ref[1]
        ref[3,0,1,0]=(0.419855380231)*x_ref[0]**o+(0.755034058786)*x_ref[0]+(-0.59112491561)*x_ref[1]**o+(-0.288034233407)*x_ref[1]
        ref[3,0,1,1]=(-0.403840813699)*x_ref[0]**o+(0.770567211689)*x_ref[0]+(0.416460080616)*x_ref[1]**o+(-0.632767925192)*x_ref[1]
        ref[3,0,2,0]=(-0.847006017592)*x_ref[0]**o+(-0.447883355706)*x_ref[0]+(-0.628290176502)*x_ref[1]**o+(0.360720747081)*x_ref[1]
        ref[3,0,2,1]=(0.843924560726)*x_ref[0]**o+(-0.887772077659)*x_ref[0]+(-0.0951024444874)*x_ref[1]**o+(0.218823836893)*x_ref[1]
        ref[3,1,0,0]=(0.574007571036)*x_ref[0]**o+(-0.248963911474)*x_ref[0]+(-0.899931219543)*x_ref[1]**o+(0.710520579857)*x_ref[1]
        ref[3,1,0,1]=(-0.667806054096)*x_ref[0]**o+(-0.241865813708)*x_ref[0]+(0.588561916587)*x_ref[1]**o+(0.788227756226)*x_ref[1]
        ref[3,1,1,0]=(-0.258519497867)*x_ref[0]**o+(0.645611508516)*x_ref[0]+(-0.69368336683)*x_ref[1]**o+(0.217859823865)*x_ref[1]
        ref[3,1,1,1]=(-0.863926174373)*x_ref[0]**o+(-0.0852128898867)*x_ref[0]+(-0.648543494951)*x_ref[1]**o+(-0.402985545677)*x_ref[1]
        ref[3,1,2,0]=(-0.499342603145)*x_ref[0]**o+(0.10223769398)*x_ref[0]+(-0.968301649305)*x_ref[1]**o+(0.295206479399)*x_ref[1]
        ref[3,1,2,1]=(-0.190629994791)*x_ref[0]**o+(-0.795765651657)*x_ref[0]+(0.793726572154)*x_ref[1]**o+(-0.367897693585)*x_ref[1]
        ref[3,2,0,0]=(0.315102787334)*x_ref[0]**o+(0.21152066219)*x_ref[0]+(0.227000389892)*x_ref[1]**o+(-0.0796967599621)*x_ref[1]
        ref[3,2,0,1]=(0.826049810996)*x_ref[0]**o+(0.281856067318)*x_ref[0]+(0.433050065825)*x_ref[1]**o+(-0.218614730929)*x_ref[1]
        ref[3,2,1,0]=(0.369148754088)*x_ref[0]**o+(0.800521572309)*x_ref[0]+(0.877288542563)*x_ref[1]**o+(-0.135428392825)*x_ref[1]
        ref[3,2,1,1]=(0.808782390733)*x_ref[0]**o+(-0.326877566881)*x_ref[0]+(0.771288189422)*x_ref[1]**o+(-0.906951778691)*x_ref[1]
        ref[3,2,2,0]=(-0.381896054964)*x_ref[0]**o+(0.142176864318)*x_ref[0]+(-0.636114766341)*x_ref[1]**o+(-0.793416969698)*x_ref[1]
        ref[3,2,2,1]=(-0.291092484455)*x_ref[0]**o+(0.679789632507)*x_ref[0]+(0.882235235177)*x_ref[1]**o+(-0.826738171565)*x_ref[1]
        ref[3,3,0,0]=(0.71811796826)*x_ref[0]**o+(0.0627548627403)*x_ref[0]+(0.376656804533)*x_ref[1]**o+(-0.00934001961218)*x_ref[1]
        ref[3,3,0,1]=(-0.509178665563)*x_ref[0]**o+(0.804608293206)*x_ref[0]+(-0.752113912219)*x_ref[1]**o+(-0.313105572748)*x_ref[1]
        ref[3,3,1,0]=(-0.896783097316)*x_ref[0]**o+(-0.560162135134)*x_ref[0]+(-0.830841561786)*x_ref[1]**o+(-0.333506968769)*x_ref[1]
        ref[3,3,1,1]=(0.322497189516)*x_ref[0]**o+(0.277292905042)*x_ref[0]+(0.631980511231)*x_ref[1]**o+(0.926483120043)*x_ref[1]
        ref[3,3,2,0]=(-0.995133214294)*x_ref[0]**o+(-0.501556834675)*x_ref[0]+(-0.279959032716)*x_ref[1]**o+(-0.606903349874)*x_ref[1]
        ref[3,3,2,1]=(-0.52901549305)*x_ref[0]**o+(-0.345583335877)*x_ref[0]+(0.613327737258)*x_ref[1]**o+(-0.235057519501)*x_ref[1]
        ref[3,4,0,0]=(0.129978557871)*x_ref[0]**o+(-0.145655660564)*x_ref[0]+(-0.217775687907)*x_ref[1]**o+(0.528084545073)*x_ref[1]
        ref[3,4,0,1]=(0.94863938123)*x_ref[0]**o+(-0.858645154335)*x_ref[0]+(0.950889681616)*x_ref[1]**o+(-0.560518438352)*x_ref[1]
        ref[3,4,1,0]=(0.775184423382)*x_ref[0]**o+(0.783822082576)*x_ref[0]+(-0.130919265341)*x_ref[1]**o+(0.491447558856)*x_ref[1]
        ref[3,4,1,1]=(-0.383273309117)*x_ref[0]**o+(0.268662318873)*x_ref[0]+(0.0821612756798)*x_ref[1]**o+(0.508481517445)*x_ref[1]
        ref[3,4,2,0]=(-0.151657688401)*x_ref[0]**o+(0.661522859198)*x_ref[0]+(0.082765340123)*x_ref[1]**o+(-0.209987690619)*x_ref[1]
        ref[3,4,2,1]=(-0.137879400495)*x_ref[0]**o+(-0.291576763731)*x_ref[0]+(-0.342092940232)*x_ref[1]**o+(-0.486411344299)*x_ref[1]
      else:
        s[0,0,0,0]=(0.403430415343)*x[0]**o+(0.744986679903)*x[0]+(0.951777136716)*x[1]**o+(0.0324970558661)*x[1]+(0.427315603098)*x[2]**o+(0.791513954581)*x[2]
        s[0,0,0,1]=(-0.707145320014)*x[0]**o+(0.636836752091)*x[0]+(-0.7246243168)*x[1]**o+(-0.100083066676)*x[1]+(0.959094334502)*x[2]**o+(0.177082351017)*x[2]
        s[0,0,1,0]=(0.909576305075)*x[0]**o+(0.640954665378)*x[0]+(0.590627231018)*x[1]**o+(0.664146435649)*x[1]+(-0.870774905972)*x[2]**o+(0.113576453433)*x[2]
        s[0,0,1,1]=(0.45593637348)*x[0]**o+(-0.894975411641)*x[0]+(0.927784417885)*x[1]**o+(0.299259166361)*x[1]+(-0.287479963033)*x[2]**o+(-0.75427474283)*x[2]
        s[0,0,2,0]=(0.680005830031)*x[0]**o+(-0.65590690275)*x[0]+(-0.858769409879)*x[1]**o+(0.366256915436)*x[1]+(-0.73451062052)*x[2]**o+(0.70576385992)*x[2]
        s[0,0,2,1]=(0.590482155655)*x[0]**o+(0.921763825616)*x[0]+(-0.0202929641668)*x[1]**o+(0.574774440931)*x[1]+(-0.757342276378)*x[2]**o+(0.909153357188)*x[2]
        s[0,1,0,0]=(-0.919797976848)*x[0]**o+(-0.20563445882)*x[0]+(-0.221698859692)*x[1]**o+(0.65424513809)*x[1]+(-0.201817203504)*x[2]**o+(-0.77021064577)*x[2]
        s[0,1,0,1]=(-0.668449125613)*x[0]**o+(0.823706754981)*x[0]+(0.866963872133)*x[1]**o+(0.727752089501)*x[1]+(-0.130687213798)*x[2]**o+(-0.271094436825)*x[2]
        s[0,1,1,0]=(-0.595449075664)*x[0]**o+(0.505611411905)*x[0]+(0.522359351315)*x[1]**o+(-0.170169080167)*x[1]+(0.977079201782)*x[2]**o+(-0.404732823773)*x[2]
        s[0,1,1,1]=(0.788208619396)*x[0]**o+(0.857119112805)*x[0]+(-0.576911525936)*x[1]**o+(0.475890073035)*x[1]+(0.290398934513)*x[2]**o+(-0.242523666652)*x[2]
        s[0,1,2,0]=(0.385763599867)*x[0]**o+(0.941078949076)*x[0]+(-0.134687057713)*x[1]**o+(-0.999870523349)*x[1]+(0.343530780836)*x[2]**o+(-0.311388769086)*x[2]
        s[0,1,2,1]=(0.480578338846)*x[0]**o+(0.150470843838)*x[0]+(-0.33968928823)*x[1]**o+(-0.757189778135)*x[1]+(-0.808410519198)*x[2]**o+(0.00343659183409)*x[2]
        s[0,2,0,0]=(0.373686923589)*x[0]**o+(0.918031283521)*x[0]+(0.698320785442)*x[1]**o+(0.442416896262)*x[1]+(0.84105523068)*x[2]**o+(-0.202049171217)*x[2]
        s[0,2,0,1]=(0.867442299045)*x[0]**o+(0.714036317768)*x[0]+(0.0187944333819)*x[1]**o+(0.188685685012)*x[1]+(0.117805334047)*x[2]**o+(-0.105379365951)*x[2]
        s[0,2,1,0]=(-0.863116901688)*x[0]**o+(0.609059887913)*x[0]+(-0.017868762026)*x[1]**o+(-0.34217932778)*x[1]+(-0.0198171807922)*x[2]**o+(0.178138484275)*x[2]
        s[0,2,1,1]=(-0.209384449669)*x[0]**o+(0.264503737594)*x[0]+(-0.715310246173)*x[1]**o+(-0.529804181531)*x[1]+(-0.17584633955)*x[2]**o+(-0.515166239726)*x[2]
        s[0,2,2,0]=(0.913993550519)*x[0]**o+(-0.596906672047)*x[0]+(-0.182913064333)*x[1]**o+(0.0192817640562)*x[1]+(-0.931584175516)*x[2]**o+(0.333860908154)*x[2]
        s[0,2,2,1]=(-0.876954338835)*x[0]**o+(0.919399270359)*x[0]+(-0.0160062319874)*x[1]**o+(0.268471126695)*x[1]+(-0.508558598184)*x[2]**o+(-0.727898700083)*x[2]
        s[0,3,0,0]=(0.748569441709)*x[0]**o+(0.224242239087)*x[0]+(-0.776035603128)*x[1]**o+(0.536003741201)*x[1]+(0.438740206402)*x[2]**o+(0.94808528867)*x[2]
        s[0,3,0,1]=(0.898759703229)*x[0]**o+(0.40124773239)*x[0]+(-0.376410882714)*x[1]**o+(0.533837743659)*x[1]+(-0.0796343726864)*x[2]**o+(0.787983009824)*x[2]
        s[0,3,1,0]=(-0.833786092351)*x[0]**o+(-0.0186594726555)*x[0]+(0.805351464759)*x[1]**o+(0.273499432333)*x[1]+(-0.552330972949)*x[2]**o+(-0.87988075243)*x[2]
        s[0,3,1,1]=(0.944034634838)*x[0]**o+(0.481700802007)*x[0]+(-0.547568278381)*x[1]**o+(0.388929238306)*x[1]+(-0.894294015492)*x[2]**o+(-0.980420668029)*x[2]
        s[0,3,2,0]=(-0.865161590372)*x[0]**o+(0.671081690016)*x[0]+(-0.836586225032)*x[1]**o+(-0.480708191732)*x[1]+(-0.629454985205)*x[2]**o+(-0.183602424476)*x[2]
        s[0,3,2,1]=(0.244774306367)*x[0]**o+(-0.103110814177)*x[0]+(-0.389358087765)*x[1]**o+(-0.189491709415)*x[1]+(-0.2840972623)*x[2]**o+(-0.150115520815)*x[2]
        s[0,4,0,0]=(0.0679603146824)*x[0]**o+(0.804250679094)*x[0]+(-0.370590033619)*x[1]**o+(0.637308401931)*x[1]+(-0.941087411622)*x[2]**o+(-0.418093500864)*x[2]
        s[0,4,0,1]=(0.798443431935)*x[0]**o+(0.310544403291)*x[0]+(-0.253173498061)*x[1]**o+(-0.835625702512)*x[1]+(0.274190640271)*x[2]**o+(-0.0844258368882)*x[2]
        s[0,4,1,0]=(0.26366083654)*x[0]**o+(-0.525404638408)*x[0]+(0.177135519023)*x[1]**o+(0.679344298543)*x[1]+(0.491221604143)*x[2]**o+(-0.543997034477)*x[2]
        s[0,4,1,1]=(0.158539354649)*x[0]**o+(-0.40131214777)*x[0]+(-0.263013003032)*x[1]**o+(-0.835688710713)*x[1]+(-0.0507790155866)*x[2]**o+(0.310277829331)*x[2]
        s[0,4,2,0]=(0.0437752467637)*x[0]**o+(0.737258657262)*x[0]+(0.715865341928)*x[1]**o+(-0.330238427805)*x[1]+(0.360959461942)*x[2]**o+(0.416229562611)*x[2]
        s[0,4,2,1]=(0.432662559218)*x[0]**o+(-0.298420110661)*x[0]+(-0.596605166622)*x[1]**o+(0.841613034292)*x[1]+(-0.167017705859)*x[2]**o+(-0.17756125844)*x[2]
        s[1,0,0,0]=(-0.22874820236)*x[0]**o+(-0.381446359867)*x[0]+(0.0528518158695)*x[1]**o+(0.370749455302)*x[1]+(-0.985653513097)*x[2]**o+(0.844180132992)*x[2]
        s[1,0,0,1]=(0.373217484901)*x[0]**o+(0.144724114872)*x[0]+(-0.237896674313)*x[1]**o+(0.761700827804)*x[1]+(0.0524853501728)*x[2]**o+(0.136621357011)*x[2]
        s[1,0,1,0]=(-0.19687582146)*x[0]**o+(-0.431315907149)*x[0]+(-0.513904824755)*x[1]**o+(-0.971808827325)*x[1]+(-0.798897024204)*x[2]**o+(-0.845709083575)*x[2]
        s[1,0,1,1]=(0.65367003893)*x[0]**o+(-0.42776141089)*x[0]+(0.863689259035)*x[1]**o+(0.337607229707)*x[1]+(-0.733641128663)*x[2]**o+(0.299566554551)*x[2]
        s[1,0,2,0]=(0.713566633836)*x[0]**o+(-0.052250830321)*x[0]+(0.251265652235)*x[1]**o+(0.934301618588)*x[1]+(0.958676193949)*x[2]**o+(-0.904767643338)*x[2]
        s[1,0,2,1]=(0.0979309721699)*x[0]**o+(0.0508911031883)*x[0]+(-0.372782100123)*x[1]**o+(-0.858580732643)*x[1]+(0.558490969157)*x[2]**o+(0.901292190031)*x[2]
        s[1,1,0,0]=(-0.324020428418)*x[0]**o+(0.802490572711)*x[0]+(0.259862474757)*x[1]**o+(-0.937763031093)*x[1]+(0.738150549938)*x[2]**o+(0.710561531635)*x[2]
        s[1,1,0,1]=(0.169427696895)*x[0]**o+(0.123024711782)*x[0]+(-0.823267286726)*x[1]**o+(-0.257039853757)*x[1]+(-0.0869596277016)*x[2]**o+(0.823320253323)*x[2]
        s[1,1,1,0]=(0.563579538822)*x[0]**o+(0.886773654583)*x[0]+(-0.89146336006)*x[1]**o+(0.876812303998)*x[1]+(0.446982305387)*x[2]**o+(0.75768223621)*x[2]
        s[1,1,1,1]=(0.621430066484)*x[0]**o+(-0.60654124878)*x[0]+(0.997552582226)*x[1]**o+(-0.265298298595)*x[1]+(0.0724771779462)*x[2]**o+(-0.965477240471)*x[2]
        s[1,1,2,0]=(-0.226280272424)*x[0]**o+(0.60069948726)*x[0]+(0.607759942315)*x[1]**o+(-0.541368071975)*x[1]+(-0.255834836246)*x[2]**o+(0.53431491015)*x[2]
        s[1,1,2,1]=(-0.426027262724)*x[0]**o+(-0.458423430569)*x[0]+(0.457896921115)*x[1]**o+(0.903315740031)*x[1]+(-0.693306919014)*x[2]**o+(-0.711382904781)*x[2]
        s[1,2,0,0]=(0.772506053115)*x[0]**o+(-0.744777290755)*x[0]+(-0.666309688375)*x[1]**o+(-0.238296589057)*x[1]+(-0.914637614238)*x[2]**o+(-0.162549994528)*x[2]
        s[1,2,0,1]=(0.411613183464)*x[0]**o+(0.226948395694)*x[0]+(-0.501253778522)*x[1]**o+(0.643756868221)*x[1]+(0.695812686066)*x[2]**o+(0.945000689761)*x[2]
        s[1,2,1,0]=(-0.835518678598)*x[0]**o+(-0.0754832962638)*x[0]+(-0.365414270293)*x[1]**o+(-0.861470767362)*x[1]+(-0.409837921894)*x[2]**o+(-0.519708812476)*x[2]
        s[1,2,1,1]=(0.572469083271)*x[0]**o+(0.919850742056)*x[0]+(0.459717731039)*x[1]**o+(0.477016014609)*x[1]+(-0.0436975535208)*x[2]**o+(-0.474209977338)*x[2]
        s[1,2,2,0]=(0.326614458424)*x[0]**o+(0.359838636942)*x[0]+(-0.72907039396)*x[1]**o+(-0.855005539924)*x[1]+(0.260209906496)*x[2]**o+(0.60056833752)*x[2]
        s[1,2,2,1]=(0.311176883245)*x[0]**o+(0.432122987861)*x[0]+(0.535888800322)*x[1]**o+(0.108245072788)*x[1]+(0.139904790171)*x[2]**o+(-0.150799908363)*x[2]
        s[1,3,0,0]=(0.716614357434)*x[0]**o+(0.927927903373)*x[0]+(0.559668659399)*x[1]**o+(0.0939656439524)*x[1]+(0.761765352044)*x[2]**o+(-0.0416088943327)*x[2]
        s[1,3,0,1]=(0.38236332196)*x[0]**o+(-0.143078754628)*x[0]+(0.475444785511)*x[1]**o+(0.822323398805)*x[1]+(-0.660091618264)*x[2]**o+(0.342720814754)*x[2]
        s[1,3,1,0]=(-0.0840865615634)*x[0]**o+(0.171007008675)*x[0]+(0.103345219825)*x[1]**o+(0.573166199398)*x[1]+(-0.758853108442)*x[2]**o+(-0.86669302895)*x[2]
        s[1,3,1,1]=(-0.482667010514)*x[0]**o+(0.191120639148)*x[0]+(-0.606944850069)*x[1]**o+(-0.892213973892)*x[1]+(0.260456577775)*x[2]**o+(0.723746975714)*x[2]
        s[1,3,2,0]=(-0.517781948247)*x[0]**o+(-0.136449517335)*x[0]+(-0.8548184656)*x[1]**o+(0.221534125062)*x[1]+(-0.0312899368616)*x[2]**o+(0.853214998227)*x[2]
        s[1,3,2,1]=(0.659794051863)*x[0]**o+(-0.0558772903633)*x[0]+(-0.486705850738)*x[1]**o+(-0.946678425217)*x[1]+(-0.256832988748)*x[2]**o+(-0.735318810669)*x[2]
        s[1,4,0,0]=(0.145098093074)*x[0]**o+(-0.269302637649)*x[0]+(-0.982450704411)*x[1]**o+(-0.393217806322)*x[1]+(0.665598559355)*x[2]**o+(-0.82063852766)*x[2]
        s[1,4,0,1]=(-0.0316739172237)*x[0]**o+(-0.51840627642)*x[0]+(0.702751002426)*x[1]**o+(0.409172009664)*x[1]+(-0.665435255057)*x[2]**o+(-0.543340819924)*x[2]
        s[1,4,1,0]=(-0.655176869439)*x[0]**o+(0.767327923347)*x[0]+(0.21052751533)*x[1]**o+(-0.592499780096)*x[1]+(-0.404009516092)*x[2]**o+(0.337897620093)*x[2]
        s[1,4,1,1]=(0.0417921512129)*x[0]**o+(0.665949436737)*x[0]+(-0.94169952127)*x[1]**o+(-0.188551213715)*x[1]+(-0.728653294411)*x[2]**o+(-0.0105959354234)*x[2]
        s[1,4,2,0]=(0.965501168534)*x[0]**o+(-0.369191143342)*x[0]+(-0.0552020978431)*x[1]**o+(-0.428038061155)*x[1]+(-0.965958284463)*x[2]**o+(0.865323691614)*x[2]
        s[1,4,2,1]=(0.335937941327)*x[0]**o+(0.707755275673)*x[0]+(-0.213328047087)*x[1]**o+(0.40671856687)*x[1]+(-0.142473237329)*x[2]**o+(0.335602672685)*x[2]
        s[2,0,0,0]=(-0.543759088147)*x[0]**o+(-0.0578118152935)*x[0]+(-0.375211120232)*x[1]**o+(0.396642940459)*x[1]+(-0.827594349096)*x[2]**o+(-0.847196141531)*x[2]
        s[2,0,0,1]=(-0.640509948934)*x[0]**o+(-0.687311225614)*x[0]+(-0.727780018954)*x[1]**o+(0.730028842732)*x[1]+(-0.768410049039)*x[2]**o+(0.630097911191)*x[2]
        s[2,0,1,0]=(-0.237913733904)*x[0]**o+(0.678812377456)*x[0]+(0.744831381131)*x[1]**o+(0.234183216636)*x[1]+(0.607665520433)*x[2]**o+(0.0367432104088)*x[2]
        s[2,0,1,1]=(0.396457163479)*x[0]**o+(-0.0157749791786)*x[0]+(0.786041452538)*x[1]**o+(-0.413081468871)*x[1]+(0.530739739653)*x[2]**o+(0.614238047999)*x[2]
        s[2,0,2,0]=(0.102175067251)*x[0]**o+(0.610349372817)*x[0]+(0.697238281695)*x[1]**o+(-0.662368901962)*x[1]+(0.446990471604)*x[2]**o+(-0.144860323439)*x[2]
        s[2,0,2,1]=(-0.338454461333)*x[0]**o+(-0.582045241566)*x[0]+(-0.558490097333)*x[1]**o+(0.957514156098)*x[1]+(0.516681828614)*x[2]**o+(-0.03648670189)*x[2]
        s[2,1,0,0]=(-0.0956232886497)*x[0]**o+(0.296963706133)*x[0]+(-0.000278551018033)*x[1]**o+(0.211372560963)*x[1]+(0.373704134826)*x[2]**o+(-0.555747115303)*x[2]
        s[2,1,0,1]=(0.386383239127)*x[0]**o+(0.679111826958)*x[0]+(-0.70684681453)*x[1]**o+(0.760005656483)*x[1]+(0.913523573598)*x[2]**o+(0.236584510992)*x[2]
        s[2,1,1,0]=(0.960607356699)*x[0]**o+(-0.247980409436)*x[0]+(0.130760532932)*x[1]**o+(0.0471429101817)*x[1]+(0.885252361096)*x[2]**o+(-0.298931030118)*x[2]
        s[2,1,1,1]=(0.762861047844)*x[0]**o+(0.574433782034)*x[0]+(0.750887418107)*x[1]**o+(-0.505676440119)*x[1]+(0.0779737405854)*x[2]**o+(0.0416850170145)*x[2]
        s[2,1,2,0]=(-0.0935408632275)*x[0]**o+(0.693841928567)*x[0]+(-0.507446679719)*x[1]**o+(-0.595040059734)*x[1]+(-0.328667869314)*x[2]**o+(0.0930433314449)*x[2]
        s[2,1,2,1]=(0.159355451881)*x[0]**o+(0.698998810541)*x[0]+(-0.186951857037)*x[1]**o+(-0.555329009971)*x[1]+(0.464366123844)*x[2]**o+(-0.827667074026)*x[2]
        s[2,2,0,0]=(0.875743435714)*x[0]**o+(0.293157772376)*x[0]+(0.263232070113)*x[1]**o+(0.50022685176)*x[1]+(-0.838853959101)*x[2]**o+(0.952954656108)*x[2]
        s[2,2,0,1]=(0.847839606815)*x[0]**o+(0.417555361706)*x[0]+(0.855225402586)*x[1]**o+(0.701425285662)*x[1]+(0.408021284474)*x[2]**o+(-0.971501813458)*x[2]
        s[2,2,1,0]=(0.665514706889)*x[0]**o+(0.740492822423)*x[0]+(0.96298726932)*x[1]**o+(-0.237481520356)*x[1]+(-0.290560671331)*x[2]**o+(0.56838586659)*x[2]
        s[2,2,1,1]=(0.615075259737)*x[0]**o+(-0.89613745809)*x[0]+(0.0626934995142)*x[1]**o+(-0.637799053304)*x[1]+(0.0452470795742)*x[2]**o+(0.881307116917)*x[2]
        s[2,2,2,0]=(-0.786344805842)*x[0]**o+(-0.243568479393)*x[0]+(0.176867766715)*x[1]**o+(-0.578984774309)*x[1]+(-0.0915593214783)*x[2]**o+(0.872508270394)*x[2]
        s[2,2,2,1]=(-0.46735634172)*x[0]**o+(0.0363482739659)*x[0]+(-0.386717434419)*x[1]**o+(-0.783085654659)*x[1]+(0.949178326846)*x[2]**o+(0.371970006385)*x[2]
        s[2,3,0,0]=(-0.363466232325)*x[0]**o+(0.858040254812)*x[0]+(0.999192046284)*x[1]**o+(0.423646129134)*x[1]+(-0.030981663765)*x[2]**o+(-0.0671093802388)*x[2]
        s[2,3,0,1]=(-0.170266706257)*x[0]**o+(0.452981813541)*x[0]+(0.963234349731)*x[1]**o+(-0.0317092848063)*x[1]+(0.975127734624)*x[2]**o+(0.149233456408)*x[2]
        s[2,3,1,0]=(-0.279121247341)*x[0]**o+(0.74782301882)*x[0]+(0.40076226147)*x[1]**o+(-0.945530955089)*x[1]+(0.985569496972)*x[2]**o+(0.0378361605736)*x[2]
        s[2,3,1,1]=(-0.366292008093)*x[0]**o+(-0.364548368795)*x[0]+(-0.513038598699)*x[1]**o+(0.790842389439)*x[1]+(-0.139885010933)*x[2]**o+(-0.716083966108)*x[2]
        s[2,3,2,0]=(-0.960606339941)*x[0]**o+(-0.641708143226)*x[0]+(0.852467313167)*x[1]**o+(0.73409727377)*x[1]+(-0.0714287908193)*x[2]**o+(-0.888053934135)*x[2]
        s[2,3,2,1]=(-0.926276707175)*x[0]**o+(0.2576246735)*x[0]+(-0.490189832043)*x[1]**o+(-0.0410634355326)*x[1]+(0.0161969403712)*x[2]**o+(-0.188218508208)*x[2]
        s[2,4,0,0]=(-0.178315003213)*x[0]**o+(0.961625199265)*x[0]+(0.428091759741)*x[1]**o+(0.651624723931)*x[1]+(0.456933213658)*x[2]**o+(-0.405775680206)*x[2]
        s[2,4,0,1]=(0.907719303739)*x[0]**o+(0.0434247496698)*x[0]+(-0.282250394589)*x[1]**o+(-0.64568318107)*x[1]+(0.94652069508)*x[2]**o+(0.423506132915)*x[2]
        s[2,4,1,0]=(-0.737492191639)*x[0]**o+(-0.767903181959)*x[0]+(-0.0607159342444)*x[1]**o+(-0.0124107341616)*x[1]+(0.098984524763)*x[2]**o+(-0.246547982017)*x[2]
        s[2,4,1,1]=(-0.470182038449)*x[0]**o+(-0.608222886031)*x[0]+(-0.139904312627)*x[1]**o+(0.215173317211)*x[1]+(0.413902933771)*x[2]**o+(-0.650952803114)*x[2]
        s[2,4,2,0]=(-0.201212478305)*x[0]**o+(-0.222342535801)*x[0]+(0.299636591944)*x[1]**o+(0.167785412028)*x[1]+(-0.217953103956)*x[2]**o+(0.232079849598)*x[2]
        s[2,4,2,1]=(-0.0952620563909)*x[0]**o+(-0.43931419017)*x[0]+(0.482142211317)*x[1]**o+(-0.897201437534)*x[1]+(-0.159275604987)*x[2]**o+(-0.422577013901)*x[2]
        s[3,0,0,0]=(-0.863901307524)*x[0]**o+(-0.677273989905)*x[0]+(-0.551763007824)*x[1]**o+(0.660384831945)*x[1]+(0.728823623738)*x[2]**o+(0.364227655039)*x[2]
        s[3,0,0,1]=(0.410308900399)*x[0]**o+(-0.40573106086)*x[0]+(-0.104542078322)*x[1]**o+(0.269142743649)*x[1]+(-0.99125145064)*x[2]**o+(-0.758527358533)*x[2]
        s[3,0,1,0]=(0.592010233483)*x[0]**o+(0.229727871455)*x[0]+(0.867383811975)*x[1]**o+(-0.402872405597)*x[1]+(-0.105532874831)*x[2]**o+(0.051875826025)*x[2]
        s[3,0,1,1]=(-0.60342186941)*x[0]**o+(-0.536485143439)*x[0]+(0.563349948189)*x[1]**o+(-0.58710738551)*x[1]+(0.614134748504)*x[2]**o+(-0.374966138094)*x[2]
        s[3,0,2,0]=(-0.166713441956)*x[0]**o+(-0.148821382566)*x[0]+(0.243521410156)*x[1]**o+(-0.505696772795)*x[1]+(-0.0792469843575)*x[2]**o+(-0.254016378582)*x[2]
        s[3,0,2,1]=(0.841460450087)*x[0]**o+(0.677810255367)*x[0]+(0.194161032049)*x[1]**o+(0.242815225847)*x[1]+(-0.324804605495)*x[2]**o+(-0.587254716549)*x[2]
        s[3,1,0,0]=(-0.151432178911)*x[0]**o+(0.100422942614)*x[0]+(0.116301924918)*x[1]**o+(0.096254541101)*x[1]+(0.962400557131)*x[2]**o+(0.947799922492)*x[2]
        s[3,1,0,1]=(0.979488886801)*x[0]**o+(0.461069157066)*x[0]+(0.218226120996)*x[1]**o+(-0.800476906997)*x[1]+(0.880297021382)*x[2]**o+(0.19863665419)*x[2]
        s[3,1,1,0]=(0.229652753234)*x[0]**o+(-0.0478730795098)*x[0]+(0.82758343905)*x[1]**o+(-0.335216944553)*x[1]+(0.0465873906105)*x[2]**o+(0.561472522829)*x[2]
        s[3,1,1,1]=(0.669691480743)*x[0]**o+(0.920801400574)*x[0]+(0.687280123214)*x[1]**o+(-0.0891093965453)*x[1]+(-0.816955571861)*x[2]**o+(-0.613182601516)*x[2]
        s[3,1,2,0]=(-0.23526422348)*x[0]**o+(0.621205030541)*x[0]+(0.918118360964)*x[1]**o+(-0.889524467509)*x[1]+(-0.376827261814)*x[2]**o+(-0.661403722693)*x[2]
        s[3,1,2,1]=(-0.876978334114)*x[0]**o+(-0.411337422377)*x[0]+(-0.839251894886)*x[1]**o+(0.247435927004)*x[1]+(-0.0211212456768)*x[2]**o+(-0.217307469444)*x[2]
        s[3,2,0,0]=(-0.150749744623)*x[0]**o+(-0.433694637584)*x[0]+(0.998228805662)*x[1]**o+(-0.23851837551)*x[1]+(0.385865650695)*x[2]**o+(0.62171292869)*x[2]
        s[3,2,0,1]=(0.671159236132)*x[0]**o+(-0.759374252135)*x[0]+(0.947027836257)*x[1]**o+(0.993028655384)*x[1]+(-0.648286275)*x[2]**o+(0.245162986154)*x[2]
        s[3,2,1,0]=(0.560968532205)*x[0]**o+(0.233370818196)*x[0]+(0.898440249493)*x[1]**o+(-0.979069406244)*x[1]+(0.861103116467)*x[2]**o+(-0.448426992741)*x[2]
        s[3,2,1,1]=(-0.866958339062)*x[0]**o+(-0.93452597685)*x[0]+(-0.399823785608)*x[1]**o+(0.27169781451)*x[1]+(-0.467269546938)*x[2]**o+(-0.308295429989)*x[2]
        s[3,2,2,0]=(0.306858101448)*x[0]**o+(0.711533808817)*x[0]+(0.233577129267)*x[1]**o+(0.14002078557)*x[1]+(-0.110374227511)*x[2]**o+(-0.511365021523)*x[2]
        s[3,2,2,1]=(-0.885360569178)*x[0]**o+(0.686318360034)*x[0]+(-0.896320358304)*x[1]**o+(0.128087358671)*x[1]+(-0.353933621249)*x[2]**o+(0.467858811535)*x[2]
        s[3,3,0,0]=(0.852473040057)*x[0]**o+(0.905588366153)*x[0]+(0.264432751845)*x[1]**o+(-0.0639928837947)*x[1]+(-0.238384055445)*x[2]**o+(0.0742239369593)*x[2]
        s[3,3,0,1]=(0.419119777186)*x[0]**o+(-0.189773380938)*x[0]+(-0.304076026821)*x[1]**o+(0.343142155872)*x[1]+(-0.526644876689)*x[2]**o+(-0.449233806854)*x[2]
        s[3,3,1,0]=(-0.752488468101)*x[0]**o+(-0.583458526738)*x[0]+(0.431028528765)*x[1]**o+(0.540769571878)*x[1]+(0.667118059022)*x[2]**o+(-0.748999771002)*x[2]
        s[3,3,1,1]=(0.123191267212)*x[0]**o+(-0.109841090863)*x[0]+(0.621651033395)*x[1]**o+(-0.0614438487297)*x[1]+(-0.733103488283)*x[2]**o+(0.780271792879)*x[2]
        s[3,3,2,0]=(0.0663423783867)*x[0]**o+(-0.490704963664)*x[0]+(-0.461917954177)*x[1]**o+(-0.214475167777)*x[1]+(0.236974800111)*x[2]**o+(-0.506510367775)*x[2]
        s[3,3,2,1]=(0.713169472587)*x[0]**o+(0.133009068442)*x[0]+(-0.230236333051)*x[1]**o+(0.219280908939)*x[1]+(-0.568836956043)*x[2]**o+(0.669851689088)*x[2]
        s[3,4,0,0]=(-0.377299932399)*x[0]**o+(0.518102869754)*x[0]+(0.599212898665)*x[1]**o+(-0.500227698541)*x[1]+(-0.0454372641183)*x[2]**o+(0.339254001531)*x[2]
        s[3,4,0,1]=(-0.583931750822)*x[0]**o+(0.704497185439)*x[0]+(-0.683901595665)*x[1]**o+(0.278109096823)*x[1]+(-0.704371990659)*x[2]**o+(-0.990621051274)*x[2]
        s[3,4,1,0]=(-0.841302933963)*x[0]**o+(0.904679013637)*x[0]+(0.895802866749)*x[1]**o+(0.833316140426)*x[1]+(-0.234856502428)*x[2]**o+(-0.860024086409)*x[2]
        s[3,4,1,1]=(0.714165952226)*x[0]**o+(0.93387684839)*x[0]+(0.991668345861)*x[1]**o+(-0.438423497739)*x[1]+(-0.179296399411)*x[2]**o+(-0.0822020085158)*x[2]
        s[3,4,2,0]=(0.811219494783)*x[0]**o+(0.455443748484)*x[0]+(-0.737735168733)*x[1]**o+(0.258834303458)*x[1]+(0.345087059995)*x[2]**o+(0.892068047055)*x[2]
        s[3,4,2,1]=(0.776444289711)*x[0]**o+(-0.871162700545)*x[0]+(-0.785948122419)*x[1]**o+(0.42316871158)*x[1]+(0.18279624398)*x[2]**o+(-0.0234847339463)*x[2]
        ref[0,0,0,0]=(0.403430415343)*x_ref[0]**o+(0.744986679903)*x_ref[0]+(0.951777136716)*x_ref[1]**o+(0.0324970558661)*x_ref[1]+(0.427315603098)*x_ref[2]**o+(0.791513954581)*x_ref[2]
        ref[0,0,0,1]=(-0.707145320014)*x_ref[0]**o+(0.636836752091)*x_ref[0]+(-0.7246243168)*x_ref[1]**o+(-0.100083066676)*x_ref[1]+(0.959094334502)*x_ref[2]**o+(0.177082351017)*x_ref[2]
        ref[0,0,1,0]=(0.909576305075)*x_ref[0]**o+(0.640954665378)*x_ref[0]+(0.590627231018)*x_ref[1]**o+(0.664146435649)*x_ref[1]+(-0.870774905972)*x_ref[2]**o+(0.113576453433)*x_ref[2]
        ref[0,0,1,1]=(0.45593637348)*x_ref[0]**o+(-0.894975411641)*x_ref[0]+(0.927784417885)*x_ref[1]**o+(0.299259166361)*x_ref[1]+(-0.287479963033)*x_ref[2]**o+(-0.75427474283)*x_ref[2]
        ref[0,0,2,0]=(0.680005830031)*x_ref[0]**o+(-0.65590690275)*x_ref[0]+(-0.858769409879)*x_ref[1]**o+(0.366256915436)*x_ref[1]+(-0.73451062052)*x_ref[2]**o+(0.70576385992)*x_ref[2]
        ref[0,0,2,1]=(0.590482155655)*x_ref[0]**o+(0.921763825616)*x_ref[0]+(-0.0202929641668)*x_ref[1]**o+(0.574774440931)*x_ref[1]+(-0.757342276378)*x_ref[2]**o+(0.909153357188)*x_ref[2]
        ref[0,1,0,0]=(-0.919797976848)*x_ref[0]**o+(-0.20563445882)*x_ref[0]+(-0.221698859692)*x_ref[1]**o+(0.65424513809)*x_ref[1]+(-0.201817203504)*x_ref[2]**o+(-0.77021064577)*x_ref[2]
        ref[0,1,0,1]=(-0.668449125613)*x_ref[0]**o+(0.823706754981)*x_ref[0]+(0.866963872133)*x_ref[1]**o+(0.727752089501)*x_ref[1]+(-0.130687213798)*x_ref[2]**o+(-0.271094436825)*x_ref[2]
        ref[0,1,1,0]=(-0.595449075664)*x_ref[0]**o+(0.505611411905)*x_ref[0]+(0.522359351315)*x_ref[1]**o+(-0.170169080167)*x_ref[1]+(0.977079201782)*x_ref[2]**o+(-0.404732823773)*x_ref[2]
        ref[0,1,1,1]=(0.788208619396)*x_ref[0]**o+(0.857119112805)*x_ref[0]+(-0.576911525936)*x_ref[1]**o+(0.475890073035)*x_ref[1]+(0.290398934513)*x_ref[2]**o+(-0.242523666652)*x_ref[2]
        ref[0,1,2,0]=(0.385763599867)*x_ref[0]**o+(0.941078949076)*x_ref[0]+(-0.134687057713)*x_ref[1]**o+(-0.999870523349)*x_ref[1]+(0.343530780836)*x_ref[2]**o+(-0.311388769086)*x_ref[2]
        ref[0,1,2,1]=(0.480578338846)*x_ref[0]**o+(0.150470843838)*x_ref[0]+(-0.33968928823)*x_ref[1]**o+(-0.757189778135)*x_ref[1]+(-0.808410519198)*x_ref[2]**o+(0.00343659183409)*x_ref[2]
        ref[0,2,0,0]=(0.373686923589)*x_ref[0]**o+(0.918031283521)*x_ref[0]+(0.698320785442)*x_ref[1]**o+(0.442416896262)*x_ref[1]+(0.84105523068)*x_ref[2]**o+(-0.202049171217)*x_ref[2]
        ref[0,2,0,1]=(0.867442299045)*x_ref[0]**o+(0.714036317768)*x_ref[0]+(0.0187944333819)*x_ref[1]**o+(0.188685685012)*x_ref[1]+(0.117805334047)*x_ref[2]**o+(-0.105379365951)*x_ref[2]
        ref[0,2,1,0]=(-0.863116901688)*x_ref[0]**o+(0.609059887913)*x_ref[0]+(-0.017868762026)*x_ref[1]**o+(-0.34217932778)*x_ref[1]+(-0.0198171807922)*x_ref[2]**o+(0.178138484275)*x_ref[2]
        ref[0,2,1,1]=(-0.209384449669)*x_ref[0]**o+(0.264503737594)*x_ref[0]+(-0.715310246173)*x_ref[1]**o+(-0.529804181531)*x_ref[1]+(-0.17584633955)*x_ref[2]**o+(-0.515166239726)*x_ref[2]
        ref[0,2,2,0]=(0.913993550519)*x_ref[0]**o+(-0.596906672047)*x_ref[0]+(-0.182913064333)*x_ref[1]**o+(0.0192817640562)*x_ref[1]+(-0.931584175516)*x_ref[2]**o+(0.333860908154)*x_ref[2]
        ref[0,2,2,1]=(-0.876954338835)*x_ref[0]**o+(0.919399270359)*x_ref[0]+(-0.0160062319874)*x_ref[1]**o+(0.268471126695)*x_ref[1]+(-0.508558598184)*x_ref[2]**o+(-0.727898700083)*x_ref[2]
        ref[0,3,0,0]=(0.748569441709)*x_ref[0]**o+(0.224242239087)*x_ref[0]+(-0.776035603128)*x_ref[1]**o+(0.536003741201)*x_ref[1]+(0.438740206402)*x_ref[2]**o+(0.94808528867)*x_ref[2]
        ref[0,3,0,1]=(0.898759703229)*x_ref[0]**o+(0.40124773239)*x_ref[0]+(-0.376410882714)*x_ref[1]**o+(0.533837743659)*x_ref[1]+(-0.0796343726864)*x_ref[2]**o+(0.787983009824)*x_ref[2]
        ref[0,3,1,0]=(-0.833786092351)*x_ref[0]**o+(-0.0186594726555)*x_ref[0]+(0.805351464759)*x_ref[1]**o+(0.273499432333)*x_ref[1]+(-0.552330972949)*x_ref[2]**o+(-0.87988075243)*x_ref[2]
        ref[0,3,1,1]=(0.944034634838)*x_ref[0]**o+(0.481700802007)*x_ref[0]+(-0.547568278381)*x_ref[1]**o+(0.388929238306)*x_ref[1]+(-0.894294015492)*x_ref[2]**o+(-0.980420668029)*x_ref[2]
        ref[0,3,2,0]=(-0.865161590372)*x_ref[0]**o+(0.671081690016)*x_ref[0]+(-0.836586225032)*x_ref[1]**o+(-0.480708191732)*x_ref[1]+(-0.629454985205)*x_ref[2]**o+(-0.183602424476)*x_ref[2]
        ref[0,3,2,1]=(0.244774306367)*x_ref[0]**o+(-0.103110814177)*x_ref[0]+(-0.389358087765)*x_ref[1]**o+(-0.189491709415)*x_ref[1]+(-0.2840972623)*x_ref[2]**o+(-0.150115520815)*x_ref[2]
        ref[0,4,0,0]=(0.0679603146824)*x_ref[0]**o+(0.804250679094)*x_ref[0]+(-0.370590033619)*x_ref[1]**o+(0.637308401931)*x_ref[1]+(-0.941087411622)*x_ref[2]**o+(-0.418093500864)*x_ref[2]
        ref[0,4,0,1]=(0.798443431935)*x_ref[0]**o+(0.310544403291)*x_ref[0]+(-0.253173498061)*x_ref[1]**o+(-0.835625702512)*x_ref[1]+(0.274190640271)*x_ref[2]**o+(-0.0844258368882)*x_ref[2]
        ref[0,4,1,0]=(0.26366083654)*x_ref[0]**o+(-0.525404638408)*x_ref[0]+(0.177135519023)*x_ref[1]**o+(0.679344298543)*x_ref[1]+(0.491221604143)*x_ref[2]**o+(-0.543997034477)*x_ref[2]
        ref[0,4,1,1]=(0.158539354649)*x_ref[0]**o+(-0.40131214777)*x_ref[0]+(-0.263013003032)*x_ref[1]**o+(-0.835688710713)*x_ref[1]+(-0.0507790155866)*x_ref[2]**o+(0.310277829331)*x_ref[2]
        ref[0,4,2,0]=(0.0437752467637)*x_ref[0]**o+(0.737258657262)*x_ref[0]+(0.715865341928)*x_ref[1]**o+(-0.330238427805)*x_ref[1]+(0.360959461942)*x_ref[2]**o+(0.416229562611)*x_ref[2]
        ref[0,4,2,1]=(0.432662559218)*x_ref[0]**o+(-0.298420110661)*x_ref[0]+(-0.596605166622)*x_ref[1]**o+(0.841613034292)*x_ref[1]+(-0.167017705859)*x_ref[2]**o+(-0.17756125844)*x_ref[2]
        ref[1,0,0,0]=(-0.22874820236)*x_ref[0]**o+(-0.381446359867)*x_ref[0]+(0.0528518158695)*x_ref[1]**o+(0.370749455302)*x_ref[1]+(-0.985653513097)*x_ref[2]**o+(0.844180132992)*x_ref[2]
        ref[1,0,0,1]=(0.373217484901)*x_ref[0]**o+(0.144724114872)*x_ref[0]+(-0.237896674313)*x_ref[1]**o+(0.761700827804)*x_ref[1]+(0.0524853501728)*x_ref[2]**o+(0.136621357011)*x_ref[2]
        ref[1,0,1,0]=(-0.19687582146)*x_ref[0]**o+(-0.431315907149)*x_ref[0]+(-0.513904824755)*x_ref[1]**o+(-0.971808827325)*x_ref[1]+(-0.798897024204)*x_ref[2]**o+(-0.845709083575)*x_ref[2]
        ref[1,0,1,1]=(0.65367003893)*x_ref[0]**o+(-0.42776141089)*x_ref[0]+(0.863689259035)*x_ref[1]**o+(0.337607229707)*x_ref[1]+(-0.733641128663)*x_ref[2]**o+(0.299566554551)*x_ref[2]
        ref[1,0,2,0]=(0.713566633836)*x_ref[0]**o+(-0.052250830321)*x_ref[0]+(0.251265652235)*x_ref[1]**o+(0.934301618588)*x_ref[1]+(0.958676193949)*x_ref[2]**o+(-0.904767643338)*x_ref[2]
        ref[1,0,2,1]=(0.0979309721699)*x_ref[0]**o+(0.0508911031883)*x_ref[0]+(-0.372782100123)*x_ref[1]**o+(-0.858580732643)*x_ref[1]+(0.558490969157)*x_ref[2]**o+(0.901292190031)*x_ref[2]
        ref[1,1,0,0]=(-0.324020428418)*x_ref[0]**o+(0.802490572711)*x_ref[0]+(0.259862474757)*x_ref[1]**o+(-0.937763031093)*x_ref[1]+(0.738150549938)*x_ref[2]**o+(0.710561531635)*x_ref[2]
        ref[1,1,0,1]=(0.169427696895)*x_ref[0]**o+(0.123024711782)*x_ref[0]+(-0.823267286726)*x_ref[1]**o+(-0.257039853757)*x_ref[1]+(-0.0869596277016)*x_ref[2]**o+(0.823320253323)*x_ref[2]
        ref[1,1,1,0]=(0.563579538822)*x_ref[0]**o+(0.886773654583)*x_ref[0]+(-0.89146336006)*x_ref[1]**o+(0.876812303998)*x_ref[1]+(0.446982305387)*x_ref[2]**o+(0.75768223621)*x_ref[2]
        ref[1,1,1,1]=(0.621430066484)*x_ref[0]**o+(-0.60654124878)*x_ref[0]+(0.997552582226)*x_ref[1]**o+(-0.265298298595)*x_ref[1]+(0.0724771779462)*x_ref[2]**o+(-0.965477240471)*x_ref[2]
        ref[1,1,2,0]=(-0.226280272424)*x_ref[0]**o+(0.60069948726)*x_ref[0]+(0.607759942315)*x_ref[1]**o+(-0.541368071975)*x_ref[1]+(-0.255834836246)*x_ref[2]**o+(0.53431491015)*x_ref[2]
        ref[1,1,2,1]=(-0.426027262724)*x_ref[0]**o+(-0.458423430569)*x_ref[0]+(0.457896921115)*x_ref[1]**o+(0.903315740031)*x_ref[1]+(-0.693306919014)*x_ref[2]**o+(-0.711382904781)*x_ref[2]
        ref[1,2,0,0]=(0.772506053115)*x_ref[0]**o+(-0.744777290755)*x_ref[0]+(-0.666309688375)*x_ref[1]**o+(-0.238296589057)*x_ref[1]+(-0.914637614238)*x_ref[2]**o+(-0.162549994528)*x_ref[2]
        ref[1,2,0,1]=(0.411613183464)*x_ref[0]**o+(0.226948395694)*x_ref[0]+(-0.501253778522)*x_ref[1]**o+(0.643756868221)*x_ref[1]+(0.695812686066)*x_ref[2]**o+(0.945000689761)*x_ref[2]
        ref[1,2,1,0]=(-0.835518678598)*x_ref[0]**o+(-0.0754832962638)*x_ref[0]+(-0.365414270293)*x_ref[1]**o+(-0.861470767362)*x_ref[1]+(-0.409837921894)*x_ref[2]**o+(-0.519708812476)*x_ref[2]
        ref[1,2,1,1]=(0.572469083271)*x_ref[0]**o+(0.919850742056)*x_ref[0]+(0.459717731039)*x_ref[1]**o+(0.477016014609)*x_ref[1]+(-0.0436975535208)*x_ref[2]**o+(-0.474209977338)*x_ref[2]
        ref[1,2,2,0]=(0.326614458424)*x_ref[0]**o+(0.359838636942)*x_ref[0]+(-0.72907039396)*x_ref[1]**o+(-0.855005539924)*x_ref[1]+(0.260209906496)*x_ref[2]**o+(0.60056833752)*x_ref[2]
        ref[1,2,2,1]=(0.311176883245)*x_ref[0]**o+(0.432122987861)*x_ref[0]+(0.535888800322)*x_ref[1]**o+(0.108245072788)*x_ref[1]+(0.139904790171)*x_ref[2]**o+(-0.150799908363)*x_ref[2]
        ref[1,3,0,0]=(0.716614357434)*x_ref[0]**o+(0.927927903373)*x_ref[0]+(0.559668659399)*x_ref[1]**o+(0.0939656439524)*x_ref[1]+(0.761765352044)*x_ref[2]**o+(-0.0416088943327)*x_ref[2]
        ref[1,3,0,1]=(0.38236332196)*x_ref[0]**o+(-0.143078754628)*x_ref[0]+(0.475444785511)*x_ref[1]**o+(0.822323398805)*x_ref[1]+(-0.660091618264)*x_ref[2]**o+(0.342720814754)*x_ref[2]
        ref[1,3,1,0]=(-0.0840865615634)*x_ref[0]**o+(0.171007008675)*x_ref[0]+(0.103345219825)*x_ref[1]**o+(0.573166199398)*x_ref[1]+(-0.758853108442)*x_ref[2]**o+(-0.86669302895)*x_ref[2]
        ref[1,3,1,1]=(-0.482667010514)*x_ref[0]**o+(0.191120639148)*x_ref[0]+(-0.606944850069)*x_ref[1]**o+(-0.892213973892)*x_ref[1]+(0.260456577775)*x_ref[2]**o+(0.723746975714)*x_ref[2]
        ref[1,3,2,0]=(-0.517781948247)*x_ref[0]**o+(-0.136449517335)*x_ref[0]+(-0.8548184656)*x_ref[1]**o+(0.221534125062)*x_ref[1]+(-0.0312899368616)*x_ref[2]**o+(0.853214998227)*x_ref[2]
        ref[1,3,2,1]=(0.659794051863)*x_ref[0]**o+(-0.0558772903633)*x_ref[0]+(-0.486705850738)*x_ref[1]**o+(-0.946678425217)*x_ref[1]+(-0.256832988748)*x_ref[2]**o+(-0.735318810669)*x_ref[2]
        ref[1,4,0,0]=(0.145098093074)*x_ref[0]**o+(-0.269302637649)*x_ref[0]+(-0.982450704411)*x_ref[1]**o+(-0.393217806322)*x_ref[1]+(0.665598559355)*x_ref[2]**o+(-0.82063852766)*x_ref[2]
        ref[1,4,0,1]=(-0.0316739172237)*x_ref[0]**o+(-0.51840627642)*x_ref[0]+(0.702751002426)*x_ref[1]**o+(0.409172009664)*x_ref[1]+(-0.665435255057)*x_ref[2]**o+(-0.543340819924)*x_ref[2]
        ref[1,4,1,0]=(-0.655176869439)*x_ref[0]**o+(0.767327923347)*x_ref[0]+(0.21052751533)*x_ref[1]**o+(-0.592499780096)*x_ref[1]+(-0.404009516092)*x_ref[2]**o+(0.337897620093)*x_ref[2]
        ref[1,4,1,1]=(0.0417921512129)*x_ref[0]**o+(0.665949436737)*x_ref[0]+(-0.94169952127)*x_ref[1]**o+(-0.188551213715)*x_ref[1]+(-0.728653294411)*x_ref[2]**o+(-0.0105959354234)*x_ref[2]
        ref[1,4,2,0]=(0.965501168534)*x_ref[0]**o+(-0.369191143342)*x_ref[0]+(-0.0552020978431)*x_ref[1]**o+(-0.428038061155)*x_ref[1]+(-0.965958284463)*x_ref[2]**o+(0.865323691614)*x_ref[2]
        ref[1,4,2,1]=(0.335937941327)*x_ref[0]**o+(0.707755275673)*x_ref[0]+(-0.213328047087)*x_ref[1]**o+(0.40671856687)*x_ref[1]+(-0.142473237329)*x_ref[2]**o+(0.335602672685)*x_ref[2]
        ref[2,0,0,0]=(-0.543759088147)*x_ref[0]**o+(-0.0578118152935)*x_ref[0]+(-0.375211120232)*x_ref[1]**o+(0.396642940459)*x_ref[1]+(-0.827594349096)*x_ref[2]**o+(-0.847196141531)*x_ref[2]
        ref[2,0,0,1]=(-0.640509948934)*x_ref[0]**o+(-0.687311225614)*x_ref[0]+(-0.727780018954)*x_ref[1]**o+(0.730028842732)*x_ref[1]+(-0.768410049039)*x_ref[2]**o+(0.630097911191)*x_ref[2]
        ref[2,0,1,0]=(-0.237913733904)*x_ref[0]**o+(0.678812377456)*x_ref[0]+(0.744831381131)*x_ref[1]**o+(0.234183216636)*x_ref[1]+(0.607665520433)*x_ref[2]**o+(0.0367432104088)*x_ref[2]
        ref[2,0,1,1]=(0.396457163479)*x_ref[0]**o+(-0.0157749791786)*x_ref[0]+(0.786041452538)*x_ref[1]**o+(-0.413081468871)*x_ref[1]+(0.530739739653)*x_ref[2]**o+(0.614238047999)*x_ref[2]
        ref[2,0,2,0]=(0.102175067251)*x_ref[0]**o+(0.610349372817)*x_ref[0]+(0.697238281695)*x_ref[1]**o+(-0.662368901962)*x_ref[1]+(0.446990471604)*x_ref[2]**o+(-0.144860323439)*x_ref[2]
        ref[2,0,2,1]=(-0.338454461333)*x_ref[0]**o+(-0.582045241566)*x_ref[0]+(-0.558490097333)*x_ref[1]**o+(0.957514156098)*x_ref[1]+(0.516681828614)*x_ref[2]**o+(-0.03648670189)*x_ref[2]
        ref[2,1,0,0]=(-0.0956232886497)*x_ref[0]**o+(0.296963706133)*x_ref[0]+(-0.000278551018033)*x_ref[1]**o+(0.211372560963)*x_ref[1]+(0.373704134826)*x_ref[2]**o+(-0.555747115303)*x_ref[2]
        ref[2,1,0,1]=(0.386383239127)*x_ref[0]**o+(0.679111826958)*x_ref[0]+(-0.70684681453)*x_ref[1]**o+(0.760005656483)*x_ref[1]+(0.913523573598)*x_ref[2]**o+(0.236584510992)*x_ref[2]
        ref[2,1,1,0]=(0.960607356699)*x_ref[0]**o+(-0.247980409436)*x_ref[0]+(0.130760532932)*x_ref[1]**o+(0.0471429101817)*x_ref[1]+(0.885252361096)*x_ref[2]**o+(-0.298931030118)*x_ref[2]
        ref[2,1,1,1]=(0.762861047844)*x_ref[0]**o+(0.574433782034)*x_ref[0]+(0.750887418107)*x_ref[1]**o+(-0.505676440119)*x_ref[1]+(0.0779737405854)*x_ref[2]**o+(0.0416850170145)*x_ref[2]
        ref[2,1,2,0]=(-0.0935408632275)*x_ref[0]**o+(0.693841928567)*x_ref[0]+(-0.507446679719)*x_ref[1]**o+(-0.595040059734)*x_ref[1]+(-0.328667869314)*x_ref[2]**o+(0.0930433314449)*x_ref[2]
        ref[2,1,2,1]=(0.159355451881)*x_ref[0]**o+(0.698998810541)*x_ref[0]+(-0.186951857037)*x_ref[1]**o+(-0.555329009971)*x_ref[1]+(0.464366123844)*x_ref[2]**o+(-0.827667074026)*x_ref[2]
        ref[2,2,0,0]=(0.875743435714)*x_ref[0]**o+(0.293157772376)*x_ref[0]+(0.263232070113)*x_ref[1]**o+(0.50022685176)*x_ref[1]+(-0.838853959101)*x_ref[2]**o+(0.952954656108)*x_ref[2]
        ref[2,2,0,1]=(0.847839606815)*x_ref[0]**o+(0.417555361706)*x_ref[0]+(0.855225402586)*x_ref[1]**o+(0.701425285662)*x_ref[1]+(0.408021284474)*x_ref[2]**o+(-0.971501813458)*x_ref[2]
        ref[2,2,1,0]=(0.665514706889)*x_ref[0]**o+(0.740492822423)*x_ref[0]+(0.96298726932)*x_ref[1]**o+(-0.237481520356)*x_ref[1]+(-0.290560671331)*x_ref[2]**o+(0.56838586659)*x_ref[2]
        ref[2,2,1,1]=(0.615075259737)*x_ref[0]**o+(-0.89613745809)*x_ref[0]+(0.0626934995142)*x_ref[1]**o+(-0.637799053304)*x_ref[1]+(0.0452470795742)*x_ref[2]**o+(0.881307116917)*x_ref[2]
        ref[2,2,2,0]=(-0.786344805842)*x_ref[0]**o+(-0.243568479393)*x_ref[0]+(0.176867766715)*x_ref[1]**o+(-0.578984774309)*x_ref[1]+(-0.0915593214783)*x_ref[2]**o+(0.872508270394)*x_ref[2]
        ref[2,2,2,1]=(-0.46735634172)*x_ref[0]**o+(0.0363482739659)*x_ref[0]+(-0.386717434419)*x_ref[1]**o+(-0.783085654659)*x_ref[1]+(0.949178326846)*x_ref[2]**o+(0.371970006385)*x_ref[2]
        ref[2,3,0,0]=(-0.363466232325)*x_ref[0]**o+(0.858040254812)*x_ref[0]+(0.999192046284)*x_ref[1]**o+(0.423646129134)*x_ref[1]+(-0.030981663765)*x_ref[2]**o+(-0.0671093802388)*x_ref[2]
        ref[2,3,0,1]=(-0.170266706257)*x_ref[0]**o+(0.452981813541)*x_ref[0]+(0.963234349731)*x_ref[1]**o+(-0.0317092848063)*x_ref[1]+(0.975127734624)*x_ref[2]**o+(0.149233456408)*x_ref[2]
        ref[2,3,1,0]=(-0.279121247341)*x_ref[0]**o+(0.74782301882)*x_ref[0]+(0.40076226147)*x_ref[1]**o+(-0.945530955089)*x_ref[1]+(0.985569496972)*x_ref[2]**o+(0.0378361605736)*x_ref[2]
        ref[2,3,1,1]=(-0.366292008093)*x_ref[0]**o+(-0.364548368795)*x_ref[0]+(-0.513038598699)*x_ref[1]**o+(0.790842389439)*x_ref[1]+(-0.139885010933)*x_ref[2]**o+(-0.716083966108)*x_ref[2]
        ref[2,3,2,0]=(-0.960606339941)*x_ref[0]**o+(-0.641708143226)*x_ref[0]+(0.852467313167)*x_ref[1]**o+(0.73409727377)*x_ref[1]+(-0.0714287908193)*x_ref[2]**o+(-0.888053934135)*x_ref[2]
        ref[2,3,2,1]=(-0.926276707175)*x_ref[0]**o+(0.2576246735)*x_ref[0]+(-0.490189832043)*x_ref[1]**o+(-0.0410634355326)*x_ref[1]+(0.0161969403712)*x_ref[2]**o+(-0.188218508208)*x_ref[2]
        ref[2,4,0,0]=(-0.178315003213)*x_ref[0]**o+(0.961625199265)*x_ref[0]+(0.428091759741)*x_ref[1]**o+(0.651624723931)*x_ref[1]+(0.456933213658)*x_ref[2]**o+(-0.405775680206)*x_ref[2]
        ref[2,4,0,1]=(0.907719303739)*x_ref[0]**o+(0.0434247496698)*x_ref[0]+(-0.282250394589)*x_ref[1]**o+(-0.64568318107)*x_ref[1]+(0.94652069508)*x_ref[2]**o+(0.423506132915)*x_ref[2]
        ref[2,4,1,0]=(-0.737492191639)*x_ref[0]**o+(-0.767903181959)*x_ref[0]+(-0.0607159342444)*x_ref[1]**o+(-0.0124107341616)*x_ref[1]+(0.098984524763)*x_ref[2]**o+(-0.246547982017)*x_ref[2]
        ref[2,4,1,1]=(-0.470182038449)*x_ref[0]**o+(-0.608222886031)*x_ref[0]+(-0.139904312627)*x_ref[1]**o+(0.215173317211)*x_ref[1]+(0.413902933771)*x_ref[2]**o+(-0.650952803114)*x_ref[2]
        ref[2,4,2,0]=(-0.201212478305)*x_ref[0]**o+(-0.222342535801)*x_ref[0]+(0.299636591944)*x_ref[1]**o+(0.167785412028)*x_ref[1]+(-0.217953103956)*x_ref[2]**o+(0.232079849598)*x_ref[2]
        ref[2,4,2,1]=(-0.0952620563909)*x_ref[0]**o+(-0.43931419017)*x_ref[0]+(0.482142211317)*x_ref[1]**o+(-0.897201437534)*x_ref[1]+(-0.159275604987)*x_ref[2]**o+(-0.422577013901)*x_ref[2]
        ref[3,0,0,0]=(-0.863901307524)*x_ref[0]**o+(-0.677273989905)*x_ref[0]+(-0.551763007824)*x_ref[1]**o+(0.660384831945)*x_ref[1]+(0.728823623738)*x_ref[2]**o+(0.364227655039)*x_ref[2]
        ref[3,0,0,1]=(0.410308900399)*x_ref[0]**o+(-0.40573106086)*x_ref[0]+(-0.104542078322)*x_ref[1]**o+(0.269142743649)*x_ref[1]+(-0.99125145064)*x_ref[2]**o+(-0.758527358533)*x_ref[2]
        ref[3,0,1,0]=(0.592010233483)*x_ref[0]**o+(0.229727871455)*x_ref[0]+(0.867383811975)*x_ref[1]**o+(-0.402872405597)*x_ref[1]+(-0.105532874831)*x_ref[2]**o+(0.051875826025)*x_ref[2]
        ref[3,0,1,1]=(-0.60342186941)*x_ref[0]**o+(-0.536485143439)*x_ref[0]+(0.563349948189)*x_ref[1]**o+(-0.58710738551)*x_ref[1]+(0.614134748504)*x_ref[2]**o+(-0.374966138094)*x_ref[2]
        ref[3,0,2,0]=(-0.166713441956)*x_ref[0]**o+(-0.148821382566)*x_ref[0]+(0.243521410156)*x_ref[1]**o+(-0.505696772795)*x_ref[1]+(-0.0792469843575)*x_ref[2]**o+(-0.254016378582)*x_ref[2]
        ref[3,0,2,1]=(0.841460450087)*x_ref[0]**o+(0.677810255367)*x_ref[0]+(0.194161032049)*x_ref[1]**o+(0.242815225847)*x_ref[1]+(-0.324804605495)*x_ref[2]**o+(-0.587254716549)*x_ref[2]
        ref[3,1,0,0]=(-0.151432178911)*x_ref[0]**o+(0.100422942614)*x_ref[0]+(0.116301924918)*x_ref[1]**o+(0.096254541101)*x_ref[1]+(0.962400557131)*x_ref[2]**o+(0.947799922492)*x_ref[2]
        ref[3,1,0,1]=(0.979488886801)*x_ref[0]**o+(0.461069157066)*x_ref[0]+(0.218226120996)*x_ref[1]**o+(-0.800476906997)*x_ref[1]+(0.880297021382)*x_ref[2]**o+(0.19863665419)*x_ref[2]
        ref[3,1,1,0]=(0.229652753234)*x_ref[0]**o+(-0.0478730795098)*x_ref[0]+(0.82758343905)*x_ref[1]**o+(-0.335216944553)*x_ref[1]+(0.0465873906105)*x_ref[2]**o+(0.561472522829)*x_ref[2]
        ref[3,1,1,1]=(0.669691480743)*x_ref[0]**o+(0.920801400574)*x_ref[0]+(0.687280123214)*x_ref[1]**o+(-0.0891093965453)*x_ref[1]+(-0.816955571861)*x_ref[2]**o+(-0.613182601516)*x_ref[2]
        ref[3,1,2,0]=(-0.23526422348)*x_ref[0]**o+(0.621205030541)*x_ref[0]+(0.918118360964)*x_ref[1]**o+(-0.889524467509)*x_ref[1]+(-0.376827261814)*x_ref[2]**o+(-0.661403722693)*x_ref[2]
        ref[3,1,2,1]=(-0.876978334114)*x_ref[0]**o+(-0.411337422377)*x_ref[0]+(-0.839251894886)*x_ref[1]**o+(0.247435927004)*x_ref[1]+(-0.0211212456768)*x_ref[2]**o+(-0.217307469444)*x_ref[2]
        ref[3,2,0,0]=(-0.150749744623)*x_ref[0]**o+(-0.433694637584)*x_ref[0]+(0.998228805662)*x_ref[1]**o+(-0.23851837551)*x_ref[1]+(0.385865650695)*x_ref[2]**o+(0.62171292869)*x_ref[2]
        ref[3,2,0,1]=(0.671159236132)*x_ref[0]**o+(-0.759374252135)*x_ref[0]+(0.947027836257)*x_ref[1]**o+(0.993028655384)*x_ref[1]+(-0.648286275)*x_ref[2]**o+(0.245162986154)*x_ref[2]
        ref[3,2,1,0]=(0.560968532205)*x_ref[0]**o+(0.233370818196)*x_ref[0]+(0.898440249493)*x_ref[1]**o+(-0.979069406244)*x_ref[1]+(0.861103116467)*x_ref[2]**o+(-0.448426992741)*x_ref[2]
        ref[3,2,1,1]=(-0.866958339062)*x_ref[0]**o+(-0.93452597685)*x_ref[0]+(-0.399823785608)*x_ref[1]**o+(0.27169781451)*x_ref[1]+(-0.467269546938)*x_ref[2]**o+(-0.308295429989)*x_ref[2]
        ref[3,2,2,0]=(0.306858101448)*x_ref[0]**o+(0.711533808817)*x_ref[0]+(0.233577129267)*x_ref[1]**o+(0.14002078557)*x_ref[1]+(-0.110374227511)*x_ref[2]**o+(-0.511365021523)*x_ref[2]
        ref[3,2,2,1]=(-0.885360569178)*x_ref[0]**o+(0.686318360034)*x_ref[0]+(-0.896320358304)*x_ref[1]**o+(0.128087358671)*x_ref[1]+(-0.353933621249)*x_ref[2]**o+(0.467858811535)*x_ref[2]
        ref[3,3,0,0]=(0.852473040057)*x_ref[0]**o+(0.905588366153)*x_ref[0]+(0.264432751845)*x_ref[1]**o+(-0.0639928837947)*x_ref[1]+(-0.238384055445)*x_ref[2]**o+(0.0742239369593)*x_ref[2]
        ref[3,3,0,1]=(0.419119777186)*x_ref[0]**o+(-0.189773380938)*x_ref[0]+(-0.304076026821)*x_ref[1]**o+(0.343142155872)*x_ref[1]+(-0.526644876689)*x_ref[2]**o+(-0.449233806854)*x_ref[2]
        ref[3,3,1,0]=(-0.752488468101)*x_ref[0]**o+(-0.583458526738)*x_ref[0]+(0.431028528765)*x_ref[1]**o+(0.540769571878)*x_ref[1]+(0.667118059022)*x_ref[2]**o+(-0.748999771002)*x_ref[2]
        ref[3,3,1,1]=(0.123191267212)*x_ref[0]**o+(-0.109841090863)*x_ref[0]+(0.621651033395)*x_ref[1]**o+(-0.0614438487297)*x_ref[1]+(-0.733103488283)*x_ref[2]**o+(0.780271792879)*x_ref[2]
        ref[3,3,2,0]=(0.0663423783867)*x_ref[0]**o+(-0.490704963664)*x_ref[0]+(-0.461917954177)*x_ref[1]**o+(-0.214475167777)*x_ref[1]+(0.236974800111)*x_ref[2]**o+(-0.506510367775)*x_ref[2]
        ref[3,3,2,1]=(0.713169472587)*x_ref[0]**o+(0.133009068442)*x_ref[0]+(-0.230236333051)*x_ref[1]**o+(0.219280908939)*x_ref[1]+(-0.568836956043)*x_ref[2]**o+(0.669851689088)*x_ref[2]
        ref[3,4,0,0]=(-0.377299932399)*x_ref[0]**o+(0.518102869754)*x_ref[0]+(0.599212898665)*x_ref[1]**o+(-0.500227698541)*x_ref[1]+(-0.0454372641183)*x_ref[2]**o+(0.339254001531)*x_ref[2]
        ref[3,4,0,1]=(-0.583931750822)*x_ref[0]**o+(0.704497185439)*x_ref[0]+(-0.683901595665)*x_ref[1]**o+(0.278109096823)*x_ref[1]+(-0.704371990659)*x_ref[2]**o+(-0.990621051274)*x_ref[2]
        ref[3,4,1,0]=(-0.841302933963)*x_ref[0]**o+(0.904679013637)*x_ref[0]+(0.895802866749)*x_ref[1]**o+(0.833316140426)*x_ref[1]+(-0.234856502428)*x_ref[2]**o+(-0.860024086409)*x_ref[2]
        ref[3,4,1,1]=(0.714165952226)*x_ref[0]**o+(0.93387684839)*x_ref[0]+(0.991668345861)*x_ref[1]**o+(-0.438423497739)*x_ref[1]+(-0.179296399411)*x_ref[2]**o+(-0.0822020085158)*x_ref[2]
        ref[3,4,2,0]=(0.811219494783)*x_ref[0]**o+(0.455443748484)*x_ref[0]+(-0.737735168733)*x_ref[1]**o+(0.258834303458)*x_ref[1]+(0.345087059995)*x_ref[2]**o+(0.892068047055)*x_ref[2]
        ref[3,4,2,1]=(0.776444289711)*x_ref[0]**o+(-0.871162700545)*x_ref[0]+(-0.785948122419)*x_ref[1]**o+(0.42316871158)*x_ref[1]+(0.18279624398)*x_ref[2]**o+(-0.0234847339463)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-1.02854286859)*x[0]+(0.105922853194)*x[1]
        ref=(-1.02854286859)*x_ref[0]+(0.105922853194)*x_ref[1]
      else:
        arg=(1.36620376771)*x[0]+(-0.336798536224)*x[1]+(0.534538499605)*x[2]
        ref=(1.36620376771)*x_ref[0]+(-0.336798536224)*x_ref[1]+(0.534538499605)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        arg[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
        ref[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        ref[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
      else:
        arg[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        arg[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
        ref[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        ref[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.452635897444)*x[0]+(-0.210513646405)*x[1]
        arg[0,1]=(-0.590704436627)*x[0]+(1.78000273884)*x[1]
        arg[0,2]=(1.01331349402)*x[0]+(1.76279765125)*x[1]
        arg[0,3]=(1.5588852701)*x[0]+(-0.550895238707)*x[1]
        arg[0,4]=(-0.352453804527)*x[0]+(-0.338832613421)*x[1]
        arg[1,0]=(-0.809837877163)*x[0]+(-0.531009602702)*x[1]
        arg[1,1]=(0.513641503665)*x[0]+(0.232114989082)*x[1]
        arg[1,2]=(-0.424635655997)*x[0]+(0.0796757576821)*x[1]
        arg[1,3]=(-1.5899938122)*x[0]+(-0.178949548482)*x[1]
        arg[1,4]=(0.989317971495)*x[0]+(0.969711267468)*x[1]
        arg[2,0]=(0.384490947785)*x[0]+(1.42230193156)*x[1]
        arg[2,1]=(-0.0832227929255)*x[0]+(0.859020891429)*x[1]
        arg[2,2]=(-1.36129563124)*x[0]+(-0.240985031888)*x[1]
        arg[2,3]=(0.319706901063)*x[0]+(1.03419924566)*x[1]
        arg[2,4]=(0.986965249719)*x[0]+(0.228684031472)*x[1]
        arg[3,0]=(0.588520127337)*x[0]+(0.016799594822)*x[1]
        arg[3,1]=(1.46193438322)*x[0]+(0.431838900383)*x[1]
        arg[3,2]=(-0.0235512206818)*x[0]+(0.00227038037474)*x[1]
        arg[3,3]=(0.977383733076)*x[0]+(0.383979958766)*x[1]
        arg[3,4]=(0.447782383836)*x[0]+(-0.504218231709)*x[1]
        ref[0,0]=(-0.452635897444)*x_ref[0]+(-0.210513646405)*x_ref[1]
        ref[0,1]=(-0.590704436627)*x_ref[0]+(1.78000273884)*x_ref[1]
        ref[0,2]=(1.01331349402)*x_ref[0]+(1.76279765125)*x_ref[1]
        ref[0,3]=(1.5588852701)*x_ref[0]+(-0.550895238707)*x_ref[1]
        ref[0,4]=(-0.352453804527)*x_ref[0]+(-0.338832613421)*x_ref[1]
        ref[1,0]=(-0.809837877163)*x_ref[0]+(-0.531009602702)*x_ref[1]
        ref[1,1]=(0.513641503665)*x_ref[0]+(0.232114989082)*x_ref[1]
        ref[1,2]=(-0.424635655997)*x_ref[0]+(0.0796757576821)*x_ref[1]
        ref[1,3]=(-1.5899938122)*x_ref[0]+(-0.178949548482)*x_ref[1]
        ref[1,4]=(0.989317971495)*x_ref[0]+(0.969711267468)*x_ref[1]
        ref[2,0]=(0.384490947785)*x_ref[0]+(1.42230193156)*x_ref[1]
        ref[2,1]=(-0.0832227929255)*x_ref[0]+(0.859020891429)*x_ref[1]
        ref[2,2]=(-1.36129563124)*x_ref[0]+(-0.240985031888)*x_ref[1]
        ref[2,3]=(0.319706901063)*x_ref[0]+(1.03419924566)*x_ref[1]
        ref[2,4]=(0.986965249719)*x_ref[0]+(0.228684031472)*x_ref[1]
        ref[3,0]=(0.588520127337)*x_ref[0]+(0.016799594822)*x_ref[1]
        ref[3,1]=(1.46193438322)*x_ref[0]+(0.431838900383)*x_ref[1]
        ref[3,2]=(-0.0235512206818)*x_ref[0]+(0.00227038037474)*x_ref[1]
        ref[3,3]=(0.977383733076)*x_ref[0]+(0.383979958766)*x_ref[1]
        ref[3,4]=(0.447782383836)*x_ref[0]+(-0.504218231709)*x_ref[1]
      else:
        arg[0,0]=(-0.566459971675)*x[0]+(0.240246648893)*x[1]+(1.84693181923)*x[2]
        arg[0,1]=(1.09715015899)*x[0]+(1.26531924413)*x[1]+(0.627813347679)*x[2]
        arg[0,2]=(-1.17535073843)*x[0]+(-0.215601498941)*x[1]+(1.36346500888)*x[2]
        arg[0,3]=(-0.637207120161)*x[0]+(1.08542566952)*x[1]+(-1.13757514425)*x[2]
        arg[0,4]=(1.18247424069)*x[0]+(-1.48951614815)*x[1]+(0.615535039719)*x[2]
        arg[1,0]=(0.0937432100426)*x[0]+(-0.510326894156)*x[1]+(-0.456208996457)*x[2]
        arg[1,1]=(-0.866120342442)*x[0]+(0.927484984985)*x[1]+(0.918081524156)*x[2]
        arg[1,2]=(0.025238374235)*x[0]+(-0.182938599269)*x[1]+(-1.50048281322)*x[2]
        arg[1,3]=(-0.302334971294)*x[0]+(-0.925062799423)*x[1]+(-1.4675138953)*x[2]
        arg[1,4]=(-0.497738566447)*x[0]+(-0.324727071875)*x[1]+(-1.43341212055)*x[2]
        arg[2,0]=(-1.4605237958)*x[0]+(-0.332826007088)*x[1]+(-0.230483401111)*x[2]
        arg[2,1]=(0.801769248581)*x[0]+(-0.930588036503)*x[1]+(-1.06804808381)*x[2]
        arg[2,2]=(-1.07972165017)*x[0]+(1.09295362882)*x[1]+(0.048550098477)*x[2]
        arg[2,3]=(0.028290230337)*x[0]+(0.179461246635)*x[1]+(-1.09165257072)*x[2]
        arg[2,4]=(0.238850986174)*x[0]+(-0.837307206184)*x[1]+(-0.617254874594)*x[2]
        arg[3,0]=(-0.125921204582)*x[0]+(0.234758657879)*x[1]+(0.0847162247843)*x[2]
        arg[3,1]=(-0.213352414907)*x[0]+(0.206777352455)*x[1]+(-0.135357638131)*x[2]
        arg[3,2]=(-0.41897394434)*x[0]+(0.152706100016)*x[1]+(1.52390356012)*x[2]
        arg[3,3]=(0.1831705248)*x[0]+(-1.13556841364)*x[1]+(-1.7661652693)*x[2]
        arg[3,4]=(0.43473708388)*x[0]+(1.04496896213)*x[1]+(-0.816894739325)*x[2]
        ref[0,0]=(-0.566459971675)*x_ref[0]+(0.240246648893)*x_ref[1]+(1.84693181923)*x_ref[2]
        ref[0,1]=(1.09715015899)*x_ref[0]+(1.26531924413)*x_ref[1]+(0.627813347679)*x_ref[2]
        ref[0,2]=(-1.17535073843)*x_ref[0]+(-0.215601498941)*x_ref[1]+(1.36346500888)*x_ref[2]
        ref[0,3]=(-0.637207120161)*x_ref[0]+(1.08542566952)*x_ref[1]+(-1.13757514425)*x_ref[2]
        ref[0,4]=(1.18247424069)*x_ref[0]+(-1.48951614815)*x_ref[1]+(0.615535039719)*x_ref[2]
        ref[1,0]=(0.0937432100426)*x_ref[0]+(-0.510326894156)*x_ref[1]+(-0.456208996457)*x_ref[2]
        ref[1,1]=(-0.866120342442)*x_ref[0]+(0.927484984985)*x_ref[1]+(0.918081524156)*x_ref[2]
        ref[1,2]=(0.025238374235)*x_ref[0]+(-0.182938599269)*x_ref[1]+(-1.50048281322)*x_ref[2]
        ref[1,3]=(-0.302334971294)*x_ref[0]+(-0.925062799423)*x_ref[1]+(-1.4675138953)*x_ref[2]
        ref[1,4]=(-0.497738566447)*x_ref[0]+(-0.324727071875)*x_ref[1]+(-1.43341212055)*x_ref[2]
        ref[2,0]=(-1.4605237958)*x_ref[0]+(-0.332826007088)*x_ref[1]+(-0.230483401111)*x_ref[2]
        ref[2,1]=(0.801769248581)*x_ref[0]+(-0.930588036503)*x_ref[1]+(-1.06804808381)*x_ref[2]
        ref[2,2]=(-1.07972165017)*x_ref[0]+(1.09295362882)*x_ref[1]+(0.048550098477)*x_ref[2]
        ref[2,3]=(0.028290230337)*x_ref[0]+(0.179461246635)*x_ref[1]+(-1.09165257072)*x_ref[2]
        ref[2,4]=(0.238850986174)*x_ref[0]+(-0.837307206184)*x_ref[1]+(-0.617254874594)*x_ref[2]
        ref[3,0]=(-0.125921204582)*x_ref[0]+(0.234758657879)*x_ref[1]+(0.0847162247843)*x_ref[2]
        ref[3,1]=(-0.213352414907)*x_ref[0]+(0.206777352455)*x_ref[1]+(-0.135357638131)*x_ref[2]
        ref[3,2]=(-0.41897394434)*x_ref[0]+(0.152706100016)*x_ref[1]+(1.52390356012)*x_ref[2]
        ref[3,3]=(0.1831705248)*x_ref[0]+(-1.13556841364)*x_ref[1]+(-1.7661652693)*x_ref[2]
        ref[3,4]=(0.43473708388)*x_ref[0]+(1.04496896213)*x_ref[1]+(-0.816894739325)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.324237136046)*x[0]+(0.67172278334)*x[1]
        arg[0,0,1]=(-1.07360258996)*x[0]+(-0.646208771542)*x[1]
        arg[0,1,0]=(-0.040139403311)*x[0]+(-0.281626553874)*x[1]
        arg[0,1,1]=(0.963343050385)*x[0]+(-0.724838290038)*x[1]
        arg[1,0,0]=(-0.909697586367)*x[0]+(0.48041903189)*x[1]
        arg[1,0,1]=(0.414758676032)*x[0]+(-0.0921084575218)*x[1]
        arg[1,1,0]=(1.06852211241)*x[0]+(-1.01026709886)*x[1]
        arg[1,1,1]=(-0.675258695211)*x[0]+(0.51361078905)*x[1]
        arg[2,0,0]=(-0.791246198786)*x[0]+(-0.0283034425042)*x[1]
        arg[2,0,1]=(-0.87756010848)*x[0]+(-0.00868464377539)*x[1]
        arg[2,1,0]=(0.44209070344)*x[0]+(0.0846344294002)*x[1]
        arg[2,1,1]=(0.777287404475)*x[0]+(0.72614634535)*x[1]
        arg[3,0,0]=(-0.363354104693)*x[0]+(-1.22959837449)*x[1]
        arg[3,0,1]=(-0.532220757425)*x[0]+(-1.54990170304)*x[1]
        arg[3,1,0]=(-0.491099325731)*x[0]+(-0.412402332731)*x[1]
        arg[3,1,1]=(0.0750673227925)*x[0]+(-0.725108664028)*x[1]
        arg[4,0,0]=(-0.419689899608)*x[0]+(-1.32423898043)*x[1]
        arg[4,0,1]=(0.532682204197)*x[0]+(0.0703779631797)*x[1]
        arg[4,1,0]=(1.68643150913)*x[0]+(0.105443733595)*x[1]
        arg[4,1,1]=(0.936329701936)*x[0]+(1.04067808474)*x[1]
        arg[5,0,0]=(1.37285560133)*x[0]+(-0.680844454631)*x[1]
        arg[5,0,1]=(0.843677131436)*x[0]+(0.272365035967)*x[1]
        arg[5,1,0]=(0.00586472074043)*x[0]+(-1.09785622721)*x[1]
        arg[5,1,1]=(1.1410854961)*x[0]+(-0.591468439837)*x[1]
        ref[0,0,0]=(0.324237136046)*x_ref[0]+(0.67172278334)*x_ref[1]
        ref[0,0,1]=(-1.07360258996)*x_ref[0]+(-0.646208771542)*x_ref[1]
        ref[0,1,0]=(-0.040139403311)*x_ref[0]+(-0.281626553874)*x_ref[1]
        ref[0,1,1]=(0.963343050385)*x_ref[0]+(-0.724838290038)*x_ref[1]
        ref[1,0,0]=(-0.909697586367)*x_ref[0]+(0.48041903189)*x_ref[1]
        ref[1,0,1]=(0.414758676032)*x_ref[0]+(-0.0921084575218)*x_ref[1]
        ref[1,1,0]=(1.06852211241)*x_ref[0]+(-1.01026709886)*x_ref[1]
        ref[1,1,1]=(-0.675258695211)*x_ref[0]+(0.51361078905)*x_ref[1]
        ref[2,0,0]=(-0.791246198786)*x_ref[0]+(-0.0283034425042)*x_ref[1]
        ref[2,0,1]=(-0.87756010848)*x_ref[0]+(-0.00868464377539)*x_ref[1]
        ref[2,1,0]=(0.44209070344)*x_ref[0]+(0.0846344294002)*x_ref[1]
        ref[2,1,1]=(0.777287404475)*x_ref[0]+(0.72614634535)*x_ref[1]
        ref[3,0,0]=(-0.363354104693)*x_ref[0]+(-1.22959837449)*x_ref[1]
        ref[3,0,1]=(-0.532220757425)*x_ref[0]+(-1.54990170304)*x_ref[1]
        ref[3,1,0]=(-0.491099325731)*x_ref[0]+(-0.412402332731)*x_ref[1]
        ref[3,1,1]=(0.0750673227925)*x_ref[0]+(-0.725108664028)*x_ref[1]
        ref[4,0,0]=(-0.419689899608)*x_ref[0]+(-1.32423898043)*x_ref[1]
        ref[4,0,1]=(0.532682204197)*x_ref[0]+(0.0703779631797)*x_ref[1]
        ref[4,1,0]=(1.68643150913)*x_ref[0]+(0.105443733595)*x_ref[1]
        ref[4,1,1]=(0.936329701936)*x_ref[0]+(1.04067808474)*x_ref[1]
        ref[5,0,0]=(1.37285560133)*x_ref[0]+(-0.680844454631)*x_ref[1]
        ref[5,0,1]=(0.843677131436)*x_ref[0]+(0.272365035967)*x_ref[1]
        ref[5,1,0]=(0.00586472074043)*x_ref[0]+(-1.09785622721)*x_ref[1]
        ref[5,1,1]=(1.1410854961)*x_ref[0]+(-0.591468439837)*x_ref[1]
      else:
        arg[0,0,0]=(-0.289427308366)*x[0]+(1.17077011703)*x[1]+(0.763196501047)*x[2]
        arg[0,0,1]=(0.642730500265)*x[0]+(1.1719001632)*x[1]+(-0.121378040894)*x[2]
        arg[0,1,0]=(0.0879005621769)*x[0]+(-0.627582537594)*x[1]+(0.220072369447)*x[2]
        arg[0,1,1]=(-0.518762071566)*x[0]+(-0.700754417623)*x[1]+(-0.340623426752)*x[2]
        arg[1,0,0]=(-0.162635708202)*x[0]+(-0.996922517356)*x[1]+(0.48000598379)*x[2]
        arg[1,0,1]=(-0.49910792923)*x[0]+(1.11334748765)*x[1]+(0.0050657496159)*x[2]
        arg[1,1,0]=(-1.55972927502)*x[0]+(0.233936234238)*x[1]+(-0.546906209213)*x[2]
        arg[1,1,1]=(-0.6012330303)*x[0]+(-0.267951301884)*x[1]+(-0.193021060351)*x[2]
        arg[2,0,0]=(1.32515959423)*x[0]+(1.2070735022)*x[1]+(1.11474945521)*x[2]
        arg[2,0,1]=(-1.05500093406)*x[0]+(-0.937935240268)*x[1]+(0.180289797658)*x[2]
        arg[2,1,0]=(-0.450015559749)*x[0]+(-1.17715272891)*x[1]+(0.197160834934)*x[2]
        arg[2,1,1]=(-1.24969258786)*x[0]+(-0.300915973735)*x[1]+(-0.195058423793)*x[2]
        arg[3,0,0]=(0.704171975525)*x[0]+(1.0327384159)*x[1]+(-1.23664355084)*x[2]
        arg[3,0,1]=(0.916288412006)*x[0]+(1.6929893349)*x[1]+(-1.16682244827)*x[2]
        arg[3,1,0]=(-0.756147901025)*x[0]+(0.0232005287342)*x[1]+(0.213796289692)*x[2]
        arg[3,1,1]=(0.92799443461)*x[0]+(-0.228093116826)*x[1]+(-1.53373185211)*x[2]
        arg[4,0,0]=(-0.272514671459)*x[0]+(0.310026631492)*x[1]+(0.735127519475)*x[2]
        arg[4,0,1]=(1.23963120106)*x[0]+(-0.0769223754537)*x[1]+(0.767002455909)*x[2]
        arg[4,1,0]=(-1.1121433063)*x[0]+(0.748482838233)*x[1]+(-0.0985898502473)*x[2]
        arg[4,1,1]=(-1.08169600038)*x[0]+(-1.39575461699)*x[1]+(0.468455733883)*x[2]
        arg[5,0,0]=(-0.737100332594)*x[0]+(1.0950941569)*x[1]+(0.530144605873)*x[2]
        arg[5,0,1]=(1.94353080429)*x[0]+(1.61187736568)*x[1]+(0.887255997784)*x[2]
        arg[5,1,0]=(-0.492812782764)*x[0]+(-1.17859119673)*x[1]+(-0.627254028293)*x[2]
        arg[5,1,1]=(0.932565304399)*x[0]+(0.760125636669)*x[1]+(-0.687485104891)*x[2]
        ref[0,0,0]=(-0.289427308366)*x_ref[0]+(1.17077011703)*x_ref[1]+(0.763196501047)*x_ref[2]
        ref[0,0,1]=(0.642730500265)*x_ref[0]+(1.1719001632)*x_ref[1]+(-0.121378040894)*x_ref[2]
        ref[0,1,0]=(0.0879005621769)*x_ref[0]+(-0.627582537594)*x_ref[1]+(0.220072369447)*x_ref[2]
        ref[0,1,1]=(-0.518762071566)*x_ref[0]+(-0.700754417623)*x_ref[1]+(-0.340623426752)*x_ref[2]
        ref[1,0,0]=(-0.162635708202)*x_ref[0]+(-0.996922517356)*x_ref[1]+(0.48000598379)*x_ref[2]
        ref[1,0,1]=(-0.49910792923)*x_ref[0]+(1.11334748765)*x_ref[1]+(0.0050657496159)*x_ref[2]
        ref[1,1,0]=(-1.55972927502)*x_ref[0]+(0.233936234238)*x_ref[1]+(-0.546906209213)*x_ref[2]
        ref[1,1,1]=(-0.6012330303)*x_ref[0]+(-0.267951301884)*x_ref[1]+(-0.193021060351)*x_ref[2]
        ref[2,0,0]=(1.32515959423)*x_ref[0]+(1.2070735022)*x_ref[1]+(1.11474945521)*x_ref[2]
        ref[2,0,1]=(-1.05500093406)*x_ref[0]+(-0.937935240268)*x_ref[1]+(0.180289797658)*x_ref[2]
        ref[2,1,0]=(-0.450015559749)*x_ref[0]+(-1.17715272891)*x_ref[1]+(0.197160834934)*x_ref[2]
        ref[2,1,1]=(-1.24969258786)*x_ref[0]+(-0.300915973735)*x_ref[1]+(-0.195058423793)*x_ref[2]
        ref[3,0,0]=(0.704171975525)*x_ref[0]+(1.0327384159)*x_ref[1]+(-1.23664355084)*x_ref[2]
        ref[3,0,1]=(0.916288412006)*x_ref[0]+(1.6929893349)*x_ref[1]+(-1.16682244827)*x_ref[2]
        ref[3,1,0]=(-0.756147901025)*x_ref[0]+(0.0232005287342)*x_ref[1]+(0.213796289692)*x_ref[2]
        ref[3,1,1]=(0.92799443461)*x_ref[0]+(-0.228093116826)*x_ref[1]+(-1.53373185211)*x_ref[2]
        ref[4,0,0]=(-0.272514671459)*x_ref[0]+(0.310026631492)*x_ref[1]+(0.735127519475)*x_ref[2]
        ref[4,0,1]=(1.23963120106)*x_ref[0]+(-0.0769223754537)*x_ref[1]+(0.767002455909)*x_ref[2]
        ref[4,1,0]=(-1.1121433063)*x_ref[0]+(0.748482838233)*x_ref[1]+(-0.0985898502473)*x_ref[2]
        ref[4,1,1]=(-1.08169600038)*x_ref[0]+(-1.39575461699)*x_ref[1]+(0.468455733883)*x_ref[2]
        ref[5,0,0]=(-0.737100332594)*x_ref[0]+(1.0950941569)*x_ref[1]+(0.530144605873)*x_ref[2]
        ref[5,0,1]=(1.94353080429)*x_ref[0]+(1.61187736568)*x_ref[1]+(0.887255997784)*x_ref[2]
        ref[5,1,0]=(-0.492812782764)*x_ref[0]+(-1.17859119673)*x_ref[1]+(-0.627254028293)*x_ref[2]
        ref[5,1,1]=(0.932565304399)*x_ref[0]+(0.760125636669)*x_ref[1]+(-0.687485104891)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.261713255264)*x[0]+(-0.650225996431)*x[1]
        arg[0,0,0,1]=(0.177056727607)*x[0]+(-0.486487540393)*x[1]
        arg[0,0,1,0]=(1.38050234107)*x[0]+(0.273417897264)*x[1]
        arg[0,0,1,1]=(-1.059824929)*x[0]+(0.432893707387)*x[1]
        arg[0,0,2,0]=(1.11328992978)*x[0]+(1.14949237037)*x[1]
        arg[0,0,2,1]=(-0.176459646227)*x[0]+(-0.835980248375)*x[1]
        arg[0,1,0,0]=(0.992757237392)*x[0]+(0.113511598452)*x[1]
        arg[0,1,0,1]=(-0.211425610821)*x[0]+(-0.472765285276)*x[1]
        arg[0,1,1,0]=(-1.52498305232)*x[0]+(-0.869737083114)*x[1]
        arg[0,1,1,1]=(-0.474203117317)*x[0]+(-0.284532442353)*x[1]
        arg[0,1,2,0]=(-0.431542961546)*x[0]+(-0.0427397090453)*x[1]
        arg[0,1,2,1]=(-1.0562625269)*x[0]+(-0.74486767124)*x[1]
        arg[0,2,0,0]=(1.46553867064)*x[0]+(-0.650182195657)*x[1]
        arg[0,2,0,1]=(-0.717482663647)*x[0]+(-0.716320000777)*x[1]
        arg[0,2,1,0]=(-1.21701970743)*x[0]+(0.636901443761)*x[1]
        arg[0,2,1,1]=(0.438592928507)*x[0]+(0.812362899723)*x[1]
        arg[0,2,2,0]=(0.104885477141)*x[0]+(-0.127968978832)*x[1]
        arg[0,2,2,1]=(0.285116712849)*x[0]+(0.169740783305)*x[1]
        arg[0,3,0,0]=(-1.22825142912)*x[0]+(-0.562911687838)*x[1]
        arg[0,3,0,1]=(-0.569159534025)*x[0]+(1.30898732707)*x[1]
        arg[0,3,1,0]=(0.164709387569)*x[0]+(-1.62733582919)*x[1]
        arg[0,3,1,1]=(-0.477197320128)*x[0]+(-0.115207237191)*x[1]
        arg[0,3,2,0]=(-0.046871003317)*x[0]+(0.0371149345252)*x[1]
        arg[0,3,2,1]=(0.839624830979)*x[0]+(0.0785937909827)*x[1]
        arg[0,4,0,0]=(-0.0250725769598)*x[0]+(-0.535781603767)*x[1]
        arg[0,4,0,1]=(-0.761849740443)*x[0]+(-0.300714501152)*x[1]
        arg[0,4,1,0]=(0.0212637838833)*x[0]+(-1.56914428088)*x[1]
        arg[0,4,1,1]=(-0.254524677425)*x[0]+(0.699288623343)*x[1]
        arg[0,4,2,0]=(1.14678964272)*x[0]+(1.24738649379)*x[1]
        arg[0,4,2,1]=(1.80545916472)*x[0]+(1.22689087643)*x[1]
        arg[1,0,0,0]=(-0.116129927341)*x[0]+(-0.875165834648)*x[1]
        arg[1,0,0,1]=(-0.49905851236)*x[0]+(-1.10711214732)*x[1]
        arg[1,0,1,0]=(-1.44706167224)*x[0]+(-1.80259802262)*x[1]
        arg[1,0,1,1]=(0.555288273554)*x[0]+(0.153437302525)*x[1]
        arg[1,0,2,0]=(0.00774838569142)*x[0]+(-1.35683757658)*x[1]
        arg[1,0,2,1]=(-0.625156986286)*x[0]+(-0.201091038824)*x[1]
        arg[1,1,0,0]=(-0.765068445971)*x[0]+(1.32890117415)*x[1]
        arg[1,1,0,1]=(-0.629548919312)*x[0]+(-0.0512345513979)*x[1]
        arg[1,1,1,0]=(-0.361717423775)*x[0]+(0.286169858965)*x[1]
        arg[1,1,1,1]=(-0.533065922681)*x[0]+(0.103362397133)*x[1]
        arg[1,1,2,0]=(0.978041972715)*x[0]+(0.405949452109)*x[1]
        arg[1,1,2,1]=(-1.08418078618)*x[0]+(0.027879770152)*x[1]
        arg[1,2,0,0]=(1.23545256768)*x[0]+(0.574400116282)*x[1]
        arg[1,2,0,1]=(0.245362478488)*x[0]+(0.241577755816)*x[1]
        arg[1,2,1,0]=(0.645309443244)*x[0]+(0.0437773509036)*x[1]
        arg[1,2,1,1]=(0.00930197364105)*x[0]+(0.472462788946)*x[1]
        arg[1,2,2,0]=(1.44241534167)*x[0]+(-0.901332994844)*x[1]
        arg[1,2,2,1]=(0.893335284585)*x[0]+(0.109130146628)*x[1]
        arg[1,3,0,0]=(0.601907323675)*x[0]+(-1.54057502586)*x[1]
        arg[1,3,0,1]=(1.38086759376)*x[0]+(0.356522488945)*x[1]
        arg[1,3,1,0]=(1.16786130652)*x[0]+(-0.788756049484)*x[1]
        arg[1,3,1,1]=(-0.22580270698)*x[0]+(0.857686971693)*x[1]
        arg[1,3,2,0]=(0.709549396131)*x[0]+(-0.644330864877)*x[1]
        arg[1,3,2,1]=(1.04901965211)*x[0]+(0.0178384468958)*x[1]
        arg[1,4,0,0]=(-0.525050731412)*x[0]+(-0.253639122312)*x[1]
        arg[1,4,0,1]=(-1.63200911423)*x[0]+(0.519020443372)*x[1]
        arg[1,4,1,0]=(0.23956666878)*x[0]+(-1.76973024011)*x[1]
        arg[1,4,1,1]=(-0.412317050346)*x[0]+(-0.391711749094)*x[1]
        arg[1,4,2,0]=(1.63293176518)*x[0]+(0.583921572851)*x[1]
        arg[1,4,2,1]=(0.434893659373)*x[0]+(-0.324706930043)*x[1]
        arg[2,0,0,0]=(0.368478865639)*x[0]+(1.87637158068)*x[1]
        arg[2,0,0,1]=(1.32519020398)*x[0]+(0.306016568948)*x[1]
        arg[2,0,1,0]=(-0.284264003593)*x[0]+(-0.899022593367)*x[1]
        arg[2,0,1,1]=(-0.0439761413236)*x[0]+(-0.696386615358)*x[1]
        arg[2,0,2,0]=(0.407393812349)*x[0]+(-1.04815464317)*x[1]
        arg[2,0,2,1]=(-0.956661282071)*x[0]+(0.243281223782)*x[1]
        arg[2,1,0,0]=(-1.29201911939)*x[0]+(-0.693049706038)*x[1]
        arg[2,1,0,1]=(-0.711059848853)*x[0]+(-1.14204907272)*x[1]
        arg[2,1,1,0]=(-0.347693333057)*x[0]+(0.818879269672)*x[1]
        arg[2,1,1,1]=(0.00742807299875)*x[0]+(0.138845309984)*x[1]
        arg[2,1,2,0]=(1.13412244048)*x[0]+(-0.481422214102)*x[1]
        arg[2,1,2,1]=(0.64440302826)*x[0]+(-0.955231893089)*x[1]
        arg[2,2,0,0]=(0.13123141544)*x[0]+(-0.602266537541)*x[1]
        arg[2,2,0,1]=(-0.204818071763)*x[0]+(-1.13480248293)*x[1]
        arg[2,2,1,0]=(0.781058287579)*x[0]+(-1.31936224435)*x[1]
        arg[2,2,1,1]=(-0.591275351662)*x[0]+(0.691816969517)*x[1]
        arg[2,2,2,0]=(-0.921964826241)*x[0]+(1.28687075541)*x[1]
        arg[2,2,2,1]=(0.246260462337)*x[0]+(-1.04699282194)*x[1]
        arg[2,3,0,0]=(0.264209126513)*x[0]+(-0.697094400566)*x[1]
        arg[2,3,0,1]=(0.456158298484)*x[0]+(0.0288036445565)*x[1]
        arg[2,3,1,0]=(-0.720337925392)*x[0]+(0.0965182642186)*x[1]
        arg[2,3,1,1]=(0.397480419547)*x[0]+(-1.55232798344)*x[1]
        arg[2,3,2,0]=(-0.337139999283)*x[0]+(0.918809751807)*x[1]
        arg[2,3,2,1]=(1.34009447198)*x[0]+(-1.41390962018)*x[1]
        arg[2,4,0,0]=(1.67781838006)*x[0]+(0.521276193814)*x[1]
        arg[2,4,0,1]=(-0.114014321939)*x[0]+(0.529316778773)*x[1]
        arg[2,4,1,0]=(-0.19938131807)*x[0]+(-0.0712417614843)*x[1]
        arg[2,4,1,1]=(-0.724676753871)*x[0]+(0.731341170171)*x[1]
        arg[2,4,2,0]=(0.988755659656)*x[0]+(-1.0327352189)*x[1]
        arg[2,4,2,1]=(-0.526286732859)*x[0]+(-1.18295671598)*x[1]
        arg[3,0,0,0]=(-0.685223304723)*x[0]+(0.508431586068)*x[1]
        arg[3,0,0,1]=(0.529240274944)*x[0]+(0.77519128969)*x[1]
        arg[3,0,1,0]=(-0.757098431928)*x[0]+(1.41375812725)*x[1]
        arg[3,0,1,1]=(1.20805599335)*x[0]+(-0.184660334182)*x[1]
        arg[3,0,2,0]=(0.718413765561)*x[0]+(0.971149168706)*x[1]
        arg[3,0,2,1]=(-0.632670534518)*x[0]+(0.0389618967363)*x[1]
        arg[3,1,0,0]=(-0.525814673755)*x[0]+(0.574122587598)*x[1]
        arg[3,1,0,1]=(-1.04117288975)*x[0]+(-1.00738089823)*x[1]
        arg[3,1,1,0]=(0.0312806064182)*x[0]+(-0.30773511847)*x[1]
        arg[3,1,1,1]=(0.889930889355)*x[0]+(0.833689153489)*x[1]
        arg[3,1,2,0]=(-1.21586293381)*x[0]+(0.512182349574)*x[1]
        arg[3,1,2,1]=(-0.00437508755415)*x[0]+(-1.13228664497)*x[1]
        arg[3,2,0,0]=(1.64174174554)*x[0]+(-0.696500062562)*x[1]
        arg[3,2,0,1]=(0.0265343210604)*x[0]+(-0.707378716636)*x[1]
        arg[3,2,1,0]=(0.952922122967)*x[0]+(0.0946904561805)*x[1]
        arg[3,2,1,1]=(-1.06905266456)*x[0]+(-0.155122789807)*x[1]
        arg[3,2,2,0]=(-1.14410789976)*x[0]+(-0.379020164446)*x[1]
        arg[3,2,2,1]=(-0.0922611198553)*x[0]+(1.36680699099)*x[1]
        arg[3,3,0,0]=(0.849028205422)*x[0]+(-0.968698293811)*x[1]
        arg[3,3,0,1]=(0.649894718675)*x[0]+(0.238249259292)*x[1]
        arg[3,3,1,0]=(1.28170508471)*x[0]+(-0.736544947537)*x[1]
        arg[3,3,1,1]=(-0.188411685414)*x[0]+(1.14145010691)*x[1]
        arg[3,3,2,0]=(-1.13973037849)*x[0]+(-0.256565914821)*x[1]
        arg[3,3,2,1]=(0.955478585358)*x[0]+(-0.83417326515)*x[1]
        arg[3,4,0,0]=(0.155710789468)*x[0]+(0.204956811648)*x[1]
        arg[3,4,0,1]=(0.567866952974)*x[0]+(-0.168398320099)*x[1]
        arg[3,4,1,0]=(-1.36997367)*x[0]+(0.141984040172)*x[1]
        arg[3,4,1,1]=(-0.809532285206)*x[0]+(0.70611485405)*x[1]
        arg[3,4,2,0]=(-0.453260232051)*x[0]+(0.0665127865858)*x[1]
        arg[3,4,2,1]=(0.00142021115277)*x[0]+(1.02307505343)*x[1]
        ref[0,0,0,0]=(-0.261713255264)*x_ref[0]+(-0.650225996431)*x_ref[1]
        ref[0,0,0,1]=(0.177056727607)*x_ref[0]+(-0.486487540393)*x_ref[1]
        ref[0,0,1,0]=(1.38050234107)*x_ref[0]+(0.273417897264)*x_ref[1]
        ref[0,0,1,1]=(-1.059824929)*x_ref[0]+(0.432893707387)*x_ref[1]
        ref[0,0,2,0]=(1.11328992978)*x_ref[0]+(1.14949237037)*x_ref[1]
        ref[0,0,2,1]=(-0.176459646227)*x_ref[0]+(-0.835980248375)*x_ref[1]
        ref[0,1,0,0]=(0.992757237392)*x_ref[0]+(0.113511598452)*x_ref[1]
        ref[0,1,0,1]=(-0.211425610821)*x_ref[0]+(-0.472765285276)*x_ref[1]
        ref[0,1,1,0]=(-1.52498305232)*x_ref[0]+(-0.869737083114)*x_ref[1]
        ref[0,1,1,1]=(-0.474203117317)*x_ref[0]+(-0.284532442353)*x_ref[1]
        ref[0,1,2,0]=(-0.431542961546)*x_ref[0]+(-0.0427397090453)*x_ref[1]
        ref[0,1,2,1]=(-1.0562625269)*x_ref[0]+(-0.74486767124)*x_ref[1]
        ref[0,2,0,0]=(1.46553867064)*x_ref[0]+(-0.650182195657)*x_ref[1]
        ref[0,2,0,1]=(-0.717482663647)*x_ref[0]+(-0.716320000777)*x_ref[1]
        ref[0,2,1,0]=(-1.21701970743)*x_ref[0]+(0.636901443761)*x_ref[1]
        ref[0,2,1,1]=(0.438592928507)*x_ref[0]+(0.812362899723)*x_ref[1]
        ref[0,2,2,0]=(0.104885477141)*x_ref[0]+(-0.127968978832)*x_ref[1]
        ref[0,2,2,1]=(0.285116712849)*x_ref[0]+(0.169740783305)*x_ref[1]
        ref[0,3,0,0]=(-1.22825142912)*x_ref[0]+(-0.562911687838)*x_ref[1]
        ref[0,3,0,1]=(-0.569159534025)*x_ref[0]+(1.30898732707)*x_ref[1]
        ref[0,3,1,0]=(0.164709387569)*x_ref[0]+(-1.62733582919)*x_ref[1]
        ref[0,3,1,1]=(-0.477197320128)*x_ref[0]+(-0.115207237191)*x_ref[1]
        ref[0,3,2,0]=(-0.046871003317)*x_ref[0]+(0.0371149345252)*x_ref[1]
        ref[0,3,2,1]=(0.839624830979)*x_ref[0]+(0.0785937909827)*x_ref[1]
        ref[0,4,0,0]=(-0.0250725769598)*x_ref[0]+(-0.535781603767)*x_ref[1]
        ref[0,4,0,1]=(-0.761849740443)*x_ref[0]+(-0.300714501152)*x_ref[1]
        ref[0,4,1,0]=(0.0212637838833)*x_ref[0]+(-1.56914428088)*x_ref[1]
        ref[0,4,1,1]=(-0.254524677425)*x_ref[0]+(0.699288623343)*x_ref[1]
        ref[0,4,2,0]=(1.14678964272)*x_ref[0]+(1.24738649379)*x_ref[1]
        ref[0,4,2,1]=(1.80545916472)*x_ref[0]+(1.22689087643)*x_ref[1]
        ref[1,0,0,0]=(-0.116129927341)*x_ref[0]+(-0.875165834648)*x_ref[1]
        ref[1,0,0,1]=(-0.49905851236)*x_ref[0]+(-1.10711214732)*x_ref[1]
        ref[1,0,1,0]=(-1.44706167224)*x_ref[0]+(-1.80259802262)*x_ref[1]
        ref[1,0,1,1]=(0.555288273554)*x_ref[0]+(0.153437302525)*x_ref[1]
        ref[1,0,2,0]=(0.00774838569142)*x_ref[0]+(-1.35683757658)*x_ref[1]
        ref[1,0,2,1]=(-0.625156986286)*x_ref[0]+(-0.201091038824)*x_ref[1]
        ref[1,1,0,0]=(-0.765068445971)*x_ref[0]+(1.32890117415)*x_ref[1]
        ref[1,1,0,1]=(-0.629548919312)*x_ref[0]+(-0.0512345513979)*x_ref[1]
        ref[1,1,1,0]=(-0.361717423775)*x_ref[0]+(0.286169858965)*x_ref[1]
        ref[1,1,1,1]=(-0.533065922681)*x_ref[0]+(0.103362397133)*x_ref[1]
        ref[1,1,2,0]=(0.978041972715)*x_ref[0]+(0.405949452109)*x_ref[1]
        ref[1,1,2,1]=(-1.08418078618)*x_ref[0]+(0.027879770152)*x_ref[1]
        ref[1,2,0,0]=(1.23545256768)*x_ref[0]+(0.574400116282)*x_ref[1]
        ref[1,2,0,1]=(0.245362478488)*x_ref[0]+(0.241577755816)*x_ref[1]
        ref[1,2,1,0]=(0.645309443244)*x_ref[0]+(0.0437773509036)*x_ref[1]
        ref[1,2,1,1]=(0.00930197364105)*x_ref[0]+(0.472462788946)*x_ref[1]
        ref[1,2,2,0]=(1.44241534167)*x_ref[0]+(-0.901332994844)*x_ref[1]
        ref[1,2,2,1]=(0.893335284585)*x_ref[0]+(0.109130146628)*x_ref[1]
        ref[1,3,0,0]=(0.601907323675)*x_ref[0]+(-1.54057502586)*x_ref[1]
        ref[1,3,0,1]=(1.38086759376)*x_ref[0]+(0.356522488945)*x_ref[1]
        ref[1,3,1,0]=(1.16786130652)*x_ref[0]+(-0.788756049484)*x_ref[1]
        ref[1,3,1,1]=(-0.22580270698)*x_ref[0]+(0.857686971693)*x_ref[1]
        ref[1,3,2,0]=(0.709549396131)*x_ref[0]+(-0.644330864877)*x_ref[1]
        ref[1,3,2,1]=(1.04901965211)*x_ref[0]+(0.0178384468958)*x_ref[1]
        ref[1,4,0,0]=(-0.525050731412)*x_ref[0]+(-0.253639122312)*x_ref[1]
        ref[1,4,0,1]=(-1.63200911423)*x_ref[0]+(0.519020443372)*x_ref[1]
        ref[1,4,1,0]=(0.23956666878)*x_ref[0]+(-1.76973024011)*x_ref[1]
        ref[1,4,1,1]=(-0.412317050346)*x_ref[0]+(-0.391711749094)*x_ref[1]
        ref[1,4,2,0]=(1.63293176518)*x_ref[0]+(0.583921572851)*x_ref[1]
        ref[1,4,2,1]=(0.434893659373)*x_ref[0]+(-0.324706930043)*x_ref[1]
        ref[2,0,0,0]=(0.368478865639)*x_ref[0]+(1.87637158068)*x_ref[1]
        ref[2,0,0,1]=(1.32519020398)*x_ref[0]+(0.306016568948)*x_ref[1]
        ref[2,0,1,0]=(-0.284264003593)*x_ref[0]+(-0.899022593367)*x_ref[1]
        ref[2,0,1,1]=(-0.0439761413236)*x_ref[0]+(-0.696386615358)*x_ref[1]
        ref[2,0,2,0]=(0.407393812349)*x_ref[0]+(-1.04815464317)*x_ref[1]
        ref[2,0,2,1]=(-0.956661282071)*x_ref[0]+(0.243281223782)*x_ref[1]
        ref[2,1,0,0]=(-1.29201911939)*x_ref[0]+(-0.693049706038)*x_ref[1]
        ref[2,1,0,1]=(-0.711059848853)*x_ref[0]+(-1.14204907272)*x_ref[1]
        ref[2,1,1,0]=(-0.347693333057)*x_ref[0]+(0.818879269672)*x_ref[1]
        ref[2,1,1,1]=(0.00742807299875)*x_ref[0]+(0.138845309984)*x_ref[1]
        ref[2,1,2,0]=(1.13412244048)*x_ref[0]+(-0.481422214102)*x_ref[1]
        ref[2,1,2,1]=(0.64440302826)*x_ref[0]+(-0.955231893089)*x_ref[1]
        ref[2,2,0,0]=(0.13123141544)*x_ref[0]+(-0.602266537541)*x_ref[1]
        ref[2,2,0,1]=(-0.204818071763)*x_ref[0]+(-1.13480248293)*x_ref[1]
        ref[2,2,1,0]=(0.781058287579)*x_ref[0]+(-1.31936224435)*x_ref[1]
        ref[2,2,1,1]=(-0.591275351662)*x_ref[0]+(0.691816969517)*x_ref[1]
        ref[2,2,2,0]=(-0.921964826241)*x_ref[0]+(1.28687075541)*x_ref[1]
        ref[2,2,2,1]=(0.246260462337)*x_ref[0]+(-1.04699282194)*x_ref[1]
        ref[2,3,0,0]=(0.264209126513)*x_ref[0]+(-0.697094400566)*x_ref[1]
        ref[2,3,0,1]=(0.456158298484)*x_ref[0]+(0.0288036445565)*x_ref[1]
        ref[2,3,1,0]=(-0.720337925392)*x_ref[0]+(0.0965182642186)*x_ref[1]
        ref[2,3,1,1]=(0.397480419547)*x_ref[0]+(-1.55232798344)*x_ref[1]
        ref[2,3,2,0]=(-0.337139999283)*x_ref[0]+(0.918809751807)*x_ref[1]
        ref[2,3,2,1]=(1.34009447198)*x_ref[0]+(-1.41390962018)*x_ref[1]
        ref[2,4,0,0]=(1.67781838006)*x_ref[0]+(0.521276193814)*x_ref[1]
        ref[2,4,0,1]=(-0.114014321939)*x_ref[0]+(0.529316778773)*x_ref[1]
        ref[2,4,1,0]=(-0.19938131807)*x_ref[0]+(-0.0712417614843)*x_ref[1]
        ref[2,4,1,1]=(-0.724676753871)*x_ref[0]+(0.731341170171)*x_ref[1]
        ref[2,4,2,0]=(0.988755659656)*x_ref[0]+(-1.0327352189)*x_ref[1]
        ref[2,4,2,1]=(-0.526286732859)*x_ref[0]+(-1.18295671598)*x_ref[1]
        ref[3,0,0,0]=(-0.685223304723)*x_ref[0]+(0.508431586068)*x_ref[1]
        ref[3,0,0,1]=(0.529240274944)*x_ref[0]+(0.77519128969)*x_ref[1]
        ref[3,0,1,0]=(-0.757098431928)*x_ref[0]+(1.41375812725)*x_ref[1]
        ref[3,0,1,1]=(1.20805599335)*x_ref[0]+(-0.184660334182)*x_ref[1]
        ref[3,0,2,0]=(0.718413765561)*x_ref[0]+(0.971149168706)*x_ref[1]
        ref[3,0,2,1]=(-0.632670534518)*x_ref[0]+(0.0389618967363)*x_ref[1]
        ref[3,1,0,0]=(-0.525814673755)*x_ref[0]+(0.574122587598)*x_ref[1]
        ref[3,1,0,1]=(-1.04117288975)*x_ref[0]+(-1.00738089823)*x_ref[1]
        ref[3,1,1,0]=(0.0312806064182)*x_ref[0]+(-0.30773511847)*x_ref[1]
        ref[3,1,1,1]=(0.889930889355)*x_ref[0]+(0.833689153489)*x_ref[1]
        ref[3,1,2,0]=(-1.21586293381)*x_ref[0]+(0.512182349574)*x_ref[1]
        ref[3,1,2,1]=(-0.00437508755415)*x_ref[0]+(-1.13228664497)*x_ref[1]
        ref[3,2,0,0]=(1.64174174554)*x_ref[0]+(-0.696500062562)*x_ref[1]
        ref[3,2,0,1]=(0.0265343210604)*x_ref[0]+(-0.707378716636)*x_ref[1]
        ref[3,2,1,0]=(0.952922122967)*x_ref[0]+(0.0946904561805)*x_ref[1]
        ref[3,2,1,1]=(-1.06905266456)*x_ref[0]+(-0.155122789807)*x_ref[1]
        ref[3,2,2,0]=(-1.14410789976)*x_ref[0]+(-0.379020164446)*x_ref[1]
        ref[3,2,2,1]=(-0.0922611198553)*x_ref[0]+(1.36680699099)*x_ref[1]
        ref[3,3,0,0]=(0.849028205422)*x_ref[0]+(-0.968698293811)*x_ref[1]
        ref[3,3,0,1]=(0.649894718675)*x_ref[0]+(0.238249259292)*x_ref[1]
        ref[3,3,1,0]=(1.28170508471)*x_ref[0]+(-0.736544947537)*x_ref[1]
        ref[3,3,1,1]=(-0.188411685414)*x_ref[0]+(1.14145010691)*x_ref[1]
        ref[3,3,2,0]=(-1.13973037849)*x_ref[0]+(-0.256565914821)*x_ref[1]
        ref[3,3,2,1]=(0.955478585358)*x_ref[0]+(-0.83417326515)*x_ref[1]
        ref[3,4,0,0]=(0.155710789468)*x_ref[0]+(0.204956811648)*x_ref[1]
        ref[3,4,0,1]=(0.567866952974)*x_ref[0]+(-0.168398320099)*x_ref[1]
        ref[3,4,1,0]=(-1.36997367)*x_ref[0]+(0.141984040172)*x_ref[1]
        ref[3,4,1,1]=(-0.809532285206)*x_ref[0]+(0.70611485405)*x_ref[1]
        ref[3,4,2,0]=(-0.453260232051)*x_ref[0]+(0.0665127865858)*x_ref[1]
        ref[3,4,2,1]=(0.00142021115277)*x_ref[0]+(1.02307505343)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.59037121238)*x[0]+(-0.432259357387)*x[1]+(1.06313464506)*x[2]
        arg[0,0,0,1]=(-0.29317617186)*x[0]+(-0.477878024138)*x[1]+(-0.943279870083)*x[2]
        arg[0,0,1,0]=(-0.0140970955833)*x[0]+(0.141752937281)*x[1]+(0.46748311988)*x[2]
        arg[0,0,1,1]=(1.06645552756)*x[0]+(-1.4360320823)*x[1]+(-1.05643251873)*x[2]
        arg[0,0,2,0]=(-0.527854671619)*x[0]+(0.68918711423)*x[1]+(0.225750468582)*x[2]
        arg[0,0,2,1]=(0.765285169635)*x[0]+(-0.292098568154)*x[1]+(-0.611702446821)*x[2]
        arg[0,1,0,0]=(0.472029729041)*x[0]+(-0.391383508709)*x[1]+(-0.0882199174494)*x[2]
        arg[0,1,0,1]=(1.76010861737)*x[0]+(1.40668629263)*x[1]+(-0.171024087043)*x[2]
        arg[0,1,1,0]=(0.251929705017)*x[0]+(-0.326420063624)*x[1]+(-0.337849942832)*x[2]
        arg[0,1,1,1]=(0.049289902392)*x[0]+(1.61775987596)*x[1]+(0.545741506198)*x[2]
        arg[0,1,2,0]=(-0.500710333189)*x[0]+(-0.551222136922)*x[1]+(0.665856619864)*x[2]
        arg[0,1,2,1]=(0.271547935249)*x[0]+(0.46154733209)*x[1]+(0.0776993672589)*x[2]
        arg[0,2,0,0]=(0.46744023541)*x[0]+(0.281754164985)*x[1]+(0.11721465855)*x[2]
        arg[0,2,0,1]=(0.58865724402)*x[0]+(-0.280911783535)*x[1]+(0.793461201474)*x[2]
        arg[0,2,1,0]=(-0.348215254122)*x[0]+(-0.274357607001)*x[1]+(-0.975813503666)*x[2]
        arg[0,2,1,1]=(-0.546269224094)*x[0]+(0.54132986357)*x[1]+(-0.346797273682)*x[2]
        arg[0,2,2,0]=(0.956395362583)*x[0]+(-0.609333787208)*x[1]+(-0.86408067145)*x[2]
        arg[0,2,2,1]=(0.570075324955)*x[0]+(-0.251638346903)*x[1]+(0.292841304398)*x[2]
        arg[0,3,0,0]=(-0.240717117869)*x[0]+(0.571307481772)*x[1]+(0.66826910816)*x[2]
        arg[0,3,0,1]=(0.113871023723)*x[0]+(0.022498355636)*x[1]+(0.95331234379)*x[2]
        arg[0,3,1,0]=(-0.264946899001)*x[0]+(1.08412373438)*x[1]+(-1.10548347411)*x[2]
        arg[0,3,1,1]=(-0.744510132854)*x[0]+(-0.716638983187)*x[1]+(0.64330838226)*x[2]
        arg[0,3,2,0]=(1.15237343787)*x[0]+(-1.42565901229)*x[1]+(0.727585401396)*x[2]
        arg[0,3,2,1]=(-0.251464863627)*x[0]+(1.4067624912)*x[1]+(0.674418739118)*x[2]
        arg[0,4,0,0]=(-0.00551820316191)*x[0]+(0.193862853101)*x[1]+(1.8004924522)*x[2]
        arg[0,4,0,1]=(-0.985879358916)*x[0]+(0.476965815991)*x[1]+(0.96363449822)*x[2]
        arg[0,4,1,0]=(-0.657967361584)*x[0]+(0.833452284341)*x[1]+(-1.40128003822)*x[2]
        arg[0,4,1,1]=(0.301379967988)*x[0]+(0.0474545542353)*x[1]+(-0.633087920569)*x[2]
        arg[0,4,2,0]=(-1.94409755409)*x[0]+(-0.468638363928)*x[1]+(-0.713880237452)*x[2]
        arg[0,4,2,1]=(-1.13536316065)*x[0]+(-0.250761280127)*x[1]+(0.328722833182)*x[2]
        arg[1,0,0,0]=(0.556886266554)*x[0]+(-0.70523094638)*x[1]+(-0.333621268224)*x[2]
        arg[1,0,0,1]=(-0.995565683628)*x[0]+(-0.158120855239)*x[1]+(0.0738870470169)*x[2]
        arg[1,0,1,0]=(-0.209556712574)*x[0]+(0.54393408711)*x[1]+(-1.1617498966)*x[2]
        arg[1,0,1,1]=(0.666809649924)*x[0]+(1.18555984779)*x[1]+(-0.442831598796)*x[2]
        arg[1,0,2,0]=(-0.21354395591)*x[0]+(-0.634626698269)*x[1]+(1.25643757008)*x[2]
        arg[1,0,2,1]=(-0.676259187073)*x[0]+(-0.458532774397)*x[1]+(0.271156919902)*x[2]
        arg[1,1,0,0]=(-1.45529139262)*x[0]+(0.311382249732)*x[1]+(-0.134620909882)*x[2]
        arg[1,1,0,1]=(-1.47708829198)*x[0]+(0.968291536601)*x[1]+(-0.566606887608)*x[2]
        arg[1,1,1,0]=(1.18904329045)*x[0]+(-0.209072247061)*x[1]+(-0.26261457093)*x[2]
        arg[1,1,1,1]=(0.313723817215)*x[0]+(-0.169177309423)*x[1]+(-0.00119161234247)*x[2]
        arg[1,1,2,0]=(0.431298197457)*x[0]+(0.264113678484)*x[1]+(0.192619025313)*x[2]
        arg[1,1,2,1]=(0.0469221259627)*x[0]+(1.53477696238)*x[1]+(0.00794662724457)*x[2]
        arg[1,2,0,0]=(0.102179220279)*x[0]+(-0.391365988191)*x[1]+(0.302910557263)*x[2]
        arg[1,2,0,1]=(0.841803002697)*x[0]+(0.581539244301)*x[1]+(0.00409108885792)*x[2]
        arg[1,2,1,0]=(0.0840378161144)*x[0]+(-0.431329807551)*x[1]+(-1.0774029159)*x[2]
        arg[1,2,1,1]=(-0.270803943532)*x[0]+(0.0124514531652)*x[1]+(0.197735227059)*x[2]
        arg[1,2,2,0]=(0.479336250918)*x[0]+(0.0435331360815)*x[1]+(0.458048853417)*x[2]
        arg[1,2,2,1]=(-0.59916419564)*x[0]+(0.415546807791)*x[1]+(-1.12105410423)*x[2]
        arg[1,3,0,0]=(0.378277354507)*x[0]+(1.00734343779)*x[1]+(-0.282811389334)*x[2]
        arg[1,3,0,1]=(0.717420487474)*x[0]+(-0.730395427407)*x[1]+(1.63850706843)*x[2]
        arg[1,3,1,0]=(-0.0126954991872)*x[0]+(-1.35916858377)*x[1]+(-0.955731750212)*x[2]
        arg[1,3,1,1]=(-0.560951559862)*x[0]+(0.271085100351)*x[1]+(0.60767567276)*x[2]
        arg[1,3,2,0]=(-0.73286080289)*x[0]+(-1.12018839274)*x[1]+(0.365586479937)*x[2]
        arg[1,3,2,1]=(-0.495391543232)*x[0]+(-0.404605421036)*x[1]+(1.21261493791)*x[2]
        arg[1,4,0,0]=(-0.322147956238)*x[0]+(0.278257549185)*x[1]+(-0.927648624956)*x[2]
        arg[1,4,0,1]=(1.23243511737)*x[0]+(1.38355111729)*x[1]+(0.0813739945129)*x[2]
        arg[1,4,1,0]=(0.668384291635)*x[0]+(0.995328961231)*x[1]+(0.803009543307)*x[2]
        arg[1,4,1,1]=(1.42643018603)*x[0]+(0.940855692156)*x[1]+(0.277556451163)*x[2]
        arg[1,4,2,0]=(0.489588226757)*x[0]+(-1.63608809124)*x[1]+(-0.262918068522)*x[2]
        arg[1,4,2,1]=(-1.06922340601)*x[0]+(-0.0241938384008)*x[1]+(-0.244107290038)*x[2]
        arg[2,0,0,0]=(-0.556338154459)*x[0]+(0.12071126785)*x[1]+(0.11215804852)*x[2]
        arg[2,0,0,1]=(-0.479017339774)*x[0]+(-0.530313718209)*x[1]+(0.490083551372)*x[2]
        arg[2,0,1,0]=(0.731269264444)*x[0]+(0.245793038193)*x[1]+(0.0479833912422)*x[2]
        arg[2,0,1,1]=(-0.286189382681)*x[0]+(0.601756275554)*x[1]+(-0.559334203137)*x[2]
        arg[2,0,2,0]=(-0.876495720673)*x[0]+(0.592513223261)*x[1]+(0.844551261652)*x[2]
        arg[2,0,2,1]=(1.22915142187)*x[0]+(0.136223623577)*x[1]+(0.88656874029)*x[2]
        arg[2,1,0,0]=(0.542222650916)*x[0]+(-0.709719163926)*x[1]+(0.262207188498)*x[2]
        arg[2,1,0,1]=(-0.208896980258)*x[0]+(1.33968890395)*x[1]+(-0.599621090138)*x[2]
        arg[2,1,1,0]=(-0.171147406604)*x[0]+(0.332227749427)*x[1]+(1.15043521586)*x[2]
        arg[2,1,1,1]=(0.571607567027)*x[0]+(0.1367434941)*x[1]+(1.47797159814)*x[2]
        arg[2,1,2,0]=(0.839845574634)*x[0]+(0.732437808234)*x[1]+(-0.0311071654701)*x[2]
        arg[2,1,2,1]=(1.45193213653)*x[0]+(-0.0192312612414)*x[1]+(0.308521161282)*x[2]
        arg[2,2,0,0]=(-0.122755384028)*x[0]+(0.133510038844)*x[1]+(0.837213365425)*x[2]
        arg[2,2,0,1]=(-0.411967923958)*x[0]+(-0.955489021071)*x[1]+(0.269005602186)*x[2]
        arg[2,2,1,0]=(-1.19993089382)*x[0]+(0.276541296246)*x[1]+(0.523388479396)*x[2]
        arg[2,2,1,1]=(0.0586514653869)*x[0]+(-0.624975232742)*x[1]+(-1.55738876751)*x[2]
        arg[2,2,2,0]=(-0.892065938887)*x[0]+(-0.4848241954)*x[1]+(0.315463883511)*x[2]
        arg[2,2,2,1]=(0.0272963181118)*x[0]+(0.0187713454669)*x[1]+(-1.06524661854)*x[2]
        arg[2,3,0,0]=(-1.44099287886)*x[0]+(0.495718421075)*x[1]+(0.100401960223)*x[2]
        arg[2,3,0,1]=(0.171175119412)*x[0]+(-1.34795558607)*x[1]+(-0.361766653264)*x[2]
        arg[2,3,1,0]=(-1.80591882588)*x[0]+(-0.511459455136)*x[1]+(-0.491310264055)*x[2]
        arg[2,3,1,1]=(-0.419508934622)*x[0]+(-1.12084042844)*x[1]+(-0.086947672548)*x[2]
        arg[2,3,2,0]=(0.492388303934)*x[0]+(-0.623505828597)*x[1]+(1.6965873745)*x[2]
        arg[2,3,2,1]=(1.59203810356)*x[0]+(0.36918142231)*x[1]+(-1.90067115052)*x[2]
        arg[2,4,0,0]=(1.00875649298)*x[0]+(-1.43246384934)*x[1]+(-0.51811906593)*x[2]
        arg[2,4,0,1]=(0.260118918034)*x[0]+(-1.57606061865)*x[1]+(-0.79257165655)*x[2]
        arg[2,4,1,0]=(0.954414074467)*x[0]+(-0.230132876613)*x[1]+(0.0956378957424)*x[2]
        arg[2,4,1,1]=(0.958843806364)*x[0]+(-0.155804429995)*x[1]+(-1.02904877039)*x[2]
        arg[2,4,2,0]=(0.252668700811)*x[0]+(0.164829885749)*x[1]+(0.747032722421)*x[2]
        arg[2,4,2,1]=(1.96393052323)*x[0]+(-0.821363958807)*x[1]+(-1.01663555063)*x[2]
        arg[3,0,0,0]=(1.58086037053)*x[0]+(-1.66495295611)*x[1]+(-0.510526191383)*x[2]
        arg[3,0,0,1]=(-0.931871976029)*x[0]+(1.33102205125)*x[1]+(-0.95170580468)*x[2]
        arg[3,0,1,0]=(0.99825404115)*x[0]+(0.915274413552)*x[1]+(1.5057227674)*x[2]
        arg[3,0,1,1]=(0.0485551924989)*x[0]+(-1.0877405758)*x[1]+(-0.887342573702)*x[2]
        arg[3,0,2,0]=(-0.745146140677)*x[0]+(-0.0108141266015)*x[1]+(0.194597576308)*x[2]
        arg[3,0,2,1]=(-0.935029218338)*x[0]+(0.726322809025)*x[1]+(0.57508467103)*x[2]
        arg[3,1,0,0]=(0.115668237156)*x[0]+(0.0970255991906)*x[1]+(-0.227291123547)*x[2]
        arg[3,1,0,1]=(0.167722335863)*x[0]+(-0.552168424291)*x[1]+(1.36293027463)*x[2]
        arg[3,1,1,0]=(-1.05781878028)*x[0]+(-0.0429389047972)*x[1]+(-0.257166337509)*x[2]
        arg[3,1,1,1]=(0.129148486118)*x[0]+(-0.0209434899369)*x[1]+(0.253153005253)*x[2]
        arg[3,1,2,0]=(1.3236615365)*x[0]+(-1.18164469224)*x[1]+(0.698012697426)*x[2]
        arg[3,1,2,1]=(1.10438661886)*x[0]+(-0.852903948395)*x[1]+(1.40365162024)*x[2]
        arg[3,2,0,0]=(-0.949308009957)*x[0]+(1.01993206161)*x[1]+(-0.821721697476)*x[2]
        arg[3,2,0,1]=(0.326450662787)*x[0]+(0.395096088107)*x[1]+(-0.368904800584)*x[2]
        arg[3,2,1,0]=(0.13107780566)*x[0]+(-0.342904134623)*x[1]+(0.00835087701185)*x[2]
        arg[3,2,1,1]=(-0.500694550647)*x[0]+(1.38322390003)*x[1]+(0.0576453183412)*x[2]
        arg[3,2,2,0]=(-1.32014489811)*x[0]+(-0.444079317864)*x[1]+(1.1135946018)*x[2]
        arg[3,2,2,1]=(-0.19593105007)*x[0]+(-0.00224390969377)*x[1]+(-0.73869338942)*x[2]
        arg[3,3,0,0]=(0.0151201622255)*x[0]+(-1.10873788153)*x[1]+(0.189116254832)*x[2]
        arg[3,3,0,1]=(1.05996205566)*x[0]+(0.398465639867)*x[1]+(-0.319773104448)*x[2]
        arg[3,3,1,0]=(0.687612827243)*x[0]+(1.10970823257)*x[1]+(1.02543288975)*x[2]
        arg[3,3,1,1]=(1.27258509981)*x[0]+(-1.7060826391)*x[1]+(-0.84948812227)*x[2]
        arg[3,3,2,0]=(1.2343036886)*x[0]+(0.521090253023)*x[1]+(-1.13377412954)*x[2]
        arg[3,3,2,1]=(-0.65568738385)*x[0]+(1.00354617712)*x[1]+(0.836719015812)*x[2]
        arg[3,4,0,0]=(-0.545324078594)*x[0]+(0.452747782208)*x[1]+(-1.59718954237)*x[2]
        arg[3,4,0,1]=(1.71651153317)*x[0]+(-0.161984538787)*x[1]+(-0.377244827904)*x[2]
        arg[3,4,1,0]=(-0.284033324558)*x[0]+(-0.405941433062)*x[1]+(0.824573264594)*x[2]
        arg[3,4,1,1]=(-0.112542702803)*x[0]+(-0.470814805786)*x[1]+(-0.827427436576)*x[2]
        arg[3,4,2,0]=(-0.763655132981)*x[0]+(-0.442818775864)*x[1]+(0.809814510041)*x[2]
        arg[3,4,2,1]=(-0.33794306298)*x[0]+(1.91169591332)*x[1]+(-0.487268088196)*x[2]
        ref[0,0,0,0]=(-0.59037121238)*x_ref[0]+(-0.432259357387)*x_ref[1]+(1.06313464506)*x_ref[2]
        ref[0,0,0,1]=(-0.29317617186)*x_ref[0]+(-0.477878024138)*x_ref[1]+(-0.943279870083)*x_ref[2]
        ref[0,0,1,0]=(-0.0140970955833)*x_ref[0]+(0.141752937281)*x_ref[1]+(0.46748311988)*x_ref[2]
        ref[0,0,1,1]=(1.06645552756)*x_ref[0]+(-1.4360320823)*x_ref[1]+(-1.05643251873)*x_ref[2]
        ref[0,0,2,0]=(-0.527854671619)*x_ref[0]+(0.68918711423)*x_ref[1]+(0.225750468582)*x_ref[2]
        ref[0,0,2,1]=(0.765285169635)*x_ref[0]+(-0.292098568154)*x_ref[1]+(-0.611702446821)*x_ref[2]
        ref[0,1,0,0]=(0.472029729041)*x_ref[0]+(-0.391383508709)*x_ref[1]+(-0.0882199174494)*x_ref[2]
        ref[0,1,0,1]=(1.76010861737)*x_ref[0]+(1.40668629263)*x_ref[1]+(-0.171024087043)*x_ref[2]
        ref[0,1,1,0]=(0.251929705017)*x_ref[0]+(-0.326420063624)*x_ref[1]+(-0.337849942832)*x_ref[2]
        ref[0,1,1,1]=(0.049289902392)*x_ref[0]+(1.61775987596)*x_ref[1]+(0.545741506198)*x_ref[2]
        ref[0,1,2,0]=(-0.500710333189)*x_ref[0]+(-0.551222136922)*x_ref[1]+(0.665856619864)*x_ref[2]
        ref[0,1,2,1]=(0.271547935249)*x_ref[0]+(0.46154733209)*x_ref[1]+(0.0776993672589)*x_ref[2]
        ref[0,2,0,0]=(0.46744023541)*x_ref[0]+(0.281754164985)*x_ref[1]+(0.11721465855)*x_ref[2]
        ref[0,2,0,1]=(0.58865724402)*x_ref[0]+(-0.280911783535)*x_ref[1]+(0.793461201474)*x_ref[2]
        ref[0,2,1,0]=(-0.348215254122)*x_ref[0]+(-0.274357607001)*x_ref[1]+(-0.975813503666)*x_ref[2]
        ref[0,2,1,1]=(-0.546269224094)*x_ref[0]+(0.54132986357)*x_ref[1]+(-0.346797273682)*x_ref[2]
        ref[0,2,2,0]=(0.956395362583)*x_ref[0]+(-0.609333787208)*x_ref[1]+(-0.86408067145)*x_ref[2]
        ref[0,2,2,1]=(0.570075324955)*x_ref[0]+(-0.251638346903)*x_ref[1]+(0.292841304398)*x_ref[2]
        ref[0,3,0,0]=(-0.240717117869)*x_ref[0]+(0.571307481772)*x_ref[1]+(0.66826910816)*x_ref[2]
        ref[0,3,0,1]=(0.113871023723)*x_ref[0]+(0.022498355636)*x_ref[1]+(0.95331234379)*x_ref[2]
        ref[0,3,1,0]=(-0.264946899001)*x_ref[0]+(1.08412373438)*x_ref[1]+(-1.10548347411)*x_ref[2]
        ref[0,3,1,1]=(-0.744510132854)*x_ref[0]+(-0.716638983187)*x_ref[1]+(0.64330838226)*x_ref[2]
        ref[0,3,2,0]=(1.15237343787)*x_ref[0]+(-1.42565901229)*x_ref[1]+(0.727585401396)*x_ref[2]
        ref[0,3,2,1]=(-0.251464863627)*x_ref[0]+(1.4067624912)*x_ref[1]+(0.674418739118)*x_ref[2]
        ref[0,4,0,0]=(-0.00551820316191)*x_ref[0]+(0.193862853101)*x_ref[1]+(1.8004924522)*x_ref[2]
        ref[0,4,0,1]=(-0.985879358916)*x_ref[0]+(0.476965815991)*x_ref[1]+(0.96363449822)*x_ref[2]
        ref[0,4,1,0]=(-0.657967361584)*x_ref[0]+(0.833452284341)*x_ref[1]+(-1.40128003822)*x_ref[2]
        ref[0,4,1,1]=(0.301379967988)*x_ref[0]+(0.0474545542353)*x_ref[1]+(-0.633087920569)*x_ref[2]
        ref[0,4,2,0]=(-1.94409755409)*x_ref[0]+(-0.468638363928)*x_ref[1]+(-0.713880237452)*x_ref[2]
        ref[0,4,2,1]=(-1.13536316065)*x_ref[0]+(-0.250761280127)*x_ref[1]+(0.328722833182)*x_ref[2]
        ref[1,0,0,0]=(0.556886266554)*x_ref[0]+(-0.70523094638)*x_ref[1]+(-0.333621268224)*x_ref[2]
        ref[1,0,0,1]=(-0.995565683628)*x_ref[0]+(-0.158120855239)*x_ref[1]+(0.0738870470169)*x_ref[2]
        ref[1,0,1,0]=(-0.209556712574)*x_ref[0]+(0.54393408711)*x_ref[1]+(-1.1617498966)*x_ref[2]
        ref[1,0,1,1]=(0.666809649924)*x_ref[0]+(1.18555984779)*x_ref[1]+(-0.442831598796)*x_ref[2]
        ref[1,0,2,0]=(-0.21354395591)*x_ref[0]+(-0.634626698269)*x_ref[1]+(1.25643757008)*x_ref[2]
        ref[1,0,2,1]=(-0.676259187073)*x_ref[0]+(-0.458532774397)*x_ref[1]+(0.271156919902)*x_ref[2]
        ref[1,1,0,0]=(-1.45529139262)*x_ref[0]+(0.311382249732)*x_ref[1]+(-0.134620909882)*x_ref[2]
        ref[1,1,0,1]=(-1.47708829198)*x_ref[0]+(0.968291536601)*x_ref[1]+(-0.566606887608)*x_ref[2]
        ref[1,1,1,0]=(1.18904329045)*x_ref[0]+(-0.209072247061)*x_ref[1]+(-0.26261457093)*x_ref[2]
        ref[1,1,1,1]=(0.313723817215)*x_ref[0]+(-0.169177309423)*x_ref[1]+(-0.00119161234247)*x_ref[2]
        ref[1,1,2,0]=(0.431298197457)*x_ref[0]+(0.264113678484)*x_ref[1]+(0.192619025313)*x_ref[2]
        ref[1,1,2,1]=(0.0469221259627)*x_ref[0]+(1.53477696238)*x_ref[1]+(0.00794662724457)*x_ref[2]
        ref[1,2,0,0]=(0.102179220279)*x_ref[0]+(-0.391365988191)*x_ref[1]+(0.302910557263)*x_ref[2]
        ref[1,2,0,1]=(0.841803002697)*x_ref[0]+(0.581539244301)*x_ref[1]+(0.00409108885792)*x_ref[2]
        ref[1,2,1,0]=(0.0840378161144)*x_ref[0]+(-0.431329807551)*x_ref[1]+(-1.0774029159)*x_ref[2]
        ref[1,2,1,1]=(-0.270803943532)*x_ref[0]+(0.0124514531652)*x_ref[1]+(0.197735227059)*x_ref[2]
        ref[1,2,2,0]=(0.479336250918)*x_ref[0]+(0.0435331360815)*x_ref[1]+(0.458048853417)*x_ref[2]
        ref[1,2,2,1]=(-0.59916419564)*x_ref[0]+(0.415546807791)*x_ref[1]+(-1.12105410423)*x_ref[2]
        ref[1,3,0,0]=(0.378277354507)*x_ref[0]+(1.00734343779)*x_ref[1]+(-0.282811389334)*x_ref[2]
        ref[1,3,0,1]=(0.717420487474)*x_ref[0]+(-0.730395427407)*x_ref[1]+(1.63850706843)*x_ref[2]
        ref[1,3,1,0]=(-0.0126954991872)*x_ref[0]+(-1.35916858377)*x_ref[1]+(-0.955731750212)*x_ref[2]
        ref[1,3,1,1]=(-0.560951559862)*x_ref[0]+(0.271085100351)*x_ref[1]+(0.60767567276)*x_ref[2]
        ref[1,3,2,0]=(-0.73286080289)*x_ref[0]+(-1.12018839274)*x_ref[1]+(0.365586479937)*x_ref[2]
        ref[1,3,2,1]=(-0.495391543232)*x_ref[0]+(-0.404605421036)*x_ref[1]+(1.21261493791)*x_ref[2]
        ref[1,4,0,0]=(-0.322147956238)*x_ref[0]+(0.278257549185)*x_ref[1]+(-0.927648624956)*x_ref[2]
        ref[1,4,0,1]=(1.23243511737)*x_ref[0]+(1.38355111729)*x_ref[1]+(0.0813739945129)*x_ref[2]
        ref[1,4,1,0]=(0.668384291635)*x_ref[0]+(0.995328961231)*x_ref[1]+(0.803009543307)*x_ref[2]
        ref[1,4,1,1]=(1.42643018603)*x_ref[0]+(0.940855692156)*x_ref[1]+(0.277556451163)*x_ref[2]
        ref[1,4,2,0]=(0.489588226757)*x_ref[0]+(-1.63608809124)*x_ref[1]+(-0.262918068522)*x_ref[2]
        ref[1,4,2,1]=(-1.06922340601)*x_ref[0]+(-0.0241938384008)*x_ref[1]+(-0.244107290038)*x_ref[2]
        ref[2,0,0,0]=(-0.556338154459)*x_ref[0]+(0.12071126785)*x_ref[1]+(0.11215804852)*x_ref[2]
        ref[2,0,0,1]=(-0.479017339774)*x_ref[0]+(-0.530313718209)*x_ref[1]+(0.490083551372)*x_ref[2]
        ref[2,0,1,0]=(0.731269264444)*x_ref[0]+(0.245793038193)*x_ref[1]+(0.0479833912422)*x_ref[2]
        ref[2,0,1,1]=(-0.286189382681)*x_ref[0]+(0.601756275554)*x_ref[1]+(-0.559334203137)*x_ref[2]
        ref[2,0,2,0]=(-0.876495720673)*x_ref[0]+(0.592513223261)*x_ref[1]+(0.844551261652)*x_ref[2]
        ref[2,0,2,1]=(1.22915142187)*x_ref[0]+(0.136223623577)*x_ref[1]+(0.88656874029)*x_ref[2]
        ref[2,1,0,0]=(0.542222650916)*x_ref[0]+(-0.709719163926)*x_ref[1]+(0.262207188498)*x_ref[2]
        ref[2,1,0,1]=(-0.208896980258)*x_ref[0]+(1.33968890395)*x_ref[1]+(-0.599621090138)*x_ref[2]
        ref[2,1,1,0]=(-0.171147406604)*x_ref[0]+(0.332227749427)*x_ref[1]+(1.15043521586)*x_ref[2]
        ref[2,1,1,1]=(0.571607567027)*x_ref[0]+(0.1367434941)*x_ref[1]+(1.47797159814)*x_ref[2]
        ref[2,1,2,0]=(0.839845574634)*x_ref[0]+(0.732437808234)*x_ref[1]+(-0.0311071654701)*x_ref[2]
        ref[2,1,2,1]=(1.45193213653)*x_ref[0]+(-0.0192312612414)*x_ref[1]+(0.308521161282)*x_ref[2]
        ref[2,2,0,0]=(-0.122755384028)*x_ref[0]+(0.133510038844)*x_ref[1]+(0.837213365425)*x_ref[2]
        ref[2,2,0,1]=(-0.411967923958)*x_ref[0]+(-0.955489021071)*x_ref[1]+(0.269005602186)*x_ref[2]
        ref[2,2,1,0]=(-1.19993089382)*x_ref[0]+(0.276541296246)*x_ref[1]+(0.523388479396)*x_ref[2]
        ref[2,2,1,1]=(0.0586514653869)*x_ref[0]+(-0.624975232742)*x_ref[1]+(-1.55738876751)*x_ref[2]
        ref[2,2,2,0]=(-0.892065938887)*x_ref[0]+(-0.4848241954)*x_ref[1]+(0.315463883511)*x_ref[2]
        ref[2,2,2,1]=(0.0272963181118)*x_ref[0]+(0.0187713454669)*x_ref[1]+(-1.06524661854)*x_ref[2]
        ref[2,3,0,0]=(-1.44099287886)*x_ref[0]+(0.495718421075)*x_ref[1]+(0.100401960223)*x_ref[2]
        ref[2,3,0,1]=(0.171175119412)*x_ref[0]+(-1.34795558607)*x_ref[1]+(-0.361766653264)*x_ref[2]
        ref[2,3,1,0]=(-1.80591882588)*x_ref[0]+(-0.511459455136)*x_ref[1]+(-0.491310264055)*x_ref[2]
        ref[2,3,1,1]=(-0.419508934622)*x_ref[0]+(-1.12084042844)*x_ref[1]+(-0.086947672548)*x_ref[2]
        ref[2,3,2,0]=(0.492388303934)*x_ref[0]+(-0.623505828597)*x_ref[1]+(1.6965873745)*x_ref[2]
        ref[2,3,2,1]=(1.59203810356)*x_ref[0]+(0.36918142231)*x_ref[1]+(-1.90067115052)*x_ref[2]
        ref[2,4,0,0]=(1.00875649298)*x_ref[0]+(-1.43246384934)*x_ref[1]+(-0.51811906593)*x_ref[2]
        ref[2,4,0,1]=(0.260118918034)*x_ref[0]+(-1.57606061865)*x_ref[1]+(-0.79257165655)*x_ref[2]
        ref[2,4,1,0]=(0.954414074467)*x_ref[0]+(-0.230132876613)*x_ref[1]+(0.0956378957424)*x_ref[2]
        ref[2,4,1,1]=(0.958843806364)*x_ref[0]+(-0.155804429995)*x_ref[1]+(-1.02904877039)*x_ref[2]
        ref[2,4,2,0]=(0.252668700811)*x_ref[0]+(0.164829885749)*x_ref[1]+(0.747032722421)*x_ref[2]
        ref[2,4,2,1]=(1.96393052323)*x_ref[0]+(-0.821363958807)*x_ref[1]+(-1.01663555063)*x_ref[2]
        ref[3,0,0,0]=(1.58086037053)*x_ref[0]+(-1.66495295611)*x_ref[1]+(-0.510526191383)*x_ref[2]
        ref[3,0,0,1]=(-0.931871976029)*x_ref[0]+(1.33102205125)*x_ref[1]+(-0.95170580468)*x_ref[2]
        ref[3,0,1,0]=(0.99825404115)*x_ref[0]+(0.915274413552)*x_ref[1]+(1.5057227674)*x_ref[2]
        ref[3,0,1,1]=(0.0485551924989)*x_ref[0]+(-1.0877405758)*x_ref[1]+(-0.887342573702)*x_ref[2]
        ref[3,0,2,0]=(-0.745146140677)*x_ref[0]+(-0.0108141266015)*x_ref[1]+(0.194597576308)*x_ref[2]
        ref[3,0,2,1]=(-0.935029218338)*x_ref[0]+(0.726322809025)*x_ref[1]+(0.57508467103)*x_ref[2]
        ref[3,1,0,0]=(0.115668237156)*x_ref[0]+(0.0970255991906)*x_ref[1]+(-0.227291123547)*x_ref[2]
        ref[3,1,0,1]=(0.167722335863)*x_ref[0]+(-0.552168424291)*x_ref[1]+(1.36293027463)*x_ref[2]
        ref[3,1,1,0]=(-1.05781878028)*x_ref[0]+(-0.0429389047972)*x_ref[1]+(-0.257166337509)*x_ref[2]
        ref[3,1,1,1]=(0.129148486118)*x_ref[0]+(-0.0209434899369)*x_ref[1]+(0.253153005253)*x_ref[2]
        ref[3,1,2,0]=(1.3236615365)*x_ref[0]+(-1.18164469224)*x_ref[1]+(0.698012697426)*x_ref[2]
        ref[3,1,2,1]=(1.10438661886)*x_ref[0]+(-0.852903948395)*x_ref[1]+(1.40365162024)*x_ref[2]
        ref[3,2,0,0]=(-0.949308009957)*x_ref[0]+(1.01993206161)*x_ref[1]+(-0.821721697476)*x_ref[2]
        ref[3,2,0,1]=(0.326450662787)*x_ref[0]+(0.395096088107)*x_ref[1]+(-0.368904800584)*x_ref[2]
        ref[3,2,1,0]=(0.13107780566)*x_ref[0]+(-0.342904134623)*x_ref[1]+(0.00835087701185)*x_ref[2]
        ref[3,2,1,1]=(-0.500694550647)*x_ref[0]+(1.38322390003)*x_ref[1]+(0.0576453183412)*x_ref[2]
        ref[3,2,2,0]=(-1.32014489811)*x_ref[0]+(-0.444079317864)*x_ref[1]+(1.1135946018)*x_ref[2]
        ref[3,2,2,1]=(-0.19593105007)*x_ref[0]+(-0.00224390969377)*x_ref[1]+(-0.73869338942)*x_ref[2]
        ref[3,3,0,0]=(0.0151201622255)*x_ref[0]+(-1.10873788153)*x_ref[1]+(0.189116254832)*x_ref[2]
        ref[3,3,0,1]=(1.05996205566)*x_ref[0]+(0.398465639867)*x_ref[1]+(-0.319773104448)*x_ref[2]
        ref[3,3,1,0]=(0.687612827243)*x_ref[0]+(1.10970823257)*x_ref[1]+(1.02543288975)*x_ref[2]
        ref[3,3,1,1]=(1.27258509981)*x_ref[0]+(-1.7060826391)*x_ref[1]+(-0.84948812227)*x_ref[2]
        ref[3,3,2,0]=(1.2343036886)*x_ref[0]+(0.521090253023)*x_ref[1]+(-1.13377412954)*x_ref[2]
        ref[3,3,2,1]=(-0.65568738385)*x_ref[0]+(1.00354617712)*x_ref[1]+(0.836719015812)*x_ref[2]
        ref[3,4,0,0]=(-0.545324078594)*x_ref[0]+(0.452747782208)*x_ref[1]+(-1.59718954237)*x_ref[2]
        ref[3,4,0,1]=(1.71651153317)*x_ref[0]+(-0.161984538787)*x_ref[1]+(-0.377244827904)*x_ref[2]
        ref[3,4,1,0]=(-0.284033324558)*x_ref[0]+(-0.405941433062)*x_ref[1]+(0.824573264594)*x_ref[2]
        ref[3,4,1,1]=(-0.112542702803)*x_ref[0]+(-0.470814805786)*x_ref[1]+(-0.827427436576)*x_ref[2]
        ref[3,4,2,0]=(-0.763655132981)*x_ref[0]+(-0.442818775864)*x_ref[1]+(0.809814510041)*x_ref[2]
        ref[3,4,2,1]=(-0.33794306298)*x_ref[0]+(1.91169591332)*x_ref[1]+(-0.487268088196)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.898871557679)*x[0]+(0.0218492481542)*x[1]
        ref=(0.898871557679)*x_ref[0]+(0.0218492481542)*x_ref[1]
      else:
        arg=(-0.490730396919)*x[0]+(0.840683566031)*x[1]+(-0.265457839434)*x[2]
        ref=(-0.490730396919)*x_ref[0]+(0.840683566031)*x_ref[1]+(-0.265457839434)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        arg[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
        ref[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        ref[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
      else:
        arg[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        arg[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
        ref[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        ref[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.609515733352)*x[0]+(0.477437125804)*x[1]
        arg[0,1]=(0.225027969847)*x[0]+(-1.2178256661)*x[1]
        arg[0,2]=(-0.414641287889)*x[0]+(-0.536814389632)*x[1]
        arg[0,3]=(0.993370642676)*x[0]+(0.430000528217)*x[1]
        arg[0,4]=(0.577804610335)*x[0]+(0.0950391061127)*x[1]
        arg[1,0]=(0.437140855856)*x[0]+(0.764946114688)*x[1]
        arg[1,1]=(-0.491195658809)*x[0]+(-1.04729131261)*x[1]
        arg[1,2]=(-0.179287249145)*x[0]+(-1.84699873991)*x[1]
        arg[1,3]=(0.650408136022)*x[0]+(0.437145116534)*x[1]
        arg[1,4]=(0.43853490169)*x[0]+(-0.633716126064)*x[1]
        arg[2,0]=(0.6275418675)*x[0]+(0.812839621482)*x[1]
        arg[2,1]=(0.151298084104)*x[0]+(0.191677853013)*x[1]
        arg[2,2]=(1.35800571802)*x[0]+(1.08489593249)*x[1]
        arg[2,3]=(0.343386100925)*x[0]+(-0.108323271881)*x[1]
        arg[2,4]=(1.5968773444)*x[0]+(0.421328566279)*x[1]
        arg[3,0]=(1.0607367804)*x[0]+(-0.347352081886)*x[1]
        arg[3,1]=(-0.307183028406)*x[0]+(0.171667489459)*x[1]
        arg[3,2]=(0.255609148103)*x[0]+(0.21702578756)*x[1]
        arg[3,3]=(0.204487068999)*x[0]+(0.380168463726)*x[1]
        arg[3,4]=(-0.727714186561)*x[0]+(-0.129616906837)*x[1]
        ref[0,0]=(-0.609515733352)*x_ref[0]+(0.477437125804)*x_ref[1]
        ref[0,1]=(0.225027969847)*x_ref[0]+(-1.2178256661)*x_ref[1]
        ref[0,2]=(-0.414641287889)*x_ref[0]+(-0.536814389632)*x_ref[1]
        ref[0,3]=(0.993370642676)*x_ref[0]+(0.430000528217)*x_ref[1]
        ref[0,4]=(0.577804610335)*x_ref[0]+(0.0950391061127)*x_ref[1]
        ref[1,0]=(0.437140855856)*x_ref[0]+(0.764946114688)*x_ref[1]
        ref[1,1]=(-0.491195658809)*x_ref[0]+(-1.04729131261)*x_ref[1]
        ref[1,2]=(-0.179287249145)*x_ref[0]+(-1.84699873991)*x_ref[1]
        ref[1,3]=(0.650408136022)*x_ref[0]+(0.437145116534)*x_ref[1]
        ref[1,4]=(0.43853490169)*x_ref[0]+(-0.633716126064)*x_ref[1]
        ref[2,0]=(0.6275418675)*x_ref[0]+(0.812839621482)*x_ref[1]
        ref[2,1]=(0.151298084104)*x_ref[0]+(0.191677853013)*x_ref[1]
        ref[2,2]=(1.35800571802)*x_ref[0]+(1.08489593249)*x_ref[1]
        ref[2,3]=(0.343386100925)*x_ref[0]+(-0.108323271881)*x_ref[1]
        ref[2,4]=(1.5968773444)*x_ref[0]+(0.421328566279)*x_ref[1]
        ref[3,0]=(1.0607367804)*x_ref[0]+(-0.347352081886)*x_ref[1]
        ref[3,1]=(-0.307183028406)*x_ref[0]+(0.171667489459)*x_ref[1]
        ref[3,2]=(0.255609148103)*x_ref[0]+(0.21702578756)*x_ref[1]
        ref[3,3]=(0.204487068999)*x_ref[0]+(0.380168463726)*x_ref[1]
        ref[3,4]=(-0.727714186561)*x_ref[0]+(-0.129616906837)*x_ref[1]
      else:
        arg[0,0]=(-0.282058347281)*x[0]+(-0.804878634484)*x[1]+(1.59321960831)*x[2]
        arg[0,1]=(0.046711660417)*x[0]+(-0.353452776565)*x[1]+(-1.22459734449)*x[2]
        arg[0,2]=(0.846231855307)*x[0]+(1.08618070898)*x[1]+(-1.50365393652)*x[2]
        arg[0,3]=(-0.358546890162)*x[0]+(1.53476318049)*x[1]+(-1.19382694871)*x[2]
        arg[0,4]=(-1.1943310015)*x[0]+(0.260842673421)*x[1]+(-0.337737972605)*x[2]
        arg[1,0]=(1.87166717118)*x[0]+(0.853452519127)*x[1]+(-0.59081660847)*x[2]
        arg[1,1]=(0.23310002871)*x[0]+(0.306981228888)*x[1]+(0.36040454683)*x[2]
        arg[1,2]=(-1.619288679)*x[0]+(-0.459254234211)*x[1]+(0.722479242965)*x[2]
        arg[1,3]=(-1.07724802769)*x[0]+(0.836465436044)*x[1]+(0.255446930493)*x[2]
        arg[1,4]=(1.07857092921)*x[0]+(1.25642859092)*x[1]+(-0.922799422574)*x[2]
        arg[2,0]=(-0.832179820568)*x[0]+(-0.511243247139)*x[1]+(0.589457094147)*x[2]
        arg[2,1]=(-0.22616483359)*x[0]+(1.75842052561)*x[1]+(-1.19438769374)*x[2]
        arg[2,2]=(-0.690678634371)*x[0]+(0.963335056891)*x[1]+(0.53653322138)*x[2]
        arg[2,3]=(-0.906140718249)*x[0]+(0.766237211441)*x[1]+(1.19445635284)*x[2]
        arg[2,4]=(-0.16966820901)*x[0]+(0.185706811112)*x[1]+(0.975546783315)*x[2]
        arg[3,0]=(1.85053644213)*x[0]+(1.000328518)*x[1]+(0.332466610087)*x[2]
        arg[3,1]=(-0.199393988242)*x[0]+(0.637601295557)*x[1]+(1.29246614993)*x[2]
        arg[3,2]=(-0.275167528438)*x[0]+(1.00557873837)*x[1]+(1.17143679274)*x[2]
        arg[3,3]=(-0.306797319235)*x[0]+(0.984827835253)*x[1]+(0.0324710900497)*x[2]
        arg[3,4]=(-1.07994829511)*x[0]+(0.341016609147)*x[1]+(0.115829923585)*x[2]
        ref[0,0]=(-0.282058347281)*x_ref[0]+(-0.804878634484)*x_ref[1]+(1.59321960831)*x_ref[2]
        ref[0,1]=(0.046711660417)*x_ref[0]+(-0.353452776565)*x_ref[1]+(-1.22459734449)*x_ref[2]
        ref[0,2]=(0.846231855307)*x_ref[0]+(1.08618070898)*x_ref[1]+(-1.50365393652)*x_ref[2]
        ref[0,3]=(-0.358546890162)*x_ref[0]+(1.53476318049)*x_ref[1]+(-1.19382694871)*x_ref[2]
        ref[0,4]=(-1.1943310015)*x_ref[0]+(0.260842673421)*x_ref[1]+(-0.337737972605)*x_ref[2]
        ref[1,0]=(1.87166717118)*x_ref[0]+(0.853452519127)*x_ref[1]+(-0.59081660847)*x_ref[2]
        ref[1,1]=(0.23310002871)*x_ref[0]+(0.306981228888)*x_ref[1]+(0.36040454683)*x_ref[2]
        ref[1,2]=(-1.619288679)*x_ref[0]+(-0.459254234211)*x_ref[1]+(0.722479242965)*x_ref[2]
        ref[1,3]=(-1.07724802769)*x_ref[0]+(0.836465436044)*x_ref[1]+(0.255446930493)*x_ref[2]
        ref[1,4]=(1.07857092921)*x_ref[0]+(1.25642859092)*x_ref[1]+(-0.922799422574)*x_ref[2]
        ref[2,0]=(-0.832179820568)*x_ref[0]+(-0.511243247139)*x_ref[1]+(0.589457094147)*x_ref[2]
        ref[2,1]=(-0.22616483359)*x_ref[0]+(1.75842052561)*x_ref[1]+(-1.19438769374)*x_ref[2]
        ref[2,2]=(-0.690678634371)*x_ref[0]+(0.963335056891)*x_ref[1]+(0.53653322138)*x_ref[2]
        ref[2,3]=(-0.906140718249)*x_ref[0]+(0.766237211441)*x_ref[1]+(1.19445635284)*x_ref[2]
        ref[2,4]=(-0.16966820901)*x_ref[0]+(0.185706811112)*x_ref[1]+(0.975546783315)*x_ref[2]
        ref[3,0]=(1.85053644213)*x_ref[0]+(1.000328518)*x_ref[1]+(0.332466610087)*x_ref[2]
        ref[3,1]=(-0.199393988242)*x_ref[0]+(0.637601295557)*x_ref[1]+(1.29246614993)*x_ref[2]
        ref[3,2]=(-0.275167528438)*x_ref[0]+(1.00557873837)*x_ref[1]+(1.17143679274)*x_ref[2]
        ref[3,3]=(-0.306797319235)*x_ref[0]+(0.984827835253)*x_ref[1]+(0.0324710900497)*x_ref[2]
        ref[3,4]=(-1.07994829511)*x_ref[0]+(0.341016609147)*x_ref[1]+(0.115829923585)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.618861298758)*x[0]+(-0.764616734859)*x[1]
        arg[0,0,1]=(-0.220720229024)*x[0]+(-0.670415412641)*x[1]
        arg[0,1,0]=(0.847082172076)*x[0]+(0.104515154796)*x[1]
        arg[0,1,1]=(-1.39487869037)*x[0]+(1.37639032205)*x[1]
        arg[1,0,0]=(-0.12944065767)*x[0]+(-0.91214412252)*x[1]
        arg[1,0,1]=(0.0421119211697)*x[0]+(1.19918915102)*x[1]
        arg[1,1,0]=(0.657099038454)*x[0]+(0.102886690913)*x[1]
        arg[1,1,1]=(0.114311284584)*x[0]+(-0.827908524893)*x[1]
        arg[2,0,0]=(-0.905281334495)*x[0]+(1.02228309099)*x[1]
        arg[2,0,1]=(-0.560619228879)*x[0]+(-0.924304525905)*x[1]
        arg[2,1,0]=(0.082868967617)*x[0]+(-0.129063346527)*x[1]
        arg[2,1,1]=(0.386687881433)*x[0]+(-0.874995344656)*x[1]
        arg[3,0,0]=(0.460976605295)*x[0]+(0.145631537258)*x[1]
        arg[3,0,1]=(0.439977672901)*x[0]+(0.714321229059)*x[1]
        arg[3,1,0]=(-0.0459363656635)*x[0]+(-0.0754715887416)*x[1]
        arg[3,1,1]=(-0.789910257931)*x[0]+(-0.806078036909)*x[1]
        arg[4,0,0]=(-1.09951071165)*x[0]+(-1.33461983123)*x[1]
        arg[4,0,1]=(0.857184302111)*x[0]+(0.823036666525)*x[1]
        arg[4,1,0]=(-1.0682486872)*x[0]+(-0.149861981932)*x[1]
        arg[4,1,1]=(-0.0758367136425)*x[0]+(0.0460556685916)*x[1]
        arg[5,0,0]=(0.856207295965)*x[0]+(-0.885450852663)*x[1]
        arg[5,0,1]=(-0.113939086685)*x[0]+(0.742808810968)*x[1]
        arg[5,1,0]=(-0.174573302919)*x[0]+(0.486803189896)*x[1]
        arg[5,1,1]=(1.1726771004)*x[0]+(-0.329765988186)*x[1]
        ref[0,0,0]=(0.618861298758)*x_ref[0]+(-0.764616734859)*x_ref[1]
        ref[0,0,1]=(-0.220720229024)*x_ref[0]+(-0.670415412641)*x_ref[1]
        ref[0,1,0]=(0.847082172076)*x_ref[0]+(0.104515154796)*x_ref[1]
        ref[0,1,1]=(-1.39487869037)*x_ref[0]+(1.37639032205)*x_ref[1]
        ref[1,0,0]=(-0.12944065767)*x_ref[0]+(-0.91214412252)*x_ref[1]
        ref[1,0,1]=(0.0421119211697)*x_ref[0]+(1.19918915102)*x_ref[1]
        ref[1,1,0]=(0.657099038454)*x_ref[0]+(0.102886690913)*x_ref[1]
        ref[1,1,1]=(0.114311284584)*x_ref[0]+(-0.827908524893)*x_ref[1]
        ref[2,0,0]=(-0.905281334495)*x_ref[0]+(1.02228309099)*x_ref[1]
        ref[2,0,1]=(-0.560619228879)*x_ref[0]+(-0.924304525905)*x_ref[1]
        ref[2,1,0]=(0.082868967617)*x_ref[0]+(-0.129063346527)*x_ref[1]
        ref[2,1,1]=(0.386687881433)*x_ref[0]+(-0.874995344656)*x_ref[1]
        ref[3,0,0]=(0.460976605295)*x_ref[0]+(0.145631537258)*x_ref[1]
        ref[3,0,1]=(0.439977672901)*x_ref[0]+(0.714321229059)*x_ref[1]
        ref[3,1,0]=(-0.0459363656635)*x_ref[0]+(-0.0754715887416)*x_ref[1]
        ref[3,1,1]=(-0.789910257931)*x_ref[0]+(-0.806078036909)*x_ref[1]
        ref[4,0,0]=(-1.09951071165)*x_ref[0]+(-1.33461983123)*x_ref[1]
        ref[4,0,1]=(0.857184302111)*x_ref[0]+(0.823036666525)*x_ref[1]
        ref[4,1,0]=(-1.0682486872)*x_ref[0]+(-0.149861981932)*x_ref[1]
        ref[4,1,1]=(-0.0758367136425)*x_ref[0]+(0.0460556685916)*x_ref[1]
        ref[5,0,0]=(0.856207295965)*x_ref[0]+(-0.885450852663)*x_ref[1]
        ref[5,0,1]=(-0.113939086685)*x_ref[0]+(0.742808810968)*x_ref[1]
        ref[5,1,0]=(-0.174573302919)*x_ref[0]+(0.486803189896)*x_ref[1]
        ref[5,1,1]=(1.1726771004)*x_ref[0]+(-0.329765988186)*x_ref[1]
      else:
        arg[0,0,0]=(-0.640205798541)*x[0]+(0.271670719935)*x[1]+(-0.163000598261)*x[2]
        arg[0,0,1]=(0.735605297981)*x[0]+(-1.05892412676)*x[1]+(-1.1083007441)*x[2]
        arg[0,1,0]=(0.0349869408333)*x[0]+(0.699147831788)*x[1]+(-0.454390826725)*x[2]
        arg[0,1,1]=(-0.703957679243)*x[0]+(-1.63654420176)*x[1]+(-1.13045075795)*x[2]
        arg[1,0,0]=(1.01167296102)*x[0]+(-0.362443738659)*x[1]+(-0.0481138040854)*x[2]
        arg[1,0,1]=(0.110033706339)*x[0]+(1.47809236549)*x[1]+(0.240737599484)*x[2]
        arg[1,1,0]=(-0.0429437939148)*x[0]+(-1.16945967196)*x[1]+(0.924577053186)*x[2]
        arg[1,1,1]=(-1.57856509244)*x[0]+(-0.460441809594)*x[1]+(-0.792590928973)*x[2]
        arg[2,0,0]=(-1.93020920065)*x[0]+(0.391612451745)*x[1]+(-0.0130238634938)*x[2]
        arg[2,0,1]=(0.0678908982414)*x[0]+(-0.929813889361)*x[1]+(0.727462587818)*x[2]
        arg[2,1,0]=(-0.297494494419)*x[0]+(0.167234033247)*x[1]+(-1.01372084798)*x[2]
        arg[2,1,1]=(-0.608810623174)*x[0]+(1.04629900384)*x[1]+(1.09464953456)*x[2]
        arg[3,0,0]=(1.2996449893)*x[0]+(-0.458197425323)*x[1]+(1.31523663461)*x[2]
        arg[3,0,1]=(-0.848124038387)*x[0]+(0.0136487458675)*x[1]+(1.13573336016)*x[2]
        arg[3,1,0]=(0.377245000633)*x[0]+(-0.766336675857)*x[1]+(0.380031677083)*x[2]
        arg[3,1,1]=(0.158289533163)*x[0]+(-0.149454925108)*x[1]+(-0.59348598427)*x[2]
        arg[4,0,0]=(-1.13101661269)*x[0]+(-0.694062133171)*x[1]+(0.00925221725394)*x[2]
        arg[4,0,1]=(-0.977317473287)*x[0]+(-0.121712752812)*x[1]+(-0.480470720169)*x[2]
        arg[4,1,0]=(1.14884269473)*x[0]+(-0.204958154665)*x[1]+(1.24328207187)*x[2]
        arg[4,1,1]=(0.0791969146134)*x[0]+(0.589873409499)*x[1]+(1.90521993698)*x[2]
        arg[5,0,0]=(1.75625788908)*x[0]+(0.0134922175773)*x[1]+(-0.699884009939)*x[2]
        arg[5,0,1]=(-0.73456900252)*x[0]+(0.923248295457)*x[1]+(1.05161845199)*x[2]
        arg[5,1,0]=(-0.0335528775888)*x[0]+(0.728742831853)*x[1]+(-0.5212350627)*x[2]
        arg[5,1,1]=(0.556540394981)*x[0]+(-0.128428520433)*x[1]+(1.06358498081)*x[2]
        ref[0,0,0]=(-0.640205798541)*x_ref[0]+(0.271670719935)*x_ref[1]+(-0.163000598261)*x_ref[2]
        ref[0,0,1]=(0.735605297981)*x_ref[0]+(-1.05892412676)*x_ref[1]+(-1.1083007441)*x_ref[2]
        ref[0,1,0]=(0.0349869408333)*x_ref[0]+(0.699147831788)*x_ref[1]+(-0.454390826725)*x_ref[2]
        ref[0,1,1]=(-0.703957679243)*x_ref[0]+(-1.63654420176)*x_ref[1]+(-1.13045075795)*x_ref[2]
        ref[1,0,0]=(1.01167296102)*x_ref[0]+(-0.362443738659)*x_ref[1]+(-0.0481138040854)*x_ref[2]
        ref[1,0,1]=(0.110033706339)*x_ref[0]+(1.47809236549)*x_ref[1]+(0.240737599484)*x_ref[2]
        ref[1,1,0]=(-0.0429437939148)*x_ref[0]+(-1.16945967196)*x_ref[1]+(0.924577053186)*x_ref[2]
        ref[1,1,1]=(-1.57856509244)*x_ref[0]+(-0.460441809594)*x_ref[1]+(-0.792590928973)*x_ref[2]
        ref[2,0,0]=(-1.93020920065)*x_ref[0]+(0.391612451745)*x_ref[1]+(-0.0130238634938)*x_ref[2]
        ref[2,0,1]=(0.0678908982414)*x_ref[0]+(-0.929813889361)*x_ref[1]+(0.727462587818)*x_ref[2]
        ref[2,1,0]=(-0.297494494419)*x_ref[0]+(0.167234033247)*x_ref[1]+(-1.01372084798)*x_ref[2]
        ref[2,1,1]=(-0.608810623174)*x_ref[0]+(1.04629900384)*x_ref[1]+(1.09464953456)*x_ref[2]
        ref[3,0,0]=(1.2996449893)*x_ref[0]+(-0.458197425323)*x_ref[1]+(1.31523663461)*x_ref[2]
        ref[3,0,1]=(-0.848124038387)*x_ref[0]+(0.0136487458675)*x_ref[1]+(1.13573336016)*x_ref[2]
        ref[3,1,0]=(0.377245000633)*x_ref[0]+(-0.766336675857)*x_ref[1]+(0.380031677083)*x_ref[2]
        ref[3,1,1]=(0.158289533163)*x_ref[0]+(-0.149454925108)*x_ref[1]+(-0.59348598427)*x_ref[2]
        ref[4,0,0]=(-1.13101661269)*x_ref[0]+(-0.694062133171)*x_ref[1]+(0.00925221725394)*x_ref[2]
        ref[4,0,1]=(-0.977317473287)*x_ref[0]+(-0.121712752812)*x_ref[1]+(-0.480470720169)*x_ref[2]
        ref[4,1,0]=(1.14884269473)*x_ref[0]+(-0.204958154665)*x_ref[1]+(1.24328207187)*x_ref[2]
        ref[4,1,1]=(0.0791969146134)*x_ref[0]+(0.589873409499)*x_ref[1]+(1.90521993698)*x_ref[2]
        ref[5,0,0]=(1.75625788908)*x_ref[0]+(0.0134922175773)*x_ref[1]+(-0.699884009939)*x_ref[2]
        ref[5,0,1]=(-0.73456900252)*x_ref[0]+(0.923248295457)*x_ref[1]+(1.05161845199)*x_ref[2]
        ref[5,1,0]=(-0.0335528775888)*x_ref[0]+(0.728742831853)*x_ref[1]+(-0.5212350627)*x_ref[2]
        ref[5,1,1]=(0.556540394981)*x_ref[0]+(-0.128428520433)*x_ref[1]+(1.06358498081)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.507324804235)*x[0]+(0.725831700089)*x[1]
        arg[0,0,0,1]=(0.847901454936)*x[0]+(0.84787024099)*x[1]
        arg[0,0,1,0]=(0.221488465784)*x[0]+(0.0449286988708)*x[1]
        arg[0,0,1,1]=(0.00897176578501)*x[0]+(-0.799118487206)*x[1]
        arg[0,0,2,0]=(0.171491682599)*x[0]+(1.0259065872)*x[1]
        arg[0,0,2,1]=(-0.213244683798)*x[0]+(-0.706816385516)*x[1]
        arg[0,1,0,0]=(1.54199875186)*x[0]+(-1.09479402082)*x[1]
        arg[0,1,0,1]=(-0.755278051349)*x[0]+(-0.371785089226)*x[1]
        arg[0,1,1,0]=(-0.798931370137)*x[0]+(-0.716047155117)*x[1]
        arg[0,1,1,1]=(-0.79104072521)*x[0]+(-1.45842710768)*x[1]
        arg[0,1,2,0]=(-0.489161301965)*x[0]+(0.115368605994)*x[1]
        arg[0,1,2,1]=(0.00167228618616)*x[0]+(-0.636896504873)*x[1]
        arg[0,2,0,0]=(-0.239182729045)*x[0]+(0.470786327067)*x[1]
        arg[0,2,0,1]=(0.182105943188)*x[0]+(0.328443662342)*x[1]
        arg[0,2,1,0]=(0.000476540430199)*x[0]+(-0.498066512126)*x[1]
        arg[0,2,1,1]=(-0.822985659305)*x[0]+(-0.118645880185)*x[1]
        arg[0,2,2,0]=(1.18375382554)*x[0]+(1.51417307474)*x[1]
        arg[0,2,2,1]=(-0.99755868797)*x[0]+(-1.4187068658)*x[1]
        arg[0,3,0,0]=(-0.0908782314954)*x[0]+(-0.25901340471)*x[1]
        arg[0,3,0,1]=(-1.77626335863)*x[0]+(-0.862132832663)*x[1]
        arg[0,3,1,0]=(0.532147327491)*x[0]+(0.454949026506)*x[1]
        arg[0,3,1,1]=(-0.882307958628)*x[0]+(0.00780544746858)*x[1]
        arg[0,3,2,0]=(0.700104420574)*x[0]+(0.720508036709)*x[1]
        arg[0,3,2,1]=(-1.27167118811)*x[0]+(-0.372999410908)*x[1]
        arg[0,4,0,0]=(0.396222903345)*x[0]+(1.70664799562)*x[1]
        arg[0,4,0,1]=(0.229159055372)*x[0]+(0.0162677758439)*x[1]
        arg[0,4,1,0]=(1.69470499818)*x[0]+(0.0971804891468)*x[1]
        arg[0,4,1,1]=(0.180710573838)*x[0]+(1.05311557127)*x[1]
        arg[0,4,2,0]=(-0.072661496917)*x[0]+(0.466256643644)*x[1]
        arg[0,4,2,1]=(0.25755631084)*x[0]+(0.306990925808)*x[1]
        arg[1,0,0,0]=(0.835353191317)*x[0]+(-0.619237060145)*x[1]
        arg[1,0,0,1]=(1.02966866371)*x[0]+(0.327724003031)*x[1]
        arg[1,0,1,0]=(-0.59214788986)*x[0]+(-0.927738485428)*x[1]
        arg[1,0,1,1]=(-0.805929278446)*x[0]+(-0.366113927406)*x[1]
        arg[1,0,2,0]=(0.952876155721)*x[0]+(0.490531201467)*x[1]
        arg[1,0,2,1]=(1.44341651657)*x[0]+(-0.269899260772)*x[1]
        arg[1,1,0,0]=(0.302587966785)*x[0]+(-0.0373883991695)*x[1]
        arg[1,1,0,1]=(0.77498995648)*x[0]+(1.81052574555)*x[1]
        arg[1,1,1,0]=(1.47950560134)*x[0]+(-0.714878219566)*x[1]
        arg[1,1,1,1]=(0.0713827629606)*x[0]+(-1.22018291576)*x[1]
        arg[1,1,2,0]=(1.30715631604)*x[0]+(-0.191666253513)*x[1]
        arg[1,1,2,1]=(0.330921962628)*x[0]+(-1.57160261912)*x[1]
        arg[1,2,0,0]=(0.593477126597)*x[0]+(0.728196634765)*x[1]
        arg[1,2,0,1]=(0.853860342143)*x[0]+(0.142625101433)*x[1]
        arg[1,2,1,0]=(-0.743590115343)*x[0]+(0.108893736366)*x[1]
        arg[1,2,1,1]=(-0.651514863459)*x[0]+(0.365517180762)*x[1]
        arg[1,2,2,0]=(1.491516719)*x[0]+(0.28900806658)*x[1]
        arg[1,2,2,1]=(1.23733736339)*x[0]+(-0.58438294882)*x[1]
        arg[1,3,0,0]=(0.185696018505)*x[0]+(-1.29119084907)*x[1]
        arg[1,3,0,1]=(1.73584656085)*x[0]+(0.720989518003)*x[1]
        arg[1,3,1,0]=(-1.16044192371)*x[0]+(-0.131128994204)*x[1]
        arg[1,3,1,1]=(-0.00913995119508)*x[0]+(0.506516682557)*x[1]
        arg[1,3,2,0]=(-1.4970171408)*x[0]+(0.692262679219)*x[1]
        arg[1,3,2,1]=(0.433442352395)*x[0]+(-0.118782513258)*x[1]
        arg[1,4,0,0]=(0.186272280804)*x[0]+(-0.136627137097)*x[1]
        arg[1,4,0,1]=(1.25477168515)*x[0]+(-0.0871831369474)*x[1]
        arg[1,4,1,0]=(1.22746022677)*x[0]+(0.799329883559)*x[1]
        arg[1,4,1,1]=(-0.756298280123)*x[0]+(0.317500967634)*x[1]
        arg[1,4,2,0]=(0.201184703435)*x[0]+(-0.116537935236)*x[1]
        arg[1,4,2,1]=(0.851291897607)*x[0]+(0.404063606516)*x[1]
        arg[2,0,0,0]=(-0.257605370026)*x[0]+(-0.202424701671)*x[1]
        arg[2,0,0,1]=(0.0219332720675)*x[0]+(0.518222103227)*x[1]
        arg[2,0,1,0]=(1.03726036144)*x[0]+(-1.01482867596)*x[1]
        arg[2,0,1,1]=(0.00677025785109)*x[0]+(0.607932484906)*x[1]
        arg[2,0,2,0]=(-0.753000974637)*x[0]+(0.112884693676)*x[1]
        arg[2,0,2,1]=(-0.585627190582)*x[0]+(0.548136242722)*x[1]
        arg[2,1,0,0]=(0.382466685883)*x[0]+(-1.15520454298)*x[1]
        arg[2,1,0,1]=(-0.722705123323)*x[0]+(-0.599069291154)*x[1]
        arg[2,1,1,0]=(-0.845885208256)*x[0]+(-1.56066350212)*x[1]
        arg[2,1,1,1]=(-1.31472469553)*x[0]+(0.18973548452)*x[1]
        arg[2,1,2,0]=(-0.419142599733)*x[0]+(1.43873372527)*x[1]
        arg[2,1,2,1]=(0.22685530286)*x[0]+(-0.53636714393)*x[1]
        arg[2,2,0,0]=(1.32884214935)*x[0]+(-0.377575726395)*x[1]
        arg[2,2,0,1]=(0.0845542736847)*x[0]+(0.0888339000429)*x[1]
        arg[2,2,1,0]=(-1.05813285521)*x[0]+(-0.980382395463)*x[1]
        arg[2,2,1,1]=(0.306785463268)*x[0]+(0.622558337967)*x[1]
        arg[2,2,2,0]=(-0.765475308614)*x[0]+(1.40320055807)*x[1]
        arg[2,2,2,1]=(-1.23199081441)*x[0]+(0.245517841221)*x[1]
        arg[2,3,0,0]=(0.655775411743)*x[0]+(-1.09496249546)*x[1]
        arg[2,3,0,1]=(-1.57906394006)*x[0]+(-0.223321547826)*x[1]
        arg[2,3,1,0]=(-0.329934697034)*x[0]+(-1.36492062074)*x[1]
        arg[2,3,1,1]=(1.2584701256)*x[0]+(0.172958559745)*x[1]
        arg[2,3,2,0]=(0.265769150885)*x[0]+(-0.273148332255)*x[1]
        arg[2,3,2,1]=(-1.18294356647)*x[0]+(-0.269813147795)*x[1]
        arg[2,4,0,0]=(-1.80352122024)*x[0]+(-0.417693287507)*x[1]
        arg[2,4,0,1]=(0.829979623343)*x[0]+(0.0526239642579)*x[1]
        arg[2,4,1,0]=(0.967999103063)*x[0]+(0.0184170526665)*x[1]
        arg[2,4,1,1]=(-0.326938715134)*x[0]+(-0.575040647639)*x[1]
        arg[2,4,2,0]=(1.44727400099)*x[0]+(0.710269500685)*x[1]
        arg[2,4,2,1]=(-0.248166665809)*x[0]+(0.247820501661)*x[1]
        arg[3,0,0,0]=(0.0463415093786)*x[0]+(0.650085996364)*x[1]
        arg[3,0,0,1]=(-0.440939509697)*x[0]+(-0.323307669473)*x[1]
        arg[3,0,1,0]=(0.273356160699)*x[0]+(-0.66721600197)*x[1]
        arg[3,0,1,1]=(1.29465477714)*x[0]+(-0.37486215111)*x[1]
        arg[3,0,2,0]=(-0.809896216106)*x[0]+(-1.51076103892)*x[1]
        arg[3,0,2,1]=(0.686428827303)*x[0]+(-0.538164742287)*x[1]
        arg[3,1,0,0]=(-0.863651610678)*x[0]+(0.673803453645)*x[1]
        arg[3,1,0,1]=(0.526827519816)*x[0]+(-0.0379595314201)*x[1]
        arg[3,1,1,0]=(-1.04759557033)*x[0]+(0.114198098457)*x[1]
        arg[3,1,1,1]=(0.552205171704)*x[0]+(0.116364312378)*x[1]
        arg[3,1,2,0]=(-0.926549911354)*x[0]+(0.296927704155)*x[1]
        arg[3,1,2,1]=(0.46702104365)*x[0]+(0.355015968814)*x[1]
        arg[3,2,0,0]=(0.603177046544)*x[0]+(-0.600600764625)*x[1]
        arg[3,2,0,1]=(0.656827398088)*x[0]+(1.19922812106)*x[1]
        arg[3,2,1,0]=(-0.0185352951675)*x[0]+(-0.575006610999)*x[1]
        arg[3,2,1,1]=(0.954361462974)*x[0]+(-0.239751892125)*x[1]
        arg[3,2,2,0]=(0.880568376655)*x[0]+(1.46879509876)*x[1]
        arg[3,2,2,1]=(-1.19051913627)*x[0]+(-0.0387119361514)*x[1]
        arg[3,3,0,0]=(1.19520855834)*x[0]+(1.3515728248)*x[1]
        arg[3,3,0,1]=(0.695487159903)*x[0]+(0.783460154378)*x[1]
        arg[3,3,1,0]=(0.482352542106)*x[0]+(-1.321707654)*x[1]
        arg[3,3,1,1]=(0.0637384915311)*x[0]+(-0.5903067787)*x[1]
        arg[3,3,2,0]=(0.571974795979)*x[0]+(1.33251727342)*x[1]
        arg[3,3,2,1]=(0.579723470137)*x[0]+(-0.260885275522)*x[1]
        arg[3,4,0,0]=(0.0717612359082)*x[0]+(-0.576982745638)*x[1]
        arg[3,4,0,1]=(-0.227660227523)*x[0]+(0.232974094548)*x[1]
        arg[3,4,1,0]=(0.767797269581)*x[0]+(0.98857043831)*x[1]
        arg[3,4,1,1]=(0.31200408415)*x[0]+(-0.8317658048)*x[1]
        arg[3,4,2,0]=(0.381688388928)*x[0]+(-0.335620162467)*x[1]
        arg[3,4,2,1]=(0.781228510591)*x[0]+(0.960966953643)*x[1]
        ref[0,0,0,0]=(0.507324804235)*x_ref[0]+(0.725831700089)*x_ref[1]
        ref[0,0,0,1]=(0.847901454936)*x_ref[0]+(0.84787024099)*x_ref[1]
        ref[0,0,1,0]=(0.221488465784)*x_ref[0]+(0.0449286988708)*x_ref[1]
        ref[0,0,1,1]=(0.00897176578501)*x_ref[0]+(-0.799118487206)*x_ref[1]
        ref[0,0,2,0]=(0.171491682599)*x_ref[0]+(1.0259065872)*x_ref[1]
        ref[0,0,2,1]=(-0.213244683798)*x_ref[0]+(-0.706816385516)*x_ref[1]
        ref[0,1,0,0]=(1.54199875186)*x_ref[0]+(-1.09479402082)*x_ref[1]
        ref[0,1,0,1]=(-0.755278051349)*x_ref[0]+(-0.371785089226)*x_ref[1]
        ref[0,1,1,0]=(-0.798931370137)*x_ref[0]+(-0.716047155117)*x_ref[1]
        ref[0,1,1,1]=(-0.79104072521)*x_ref[0]+(-1.45842710768)*x_ref[1]
        ref[0,1,2,0]=(-0.489161301965)*x_ref[0]+(0.115368605994)*x_ref[1]
        ref[0,1,2,1]=(0.00167228618616)*x_ref[0]+(-0.636896504873)*x_ref[1]
        ref[0,2,0,0]=(-0.239182729045)*x_ref[0]+(0.470786327067)*x_ref[1]
        ref[0,2,0,1]=(0.182105943188)*x_ref[0]+(0.328443662342)*x_ref[1]
        ref[0,2,1,0]=(0.000476540430199)*x_ref[0]+(-0.498066512126)*x_ref[1]
        ref[0,2,1,1]=(-0.822985659305)*x_ref[0]+(-0.118645880185)*x_ref[1]
        ref[0,2,2,0]=(1.18375382554)*x_ref[0]+(1.51417307474)*x_ref[1]
        ref[0,2,2,1]=(-0.99755868797)*x_ref[0]+(-1.4187068658)*x_ref[1]
        ref[0,3,0,0]=(-0.0908782314954)*x_ref[0]+(-0.25901340471)*x_ref[1]
        ref[0,3,0,1]=(-1.77626335863)*x_ref[0]+(-0.862132832663)*x_ref[1]
        ref[0,3,1,0]=(0.532147327491)*x_ref[0]+(0.454949026506)*x_ref[1]
        ref[0,3,1,1]=(-0.882307958628)*x_ref[0]+(0.00780544746858)*x_ref[1]
        ref[0,3,2,0]=(0.700104420574)*x_ref[0]+(0.720508036709)*x_ref[1]
        ref[0,3,2,1]=(-1.27167118811)*x_ref[0]+(-0.372999410908)*x_ref[1]
        ref[0,4,0,0]=(0.396222903345)*x_ref[0]+(1.70664799562)*x_ref[1]
        ref[0,4,0,1]=(0.229159055372)*x_ref[0]+(0.0162677758439)*x_ref[1]
        ref[0,4,1,0]=(1.69470499818)*x_ref[0]+(0.0971804891468)*x_ref[1]
        ref[0,4,1,1]=(0.180710573838)*x_ref[0]+(1.05311557127)*x_ref[1]
        ref[0,4,2,0]=(-0.072661496917)*x_ref[0]+(0.466256643644)*x_ref[1]
        ref[0,4,2,1]=(0.25755631084)*x_ref[0]+(0.306990925808)*x_ref[1]
        ref[1,0,0,0]=(0.835353191317)*x_ref[0]+(-0.619237060145)*x_ref[1]
        ref[1,0,0,1]=(1.02966866371)*x_ref[0]+(0.327724003031)*x_ref[1]
        ref[1,0,1,0]=(-0.59214788986)*x_ref[0]+(-0.927738485428)*x_ref[1]
        ref[1,0,1,1]=(-0.805929278446)*x_ref[0]+(-0.366113927406)*x_ref[1]
        ref[1,0,2,0]=(0.952876155721)*x_ref[0]+(0.490531201467)*x_ref[1]
        ref[1,0,2,1]=(1.44341651657)*x_ref[0]+(-0.269899260772)*x_ref[1]
        ref[1,1,0,0]=(0.302587966785)*x_ref[0]+(-0.0373883991695)*x_ref[1]
        ref[1,1,0,1]=(0.77498995648)*x_ref[0]+(1.81052574555)*x_ref[1]
        ref[1,1,1,0]=(1.47950560134)*x_ref[0]+(-0.714878219566)*x_ref[1]
        ref[1,1,1,1]=(0.0713827629606)*x_ref[0]+(-1.22018291576)*x_ref[1]
        ref[1,1,2,0]=(1.30715631604)*x_ref[0]+(-0.191666253513)*x_ref[1]
        ref[1,1,2,1]=(0.330921962628)*x_ref[0]+(-1.57160261912)*x_ref[1]
        ref[1,2,0,0]=(0.593477126597)*x_ref[0]+(0.728196634765)*x_ref[1]
        ref[1,2,0,1]=(0.853860342143)*x_ref[0]+(0.142625101433)*x_ref[1]
        ref[1,2,1,0]=(-0.743590115343)*x_ref[0]+(0.108893736366)*x_ref[1]
        ref[1,2,1,1]=(-0.651514863459)*x_ref[0]+(0.365517180762)*x_ref[1]
        ref[1,2,2,0]=(1.491516719)*x_ref[0]+(0.28900806658)*x_ref[1]
        ref[1,2,2,1]=(1.23733736339)*x_ref[0]+(-0.58438294882)*x_ref[1]
        ref[1,3,0,0]=(0.185696018505)*x_ref[0]+(-1.29119084907)*x_ref[1]
        ref[1,3,0,1]=(1.73584656085)*x_ref[0]+(0.720989518003)*x_ref[1]
        ref[1,3,1,0]=(-1.16044192371)*x_ref[0]+(-0.131128994204)*x_ref[1]
        ref[1,3,1,1]=(-0.00913995119508)*x_ref[0]+(0.506516682557)*x_ref[1]
        ref[1,3,2,0]=(-1.4970171408)*x_ref[0]+(0.692262679219)*x_ref[1]
        ref[1,3,2,1]=(0.433442352395)*x_ref[0]+(-0.118782513258)*x_ref[1]
        ref[1,4,0,0]=(0.186272280804)*x_ref[0]+(-0.136627137097)*x_ref[1]
        ref[1,4,0,1]=(1.25477168515)*x_ref[0]+(-0.0871831369474)*x_ref[1]
        ref[1,4,1,0]=(1.22746022677)*x_ref[0]+(0.799329883559)*x_ref[1]
        ref[1,4,1,1]=(-0.756298280123)*x_ref[0]+(0.317500967634)*x_ref[1]
        ref[1,4,2,0]=(0.201184703435)*x_ref[0]+(-0.116537935236)*x_ref[1]
        ref[1,4,2,1]=(0.851291897607)*x_ref[0]+(0.404063606516)*x_ref[1]
        ref[2,0,0,0]=(-0.257605370026)*x_ref[0]+(-0.202424701671)*x_ref[1]
        ref[2,0,0,1]=(0.0219332720675)*x_ref[0]+(0.518222103227)*x_ref[1]
        ref[2,0,1,0]=(1.03726036144)*x_ref[0]+(-1.01482867596)*x_ref[1]
        ref[2,0,1,1]=(0.00677025785109)*x_ref[0]+(0.607932484906)*x_ref[1]
        ref[2,0,2,0]=(-0.753000974637)*x_ref[0]+(0.112884693676)*x_ref[1]
        ref[2,0,2,1]=(-0.585627190582)*x_ref[0]+(0.548136242722)*x_ref[1]
        ref[2,1,0,0]=(0.382466685883)*x_ref[0]+(-1.15520454298)*x_ref[1]
        ref[2,1,0,1]=(-0.722705123323)*x_ref[0]+(-0.599069291154)*x_ref[1]
        ref[2,1,1,0]=(-0.845885208256)*x_ref[0]+(-1.56066350212)*x_ref[1]
        ref[2,1,1,1]=(-1.31472469553)*x_ref[0]+(0.18973548452)*x_ref[1]
        ref[2,1,2,0]=(-0.419142599733)*x_ref[0]+(1.43873372527)*x_ref[1]
        ref[2,1,2,1]=(0.22685530286)*x_ref[0]+(-0.53636714393)*x_ref[1]
        ref[2,2,0,0]=(1.32884214935)*x_ref[0]+(-0.377575726395)*x_ref[1]
        ref[2,2,0,1]=(0.0845542736847)*x_ref[0]+(0.0888339000429)*x_ref[1]
        ref[2,2,1,0]=(-1.05813285521)*x_ref[0]+(-0.980382395463)*x_ref[1]
        ref[2,2,1,1]=(0.306785463268)*x_ref[0]+(0.622558337967)*x_ref[1]
        ref[2,2,2,0]=(-0.765475308614)*x_ref[0]+(1.40320055807)*x_ref[1]
        ref[2,2,2,1]=(-1.23199081441)*x_ref[0]+(0.245517841221)*x_ref[1]
        ref[2,3,0,0]=(0.655775411743)*x_ref[0]+(-1.09496249546)*x_ref[1]
        ref[2,3,0,1]=(-1.57906394006)*x_ref[0]+(-0.223321547826)*x_ref[1]
        ref[2,3,1,0]=(-0.329934697034)*x_ref[0]+(-1.36492062074)*x_ref[1]
        ref[2,3,1,1]=(1.2584701256)*x_ref[0]+(0.172958559745)*x_ref[1]
        ref[2,3,2,0]=(0.265769150885)*x_ref[0]+(-0.273148332255)*x_ref[1]
        ref[2,3,2,1]=(-1.18294356647)*x_ref[0]+(-0.269813147795)*x_ref[1]
        ref[2,4,0,0]=(-1.80352122024)*x_ref[0]+(-0.417693287507)*x_ref[1]
        ref[2,4,0,1]=(0.829979623343)*x_ref[0]+(0.0526239642579)*x_ref[1]
        ref[2,4,1,0]=(0.967999103063)*x_ref[0]+(0.0184170526665)*x_ref[1]
        ref[2,4,1,1]=(-0.326938715134)*x_ref[0]+(-0.575040647639)*x_ref[1]
        ref[2,4,2,0]=(1.44727400099)*x_ref[0]+(0.710269500685)*x_ref[1]
        ref[2,4,2,1]=(-0.248166665809)*x_ref[0]+(0.247820501661)*x_ref[1]
        ref[3,0,0,0]=(0.0463415093786)*x_ref[0]+(0.650085996364)*x_ref[1]
        ref[3,0,0,1]=(-0.440939509697)*x_ref[0]+(-0.323307669473)*x_ref[1]
        ref[3,0,1,0]=(0.273356160699)*x_ref[0]+(-0.66721600197)*x_ref[1]
        ref[3,0,1,1]=(1.29465477714)*x_ref[0]+(-0.37486215111)*x_ref[1]
        ref[3,0,2,0]=(-0.809896216106)*x_ref[0]+(-1.51076103892)*x_ref[1]
        ref[3,0,2,1]=(0.686428827303)*x_ref[0]+(-0.538164742287)*x_ref[1]
        ref[3,1,0,0]=(-0.863651610678)*x_ref[0]+(0.673803453645)*x_ref[1]
        ref[3,1,0,1]=(0.526827519816)*x_ref[0]+(-0.0379595314201)*x_ref[1]
        ref[3,1,1,0]=(-1.04759557033)*x_ref[0]+(0.114198098457)*x_ref[1]
        ref[3,1,1,1]=(0.552205171704)*x_ref[0]+(0.116364312378)*x_ref[1]
        ref[3,1,2,0]=(-0.926549911354)*x_ref[0]+(0.296927704155)*x_ref[1]
        ref[3,1,2,1]=(0.46702104365)*x_ref[0]+(0.355015968814)*x_ref[1]
        ref[3,2,0,0]=(0.603177046544)*x_ref[0]+(-0.600600764625)*x_ref[1]
        ref[3,2,0,1]=(0.656827398088)*x_ref[0]+(1.19922812106)*x_ref[1]
        ref[3,2,1,0]=(-0.0185352951675)*x_ref[0]+(-0.575006610999)*x_ref[1]
        ref[3,2,1,1]=(0.954361462974)*x_ref[0]+(-0.239751892125)*x_ref[1]
        ref[3,2,2,0]=(0.880568376655)*x_ref[0]+(1.46879509876)*x_ref[1]
        ref[3,2,2,1]=(-1.19051913627)*x_ref[0]+(-0.0387119361514)*x_ref[1]
        ref[3,3,0,0]=(1.19520855834)*x_ref[0]+(1.3515728248)*x_ref[1]
        ref[3,3,0,1]=(0.695487159903)*x_ref[0]+(0.783460154378)*x_ref[1]
        ref[3,3,1,0]=(0.482352542106)*x_ref[0]+(-1.321707654)*x_ref[1]
        ref[3,3,1,1]=(0.0637384915311)*x_ref[0]+(-0.5903067787)*x_ref[1]
        ref[3,3,2,0]=(0.571974795979)*x_ref[0]+(1.33251727342)*x_ref[1]
        ref[3,3,2,1]=(0.579723470137)*x_ref[0]+(-0.260885275522)*x_ref[1]
        ref[3,4,0,0]=(0.0717612359082)*x_ref[0]+(-0.576982745638)*x_ref[1]
        ref[3,4,0,1]=(-0.227660227523)*x_ref[0]+(0.232974094548)*x_ref[1]
        ref[3,4,1,0]=(0.767797269581)*x_ref[0]+(0.98857043831)*x_ref[1]
        ref[3,4,1,1]=(0.31200408415)*x_ref[0]+(-0.8317658048)*x_ref[1]
        ref[3,4,2,0]=(0.381688388928)*x_ref[0]+(-0.335620162467)*x_ref[1]
        ref[3,4,2,1]=(0.781228510591)*x_ref[0]+(0.960966953643)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.875410081875)*x[0]+(0.859961622438)*x[1]+(0.716805919947)*x[2]
        arg[0,0,0,1]=(0.343337453662)*x[0]+(0.802028072985)*x[1]+(0.588370750348)*x[2]
        arg[0,0,1,0]=(0.467615640242)*x[0]+(0.0457804028001)*x[1]+(-0.163677176662)*x[2]
        arg[0,0,1,1]=(-0.46378923457)*x[0]+(1.29379360824)*x[1]+(0.183839231748)*x[2]
        arg[0,0,2,0]=(1.65864647053)*x[0]+(1.23523675105)*x[1]+(-0.556155025527)*x[2]
        arg[0,0,2,1]=(-0.41581445233)*x[0]+(-0.0709880820438)*x[1]+(0.45785589934)*x[2]
        arg[0,1,0,0]=(1.83790068156)*x[0]+(-0.1690836925)*x[1]+(1.25356286366)*x[2]
        arg[0,1,0,1]=(-0.22584314958)*x[0]+(0.684595424346)*x[1]+(1.13919885217)*x[2]
        arg[0,1,1,0]=(1.02792150528)*x[0]+(-0.312817626145)*x[1]+(-0.121465758957)*x[2]
        arg[0,1,1,1]=(0.533491564413)*x[0]+(-0.62208812136)*x[1]+(-0.717947781071)*x[2]
        arg[0,1,2,0]=(-0.864505701555)*x[0]+(-0.296047724934)*x[1]+(-0.122917018588)*x[2]
        arg[0,1,2,1]=(1.05520768182)*x[0]+(0.662053590424)*x[1]+(0.0251841065476)*x[2]
        arg[0,2,0,0]=(-0.408305818721)*x[0]+(-0.355992587618)*x[1]+(-0.075231462678)*x[2]
        arg[0,2,0,1]=(0.926563898549)*x[0]+(0.203358704647)*x[1]+(-0.365269987808)*x[2]
        arg[0,2,1,0]=(-0.403860257745)*x[0]+(-0.0312835722524)*x[1]+(0.241379582617)*x[2]
        arg[0,2,1,1]=(-0.287297815371)*x[0]+(-0.381813267253)*x[1]+(0.46047051318)*x[2]
        arg[0,2,2,0]=(-0.168164506906)*x[0]+(-1.24152426841)*x[1]+(-0.629704769515)*x[2]
        arg[0,2,2,1]=(-0.937219354291)*x[0]+(-0.736950161299)*x[1]+(0.284057170092)*x[2]
        arg[0,3,0,0]=(-0.292205445759)*x[0]+(1.21842674086)*x[1]+(0.608979796402)*x[2]
        arg[0,3,0,1]=(0.53232240924)*x[0]+(0.226439017439)*x[1]+(-0.42603939944)*x[2]
        arg[0,3,1,0]=(1.67410358342)*x[0]+(-0.00813619771347)*x[1]+(-0.463361889732)*x[2]
        arg[0,3,1,1]=(1.3991502461)*x[0]+(0.239999857694)*x[1]+(-0.369907067208)*x[2]
        arg[0,3,2,0]=(-0.248767913403)*x[0]+(0.893584158229)*x[1]+(1.97656682548)*x[2]
        arg[0,3,2,1]=(-0.835046144341)*x[0]+(-0.00979850832247)*x[1]+(-0.281439904932)*x[2]
        arg[0,4,0,0]=(0.974179494371)*x[0]+(-0.0432121220464)*x[1]+(-0.475018088554)*x[2]
        arg[0,4,0,1]=(0.245490151034)*x[0]+(-0.367572047585)*x[1]+(-0.926123425123)*x[2]
        arg[0,4,1,0]=(-0.239057491277)*x[0]+(0.339734302574)*x[1]+(0.760546363123)*x[2]
        arg[0,4,1,1]=(-1.0425426898)*x[0]+(0.774210702874)*x[1]+(-1.15990543484)*x[2]
        arg[0,4,2,0]=(1.02994063274)*x[0]+(1.69313195574)*x[1]+(-1.02468862318)*x[2]
        arg[0,4,2,1]=(0.221805673731)*x[0]+(-1.06588014758)*x[1]+(-0.546973870537)*x[2]
        arg[1,0,0,0]=(-0.140765393552)*x[0]+(-0.309057115268)*x[1]+(0.07814948965)*x[2]
        arg[1,0,0,1]=(-0.29622101259)*x[0]+(-1.0654452087)*x[1]+(1.77500323701)*x[2]
        arg[1,0,1,0]=(1.13202947738)*x[0]+(-0.515454892388)*x[1]+(-1.34230346013)*x[2]
        arg[1,0,1,1]=(1.50210238409)*x[0]+(-0.75673088573)*x[1]+(-0.792078632258)*x[2]
        arg[1,0,2,0]=(0.249111383954)*x[0]+(-0.972418007776)*x[1]+(0.766060429469)*x[2]
        arg[1,0,2,1]=(-0.0710717292409)*x[0]+(0.164949706666)*x[1]+(1.46045331387)*x[2]
        arg[1,1,0,0]=(-0.587668209291)*x[0]+(0.303249169818)*x[1]+(0.352197808782)*x[2]
        arg[1,1,0,1]=(-0.145609904248)*x[0]+(0.91895760024)*x[1]+(-1.25063091478)*x[2]
        arg[1,1,1,0]=(0.450869946651)*x[0]+(0.752704827905)*x[1]+(0.132042943533)*x[2]
        arg[1,1,1,1]=(-0.87296460294)*x[0]+(-0.459840609019)*x[1]+(0.422574948491)*x[2]
        arg[1,1,2,0]=(0.828459269131)*x[0]+(0.45869153936)*x[1]+(0.854640684176)*x[2]
        arg[1,1,2,1]=(1.84720399473)*x[0]+(0.0416642546234)*x[1]+(-0.309583524852)*x[2]
        arg[1,2,0,0]=(-1.12563293246)*x[0]+(0.0142184473972)*x[1]+(-1.18150362584)*x[2]
        arg[1,2,0,1]=(-0.0750965768147)*x[0]+(-0.419276687751)*x[1]+(0.521012213863)*x[2]
        arg[1,2,1,0]=(-1.06905807284)*x[0]+(0.692258030743)*x[1]+(0.262254909068)*x[2]
        arg[1,2,1,1]=(0.0552441016888)*x[0]+(-0.131040560322)*x[1]+(-0.321517725963)*x[2]
        arg[1,2,2,0]=(-0.740410514693)*x[0]+(0.219020143899)*x[1]+(1.35104682965)*x[2]
        arg[1,2,2,1]=(-1.38198583939)*x[0]+(-0.19342887842)*x[1]+(-0.780008129494)*x[2]
        arg[1,3,0,0]=(0.378358305605)*x[0]+(1.32138618734)*x[1]+(0.627209784997)*x[2]
        arg[1,3,0,1]=(-0.0383657917537)*x[0]+(-1.41770189628)*x[1]+(-1.21625280681)*x[2]
        arg[1,3,1,0]=(0.819666413875)*x[0]+(0.160369878383)*x[1]+(-0.229388558682)*x[2]
        arg[1,3,1,1]=(0.393409555876)*x[0]+(-0.525985341296)*x[1]+(-1.11892438309)*x[2]
        arg[1,3,2,0]=(0.145077010777)*x[0]+(0.229800008438)*x[1]+(-0.202652862622)*x[2]
        arg[1,3,2,1]=(-1.48698378177)*x[0]+(0.553698672636)*x[1]+(-0.427263736796)*x[2]
        arg[1,4,0,0]=(-0.129527725249)*x[0]+(-0.639824801211)*x[1]+(0.759081980717)*x[2]
        arg[1,4,0,1]=(-0.117772473784)*x[0]+(0.387474104611)*x[1]+(0.507306747135)*x[2]
        arg[1,4,1,0]=(-0.82518622198)*x[0]+(-0.754474320343)*x[1]+(-0.978049152511)*x[2]
        arg[1,4,1,1]=(-0.370385633483)*x[0]+(-0.348429423043)*x[1]+(-0.254278257899)*x[2]
        arg[1,4,2,0]=(1.42596167697)*x[0]+(-0.871785904514)*x[1]+(-0.5343959325)*x[2]
        arg[1,4,2,1]=(0.817877672506)*x[0]+(-0.222622791737)*x[1]+(-0.677237468595)*x[2]
        arg[2,0,0,0]=(0.877660484813)*x[0]+(0.818663846913)*x[1]+(-1.26520874625)*x[2]
        arg[2,0,0,1]=(-0.822344385073)*x[0]+(-0.274691696266)*x[1]+(-0.35079248697)*x[2]
        arg[2,0,1,0]=(-0.319040285312)*x[0]+(-0.750406131559)*x[1]+(-0.920332967032)*x[2]
        arg[2,0,1,1]=(1.48758278445)*x[0]+(1.30587557873)*x[1]+(0.230692557888)*x[2]
        arg[2,0,2,0]=(1.44269354043)*x[0]+(-0.63270356894)*x[1]+(0.229323632455)*x[2]
        arg[2,0,2,1]=(-0.645065877536)*x[0]+(-0.393083499676)*x[1]+(-0.305117645453)*x[2]
        arg[2,1,0,0]=(-0.63304134143)*x[0]+(0.0664768401917)*x[1]+(0.87587743066)*x[2]
        arg[2,1,0,1]=(-0.531856958285)*x[0]+(0.747953146648)*x[1]+(1.23560779546)*x[2]
        arg[2,1,1,0]=(0.0564479101025)*x[0]+(0.546307803332)*x[1]+(0.655044890368)*x[2]
        arg[2,1,1,1]=(-1.12793057228)*x[0]+(-0.736676194013)*x[1]+(0.0375913829411)*x[2]
        arg[2,1,2,0]=(-0.531384188067)*x[0]+(0.249416904081)*x[1]+(0.711430712986)*x[2]
        arg[2,1,2,1]=(0.884244664065)*x[0]+(0.456066941968)*x[1]+(0.181630861785)*x[2]
        arg[2,2,0,0]=(0.0844543345062)*x[0]+(-0.435288041256)*x[1]+(-0.434816763062)*x[2]
        arg[2,2,0,1]=(0.373388105049)*x[0]+(0.239833846429)*x[1]+(-0.437771573282)*x[2]
        arg[2,2,1,0]=(0.534401521907)*x[0]+(0.393077203448)*x[1]+(-0.42440387717)*x[2]
        arg[2,2,1,1]=(0.318683369617)*x[0]+(-0.538864643793)*x[1]+(1.83466945806)*x[2]
        arg[2,2,2,0]=(-1.79048122074)*x[0]+(-1.05441213478)*x[1]+(-0.573963806281)*x[2]
        arg[2,2,2,1]=(0.944017405182)*x[0]+(0.772756791538)*x[1]+(0.0269673237336)*x[2]
        arg[2,3,0,0]=(0.989703689142)*x[0]+(1.21302325945)*x[1]+(-0.92209419073)*x[2]
        arg[2,3,0,1]=(-0.0957244757576)*x[0]+(1.50132638315)*x[1]+(-1.38716099756)*x[2]
        arg[2,3,1,0]=(1.6445441776)*x[0]+(-0.490207070694)*x[1]+(0.626469347781)*x[2]
        arg[2,3,1,1]=(0.0185568939147)*x[0]+(0.412024752941)*x[1]+(-0.22626102813)*x[2]
        arg[2,3,2,0]=(0.0143609822966)*x[0]+(0.115739025893)*x[1]+(-0.544468631925)*x[2]
        arg[2,3,2,1]=(-1.80522341493)*x[0]+(0.33352152695)*x[1]+(-0.841286464932)*x[2]
        arg[2,4,0,0]=(0.0618444122042)*x[0]+(1.91232484622)*x[1]+(-0.406317200023)*x[2]
        arg[2,4,0,1]=(0.279131062922)*x[0]+(0.0194232637634)*x[1]+(-0.00179069113634)*x[2]
        arg[2,4,1,0]=(-0.701344579135)*x[0]+(-0.27410935473)*x[1]+(0.763362100733)*x[2]
        arg[2,4,1,1]=(0.831198601185)*x[0]+(0.90336781771)*x[1]+(1.21856633382)*x[2]
        arg[2,4,2,0]=(0.526278615033)*x[0]+(-1.08673855801)*x[1]+(0.0784355900913)*x[2]
        arg[2,4,2,1]=(0.853226856851)*x[0]+(-1.15919062627)*x[1]+(0.185686031732)*x[2]
        arg[3,0,0,0]=(0.272127498117)*x[0]+(-0.457140162438)*x[1]+(0.318358670541)*x[2]
        arg[3,0,0,1]=(1.71056378307)*x[0]+(0.200052730234)*x[1]+(-0.127733325792)*x[2]
        arg[3,0,1,0]=(-0.285340963261)*x[0]+(-0.524544624385)*x[1]+(0.916366474944)*x[2]
        arg[3,0,1,1]=(-0.625506557557)*x[0]+(1.0858476479)*x[1]+(-1.24478983268)*x[2]
        arg[3,0,2,0]=(-1.37299619916)*x[0]+(-0.482246528202)*x[1]+(-1.36857669075)*x[2]
        arg[3,0,2,1]=(-0.924421844916)*x[0]+(-1.50402098115)*x[1]+(1.00045130673)*x[2]
        arg[3,1,0,0]=(0.0417304700827)*x[0]+(0.554527451004)*x[1]+(0.299758163628)*x[2]
        arg[3,1,0,1]=(0.639611639279)*x[0]+(-0.444675638162)*x[1]+(1.43150046757)*x[2]
        arg[3,1,1,0]=(0.288697220898)*x[0]+(1.32660433773)*x[1]+(0.0889412116759)*x[2]
        arg[3,1,1,1]=(-0.110516770201)*x[0]+(-0.0214248733565)*x[1]+(-0.758638794785)*x[2]
        arg[3,1,2,0]=(1.34357138654)*x[0]+(-0.13759272972)*x[1]+(-0.00183845032784)*x[2]
        arg[3,1,2,1]=(1.44174769627)*x[0]+(0.49190176388)*x[1]+(0.477331360141)*x[2]
        arg[3,2,0,0]=(-0.10284777121)*x[0]+(-0.63294157419)*x[1]+(0.246064460451)*x[2]
        arg[3,2,0,1]=(0.974270980699)*x[0]+(1.03835886161)*x[1]+(0.420000854838)*x[2]
        arg[3,2,1,0]=(-0.00617831615923)*x[0]+(-0.354063634073)*x[1]+(0.764285884619)*x[2]
        arg[3,2,1,1]=(0.210815790746)*x[0]+(0.76609515744)*x[1]+(-0.564401558401)*x[2]
        arg[3,2,2,0]=(1.30471234011)*x[0]+(1.54864119332)*x[1]+(-0.493583650519)*x[2]
        arg[3,2,2,1]=(0.42287513668)*x[0]+(-0.603093868171)*x[1]+(-0.811892974203)*x[2]
        arg[3,3,0,0]=(-1.38358134693)*x[0]+(-1.53706018098)*x[1]+(-1.29701227501)*x[2]
        arg[3,3,0,1]=(0.598581779902)*x[0]+(1.02424010922)*x[1]+(-1.1814313009)*x[2]
        arg[3,3,1,0]=(1.30938529116)*x[0]+(-0.504468477244)*x[1]+(1.63655848854)*x[2]
        arg[3,3,1,1]=(-1.47006424113)*x[0]+(0.225951520203)*x[1]+(-0.412273003169)*x[2]
        arg[3,3,2,0]=(-0.116902234482)*x[0]+(0.488144536044)*x[1]+(1.01732461694)*x[2]
        arg[3,3,2,1]=(1.0935814877)*x[0]+(1.06389385916)*x[1]+(-0.17981400591)*x[2]
        arg[3,4,0,0]=(-0.833265062101)*x[0]+(-0.754461148648)*x[1]+(1.26702012815)*x[2]
        arg[3,4,0,1]=(-0.793744278171)*x[0]+(0.112704521653)*x[1]+(-0.448684922187)*x[2]
        arg[3,4,1,0]=(-0.396153523625)*x[0]+(-0.0370870723568)*x[1]+(0.554347618511)*x[2]
        arg[3,4,1,1]=(-0.323276139711)*x[0]+(-0.397389928392)*x[1]+(-1.1707028324)*x[2]
        arg[3,4,2,0]=(0.823350512513)*x[0]+(1.47105759864)*x[1]+(-0.977456877855)*x[2]
        arg[3,4,2,1]=(-0.719791170741)*x[0]+(-1.00363340645)*x[1]+(0.814612770572)*x[2]
        ref[0,0,0,0]=(-0.875410081875)*x_ref[0]+(0.859961622438)*x_ref[1]+(0.716805919947)*x_ref[2]
        ref[0,0,0,1]=(0.343337453662)*x_ref[0]+(0.802028072985)*x_ref[1]+(0.588370750348)*x_ref[2]
        ref[0,0,1,0]=(0.467615640242)*x_ref[0]+(0.0457804028001)*x_ref[1]+(-0.163677176662)*x_ref[2]
        ref[0,0,1,1]=(-0.46378923457)*x_ref[0]+(1.29379360824)*x_ref[1]+(0.183839231748)*x_ref[2]
        ref[0,0,2,0]=(1.65864647053)*x_ref[0]+(1.23523675105)*x_ref[1]+(-0.556155025527)*x_ref[2]
        ref[0,0,2,1]=(-0.41581445233)*x_ref[0]+(-0.0709880820438)*x_ref[1]+(0.45785589934)*x_ref[2]
        ref[0,1,0,0]=(1.83790068156)*x_ref[0]+(-0.1690836925)*x_ref[1]+(1.25356286366)*x_ref[2]
        ref[0,1,0,1]=(-0.22584314958)*x_ref[0]+(0.684595424346)*x_ref[1]+(1.13919885217)*x_ref[2]
        ref[0,1,1,0]=(1.02792150528)*x_ref[0]+(-0.312817626145)*x_ref[1]+(-0.121465758957)*x_ref[2]
        ref[0,1,1,1]=(0.533491564413)*x_ref[0]+(-0.62208812136)*x_ref[1]+(-0.717947781071)*x_ref[2]
        ref[0,1,2,0]=(-0.864505701555)*x_ref[0]+(-0.296047724934)*x_ref[1]+(-0.122917018588)*x_ref[2]
        ref[0,1,2,1]=(1.05520768182)*x_ref[0]+(0.662053590424)*x_ref[1]+(0.0251841065476)*x_ref[2]
        ref[0,2,0,0]=(-0.408305818721)*x_ref[0]+(-0.355992587618)*x_ref[1]+(-0.075231462678)*x_ref[2]
        ref[0,2,0,1]=(0.926563898549)*x_ref[0]+(0.203358704647)*x_ref[1]+(-0.365269987808)*x_ref[2]
        ref[0,2,1,0]=(-0.403860257745)*x_ref[0]+(-0.0312835722524)*x_ref[1]+(0.241379582617)*x_ref[2]
        ref[0,2,1,1]=(-0.287297815371)*x_ref[0]+(-0.381813267253)*x_ref[1]+(0.46047051318)*x_ref[2]
        ref[0,2,2,0]=(-0.168164506906)*x_ref[0]+(-1.24152426841)*x_ref[1]+(-0.629704769515)*x_ref[2]
        ref[0,2,2,1]=(-0.937219354291)*x_ref[0]+(-0.736950161299)*x_ref[1]+(0.284057170092)*x_ref[2]
        ref[0,3,0,0]=(-0.292205445759)*x_ref[0]+(1.21842674086)*x_ref[1]+(0.608979796402)*x_ref[2]
        ref[0,3,0,1]=(0.53232240924)*x_ref[0]+(0.226439017439)*x_ref[1]+(-0.42603939944)*x_ref[2]
        ref[0,3,1,0]=(1.67410358342)*x_ref[0]+(-0.00813619771347)*x_ref[1]+(-0.463361889732)*x_ref[2]
        ref[0,3,1,1]=(1.3991502461)*x_ref[0]+(0.239999857694)*x_ref[1]+(-0.369907067208)*x_ref[2]
        ref[0,3,2,0]=(-0.248767913403)*x_ref[0]+(0.893584158229)*x_ref[1]+(1.97656682548)*x_ref[2]
        ref[0,3,2,1]=(-0.835046144341)*x_ref[0]+(-0.00979850832247)*x_ref[1]+(-0.281439904932)*x_ref[2]
        ref[0,4,0,0]=(0.974179494371)*x_ref[0]+(-0.0432121220464)*x_ref[1]+(-0.475018088554)*x_ref[2]
        ref[0,4,0,1]=(0.245490151034)*x_ref[0]+(-0.367572047585)*x_ref[1]+(-0.926123425123)*x_ref[2]
        ref[0,4,1,0]=(-0.239057491277)*x_ref[0]+(0.339734302574)*x_ref[1]+(0.760546363123)*x_ref[2]
        ref[0,4,1,1]=(-1.0425426898)*x_ref[0]+(0.774210702874)*x_ref[1]+(-1.15990543484)*x_ref[2]
        ref[0,4,2,0]=(1.02994063274)*x_ref[0]+(1.69313195574)*x_ref[1]+(-1.02468862318)*x_ref[2]
        ref[0,4,2,1]=(0.221805673731)*x_ref[0]+(-1.06588014758)*x_ref[1]+(-0.546973870537)*x_ref[2]
        ref[1,0,0,0]=(-0.140765393552)*x_ref[0]+(-0.309057115268)*x_ref[1]+(0.07814948965)*x_ref[2]
        ref[1,0,0,1]=(-0.29622101259)*x_ref[0]+(-1.0654452087)*x_ref[1]+(1.77500323701)*x_ref[2]
        ref[1,0,1,0]=(1.13202947738)*x_ref[0]+(-0.515454892388)*x_ref[1]+(-1.34230346013)*x_ref[2]
        ref[1,0,1,1]=(1.50210238409)*x_ref[0]+(-0.75673088573)*x_ref[1]+(-0.792078632258)*x_ref[2]
        ref[1,0,2,0]=(0.249111383954)*x_ref[0]+(-0.972418007776)*x_ref[1]+(0.766060429469)*x_ref[2]
        ref[1,0,2,1]=(-0.0710717292409)*x_ref[0]+(0.164949706666)*x_ref[1]+(1.46045331387)*x_ref[2]
        ref[1,1,0,0]=(-0.587668209291)*x_ref[0]+(0.303249169818)*x_ref[1]+(0.352197808782)*x_ref[2]
        ref[1,1,0,1]=(-0.145609904248)*x_ref[0]+(0.91895760024)*x_ref[1]+(-1.25063091478)*x_ref[2]
        ref[1,1,1,0]=(0.450869946651)*x_ref[0]+(0.752704827905)*x_ref[1]+(0.132042943533)*x_ref[2]
        ref[1,1,1,1]=(-0.87296460294)*x_ref[0]+(-0.459840609019)*x_ref[1]+(0.422574948491)*x_ref[2]
        ref[1,1,2,0]=(0.828459269131)*x_ref[0]+(0.45869153936)*x_ref[1]+(0.854640684176)*x_ref[2]
        ref[1,1,2,1]=(1.84720399473)*x_ref[0]+(0.0416642546234)*x_ref[1]+(-0.309583524852)*x_ref[2]
        ref[1,2,0,0]=(-1.12563293246)*x_ref[0]+(0.0142184473972)*x_ref[1]+(-1.18150362584)*x_ref[2]
        ref[1,2,0,1]=(-0.0750965768147)*x_ref[0]+(-0.419276687751)*x_ref[1]+(0.521012213863)*x_ref[2]
        ref[1,2,1,0]=(-1.06905807284)*x_ref[0]+(0.692258030743)*x_ref[1]+(0.262254909068)*x_ref[2]
        ref[1,2,1,1]=(0.0552441016888)*x_ref[0]+(-0.131040560322)*x_ref[1]+(-0.321517725963)*x_ref[2]
        ref[1,2,2,0]=(-0.740410514693)*x_ref[0]+(0.219020143899)*x_ref[1]+(1.35104682965)*x_ref[2]
        ref[1,2,2,1]=(-1.38198583939)*x_ref[0]+(-0.19342887842)*x_ref[1]+(-0.780008129494)*x_ref[2]
        ref[1,3,0,0]=(0.378358305605)*x_ref[0]+(1.32138618734)*x_ref[1]+(0.627209784997)*x_ref[2]
        ref[1,3,0,1]=(-0.0383657917537)*x_ref[0]+(-1.41770189628)*x_ref[1]+(-1.21625280681)*x_ref[2]
        ref[1,3,1,0]=(0.819666413875)*x_ref[0]+(0.160369878383)*x_ref[1]+(-0.229388558682)*x_ref[2]
        ref[1,3,1,1]=(0.393409555876)*x_ref[0]+(-0.525985341296)*x_ref[1]+(-1.11892438309)*x_ref[2]
        ref[1,3,2,0]=(0.145077010777)*x_ref[0]+(0.229800008438)*x_ref[1]+(-0.202652862622)*x_ref[2]
        ref[1,3,2,1]=(-1.48698378177)*x_ref[0]+(0.553698672636)*x_ref[1]+(-0.427263736796)*x_ref[2]
        ref[1,4,0,0]=(-0.129527725249)*x_ref[0]+(-0.639824801211)*x_ref[1]+(0.759081980717)*x_ref[2]
        ref[1,4,0,1]=(-0.117772473784)*x_ref[0]+(0.387474104611)*x_ref[1]+(0.507306747135)*x_ref[2]
        ref[1,4,1,0]=(-0.82518622198)*x_ref[0]+(-0.754474320343)*x_ref[1]+(-0.978049152511)*x_ref[2]
        ref[1,4,1,1]=(-0.370385633483)*x_ref[0]+(-0.348429423043)*x_ref[1]+(-0.254278257899)*x_ref[2]
        ref[1,4,2,0]=(1.42596167697)*x_ref[0]+(-0.871785904514)*x_ref[1]+(-0.5343959325)*x_ref[2]
        ref[1,4,2,1]=(0.817877672506)*x_ref[0]+(-0.222622791737)*x_ref[1]+(-0.677237468595)*x_ref[2]
        ref[2,0,0,0]=(0.877660484813)*x_ref[0]+(0.818663846913)*x_ref[1]+(-1.26520874625)*x_ref[2]
        ref[2,0,0,1]=(-0.822344385073)*x_ref[0]+(-0.274691696266)*x_ref[1]+(-0.35079248697)*x_ref[2]
        ref[2,0,1,0]=(-0.319040285312)*x_ref[0]+(-0.750406131559)*x_ref[1]+(-0.920332967032)*x_ref[2]
        ref[2,0,1,1]=(1.48758278445)*x_ref[0]+(1.30587557873)*x_ref[1]+(0.230692557888)*x_ref[2]
        ref[2,0,2,0]=(1.44269354043)*x_ref[0]+(-0.63270356894)*x_ref[1]+(0.229323632455)*x_ref[2]
        ref[2,0,2,1]=(-0.645065877536)*x_ref[0]+(-0.393083499676)*x_ref[1]+(-0.305117645453)*x_ref[2]
        ref[2,1,0,0]=(-0.63304134143)*x_ref[0]+(0.0664768401917)*x_ref[1]+(0.87587743066)*x_ref[2]
        ref[2,1,0,1]=(-0.531856958285)*x_ref[0]+(0.747953146648)*x_ref[1]+(1.23560779546)*x_ref[2]
        ref[2,1,1,0]=(0.0564479101025)*x_ref[0]+(0.546307803332)*x_ref[1]+(0.655044890368)*x_ref[2]
        ref[2,1,1,1]=(-1.12793057228)*x_ref[0]+(-0.736676194013)*x_ref[1]+(0.0375913829411)*x_ref[2]
        ref[2,1,2,0]=(-0.531384188067)*x_ref[0]+(0.249416904081)*x_ref[1]+(0.711430712986)*x_ref[2]
        ref[2,1,2,1]=(0.884244664065)*x_ref[0]+(0.456066941968)*x_ref[1]+(0.181630861785)*x_ref[2]
        ref[2,2,0,0]=(0.0844543345062)*x_ref[0]+(-0.435288041256)*x_ref[1]+(-0.434816763062)*x_ref[2]
        ref[2,2,0,1]=(0.373388105049)*x_ref[0]+(0.239833846429)*x_ref[1]+(-0.437771573282)*x_ref[2]
        ref[2,2,1,0]=(0.534401521907)*x_ref[0]+(0.393077203448)*x_ref[1]+(-0.42440387717)*x_ref[2]
        ref[2,2,1,1]=(0.318683369617)*x_ref[0]+(-0.538864643793)*x_ref[1]+(1.83466945806)*x_ref[2]
        ref[2,2,2,0]=(-1.79048122074)*x_ref[0]+(-1.05441213478)*x_ref[1]+(-0.573963806281)*x_ref[2]
        ref[2,2,2,1]=(0.944017405182)*x_ref[0]+(0.772756791538)*x_ref[1]+(0.0269673237336)*x_ref[2]
        ref[2,3,0,0]=(0.989703689142)*x_ref[0]+(1.21302325945)*x_ref[1]+(-0.92209419073)*x_ref[2]
        ref[2,3,0,1]=(-0.0957244757576)*x_ref[0]+(1.50132638315)*x_ref[1]+(-1.38716099756)*x_ref[2]
        ref[2,3,1,0]=(1.6445441776)*x_ref[0]+(-0.490207070694)*x_ref[1]+(0.626469347781)*x_ref[2]
        ref[2,3,1,1]=(0.0185568939147)*x_ref[0]+(0.412024752941)*x_ref[1]+(-0.22626102813)*x_ref[2]
        ref[2,3,2,0]=(0.0143609822966)*x_ref[0]+(0.115739025893)*x_ref[1]+(-0.544468631925)*x_ref[2]
        ref[2,3,2,1]=(-1.80522341493)*x_ref[0]+(0.33352152695)*x_ref[1]+(-0.841286464932)*x_ref[2]
        ref[2,4,0,0]=(0.0618444122042)*x_ref[0]+(1.91232484622)*x_ref[1]+(-0.406317200023)*x_ref[2]
        ref[2,4,0,1]=(0.279131062922)*x_ref[0]+(0.0194232637634)*x_ref[1]+(-0.00179069113634)*x_ref[2]
        ref[2,4,1,0]=(-0.701344579135)*x_ref[0]+(-0.27410935473)*x_ref[1]+(0.763362100733)*x_ref[2]
        ref[2,4,1,1]=(0.831198601185)*x_ref[0]+(0.90336781771)*x_ref[1]+(1.21856633382)*x_ref[2]
        ref[2,4,2,0]=(0.526278615033)*x_ref[0]+(-1.08673855801)*x_ref[1]+(0.0784355900913)*x_ref[2]
        ref[2,4,2,1]=(0.853226856851)*x_ref[0]+(-1.15919062627)*x_ref[1]+(0.185686031732)*x_ref[2]
        ref[3,0,0,0]=(0.272127498117)*x_ref[0]+(-0.457140162438)*x_ref[1]+(0.318358670541)*x_ref[2]
        ref[3,0,0,1]=(1.71056378307)*x_ref[0]+(0.200052730234)*x_ref[1]+(-0.127733325792)*x_ref[2]
        ref[3,0,1,0]=(-0.285340963261)*x_ref[0]+(-0.524544624385)*x_ref[1]+(0.916366474944)*x_ref[2]
        ref[3,0,1,1]=(-0.625506557557)*x_ref[0]+(1.0858476479)*x_ref[1]+(-1.24478983268)*x_ref[2]
        ref[3,0,2,0]=(-1.37299619916)*x_ref[0]+(-0.482246528202)*x_ref[1]+(-1.36857669075)*x_ref[2]
        ref[3,0,2,1]=(-0.924421844916)*x_ref[0]+(-1.50402098115)*x_ref[1]+(1.00045130673)*x_ref[2]
        ref[3,1,0,0]=(0.0417304700827)*x_ref[0]+(0.554527451004)*x_ref[1]+(0.299758163628)*x_ref[2]
        ref[3,1,0,1]=(0.639611639279)*x_ref[0]+(-0.444675638162)*x_ref[1]+(1.43150046757)*x_ref[2]
        ref[3,1,1,0]=(0.288697220898)*x_ref[0]+(1.32660433773)*x_ref[1]+(0.0889412116759)*x_ref[2]
        ref[3,1,1,1]=(-0.110516770201)*x_ref[0]+(-0.0214248733565)*x_ref[1]+(-0.758638794785)*x_ref[2]
        ref[3,1,2,0]=(1.34357138654)*x_ref[0]+(-0.13759272972)*x_ref[1]+(-0.00183845032784)*x_ref[2]
        ref[3,1,2,1]=(1.44174769627)*x_ref[0]+(0.49190176388)*x_ref[1]+(0.477331360141)*x_ref[2]
        ref[3,2,0,0]=(-0.10284777121)*x_ref[0]+(-0.63294157419)*x_ref[1]+(0.246064460451)*x_ref[2]
        ref[3,2,0,1]=(0.974270980699)*x_ref[0]+(1.03835886161)*x_ref[1]+(0.420000854838)*x_ref[2]
        ref[3,2,1,0]=(-0.00617831615923)*x_ref[0]+(-0.354063634073)*x_ref[1]+(0.764285884619)*x_ref[2]
        ref[3,2,1,1]=(0.210815790746)*x_ref[0]+(0.76609515744)*x_ref[1]+(-0.564401558401)*x_ref[2]
        ref[3,2,2,0]=(1.30471234011)*x_ref[0]+(1.54864119332)*x_ref[1]+(-0.493583650519)*x_ref[2]
        ref[3,2,2,1]=(0.42287513668)*x_ref[0]+(-0.603093868171)*x_ref[1]+(-0.811892974203)*x_ref[2]
        ref[3,3,0,0]=(-1.38358134693)*x_ref[0]+(-1.53706018098)*x_ref[1]+(-1.29701227501)*x_ref[2]
        ref[3,3,0,1]=(0.598581779902)*x_ref[0]+(1.02424010922)*x_ref[1]+(-1.1814313009)*x_ref[2]
        ref[3,3,1,0]=(1.30938529116)*x_ref[0]+(-0.504468477244)*x_ref[1]+(1.63655848854)*x_ref[2]
        ref[3,3,1,1]=(-1.47006424113)*x_ref[0]+(0.225951520203)*x_ref[1]+(-0.412273003169)*x_ref[2]
        ref[3,3,2,0]=(-0.116902234482)*x_ref[0]+(0.488144536044)*x_ref[1]+(1.01732461694)*x_ref[2]
        ref[3,3,2,1]=(1.0935814877)*x_ref[0]+(1.06389385916)*x_ref[1]+(-0.17981400591)*x_ref[2]
        ref[3,4,0,0]=(-0.833265062101)*x_ref[0]+(-0.754461148648)*x_ref[1]+(1.26702012815)*x_ref[2]
        ref[3,4,0,1]=(-0.793744278171)*x_ref[0]+(0.112704521653)*x_ref[1]+(-0.448684922187)*x_ref[2]
        ref[3,4,1,0]=(-0.396153523625)*x_ref[0]+(-0.0370870723568)*x_ref[1]+(0.554347618511)*x_ref[2]
        ref[3,4,1,1]=(-0.323276139711)*x_ref[0]+(-0.397389928392)*x_ref[1]+(-1.1707028324)*x_ref[2]
        ref[3,4,2,0]=(0.823350512513)*x_ref[0]+(1.47105759864)*x_ref[1]+(-0.977456877855)*x_ref[2]
        ref[3,4,2,1]=(-0.719791170741)*x_ref[0]+(-1.00363340645)*x_ref[1]+(0.814612770572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedContinuousFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.459241599813)*x[0]+(-0.773093048708)*x[1]
        ref=(-0.459241599813)*x_ref[0]+(-0.773093048708)*x_ref[1]
      else:
        s=(1.54748967569)*x[0]+(0.307874797615)*x[1]+(-0.414822431494)*x[2]
        ref=(1.54748967569)*x_ref[0]+(0.307874797615)*x_ref[1]+(-0.414822431494)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-1.51228566012)*x[0]+(-0.814844304147)*x[1]
        s[1]=(-0.881464001412)*x[0]+(0.764098134348)*x[1]
        ref[0]=(-1.51228566012)*x[0]+(-0.814844304147)*x[1]
        ref[1]=(-0.881464001412)*x[0]+(0.764098134348)*x[1]
      else:
        s[0]=(-0.91410305648)*x[0]+(-0.253301678985)*x[1]+(-0.00826967639394)*x[2]
        s[1]=(-0.592096115904)*x[0]+(-1.05985873602)*x[1]+(0.00147658091963)*x[2]
        ref[0]=(-0.91410305648)*x[0]+(-0.253301678985)*x[1]+(-0.00826967639394)*x[2]
        ref[1]=(-0.592096115904)*x[0]+(-1.05985873602)*x[1]+(0.00147658091963)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.209549210666)*x[0]+(1.05992146041)*x[1]
        s[0,1]=(-0.389388721283)*x[0]+(-0.122353380381)*x[1]
        s[0,2]=(1.48834814863)*x[0]+(1.11546828469)*x[1]
        s[0,3]=(-0.326661291423)*x[0]+(-0.20061861948)*x[1]
        s[0,4]=(0.941218540101)*x[0]+(-1.04582695848)*x[1]
        s[1,0]=(-0.196538410782)*x[0]+(0.110519724291)*x[1]
        s[1,1]=(0.361129728605)*x[0]+(0.740228471382)*x[1]
        s[1,2]=(-1.56560253129)*x[0]+(-0.585194140493)*x[1]
        s[1,3]=(0.177982070606)*x[0]+(-0.392964886439)*x[1]
        s[1,4]=(0.700111909929)*x[0]+(-0.366757279647)*x[1]
        s[2,0]=(-0.809083106921)*x[0]+(-0.447731857848)*x[1]
        s[2,1]=(0.867383974301)*x[0]+(-1.21953693153)*x[1]
        s[2,2]=(-0.859520524516)*x[0]+(-0.071452049869)*x[1]
        s[2,3]=(0.00398656657353)*x[0]+(1.23623708268)*x[1]
        s[2,4]=(0.166456051527)*x[0]+(1.47611415007)*x[1]
        s[3,0]=(-0.778396522621)*x[0]+(-0.386585278282)*x[1]
        s[3,1]=(0.231266833602)*x[0]+(0.19536493852)*x[1]
        s[3,2]=(-1.20923484879)*x[0]+(-0.490207641665)*x[1]
        s[3,3]=(0.455704265159)*x[0]+(0.109883193088)*x[1]
        s[3,4]=(-0.329454340826)*x[0]+(0.467154318354)*x[1]
        ref[0,0]=(0.209549210666)*x_ref[0]+(1.05992146041)*x_ref[1]
        ref[0,1]=(-0.389388721283)*x_ref[0]+(-0.122353380381)*x_ref[1]
        ref[0,2]=(1.48834814863)*x_ref[0]+(1.11546828469)*x_ref[1]
        ref[0,3]=(-0.326661291423)*x_ref[0]+(-0.20061861948)*x_ref[1]
        ref[0,4]=(0.941218540101)*x_ref[0]+(-1.04582695848)*x_ref[1]
        ref[1,0]=(-0.196538410782)*x_ref[0]+(0.110519724291)*x_ref[1]
        ref[1,1]=(0.361129728605)*x_ref[0]+(0.740228471382)*x_ref[1]
        ref[1,2]=(-1.56560253129)*x_ref[0]+(-0.585194140493)*x_ref[1]
        ref[1,3]=(0.177982070606)*x_ref[0]+(-0.392964886439)*x_ref[1]
        ref[1,4]=(0.700111909929)*x_ref[0]+(-0.366757279647)*x_ref[1]
        ref[2,0]=(-0.809083106921)*x_ref[0]+(-0.447731857848)*x_ref[1]
        ref[2,1]=(0.867383974301)*x_ref[0]+(-1.21953693153)*x_ref[1]
        ref[2,2]=(-0.859520524516)*x_ref[0]+(-0.071452049869)*x_ref[1]
        ref[2,3]=(0.00398656657353)*x_ref[0]+(1.23623708268)*x_ref[1]
        ref[2,4]=(0.166456051527)*x_ref[0]+(1.47611415007)*x_ref[1]
        ref[3,0]=(-0.778396522621)*x_ref[0]+(-0.386585278282)*x_ref[1]
        ref[3,1]=(0.231266833602)*x_ref[0]+(0.19536493852)*x_ref[1]
        ref[3,2]=(-1.20923484879)*x_ref[0]+(-0.490207641665)*x_ref[1]
        ref[3,3]=(0.455704265159)*x_ref[0]+(0.109883193088)*x_ref[1]
        ref[3,4]=(-0.329454340826)*x_ref[0]+(0.467154318354)*x_ref[1]
      else:
        s[0,0]=(-1.51270660613)*x[0]+(-1.23166098758)*x[1]+(0.00126913231069)*x[2]
        s[0,1]=(0.0783399703925)*x[0]+(0.704421352501)*x[1]+(-0.0527087709157)*x[2]
        s[0,2]=(0.212858037142)*x[0]+(0.879768825167)*x[1]+(0.3376329277)*x[2]
        s[0,3]=(-1.02885491754)*x[0]+(-0.349446901437)*x[1]+(-0.518150496566)*x[2]
        s[0,4]=(-0.632641407112)*x[0]+(0.736682065273)*x[1]+(0.849728972959)*x[2]
        s[1,0]=(-1.31434627789)*x[0]+(-1.40408223096)*x[1]+(0.184109480479)*x[2]
        s[1,1]=(-0.258805252155)*x[0]+(-0.794639774)*x[1]+(-0.370761920362)*x[2]
        s[1,2]=(0.645662860304)*x[0]+(0.196908807781)*x[1]+(-0.348791058482)*x[2]
        s[1,3]=(-0.397782835091)*x[0]+(0.292971279611)*x[1]+(-0.952746843146)*x[2]
        s[1,4]=(0.226551429583)*x[0]+(-0.891234988447)*x[1]+(-0.00309756588396)*x[2]
        s[2,0]=(-0.418562068509)*x[0]+(0.43739854393)*x[1]+(0.0509772218411)*x[2]
        s[2,1]=(0.783305063847)*x[0]+(0.824783771448)*x[1]+(0.21157958706)*x[2]
        s[2,2]=(-0.557598710851)*x[0]+(0.973487903111)*x[1]+(-0.400806427014)*x[2]
        s[2,3]=(1.02110121409)*x[0]+(-0.189675236979)*x[1]+(-0.692006170319)*x[2]
        s[2,4]=(0.204984594869)*x[0]+(-0.591307306263)*x[1]+(-0.0198714935717)*x[2]
        s[3,0]=(-0.524519709351)*x[0]+(0.263836205807)*x[1]+(-0.163880326974)*x[2]
        s[3,1]=(0.268669519472)*x[0]+(0.379462793693)*x[1]+(-0.297560957033)*x[2]
        s[3,2]=(-0.13621646905)*x[0]+(0.650614413717)*x[1]+(0.44018137263)*x[2]
        s[3,3]=(0.261445807973)*x[0]+(-0.00264446194008)*x[1]+(-0.168858453474)*x[2]
        s[3,4]=(-0.247581704325)*x[0]+(-1.9405050847)*x[1]+(1.5206187195)*x[2]
        ref[0,0]=(-1.51270660613)*x_ref[0]+(-1.23166098758)*x_ref[1]+(0.00126913231069)*x_ref[2]
        ref[0,1]=(0.0783399703925)*x_ref[0]+(0.704421352501)*x_ref[1]+(-0.0527087709157)*x_ref[2]
        ref[0,2]=(0.212858037142)*x_ref[0]+(0.879768825167)*x_ref[1]+(0.3376329277)*x_ref[2]
        ref[0,3]=(-1.02885491754)*x_ref[0]+(-0.349446901437)*x_ref[1]+(-0.518150496566)*x_ref[2]
        ref[0,4]=(-0.632641407112)*x_ref[0]+(0.736682065273)*x_ref[1]+(0.849728972959)*x_ref[2]
        ref[1,0]=(-1.31434627789)*x_ref[0]+(-1.40408223096)*x_ref[1]+(0.184109480479)*x_ref[2]
        ref[1,1]=(-0.258805252155)*x_ref[0]+(-0.794639774)*x_ref[1]+(-0.370761920362)*x_ref[2]
        ref[1,2]=(0.645662860304)*x_ref[0]+(0.196908807781)*x_ref[1]+(-0.348791058482)*x_ref[2]
        ref[1,3]=(-0.397782835091)*x_ref[0]+(0.292971279611)*x_ref[1]+(-0.952746843146)*x_ref[2]
        ref[1,4]=(0.226551429583)*x_ref[0]+(-0.891234988447)*x_ref[1]+(-0.00309756588396)*x_ref[2]
        ref[2,0]=(-0.418562068509)*x_ref[0]+(0.43739854393)*x_ref[1]+(0.0509772218411)*x_ref[2]
        ref[2,1]=(0.783305063847)*x_ref[0]+(0.824783771448)*x_ref[1]+(0.21157958706)*x_ref[2]
        ref[2,2]=(-0.557598710851)*x_ref[0]+(0.973487903111)*x_ref[1]+(-0.400806427014)*x_ref[2]
        ref[2,3]=(1.02110121409)*x_ref[0]+(-0.189675236979)*x_ref[1]+(-0.692006170319)*x_ref[2]
        ref[2,4]=(0.204984594869)*x_ref[0]+(-0.591307306263)*x_ref[1]+(-0.0198714935717)*x_ref[2]
        ref[3,0]=(-0.524519709351)*x_ref[0]+(0.263836205807)*x_ref[1]+(-0.163880326974)*x_ref[2]
        ref[3,1]=(0.268669519472)*x_ref[0]+(0.379462793693)*x_ref[1]+(-0.297560957033)*x_ref[2]
        ref[3,2]=(-0.13621646905)*x_ref[0]+(0.650614413717)*x_ref[1]+(0.44018137263)*x_ref[2]
        ref[3,3]=(0.261445807973)*x_ref[0]+(-0.00264446194008)*x_ref[1]+(-0.168858453474)*x_ref[2]
        ref[3,4]=(-0.247581704325)*x_ref[0]+(-1.9405050847)*x_ref[1]+(1.5206187195)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(1.27061375599)*x[0]+(0.693010110721)*x[1]
        s[0,0,1]=(0.385267649608)*x[0]+(-0.409112804165)*x[1]
        s[0,1,0]=(0.235254219396)*x[0]+(0.00806676427989)*x[1]
        s[0,1,1]=(0.546338474187)*x[0]+(0.269505133228)*x[1]
        s[1,0,0]=(-0.252450318463)*x[0]+(0.565357325098)*x[1]
        s[1,0,1]=(0.519066028574)*x[0]+(-0.151896081553)*x[1]
        s[1,1,0]=(0.583246542849)*x[0]+(0.74106904645)*x[1]
        s[1,1,1]=(0.283330295456)*x[0]+(1.3671331438)*x[1]
        s[2,0,0]=(0.295333306073)*x[0]+(-0.561691152526)*x[1]
        s[2,0,1]=(0.637514074862)*x[0]+(-0.484517233318)*x[1]
        s[2,1,0]=(-0.206769011353)*x[0]+(-0.0115621456599)*x[1]
        s[2,1,1]=(1.40165335416)*x[0]+(-1.43439048946)*x[1]
        s[3,0,0]=(0.567950789213)*x[0]+(-1.6427426351)*x[1]
        s[3,0,1]=(0.913125584361)*x[0]+(-0.192143890031)*x[1]
        s[3,1,0]=(-0.480605427139)*x[0]+(-0.474866017698)*x[1]
        s[3,1,1]=(-0.137178314752)*x[0]+(0.28719953047)*x[1]
        s[4,0,0]=(0.697948943346)*x[0]+(-0.907672862452)*x[1]
        s[4,0,1]=(-1.62690254564)*x[0]+(0.914366733304)*x[1]
        s[4,1,0]=(0.900337784632)*x[0]+(-0.0470217841933)*x[1]
        s[4,1,1]=(0.313193728362)*x[0]+(-0.122194456503)*x[1]
        s[5,0,0]=(0.606327952796)*x[0]+(-0.917393520302)*x[1]
        s[5,0,1]=(-0.262565816623)*x[0]+(0.232087670002)*x[1]
        s[5,1,0]=(-0.0786114664469)*x[0]+(0.252622361658)*x[1]
        s[5,1,1]=(0.193290467334)*x[0]+(0.529096699149)*x[1]
        ref[0,0,0]=(1.27061375599)*x_ref[0]+(0.693010110721)*x_ref[1]
        ref[0,0,1]=(0.385267649608)*x_ref[0]+(-0.409112804165)*x_ref[1]
        ref[0,1,0]=(0.235254219396)*x_ref[0]+(0.00806676427989)*x_ref[1]
        ref[0,1,1]=(0.546338474187)*x_ref[0]+(0.269505133228)*x_ref[1]
        ref[1,0,0]=(-0.252450318463)*x_ref[0]+(0.565357325098)*x_ref[1]
        ref[1,0,1]=(0.519066028574)*x_ref[0]+(-0.151896081553)*x_ref[1]
        ref[1,1,0]=(0.583246542849)*x_ref[0]+(0.74106904645)*x_ref[1]
        ref[1,1,1]=(0.283330295456)*x_ref[0]+(1.3671331438)*x_ref[1]
        ref[2,0,0]=(0.295333306073)*x_ref[0]+(-0.561691152526)*x_ref[1]
        ref[2,0,1]=(0.637514074862)*x_ref[0]+(-0.484517233318)*x_ref[1]
        ref[2,1,0]=(-0.206769011353)*x_ref[0]+(-0.0115621456599)*x_ref[1]
        ref[2,1,1]=(1.40165335416)*x_ref[0]+(-1.43439048946)*x_ref[1]
        ref[3,0,0]=(0.567950789213)*x_ref[0]+(-1.6427426351)*x_ref[1]
        ref[3,0,1]=(0.913125584361)*x_ref[0]+(-0.192143890031)*x_ref[1]
        ref[3,1,0]=(-0.480605427139)*x_ref[0]+(-0.474866017698)*x_ref[1]
        ref[3,1,1]=(-0.137178314752)*x_ref[0]+(0.28719953047)*x_ref[1]
        ref[4,0,0]=(0.697948943346)*x_ref[0]+(-0.907672862452)*x_ref[1]
        ref[4,0,1]=(-1.62690254564)*x_ref[0]+(0.914366733304)*x_ref[1]
        ref[4,1,0]=(0.900337784632)*x_ref[0]+(-0.0470217841933)*x_ref[1]
        ref[4,1,1]=(0.313193728362)*x_ref[0]+(-0.122194456503)*x_ref[1]
        ref[5,0,0]=(0.606327952796)*x_ref[0]+(-0.917393520302)*x_ref[1]
        ref[5,0,1]=(-0.262565816623)*x_ref[0]+(0.232087670002)*x_ref[1]
        ref[5,1,0]=(-0.0786114664469)*x_ref[0]+(0.252622361658)*x_ref[1]
        ref[5,1,1]=(0.193290467334)*x_ref[0]+(0.529096699149)*x_ref[1]
      else:
        s[0,0,0]=(-0.50338473477)*x[0]+(-0.164749126249)*x[1]+(-0.554967226755)*x[2]
        s[0,0,1]=(-0.776721662344)*x[0]+(0.84671767153)*x[1]+(-1.62325361007)*x[2]
        s[0,1,0]=(-0.553665559679)*x[0]+(0.762038412432)*x[1]+(0.219451112807)*x[2]
        s[0,1,1]=(-0.295200279284)*x[0]+(-0.366906381317)*x[1]+(-0.114064671233)*x[2]
        s[1,0,0]=(-0.627912305852)*x[0]+(1.05590035705)*x[1]+(-0.357432981292)*x[2]
        s[1,0,1]=(-0.103518100064)*x[0]+(0.0285901975227)*x[1]+(0.272743015592)*x[2]
        s[1,1,0]=(-1.48220340227)*x[0]+(1.16859770671)*x[1]+(-1.15282038238)*x[2]
        s[1,1,1]=(0.758570299808)*x[0]+(-0.211340586797)*x[1]+(0.263787144967)*x[2]
        s[2,0,0]=(-1.28649435546)*x[0]+(-1.34545925137)*x[1]+(-1.23248562938)*x[2]
        s[2,0,1]=(-0.685813167263)*x[0]+(-0.811060892059)*x[1]+(-0.905743168112)*x[2]
        s[2,1,0]=(-1.27012789685)*x[0]+(-1.45423179754)*x[1]+(-0.461566077486)*x[2]
        s[2,1,1]=(0.764272676292)*x[0]+(1.79354721116)*x[1]+(1.54561480836)*x[2]
        s[3,0,0]=(0.483458668467)*x[0]+(-0.520255901158)*x[1]+(0.931135044848)*x[2]
        s[3,0,1]=(0.0731034526543)*x[0]+(0.61085360436)*x[1]+(1.15630425582)*x[2]
        s[3,1,0]=(-1.13345961818)*x[0]+(1.53627408722)*x[1]+(0.0394537895884)*x[2]
        s[3,1,1]=(1.93910425421)*x[0]+(-0.245667398856)*x[1]+(0.26326290067)*x[2]
        s[4,0,0]=(-0.316109947551)*x[0]+(0.523912230594)*x[1]+(0.428093796157)*x[2]
        s[4,0,1]=(-0.421586699534)*x[0]+(0.0956846699257)*x[1]+(0.153930459299)*x[2]
        s[4,1,0]=(1.29911179477)*x[0]+(-0.668540796991)*x[1]+(0.226289836438)*x[2]
        s[4,1,1]=(-1.2174140353)*x[0]+(0.577999358777)*x[1]+(-0.756367967534)*x[2]
        s[5,0,0]=(-0.158832852108)*x[0]+(-0.444245105421)*x[1]+(0.00253446453922)*x[2]
        s[5,0,1]=(0.654262047318)*x[0]+(-0.573424848715)*x[1]+(-1.74703307574)*x[2]
        s[5,1,0]=(1.18879231974)*x[0]+(1.01556300264)*x[1]+(-0.902278634733)*x[2]
        s[5,1,1]=(-0.0136999102655)*x[0]+(-0.123143996985)*x[1]+(1.70755156491)*x[2]
        ref[0,0,0]=(-0.50338473477)*x_ref[0]+(-0.164749126249)*x_ref[1]+(-0.554967226755)*x_ref[2]
        ref[0,0,1]=(-0.776721662344)*x_ref[0]+(0.84671767153)*x_ref[1]+(-1.62325361007)*x_ref[2]
        ref[0,1,0]=(-0.553665559679)*x_ref[0]+(0.762038412432)*x_ref[1]+(0.219451112807)*x_ref[2]
        ref[0,1,1]=(-0.295200279284)*x_ref[0]+(-0.366906381317)*x_ref[1]+(-0.114064671233)*x_ref[2]
        ref[1,0,0]=(-0.627912305852)*x_ref[0]+(1.05590035705)*x_ref[1]+(-0.357432981292)*x_ref[2]
        ref[1,0,1]=(-0.103518100064)*x_ref[0]+(0.0285901975227)*x_ref[1]+(0.272743015592)*x_ref[2]
        ref[1,1,0]=(-1.48220340227)*x_ref[0]+(1.16859770671)*x_ref[1]+(-1.15282038238)*x_ref[2]
        ref[1,1,1]=(0.758570299808)*x_ref[0]+(-0.211340586797)*x_ref[1]+(0.263787144967)*x_ref[2]
        ref[2,0,0]=(-1.28649435546)*x_ref[0]+(-1.34545925137)*x_ref[1]+(-1.23248562938)*x_ref[2]
        ref[2,0,1]=(-0.685813167263)*x_ref[0]+(-0.811060892059)*x_ref[1]+(-0.905743168112)*x_ref[2]
        ref[2,1,0]=(-1.27012789685)*x_ref[0]+(-1.45423179754)*x_ref[1]+(-0.461566077486)*x_ref[2]
        ref[2,1,1]=(0.764272676292)*x_ref[0]+(1.79354721116)*x_ref[1]+(1.54561480836)*x_ref[2]
        ref[3,0,0]=(0.483458668467)*x_ref[0]+(-0.520255901158)*x_ref[1]+(0.931135044848)*x_ref[2]
        ref[3,0,1]=(0.0731034526543)*x_ref[0]+(0.61085360436)*x_ref[1]+(1.15630425582)*x_ref[2]
        ref[3,1,0]=(-1.13345961818)*x_ref[0]+(1.53627408722)*x_ref[1]+(0.0394537895884)*x_ref[2]
        ref[3,1,1]=(1.93910425421)*x_ref[0]+(-0.245667398856)*x_ref[1]+(0.26326290067)*x_ref[2]
        ref[4,0,0]=(-0.316109947551)*x_ref[0]+(0.523912230594)*x_ref[1]+(0.428093796157)*x_ref[2]
        ref[4,0,1]=(-0.421586699534)*x_ref[0]+(0.0956846699257)*x_ref[1]+(0.153930459299)*x_ref[2]
        ref[4,1,0]=(1.29911179477)*x_ref[0]+(-0.668540796991)*x_ref[1]+(0.226289836438)*x_ref[2]
        ref[4,1,1]=(-1.2174140353)*x_ref[0]+(0.577999358777)*x_ref[1]+(-0.756367967534)*x_ref[2]
        ref[5,0,0]=(-0.158832852108)*x_ref[0]+(-0.444245105421)*x_ref[1]+(0.00253446453922)*x_ref[2]
        ref[5,0,1]=(0.654262047318)*x_ref[0]+(-0.573424848715)*x_ref[1]+(-1.74703307574)*x_ref[2]
        ref[5,1,0]=(1.18879231974)*x_ref[0]+(1.01556300264)*x_ref[1]+(-0.902278634733)*x_ref[2]
        ref[5,1,1]=(-0.0136999102655)*x_ref[0]+(-0.123143996985)*x_ref[1]+(1.70755156491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.224003216764)*x[0]+(-1.53537141395)*x[1]
        s[0,0,0,1]=(-0.415108576452)*x[0]+(0.512267468115)*x[1]
        s[0,0,1,0]=(1.23475149747)*x[0]+(0.834141411624)*x[1]
        s[0,0,1,1]=(1.23314379215)*x[0]+(0.097729946154)*x[1]
        s[0,0,2,0]=(1.78249035938)*x[0]+(0.149287882179)*x[1]
        s[0,0,2,1]=(0.268374768258)*x[0]+(0.844004814805)*x[1]
        s[0,1,0,0]=(-0.0921256378797)*x[0]+(0.54607120248)*x[1]
        s[0,1,0,1]=(-0.108118391324)*x[0]+(1.44715323804)*x[1]
        s[0,1,1,0]=(0.642892153115)*x[0]+(1.29939851925)*x[1]
        s[0,1,1,1]=(-0.346502475731)*x[0]+(-1.59694261779)*x[1]
        s[0,1,2,0]=(1.17858221695)*x[0]+(-0.0467327691871)*x[1]
        s[0,1,2,1]=(-0.0676520776338)*x[0]+(-1.19701821874)*x[1]
        s[0,2,0,0]=(0.0594877869572)*x[0]+(0.4691178171)*x[1]
        s[0,2,0,1]=(-0.471134574689)*x[0]+(-0.0774880925496)*x[1]
        s[0,2,1,0]=(0.957705182694)*x[0]+(-0.148346694036)*x[1]
        s[0,2,1,1]=(1.35974448438)*x[0]+(0.976348738984)*x[1]
        s[0,2,2,0]=(-0.0888644417678)*x[0]+(0.185465627916)*x[1]
        s[0,2,2,1]=(0.366706758508)*x[0]+(0.424423271146)*x[1]
        s[0,3,0,0]=(0.300438450637)*x[0]+(0.489300186408)*x[1]
        s[0,3,0,1]=(-0.837684340885)*x[0]+(1.11058469871)*x[1]
        s[0,3,1,0]=(-0.685409716574)*x[0]+(1.29529487058)*x[1]
        s[0,3,1,1]=(-0.734298461298)*x[0]+(1.49154390821)*x[1]
        s[0,3,2,0]=(-0.360678115755)*x[0]+(-1.31090940382)*x[1]
        s[0,3,2,1]=(-1.76090388864)*x[0]+(-0.197058042294)*x[1]
        s[0,4,0,0]=(-0.647957185052)*x[0]+(-1.07397957837)*x[1]
        s[0,4,0,1]=(1.15138043713)*x[0]+(1.27703705411)*x[1]
        s[0,4,1,0]=(0.880204392036)*x[0]+(0.317776491177)*x[1]
        s[0,4,1,1]=(1.75846459794)*x[0]+(0.58329739948)*x[1]
        s[0,4,2,0]=(0.321739539217)*x[0]+(0.155992474445)*x[1]
        s[0,4,2,1]=(-0.831765424787)*x[0]+(-0.174239473413)*x[1]
        s[1,0,0,0]=(-0.315436217973)*x[0]+(-1.50414837046)*x[1]
        s[1,0,0,1]=(-0.20319068189)*x[0]+(-0.471694875132)*x[1]
        s[1,0,1,0]=(1.11728129989)*x[0]+(-1.56713796242)*x[1]
        s[1,0,1,1]=(1.57812938291)*x[0]+(0.768357124158)*x[1]
        s[1,0,2,0]=(0.846908340877)*x[0]+(0.00535056985765)*x[1]
        s[1,0,2,1]=(1.65497801331)*x[0]+(-0.787783905283)*x[1]
        s[1,1,0,0]=(-1.88514071539)*x[0]+(0.169147044809)*x[1]
        s[1,1,0,1]=(0.360553360055)*x[0]+(0.239599872139)*x[1]
        s[1,1,1,0]=(-0.870583030697)*x[0]+(-0.956757500255)*x[1]
        s[1,1,1,1]=(0.398902070984)*x[0]+(-0.299144054803)*x[1]
        s[1,1,2,0]=(0.683520893803)*x[0]+(0.569312151808)*x[1]
        s[1,1,2,1]=(-0.142857438428)*x[0]+(0.216222456467)*x[1]
        s[1,2,0,0]=(-0.512922081732)*x[0]+(1.55472942936)*x[1]
        s[1,2,0,1]=(-0.00159718420231)*x[0]+(0.126511414314)*x[1]
        s[1,2,1,0]=(0.687145297745)*x[0]+(-0.392573700665)*x[1]
        s[1,2,1,1]=(-0.308147808315)*x[0]+(0.0500851036579)*x[1]
        s[1,2,2,0]=(-0.121563796894)*x[0]+(-0.671555650221)*x[1]
        s[1,2,2,1]=(-1.14943669955)*x[0]+(-1.12860967665)*x[1]
        s[1,3,0,0]=(0.796492123166)*x[0]+(-0.594727403527)*x[1]
        s[1,3,0,1]=(0.0776108758324)*x[0]+(0.758254260971)*x[1]
        s[1,3,1,0]=(-0.974221864735)*x[0]+(1.02650607362)*x[1]
        s[1,3,1,1]=(-0.476728696056)*x[0]+(1.39394954201)*x[1]
        s[1,3,2,0]=(-0.276890313424)*x[0]+(-0.42434071217)*x[1]
        s[1,3,2,1]=(-0.45671100973)*x[0]+(-0.271986551291)*x[1]
        s[1,4,0,0]=(0.415315726206)*x[0]+(0.25174859473)*x[1]
        s[1,4,0,1]=(0.718116945034)*x[0]+(0.300885379458)*x[1]
        s[1,4,1,0]=(0.0362286683347)*x[0]+(0.0860658750969)*x[1]
        s[1,4,1,1]=(0.966944206449)*x[0]+(0.605527604606)*x[1]
        s[1,4,2,0]=(-0.681446172251)*x[0]+(-0.0201721796333)*x[1]
        s[1,4,2,1]=(1.16600295744)*x[0]+(1.10261292926)*x[1]
        s[2,0,0,0]=(-0.239525904251)*x[0]+(0.428475333805)*x[1]
        s[2,0,0,1]=(0.0817932891372)*x[0]+(0.472895457716)*x[1]
        s[2,0,1,0]=(0.96571284204)*x[0]+(1.30765537984)*x[1]
        s[2,0,1,1]=(-0.27274079215)*x[0]+(-0.973090020231)*x[1]
        s[2,0,2,0]=(-0.586811136701)*x[0]+(0.0282301965253)*x[1]
        s[2,0,2,1]=(-0.988289117174)*x[0]+(-0.982428801403)*x[1]
        s[2,1,0,0]=(-0.97313872216)*x[0]+(0.264946309189)*x[1]
        s[2,1,0,1]=(0.477464114989)*x[0]+(0.735813896712)*x[1]
        s[2,1,1,0]=(-0.999655360356)*x[0]+(0.0906084300723)*x[1]
        s[2,1,1,1]=(-0.672480642088)*x[0]+(-0.0511287334859)*x[1]
        s[2,1,2,0]=(-1.06854561704)*x[0]+(-0.188249587015)*x[1]
        s[2,1,2,1]=(-0.439030823181)*x[0]+(0.76989999863)*x[1]
        s[2,2,0,0]=(0.935351178562)*x[0]+(-0.732961189428)*x[1]
        s[2,2,0,1]=(-0.201449422629)*x[0]+(-0.920429771001)*x[1]
        s[2,2,1,0]=(-0.0397141758777)*x[0]+(1.23810013859)*x[1]
        s[2,2,1,1]=(-0.853635033788)*x[0]+(0.982464467568)*x[1]
        s[2,2,2,0]=(-1.35330887855)*x[0]+(0.200038841813)*x[1]
        s[2,2,2,1]=(0.0388084224165)*x[0]+(-0.312397495662)*x[1]
        s[2,3,0,0]=(-0.762894528313)*x[0]+(0.206223517392)*x[1]
        s[2,3,0,1]=(0.0823646028803)*x[0]+(-1.1834743671)*x[1]
        s[2,3,1,0]=(0.302380236653)*x[0]+(-0.725627755476)*x[1]
        s[2,3,1,1]=(-0.12698195797)*x[0]+(-0.289106042968)*x[1]
        s[2,3,2,0]=(-1.33992252725)*x[0]+(-0.903068424118)*x[1]
        s[2,3,2,1]=(-0.128913469794)*x[0]+(0.392464023413)*x[1]
        s[2,4,0,0]=(-0.326550273177)*x[0]+(-0.856389747069)*x[1]
        s[2,4,0,1]=(-1.14966749976)*x[0]+(0.809029270783)*x[1]
        s[2,4,1,0]=(1.50037083169)*x[0]+(0.564371263186)*x[1]
        s[2,4,1,1]=(-0.0660188964857)*x[0]+(-0.871309804802)*x[1]
        s[2,4,2,0]=(0.527342096457)*x[0]+(1.41520765334)*x[1]
        s[2,4,2,1]=(-0.177020653581)*x[0]+(-0.218017520886)*x[1]
        s[3,0,0,0]=(-0.901819829595)*x[0]+(-0.0897360627182)*x[1]
        s[3,0,0,1]=(-1.84231567451)*x[0]+(-1.57138993977)*x[1]
        s[3,0,1,0]=(-0.692347444292)*x[0]+(-0.35994981731)*x[1]
        s[3,0,1,1]=(-1.20666616877)*x[0]+(1.31974198302)*x[1]
        s[3,0,2,0]=(-1.49613152411)*x[0]+(-1.29087493111)*x[1]
        s[3,0,2,1]=(0.00302152289294)*x[0]+(0.0596509711717)*x[1]
        s[3,1,0,0]=(-0.716314986327)*x[0]+(0.540487536594)*x[1]
        s[3,1,0,1]=(1.26797434361)*x[0]+(1.24611664766)*x[1]
        s[3,1,1,0]=(1.29133591716)*x[0]+(-0.822278152235)*x[1]
        s[3,1,1,1]=(0.736506521682)*x[0]+(0.295109058687)*x[1]
        s[3,1,2,0]=(-1.31999878466)*x[0]+(-1.06268553583)*x[1]
        s[3,1,2,1]=(-1.02353629588)*x[0]+(-0.180118460659)*x[1]
        s[3,2,0,0]=(1.2591074793)*x[0]+(0.716164110599)*x[1]
        s[3,2,0,1]=(0.0737591553738)*x[0]+(-0.180183018938)*x[1]
        s[3,2,1,0]=(0.842324803092)*x[0]+(-0.179012336351)*x[1]
        s[3,2,1,1]=(-1.04205716802)*x[0]+(0.91772961597)*x[1]
        s[3,2,2,0]=(0.774019388104)*x[0]+(0.392202481643)*x[1]
        s[3,2,2,1]=(0.583625745968)*x[0]+(0.857713356409)*x[1]
        s[3,3,0,0]=(-1.08653988672)*x[0]+(-1.47741013604)*x[1]
        s[3,3,0,1]=(-0.529552146407)*x[0]+(-0.658937785849)*x[1]
        s[3,3,1,0]=(1.8439291246)*x[0]+(-1.36075468038)*x[1]
        s[3,3,1,1]=(-0.656328225025)*x[0]+(1.18200793838)*x[1]
        s[3,3,2,0]=(-1.57290668837)*x[0]+(0.527467780284)*x[1]
        s[3,3,2,1]=(1.44984969309)*x[0]+(-1.37301308873)*x[1]
        s[3,4,0,0]=(-0.225012249574)*x[0]+(-1.80387444921)*x[1]
        s[3,4,0,1]=(-1.27157572533)*x[0]+(0.568859005411)*x[1]
        s[3,4,1,0]=(-0.381626896774)*x[0]+(-0.335697445475)*x[1]
        s[3,4,1,1]=(0.555084489073)*x[0]+(-0.659578343093)*x[1]
        s[3,4,2,0]=(0.784894297286)*x[0]+(0.955314078432)*x[1]
        s[3,4,2,1]=(0.33356366872)*x[0]+(1.38175924325)*x[1]
        ref[0,0,0,0]=(-0.224003216764)*x_ref[0]+(-1.53537141395)*x_ref[1]
        ref[0,0,0,1]=(-0.415108576452)*x_ref[0]+(0.512267468115)*x_ref[1]
        ref[0,0,1,0]=(1.23475149747)*x_ref[0]+(0.834141411624)*x_ref[1]
        ref[0,0,1,1]=(1.23314379215)*x_ref[0]+(0.097729946154)*x_ref[1]
        ref[0,0,2,0]=(1.78249035938)*x_ref[0]+(0.149287882179)*x_ref[1]
        ref[0,0,2,1]=(0.268374768258)*x_ref[0]+(0.844004814805)*x_ref[1]
        ref[0,1,0,0]=(-0.0921256378797)*x_ref[0]+(0.54607120248)*x_ref[1]
        ref[0,1,0,1]=(-0.108118391324)*x_ref[0]+(1.44715323804)*x_ref[1]
        ref[0,1,1,0]=(0.642892153115)*x_ref[0]+(1.29939851925)*x_ref[1]
        ref[0,1,1,1]=(-0.346502475731)*x_ref[0]+(-1.59694261779)*x_ref[1]
        ref[0,1,2,0]=(1.17858221695)*x_ref[0]+(-0.0467327691871)*x_ref[1]
        ref[0,1,2,1]=(-0.0676520776338)*x_ref[0]+(-1.19701821874)*x_ref[1]
        ref[0,2,0,0]=(0.0594877869572)*x_ref[0]+(0.4691178171)*x_ref[1]
        ref[0,2,0,1]=(-0.471134574689)*x_ref[0]+(-0.0774880925496)*x_ref[1]
        ref[0,2,1,0]=(0.957705182694)*x_ref[0]+(-0.148346694036)*x_ref[1]
        ref[0,2,1,1]=(1.35974448438)*x_ref[0]+(0.976348738984)*x_ref[1]
        ref[0,2,2,0]=(-0.0888644417678)*x_ref[0]+(0.185465627916)*x_ref[1]
        ref[0,2,2,1]=(0.366706758508)*x_ref[0]+(0.424423271146)*x_ref[1]
        ref[0,3,0,0]=(0.300438450637)*x_ref[0]+(0.489300186408)*x_ref[1]
        ref[0,3,0,1]=(-0.837684340885)*x_ref[0]+(1.11058469871)*x_ref[1]
        ref[0,3,1,0]=(-0.685409716574)*x_ref[0]+(1.29529487058)*x_ref[1]
        ref[0,3,1,1]=(-0.734298461298)*x_ref[0]+(1.49154390821)*x_ref[1]
        ref[0,3,2,0]=(-0.360678115755)*x_ref[0]+(-1.31090940382)*x_ref[1]
        ref[0,3,2,1]=(-1.76090388864)*x_ref[0]+(-0.197058042294)*x_ref[1]
        ref[0,4,0,0]=(-0.647957185052)*x_ref[0]+(-1.07397957837)*x_ref[1]
        ref[0,4,0,1]=(1.15138043713)*x_ref[0]+(1.27703705411)*x_ref[1]
        ref[0,4,1,0]=(0.880204392036)*x_ref[0]+(0.317776491177)*x_ref[1]
        ref[0,4,1,1]=(1.75846459794)*x_ref[0]+(0.58329739948)*x_ref[1]
        ref[0,4,2,0]=(0.321739539217)*x_ref[0]+(0.155992474445)*x_ref[1]
        ref[0,4,2,1]=(-0.831765424787)*x_ref[0]+(-0.174239473413)*x_ref[1]
        ref[1,0,0,0]=(-0.315436217973)*x_ref[0]+(-1.50414837046)*x_ref[1]
        ref[1,0,0,1]=(-0.20319068189)*x_ref[0]+(-0.471694875132)*x_ref[1]
        ref[1,0,1,0]=(1.11728129989)*x_ref[0]+(-1.56713796242)*x_ref[1]
        ref[1,0,1,1]=(1.57812938291)*x_ref[0]+(0.768357124158)*x_ref[1]
        ref[1,0,2,0]=(0.846908340877)*x_ref[0]+(0.00535056985765)*x_ref[1]
        ref[1,0,2,1]=(1.65497801331)*x_ref[0]+(-0.787783905283)*x_ref[1]
        ref[1,1,0,0]=(-1.88514071539)*x_ref[0]+(0.169147044809)*x_ref[1]
        ref[1,1,0,1]=(0.360553360055)*x_ref[0]+(0.239599872139)*x_ref[1]
        ref[1,1,1,0]=(-0.870583030697)*x_ref[0]+(-0.956757500255)*x_ref[1]
        ref[1,1,1,1]=(0.398902070984)*x_ref[0]+(-0.299144054803)*x_ref[1]
        ref[1,1,2,0]=(0.683520893803)*x_ref[0]+(0.569312151808)*x_ref[1]
        ref[1,1,2,1]=(-0.142857438428)*x_ref[0]+(0.216222456467)*x_ref[1]
        ref[1,2,0,0]=(-0.512922081732)*x_ref[0]+(1.55472942936)*x_ref[1]
        ref[1,2,0,1]=(-0.00159718420231)*x_ref[0]+(0.126511414314)*x_ref[1]
        ref[1,2,1,0]=(0.687145297745)*x_ref[0]+(-0.392573700665)*x_ref[1]
        ref[1,2,1,1]=(-0.308147808315)*x_ref[0]+(0.0500851036579)*x_ref[1]
        ref[1,2,2,0]=(-0.121563796894)*x_ref[0]+(-0.671555650221)*x_ref[1]
        ref[1,2,2,1]=(-1.14943669955)*x_ref[0]+(-1.12860967665)*x_ref[1]
        ref[1,3,0,0]=(0.796492123166)*x_ref[0]+(-0.594727403527)*x_ref[1]
        ref[1,3,0,1]=(0.0776108758324)*x_ref[0]+(0.758254260971)*x_ref[1]
        ref[1,3,1,0]=(-0.974221864735)*x_ref[0]+(1.02650607362)*x_ref[1]
        ref[1,3,1,1]=(-0.476728696056)*x_ref[0]+(1.39394954201)*x_ref[1]
        ref[1,3,2,0]=(-0.276890313424)*x_ref[0]+(-0.42434071217)*x_ref[1]
        ref[1,3,2,1]=(-0.45671100973)*x_ref[0]+(-0.271986551291)*x_ref[1]
        ref[1,4,0,0]=(0.415315726206)*x_ref[0]+(0.25174859473)*x_ref[1]
        ref[1,4,0,1]=(0.718116945034)*x_ref[0]+(0.300885379458)*x_ref[1]
        ref[1,4,1,0]=(0.0362286683347)*x_ref[0]+(0.0860658750969)*x_ref[1]
        ref[1,4,1,1]=(0.966944206449)*x_ref[0]+(0.605527604606)*x_ref[1]
        ref[1,4,2,0]=(-0.681446172251)*x_ref[0]+(-0.0201721796333)*x_ref[1]
        ref[1,4,2,1]=(1.16600295744)*x_ref[0]+(1.10261292926)*x_ref[1]
        ref[2,0,0,0]=(-0.239525904251)*x_ref[0]+(0.428475333805)*x_ref[1]
        ref[2,0,0,1]=(0.0817932891372)*x_ref[0]+(0.472895457716)*x_ref[1]
        ref[2,0,1,0]=(0.96571284204)*x_ref[0]+(1.30765537984)*x_ref[1]
        ref[2,0,1,1]=(-0.27274079215)*x_ref[0]+(-0.973090020231)*x_ref[1]
        ref[2,0,2,0]=(-0.586811136701)*x_ref[0]+(0.0282301965253)*x_ref[1]
        ref[2,0,2,1]=(-0.988289117174)*x_ref[0]+(-0.982428801403)*x_ref[1]
        ref[2,1,0,0]=(-0.97313872216)*x_ref[0]+(0.264946309189)*x_ref[1]
        ref[2,1,0,1]=(0.477464114989)*x_ref[0]+(0.735813896712)*x_ref[1]
        ref[2,1,1,0]=(-0.999655360356)*x_ref[0]+(0.0906084300723)*x_ref[1]
        ref[2,1,1,1]=(-0.672480642088)*x_ref[0]+(-0.0511287334859)*x_ref[1]
        ref[2,1,2,0]=(-1.06854561704)*x_ref[0]+(-0.188249587015)*x_ref[1]
        ref[2,1,2,1]=(-0.439030823181)*x_ref[0]+(0.76989999863)*x_ref[1]
        ref[2,2,0,0]=(0.935351178562)*x_ref[0]+(-0.732961189428)*x_ref[1]
        ref[2,2,0,1]=(-0.201449422629)*x_ref[0]+(-0.920429771001)*x_ref[1]
        ref[2,2,1,0]=(-0.0397141758777)*x_ref[0]+(1.23810013859)*x_ref[1]
        ref[2,2,1,1]=(-0.853635033788)*x_ref[0]+(0.982464467568)*x_ref[1]
        ref[2,2,2,0]=(-1.35330887855)*x_ref[0]+(0.200038841813)*x_ref[1]
        ref[2,2,2,1]=(0.0388084224165)*x_ref[0]+(-0.312397495662)*x_ref[1]
        ref[2,3,0,0]=(-0.762894528313)*x_ref[0]+(0.206223517392)*x_ref[1]
        ref[2,3,0,1]=(0.0823646028803)*x_ref[0]+(-1.1834743671)*x_ref[1]
        ref[2,3,1,0]=(0.302380236653)*x_ref[0]+(-0.725627755476)*x_ref[1]
        ref[2,3,1,1]=(-0.12698195797)*x_ref[0]+(-0.289106042968)*x_ref[1]
        ref[2,3,2,0]=(-1.33992252725)*x_ref[0]+(-0.903068424118)*x_ref[1]
        ref[2,3,2,1]=(-0.128913469794)*x_ref[0]+(0.392464023413)*x_ref[1]
        ref[2,4,0,0]=(-0.326550273177)*x_ref[0]+(-0.856389747069)*x_ref[1]
        ref[2,4,0,1]=(-1.14966749976)*x_ref[0]+(0.809029270783)*x_ref[1]
        ref[2,4,1,0]=(1.50037083169)*x_ref[0]+(0.564371263186)*x_ref[1]
        ref[2,4,1,1]=(-0.0660188964857)*x_ref[0]+(-0.871309804802)*x_ref[1]
        ref[2,4,2,0]=(0.527342096457)*x_ref[0]+(1.41520765334)*x_ref[1]
        ref[2,4,2,1]=(-0.177020653581)*x_ref[0]+(-0.218017520886)*x_ref[1]
        ref[3,0,0,0]=(-0.901819829595)*x_ref[0]+(-0.0897360627182)*x_ref[1]
        ref[3,0,0,1]=(-1.84231567451)*x_ref[0]+(-1.57138993977)*x_ref[1]
        ref[3,0,1,0]=(-0.692347444292)*x_ref[0]+(-0.35994981731)*x_ref[1]
        ref[3,0,1,1]=(-1.20666616877)*x_ref[0]+(1.31974198302)*x_ref[1]
        ref[3,0,2,0]=(-1.49613152411)*x_ref[0]+(-1.29087493111)*x_ref[1]
        ref[3,0,2,1]=(0.00302152289294)*x_ref[0]+(0.0596509711717)*x_ref[1]
        ref[3,1,0,0]=(-0.716314986327)*x_ref[0]+(0.540487536594)*x_ref[1]
        ref[3,1,0,1]=(1.26797434361)*x_ref[0]+(1.24611664766)*x_ref[1]
        ref[3,1,1,0]=(1.29133591716)*x_ref[0]+(-0.822278152235)*x_ref[1]
        ref[3,1,1,1]=(0.736506521682)*x_ref[0]+(0.295109058687)*x_ref[1]
        ref[3,1,2,0]=(-1.31999878466)*x_ref[0]+(-1.06268553583)*x_ref[1]
        ref[3,1,2,1]=(-1.02353629588)*x_ref[0]+(-0.180118460659)*x_ref[1]
        ref[3,2,0,0]=(1.2591074793)*x_ref[0]+(0.716164110599)*x_ref[1]
        ref[3,2,0,1]=(0.0737591553738)*x_ref[0]+(-0.180183018938)*x_ref[1]
        ref[3,2,1,0]=(0.842324803092)*x_ref[0]+(-0.179012336351)*x_ref[1]
        ref[3,2,1,1]=(-1.04205716802)*x_ref[0]+(0.91772961597)*x_ref[1]
        ref[3,2,2,0]=(0.774019388104)*x_ref[0]+(0.392202481643)*x_ref[1]
        ref[3,2,2,1]=(0.583625745968)*x_ref[0]+(0.857713356409)*x_ref[1]
        ref[3,3,0,0]=(-1.08653988672)*x_ref[0]+(-1.47741013604)*x_ref[1]
        ref[3,3,0,1]=(-0.529552146407)*x_ref[0]+(-0.658937785849)*x_ref[1]
        ref[3,3,1,0]=(1.8439291246)*x_ref[0]+(-1.36075468038)*x_ref[1]
        ref[3,3,1,1]=(-0.656328225025)*x_ref[0]+(1.18200793838)*x_ref[1]
        ref[3,3,2,0]=(-1.57290668837)*x_ref[0]+(0.527467780284)*x_ref[1]
        ref[3,3,2,1]=(1.44984969309)*x_ref[0]+(-1.37301308873)*x_ref[1]
        ref[3,4,0,0]=(-0.225012249574)*x_ref[0]+(-1.80387444921)*x_ref[1]
        ref[3,4,0,1]=(-1.27157572533)*x_ref[0]+(0.568859005411)*x_ref[1]
        ref[3,4,1,0]=(-0.381626896774)*x_ref[0]+(-0.335697445475)*x_ref[1]
        ref[3,4,1,1]=(0.555084489073)*x_ref[0]+(-0.659578343093)*x_ref[1]
        ref[3,4,2,0]=(0.784894297286)*x_ref[0]+(0.955314078432)*x_ref[1]
        ref[3,4,2,1]=(0.33356366872)*x_ref[0]+(1.38175924325)*x_ref[1]
      else:
        s[0,0,0,0]=(0.598866527602)*x[0]+(0.44051361201)*x[1]+(-0.131423809939)*x[2]
        s[0,0,0,1]=(1.25181618674)*x[0]+(1.52822376488)*x[1]+(1.30574312696)*x[2]
        s[0,0,1,0]=(-0.331709443946)*x[0]+(-1.34705365817)*x[1]+(0.726301465323)*x[2]
        s[0,0,1,1]=(-0.577398988618)*x[0]+(-1.22594248786)*x[1]+(-0.961458379201)*x[2]
        s[0,0,2,0]=(-0.451500880848)*x[0]+(-0.774270726492)*x[1]+(-1.2165638211)*x[2]
        s[0,0,2,1]=(-1.10567205671)*x[0]+(-0.161917736574)*x[1]+(1.37957442359)*x[2]
        s[0,1,0,0]=(0.745737514206)*x[0]+(-0.796248435669)*x[1]+(-1.63816932586)*x[2]
        s[0,1,0,1]=(0.0831105458366)*x[0]+(-0.136660461565)*x[1]+(-1.56685434932)*x[2]
        s[0,1,1,0]=(-1.44351617889)*x[0]+(1.45409315626)*x[1]+(0.442708463336)*x[2]
        s[0,1,1,1]=(-0.558066341802)*x[0]+(1.19630033662)*x[1]+(0.186071415659)*x[2]
        s[0,1,2,0]=(-0.374869442605)*x[0]+(0.648318396628)*x[1]+(0.0555371118519)*x[2]
        s[0,1,2,1]=(-0.36160618163)*x[0]+(0.437013254054)*x[1]+(-0.638236225808)*x[2]
        s[0,2,0,0]=(0.189302763729)*x[0]+(-0.255111592053)*x[1]+(0.682670294776)*x[2]
        s[0,2,0,1]=(0.025995815858)*x[0]+(0.34462072647)*x[1]+(0.341995413585)*x[2]
        s[0,2,1,0]=(-1.28972101272)*x[0]+(1.01784177655)*x[1]+(1.11775675824)*x[2]
        s[0,2,1,1]=(0.557076830097)*x[0]+(0.838542238067)*x[1]+(-0.155585369573)*x[2]
        s[0,2,2,0]=(-0.0141241838818)*x[0]+(-0.191831242085)*x[1]+(-0.847287909515)*x[2]
        s[0,2,2,1]=(0.429778841571)*x[0]+(0.376566699487)*x[1]+(0.148313161917)*x[2]
        s[0,3,0,0]=(0.233606469427)*x[0]+(0.266803505307)*x[1]+(-0.205218686924)*x[2]
        s[0,3,0,1]=(-0.124511680509)*x[0]+(1.02896325334)*x[1]+(0.935771094012)*x[2]
        s[0,3,1,0]=(0.976537956221)*x[0]+(-0.345122171465)*x[1]+(0.777459950315)*x[2]
        s[0,3,1,1]=(-0.263691953956)*x[0]+(0.287964213001)*x[1]+(-0.237113357698)*x[2]
        s[0,3,2,0]=(-0.14155225233)*x[0]+(0.122587298994)*x[1]+(-0.0264137141467)*x[2]
        s[0,3,2,1]=(-0.760118620618)*x[0]+(1.57384021648)*x[1]+(1.32809481134)*x[2]
        s[0,4,0,0]=(0.673626831345)*x[0]+(-0.454271151216)*x[1]+(-0.863944524769)*x[2]
        s[0,4,0,1]=(-0.630430988823)*x[0]+(0.45809212038)*x[1]+(-1.27010861122)*x[2]
        s[0,4,1,0]=(-0.171563094037)*x[0]+(-0.360716542419)*x[1]+(0.162811650708)*x[2]
        s[0,4,1,1]=(0.890628840767)*x[0]+(-0.33837153333)*x[1]+(-1.57253750851)*x[2]
        s[0,4,2,0]=(-0.0358372393822)*x[0]+(1.12988684611)*x[1]+(0.153713271682)*x[2]
        s[0,4,2,1]=(-0.24590925839)*x[0]+(0.616341690491)*x[1]+(-1.12513207308)*x[2]
        s[1,0,0,0]=(1.17052294387)*x[0]+(-0.176023586935)*x[1]+(-0.413320632417)*x[2]
        s[1,0,0,1]=(1.2820208139)*x[0]+(0.228009714737)*x[1]+(-0.149423928125)*x[2]
        s[1,0,1,0]=(-0.0312869788993)*x[0]+(0.526820009564)*x[1]+(0.0903543587276)*x[2]
        s[1,0,1,1]=(-1.21319041974)*x[0]+(-0.428253988159)*x[1]+(0.464434175335)*x[2]
        s[1,0,2,0]=(0.9190018302)*x[0]+(0.449296159226)*x[1]+(-0.481445053339)*x[2]
        s[1,0,2,1]=(-0.616752258293)*x[0]+(0.359850140048)*x[1]+(-1.27542462984)*x[2]
        s[1,1,0,0]=(1.35404696733)*x[0]+(1.25468169576)*x[1]+(-0.410473495612)*x[2]
        s[1,1,0,1]=(0.320849051626)*x[0]+(1.56909917585)*x[1]+(1.00666256572)*x[2]
        s[1,1,1,0]=(0.235361535902)*x[0]+(-0.826787974769)*x[1]+(1.07359804999)*x[2]
        s[1,1,1,1]=(-0.334284708722)*x[0]+(0.316786292756)*x[1]+(-0.886350336088)*x[2]
        s[1,1,2,0]=(-0.890644854382)*x[0]+(0.509273281135)*x[1]+(-1.19292772513)*x[2]
        s[1,1,2,1]=(0.252216653779)*x[0]+(1.57054197023)*x[1]+(-1.90744671962)*x[2]
        s[1,2,0,0]=(0.551491892715)*x[0]+(-0.222655049798)*x[1]+(0.398387383719)*x[2]
        s[1,2,0,1]=(-1.55041598041)*x[0]+(0.994016310342)*x[1]+(-0.474205618301)*x[2]
        s[1,2,1,0]=(0.268662020467)*x[0]+(-0.755250236528)*x[1]+(-0.540383977121)*x[2]
        s[1,2,1,1]=(-0.551512759298)*x[0]+(1.4277999148)*x[1]+(-1.58961615319)*x[2]
        s[1,2,2,0]=(0.304822528983)*x[0]+(0.232100103775)*x[1]+(0.463883691163)*x[2]
        s[1,2,2,1]=(0.561644600923)*x[0]+(-0.359426536495)*x[1]+(-1.34828152802)*x[2]
        s[1,3,0,0]=(0.850248623505)*x[0]+(0.116021948452)*x[1]+(0.142044197471)*x[2]
        s[1,3,0,1]=(0.724145173317)*x[0]+(-0.792974540874)*x[1]+(-0.195454750691)*x[2]
        s[1,3,1,0]=(-0.110714388555)*x[0]+(-1.40121178374)*x[1]+(0.0347355958316)*x[2]
        s[1,3,1,1]=(-1.12980968478)*x[0]+(-0.570714541482)*x[1]+(-1.2280735041)*x[2]
        s[1,3,2,0]=(-1.21834172715)*x[0]+(1.83863423781)*x[1]+(0.442239363821)*x[2]
        s[1,3,2,1]=(0.426294021669)*x[0]+(1.63350652457)*x[1]+(-0.580391188165)*x[2]
        s[1,4,0,0]=(-0.791296449419)*x[0]+(0.395345927547)*x[1]+(-0.575778346002)*x[2]
        s[1,4,0,1]=(1.05262027172)*x[0]+(-1.1912135275)*x[1]+(-0.281282812313)*x[2]
        s[1,4,1,0]=(-0.229073653507)*x[0]+(0.333793277005)*x[1]+(-0.540787491142)*x[2]
        s[1,4,1,1]=(-0.782297770239)*x[0]+(0.965050908578)*x[1]+(0.648337674864)*x[2]
        s[1,4,2,0]=(-0.425598399058)*x[0]+(1.24755177094)*x[1]+(-0.599571710505)*x[2]
        s[1,4,2,1]=(-0.584529028943)*x[0]+(-0.115930785496)*x[1]+(-0.412776678144)*x[2]
        s[2,0,0,0]=(-1.22220614711)*x[0]+(0.991280082493)*x[1]+(-0.771530362129)*x[2]
        s[2,0,0,1]=(-0.0408833973017)*x[0]+(-0.878550136082)*x[1]+(-1.27973588501)*x[2]
        s[2,0,1,0]=(0.380052050339)*x[0]+(1.34070050501)*x[1]+(-1.39080485735)*x[2]
        s[2,0,1,1]=(-0.725808759614)*x[0]+(1.37568148722)*x[1]+(-0.369008444335)*x[2]
        s[2,0,2,0]=(-0.563453230072)*x[0]+(0.25114466425)*x[1]+(-0.0602643663972)*x[2]
        s[2,0,2,1]=(-0.802381432423)*x[0]+(0.22715480998)*x[1]+(0.512734770933)*x[2]
        s[2,1,0,0]=(-0.522742058617)*x[0]+(0.37241878501)*x[1]+(0.691287986533)*x[2]
        s[2,1,0,1]=(-0.168275058841)*x[0]+(-0.764400123274)*x[1]+(-0.681242773507)*x[2]
        s[2,1,1,0]=(0.551464415842)*x[0]+(1.03330737276)*x[1]+(0.958549917295)*x[2]
        s[2,1,1,1]=(-0.909196447141)*x[0]+(0.145246231629)*x[1]+(-1.05582073367)*x[2]
        s[2,1,2,0]=(-0.638872672519)*x[0]+(0.406342003401)*x[1]+(-0.393150454497)*x[2]
        s[2,1,2,1]=(0.366774488001)*x[0]+(0.365786897102)*x[1]+(0.441878454676)*x[2]
        s[2,2,0,0]=(-0.261857371972)*x[0]+(0.0318314524134)*x[1]+(-0.101372223186)*x[2]
        s[2,2,0,1]=(-0.135781961265)*x[0]+(-0.00345171998114)*x[1]+(0.0101907167492)*x[2]
        s[2,2,1,0]=(-1.37428228456)*x[0]+(-0.524953075499)*x[1]+(-0.372440345968)*x[2]
        s[2,2,1,1]=(0.0540542424565)*x[0]+(-1.03413875822)*x[1]+(-0.899141477767)*x[2]
        s[2,2,2,0]=(-0.431640543873)*x[0]+(0.934769305572)*x[1]+(-0.848404080116)*x[2]
        s[2,2,2,1]=(0.984445227688)*x[0]+(-0.207064309031)*x[1]+(-1.75497257678)*x[2]
        s[2,3,0,0]=(-1.61836350351)*x[0]+(0.885389572019)*x[1]+(0.133218748941)*x[2]
        s[2,3,0,1]=(0.100716060215)*x[0]+(-0.713421128342)*x[1]+(1.16765406123)*x[2]
        s[2,3,1,0]=(0.497272818223)*x[0]+(-0.698345898985)*x[1]+(-0.643672453237)*x[2]
        s[2,3,1,1]=(-0.680652178847)*x[0]+(-0.56686153354)*x[1]+(-0.259410997801)*x[2]
        s[2,3,2,0]=(1.08214823373)*x[0]+(-0.875408253309)*x[1]+(-0.165330678277)*x[2]
        s[2,3,2,1]=(-0.251850346585)*x[0]+(-0.233254346449)*x[1]+(0.348350958039)*x[2]
        s[2,4,0,0]=(0.283802782063)*x[0]+(0.338243660721)*x[1]+(0.0400948594919)*x[2]
        s[2,4,0,1]=(-1.46364736479)*x[0]+(0.738031226032)*x[1]+(0.897193522713)*x[2]
        s[2,4,1,0]=(1.17637134366)*x[0]+(-0.607880128773)*x[1]+(-1.00344218409)*x[2]
        s[2,4,1,1]=(0.716275197822)*x[0]+(1.21991657141)*x[1]+(-0.0467328809597)*x[2]
        s[2,4,2,0]=(0.380704185608)*x[0]+(0.603603708529)*x[1]+(-0.820752063279)*x[2]
        s[2,4,2,1]=(1.39582555808)*x[0]+(-0.473415991144)*x[1]+(1.24135585822)*x[2]
        s[3,0,0,0]=(1.39071291597)*x[0]+(-1.59041985271)*x[1]+(0.60185319823)*x[2]
        s[3,0,0,1]=(-1.46801716969)*x[0]+(-0.552777641699)*x[1]+(1.23129633317)*x[2]
        s[3,0,1,0]=(1.64196247631)*x[0]+(-0.343527059014)*x[1]+(1.11486568074)*x[2]
        s[3,0,1,1]=(-1.57106459992)*x[0]+(1.18660381744)*x[1]+(0.495305193784)*x[2]
        s[3,0,2,0]=(-0.246662602621)*x[0]+(0.980536147604)*x[1]+(-1.29582037379)*x[2]
        s[3,0,2,1]=(0.351277952246)*x[0]+(0.0189052721516)*x[1]+(-0.109275345669)*x[2]
        s[3,1,0,0]=(-0.559164628236)*x[0]+(-0.390894772538)*x[1]+(0.657026552721)*x[2]
        s[3,1,0,1]=(-1.63523473278)*x[0]+(-0.958777701493)*x[1]+(0.0466950147837)*x[2]
        s[3,1,1,0]=(-0.913387269773)*x[0]+(-0.752264610272)*x[1]+(-1.00623906741)*x[2]
        s[3,1,1,1]=(0.703502728627)*x[0]+(0.985524845396)*x[1]+(-0.0966526090953)*x[2]
        s[3,1,2,0]=(-0.993367844074)*x[0]+(-0.205401158091)*x[1]+(0.416042784765)*x[2]
        s[3,1,2,1]=(0.299615272051)*x[0]+(1.32572777489)*x[1]+(0.338960058542)*x[2]
        s[3,2,0,0]=(-0.757820100577)*x[0]+(0.644913131328)*x[1]+(-0.660404804486)*x[2]
        s[3,2,0,1]=(-1.75641089093)*x[0]+(-0.366399254245)*x[1]+(-0.646900183385)*x[2]
        s[3,2,1,0]=(0.0912144040068)*x[0]+(-0.857875163959)*x[1]+(0.350343357173)*x[2]
        s[3,2,1,1]=(-1.80008916515)*x[0]+(-1.83308297974)*x[1]+(0.052958998062)*x[2]
        s[3,2,2,0]=(1.44092323696)*x[0]+(-1.1831374133)*x[1]+(-0.0323647992786)*x[2]
        s[3,2,2,1]=(-0.353641143336)*x[0]+(0.774344611259)*x[1]+(-0.842874663076)*x[2]
        s[3,3,0,0]=(-1.17418376796)*x[0]+(0.334599071712)*x[1]+(0.0621712582176)*x[2]
        s[3,3,0,1]=(1.1539137936)*x[0]+(0.781617539757)*x[1]+(1.6289484244)*x[2]
        s[3,3,1,0]=(1.6860969053)*x[0]+(-0.936239288854)*x[1]+(1.7596463612)*x[2]
        s[3,3,1,1]=(-1.11111948999)*x[0]+(0.285869758308)*x[1]+(1.69003878827)*x[2]
        s[3,3,2,0]=(-1.00166505759)*x[0]+(-1.21067712793)*x[1]+(0.422711238131)*x[2]
        s[3,3,2,1]=(0.49093698547)*x[0]+(-0.952249146613)*x[1]+(-0.955917587594)*x[2]
        s[3,4,0,0]=(-0.511109469881)*x[0]+(0.630807398954)*x[1]+(-0.0365100439309)*x[2]
        s[3,4,0,1]=(0.260323499449)*x[0]+(-0.232312235272)*x[1]+(-0.676872879397)*x[2]
        s[3,4,1,0]=(-0.327180688658)*x[0]+(0.0260172986398)*x[1]+(-1.59267263743)*x[2]
        s[3,4,1,1]=(0.135287590014)*x[0]+(1.69850923514)*x[1]+(-0.498682843917)*x[2]
        s[3,4,2,0]=(0.314853952103)*x[0]+(-0.562331392142)*x[1]+(0.726740096722)*x[2]
        s[3,4,2,1]=(-0.638129182205)*x[0]+(-0.881463492466)*x[1]+(-0.792625664675)*x[2]
        ref[0,0,0,0]=(0.598866527602)*x_ref[0]+(0.44051361201)*x_ref[1]+(-0.131423809939)*x_ref[2]
        ref[0,0,0,1]=(1.25181618674)*x_ref[0]+(1.52822376488)*x_ref[1]+(1.30574312696)*x_ref[2]
        ref[0,0,1,0]=(-0.331709443946)*x_ref[0]+(-1.34705365817)*x_ref[1]+(0.726301465323)*x_ref[2]
        ref[0,0,1,1]=(-0.577398988618)*x_ref[0]+(-1.22594248786)*x_ref[1]+(-0.961458379201)*x_ref[2]
        ref[0,0,2,0]=(-0.451500880848)*x_ref[0]+(-0.774270726492)*x_ref[1]+(-1.2165638211)*x_ref[2]
        ref[0,0,2,1]=(-1.10567205671)*x_ref[0]+(-0.161917736574)*x_ref[1]+(1.37957442359)*x_ref[2]
        ref[0,1,0,0]=(0.745737514206)*x_ref[0]+(-0.796248435669)*x_ref[1]+(-1.63816932586)*x_ref[2]
        ref[0,1,0,1]=(0.0831105458366)*x_ref[0]+(-0.136660461565)*x_ref[1]+(-1.56685434932)*x_ref[2]
        ref[0,1,1,0]=(-1.44351617889)*x_ref[0]+(1.45409315626)*x_ref[1]+(0.442708463336)*x_ref[2]
        ref[0,1,1,1]=(-0.558066341802)*x_ref[0]+(1.19630033662)*x_ref[1]+(0.186071415659)*x_ref[2]
        ref[0,1,2,0]=(-0.374869442605)*x_ref[0]+(0.648318396628)*x_ref[1]+(0.0555371118519)*x_ref[2]
        ref[0,1,2,1]=(-0.36160618163)*x_ref[0]+(0.437013254054)*x_ref[1]+(-0.638236225808)*x_ref[2]
        ref[0,2,0,0]=(0.189302763729)*x_ref[0]+(-0.255111592053)*x_ref[1]+(0.682670294776)*x_ref[2]
        ref[0,2,0,1]=(0.025995815858)*x_ref[0]+(0.34462072647)*x_ref[1]+(0.341995413585)*x_ref[2]
        ref[0,2,1,0]=(-1.28972101272)*x_ref[0]+(1.01784177655)*x_ref[1]+(1.11775675824)*x_ref[2]
        ref[0,2,1,1]=(0.557076830097)*x_ref[0]+(0.838542238067)*x_ref[1]+(-0.155585369573)*x_ref[2]
        ref[0,2,2,0]=(-0.0141241838818)*x_ref[0]+(-0.191831242085)*x_ref[1]+(-0.847287909515)*x_ref[2]
        ref[0,2,2,1]=(0.429778841571)*x_ref[0]+(0.376566699487)*x_ref[1]+(0.148313161917)*x_ref[2]
        ref[0,3,0,0]=(0.233606469427)*x_ref[0]+(0.266803505307)*x_ref[1]+(-0.205218686924)*x_ref[2]
        ref[0,3,0,1]=(-0.124511680509)*x_ref[0]+(1.02896325334)*x_ref[1]+(0.935771094012)*x_ref[2]
        ref[0,3,1,0]=(0.976537956221)*x_ref[0]+(-0.345122171465)*x_ref[1]+(0.777459950315)*x_ref[2]
        ref[0,3,1,1]=(-0.263691953956)*x_ref[0]+(0.287964213001)*x_ref[1]+(-0.237113357698)*x_ref[2]
        ref[0,3,2,0]=(-0.14155225233)*x_ref[0]+(0.122587298994)*x_ref[1]+(-0.0264137141467)*x_ref[2]
        ref[0,3,2,1]=(-0.760118620618)*x_ref[0]+(1.57384021648)*x_ref[1]+(1.32809481134)*x_ref[2]
        ref[0,4,0,0]=(0.673626831345)*x_ref[0]+(-0.454271151216)*x_ref[1]+(-0.863944524769)*x_ref[2]
        ref[0,4,0,1]=(-0.630430988823)*x_ref[0]+(0.45809212038)*x_ref[1]+(-1.27010861122)*x_ref[2]
        ref[0,4,1,0]=(-0.171563094037)*x_ref[0]+(-0.360716542419)*x_ref[1]+(0.162811650708)*x_ref[2]
        ref[0,4,1,1]=(0.890628840767)*x_ref[0]+(-0.33837153333)*x_ref[1]+(-1.57253750851)*x_ref[2]
        ref[0,4,2,0]=(-0.0358372393822)*x_ref[0]+(1.12988684611)*x_ref[1]+(0.153713271682)*x_ref[2]
        ref[0,4,2,1]=(-0.24590925839)*x_ref[0]+(0.616341690491)*x_ref[1]+(-1.12513207308)*x_ref[2]
        ref[1,0,0,0]=(1.17052294387)*x_ref[0]+(-0.176023586935)*x_ref[1]+(-0.413320632417)*x_ref[2]
        ref[1,0,0,1]=(1.2820208139)*x_ref[0]+(0.228009714737)*x_ref[1]+(-0.149423928125)*x_ref[2]
        ref[1,0,1,0]=(-0.0312869788993)*x_ref[0]+(0.526820009564)*x_ref[1]+(0.0903543587276)*x_ref[2]
        ref[1,0,1,1]=(-1.21319041974)*x_ref[0]+(-0.428253988159)*x_ref[1]+(0.464434175335)*x_ref[2]
        ref[1,0,2,0]=(0.9190018302)*x_ref[0]+(0.449296159226)*x_ref[1]+(-0.481445053339)*x_ref[2]
        ref[1,0,2,1]=(-0.616752258293)*x_ref[0]+(0.359850140048)*x_ref[1]+(-1.27542462984)*x_ref[2]
        ref[1,1,0,0]=(1.35404696733)*x_ref[0]+(1.25468169576)*x_ref[1]+(-0.410473495612)*x_ref[2]
        ref[1,1,0,1]=(0.320849051626)*x_ref[0]+(1.56909917585)*x_ref[1]+(1.00666256572)*x_ref[2]
        ref[1,1,1,0]=(0.235361535902)*x_ref[0]+(-0.826787974769)*x_ref[1]+(1.07359804999)*x_ref[2]
        ref[1,1,1,1]=(-0.334284708722)*x_ref[0]+(0.316786292756)*x_ref[1]+(-0.886350336088)*x_ref[2]
        ref[1,1,2,0]=(-0.890644854382)*x_ref[0]+(0.509273281135)*x_ref[1]+(-1.19292772513)*x_ref[2]
        ref[1,1,2,1]=(0.252216653779)*x_ref[0]+(1.57054197023)*x_ref[1]+(-1.90744671962)*x_ref[2]
        ref[1,2,0,0]=(0.551491892715)*x_ref[0]+(-0.222655049798)*x_ref[1]+(0.398387383719)*x_ref[2]
        ref[1,2,0,1]=(-1.55041598041)*x_ref[0]+(0.994016310342)*x_ref[1]+(-0.474205618301)*x_ref[2]
        ref[1,2,1,0]=(0.268662020467)*x_ref[0]+(-0.755250236528)*x_ref[1]+(-0.540383977121)*x_ref[2]
        ref[1,2,1,1]=(-0.551512759298)*x_ref[0]+(1.4277999148)*x_ref[1]+(-1.58961615319)*x_ref[2]
        ref[1,2,2,0]=(0.304822528983)*x_ref[0]+(0.232100103775)*x_ref[1]+(0.463883691163)*x_ref[2]
        ref[1,2,2,1]=(0.561644600923)*x_ref[0]+(-0.359426536495)*x_ref[1]+(-1.34828152802)*x_ref[2]
        ref[1,3,0,0]=(0.850248623505)*x_ref[0]+(0.116021948452)*x_ref[1]+(0.142044197471)*x_ref[2]
        ref[1,3,0,1]=(0.724145173317)*x_ref[0]+(-0.792974540874)*x_ref[1]+(-0.195454750691)*x_ref[2]
        ref[1,3,1,0]=(-0.110714388555)*x_ref[0]+(-1.40121178374)*x_ref[1]+(0.0347355958316)*x_ref[2]
        ref[1,3,1,1]=(-1.12980968478)*x_ref[0]+(-0.570714541482)*x_ref[1]+(-1.2280735041)*x_ref[2]
        ref[1,3,2,0]=(-1.21834172715)*x_ref[0]+(1.83863423781)*x_ref[1]+(0.442239363821)*x_ref[2]
        ref[1,3,2,1]=(0.426294021669)*x_ref[0]+(1.63350652457)*x_ref[1]+(-0.580391188165)*x_ref[2]
        ref[1,4,0,0]=(-0.791296449419)*x_ref[0]+(0.395345927547)*x_ref[1]+(-0.575778346002)*x_ref[2]
        ref[1,4,0,1]=(1.05262027172)*x_ref[0]+(-1.1912135275)*x_ref[1]+(-0.281282812313)*x_ref[2]
        ref[1,4,1,0]=(-0.229073653507)*x_ref[0]+(0.333793277005)*x_ref[1]+(-0.540787491142)*x_ref[2]
        ref[1,4,1,1]=(-0.782297770239)*x_ref[0]+(0.965050908578)*x_ref[1]+(0.648337674864)*x_ref[2]
        ref[1,4,2,0]=(-0.425598399058)*x_ref[0]+(1.24755177094)*x_ref[1]+(-0.599571710505)*x_ref[2]
        ref[1,4,2,1]=(-0.584529028943)*x_ref[0]+(-0.115930785496)*x_ref[1]+(-0.412776678144)*x_ref[2]
        ref[2,0,0,0]=(-1.22220614711)*x_ref[0]+(0.991280082493)*x_ref[1]+(-0.771530362129)*x_ref[2]
        ref[2,0,0,1]=(-0.0408833973017)*x_ref[0]+(-0.878550136082)*x_ref[1]+(-1.27973588501)*x_ref[2]
        ref[2,0,1,0]=(0.380052050339)*x_ref[0]+(1.34070050501)*x_ref[1]+(-1.39080485735)*x_ref[2]
        ref[2,0,1,1]=(-0.725808759614)*x_ref[0]+(1.37568148722)*x_ref[1]+(-0.369008444335)*x_ref[2]
        ref[2,0,2,0]=(-0.563453230072)*x_ref[0]+(0.25114466425)*x_ref[1]+(-0.0602643663972)*x_ref[2]
        ref[2,0,2,1]=(-0.802381432423)*x_ref[0]+(0.22715480998)*x_ref[1]+(0.512734770933)*x_ref[2]
        ref[2,1,0,0]=(-0.522742058617)*x_ref[0]+(0.37241878501)*x_ref[1]+(0.691287986533)*x_ref[2]
        ref[2,1,0,1]=(-0.168275058841)*x_ref[0]+(-0.764400123274)*x_ref[1]+(-0.681242773507)*x_ref[2]
        ref[2,1,1,0]=(0.551464415842)*x_ref[0]+(1.03330737276)*x_ref[1]+(0.958549917295)*x_ref[2]
        ref[2,1,1,1]=(-0.909196447141)*x_ref[0]+(0.145246231629)*x_ref[1]+(-1.05582073367)*x_ref[2]
        ref[2,1,2,0]=(-0.638872672519)*x_ref[0]+(0.406342003401)*x_ref[1]+(-0.393150454497)*x_ref[2]
        ref[2,1,2,1]=(0.366774488001)*x_ref[0]+(0.365786897102)*x_ref[1]+(0.441878454676)*x_ref[2]
        ref[2,2,0,0]=(-0.261857371972)*x_ref[0]+(0.0318314524134)*x_ref[1]+(-0.101372223186)*x_ref[2]
        ref[2,2,0,1]=(-0.135781961265)*x_ref[0]+(-0.00345171998114)*x_ref[1]+(0.0101907167492)*x_ref[2]
        ref[2,2,1,0]=(-1.37428228456)*x_ref[0]+(-0.524953075499)*x_ref[1]+(-0.372440345968)*x_ref[2]
        ref[2,2,1,1]=(0.0540542424565)*x_ref[0]+(-1.03413875822)*x_ref[1]+(-0.899141477767)*x_ref[2]
        ref[2,2,2,0]=(-0.431640543873)*x_ref[0]+(0.934769305572)*x_ref[1]+(-0.848404080116)*x_ref[2]
        ref[2,2,2,1]=(0.984445227688)*x_ref[0]+(-0.207064309031)*x_ref[1]+(-1.75497257678)*x_ref[2]
        ref[2,3,0,0]=(-1.61836350351)*x_ref[0]+(0.885389572019)*x_ref[1]+(0.133218748941)*x_ref[2]
        ref[2,3,0,1]=(0.100716060215)*x_ref[0]+(-0.713421128342)*x_ref[1]+(1.16765406123)*x_ref[2]
        ref[2,3,1,0]=(0.497272818223)*x_ref[0]+(-0.698345898985)*x_ref[1]+(-0.643672453237)*x_ref[2]
        ref[2,3,1,1]=(-0.680652178847)*x_ref[0]+(-0.56686153354)*x_ref[1]+(-0.259410997801)*x_ref[2]
        ref[2,3,2,0]=(1.08214823373)*x_ref[0]+(-0.875408253309)*x_ref[1]+(-0.165330678277)*x_ref[2]
        ref[2,3,2,1]=(-0.251850346585)*x_ref[0]+(-0.233254346449)*x_ref[1]+(0.348350958039)*x_ref[2]
        ref[2,4,0,0]=(0.283802782063)*x_ref[0]+(0.338243660721)*x_ref[1]+(0.0400948594919)*x_ref[2]
        ref[2,4,0,1]=(-1.46364736479)*x_ref[0]+(0.738031226032)*x_ref[1]+(0.897193522713)*x_ref[2]
        ref[2,4,1,0]=(1.17637134366)*x_ref[0]+(-0.607880128773)*x_ref[1]+(-1.00344218409)*x_ref[2]
        ref[2,4,1,1]=(0.716275197822)*x_ref[0]+(1.21991657141)*x_ref[1]+(-0.0467328809597)*x_ref[2]
        ref[2,4,2,0]=(0.380704185608)*x_ref[0]+(0.603603708529)*x_ref[1]+(-0.820752063279)*x_ref[2]
        ref[2,4,2,1]=(1.39582555808)*x_ref[0]+(-0.473415991144)*x_ref[1]+(1.24135585822)*x_ref[2]
        ref[3,0,0,0]=(1.39071291597)*x_ref[0]+(-1.59041985271)*x_ref[1]+(0.60185319823)*x_ref[2]
        ref[3,0,0,1]=(-1.46801716969)*x_ref[0]+(-0.552777641699)*x_ref[1]+(1.23129633317)*x_ref[2]
        ref[3,0,1,0]=(1.64196247631)*x_ref[0]+(-0.343527059014)*x_ref[1]+(1.11486568074)*x_ref[2]
        ref[3,0,1,1]=(-1.57106459992)*x_ref[0]+(1.18660381744)*x_ref[1]+(0.495305193784)*x_ref[2]
        ref[3,0,2,0]=(-0.246662602621)*x_ref[0]+(0.980536147604)*x_ref[1]+(-1.29582037379)*x_ref[2]
        ref[3,0,2,1]=(0.351277952246)*x_ref[0]+(0.0189052721516)*x_ref[1]+(-0.109275345669)*x_ref[2]
        ref[3,1,0,0]=(-0.559164628236)*x_ref[0]+(-0.390894772538)*x_ref[1]+(0.657026552721)*x_ref[2]
        ref[3,1,0,1]=(-1.63523473278)*x_ref[0]+(-0.958777701493)*x_ref[1]+(0.0466950147837)*x_ref[2]
        ref[3,1,1,0]=(-0.913387269773)*x_ref[0]+(-0.752264610272)*x_ref[1]+(-1.00623906741)*x_ref[2]
        ref[3,1,1,1]=(0.703502728627)*x_ref[0]+(0.985524845396)*x_ref[1]+(-0.0966526090953)*x_ref[2]
        ref[3,1,2,0]=(-0.993367844074)*x_ref[0]+(-0.205401158091)*x_ref[1]+(0.416042784765)*x_ref[2]
        ref[3,1,2,1]=(0.299615272051)*x_ref[0]+(1.32572777489)*x_ref[1]+(0.338960058542)*x_ref[2]
        ref[3,2,0,0]=(-0.757820100577)*x_ref[0]+(0.644913131328)*x_ref[1]+(-0.660404804486)*x_ref[2]
        ref[3,2,0,1]=(-1.75641089093)*x_ref[0]+(-0.366399254245)*x_ref[1]+(-0.646900183385)*x_ref[2]
        ref[3,2,1,0]=(0.0912144040068)*x_ref[0]+(-0.857875163959)*x_ref[1]+(0.350343357173)*x_ref[2]
        ref[3,2,1,1]=(-1.80008916515)*x_ref[0]+(-1.83308297974)*x_ref[1]+(0.052958998062)*x_ref[2]
        ref[3,2,2,0]=(1.44092323696)*x_ref[0]+(-1.1831374133)*x_ref[1]+(-0.0323647992786)*x_ref[2]
        ref[3,2,2,1]=(-0.353641143336)*x_ref[0]+(0.774344611259)*x_ref[1]+(-0.842874663076)*x_ref[2]
        ref[3,3,0,0]=(-1.17418376796)*x_ref[0]+(0.334599071712)*x_ref[1]+(0.0621712582176)*x_ref[2]
        ref[3,3,0,1]=(1.1539137936)*x_ref[0]+(0.781617539757)*x_ref[1]+(1.6289484244)*x_ref[2]
        ref[3,3,1,0]=(1.6860969053)*x_ref[0]+(-0.936239288854)*x_ref[1]+(1.7596463612)*x_ref[2]
        ref[3,3,1,1]=(-1.11111948999)*x_ref[0]+(0.285869758308)*x_ref[1]+(1.69003878827)*x_ref[2]
        ref[3,3,2,0]=(-1.00166505759)*x_ref[0]+(-1.21067712793)*x_ref[1]+(0.422711238131)*x_ref[2]
        ref[3,3,2,1]=(0.49093698547)*x_ref[0]+(-0.952249146613)*x_ref[1]+(-0.955917587594)*x_ref[2]
        ref[3,4,0,0]=(-0.511109469881)*x_ref[0]+(0.630807398954)*x_ref[1]+(-0.0365100439309)*x_ref[2]
        ref[3,4,0,1]=(0.260323499449)*x_ref[0]+(-0.232312235272)*x_ref[1]+(-0.676872879397)*x_ref[2]
        ref[3,4,1,0]=(-0.327180688658)*x_ref[0]+(0.0260172986398)*x_ref[1]+(-1.59267263743)*x_ref[2]
        ref[3,4,1,1]=(0.135287590014)*x_ref[0]+(1.69850923514)*x_ref[1]+(-0.498682843917)*x_ref[2]
        ref[3,4,2,0]=(0.314853952103)*x_ref[0]+(-0.562331392142)*x_ref[1]+(0.726740096722)*x_ref[2]
        ref[3,4,2,1]=(-0.638129182205)*x_ref[0]+(-0.881463492466)*x_ref[1]+(-0.792625664675)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.648487700909)*x[0]+(-1.43364306566)*x[1]
        ref=(0.648487700909)*x_ref[0]+(-1.43364306566)*x_ref[1]
      else:
        s=(0.0554263047837)*x[0]+(0.627379593937)*x[1]+(0.740861554164)*x[2]
        ref=(0.0554263047837)*x_ref[0]+(0.627379593937)*x_ref[1]+(0.740861554164)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-1.61406386148)*x[0]+(0.644766840626)*x[1]
        s[1]=(-0.328446322604)*x[0]+(-0.277890186177)*x[1]
        ref[0]=(-1.61406386148)*x[0]+(0.644766840626)*x[1]
        ref[1]=(-0.328446322604)*x[0]+(-0.277890186177)*x[1]
      else:
        s[0]=(0.280680714031)*x[0]+(0.240612588198)*x[1]+(-0.289142363413)*x[2]
        s[1]=(-0.845801139442)*x[0]+(-1.55820065311)*x[1]+(0.103375183543)*x[2]
        ref[0]=(0.280680714031)*x[0]+(0.240612588198)*x[1]+(-0.289142363413)*x[2]
        ref[1]=(-0.845801139442)*x[0]+(-1.55820065311)*x[1]+(0.103375183543)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0481111909695)*x[0]+(-0.393526010503)*x[1]
        s[0,1]=(-0.284097772097)*x[0]+(0.618264565258)*x[1]
        s[0,2]=(0.401575436099)*x[0]+(-0.697718912762)*x[1]
        s[0,3]=(-1.49798792514)*x[0]+(-0.0425176867511)*x[1]
        s[0,4]=(0.213840882194)*x[0]+(-0.914489018365)*x[1]
        s[1,0]=(0.857166519205)*x[0]+(1.5282054335)*x[1]
        s[1,1]=(1.61091698609)*x[0]+(-0.548453984777)*x[1]
        s[1,2]=(-1.56266453199)*x[0]+(-1.66187429416)*x[1]
        s[1,3]=(0.343955167735)*x[0]+(0.745969147169)*x[1]
        s[1,4]=(-0.449555353075)*x[0]+(0.67139606475)*x[1]
        s[2,0]=(-1.35267387997)*x[0]+(0.827592755339)*x[1]
        s[2,1]=(-0.332416565825)*x[0]+(-0.570384450901)*x[1]
        s[2,2]=(0.228812981844)*x[0]+(-0.506320631557)*x[1]
        s[2,3]=(0.346948032916)*x[0]+(-0.0370468335786)*x[1]
        s[2,4]=(-0.38539371487)*x[0]+(-1.53408220383)*x[1]
        s[3,0]=(-0.446967528892)*x[0]+(-0.577472772294)*x[1]
        s[3,1]=(1.3016519192)*x[0]+(-1.21796536064)*x[1]
        s[3,2]=(-1.11173710879)*x[0]+(0.0982539955389)*x[1]
        s[3,3]=(0.398891317196)*x[0]+(0.0151699052193)*x[1]
        s[3,4]=(0.702195275642)*x[0]+(-0.355526973434)*x[1]
        ref[0,0]=(0.0481111909695)*x_ref[0]+(-0.393526010503)*x_ref[1]
        ref[0,1]=(-0.284097772097)*x_ref[0]+(0.618264565258)*x_ref[1]
        ref[0,2]=(0.401575436099)*x_ref[0]+(-0.697718912762)*x_ref[1]
        ref[0,3]=(-1.49798792514)*x_ref[0]+(-0.0425176867511)*x_ref[1]
        ref[0,4]=(0.213840882194)*x_ref[0]+(-0.914489018365)*x_ref[1]
        ref[1,0]=(0.857166519205)*x_ref[0]+(1.5282054335)*x_ref[1]
        ref[1,1]=(1.61091698609)*x_ref[0]+(-0.548453984777)*x_ref[1]
        ref[1,2]=(-1.56266453199)*x_ref[0]+(-1.66187429416)*x_ref[1]
        ref[1,3]=(0.343955167735)*x_ref[0]+(0.745969147169)*x_ref[1]
        ref[1,4]=(-0.449555353075)*x_ref[0]+(0.67139606475)*x_ref[1]
        ref[2,0]=(-1.35267387997)*x_ref[0]+(0.827592755339)*x_ref[1]
        ref[2,1]=(-0.332416565825)*x_ref[0]+(-0.570384450901)*x_ref[1]
        ref[2,2]=(0.228812981844)*x_ref[0]+(-0.506320631557)*x_ref[1]
        ref[2,3]=(0.346948032916)*x_ref[0]+(-0.0370468335786)*x_ref[1]
        ref[2,4]=(-0.38539371487)*x_ref[0]+(-1.53408220383)*x_ref[1]
        ref[3,0]=(-0.446967528892)*x_ref[0]+(-0.577472772294)*x_ref[1]
        ref[3,1]=(1.3016519192)*x_ref[0]+(-1.21796536064)*x_ref[1]
        ref[3,2]=(-1.11173710879)*x_ref[0]+(0.0982539955389)*x_ref[1]
        ref[3,3]=(0.398891317196)*x_ref[0]+(0.0151699052193)*x_ref[1]
        ref[3,4]=(0.702195275642)*x_ref[0]+(-0.355526973434)*x_ref[1]
      else:
        s[0,0]=(0.525077989935)*x[0]+(-0.129538863404)*x[1]+(-1.47575933759)*x[2]
        s[0,1]=(0.575439663003)*x[0]+(-1.40344366825)*x[1]+(0.221748855921)*x[2]
        s[0,2]=(0.171201653563)*x[0]+(-0.185657063083)*x[1]+(0.42305480444)*x[2]
        s[0,3]=(0.787118287986)*x[0]+(-1.43947391779)*x[1]+(0.322131428849)*x[2]
        s[0,4]=(-0.693857541121)*x[0]+(1.29660018396)*x[1]+(1.01092834477)*x[2]
        s[1,0]=(0.371970743598)*x[0]+(-0.744756043562)*x[1]+(0.194145828387)*x[2]
        s[1,1]=(1.23828194153)*x[0]+(-0.937084509616)*x[1]+(-1.47063754607)*x[2]
        s[1,2]=(-0.287641758427)*x[0]+(-0.42656557181)*x[1]+(1.49807905914)*x[2]
        s[1,3]=(-0.412169130593)*x[0]+(0.590235477443)*x[1]+(-0.575739422159)*x[2]
        s[1,4]=(-0.372408793303)*x[0]+(1.48960296195)*x[1]+(1.42957998207)*x[2]
        s[2,0]=(-0.356728277451)*x[0]+(0.562758907106)*x[1]+(-0.0825032302409)*x[2]
        s[2,1]=(-1.14497343552)*x[0]+(0.605393479159)*x[1]+(-0.286645936691)*x[2]
        s[2,2]=(-0.00750519201495)*x[0]+(-0.742525175154)*x[1]+(1.33814411456)*x[2]
        s[2,3]=(-1.12547614642)*x[0]+(0.291849175119)*x[1]+(0.745190340427)*x[2]
        s[2,4]=(0.0863930736605)*x[0]+(0.726972597715)*x[1]+(0.335180039249)*x[2]
        s[3,0]=(0.164315801392)*x[0]+(-0.276344360153)*x[1]+(0.364776703598)*x[2]
        s[3,1]=(0.980174491212)*x[0]+(0.920966782066)*x[1]+(1.6251437721)*x[2]
        s[3,2]=(0.722553089652)*x[0]+(-1.29237901172)*x[1]+(1.38442221241)*x[2]
        s[3,3]=(-1.1138747054)*x[0]+(0.184113285873)*x[1]+(0.113137432772)*x[2]
        s[3,4]=(0.560373055499)*x[0]+(0.347806062763)*x[1]+(-0.103028298482)*x[2]
        ref[0,0]=(0.525077989935)*x_ref[0]+(-0.129538863404)*x_ref[1]+(-1.47575933759)*x_ref[2]
        ref[0,1]=(0.575439663003)*x_ref[0]+(-1.40344366825)*x_ref[1]+(0.221748855921)*x_ref[2]
        ref[0,2]=(0.171201653563)*x_ref[0]+(-0.185657063083)*x_ref[1]+(0.42305480444)*x_ref[2]
        ref[0,3]=(0.787118287986)*x_ref[0]+(-1.43947391779)*x_ref[1]+(0.322131428849)*x_ref[2]
        ref[0,4]=(-0.693857541121)*x_ref[0]+(1.29660018396)*x_ref[1]+(1.01092834477)*x_ref[2]
        ref[1,0]=(0.371970743598)*x_ref[0]+(-0.744756043562)*x_ref[1]+(0.194145828387)*x_ref[2]
        ref[1,1]=(1.23828194153)*x_ref[0]+(-0.937084509616)*x_ref[1]+(-1.47063754607)*x_ref[2]
        ref[1,2]=(-0.287641758427)*x_ref[0]+(-0.42656557181)*x_ref[1]+(1.49807905914)*x_ref[2]
        ref[1,3]=(-0.412169130593)*x_ref[0]+(0.590235477443)*x_ref[1]+(-0.575739422159)*x_ref[2]
        ref[1,4]=(-0.372408793303)*x_ref[0]+(1.48960296195)*x_ref[1]+(1.42957998207)*x_ref[2]
        ref[2,0]=(-0.356728277451)*x_ref[0]+(0.562758907106)*x_ref[1]+(-0.0825032302409)*x_ref[2]
        ref[2,1]=(-1.14497343552)*x_ref[0]+(0.605393479159)*x_ref[1]+(-0.286645936691)*x_ref[2]
        ref[2,2]=(-0.00750519201495)*x_ref[0]+(-0.742525175154)*x_ref[1]+(1.33814411456)*x_ref[2]
        ref[2,3]=(-1.12547614642)*x_ref[0]+(0.291849175119)*x_ref[1]+(0.745190340427)*x_ref[2]
        ref[2,4]=(0.0863930736605)*x_ref[0]+(0.726972597715)*x_ref[1]+(0.335180039249)*x_ref[2]
        ref[3,0]=(0.164315801392)*x_ref[0]+(-0.276344360153)*x_ref[1]+(0.364776703598)*x_ref[2]
        ref[3,1]=(0.980174491212)*x_ref[0]+(0.920966782066)*x_ref[1]+(1.6251437721)*x_ref[2]
        ref[3,2]=(0.722553089652)*x_ref[0]+(-1.29237901172)*x_ref[1]+(1.38442221241)*x_ref[2]
        ref[3,3]=(-1.1138747054)*x_ref[0]+(0.184113285873)*x_ref[1]+(0.113137432772)*x_ref[2]
        ref[3,4]=(0.560373055499)*x_ref[0]+(0.347806062763)*x_ref[1]+(-0.103028298482)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.616803876826)*x[0]+(0.239741751767)*x[1]
        s[0,0,1]=(0.112656354167)*x[0]+(0.0390998528525)*x[1]
        s[0,1,0]=(1.02334404242)*x[0]+(1.16228451879)*x[1]
        s[0,1,1]=(1.26927845423)*x[0]+(-0.0309071031648)*x[1]
        s[1,0,0]=(0.154947909629)*x[0]+(1.41925252494)*x[1]
        s[1,0,1]=(0.299309779213)*x[0]+(0.558282387192)*x[1]
        s[1,1,0]=(0.748405958274)*x[0]+(-0.241956071057)*x[1]
        s[1,1,1]=(-0.100230866827)*x[0]+(-0.616595937224)*x[1]
        s[2,0,0]=(-0.485117813294)*x[0]+(-0.814821957713)*x[1]
        s[2,0,1]=(0.573476779093)*x[0]+(0.480686896264)*x[1]
        s[2,1,0]=(0.79651609962)*x[0]+(0.257600415705)*x[1]
        s[2,1,1]=(-0.162020224334)*x[0]+(1.2460751805)*x[1]
        s[3,0,0]=(0.467277953063)*x[0]+(0.254739285648)*x[1]
        s[3,0,1]=(-0.619802362841)*x[0]+(-1.30062481861)*x[1]
        s[3,1,0]=(0.160765639764)*x[0]+(-0.716696954941)*x[1]
        s[3,1,1]=(-0.0317856002877)*x[0]+(0.561299373553)*x[1]
        s[4,0,0]=(0.238090555098)*x[0]+(-0.99128263972)*x[1]
        s[4,0,1]=(-0.90912380586)*x[0]+(-0.32214847292)*x[1]
        s[4,1,0]=(1.67021862968)*x[0]+(-0.396622542538)*x[1]
        s[4,1,1]=(-0.118992684804)*x[0]+(-0.229200542459)*x[1]
        s[5,0,0]=(-0.124626687555)*x[0]+(0.429266692145)*x[1]
        s[5,0,1]=(1.40894998937)*x[0]+(-0.222278182789)*x[1]
        s[5,1,0]=(-0.291128816468)*x[0]+(1.33531261196)*x[1]
        s[5,1,1]=(-0.985994002551)*x[0]+(1.10232269681)*x[1]
        ref[0,0,0]=(-0.616803876826)*x_ref[0]+(0.239741751767)*x_ref[1]
        ref[0,0,1]=(0.112656354167)*x_ref[0]+(0.0390998528525)*x_ref[1]
        ref[0,1,0]=(1.02334404242)*x_ref[0]+(1.16228451879)*x_ref[1]
        ref[0,1,1]=(1.26927845423)*x_ref[0]+(-0.0309071031648)*x_ref[1]
        ref[1,0,0]=(0.154947909629)*x_ref[0]+(1.41925252494)*x_ref[1]
        ref[1,0,1]=(0.299309779213)*x_ref[0]+(0.558282387192)*x_ref[1]
        ref[1,1,0]=(0.748405958274)*x_ref[0]+(-0.241956071057)*x_ref[1]
        ref[1,1,1]=(-0.100230866827)*x_ref[0]+(-0.616595937224)*x_ref[1]
        ref[2,0,0]=(-0.485117813294)*x_ref[0]+(-0.814821957713)*x_ref[1]
        ref[2,0,1]=(0.573476779093)*x_ref[0]+(0.480686896264)*x_ref[1]
        ref[2,1,0]=(0.79651609962)*x_ref[0]+(0.257600415705)*x_ref[1]
        ref[2,1,1]=(-0.162020224334)*x_ref[0]+(1.2460751805)*x_ref[1]
        ref[3,0,0]=(0.467277953063)*x_ref[0]+(0.254739285648)*x_ref[1]
        ref[3,0,1]=(-0.619802362841)*x_ref[0]+(-1.30062481861)*x_ref[1]
        ref[3,1,0]=(0.160765639764)*x_ref[0]+(-0.716696954941)*x_ref[1]
        ref[3,1,1]=(-0.0317856002877)*x_ref[0]+(0.561299373553)*x_ref[1]
        ref[4,0,0]=(0.238090555098)*x_ref[0]+(-0.99128263972)*x_ref[1]
        ref[4,0,1]=(-0.90912380586)*x_ref[0]+(-0.32214847292)*x_ref[1]
        ref[4,1,0]=(1.67021862968)*x_ref[0]+(-0.396622542538)*x_ref[1]
        ref[4,1,1]=(-0.118992684804)*x_ref[0]+(-0.229200542459)*x_ref[1]
        ref[5,0,0]=(-0.124626687555)*x_ref[0]+(0.429266692145)*x_ref[1]
        ref[5,0,1]=(1.40894998937)*x_ref[0]+(-0.222278182789)*x_ref[1]
        ref[5,1,0]=(-0.291128816468)*x_ref[0]+(1.33531261196)*x_ref[1]
        ref[5,1,1]=(-0.985994002551)*x_ref[0]+(1.10232269681)*x_ref[1]
      else:
        s[0,0,0]=(-0.622303225663)*x[0]+(-0.468862605979)*x[1]+(-0.822269949108)*x[2]
        s[0,0,1]=(1.42712676639)*x[0]+(0.466554590146)*x[1]+(-0.416360290497)*x[2]
        s[0,1,0]=(-0.56276315017)*x[0]+(0.681130342695)*x[1]+(-1.150317219)*x[2]
        s[0,1,1]=(0.840875694603)*x[0]+(1.06129869177)*x[1]+(0.186757925344)*x[2]
        s[1,0,0]=(1.10693557715)*x[0]+(0.515133841775)*x[1]+(-1.46650509667)*x[2]
        s[1,0,1]=(0.469874083814)*x[0]+(0.712094004219)*x[1]+(0.0428537567242)*x[2]
        s[1,1,0]=(1.09283636268)*x[0]+(-0.922481846245)*x[1]+(0.611758101252)*x[2]
        s[1,1,1]=(0.32198205915)*x[0]+(0.75194760009)*x[1]+(-0.118025900786)*x[2]
        s[2,0,0]=(-1.52389124946)*x[0]+(1.49105249151)*x[1]+(-0.874107742996)*x[2]
        s[2,0,1]=(-0.90495363127)*x[0]+(0.078342943764)*x[1]+(0.219749629688)*x[2]
        s[2,1,0]=(-0.00849262013635)*x[0]+(0.00531993772995)*x[1]+(1.76477138966)*x[2]
        s[2,1,1]=(0.0396318942805)*x[0]+(0.630819877838)*x[1]+(-1.44621732026)*x[2]
        s[3,0,0]=(0.0109746356852)*x[0]+(-0.362009865261)*x[1]+(-0.538016468887)*x[2]
        s[3,0,1]=(-0.234158150184)*x[0]+(-1.44595936898)*x[1]+(0.329273103366)*x[2]
        s[3,1,0]=(0.146354345455)*x[0]+(0.138204478429)*x[1]+(1.61082500463)*x[2]
        s[3,1,1]=(1.16444593046)*x[0]+(-0.353850126255)*x[1]+(-0.0472752976526)*x[2]
        s[4,0,0]=(0.440054653358)*x[0]+(-1.16079243521)*x[1]+(-0.135698496809)*x[2]
        s[4,0,1]=(-1.89214756492)*x[0]+(-0.118895763118)*x[1]+(0.44262404148)*x[2]
        s[4,1,0]=(-1.13364355366)*x[0]+(1.47153317046)*x[1]+(0.519901952659)*x[2]
        s[4,1,1]=(0.340556700422)*x[0]+(0.270041963096)*x[1]+(0.303701099668)*x[2]
        s[5,0,0]=(0.779284769306)*x[0]+(0.91773222667)*x[1]+(0.870658706587)*x[2]
        s[5,0,1]=(-1.19937838023)*x[0]+(-0.182653397415)*x[1]+(0.229890290729)*x[2]
        s[5,1,0]=(-1.22173252282)*x[0]+(-0.465952231432)*x[1]+(-0.173356024515)*x[2]
        s[5,1,1]=(-0.175957737331)*x[0]+(0.512508313109)*x[1]+(-0.81846318618)*x[2]
        ref[0,0,0]=(-0.622303225663)*x_ref[0]+(-0.468862605979)*x_ref[1]+(-0.822269949108)*x_ref[2]
        ref[0,0,1]=(1.42712676639)*x_ref[0]+(0.466554590146)*x_ref[1]+(-0.416360290497)*x_ref[2]
        ref[0,1,0]=(-0.56276315017)*x_ref[0]+(0.681130342695)*x_ref[1]+(-1.150317219)*x_ref[2]
        ref[0,1,1]=(0.840875694603)*x_ref[0]+(1.06129869177)*x_ref[1]+(0.186757925344)*x_ref[2]
        ref[1,0,0]=(1.10693557715)*x_ref[0]+(0.515133841775)*x_ref[1]+(-1.46650509667)*x_ref[2]
        ref[1,0,1]=(0.469874083814)*x_ref[0]+(0.712094004219)*x_ref[1]+(0.0428537567242)*x_ref[2]
        ref[1,1,0]=(1.09283636268)*x_ref[0]+(-0.922481846245)*x_ref[1]+(0.611758101252)*x_ref[2]
        ref[1,1,1]=(0.32198205915)*x_ref[0]+(0.75194760009)*x_ref[1]+(-0.118025900786)*x_ref[2]
        ref[2,0,0]=(-1.52389124946)*x_ref[0]+(1.49105249151)*x_ref[1]+(-0.874107742996)*x_ref[2]
        ref[2,0,1]=(-0.90495363127)*x_ref[0]+(0.078342943764)*x_ref[1]+(0.219749629688)*x_ref[2]
        ref[2,1,0]=(-0.00849262013635)*x_ref[0]+(0.00531993772995)*x_ref[1]+(1.76477138966)*x_ref[2]
        ref[2,1,1]=(0.0396318942805)*x_ref[0]+(0.630819877838)*x_ref[1]+(-1.44621732026)*x_ref[2]
        ref[3,0,0]=(0.0109746356852)*x_ref[0]+(-0.362009865261)*x_ref[1]+(-0.538016468887)*x_ref[2]
        ref[3,0,1]=(-0.234158150184)*x_ref[0]+(-1.44595936898)*x_ref[1]+(0.329273103366)*x_ref[2]
        ref[3,1,0]=(0.146354345455)*x_ref[0]+(0.138204478429)*x_ref[1]+(1.61082500463)*x_ref[2]
        ref[3,1,1]=(1.16444593046)*x_ref[0]+(-0.353850126255)*x_ref[1]+(-0.0472752976526)*x_ref[2]
        ref[4,0,0]=(0.440054653358)*x_ref[0]+(-1.16079243521)*x_ref[1]+(-0.135698496809)*x_ref[2]
        ref[4,0,1]=(-1.89214756492)*x_ref[0]+(-0.118895763118)*x_ref[1]+(0.44262404148)*x_ref[2]
        ref[4,1,0]=(-1.13364355366)*x_ref[0]+(1.47153317046)*x_ref[1]+(0.519901952659)*x_ref[2]
        ref[4,1,1]=(0.340556700422)*x_ref[0]+(0.270041963096)*x_ref[1]+(0.303701099668)*x_ref[2]
        ref[5,0,0]=(0.779284769306)*x_ref[0]+(0.91773222667)*x_ref[1]+(0.870658706587)*x_ref[2]
        ref[5,0,1]=(-1.19937838023)*x_ref[0]+(-0.182653397415)*x_ref[1]+(0.229890290729)*x_ref[2]
        ref[5,1,0]=(-1.22173252282)*x_ref[0]+(-0.465952231432)*x_ref[1]+(-0.173356024515)*x_ref[2]
        ref[5,1,1]=(-0.175957737331)*x_ref[0]+(0.512508313109)*x_ref[1]+(-0.81846318618)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.263699858395)*x[0]+(0.863340165591)*x[1]
        s[0,0,0,1]=(0.168551523361)*x[0]+(0.822827042977)*x[1]
        s[0,0,1,0]=(0.922495824101)*x[0]+(1.10277493051)*x[1]
        s[0,0,1,1]=(0.858215828807)*x[0]+(0.27444875925)*x[1]
        s[0,0,2,0]=(0.203856130696)*x[0]+(-1.32433677722)*x[1]
        s[0,0,2,1]=(-1.2222254241)*x[0]+(-0.53464384301)*x[1]
        s[0,1,0,0]=(0.0467230980779)*x[0]+(1.64188619595)*x[1]
        s[0,1,0,1]=(0.917582861623)*x[0]+(-1.01711076348)*x[1]
        s[0,1,1,0]=(-0.70445337121)*x[0]+(-0.499786458329)*x[1]
        s[0,1,1,1]=(1.03899670749)*x[0]+(-1.43262408006)*x[1]
        s[0,1,2,0]=(0.409224950113)*x[0]+(1.48049510634)*x[1]
        s[0,1,2,1]=(-0.639323586948)*x[0]+(0.41324495759)*x[1]
        s[0,2,0,0]=(-1.04843809688)*x[0]+(0.409833243907)*x[1]
        s[0,2,0,1]=(1.2606980391)*x[0]+(1.05476378621)*x[1]
        s[0,2,1,0]=(-0.0653124068721)*x[0]+(1.68355228676)*x[1]
        s[0,2,1,1]=(1.56886755247)*x[0]+(-1.67289454758)*x[1]
        s[0,2,2,0]=(0.542254512691)*x[0]+(-0.574877434565)*x[1]
        s[0,2,2,1]=(-1.79104846611)*x[0]+(0.752170029246)*x[1]
        s[0,3,0,0]=(-0.804486243116)*x[0]+(1.58214181283)*x[1]
        s[0,3,0,1]=(0.268657213023)*x[0]+(-0.632746875992)*x[1]
        s[0,3,1,0]=(0.748200841554)*x[0]+(-0.372804380425)*x[1]
        s[0,3,1,1]=(0.532346891312)*x[0]+(0.0944645102486)*x[1]
        s[0,3,2,0]=(0.255430361824)*x[0]+(0.493809212019)*x[1]
        s[0,3,2,1]=(1.06219716938)*x[0]+(0.0256120510252)*x[1]
        s[0,4,0,0]=(-1.11393336407)*x[0]+(0.0040462142619)*x[1]
        s[0,4,0,1]=(0.290958676921)*x[0]+(1.43659749943)*x[1]
        s[0,4,1,0]=(-0.921614727859)*x[0]+(-0.730623491689)*x[1]
        s[0,4,1,1]=(-0.539009340284)*x[0]+(0.162040517816)*x[1]
        s[0,4,2,0]=(-0.166431286126)*x[0]+(-0.881280007454)*x[1]
        s[0,4,2,1]=(0.939057506826)*x[0]+(-0.798845422966)*x[1]
        s[1,0,0,0]=(-1.48556828853)*x[0]+(0.398947142701)*x[1]
        s[1,0,0,1]=(-0.867994319546)*x[0]+(-1.51760635946)*x[1]
        s[1,0,1,0]=(-1.3436690984)*x[0]+(0.0587809657777)*x[1]
        s[1,0,1,1]=(-0.358397464109)*x[0]+(-0.110074947013)*x[1]
        s[1,0,2,0]=(-1.65339451931)*x[0]+(-0.628438027419)*x[1]
        s[1,0,2,1]=(0.75949197524)*x[0]+(1.47344312464)*x[1]
        s[1,1,0,0]=(-0.184897359409)*x[0]+(0.997493671981)*x[1]
        s[1,1,0,1]=(0.0348263424069)*x[0]+(-0.113374992622)*x[1]
        s[1,1,1,0]=(1.48734955276)*x[0]+(-0.789492365846)*x[1]
        s[1,1,1,1]=(0.865097624364)*x[0]+(0.117588671443)*x[1]
        s[1,1,2,0]=(0.195936219884)*x[0]+(-0.948389540535)*x[1]
        s[1,1,2,1]=(1.11314083647)*x[0]+(0.215110041922)*x[1]
        s[1,2,0,0]=(-0.920590092452)*x[0]+(-0.612482637981)*x[1]
        s[1,2,0,1]=(1.04301519124)*x[0]+(-0.0629305143129)*x[1]
        s[1,2,1,0]=(0.382136489277)*x[0]+(-0.230145907219)*x[1]
        s[1,2,1,1]=(-0.0252796835879)*x[0]+(-0.900125981212)*x[1]
        s[1,2,2,0]=(0.965035634162)*x[0]+(-1.22815714841)*x[1]
        s[1,2,2,1]=(0.357785163717)*x[0]+(1.27915764344)*x[1]
        s[1,3,0,0]=(-1.27549226474)*x[0]+(1.24432292599)*x[1]
        s[1,3,0,1]=(-0.0383725074991)*x[0]+(-0.349361287486)*x[1]
        s[1,3,1,0]=(-0.200272394326)*x[0]+(1.06221698364)*x[1]
        s[1,3,1,1]=(-0.215946498482)*x[0]+(1.01531234763)*x[1]
        s[1,3,2,0]=(-1.22728454222)*x[0]+(0.907627178806)*x[1]
        s[1,3,2,1]=(-1.00398418868)*x[0]+(-0.199827024316)*x[1]
        s[1,4,0,0]=(-0.00303564690889)*x[0]+(0.760141903922)*x[1]
        s[1,4,0,1]=(-0.456280795832)*x[0]+(0.986442355989)*x[1]
        s[1,4,1,0]=(-0.318954368532)*x[0]+(-0.966894508646)*x[1]
        s[1,4,1,1]=(0.65555609342)*x[0]+(-0.526267162413)*x[1]
        s[1,4,2,0]=(1.72927006133)*x[0]+(1.65889052032)*x[1]
        s[1,4,2,1]=(0.340218940988)*x[0]+(-0.815180133977)*x[1]
        s[2,0,0,0]=(-0.910985187877)*x[0]+(-1.31360663975)*x[1]
        s[2,0,0,1]=(-0.174389735675)*x[0]+(0.268073959103)*x[1]
        s[2,0,1,0]=(-0.111243591678)*x[0]+(-0.7123465972)*x[1]
        s[2,0,1,1]=(1.03340508201)*x[0]+(-1.15516978578)*x[1]
        s[2,0,2,0]=(0.0791763687033)*x[0]+(-0.0239321376253)*x[1]
        s[2,0,2,1]=(-0.0296109649235)*x[0]+(-0.292837031166)*x[1]
        s[2,1,0,0]=(-0.00911908859083)*x[0]+(-1.17279969051)*x[1]
        s[2,1,0,1]=(-1.06648321801)*x[0]+(0.126622301021)*x[1]
        s[2,1,1,0]=(0.275881626889)*x[0]+(1.69565501602)*x[1]
        s[2,1,1,1]=(-0.409495321219)*x[0]+(-0.407778373033)*x[1]
        s[2,1,2,0]=(1.33276969465)*x[0]+(-1.17532941128)*x[1]
        s[2,1,2,1]=(-0.866150634496)*x[0]+(0.797755014562)*x[1]
        s[2,2,0,0]=(-0.106882476852)*x[0]+(-0.249249128227)*x[1]
        s[2,2,0,1]=(-0.520287837168)*x[0]+(0.338818467136)*x[1]
        s[2,2,1,0]=(-0.852757899877)*x[0]+(0.321704892591)*x[1]
        s[2,2,1,1]=(0.142520864112)*x[0]+(-1.05840136269)*x[1]
        s[2,2,2,0]=(-0.715836025391)*x[0]+(0.22306435555)*x[1]
        s[2,2,2,1]=(-0.586149670377)*x[0]+(-0.516267803245)*x[1]
        s[2,3,0,0]=(-0.950483491273)*x[0]+(0.624915645607)*x[1]
        s[2,3,0,1]=(-0.217697021618)*x[0]+(0.0442099818289)*x[1]
        s[2,3,1,0]=(-1.01046266386)*x[0]+(-1.48154228284)*x[1]
        s[2,3,1,1]=(-1.3050947534)*x[0]+(-0.998384196826)*x[1]
        s[2,3,2,0]=(0.267414868071)*x[0]+(1.15992550636)*x[1]
        s[2,3,2,1]=(-0.862214432382)*x[0]+(0.292670280728)*x[1]
        s[2,4,0,0]=(-0.134428442874)*x[0]+(0.719512272598)*x[1]
        s[2,4,0,1]=(-1.67096566574)*x[0]+(0.43067843233)*x[1]
        s[2,4,1,0]=(0.188806295154)*x[0]+(-1.54252342316)*x[1]
        s[2,4,1,1]=(0.0976229597501)*x[0]+(-0.937051784938)*x[1]
        s[2,4,2,0]=(-0.00596882510121)*x[0]+(-0.312957453878)*x[1]
        s[2,4,2,1]=(-1.02156076253)*x[0]+(-0.896718827796)*x[1]
        s[3,0,0,0]=(0.0133947002051)*x[0]+(0.238649243438)*x[1]
        s[3,0,0,1]=(1.00467946164)*x[0]+(1.34133983171)*x[1]
        s[3,0,1,0]=(0.341713854404)*x[0]+(0.511010566842)*x[1]
        s[3,0,1,1]=(0.546165677394)*x[0]+(0.321446986518)*x[1]
        s[3,0,2,0]=(-0.764072559141)*x[0]+(-0.312532010803)*x[1]
        s[3,0,2,1]=(-0.811247938505)*x[0]+(-0.939324734555)*x[1]
        s[3,1,0,0]=(0.572111068237)*x[0]+(-0.0530416981116)*x[1]
        s[3,1,0,1]=(0.837385483064)*x[0]+(0.928431354228)*x[1]
        s[3,1,1,0]=(0.238702836098)*x[0]+(0.182536052546)*x[1]
        s[3,1,1,1]=(0.461955717914)*x[0]+(0.0256860825024)*x[1]
        s[3,1,2,0]=(-0.787501340181)*x[0]+(-0.828497466795)*x[1]
        s[3,1,2,1]=(0.348032537213)*x[0]+(-0.984456869176)*x[1]
        s[3,2,0,0]=(0.588847373206)*x[0]+(1.84398386935)*x[1]
        s[3,2,0,1]=(-0.2085469075)*x[0]+(-0.244926359847)*x[1]
        s[3,2,1,0]=(-1.62397953336)*x[0]+(-0.102475340253)*x[1]
        s[3,2,1,1]=(0.455695325345)*x[0]+(0.552529049227)*x[1]
        s[3,2,2,0]=(0.734375962799)*x[0]+(0.279367642817)*x[1]
        s[3,2,2,1]=(1.52878551603)*x[0]+(-1.43594288923)*x[1]
        s[3,3,0,0]=(-0.331774589719)*x[0]+(-0.852014090099)*x[1]
        s[3,3,0,1]=(0.0495293070849)*x[0]+(1.70225124669)*x[1]
        s[3,3,1,0]=(1.37894662758)*x[0]+(0.350193222832)*x[1]
        s[3,3,1,1]=(0.00570139162378)*x[0]+(-0.37112295664)*x[1]
        s[3,3,2,0]=(-0.205033339669)*x[0]+(-0.695461477152)*x[1]
        s[3,3,2,1]=(0.12868380026)*x[0]+(0.764317872049)*x[1]
        s[3,4,0,0]=(-1.04309352143)*x[0]+(-0.805487628821)*x[1]
        s[3,4,0,1]=(-0.496744676221)*x[0]+(-0.13812443637)*x[1]
        s[3,4,1,0]=(0.0339710647687)*x[0]+(0.0274093213606)*x[1]
        s[3,4,1,1]=(-0.516837676304)*x[0]+(-0.119504191446)*x[1]
        s[3,4,2,0]=(-0.472643263791)*x[0]+(-0.337262686677)*x[1]
        s[3,4,2,1]=(-1.34135241071)*x[0]+(-0.185936541491)*x[1]
        ref[0,0,0,0]=(-0.263699858395)*x_ref[0]+(0.863340165591)*x_ref[1]
        ref[0,0,0,1]=(0.168551523361)*x_ref[0]+(0.822827042977)*x_ref[1]
        ref[0,0,1,0]=(0.922495824101)*x_ref[0]+(1.10277493051)*x_ref[1]
        ref[0,0,1,1]=(0.858215828807)*x_ref[0]+(0.27444875925)*x_ref[1]
        ref[0,0,2,0]=(0.203856130696)*x_ref[0]+(-1.32433677722)*x_ref[1]
        ref[0,0,2,1]=(-1.2222254241)*x_ref[0]+(-0.53464384301)*x_ref[1]
        ref[0,1,0,0]=(0.0467230980779)*x_ref[0]+(1.64188619595)*x_ref[1]
        ref[0,1,0,1]=(0.917582861623)*x_ref[0]+(-1.01711076348)*x_ref[1]
        ref[0,1,1,0]=(-0.70445337121)*x_ref[0]+(-0.499786458329)*x_ref[1]
        ref[0,1,1,1]=(1.03899670749)*x_ref[0]+(-1.43262408006)*x_ref[1]
        ref[0,1,2,0]=(0.409224950113)*x_ref[0]+(1.48049510634)*x_ref[1]
        ref[0,1,2,1]=(-0.639323586948)*x_ref[0]+(0.41324495759)*x_ref[1]
        ref[0,2,0,0]=(-1.04843809688)*x_ref[0]+(0.409833243907)*x_ref[1]
        ref[0,2,0,1]=(1.2606980391)*x_ref[0]+(1.05476378621)*x_ref[1]
        ref[0,2,1,0]=(-0.0653124068721)*x_ref[0]+(1.68355228676)*x_ref[1]
        ref[0,2,1,1]=(1.56886755247)*x_ref[0]+(-1.67289454758)*x_ref[1]
        ref[0,2,2,0]=(0.542254512691)*x_ref[0]+(-0.574877434565)*x_ref[1]
        ref[0,2,2,1]=(-1.79104846611)*x_ref[0]+(0.752170029246)*x_ref[1]
        ref[0,3,0,0]=(-0.804486243116)*x_ref[0]+(1.58214181283)*x_ref[1]
        ref[0,3,0,1]=(0.268657213023)*x_ref[0]+(-0.632746875992)*x_ref[1]
        ref[0,3,1,0]=(0.748200841554)*x_ref[0]+(-0.372804380425)*x_ref[1]
        ref[0,3,1,1]=(0.532346891312)*x_ref[0]+(0.0944645102486)*x_ref[1]
        ref[0,3,2,0]=(0.255430361824)*x_ref[0]+(0.493809212019)*x_ref[1]
        ref[0,3,2,1]=(1.06219716938)*x_ref[0]+(0.0256120510252)*x_ref[1]
        ref[0,4,0,0]=(-1.11393336407)*x_ref[0]+(0.0040462142619)*x_ref[1]
        ref[0,4,0,1]=(0.290958676921)*x_ref[0]+(1.43659749943)*x_ref[1]
        ref[0,4,1,0]=(-0.921614727859)*x_ref[0]+(-0.730623491689)*x_ref[1]
        ref[0,4,1,1]=(-0.539009340284)*x_ref[0]+(0.162040517816)*x_ref[1]
        ref[0,4,2,0]=(-0.166431286126)*x_ref[0]+(-0.881280007454)*x_ref[1]
        ref[0,4,2,1]=(0.939057506826)*x_ref[0]+(-0.798845422966)*x_ref[1]
        ref[1,0,0,0]=(-1.48556828853)*x_ref[0]+(0.398947142701)*x_ref[1]
        ref[1,0,0,1]=(-0.867994319546)*x_ref[0]+(-1.51760635946)*x_ref[1]
        ref[1,0,1,0]=(-1.3436690984)*x_ref[0]+(0.0587809657777)*x_ref[1]
        ref[1,0,1,1]=(-0.358397464109)*x_ref[0]+(-0.110074947013)*x_ref[1]
        ref[1,0,2,0]=(-1.65339451931)*x_ref[0]+(-0.628438027419)*x_ref[1]
        ref[1,0,2,1]=(0.75949197524)*x_ref[0]+(1.47344312464)*x_ref[1]
        ref[1,1,0,0]=(-0.184897359409)*x_ref[0]+(0.997493671981)*x_ref[1]
        ref[1,1,0,1]=(0.0348263424069)*x_ref[0]+(-0.113374992622)*x_ref[1]
        ref[1,1,1,0]=(1.48734955276)*x_ref[0]+(-0.789492365846)*x_ref[1]
        ref[1,1,1,1]=(0.865097624364)*x_ref[0]+(0.117588671443)*x_ref[1]
        ref[1,1,2,0]=(0.195936219884)*x_ref[0]+(-0.948389540535)*x_ref[1]
        ref[1,1,2,1]=(1.11314083647)*x_ref[0]+(0.215110041922)*x_ref[1]
        ref[1,2,0,0]=(-0.920590092452)*x_ref[0]+(-0.612482637981)*x_ref[1]
        ref[1,2,0,1]=(1.04301519124)*x_ref[0]+(-0.0629305143129)*x_ref[1]
        ref[1,2,1,0]=(0.382136489277)*x_ref[0]+(-0.230145907219)*x_ref[1]
        ref[1,2,1,1]=(-0.0252796835879)*x_ref[0]+(-0.900125981212)*x_ref[1]
        ref[1,2,2,0]=(0.965035634162)*x_ref[0]+(-1.22815714841)*x_ref[1]
        ref[1,2,2,1]=(0.357785163717)*x_ref[0]+(1.27915764344)*x_ref[1]
        ref[1,3,0,0]=(-1.27549226474)*x_ref[0]+(1.24432292599)*x_ref[1]
        ref[1,3,0,1]=(-0.0383725074991)*x_ref[0]+(-0.349361287486)*x_ref[1]
        ref[1,3,1,0]=(-0.200272394326)*x_ref[0]+(1.06221698364)*x_ref[1]
        ref[1,3,1,1]=(-0.215946498482)*x_ref[0]+(1.01531234763)*x_ref[1]
        ref[1,3,2,0]=(-1.22728454222)*x_ref[0]+(0.907627178806)*x_ref[1]
        ref[1,3,2,1]=(-1.00398418868)*x_ref[0]+(-0.199827024316)*x_ref[1]
        ref[1,4,0,0]=(-0.00303564690889)*x_ref[0]+(0.760141903922)*x_ref[1]
        ref[1,4,0,1]=(-0.456280795832)*x_ref[0]+(0.986442355989)*x_ref[1]
        ref[1,4,1,0]=(-0.318954368532)*x_ref[0]+(-0.966894508646)*x_ref[1]
        ref[1,4,1,1]=(0.65555609342)*x_ref[0]+(-0.526267162413)*x_ref[1]
        ref[1,4,2,0]=(1.72927006133)*x_ref[0]+(1.65889052032)*x_ref[1]
        ref[1,4,2,1]=(0.340218940988)*x_ref[0]+(-0.815180133977)*x_ref[1]
        ref[2,0,0,0]=(-0.910985187877)*x_ref[0]+(-1.31360663975)*x_ref[1]
        ref[2,0,0,1]=(-0.174389735675)*x_ref[0]+(0.268073959103)*x_ref[1]
        ref[2,0,1,0]=(-0.111243591678)*x_ref[0]+(-0.7123465972)*x_ref[1]
        ref[2,0,1,1]=(1.03340508201)*x_ref[0]+(-1.15516978578)*x_ref[1]
        ref[2,0,2,0]=(0.0791763687033)*x_ref[0]+(-0.0239321376253)*x_ref[1]
        ref[2,0,2,1]=(-0.0296109649235)*x_ref[0]+(-0.292837031166)*x_ref[1]
        ref[2,1,0,0]=(-0.00911908859083)*x_ref[0]+(-1.17279969051)*x_ref[1]
        ref[2,1,0,1]=(-1.06648321801)*x_ref[0]+(0.126622301021)*x_ref[1]
        ref[2,1,1,0]=(0.275881626889)*x_ref[0]+(1.69565501602)*x_ref[1]
        ref[2,1,1,1]=(-0.409495321219)*x_ref[0]+(-0.407778373033)*x_ref[1]
        ref[2,1,2,0]=(1.33276969465)*x_ref[0]+(-1.17532941128)*x_ref[1]
        ref[2,1,2,1]=(-0.866150634496)*x_ref[0]+(0.797755014562)*x_ref[1]
        ref[2,2,0,0]=(-0.106882476852)*x_ref[0]+(-0.249249128227)*x_ref[1]
        ref[2,2,0,1]=(-0.520287837168)*x_ref[0]+(0.338818467136)*x_ref[1]
        ref[2,2,1,0]=(-0.852757899877)*x_ref[0]+(0.321704892591)*x_ref[1]
        ref[2,2,1,1]=(0.142520864112)*x_ref[0]+(-1.05840136269)*x_ref[1]
        ref[2,2,2,0]=(-0.715836025391)*x_ref[0]+(0.22306435555)*x_ref[1]
        ref[2,2,2,1]=(-0.586149670377)*x_ref[0]+(-0.516267803245)*x_ref[1]
        ref[2,3,0,0]=(-0.950483491273)*x_ref[0]+(0.624915645607)*x_ref[1]
        ref[2,3,0,1]=(-0.217697021618)*x_ref[0]+(0.0442099818289)*x_ref[1]
        ref[2,3,1,0]=(-1.01046266386)*x_ref[0]+(-1.48154228284)*x_ref[1]
        ref[2,3,1,1]=(-1.3050947534)*x_ref[0]+(-0.998384196826)*x_ref[1]
        ref[2,3,2,0]=(0.267414868071)*x_ref[0]+(1.15992550636)*x_ref[1]
        ref[2,3,2,1]=(-0.862214432382)*x_ref[0]+(0.292670280728)*x_ref[1]
        ref[2,4,0,0]=(-0.134428442874)*x_ref[0]+(0.719512272598)*x_ref[1]
        ref[2,4,0,1]=(-1.67096566574)*x_ref[0]+(0.43067843233)*x_ref[1]
        ref[2,4,1,0]=(0.188806295154)*x_ref[0]+(-1.54252342316)*x_ref[1]
        ref[2,4,1,1]=(0.0976229597501)*x_ref[0]+(-0.937051784938)*x_ref[1]
        ref[2,4,2,0]=(-0.00596882510121)*x_ref[0]+(-0.312957453878)*x_ref[1]
        ref[2,4,2,1]=(-1.02156076253)*x_ref[0]+(-0.896718827796)*x_ref[1]
        ref[3,0,0,0]=(0.0133947002051)*x_ref[0]+(0.238649243438)*x_ref[1]
        ref[3,0,0,1]=(1.00467946164)*x_ref[0]+(1.34133983171)*x_ref[1]
        ref[3,0,1,0]=(0.341713854404)*x_ref[0]+(0.511010566842)*x_ref[1]
        ref[3,0,1,1]=(0.546165677394)*x_ref[0]+(0.321446986518)*x_ref[1]
        ref[3,0,2,0]=(-0.764072559141)*x_ref[0]+(-0.312532010803)*x_ref[1]
        ref[3,0,2,1]=(-0.811247938505)*x_ref[0]+(-0.939324734555)*x_ref[1]
        ref[3,1,0,0]=(0.572111068237)*x_ref[0]+(-0.0530416981116)*x_ref[1]
        ref[3,1,0,1]=(0.837385483064)*x_ref[0]+(0.928431354228)*x_ref[1]
        ref[3,1,1,0]=(0.238702836098)*x_ref[0]+(0.182536052546)*x_ref[1]
        ref[3,1,1,1]=(0.461955717914)*x_ref[0]+(0.0256860825024)*x_ref[1]
        ref[3,1,2,0]=(-0.787501340181)*x_ref[0]+(-0.828497466795)*x_ref[1]
        ref[3,1,2,1]=(0.348032537213)*x_ref[0]+(-0.984456869176)*x_ref[1]
        ref[3,2,0,0]=(0.588847373206)*x_ref[0]+(1.84398386935)*x_ref[1]
        ref[3,2,0,1]=(-0.2085469075)*x_ref[0]+(-0.244926359847)*x_ref[1]
        ref[3,2,1,0]=(-1.62397953336)*x_ref[0]+(-0.102475340253)*x_ref[1]
        ref[3,2,1,1]=(0.455695325345)*x_ref[0]+(0.552529049227)*x_ref[1]
        ref[3,2,2,0]=(0.734375962799)*x_ref[0]+(0.279367642817)*x_ref[1]
        ref[3,2,2,1]=(1.52878551603)*x_ref[0]+(-1.43594288923)*x_ref[1]
        ref[3,3,0,0]=(-0.331774589719)*x_ref[0]+(-0.852014090099)*x_ref[1]
        ref[3,3,0,1]=(0.0495293070849)*x_ref[0]+(1.70225124669)*x_ref[1]
        ref[3,3,1,0]=(1.37894662758)*x_ref[0]+(0.350193222832)*x_ref[1]
        ref[3,3,1,1]=(0.00570139162378)*x_ref[0]+(-0.37112295664)*x_ref[1]
        ref[3,3,2,0]=(-0.205033339669)*x_ref[0]+(-0.695461477152)*x_ref[1]
        ref[3,3,2,1]=(0.12868380026)*x_ref[0]+(0.764317872049)*x_ref[1]
        ref[3,4,0,0]=(-1.04309352143)*x_ref[0]+(-0.805487628821)*x_ref[1]
        ref[3,4,0,1]=(-0.496744676221)*x_ref[0]+(-0.13812443637)*x_ref[1]
        ref[3,4,1,0]=(0.0339710647687)*x_ref[0]+(0.0274093213606)*x_ref[1]
        ref[3,4,1,1]=(-0.516837676304)*x_ref[0]+(-0.119504191446)*x_ref[1]
        ref[3,4,2,0]=(-0.472643263791)*x_ref[0]+(-0.337262686677)*x_ref[1]
        ref[3,4,2,1]=(-1.34135241071)*x_ref[0]+(-0.185936541491)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.0655735713454)*x[0]+(-0.69619555181)*x[1]+(0.881033804166)*x[2]
        s[0,0,0,1]=(-0.886101209867)*x[0]+(0.0852736523075)*x[1]+(0.514766114136)*x[2]
        s[0,0,1,0]=(-0.708903115024)*x[0]+(-0.949542800872)*x[1]+(-1.07763550768)*x[2]
        s[0,0,1,1]=(0.970797729322)*x[0]+(-0.237719575345)*x[1]+(1.54252127667)*x[2]
        s[0,0,2,0]=(-0.112910964189)*x[0]+(0.813990286281)*x[1]+(0.437139526752)*x[2]
        s[0,0,2,1]=(0.270076285201)*x[0]+(-1.04231968668)*x[1]+(1.08793884214)*x[2]
        s[0,1,0,0]=(-0.376091612547)*x[0]+(0.965045498929)*x[1]+(-0.516825182321)*x[2]
        s[0,1,0,1]=(-0.922174392864)*x[0]+(1.12745699436)*x[1]+(0.647593149683)*x[2]
        s[0,1,1,0]=(-1.62204275282)*x[0]+(0.64585305798)*x[1]+(-1.41457161128)*x[2]
        s[0,1,1,1]=(-0.412715098923)*x[0]+(0.00317118495093)*x[1]+(-0.848053699577)*x[2]
        s[0,1,2,0]=(-1.33788126072)*x[0]+(-0.370246465439)*x[1]+(0.149368781369)*x[2]
        s[0,1,2,1]=(-0.0432149867078)*x[0]+(0.20428198228)*x[1]+(0.685699546484)*x[2]
        s[0,2,0,0]=(-0.849779864072)*x[0]+(-0.465408292856)*x[1]+(0.12942183587)*x[2]
        s[0,2,0,1]=(-0.533958302867)*x[0]+(0.742004817382)*x[1]+(-0.146827358028)*x[2]
        s[0,2,1,0]=(-0.100072250083)*x[0]+(-1.24645970838)*x[1]+(1.45346380514)*x[2]
        s[0,2,1,1]=(-0.718115035714)*x[0]+(0.572444837212)*x[1]+(-0.0961841883909)*x[2]
        s[0,2,2,0]=(-1.6396722492)*x[0]+(-0.657016455181)*x[1]+(-0.52001961538)*x[2]
        s[0,2,2,1]=(-0.0161596921482)*x[0]+(-0.130395174634)*x[1]+(1.68490335057)*x[2]
        s[0,3,0,0]=(0.16054789145)*x[0]+(-1.35250823955)*x[1]+(0.663123269205)*x[2]
        s[0,3,0,1]=(0.307153218553)*x[0]+(0.705376328926)*x[1]+(-0.335478576645)*x[2]
        s[0,3,1,0]=(0.499629878974)*x[0]+(-0.389863260213)*x[1]+(-0.398638918232)*x[2]
        s[0,3,1,1]=(0.229760417346)*x[0]+(-0.382704654178)*x[1]+(0.044859414166)*x[2]
        s[0,3,2,0]=(0.402905751731)*x[0]+(-0.433748870487)*x[1]+(1.05246387058)*x[2]
        s[0,3,2,1]=(0.620966047942)*x[0]+(0.778166847286)*x[1]+(-0.684881429118)*x[2]
        s[0,4,0,0]=(-0.0429218580761)*x[0]+(0.263402329307)*x[1]+(0.314852944534)*x[2]
        s[0,4,0,1]=(0.112989114824)*x[0]+(-0.407308926569)*x[1]+(-0.581889344395)*x[2]
        s[0,4,1,0]=(1.46986587487)*x[0]+(-0.360451075631)*x[1]+(1.41473323373)*x[2]
        s[0,4,1,1]=(-0.0898945660646)*x[0]+(-1.24644050789)*x[1]+(-0.344493513462)*x[2]
        s[0,4,2,0]=(-0.383865368385)*x[0]+(0.758209662119)*x[1]+(0.737714808779)*x[2]
        s[0,4,2,1]=(-0.282690133746)*x[0]+(0.520513691755)*x[1]+(-0.38353403004)*x[2]
        s[1,0,0,0]=(-0.233823433606)*x[0]+(0.627115553859)*x[1]+(0.240814969095)*x[2]
        s[1,0,0,1]=(1.65536823794)*x[0]+(0.577386580179)*x[1]+(0.64299358654)*x[2]
        s[1,0,1,0]=(-0.304849514801)*x[0]+(-1.26879113223)*x[1]+(-0.257428355176)*x[2]
        s[1,0,1,1]=(0.313165425271)*x[0]+(0.648645104314)*x[1]+(0.152429094691)*x[2]
        s[1,0,2,0]=(-1.18245997349)*x[0]+(-0.990110942468)*x[1]+(-0.528372216957)*x[2]
        s[1,0,2,1]=(-0.497751288211)*x[0]+(1.27633543739)*x[1]+(1.29106845801)*x[2]
        s[1,1,0,0]=(-0.401807344422)*x[0]+(1.00839022632)*x[1]+(0.855275309873)*x[2]
        s[1,1,0,1]=(-1.22469413181)*x[0]+(0.946705273705)*x[1]+(1.02166037335)*x[2]
        s[1,1,1,0]=(0.0800997438684)*x[0]+(0.232002850135)*x[1]+(1.72304848773)*x[2]
        s[1,1,1,1]=(-1.14517485603)*x[0]+(-1.68921566046)*x[1]+(0.297914580394)*x[2]
        s[1,1,2,0]=(-0.925214063847)*x[0]+(-1.31725976684)*x[1]+(-1.15890525345)*x[2]
        s[1,1,2,1]=(-0.0211857989738)*x[0]+(-0.45544304451)*x[1]+(-1.69006065778)*x[2]
        s[1,2,0,0]=(0.545466934146)*x[0]+(-0.661570837411)*x[1]+(-0.387531002262)*x[2]
        s[1,2,0,1]=(-0.978830810453)*x[0]+(0.958890753286)*x[1]+(-0.3992633515)*x[2]
        s[1,2,1,0]=(0.395262411038)*x[0]+(0.3119070642)*x[1]+(0.544660098992)*x[2]
        s[1,2,1,1]=(-1.07426944365)*x[0]+(0.223216516043)*x[1]+(0.589789766065)*x[2]
        s[1,2,2,0]=(0.0742935523479)*x[0]+(1.20227049003)*x[1]+(0.134941956037)*x[2]
        s[1,2,2,1]=(1.02997244989)*x[0]+(1.24846778966)*x[1]+(-1.76817806687)*x[2]
        s[1,3,0,0]=(-0.0753143898819)*x[0]+(-0.138231791536)*x[1]+(-1.06788473564)*x[2]
        s[1,3,0,1]=(-1.00651208369)*x[0]+(-0.104956186368)*x[1]+(-0.51036966316)*x[2]
        s[1,3,1,0]=(-0.221214050083)*x[0]+(0.227150982313)*x[1]+(1.08579667343)*x[2]
        s[1,3,1,1]=(0.510525992628)*x[0]+(1.49179453941)*x[1]+(-0.0676345340155)*x[2]
        s[1,3,2,0]=(0.917441023004)*x[0]+(0.51231105646)*x[1]+(-0.620649550464)*x[2]
        s[1,3,2,1]=(-1.10288875837)*x[0]+(-0.142710630558)*x[1]+(0.623821771889)*x[2]
        s[1,4,0,0]=(-0.255410535992)*x[0]+(0.530829574503)*x[1]+(0.775272541509)*x[2]
        s[1,4,0,1]=(0.284567984278)*x[0]+(0.125859042954)*x[1]+(0.533462070129)*x[2]
        s[1,4,1,0]=(-0.508598664109)*x[0]+(-1.40877217657)*x[1]+(-0.484415496121)*x[2]
        s[1,4,1,1]=(0.556787007449)*x[0]+(-0.116886679539)*x[1]+(1.00640244397)*x[2]
        s[1,4,2,0]=(0.317366265155)*x[0]+(-1.27232976963)*x[1]+(-0.114651450491)*x[2]
        s[1,4,2,1]=(-1.28370651742)*x[0]+(-0.362861660348)*x[1]+(-0.0962080517851)*x[2]
        s[2,0,0,0]=(0.271907114067)*x[0]+(0.265344846384)*x[1]+(0.0924933201796)*x[2]
        s[2,0,0,1]=(-1.58861288248)*x[0]+(0.22963952324)*x[1]+(-0.653279584852)*x[2]
        s[2,0,1,0]=(0.944381839359)*x[0]+(-0.302082255723)*x[1]+(-1.81040568842)*x[2]
        s[2,0,1,1]=(-0.650306656954)*x[0]+(-0.0845826799882)*x[1]+(-1.25732622868)*x[2]
        s[2,0,2,0]=(-0.0135274380735)*x[0]+(-1.2176048332)*x[1]+(-0.330033040382)*x[2]
        s[2,0,2,1]=(0.8254202497)*x[0]+(0.108602990536)*x[1]+(-0.538593746905)*x[2]
        s[2,1,0,0]=(-0.586282432139)*x[0]+(-0.306182725701)*x[1]+(0.433425406694)*x[2]
        s[2,1,0,1]=(-0.252073403002)*x[0]+(0.106710584551)*x[1]+(-0.125992533885)*x[2]
        s[2,1,1,0]=(0.424274043808)*x[0]+(-0.987960051363)*x[1]+(0.602880255366)*x[2]
        s[2,1,1,1]=(-1.18215444375)*x[0]+(-1.0587178922)*x[1]+(1.07462469891)*x[2]
        s[2,1,2,0]=(-1.21183003522)*x[0]+(0.806502177052)*x[1]+(0.396337647649)*x[2]
        s[2,1,2,1]=(0.47346358421)*x[0]+(1.19060722938)*x[1]+(-0.830743235441)*x[2]
        s[2,2,0,0]=(0.394593530336)*x[0]+(-0.488087761788)*x[1]+(0.0214792800471)*x[2]
        s[2,2,0,1]=(-0.959164075423)*x[0]+(-1.29228916512)*x[1]+(-0.299706293498)*x[2]
        s[2,2,1,0]=(-0.458184497165)*x[0]+(-1.35468397098)*x[1]+(1.08155385172)*x[2]
        s[2,2,1,1]=(-0.543174665411)*x[0]+(-0.546979238048)*x[1]+(-0.00870655686839)*x[2]
        s[2,2,2,0]=(1.04289134262)*x[0]+(-0.795937095748)*x[1]+(0.437742223432)*x[2]
        s[2,2,2,1]=(0.987908115924)*x[0]+(0.717930289783)*x[1]+(-0.371823996424)*x[2]
        s[2,3,0,0]=(-0.169000222452)*x[0]+(-1.17983694078)*x[1]+(-0.525903072845)*x[2]
        s[2,3,0,1]=(-1.18221084142)*x[0]+(-0.723856791479)*x[1]+(-0.0735764242849)*x[2]
        s[2,3,1,0]=(-1.23369418263)*x[0]+(-1.61175404993)*x[1]+(-0.38186249552)*x[2]
        s[2,3,1,1]=(-0.638043359549)*x[0]+(-0.711623038143)*x[1]+(1.37146468285)*x[2]
        s[2,3,2,0]=(0.231534056605)*x[0]+(-0.468413103344)*x[1]+(0.181817447577)*x[2]
        s[2,3,2,1]=(-0.204204815383)*x[0]+(0.426315456914)*x[1]+(0.112813411509)*x[2]
        s[2,4,0,0]=(0.358495614116)*x[0]+(0.530612217508)*x[1]+(-1.18131247911)*x[2]
        s[2,4,0,1]=(-1.47421904004)*x[0]+(-0.135887217467)*x[1]+(1.1526366976)*x[2]
        s[2,4,1,0]=(1.05475775091)*x[0]+(-1.14201019549)*x[1]+(-0.00147054925041)*x[2]
        s[2,4,1,1]=(-0.845002383828)*x[0]+(1.27489434704)*x[1]+(-0.0759015370413)*x[2]
        s[2,4,2,0]=(-1.78781582559)*x[0]+(0.615399643816)*x[1]+(0.943547107618)*x[2]
        s[2,4,2,1]=(-0.269588308338)*x[0]+(0.582501413736)*x[1]+(0.309865222562)*x[2]
        s[3,0,0,0]=(0.0339770002586)*x[0]+(-1.05588139882)*x[1]+(-0.350074839678)*x[2]
        s[3,0,0,1]=(1.26035921199)*x[0]+(-1.44475344932)*x[1]+(0.230467305805)*x[2]
        s[3,0,1,0]=(-0.212637505994)*x[0]+(0.836913336266)*x[1]+(0.782933559258)*x[2]
        s[3,0,1,1]=(0.350804832921)*x[0]+(-0.711735663328)*x[1]+(-0.7055885795)*x[2]
        s[3,0,2,0]=(1.15262669529)*x[0]+(-0.609133167001)*x[1]+(-0.0481345387977)*x[2]
        s[3,0,2,1]=(-0.328624274521)*x[0]+(-0.1435975351)*x[1]+(1.1531857455)*x[2]
        s[3,1,0,0]=(0.0574967918966)*x[0]+(-0.892284893383)*x[1]+(0.939965393644)*x[2]
        s[3,1,0,1]=(0.533477118513)*x[0]+(-0.396195978184)*x[1]+(1.40017632877)*x[2]
        s[3,1,1,0]=(-0.503230088158)*x[0]+(-0.90636136387)*x[1]+(0.0881754805722)*x[2]
        s[3,1,1,1]=(1.98755626341)*x[0]+(0.807046266085)*x[1]+(-0.257226825918)*x[2]
        s[3,1,2,0]=(0.589818915815)*x[0]+(0.978508047098)*x[1]+(0.155263385992)*x[2]
        s[3,1,2,1]=(0.134507669198)*x[0]+(-0.0636645368891)*x[1]+(0.615504770845)*x[2]
        s[3,2,0,0]=(0.930104004822)*x[0]+(-0.382205774724)*x[1]+(0.668127591046)*x[2]
        s[3,2,0,1]=(-0.190422928513)*x[0]+(0.293413650478)*x[1]+(-0.505554260659)*x[2]
        s[3,2,1,0]=(-1.2934566905)*x[0]+(0.76603122212)*x[1]+(1.20584140236)*x[2]
        s[3,2,1,1]=(0.114254707378)*x[0]+(-0.0342824618295)*x[1]+(1.52172882565)*x[2]
        s[3,2,2,0]=(0.248047215938)*x[0]+(-0.978266687903)*x[1]+(-1.30670985993)*x[2]
        s[3,2,2,1]=(0.300650838074)*x[0]+(-0.220707794615)*x[1]+(-0.30715962416)*x[2]
        s[3,3,0,0]=(0.448483391935)*x[0]+(-0.439040526715)*x[1]+(-0.528237104437)*x[2]
        s[3,3,0,1]=(0.0927811849092)*x[0]+(-0.0543684949756)*x[1]+(-0.360230052054)*x[2]
        s[3,3,1,0]=(0.24809679767)*x[0]+(-1.2533905798)*x[1]+(-0.492657878912)*x[2]
        s[3,3,1,1]=(0.191376087122)*x[0]+(-0.897239039489)*x[1]+(1.40598336079)*x[2]
        s[3,3,2,0]=(-1.63824093083)*x[0]+(-1.77338068521)*x[1]+(0.454764854525)*x[2]
        s[3,3,2,1]=(1.24005155499)*x[0]+(0.259110200765)*x[1]+(-0.0880271632004)*x[2]
        s[3,4,0,0]=(-1.54899613736)*x[0]+(0.803173298876)*x[1]+(0.429702060218)*x[2]
        s[3,4,0,1]=(-1.31381922671)*x[0]+(-0.258985809712)*x[1]+(-0.769254247799)*x[2]
        s[3,4,1,0]=(-0.913527471875)*x[0]+(0.534597194574)*x[1]+(-0.937480880823)*x[2]
        s[3,4,1,1]=(0.516591744517)*x[0]+(-0.640572168797)*x[1]+(-1.24601319692)*x[2]
        s[3,4,2,0]=(0.349486782323)*x[0]+(1.25410491943)*x[1]+(0.145174151067)*x[2]
        s[3,4,2,1]=(-0.799350818525)*x[0]+(-0.505153470695)*x[1]+(-1.20903427385)*x[2]
        ref[0,0,0,0]=(-0.0655735713454)*x_ref[0]+(-0.69619555181)*x_ref[1]+(0.881033804166)*x_ref[2]
        ref[0,0,0,1]=(-0.886101209867)*x_ref[0]+(0.0852736523075)*x_ref[1]+(0.514766114136)*x_ref[2]
        ref[0,0,1,0]=(-0.708903115024)*x_ref[0]+(-0.949542800872)*x_ref[1]+(-1.07763550768)*x_ref[2]
        ref[0,0,1,1]=(0.970797729322)*x_ref[0]+(-0.237719575345)*x_ref[1]+(1.54252127667)*x_ref[2]
        ref[0,0,2,0]=(-0.112910964189)*x_ref[0]+(0.813990286281)*x_ref[1]+(0.437139526752)*x_ref[2]
        ref[0,0,2,1]=(0.270076285201)*x_ref[0]+(-1.04231968668)*x_ref[1]+(1.08793884214)*x_ref[2]
        ref[0,1,0,0]=(-0.376091612547)*x_ref[0]+(0.965045498929)*x_ref[1]+(-0.516825182321)*x_ref[2]
        ref[0,1,0,1]=(-0.922174392864)*x_ref[0]+(1.12745699436)*x_ref[1]+(0.647593149683)*x_ref[2]
        ref[0,1,1,0]=(-1.62204275282)*x_ref[0]+(0.64585305798)*x_ref[1]+(-1.41457161128)*x_ref[2]
        ref[0,1,1,1]=(-0.412715098923)*x_ref[0]+(0.00317118495093)*x_ref[1]+(-0.848053699577)*x_ref[2]
        ref[0,1,2,0]=(-1.33788126072)*x_ref[0]+(-0.370246465439)*x_ref[1]+(0.149368781369)*x_ref[2]
        ref[0,1,2,1]=(-0.0432149867078)*x_ref[0]+(0.20428198228)*x_ref[1]+(0.685699546484)*x_ref[2]
        ref[0,2,0,0]=(-0.849779864072)*x_ref[0]+(-0.465408292856)*x_ref[1]+(0.12942183587)*x_ref[2]
        ref[0,2,0,1]=(-0.533958302867)*x_ref[0]+(0.742004817382)*x_ref[1]+(-0.146827358028)*x_ref[2]
        ref[0,2,1,0]=(-0.100072250083)*x_ref[0]+(-1.24645970838)*x_ref[1]+(1.45346380514)*x_ref[2]
        ref[0,2,1,1]=(-0.718115035714)*x_ref[0]+(0.572444837212)*x_ref[1]+(-0.0961841883909)*x_ref[2]
        ref[0,2,2,0]=(-1.6396722492)*x_ref[0]+(-0.657016455181)*x_ref[1]+(-0.52001961538)*x_ref[2]
        ref[0,2,2,1]=(-0.0161596921482)*x_ref[0]+(-0.130395174634)*x_ref[1]+(1.68490335057)*x_ref[2]
        ref[0,3,0,0]=(0.16054789145)*x_ref[0]+(-1.35250823955)*x_ref[1]+(0.663123269205)*x_ref[2]
        ref[0,3,0,1]=(0.307153218553)*x_ref[0]+(0.705376328926)*x_ref[1]+(-0.335478576645)*x_ref[2]
        ref[0,3,1,0]=(0.499629878974)*x_ref[0]+(-0.389863260213)*x_ref[1]+(-0.398638918232)*x_ref[2]
        ref[0,3,1,1]=(0.229760417346)*x_ref[0]+(-0.382704654178)*x_ref[1]+(0.044859414166)*x_ref[2]
        ref[0,3,2,0]=(0.402905751731)*x_ref[0]+(-0.433748870487)*x_ref[1]+(1.05246387058)*x_ref[2]
        ref[0,3,2,1]=(0.620966047942)*x_ref[0]+(0.778166847286)*x_ref[1]+(-0.684881429118)*x_ref[2]
        ref[0,4,0,0]=(-0.0429218580761)*x_ref[0]+(0.263402329307)*x_ref[1]+(0.314852944534)*x_ref[2]
        ref[0,4,0,1]=(0.112989114824)*x_ref[0]+(-0.407308926569)*x_ref[1]+(-0.581889344395)*x_ref[2]
        ref[0,4,1,0]=(1.46986587487)*x_ref[0]+(-0.360451075631)*x_ref[1]+(1.41473323373)*x_ref[2]
        ref[0,4,1,1]=(-0.0898945660646)*x_ref[0]+(-1.24644050789)*x_ref[1]+(-0.344493513462)*x_ref[2]
        ref[0,4,2,0]=(-0.383865368385)*x_ref[0]+(0.758209662119)*x_ref[1]+(0.737714808779)*x_ref[2]
        ref[0,4,2,1]=(-0.282690133746)*x_ref[0]+(0.520513691755)*x_ref[1]+(-0.38353403004)*x_ref[2]
        ref[1,0,0,0]=(-0.233823433606)*x_ref[0]+(0.627115553859)*x_ref[1]+(0.240814969095)*x_ref[2]
        ref[1,0,0,1]=(1.65536823794)*x_ref[0]+(0.577386580179)*x_ref[1]+(0.64299358654)*x_ref[2]
        ref[1,0,1,0]=(-0.304849514801)*x_ref[0]+(-1.26879113223)*x_ref[1]+(-0.257428355176)*x_ref[2]
        ref[1,0,1,1]=(0.313165425271)*x_ref[0]+(0.648645104314)*x_ref[1]+(0.152429094691)*x_ref[2]
        ref[1,0,2,0]=(-1.18245997349)*x_ref[0]+(-0.990110942468)*x_ref[1]+(-0.528372216957)*x_ref[2]
        ref[1,0,2,1]=(-0.497751288211)*x_ref[0]+(1.27633543739)*x_ref[1]+(1.29106845801)*x_ref[2]
        ref[1,1,0,0]=(-0.401807344422)*x_ref[0]+(1.00839022632)*x_ref[1]+(0.855275309873)*x_ref[2]
        ref[1,1,0,1]=(-1.22469413181)*x_ref[0]+(0.946705273705)*x_ref[1]+(1.02166037335)*x_ref[2]
        ref[1,1,1,0]=(0.0800997438684)*x_ref[0]+(0.232002850135)*x_ref[1]+(1.72304848773)*x_ref[2]
        ref[1,1,1,1]=(-1.14517485603)*x_ref[0]+(-1.68921566046)*x_ref[1]+(0.297914580394)*x_ref[2]
        ref[1,1,2,0]=(-0.925214063847)*x_ref[0]+(-1.31725976684)*x_ref[1]+(-1.15890525345)*x_ref[2]
        ref[1,1,2,1]=(-0.0211857989738)*x_ref[0]+(-0.45544304451)*x_ref[1]+(-1.69006065778)*x_ref[2]
        ref[1,2,0,0]=(0.545466934146)*x_ref[0]+(-0.661570837411)*x_ref[1]+(-0.387531002262)*x_ref[2]
        ref[1,2,0,1]=(-0.978830810453)*x_ref[0]+(0.958890753286)*x_ref[1]+(-0.3992633515)*x_ref[2]
        ref[1,2,1,0]=(0.395262411038)*x_ref[0]+(0.3119070642)*x_ref[1]+(0.544660098992)*x_ref[2]
        ref[1,2,1,1]=(-1.07426944365)*x_ref[0]+(0.223216516043)*x_ref[1]+(0.589789766065)*x_ref[2]
        ref[1,2,2,0]=(0.0742935523479)*x_ref[0]+(1.20227049003)*x_ref[1]+(0.134941956037)*x_ref[2]
        ref[1,2,2,1]=(1.02997244989)*x_ref[0]+(1.24846778966)*x_ref[1]+(-1.76817806687)*x_ref[2]
        ref[1,3,0,0]=(-0.0753143898819)*x_ref[0]+(-0.138231791536)*x_ref[1]+(-1.06788473564)*x_ref[2]
        ref[1,3,0,1]=(-1.00651208369)*x_ref[0]+(-0.104956186368)*x_ref[1]+(-0.51036966316)*x_ref[2]
        ref[1,3,1,0]=(-0.221214050083)*x_ref[0]+(0.227150982313)*x_ref[1]+(1.08579667343)*x_ref[2]
        ref[1,3,1,1]=(0.510525992628)*x_ref[0]+(1.49179453941)*x_ref[1]+(-0.0676345340155)*x_ref[2]
        ref[1,3,2,0]=(0.917441023004)*x_ref[0]+(0.51231105646)*x_ref[1]+(-0.620649550464)*x_ref[2]
        ref[1,3,2,1]=(-1.10288875837)*x_ref[0]+(-0.142710630558)*x_ref[1]+(0.623821771889)*x_ref[2]
        ref[1,4,0,0]=(-0.255410535992)*x_ref[0]+(0.530829574503)*x_ref[1]+(0.775272541509)*x_ref[2]
        ref[1,4,0,1]=(0.284567984278)*x_ref[0]+(0.125859042954)*x_ref[1]+(0.533462070129)*x_ref[2]
        ref[1,4,1,0]=(-0.508598664109)*x_ref[0]+(-1.40877217657)*x_ref[1]+(-0.484415496121)*x_ref[2]
        ref[1,4,1,1]=(0.556787007449)*x_ref[0]+(-0.116886679539)*x_ref[1]+(1.00640244397)*x_ref[2]
        ref[1,4,2,0]=(0.317366265155)*x_ref[0]+(-1.27232976963)*x_ref[1]+(-0.114651450491)*x_ref[2]
        ref[1,4,2,1]=(-1.28370651742)*x_ref[0]+(-0.362861660348)*x_ref[1]+(-0.0962080517851)*x_ref[2]
        ref[2,0,0,0]=(0.271907114067)*x_ref[0]+(0.265344846384)*x_ref[1]+(0.0924933201796)*x_ref[2]
        ref[2,0,0,1]=(-1.58861288248)*x_ref[0]+(0.22963952324)*x_ref[1]+(-0.653279584852)*x_ref[2]
        ref[2,0,1,0]=(0.944381839359)*x_ref[0]+(-0.302082255723)*x_ref[1]+(-1.81040568842)*x_ref[2]
        ref[2,0,1,1]=(-0.650306656954)*x_ref[0]+(-0.0845826799882)*x_ref[1]+(-1.25732622868)*x_ref[2]
        ref[2,0,2,0]=(-0.0135274380735)*x_ref[0]+(-1.2176048332)*x_ref[1]+(-0.330033040382)*x_ref[2]
        ref[2,0,2,1]=(0.8254202497)*x_ref[0]+(0.108602990536)*x_ref[1]+(-0.538593746905)*x_ref[2]
        ref[2,1,0,0]=(-0.586282432139)*x_ref[0]+(-0.306182725701)*x_ref[1]+(0.433425406694)*x_ref[2]
        ref[2,1,0,1]=(-0.252073403002)*x_ref[0]+(0.106710584551)*x_ref[1]+(-0.125992533885)*x_ref[2]
        ref[2,1,1,0]=(0.424274043808)*x_ref[0]+(-0.987960051363)*x_ref[1]+(0.602880255366)*x_ref[2]
        ref[2,1,1,1]=(-1.18215444375)*x_ref[0]+(-1.0587178922)*x_ref[1]+(1.07462469891)*x_ref[2]
        ref[2,1,2,0]=(-1.21183003522)*x_ref[0]+(0.806502177052)*x_ref[1]+(0.396337647649)*x_ref[2]
        ref[2,1,2,1]=(0.47346358421)*x_ref[0]+(1.19060722938)*x_ref[1]+(-0.830743235441)*x_ref[2]
        ref[2,2,0,0]=(0.394593530336)*x_ref[0]+(-0.488087761788)*x_ref[1]+(0.0214792800471)*x_ref[2]
        ref[2,2,0,1]=(-0.959164075423)*x_ref[0]+(-1.29228916512)*x_ref[1]+(-0.299706293498)*x_ref[2]
        ref[2,2,1,0]=(-0.458184497165)*x_ref[0]+(-1.35468397098)*x_ref[1]+(1.08155385172)*x_ref[2]
        ref[2,2,1,1]=(-0.543174665411)*x_ref[0]+(-0.546979238048)*x_ref[1]+(-0.00870655686839)*x_ref[2]
        ref[2,2,2,0]=(1.04289134262)*x_ref[0]+(-0.795937095748)*x_ref[1]+(0.437742223432)*x_ref[2]
        ref[2,2,2,1]=(0.987908115924)*x_ref[0]+(0.717930289783)*x_ref[1]+(-0.371823996424)*x_ref[2]
        ref[2,3,0,0]=(-0.169000222452)*x_ref[0]+(-1.17983694078)*x_ref[1]+(-0.525903072845)*x_ref[2]
        ref[2,3,0,1]=(-1.18221084142)*x_ref[0]+(-0.723856791479)*x_ref[1]+(-0.0735764242849)*x_ref[2]
        ref[2,3,1,0]=(-1.23369418263)*x_ref[0]+(-1.61175404993)*x_ref[1]+(-0.38186249552)*x_ref[2]
        ref[2,3,1,1]=(-0.638043359549)*x_ref[0]+(-0.711623038143)*x_ref[1]+(1.37146468285)*x_ref[2]
        ref[2,3,2,0]=(0.231534056605)*x_ref[0]+(-0.468413103344)*x_ref[1]+(0.181817447577)*x_ref[2]
        ref[2,3,2,1]=(-0.204204815383)*x_ref[0]+(0.426315456914)*x_ref[1]+(0.112813411509)*x_ref[2]
        ref[2,4,0,0]=(0.358495614116)*x_ref[0]+(0.530612217508)*x_ref[1]+(-1.18131247911)*x_ref[2]
        ref[2,4,0,1]=(-1.47421904004)*x_ref[0]+(-0.135887217467)*x_ref[1]+(1.1526366976)*x_ref[2]
        ref[2,4,1,0]=(1.05475775091)*x_ref[0]+(-1.14201019549)*x_ref[1]+(-0.00147054925041)*x_ref[2]
        ref[2,4,1,1]=(-0.845002383828)*x_ref[0]+(1.27489434704)*x_ref[1]+(-0.0759015370413)*x_ref[2]
        ref[2,4,2,0]=(-1.78781582559)*x_ref[0]+(0.615399643816)*x_ref[1]+(0.943547107618)*x_ref[2]
        ref[2,4,2,1]=(-0.269588308338)*x_ref[0]+(0.582501413736)*x_ref[1]+(0.309865222562)*x_ref[2]
        ref[3,0,0,0]=(0.0339770002586)*x_ref[0]+(-1.05588139882)*x_ref[1]+(-0.350074839678)*x_ref[2]
        ref[3,0,0,1]=(1.26035921199)*x_ref[0]+(-1.44475344932)*x_ref[1]+(0.230467305805)*x_ref[2]
        ref[3,0,1,0]=(-0.212637505994)*x_ref[0]+(0.836913336266)*x_ref[1]+(0.782933559258)*x_ref[2]
        ref[3,0,1,1]=(0.350804832921)*x_ref[0]+(-0.711735663328)*x_ref[1]+(-0.7055885795)*x_ref[2]
        ref[3,0,2,0]=(1.15262669529)*x_ref[0]+(-0.609133167001)*x_ref[1]+(-0.0481345387977)*x_ref[2]
        ref[3,0,2,1]=(-0.328624274521)*x_ref[0]+(-0.1435975351)*x_ref[1]+(1.1531857455)*x_ref[2]
        ref[3,1,0,0]=(0.0574967918966)*x_ref[0]+(-0.892284893383)*x_ref[1]+(0.939965393644)*x_ref[2]
        ref[3,1,0,1]=(0.533477118513)*x_ref[0]+(-0.396195978184)*x_ref[1]+(1.40017632877)*x_ref[2]
        ref[3,1,1,0]=(-0.503230088158)*x_ref[0]+(-0.90636136387)*x_ref[1]+(0.0881754805722)*x_ref[2]
        ref[3,1,1,1]=(1.98755626341)*x_ref[0]+(0.807046266085)*x_ref[1]+(-0.257226825918)*x_ref[2]
        ref[3,1,2,0]=(0.589818915815)*x_ref[0]+(0.978508047098)*x_ref[1]+(0.155263385992)*x_ref[2]
        ref[3,1,2,1]=(0.134507669198)*x_ref[0]+(-0.0636645368891)*x_ref[1]+(0.615504770845)*x_ref[2]
        ref[3,2,0,0]=(0.930104004822)*x_ref[0]+(-0.382205774724)*x_ref[1]+(0.668127591046)*x_ref[2]
        ref[3,2,0,1]=(-0.190422928513)*x_ref[0]+(0.293413650478)*x_ref[1]+(-0.505554260659)*x_ref[2]
        ref[3,2,1,0]=(-1.2934566905)*x_ref[0]+(0.76603122212)*x_ref[1]+(1.20584140236)*x_ref[2]
        ref[3,2,1,1]=(0.114254707378)*x_ref[0]+(-0.0342824618295)*x_ref[1]+(1.52172882565)*x_ref[2]
        ref[3,2,2,0]=(0.248047215938)*x_ref[0]+(-0.978266687903)*x_ref[1]+(-1.30670985993)*x_ref[2]
        ref[3,2,2,1]=(0.300650838074)*x_ref[0]+(-0.220707794615)*x_ref[1]+(-0.30715962416)*x_ref[2]
        ref[3,3,0,0]=(0.448483391935)*x_ref[0]+(-0.439040526715)*x_ref[1]+(-0.528237104437)*x_ref[2]
        ref[3,3,0,1]=(0.0927811849092)*x_ref[0]+(-0.0543684949756)*x_ref[1]+(-0.360230052054)*x_ref[2]
        ref[3,3,1,0]=(0.24809679767)*x_ref[0]+(-1.2533905798)*x_ref[1]+(-0.492657878912)*x_ref[2]
        ref[3,3,1,1]=(0.191376087122)*x_ref[0]+(-0.897239039489)*x_ref[1]+(1.40598336079)*x_ref[2]
        ref[3,3,2,0]=(-1.63824093083)*x_ref[0]+(-1.77338068521)*x_ref[1]+(0.454764854525)*x_ref[2]
        ref[3,3,2,1]=(1.24005155499)*x_ref[0]+(0.259110200765)*x_ref[1]+(-0.0880271632004)*x_ref[2]
        ref[3,4,0,0]=(-1.54899613736)*x_ref[0]+(0.803173298876)*x_ref[1]+(0.429702060218)*x_ref[2]
        ref[3,4,0,1]=(-1.31381922671)*x_ref[0]+(-0.258985809712)*x_ref[1]+(-0.769254247799)*x_ref[2]
        ref[3,4,1,0]=(-0.913527471875)*x_ref[0]+(0.534597194574)*x_ref[1]+(-0.937480880823)*x_ref[2]
        ref[3,4,1,1]=(0.516591744517)*x_ref[0]+(-0.640572168797)*x_ref[1]+(-1.24601319692)*x_ref[2]
        ref[3,4,2,0]=(0.349486782323)*x_ref[0]+(1.25410491943)*x_ref[1]+(0.145174151067)*x_ref[2]
        ref[3,4,2,1]=(-0.799350818525)*x_ref[0]+(-0.505153470695)*x_ref[1]+(-1.20903427385)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-1.05153531037)*x[0]+(1.5374190245)*x[1]
        ref=(-1.05153531037)*x_ref[0]+(1.5374190245)*x_ref[1]
      else:
        s=(-1.91883881043)*x[0]+(-0.414033171764)*x[1]+(1.96173093572)*x[2]
        ref=(-1.91883881043)*x_ref[0]+(-0.414033171764)*x_ref[1]+(1.96173093572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        s[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
        ref[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        ref[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
      else:
        s[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        s[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
        ref[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        ref[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.581359229868)*x[0]+(-0.405627827004)*x[1]
        s[0,1]=(1.15883572458)*x[0]+(-0.212383708658)*x[1]
        s[0,2]=(-0.675196965148)*x[0]+(-0.0589289579115)*x[1]
        s[0,3]=(-0.466142003406)*x[0]+(1.143298366)*x[1]
        s[0,4]=(0.706585901476)*x[0]+(-0.596587020578)*x[1]
        s[1,0]=(0.150862770285)*x[0]+(1.3486848087)*x[1]
        s[1,1]=(0.694050128425)*x[0]+(0.919192282198)*x[1]
        s[1,2]=(0.123568652464)*x[0]+(-1.1901501926)*x[1]
        s[1,3]=(-0.0147792668086)*x[0]+(-1.49381121537)*x[1]
        s[1,4]=(0.0737831649537)*x[0]+(-0.752053730242)*x[1]
        s[2,0]=(-0.758812518739)*x[0]+(-1.26143124516)*x[1]
        s[2,1]=(0.507459828891)*x[0]+(-0.0315665691912)*x[1]
        s[2,2]=(-0.492645010187)*x[0]+(-0.649596364388)*x[1]
        s[2,3]=(-0.0632358844144)*x[0]+(0.899340082527)*x[1]
        s[2,4]=(-0.763180407153)*x[0]+(-0.269808126268)*x[1]
        s[3,0]=(-1.5724790938)*x[0]+(1.16942939992)*x[1]
        s[3,1]=(0.759317295371)*x[0]+(0.14287823082)*x[1]
        s[3,2]=(0.70712973662)*x[0]+(1.08455803224)*x[1]
        s[3,3]=(-0.736730602555)*x[0]+(-0.183785007602)*x[1]
        s[3,4]=(-0.950538745165)*x[0]+(0.294140958329)*x[1]
        ref[0,0]=(-0.581359229868)*x_ref[0]+(-0.405627827004)*x_ref[1]
        ref[0,1]=(1.15883572458)*x_ref[0]+(-0.212383708658)*x_ref[1]
        ref[0,2]=(-0.675196965148)*x_ref[0]+(-0.0589289579115)*x_ref[1]
        ref[0,3]=(-0.466142003406)*x_ref[0]+(1.143298366)*x_ref[1]
        ref[0,4]=(0.706585901476)*x_ref[0]+(-0.596587020578)*x_ref[1]
        ref[1,0]=(0.150862770285)*x_ref[0]+(1.3486848087)*x_ref[1]
        ref[1,1]=(0.694050128425)*x_ref[0]+(0.919192282198)*x_ref[1]
        ref[1,2]=(0.123568652464)*x_ref[0]+(-1.1901501926)*x_ref[1]
        ref[1,3]=(-0.0147792668086)*x_ref[0]+(-1.49381121537)*x_ref[1]
        ref[1,4]=(0.0737831649537)*x_ref[0]+(-0.752053730242)*x_ref[1]
        ref[2,0]=(-0.758812518739)*x_ref[0]+(-1.26143124516)*x_ref[1]
        ref[2,1]=(0.507459828891)*x_ref[0]+(-0.0315665691912)*x_ref[1]
        ref[2,2]=(-0.492645010187)*x_ref[0]+(-0.649596364388)*x_ref[1]
        ref[2,3]=(-0.0632358844144)*x_ref[0]+(0.899340082527)*x_ref[1]
        ref[2,4]=(-0.763180407153)*x_ref[0]+(-0.269808126268)*x_ref[1]
        ref[3,0]=(-1.5724790938)*x_ref[0]+(1.16942939992)*x_ref[1]
        ref[3,1]=(0.759317295371)*x_ref[0]+(0.14287823082)*x_ref[1]
        ref[3,2]=(0.70712973662)*x_ref[0]+(1.08455803224)*x_ref[1]
        ref[3,3]=(-0.736730602555)*x_ref[0]+(-0.183785007602)*x_ref[1]
        ref[3,4]=(-0.950538745165)*x_ref[0]+(0.294140958329)*x_ref[1]
      else:
        s[0,0]=(0.310668873066)*x[0]+(1.66326709742)*x[1]+(1.10140064967)*x[2]
        s[0,1]=(0.46962350981)*x[0]+(-1.02292998305)*x[1]+(0.919666772954)*x[2]
        s[0,2]=(-1.18114779319)*x[0]+(1.23819669136)*x[1]+(0.0622519749551)*x[2]
        s[0,3]=(1.08356504043)*x[0]+(0.450517408398)*x[1]+(-0.528809988127)*x[2]
        s[0,4]=(-0.179349520643)*x[0]+(0.192827326205)*x[1]+(-0.543850326813)*x[2]
        s[1,0]=(-1.60006459477)*x[0]+(0.800497908511)*x[1]+(0.745234766014)*x[2]
        s[1,1]=(0.667355662535)*x[0]+(0.800303765538)*x[1]+(-1.17397311427)*x[2]
        s[1,2]=(0.19342327722)*x[0]+(-0.149618364398)*x[1]+(-1.06968763194)*x[2]
        s[1,3]=(0.296694301988)*x[0]+(-1.35691147932)*x[1]+(0.748623252633)*x[2]
        s[1,4]=(0.724332288795)*x[0]+(0.349026497158)*x[1]+(0.377755550832)*x[2]
        s[2,0]=(0.560218327038)*x[0]+(0.184262569787)*x[1]+(-0.821718588725)*x[2]
        s[2,1]=(0.421969037007)*x[0]+(-0.00510504728331)*x[1]+(1.43225098227)*x[2]
        s[2,2]=(0.889590883145)*x[0]+(-0.408365128114)*x[1]+(-1.80100224461)*x[2]
        s[2,3]=(-1.91735498201)*x[0]+(0.590584386339)*x[1]+(-0.905610066725)*x[2]
        s[2,4]=(-0.367106108866)*x[0]+(-0.92493020573)*x[1]+(-0.0746199487091)*x[2]
        s[3,0]=(0.919247371049)*x[0]+(0.326674777975)*x[1]+(-1.47385875396)*x[2]
        s[3,1]=(-0.850835329183)*x[0]+(-0.536914325383)*x[1]+(1.02264919467)*x[2]
        s[3,2]=(-1.41905423752)*x[0]+(-0.632415273773)*x[1]+(0.842187325237)*x[2]
        s[3,3]=(-0.396129693721)*x[0]+(-0.110878220479)*x[1]+(0.49355539737)*x[2]
        s[3,4]=(-0.0781905590856)*x[0]+(-0.0348772309292)*x[1]+(0.184581658536)*x[2]
        ref[0,0]=(0.310668873066)*x_ref[0]+(1.66326709742)*x_ref[1]+(1.10140064967)*x_ref[2]
        ref[0,1]=(0.46962350981)*x_ref[0]+(-1.02292998305)*x_ref[1]+(0.919666772954)*x_ref[2]
        ref[0,2]=(-1.18114779319)*x_ref[0]+(1.23819669136)*x_ref[1]+(0.0622519749551)*x_ref[2]
        ref[0,3]=(1.08356504043)*x_ref[0]+(0.450517408398)*x_ref[1]+(-0.528809988127)*x_ref[2]
        ref[0,4]=(-0.179349520643)*x_ref[0]+(0.192827326205)*x_ref[1]+(-0.543850326813)*x_ref[2]
        ref[1,0]=(-1.60006459477)*x_ref[0]+(0.800497908511)*x_ref[1]+(0.745234766014)*x_ref[2]
        ref[1,1]=(0.667355662535)*x_ref[0]+(0.800303765538)*x_ref[1]+(-1.17397311427)*x_ref[2]
        ref[1,2]=(0.19342327722)*x_ref[0]+(-0.149618364398)*x_ref[1]+(-1.06968763194)*x_ref[2]
        ref[1,3]=(0.296694301988)*x_ref[0]+(-1.35691147932)*x_ref[1]+(0.748623252633)*x_ref[2]
        ref[1,4]=(0.724332288795)*x_ref[0]+(0.349026497158)*x_ref[1]+(0.377755550832)*x_ref[2]
        ref[2,0]=(0.560218327038)*x_ref[0]+(0.184262569787)*x_ref[1]+(-0.821718588725)*x_ref[2]
        ref[2,1]=(0.421969037007)*x_ref[0]+(-0.00510504728331)*x_ref[1]+(1.43225098227)*x_ref[2]
        ref[2,2]=(0.889590883145)*x_ref[0]+(-0.408365128114)*x_ref[1]+(-1.80100224461)*x_ref[2]
        ref[2,3]=(-1.91735498201)*x_ref[0]+(0.590584386339)*x_ref[1]+(-0.905610066725)*x_ref[2]
        ref[2,4]=(-0.367106108866)*x_ref[0]+(-0.92493020573)*x_ref[1]+(-0.0746199487091)*x_ref[2]
        ref[3,0]=(0.919247371049)*x_ref[0]+(0.326674777975)*x_ref[1]+(-1.47385875396)*x_ref[2]
        ref[3,1]=(-0.850835329183)*x_ref[0]+(-0.536914325383)*x_ref[1]+(1.02264919467)*x_ref[2]
        ref[3,2]=(-1.41905423752)*x_ref[0]+(-0.632415273773)*x_ref[1]+(0.842187325237)*x_ref[2]
        ref[3,3]=(-0.396129693721)*x_ref[0]+(-0.110878220479)*x_ref[1]+(0.49355539737)*x_ref[2]
        ref[3,4]=(-0.0781905590856)*x_ref[0]+(-0.0348772309292)*x_ref[1]+(0.184581658536)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.540497015605)*x[0]+(0.282753164584)*x[1]
        s[0,0,1]=(-0.127429073679)*x[0]+(-0.709993735868)*x[1]
        s[0,1,0]=(1.02019673589)*x[0]+(-1.53422031862)*x[1]
        s[0,1,1]=(0.926292918083)*x[0]+(-0.0351188336657)*x[1]
        s[1,0,0]=(-0.736080207248)*x[0]+(1.25260396336)*x[1]
        s[1,0,1]=(0.152653303192)*x[0]+(0.22547444598)*x[1]
        s[1,1,0]=(0.729526332632)*x[0]+(0.919207155406)*x[1]
        s[1,1,1]=(-0.346325378686)*x[0]+(0.141030274609)*x[1]
        s[2,0,0]=(-0.753354302966)*x[0]+(-0.766294422747)*x[1]
        s[2,0,1]=(1.17251569201)*x[0]+(-0.51673325599)*x[1]
        s[2,1,0]=(0.66192662328)*x[0]+(-0.572345218315)*x[1]
        s[2,1,1]=(1.74689126451)*x[0]+(-0.0537684141408)*x[1]
        s[3,0,0]=(0.300593672769)*x[0]+(0.0833505471404)*x[1]
        s[3,0,1]=(0.770671590417)*x[0]+(-0.667020549164)*x[1]
        s[3,1,0]=(0.281444608616)*x[0]+(0.701762037742)*x[1]
        s[3,1,1]=(0.404631442844)*x[0]+(0.337744807226)*x[1]
        s[4,0,0]=(-1.43357362276)*x[0]+(0.687631941644)*x[1]
        s[4,0,1]=(0.694571843151)*x[0]+(0.979495014797)*x[1]
        s[4,1,0]=(-0.0880858702517)*x[0]+(0.729298516002)*x[1]
        s[4,1,1]=(0.641170415423)*x[0]+(-0.407278118417)*x[1]
        s[5,0,0]=(-0.60364366976)*x[0]+(-0.663798735078)*x[1]
        s[5,0,1]=(-0.841035861641)*x[0]+(-1.09892876031)*x[1]
        s[5,1,0]=(0.0838180897314)*x[0]+(0.15765944496)*x[1]
        s[5,1,1]=(-0.847291245273)*x[0]+(1.01145652383)*x[1]
        ref[0,0,0]=(0.540497015605)*x_ref[0]+(0.282753164584)*x_ref[1]
        ref[0,0,1]=(-0.127429073679)*x_ref[0]+(-0.709993735868)*x_ref[1]
        ref[0,1,0]=(1.02019673589)*x_ref[0]+(-1.53422031862)*x_ref[1]
        ref[0,1,1]=(0.926292918083)*x_ref[0]+(-0.0351188336657)*x_ref[1]
        ref[1,0,0]=(-0.736080207248)*x_ref[0]+(1.25260396336)*x_ref[1]
        ref[1,0,1]=(0.152653303192)*x_ref[0]+(0.22547444598)*x_ref[1]
        ref[1,1,0]=(0.729526332632)*x_ref[0]+(0.919207155406)*x_ref[1]
        ref[1,1,1]=(-0.346325378686)*x_ref[0]+(0.141030274609)*x_ref[1]
        ref[2,0,0]=(-0.753354302966)*x_ref[0]+(-0.766294422747)*x_ref[1]
        ref[2,0,1]=(1.17251569201)*x_ref[0]+(-0.51673325599)*x_ref[1]
        ref[2,1,0]=(0.66192662328)*x_ref[0]+(-0.572345218315)*x_ref[1]
        ref[2,1,1]=(1.74689126451)*x_ref[0]+(-0.0537684141408)*x_ref[1]
        ref[3,0,0]=(0.300593672769)*x_ref[0]+(0.0833505471404)*x_ref[1]
        ref[3,0,1]=(0.770671590417)*x_ref[0]+(-0.667020549164)*x_ref[1]
        ref[3,1,0]=(0.281444608616)*x_ref[0]+(0.701762037742)*x_ref[1]
        ref[3,1,1]=(0.404631442844)*x_ref[0]+(0.337744807226)*x_ref[1]
        ref[4,0,0]=(-1.43357362276)*x_ref[0]+(0.687631941644)*x_ref[1]
        ref[4,0,1]=(0.694571843151)*x_ref[0]+(0.979495014797)*x_ref[1]
        ref[4,1,0]=(-0.0880858702517)*x_ref[0]+(0.729298516002)*x_ref[1]
        ref[4,1,1]=(0.641170415423)*x_ref[0]+(-0.407278118417)*x_ref[1]
        ref[5,0,0]=(-0.60364366976)*x_ref[0]+(-0.663798735078)*x_ref[1]
        ref[5,0,1]=(-0.841035861641)*x_ref[0]+(-1.09892876031)*x_ref[1]
        ref[5,1,0]=(0.0838180897314)*x_ref[0]+(0.15765944496)*x_ref[1]
        ref[5,1,1]=(-0.847291245273)*x_ref[0]+(1.01145652383)*x_ref[1]
      else:
        s[0,0,0]=(-0.29326741782)*x[0]+(0.716468869525)*x[1]+(-0.194120207009)*x[2]
        s[0,0,1]=(-0.0687304327556)*x[0]+(0.181789058707)*x[1]+(-0.940504268245)*x[2]
        s[0,1,0]=(0.437145055905)*x[0]+(1.72039421904)*x[1]+(-0.0326204183567)*x[2]
        s[0,1,1]=(1.25919532639)*x[0]+(0.082561920762)*x[1]+(1.01503353491)*x[2]
        s[1,0,0]=(-0.552129852247)*x[0]+(-0.661963725391)*x[1]+(0.191012614278)*x[2]
        s[1,0,1]=(-0.322228468989)*x[0]+(-0.79555280135)*x[1]+(0.0689420056626)*x[2]
        s[1,1,0]=(0.294413007041)*x[0]+(-0.453661316564)*x[1]+(1.74072519115)*x[2]
        s[1,1,1]=(-0.13491214165)*x[0]+(-1.18907003959)*x[1]+(-1.59325361447)*x[2]
        s[2,0,0]=(-0.522689030297)*x[0]+(-0.784658318696)*x[1]+(0.618343699514)*x[2]
        s[2,0,1]=(-0.227214629419)*x[0]+(0.770163719023)*x[1]+(-0.272804476799)*x[2]
        s[2,1,0]=(-1.28131384809)*x[0]+(0.0591076357001)*x[1]+(-0.206570950984)*x[2]
        s[2,1,1]=(0.00241150481435)*x[0]+(0.0678794618244)*x[1]+(-0.272677008307)*x[2]
        s[3,0,0]=(-0.537194244467)*x[0]+(-0.886303700791)*x[1]+(0.704423993884)*x[2]
        s[3,0,1]=(0.0924777549448)*x[0]+(-0.783168034362)*x[1]+(-0.450930098538)*x[2]
        s[3,1,0]=(0.611562862859)*x[0]+(-1.04224402701)*x[1]+(-0.775594008735)*x[2]
        s[3,1,1]=(1.44402445157)*x[0]+(-0.736290409307)*x[1]+(0.51690502447)*x[2]
        s[4,0,0]=(-0.470210375338)*x[0]+(-0.433261543227)*x[1]+(-0.582080162476)*x[2]
        s[4,0,1]=(0.313323330562)*x[0]+(-1.47502042219)*x[1]+(-1.50202564103)*x[2]
        s[4,1,0]=(-0.807949755771)*x[0]+(-0.162904810555)*x[1]+(0.748809317029)*x[2]
        s[4,1,1]=(-1.24228935643)*x[0]+(0.193414583713)*x[1]+(0.85984063481)*x[2]
        s[5,0,0]=(1.08068739991)*x[0]+(-1.30655986333)*x[1]+(0.289800143967)*x[2]
        s[5,0,1]=(1.12114019122)*x[0]+(0.280561086566)*x[1]+(0.844991094994)*x[2]
        s[5,1,0]=(-0.133922711479)*x[0]+(-1.39747129717)*x[1]+(-1.11382363437)*x[2]
        s[5,1,1]=(-1.34164423823)*x[0]+(-0.335643322772)*x[1]+(0.757067990421)*x[2]
        ref[0,0,0]=(-0.29326741782)*x_ref[0]+(0.716468869525)*x_ref[1]+(-0.194120207009)*x_ref[2]
        ref[0,0,1]=(-0.0687304327556)*x_ref[0]+(0.181789058707)*x_ref[1]+(-0.940504268245)*x_ref[2]
        ref[0,1,0]=(0.437145055905)*x_ref[0]+(1.72039421904)*x_ref[1]+(-0.0326204183567)*x_ref[2]
        ref[0,1,1]=(1.25919532639)*x_ref[0]+(0.082561920762)*x_ref[1]+(1.01503353491)*x_ref[2]
        ref[1,0,0]=(-0.552129852247)*x_ref[0]+(-0.661963725391)*x_ref[1]+(0.191012614278)*x_ref[2]
        ref[1,0,1]=(-0.322228468989)*x_ref[0]+(-0.79555280135)*x_ref[1]+(0.0689420056626)*x_ref[2]
        ref[1,1,0]=(0.294413007041)*x_ref[0]+(-0.453661316564)*x_ref[1]+(1.74072519115)*x_ref[2]
        ref[1,1,1]=(-0.13491214165)*x_ref[0]+(-1.18907003959)*x_ref[1]+(-1.59325361447)*x_ref[2]
        ref[2,0,0]=(-0.522689030297)*x_ref[0]+(-0.784658318696)*x_ref[1]+(0.618343699514)*x_ref[2]
        ref[2,0,1]=(-0.227214629419)*x_ref[0]+(0.770163719023)*x_ref[1]+(-0.272804476799)*x_ref[2]
        ref[2,1,0]=(-1.28131384809)*x_ref[0]+(0.0591076357001)*x_ref[1]+(-0.206570950984)*x_ref[2]
        ref[2,1,1]=(0.00241150481435)*x_ref[0]+(0.0678794618244)*x_ref[1]+(-0.272677008307)*x_ref[2]
        ref[3,0,0]=(-0.537194244467)*x_ref[0]+(-0.886303700791)*x_ref[1]+(0.704423993884)*x_ref[2]
        ref[3,0,1]=(0.0924777549448)*x_ref[0]+(-0.783168034362)*x_ref[1]+(-0.450930098538)*x_ref[2]
        ref[3,1,0]=(0.611562862859)*x_ref[0]+(-1.04224402701)*x_ref[1]+(-0.775594008735)*x_ref[2]
        ref[3,1,1]=(1.44402445157)*x_ref[0]+(-0.736290409307)*x_ref[1]+(0.51690502447)*x_ref[2]
        ref[4,0,0]=(-0.470210375338)*x_ref[0]+(-0.433261543227)*x_ref[1]+(-0.582080162476)*x_ref[2]
        ref[4,0,1]=(0.313323330562)*x_ref[0]+(-1.47502042219)*x_ref[1]+(-1.50202564103)*x_ref[2]
        ref[4,1,0]=(-0.807949755771)*x_ref[0]+(-0.162904810555)*x_ref[1]+(0.748809317029)*x_ref[2]
        ref[4,1,1]=(-1.24228935643)*x_ref[0]+(0.193414583713)*x_ref[1]+(0.85984063481)*x_ref[2]
        ref[5,0,0]=(1.08068739991)*x_ref[0]+(-1.30655986333)*x_ref[1]+(0.289800143967)*x_ref[2]
        ref[5,0,1]=(1.12114019122)*x_ref[0]+(0.280561086566)*x_ref[1]+(0.844991094994)*x_ref[2]
        ref[5,1,0]=(-0.133922711479)*x_ref[0]+(-1.39747129717)*x_ref[1]+(-1.11382363437)*x_ref[2]
        ref[5,1,1]=(-1.34164423823)*x_ref[0]+(-0.335643322772)*x_ref[1]+(0.757067990421)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.9809347951)*x[0]+(-1.05666590015)*x[1]
        s[0,0,0,1]=(1.15925248094)*x[0]+(1.22070923582)*x[1]
        s[0,0,1,0]=(-0.0724361628927)*x[0]+(-0.439807238916)*x[1]
        s[0,0,1,1]=(-0.608816713414)*x[0]+(1.78792604745)*x[1]
        s[0,0,2,0]=(-0.78903506363)*x[0]+(-0.75896072902)*x[1]
        s[0,0,2,1]=(0.516424260522)*x[0]+(1.16788015378)*x[1]
        s[0,1,0,0]=(0.383586654748)*x[0]+(1.4179924146)*x[1]
        s[0,1,0,1]=(0.97236907624)*x[0]+(-0.438982550276)*x[1]
        s[0,1,1,0]=(-0.645083297783)*x[0]+(-0.740808997218)*x[1]
        s[0,1,1,1]=(0.330652781935)*x[0]+(0.519132367114)*x[1]
        s[0,1,2,0]=(1.25159174261)*x[0]+(1.26021904904)*x[1]
        s[0,1,2,1]=(-0.239621293462)*x[0]+(0.0244107517523)*x[1]
        s[0,2,0,0]=(-0.286208240697)*x[0]+(-0.314157258671)*x[1]
        s[0,2,0,1]=(-0.0856319829397)*x[0]+(0.278603809503)*x[1]
        s[0,2,1,0]=(0.107199217139)*x[0]+(0.0948664401127)*x[1]
        s[0,2,1,1]=(-0.24620060163)*x[0]+(-1.42942082558)*x[1]
        s[0,2,2,0]=(0.845285367352)*x[0]+(0.408676159609)*x[1]
        s[0,2,2,1]=(-1.49467460394)*x[0]+(0.270722457265)*x[1]
        s[0,3,0,0]=(0.469802598061)*x[0]+(-0.801790345348)*x[1]
        s[0,3,0,1]=(1.71249257903)*x[0]+(-0.776797378881)*x[1]
        s[0,3,1,0]=(0.0540049036319)*x[0]+(1.19031969164)*x[1]
        s[0,3,1,1]=(0.161929130968)*x[0]+(-0.525381566632)*x[1]
        s[0,3,2,0]=(-0.646041965665)*x[0]+(-0.123993555032)*x[1]
        s[0,3,2,1]=(0.0966044205198)*x[0]+(1.24780552685)*x[1]
        s[0,4,0,0]=(-0.289531041539)*x[0]+(1.21117470928)*x[1]
        s[0,4,0,1]=(-0.112215925799)*x[0]+(0.598717725332)*x[1]
        s[0,4,1,0]=(0.162936056701)*x[0]+(-1.02395851472)*x[1]
        s[0,4,1,1]=(0.694066364547)*x[0]+(1.27205020727)*x[1]
        s[0,4,2,0]=(1.17253981195)*x[0]+(0.130572791304)*x[1]
        s[0,4,2,1]=(-0.555241229093)*x[0]+(-0.881115061403)*x[1]
        s[1,0,0,0]=(0.574137047465)*x[0]+(-1.10459006911)*x[1]
        s[1,0,0,1]=(0.441431076852)*x[0]+(-0.133156832361)*x[1]
        s[1,0,1,0]=(0.217136655494)*x[0]+(0.400834952015)*x[1]
        s[1,0,1,1]=(0.516360280675)*x[0]+(0.532907430046)*x[1]
        s[1,0,2,0]=(-0.340764752946)*x[0]+(-1.26671004418)*x[1]
        s[1,0,2,1]=(-1.23231139783)*x[0]+(0.61934930883)*x[1]
        s[1,1,0,0]=(1.76821949014)*x[0]+(1.67542625016)*x[1]
        s[1,1,0,1]=(0.901675970272)*x[0]+(-0.0567109697208)*x[1]
        s[1,1,1,0]=(1.21143551644)*x[0]+(0.375877605521)*x[1]
        s[1,1,1,1]=(-0.699830133987)*x[0]+(0.174660968691)*x[1]
        s[1,1,2,0]=(1.07839763024)*x[0]+(-0.063554858486)*x[1]
        s[1,1,2,1]=(0.131335396282)*x[0]+(0.109893568063)*x[1]
        s[1,2,0,0]=(-0.562355685272)*x[0]+(0.495744936869)*x[1]
        s[1,2,0,1]=(0.0747584188968)*x[0]+(1.94924356864)*x[1]
        s[1,2,1,0]=(-0.103515803094)*x[0]+(0.545380886507)*x[1]
        s[1,2,1,1]=(-0.708926705928)*x[0]+(-0.0610123075525)*x[1]
        s[1,2,2,0]=(1.06802154614)*x[0]+(0.0622475519235)*x[1]
        s[1,2,2,1]=(0.103937010427)*x[0]+(-0.70201008693)*x[1]
        s[1,3,0,0]=(0.0730726824626)*x[0]+(0.39791188789)*x[1]
        s[1,3,0,1]=(0.455273194688)*x[0]+(1.65845657261)*x[1]
        s[1,3,1,0]=(0.0807767752424)*x[0]+(-0.765026395125)*x[1]
        s[1,3,1,1]=(0.0404890748976)*x[0]+(-0.383198358048)*x[1]
        s[1,3,2,0]=(1.26048751933)*x[0]+(0.00329382174075)*x[1]
        s[1,3,2,1]=(-1.21518062161)*x[0]+(-0.325769725861)*x[1]
        s[1,4,0,0]=(0.682181651323)*x[0]+(-0.375432356672)*x[1]
        s[1,4,0,1]=(-1.09892909715)*x[0]+(0.207984305416)*x[1]
        s[1,4,1,0]=(-1.2375595164)*x[0]+(0.215243019189)*x[1]
        s[1,4,1,1]=(-0.595125980205)*x[0]+(0.29104395616)*x[1]
        s[1,4,2,0]=(0.127947393316)*x[0]+(1.43408039596)*x[1]
        s[1,4,2,1]=(0.468574871016)*x[0]+(0.333319611281)*x[1]
        s[2,0,0,0]=(0.369954209324)*x[0]+(-1.09614147452)*x[1]
        s[2,0,0,1]=(-0.120003056441)*x[0]+(-0.172964298439)*x[1]
        s[2,0,1,0]=(1.38230317911)*x[0]+(1.25730140777)*x[1]
        s[2,0,1,1]=(1.10857329622)*x[0]+(0.449561992049)*x[1]
        s[2,0,2,0]=(-1.48479939533)*x[0]+(-0.839474409588)*x[1]
        s[2,0,2,1]=(-0.662177361426)*x[0]+(0.380170348062)*x[1]
        s[2,1,0,0]=(1.13771311046)*x[0]+(-1.59960740614)*x[1]
        s[2,1,0,1]=(0.925401031957)*x[0]+(1.30170041054)*x[1]
        s[2,1,1,0]=(0.455911230864)*x[0]+(0.755253828053)*x[1]
        s[2,1,1,1]=(0.64783771432)*x[0]+(0.251664567771)*x[1]
        s[2,1,2,0]=(-0.0449523798045)*x[0]+(-0.614423607357)*x[1]
        s[2,1,2,1]=(0.635435155497)*x[0]+(0.218411499589)*x[1]
        s[2,2,0,0]=(1.07064943699)*x[0]+(1.33695839858)*x[1]
        s[2,2,0,1]=(0.286067300461)*x[0]+(-1.19968009538)*x[1]
        s[2,2,1,0]=(0.373936537767)*x[0]+(-0.684346831863)*x[1]
        s[2,2,1,1]=(0.709764408449)*x[0]+(0.560181368149)*x[1]
        s[2,2,2,0]=(0.687217025484)*x[0]+(0.638368794102)*x[1]
        s[2,2,2,1]=(-1.13694076648)*x[0]+(1.31012182384)*x[1]
        s[2,3,0,0]=(0.784452969448)*x[0]+(-0.256804074025)*x[1]
        s[2,3,0,1]=(-0.528533290176)*x[0]+(0.230293895766)*x[1]
        s[2,3,1,0]=(-0.152281766055)*x[0]+(-1.29007246581)*x[1]
        s[2,3,1,1]=(-0.113853891603)*x[0]+(-0.146975472116)*x[1]
        s[2,3,2,0]=(-1.16737851429)*x[0]+(1.18358630738)*x[1]
        s[2,3,2,1]=(-0.110242592809)*x[0]+(-0.228040285919)*x[1]
        s[2,4,0,0]=(-0.615024817893)*x[0]+(1.19310484943)*x[1]
        s[2,4,0,1]=(0.405843983847)*x[0]+(-0.7683655961)*x[1]
        s[2,4,1,0]=(-1.94063527375)*x[0]+(0.135437762233)*x[1]
        s[2,4,1,1]=(-0.680568187039)*x[0]+(0.411161107413)*x[1]
        s[2,4,2,0]=(-1.95575235842)*x[0]+(0.384766876765)*x[1]
        s[2,4,2,1]=(1.37236985627)*x[0]+(0.0082159494728)*x[1]
        s[3,0,0,0]=(1.63255667845)*x[0]+(0.0225121724188)*x[1]
        s[3,0,0,1]=(1.2376463155)*x[0]+(1.32594028087)*x[1]
        s[3,0,1,0]=(1.54455876996)*x[0]+(-1.21021289438)*x[1]
        s[3,0,1,1]=(-1.81162592117)*x[0]+(-0.297403340122)*x[1]
        s[3,0,2,0]=(0.806731759445)*x[0]+(-0.981072370255)*x[1]
        s[3,0,2,1]=(-0.15215786955)*x[0]+(0.163165607014)*x[1]
        s[3,1,0,0]=(0.233006578932)*x[0]+(-1.29832996415)*x[1]
        s[3,1,0,1]=(0.96755903915)*x[0]+(-0.00597412239186)*x[1]
        s[3,1,1,0]=(0.0413376466172)*x[0]+(1.26720441986)*x[1]
        s[3,1,1,1]=(0.676509459424)*x[0]+(1.2621989434)*x[1]
        s[3,1,2,0]=(-0.211034948146)*x[0]+(-1.37470768893)*x[1]
        s[3,1,2,1]=(-0.402208754333)*x[0]+(1.25258243878)*x[1]
        s[3,2,0,0]=(-0.529199129991)*x[0]+(-0.431490132335)*x[1]
        s[3,2,0,1]=(-0.850343939948)*x[0]+(0.0602529984167)*x[1]
        s[3,2,1,0]=(-0.420125360943)*x[0]+(1.24318875197)*x[1]
        s[3,2,1,1]=(0.753685328318)*x[0]+(0.354913692774)*x[1]
        s[3,2,2,0]=(-1.17147381854)*x[0]+(0.850516886779)*x[1]
        s[3,2,2,1]=(1.48587305579)*x[0]+(1.31835868193)*x[1]
        s[3,3,0,0]=(1.80808607564)*x[0]+(-0.124508319578)*x[1]
        s[3,3,0,1]=(-1.84429695094)*x[0]+(-0.254797649176)*x[1]
        s[3,3,1,0]=(1.17110770558)*x[0]+(0.89756258226)*x[1]
        s[3,3,1,1]=(0.692264407822)*x[0]+(-1.00105988426)*x[1]
        s[3,3,2,0]=(-0.505689548372)*x[0]+(0.365519850058)*x[1]
        s[3,3,2,1]=(0.651544272536)*x[0]+(1.09868410461)*x[1]
        s[3,4,0,0]=(0.155129169877)*x[0]+(0.4435929499)*x[1]
        s[3,4,0,1]=(1.65974703109)*x[0]+(-0.776830635917)*x[1]
        s[3,4,1,0]=(-1.05612271738)*x[0]+(1.63557694551)*x[1]
        s[3,4,1,1]=(0.69435492251)*x[0]+(-1.20394873527)*x[1]
        s[3,4,2,0]=(-0.237598987964)*x[0]+(0.404820539664)*x[1]
        s[3,4,2,1]=(-1.32483403281)*x[0]+(0.431217340618)*x[1]
        ref[0,0,0,0]=(-0.9809347951)*x_ref[0]+(-1.05666590015)*x_ref[1]
        ref[0,0,0,1]=(1.15925248094)*x_ref[0]+(1.22070923582)*x_ref[1]
        ref[0,0,1,0]=(-0.0724361628927)*x_ref[0]+(-0.439807238916)*x_ref[1]
        ref[0,0,1,1]=(-0.608816713414)*x_ref[0]+(1.78792604745)*x_ref[1]
        ref[0,0,2,0]=(-0.78903506363)*x_ref[0]+(-0.75896072902)*x_ref[1]
        ref[0,0,2,1]=(0.516424260522)*x_ref[0]+(1.16788015378)*x_ref[1]
        ref[0,1,0,0]=(0.383586654748)*x_ref[0]+(1.4179924146)*x_ref[1]
        ref[0,1,0,1]=(0.97236907624)*x_ref[0]+(-0.438982550276)*x_ref[1]
        ref[0,1,1,0]=(-0.645083297783)*x_ref[0]+(-0.740808997218)*x_ref[1]
        ref[0,1,1,1]=(0.330652781935)*x_ref[0]+(0.519132367114)*x_ref[1]
        ref[0,1,2,0]=(1.25159174261)*x_ref[0]+(1.26021904904)*x_ref[1]
        ref[0,1,2,1]=(-0.239621293462)*x_ref[0]+(0.0244107517523)*x_ref[1]
        ref[0,2,0,0]=(-0.286208240697)*x_ref[0]+(-0.314157258671)*x_ref[1]
        ref[0,2,0,1]=(-0.0856319829397)*x_ref[0]+(0.278603809503)*x_ref[1]
        ref[0,2,1,0]=(0.107199217139)*x_ref[0]+(0.0948664401127)*x_ref[1]
        ref[0,2,1,1]=(-0.24620060163)*x_ref[0]+(-1.42942082558)*x_ref[1]
        ref[0,2,2,0]=(0.845285367352)*x_ref[0]+(0.408676159609)*x_ref[1]
        ref[0,2,2,1]=(-1.49467460394)*x_ref[0]+(0.270722457265)*x_ref[1]
        ref[0,3,0,0]=(0.469802598061)*x_ref[0]+(-0.801790345348)*x_ref[1]
        ref[0,3,0,1]=(1.71249257903)*x_ref[0]+(-0.776797378881)*x_ref[1]
        ref[0,3,1,0]=(0.0540049036319)*x_ref[0]+(1.19031969164)*x_ref[1]
        ref[0,3,1,1]=(0.161929130968)*x_ref[0]+(-0.525381566632)*x_ref[1]
        ref[0,3,2,0]=(-0.646041965665)*x_ref[0]+(-0.123993555032)*x_ref[1]
        ref[0,3,2,1]=(0.0966044205198)*x_ref[0]+(1.24780552685)*x_ref[1]
        ref[0,4,0,0]=(-0.289531041539)*x_ref[0]+(1.21117470928)*x_ref[1]
        ref[0,4,0,1]=(-0.112215925799)*x_ref[0]+(0.598717725332)*x_ref[1]
        ref[0,4,1,0]=(0.162936056701)*x_ref[0]+(-1.02395851472)*x_ref[1]
        ref[0,4,1,1]=(0.694066364547)*x_ref[0]+(1.27205020727)*x_ref[1]
        ref[0,4,2,0]=(1.17253981195)*x_ref[0]+(0.130572791304)*x_ref[1]
        ref[0,4,2,1]=(-0.555241229093)*x_ref[0]+(-0.881115061403)*x_ref[1]
        ref[1,0,0,0]=(0.574137047465)*x_ref[0]+(-1.10459006911)*x_ref[1]
        ref[1,0,0,1]=(0.441431076852)*x_ref[0]+(-0.133156832361)*x_ref[1]
        ref[1,0,1,0]=(0.217136655494)*x_ref[0]+(0.400834952015)*x_ref[1]
        ref[1,0,1,1]=(0.516360280675)*x_ref[0]+(0.532907430046)*x_ref[1]
        ref[1,0,2,0]=(-0.340764752946)*x_ref[0]+(-1.26671004418)*x_ref[1]
        ref[1,0,2,1]=(-1.23231139783)*x_ref[0]+(0.61934930883)*x_ref[1]
        ref[1,1,0,0]=(1.76821949014)*x_ref[0]+(1.67542625016)*x_ref[1]
        ref[1,1,0,1]=(0.901675970272)*x_ref[0]+(-0.0567109697208)*x_ref[1]
        ref[1,1,1,0]=(1.21143551644)*x_ref[0]+(0.375877605521)*x_ref[1]
        ref[1,1,1,1]=(-0.699830133987)*x_ref[0]+(0.174660968691)*x_ref[1]
        ref[1,1,2,0]=(1.07839763024)*x_ref[0]+(-0.063554858486)*x_ref[1]
        ref[1,1,2,1]=(0.131335396282)*x_ref[0]+(0.109893568063)*x_ref[1]
        ref[1,2,0,0]=(-0.562355685272)*x_ref[0]+(0.495744936869)*x_ref[1]
        ref[1,2,0,1]=(0.0747584188968)*x_ref[0]+(1.94924356864)*x_ref[1]
        ref[1,2,1,0]=(-0.103515803094)*x_ref[0]+(0.545380886507)*x_ref[1]
        ref[1,2,1,1]=(-0.708926705928)*x_ref[0]+(-0.0610123075525)*x_ref[1]
        ref[1,2,2,0]=(1.06802154614)*x_ref[0]+(0.0622475519235)*x_ref[1]
        ref[1,2,2,1]=(0.103937010427)*x_ref[0]+(-0.70201008693)*x_ref[1]
        ref[1,3,0,0]=(0.0730726824626)*x_ref[0]+(0.39791188789)*x_ref[1]
        ref[1,3,0,1]=(0.455273194688)*x_ref[0]+(1.65845657261)*x_ref[1]
        ref[1,3,1,0]=(0.0807767752424)*x_ref[0]+(-0.765026395125)*x_ref[1]
        ref[1,3,1,1]=(0.0404890748976)*x_ref[0]+(-0.383198358048)*x_ref[1]
        ref[1,3,2,0]=(1.26048751933)*x_ref[0]+(0.00329382174075)*x_ref[1]
        ref[1,3,2,1]=(-1.21518062161)*x_ref[0]+(-0.325769725861)*x_ref[1]
        ref[1,4,0,0]=(0.682181651323)*x_ref[0]+(-0.375432356672)*x_ref[1]
        ref[1,4,0,1]=(-1.09892909715)*x_ref[0]+(0.207984305416)*x_ref[1]
        ref[1,4,1,0]=(-1.2375595164)*x_ref[0]+(0.215243019189)*x_ref[1]
        ref[1,4,1,1]=(-0.595125980205)*x_ref[0]+(0.29104395616)*x_ref[1]
        ref[1,4,2,0]=(0.127947393316)*x_ref[0]+(1.43408039596)*x_ref[1]
        ref[1,4,2,1]=(0.468574871016)*x_ref[0]+(0.333319611281)*x_ref[1]
        ref[2,0,0,0]=(0.369954209324)*x_ref[0]+(-1.09614147452)*x_ref[1]
        ref[2,0,0,1]=(-0.120003056441)*x_ref[0]+(-0.172964298439)*x_ref[1]
        ref[2,0,1,0]=(1.38230317911)*x_ref[0]+(1.25730140777)*x_ref[1]
        ref[2,0,1,1]=(1.10857329622)*x_ref[0]+(0.449561992049)*x_ref[1]
        ref[2,0,2,0]=(-1.48479939533)*x_ref[0]+(-0.839474409588)*x_ref[1]
        ref[2,0,2,1]=(-0.662177361426)*x_ref[0]+(0.380170348062)*x_ref[1]
        ref[2,1,0,0]=(1.13771311046)*x_ref[0]+(-1.59960740614)*x_ref[1]
        ref[2,1,0,1]=(0.925401031957)*x_ref[0]+(1.30170041054)*x_ref[1]
        ref[2,1,1,0]=(0.455911230864)*x_ref[0]+(0.755253828053)*x_ref[1]
        ref[2,1,1,1]=(0.64783771432)*x_ref[0]+(0.251664567771)*x_ref[1]
        ref[2,1,2,0]=(-0.0449523798045)*x_ref[0]+(-0.614423607357)*x_ref[1]
        ref[2,1,2,1]=(0.635435155497)*x_ref[0]+(0.218411499589)*x_ref[1]
        ref[2,2,0,0]=(1.07064943699)*x_ref[0]+(1.33695839858)*x_ref[1]
        ref[2,2,0,1]=(0.286067300461)*x_ref[0]+(-1.19968009538)*x_ref[1]
        ref[2,2,1,0]=(0.373936537767)*x_ref[0]+(-0.684346831863)*x_ref[1]
        ref[2,2,1,1]=(0.709764408449)*x_ref[0]+(0.560181368149)*x_ref[1]
        ref[2,2,2,0]=(0.687217025484)*x_ref[0]+(0.638368794102)*x_ref[1]
        ref[2,2,2,1]=(-1.13694076648)*x_ref[0]+(1.31012182384)*x_ref[1]
        ref[2,3,0,0]=(0.784452969448)*x_ref[0]+(-0.256804074025)*x_ref[1]
        ref[2,3,0,1]=(-0.528533290176)*x_ref[0]+(0.230293895766)*x_ref[1]
        ref[2,3,1,0]=(-0.152281766055)*x_ref[0]+(-1.29007246581)*x_ref[1]
        ref[2,3,1,1]=(-0.113853891603)*x_ref[0]+(-0.146975472116)*x_ref[1]
        ref[2,3,2,0]=(-1.16737851429)*x_ref[0]+(1.18358630738)*x_ref[1]
        ref[2,3,2,1]=(-0.110242592809)*x_ref[0]+(-0.228040285919)*x_ref[1]
        ref[2,4,0,0]=(-0.615024817893)*x_ref[0]+(1.19310484943)*x_ref[1]
        ref[2,4,0,1]=(0.405843983847)*x_ref[0]+(-0.7683655961)*x_ref[1]
        ref[2,4,1,0]=(-1.94063527375)*x_ref[0]+(0.135437762233)*x_ref[1]
        ref[2,4,1,1]=(-0.680568187039)*x_ref[0]+(0.411161107413)*x_ref[1]
        ref[2,4,2,0]=(-1.95575235842)*x_ref[0]+(0.384766876765)*x_ref[1]
        ref[2,4,2,1]=(1.37236985627)*x_ref[0]+(0.0082159494728)*x_ref[1]
        ref[3,0,0,0]=(1.63255667845)*x_ref[0]+(0.0225121724188)*x_ref[1]
        ref[3,0,0,1]=(1.2376463155)*x_ref[0]+(1.32594028087)*x_ref[1]
        ref[3,0,1,0]=(1.54455876996)*x_ref[0]+(-1.21021289438)*x_ref[1]
        ref[3,0,1,1]=(-1.81162592117)*x_ref[0]+(-0.297403340122)*x_ref[1]
        ref[3,0,2,0]=(0.806731759445)*x_ref[0]+(-0.981072370255)*x_ref[1]
        ref[3,0,2,1]=(-0.15215786955)*x_ref[0]+(0.163165607014)*x_ref[1]
        ref[3,1,0,0]=(0.233006578932)*x_ref[0]+(-1.29832996415)*x_ref[1]
        ref[3,1,0,1]=(0.96755903915)*x_ref[0]+(-0.00597412239186)*x_ref[1]
        ref[3,1,1,0]=(0.0413376466172)*x_ref[0]+(1.26720441986)*x_ref[1]
        ref[3,1,1,1]=(0.676509459424)*x_ref[0]+(1.2621989434)*x_ref[1]
        ref[3,1,2,0]=(-0.211034948146)*x_ref[0]+(-1.37470768893)*x_ref[1]
        ref[3,1,2,1]=(-0.402208754333)*x_ref[0]+(1.25258243878)*x_ref[1]
        ref[3,2,0,0]=(-0.529199129991)*x_ref[0]+(-0.431490132335)*x_ref[1]
        ref[3,2,0,1]=(-0.850343939948)*x_ref[0]+(0.0602529984167)*x_ref[1]
        ref[3,2,1,0]=(-0.420125360943)*x_ref[0]+(1.24318875197)*x_ref[1]
        ref[3,2,1,1]=(0.753685328318)*x_ref[0]+(0.354913692774)*x_ref[1]
        ref[3,2,2,0]=(-1.17147381854)*x_ref[0]+(0.850516886779)*x_ref[1]
        ref[3,2,2,1]=(1.48587305579)*x_ref[0]+(1.31835868193)*x_ref[1]
        ref[3,3,0,0]=(1.80808607564)*x_ref[0]+(-0.124508319578)*x_ref[1]
        ref[3,3,0,1]=(-1.84429695094)*x_ref[0]+(-0.254797649176)*x_ref[1]
        ref[3,3,1,0]=(1.17110770558)*x_ref[0]+(0.89756258226)*x_ref[1]
        ref[3,3,1,1]=(0.692264407822)*x_ref[0]+(-1.00105988426)*x_ref[1]
        ref[3,3,2,0]=(-0.505689548372)*x_ref[0]+(0.365519850058)*x_ref[1]
        ref[3,3,2,1]=(0.651544272536)*x_ref[0]+(1.09868410461)*x_ref[1]
        ref[3,4,0,0]=(0.155129169877)*x_ref[0]+(0.4435929499)*x_ref[1]
        ref[3,4,0,1]=(1.65974703109)*x_ref[0]+(-0.776830635917)*x_ref[1]
        ref[3,4,1,0]=(-1.05612271738)*x_ref[0]+(1.63557694551)*x_ref[1]
        ref[3,4,1,1]=(0.69435492251)*x_ref[0]+(-1.20394873527)*x_ref[1]
        ref[3,4,2,0]=(-0.237598987964)*x_ref[0]+(0.404820539664)*x_ref[1]
        ref[3,4,2,1]=(-1.32483403281)*x_ref[0]+(0.431217340618)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.270116752643)*x[0]+(0.0715809194071)*x[1]+(0.930011298681)*x[2]
        s[0,0,0,1]=(-0.0435807732016)*x[0]+(-0.492622221729)*x[1]+(-1.36754174971)*x[2]
        s[0,0,1,0]=(-0.949818131569)*x[0]+(1.08437418341)*x[1]+(0.527267447353)*x[2]
        s[0,0,1,1]=(-0.164876627537)*x[0]+(-1.44143896699)*x[1]+(0.0355140164081)*x[2]
        s[0,0,2,0]=(-0.530229192916)*x[0]+(-0.988412045132)*x[1]+(-0.466520393878)*x[2]
        s[0,0,2,1]=(-0.0166373496749)*x[0]+(0.322621663585)*x[1]+(-0.102019336059)*x[2]
        s[0,1,0,0]=(-1.36627178258)*x[0]+(-0.803463758822)*x[1]+(0.734446026661)*x[2]
        s[0,1,0,1]=(-0.804729478825)*x[0]+(-1.24406753288)*x[1]+(-0.755743791016)*x[2]
        s[0,1,1,0]=(1.03869027308)*x[0]+(-0.25237293575)*x[1]+(1.43737134487)*x[2]
        s[0,1,1,1]=(1.48269192251)*x[0]+(0.045590449115)*x[1]+(0.47425103841)*x[2]
        s[0,1,2,0]=(-0.738695287005)*x[0]+(1.01938431903)*x[1]+(1.20767534795)*x[2]
        s[0,1,2,1]=(-0.237445941287)*x[0]+(0.766593746786)*x[1]+(0.0297279831889)*x[2]
        s[0,2,0,0]=(-0.804902383375)*x[0]+(0.0962339257662)*x[1]+(1.37407473707)*x[2]
        s[0,2,0,1]=(-0.0821182376238)*x[0]+(0.35489327051)*x[1]+(-1.43235115632)*x[2]
        s[0,2,1,0]=(1.11593707013)*x[0]+(0.744039820943)*x[1]+(0.294815281064)*x[2]
        s[0,2,1,1]=(0.0789595478782)*x[0]+(0.719656781874)*x[1]+(0.355275528053)*x[2]
        s[0,2,2,0]=(1.20514986435)*x[0]+(0.203004753589)*x[1]+(-0.0731941650839)*x[2]
        s[0,2,2,1]=(-1.23231910051)*x[0]+(-0.470834142341)*x[1]+(-0.633518095593)*x[2]
        s[0,3,0,0]=(0.468164093965)*x[0]+(-1.15996958069)*x[1]+(0.922274695212)*x[2]
        s[0,3,0,1]=(-0.964250141036)*x[0]+(-0.769747467017)*x[1]+(-1.16369265956)*x[2]
        s[0,3,1,0]=(1.1622810944)*x[0]+(-1.72794997517)*x[1]+(-0.297833274612)*x[2]
        s[0,3,1,1]=(1.20239531447)*x[0]+(-0.195111187444)*x[1]+(-0.720186501147)*x[2]
        s[0,3,2,0]=(-0.0760001358746)*x[0]+(-1.65266437471)*x[1]+(-0.496163951015)*x[2]
        s[0,3,2,1]=(-1.4998248458)*x[0]+(-1.73305168566)*x[1]+(-0.107384086634)*x[2]
        s[0,4,0,0]=(0.235699357585)*x[0]+(-0.213348404403)*x[1]+(0.35368491512)*x[2]
        s[0,4,0,1]=(0.264641788778)*x[0]+(0.435251584452)*x[1]+(0.22539974851)*x[2]
        s[0,4,1,0]=(0.540203510472)*x[0]+(-1.33243180984)*x[1]+(-0.520357911324)*x[2]
        s[0,4,1,1]=(-0.629868492098)*x[0]+(-0.403402885662)*x[1]+(-0.184983842426)*x[2]
        s[0,4,2,0]=(1.17843494613)*x[0]+(0.562783141759)*x[1]+(-1.48575741726)*x[2]
        s[0,4,2,1]=(0.623796209289)*x[0]+(1.02627846638)*x[1]+(1.39369241754)*x[2]
        s[1,0,0,0]=(1.9855507901)*x[0]+(0.152976407288)*x[1]+(0.0931547545156)*x[2]
        s[1,0,0,1]=(-0.422055088151)*x[0]+(-0.509449435814)*x[1]+(-1.31705141661)*x[2]
        s[1,0,1,0]=(-0.24193816853)*x[0]+(-0.68450536388)*x[1]+(-0.309024887085)*x[2]
        s[1,0,1,1]=(1.24557169323)*x[0]+(0.353538565492)*x[1]+(0.115821365646)*x[2]
        s[1,0,2,0]=(-0.280246915004)*x[0]+(-0.0972556765171)*x[1]+(0.758816560386)*x[2]
        s[1,0,2,1]=(1.0396259589)*x[0]+(-0.239507470189)*x[1]+(-0.987815676492)*x[2]
        s[1,1,0,0]=(0.245203366317)*x[0]+(-1.53269800659)*x[1]+(0.619360300925)*x[2]
        s[1,1,0,1]=(0.246406467482)*x[0]+(-0.729988292889)*x[1]+(0.585422287189)*x[2]
        s[1,1,1,0]=(1.19810038121)*x[0]+(-0.72772445701)*x[1]+(1.11619392954)*x[2]
        s[1,1,1,1]=(-0.311984976134)*x[0]+(-0.188443023344)*x[1]+(-0.100078290654)*x[2]
        s[1,1,2,0]=(0.158578485066)*x[0]+(0.214804573675)*x[1]+(-0.22418434724)*x[2]
        s[1,1,2,1]=(-0.0550611413075)*x[0]+(-0.147088267891)*x[1]+(-0.269933900128)*x[2]
        s[1,2,0,0]=(-0.8618782319)*x[0]+(-0.770284701216)*x[1]+(-0.407458817279)*x[2]
        s[1,2,0,1]=(-0.242336600721)*x[0]+(-1.57823911131)*x[1]+(0.452055841078)*x[2]
        s[1,2,1,0]=(0.0537219652862)*x[0]+(0.18864518829)*x[1]+(0.021034116382)*x[2]
        s[1,2,1,1]=(0.143799709809)*x[0]+(0.787934959856)*x[1]+(-0.283457275339)*x[2]
        s[1,2,2,0]=(0.475965504456)*x[0]+(-0.744950301051)*x[1]+(0.45308618026)*x[2]
        s[1,2,2,1]=(0.561593580643)*x[0]+(0.121626528038)*x[1]+(-1.24221676348)*x[2]
        s[1,3,0,0]=(-0.340056043985)*x[0]+(-0.434795551938)*x[1]+(-0.291179084342)*x[2]
        s[1,3,0,1]=(0.312318149724)*x[0]+(1.66950279503)*x[1]+(-1.1622076779)*x[2]
        s[1,3,1,0]=(0.0965172054443)*x[0]+(-0.64102022474)*x[1]+(-0.173612340622)*x[2]
        s[1,3,1,1]=(1.88701305968)*x[0]+(-1.66794136241)*x[1]+(1.56059966446)*x[2]
        s[1,3,2,0]=(-0.497773999274)*x[0]+(0.395345962943)*x[1]+(-0.499105933847)*x[2]
        s[1,3,2,1]=(-0.312096437005)*x[0]+(-1.23301763457)*x[1]+(0.198835942067)*x[2]
        s[1,4,0,0]=(0.076377295342)*x[0]+(-0.487847012612)*x[1]+(-0.084314987823)*x[2]
        s[1,4,0,1]=(0.666434654203)*x[0]+(-0.220376188514)*x[1]+(-0.355749198488)*x[2]
        s[1,4,1,0]=(-0.701104468581)*x[0]+(0.828174884422)*x[1]+(0.309374641659)*x[2]
        s[1,4,1,1]=(0.526340204296)*x[0]+(0.571598927007)*x[1]+(-0.0793826750713)*x[2]
        s[1,4,2,0]=(0.775563727486)*x[0]+(0.248228533771)*x[1]+(-1.18169855975)*x[2]
        s[1,4,2,1]=(-0.399357449979)*x[0]+(-0.723554468137)*x[1]+(0.0231913550517)*x[2]
        s[2,0,0,0]=(-0.0504735749489)*x[0]+(0.506564345355)*x[1]+(0.790336060204)*x[2]
        s[2,0,0,1]=(0.426187262022)*x[0]+(0.966680952632)*x[1]+(0.410542727369)*x[2]
        s[2,0,1,0]=(0.594102954821)*x[0]+(0.0639755896611)*x[1]+(-0.675769510238)*x[2]
        s[2,0,1,1]=(0.367479711308)*x[0]+(0.150736830273)*x[1]+(1.15797942999)*x[2]
        s[2,0,2,0]=(0.875207101066)*x[0]+(-1.12939492894)*x[1]+(0.73561332295)*x[2]
        s[2,0,2,1]=(-1.70885193598)*x[0]+(0.835257764029)*x[1]+(-1.58521909429)*x[2]
        s[2,1,0,0]=(-0.144255046397)*x[0]+(0.590074178586)*x[1]+(0.592144641271)*x[2]
        s[2,1,0,1]=(-0.0688470601204)*x[0]+(-0.153340975531)*x[1]+(1.15241895553)*x[2]
        s[2,1,1,0]=(-0.73257033584)*x[0]+(-0.563316507818)*x[1]+(-1.05620654545)*x[2]
        s[2,1,1,1]=(-1.29833350092)*x[0]+(0.42237934165)*x[1]+(0.0725296431412)*x[2]
        s[2,1,2,0]=(-0.837473685756)*x[0]+(-0.425072509905)*x[1]+(-0.463397001131)*x[2]
        s[2,1,2,1]=(-0.937038584847)*x[0]+(0.531989188194)*x[1]+(-1.06694316698)*x[2]
        s[2,2,0,0]=(-0.846346120269)*x[0]+(-0.639618775615)*x[1]+(-0.928154018676)*x[2]
        s[2,2,0,1]=(-0.726631889289)*x[0]+(-0.1318624965)*x[1]+(-0.343729797494)*x[2]
        s[2,2,1,0]=(0.802665744539)*x[0]+(-0.115632355474)*x[1]+(0.303919795863)*x[2]
        s[2,2,1,1]=(0.0371825373918)*x[0]+(-1.17799388388)*x[1]+(-0.223043449617)*x[2]
        s[2,2,2,0]=(0.75988617552)*x[0]+(-0.608977501492)*x[1]+(1.75211913311)*x[2]
        s[2,2,2,1]=(-0.291397991509)*x[0]+(-1.55173354063)*x[1]+(-0.461269447756)*x[2]
        s[2,3,0,0]=(0.351624965509)*x[0]+(-1.32799214259)*x[1]+(1.58946725409)*x[2]
        s[2,3,0,1]=(0.180149992744)*x[0]+(1.07861987473)*x[1]+(0.834357928788)*x[2]
        s[2,3,1,0]=(0.0475016328103)*x[0]+(0.707449786509)*x[1]+(-0.370421903676)*x[2]
        s[2,3,1,1]=(0.553988297947)*x[0]+(1.13388241001)*x[1]+(0.575604300083)*x[2]
        s[2,3,2,0]=(-0.249047035016)*x[0]+(0.15013128589)*x[1]+(1.3953926353)*x[2]
        s[2,3,2,1]=(-0.783008188949)*x[0]+(0.861992436037)*x[1]+(-0.929496477422)*x[2]
        s[2,4,0,0]=(0.264348996811)*x[0]+(-1.23207394036)*x[1]+(-1.37804290092)*x[2]
        s[2,4,0,1]=(0.360782273055)*x[0]+(-0.270520439821)*x[1]+(-1.01285201623)*x[2]
        s[2,4,1,0]=(0.280578579439)*x[0]+(0.584696145403)*x[1]+(-1.21852764117)*x[2]
        s[2,4,1,1]=(0.112102121128)*x[0]+(-0.0619663521973)*x[1]+(1.08490293257)*x[2]
        s[2,4,2,0]=(-1.89749569924)*x[0]+(-0.21217889191)*x[1]+(0.635524779965)*x[2]
        s[2,4,2,1]=(-0.309239337357)*x[0]+(0.0116284822724)*x[1]+(0.0107247712951)*x[2]
        s[3,0,0,0]=(1.23545873872)*x[0]+(-1.29896744411)*x[1]+(0.855094335492)*x[2]
        s[3,0,0,1]=(-0.51465695324)*x[0]+(1.47799693308)*x[1]+(0.0718360423316)*x[2]
        s[3,0,1,0]=(0.220505442402)*x[0]+(-0.3171385194)*x[1]+(0.27954628665)*x[2]
        s[3,0,1,1]=(0.860582046972)*x[0]+(1.04565549687)*x[1]+(1.18108647078)*x[2]
        s[3,0,2,0]=(-0.593547630941)*x[0]+(-0.0390465590147)*x[1]+(0.00129707682534)*x[2]
        s[3,0,2,1]=(1.39210293617)*x[0]+(-1.05660957806)*x[1]+(-1.20489109051)*x[2]
        s[3,1,0,0]=(-0.536035988946)*x[0]+(-0.594403345465)*x[1]+(1.11719111746)*x[2]
        s[3,1,0,1]=(-0.0403557954127)*x[0]+(-1.32391025453)*x[1]+(-0.639127329416)*x[2]
        s[3,1,1,0]=(-0.495803710666)*x[0]+(-0.199832989186)*x[1]+(-0.832862076561)*x[2]
        s[3,1,1,1]=(0.0216882402803)*x[0]+(0.367268873436)*x[1]+(0.295944335418)*x[2]
        s[3,1,2,0]=(0.172936506269)*x[0]+(-0.339886198196)*x[1]+(0.268201481244)*x[2]
        s[3,1,2,1]=(0.665786569897)*x[0]+(-0.290637226736)*x[1]+(0.617677178537)*x[2]
        s[3,2,0,0]=(-0.657902185911)*x[0]+(-1.34292379802)*x[1]+(0.0641950819134)*x[2]
        s[3,2,0,1]=(0.352039175923)*x[0]+(-0.483047873735)*x[1]+(0.47997600743)*x[2]
        s[3,2,1,0]=(0.944619484106)*x[0]+(-0.308755547249)*x[1]+(-0.765906593989)*x[2]
        s[3,2,1,1]=(-1.09674781865)*x[0]+(0.013069257236)*x[1]+(-0.0741247275628)*x[2]
        s[3,2,2,0]=(0.400057737624)*x[0]+(-1.16165281514)*x[1]+(-0.789757454741)*x[2]
        s[3,2,2,1]=(0.367100312493)*x[0]+(-0.547338908792)*x[1]+(-0.409173859392)*x[2]
        s[3,3,0,0]=(0.890118369638)*x[0]+(1.201296587)*x[1]+(-0.0319246476545)*x[2]
        s[3,3,0,1]=(0.620955702501)*x[0]+(0.438440818114)*x[1]+(0.499747510411)*x[2]
        s[3,3,1,0]=(1.61217537222)*x[0]+(-1.21323714148)*x[1]+(-1.06137097571)*x[2]
        s[3,3,1,1]=(-1.75942413542)*x[0]+(0.440345678738)*x[1]+(0.625299774096)*x[2]
        s[3,3,2,0]=(-0.878497266865)*x[0]+(-1.15879244679)*x[1]+(1.03993124425)*x[2]
        s[3,3,2,1]=(-0.0138445968491)*x[0]+(0.547910692711)*x[1]+(0.971642264613)*x[2]
        s[3,4,0,0]=(-0.107769272336)*x[0]+(-0.987911384393)*x[1]+(1.2076201556)*x[2]
        s[3,4,0,1]=(-0.556570952071)*x[0]+(-0.284568452888)*x[1]+(1.23341917108)*x[2]
        s[3,4,1,0]=(1.56461193947)*x[0]+(0.185382898933)*x[1]+(-1.67714180845)*x[2]
        s[3,4,1,1]=(-0.894934034676)*x[0]+(1.05560233139)*x[1]+(0.651029150593)*x[2]
        s[3,4,2,0]=(0.0559603779066)*x[0]+(-1.72752549101)*x[1]+(-1.02017554749)*x[2]
        s[3,4,2,1]=(-1.75424333815)*x[0]+(-0.198869840282)*x[1]+(0.475065396299)*x[2]
        ref[0,0,0,0]=(-0.270116752643)*x_ref[0]+(0.0715809194071)*x_ref[1]+(0.930011298681)*x_ref[2]
        ref[0,0,0,1]=(-0.0435807732016)*x_ref[0]+(-0.492622221729)*x_ref[1]+(-1.36754174971)*x_ref[2]
        ref[0,0,1,0]=(-0.949818131569)*x_ref[0]+(1.08437418341)*x_ref[1]+(0.527267447353)*x_ref[2]
        ref[0,0,1,1]=(-0.164876627537)*x_ref[0]+(-1.44143896699)*x_ref[1]+(0.0355140164081)*x_ref[2]
        ref[0,0,2,0]=(-0.530229192916)*x_ref[0]+(-0.988412045132)*x_ref[1]+(-0.466520393878)*x_ref[2]
        ref[0,0,2,1]=(-0.0166373496749)*x_ref[0]+(0.322621663585)*x_ref[1]+(-0.102019336059)*x_ref[2]
        ref[0,1,0,0]=(-1.36627178258)*x_ref[0]+(-0.803463758822)*x_ref[1]+(0.734446026661)*x_ref[2]
        ref[0,1,0,1]=(-0.804729478825)*x_ref[0]+(-1.24406753288)*x_ref[1]+(-0.755743791016)*x_ref[2]
        ref[0,1,1,0]=(1.03869027308)*x_ref[0]+(-0.25237293575)*x_ref[1]+(1.43737134487)*x_ref[2]
        ref[0,1,1,1]=(1.48269192251)*x_ref[0]+(0.045590449115)*x_ref[1]+(0.47425103841)*x_ref[2]
        ref[0,1,2,0]=(-0.738695287005)*x_ref[0]+(1.01938431903)*x_ref[1]+(1.20767534795)*x_ref[2]
        ref[0,1,2,1]=(-0.237445941287)*x_ref[0]+(0.766593746786)*x_ref[1]+(0.0297279831889)*x_ref[2]
        ref[0,2,0,0]=(-0.804902383375)*x_ref[0]+(0.0962339257662)*x_ref[1]+(1.37407473707)*x_ref[2]
        ref[0,2,0,1]=(-0.0821182376238)*x_ref[0]+(0.35489327051)*x_ref[1]+(-1.43235115632)*x_ref[2]
        ref[0,2,1,0]=(1.11593707013)*x_ref[0]+(0.744039820943)*x_ref[1]+(0.294815281064)*x_ref[2]
        ref[0,2,1,1]=(0.0789595478782)*x_ref[0]+(0.719656781874)*x_ref[1]+(0.355275528053)*x_ref[2]
        ref[0,2,2,0]=(1.20514986435)*x_ref[0]+(0.203004753589)*x_ref[1]+(-0.0731941650839)*x_ref[2]
        ref[0,2,2,1]=(-1.23231910051)*x_ref[0]+(-0.470834142341)*x_ref[1]+(-0.633518095593)*x_ref[2]
        ref[0,3,0,0]=(0.468164093965)*x_ref[0]+(-1.15996958069)*x_ref[1]+(0.922274695212)*x_ref[2]
        ref[0,3,0,1]=(-0.964250141036)*x_ref[0]+(-0.769747467017)*x_ref[1]+(-1.16369265956)*x_ref[2]
        ref[0,3,1,0]=(1.1622810944)*x_ref[0]+(-1.72794997517)*x_ref[1]+(-0.297833274612)*x_ref[2]
        ref[0,3,1,1]=(1.20239531447)*x_ref[0]+(-0.195111187444)*x_ref[1]+(-0.720186501147)*x_ref[2]
        ref[0,3,2,0]=(-0.0760001358746)*x_ref[0]+(-1.65266437471)*x_ref[1]+(-0.496163951015)*x_ref[2]
        ref[0,3,2,1]=(-1.4998248458)*x_ref[0]+(-1.73305168566)*x_ref[1]+(-0.107384086634)*x_ref[2]
        ref[0,4,0,0]=(0.235699357585)*x_ref[0]+(-0.213348404403)*x_ref[1]+(0.35368491512)*x_ref[2]
        ref[0,4,0,1]=(0.264641788778)*x_ref[0]+(0.435251584452)*x_ref[1]+(0.22539974851)*x_ref[2]
        ref[0,4,1,0]=(0.540203510472)*x_ref[0]+(-1.33243180984)*x_ref[1]+(-0.520357911324)*x_ref[2]
        ref[0,4,1,1]=(-0.629868492098)*x_ref[0]+(-0.403402885662)*x_ref[1]+(-0.184983842426)*x_ref[2]
        ref[0,4,2,0]=(1.17843494613)*x_ref[0]+(0.562783141759)*x_ref[1]+(-1.48575741726)*x_ref[2]
        ref[0,4,2,1]=(0.623796209289)*x_ref[0]+(1.02627846638)*x_ref[1]+(1.39369241754)*x_ref[2]
        ref[1,0,0,0]=(1.9855507901)*x_ref[0]+(0.152976407288)*x_ref[1]+(0.0931547545156)*x_ref[2]
        ref[1,0,0,1]=(-0.422055088151)*x_ref[0]+(-0.509449435814)*x_ref[1]+(-1.31705141661)*x_ref[2]
        ref[1,0,1,0]=(-0.24193816853)*x_ref[0]+(-0.68450536388)*x_ref[1]+(-0.309024887085)*x_ref[2]
        ref[1,0,1,1]=(1.24557169323)*x_ref[0]+(0.353538565492)*x_ref[1]+(0.115821365646)*x_ref[2]
        ref[1,0,2,0]=(-0.280246915004)*x_ref[0]+(-0.0972556765171)*x_ref[1]+(0.758816560386)*x_ref[2]
        ref[1,0,2,1]=(1.0396259589)*x_ref[0]+(-0.239507470189)*x_ref[1]+(-0.987815676492)*x_ref[2]
        ref[1,1,0,0]=(0.245203366317)*x_ref[0]+(-1.53269800659)*x_ref[1]+(0.619360300925)*x_ref[2]
        ref[1,1,0,1]=(0.246406467482)*x_ref[0]+(-0.729988292889)*x_ref[1]+(0.585422287189)*x_ref[2]
        ref[1,1,1,0]=(1.19810038121)*x_ref[0]+(-0.72772445701)*x_ref[1]+(1.11619392954)*x_ref[2]
        ref[1,1,1,1]=(-0.311984976134)*x_ref[0]+(-0.188443023344)*x_ref[1]+(-0.100078290654)*x_ref[2]
        ref[1,1,2,0]=(0.158578485066)*x_ref[0]+(0.214804573675)*x_ref[1]+(-0.22418434724)*x_ref[2]
        ref[1,1,2,1]=(-0.0550611413075)*x_ref[0]+(-0.147088267891)*x_ref[1]+(-0.269933900128)*x_ref[2]
        ref[1,2,0,0]=(-0.8618782319)*x_ref[0]+(-0.770284701216)*x_ref[1]+(-0.407458817279)*x_ref[2]
        ref[1,2,0,1]=(-0.242336600721)*x_ref[0]+(-1.57823911131)*x_ref[1]+(0.452055841078)*x_ref[2]
        ref[1,2,1,0]=(0.0537219652862)*x_ref[0]+(0.18864518829)*x_ref[1]+(0.021034116382)*x_ref[2]
        ref[1,2,1,1]=(0.143799709809)*x_ref[0]+(0.787934959856)*x_ref[1]+(-0.283457275339)*x_ref[2]
        ref[1,2,2,0]=(0.475965504456)*x_ref[0]+(-0.744950301051)*x_ref[1]+(0.45308618026)*x_ref[2]
        ref[1,2,2,1]=(0.561593580643)*x_ref[0]+(0.121626528038)*x_ref[1]+(-1.24221676348)*x_ref[2]
        ref[1,3,0,0]=(-0.340056043985)*x_ref[0]+(-0.434795551938)*x_ref[1]+(-0.291179084342)*x_ref[2]
        ref[1,3,0,1]=(0.312318149724)*x_ref[0]+(1.66950279503)*x_ref[1]+(-1.1622076779)*x_ref[2]
        ref[1,3,1,0]=(0.0965172054443)*x_ref[0]+(-0.64102022474)*x_ref[1]+(-0.173612340622)*x_ref[2]
        ref[1,3,1,1]=(1.88701305968)*x_ref[0]+(-1.66794136241)*x_ref[1]+(1.56059966446)*x_ref[2]
        ref[1,3,2,0]=(-0.497773999274)*x_ref[0]+(0.395345962943)*x_ref[1]+(-0.499105933847)*x_ref[2]
        ref[1,3,2,1]=(-0.312096437005)*x_ref[0]+(-1.23301763457)*x_ref[1]+(0.198835942067)*x_ref[2]
        ref[1,4,0,0]=(0.076377295342)*x_ref[0]+(-0.487847012612)*x_ref[1]+(-0.084314987823)*x_ref[2]
        ref[1,4,0,1]=(0.666434654203)*x_ref[0]+(-0.220376188514)*x_ref[1]+(-0.355749198488)*x_ref[2]
        ref[1,4,1,0]=(-0.701104468581)*x_ref[0]+(0.828174884422)*x_ref[1]+(0.309374641659)*x_ref[2]
        ref[1,4,1,1]=(0.526340204296)*x_ref[0]+(0.571598927007)*x_ref[1]+(-0.0793826750713)*x_ref[2]
        ref[1,4,2,0]=(0.775563727486)*x_ref[0]+(0.248228533771)*x_ref[1]+(-1.18169855975)*x_ref[2]
        ref[1,4,2,1]=(-0.399357449979)*x_ref[0]+(-0.723554468137)*x_ref[1]+(0.0231913550517)*x_ref[2]
        ref[2,0,0,0]=(-0.0504735749489)*x_ref[0]+(0.506564345355)*x_ref[1]+(0.790336060204)*x_ref[2]
        ref[2,0,0,1]=(0.426187262022)*x_ref[0]+(0.966680952632)*x_ref[1]+(0.410542727369)*x_ref[2]
        ref[2,0,1,0]=(0.594102954821)*x_ref[0]+(0.0639755896611)*x_ref[1]+(-0.675769510238)*x_ref[2]
        ref[2,0,1,1]=(0.367479711308)*x_ref[0]+(0.150736830273)*x_ref[1]+(1.15797942999)*x_ref[2]
        ref[2,0,2,0]=(0.875207101066)*x_ref[0]+(-1.12939492894)*x_ref[1]+(0.73561332295)*x_ref[2]
        ref[2,0,2,1]=(-1.70885193598)*x_ref[0]+(0.835257764029)*x_ref[1]+(-1.58521909429)*x_ref[2]
        ref[2,1,0,0]=(-0.144255046397)*x_ref[0]+(0.590074178586)*x_ref[1]+(0.592144641271)*x_ref[2]
        ref[2,1,0,1]=(-0.0688470601204)*x_ref[0]+(-0.153340975531)*x_ref[1]+(1.15241895553)*x_ref[2]
        ref[2,1,1,0]=(-0.73257033584)*x_ref[0]+(-0.563316507818)*x_ref[1]+(-1.05620654545)*x_ref[2]
        ref[2,1,1,1]=(-1.29833350092)*x_ref[0]+(0.42237934165)*x_ref[1]+(0.0725296431412)*x_ref[2]
        ref[2,1,2,0]=(-0.837473685756)*x_ref[0]+(-0.425072509905)*x_ref[1]+(-0.463397001131)*x_ref[2]
        ref[2,1,2,1]=(-0.937038584847)*x_ref[0]+(0.531989188194)*x_ref[1]+(-1.06694316698)*x_ref[2]
        ref[2,2,0,0]=(-0.846346120269)*x_ref[0]+(-0.639618775615)*x_ref[1]+(-0.928154018676)*x_ref[2]
        ref[2,2,0,1]=(-0.726631889289)*x_ref[0]+(-0.1318624965)*x_ref[1]+(-0.343729797494)*x_ref[2]
        ref[2,2,1,0]=(0.802665744539)*x_ref[0]+(-0.115632355474)*x_ref[1]+(0.303919795863)*x_ref[2]
        ref[2,2,1,1]=(0.0371825373918)*x_ref[0]+(-1.17799388388)*x_ref[1]+(-0.223043449617)*x_ref[2]
        ref[2,2,2,0]=(0.75988617552)*x_ref[0]+(-0.608977501492)*x_ref[1]+(1.75211913311)*x_ref[2]
        ref[2,2,2,1]=(-0.291397991509)*x_ref[0]+(-1.55173354063)*x_ref[1]+(-0.461269447756)*x_ref[2]
        ref[2,3,0,0]=(0.351624965509)*x_ref[0]+(-1.32799214259)*x_ref[1]+(1.58946725409)*x_ref[2]
        ref[2,3,0,1]=(0.180149992744)*x_ref[0]+(1.07861987473)*x_ref[1]+(0.834357928788)*x_ref[2]
        ref[2,3,1,0]=(0.0475016328103)*x_ref[0]+(0.707449786509)*x_ref[1]+(-0.370421903676)*x_ref[2]
        ref[2,3,1,1]=(0.553988297947)*x_ref[0]+(1.13388241001)*x_ref[1]+(0.575604300083)*x_ref[2]
        ref[2,3,2,0]=(-0.249047035016)*x_ref[0]+(0.15013128589)*x_ref[1]+(1.3953926353)*x_ref[2]
        ref[2,3,2,1]=(-0.783008188949)*x_ref[0]+(0.861992436037)*x_ref[1]+(-0.929496477422)*x_ref[2]
        ref[2,4,0,0]=(0.264348996811)*x_ref[0]+(-1.23207394036)*x_ref[1]+(-1.37804290092)*x_ref[2]
        ref[2,4,0,1]=(0.360782273055)*x_ref[0]+(-0.270520439821)*x_ref[1]+(-1.01285201623)*x_ref[2]
        ref[2,4,1,0]=(0.280578579439)*x_ref[0]+(0.584696145403)*x_ref[1]+(-1.21852764117)*x_ref[2]
        ref[2,4,1,1]=(0.112102121128)*x_ref[0]+(-0.0619663521973)*x_ref[1]+(1.08490293257)*x_ref[2]
        ref[2,4,2,0]=(-1.89749569924)*x_ref[0]+(-0.21217889191)*x_ref[1]+(0.635524779965)*x_ref[2]
        ref[2,4,2,1]=(-0.309239337357)*x_ref[0]+(0.0116284822724)*x_ref[1]+(0.0107247712951)*x_ref[2]
        ref[3,0,0,0]=(1.23545873872)*x_ref[0]+(-1.29896744411)*x_ref[1]+(0.855094335492)*x_ref[2]
        ref[3,0,0,1]=(-0.51465695324)*x_ref[0]+(1.47799693308)*x_ref[1]+(0.0718360423316)*x_ref[2]
        ref[3,0,1,0]=(0.220505442402)*x_ref[0]+(-0.3171385194)*x_ref[1]+(0.27954628665)*x_ref[2]
        ref[3,0,1,1]=(0.860582046972)*x_ref[0]+(1.04565549687)*x_ref[1]+(1.18108647078)*x_ref[2]
        ref[3,0,2,0]=(-0.593547630941)*x_ref[0]+(-0.0390465590147)*x_ref[1]+(0.00129707682534)*x_ref[2]
        ref[3,0,2,1]=(1.39210293617)*x_ref[0]+(-1.05660957806)*x_ref[1]+(-1.20489109051)*x_ref[2]
        ref[3,1,0,0]=(-0.536035988946)*x_ref[0]+(-0.594403345465)*x_ref[1]+(1.11719111746)*x_ref[2]
        ref[3,1,0,1]=(-0.0403557954127)*x_ref[0]+(-1.32391025453)*x_ref[1]+(-0.639127329416)*x_ref[2]
        ref[3,1,1,0]=(-0.495803710666)*x_ref[0]+(-0.199832989186)*x_ref[1]+(-0.832862076561)*x_ref[2]
        ref[3,1,1,1]=(0.0216882402803)*x_ref[0]+(0.367268873436)*x_ref[1]+(0.295944335418)*x_ref[2]
        ref[3,1,2,0]=(0.172936506269)*x_ref[0]+(-0.339886198196)*x_ref[1]+(0.268201481244)*x_ref[2]
        ref[3,1,2,1]=(0.665786569897)*x_ref[0]+(-0.290637226736)*x_ref[1]+(0.617677178537)*x_ref[2]
        ref[3,2,0,0]=(-0.657902185911)*x_ref[0]+(-1.34292379802)*x_ref[1]+(0.0641950819134)*x_ref[2]
        ref[3,2,0,1]=(0.352039175923)*x_ref[0]+(-0.483047873735)*x_ref[1]+(0.47997600743)*x_ref[2]
        ref[3,2,1,0]=(0.944619484106)*x_ref[0]+(-0.308755547249)*x_ref[1]+(-0.765906593989)*x_ref[2]
        ref[3,2,1,1]=(-1.09674781865)*x_ref[0]+(0.013069257236)*x_ref[1]+(-0.0741247275628)*x_ref[2]
        ref[3,2,2,0]=(0.400057737624)*x_ref[0]+(-1.16165281514)*x_ref[1]+(-0.789757454741)*x_ref[2]
        ref[3,2,2,1]=(0.367100312493)*x_ref[0]+(-0.547338908792)*x_ref[1]+(-0.409173859392)*x_ref[2]
        ref[3,3,0,0]=(0.890118369638)*x_ref[0]+(1.201296587)*x_ref[1]+(-0.0319246476545)*x_ref[2]
        ref[3,3,0,1]=(0.620955702501)*x_ref[0]+(0.438440818114)*x_ref[1]+(0.499747510411)*x_ref[2]
        ref[3,3,1,0]=(1.61217537222)*x_ref[0]+(-1.21323714148)*x_ref[1]+(-1.06137097571)*x_ref[2]
        ref[3,3,1,1]=(-1.75942413542)*x_ref[0]+(0.440345678738)*x_ref[1]+(0.625299774096)*x_ref[2]
        ref[3,3,2,0]=(-0.878497266865)*x_ref[0]+(-1.15879244679)*x_ref[1]+(1.03993124425)*x_ref[2]
        ref[3,3,2,1]=(-0.0138445968491)*x_ref[0]+(0.547910692711)*x_ref[1]+(0.971642264613)*x_ref[2]
        ref[3,4,0,0]=(-0.107769272336)*x_ref[0]+(-0.987911384393)*x_ref[1]+(1.2076201556)*x_ref[2]
        ref[3,4,0,1]=(-0.556570952071)*x_ref[0]+(-0.284568452888)*x_ref[1]+(1.23341917108)*x_ref[2]
        ref[3,4,1,0]=(1.56461193947)*x_ref[0]+(0.185382898933)*x_ref[1]+(-1.67714180845)*x_ref[2]
        ref[3,4,1,1]=(-0.894934034676)*x_ref[0]+(1.05560233139)*x_ref[1]+(0.651029150593)*x_ref[2]
        ref[3,4,2,0]=(0.0559603779066)*x_ref[0]+(-1.72752549101)*x_ref[1]+(-1.02017554749)*x_ref[2]
        ref[3,4,2,1]=(-1.75424333815)*x_ref[0]+(-0.198869840282)*x_ref[1]+(0.475065396299)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-1.05153531037)*x[0]+(1.5374190245)*x[1]
        ref=(-1.05153531037)*x_ref[0]+(1.5374190245)*x_ref[1]
      else:
        s=(-1.91883881043)*x[0]+(-0.414033171764)*x[1]+(1.96173093572)*x[2]
        ref=(-1.91883881043)*x_ref[0]+(-0.414033171764)*x_ref[1]+(1.96173093572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        s[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
        ref[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        ref[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
      else:
        s[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        s[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
        ref[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        ref[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.581359229868)*x[0]+(-0.405627827004)*x[1]
        s[0,1]=(1.15883572458)*x[0]+(-0.212383708658)*x[1]
        s[0,2]=(-0.675196965148)*x[0]+(-0.0589289579115)*x[1]
        s[0,3]=(-0.466142003406)*x[0]+(1.143298366)*x[1]
        s[0,4]=(0.706585901476)*x[0]+(-0.596587020578)*x[1]
        s[1,0]=(0.150862770285)*x[0]+(1.3486848087)*x[1]
        s[1,1]=(0.694050128425)*x[0]+(0.919192282198)*x[1]
        s[1,2]=(0.123568652464)*x[0]+(-1.1901501926)*x[1]
        s[1,3]=(-0.0147792668086)*x[0]+(-1.49381121537)*x[1]
        s[1,4]=(0.0737831649537)*x[0]+(-0.752053730242)*x[1]
        s[2,0]=(-0.758812518739)*x[0]+(-1.26143124516)*x[1]
        s[2,1]=(0.507459828891)*x[0]+(-0.0315665691912)*x[1]
        s[2,2]=(-0.492645010187)*x[0]+(-0.649596364388)*x[1]
        s[2,3]=(-0.0632358844144)*x[0]+(0.899340082527)*x[1]
        s[2,4]=(-0.763180407153)*x[0]+(-0.269808126268)*x[1]
        s[3,0]=(-1.5724790938)*x[0]+(1.16942939992)*x[1]
        s[3,1]=(0.759317295371)*x[0]+(0.14287823082)*x[1]
        s[3,2]=(0.70712973662)*x[0]+(1.08455803224)*x[1]
        s[3,3]=(-0.736730602555)*x[0]+(-0.183785007602)*x[1]
        s[3,4]=(-0.950538745165)*x[0]+(0.294140958329)*x[1]
        ref[0,0]=(-0.581359229868)*x_ref[0]+(-0.405627827004)*x_ref[1]
        ref[0,1]=(1.15883572458)*x_ref[0]+(-0.212383708658)*x_ref[1]
        ref[0,2]=(-0.675196965148)*x_ref[0]+(-0.0589289579115)*x_ref[1]
        ref[0,3]=(-0.466142003406)*x_ref[0]+(1.143298366)*x_ref[1]
        ref[0,4]=(0.706585901476)*x_ref[0]+(-0.596587020578)*x_ref[1]
        ref[1,0]=(0.150862770285)*x_ref[0]+(1.3486848087)*x_ref[1]
        ref[1,1]=(0.694050128425)*x_ref[0]+(0.919192282198)*x_ref[1]
        ref[1,2]=(0.123568652464)*x_ref[0]+(-1.1901501926)*x_ref[1]
        ref[1,3]=(-0.0147792668086)*x_ref[0]+(-1.49381121537)*x_ref[1]
        ref[1,4]=(0.0737831649537)*x_ref[0]+(-0.752053730242)*x_ref[1]
        ref[2,0]=(-0.758812518739)*x_ref[0]+(-1.26143124516)*x_ref[1]
        ref[2,1]=(0.507459828891)*x_ref[0]+(-0.0315665691912)*x_ref[1]
        ref[2,2]=(-0.492645010187)*x_ref[0]+(-0.649596364388)*x_ref[1]
        ref[2,3]=(-0.0632358844144)*x_ref[0]+(0.899340082527)*x_ref[1]
        ref[2,4]=(-0.763180407153)*x_ref[0]+(-0.269808126268)*x_ref[1]
        ref[3,0]=(-1.5724790938)*x_ref[0]+(1.16942939992)*x_ref[1]
        ref[3,1]=(0.759317295371)*x_ref[0]+(0.14287823082)*x_ref[1]
        ref[3,2]=(0.70712973662)*x_ref[0]+(1.08455803224)*x_ref[1]
        ref[3,3]=(-0.736730602555)*x_ref[0]+(-0.183785007602)*x_ref[1]
        ref[3,4]=(-0.950538745165)*x_ref[0]+(0.294140958329)*x_ref[1]
      else:
        s[0,0]=(0.310668873066)*x[0]+(1.66326709742)*x[1]+(1.10140064967)*x[2]
        s[0,1]=(0.46962350981)*x[0]+(-1.02292998305)*x[1]+(0.919666772954)*x[2]
        s[0,2]=(-1.18114779319)*x[0]+(1.23819669136)*x[1]+(0.0622519749551)*x[2]
        s[0,3]=(1.08356504043)*x[0]+(0.450517408398)*x[1]+(-0.528809988127)*x[2]
        s[0,4]=(-0.179349520643)*x[0]+(0.192827326205)*x[1]+(-0.543850326813)*x[2]
        s[1,0]=(-1.60006459477)*x[0]+(0.800497908511)*x[1]+(0.745234766014)*x[2]
        s[1,1]=(0.667355662535)*x[0]+(0.800303765538)*x[1]+(-1.17397311427)*x[2]
        s[1,2]=(0.19342327722)*x[0]+(-0.149618364398)*x[1]+(-1.06968763194)*x[2]
        s[1,3]=(0.296694301988)*x[0]+(-1.35691147932)*x[1]+(0.748623252633)*x[2]
        s[1,4]=(0.724332288795)*x[0]+(0.349026497158)*x[1]+(0.377755550832)*x[2]
        s[2,0]=(0.560218327038)*x[0]+(0.184262569787)*x[1]+(-0.821718588725)*x[2]
        s[2,1]=(0.421969037007)*x[0]+(-0.00510504728331)*x[1]+(1.43225098227)*x[2]
        s[2,2]=(0.889590883145)*x[0]+(-0.408365128114)*x[1]+(-1.80100224461)*x[2]
        s[2,3]=(-1.91735498201)*x[0]+(0.590584386339)*x[1]+(-0.905610066725)*x[2]
        s[2,4]=(-0.367106108866)*x[0]+(-0.92493020573)*x[1]+(-0.0746199487091)*x[2]
        s[3,0]=(0.919247371049)*x[0]+(0.326674777975)*x[1]+(-1.47385875396)*x[2]
        s[3,1]=(-0.850835329183)*x[0]+(-0.536914325383)*x[1]+(1.02264919467)*x[2]
        s[3,2]=(-1.41905423752)*x[0]+(-0.632415273773)*x[1]+(0.842187325237)*x[2]
        s[3,3]=(-0.396129693721)*x[0]+(-0.110878220479)*x[1]+(0.49355539737)*x[2]
        s[3,4]=(-0.0781905590856)*x[0]+(-0.0348772309292)*x[1]+(0.184581658536)*x[2]
        ref[0,0]=(0.310668873066)*x_ref[0]+(1.66326709742)*x_ref[1]+(1.10140064967)*x_ref[2]
        ref[0,1]=(0.46962350981)*x_ref[0]+(-1.02292998305)*x_ref[1]+(0.919666772954)*x_ref[2]
        ref[0,2]=(-1.18114779319)*x_ref[0]+(1.23819669136)*x_ref[1]+(0.0622519749551)*x_ref[2]
        ref[0,3]=(1.08356504043)*x_ref[0]+(0.450517408398)*x_ref[1]+(-0.528809988127)*x_ref[2]
        ref[0,4]=(-0.179349520643)*x_ref[0]+(0.192827326205)*x_ref[1]+(-0.543850326813)*x_ref[2]
        ref[1,0]=(-1.60006459477)*x_ref[0]+(0.800497908511)*x_ref[1]+(0.745234766014)*x_ref[2]
        ref[1,1]=(0.667355662535)*x_ref[0]+(0.800303765538)*x_ref[1]+(-1.17397311427)*x_ref[2]
        ref[1,2]=(0.19342327722)*x_ref[0]+(-0.149618364398)*x_ref[1]+(-1.06968763194)*x_ref[2]
        ref[1,3]=(0.296694301988)*x_ref[0]+(-1.35691147932)*x_ref[1]+(0.748623252633)*x_ref[2]
        ref[1,4]=(0.724332288795)*x_ref[0]+(0.349026497158)*x_ref[1]+(0.377755550832)*x_ref[2]
        ref[2,0]=(0.560218327038)*x_ref[0]+(0.184262569787)*x_ref[1]+(-0.821718588725)*x_ref[2]
        ref[2,1]=(0.421969037007)*x_ref[0]+(-0.00510504728331)*x_ref[1]+(1.43225098227)*x_ref[2]
        ref[2,2]=(0.889590883145)*x_ref[0]+(-0.408365128114)*x_ref[1]+(-1.80100224461)*x_ref[2]
        ref[2,3]=(-1.91735498201)*x_ref[0]+(0.590584386339)*x_ref[1]+(-0.905610066725)*x_ref[2]
        ref[2,4]=(-0.367106108866)*x_ref[0]+(-0.92493020573)*x_ref[1]+(-0.0746199487091)*x_ref[2]
        ref[3,0]=(0.919247371049)*x_ref[0]+(0.326674777975)*x_ref[1]+(-1.47385875396)*x_ref[2]
        ref[3,1]=(-0.850835329183)*x_ref[0]+(-0.536914325383)*x_ref[1]+(1.02264919467)*x_ref[2]
        ref[3,2]=(-1.41905423752)*x_ref[0]+(-0.632415273773)*x_ref[1]+(0.842187325237)*x_ref[2]
        ref[3,3]=(-0.396129693721)*x_ref[0]+(-0.110878220479)*x_ref[1]+(0.49355539737)*x_ref[2]
        ref[3,4]=(-0.0781905590856)*x_ref[0]+(-0.0348772309292)*x_ref[1]+(0.184581658536)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.540497015605)*x[0]+(0.282753164584)*x[1]
        s[0,0,1]=(-0.127429073679)*x[0]+(-0.709993735868)*x[1]
        s[0,1,0]=(1.02019673589)*x[0]+(-1.53422031862)*x[1]
        s[0,1,1]=(0.926292918083)*x[0]+(-0.0351188336657)*x[1]
        s[1,0,0]=(-0.736080207248)*x[0]+(1.25260396336)*x[1]
        s[1,0,1]=(0.152653303192)*x[0]+(0.22547444598)*x[1]
        s[1,1,0]=(0.729526332632)*x[0]+(0.919207155406)*x[1]
        s[1,1,1]=(-0.346325378686)*x[0]+(0.141030274609)*x[1]
        s[2,0,0]=(-0.753354302966)*x[0]+(-0.766294422747)*x[1]
        s[2,0,1]=(1.17251569201)*x[0]+(-0.51673325599)*x[1]
        s[2,1,0]=(0.66192662328)*x[0]+(-0.572345218315)*x[1]
        s[2,1,1]=(1.74689126451)*x[0]+(-0.0537684141408)*x[1]
        s[3,0,0]=(0.300593672769)*x[0]+(0.0833505471404)*x[1]
        s[3,0,1]=(0.770671590417)*x[0]+(-0.667020549164)*x[1]
        s[3,1,0]=(0.281444608616)*x[0]+(0.701762037742)*x[1]
        s[3,1,1]=(0.404631442844)*x[0]+(0.337744807226)*x[1]
        s[4,0,0]=(-1.43357362276)*x[0]+(0.687631941644)*x[1]
        s[4,0,1]=(0.694571843151)*x[0]+(0.979495014797)*x[1]
        s[4,1,0]=(-0.0880858702517)*x[0]+(0.729298516002)*x[1]
        s[4,1,1]=(0.641170415423)*x[0]+(-0.407278118417)*x[1]
        s[5,0,0]=(-0.60364366976)*x[0]+(-0.663798735078)*x[1]
        s[5,0,1]=(-0.841035861641)*x[0]+(-1.09892876031)*x[1]
        s[5,1,0]=(0.0838180897314)*x[0]+(0.15765944496)*x[1]
        s[5,1,1]=(-0.847291245273)*x[0]+(1.01145652383)*x[1]
        ref[0,0,0]=(0.540497015605)*x_ref[0]+(0.282753164584)*x_ref[1]
        ref[0,0,1]=(-0.127429073679)*x_ref[0]+(-0.709993735868)*x_ref[1]
        ref[0,1,0]=(1.02019673589)*x_ref[0]+(-1.53422031862)*x_ref[1]
        ref[0,1,1]=(0.926292918083)*x_ref[0]+(-0.0351188336657)*x_ref[1]
        ref[1,0,0]=(-0.736080207248)*x_ref[0]+(1.25260396336)*x_ref[1]
        ref[1,0,1]=(0.152653303192)*x_ref[0]+(0.22547444598)*x_ref[1]
        ref[1,1,0]=(0.729526332632)*x_ref[0]+(0.919207155406)*x_ref[1]
        ref[1,1,1]=(-0.346325378686)*x_ref[0]+(0.141030274609)*x_ref[1]
        ref[2,0,0]=(-0.753354302966)*x_ref[0]+(-0.766294422747)*x_ref[1]
        ref[2,0,1]=(1.17251569201)*x_ref[0]+(-0.51673325599)*x_ref[1]
        ref[2,1,0]=(0.66192662328)*x_ref[0]+(-0.572345218315)*x_ref[1]
        ref[2,1,1]=(1.74689126451)*x_ref[0]+(-0.0537684141408)*x_ref[1]
        ref[3,0,0]=(0.300593672769)*x_ref[0]+(0.0833505471404)*x_ref[1]
        ref[3,0,1]=(0.770671590417)*x_ref[0]+(-0.667020549164)*x_ref[1]
        ref[3,1,0]=(0.281444608616)*x_ref[0]+(0.701762037742)*x_ref[1]
        ref[3,1,1]=(0.404631442844)*x_ref[0]+(0.337744807226)*x_ref[1]
        ref[4,0,0]=(-1.43357362276)*x_ref[0]+(0.687631941644)*x_ref[1]
        ref[4,0,1]=(0.694571843151)*x_ref[0]+(0.979495014797)*x_ref[1]
        ref[4,1,0]=(-0.0880858702517)*x_ref[0]+(0.729298516002)*x_ref[1]
        ref[4,1,1]=(0.641170415423)*x_ref[0]+(-0.407278118417)*x_ref[1]
        ref[5,0,0]=(-0.60364366976)*x_ref[0]+(-0.663798735078)*x_ref[1]
        ref[5,0,1]=(-0.841035861641)*x_ref[0]+(-1.09892876031)*x_ref[1]
        ref[5,1,0]=(0.0838180897314)*x_ref[0]+(0.15765944496)*x_ref[1]
        ref[5,1,1]=(-0.847291245273)*x_ref[0]+(1.01145652383)*x_ref[1]
      else:
        s[0,0,0]=(-0.29326741782)*x[0]+(0.716468869525)*x[1]+(-0.194120207009)*x[2]
        s[0,0,1]=(-0.0687304327556)*x[0]+(0.181789058707)*x[1]+(-0.940504268245)*x[2]
        s[0,1,0]=(0.437145055905)*x[0]+(1.72039421904)*x[1]+(-0.0326204183567)*x[2]
        s[0,1,1]=(1.25919532639)*x[0]+(0.082561920762)*x[1]+(1.01503353491)*x[2]
        s[1,0,0]=(-0.552129852247)*x[0]+(-0.661963725391)*x[1]+(0.191012614278)*x[2]
        s[1,0,1]=(-0.322228468989)*x[0]+(-0.79555280135)*x[1]+(0.0689420056626)*x[2]
        s[1,1,0]=(0.294413007041)*x[0]+(-0.453661316564)*x[1]+(1.74072519115)*x[2]
        s[1,1,1]=(-0.13491214165)*x[0]+(-1.18907003959)*x[1]+(-1.59325361447)*x[2]
        s[2,0,0]=(-0.522689030297)*x[0]+(-0.784658318696)*x[1]+(0.618343699514)*x[2]
        s[2,0,1]=(-0.227214629419)*x[0]+(0.770163719023)*x[1]+(-0.272804476799)*x[2]
        s[2,1,0]=(-1.28131384809)*x[0]+(0.0591076357001)*x[1]+(-0.206570950984)*x[2]
        s[2,1,1]=(0.00241150481435)*x[0]+(0.0678794618244)*x[1]+(-0.272677008307)*x[2]
        s[3,0,0]=(-0.537194244467)*x[0]+(-0.886303700791)*x[1]+(0.704423993884)*x[2]
        s[3,0,1]=(0.0924777549448)*x[0]+(-0.783168034362)*x[1]+(-0.450930098538)*x[2]
        s[3,1,0]=(0.611562862859)*x[0]+(-1.04224402701)*x[1]+(-0.775594008735)*x[2]
        s[3,1,1]=(1.44402445157)*x[0]+(-0.736290409307)*x[1]+(0.51690502447)*x[2]
        s[4,0,0]=(-0.470210375338)*x[0]+(-0.433261543227)*x[1]+(-0.582080162476)*x[2]
        s[4,0,1]=(0.313323330562)*x[0]+(-1.47502042219)*x[1]+(-1.50202564103)*x[2]
        s[4,1,0]=(-0.807949755771)*x[0]+(-0.162904810555)*x[1]+(0.748809317029)*x[2]
        s[4,1,1]=(-1.24228935643)*x[0]+(0.193414583713)*x[1]+(0.85984063481)*x[2]
        s[5,0,0]=(1.08068739991)*x[0]+(-1.30655986333)*x[1]+(0.289800143967)*x[2]
        s[5,0,1]=(1.12114019122)*x[0]+(0.280561086566)*x[1]+(0.844991094994)*x[2]
        s[5,1,0]=(-0.133922711479)*x[0]+(-1.39747129717)*x[1]+(-1.11382363437)*x[2]
        s[5,1,1]=(-1.34164423823)*x[0]+(-0.335643322772)*x[1]+(0.757067990421)*x[2]
        ref[0,0,0]=(-0.29326741782)*x_ref[0]+(0.716468869525)*x_ref[1]+(-0.194120207009)*x_ref[2]
        ref[0,0,1]=(-0.0687304327556)*x_ref[0]+(0.181789058707)*x_ref[1]+(-0.940504268245)*x_ref[2]
        ref[0,1,0]=(0.437145055905)*x_ref[0]+(1.72039421904)*x_ref[1]+(-0.0326204183567)*x_ref[2]
        ref[0,1,1]=(1.25919532639)*x_ref[0]+(0.082561920762)*x_ref[1]+(1.01503353491)*x_ref[2]
        ref[1,0,0]=(-0.552129852247)*x_ref[0]+(-0.661963725391)*x_ref[1]+(0.191012614278)*x_ref[2]
        ref[1,0,1]=(-0.322228468989)*x_ref[0]+(-0.79555280135)*x_ref[1]+(0.0689420056626)*x_ref[2]
        ref[1,1,0]=(0.294413007041)*x_ref[0]+(-0.453661316564)*x_ref[1]+(1.74072519115)*x_ref[2]
        ref[1,1,1]=(-0.13491214165)*x_ref[0]+(-1.18907003959)*x_ref[1]+(-1.59325361447)*x_ref[2]
        ref[2,0,0]=(-0.522689030297)*x_ref[0]+(-0.784658318696)*x_ref[1]+(0.618343699514)*x_ref[2]
        ref[2,0,1]=(-0.227214629419)*x_ref[0]+(0.770163719023)*x_ref[1]+(-0.272804476799)*x_ref[2]
        ref[2,1,0]=(-1.28131384809)*x_ref[0]+(0.0591076357001)*x_ref[1]+(-0.206570950984)*x_ref[2]
        ref[2,1,1]=(0.00241150481435)*x_ref[0]+(0.0678794618244)*x_ref[1]+(-0.272677008307)*x_ref[2]
        ref[3,0,0]=(-0.537194244467)*x_ref[0]+(-0.886303700791)*x_ref[1]+(0.704423993884)*x_ref[2]
        ref[3,0,1]=(0.0924777549448)*x_ref[0]+(-0.783168034362)*x_ref[1]+(-0.450930098538)*x_ref[2]
        ref[3,1,0]=(0.611562862859)*x_ref[0]+(-1.04224402701)*x_ref[1]+(-0.775594008735)*x_ref[2]
        ref[3,1,1]=(1.44402445157)*x_ref[0]+(-0.736290409307)*x_ref[1]+(0.51690502447)*x_ref[2]
        ref[4,0,0]=(-0.470210375338)*x_ref[0]+(-0.433261543227)*x_ref[1]+(-0.582080162476)*x_ref[2]
        ref[4,0,1]=(0.313323330562)*x_ref[0]+(-1.47502042219)*x_ref[1]+(-1.50202564103)*x_ref[2]
        ref[4,1,0]=(-0.807949755771)*x_ref[0]+(-0.162904810555)*x_ref[1]+(0.748809317029)*x_ref[2]
        ref[4,1,1]=(-1.24228935643)*x_ref[0]+(0.193414583713)*x_ref[1]+(0.85984063481)*x_ref[2]
        ref[5,0,0]=(1.08068739991)*x_ref[0]+(-1.30655986333)*x_ref[1]+(0.289800143967)*x_ref[2]
        ref[5,0,1]=(1.12114019122)*x_ref[0]+(0.280561086566)*x_ref[1]+(0.844991094994)*x_ref[2]
        ref[5,1,0]=(-0.133922711479)*x_ref[0]+(-1.39747129717)*x_ref[1]+(-1.11382363437)*x_ref[2]
        ref[5,1,1]=(-1.34164423823)*x_ref[0]+(-0.335643322772)*x_ref[1]+(0.757067990421)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedSolution_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedSolution(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.9809347951)*x[0]+(-1.05666590015)*x[1]
        s[0,0,0,1]=(1.15925248094)*x[0]+(1.22070923582)*x[1]
        s[0,0,1,0]=(-0.0724361628927)*x[0]+(-0.439807238916)*x[1]
        s[0,0,1,1]=(-0.608816713414)*x[0]+(1.78792604745)*x[1]
        s[0,0,2,0]=(-0.78903506363)*x[0]+(-0.75896072902)*x[1]
        s[0,0,2,1]=(0.516424260522)*x[0]+(1.16788015378)*x[1]
        s[0,1,0,0]=(0.383586654748)*x[0]+(1.4179924146)*x[1]
        s[0,1,0,1]=(0.97236907624)*x[0]+(-0.438982550276)*x[1]
        s[0,1,1,0]=(-0.645083297783)*x[0]+(-0.740808997218)*x[1]
        s[0,1,1,1]=(0.330652781935)*x[0]+(0.519132367114)*x[1]
        s[0,1,2,0]=(1.25159174261)*x[0]+(1.26021904904)*x[1]
        s[0,1,2,1]=(-0.239621293462)*x[0]+(0.0244107517523)*x[1]
        s[0,2,0,0]=(-0.286208240697)*x[0]+(-0.314157258671)*x[1]
        s[0,2,0,1]=(-0.0856319829397)*x[0]+(0.278603809503)*x[1]
        s[0,2,1,0]=(0.107199217139)*x[0]+(0.0948664401127)*x[1]
        s[0,2,1,1]=(-0.24620060163)*x[0]+(-1.42942082558)*x[1]
        s[0,2,2,0]=(0.845285367352)*x[0]+(0.408676159609)*x[1]
        s[0,2,2,1]=(-1.49467460394)*x[0]+(0.270722457265)*x[1]
        s[0,3,0,0]=(0.469802598061)*x[0]+(-0.801790345348)*x[1]
        s[0,3,0,1]=(1.71249257903)*x[0]+(-0.776797378881)*x[1]
        s[0,3,1,0]=(0.0540049036319)*x[0]+(1.19031969164)*x[1]
        s[0,3,1,1]=(0.161929130968)*x[0]+(-0.525381566632)*x[1]
        s[0,3,2,0]=(-0.646041965665)*x[0]+(-0.123993555032)*x[1]
        s[0,3,2,1]=(0.0966044205198)*x[0]+(1.24780552685)*x[1]
        s[0,4,0,0]=(-0.289531041539)*x[0]+(1.21117470928)*x[1]
        s[0,4,0,1]=(-0.112215925799)*x[0]+(0.598717725332)*x[1]
        s[0,4,1,0]=(0.162936056701)*x[0]+(-1.02395851472)*x[1]
        s[0,4,1,1]=(0.694066364547)*x[0]+(1.27205020727)*x[1]
        s[0,4,2,0]=(1.17253981195)*x[0]+(0.130572791304)*x[1]
        s[0,4,2,1]=(-0.555241229093)*x[0]+(-0.881115061403)*x[1]
        s[1,0,0,0]=(0.574137047465)*x[0]+(-1.10459006911)*x[1]
        s[1,0,0,1]=(0.441431076852)*x[0]+(-0.133156832361)*x[1]
        s[1,0,1,0]=(0.217136655494)*x[0]+(0.400834952015)*x[1]
        s[1,0,1,1]=(0.516360280675)*x[0]+(0.532907430046)*x[1]
        s[1,0,2,0]=(-0.340764752946)*x[0]+(-1.26671004418)*x[1]
        s[1,0,2,1]=(-1.23231139783)*x[0]+(0.61934930883)*x[1]
        s[1,1,0,0]=(1.76821949014)*x[0]+(1.67542625016)*x[1]
        s[1,1,0,1]=(0.901675970272)*x[0]+(-0.0567109697208)*x[1]
        s[1,1,1,0]=(1.21143551644)*x[0]+(0.375877605521)*x[1]
        s[1,1,1,1]=(-0.699830133987)*x[0]+(0.174660968691)*x[1]
        s[1,1,2,0]=(1.07839763024)*x[0]+(-0.063554858486)*x[1]
        s[1,1,2,1]=(0.131335396282)*x[0]+(0.109893568063)*x[1]
        s[1,2,0,0]=(-0.562355685272)*x[0]+(0.495744936869)*x[1]
        s[1,2,0,1]=(0.0747584188968)*x[0]+(1.94924356864)*x[1]
        s[1,2,1,0]=(-0.103515803094)*x[0]+(0.545380886507)*x[1]
        s[1,2,1,1]=(-0.708926705928)*x[0]+(-0.0610123075525)*x[1]
        s[1,2,2,0]=(1.06802154614)*x[0]+(0.0622475519235)*x[1]
        s[1,2,2,1]=(0.103937010427)*x[0]+(-0.70201008693)*x[1]
        s[1,3,0,0]=(0.0730726824626)*x[0]+(0.39791188789)*x[1]
        s[1,3,0,1]=(0.455273194688)*x[0]+(1.65845657261)*x[1]
        s[1,3,1,0]=(0.0807767752424)*x[0]+(-0.765026395125)*x[1]
        s[1,3,1,1]=(0.0404890748976)*x[0]+(-0.383198358048)*x[1]
        s[1,3,2,0]=(1.26048751933)*x[0]+(0.00329382174075)*x[1]
        s[1,3,2,1]=(-1.21518062161)*x[0]+(-0.325769725861)*x[1]
        s[1,4,0,0]=(0.682181651323)*x[0]+(-0.375432356672)*x[1]
        s[1,4,0,1]=(-1.09892909715)*x[0]+(0.207984305416)*x[1]
        s[1,4,1,0]=(-1.2375595164)*x[0]+(0.215243019189)*x[1]
        s[1,4,1,1]=(-0.595125980205)*x[0]+(0.29104395616)*x[1]
        s[1,4,2,0]=(0.127947393316)*x[0]+(1.43408039596)*x[1]
        s[1,4,2,1]=(0.468574871016)*x[0]+(0.333319611281)*x[1]
        s[2,0,0,0]=(0.369954209324)*x[0]+(-1.09614147452)*x[1]
        s[2,0,0,1]=(-0.120003056441)*x[0]+(-0.172964298439)*x[1]
        s[2,0,1,0]=(1.38230317911)*x[0]+(1.25730140777)*x[1]
        s[2,0,1,1]=(1.10857329622)*x[0]+(0.449561992049)*x[1]
        s[2,0,2,0]=(-1.48479939533)*x[0]+(-0.839474409588)*x[1]
        s[2,0,2,1]=(-0.662177361426)*x[0]+(0.380170348062)*x[1]
        s[2,1,0,0]=(1.13771311046)*x[0]+(-1.59960740614)*x[1]
        s[2,1,0,1]=(0.925401031957)*x[0]+(1.30170041054)*x[1]
        s[2,1,1,0]=(0.455911230864)*x[0]+(0.755253828053)*x[1]
        s[2,1,1,1]=(0.64783771432)*x[0]+(0.251664567771)*x[1]
        s[2,1,2,0]=(-0.0449523798045)*x[0]+(-0.614423607357)*x[1]
        s[2,1,2,1]=(0.635435155497)*x[0]+(0.218411499589)*x[1]
        s[2,2,0,0]=(1.07064943699)*x[0]+(1.33695839858)*x[1]
        s[2,2,0,1]=(0.286067300461)*x[0]+(-1.19968009538)*x[1]
        s[2,2,1,0]=(0.373936537767)*x[0]+(-0.684346831863)*x[1]
        s[2,2,1,1]=(0.709764408449)*x[0]+(0.560181368149)*x[1]
        s[2,2,2,0]=(0.687217025484)*x[0]+(0.638368794102)*x[1]
        s[2,2,2,1]=(-1.13694076648)*x[0]+(1.31012182384)*x[1]
        s[2,3,0,0]=(0.784452969448)*x[0]+(-0.256804074025)*x[1]
        s[2,3,0,1]=(-0.528533290176)*x[0]+(0.230293895766)*x[1]
        s[2,3,1,0]=(-0.152281766055)*x[0]+(-1.29007246581)*x[1]
        s[2,3,1,1]=(-0.113853891603)*x[0]+(-0.146975472116)*x[1]
        s[2,3,2,0]=(-1.16737851429)*x[0]+(1.18358630738)*x[1]
        s[2,3,2,1]=(-0.110242592809)*x[0]+(-0.228040285919)*x[1]
        s[2,4,0,0]=(-0.615024817893)*x[0]+(1.19310484943)*x[1]
        s[2,4,0,1]=(0.405843983847)*x[0]+(-0.7683655961)*x[1]
        s[2,4,1,0]=(-1.94063527375)*x[0]+(0.135437762233)*x[1]
        s[2,4,1,1]=(-0.680568187039)*x[0]+(0.411161107413)*x[1]
        s[2,4,2,0]=(-1.95575235842)*x[0]+(0.384766876765)*x[1]
        s[2,4,2,1]=(1.37236985627)*x[0]+(0.0082159494728)*x[1]
        s[3,0,0,0]=(1.63255667845)*x[0]+(0.0225121724188)*x[1]
        s[3,0,0,1]=(1.2376463155)*x[0]+(1.32594028087)*x[1]
        s[3,0,1,0]=(1.54455876996)*x[0]+(-1.21021289438)*x[1]
        s[3,0,1,1]=(-1.81162592117)*x[0]+(-0.297403340122)*x[1]
        s[3,0,2,0]=(0.806731759445)*x[0]+(-0.981072370255)*x[1]
        s[3,0,2,1]=(-0.15215786955)*x[0]+(0.163165607014)*x[1]
        s[3,1,0,0]=(0.233006578932)*x[0]+(-1.29832996415)*x[1]
        s[3,1,0,1]=(0.96755903915)*x[0]+(-0.00597412239186)*x[1]
        s[3,1,1,0]=(0.0413376466172)*x[0]+(1.26720441986)*x[1]
        s[3,1,1,1]=(0.676509459424)*x[0]+(1.2621989434)*x[1]
        s[3,1,2,0]=(-0.211034948146)*x[0]+(-1.37470768893)*x[1]
        s[3,1,2,1]=(-0.402208754333)*x[0]+(1.25258243878)*x[1]
        s[3,2,0,0]=(-0.529199129991)*x[0]+(-0.431490132335)*x[1]
        s[3,2,0,1]=(-0.850343939948)*x[0]+(0.0602529984167)*x[1]
        s[3,2,1,0]=(-0.420125360943)*x[0]+(1.24318875197)*x[1]
        s[3,2,1,1]=(0.753685328318)*x[0]+(0.354913692774)*x[1]
        s[3,2,2,0]=(-1.17147381854)*x[0]+(0.850516886779)*x[1]
        s[3,2,2,1]=(1.48587305579)*x[0]+(1.31835868193)*x[1]
        s[3,3,0,0]=(1.80808607564)*x[0]+(-0.124508319578)*x[1]
        s[3,3,0,1]=(-1.84429695094)*x[0]+(-0.254797649176)*x[1]
        s[3,3,1,0]=(1.17110770558)*x[0]+(0.89756258226)*x[1]
        s[3,3,1,1]=(0.692264407822)*x[0]+(-1.00105988426)*x[1]
        s[3,3,2,0]=(-0.505689548372)*x[0]+(0.365519850058)*x[1]
        s[3,3,2,1]=(0.651544272536)*x[0]+(1.09868410461)*x[1]
        s[3,4,0,0]=(0.155129169877)*x[0]+(0.4435929499)*x[1]
        s[3,4,0,1]=(1.65974703109)*x[0]+(-0.776830635917)*x[1]
        s[3,4,1,0]=(-1.05612271738)*x[0]+(1.63557694551)*x[1]
        s[3,4,1,1]=(0.69435492251)*x[0]+(-1.20394873527)*x[1]
        s[3,4,2,0]=(-0.237598987964)*x[0]+(0.404820539664)*x[1]
        s[3,4,2,1]=(-1.32483403281)*x[0]+(0.431217340618)*x[1]
        ref[0,0,0,0]=(-0.9809347951)*x_ref[0]+(-1.05666590015)*x_ref[1]
        ref[0,0,0,1]=(1.15925248094)*x_ref[0]+(1.22070923582)*x_ref[1]
        ref[0,0,1,0]=(-0.0724361628927)*x_ref[0]+(-0.439807238916)*x_ref[1]
        ref[0,0,1,1]=(-0.608816713414)*x_ref[0]+(1.78792604745)*x_ref[1]
        ref[0,0,2,0]=(-0.78903506363)*x_ref[0]+(-0.75896072902)*x_ref[1]
        ref[0,0,2,1]=(0.516424260522)*x_ref[0]+(1.16788015378)*x_ref[1]
        ref[0,1,0,0]=(0.383586654748)*x_ref[0]+(1.4179924146)*x_ref[1]
        ref[0,1,0,1]=(0.97236907624)*x_ref[0]+(-0.438982550276)*x_ref[1]
        ref[0,1,1,0]=(-0.645083297783)*x_ref[0]+(-0.740808997218)*x_ref[1]
        ref[0,1,1,1]=(0.330652781935)*x_ref[0]+(0.519132367114)*x_ref[1]
        ref[0,1,2,0]=(1.25159174261)*x_ref[0]+(1.26021904904)*x_ref[1]
        ref[0,1,2,1]=(-0.239621293462)*x_ref[0]+(0.0244107517523)*x_ref[1]
        ref[0,2,0,0]=(-0.286208240697)*x_ref[0]+(-0.314157258671)*x_ref[1]
        ref[0,2,0,1]=(-0.0856319829397)*x_ref[0]+(0.278603809503)*x_ref[1]
        ref[0,2,1,0]=(0.107199217139)*x_ref[0]+(0.0948664401127)*x_ref[1]
        ref[0,2,1,1]=(-0.24620060163)*x_ref[0]+(-1.42942082558)*x_ref[1]
        ref[0,2,2,0]=(0.845285367352)*x_ref[0]+(0.408676159609)*x_ref[1]
        ref[0,2,2,1]=(-1.49467460394)*x_ref[0]+(0.270722457265)*x_ref[1]
        ref[0,3,0,0]=(0.469802598061)*x_ref[0]+(-0.801790345348)*x_ref[1]
        ref[0,3,0,1]=(1.71249257903)*x_ref[0]+(-0.776797378881)*x_ref[1]
        ref[0,3,1,0]=(0.0540049036319)*x_ref[0]+(1.19031969164)*x_ref[1]
        ref[0,3,1,1]=(0.161929130968)*x_ref[0]+(-0.525381566632)*x_ref[1]
        ref[0,3,2,0]=(-0.646041965665)*x_ref[0]+(-0.123993555032)*x_ref[1]
        ref[0,3,2,1]=(0.0966044205198)*x_ref[0]+(1.24780552685)*x_ref[1]
        ref[0,4,0,0]=(-0.289531041539)*x_ref[0]+(1.21117470928)*x_ref[1]
        ref[0,4,0,1]=(-0.112215925799)*x_ref[0]+(0.598717725332)*x_ref[1]
        ref[0,4,1,0]=(0.162936056701)*x_ref[0]+(-1.02395851472)*x_ref[1]
        ref[0,4,1,1]=(0.694066364547)*x_ref[0]+(1.27205020727)*x_ref[1]
        ref[0,4,2,0]=(1.17253981195)*x_ref[0]+(0.130572791304)*x_ref[1]
        ref[0,4,2,1]=(-0.555241229093)*x_ref[0]+(-0.881115061403)*x_ref[1]
        ref[1,0,0,0]=(0.574137047465)*x_ref[0]+(-1.10459006911)*x_ref[1]
        ref[1,0,0,1]=(0.441431076852)*x_ref[0]+(-0.133156832361)*x_ref[1]
        ref[1,0,1,0]=(0.217136655494)*x_ref[0]+(0.400834952015)*x_ref[1]
        ref[1,0,1,1]=(0.516360280675)*x_ref[0]+(0.532907430046)*x_ref[1]
        ref[1,0,2,0]=(-0.340764752946)*x_ref[0]+(-1.26671004418)*x_ref[1]
        ref[1,0,2,1]=(-1.23231139783)*x_ref[0]+(0.61934930883)*x_ref[1]
        ref[1,1,0,0]=(1.76821949014)*x_ref[0]+(1.67542625016)*x_ref[1]
        ref[1,1,0,1]=(0.901675970272)*x_ref[0]+(-0.0567109697208)*x_ref[1]
        ref[1,1,1,0]=(1.21143551644)*x_ref[0]+(0.375877605521)*x_ref[1]
        ref[1,1,1,1]=(-0.699830133987)*x_ref[0]+(0.174660968691)*x_ref[1]
        ref[1,1,2,0]=(1.07839763024)*x_ref[0]+(-0.063554858486)*x_ref[1]
        ref[1,1,2,1]=(0.131335396282)*x_ref[0]+(0.109893568063)*x_ref[1]
        ref[1,2,0,0]=(-0.562355685272)*x_ref[0]+(0.495744936869)*x_ref[1]
        ref[1,2,0,1]=(0.0747584188968)*x_ref[0]+(1.94924356864)*x_ref[1]
        ref[1,2,1,0]=(-0.103515803094)*x_ref[0]+(0.545380886507)*x_ref[1]
        ref[1,2,1,1]=(-0.708926705928)*x_ref[0]+(-0.0610123075525)*x_ref[1]
        ref[1,2,2,0]=(1.06802154614)*x_ref[0]+(0.0622475519235)*x_ref[1]
        ref[1,2,2,1]=(0.103937010427)*x_ref[0]+(-0.70201008693)*x_ref[1]
        ref[1,3,0,0]=(0.0730726824626)*x_ref[0]+(0.39791188789)*x_ref[1]
        ref[1,3,0,1]=(0.455273194688)*x_ref[0]+(1.65845657261)*x_ref[1]
        ref[1,3,1,0]=(0.0807767752424)*x_ref[0]+(-0.765026395125)*x_ref[1]
        ref[1,3,1,1]=(0.0404890748976)*x_ref[0]+(-0.383198358048)*x_ref[1]
        ref[1,3,2,0]=(1.26048751933)*x_ref[0]+(0.00329382174075)*x_ref[1]
        ref[1,3,2,1]=(-1.21518062161)*x_ref[0]+(-0.325769725861)*x_ref[1]
        ref[1,4,0,0]=(0.682181651323)*x_ref[0]+(-0.375432356672)*x_ref[1]
        ref[1,4,0,1]=(-1.09892909715)*x_ref[0]+(0.207984305416)*x_ref[1]
        ref[1,4,1,0]=(-1.2375595164)*x_ref[0]+(0.215243019189)*x_ref[1]
        ref[1,4,1,1]=(-0.595125980205)*x_ref[0]+(0.29104395616)*x_ref[1]
        ref[1,4,2,0]=(0.127947393316)*x_ref[0]+(1.43408039596)*x_ref[1]
        ref[1,4,2,1]=(0.468574871016)*x_ref[0]+(0.333319611281)*x_ref[1]
        ref[2,0,0,0]=(0.369954209324)*x_ref[0]+(-1.09614147452)*x_ref[1]
        ref[2,0,0,1]=(-0.120003056441)*x_ref[0]+(-0.172964298439)*x_ref[1]
        ref[2,0,1,0]=(1.38230317911)*x_ref[0]+(1.25730140777)*x_ref[1]
        ref[2,0,1,1]=(1.10857329622)*x_ref[0]+(0.449561992049)*x_ref[1]
        ref[2,0,2,0]=(-1.48479939533)*x_ref[0]+(-0.839474409588)*x_ref[1]
        ref[2,0,2,1]=(-0.662177361426)*x_ref[0]+(0.380170348062)*x_ref[1]
        ref[2,1,0,0]=(1.13771311046)*x_ref[0]+(-1.59960740614)*x_ref[1]
        ref[2,1,0,1]=(0.925401031957)*x_ref[0]+(1.30170041054)*x_ref[1]
        ref[2,1,1,0]=(0.455911230864)*x_ref[0]+(0.755253828053)*x_ref[1]
        ref[2,1,1,1]=(0.64783771432)*x_ref[0]+(0.251664567771)*x_ref[1]
        ref[2,1,2,0]=(-0.0449523798045)*x_ref[0]+(-0.614423607357)*x_ref[1]
        ref[2,1,2,1]=(0.635435155497)*x_ref[0]+(0.218411499589)*x_ref[1]
        ref[2,2,0,0]=(1.07064943699)*x_ref[0]+(1.33695839858)*x_ref[1]
        ref[2,2,0,1]=(0.286067300461)*x_ref[0]+(-1.19968009538)*x_ref[1]
        ref[2,2,1,0]=(0.373936537767)*x_ref[0]+(-0.684346831863)*x_ref[1]
        ref[2,2,1,1]=(0.709764408449)*x_ref[0]+(0.560181368149)*x_ref[1]
        ref[2,2,2,0]=(0.687217025484)*x_ref[0]+(0.638368794102)*x_ref[1]
        ref[2,2,2,1]=(-1.13694076648)*x_ref[0]+(1.31012182384)*x_ref[1]
        ref[2,3,0,0]=(0.784452969448)*x_ref[0]+(-0.256804074025)*x_ref[1]
        ref[2,3,0,1]=(-0.528533290176)*x_ref[0]+(0.230293895766)*x_ref[1]
        ref[2,3,1,0]=(-0.152281766055)*x_ref[0]+(-1.29007246581)*x_ref[1]
        ref[2,3,1,1]=(-0.113853891603)*x_ref[0]+(-0.146975472116)*x_ref[1]
        ref[2,3,2,0]=(-1.16737851429)*x_ref[0]+(1.18358630738)*x_ref[1]
        ref[2,3,2,1]=(-0.110242592809)*x_ref[0]+(-0.228040285919)*x_ref[1]
        ref[2,4,0,0]=(-0.615024817893)*x_ref[0]+(1.19310484943)*x_ref[1]
        ref[2,4,0,1]=(0.405843983847)*x_ref[0]+(-0.7683655961)*x_ref[1]
        ref[2,4,1,0]=(-1.94063527375)*x_ref[0]+(0.135437762233)*x_ref[1]
        ref[2,4,1,1]=(-0.680568187039)*x_ref[0]+(0.411161107413)*x_ref[1]
        ref[2,4,2,0]=(-1.95575235842)*x_ref[0]+(0.384766876765)*x_ref[1]
        ref[2,4,2,1]=(1.37236985627)*x_ref[0]+(0.0082159494728)*x_ref[1]
        ref[3,0,0,0]=(1.63255667845)*x_ref[0]+(0.0225121724188)*x_ref[1]
        ref[3,0,0,1]=(1.2376463155)*x_ref[0]+(1.32594028087)*x_ref[1]
        ref[3,0,1,0]=(1.54455876996)*x_ref[0]+(-1.21021289438)*x_ref[1]
        ref[3,0,1,1]=(-1.81162592117)*x_ref[0]+(-0.297403340122)*x_ref[1]
        ref[3,0,2,0]=(0.806731759445)*x_ref[0]+(-0.981072370255)*x_ref[1]
        ref[3,0,2,1]=(-0.15215786955)*x_ref[0]+(0.163165607014)*x_ref[1]
        ref[3,1,0,0]=(0.233006578932)*x_ref[0]+(-1.29832996415)*x_ref[1]
        ref[3,1,0,1]=(0.96755903915)*x_ref[0]+(-0.00597412239186)*x_ref[1]
        ref[3,1,1,0]=(0.0413376466172)*x_ref[0]+(1.26720441986)*x_ref[1]
        ref[3,1,1,1]=(0.676509459424)*x_ref[0]+(1.2621989434)*x_ref[1]
        ref[3,1,2,0]=(-0.211034948146)*x_ref[0]+(-1.37470768893)*x_ref[1]
        ref[3,1,2,1]=(-0.402208754333)*x_ref[0]+(1.25258243878)*x_ref[1]
        ref[3,2,0,0]=(-0.529199129991)*x_ref[0]+(-0.431490132335)*x_ref[1]
        ref[3,2,0,1]=(-0.850343939948)*x_ref[0]+(0.0602529984167)*x_ref[1]
        ref[3,2,1,0]=(-0.420125360943)*x_ref[0]+(1.24318875197)*x_ref[1]
        ref[3,2,1,1]=(0.753685328318)*x_ref[0]+(0.354913692774)*x_ref[1]
        ref[3,2,2,0]=(-1.17147381854)*x_ref[0]+(0.850516886779)*x_ref[1]
        ref[3,2,2,1]=(1.48587305579)*x_ref[0]+(1.31835868193)*x_ref[1]
        ref[3,3,0,0]=(1.80808607564)*x_ref[0]+(-0.124508319578)*x_ref[1]
        ref[3,3,0,1]=(-1.84429695094)*x_ref[0]+(-0.254797649176)*x_ref[1]
        ref[3,3,1,0]=(1.17110770558)*x_ref[0]+(0.89756258226)*x_ref[1]
        ref[3,3,1,1]=(0.692264407822)*x_ref[0]+(-1.00105988426)*x_ref[1]
        ref[3,3,2,0]=(-0.505689548372)*x_ref[0]+(0.365519850058)*x_ref[1]
        ref[3,3,2,1]=(0.651544272536)*x_ref[0]+(1.09868410461)*x_ref[1]
        ref[3,4,0,0]=(0.155129169877)*x_ref[0]+(0.4435929499)*x_ref[1]
        ref[3,4,0,1]=(1.65974703109)*x_ref[0]+(-0.776830635917)*x_ref[1]
        ref[3,4,1,0]=(-1.05612271738)*x_ref[0]+(1.63557694551)*x_ref[1]
        ref[3,4,1,1]=(0.69435492251)*x_ref[0]+(-1.20394873527)*x_ref[1]
        ref[3,4,2,0]=(-0.237598987964)*x_ref[0]+(0.404820539664)*x_ref[1]
        ref[3,4,2,1]=(-1.32483403281)*x_ref[0]+(0.431217340618)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.270116752643)*x[0]+(0.0715809194071)*x[1]+(0.930011298681)*x[2]
        s[0,0,0,1]=(-0.0435807732016)*x[0]+(-0.492622221729)*x[1]+(-1.36754174971)*x[2]
        s[0,0,1,0]=(-0.949818131569)*x[0]+(1.08437418341)*x[1]+(0.527267447353)*x[2]
        s[0,0,1,1]=(-0.164876627537)*x[0]+(-1.44143896699)*x[1]+(0.0355140164081)*x[2]
        s[0,0,2,0]=(-0.530229192916)*x[0]+(-0.988412045132)*x[1]+(-0.466520393878)*x[2]
        s[0,0,2,1]=(-0.0166373496749)*x[0]+(0.322621663585)*x[1]+(-0.102019336059)*x[2]
        s[0,1,0,0]=(-1.36627178258)*x[0]+(-0.803463758822)*x[1]+(0.734446026661)*x[2]
        s[0,1,0,1]=(-0.804729478825)*x[0]+(-1.24406753288)*x[1]+(-0.755743791016)*x[2]
        s[0,1,1,0]=(1.03869027308)*x[0]+(-0.25237293575)*x[1]+(1.43737134487)*x[2]
        s[0,1,1,1]=(1.48269192251)*x[0]+(0.045590449115)*x[1]+(0.47425103841)*x[2]
        s[0,1,2,0]=(-0.738695287005)*x[0]+(1.01938431903)*x[1]+(1.20767534795)*x[2]
        s[0,1,2,1]=(-0.237445941287)*x[0]+(0.766593746786)*x[1]+(0.0297279831889)*x[2]
        s[0,2,0,0]=(-0.804902383375)*x[0]+(0.0962339257662)*x[1]+(1.37407473707)*x[2]
        s[0,2,0,1]=(-0.0821182376238)*x[0]+(0.35489327051)*x[1]+(-1.43235115632)*x[2]
        s[0,2,1,0]=(1.11593707013)*x[0]+(0.744039820943)*x[1]+(0.294815281064)*x[2]
        s[0,2,1,1]=(0.0789595478782)*x[0]+(0.719656781874)*x[1]+(0.355275528053)*x[2]
        s[0,2,2,0]=(1.20514986435)*x[0]+(0.203004753589)*x[1]+(-0.0731941650839)*x[2]
        s[0,2,2,1]=(-1.23231910051)*x[0]+(-0.470834142341)*x[1]+(-0.633518095593)*x[2]
        s[0,3,0,0]=(0.468164093965)*x[0]+(-1.15996958069)*x[1]+(0.922274695212)*x[2]
        s[0,3,0,1]=(-0.964250141036)*x[0]+(-0.769747467017)*x[1]+(-1.16369265956)*x[2]
        s[0,3,1,0]=(1.1622810944)*x[0]+(-1.72794997517)*x[1]+(-0.297833274612)*x[2]
        s[0,3,1,1]=(1.20239531447)*x[0]+(-0.195111187444)*x[1]+(-0.720186501147)*x[2]
        s[0,3,2,0]=(-0.0760001358746)*x[0]+(-1.65266437471)*x[1]+(-0.496163951015)*x[2]
        s[0,3,2,1]=(-1.4998248458)*x[0]+(-1.73305168566)*x[1]+(-0.107384086634)*x[2]
        s[0,4,0,0]=(0.235699357585)*x[0]+(-0.213348404403)*x[1]+(0.35368491512)*x[2]
        s[0,4,0,1]=(0.264641788778)*x[0]+(0.435251584452)*x[1]+(0.22539974851)*x[2]
        s[0,4,1,0]=(0.540203510472)*x[0]+(-1.33243180984)*x[1]+(-0.520357911324)*x[2]
        s[0,4,1,1]=(-0.629868492098)*x[0]+(-0.403402885662)*x[1]+(-0.184983842426)*x[2]
        s[0,4,2,0]=(1.17843494613)*x[0]+(0.562783141759)*x[1]+(-1.48575741726)*x[2]
        s[0,4,2,1]=(0.623796209289)*x[0]+(1.02627846638)*x[1]+(1.39369241754)*x[2]
        s[1,0,0,0]=(1.9855507901)*x[0]+(0.152976407288)*x[1]+(0.0931547545156)*x[2]
        s[1,0,0,1]=(-0.422055088151)*x[0]+(-0.509449435814)*x[1]+(-1.31705141661)*x[2]
        s[1,0,1,0]=(-0.24193816853)*x[0]+(-0.68450536388)*x[1]+(-0.309024887085)*x[2]
        s[1,0,1,1]=(1.24557169323)*x[0]+(0.353538565492)*x[1]+(0.115821365646)*x[2]
        s[1,0,2,0]=(-0.280246915004)*x[0]+(-0.0972556765171)*x[1]+(0.758816560386)*x[2]
        s[1,0,2,1]=(1.0396259589)*x[0]+(-0.239507470189)*x[1]+(-0.987815676492)*x[2]
        s[1,1,0,0]=(0.245203366317)*x[0]+(-1.53269800659)*x[1]+(0.619360300925)*x[2]
        s[1,1,0,1]=(0.246406467482)*x[0]+(-0.729988292889)*x[1]+(0.585422287189)*x[2]
        s[1,1,1,0]=(1.19810038121)*x[0]+(-0.72772445701)*x[1]+(1.11619392954)*x[2]
        s[1,1,1,1]=(-0.311984976134)*x[0]+(-0.188443023344)*x[1]+(-0.100078290654)*x[2]
        s[1,1,2,0]=(0.158578485066)*x[0]+(0.214804573675)*x[1]+(-0.22418434724)*x[2]
        s[1,1,2,1]=(-0.0550611413075)*x[0]+(-0.147088267891)*x[1]+(-0.269933900128)*x[2]
        s[1,2,0,0]=(-0.8618782319)*x[0]+(-0.770284701216)*x[1]+(-0.407458817279)*x[2]
        s[1,2,0,1]=(-0.242336600721)*x[0]+(-1.57823911131)*x[1]+(0.452055841078)*x[2]
        s[1,2,1,0]=(0.0537219652862)*x[0]+(0.18864518829)*x[1]+(0.021034116382)*x[2]
        s[1,2,1,1]=(0.143799709809)*x[0]+(0.787934959856)*x[1]+(-0.283457275339)*x[2]
        s[1,2,2,0]=(0.475965504456)*x[0]+(-0.744950301051)*x[1]+(0.45308618026)*x[2]
        s[1,2,2,1]=(0.561593580643)*x[0]+(0.121626528038)*x[1]+(-1.24221676348)*x[2]
        s[1,3,0,0]=(-0.340056043985)*x[0]+(-0.434795551938)*x[1]+(-0.291179084342)*x[2]
        s[1,3,0,1]=(0.312318149724)*x[0]+(1.66950279503)*x[1]+(-1.1622076779)*x[2]
        s[1,3,1,0]=(0.0965172054443)*x[0]+(-0.64102022474)*x[1]+(-0.173612340622)*x[2]
        s[1,3,1,1]=(1.88701305968)*x[0]+(-1.66794136241)*x[1]+(1.56059966446)*x[2]
        s[1,3,2,0]=(-0.497773999274)*x[0]+(0.395345962943)*x[1]+(-0.499105933847)*x[2]
        s[1,3,2,1]=(-0.312096437005)*x[0]+(-1.23301763457)*x[1]+(0.198835942067)*x[2]
        s[1,4,0,0]=(0.076377295342)*x[0]+(-0.487847012612)*x[1]+(-0.084314987823)*x[2]
        s[1,4,0,1]=(0.666434654203)*x[0]+(-0.220376188514)*x[1]+(-0.355749198488)*x[2]
        s[1,4,1,0]=(-0.701104468581)*x[0]+(0.828174884422)*x[1]+(0.309374641659)*x[2]
        s[1,4,1,1]=(0.526340204296)*x[0]+(0.571598927007)*x[1]+(-0.0793826750713)*x[2]
        s[1,4,2,0]=(0.775563727486)*x[0]+(0.248228533771)*x[1]+(-1.18169855975)*x[2]
        s[1,4,2,1]=(-0.399357449979)*x[0]+(-0.723554468137)*x[1]+(0.0231913550517)*x[2]
        s[2,0,0,0]=(-0.0504735749489)*x[0]+(0.506564345355)*x[1]+(0.790336060204)*x[2]
        s[2,0,0,1]=(0.426187262022)*x[0]+(0.966680952632)*x[1]+(0.410542727369)*x[2]
        s[2,0,1,0]=(0.594102954821)*x[0]+(0.0639755896611)*x[1]+(-0.675769510238)*x[2]
        s[2,0,1,1]=(0.367479711308)*x[0]+(0.150736830273)*x[1]+(1.15797942999)*x[2]
        s[2,0,2,0]=(0.875207101066)*x[0]+(-1.12939492894)*x[1]+(0.73561332295)*x[2]
        s[2,0,2,1]=(-1.70885193598)*x[0]+(0.835257764029)*x[1]+(-1.58521909429)*x[2]
        s[2,1,0,0]=(-0.144255046397)*x[0]+(0.590074178586)*x[1]+(0.592144641271)*x[2]
        s[2,1,0,1]=(-0.0688470601204)*x[0]+(-0.153340975531)*x[1]+(1.15241895553)*x[2]
        s[2,1,1,0]=(-0.73257033584)*x[0]+(-0.563316507818)*x[1]+(-1.05620654545)*x[2]
        s[2,1,1,1]=(-1.29833350092)*x[0]+(0.42237934165)*x[1]+(0.0725296431412)*x[2]
        s[2,1,2,0]=(-0.837473685756)*x[0]+(-0.425072509905)*x[1]+(-0.463397001131)*x[2]
        s[2,1,2,1]=(-0.937038584847)*x[0]+(0.531989188194)*x[1]+(-1.06694316698)*x[2]
        s[2,2,0,0]=(-0.846346120269)*x[0]+(-0.639618775615)*x[1]+(-0.928154018676)*x[2]
        s[2,2,0,1]=(-0.726631889289)*x[0]+(-0.1318624965)*x[1]+(-0.343729797494)*x[2]
        s[2,2,1,0]=(0.802665744539)*x[0]+(-0.115632355474)*x[1]+(0.303919795863)*x[2]
        s[2,2,1,1]=(0.0371825373918)*x[0]+(-1.17799388388)*x[1]+(-0.223043449617)*x[2]
        s[2,2,2,0]=(0.75988617552)*x[0]+(-0.608977501492)*x[1]+(1.75211913311)*x[2]
        s[2,2,2,1]=(-0.291397991509)*x[0]+(-1.55173354063)*x[1]+(-0.461269447756)*x[2]
        s[2,3,0,0]=(0.351624965509)*x[0]+(-1.32799214259)*x[1]+(1.58946725409)*x[2]
        s[2,3,0,1]=(0.180149992744)*x[0]+(1.07861987473)*x[1]+(0.834357928788)*x[2]
        s[2,3,1,0]=(0.0475016328103)*x[0]+(0.707449786509)*x[1]+(-0.370421903676)*x[2]
        s[2,3,1,1]=(0.553988297947)*x[0]+(1.13388241001)*x[1]+(0.575604300083)*x[2]
        s[2,3,2,0]=(-0.249047035016)*x[0]+(0.15013128589)*x[1]+(1.3953926353)*x[2]
        s[2,3,2,1]=(-0.783008188949)*x[0]+(0.861992436037)*x[1]+(-0.929496477422)*x[2]
        s[2,4,0,0]=(0.264348996811)*x[0]+(-1.23207394036)*x[1]+(-1.37804290092)*x[2]
        s[2,4,0,1]=(0.360782273055)*x[0]+(-0.270520439821)*x[1]+(-1.01285201623)*x[2]
        s[2,4,1,0]=(0.280578579439)*x[0]+(0.584696145403)*x[1]+(-1.21852764117)*x[2]
        s[2,4,1,1]=(0.112102121128)*x[0]+(-0.0619663521973)*x[1]+(1.08490293257)*x[2]
        s[2,4,2,0]=(-1.89749569924)*x[0]+(-0.21217889191)*x[1]+(0.635524779965)*x[2]
        s[2,4,2,1]=(-0.309239337357)*x[0]+(0.0116284822724)*x[1]+(0.0107247712951)*x[2]
        s[3,0,0,0]=(1.23545873872)*x[0]+(-1.29896744411)*x[1]+(0.855094335492)*x[2]
        s[3,0,0,1]=(-0.51465695324)*x[0]+(1.47799693308)*x[1]+(0.0718360423316)*x[2]
        s[3,0,1,0]=(0.220505442402)*x[0]+(-0.3171385194)*x[1]+(0.27954628665)*x[2]
        s[3,0,1,1]=(0.860582046972)*x[0]+(1.04565549687)*x[1]+(1.18108647078)*x[2]
        s[3,0,2,0]=(-0.593547630941)*x[0]+(-0.0390465590147)*x[1]+(0.00129707682534)*x[2]
        s[3,0,2,1]=(1.39210293617)*x[0]+(-1.05660957806)*x[1]+(-1.20489109051)*x[2]
        s[3,1,0,0]=(-0.536035988946)*x[0]+(-0.594403345465)*x[1]+(1.11719111746)*x[2]
        s[3,1,0,1]=(-0.0403557954127)*x[0]+(-1.32391025453)*x[1]+(-0.639127329416)*x[2]
        s[3,1,1,0]=(-0.495803710666)*x[0]+(-0.199832989186)*x[1]+(-0.832862076561)*x[2]
        s[3,1,1,1]=(0.0216882402803)*x[0]+(0.367268873436)*x[1]+(0.295944335418)*x[2]
        s[3,1,2,0]=(0.172936506269)*x[0]+(-0.339886198196)*x[1]+(0.268201481244)*x[2]
        s[3,1,2,1]=(0.665786569897)*x[0]+(-0.290637226736)*x[1]+(0.617677178537)*x[2]
        s[3,2,0,0]=(-0.657902185911)*x[0]+(-1.34292379802)*x[1]+(0.0641950819134)*x[2]
        s[3,2,0,1]=(0.352039175923)*x[0]+(-0.483047873735)*x[1]+(0.47997600743)*x[2]
        s[3,2,1,0]=(0.944619484106)*x[0]+(-0.308755547249)*x[1]+(-0.765906593989)*x[2]
        s[3,2,1,1]=(-1.09674781865)*x[0]+(0.013069257236)*x[1]+(-0.0741247275628)*x[2]
        s[3,2,2,0]=(0.400057737624)*x[0]+(-1.16165281514)*x[1]+(-0.789757454741)*x[2]
        s[3,2,2,1]=(0.367100312493)*x[0]+(-0.547338908792)*x[1]+(-0.409173859392)*x[2]
        s[3,3,0,0]=(0.890118369638)*x[0]+(1.201296587)*x[1]+(-0.0319246476545)*x[2]
        s[3,3,0,1]=(0.620955702501)*x[0]+(0.438440818114)*x[1]+(0.499747510411)*x[2]
        s[3,3,1,0]=(1.61217537222)*x[0]+(-1.21323714148)*x[1]+(-1.06137097571)*x[2]
        s[3,3,1,1]=(-1.75942413542)*x[0]+(0.440345678738)*x[1]+(0.625299774096)*x[2]
        s[3,3,2,0]=(-0.878497266865)*x[0]+(-1.15879244679)*x[1]+(1.03993124425)*x[2]
        s[3,3,2,1]=(-0.0138445968491)*x[0]+(0.547910692711)*x[1]+(0.971642264613)*x[2]
        s[3,4,0,0]=(-0.107769272336)*x[0]+(-0.987911384393)*x[1]+(1.2076201556)*x[2]
        s[3,4,0,1]=(-0.556570952071)*x[0]+(-0.284568452888)*x[1]+(1.23341917108)*x[2]
        s[3,4,1,0]=(1.56461193947)*x[0]+(0.185382898933)*x[1]+(-1.67714180845)*x[2]
        s[3,4,1,1]=(-0.894934034676)*x[0]+(1.05560233139)*x[1]+(0.651029150593)*x[2]
        s[3,4,2,0]=(0.0559603779066)*x[0]+(-1.72752549101)*x[1]+(-1.02017554749)*x[2]
        s[3,4,2,1]=(-1.75424333815)*x[0]+(-0.198869840282)*x[1]+(0.475065396299)*x[2]
        ref[0,0,0,0]=(-0.270116752643)*x_ref[0]+(0.0715809194071)*x_ref[1]+(0.930011298681)*x_ref[2]
        ref[0,0,0,1]=(-0.0435807732016)*x_ref[0]+(-0.492622221729)*x_ref[1]+(-1.36754174971)*x_ref[2]
        ref[0,0,1,0]=(-0.949818131569)*x_ref[0]+(1.08437418341)*x_ref[1]+(0.527267447353)*x_ref[2]
        ref[0,0,1,1]=(-0.164876627537)*x_ref[0]+(-1.44143896699)*x_ref[1]+(0.0355140164081)*x_ref[2]
        ref[0,0,2,0]=(-0.530229192916)*x_ref[0]+(-0.988412045132)*x_ref[1]+(-0.466520393878)*x_ref[2]
        ref[0,0,2,1]=(-0.0166373496749)*x_ref[0]+(0.322621663585)*x_ref[1]+(-0.102019336059)*x_ref[2]
        ref[0,1,0,0]=(-1.36627178258)*x_ref[0]+(-0.803463758822)*x_ref[1]+(0.734446026661)*x_ref[2]
        ref[0,1,0,1]=(-0.804729478825)*x_ref[0]+(-1.24406753288)*x_ref[1]+(-0.755743791016)*x_ref[2]
        ref[0,1,1,0]=(1.03869027308)*x_ref[0]+(-0.25237293575)*x_ref[1]+(1.43737134487)*x_ref[2]
        ref[0,1,1,1]=(1.48269192251)*x_ref[0]+(0.045590449115)*x_ref[1]+(0.47425103841)*x_ref[2]
        ref[0,1,2,0]=(-0.738695287005)*x_ref[0]+(1.01938431903)*x_ref[1]+(1.20767534795)*x_ref[2]
        ref[0,1,2,1]=(-0.237445941287)*x_ref[0]+(0.766593746786)*x_ref[1]+(0.0297279831889)*x_ref[2]
        ref[0,2,0,0]=(-0.804902383375)*x_ref[0]+(0.0962339257662)*x_ref[1]+(1.37407473707)*x_ref[2]
        ref[0,2,0,1]=(-0.0821182376238)*x_ref[0]+(0.35489327051)*x_ref[1]+(-1.43235115632)*x_ref[2]
        ref[0,2,1,0]=(1.11593707013)*x_ref[0]+(0.744039820943)*x_ref[1]+(0.294815281064)*x_ref[2]
        ref[0,2,1,1]=(0.0789595478782)*x_ref[0]+(0.719656781874)*x_ref[1]+(0.355275528053)*x_ref[2]
        ref[0,2,2,0]=(1.20514986435)*x_ref[0]+(0.203004753589)*x_ref[1]+(-0.0731941650839)*x_ref[2]
        ref[0,2,2,1]=(-1.23231910051)*x_ref[0]+(-0.470834142341)*x_ref[1]+(-0.633518095593)*x_ref[2]
        ref[0,3,0,0]=(0.468164093965)*x_ref[0]+(-1.15996958069)*x_ref[1]+(0.922274695212)*x_ref[2]
        ref[0,3,0,1]=(-0.964250141036)*x_ref[0]+(-0.769747467017)*x_ref[1]+(-1.16369265956)*x_ref[2]
        ref[0,3,1,0]=(1.1622810944)*x_ref[0]+(-1.72794997517)*x_ref[1]+(-0.297833274612)*x_ref[2]
        ref[0,3,1,1]=(1.20239531447)*x_ref[0]+(-0.195111187444)*x_ref[1]+(-0.720186501147)*x_ref[2]
        ref[0,3,2,0]=(-0.0760001358746)*x_ref[0]+(-1.65266437471)*x_ref[1]+(-0.496163951015)*x_ref[2]
        ref[0,3,2,1]=(-1.4998248458)*x_ref[0]+(-1.73305168566)*x_ref[1]+(-0.107384086634)*x_ref[2]
        ref[0,4,0,0]=(0.235699357585)*x_ref[0]+(-0.213348404403)*x_ref[1]+(0.35368491512)*x_ref[2]
        ref[0,4,0,1]=(0.264641788778)*x_ref[0]+(0.435251584452)*x_ref[1]+(0.22539974851)*x_ref[2]
        ref[0,4,1,0]=(0.540203510472)*x_ref[0]+(-1.33243180984)*x_ref[1]+(-0.520357911324)*x_ref[2]
        ref[0,4,1,1]=(-0.629868492098)*x_ref[0]+(-0.403402885662)*x_ref[1]+(-0.184983842426)*x_ref[2]
        ref[0,4,2,0]=(1.17843494613)*x_ref[0]+(0.562783141759)*x_ref[1]+(-1.48575741726)*x_ref[2]
        ref[0,4,2,1]=(0.623796209289)*x_ref[0]+(1.02627846638)*x_ref[1]+(1.39369241754)*x_ref[2]
        ref[1,0,0,0]=(1.9855507901)*x_ref[0]+(0.152976407288)*x_ref[1]+(0.0931547545156)*x_ref[2]
        ref[1,0,0,1]=(-0.422055088151)*x_ref[0]+(-0.509449435814)*x_ref[1]+(-1.31705141661)*x_ref[2]
        ref[1,0,1,0]=(-0.24193816853)*x_ref[0]+(-0.68450536388)*x_ref[1]+(-0.309024887085)*x_ref[2]
        ref[1,0,1,1]=(1.24557169323)*x_ref[0]+(0.353538565492)*x_ref[1]+(0.115821365646)*x_ref[2]
        ref[1,0,2,0]=(-0.280246915004)*x_ref[0]+(-0.0972556765171)*x_ref[1]+(0.758816560386)*x_ref[2]
        ref[1,0,2,1]=(1.0396259589)*x_ref[0]+(-0.239507470189)*x_ref[1]+(-0.987815676492)*x_ref[2]
        ref[1,1,0,0]=(0.245203366317)*x_ref[0]+(-1.53269800659)*x_ref[1]+(0.619360300925)*x_ref[2]
        ref[1,1,0,1]=(0.246406467482)*x_ref[0]+(-0.729988292889)*x_ref[1]+(0.585422287189)*x_ref[2]
        ref[1,1,1,0]=(1.19810038121)*x_ref[0]+(-0.72772445701)*x_ref[1]+(1.11619392954)*x_ref[2]
        ref[1,1,1,1]=(-0.311984976134)*x_ref[0]+(-0.188443023344)*x_ref[1]+(-0.100078290654)*x_ref[2]
        ref[1,1,2,0]=(0.158578485066)*x_ref[0]+(0.214804573675)*x_ref[1]+(-0.22418434724)*x_ref[2]
        ref[1,1,2,1]=(-0.0550611413075)*x_ref[0]+(-0.147088267891)*x_ref[1]+(-0.269933900128)*x_ref[2]
        ref[1,2,0,0]=(-0.8618782319)*x_ref[0]+(-0.770284701216)*x_ref[1]+(-0.407458817279)*x_ref[2]
        ref[1,2,0,1]=(-0.242336600721)*x_ref[0]+(-1.57823911131)*x_ref[1]+(0.452055841078)*x_ref[2]
        ref[1,2,1,0]=(0.0537219652862)*x_ref[0]+(0.18864518829)*x_ref[1]+(0.021034116382)*x_ref[2]
        ref[1,2,1,1]=(0.143799709809)*x_ref[0]+(0.787934959856)*x_ref[1]+(-0.283457275339)*x_ref[2]
        ref[1,2,2,0]=(0.475965504456)*x_ref[0]+(-0.744950301051)*x_ref[1]+(0.45308618026)*x_ref[2]
        ref[1,2,2,1]=(0.561593580643)*x_ref[0]+(0.121626528038)*x_ref[1]+(-1.24221676348)*x_ref[2]
        ref[1,3,0,0]=(-0.340056043985)*x_ref[0]+(-0.434795551938)*x_ref[1]+(-0.291179084342)*x_ref[2]
        ref[1,3,0,1]=(0.312318149724)*x_ref[0]+(1.66950279503)*x_ref[1]+(-1.1622076779)*x_ref[2]
        ref[1,3,1,0]=(0.0965172054443)*x_ref[0]+(-0.64102022474)*x_ref[1]+(-0.173612340622)*x_ref[2]
        ref[1,3,1,1]=(1.88701305968)*x_ref[0]+(-1.66794136241)*x_ref[1]+(1.56059966446)*x_ref[2]
        ref[1,3,2,0]=(-0.497773999274)*x_ref[0]+(0.395345962943)*x_ref[1]+(-0.499105933847)*x_ref[2]
        ref[1,3,2,1]=(-0.312096437005)*x_ref[0]+(-1.23301763457)*x_ref[1]+(0.198835942067)*x_ref[2]
        ref[1,4,0,0]=(0.076377295342)*x_ref[0]+(-0.487847012612)*x_ref[1]+(-0.084314987823)*x_ref[2]
        ref[1,4,0,1]=(0.666434654203)*x_ref[0]+(-0.220376188514)*x_ref[1]+(-0.355749198488)*x_ref[2]
        ref[1,4,1,0]=(-0.701104468581)*x_ref[0]+(0.828174884422)*x_ref[1]+(0.309374641659)*x_ref[2]
        ref[1,4,1,1]=(0.526340204296)*x_ref[0]+(0.571598927007)*x_ref[1]+(-0.0793826750713)*x_ref[2]
        ref[1,4,2,0]=(0.775563727486)*x_ref[0]+(0.248228533771)*x_ref[1]+(-1.18169855975)*x_ref[2]
        ref[1,4,2,1]=(-0.399357449979)*x_ref[0]+(-0.723554468137)*x_ref[1]+(0.0231913550517)*x_ref[2]
        ref[2,0,0,0]=(-0.0504735749489)*x_ref[0]+(0.506564345355)*x_ref[1]+(0.790336060204)*x_ref[2]
        ref[2,0,0,1]=(0.426187262022)*x_ref[0]+(0.966680952632)*x_ref[1]+(0.410542727369)*x_ref[2]
        ref[2,0,1,0]=(0.594102954821)*x_ref[0]+(0.0639755896611)*x_ref[1]+(-0.675769510238)*x_ref[2]
        ref[2,0,1,1]=(0.367479711308)*x_ref[0]+(0.150736830273)*x_ref[1]+(1.15797942999)*x_ref[2]
        ref[2,0,2,0]=(0.875207101066)*x_ref[0]+(-1.12939492894)*x_ref[1]+(0.73561332295)*x_ref[2]
        ref[2,0,2,1]=(-1.70885193598)*x_ref[0]+(0.835257764029)*x_ref[1]+(-1.58521909429)*x_ref[2]
        ref[2,1,0,0]=(-0.144255046397)*x_ref[0]+(0.590074178586)*x_ref[1]+(0.592144641271)*x_ref[2]
        ref[2,1,0,1]=(-0.0688470601204)*x_ref[0]+(-0.153340975531)*x_ref[1]+(1.15241895553)*x_ref[2]
        ref[2,1,1,0]=(-0.73257033584)*x_ref[0]+(-0.563316507818)*x_ref[1]+(-1.05620654545)*x_ref[2]
        ref[2,1,1,1]=(-1.29833350092)*x_ref[0]+(0.42237934165)*x_ref[1]+(0.0725296431412)*x_ref[2]
        ref[2,1,2,0]=(-0.837473685756)*x_ref[0]+(-0.425072509905)*x_ref[1]+(-0.463397001131)*x_ref[2]
        ref[2,1,2,1]=(-0.937038584847)*x_ref[0]+(0.531989188194)*x_ref[1]+(-1.06694316698)*x_ref[2]
        ref[2,2,0,0]=(-0.846346120269)*x_ref[0]+(-0.639618775615)*x_ref[1]+(-0.928154018676)*x_ref[2]
        ref[2,2,0,1]=(-0.726631889289)*x_ref[0]+(-0.1318624965)*x_ref[1]+(-0.343729797494)*x_ref[2]
        ref[2,2,1,0]=(0.802665744539)*x_ref[0]+(-0.115632355474)*x_ref[1]+(0.303919795863)*x_ref[2]
        ref[2,2,1,1]=(0.0371825373918)*x_ref[0]+(-1.17799388388)*x_ref[1]+(-0.223043449617)*x_ref[2]
        ref[2,2,2,0]=(0.75988617552)*x_ref[0]+(-0.608977501492)*x_ref[1]+(1.75211913311)*x_ref[2]
        ref[2,2,2,1]=(-0.291397991509)*x_ref[0]+(-1.55173354063)*x_ref[1]+(-0.461269447756)*x_ref[2]
        ref[2,3,0,0]=(0.351624965509)*x_ref[0]+(-1.32799214259)*x_ref[1]+(1.58946725409)*x_ref[2]
        ref[2,3,0,1]=(0.180149992744)*x_ref[0]+(1.07861987473)*x_ref[1]+(0.834357928788)*x_ref[2]
        ref[2,3,1,0]=(0.0475016328103)*x_ref[0]+(0.707449786509)*x_ref[1]+(-0.370421903676)*x_ref[2]
        ref[2,3,1,1]=(0.553988297947)*x_ref[0]+(1.13388241001)*x_ref[1]+(0.575604300083)*x_ref[2]
        ref[2,3,2,0]=(-0.249047035016)*x_ref[0]+(0.15013128589)*x_ref[1]+(1.3953926353)*x_ref[2]
        ref[2,3,2,1]=(-0.783008188949)*x_ref[0]+(0.861992436037)*x_ref[1]+(-0.929496477422)*x_ref[2]
        ref[2,4,0,0]=(0.264348996811)*x_ref[0]+(-1.23207394036)*x_ref[1]+(-1.37804290092)*x_ref[2]
        ref[2,4,0,1]=(0.360782273055)*x_ref[0]+(-0.270520439821)*x_ref[1]+(-1.01285201623)*x_ref[2]
        ref[2,4,1,0]=(0.280578579439)*x_ref[0]+(0.584696145403)*x_ref[1]+(-1.21852764117)*x_ref[2]
        ref[2,4,1,1]=(0.112102121128)*x_ref[0]+(-0.0619663521973)*x_ref[1]+(1.08490293257)*x_ref[2]
        ref[2,4,2,0]=(-1.89749569924)*x_ref[0]+(-0.21217889191)*x_ref[1]+(0.635524779965)*x_ref[2]
        ref[2,4,2,1]=(-0.309239337357)*x_ref[0]+(0.0116284822724)*x_ref[1]+(0.0107247712951)*x_ref[2]
        ref[3,0,0,0]=(1.23545873872)*x_ref[0]+(-1.29896744411)*x_ref[1]+(0.855094335492)*x_ref[2]
        ref[3,0,0,1]=(-0.51465695324)*x_ref[0]+(1.47799693308)*x_ref[1]+(0.0718360423316)*x_ref[2]
        ref[3,0,1,0]=(0.220505442402)*x_ref[0]+(-0.3171385194)*x_ref[1]+(0.27954628665)*x_ref[2]
        ref[3,0,1,1]=(0.860582046972)*x_ref[0]+(1.04565549687)*x_ref[1]+(1.18108647078)*x_ref[2]
        ref[3,0,2,0]=(-0.593547630941)*x_ref[0]+(-0.0390465590147)*x_ref[1]+(0.00129707682534)*x_ref[2]
        ref[3,0,2,1]=(1.39210293617)*x_ref[0]+(-1.05660957806)*x_ref[1]+(-1.20489109051)*x_ref[2]
        ref[3,1,0,0]=(-0.536035988946)*x_ref[0]+(-0.594403345465)*x_ref[1]+(1.11719111746)*x_ref[2]
        ref[3,1,0,1]=(-0.0403557954127)*x_ref[0]+(-1.32391025453)*x_ref[1]+(-0.639127329416)*x_ref[2]
        ref[3,1,1,0]=(-0.495803710666)*x_ref[0]+(-0.199832989186)*x_ref[1]+(-0.832862076561)*x_ref[2]
        ref[3,1,1,1]=(0.0216882402803)*x_ref[0]+(0.367268873436)*x_ref[1]+(0.295944335418)*x_ref[2]
        ref[3,1,2,0]=(0.172936506269)*x_ref[0]+(-0.339886198196)*x_ref[1]+(0.268201481244)*x_ref[2]
        ref[3,1,2,1]=(0.665786569897)*x_ref[0]+(-0.290637226736)*x_ref[1]+(0.617677178537)*x_ref[2]
        ref[3,2,0,0]=(-0.657902185911)*x_ref[0]+(-1.34292379802)*x_ref[1]+(0.0641950819134)*x_ref[2]
        ref[3,2,0,1]=(0.352039175923)*x_ref[0]+(-0.483047873735)*x_ref[1]+(0.47997600743)*x_ref[2]
        ref[3,2,1,0]=(0.944619484106)*x_ref[0]+(-0.308755547249)*x_ref[1]+(-0.765906593989)*x_ref[2]
        ref[3,2,1,1]=(-1.09674781865)*x_ref[0]+(0.013069257236)*x_ref[1]+(-0.0741247275628)*x_ref[2]
        ref[3,2,2,0]=(0.400057737624)*x_ref[0]+(-1.16165281514)*x_ref[1]+(-0.789757454741)*x_ref[2]
        ref[3,2,2,1]=(0.367100312493)*x_ref[0]+(-0.547338908792)*x_ref[1]+(-0.409173859392)*x_ref[2]
        ref[3,3,0,0]=(0.890118369638)*x_ref[0]+(1.201296587)*x_ref[1]+(-0.0319246476545)*x_ref[2]
        ref[3,3,0,1]=(0.620955702501)*x_ref[0]+(0.438440818114)*x_ref[1]+(0.499747510411)*x_ref[2]
        ref[3,3,1,0]=(1.61217537222)*x_ref[0]+(-1.21323714148)*x_ref[1]+(-1.06137097571)*x_ref[2]
        ref[3,3,1,1]=(-1.75942413542)*x_ref[0]+(0.440345678738)*x_ref[1]+(0.625299774096)*x_ref[2]
        ref[3,3,2,0]=(-0.878497266865)*x_ref[0]+(-1.15879244679)*x_ref[1]+(1.03993124425)*x_ref[2]
        ref[3,3,2,1]=(-0.0138445968491)*x_ref[0]+(0.547910692711)*x_ref[1]+(0.971642264613)*x_ref[2]
        ref[3,4,0,0]=(-0.107769272336)*x_ref[0]+(-0.987911384393)*x_ref[1]+(1.2076201556)*x_ref[2]
        ref[3,4,0,1]=(-0.556570952071)*x_ref[0]+(-0.284568452888)*x_ref[1]+(1.23341917108)*x_ref[2]
        ref[3,4,1,0]=(1.56461193947)*x_ref[0]+(0.185382898933)*x_ref[1]+(-1.67714180845)*x_ref[2]
        ref[3,4,1,1]=(-0.894934034676)*x_ref[0]+(1.05560233139)*x_ref[1]+(0.651029150593)*x_ref[2]
        ref[3,4,2,0]=(0.0559603779066)*x_ref[0]+(-1.72752549101)*x_ref[1]+(-1.02017554749)*x_ref[2]
        ref[3,4,2,1]=(-1.75424333815)*x_ref[0]+(-0.198869840282)*x_ref[1]+(0.475065396299)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(-1.02854286859)*x[0]+(0.105922853194)*x[1]
        ref=(-1.02854286859)*x_ref[0]+(0.105922853194)*x_ref[1]
      else:
        arg=(1.36620376771)*x[0]+(-0.336798536224)*x[1]+(0.534538499605)*x[2]
        ref=(1.36620376771)*x_ref[0]+(-0.336798536224)*x_ref[1]+(0.534538499605)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        arg[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
        ref[0]=(-0.403120829068)*x[0]+(-0.70319880694)*x[1]
        ref[1]=(0.936461664193)*x[0]+(1.3454908089)*x[1]
      else:
        arg[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        arg[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
        ref[0]=(0.154137304713)*x[0]+(0.332968506642)*x[1]+(-0.4661136382)*x[2]
        ref[1]=(0.306790292003)*x[0]+(0.466856258098)*x[1]+(0.205782933414)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.452635897444)*x[0]+(-0.210513646405)*x[1]
        arg[0,1]=(-0.590704436627)*x[0]+(1.78000273884)*x[1]
        arg[0,2]=(1.01331349402)*x[0]+(1.76279765125)*x[1]
        arg[0,3]=(1.5588852701)*x[0]+(-0.550895238707)*x[1]
        arg[0,4]=(-0.352453804527)*x[0]+(-0.338832613421)*x[1]
        arg[1,0]=(-0.809837877163)*x[0]+(-0.531009602702)*x[1]
        arg[1,1]=(0.513641503665)*x[0]+(0.232114989082)*x[1]
        arg[1,2]=(-0.424635655997)*x[0]+(0.0796757576821)*x[1]
        arg[1,3]=(-1.5899938122)*x[0]+(-0.178949548482)*x[1]
        arg[1,4]=(0.989317971495)*x[0]+(0.969711267468)*x[1]
        arg[2,0]=(0.384490947785)*x[0]+(1.42230193156)*x[1]
        arg[2,1]=(-0.0832227929255)*x[0]+(0.859020891429)*x[1]
        arg[2,2]=(-1.36129563124)*x[0]+(-0.240985031888)*x[1]
        arg[2,3]=(0.319706901063)*x[0]+(1.03419924566)*x[1]
        arg[2,4]=(0.986965249719)*x[0]+(0.228684031472)*x[1]
        arg[3,0]=(0.588520127337)*x[0]+(0.016799594822)*x[1]
        arg[3,1]=(1.46193438322)*x[0]+(0.431838900383)*x[1]
        arg[3,2]=(-0.0235512206818)*x[0]+(0.00227038037474)*x[1]
        arg[3,3]=(0.977383733076)*x[0]+(0.383979958766)*x[1]
        arg[3,4]=(0.447782383836)*x[0]+(-0.504218231709)*x[1]
        ref[0,0]=(-0.452635897444)*x_ref[0]+(-0.210513646405)*x_ref[1]
        ref[0,1]=(-0.590704436627)*x_ref[0]+(1.78000273884)*x_ref[1]
        ref[0,2]=(1.01331349402)*x_ref[0]+(1.76279765125)*x_ref[1]
        ref[0,3]=(1.5588852701)*x_ref[0]+(-0.550895238707)*x_ref[1]
        ref[0,4]=(-0.352453804527)*x_ref[0]+(-0.338832613421)*x_ref[1]
        ref[1,0]=(-0.809837877163)*x_ref[0]+(-0.531009602702)*x_ref[1]
        ref[1,1]=(0.513641503665)*x_ref[0]+(0.232114989082)*x_ref[1]
        ref[1,2]=(-0.424635655997)*x_ref[0]+(0.0796757576821)*x_ref[1]
        ref[1,3]=(-1.5899938122)*x_ref[0]+(-0.178949548482)*x_ref[1]
        ref[1,4]=(0.989317971495)*x_ref[0]+(0.969711267468)*x_ref[1]
        ref[2,0]=(0.384490947785)*x_ref[0]+(1.42230193156)*x_ref[1]
        ref[2,1]=(-0.0832227929255)*x_ref[0]+(0.859020891429)*x_ref[1]
        ref[2,2]=(-1.36129563124)*x_ref[0]+(-0.240985031888)*x_ref[1]
        ref[2,3]=(0.319706901063)*x_ref[0]+(1.03419924566)*x_ref[1]
        ref[2,4]=(0.986965249719)*x_ref[0]+(0.228684031472)*x_ref[1]
        ref[3,0]=(0.588520127337)*x_ref[0]+(0.016799594822)*x_ref[1]
        ref[3,1]=(1.46193438322)*x_ref[0]+(0.431838900383)*x_ref[1]
        ref[3,2]=(-0.0235512206818)*x_ref[0]+(0.00227038037474)*x_ref[1]
        ref[3,3]=(0.977383733076)*x_ref[0]+(0.383979958766)*x_ref[1]
        ref[3,4]=(0.447782383836)*x_ref[0]+(-0.504218231709)*x_ref[1]
      else:
        arg[0,0]=(-0.566459971675)*x[0]+(0.240246648893)*x[1]+(1.84693181923)*x[2]
        arg[0,1]=(1.09715015899)*x[0]+(1.26531924413)*x[1]+(0.627813347679)*x[2]
        arg[0,2]=(-1.17535073843)*x[0]+(-0.215601498941)*x[1]+(1.36346500888)*x[2]
        arg[0,3]=(-0.637207120161)*x[0]+(1.08542566952)*x[1]+(-1.13757514425)*x[2]
        arg[0,4]=(1.18247424069)*x[0]+(-1.48951614815)*x[1]+(0.615535039719)*x[2]
        arg[1,0]=(0.0937432100426)*x[0]+(-0.510326894156)*x[1]+(-0.456208996457)*x[2]
        arg[1,1]=(-0.866120342442)*x[0]+(0.927484984985)*x[1]+(0.918081524156)*x[2]
        arg[1,2]=(0.025238374235)*x[0]+(-0.182938599269)*x[1]+(-1.50048281322)*x[2]
        arg[1,3]=(-0.302334971294)*x[0]+(-0.925062799423)*x[1]+(-1.4675138953)*x[2]
        arg[1,4]=(-0.497738566447)*x[0]+(-0.324727071875)*x[1]+(-1.43341212055)*x[2]
        arg[2,0]=(-1.4605237958)*x[0]+(-0.332826007088)*x[1]+(-0.230483401111)*x[2]
        arg[2,1]=(0.801769248581)*x[0]+(-0.930588036503)*x[1]+(-1.06804808381)*x[2]
        arg[2,2]=(-1.07972165017)*x[0]+(1.09295362882)*x[1]+(0.048550098477)*x[2]
        arg[2,3]=(0.028290230337)*x[0]+(0.179461246635)*x[1]+(-1.09165257072)*x[2]
        arg[2,4]=(0.238850986174)*x[0]+(-0.837307206184)*x[1]+(-0.617254874594)*x[2]
        arg[3,0]=(-0.125921204582)*x[0]+(0.234758657879)*x[1]+(0.0847162247843)*x[2]
        arg[3,1]=(-0.213352414907)*x[0]+(0.206777352455)*x[1]+(-0.135357638131)*x[2]
        arg[3,2]=(-0.41897394434)*x[0]+(0.152706100016)*x[1]+(1.52390356012)*x[2]
        arg[3,3]=(0.1831705248)*x[0]+(-1.13556841364)*x[1]+(-1.7661652693)*x[2]
        arg[3,4]=(0.43473708388)*x[0]+(1.04496896213)*x[1]+(-0.816894739325)*x[2]
        ref[0,0]=(-0.566459971675)*x_ref[0]+(0.240246648893)*x_ref[1]+(1.84693181923)*x_ref[2]
        ref[0,1]=(1.09715015899)*x_ref[0]+(1.26531924413)*x_ref[1]+(0.627813347679)*x_ref[2]
        ref[0,2]=(-1.17535073843)*x_ref[0]+(-0.215601498941)*x_ref[1]+(1.36346500888)*x_ref[2]
        ref[0,3]=(-0.637207120161)*x_ref[0]+(1.08542566952)*x_ref[1]+(-1.13757514425)*x_ref[2]
        ref[0,4]=(1.18247424069)*x_ref[0]+(-1.48951614815)*x_ref[1]+(0.615535039719)*x_ref[2]
        ref[1,0]=(0.0937432100426)*x_ref[0]+(-0.510326894156)*x_ref[1]+(-0.456208996457)*x_ref[2]
        ref[1,1]=(-0.866120342442)*x_ref[0]+(0.927484984985)*x_ref[1]+(0.918081524156)*x_ref[2]
        ref[1,2]=(0.025238374235)*x_ref[0]+(-0.182938599269)*x_ref[1]+(-1.50048281322)*x_ref[2]
        ref[1,3]=(-0.302334971294)*x_ref[0]+(-0.925062799423)*x_ref[1]+(-1.4675138953)*x_ref[2]
        ref[1,4]=(-0.497738566447)*x_ref[0]+(-0.324727071875)*x_ref[1]+(-1.43341212055)*x_ref[2]
        ref[2,0]=(-1.4605237958)*x_ref[0]+(-0.332826007088)*x_ref[1]+(-0.230483401111)*x_ref[2]
        ref[2,1]=(0.801769248581)*x_ref[0]+(-0.930588036503)*x_ref[1]+(-1.06804808381)*x_ref[2]
        ref[2,2]=(-1.07972165017)*x_ref[0]+(1.09295362882)*x_ref[1]+(0.048550098477)*x_ref[2]
        ref[2,3]=(0.028290230337)*x_ref[0]+(0.179461246635)*x_ref[1]+(-1.09165257072)*x_ref[2]
        ref[2,4]=(0.238850986174)*x_ref[0]+(-0.837307206184)*x_ref[1]+(-0.617254874594)*x_ref[2]
        ref[3,0]=(-0.125921204582)*x_ref[0]+(0.234758657879)*x_ref[1]+(0.0847162247843)*x_ref[2]
        ref[3,1]=(-0.213352414907)*x_ref[0]+(0.206777352455)*x_ref[1]+(-0.135357638131)*x_ref[2]
        ref[3,2]=(-0.41897394434)*x_ref[0]+(0.152706100016)*x_ref[1]+(1.52390356012)*x_ref[2]
        ref[3,3]=(0.1831705248)*x_ref[0]+(-1.13556841364)*x_ref[1]+(-1.7661652693)*x_ref[2]
        ref[3,4]=(0.43473708388)*x_ref[0]+(1.04496896213)*x_ref[1]+(-0.816894739325)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.324237136046)*x[0]+(0.67172278334)*x[1]
        arg[0,0,1]=(-1.07360258996)*x[0]+(-0.646208771542)*x[1]
        arg[0,1,0]=(-0.040139403311)*x[0]+(-0.281626553874)*x[1]
        arg[0,1,1]=(0.963343050385)*x[0]+(-0.724838290038)*x[1]
        arg[1,0,0]=(-0.909697586367)*x[0]+(0.48041903189)*x[1]
        arg[1,0,1]=(0.414758676032)*x[0]+(-0.0921084575218)*x[1]
        arg[1,1,0]=(1.06852211241)*x[0]+(-1.01026709886)*x[1]
        arg[1,1,1]=(-0.675258695211)*x[0]+(0.51361078905)*x[1]
        arg[2,0,0]=(-0.791246198786)*x[0]+(-0.0283034425042)*x[1]
        arg[2,0,1]=(-0.87756010848)*x[0]+(-0.00868464377539)*x[1]
        arg[2,1,0]=(0.44209070344)*x[0]+(0.0846344294002)*x[1]
        arg[2,1,1]=(0.777287404475)*x[0]+(0.72614634535)*x[1]
        arg[3,0,0]=(-0.363354104693)*x[0]+(-1.22959837449)*x[1]
        arg[3,0,1]=(-0.532220757425)*x[0]+(-1.54990170304)*x[1]
        arg[3,1,0]=(-0.491099325731)*x[0]+(-0.412402332731)*x[1]
        arg[3,1,1]=(0.0750673227925)*x[0]+(-0.725108664028)*x[1]
        arg[4,0,0]=(-0.419689899608)*x[0]+(-1.32423898043)*x[1]
        arg[4,0,1]=(0.532682204197)*x[0]+(0.0703779631797)*x[1]
        arg[4,1,0]=(1.68643150913)*x[0]+(0.105443733595)*x[1]
        arg[4,1,1]=(0.936329701936)*x[0]+(1.04067808474)*x[1]
        arg[5,0,0]=(1.37285560133)*x[0]+(-0.680844454631)*x[1]
        arg[5,0,1]=(0.843677131436)*x[0]+(0.272365035967)*x[1]
        arg[5,1,0]=(0.00586472074043)*x[0]+(-1.09785622721)*x[1]
        arg[5,1,1]=(1.1410854961)*x[0]+(-0.591468439837)*x[1]
        ref[0,0,0]=(0.324237136046)*x_ref[0]+(0.67172278334)*x_ref[1]
        ref[0,0,1]=(-1.07360258996)*x_ref[0]+(-0.646208771542)*x_ref[1]
        ref[0,1,0]=(-0.040139403311)*x_ref[0]+(-0.281626553874)*x_ref[1]
        ref[0,1,1]=(0.963343050385)*x_ref[0]+(-0.724838290038)*x_ref[1]
        ref[1,0,0]=(-0.909697586367)*x_ref[0]+(0.48041903189)*x_ref[1]
        ref[1,0,1]=(0.414758676032)*x_ref[0]+(-0.0921084575218)*x_ref[1]
        ref[1,1,0]=(1.06852211241)*x_ref[0]+(-1.01026709886)*x_ref[1]
        ref[1,1,1]=(-0.675258695211)*x_ref[0]+(0.51361078905)*x_ref[1]
        ref[2,0,0]=(-0.791246198786)*x_ref[0]+(-0.0283034425042)*x_ref[1]
        ref[2,0,1]=(-0.87756010848)*x_ref[0]+(-0.00868464377539)*x_ref[1]
        ref[2,1,0]=(0.44209070344)*x_ref[0]+(0.0846344294002)*x_ref[1]
        ref[2,1,1]=(0.777287404475)*x_ref[0]+(0.72614634535)*x_ref[1]
        ref[3,0,0]=(-0.363354104693)*x_ref[0]+(-1.22959837449)*x_ref[1]
        ref[3,0,1]=(-0.532220757425)*x_ref[0]+(-1.54990170304)*x_ref[1]
        ref[3,1,0]=(-0.491099325731)*x_ref[0]+(-0.412402332731)*x_ref[1]
        ref[3,1,1]=(0.0750673227925)*x_ref[0]+(-0.725108664028)*x_ref[1]
        ref[4,0,0]=(-0.419689899608)*x_ref[0]+(-1.32423898043)*x_ref[1]
        ref[4,0,1]=(0.532682204197)*x_ref[0]+(0.0703779631797)*x_ref[1]
        ref[4,1,0]=(1.68643150913)*x_ref[0]+(0.105443733595)*x_ref[1]
        ref[4,1,1]=(0.936329701936)*x_ref[0]+(1.04067808474)*x_ref[1]
        ref[5,0,0]=(1.37285560133)*x_ref[0]+(-0.680844454631)*x_ref[1]
        ref[5,0,1]=(0.843677131436)*x_ref[0]+(0.272365035967)*x_ref[1]
        ref[5,1,0]=(0.00586472074043)*x_ref[0]+(-1.09785622721)*x_ref[1]
        ref[5,1,1]=(1.1410854961)*x_ref[0]+(-0.591468439837)*x_ref[1]
      else:
        arg[0,0,0]=(-0.289427308366)*x[0]+(1.17077011703)*x[1]+(0.763196501047)*x[2]
        arg[0,0,1]=(0.642730500265)*x[0]+(1.1719001632)*x[1]+(-0.121378040894)*x[2]
        arg[0,1,0]=(0.0879005621769)*x[0]+(-0.627582537594)*x[1]+(0.220072369447)*x[2]
        arg[0,1,1]=(-0.518762071566)*x[0]+(-0.700754417623)*x[1]+(-0.340623426752)*x[2]
        arg[1,0,0]=(-0.162635708202)*x[0]+(-0.996922517356)*x[1]+(0.48000598379)*x[2]
        arg[1,0,1]=(-0.49910792923)*x[0]+(1.11334748765)*x[1]+(0.0050657496159)*x[2]
        arg[1,1,0]=(-1.55972927502)*x[0]+(0.233936234238)*x[1]+(-0.546906209213)*x[2]
        arg[1,1,1]=(-0.6012330303)*x[0]+(-0.267951301884)*x[1]+(-0.193021060351)*x[2]
        arg[2,0,0]=(1.32515959423)*x[0]+(1.2070735022)*x[1]+(1.11474945521)*x[2]
        arg[2,0,1]=(-1.05500093406)*x[0]+(-0.937935240268)*x[1]+(0.180289797658)*x[2]
        arg[2,1,0]=(-0.450015559749)*x[0]+(-1.17715272891)*x[1]+(0.197160834934)*x[2]
        arg[2,1,1]=(-1.24969258786)*x[0]+(-0.300915973735)*x[1]+(-0.195058423793)*x[2]
        arg[3,0,0]=(0.704171975525)*x[0]+(1.0327384159)*x[1]+(-1.23664355084)*x[2]
        arg[3,0,1]=(0.916288412006)*x[0]+(1.6929893349)*x[1]+(-1.16682244827)*x[2]
        arg[3,1,0]=(-0.756147901025)*x[0]+(0.0232005287342)*x[1]+(0.213796289692)*x[2]
        arg[3,1,1]=(0.92799443461)*x[0]+(-0.228093116826)*x[1]+(-1.53373185211)*x[2]
        arg[4,0,0]=(-0.272514671459)*x[0]+(0.310026631492)*x[1]+(0.735127519475)*x[2]
        arg[4,0,1]=(1.23963120106)*x[0]+(-0.0769223754537)*x[1]+(0.767002455909)*x[2]
        arg[4,1,0]=(-1.1121433063)*x[0]+(0.748482838233)*x[1]+(-0.0985898502473)*x[2]
        arg[4,1,1]=(-1.08169600038)*x[0]+(-1.39575461699)*x[1]+(0.468455733883)*x[2]
        arg[5,0,0]=(-0.737100332594)*x[0]+(1.0950941569)*x[1]+(0.530144605873)*x[2]
        arg[5,0,1]=(1.94353080429)*x[0]+(1.61187736568)*x[1]+(0.887255997784)*x[2]
        arg[5,1,0]=(-0.492812782764)*x[0]+(-1.17859119673)*x[1]+(-0.627254028293)*x[2]
        arg[5,1,1]=(0.932565304399)*x[0]+(0.760125636669)*x[1]+(-0.687485104891)*x[2]
        ref[0,0,0]=(-0.289427308366)*x_ref[0]+(1.17077011703)*x_ref[1]+(0.763196501047)*x_ref[2]
        ref[0,0,1]=(0.642730500265)*x_ref[0]+(1.1719001632)*x_ref[1]+(-0.121378040894)*x_ref[2]
        ref[0,1,0]=(0.0879005621769)*x_ref[0]+(-0.627582537594)*x_ref[1]+(0.220072369447)*x_ref[2]
        ref[0,1,1]=(-0.518762071566)*x_ref[0]+(-0.700754417623)*x_ref[1]+(-0.340623426752)*x_ref[2]
        ref[1,0,0]=(-0.162635708202)*x_ref[0]+(-0.996922517356)*x_ref[1]+(0.48000598379)*x_ref[2]
        ref[1,0,1]=(-0.49910792923)*x_ref[0]+(1.11334748765)*x_ref[1]+(0.0050657496159)*x_ref[2]
        ref[1,1,0]=(-1.55972927502)*x_ref[0]+(0.233936234238)*x_ref[1]+(-0.546906209213)*x_ref[2]
        ref[1,1,1]=(-0.6012330303)*x_ref[0]+(-0.267951301884)*x_ref[1]+(-0.193021060351)*x_ref[2]
        ref[2,0,0]=(1.32515959423)*x_ref[0]+(1.2070735022)*x_ref[1]+(1.11474945521)*x_ref[2]
        ref[2,0,1]=(-1.05500093406)*x_ref[0]+(-0.937935240268)*x_ref[1]+(0.180289797658)*x_ref[2]
        ref[2,1,0]=(-0.450015559749)*x_ref[0]+(-1.17715272891)*x_ref[1]+(0.197160834934)*x_ref[2]
        ref[2,1,1]=(-1.24969258786)*x_ref[0]+(-0.300915973735)*x_ref[1]+(-0.195058423793)*x_ref[2]
        ref[3,0,0]=(0.704171975525)*x_ref[0]+(1.0327384159)*x_ref[1]+(-1.23664355084)*x_ref[2]
        ref[3,0,1]=(0.916288412006)*x_ref[0]+(1.6929893349)*x_ref[1]+(-1.16682244827)*x_ref[2]
        ref[3,1,0]=(-0.756147901025)*x_ref[0]+(0.0232005287342)*x_ref[1]+(0.213796289692)*x_ref[2]
        ref[3,1,1]=(0.92799443461)*x_ref[0]+(-0.228093116826)*x_ref[1]+(-1.53373185211)*x_ref[2]
        ref[4,0,0]=(-0.272514671459)*x_ref[0]+(0.310026631492)*x_ref[1]+(0.735127519475)*x_ref[2]
        ref[4,0,1]=(1.23963120106)*x_ref[0]+(-0.0769223754537)*x_ref[1]+(0.767002455909)*x_ref[2]
        ref[4,1,0]=(-1.1121433063)*x_ref[0]+(0.748482838233)*x_ref[1]+(-0.0985898502473)*x_ref[2]
        ref[4,1,1]=(-1.08169600038)*x_ref[0]+(-1.39575461699)*x_ref[1]+(0.468455733883)*x_ref[2]
        ref[5,0,0]=(-0.737100332594)*x_ref[0]+(1.0950941569)*x_ref[1]+(0.530144605873)*x_ref[2]
        ref[5,0,1]=(1.94353080429)*x_ref[0]+(1.61187736568)*x_ref[1]+(0.887255997784)*x_ref[2]
        ref[5,1,0]=(-0.492812782764)*x_ref[0]+(-1.17859119673)*x_ref[1]+(-0.627254028293)*x_ref[2]
        ref[5,1,1]=(0.932565304399)*x_ref[0]+(0.760125636669)*x_ref[1]+(-0.687485104891)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.261713255264)*x[0]+(-0.650225996431)*x[1]
        arg[0,0,0,1]=(0.177056727607)*x[0]+(-0.486487540393)*x[1]
        arg[0,0,1,0]=(1.38050234107)*x[0]+(0.273417897264)*x[1]
        arg[0,0,1,1]=(-1.059824929)*x[0]+(0.432893707387)*x[1]
        arg[0,0,2,0]=(1.11328992978)*x[0]+(1.14949237037)*x[1]
        arg[0,0,2,1]=(-0.176459646227)*x[0]+(-0.835980248375)*x[1]
        arg[0,1,0,0]=(0.992757237392)*x[0]+(0.113511598452)*x[1]
        arg[0,1,0,1]=(-0.211425610821)*x[0]+(-0.472765285276)*x[1]
        arg[0,1,1,0]=(-1.52498305232)*x[0]+(-0.869737083114)*x[1]
        arg[0,1,1,1]=(-0.474203117317)*x[0]+(-0.284532442353)*x[1]
        arg[0,1,2,0]=(-0.431542961546)*x[0]+(-0.0427397090453)*x[1]
        arg[0,1,2,1]=(-1.0562625269)*x[0]+(-0.74486767124)*x[1]
        arg[0,2,0,0]=(1.46553867064)*x[0]+(-0.650182195657)*x[1]
        arg[0,2,0,1]=(-0.717482663647)*x[0]+(-0.716320000777)*x[1]
        arg[0,2,1,0]=(-1.21701970743)*x[0]+(0.636901443761)*x[1]
        arg[0,2,1,1]=(0.438592928507)*x[0]+(0.812362899723)*x[1]
        arg[0,2,2,0]=(0.104885477141)*x[0]+(-0.127968978832)*x[1]
        arg[0,2,2,1]=(0.285116712849)*x[0]+(0.169740783305)*x[1]
        arg[0,3,0,0]=(-1.22825142912)*x[0]+(-0.562911687838)*x[1]
        arg[0,3,0,1]=(-0.569159534025)*x[0]+(1.30898732707)*x[1]
        arg[0,3,1,0]=(0.164709387569)*x[0]+(-1.62733582919)*x[1]
        arg[0,3,1,1]=(-0.477197320128)*x[0]+(-0.115207237191)*x[1]
        arg[0,3,2,0]=(-0.046871003317)*x[0]+(0.0371149345252)*x[1]
        arg[0,3,2,1]=(0.839624830979)*x[0]+(0.0785937909827)*x[1]
        arg[0,4,0,0]=(-0.0250725769598)*x[0]+(-0.535781603767)*x[1]
        arg[0,4,0,1]=(-0.761849740443)*x[0]+(-0.300714501152)*x[1]
        arg[0,4,1,0]=(0.0212637838833)*x[0]+(-1.56914428088)*x[1]
        arg[0,4,1,1]=(-0.254524677425)*x[0]+(0.699288623343)*x[1]
        arg[0,4,2,0]=(1.14678964272)*x[0]+(1.24738649379)*x[1]
        arg[0,4,2,1]=(1.80545916472)*x[0]+(1.22689087643)*x[1]
        arg[1,0,0,0]=(-0.116129927341)*x[0]+(-0.875165834648)*x[1]
        arg[1,0,0,1]=(-0.49905851236)*x[0]+(-1.10711214732)*x[1]
        arg[1,0,1,0]=(-1.44706167224)*x[0]+(-1.80259802262)*x[1]
        arg[1,0,1,1]=(0.555288273554)*x[0]+(0.153437302525)*x[1]
        arg[1,0,2,0]=(0.00774838569142)*x[0]+(-1.35683757658)*x[1]
        arg[1,0,2,1]=(-0.625156986286)*x[0]+(-0.201091038824)*x[1]
        arg[1,1,0,0]=(-0.765068445971)*x[0]+(1.32890117415)*x[1]
        arg[1,1,0,1]=(-0.629548919312)*x[0]+(-0.0512345513979)*x[1]
        arg[1,1,1,0]=(-0.361717423775)*x[0]+(0.286169858965)*x[1]
        arg[1,1,1,1]=(-0.533065922681)*x[0]+(0.103362397133)*x[1]
        arg[1,1,2,0]=(0.978041972715)*x[0]+(0.405949452109)*x[1]
        arg[1,1,2,1]=(-1.08418078618)*x[0]+(0.027879770152)*x[1]
        arg[1,2,0,0]=(1.23545256768)*x[0]+(0.574400116282)*x[1]
        arg[1,2,0,1]=(0.245362478488)*x[0]+(0.241577755816)*x[1]
        arg[1,2,1,0]=(0.645309443244)*x[0]+(0.0437773509036)*x[1]
        arg[1,2,1,1]=(0.00930197364105)*x[0]+(0.472462788946)*x[1]
        arg[1,2,2,0]=(1.44241534167)*x[0]+(-0.901332994844)*x[1]
        arg[1,2,2,1]=(0.893335284585)*x[0]+(0.109130146628)*x[1]
        arg[1,3,0,0]=(0.601907323675)*x[0]+(-1.54057502586)*x[1]
        arg[1,3,0,1]=(1.38086759376)*x[0]+(0.356522488945)*x[1]
        arg[1,3,1,0]=(1.16786130652)*x[0]+(-0.788756049484)*x[1]
        arg[1,3,1,1]=(-0.22580270698)*x[0]+(0.857686971693)*x[1]
        arg[1,3,2,0]=(0.709549396131)*x[0]+(-0.644330864877)*x[1]
        arg[1,3,2,1]=(1.04901965211)*x[0]+(0.0178384468958)*x[1]
        arg[1,4,0,0]=(-0.525050731412)*x[0]+(-0.253639122312)*x[1]
        arg[1,4,0,1]=(-1.63200911423)*x[0]+(0.519020443372)*x[1]
        arg[1,4,1,0]=(0.23956666878)*x[0]+(-1.76973024011)*x[1]
        arg[1,4,1,1]=(-0.412317050346)*x[0]+(-0.391711749094)*x[1]
        arg[1,4,2,0]=(1.63293176518)*x[0]+(0.583921572851)*x[1]
        arg[1,4,2,1]=(0.434893659373)*x[0]+(-0.324706930043)*x[1]
        arg[2,0,0,0]=(0.368478865639)*x[0]+(1.87637158068)*x[1]
        arg[2,0,0,1]=(1.32519020398)*x[0]+(0.306016568948)*x[1]
        arg[2,0,1,0]=(-0.284264003593)*x[0]+(-0.899022593367)*x[1]
        arg[2,0,1,1]=(-0.0439761413236)*x[0]+(-0.696386615358)*x[1]
        arg[2,0,2,0]=(0.407393812349)*x[0]+(-1.04815464317)*x[1]
        arg[2,0,2,1]=(-0.956661282071)*x[0]+(0.243281223782)*x[1]
        arg[2,1,0,0]=(-1.29201911939)*x[0]+(-0.693049706038)*x[1]
        arg[2,1,0,1]=(-0.711059848853)*x[0]+(-1.14204907272)*x[1]
        arg[2,1,1,0]=(-0.347693333057)*x[0]+(0.818879269672)*x[1]
        arg[2,1,1,1]=(0.00742807299875)*x[0]+(0.138845309984)*x[1]
        arg[2,1,2,0]=(1.13412244048)*x[0]+(-0.481422214102)*x[1]
        arg[2,1,2,1]=(0.64440302826)*x[0]+(-0.955231893089)*x[1]
        arg[2,2,0,0]=(0.13123141544)*x[0]+(-0.602266537541)*x[1]
        arg[2,2,0,1]=(-0.204818071763)*x[0]+(-1.13480248293)*x[1]
        arg[2,2,1,0]=(0.781058287579)*x[0]+(-1.31936224435)*x[1]
        arg[2,2,1,1]=(-0.591275351662)*x[0]+(0.691816969517)*x[1]
        arg[2,2,2,0]=(-0.921964826241)*x[0]+(1.28687075541)*x[1]
        arg[2,2,2,1]=(0.246260462337)*x[0]+(-1.04699282194)*x[1]
        arg[2,3,0,0]=(0.264209126513)*x[0]+(-0.697094400566)*x[1]
        arg[2,3,0,1]=(0.456158298484)*x[0]+(0.0288036445565)*x[1]
        arg[2,3,1,0]=(-0.720337925392)*x[0]+(0.0965182642186)*x[1]
        arg[2,3,1,1]=(0.397480419547)*x[0]+(-1.55232798344)*x[1]
        arg[2,3,2,0]=(-0.337139999283)*x[0]+(0.918809751807)*x[1]
        arg[2,3,2,1]=(1.34009447198)*x[0]+(-1.41390962018)*x[1]
        arg[2,4,0,0]=(1.67781838006)*x[0]+(0.521276193814)*x[1]
        arg[2,4,0,1]=(-0.114014321939)*x[0]+(0.529316778773)*x[1]
        arg[2,4,1,0]=(-0.19938131807)*x[0]+(-0.0712417614843)*x[1]
        arg[2,4,1,1]=(-0.724676753871)*x[0]+(0.731341170171)*x[1]
        arg[2,4,2,0]=(0.988755659656)*x[0]+(-1.0327352189)*x[1]
        arg[2,4,2,1]=(-0.526286732859)*x[0]+(-1.18295671598)*x[1]
        arg[3,0,0,0]=(-0.685223304723)*x[0]+(0.508431586068)*x[1]
        arg[3,0,0,1]=(0.529240274944)*x[0]+(0.77519128969)*x[1]
        arg[3,0,1,0]=(-0.757098431928)*x[0]+(1.41375812725)*x[1]
        arg[3,0,1,1]=(1.20805599335)*x[0]+(-0.184660334182)*x[1]
        arg[3,0,2,0]=(0.718413765561)*x[0]+(0.971149168706)*x[1]
        arg[3,0,2,1]=(-0.632670534518)*x[0]+(0.0389618967363)*x[1]
        arg[3,1,0,0]=(-0.525814673755)*x[0]+(0.574122587598)*x[1]
        arg[3,1,0,1]=(-1.04117288975)*x[0]+(-1.00738089823)*x[1]
        arg[3,1,1,0]=(0.0312806064182)*x[0]+(-0.30773511847)*x[1]
        arg[3,1,1,1]=(0.889930889355)*x[0]+(0.833689153489)*x[1]
        arg[3,1,2,0]=(-1.21586293381)*x[0]+(0.512182349574)*x[1]
        arg[3,1,2,1]=(-0.00437508755415)*x[0]+(-1.13228664497)*x[1]
        arg[3,2,0,0]=(1.64174174554)*x[0]+(-0.696500062562)*x[1]
        arg[3,2,0,1]=(0.0265343210604)*x[0]+(-0.707378716636)*x[1]
        arg[3,2,1,0]=(0.952922122967)*x[0]+(0.0946904561805)*x[1]
        arg[3,2,1,1]=(-1.06905266456)*x[0]+(-0.155122789807)*x[1]
        arg[3,2,2,0]=(-1.14410789976)*x[0]+(-0.379020164446)*x[1]
        arg[3,2,2,1]=(-0.0922611198553)*x[0]+(1.36680699099)*x[1]
        arg[3,3,0,0]=(0.849028205422)*x[0]+(-0.968698293811)*x[1]
        arg[3,3,0,1]=(0.649894718675)*x[0]+(0.238249259292)*x[1]
        arg[3,3,1,0]=(1.28170508471)*x[0]+(-0.736544947537)*x[1]
        arg[3,3,1,1]=(-0.188411685414)*x[0]+(1.14145010691)*x[1]
        arg[3,3,2,0]=(-1.13973037849)*x[0]+(-0.256565914821)*x[1]
        arg[3,3,2,1]=(0.955478585358)*x[0]+(-0.83417326515)*x[1]
        arg[3,4,0,0]=(0.155710789468)*x[0]+(0.204956811648)*x[1]
        arg[3,4,0,1]=(0.567866952974)*x[0]+(-0.168398320099)*x[1]
        arg[3,4,1,0]=(-1.36997367)*x[0]+(0.141984040172)*x[1]
        arg[3,4,1,1]=(-0.809532285206)*x[0]+(0.70611485405)*x[1]
        arg[3,4,2,0]=(-0.453260232051)*x[0]+(0.0665127865858)*x[1]
        arg[3,4,2,1]=(0.00142021115277)*x[0]+(1.02307505343)*x[1]
        ref[0,0,0,0]=(-0.261713255264)*x_ref[0]+(-0.650225996431)*x_ref[1]
        ref[0,0,0,1]=(0.177056727607)*x_ref[0]+(-0.486487540393)*x_ref[1]
        ref[0,0,1,0]=(1.38050234107)*x_ref[0]+(0.273417897264)*x_ref[1]
        ref[0,0,1,1]=(-1.059824929)*x_ref[0]+(0.432893707387)*x_ref[1]
        ref[0,0,2,0]=(1.11328992978)*x_ref[0]+(1.14949237037)*x_ref[1]
        ref[0,0,2,1]=(-0.176459646227)*x_ref[0]+(-0.835980248375)*x_ref[1]
        ref[0,1,0,0]=(0.992757237392)*x_ref[0]+(0.113511598452)*x_ref[1]
        ref[0,1,0,1]=(-0.211425610821)*x_ref[0]+(-0.472765285276)*x_ref[1]
        ref[0,1,1,0]=(-1.52498305232)*x_ref[0]+(-0.869737083114)*x_ref[1]
        ref[0,1,1,1]=(-0.474203117317)*x_ref[0]+(-0.284532442353)*x_ref[1]
        ref[0,1,2,0]=(-0.431542961546)*x_ref[0]+(-0.0427397090453)*x_ref[1]
        ref[0,1,2,1]=(-1.0562625269)*x_ref[0]+(-0.74486767124)*x_ref[1]
        ref[0,2,0,0]=(1.46553867064)*x_ref[0]+(-0.650182195657)*x_ref[1]
        ref[0,2,0,1]=(-0.717482663647)*x_ref[0]+(-0.716320000777)*x_ref[1]
        ref[0,2,1,0]=(-1.21701970743)*x_ref[0]+(0.636901443761)*x_ref[1]
        ref[0,2,1,1]=(0.438592928507)*x_ref[0]+(0.812362899723)*x_ref[1]
        ref[0,2,2,0]=(0.104885477141)*x_ref[0]+(-0.127968978832)*x_ref[1]
        ref[0,2,2,1]=(0.285116712849)*x_ref[0]+(0.169740783305)*x_ref[1]
        ref[0,3,0,0]=(-1.22825142912)*x_ref[0]+(-0.562911687838)*x_ref[1]
        ref[0,3,0,1]=(-0.569159534025)*x_ref[0]+(1.30898732707)*x_ref[1]
        ref[0,3,1,0]=(0.164709387569)*x_ref[0]+(-1.62733582919)*x_ref[1]
        ref[0,3,1,1]=(-0.477197320128)*x_ref[0]+(-0.115207237191)*x_ref[1]
        ref[0,3,2,0]=(-0.046871003317)*x_ref[0]+(0.0371149345252)*x_ref[1]
        ref[0,3,2,1]=(0.839624830979)*x_ref[0]+(0.0785937909827)*x_ref[1]
        ref[0,4,0,0]=(-0.0250725769598)*x_ref[0]+(-0.535781603767)*x_ref[1]
        ref[0,4,0,1]=(-0.761849740443)*x_ref[0]+(-0.300714501152)*x_ref[1]
        ref[0,4,1,0]=(0.0212637838833)*x_ref[0]+(-1.56914428088)*x_ref[1]
        ref[0,4,1,1]=(-0.254524677425)*x_ref[0]+(0.699288623343)*x_ref[1]
        ref[0,4,2,0]=(1.14678964272)*x_ref[0]+(1.24738649379)*x_ref[1]
        ref[0,4,2,1]=(1.80545916472)*x_ref[0]+(1.22689087643)*x_ref[1]
        ref[1,0,0,0]=(-0.116129927341)*x_ref[0]+(-0.875165834648)*x_ref[1]
        ref[1,0,0,1]=(-0.49905851236)*x_ref[0]+(-1.10711214732)*x_ref[1]
        ref[1,0,1,0]=(-1.44706167224)*x_ref[0]+(-1.80259802262)*x_ref[1]
        ref[1,0,1,1]=(0.555288273554)*x_ref[0]+(0.153437302525)*x_ref[1]
        ref[1,0,2,0]=(0.00774838569142)*x_ref[0]+(-1.35683757658)*x_ref[1]
        ref[1,0,2,1]=(-0.625156986286)*x_ref[0]+(-0.201091038824)*x_ref[1]
        ref[1,1,0,0]=(-0.765068445971)*x_ref[0]+(1.32890117415)*x_ref[1]
        ref[1,1,0,1]=(-0.629548919312)*x_ref[0]+(-0.0512345513979)*x_ref[1]
        ref[1,1,1,0]=(-0.361717423775)*x_ref[0]+(0.286169858965)*x_ref[1]
        ref[1,1,1,1]=(-0.533065922681)*x_ref[0]+(0.103362397133)*x_ref[1]
        ref[1,1,2,0]=(0.978041972715)*x_ref[0]+(0.405949452109)*x_ref[1]
        ref[1,1,2,1]=(-1.08418078618)*x_ref[0]+(0.027879770152)*x_ref[1]
        ref[1,2,0,0]=(1.23545256768)*x_ref[0]+(0.574400116282)*x_ref[1]
        ref[1,2,0,1]=(0.245362478488)*x_ref[0]+(0.241577755816)*x_ref[1]
        ref[1,2,1,0]=(0.645309443244)*x_ref[0]+(0.0437773509036)*x_ref[1]
        ref[1,2,1,1]=(0.00930197364105)*x_ref[0]+(0.472462788946)*x_ref[1]
        ref[1,2,2,0]=(1.44241534167)*x_ref[0]+(-0.901332994844)*x_ref[1]
        ref[1,2,2,1]=(0.893335284585)*x_ref[0]+(0.109130146628)*x_ref[1]
        ref[1,3,0,0]=(0.601907323675)*x_ref[0]+(-1.54057502586)*x_ref[1]
        ref[1,3,0,1]=(1.38086759376)*x_ref[0]+(0.356522488945)*x_ref[1]
        ref[1,3,1,0]=(1.16786130652)*x_ref[0]+(-0.788756049484)*x_ref[1]
        ref[1,3,1,1]=(-0.22580270698)*x_ref[0]+(0.857686971693)*x_ref[1]
        ref[1,3,2,0]=(0.709549396131)*x_ref[0]+(-0.644330864877)*x_ref[1]
        ref[1,3,2,1]=(1.04901965211)*x_ref[0]+(0.0178384468958)*x_ref[1]
        ref[1,4,0,0]=(-0.525050731412)*x_ref[0]+(-0.253639122312)*x_ref[1]
        ref[1,4,0,1]=(-1.63200911423)*x_ref[0]+(0.519020443372)*x_ref[1]
        ref[1,4,1,0]=(0.23956666878)*x_ref[0]+(-1.76973024011)*x_ref[1]
        ref[1,4,1,1]=(-0.412317050346)*x_ref[0]+(-0.391711749094)*x_ref[1]
        ref[1,4,2,0]=(1.63293176518)*x_ref[0]+(0.583921572851)*x_ref[1]
        ref[1,4,2,1]=(0.434893659373)*x_ref[0]+(-0.324706930043)*x_ref[1]
        ref[2,0,0,0]=(0.368478865639)*x_ref[0]+(1.87637158068)*x_ref[1]
        ref[2,0,0,1]=(1.32519020398)*x_ref[0]+(0.306016568948)*x_ref[1]
        ref[2,0,1,0]=(-0.284264003593)*x_ref[0]+(-0.899022593367)*x_ref[1]
        ref[2,0,1,1]=(-0.0439761413236)*x_ref[0]+(-0.696386615358)*x_ref[1]
        ref[2,0,2,0]=(0.407393812349)*x_ref[0]+(-1.04815464317)*x_ref[1]
        ref[2,0,2,1]=(-0.956661282071)*x_ref[0]+(0.243281223782)*x_ref[1]
        ref[2,1,0,0]=(-1.29201911939)*x_ref[0]+(-0.693049706038)*x_ref[1]
        ref[2,1,0,1]=(-0.711059848853)*x_ref[0]+(-1.14204907272)*x_ref[1]
        ref[2,1,1,0]=(-0.347693333057)*x_ref[0]+(0.818879269672)*x_ref[1]
        ref[2,1,1,1]=(0.00742807299875)*x_ref[0]+(0.138845309984)*x_ref[1]
        ref[2,1,2,0]=(1.13412244048)*x_ref[0]+(-0.481422214102)*x_ref[1]
        ref[2,1,2,1]=(0.64440302826)*x_ref[0]+(-0.955231893089)*x_ref[1]
        ref[2,2,0,0]=(0.13123141544)*x_ref[0]+(-0.602266537541)*x_ref[1]
        ref[2,2,0,1]=(-0.204818071763)*x_ref[0]+(-1.13480248293)*x_ref[1]
        ref[2,2,1,0]=(0.781058287579)*x_ref[0]+(-1.31936224435)*x_ref[1]
        ref[2,2,1,1]=(-0.591275351662)*x_ref[0]+(0.691816969517)*x_ref[1]
        ref[2,2,2,0]=(-0.921964826241)*x_ref[0]+(1.28687075541)*x_ref[1]
        ref[2,2,2,1]=(0.246260462337)*x_ref[0]+(-1.04699282194)*x_ref[1]
        ref[2,3,0,0]=(0.264209126513)*x_ref[0]+(-0.697094400566)*x_ref[1]
        ref[2,3,0,1]=(0.456158298484)*x_ref[0]+(0.0288036445565)*x_ref[1]
        ref[2,3,1,0]=(-0.720337925392)*x_ref[0]+(0.0965182642186)*x_ref[1]
        ref[2,3,1,1]=(0.397480419547)*x_ref[0]+(-1.55232798344)*x_ref[1]
        ref[2,3,2,0]=(-0.337139999283)*x_ref[0]+(0.918809751807)*x_ref[1]
        ref[2,3,2,1]=(1.34009447198)*x_ref[0]+(-1.41390962018)*x_ref[1]
        ref[2,4,0,0]=(1.67781838006)*x_ref[0]+(0.521276193814)*x_ref[1]
        ref[2,4,0,1]=(-0.114014321939)*x_ref[0]+(0.529316778773)*x_ref[1]
        ref[2,4,1,0]=(-0.19938131807)*x_ref[0]+(-0.0712417614843)*x_ref[1]
        ref[2,4,1,1]=(-0.724676753871)*x_ref[0]+(0.731341170171)*x_ref[1]
        ref[2,4,2,0]=(0.988755659656)*x_ref[0]+(-1.0327352189)*x_ref[1]
        ref[2,4,2,1]=(-0.526286732859)*x_ref[0]+(-1.18295671598)*x_ref[1]
        ref[3,0,0,0]=(-0.685223304723)*x_ref[0]+(0.508431586068)*x_ref[1]
        ref[3,0,0,1]=(0.529240274944)*x_ref[0]+(0.77519128969)*x_ref[1]
        ref[3,0,1,0]=(-0.757098431928)*x_ref[0]+(1.41375812725)*x_ref[1]
        ref[3,0,1,1]=(1.20805599335)*x_ref[0]+(-0.184660334182)*x_ref[1]
        ref[3,0,2,0]=(0.718413765561)*x_ref[0]+(0.971149168706)*x_ref[1]
        ref[3,0,2,1]=(-0.632670534518)*x_ref[0]+(0.0389618967363)*x_ref[1]
        ref[3,1,0,0]=(-0.525814673755)*x_ref[0]+(0.574122587598)*x_ref[1]
        ref[3,1,0,1]=(-1.04117288975)*x_ref[0]+(-1.00738089823)*x_ref[1]
        ref[3,1,1,0]=(0.0312806064182)*x_ref[0]+(-0.30773511847)*x_ref[1]
        ref[3,1,1,1]=(0.889930889355)*x_ref[0]+(0.833689153489)*x_ref[1]
        ref[3,1,2,0]=(-1.21586293381)*x_ref[0]+(0.512182349574)*x_ref[1]
        ref[3,1,2,1]=(-0.00437508755415)*x_ref[0]+(-1.13228664497)*x_ref[1]
        ref[3,2,0,0]=(1.64174174554)*x_ref[0]+(-0.696500062562)*x_ref[1]
        ref[3,2,0,1]=(0.0265343210604)*x_ref[0]+(-0.707378716636)*x_ref[1]
        ref[3,2,1,0]=(0.952922122967)*x_ref[0]+(0.0946904561805)*x_ref[1]
        ref[3,2,1,1]=(-1.06905266456)*x_ref[0]+(-0.155122789807)*x_ref[1]
        ref[3,2,2,0]=(-1.14410789976)*x_ref[0]+(-0.379020164446)*x_ref[1]
        ref[3,2,2,1]=(-0.0922611198553)*x_ref[0]+(1.36680699099)*x_ref[1]
        ref[3,3,0,0]=(0.849028205422)*x_ref[0]+(-0.968698293811)*x_ref[1]
        ref[3,3,0,1]=(0.649894718675)*x_ref[0]+(0.238249259292)*x_ref[1]
        ref[3,3,1,0]=(1.28170508471)*x_ref[0]+(-0.736544947537)*x_ref[1]
        ref[3,3,1,1]=(-0.188411685414)*x_ref[0]+(1.14145010691)*x_ref[1]
        ref[3,3,2,0]=(-1.13973037849)*x_ref[0]+(-0.256565914821)*x_ref[1]
        ref[3,3,2,1]=(0.955478585358)*x_ref[0]+(-0.83417326515)*x_ref[1]
        ref[3,4,0,0]=(0.155710789468)*x_ref[0]+(0.204956811648)*x_ref[1]
        ref[3,4,0,1]=(0.567866952974)*x_ref[0]+(-0.168398320099)*x_ref[1]
        ref[3,4,1,0]=(-1.36997367)*x_ref[0]+(0.141984040172)*x_ref[1]
        ref[3,4,1,1]=(-0.809532285206)*x_ref[0]+(0.70611485405)*x_ref[1]
        ref[3,4,2,0]=(-0.453260232051)*x_ref[0]+(0.0665127865858)*x_ref[1]
        ref[3,4,2,1]=(0.00142021115277)*x_ref[0]+(1.02307505343)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.59037121238)*x[0]+(-0.432259357387)*x[1]+(1.06313464506)*x[2]
        arg[0,0,0,1]=(-0.29317617186)*x[0]+(-0.477878024138)*x[1]+(-0.943279870083)*x[2]
        arg[0,0,1,0]=(-0.0140970955833)*x[0]+(0.141752937281)*x[1]+(0.46748311988)*x[2]
        arg[0,0,1,1]=(1.06645552756)*x[0]+(-1.4360320823)*x[1]+(-1.05643251873)*x[2]
        arg[0,0,2,0]=(-0.527854671619)*x[0]+(0.68918711423)*x[1]+(0.225750468582)*x[2]
        arg[0,0,2,1]=(0.765285169635)*x[0]+(-0.292098568154)*x[1]+(-0.611702446821)*x[2]
        arg[0,1,0,0]=(0.472029729041)*x[0]+(-0.391383508709)*x[1]+(-0.0882199174494)*x[2]
        arg[0,1,0,1]=(1.76010861737)*x[0]+(1.40668629263)*x[1]+(-0.171024087043)*x[2]
        arg[0,1,1,0]=(0.251929705017)*x[0]+(-0.326420063624)*x[1]+(-0.337849942832)*x[2]
        arg[0,1,1,1]=(0.049289902392)*x[0]+(1.61775987596)*x[1]+(0.545741506198)*x[2]
        arg[0,1,2,0]=(-0.500710333189)*x[0]+(-0.551222136922)*x[1]+(0.665856619864)*x[2]
        arg[0,1,2,1]=(0.271547935249)*x[0]+(0.46154733209)*x[1]+(0.0776993672589)*x[2]
        arg[0,2,0,0]=(0.46744023541)*x[0]+(0.281754164985)*x[1]+(0.11721465855)*x[2]
        arg[0,2,0,1]=(0.58865724402)*x[0]+(-0.280911783535)*x[1]+(0.793461201474)*x[2]
        arg[0,2,1,0]=(-0.348215254122)*x[0]+(-0.274357607001)*x[1]+(-0.975813503666)*x[2]
        arg[0,2,1,1]=(-0.546269224094)*x[0]+(0.54132986357)*x[1]+(-0.346797273682)*x[2]
        arg[0,2,2,0]=(0.956395362583)*x[0]+(-0.609333787208)*x[1]+(-0.86408067145)*x[2]
        arg[0,2,2,1]=(0.570075324955)*x[0]+(-0.251638346903)*x[1]+(0.292841304398)*x[2]
        arg[0,3,0,0]=(-0.240717117869)*x[0]+(0.571307481772)*x[1]+(0.66826910816)*x[2]
        arg[0,3,0,1]=(0.113871023723)*x[0]+(0.022498355636)*x[1]+(0.95331234379)*x[2]
        arg[0,3,1,0]=(-0.264946899001)*x[0]+(1.08412373438)*x[1]+(-1.10548347411)*x[2]
        arg[0,3,1,1]=(-0.744510132854)*x[0]+(-0.716638983187)*x[1]+(0.64330838226)*x[2]
        arg[0,3,2,0]=(1.15237343787)*x[0]+(-1.42565901229)*x[1]+(0.727585401396)*x[2]
        arg[0,3,2,1]=(-0.251464863627)*x[0]+(1.4067624912)*x[1]+(0.674418739118)*x[2]
        arg[0,4,0,0]=(-0.00551820316191)*x[0]+(0.193862853101)*x[1]+(1.8004924522)*x[2]
        arg[0,4,0,1]=(-0.985879358916)*x[0]+(0.476965815991)*x[1]+(0.96363449822)*x[2]
        arg[0,4,1,0]=(-0.657967361584)*x[0]+(0.833452284341)*x[1]+(-1.40128003822)*x[2]
        arg[0,4,1,1]=(0.301379967988)*x[0]+(0.0474545542353)*x[1]+(-0.633087920569)*x[2]
        arg[0,4,2,0]=(-1.94409755409)*x[0]+(-0.468638363928)*x[1]+(-0.713880237452)*x[2]
        arg[0,4,2,1]=(-1.13536316065)*x[0]+(-0.250761280127)*x[1]+(0.328722833182)*x[2]
        arg[1,0,0,0]=(0.556886266554)*x[0]+(-0.70523094638)*x[1]+(-0.333621268224)*x[2]
        arg[1,0,0,1]=(-0.995565683628)*x[0]+(-0.158120855239)*x[1]+(0.0738870470169)*x[2]
        arg[1,0,1,0]=(-0.209556712574)*x[0]+(0.54393408711)*x[1]+(-1.1617498966)*x[2]
        arg[1,0,1,1]=(0.666809649924)*x[0]+(1.18555984779)*x[1]+(-0.442831598796)*x[2]
        arg[1,0,2,0]=(-0.21354395591)*x[0]+(-0.634626698269)*x[1]+(1.25643757008)*x[2]
        arg[1,0,2,1]=(-0.676259187073)*x[0]+(-0.458532774397)*x[1]+(0.271156919902)*x[2]
        arg[1,1,0,0]=(-1.45529139262)*x[0]+(0.311382249732)*x[1]+(-0.134620909882)*x[2]
        arg[1,1,0,1]=(-1.47708829198)*x[0]+(0.968291536601)*x[1]+(-0.566606887608)*x[2]
        arg[1,1,1,0]=(1.18904329045)*x[0]+(-0.209072247061)*x[1]+(-0.26261457093)*x[2]
        arg[1,1,1,1]=(0.313723817215)*x[0]+(-0.169177309423)*x[1]+(-0.00119161234247)*x[2]
        arg[1,1,2,0]=(0.431298197457)*x[0]+(0.264113678484)*x[1]+(0.192619025313)*x[2]
        arg[1,1,2,1]=(0.0469221259627)*x[0]+(1.53477696238)*x[1]+(0.00794662724457)*x[2]
        arg[1,2,0,0]=(0.102179220279)*x[0]+(-0.391365988191)*x[1]+(0.302910557263)*x[2]
        arg[1,2,0,1]=(0.841803002697)*x[0]+(0.581539244301)*x[1]+(0.00409108885792)*x[2]
        arg[1,2,1,0]=(0.0840378161144)*x[0]+(-0.431329807551)*x[1]+(-1.0774029159)*x[2]
        arg[1,2,1,1]=(-0.270803943532)*x[0]+(0.0124514531652)*x[1]+(0.197735227059)*x[2]
        arg[1,2,2,0]=(0.479336250918)*x[0]+(0.0435331360815)*x[1]+(0.458048853417)*x[2]
        arg[1,2,2,1]=(-0.59916419564)*x[0]+(0.415546807791)*x[1]+(-1.12105410423)*x[2]
        arg[1,3,0,0]=(0.378277354507)*x[0]+(1.00734343779)*x[1]+(-0.282811389334)*x[2]
        arg[1,3,0,1]=(0.717420487474)*x[0]+(-0.730395427407)*x[1]+(1.63850706843)*x[2]
        arg[1,3,1,0]=(-0.0126954991872)*x[0]+(-1.35916858377)*x[1]+(-0.955731750212)*x[2]
        arg[1,3,1,1]=(-0.560951559862)*x[0]+(0.271085100351)*x[1]+(0.60767567276)*x[2]
        arg[1,3,2,0]=(-0.73286080289)*x[0]+(-1.12018839274)*x[1]+(0.365586479937)*x[2]
        arg[1,3,2,1]=(-0.495391543232)*x[0]+(-0.404605421036)*x[1]+(1.21261493791)*x[2]
        arg[1,4,0,0]=(-0.322147956238)*x[0]+(0.278257549185)*x[1]+(-0.927648624956)*x[2]
        arg[1,4,0,1]=(1.23243511737)*x[0]+(1.38355111729)*x[1]+(0.0813739945129)*x[2]
        arg[1,4,1,0]=(0.668384291635)*x[0]+(0.995328961231)*x[1]+(0.803009543307)*x[2]
        arg[1,4,1,1]=(1.42643018603)*x[0]+(0.940855692156)*x[1]+(0.277556451163)*x[2]
        arg[1,4,2,0]=(0.489588226757)*x[0]+(-1.63608809124)*x[1]+(-0.262918068522)*x[2]
        arg[1,4,2,1]=(-1.06922340601)*x[0]+(-0.0241938384008)*x[1]+(-0.244107290038)*x[2]
        arg[2,0,0,0]=(-0.556338154459)*x[0]+(0.12071126785)*x[1]+(0.11215804852)*x[2]
        arg[2,0,0,1]=(-0.479017339774)*x[0]+(-0.530313718209)*x[1]+(0.490083551372)*x[2]
        arg[2,0,1,0]=(0.731269264444)*x[0]+(0.245793038193)*x[1]+(0.0479833912422)*x[2]
        arg[2,0,1,1]=(-0.286189382681)*x[0]+(0.601756275554)*x[1]+(-0.559334203137)*x[2]
        arg[2,0,2,0]=(-0.876495720673)*x[0]+(0.592513223261)*x[1]+(0.844551261652)*x[2]
        arg[2,0,2,1]=(1.22915142187)*x[0]+(0.136223623577)*x[1]+(0.88656874029)*x[2]
        arg[2,1,0,0]=(0.542222650916)*x[0]+(-0.709719163926)*x[1]+(0.262207188498)*x[2]
        arg[2,1,0,1]=(-0.208896980258)*x[0]+(1.33968890395)*x[1]+(-0.599621090138)*x[2]
        arg[2,1,1,0]=(-0.171147406604)*x[0]+(0.332227749427)*x[1]+(1.15043521586)*x[2]
        arg[2,1,1,1]=(0.571607567027)*x[0]+(0.1367434941)*x[1]+(1.47797159814)*x[2]
        arg[2,1,2,0]=(0.839845574634)*x[0]+(0.732437808234)*x[1]+(-0.0311071654701)*x[2]
        arg[2,1,2,1]=(1.45193213653)*x[0]+(-0.0192312612414)*x[1]+(0.308521161282)*x[2]
        arg[2,2,0,0]=(-0.122755384028)*x[0]+(0.133510038844)*x[1]+(0.837213365425)*x[2]
        arg[2,2,0,1]=(-0.411967923958)*x[0]+(-0.955489021071)*x[1]+(0.269005602186)*x[2]
        arg[2,2,1,0]=(-1.19993089382)*x[0]+(0.276541296246)*x[1]+(0.523388479396)*x[2]
        arg[2,2,1,1]=(0.0586514653869)*x[0]+(-0.624975232742)*x[1]+(-1.55738876751)*x[2]
        arg[2,2,2,0]=(-0.892065938887)*x[0]+(-0.4848241954)*x[1]+(0.315463883511)*x[2]
        arg[2,2,2,1]=(0.0272963181118)*x[0]+(0.0187713454669)*x[1]+(-1.06524661854)*x[2]
        arg[2,3,0,0]=(-1.44099287886)*x[0]+(0.495718421075)*x[1]+(0.100401960223)*x[2]
        arg[2,3,0,1]=(0.171175119412)*x[0]+(-1.34795558607)*x[1]+(-0.361766653264)*x[2]
        arg[2,3,1,0]=(-1.80591882588)*x[0]+(-0.511459455136)*x[1]+(-0.491310264055)*x[2]
        arg[2,3,1,1]=(-0.419508934622)*x[0]+(-1.12084042844)*x[1]+(-0.086947672548)*x[2]
        arg[2,3,2,0]=(0.492388303934)*x[0]+(-0.623505828597)*x[1]+(1.6965873745)*x[2]
        arg[2,3,2,1]=(1.59203810356)*x[0]+(0.36918142231)*x[1]+(-1.90067115052)*x[2]
        arg[2,4,0,0]=(1.00875649298)*x[0]+(-1.43246384934)*x[1]+(-0.51811906593)*x[2]
        arg[2,4,0,1]=(0.260118918034)*x[0]+(-1.57606061865)*x[1]+(-0.79257165655)*x[2]
        arg[2,4,1,0]=(0.954414074467)*x[0]+(-0.230132876613)*x[1]+(0.0956378957424)*x[2]
        arg[2,4,1,1]=(0.958843806364)*x[0]+(-0.155804429995)*x[1]+(-1.02904877039)*x[2]
        arg[2,4,2,0]=(0.252668700811)*x[0]+(0.164829885749)*x[1]+(0.747032722421)*x[2]
        arg[2,4,2,1]=(1.96393052323)*x[0]+(-0.821363958807)*x[1]+(-1.01663555063)*x[2]
        arg[3,0,0,0]=(1.58086037053)*x[0]+(-1.66495295611)*x[1]+(-0.510526191383)*x[2]
        arg[3,0,0,1]=(-0.931871976029)*x[0]+(1.33102205125)*x[1]+(-0.95170580468)*x[2]
        arg[3,0,1,0]=(0.99825404115)*x[0]+(0.915274413552)*x[1]+(1.5057227674)*x[2]
        arg[3,0,1,1]=(0.0485551924989)*x[0]+(-1.0877405758)*x[1]+(-0.887342573702)*x[2]
        arg[3,0,2,0]=(-0.745146140677)*x[0]+(-0.0108141266015)*x[1]+(0.194597576308)*x[2]
        arg[3,0,2,1]=(-0.935029218338)*x[0]+(0.726322809025)*x[1]+(0.57508467103)*x[2]
        arg[3,1,0,0]=(0.115668237156)*x[0]+(0.0970255991906)*x[1]+(-0.227291123547)*x[2]
        arg[3,1,0,1]=(0.167722335863)*x[0]+(-0.552168424291)*x[1]+(1.36293027463)*x[2]
        arg[3,1,1,0]=(-1.05781878028)*x[0]+(-0.0429389047972)*x[1]+(-0.257166337509)*x[2]
        arg[3,1,1,1]=(0.129148486118)*x[0]+(-0.0209434899369)*x[1]+(0.253153005253)*x[2]
        arg[3,1,2,0]=(1.3236615365)*x[0]+(-1.18164469224)*x[1]+(0.698012697426)*x[2]
        arg[3,1,2,1]=(1.10438661886)*x[0]+(-0.852903948395)*x[1]+(1.40365162024)*x[2]
        arg[3,2,0,0]=(-0.949308009957)*x[0]+(1.01993206161)*x[1]+(-0.821721697476)*x[2]
        arg[3,2,0,1]=(0.326450662787)*x[0]+(0.395096088107)*x[1]+(-0.368904800584)*x[2]
        arg[3,2,1,0]=(0.13107780566)*x[0]+(-0.342904134623)*x[1]+(0.00835087701185)*x[2]
        arg[3,2,1,1]=(-0.500694550647)*x[0]+(1.38322390003)*x[1]+(0.0576453183412)*x[2]
        arg[3,2,2,0]=(-1.32014489811)*x[0]+(-0.444079317864)*x[1]+(1.1135946018)*x[2]
        arg[3,2,2,1]=(-0.19593105007)*x[0]+(-0.00224390969377)*x[1]+(-0.73869338942)*x[2]
        arg[3,3,0,0]=(0.0151201622255)*x[0]+(-1.10873788153)*x[1]+(0.189116254832)*x[2]
        arg[3,3,0,1]=(1.05996205566)*x[0]+(0.398465639867)*x[1]+(-0.319773104448)*x[2]
        arg[3,3,1,0]=(0.687612827243)*x[0]+(1.10970823257)*x[1]+(1.02543288975)*x[2]
        arg[3,3,1,1]=(1.27258509981)*x[0]+(-1.7060826391)*x[1]+(-0.84948812227)*x[2]
        arg[3,3,2,0]=(1.2343036886)*x[0]+(0.521090253023)*x[1]+(-1.13377412954)*x[2]
        arg[3,3,2,1]=(-0.65568738385)*x[0]+(1.00354617712)*x[1]+(0.836719015812)*x[2]
        arg[3,4,0,0]=(-0.545324078594)*x[0]+(0.452747782208)*x[1]+(-1.59718954237)*x[2]
        arg[3,4,0,1]=(1.71651153317)*x[0]+(-0.161984538787)*x[1]+(-0.377244827904)*x[2]
        arg[3,4,1,0]=(-0.284033324558)*x[0]+(-0.405941433062)*x[1]+(0.824573264594)*x[2]
        arg[3,4,1,1]=(-0.112542702803)*x[0]+(-0.470814805786)*x[1]+(-0.827427436576)*x[2]
        arg[3,4,2,0]=(-0.763655132981)*x[0]+(-0.442818775864)*x[1]+(0.809814510041)*x[2]
        arg[3,4,2,1]=(-0.33794306298)*x[0]+(1.91169591332)*x[1]+(-0.487268088196)*x[2]
        ref[0,0,0,0]=(-0.59037121238)*x_ref[0]+(-0.432259357387)*x_ref[1]+(1.06313464506)*x_ref[2]
        ref[0,0,0,1]=(-0.29317617186)*x_ref[0]+(-0.477878024138)*x_ref[1]+(-0.943279870083)*x_ref[2]
        ref[0,0,1,0]=(-0.0140970955833)*x_ref[0]+(0.141752937281)*x_ref[1]+(0.46748311988)*x_ref[2]
        ref[0,0,1,1]=(1.06645552756)*x_ref[0]+(-1.4360320823)*x_ref[1]+(-1.05643251873)*x_ref[2]
        ref[0,0,2,0]=(-0.527854671619)*x_ref[0]+(0.68918711423)*x_ref[1]+(0.225750468582)*x_ref[2]
        ref[0,0,2,1]=(0.765285169635)*x_ref[0]+(-0.292098568154)*x_ref[1]+(-0.611702446821)*x_ref[2]
        ref[0,1,0,0]=(0.472029729041)*x_ref[0]+(-0.391383508709)*x_ref[1]+(-0.0882199174494)*x_ref[2]
        ref[0,1,0,1]=(1.76010861737)*x_ref[0]+(1.40668629263)*x_ref[1]+(-0.171024087043)*x_ref[2]
        ref[0,1,1,0]=(0.251929705017)*x_ref[0]+(-0.326420063624)*x_ref[1]+(-0.337849942832)*x_ref[2]
        ref[0,1,1,1]=(0.049289902392)*x_ref[0]+(1.61775987596)*x_ref[1]+(0.545741506198)*x_ref[2]
        ref[0,1,2,0]=(-0.500710333189)*x_ref[0]+(-0.551222136922)*x_ref[1]+(0.665856619864)*x_ref[2]
        ref[0,1,2,1]=(0.271547935249)*x_ref[0]+(0.46154733209)*x_ref[1]+(0.0776993672589)*x_ref[2]
        ref[0,2,0,0]=(0.46744023541)*x_ref[0]+(0.281754164985)*x_ref[1]+(0.11721465855)*x_ref[2]
        ref[0,2,0,1]=(0.58865724402)*x_ref[0]+(-0.280911783535)*x_ref[1]+(0.793461201474)*x_ref[2]
        ref[0,2,1,0]=(-0.348215254122)*x_ref[0]+(-0.274357607001)*x_ref[1]+(-0.975813503666)*x_ref[2]
        ref[0,2,1,1]=(-0.546269224094)*x_ref[0]+(0.54132986357)*x_ref[1]+(-0.346797273682)*x_ref[2]
        ref[0,2,2,0]=(0.956395362583)*x_ref[0]+(-0.609333787208)*x_ref[1]+(-0.86408067145)*x_ref[2]
        ref[0,2,2,1]=(0.570075324955)*x_ref[0]+(-0.251638346903)*x_ref[1]+(0.292841304398)*x_ref[2]
        ref[0,3,0,0]=(-0.240717117869)*x_ref[0]+(0.571307481772)*x_ref[1]+(0.66826910816)*x_ref[2]
        ref[0,3,0,1]=(0.113871023723)*x_ref[0]+(0.022498355636)*x_ref[1]+(0.95331234379)*x_ref[2]
        ref[0,3,1,0]=(-0.264946899001)*x_ref[0]+(1.08412373438)*x_ref[1]+(-1.10548347411)*x_ref[2]
        ref[0,3,1,1]=(-0.744510132854)*x_ref[0]+(-0.716638983187)*x_ref[1]+(0.64330838226)*x_ref[2]
        ref[0,3,2,0]=(1.15237343787)*x_ref[0]+(-1.42565901229)*x_ref[1]+(0.727585401396)*x_ref[2]
        ref[0,3,2,1]=(-0.251464863627)*x_ref[0]+(1.4067624912)*x_ref[1]+(0.674418739118)*x_ref[2]
        ref[0,4,0,0]=(-0.00551820316191)*x_ref[0]+(0.193862853101)*x_ref[1]+(1.8004924522)*x_ref[2]
        ref[0,4,0,1]=(-0.985879358916)*x_ref[0]+(0.476965815991)*x_ref[1]+(0.96363449822)*x_ref[2]
        ref[0,4,1,0]=(-0.657967361584)*x_ref[0]+(0.833452284341)*x_ref[1]+(-1.40128003822)*x_ref[2]
        ref[0,4,1,1]=(0.301379967988)*x_ref[0]+(0.0474545542353)*x_ref[1]+(-0.633087920569)*x_ref[2]
        ref[0,4,2,0]=(-1.94409755409)*x_ref[0]+(-0.468638363928)*x_ref[1]+(-0.713880237452)*x_ref[2]
        ref[0,4,2,1]=(-1.13536316065)*x_ref[0]+(-0.250761280127)*x_ref[1]+(0.328722833182)*x_ref[2]
        ref[1,0,0,0]=(0.556886266554)*x_ref[0]+(-0.70523094638)*x_ref[1]+(-0.333621268224)*x_ref[2]
        ref[1,0,0,1]=(-0.995565683628)*x_ref[0]+(-0.158120855239)*x_ref[1]+(0.0738870470169)*x_ref[2]
        ref[1,0,1,0]=(-0.209556712574)*x_ref[0]+(0.54393408711)*x_ref[1]+(-1.1617498966)*x_ref[2]
        ref[1,0,1,1]=(0.666809649924)*x_ref[0]+(1.18555984779)*x_ref[1]+(-0.442831598796)*x_ref[2]
        ref[1,0,2,0]=(-0.21354395591)*x_ref[0]+(-0.634626698269)*x_ref[1]+(1.25643757008)*x_ref[2]
        ref[1,0,2,1]=(-0.676259187073)*x_ref[0]+(-0.458532774397)*x_ref[1]+(0.271156919902)*x_ref[2]
        ref[1,1,0,0]=(-1.45529139262)*x_ref[0]+(0.311382249732)*x_ref[1]+(-0.134620909882)*x_ref[2]
        ref[1,1,0,1]=(-1.47708829198)*x_ref[0]+(0.968291536601)*x_ref[1]+(-0.566606887608)*x_ref[2]
        ref[1,1,1,0]=(1.18904329045)*x_ref[0]+(-0.209072247061)*x_ref[1]+(-0.26261457093)*x_ref[2]
        ref[1,1,1,1]=(0.313723817215)*x_ref[0]+(-0.169177309423)*x_ref[1]+(-0.00119161234247)*x_ref[2]
        ref[1,1,2,0]=(0.431298197457)*x_ref[0]+(0.264113678484)*x_ref[1]+(0.192619025313)*x_ref[2]
        ref[1,1,2,1]=(0.0469221259627)*x_ref[0]+(1.53477696238)*x_ref[1]+(0.00794662724457)*x_ref[2]
        ref[1,2,0,0]=(0.102179220279)*x_ref[0]+(-0.391365988191)*x_ref[1]+(0.302910557263)*x_ref[2]
        ref[1,2,0,1]=(0.841803002697)*x_ref[0]+(0.581539244301)*x_ref[1]+(0.00409108885792)*x_ref[2]
        ref[1,2,1,0]=(0.0840378161144)*x_ref[0]+(-0.431329807551)*x_ref[1]+(-1.0774029159)*x_ref[2]
        ref[1,2,1,1]=(-0.270803943532)*x_ref[0]+(0.0124514531652)*x_ref[1]+(0.197735227059)*x_ref[2]
        ref[1,2,2,0]=(0.479336250918)*x_ref[0]+(0.0435331360815)*x_ref[1]+(0.458048853417)*x_ref[2]
        ref[1,2,2,1]=(-0.59916419564)*x_ref[0]+(0.415546807791)*x_ref[1]+(-1.12105410423)*x_ref[2]
        ref[1,3,0,0]=(0.378277354507)*x_ref[0]+(1.00734343779)*x_ref[1]+(-0.282811389334)*x_ref[2]
        ref[1,3,0,1]=(0.717420487474)*x_ref[0]+(-0.730395427407)*x_ref[1]+(1.63850706843)*x_ref[2]
        ref[1,3,1,0]=(-0.0126954991872)*x_ref[0]+(-1.35916858377)*x_ref[1]+(-0.955731750212)*x_ref[2]
        ref[1,3,1,1]=(-0.560951559862)*x_ref[0]+(0.271085100351)*x_ref[1]+(0.60767567276)*x_ref[2]
        ref[1,3,2,0]=(-0.73286080289)*x_ref[0]+(-1.12018839274)*x_ref[1]+(0.365586479937)*x_ref[2]
        ref[1,3,2,1]=(-0.495391543232)*x_ref[0]+(-0.404605421036)*x_ref[1]+(1.21261493791)*x_ref[2]
        ref[1,4,0,0]=(-0.322147956238)*x_ref[0]+(0.278257549185)*x_ref[1]+(-0.927648624956)*x_ref[2]
        ref[1,4,0,1]=(1.23243511737)*x_ref[0]+(1.38355111729)*x_ref[1]+(0.0813739945129)*x_ref[2]
        ref[1,4,1,0]=(0.668384291635)*x_ref[0]+(0.995328961231)*x_ref[1]+(0.803009543307)*x_ref[2]
        ref[1,4,1,1]=(1.42643018603)*x_ref[0]+(0.940855692156)*x_ref[1]+(0.277556451163)*x_ref[2]
        ref[1,4,2,0]=(0.489588226757)*x_ref[0]+(-1.63608809124)*x_ref[1]+(-0.262918068522)*x_ref[2]
        ref[1,4,2,1]=(-1.06922340601)*x_ref[0]+(-0.0241938384008)*x_ref[1]+(-0.244107290038)*x_ref[2]
        ref[2,0,0,0]=(-0.556338154459)*x_ref[0]+(0.12071126785)*x_ref[1]+(0.11215804852)*x_ref[2]
        ref[2,0,0,1]=(-0.479017339774)*x_ref[0]+(-0.530313718209)*x_ref[1]+(0.490083551372)*x_ref[2]
        ref[2,0,1,0]=(0.731269264444)*x_ref[0]+(0.245793038193)*x_ref[1]+(0.0479833912422)*x_ref[2]
        ref[2,0,1,1]=(-0.286189382681)*x_ref[0]+(0.601756275554)*x_ref[1]+(-0.559334203137)*x_ref[2]
        ref[2,0,2,0]=(-0.876495720673)*x_ref[0]+(0.592513223261)*x_ref[1]+(0.844551261652)*x_ref[2]
        ref[2,0,2,1]=(1.22915142187)*x_ref[0]+(0.136223623577)*x_ref[1]+(0.88656874029)*x_ref[2]
        ref[2,1,0,0]=(0.542222650916)*x_ref[0]+(-0.709719163926)*x_ref[1]+(0.262207188498)*x_ref[2]
        ref[2,1,0,1]=(-0.208896980258)*x_ref[0]+(1.33968890395)*x_ref[1]+(-0.599621090138)*x_ref[2]
        ref[2,1,1,0]=(-0.171147406604)*x_ref[0]+(0.332227749427)*x_ref[1]+(1.15043521586)*x_ref[2]
        ref[2,1,1,1]=(0.571607567027)*x_ref[0]+(0.1367434941)*x_ref[1]+(1.47797159814)*x_ref[2]
        ref[2,1,2,0]=(0.839845574634)*x_ref[0]+(0.732437808234)*x_ref[1]+(-0.0311071654701)*x_ref[2]
        ref[2,1,2,1]=(1.45193213653)*x_ref[0]+(-0.0192312612414)*x_ref[1]+(0.308521161282)*x_ref[2]
        ref[2,2,0,0]=(-0.122755384028)*x_ref[0]+(0.133510038844)*x_ref[1]+(0.837213365425)*x_ref[2]
        ref[2,2,0,1]=(-0.411967923958)*x_ref[0]+(-0.955489021071)*x_ref[1]+(0.269005602186)*x_ref[2]
        ref[2,2,1,0]=(-1.19993089382)*x_ref[0]+(0.276541296246)*x_ref[1]+(0.523388479396)*x_ref[2]
        ref[2,2,1,1]=(0.0586514653869)*x_ref[0]+(-0.624975232742)*x_ref[1]+(-1.55738876751)*x_ref[2]
        ref[2,2,2,0]=(-0.892065938887)*x_ref[0]+(-0.4848241954)*x_ref[1]+(0.315463883511)*x_ref[2]
        ref[2,2,2,1]=(0.0272963181118)*x_ref[0]+(0.0187713454669)*x_ref[1]+(-1.06524661854)*x_ref[2]
        ref[2,3,0,0]=(-1.44099287886)*x_ref[0]+(0.495718421075)*x_ref[1]+(0.100401960223)*x_ref[2]
        ref[2,3,0,1]=(0.171175119412)*x_ref[0]+(-1.34795558607)*x_ref[1]+(-0.361766653264)*x_ref[2]
        ref[2,3,1,0]=(-1.80591882588)*x_ref[0]+(-0.511459455136)*x_ref[1]+(-0.491310264055)*x_ref[2]
        ref[2,3,1,1]=(-0.419508934622)*x_ref[0]+(-1.12084042844)*x_ref[1]+(-0.086947672548)*x_ref[2]
        ref[2,3,2,0]=(0.492388303934)*x_ref[0]+(-0.623505828597)*x_ref[1]+(1.6965873745)*x_ref[2]
        ref[2,3,2,1]=(1.59203810356)*x_ref[0]+(0.36918142231)*x_ref[1]+(-1.90067115052)*x_ref[2]
        ref[2,4,0,0]=(1.00875649298)*x_ref[0]+(-1.43246384934)*x_ref[1]+(-0.51811906593)*x_ref[2]
        ref[2,4,0,1]=(0.260118918034)*x_ref[0]+(-1.57606061865)*x_ref[1]+(-0.79257165655)*x_ref[2]
        ref[2,4,1,0]=(0.954414074467)*x_ref[0]+(-0.230132876613)*x_ref[1]+(0.0956378957424)*x_ref[2]
        ref[2,4,1,1]=(0.958843806364)*x_ref[0]+(-0.155804429995)*x_ref[1]+(-1.02904877039)*x_ref[2]
        ref[2,4,2,0]=(0.252668700811)*x_ref[0]+(0.164829885749)*x_ref[1]+(0.747032722421)*x_ref[2]
        ref[2,4,2,1]=(1.96393052323)*x_ref[0]+(-0.821363958807)*x_ref[1]+(-1.01663555063)*x_ref[2]
        ref[3,0,0,0]=(1.58086037053)*x_ref[0]+(-1.66495295611)*x_ref[1]+(-0.510526191383)*x_ref[2]
        ref[3,0,0,1]=(-0.931871976029)*x_ref[0]+(1.33102205125)*x_ref[1]+(-0.95170580468)*x_ref[2]
        ref[3,0,1,0]=(0.99825404115)*x_ref[0]+(0.915274413552)*x_ref[1]+(1.5057227674)*x_ref[2]
        ref[3,0,1,1]=(0.0485551924989)*x_ref[0]+(-1.0877405758)*x_ref[1]+(-0.887342573702)*x_ref[2]
        ref[3,0,2,0]=(-0.745146140677)*x_ref[0]+(-0.0108141266015)*x_ref[1]+(0.194597576308)*x_ref[2]
        ref[3,0,2,1]=(-0.935029218338)*x_ref[0]+(0.726322809025)*x_ref[1]+(0.57508467103)*x_ref[2]
        ref[3,1,0,0]=(0.115668237156)*x_ref[0]+(0.0970255991906)*x_ref[1]+(-0.227291123547)*x_ref[2]
        ref[3,1,0,1]=(0.167722335863)*x_ref[0]+(-0.552168424291)*x_ref[1]+(1.36293027463)*x_ref[2]
        ref[3,1,1,0]=(-1.05781878028)*x_ref[0]+(-0.0429389047972)*x_ref[1]+(-0.257166337509)*x_ref[2]
        ref[3,1,1,1]=(0.129148486118)*x_ref[0]+(-0.0209434899369)*x_ref[1]+(0.253153005253)*x_ref[2]
        ref[3,1,2,0]=(1.3236615365)*x_ref[0]+(-1.18164469224)*x_ref[1]+(0.698012697426)*x_ref[2]
        ref[3,1,2,1]=(1.10438661886)*x_ref[0]+(-0.852903948395)*x_ref[1]+(1.40365162024)*x_ref[2]
        ref[3,2,0,0]=(-0.949308009957)*x_ref[0]+(1.01993206161)*x_ref[1]+(-0.821721697476)*x_ref[2]
        ref[3,2,0,1]=(0.326450662787)*x_ref[0]+(0.395096088107)*x_ref[1]+(-0.368904800584)*x_ref[2]
        ref[3,2,1,0]=(0.13107780566)*x_ref[0]+(-0.342904134623)*x_ref[1]+(0.00835087701185)*x_ref[2]
        ref[3,2,1,1]=(-0.500694550647)*x_ref[0]+(1.38322390003)*x_ref[1]+(0.0576453183412)*x_ref[2]
        ref[3,2,2,0]=(-1.32014489811)*x_ref[0]+(-0.444079317864)*x_ref[1]+(1.1135946018)*x_ref[2]
        ref[3,2,2,1]=(-0.19593105007)*x_ref[0]+(-0.00224390969377)*x_ref[1]+(-0.73869338942)*x_ref[2]
        ref[3,3,0,0]=(0.0151201622255)*x_ref[0]+(-1.10873788153)*x_ref[1]+(0.189116254832)*x_ref[2]
        ref[3,3,0,1]=(1.05996205566)*x_ref[0]+(0.398465639867)*x_ref[1]+(-0.319773104448)*x_ref[2]
        ref[3,3,1,0]=(0.687612827243)*x_ref[0]+(1.10970823257)*x_ref[1]+(1.02543288975)*x_ref[2]
        ref[3,3,1,1]=(1.27258509981)*x_ref[0]+(-1.7060826391)*x_ref[1]+(-0.84948812227)*x_ref[2]
        ref[3,3,2,0]=(1.2343036886)*x_ref[0]+(0.521090253023)*x_ref[1]+(-1.13377412954)*x_ref[2]
        ref[3,3,2,1]=(-0.65568738385)*x_ref[0]+(1.00354617712)*x_ref[1]+(0.836719015812)*x_ref[2]
        ref[3,4,0,0]=(-0.545324078594)*x_ref[0]+(0.452747782208)*x_ref[1]+(-1.59718954237)*x_ref[2]
        ref[3,4,0,1]=(1.71651153317)*x_ref[0]+(-0.161984538787)*x_ref[1]+(-0.377244827904)*x_ref[2]
        ref[3,4,1,0]=(-0.284033324558)*x_ref[0]+(-0.405941433062)*x_ref[1]+(0.824573264594)*x_ref[2]
        ref[3,4,1,1]=(-0.112542702803)*x_ref[0]+(-0.470814805786)*x_ref[1]+(-0.827427436576)*x_ref[2]
        ref[3,4,2,0]=(-0.763655132981)*x_ref[0]+(-0.442818775864)*x_ref[1]+(0.809814510041)*x_ref[2]
        ref[3,4,2,1]=(-0.33794306298)*x_ref[0]+(1.91169591332)*x_ref[1]+(-0.487268088196)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.898871557679)*x[0]+(0.0218492481542)*x[1]
        ref=(0.898871557679)*x_ref[0]+(0.0218492481542)*x_ref[1]
      else:
        arg=(-0.490730396919)*x[0]+(0.840683566031)*x[1]+(-0.265457839434)*x[2]
        ref=(-0.490730396919)*x_ref[0]+(0.840683566031)*x_ref[1]+(-0.265457839434)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        arg[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
        ref[0]=(-0.505098987801)*x[0]+(-0.647246526986)*x[1]
        ref[1]=(-0.823205698667)*x[0]+(-0.88472991168)*x[1]
      else:
        arg[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        arg[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
        ref[0]=(-1.01688152316)*x[0]+(0.197478760214)*x[1]+(-0.833266923905)*x[2]
        ref[1]=(0.763755283448)*x[0]+(-0.920100276504)*x[1]+(1.0042717309)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.609515733352)*x[0]+(0.477437125804)*x[1]
        arg[0,1]=(0.225027969847)*x[0]+(-1.2178256661)*x[1]
        arg[0,2]=(-0.414641287889)*x[0]+(-0.536814389632)*x[1]
        arg[0,3]=(0.993370642676)*x[0]+(0.430000528217)*x[1]
        arg[0,4]=(0.577804610335)*x[0]+(0.0950391061127)*x[1]
        arg[1,0]=(0.437140855856)*x[0]+(0.764946114688)*x[1]
        arg[1,1]=(-0.491195658809)*x[0]+(-1.04729131261)*x[1]
        arg[1,2]=(-0.179287249145)*x[0]+(-1.84699873991)*x[1]
        arg[1,3]=(0.650408136022)*x[0]+(0.437145116534)*x[1]
        arg[1,4]=(0.43853490169)*x[0]+(-0.633716126064)*x[1]
        arg[2,0]=(0.6275418675)*x[0]+(0.812839621482)*x[1]
        arg[2,1]=(0.151298084104)*x[0]+(0.191677853013)*x[1]
        arg[2,2]=(1.35800571802)*x[0]+(1.08489593249)*x[1]
        arg[2,3]=(0.343386100925)*x[0]+(-0.108323271881)*x[1]
        arg[2,4]=(1.5968773444)*x[0]+(0.421328566279)*x[1]
        arg[3,0]=(1.0607367804)*x[0]+(-0.347352081886)*x[1]
        arg[3,1]=(-0.307183028406)*x[0]+(0.171667489459)*x[1]
        arg[3,2]=(0.255609148103)*x[0]+(0.21702578756)*x[1]
        arg[3,3]=(0.204487068999)*x[0]+(0.380168463726)*x[1]
        arg[3,4]=(-0.727714186561)*x[0]+(-0.129616906837)*x[1]
        ref[0,0]=(-0.609515733352)*x_ref[0]+(0.477437125804)*x_ref[1]
        ref[0,1]=(0.225027969847)*x_ref[0]+(-1.2178256661)*x_ref[1]
        ref[0,2]=(-0.414641287889)*x_ref[0]+(-0.536814389632)*x_ref[1]
        ref[0,3]=(0.993370642676)*x_ref[0]+(0.430000528217)*x_ref[1]
        ref[0,4]=(0.577804610335)*x_ref[0]+(0.0950391061127)*x_ref[1]
        ref[1,0]=(0.437140855856)*x_ref[0]+(0.764946114688)*x_ref[1]
        ref[1,1]=(-0.491195658809)*x_ref[0]+(-1.04729131261)*x_ref[1]
        ref[1,2]=(-0.179287249145)*x_ref[0]+(-1.84699873991)*x_ref[1]
        ref[1,3]=(0.650408136022)*x_ref[0]+(0.437145116534)*x_ref[1]
        ref[1,4]=(0.43853490169)*x_ref[0]+(-0.633716126064)*x_ref[1]
        ref[2,0]=(0.6275418675)*x_ref[0]+(0.812839621482)*x_ref[1]
        ref[2,1]=(0.151298084104)*x_ref[0]+(0.191677853013)*x_ref[1]
        ref[2,2]=(1.35800571802)*x_ref[0]+(1.08489593249)*x_ref[1]
        ref[2,3]=(0.343386100925)*x_ref[0]+(-0.108323271881)*x_ref[1]
        ref[2,4]=(1.5968773444)*x_ref[0]+(0.421328566279)*x_ref[1]
        ref[3,0]=(1.0607367804)*x_ref[0]+(-0.347352081886)*x_ref[1]
        ref[3,1]=(-0.307183028406)*x_ref[0]+(0.171667489459)*x_ref[1]
        ref[3,2]=(0.255609148103)*x_ref[0]+(0.21702578756)*x_ref[1]
        ref[3,3]=(0.204487068999)*x_ref[0]+(0.380168463726)*x_ref[1]
        ref[3,4]=(-0.727714186561)*x_ref[0]+(-0.129616906837)*x_ref[1]
      else:
        arg[0,0]=(-0.282058347281)*x[0]+(-0.804878634484)*x[1]+(1.59321960831)*x[2]
        arg[0,1]=(0.046711660417)*x[0]+(-0.353452776565)*x[1]+(-1.22459734449)*x[2]
        arg[0,2]=(0.846231855307)*x[0]+(1.08618070898)*x[1]+(-1.50365393652)*x[2]
        arg[0,3]=(-0.358546890162)*x[0]+(1.53476318049)*x[1]+(-1.19382694871)*x[2]
        arg[0,4]=(-1.1943310015)*x[0]+(0.260842673421)*x[1]+(-0.337737972605)*x[2]
        arg[1,0]=(1.87166717118)*x[0]+(0.853452519127)*x[1]+(-0.59081660847)*x[2]
        arg[1,1]=(0.23310002871)*x[0]+(0.306981228888)*x[1]+(0.36040454683)*x[2]
        arg[1,2]=(-1.619288679)*x[0]+(-0.459254234211)*x[1]+(0.722479242965)*x[2]
        arg[1,3]=(-1.07724802769)*x[0]+(0.836465436044)*x[1]+(0.255446930493)*x[2]
        arg[1,4]=(1.07857092921)*x[0]+(1.25642859092)*x[1]+(-0.922799422574)*x[2]
        arg[2,0]=(-0.832179820568)*x[0]+(-0.511243247139)*x[1]+(0.589457094147)*x[2]
        arg[2,1]=(-0.22616483359)*x[0]+(1.75842052561)*x[1]+(-1.19438769374)*x[2]
        arg[2,2]=(-0.690678634371)*x[0]+(0.963335056891)*x[1]+(0.53653322138)*x[2]
        arg[2,3]=(-0.906140718249)*x[0]+(0.766237211441)*x[1]+(1.19445635284)*x[2]
        arg[2,4]=(-0.16966820901)*x[0]+(0.185706811112)*x[1]+(0.975546783315)*x[2]
        arg[3,0]=(1.85053644213)*x[0]+(1.000328518)*x[1]+(0.332466610087)*x[2]
        arg[3,1]=(-0.199393988242)*x[0]+(0.637601295557)*x[1]+(1.29246614993)*x[2]
        arg[3,2]=(-0.275167528438)*x[0]+(1.00557873837)*x[1]+(1.17143679274)*x[2]
        arg[3,3]=(-0.306797319235)*x[0]+(0.984827835253)*x[1]+(0.0324710900497)*x[2]
        arg[3,4]=(-1.07994829511)*x[0]+(0.341016609147)*x[1]+(0.115829923585)*x[2]
        ref[0,0]=(-0.282058347281)*x_ref[0]+(-0.804878634484)*x_ref[1]+(1.59321960831)*x_ref[2]
        ref[0,1]=(0.046711660417)*x_ref[0]+(-0.353452776565)*x_ref[1]+(-1.22459734449)*x_ref[2]
        ref[0,2]=(0.846231855307)*x_ref[0]+(1.08618070898)*x_ref[1]+(-1.50365393652)*x_ref[2]
        ref[0,3]=(-0.358546890162)*x_ref[0]+(1.53476318049)*x_ref[1]+(-1.19382694871)*x_ref[2]
        ref[0,4]=(-1.1943310015)*x_ref[0]+(0.260842673421)*x_ref[1]+(-0.337737972605)*x_ref[2]
        ref[1,0]=(1.87166717118)*x_ref[0]+(0.853452519127)*x_ref[1]+(-0.59081660847)*x_ref[2]
        ref[1,1]=(0.23310002871)*x_ref[0]+(0.306981228888)*x_ref[1]+(0.36040454683)*x_ref[2]
        ref[1,2]=(-1.619288679)*x_ref[0]+(-0.459254234211)*x_ref[1]+(0.722479242965)*x_ref[2]
        ref[1,3]=(-1.07724802769)*x_ref[0]+(0.836465436044)*x_ref[1]+(0.255446930493)*x_ref[2]
        ref[1,4]=(1.07857092921)*x_ref[0]+(1.25642859092)*x_ref[1]+(-0.922799422574)*x_ref[2]
        ref[2,0]=(-0.832179820568)*x_ref[0]+(-0.511243247139)*x_ref[1]+(0.589457094147)*x_ref[2]
        ref[2,1]=(-0.22616483359)*x_ref[0]+(1.75842052561)*x_ref[1]+(-1.19438769374)*x_ref[2]
        ref[2,2]=(-0.690678634371)*x_ref[0]+(0.963335056891)*x_ref[1]+(0.53653322138)*x_ref[2]
        ref[2,3]=(-0.906140718249)*x_ref[0]+(0.766237211441)*x_ref[1]+(1.19445635284)*x_ref[2]
        ref[2,4]=(-0.16966820901)*x_ref[0]+(0.185706811112)*x_ref[1]+(0.975546783315)*x_ref[2]
        ref[3,0]=(1.85053644213)*x_ref[0]+(1.000328518)*x_ref[1]+(0.332466610087)*x_ref[2]
        ref[3,1]=(-0.199393988242)*x_ref[0]+(0.637601295557)*x_ref[1]+(1.29246614993)*x_ref[2]
        ref[3,2]=(-0.275167528438)*x_ref[0]+(1.00557873837)*x_ref[1]+(1.17143679274)*x_ref[2]
        ref[3,3]=(-0.306797319235)*x_ref[0]+(0.984827835253)*x_ref[1]+(0.0324710900497)*x_ref[2]
        ref[3,4]=(-1.07994829511)*x_ref[0]+(0.341016609147)*x_ref[1]+(0.115829923585)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(0.618861298758)*x[0]+(-0.764616734859)*x[1]
        arg[0,0,1]=(-0.220720229024)*x[0]+(-0.670415412641)*x[1]
        arg[0,1,0]=(0.847082172076)*x[0]+(0.104515154796)*x[1]
        arg[0,1,1]=(-1.39487869037)*x[0]+(1.37639032205)*x[1]
        arg[1,0,0]=(-0.12944065767)*x[0]+(-0.91214412252)*x[1]
        arg[1,0,1]=(0.0421119211697)*x[0]+(1.19918915102)*x[1]
        arg[1,1,0]=(0.657099038454)*x[0]+(0.102886690913)*x[1]
        arg[1,1,1]=(0.114311284584)*x[0]+(-0.827908524893)*x[1]
        arg[2,0,0]=(-0.905281334495)*x[0]+(1.02228309099)*x[1]
        arg[2,0,1]=(-0.560619228879)*x[0]+(-0.924304525905)*x[1]
        arg[2,1,0]=(0.082868967617)*x[0]+(-0.129063346527)*x[1]
        arg[2,1,1]=(0.386687881433)*x[0]+(-0.874995344656)*x[1]
        arg[3,0,0]=(0.460976605295)*x[0]+(0.145631537258)*x[1]
        arg[3,0,1]=(0.439977672901)*x[0]+(0.714321229059)*x[1]
        arg[3,1,0]=(-0.0459363656635)*x[0]+(-0.0754715887416)*x[1]
        arg[3,1,1]=(-0.789910257931)*x[0]+(-0.806078036909)*x[1]
        arg[4,0,0]=(-1.09951071165)*x[0]+(-1.33461983123)*x[1]
        arg[4,0,1]=(0.857184302111)*x[0]+(0.823036666525)*x[1]
        arg[4,1,0]=(-1.0682486872)*x[0]+(-0.149861981932)*x[1]
        arg[4,1,1]=(-0.0758367136425)*x[0]+(0.0460556685916)*x[1]
        arg[5,0,0]=(0.856207295965)*x[0]+(-0.885450852663)*x[1]
        arg[5,0,1]=(-0.113939086685)*x[0]+(0.742808810968)*x[1]
        arg[5,1,0]=(-0.174573302919)*x[0]+(0.486803189896)*x[1]
        arg[5,1,1]=(1.1726771004)*x[0]+(-0.329765988186)*x[1]
        ref[0,0,0]=(0.618861298758)*x_ref[0]+(-0.764616734859)*x_ref[1]
        ref[0,0,1]=(-0.220720229024)*x_ref[0]+(-0.670415412641)*x_ref[1]
        ref[0,1,0]=(0.847082172076)*x_ref[0]+(0.104515154796)*x_ref[1]
        ref[0,1,1]=(-1.39487869037)*x_ref[0]+(1.37639032205)*x_ref[1]
        ref[1,0,0]=(-0.12944065767)*x_ref[0]+(-0.91214412252)*x_ref[1]
        ref[1,0,1]=(0.0421119211697)*x_ref[0]+(1.19918915102)*x_ref[1]
        ref[1,1,0]=(0.657099038454)*x_ref[0]+(0.102886690913)*x_ref[1]
        ref[1,1,1]=(0.114311284584)*x_ref[0]+(-0.827908524893)*x_ref[1]
        ref[2,0,0]=(-0.905281334495)*x_ref[0]+(1.02228309099)*x_ref[1]
        ref[2,0,1]=(-0.560619228879)*x_ref[0]+(-0.924304525905)*x_ref[1]
        ref[2,1,0]=(0.082868967617)*x_ref[0]+(-0.129063346527)*x_ref[1]
        ref[2,1,1]=(0.386687881433)*x_ref[0]+(-0.874995344656)*x_ref[1]
        ref[3,0,0]=(0.460976605295)*x_ref[0]+(0.145631537258)*x_ref[1]
        ref[3,0,1]=(0.439977672901)*x_ref[0]+(0.714321229059)*x_ref[1]
        ref[3,1,0]=(-0.0459363656635)*x_ref[0]+(-0.0754715887416)*x_ref[1]
        ref[3,1,1]=(-0.789910257931)*x_ref[0]+(-0.806078036909)*x_ref[1]
        ref[4,0,0]=(-1.09951071165)*x_ref[0]+(-1.33461983123)*x_ref[1]
        ref[4,0,1]=(0.857184302111)*x_ref[0]+(0.823036666525)*x_ref[1]
        ref[4,1,0]=(-1.0682486872)*x_ref[0]+(-0.149861981932)*x_ref[1]
        ref[4,1,1]=(-0.0758367136425)*x_ref[0]+(0.0460556685916)*x_ref[1]
        ref[5,0,0]=(0.856207295965)*x_ref[0]+(-0.885450852663)*x_ref[1]
        ref[5,0,1]=(-0.113939086685)*x_ref[0]+(0.742808810968)*x_ref[1]
        ref[5,1,0]=(-0.174573302919)*x_ref[0]+(0.486803189896)*x_ref[1]
        ref[5,1,1]=(1.1726771004)*x_ref[0]+(-0.329765988186)*x_ref[1]
      else:
        arg[0,0,0]=(-0.640205798541)*x[0]+(0.271670719935)*x[1]+(-0.163000598261)*x[2]
        arg[0,0,1]=(0.735605297981)*x[0]+(-1.05892412676)*x[1]+(-1.1083007441)*x[2]
        arg[0,1,0]=(0.0349869408333)*x[0]+(0.699147831788)*x[1]+(-0.454390826725)*x[2]
        arg[0,1,1]=(-0.703957679243)*x[0]+(-1.63654420176)*x[1]+(-1.13045075795)*x[2]
        arg[1,0,0]=(1.01167296102)*x[0]+(-0.362443738659)*x[1]+(-0.0481138040854)*x[2]
        arg[1,0,1]=(0.110033706339)*x[0]+(1.47809236549)*x[1]+(0.240737599484)*x[2]
        arg[1,1,0]=(-0.0429437939148)*x[0]+(-1.16945967196)*x[1]+(0.924577053186)*x[2]
        arg[1,1,1]=(-1.57856509244)*x[0]+(-0.460441809594)*x[1]+(-0.792590928973)*x[2]
        arg[2,0,0]=(-1.93020920065)*x[0]+(0.391612451745)*x[1]+(-0.0130238634938)*x[2]
        arg[2,0,1]=(0.0678908982414)*x[0]+(-0.929813889361)*x[1]+(0.727462587818)*x[2]
        arg[2,1,0]=(-0.297494494419)*x[0]+(0.167234033247)*x[1]+(-1.01372084798)*x[2]
        arg[2,1,1]=(-0.608810623174)*x[0]+(1.04629900384)*x[1]+(1.09464953456)*x[2]
        arg[3,0,0]=(1.2996449893)*x[0]+(-0.458197425323)*x[1]+(1.31523663461)*x[2]
        arg[3,0,1]=(-0.848124038387)*x[0]+(0.0136487458675)*x[1]+(1.13573336016)*x[2]
        arg[3,1,0]=(0.377245000633)*x[0]+(-0.766336675857)*x[1]+(0.380031677083)*x[2]
        arg[3,1,1]=(0.158289533163)*x[0]+(-0.149454925108)*x[1]+(-0.59348598427)*x[2]
        arg[4,0,0]=(-1.13101661269)*x[0]+(-0.694062133171)*x[1]+(0.00925221725394)*x[2]
        arg[4,0,1]=(-0.977317473287)*x[0]+(-0.121712752812)*x[1]+(-0.480470720169)*x[2]
        arg[4,1,0]=(1.14884269473)*x[0]+(-0.204958154665)*x[1]+(1.24328207187)*x[2]
        arg[4,1,1]=(0.0791969146134)*x[0]+(0.589873409499)*x[1]+(1.90521993698)*x[2]
        arg[5,0,0]=(1.75625788908)*x[0]+(0.0134922175773)*x[1]+(-0.699884009939)*x[2]
        arg[5,0,1]=(-0.73456900252)*x[0]+(0.923248295457)*x[1]+(1.05161845199)*x[2]
        arg[5,1,0]=(-0.0335528775888)*x[0]+(0.728742831853)*x[1]+(-0.5212350627)*x[2]
        arg[5,1,1]=(0.556540394981)*x[0]+(-0.128428520433)*x[1]+(1.06358498081)*x[2]
        ref[0,0,0]=(-0.640205798541)*x_ref[0]+(0.271670719935)*x_ref[1]+(-0.163000598261)*x_ref[2]
        ref[0,0,1]=(0.735605297981)*x_ref[0]+(-1.05892412676)*x_ref[1]+(-1.1083007441)*x_ref[2]
        ref[0,1,0]=(0.0349869408333)*x_ref[0]+(0.699147831788)*x_ref[1]+(-0.454390826725)*x_ref[2]
        ref[0,1,1]=(-0.703957679243)*x_ref[0]+(-1.63654420176)*x_ref[1]+(-1.13045075795)*x_ref[2]
        ref[1,0,0]=(1.01167296102)*x_ref[0]+(-0.362443738659)*x_ref[1]+(-0.0481138040854)*x_ref[2]
        ref[1,0,1]=(0.110033706339)*x_ref[0]+(1.47809236549)*x_ref[1]+(0.240737599484)*x_ref[2]
        ref[1,1,0]=(-0.0429437939148)*x_ref[0]+(-1.16945967196)*x_ref[1]+(0.924577053186)*x_ref[2]
        ref[1,1,1]=(-1.57856509244)*x_ref[0]+(-0.460441809594)*x_ref[1]+(-0.792590928973)*x_ref[2]
        ref[2,0,0]=(-1.93020920065)*x_ref[0]+(0.391612451745)*x_ref[1]+(-0.0130238634938)*x_ref[2]
        ref[2,0,1]=(0.0678908982414)*x_ref[0]+(-0.929813889361)*x_ref[1]+(0.727462587818)*x_ref[2]
        ref[2,1,0]=(-0.297494494419)*x_ref[0]+(0.167234033247)*x_ref[1]+(-1.01372084798)*x_ref[2]
        ref[2,1,1]=(-0.608810623174)*x_ref[0]+(1.04629900384)*x_ref[1]+(1.09464953456)*x_ref[2]
        ref[3,0,0]=(1.2996449893)*x_ref[0]+(-0.458197425323)*x_ref[1]+(1.31523663461)*x_ref[2]
        ref[3,0,1]=(-0.848124038387)*x_ref[0]+(0.0136487458675)*x_ref[1]+(1.13573336016)*x_ref[2]
        ref[3,1,0]=(0.377245000633)*x_ref[0]+(-0.766336675857)*x_ref[1]+(0.380031677083)*x_ref[2]
        ref[3,1,1]=(0.158289533163)*x_ref[0]+(-0.149454925108)*x_ref[1]+(-0.59348598427)*x_ref[2]
        ref[4,0,0]=(-1.13101661269)*x_ref[0]+(-0.694062133171)*x_ref[1]+(0.00925221725394)*x_ref[2]
        ref[4,0,1]=(-0.977317473287)*x_ref[0]+(-0.121712752812)*x_ref[1]+(-0.480470720169)*x_ref[2]
        ref[4,1,0]=(1.14884269473)*x_ref[0]+(-0.204958154665)*x_ref[1]+(1.24328207187)*x_ref[2]
        ref[4,1,1]=(0.0791969146134)*x_ref[0]+(0.589873409499)*x_ref[1]+(1.90521993698)*x_ref[2]
        ref[5,0,0]=(1.75625788908)*x_ref[0]+(0.0134922175773)*x_ref[1]+(-0.699884009939)*x_ref[2]
        ref[5,0,1]=(-0.73456900252)*x_ref[0]+(0.923248295457)*x_ref[1]+(1.05161845199)*x_ref[2]
        ref[5,1,0]=(-0.0335528775888)*x_ref[0]+(0.728742831853)*x_ref[1]+(-0.5212350627)*x_ref[2]
        ref[5,1,1]=(0.556540394981)*x_ref[0]+(-0.128428520433)*x_ref[1]+(1.06358498081)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_Solution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(0.507324804235)*x[0]+(0.725831700089)*x[1]
        arg[0,0,0,1]=(0.847901454936)*x[0]+(0.84787024099)*x[1]
        arg[0,0,1,0]=(0.221488465784)*x[0]+(0.0449286988708)*x[1]
        arg[0,0,1,1]=(0.00897176578501)*x[0]+(-0.799118487206)*x[1]
        arg[0,0,2,0]=(0.171491682599)*x[0]+(1.0259065872)*x[1]
        arg[0,0,2,1]=(-0.213244683798)*x[0]+(-0.706816385516)*x[1]
        arg[0,1,0,0]=(1.54199875186)*x[0]+(-1.09479402082)*x[1]
        arg[0,1,0,1]=(-0.755278051349)*x[0]+(-0.371785089226)*x[1]
        arg[0,1,1,0]=(-0.798931370137)*x[0]+(-0.716047155117)*x[1]
        arg[0,1,1,1]=(-0.79104072521)*x[0]+(-1.45842710768)*x[1]
        arg[0,1,2,0]=(-0.489161301965)*x[0]+(0.115368605994)*x[1]
        arg[0,1,2,1]=(0.00167228618616)*x[0]+(-0.636896504873)*x[1]
        arg[0,2,0,0]=(-0.239182729045)*x[0]+(0.470786327067)*x[1]
        arg[0,2,0,1]=(0.182105943188)*x[0]+(0.328443662342)*x[1]
        arg[0,2,1,0]=(0.000476540430199)*x[0]+(-0.498066512126)*x[1]
        arg[0,2,1,1]=(-0.822985659305)*x[0]+(-0.118645880185)*x[1]
        arg[0,2,2,0]=(1.18375382554)*x[0]+(1.51417307474)*x[1]
        arg[0,2,2,1]=(-0.99755868797)*x[0]+(-1.4187068658)*x[1]
        arg[0,3,0,0]=(-0.0908782314954)*x[0]+(-0.25901340471)*x[1]
        arg[0,3,0,1]=(-1.77626335863)*x[0]+(-0.862132832663)*x[1]
        arg[0,3,1,0]=(0.532147327491)*x[0]+(0.454949026506)*x[1]
        arg[0,3,1,1]=(-0.882307958628)*x[0]+(0.00780544746858)*x[1]
        arg[0,3,2,0]=(0.700104420574)*x[0]+(0.720508036709)*x[1]
        arg[0,3,2,1]=(-1.27167118811)*x[0]+(-0.372999410908)*x[1]
        arg[0,4,0,0]=(0.396222903345)*x[0]+(1.70664799562)*x[1]
        arg[0,4,0,1]=(0.229159055372)*x[0]+(0.0162677758439)*x[1]
        arg[0,4,1,0]=(1.69470499818)*x[0]+(0.0971804891468)*x[1]
        arg[0,4,1,1]=(0.180710573838)*x[0]+(1.05311557127)*x[1]
        arg[0,4,2,0]=(-0.072661496917)*x[0]+(0.466256643644)*x[1]
        arg[0,4,2,1]=(0.25755631084)*x[0]+(0.306990925808)*x[1]
        arg[1,0,0,0]=(0.835353191317)*x[0]+(-0.619237060145)*x[1]
        arg[1,0,0,1]=(1.02966866371)*x[0]+(0.327724003031)*x[1]
        arg[1,0,1,0]=(-0.59214788986)*x[0]+(-0.927738485428)*x[1]
        arg[1,0,1,1]=(-0.805929278446)*x[0]+(-0.366113927406)*x[1]
        arg[1,0,2,0]=(0.952876155721)*x[0]+(0.490531201467)*x[1]
        arg[1,0,2,1]=(1.44341651657)*x[0]+(-0.269899260772)*x[1]
        arg[1,1,0,0]=(0.302587966785)*x[0]+(-0.0373883991695)*x[1]
        arg[1,1,0,1]=(0.77498995648)*x[0]+(1.81052574555)*x[1]
        arg[1,1,1,0]=(1.47950560134)*x[0]+(-0.714878219566)*x[1]
        arg[1,1,1,1]=(0.0713827629606)*x[0]+(-1.22018291576)*x[1]
        arg[1,1,2,0]=(1.30715631604)*x[0]+(-0.191666253513)*x[1]
        arg[1,1,2,1]=(0.330921962628)*x[0]+(-1.57160261912)*x[1]
        arg[1,2,0,0]=(0.593477126597)*x[0]+(0.728196634765)*x[1]
        arg[1,2,0,1]=(0.853860342143)*x[0]+(0.142625101433)*x[1]
        arg[1,2,1,0]=(-0.743590115343)*x[0]+(0.108893736366)*x[1]
        arg[1,2,1,1]=(-0.651514863459)*x[0]+(0.365517180762)*x[1]
        arg[1,2,2,0]=(1.491516719)*x[0]+(0.28900806658)*x[1]
        arg[1,2,2,1]=(1.23733736339)*x[0]+(-0.58438294882)*x[1]
        arg[1,3,0,0]=(0.185696018505)*x[0]+(-1.29119084907)*x[1]
        arg[1,3,0,1]=(1.73584656085)*x[0]+(0.720989518003)*x[1]
        arg[1,3,1,0]=(-1.16044192371)*x[0]+(-0.131128994204)*x[1]
        arg[1,3,1,1]=(-0.00913995119508)*x[0]+(0.506516682557)*x[1]
        arg[1,3,2,0]=(-1.4970171408)*x[0]+(0.692262679219)*x[1]
        arg[1,3,2,1]=(0.433442352395)*x[0]+(-0.118782513258)*x[1]
        arg[1,4,0,0]=(0.186272280804)*x[0]+(-0.136627137097)*x[1]
        arg[1,4,0,1]=(1.25477168515)*x[0]+(-0.0871831369474)*x[1]
        arg[1,4,1,0]=(1.22746022677)*x[0]+(0.799329883559)*x[1]
        arg[1,4,1,1]=(-0.756298280123)*x[0]+(0.317500967634)*x[1]
        arg[1,4,2,0]=(0.201184703435)*x[0]+(-0.116537935236)*x[1]
        arg[1,4,2,1]=(0.851291897607)*x[0]+(0.404063606516)*x[1]
        arg[2,0,0,0]=(-0.257605370026)*x[0]+(-0.202424701671)*x[1]
        arg[2,0,0,1]=(0.0219332720675)*x[0]+(0.518222103227)*x[1]
        arg[2,0,1,0]=(1.03726036144)*x[0]+(-1.01482867596)*x[1]
        arg[2,0,1,1]=(0.00677025785109)*x[0]+(0.607932484906)*x[1]
        arg[2,0,2,0]=(-0.753000974637)*x[0]+(0.112884693676)*x[1]
        arg[2,0,2,1]=(-0.585627190582)*x[0]+(0.548136242722)*x[1]
        arg[2,1,0,0]=(0.382466685883)*x[0]+(-1.15520454298)*x[1]
        arg[2,1,0,1]=(-0.722705123323)*x[0]+(-0.599069291154)*x[1]
        arg[2,1,1,0]=(-0.845885208256)*x[0]+(-1.56066350212)*x[1]
        arg[2,1,1,1]=(-1.31472469553)*x[0]+(0.18973548452)*x[1]
        arg[2,1,2,0]=(-0.419142599733)*x[0]+(1.43873372527)*x[1]
        arg[2,1,2,1]=(0.22685530286)*x[0]+(-0.53636714393)*x[1]
        arg[2,2,0,0]=(1.32884214935)*x[0]+(-0.377575726395)*x[1]
        arg[2,2,0,1]=(0.0845542736847)*x[0]+(0.0888339000429)*x[1]
        arg[2,2,1,0]=(-1.05813285521)*x[0]+(-0.980382395463)*x[1]
        arg[2,2,1,1]=(0.306785463268)*x[0]+(0.622558337967)*x[1]
        arg[2,2,2,0]=(-0.765475308614)*x[0]+(1.40320055807)*x[1]
        arg[2,2,2,1]=(-1.23199081441)*x[0]+(0.245517841221)*x[1]
        arg[2,3,0,0]=(0.655775411743)*x[0]+(-1.09496249546)*x[1]
        arg[2,3,0,1]=(-1.57906394006)*x[0]+(-0.223321547826)*x[1]
        arg[2,3,1,0]=(-0.329934697034)*x[0]+(-1.36492062074)*x[1]
        arg[2,3,1,1]=(1.2584701256)*x[0]+(0.172958559745)*x[1]
        arg[2,3,2,0]=(0.265769150885)*x[0]+(-0.273148332255)*x[1]
        arg[2,3,2,1]=(-1.18294356647)*x[0]+(-0.269813147795)*x[1]
        arg[2,4,0,0]=(-1.80352122024)*x[0]+(-0.417693287507)*x[1]
        arg[2,4,0,1]=(0.829979623343)*x[0]+(0.0526239642579)*x[1]
        arg[2,4,1,0]=(0.967999103063)*x[0]+(0.0184170526665)*x[1]
        arg[2,4,1,1]=(-0.326938715134)*x[0]+(-0.575040647639)*x[1]
        arg[2,4,2,0]=(1.44727400099)*x[0]+(0.710269500685)*x[1]
        arg[2,4,2,1]=(-0.248166665809)*x[0]+(0.247820501661)*x[1]
        arg[3,0,0,0]=(0.0463415093786)*x[0]+(0.650085996364)*x[1]
        arg[3,0,0,1]=(-0.440939509697)*x[0]+(-0.323307669473)*x[1]
        arg[3,0,1,0]=(0.273356160699)*x[0]+(-0.66721600197)*x[1]
        arg[3,0,1,1]=(1.29465477714)*x[0]+(-0.37486215111)*x[1]
        arg[3,0,2,0]=(-0.809896216106)*x[0]+(-1.51076103892)*x[1]
        arg[3,0,2,1]=(0.686428827303)*x[0]+(-0.538164742287)*x[1]
        arg[3,1,0,0]=(-0.863651610678)*x[0]+(0.673803453645)*x[1]
        arg[3,1,0,1]=(0.526827519816)*x[0]+(-0.0379595314201)*x[1]
        arg[3,1,1,0]=(-1.04759557033)*x[0]+(0.114198098457)*x[1]
        arg[3,1,1,1]=(0.552205171704)*x[0]+(0.116364312378)*x[1]
        arg[3,1,2,0]=(-0.926549911354)*x[0]+(0.296927704155)*x[1]
        arg[3,1,2,1]=(0.46702104365)*x[0]+(0.355015968814)*x[1]
        arg[3,2,0,0]=(0.603177046544)*x[0]+(-0.600600764625)*x[1]
        arg[3,2,0,1]=(0.656827398088)*x[0]+(1.19922812106)*x[1]
        arg[3,2,1,0]=(-0.0185352951675)*x[0]+(-0.575006610999)*x[1]
        arg[3,2,1,1]=(0.954361462974)*x[0]+(-0.239751892125)*x[1]
        arg[3,2,2,0]=(0.880568376655)*x[0]+(1.46879509876)*x[1]
        arg[3,2,2,1]=(-1.19051913627)*x[0]+(-0.0387119361514)*x[1]
        arg[3,3,0,0]=(1.19520855834)*x[0]+(1.3515728248)*x[1]
        arg[3,3,0,1]=(0.695487159903)*x[0]+(0.783460154378)*x[1]
        arg[3,3,1,0]=(0.482352542106)*x[0]+(-1.321707654)*x[1]
        arg[3,3,1,1]=(0.0637384915311)*x[0]+(-0.5903067787)*x[1]
        arg[3,3,2,0]=(0.571974795979)*x[0]+(1.33251727342)*x[1]
        arg[3,3,2,1]=(0.579723470137)*x[0]+(-0.260885275522)*x[1]
        arg[3,4,0,0]=(0.0717612359082)*x[0]+(-0.576982745638)*x[1]
        arg[3,4,0,1]=(-0.227660227523)*x[0]+(0.232974094548)*x[1]
        arg[3,4,1,0]=(0.767797269581)*x[0]+(0.98857043831)*x[1]
        arg[3,4,1,1]=(0.31200408415)*x[0]+(-0.8317658048)*x[1]
        arg[3,4,2,0]=(0.381688388928)*x[0]+(-0.335620162467)*x[1]
        arg[3,4,2,1]=(0.781228510591)*x[0]+(0.960966953643)*x[1]
        ref[0,0,0,0]=(0.507324804235)*x_ref[0]+(0.725831700089)*x_ref[1]
        ref[0,0,0,1]=(0.847901454936)*x_ref[0]+(0.84787024099)*x_ref[1]
        ref[0,0,1,0]=(0.221488465784)*x_ref[0]+(0.0449286988708)*x_ref[1]
        ref[0,0,1,1]=(0.00897176578501)*x_ref[0]+(-0.799118487206)*x_ref[1]
        ref[0,0,2,0]=(0.171491682599)*x_ref[0]+(1.0259065872)*x_ref[1]
        ref[0,0,2,1]=(-0.213244683798)*x_ref[0]+(-0.706816385516)*x_ref[1]
        ref[0,1,0,0]=(1.54199875186)*x_ref[0]+(-1.09479402082)*x_ref[1]
        ref[0,1,0,1]=(-0.755278051349)*x_ref[0]+(-0.371785089226)*x_ref[1]
        ref[0,1,1,0]=(-0.798931370137)*x_ref[0]+(-0.716047155117)*x_ref[1]
        ref[0,1,1,1]=(-0.79104072521)*x_ref[0]+(-1.45842710768)*x_ref[1]
        ref[0,1,2,0]=(-0.489161301965)*x_ref[0]+(0.115368605994)*x_ref[1]
        ref[0,1,2,1]=(0.00167228618616)*x_ref[0]+(-0.636896504873)*x_ref[1]
        ref[0,2,0,0]=(-0.239182729045)*x_ref[0]+(0.470786327067)*x_ref[1]
        ref[0,2,0,1]=(0.182105943188)*x_ref[0]+(0.328443662342)*x_ref[1]
        ref[0,2,1,0]=(0.000476540430199)*x_ref[0]+(-0.498066512126)*x_ref[1]
        ref[0,2,1,1]=(-0.822985659305)*x_ref[0]+(-0.118645880185)*x_ref[1]
        ref[0,2,2,0]=(1.18375382554)*x_ref[0]+(1.51417307474)*x_ref[1]
        ref[0,2,2,1]=(-0.99755868797)*x_ref[0]+(-1.4187068658)*x_ref[1]
        ref[0,3,0,0]=(-0.0908782314954)*x_ref[0]+(-0.25901340471)*x_ref[1]
        ref[0,3,0,1]=(-1.77626335863)*x_ref[0]+(-0.862132832663)*x_ref[1]
        ref[0,3,1,0]=(0.532147327491)*x_ref[0]+(0.454949026506)*x_ref[1]
        ref[0,3,1,1]=(-0.882307958628)*x_ref[0]+(0.00780544746858)*x_ref[1]
        ref[0,3,2,0]=(0.700104420574)*x_ref[0]+(0.720508036709)*x_ref[1]
        ref[0,3,2,1]=(-1.27167118811)*x_ref[0]+(-0.372999410908)*x_ref[1]
        ref[0,4,0,0]=(0.396222903345)*x_ref[0]+(1.70664799562)*x_ref[1]
        ref[0,4,0,1]=(0.229159055372)*x_ref[0]+(0.0162677758439)*x_ref[1]
        ref[0,4,1,0]=(1.69470499818)*x_ref[0]+(0.0971804891468)*x_ref[1]
        ref[0,4,1,1]=(0.180710573838)*x_ref[0]+(1.05311557127)*x_ref[1]
        ref[0,4,2,0]=(-0.072661496917)*x_ref[0]+(0.466256643644)*x_ref[1]
        ref[0,4,2,1]=(0.25755631084)*x_ref[0]+(0.306990925808)*x_ref[1]
        ref[1,0,0,0]=(0.835353191317)*x_ref[0]+(-0.619237060145)*x_ref[1]
        ref[1,0,0,1]=(1.02966866371)*x_ref[0]+(0.327724003031)*x_ref[1]
        ref[1,0,1,0]=(-0.59214788986)*x_ref[0]+(-0.927738485428)*x_ref[1]
        ref[1,0,1,1]=(-0.805929278446)*x_ref[0]+(-0.366113927406)*x_ref[1]
        ref[1,0,2,0]=(0.952876155721)*x_ref[0]+(0.490531201467)*x_ref[1]
        ref[1,0,2,1]=(1.44341651657)*x_ref[0]+(-0.269899260772)*x_ref[1]
        ref[1,1,0,0]=(0.302587966785)*x_ref[0]+(-0.0373883991695)*x_ref[1]
        ref[1,1,0,1]=(0.77498995648)*x_ref[0]+(1.81052574555)*x_ref[1]
        ref[1,1,1,0]=(1.47950560134)*x_ref[0]+(-0.714878219566)*x_ref[1]
        ref[1,1,1,1]=(0.0713827629606)*x_ref[0]+(-1.22018291576)*x_ref[1]
        ref[1,1,2,0]=(1.30715631604)*x_ref[0]+(-0.191666253513)*x_ref[1]
        ref[1,1,2,1]=(0.330921962628)*x_ref[0]+(-1.57160261912)*x_ref[1]
        ref[1,2,0,0]=(0.593477126597)*x_ref[0]+(0.728196634765)*x_ref[1]
        ref[1,2,0,1]=(0.853860342143)*x_ref[0]+(0.142625101433)*x_ref[1]
        ref[1,2,1,0]=(-0.743590115343)*x_ref[0]+(0.108893736366)*x_ref[1]
        ref[1,2,1,1]=(-0.651514863459)*x_ref[0]+(0.365517180762)*x_ref[1]
        ref[1,2,2,0]=(1.491516719)*x_ref[0]+(0.28900806658)*x_ref[1]
        ref[1,2,2,1]=(1.23733736339)*x_ref[0]+(-0.58438294882)*x_ref[1]
        ref[1,3,0,0]=(0.185696018505)*x_ref[0]+(-1.29119084907)*x_ref[1]
        ref[1,3,0,1]=(1.73584656085)*x_ref[0]+(0.720989518003)*x_ref[1]
        ref[1,3,1,0]=(-1.16044192371)*x_ref[0]+(-0.131128994204)*x_ref[1]
        ref[1,3,1,1]=(-0.00913995119508)*x_ref[0]+(0.506516682557)*x_ref[1]
        ref[1,3,2,0]=(-1.4970171408)*x_ref[0]+(0.692262679219)*x_ref[1]
        ref[1,3,2,1]=(0.433442352395)*x_ref[0]+(-0.118782513258)*x_ref[1]
        ref[1,4,0,0]=(0.186272280804)*x_ref[0]+(-0.136627137097)*x_ref[1]
        ref[1,4,0,1]=(1.25477168515)*x_ref[0]+(-0.0871831369474)*x_ref[1]
        ref[1,4,1,0]=(1.22746022677)*x_ref[0]+(0.799329883559)*x_ref[1]
        ref[1,4,1,1]=(-0.756298280123)*x_ref[0]+(0.317500967634)*x_ref[1]
        ref[1,4,2,0]=(0.201184703435)*x_ref[0]+(-0.116537935236)*x_ref[1]
        ref[1,4,2,1]=(0.851291897607)*x_ref[0]+(0.404063606516)*x_ref[1]
        ref[2,0,0,0]=(-0.257605370026)*x_ref[0]+(-0.202424701671)*x_ref[1]
        ref[2,0,0,1]=(0.0219332720675)*x_ref[0]+(0.518222103227)*x_ref[1]
        ref[2,0,1,0]=(1.03726036144)*x_ref[0]+(-1.01482867596)*x_ref[1]
        ref[2,0,1,1]=(0.00677025785109)*x_ref[0]+(0.607932484906)*x_ref[1]
        ref[2,0,2,0]=(-0.753000974637)*x_ref[0]+(0.112884693676)*x_ref[1]
        ref[2,0,2,1]=(-0.585627190582)*x_ref[0]+(0.548136242722)*x_ref[1]
        ref[2,1,0,0]=(0.382466685883)*x_ref[0]+(-1.15520454298)*x_ref[1]
        ref[2,1,0,1]=(-0.722705123323)*x_ref[0]+(-0.599069291154)*x_ref[1]
        ref[2,1,1,0]=(-0.845885208256)*x_ref[0]+(-1.56066350212)*x_ref[1]
        ref[2,1,1,1]=(-1.31472469553)*x_ref[0]+(0.18973548452)*x_ref[1]
        ref[2,1,2,0]=(-0.419142599733)*x_ref[0]+(1.43873372527)*x_ref[1]
        ref[2,1,2,1]=(0.22685530286)*x_ref[0]+(-0.53636714393)*x_ref[1]
        ref[2,2,0,0]=(1.32884214935)*x_ref[0]+(-0.377575726395)*x_ref[1]
        ref[2,2,0,1]=(0.0845542736847)*x_ref[0]+(0.0888339000429)*x_ref[1]
        ref[2,2,1,0]=(-1.05813285521)*x_ref[0]+(-0.980382395463)*x_ref[1]
        ref[2,2,1,1]=(0.306785463268)*x_ref[0]+(0.622558337967)*x_ref[1]
        ref[2,2,2,0]=(-0.765475308614)*x_ref[0]+(1.40320055807)*x_ref[1]
        ref[2,2,2,1]=(-1.23199081441)*x_ref[0]+(0.245517841221)*x_ref[1]
        ref[2,3,0,0]=(0.655775411743)*x_ref[0]+(-1.09496249546)*x_ref[1]
        ref[2,3,0,1]=(-1.57906394006)*x_ref[0]+(-0.223321547826)*x_ref[1]
        ref[2,3,1,0]=(-0.329934697034)*x_ref[0]+(-1.36492062074)*x_ref[1]
        ref[2,3,1,1]=(1.2584701256)*x_ref[0]+(0.172958559745)*x_ref[1]
        ref[2,3,2,0]=(0.265769150885)*x_ref[0]+(-0.273148332255)*x_ref[1]
        ref[2,3,2,1]=(-1.18294356647)*x_ref[0]+(-0.269813147795)*x_ref[1]
        ref[2,4,0,0]=(-1.80352122024)*x_ref[0]+(-0.417693287507)*x_ref[1]
        ref[2,4,0,1]=(0.829979623343)*x_ref[0]+(0.0526239642579)*x_ref[1]
        ref[2,4,1,0]=(0.967999103063)*x_ref[0]+(0.0184170526665)*x_ref[1]
        ref[2,4,1,1]=(-0.326938715134)*x_ref[0]+(-0.575040647639)*x_ref[1]
        ref[2,4,2,0]=(1.44727400099)*x_ref[0]+(0.710269500685)*x_ref[1]
        ref[2,4,2,1]=(-0.248166665809)*x_ref[0]+(0.247820501661)*x_ref[1]
        ref[3,0,0,0]=(0.0463415093786)*x_ref[0]+(0.650085996364)*x_ref[1]
        ref[3,0,0,1]=(-0.440939509697)*x_ref[0]+(-0.323307669473)*x_ref[1]
        ref[3,0,1,0]=(0.273356160699)*x_ref[0]+(-0.66721600197)*x_ref[1]
        ref[3,0,1,1]=(1.29465477714)*x_ref[0]+(-0.37486215111)*x_ref[1]
        ref[3,0,2,0]=(-0.809896216106)*x_ref[0]+(-1.51076103892)*x_ref[1]
        ref[3,0,2,1]=(0.686428827303)*x_ref[0]+(-0.538164742287)*x_ref[1]
        ref[3,1,0,0]=(-0.863651610678)*x_ref[0]+(0.673803453645)*x_ref[1]
        ref[3,1,0,1]=(0.526827519816)*x_ref[0]+(-0.0379595314201)*x_ref[1]
        ref[3,1,1,0]=(-1.04759557033)*x_ref[0]+(0.114198098457)*x_ref[1]
        ref[3,1,1,1]=(0.552205171704)*x_ref[0]+(0.116364312378)*x_ref[1]
        ref[3,1,2,0]=(-0.926549911354)*x_ref[0]+(0.296927704155)*x_ref[1]
        ref[3,1,2,1]=(0.46702104365)*x_ref[0]+(0.355015968814)*x_ref[1]
        ref[3,2,0,0]=(0.603177046544)*x_ref[0]+(-0.600600764625)*x_ref[1]
        ref[3,2,0,1]=(0.656827398088)*x_ref[0]+(1.19922812106)*x_ref[1]
        ref[3,2,1,0]=(-0.0185352951675)*x_ref[0]+(-0.575006610999)*x_ref[1]
        ref[3,2,1,1]=(0.954361462974)*x_ref[0]+(-0.239751892125)*x_ref[1]
        ref[3,2,2,0]=(0.880568376655)*x_ref[0]+(1.46879509876)*x_ref[1]
        ref[3,2,2,1]=(-1.19051913627)*x_ref[0]+(-0.0387119361514)*x_ref[1]
        ref[3,3,0,0]=(1.19520855834)*x_ref[0]+(1.3515728248)*x_ref[1]
        ref[3,3,0,1]=(0.695487159903)*x_ref[0]+(0.783460154378)*x_ref[1]
        ref[3,3,1,0]=(0.482352542106)*x_ref[0]+(-1.321707654)*x_ref[1]
        ref[3,3,1,1]=(0.0637384915311)*x_ref[0]+(-0.5903067787)*x_ref[1]
        ref[3,3,2,0]=(0.571974795979)*x_ref[0]+(1.33251727342)*x_ref[1]
        ref[3,3,2,1]=(0.579723470137)*x_ref[0]+(-0.260885275522)*x_ref[1]
        ref[3,4,0,0]=(0.0717612359082)*x_ref[0]+(-0.576982745638)*x_ref[1]
        ref[3,4,0,1]=(-0.227660227523)*x_ref[0]+(0.232974094548)*x_ref[1]
        ref[3,4,1,0]=(0.767797269581)*x_ref[0]+(0.98857043831)*x_ref[1]
        ref[3,4,1,1]=(0.31200408415)*x_ref[0]+(-0.8317658048)*x_ref[1]
        ref[3,4,2,0]=(0.381688388928)*x_ref[0]+(-0.335620162467)*x_ref[1]
        ref[3,4,2,1]=(0.781228510591)*x_ref[0]+(0.960966953643)*x_ref[1]
      else:
        arg[0,0,0,0]=(-0.875410081875)*x[0]+(0.859961622438)*x[1]+(0.716805919947)*x[2]
        arg[0,0,0,1]=(0.343337453662)*x[0]+(0.802028072985)*x[1]+(0.588370750348)*x[2]
        arg[0,0,1,0]=(0.467615640242)*x[0]+(0.0457804028001)*x[1]+(-0.163677176662)*x[2]
        arg[0,0,1,1]=(-0.46378923457)*x[0]+(1.29379360824)*x[1]+(0.183839231748)*x[2]
        arg[0,0,2,0]=(1.65864647053)*x[0]+(1.23523675105)*x[1]+(-0.556155025527)*x[2]
        arg[0,0,2,1]=(-0.41581445233)*x[0]+(-0.0709880820438)*x[1]+(0.45785589934)*x[2]
        arg[0,1,0,0]=(1.83790068156)*x[0]+(-0.1690836925)*x[1]+(1.25356286366)*x[2]
        arg[0,1,0,1]=(-0.22584314958)*x[0]+(0.684595424346)*x[1]+(1.13919885217)*x[2]
        arg[0,1,1,0]=(1.02792150528)*x[0]+(-0.312817626145)*x[1]+(-0.121465758957)*x[2]
        arg[0,1,1,1]=(0.533491564413)*x[0]+(-0.62208812136)*x[1]+(-0.717947781071)*x[2]
        arg[0,1,2,0]=(-0.864505701555)*x[0]+(-0.296047724934)*x[1]+(-0.122917018588)*x[2]
        arg[0,1,2,1]=(1.05520768182)*x[0]+(0.662053590424)*x[1]+(0.0251841065476)*x[2]
        arg[0,2,0,0]=(-0.408305818721)*x[0]+(-0.355992587618)*x[1]+(-0.075231462678)*x[2]
        arg[0,2,0,1]=(0.926563898549)*x[0]+(0.203358704647)*x[1]+(-0.365269987808)*x[2]
        arg[0,2,1,0]=(-0.403860257745)*x[0]+(-0.0312835722524)*x[1]+(0.241379582617)*x[2]
        arg[0,2,1,1]=(-0.287297815371)*x[0]+(-0.381813267253)*x[1]+(0.46047051318)*x[2]
        arg[0,2,2,0]=(-0.168164506906)*x[0]+(-1.24152426841)*x[1]+(-0.629704769515)*x[2]
        arg[0,2,2,1]=(-0.937219354291)*x[0]+(-0.736950161299)*x[1]+(0.284057170092)*x[2]
        arg[0,3,0,0]=(-0.292205445759)*x[0]+(1.21842674086)*x[1]+(0.608979796402)*x[2]
        arg[0,3,0,1]=(0.53232240924)*x[0]+(0.226439017439)*x[1]+(-0.42603939944)*x[2]
        arg[0,3,1,0]=(1.67410358342)*x[0]+(-0.00813619771347)*x[1]+(-0.463361889732)*x[2]
        arg[0,3,1,1]=(1.3991502461)*x[0]+(0.239999857694)*x[1]+(-0.369907067208)*x[2]
        arg[0,3,2,0]=(-0.248767913403)*x[0]+(0.893584158229)*x[1]+(1.97656682548)*x[2]
        arg[0,3,2,1]=(-0.835046144341)*x[0]+(-0.00979850832247)*x[1]+(-0.281439904932)*x[2]
        arg[0,4,0,0]=(0.974179494371)*x[0]+(-0.0432121220464)*x[1]+(-0.475018088554)*x[2]
        arg[0,4,0,1]=(0.245490151034)*x[0]+(-0.367572047585)*x[1]+(-0.926123425123)*x[2]
        arg[0,4,1,0]=(-0.239057491277)*x[0]+(0.339734302574)*x[1]+(0.760546363123)*x[2]
        arg[0,4,1,1]=(-1.0425426898)*x[0]+(0.774210702874)*x[1]+(-1.15990543484)*x[2]
        arg[0,4,2,0]=(1.02994063274)*x[0]+(1.69313195574)*x[1]+(-1.02468862318)*x[2]
        arg[0,4,2,1]=(0.221805673731)*x[0]+(-1.06588014758)*x[1]+(-0.546973870537)*x[2]
        arg[1,0,0,0]=(-0.140765393552)*x[0]+(-0.309057115268)*x[1]+(0.07814948965)*x[2]
        arg[1,0,0,1]=(-0.29622101259)*x[0]+(-1.0654452087)*x[1]+(1.77500323701)*x[2]
        arg[1,0,1,0]=(1.13202947738)*x[0]+(-0.515454892388)*x[1]+(-1.34230346013)*x[2]
        arg[1,0,1,1]=(1.50210238409)*x[0]+(-0.75673088573)*x[1]+(-0.792078632258)*x[2]
        arg[1,0,2,0]=(0.249111383954)*x[0]+(-0.972418007776)*x[1]+(0.766060429469)*x[2]
        arg[1,0,2,1]=(-0.0710717292409)*x[0]+(0.164949706666)*x[1]+(1.46045331387)*x[2]
        arg[1,1,0,0]=(-0.587668209291)*x[0]+(0.303249169818)*x[1]+(0.352197808782)*x[2]
        arg[1,1,0,1]=(-0.145609904248)*x[0]+(0.91895760024)*x[1]+(-1.25063091478)*x[2]
        arg[1,1,1,0]=(0.450869946651)*x[0]+(0.752704827905)*x[1]+(0.132042943533)*x[2]
        arg[1,1,1,1]=(-0.87296460294)*x[0]+(-0.459840609019)*x[1]+(0.422574948491)*x[2]
        arg[1,1,2,0]=(0.828459269131)*x[0]+(0.45869153936)*x[1]+(0.854640684176)*x[2]
        arg[1,1,2,1]=(1.84720399473)*x[0]+(0.0416642546234)*x[1]+(-0.309583524852)*x[2]
        arg[1,2,0,0]=(-1.12563293246)*x[0]+(0.0142184473972)*x[1]+(-1.18150362584)*x[2]
        arg[1,2,0,1]=(-0.0750965768147)*x[0]+(-0.419276687751)*x[1]+(0.521012213863)*x[2]
        arg[1,2,1,0]=(-1.06905807284)*x[0]+(0.692258030743)*x[1]+(0.262254909068)*x[2]
        arg[1,2,1,1]=(0.0552441016888)*x[0]+(-0.131040560322)*x[1]+(-0.321517725963)*x[2]
        arg[1,2,2,0]=(-0.740410514693)*x[0]+(0.219020143899)*x[1]+(1.35104682965)*x[2]
        arg[1,2,2,1]=(-1.38198583939)*x[0]+(-0.19342887842)*x[1]+(-0.780008129494)*x[2]
        arg[1,3,0,0]=(0.378358305605)*x[0]+(1.32138618734)*x[1]+(0.627209784997)*x[2]
        arg[1,3,0,1]=(-0.0383657917537)*x[0]+(-1.41770189628)*x[1]+(-1.21625280681)*x[2]
        arg[1,3,1,0]=(0.819666413875)*x[0]+(0.160369878383)*x[1]+(-0.229388558682)*x[2]
        arg[1,3,1,1]=(0.393409555876)*x[0]+(-0.525985341296)*x[1]+(-1.11892438309)*x[2]
        arg[1,3,2,0]=(0.145077010777)*x[0]+(0.229800008438)*x[1]+(-0.202652862622)*x[2]
        arg[1,3,2,1]=(-1.48698378177)*x[0]+(0.553698672636)*x[1]+(-0.427263736796)*x[2]
        arg[1,4,0,0]=(-0.129527725249)*x[0]+(-0.639824801211)*x[1]+(0.759081980717)*x[2]
        arg[1,4,0,1]=(-0.117772473784)*x[0]+(0.387474104611)*x[1]+(0.507306747135)*x[2]
        arg[1,4,1,0]=(-0.82518622198)*x[0]+(-0.754474320343)*x[1]+(-0.978049152511)*x[2]
        arg[1,4,1,1]=(-0.370385633483)*x[0]+(-0.348429423043)*x[1]+(-0.254278257899)*x[2]
        arg[1,4,2,0]=(1.42596167697)*x[0]+(-0.871785904514)*x[1]+(-0.5343959325)*x[2]
        arg[1,4,2,1]=(0.817877672506)*x[0]+(-0.222622791737)*x[1]+(-0.677237468595)*x[2]
        arg[2,0,0,0]=(0.877660484813)*x[0]+(0.818663846913)*x[1]+(-1.26520874625)*x[2]
        arg[2,0,0,1]=(-0.822344385073)*x[0]+(-0.274691696266)*x[1]+(-0.35079248697)*x[2]
        arg[2,0,1,0]=(-0.319040285312)*x[0]+(-0.750406131559)*x[1]+(-0.920332967032)*x[2]
        arg[2,0,1,1]=(1.48758278445)*x[0]+(1.30587557873)*x[1]+(0.230692557888)*x[2]
        arg[2,0,2,0]=(1.44269354043)*x[0]+(-0.63270356894)*x[1]+(0.229323632455)*x[2]
        arg[2,0,2,1]=(-0.645065877536)*x[0]+(-0.393083499676)*x[1]+(-0.305117645453)*x[2]
        arg[2,1,0,0]=(-0.63304134143)*x[0]+(0.0664768401917)*x[1]+(0.87587743066)*x[2]
        arg[2,1,0,1]=(-0.531856958285)*x[0]+(0.747953146648)*x[1]+(1.23560779546)*x[2]
        arg[2,1,1,0]=(0.0564479101025)*x[0]+(0.546307803332)*x[1]+(0.655044890368)*x[2]
        arg[2,1,1,1]=(-1.12793057228)*x[0]+(-0.736676194013)*x[1]+(0.0375913829411)*x[2]
        arg[2,1,2,0]=(-0.531384188067)*x[0]+(0.249416904081)*x[1]+(0.711430712986)*x[2]
        arg[2,1,2,1]=(0.884244664065)*x[0]+(0.456066941968)*x[1]+(0.181630861785)*x[2]
        arg[2,2,0,0]=(0.0844543345062)*x[0]+(-0.435288041256)*x[1]+(-0.434816763062)*x[2]
        arg[2,2,0,1]=(0.373388105049)*x[0]+(0.239833846429)*x[1]+(-0.437771573282)*x[2]
        arg[2,2,1,0]=(0.534401521907)*x[0]+(0.393077203448)*x[1]+(-0.42440387717)*x[2]
        arg[2,2,1,1]=(0.318683369617)*x[0]+(-0.538864643793)*x[1]+(1.83466945806)*x[2]
        arg[2,2,2,0]=(-1.79048122074)*x[0]+(-1.05441213478)*x[1]+(-0.573963806281)*x[2]
        arg[2,2,2,1]=(0.944017405182)*x[0]+(0.772756791538)*x[1]+(0.0269673237336)*x[2]
        arg[2,3,0,0]=(0.989703689142)*x[0]+(1.21302325945)*x[1]+(-0.92209419073)*x[2]
        arg[2,3,0,1]=(-0.0957244757576)*x[0]+(1.50132638315)*x[1]+(-1.38716099756)*x[2]
        arg[2,3,1,0]=(1.6445441776)*x[0]+(-0.490207070694)*x[1]+(0.626469347781)*x[2]
        arg[2,3,1,1]=(0.0185568939147)*x[0]+(0.412024752941)*x[1]+(-0.22626102813)*x[2]
        arg[2,3,2,0]=(0.0143609822966)*x[0]+(0.115739025893)*x[1]+(-0.544468631925)*x[2]
        arg[2,3,2,1]=(-1.80522341493)*x[0]+(0.33352152695)*x[1]+(-0.841286464932)*x[2]
        arg[2,4,0,0]=(0.0618444122042)*x[0]+(1.91232484622)*x[1]+(-0.406317200023)*x[2]
        arg[2,4,0,1]=(0.279131062922)*x[0]+(0.0194232637634)*x[1]+(-0.00179069113634)*x[2]
        arg[2,4,1,0]=(-0.701344579135)*x[0]+(-0.27410935473)*x[1]+(0.763362100733)*x[2]
        arg[2,4,1,1]=(0.831198601185)*x[0]+(0.90336781771)*x[1]+(1.21856633382)*x[2]
        arg[2,4,2,0]=(0.526278615033)*x[0]+(-1.08673855801)*x[1]+(0.0784355900913)*x[2]
        arg[2,4,2,1]=(0.853226856851)*x[0]+(-1.15919062627)*x[1]+(0.185686031732)*x[2]
        arg[3,0,0,0]=(0.272127498117)*x[0]+(-0.457140162438)*x[1]+(0.318358670541)*x[2]
        arg[3,0,0,1]=(1.71056378307)*x[0]+(0.200052730234)*x[1]+(-0.127733325792)*x[2]
        arg[3,0,1,0]=(-0.285340963261)*x[0]+(-0.524544624385)*x[1]+(0.916366474944)*x[2]
        arg[3,0,1,1]=(-0.625506557557)*x[0]+(1.0858476479)*x[1]+(-1.24478983268)*x[2]
        arg[3,0,2,0]=(-1.37299619916)*x[0]+(-0.482246528202)*x[1]+(-1.36857669075)*x[2]
        arg[3,0,2,1]=(-0.924421844916)*x[0]+(-1.50402098115)*x[1]+(1.00045130673)*x[2]
        arg[3,1,0,0]=(0.0417304700827)*x[0]+(0.554527451004)*x[1]+(0.299758163628)*x[2]
        arg[3,1,0,1]=(0.639611639279)*x[0]+(-0.444675638162)*x[1]+(1.43150046757)*x[2]
        arg[3,1,1,0]=(0.288697220898)*x[0]+(1.32660433773)*x[1]+(0.0889412116759)*x[2]
        arg[3,1,1,1]=(-0.110516770201)*x[0]+(-0.0214248733565)*x[1]+(-0.758638794785)*x[2]
        arg[3,1,2,0]=(1.34357138654)*x[0]+(-0.13759272972)*x[1]+(-0.00183845032784)*x[2]
        arg[3,1,2,1]=(1.44174769627)*x[0]+(0.49190176388)*x[1]+(0.477331360141)*x[2]
        arg[3,2,0,0]=(-0.10284777121)*x[0]+(-0.63294157419)*x[1]+(0.246064460451)*x[2]
        arg[3,2,0,1]=(0.974270980699)*x[0]+(1.03835886161)*x[1]+(0.420000854838)*x[2]
        arg[3,2,1,0]=(-0.00617831615923)*x[0]+(-0.354063634073)*x[1]+(0.764285884619)*x[2]
        arg[3,2,1,1]=(0.210815790746)*x[0]+(0.76609515744)*x[1]+(-0.564401558401)*x[2]
        arg[3,2,2,0]=(1.30471234011)*x[0]+(1.54864119332)*x[1]+(-0.493583650519)*x[2]
        arg[3,2,2,1]=(0.42287513668)*x[0]+(-0.603093868171)*x[1]+(-0.811892974203)*x[2]
        arg[3,3,0,0]=(-1.38358134693)*x[0]+(-1.53706018098)*x[1]+(-1.29701227501)*x[2]
        arg[3,3,0,1]=(0.598581779902)*x[0]+(1.02424010922)*x[1]+(-1.1814313009)*x[2]
        arg[3,3,1,0]=(1.30938529116)*x[0]+(-0.504468477244)*x[1]+(1.63655848854)*x[2]
        arg[3,3,1,1]=(-1.47006424113)*x[0]+(0.225951520203)*x[1]+(-0.412273003169)*x[2]
        arg[3,3,2,0]=(-0.116902234482)*x[0]+(0.488144536044)*x[1]+(1.01732461694)*x[2]
        arg[3,3,2,1]=(1.0935814877)*x[0]+(1.06389385916)*x[1]+(-0.17981400591)*x[2]
        arg[3,4,0,0]=(-0.833265062101)*x[0]+(-0.754461148648)*x[1]+(1.26702012815)*x[2]
        arg[3,4,0,1]=(-0.793744278171)*x[0]+(0.112704521653)*x[1]+(-0.448684922187)*x[2]
        arg[3,4,1,0]=(-0.396153523625)*x[0]+(-0.0370870723568)*x[1]+(0.554347618511)*x[2]
        arg[3,4,1,1]=(-0.323276139711)*x[0]+(-0.397389928392)*x[1]+(-1.1707028324)*x[2]
        arg[3,4,2,0]=(0.823350512513)*x[0]+(1.47105759864)*x[1]+(-0.977456877855)*x[2]
        arg[3,4,2,1]=(-0.719791170741)*x[0]+(-1.00363340645)*x[1]+(0.814612770572)*x[2]
        ref[0,0,0,0]=(-0.875410081875)*x_ref[0]+(0.859961622438)*x_ref[1]+(0.716805919947)*x_ref[2]
        ref[0,0,0,1]=(0.343337453662)*x_ref[0]+(0.802028072985)*x_ref[1]+(0.588370750348)*x_ref[2]
        ref[0,0,1,0]=(0.467615640242)*x_ref[0]+(0.0457804028001)*x_ref[1]+(-0.163677176662)*x_ref[2]
        ref[0,0,1,1]=(-0.46378923457)*x_ref[0]+(1.29379360824)*x_ref[1]+(0.183839231748)*x_ref[2]
        ref[0,0,2,0]=(1.65864647053)*x_ref[0]+(1.23523675105)*x_ref[1]+(-0.556155025527)*x_ref[2]
        ref[0,0,2,1]=(-0.41581445233)*x_ref[0]+(-0.0709880820438)*x_ref[1]+(0.45785589934)*x_ref[2]
        ref[0,1,0,0]=(1.83790068156)*x_ref[0]+(-0.1690836925)*x_ref[1]+(1.25356286366)*x_ref[2]
        ref[0,1,0,1]=(-0.22584314958)*x_ref[0]+(0.684595424346)*x_ref[1]+(1.13919885217)*x_ref[2]
        ref[0,1,1,0]=(1.02792150528)*x_ref[0]+(-0.312817626145)*x_ref[1]+(-0.121465758957)*x_ref[2]
        ref[0,1,1,1]=(0.533491564413)*x_ref[0]+(-0.62208812136)*x_ref[1]+(-0.717947781071)*x_ref[2]
        ref[0,1,2,0]=(-0.864505701555)*x_ref[0]+(-0.296047724934)*x_ref[1]+(-0.122917018588)*x_ref[2]
        ref[0,1,2,1]=(1.05520768182)*x_ref[0]+(0.662053590424)*x_ref[1]+(0.0251841065476)*x_ref[2]
        ref[0,2,0,0]=(-0.408305818721)*x_ref[0]+(-0.355992587618)*x_ref[1]+(-0.075231462678)*x_ref[2]
        ref[0,2,0,1]=(0.926563898549)*x_ref[0]+(0.203358704647)*x_ref[1]+(-0.365269987808)*x_ref[2]
        ref[0,2,1,0]=(-0.403860257745)*x_ref[0]+(-0.0312835722524)*x_ref[1]+(0.241379582617)*x_ref[2]
        ref[0,2,1,1]=(-0.287297815371)*x_ref[0]+(-0.381813267253)*x_ref[1]+(0.46047051318)*x_ref[2]
        ref[0,2,2,0]=(-0.168164506906)*x_ref[0]+(-1.24152426841)*x_ref[1]+(-0.629704769515)*x_ref[2]
        ref[0,2,2,1]=(-0.937219354291)*x_ref[0]+(-0.736950161299)*x_ref[1]+(0.284057170092)*x_ref[2]
        ref[0,3,0,0]=(-0.292205445759)*x_ref[0]+(1.21842674086)*x_ref[1]+(0.608979796402)*x_ref[2]
        ref[0,3,0,1]=(0.53232240924)*x_ref[0]+(0.226439017439)*x_ref[1]+(-0.42603939944)*x_ref[2]
        ref[0,3,1,0]=(1.67410358342)*x_ref[0]+(-0.00813619771347)*x_ref[1]+(-0.463361889732)*x_ref[2]
        ref[0,3,1,1]=(1.3991502461)*x_ref[0]+(0.239999857694)*x_ref[1]+(-0.369907067208)*x_ref[2]
        ref[0,3,2,0]=(-0.248767913403)*x_ref[0]+(0.893584158229)*x_ref[1]+(1.97656682548)*x_ref[2]
        ref[0,3,2,1]=(-0.835046144341)*x_ref[0]+(-0.00979850832247)*x_ref[1]+(-0.281439904932)*x_ref[2]
        ref[0,4,0,0]=(0.974179494371)*x_ref[0]+(-0.0432121220464)*x_ref[1]+(-0.475018088554)*x_ref[2]
        ref[0,4,0,1]=(0.245490151034)*x_ref[0]+(-0.367572047585)*x_ref[1]+(-0.926123425123)*x_ref[2]
        ref[0,4,1,0]=(-0.239057491277)*x_ref[0]+(0.339734302574)*x_ref[1]+(0.760546363123)*x_ref[2]
        ref[0,4,1,1]=(-1.0425426898)*x_ref[0]+(0.774210702874)*x_ref[1]+(-1.15990543484)*x_ref[2]
        ref[0,4,2,0]=(1.02994063274)*x_ref[0]+(1.69313195574)*x_ref[1]+(-1.02468862318)*x_ref[2]
        ref[0,4,2,1]=(0.221805673731)*x_ref[0]+(-1.06588014758)*x_ref[1]+(-0.546973870537)*x_ref[2]
        ref[1,0,0,0]=(-0.140765393552)*x_ref[0]+(-0.309057115268)*x_ref[1]+(0.07814948965)*x_ref[2]
        ref[1,0,0,1]=(-0.29622101259)*x_ref[0]+(-1.0654452087)*x_ref[1]+(1.77500323701)*x_ref[2]
        ref[1,0,1,0]=(1.13202947738)*x_ref[0]+(-0.515454892388)*x_ref[1]+(-1.34230346013)*x_ref[2]
        ref[1,0,1,1]=(1.50210238409)*x_ref[0]+(-0.75673088573)*x_ref[1]+(-0.792078632258)*x_ref[2]
        ref[1,0,2,0]=(0.249111383954)*x_ref[0]+(-0.972418007776)*x_ref[1]+(0.766060429469)*x_ref[2]
        ref[1,0,2,1]=(-0.0710717292409)*x_ref[0]+(0.164949706666)*x_ref[1]+(1.46045331387)*x_ref[2]
        ref[1,1,0,0]=(-0.587668209291)*x_ref[0]+(0.303249169818)*x_ref[1]+(0.352197808782)*x_ref[2]
        ref[1,1,0,1]=(-0.145609904248)*x_ref[0]+(0.91895760024)*x_ref[1]+(-1.25063091478)*x_ref[2]
        ref[1,1,1,0]=(0.450869946651)*x_ref[0]+(0.752704827905)*x_ref[1]+(0.132042943533)*x_ref[2]
        ref[1,1,1,1]=(-0.87296460294)*x_ref[0]+(-0.459840609019)*x_ref[1]+(0.422574948491)*x_ref[2]
        ref[1,1,2,0]=(0.828459269131)*x_ref[0]+(0.45869153936)*x_ref[1]+(0.854640684176)*x_ref[2]
        ref[1,1,2,1]=(1.84720399473)*x_ref[0]+(0.0416642546234)*x_ref[1]+(-0.309583524852)*x_ref[2]
        ref[1,2,0,0]=(-1.12563293246)*x_ref[0]+(0.0142184473972)*x_ref[1]+(-1.18150362584)*x_ref[2]
        ref[1,2,0,1]=(-0.0750965768147)*x_ref[0]+(-0.419276687751)*x_ref[1]+(0.521012213863)*x_ref[2]
        ref[1,2,1,0]=(-1.06905807284)*x_ref[0]+(0.692258030743)*x_ref[1]+(0.262254909068)*x_ref[2]
        ref[1,2,1,1]=(0.0552441016888)*x_ref[0]+(-0.131040560322)*x_ref[1]+(-0.321517725963)*x_ref[2]
        ref[1,2,2,0]=(-0.740410514693)*x_ref[0]+(0.219020143899)*x_ref[1]+(1.35104682965)*x_ref[2]
        ref[1,2,2,1]=(-1.38198583939)*x_ref[0]+(-0.19342887842)*x_ref[1]+(-0.780008129494)*x_ref[2]
        ref[1,3,0,0]=(0.378358305605)*x_ref[0]+(1.32138618734)*x_ref[1]+(0.627209784997)*x_ref[2]
        ref[1,3,0,1]=(-0.0383657917537)*x_ref[0]+(-1.41770189628)*x_ref[1]+(-1.21625280681)*x_ref[2]
        ref[1,3,1,0]=(0.819666413875)*x_ref[0]+(0.160369878383)*x_ref[1]+(-0.229388558682)*x_ref[2]
        ref[1,3,1,1]=(0.393409555876)*x_ref[0]+(-0.525985341296)*x_ref[1]+(-1.11892438309)*x_ref[2]
        ref[1,3,2,0]=(0.145077010777)*x_ref[0]+(0.229800008438)*x_ref[1]+(-0.202652862622)*x_ref[2]
        ref[1,3,2,1]=(-1.48698378177)*x_ref[0]+(0.553698672636)*x_ref[1]+(-0.427263736796)*x_ref[2]
        ref[1,4,0,0]=(-0.129527725249)*x_ref[0]+(-0.639824801211)*x_ref[1]+(0.759081980717)*x_ref[2]
        ref[1,4,0,1]=(-0.117772473784)*x_ref[0]+(0.387474104611)*x_ref[1]+(0.507306747135)*x_ref[2]
        ref[1,4,1,0]=(-0.82518622198)*x_ref[0]+(-0.754474320343)*x_ref[1]+(-0.978049152511)*x_ref[2]
        ref[1,4,1,1]=(-0.370385633483)*x_ref[0]+(-0.348429423043)*x_ref[1]+(-0.254278257899)*x_ref[2]
        ref[1,4,2,0]=(1.42596167697)*x_ref[0]+(-0.871785904514)*x_ref[1]+(-0.5343959325)*x_ref[2]
        ref[1,4,2,1]=(0.817877672506)*x_ref[0]+(-0.222622791737)*x_ref[1]+(-0.677237468595)*x_ref[2]
        ref[2,0,0,0]=(0.877660484813)*x_ref[0]+(0.818663846913)*x_ref[1]+(-1.26520874625)*x_ref[2]
        ref[2,0,0,1]=(-0.822344385073)*x_ref[0]+(-0.274691696266)*x_ref[1]+(-0.35079248697)*x_ref[2]
        ref[2,0,1,0]=(-0.319040285312)*x_ref[0]+(-0.750406131559)*x_ref[1]+(-0.920332967032)*x_ref[2]
        ref[2,0,1,1]=(1.48758278445)*x_ref[0]+(1.30587557873)*x_ref[1]+(0.230692557888)*x_ref[2]
        ref[2,0,2,0]=(1.44269354043)*x_ref[0]+(-0.63270356894)*x_ref[1]+(0.229323632455)*x_ref[2]
        ref[2,0,2,1]=(-0.645065877536)*x_ref[0]+(-0.393083499676)*x_ref[1]+(-0.305117645453)*x_ref[2]
        ref[2,1,0,0]=(-0.63304134143)*x_ref[0]+(0.0664768401917)*x_ref[1]+(0.87587743066)*x_ref[2]
        ref[2,1,0,1]=(-0.531856958285)*x_ref[0]+(0.747953146648)*x_ref[1]+(1.23560779546)*x_ref[2]
        ref[2,1,1,0]=(0.0564479101025)*x_ref[0]+(0.546307803332)*x_ref[1]+(0.655044890368)*x_ref[2]
        ref[2,1,1,1]=(-1.12793057228)*x_ref[0]+(-0.736676194013)*x_ref[1]+(0.0375913829411)*x_ref[2]
        ref[2,1,2,0]=(-0.531384188067)*x_ref[0]+(0.249416904081)*x_ref[1]+(0.711430712986)*x_ref[2]
        ref[2,1,2,1]=(0.884244664065)*x_ref[0]+(0.456066941968)*x_ref[1]+(0.181630861785)*x_ref[2]
        ref[2,2,0,0]=(0.0844543345062)*x_ref[0]+(-0.435288041256)*x_ref[1]+(-0.434816763062)*x_ref[2]
        ref[2,2,0,1]=(0.373388105049)*x_ref[0]+(0.239833846429)*x_ref[1]+(-0.437771573282)*x_ref[2]
        ref[2,2,1,0]=(0.534401521907)*x_ref[0]+(0.393077203448)*x_ref[1]+(-0.42440387717)*x_ref[2]
        ref[2,2,1,1]=(0.318683369617)*x_ref[0]+(-0.538864643793)*x_ref[1]+(1.83466945806)*x_ref[2]
        ref[2,2,2,0]=(-1.79048122074)*x_ref[0]+(-1.05441213478)*x_ref[1]+(-0.573963806281)*x_ref[2]
        ref[2,2,2,1]=(0.944017405182)*x_ref[0]+(0.772756791538)*x_ref[1]+(0.0269673237336)*x_ref[2]
        ref[2,3,0,0]=(0.989703689142)*x_ref[0]+(1.21302325945)*x_ref[1]+(-0.92209419073)*x_ref[2]
        ref[2,3,0,1]=(-0.0957244757576)*x_ref[0]+(1.50132638315)*x_ref[1]+(-1.38716099756)*x_ref[2]
        ref[2,3,1,0]=(1.6445441776)*x_ref[0]+(-0.490207070694)*x_ref[1]+(0.626469347781)*x_ref[2]
        ref[2,3,1,1]=(0.0185568939147)*x_ref[0]+(0.412024752941)*x_ref[1]+(-0.22626102813)*x_ref[2]
        ref[2,3,2,0]=(0.0143609822966)*x_ref[0]+(0.115739025893)*x_ref[1]+(-0.544468631925)*x_ref[2]
        ref[2,3,2,1]=(-1.80522341493)*x_ref[0]+(0.33352152695)*x_ref[1]+(-0.841286464932)*x_ref[2]
        ref[2,4,0,0]=(0.0618444122042)*x_ref[0]+(1.91232484622)*x_ref[1]+(-0.406317200023)*x_ref[2]
        ref[2,4,0,1]=(0.279131062922)*x_ref[0]+(0.0194232637634)*x_ref[1]+(-0.00179069113634)*x_ref[2]
        ref[2,4,1,0]=(-0.701344579135)*x_ref[0]+(-0.27410935473)*x_ref[1]+(0.763362100733)*x_ref[2]
        ref[2,4,1,1]=(0.831198601185)*x_ref[0]+(0.90336781771)*x_ref[1]+(1.21856633382)*x_ref[2]
        ref[2,4,2,0]=(0.526278615033)*x_ref[0]+(-1.08673855801)*x_ref[1]+(0.0784355900913)*x_ref[2]
        ref[2,4,2,1]=(0.853226856851)*x_ref[0]+(-1.15919062627)*x_ref[1]+(0.185686031732)*x_ref[2]
        ref[3,0,0,0]=(0.272127498117)*x_ref[0]+(-0.457140162438)*x_ref[1]+(0.318358670541)*x_ref[2]
        ref[3,0,0,1]=(1.71056378307)*x_ref[0]+(0.200052730234)*x_ref[1]+(-0.127733325792)*x_ref[2]
        ref[3,0,1,0]=(-0.285340963261)*x_ref[0]+(-0.524544624385)*x_ref[1]+(0.916366474944)*x_ref[2]
        ref[3,0,1,1]=(-0.625506557557)*x_ref[0]+(1.0858476479)*x_ref[1]+(-1.24478983268)*x_ref[2]
        ref[3,0,2,0]=(-1.37299619916)*x_ref[0]+(-0.482246528202)*x_ref[1]+(-1.36857669075)*x_ref[2]
        ref[3,0,2,1]=(-0.924421844916)*x_ref[0]+(-1.50402098115)*x_ref[1]+(1.00045130673)*x_ref[2]
        ref[3,1,0,0]=(0.0417304700827)*x_ref[0]+(0.554527451004)*x_ref[1]+(0.299758163628)*x_ref[2]
        ref[3,1,0,1]=(0.639611639279)*x_ref[0]+(-0.444675638162)*x_ref[1]+(1.43150046757)*x_ref[2]
        ref[3,1,1,0]=(0.288697220898)*x_ref[0]+(1.32660433773)*x_ref[1]+(0.0889412116759)*x_ref[2]
        ref[3,1,1,1]=(-0.110516770201)*x_ref[0]+(-0.0214248733565)*x_ref[1]+(-0.758638794785)*x_ref[2]
        ref[3,1,2,0]=(1.34357138654)*x_ref[0]+(-0.13759272972)*x_ref[1]+(-0.00183845032784)*x_ref[2]
        ref[3,1,2,1]=(1.44174769627)*x_ref[0]+(0.49190176388)*x_ref[1]+(0.477331360141)*x_ref[2]
        ref[3,2,0,0]=(-0.10284777121)*x_ref[0]+(-0.63294157419)*x_ref[1]+(0.246064460451)*x_ref[2]
        ref[3,2,0,1]=(0.974270980699)*x_ref[0]+(1.03835886161)*x_ref[1]+(0.420000854838)*x_ref[2]
        ref[3,2,1,0]=(-0.00617831615923)*x_ref[0]+(-0.354063634073)*x_ref[1]+(0.764285884619)*x_ref[2]
        ref[3,2,1,1]=(0.210815790746)*x_ref[0]+(0.76609515744)*x_ref[1]+(-0.564401558401)*x_ref[2]
        ref[3,2,2,0]=(1.30471234011)*x_ref[0]+(1.54864119332)*x_ref[1]+(-0.493583650519)*x_ref[2]
        ref[3,2,2,1]=(0.42287513668)*x_ref[0]+(-0.603093868171)*x_ref[1]+(-0.811892974203)*x_ref[2]
        ref[3,3,0,0]=(-1.38358134693)*x_ref[0]+(-1.53706018098)*x_ref[1]+(-1.29701227501)*x_ref[2]
        ref[3,3,0,1]=(0.598581779902)*x_ref[0]+(1.02424010922)*x_ref[1]+(-1.1814313009)*x_ref[2]
        ref[3,3,1,0]=(1.30938529116)*x_ref[0]+(-0.504468477244)*x_ref[1]+(1.63655848854)*x_ref[2]
        ref[3,3,1,1]=(-1.47006424113)*x_ref[0]+(0.225951520203)*x_ref[1]+(-0.412273003169)*x_ref[2]
        ref[3,3,2,0]=(-0.116902234482)*x_ref[0]+(0.488144536044)*x_ref[1]+(1.01732461694)*x_ref[2]
        ref[3,3,2,1]=(1.0935814877)*x_ref[0]+(1.06389385916)*x_ref[1]+(-0.17981400591)*x_ref[2]
        ref[3,4,0,0]=(-0.833265062101)*x_ref[0]+(-0.754461148648)*x_ref[1]+(1.26702012815)*x_ref[2]
        ref[3,4,0,1]=(-0.793744278171)*x_ref[0]+(0.112704521653)*x_ref[1]+(-0.448684922187)*x_ref[2]
        ref[3,4,1,0]=(-0.396153523625)*x_ref[0]+(-0.0370870723568)*x_ref[1]+(0.554347618511)*x_ref[2]
        ref[3,4,1,1]=(-0.323276139711)*x_ref[0]+(-0.397389928392)*x_ref[1]+(-1.1707028324)*x_ref[2]
        ref[3,4,2,0]=(0.823350512513)*x_ref[0]+(1.47105759864)*x_ref[1]+(-0.977456877855)*x_ref[2]
        ref[3,4,2,1]=(-0.719791170741)*x_ref[0]+(-1.00363340645)*x_ref[1]+(0.814612770572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedContinuousFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        arg=(0.0473893058941)*x[0]+(0.951986966173)*x[1]
        ref=(0.0473893058941)*x_ref[0]+(0.951986966173)*x_ref[1]
      else:
        arg=(1.17583059)*x[0]+(-0.0937717081252)*x[1]+(-1.27613571861)*x[2]
        ref=(1.17583059)*x_ref[0]+(-0.0937717081252)*x_ref[1]+(-1.27613571861)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        arg[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        arg[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
        ref[0]=(-1.12026442538)*x[0]+(-0.118489919905)*x[1]
        ref[1]=(-1.31300753934)*x[0]+(0.938487315467)*x[1]
      else:
        arg[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        arg[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
        ref[0]=(1.03792743444)*x[0]+(0.131747885089)*x[1]+(-0.122522375594)*x[2]
        ref[1]=(-0.948813532769)*x[0]+(0.355251567439)*x[1]+(0.256406780636)*x[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        arg[0,0]=(-0.352147407922)*x[0]+(1.05748871495)*x[1]
        arg[0,1]=(-0.474780656205)*x[0]+(0.459896921083)*x[1]
        arg[0,2]=(0.460595468629)*x[0]+(-0.719265552984)*x[1]
        arg[0,3]=(-0.182481554307)*x[0]+(0.658370858696)*x[1]
        arg[0,4]=(0.827070331069)*x[0]+(0.461652365051)*x[1]
        arg[1,0]=(-1.04815178098)*x[0]+(-0.401056329621)*x[1]
        arg[1,1]=(1.44973151029)*x[0]+(-0.240684558811)*x[1]
        arg[1,2]=(1.66374138794)*x[0]+(0.77295074874)*x[1]
        arg[1,3]=(0.367619281099)*x[0]+(1.68735398814)*x[1]
        arg[1,4]=(-0.609166774517)*x[0]+(-0.236664892047)*x[1]
        arg[2,0]=(0.517168891914)*x[0]+(0.14465722282)*x[1]
        arg[2,1]=(1.44226468541)*x[0]+(0.138272882962)*x[1]
        arg[2,2]=(1.48139408702)*x[0]+(0.858274661903)*x[1]
        arg[2,3]=(-0.313581651419)*x[0]+(0.23082842733)*x[1]
        arg[2,4]=(-0.174268737758)*x[0]+(-0.21537909808)*x[1]
        arg[3,0]=(-0.228982633638)*x[0]+(0.469215862849)*x[1]
        arg[3,1]=(1.13696568169)*x[0]+(0.937455803513)*x[1]
        arg[3,2]=(-0.104407288228)*x[0]+(0.695766113469)*x[1]
        arg[3,3]=(-0.854996521846)*x[0]+(-0.112176127951)*x[1]
        arg[3,4]=(-0.995531601234)*x[0]+(0.0800402854558)*x[1]
        ref[0,0]=(-0.352147407922)*x_ref[0]+(1.05748871495)*x_ref[1]
        ref[0,1]=(-0.474780656205)*x_ref[0]+(0.459896921083)*x_ref[1]
        ref[0,2]=(0.460595468629)*x_ref[0]+(-0.719265552984)*x_ref[1]
        ref[0,3]=(-0.182481554307)*x_ref[0]+(0.658370858696)*x_ref[1]
        ref[0,4]=(0.827070331069)*x_ref[0]+(0.461652365051)*x_ref[1]
        ref[1,0]=(-1.04815178098)*x_ref[0]+(-0.401056329621)*x_ref[1]
        ref[1,1]=(1.44973151029)*x_ref[0]+(-0.240684558811)*x_ref[1]
        ref[1,2]=(1.66374138794)*x_ref[0]+(0.77295074874)*x_ref[1]
        ref[1,3]=(0.367619281099)*x_ref[0]+(1.68735398814)*x_ref[1]
        ref[1,4]=(-0.609166774517)*x_ref[0]+(-0.236664892047)*x_ref[1]
        ref[2,0]=(0.517168891914)*x_ref[0]+(0.14465722282)*x_ref[1]
        ref[2,1]=(1.44226468541)*x_ref[0]+(0.138272882962)*x_ref[1]
        ref[2,2]=(1.48139408702)*x_ref[0]+(0.858274661903)*x_ref[1]
        ref[2,3]=(-0.313581651419)*x_ref[0]+(0.23082842733)*x_ref[1]
        ref[2,4]=(-0.174268737758)*x_ref[0]+(-0.21537909808)*x_ref[1]
        ref[3,0]=(-0.228982633638)*x_ref[0]+(0.469215862849)*x_ref[1]
        ref[3,1]=(1.13696568169)*x_ref[0]+(0.937455803513)*x_ref[1]
        ref[3,2]=(-0.104407288228)*x_ref[0]+(0.695766113469)*x_ref[1]
        ref[3,3]=(-0.854996521846)*x_ref[0]+(-0.112176127951)*x_ref[1]
        ref[3,4]=(-0.995531601234)*x_ref[0]+(0.0800402854558)*x_ref[1]
      else:
        arg[0,0]=(-1.37643176268)*x[0]+(-0.1427781822)*x[1]+(-0.959699369061)*x[2]
        arg[0,1]=(-0.182711571685)*x[0]+(0.961312900413)*x[1]+(1.51040175428)*x[2]
        arg[0,2]=(0.705353375057)*x[0]+(-0.486345048698)*x[1]+(0.276322756959)*x[2]
        arg[0,3]=(0.617751891549)*x[0]+(0.842087813554)*x[1]+(0.181498577908)*x[2]
        arg[0,4]=(-0.806056649748)*x[0]+(-1.46464582218)*x[1]+(1.16207385717)*x[2]
        arg[1,0]=(0.443056731112)*x[0]+(-0.227362546034)*x[1]+(-0.34165958927)*x[2]
        arg[1,1]=(-0.0914572456267)*x[0]+(-1.45454323755)*x[1]+(-0.197767792435)*x[2]
        arg[1,2]=(-0.0294645792971)*x[0]+(0.154875742902)*x[1]+(0.227583590411)*x[2]
        arg[1,3]=(0.0412025551742)*x[0]+(0.263462610593)*x[1]+(-0.916261691981)*x[2]
        arg[1,4]=(0.436907441126)*x[0]+(1.69979623684)*x[1]+(1.28829395572)*x[2]
        arg[2,0]=(0.328651176462)*x[0]+(1.15129657872)*x[1]+(0.288790490546)*x[2]
        arg[2,1]=(0.603244456904)*x[0]+(0.579263327926)*x[1]+(1.29848862962)*x[2]
        arg[2,2]=(1.20931621113)*x[0]+(1.45214239482)*x[1]+(-0.859270540382)*x[2]
        arg[2,3]=(-0.313390836289)*x[0]+(1.29725175689)*x[1]+(-0.677628108203)*x[2]
        arg[2,4]=(-0.182499400988)*x[0]+(-0.150297906549)*x[1]+(-0.0239318512478)*x[2]
        arg[3,0]=(1.24015393707)*x[0]+(0.0747824863962)*x[1]+(0.0762786991199)*x[2]
        arg[3,1]=(0.0278504259683)*x[0]+(-1.03121188802)*x[1]+(-1.1241189107)*x[2]
        arg[3,2]=(-0.014656978991)*x[0]+(-0.0946814312886)*x[1]+(0.281133656449)*x[2]
        arg[3,3]=(0.0763735895586)*x[0]+(-0.878856395577)*x[1]+(-0.0317884442673)*x[2]
        arg[3,4]=(-0.0394676966134)*x[0]+(0.491771511608)*x[1]+(0.270055976645)*x[2]
        ref[0,0]=(-1.37643176268)*x_ref[0]+(-0.1427781822)*x_ref[1]+(-0.959699369061)*x_ref[2]
        ref[0,1]=(-0.182711571685)*x_ref[0]+(0.961312900413)*x_ref[1]+(1.51040175428)*x_ref[2]
        ref[0,2]=(0.705353375057)*x_ref[0]+(-0.486345048698)*x_ref[1]+(0.276322756959)*x_ref[2]
        ref[0,3]=(0.617751891549)*x_ref[0]+(0.842087813554)*x_ref[1]+(0.181498577908)*x_ref[2]
        ref[0,4]=(-0.806056649748)*x_ref[0]+(-1.46464582218)*x_ref[1]+(1.16207385717)*x_ref[2]
        ref[1,0]=(0.443056731112)*x_ref[0]+(-0.227362546034)*x_ref[1]+(-0.34165958927)*x_ref[2]
        ref[1,1]=(-0.0914572456267)*x_ref[0]+(-1.45454323755)*x_ref[1]+(-0.197767792435)*x_ref[2]
        ref[1,2]=(-0.0294645792971)*x_ref[0]+(0.154875742902)*x_ref[1]+(0.227583590411)*x_ref[2]
        ref[1,3]=(0.0412025551742)*x_ref[0]+(0.263462610593)*x_ref[1]+(-0.916261691981)*x_ref[2]
        ref[1,4]=(0.436907441126)*x_ref[0]+(1.69979623684)*x_ref[1]+(1.28829395572)*x_ref[2]
        ref[2,0]=(0.328651176462)*x_ref[0]+(1.15129657872)*x_ref[1]+(0.288790490546)*x_ref[2]
        ref[2,1]=(0.603244456904)*x_ref[0]+(0.579263327926)*x_ref[1]+(1.29848862962)*x_ref[2]
        ref[2,2]=(1.20931621113)*x_ref[0]+(1.45214239482)*x_ref[1]+(-0.859270540382)*x_ref[2]
        ref[2,3]=(-0.313390836289)*x_ref[0]+(1.29725175689)*x_ref[1]+(-0.677628108203)*x_ref[2]
        ref[2,4]=(-0.182499400988)*x_ref[0]+(-0.150297906549)*x_ref[1]+(-0.0239318512478)*x_ref[2]
        ref[3,0]=(1.24015393707)*x_ref[0]+(0.0747824863962)*x_ref[1]+(0.0762786991199)*x_ref[2]
        ref[3,1]=(0.0278504259683)*x_ref[0]+(-1.03121188802)*x_ref[1]+(-1.1241189107)*x_ref[2]
        ref[3,2]=(-0.014656978991)*x_ref[0]+(-0.0946814312886)*x_ref[1]+(0.281133656449)*x_ref[2]
        ref[3,3]=(0.0763735895586)*x_ref[0]+(-0.878856395577)*x_ref[1]+(-0.0317884442673)*x_ref[2]
        ref[3,4]=(-0.0394676966134)*x_ref[0]+(0.491771511608)*x_ref[1]+(0.270055976645)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        arg[0,0,0]=(-0.648027713097)*x[0]+(0.454772464521)*x[1]
        arg[0,0,1]=(-1.04215194027)*x[0]+(1.77272186309)*x[1]
        arg[0,1,0]=(0.957518940744)*x[0]+(-0.00690667683404)*x[1]
        arg[0,1,1]=(0.498748880149)*x[0]+(0.259327852057)*x[1]
        arg[1,0,0]=(0.0289932206553)*x[0]+(-0.0536452199935)*x[1]
        arg[1,0,1]=(1.87294208563)*x[0]+(-1.51381628261)*x[1]
        arg[1,1,0]=(-1.23040396582)*x[0]+(0.554662981657)*x[1]
        arg[1,1,1]=(-0.860941498913)*x[0]+(0.574367240324)*x[1]
        arg[2,0,0]=(0.137130620331)*x[0]+(-0.670662009715)*x[1]
        arg[2,0,1]=(-0.01425641324)*x[0]+(0.938457397733)*x[1]
        arg[2,1,0]=(-1.19963811258)*x[0]+(0.644324590438)*x[1]
        arg[2,1,1]=(1.95704700194e-05)*x[0]+(-1.38366302093)*x[1]
        arg[3,0,0]=(-0.340745623814)*x[0]+(0.250094821868)*x[1]
        arg[3,0,1]=(-1.29605858834)*x[0]+(0.960366146641)*x[1]
        arg[3,1,0]=(0.550081967023)*x[0]+(0.288502329322)*x[1]
        arg[3,1,1]=(0.732775194015)*x[0]+(-0.0386387970031)*x[1]
        arg[4,0,0]=(-0.849770148692)*x[0]+(-0.436107981485)*x[1]
        arg[4,0,1]=(0.11202465528)*x[0]+(1.66934758318)*x[1]
        arg[4,1,0]=(-0.0549275626402)*x[0]+(0.75026785816)*x[1]
        arg[4,1,1]=(-0.351732868571)*x[0]+(0.892311743034)*x[1]
        arg[5,0,0]=(-0.744188679625)*x[0]+(0.335492215064)*x[1]
        arg[5,0,1]=(-0.998395994344)*x[0]+(-1.10552505805)*x[1]
        arg[5,1,0]=(-0.125733052457)*x[0]+(0.433627801831)*x[1]
        arg[5,1,1]=(-0.0704342146568)*x[0]+(-1.38357990003)*x[1]
        ref[0,0,0]=(-0.648027713097)*x_ref[0]+(0.454772464521)*x_ref[1]
        ref[0,0,1]=(-1.04215194027)*x_ref[0]+(1.77272186309)*x_ref[1]
        ref[0,1,0]=(0.957518940744)*x_ref[0]+(-0.00690667683404)*x_ref[1]
        ref[0,1,1]=(0.498748880149)*x_ref[0]+(0.259327852057)*x_ref[1]
        ref[1,0,0]=(0.0289932206553)*x_ref[0]+(-0.0536452199935)*x_ref[1]
        ref[1,0,1]=(1.87294208563)*x_ref[0]+(-1.51381628261)*x_ref[1]
        ref[1,1,0]=(-1.23040396582)*x_ref[0]+(0.554662981657)*x_ref[1]
        ref[1,1,1]=(-0.860941498913)*x_ref[0]+(0.574367240324)*x_ref[1]
        ref[2,0,0]=(0.137130620331)*x_ref[0]+(-0.670662009715)*x_ref[1]
        ref[2,0,1]=(-0.01425641324)*x_ref[0]+(0.938457397733)*x_ref[1]
        ref[2,1,0]=(-1.19963811258)*x_ref[0]+(0.644324590438)*x_ref[1]
        ref[2,1,1]=(1.95704700194e-05)*x_ref[0]+(-1.38366302093)*x_ref[1]
        ref[3,0,0]=(-0.340745623814)*x_ref[0]+(0.250094821868)*x_ref[1]
        ref[3,0,1]=(-1.29605858834)*x_ref[0]+(0.960366146641)*x_ref[1]
        ref[3,1,0]=(0.550081967023)*x_ref[0]+(0.288502329322)*x_ref[1]
        ref[3,1,1]=(0.732775194015)*x_ref[0]+(-0.0386387970031)*x_ref[1]
        ref[4,0,0]=(-0.849770148692)*x_ref[0]+(-0.436107981485)*x_ref[1]
        ref[4,0,1]=(0.11202465528)*x_ref[0]+(1.66934758318)*x_ref[1]
        ref[4,1,0]=(-0.0549275626402)*x_ref[0]+(0.75026785816)*x_ref[1]
        ref[4,1,1]=(-0.351732868571)*x_ref[0]+(0.892311743034)*x_ref[1]
        ref[5,0,0]=(-0.744188679625)*x_ref[0]+(0.335492215064)*x_ref[1]
        ref[5,0,1]=(-0.998395994344)*x_ref[0]+(-1.10552505805)*x_ref[1]
        ref[5,1,0]=(-0.125733052457)*x_ref[0]+(0.433627801831)*x_ref[1]
        ref[5,1,1]=(-0.0704342146568)*x_ref[0]+(-1.38357990003)*x_ref[1]
      else:
        arg[0,0,0]=(0.514282046117)*x[0]+(-0.323828864786)*x[1]+(1.01124693935)*x[2]
        arg[0,0,1]=(-0.621932286265)*x[0]+(-0.797795969893)*x[1]+(-0.0773166049033)*x[2]
        arg[0,1,0]=(-0.874326235403)*x[0]+(1.20497718671)*x[1]+(-0.327967753068)*x[2]
        arg[0,1,1]=(0.951360009956)*x[0]+(-1.21275900525)*x[1]+(-0.018534105952)*x[2]
        arg[1,0,0]=(0.281987835612)*x[0]+(-1.19722794418)*x[1]+(1.1763726345)*x[2]
        arg[1,0,1]=(-0.32305643761)*x[0]+(-1.03791873988)*x[1]+(-0.27820297843)*x[2]
        arg[1,1,0]=(-0.250435095454)*x[0]+(-0.214489103641)*x[1]+(0.24396821423)*x[2]
        arg[1,1,1]=(0.579357080067)*x[0]+(-0.0941366284697)*x[1]+(0.723253564904)*x[2]
        arg[2,0,0]=(0.0300521887548)*x[0]+(0.856925059877)*x[1]+(1.62627876391)*x[2]
        arg[2,0,1]=(0.852347895674)*x[0]+(0.206107771527)*x[1]+(0.117053849317)*x[2]
        arg[2,1,0]=(0.398984864506)*x[0]+(-0.547599955809)*x[1]+(0.960920641374)*x[2]
        arg[2,1,1]=(-0.958823099193)*x[0]+(0.781951729915)*x[1]+(0.450383856811)*x[2]
        arg[3,0,0]=(0.229905485668)*x[0]+(0.13732685767)*x[1]+(0.512823357151)*x[2]
        arg[3,0,1]=(-1.85513528612)*x[0]+(1.26641983818)*x[1]+(1.23786450856)*x[2]
        arg[3,1,0]=(-1.42722463673)*x[0]+(1.39168597069)*x[1]+(0.268597738146)*x[2]
        arg[3,1,1]=(0.290999479499)*x[0]+(0.948263729849)*x[1]+(-1.60034455324)*x[2]
        arg[4,0,0]=(-0.755582573265)*x[0]+(-0.0499101766292)*x[1]+(0.0615553965851)*x[2]
        arg[4,0,1]=(0.437479529174)*x[0]+(-1.05498330706)*x[1]+(-0.707627367761)*x[2]
        arg[4,1,0]=(0.608125820223)*x[0]+(0.893214773163)*x[1]+(-1.28896829926)*x[2]
        arg[4,1,1]=(0.953486972425)*x[0]+(1.41823237022)*x[1]+(-0.0439589309603)*x[2]
        arg[5,0,0]=(1.06772863677)*x[0]+(0.25868918695)*x[1]+(-0.118887800607)*x[2]
        arg[5,0,1]=(0.407678657671)*x[0]+(-0.464724724076)*x[1]+(0.249621788892)*x[2]
        arg[5,1,0]=(-0.516649619366)*x[0]+(-1.7414783126)*x[1]+(-1.1378435051)*x[2]
        arg[5,1,1]=(-0.805463452054)*x[0]+(-0.189909983229)*x[1]+(-0.127169486885)*x[2]
        ref[0,0,0]=(0.514282046117)*x_ref[0]+(-0.323828864786)*x_ref[1]+(1.01124693935)*x_ref[2]
        ref[0,0,1]=(-0.621932286265)*x_ref[0]+(-0.797795969893)*x_ref[1]+(-0.0773166049033)*x_ref[2]
        ref[0,1,0]=(-0.874326235403)*x_ref[0]+(1.20497718671)*x_ref[1]+(-0.327967753068)*x_ref[2]
        ref[0,1,1]=(0.951360009956)*x_ref[0]+(-1.21275900525)*x_ref[1]+(-0.018534105952)*x_ref[2]
        ref[1,0,0]=(0.281987835612)*x_ref[0]+(-1.19722794418)*x_ref[1]+(1.1763726345)*x_ref[2]
        ref[1,0,1]=(-0.32305643761)*x_ref[0]+(-1.03791873988)*x_ref[1]+(-0.27820297843)*x_ref[2]
        ref[1,1,0]=(-0.250435095454)*x_ref[0]+(-0.214489103641)*x_ref[1]+(0.24396821423)*x_ref[2]
        ref[1,1,1]=(0.579357080067)*x_ref[0]+(-0.0941366284697)*x_ref[1]+(0.723253564904)*x_ref[2]
        ref[2,0,0]=(0.0300521887548)*x_ref[0]+(0.856925059877)*x_ref[1]+(1.62627876391)*x_ref[2]
        ref[2,0,1]=(0.852347895674)*x_ref[0]+(0.206107771527)*x_ref[1]+(0.117053849317)*x_ref[2]
        ref[2,1,0]=(0.398984864506)*x_ref[0]+(-0.547599955809)*x_ref[1]+(0.960920641374)*x_ref[2]
        ref[2,1,1]=(-0.958823099193)*x_ref[0]+(0.781951729915)*x_ref[1]+(0.450383856811)*x_ref[2]
        ref[3,0,0]=(0.229905485668)*x_ref[0]+(0.13732685767)*x_ref[1]+(0.512823357151)*x_ref[2]
        ref[3,0,1]=(-1.85513528612)*x_ref[0]+(1.26641983818)*x_ref[1]+(1.23786450856)*x_ref[2]
        ref[3,1,0]=(-1.42722463673)*x_ref[0]+(1.39168597069)*x_ref[1]+(0.268597738146)*x_ref[2]
        ref[3,1,1]=(0.290999479499)*x_ref[0]+(0.948263729849)*x_ref[1]+(-1.60034455324)*x_ref[2]
        ref[4,0,0]=(-0.755582573265)*x_ref[0]+(-0.0499101766292)*x_ref[1]+(0.0615553965851)*x_ref[2]
        ref[4,0,1]=(0.437479529174)*x_ref[0]+(-1.05498330706)*x_ref[1]+(-0.707627367761)*x_ref[2]
        ref[4,1,0]=(0.608125820223)*x_ref[0]+(0.893214773163)*x_ref[1]+(-1.28896829926)*x_ref[2]
        ref[4,1,1]=(0.953486972425)*x_ref[0]+(1.41823237022)*x_ref[1]+(-0.0439589309603)*x_ref[2]
        ref[5,0,0]=(1.06772863677)*x_ref[0]+(0.25868918695)*x_ref[1]+(-0.118887800607)*x_ref[2]
        ref[5,0,1]=(0.407678657671)*x_ref[0]+(-0.464724724076)*x_ref[1]+(0.249621788892)*x_ref[2]
        ref[5,1,0]=(-0.516649619366)*x_ref[0]+(-1.7414783126)*x_ref[1]+(-1.1378435051)*x_ref[2]
        ref[5,1,1]=(-0.805463452054)*x_ref[0]+(-0.189909983229)*x_ref[1]+(-0.127169486885)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromData_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Data onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        arg[0,0,0,0]=(-0.0345299040391)*x[0]+(-0.454395763203)*x[1]
        arg[0,0,0,1]=(0.816055157967)*x[0]+(1.65538436278)*x[1]
        arg[0,0,1,0]=(0.0173895492935)*x[0]+(0.390351555311)*x[1]
        arg[0,0,1,1]=(0.0601948568001)*x[0]+(0.221935674716)*x[1]
        arg[0,0,2,0]=(-0.86832157368)*x[0]+(0.719344815038)*x[1]
        arg[0,0,2,1]=(1.25819429092)*x[0]+(-0.626192861251)*x[1]
        arg[0,1,0,0]=(-0.408148703502)*x[0]+(0.614498083344)*x[1]
        arg[0,1,0,1]=(-0.948405669606)*x[0]+(-0.431912186759)*x[1]
        arg[0,1,1,0]=(1.31545349201)*x[0]+(-1.43633966996)*x[1]
        arg[0,1,1,1]=(-0.262029058642)*x[0]+(-0.211155762782)*x[1]
        arg[0,1,2,0]=(-1.29637408623)*x[0]+(0.015300306679)*x[1]
        arg[0,1,2,1]=(-0.338105266735)*x[0]+(-1.17464147824)*x[1]
        arg[0,2,0,0]=(-1.29420036616)*x[0]+(0.162651211594)*x[1]
        arg[0,2,0,1]=(-0.524474022859)*x[0]+(-0.606176834607)*x[1]
        arg[0,2,1,0]=(1.43089226048)*x[0]+(-1.4419171599)*x[1]
        arg[0,2,1,1]=(1.2204089172)*x[0]+(0.642504226769)*x[1]
        arg[0,2,2,0]=(-1.43093048803)*x[0]+(-0.220023978143)*x[1]
        arg[0,2,2,1]=(-1.03649750995)*x[0]+(0.169884643951)*x[1]
        arg[0,3,0,0]=(-0.250124925508)*x[0]+(-0.46412090099)*x[1]
        arg[0,3,0,1]=(-0.534018976189)*x[0]+(-0.118541672182)*x[1]
        arg[0,3,1,0]=(-1.79137479214)*x[0]+(0.662559958387)*x[1]
        arg[0,3,1,1]=(-0.881482101487)*x[0]+(0.372532785249)*x[1]
        arg[0,3,2,0]=(-1.71336096366)*x[0]+(-1.24175201959)*x[1]
        arg[0,3,2,1]=(0.384581239528)*x[0]+(-0.131008110136)*x[1]
        arg[0,4,0,0]=(0.315863164115)*x[0]+(-0.494712527335)*x[1]
        arg[0,4,0,1]=(-1.10002713328)*x[0]+(0.129861779935)*x[1]
        arg[0,4,1,0]=(1.10412289984)*x[0]+(1.34015931457)*x[1]
        arg[0,4,1,1]=(-0.692340206781)*x[0]+(1.08467851048)*x[1]
        arg[0,4,2,0]=(-0.0972657457169)*x[0]+(1.0145314899)*x[1]
        arg[0,4,2,1]=(0.299871193443)*x[0]+(-0.470271033392)*x[1]
        arg[1,0,0,0]=(-0.724229380092)*x[0]+(1.14448991134)*x[1]
        arg[1,0,0,1]=(0.154147313948)*x[0]+(-0.318350887819)*x[1]
        arg[1,0,1,0]=(-1.2997484551)*x[0]+(0.00645216870018)*x[1]
        arg[1,0,1,1]=(1.53322180988)*x[0]+(-0.356405349895)*x[1]
        arg[1,0,2,0]=(0.971184272781)*x[0]+(1.07821077146)*x[1]
        arg[1,0,2,1]=(0.000530732762419)*x[0]+(0.955934156926)*x[1]
        arg[1,1,0,0]=(-0.180982534825)*x[0]+(1.33554642886)*x[1]
        arg[1,1,0,1]=(-0.419614369168)*x[0]+(-0.51201636845)*x[1]
        arg[1,1,1,0]=(-0.263686990653)*x[0]+(-0.463089568134)*x[1]
        arg[1,1,1,1]=(-0.792982985293)*x[0]+(-0.720992794822)*x[1]
        arg[1,1,2,0]=(1.72981591311)*x[0]+(-0.305755746837)*x[1]
        arg[1,1,2,1]=(0.874057242127)*x[0]+(0.247839316695)*x[1]
        arg[1,2,0,0]=(1.20726244725)*x[0]+(0.44394602451)*x[1]
        arg[1,2,0,1]=(-0.516016832395)*x[0]+(0.036120494771)*x[1]
        arg[1,2,1,0]=(-0.157245073014)*x[0]+(0.845240869121)*x[1]
        arg[1,2,1,1]=(0.588082465766)*x[0]+(0.195937663431)*x[1]
        arg[1,2,2,0]=(0.095726763028)*x[0]+(-0.177757031723)*x[1]
        arg[1,2,2,1]=(1.73215168642)*x[0]+(-0.189154112108)*x[1]
        arg[1,3,0,0]=(-1.76621021161)*x[0]+(0.367931299124)*x[1]
        arg[1,3,0,1]=(0.0222607642178)*x[0]+(0.466368587137)*x[1]
        arg[1,3,1,0]=(0.412132176201)*x[0]+(0.553362139129)*x[1]
        arg[1,3,1,1]=(-1.08306042349)*x[0]+(-0.834943125248)*x[1]
        arg[1,3,2,0]=(-0.417451930632)*x[0]+(-0.659233486131)*x[1]
        arg[1,3,2,1]=(-0.400333106519)*x[0]+(0.854889031906)*x[1]
        arg[1,4,0,0]=(-0.758414654342)*x[0]+(0.838957251391)*x[1]
        arg[1,4,0,1]=(-0.353472866001)*x[0]+(0.397138506541)*x[1]
        arg[1,4,1,0]=(0.127641591916)*x[0]+(0.327603318651)*x[1]
        arg[1,4,1,1]=(-0.341040162377)*x[0]+(-0.846062323778)*x[1]
        arg[1,4,2,0]=(-0.293356226448)*x[0]+(-0.700648323903)*x[1]
        arg[1,4,2,1]=(-0.228549394688)*x[0]+(1.43724938876)*x[1]
        arg[2,0,0,0]=(-0.322023776234)*x[0]+(0.564889094588)*x[1]
        arg[2,0,0,1]=(-1.11907082102)*x[0]+(-1.39374772922)*x[1]
        arg[2,0,1,0]=(-0.705740617022)*x[0]+(0.511499218842)*x[1]
        arg[2,0,1,1]=(0.454111255253)*x[0]+(-1.69843226881)*x[1]
        arg[2,0,2,0]=(-0.317781478691)*x[0]+(-1.35991309273)*x[1]
        arg[2,0,2,1]=(1.54613408914)*x[0]+(1.32901459805)*x[1]
        arg[2,1,0,0]=(0.837517924208)*x[0]+(-0.386149637251)*x[1]
        arg[2,1,0,1]=(0.961896547496)*x[0]+(0.617366532183)*x[1]
        arg[2,1,1,0]=(-0.124021064331)*x[0]+(1.45238030663)*x[1]
        arg[2,1,1,1]=(-1.34194701555)*x[0]+(-0.479310148218)*x[1]
        arg[2,1,2,0]=(-0.813567438318)*x[0]+(0.35270597356)*x[1]
        arg[2,1,2,1]=(0.833034787475)*x[0]+(-0.108648813503)*x[1]
        arg[2,2,0,0]=(-1.36052757115)*x[0]+(-0.528858737068)*x[1]
        arg[2,2,0,1]=(0.348085718404)*x[0]+(-0.550199132198)*x[1]
        arg[2,2,1,0]=(-0.592064932053)*x[0]+(0.268729595158)*x[1]
        arg[2,2,1,1]=(0.762201961036)*x[0]+(-0.170275800619)*x[1]
        arg[2,2,2,0]=(0.292508257449)*x[0]+(0.546726460436)*x[1]
        arg[2,2,2,1]=(-0.832981468483)*x[0]+(-1.01580974142)*x[1]
        arg[2,3,0,0]=(-0.0778970282444)*x[0]+(-0.0588582131822)*x[1]
        arg[2,3,0,1]=(-0.462860486423)*x[0]+(0.938910405107)*x[1]
        arg[2,3,1,0]=(0.274056697467)*x[0]+(0.0938367562972)*x[1]
        arg[2,3,1,1]=(0.168689167517)*x[0]+(-0.486345829711)*x[1]
        arg[2,3,2,0]=(-1.33699397485)*x[0]+(-0.863623492326)*x[1]
        arg[2,3,2,1]=(-0.317889793298)*x[0]+(0.47526951533)*x[1]
        arg[2,4,0,0]=(-1.26561106959)*x[0]+(0.0486883091899)*x[1]
        arg[2,4,0,1]=(-1.37742147186)*x[0]+(-1.14365670741)*x[1]
        arg[2,4,1,0]=(0.891487506649)*x[0]+(-1.02149537779)*x[1]
        arg[2,4,1,1]=(0.120446738176)*x[0]+(1.128322636)*x[1]
        arg[2,4,2,0]=(-0.322931882495)*x[0]+(-0.00993625804164)*x[1]
        arg[2,4,2,1]=(0.17764198759)*x[0]+(0.311937789824)*x[1]
        arg[3,0,0,0]=(0.482490440467)*x[0]+(0.98496437249)*x[1]
        arg[3,0,0,1]=(-0.174114820802)*x[0]+(0.442866651846)*x[1]
        arg[3,0,1,0]=(-0.934180386688)*x[0]+(-1.33361314957)*x[1]
        arg[3,0,1,1]=(0.0452285929122)*x[0]+(0.686705971575)*x[1]
        arg[3,0,2,0]=(0.839801283293)*x[0]+(-0.967697483261)*x[1]
        arg[3,0,2,1]=(0.14250663194)*x[0]+(-1.3515960506)*x[1]
        arg[3,1,0,0]=(-0.322855905664)*x[0]+(0.177298330914)*x[1]
        arg[3,1,0,1]=(-0.376491734039)*x[0]+(0.451059427164)*x[1]
        arg[3,1,1,0]=(0.817537444737)*x[0]+(-0.93908061343)*x[1]
        arg[3,1,1,1]=(-0.0753530118606)*x[0]+(-0.0718596035883)*x[1]
        arg[3,1,2,0]=(-1.48772358944)*x[0]+(0.414823807711)*x[1]
        arg[3,1,2,1]=(0.857324284797)*x[0]+(0.47131796901)*x[1]
        arg[3,2,0,0]=(0.440034775536)*x[0]+(0.712881563582)*x[1]
        arg[3,2,0,1]=(-1.05547066)*x[0]+(0.718956424358)*x[1]
        arg[3,2,1,0]=(-1.40526673086)*x[0]+(-0.366871542135)*x[1]
        arg[3,2,1,1]=(1.65950470328)*x[0]+(-1.65805816594)*x[1]
        arg[3,2,2,0]=(1.09096918876)*x[0]+(-1.33325242756)*x[1]
        arg[3,2,2,1]=(-0.591939877469)*x[0]+(1.0303591776)*x[1]
        arg[3,3,0,0]=(-0.969754780364)*x[0]+(-0.0145852822005)*x[1]
        arg[3,3,0,1]=(-1.22384841753)*x[0]+(-0.0867406833581)*x[1]
        arg[3,3,1,0]=(0.673272022388)*x[0]+(0.947486168095)*x[1]
        arg[3,3,1,1]=(-0.41449883684)*x[0]+(0.137714746421)*x[1]
        arg[3,3,2,0]=(0.874598196414)*x[0]+(0.436570682065)*x[1]
        arg[3,3,2,1]=(0.127986513106)*x[0]+(0.955109334387)*x[1]
        arg[3,4,0,0]=(0.226160174687)*x[0]+(-0.258282204272)*x[1]
        arg[3,4,0,1]=(-0.157570106985)*x[0]+(-0.299582244336)*x[1]
        arg[3,4,1,0]=(0.305056978098)*x[0]+(1.34807929179)*x[1]
        arg[3,4,1,1]=(1.16863934732)*x[0]+(-0.265384337877)*x[1]
        arg[3,4,2,0]=(1.26265936998)*x[0]+(0.379703748135)*x[1]
        arg[3,4,2,1]=(-0.392236938104)*x[0]+(0.829704189106)*x[1]
        ref[0,0,0,0]=(-0.0345299040391)*x_ref[0]+(-0.454395763203)*x_ref[1]
        ref[0,0,0,1]=(0.816055157967)*x_ref[0]+(1.65538436278)*x_ref[1]
        ref[0,0,1,0]=(0.0173895492935)*x_ref[0]+(0.390351555311)*x_ref[1]
        ref[0,0,1,1]=(0.0601948568001)*x_ref[0]+(0.221935674716)*x_ref[1]
        ref[0,0,2,0]=(-0.86832157368)*x_ref[0]+(0.719344815038)*x_ref[1]
        ref[0,0,2,1]=(1.25819429092)*x_ref[0]+(-0.626192861251)*x_ref[1]
        ref[0,1,0,0]=(-0.408148703502)*x_ref[0]+(0.614498083344)*x_ref[1]
        ref[0,1,0,1]=(-0.948405669606)*x_ref[0]+(-0.431912186759)*x_ref[1]
        ref[0,1,1,0]=(1.31545349201)*x_ref[0]+(-1.43633966996)*x_ref[1]
        ref[0,1,1,1]=(-0.262029058642)*x_ref[0]+(-0.211155762782)*x_ref[1]
        ref[0,1,2,0]=(-1.29637408623)*x_ref[0]+(0.015300306679)*x_ref[1]
        ref[0,1,2,1]=(-0.338105266735)*x_ref[0]+(-1.17464147824)*x_ref[1]
        ref[0,2,0,0]=(-1.29420036616)*x_ref[0]+(0.162651211594)*x_ref[1]
        ref[0,2,0,1]=(-0.524474022859)*x_ref[0]+(-0.606176834607)*x_ref[1]
        ref[0,2,1,0]=(1.43089226048)*x_ref[0]+(-1.4419171599)*x_ref[1]
        ref[0,2,1,1]=(1.2204089172)*x_ref[0]+(0.642504226769)*x_ref[1]
        ref[0,2,2,0]=(-1.43093048803)*x_ref[0]+(-0.220023978143)*x_ref[1]
        ref[0,2,2,1]=(-1.03649750995)*x_ref[0]+(0.169884643951)*x_ref[1]
        ref[0,3,0,0]=(-0.250124925508)*x_ref[0]+(-0.46412090099)*x_ref[1]
        ref[0,3,0,1]=(-0.534018976189)*x_ref[0]+(-0.118541672182)*x_ref[1]
        ref[0,3,1,0]=(-1.79137479214)*x_ref[0]+(0.662559958387)*x_ref[1]
        ref[0,3,1,1]=(-0.881482101487)*x_ref[0]+(0.372532785249)*x_ref[1]
        ref[0,3,2,0]=(-1.71336096366)*x_ref[0]+(-1.24175201959)*x_ref[1]
        ref[0,3,2,1]=(0.384581239528)*x_ref[0]+(-0.131008110136)*x_ref[1]
        ref[0,4,0,0]=(0.315863164115)*x_ref[0]+(-0.494712527335)*x_ref[1]
        ref[0,4,0,1]=(-1.10002713328)*x_ref[0]+(0.129861779935)*x_ref[1]
        ref[0,4,1,0]=(1.10412289984)*x_ref[0]+(1.34015931457)*x_ref[1]
        ref[0,4,1,1]=(-0.692340206781)*x_ref[0]+(1.08467851048)*x_ref[1]
        ref[0,4,2,0]=(-0.0972657457169)*x_ref[0]+(1.0145314899)*x_ref[1]
        ref[0,4,2,1]=(0.299871193443)*x_ref[0]+(-0.470271033392)*x_ref[1]
        ref[1,0,0,0]=(-0.724229380092)*x_ref[0]+(1.14448991134)*x_ref[1]
        ref[1,0,0,1]=(0.154147313948)*x_ref[0]+(-0.318350887819)*x_ref[1]
        ref[1,0,1,0]=(-1.2997484551)*x_ref[0]+(0.00645216870018)*x_ref[1]
        ref[1,0,1,1]=(1.53322180988)*x_ref[0]+(-0.356405349895)*x_ref[1]
        ref[1,0,2,0]=(0.971184272781)*x_ref[0]+(1.07821077146)*x_ref[1]
        ref[1,0,2,1]=(0.000530732762419)*x_ref[0]+(0.955934156926)*x_ref[1]
        ref[1,1,0,0]=(-0.180982534825)*x_ref[0]+(1.33554642886)*x_ref[1]
        ref[1,1,0,1]=(-0.419614369168)*x_ref[0]+(-0.51201636845)*x_ref[1]
        ref[1,1,1,0]=(-0.263686990653)*x_ref[0]+(-0.463089568134)*x_ref[1]
        ref[1,1,1,1]=(-0.792982985293)*x_ref[0]+(-0.720992794822)*x_ref[1]
        ref[1,1,2,0]=(1.72981591311)*x_ref[0]+(-0.305755746837)*x_ref[1]
        ref[1,1,2,1]=(0.874057242127)*x_ref[0]+(0.247839316695)*x_ref[1]
        ref[1,2,0,0]=(1.20726244725)*x_ref[0]+(0.44394602451)*x_ref[1]
        ref[1,2,0,1]=(-0.516016832395)*x_ref[0]+(0.036120494771)*x_ref[1]
        ref[1,2,1,0]=(-0.157245073014)*x_ref[0]+(0.845240869121)*x_ref[1]
        ref[1,2,1,1]=(0.588082465766)*x_ref[0]+(0.195937663431)*x_ref[1]
        ref[1,2,2,0]=(0.095726763028)*x_ref[0]+(-0.177757031723)*x_ref[1]
        ref[1,2,2,1]=(1.73215168642)*x_ref[0]+(-0.189154112108)*x_ref[1]
        ref[1,3,0,0]=(-1.76621021161)*x_ref[0]+(0.367931299124)*x_ref[1]
        ref[1,3,0,1]=(0.0222607642178)*x_ref[0]+(0.466368587137)*x_ref[1]
        ref[1,3,1,0]=(0.412132176201)*x_ref[0]+(0.553362139129)*x_ref[1]
        ref[1,3,1,1]=(-1.08306042349)*x_ref[0]+(-0.834943125248)*x_ref[1]
        ref[1,3,2,0]=(-0.417451930632)*x_ref[0]+(-0.659233486131)*x_ref[1]
        ref[1,3,2,1]=(-0.400333106519)*x_ref[0]+(0.854889031906)*x_ref[1]
        ref[1,4,0,0]=(-0.758414654342)*x_ref[0]+(0.838957251391)*x_ref[1]
        ref[1,4,0,1]=(-0.353472866001)*x_ref[0]+(0.397138506541)*x_ref[1]
        ref[1,4,1,0]=(0.127641591916)*x_ref[0]+(0.327603318651)*x_ref[1]
        ref[1,4,1,1]=(-0.341040162377)*x_ref[0]+(-0.846062323778)*x_ref[1]
        ref[1,4,2,0]=(-0.293356226448)*x_ref[0]+(-0.700648323903)*x_ref[1]
        ref[1,4,2,1]=(-0.228549394688)*x_ref[0]+(1.43724938876)*x_ref[1]
        ref[2,0,0,0]=(-0.322023776234)*x_ref[0]+(0.564889094588)*x_ref[1]
        ref[2,0,0,1]=(-1.11907082102)*x_ref[0]+(-1.39374772922)*x_ref[1]
        ref[2,0,1,0]=(-0.705740617022)*x_ref[0]+(0.511499218842)*x_ref[1]
        ref[2,0,1,1]=(0.454111255253)*x_ref[0]+(-1.69843226881)*x_ref[1]
        ref[2,0,2,0]=(-0.317781478691)*x_ref[0]+(-1.35991309273)*x_ref[1]
        ref[2,0,2,1]=(1.54613408914)*x_ref[0]+(1.32901459805)*x_ref[1]
        ref[2,1,0,0]=(0.837517924208)*x_ref[0]+(-0.386149637251)*x_ref[1]
        ref[2,1,0,1]=(0.961896547496)*x_ref[0]+(0.617366532183)*x_ref[1]
        ref[2,1,1,0]=(-0.124021064331)*x_ref[0]+(1.45238030663)*x_ref[1]
        ref[2,1,1,1]=(-1.34194701555)*x_ref[0]+(-0.479310148218)*x_ref[1]
        ref[2,1,2,0]=(-0.813567438318)*x_ref[0]+(0.35270597356)*x_ref[1]
        ref[2,1,2,1]=(0.833034787475)*x_ref[0]+(-0.108648813503)*x_ref[1]
        ref[2,2,0,0]=(-1.36052757115)*x_ref[0]+(-0.528858737068)*x_ref[1]
        ref[2,2,0,1]=(0.348085718404)*x_ref[0]+(-0.550199132198)*x_ref[1]
        ref[2,2,1,0]=(-0.592064932053)*x_ref[0]+(0.268729595158)*x_ref[1]
        ref[2,2,1,1]=(0.762201961036)*x_ref[0]+(-0.170275800619)*x_ref[1]
        ref[2,2,2,0]=(0.292508257449)*x_ref[0]+(0.546726460436)*x_ref[1]
        ref[2,2,2,1]=(-0.832981468483)*x_ref[0]+(-1.01580974142)*x_ref[1]
        ref[2,3,0,0]=(-0.0778970282444)*x_ref[0]+(-0.0588582131822)*x_ref[1]
        ref[2,3,0,1]=(-0.462860486423)*x_ref[0]+(0.938910405107)*x_ref[1]
        ref[2,3,1,0]=(0.274056697467)*x_ref[0]+(0.0938367562972)*x_ref[1]
        ref[2,3,1,1]=(0.168689167517)*x_ref[0]+(-0.486345829711)*x_ref[1]
        ref[2,3,2,0]=(-1.33699397485)*x_ref[0]+(-0.863623492326)*x_ref[1]
        ref[2,3,2,1]=(-0.317889793298)*x_ref[0]+(0.47526951533)*x_ref[1]
        ref[2,4,0,0]=(-1.26561106959)*x_ref[0]+(0.0486883091899)*x_ref[1]
        ref[2,4,0,1]=(-1.37742147186)*x_ref[0]+(-1.14365670741)*x_ref[1]
        ref[2,4,1,0]=(0.891487506649)*x_ref[0]+(-1.02149537779)*x_ref[1]
        ref[2,4,1,1]=(0.120446738176)*x_ref[0]+(1.128322636)*x_ref[1]
        ref[2,4,2,0]=(-0.322931882495)*x_ref[0]+(-0.00993625804164)*x_ref[1]
        ref[2,4,2,1]=(0.17764198759)*x_ref[0]+(0.311937789824)*x_ref[1]
        ref[3,0,0,0]=(0.482490440467)*x_ref[0]+(0.98496437249)*x_ref[1]
        ref[3,0,0,1]=(-0.174114820802)*x_ref[0]+(0.442866651846)*x_ref[1]
        ref[3,0,1,0]=(-0.934180386688)*x_ref[0]+(-1.33361314957)*x_ref[1]
        ref[3,0,1,1]=(0.0452285929122)*x_ref[0]+(0.686705971575)*x_ref[1]
        ref[3,0,2,0]=(0.839801283293)*x_ref[0]+(-0.967697483261)*x_ref[1]
        ref[3,0,2,1]=(0.14250663194)*x_ref[0]+(-1.3515960506)*x_ref[1]
        ref[3,1,0,0]=(-0.322855905664)*x_ref[0]+(0.177298330914)*x_ref[1]
        ref[3,1,0,1]=(-0.376491734039)*x_ref[0]+(0.451059427164)*x_ref[1]
        ref[3,1,1,0]=(0.817537444737)*x_ref[0]+(-0.93908061343)*x_ref[1]
        ref[3,1,1,1]=(-0.0753530118606)*x_ref[0]+(-0.0718596035883)*x_ref[1]
        ref[3,1,2,0]=(-1.48772358944)*x_ref[0]+(0.414823807711)*x_ref[1]
        ref[3,1,2,1]=(0.857324284797)*x_ref[0]+(0.47131796901)*x_ref[1]
        ref[3,2,0,0]=(0.440034775536)*x_ref[0]+(0.712881563582)*x_ref[1]
        ref[3,2,0,1]=(-1.05547066)*x_ref[0]+(0.718956424358)*x_ref[1]
        ref[3,2,1,0]=(-1.40526673086)*x_ref[0]+(-0.366871542135)*x_ref[1]
        ref[3,2,1,1]=(1.65950470328)*x_ref[0]+(-1.65805816594)*x_ref[1]
        ref[3,2,2,0]=(1.09096918876)*x_ref[0]+(-1.33325242756)*x_ref[1]
        ref[3,2,2,1]=(-0.591939877469)*x_ref[0]+(1.0303591776)*x_ref[1]
        ref[3,3,0,0]=(-0.969754780364)*x_ref[0]+(-0.0145852822005)*x_ref[1]
        ref[3,3,0,1]=(-1.22384841753)*x_ref[0]+(-0.0867406833581)*x_ref[1]
        ref[3,3,1,0]=(0.673272022388)*x_ref[0]+(0.947486168095)*x_ref[1]
        ref[3,3,1,1]=(-0.41449883684)*x_ref[0]+(0.137714746421)*x_ref[1]
        ref[3,3,2,0]=(0.874598196414)*x_ref[0]+(0.436570682065)*x_ref[1]
        ref[3,3,2,1]=(0.127986513106)*x_ref[0]+(0.955109334387)*x_ref[1]
        ref[3,4,0,0]=(0.226160174687)*x_ref[0]+(-0.258282204272)*x_ref[1]
        ref[3,4,0,1]=(-0.157570106985)*x_ref[0]+(-0.299582244336)*x_ref[1]
        ref[3,4,1,0]=(0.305056978098)*x_ref[0]+(1.34807929179)*x_ref[1]
        ref[3,4,1,1]=(1.16863934732)*x_ref[0]+(-0.265384337877)*x_ref[1]
        ref[3,4,2,0]=(1.26265936998)*x_ref[0]+(0.379703748135)*x_ref[1]
        ref[3,4,2,1]=(-0.392236938104)*x_ref[0]+(0.829704189106)*x_ref[1]
      else:
        arg[0,0,0,0]=(0.657131010834)*x[0]+(0.925056704895)*x[1]+(0.405133059909)*x[2]
        arg[0,0,0,1]=(1.00133275423)*x[0]+(0.804604848583)*x[1]+(-0.127865284577)*x[2]
        arg[0,0,1,0]=(-1.43854141172)*x[0]+(0.461779683792)*x[1]+(-1.35332048209)*x[2]
        arg[0,0,1,1]=(0.00990645894894)*x[0]+(-0.127101800988)*x[1]+(-0.105358765313)*x[2]
        arg[0,0,2,0]=(0.116780997053)*x[0]+(1.14704063829)*x[1]+(-0.0190951296257)*x[2]
        arg[0,0,2,1]=(0.485352133633)*x[0]+(0.232507104974)*x[1]+(-1.13067055888)*x[2]
        arg[0,1,0,0]=(-0.350822544873)*x[0]+(-0.812391120478)*x[1]+(0.215924121867)*x[2]
        arg[0,1,0,1]=(-0.351284651716)*x[0]+(-0.221990440191)*x[1]+(-0.223075105598)*x[2]
        arg[0,1,1,0]=(-0.148376043654)*x[0]+(0.571632161665)*x[1]+(0.519163900078)*x[2]
        arg[0,1,1,1]=(-1.45038964408)*x[0]+(0.302237803429)*x[1]+(0.742483356143)*x[2]
        arg[0,1,2,0]=(-0.504270471573)*x[0]+(0.450418202232)*x[1]+(-0.111778328066)*x[2]
        arg[0,1,2,1]=(-0.0249264677684)*x[0]+(-1.50643033256)*x[1]+(-1.35775630013)*x[2]
        arg[0,2,0,0]=(0.303211268309)*x[0]+(-0.464853430563)*x[1]+(0.649958565051)*x[2]
        arg[0,2,0,1]=(-0.099640500895)*x[0]+(-0.190653996191)*x[1]+(1.25662361146)*x[2]
        arg[0,2,1,0]=(0.196051285683)*x[0]+(-0.0405207084252)*x[1]+(-0.247626612712)*x[2]
        arg[0,2,1,1]=(0.466541264508)*x[0]+(1.12203001637)*x[1]+(0.214880867458)*x[2]
        arg[0,2,2,0]=(-0.066399925725)*x[0]+(-0.246038048418)*x[1]+(-1.36436628718)*x[2]
        arg[0,2,2,1]=(1.11813537863)*x[0]+(-1.09166896181)*x[1]+(0.321936214135)*x[2]
        arg[0,3,0,0]=(0.374873324445)*x[0]+(-0.465943548272)*x[1]+(0.0190372923331)*x[2]
        arg[0,3,0,1]=(0.833988907462)*x[0]+(-1.01195758004)*x[1]+(0.237305351727)*x[2]
        arg[0,3,1,0]=(0.311900067643)*x[0]+(-1.11508238642)*x[1]+(-1.04912391557)*x[2]
        arg[0,3,1,1]=(-0.211460991448)*x[0]+(-0.768164308901)*x[1]+(-0.450128967205)*x[2]
        arg[0,3,2,0]=(-0.412678195326)*x[0]+(-0.0808659043088)*x[1]+(0.588002415672)*x[2]
        arg[0,3,2,1]=(-1.25661069251)*x[0]+(-1.51596081235)*x[1]+(-1.38001371964)*x[2]
        arg[0,4,0,0]=(0.0822281994319)*x[0]+(-0.726137997384)*x[1]+(0.20532237702)*x[2]
        arg[0,4,0,1]=(-0.538872250604)*x[0]+(0.669139417413)*x[1]+(0.235909494023)*x[2]
        arg[0,4,1,0]=(0.171512295446)*x[0]+(-1.37296529156)*x[1]+(1.25081328539)*x[2]
        arg[0,4,1,1]=(-0.440062802901)*x[0]+(1.32591598513)*x[1]+(0.802645502519)*x[2]
        arg[0,4,2,0]=(1.55151533491)*x[0]+(0.923232387757)*x[1]+(-0.238979557225)*x[2]
        arg[0,4,2,1]=(-0.193306423923)*x[0]+(-0.565399635257)*x[1]+(-0.149043846608)*x[2]
        arg[1,0,0,0]=(0.152247539316)*x[0]+(-1.219763414)*x[1]+(0.836742536417)*x[2]
        arg[1,0,0,1]=(0.279236863017)*x[0]+(0.14118432621)*x[1]+(-0.33072070595)*x[2]
        arg[1,0,1,0]=(-0.242848555181)*x[0]+(-0.981144530476)*x[1]+(-0.925220699112)*x[2]
        arg[1,0,1,1]=(-1.8765425995)*x[0]+(-0.0878096696901)*x[1]+(0.133461124587)*x[2]
        arg[1,0,2,0]=(-0.282373512491)*x[0]+(0.640862824684)*x[1]+(-1.41391625471)*x[2]
        arg[1,0,2,1]=(0.168260007994)*x[0]+(0.660595443409)*x[1]+(1.14161006755)*x[2]
        arg[1,1,0,0]=(-0.160956063437)*x[0]+(-0.60581390216)*x[1]+(1.28589122111)*x[2]
        arg[1,1,0,1]=(0.0969448014041)*x[0]+(-1.63715420302)*x[1]+(-0.350879443881)*x[2]
        arg[1,1,1,0]=(0.0878659583502)*x[0]+(1.13265578932)*x[1]+(-0.615846752768)*x[2]
        arg[1,1,1,1]=(-0.434597962742)*x[0]+(-0.177532580699)*x[1]+(0.543921860391)*x[2]
        arg[1,1,2,0]=(-0.7739187322)*x[0]+(-0.285452235523)*x[1]+(-0.0426663737344)*x[2]
        arg[1,1,2,1]=(0.796313217511)*x[0]+(-1.12909584404)*x[1]+(1.26302048961)*x[2]
        arg[1,2,0,0]=(1.62555299859)*x[0]+(-1.1660403401)*x[1]+(-0.668592413245)*x[2]
        arg[1,2,0,1]=(-0.807742954877)*x[0]+(-0.238532476847)*x[1]+(-1.13575189406)*x[2]
        arg[1,2,1,0]=(-0.139280450456)*x[0]+(-0.043306992095)*x[1]+(0.832351898761)*x[2]
        arg[1,2,1,1]=(0.16731150192)*x[0]+(0.0705653171485)*x[1]+(0.12943106453)*x[2]
        arg[1,2,2,0]=(0.669247838367)*x[0]+(1.10177451339)*x[1]+(1.21850224582)*x[2]
        arg[1,2,2,1]=(-0.779527532537)*x[0]+(0.10943711539)*x[1]+(0.0286417760627)*x[2]
        arg[1,3,0,0]=(-0.500367153143)*x[0]+(-0.0315766921687)*x[1]+(0.0814977959486)*x[2]
        arg[1,3,0,1]=(1.44010930702)*x[0]+(0.985266827661)*x[1]+(-0.4777280416)*x[2]
        arg[1,3,1,0]=(-1.09856159209)*x[0]+(0.914767947742)*x[1]+(0.257975836816)*x[2]
        arg[1,3,1,1]=(1.20572454864)*x[0]+(-1.18899739672)*x[1]+(-0.881322688875)*x[2]
        arg[1,3,2,0]=(0.0683727807399)*x[0]+(0.27893403097)*x[1]+(0.532141657019)*x[2]
        arg[1,3,2,1]=(1.22927878362)*x[0]+(-0.746664730318)*x[1]+(0.706635763178)*x[2]
        arg[1,4,0,0]=(-0.056381118157)*x[0]+(0.497463663176)*x[1]+(-0.0170016763418)*x[2]
        arg[1,4,0,1]=(0.220396391337)*x[0]+(-0.311295539709)*x[1]+(-0.0925408150537)*x[2]
        arg[1,4,1,0]=(-1.26814101481)*x[0]+(1.36544235985)*x[1]+(-0.706223605983)*x[2]
        arg[1,4,1,1]=(-0.151531826396)*x[0]+(0.709418004406)*x[1]+(0.361464674136)*x[2]
        arg[1,4,2,0]=(-1.29637464683)*x[0]+(1.12643501861)*x[1]+(0.0119437305981)*x[2]
        arg[1,4,2,1]=(-0.127448007113)*x[0]+(0.423470578567)*x[1]+(0.57274403474)*x[2]
        arg[2,0,0,0]=(-0.753513005537)*x[0]+(-1.42995648081)*x[1]+(-0.748578276092)*x[2]
        arg[2,0,0,1]=(-0.529225216075)*x[0]+(0.723600505708)*x[1]+(-0.265987788752)*x[2]
        arg[2,0,1,0]=(0.570242632099)*x[0]+(0.7183918375)*x[1]+(-0.127239408099)*x[2]
        arg[2,0,1,1]=(0.23496854128)*x[0]+(0.642830624861)*x[1]+(0.734149374044)*x[2]
        arg[2,0,2,0]=(-0.302583224046)*x[0]+(0.0684944027143)*x[1]+(0.156995461545)*x[2]
        arg[2,0,2,1]=(-0.549180930856)*x[0]+(0.334205359442)*x[1]+(1.05254428528)*x[2]
        arg[2,1,0,0]=(-0.234799499951)*x[0]+(1.17826730605)*x[1]+(1.09739059469)*x[2]
        arg[2,1,0,1]=(-0.187707076501)*x[0]+(0.011330011831)*x[1]+(-0.233605047453)*x[2]
        arg[2,1,1,0]=(0.290424858876)*x[0]+(0.12865400019)*x[1]+(-1.01515096279)*x[2]
        arg[2,1,1,1]=(-1.05797585987)*x[0]+(0.679572972092)*x[1]+(1.04481198693)*x[2]
        arg[2,1,2,0]=(0.240336170335)*x[0]+(0.567306847159)*x[1]+(0.224146880555)*x[2]
        arg[2,1,2,1]=(-0.568589696446)*x[0]+(1.37586080732)*x[1]+(-0.292086990344)*x[2]
        arg[2,2,0,0]=(1.16127933088)*x[0]+(-0.205088565357)*x[1]+(-0.0262686243919)*x[2]
        arg[2,2,0,1]=(1.17913541301)*x[0]+(-0.953527588914)*x[1]+(-1.0594095352)*x[2]
        arg[2,2,1,0]=(-0.116342811686)*x[0]+(0.459831226378)*x[1]+(0.532324650374)*x[2]
        arg[2,2,1,1]=(-0.904379150749)*x[0]+(0.0197201393782)*x[1]+(1.16488799459)*x[2]
        arg[2,2,2,0]=(0.288441825708)*x[0]+(-0.00162365695053)*x[1]+(-0.696961968544)*x[2]
        arg[2,2,2,1]=(-0.278951886334)*x[0]+(-0.438973263931)*x[1]+(-0.617387268055)*x[2]
        arg[2,3,0,0]=(-0.575840111545)*x[0]+(-0.430583405817)*x[1]+(0.678495026384)*x[2]
        arg[2,3,0,1]=(-0.0712810127233)*x[0]+(-0.479271089471)*x[1]+(-0.561687276819)*x[2]
        arg[2,3,1,0]=(0.00293903214402)*x[0]+(0.707224264038)*x[1]+(0.863200461564)*x[2]
        arg[2,3,1,1]=(-0.982725212738)*x[0]+(1.12174250462)*x[1]+(0.692163622484)*x[2]
        arg[2,3,2,0]=(0.563182079719)*x[0]+(0.592670114886)*x[1]+(1.12490471848)*x[2]
        arg[2,3,2,1]=(0.674175588576)*x[0]+(0.125281804818)*x[1]+(0.260310807977)*x[2]
        arg[2,4,0,0]=(-0.468871891598)*x[0]+(0.628476449634)*x[1]+(1.44798783935)*x[2]
        arg[2,4,0,1]=(0.331618946805)*x[0]+(1.18126898042)*x[1]+(1.73036582285)*x[2]
        arg[2,4,1,0]=(0.643190142583)*x[0]+(0.958102048788)*x[1]+(-1.17509922201)*x[2]
        arg[2,4,1,1]=(0.140678141755)*x[0]+(-0.964205570281)*x[1]+(1.33722038246)*x[2]
        arg[2,4,2,0]=(1.00405626052)*x[0]+(-0.766180509924)*x[1]+(-1.12787337858)*x[2]
        arg[2,4,2,1]=(0.015701651285)*x[0]+(-0.0327586063712)*x[1]+(-1.60288759856)*x[2]
        arg[3,0,0,0]=(-0.0648346068133)*x[0]+(0.0319743940508)*x[1]+(-0.0885983990043)*x[2]
        arg[3,0,0,1]=(0.440191590799)*x[0]+(-0.224950534025)*x[1]+(-1.59285278381)*x[2]
        arg[3,0,1,0]=(-1.39437403492)*x[0]+(0.718181967012)*x[1]+(-0.320282124786)*x[2]
        arg[3,0,1,1]=(0.212062143443)*x[0]+(0.552967368897)*x[1]+(-0.805494016261)*x[2]
        arg[3,0,2,0]=(0.123432537169)*x[0]+(-0.124637759512)*x[1]+(-0.627011571378)*x[2]
        arg[3,0,2,1]=(-0.112696666742)*x[0]+(-0.264195102375)*x[1]+(0.0720389973432)*x[2]
        arg[3,1,0,0]=(-0.13525504711)*x[0]+(1.02997422411)*x[1]+(-1.00797986845)*x[2]
        arg[3,1,0,1]=(-0.0675047394459)*x[0]+(1.39934707441)*x[1]+(-0.563292489037)*x[2]
        arg[3,1,1,0]=(-0.597786255631)*x[0]+(0.969544698551)*x[1]+(-0.0636563520143)*x[2]
        arg[3,1,1,1]=(0.615372705023)*x[0]+(-0.312239254268)*x[1]+(1.11108440155)*x[2]
        arg[3,1,2,0]=(0.876145642141)*x[0]+(0.117271425687)*x[1]+(1.10498467015)*x[2]
        arg[3,1,2,1]=(-0.881281403659)*x[0]+(0.240299603044)*x[1]+(0.14672255814)*x[2]
        arg[3,2,0,0]=(-1.72712606675)*x[0]+(0.20271220583)*x[1]+(1.10555001416)*x[2]
        arg[3,2,0,1]=(0.251203467531)*x[0]+(1.14031566726)*x[1]+(0.13511684108)*x[2]
        arg[3,2,1,0]=(-0.318589273012)*x[0]+(1.22432558956)*x[1]+(0.0532448118609)*x[2]
        arg[3,2,1,1]=(0.238528986535)*x[0]+(0.181951236101)*x[1]+(0.0816544114367)*x[2]
        arg[3,2,2,0]=(0.691665718353)*x[0]+(0.238132802916)*x[1]+(1.0244502435)*x[2]
        arg[3,2,2,1]=(-1.64759886895)*x[0]+(1.06200853706)*x[1]+(-0.0980251738655)*x[2]
        arg[3,3,0,0]=(-0.0493788966334)*x[0]+(-0.471962146153)*x[1]+(0.156759333734)*x[2]
        arg[3,3,0,1]=(-0.304056988684)*x[0]+(-0.298060200774)*x[1]+(-0.455474275206)*x[2]
        arg[3,3,1,0]=(-0.62366321672)*x[0]+(0.229030608831)*x[1]+(0.737950069935)*x[2]
        arg[3,3,1,1]=(0.132130038359)*x[0]+(0.749997139232)*x[1]+(-0.793138255465)*x[2]
        arg[3,3,2,0]=(1.71562899872)*x[0]+(0.659770561383)*x[1]+(0.59886052289)*x[2]
        arg[3,3,2,1]=(0.863622138509)*x[0]+(-0.626586688059)*x[1]+(0.744519466248)*x[2]
        arg[3,4,0,0]=(0.0214494537522)*x[0]+(0.732351688803)*x[1]+(-0.46140559894)*x[2]
        arg[3,4,0,1]=(1.33581215441)*x[0]+(-1.16528842961)*x[1]+(-0.888687584892)*x[2]
        arg[3,4,1,0]=(-0.971883445892)*x[0]+(0.715423530047)*x[1]+(-0.0743908860543)*x[2]
        arg[3,4,1,1]=(-0.941321450189)*x[0]+(-0.120890461937)*x[1]+(0.929504478456)*x[2]
        arg[3,4,2,0]=(0.515343485731)*x[0]+(0.323772401243)*x[1]+(-1.03459959098)*x[2]
        arg[3,4,2,1]=(0.50815257234)*x[0]+(0.461750357229)*x[1]+(0.0260081346396)*x[2]
        ref[0,0,0,0]=(0.657131010834)*x_ref[0]+(0.925056704895)*x_ref[1]+(0.405133059909)*x_ref[2]
        ref[0,0,0,1]=(1.00133275423)*x_ref[0]+(0.804604848583)*x_ref[1]+(-0.127865284577)*x_ref[2]
        ref[0,0,1,0]=(-1.43854141172)*x_ref[0]+(0.461779683792)*x_ref[1]+(-1.35332048209)*x_ref[2]
        ref[0,0,1,1]=(0.00990645894894)*x_ref[0]+(-0.127101800988)*x_ref[1]+(-0.105358765313)*x_ref[2]
        ref[0,0,2,0]=(0.116780997053)*x_ref[0]+(1.14704063829)*x_ref[1]+(-0.0190951296257)*x_ref[2]
        ref[0,0,2,1]=(0.485352133633)*x_ref[0]+(0.232507104974)*x_ref[1]+(-1.13067055888)*x_ref[2]
        ref[0,1,0,0]=(-0.350822544873)*x_ref[0]+(-0.812391120478)*x_ref[1]+(0.215924121867)*x_ref[2]
        ref[0,1,0,1]=(-0.351284651716)*x_ref[0]+(-0.221990440191)*x_ref[1]+(-0.223075105598)*x_ref[2]
        ref[0,1,1,0]=(-0.148376043654)*x_ref[0]+(0.571632161665)*x_ref[1]+(0.519163900078)*x_ref[2]
        ref[0,1,1,1]=(-1.45038964408)*x_ref[0]+(0.302237803429)*x_ref[1]+(0.742483356143)*x_ref[2]
        ref[0,1,2,0]=(-0.504270471573)*x_ref[0]+(0.450418202232)*x_ref[1]+(-0.111778328066)*x_ref[2]
        ref[0,1,2,1]=(-0.0249264677684)*x_ref[0]+(-1.50643033256)*x_ref[1]+(-1.35775630013)*x_ref[2]
        ref[0,2,0,0]=(0.303211268309)*x_ref[0]+(-0.464853430563)*x_ref[1]+(0.649958565051)*x_ref[2]
        ref[0,2,0,1]=(-0.099640500895)*x_ref[0]+(-0.190653996191)*x_ref[1]+(1.25662361146)*x_ref[2]
        ref[0,2,1,0]=(0.196051285683)*x_ref[0]+(-0.0405207084252)*x_ref[1]+(-0.247626612712)*x_ref[2]
        ref[0,2,1,1]=(0.466541264508)*x_ref[0]+(1.12203001637)*x_ref[1]+(0.214880867458)*x_ref[2]
        ref[0,2,2,0]=(-0.066399925725)*x_ref[0]+(-0.246038048418)*x_ref[1]+(-1.36436628718)*x_ref[2]
        ref[0,2,2,1]=(1.11813537863)*x_ref[0]+(-1.09166896181)*x_ref[1]+(0.321936214135)*x_ref[2]
        ref[0,3,0,0]=(0.374873324445)*x_ref[0]+(-0.465943548272)*x_ref[1]+(0.0190372923331)*x_ref[2]
        ref[0,3,0,1]=(0.833988907462)*x_ref[0]+(-1.01195758004)*x_ref[1]+(0.237305351727)*x_ref[2]
        ref[0,3,1,0]=(0.311900067643)*x_ref[0]+(-1.11508238642)*x_ref[1]+(-1.04912391557)*x_ref[2]
        ref[0,3,1,1]=(-0.211460991448)*x_ref[0]+(-0.768164308901)*x_ref[1]+(-0.450128967205)*x_ref[2]
        ref[0,3,2,0]=(-0.412678195326)*x_ref[0]+(-0.0808659043088)*x_ref[1]+(0.588002415672)*x_ref[2]
        ref[0,3,2,1]=(-1.25661069251)*x_ref[0]+(-1.51596081235)*x_ref[1]+(-1.38001371964)*x_ref[2]
        ref[0,4,0,0]=(0.0822281994319)*x_ref[0]+(-0.726137997384)*x_ref[1]+(0.20532237702)*x_ref[2]
        ref[0,4,0,1]=(-0.538872250604)*x_ref[0]+(0.669139417413)*x_ref[1]+(0.235909494023)*x_ref[2]
        ref[0,4,1,0]=(0.171512295446)*x_ref[0]+(-1.37296529156)*x_ref[1]+(1.25081328539)*x_ref[2]
        ref[0,4,1,1]=(-0.440062802901)*x_ref[0]+(1.32591598513)*x_ref[1]+(0.802645502519)*x_ref[2]
        ref[0,4,2,0]=(1.55151533491)*x_ref[0]+(0.923232387757)*x_ref[1]+(-0.238979557225)*x_ref[2]
        ref[0,4,2,1]=(-0.193306423923)*x_ref[0]+(-0.565399635257)*x_ref[1]+(-0.149043846608)*x_ref[2]
        ref[1,0,0,0]=(0.152247539316)*x_ref[0]+(-1.219763414)*x_ref[1]+(0.836742536417)*x_ref[2]
        ref[1,0,0,1]=(0.279236863017)*x_ref[0]+(0.14118432621)*x_ref[1]+(-0.33072070595)*x_ref[2]
        ref[1,0,1,0]=(-0.242848555181)*x_ref[0]+(-0.981144530476)*x_ref[1]+(-0.925220699112)*x_ref[2]
        ref[1,0,1,1]=(-1.8765425995)*x_ref[0]+(-0.0878096696901)*x_ref[1]+(0.133461124587)*x_ref[2]
        ref[1,0,2,0]=(-0.282373512491)*x_ref[0]+(0.640862824684)*x_ref[1]+(-1.41391625471)*x_ref[2]
        ref[1,0,2,1]=(0.168260007994)*x_ref[0]+(0.660595443409)*x_ref[1]+(1.14161006755)*x_ref[2]
        ref[1,1,0,0]=(-0.160956063437)*x_ref[0]+(-0.60581390216)*x_ref[1]+(1.28589122111)*x_ref[2]
        ref[1,1,0,1]=(0.0969448014041)*x_ref[0]+(-1.63715420302)*x_ref[1]+(-0.350879443881)*x_ref[2]
        ref[1,1,1,0]=(0.0878659583502)*x_ref[0]+(1.13265578932)*x_ref[1]+(-0.615846752768)*x_ref[2]
        ref[1,1,1,1]=(-0.434597962742)*x_ref[0]+(-0.177532580699)*x_ref[1]+(0.543921860391)*x_ref[2]
        ref[1,1,2,0]=(-0.7739187322)*x_ref[0]+(-0.285452235523)*x_ref[1]+(-0.0426663737344)*x_ref[2]
        ref[1,1,2,1]=(0.796313217511)*x_ref[0]+(-1.12909584404)*x_ref[1]+(1.26302048961)*x_ref[2]
        ref[1,2,0,0]=(1.62555299859)*x_ref[0]+(-1.1660403401)*x_ref[1]+(-0.668592413245)*x_ref[2]
        ref[1,2,0,1]=(-0.807742954877)*x_ref[0]+(-0.238532476847)*x_ref[1]+(-1.13575189406)*x_ref[2]
        ref[1,2,1,0]=(-0.139280450456)*x_ref[0]+(-0.043306992095)*x_ref[1]+(0.832351898761)*x_ref[2]
        ref[1,2,1,1]=(0.16731150192)*x_ref[0]+(0.0705653171485)*x_ref[1]+(0.12943106453)*x_ref[2]
        ref[1,2,2,0]=(0.669247838367)*x_ref[0]+(1.10177451339)*x_ref[1]+(1.21850224582)*x_ref[2]
        ref[1,2,2,1]=(-0.779527532537)*x_ref[0]+(0.10943711539)*x_ref[1]+(0.0286417760627)*x_ref[2]
        ref[1,3,0,0]=(-0.500367153143)*x_ref[0]+(-0.0315766921687)*x_ref[1]+(0.0814977959486)*x_ref[2]
        ref[1,3,0,1]=(1.44010930702)*x_ref[0]+(0.985266827661)*x_ref[1]+(-0.4777280416)*x_ref[2]
        ref[1,3,1,0]=(-1.09856159209)*x_ref[0]+(0.914767947742)*x_ref[1]+(0.257975836816)*x_ref[2]
        ref[1,3,1,1]=(1.20572454864)*x_ref[0]+(-1.18899739672)*x_ref[1]+(-0.881322688875)*x_ref[2]
        ref[1,3,2,0]=(0.0683727807399)*x_ref[0]+(0.27893403097)*x_ref[1]+(0.532141657019)*x_ref[2]
        ref[1,3,2,1]=(1.22927878362)*x_ref[0]+(-0.746664730318)*x_ref[1]+(0.706635763178)*x_ref[2]
        ref[1,4,0,0]=(-0.056381118157)*x_ref[0]+(0.497463663176)*x_ref[1]+(-0.0170016763418)*x_ref[2]
        ref[1,4,0,1]=(0.220396391337)*x_ref[0]+(-0.311295539709)*x_ref[1]+(-0.0925408150537)*x_ref[2]
        ref[1,4,1,0]=(-1.26814101481)*x_ref[0]+(1.36544235985)*x_ref[1]+(-0.706223605983)*x_ref[2]
        ref[1,4,1,1]=(-0.151531826396)*x_ref[0]+(0.709418004406)*x_ref[1]+(0.361464674136)*x_ref[2]
        ref[1,4,2,0]=(-1.29637464683)*x_ref[0]+(1.12643501861)*x_ref[1]+(0.0119437305981)*x_ref[2]
        ref[1,4,2,1]=(-0.127448007113)*x_ref[0]+(0.423470578567)*x_ref[1]+(0.57274403474)*x_ref[2]
        ref[2,0,0,0]=(-0.753513005537)*x_ref[0]+(-1.42995648081)*x_ref[1]+(-0.748578276092)*x_ref[2]
        ref[2,0,0,1]=(-0.529225216075)*x_ref[0]+(0.723600505708)*x_ref[1]+(-0.265987788752)*x_ref[2]
        ref[2,0,1,0]=(0.570242632099)*x_ref[0]+(0.7183918375)*x_ref[1]+(-0.127239408099)*x_ref[2]
        ref[2,0,1,1]=(0.23496854128)*x_ref[0]+(0.642830624861)*x_ref[1]+(0.734149374044)*x_ref[2]
        ref[2,0,2,0]=(-0.302583224046)*x_ref[0]+(0.0684944027143)*x_ref[1]+(0.156995461545)*x_ref[2]
        ref[2,0,2,1]=(-0.549180930856)*x_ref[0]+(0.334205359442)*x_ref[1]+(1.05254428528)*x_ref[2]
        ref[2,1,0,0]=(-0.234799499951)*x_ref[0]+(1.17826730605)*x_ref[1]+(1.09739059469)*x_ref[2]
        ref[2,1,0,1]=(-0.187707076501)*x_ref[0]+(0.011330011831)*x_ref[1]+(-0.233605047453)*x_ref[2]
        ref[2,1,1,0]=(0.290424858876)*x_ref[0]+(0.12865400019)*x_ref[1]+(-1.01515096279)*x_ref[2]
        ref[2,1,1,1]=(-1.05797585987)*x_ref[0]+(0.679572972092)*x_ref[1]+(1.04481198693)*x_ref[2]
        ref[2,1,2,0]=(0.240336170335)*x_ref[0]+(0.567306847159)*x_ref[1]+(0.224146880555)*x_ref[2]
        ref[2,1,2,1]=(-0.568589696446)*x_ref[0]+(1.37586080732)*x_ref[1]+(-0.292086990344)*x_ref[2]
        ref[2,2,0,0]=(1.16127933088)*x_ref[0]+(-0.205088565357)*x_ref[1]+(-0.0262686243919)*x_ref[2]
        ref[2,2,0,1]=(1.17913541301)*x_ref[0]+(-0.953527588914)*x_ref[1]+(-1.0594095352)*x_ref[2]
        ref[2,2,1,0]=(-0.116342811686)*x_ref[0]+(0.459831226378)*x_ref[1]+(0.532324650374)*x_ref[2]
        ref[2,2,1,1]=(-0.904379150749)*x_ref[0]+(0.0197201393782)*x_ref[1]+(1.16488799459)*x_ref[2]
        ref[2,2,2,0]=(0.288441825708)*x_ref[0]+(-0.00162365695053)*x_ref[1]+(-0.696961968544)*x_ref[2]
        ref[2,2,2,1]=(-0.278951886334)*x_ref[0]+(-0.438973263931)*x_ref[1]+(-0.617387268055)*x_ref[2]
        ref[2,3,0,0]=(-0.575840111545)*x_ref[0]+(-0.430583405817)*x_ref[1]+(0.678495026384)*x_ref[2]
        ref[2,3,0,1]=(-0.0712810127233)*x_ref[0]+(-0.479271089471)*x_ref[1]+(-0.561687276819)*x_ref[2]
        ref[2,3,1,0]=(0.00293903214402)*x_ref[0]+(0.707224264038)*x_ref[1]+(0.863200461564)*x_ref[2]
        ref[2,3,1,1]=(-0.982725212738)*x_ref[0]+(1.12174250462)*x_ref[1]+(0.692163622484)*x_ref[2]
        ref[2,3,2,0]=(0.563182079719)*x_ref[0]+(0.592670114886)*x_ref[1]+(1.12490471848)*x_ref[2]
        ref[2,3,2,1]=(0.674175588576)*x_ref[0]+(0.125281804818)*x_ref[1]+(0.260310807977)*x_ref[2]
        ref[2,4,0,0]=(-0.468871891598)*x_ref[0]+(0.628476449634)*x_ref[1]+(1.44798783935)*x_ref[2]
        ref[2,4,0,1]=(0.331618946805)*x_ref[0]+(1.18126898042)*x_ref[1]+(1.73036582285)*x_ref[2]
        ref[2,4,1,0]=(0.643190142583)*x_ref[0]+(0.958102048788)*x_ref[1]+(-1.17509922201)*x_ref[2]
        ref[2,4,1,1]=(0.140678141755)*x_ref[0]+(-0.964205570281)*x_ref[1]+(1.33722038246)*x_ref[2]
        ref[2,4,2,0]=(1.00405626052)*x_ref[0]+(-0.766180509924)*x_ref[1]+(-1.12787337858)*x_ref[2]
        ref[2,4,2,1]=(0.015701651285)*x_ref[0]+(-0.0327586063712)*x_ref[1]+(-1.60288759856)*x_ref[2]
        ref[3,0,0,0]=(-0.0648346068133)*x_ref[0]+(0.0319743940508)*x_ref[1]+(-0.0885983990043)*x_ref[2]
        ref[3,0,0,1]=(0.440191590799)*x_ref[0]+(-0.224950534025)*x_ref[1]+(-1.59285278381)*x_ref[2]
        ref[3,0,1,0]=(-1.39437403492)*x_ref[0]+(0.718181967012)*x_ref[1]+(-0.320282124786)*x_ref[2]
        ref[3,0,1,1]=(0.212062143443)*x_ref[0]+(0.552967368897)*x_ref[1]+(-0.805494016261)*x_ref[2]
        ref[3,0,2,0]=(0.123432537169)*x_ref[0]+(-0.124637759512)*x_ref[1]+(-0.627011571378)*x_ref[2]
        ref[3,0,2,1]=(-0.112696666742)*x_ref[0]+(-0.264195102375)*x_ref[1]+(0.0720389973432)*x_ref[2]
        ref[3,1,0,0]=(-0.13525504711)*x_ref[0]+(1.02997422411)*x_ref[1]+(-1.00797986845)*x_ref[2]
        ref[3,1,0,1]=(-0.0675047394459)*x_ref[0]+(1.39934707441)*x_ref[1]+(-0.563292489037)*x_ref[2]
        ref[3,1,1,0]=(-0.597786255631)*x_ref[0]+(0.969544698551)*x_ref[1]+(-0.0636563520143)*x_ref[2]
        ref[3,1,1,1]=(0.615372705023)*x_ref[0]+(-0.312239254268)*x_ref[1]+(1.11108440155)*x_ref[2]
        ref[3,1,2,0]=(0.876145642141)*x_ref[0]+(0.117271425687)*x_ref[1]+(1.10498467015)*x_ref[2]
        ref[3,1,2,1]=(-0.881281403659)*x_ref[0]+(0.240299603044)*x_ref[1]+(0.14672255814)*x_ref[2]
        ref[3,2,0,0]=(-1.72712606675)*x_ref[0]+(0.20271220583)*x_ref[1]+(1.10555001416)*x_ref[2]
        ref[3,2,0,1]=(0.251203467531)*x_ref[0]+(1.14031566726)*x_ref[1]+(0.13511684108)*x_ref[2]
        ref[3,2,1,0]=(-0.318589273012)*x_ref[0]+(1.22432558956)*x_ref[1]+(0.0532448118609)*x_ref[2]
        ref[3,2,1,1]=(0.238528986535)*x_ref[0]+(0.181951236101)*x_ref[1]+(0.0816544114367)*x_ref[2]
        ref[3,2,2,0]=(0.691665718353)*x_ref[0]+(0.238132802916)*x_ref[1]+(1.0244502435)*x_ref[2]
        ref[3,2,2,1]=(-1.64759886895)*x_ref[0]+(1.06200853706)*x_ref[1]+(-0.0980251738655)*x_ref[2]
        ref[3,3,0,0]=(-0.0493788966334)*x_ref[0]+(-0.471962146153)*x_ref[1]+(0.156759333734)*x_ref[2]
        ref[3,3,0,1]=(-0.304056988684)*x_ref[0]+(-0.298060200774)*x_ref[1]+(-0.455474275206)*x_ref[2]
        ref[3,3,1,0]=(-0.62366321672)*x_ref[0]+(0.229030608831)*x_ref[1]+(0.737950069935)*x_ref[2]
        ref[3,3,1,1]=(0.132130038359)*x_ref[0]+(0.749997139232)*x_ref[1]+(-0.793138255465)*x_ref[2]
        ref[3,3,2,0]=(1.71562899872)*x_ref[0]+(0.659770561383)*x_ref[1]+(0.59886052289)*x_ref[2]
        ref[3,3,2,1]=(0.863622138509)*x_ref[0]+(-0.626586688059)*x_ref[1]+(0.744519466248)*x_ref[2]
        ref[3,4,0,0]=(0.0214494537522)*x_ref[0]+(0.732351688803)*x_ref[1]+(-0.46140559894)*x_ref[2]
        ref[3,4,0,1]=(1.33581215441)*x_ref[0]+(-1.16528842961)*x_ref[1]+(-0.888687584892)*x_ref[2]
        ref[3,4,1,0]=(-0.971883445892)*x_ref[0]+(0.715423530047)*x_ref[1]+(-0.0743908860543)*x_ref[2]
        ref[3,4,1,1]=(-0.941321450189)*x_ref[0]+(-0.120890461937)*x_ref[1]+(0.929504478456)*x_ref[2]
        ref[3,4,2,0]=(0.515343485731)*x_ref[0]+(0.323772401243)*x_ref[1]+(-1.03459959098)*x_ref[2]
        ref[3,4,2,1]=(0.50815257234)*x_ref[0]+(0.461750357229)*x_ref[1]+(0.0260081346396)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-0.459241599813)*x[0]+(-0.773093048708)*x[1]
        ref=(-0.459241599813)*x_ref[0]+(-0.773093048708)*x_ref[1]
      else:
        s=(1.54748967569)*x[0]+(0.307874797615)*x[1]+(-0.414822431494)*x[2]
        ref=(1.54748967569)*x_ref[0]+(0.307874797615)*x_ref[1]+(-0.414822431494)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-1.51228566012)*x[0]+(-0.814844304147)*x[1]
        s[1]=(-0.881464001412)*x[0]+(0.764098134348)*x[1]
        ref[0]=(-1.51228566012)*x[0]+(-0.814844304147)*x[1]
        ref[1]=(-0.881464001412)*x[0]+(0.764098134348)*x[1]
      else:
        s[0]=(-0.91410305648)*x[0]+(-0.253301678985)*x[1]+(-0.00826967639394)*x[2]
        s[1]=(-0.592096115904)*x[0]+(-1.05985873602)*x[1]+(0.00147658091963)*x[2]
        ref[0]=(-0.91410305648)*x[0]+(-0.253301678985)*x[1]+(-0.00826967639394)*x[2]
        ref[1]=(-0.592096115904)*x[0]+(-1.05985873602)*x[1]+(0.00147658091963)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.209549210666)*x[0]+(1.05992146041)*x[1]
        s[0,1]=(-0.389388721283)*x[0]+(-0.122353380381)*x[1]
        s[0,2]=(1.48834814863)*x[0]+(1.11546828469)*x[1]
        s[0,3]=(-0.326661291423)*x[0]+(-0.20061861948)*x[1]
        s[0,4]=(0.941218540101)*x[0]+(-1.04582695848)*x[1]
        s[1,0]=(-0.196538410782)*x[0]+(0.110519724291)*x[1]
        s[1,1]=(0.361129728605)*x[0]+(0.740228471382)*x[1]
        s[1,2]=(-1.56560253129)*x[0]+(-0.585194140493)*x[1]
        s[1,3]=(0.177982070606)*x[0]+(-0.392964886439)*x[1]
        s[1,4]=(0.700111909929)*x[0]+(-0.366757279647)*x[1]
        s[2,0]=(-0.809083106921)*x[0]+(-0.447731857848)*x[1]
        s[2,1]=(0.867383974301)*x[0]+(-1.21953693153)*x[1]
        s[2,2]=(-0.859520524516)*x[0]+(-0.071452049869)*x[1]
        s[2,3]=(0.00398656657353)*x[0]+(1.23623708268)*x[1]
        s[2,4]=(0.166456051527)*x[0]+(1.47611415007)*x[1]
        s[3,0]=(-0.778396522621)*x[0]+(-0.386585278282)*x[1]
        s[3,1]=(0.231266833602)*x[0]+(0.19536493852)*x[1]
        s[3,2]=(-1.20923484879)*x[0]+(-0.490207641665)*x[1]
        s[3,3]=(0.455704265159)*x[0]+(0.109883193088)*x[1]
        s[3,4]=(-0.329454340826)*x[0]+(0.467154318354)*x[1]
        ref[0,0]=(0.209549210666)*x_ref[0]+(1.05992146041)*x_ref[1]
        ref[0,1]=(-0.389388721283)*x_ref[0]+(-0.122353380381)*x_ref[1]
        ref[0,2]=(1.48834814863)*x_ref[0]+(1.11546828469)*x_ref[1]
        ref[0,3]=(-0.326661291423)*x_ref[0]+(-0.20061861948)*x_ref[1]
        ref[0,4]=(0.941218540101)*x_ref[0]+(-1.04582695848)*x_ref[1]
        ref[1,0]=(-0.196538410782)*x_ref[0]+(0.110519724291)*x_ref[1]
        ref[1,1]=(0.361129728605)*x_ref[0]+(0.740228471382)*x_ref[1]
        ref[1,2]=(-1.56560253129)*x_ref[0]+(-0.585194140493)*x_ref[1]
        ref[1,3]=(0.177982070606)*x_ref[0]+(-0.392964886439)*x_ref[1]
        ref[1,4]=(0.700111909929)*x_ref[0]+(-0.366757279647)*x_ref[1]
        ref[2,0]=(-0.809083106921)*x_ref[0]+(-0.447731857848)*x_ref[1]
        ref[2,1]=(0.867383974301)*x_ref[0]+(-1.21953693153)*x_ref[1]
        ref[2,2]=(-0.859520524516)*x_ref[0]+(-0.071452049869)*x_ref[1]
        ref[2,3]=(0.00398656657353)*x_ref[0]+(1.23623708268)*x_ref[1]
        ref[2,4]=(0.166456051527)*x_ref[0]+(1.47611415007)*x_ref[1]
        ref[3,0]=(-0.778396522621)*x_ref[0]+(-0.386585278282)*x_ref[1]
        ref[3,1]=(0.231266833602)*x_ref[0]+(0.19536493852)*x_ref[1]
        ref[3,2]=(-1.20923484879)*x_ref[0]+(-0.490207641665)*x_ref[1]
        ref[3,3]=(0.455704265159)*x_ref[0]+(0.109883193088)*x_ref[1]
        ref[3,4]=(-0.329454340826)*x_ref[0]+(0.467154318354)*x_ref[1]
      else:
        s[0,0]=(-1.51270660613)*x[0]+(-1.23166098758)*x[1]+(0.00126913231069)*x[2]
        s[0,1]=(0.0783399703925)*x[0]+(0.704421352501)*x[1]+(-0.0527087709157)*x[2]
        s[0,2]=(0.212858037142)*x[0]+(0.879768825167)*x[1]+(0.3376329277)*x[2]
        s[0,3]=(-1.02885491754)*x[0]+(-0.349446901437)*x[1]+(-0.518150496566)*x[2]
        s[0,4]=(-0.632641407112)*x[0]+(0.736682065273)*x[1]+(0.849728972959)*x[2]
        s[1,0]=(-1.31434627789)*x[0]+(-1.40408223096)*x[1]+(0.184109480479)*x[2]
        s[1,1]=(-0.258805252155)*x[0]+(-0.794639774)*x[1]+(-0.370761920362)*x[2]
        s[1,2]=(0.645662860304)*x[0]+(0.196908807781)*x[1]+(-0.348791058482)*x[2]
        s[1,3]=(-0.397782835091)*x[0]+(0.292971279611)*x[1]+(-0.952746843146)*x[2]
        s[1,4]=(0.226551429583)*x[0]+(-0.891234988447)*x[1]+(-0.00309756588396)*x[2]
        s[2,0]=(-0.418562068509)*x[0]+(0.43739854393)*x[1]+(0.0509772218411)*x[2]
        s[2,1]=(0.783305063847)*x[0]+(0.824783771448)*x[1]+(0.21157958706)*x[2]
        s[2,2]=(-0.557598710851)*x[0]+(0.973487903111)*x[1]+(-0.400806427014)*x[2]
        s[2,3]=(1.02110121409)*x[0]+(-0.189675236979)*x[1]+(-0.692006170319)*x[2]
        s[2,4]=(0.204984594869)*x[0]+(-0.591307306263)*x[1]+(-0.0198714935717)*x[2]
        s[3,0]=(-0.524519709351)*x[0]+(0.263836205807)*x[1]+(-0.163880326974)*x[2]
        s[3,1]=(0.268669519472)*x[0]+(0.379462793693)*x[1]+(-0.297560957033)*x[2]
        s[3,2]=(-0.13621646905)*x[0]+(0.650614413717)*x[1]+(0.44018137263)*x[2]
        s[3,3]=(0.261445807973)*x[0]+(-0.00264446194008)*x[1]+(-0.168858453474)*x[2]
        s[3,4]=(-0.247581704325)*x[0]+(-1.9405050847)*x[1]+(1.5206187195)*x[2]
        ref[0,0]=(-1.51270660613)*x_ref[0]+(-1.23166098758)*x_ref[1]+(0.00126913231069)*x_ref[2]
        ref[0,1]=(0.0783399703925)*x_ref[0]+(0.704421352501)*x_ref[1]+(-0.0527087709157)*x_ref[2]
        ref[0,2]=(0.212858037142)*x_ref[0]+(0.879768825167)*x_ref[1]+(0.3376329277)*x_ref[2]
        ref[0,3]=(-1.02885491754)*x_ref[0]+(-0.349446901437)*x_ref[1]+(-0.518150496566)*x_ref[2]
        ref[0,4]=(-0.632641407112)*x_ref[0]+(0.736682065273)*x_ref[1]+(0.849728972959)*x_ref[2]
        ref[1,0]=(-1.31434627789)*x_ref[0]+(-1.40408223096)*x_ref[1]+(0.184109480479)*x_ref[2]
        ref[1,1]=(-0.258805252155)*x_ref[0]+(-0.794639774)*x_ref[1]+(-0.370761920362)*x_ref[2]
        ref[1,2]=(0.645662860304)*x_ref[0]+(0.196908807781)*x_ref[1]+(-0.348791058482)*x_ref[2]
        ref[1,3]=(-0.397782835091)*x_ref[0]+(0.292971279611)*x_ref[1]+(-0.952746843146)*x_ref[2]
        ref[1,4]=(0.226551429583)*x_ref[0]+(-0.891234988447)*x_ref[1]+(-0.00309756588396)*x_ref[2]
        ref[2,0]=(-0.418562068509)*x_ref[0]+(0.43739854393)*x_ref[1]+(0.0509772218411)*x_ref[2]
        ref[2,1]=(0.783305063847)*x_ref[0]+(0.824783771448)*x_ref[1]+(0.21157958706)*x_ref[2]
        ref[2,2]=(-0.557598710851)*x_ref[0]+(0.973487903111)*x_ref[1]+(-0.400806427014)*x_ref[2]
        ref[2,3]=(1.02110121409)*x_ref[0]+(-0.189675236979)*x_ref[1]+(-0.692006170319)*x_ref[2]
        ref[2,4]=(0.204984594869)*x_ref[0]+(-0.591307306263)*x_ref[1]+(-0.0198714935717)*x_ref[2]
        ref[3,0]=(-0.524519709351)*x_ref[0]+(0.263836205807)*x_ref[1]+(-0.163880326974)*x_ref[2]
        ref[3,1]=(0.268669519472)*x_ref[0]+(0.379462793693)*x_ref[1]+(-0.297560957033)*x_ref[2]
        ref[3,2]=(-0.13621646905)*x_ref[0]+(0.650614413717)*x_ref[1]+(0.44018137263)*x_ref[2]
        ref[3,3]=(0.261445807973)*x_ref[0]+(-0.00264446194008)*x_ref[1]+(-0.168858453474)*x_ref[2]
        ref[3,4]=(-0.247581704325)*x_ref[0]+(-1.9405050847)*x_ref[1]+(1.5206187195)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(1.27061375599)*x[0]+(0.693010110721)*x[1]
        s[0,0,1]=(0.385267649608)*x[0]+(-0.409112804165)*x[1]
        s[0,1,0]=(0.235254219396)*x[0]+(0.00806676427989)*x[1]
        s[0,1,1]=(0.546338474187)*x[0]+(0.269505133228)*x[1]
        s[1,0,0]=(-0.252450318463)*x[0]+(0.565357325098)*x[1]
        s[1,0,1]=(0.519066028574)*x[0]+(-0.151896081553)*x[1]
        s[1,1,0]=(0.583246542849)*x[0]+(0.74106904645)*x[1]
        s[1,1,1]=(0.283330295456)*x[0]+(1.3671331438)*x[1]
        s[2,0,0]=(0.295333306073)*x[0]+(-0.561691152526)*x[1]
        s[2,0,1]=(0.637514074862)*x[0]+(-0.484517233318)*x[1]
        s[2,1,0]=(-0.206769011353)*x[0]+(-0.0115621456599)*x[1]
        s[2,1,1]=(1.40165335416)*x[0]+(-1.43439048946)*x[1]
        s[3,0,0]=(0.567950789213)*x[0]+(-1.6427426351)*x[1]
        s[3,0,1]=(0.913125584361)*x[0]+(-0.192143890031)*x[1]
        s[3,1,0]=(-0.480605427139)*x[0]+(-0.474866017698)*x[1]
        s[3,1,1]=(-0.137178314752)*x[0]+(0.28719953047)*x[1]
        s[4,0,0]=(0.697948943346)*x[0]+(-0.907672862452)*x[1]
        s[4,0,1]=(-1.62690254564)*x[0]+(0.914366733304)*x[1]
        s[4,1,0]=(0.900337784632)*x[0]+(-0.0470217841933)*x[1]
        s[4,1,1]=(0.313193728362)*x[0]+(-0.122194456503)*x[1]
        s[5,0,0]=(0.606327952796)*x[0]+(-0.917393520302)*x[1]
        s[5,0,1]=(-0.262565816623)*x[0]+(0.232087670002)*x[1]
        s[5,1,0]=(-0.0786114664469)*x[0]+(0.252622361658)*x[1]
        s[5,1,1]=(0.193290467334)*x[0]+(0.529096699149)*x[1]
        ref[0,0,0]=(1.27061375599)*x_ref[0]+(0.693010110721)*x_ref[1]
        ref[0,0,1]=(0.385267649608)*x_ref[0]+(-0.409112804165)*x_ref[1]
        ref[0,1,0]=(0.235254219396)*x_ref[0]+(0.00806676427989)*x_ref[1]
        ref[0,1,1]=(0.546338474187)*x_ref[0]+(0.269505133228)*x_ref[1]
        ref[1,0,0]=(-0.252450318463)*x_ref[0]+(0.565357325098)*x_ref[1]
        ref[1,0,1]=(0.519066028574)*x_ref[0]+(-0.151896081553)*x_ref[1]
        ref[1,1,0]=(0.583246542849)*x_ref[0]+(0.74106904645)*x_ref[1]
        ref[1,1,1]=(0.283330295456)*x_ref[0]+(1.3671331438)*x_ref[1]
        ref[2,0,0]=(0.295333306073)*x_ref[0]+(-0.561691152526)*x_ref[1]
        ref[2,0,1]=(0.637514074862)*x_ref[0]+(-0.484517233318)*x_ref[1]
        ref[2,1,0]=(-0.206769011353)*x_ref[0]+(-0.0115621456599)*x_ref[1]
        ref[2,1,1]=(1.40165335416)*x_ref[0]+(-1.43439048946)*x_ref[1]
        ref[3,0,0]=(0.567950789213)*x_ref[0]+(-1.6427426351)*x_ref[1]
        ref[3,0,1]=(0.913125584361)*x_ref[0]+(-0.192143890031)*x_ref[1]
        ref[3,1,0]=(-0.480605427139)*x_ref[0]+(-0.474866017698)*x_ref[1]
        ref[3,1,1]=(-0.137178314752)*x_ref[0]+(0.28719953047)*x_ref[1]
        ref[4,0,0]=(0.697948943346)*x_ref[0]+(-0.907672862452)*x_ref[1]
        ref[4,0,1]=(-1.62690254564)*x_ref[0]+(0.914366733304)*x_ref[1]
        ref[4,1,0]=(0.900337784632)*x_ref[0]+(-0.0470217841933)*x_ref[1]
        ref[4,1,1]=(0.313193728362)*x_ref[0]+(-0.122194456503)*x_ref[1]
        ref[5,0,0]=(0.606327952796)*x_ref[0]+(-0.917393520302)*x_ref[1]
        ref[5,0,1]=(-0.262565816623)*x_ref[0]+(0.232087670002)*x_ref[1]
        ref[5,1,0]=(-0.0786114664469)*x_ref[0]+(0.252622361658)*x_ref[1]
        ref[5,1,1]=(0.193290467334)*x_ref[0]+(0.529096699149)*x_ref[1]
      else:
        s[0,0,0]=(-0.50338473477)*x[0]+(-0.164749126249)*x[1]+(-0.554967226755)*x[2]
        s[0,0,1]=(-0.776721662344)*x[0]+(0.84671767153)*x[1]+(-1.62325361007)*x[2]
        s[0,1,0]=(-0.553665559679)*x[0]+(0.762038412432)*x[1]+(0.219451112807)*x[2]
        s[0,1,1]=(-0.295200279284)*x[0]+(-0.366906381317)*x[1]+(-0.114064671233)*x[2]
        s[1,0,0]=(-0.627912305852)*x[0]+(1.05590035705)*x[1]+(-0.357432981292)*x[2]
        s[1,0,1]=(-0.103518100064)*x[0]+(0.0285901975227)*x[1]+(0.272743015592)*x[2]
        s[1,1,0]=(-1.48220340227)*x[0]+(1.16859770671)*x[1]+(-1.15282038238)*x[2]
        s[1,1,1]=(0.758570299808)*x[0]+(-0.211340586797)*x[1]+(0.263787144967)*x[2]
        s[2,0,0]=(-1.28649435546)*x[0]+(-1.34545925137)*x[1]+(-1.23248562938)*x[2]
        s[2,0,1]=(-0.685813167263)*x[0]+(-0.811060892059)*x[1]+(-0.905743168112)*x[2]
        s[2,1,0]=(-1.27012789685)*x[0]+(-1.45423179754)*x[1]+(-0.461566077486)*x[2]
        s[2,1,1]=(0.764272676292)*x[0]+(1.79354721116)*x[1]+(1.54561480836)*x[2]
        s[3,0,0]=(0.483458668467)*x[0]+(-0.520255901158)*x[1]+(0.931135044848)*x[2]
        s[3,0,1]=(0.0731034526543)*x[0]+(0.61085360436)*x[1]+(1.15630425582)*x[2]
        s[3,1,0]=(-1.13345961818)*x[0]+(1.53627408722)*x[1]+(0.0394537895884)*x[2]
        s[3,1,1]=(1.93910425421)*x[0]+(-0.245667398856)*x[1]+(0.26326290067)*x[2]
        s[4,0,0]=(-0.316109947551)*x[0]+(0.523912230594)*x[1]+(0.428093796157)*x[2]
        s[4,0,1]=(-0.421586699534)*x[0]+(0.0956846699257)*x[1]+(0.153930459299)*x[2]
        s[4,1,0]=(1.29911179477)*x[0]+(-0.668540796991)*x[1]+(0.226289836438)*x[2]
        s[4,1,1]=(-1.2174140353)*x[0]+(0.577999358777)*x[1]+(-0.756367967534)*x[2]
        s[5,0,0]=(-0.158832852108)*x[0]+(-0.444245105421)*x[1]+(0.00253446453922)*x[2]
        s[5,0,1]=(0.654262047318)*x[0]+(-0.573424848715)*x[1]+(-1.74703307574)*x[2]
        s[5,1,0]=(1.18879231974)*x[0]+(1.01556300264)*x[1]+(-0.902278634733)*x[2]
        s[5,1,1]=(-0.0136999102655)*x[0]+(-0.123143996985)*x[1]+(1.70755156491)*x[2]
        ref[0,0,0]=(-0.50338473477)*x_ref[0]+(-0.164749126249)*x_ref[1]+(-0.554967226755)*x_ref[2]
        ref[0,0,1]=(-0.776721662344)*x_ref[0]+(0.84671767153)*x_ref[1]+(-1.62325361007)*x_ref[2]
        ref[0,1,0]=(-0.553665559679)*x_ref[0]+(0.762038412432)*x_ref[1]+(0.219451112807)*x_ref[2]
        ref[0,1,1]=(-0.295200279284)*x_ref[0]+(-0.366906381317)*x_ref[1]+(-0.114064671233)*x_ref[2]
        ref[1,0,0]=(-0.627912305852)*x_ref[0]+(1.05590035705)*x_ref[1]+(-0.357432981292)*x_ref[2]
        ref[1,0,1]=(-0.103518100064)*x_ref[0]+(0.0285901975227)*x_ref[1]+(0.272743015592)*x_ref[2]
        ref[1,1,0]=(-1.48220340227)*x_ref[0]+(1.16859770671)*x_ref[1]+(-1.15282038238)*x_ref[2]
        ref[1,1,1]=(0.758570299808)*x_ref[0]+(-0.211340586797)*x_ref[1]+(0.263787144967)*x_ref[2]
        ref[2,0,0]=(-1.28649435546)*x_ref[0]+(-1.34545925137)*x_ref[1]+(-1.23248562938)*x_ref[2]
        ref[2,0,1]=(-0.685813167263)*x_ref[0]+(-0.811060892059)*x_ref[1]+(-0.905743168112)*x_ref[2]
        ref[2,1,0]=(-1.27012789685)*x_ref[0]+(-1.45423179754)*x_ref[1]+(-0.461566077486)*x_ref[2]
        ref[2,1,1]=(0.764272676292)*x_ref[0]+(1.79354721116)*x_ref[1]+(1.54561480836)*x_ref[2]
        ref[3,0,0]=(0.483458668467)*x_ref[0]+(-0.520255901158)*x_ref[1]+(0.931135044848)*x_ref[2]
        ref[3,0,1]=(0.0731034526543)*x_ref[0]+(0.61085360436)*x_ref[1]+(1.15630425582)*x_ref[2]
        ref[3,1,0]=(-1.13345961818)*x_ref[0]+(1.53627408722)*x_ref[1]+(0.0394537895884)*x_ref[2]
        ref[3,1,1]=(1.93910425421)*x_ref[0]+(-0.245667398856)*x_ref[1]+(0.26326290067)*x_ref[2]
        ref[4,0,0]=(-0.316109947551)*x_ref[0]+(0.523912230594)*x_ref[1]+(0.428093796157)*x_ref[2]
        ref[4,0,1]=(-0.421586699534)*x_ref[0]+(0.0956846699257)*x_ref[1]+(0.153930459299)*x_ref[2]
        ref[4,1,0]=(1.29911179477)*x_ref[0]+(-0.668540796991)*x_ref[1]+(0.226289836438)*x_ref[2]
        ref[4,1,1]=(-1.2174140353)*x_ref[0]+(0.577999358777)*x_ref[1]+(-0.756367967534)*x_ref[2]
        ref[5,0,0]=(-0.158832852108)*x_ref[0]+(-0.444245105421)*x_ref[1]+(0.00253446453922)*x_ref[2]
        ref[5,0,1]=(0.654262047318)*x_ref[0]+(-0.573424848715)*x_ref[1]+(-1.74703307574)*x_ref[2]
        ref[5,1,0]=(1.18879231974)*x_ref[0]+(1.01556300264)*x_ref[1]+(-0.902278634733)*x_ref[2]
        ref[5,1,1]=(-0.0136999102655)*x_ref[0]+(-0.123143996985)*x_ref[1]+(1.70755156491)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.224003216764)*x[0]+(-1.53537141395)*x[1]
        s[0,0,0,1]=(-0.415108576452)*x[0]+(0.512267468115)*x[1]
        s[0,0,1,0]=(1.23475149747)*x[0]+(0.834141411624)*x[1]
        s[0,0,1,1]=(1.23314379215)*x[0]+(0.097729946154)*x[1]
        s[0,0,2,0]=(1.78249035938)*x[0]+(0.149287882179)*x[1]
        s[0,0,2,1]=(0.268374768258)*x[0]+(0.844004814805)*x[1]
        s[0,1,0,0]=(-0.0921256378797)*x[0]+(0.54607120248)*x[1]
        s[0,1,0,1]=(-0.108118391324)*x[0]+(1.44715323804)*x[1]
        s[0,1,1,0]=(0.642892153115)*x[0]+(1.29939851925)*x[1]
        s[0,1,1,1]=(-0.346502475731)*x[0]+(-1.59694261779)*x[1]
        s[0,1,2,0]=(1.17858221695)*x[0]+(-0.0467327691871)*x[1]
        s[0,1,2,1]=(-0.0676520776338)*x[0]+(-1.19701821874)*x[1]
        s[0,2,0,0]=(0.0594877869572)*x[0]+(0.4691178171)*x[1]
        s[0,2,0,1]=(-0.471134574689)*x[0]+(-0.0774880925496)*x[1]
        s[0,2,1,0]=(0.957705182694)*x[0]+(-0.148346694036)*x[1]
        s[0,2,1,1]=(1.35974448438)*x[0]+(0.976348738984)*x[1]
        s[0,2,2,0]=(-0.0888644417678)*x[0]+(0.185465627916)*x[1]
        s[0,2,2,1]=(0.366706758508)*x[0]+(0.424423271146)*x[1]
        s[0,3,0,0]=(0.300438450637)*x[0]+(0.489300186408)*x[1]
        s[0,3,0,1]=(-0.837684340885)*x[0]+(1.11058469871)*x[1]
        s[0,3,1,0]=(-0.685409716574)*x[0]+(1.29529487058)*x[1]
        s[0,3,1,1]=(-0.734298461298)*x[0]+(1.49154390821)*x[1]
        s[0,3,2,0]=(-0.360678115755)*x[0]+(-1.31090940382)*x[1]
        s[0,3,2,1]=(-1.76090388864)*x[0]+(-0.197058042294)*x[1]
        s[0,4,0,0]=(-0.647957185052)*x[0]+(-1.07397957837)*x[1]
        s[0,4,0,1]=(1.15138043713)*x[0]+(1.27703705411)*x[1]
        s[0,4,1,0]=(0.880204392036)*x[0]+(0.317776491177)*x[1]
        s[0,4,1,1]=(1.75846459794)*x[0]+(0.58329739948)*x[1]
        s[0,4,2,0]=(0.321739539217)*x[0]+(0.155992474445)*x[1]
        s[0,4,2,1]=(-0.831765424787)*x[0]+(-0.174239473413)*x[1]
        s[1,0,0,0]=(-0.315436217973)*x[0]+(-1.50414837046)*x[1]
        s[1,0,0,1]=(-0.20319068189)*x[0]+(-0.471694875132)*x[1]
        s[1,0,1,0]=(1.11728129989)*x[0]+(-1.56713796242)*x[1]
        s[1,0,1,1]=(1.57812938291)*x[0]+(0.768357124158)*x[1]
        s[1,0,2,0]=(0.846908340877)*x[0]+(0.00535056985765)*x[1]
        s[1,0,2,1]=(1.65497801331)*x[0]+(-0.787783905283)*x[1]
        s[1,1,0,0]=(-1.88514071539)*x[0]+(0.169147044809)*x[1]
        s[1,1,0,1]=(0.360553360055)*x[0]+(0.239599872139)*x[1]
        s[1,1,1,0]=(-0.870583030697)*x[0]+(-0.956757500255)*x[1]
        s[1,1,1,1]=(0.398902070984)*x[0]+(-0.299144054803)*x[1]
        s[1,1,2,0]=(0.683520893803)*x[0]+(0.569312151808)*x[1]
        s[1,1,2,1]=(-0.142857438428)*x[0]+(0.216222456467)*x[1]
        s[1,2,0,0]=(-0.512922081732)*x[0]+(1.55472942936)*x[1]
        s[1,2,0,1]=(-0.00159718420231)*x[0]+(0.126511414314)*x[1]
        s[1,2,1,0]=(0.687145297745)*x[0]+(-0.392573700665)*x[1]
        s[1,2,1,1]=(-0.308147808315)*x[0]+(0.0500851036579)*x[1]
        s[1,2,2,0]=(-0.121563796894)*x[0]+(-0.671555650221)*x[1]
        s[1,2,2,1]=(-1.14943669955)*x[0]+(-1.12860967665)*x[1]
        s[1,3,0,0]=(0.796492123166)*x[0]+(-0.594727403527)*x[1]
        s[1,3,0,1]=(0.0776108758324)*x[0]+(0.758254260971)*x[1]
        s[1,3,1,0]=(-0.974221864735)*x[0]+(1.02650607362)*x[1]
        s[1,3,1,1]=(-0.476728696056)*x[0]+(1.39394954201)*x[1]
        s[1,3,2,0]=(-0.276890313424)*x[0]+(-0.42434071217)*x[1]
        s[1,3,2,1]=(-0.45671100973)*x[0]+(-0.271986551291)*x[1]
        s[1,4,0,0]=(0.415315726206)*x[0]+(0.25174859473)*x[1]
        s[1,4,0,1]=(0.718116945034)*x[0]+(0.300885379458)*x[1]
        s[1,4,1,0]=(0.0362286683347)*x[0]+(0.0860658750969)*x[1]
        s[1,4,1,1]=(0.966944206449)*x[0]+(0.605527604606)*x[1]
        s[1,4,2,0]=(-0.681446172251)*x[0]+(-0.0201721796333)*x[1]
        s[1,4,2,1]=(1.16600295744)*x[0]+(1.10261292926)*x[1]
        s[2,0,0,0]=(-0.239525904251)*x[0]+(0.428475333805)*x[1]
        s[2,0,0,1]=(0.0817932891372)*x[0]+(0.472895457716)*x[1]
        s[2,0,1,0]=(0.96571284204)*x[0]+(1.30765537984)*x[1]
        s[2,0,1,1]=(-0.27274079215)*x[0]+(-0.973090020231)*x[1]
        s[2,0,2,0]=(-0.586811136701)*x[0]+(0.0282301965253)*x[1]
        s[2,0,2,1]=(-0.988289117174)*x[0]+(-0.982428801403)*x[1]
        s[2,1,0,0]=(-0.97313872216)*x[0]+(0.264946309189)*x[1]
        s[2,1,0,1]=(0.477464114989)*x[0]+(0.735813896712)*x[1]
        s[2,1,1,0]=(-0.999655360356)*x[0]+(0.0906084300723)*x[1]
        s[2,1,1,1]=(-0.672480642088)*x[0]+(-0.0511287334859)*x[1]
        s[2,1,2,0]=(-1.06854561704)*x[0]+(-0.188249587015)*x[1]
        s[2,1,2,1]=(-0.439030823181)*x[0]+(0.76989999863)*x[1]
        s[2,2,0,0]=(0.935351178562)*x[0]+(-0.732961189428)*x[1]
        s[2,2,0,1]=(-0.201449422629)*x[0]+(-0.920429771001)*x[1]
        s[2,2,1,0]=(-0.0397141758777)*x[0]+(1.23810013859)*x[1]
        s[2,2,1,1]=(-0.853635033788)*x[0]+(0.982464467568)*x[1]
        s[2,2,2,0]=(-1.35330887855)*x[0]+(0.200038841813)*x[1]
        s[2,2,2,1]=(0.0388084224165)*x[0]+(-0.312397495662)*x[1]
        s[2,3,0,0]=(-0.762894528313)*x[0]+(0.206223517392)*x[1]
        s[2,3,0,1]=(0.0823646028803)*x[0]+(-1.1834743671)*x[1]
        s[2,3,1,0]=(0.302380236653)*x[0]+(-0.725627755476)*x[1]
        s[2,3,1,1]=(-0.12698195797)*x[0]+(-0.289106042968)*x[1]
        s[2,3,2,0]=(-1.33992252725)*x[0]+(-0.903068424118)*x[1]
        s[2,3,2,1]=(-0.128913469794)*x[0]+(0.392464023413)*x[1]
        s[2,4,0,0]=(-0.326550273177)*x[0]+(-0.856389747069)*x[1]
        s[2,4,0,1]=(-1.14966749976)*x[0]+(0.809029270783)*x[1]
        s[2,4,1,0]=(1.50037083169)*x[0]+(0.564371263186)*x[1]
        s[2,4,1,1]=(-0.0660188964857)*x[0]+(-0.871309804802)*x[1]
        s[2,4,2,0]=(0.527342096457)*x[0]+(1.41520765334)*x[1]
        s[2,4,2,1]=(-0.177020653581)*x[0]+(-0.218017520886)*x[1]
        s[3,0,0,0]=(-0.901819829595)*x[0]+(-0.0897360627182)*x[1]
        s[3,0,0,1]=(-1.84231567451)*x[0]+(-1.57138993977)*x[1]
        s[3,0,1,0]=(-0.692347444292)*x[0]+(-0.35994981731)*x[1]
        s[3,0,1,1]=(-1.20666616877)*x[0]+(1.31974198302)*x[1]
        s[3,0,2,0]=(-1.49613152411)*x[0]+(-1.29087493111)*x[1]
        s[3,0,2,1]=(0.00302152289294)*x[0]+(0.0596509711717)*x[1]
        s[3,1,0,0]=(-0.716314986327)*x[0]+(0.540487536594)*x[1]
        s[3,1,0,1]=(1.26797434361)*x[0]+(1.24611664766)*x[1]
        s[3,1,1,0]=(1.29133591716)*x[0]+(-0.822278152235)*x[1]
        s[3,1,1,1]=(0.736506521682)*x[0]+(0.295109058687)*x[1]
        s[3,1,2,0]=(-1.31999878466)*x[0]+(-1.06268553583)*x[1]
        s[3,1,2,1]=(-1.02353629588)*x[0]+(-0.180118460659)*x[1]
        s[3,2,0,0]=(1.2591074793)*x[0]+(0.716164110599)*x[1]
        s[3,2,0,1]=(0.0737591553738)*x[0]+(-0.180183018938)*x[1]
        s[3,2,1,0]=(0.842324803092)*x[0]+(-0.179012336351)*x[1]
        s[3,2,1,1]=(-1.04205716802)*x[0]+(0.91772961597)*x[1]
        s[3,2,2,0]=(0.774019388104)*x[0]+(0.392202481643)*x[1]
        s[3,2,2,1]=(0.583625745968)*x[0]+(0.857713356409)*x[1]
        s[3,3,0,0]=(-1.08653988672)*x[0]+(-1.47741013604)*x[1]
        s[3,3,0,1]=(-0.529552146407)*x[0]+(-0.658937785849)*x[1]
        s[3,3,1,0]=(1.8439291246)*x[0]+(-1.36075468038)*x[1]
        s[3,3,1,1]=(-0.656328225025)*x[0]+(1.18200793838)*x[1]
        s[3,3,2,0]=(-1.57290668837)*x[0]+(0.527467780284)*x[1]
        s[3,3,2,1]=(1.44984969309)*x[0]+(-1.37301308873)*x[1]
        s[3,4,0,0]=(-0.225012249574)*x[0]+(-1.80387444921)*x[1]
        s[3,4,0,1]=(-1.27157572533)*x[0]+(0.568859005411)*x[1]
        s[3,4,1,0]=(-0.381626896774)*x[0]+(-0.335697445475)*x[1]
        s[3,4,1,1]=(0.555084489073)*x[0]+(-0.659578343093)*x[1]
        s[3,4,2,0]=(0.784894297286)*x[0]+(0.955314078432)*x[1]
        s[3,4,2,1]=(0.33356366872)*x[0]+(1.38175924325)*x[1]
        ref[0,0,0,0]=(-0.224003216764)*x_ref[0]+(-1.53537141395)*x_ref[1]
        ref[0,0,0,1]=(-0.415108576452)*x_ref[0]+(0.512267468115)*x_ref[1]
        ref[0,0,1,0]=(1.23475149747)*x_ref[0]+(0.834141411624)*x_ref[1]
        ref[0,0,1,1]=(1.23314379215)*x_ref[0]+(0.097729946154)*x_ref[1]
        ref[0,0,2,0]=(1.78249035938)*x_ref[0]+(0.149287882179)*x_ref[1]
        ref[0,0,2,1]=(0.268374768258)*x_ref[0]+(0.844004814805)*x_ref[1]
        ref[0,1,0,0]=(-0.0921256378797)*x_ref[0]+(0.54607120248)*x_ref[1]
        ref[0,1,0,1]=(-0.108118391324)*x_ref[0]+(1.44715323804)*x_ref[1]
        ref[0,1,1,0]=(0.642892153115)*x_ref[0]+(1.29939851925)*x_ref[1]
        ref[0,1,1,1]=(-0.346502475731)*x_ref[0]+(-1.59694261779)*x_ref[1]
        ref[0,1,2,0]=(1.17858221695)*x_ref[0]+(-0.0467327691871)*x_ref[1]
        ref[0,1,2,1]=(-0.0676520776338)*x_ref[0]+(-1.19701821874)*x_ref[1]
        ref[0,2,0,0]=(0.0594877869572)*x_ref[0]+(0.4691178171)*x_ref[1]
        ref[0,2,0,1]=(-0.471134574689)*x_ref[0]+(-0.0774880925496)*x_ref[1]
        ref[0,2,1,0]=(0.957705182694)*x_ref[0]+(-0.148346694036)*x_ref[1]
        ref[0,2,1,1]=(1.35974448438)*x_ref[0]+(0.976348738984)*x_ref[1]
        ref[0,2,2,0]=(-0.0888644417678)*x_ref[0]+(0.185465627916)*x_ref[1]
        ref[0,2,2,1]=(0.366706758508)*x_ref[0]+(0.424423271146)*x_ref[1]
        ref[0,3,0,0]=(0.300438450637)*x_ref[0]+(0.489300186408)*x_ref[1]
        ref[0,3,0,1]=(-0.837684340885)*x_ref[0]+(1.11058469871)*x_ref[1]
        ref[0,3,1,0]=(-0.685409716574)*x_ref[0]+(1.29529487058)*x_ref[1]
        ref[0,3,1,1]=(-0.734298461298)*x_ref[0]+(1.49154390821)*x_ref[1]
        ref[0,3,2,0]=(-0.360678115755)*x_ref[0]+(-1.31090940382)*x_ref[1]
        ref[0,3,2,1]=(-1.76090388864)*x_ref[0]+(-0.197058042294)*x_ref[1]
        ref[0,4,0,0]=(-0.647957185052)*x_ref[0]+(-1.07397957837)*x_ref[1]
        ref[0,4,0,1]=(1.15138043713)*x_ref[0]+(1.27703705411)*x_ref[1]
        ref[0,4,1,0]=(0.880204392036)*x_ref[0]+(0.317776491177)*x_ref[1]
        ref[0,4,1,1]=(1.75846459794)*x_ref[0]+(0.58329739948)*x_ref[1]
        ref[0,4,2,0]=(0.321739539217)*x_ref[0]+(0.155992474445)*x_ref[1]
        ref[0,4,2,1]=(-0.831765424787)*x_ref[0]+(-0.174239473413)*x_ref[1]
        ref[1,0,0,0]=(-0.315436217973)*x_ref[0]+(-1.50414837046)*x_ref[1]
        ref[1,0,0,1]=(-0.20319068189)*x_ref[0]+(-0.471694875132)*x_ref[1]
        ref[1,0,1,0]=(1.11728129989)*x_ref[0]+(-1.56713796242)*x_ref[1]
        ref[1,0,1,1]=(1.57812938291)*x_ref[0]+(0.768357124158)*x_ref[1]
        ref[1,0,2,0]=(0.846908340877)*x_ref[0]+(0.00535056985765)*x_ref[1]
        ref[1,0,2,1]=(1.65497801331)*x_ref[0]+(-0.787783905283)*x_ref[1]
        ref[1,1,0,0]=(-1.88514071539)*x_ref[0]+(0.169147044809)*x_ref[1]
        ref[1,1,0,1]=(0.360553360055)*x_ref[0]+(0.239599872139)*x_ref[1]
        ref[1,1,1,0]=(-0.870583030697)*x_ref[0]+(-0.956757500255)*x_ref[1]
        ref[1,1,1,1]=(0.398902070984)*x_ref[0]+(-0.299144054803)*x_ref[1]
        ref[1,1,2,0]=(0.683520893803)*x_ref[0]+(0.569312151808)*x_ref[1]
        ref[1,1,2,1]=(-0.142857438428)*x_ref[0]+(0.216222456467)*x_ref[1]
        ref[1,2,0,0]=(-0.512922081732)*x_ref[0]+(1.55472942936)*x_ref[1]
        ref[1,2,0,1]=(-0.00159718420231)*x_ref[0]+(0.126511414314)*x_ref[1]
        ref[1,2,1,0]=(0.687145297745)*x_ref[0]+(-0.392573700665)*x_ref[1]
        ref[1,2,1,1]=(-0.308147808315)*x_ref[0]+(0.0500851036579)*x_ref[1]
        ref[1,2,2,0]=(-0.121563796894)*x_ref[0]+(-0.671555650221)*x_ref[1]
        ref[1,2,2,1]=(-1.14943669955)*x_ref[0]+(-1.12860967665)*x_ref[1]
        ref[1,3,0,0]=(0.796492123166)*x_ref[0]+(-0.594727403527)*x_ref[1]
        ref[1,3,0,1]=(0.0776108758324)*x_ref[0]+(0.758254260971)*x_ref[1]
        ref[1,3,1,0]=(-0.974221864735)*x_ref[0]+(1.02650607362)*x_ref[1]
        ref[1,3,1,1]=(-0.476728696056)*x_ref[0]+(1.39394954201)*x_ref[1]
        ref[1,3,2,0]=(-0.276890313424)*x_ref[0]+(-0.42434071217)*x_ref[1]
        ref[1,3,2,1]=(-0.45671100973)*x_ref[0]+(-0.271986551291)*x_ref[1]
        ref[1,4,0,0]=(0.415315726206)*x_ref[0]+(0.25174859473)*x_ref[1]
        ref[1,4,0,1]=(0.718116945034)*x_ref[0]+(0.300885379458)*x_ref[1]
        ref[1,4,1,0]=(0.0362286683347)*x_ref[0]+(0.0860658750969)*x_ref[1]
        ref[1,4,1,1]=(0.966944206449)*x_ref[0]+(0.605527604606)*x_ref[1]
        ref[1,4,2,0]=(-0.681446172251)*x_ref[0]+(-0.0201721796333)*x_ref[1]
        ref[1,4,2,1]=(1.16600295744)*x_ref[0]+(1.10261292926)*x_ref[1]
        ref[2,0,0,0]=(-0.239525904251)*x_ref[0]+(0.428475333805)*x_ref[1]
        ref[2,0,0,1]=(0.0817932891372)*x_ref[0]+(0.472895457716)*x_ref[1]
        ref[2,0,1,0]=(0.96571284204)*x_ref[0]+(1.30765537984)*x_ref[1]
        ref[2,0,1,1]=(-0.27274079215)*x_ref[0]+(-0.973090020231)*x_ref[1]
        ref[2,0,2,0]=(-0.586811136701)*x_ref[0]+(0.0282301965253)*x_ref[1]
        ref[2,0,2,1]=(-0.988289117174)*x_ref[0]+(-0.982428801403)*x_ref[1]
        ref[2,1,0,0]=(-0.97313872216)*x_ref[0]+(0.264946309189)*x_ref[1]
        ref[2,1,0,1]=(0.477464114989)*x_ref[0]+(0.735813896712)*x_ref[1]
        ref[2,1,1,0]=(-0.999655360356)*x_ref[0]+(0.0906084300723)*x_ref[1]
        ref[2,1,1,1]=(-0.672480642088)*x_ref[0]+(-0.0511287334859)*x_ref[1]
        ref[2,1,2,0]=(-1.06854561704)*x_ref[0]+(-0.188249587015)*x_ref[1]
        ref[2,1,2,1]=(-0.439030823181)*x_ref[0]+(0.76989999863)*x_ref[1]
        ref[2,2,0,0]=(0.935351178562)*x_ref[0]+(-0.732961189428)*x_ref[1]
        ref[2,2,0,1]=(-0.201449422629)*x_ref[0]+(-0.920429771001)*x_ref[1]
        ref[2,2,1,0]=(-0.0397141758777)*x_ref[0]+(1.23810013859)*x_ref[1]
        ref[2,2,1,1]=(-0.853635033788)*x_ref[0]+(0.982464467568)*x_ref[1]
        ref[2,2,2,0]=(-1.35330887855)*x_ref[0]+(0.200038841813)*x_ref[1]
        ref[2,2,2,1]=(0.0388084224165)*x_ref[0]+(-0.312397495662)*x_ref[1]
        ref[2,3,0,0]=(-0.762894528313)*x_ref[0]+(0.206223517392)*x_ref[1]
        ref[2,3,0,1]=(0.0823646028803)*x_ref[0]+(-1.1834743671)*x_ref[1]
        ref[2,3,1,0]=(0.302380236653)*x_ref[0]+(-0.725627755476)*x_ref[1]
        ref[2,3,1,1]=(-0.12698195797)*x_ref[0]+(-0.289106042968)*x_ref[1]
        ref[2,3,2,0]=(-1.33992252725)*x_ref[0]+(-0.903068424118)*x_ref[1]
        ref[2,3,2,1]=(-0.128913469794)*x_ref[0]+(0.392464023413)*x_ref[1]
        ref[2,4,0,0]=(-0.326550273177)*x_ref[0]+(-0.856389747069)*x_ref[1]
        ref[2,4,0,1]=(-1.14966749976)*x_ref[0]+(0.809029270783)*x_ref[1]
        ref[2,4,1,0]=(1.50037083169)*x_ref[0]+(0.564371263186)*x_ref[1]
        ref[2,4,1,1]=(-0.0660188964857)*x_ref[0]+(-0.871309804802)*x_ref[1]
        ref[2,4,2,0]=(0.527342096457)*x_ref[0]+(1.41520765334)*x_ref[1]
        ref[2,4,2,1]=(-0.177020653581)*x_ref[0]+(-0.218017520886)*x_ref[1]
        ref[3,0,0,0]=(-0.901819829595)*x_ref[0]+(-0.0897360627182)*x_ref[1]
        ref[3,0,0,1]=(-1.84231567451)*x_ref[0]+(-1.57138993977)*x_ref[1]
        ref[3,0,1,0]=(-0.692347444292)*x_ref[0]+(-0.35994981731)*x_ref[1]
        ref[3,0,1,1]=(-1.20666616877)*x_ref[0]+(1.31974198302)*x_ref[1]
        ref[3,0,2,0]=(-1.49613152411)*x_ref[0]+(-1.29087493111)*x_ref[1]
        ref[3,0,2,1]=(0.00302152289294)*x_ref[0]+(0.0596509711717)*x_ref[1]
        ref[3,1,0,0]=(-0.716314986327)*x_ref[0]+(0.540487536594)*x_ref[1]
        ref[3,1,0,1]=(1.26797434361)*x_ref[0]+(1.24611664766)*x_ref[1]
        ref[3,1,1,0]=(1.29133591716)*x_ref[0]+(-0.822278152235)*x_ref[1]
        ref[3,1,1,1]=(0.736506521682)*x_ref[0]+(0.295109058687)*x_ref[1]
        ref[3,1,2,0]=(-1.31999878466)*x_ref[0]+(-1.06268553583)*x_ref[1]
        ref[3,1,2,1]=(-1.02353629588)*x_ref[0]+(-0.180118460659)*x_ref[1]
        ref[3,2,0,0]=(1.2591074793)*x_ref[0]+(0.716164110599)*x_ref[1]
        ref[3,2,0,1]=(0.0737591553738)*x_ref[0]+(-0.180183018938)*x_ref[1]
        ref[3,2,1,0]=(0.842324803092)*x_ref[0]+(-0.179012336351)*x_ref[1]
        ref[3,2,1,1]=(-1.04205716802)*x_ref[0]+(0.91772961597)*x_ref[1]
        ref[3,2,2,0]=(0.774019388104)*x_ref[0]+(0.392202481643)*x_ref[1]
        ref[3,2,2,1]=(0.583625745968)*x_ref[0]+(0.857713356409)*x_ref[1]
        ref[3,3,0,0]=(-1.08653988672)*x_ref[0]+(-1.47741013604)*x_ref[1]
        ref[3,3,0,1]=(-0.529552146407)*x_ref[0]+(-0.658937785849)*x_ref[1]
        ref[3,3,1,0]=(1.8439291246)*x_ref[0]+(-1.36075468038)*x_ref[1]
        ref[3,3,1,1]=(-0.656328225025)*x_ref[0]+(1.18200793838)*x_ref[1]
        ref[3,3,2,0]=(-1.57290668837)*x_ref[0]+(0.527467780284)*x_ref[1]
        ref[3,3,2,1]=(1.44984969309)*x_ref[0]+(-1.37301308873)*x_ref[1]
        ref[3,4,0,0]=(-0.225012249574)*x_ref[0]+(-1.80387444921)*x_ref[1]
        ref[3,4,0,1]=(-1.27157572533)*x_ref[0]+(0.568859005411)*x_ref[1]
        ref[3,4,1,0]=(-0.381626896774)*x_ref[0]+(-0.335697445475)*x_ref[1]
        ref[3,4,1,1]=(0.555084489073)*x_ref[0]+(-0.659578343093)*x_ref[1]
        ref[3,4,2,0]=(0.784894297286)*x_ref[0]+(0.955314078432)*x_ref[1]
        ref[3,4,2,1]=(0.33356366872)*x_ref[0]+(1.38175924325)*x_ref[1]
      else:
        s[0,0,0,0]=(0.598866527602)*x[0]+(0.44051361201)*x[1]+(-0.131423809939)*x[2]
        s[0,0,0,1]=(1.25181618674)*x[0]+(1.52822376488)*x[1]+(1.30574312696)*x[2]
        s[0,0,1,0]=(-0.331709443946)*x[0]+(-1.34705365817)*x[1]+(0.726301465323)*x[2]
        s[0,0,1,1]=(-0.577398988618)*x[0]+(-1.22594248786)*x[1]+(-0.961458379201)*x[2]
        s[0,0,2,0]=(-0.451500880848)*x[0]+(-0.774270726492)*x[1]+(-1.2165638211)*x[2]
        s[0,0,2,1]=(-1.10567205671)*x[0]+(-0.161917736574)*x[1]+(1.37957442359)*x[2]
        s[0,1,0,0]=(0.745737514206)*x[0]+(-0.796248435669)*x[1]+(-1.63816932586)*x[2]
        s[0,1,0,1]=(0.0831105458366)*x[0]+(-0.136660461565)*x[1]+(-1.56685434932)*x[2]
        s[0,1,1,0]=(-1.44351617889)*x[0]+(1.45409315626)*x[1]+(0.442708463336)*x[2]
        s[0,1,1,1]=(-0.558066341802)*x[0]+(1.19630033662)*x[1]+(0.186071415659)*x[2]
        s[0,1,2,0]=(-0.374869442605)*x[0]+(0.648318396628)*x[1]+(0.0555371118519)*x[2]
        s[0,1,2,1]=(-0.36160618163)*x[0]+(0.437013254054)*x[1]+(-0.638236225808)*x[2]
        s[0,2,0,0]=(0.189302763729)*x[0]+(-0.255111592053)*x[1]+(0.682670294776)*x[2]
        s[0,2,0,1]=(0.025995815858)*x[0]+(0.34462072647)*x[1]+(0.341995413585)*x[2]
        s[0,2,1,0]=(-1.28972101272)*x[0]+(1.01784177655)*x[1]+(1.11775675824)*x[2]
        s[0,2,1,1]=(0.557076830097)*x[0]+(0.838542238067)*x[1]+(-0.155585369573)*x[2]
        s[0,2,2,0]=(-0.0141241838818)*x[0]+(-0.191831242085)*x[1]+(-0.847287909515)*x[2]
        s[0,2,2,1]=(0.429778841571)*x[0]+(0.376566699487)*x[1]+(0.148313161917)*x[2]
        s[0,3,0,0]=(0.233606469427)*x[0]+(0.266803505307)*x[1]+(-0.205218686924)*x[2]
        s[0,3,0,1]=(-0.124511680509)*x[0]+(1.02896325334)*x[1]+(0.935771094012)*x[2]
        s[0,3,1,0]=(0.976537956221)*x[0]+(-0.345122171465)*x[1]+(0.777459950315)*x[2]
        s[0,3,1,1]=(-0.263691953956)*x[0]+(0.287964213001)*x[1]+(-0.237113357698)*x[2]
        s[0,3,2,0]=(-0.14155225233)*x[0]+(0.122587298994)*x[1]+(-0.0264137141467)*x[2]
        s[0,3,2,1]=(-0.760118620618)*x[0]+(1.57384021648)*x[1]+(1.32809481134)*x[2]
        s[0,4,0,0]=(0.673626831345)*x[0]+(-0.454271151216)*x[1]+(-0.863944524769)*x[2]
        s[0,4,0,1]=(-0.630430988823)*x[0]+(0.45809212038)*x[1]+(-1.27010861122)*x[2]
        s[0,4,1,0]=(-0.171563094037)*x[0]+(-0.360716542419)*x[1]+(0.162811650708)*x[2]
        s[0,4,1,1]=(0.890628840767)*x[0]+(-0.33837153333)*x[1]+(-1.57253750851)*x[2]
        s[0,4,2,0]=(-0.0358372393822)*x[0]+(1.12988684611)*x[1]+(0.153713271682)*x[2]
        s[0,4,2,1]=(-0.24590925839)*x[0]+(0.616341690491)*x[1]+(-1.12513207308)*x[2]
        s[1,0,0,0]=(1.17052294387)*x[0]+(-0.176023586935)*x[1]+(-0.413320632417)*x[2]
        s[1,0,0,1]=(1.2820208139)*x[0]+(0.228009714737)*x[1]+(-0.149423928125)*x[2]
        s[1,0,1,0]=(-0.0312869788993)*x[0]+(0.526820009564)*x[1]+(0.0903543587276)*x[2]
        s[1,0,1,1]=(-1.21319041974)*x[0]+(-0.428253988159)*x[1]+(0.464434175335)*x[2]
        s[1,0,2,0]=(0.9190018302)*x[0]+(0.449296159226)*x[1]+(-0.481445053339)*x[2]
        s[1,0,2,1]=(-0.616752258293)*x[0]+(0.359850140048)*x[1]+(-1.27542462984)*x[2]
        s[1,1,0,0]=(1.35404696733)*x[0]+(1.25468169576)*x[1]+(-0.410473495612)*x[2]
        s[1,1,0,1]=(0.320849051626)*x[0]+(1.56909917585)*x[1]+(1.00666256572)*x[2]
        s[1,1,1,0]=(0.235361535902)*x[0]+(-0.826787974769)*x[1]+(1.07359804999)*x[2]
        s[1,1,1,1]=(-0.334284708722)*x[0]+(0.316786292756)*x[1]+(-0.886350336088)*x[2]
        s[1,1,2,0]=(-0.890644854382)*x[0]+(0.509273281135)*x[1]+(-1.19292772513)*x[2]
        s[1,1,2,1]=(0.252216653779)*x[0]+(1.57054197023)*x[1]+(-1.90744671962)*x[2]
        s[1,2,0,0]=(0.551491892715)*x[0]+(-0.222655049798)*x[1]+(0.398387383719)*x[2]
        s[1,2,0,1]=(-1.55041598041)*x[0]+(0.994016310342)*x[1]+(-0.474205618301)*x[2]
        s[1,2,1,0]=(0.268662020467)*x[0]+(-0.755250236528)*x[1]+(-0.540383977121)*x[2]
        s[1,2,1,1]=(-0.551512759298)*x[0]+(1.4277999148)*x[1]+(-1.58961615319)*x[2]
        s[1,2,2,0]=(0.304822528983)*x[0]+(0.232100103775)*x[1]+(0.463883691163)*x[2]
        s[1,2,2,1]=(0.561644600923)*x[0]+(-0.359426536495)*x[1]+(-1.34828152802)*x[2]
        s[1,3,0,0]=(0.850248623505)*x[0]+(0.116021948452)*x[1]+(0.142044197471)*x[2]
        s[1,3,0,1]=(0.724145173317)*x[0]+(-0.792974540874)*x[1]+(-0.195454750691)*x[2]
        s[1,3,1,0]=(-0.110714388555)*x[0]+(-1.40121178374)*x[1]+(0.0347355958316)*x[2]
        s[1,3,1,1]=(-1.12980968478)*x[0]+(-0.570714541482)*x[1]+(-1.2280735041)*x[2]
        s[1,3,2,0]=(-1.21834172715)*x[0]+(1.83863423781)*x[1]+(0.442239363821)*x[2]
        s[1,3,2,1]=(0.426294021669)*x[0]+(1.63350652457)*x[1]+(-0.580391188165)*x[2]
        s[1,4,0,0]=(-0.791296449419)*x[0]+(0.395345927547)*x[1]+(-0.575778346002)*x[2]
        s[1,4,0,1]=(1.05262027172)*x[0]+(-1.1912135275)*x[1]+(-0.281282812313)*x[2]
        s[1,4,1,0]=(-0.229073653507)*x[0]+(0.333793277005)*x[1]+(-0.540787491142)*x[2]
        s[1,4,1,1]=(-0.782297770239)*x[0]+(0.965050908578)*x[1]+(0.648337674864)*x[2]
        s[1,4,2,0]=(-0.425598399058)*x[0]+(1.24755177094)*x[1]+(-0.599571710505)*x[2]
        s[1,4,2,1]=(-0.584529028943)*x[0]+(-0.115930785496)*x[1]+(-0.412776678144)*x[2]
        s[2,0,0,0]=(-1.22220614711)*x[0]+(0.991280082493)*x[1]+(-0.771530362129)*x[2]
        s[2,0,0,1]=(-0.0408833973017)*x[0]+(-0.878550136082)*x[1]+(-1.27973588501)*x[2]
        s[2,0,1,0]=(0.380052050339)*x[0]+(1.34070050501)*x[1]+(-1.39080485735)*x[2]
        s[2,0,1,1]=(-0.725808759614)*x[0]+(1.37568148722)*x[1]+(-0.369008444335)*x[2]
        s[2,0,2,0]=(-0.563453230072)*x[0]+(0.25114466425)*x[1]+(-0.0602643663972)*x[2]
        s[2,0,2,1]=(-0.802381432423)*x[0]+(0.22715480998)*x[1]+(0.512734770933)*x[2]
        s[2,1,0,0]=(-0.522742058617)*x[0]+(0.37241878501)*x[1]+(0.691287986533)*x[2]
        s[2,1,0,1]=(-0.168275058841)*x[0]+(-0.764400123274)*x[1]+(-0.681242773507)*x[2]
        s[2,1,1,0]=(0.551464415842)*x[0]+(1.03330737276)*x[1]+(0.958549917295)*x[2]
        s[2,1,1,1]=(-0.909196447141)*x[0]+(0.145246231629)*x[1]+(-1.05582073367)*x[2]
        s[2,1,2,0]=(-0.638872672519)*x[0]+(0.406342003401)*x[1]+(-0.393150454497)*x[2]
        s[2,1,2,1]=(0.366774488001)*x[0]+(0.365786897102)*x[1]+(0.441878454676)*x[2]
        s[2,2,0,0]=(-0.261857371972)*x[0]+(0.0318314524134)*x[1]+(-0.101372223186)*x[2]
        s[2,2,0,1]=(-0.135781961265)*x[0]+(-0.00345171998114)*x[1]+(0.0101907167492)*x[2]
        s[2,2,1,0]=(-1.37428228456)*x[0]+(-0.524953075499)*x[1]+(-0.372440345968)*x[2]
        s[2,2,1,1]=(0.0540542424565)*x[0]+(-1.03413875822)*x[1]+(-0.899141477767)*x[2]
        s[2,2,2,0]=(-0.431640543873)*x[0]+(0.934769305572)*x[1]+(-0.848404080116)*x[2]
        s[2,2,2,1]=(0.984445227688)*x[0]+(-0.207064309031)*x[1]+(-1.75497257678)*x[2]
        s[2,3,0,0]=(-1.61836350351)*x[0]+(0.885389572019)*x[1]+(0.133218748941)*x[2]
        s[2,3,0,1]=(0.100716060215)*x[0]+(-0.713421128342)*x[1]+(1.16765406123)*x[2]
        s[2,3,1,0]=(0.497272818223)*x[0]+(-0.698345898985)*x[1]+(-0.643672453237)*x[2]
        s[2,3,1,1]=(-0.680652178847)*x[0]+(-0.56686153354)*x[1]+(-0.259410997801)*x[2]
        s[2,3,2,0]=(1.08214823373)*x[0]+(-0.875408253309)*x[1]+(-0.165330678277)*x[2]
        s[2,3,2,1]=(-0.251850346585)*x[0]+(-0.233254346449)*x[1]+(0.348350958039)*x[2]
        s[2,4,0,0]=(0.283802782063)*x[0]+(0.338243660721)*x[1]+(0.0400948594919)*x[2]
        s[2,4,0,1]=(-1.46364736479)*x[0]+(0.738031226032)*x[1]+(0.897193522713)*x[2]
        s[2,4,1,0]=(1.17637134366)*x[0]+(-0.607880128773)*x[1]+(-1.00344218409)*x[2]
        s[2,4,1,1]=(0.716275197822)*x[0]+(1.21991657141)*x[1]+(-0.0467328809597)*x[2]
        s[2,4,2,0]=(0.380704185608)*x[0]+(0.603603708529)*x[1]+(-0.820752063279)*x[2]
        s[2,4,2,1]=(1.39582555808)*x[0]+(-0.473415991144)*x[1]+(1.24135585822)*x[2]
        s[3,0,0,0]=(1.39071291597)*x[0]+(-1.59041985271)*x[1]+(0.60185319823)*x[2]
        s[3,0,0,1]=(-1.46801716969)*x[0]+(-0.552777641699)*x[1]+(1.23129633317)*x[2]
        s[3,0,1,0]=(1.64196247631)*x[0]+(-0.343527059014)*x[1]+(1.11486568074)*x[2]
        s[3,0,1,1]=(-1.57106459992)*x[0]+(1.18660381744)*x[1]+(0.495305193784)*x[2]
        s[3,0,2,0]=(-0.246662602621)*x[0]+(0.980536147604)*x[1]+(-1.29582037379)*x[2]
        s[3,0,2,1]=(0.351277952246)*x[0]+(0.0189052721516)*x[1]+(-0.109275345669)*x[2]
        s[3,1,0,0]=(-0.559164628236)*x[0]+(-0.390894772538)*x[1]+(0.657026552721)*x[2]
        s[3,1,0,1]=(-1.63523473278)*x[0]+(-0.958777701493)*x[1]+(0.0466950147837)*x[2]
        s[3,1,1,0]=(-0.913387269773)*x[0]+(-0.752264610272)*x[1]+(-1.00623906741)*x[2]
        s[3,1,1,1]=(0.703502728627)*x[0]+(0.985524845396)*x[1]+(-0.0966526090953)*x[2]
        s[3,1,2,0]=(-0.993367844074)*x[0]+(-0.205401158091)*x[1]+(0.416042784765)*x[2]
        s[3,1,2,1]=(0.299615272051)*x[0]+(1.32572777489)*x[1]+(0.338960058542)*x[2]
        s[3,2,0,0]=(-0.757820100577)*x[0]+(0.644913131328)*x[1]+(-0.660404804486)*x[2]
        s[3,2,0,1]=(-1.75641089093)*x[0]+(-0.366399254245)*x[1]+(-0.646900183385)*x[2]
        s[3,2,1,0]=(0.0912144040068)*x[0]+(-0.857875163959)*x[1]+(0.350343357173)*x[2]
        s[3,2,1,1]=(-1.80008916515)*x[0]+(-1.83308297974)*x[1]+(0.052958998062)*x[2]
        s[3,2,2,0]=(1.44092323696)*x[0]+(-1.1831374133)*x[1]+(-0.0323647992786)*x[2]
        s[3,2,2,1]=(-0.353641143336)*x[0]+(0.774344611259)*x[1]+(-0.842874663076)*x[2]
        s[3,3,0,0]=(-1.17418376796)*x[0]+(0.334599071712)*x[1]+(0.0621712582176)*x[2]
        s[3,3,0,1]=(1.1539137936)*x[0]+(0.781617539757)*x[1]+(1.6289484244)*x[2]
        s[3,3,1,0]=(1.6860969053)*x[0]+(-0.936239288854)*x[1]+(1.7596463612)*x[2]
        s[3,3,1,1]=(-1.11111948999)*x[0]+(0.285869758308)*x[1]+(1.69003878827)*x[2]
        s[3,3,2,0]=(-1.00166505759)*x[0]+(-1.21067712793)*x[1]+(0.422711238131)*x[2]
        s[3,3,2,1]=(0.49093698547)*x[0]+(-0.952249146613)*x[1]+(-0.955917587594)*x[2]
        s[3,4,0,0]=(-0.511109469881)*x[0]+(0.630807398954)*x[1]+(-0.0365100439309)*x[2]
        s[3,4,0,1]=(0.260323499449)*x[0]+(-0.232312235272)*x[1]+(-0.676872879397)*x[2]
        s[3,4,1,0]=(-0.327180688658)*x[0]+(0.0260172986398)*x[1]+(-1.59267263743)*x[2]
        s[3,4,1,1]=(0.135287590014)*x[0]+(1.69850923514)*x[1]+(-0.498682843917)*x[2]
        s[3,4,2,0]=(0.314853952103)*x[0]+(-0.562331392142)*x[1]+(0.726740096722)*x[2]
        s[3,4,2,1]=(-0.638129182205)*x[0]+(-0.881463492466)*x[1]+(-0.792625664675)*x[2]
        ref[0,0,0,0]=(0.598866527602)*x_ref[0]+(0.44051361201)*x_ref[1]+(-0.131423809939)*x_ref[2]
        ref[0,0,0,1]=(1.25181618674)*x_ref[0]+(1.52822376488)*x_ref[1]+(1.30574312696)*x_ref[2]
        ref[0,0,1,0]=(-0.331709443946)*x_ref[0]+(-1.34705365817)*x_ref[1]+(0.726301465323)*x_ref[2]
        ref[0,0,1,1]=(-0.577398988618)*x_ref[0]+(-1.22594248786)*x_ref[1]+(-0.961458379201)*x_ref[2]
        ref[0,0,2,0]=(-0.451500880848)*x_ref[0]+(-0.774270726492)*x_ref[1]+(-1.2165638211)*x_ref[2]
        ref[0,0,2,1]=(-1.10567205671)*x_ref[0]+(-0.161917736574)*x_ref[1]+(1.37957442359)*x_ref[2]
        ref[0,1,0,0]=(0.745737514206)*x_ref[0]+(-0.796248435669)*x_ref[1]+(-1.63816932586)*x_ref[2]
        ref[0,1,0,1]=(0.0831105458366)*x_ref[0]+(-0.136660461565)*x_ref[1]+(-1.56685434932)*x_ref[2]
        ref[0,1,1,0]=(-1.44351617889)*x_ref[0]+(1.45409315626)*x_ref[1]+(0.442708463336)*x_ref[2]
        ref[0,1,1,1]=(-0.558066341802)*x_ref[0]+(1.19630033662)*x_ref[1]+(0.186071415659)*x_ref[2]
        ref[0,1,2,0]=(-0.374869442605)*x_ref[0]+(0.648318396628)*x_ref[1]+(0.0555371118519)*x_ref[2]
        ref[0,1,2,1]=(-0.36160618163)*x_ref[0]+(0.437013254054)*x_ref[1]+(-0.638236225808)*x_ref[2]
        ref[0,2,0,0]=(0.189302763729)*x_ref[0]+(-0.255111592053)*x_ref[1]+(0.682670294776)*x_ref[2]
        ref[0,2,0,1]=(0.025995815858)*x_ref[0]+(0.34462072647)*x_ref[1]+(0.341995413585)*x_ref[2]
        ref[0,2,1,0]=(-1.28972101272)*x_ref[0]+(1.01784177655)*x_ref[1]+(1.11775675824)*x_ref[2]
        ref[0,2,1,1]=(0.557076830097)*x_ref[0]+(0.838542238067)*x_ref[1]+(-0.155585369573)*x_ref[2]
        ref[0,2,2,0]=(-0.0141241838818)*x_ref[0]+(-0.191831242085)*x_ref[1]+(-0.847287909515)*x_ref[2]
        ref[0,2,2,1]=(0.429778841571)*x_ref[0]+(0.376566699487)*x_ref[1]+(0.148313161917)*x_ref[2]
        ref[0,3,0,0]=(0.233606469427)*x_ref[0]+(0.266803505307)*x_ref[1]+(-0.205218686924)*x_ref[2]
        ref[0,3,0,1]=(-0.124511680509)*x_ref[0]+(1.02896325334)*x_ref[1]+(0.935771094012)*x_ref[2]
        ref[0,3,1,0]=(0.976537956221)*x_ref[0]+(-0.345122171465)*x_ref[1]+(0.777459950315)*x_ref[2]
        ref[0,3,1,1]=(-0.263691953956)*x_ref[0]+(0.287964213001)*x_ref[1]+(-0.237113357698)*x_ref[2]
        ref[0,3,2,0]=(-0.14155225233)*x_ref[0]+(0.122587298994)*x_ref[1]+(-0.0264137141467)*x_ref[2]
        ref[0,3,2,1]=(-0.760118620618)*x_ref[0]+(1.57384021648)*x_ref[1]+(1.32809481134)*x_ref[2]
        ref[0,4,0,0]=(0.673626831345)*x_ref[0]+(-0.454271151216)*x_ref[1]+(-0.863944524769)*x_ref[2]
        ref[0,4,0,1]=(-0.630430988823)*x_ref[0]+(0.45809212038)*x_ref[1]+(-1.27010861122)*x_ref[2]
        ref[0,4,1,0]=(-0.171563094037)*x_ref[0]+(-0.360716542419)*x_ref[1]+(0.162811650708)*x_ref[2]
        ref[0,4,1,1]=(0.890628840767)*x_ref[0]+(-0.33837153333)*x_ref[1]+(-1.57253750851)*x_ref[2]
        ref[0,4,2,0]=(-0.0358372393822)*x_ref[0]+(1.12988684611)*x_ref[1]+(0.153713271682)*x_ref[2]
        ref[0,4,2,1]=(-0.24590925839)*x_ref[0]+(0.616341690491)*x_ref[1]+(-1.12513207308)*x_ref[2]
        ref[1,0,0,0]=(1.17052294387)*x_ref[0]+(-0.176023586935)*x_ref[1]+(-0.413320632417)*x_ref[2]
        ref[1,0,0,1]=(1.2820208139)*x_ref[0]+(0.228009714737)*x_ref[1]+(-0.149423928125)*x_ref[2]
        ref[1,0,1,0]=(-0.0312869788993)*x_ref[0]+(0.526820009564)*x_ref[1]+(0.0903543587276)*x_ref[2]
        ref[1,0,1,1]=(-1.21319041974)*x_ref[0]+(-0.428253988159)*x_ref[1]+(0.464434175335)*x_ref[2]
        ref[1,0,2,0]=(0.9190018302)*x_ref[0]+(0.449296159226)*x_ref[1]+(-0.481445053339)*x_ref[2]
        ref[1,0,2,1]=(-0.616752258293)*x_ref[0]+(0.359850140048)*x_ref[1]+(-1.27542462984)*x_ref[2]
        ref[1,1,0,0]=(1.35404696733)*x_ref[0]+(1.25468169576)*x_ref[1]+(-0.410473495612)*x_ref[2]
        ref[1,1,0,1]=(0.320849051626)*x_ref[0]+(1.56909917585)*x_ref[1]+(1.00666256572)*x_ref[2]
        ref[1,1,1,0]=(0.235361535902)*x_ref[0]+(-0.826787974769)*x_ref[1]+(1.07359804999)*x_ref[2]
        ref[1,1,1,1]=(-0.334284708722)*x_ref[0]+(0.316786292756)*x_ref[1]+(-0.886350336088)*x_ref[2]
        ref[1,1,2,0]=(-0.890644854382)*x_ref[0]+(0.509273281135)*x_ref[1]+(-1.19292772513)*x_ref[2]
        ref[1,1,2,1]=(0.252216653779)*x_ref[0]+(1.57054197023)*x_ref[1]+(-1.90744671962)*x_ref[2]
        ref[1,2,0,0]=(0.551491892715)*x_ref[0]+(-0.222655049798)*x_ref[1]+(0.398387383719)*x_ref[2]
        ref[1,2,0,1]=(-1.55041598041)*x_ref[0]+(0.994016310342)*x_ref[1]+(-0.474205618301)*x_ref[2]
        ref[1,2,1,0]=(0.268662020467)*x_ref[0]+(-0.755250236528)*x_ref[1]+(-0.540383977121)*x_ref[2]
        ref[1,2,1,1]=(-0.551512759298)*x_ref[0]+(1.4277999148)*x_ref[1]+(-1.58961615319)*x_ref[2]
        ref[1,2,2,0]=(0.304822528983)*x_ref[0]+(0.232100103775)*x_ref[1]+(0.463883691163)*x_ref[2]
        ref[1,2,2,1]=(0.561644600923)*x_ref[0]+(-0.359426536495)*x_ref[1]+(-1.34828152802)*x_ref[2]
        ref[1,3,0,0]=(0.850248623505)*x_ref[0]+(0.116021948452)*x_ref[1]+(0.142044197471)*x_ref[2]
        ref[1,3,0,1]=(0.724145173317)*x_ref[0]+(-0.792974540874)*x_ref[1]+(-0.195454750691)*x_ref[2]
        ref[1,3,1,0]=(-0.110714388555)*x_ref[0]+(-1.40121178374)*x_ref[1]+(0.0347355958316)*x_ref[2]
        ref[1,3,1,1]=(-1.12980968478)*x_ref[0]+(-0.570714541482)*x_ref[1]+(-1.2280735041)*x_ref[2]
        ref[1,3,2,0]=(-1.21834172715)*x_ref[0]+(1.83863423781)*x_ref[1]+(0.442239363821)*x_ref[2]
        ref[1,3,2,1]=(0.426294021669)*x_ref[0]+(1.63350652457)*x_ref[1]+(-0.580391188165)*x_ref[2]
        ref[1,4,0,0]=(-0.791296449419)*x_ref[0]+(0.395345927547)*x_ref[1]+(-0.575778346002)*x_ref[2]
        ref[1,4,0,1]=(1.05262027172)*x_ref[0]+(-1.1912135275)*x_ref[1]+(-0.281282812313)*x_ref[2]
        ref[1,4,1,0]=(-0.229073653507)*x_ref[0]+(0.333793277005)*x_ref[1]+(-0.540787491142)*x_ref[2]
        ref[1,4,1,1]=(-0.782297770239)*x_ref[0]+(0.965050908578)*x_ref[1]+(0.648337674864)*x_ref[2]
        ref[1,4,2,0]=(-0.425598399058)*x_ref[0]+(1.24755177094)*x_ref[1]+(-0.599571710505)*x_ref[2]
        ref[1,4,2,1]=(-0.584529028943)*x_ref[0]+(-0.115930785496)*x_ref[1]+(-0.412776678144)*x_ref[2]
        ref[2,0,0,0]=(-1.22220614711)*x_ref[0]+(0.991280082493)*x_ref[1]+(-0.771530362129)*x_ref[2]
        ref[2,0,0,1]=(-0.0408833973017)*x_ref[0]+(-0.878550136082)*x_ref[1]+(-1.27973588501)*x_ref[2]
        ref[2,0,1,0]=(0.380052050339)*x_ref[0]+(1.34070050501)*x_ref[1]+(-1.39080485735)*x_ref[2]
        ref[2,0,1,1]=(-0.725808759614)*x_ref[0]+(1.37568148722)*x_ref[1]+(-0.369008444335)*x_ref[2]
        ref[2,0,2,0]=(-0.563453230072)*x_ref[0]+(0.25114466425)*x_ref[1]+(-0.0602643663972)*x_ref[2]
        ref[2,0,2,1]=(-0.802381432423)*x_ref[0]+(0.22715480998)*x_ref[1]+(0.512734770933)*x_ref[2]
        ref[2,1,0,0]=(-0.522742058617)*x_ref[0]+(0.37241878501)*x_ref[1]+(0.691287986533)*x_ref[2]
        ref[2,1,0,1]=(-0.168275058841)*x_ref[0]+(-0.764400123274)*x_ref[1]+(-0.681242773507)*x_ref[2]
        ref[2,1,1,0]=(0.551464415842)*x_ref[0]+(1.03330737276)*x_ref[1]+(0.958549917295)*x_ref[2]
        ref[2,1,1,1]=(-0.909196447141)*x_ref[0]+(0.145246231629)*x_ref[1]+(-1.05582073367)*x_ref[2]
        ref[2,1,2,0]=(-0.638872672519)*x_ref[0]+(0.406342003401)*x_ref[1]+(-0.393150454497)*x_ref[2]
        ref[2,1,2,1]=(0.366774488001)*x_ref[0]+(0.365786897102)*x_ref[1]+(0.441878454676)*x_ref[2]
        ref[2,2,0,0]=(-0.261857371972)*x_ref[0]+(0.0318314524134)*x_ref[1]+(-0.101372223186)*x_ref[2]
        ref[2,2,0,1]=(-0.135781961265)*x_ref[0]+(-0.00345171998114)*x_ref[1]+(0.0101907167492)*x_ref[2]
        ref[2,2,1,0]=(-1.37428228456)*x_ref[0]+(-0.524953075499)*x_ref[1]+(-0.372440345968)*x_ref[2]
        ref[2,2,1,1]=(0.0540542424565)*x_ref[0]+(-1.03413875822)*x_ref[1]+(-0.899141477767)*x_ref[2]
        ref[2,2,2,0]=(-0.431640543873)*x_ref[0]+(0.934769305572)*x_ref[1]+(-0.848404080116)*x_ref[2]
        ref[2,2,2,1]=(0.984445227688)*x_ref[0]+(-0.207064309031)*x_ref[1]+(-1.75497257678)*x_ref[2]
        ref[2,3,0,0]=(-1.61836350351)*x_ref[0]+(0.885389572019)*x_ref[1]+(0.133218748941)*x_ref[2]
        ref[2,3,0,1]=(0.100716060215)*x_ref[0]+(-0.713421128342)*x_ref[1]+(1.16765406123)*x_ref[2]
        ref[2,3,1,0]=(0.497272818223)*x_ref[0]+(-0.698345898985)*x_ref[1]+(-0.643672453237)*x_ref[2]
        ref[2,3,1,1]=(-0.680652178847)*x_ref[0]+(-0.56686153354)*x_ref[1]+(-0.259410997801)*x_ref[2]
        ref[2,3,2,0]=(1.08214823373)*x_ref[0]+(-0.875408253309)*x_ref[1]+(-0.165330678277)*x_ref[2]
        ref[2,3,2,1]=(-0.251850346585)*x_ref[0]+(-0.233254346449)*x_ref[1]+(0.348350958039)*x_ref[2]
        ref[2,4,0,0]=(0.283802782063)*x_ref[0]+(0.338243660721)*x_ref[1]+(0.0400948594919)*x_ref[2]
        ref[2,4,0,1]=(-1.46364736479)*x_ref[0]+(0.738031226032)*x_ref[1]+(0.897193522713)*x_ref[2]
        ref[2,4,1,0]=(1.17637134366)*x_ref[0]+(-0.607880128773)*x_ref[1]+(-1.00344218409)*x_ref[2]
        ref[2,4,1,1]=(0.716275197822)*x_ref[0]+(1.21991657141)*x_ref[1]+(-0.0467328809597)*x_ref[2]
        ref[2,4,2,0]=(0.380704185608)*x_ref[0]+(0.603603708529)*x_ref[1]+(-0.820752063279)*x_ref[2]
        ref[2,4,2,1]=(1.39582555808)*x_ref[0]+(-0.473415991144)*x_ref[1]+(1.24135585822)*x_ref[2]
        ref[3,0,0,0]=(1.39071291597)*x_ref[0]+(-1.59041985271)*x_ref[1]+(0.60185319823)*x_ref[2]
        ref[3,0,0,1]=(-1.46801716969)*x_ref[0]+(-0.552777641699)*x_ref[1]+(1.23129633317)*x_ref[2]
        ref[3,0,1,0]=(1.64196247631)*x_ref[0]+(-0.343527059014)*x_ref[1]+(1.11486568074)*x_ref[2]
        ref[3,0,1,1]=(-1.57106459992)*x_ref[0]+(1.18660381744)*x_ref[1]+(0.495305193784)*x_ref[2]
        ref[3,0,2,0]=(-0.246662602621)*x_ref[0]+(0.980536147604)*x_ref[1]+(-1.29582037379)*x_ref[2]
        ref[3,0,2,1]=(0.351277952246)*x_ref[0]+(0.0189052721516)*x_ref[1]+(-0.109275345669)*x_ref[2]
        ref[3,1,0,0]=(-0.559164628236)*x_ref[0]+(-0.390894772538)*x_ref[1]+(0.657026552721)*x_ref[2]
        ref[3,1,0,1]=(-1.63523473278)*x_ref[0]+(-0.958777701493)*x_ref[1]+(0.0466950147837)*x_ref[2]
        ref[3,1,1,0]=(-0.913387269773)*x_ref[0]+(-0.752264610272)*x_ref[1]+(-1.00623906741)*x_ref[2]
        ref[3,1,1,1]=(0.703502728627)*x_ref[0]+(0.985524845396)*x_ref[1]+(-0.0966526090953)*x_ref[2]
        ref[3,1,2,0]=(-0.993367844074)*x_ref[0]+(-0.205401158091)*x_ref[1]+(0.416042784765)*x_ref[2]
        ref[3,1,2,1]=(0.299615272051)*x_ref[0]+(1.32572777489)*x_ref[1]+(0.338960058542)*x_ref[2]
        ref[3,2,0,0]=(-0.757820100577)*x_ref[0]+(0.644913131328)*x_ref[1]+(-0.660404804486)*x_ref[2]
        ref[3,2,0,1]=(-1.75641089093)*x_ref[0]+(-0.366399254245)*x_ref[1]+(-0.646900183385)*x_ref[2]
        ref[3,2,1,0]=(0.0912144040068)*x_ref[0]+(-0.857875163959)*x_ref[1]+(0.350343357173)*x_ref[2]
        ref[3,2,1,1]=(-1.80008916515)*x_ref[0]+(-1.83308297974)*x_ref[1]+(0.052958998062)*x_ref[2]
        ref[3,2,2,0]=(1.44092323696)*x_ref[0]+(-1.1831374133)*x_ref[1]+(-0.0323647992786)*x_ref[2]
        ref[3,2,2,1]=(-0.353641143336)*x_ref[0]+(0.774344611259)*x_ref[1]+(-0.842874663076)*x_ref[2]
        ref[3,3,0,0]=(-1.17418376796)*x_ref[0]+(0.334599071712)*x_ref[1]+(0.0621712582176)*x_ref[2]
        ref[3,3,0,1]=(1.1539137936)*x_ref[0]+(0.781617539757)*x_ref[1]+(1.6289484244)*x_ref[2]
        ref[3,3,1,0]=(1.6860969053)*x_ref[0]+(-0.936239288854)*x_ref[1]+(1.7596463612)*x_ref[2]
        ref[3,3,1,1]=(-1.11111948999)*x_ref[0]+(0.285869758308)*x_ref[1]+(1.69003878827)*x_ref[2]
        ref[3,3,2,0]=(-1.00166505759)*x_ref[0]+(-1.21067712793)*x_ref[1]+(0.422711238131)*x_ref[2]
        ref[3,3,2,1]=(0.49093698547)*x_ref[0]+(-0.952249146613)*x_ref[1]+(-0.955917587594)*x_ref[2]
        ref[3,4,0,0]=(-0.511109469881)*x_ref[0]+(0.630807398954)*x_ref[1]+(-0.0365100439309)*x_ref[2]
        ref[3,4,0,1]=(0.260323499449)*x_ref[0]+(-0.232312235272)*x_ref[1]+(-0.676872879397)*x_ref[2]
        ref[3,4,1,0]=(-0.327180688658)*x_ref[0]+(0.0260172986398)*x_ref[1]+(-1.59267263743)*x_ref[2]
        ref[3,4,1,1]=(0.135287590014)*x_ref[0]+(1.69850923514)*x_ref[1]+(-0.498682843917)*x_ref[2]
        ref[3,4,2,0]=(0.314853952103)*x_ref[0]+(-0.562331392142)*x_ref[1]+(0.726740096722)*x_ref[2]
        ref[3,4,2,1]=(-0.638129182205)*x_ref[0]+(-0.881463492466)*x_ref[1]+(-0.792625664675)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_Solution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(0.648487700909)*x[0]+(-1.43364306566)*x[1]
        ref=(0.648487700909)*x_ref[0]+(-1.43364306566)*x_ref[1]
      else:
        s=(0.0554263047837)*x[0]+(0.627379593937)*x[1]+(0.740861554164)*x[2]
        ref=(0.0554263047837)*x_ref[0]+(0.627379593937)*x_ref[1]+(0.740861554164)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_Solution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-1.61406386148)*x[0]+(0.644766840626)*x[1]
        s[1]=(-0.328446322604)*x[0]+(-0.277890186177)*x[1]
        ref[0]=(-1.61406386148)*x[0]+(0.644766840626)*x[1]
        ref[1]=(-0.328446322604)*x[0]+(-0.277890186177)*x[1]
      else:
        s[0]=(0.280680714031)*x[0]+(0.240612588198)*x[1]+(-0.289142363413)*x[2]
        s[1]=(-0.845801139442)*x[0]+(-1.55820065311)*x[1]+(0.103375183543)*x[2]
        ref[0]=(0.280680714031)*x[0]+(0.240612588198)*x[1]+(-0.289142363413)*x[2]
        ref[1]=(-0.845801139442)*x[0]+(-1.55820065311)*x[1]+(0.103375183543)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_Solution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(0.0481111909695)*x[0]+(-0.393526010503)*x[1]
        s[0,1]=(-0.284097772097)*x[0]+(0.618264565258)*x[1]
        s[0,2]=(0.401575436099)*x[0]+(-0.697718912762)*x[1]
        s[0,3]=(-1.49798792514)*x[0]+(-0.0425176867511)*x[1]
        s[0,4]=(0.213840882194)*x[0]+(-0.914489018365)*x[1]
        s[1,0]=(0.857166519205)*x[0]+(1.5282054335)*x[1]
        s[1,1]=(1.61091698609)*x[0]+(-0.548453984777)*x[1]
        s[1,2]=(-1.56266453199)*x[0]+(-1.66187429416)*x[1]
        s[1,3]=(0.343955167735)*x[0]+(0.745969147169)*x[1]
        s[1,4]=(-0.449555353075)*x[0]+(0.67139606475)*x[1]
        s[2,0]=(-1.35267387997)*x[0]+(0.827592755339)*x[1]
        s[2,1]=(-0.332416565825)*x[0]+(-0.570384450901)*x[1]
        s[2,2]=(0.228812981844)*x[0]+(-0.506320631557)*x[1]
        s[2,3]=(0.346948032916)*x[0]+(-0.0370468335786)*x[1]
        s[2,4]=(-0.38539371487)*x[0]+(-1.53408220383)*x[1]
        s[3,0]=(-0.446967528892)*x[0]+(-0.577472772294)*x[1]
        s[3,1]=(1.3016519192)*x[0]+(-1.21796536064)*x[1]
        s[3,2]=(-1.11173710879)*x[0]+(0.0982539955389)*x[1]
        s[3,3]=(0.398891317196)*x[0]+(0.0151699052193)*x[1]
        s[3,4]=(0.702195275642)*x[0]+(-0.355526973434)*x[1]
        ref[0,0]=(0.0481111909695)*x_ref[0]+(-0.393526010503)*x_ref[1]
        ref[0,1]=(-0.284097772097)*x_ref[0]+(0.618264565258)*x_ref[1]
        ref[0,2]=(0.401575436099)*x_ref[0]+(-0.697718912762)*x_ref[1]
        ref[0,3]=(-1.49798792514)*x_ref[0]+(-0.0425176867511)*x_ref[1]
        ref[0,4]=(0.213840882194)*x_ref[0]+(-0.914489018365)*x_ref[1]
        ref[1,0]=(0.857166519205)*x_ref[0]+(1.5282054335)*x_ref[1]
        ref[1,1]=(1.61091698609)*x_ref[0]+(-0.548453984777)*x_ref[1]
        ref[1,2]=(-1.56266453199)*x_ref[0]+(-1.66187429416)*x_ref[1]
        ref[1,3]=(0.343955167735)*x_ref[0]+(0.745969147169)*x_ref[1]
        ref[1,4]=(-0.449555353075)*x_ref[0]+(0.67139606475)*x_ref[1]
        ref[2,0]=(-1.35267387997)*x_ref[0]+(0.827592755339)*x_ref[1]
        ref[2,1]=(-0.332416565825)*x_ref[0]+(-0.570384450901)*x_ref[1]
        ref[2,2]=(0.228812981844)*x_ref[0]+(-0.506320631557)*x_ref[1]
        ref[2,3]=(0.346948032916)*x_ref[0]+(-0.0370468335786)*x_ref[1]
        ref[2,4]=(-0.38539371487)*x_ref[0]+(-1.53408220383)*x_ref[1]
        ref[3,0]=(-0.446967528892)*x_ref[0]+(-0.577472772294)*x_ref[1]
        ref[3,1]=(1.3016519192)*x_ref[0]+(-1.21796536064)*x_ref[1]
        ref[3,2]=(-1.11173710879)*x_ref[0]+(0.0982539955389)*x_ref[1]
        ref[3,3]=(0.398891317196)*x_ref[0]+(0.0151699052193)*x_ref[1]
        ref[3,4]=(0.702195275642)*x_ref[0]+(-0.355526973434)*x_ref[1]
      else:
        s[0,0]=(0.525077989935)*x[0]+(-0.129538863404)*x[1]+(-1.47575933759)*x[2]
        s[0,1]=(0.575439663003)*x[0]+(-1.40344366825)*x[1]+(0.221748855921)*x[2]
        s[0,2]=(0.171201653563)*x[0]+(-0.185657063083)*x[1]+(0.42305480444)*x[2]
        s[0,3]=(0.787118287986)*x[0]+(-1.43947391779)*x[1]+(0.322131428849)*x[2]
        s[0,4]=(-0.693857541121)*x[0]+(1.29660018396)*x[1]+(1.01092834477)*x[2]
        s[1,0]=(0.371970743598)*x[0]+(-0.744756043562)*x[1]+(0.194145828387)*x[2]
        s[1,1]=(1.23828194153)*x[0]+(-0.937084509616)*x[1]+(-1.47063754607)*x[2]
        s[1,2]=(-0.287641758427)*x[0]+(-0.42656557181)*x[1]+(1.49807905914)*x[2]
        s[1,3]=(-0.412169130593)*x[0]+(0.590235477443)*x[1]+(-0.575739422159)*x[2]
        s[1,4]=(-0.372408793303)*x[0]+(1.48960296195)*x[1]+(1.42957998207)*x[2]
        s[2,0]=(-0.356728277451)*x[0]+(0.562758907106)*x[1]+(-0.0825032302409)*x[2]
        s[2,1]=(-1.14497343552)*x[0]+(0.605393479159)*x[1]+(-0.286645936691)*x[2]
        s[2,2]=(-0.00750519201495)*x[0]+(-0.742525175154)*x[1]+(1.33814411456)*x[2]
        s[2,3]=(-1.12547614642)*x[0]+(0.291849175119)*x[1]+(0.745190340427)*x[2]
        s[2,4]=(0.0863930736605)*x[0]+(0.726972597715)*x[1]+(0.335180039249)*x[2]
        s[3,0]=(0.164315801392)*x[0]+(-0.276344360153)*x[1]+(0.364776703598)*x[2]
        s[3,1]=(0.980174491212)*x[0]+(0.920966782066)*x[1]+(1.6251437721)*x[2]
        s[3,2]=(0.722553089652)*x[0]+(-1.29237901172)*x[1]+(1.38442221241)*x[2]
        s[3,3]=(-1.1138747054)*x[0]+(0.184113285873)*x[1]+(0.113137432772)*x[2]
        s[3,4]=(0.560373055499)*x[0]+(0.347806062763)*x[1]+(-0.103028298482)*x[2]
        ref[0,0]=(0.525077989935)*x_ref[0]+(-0.129538863404)*x_ref[1]+(-1.47575933759)*x_ref[2]
        ref[0,1]=(0.575439663003)*x_ref[0]+(-1.40344366825)*x_ref[1]+(0.221748855921)*x_ref[2]
        ref[0,2]=(0.171201653563)*x_ref[0]+(-0.185657063083)*x_ref[1]+(0.42305480444)*x_ref[2]
        ref[0,3]=(0.787118287986)*x_ref[0]+(-1.43947391779)*x_ref[1]+(0.322131428849)*x_ref[2]
        ref[0,4]=(-0.693857541121)*x_ref[0]+(1.29660018396)*x_ref[1]+(1.01092834477)*x_ref[2]
        ref[1,0]=(0.371970743598)*x_ref[0]+(-0.744756043562)*x_ref[1]+(0.194145828387)*x_ref[2]
        ref[1,1]=(1.23828194153)*x_ref[0]+(-0.937084509616)*x_ref[1]+(-1.47063754607)*x_ref[2]
        ref[1,2]=(-0.287641758427)*x_ref[0]+(-0.42656557181)*x_ref[1]+(1.49807905914)*x_ref[2]
        ref[1,3]=(-0.412169130593)*x_ref[0]+(0.590235477443)*x_ref[1]+(-0.575739422159)*x_ref[2]
        ref[1,4]=(-0.372408793303)*x_ref[0]+(1.48960296195)*x_ref[1]+(1.42957998207)*x_ref[2]
        ref[2,0]=(-0.356728277451)*x_ref[0]+(0.562758907106)*x_ref[1]+(-0.0825032302409)*x_ref[2]
        ref[2,1]=(-1.14497343552)*x_ref[0]+(0.605393479159)*x_ref[1]+(-0.286645936691)*x_ref[2]
        ref[2,2]=(-0.00750519201495)*x_ref[0]+(-0.742525175154)*x_ref[1]+(1.33814411456)*x_ref[2]
        ref[2,3]=(-1.12547614642)*x_ref[0]+(0.291849175119)*x_ref[1]+(0.745190340427)*x_ref[2]
        ref[2,4]=(0.0863930736605)*x_ref[0]+(0.726972597715)*x_ref[1]+(0.335180039249)*x_ref[2]
        ref[3,0]=(0.164315801392)*x_ref[0]+(-0.276344360153)*x_ref[1]+(0.364776703598)*x_ref[2]
        ref[3,1]=(0.980174491212)*x_ref[0]+(0.920966782066)*x_ref[1]+(1.6251437721)*x_ref[2]
        ref[3,2]=(0.722553089652)*x_ref[0]+(-1.29237901172)*x_ref[1]+(1.38442221241)*x_ref[2]
        ref[3,3]=(-1.1138747054)*x_ref[0]+(0.184113285873)*x_ref[1]+(0.113137432772)*x_ref[2]
        ref[3,4]=(0.560373055499)*x_ref[0]+(0.347806062763)*x_ref[1]+(-0.103028298482)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_Solution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(-0.616803876826)*x[0]+(0.239741751767)*x[1]
        s[0,0,1]=(0.112656354167)*x[0]+(0.0390998528525)*x[1]
        s[0,1,0]=(1.02334404242)*x[0]+(1.16228451879)*x[1]
        s[0,1,1]=(1.26927845423)*x[0]+(-0.0309071031648)*x[1]
        s[1,0,0]=(0.154947909629)*x[0]+(1.41925252494)*x[1]
        s[1,0,1]=(0.299309779213)*x[0]+(0.558282387192)*x[1]
        s[1,1,0]=(0.748405958274)*x[0]+(-0.241956071057)*x[1]
        s[1,1,1]=(-0.100230866827)*x[0]+(-0.616595937224)*x[1]
        s[2,0,0]=(-0.485117813294)*x[0]+(-0.814821957713)*x[1]
        s[2,0,1]=(0.573476779093)*x[0]+(0.480686896264)*x[1]
        s[2,1,0]=(0.79651609962)*x[0]+(0.257600415705)*x[1]
        s[2,1,1]=(-0.162020224334)*x[0]+(1.2460751805)*x[1]
        s[3,0,0]=(0.467277953063)*x[0]+(0.254739285648)*x[1]
        s[3,0,1]=(-0.619802362841)*x[0]+(-1.30062481861)*x[1]
        s[3,1,0]=(0.160765639764)*x[0]+(-0.716696954941)*x[1]
        s[3,1,1]=(-0.0317856002877)*x[0]+(0.561299373553)*x[1]
        s[4,0,0]=(0.238090555098)*x[0]+(-0.99128263972)*x[1]
        s[4,0,1]=(-0.90912380586)*x[0]+(-0.32214847292)*x[1]
        s[4,1,0]=(1.67021862968)*x[0]+(-0.396622542538)*x[1]
        s[4,1,1]=(-0.118992684804)*x[0]+(-0.229200542459)*x[1]
        s[5,0,0]=(-0.124626687555)*x[0]+(0.429266692145)*x[1]
        s[5,0,1]=(1.40894998937)*x[0]+(-0.222278182789)*x[1]
        s[5,1,0]=(-0.291128816468)*x[0]+(1.33531261196)*x[1]
        s[5,1,1]=(-0.985994002551)*x[0]+(1.10232269681)*x[1]
        ref[0,0,0]=(-0.616803876826)*x_ref[0]+(0.239741751767)*x_ref[1]
        ref[0,0,1]=(0.112656354167)*x_ref[0]+(0.0390998528525)*x_ref[1]
        ref[0,1,0]=(1.02334404242)*x_ref[0]+(1.16228451879)*x_ref[1]
        ref[0,1,1]=(1.26927845423)*x_ref[0]+(-0.0309071031648)*x_ref[1]
        ref[1,0,0]=(0.154947909629)*x_ref[0]+(1.41925252494)*x_ref[1]
        ref[1,0,1]=(0.299309779213)*x_ref[0]+(0.558282387192)*x_ref[1]
        ref[1,1,0]=(0.748405958274)*x_ref[0]+(-0.241956071057)*x_ref[1]
        ref[1,1,1]=(-0.100230866827)*x_ref[0]+(-0.616595937224)*x_ref[1]
        ref[2,0,0]=(-0.485117813294)*x_ref[0]+(-0.814821957713)*x_ref[1]
        ref[2,0,1]=(0.573476779093)*x_ref[0]+(0.480686896264)*x_ref[1]
        ref[2,1,0]=(0.79651609962)*x_ref[0]+(0.257600415705)*x_ref[1]
        ref[2,1,1]=(-0.162020224334)*x_ref[0]+(1.2460751805)*x_ref[1]
        ref[3,0,0]=(0.467277953063)*x_ref[0]+(0.254739285648)*x_ref[1]
        ref[3,0,1]=(-0.619802362841)*x_ref[0]+(-1.30062481861)*x_ref[1]
        ref[3,1,0]=(0.160765639764)*x_ref[0]+(-0.716696954941)*x_ref[1]
        ref[3,1,1]=(-0.0317856002877)*x_ref[0]+(0.561299373553)*x_ref[1]
        ref[4,0,0]=(0.238090555098)*x_ref[0]+(-0.99128263972)*x_ref[1]
        ref[4,0,1]=(-0.90912380586)*x_ref[0]+(-0.32214847292)*x_ref[1]
        ref[4,1,0]=(1.67021862968)*x_ref[0]+(-0.396622542538)*x_ref[1]
        ref[4,1,1]=(-0.118992684804)*x_ref[0]+(-0.229200542459)*x_ref[1]
        ref[5,0,0]=(-0.124626687555)*x_ref[0]+(0.429266692145)*x_ref[1]
        ref[5,0,1]=(1.40894998937)*x_ref[0]+(-0.222278182789)*x_ref[1]
        ref[5,1,0]=(-0.291128816468)*x_ref[0]+(1.33531261196)*x_ref[1]
        ref[5,1,1]=(-0.985994002551)*x_ref[0]+(1.10232269681)*x_ref[1]
      else:
        s[0,0,0]=(-0.622303225663)*x[0]+(-0.468862605979)*x[1]+(-0.822269949108)*x[2]
        s[0,0,1]=(1.42712676639)*x[0]+(0.466554590146)*x[1]+(-0.416360290497)*x[2]
        s[0,1,0]=(-0.56276315017)*x[0]+(0.681130342695)*x[1]+(-1.150317219)*x[2]
        s[0,1,1]=(0.840875694603)*x[0]+(1.06129869177)*x[1]+(0.186757925344)*x[2]
        s[1,0,0]=(1.10693557715)*x[0]+(0.515133841775)*x[1]+(-1.46650509667)*x[2]
        s[1,0,1]=(0.469874083814)*x[0]+(0.712094004219)*x[1]+(0.0428537567242)*x[2]
        s[1,1,0]=(1.09283636268)*x[0]+(-0.922481846245)*x[1]+(0.611758101252)*x[2]
        s[1,1,1]=(0.32198205915)*x[0]+(0.75194760009)*x[1]+(-0.118025900786)*x[2]
        s[2,0,0]=(-1.52389124946)*x[0]+(1.49105249151)*x[1]+(-0.874107742996)*x[2]
        s[2,0,1]=(-0.90495363127)*x[0]+(0.078342943764)*x[1]+(0.219749629688)*x[2]
        s[2,1,0]=(-0.00849262013635)*x[0]+(0.00531993772995)*x[1]+(1.76477138966)*x[2]
        s[2,1,1]=(0.0396318942805)*x[0]+(0.630819877838)*x[1]+(-1.44621732026)*x[2]
        s[3,0,0]=(0.0109746356852)*x[0]+(-0.362009865261)*x[1]+(-0.538016468887)*x[2]
        s[3,0,1]=(-0.234158150184)*x[0]+(-1.44595936898)*x[1]+(0.329273103366)*x[2]
        s[3,1,0]=(0.146354345455)*x[0]+(0.138204478429)*x[1]+(1.61082500463)*x[2]
        s[3,1,1]=(1.16444593046)*x[0]+(-0.353850126255)*x[1]+(-0.0472752976526)*x[2]
        s[4,0,0]=(0.440054653358)*x[0]+(-1.16079243521)*x[1]+(-0.135698496809)*x[2]
        s[4,0,1]=(-1.89214756492)*x[0]+(-0.118895763118)*x[1]+(0.44262404148)*x[2]
        s[4,1,0]=(-1.13364355366)*x[0]+(1.47153317046)*x[1]+(0.519901952659)*x[2]
        s[4,1,1]=(0.340556700422)*x[0]+(0.270041963096)*x[1]+(0.303701099668)*x[2]
        s[5,0,0]=(0.779284769306)*x[0]+(0.91773222667)*x[1]+(0.870658706587)*x[2]
        s[5,0,1]=(-1.19937838023)*x[0]+(-0.182653397415)*x[1]+(0.229890290729)*x[2]
        s[5,1,0]=(-1.22173252282)*x[0]+(-0.465952231432)*x[1]+(-0.173356024515)*x[2]
        s[5,1,1]=(-0.175957737331)*x[0]+(0.512508313109)*x[1]+(-0.81846318618)*x[2]
        ref[0,0,0]=(-0.622303225663)*x_ref[0]+(-0.468862605979)*x_ref[1]+(-0.822269949108)*x_ref[2]
        ref[0,0,1]=(1.42712676639)*x_ref[0]+(0.466554590146)*x_ref[1]+(-0.416360290497)*x_ref[2]
        ref[0,1,0]=(-0.56276315017)*x_ref[0]+(0.681130342695)*x_ref[1]+(-1.150317219)*x_ref[2]
        ref[0,1,1]=(0.840875694603)*x_ref[0]+(1.06129869177)*x_ref[1]+(0.186757925344)*x_ref[2]
        ref[1,0,0]=(1.10693557715)*x_ref[0]+(0.515133841775)*x_ref[1]+(-1.46650509667)*x_ref[2]
        ref[1,0,1]=(0.469874083814)*x_ref[0]+(0.712094004219)*x_ref[1]+(0.0428537567242)*x_ref[2]
        ref[1,1,0]=(1.09283636268)*x_ref[0]+(-0.922481846245)*x_ref[1]+(0.611758101252)*x_ref[2]
        ref[1,1,1]=(0.32198205915)*x_ref[0]+(0.75194760009)*x_ref[1]+(-0.118025900786)*x_ref[2]
        ref[2,0,0]=(-1.52389124946)*x_ref[0]+(1.49105249151)*x_ref[1]+(-0.874107742996)*x_ref[2]
        ref[2,0,1]=(-0.90495363127)*x_ref[0]+(0.078342943764)*x_ref[1]+(0.219749629688)*x_ref[2]
        ref[2,1,0]=(-0.00849262013635)*x_ref[0]+(0.00531993772995)*x_ref[1]+(1.76477138966)*x_ref[2]
        ref[2,1,1]=(0.0396318942805)*x_ref[0]+(0.630819877838)*x_ref[1]+(-1.44621732026)*x_ref[2]
        ref[3,0,0]=(0.0109746356852)*x_ref[0]+(-0.362009865261)*x_ref[1]+(-0.538016468887)*x_ref[2]
        ref[3,0,1]=(-0.234158150184)*x_ref[0]+(-1.44595936898)*x_ref[1]+(0.329273103366)*x_ref[2]
        ref[3,1,0]=(0.146354345455)*x_ref[0]+(0.138204478429)*x_ref[1]+(1.61082500463)*x_ref[2]
        ref[3,1,1]=(1.16444593046)*x_ref[0]+(-0.353850126255)*x_ref[1]+(-0.0472752976526)*x_ref[2]
        ref[4,0,0]=(0.440054653358)*x_ref[0]+(-1.16079243521)*x_ref[1]+(-0.135698496809)*x_ref[2]
        ref[4,0,1]=(-1.89214756492)*x_ref[0]+(-0.118895763118)*x_ref[1]+(0.44262404148)*x_ref[2]
        ref[4,1,0]=(-1.13364355366)*x_ref[0]+(1.47153317046)*x_ref[1]+(0.519901952659)*x_ref[2]
        ref[4,1,1]=(0.340556700422)*x_ref[0]+(0.270041963096)*x_ref[1]+(0.303701099668)*x_ref[2]
        ref[5,0,0]=(0.779284769306)*x_ref[0]+(0.91773222667)*x_ref[1]+(0.870658706587)*x_ref[2]
        ref[5,0,1]=(-1.19937838023)*x_ref[0]+(-0.182653397415)*x_ref[1]+(0.229890290729)*x_ref[2]
        ref[5,1,0]=(-1.22173252282)*x_ref[0]+(-0.465952231432)*x_ref[1]+(-0.173356024515)*x_ref[2]
        ref[5,1,1]=(-0.175957737331)*x_ref[0]+(0.512508313109)*x_ref[1]+(-0.81846318618)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_Solution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from Solution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.263699858395)*x[0]+(0.863340165591)*x[1]
        s[0,0,0,1]=(0.168551523361)*x[0]+(0.822827042977)*x[1]
        s[0,0,1,0]=(0.922495824101)*x[0]+(1.10277493051)*x[1]
        s[0,0,1,1]=(0.858215828807)*x[0]+(0.27444875925)*x[1]
        s[0,0,2,0]=(0.203856130696)*x[0]+(-1.32433677722)*x[1]
        s[0,0,2,1]=(-1.2222254241)*x[0]+(-0.53464384301)*x[1]
        s[0,1,0,0]=(0.0467230980779)*x[0]+(1.64188619595)*x[1]
        s[0,1,0,1]=(0.917582861623)*x[0]+(-1.01711076348)*x[1]
        s[0,1,1,0]=(-0.70445337121)*x[0]+(-0.499786458329)*x[1]
        s[0,1,1,1]=(1.03899670749)*x[0]+(-1.43262408006)*x[1]
        s[0,1,2,0]=(0.409224950113)*x[0]+(1.48049510634)*x[1]
        s[0,1,2,1]=(-0.639323586948)*x[0]+(0.41324495759)*x[1]
        s[0,2,0,0]=(-1.04843809688)*x[0]+(0.409833243907)*x[1]
        s[0,2,0,1]=(1.2606980391)*x[0]+(1.05476378621)*x[1]
        s[0,2,1,0]=(-0.0653124068721)*x[0]+(1.68355228676)*x[1]
        s[0,2,1,1]=(1.56886755247)*x[0]+(-1.67289454758)*x[1]
        s[0,2,2,0]=(0.542254512691)*x[0]+(-0.574877434565)*x[1]
        s[0,2,2,1]=(-1.79104846611)*x[0]+(0.752170029246)*x[1]
        s[0,3,0,0]=(-0.804486243116)*x[0]+(1.58214181283)*x[1]
        s[0,3,0,1]=(0.268657213023)*x[0]+(-0.632746875992)*x[1]
        s[0,3,1,0]=(0.748200841554)*x[0]+(-0.372804380425)*x[1]
        s[0,3,1,1]=(0.532346891312)*x[0]+(0.0944645102486)*x[1]
        s[0,3,2,0]=(0.255430361824)*x[0]+(0.493809212019)*x[1]
        s[0,3,2,1]=(1.06219716938)*x[0]+(0.0256120510252)*x[1]
        s[0,4,0,0]=(-1.11393336407)*x[0]+(0.0040462142619)*x[1]
        s[0,4,0,1]=(0.290958676921)*x[0]+(1.43659749943)*x[1]
        s[0,4,1,0]=(-0.921614727859)*x[0]+(-0.730623491689)*x[1]
        s[0,4,1,1]=(-0.539009340284)*x[0]+(0.162040517816)*x[1]
        s[0,4,2,0]=(-0.166431286126)*x[0]+(-0.881280007454)*x[1]
        s[0,4,2,1]=(0.939057506826)*x[0]+(-0.798845422966)*x[1]
        s[1,0,0,0]=(-1.48556828853)*x[0]+(0.398947142701)*x[1]
        s[1,0,0,1]=(-0.867994319546)*x[0]+(-1.51760635946)*x[1]
        s[1,0,1,0]=(-1.3436690984)*x[0]+(0.0587809657777)*x[1]
        s[1,0,1,1]=(-0.358397464109)*x[0]+(-0.110074947013)*x[1]
        s[1,0,2,0]=(-1.65339451931)*x[0]+(-0.628438027419)*x[1]
        s[1,0,2,1]=(0.75949197524)*x[0]+(1.47344312464)*x[1]
        s[1,1,0,0]=(-0.184897359409)*x[0]+(0.997493671981)*x[1]
        s[1,1,0,1]=(0.0348263424069)*x[0]+(-0.113374992622)*x[1]
        s[1,1,1,0]=(1.48734955276)*x[0]+(-0.789492365846)*x[1]
        s[1,1,1,1]=(0.865097624364)*x[0]+(0.117588671443)*x[1]
        s[1,1,2,0]=(0.195936219884)*x[0]+(-0.948389540535)*x[1]
        s[1,1,2,1]=(1.11314083647)*x[0]+(0.215110041922)*x[1]
        s[1,2,0,0]=(-0.920590092452)*x[0]+(-0.612482637981)*x[1]
        s[1,2,0,1]=(1.04301519124)*x[0]+(-0.0629305143129)*x[1]
        s[1,2,1,0]=(0.382136489277)*x[0]+(-0.230145907219)*x[1]
        s[1,2,1,1]=(-0.0252796835879)*x[0]+(-0.900125981212)*x[1]
        s[1,2,2,0]=(0.965035634162)*x[0]+(-1.22815714841)*x[1]
        s[1,2,2,1]=(0.357785163717)*x[0]+(1.27915764344)*x[1]
        s[1,3,0,0]=(-1.27549226474)*x[0]+(1.24432292599)*x[1]
        s[1,3,0,1]=(-0.0383725074991)*x[0]+(-0.349361287486)*x[1]
        s[1,3,1,0]=(-0.200272394326)*x[0]+(1.06221698364)*x[1]
        s[1,3,1,1]=(-0.215946498482)*x[0]+(1.01531234763)*x[1]
        s[1,3,2,0]=(-1.22728454222)*x[0]+(0.907627178806)*x[1]
        s[1,3,2,1]=(-1.00398418868)*x[0]+(-0.199827024316)*x[1]
        s[1,4,0,0]=(-0.00303564690889)*x[0]+(0.760141903922)*x[1]
        s[1,4,0,1]=(-0.456280795832)*x[0]+(0.986442355989)*x[1]
        s[1,4,1,0]=(-0.318954368532)*x[0]+(-0.966894508646)*x[1]
        s[1,4,1,1]=(0.65555609342)*x[0]+(-0.526267162413)*x[1]
        s[1,4,2,0]=(1.72927006133)*x[0]+(1.65889052032)*x[1]
        s[1,4,2,1]=(0.340218940988)*x[0]+(-0.815180133977)*x[1]
        s[2,0,0,0]=(-0.910985187877)*x[0]+(-1.31360663975)*x[1]
        s[2,0,0,1]=(-0.174389735675)*x[0]+(0.268073959103)*x[1]
        s[2,0,1,0]=(-0.111243591678)*x[0]+(-0.7123465972)*x[1]
        s[2,0,1,1]=(1.03340508201)*x[0]+(-1.15516978578)*x[1]
        s[2,0,2,0]=(0.0791763687033)*x[0]+(-0.0239321376253)*x[1]
        s[2,0,2,1]=(-0.0296109649235)*x[0]+(-0.292837031166)*x[1]
        s[2,1,0,0]=(-0.00911908859083)*x[0]+(-1.17279969051)*x[1]
        s[2,1,0,1]=(-1.06648321801)*x[0]+(0.126622301021)*x[1]
        s[2,1,1,0]=(0.275881626889)*x[0]+(1.69565501602)*x[1]
        s[2,1,1,1]=(-0.409495321219)*x[0]+(-0.407778373033)*x[1]
        s[2,1,2,0]=(1.33276969465)*x[0]+(-1.17532941128)*x[1]
        s[2,1,2,1]=(-0.866150634496)*x[0]+(0.797755014562)*x[1]
        s[2,2,0,0]=(-0.106882476852)*x[0]+(-0.249249128227)*x[1]
        s[2,2,0,1]=(-0.520287837168)*x[0]+(0.338818467136)*x[1]
        s[2,2,1,0]=(-0.852757899877)*x[0]+(0.321704892591)*x[1]
        s[2,2,1,1]=(0.142520864112)*x[0]+(-1.05840136269)*x[1]
        s[2,2,2,0]=(-0.715836025391)*x[0]+(0.22306435555)*x[1]
        s[2,2,2,1]=(-0.586149670377)*x[0]+(-0.516267803245)*x[1]
        s[2,3,0,0]=(-0.950483491273)*x[0]+(0.624915645607)*x[1]
        s[2,3,0,1]=(-0.217697021618)*x[0]+(0.0442099818289)*x[1]
        s[2,3,1,0]=(-1.01046266386)*x[0]+(-1.48154228284)*x[1]
        s[2,3,1,1]=(-1.3050947534)*x[0]+(-0.998384196826)*x[1]
        s[2,3,2,0]=(0.267414868071)*x[0]+(1.15992550636)*x[1]
        s[2,3,2,1]=(-0.862214432382)*x[0]+(0.292670280728)*x[1]
        s[2,4,0,0]=(-0.134428442874)*x[0]+(0.719512272598)*x[1]
        s[2,4,0,1]=(-1.67096566574)*x[0]+(0.43067843233)*x[1]
        s[2,4,1,0]=(0.188806295154)*x[0]+(-1.54252342316)*x[1]
        s[2,4,1,1]=(0.0976229597501)*x[0]+(-0.937051784938)*x[1]
        s[2,4,2,0]=(-0.00596882510121)*x[0]+(-0.312957453878)*x[1]
        s[2,4,2,1]=(-1.02156076253)*x[0]+(-0.896718827796)*x[1]
        s[3,0,0,0]=(0.0133947002051)*x[0]+(0.238649243438)*x[1]
        s[3,0,0,1]=(1.00467946164)*x[0]+(1.34133983171)*x[1]
        s[3,0,1,0]=(0.341713854404)*x[0]+(0.511010566842)*x[1]
        s[3,0,1,1]=(0.546165677394)*x[0]+(0.321446986518)*x[1]
        s[3,0,2,0]=(-0.764072559141)*x[0]+(-0.312532010803)*x[1]
        s[3,0,2,1]=(-0.811247938505)*x[0]+(-0.939324734555)*x[1]
        s[3,1,0,0]=(0.572111068237)*x[0]+(-0.0530416981116)*x[1]
        s[3,1,0,1]=(0.837385483064)*x[0]+(0.928431354228)*x[1]
        s[3,1,1,0]=(0.238702836098)*x[0]+(0.182536052546)*x[1]
        s[3,1,1,1]=(0.461955717914)*x[0]+(0.0256860825024)*x[1]
        s[3,1,2,0]=(-0.787501340181)*x[0]+(-0.828497466795)*x[1]
        s[3,1,2,1]=(0.348032537213)*x[0]+(-0.984456869176)*x[1]
        s[3,2,0,0]=(0.588847373206)*x[0]+(1.84398386935)*x[1]
        s[3,2,0,1]=(-0.2085469075)*x[0]+(-0.244926359847)*x[1]
        s[3,2,1,0]=(-1.62397953336)*x[0]+(-0.102475340253)*x[1]
        s[3,2,1,1]=(0.455695325345)*x[0]+(0.552529049227)*x[1]
        s[3,2,2,0]=(0.734375962799)*x[0]+(0.279367642817)*x[1]
        s[3,2,2,1]=(1.52878551603)*x[0]+(-1.43594288923)*x[1]
        s[3,3,0,0]=(-0.331774589719)*x[0]+(-0.852014090099)*x[1]
        s[3,3,0,1]=(0.0495293070849)*x[0]+(1.70225124669)*x[1]
        s[3,3,1,0]=(1.37894662758)*x[0]+(0.350193222832)*x[1]
        s[3,3,1,1]=(0.00570139162378)*x[0]+(-0.37112295664)*x[1]
        s[3,3,2,0]=(-0.205033339669)*x[0]+(-0.695461477152)*x[1]
        s[3,3,2,1]=(0.12868380026)*x[0]+(0.764317872049)*x[1]
        s[3,4,0,0]=(-1.04309352143)*x[0]+(-0.805487628821)*x[1]
        s[3,4,0,1]=(-0.496744676221)*x[0]+(-0.13812443637)*x[1]
        s[3,4,1,0]=(0.0339710647687)*x[0]+(0.0274093213606)*x[1]
        s[3,4,1,1]=(-0.516837676304)*x[0]+(-0.119504191446)*x[1]
        s[3,4,2,0]=(-0.472643263791)*x[0]+(-0.337262686677)*x[1]
        s[3,4,2,1]=(-1.34135241071)*x[0]+(-0.185936541491)*x[1]
        ref[0,0,0,0]=(-0.263699858395)*x_ref[0]+(0.863340165591)*x_ref[1]
        ref[0,0,0,1]=(0.168551523361)*x_ref[0]+(0.822827042977)*x_ref[1]
        ref[0,0,1,0]=(0.922495824101)*x_ref[0]+(1.10277493051)*x_ref[1]
        ref[0,0,1,1]=(0.858215828807)*x_ref[0]+(0.27444875925)*x_ref[1]
        ref[0,0,2,0]=(0.203856130696)*x_ref[0]+(-1.32433677722)*x_ref[1]
        ref[0,0,2,1]=(-1.2222254241)*x_ref[0]+(-0.53464384301)*x_ref[1]
        ref[0,1,0,0]=(0.0467230980779)*x_ref[0]+(1.64188619595)*x_ref[1]
        ref[0,1,0,1]=(0.917582861623)*x_ref[0]+(-1.01711076348)*x_ref[1]
        ref[0,1,1,0]=(-0.70445337121)*x_ref[0]+(-0.499786458329)*x_ref[1]
        ref[0,1,1,1]=(1.03899670749)*x_ref[0]+(-1.43262408006)*x_ref[1]
        ref[0,1,2,0]=(0.409224950113)*x_ref[0]+(1.48049510634)*x_ref[1]
        ref[0,1,2,1]=(-0.639323586948)*x_ref[0]+(0.41324495759)*x_ref[1]
        ref[0,2,0,0]=(-1.04843809688)*x_ref[0]+(0.409833243907)*x_ref[1]
        ref[0,2,0,1]=(1.2606980391)*x_ref[0]+(1.05476378621)*x_ref[1]
        ref[0,2,1,0]=(-0.0653124068721)*x_ref[0]+(1.68355228676)*x_ref[1]
        ref[0,2,1,1]=(1.56886755247)*x_ref[0]+(-1.67289454758)*x_ref[1]
        ref[0,2,2,0]=(0.542254512691)*x_ref[0]+(-0.574877434565)*x_ref[1]
        ref[0,2,2,1]=(-1.79104846611)*x_ref[0]+(0.752170029246)*x_ref[1]
        ref[0,3,0,0]=(-0.804486243116)*x_ref[0]+(1.58214181283)*x_ref[1]
        ref[0,3,0,1]=(0.268657213023)*x_ref[0]+(-0.632746875992)*x_ref[1]
        ref[0,3,1,0]=(0.748200841554)*x_ref[0]+(-0.372804380425)*x_ref[1]
        ref[0,3,1,1]=(0.532346891312)*x_ref[0]+(0.0944645102486)*x_ref[1]
        ref[0,3,2,0]=(0.255430361824)*x_ref[0]+(0.493809212019)*x_ref[1]
        ref[0,3,2,1]=(1.06219716938)*x_ref[0]+(0.0256120510252)*x_ref[1]
        ref[0,4,0,0]=(-1.11393336407)*x_ref[0]+(0.0040462142619)*x_ref[1]
        ref[0,4,0,1]=(0.290958676921)*x_ref[0]+(1.43659749943)*x_ref[1]
        ref[0,4,1,0]=(-0.921614727859)*x_ref[0]+(-0.730623491689)*x_ref[1]
        ref[0,4,1,1]=(-0.539009340284)*x_ref[0]+(0.162040517816)*x_ref[1]
        ref[0,4,2,0]=(-0.166431286126)*x_ref[0]+(-0.881280007454)*x_ref[1]
        ref[0,4,2,1]=(0.939057506826)*x_ref[0]+(-0.798845422966)*x_ref[1]
        ref[1,0,0,0]=(-1.48556828853)*x_ref[0]+(0.398947142701)*x_ref[1]
        ref[1,0,0,1]=(-0.867994319546)*x_ref[0]+(-1.51760635946)*x_ref[1]
        ref[1,0,1,0]=(-1.3436690984)*x_ref[0]+(0.0587809657777)*x_ref[1]
        ref[1,0,1,1]=(-0.358397464109)*x_ref[0]+(-0.110074947013)*x_ref[1]
        ref[1,0,2,0]=(-1.65339451931)*x_ref[0]+(-0.628438027419)*x_ref[1]
        ref[1,0,2,1]=(0.75949197524)*x_ref[0]+(1.47344312464)*x_ref[1]
        ref[1,1,0,0]=(-0.184897359409)*x_ref[0]+(0.997493671981)*x_ref[1]
        ref[1,1,0,1]=(0.0348263424069)*x_ref[0]+(-0.113374992622)*x_ref[1]
        ref[1,1,1,0]=(1.48734955276)*x_ref[0]+(-0.789492365846)*x_ref[1]
        ref[1,1,1,1]=(0.865097624364)*x_ref[0]+(0.117588671443)*x_ref[1]
        ref[1,1,2,0]=(0.195936219884)*x_ref[0]+(-0.948389540535)*x_ref[1]
        ref[1,1,2,1]=(1.11314083647)*x_ref[0]+(0.215110041922)*x_ref[1]
        ref[1,2,0,0]=(-0.920590092452)*x_ref[0]+(-0.612482637981)*x_ref[1]
        ref[1,2,0,1]=(1.04301519124)*x_ref[0]+(-0.0629305143129)*x_ref[1]
        ref[1,2,1,0]=(0.382136489277)*x_ref[0]+(-0.230145907219)*x_ref[1]
        ref[1,2,1,1]=(-0.0252796835879)*x_ref[0]+(-0.900125981212)*x_ref[1]
        ref[1,2,2,0]=(0.965035634162)*x_ref[0]+(-1.22815714841)*x_ref[1]
        ref[1,2,2,1]=(0.357785163717)*x_ref[0]+(1.27915764344)*x_ref[1]
        ref[1,3,0,0]=(-1.27549226474)*x_ref[0]+(1.24432292599)*x_ref[1]
        ref[1,3,0,1]=(-0.0383725074991)*x_ref[0]+(-0.349361287486)*x_ref[1]
        ref[1,3,1,0]=(-0.200272394326)*x_ref[0]+(1.06221698364)*x_ref[1]
        ref[1,3,1,1]=(-0.215946498482)*x_ref[0]+(1.01531234763)*x_ref[1]
        ref[1,3,2,0]=(-1.22728454222)*x_ref[0]+(0.907627178806)*x_ref[1]
        ref[1,3,2,1]=(-1.00398418868)*x_ref[0]+(-0.199827024316)*x_ref[1]
        ref[1,4,0,0]=(-0.00303564690889)*x_ref[0]+(0.760141903922)*x_ref[1]
        ref[1,4,0,1]=(-0.456280795832)*x_ref[0]+(0.986442355989)*x_ref[1]
        ref[1,4,1,0]=(-0.318954368532)*x_ref[0]+(-0.966894508646)*x_ref[1]
        ref[1,4,1,1]=(0.65555609342)*x_ref[0]+(-0.526267162413)*x_ref[1]
        ref[1,4,2,0]=(1.72927006133)*x_ref[0]+(1.65889052032)*x_ref[1]
        ref[1,4,2,1]=(0.340218940988)*x_ref[0]+(-0.815180133977)*x_ref[1]
        ref[2,0,0,0]=(-0.910985187877)*x_ref[0]+(-1.31360663975)*x_ref[1]
        ref[2,0,0,1]=(-0.174389735675)*x_ref[0]+(0.268073959103)*x_ref[1]
        ref[2,0,1,0]=(-0.111243591678)*x_ref[0]+(-0.7123465972)*x_ref[1]
        ref[2,0,1,1]=(1.03340508201)*x_ref[0]+(-1.15516978578)*x_ref[1]
        ref[2,0,2,0]=(0.0791763687033)*x_ref[0]+(-0.0239321376253)*x_ref[1]
        ref[2,0,2,1]=(-0.0296109649235)*x_ref[0]+(-0.292837031166)*x_ref[1]
        ref[2,1,0,0]=(-0.00911908859083)*x_ref[0]+(-1.17279969051)*x_ref[1]
        ref[2,1,0,1]=(-1.06648321801)*x_ref[0]+(0.126622301021)*x_ref[1]
        ref[2,1,1,0]=(0.275881626889)*x_ref[0]+(1.69565501602)*x_ref[1]
        ref[2,1,1,1]=(-0.409495321219)*x_ref[0]+(-0.407778373033)*x_ref[1]
        ref[2,1,2,0]=(1.33276969465)*x_ref[0]+(-1.17532941128)*x_ref[1]
        ref[2,1,2,1]=(-0.866150634496)*x_ref[0]+(0.797755014562)*x_ref[1]
        ref[2,2,0,0]=(-0.106882476852)*x_ref[0]+(-0.249249128227)*x_ref[1]
        ref[2,2,0,1]=(-0.520287837168)*x_ref[0]+(0.338818467136)*x_ref[1]
        ref[2,2,1,0]=(-0.852757899877)*x_ref[0]+(0.321704892591)*x_ref[1]
        ref[2,2,1,1]=(0.142520864112)*x_ref[0]+(-1.05840136269)*x_ref[1]
        ref[2,2,2,0]=(-0.715836025391)*x_ref[0]+(0.22306435555)*x_ref[1]
        ref[2,2,2,1]=(-0.586149670377)*x_ref[0]+(-0.516267803245)*x_ref[1]
        ref[2,3,0,0]=(-0.950483491273)*x_ref[0]+(0.624915645607)*x_ref[1]
        ref[2,3,0,1]=(-0.217697021618)*x_ref[0]+(0.0442099818289)*x_ref[1]
        ref[2,3,1,0]=(-1.01046266386)*x_ref[0]+(-1.48154228284)*x_ref[1]
        ref[2,3,1,1]=(-1.3050947534)*x_ref[0]+(-0.998384196826)*x_ref[1]
        ref[2,3,2,0]=(0.267414868071)*x_ref[0]+(1.15992550636)*x_ref[1]
        ref[2,3,2,1]=(-0.862214432382)*x_ref[0]+(0.292670280728)*x_ref[1]
        ref[2,4,0,0]=(-0.134428442874)*x_ref[0]+(0.719512272598)*x_ref[1]
        ref[2,4,0,1]=(-1.67096566574)*x_ref[0]+(0.43067843233)*x_ref[1]
        ref[2,4,1,0]=(0.188806295154)*x_ref[0]+(-1.54252342316)*x_ref[1]
        ref[2,4,1,1]=(0.0976229597501)*x_ref[0]+(-0.937051784938)*x_ref[1]
        ref[2,4,2,0]=(-0.00596882510121)*x_ref[0]+(-0.312957453878)*x_ref[1]
        ref[2,4,2,1]=(-1.02156076253)*x_ref[0]+(-0.896718827796)*x_ref[1]
        ref[3,0,0,0]=(0.0133947002051)*x_ref[0]+(0.238649243438)*x_ref[1]
        ref[3,0,0,1]=(1.00467946164)*x_ref[0]+(1.34133983171)*x_ref[1]
        ref[3,0,1,0]=(0.341713854404)*x_ref[0]+(0.511010566842)*x_ref[1]
        ref[3,0,1,1]=(0.546165677394)*x_ref[0]+(0.321446986518)*x_ref[1]
        ref[3,0,2,0]=(-0.764072559141)*x_ref[0]+(-0.312532010803)*x_ref[1]
        ref[3,0,2,1]=(-0.811247938505)*x_ref[0]+(-0.939324734555)*x_ref[1]
        ref[3,1,0,0]=(0.572111068237)*x_ref[0]+(-0.0530416981116)*x_ref[1]
        ref[3,1,0,1]=(0.837385483064)*x_ref[0]+(0.928431354228)*x_ref[1]
        ref[3,1,1,0]=(0.238702836098)*x_ref[0]+(0.182536052546)*x_ref[1]
        ref[3,1,1,1]=(0.461955717914)*x_ref[0]+(0.0256860825024)*x_ref[1]
        ref[3,1,2,0]=(-0.787501340181)*x_ref[0]+(-0.828497466795)*x_ref[1]
        ref[3,1,2,1]=(0.348032537213)*x_ref[0]+(-0.984456869176)*x_ref[1]
        ref[3,2,0,0]=(0.588847373206)*x_ref[0]+(1.84398386935)*x_ref[1]
        ref[3,2,0,1]=(-0.2085469075)*x_ref[0]+(-0.244926359847)*x_ref[1]
        ref[3,2,1,0]=(-1.62397953336)*x_ref[0]+(-0.102475340253)*x_ref[1]
        ref[3,2,1,1]=(0.455695325345)*x_ref[0]+(0.552529049227)*x_ref[1]
        ref[3,2,2,0]=(0.734375962799)*x_ref[0]+(0.279367642817)*x_ref[1]
        ref[3,2,2,1]=(1.52878551603)*x_ref[0]+(-1.43594288923)*x_ref[1]
        ref[3,3,0,0]=(-0.331774589719)*x_ref[0]+(-0.852014090099)*x_ref[1]
        ref[3,3,0,1]=(0.0495293070849)*x_ref[0]+(1.70225124669)*x_ref[1]
        ref[3,3,1,0]=(1.37894662758)*x_ref[0]+(0.350193222832)*x_ref[1]
        ref[3,3,1,1]=(0.00570139162378)*x_ref[0]+(-0.37112295664)*x_ref[1]
        ref[3,3,2,0]=(-0.205033339669)*x_ref[0]+(-0.695461477152)*x_ref[1]
        ref[3,3,2,1]=(0.12868380026)*x_ref[0]+(0.764317872049)*x_ref[1]
        ref[3,4,0,0]=(-1.04309352143)*x_ref[0]+(-0.805487628821)*x_ref[1]
        ref[3,4,0,1]=(-0.496744676221)*x_ref[0]+(-0.13812443637)*x_ref[1]
        ref[3,4,1,0]=(0.0339710647687)*x_ref[0]+(0.0274093213606)*x_ref[1]
        ref[3,4,1,1]=(-0.516837676304)*x_ref[0]+(-0.119504191446)*x_ref[1]
        ref[3,4,2,0]=(-0.472643263791)*x_ref[0]+(-0.337262686677)*x_ref[1]
        ref[3,4,2,1]=(-1.34135241071)*x_ref[0]+(-0.185936541491)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.0655735713454)*x[0]+(-0.69619555181)*x[1]+(0.881033804166)*x[2]
        s[0,0,0,1]=(-0.886101209867)*x[0]+(0.0852736523075)*x[1]+(0.514766114136)*x[2]
        s[0,0,1,0]=(-0.708903115024)*x[0]+(-0.949542800872)*x[1]+(-1.07763550768)*x[2]
        s[0,0,1,1]=(0.970797729322)*x[0]+(-0.237719575345)*x[1]+(1.54252127667)*x[2]
        s[0,0,2,0]=(-0.112910964189)*x[0]+(0.813990286281)*x[1]+(0.437139526752)*x[2]
        s[0,0,2,1]=(0.270076285201)*x[0]+(-1.04231968668)*x[1]+(1.08793884214)*x[2]
        s[0,1,0,0]=(-0.376091612547)*x[0]+(0.965045498929)*x[1]+(-0.516825182321)*x[2]
        s[0,1,0,1]=(-0.922174392864)*x[0]+(1.12745699436)*x[1]+(0.647593149683)*x[2]
        s[0,1,1,0]=(-1.62204275282)*x[0]+(0.64585305798)*x[1]+(-1.41457161128)*x[2]
        s[0,1,1,1]=(-0.412715098923)*x[0]+(0.00317118495093)*x[1]+(-0.848053699577)*x[2]
        s[0,1,2,0]=(-1.33788126072)*x[0]+(-0.370246465439)*x[1]+(0.149368781369)*x[2]
        s[0,1,2,1]=(-0.0432149867078)*x[0]+(0.20428198228)*x[1]+(0.685699546484)*x[2]
        s[0,2,0,0]=(-0.849779864072)*x[0]+(-0.465408292856)*x[1]+(0.12942183587)*x[2]
        s[0,2,0,1]=(-0.533958302867)*x[0]+(0.742004817382)*x[1]+(-0.146827358028)*x[2]
        s[0,2,1,0]=(-0.100072250083)*x[0]+(-1.24645970838)*x[1]+(1.45346380514)*x[2]
        s[0,2,1,1]=(-0.718115035714)*x[0]+(0.572444837212)*x[1]+(-0.0961841883909)*x[2]
        s[0,2,2,0]=(-1.6396722492)*x[0]+(-0.657016455181)*x[1]+(-0.52001961538)*x[2]
        s[0,2,2,1]=(-0.0161596921482)*x[0]+(-0.130395174634)*x[1]+(1.68490335057)*x[2]
        s[0,3,0,0]=(0.16054789145)*x[0]+(-1.35250823955)*x[1]+(0.663123269205)*x[2]
        s[0,3,0,1]=(0.307153218553)*x[0]+(0.705376328926)*x[1]+(-0.335478576645)*x[2]
        s[0,3,1,0]=(0.499629878974)*x[0]+(-0.389863260213)*x[1]+(-0.398638918232)*x[2]
        s[0,3,1,1]=(0.229760417346)*x[0]+(-0.382704654178)*x[1]+(0.044859414166)*x[2]
        s[0,3,2,0]=(0.402905751731)*x[0]+(-0.433748870487)*x[1]+(1.05246387058)*x[2]
        s[0,3,2,1]=(0.620966047942)*x[0]+(0.778166847286)*x[1]+(-0.684881429118)*x[2]
        s[0,4,0,0]=(-0.0429218580761)*x[0]+(0.263402329307)*x[1]+(0.314852944534)*x[2]
        s[0,4,0,1]=(0.112989114824)*x[0]+(-0.407308926569)*x[1]+(-0.581889344395)*x[2]
        s[0,4,1,0]=(1.46986587487)*x[0]+(-0.360451075631)*x[1]+(1.41473323373)*x[2]
        s[0,4,1,1]=(-0.0898945660646)*x[0]+(-1.24644050789)*x[1]+(-0.344493513462)*x[2]
        s[0,4,2,0]=(-0.383865368385)*x[0]+(0.758209662119)*x[1]+(0.737714808779)*x[2]
        s[0,4,2,1]=(-0.282690133746)*x[0]+(0.520513691755)*x[1]+(-0.38353403004)*x[2]
        s[1,0,0,0]=(-0.233823433606)*x[0]+(0.627115553859)*x[1]+(0.240814969095)*x[2]
        s[1,0,0,1]=(1.65536823794)*x[0]+(0.577386580179)*x[1]+(0.64299358654)*x[2]
        s[1,0,1,0]=(-0.304849514801)*x[0]+(-1.26879113223)*x[1]+(-0.257428355176)*x[2]
        s[1,0,1,1]=(0.313165425271)*x[0]+(0.648645104314)*x[1]+(0.152429094691)*x[2]
        s[1,0,2,0]=(-1.18245997349)*x[0]+(-0.990110942468)*x[1]+(-0.528372216957)*x[2]
        s[1,0,2,1]=(-0.497751288211)*x[0]+(1.27633543739)*x[1]+(1.29106845801)*x[2]
        s[1,1,0,0]=(-0.401807344422)*x[0]+(1.00839022632)*x[1]+(0.855275309873)*x[2]
        s[1,1,0,1]=(-1.22469413181)*x[0]+(0.946705273705)*x[1]+(1.02166037335)*x[2]
        s[1,1,1,0]=(0.0800997438684)*x[0]+(0.232002850135)*x[1]+(1.72304848773)*x[2]
        s[1,1,1,1]=(-1.14517485603)*x[0]+(-1.68921566046)*x[1]+(0.297914580394)*x[2]
        s[1,1,2,0]=(-0.925214063847)*x[0]+(-1.31725976684)*x[1]+(-1.15890525345)*x[2]
        s[1,1,2,1]=(-0.0211857989738)*x[0]+(-0.45544304451)*x[1]+(-1.69006065778)*x[2]
        s[1,2,0,0]=(0.545466934146)*x[0]+(-0.661570837411)*x[1]+(-0.387531002262)*x[2]
        s[1,2,0,1]=(-0.978830810453)*x[0]+(0.958890753286)*x[1]+(-0.3992633515)*x[2]
        s[1,2,1,0]=(0.395262411038)*x[0]+(0.3119070642)*x[1]+(0.544660098992)*x[2]
        s[1,2,1,1]=(-1.07426944365)*x[0]+(0.223216516043)*x[1]+(0.589789766065)*x[2]
        s[1,2,2,0]=(0.0742935523479)*x[0]+(1.20227049003)*x[1]+(0.134941956037)*x[2]
        s[1,2,2,1]=(1.02997244989)*x[0]+(1.24846778966)*x[1]+(-1.76817806687)*x[2]
        s[1,3,0,0]=(-0.0753143898819)*x[0]+(-0.138231791536)*x[1]+(-1.06788473564)*x[2]
        s[1,3,0,1]=(-1.00651208369)*x[0]+(-0.104956186368)*x[1]+(-0.51036966316)*x[2]
        s[1,3,1,0]=(-0.221214050083)*x[0]+(0.227150982313)*x[1]+(1.08579667343)*x[2]
        s[1,3,1,1]=(0.510525992628)*x[0]+(1.49179453941)*x[1]+(-0.0676345340155)*x[2]
        s[1,3,2,0]=(0.917441023004)*x[0]+(0.51231105646)*x[1]+(-0.620649550464)*x[2]
        s[1,3,2,1]=(-1.10288875837)*x[0]+(-0.142710630558)*x[1]+(0.623821771889)*x[2]
        s[1,4,0,0]=(-0.255410535992)*x[0]+(0.530829574503)*x[1]+(0.775272541509)*x[2]
        s[1,4,0,1]=(0.284567984278)*x[0]+(0.125859042954)*x[1]+(0.533462070129)*x[2]
        s[1,4,1,0]=(-0.508598664109)*x[0]+(-1.40877217657)*x[1]+(-0.484415496121)*x[2]
        s[1,4,1,1]=(0.556787007449)*x[0]+(-0.116886679539)*x[1]+(1.00640244397)*x[2]
        s[1,4,2,0]=(0.317366265155)*x[0]+(-1.27232976963)*x[1]+(-0.114651450491)*x[2]
        s[1,4,2,1]=(-1.28370651742)*x[0]+(-0.362861660348)*x[1]+(-0.0962080517851)*x[2]
        s[2,0,0,0]=(0.271907114067)*x[0]+(0.265344846384)*x[1]+(0.0924933201796)*x[2]
        s[2,0,0,1]=(-1.58861288248)*x[0]+(0.22963952324)*x[1]+(-0.653279584852)*x[2]
        s[2,0,1,0]=(0.944381839359)*x[0]+(-0.302082255723)*x[1]+(-1.81040568842)*x[2]
        s[2,0,1,1]=(-0.650306656954)*x[0]+(-0.0845826799882)*x[1]+(-1.25732622868)*x[2]
        s[2,0,2,0]=(-0.0135274380735)*x[0]+(-1.2176048332)*x[1]+(-0.330033040382)*x[2]
        s[2,0,2,1]=(0.8254202497)*x[0]+(0.108602990536)*x[1]+(-0.538593746905)*x[2]
        s[2,1,0,0]=(-0.586282432139)*x[0]+(-0.306182725701)*x[1]+(0.433425406694)*x[2]
        s[2,1,0,1]=(-0.252073403002)*x[0]+(0.106710584551)*x[1]+(-0.125992533885)*x[2]
        s[2,1,1,0]=(0.424274043808)*x[0]+(-0.987960051363)*x[1]+(0.602880255366)*x[2]
        s[2,1,1,1]=(-1.18215444375)*x[0]+(-1.0587178922)*x[1]+(1.07462469891)*x[2]
        s[2,1,2,0]=(-1.21183003522)*x[0]+(0.806502177052)*x[1]+(0.396337647649)*x[2]
        s[2,1,2,1]=(0.47346358421)*x[0]+(1.19060722938)*x[1]+(-0.830743235441)*x[2]
        s[2,2,0,0]=(0.394593530336)*x[0]+(-0.488087761788)*x[1]+(0.0214792800471)*x[2]
        s[2,2,0,1]=(-0.959164075423)*x[0]+(-1.29228916512)*x[1]+(-0.299706293498)*x[2]
        s[2,2,1,0]=(-0.458184497165)*x[0]+(-1.35468397098)*x[1]+(1.08155385172)*x[2]
        s[2,2,1,1]=(-0.543174665411)*x[0]+(-0.546979238048)*x[1]+(-0.00870655686839)*x[2]
        s[2,2,2,0]=(1.04289134262)*x[0]+(-0.795937095748)*x[1]+(0.437742223432)*x[2]
        s[2,2,2,1]=(0.987908115924)*x[0]+(0.717930289783)*x[1]+(-0.371823996424)*x[2]
        s[2,3,0,0]=(-0.169000222452)*x[0]+(-1.17983694078)*x[1]+(-0.525903072845)*x[2]
        s[2,3,0,1]=(-1.18221084142)*x[0]+(-0.723856791479)*x[1]+(-0.0735764242849)*x[2]
        s[2,3,1,0]=(-1.23369418263)*x[0]+(-1.61175404993)*x[1]+(-0.38186249552)*x[2]
        s[2,3,1,1]=(-0.638043359549)*x[0]+(-0.711623038143)*x[1]+(1.37146468285)*x[2]
        s[2,3,2,0]=(0.231534056605)*x[0]+(-0.468413103344)*x[1]+(0.181817447577)*x[2]
        s[2,3,2,1]=(-0.204204815383)*x[0]+(0.426315456914)*x[1]+(0.112813411509)*x[2]
        s[2,4,0,0]=(0.358495614116)*x[0]+(0.530612217508)*x[1]+(-1.18131247911)*x[2]
        s[2,4,0,1]=(-1.47421904004)*x[0]+(-0.135887217467)*x[1]+(1.1526366976)*x[2]
        s[2,4,1,0]=(1.05475775091)*x[0]+(-1.14201019549)*x[1]+(-0.00147054925041)*x[2]
        s[2,4,1,1]=(-0.845002383828)*x[0]+(1.27489434704)*x[1]+(-0.0759015370413)*x[2]
        s[2,4,2,0]=(-1.78781582559)*x[0]+(0.615399643816)*x[1]+(0.943547107618)*x[2]
        s[2,4,2,1]=(-0.269588308338)*x[0]+(0.582501413736)*x[1]+(0.309865222562)*x[2]
        s[3,0,0,0]=(0.0339770002586)*x[0]+(-1.05588139882)*x[1]+(-0.350074839678)*x[2]
        s[3,0,0,1]=(1.26035921199)*x[0]+(-1.44475344932)*x[1]+(0.230467305805)*x[2]
        s[3,0,1,0]=(-0.212637505994)*x[0]+(0.836913336266)*x[1]+(0.782933559258)*x[2]
        s[3,0,1,1]=(0.350804832921)*x[0]+(-0.711735663328)*x[1]+(-0.7055885795)*x[2]
        s[3,0,2,0]=(1.15262669529)*x[0]+(-0.609133167001)*x[1]+(-0.0481345387977)*x[2]
        s[3,0,2,1]=(-0.328624274521)*x[0]+(-0.1435975351)*x[1]+(1.1531857455)*x[2]
        s[3,1,0,0]=(0.0574967918966)*x[0]+(-0.892284893383)*x[1]+(0.939965393644)*x[2]
        s[3,1,0,1]=(0.533477118513)*x[0]+(-0.396195978184)*x[1]+(1.40017632877)*x[2]
        s[3,1,1,0]=(-0.503230088158)*x[0]+(-0.90636136387)*x[1]+(0.0881754805722)*x[2]
        s[3,1,1,1]=(1.98755626341)*x[0]+(0.807046266085)*x[1]+(-0.257226825918)*x[2]
        s[3,1,2,0]=(0.589818915815)*x[0]+(0.978508047098)*x[1]+(0.155263385992)*x[2]
        s[3,1,2,1]=(0.134507669198)*x[0]+(-0.0636645368891)*x[1]+(0.615504770845)*x[2]
        s[3,2,0,0]=(0.930104004822)*x[0]+(-0.382205774724)*x[1]+(0.668127591046)*x[2]
        s[3,2,0,1]=(-0.190422928513)*x[0]+(0.293413650478)*x[1]+(-0.505554260659)*x[2]
        s[3,2,1,0]=(-1.2934566905)*x[0]+(0.76603122212)*x[1]+(1.20584140236)*x[2]
        s[3,2,1,1]=(0.114254707378)*x[0]+(-0.0342824618295)*x[1]+(1.52172882565)*x[2]
        s[3,2,2,0]=(0.248047215938)*x[0]+(-0.978266687903)*x[1]+(-1.30670985993)*x[2]
        s[3,2,2,1]=(0.300650838074)*x[0]+(-0.220707794615)*x[1]+(-0.30715962416)*x[2]
        s[3,3,0,0]=(0.448483391935)*x[0]+(-0.439040526715)*x[1]+(-0.528237104437)*x[2]
        s[3,3,0,1]=(0.0927811849092)*x[0]+(-0.0543684949756)*x[1]+(-0.360230052054)*x[2]
        s[3,3,1,0]=(0.24809679767)*x[0]+(-1.2533905798)*x[1]+(-0.492657878912)*x[2]
        s[3,3,1,1]=(0.191376087122)*x[0]+(-0.897239039489)*x[1]+(1.40598336079)*x[2]
        s[3,3,2,0]=(-1.63824093083)*x[0]+(-1.77338068521)*x[1]+(0.454764854525)*x[2]
        s[3,3,2,1]=(1.24005155499)*x[0]+(0.259110200765)*x[1]+(-0.0880271632004)*x[2]
        s[3,4,0,0]=(-1.54899613736)*x[0]+(0.803173298876)*x[1]+(0.429702060218)*x[2]
        s[3,4,0,1]=(-1.31381922671)*x[0]+(-0.258985809712)*x[1]+(-0.769254247799)*x[2]
        s[3,4,1,0]=(-0.913527471875)*x[0]+(0.534597194574)*x[1]+(-0.937480880823)*x[2]
        s[3,4,1,1]=(0.516591744517)*x[0]+(-0.640572168797)*x[1]+(-1.24601319692)*x[2]
        s[3,4,2,0]=(0.349486782323)*x[0]+(1.25410491943)*x[1]+(0.145174151067)*x[2]
        s[3,4,2,1]=(-0.799350818525)*x[0]+(-0.505153470695)*x[1]+(-1.20903427385)*x[2]
        ref[0,0,0,0]=(-0.0655735713454)*x_ref[0]+(-0.69619555181)*x_ref[1]+(0.881033804166)*x_ref[2]
        ref[0,0,0,1]=(-0.886101209867)*x_ref[0]+(0.0852736523075)*x_ref[1]+(0.514766114136)*x_ref[2]
        ref[0,0,1,0]=(-0.708903115024)*x_ref[0]+(-0.949542800872)*x_ref[1]+(-1.07763550768)*x_ref[2]
        ref[0,0,1,1]=(0.970797729322)*x_ref[0]+(-0.237719575345)*x_ref[1]+(1.54252127667)*x_ref[2]
        ref[0,0,2,0]=(-0.112910964189)*x_ref[0]+(0.813990286281)*x_ref[1]+(0.437139526752)*x_ref[2]
        ref[0,0,2,1]=(0.270076285201)*x_ref[0]+(-1.04231968668)*x_ref[1]+(1.08793884214)*x_ref[2]
        ref[0,1,0,0]=(-0.376091612547)*x_ref[0]+(0.965045498929)*x_ref[1]+(-0.516825182321)*x_ref[2]
        ref[0,1,0,1]=(-0.922174392864)*x_ref[0]+(1.12745699436)*x_ref[1]+(0.647593149683)*x_ref[2]
        ref[0,1,1,0]=(-1.62204275282)*x_ref[0]+(0.64585305798)*x_ref[1]+(-1.41457161128)*x_ref[2]
        ref[0,1,1,1]=(-0.412715098923)*x_ref[0]+(0.00317118495093)*x_ref[1]+(-0.848053699577)*x_ref[2]
        ref[0,1,2,0]=(-1.33788126072)*x_ref[0]+(-0.370246465439)*x_ref[1]+(0.149368781369)*x_ref[2]
        ref[0,1,2,1]=(-0.0432149867078)*x_ref[0]+(0.20428198228)*x_ref[1]+(0.685699546484)*x_ref[2]
        ref[0,2,0,0]=(-0.849779864072)*x_ref[0]+(-0.465408292856)*x_ref[1]+(0.12942183587)*x_ref[2]
        ref[0,2,0,1]=(-0.533958302867)*x_ref[0]+(0.742004817382)*x_ref[1]+(-0.146827358028)*x_ref[2]
        ref[0,2,1,0]=(-0.100072250083)*x_ref[0]+(-1.24645970838)*x_ref[1]+(1.45346380514)*x_ref[2]
        ref[0,2,1,1]=(-0.718115035714)*x_ref[0]+(0.572444837212)*x_ref[1]+(-0.0961841883909)*x_ref[2]
        ref[0,2,2,0]=(-1.6396722492)*x_ref[0]+(-0.657016455181)*x_ref[1]+(-0.52001961538)*x_ref[2]
        ref[0,2,2,1]=(-0.0161596921482)*x_ref[0]+(-0.130395174634)*x_ref[1]+(1.68490335057)*x_ref[2]
        ref[0,3,0,0]=(0.16054789145)*x_ref[0]+(-1.35250823955)*x_ref[1]+(0.663123269205)*x_ref[2]
        ref[0,3,0,1]=(0.307153218553)*x_ref[0]+(0.705376328926)*x_ref[1]+(-0.335478576645)*x_ref[2]
        ref[0,3,1,0]=(0.499629878974)*x_ref[0]+(-0.389863260213)*x_ref[1]+(-0.398638918232)*x_ref[2]
        ref[0,3,1,1]=(0.229760417346)*x_ref[0]+(-0.382704654178)*x_ref[1]+(0.044859414166)*x_ref[2]
        ref[0,3,2,0]=(0.402905751731)*x_ref[0]+(-0.433748870487)*x_ref[1]+(1.05246387058)*x_ref[2]
        ref[0,3,2,1]=(0.620966047942)*x_ref[0]+(0.778166847286)*x_ref[1]+(-0.684881429118)*x_ref[2]
        ref[0,4,0,0]=(-0.0429218580761)*x_ref[0]+(0.263402329307)*x_ref[1]+(0.314852944534)*x_ref[2]
        ref[0,4,0,1]=(0.112989114824)*x_ref[0]+(-0.407308926569)*x_ref[1]+(-0.581889344395)*x_ref[2]
        ref[0,4,1,0]=(1.46986587487)*x_ref[0]+(-0.360451075631)*x_ref[1]+(1.41473323373)*x_ref[2]
        ref[0,4,1,1]=(-0.0898945660646)*x_ref[0]+(-1.24644050789)*x_ref[1]+(-0.344493513462)*x_ref[2]
        ref[0,4,2,0]=(-0.383865368385)*x_ref[0]+(0.758209662119)*x_ref[1]+(0.737714808779)*x_ref[2]
        ref[0,4,2,1]=(-0.282690133746)*x_ref[0]+(0.520513691755)*x_ref[1]+(-0.38353403004)*x_ref[2]
        ref[1,0,0,0]=(-0.233823433606)*x_ref[0]+(0.627115553859)*x_ref[1]+(0.240814969095)*x_ref[2]
        ref[1,0,0,1]=(1.65536823794)*x_ref[0]+(0.577386580179)*x_ref[1]+(0.64299358654)*x_ref[2]
        ref[1,0,1,0]=(-0.304849514801)*x_ref[0]+(-1.26879113223)*x_ref[1]+(-0.257428355176)*x_ref[2]
        ref[1,0,1,1]=(0.313165425271)*x_ref[0]+(0.648645104314)*x_ref[1]+(0.152429094691)*x_ref[2]
        ref[1,0,2,0]=(-1.18245997349)*x_ref[0]+(-0.990110942468)*x_ref[1]+(-0.528372216957)*x_ref[2]
        ref[1,0,2,1]=(-0.497751288211)*x_ref[0]+(1.27633543739)*x_ref[1]+(1.29106845801)*x_ref[2]
        ref[1,1,0,0]=(-0.401807344422)*x_ref[0]+(1.00839022632)*x_ref[1]+(0.855275309873)*x_ref[2]
        ref[1,1,0,1]=(-1.22469413181)*x_ref[0]+(0.946705273705)*x_ref[1]+(1.02166037335)*x_ref[2]
        ref[1,1,1,0]=(0.0800997438684)*x_ref[0]+(0.232002850135)*x_ref[1]+(1.72304848773)*x_ref[2]
        ref[1,1,1,1]=(-1.14517485603)*x_ref[0]+(-1.68921566046)*x_ref[1]+(0.297914580394)*x_ref[2]
        ref[1,1,2,0]=(-0.925214063847)*x_ref[0]+(-1.31725976684)*x_ref[1]+(-1.15890525345)*x_ref[2]
        ref[1,1,2,1]=(-0.0211857989738)*x_ref[0]+(-0.45544304451)*x_ref[1]+(-1.69006065778)*x_ref[2]
        ref[1,2,0,0]=(0.545466934146)*x_ref[0]+(-0.661570837411)*x_ref[1]+(-0.387531002262)*x_ref[2]
        ref[1,2,0,1]=(-0.978830810453)*x_ref[0]+(0.958890753286)*x_ref[1]+(-0.3992633515)*x_ref[2]
        ref[1,2,1,0]=(0.395262411038)*x_ref[0]+(0.3119070642)*x_ref[1]+(0.544660098992)*x_ref[2]
        ref[1,2,1,1]=(-1.07426944365)*x_ref[0]+(0.223216516043)*x_ref[1]+(0.589789766065)*x_ref[2]
        ref[1,2,2,0]=(0.0742935523479)*x_ref[0]+(1.20227049003)*x_ref[1]+(0.134941956037)*x_ref[2]
        ref[1,2,2,1]=(1.02997244989)*x_ref[0]+(1.24846778966)*x_ref[1]+(-1.76817806687)*x_ref[2]
        ref[1,3,0,0]=(-0.0753143898819)*x_ref[0]+(-0.138231791536)*x_ref[1]+(-1.06788473564)*x_ref[2]
        ref[1,3,0,1]=(-1.00651208369)*x_ref[0]+(-0.104956186368)*x_ref[1]+(-0.51036966316)*x_ref[2]
        ref[1,3,1,0]=(-0.221214050083)*x_ref[0]+(0.227150982313)*x_ref[1]+(1.08579667343)*x_ref[2]
        ref[1,3,1,1]=(0.510525992628)*x_ref[0]+(1.49179453941)*x_ref[1]+(-0.0676345340155)*x_ref[2]
        ref[1,3,2,0]=(0.917441023004)*x_ref[0]+(0.51231105646)*x_ref[1]+(-0.620649550464)*x_ref[2]
        ref[1,3,2,1]=(-1.10288875837)*x_ref[0]+(-0.142710630558)*x_ref[1]+(0.623821771889)*x_ref[2]
        ref[1,4,0,0]=(-0.255410535992)*x_ref[0]+(0.530829574503)*x_ref[1]+(0.775272541509)*x_ref[2]
        ref[1,4,0,1]=(0.284567984278)*x_ref[0]+(0.125859042954)*x_ref[1]+(0.533462070129)*x_ref[2]
        ref[1,4,1,0]=(-0.508598664109)*x_ref[0]+(-1.40877217657)*x_ref[1]+(-0.484415496121)*x_ref[2]
        ref[1,4,1,1]=(0.556787007449)*x_ref[0]+(-0.116886679539)*x_ref[1]+(1.00640244397)*x_ref[2]
        ref[1,4,2,0]=(0.317366265155)*x_ref[0]+(-1.27232976963)*x_ref[1]+(-0.114651450491)*x_ref[2]
        ref[1,4,2,1]=(-1.28370651742)*x_ref[0]+(-0.362861660348)*x_ref[1]+(-0.0962080517851)*x_ref[2]
        ref[2,0,0,0]=(0.271907114067)*x_ref[0]+(0.265344846384)*x_ref[1]+(0.0924933201796)*x_ref[2]
        ref[2,0,0,1]=(-1.58861288248)*x_ref[0]+(0.22963952324)*x_ref[1]+(-0.653279584852)*x_ref[2]
        ref[2,0,1,0]=(0.944381839359)*x_ref[0]+(-0.302082255723)*x_ref[1]+(-1.81040568842)*x_ref[2]
        ref[2,0,1,1]=(-0.650306656954)*x_ref[0]+(-0.0845826799882)*x_ref[1]+(-1.25732622868)*x_ref[2]
        ref[2,0,2,0]=(-0.0135274380735)*x_ref[0]+(-1.2176048332)*x_ref[1]+(-0.330033040382)*x_ref[2]
        ref[2,0,2,1]=(0.8254202497)*x_ref[0]+(0.108602990536)*x_ref[1]+(-0.538593746905)*x_ref[2]
        ref[2,1,0,0]=(-0.586282432139)*x_ref[0]+(-0.306182725701)*x_ref[1]+(0.433425406694)*x_ref[2]
        ref[2,1,0,1]=(-0.252073403002)*x_ref[0]+(0.106710584551)*x_ref[1]+(-0.125992533885)*x_ref[2]
        ref[2,1,1,0]=(0.424274043808)*x_ref[0]+(-0.987960051363)*x_ref[1]+(0.602880255366)*x_ref[2]
        ref[2,1,1,1]=(-1.18215444375)*x_ref[0]+(-1.0587178922)*x_ref[1]+(1.07462469891)*x_ref[2]
        ref[2,1,2,0]=(-1.21183003522)*x_ref[0]+(0.806502177052)*x_ref[1]+(0.396337647649)*x_ref[2]
        ref[2,1,2,1]=(0.47346358421)*x_ref[0]+(1.19060722938)*x_ref[1]+(-0.830743235441)*x_ref[2]
        ref[2,2,0,0]=(0.394593530336)*x_ref[0]+(-0.488087761788)*x_ref[1]+(0.0214792800471)*x_ref[2]
        ref[2,2,0,1]=(-0.959164075423)*x_ref[0]+(-1.29228916512)*x_ref[1]+(-0.299706293498)*x_ref[2]
        ref[2,2,1,0]=(-0.458184497165)*x_ref[0]+(-1.35468397098)*x_ref[1]+(1.08155385172)*x_ref[2]
        ref[2,2,1,1]=(-0.543174665411)*x_ref[0]+(-0.546979238048)*x_ref[1]+(-0.00870655686839)*x_ref[2]
        ref[2,2,2,0]=(1.04289134262)*x_ref[0]+(-0.795937095748)*x_ref[1]+(0.437742223432)*x_ref[2]
        ref[2,2,2,1]=(0.987908115924)*x_ref[0]+(0.717930289783)*x_ref[1]+(-0.371823996424)*x_ref[2]
        ref[2,3,0,0]=(-0.169000222452)*x_ref[0]+(-1.17983694078)*x_ref[1]+(-0.525903072845)*x_ref[2]
        ref[2,3,0,1]=(-1.18221084142)*x_ref[0]+(-0.723856791479)*x_ref[1]+(-0.0735764242849)*x_ref[2]
        ref[2,3,1,0]=(-1.23369418263)*x_ref[0]+(-1.61175404993)*x_ref[1]+(-0.38186249552)*x_ref[2]
        ref[2,3,1,1]=(-0.638043359549)*x_ref[0]+(-0.711623038143)*x_ref[1]+(1.37146468285)*x_ref[2]
        ref[2,3,2,0]=(0.231534056605)*x_ref[0]+(-0.468413103344)*x_ref[1]+(0.181817447577)*x_ref[2]
        ref[2,3,2,1]=(-0.204204815383)*x_ref[0]+(0.426315456914)*x_ref[1]+(0.112813411509)*x_ref[2]
        ref[2,4,0,0]=(0.358495614116)*x_ref[0]+(0.530612217508)*x_ref[1]+(-1.18131247911)*x_ref[2]
        ref[2,4,0,1]=(-1.47421904004)*x_ref[0]+(-0.135887217467)*x_ref[1]+(1.1526366976)*x_ref[2]
        ref[2,4,1,0]=(1.05475775091)*x_ref[0]+(-1.14201019549)*x_ref[1]+(-0.00147054925041)*x_ref[2]
        ref[2,4,1,1]=(-0.845002383828)*x_ref[0]+(1.27489434704)*x_ref[1]+(-0.0759015370413)*x_ref[2]
        ref[2,4,2,0]=(-1.78781582559)*x_ref[0]+(0.615399643816)*x_ref[1]+(0.943547107618)*x_ref[2]
        ref[2,4,2,1]=(-0.269588308338)*x_ref[0]+(0.582501413736)*x_ref[1]+(0.309865222562)*x_ref[2]
        ref[3,0,0,0]=(0.0339770002586)*x_ref[0]+(-1.05588139882)*x_ref[1]+(-0.350074839678)*x_ref[2]
        ref[3,0,0,1]=(1.26035921199)*x_ref[0]+(-1.44475344932)*x_ref[1]+(0.230467305805)*x_ref[2]
        ref[3,0,1,0]=(-0.212637505994)*x_ref[0]+(0.836913336266)*x_ref[1]+(0.782933559258)*x_ref[2]
        ref[3,0,1,1]=(0.350804832921)*x_ref[0]+(-0.711735663328)*x_ref[1]+(-0.7055885795)*x_ref[2]
        ref[3,0,2,0]=(1.15262669529)*x_ref[0]+(-0.609133167001)*x_ref[1]+(-0.0481345387977)*x_ref[2]
        ref[3,0,2,1]=(-0.328624274521)*x_ref[0]+(-0.1435975351)*x_ref[1]+(1.1531857455)*x_ref[2]
        ref[3,1,0,0]=(0.0574967918966)*x_ref[0]+(-0.892284893383)*x_ref[1]+(0.939965393644)*x_ref[2]
        ref[3,1,0,1]=(0.533477118513)*x_ref[0]+(-0.396195978184)*x_ref[1]+(1.40017632877)*x_ref[2]
        ref[3,1,1,0]=(-0.503230088158)*x_ref[0]+(-0.90636136387)*x_ref[1]+(0.0881754805722)*x_ref[2]
        ref[3,1,1,1]=(1.98755626341)*x_ref[0]+(0.807046266085)*x_ref[1]+(-0.257226825918)*x_ref[2]
        ref[3,1,2,0]=(0.589818915815)*x_ref[0]+(0.978508047098)*x_ref[1]+(0.155263385992)*x_ref[2]
        ref[3,1,2,1]=(0.134507669198)*x_ref[0]+(-0.0636645368891)*x_ref[1]+(0.615504770845)*x_ref[2]
        ref[3,2,0,0]=(0.930104004822)*x_ref[0]+(-0.382205774724)*x_ref[1]+(0.668127591046)*x_ref[2]
        ref[3,2,0,1]=(-0.190422928513)*x_ref[0]+(0.293413650478)*x_ref[1]+(-0.505554260659)*x_ref[2]
        ref[3,2,1,0]=(-1.2934566905)*x_ref[0]+(0.76603122212)*x_ref[1]+(1.20584140236)*x_ref[2]
        ref[3,2,1,1]=(0.114254707378)*x_ref[0]+(-0.0342824618295)*x_ref[1]+(1.52172882565)*x_ref[2]
        ref[3,2,2,0]=(0.248047215938)*x_ref[0]+(-0.978266687903)*x_ref[1]+(-1.30670985993)*x_ref[2]
        ref[3,2,2,1]=(0.300650838074)*x_ref[0]+(-0.220707794615)*x_ref[1]+(-0.30715962416)*x_ref[2]
        ref[3,3,0,0]=(0.448483391935)*x_ref[0]+(-0.439040526715)*x_ref[1]+(-0.528237104437)*x_ref[2]
        ref[3,3,0,1]=(0.0927811849092)*x_ref[0]+(-0.0543684949756)*x_ref[1]+(-0.360230052054)*x_ref[2]
        ref[3,3,1,0]=(0.24809679767)*x_ref[0]+(-1.2533905798)*x_ref[1]+(-0.492657878912)*x_ref[2]
        ref[3,3,1,1]=(0.191376087122)*x_ref[0]+(-0.897239039489)*x_ref[1]+(1.40598336079)*x_ref[2]
        ref[3,3,2,0]=(-1.63824093083)*x_ref[0]+(-1.77338068521)*x_ref[1]+(0.454764854525)*x_ref[2]
        ref[3,3,2,1]=(1.24005155499)*x_ref[0]+(0.259110200765)*x_ref[1]+(-0.0880271632004)*x_ref[2]
        ref[3,4,0,0]=(-1.54899613736)*x_ref[0]+(0.803173298876)*x_ref[1]+(0.429702060218)*x_ref[2]
        ref[3,4,0,1]=(-1.31381922671)*x_ref[0]+(-0.258985809712)*x_ref[1]+(-0.769254247799)*x_ref[2]
        ref[3,4,1,0]=(-0.913527471875)*x_ref[0]+(0.534597194574)*x_ref[1]+(-0.937480880823)*x_ref[2]
        ref[3,4,1,1]=(0.516591744517)*x_ref[0]+(-0.640572168797)*x_ref[1]+(-1.24601319692)*x_ref[2]
        ref[3,4,2,0]=(0.349486782323)*x_ref[0]+(1.25410491943)*x_ref[1]+(0.145174151067)*x_ref[2]
        ref[3,4,2,1]=(-0.799350818525)*x_ref[0]+(-0.505153470695)*x_ref[1]+(-1.20903427385)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedSolution_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-1.05153531037)*x[0]+(1.5374190245)*x[1]
        ref=(-1.05153531037)*x_ref[0]+(1.5374190245)*x_ref[1]
      else:
        s=(-1.91883881043)*x[0]+(-0.414033171764)*x[1]+(1.96173093572)*x[2]
        ref=(-1.91883881043)*x_ref[0]+(-0.414033171764)*x_ref[1]+(1.96173093572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedSolution_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        s[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
        ref[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        ref[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
      else:
        s[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        s[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
        ref[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        ref[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedSolution_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.581359229868)*x[0]+(-0.405627827004)*x[1]
        s[0,1]=(1.15883572458)*x[0]+(-0.212383708658)*x[1]
        s[0,2]=(-0.675196965148)*x[0]+(-0.0589289579115)*x[1]
        s[0,3]=(-0.466142003406)*x[0]+(1.143298366)*x[1]
        s[0,4]=(0.706585901476)*x[0]+(-0.596587020578)*x[1]
        s[1,0]=(0.150862770285)*x[0]+(1.3486848087)*x[1]
        s[1,1]=(0.694050128425)*x[0]+(0.919192282198)*x[1]
        s[1,2]=(0.123568652464)*x[0]+(-1.1901501926)*x[1]
        s[1,3]=(-0.0147792668086)*x[0]+(-1.49381121537)*x[1]
        s[1,4]=(0.0737831649537)*x[0]+(-0.752053730242)*x[1]
        s[2,0]=(-0.758812518739)*x[0]+(-1.26143124516)*x[1]
        s[2,1]=(0.507459828891)*x[0]+(-0.0315665691912)*x[1]
        s[2,2]=(-0.492645010187)*x[0]+(-0.649596364388)*x[1]
        s[2,3]=(-0.0632358844144)*x[0]+(0.899340082527)*x[1]
        s[2,4]=(-0.763180407153)*x[0]+(-0.269808126268)*x[1]
        s[3,0]=(-1.5724790938)*x[0]+(1.16942939992)*x[1]
        s[3,1]=(0.759317295371)*x[0]+(0.14287823082)*x[1]
        s[3,2]=(0.70712973662)*x[0]+(1.08455803224)*x[1]
        s[3,3]=(-0.736730602555)*x[0]+(-0.183785007602)*x[1]
        s[3,4]=(-0.950538745165)*x[0]+(0.294140958329)*x[1]
        ref[0,0]=(-0.581359229868)*x_ref[0]+(-0.405627827004)*x_ref[1]
        ref[0,1]=(1.15883572458)*x_ref[0]+(-0.212383708658)*x_ref[1]
        ref[0,2]=(-0.675196965148)*x_ref[0]+(-0.0589289579115)*x_ref[1]
        ref[0,3]=(-0.466142003406)*x_ref[0]+(1.143298366)*x_ref[1]
        ref[0,4]=(0.706585901476)*x_ref[0]+(-0.596587020578)*x_ref[1]
        ref[1,0]=(0.150862770285)*x_ref[0]+(1.3486848087)*x_ref[1]
        ref[1,1]=(0.694050128425)*x_ref[0]+(0.919192282198)*x_ref[1]
        ref[1,2]=(0.123568652464)*x_ref[0]+(-1.1901501926)*x_ref[1]
        ref[1,3]=(-0.0147792668086)*x_ref[0]+(-1.49381121537)*x_ref[1]
        ref[1,4]=(0.0737831649537)*x_ref[0]+(-0.752053730242)*x_ref[1]
        ref[2,0]=(-0.758812518739)*x_ref[0]+(-1.26143124516)*x_ref[1]
        ref[2,1]=(0.507459828891)*x_ref[0]+(-0.0315665691912)*x_ref[1]
        ref[2,2]=(-0.492645010187)*x_ref[0]+(-0.649596364388)*x_ref[1]
        ref[2,3]=(-0.0632358844144)*x_ref[0]+(0.899340082527)*x_ref[1]
        ref[2,4]=(-0.763180407153)*x_ref[0]+(-0.269808126268)*x_ref[1]
        ref[3,0]=(-1.5724790938)*x_ref[0]+(1.16942939992)*x_ref[1]
        ref[3,1]=(0.759317295371)*x_ref[0]+(0.14287823082)*x_ref[1]
        ref[3,2]=(0.70712973662)*x_ref[0]+(1.08455803224)*x_ref[1]
        ref[3,3]=(-0.736730602555)*x_ref[0]+(-0.183785007602)*x_ref[1]
        ref[3,4]=(-0.950538745165)*x_ref[0]+(0.294140958329)*x_ref[1]
      else:
        s[0,0]=(0.310668873066)*x[0]+(1.66326709742)*x[1]+(1.10140064967)*x[2]
        s[0,1]=(0.46962350981)*x[0]+(-1.02292998305)*x[1]+(0.919666772954)*x[2]
        s[0,2]=(-1.18114779319)*x[0]+(1.23819669136)*x[1]+(0.0622519749551)*x[2]
        s[0,3]=(1.08356504043)*x[0]+(0.450517408398)*x[1]+(-0.528809988127)*x[2]
        s[0,4]=(-0.179349520643)*x[0]+(0.192827326205)*x[1]+(-0.543850326813)*x[2]
        s[1,0]=(-1.60006459477)*x[0]+(0.800497908511)*x[1]+(0.745234766014)*x[2]
        s[1,1]=(0.667355662535)*x[0]+(0.800303765538)*x[1]+(-1.17397311427)*x[2]
        s[1,2]=(0.19342327722)*x[0]+(-0.149618364398)*x[1]+(-1.06968763194)*x[2]
        s[1,3]=(0.296694301988)*x[0]+(-1.35691147932)*x[1]+(0.748623252633)*x[2]
        s[1,4]=(0.724332288795)*x[0]+(0.349026497158)*x[1]+(0.377755550832)*x[2]
        s[2,0]=(0.560218327038)*x[0]+(0.184262569787)*x[1]+(-0.821718588725)*x[2]
        s[2,1]=(0.421969037007)*x[0]+(-0.00510504728331)*x[1]+(1.43225098227)*x[2]
        s[2,2]=(0.889590883145)*x[0]+(-0.408365128114)*x[1]+(-1.80100224461)*x[2]
        s[2,3]=(-1.91735498201)*x[0]+(0.590584386339)*x[1]+(-0.905610066725)*x[2]
        s[2,4]=(-0.367106108866)*x[0]+(-0.92493020573)*x[1]+(-0.0746199487091)*x[2]
        s[3,0]=(0.919247371049)*x[0]+(0.326674777975)*x[1]+(-1.47385875396)*x[2]
        s[3,1]=(-0.850835329183)*x[0]+(-0.536914325383)*x[1]+(1.02264919467)*x[2]
        s[3,2]=(-1.41905423752)*x[0]+(-0.632415273773)*x[1]+(0.842187325237)*x[2]
        s[3,3]=(-0.396129693721)*x[0]+(-0.110878220479)*x[1]+(0.49355539737)*x[2]
        s[3,4]=(-0.0781905590856)*x[0]+(-0.0348772309292)*x[1]+(0.184581658536)*x[2]
        ref[0,0]=(0.310668873066)*x_ref[0]+(1.66326709742)*x_ref[1]+(1.10140064967)*x_ref[2]
        ref[0,1]=(0.46962350981)*x_ref[0]+(-1.02292998305)*x_ref[1]+(0.919666772954)*x_ref[2]
        ref[0,2]=(-1.18114779319)*x_ref[0]+(1.23819669136)*x_ref[1]+(0.0622519749551)*x_ref[2]
        ref[0,3]=(1.08356504043)*x_ref[0]+(0.450517408398)*x_ref[1]+(-0.528809988127)*x_ref[2]
        ref[0,4]=(-0.179349520643)*x_ref[0]+(0.192827326205)*x_ref[1]+(-0.543850326813)*x_ref[2]
        ref[1,0]=(-1.60006459477)*x_ref[0]+(0.800497908511)*x_ref[1]+(0.745234766014)*x_ref[2]
        ref[1,1]=(0.667355662535)*x_ref[0]+(0.800303765538)*x_ref[1]+(-1.17397311427)*x_ref[2]
        ref[1,2]=(0.19342327722)*x_ref[0]+(-0.149618364398)*x_ref[1]+(-1.06968763194)*x_ref[2]
        ref[1,3]=(0.296694301988)*x_ref[0]+(-1.35691147932)*x_ref[1]+(0.748623252633)*x_ref[2]
        ref[1,4]=(0.724332288795)*x_ref[0]+(0.349026497158)*x_ref[1]+(0.377755550832)*x_ref[2]
        ref[2,0]=(0.560218327038)*x_ref[0]+(0.184262569787)*x_ref[1]+(-0.821718588725)*x_ref[2]
        ref[2,1]=(0.421969037007)*x_ref[0]+(-0.00510504728331)*x_ref[1]+(1.43225098227)*x_ref[2]
        ref[2,2]=(0.889590883145)*x_ref[0]+(-0.408365128114)*x_ref[1]+(-1.80100224461)*x_ref[2]
        ref[2,3]=(-1.91735498201)*x_ref[0]+(0.590584386339)*x_ref[1]+(-0.905610066725)*x_ref[2]
        ref[2,4]=(-0.367106108866)*x_ref[0]+(-0.92493020573)*x_ref[1]+(-0.0746199487091)*x_ref[2]
        ref[3,0]=(0.919247371049)*x_ref[0]+(0.326674777975)*x_ref[1]+(-1.47385875396)*x_ref[2]
        ref[3,1]=(-0.850835329183)*x_ref[0]+(-0.536914325383)*x_ref[1]+(1.02264919467)*x_ref[2]
        ref[3,2]=(-1.41905423752)*x_ref[0]+(-0.632415273773)*x_ref[1]+(0.842187325237)*x_ref[2]
        ref[3,3]=(-0.396129693721)*x_ref[0]+(-0.110878220479)*x_ref[1]+(0.49355539737)*x_ref[2]
        ref[3,4]=(-0.0781905590856)*x_ref[0]+(-0.0348772309292)*x_ref[1]+(0.184581658536)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedSolution_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.540497015605)*x[0]+(0.282753164584)*x[1]
        s[0,0,1]=(-0.127429073679)*x[0]+(-0.709993735868)*x[1]
        s[0,1,0]=(1.02019673589)*x[0]+(-1.53422031862)*x[1]
        s[0,1,1]=(0.926292918083)*x[0]+(-0.0351188336657)*x[1]
        s[1,0,0]=(-0.736080207248)*x[0]+(1.25260396336)*x[1]
        s[1,0,1]=(0.152653303192)*x[0]+(0.22547444598)*x[1]
        s[1,1,0]=(0.729526332632)*x[0]+(0.919207155406)*x[1]
        s[1,1,1]=(-0.346325378686)*x[0]+(0.141030274609)*x[1]
        s[2,0,0]=(-0.753354302966)*x[0]+(-0.766294422747)*x[1]
        s[2,0,1]=(1.17251569201)*x[0]+(-0.51673325599)*x[1]
        s[2,1,0]=(0.66192662328)*x[0]+(-0.572345218315)*x[1]
        s[2,1,1]=(1.74689126451)*x[0]+(-0.0537684141408)*x[1]
        s[3,0,0]=(0.300593672769)*x[0]+(0.0833505471404)*x[1]
        s[3,0,1]=(0.770671590417)*x[0]+(-0.667020549164)*x[1]
        s[3,1,0]=(0.281444608616)*x[0]+(0.701762037742)*x[1]
        s[3,1,1]=(0.404631442844)*x[0]+(0.337744807226)*x[1]
        s[4,0,0]=(-1.43357362276)*x[0]+(0.687631941644)*x[1]
        s[4,0,1]=(0.694571843151)*x[0]+(0.979495014797)*x[1]
        s[4,1,0]=(-0.0880858702517)*x[0]+(0.729298516002)*x[1]
        s[4,1,1]=(0.641170415423)*x[0]+(-0.407278118417)*x[1]
        s[5,0,0]=(-0.60364366976)*x[0]+(-0.663798735078)*x[1]
        s[5,0,1]=(-0.841035861641)*x[0]+(-1.09892876031)*x[1]
        s[5,1,0]=(0.0838180897314)*x[0]+(0.15765944496)*x[1]
        s[5,1,1]=(-0.847291245273)*x[0]+(1.01145652383)*x[1]
        ref[0,0,0]=(0.540497015605)*x_ref[0]+(0.282753164584)*x_ref[1]
        ref[0,0,1]=(-0.127429073679)*x_ref[0]+(-0.709993735868)*x_ref[1]
        ref[0,1,0]=(1.02019673589)*x_ref[0]+(-1.53422031862)*x_ref[1]
        ref[0,1,1]=(0.926292918083)*x_ref[0]+(-0.0351188336657)*x_ref[1]
        ref[1,0,0]=(-0.736080207248)*x_ref[0]+(1.25260396336)*x_ref[1]
        ref[1,0,1]=(0.152653303192)*x_ref[0]+(0.22547444598)*x_ref[1]
        ref[1,1,0]=(0.729526332632)*x_ref[0]+(0.919207155406)*x_ref[1]
        ref[1,1,1]=(-0.346325378686)*x_ref[0]+(0.141030274609)*x_ref[1]
        ref[2,0,0]=(-0.753354302966)*x_ref[0]+(-0.766294422747)*x_ref[1]
        ref[2,0,1]=(1.17251569201)*x_ref[0]+(-0.51673325599)*x_ref[1]
        ref[2,1,0]=(0.66192662328)*x_ref[0]+(-0.572345218315)*x_ref[1]
        ref[2,1,1]=(1.74689126451)*x_ref[0]+(-0.0537684141408)*x_ref[1]
        ref[3,0,0]=(0.300593672769)*x_ref[0]+(0.0833505471404)*x_ref[1]
        ref[3,0,1]=(0.770671590417)*x_ref[0]+(-0.667020549164)*x_ref[1]
        ref[3,1,0]=(0.281444608616)*x_ref[0]+(0.701762037742)*x_ref[1]
        ref[3,1,1]=(0.404631442844)*x_ref[0]+(0.337744807226)*x_ref[1]
        ref[4,0,0]=(-1.43357362276)*x_ref[0]+(0.687631941644)*x_ref[1]
        ref[4,0,1]=(0.694571843151)*x_ref[0]+(0.979495014797)*x_ref[1]
        ref[4,1,0]=(-0.0880858702517)*x_ref[0]+(0.729298516002)*x_ref[1]
        ref[4,1,1]=(0.641170415423)*x_ref[0]+(-0.407278118417)*x_ref[1]
        ref[5,0,0]=(-0.60364366976)*x_ref[0]+(-0.663798735078)*x_ref[1]
        ref[5,0,1]=(-0.841035861641)*x_ref[0]+(-1.09892876031)*x_ref[1]
        ref[5,1,0]=(0.0838180897314)*x_ref[0]+(0.15765944496)*x_ref[1]
        ref[5,1,1]=(-0.847291245273)*x_ref[0]+(1.01145652383)*x_ref[1]
      else:
        s[0,0,0]=(-0.29326741782)*x[0]+(0.716468869525)*x[1]+(-0.194120207009)*x[2]
        s[0,0,1]=(-0.0687304327556)*x[0]+(0.181789058707)*x[1]+(-0.940504268245)*x[2]
        s[0,1,0]=(0.437145055905)*x[0]+(1.72039421904)*x[1]+(-0.0326204183567)*x[2]
        s[0,1,1]=(1.25919532639)*x[0]+(0.082561920762)*x[1]+(1.01503353491)*x[2]
        s[1,0,0]=(-0.552129852247)*x[0]+(-0.661963725391)*x[1]+(0.191012614278)*x[2]
        s[1,0,1]=(-0.322228468989)*x[0]+(-0.79555280135)*x[1]+(0.0689420056626)*x[2]
        s[1,1,0]=(0.294413007041)*x[0]+(-0.453661316564)*x[1]+(1.74072519115)*x[2]
        s[1,1,1]=(-0.13491214165)*x[0]+(-1.18907003959)*x[1]+(-1.59325361447)*x[2]
        s[2,0,0]=(-0.522689030297)*x[0]+(-0.784658318696)*x[1]+(0.618343699514)*x[2]
        s[2,0,1]=(-0.227214629419)*x[0]+(0.770163719023)*x[1]+(-0.272804476799)*x[2]
        s[2,1,0]=(-1.28131384809)*x[0]+(0.0591076357001)*x[1]+(-0.206570950984)*x[2]
        s[2,1,1]=(0.00241150481435)*x[0]+(0.0678794618244)*x[1]+(-0.272677008307)*x[2]
        s[3,0,0]=(-0.537194244467)*x[0]+(-0.886303700791)*x[1]+(0.704423993884)*x[2]
        s[3,0,1]=(0.0924777549448)*x[0]+(-0.783168034362)*x[1]+(-0.450930098538)*x[2]
        s[3,1,0]=(0.611562862859)*x[0]+(-1.04224402701)*x[1]+(-0.775594008735)*x[2]
        s[3,1,1]=(1.44402445157)*x[0]+(-0.736290409307)*x[1]+(0.51690502447)*x[2]
        s[4,0,0]=(-0.470210375338)*x[0]+(-0.433261543227)*x[1]+(-0.582080162476)*x[2]
        s[4,0,1]=(0.313323330562)*x[0]+(-1.47502042219)*x[1]+(-1.50202564103)*x[2]
        s[4,1,0]=(-0.807949755771)*x[0]+(-0.162904810555)*x[1]+(0.748809317029)*x[2]
        s[4,1,1]=(-1.24228935643)*x[0]+(0.193414583713)*x[1]+(0.85984063481)*x[2]
        s[5,0,0]=(1.08068739991)*x[0]+(-1.30655986333)*x[1]+(0.289800143967)*x[2]
        s[5,0,1]=(1.12114019122)*x[0]+(0.280561086566)*x[1]+(0.844991094994)*x[2]
        s[5,1,0]=(-0.133922711479)*x[0]+(-1.39747129717)*x[1]+(-1.11382363437)*x[2]
        s[5,1,1]=(-1.34164423823)*x[0]+(-0.335643322772)*x[1]+(0.757067990421)*x[2]
        ref[0,0,0]=(-0.29326741782)*x_ref[0]+(0.716468869525)*x_ref[1]+(-0.194120207009)*x_ref[2]
        ref[0,0,1]=(-0.0687304327556)*x_ref[0]+(0.181789058707)*x_ref[1]+(-0.940504268245)*x_ref[2]
        ref[0,1,0]=(0.437145055905)*x_ref[0]+(1.72039421904)*x_ref[1]+(-0.0326204183567)*x_ref[2]
        ref[0,1,1]=(1.25919532639)*x_ref[0]+(0.082561920762)*x_ref[1]+(1.01503353491)*x_ref[2]
        ref[1,0,0]=(-0.552129852247)*x_ref[0]+(-0.661963725391)*x_ref[1]+(0.191012614278)*x_ref[2]
        ref[1,0,1]=(-0.322228468989)*x_ref[0]+(-0.79555280135)*x_ref[1]+(0.0689420056626)*x_ref[2]
        ref[1,1,0]=(0.294413007041)*x_ref[0]+(-0.453661316564)*x_ref[1]+(1.74072519115)*x_ref[2]
        ref[1,1,1]=(-0.13491214165)*x_ref[0]+(-1.18907003959)*x_ref[1]+(-1.59325361447)*x_ref[2]
        ref[2,0,0]=(-0.522689030297)*x_ref[0]+(-0.784658318696)*x_ref[1]+(0.618343699514)*x_ref[2]
        ref[2,0,1]=(-0.227214629419)*x_ref[0]+(0.770163719023)*x_ref[1]+(-0.272804476799)*x_ref[2]
        ref[2,1,0]=(-1.28131384809)*x_ref[0]+(0.0591076357001)*x_ref[1]+(-0.206570950984)*x_ref[2]
        ref[2,1,1]=(0.00241150481435)*x_ref[0]+(0.0678794618244)*x_ref[1]+(-0.272677008307)*x_ref[2]
        ref[3,0,0]=(-0.537194244467)*x_ref[0]+(-0.886303700791)*x_ref[1]+(0.704423993884)*x_ref[2]
        ref[3,0,1]=(0.0924777549448)*x_ref[0]+(-0.783168034362)*x_ref[1]+(-0.450930098538)*x_ref[2]
        ref[3,1,0]=(0.611562862859)*x_ref[0]+(-1.04224402701)*x_ref[1]+(-0.775594008735)*x_ref[2]
        ref[3,1,1]=(1.44402445157)*x_ref[0]+(-0.736290409307)*x_ref[1]+(0.51690502447)*x_ref[2]
        ref[4,0,0]=(-0.470210375338)*x_ref[0]+(-0.433261543227)*x_ref[1]+(-0.582080162476)*x_ref[2]
        ref[4,0,1]=(0.313323330562)*x_ref[0]+(-1.47502042219)*x_ref[1]+(-1.50202564103)*x_ref[2]
        ref[4,1,0]=(-0.807949755771)*x_ref[0]+(-0.162904810555)*x_ref[1]+(0.748809317029)*x_ref[2]
        ref[4,1,1]=(-1.24228935643)*x_ref[0]+(0.193414583713)*x_ref[1]+(0.85984063481)*x_ref[2]
        ref[5,0,0]=(1.08068739991)*x_ref[0]+(-1.30655986333)*x_ref[1]+(0.289800143967)*x_ref[2]
        ref[5,0,1]=(1.12114019122)*x_ref[0]+(0.280561086566)*x_ref[1]+(0.844991094994)*x_ref[2]
        ref[5,1,0]=(-0.133922711479)*x_ref[0]+(-1.39747129717)*x_ref[1]+(-1.11382363437)*x_ref[2]
        ref[5,1,1]=(-1.34164423823)*x_ref[0]+(-0.335643322772)*x_ref[1]+(0.757067990421)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedSolution_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedSolution onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.9809347951)*x[0]+(-1.05666590015)*x[1]
        s[0,0,0,1]=(1.15925248094)*x[0]+(1.22070923582)*x[1]
        s[0,0,1,0]=(-0.0724361628927)*x[0]+(-0.439807238916)*x[1]
        s[0,0,1,1]=(-0.608816713414)*x[0]+(1.78792604745)*x[1]
        s[0,0,2,0]=(-0.78903506363)*x[0]+(-0.75896072902)*x[1]
        s[0,0,2,1]=(0.516424260522)*x[0]+(1.16788015378)*x[1]
        s[0,1,0,0]=(0.383586654748)*x[0]+(1.4179924146)*x[1]
        s[0,1,0,1]=(0.97236907624)*x[0]+(-0.438982550276)*x[1]
        s[0,1,1,0]=(-0.645083297783)*x[0]+(-0.740808997218)*x[1]
        s[0,1,1,1]=(0.330652781935)*x[0]+(0.519132367114)*x[1]
        s[0,1,2,0]=(1.25159174261)*x[0]+(1.26021904904)*x[1]
        s[0,1,2,1]=(-0.239621293462)*x[0]+(0.0244107517523)*x[1]
        s[0,2,0,0]=(-0.286208240697)*x[0]+(-0.314157258671)*x[1]
        s[0,2,0,1]=(-0.0856319829397)*x[0]+(0.278603809503)*x[1]
        s[0,2,1,0]=(0.107199217139)*x[0]+(0.0948664401127)*x[1]
        s[0,2,1,1]=(-0.24620060163)*x[0]+(-1.42942082558)*x[1]
        s[0,2,2,0]=(0.845285367352)*x[0]+(0.408676159609)*x[1]
        s[0,2,2,1]=(-1.49467460394)*x[0]+(0.270722457265)*x[1]
        s[0,3,0,0]=(0.469802598061)*x[0]+(-0.801790345348)*x[1]
        s[0,3,0,1]=(1.71249257903)*x[0]+(-0.776797378881)*x[1]
        s[0,3,1,0]=(0.0540049036319)*x[0]+(1.19031969164)*x[1]
        s[0,3,1,1]=(0.161929130968)*x[0]+(-0.525381566632)*x[1]
        s[0,3,2,0]=(-0.646041965665)*x[0]+(-0.123993555032)*x[1]
        s[0,3,2,1]=(0.0966044205198)*x[0]+(1.24780552685)*x[1]
        s[0,4,0,0]=(-0.289531041539)*x[0]+(1.21117470928)*x[1]
        s[0,4,0,1]=(-0.112215925799)*x[0]+(0.598717725332)*x[1]
        s[0,4,1,0]=(0.162936056701)*x[0]+(-1.02395851472)*x[1]
        s[0,4,1,1]=(0.694066364547)*x[0]+(1.27205020727)*x[1]
        s[0,4,2,0]=(1.17253981195)*x[0]+(0.130572791304)*x[1]
        s[0,4,2,1]=(-0.555241229093)*x[0]+(-0.881115061403)*x[1]
        s[1,0,0,0]=(0.574137047465)*x[0]+(-1.10459006911)*x[1]
        s[1,0,0,1]=(0.441431076852)*x[0]+(-0.133156832361)*x[1]
        s[1,0,1,0]=(0.217136655494)*x[0]+(0.400834952015)*x[1]
        s[1,0,1,1]=(0.516360280675)*x[0]+(0.532907430046)*x[1]
        s[1,0,2,0]=(-0.340764752946)*x[0]+(-1.26671004418)*x[1]
        s[1,0,2,1]=(-1.23231139783)*x[0]+(0.61934930883)*x[1]
        s[1,1,0,0]=(1.76821949014)*x[0]+(1.67542625016)*x[1]
        s[1,1,0,1]=(0.901675970272)*x[0]+(-0.0567109697208)*x[1]
        s[1,1,1,0]=(1.21143551644)*x[0]+(0.375877605521)*x[1]
        s[1,1,1,1]=(-0.699830133987)*x[0]+(0.174660968691)*x[1]
        s[1,1,2,0]=(1.07839763024)*x[0]+(-0.063554858486)*x[1]
        s[1,1,2,1]=(0.131335396282)*x[0]+(0.109893568063)*x[1]
        s[1,2,0,0]=(-0.562355685272)*x[0]+(0.495744936869)*x[1]
        s[1,2,0,1]=(0.0747584188968)*x[0]+(1.94924356864)*x[1]
        s[1,2,1,0]=(-0.103515803094)*x[0]+(0.545380886507)*x[1]
        s[1,2,1,1]=(-0.708926705928)*x[0]+(-0.0610123075525)*x[1]
        s[1,2,2,0]=(1.06802154614)*x[0]+(0.0622475519235)*x[1]
        s[1,2,2,1]=(0.103937010427)*x[0]+(-0.70201008693)*x[1]
        s[1,3,0,0]=(0.0730726824626)*x[0]+(0.39791188789)*x[1]
        s[1,3,0,1]=(0.455273194688)*x[0]+(1.65845657261)*x[1]
        s[1,3,1,0]=(0.0807767752424)*x[0]+(-0.765026395125)*x[1]
        s[1,3,1,1]=(0.0404890748976)*x[0]+(-0.383198358048)*x[1]
        s[1,3,2,0]=(1.26048751933)*x[0]+(0.00329382174075)*x[1]
        s[1,3,2,1]=(-1.21518062161)*x[0]+(-0.325769725861)*x[1]
        s[1,4,0,0]=(0.682181651323)*x[0]+(-0.375432356672)*x[1]
        s[1,4,0,1]=(-1.09892909715)*x[0]+(0.207984305416)*x[1]
        s[1,4,1,0]=(-1.2375595164)*x[0]+(0.215243019189)*x[1]
        s[1,4,1,1]=(-0.595125980205)*x[0]+(0.29104395616)*x[1]
        s[1,4,2,0]=(0.127947393316)*x[0]+(1.43408039596)*x[1]
        s[1,4,2,1]=(0.468574871016)*x[0]+(0.333319611281)*x[1]
        s[2,0,0,0]=(0.369954209324)*x[0]+(-1.09614147452)*x[1]
        s[2,0,0,1]=(-0.120003056441)*x[0]+(-0.172964298439)*x[1]
        s[2,0,1,0]=(1.38230317911)*x[0]+(1.25730140777)*x[1]
        s[2,0,1,1]=(1.10857329622)*x[0]+(0.449561992049)*x[1]
        s[2,0,2,0]=(-1.48479939533)*x[0]+(-0.839474409588)*x[1]
        s[2,0,2,1]=(-0.662177361426)*x[0]+(0.380170348062)*x[1]
        s[2,1,0,0]=(1.13771311046)*x[0]+(-1.59960740614)*x[1]
        s[2,1,0,1]=(0.925401031957)*x[0]+(1.30170041054)*x[1]
        s[2,1,1,0]=(0.455911230864)*x[0]+(0.755253828053)*x[1]
        s[2,1,1,1]=(0.64783771432)*x[0]+(0.251664567771)*x[1]
        s[2,1,2,0]=(-0.0449523798045)*x[0]+(-0.614423607357)*x[1]
        s[2,1,2,1]=(0.635435155497)*x[0]+(0.218411499589)*x[1]
        s[2,2,0,0]=(1.07064943699)*x[0]+(1.33695839858)*x[1]
        s[2,2,0,1]=(0.286067300461)*x[0]+(-1.19968009538)*x[1]
        s[2,2,1,0]=(0.373936537767)*x[0]+(-0.684346831863)*x[1]
        s[2,2,1,1]=(0.709764408449)*x[0]+(0.560181368149)*x[1]
        s[2,2,2,0]=(0.687217025484)*x[0]+(0.638368794102)*x[1]
        s[2,2,2,1]=(-1.13694076648)*x[0]+(1.31012182384)*x[1]
        s[2,3,0,0]=(0.784452969448)*x[0]+(-0.256804074025)*x[1]
        s[2,3,0,1]=(-0.528533290176)*x[0]+(0.230293895766)*x[1]
        s[2,3,1,0]=(-0.152281766055)*x[0]+(-1.29007246581)*x[1]
        s[2,3,1,1]=(-0.113853891603)*x[0]+(-0.146975472116)*x[1]
        s[2,3,2,0]=(-1.16737851429)*x[0]+(1.18358630738)*x[1]
        s[2,3,2,1]=(-0.110242592809)*x[0]+(-0.228040285919)*x[1]
        s[2,4,0,0]=(-0.615024817893)*x[0]+(1.19310484943)*x[1]
        s[2,4,0,1]=(0.405843983847)*x[0]+(-0.7683655961)*x[1]
        s[2,4,1,0]=(-1.94063527375)*x[0]+(0.135437762233)*x[1]
        s[2,4,1,1]=(-0.680568187039)*x[0]+(0.411161107413)*x[1]
        s[2,4,2,0]=(-1.95575235842)*x[0]+(0.384766876765)*x[1]
        s[2,4,2,1]=(1.37236985627)*x[0]+(0.0082159494728)*x[1]
        s[3,0,0,0]=(1.63255667845)*x[0]+(0.0225121724188)*x[1]
        s[3,0,0,1]=(1.2376463155)*x[0]+(1.32594028087)*x[1]
        s[3,0,1,0]=(1.54455876996)*x[0]+(-1.21021289438)*x[1]
        s[3,0,1,1]=(-1.81162592117)*x[0]+(-0.297403340122)*x[1]
        s[3,0,2,0]=(0.806731759445)*x[0]+(-0.981072370255)*x[1]
        s[3,0,2,1]=(-0.15215786955)*x[0]+(0.163165607014)*x[1]
        s[3,1,0,0]=(0.233006578932)*x[0]+(-1.29832996415)*x[1]
        s[3,1,0,1]=(0.96755903915)*x[0]+(-0.00597412239186)*x[1]
        s[3,1,1,0]=(0.0413376466172)*x[0]+(1.26720441986)*x[1]
        s[3,1,1,1]=(0.676509459424)*x[0]+(1.2621989434)*x[1]
        s[3,1,2,0]=(-0.211034948146)*x[0]+(-1.37470768893)*x[1]
        s[3,1,2,1]=(-0.402208754333)*x[0]+(1.25258243878)*x[1]
        s[3,2,0,0]=(-0.529199129991)*x[0]+(-0.431490132335)*x[1]
        s[3,2,0,1]=(-0.850343939948)*x[0]+(0.0602529984167)*x[1]
        s[3,2,1,0]=(-0.420125360943)*x[0]+(1.24318875197)*x[1]
        s[3,2,1,1]=(0.753685328318)*x[0]+(0.354913692774)*x[1]
        s[3,2,2,0]=(-1.17147381854)*x[0]+(0.850516886779)*x[1]
        s[3,2,2,1]=(1.48587305579)*x[0]+(1.31835868193)*x[1]
        s[3,3,0,0]=(1.80808607564)*x[0]+(-0.124508319578)*x[1]
        s[3,3,0,1]=(-1.84429695094)*x[0]+(-0.254797649176)*x[1]
        s[3,3,1,0]=(1.17110770558)*x[0]+(0.89756258226)*x[1]
        s[3,3,1,1]=(0.692264407822)*x[0]+(-1.00105988426)*x[1]
        s[3,3,2,0]=(-0.505689548372)*x[0]+(0.365519850058)*x[1]
        s[3,3,2,1]=(0.651544272536)*x[0]+(1.09868410461)*x[1]
        s[3,4,0,0]=(0.155129169877)*x[0]+(0.4435929499)*x[1]
        s[3,4,0,1]=(1.65974703109)*x[0]+(-0.776830635917)*x[1]
        s[3,4,1,0]=(-1.05612271738)*x[0]+(1.63557694551)*x[1]
        s[3,4,1,1]=(0.69435492251)*x[0]+(-1.20394873527)*x[1]
        s[3,4,2,0]=(-0.237598987964)*x[0]+(0.404820539664)*x[1]
        s[3,4,2,1]=(-1.32483403281)*x[0]+(0.431217340618)*x[1]
        ref[0,0,0,0]=(-0.9809347951)*x_ref[0]+(-1.05666590015)*x_ref[1]
        ref[0,0,0,1]=(1.15925248094)*x_ref[0]+(1.22070923582)*x_ref[1]
        ref[0,0,1,0]=(-0.0724361628927)*x_ref[0]+(-0.439807238916)*x_ref[1]
        ref[0,0,1,1]=(-0.608816713414)*x_ref[0]+(1.78792604745)*x_ref[1]
        ref[0,0,2,0]=(-0.78903506363)*x_ref[0]+(-0.75896072902)*x_ref[1]
        ref[0,0,2,1]=(0.516424260522)*x_ref[0]+(1.16788015378)*x_ref[1]
        ref[0,1,0,0]=(0.383586654748)*x_ref[0]+(1.4179924146)*x_ref[1]
        ref[0,1,0,1]=(0.97236907624)*x_ref[0]+(-0.438982550276)*x_ref[1]
        ref[0,1,1,0]=(-0.645083297783)*x_ref[0]+(-0.740808997218)*x_ref[1]
        ref[0,1,1,1]=(0.330652781935)*x_ref[0]+(0.519132367114)*x_ref[1]
        ref[0,1,2,0]=(1.25159174261)*x_ref[0]+(1.26021904904)*x_ref[1]
        ref[0,1,2,1]=(-0.239621293462)*x_ref[0]+(0.0244107517523)*x_ref[1]
        ref[0,2,0,0]=(-0.286208240697)*x_ref[0]+(-0.314157258671)*x_ref[1]
        ref[0,2,0,1]=(-0.0856319829397)*x_ref[0]+(0.278603809503)*x_ref[1]
        ref[0,2,1,0]=(0.107199217139)*x_ref[0]+(0.0948664401127)*x_ref[1]
        ref[0,2,1,1]=(-0.24620060163)*x_ref[0]+(-1.42942082558)*x_ref[1]
        ref[0,2,2,0]=(0.845285367352)*x_ref[0]+(0.408676159609)*x_ref[1]
        ref[0,2,2,1]=(-1.49467460394)*x_ref[0]+(0.270722457265)*x_ref[1]
        ref[0,3,0,0]=(0.469802598061)*x_ref[0]+(-0.801790345348)*x_ref[1]
        ref[0,3,0,1]=(1.71249257903)*x_ref[0]+(-0.776797378881)*x_ref[1]
        ref[0,3,1,0]=(0.0540049036319)*x_ref[0]+(1.19031969164)*x_ref[1]
        ref[0,3,1,1]=(0.161929130968)*x_ref[0]+(-0.525381566632)*x_ref[1]
        ref[0,3,2,0]=(-0.646041965665)*x_ref[0]+(-0.123993555032)*x_ref[1]
        ref[0,3,2,1]=(0.0966044205198)*x_ref[0]+(1.24780552685)*x_ref[1]
        ref[0,4,0,0]=(-0.289531041539)*x_ref[0]+(1.21117470928)*x_ref[1]
        ref[0,4,0,1]=(-0.112215925799)*x_ref[0]+(0.598717725332)*x_ref[1]
        ref[0,4,1,0]=(0.162936056701)*x_ref[0]+(-1.02395851472)*x_ref[1]
        ref[0,4,1,1]=(0.694066364547)*x_ref[0]+(1.27205020727)*x_ref[1]
        ref[0,4,2,0]=(1.17253981195)*x_ref[0]+(0.130572791304)*x_ref[1]
        ref[0,4,2,1]=(-0.555241229093)*x_ref[0]+(-0.881115061403)*x_ref[1]
        ref[1,0,0,0]=(0.574137047465)*x_ref[0]+(-1.10459006911)*x_ref[1]
        ref[1,0,0,1]=(0.441431076852)*x_ref[0]+(-0.133156832361)*x_ref[1]
        ref[1,0,1,0]=(0.217136655494)*x_ref[0]+(0.400834952015)*x_ref[1]
        ref[1,0,1,1]=(0.516360280675)*x_ref[0]+(0.532907430046)*x_ref[1]
        ref[1,0,2,0]=(-0.340764752946)*x_ref[0]+(-1.26671004418)*x_ref[1]
        ref[1,0,2,1]=(-1.23231139783)*x_ref[0]+(0.61934930883)*x_ref[1]
        ref[1,1,0,0]=(1.76821949014)*x_ref[0]+(1.67542625016)*x_ref[1]
        ref[1,1,0,1]=(0.901675970272)*x_ref[0]+(-0.0567109697208)*x_ref[1]
        ref[1,1,1,0]=(1.21143551644)*x_ref[0]+(0.375877605521)*x_ref[1]
        ref[1,1,1,1]=(-0.699830133987)*x_ref[0]+(0.174660968691)*x_ref[1]
        ref[1,1,2,0]=(1.07839763024)*x_ref[0]+(-0.063554858486)*x_ref[1]
        ref[1,1,2,1]=(0.131335396282)*x_ref[0]+(0.109893568063)*x_ref[1]
        ref[1,2,0,0]=(-0.562355685272)*x_ref[0]+(0.495744936869)*x_ref[1]
        ref[1,2,0,1]=(0.0747584188968)*x_ref[0]+(1.94924356864)*x_ref[1]
        ref[1,2,1,0]=(-0.103515803094)*x_ref[0]+(0.545380886507)*x_ref[1]
        ref[1,2,1,1]=(-0.708926705928)*x_ref[0]+(-0.0610123075525)*x_ref[1]
        ref[1,2,2,0]=(1.06802154614)*x_ref[0]+(0.0622475519235)*x_ref[1]
        ref[1,2,2,1]=(0.103937010427)*x_ref[0]+(-0.70201008693)*x_ref[1]
        ref[1,3,0,0]=(0.0730726824626)*x_ref[0]+(0.39791188789)*x_ref[1]
        ref[1,3,0,1]=(0.455273194688)*x_ref[0]+(1.65845657261)*x_ref[1]
        ref[1,3,1,0]=(0.0807767752424)*x_ref[0]+(-0.765026395125)*x_ref[1]
        ref[1,3,1,1]=(0.0404890748976)*x_ref[0]+(-0.383198358048)*x_ref[1]
        ref[1,3,2,0]=(1.26048751933)*x_ref[0]+(0.00329382174075)*x_ref[1]
        ref[1,3,2,1]=(-1.21518062161)*x_ref[0]+(-0.325769725861)*x_ref[1]
        ref[1,4,0,0]=(0.682181651323)*x_ref[0]+(-0.375432356672)*x_ref[1]
        ref[1,4,0,1]=(-1.09892909715)*x_ref[0]+(0.207984305416)*x_ref[1]
        ref[1,4,1,0]=(-1.2375595164)*x_ref[0]+(0.215243019189)*x_ref[1]
        ref[1,4,1,1]=(-0.595125980205)*x_ref[0]+(0.29104395616)*x_ref[1]
        ref[1,4,2,0]=(0.127947393316)*x_ref[0]+(1.43408039596)*x_ref[1]
        ref[1,4,2,1]=(0.468574871016)*x_ref[0]+(0.333319611281)*x_ref[1]
        ref[2,0,0,0]=(0.369954209324)*x_ref[0]+(-1.09614147452)*x_ref[1]
        ref[2,0,0,1]=(-0.120003056441)*x_ref[0]+(-0.172964298439)*x_ref[1]
        ref[2,0,1,0]=(1.38230317911)*x_ref[0]+(1.25730140777)*x_ref[1]
        ref[2,0,1,1]=(1.10857329622)*x_ref[0]+(0.449561992049)*x_ref[1]
        ref[2,0,2,0]=(-1.48479939533)*x_ref[0]+(-0.839474409588)*x_ref[1]
        ref[2,0,2,1]=(-0.662177361426)*x_ref[0]+(0.380170348062)*x_ref[1]
        ref[2,1,0,0]=(1.13771311046)*x_ref[0]+(-1.59960740614)*x_ref[1]
        ref[2,1,0,1]=(0.925401031957)*x_ref[0]+(1.30170041054)*x_ref[1]
        ref[2,1,1,0]=(0.455911230864)*x_ref[0]+(0.755253828053)*x_ref[1]
        ref[2,1,1,1]=(0.64783771432)*x_ref[0]+(0.251664567771)*x_ref[1]
        ref[2,1,2,0]=(-0.0449523798045)*x_ref[0]+(-0.614423607357)*x_ref[1]
        ref[2,1,2,1]=(0.635435155497)*x_ref[0]+(0.218411499589)*x_ref[1]
        ref[2,2,0,0]=(1.07064943699)*x_ref[0]+(1.33695839858)*x_ref[1]
        ref[2,2,0,1]=(0.286067300461)*x_ref[0]+(-1.19968009538)*x_ref[1]
        ref[2,2,1,0]=(0.373936537767)*x_ref[0]+(-0.684346831863)*x_ref[1]
        ref[2,2,1,1]=(0.709764408449)*x_ref[0]+(0.560181368149)*x_ref[1]
        ref[2,2,2,0]=(0.687217025484)*x_ref[0]+(0.638368794102)*x_ref[1]
        ref[2,2,2,1]=(-1.13694076648)*x_ref[0]+(1.31012182384)*x_ref[1]
        ref[2,3,0,0]=(0.784452969448)*x_ref[0]+(-0.256804074025)*x_ref[1]
        ref[2,3,0,1]=(-0.528533290176)*x_ref[0]+(0.230293895766)*x_ref[1]
        ref[2,3,1,0]=(-0.152281766055)*x_ref[0]+(-1.29007246581)*x_ref[1]
        ref[2,3,1,1]=(-0.113853891603)*x_ref[0]+(-0.146975472116)*x_ref[1]
        ref[2,3,2,0]=(-1.16737851429)*x_ref[0]+(1.18358630738)*x_ref[1]
        ref[2,3,2,1]=(-0.110242592809)*x_ref[0]+(-0.228040285919)*x_ref[1]
        ref[2,4,0,0]=(-0.615024817893)*x_ref[0]+(1.19310484943)*x_ref[1]
        ref[2,4,0,1]=(0.405843983847)*x_ref[0]+(-0.7683655961)*x_ref[1]
        ref[2,4,1,0]=(-1.94063527375)*x_ref[0]+(0.135437762233)*x_ref[1]
        ref[2,4,1,1]=(-0.680568187039)*x_ref[0]+(0.411161107413)*x_ref[1]
        ref[2,4,2,0]=(-1.95575235842)*x_ref[0]+(0.384766876765)*x_ref[1]
        ref[2,4,2,1]=(1.37236985627)*x_ref[0]+(0.0082159494728)*x_ref[1]
        ref[3,0,0,0]=(1.63255667845)*x_ref[0]+(0.0225121724188)*x_ref[1]
        ref[3,0,0,1]=(1.2376463155)*x_ref[0]+(1.32594028087)*x_ref[1]
        ref[3,0,1,0]=(1.54455876996)*x_ref[0]+(-1.21021289438)*x_ref[1]
        ref[3,0,1,1]=(-1.81162592117)*x_ref[0]+(-0.297403340122)*x_ref[1]
        ref[3,0,2,0]=(0.806731759445)*x_ref[0]+(-0.981072370255)*x_ref[1]
        ref[3,0,2,1]=(-0.15215786955)*x_ref[0]+(0.163165607014)*x_ref[1]
        ref[3,1,0,0]=(0.233006578932)*x_ref[0]+(-1.29832996415)*x_ref[1]
        ref[3,1,0,1]=(0.96755903915)*x_ref[0]+(-0.00597412239186)*x_ref[1]
        ref[3,1,1,0]=(0.0413376466172)*x_ref[0]+(1.26720441986)*x_ref[1]
        ref[3,1,1,1]=(0.676509459424)*x_ref[0]+(1.2621989434)*x_ref[1]
        ref[3,1,2,0]=(-0.211034948146)*x_ref[0]+(-1.37470768893)*x_ref[1]
        ref[3,1,2,1]=(-0.402208754333)*x_ref[0]+(1.25258243878)*x_ref[1]
        ref[3,2,0,0]=(-0.529199129991)*x_ref[0]+(-0.431490132335)*x_ref[1]
        ref[3,2,0,1]=(-0.850343939948)*x_ref[0]+(0.0602529984167)*x_ref[1]
        ref[3,2,1,0]=(-0.420125360943)*x_ref[0]+(1.24318875197)*x_ref[1]
        ref[3,2,1,1]=(0.753685328318)*x_ref[0]+(0.354913692774)*x_ref[1]
        ref[3,2,2,0]=(-1.17147381854)*x_ref[0]+(0.850516886779)*x_ref[1]
        ref[3,2,2,1]=(1.48587305579)*x_ref[0]+(1.31835868193)*x_ref[1]
        ref[3,3,0,0]=(1.80808607564)*x_ref[0]+(-0.124508319578)*x_ref[1]
        ref[3,3,0,1]=(-1.84429695094)*x_ref[0]+(-0.254797649176)*x_ref[1]
        ref[3,3,1,0]=(1.17110770558)*x_ref[0]+(0.89756258226)*x_ref[1]
        ref[3,3,1,1]=(0.692264407822)*x_ref[0]+(-1.00105988426)*x_ref[1]
        ref[3,3,2,0]=(-0.505689548372)*x_ref[0]+(0.365519850058)*x_ref[1]
        ref[3,3,2,1]=(0.651544272536)*x_ref[0]+(1.09868410461)*x_ref[1]
        ref[3,4,0,0]=(0.155129169877)*x_ref[0]+(0.4435929499)*x_ref[1]
        ref[3,4,0,1]=(1.65974703109)*x_ref[0]+(-0.776830635917)*x_ref[1]
        ref[3,4,1,0]=(-1.05612271738)*x_ref[0]+(1.63557694551)*x_ref[1]
        ref[3,4,1,1]=(0.69435492251)*x_ref[0]+(-1.20394873527)*x_ref[1]
        ref[3,4,2,0]=(-0.237598987964)*x_ref[0]+(0.404820539664)*x_ref[1]
        ref[3,4,2,1]=(-1.32483403281)*x_ref[0]+(0.431217340618)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.270116752643)*x[0]+(0.0715809194071)*x[1]+(0.930011298681)*x[2]
        s[0,0,0,1]=(-0.0435807732016)*x[0]+(-0.492622221729)*x[1]+(-1.36754174971)*x[2]
        s[0,0,1,0]=(-0.949818131569)*x[0]+(1.08437418341)*x[1]+(0.527267447353)*x[2]
        s[0,0,1,1]=(-0.164876627537)*x[0]+(-1.44143896699)*x[1]+(0.0355140164081)*x[2]
        s[0,0,2,0]=(-0.530229192916)*x[0]+(-0.988412045132)*x[1]+(-0.466520393878)*x[2]
        s[0,0,2,1]=(-0.0166373496749)*x[0]+(0.322621663585)*x[1]+(-0.102019336059)*x[2]
        s[0,1,0,0]=(-1.36627178258)*x[0]+(-0.803463758822)*x[1]+(0.734446026661)*x[2]
        s[0,1,0,1]=(-0.804729478825)*x[0]+(-1.24406753288)*x[1]+(-0.755743791016)*x[2]
        s[0,1,1,0]=(1.03869027308)*x[0]+(-0.25237293575)*x[1]+(1.43737134487)*x[2]
        s[0,1,1,1]=(1.48269192251)*x[0]+(0.045590449115)*x[1]+(0.47425103841)*x[2]
        s[0,1,2,0]=(-0.738695287005)*x[0]+(1.01938431903)*x[1]+(1.20767534795)*x[2]
        s[0,1,2,1]=(-0.237445941287)*x[0]+(0.766593746786)*x[1]+(0.0297279831889)*x[2]
        s[0,2,0,0]=(-0.804902383375)*x[0]+(0.0962339257662)*x[1]+(1.37407473707)*x[2]
        s[0,2,0,1]=(-0.0821182376238)*x[0]+(0.35489327051)*x[1]+(-1.43235115632)*x[2]
        s[0,2,1,0]=(1.11593707013)*x[0]+(0.744039820943)*x[1]+(0.294815281064)*x[2]
        s[0,2,1,1]=(0.0789595478782)*x[0]+(0.719656781874)*x[1]+(0.355275528053)*x[2]
        s[0,2,2,0]=(1.20514986435)*x[0]+(0.203004753589)*x[1]+(-0.0731941650839)*x[2]
        s[0,2,2,1]=(-1.23231910051)*x[0]+(-0.470834142341)*x[1]+(-0.633518095593)*x[2]
        s[0,3,0,0]=(0.468164093965)*x[0]+(-1.15996958069)*x[1]+(0.922274695212)*x[2]
        s[0,3,0,1]=(-0.964250141036)*x[0]+(-0.769747467017)*x[1]+(-1.16369265956)*x[2]
        s[0,3,1,0]=(1.1622810944)*x[0]+(-1.72794997517)*x[1]+(-0.297833274612)*x[2]
        s[0,3,1,1]=(1.20239531447)*x[0]+(-0.195111187444)*x[1]+(-0.720186501147)*x[2]
        s[0,3,2,0]=(-0.0760001358746)*x[0]+(-1.65266437471)*x[1]+(-0.496163951015)*x[2]
        s[0,3,2,1]=(-1.4998248458)*x[0]+(-1.73305168566)*x[1]+(-0.107384086634)*x[2]
        s[0,4,0,0]=(0.235699357585)*x[0]+(-0.213348404403)*x[1]+(0.35368491512)*x[2]
        s[0,4,0,1]=(0.264641788778)*x[0]+(0.435251584452)*x[1]+(0.22539974851)*x[2]
        s[0,4,1,0]=(0.540203510472)*x[0]+(-1.33243180984)*x[1]+(-0.520357911324)*x[2]
        s[0,4,1,1]=(-0.629868492098)*x[0]+(-0.403402885662)*x[1]+(-0.184983842426)*x[2]
        s[0,4,2,0]=(1.17843494613)*x[0]+(0.562783141759)*x[1]+(-1.48575741726)*x[2]
        s[0,4,2,1]=(0.623796209289)*x[0]+(1.02627846638)*x[1]+(1.39369241754)*x[2]
        s[1,0,0,0]=(1.9855507901)*x[0]+(0.152976407288)*x[1]+(0.0931547545156)*x[2]
        s[1,0,0,1]=(-0.422055088151)*x[0]+(-0.509449435814)*x[1]+(-1.31705141661)*x[2]
        s[1,0,1,0]=(-0.24193816853)*x[0]+(-0.68450536388)*x[1]+(-0.309024887085)*x[2]
        s[1,0,1,1]=(1.24557169323)*x[0]+(0.353538565492)*x[1]+(0.115821365646)*x[2]
        s[1,0,2,0]=(-0.280246915004)*x[0]+(-0.0972556765171)*x[1]+(0.758816560386)*x[2]
        s[1,0,2,1]=(1.0396259589)*x[0]+(-0.239507470189)*x[1]+(-0.987815676492)*x[2]
        s[1,1,0,0]=(0.245203366317)*x[0]+(-1.53269800659)*x[1]+(0.619360300925)*x[2]
        s[1,1,0,1]=(0.246406467482)*x[0]+(-0.729988292889)*x[1]+(0.585422287189)*x[2]
        s[1,1,1,0]=(1.19810038121)*x[0]+(-0.72772445701)*x[1]+(1.11619392954)*x[2]
        s[1,1,1,1]=(-0.311984976134)*x[0]+(-0.188443023344)*x[1]+(-0.100078290654)*x[2]
        s[1,1,2,0]=(0.158578485066)*x[0]+(0.214804573675)*x[1]+(-0.22418434724)*x[2]
        s[1,1,2,1]=(-0.0550611413075)*x[0]+(-0.147088267891)*x[1]+(-0.269933900128)*x[2]
        s[1,2,0,0]=(-0.8618782319)*x[0]+(-0.770284701216)*x[1]+(-0.407458817279)*x[2]
        s[1,2,0,1]=(-0.242336600721)*x[0]+(-1.57823911131)*x[1]+(0.452055841078)*x[2]
        s[1,2,1,0]=(0.0537219652862)*x[0]+(0.18864518829)*x[1]+(0.021034116382)*x[2]
        s[1,2,1,1]=(0.143799709809)*x[0]+(0.787934959856)*x[1]+(-0.283457275339)*x[2]
        s[1,2,2,0]=(0.475965504456)*x[0]+(-0.744950301051)*x[1]+(0.45308618026)*x[2]
        s[1,2,2,1]=(0.561593580643)*x[0]+(0.121626528038)*x[1]+(-1.24221676348)*x[2]
        s[1,3,0,0]=(-0.340056043985)*x[0]+(-0.434795551938)*x[1]+(-0.291179084342)*x[2]
        s[1,3,0,1]=(0.312318149724)*x[0]+(1.66950279503)*x[1]+(-1.1622076779)*x[2]
        s[1,3,1,0]=(0.0965172054443)*x[0]+(-0.64102022474)*x[1]+(-0.173612340622)*x[2]
        s[1,3,1,1]=(1.88701305968)*x[0]+(-1.66794136241)*x[1]+(1.56059966446)*x[2]
        s[1,3,2,0]=(-0.497773999274)*x[0]+(0.395345962943)*x[1]+(-0.499105933847)*x[2]
        s[1,3,2,1]=(-0.312096437005)*x[0]+(-1.23301763457)*x[1]+(0.198835942067)*x[2]
        s[1,4,0,0]=(0.076377295342)*x[0]+(-0.487847012612)*x[1]+(-0.084314987823)*x[2]
        s[1,4,0,1]=(0.666434654203)*x[0]+(-0.220376188514)*x[1]+(-0.355749198488)*x[2]
        s[1,4,1,0]=(-0.701104468581)*x[0]+(0.828174884422)*x[1]+(0.309374641659)*x[2]
        s[1,4,1,1]=(0.526340204296)*x[0]+(0.571598927007)*x[1]+(-0.0793826750713)*x[2]
        s[1,4,2,0]=(0.775563727486)*x[0]+(0.248228533771)*x[1]+(-1.18169855975)*x[2]
        s[1,4,2,1]=(-0.399357449979)*x[0]+(-0.723554468137)*x[1]+(0.0231913550517)*x[2]
        s[2,0,0,0]=(-0.0504735749489)*x[0]+(0.506564345355)*x[1]+(0.790336060204)*x[2]
        s[2,0,0,1]=(0.426187262022)*x[0]+(0.966680952632)*x[1]+(0.410542727369)*x[2]
        s[2,0,1,0]=(0.594102954821)*x[0]+(0.0639755896611)*x[1]+(-0.675769510238)*x[2]
        s[2,0,1,1]=(0.367479711308)*x[0]+(0.150736830273)*x[1]+(1.15797942999)*x[2]
        s[2,0,2,0]=(0.875207101066)*x[0]+(-1.12939492894)*x[1]+(0.73561332295)*x[2]
        s[2,0,2,1]=(-1.70885193598)*x[0]+(0.835257764029)*x[1]+(-1.58521909429)*x[2]
        s[2,1,0,0]=(-0.144255046397)*x[0]+(0.590074178586)*x[1]+(0.592144641271)*x[2]
        s[2,1,0,1]=(-0.0688470601204)*x[0]+(-0.153340975531)*x[1]+(1.15241895553)*x[2]
        s[2,1,1,0]=(-0.73257033584)*x[0]+(-0.563316507818)*x[1]+(-1.05620654545)*x[2]
        s[2,1,1,1]=(-1.29833350092)*x[0]+(0.42237934165)*x[1]+(0.0725296431412)*x[2]
        s[2,1,2,0]=(-0.837473685756)*x[0]+(-0.425072509905)*x[1]+(-0.463397001131)*x[2]
        s[2,1,2,1]=(-0.937038584847)*x[0]+(0.531989188194)*x[1]+(-1.06694316698)*x[2]
        s[2,2,0,0]=(-0.846346120269)*x[0]+(-0.639618775615)*x[1]+(-0.928154018676)*x[2]
        s[2,2,0,1]=(-0.726631889289)*x[0]+(-0.1318624965)*x[1]+(-0.343729797494)*x[2]
        s[2,2,1,0]=(0.802665744539)*x[0]+(-0.115632355474)*x[1]+(0.303919795863)*x[2]
        s[2,2,1,1]=(0.0371825373918)*x[0]+(-1.17799388388)*x[1]+(-0.223043449617)*x[2]
        s[2,2,2,0]=(0.75988617552)*x[0]+(-0.608977501492)*x[1]+(1.75211913311)*x[2]
        s[2,2,2,1]=(-0.291397991509)*x[0]+(-1.55173354063)*x[1]+(-0.461269447756)*x[2]
        s[2,3,0,0]=(0.351624965509)*x[0]+(-1.32799214259)*x[1]+(1.58946725409)*x[2]
        s[2,3,0,1]=(0.180149992744)*x[0]+(1.07861987473)*x[1]+(0.834357928788)*x[2]
        s[2,3,1,0]=(0.0475016328103)*x[0]+(0.707449786509)*x[1]+(-0.370421903676)*x[2]
        s[2,3,1,1]=(0.553988297947)*x[0]+(1.13388241001)*x[1]+(0.575604300083)*x[2]
        s[2,3,2,0]=(-0.249047035016)*x[0]+(0.15013128589)*x[1]+(1.3953926353)*x[2]
        s[2,3,2,1]=(-0.783008188949)*x[0]+(0.861992436037)*x[1]+(-0.929496477422)*x[2]
        s[2,4,0,0]=(0.264348996811)*x[0]+(-1.23207394036)*x[1]+(-1.37804290092)*x[2]
        s[2,4,0,1]=(0.360782273055)*x[0]+(-0.270520439821)*x[1]+(-1.01285201623)*x[2]
        s[2,4,1,0]=(0.280578579439)*x[0]+(0.584696145403)*x[1]+(-1.21852764117)*x[2]
        s[2,4,1,1]=(0.112102121128)*x[0]+(-0.0619663521973)*x[1]+(1.08490293257)*x[2]
        s[2,4,2,0]=(-1.89749569924)*x[0]+(-0.21217889191)*x[1]+(0.635524779965)*x[2]
        s[2,4,2,1]=(-0.309239337357)*x[0]+(0.0116284822724)*x[1]+(0.0107247712951)*x[2]
        s[3,0,0,0]=(1.23545873872)*x[0]+(-1.29896744411)*x[1]+(0.855094335492)*x[2]
        s[3,0,0,1]=(-0.51465695324)*x[0]+(1.47799693308)*x[1]+(0.0718360423316)*x[2]
        s[3,0,1,0]=(0.220505442402)*x[0]+(-0.3171385194)*x[1]+(0.27954628665)*x[2]
        s[3,0,1,1]=(0.860582046972)*x[0]+(1.04565549687)*x[1]+(1.18108647078)*x[2]
        s[3,0,2,0]=(-0.593547630941)*x[0]+(-0.0390465590147)*x[1]+(0.00129707682534)*x[2]
        s[3,0,2,1]=(1.39210293617)*x[0]+(-1.05660957806)*x[1]+(-1.20489109051)*x[2]
        s[3,1,0,0]=(-0.536035988946)*x[0]+(-0.594403345465)*x[1]+(1.11719111746)*x[2]
        s[3,1,0,1]=(-0.0403557954127)*x[0]+(-1.32391025453)*x[1]+(-0.639127329416)*x[2]
        s[3,1,1,0]=(-0.495803710666)*x[0]+(-0.199832989186)*x[1]+(-0.832862076561)*x[2]
        s[3,1,1,1]=(0.0216882402803)*x[0]+(0.367268873436)*x[1]+(0.295944335418)*x[2]
        s[3,1,2,0]=(0.172936506269)*x[0]+(-0.339886198196)*x[1]+(0.268201481244)*x[2]
        s[3,1,2,1]=(0.665786569897)*x[0]+(-0.290637226736)*x[1]+(0.617677178537)*x[2]
        s[3,2,0,0]=(-0.657902185911)*x[0]+(-1.34292379802)*x[1]+(0.0641950819134)*x[2]
        s[3,2,0,1]=(0.352039175923)*x[0]+(-0.483047873735)*x[1]+(0.47997600743)*x[2]
        s[3,2,1,0]=(0.944619484106)*x[0]+(-0.308755547249)*x[1]+(-0.765906593989)*x[2]
        s[3,2,1,1]=(-1.09674781865)*x[0]+(0.013069257236)*x[1]+(-0.0741247275628)*x[2]
        s[3,2,2,0]=(0.400057737624)*x[0]+(-1.16165281514)*x[1]+(-0.789757454741)*x[2]
        s[3,2,2,1]=(0.367100312493)*x[0]+(-0.547338908792)*x[1]+(-0.409173859392)*x[2]
        s[3,3,0,0]=(0.890118369638)*x[0]+(1.201296587)*x[1]+(-0.0319246476545)*x[2]
        s[3,3,0,1]=(0.620955702501)*x[0]+(0.438440818114)*x[1]+(0.499747510411)*x[2]
        s[3,3,1,0]=(1.61217537222)*x[0]+(-1.21323714148)*x[1]+(-1.06137097571)*x[2]
        s[3,3,1,1]=(-1.75942413542)*x[0]+(0.440345678738)*x[1]+(0.625299774096)*x[2]
        s[3,3,2,0]=(-0.878497266865)*x[0]+(-1.15879244679)*x[1]+(1.03993124425)*x[2]
        s[3,3,2,1]=(-0.0138445968491)*x[0]+(0.547910692711)*x[1]+(0.971642264613)*x[2]
        s[3,4,0,0]=(-0.107769272336)*x[0]+(-0.987911384393)*x[1]+(1.2076201556)*x[2]
        s[3,4,0,1]=(-0.556570952071)*x[0]+(-0.284568452888)*x[1]+(1.23341917108)*x[2]
        s[3,4,1,0]=(1.56461193947)*x[0]+(0.185382898933)*x[1]+(-1.67714180845)*x[2]
        s[3,4,1,1]=(-0.894934034676)*x[0]+(1.05560233139)*x[1]+(0.651029150593)*x[2]
        s[3,4,2,0]=(0.0559603779066)*x[0]+(-1.72752549101)*x[1]+(-1.02017554749)*x[2]
        s[3,4,2,1]=(-1.75424333815)*x[0]+(-0.198869840282)*x[1]+(0.475065396299)*x[2]
        ref[0,0,0,0]=(-0.270116752643)*x_ref[0]+(0.0715809194071)*x_ref[1]+(0.930011298681)*x_ref[2]
        ref[0,0,0,1]=(-0.0435807732016)*x_ref[0]+(-0.492622221729)*x_ref[1]+(-1.36754174971)*x_ref[2]
        ref[0,0,1,0]=(-0.949818131569)*x_ref[0]+(1.08437418341)*x_ref[1]+(0.527267447353)*x_ref[2]
        ref[0,0,1,1]=(-0.164876627537)*x_ref[0]+(-1.44143896699)*x_ref[1]+(0.0355140164081)*x_ref[2]
        ref[0,0,2,0]=(-0.530229192916)*x_ref[0]+(-0.988412045132)*x_ref[1]+(-0.466520393878)*x_ref[2]
        ref[0,0,2,1]=(-0.0166373496749)*x_ref[0]+(0.322621663585)*x_ref[1]+(-0.102019336059)*x_ref[2]
        ref[0,1,0,0]=(-1.36627178258)*x_ref[0]+(-0.803463758822)*x_ref[1]+(0.734446026661)*x_ref[2]
        ref[0,1,0,1]=(-0.804729478825)*x_ref[0]+(-1.24406753288)*x_ref[1]+(-0.755743791016)*x_ref[2]
        ref[0,1,1,0]=(1.03869027308)*x_ref[0]+(-0.25237293575)*x_ref[1]+(1.43737134487)*x_ref[2]
        ref[0,1,1,1]=(1.48269192251)*x_ref[0]+(0.045590449115)*x_ref[1]+(0.47425103841)*x_ref[2]
        ref[0,1,2,0]=(-0.738695287005)*x_ref[0]+(1.01938431903)*x_ref[1]+(1.20767534795)*x_ref[2]
        ref[0,1,2,1]=(-0.237445941287)*x_ref[0]+(0.766593746786)*x_ref[1]+(0.0297279831889)*x_ref[2]
        ref[0,2,0,0]=(-0.804902383375)*x_ref[0]+(0.0962339257662)*x_ref[1]+(1.37407473707)*x_ref[2]
        ref[0,2,0,1]=(-0.0821182376238)*x_ref[0]+(0.35489327051)*x_ref[1]+(-1.43235115632)*x_ref[2]
        ref[0,2,1,0]=(1.11593707013)*x_ref[0]+(0.744039820943)*x_ref[1]+(0.294815281064)*x_ref[2]
        ref[0,2,1,1]=(0.0789595478782)*x_ref[0]+(0.719656781874)*x_ref[1]+(0.355275528053)*x_ref[2]
        ref[0,2,2,0]=(1.20514986435)*x_ref[0]+(0.203004753589)*x_ref[1]+(-0.0731941650839)*x_ref[2]
        ref[0,2,2,1]=(-1.23231910051)*x_ref[0]+(-0.470834142341)*x_ref[1]+(-0.633518095593)*x_ref[2]
        ref[0,3,0,0]=(0.468164093965)*x_ref[0]+(-1.15996958069)*x_ref[1]+(0.922274695212)*x_ref[2]
        ref[0,3,0,1]=(-0.964250141036)*x_ref[0]+(-0.769747467017)*x_ref[1]+(-1.16369265956)*x_ref[2]
        ref[0,3,1,0]=(1.1622810944)*x_ref[0]+(-1.72794997517)*x_ref[1]+(-0.297833274612)*x_ref[2]
        ref[0,3,1,1]=(1.20239531447)*x_ref[0]+(-0.195111187444)*x_ref[1]+(-0.720186501147)*x_ref[2]
        ref[0,3,2,0]=(-0.0760001358746)*x_ref[0]+(-1.65266437471)*x_ref[1]+(-0.496163951015)*x_ref[2]
        ref[0,3,2,1]=(-1.4998248458)*x_ref[0]+(-1.73305168566)*x_ref[1]+(-0.107384086634)*x_ref[2]
        ref[0,4,0,0]=(0.235699357585)*x_ref[0]+(-0.213348404403)*x_ref[1]+(0.35368491512)*x_ref[2]
        ref[0,4,0,1]=(0.264641788778)*x_ref[0]+(0.435251584452)*x_ref[1]+(0.22539974851)*x_ref[2]
        ref[0,4,1,0]=(0.540203510472)*x_ref[0]+(-1.33243180984)*x_ref[1]+(-0.520357911324)*x_ref[2]
        ref[0,4,1,1]=(-0.629868492098)*x_ref[0]+(-0.403402885662)*x_ref[1]+(-0.184983842426)*x_ref[2]
        ref[0,4,2,0]=(1.17843494613)*x_ref[0]+(0.562783141759)*x_ref[1]+(-1.48575741726)*x_ref[2]
        ref[0,4,2,1]=(0.623796209289)*x_ref[0]+(1.02627846638)*x_ref[1]+(1.39369241754)*x_ref[2]
        ref[1,0,0,0]=(1.9855507901)*x_ref[0]+(0.152976407288)*x_ref[1]+(0.0931547545156)*x_ref[2]
        ref[1,0,0,1]=(-0.422055088151)*x_ref[0]+(-0.509449435814)*x_ref[1]+(-1.31705141661)*x_ref[2]
        ref[1,0,1,0]=(-0.24193816853)*x_ref[0]+(-0.68450536388)*x_ref[1]+(-0.309024887085)*x_ref[2]
        ref[1,0,1,1]=(1.24557169323)*x_ref[0]+(0.353538565492)*x_ref[1]+(0.115821365646)*x_ref[2]
        ref[1,0,2,0]=(-0.280246915004)*x_ref[0]+(-0.0972556765171)*x_ref[1]+(0.758816560386)*x_ref[2]
        ref[1,0,2,1]=(1.0396259589)*x_ref[0]+(-0.239507470189)*x_ref[1]+(-0.987815676492)*x_ref[2]
        ref[1,1,0,0]=(0.245203366317)*x_ref[0]+(-1.53269800659)*x_ref[1]+(0.619360300925)*x_ref[2]
        ref[1,1,0,1]=(0.246406467482)*x_ref[0]+(-0.729988292889)*x_ref[1]+(0.585422287189)*x_ref[2]
        ref[1,1,1,0]=(1.19810038121)*x_ref[0]+(-0.72772445701)*x_ref[1]+(1.11619392954)*x_ref[2]
        ref[1,1,1,1]=(-0.311984976134)*x_ref[0]+(-0.188443023344)*x_ref[1]+(-0.100078290654)*x_ref[2]
        ref[1,1,2,0]=(0.158578485066)*x_ref[0]+(0.214804573675)*x_ref[1]+(-0.22418434724)*x_ref[2]
        ref[1,1,2,1]=(-0.0550611413075)*x_ref[0]+(-0.147088267891)*x_ref[1]+(-0.269933900128)*x_ref[2]
        ref[1,2,0,0]=(-0.8618782319)*x_ref[0]+(-0.770284701216)*x_ref[1]+(-0.407458817279)*x_ref[2]
        ref[1,2,0,1]=(-0.242336600721)*x_ref[0]+(-1.57823911131)*x_ref[1]+(0.452055841078)*x_ref[2]
        ref[1,2,1,0]=(0.0537219652862)*x_ref[0]+(0.18864518829)*x_ref[1]+(0.021034116382)*x_ref[2]
        ref[1,2,1,1]=(0.143799709809)*x_ref[0]+(0.787934959856)*x_ref[1]+(-0.283457275339)*x_ref[2]
        ref[1,2,2,0]=(0.475965504456)*x_ref[0]+(-0.744950301051)*x_ref[1]+(0.45308618026)*x_ref[2]
        ref[1,2,2,1]=(0.561593580643)*x_ref[0]+(0.121626528038)*x_ref[1]+(-1.24221676348)*x_ref[2]
        ref[1,3,0,0]=(-0.340056043985)*x_ref[0]+(-0.434795551938)*x_ref[1]+(-0.291179084342)*x_ref[2]
        ref[1,3,0,1]=(0.312318149724)*x_ref[0]+(1.66950279503)*x_ref[1]+(-1.1622076779)*x_ref[2]
        ref[1,3,1,0]=(0.0965172054443)*x_ref[0]+(-0.64102022474)*x_ref[1]+(-0.173612340622)*x_ref[2]
        ref[1,3,1,1]=(1.88701305968)*x_ref[0]+(-1.66794136241)*x_ref[1]+(1.56059966446)*x_ref[2]
        ref[1,3,2,0]=(-0.497773999274)*x_ref[0]+(0.395345962943)*x_ref[1]+(-0.499105933847)*x_ref[2]
        ref[1,3,2,1]=(-0.312096437005)*x_ref[0]+(-1.23301763457)*x_ref[1]+(0.198835942067)*x_ref[2]
        ref[1,4,0,0]=(0.076377295342)*x_ref[0]+(-0.487847012612)*x_ref[1]+(-0.084314987823)*x_ref[2]
        ref[1,4,0,1]=(0.666434654203)*x_ref[0]+(-0.220376188514)*x_ref[1]+(-0.355749198488)*x_ref[2]
        ref[1,4,1,0]=(-0.701104468581)*x_ref[0]+(0.828174884422)*x_ref[1]+(0.309374641659)*x_ref[2]
        ref[1,4,1,1]=(0.526340204296)*x_ref[0]+(0.571598927007)*x_ref[1]+(-0.0793826750713)*x_ref[2]
        ref[1,4,2,0]=(0.775563727486)*x_ref[0]+(0.248228533771)*x_ref[1]+(-1.18169855975)*x_ref[2]
        ref[1,4,2,1]=(-0.399357449979)*x_ref[0]+(-0.723554468137)*x_ref[1]+(0.0231913550517)*x_ref[2]
        ref[2,0,0,0]=(-0.0504735749489)*x_ref[0]+(0.506564345355)*x_ref[1]+(0.790336060204)*x_ref[2]
        ref[2,0,0,1]=(0.426187262022)*x_ref[0]+(0.966680952632)*x_ref[1]+(0.410542727369)*x_ref[2]
        ref[2,0,1,0]=(0.594102954821)*x_ref[0]+(0.0639755896611)*x_ref[1]+(-0.675769510238)*x_ref[2]
        ref[2,0,1,1]=(0.367479711308)*x_ref[0]+(0.150736830273)*x_ref[1]+(1.15797942999)*x_ref[2]
        ref[2,0,2,0]=(0.875207101066)*x_ref[0]+(-1.12939492894)*x_ref[1]+(0.73561332295)*x_ref[2]
        ref[2,0,2,1]=(-1.70885193598)*x_ref[0]+(0.835257764029)*x_ref[1]+(-1.58521909429)*x_ref[2]
        ref[2,1,0,0]=(-0.144255046397)*x_ref[0]+(0.590074178586)*x_ref[1]+(0.592144641271)*x_ref[2]
        ref[2,1,0,1]=(-0.0688470601204)*x_ref[0]+(-0.153340975531)*x_ref[1]+(1.15241895553)*x_ref[2]
        ref[2,1,1,0]=(-0.73257033584)*x_ref[0]+(-0.563316507818)*x_ref[1]+(-1.05620654545)*x_ref[2]
        ref[2,1,1,1]=(-1.29833350092)*x_ref[0]+(0.42237934165)*x_ref[1]+(0.0725296431412)*x_ref[2]
        ref[2,1,2,0]=(-0.837473685756)*x_ref[0]+(-0.425072509905)*x_ref[1]+(-0.463397001131)*x_ref[2]
        ref[2,1,2,1]=(-0.937038584847)*x_ref[0]+(0.531989188194)*x_ref[1]+(-1.06694316698)*x_ref[2]
        ref[2,2,0,0]=(-0.846346120269)*x_ref[0]+(-0.639618775615)*x_ref[1]+(-0.928154018676)*x_ref[2]
        ref[2,2,0,1]=(-0.726631889289)*x_ref[0]+(-0.1318624965)*x_ref[1]+(-0.343729797494)*x_ref[2]
        ref[2,2,1,0]=(0.802665744539)*x_ref[0]+(-0.115632355474)*x_ref[1]+(0.303919795863)*x_ref[2]
        ref[2,2,1,1]=(0.0371825373918)*x_ref[0]+(-1.17799388388)*x_ref[1]+(-0.223043449617)*x_ref[2]
        ref[2,2,2,0]=(0.75988617552)*x_ref[0]+(-0.608977501492)*x_ref[1]+(1.75211913311)*x_ref[2]
        ref[2,2,2,1]=(-0.291397991509)*x_ref[0]+(-1.55173354063)*x_ref[1]+(-0.461269447756)*x_ref[2]
        ref[2,3,0,0]=(0.351624965509)*x_ref[0]+(-1.32799214259)*x_ref[1]+(1.58946725409)*x_ref[2]
        ref[2,3,0,1]=(0.180149992744)*x_ref[0]+(1.07861987473)*x_ref[1]+(0.834357928788)*x_ref[2]
        ref[2,3,1,0]=(0.0475016328103)*x_ref[0]+(0.707449786509)*x_ref[1]+(-0.370421903676)*x_ref[2]
        ref[2,3,1,1]=(0.553988297947)*x_ref[0]+(1.13388241001)*x_ref[1]+(0.575604300083)*x_ref[2]
        ref[2,3,2,0]=(-0.249047035016)*x_ref[0]+(0.15013128589)*x_ref[1]+(1.3953926353)*x_ref[2]
        ref[2,3,2,1]=(-0.783008188949)*x_ref[0]+(0.861992436037)*x_ref[1]+(-0.929496477422)*x_ref[2]
        ref[2,4,0,0]=(0.264348996811)*x_ref[0]+(-1.23207394036)*x_ref[1]+(-1.37804290092)*x_ref[2]
        ref[2,4,0,1]=(0.360782273055)*x_ref[0]+(-0.270520439821)*x_ref[1]+(-1.01285201623)*x_ref[2]
        ref[2,4,1,0]=(0.280578579439)*x_ref[0]+(0.584696145403)*x_ref[1]+(-1.21852764117)*x_ref[2]
        ref[2,4,1,1]=(0.112102121128)*x_ref[0]+(-0.0619663521973)*x_ref[1]+(1.08490293257)*x_ref[2]
        ref[2,4,2,0]=(-1.89749569924)*x_ref[0]+(-0.21217889191)*x_ref[1]+(0.635524779965)*x_ref[2]
        ref[2,4,2,1]=(-0.309239337357)*x_ref[0]+(0.0116284822724)*x_ref[1]+(0.0107247712951)*x_ref[2]
        ref[3,0,0,0]=(1.23545873872)*x_ref[0]+(-1.29896744411)*x_ref[1]+(0.855094335492)*x_ref[2]
        ref[3,0,0,1]=(-0.51465695324)*x_ref[0]+(1.47799693308)*x_ref[1]+(0.0718360423316)*x_ref[2]
        ref[3,0,1,0]=(0.220505442402)*x_ref[0]+(-0.3171385194)*x_ref[1]+(0.27954628665)*x_ref[2]
        ref[3,0,1,1]=(0.860582046972)*x_ref[0]+(1.04565549687)*x_ref[1]+(1.18108647078)*x_ref[2]
        ref[3,0,2,0]=(-0.593547630941)*x_ref[0]+(-0.0390465590147)*x_ref[1]+(0.00129707682534)*x_ref[2]
        ref[3,0,2,1]=(1.39210293617)*x_ref[0]+(-1.05660957806)*x_ref[1]+(-1.20489109051)*x_ref[2]
        ref[3,1,0,0]=(-0.536035988946)*x_ref[0]+(-0.594403345465)*x_ref[1]+(1.11719111746)*x_ref[2]
        ref[3,1,0,1]=(-0.0403557954127)*x_ref[0]+(-1.32391025453)*x_ref[1]+(-0.639127329416)*x_ref[2]
        ref[3,1,1,0]=(-0.495803710666)*x_ref[0]+(-0.199832989186)*x_ref[1]+(-0.832862076561)*x_ref[2]
        ref[3,1,1,1]=(0.0216882402803)*x_ref[0]+(0.367268873436)*x_ref[1]+(0.295944335418)*x_ref[2]
        ref[3,1,2,0]=(0.172936506269)*x_ref[0]+(-0.339886198196)*x_ref[1]+(0.268201481244)*x_ref[2]
        ref[3,1,2,1]=(0.665786569897)*x_ref[0]+(-0.290637226736)*x_ref[1]+(0.617677178537)*x_ref[2]
        ref[3,2,0,0]=(-0.657902185911)*x_ref[0]+(-1.34292379802)*x_ref[1]+(0.0641950819134)*x_ref[2]
        ref[3,2,0,1]=(0.352039175923)*x_ref[0]+(-0.483047873735)*x_ref[1]+(0.47997600743)*x_ref[2]
        ref[3,2,1,0]=(0.944619484106)*x_ref[0]+(-0.308755547249)*x_ref[1]+(-0.765906593989)*x_ref[2]
        ref[3,2,1,1]=(-1.09674781865)*x_ref[0]+(0.013069257236)*x_ref[1]+(-0.0741247275628)*x_ref[2]
        ref[3,2,2,0]=(0.400057737624)*x_ref[0]+(-1.16165281514)*x_ref[1]+(-0.789757454741)*x_ref[2]
        ref[3,2,2,1]=(0.367100312493)*x_ref[0]+(-0.547338908792)*x_ref[1]+(-0.409173859392)*x_ref[2]
        ref[3,3,0,0]=(0.890118369638)*x_ref[0]+(1.201296587)*x_ref[1]+(-0.0319246476545)*x_ref[2]
        ref[3,3,0,1]=(0.620955702501)*x_ref[0]+(0.438440818114)*x_ref[1]+(0.499747510411)*x_ref[2]
        ref[3,3,1,0]=(1.61217537222)*x_ref[0]+(-1.21323714148)*x_ref[1]+(-1.06137097571)*x_ref[2]
        ref[3,3,1,1]=(-1.75942413542)*x_ref[0]+(0.440345678738)*x_ref[1]+(0.625299774096)*x_ref[2]
        ref[3,3,2,0]=(-0.878497266865)*x_ref[0]+(-1.15879244679)*x_ref[1]+(1.03993124425)*x_ref[2]
        ref[3,3,2,1]=(-0.0138445968491)*x_ref[0]+(0.547910692711)*x_ref[1]+(0.971642264613)*x_ref[2]
        ref[3,4,0,0]=(-0.107769272336)*x_ref[0]+(-0.987911384393)*x_ref[1]+(1.2076201556)*x_ref[2]
        ref[3,4,0,1]=(-0.556570952071)*x_ref[0]+(-0.284568452888)*x_ref[1]+(1.23341917108)*x_ref[2]
        ref[3,4,1,0]=(1.56461193947)*x_ref[0]+(0.185382898933)*x_ref[1]+(-1.67714180845)*x_ref[2]
        ref[3,4,1,1]=(-0.894934034676)*x_ref[0]+(1.05560233139)*x_ref[1]+(0.651029150593)*x_ref[2]
        ref[3,4,2,0]=(0.0559603779066)*x_ref[0]+(-1.72752549101)*x_ref[1]+(-1.02017554749)*x_ref[2]
        ref[3,4,2,1]=(-1.75424333815)*x_ref[0]+(-0.198869840282)*x_ref[1]+(0.475065396299)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedContinuousFunction_rank0(self):
      """
      tests interpolate for rank 0 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(),dim=dim)
      s=Data(0,(),w)
      ref=Data(0,(),w_ref)
      if dim==2:
        s=(-1.05153531037)*x[0]+(1.5374190245)*x[1]
        ref=(-1.05153531037)*x_ref[0]+(1.5374190245)*x_ref[1]
      else:
        s=(-1.91883881043)*x[0]+(-0.414033171764)*x[1]+(1.96173093572)*x[2]
        ref=(-1.91883881043)*x_ref[0]+(-0.414033171764)*x_ref[1]+(1.96173093572)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedContinuousFunction_rank1(self):
      """
      tests interpolate for rank 1 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(2,),dim=dim)
      s=Data(0,(2,),w)
      ref=Data(0,(2,),w_ref)
      if dim==2:
        s[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        s[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
        ref[0]=(-0.465304303454)*x[0]+(0.0254131991036)*x[1]
        ref[1]=(-0.538872626463)*x[0]+(0.187837758076)*x[1]
      else:
        s[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        s[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
        ref[0]=(1.62562954907)*x[0]+(0.0557929062682)*x[1]+(1.28261964503)*x[2]
        ref[1]=(1.22016241295)*x[0]+(-1.55194071457)*x[1]+(0.798935989838)*x[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2,),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedContinuousFunction_rank2(self):
      """
      tests interpolate for rank 2 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5),dim=dim)
      s=Data(0,(4, 5),w)
      ref=Data(0,(4, 5),w_ref)
      if dim==2:
        s[0,0]=(-0.581359229868)*x[0]+(-0.405627827004)*x[1]
        s[0,1]=(1.15883572458)*x[0]+(-0.212383708658)*x[1]
        s[0,2]=(-0.675196965148)*x[0]+(-0.0589289579115)*x[1]
        s[0,3]=(-0.466142003406)*x[0]+(1.143298366)*x[1]
        s[0,4]=(0.706585901476)*x[0]+(-0.596587020578)*x[1]
        s[1,0]=(0.150862770285)*x[0]+(1.3486848087)*x[1]
        s[1,1]=(0.694050128425)*x[0]+(0.919192282198)*x[1]
        s[1,2]=(0.123568652464)*x[0]+(-1.1901501926)*x[1]
        s[1,3]=(-0.0147792668086)*x[0]+(-1.49381121537)*x[1]
        s[1,4]=(0.0737831649537)*x[0]+(-0.752053730242)*x[1]
        s[2,0]=(-0.758812518739)*x[0]+(-1.26143124516)*x[1]
        s[2,1]=(0.507459828891)*x[0]+(-0.0315665691912)*x[1]
        s[2,2]=(-0.492645010187)*x[0]+(-0.649596364388)*x[1]
        s[2,3]=(-0.0632358844144)*x[0]+(0.899340082527)*x[1]
        s[2,4]=(-0.763180407153)*x[0]+(-0.269808126268)*x[1]
        s[3,0]=(-1.5724790938)*x[0]+(1.16942939992)*x[1]
        s[3,1]=(0.759317295371)*x[0]+(0.14287823082)*x[1]
        s[3,2]=(0.70712973662)*x[0]+(1.08455803224)*x[1]
        s[3,3]=(-0.736730602555)*x[0]+(-0.183785007602)*x[1]
        s[3,4]=(-0.950538745165)*x[0]+(0.294140958329)*x[1]
        ref[0,0]=(-0.581359229868)*x_ref[0]+(-0.405627827004)*x_ref[1]
        ref[0,1]=(1.15883572458)*x_ref[0]+(-0.212383708658)*x_ref[1]
        ref[0,2]=(-0.675196965148)*x_ref[0]+(-0.0589289579115)*x_ref[1]
        ref[0,3]=(-0.466142003406)*x_ref[0]+(1.143298366)*x_ref[1]
        ref[0,4]=(0.706585901476)*x_ref[0]+(-0.596587020578)*x_ref[1]
        ref[1,0]=(0.150862770285)*x_ref[0]+(1.3486848087)*x_ref[1]
        ref[1,1]=(0.694050128425)*x_ref[0]+(0.919192282198)*x_ref[1]
        ref[1,2]=(0.123568652464)*x_ref[0]+(-1.1901501926)*x_ref[1]
        ref[1,3]=(-0.0147792668086)*x_ref[0]+(-1.49381121537)*x_ref[1]
        ref[1,4]=(0.0737831649537)*x_ref[0]+(-0.752053730242)*x_ref[1]
        ref[2,0]=(-0.758812518739)*x_ref[0]+(-1.26143124516)*x_ref[1]
        ref[2,1]=(0.507459828891)*x_ref[0]+(-0.0315665691912)*x_ref[1]
        ref[2,2]=(-0.492645010187)*x_ref[0]+(-0.649596364388)*x_ref[1]
        ref[2,3]=(-0.0632358844144)*x_ref[0]+(0.899340082527)*x_ref[1]
        ref[2,4]=(-0.763180407153)*x_ref[0]+(-0.269808126268)*x_ref[1]
        ref[3,0]=(-1.5724790938)*x_ref[0]+(1.16942939992)*x_ref[1]
        ref[3,1]=(0.759317295371)*x_ref[0]+(0.14287823082)*x_ref[1]
        ref[3,2]=(0.70712973662)*x_ref[0]+(1.08455803224)*x_ref[1]
        ref[3,3]=(-0.736730602555)*x_ref[0]+(-0.183785007602)*x_ref[1]
        ref[3,4]=(-0.950538745165)*x_ref[0]+(0.294140958329)*x_ref[1]
      else:
        s[0,0]=(0.310668873066)*x[0]+(1.66326709742)*x[1]+(1.10140064967)*x[2]
        s[0,1]=(0.46962350981)*x[0]+(-1.02292998305)*x[1]+(0.919666772954)*x[2]
        s[0,2]=(-1.18114779319)*x[0]+(1.23819669136)*x[1]+(0.0622519749551)*x[2]
        s[0,3]=(1.08356504043)*x[0]+(0.450517408398)*x[1]+(-0.528809988127)*x[2]
        s[0,4]=(-0.179349520643)*x[0]+(0.192827326205)*x[1]+(-0.543850326813)*x[2]
        s[1,0]=(-1.60006459477)*x[0]+(0.800497908511)*x[1]+(0.745234766014)*x[2]
        s[1,1]=(0.667355662535)*x[0]+(0.800303765538)*x[1]+(-1.17397311427)*x[2]
        s[1,2]=(0.19342327722)*x[0]+(-0.149618364398)*x[1]+(-1.06968763194)*x[2]
        s[1,3]=(0.296694301988)*x[0]+(-1.35691147932)*x[1]+(0.748623252633)*x[2]
        s[1,4]=(0.724332288795)*x[0]+(0.349026497158)*x[1]+(0.377755550832)*x[2]
        s[2,0]=(0.560218327038)*x[0]+(0.184262569787)*x[1]+(-0.821718588725)*x[2]
        s[2,1]=(0.421969037007)*x[0]+(-0.00510504728331)*x[1]+(1.43225098227)*x[2]
        s[2,2]=(0.889590883145)*x[0]+(-0.408365128114)*x[1]+(-1.80100224461)*x[2]
        s[2,3]=(-1.91735498201)*x[0]+(0.590584386339)*x[1]+(-0.905610066725)*x[2]
        s[2,4]=(-0.367106108866)*x[0]+(-0.92493020573)*x[1]+(-0.0746199487091)*x[2]
        s[3,0]=(0.919247371049)*x[0]+(0.326674777975)*x[1]+(-1.47385875396)*x[2]
        s[3,1]=(-0.850835329183)*x[0]+(-0.536914325383)*x[1]+(1.02264919467)*x[2]
        s[3,2]=(-1.41905423752)*x[0]+(-0.632415273773)*x[1]+(0.842187325237)*x[2]
        s[3,3]=(-0.396129693721)*x[0]+(-0.110878220479)*x[1]+(0.49355539737)*x[2]
        s[3,4]=(-0.0781905590856)*x[0]+(-0.0348772309292)*x[1]+(0.184581658536)*x[2]
        ref[0,0]=(0.310668873066)*x_ref[0]+(1.66326709742)*x_ref[1]+(1.10140064967)*x_ref[2]
        ref[0,1]=(0.46962350981)*x_ref[0]+(-1.02292998305)*x_ref[1]+(0.919666772954)*x_ref[2]
        ref[0,2]=(-1.18114779319)*x_ref[0]+(1.23819669136)*x_ref[1]+(0.0622519749551)*x_ref[2]
        ref[0,3]=(1.08356504043)*x_ref[0]+(0.450517408398)*x_ref[1]+(-0.528809988127)*x_ref[2]
        ref[0,4]=(-0.179349520643)*x_ref[0]+(0.192827326205)*x_ref[1]+(-0.543850326813)*x_ref[2]
        ref[1,0]=(-1.60006459477)*x_ref[0]+(0.800497908511)*x_ref[1]+(0.745234766014)*x_ref[2]
        ref[1,1]=(0.667355662535)*x_ref[0]+(0.800303765538)*x_ref[1]+(-1.17397311427)*x_ref[2]
        ref[1,2]=(0.19342327722)*x_ref[0]+(-0.149618364398)*x_ref[1]+(-1.06968763194)*x_ref[2]
        ref[1,3]=(0.296694301988)*x_ref[0]+(-1.35691147932)*x_ref[1]+(0.748623252633)*x_ref[2]
        ref[1,4]=(0.724332288795)*x_ref[0]+(0.349026497158)*x_ref[1]+(0.377755550832)*x_ref[2]
        ref[2,0]=(0.560218327038)*x_ref[0]+(0.184262569787)*x_ref[1]+(-0.821718588725)*x_ref[2]
        ref[2,1]=(0.421969037007)*x_ref[0]+(-0.00510504728331)*x_ref[1]+(1.43225098227)*x_ref[2]
        ref[2,2]=(0.889590883145)*x_ref[0]+(-0.408365128114)*x_ref[1]+(-1.80100224461)*x_ref[2]
        ref[2,3]=(-1.91735498201)*x_ref[0]+(0.590584386339)*x_ref[1]+(-0.905610066725)*x_ref[2]
        ref[2,4]=(-0.367106108866)*x_ref[0]+(-0.92493020573)*x_ref[1]+(-0.0746199487091)*x_ref[2]
        ref[3,0]=(0.919247371049)*x_ref[0]+(0.326674777975)*x_ref[1]+(-1.47385875396)*x_ref[2]
        ref[3,1]=(-0.850835329183)*x_ref[0]+(-0.536914325383)*x_ref[1]+(1.02264919467)*x_ref[2]
        ref[3,2]=(-1.41905423752)*x_ref[0]+(-0.632415273773)*x_ref[1]+(0.842187325237)*x_ref[2]
        ref[3,3]=(-0.396129693721)*x_ref[0]+(-0.110878220479)*x_ref[1]+(0.49355539737)*x_ref[2]
        ref[3,4]=(-0.0781905590856)*x_ref[0]+(-0.0348772309292)*x_ref[1]+(0.184581658536)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedContinuousFunction_rank3(self):
      """
      tests interpolate for rank 3 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(6, 2, 2),dim=dim)
      s=Data(0,(6, 2, 2),w)
      ref=Data(0,(6, 2, 2),w_ref)
      if dim==2:
        s[0,0,0]=(0.540497015605)*x[0]+(0.282753164584)*x[1]
        s[0,0,1]=(-0.127429073679)*x[0]+(-0.709993735868)*x[1]
        s[0,1,0]=(1.02019673589)*x[0]+(-1.53422031862)*x[1]
        s[0,1,1]=(0.926292918083)*x[0]+(-0.0351188336657)*x[1]
        s[1,0,0]=(-0.736080207248)*x[0]+(1.25260396336)*x[1]
        s[1,0,1]=(0.152653303192)*x[0]+(0.22547444598)*x[1]
        s[1,1,0]=(0.729526332632)*x[0]+(0.919207155406)*x[1]
        s[1,1,1]=(-0.346325378686)*x[0]+(0.141030274609)*x[1]
        s[2,0,0]=(-0.753354302966)*x[0]+(-0.766294422747)*x[1]
        s[2,0,1]=(1.17251569201)*x[0]+(-0.51673325599)*x[1]
        s[2,1,0]=(0.66192662328)*x[0]+(-0.572345218315)*x[1]
        s[2,1,1]=(1.74689126451)*x[0]+(-0.0537684141408)*x[1]
        s[3,0,0]=(0.300593672769)*x[0]+(0.0833505471404)*x[1]
        s[3,0,1]=(0.770671590417)*x[0]+(-0.667020549164)*x[1]
        s[3,1,0]=(0.281444608616)*x[0]+(0.701762037742)*x[1]
        s[3,1,1]=(0.404631442844)*x[0]+(0.337744807226)*x[1]
        s[4,0,0]=(-1.43357362276)*x[0]+(0.687631941644)*x[1]
        s[4,0,1]=(0.694571843151)*x[0]+(0.979495014797)*x[1]
        s[4,1,0]=(-0.0880858702517)*x[0]+(0.729298516002)*x[1]
        s[4,1,1]=(0.641170415423)*x[0]+(-0.407278118417)*x[1]
        s[5,0,0]=(-0.60364366976)*x[0]+(-0.663798735078)*x[1]
        s[5,0,1]=(-0.841035861641)*x[0]+(-1.09892876031)*x[1]
        s[5,1,0]=(0.0838180897314)*x[0]+(0.15765944496)*x[1]
        s[5,1,1]=(-0.847291245273)*x[0]+(1.01145652383)*x[1]
        ref[0,0,0]=(0.540497015605)*x_ref[0]+(0.282753164584)*x_ref[1]
        ref[0,0,1]=(-0.127429073679)*x_ref[0]+(-0.709993735868)*x_ref[1]
        ref[0,1,0]=(1.02019673589)*x_ref[0]+(-1.53422031862)*x_ref[1]
        ref[0,1,1]=(0.926292918083)*x_ref[0]+(-0.0351188336657)*x_ref[1]
        ref[1,0,0]=(-0.736080207248)*x_ref[0]+(1.25260396336)*x_ref[1]
        ref[1,0,1]=(0.152653303192)*x_ref[0]+(0.22547444598)*x_ref[1]
        ref[1,1,0]=(0.729526332632)*x_ref[0]+(0.919207155406)*x_ref[1]
        ref[1,1,1]=(-0.346325378686)*x_ref[0]+(0.141030274609)*x_ref[1]
        ref[2,0,0]=(-0.753354302966)*x_ref[0]+(-0.766294422747)*x_ref[1]
        ref[2,0,1]=(1.17251569201)*x_ref[0]+(-0.51673325599)*x_ref[1]
        ref[2,1,0]=(0.66192662328)*x_ref[0]+(-0.572345218315)*x_ref[1]
        ref[2,1,1]=(1.74689126451)*x_ref[0]+(-0.0537684141408)*x_ref[1]
        ref[3,0,0]=(0.300593672769)*x_ref[0]+(0.0833505471404)*x_ref[1]
        ref[3,0,1]=(0.770671590417)*x_ref[0]+(-0.667020549164)*x_ref[1]
        ref[3,1,0]=(0.281444608616)*x_ref[0]+(0.701762037742)*x_ref[1]
        ref[3,1,1]=(0.404631442844)*x_ref[0]+(0.337744807226)*x_ref[1]
        ref[4,0,0]=(-1.43357362276)*x_ref[0]+(0.687631941644)*x_ref[1]
        ref[4,0,1]=(0.694571843151)*x_ref[0]+(0.979495014797)*x_ref[1]
        ref[4,1,0]=(-0.0880858702517)*x_ref[0]+(0.729298516002)*x_ref[1]
        ref[4,1,1]=(0.641170415423)*x_ref[0]+(-0.407278118417)*x_ref[1]
        ref[5,0,0]=(-0.60364366976)*x_ref[0]+(-0.663798735078)*x_ref[1]
        ref[5,0,1]=(-0.841035861641)*x_ref[0]+(-1.09892876031)*x_ref[1]
        ref[5,1,0]=(0.0838180897314)*x_ref[0]+(0.15765944496)*x_ref[1]
        ref[5,1,1]=(-0.847291245273)*x_ref[0]+(1.01145652383)*x_ref[1]
      else:
        s[0,0,0]=(-0.29326741782)*x[0]+(0.716468869525)*x[1]+(-0.194120207009)*x[2]
        s[0,0,1]=(-0.0687304327556)*x[0]+(0.181789058707)*x[1]+(-0.940504268245)*x[2]
        s[0,1,0]=(0.437145055905)*x[0]+(1.72039421904)*x[1]+(-0.0326204183567)*x[2]
        s[0,1,1]=(1.25919532639)*x[0]+(0.082561920762)*x[1]+(1.01503353491)*x[2]
        s[1,0,0]=(-0.552129852247)*x[0]+(-0.661963725391)*x[1]+(0.191012614278)*x[2]
        s[1,0,1]=(-0.322228468989)*x[0]+(-0.79555280135)*x[1]+(0.0689420056626)*x[2]
        s[1,1,0]=(0.294413007041)*x[0]+(-0.453661316564)*x[1]+(1.74072519115)*x[2]
        s[1,1,1]=(-0.13491214165)*x[0]+(-1.18907003959)*x[1]+(-1.59325361447)*x[2]
        s[2,0,0]=(-0.522689030297)*x[0]+(-0.784658318696)*x[1]+(0.618343699514)*x[2]
        s[2,0,1]=(-0.227214629419)*x[0]+(0.770163719023)*x[1]+(-0.272804476799)*x[2]
        s[2,1,0]=(-1.28131384809)*x[0]+(0.0591076357001)*x[1]+(-0.206570950984)*x[2]
        s[2,1,1]=(0.00241150481435)*x[0]+(0.0678794618244)*x[1]+(-0.272677008307)*x[2]
        s[3,0,0]=(-0.537194244467)*x[0]+(-0.886303700791)*x[1]+(0.704423993884)*x[2]
        s[3,0,1]=(0.0924777549448)*x[0]+(-0.783168034362)*x[1]+(-0.450930098538)*x[2]
        s[3,1,0]=(0.611562862859)*x[0]+(-1.04224402701)*x[1]+(-0.775594008735)*x[2]
        s[3,1,1]=(1.44402445157)*x[0]+(-0.736290409307)*x[1]+(0.51690502447)*x[2]
        s[4,0,0]=(-0.470210375338)*x[0]+(-0.433261543227)*x[1]+(-0.582080162476)*x[2]
        s[4,0,1]=(0.313323330562)*x[0]+(-1.47502042219)*x[1]+(-1.50202564103)*x[2]
        s[4,1,0]=(-0.807949755771)*x[0]+(-0.162904810555)*x[1]+(0.748809317029)*x[2]
        s[4,1,1]=(-1.24228935643)*x[0]+(0.193414583713)*x[1]+(0.85984063481)*x[2]
        s[5,0,0]=(1.08068739991)*x[0]+(-1.30655986333)*x[1]+(0.289800143967)*x[2]
        s[5,0,1]=(1.12114019122)*x[0]+(0.280561086566)*x[1]+(0.844991094994)*x[2]
        s[5,1,0]=(-0.133922711479)*x[0]+(-1.39747129717)*x[1]+(-1.11382363437)*x[2]
        s[5,1,1]=(-1.34164423823)*x[0]+(-0.335643322772)*x[1]+(0.757067990421)*x[2]
        ref[0,0,0]=(-0.29326741782)*x_ref[0]+(0.716468869525)*x_ref[1]+(-0.194120207009)*x_ref[2]
        ref[0,0,1]=(-0.0687304327556)*x_ref[0]+(0.181789058707)*x_ref[1]+(-0.940504268245)*x_ref[2]
        ref[0,1,0]=(0.437145055905)*x_ref[0]+(1.72039421904)*x_ref[1]+(-0.0326204183567)*x_ref[2]
        ref[0,1,1]=(1.25919532639)*x_ref[0]+(0.082561920762)*x_ref[1]+(1.01503353491)*x_ref[2]
        ref[1,0,0]=(-0.552129852247)*x_ref[0]+(-0.661963725391)*x_ref[1]+(0.191012614278)*x_ref[2]
        ref[1,0,1]=(-0.322228468989)*x_ref[0]+(-0.79555280135)*x_ref[1]+(0.0689420056626)*x_ref[2]
        ref[1,1,0]=(0.294413007041)*x_ref[0]+(-0.453661316564)*x_ref[1]+(1.74072519115)*x_ref[2]
        ref[1,1,1]=(-0.13491214165)*x_ref[0]+(-1.18907003959)*x_ref[1]+(-1.59325361447)*x_ref[2]
        ref[2,0,0]=(-0.522689030297)*x_ref[0]+(-0.784658318696)*x_ref[1]+(0.618343699514)*x_ref[2]
        ref[2,0,1]=(-0.227214629419)*x_ref[0]+(0.770163719023)*x_ref[1]+(-0.272804476799)*x_ref[2]
        ref[2,1,0]=(-1.28131384809)*x_ref[0]+(0.0591076357001)*x_ref[1]+(-0.206570950984)*x_ref[2]
        ref[2,1,1]=(0.00241150481435)*x_ref[0]+(0.0678794618244)*x_ref[1]+(-0.272677008307)*x_ref[2]
        ref[3,0,0]=(-0.537194244467)*x_ref[0]+(-0.886303700791)*x_ref[1]+(0.704423993884)*x_ref[2]
        ref[3,0,1]=(0.0924777549448)*x_ref[0]+(-0.783168034362)*x_ref[1]+(-0.450930098538)*x_ref[2]
        ref[3,1,0]=(0.611562862859)*x_ref[0]+(-1.04224402701)*x_ref[1]+(-0.775594008735)*x_ref[2]
        ref[3,1,1]=(1.44402445157)*x_ref[0]+(-0.736290409307)*x_ref[1]+(0.51690502447)*x_ref[2]
        ref[4,0,0]=(-0.470210375338)*x_ref[0]+(-0.433261543227)*x_ref[1]+(-0.582080162476)*x_ref[2]
        ref[4,0,1]=(0.313323330562)*x_ref[0]+(-1.47502042219)*x_ref[1]+(-1.50202564103)*x_ref[2]
        ref[4,1,0]=(-0.807949755771)*x_ref[0]+(-0.162904810555)*x_ref[1]+(0.748809317029)*x_ref[2]
        ref[4,1,1]=(-1.24228935643)*x_ref[0]+(0.193414583713)*x_ref[1]+(0.85984063481)*x_ref[2]
        ref[5,0,0]=(1.08068739991)*x_ref[0]+(-1.30655986333)*x_ref[1]+(0.289800143967)*x_ref[2]
        ref[5,0,1]=(1.12114019122)*x_ref[0]+(0.280561086566)*x_ref[1]+(0.844991094994)*x_ref[2]
        ref[5,1,0]=(-0.133922711479)*x_ref[0]+(-1.39747129717)*x_ref[1]+(-1.11382363437)*x_ref[2]
        ref[5,1,1]=(-1.34164423823)*x_ref[0]+(-0.335643322772)*x_ref[1]+(0.757067990421)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(6, 2, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_interpolate_onReducedContinuousFunction_fromSymbol_ReducedContinuousFunction_rank4(self):
      """
      tests interpolate for rank 4 Symbol onto the ReducedSolution

      assumptions: self.domain supports inpterpolation from ReducedContinuousFunction onto ReducedSolution
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedContinuousFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
      s=Data(0,(4, 5, 3, 2),w)
      ref=Data(0,(4, 5, 3, 2),w_ref)
      if dim==2:
        s[0,0,0,0]=(-0.9809347951)*x[0]+(-1.05666590015)*x[1]
        s[0,0,0,1]=(1.15925248094)*x[0]+(1.22070923582)*x[1]
        s[0,0,1,0]=(-0.0724361628927)*x[0]+(-0.439807238916)*x[1]
        s[0,0,1,1]=(-0.608816713414)*x[0]+(1.78792604745)*x[1]
        s[0,0,2,0]=(-0.78903506363)*x[0]+(-0.75896072902)*x[1]
        s[0,0,2,1]=(0.516424260522)*x[0]+(1.16788015378)*x[1]
        s[0,1,0,0]=(0.383586654748)*x[0]+(1.4179924146)*x[1]
        s[0,1,0,1]=(0.97236907624)*x[0]+(-0.438982550276)*x[1]
        s[0,1,1,0]=(-0.645083297783)*x[0]+(-0.740808997218)*x[1]
        s[0,1,1,1]=(0.330652781935)*x[0]+(0.519132367114)*x[1]
        s[0,1,2,0]=(1.25159174261)*x[0]+(1.26021904904)*x[1]
        s[0,1,2,1]=(-0.239621293462)*x[0]+(0.0244107517523)*x[1]
        s[0,2,0,0]=(-0.286208240697)*x[0]+(-0.314157258671)*x[1]
        s[0,2,0,1]=(-0.0856319829397)*x[0]+(0.278603809503)*x[1]
        s[0,2,1,0]=(0.107199217139)*x[0]+(0.0948664401127)*x[1]
        s[0,2,1,1]=(-0.24620060163)*x[0]+(-1.42942082558)*x[1]
        s[0,2,2,0]=(0.845285367352)*x[0]+(0.408676159609)*x[1]
        s[0,2,2,1]=(-1.49467460394)*x[0]+(0.270722457265)*x[1]
        s[0,3,0,0]=(0.469802598061)*x[0]+(-0.801790345348)*x[1]
        s[0,3,0,1]=(1.71249257903)*x[0]+(-0.776797378881)*x[1]
        s[0,3,1,0]=(0.0540049036319)*x[0]+(1.19031969164)*x[1]
        s[0,3,1,1]=(0.161929130968)*x[0]+(-0.525381566632)*x[1]
        s[0,3,2,0]=(-0.646041965665)*x[0]+(-0.123993555032)*x[1]
        s[0,3,2,1]=(0.0966044205198)*x[0]+(1.24780552685)*x[1]
        s[0,4,0,0]=(-0.289531041539)*x[0]+(1.21117470928)*x[1]
        s[0,4,0,1]=(-0.112215925799)*x[0]+(0.598717725332)*x[1]
        s[0,4,1,0]=(0.162936056701)*x[0]+(-1.02395851472)*x[1]
        s[0,4,1,1]=(0.694066364547)*x[0]+(1.27205020727)*x[1]
        s[0,4,2,0]=(1.17253981195)*x[0]+(0.130572791304)*x[1]
        s[0,4,2,1]=(-0.555241229093)*x[0]+(-0.881115061403)*x[1]
        s[1,0,0,0]=(0.574137047465)*x[0]+(-1.10459006911)*x[1]
        s[1,0,0,1]=(0.441431076852)*x[0]+(-0.133156832361)*x[1]
        s[1,0,1,0]=(0.217136655494)*x[0]+(0.400834952015)*x[1]
        s[1,0,1,1]=(0.516360280675)*x[0]+(0.532907430046)*x[1]
        s[1,0,2,0]=(-0.340764752946)*x[0]+(-1.26671004418)*x[1]
        s[1,0,2,1]=(-1.23231139783)*x[0]+(0.61934930883)*x[1]
        s[1,1,0,0]=(1.76821949014)*x[0]+(1.67542625016)*x[1]
        s[1,1,0,1]=(0.901675970272)*x[0]+(-0.0567109697208)*x[1]
        s[1,1,1,0]=(1.21143551644)*x[0]+(0.375877605521)*x[1]
        s[1,1,1,1]=(-0.699830133987)*x[0]+(0.174660968691)*x[1]
        s[1,1,2,0]=(1.07839763024)*x[0]+(-0.063554858486)*x[1]
        s[1,1,2,1]=(0.131335396282)*x[0]+(0.109893568063)*x[1]
        s[1,2,0,0]=(-0.562355685272)*x[0]+(0.495744936869)*x[1]
        s[1,2,0,1]=(0.0747584188968)*x[0]+(1.94924356864)*x[1]
        s[1,2,1,0]=(-0.103515803094)*x[0]+(0.545380886507)*x[1]
        s[1,2,1,1]=(-0.708926705928)*x[0]+(-0.0610123075525)*x[1]
        s[1,2,2,0]=(1.06802154614)*x[0]+(0.0622475519235)*x[1]
        s[1,2,2,1]=(0.103937010427)*x[0]+(-0.70201008693)*x[1]
        s[1,3,0,0]=(0.0730726824626)*x[0]+(0.39791188789)*x[1]
        s[1,3,0,1]=(0.455273194688)*x[0]+(1.65845657261)*x[1]
        s[1,3,1,0]=(0.0807767752424)*x[0]+(-0.765026395125)*x[1]
        s[1,3,1,1]=(0.0404890748976)*x[0]+(-0.383198358048)*x[1]
        s[1,3,2,0]=(1.26048751933)*x[0]+(0.00329382174075)*x[1]
        s[1,3,2,1]=(-1.21518062161)*x[0]+(-0.325769725861)*x[1]
        s[1,4,0,0]=(0.682181651323)*x[0]+(-0.375432356672)*x[1]
        s[1,4,0,1]=(-1.09892909715)*x[0]+(0.207984305416)*x[1]
        s[1,4,1,0]=(-1.2375595164)*x[0]+(0.215243019189)*x[1]
        s[1,4,1,1]=(-0.595125980205)*x[0]+(0.29104395616)*x[1]
        s[1,4,2,0]=(0.127947393316)*x[0]+(1.43408039596)*x[1]
        s[1,4,2,1]=(0.468574871016)*x[0]+(0.333319611281)*x[1]
        s[2,0,0,0]=(0.369954209324)*x[0]+(-1.09614147452)*x[1]
        s[2,0,0,1]=(-0.120003056441)*x[0]+(-0.172964298439)*x[1]
        s[2,0,1,0]=(1.38230317911)*x[0]+(1.25730140777)*x[1]
        s[2,0,1,1]=(1.10857329622)*x[0]+(0.449561992049)*x[1]
        s[2,0,2,0]=(-1.48479939533)*x[0]+(-0.839474409588)*x[1]
        s[2,0,2,1]=(-0.662177361426)*x[0]+(0.380170348062)*x[1]
        s[2,1,0,0]=(1.13771311046)*x[0]+(-1.59960740614)*x[1]
        s[2,1,0,1]=(0.925401031957)*x[0]+(1.30170041054)*x[1]
        s[2,1,1,0]=(0.455911230864)*x[0]+(0.755253828053)*x[1]
        s[2,1,1,1]=(0.64783771432)*x[0]+(0.251664567771)*x[1]
        s[2,1,2,0]=(-0.0449523798045)*x[0]+(-0.614423607357)*x[1]
        s[2,1,2,1]=(0.635435155497)*x[0]+(0.218411499589)*x[1]
        s[2,2,0,0]=(1.07064943699)*x[0]+(1.33695839858)*x[1]
        s[2,2,0,1]=(0.286067300461)*x[0]+(-1.19968009538)*x[1]
        s[2,2,1,0]=(0.373936537767)*x[0]+(-0.684346831863)*x[1]
        s[2,2,1,1]=(0.709764408449)*x[0]+(0.560181368149)*x[1]
        s[2,2,2,0]=(0.687217025484)*x[0]+(0.638368794102)*x[1]
        s[2,2,2,1]=(-1.13694076648)*x[0]+(1.31012182384)*x[1]
        s[2,3,0,0]=(0.784452969448)*x[0]+(-0.256804074025)*x[1]
        s[2,3,0,1]=(-0.528533290176)*x[0]+(0.230293895766)*x[1]
        s[2,3,1,0]=(-0.152281766055)*x[0]+(-1.29007246581)*x[1]
        s[2,3,1,1]=(-0.113853891603)*x[0]+(-0.146975472116)*x[1]
        s[2,3,2,0]=(-1.16737851429)*x[0]+(1.18358630738)*x[1]
        s[2,3,2,1]=(-0.110242592809)*x[0]+(-0.228040285919)*x[1]
        s[2,4,0,0]=(-0.615024817893)*x[0]+(1.19310484943)*x[1]
        s[2,4,0,1]=(0.405843983847)*x[0]+(-0.7683655961)*x[1]
        s[2,4,1,0]=(-1.94063527375)*x[0]+(0.135437762233)*x[1]
        s[2,4,1,1]=(-0.680568187039)*x[0]+(0.411161107413)*x[1]
        s[2,4,2,0]=(-1.95575235842)*x[0]+(0.384766876765)*x[1]
        s[2,4,2,1]=(1.37236985627)*x[0]+(0.0082159494728)*x[1]
        s[3,0,0,0]=(1.63255667845)*x[0]+(0.0225121724188)*x[1]
        s[3,0,0,1]=(1.2376463155)*x[0]+(1.32594028087)*x[1]
        s[3,0,1,0]=(1.54455876996)*x[0]+(-1.21021289438)*x[1]
        s[3,0,1,1]=(-1.81162592117)*x[0]+(-0.297403340122)*x[1]
        s[3,0,2,0]=(0.806731759445)*x[0]+(-0.981072370255)*x[1]
        s[3,0,2,1]=(-0.15215786955)*x[0]+(0.163165607014)*x[1]
        s[3,1,0,0]=(0.233006578932)*x[0]+(-1.29832996415)*x[1]
        s[3,1,0,1]=(0.96755903915)*x[0]+(-0.00597412239186)*x[1]
        s[3,1,1,0]=(0.0413376466172)*x[0]+(1.26720441986)*x[1]
        s[3,1,1,1]=(0.676509459424)*x[0]+(1.2621989434)*x[1]
        s[3,1,2,0]=(-0.211034948146)*x[0]+(-1.37470768893)*x[1]
        s[3,1,2,1]=(-0.402208754333)*x[0]+(1.25258243878)*x[1]
        s[3,2,0,0]=(-0.529199129991)*x[0]+(-0.431490132335)*x[1]
        s[3,2,0,1]=(-0.850343939948)*x[0]+(0.0602529984167)*x[1]
        s[3,2,1,0]=(-0.420125360943)*x[0]+(1.24318875197)*x[1]
        s[3,2,1,1]=(0.753685328318)*x[0]+(0.354913692774)*x[1]
        s[3,2,2,0]=(-1.17147381854)*x[0]+(0.850516886779)*x[1]
        s[3,2,2,1]=(1.48587305579)*x[0]+(1.31835868193)*x[1]
        s[3,3,0,0]=(1.80808607564)*x[0]+(-0.124508319578)*x[1]
        s[3,3,0,1]=(-1.84429695094)*x[0]+(-0.254797649176)*x[1]
        s[3,3,1,0]=(1.17110770558)*x[0]+(0.89756258226)*x[1]
        s[3,3,1,1]=(0.692264407822)*x[0]+(-1.00105988426)*x[1]
        s[3,3,2,0]=(-0.505689548372)*x[0]+(0.365519850058)*x[1]
        s[3,3,2,1]=(0.651544272536)*x[0]+(1.09868410461)*x[1]
        s[3,4,0,0]=(0.155129169877)*x[0]+(0.4435929499)*x[1]
        s[3,4,0,1]=(1.65974703109)*x[0]+(-0.776830635917)*x[1]
        s[3,4,1,0]=(-1.05612271738)*x[0]+(1.63557694551)*x[1]
        s[3,4,1,1]=(0.69435492251)*x[0]+(-1.20394873527)*x[1]
        s[3,4,2,0]=(-0.237598987964)*x[0]+(0.404820539664)*x[1]
        s[3,4,2,1]=(-1.32483403281)*x[0]+(0.431217340618)*x[1]
        ref[0,0,0,0]=(-0.9809347951)*x_ref[0]+(-1.05666590015)*x_ref[1]
        ref[0,0,0,1]=(1.15925248094)*x_ref[0]+(1.22070923582)*x_ref[1]
        ref[0,0,1,0]=(-0.0724361628927)*x_ref[0]+(-0.439807238916)*x_ref[1]
        ref[0,0,1,1]=(-0.608816713414)*x_ref[0]+(1.78792604745)*x_ref[1]
        ref[0,0,2,0]=(-0.78903506363)*x_ref[0]+(-0.75896072902)*x_ref[1]
        ref[0,0,2,1]=(0.516424260522)*x_ref[0]+(1.16788015378)*x_ref[1]
        ref[0,1,0,0]=(0.383586654748)*x_ref[0]+(1.4179924146)*x_ref[1]
        ref[0,1,0,1]=(0.97236907624)*x_ref[0]+(-0.438982550276)*x_ref[1]
        ref[0,1,1,0]=(-0.645083297783)*x_ref[0]+(-0.740808997218)*x_ref[1]
        ref[0,1,1,1]=(0.330652781935)*x_ref[0]+(0.519132367114)*x_ref[1]
        ref[0,1,2,0]=(1.25159174261)*x_ref[0]+(1.26021904904)*x_ref[1]
        ref[0,1,2,1]=(-0.239621293462)*x_ref[0]+(0.0244107517523)*x_ref[1]
        ref[0,2,0,0]=(-0.286208240697)*x_ref[0]+(-0.314157258671)*x_ref[1]
        ref[0,2,0,1]=(-0.0856319829397)*x_ref[0]+(0.278603809503)*x_ref[1]
        ref[0,2,1,0]=(0.107199217139)*x_ref[0]+(0.0948664401127)*x_ref[1]
        ref[0,2,1,1]=(-0.24620060163)*x_ref[0]+(-1.42942082558)*x_ref[1]
        ref[0,2,2,0]=(0.845285367352)*x_ref[0]+(0.408676159609)*x_ref[1]
        ref[0,2,2,1]=(-1.49467460394)*x_ref[0]+(0.270722457265)*x_ref[1]
        ref[0,3,0,0]=(0.469802598061)*x_ref[0]+(-0.801790345348)*x_ref[1]
        ref[0,3,0,1]=(1.71249257903)*x_ref[0]+(-0.776797378881)*x_ref[1]
        ref[0,3,1,0]=(0.0540049036319)*x_ref[0]+(1.19031969164)*x_ref[1]
        ref[0,3,1,1]=(0.161929130968)*x_ref[0]+(-0.525381566632)*x_ref[1]
        ref[0,3,2,0]=(-0.646041965665)*x_ref[0]+(-0.123993555032)*x_ref[1]
        ref[0,3,2,1]=(0.0966044205198)*x_ref[0]+(1.24780552685)*x_ref[1]
        ref[0,4,0,0]=(-0.289531041539)*x_ref[0]+(1.21117470928)*x_ref[1]
        ref[0,4,0,1]=(-0.112215925799)*x_ref[0]+(0.598717725332)*x_ref[1]
        ref[0,4,1,0]=(0.162936056701)*x_ref[0]+(-1.02395851472)*x_ref[1]
        ref[0,4,1,1]=(0.694066364547)*x_ref[0]+(1.27205020727)*x_ref[1]
        ref[0,4,2,0]=(1.17253981195)*x_ref[0]+(0.130572791304)*x_ref[1]
        ref[0,4,2,1]=(-0.555241229093)*x_ref[0]+(-0.881115061403)*x_ref[1]
        ref[1,0,0,0]=(0.574137047465)*x_ref[0]+(-1.10459006911)*x_ref[1]
        ref[1,0,0,1]=(0.441431076852)*x_ref[0]+(-0.133156832361)*x_ref[1]
        ref[1,0,1,0]=(0.217136655494)*x_ref[0]+(0.400834952015)*x_ref[1]
        ref[1,0,1,1]=(0.516360280675)*x_ref[0]+(0.532907430046)*x_ref[1]
        ref[1,0,2,0]=(-0.340764752946)*x_ref[0]+(-1.26671004418)*x_ref[1]
        ref[1,0,2,1]=(-1.23231139783)*x_ref[0]+(0.61934930883)*x_ref[1]
        ref[1,1,0,0]=(1.76821949014)*x_ref[0]+(1.67542625016)*x_ref[1]
        ref[1,1,0,1]=(0.901675970272)*x_ref[0]+(-0.0567109697208)*x_ref[1]
        ref[1,1,1,0]=(1.21143551644)*x_ref[0]+(0.375877605521)*x_ref[1]
        ref[1,1,1,1]=(-0.699830133987)*x_ref[0]+(0.174660968691)*x_ref[1]
        ref[1,1,2,0]=(1.07839763024)*x_ref[0]+(-0.063554858486)*x_ref[1]
        ref[1,1,2,1]=(0.131335396282)*x_ref[0]+(0.109893568063)*x_ref[1]
        ref[1,2,0,0]=(-0.562355685272)*x_ref[0]+(0.495744936869)*x_ref[1]
        ref[1,2,0,1]=(0.0747584188968)*x_ref[0]+(1.94924356864)*x_ref[1]
        ref[1,2,1,0]=(-0.103515803094)*x_ref[0]+(0.545380886507)*x_ref[1]
        ref[1,2,1,1]=(-0.708926705928)*x_ref[0]+(-0.0610123075525)*x_ref[1]
        ref[1,2,2,0]=(1.06802154614)*x_ref[0]+(0.0622475519235)*x_ref[1]
        ref[1,2,2,1]=(0.103937010427)*x_ref[0]+(-0.70201008693)*x_ref[1]
        ref[1,3,0,0]=(0.0730726824626)*x_ref[0]+(0.39791188789)*x_ref[1]
        ref[1,3,0,1]=(0.455273194688)*x_ref[0]+(1.65845657261)*x_ref[1]
        ref[1,3,1,0]=(0.0807767752424)*x_ref[0]+(-0.765026395125)*x_ref[1]
        ref[1,3,1,1]=(0.0404890748976)*x_ref[0]+(-0.383198358048)*x_ref[1]
        ref[1,3,2,0]=(1.26048751933)*x_ref[0]+(0.00329382174075)*x_ref[1]
        ref[1,3,2,1]=(-1.21518062161)*x_ref[0]+(-0.325769725861)*x_ref[1]
        ref[1,4,0,0]=(0.682181651323)*x_ref[0]+(-0.375432356672)*x_ref[1]
        ref[1,4,0,1]=(-1.09892909715)*x_ref[0]+(0.207984305416)*x_ref[1]
        ref[1,4,1,0]=(-1.2375595164)*x_ref[0]+(0.215243019189)*x_ref[1]
        ref[1,4,1,1]=(-0.595125980205)*x_ref[0]+(0.29104395616)*x_ref[1]
        ref[1,4,2,0]=(0.127947393316)*x_ref[0]+(1.43408039596)*x_ref[1]
        ref[1,4,2,1]=(0.468574871016)*x_ref[0]+(0.333319611281)*x_ref[1]
        ref[2,0,0,0]=(0.369954209324)*x_ref[0]+(-1.09614147452)*x_ref[1]
        ref[2,0,0,1]=(-0.120003056441)*x_ref[0]+(-0.172964298439)*x_ref[1]
        ref[2,0,1,0]=(1.38230317911)*x_ref[0]+(1.25730140777)*x_ref[1]
        ref[2,0,1,1]=(1.10857329622)*x_ref[0]+(0.449561992049)*x_ref[1]
        ref[2,0,2,0]=(-1.48479939533)*x_ref[0]+(-0.839474409588)*x_ref[1]
        ref[2,0,2,1]=(-0.662177361426)*x_ref[0]+(0.380170348062)*x_ref[1]
        ref[2,1,0,0]=(1.13771311046)*x_ref[0]+(-1.59960740614)*x_ref[1]
        ref[2,1,0,1]=(0.925401031957)*x_ref[0]+(1.30170041054)*x_ref[1]
        ref[2,1,1,0]=(0.455911230864)*x_ref[0]+(0.755253828053)*x_ref[1]
        ref[2,1,1,1]=(0.64783771432)*x_ref[0]+(0.251664567771)*x_ref[1]
        ref[2,1,2,0]=(-0.0449523798045)*x_ref[0]+(-0.614423607357)*x_ref[1]
        ref[2,1,2,1]=(0.635435155497)*x_ref[0]+(0.218411499589)*x_ref[1]
        ref[2,2,0,0]=(1.07064943699)*x_ref[0]+(1.33695839858)*x_ref[1]
        ref[2,2,0,1]=(0.286067300461)*x_ref[0]+(-1.19968009538)*x_ref[1]
        ref[2,2,1,0]=(0.373936537767)*x_ref[0]+(-0.684346831863)*x_ref[1]
        ref[2,2,1,1]=(0.709764408449)*x_ref[0]+(0.560181368149)*x_ref[1]
        ref[2,2,2,0]=(0.687217025484)*x_ref[0]+(0.638368794102)*x_ref[1]
        ref[2,2,2,1]=(-1.13694076648)*x_ref[0]+(1.31012182384)*x_ref[1]
        ref[2,3,0,0]=(0.784452969448)*x_ref[0]+(-0.256804074025)*x_ref[1]
        ref[2,3,0,1]=(-0.528533290176)*x_ref[0]+(0.230293895766)*x_ref[1]
        ref[2,3,1,0]=(-0.152281766055)*x_ref[0]+(-1.29007246581)*x_ref[1]
        ref[2,3,1,1]=(-0.113853891603)*x_ref[0]+(-0.146975472116)*x_ref[1]
        ref[2,3,2,0]=(-1.16737851429)*x_ref[0]+(1.18358630738)*x_ref[1]
        ref[2,3,2,1]=(-0.110242592809)*x_ref[0]+(-0.228040285919)*x_ref[1]
        ref[2,4,0,0]=(-0.615024817893)*x_ref[0]+(1.19310484943)*x_ref[1]
        ref[2,4,0,1]=(0.405843983847)*x_ref[0]+(-0.7683655961)*x_ref[1]
        ref[2,4,1,0]=(-1.94063527375)*x_ref[0]+(0.135437762233)*x_ref[1]
        ref[2,4,1,1]=(-0.680568187039)*x_ref[0]+(0.411161107413)*x_ref[1]
        ref[2,4,2,0]=(-1.95575235842)*x_ref[0]+(0.384766876765)*x_ref[1]
        ref[2,4,2,1]=(1.37236985627)*x_ref[0]+(0.0082159494728)*x_ref[1]
        ref[3,0,0,0]=(1.63255667845)*x_ref[0]+(0.0225121724188)*x_ref[1]
        ref[3,0,0,1]=(1.2376463155)*x_ref[0]+(1.32594028087)*x_ref[1]
        ref[3,0,1,0]=(1.54455876996)*x_ref[0]+(-1.21021289438)*x_ref[1]
        ref[3,0,1,1]=(-1.81162592117)*x_ref[0]+(-0.297403340122)*x_ref[1]
        ref[3,0,2,0]=(0.806731759445)*x_ref[0]+(-0.981072370255)*x_ref[1]
        ref[3,0,2,1]=(-0.15215786955)*x_ref[0]+(0.163165607014)*x_ref[1]
        ref[3,1,0,0]=(0.233006578932)*x_ref[0]+(-1.29832996415)*x_ref[1]
        ref[3,1,0,1]=(0.96755903915)*x_ref[0]+(-0.00597412239186)*x_ref[1]
        ref[3,1,1,0]=(0.0413376466172)*x_ref[0]+(1.26720441986)*x_ref[1]
        ref[3,1,1,1]=(0.676509459424)*x_ref[0]+(1.2621989434)*x_ref[1]
        ref[3,1,2,0]=(-0.211034948146)*x_ref[0]+(-1.37470768893)*x_ref[1]
        ref[3,1,2,1]=(-0.402208754333)*x_ref[0]+(1.25258243878)*x_ref[1]
        ref[3,2,0,0]=(-0.529199129991)*x_ref[0]+(-0.431490132335)*x_ref[1]
        ref[3,2,0,1]=(-0.850343939948)*x_ref[0]+(0.0602529984167)*x_ref[1]
        ref[3,2,1,0]=(-0.420125360943)*x_ref[0]+(1.24318875197)*x_ref[1]
        ref[3,2,1,1]=(0.753685328318)*x_ref[0]+(0.354913692774)*x_ref[1]
        ref[3,2,2,0]=(-1.17147381854)*x_ref[0]+(0.850516886779)*x_ref[1]
        ref[3,2,2,1]=(1.48587305579)*x_ref[0]+(1.31835868193)*x_ref[1]
        ref[3,3,0,0]=(1.80808607564)*x_ref[0]+(-0.124508319578)*x_ref[1]
        ref[3,3,0,1]=(-1.84429695094)*x_ref[0]+(-0.254797649176)*x_ref[1]
        ref[3,3,1,0]=(1.17110770558)*x_ref[0]+(0.89756258226)*x_ref[1]
        ref[3,3,1,1]=(0.692264407822)*x_ref[0]+(-1.00105988426)*x_ref[1]
        ref[3,3,2,0]=(-0.505689548372)*x_ref[0]+(0.365519850058)*x_ref[1]
        ref[3,3,2,1]=(0.651544272536)*x_ref[0]+(1.09868410461)*x_ref[1]
        ref[3,4,0,0]=(0.155129169877)*x_ref[0]+(0.4435929499)*x_ref[1]
        ref[3,4,0,1]=(1.65974703109)*x_ref[0]+(-0.776830635917)*x_ref[1]
        ref[3,4,1,0]=(-1.05612271738)*x_ref[0]+(1.63557694551)*x_ref[1]
        ref[3,4,1,1]=(0.69435492251)*x_ref[0]+(-1.20394873527)*x_ref[1]
        ref[3,4,2,0]=(-0.237598987964)*x_ref[0]+(0.404820539664)*x_ref[1]
        ref[3,4,2,1]=(-1.32483403281)*x_ref[0]+(0.431217340618)*x_ref[1]
      else:
        s[0,0,0,0]=(-0.270116752643)*x[0]+(0.0715809194071)*x[1]+(0.930011298681)*x[2]
        s[0,0,0,1]=(-0.0435807732016)*x[0]+(-0.492622221729)*x[1]+(-1.36754174971)*x[2]
        s[0,0,1,0]=(-0.949818131569)*x[0]+(1.08437418341)*x[1]+(0.527267447353)*x[2]
        s[0,0,1,1]=(-0.164876627537)*x[0]+(-1.44143896699)*x[1]+(0.0355140164081)*x[2]
        s[0,0,2,0]=(-0.530229192916)*x[0]+(-0.988412045132)*x[1]+(-0.466520393878)*x[2]
        s[0,0,2,1]=(-0.0166373496749)*x[0]+(0.322621663585)*x[1]+(-0.102019336059)*x[2]
        s[0,1,0,0]=(-1.36627178258)*x[0]+(-0.803463758822)*x[1]+(0.734446026661)*x[2]
        s[0,1,0,1]=(-0.804729478825)*x[0]+(-1.24406753288)*x[1]+(-0.755743791016)*x[2]
        s[0,1,1,0]=(1.03869027308)*x[0]+(-0.25237293575)*x[1]+(1.43737134487)*x[2]
        s[0,1,1,1]=(1.48269192251)*x[0]+(0.045590449115)*x[1]+(0.47425103841)*x[2]
        s[0,1,2,0]=(-0.738695287005)*x[0]+(1.01938431903)*x[1]+(1.20767534795)*x[2]
        s[0,1,2,1]=(-0.237445941287)*x[0]+(0.766593746786)*x[1]+(0.0297279831889)*x[2]
        s[0,2,0,0]=(-0.804902383375)*x[0]+(0.0962339257662)*x[1]+(1.37407473707)*x[2]
        s[0,2,0,1]=(-0.0821182376238)*x[0]+(0.35489327051)*x[1]+(-1.43235115632)*x[2]
        s[0,2,1,0]=(1.11593707013)*x[0]+(0.744039820943)*x[1]+(0.294815281064)*x[2]
        s[0,2,1,1]=(0.0789595478782)*x[0]+(0.719656781874)*x[1]+(0.355275528053)*x[2]
        s[0,2,2,0]=(1.20514986435)*x[0]+(0.203004753589)*x[1]+(-0.0731941650839)*x[2]
        s[0,2,2,1]=(-1.23231910051)*x[0]+(-0.470834142341)*x[1]+(-0.633518095593)*x[2]
        s[0,3,0,0]=(0.468164093965)*x[0]+(-1.15996958069)*x[1]+(0.922274695212)*x[2]
        s[0,3,0,1]=(-0.964250141036)*x[0]+(-0.769747467017)*x[1]+(-1.16369265956)*x[2]
        s[0,3,1,0]=(1.1622810944)*x[0]+(-1.72794997517)*x[1]+(-0.297833274612)*x[2]
        s[0,3,1,1]=(1.20239531447)*x[0]+(-0.195111187444)*x[1]+(-0.720186501147)*x[2]
        s[0,3,2,0]=(-0.0760001358746)*x[0]+(-1.65266437471)*x[1]+(-0.496163951015)*x[2]
        s[0,3,2,1]=(-1.4998248458)*x[0]+(-1.73305168566)*x[1]+(-0.107384086634)*x[2]
        s[0,4,0,0]=(0.235699357585)*x[0]+(-0.213348404403)*x[1]+(0.35368491512)*x[2]
        s[0,4,0,1]=(0.264641788778)*x[0]+(0.435251584452)*x[1]+(0.22539974851)*x[2]
        s[0,4,1,0]=(0.540203510472)*x[0]+(-1.33243180984)*x[1]+(-0.520357911324)*x[2]
        s[0,4,1,1]=(-0.629868492098)*x[0]+(-0.403402885662)*x[1]+(-0.184983842426)*x[2]
        s[0,4,2,0]=(1.17843494613)*x[0]+(0.562783141759)*x[1]+(-1.48575741726)*x[2]
        s[0,4,2,1]=(0.623796209289)*x[0]+(1.02627846638)*x[1]+(1.39369241754)*x[2]
        s[1,0,0,0]=(1.9855507901)*x[0]+(0.152976407288)*x[1]+(0.0931547545156)*x[2]
        s[1,0,0,1]=(-0.422055088151)*x[0]+(-0.509449435814)*x[1]+(-1.31705141661)*x[2]
        s[1,0,1,0]=(-0.24193816853)*x[0]+(-0.68450536388)*x[1]+(-0.309024887085)*x[2]
        s[1,0,1,1]=(1.24557169323)*x[0]+(0.353538565492)*x[1]+(0.115821365646)*x[2]
        s[1,0,2,0]=(-0.280246915004)*x[0]+(-0.0972556765171)*x[1]+(0.758816560386)*x[2]
        s[1,0,2,1]=(1.0396259589)*x[0]+(-0.239507470189)*x[1]+(-0.987815676492)*x[2]
        s[1,1,0,0]=(0.245203366317)*x[0]+(-1.53269800659)*x[1]+(0.619360300925)*x[2]
        s[1,1,0,1]=(0.246406467482)*x[0]+(-0.729988292889)*x[1]+(0.585422287189)*x[2]
        s[1,1,1,0]=(1.19810038121)*x[0]+(-0.72772445701)*x[1]+(1.11619392954)*x[2]
        s[1,1,1,1]=(-0.311984976134)*x[0]+(-0.188443023344)*x[1]+(-0.100078290654)*x[2]
        s[1,1,2,0]=(0.158578485066)*x[0]+(0.214804573675)*x[1]+(-0.22418434724)*x[2]
        s[1,1,2,1]=(-0.0550611413075)*x[0]+(-0.147088267891)*x[1]+(-0.269933900128)*x[2]
        s[1,2,0,0]=(-0.8618782319)*x[0]+(-0.770284701216)*x[1]+(-0.407458817279)*x[2]
        s[1,2,0,1]=(-0.242336600721)*x[0]+(-1.57823911131)*x[1]+(0.452055841078)*x[2]
        s[1,2,1,0]=(0.0537219652862)*x[0]+(0.18864518829)*x[1]+(0.021034116382)*x[2]
        s[1,2,1,1]=(0.143799709809)*x[0]+(0.787934959856)*x[1]+(-0.283457275339)*x[2]
        s[1,2,2,0]=(0.475965504456)*x[0]+(-0.744950301051)*x[1]+(0.45308618026)*x[2]
        s[1,2,2,1]=(0.561593580643)*x[0]+(0.121626528038)*x[1]+(-1.24221676348)*x[2]
        s[1,3,0,0]=(-0.340056043985)*x[0]+(-0.434795551938)*x[1]+(-0.291179084342)*x[2]
        s[1,3,0,1]=(0.312318149724)*x[0]+(1.66950279503)*x[1]+(-1.1622076779)*x[2]
        s[1,3,1,0]=(0.0965172054443)*x[0]+(-0.64102022474)*x[1]+(-0.173612340622)*x[2]
        s[1,3,1,1]=(1.88701305968)*x[0]+(-1.66794136241)*x[1]+(1.56059966446)*x[2]
        s[1,3,2,0]=(-0.497773999274)*x[0]+(0.395345962943)*x[1]+(-0.499105933847)*x[2]
        s[1,3,2,1]=(-0.312096437005)*x[0]+(-1.23301763457)*x[1]+(0.198835942067)*x[2]
        s[1,4,0,0]=(0.076377295342)*x[0]+(-0.487847012612)*x[1]+(-0.084314987823)*x[2]
        s[1,4,0,1]=(0.666434654203)*x[0]+(-0.220376188514)*x[1]+(-0.355749198488)*x[2]
        s[1,4,1,0]=(-0.701104468581)*x[0]+(0.828174884422)*x[1]+(0.309374641659)*x[2]
        s[1,4,1,1]=(0.526340204296)*x[0]+(0.571598927007)*x[1]+(-0.0793826750713)*x[2]
        s[1,4,2,0]=(0.775563727486)*x[0]+(0.248228533771)*x[1]+(-1.18169855975)*x[2]
        s[1,4,2,1]=(-0.399357449979)*x[0]+(-0.723554468137)*x[1]+(0.0231913550517)*x[2]
        s[2,0,0,0]=(-0.0504735749489)*x[0]+(0.506564345355)*x[1]+(0.790336060204)*x[2]
        s[2,0,0,1]=(0.426187262022)*x[0]+(0.966680952632)*x[1]+(0.410542727369)*x[2]
        s[2,0,1,0]=(0.594102954821)*x[0]+(0.0639755896611)*x[1]+(-0.675769510238)*x[2]
        s[2,0,1,1]=(0.367479711308)*x[0]+(0.150736830273)*x[1]+(1.15797942999)*x[2]
        s[2,0,2,0]=(0.875207101066)*x[0]+(-1.12939492894)*x[1]+(0.73561332295)*x[2]
        s[2,0,2,1]=(-1.70885193598)*x[0]+(0.835257764029)*x[1]+(-1.58521909429)*x[2]
        s[2,1,0,0]=(-0.144255046397)*x[0]+(0.590074178586)*x[1]+(0.592144641271)*x[2]
        s[2,1,0,1]=(-0.0688470601204)*x[0]+(-0.153340975531)*x[1]+(1.15241895553)*x[2]
        s[2,1,1,0]=(-0.73257033584)*x[0]+(-0.563316507818)*x[1]+(-1.05620654545)*x[2]
        s[2,1,1,1]=(-1.29833350092)*x[0]+(0.42237934165)*x[1]+(0.0725296431412)*x[2]
        s[2,1,2,0]=(-0.837473685756)*x[0]+(-0.425072509905)*x[1]+(-0.463397001131)*x[2]
        s[2,1,2,1]=(-0.937038584847)*x[0]+(0.531989188194)*x[1]+(-1.06694316698)*x[2]
        s[2,2,0,0]=(-0.846346120269)*x[0]+(-0.639618775615)*x[1]+(-0.928154018676)*x[2]
        s[2,2,0,1]=(-0.726631889289)*x[0]+(-0.1318624965)*x[1]+(-0.343729797494)*x[2]
        s[2,2,1,0]=(0.802665744539)*x[0]+(-0.115632355474)*x[1]+(0.303919795863)*x[2]
        s[2,2,1,1]=(0.0371825373918)*x[0]+(-1.17799388388)*x[1]+(-0.223043449617)*x[2]
        s[2,2,2,0]=(0.75988617552)*x[0]+(-0.608977501492)*x[1]+(1.75211913311)*x[2]
        s[2,2,2,1]=(-0.291397991509)*x[0]+(-1.55173354063)*x[1]+(-0.461269447756)*x[2]
        s[2,3,0,0]=(0.351624965509)*x[0]+(-1.32799214259)*x[1]+(1.58946725409)*x[2]
        s[2,3,0,1]=(0.180149992744)*x[0]+(1.07861987473)*x[1]+(0.834357928788)*x[2]
        s[2,3,1,0]=(0.0475016328103)*x[0]+(0.707449786509)*x[1]+(-0.370421903676)*x[2]
        s[2,3,1,1]=(0.553988297947)*x[0]+(1.13388241001)*x[1]+(0.575604300083)*x[2]
        s[2,3,2,0]=(-0.249047035016)*x[0]+(0.15013128589)*x[1]+(1.3953926353)*x[2]
        s[2,3,2,1]=(-0.783008188949)*x[0]+(0.861992436037)*x[1]+(-0.929496477422)*x[2]
        s[2,4,0,0]=(0.264348996811)*x[0]+(-1.23207394036)*x[1]+(-1.37804290092)*x[2]
        s[2,4,0,1]=(0.360782273055)*x[0]+(-0.270520439821)*x[1]+(-1.01285201623)*x[2]
        s[2,4,1,0]=(0.280578579439)*x[0]+(0.584696145403)*x[1]+(-1.21852764117)*x[2]
        s[2,4,1,1]=(0.112102121128)*x[0]+(-0.0619663521973)*x[1]+(1.08490293257)*x[2]
        s[2,4,2,0]=(-1.89749569924)*x[0]+(-0.21217889191)*x[1]+(0.635524779965)*x[2]
        s[2,4,2,1]=(-0.309239337357)*x[0]+(0.0116284822724)*x[1]+(0.0107247712951)*x[2]
        s[3,0,0,0]=(1.23545873872)*x[0]+(-1.29896744411)*x[1]+(0.855094335492)*x[2]
        s[3,0,0,1]=(-0.51465695324)*x[0]+(1.47799693308)*x[1]+(0.0718360423316)*x[2]
        s[3,0,1,0]=(0.220505442402)*x[0]+(-0.3171385194)*x[1]+(0.27954628665)*x[2]
        s[3,0,1,1]=(0.860582046972)*x[0]+(1.04565549687)*x[1]+(1.18108647078)*x[2]
        s[3,0,2,0]=(-0.593547630941)*x[0]+(-0.0390465590147)*x[1]+(0.00129707682534)*x[2]
        s[3,0,2,1]=(1.39210293617)*x[0]+(-1.05660957806)*x[1]+(-1.20489109051)*x[2]
        s[3,1,0,0]=(-0.536035988946)*x[0]+(-0.594403345465)*x[1]+(1.11719111746)*x[2]
        s[3,1,0,1]=(-0.0403557954127)*x[0]+(-1.32391025453)*x[1]+(-0.639127329416)*x[2]
        s[3,1,1,0]=(-0.495803710666)*x[0]+(-0.199832989186)*x[1]+(-0.832862076561)*x[2]
        s[3,1,1,1]=(0.0216882402803)*x[0]+(0.367268873436)*x[1]+(0.295944335418)*x[2]
        s[3,1,2,0]=(0.172936506269)*x[0]+(-0.339886198196)*x[1]+(0.268201481244)*x[2]
        s[3,1,2,1]=(0.665786569897)*x[0]+(-0.290637226736)*x[1]+(0.617677178537)*x[2]
        s[3,2,0,0]=(-0.657902185911)*x[0]+(-1.34292379802)*x[1]+(0.0641950819134)*x[2]
        s[3,2,0,1]=(0.352039175923)*x[0]+(-0.483047873735)*x[1]+(0.47997600743)*x[2]
        s[3,2,1,0]=(0.944619484106)*x[0]+(-0.308755547249)*x[1]+(-0.765906593989)*x[2]
        s[3,2,1,1]=(-1.09674781865)*x[0]+(0.013069257236)*x[1]+(-0.0741247275628)*x[2]
        s[3,2,2,0]=(0.400057737624)*x[0]+(-1.16165281514)*x[1]+(-0.789757454741)*x[2]
        s[3,2,2,1]=(0.367100312493)*x[0]+(-0.547338908792)*x[1]+(-0.409173859392)*x[2]
        s[3,3,0,0]=(0.890118369638)*x[0]+(1.201296587)*x[1]+(-0.0319246476545)*x[2]
        s[3,3,0,1]=(0.620955702501)*x[0]+(0.438440818114)*x[1]+(0.499747510411)*x[2]
        s[3,3,1,0]=(1.61217537222)*x[0]+(-1.21323714148)*x[1]+(-1.06137097571)*x[2]
        s[3,3,1,1]=(-1.75942413542)*x[0]+(0.440345678738)*x[1]+(0.625299774096)*x[2]
        s[3,3,2,0]=(-0.878497266865)*x[0]+(-1.15879244679)*x[1]+(1.03993124425)*x[2]
        s[3,3,2,1]=(-0.0138445968491)*x[0]+(0.547910692711)*x[1]+(0.971642264613)*x[2]
        s[3,4,0,0]=(-0.107769272336)*x[0]+(-0.987911384393)*x[1]+(1.2076201556)*x[2]
        s[3,4,0,1]=(-0.556570952071)*x[0]+(-0.284568452888)*x[1]+(1.23341917108)*x[2]
        s[3,4,1,0]=(1.56461193947)*x[0]+(0.185382898933)*x[1]+(-1.67714180845)*x[2]
        s[3,4,1,1]=(-0.894934034676)*x[0]+(1.05560233139)*x[1]+(0.651029150593)*x[2]
        s[3,4,2,0]=(0.0559603779066)*x[0]+(-1.72752549101)*x[1]+(-1.02017554749)*x[2]
        s[3,4,2,1]=(-1.75424333815)*x[0]+(-0.198869840282)*x[1]+(0.475065396299)*x[2]
        ref[0,0,0,0]=(-0.270116752643)*x_ref[0]+(0.0715809194071)*x_ref[1]+(0.930011298681)*x_ref[2]
        ref[0,0,0,1]=(-0.0435807732016)*x_ref[0]+(-0.492622221729)*x_ref[1]+(-1.36754174971)*x_ref[2]
        ref[0,0,1,0]=(-0.949818131569)*x_ref[0]+(1.08437418341)*x_ref[1]+(0.527267447353)*x_ref[2]
        ref[0,0,1,1]=(-0.164876627537)*x_ref[0]+(-1.44143896699)*x_ref[1]+(0.0355140164081)*x_ref[2]
        ref[0,0,2,0]=(-0.530229192916)*x_ref[0]+(-0.988412045132)*x_ref[1]+(-0.466520393878)*x_ref[2]
        ref[0,0,2,1]=(-0.0166373496749)*x_ref[0]+(0.322621663585)*x_ref[1]+(-0.102019336059)*x_ref[2]
        ref[0,1,0,0]=(-1.36627178258)*x_ref[0]+(-0.803463758822)*x_ref[1]+(0.734446026661)*x_ref[2]
        ref[0,1,0,1]=(-0.804729478825)*x_ref[0]+(-1.24406753288)*x_ref[1]+(-0.755743791016)*x_ref[2]
        ref[0,1,1,0]=(1.03869027308)*x_ref[0]+(-0.25237293575)*x_ref[1]+(1.43737134487)*x_ref[2]
        ref[0,1,1,1]=(1.48269192251)*x_ref[0]+(0.045590449115)*x_ref[1]+(0.47425103841)*x_ref[2]
        ref[0,1,2,0]=(-0.738695287005)*x_ref[0]+(1.01938431903)*x_ref[1]+(1.20767534795)*x_ref[2]
        ref[0,1,2,1]=(-0.237445941287)*x_ref[0]+(0.766593746786)*x_ref[1]+(0.0297279831889)*x_ref[2]
        ref[0,2,0,0]=(-0.804902383375)*x_ref[0]+(0.0962339257662)*x_ref[1]+(1.37407473707)*x_ref[2]
        ref[0,2,0,1]=(-0.0821182376238)*x_ref[0]+(0.35489327051)*x_ref[1]+(-1.43235115632)*x_ref[2]
        ref[0,2,1,0]=(1.11593707013)*x_ref[0]+(0.744039820943)*x_ref[1]+(0.294815281064)*x_ref[2]
        ref[0,2,1,1]=(0.0789595478782)*x_ref[0]+(0.719656781874)*x_ref[1]+(0.355275528053)*x_ref[2]
        ref[0,2,2,0]=(1.20514986435)*x_ref[0]+(0.203004753589)*x_ref[1]+(-0.0731941650839)*x_ref[2]
        ref[0,2,2,1]=(-1.23231910051)*x_ref[0]+(-0.470834142341)*x_ref[1]+(-0.633518095593)*x_ref[2]
        ref[0,3,0,0]=(0.468164093965)*x_ref[0]+(-1.15996958069)*x_ref[1]+(0.922274695212)*x_ref[2]
        ref[0,3,0,1]=(-0.964250141036)*x_ref[0]+(-0.769747467017)*x_ref[1]+(-1.16369265956)*x_ref[2]
        ref[0,3,1,0]=(1.1622810944)*x_ref[0]+(-1.72794997517)*x_ref[1]+(-0.297833274612)*x_ref[2]
        ref[0,3,1,1]=(1.20239531447)*x_ref[0]+(-0.195111187444)*x_ref[1]+(-0.720186501147)*x_ref[2]
        ref[0,3,2,0]=(-0.0760001358746)*x_ref[0]+(-1.65266437471)*x_ref[1]+(-0.496163951015)*x_ref[2]
        ref[0,3,2,1]=(-1.4998248458)*x_ref[0]+(-1.73305168566)*x_ref[1]+(-0.107384086634)*x_ref[2]
        ref[0,4,0,0]=(0.235699357585)*x_ref[0]+(-0.213348404403)*x_ref[1]+(0.35368491512)*x_ref[2]
        ref[0,4,0,1]=(0.264641788778)*x_ref[0]+(0.435251584452)*x_ref[1]+(0.22539974851)*x_ref[2]
        ref[0,4,1,0]=(0.540203510472)*x_ref[0]+(-1.33243180984)*x_ref[1]+(-0.520357911324)*x_ref[2]
        ref[0,4,1,1]=(-0.629868492098)*x_ref[0]+(-0.403402885662)*x_ref[1]+(-0.184983842426)*x_ref[2]
        ref[0,4,2,0]=(1.17843494613)*x_ref[0]+(0.562783141759)*x_ref[1]+(-1.48575741726)*x_ref[2]
        ref[0,4,2,1]=(0.623796209289)*x_ref[0]+(1.02627846638)*x_ref[1]+(1.39369241754)*x_ref[2]
        ref[1,0,0,0]=(1.9855507901)*x_ref[0]+(0.152976407288)*x_ref[1]+(0.0931547545156)*x_ref[2]
        ref[1,0,0,1]=(-0.422055088151)*x_ref[0]+(-0.509449435814)*x_ref[1]+(-1.31705141661)*x_ref[2]
        ref[1,0,1,0]=(-0.24193816853)*x_ref[0]+(-0.68450536388)*x_ref[1]+(-0.309024887085)*x_ref[2]
        ref[1,0,1,1]=(1.24557169323)*x_ref[0]+(0.353538565492)*x_ref[1]+(0.115821365646)*x_ref[2]
        ref[1,0,2,0]=(-0.280246915004)*x_ref[0]+(-0.0972556765171)*x_ref[1]+(0.758816560386)*x_ref[2]
        ref[1,0,2,1]=(1.0396259589)*x_ref[0]+(-0.239507470189)*x_ref[1]+(-0.987815676492)*x_ref[2]
        ref[1,1,0,0]=(0.245203366317)*x_ref[0]+(-1.53269800659)*x_ref[1]+(0.619360300925)*x_ref[2]
        ref[1,1,0,1]=(0.246406467482)*x_ref[0]+(-0.729988292889)*x_ref[1]+(0.585422287189)*x_ref[2]
        ref[1,1,1,0]=(1.19810038121)*x_ref[0]+(-0.72772445701)*x_ref[1]+(1.11619392954)*x_ref[2]
        ref[1,1,1,1]=(-0.311984976134)*x_ref[0]+(-0.188443023344)*x_ref[1]+(-0.100078290654)*x_ref[2]
        ref[1,1,2,0]=(0.158578485066)*x_ref[0]+(0.214804573675)*x_ref[1]+(-0.22418434724)*x_ref[2]
        ref[1,1,2,1]=(-0.0550611413075)*x_ref[0]+(-0.147088267891)*x_ref[1]+(-0.269933900128)*x_ref[2]
        ref[1,2,0,0]=(-0.8618782319)*x_ref[0]+(-0.770284701216)*x_ref[1]+(-0.407458817279)*x_ref[2]
        ref[1,2,0,1]=(-0.242336600721)*x_ref[0]+(-1.57823911131)*x_ref[1]+(0.452055841078)*x_ref[2]
        ref[1,2,1,0]=(0.0537219652862)*x_ref[0]+(0.18864518829)*x_ref[1]+(0.021034116382)*x_ref[2]
        ref[1,2,1,1]=(0.143799709809)*x_ref[0]+(0.787934959856)*x_ref[1]+(-0.283457275339)*x_ref[2]
        ref[1,2,2,0]=(0.475965504456)*x_ref[0]+(-0.744950301051)*x_ref[1]+(0.45308618026)*x_ref[2]
        ref[1,2,2,1]=(0.561593580643)*x_ref[0]+(0.121626528038)*x_ref[1]+(-1.24221676348)*x_ref[2]
        ref[1,3,0,0]=(-0.340056043985)*x_ref[0]+(-0.434795551938)*x_ref[1]+(-0.291179084342)*x_ref[2]
        ref[1,3,0,1]=(0.312318149724)*x_ref[0]+(1.66950279503)*x_ref[1]+(-1.1622076779)*x_ref[2]
        ref[1,3,1,0]=(0.0965172054443)*x_ref[0]+(-0.64102022474)*x_ref[1]+(-0.173612340622)*x_ref[2]
        ref[1,3,1,1]=(1.88701305968)*x_ref[0]+(-1.66794136241)*x_ref[1]+(1.56059966446)*x_ref[2]
        ref[1,3,2,0]=(-0.497773999274)*x_ref[0]+(0.395345962943)*x_ref[1]+(-0.499105933847)*x_ref[2]
        ref[1,3,2,1]=(-0.312096437005)*x_ref[0]+(-1.23301763457)*x_ref[1]+(0.198835942067)*x_ref[2]
        ref[1,4,0,0]=(0.076377295342)*x_ref[0]+(-0.487847012612)*x_ref[1]+(-0.084314987823)*x_ref[2]
        ref[1,4,0,1]=(0.666434654203)*x_ref[0]+(-0.220376188514)*x_ref[1]+(-0.355749198488)*x_ref[2]
        ref[1,4,1,0]=(-0.701104468581)*x_ref[0]+(0.828174884422)*x_ref[1]+(0.309374641659)*x_ref[2]
        ref[1,4,1,1]=(0.526340204296)*x_ref[0]+(0.571598927007)*x_ref[1]+(-0.0793826750713)*x_ref[2]
        ref[1,4,2,0]=(0.775563727486)*x_ref[0]+(0.248228533771)*x_ref[1]+(-1.18169855975)*x_ref[2]
        ref[1,4,2,1]=(-0.399357449979)*x_ref[0]+(-0.723554468137)*x_ref[1]+(0.0231913550517)*x_ref[2]
        ref[2,0,0,0]=(-0.0504735749489)*x_ref[0]+(0.506564345355)*x_ref[1]+(0.790336060204)*x_ref[2]
        ref[2,0,0,1]=(0.426187262022)*x_ref[0]+(0.966680952632)*x_ref[1]+(0.410542727369)*x_ref[2]
        ref[2,0,1,0]=(0.594102954821)*x_ref[0]+(0.0639755896611)*x_ref[1]+(-0.675769510238)*x_ref[2]
        ref[2,0,1,1]=(0.367479711308)*x_ref[0]+(0.150736830273)*x_ref[1]+(1.15797942999)*x_ref[2]
        ref[2,0,2,0]=(0.875207101066)*x_ref[0]+(-1.12939492894)*x_ref[1]+(0.73561332295)*x_ref[2]
        ref[2,0,2,1]=(-1.70885193598)*x_ref[0]+(0.835257764029)*x_ref[1]+(-1.58521909429)*x_ref[2]
        ref[2,1,0,0]=(-0.144255046397)*x_ref[0]+(0.590074178586)*x_ref[1]+(0.592144641271)*x_ref[2]
        ref[2,1,0,1]=(-0.0688470601204)*x_ref[0]+(-0.153340975531)*x_ref[1]+(1.15241895553)*x_ref[2]
        ref[2,1,1,0]=(-0.73257033584)*x_ref[0]+(-0.563316507818)*x_ref[1]+(-1.05620654545)*x_ref[2]
        ref[2,1,1,1]=(-1.29833350092)*x_ref[0]+(0.42237934165)*x_ref[1]+(0.0725296431412)*x_ref[2]
        ref[2,1,2,0]=(-0.837473685756)*x_ref[0]+(-0.425072509905)*x_ref[1]+(-0.463397001131)*x_ref[2]
        ref[2,1,2,1]=(-0.937038584847)*x_ref[0]+(0.531989188194)*x_ref[1]+(-1.06694316698)*x_ref[2]
        ref[2,2,0,0]=(-0.846346120269)*x_ref[0]+(-0.639618775615)*x_ref[1]+(-0.928154018676)*x_ref[2]
        ref[2,2,0,1]=(-0.726631889289)*x_ref[0]+(-0.1318624965)*x_ref[1]+(-0.343729797494)*x_ref[2]
        ref[2,2,1,0]=(0.802665744539)*x_ref[0]+(-0.115632355474)*x_ref[1]+(0.303919795863)*x_ref[2]
        ref[2,2,1,1]=(0.0371825373918)*x_ref[0]+(-1.17799388388)*x_ref[1]+(-0.223043449617)*x_ref[2]
        ref[2,2,2,0]=(0.75988617552)*x_ref[0]+(-0.608977501492)*x_ref[1]+(1.75211913311)*x_ref[2]
        ref[2,2,2,1]=(-0.291397991509)*x_ref[0]+(-1.55173354063)*x_ref[1]+(-0.461269447756)*x_ref[2]
        ref[2,3,0,0]=(0.351624965509)*x_ref[0]+(-1.32799214259)*x_ref[1]+(1.58946725409)*x_ref[2]
        ref[2,3,0,1]=(0.180149992744)*x_ref[0]+(1.07861987473)*x_ref[1]+(0.834357928788)*x_ref[2]
        ref[2,3,1,0]=(0.0475016328103)*x_ref[0]+(0.707449786509)*x_ref[1]+(-0.370421903676)*x_ref[2]
        ref[2,3,1,1]=(0.553988297947)*x_ref[0]+(1.13388241001)*x_ref[1]+(0.575604300083)*x_ref[2]
        ref[2,3,2,0]=(-0.249047035016)*x_ref[0]+(0.15013128589)*x_ref[1]+(1.3953926353)*x_ref[2]
        ref[2,3,2,1]=(-0.783008188949)*x_ref[0]+(0.861992436037)*x_ref[1]+(-0.929496477422)*x_ref[2]
        ref[2,4,0,0]=(0.264348996811)*x_ref[0]+(-1.23207394036)*x_ref[1]+(-1.37804290092)*x_ref[2]
        ref[2,4,0,1]=(0.360782273055)*x_ref[0]+(-0.270520439821)*x_ref[1]+(-1.01285201623)*x_ref[2]
        ref[2,4,1,0]=(0.280578579439)*x_ref[0]+(0.584696145403)*x_ref[1]+(-1.21852764117)*x_ref[2]
        ref[2,4,1,1]=(0.112102121128)*x_ref[0]+(-0.0619663521973)*x_ref[1]+(1.08490293257)*x_ref[2]
        ref[2,4,2,0]=(-1.89749569924)*x_ref[0]+(-0.21217889191)*x_ref[1]+(0.635524779965)*x_ref[2]
        ref[2,4,2,1]=(-0.309239337357)*x_ref[0]+(0.0116284822724)*x_ref[1]+(0.0107247712951)*x_ref[2]
        ref[3,0,0,0]=(1.23545873872)*x_ref[0]+(-1.29896744411)*x_ref[1]+(0.855094335492)*x_ref[2]
        ref[3,0,0,1]=(-0.51465695324)*x_ref[0]+(1.47799693308)*x_ref[1]+(0.0718360423316)*x_ref[2]
        ref[3,0,1,0]=(0.220505442402)*x_ref[0]+(-0.3171385194)*x_ref[1]+(0.27954628665)*x_ref[2]
        ref[3,0,1,1]=(0.860582046972)*x_ref[0]+(1.04565549687)*x_ref[1]+(1.18108647078)*x_ref[2]
        ref[3,0,2,0]=(-0.593547630941)*x_ref[0]+(-0.0390465590147)*x_ref[1]+(0.00129707682534)*x_ref[2]
        ref[3,0,2,1]=(1.39210293617)*x_ref[0]+(-1.05660957806)*x_ref[1]+(-1.20489109051)*x_ref[2]
        ref[3,1,0,0]=(-0.536035988946)*x_ref[0]+(-0.594403345465)*x_ref[1]+(1.11719111746)*x_ref[2]
        ref[3,1,0,1]=(-0.0403557954127)*x_ref[0]+(-1.32391025453)*x_ref[1]+(-0.639127329416)*x_ref[2]
        ref[3,1,1,0]=(-0.495803710666)*x_ref[0]+(-0.199832989186)*x_ref[1]+(-0.832862076561)*x_ref[2]
        ref[3,1,1,1]=(0.0216882402803)*x_ref[0]+(0.367268873436)*x_ref[1]+(0.295944335418)*x_ref[2]
        ref[3,1,2,0]=(0.172936506269)*x_ref[0]+(-0.339886198196)*x_ref[1]+(0.268201481244)*x_ref[2]
        ref[3,1,2,1]=(0.665786569897)*x_ref[0]+(-0.290637226736)*x_ref[1]+(0.617677178537)*x_ref[2]
        ref[3,2,0,0]=(-0.657902185911)*x_ref[0]+(-1.34292379802)*x_ref[1]+(0.0641950819134)*x_ref[2]
        ref[3,2,0,1]=(0.352039175923)*x_ref[0]+(-0.483047873735)*x_ref[1]+(0.47997600743)*x_ref[2]
        ref[3,2,1,0]=(0.944619484106)*x_ref[0]+(-0.308755547249)*x_ref[1]+(-0.765906593989)*x_ref[2]
        ref[3,2,1,1]=(-1.09674781865)*x_ref[0]+(0.013069257236)*x_ref[1]+(-0.0741247275628)*x_ref[2]
        ref[3,2,2,0]=(0.400057737624)*x_ref[0]+(-1.16165281514)*x_ref[1]+(-0.789757454741)*x_ref[2]
        ref[3,2,2,1]=(0.367100312493)*x_ref[0]+(-0.547338908792)*x_ref[1]+(-0.409173859392)*x_ref[2]
        ref[3,3,0,0]=(0.890118369638)*x_ref[0]+(1.201296587)*x_ref[1]+(-0.0319246476545)*x_ref[2]
        ref[3,3,0,1]=(0.620955702501)*x_ref[0]+(0.438440818114)*x_ref[1]+(0.499747510411)*x_ref[2]
        ref[3,3,1,0]=(1.61217537222)*x_ref[0]+(-1.21323714148)*x_ref[1]+(-1.06137097571)*x_ref[2]
        ref[3,3,1,1]=(-1.75942413542)*x_ref[0]+(0.440345678738)*x_ref[1]+(0.625299774096)*x_ref[2]
        ref[3,3,2,0]=(-0.878497266865)*x_ref[0]+(-1.15879244679)*x_ref[1]+(1.03993124425)*x_ref[2]
        ref[3,3,2,1]=(-0.0138445968491)*x_ref[0]+(0.547910692711)*x_ref[1]+(0.971642264613)*x_ref[2]
        ref[3,4,0,0]=(-0.107769272336)*x_ref[0]+(-0.987911384393)*x_ref[1]+(1.2076201556)*x_ref[2]
        ref[3,4,0,1]=(-0.556570952071)*x_ref[0]+(-0.284568452888)*x_ref[1]+(1.23341917108)*x_ref[2]
        ref[3,4,1,0]=(1.56461193947)*x_ref[0]+(0.185382898933)*x_ref[1]+(-1.67714180845)*x_ref[2]
        ref[3,4,1,1]=(-0.894934034676)*x_ref[0]+(1.05560233139)*x_ref[1]+(0.651029150593)*x_ref[2]
        ref[3,4,2,0]=(0.0559603779066)*x_ref[0]+(-1.72752549101)*x_ref[1]+(-1.02017554749)*x_ref[2]
        ref[3,4,2,1]=(-1.75424333815)*x_ref[0]+(-0.198869840282)*x_ref[1]+(0.475065396299)*x_ref[2]
      res=interpolate(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 5, 3, 2),"wrong shape of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_ContinuousFunction(self):
      """
      tests divergence of Data on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports div on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.627139584179)*x[0]**o+(0.0871027334774)*x[0]+(-0.540784323698)*x[1]**o+(0.951143032871)*x[1]
        arg[1]=(-0.119239347115)*x[0]**o+(0.277154010744)*x[0]+(-0.0745557686973)*x[1]**o+(-0.15832879535)*x[1]
        ref=o*(-0.627139584179)*x_ref[0]**(o-1)+o*(-0.0745557686973)*x_ref[1]**(o-1)+(-0.0712260618722)
      else:
        arg[0]=(0.687070465546)*x[0]**o+(0.789237250884)*x[0]+(-0.00415360035876)*x[1]**o+(-0.809802614239)*x[1]+(0.418087966616)*x[2]**o+(0.730203229934)*x[2]
        arg[1]=(-0.0962217374268)*x[0]**o+(-0.180692492986)*x[0]+(-0.59921793179)*x[1]**o+(-0.799342788982)*x[1]+(0.421364766686)*x[2]**o+(-0.586211831795)*x[2]
        arg[2]=(0.774915504102)*x[0]**o+(-0.968436517935)*x[0]+(0.325821385033)*x[1]**o+(-0.159885723341)*x[1]+(0.411134413872)*x[2]**o+(0.083388357951)*x[2]
        ref=o*(0.687070465546)*x_ref[0]**(o-1)+o*(-0.59921793179)*x_ref[1]**(o-1)+o*(0.411134413872)*x_ref[2]**(o-1)+(0.0732828198535)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_Solution(self):
      """
      tests divergence of Data on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports div on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.037254966228)*x[0]**o+(-0.175442318226)*x[0]+(-0.723089592066)*x[1]**o+(-0.487393332634)*x[1]
        arg[1]=(0.617589532887)*x[0]**o+(0.392126578405)*x[0]+(0.612110931251)*x[1]**o+(-0.932236829843)*x[1]
        ref=o*(-0.037254966228)*x_ref[0]**(o-1)+o*(0.612110931251)*x_ref[1]**(o-1)+(-1.10767914807)
      else:
        arg[0]=(-0.855344637098)*x[0]**o+(0.152992485241)*x[0]+(0.519809077415)*x[1]**o+(-0.0514985311411)*x[1]+(-0.473095642241)*x[2]**o+(-0.743711945025)*x[2]
        arg[1]=(0.629819444749)*x[0]**o+(-0.649277860811)*x[0]+(-0.382231371814)*x[1]**o+(-0.354475847319)*x[1]+(0.523961319618)*x[2]**o+(-0.0301027430993)*x[2]
        arg[2]=(0.489214433445)*x[0]**o+(-0.340824606014)*x[0]+(-0.78445257629)*x[1]**o+(-0.821498813523)*x[1]+(-0.147119976738)*x[2]**o+(0.464427291378)*x[2]
        ref=o*(-0.855344637098)*x_ref[0]**(o-1)+o*(-0.382231371814)*x_ref[1]**(o-1)+o*(-0.147119976738)*x_ref[2]**(o-1)+(0.2629439293)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_ReducedSolution(self):
      """
      tests divergence of Data on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports div on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromData_ReducedContinuousFunction(self):
      """
      tests divergence of Data on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports div on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromSymbol_ContinuousFunction(self):
      """
      tests divergence of Symbol on the Function

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports div on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(0.056440576841)*x[0]**o+(0.246232077047)*x[0]+(0.855071175271)*x[1]**o+(0.501469853761)*x[1]
        s[1]=(0.138795877959)*x[0]**o+(-0.272898431539)*x[0]+(0.446848353852)*x[1]**o+(0.590328561048)*x[1]
        ref=o*(0.056440576841)*x_ref[0]**(o-1)+o*(0.446848353852)*x_ref[1]**(o-1)+(0.836560638095)
      else:
        s[0]=(-0.50629608129)*x[0]**o+(-0.741872423345)*x[0]+(-0.386081173959)*x[1]**o+(0.318875107198)*x[1]+(-0.385714540576)*x[2]**o+(-0.628522049303)*x[2]
        s[1]=(-0.174000956733)*x[0]**o+(-0.883773784269)*x[0]+(-0.811120942906)*x[1]**o+(-0.16019437105)*x[1]+(-0.385051346205)*x[2]**o+(0.988540730072)*x[2]
        s[2]=(0.354661225031)*x[0]**o+(0.900659468106)*x[0]+(0.589271296273)*x[1]**o+(0.339143175651)*x[1]+(-0.50750318566)*x[2]**o+(0.934057146824)*x[2]
        ref=o*(-0.50629608129)*x_ref[0]**(o-1)+o*(-0.811120942906)*x_ref[1]**(o-1)+o*(-0.50750318566)*x_ref[2]**(o-1)+(0.0319903524287)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromSymbol_Solution(self):
      """
      tests divergence of Symbol on the Function

      assumptions: Solution(self.domain) exists
                   self.domain supports div on Function
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(-0.170850365394)*x[0]**o+(0.150036471084)*x[0]+(0.0923309607854)*x[1]**o+(-0.296802366309)*x[1]
        s[1]=(0.61420032995)*x[0]**o+(-0.430079619352)*x[0]+(0.0912728436628)*x[1]**o+(0.817677896693)*x[1]
        ref=o*(-0.170850365394)*x_ref[0]**(o-1)+o*(0.0912728436628)*x_ref[1]**(o-1)+(0.967714367777)
      else:
        s[0]=(-0.740025332388)*x[0]**o+(0.628542391186)*x[0]+(0.201133646437)*x[1]**o+(0.518904232415)*x[1]+(0.745948286476)*x[2]**o+(0.6146222318)*x[2]
        s[1]=(-0.646212542688)*x[0]**o+(-0.687137148869)*x[0]+(-0.162072841283)*x[1]**o+(0.627944920212)*x[1]+(-0.713818605572)*x[2]**o+(0.246460903592)*x[2]
        s[2]=(-0.736234793544)*x[0]**o+(-0.382762645761)*x[0]+(-0.837579788578)*x[1]**o+(-0.913690456246)*x[1]+(0.364622848764)*x[2]**o+(0.241650777953)*x[2]
        ref=o*(-0.740025332388)*x_ref[0]**(o-1)+o*(-0.162072841283)*x_ref[1]**(o-1)+o*(0.364622848764)*x_ref[2]**(o-1)+(1.49813808935)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromSymbol_ReducedSolution(self):
      """
      tests divergence of Symbol on the Function

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports div on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(0.0131350820952)*x[0]+(-1.82221871934)*x[1]
        s[1]=(-1.89777545856)*x[0]+(0.191135975577)*x[1]
        ref=(0.204271057672)
      else:
        s[0]=(-1.30169999796)*x[0]+(-0.550711448862)*x[1]+(0.241706428593)*x[2]
        s[1]=(0.843465810567)*x[0]+(-1.42460661253)*x[1]+(1.72435488277)*x[2]
        s[2]=(0.290030514202)*x[0]+(1.75177720442)*x[1]+(0.568532008682)*x[2]
        ref=(-2.15777460181)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onFunction_fromSymbol_ReducedContinuousFunction(self):
      """
      tests divergence of Symbol on the Function

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports div on Function
      """
      o=1
      dim=self.domain.getDim()
      w_ref=Function(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(0.0131350820952)*x[0]+(-1.82221871934)*x[1]
        s[1]=(-1.89777545856)*x[0]+(0.191135975577)*x[1]
        ref=(0.204271057672)
      else:
        s[0]=(-1.30169999796)*x[0]+(-0.550711448862)*x[1]+(0.241706428593)*x[2]
        s[1]=(0.843465810567)*x[0]+(-1.42460661253)*x[1]+(1.72435488277)*x[2]
        s[2]=(0.290030514202)*x[0]+(1.75177720442)*x[1]+(0.568532008682)*x[2]
        ref=(-2.15777460181)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_ContinuousFunction(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.627139584179)*x[0]**o+(0.0871027334774)*x[0]+(-0.540784323698)*x[1]**o+(0.951143032871)*x[1]
        arg[1]=(-0.119239347115)*x[0]**o+(0.277154010744)*x[0]+(-0.0745557686973)*x[1]**o+(-0.15832879535)*x[1]
        ref=o*(-0.627139584179)*x_ref[0]**(o-1)+o*(-0.0745557686973)*x_ref[1]**(o-1)+(-0.0712260618722)
      else:
        arg[0]=(0.687070465546)*x[0]**o+(0.789237250884)*x[0]+(-0.00415360035876)*x[1]**o+(-0.809802614239)*x[1]+(0.418087966616)*x[2]**o+(0.730203229934)*x[2]
        arg[1]=(-0.0962217374268)*x[0]**o+(-0.180692492986)*x[0]+(-0.59921793179)*x[1]**o+(-0.799342788982)*x[1]+(0.421364766686)*x[2]**o+(-0.586211831795)*x[2]
        arg[2]=(0.774915504102)*x[0]**o+(-0.968436517935)*x[0]+(0.325821385033)*x[1]**o+(-0.159885723341)*x[1]+(0.411134413872)*x[2]**o+(0.083388357951)*x[2]
        ref=o*(0.687070465546)*x_ref[0]**(o-1)+o*(-0.59921793179)*x_ref[1]**(o-1)+o*(0.411134413872)*x_ref[2]**(o-1)+(0.0732828198535)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_Solution(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(-0.037254966228)*x[0]**o+(-0.175442318226)*x[0]+(-0.723089592066)*x[1]**o+(-0.487393332634)*x[1]
        arg[1]=(0.617589532887)*x[0]**o+(0.392126578405)*x[0]+(0.612110931251)*x[1]**o+(-0.932236829843)*x[1]
        ref=o*(-0.037254966228)*x_ref[0]**(o-1)+o*(0.612110931251)*x_ref[1]**(o-1)+(-1.10767914807)
      else:
        arg[0]=(-0.855344637098)*x[0]**o+(0.152992485241)*x[0]+(0.519809077415)*x[1]**o+(-0.0514985311411)*x[1]+(-0.473095642241)*x[2]**o+(-0.743711945025)*x[2]
        arg[1]=(0.629819444749)*x[0]**o+(-0.649277860811)*x[0]+(-0.382231371814)*x[1]**o+(-0.354475847319)*x[1]+(0.523961319618)*x[2]**o+(-0.0301027430993)*x[2]
        arg[2]=(0.489214433445)*x[0]**o+(-0.340824606014)*x[0]+(-0.78445257629)*x[1]**o+(-0.821498813523)*x[1]+(-0.147119976738)*x[2]**o+(0.464427291378)*x[2]
        ref=o*(-0.855344637098)*x_ref[0]**(o-1)+o*(-0.382231371814)*x_ref[1]**(o-1)+o*(-0.147119976738)*x_ref[2]**(o-1)+(0.2629439293)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_ReducedSolution(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromData_ReducedContinuousFunction(self):
      """
      tests divergence of Data on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Vector(0,w)
      if dim==2:
        arg[0]=(0.983549953155)*x[0]+(0.741104834322)*x[1]
        arg[1]=(0.548447952865)*x[0]+(-0.817326874507)*x[1]
        ref=(0.166223078648)
      else:
        arg[0]=(1.39002615085)*x[0]+(0.609245504565)*x[1]+(-1.06625041498)*x[2]
        arg[1]=(-1.13329921103)*x[0]+(0.160224729695)*x[1]+(0.408632198066)*x[2]
        arg[2]=(0.124694320475)*x[0]+(-1.53975404419)*x[1]+(1.00224340458)*x[2]
        ref=(2.55249428513)
      res=div(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromSymbol_ContinuousFunction(self):
      """
      tests divergence of Symbol on the ReducedFunction

      assumptions: ContinuousFunction(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(0.056440576841)*x[0]**o+(0.246232077047)*x[0]+(0.855071175271)*x[1]**o+(0.501469853761)*x[1]
        s[1]=(0.138795877959)*x[0]**o+(-0.272898431539)*x[0]+(0.446848353852)*x[1]**o+(0.590328561048)*x[1]
        ref=o*(0.056440576841)*x_ref[0]**(o-1)+o*(0.446848353852)*x_ref[1]**(o-1)+(0.836560638095)
      else:
        s[0]=(-0.50629608129)*x[0]**o+(-0.741872423345)*x[0]+(-0.386081173959)*x[1]**o+(0.318875107198)*x[1]+(-0.385714540576)*x[2]**o+(-0.628522049303)*x[2]
        s[1]=(-0.174000956733)*x[0]**o+(-0.883773784269)*x[0]+(-0.811120942906)*x[1]**o+(-0.16019437105)*x[1]+(-0.385051346205)*x[2]**o+(0.988540730072)*x[2]
        s[2]=(0.354661225031)*x[0]**o+(0.900659468106)*x[0]+(0.589271296273)*x[1]**o+(0.339143175651)*x[1]+(-0.50750318566)*x[2]**o+(0.934057146824)*x[2]
        ref=o*(-0.50629608129)*x_ref[0]**(o-1)+o*(-0.811120942906)*x_ref[1]**(o-1)+o*(-0.50750318566)*x_ref[2]**(o-1)+(0.0319903524287)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromSymbol_Solution(self):
      """
      tests divergence of Symbol on the ReducedFunction

      assumptions: Solution(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=self.order
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=Solution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(-0.170850365394)*x[0]**o+(0.150036471084)*x[0]+(0.0923309607854)*x[1]**o+(-0.296802366309)*x[1]
        s[1]=(0.61420032995)*x[0]**o+(-0.430079619352)*x[0]+(0.0912728436628)*x[1]**o+(0.817677896693)*x[1]
        ref=o*(-0.170850365394)*x_ref[0]**(o-1)+o*(0.0912728436628)*x_ref[1]**(o-1)+(0.967714367777)
      else:
        s[0]=(-0.740025332388)*x[0]**o+(0.628542391186)*x[0]+(0.201133646437)*x[1]**o+(0.518904232415)*x[1]+(0.745948286476)*x[2]**o+(0.6146222318)*x[2]
        s[1]=(-0.646212542688)*x[0]**o+(-0.687137148869)*x[0]+(-0.162072841283)*x[1]**o+(0.627944920212)*x[1]+(-0.713818605572)*x[2]**o+(0.246460903592)*x[2]
        s[2]=(-0.736234793544)*x[0]**o+(-0.382762645761)*x[0]+(-0.837579788578)*x[1]**o+(-0.913690456246)*x[1]+(0.364622848764)*x[2]**o+(0.241650777953)*x[2]
        ref=o*(-0.740025332388)*x_ref[0]**(o-1)+o*(-0.162072841283)*x_ref[1]**(o-1)+o*(0.364622848764)*x_ref[2]**(o-1)+(1.49813808935)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromSymbol_ReducedSolution(self):
      """
      tests divergence of Symbol on the ReducedFunction

      assumptions: ReducedSolution(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedSolution(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(0.0131350820952)*x[0]+(-1.82221871934)*x[1]
        s[1]=(-1.89777545856)*x[0]+(0.191135975577)*x[1]
        ref=(0.204271057672)
      else:
        s[0]=(-1.30169999796)*x[0]+(-0.550711448862)*x[1]+(0.241706428593)*x[2]
        s[1]=(0.843465810567)*x[0]+(-1.42460661253)*x[1]+(1.72435488277)*x[2]
        s[2]=(0.290030514202)*x[0]+(1.75177720442)*x[1]+(0.568532008682)*x[2]
        ref=(-2.15777460181)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_div_onReducedFunction_fromSymbol_ReducedContinuousFunction(self):
      """
      tests divergence of Symbol on the ReducedFunction

      assumptions: ReducedContinuousFunction(self.domain) exists
                   self.domain supports div on ReducedFunction
      """
      o=1
      dim=self.domain.getDim()
      w_ref=ReducedFunction(self.domain)
      x_ref=w_ref.getX()
      w=ReducedContinuousFunction(self.domain)
      x=w.getX()
      arg=Symbol(shape=(dim,),dim=dim)
      s=Vector(0,w)
      if dim==2:
        s[0]=(0.0131350820952)*x[0]+(-1.82221871934)*x[1]
        s[1]=(-1.89777545856)*x[0]+(0.191135975577)*x[1]
        ref=(0.204271057672)
      else:
        s[0]=(-1.30169999796)*x[0]+(-0.550711448862)*x[1]+(0.241706428593)*x[2]
        s[1]=(0.843465810567)*x[0]+(-1.42460661253)*x[1]+(1.72435488277)*x[2]
        s[2]=(0.290030514202)*x[0]+(1.75177720442)*x[1]+(0.568532008682)*x[2]
        ref=(-2.15777460181)
      res=div(arg,where=w_ref)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertEqual(sub.getFunctionSpace(),w_ref,"wrong function space of result.")
      self.assertTrue(Lsup(sub-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank0(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(0.608797336225)*x[0]
        ref=sqrt((0.123544732198))

      else:
        arg=Data(0,(),w)
        arg=(0.136031275673)*x[0]
        ref=sqrt((0.00616816932037))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank1(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(-0.212143919436)*x[0]
        arg[1]=(-0.256194155686)*x[1]
        ref=sqrt((0.0368801626538))

      else:
        arg=Data(0,(3,),w)
        arg[0]=(0.0452831341416)*x[0]
        arg[1]=(-0.278640180656)*x[1]
        arg[2]=(-0.607035001062)*x[2]
        ref=sqrt((0.149394135009))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank2(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.239448813076)*x[0]
        arg[0,1]=(-0.529349708753)*x[1]
        arg[1,0]=(-0.381557161859)*x[0]
        arg[1,1]=(0.731658534249)*x[1]
        arg[2,0]=(-0.813679062342)*x[0]
        arg[2,1]=(0.528100089704)*x[1]
        arg[3,0]=(-0.480867528161)*x[0]
        arg[3,1]=(-0.167862206972)*x[1]
        ref=sqrt((0.739610516051))

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(0.951209543612)*x[0]
        arg[0,1]=(0.735178735637)*x[1]
        arg[0,2]=(0.13074673272)*x[2]
        arg[1,0]=(0.412295676715)*x[0]
        arg[1,1]=(-0.657695950153)*x[1]
        arg[1,2]=(-0.900044734695)*x[2]
        arg[2,0]=(0.741773926224)*x[0]
        arg[2,1]=(0.0521828807406)*x[1]
        arg[2,2]=(0.797728501985)*x[2]
        arg[3,0]=(-0.61235554051)*x[0]
        arg[3,1]=(0.456652747412)*x[1]
        arg[3,2]=(-0.734303857319)*x[2]
        ref=sqrt((1.72901661926))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank3(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(0.449174971953)*x[0]
        arg[0,0,1]=(-0.0109398763289)*x[1]
        arg[0,1,0]=(-0.202497187709)*x[0]
        arg[0,1,1]=(-0.12970879334)*x[1]
        arg[1,0,0]=(-0.138092481719)*x[0]
        arg[1,0,1]=(-0.528752200917)*x[1]
        arg[1,1,0]=(-0.605919441662)*x[0]
        arg[1,1,1]=(0.215615032334)*x[1]
        arg[2,0,0]=(-0.998734541972)*x[0]
        arg[2,0,1]=(0.725811901251)*x[1]
        arg[2,1,0]=(-0.966536503228)*x[0]
        arg[2,1,1]=(-0.528692217355)*x[1]
        arg[3,0,0]=(0.757633851466)*x[0]
        arg[3,0,1]=(-0.524660157377)*x[1]
        arg[3,1,0]=(0.983733431677)*x[0]
        arg[3,1,1]=(0.061279109546)*x[1]
        arg[4,0,0]=(0.85914215305)*x[0]
        arg[4,0,1]=(0.941714045112)*x[1]
        arg[4,1,0]=(0.172235529555)*x[0]
        arg[4,1,1]=(-0.108381454437)*x[1]
        arg[5,0,0]=(-0.736373697727)*x[0]
        arg[5,0,1]=(-0.599337929679)*x[1]
        arg[5,1,0]=(0.661072686392)*x[0]
        arg[5,1,1]=(-0.55107327409)*x[1]
        ref=sqrt((2.94641432714))

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.69227064904)*x[0]
        arg[0,0,1]=(-0.968336177418)*x[1]
        arg[0,0,2]=(-0.634883146685)*x[2]
        arg[0,1,0]=(-0.12640661422)*x[0]
        arg[0,1,1]=(-0.637386589888)*x[1]
        arg[0,1,2]=(0.26060859356)*x[2]
        arg[1,0,0]=(-0.986864633297)*x[0]
        arg[1,0,1]=(-0.441589142379)*x[1]
        arg[1,0,2]=(-0.587865539582)*x[2]
        arg[1,1,0]=(0.596052465031)*x[0]
        arg[1,1,1]=(0.312732336652)*x[1]
        arg[1,1,2]=(-0.514423945092)*x[2]
        arg[2,0,0]=(-0.892391254794)*x[0]
        arg[2,0,1]=(0.377920185756)*x[1]
        arg[2,0,2]=(-0.120174597181)*x[2]
        arg[2,1,0]=(-0.469951576468)*x[0]
        arg[2,1,1]=(-0.788362249555)*x[1]
        arg[2,1,2]=(0.745625354986)*x[2]
        arg[3,0,0]=(0.542802498569)*x[0]
        arg[3,0,1]=(-0.814541028706)*x[1]
        arg[3,0,2]=(0.298410992196)*x[2]
        arg[3,1,0]=(0.981190341206)*x[0]
        arg[3,1,1]=(0.666421298608)*x[1]
        arg[3,1,2]=(-0.369751722626)*x[2]
        arg[4,0,0]=(-0.75379530597)*x[0]
        arg[4,0,1]=(0.283357267139)*x[1]
        arg[4,0,2]=(0.247787072861)*x[2]
        arg[4,1,0]=(0.301766692533)*x[0]
        arg[4,1,1]=(0.828183439224)*x[1]
        arg[4,1,2]=(-0.580824060547)*x[2]
        arg[5,0,0]=(0.637345610764)*x[0]
        arg[5,0,1]=(-0.234409115997)*x[1]
        arg[5,0,2]=(-0.192639300316)*x[2]
        arg[5,1,0]=(-0.62609237162)*x[0]
        arg[5,1,1]=(0.463404958552)*x[1]
        arg[5,1,2]=(-0.547814448738)*x[2]
        ref=sqrt((4.2381131862))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromData_rank4(self):
      """
      tests L2-norm of Data on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.232618585183)*x[0]
        arg[0,0,0,1]=(0.39796117869)*x[1]
        arg[0,0,1,0]=(-0.997336958262)*x[0]
        arg[0,0,1,1]=(-0.351780915076)*x[1]
        arg[0,0,2,0]=(-0.876764070136)*x[0]
        arg[0,0,2,1]=(0.808730805817)*x[1]
        arg[0,1,0,0]=(-0.197154744966)*x[0]
        arg[0,1,0,1]=(0.416246096086)*x[1]
        arg[0,1,1,0]=(0.708038457121)*x[0]
        arg[0,1,1,1]=(-0.00954021503188)*x[1]
        arg[0,1,2,0]=(-0.62630809425)*x[0]
        arg[0,1,2,1]=(0.430228727912)*x[1]
        arg[0,2,0,0]=(0.0306704747648)*x[0]
        arg[0,2,0,1]=(-0.913877199453)*x[1]
        arg[0,2,1,0]=(-0.697612800829)*x[0]
        arg[0,2,1,1]=(-0.17996376822)*x[1]
        arg[0,2,2,0]=(-0.304509578871)*x[0]
        arg[0,2,2,1]=(-0.610556755811)*x[1]
        arg[0,3,0,0]=(-0.452355972234)*x[0]
        arg[0,3,0,1]=(-0.368921242518)*x[1]
        arg[0,3,1,0]=(-0.478275554932)*x[0]
        arg[0,3,1,1]=(0.257178549127)*x[1]
        arg[0,3,2,0]=(0.530736487177)*x[0]
        arg[0,3,2,1]=(-0.567126272463)*x[1]
        arg[0,4,0,0]=(0.801519165938)*x[0]
        arg[0,4,0,1]=(-0.509816703951)*x[1]
        arg[0,4,1,0]=(-0.255412646934)*x[0]
        arg[0,4,1,1]=(0.437540101896)*x[1]
        arg[0,4,2,0]=(-0.815574969538)*x[0]
        arg[0,4,2,1]=(-0.94691547137)*x[1]
        arg[1,0,0,0]=(-0.732550722593)*x[0]
        arg[1,0,0,1]=(0.515752381704)*x[1]
        arg[1,0,1,0]=(-0.343590210899)*x[0]
        arg[1,0,1,1]=(-0.0601907964915)*x[1]
        arg[1,0,2,0]=(0.0199916154421)*x[0]
        arg[1,0,2,1]=(-0.136927227821)*x[1]
        arg[1,1,0,0]=(0.397994441702)*x[0]
        arg[1,1,0,1]=(0.953873148948)*x[1]
        arg[1,1,1,0]=(0.419416235967)*x[0]
        arg[1,1,1,1]=(0.700998577193)*x[1]
        arg[1,1,2,0]=(-0.497358799271)*x[0]
        arg[1,1,2,1]=(0.0851768858379)*x[1]
        arg[1,2,0,0]=(0.0936678875202)*x[0]
        arg[1,2,0,1]=(0.869883786896)*x[1]
        arg[1,2,1,0]=(0.582700123485)*x[0]
        arg[1,2,1,1]=(-0.433381106794)*x[1]
        arg[1,2,2,0]=(-0.527031777974)*x[0]
        arg[1,2,2,1]=(0.105105137652)*x[1]
        arg[1,3,0,0]=(-0.716750829134)*x[0]
        arg[1,3,0,1]=(0.774519209008)*x[1]
        arg[1,3,1,0]=(-0.568743372716)*x[0]
        arg[1,3,1,1]=(0.794732483944)*x[1]
        arg[1,3,2,0]=(0.246606002015)*x[0]
        arg[1,3,2,1]=(-0.988869494994)*x[1]
        arg[1,4,0,0]=(0.482379298083)*x[0]
        arg[1,4,0,1]=(-0.386268387903)*x[1]
        arg[1,4,1,0]=(0.137184889675)*x[0]
        arg[1,4,1,1]=(-0.140520035321)*x[1]
        arg[1,4,2,0]=(0.822755050415)*x[0]
        arg[1,4,2,1]=(-0.815562139522)*x[1]
        arg[2,0,0,0]=(-0.462891511962)*x[0]
        arg[2,0,0,1]=(-0.122643411631)*x[1]
        arg[2,0,1,0]=(-0.520861119962)*x[0]
        arg[2,0,1,1]=(-0.881189618018)*x[1]
        arg[2,0,2,0]=(-0.776157842774)*x[0]
        arg[2,0,2,1]=(-0.12354053207)*x[1]
        arg[2,1,0,0]=(0.395495230826)*x[0]
        arg[2,1,0,1]=(-0.388106659423)*x[1]
        arg[2,1,1,0]=(0.354250242834)*x[0]
        arg[2,1,1,1]=(-0.666514210192)*x[1]
        arg[2,1,2,0]=(0.951294655083)*x[0]
        arg[2,1,2,1]=(0.074024416386)*x[1]
        arg[2,2,0,0]=(0.335448485459)*x[0]
        arg[2,2,0,1]=(-0.40988282528)*x[1]
        arg[2,2,1,0]=(-0.805725968875)*x[0]
        arg[2,2,1,1]=(-0.949883082118)*x[1]
        arg[2,2,2,0]=(0.531549210683)*x[0]
        arg[2,2,2,1]=(-0.398401016682)*x[1]
        arg[2,3,0,0]=(-0.953963433205)*x[0]
        arg[2,3,0,1]=(0.643431126406)*x[1]
        arg[2,3,1,0]=(-0.167611998738)*x[0]
        arg[2,3,1,1]=(0.226130056552)*x[1]
        arg[2,3,2,0]=(0.0752687641131)*x[0]
        arg[2,3,2,1]=(-0.115742756362)*x[1]
        arg[2,4,0,0]=(0.579694491028)*x[0]
        arg[2,4,0,1]=(-0.112005738299)*x[1]
        arg[2,4,1,0]=(0.657291764224)*x[0]
        arg[2,4,1,1]=(0.62671154177)*x[1]
        arg[2,4,2,0]=(0.103695027944)*x[0]
        arg[2,4,2,1]=(0.462828491544)*x[1]
        arg[3,0,0,0]=(0.697692979998)*x[0]
        arg[3,0,0,1]=(-0.123481859619)*x[1]
        arg[3,0,1,0]=(-0.749745629459)*x[0]
        arg[3,0,1,1]=(-0.541969524069)*x[1]
        arg[3,0,2,0]=(0.819484470759)*x[0]
        arg[3,0,2,1]=(-0.860592326469)*x[1]
        arg[3,1,0,0]=(-0.716566084771)*x[0]
        arg[3,1,0,1]=(-0.949235434827)*x[1]
        arg[3,1,1,0]=(-0.826699498174)*x[0]
        arg[3,1,1,1]=(-0.138511521583)*x[1]
        arg[3,1,2,0]=(-0.951682890904)*x[0]
        arg[3,1,2,1]=(0.413293316925)*x[1]
        arg[3,2,0,0]=(0.909516836775)*x[0]
        arg[3,2,0,1]=(-0.919989721277)*x[1]
        arg[3,2,1,0]=(0.0994860369337)*x[0]
        arg[3,2,1,1]=(-0.933647246623)*x[1]
        arg[3,2,2,0]=(-0.759215183015)*x[0]
        arg[3,2,2,1]=(0.0975793309286)*x[1]
        arg[3,3,0,0]=(-0.130256739381)*x[0]
        arg[3,3,0,1]=(-0.582280862311)*x[1]
        arg[3,3,1,0]=(0.206970526192)*x[0]
        arg[3,3,1,1]=(-0.8678322258)*x[1]
        arg[3,3,2,0]=(0.133004501279)*x[0]
        arg[3,3,2,1]=(0.802921710935)*x[1]
        arg[3,4,0,0]=(-0.255737792764)*x[0]
        arg[3,4,0,1]=(-0.34168114937)*x[1]
        arg[3,4,1,0]=(-0.859309090399)*x[0]
        arg[3,4,1,1]=(0.245043986435)*x[1]
        arg[3,4,2,0]=(0.893062018695)*x[0]
        arg[3,4,2,1]=(0.709422742588)*x[1]
        ref=sqrt((13.7289280362))

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.0312828390439)*x[0]
        arg[0,0,0,1]=(-0.524970416212)*x[1]
        arg[0,0,0,2]=(0.561865217554)*x[2]
        arg[0,0,1,0]=(0.692457187384)*x[0]
        arg[0,0,1,1]=(0.946967182157)*x[1]
        arg[0,0,1,2]=(-0.863842279464)*x[2]
        arg[0,0,2,0]=(0.993922921598)*x[0]
        arg[0,0,2,1]=(0.322812768679)*x[1]
        arg[0,0,2,2]=(0.901876132204)*x[2]
        arg[0,1,0,0]=(0.967569979365)*x[0]
        arg[0,1,0,1]=(0.840979131355)*x[1]
        arg[0,1,0,2]=(0.0494811460856)*x[2]
        arg[0,1,1,0]=(0.315178456102)*x[0]
        arg[0,1,1,1]=(0.449848313024)*x[1]
        arg[0,1,1,2]=(0.765887852886)*x[2]
        arg[0,1,2,0]=(0.975541574352)*x[0]
        arg[0,1,2,1]=(-0.797851290751)*x[1]
        arg[0,1,2,2]=(0.628918775319)*x[2]
        arg[0,2,0,0]=(0.685635794312)*x[0]
        arg[0,2,0,1]=(0.10341799962)*x[1]
        arg[0,2,0,2]=(-0.964822756043)*x[2]
        arg[0,2,1,0]=(-0.56160368212)*x[0]
        arg[0,2,1,1]=(0.676344298102)*x[1]
        arg[0,2,1,2]=(-0.713924121843)*x[2]
        arg[0,2,2,0]=(-0.276655136263)*x[0]
        arg[0,2,2,1]=(0.336046973788)*x[1]
        arg[0,2,2,2]=(-0.68789392396)*x[2]
        arg[0,3,0,0]=(0.0172861311571)*x[0]
        arg[0,3,0,1]=(-0.301075956456)*x[1]
        arg[0,3,0,2]=(0.779442985415)*x[2]
        arg[0,3,1,0]=(-0.517629576558)*x[0]
        arg[0,3,1,1]=(0.584779586639)*x[1]
        arg[0,3,1,2]=(-0.53266435436)*x[2]
        arg[0,3,2,0]=(0.841533567102)*x[0]
        arg[0,3,2,1]=(0.0458746415489)*x[1]
        arg[0,3,2,2]=(0.921237870758)*x[2]
        arg[0,4,0,0]=(0.0548343238805)*x[0]
        arg[0,4,0,1]=(0.687022707412)*x[1]
        arg[0,4,0,2]=(-0.319803609795)*x[2]
        arg[0,4,1,0]=(0.409763007811)*x[0]
        arg[0,4,1,1]=(0.165501957435)*x[1]
        arg[0,4,1,2]=(0.116001692781)*x[2]
        arg[0,4,2,0]=(-0.515571394238)*x[0]
        arg[0,4,2,1]=(0.209467945147)*x[1]
        arg[0,4,2,2]=(-0.344827191247)*x[2]
        arg[1,0,0,0]=(0.57193838014)*x[0]
        arg[1,0,0,1]=(-0.0880683799076)*x[1]
        arg[1,0,0,2]=(0.956899617441)*x[2]
        arg[1,0,1,0]=(-0.783689636357)*x[0]
        arg[1,0,1,1]=(-0.25177506885)*x[1]
        arg[1,0,1,2]=(-0.97074584634)*x[2]
        arg[1,0,2,0]=(0.432543519806)*x[0]
        arg[1,0,2,1]=(0.481003021954)*x[1]
        arg[1,0,2,2]=(-0.0630751518268)*x[2]
        arg[1,1,0,0]=(-0.65152446796)*x[0]
        arg[1,1,0,1]=(-0.0323685084425)*x[1]
        arg[1,1,0,2]=(-0.508674033909)*x[2]
        arg[1,1,1,0]=(-0.533367818916)*x[0]
        arg[1,1,1,1]=(0.310738340288)*x[1]
        arg[1,1,1,2]=(0.694612234326)*x[2]
        arg[1,1,2,0]=(-0.622052473032)*x[0]
        arg[1,1,2,1]=(0.0498443793671)*x[1]
        arg[1,1,2,2]=(0.61023707512)*x[2]
        arg[1,2,0,0]=(0.0730267406859)*x[0]
        arg[1,2,0,1]=(0.146909334607)*x[1]
        arg[1,2,0,2]=(-0.641860284448)*x[2]
        arg[1,2,1,0]=(0.917976589737)*x[0]
        arg[1,2,1,1]=(0.50219672122)*x[1]
        arg[1,2,1,2]=(0.634559579812)*x[2]
        arg[1,2,2,0]=(0.0578772734534)*x[0]
        arg[1,2,2,1]=(0.288730973517)*x[1]
        arg[1,2,2,2]=(-0.0525978796154)*x[2]
        arg[1,3,0,0]=(-0.926152433388)*x[0]
        arg[1,3,0,1]=(0.0616647680855)*x[1]
        arg[1,3,0,2]=(-0.875889217846)*x[2]
        arg[1,3,1,0]=(-0.638931542845)*x[0]
        arg[1,3,1,1]=(0.708848122964)*x[1]
        arg[1,3,1,2]=(0.119066979792)*x[2]
        arg[1,3,2,0]=(0.853716218591)*x[0]
        arg[1,3,2,1]=(-0.92754322201)*x[1]
        arg[1,3,2,2]=(-0.671530626265)*x[2]
        arg[1,4,0,0]=(0.337424536231)*x[0]
        arg[1,4,0,1]=(0.335704451719)*x[1]
        arg[1,4,0,2]=(-0.484565969466)*x[2]
        arg[1,4,1,0]=(-0.855476192012)*x[0]
        arg[1,4,1,1]=(0.405674615553)*x[1]
        arg[1,4,1,2]=(0.728310771323)*x[2]
        arg[1,4,2,0]=(0.363651308265)*x[0]
        arg[1,4,2,1]=(0.174460594531)*x[1]
        arg[1,4,2,2]=(-0.0418244838617)*x[2]
        arg[2,0,0,0]=(-0.531341992511)*x[0]
        arg[2,0,0,1]=(0.584996796272)*x[1]
        arg[2,0,0,2]=(-0.752430968716)*x[2]
        arg[2,0,1,0]=(-0.341989849747)*x[0]
        arg[2,0,1,1]=(0.153572646953)*x[1]
        arg[2,0,1,2]=(-0.197130051737)*x[2]
        arg[2,0,2,0]=(-0.338082424082)*x[0]
        arg[2,0,2,1]=(0.000173657394772)*x[1]
        arg[2,0,2,2]=(0.365272907692)*x[2]
        arg[2,1,0,0]=(0.904304126564)*x[0]
        arg[2,1,0,1]=(0.161252368484)*x[1]
        arg[2,1,0,2]=(0.246854092422)*x[2]
        arg[2,1,1,0]=(-0.299880647529)*x[0]
        arg[2,1,1,1]=(-0.566917528608)*x[1]
        arg[2,1,1,2]=(0.243183337285)*x[2]
        arg[2,1,2,0]=(0.437406011474)*x[0]
        arg[2,1,2,1]=(0.727447394053)*x[1]
        arg[2,1,2,2]=(0.380752950664)*x[2]
        arg[2,2,0,0]=(0.172292846911)*x[0]
        arg[2,2,0,1]=(0.334201791643)*x[1]
        arg[2,2,0,2]=(0.739989926962)*x[2]
        arg[2,2,1,0]=(-0.0669843715042)*x[0]
        arg[2,2,1,1]=(-0.540497281635)*x[1]
        arg[2,2,1,2]=(-0.744217027088)*x[2]
        arg[2,2,2,0]=(-0.287295952259)*x[0]
        arg[2,2,2,1]=(-0.512411849183)*x[1]
        arg[2,2,2,2]=(0.953107417666)*x[2]
        arg[2,3,0,0]=(0.998168116695)*x[0]
        arg[2,3,0,1]=(0.960065646359)*x[1]
        arg[2,3,0,2]=(0.110048258832)*x[2]
        arg[2,3,1,0]=(-0.477271134724)*x[0]
        arg[2,3,1,1]=(0.707182612251)*x[1]
        arg[2,3,1,2]=(0.285500891755)*x[2]
        arg[2,3,2,0]=(-0.863497506661)*x[0]
        arg[2,3,2,1]=(-0.293917669879)*x[1]
        arg[2,3,2,2]=(-0.403384244295)*x[2]
        arg[2,4,0,0]=(0.848455277702)*x[0]
        arg[2,4,0,1]=(-0.530101455578)*x[1]
        arg[2,4,0,2]=(0.33887313048)*x[2]
        arg[2,4,1,0]=(-0.195313538124)*x[0]
        arg[2,4,1,1]=(-0.62754572008)*x[1]
        arg[2,4,1,2]=(-0.385132960582)*x[2]
        arg[2,4,2,0]=(0.240048012886)*x[0]
        arg[2,4,2,1]=(0.900766252969)*x[1]
        arg[2,4,2,2]=(0.669620533505)*x[2]
        arg[3,0,0,0]=(0.375766827301)*x[0]
        arg[3,0,0,1]=(0.705484960308)*x[1]
        arg[3,0,0,2]=(0.440931516034)*x[2]
        arg[3,0,1,0]=(-0.44724403177)*x[0]
        arg[3,0,1,1]=(-0.31558249626)*x[1]
        arg[3,0,1,2]=(-0.00419436365172)*x[2]
        arg[3,0,2,0]=(0.750599752032)*x[0]
        arg[3,0,2,1]=(0.367649951795)*x[1]
        arg[3,0,2,2]=(0.0488013073654)*x[2]
        arg[3,1,0,0]=(-0.992890068274)*x[0]
        arg[3,1,0,1]=(0.671447745511)*x[1]
        arg[3,1,0,2]=(0.85613331404)*x[2]
        arg[3,1,1,0]=(-0.46064764242)*x[0]
        arg[3,1,1,1]=(0.48138877715)*x[1]
        arg[3,1,1,2]=(0.396741761803)*x[2]
        arg[3,1,2,0]=(-0.879391967543)*x[0]
        arg[3,1,2,1]=(-0.44039462138)*x[1]
        arg[3,1,2,2]=(0.0330511573872)*x[2]
        arg[3,2,0,0]=(-0.367413701648)*x[0]
        arg[3,2,0,1]=(0.0359818324891)*x[1]
        arg[3,2,0,2]=(-0.307532667032)*x[2]
        arg[3,2,1,0]=(0.334663597166)*x[0]
        arg[3,2,1,1]=(0.541941978066)*x[1]
        arg[3,2,1,2]=(-0.609184079318)*x[2]
        arg[3,2,2,0]=(0.359349239826)*x[0]
        arg[3,2,2,1]=(0.0419272305685)*x[1]
        arg[3,2,2,2]=(0.557189794296)*x[2]
        arg[3,3,0,0]=(-0.85864165554)*x[0]
        arg[3,3,0,1]=(-0.185411404213)*x[1]
        arg[3,3,0,2]=(0.254294865253)*x[2]
        arg[3,3,1,0]=(0.870362177541)*x[0]
        arg[3,3,1,1]=(-0.439688612864)*x[1]
        arg[3,3,1,2]=(0.26006729357)*x[2]
        arg[3,3,2,0]=(-0.0724034754175)*x[0]
        arg[3,3,2,1]=(0.444871564246)*x[1]
        arg[3,3,2,2]=(0.485634530531)*x[2]
        arg[3,4,0,0]=(-0.744756961758)*x[0]
        arg[3,4,0,1]=(0.429761406102)*x[1]
        arg[3,4,0,2]=(-0.584963735834)*x[2]
        arg[3,4,1,0]=(0.684578379159)*x[0]
        arg[3,4,1,1]=(0.949460132601)*x[1]
        arg[3,4,1,2]=(-0.592179909559)*x[2]
        arg[3,4,2,0]=(0.707154437797)*x[0]
        arg[3,4,2,1]=(0.619200407063)*x[1]
        arg[3,4,2,2]=(-0.338547165)*x[2]
        ref=sqrt((19.2170638478))

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromSymbol_rank0(self):
      """
      tests L2-norm of Symbol on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(-0.292546362165)*x[0]
        ref=sqrt((0.0285277913387))

      else:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(0.751025585654)*x[0]
        ref=sqrt((0.188013143436))

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromSymbol_rank1(self):
      """
      tests L2-norm of Symbol on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(2,),dim=dim)
        s=Data(0,(2,),w)
        s[0]=(-0.1759467192)*x[0]
        s[1]=(0.37884376337)*x[1]
        ref=sqrt((0.0581599483473))

      else:
        arg=Symbol(shape=(3,),dim=dim)
        s=Data(0,(3,),w)
        s[0]=(-0.879259245697)*x[0]
        s[1]=(0.528911893259)*x[1]
        s[2]=(-0.751304571514)*x[2]
        ref=sqrt((0.539101057051))

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromSymbol_rank2(self):
      """
      tests L2-norm of Symbol on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 2),dim=dim)
        s=Data(0,(4, 2),w)
        s[0,0]=(0.792724994628)*x[0]
        s[0,1]=(-0.662113166656)*x[1]
        s[1,0]=(-0.548842538341)*x[0]
        s[1,1]=(-0.555018478507)*x[1]
        s[2,0]=(0.42962640812)*x[0]
        s[2,1]=(-0.582731787726)*x[1]
        s[3,0]=(0.603890261621)*x[0]
        s[3,1]=(0.802141268835)*x[1]
        ref=sqrt((1.06944988539))

      else:
        arg=Symbol(shape=(4, 3),dim=dim)
        s=Data(0,(4, 3),w)
        s[0,0]=(-0.432893320336)*x[0]
        s[0,1]=(0.277328929482)*x[1]
        s[0,2]=(0.950129871073)*x[2]
        s[1,0]=(0.675116050333)*x[0]
        s[1,1]=(-0.234274111148)*x[1]
        s[1,2]=(-0.477441654222)*x[2]
        s[2,0]=(0.603834062829)*x[0]
        s[2,1]=(-0.606015016913)*x[1]
        s[2,2]=(0.637487308345)*x[2]
        s[3,0]=(-0.288536804159)*x[0]
        s[3,1]=(-0.442827019286)*x[1]
        s[3,2]=(-0.0978007221136)*x[2]
        ref=sqrt((1.11094846322))

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromSymbol_rank3(self):
      """
      tests L2-norm of Symbol on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(6, 2, 2),dim=dim)
        s=Data(0,(6, 2, 2),w)
        s[0,0,0]=(0.845216886022)*x[0]
        s[0,0,1]=(0.0597182977987)*x[1]
        s[0,1,0]=(0.146117477376)*x[0]
        s[0,1,1]=(0.918064672934)*x[1]
        s[1,0,0]=(-0.629068350604)*x[0]
        s[1,0,1]=(-0.116285587839)*x[1]
        s[1,1,0]=(0.678041887149)*x[0]
        s[1,1,1]=(-0.756045170867)*x[1]
        s[2,0,0]=(0.66884645591)*x[0]
        s[2,0,1]=(0.527297857958)*x[1]
        s[2,1,0]=(-0.453462310421)*x[0]
        s[2,1,1]=(-0.801421668573)*x[1]
        s[3,0,0]=(-0.418575000488)*x[0]
        s[3,0,1]=(0.977995539022)*x[1]
        s[3,1,0]=(0.863549457649)*x[0]
        s[3,1,1]=(0.975604004421)*x[1]
        s[4,0,0]=(0.642351898331)*x[0]
        s[4,0,1]=(-0.63624912796)*x[1]
        s[4,1,0]=(0.561260478169)*x[0]
        s[4,1,1]=(-0.942077895874)*x[1]
        s[5,0,0]=(0.0142173692077)*x[0]
        s[5,0,1]=(0.106312727314)*x[1]
        s[5,1,0]=(0.110676503653)*x[0]
        s[5,1,1]=(-0.534921324957)*x[1]
        ref=sqrt((3.25169916884))

      else:
        arg=Symbol(shape=(6, 2, 3),dim=dim)
        s=Data(0,(6, 2, 3),w)
        s[0,0,0]=(-0.351762958436)*x[0]
        s[0,0,1]=(0.797778461739)*x[1]
        s[0,0,2]=(-0.216759353067)*x[2]
        s[0,1,0]=(0.913252336442)*x[0]
        s[0,1,1]=(0.988076789445)*x[1]
        s[0,1,2]=(-0.901479800398)*x[2]
        s[1,0,0]=(0.179074582506)*x[0]
        s[1,0,1]=(-0.0198494753788)*x[1]
        s[1,0,2]=(0.453475519845)*x[2]
        s[1,1,0]=(0.0703079255814)*x[0]
        s[1,1,1]=(0.132531352932)*x[1]
        s[1,1,2]=(0.746319860866)*x[2]
        s[2,0,0]=(0.414228087724)*x[0]
        s[2,0,1]=(0.466702867918)*x[1]
        s[2,0,2]=(0.0868129987291)*x[2]
        s[2,1,0]=(0.903287212799)*x[0]
        s[2,1,1]=(0.912709008679)*x[1]
        s[2,1,2]=(-0.0365579422157)*x[2]
        s[3,0,0]=(0.0164686487169)*x[0]
        s[3,0,1]=(-0.914613841493)*x[1]
        s[3,0,2]=(-0.0822005463079)*x[2]
        s[3,1,0]=(-0.315907377871)*x[0]
        s[3,1,1]=(-0.102105289546)*x[1]
        s[3,1,2]=(-0.0353732873141)*x[2]
        s[4,0,0]=(0.0323628908944)*x[0]
        s[4,0,1]=(-0.954360115684)*x[1]
        s[4,0,2]=(0.923756483241)*x[2]
        s[4,1,0]=(0.980063248251)*x[0]
        s[4,1,1]=(0.0664804433015)*x[1]
        s[4,1,2]=(-0.765454557157)*x[2]
        s[5,0,0]=(-0.806747348451)*x[0]
        s[5,0,1]=(-0.423078051628)*x[1]
        s[5,0,2]=(0.765544317002)*x[2]
        s[5,1,0]=(0.00826992946354)*x[0]
        s[5,1,1]=(-0.117741801904)*x[1]
        s[5,1,2]=(-0.975554426921)*x[2]
        ref=sqrt((4.31586522476))

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunction_fromSymbol_rank4(self):
      """
      tests L2-norm of Symbol on the Function

      assumptions: self.domain supports integration on Function
      """
      dim=self.domain.getDim()
      w=Function(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
        s=Data(0,(4, 5, 3, 2),w)
        s[0,0,0,0]=(0.392249504065)*x[0]
        s[0,0,0,1]=(-0.0829499082807)*x[1]
        s[0,0,1,0]=(0.075391736082)*x[0]
        s[0,0,1,1]=(-0.267274680434)*x[1]
        s[0,0,2,0]=(0.758405394297)*x[0]
        s[0,0,2,1]=(0.966859664545)*x[1]
        s[0,1,0,0]=(-0.665158005166)*x[0]
        s[0,1,0,1]=(-0.861735224083)*x[1]
        s[0,1,1,0]=(0.889852595371)*x[0]
        s[0,1,1,1]=(-0.456217187733)*x[1]
        s[0,1,2,0]=(-0.444839601276)*x[0]
        s[0,1,2,1]=(-0.646347424003)*x[1]
        s[0,2,0,0]=(0.016728267633)*x[0]
        s[0,2,0,1]=(0.598805251354)*x[1]
        s[0,2,1,0]=(-0.0850259337934)*x[0]
        s[0,2,1,1]=(-0.958442577173)*x[1]
        s[0,2,2,0]=(-0.316155802878)*x[0]
        s[0,2,2,1]=(0.882575663199)*x[1]
        s[0,3,0,0]=(-0.350482054568)*x[0]
        s[0,3,0,1]=(-0.0018770186166)*x[1]
        s[0,3,1,0]=(0.434158350351)*x[0]
        s[0,3,1,1]=(0.510916541891)*x[1]
        s[0,3,2,0]=(0.415800847408)*x[0]
        s[0,3,2,1]=(0.0045200661478)*x[1]
        s[0,4,0,0]=(0.719981048541)*x[0]
        s[0,4,0,1]=(-0.394991607605)*x[1]
        s[0,4,1,0]=(0.225637020507)*x[0]
        s[0,4,1,1]=(-0.942644070947)*x[1]
        s[0,4,2,0]=(0.914486047699)*x[0]
        s[0,4,2,1]=(-0.0605237231206)*x[1]
        s[1,0,0,0]=(0.49162536785)*x[0]
        s[1,0,0,1]=(-0.616492901013)*x[1]
        s[1,0,1,0]=(-0.1288077626)*x[0]
        s[1,0,1,1]=(-0.345407781206)*x[1]
        s[1,0,2,0]=(-0.0725937970644)*x[0]
        s[1,0,2,1]=(0.40531578317)*x[1]
        s[1,1,0,0]=(0.598712741491)*x[0]
        s[1,1,0,1]=(-0.93879902651)*x[1]
        s[1,1,1,0]=(-0.0264475169249)*x[0]
        s[1,1,1,1]=(0.249909294638)*x[1]
        s[1,1,2,0]=(0.0027245979433)*x[0]
        s[1,1,2,1]=(-0.373044612275)*x[1]
        s[1,2,0,0]=(-0.281778393525)*x[0]
        s[1,2,0,1]=(0.627994360637)*x[1]
        s[1,2,1,0]=(0.589488628124)*x[0]
        s[1,2,1,1]=(-0.758174695207)*x[1]
        s[1,2,2,0]=(-0.406965936453)*x[0]
        s[1,2,2,1]=(-0.69002957721)*x[1]
        s[1,3,0,0]=(-0.199124201524)*x[0]
        s[1,3,0,1]=(0.869253645966)*x[1]
        s[1,3,1,0]=(-0.452605020068)*x[0]
        s[1,3,1,1]=(-0.970641886846)*x[1]
        s[1,3,2,0]=(0.175601171946)*x[0]
        s[1,3,2,1]=(0.672851105536)*x[1]
        s[1,4,0,0]=(0.58320491243)*x[0]
        s[1,4,0,1]=(-0.259858520694)*x[1]
        s[1,4,1,0]=(-0.808152456276)*x[0]
        s[1,4,1,1]=(-0.061862146349)*x[1]
        s[1,4,2,0]=(-0.285657379415)*x[0]
        s[1,4,2,1]=(0.0400598209814)*x[1]
        s[2,0,0,0]=(-0.982895080214)*x[0]
        s[2,0,0,1]=(-0.221420258961)*x[1]
        s[2,0,1,0]=(0.241664738784)*x[0]
        s[2,0,1,1]=(0.360336298851)*x[1]
        s[2,0,2,0]=(-0.1930995368)*x[0]
        s[2,0,2,1]=(0.165323422951)*x[1]
        s[2,1,0,0]=(-0.580617143448)*x[0]
        s[2,1,0,1]=(-0.445386310829)*x[1]
        s[2,1,1,0]=(-0.679424485212)*x[0]
        s[2,1,1,1]=(0.999216345845)*x[1]
        s[2,1,2,0]=(-0.00488871164474)*x[0]
        s[2,1,2,1]=(0.94248053909)*x[1]
        s[2,2,0,0]=(0.241836822065)*x[0]
        s[2,2,0,1]=(-0.323265890223)*x[1]
        s[2,2,1,0]=(0.282530883606)*x[0]
        s[2,2,1,1]=(0.499481122286)*x[1]
        s[2,2,2,0]=(0.502100738495)*x[0]
        s[2,2,2,1]=(-0.714419126623)*x[1]
        s[2,3,0,0]=(0.427736585056)*x[0]
        s[2,3,0,1]=(0.846933115255)*x[1]
        s[2,3,1,0]=(0.293400967424)*x[0]
        s[2,3,1,1]=(0.258925376632)*x[1]
        s[2,3,2,0]=(-0.748167607341)*x[0]
        s[2,3,2,1]=(0.354090561451)*x[1]
        s[2,4,0,0]=(0.432737336014)*x[0]
        s[2,4,0,1]=(0.856383164324)*x[1]
        s[2,4,1,0]=(-0.0596776788823)*x[0]
        s[2,4,1,1]=(-0.299291531668)*x[1]
        s[2,4,2,0]=(0.0332610005043)*x[0]
        s[2,4,2,1]=(-0.784223449976)*x[1]
        s[3,0,0,0]=(0.856346607718)*x[0]
        s[3,0,0,1]=(-0.711038642485)*x[1]
        s[3,0,1,0]=(-0.722899969631)*x[0]
        s[3,0,1,1]=(-0.976358349372)*x[1]
        s[3,0,2,0]=(-0.566657009276)*x[0]
        s[3,0,2,1]=(-0.88977921418)*x[1]
        s[3,1,0,0]=(0.690995364783)*x[0]
        s[3,1,0,1]=(0.0178260768232)*x[1]
        s[3,1,1,0]=(0.210672134956)*x[0]
        s[3,1,1,1]=(-0.271769743548)*x[1]
        s[3,1,2,0]=(-0.119905539717)*x[0]
        s[3,1,2,1]=(0.408792239219)*x[1]
        s[3,2,0,0]=(0.134901834799)*x[0]
        s[3,2,0,1]=(0.0969235330927)*x[1]
        s[3,2,1,0]=(-0.418267686354)*x[0]
        s[3,2,1,1]=(-0.599174142878)*x[1]
        s[3,2,2,0]=(-0.409148104875)*x[0]
        s[3,2,2,1]=(0.504760142275)*x[1]
        s[3,3,0,0]=(-0.174999245323)*x[0]
        s[3,3,0,1]=(-0.994094684291)*x[1]
        s[3,3,1,0]=(0.470460754362)*x[0]
        s[3,3,1,1]=(-0.867881666651)*x[1]
        s[3,3,2,0]=(0.691171063635)*x[0]
        s[3,3,2,1]=(-0.644952618882)*x[1]
        s[3,4,0,0]=(0.883581127272)*x[0]
        s[3,4,0,1]=(-0.534009409512)*x[1]
        s[3,4,1,0]=(0.324806602648)*x[0]
        s[3,4,1,1]=(0.606235157996)*x[1]
        s[3,4,2,0]=(-0.0136704045984)*x[0]
        s[3,4,2,1]=(0.36671612245)*x[1]
        ref=sqrt((12.2721641113))

      else:
        arg=Symbol(shape=(4, 5, 3, 3),dim=dim)
        s=Data(0,(4, 5, 3, 3),w)
        s[0,0,0,0]=(-0.0241692471764)*x[0]
        s[0,0,0,1]=(0.816194825296)*x[1]
        s[0,0,0,2]=(-0.0193776462565)*x[2]
        s[0,0,1,0]=(0.885158780616)*x[0]
        s[0,0,1,1]=(0.750445408795)*x[1]
        s[0,0,1,2]=(-0.242904502332)*x[2]
        s[0,0,2,0]=(-0.101306101076)*x[0]
        s[0,0,2,1]=(0.254549514037)*x[1]
        s[0,0,2,2]=(-0.301743139759)*x[2]
        s[0,1,0,0]=(0.960790978081)*x[0]
        s[0,1,0,1]=(-0.873861925572)*x[1]
        s[0,1,0,2]=(-0.266926259205)*x[2]
        s[0,1,1,0]=(-0.533605756527)*x[0]
        s[0,1,1,1]=(-0.196348553154)*x[1]
        s[0,1,1,2]=(0.462438110297)*x[2]
        s[0,1,2,0]=(0.766114991742)*x[0]
        s[0,1,2,1]=(-0.282544763069)*x[1]
        s[0,1,2,2]=(-0.918562937205)*x[2]
        s[0,2,0,0]=(-0.723736937045)*x[0]
        s[0,2,0,1]=(-0.0271772220034)*x[1]
        s[0,2,0,2]=(-0.67448574094)*x[2]
        s[0,2,1,0]=(-0.0465782041976)*x[0]
        s[0,2,1,1]=(-0.701475255709)*x[1]
        s[0,2,1,2]=(-0.54745824908)*x[2]
        s[0,2,2,0]=(0.823427037717)*x[0]
        s[0,2,2,1]=(0.873366807256)*x[1]
        s[0,2,2,2]=(0.902453041345)*x[2]
        s[0,3,0,0]=(-0.383388300734)*x[0]
        s[0,3,0,1]=(-0.457344890379)*x[1]
        s[0,3,0,2]=(0.449736739028)*x[2]
        s[0,3,1,0]=(-0.994113075606)*x[0]
        s[0,3,1,1]=(0.0265853285952)*x[1]
        s[0,3,1,2]=(-0.738089523761)*x[2]
        s[0,3,2,0]=(0.663756875286)*x[0]
        s[0,3,2,1]=(0.597615440207)*x[1]
        s[0,3,2,2]=(0.246105261445)*x[2]
        s[0,4,0,0]=(0.397226701305)*x[0]
        s[0,4,0,1]=(0.0967432951001)*x[1]
        s[0,4,0,2]=(0.619665795097)*x[2]
        s[0,4,1,0]=(0.155541932506)*x[0]
        s[0,4,1,1]=(-0.114961356145)*x[1]
        s[0,4,1,2]=(-0.491668824315)*x[2]
        s[0,4,2,0]=(-0.726549525506)*x[0]
        s[0,4,2,1]=(-0.876194753259)*x[1]
        s[0,4,2,2]=(-0.506364111318)*x[2]
        s[1,0,0,0]=(-0.380662783497)*x[0]
        s[1,0,0,1]=(-0.57927422387)*x[1]
        s[1,0,0,2]=(-0.282039117242)*x[2]
        s[1,0,1,0]=(-0.873334149601)*x[0]
        s[1,0,1,1]=(-0.0145182664909)*x[1]
        s[1,0,1,2]=(0.903533313452)*x[2]
        s[1,0,2,0]=(0.123617284156)*x[0]
        s[1,0,2,1]=(0.716639379262)*x[1]
        s[1,0,2,2]=(-0.940470701414)*x[2]
        s[1,1,0,0]=(-0.831075410065)*x[0]
        s[1,1,0,1]=(0.808724910022)*x[1]
        s[1,1,0,2]=(0.218369441977)*x[2]
        s[1,1,1,0]=(0.690977533444)*x[0]
        s[1,1,1,1]=(-0.379327957709)*x[1]
        s[1,1,1,2]=(0.0964273177138)*x[2]
        s[1,1,2,0]=(-0.244790745268)*x[0]
        s[1,1,2,1]=(-0.208570215865)*x[1]
        s[1,1,2,2]=(0.280320965835)*x[2]
        s[1,2,0,0]=(-0.571572336783)*x[0]
        s[1,2,0,1]=(-0.512302129833)*x[1]
        s[1,2,0,2]=(0.458134267946)*x[2]
        s[1,2,1,0]=(-0.909450574449)*x[0]
        s[1,2,1,1]=(-0.54240194702)*x[1]
        s[1,2,1,2]=(0.015279542779)*x[2]
        s[1,2,2,0]=(-0.450829963713)*x[0]
        s[1,2,2,1]=(0.265958897553)*x[1]
        s[1,2,2,2]=(0.0115876819282)*x[2]
        s[1,3,0,0]=(-0.493820517172)*x[0]
        s[1,3,0,1]=(0.924501975039)*x[1]
        s[1,3,0,2]=(0.976144734904)*x[2]
        s[1,3,1,0]=(-0.0369162848651)*x[0]
        s[1,3,1,1]=(-0.282951956468)*x[1]
        s[1,3,1,2]=(0.46799488403)*x[2]
        s[1,3,2,0]=(0.859289726723)*x[0]
        s[1,3,2,1]=(0.460555626035)*x[1]
        s[1,3,2,2]=(0.515796525158)*x[2]
        s[1,4,0,0]=(0.707364903535)*x[0]
        s[1,4,0,1]=(0.749724556391)*x[1]
        s[1,4,0,2]=(-0.278863489815)*x[2]
        s[1,4,1,0]=(0.972950918189)*x[0]
        s[1,4,1,1]=(0.844177970875)*x[1]
        s[1,4,1,2]=(0.0763619420058)*x[2]
        s[1,4,2,0]=(-0.0852939975129)*x[0]
        s[1,4,2,1]=(0.583399144502)*x[1]
        s[1,4,2,2]=(0.848661907972)*x[2]
        s[2,0,0,0]=(-0.777408601463)*x[0]
        s[2,0,0,1]=(-0.855232334446)*x[1]
        s[2,0,0,2]=(-0.742431398498)*x[2]
        s[2,0,1,0]=(-0.96680966892)*x[0]
        s[2,0,1,1]=(-0.33301062934)*x[1]
        s[2,0,1,2]=(-0.675842472611)*x[2]
        s[2,0,2,0]=(-0.0163730639481)*x[0]
        s[2,0,2,1]=(-0.0854274064817)*x[1]
        s[2,0,2,2]=(-0.93094914338)*x[2]
        s[2,1,0,0]=(0.881651879025)*x[0]
        s[2,1,0,1]=(0.782286554309)*x[1]
        s[2,1,0,2]=(0.71813995101)*x[2]
        s[2,1,1,0]=(-0.0790321088042)*x[0]
        s[2,1,1,1]=(-0.181510921696)*x[1]
        s[2,1,1,2]=(0.822016852235)*x[2]
        s[2,1,2,0]=(0.634313542034)*x[0]
        s[2,1,2,1]=(0.797139044679)*x[1]
        s[2,1,2,2]=(0.400542618196)*x[2]
        s[2,2,0,0]=(-0.823507870125)*x[0]
        s[2,2,0,1]=(0.981256538471)*x[1]
        s[2,2,0,2]=(0.158779409645)*x[2]
        s[2,2,1,0]=(-0.0788962261172)*x[0]
        s[2,2,1,1]=(-0.295903805518)*x[1]
        s[2,2,1,2]=(0.338095367799)*x[2]
        s[2,2,2,0]=(0.259914616014)*x[0]
        s[2,2,2,1]=(-0.935338026316)*x[1]
        s[2,2,2,2]=(0.752104609965)*x[2]
        s[2,3,0,0]=(0.770409830614)*x[0]
        s[2,3,0,1]=(0.827165793631)*x[1]
        s[2,3,0,2]=(-0.570252686313)*x[2]
        s[2,3,1,0]=(-0.645381957261)*x[0]
        s[2,3,1,1]=(-0.547153600936)*x[1]
        s[2,3,1,2]=(-0.424301255985)*x[2]
        s[2,3,2,0]=(0.938907909024)*x[0]
        s[2,3,2,1]=(-0.165860294169)*x[1]
        s[2,3,2,2]=(-0.507668918028)*x[2]
        s[2,4,0,0]=(0.504198965293)*x[0]
        s[2,4,0,1]=(-0.668203518046)*x[1]
        s[2,4,0,2]=(0.229406581915)*x[2]
        s[2,4,1,0]=(0.454298324734)*x[0]
        s[2,4,1,1]=(0.631911737372)*x[1]
        s[2,4,1,2]=(-0.849242047895)*x[2]
        s[2,4,2,0]=(0.0187819832254)*x[0]
        s[2,4,2,1]=(-0.0257470291285)*x[1]
        s[2,4,2,2]=(-0.968820620099)*x[2]
        s[3,0,0,0]=(0.55579863211)*x[0]
        s[3,0,0,1]=(0.830363451819)*x[1]
        s[3,0,0,2]=(0.37501906831)*x[2]
        s[3,0,1,0]=(-0.609478442879)*x[0]
        s[3,0,1,1]=(0.895531311344)*x[1]
        s[3,0,1,2]=(0.457055008762)*x[2]
        s[3,0,2,0]=(-0.589517607648)*x[0]
        s[3,0,2,1]=(-0.840365410994)*x[1]
        s[3,0,2,2]=(-0.804004197791)*x[2]
        s[3,1,0,0]=(0.231479229381)*x[0]
        s[3,1,0,1]=(0.851135131086)*x[1]
        s[3,1,0,2]=(-0.558112154058)*x[2]
        s[3,1,1,0]=(-0.365328057806)*x[0]
        s[3,1,1,1]=(0.665370944345)*x[1]
        s[3,1,1,2]=(0.815245254888)*x[2]
        s[3,1,2,0]=(-0.622675567594)*x[0]
        s[3,1,2,1]=(0.15478575296)*x[1]
        s[3,1,2,2]=(-0.919833514628)*x[2]
        s[3,2,0,0]=(0.893976278733)*x[0]
        s[3,2,0,1]=(-0.881412936262)*x[1]
        s[3,2,0,2]=(-0.940014543462)*x[2]
        s[3,2,1,0]=(0.310442650776)*x[0]
        s[3,2,1,1]=(0.895222325851)*x[1]
        s[3,2,1,2]=(0.308995631858)*x[2]
        s[3,2,2,0]=(0.0373558936063)*x[0]
        s[3,2,2,1]=(-0.0592368839463)*x[1]
        s[3,2,2,2]=(0.793681649539)*x[2]
        s[3,3,0,0]=(-0.542535013221)*x[0]
        s[3,3,0,1]=(0.730645327078)*x[1]
        s[3,3,0,2]=(-0.776045858279)*x[2]
        s[3,3,1,0]=(-0.657528555613)*x[0]
        s[3,3,1,1]=(-0.282142368825)*x[1]
        s[3,3,1,2]=(-0.692363751497)*x[2]
        s[3,3,2,0]=(-0.168946040457)*x[0]
        s[3,3,2,1]=(-0.318493817076)*x[1]
        s[3,3,2,2]=(0.535922920691)*x[2]
        s[3,4,0,0]=(0.414492782118)*x[0]
        s[3,4,0,1]=(0.295207630842)*x[1]
        s[3,4,0,2]=(0.453596918765)*x[2]
        s[3,4,1,0]=(-0.083655892279)*x[0]
        s[3,4,1,1]=(-0.962961908465)*x[1]
        s[3,4,1,2]=(0.138455397004)*x[2]
        s[3,4,2,0]=(-0.383498097386)*x[0]
        s[3,4,2,1]=(0.650869920691)*x[1]
        s[3,4,2,2]=(-0.998836301837)*x[2]
        ref=sqrt((22.2241672931))

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank0(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=1.*sqrt(x[0])
        ref=sqrt(0.5)

      else:
        arg=Data(0,(),w)
        arg=1.*sqrt(x[0])
        ref=sqrt(0.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank1(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=1.*sqrt(x[0])
        arg[1]=2.*sqrt(x[1])
        ref=sqrt(2.5)

      else:
        arg=Data(0,(3,),w)
        arg[0]=1.*sqrt(x[0])
        arg[1]=2.*sqrt(x[1])
        arg[2]=3.*sqrt(x[2])
        ref=sqrt(7.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank2(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=11.*sqrt(x[0])
        arg[0,1]=1.*sqrt(x[1])
        arg[1,0]=10.*sqrt(x[0])
        arg[1,1]=11.*sqrt(x[1])
        arg[2,0]=20.*sqrt(x[0])
        arg[2,1]=21.*sqrt(x[1])
        arg[3,0]=30.*sqrt(x[0])
        arg[3,1]=31.*sqrt(x[1])
        ref=sqrt(1522.5)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=11.*sqrt(x[0])
        arg[0,1]=1.*sqrt(x[1])
        arg[0,2]=2.*sqrt(x[2])
        arg[1,0]=10.*sqrt(x[0])
        arg[1,1]=11.*sqrt(x[1])
        arg[1,2]=12.*sqrt(x[2])
        arg[2,0]=20.*sqrt(x[0])
        arg[2,1]=21.*sqrt(x[1])
        arg[2,2]=22.*sqrt(x[2])
        arg[3,0]=30.*sqrt(x[0])
        arg[3,1]=31.*sqrt(x[1])
        arg[3,2]=32.*sqrt(x[2])
        ref=sqrt(2350.5)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank3(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(0.449174971953)*sqrt(x[0])
        arg[0,0,1]=(-0.0109398763289)*sqrt(x[1])
        arg[0,1,0]=(-0.202497187709)*sqrt(x[0])
        arg[0,1,1]=(-0.12970879334)*sqrt(x[1])
        arg[1,0,0]=(-0.138092481719)*sqrt(x[0])
        arg[1,0,1]=(-0.528752200917)*sqrt(x[1])
        arg[1,1,0]=(-0.605919441662)*sqrt(x[0])
        arg[1,1,1]=(0.215615032334)*sqrt(x[1])
        arg[2,0,0]=(-0.998734541972)*sqrt(x[0])
        arg[2,0,1]=(0.725811901251)*sqrt(x[1])
        arg[2,1,0]=(-0.966536503228)*sqrt(x[0])
        arg[2,1,1]=(-0.528692217355)*sqrt(x[1])
        arg[3,0,0]=(0.757633851466)*sqrt(x[0])
        arg[3,0,1]=(-0.524660157377)*sqrt(x[1])
        arg[3,1,0]=(0.983733431677)*sqrt(x[0])
        arg[3,1,1]=(0.061279109546)*sqrt(x[1])
        arg[4,0,0]=(0.85914215305)*sqrt(x[0])
        arg[4,0,1]=(0.941714045112)*sqrt(x[1])
        arg[4,1,0]=(0.172235529555)*sqrt(x[0])
        arg[4,1,1]=(-0.108381454437)*sqrt(x[1])
        arg[5,0,0]=(-0.736373697727)*sqrt(x[0])
        arg[5,0,1]=(-0.599337929679)*sqrt(x[1])
        arg[5,1,0]=(0.661072686392)*sqrt(x[0])
        arg[5,1,1]=(-0.55107327409)*sqrt(x[1])
        ref=sqrt(4.4196214907099591)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.69227064904)*sqrt(x[0])
        arg[0,0,1]=(-0.968336177418)*sqrt(x[1])
        arg[0,0,2]=(-0.634883146685)*sqrt(x[2])
        arg[0,1,0]=(-0.12640661422)*sqrt(x[0])
        arg[0,1,1]=(-0.637386589888)*sqrt(x[1])
        arg[0,1,2]=(0.26060859356)*sqrt(x[2])
        arg[1,0,0]=(-0.986864633297)*sqrt(x[0])
        arg[1,0,1]=(-0.441589142379)*sqrt(x[1])
        arg[1,0,2]=(-0.587865539582)*sqrt(x[2])
        arg[1,1,0]=(0.596052465031)*sqrt(x[0])
        arg[1,1,1]=(0.312732336652)*sqrt(x[1])
        arg[1,1,2]=(-0.514423945092)*sqrt(x[2])
        arg[2,0,0]=(-0.892391254794)*sqrt(x[0])
        arg[2,0,1]=(0.377920185756)*sqrt(x[1])
        arg[2,0,2]=(-0.120174597181)*sqrt(x[2])
        arg[2,1,0]=(-0.469951576468)*sqrt(x[0])
        arg[2,1,1]=(-0.788362249555)*sqrt(x[1])
        arg[2,1,2]=(0.745625354986)*sqrt(x[2])
        arg[3,0,0]=(0.542802498569)*sqrt(x[0])
        arg[3,0,1]=(-0.814541028706)*sqrt(x[1])
        arg[3,0,2]=(0.298410992196)*sqrt(x[2])
        arg[3,1,0]=(0.981190341206)*sqrt(x[0])
        arg[3,1,1]=(0.666421298608)*sqrt(x[1])
        arg[3,1,2]=(-0.369751722626)*sqrt(x[2])
        arg[4,0,0]=(-0.75379530597)*sqrt(x[0])
        arg[4,0,1]=(0.283357267139)*sqrt(x[1])
        arg[4,0,2]=(0.247787072861)*sqrt(x[2])
        arg[4,1,0]=(0.301766692533)*sqrt(x[0])
        arg[4,1,1]=(0.828183439224)*sqrt(x[1])
        arg[4,1,2]=(-0.580824060547)*sqrt(x[2])
        arg[5,0,0]=(0.637345610764)*sqrt(x[0])
        arg[5,0,1]=(-0.234409115997)*sqrt(x[1])
        arg[5,0,2]=(-0.192639300316)*sqrt(x[2])
        arg[5,1,0]=(-0.62609237162)*sqrt(x[0])
        arg[5,1,1]=(0.463404958552)*sqrt(x[1])
        arg[5,1,2]=(-0.547814448738)*sqrt(x[2])
        ref=sqrt(6.3571697792950923)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromData_rank4(self):
      """
      tests L2-norm of Data on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.232618585183)*sqrt(x[0])
        arg[0,0,0,1]=(0.39796117869)*sqrt(x[1])
        arg[0,0,1,0]=(-0.997336958262)*sqrt(x[0])
        arg[0,0,1,1]=(-0.351780915076)*sqrt(x[1])
        arg[0,0,2,0]=(-0.876764070136)*sqrt(x[0])
        arg[0,0,2,1]=(0.808730805817)*sqrt(x[1])
        arg[0,1,0,0]=(-0.197154744966)*sqrt(x[0])
        arg[0,1,0,1]=(0.416246096086)*sqrt(x[1])
        arg[0,1,1,0]=(0.708038457121)*sqrt(x[0])
        arg[0,1,1,1]=(-0.00954021503188)*sqrt(x[1])
        arg[0,1,2,0]=(-0.62630809425)*sqrt(x[0])
        arg[0,1,2,1]=(0.430228727912)*sqrt(x[1])
        arg[0,2,0,0]=(0.0306704747648)*sqrt(x[0])
        arg[0,2,0,1]=(-0.913877199453)*sqrt(x[1])
        arg[0,2,1,0]=(-0.697612800829)*sqrt(x[0])
        arg[0,2,1,1]=(-0.17996376822)*sqrt(x[1])
        arg[0,2,2,0]=(-0.304509578871)*sqrt(x[0])
        arg[0,2,2,1]=(-0.610556755811)*sqrt(x[1])
        arg[0,3,0,0]=(-0.452355972234)*sqrt(x[0])
        arg[0,3,0,1]=(-0.368921242518)*sqrt(x[1])
        arg[0,3,1,0]=(-0.478275554932)*sqrt(x[0])
        arg[0,3,1,1]=(0.257178549127)*sqrt(x[1])
        arg[0,3,2,0]=(0.530736487177)*sqrt(x[0])
        arg[0,3,2,1]=(-0.567126272463)*sqrt(x[1])
        arg[0,4,0,0]=(0.801519165938)*sqrt(x[0])
        arg[0,4,0,1]=(-0.509816703951)*sqrt(x[1])
        arg[0,4,1,0]=(-0.255412646934)*sqrt(x[0])
        arg[0,4,1,1]=(0.437540101896)*sqrt(x[1])
        arg[0,4,2,0]=(-0.815574969538)*sqrt(x[0])
        arg[0,4,2,1]=(-0.94691547137)*sqrt(x[1])
        arg[1,0,0,0]=(-0.732550722593)*sqrt(x[0])
        arg[1,0,0,1]=(0.515752381704)*sqrt(x[1])
        arg[1,0,1,0]=(-0.343590210899)*sqrt(x[0])
        arg[1,0,1,1]=(-0.0601907964915)*sqrt(x[1])
        arg[1,0,2,0]=(0.0199916154421)*sqrt(x[0])
        arg[1,0,2,1]=(-0.136927227821)*sqrt(x[1])
        arg[1,1,0,0]=(0.397994441702)*sqrt(x[0])
        arg[1,1,0,1]=(0.953873148948)*sqrt(x[1])
        arg[1,1,1,0]=(0.419416235967)*sqrt(x[0])
        arg[1,1,1,1]=(0.700998577193)*sqrt(x[1])
        arg[1,1,2,0]=(-0.497358799271)*sqrt(x[0])
        arg[1,1,2,1]=(0.0851768858379)*sqrt(x[1])
        arg[1,2,0,0]=(0.0936678875202)*sqrt(x[0])
        arg[1,2,0,1]=(0.869883786896)*sqrt(x[1])
        arg[1,2,1,0]=(0.582700123485)*sqrt(x[0])
        arg[1,2,1,1]=(-0.433381106794)*sqrt(x[1])
        arg[1,2,2,0]=(-0.527031777974)*sqrt(x[0])
        arg[1,2,2,1]=(0.105105137652)*sqrt(x[1])
        arg[1,3,0,0]=(-0.716750829134)*sqrt(x[0])
        arg[1,3,0,1]=(0.774519209008)*sqrt(x[1])
        arg[1,3,1,0]=(-0.568743372716)*sqrt(x[0])
        arg[1,3,1,1]=(0.794732483944)*sqrt(x[1])
        arg[1,3,2,0]=(0.246606002015)*sqrt(x[0])
        arg[1,3,2,1]=(-0.988869494994)*sqrt(x[1])
        arg[1,4,0,0]=(0.482379298083)*sqrt(x[0])
        arg[1,4,0,1]=(-0.386268387903)*sqrt(x[1])
        arg[1,4,1,0]=(0.137184889675)*sqrt(x[0])
        arg[1,4,1,1]=(-0.140520035321)*sqrt(x[1])
        arg[1,4,2,0]=(0.822755050415)*sqrt(x[0])
        arg[1,4,2,1]=(-0.815562139522)*sqrt(x[1])
        arg[2,0,0,0]=(-0.462891511962)*sqrt(x[0])
        arg[2,0,0,1]=(-0.122643411631)*sqrt(x[1])
        arg[2,0,1,0]=(-0.520861119962)*sqrt(x[0])
        arg[2,0,1,1]=(-0.881189618018)*sqrt(x[1])
        arg[2,0,2,0]=(-0.776157842774)*sqrt(x[0])
        arg[2,0,2,1]=(-0.12354053207)*sqrt(x[1])
        arg[2,1,0,0]=(0.395495230826)*sqrt(x[0])
        arg[2,1,0,1]=(-0.388106659423)*sqrt(x[1])
        arg[2,1,1,0]=(0.354250242834)*sqrt(x[0])
        arg[2,1,1,1]=(-0.666514210192)*sqrt(x[1])
        arg[2,1,2,0]=(0.951294655083)*sqrt(x[0])
        arg[2,1,2,1]=(0.074024416386)*sqrt(x[1])
        arg[2,2,0,0]=(0.335448485459)*sqrt(x[0])
        arg[2,2,0,1]=(-0.40988282528)*sqrt(x[1])
        arg[2,2,1,0]=(-0.805725968875)*sqrt(x[0])
        arg[2,2,1,1]=(-0.949883082118)*sqrt(x[1])
        arg[2,2,2,0]=(0.531549210683)*sqrt(x[0])
        arg[2,2,2,1]=(-0.398401016682)*sqrt(x[1])
        arg[2,3,0,0]=(-0.953963433205)*sqrt(x[0])
        arg[2,3,0,1]=(0.643431126406)*sqrt(x[1])
        arg[2,3,1,0]=(-0.167611998738)*sqrt(x[0])
        arg[2,3,1,1]=(0.226130056552)*sqrt(x[1])
        arg[2,3,2,0]=(0.0752687641131)*sqrt(x[0])
        arg[2,3,2,1]=(-0.115742756362)*sqrt(x[1])
        arg[2,4,0,0]=(0.579694491028)*sqrt(x[0])
        arg[2,4,0,1]=(-0.112005738299)*sqrt(x[1])
        arg[2,4,1,0]=(0.657291764224)*sqrt(x[0])
        arg[2,4,1,1]=(0.62671154177)*sqrt(x[1])
        arg[2,4,2,0]=(0.103695027944)*sqrt(x[0])
        arg[2,4,2,1]=(0.462828491544)*sqrt(x[1])
        arg[3,0,0,0]=(0.697692979998)*sqrt(x[0])
        arg[3,0,0,1]=(-0.123481859619)*sqrt(x[1])
        arg[3,0,1,0]=(-0.749745629459)*sqrt(x[0])
        arg[3,0,1,1]=(-0.541969524069)*sqrt(x[1])
        arg[3,0,2,0]=(0.819484470759)*sqrt(x[0])
        arg[3,0,2,1]=(-0.860592326469)*sqrt(x[1])
        arg[3,1,0,0]=(-0.716566084771)*sqrt(x[0])
        arg[3,1,0,1]=(-0.949235434827)*sqrt(x[1])
        arg[3,1,1,0]=(-0.826699498174)*sqrt(x[0])
        arg[3,1,1,1]=(-0.138511521583)*sqrt(x[1])
        arg[3,1,2,0]=(-0.951682890904)*sqrt(x[0])
        arg[3,1,2,1]=(0.413293316925)*sqrt(x[1])
        arg[3,2,0,0]=(0.909516836775)*sqrt(x[0])
        arg[3,2,0,1]=(-0.919989721277)*sqrt(x[1])
        arg[3,2,1,0]=(0.0994860369337)*sqrt(x[0])
        arg[3,2,1,1]=(-0.933647246623)*sqrt(x[1])
        arg[3,2,2,0]=(-0.759215183015)*sqrt(x[0])
        arg[3,2,2,1]=(0.0975793309286)*sqrt(x[1])
        arg[3,3,0,0]=(-0.130256739381)*sqrt(x[0])
        arg[3,3,0,1]=(-0.582280862311)*sqrt(x[1])
        arg[3,3,1,0]=(0.206970526192)*sqrt(x[0])
        arg[3,3,1,1]=(-0.8678322258)*sqrt(x[1])
        arg[3,3,2,0]=(0.133004501279)*sqrt(x[0])
        arg[3,3,2,1]=(0.802921710935)*sqrt(x[1])
        arg[3,4,0,0]=(-0.255737792764)*sqrt(x[0])
        arg[3,4,0,1]=(-0.34168114937)*sqrt(x[1])
        arg[3,4,1,0]=(-0.859309090399)*sqrt(x[0])
        arg[3,4,1,1]=(0.245043986435)*sqrt(x[1])
        arg[3,4,2,0]=(0.893062018695)*sqrt(x[0])
        arg[3,4,2,1]=(0.709422742588)*sqrt(x[1])
        ref=sqrt(20.5933920543)
      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.0312828390439)*sqrt(x[0])
        arg[0,0,0,1]=(-0.524970416212)*sqrt(x[1])
        arg[0,0,0,2]=(0.561865217554)*sqrt(x[2])
        arg[0,0,1,0]=(0.692457187384)*sqrt(x[0])
        arg[0,0,1,1]=(0.946967182157)*sqrt(x[1])
        arg[0,0,1,2]=(-0.863842279464)*sqrt(x[2])
        arg[0,0,2,0]=(0.993922921598)*sqrt(x[0])
        arg[0,0,2,1]=(0.322812768679)*sqrt(x[1])
        arg[0,0,2,2]=(0.901876132204)*sqrt(x[2])
        arg[0,1,0,0]=(0.967569979365)*sqrt(x[0])
        arg[0,1,0,1]=(0.840979131355)*sqrt(x[1])
        arg[0,1,0,2]=(0.0494811460856)*sqrt(x[2])
        arg[0,1,1,0]=(0.315178456102)*sqrt(x[0])
        arg[0,1,1,1]=(0.449848313024)*sqrt(x[1])
        arg[0,1,1,2]=(0.765887852886)*sqrt(x[2])
        arg[0,1,2,0]=(0.975541574352)*sqrt(x[0])
        arg[0,1,2,1]=(-0.797851290751)*sqrt(x[1])
        arg[0,1,2,2]=(0.628918775319)*sqrt(x[2])
        arg[0,2,0,0]=(0.685635794312)*sqrt(x[0])
        arg[0,2,0,1]=(0.10341799962)*sqrt(x[1])
        arg[0,2,0,2]=(-0.964822756043)*sqrt(x[2])
        arg[0,2,1,0]=(-0.56160368212)*sqrt(x[0])
        arg[0,2,1,1]=(0.676344298102)*sqrt(x[1])
        arg[0,2,1,2]=(-0.713924121843)*sqrt(x[2])
        arg[0,2,2,0]=(-0.276655136263)*sqrt(x[0])
        arg[0,2,2,1]=(0.336046973788)*sqrt(x[1])
        arg[0,2,2,2]=(-0.68789392396)*sqrt(x[2])
        arg[0,3,0,0]=(0.0172861311571)*sqrt(x[0])
        arg[0,3,0,1]=(-0.301075956456)*sqrt(x[1])
        arg[0,3,0,2]=(0.779442985415)*sqrt(x[2])
        arg[0,3,1,0]=(-0.517629576558)*sqrt(x[0])
        arg[0,3,1,1]=(0.584779586639)*sqrt(x[1])
        arg[0,3,1,2]=(-0.53266435436)*sqrt(x[2])
        arg[0,3,2,0]=(0.841533567102)*sqrt(x[0])
        arg[0,3,2,1]=(0.0458746415489)*sqrt(x[1])
        arg[0,3,2,2]=(0.921237870758)*sqrt(x[2])
        arg[0,4,0,0]=(0.0548343238805)*sqrt(x[0])
        arg[0,4,0,1]=(0.687022707412)*sqrt(x[1])
        arg[0,4,0,2]=(-0.319803609795)*sqrt(x[2])
        arg[0,4,1,0]=(0.409763007811)*sqrt(x[0])
        arg[0,4,1,1]=(0.165501957435)*sqrt(x[1])
        arg[0,4,1,2]=(0.116001692781)*sqrt(x[2])
        arg[0,4,2,0]=(-0.515571394238)*sqrt(x[0])
        arg[0,4,2,1]=(0.209467945147)*sqrt(x[1])
        arg[0,4,2,2]=(-0.344827191247)*sqrt(x[2])
        arg[1,0,0,0]=(0.57193838014)*sqrt(x[0])
        arg[1,0,0,1]=(-0.0880683799076)*sqrt(x[1])
        arg[1,0,0,2]=(0.956899617441)*sqrt(x[2])
        arg[1,0,1,0]=(-0.783689636357)*sqrt(x[0])
        arg[1,0,1,1]=(-0.25177506885)*sqrt(x[1])
        arg[1,0,1,2]=(-0.97074584634)*sqrt(x[2])
        arg[1,0,2,0]=(0.432543519806)*sqrt(x[0])
        arg[1,0,2,1]=(0.481003021954)*sqrt(x[1])
        arg[1,0,2,2]=(-0.0630751518268)*sqrt(x[2])
        arg[1,1,0,0]=(-0.65152446796)*sqrt(x[0])
        arg[1,1,0,1]=(-0.0323685084425)*sqrt(x[1])
        arg[1,1,0,2]=(-0.508674033909)*sqrt(x[2])
        arg[1,1,1,0]=(-0.533367818916)*sqrt(x[0])
        arg[1,1,1,1]=(0.310738340288)*sqrt(x[1])
        arg[1,1,1,2]=(0.694612234326)*sqrt(x[2])
        arg[1,1,2,0]=(-0.622052473032)*sqrt(x[0])
        arg[1,1,2,1]=(0.0498443793671)*sqrt(x[1])
        arg[1,1,2,2]=(0.61023707512)*sqrt(x[2])
        arg[1,2,0,0]=(0.0730267406859)*sqrt(x[0])
        arg[1,2,0,1]=(0.146909334607)*sqrt(x[1])
        arg[1,2,0,2]=(-0.641860284448)*sqrt(x[2])
        arg[1,2,1,0]=(0.917976589737)*sqrt(x[0])
        arg[1,2,1,1]=(0.50219672122)*sqrt(x[1])
        arg[1,2,1,2]=(0.634559579812)*sqrt(x[2])
        arg[1,2,2,0]=(0.0578772734534)*sqrt(x[0])
        arg[1,2,2,1]=(0.288730973517)*sqrt(x[1])
        arg[1,2,2,2]=(-0.0525978796154)*sqrt(x[2])
        arg[1,3,0,0]=(-0.926152433388)*sqrt(x[0])
        arg[1,3,0,1]=(0.0616647680855)*sqrt(x[1])
        arg[1,3,0,2]=(-0.875889217846)*sqrt(x[2])
        arg[1,3,1,0]=(-0.638931542845)*sqrt(x[0])
        arg[1,3,1,1]=(0.708848122964)*sqrt(x[1])
        arg[1,3,1,2]=(0.119066979792)*sqrt(x[2])
        arg[1,3,2,0]=(0.853716218591)*sqrt(x[0])
        arg[1,3,2,1]=(-0.92754322201)*sqrt(x[1])
        arg[1,3,2,2]=(-0.671530626265)*sqrt(x[2])
        arg[1,4,0,0]=(0.337424536231)*sqrt(x[0])
        arg[1,4,0,1]=(0.335704451719)*sqrt(x[1])
        arg[1,4,0,2]=(-0.484565969466)*sqrt(x[2])
        arg[1,4,1,0]=(-0.855476192012)*sqrt(x[0])
        arg[1,4,1,1]=(0.405674615553)*sqrt(x[1])
        arg[1,4,1,2]=(0.728310771323)*sqrt(x[2])
        arg[1,4,2,0]=(0.363651308265)*sqrt(x[0])
        arg[1,4,2,1]=(0.174460594531)*sqrt(x[1])
        arg[1,4,2,2]=(-0.0418244838617)*sqrt(x[2])
        arg[2,0,0,0]=(-0.531341992511)*sqrt(x[0])
        arg[2,0,0,1]=(0.584996796272)*sqrt(x[1])
        arg[2,0,0,2]=(-0.752430968716)*sqrt(x[2])
        arg[2,0,1,0]=(-0.341989849747)*sqrt(x[0])
        arg[2,0,1,1]=(0.153572646953)*sqrt(x[1])
        arg[2,0,1,2]=(-0.197130051737)*sqrt(x[2])
        arg[2,0,2,0]=(-0.338082424082)*sqrt(x[0])
        arg[2,0,2,1]=(0.000173657394772)*sqrt(x[1])
        arg[2,0,2,2]=(0.365272907692)*sqrt(x[2])
        arg[2,1,0,0]=(0.904304126564)*sqrt(x[0])
        arg[2,1,0,1]=(0.161252368484)*sqrt(x[1])
        arg[2,1,0,2]=(0.246854092422)*sqrt(x[2])
        arg[2,1,1,0]=(-0.299880647529)*sqrt(x[0])
        arg[2,1,1,1]=(-0.566917528608)*sqrt(x[1])
        arg[2,1,1,2]=(0.243183337285)*sqrt(x[2])
        arg[2,1,2,0]=(0.437406011474)*sqrt(x[0])
        arg[2,1,2,1]=(0.727447394053)*sqrt(x[1])
        arg[2,1,2,2]=(0.380752950664)*sqrt(x[2])
        arg[2,2,0,0]=(0.172292846911)*sqrt(x[0])
        arg[2,2,0,1]=(0.334201791643)*sqrt(x[1])
        arg[2,2,0,2]=(0.739989926962)*sqrt(x[2])
        arg[2,2,1,0]=(-0.0669843715042)*sqrt(x[0])
        arg[2,2,1,1]=(-0.540497281635)*sqrt(x[1])
        arg[2,2,1,2]=(-0.744217027088)*sqrt(x[2])
        arg[2,2,2,0]=(-0.287295952259)*sqrt(x[0])
        arg[2,2,2,1]=(-0.512411849183)*sqrt(x[1])
        arg[2,2,2,2]=(0.953107417666)*sqrt(x[2])
        arg[2,3,0,0]=(0.998168116695)*sqrt(x[0])
        arg[2,3,0,1]=(0.960065646359)*sqrt(x[1])
        arg[2,3,0,2]=(0.110048258832)*sqrt(x[2])
        arg[2,3,1,0]=(-0.477271134724)*sqrt(x[0])
        arg[2,3,1,1]=(0.707182612251)*sqrt(x[1])
        arg[2,3,1,2]=(0.285500891755)*sqrt(x[2])
        arg[2,3,2,0]=(-0.863497506661)*sqrt(x[0])
        arg[2,3,2,1]=(-0.293917669879)*sqrt(x[1])
        arg[2,3,2,2]=(-0.403384244295)*sqrt(x[2])
        arg[2,4,0,0]=(0.848455277702)*sqrt(x[0])
        arg[2,4,0,1]=(-0.530101455578)*sqrt(x[1])
        arg[2,4,0,2]=(0.33887313048)*sqrt(x[2])
        arg[2,4,1,0]=(-0.195313538124)*sqrt(x[0])
        arg[2,4,1,1]=(-0.62754572008)*sqrt(x[1])
        arg[2,4,1,2]=(-0.385132960582)*sqrt(x[2])
        arg[2,4,2,0]=(0.240048012886)*sqrt(x[0])
        arg[2,4,2,1]=(0.900766252969)*sqrt(x[1])
        arg[2,4,2,2]=(0.669620533505)*sqrt(x[2])
        arg[3,0,0,0]=(0.375766827301)*sqrt(x[0])
        arg[3,0,0,1]=(0.705484960308)*sqrt(x[1])
        arg[3,0,0,2]=(0.440931516034)*sqrt(x[2])
        arg[3,0,1,0]=(-0.44724403177)*sqrt(x[0])
        arg[3,0,1,1]=(-0.31558249626)*sqrt(x[1])
        arg[3,0,1,2]=(-0.00419436365172)*sqrt(x[2])
        arg[3,0,2,0]=(0.750599752032)*sqrt(x[0])
        arg[3,0,2,1]=(0.367649951795)*sqrt(x[1])
        arg[3,0,2,2]=(0.0488013073654)*sqrt(x[2])
        arg[3,1,0,0]=(-0.992890068274)*sqrt(x[0])
        arg[3,1,0,1]=(0.671447745511)*sqrt(x[1])
        arg[3,1,0,2]=(0.85613331404)*sqrt(x[2])
        arg[3,1,1,0]=(-0.46064764242)*sqrt(x[0])
        arg[3,1,1,1]=(0.48138877715)*sqrt(x[1])
        arg[3,1,1,2]=(0.396741761803)*sqrt(x[2])
        arg[3,1,2,0]=(-0.879391967543)*sqrt(x[0])
        arg[3,1,2,1]=(-0.44039462138)*sqrt(x[1])
        arg[3,1,2,2]=(0.0330511573872)*sqrt(x[2])
        arg[3,2,0,0]=(-0.367413701648)*sqrt(x[0])
        arg[3,2,0,1]=(0.0359818324891)*sqrt(x[1])
        arg[3,2,0,2]=(-0.307532667032)*sqrt(x[2])
        arg[3,2,1,0]=(0.334663597166)*sqrt(x[0])
        arg[3,2,1,1]=(0.541941978066)*sqrt(x[1])
        arg[3,2,1,2]=(-0.609184079318)*sqrt(x[2])
        arg[3,2,2,0]=(0.359349239826)*sqrt(x[0])
        arg[3,2,2,1]=(0.0419272305685)*sqrt(x[1])
        arg[3,2,2,2]=(0.557189794296)*sqrt(x[2])
        arg[3,3,0,0]=(-0.85864165554)*sqrt(x[0])
        arg[3,3,0,1]=(-0.185411404213)*sqrt(x[1])
        arg[3,3,0,2]=(0.254294865253)*sqrt(x[2])
        arg[3,3,1,0]=(0.870362177541)*sqrt(x[0])
        arg[3,3,1,1]=(-0.439688612864)*sqrt(x[1])
        arg[3,3,1,2]=(0.26006729357)*sqrt(x[2])
        arg[3,3,2,0]=(-0.0724034754175)*sqrt(x[0])
        arg[3,3,2,1]=(0.444871564246)*sqrt(x[1])
        arg[3,3,2,2]=(0.485634530531)*sqrt(x[2])
        arg[3,4,0,0]=(-0.744756961758)*sqrt(x[0])
        arg[3,4,0,1]=(0.429761406102)*sqrt(x[1])
        arg[3,4,0,2]=(-0.584963735834)*sqrt(x[2])
        arg[3,4,1,0]=(0.684578379159)*sqrt(x[0])
        arg[3,4,1,1]=(0.949460132601)*sqrt(x[1])
        arg[3,4,1,2]=(-0.592179909559)*sqrt(x[2])
        arg[3,4,2,0]=(0.707154437797)*sqrt(x[0])
        arg[3,4,2,1]=(0.619200407063)*sqrt(x[1])
        arg[3,4,2,2]=(-0.338547165)*sqrt(x[2])
        ref=sqrt(28.8255957718)
      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromSymbol_rank0(self):
      """
      tests L2-norm of Symbol on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(-0.292546362165)*sqrt(x[0])
        ref=sqrt(0.042791687007987675)

      else:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(0.751025585654)*sqrt(x[0])
        ref=sqrt(0.28201971515346685)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromSymbol_rank1(self):
      """
      tests L2-norm of Symbol on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(2,),dim=dim)
        s=Data(0,(2,),w)
        s[0]=(-0.1759467192)*sqrt(x[0])
        s[1]=(0.37884376337)*sqrt(x[1])
        ref=sqrt(0.0872399225208)

      else:
        arg=Symbol(shape=(3,),dim=dim)
        s=Data(0,(3,),w)
        s[0]=(-0.879259245697)*sqrt(x[0])
        s[1]=(0.528911893259)*sqrt(x[1])
        s[2]=(-0.751304571514)*sqrt(x[2])
        ref=sqrt(0.808651585576)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromSymbol_rank2(self):
      """
      tests L2-norm of Symbol on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 2),dim=dim)
        s=Data(0,(4, 2),w)
        s[0,0]=(0.792724994628)*sqrt(x[0])
        s[0,1]=(-0.662113166656)*sqrt(x[1])
        s[1,0]=(-0.548842538341)*sqrt(x[0])
        s[1,1]=(-0.555018478507)*sqrt(x[1])
        s[2,0]=(0.42962640812)*sqrt(x[0])
        s[2,1]=(-0.582731787726)*sqrt(x[1])
        s[3,0]=(0.603890261621)*sqrt(x[0])
        s[3,1]=(0.802141268835)*sqrt(x[1])
        ref=sqrt(1.60417482809)

      else:
        arg=Symbol(shape=(4, 3),dim=dim)
        s=Data(0,(4, 3),w)
        s[0,0]=(-0.432893320336)*sqrt(x[0])
        s[0,1]=(0.277328929482)*sqrt(x[1])
        s[0,2]=(0.950129871073)*sqrt(x[2])
        s[1,0]=(0.675116050333)*sqrt(x[0])
        s[1,1]=(-0.234274111148)*sqrt(x[1])
        s[1,2]=(-0.477441654222)*sqrt(x[2])
        s[2,0]=(0.603834062829)*sqrt(x[0])
        s[2,1]=(-0.606015016913)*sqrt(x[1])
        s[2,2]=(0.637487308345)*sqrt(x[2])
        s[3,0]=(-0.288536804159)*sqrt(x[0])
        s[3,1]=(-0.442827019286)*sqrt(x[1])
        s[3,2]=(-0.0978007221136)*sqrt(x[2])
        ref=sqrt(1.66642269482)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromSymbol_rank3(self):
      """
      tests L2-norm of Symbol on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(6, 2, 2),dim=dim)
        s=Data(0,(6, 2, 2),w)
        s[0,0,0]=(0.845216886022)*sqrt(x[0])
        s[0,0,1]=(0.0597182977987)*sqrt(x[1])
        s[0,1,0]=(0.146117477376)*sqrt(x[0])
        s[0,1,1]=(0.918064672934)*sqrt(x[1])
        s[1,0,0]=(-0.629068350604)*sqrt(x[0])
        s[1,0,1]=(-0.116285587839)*sqrt(x[1])
        s[1,1,0]=(0.678041887149)*sqrt(x[0])
        s[1,1,1]=(-0.756045170867)*sqrt(x[1])
        s[2,0,0]=(0.66884645591)*sqrt(x[0])
        s[2,0,1]=(0.527297857958)*sqrt(x[1])
        s[2,1,0]=(-0.453462310421)*sqrt(x[0])
        s[2,1,1]=(-0.801421668573)*sqrt(x[1])
        s[3,0,0]=(-0.418575000488)*sqrt(x[0])
        s[3,0,1]=(0.977995539022)*sqrt(x[1])
        s[3,1,0]=(0.863549457649)*sqrt(x[0])
        s[3,1,1]=(0.975604004421)*sqrt(x[1])
        s[4,0,0]=(0.642351898331)*sqrt(x[0])
        s[4,0,1]=(-0.63624912796)*sqrt(x[1])
        s[4,1,0]=(0.561260478169)*sqrt(x[0])
        s[4,1,1]=(-0.942077895874)*sqrt(x[1])
        s[5,0,0]=(0.0142173692077)*sqrt(x[0])
        s[5,0,1]=(0.106312727314)*sqrt(x[1])
        s[5,1,0]=(0.110676503653)*sqrt(x[0])
        s[5,1,1]=(-0.534921324957)*sqrt(x[1])
        ref=sqrt(4.87754875327)

      else:
        arg=Symbol(shape=(6, 2, 3),dim=dim)
        s=Data(0,(6, 2, 3),w)
        s[0,0,0]=(-0.351762958436)*sqrt(x[0])
        s[0,0,1]=(0.797778461739)*sqrt(x[1])
        s[0,0,2]=(-0.216759353067)*sqrt(x[2])
        s[0,1,0]=(0.913252336442)*sqrt(x[0])
        s[0,1,1]=(0.988076789445)*sqrt(x[1])
        s[0,1,2]=(-0.901479800398)*sqrt(x[2])
        s[1,0,0]=(0.179074582506)*sqrt(x[0])
        s[1,0,1]=(-0.0198494753788)*sqrt(x[1])
        s[1,0,2]=(0.453475519845)*sqrt(x[2])
        s[1,1,0]=(0.0703079255814)*sqrt(x[0])
        s[1,1,1]=(0.132531352932)*sqrt(x[1])
        s[1,1,2]=(0.746319860866)*sqrt(x[2])
        s[2,0,0]=(0.414228087724)*sqrt(x[0])
        s[2,0,1]=(0.466702867918)*sqrt(x[1])
        s[2,0,2]=(0.0868129987291)*sqrt(x[2])
        s[2,1,0]=(0.903287212799)*sqrt(x[0])
        s[2,1,1]=(0.912709008679)*sqrt(x[1])
        s[2,1,2]=(-0.0365579422157)*sqrt(x[2])
        s[3,0,0]=(0.0164686487169)*sqrt(x[0])
        s[3,0,1]=(-0.914613841493)*sqrt(x[1])
        s[3,0,2]=(-0.0822005463079)*sqrt(x[2])
        s[3,1,0]=(-0.315907377871)*sqrt(x[0])
        s[3,1,1]=(-0.102105289546)*sqrt(x[1])
        s[3,1,2]=(-0.0353732873141)*sqrt(x[2])
        s[4,0,0]=(0.0323628908944)*sqrt(x[0])
        s[4,0,1]=(-0.954360115684)*sqrt(x[1])
        s[4,0,2]=(0.923756483241)*sqrt(x[2])
        s[4,1,0]=(0.980063248251)*sqrt(x[0])
        s[4,1,1]=(0.0664804433015)*sqrt(x[1])
        s[4,1,2]=(-0.765454557157)*sqrt(x[2])
        s[5,0,0]=(-0.806747348451)*sqrt(x[0])
        s[5,0,1]=(-0.423078051628)*sqrt(x[1])
        s[5,0,2]=(0.765544317002)*sqrt(x[2])
        s[5,1,0]=(0.00826992946354)*sqrt(x[0])
        s[5,1,1]=(-0.117741801904)*sqrt(x[1])
        s[5,1,2]=(-0.975554426921)*sqrt(x[2])
        ref=sqrt(6.47379783714)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunction_fromSymbol_rank4(self):
      """
      tests L2-norm of Symbol on the ReducedFunction

      assumptions: self.domain supports integration on ReducedFunction
      """
      dim=self.domain.getDim()
      w=ReducedFunction(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
        s=Data(0,(4, 5, 3, 2),w)
        s[0,0,0,0]=(0.392249504065)*sqrt(x[0])
        s[0,0,0,1]=(-0.0829499082807)*sqrt(x[1])
        s[0,0,1,0]=(0.075391736082)*sqrt(x[0])
        s[0,0,1,1]=(-0.267274680434)*sqrt(x[1])
        s[0,0,2,0]=(0.758405394297)*sqrt(x[0])
        s[0,0,2,1]=(0.966859664545)*sqrt(x[1])
        s[0,1,0,0]=(-0.665158005166)*sqrt(x[0])
        s[0,1,0,1]=(-0.861735224083)*sqrt(x[1])
        s[0,1,1,0]=(0.889852595371)*sqrt(x[0])
        s[0,1,1,1]=(-0.456217187733)*sqrt(x[1])
        s[0,1,2,0]=(-0.444839601276)*sqrt(x[0])
        s[0,1,2,1]=(-0.646347424003)*sqrt(x[1])
        s[0,2,0,0]=(0.016728267633)*sqrt(x[0])
        s[0,2,0,1]=(0.598805251354)*sqrt(x[1])
        s[0,2,1,0]=(-0.0850259337934)*sqrt(x[0])
        s[0,2,1,1]=(-0.958442577173)*sqrt(x[1])
        s[0,2,2,0]=(-0.316155802878)*sqrt(x[0])
        s[0,2,2,1]=(0.882575663199)*sqrt(x[1])
        s[0,3,0,0]=(-0.350482054568)*sqrt(x[0])
        s[0,3,0,1]=(-0.0018770186166)*sqrt(x[1])
        s[0,3,1,0]=(0.434158350351)*sqrt(x[0])
        s[0,3,1,1]=(0.510916541891)*sqrt(x[1])
        s[0,3,2,0]=(0.415800847408)*sqrt(x[0])
        s[0,3,2,1]=(0.0045200661478)*sqrt(x[1])
        s[0,4,0,0]=(0.719981048541)*sqrt(x[0])
        s[0,4,0,1]=(-0.394991607605)*sqrt(x[1])
        s[0,4,1,0]=(0.225637020507)*sqrt(x[0])
        s[0,4,1,1]=(-0.942644070947)*sqrt(x[1])
        s[0,4,2,0]=(0.914486047699)*sqrt(x[0])
        s[0,4,2,1]=(-0.0605237231206)*sqrt(x[1])
        s[1,0,0,0]=(0.49162536785)*sqrt(x[0])
        s[1,0,0,1]=(-0.616492901013)*sqrt(x[1])
        s[1,0,1,0]=(-0.1288077626)*sqrt(x[0])
        s[1,0,1,1]=(-0.345407781206)*sqrt(x[1])
        s[1,0,2,0]=(-0.0725937970644)*sqrt(x[0])
        s[1,0,2,1]=(0.40531578317)*sqrt(x[1])
        s[1,1,0,0]=(0.598712741491)*sqrt(x[0])
        s[1,1,0,1]=(-0.93879902651)*sqrt(x[1])
        s[1,1,1,0]=(-0.0264475169249)*sqrt(x[0])
        s[1,1,1,1]=(0.249909294638)*sqrt(x[1])
        s[1,1,2,0]=(0.0027245979433)*sqrt(x[0])
        s[1,1,2,1]=(-0.373044612275)*sqrt(x[1])
        s[1,2,0,0]=(-0.281778393525)*sqrt(x[0])
        s[1,2,0,1]=(0.627994360637)*sqrt(x[1])
        s[1,2,1,0]=(0.589488628124)*sqrt(x[0])
        s[1,2,1,1]=(-0.758174695207)*sqrt(x[1])
        s[1,2,2,0]=(-0.406965936453)*sqrt(x[0])
        s[1,2,2,1]=(-0.69002957721)*sqrt(x[1])
        s[1,3,0,0]=(-0.199124201524)*sqrt(x[0])
        s[1,3,0,1]=(0.869253645966)*sqrt(x[1])
        s[1,3,1,0]=(-0.452605020068)*sqrt(x[0])
        s[1,3,1,1]=(-0.970641886846)*sqrt(x[1])
        s[1,3,2,0]=(0.175601171946)*sqrt(x[0])
        s[1,3,2,1]=(0.672851105536)*sqrt(x[1])
        s[1,4,0,0]=(0.58320491243)*sqrt(x[0])
        s[1,4,0,1]=(-0.259858520694)*sqrt(x[1])
        s[1,4,1,0]=(-0.808152456276)*sqrt(x[0])
        s[1,4,1,1]=(-0.061862146349)*sqrt(x[1])
        s[1,4,2,0]=(-0.285657379415)*sqrt(x[0])
        s[1,4,2,1]=(0.0400598209814)*sqrt(x[1])
        s[2,0,0,0]=(-0.982895080214)*sqrt(x[0])
        s[2,0,0,1]=(-0.221420258961)*sqrt(x[1])
        s[2,0,1,0]=(0.241664738784)*sqrt(x[0])
        s[2,0,1,1]=(0.360336298851)*sqrt(x[1])
        s[2,0,2,0]=(-0.1930995368)*sqrt(x[0])
        s[2,0,2,1]=(0.165323422951)*sqrt(x[1])
        s[2,1,0,0]=(-0.580617143448)*sqrt(x[0])
        s[2,1,0,1]=(-0.445386310829)*sqrt(x[1])
        s[2,1,1,0]=(-0.679424485212)*sqrt(x[0])
        s[2,1,1,1]=(0.999216345845)*sqrt(x[1])
        s[2,1,2,0]=(-0.00488871164474)*sqrt(x[0])
        s[2,1,2,1]=(0.94248053909)*sqrt(x[1])
        s[2,2,0,0]=(0.241836822065)*sqrt(x[0])
        s[2,2,0,1]=(-0.323265890223)*sqrt(x[1])
        s[2,2,1,0]=(0.282530883606)*sqrt(x[0])
        s[2,2,1,1]=(0.499481122286)*sqrt(x[1])
        s[2,2,2,0]=(0.502100738495)*sqrt(x[0])
        s[2,2,2,1]=(-0.714419126623)*sqrt(x[1])
        s[2,3,0,0]=(0.427736585056)*sqrt(x[0])
        s[2,3,0,1]=(0.846933115255)*sqrt(x[1])
        s[2,3,1,0]=(0.293400967424)*sqrt(x[0])
        s[2,3,1,1]=(0.258925376632)*sqrt(x[1])
        s[2,3,2,0]=(-0.748167607341)*sqrt(x[0])
        s[2,3,2,1]=(0.354090561451)*sqrt(x[1])
        s[2,4,0,0]=(0.432737336014)*sqrt(x[0])
        s[2,4,0,1]=(0.856383164324)*sqrt(x[1])
        s[2,4,1,0]=(-0.0596776788823)*sqrt(x[0])
        s[2,4,1,1]=(-0.299291531668)*sqrt(x[1])
        s[2,4,2,0]=(0.0332610005043)*sqrt(x[0])
        s[2,4,2,1]=(-0.784223449976)*sqrt(x[1])
        s[3,0,0,0]=(0.856346607718)*sqrt(x[0])
        s[3,0,0,1]=(-0.711038642485)*sqrt(x[1])
        s[3,0,1,0]=(-0.722899969631)*sqrt(x[0])
        s[3,0,1,1]=(-0.976358349372)*sqrt(x[1])
        s[3,0,2,0]=(-0.566657009276)*sqrt(x[0])
        s[3,0,2,1]=(-0.88977921418)*sqrt(x[1])
        s[3,1,0,0]=(0.690995364783)*sqrt(x[0])
        s[3,1,0,1]=(0.0178260768232)*sqrt(x[1])
        s[3,1,1,0]=(0.210672134956)*sqrt(x[0])
        s[3,1,1,1]=(-0.271769743548)*sqrt(x[1])
        s[3,1,2,0]=(-0.119905539717)*sqrt(x[0])
        s[3,1,2,1]=(0.408792239219)*sqrt(x[1])
        s[3,2,0,0]=(0.134901834799)*sqrt(x[0])
        s[3,2,0,1]=(0.0969235330927)*sqrt(x[1])
        s[3,2,1,0]=(-0.418267686354)*sqrt(x[0])
        s[3,2,1,1]=(-0.599174142878)*sqrt(x[1])
        s[3,2,2,0]=(-0.409148104875)*sqrt(x[0])
        s[3,2,2,1]=(0.504760142275)*sqrt(x[1])
        s[3,3,0,0]=(-0.174999245323)*sqrt(x[0])
        s[3,3,0,1]=(-0.994094684291)*sqrt(x[1])
        s[3,3,1,0]=(0.470460754362)*sqrt(x[0])
        s[3,3,1,1]=(-0.867881666651)*sqrt(x[1])
        s[3,3,2,0]=(0.691171063635)*sqrt(x[0])
        s[3,3,2,1]=(-0.644952618882)*sqrt(x[1])
        s[3,4,0,0]=(0.883581127272)*sqrt(x[0])
        s[3,4,0,1]=(-0.534009409512)*sqrt(x[1])
        s[3,4,1,0]=(0.324806602648)*sqrt(x[0])
        s[3,4,1,1]=(0.606235157996)*sqrt(x[1])
        s[3,4,2,0]=(-0.0136704045984)*sqrt(x[0])
        s[3,4,2,1]=(0.36671612245)*sqrt(x[1])
        ref=sqrt(18.4082461669)

      else:
        arg=Symbol(shape=(4, 5, 3, 3),dim=dim)
        s=Data(0,(4, 5, 3, 3),w)
        s[0,0,0,0]=(-0.0241692471764)*sqrt(x[0])
        s[0,0,0,1]=(0.816194825296)*sqrt(x[1])
        s[0,0,0,2]=(-0.0193776462565)*sqrt(x[2])
        s[0,0,1,0]=(0.885158780616)*sqrt(x[0])
        s[0,0,1,1]=(0.750445408795)*sqrt(x[1])
        s[0,0,1,2]=(-0.242904502332)*sqrt(x[2])
        s[0,0,2,0]=(-0.101306101076)*sqrt(x[0])
        s[0,0,2,1]=(0.254549514037)*sqrt(x[1])
        s[0,0,2,2]=(-0.301743139759)*sqrt(x[2])
        s[0,1,0,0]=(0.960790978081)*sqrt(x[0])
        s[0,1,0,1]=(-0.873861925572)*sqrt(x[1])
        s[0,1,0,2]=(-0.266926259205)*sqrt(x[2])
        s[0,1,1,0]=(-0.533605756527)*sqrt(x[0])
        s[0,1,1,1]=(-0.196348553154)*sqrt(x[1])
        s[0,1,1,2]=(0.462438110297)*sqrt(x[2])
        s[0,1,2,0]=(0.766114991742)*sqrt(x[0])
        s[0,1,2,1]=(-0.282544763069)*sqrt(x[1])
        s[0,1,2,2]=(-0.918562937205)*sqrt(x[2])
        s[0,2,0,0]=(-0.723736937045)*sqrt(x[0])
        s[0,2,0,1]=(-0.0271772220034)*sqrt(x[1])
        s[0,2,0,2]=(-0.67448574094)*sqrt(x[2])
        s[0,2,1,0]=(-0.0465782041976)*sqrt(x[0])
        s[0,2,1,1]=(-0.701475255709)*sqrt(x[1])
        s[0,2,1,2]=(-0.54745824908)*sqrt(x[2])
        s[0,2,2,0]=(0.823427037717)*sqrt(x[0])
        s[0,2,2,1]=(0.873366807256)*sqrt(x[1])
        s[0,2,2,2]=(0.902453041345)*sqrt(x[2])
        s[0,3,0,0]=(-0.383388300734)*sqrt(x[0])
        s[0,3,0,1]=(-0.457344890379)*sqrt(x[1])
        s[0,3,0,2]=(0.449736739028)*sqrt(x[2])
        s[0,3,1,0]=(-0.994113075606)*sqrt(x[0])
        s[0,3,1,1]=(0.0265853285952)*sqrt(x[1])
        s[0,3,1,2]=(-0.738089523761)*sqrt(x[2])
        s[0,3,2,0]=(0.663756875286)*sqrt(x[0])
        s[0,3,2,1]=(0.597615440207)*sqrt(x[1])
        s[0,3,2,2]=(0.246105261445)*sqrt(x[2])
        s[0,4,0,0]=(0.397226701305)*sqrt(x[0])
        s[0,4,0,1]=(0.0967432951001)*sqrt(x[1])
        s[0,4,0,2]=(0.619665795097)*sqrt(x[2])
        s[0,4,1,0]=(0.155541932506)*sqrt(x[0])
        s[0,4,1,1]=(-0.114961356145)*sqrt(x[1])
        s[0,4,1,2]=(-0.491668824315)*sqrt(x[2])
        s[0,4,2,0]=(-0.726549525506)*sqrt(x[0])
        s[0,4,2,1]=(-0.876194753259)*sqrt(x[1])
        s[0,4,2,2]=(-0.506364111318)*sqrt(x[2])
        s[1,0,0,0]=(-0.380662783497)*sqrt(x[0])
        s[1,0,0,1]=(-0.57927422387)*sqrt(x[1])
        s[1,0,0,2]=(-0.282039117242)*sqrt(x[2])
        s[1,0,1,0]=(-0.873334149601)*sqrt(x[0])
        s[1,0,1,1]=(-0.0145182664909)*sqrt(x[1])
        s[1,0,1,2]=(0.903533313452)*sqrt(x[2])
        s[1,0,2,0]=(0.123617284156)*sqrt(x[0])
        s[1,0,2,1]=(0.716639379262)*sqrt(x[1])
        s[1,0,2,2]=(-0.940470701414)*sqrt(x[2])
        s[1,1,0,0]=(-0.831075410065)*sqrt(x[0])
        s[1,1,0,1]=(0.808724910022)*sqrt(x[1])
        s[1,1,0,2]=(0.218369441977)*sqrt(x[2])
        s[1,1,1,0]=(0.690977533444)*sqrt(x[0])
        s[1,1,1,1]=(-0.379327957709)*sqrt(x[1])
        s[1,1,1,2]=(0.0964273177138)*sqrt(x[2])
        s[1,1,2,0]=(-0.244790745268)*sqrt(x[0])
        s[1,1,2,1]=(-0.208570215865)*sqrt(x[1])
        s[1,1,2,2]=(0.280320965835)*sqrt(x[2])
        s[1,2,0,0]=(-0.571572336783)*sqrt(x[0])
        s[1,2,0,1]=(-0.512302129833)*sqrt(x[1])
        s[1,2,0,2]=(0.458134267946)*sqrt(x[2])
        s[1,2,1,0]=(-0.909450574449)*sqrt(x[0])
        s[1,2,1,1]=(-0.54240194702)*sqrt(x[1])
        s[1,2,1,2]=(0.015279542779)*sqrt(x[2])
        s[1,2,2,0]=(-0.450829963713)*sqrt(x[0])
        s[1,2,2,1]=(0.265958897553)*sqrt(x[1])
        s[1,2,2,2]=(0.0115876819282)*sqrt(x[2])
        s[1,3,0,0]=(-0.493820517172)*sqrt(x[0])
        s[1,3,0,1]=(0.924501975039)*sqrt(x[1])
        s[1,3,0,2]=(0.976144734904)*sqrt(x[2])
        s[1,3,1,0]=(-0.0369162848651)*sqrt(x[0])
        s[1,3,1,1]=(-0.282951956468)*sqrt(x[1])
        s[1,3,1,2]=(0.46799488403)*sqrt(x[2])
        s[1,3,2,0]=(0.859289726723)*sqrt(x[0])
        s[1,3,2,1]=(0.460555626035)*sqrt(x[1])
        s[1,3,2,2]=(0.515796525158)*sqrt(x[2])
        s[1,4,0,0]=(0.707364903535)*sqrt(x[0])
        s[1,4,0,1]=(0.749724556391)*sqrt(x[1])
        s[1,4,0,2]=(-0.278863489815)*sqrt(x[2])
        s[1,4,1,0]=(0.972950918189)*sqrt(x[0])
        s[1,4,1,1]=(0.844177970875)*sqrt(x[1])
        s[1,4,1,2]=(0.0763619420058)*sqrt(x[2])
        s[1,4,2,0]=(-0.0852939975129)*sqrt(x[0])
        s[1,4,2,1]=(0.583399144502)*sqrt(x[1])
        s[1,4,2,2]=(0.848661907972)*sqrt(x[2])
        s[2,0,0,0]=(-0.777408601463)*sqrt(x[0])
        s[2,0,0,1]=(-0.855232334446)*sqrt(x[1])
        s[2,0,0,2]=(-0.742431398498)*sqrt(x[2])
        s[2,0,1,0]=(-0.96680966892)*sqrt(x[0])
        s[2,0,1,1]=(-0.33301062934)*sqrt(x[1])
        s[2,0,1,2]=(-0.675842472611)*sqrt(x[2])
        s[2,0,2,0]=(-0.0163730639481)*sqrt(x[0])
        s[2,0,2,1]=(-0.0854274064817)*sqrt(x[1])
        s[2,0,2,2]=(-0.93094914338)*sqrt(x[2])
        s[2,1,0,0]=(0.881651879025)*sqrt(x[0])
        s[2,1,0,1]=(0.782286554309)*sqrt(x[1])
        s[2,1,0,2]=(0.71813995101)*sqrt(x[2])
        s[2,1,1,0]=(-0.0790321088042)*sqrt(x[0])
        s[2,1,1,1]=(-0.181510921696)*sqrt(x[1])
        s[2,1,1,2]=(0.822016852235)*sqrt(x[2])
        s[2,1,2,0]=(0.634313542034)*sqrt(x[0])
        s[2,1,2,1]=(0.797139044679)*sqrt(x[1])
        s[2,1,2,2]=(0.400542618196)*sqrt(x[2])
        s[2,2,0,0]=(-0.823507870125)*sqrt(x[0])
        s[2,2,0,1]=(0.981256538471)*sqrt(x[1])
        s[2,2,0,2]=(0.158779409645)*sqrt(x[2])
        s[2,2,1,0]=(-0.0788962261172)*sqrt(x[0])
        s[2,2,1,1]=(-0.295903805518)*sqrt(x[1])
        s[2,2,1,2]=(0.338095367799)*sqrt(x[2])
        s[2,2,2,0]=(0.259914616014)*sqrt(x[0])
        s[2,2,2,1]=(-0.935338026316)*sqrt(x[1])
        s[2,2,2,2]=(0.752104609965)*sqrt(x[2])
        s[2,3,0,0]=(0.770409830614)*sqrt(x[0])
        s[2,3,0,1]=(0.827165793631)*sqrt(x[1])
        s[2,3,0,2]=(-0.570252686313)*sqrt(x[2])
        s[2,3,1,0]=(-0.645381957261)*sqrt(x[0])
        s[2,3,1,1]=(-0.547153600936)*sqrt(x[1])
        s[2,3,1,2]=(-0.424301255985)*sqrt(x[2])
        s[2,3,2,0]=(0.938907909024)*sqrt(x[0])
        s[2,3,2,1]=(-0.165860294169)*sqrt(x[1])
        s[2,3,2,2]=(-0.507668918028)*sqrt(x[2])
        s[2,4,0,0]=(0.504198965293)*sqrt(x[0])
        s[2,4,0,1]=(-0.668203518046)*sqrt(x[1])
        s[2,4,0,2]=(0.229406581915)*sqrt(x[2])
        s[2,4,1,0]=(0.454298324734)*sqrt(x[0])
        s[2,4,1,1]=(0.631911737372)*sqrt(x[1])
        s[2,4,1,2]=(-0.849242047895)*sqrt(x[2])
        s[2,4,2,0]=(0.0187819832254)*sqrt(x[0])
        s[2,4,2,1]=(-0.0257470291285)*sqrt(x[1])
        s[2,4,2,2]=(-0.968820620099)*sqrt(x[2])
        s[3,0,0,0]=(0.55579863211)*sqrt(x[0])
        s[3,0,0,1]=(0.830363451819)*sqrt(x[1])
        s[3,0,0,2]=(0.37501906831)*sqrt(x[2])
        s[3,0,1,0]=(-0.609478442879)*sqrt(x[0])
        s[3,0,1,1]=(0.895531311344)*sqrt(x[1])
        s[3,0,1,2]=(0.457055008762)*sqrt(x[2])
        s[3,0,2,0]=(-0.589517607648)*sqrt(x[0])
        s[3,0,2,1]=(-0.840365410994)*sqrt(x[1])
        s[3,0,2,2]=(-0.804004197791)*sqrt(x[2])
        s[3,1,0,0]=(0.231479229381)*sqrt(x[0])
        s[3,1,0,1]=(0.851135131086)*sqrt(x[1])
        s[3,1,0,2]=(-0.558112154058)*sqrt(x[2])
        s[3,1,1,0]=(-0.365328057806)*sqrt(x[0])
        s[3,1,1,1]=(0.665370944345)*sqrt(x[1])
        s[3,1,1,2]=(0.815245254888)*sqrt(x[2])
        s[3,1,2,0]=(-0.622675567594)*sqrt(x[0])
        s[3,1,2,1]=(0.15478575296)*sqrt(x[1])
        s[3,1,2,2]=(-0.919833514628)*sqrt(x[2])
        s[3,2,0,0]=(0.893976278733)*sqrt(x[0])
        s[3,2,0,1]=(-0.881412936262)*sqrt(x[1])
        s[3,2,0,2]=(-0.940014543462)*sqrt(x[2])
        s[3,2,1,0]=(0.310442650776)*sqrt(x[0])
        s[3,2,1,1]=(0.895222325851)*sqrt(x[1])
        s[3,2,1,2]=(0.308995631858)*sqrt(x[2])
        s[3,2,2,0]=(0.0373558936063)*sqrt(x[0])
        s[3,2,2,1]=(-0.0592368839463)*sqrt(x[1])
        s[3,2,2,2]=(0.793681649539)*sqrt(x[2])
        s[3,3,0,0]=(-0.542535013221)*sqrt(x[0])
        s[3,3,0,1]=(0.730645327078)*sqrt(x[1])
        s[3,3,0,2]=(-0.776045858279)*sqrt(x[2])
        s[3,3,1,0]=(-0.657528555613)*sqrt(x[0])
        s[3,3,1,1]=(-0.282142368825)*sqrt(x[1])
        s[3,3,1,2]=(-0.692363751497)*sqrt(x[2])
        s[3,3,2,0]=(-0.168946040457)*sqrt(x[0])
        s[3,3,2,1]=(-0.318493817076)*sqrt(x[1])
        s[3,3,2,2]=(0.535922920691)*sqrt(x[2])
        s[3,4,0,0]=(0.414492782118)*sqrt(x[0])
        s[3,4,0,1]=(0.295207630842)*sqrt(x[1])
        s[3,4,0,2]=(0.453596918765)*sqrt(x[2])
        s[3,4,1,0]=(-0.083655892279)*sqrt(x[0])
        s[3,4,1,1]=(-0.962961908465)*sqrt(x[1])
        s[3,4,1,2]=(0.138455397004)*sqrt(x[2])
        s[3,4,2,0]=(-0.383498097386)*sqrt(x[0])
        s[3,4,2,1]=(0.650869920691)*sqrt(x[1])
        s[3,4,2,2]=(-0.998836301837)*sqrt(x[2])
        ref=sqrt(33.3362509396)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank0(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(-0.245574919477)*x[0]
        ref=sqrt((0.0603070410759)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(),w)
        arg=(0.757324521515)*x[0]
        ref=sqrt((0.573540430888)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank1(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(0.723421565407)*x[0]
        arg[1]=(-0.460477393103)*x[1]
        ref=sqrt((0.735378190855)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(3,),w)
        arg[0]=(-0.88528497163)*x[0]
        arg[1]=(-0.65510214636)*x[1]
        arg[2]=(0.399538866363)*x[2]
        ref=sqrt((1.37251960889)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank2(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.955620993904)*x[0]
        arg[0,1]=(-0.0987865813703)*x[1]
        arg[1,0]=(0.0288267231531)*x[0]
        arg[1,1]=(0.655440599879)*x[1]
        arg[2,0]=(0.685627284533)*x[0]
        arg[2,1]=(-0.989832824892)*x[1]
        arg[3,0]=(0.292184093194)*x[0]
        arg[3,1]=(0.149553857773)*x[1]
        ref=sqrt((2.91099532781)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(-0.325908541533)*x[0]
        arg[0,1]=(-0.992480479749)*x[1]
        arg[0,2]=(0.660360271799)*x[2]
        arg[1,0]=(0.173485908581)*x[0]
        arg[1,1]=(-0.328755199781)*x[1]
        arg[1,2]=(-0.943354674948)*x[2]
        arg[2,0]=(0.680713222646)*x[0]
        arg[2,1]=(-0.765971835693)*x[1]
        arg[2,2]=(0.0413284847528)*x[2]
        arg[3,0]=(0.990074004708)*x[0]
        arg[3,1]=(0.941801786766)*x[1]
        arg[3,2]=(0.886926192201)*x[2]
        ref=sqrt((6.26107155228)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank3(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(-0.0781611551598)*x[0]
        arg[0,0,1]=(0.291016249575)*x[1]
        arg[0,1,0]=(-0.107555233086)*x[0]
        arg[0,1,1]=(-0.559067108546)*x[1]
        arg[1,0,0]=(-0.0818406701266)*x[0]
        arg[1,0,1]=(-0.594866806483)*x[1]
        arg[1,1,0]=(-0.725814803863)*x[0]
        arg[1,1,1]=(0.59128101992)*x[1]
        arg[2,0,0]=(-0.15381555291)*x[0]
        arg[2,0,1]=(-0.679882948503)*x[1]
        arg[2,1,0]=(-0.58437193917)*x[0]
        arg[2,1,1]=(0.136304615849)*x[1]
        arg[3,0,0]=(0.0671365410096)*x[0]
        arg[3,0,1]=(-0.645687212187)*x[1]
        arg[3,1,0]=(-0.642492412392)*x[0]
        arg[3,1,1]=(-0.125760054735)*x[1]
        arg[4,0,0]=(0.731110824794)*x[0]
        arg[4,0,1]=(0.491668422979)*x[1]
        arg[4,1,0]=(-0.775841478292)*x[0]
        arg[4,1,1]=(0.728265567974)*x[1]
        arg[5,0,0]=(0.84511832373)*x[0]
        arg[5,0,1]=(-0.513796801068)*x[1]
        arg[5,1,0]=(0.113072243554)*x[0]
        arg[5,1,1]=(0.246630838744)*x[1]
        ref=sqrt((6.30829536252)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.369748116859)*x[0]
        arg[0,0,1]=(-0.758056560031)*x[1]
        arg[0,0,2]=(-0.873984709951)*x[2]
        arg[0,1,0]=(0.311680165784)*x[0]
        arg[0,1,1]=(0.374400673651)*x[1]
        arg[0,1,2]=(0.712484217076)*x[2]
        arg[1,0,0]=(0.829379714484)*x[0]
        arg[1,0,1]=(-0.0551589596149)*x[1]
        arg[1,0,2]=(0.965672208426)*x[2]
        arg[1,1,0]=(-0.205044281547)*x[0]
        arg[1,1,1]=(0.238197452756)*x[1]
        arg[1,1,2]=(-0.33456139292)*x[2]
        arg[2,0,0]=(0.649928288926)*x[0]
        arg[2,0,1]=(-0.661384953389)*x[1]
        arg[2,0,2]=(-0.253241222975)*x[2]
        arg[2,1,0]=(-0.491716575992)*x[0]
        arg[2,1,1]=(-0.970872527468)*x[1]
        arg[2,1,2]=(0.222410198921)*x[2]
        arg[3,0,0]=(0.205752630262)*x[0]
        arg[3,0,1]=(0.864804362697)*x[1]
        arg[3,0,2]=(-0.417975564033)*x[2]
        arg[3,1,0]=(0.586425694033)*x[0]
        arg[3,1,1]=(0.952661122184)*x[1]
        arg[3,1,2]=(0.608680080453)*x[2]
        arg[4,0,0]=(0.625968903369)*x[0]
        arg[4,0,1]=(-0.573909405003)*x[1]
        arg[4,0,2]=(-0.762256394595)*x[2]
        arg[4,1,0]=(0.0710742394418)*x[0]
        arg[4,1,1]=(0.583378040574)*x[1]
        arg[4,1,2]=(0.719032893115)*x[2]
        arg[5,0,0]=(0.032173368884)*x[0]
        arg[5,0,1]=(-0.434042549492)*x[1]
        arg[5,0,2]=(0.363504765447)*x[2]
        arg[5,1,0]=(0.598817469198)*x[0]
        arg[5,1,1]=(-0.163967008775)*x[1]
        arg[5,1,2]=(0.546778730604)*x[2]
        ref=sqrt((11.9696343123)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromData_rank4(self):
      """
      tests L2-norm of Data on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.273446520069)*x[0]
        arg[0,0,0,1]=(-0.913305910831)*x[1]
        arg[0,0,1,0]=(0.0745566747537)*x[0]
        arg[0,0,1,1]=(0.98803601919)*x[1]
        arg[0,0,2,0]=(-0.244120818875)*x[0]
        arg[0,0,2,1]=(0.247509644998)*x[1]
        arg[0,1,0,0]=(-0.548756417777)*x[0]
        arg[0,1,0,1]=(-0.354587911923)*x[1]
        arg[0,1,1,0]=(0.104268198867)*x[0]
        arg[0,1,1,1]=(-0.541700877072)*x[1]
        arg[0,1,2,0]=(-0.25900060689)*x[0]
        arg[0,1,2,1]=(-0.859660175231)*x[1]
        arg[0,2,0,0]=(0.195235560321)*x[0]
        arg[0,2,0,1]=(0.175518738589)*x[1]
        arg[0,2,1,0]=(-0.0638854232272)*x[0]
        arg[0,2,1,1]=(-0.586161016541)*x[1]
        arg[0,2,2,0]=(0.580258892247)*x[0]
        arg[0,2,2,1]=(-0.931927145435)*x[1]
        arg[0,3,0,0]=(-0.408209600298)*x[0]
        arg[0,3,0,1]=(-0.0344882667014)*x[1]
        arg[0,3,1,0]=(-0.131763534163)*x[0]
        arg[0,3,1,1]=(-0.787653739965)*x[1]
        arg[0,3,2,0]=(0.0910808104711)*x[0]
        arg[0,3,2,1]=(-0.280409023611)*x[1]
        arg[0,4,0,0]=(0.97745754012)*x[0]
        arg[0,4,0,1]=(-0.59829020936)*x[1]
        arg[0,4,1,0]=(0.890520260543)*x[0]
        arg[0,4,1,1]=(-0.600760090231)*x[1]
        arg[0,4,2,0]=(-0.897992297974)*x[0]
        arg[0,4,2,1]=(-0.841169898923)*x[1]
        arg[1,0,0,0]=(-0.249868816409)*x[0]
        arg[1,0,0,1]=(0.620375082228)*x[1]
        arg[1,0,1,0]=(-0.660480789306)*x[0]
        arg[1,0,1,1]=(-0.73638571806)*x[1]
        arg[1,0,2,0]=(0.339987316643)*x[0]
        arg[1,0,2,1]=(0.541112529894)*x[1]
        arg[1,1,0,0]=(-0.468808186705)*x[0]
        arg[1,1,0,1]=(-0.32919679792)*x[1]
        arg[1,1,1,0]=(0.917292803419)*x[0]
        arg[1,1,1,1]=(-0.834265058005)*x[1]
        arg[1,1,2,0]=(-0.247536849264)*x[0]
        arg[1,1,2,1]=(-0.197503469238)*x[1]
        arg[1,2,0,0]=(0.897591919909)*x[0]
        arg[1,2,0,1]=(-0.807446231234)*x[1]
        arg[1,2,1,0]=(-0.369878499382)*x[0]
        arg[1,2,1,1]=(0.985678692179)*x[1]
        arg[1,2,2,0]=(-0.709976427525)*x[0]
        arg[1,2,2,1]=(-0.368744647016)*x[1]
        arg[1,3,0,0]=(0.299630726462)*x[0]
        arg[1,3,0,1]=(-0.445295757899)*x[1]
        arg[1,3,1,0]=(-0.922386577254)*x[0]
        arg[1,3,1,1]=(0.234794853697)*x[1]
        arg[1,3,2,0]=(0.953377720197)*x[0]
        arg[1,3,2,1]=(0.409778183998)*x[1]
        arg[1,4,0,0]=(0.271967945488)*x[0]
        arg[1,4,0,1]=(-0.578629001202)*x[1]
        arg[1,4,1,0]=(0.210755226769)*x[0]
        arg[1,4,1,1]=(-0.0902751419945)*x[1]
        arg[1,4,2,0]=(0.70033387381)*x[0]
        arg[1,4,2,1]=(0.305733565661)*x[1]
        arg[2,0,0,0]=(-0.662483167298)*x[0]
        arg[2,0,0,1]=(0.585252048652)*x[1]
        arg[2,0,1,0]=(-0.398813785959)*x[0]
        arg[2,0,1,1]=(-0.797438697186)*x[1]
        arg[2,0,2,0]=(-0.508308971009)*x[0]
        arg[2,0,2,1]=(0.302249407524)*x[1]
        arg[2,1,0,0]=(0.208644491879)*x[0]
        arg[2,1,0,1]=(-0.604749055374)*x[1]
        arg[2,1,1,0]=(0.641654284594)*x[0]
        arg[2,1,1,1]=(0.456898593356)*x[1]
        arg[2,1,2,0]=(-0.398043867778)*x[0]
        arg[2,1,2,1]=(-0.0712344657587)*x[1]
        arg[2,2,0,0]=(0.0967860954865)*x[0]
        arg[2,2,0,1]=(0.520449905952)*x[1]
        arg[2,2,1,0]=(0.770198029595)*x[0]
        arg[2,2,1,1]=(-0.594004671621)*x[1]
        arg[2,2,2,0]=(-0.744571452885)*x[0]
        arg[2,2,2,1]=(0.544447367825)*x[1]
        arg[2,3,0,0]=(-0.137087966968)*x[0]
        arg[2,3,0,1]=(0.120672667497)*x[1]
        arg[2,3,1,0]=(0.204800088057)*x[0]
        arg[2,3,1,1]=(0.626526346076)*x[1]
        arg[2,3,2,0]=(-0.696480393227)*x[0]
        arg[2,3,2,1]=(0.188533741996)*x[1]
        arg[2,4,0,0]=(-0.403523821067)*x[0]
        arg[2,4,0,1]=(-0.428048989483)*x[1]
        arg[2,4,1,0]=(-0.244186366584)*x[0]
        arg[2,4,1,1]=(0.00866444909003)*x[1]
        arg[2,4,2,0]=(-0.445991308853)*x[0]
        arg[2,4,2,1]=(-0.899951068935)*x[1]
        arg[3,0,0,0]=(0.609340085418)*x[0]
        arg[3,0,0,1]=(0.878750391425)*x[1]
        arg[3,0,1,0]=(0.258064654464)*x[0]
        arg[3,0,1,1]=(-0.482402612985)*x[1]
        arg[3,0,2,0]=(0.943732283389)*x[0]
        arg[3,0,2,1]=(0.65514211843)*x[1]
        arg[3,1,0,0]=(-0.894551979619)*x[0]
        arg[3,1,0,1]=(0.220116541042)*x[1]
        arg[3,1,1,0]=(0.386887699577)*x[0]
        arg[3,1,1,1]=(-0.422560075108)*x[1]
        arg[3,1,2,0]=(0.00387273783493)*x[0]
        arg[3,1,2,1]=(0.465673505613)*x[1]
        arg[3,2,0,0]=(0.987383428982)*x[0]
        arg[3,2,0,1]=(0.376320055964)*x[1]
        arg[3,2,1,0]=(-0.463778689128)*x[0]
        arg[3,2,1,1]=(0.179816566227)*x[1]
        arg[3,2,2,0]=(0.961522856801)*x[0]
        arg[3,2,2,1]=(-0.257779627946)*x[1]
        arg[3,3,0,0]=(0.748886531458)*x[0]
        arg[3,3,0,1]=(-0.257342282566)*x[1]
        arg[3,3,1,0]=(0.377494024401)*x[0]
        arg[3,3,1,1]=(-0.334588346017)*x[1]
        arg[3,3,2,0]=(0.502495149189)*x[0]
        arg[3,3,2,1]=(0.534612429702)*x[1]
        arg[3,4,0,0]=(-0.308551337355)*x[0]
        arg[3,4,0,1]=(-0.471825826745)*x[1]
        arg[3,4,1,0]=(-0.262606531584)*x[0]
        arg[3,4,1,1]=(0.766089616367)*x[1]
        arg[3,4,2,0]=(-0.136526755642)*x[0]
        arg[3,4,2,1]=(0.675111459363)*x[1]
        ref=sqrt((37.3453550914)*(2.*dim+1.)/3.)

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.532946231146)*x[0]
        arg[0,0,0,1]=(0.269364089513)*x[1]
        arg[0,0,0,2]=(-0.207412457081)*x[2]
        arg[0,0,1,0]=(-0.843104704858)*x[0]
        arg[0,0,1,1]=(0.0416216508473)*x[1]
        arg[0,0,1,2]=(-0.836074693662)*x[2]
        arg[0,0,2,0]=(0.943609268731)*x[0]
        arg[0,0,2,1]=(0.0154543737816)*x[1]
        arg[0,0,2,2]=(0.0726268788381)*x[2]
        arg[0,1,0,0]=(0.108422740078)*x[0]
        arg[0,1,0,1]=(-0.296667916638)*x[1]
        arg[0,1,0,2]=(-0.769732600535)*x[2]
        arg[0,1,1,0]=(-0.428575493834)*x[0]
        arg[0,1,1,1]=(0.421245456722)*x[1]
        arg[0,1,1,2]=(-0.588277652277)*x[2]
        arg[0,1,2,0]=(0.145294576795)*x[0]
        arg[0,1,2,1]=(0.323206623794)*x[1]
        arg[0,1,2,2]=(0.788115602892)*x[2]
        arg[0,2,0,0]=(-0.227877282292)*x[0]
        arg[0,2,0,1]=(-0.630647460719)*x[1]
        arg[0,2,0,2]=(0.58754882135)*x[2]
        arg[0,2,1,0]=(0.347191113403)*x[0]
        arg[0,2,1,1]=(0.464093634725)*x[1]
        arg[0,2,1,2]=(-0.0412800774497)*x[2]
        arg[0,2,2,0]=(0.223364317185)*x[0]
        arg[0,2,2,1]=(0.257201130157)*x[1]
        arg[0,2,2,2]=(0.063203467463)*x[2]
        arg[0,3,0,0]=(-0.723240451643)*x[0]
        arg[0,3,0,1]=(-0.862468295097)*x[1]
        arg[0,3,0,2]=(-0.149283247587)*x[2]
        arg[0,3,1,0]=(0.15680097839)*x[0]
        arg[0,3,1,1]=(0.421563637547)*x[1]
        arg[0,3,1,2]=(0.111549188549)*x[2]
        arg[0,3,2,0]=(-0.272783329363)*x[0]
        arg[0,3,2,1]=(-0.420352789853)*x[1]
        arg[0,3,2,2]=(0.570865117722)*x[2]
        arg[0,4,0,0]=(-0.321910078414)*x[0]
        arg[0,4,0,1]=(0.988695599439)*x[1]
        arg[0,4,0,2]=(0.920200893398)*x[2]
        arg[0,4,1,0]=(0.0260910072651)*x[0]
        arg[0,4,1,1]=(0.460012578184)*x[1]
        arg[0,4,1,2]=(0.848099524112)*x[2]
        arg[0,4,2,0]=(0.242157803251)*x[0]
        arg[0,4,2,1]=(0.394528777004)*x[1]
        arg[0,4,2,2]=(0.562996837311)*x[2]
        arg[1,0,0,0]=(0.459886225958)*x[0]
        arg[1,0,0,1]=(-0.721868942003)*x[1]
        arg[1,0,0,2]=(0.432203082994)*x[2]
        arg[1,0,1,0]=(0.409831045482)*x[0]
        arg[1,0,1,1]=(-0.481677513473)*x[1]
        arg[1,0,1,2]=(0.439387853437)*x[2]
        arg[1,0,2,0]=(0.261583198434)*x[0]
        arg[1,0,2,1]=(0.290993423577)*x[1]
        arg[1,0,2,2]=(0.477993114134)*x[2]
        arg[1,1,0,0]=(0.586344598248)*x[0]
        arg[1,1,0,1]=(-0.105390792831)*x[1]
        arg[1,1,0,2]=(0.335990751314)*x[2]
        arg[1,1,1,0]=(-0.191500562856)*x[0]
        arg[1,1,1,1]=(0.244514598216)*x[1]
        arg[1,1,1,2]=(-0.804402720669)*x[2]
        arg[1,1,2,0]=(-0.455225710648)*x[0]
        arg[1,1,2,1]=(-0.505052700585)*x[1]
        arg[1,1,2,2]=(-0.0240295199362)*x[2]
        arg[1,2,0,0]=(-0.718487964893)*x[0]
        arg[1,2,0,1]=(-0.0899522570462)*x[1]
        arg[1,2,0,2]=(-0.293353754696)*x[2]
        arg[1,2,1,0]=(-0.180013826342)*x[0]
        arg[1,2,1,1]=(0.793689231922)*x[1]
        arg[1,2,1,2]=(0.673066555571)*x[2]
        arg[1,2,2,0]=(0.705362155032)*x[0]
        arg[1,2,2,1]=(0.54476742883)*x[1]
        arg[1,2,2,2]=(-0.331195064878)*x[2]
        arg[1,3,0,0]=(-0.360927441647)*x[0]
        arg[1,3,0,1]=(0.230772030282)*x[1]
        arg[1,3,0,2]=(0.912342489431)*x[2]
        arg[1,3,1,0]=(-0.817510690014)*x[0]
        arg[1,3,1,1]=(0.397583721353)*x[1]
        arg[1,3,1,2]=(-0.982551067917)*x[2]
        arg[1,3,2,0]=(0.86380240427)*x[0]
        arg[1,3,2,1]=(-0.415018976841)*x[1]
        arg[1,3,2,2]=(0.271582572267)*x[2]
        arg[1,4,0,0]=(0.252845347406)*x[0]
        arg[1,4,0,1]=(0.687786802906)*x[1]
        arg[1,4,0,2]=(0.465501171342)*x[2]
        arg[1,4,1,0]=(-0.613703721675)*x[0]
        arg[1,4,1,1]=(-0.110297640533)*x[1]
        arg[1,4,1,2]=(-0.836768056501)*x[2]
        arg[1,4,2,0]=(-0.0400898232224)*x[0]
        arg[1,4,2,1]=(0.0358172759009)*x[1]
        arg[1,4,2,2]=(-0.335751455408)*x[2]
        arg[2,0,0,0]=(-0.309992915015)*x[0]
        arg[2,0,0,1]=(-0.721404217867)*x[1]
        arg[2,0,0,2]=(-0.548000635629)*x[2]
        arg[2,0,1,0]=(0.651175831531)*x[0]
        arg[2,0,1,1]=(0.158960783491)*x[1]
        arg[2,0,1,2]=(-0.310676926155)*x[2]
        arg[2,0,2,0]=(-0.122289734411)*x[0]
        arg[2,0,2,1]=(-0.252405938421)*x[1]
        arg[2,0,2,2]=(-0.938280244213)*x[2]
        arg[2,1,0,0]=(0.559495801686)*x[0]
        arg[2,1,0,1]=(-0.547182622716)*x[1]
        arg[2,1,0,2]=(0.397441517898)*x[2]
        arg[2,1,1,0]=(-0.406112472071)*x[0]
        arg[2,1,1,1]=(0.355063810677)*x[1]
        arg[2,1,1,2]=(0.760400203215)*x[2]
        arg[2,1,2,0]=(0.992201320481)*x[0]
        arg[2,1,2,1]=(0.0580660882576)*x[1]
        arg[2,1,2,2]=(-0.643170879939)*x[2]
        arg[2,2,0,0]=(-0.280644461832)*x[0]
        arg[2,2,0,1]=(-0.0467430285531)*x[1]
        arg[2,2,0,2]=(0.314050593255)*x[2]
        arg[2,2,1,0]=(-0.230032618609)*x[0]
        arg[2,2,1,1]=(0.0996058698273)*x[1]
        arg[2,2,1,2]=(-0.0270266073208)*x[2]
        arg[2,2,2,0]=(0.767914132956)*x[0]
        arg[2,2,2,1]=(0.496930363612)*x[1]
        arg[2,2,2,2]=(-0.599525033616)*x[2]
        arg[2,3,0,0]=(-0.326433376073)*x[0]
        arg[2,3,0,1]=(-0.0366374501025)*x[1]
        arg[2,3,0,2]=(0.22555705749)*x[2]
        arg[2,3,1,0]=(-0.162548813895)*x[0]
        arg[2,3,1,1]=(-0.110074212194)*x[1]
        arg[2,3,1,2]=(-0.143600895553)*x[2]
        arg[2,3,2,0]=(0.771148880174)*x[0]
        arg[2,3,2,1]=(0.112528116552)*x[1]
        arg[2,3,2,2]=(-0.955735294341)*x[2]
        arg[2,4,0,0]=(-0.968392951034)*x[0]
        arg[2,4,0,1]=(-0.36901708507)*x[1]
        arg[2,4,0,2]=(0.283692515492)*x[2]
        arg[2,4,1,0]=(0.997238032837)*x[0]
        arg[2,4,1,1]=(-0.625794124653)*x[1]
        arg[2,4,1,2]=(0.533386027556)*x[2]
        arg[2,4,2,0]=(0.977311695557)*x[0]
        arg[2,4,2,1]=(0.693009976689)*x[1]
        arg[2,4,2,2]=(0.711179347652)*x[2]
        arg[3,0,0,0]=(-0.155585788931)*x[0]
        arg[3,0,0,1]=(0.0228078851234)*x[1]
        arg[3,0,0,2]=(0.510104938032)*x[2]
        arg[3,0,1,0]=(0.74865995369)*x[0]
        arg[3,0,1,1]=(0.672153736284)*x[1]
        arg[3,0,1,2]=(0.588012355098)*x[2]
        arg[3,0,2,0]=(-0.924508475715)*x[0]
        arg[3,0,2,1]=(-0.392784674758)*x[1]
        arg[3,0,2,2]=(-0.36371454642)*x[2]
        arg[3,1,0,0]=(-0.709783490337)*x[0]
        arg[3,1,0,1]=(0.844136172222)*x[1]
        arg[3,1,0,2]=(0.621011730043)*x[2]
        arg[3,1,1,0]=(0.428807337181)*x[0]
        arg[3,1,1,1]=(0.126300214574)*x[1]
        arg[3,1,1,2]=(0.795972806221)*x[2]
        arg[3,1,2,0]=(-0.252334324004)*x[0]
        arg[3,1,2,1]=(-0.722829467938)*x[1]
        arg[3,1,2,2]=(-0.551540062366)*x[2]
        arg[3,2,0,0]=(-0.134668475963)*x[0]
        arg[3,2,0,1]=(-0.598747540536)*x[1]
        arg[3,2,0,2]=(0.426422436624)*x[2]
        arg[3,2,1,0]=(-0.363050323762)*x[0]
        arg[3,2,1,1]=(0.980891457977)*x[1]
        arg[3,2,1,2]=(0.162831912555)*x[2]
        arg[3,2,2,0]=(-0.126505493475)*x[0]
        arg[3,2,2,1]=(-0.578567864811)*x[1]
        arg[3,2,2,2]=(-0.509843129095)*x[2]
        arg[3,3,0,0]=(-0.446171262265)*x[0]
        arg[3,3,0,1]=(-0.715175197494)*x[1]
        arg[3,3,0,2]=(-0.881016888806)*x[2]
        arg[3,3,1,0]=(-0.942020866327)*x[0]
        arg[3,3,1,1]=(0.156434646828)*x[1]
        arg[3,3,1,2]=(0.523624761583)*x[2]
        arg[3,3,2,0]=(-0.683550923926)*x[0]
        arg[3,3,2,1]=(0.857075218033)*x[1]
        arg[3,3,2,2]=(0.297672594023)*x[2]
        arg[3,4,0,0]=(0.74317121113)*x[0]
        arg[3,4,0,1]=(0.076464540756)*x[1]
        arg[3,4,0,2]=(0.781965468281)*x[2]
        arg[3,4,1,0]=(0.417750169098)*x[0]
        arg[3,4,1,1]=(0.82275428729)*x[1]
        arg[3,4,1,2]=(0.919072321093)*x[2]
        arg[3,4,2,0]=(-0.0246706472217)*x[0]
        arg[3,4,2,1]=(0.179863245513)*x[1]
        arg[3,4,2,2]=(0.539115287766)*x[2]
        ref=sqrt((52.492676775)*(2.*dim+1.)/3.)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromSymbol_rank0(self):
      """
      tests L2-norm of Symbol on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(0.0248400067683)*x[0]
        ref=sqrt((0.000617025936249)*(2.*dim+1.)/3.)

      else:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(-0.48069197083)*x[0]
        ref=sqrt((0.23106477082)*(2.*dim+1.)/3.)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromSymbol_rank1(self):
      """
      tests L2-norm of Symbol on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(2,),dim=dim)
        s=Data(0,(2,),w)
        s[0]=(0.541099842991)*x[0]
        s[1]=(0.123212139947)*x[1]
        ref=sqrt((0.307970271515)*(2.*dim+1.)/3.)

      else:
        arg=Symbol(shape=(3,),dim=dim)
        s=Data(0,(3,),w)
        s[0]=(-0.74011573744)*x[0]
        s[1]=(0.63543826621)*x[1]
        s[2]=(-0.210366750156)*x[2]
        ref=sqrt((0.995807264541)*(2.*dim+1.)/3.)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromSymbol_rank2(self):
      """
      tests L2-norm of Symbol on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 2),dim=dim)
        s=Data(0,(4, 2),w)
        s[0,0]=(-0.324596599761)*x[0]
        s[0,1]=(0.456181444086)*x[1]
        s[1,0]=(-0.85366968746)*x[0]
        s[1,1]=(-0.984222055109)*x[1]
        s[2,0]=(0.302615209563)*x[0]
        s[2,1]=(-0.259853865003)*x[1]
        s[3,0]=(-0.2851901062)*x[0]
        s[3,1]=(0.790097073483)*x[1]
        ref=sqrt((2.87559622997)*(2.*dim+1.)/3.)

      else:
        arg=Symbol(shape=(4, 3),dim=dim)
        s=Data(0,(4, 3),w)
        s[0,0]=(0.806171559469)*x[0]
        s[0,1]=(0.734186538041)*x[1]
        s[0,2]=(0.376885565623)*x[2]
        s[1,0]=(-0.168339360861)*x[0]
        s[1,1]=(-0.184538096909)*x[1]
        s[1,2]=(-0.649204637511)*x[2]
        s[2,0]=(0.0187762580033)*x[0]
        s[2,1]=(-0.851287373276)*x[1]
        s[2,2]=(0.84351620048)*x[2]
        s[3,0]=(0.759367087847)*x[0]
        s[3,1]=(-0.418326808815)*x[1]
        s[3,2]=(0.207925273666)*x[2]
        ref=sqrt((4.04627522925)*(2.*dim+1.)/3.)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromSymbol_rank3(self):
      """
      tests L2-norm of Symbol on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(6, 2, 2),dim=dim)
        s=Data(0,(6, 2, 2),w)
        s[0,0,0]=(0.656602395169)*x[0]
        s[0,0,1]=(-0.153988287938)*x[1]
        s[0,1,0]=(-0.911070962892)*x[0]
        s[0,1,1]=(0.734617382897)*x[1]
        s[1,0,0]=(0.708777892683)*x[0]
        s[1,0,1]=(0.407403464363)*x[1]
        s[1,1,0]=(-0.73004005055)*x[0]
        s[1,1,1]=(0.108502620047)*x[1]
        s[2,0,0]=(0.584913648885)*x[0]
        s[2,0,1]=(0.655698697233)*x[1]
        s[2,1,0]=(0.504301542359)*x[0]
        s[2,1,1]=(-0.637819233396)*x[1]
        s[3,0,0]=(0.395391847846)*x[0]
        s[3,0,1]=(-0.136070177387)*x[1]
        s[3,1,0]=(0.77444662439)*x[0]
        s[3,1,1]=(0.245316754174)*x[1]
        s[4,0,0]=(0.542716832119)*x[0]
        s[4,0,1]=(0.0202895164815)*x[1]
        s[4,1,0]=(0.922586929777)*x[0]
        s[4,1,1]=(-0.948487993154)*x[1]
        s[5,0,0]=(0.629779082528)*x[0]
        s[5,0,1]=(-0.90719864412)*x[1]
        s[5,1,0]=(-0.29521666306)*x[0]
        s[5,1,1]=(-0.431748113769)*x[1]
        ref=sqrt((8.84456266856)*(2.*dim+1.)/3.)

      else:
        arg=Symbol(shape=(6, 2, 3),dim=dim)
        s=Data(0,(6, 2, 3),w)
        s[0,0,0]=(0.529698924744)*x[0]
        s[0,0,1]=(0.779440018368)*x[1]
        s[0,0,2]=(-0.447704804318)*x[2]
        s[0,1,0]=(0.375433500828)*x[0]
        s[0,1,1]=(-0.567989496415)*x[1]
        s[0,1,2]=(-0.47133957857)*x[2]
        s[1,0,0]=(0.379156720923)*x[0]
        s[1,0,1]=(-0.629670565361)*x[1]
        s[1,0,2]=(0.814298387366)*x[2]
        s[1,1,0]=(-0.482876103672)*x[0]
        s[1,1,1]=(-0.592362629566)*x[1]
        s[1,1,2]=(0.368859438611)*x[2]
        s[2,0,0]=(0.145060478226)*x[0]
        s[2,0,1]=(-0.582989521856)*x[1]
        s[2,0,2]=(-0.977335802848)*x[2]
        s[2,1,0]=(-0.471820598698)*x[0]
        s[2,1,1]=(-0.680071585031)*x[1]
        s[2,1,2]=(-0.686102402643)*x[2]
        s[3,0,0]=(-0.647535125154)*x[0]
        s[3,0,1]=(-0.739392553517)*x[1]
        s[3,0,2]=(-0.0852792216585)*x[2]
        s[3,1,0]=(-0.0197635627728)*x[0]
        s[3,1,1]=(-0.811782244587)*x[1]
        s[3,1,2]=(0.842974983966)*x[2]
        s[4,0,0]=(-0.765665515246)*x[0]
        s[4,0,1]=(-0.379368211948)*x[1]
        s[4,0,2]=(-0.864157207524)*x[2]
        s[4,1,0]=(-0.393989909405)*x[0]
        s[4,1,1]=(0.482309706914)*x[1]
        s[4,1,2]=(-0.80068507467)*x[2]
        s[5,0,0]=(-0.686334218925)*x[0]
        s[5,0,1]=(-0.414623392196)*x[1]
        s[5,0,2]=(-0.689227675292)*x[2]
        s[5,1,0]=(0.347067831012)*x[0]
        s[5,1,1]=(0.675600902408)*x[1]
        s[5,1,2]=(-0.763467841004)*x[2]
        ref=sqrt((13.2965907789)*(2.*dim+1.)/3.)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onFunctionOnBoundary_fromSymbol_rank4(self):
      """
      tests L2-norm of Symbol on the FunctionOnBoundary

      assumptions: self.domain supports integration on FunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=FunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
        s=Data(0,(4, 5, 3, 2),w)
        s[0,0,0,0]=(0.478851101914)*x[0]
        s[0,0,0,1]=(-0.669115623243)*x[1]
        s[0,0,1,0]=(0.169431656457)*x[0]
        s[0,0,1,1]=(-0.154130248518)*x[1]
        s[0,0,2,0]=(0.590784906322)*x[0]
        s[0,0,2,1]=(-0.843193872964)*x[1]
        s[0,1,0,0]=(0.040548638805)*x[0]
        s[0,1,0,1]=(0.0223688163494)*x[1]
        s[0,1,1,0]=(0.591094685655)*x[0]
        s[0,1,1,1]=(0.692105812582)*x[1]
        s[0,1,2,0]=(-0.326832256739)*x[0]
        s[0,1,2,1]=(0.824824822029)*x[1]
        s[0,2,0,0]=(-0.352757110519)*x[0]
        s[0,2,0,1]=(-0.898616847036)*x[1]
        s[0,2,1,0]=(0.972018635849)*x[0]
        s[0,2,1,1]=(0.64223146835)*x[1]
        s[0,2,2,0]=(0.599638014876)*x[0]
        s[0,2,2,1]=(0.665927987689)*x[1]
        s[0,3,0,0]=(-0.757282087611)*x[0]
        s[0,3,0,1]=(-0.541306038159)*x[1]
        s[0,3,1,0]=(0.443494146532)*x[0]
        s[0,3,1,1]=(-0.125404019651)*x[1]
        s[0,3,2,0]=(0.234422087306)*x[0]
        s[0,3,2,1]=(0.443982420301)*x[1]
        s[0,4,0,0]=(0.46454740036)*x[0]
        s[0,4,0,1]=(-0.479031824057)*x[1]
        s[0,4,1,0]=(-0.304056971991)*x[0]
        s[0,4,1,1]=(-0.763293280009)*x[1]
        s[0,4,2,0]=(-0.142796685199)*x[0]
        s[0,4,2,1]=(0.301022025426)*x[1]
        s[1,0,0,0]=(-0.843399285738)*x[0]
        s[1,0,0,1]=(0.639837980509)*x[1]
        s[1,0,1,0]=(-0.649024291296)*x[0]
        s[1,0,1,1]=(0.645960995308)*x[1]
        s[1,0,2,0]=(0.0862852468825)*x[0]
        s[1,0,2,1]=(-0.399237566917)*x[1]
        s[1,1,0,0]=(0.959425187971)*x[0]
        s[1,1,0,1]=(0.65407910284)*x[1]
        s[1,1,1,0]=(0.873927579655)*x[0]
        s[1,1,1,1]=(0.399448652787)*x[1]
        s[1,1,2,0]=(-0.363863298937)*x[0]
        s[1,1,2,1]=(0.555310285452)*x[1]
        s[1,2,0,0]=(0.962712099815)*x[0]
        s[1,2,0,1]=(0.141495259052)*x[1]
        s[1,2,1,0]=(-0.819117296209)*x[0]
        s[1,2,1,1]=(-0.752934712352)*x[1]
        s[1,2,2,0]=(0.065886700262)*x[0]
        s[1,2,2,1]=(-0.530225381274)*x[1]
        s[1,3,0,0]=(0.830686848377)*x[0]
        s[1,3,0,1]=(0.139854694845)*x[1]
        s[1,3,1,0]=(0.869755726605)*x[0]
        s[1,3,1,1]=(0.145889500911)*x[1]
        s[1,3,2,0]=(-0.690283765486)*x[0]
        s[1,3,2,1]=(-0.790245562042)*x[1]
        s[1,4,0,0]=(0.704042980492)*x[0]
        s[1,4,0,1]=(-0.692268480039)*x[1]
        s[1,4,1,0]=(-0.708898941564)*x[0]
        s[1,4,1,1]=(-0.675824414677)*x[1]
        s[1,4,2,0]=(0.320394083261)*x[0]
        s[1,4,2,1]=(0.173795411968)*x[1]
        s[2,0,0,0]=(0.707844986267)*x[0]
        s[2,0,0,1]=(0.991847131293)*x[1]
        s[2,0,1,0]=(0.470933752596)*x[0]
        s[2,0,1,1]=(0.847427404272)*x[1]
        s[2,0,2,0]=(0.383056136263)*x[0]
        s[2,0,2,1]=(-0.717411302782)*x[1]
        s[2,1,0,0]=(0.118326527818)*x[0]
        s[2,1,0,1]=(0.0957434761651)*x[1]
        s[2,1,1,0]=(-0.420654904504)*x[0]
        s[2,1,1,1]=(0.31903393188)*x[1]
        s[2,1,2,0]=(0.0597937431476)*x[0]
        s[2,1,2,1]=(0.965201628622)*x[1]
        s[2,2,0,0]=(0.884709108482)*x[0]
        s[2,2,0,1]=(-0.541160500366)*x[1]
        s[2,2,1,0]=(-0.180788996448)*x[0]
        s[2,2,1,1]=(-0.44537512426)*x[1]
        s[2,2,2,0]=(-0.291513668432)*x[0]
        s[2,2,2,1]=(0.349090737561)*x[1]
        s[2,3,0,0]=(0.0994765186043)*x[0]
        s[2,3,0,1]=(-0.317258869646)*x[1]
        s[2,3,1,0]=(0.661005398598)*x[0]
        s[2,3,1,1]=(-0.914254206568)*x[1]
        s[2,3,2,0]=(-0.374142249366)*x[0]
        s[2,3,2,1]=(-0.587651938713)*x[1]
        s[2,4,0,0]=(-0.0990354726919)*x[0]
        s[2,4,0,1]=(-0.104596309056)*x[1]
        s[2,4,1,0]=(0.183201845594)*x[0]
        s[2,4,1,1]=(-0.265185815743)*x[1]
        s[2,4,2,0]=(0.804060169933)*x[0]
        s[2,4,2,1]=(-0.0798336799535)*x[1]
        s[3,0,0,0]=(-0.470265711106)*x[0]
        s[3,0,0,1]=(0.123030821016)*x[1]
        s[3,0,1,0]=(0.971878570072)*x[0]
        s[3,0,1,1]=(0.557374437558)*x[1]
        s[3,0,2,0]=(0.771881325143)*x[0]
        s[3,0,2,1]=(-0.181917882935)*x[1]
        s[3,1,0,0]=(-0.30330086877)*x[0]
        s[3,1,0,1]=(0.785561830696)*x[1]
        s[3,1,1,0]=(0.0783160883668)*x[0]
        s[3,1,1,1]=(-0.192122642565)*x[1]
        s[3,1,2,0]=(-0.294486881116)*x[0]
        s[3,1,2,1]=(0.533158403133)*x[1]
        s[3,2,0,0]=(-0.413737404117)*x[0]
        s[3,2,0,1]=(0.622363162902)*x[1]
        s[3,2,1,0]=(0.241741847782)*x[0]
        s[3,2,1,1]=(-0.0541594558102)*x[1]
        s[3,2,2,0]=(0.157759963106)*x[0]
        s[3,2,2,1]=(-0.833376953416)*x[1]
        s[3,3,0,0]=(-0.671630074873)*x[0]
        s[3,3,0,1]=(-0.10840086864)*x[1]
        s[3,3,1,0]=(-0.995501368513)*x[0]
        s[3,3,1,1]=(-0.153147007921)*x[1]
        s[3,3,2,0]=(-0.504815129335)*x[0]
        s[3,3,2,1]=(-0.489511115245)*x[1]
        s[3,4,0,0]=(0.206555206064)*x[0]
        s[3,4,0,1]=(-0.294058867178)*x[1]
        s[3,4,1,0]=(-0.571176824119)*x[0]
        s[3,4,1,1]=(0.353847783519)*x[1]
        s[3,4,2,0]=(-0.441012644459)*x[0]
        s[3,4,2,1]=(0.198182379551)*x[1]
        ref=sqrt((36.8775562848)*(2.*dim+1.)/3.)

      else:
        arg=Symbol(shape=(4, 5, 3, 3),dim=dim)
        s=Data(0,(4, 5, 3, 3),w)
        s[0,0,0,0]=(-0.391342187918)*x[0]
        s[0,0,0,1]=(-0.807670030995)*x[1]
        s[0,0,0,2]=(-0.730776810158)*x[2]
        s[0,0,1,0]=(-0.640249702579)*x[0]
        s[0,0,1,1]=(-0.0944608864404)*x[1]
        s[0,0,1,2]=(0.852922834544)*x[2]
        s[0,0,2,0]=(0.0996848239317)*x[0]
        s[0,0,2,1]=(-0.323310001677)*x[1]
        s[0,0,2,2]=(-0.311089889753)*x[2]
        s[0,1,0,0]=(-0.294816251058)*x[0]
        s[0,1,0,1]=(0.939389998027)*x[1]
        s[0,1,0,2]=(0.900617063513)*x[2]
        s[0,1,1,0]=(-0.961986583089)*x[0]
        s[0,1,1,1]=(0.629473020335)*x[1]
        s[0,1,1,2]=(-0.215446848397)*x[2]
        s[0,1,2,0]=(-0.256062202737)*x[0]
        s[0,1,2,1]=(0.497231616486)*x[1]
        s[0,1,2,2]=(-0.813869122359)*x[2]
        s[0,2,0,0]=(0.352147663827)*x[0]
        s[0,2,0,1]=(0.553022457414)*x[1]
        s[0,2,0,2]=(-0.340077213113)*x[2]
        s[0,2,1,0]=(0.654397833915)*x[0]
        s[0,2,1,1]=(-0.34308640744)*x[1]
        s[0,2,1,2]=(-0.689126280824)*x[2]
        s[0,2,2,0]=(0.428183196531)*x[0]
        s[0,2,2,1]=(-0.322384435092)*x[1]
        s[0,2,2,2]=(0.636495513573)*x[2]
        s[0,3,0,0]=(0.214083660816)*x[0]
        s[0,3,0,1]=(0.312445001293)*x[1]
        s[0,3,0,2]=(-0.939086993471)*x[2]
        s[0,3,1,0]=(-0.0668860159236)*x[0]
        s[0,3,1,1]=(0.591012248354)*x[1]
        s[0,3,1,2]=(0.696149816619)*x[2]
        s[0,3,2,0]=(0.917050906357)*x[0]
        s[0,3,2,1]=(-0.695398022945)*x[1]
        s[0,3,2,2]=(0.578841930671)*x[2]
        s[0,4,0,0]=(-0.321769393495)*x[0]
        s[0,4,0,1]=(-0.761161601481)*x[1]
        s[0,4,0,2]=(0.000116192661055)*x[2]
        s[0,4,1,0]=(0.347135267821)*x[0]
        s[0,4,1,1]=(-0.0105215354141)*x[1]
        s[0,4,1,2]=(0.582827998159)*x[2]
        s[0,4,2,0]=(0.765317199582)*x[0]
        s[0,4,2,1]=(0.546093510978)*x[1]
        s[0,4,2,2]=(-0.753832119978)*x[2]
        s[1,0,0,0]=(-0.163230444043)*x[0]
        s[1,0,0,1]=(0.415060470539)*x[1]
        s[1,0,0,2]=(0.638843786094)*x[2]
        s[1,0,1,0]=(-0.320730240187)*x[0]
        s[1,0,1,1]=(-0.369428894975)*x[1]
        s[1,0,1,2]=(-0.944414569234)*x[2]
        s[1,0,2,0]=(0.580606041923)*x[0]
        s[1,0,2,1]=(-0.351082080131)*x[1]
        s[1,0,2,2]=(-0.170588782574)*x[2]
        s[1,1,0,0]=(0.323136286071)*x[0]
        s[1,1,0,1]=(0.397263600918)*x[1]
        s[1,1,0,2]=(-0.687508659585)*x[2]
        s[1,1,1,0]=(-0.997412227958)*x[0]
        s[1,1,1,1]=(0.968968702091)*x[1]
        s[1,1,1,2]=(-0.223523055076)*x[2]
        s[1,1,2,0]=(-0.495473193251)*x[0]
        s[1,1,2,1]=(-0.00758436986969)*x[1]
        s[1,1,2,2]=(0.0770020818673)*x[2]
        s[1,2,0,0]=(-0.522152049391)*x[0]
        s[1,2,0,1]=(0.820143671384)*x[1]
        s[1,2,0,2]=(0.672347144349)*x[2]
        s[1,2,1,0]=(-0.868655459375)*x[0]
        s[1,2,1,1]=(-0.698503510192)*x[1]
        s[1,2,1,2]=(0.16866148235)*x[2]
        s[1,2,2,0]=(0.291645089215)*x[0]
        s[1,2,2,1]=(0.375705968271)*x[1]
        s[1,2,2,2]=(0.047575084848)*x[2]
        s[1,3,0,0]=(-0.856292393733)*x[0]
        s[1,3,0,1]=(0.741997693375)*x[1]
        s[1,3,0,2]=(0.922057641142)*x[2]
        s[1,3,1,0]=(-0.0246230217779)*x[0]
        s[1,3,1,1]=(0.233511225618)*x[1]
        s[1,3,1,2]=(-0.128732881848)*x[2]
        s[1,3,2,0]=(0.437273172814)*x[0]
        s[1,3,2,1]=(-0.24673804436)*x[1]
        s[1,3,2,2]=(0.834689693971)*x[2]
        s[1,4,0,0]=(0.0567731436429)*x[0]
        s[1,4,0,1]=(-0.523439015055)*x[1]
        s[1,4,0,2]=(-0.175105109157)*x[2]
        s[1,4,1,0]=(-0.783819538589)*x[0]
        s[1,4,1,1]=(0.722409822905)*x[1]
        s[1,4,1,2]=(-0.99480794117)*x[2]
        s[1,4,2,0]=(-0.0948610680733)*x[0]
        s[1,4,2,1]=(-0.0934034446625)*x[1]
        s[1,4,2,2]=(0.993222484498)*x[2]
        s[2,0,0,0]=(-0.739506857296)*x[0]
        s[2,0,0,1]=(-0.402129626454)*x[1]
        s[2,0,0,2]=(-0.131164073645)*x[2]
        s[2,0,1,0]=(-0.185068425131)*x[0]
        s[2,0,1,1]=(-0.888669413054)*x[1]
        s[2,0,1,2]=(0.149638740783)*x[2]
        s[2,0,2,0]=(-0.977520384)*x[0]
        s[2,0,2,1]=(0.495784511707)*x[1]
        s[2,0,2,2]=(0.140779930302)*x[2]
        s[2,1,0,0]=(-0.668882409761)*x[0]
        s[2,1,0,1]=(0.87881226199)*x[1]
        s[2,1,0,2]=(0.596921491285)*x[2]
        s[2,1,1,0]=(0.327990184264)*x[0]
        s[2,1,1,1]=(-0.842310366654)*x[1]
        s[2,1,1,2]=(-0.104630179971)*x[2]
        s[2,1,2,0]=(-0.997782736966)*x[0]
        s[2,1,2,1]=(0.362490263673)*x[1]
        s[2,1,2,2]=(-0.919527812431)*x[2]
        s[2,2,0,0]=(-0.694141113242)*x[0]
        s[2,2,0,1]=(-0.189176840568)*x[1]
        s[2,2,0,2]=(-0.305307574884)*x[2]
        s[2,2,1,0]=(0.97520770841)*x[0]
        s[2,2,1,1]=(0.709672831962)*x[1]
        s[2,2,1,2]=(-0.056408975067)*x[2]
        s[2,2,2,0]=(-0.491153842737)*x[0]
        s[2,2,2,1]=(-0.487533744248)*x[1]
        s[2,2,2,2]=(-0.16258274458)*x[2]
        s[2,3,0,0]=(0.69721543109)*x[0]
        s[2,3,0,1]=(-0.927116706183)*x[1]
        s[2,3,0,2]=(-0.277453826622)*x[2]
        s[2,3,1,0]=(-0.530246149296)*x[0]
        s[2,3,1,1]=(0.971840533777)*x[1]
        s[2,3,1,2]=(-0.472599623799)*x[2]
        s[2,3,2,0]=(0.772149184332)*x[0]
        s[2,3,2,1]=(-0.644107019603)*x[1]
        s[2,3,2,2]=(-0.537937585926)*x[2]
        s[2,4,0,0]=(-0.786226779015)*x[0]
        s[2,4,0,1]=(-0.132798607299)*x[1]
        s[2,4,0,2]=(-0.954772620074)*x[2]
        s[2,4,1,0]=(-0.378657907627)*x[0]
        s[2,4,1,1]=(0.896332243914)*x[1]
        s[2,4,1,2]=(-0.980811837061)*x[2]
        s[2,4,2,0]=(-0.033610987457)*x[0]
        s[2,4,2,1]=(0.301124520167)*x[1]
        s[2,4,2,2]=(0.505461084966)*x[2]
        s[3,0,0,0]=(0.92844681553)*x[0]
        s[3,0,0,1]=(-0.833141539681)*x[1]
        s[3,0,0,2]=(-0.219378232103)*x[2]
        s[3,0,1,0]=(-0.342722442206)*x[0]
        s[3,0,1,1]=(0.662705183551)*x[1]
        s[3,0,1,2]=(0.544702876622)*x[2]
        s[3,0,2,0]=(-0.984992104327)*x[0]
        s[3,0,2,1]=(-0.118846731)*x[1]
        s[3,0,2,2]=(-0.850203016683)*x[2]
        s[3,1,0,0]=(0.797626533525)*x[0]
        s[3,1,0,1]=(-0.415519060262)*x[1]
        s[3,1,0,2]=(-0.629491781709)*x[2]
        s[3,1,1,0]=(0.223213637134)*x[0]
        s[3,1,1,1]=(0.824993414773)*x[1]
        s[3,1,1,2]=(0.141523286845)*x[2]
        s[3,1,2,0]=(-0.66436462404)*x[0]
        s[3,1,2,1]=(-0.854989422144)*x[1]
        s[3,1,2,2]=(-0.0354537218933)*x[2]
        s[3,2,0,0]=(-0.0988394772118)*x[0]
        s[3,2,0,1]=(0.125214003041)*x[1]
        s[3,2,0,2]=(-0.701121659493)*x[2]
        s[3,2,1,0]=(-0.603538097381)*x[0]
        s[3,2,1,1]=(-0.659139390328)*x[1]
        s[3,2,1,2]=(-0.849986094805)*x[2]
        s[3,2,2,0]=(-0.545364489526)*x[0]
        s[3,2,2,1]=(-0.90842178441)*x[1]
        s[3,2,2,2]=(-0.650251007613)*x[2]
        s[3,3,0,0]=(0.943460211337)*x[0]
        s[3,3,0,1]=(-0.707864979205)*x[1]
        s[3,3,0,2]=(0.9885932761)*x[2]
        s[3,3,1,0]=(0.127650934393)*x[0]
        s[3,3,1,1]=(-0.495451047906)*x[1]
        s[3,3,1,2]=(0.641823757246)*x[2]
        s[3,3,2,0]=(-0.33331510031)*x[0]
        s[3,3,2,1]=(0.922346384165)*x[1]
        s[3,3,2,2]=(-0.987453834808)*x[2]
        s[3,4,0,0]=(-0.376485044284)*x[0]
        s[3,4,0,1]=(0.693296795267)*x[1]
        s[3,4,0,2]=(-0.31498404544)*x[2]
        s[3,4,1,0]=(0.512477956635)*x[0]
        s[3,4,1,1]=(-0.333819138285)*x[1]
        s[3,4,1,2]=(0.560322373773)*x[2]
        s[3,4,2,0]=(0.889090668029)*x[0]
        s[3,4,2,1]=(0.981814634684)*x[1]
        s[3,4,2,2]=(-0.41272129735)*x[2]
        ref=sqrt((66.7960236856)*(2.*dim+1.)/3.)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")

   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank0(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(),w)
        arg=(-0.245574919477)*sqrt(x[0])
        ref=sqrt((0.0603070410759)*dim)

      else:
        arg=Data(0,(),w)
        arg=(0.757324521515)*sqrt(x[0])
        ref=sqrt((0.573540430888)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank1(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(2,),w)
        arg[0]=(0.723421565407)*sqrt(x[0])
        arg[1]=(-0.460477393103)*sqrt(x[1])
        ref=sqrt((0.735378190855)*dim)

      else:
        arg=Data(0,(3,),w)
        arg[0]=(-0.88528497163)*sqrt(x[0])
        arg[1]=(-0.65510214636)*sqrt(x[1])
        arg[2]=(0.399538866363)*sqrt(x[2])
        ref=sqrt((1.37251960889)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank2(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 2),w)
        arg[0,0]=(0.955620993904)*sqrt(x[0])
        arg[0,1]=(-0.0987865813703)*sqrt(x[1])
        arg[1,0]=(0.0288267231531)*sqrt(x[0])
        arg[1,1]=(0.655440599879)*sqrt(x[1])
        arg[2,0]=(0.685627284533)*sqrt(x[0])
        arg[2,1]=(-0.989832824892)*sqrt(x[1])
        arg[3,0]=(0.292184093194)*sqrt(x[0])
        arg[3,1]=(0.149553857773)*sqrt(x[1])
        ref=sqrt((2.91099532781)*dim)

      else:
        arg=Data(0,(4, 3),w)
        arg[0,0]=(-0.325908541533)*sqrt(x[0])
        arg[0,1]=(-0.992480479749)*sqrt(x[1])
        arg[0,2]=(0.660360271799)*sqrt(x[2])
        arg[1,0]=(0.173485908581)*sqrt(x[0])
        arg[1,1]=(-0.328755199781)*sqrt(x[1])
        arg[1,2]=(-0.943354674948)*sqrt(x[2])
        arg[2,0]=(0.680713222646)*sqrt(x[0])
        arg[2,1]=(-0.765971835693)*sqrt(x[1])
        arg[2,2]=(0.0413284847528)*sqrt(x[2])
        arg[3,0]=(0.990074004708)*sqrt(x[0])
        arg[3,1]=(0.941801786766)*sqrt(x[1])
        arg[3,2]=(0.886926192201)*sqrt(x[2])
        ref=sqrt((6.26107155228)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank3(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(6, 2, 2),w)
        arg[0,0,0]=(-0.0781611551598)*sqrt(x[0])
        arg[0,0,1]=(0.291016249575)*sqrt(x[1])
        arg[0,1,0]=(-0.107555233086)*sqrt(x[0])
        arg[0,1,1]=(-0.559067108546)*sqrt(x[1])
        arg[1,0,0]=(-0.0818406701266)*sqrt(x[0])
        arg[1,0,1]=(-0.594866806483)*sqrt(x[1])
        arg[1,1,0]=(-0.725814803863)*sqrt(x[0])
        arg[1,1,1]=(0.59128101992)*sqrt(x[1])
        arg[2,0,0]=(-0.15381555291)*sqrt(x[0])
        arg[2,0,1]=(-0.679882948503)*sqrt(x[1])
        arg[2,1,0]=(-0.58437193917)*sqrt(x[0])
        arg[2,1,1]=(0.136304615849)*sqrt(x[1])
        arg[3,0,0]=(0.0671365410096)*sqrt(x[0])
        arg[3,0,1]=(-0.645687212187)*sqrt(x[1])
        arg[3,1,0]=(-0.642492412392)*sqrt(x[0])
        arg[3,1,1]=(-0.125760054735)*sqrt(x[1])
        arg[4,0,0]=(0.731110824794)*sqrt(x[0])
        arg[4,0,1]=(0.491668422979)*sqrt(x[1])
        arg[4,1,0]=(-0.775841478292)*sqrt(x[0])
        arg[4,1,1]=(0.728265567974)*sqrt(x[1])
        arg[5,0,0]=(0.84511832373)*sqrt(x[0])
        arg[5,0,1]=(-0.513796801068)*sqrt(x[1])
        arg[5,1,0]=(0.113072243554)*sqrt(x[0])
        arg[5,1,1]=(0.246630838744)*sqrt(x[1])
        ref=sqrt((6.30829536252)*dim)

      else:
        arg=Data(0,(6, 2, 3),w)
        arg[0,0,0]=(0.369748116859)*sqrt(x[0])
        arg[0,0,1]=(-0.758056560031)*sqrt(x[1])
        arg[0,0,2]=(-0.873984709951)*sqrt(x[2])
        arg[0,1,0]=(0.311680165784)*sqrt(x[0])
        arg[0,1,1]=(0.374400673651)*sqrt(x[1])
        arg[0,1,2]=(0.712484217076)*sqrt(x[2])
        arg[1,0,0]=(0.829379714484)*sqrt(x[0])
        arg[1,0,1]=(-0.0551589596149)*sqrt(x[1])
        arg[1,0,2]=(0.965672208426)*sqrt(x[2])
        arg[1,1,0]=(-0.205044281547)*sqrt(x[0])
        arg[1,1,1]=(0.238197452756)*sqrt(x[1])
        arg[1,1,2]=(-0.33456139292)*sqrt(x[2])
        arg[2,0,0]=(0.649928288926)*sqrt(x[0])
        arg[2,0,1]=(-0.661384953389)*sqrt(x[1])
        arg[2,0,2]=(-0.253241222975)*sqrt(x[2])
        arg[2,1,0]=(-0.491716575992)*sqrt(x[0])
        arg[2,1,1]=(-0.970872527468)*sqrt(x[1])
        arg[2,1,2]=(0.222410198921)*sqrt(x[2])
        arg[3,0,0]=(0.205752630262)*sqrt(x[0])
        arg[3,0,1]=(0.864804362697)*sqrt(x[1])
        arg[3,0,2]=(-0.417975564033)*sqrt(x[2])
        arg[3,1,0]=(0.586425694033)*sqrt(x[0])
        arg[3,1,1]=(0.952661122184)*sqrt(x[1])
        arg[3,1,2]=(0.608680080453)*sqrt(x[2])
        arg[4,0,0]=(0.625968903369)*sqrt(x[0])
        arg[4,0,1]=(-0.573909405003)*sqrt(x[1])
        arg[4,0,2]=(-0.762256394595)*sqrt(x[2])
        arg[4,1,0]=(0.0710742394418)*sqrt(x[0])
        arg[4,1,1]=(0.583378040574)*sqrt(x[1])
        arg[4,1,2]=(0.719032893115)*sqrt(x[2])
        arg[5,0,0]=(0.032173368884)*sqrt(x[0])
        arg[5,0,1]=(-0.434042549492)*sqrt(x[1])
        arg[5,0,2]=(0.363504765447)*sqrt(x[2])
        arg[5,1,0]=(0.598817469198)*sqrt(x[0])
        arg[5,1,1]=(-0.163967008775)*sqrt(x[1])
        arg[5,1,2]=(0.546778730604)*sqrt(x[2])
        ref=sqrt((11.9696343123)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromData_rank4(self):
      """
      tests L2-norm of Data on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Data(0,(4, 5, 3, 2),w)
        arg[0,0,0,0]=(-0.273446520069)*sqrt(x[0])
        arg[0,0,0,1]=(-0.913305910831)*sqrt(x[1])
        arg[0,0,1,0]=(0.0745566747537)*sqrt(x[0])
        arg[0,0,1,1]=(0.98803601919)*sqrt(x[1])
        arg[0,0,2,0]=(-0.244120818875)*sqrt(x[0])
        arg[0,0,2,1]=(0.247509644998)*sqrt(x[1])
        arg[0,1,0,0]=(-0.548756417777)*sqrt(x[0])
        arg[0,1,0,1]=(-0.354587911923)*sqrt(x[1])
        arg[0,1,1,0]=(0.104268198867)*sqrt(x[0])
        arg[0,1,1,1]=(-0.541700877072)*sqrt(x[1])
        arg[0,1,2,0]=(-0.25900060689)*sqrt(x[0])
        arg[0,1,2,1]=(-0.859660175231)*sqrt(x[1])
        arg[0,2,0,0]=(0.195235560321)*sqrt(x[0])
        arg[0,2,0,1]=(0.175518738589)*sqrt(x[1])
        arg[0,2,1,0]=(-0.0638854232272)*sqrt(x[0])
        arg[0,2,1,1]=(-0.586161016541)*sqrt(x[1])
        arg[0,2,2,0]=(0.580258892247)*sqrt(x[0])
        arg[0,2,2,1]=(-0.931927145435)*sqrt(x[1])
        arg[0,3,0,0]=(-0.408209600298)*sqrt(x[0])
        arg[0,3,0,1]=(-0.0344882667014)*sqrt(x[1])
        arg[0,3,1,0]=(-0.131763534163)*sqrt(x[0])
        arg[0,3,1,1]=(-0.787653739965)*sqrt(x[1])
        arg[0,3,2,0]=(0.0910808104711)*sqrt(x[0])
        arg[0,3,2,1]=(-0.280409023611)*sqrt(x[1])
        arg[0,4,0,0]=(0.97745754012)*sqrt(x[0])
        arg[0,4,0,1]=(-0.59829020936)*sqrt(x[1])
        arg[0,4,1,0]=(0.890520260543)*sqrt(x[0])
        arg[0,4,1,1]=(-0.600760090231)*sqrt(x[1])
        arg[0,4,2,0]=(-0.897992297974)*sqrt(x[0])
        arg[0,4,2,1]=(-0.841169898923)*sqrt(x[1])
        arg[1,0,0,0]=(-0.249868816409)*sqrt(x[0])
        arg[1,0,0,1]=(0.620375082228)*sqrt(x[1])
        arg[1,0,1,0]=(-0.660480789306)*sqrt(x[0])
        arg[1,0,1,1]=(-0.73638571806)*sqrt(x[1])
        arg[1,0,2,0]=(0.339987316643)*sqrt(x[0])
        arg[1,0,2,1]=(0.541112529894)*sqrt(x[1])
        arg[1,1,0,0]=(-0.468808186705)*sqrt(x[0])
        arg[1,1,0,1]=(-0.32919679792)*sqrt(x[1])
        arg[1,1,1,0]=(0.917292803419)*sqrt(x[0])
        arg[1,1,1,1]=(-0.834265058005)*sqrt(x[1])
        arg[1,1,2,0]=(-0.247536849264)*sqrt(x[0])
        arg[1,1,2,1]=(-0.197503469238)*sqrt(x[1])
        arg[1,2,0,0]=(0.897591919909)*sqrt(x[0])
        arg[1,2,0,1]=(-0.807446231234)*sqrt(x[1])
        arg[1,2,1,0]=(-0.369878499382)*sqrt(x[0])
        arg[1,2,1,1]=(0.985678692179)*sqrt(x[1])
        arg[1,2,2,0]=(-0.709976427525)*sqrt(x[0])
        arg[1,2,2,1]=(-0.368744647016)*sqrt(x[1])
        arg[1,3,0,0]=(0.299630726462)*sqrt(x[0])
        arg[1,3,0,1]=(-0.445295757899)*sqrt(x[1])
        arg[1,3,1,0]=(-0.922386577254)*sqrt(x[0])
        arg[1,3,1,1]=(0.234794853697)*sqrt(x[1])
        arg[1,3,2,0]=(0.953377720197)*sqrt(x[0])
        arg[1,3,2,1]=(0.409778183998)*sqrt(x[1])
        arg[1,4,0,0]=(0.271967945488)*sqrt(x[0])
        arg[1,4,0,1]=(-0.578629001202)*sqrt(x[1])
        arg[1,4,1,0]=(0.210755226769)*sqrt(x[0])
        arg[1,4,1,1]=(-0.0902751419945)*sqrt(x[1])
        arg[1,4,2,0]=(0.70033387381)*sqrt(x[0])
        arg[1,4,2,1]=(0.305733565661)*sqrt(x[1])
        arg[2,0,0,0]=(-0.662483167298)*sqrt(x[0])
        arg[2,0,0,1]=(0.585252048652)*sqrt(x[1])
        arg[2,0,1,0]=(-0.398813785959)*sqrt(x[0])
        arg[2,0,1,1]=(-0.797438697186)*sqrt(x[1])
        arg[2,0,2,0]=(-0.508308971009)*sqrt(x[0])
        arg[2,0,2,1]=(0.302249407524)*sqrt(x[1])
        arg[2,1,0,0]=(0.208644491879)*sqrt(x[0])
        arg[2,1,0,1]=(-0.604749055374)*sqrt(x[1])
        arg[2,1,1,0]=(0.641654284594)*sqrt(x[0])
        arg[2,1,1,1]=(0.456898593356)*sqrt(x[1])
        arg[2,1,2,0]=(-0.398043867778)*sqrt(x[0])
        arg[2,1,2,1]=(-0.0712344657587)*sqrt(x[1])
        arg[2,2,0,0]=(0.0967860954865)*sqrt(x[0])
        arg[2,2,0,1]=(0.520449905952)*sqrt(x[1])
        arg[2,2,1,0]=(0.770198029595)*sqrt(x[0])
        arg[2,2,1,1]=(-0.594004671621)*sqrt(x[1])
        arg[2,2,2,0]=(-0.744571452885)*sqrt(x[0])
        arg[2,2,2,1]=(0.544447367825)*sqrt(x[1])
        arg[2,3,0,0]=(-0.137087966968)*sqrt(x[0])
        arg[2,3,0,1]=(0.120672667497)*sqrt(x[1])
        arg[2,3,1,0]=(0.204800088057)*sqrt(x[0])
        arg[2,3,1,1]=(0.626526346076)*sqrt(x[1])
        arg[2,3,2,0]=(-0.696480393227)*sqrt(x[0])
        arg[2,3,2,1]=(0.188533741996)*sqrt(x[1])
        arg[2,4,0,0]=(-0.403523821067)*sqrt(x[0])
        arg[2,4,0,1]=(-0.428048989483)*sqrt(x[1])
        arg[2,4,1,0]=(-0.244186366584)*sqrt(x[0])
        arg[2,4,1,1]=(0.00866444909003)*sqrt(x[1])
        arg[2,4,2,0]=(-0.445991308853)*sqrt(x[0])
        arg[2,4,2,1]=(-0.899951068935)*sqrt(x[1])
        arg[3,0,0,0]=(0.609340085418)*sqrt(x[0])
        arg[3,0,0,1]=(0.878750391425)*sqrt(x[1])
        arg[3,0,1,0]=(0.258064654464)*sqrt(x[0])
        arg[3,0,1,1]=(-0.482402612985)*sqrt(x[1])
        arg[3,0,2,0]=(0.943732283389)*sqrt(x[0])
        arg[3,0,2,1]=(0.65514211843)*sqrt(x[1])
        arg[3,1,0,0]=(-0.894551979619)*sqrt(x[0])
        arg[3,1,0,1]=(0.220116541042)*sqrt(x[1])
        arg[3,1,1,0]=(0.386887699577)*sqrt(x[0])
        arg[3,1,1,1]=(-0.422560075108)*sqrt(x[1])
        arg[3,1,2,0]=(0.00387273783493)*sqrt(x[0])
        arg[3,1,2,1]=(0.465673505613)*sqrt(x[1])
        arg[3,2,0,0]=(0.987383428982)*sqrt(x[0])
        arg[3,2,0,1]=(0.376320055964)*sqrt(x[1])
        arg[3,2,1,0]=(-0.463778689128)*sqrt(x[0])
        arg[3,2,1,1]=(0.179816566227)*sqrt(x[1])
        arg[3,2,2,0]=(0.961522856801)*sqrt(x[0])
        arg[3,2,2,1]=(-0.257779627946)*sqrt(x[1])
        arg[3,3,0,0]=(0.748886531458)*sqrt(x[0])
        arg[3,3,0,1]=(-0.257342282566)*sqrt(x[1])
        arg[3,3,1,0]=(0.377494024401)*sqrt(x[0])
        arg[3,3,1,1]=(-0.334588346017)*sqrt(x[1])
        arg[3,3,2,0]=(0.502495149189)*sqrt(x[0])
        arg[3,3,2,1]=(0.534612429702)*sqrt(x[1])
        arg[3,4,0,0]=(-0.308551337355)*sqrt(x[0])
        arg[3,4,0,1]=(-0.471825826745)*sqrt(x[1])
        arg[3,4,1,0]=(-0.262606531584)*sqrt(x[0])
        arg[3,4,1,1]=(0.766089616367)*sqrt(x[1])
        arg[3,4,2,0]=(-0.136526755642)*sqrt(x[0])
        arg[3,4,2,1]=(0.675111459363)*sqrt(x[1])
        ref=sqrt((37.3453550914)*dim)

      else:
        arg=Data(0,(4, 5, 3, 3),w)
        arg[0,0,0,0]=(0.532946231146)*sqrt(x[0])
        arg[0,0,0,1]=(0.269364089513)*sqrt(x[1])
        arg[0,0,0,2]=(-0.207412457081)*sqrt(x[2])
        arg[0,0,1,0]=(-0.843104704858)*sqrt(x[0])
        arg[0,0,1,1]=(0.0416216508473)*sqrt(x[1])
        arg[0,0,1,2]=(-0.836074693662)*sqrt(x[2])
        arg[0,0,2,0]=(0.943609268731)*sqrt(x[0])
        arg[0,0,2,1]=(0.0154543737816)*sqrt(x[1])
        arg[0,0,2,2]=(0.0726268788381)*sqrt(x[2])
        arg[0,1,0,0]=(0.108422740078)*sqrt(x[0])
        arg[0,1,0,1]=(-0.296667916638)*sqrt(x[1])
        arg[0,1,0,2]=(-0.769732600535)*sqrt(x[2])
        arg[0,1,1,0]=(-0.428575493834)*sqrt(x[0])
        arg[0,1,1,1]=(0.421245456722)*sqrt(x[1])
        arg[0,1,1,2]=(-0.588277652277)*sqrt(x[2])
        arg[0,1,2,0]=(0.145294576795)*sqrt(x[0])
        arg[0,1,2,1]=(0.323206623794)*sqrt(x[1])
        arg[0,1,2,2]=(0.788115602892)*sqrt(x[2])
        arg[0,2,0,0]=(-0.227877282292)*sqrt(x[0])
        arg[0,2,0,1]=(-0.630647460719)*sqrt(x[1])
        arg[0,2,0,2]=(0.58754882135)*sqrt(x[2])
        arg[0,2,1,0]=(0.347191113403)*sqrt(x[0])
        arg[0,2,1,1]=(0.464093634725)*sqrt(x[1])
        arg[0,2,1,2]=(-0.0412800774497)*sqrt(x[2])
        arg[0,2,2,0]=(0.223364317185)*sqrt(x[0])
        arg[0,2,2,1]=(0.257201130157)*sqrt(x[1])
        arg[0,2,2,2]=(0.063203467463)*sqrt(x[2])
        arg[0,3,0,0]=(-0.723240451643)*sqrt(x[0])
        arg[0,3,0,1]=(-0.862468295097)*sqrt(x[1])
        arg[0,3,0,2]=(-0.149283247587)*sqrt(x[2])
        arg[0,3,1,0]=(0.15680097839)*sqrt(x[0])
        arg[0,3,1,1]=(0.421563637547)*sqrt(x[1])
        arg[0,3,1,2]=(0.111549188549)*sqrt(x[2])
        arg[0,3,2,0]=(-0.272783329363)*sqrt(x[0])
        arg[0,3,2,1]=(-0.420352789853)*sqrt(x[1])
        arg[0,3,2,2]=(0.570865117722)*sqrt(x[2])
        arg[0,4,0,0]=(-0.321910078414)*sqrt(x[0])
        arg[0,4,0,1]=(0.988695599439)*sqrt(x[1])
        arg[0,4,0,2]=(0.920200893398)*sqrt(x[2])
        arg[0,4,1,0]=(0.0260910072651)*sqrt(x[0])
        arg[0,4,1,1]=(0.460012578184)*sqrt(x[1])
        arg[0,4,1,2]=(0.848099524112)*sqrt(x[2])
        arg[0,4,2,0]=(0.242157803251)*sqrt(x[0])
        arg[0,4,2,1]=(0.394528777004)*sqrt(x[1])
        arg[0,4,2,2]=(0.562996837311)*sqrt(x[2])
        arg[1,0,0,0]=(0.459886225958)*sqrt(x[0])
        arg[1,0,0,1]=(-0.721868942003)*sqrt(x[1])
        arg[1,0,0,2]=(0.432203082994)*sqrt(x[2])
        arg[1,0,1,0]=(0.409831045482)*sqrt(x[0])
        arg[1,0,1,1]=(-0.481677513473)*sqrt(x[1])
        arg[1,0,1,2]=(0.439387853437)*sqrt(x[2])
        arg[1,0,2,0]=(0.261583198434)*sqrt(x[0])
        arg[1,0,2,1]=(0.290993423577)*sqrt(x[1])
        arg[1,0,2,2]=(0.477993114134)*sqrt(x[2])
        arg[1,1,0,0]=(0.586344598248)*sqrt(x[0])
        arg[1,1,0,1]=(-0.105390792831)*sqrt(x[1])
        arg[1,1,0,2]=(0.335990751314)*sqrt(x[2])
        arg[1,1,1,0]=(-0.191500562856)*sqrt(x[0])
        arg[1,1,1,1]=(0.244514598216)*sqrt(x[1])
        arg[1,1,1,2]=(-0.804402720669)*sqrt(x[2])
        arg[1,1,2,0]=(-0.455225710648)*sqrt(x[0])
        arg[1,1,2,1]=(-0.505052700585)*sqrt(x[1])
        arg[1,1,2,2]=(-0.0240295199362)*sqrt(x[2])
        arg[1,2,0,0]=(-0.718487964893)*sqrt(x[0])
        arg[1,2,0,1]=(-0.0899522570462)*sqrt(x[1])
        arg[1,2,0,2]=(-0.293353754696)*sqrt(x[2])
        arg[1,2,1,0]=(-0.180013826342)*sqrt(x[0])
        arg[1,2,1,1]=(0.793689231922)*sqrt(x[1])
        arg[1,2,1,2]=(0.673066555571)*sqrt(x[2])
        arg[1,2,2,0]=(0.705362155032)*sqrt(x[0])
        arg[1,2,2,1]=(0.54476742883)*sqrt(x[1])
        arg[1,2,2,2]=(-0.331195064878)*sqrt(x[2])
        arg[1,3,0,0]=(-0.360927441647)*sqrt(x[0])
        arg[1,3,0,1]=(0.230772030282)*sqrt(x[1])
        arg[1,3,0,2]=(0.912342489431)*sqrt(x[2])
        arg[1,3,1,0]=(-0.817510690014)*sqrt(x[0])
        arg[1,3,1,1]=(0.397583721353)*sqrt(x[1])
        arg[1,3,1,2]=(-0.982551067917)*sqrt(x[2])
        arg[1,3,2,0]=(0.86380240427)*sqrt(x[0])
        arg[1,3,2,1]=(-0.415018976841)*sqrt(x[1])
        arg[1,3,2,2]=(0.271582572267)*sqrt(x[2])
        arg[1,4,0,0]=(0.252845347406)*sqrt(x[0])
        arg[1,4,0,1]=(0.687786802906)*sqrt(x[1])
        arg[1,4,0,2]=(0.465501171342)*sqrt(x[2])
        arg[1,4,1,0]=(-0.613703721675)*sqrt(x[0])
        arg[1,4,1,1]=(-0.110297640533)*sqrt(x[1])
        arg[1,4,1,2]=(-0.836768056501)*sqrt(x[2])
        arg[1,4,2,0]=(-0.0400898232224)*sqrt(x[0])
        arg[1,4,2,1]=(0.0358172759009)*sqrt(x[1])
        arg[1,4,2,2]=(-0.335751455408)*sqrt(x[2])
        arg[2,0,0,0]=(-0.309992915015)*sqrt(x[0])
        arg[2,0,0,1]=(-0.721404217867)*sqrt(x[1])
        arg[2,0,0,2]=(-0.548000635629)*sqrt(x[2])
        arg[2,0,1,0]=(0.651175831531)*sqrt(x[0])
        arg[2,0,1,1]=(0.158960783491)*sqrt(x[1])
        arg[2,0,1,2]=(-0.310676926155)*sqrt(x[2])
        arg[2,0,2,0]=(-0.122289734411)*sqrt(x[0])
        arg[2,0,2,1]=(-0.252405938421)*sqrt(x[1])
        arg[2,0,2,2]=(-0.938280244213)*sqrt(x[2])
        arg[2,1,0,0]=(0.559495801686)*sqrt(x[0])
        arg[2,1,0,1]=(-0.547182622716)*sqrt(x[1])
        arg[2,1,0,2]=(0.397441517898)*sqrt(x[2])
        arg[2,1,1,0]=(-0.406112472071)*sqrt(x[0])
        arg[2,1,1,1]=(0.355063810677)*sqrt(x[1])
        arg[2,1,1,2]=(0.760400203215)*sqrt(x[2])
        arg[2,1,2,0]=(0.992201320481)*sqrt(x[0])
        arg[2,1,2,1]=(0.0580660882576)*sqrt(x[1])
        arg[2,1,2,2]=(-0.643170879939)*sqrt(x[2])
        arg[2,2,0,0]=(-0.280644461832)*sqrt(x[0])
        arg[2,2,0,1]=(-0.0467430285531)*sqrt(x[1])
        arg[2,2,0,2]=(0.314050593255)*sqrt(x[2])
        arg[2,2,1,0]=(-0.230032618609)*sqrt(x[0])
        arg[2,2,1,1]=(0.0996058698273)*sqrt(x[1])
        arg[2,2,1,2]=(-0.0270266073208)*sqrt(x[2])
        arg[2,2,2,0]=(0.767914132956)*sqrt(x[0])
        arg[2,2,2,1]=(0.496930363612)*sqrt(x[1])
        arg[2,2,2,2]=(-0.599525033616)*sqrt(x[2])
        arg[2,3,0,0]=(-0.326433376073)*sqrt(x[0])
        arg[2,3,0,1]=(-0.0366374501025)*sqrt(x[1])
        arg[2,3,0,2]=(0.22555705749)*sqrt(x[2])
        arg[2,3,1,0]=(-0.162548813895)*sqrt(x[0])
        arg[2,3,1,1]=(-0.110074212194)*sqrt(x[1])
        arg[2,3,1,2]=(-0.143600895553)*sqrt(x[2])
        arg[2,3,2,0]=(0.771148880174)*sqrt(x[0])
        arg[2,3,2,1]=(0.112528116552)*sqrt(x[1])
        arg[2,3,2,2]=(-0.955735294341)*sqrt(x[2])
        arg[2,4,0,0]=(-0.968392951034)*sqrt(x[0])
        arg[2,4,0,1]=(-0.36901708507)*sqrt(x[1])
        arg[2,4,0,2]=(0.283692515492)*sqrt(x[2])
        arg[2,4,1,0]=(0.997238032837)*sqrt(x[0])
        arg[2,4,1,1]=(-0.625794124653)*sqrt(x[1])
        arg[2,4,1,2]=(0.533386027556)*sqrt(x[2])
        arg[2,4,2,0]=(0.977311695557)*sqrt(x[0])
        arg[2,4,2,1]=(0.693009976689)*sqrt(x[1])
        arg[2,4,2,2]=(0.711179347652)*sqrt(x[2])
        arg[3,0,0,0]=(-0.155585788931)*sqrt(x[0])
        arg[3,0,0,1]=(0.0228078851234)*sqrt(x[1])
        arg[3,0,0,2]=(0.510104938032)*sqrt(x[2])
        arg[3,0,1,0]=(0.74865995369)*sqrt(x[0])
        arg[3,0,1,1]=(0.672153736284)*sqrt(x[1])
        arg[3,0,1,2]=(0.588012355098)*sqrt(x[2])
        arg[3,0,2,0]=(-0.924508475715)*sqrt(x[0])
        arg[3,0,2,1]=(-0.392784674758)*sqrt(x[1])
        arg[3,0,2,2]=(-0.36371454642)*sqrt(x[2])
        arg[3,1,0,0]=(-0.709783490337)*sqrt(x[0])
        arg[3,1,0,1]=(0.844136172222)*sqrt(x[1])
        arg[3,1,0,2]=(0.621011730043)*sqrt(x[2])
        arg[3,1,1,0]=(0.428807337181)*sqrt(x[0])
        arg[3,1,1,1]=(0.126300214574)*sqrt(x[1])
        arg[3,1,1,2]=(0.795972806221)*sqrt(x[2])
        arg[3,1,2,0]=(-0.252334324004)*sqrt(x[0])
        arg[3,1,2,1]=(-0.722829467938)*sqrt(x[1])
        arg[3,1,2,2]=(-0.551540062366)*sqrt(x[2])
        arg[3,2,0,0]=(-0.134668475963)*sqrt(x[0])
        arg[3,2,0,1]=(-0.598747540536)*sqrt(x[1])
        arg[3,2,0,2]=(0.426422436624)*sqrt(x[2])
        arg[3,2,1,0]=(-0.363050323762)*sqrt(x[0])
        arg[3,2,1,1]=(0.980891457977)*sqrt(x[1])
        arg[3,2,1,2]=(0.162831912555)*sqrt(x[2])
        arg[3,2,2,0]=(-0.126505493475)*sqrt(x[0])
        arg[3,2,2,1]=(-0.578567864811)*sqrt(x[1])
        arg[3,2,2,2]=(-0.509843129095)*sqrt(x[2])
        arg[3,3,0,0]=(-0.446171262265)*sqrt(x[0])
        arg[3,3,0,1]=(-0.715175197494)*sqrt(x[1])
        arg[3,3,0,2]=(-0.881016888806)*sqrt(x[2])
        arg[3,3,1,0]=(-0.942020866327)*sqrt(x[0])
        arg[3,3,1,1]=(0.156434646828)*sqrt(x[1])
        arg[3,3,1,2]=(0.523624761583)*sqrt(x[2])
        arg[3,3,2,0]=(-0.683550923926)*sqrt(x[0])
        arg[3,3,2,1]=(0.857075218033)*sqrt(x[1])
        arg[3,3,2,2]=(0.297672594023)*sqrt(x[2])
        arg[3,4,0,0]=(0.74317121113)*sqrt(x[0])
        arg[3,4,0,1]=(0.076464540756)*sqrt(x[1])
        arg[3,4,0,2]=(0.781965468281)*sqrt(x[2])
        arg[3,4,1,0]=(0.417750169098)*sqrt(x[0])
        arg[3,4,1,1]=(0.82275428729)*sqrt(x[1])
        arg[3,4,1,2]=(0.919072321093)*sqrt(x[2])
        arg[3,4,2,0]=(-0.0246706472217)*sqrt(x[0])
        arg[3,4,2,1]=(0.179863245513)*sqrt(x[1])
        arg[3,4,2,2]=(0.539115287766)*sqrt(x[2])
        ref=sqrt((52.492676775)*dim)

      res=L2(arg)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertAlmostEqual(res,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromSymbol_rank0(self):
      """
      tests L2-norm of Symbol on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(0.0248400067683)*sqrt(x[0])
        ref=sqrt((0.000617025936249)*dim)

      else:
        arg=Symbol(shape=(),dim=dim)
        s=Data(0,(),w)
        s=(-0.48069197083)*sqrt(x[0])
        ref=sqrt((0.23106477082)*dim)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromSymbol_rank1(self):
      """
      tests L2-norm of Symbol on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(2,),dim=dim)
        s=Data(0,(2,),w)
        s[0]=(0.541099842991)*sqrt(x[0])
        s[1]=(0.123212139947)*sqrt(x[1])
        ref=sqrt((0.307970271515)*dim)

      else:
        arg=Symbol(shape=(3,),dim=dim)
        s=Data(0,(3,),w)
        s[0]=(-0.74011573744)*sqrt(x[0])
        s[1]=(0.63543826621)*sqrt(x[1])
        s[2]=(-0.210366750156)*sqrt(x[2])
        ref=sqrt((0.995807264541)*dim)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromSymbol_rank2(self):
      """
      tests L2-norm of Symbol on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 2),dim=dim)
        s=Data(0,(4, 2),w)
        s[0,0]=(-0.324596599761)*sqrt(x[0])
        s[0,1]=(0.456181444086)*sqrt(x[1])
        s[1,0]=(-0.85366968746)*sqrt(x[0])
        s[1,1]=(-0.984222055109)*sqrt(x[1])
        s[2,0]=(0.302615209563)*sqrt(x[0])
        s[2,1]=(-0.259853865003)*sqrt(x[1])
        s[3,0]=(-0.2851901062)*sqrt(x[0])
        s[3,1]=(0.790097073483)*sqrt(x[1])
        ref=sqrt((2.87559622997)*dim)

      else:
        arg=Symbol(shape=(4, 3),dim=dim)
        s=Data(0,(4, 3),w)
        s[0,0]=(0.806171559469)*sqrt(x[0])
        s[0,1]=(0.734186538041)*sqrt(x[1])
        s[0,2]=(0.376885565623)*sqrt(x[2])
        s[1,0]=(-0.168339360861)*sqrt(x[0])
        s[1,1]=(-0.184538096909)*sqrt(x[1])
        s[1,2]=(-0.649204637511)*sqrt(x[2])
        s[2,0]=(0.0187762580033)*sqrt(x[0])
        s[2,1]=(-0.851287373276)*sqrt(x[1])
        s[2,2]=(0.84351620048)*sqrt(x[2])
        s[3,0]=(0.759367087847)*sqrt(x[0])
        s[3,1]=(-0.418326808815)*sqrt(x[1])
        s[3,2]=(0.207925273666)*sqrt(x[2])
        ref=sqrt((4.04627522925)*dim)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromSymbol_rank3(self):
      """
      tests L2-norm of Symbol on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(6, 2, 2),dim=dim)
        s=Data(0,(6, 2, 2),w)
        s[0,0,0]=(0.656602395169)*sqrt(x[0])
        s[0,0,1]=(-0.153988287938)*sqrt(x[1])
        s[0,1,0]=(-0.911070962892)*sqrt(x[0])
        s[0,1,1]=(0.734617382897)*sqrt(x[1])
        s[1,0,0]=(0.708777892683)*sqrt(x[0])
        s[1,0,1]=(0.407403464363)*sqrt(x[1])
        s[1,1,0]=(-0.73004005055)*sqrt(x[0])
        s[1,1,1]=(0.108502620047)*sqrt(x[1])
        s[2,0,0]=(0.584913648885)*sqrt(x[0])
        s[2,0,1]=(0.655698697233)*sqrt(x[1])
        s[2,1,0]=(0.504301542359)*sqrt(x[0])
        s[2,1,1]=(-0.637819233396)*sqrt(x[1])
        s[3,0,0]=(0.395391847846)*sqrt(x[0])
        s[3,0,1]=(-0.136070177387)*sqrt(x[1])
        s[3,1,0]=(0.77444662439)*sqrt(x[0])
        s[3,1,1]=(0.245316754174)*sqrt(x[1])
        s[4,0,0]=(0.542716832119)*sqrt(x[0])
        s[4,0,1]=(0.0202895164815)*sqrt(x[1])
        s[4,1,0]=(0.922586929777)*sqrt(x[0])
        s[4,1,1]=(-0.948487993154)*sqrt(x[1])
        s[5,0,0]=(0.629779082528)*sqrt(x[0])
        s[5,0,1]=(-0.90719864412)*sqrt(x[1])
        s[5,1,0]=(-0.29521666306)*sqrt(x[0])
        s[5,1,1]=(-0.431748113769)*sqrt(x[1])
        ref=sqrt((8.84456266856)*dim)

      else:
        arg=Symbol(shape=(6, 2, 3),dim=dim)
        s=Data(0,(6, 2, 3),w)
        s[0,0,0]=(0.529698924744)*sqrt(x[0])
        s[0,0,1]=(0.779440018368)*sqrt(x[1])
        s[0,0,2]=(-0.447704804318)*sqrt(x[2])
        s[0,1,0]=(0.375433500828)*sqrt(x[0])
        s[0,1,1]=(-0.567989496415)*sqrt(x[1])
        s[0,1,2]=(-0.47133957857)*sqrt(x[2])
        s[1,0,0]=(0.379156720923)*sqrt(x[0])
        s[1,0,1]=(-0.629670565361)*sqrt(x[1])
        s[1,0,2]=(0.814298387366)*sqrt(x[2])
        s[1,1,0]=(-0.482876103672)*sqrt(x[0])
        s[1,1,1]=(-0.592362629566)*sqrt(x[1])
        s[1,1,2]=(0.368859438611)*sqrt(x[2])
        s[2,0,0]=(0.145060478226)*sqrt(x[0])
        s[2,0,1]=(-0.582989521856)*sqrt(x[1])
        s[2,0,2]=(-0.977335802848)*sqrt(x[2])
        s[2,1,0]=(-0.471820598698)*sqrt(x[0])
        s[2,1,1]=(-0.680071585031)*sqrt(x[1])
        s[2,1,2]=(-0.686102402643)*sqrt(x[2])
        s[3,0,0]=(-0.647535125154)*sqrt(x[0])
        s[3,0,1]=(-0.739392553517)*sqrt(x[1])
        s[3,0,2]=(-0.0852792216585)*sqrt(x[2])
        s[3,1,0]=(-0.0197635627728)*sqrt(x[0])
        s[3,1,1]=(-0.811782244587)*sqrt(x[1])
        s[3,1,2]=(0.842974983966)*sqrt(x[2])
        s[4,0,0]=(-0.765665515246)*sqrt(x[0])
        s[4,0,1]=(-0.379368211948)*sqrt(x[1])
        s[4,0,2]=(-0.864157207524)*sqrt(x[2])
        s[4,1,0]=(-0.393989909405)*sqrt(x[0])
        s[4,1,1]=(0.482309706914)*sqrt(x[1])
        s[4,1,2]=(-0.80068507467)*sqrt(x[2])
        s[5,0,0]=(-0.686334218925)*sqrt(x[0])
        s[5,0,1]=(-0.414623392196)*sqrt(x[1])
        s[5,0,2]=(-0.689227675292)*sqrt(x[2])
        s[5,1,0]=(0.347067831012)*sqrt(x[0])
        s[5,1,1]=(0.675600902408)*sqrt(x[1])
        s[5,1,2]=(-0.763467841004)*sqrt(x[2])
        ref=sqrt((13.2965907789)*dim)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   def test_L2_onReducedFunctionOnBoundary_fromSymbol_rank4(self):
      """
      tests L2-norm of Symbol on the ReducedFunctionOnBoundary

      assumptions: self.domain supports integration on ReducedFunctionOnBoundary
      """
      dim=self.domain.getDim()
      w=ReducedFunctionOnBoundary(self.domain)
      x=w.getX()
      if dim==2:
        arg=Symbol(shape=(4, 5, 3, 2),dim=dim)
        s=Data(0,(4, 5, 3, 2),w)
        s[0,0,0,0]=(0.478851101914)*sqrt(x[0])
        s[0,0,0,1]=(-0.669115623243)*sqrt(x[1])
        s[0,0,1,0]=(0.169431656457)*sqrt(x[0])
        s[0,0,1,1]=(-0.154130248518)*sqrt(x[1])
        s[0,0,2,0]=(0.590784906322)*sqrt(x[0])
        s[0,0,2,1]=(-0.843193872964)*sqrt(x[1])
        s[0,1,0,0]=(0.040548638805)*sqrt(x[0])
        s[0,1,0,1]=(0.0223688163494)*sqrt(x[1])
        s[0,1,1,0]=(0.591094685655)*sqrt(x[0])
        s[0,1,1,1]=(0.692105812582)*sqrt(x[1])
        s[0,1,2,0]=(-0.326832256739)*sqrt(x[0])
        s[0,1,2,1]=(0.824824822029)*sqrt(x[1])
        s[0,2,0,0]=(-0.352757110519)*sqrt(x[0])
        s[0,2,0,1]=(-0.898616847036)*sqrt(x[1])
        s[0,2,1,0]=(0.972018635849)*sqrt(x[0])
        s[0,2,1,1]=(0.64223146835)*sqrt(x[1])
        s[0,2,2,0]=(0.599638014876)*sqrt(x[0])
        s[0,2,2,1]=(0.665927987689)*sqrt(x[1])
        s[0,3,0,0]=(-0.757282087611)*sqrt(x[0])
        s[0,3,0,1]=(-0.541306038159)*sqrt(x[1])
        s[0,3,1,0]=(0.443494146532)*sqrt(x[0])
        s[0,3,1,1]=(-0.125404019651)*sqrt(x[1])
        s[0,3,2,0]=(0.234422087306)*sqrt(x[0])
        s[0,3,2,1]=(0.443982420301)*sqrt(x[1])
        s[0,4,0,0]=(0.46454740036)*sqrt(x[0])
        s[0,4,0,1]=(-0.479031824057)*sqrt(x[1])
        s[0,4,1,0]=(-0.304056971991)*sqrt(x[0])
        s[0,4,1,1]=(-0.763293280009)*sqrt(x[1])
        s[0,4,2,0]=(-0.142796685199)*sqrt(x[0])
        s[0,4,2,1]=(0.301022025426)*sqrt(x[1])
        s[1,0,0,0]=(-0.843399285738)*sqrt(x[0])
        s[1,0,0,1]=(0.639837980509)*sqrt(x[1])
        s[1,0,1,0]=(-0.649024291296)*sqrt(x[0])
        s[1,0,1,1]=(0.645960995308)*sqrt(x[1])
        s[1,0,2,0]=(0.0862852468825)*sqrt(x[0])
        s[1,0,2,1]=(-0.399237566917)*sqrt(x[1])
        s[1,1,0,0]=(0.959425187971)*sqrt(x[0])
        s[1,1,0,1]=(0.65407910284)*sqrt(x[1])
        s[1,1,1,0]=(0.873927579655)*sqrt(x[0])
        s[1,1,1,1]=(0.399448652787)*sqrt(x[1])
        s[1,1,2,0]=(-0.363863298937)*sqrt(x[0])
        s[1,1,2,1]=(0.555310285452)*sqrt(x[1])
        s[1,2,0,0]=(0.962712099815)*sqrt(x[0])
        s[1,2,0,1]=(0.141495259052)*sqrt(x[1])
        s[1,2,1,0]=(-0.819117296209)*sqrt(x[0])
        s[1,2,1,1]=(-0.752934712352)*sqrt(x[1])
        s[1,2,2,0]=(0.065886700262)*sqrt(x[0])
        s[1,2,2,1]=(-0.530225381274)*sqrt(x[1])
        s[1,3,0,0]=(0.830686848377)*sqrt(x[0])
        s[1,3,0,1]=(0.139854694845)*sqrt(x[1])
        s[1,3,1,0]=(0.869755726605)*sqrt(x[0])
        s[1,3,1,1]=(0.145889500911)*sqrt(x[1])
        s[1,3,2,0]=(-0.690283765486)*sqrt(x[0])
        s[1,3,2,1]=(-0.790245562042)*sqrt(x[1])
        s[1,4,0,0]=(0.704042980492)*sqrt(x[0])
        s[1,4,0,1]=(-0.692268480039)*sqrt(x[1])
        s[1,4,1,0]=(-0.708898941564)*sqrt(x[0])
        s[1,4,1,1]=(-0.675824414677)*sqrt(x[1])
        s[1,4,2,0]=(0.320394083261)*sqrt(x[0])
        s[1,4,2,1]=(0.173795411968)*sqrt(x[1])
        s[2,0,0,0]=(0.707844986267)*sqrt(x[0])
        s[2,0,0,1]=(0.991847131293)*sqrt(x[1])
        s[2,0,1,0]=(0.470933752596)*sqrt(x[0])
        s[2,0,1,1]=(0.847427404272)*sqrt(x[1])
        s[2,0,2,0]=(0.383056136263)*sqrt(x[0])
        s[2,0,2,1]=(-0.717411302782)*sqrt(x[1])
        s[2,1,0,0]=(0.118326527818)*sqrt(x[0])
        s[2,1,0,1]=(0.0957434761651)*sqrt(x[1])
        s[2,1,1,0]=(-0.420654904504)*sqrt(x[0])
        s[2,1,1,1]=(0.31903393188)*sqrt(x[1])
        s[2,1,2,0]=(0.0597937431476)*sqrt(x[0])
        s[2,1,2,1]=(0.965201628622)*sqrt(x[1])
        s[2,2,0,0]=(0.884709108482)*sqrt(x[0])
        s[2,2,0,1]=(-0.541160500366)*sqrt(x[1])
        s[2,2,1,0]=(-0.180788996448)*sqrt(x[0])
        s[2,2,1,1]=(-0.44537512426)*sqrt(x[1])
        s[2,2,2,0]=(-0.291513668432)*sqrt(x[0])
        s[2,2,2,1]=(0.349090737561)*sqrt(x[1])
        s[2,3,0,0]=(0.0994765186043)*sqrt(x[0])
        s[2,3,0,1]=(-0.317258869646)*sqrt(x[1])
        s[2,3,1,0]=(0.661005398598)*sqrt(x[0])
        s[2,3,1,1]=(-0.914254206568)*sqrt(x[1])
        s[2,3,2,0]=(-0.374142249366)*sqrt(x[0])
        s[2,3,2,1]=(-0.587651938713)*sqrt(x[1])
        s[2,4,0,0]=(-0.0990354726919)*sqrt(x[0])
        s[2,4,0,1]=(-0.104596309056)*sqrt(x[1])
        s[2,4,1,0]=(0.183201845594)*sqrt(x[0])
        s[2,4,1,1]=(-0.265185815743)*sqrt(x[1])
        s[2,4,2,0]=(0.804060169933)*sqrt(x[0])
        s[2,4,2,1]=(-0.0798336799535)*sqrt(x[1])
        s[3,0,0,0]=(-0.470265711106)*sqrt(x[0])
        s[3,0,0,1]=(0.123030821016)*sqrt(x[1])
        s[3,0,1,0]=(0.971878570072)*sqrt(x[0])
        s[3,0,1,1]=(0.557374437558)*sqrt(x[1])
        s[3,0,2,0]=(0.771881325143)*sqrt(x[0])
        s[3,0,2,1]=(-0.181917882935)*sqrt(x[1])
        s[3,1,0,0]=(-0.30330086877)*sqrt(x[0])
        s[3,1,0,1]=(0.785561830696)*sqrt(x[1])
        s[3,1,1,0]=(0.0783160883668)*sqrt(x[0])
        s[3,1,1,1]=(-0.192122642565)*sqrt(x[1])
        s[3,1,2,0]=(-0.294486881116)*sqrt(x[0])
        s[3,1,2,1]=(0.533158403133)*sqrt(x[1])
        s[3,2,0,0]=(-0.413737404117)*sqrt(x[0])
        s[3,2,0,1]=(0.622363162902)*sqrt(x[1])
        s[3,2,1,0]=(0.241741847782)*sqrt(x[0])
        s[3,2,1,1]=(-0.0541594558102)*sqrt(x[1])
        s[3,2,2,0]=(0.157759963106)*sqrt(x[0])
        s[3,2,2,1]=(-0.833376953416)*sqrt(x[1])
        s[3,3,0,0]=(-0.671630074873)*sqrt(x[0])
        s[3,3,0,1]=(-0.10840086864)*sqrt(x[1])
        s[3,3,1,0]=(-0.995501368513)*sqrt(x[0])
        s[3,3,1,1]=(-0.153147007921)*sqrt(x[1])
        s[3,3,2,0]=(-0.504815129335)*sqrt(x[0])
        s[3,3,2,1]=(-0.489511115245)*sqrt(x[1])
        s[3,4,0,0]=(0.206555206064)*sqrt(x[0])
        s[3,4,0,1]=(-0.294058867178)*sqrt(x[1])
        s[3,4,1,0]=(-0.571176824119)*sqrt(x[0])
        s[3,4,1,1]=(0.353847783519)*sqrt(x[1])
        s[3,4,2,0]=(-0.441012644459)*sqrt(x[0])
        s[3,4,2,1]=(0.198182379551)*sqrt(x[1])
        ref=sqrt((36.8775562848)*dim)

      else:
        arg=Symbol(shape=(4, 5, 3, 3),dim=dim)
        s=Data(0,(4, 5, 3, 3),w)
        s[0,0,0,0]=(-0.391342187918)*sqrt(x[0])
        s[0,0,0,1]=(-0.807670030995)*sqrt(x[1])
        s[0,0,0,2]=(-0.730776810158)*sqrt(x[2])
        s[0,0,1,0]=(-0.640249702579)*sqrt(x[0])
        s[0,0,1,1]=(-0.0944608864404)*sqrt(x[1])
        s[0,0,1,2]=(0.852922834544)*sqrt(x[2])
        s[0,0,2,0]=(0.0996848239317)*sqrt(x[0])
        s[0,0,2,1]=(-0.323310001677)*sqrt(x[1])
        s[0,0,2,2]=(-0.311089889753)*sqrt(x[2])
        s[0,1,0,0]=(-0.294816251058)*sqrt(x[0])
        s[0,1,0,1]=(0.939389998027)*sqrt(x[1])
        s[0,1,0,2]=(0.900617063513)*sqrt(x[2])
        s[0,1,1,0]=(-0.961986583089)*sqrt(x[0])
        s[0,1,1,1]=(0.629473020335)*sqrt(x[1])
        s[0,1,1,2]=(-0.215446848397)*sqrt(x[2])
        s[0,1,2,0]=(-0.256062202737)*sqrt(x[0])
        s[0,1,2,1]=(0.497231616486)*sqrt(x[1])
        s[0,1,2,2]=(-0.813869122359)*sqrt(x[2])
        s[0,2,0,0]=(0.352147663827)*sqrt(x[0])
        s[0,2,0,1]=(0.553022457414)*sqrt(x[1])
        s[0,2,0,2]=(-0.340077213113)*sqrt(x[2])
        s[0,2,1,0]=(0.654397833915)*sqrt(x[0])
        s[0,2,1,1]=(-0.34308640744)*sqrt(x[1])
        s[0,2,1,2]=(-0.689126280824)*sqrt(x[2])
        s[0,2,2,0]=(0.428183196531)*sqrt(x[0])
        s[0,2,2,1]=(-0.322384435092)*sqrt(x[1])
        s[0,2,2,2]=(0.636495513573)*sqrt(x[2])
        s[0,3,0,0]=(0.214083660816)*sqrt(x[0])
        s[0,3,0,1]=(0.312445001293)*sqrt(x[1])
        s[0,3,0,2]=(-0.939086993471)*sqrt(x[2])
        s[0,3,1,0]=(-0.0668860159236)*sqrt(x[0])
        s[0,3,1,1]=(0.591012248354)*sqrt(x[1])
        s[0,3,1,2]=(0.696149816619)*sqrt(x[2])
        s[0,3,2,0]=(0.917050906357)*sqrt(x[0])
        s[0,3,2,1]=(-0.695398022945)*sqrt(x[1])
        s[0,3,2,2]=(0.578841930671)*sqrt(x[2])
        s[0,4,0,0]=(-0.321769393495)*sqrt(x[0])
        s[0,4,0,1]=(-0.761161601481)*sqrt(x[1])
        s[0,4,0,2]=(0.000116192661055)*sqrt(x[2])
        s[0,4,1,0]=(0.347135267821)*sqrt(x[0])
        s[0,4,1,1]=(-0.0105215354141)*sqrt(x[1])
        s[0,4,1,2]=(0.582827998159)*sqrt(x[2])
        s[0,4,2,0]=(0.765317199582)*sqrt(x[0])
        s[0,4,2,1]=(0.546093510978)*sqrt(x[1])
        s[0,4,2,2]=(-0.753832119978)*sqrt(x[2])
        s[1,0,0,0]=(-0.163230444043)*sqrt(x[0])
        s[1,0,0,1]=(0.415060470539)*sqrt(x[1])
        s[1,0,0,2]=(0.638843786094)*sqrt(x[2])
        s[1,0,1,0]=(-0.320730240187)*sqrt(x[0])
        s[1,0,1,1]=(-0.369428894975)*sqrt(x[1])
        s[1,0,1,2]=(-0.944414569234)*sqrt(x[2])
        s[1,0,2,0]=(0.580606041923)*sqrt(x[0])
        s[1,0,2,1]=(-0.351082080131)*sqrt(x[1])
        s[1,0,2,2]=(-0.170588782574)*sqrt(x[2])
        s[1,1,0,0]=(0.323136286071)*sqrt(x[0])
        s[1,1,0,1]=(0.397263600918)*sqrt(x[1])
        s[1,1,0,2]=(-0.687508659585)*sqrt(x[2])
        s[1,1,1,0]=(-0.997412227958)*sqrt(x[0])
        s[1,1,1,1]=(0.968968702091)*sqrt(x[1])
        s[1,1,1,2]=(-0.223523055076)*sqrt(x[2])
        s[1,1,2,0]=(-0.495473193251)*sqrt(x[0])
        s[1,1,2,1]=(-0.00758436986969)*sqrt(x[1])
        s[1,1,2,2]=(0.0770020818673)*sqrt(x[2])
        s[1,2,0,0]=(-0.522152049391)*sqrt(x[0])
        s[1,2,0,1]=(0.820143671384)*sqrt(x[1])
        s[1,2,0,2]=(0.672347144349)*sqrt(x[2])
        s[1,2,1,0]=(-0.868655459375)*sqrt(x[0])
        s[1,2,1,1]=(-0.698503510192)*sqrt(x[1])
        s[1,2,1,2]=(0.16866148235)*sqrt(x[2])
        s[1,2,2,0]=(0.291645089215)*sqrt(x[0])
        s[1,2,2,1]=(0.375705968271)*sqrt(x[1])
        s[1,2,2,2]=(0.047575084848)*sqrt(x[2])
        s[1,3,0,0]=(-0.856292393733)*sqrt(x[0])
        s[1,3,0,1]=(0.741997693375)*sqrt(x[1])
        s[1,3,0,2]=(0.922057641142)*sqrt(x[2])
        s[1,3,1,0]=(-0.0246230217779)*sqrt(x[0])
        s[1,3,1,1]=(0.233511225618)*sqrt(x[1])
        s[1,3,1,2]=(-0.128732881848)*sqrt(x[2])
        s[1,3,2,0]=(0.437273172814)*sqrt(x[0])
        s[1,3,2,1]=(-0.24673804436)*sqrt(x[1])
        s[1,3,2,2]=(0.834689693971)*sqrt(x[2])
        s[1,4,0,0]=(0.0567731436429)*sqrt(x[0])
        s[1,4,0,1]=(-0.523439015055)*sqrt(x[1])
        s[1,4,0,2]=(-0.175105109157)*sqrt(x[2])
        s[1,4,1,0]=(-0.783819538589)*sqrt(x[0])
        s[1,4,1,1]=(0.722409822905)*sqrt(x[1])
        s[1,4,1,2]=(-0.99480794117)*sqrt(x[2])
        s[1,4,2,0]=(-0.0948610680733)*sqrt(x[0])
        s[1,4,2,1]=(-0.0934034446625)*sqrt(x[1])
        s[1,4,2,2]=(0.993222484498)*sqrt(x[2])
        s[2,0,0,0]=(-0.739506857296)*sqrt(x[0])
        s[2,0,0,1]=(-0.402129626454)*sqrt(x[1])
        s[2,0,0,2]=(-0.131164073645)*sqrt(x[2])
        s[2,0,1,0]=(-0.185068425131)*sqrt(x[0])
        s[2,0,1,1]=(-0.888669413054)*sqrt(x[1])
        s[2,0,1,2]=(0.149638740783)*sqrt(x[2])
        s[2,0,2,0]=(-0.977520384)*sqrt(x[0])
        s[2,0,2,1]=(0.495784511707)*sqrt(x[1])
        s[2,0,2,2]=(0.140779930302)*sqrt(x[2])
        s[2,1,0,0]=(-0.668882409761)*sqrt(x[0])
        s[2,1,0,1]=(0.87881226199)*sqrt(x[1])
        s[2,1,0,2]=(0.596921491285)*sqrt(x[2])
        s[2,1,1,0]=(0.327990184264)*sqrt(x[0])
        s[2,1,1,1]=(-0.842310366654)*sqrt(x[1])
        s[2,1,1,2]=(-0.104630179971)*sqrt(x[2])
        s[2,1,2,0]=(-0.997782736966)*sqrt(x[0])
        s[2,1,2,1]=(0.362490263673)*sqrt(x[1])
        s[2,1,2,2]=(-0.919527812431)*sqrt(x[2])
        s[2,2,0,0]=(-0.694141113242)*sqrt(x[0])
        s[2,2,0,1]=(-0.189176840568)*sqrt(x[1])
        s[2,2,0,2]=(-0.305307574884)*sqrt(x[2])
        s[2,2,1,0]=(0.97520770841)*sqrt(x[0])
        s[2,2,1,1]=(0.709672831962)*sqrt(x[1])
        s[2,2,1,2]=(-0.056408975067)*sqrt(x[2])
        s[2,2,2,0]=(-0.491153842737)*sqrt(x[0])
        s[2,2,2,1]=(-0.487533744248)*sqrt(x[1])
        s[2,2,2,2]=(-0.16258274458)*sqrt(x[2])
        s[2,3,0,0]=(0.69721543109)*sqrt(x[0])
        s[2,3,0,1]=(-0.927116706183)*sqrt(x[1])
        s[2,3,0,2]=(-0.277453826622)*sqrt(x[2])
        s[2,3,1,0]=(-0.530246149296)*sqrt(x[0])
        s[2,3,1,1]=(0.971840533777)*sqrt(x[1])
        s[2,3,1,2]=(-0.472599623799)*sqrt(x[2])
        s[2,3,2,0]=(0.772149184332)*sqrt(x[0])
        s[2,3,2,1]=(-0.644107019603)*sqrt(x[1])
        s[2,3,2,2]=(-0.537937585926)*sqrt(x[2])
        s[2,4,0,0]=(-0.786226779015)*sqrt(x[0])
        s[2,4,0,1]=(-0.132798607299)*sqrt(x[1])
        s[2,4,0,2]=(-0.954772620074)*sqrt(x[2])
        s[2,4,1,0]=(-0.378657907627)*sqrt(x[0])
        s[2,4,1,1]=(0.896332243914)*sqrt(x[1])
        s[2,4,1,2]=(-0.980811837061)*sqrt(x[2])
        s[2,4,2,0]=(-0.033610987457)*sqrt(x[0])
        s[2,4,2,1]=(0.301124520167)*sqrt(x[1])
        s[2,4,2,2]=(0.505461084966)*sqrt(x[2])
        s[3,0,0,0]=(0.92844681553)*sqrt(x[0])
        s[3,0,0,1]=(-0.833141539681)*sqrt(x[1])
        s[3,0,0,2]=(-0.219378232103)*sqrt(x[2])
        s[3,0,1,0]=(-0.342722442206)*sqrt(x[0])
        s[3,0,1,1]=(0.662705183551)*sqrt(x[1])
        s[3,0,1,2]=(0.544702876622)*sqrt(x[2])
        s[3,0,2,0]=(-0.984992104327)*sqrt(x[0])
        s[3,0,2,1]=(-0.118846731)*sqrt(x[1])
        s[3,0,2,2]=(-0.850203016683)*sqrt(x[2])
        s[3,1,0,0]=(0.797626533525)*sqrt(x[0])
        s[3,1,0,1]=(-0.415519060262)*sqrt(x[1])
        s[3,1,0,2]=(-0.629491781709)*sqrt(x[2])
        s[3,1,1,0]=(0.223213637134)*sqrt(x[0])
        s[3,1,1,1]=(0.824993414773)*sqrt(x[1])
        s[3,1,1,2]=(0.141523286845)*sqrt(x[2])
        s[3,1,2,0]=(-0.66436462404)*sqrt(x[0])
        s[3,1,2,1]=(-0.854989422144)*sqrt(x[1])
        s[3,1,2,2]=(-0.0354537218933)*sqrt(x[2])
        s[3,2,0,0]=(-0.0988394772118)*sqrt(x[0])
        s[3,2,0,1]=(0.125214003041)*sqrt(x[1])
        s[3,2,0,2]=(-0.701121659493)*sqrt(x[2])
        s[3,2,1,0]=(-0.603538097381)*sqrt(x[0])
        s[3,2,1,1]=(-0.659139390328)*sqrt(x[1])
        s[3,2,1,2]=(-0.849986094805)*sqrt(x[2])
        s[3,2,2,0]=(-0.545364489526)*sqrt(x[0])
        s[3,2,2,1]=(-0.90842178441)*sqrt(x[1])
        s[3,2,2,2]=(-0.650251007613)*sqrt(x[2])
        s[3,3,0,0]=(0.943460211337)*sqrt(x[0])
        s[3,3,0,1]=(-0.707864979205)*sqrt(x[1])
        s[3,3,0,2]=(0.9885932761)*sqrt(x[2])
        s[3,3,1,0]=(0.127650934393)*sqrt(x[0])
        s[3,3,1,1]=(-0.495451047906)*sqrt(x[1])
        s[3,3,1,2]=(0.641823757246)*sqrt(x[2])
        s[3,3,2,0]=(-0.33331510031)*sqrt(x[0])
        s[3,3,2,1]=(0.922346384165)*sqrt(x[1])
        s[3,3,2,2]=(-0.987453834808)*sqrt(x[2])
        s[3,4,0,0]=(-0.376485044284)*sqrt(x[0])
        s[3,4,0,1]=(0.693296795267)*sqrt(x[1])
        s[3,4,0,2]=(-0.31498404544)*sqrt(x[2])
        s[3,4,1,0]=(0.512477956635)*sqrt(x[0])
        s[3,4,1,1]=(-0.333819138285)*sqrt(x[1])
        s[3,4,1,2]=(0.560322373773)*sqrt(x[2])
        s[3,4,2,0]=(0.889090668029)*sqrt(x[0])
        s[3,4,2,1]=(0.981814634684)*sqrt(x[1])
        s[3,4,2,2]=(-0.41272129735)*sqrt(x[2])
        ref=sqrt((66.7960236856)*dim)

      res=L2(arg)
      sub=res.substitute({arg:s})
      self.assertTrue(isinstance(res,Symbol),"wrong type of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertAlmostEqual(sub,ref,int(-log10(self.RES_TOL)),"wrong result")


   def test_integrate_onFunction_fromConstant_Solution_rank0(self):
      """
      test integral of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 8.515514e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Solution_rank1(self):
      """
      test integral of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.3503347430465249, 0.22515496985130545, 0.7716259978630422] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Solution_rank2(self):
      """
      test integral of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.10477678617197395, 0.8079771602462849], [0.36834575664268165, 0.22684521892495912], [0.5071818431095225, 0.2600818357623923], [0.28415128659786737, 0.3055516629560848]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Solution_rank3(self):
      """
      test integral of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.4486631794938205, 0.09199721073458356, 0.00504828825798076, 0.6234618966091582], [0.8962009095779152, 0.36633565590475337, 0.4199307138191847, 0.4806106946156349]], [[0.8296911462191642, 0.5264685949220673, 0.038405707495928976, 0.3579909865055174], [0.40452790610726186, 0.8249014030447385, 0.10707696100458464, 0.18093045318668055]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Solution_rank4(self):
      """
      test integral of constant rank 4 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.4417240747154765, 0.4157972383188666, 0.11390715625132786, 0.13614529752782534], [0.5783914405888552, 0.4790527059197023, 0.8539655005107416, 0.11545293195545214], [0.23122082094092655, 0.5025015303859531, 0.3233301461109023, 0.9161940306090219]], [[0.31859762111288015, 0.6498275806237892, 0.049253826182892424, 0.3399726366927953], [0.13203111529103795, 0.9187197733357204, 0.15828381515363066, 0.8552265714982886], [0.3211672860522733, 0.686754976512216, 0.8141370937685493, 0.6826994695222457]]], [[[0.9058221441434193, 0.29350893736780115, 0.7667179622149897, 0.18687607510938786], [0.9325800705518973, 0.14339473141811554, 0.88773972215784, 0.2676930501234882], [0.33429302842191666, 0.5341924721287644, 0.14281936869723577, 0.4602475363281884]], [[0.07349007933908325, 0.11422326377739855, 0.4066520767177183, 0.687769285423676], [0.3100033661351028, 0.3646038618828017, 0.2956387906059591, 0.917358843687122], [0.9501783295202768, 0.895283001002032, 0.7728862313797511, 0.9336719058414357]]], [[[0.19078607363328925, 0.8875910874695648, 0.9750372220028721, 0.944069680329192], [0.3806874249477603, 0.4216711838710584, 0.3743614577977087, 0.5079568699641672], [0.4138776666380225, 0.5930728321806223, 0.4039194566841643, 0.32843997994946483]], [[0.06410326414718737, 0.7699605748617742, 0.2847344843457197, 0.035408873873628566], [0.19420228752388768, 0.6677685910969101, 0.7119115665479769, 0.11058434746482482], [0.560967277558412, 0.7677515641237836, 0.1782189427232801, 0.25301597718652324]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedSolution_rank0(self):
      """
      test integral of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 2.535857e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedSolution_rank1(self):
      """
      test integral of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.4268299975671688, 0.09361293708477614, 0.9549701437881247] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedSolution_rank2(self):
      """
      test integral of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.6834043113728545, 0.5752109964172467], [0.7458319979411958, 0.32024558653357926], [0.9238053650256417, 0.6528949553656769], [0.3987323726031198, 0.08223551244447858]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedSolution_rank3(self):
      """
      test integral of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.07097131952568136, 0.3972818199819128, 0.11677105546260391, 0.7635272367931409], [0.5728766110013108, 0.3638552174941029, 0.9727992016781731, 0.3402841213977943]], [[0.22809542851048248, 0.6866415914177225, 0.631977686295478, 0.5393170991689022], [0.8681581778918123, 0.886519338325564, 0.3299698946762052, 0.9974286439711668]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedSolution_rank4(self):
      """
      test integral of constant rank 4 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.04888598251323417, 0.425169355069582, 0.9014511799771826, 0.9689405703386421], [0.3814504254086013, 0.12410577360566977, 0.020937949217290197, 0.21001566389335147], [0.23796777445753858, 0.5352008086510078, 0.2336168458624136, 0.2578025355879793]], [[0.2330028736898433, 0.17748573334982687, 0.9475630785413549, 0.05342917352446275], [0.5182240164603986, 0.3666684173749777, 0.9332913759400298, 0.20675801650623682], [0.3597366175104, 0.8913689107298399, 0.2677254998209949, 0.7440723275688989]]], [[[0.5915544655907723, 0.3395254412380312, 0.7636878463116414, 0.07705652861339685], [0.6437698551618922, 0.7807416279208318, 0.4345787862319862, 0.5966598477731073], [0.7742395309328554, 0.6324658375731044, 0.3292287946397966, 0.020309797874582425]], [[0.6064161678114954, 0.7494592233471215, 0.36579848855690866, 0.2914627555693253], [0.4702615872987206, 0.16591204278516503, 0.5156597012702854, 0.5597309384174523], [0.3598116778251317, 0.14460953819529665, 0.45308757234925123, 0.5340065856587909]]], [[[0.948879016062239, 0.06749721905368877, 0.3041850652424154, 0.24398189016441973], [0.9623927111054017, 0.741984063558802, 0.05119468413577388, 0.010587405038639863], [0.7035652945133798, 0.8327013680055799, 0.07450985275348243, 0.9800813112195097]], [[0.1468801947403966, 0.7547497939666081, 0.8202564613249552, 0.9062120719623148], [0.14811076330037287, 0.10797461995232283, 0.18881072032166535, 0.3364756767983582], [0.28412532277396485, 0.6192103569358895, 0.13652120492980369, 0.6647739115544373]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 5.697525e-02
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.6274206750508148, 0.5126509781188422, 0.47672887492315563] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.6339556318202565, 0.6963610525970986], [0.6612730508498957, 0.2899527068701305], [0.3984650955518587, 0.5168579563940039], [0.4896057977959707, 0.5083072856451893]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.6890806501388854, 0.7461005816498653, 0.7744320396270987, 0.53761338042121], [0.4478687383318837, 0.4981906041074615, 0.4397013505761801, 0.8592448919691478]], [[0.9953472659923568, 0.008334020940117703, 0.9487518218570131, 0.31444750536588273], [0.6963296218936683, 0.74883901287094, 0.470037015804609, 0.9924073301981647]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.5538787010321746, 0.6809331959633712, 0.31499000100729213, 0.5881432405864642], [0.6904417675116347, 0.5872182155311177, 0.7614906435269186, 0.8032038054087768], [0.9358651525968185, 0.6085469879095277, 0.5707599044640953, 0.31120929020009136]], [[0.23378587879059565, 0.7582829367137257, 0.5161224118197779, 0.9148037089923455], [0.4380895969708932, 0.8565411631310034, 0.9146652665909885, 0.9190626384856223], [0.6210593154808267, 0.2517767138795306, 0.913800679293643, 0.30889708201212895]]], [[[0.6559129371831308, 0.7012321039570502, 0.5786655205988063, 0.7627653044758473], [0.07958020288114764, 0.9651085511533006, 0.12968150404755907, 0.9924384269005591], [0.3143087763790522, 0.8905950662166556, 0.22347872491130727, 0.22195310704734472]], [[0.7732242502611125, 0.18116587003279272, 0.8801575055149456, 0.25749781670448735], [0.9283188387557902, 0.8004381808647438, 0.5075826393071056, 0.5271505962593412], [0.712040781694003, 0.03776484099001054, 0.8666054153596006, 0.6425544071607274]]], [[[0.2720285539566093, 0.7161060018340641, 0.3745766379314661, 0.6134627957356624], [0.9991456123542806, 0.267501542563255, 0.04275281530361208, 0.2820108419344193], [0.35601651032117776, 0.12117892361281235, 0.9637212436643023, 0.5720578145504013]], [[0.2121469812997656, 0.7190923591312838, 0.7117507992237072, 0.5363127159057264], [0.5755807868239908, 0.20373845383016975, 0.44698435022679395, 0.5467200334848109], [0.7190654266231091, 0.5545016831953747, 0.20405653476336094, 0.42666086738863374]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 4.281111e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.9119166132437209, 0.7193640444407688, 0.7752106239006539] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.920019618468412, 0.6183536153487706], [0.8351179241066686, 0.8417390670486351], [0.7689477851071664, 0.4100294449138593], [0.788980334037139, 0.5355327787781378]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.15603251749908054, 0.7713665980850889, 0.40913341535830683, 0.14348814170458113], [0.24087972864752505, 0.6837212106612998, 0.8468469725002665, 0.8486963519535567]], [[0.814858283175183, 0.7749076001478378, 0.10429015653855822, 0.9604614664411455], [0.5262382060538094, 0.04475610308836975, 0.15657970935724874, 0.2955351611234359]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_ReducedContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.8499927230080268, 0.7651408114531798, 0.19472220813514152, 0.2912066962951263], [0.17161975008821162, 0.6235904851241167, 0.6926718978383687, 0.03596315749363932], [0.47645485793187003, 0.5934733508578781, 0.9755540247403202, 0.16126697777372656]], [[0.3449079320485575, 0.7022417911573737, 0.5172424186550607, 0.9562513796690906], [0.8471237244409835, 0.48388131654156963, 0.9030524475654612, 0.2601395016535005], [0.5007464348480165, 0.1267055820596339, 0.5085935955678504, 0.8852851284070675]]], [[[0.4642488641279663, 0.6146180255135646, 0.6464710087424905, 0.08189088323751592], [0.7670130656856871, 0.2555172595625965, 0.4181058765657899, 0.00461747266586654], [0.13619448265631662, 0.22904330663704042, 0.43505775862759677, 0.023380856157918517]], [[0.7350521817816734, 0.37209335964378154, 0.6718948168733091, 0.06599105481833689], [0.07366391068822298, 0.503038507995098, 0.6105391461572015, 0.8705868448634384], [0.7754525785156687, 0.7962972586915807, 0.27168002159922233, 0.4834376117600926]]], [[[0.8926567334195302, 0.37608452273771553, 0.5744192131016351, 0.000916148691289087], [0.5346365267330431, 0.9979221034616841, 0.1593844359344746, 0.9476657869500834], [0.17222138097477802, 0.6922792638239684, 0.8907545687407497, 0.9596308200338224]], [[0.9430159192937225, 0.6725050190922935, 0.4740029634637012, 0.30194248558479697], [0.7235128544520342, 0.5676295167243214, 0.9755111318024311, 0.26101876345519226], [0.8267770668268078, 0.9549760906086663, 0.79168797982497, 0.36061663064414984]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Function_rank0(self):
      """
      test integral of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 2.037380e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Function_rank1(self):
      """
      test integral of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.8230483022652137, 0.41491828309276146, 0.929037842165629] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Function_rank2(self):
      """
      test integral of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.3519045857179769, 0.6753297681514084], [0.8206574875957522, 0.6225811604809537], [0.938370226644646, 0.6171627592957175], [0.4777850154573239, 0.5350269790755375]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Function_rank3(self):
      """
      test integral of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.21184989477285077, 0.4153520280040581, 0.23619783115789095, 0.3525885843845221], [0.26854726251996786, 0.2553387851894976, 0.8053594174222334, 0.5110094781509322]], [[0.5834385993196044, 0.21062616402780865, 0.836593001725716, 0.9771745725112635], [0.8407188794278746, 0.5561698617311589, 0.9516362488095351, 0.15595029539556793]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunction_fromConstant_Function_rank4(self):
      """
      test integral of constant rank 4 Data on Function
      """
      DIM = self.domain.getDim()
      w_int = Function(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.5472765440417001, 0.5602290920780002, 0.4376560194950483, 0.4460893800290828], [0.07076878419270882, 0.1940474640201617, 0.961204370627523, 0.3287172699053642], [0.6122226036271041, 0.381597865147278, 0.09917254335820536, 0.1370806224905209]], [[0.3402868627291231, 0.5344957401159944, 0.4797500144978618, 0.006085479526618642], [0.533712145983013, 0.30298209329564163, 0.720450617559768, 0.40285925098045927], [0.002608844295875934, 0.48379830433677595, 0.39529982107792216, 0.4474024956137199]]], [[[0.5425521712407491, 0.8792342054655766, 0.4549268556811875, 0.7421596289979908], [0.06669135276309357, 0.21851352554485803, 0.2747365808471085, 0.2526301236613091], [0.9721775719969312, 0.1704749138278141, 0.2738441799692226, 0.6934188843051623]], [[0.2581261241139461, 0.22622961003117148, 0.7362724008432099, 0.2515803441897436], [0.9569560119849841, 0.27583817114420484, 0.10913422284131336, 0.1253274161541137], [0.11249209821502093, 0.48027568311489377, 0.18828307518069587, 0.5154452143876268]]], [[[0.7185064429740038, 0.31534017308507833, 0.1244253370984233, 0.2215772400213849], [0.9156400900969264, 0.1679873639227749, 0.4387140275817699, 0.131617072349817], [0.889111915603661, 0.5739966914380886, 0.614935584810949, 0.6625601853857063]], [[0.6350904695416051, 0.5187169214950459, 0.05287033871063962, 0.32161249120087787], [0.26001525119770363, 0.9842192431832548, 0.22116734931269033, 0.5681644033136674], [0.2753818012576793, 0.7844260672376742, 0.5648016743355341, 0.9379734426126127]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Solution_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 8.517828e-02
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Solution_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.909009908881254, 0.6252787656214178, 0.5236344882153815] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Solution_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.4569280343233061, 0.99994398006045], [0.8346270018173735, 0.26892721364406], [0.18327521424468574, 0.9135755108734025], [0.7079642912526467, 0.09548139910732822]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Solution_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.6398809515672143, 0.1907895317197088, 0.3619926970150006, 0.7883202090854804], [0.947100016653321, 0.2756526945165554, 0.3351820127132842, 0.7123071416102558]], [[0.3088586388137251, 0.04681089899773838, 0.9846412222842916, 0.6919804593747575], [0.22756602070702903, 0.6491120096026316, 0.3026545788042522, 0.4729207455635366]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Solution_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.4736636541199598, 0.7225893011676272, 0.160690298387566, 0.8694380239175357], [0.7402807376640161, 0.08188882849219259, 0.5257501987189352, 0.5064844383118615], [0.4993648502761342, 0.280405287738689, 0.21111271180814728, 0.07680263985268043]], [[0.8059379576595804, 0.437863585007225, 0.1501232095605769, 0.0013302063412240495], [0.141964039832835, 0.5621070793393457, 0.9289289761649706, 0.0186544454801002], [0.5374234175652906, 0.6219773648694972, 0.5723435517186105, 0.11704037785040378]]], [[[0.7763689325671491, 0.962367653968557, 0.29837583166085335, 0.7574021806959697], [0.37666508982495917, 0.006446153001960275, 0.6132983370427808, 0.8198743796987265], [0.6393716904226334, 0.49334635697946727, 0.4111429625292413, 0.9895272149003305]], [[0.7793161564197099, 0.449049955688265, 0.5552981301421128, 0.479304486957001], [0.40523304830340656, 0.47228943521154054, 0.4125036007132251, 0.628358447284809], [0.9441569880521398, 0.09866520303716442, 0.5912008389081345, 0.08241787906634179]]], [[[0.8252304858131965, 0.18664778296772988, 0.14247883967306219, 0.6460308957180753], [0.5054485237841788, 0.297686855438963, 0.6354408468696987, 0.11187238444921477], [0.2202437279722571, 0.5404111371502452, 0.489806258225682, 0.15371480161606266]], [[0.9181477081180152, 0.7608499387543418, 0.663586587756242, 0.8044051823245268], [0.15622045821546227, 0.10818532651930468, 0.9042873401245518, 0.5294834135962319], [0.677664385822654, 0.2686737737933579, 0.9138923933269006, 0.6264786515039725]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedSolution_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 1.512851e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedSolution_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.8626922935900396, 0.9261868789114589, 0.9950559696728337] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedSolution_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.0016975599098827399, 0.1411400054684817], [0.6815369443328064, 0.17067286133044235], [0.2857640098739478, 0.8963948973539344], [0.552274821459423, 0.4152743654542329]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedSolution_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.01017408888955662, 0.06012522609838544, 0.6362937233205982, 0.5214972339083497], [0.16996285981282255, 0.5569108819054138, 0.8894554796739045, 0.31722252353974056]], [[0.09183413889776915, 0.971354783337366, 0.2965039697192422, 0.8235390399776398], [0.2902058419117539, 0.08973630080163553, 0.9739301580248261, 0.6330707014809956]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedSolution_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.0649152522072598, 0.4217961081298678, 0.14554484207859186, 0.12892489532195006], [0.6212360168568195, 0.9610665689200276, 0.7657421699950827, 0.25715091493172126], [0.7445559085827157, 0.5642968653073819, 0.048336462666391355, 0.9520872016510461]], [[0.9077718137279377, 0.10577254183973261, 0.23657142588086366, 0.1739200096190433], [0.17493454763563987, 0.40890523568581283, 0.3585753972292297, 0.08143087275967997], [0.7428110717915413, 0.32444818018047816, 0.3210231430590723, 0.1079338780535547]]], [[[0.4283100352321437, 0.21532366162275174, 0.12878181795234522, 0.6765834169737814], [0.0533321036065747, 0.13639167732869295, 0.15955051797330522, 0.2071944766488114], [0.5850241644563438, 0.4549434666426133, 0.743635003415812, 0.27264768664591144]], [[0.4042471370743217, 0.013137195614978725, 0.4838199045105914, 0.028424319719273505], [0.4348352298541047, 0.3268256581973241, 0.6034238334093851, 0.5243293452037754], [0.5370300628296355, 0.7698259360784074, 0.9689112767934063, 0.6051916915084051]]], [[[0.6903227225991604, 0.4459629499247779, 0.8019772657955326, 0.4577573631095593], [0.23742904188640923, 0.48382069248710646, 0.4162436077001709, 0.667680967793053], [0.14051165857114412, 0.35112768604616473, 0.057969226280832786, 0.7472763458201801]], [[0.45466068614071975, 0.44626487060768416, 0.35815027706293356, 0.28293002538300416], [0.48152996277407867, 0.45502806573199217, 0.4117373256975757, 0.040265796575904456], [0.5582939474986471, 0.550192317083097, 0.8112033255521558, 0.08302796562199599]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 4.412771e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.8096320144805804, 0.7949919950140597, 0.12419505451588397] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.928335061070321, 0.0891693154227382], [0.7452065883380177, 0.38370835605323905], [0.7506470548203787, 0.5924185306979014], [0.4161128712832869, 0.8098622889054996]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.7505102909150166, 0.8877248890752348, 0.7614152720688268, 0.009099978759054217], [0.6521693646058537, 0.8209114713350028, 0.9641664308360277, 0.7282302273377724]], [[0.1471232021951565, 0.7169616849992533, 0.5075492439550722, 0.01252007188997828], [0.22081657663158727, 0.08927232094446291, 0.06562874797373774, 0.12774046948515116]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.3240639739727852, 0.8166361837511436, 0.8078643288042248, 0.8369397328906489], [0.908087477151493, 0.9948300315453555, 0.014936795975283745, 0.9133848472090799], [0.16288354082646805, 0.32125613507928064, 0.5326080936132516, 0.14596079004162832]], [[0.8233368201432943, 0.1284909883980726, 0.6394161234002498, 0.8423335386747732], [0.3044349048615199, 0.4909206202074553, 0.5213122780585577, 0.03631937071796554], [0.023771745864425764, 0.12110598541579676, 0.08872630078365773, 0.9491548342950022]]], [[[0.37789775532824443, 0.862620156229067, 0.849987142672508, 0.040015594180522474], [0.21604277050303666, 0.6213166755277112, 0.11120452938230474, 0.49350813849821373], [0.4490406877573889, 0.571636527946194, 0.9217940566002697, 0.5943290646364829]], [[0.12245365743330128, 0.47264150044844255, 0.6386533914666578, 0.14732274247030985], [0.884108185009124, 0.4102444599690005, 0.9305489803997793, 0.3102784475337801], [0.5643941860894383, 0.4355092777175219, 0.28644558933293807, 0.5178398321587948]]], [[[0.38160257630529615, 0.6014446243396778, 0.10037529838725223, 0.48485792563344676], [0.48868561043981706, 0.24793557059949678, 0.27876129109296677, 0.186323697866716], [0.2162082557725693, 0.32592030420567975, 0.9942501479625927, 0.8838852609090097]], [[0.8603384332064178, 0.2844693325411508, 0.06607212397858586, 0.9928783072287632], [0.40213533764335074, 0.3487716569706739, 0.33602022475589455, 0.7588798355311545], [0.5750222436077979, 0.38245661816819176, 0.4275499321898034, 0.9661391209215562]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 2.263831e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.560160926240131, 0.4544937506040815, 0.41428585915821403] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.25425363272669266, 0.06679575800652149], [0.5908458559644406, 0.04827893409058204], [0.08506081136354782, 0.9785550550114211], [0.600194310885647, 0.7230276035118565]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.25145799710326555, 0.4691756359058713, 0.6109016354584982, 0.2595524377370193], [0.8716397205291253, 0.9484470557126324, 0.8004533470969156, 0.09701398057708377]], [[0.792760457513162, 0.20865004351148364, 0.6137699963302109, 0.09780000284045842], [0.9647820638298652, 0.3579432738674304, 0.01289075952701857, 0.6146962788215177]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.9193935413798696, 0.8374491187507355, 0.9891684749379919, 0.7326781077496545], [0.22280825901230972, 0.5103009634084649, 0.08253619866563344, 0.4711871651108336], [0.8153799442178684, 0.1746245560376336, 0.0732999100758519, 0.6106553404644791]], [[0.8989557902416844, 0.5651359713649547, 0.12200062281837076, 0.7126870177144767], [0.2105956006916413, 0.43225587170659274, 0.0001768248529889771, 0.591498149433938], [0.41415106445170835, 0.21054081005809333, 0.07122046959799333, 0.08678742887266622]]], [[[0.8807770067247797, 0.5140564481913797, 0.6061410280165955, 0.7821078618776464], [0.4570722654263897, 0.03339427239611714, 0.9242235253634065, 0.9193752471978738], [0.08029535989742353, 0.041133798715233594, 0.6300616993695234, 0.3893725551235887]], [[0.17375904646317974, 0.7626999981203012, 0.1682852550089161, 0.5082358584141876], [0.3003734196702933, 0.35671129674179936, 0.1435541362157392, 0.9540226000495032], [0.9562773651317534, 0.8634817924867246, 0.8229345431600459, 0.42125909020318986]]], [[[0.08170906259346145, 0.7363915186668427, 0.92009101627818, 0.3285223639139645], [0.8716721957638899, 0.8125467491248831, 0.7608852632251512, 0.6501111868348199], [0.9451981543620576, 0.5736288437793872, 0.23771944871322004, 0.49366887506818236]], [[0.08228858706141651, 0.10899932062780082, 0.32543608683734304, 0.829528552723553], [0.02621791412544161, 0.6815832803810583, 0.518722636445362, 0.6005740667385301], [0.15578465750061699, 0.733329914505852, 0.9680139127668663, 0.28118083166205]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Function_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 1.062875e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Function_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.9932292010205616, 0.9859703624522872, 0.014483648403172511] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Function_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.6350433900483379, 0.8341743393234992], [0.4565368003632364, 0.10720521818038953], [0.5272905787552593, 0.868515959742127], [0.42165865172971695, 0.41888149068610736]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Function_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.8543567984048774, 0.7519146259306376, 0.5421869174543705, 0.8037458973089099], [0.4661254267484708, 0.2695731008539145, 0.580359786623633, 0.08980477765083406]], [[0.3389082618461453, 0.3623055015985228, 0.2882092920170424, 0.5959792213710673], [0.5889191013255423, 0.8989882017502737, 0.8003187477670938, 0.7591733118333461]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_Function_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.2239149682846452, 0.609512192995055, 0.8935097480772161, 0.5791170465919767], [0.05193796763248604, 0.7366455809393655, 0.5008111967513614, 0.5318533692717], [0.7872772624758406, 0.10351224192682162, 0.5805814492443875, 0.7153935488688926]], [[0.5631835031235787, 0.6325542047690368, 0.8764219869562522, 0.6607641724154036], [0.9403401765370057, 0.2665338123065166, 0.2484043667632695, 0.8135141574418217], [0.36705333763650305, 0.16554420905498202, 0.03253178327097417, 0.24241617730585352]]], [[[0.977164890375996, 0.41800037213850794, 0.6481101488642553, 0.265521086144456], [0.5488510610002649, 0.2227291095677365, 0.19147751631591703, 0.4218282702325916], [0.554114710716545, 0.5092585460215068, 0.792117293182317, 0.2758999729756514]], [[0.2461078137637408, 0.9127652904987122, 0.05665626147596303, 0.1014307161583683], [0.2917984812634964, 0.9882205084665034, 0.31020330269352436, 0.2650834506935458], [0.6485569139431261, 0.08457969368208229, 0.5990827160103767, 0.7856742923620585]]], [[[0.3907066453222141, 0.19352431954153126, 0.5875104360465303, 0.09826006226495576], [0.0809662010182276, 0.17557888344619466, 0.4570453433575252, 0.5426379464612267], [0.20035115514983803, 0.9148319309700159, 0.592236010303012, 0.7311409753461974]], [[0.9157847436429263, 0.6661285058998175, 0.6788218529593951, 0.3182427993572823], [0.06689537602960194, 0.5319215191132524, 0.647958420084565, 0.3373146909277651], [0.33791064694785566, 0.6511328915478607, 0.8525432758791744, 0.4740756159633067]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedFunction_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = 1.726541e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedFunction_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([0.7701007257051183, 0.17510251639968355, 0.36130879175427344] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedFunction_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[0.10304176491227035, 0.26008328919736046], [0.2848252954331827, 0.47168667816023635], [0.19645959012617642, 0.5886734455812066], [0.2118547008742916, 0.4857099576120658]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedFunction_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[0.8359834041464674, 0.9850094569085539, 0.9552283415691541, 0.7827015389718909], [0.7581396540232714, 0.04891620585421541, 0.7231870144654986, 0.9027441630302253]], [[0.5937330544065352, 0.40812310216812275, 0.3222516549831246, 0.42965186793241184], [0.719134332409656, 0.21745463072671134, 0.5448849299640816, 0.054180894169713056]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunction_fromConstant_ReducedFunction_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = (M[0]-m[0]) * (M[1]-m[1])
      else:
          V = (M[0]-m[0]) * (M[1]-m[1]) * (M[2]-m[2])
      val = numpy.array([[[[0.16622168988131203, 0.833601276023346, 0.12029638158794509, 0.3436296913883181], [0.45016423655534865, 0.8757486464699629, 0.24763236746255413, 0.021859691517753577], [0.8800541875959128, 0.5091834832930291, 0.7417851770332284, 0.9282246894222174]], [[0.2870958866395711, 0.3827825210689664, 0.2799926588199353, 0.42997238170669283], [0.8225054119525542, 0.49943389457758014, 0.517487708369212, 0.9725002359704886], [0.7331407392131398, 0.7582190665983551, 0.8132884553647229, 0.42199212649817597]]], [[[0.6125597061831242, 0.2606915883498433, 0.1840308403735167, 0.5703971635266702], [0.9494913392265497, 0.46846018822578217, 0.2066063669917363, 0.197809652226558], [0.44233863685954067, 0.06474815693079683, 0.25014873034484986, 0.6186385905584857]], [[0.9960092861777984, 0.3842740002136741, 0.33599894816554976, 0.03685222692005119], [0.8105418009705768, 0.17418103022261622, 0.3912326672533222, 0.9325916576394074], [0.4187611400906006, 0.2707020505343579, 0.4050627036180616, 0.6928376088268758]]], [[[0.014020759639266744, 0.7949954961177269, 0.0639898114047136, 0.6909371768285926], [0.5535019185905485, 0.9871813374854082, 0.19285295518969714, 0.8096793755837617], [0.39543772854079573, 0.32667629083338057, 0.6644533266372749, 0.1950638482841156]], [[0.27634334790818604, 0.20504772162014806, 0.2047851505985102, 0.9020163267226452], [0.6235960792034383, 0.8347768741904003, 0.8092235605022734, 0.5766191498309793], [0.878952817788005, 0.6292656238267357, 0.4910572300609879, 0.6033998014793811]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_Solution_rank0(self):
      """
      test integral of constant rank 0 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 6.418292e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_Solution_rank1(self):
      """
      test integral of constant rank 1 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.7084809410883933, 0.5016752351830107, 0.5323627927361344] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_Solution_rank2(self):
      """
      test integral of constant rank 2 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.5695567749380169, 0.49944570898315155], [0.29293911435207953, 0.5781248194000584], [0.46812675051050223, 0.9303176405737067], [0.9146537021424015, 0.3473271826817299]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_Solution_rank3(self):
      """
      test integral of constant rank 3 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.03688634331385876, 0.854388811904553, 0.4462128563014265, 0.6641818874753929], [0.3716940396906563, 0.49671578455054133, 0.22758867189062593, 0.47111980012527876]], [[0.27103861928327255, 0.732502776362671, 0.724305899224255, 0.33371770111133703], [0.6326089986424137, 0.18668341220728168, 0.47337376288236643, 0.17536101474056975]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_Solution_rank4(self):
      """
      test integral of constant rank 4 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.36210210606374327, 0.899342450821971, 0.10701596650232481, 0.7085064083995116], [0.6898989557966647, 0.44972004569887125, 0.07781269632927479, 0.8266992922824304], [0.7160909371223189, 0.5467583350599373, 0.3789262995536581, 0.9697565856809153]], [[0.6966425715064388, 0.23992357852723611, 0.06138390986829778, 0.6076176251135184], [0.04970193383791521, 0.21902550671078147, 0.5708342470016115, 0.5770667495314805], [0.08794157377159373, 0.37454856400860215, 0.31636625912967264, 0.1916667874449487]]], [[[0.7964737393348712, 0.3356878815656187, 0.012989835015197126, 0.8303810592499002], [0.4147443862568705, 0.9477103205662467, 0.8888381239469976, 0.57482042884047], [0.7745082103626527, 0.5315926671390498, 0.5858399412903301, 0.8373905964235883]], [[0.7146956116383965, 0.7049471408628909, 0.6955345611519648, 0.8492225709166494], [0.6276069436066763, 0.3361245991341132, 0.5519637012387604, 0.15465428481087629], [0.6017726490193221, 0.5917978592085213, 0.4968699295155734, 0.27734654297728667]]], [[[0.4575794806927387, 0.37750282705278304, 0.5643926586239748, 0.7854152220050639], [0.03440630248789167, 0.7671628387742069, 0.5509000555440794, 0.9826960180515398], [0.7760865110123799, 0.5480116969294784, 0.09984378591657139, 0.9821811062877345]], [[0.403885943789862, 0.8889654869571993, 0.6482146861581468, 0.09091947784542487], [0.5107783332840654, 0.07245936948308962, 0.3807007107525613, 0.010548955510103042], [0.9058208242578547, 0.9229530976998306, 0.22271032149597236, 0.9469833470784652]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedSolution_rank0(self):
      """
      test integral of constant rank 0 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 5.222967e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedSolution_rank1(self):
      """
      test integral of constant rank 1 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.32908815162932004, 0.23381584742410455, 0.779532112085225] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedSolution_rank2(self):
      """
      test integral of constant rank 2 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.34461236996314637, 0.6099892413833922], [0.22918227716063, 0.16790197671710316], [0.7180026001107437, 0.07165288910842527], [0.8800185787818955, 0.33432525040084704]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedSolution_rank3(self):
      """
      test integral of constant rank 3 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.836047003112094, 0.2874456415570974, 0.18930815681621238, 0.007787873920402033], [0.24777609833333947, 0.3465711644454219, 0.005333725005916667, 0.5297153286949997]], [[0.2743349603208164, 0.8347763796469766, 0.47895248377561606, 0.6464509388298519], [0.1748164614623693, 0.07293736530373207, 0.9543770703569877, 0.37899868821083116]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedSolution_rank4(self):
      """
      test integral of constant rank 4 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.5500132380472977, 0.7592534138959203, 0.5966952968226467, 0.566827817782952], [0.701227578773762, 0.8555504552418488, 0.1265555991874847, 0.203622664259563], [0.6104929095527369, 0.3689844768922672, 0.4348645257249678, 0.29946637263876075]], [[0.3716481578640991, 0.5004202035452506, 0.16209533182471525, 0.23056859984638023], [0.9887244763470506, 0.36784771471114563, 0.3540733503000364, 0.16702636818879146], [0.8325747574520844, 0.311433346785685, 0.7625642148711727, 0.24502718519855737]]], [[[0.041197091947053455, 0.2935180102018019, 0.9637513227528277, 0.2812239936099513], [0.2495590724110056, 0.45051353981139297, 0.1593311672288671, 0.13486845289382143], [0.3983654015057956, 0.22471538098974064, 0.14269589277302863, 0.2679160505595194]], [[0.8275478702132529, 0.7529801943173022, 0.8535111315512882, 0.42251795005917636], [0.6082564581083224, 0.15590236147278114, 0.3931542965505761, 0.29805277517545714], [0.08598287487667167, 0.10712826443259349, 0.4449696955477027, 0.57620916009328]]], [[[0.33660560315747245, 0.16390520747002058, 0.8591992711611391, 0.23085858177969687], [0.23521966866189714, 0.13849532763683048, 0.630251100648929, 0.17649300324482275], [0.46941848487440774, 0.8426851396204019, 0.10638802870783848, 0.9474264904729046]], [[0.24516653518946463, 0.1784557253530057, 0.8394888085622392, 0.5024408150111394], [0.4259830471048387, 0.7691817604489838, 0.27328000307817735, 0.7765398250532942], [0.30658691816118655, 0.9616778100885319, 0.9445330518122601, 0.9355540201488441]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 9.070670e-02
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.6798759806768097, 0.945849510039827, 0.6480925836359033] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.03553172482845135, 0.5632470145215643], [0.9121094412482305, 0.8885408533232332], [0.09138389125986845, 0.3298465909192546], [0.9284876255301888, 0.7236200962204311]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.7448729957375646, 0.5090051351817642, 0.7957005950534762, 0.8361418700818971], [0.13690074472773817, 0.19867212548568736, 0.8022301234989492, 0.7777630020237758]], [[0.2186822945406165, 0.880223913989578, 0.03910270802848481, 0.8845660067359125], [0.7016185463419419, 0.774157683821756, 0.09143548361400677, 0.595892585600503]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.06383647545979498, 0.38906990060488433, 0.46149399444958683, 0.776473440100299], [0.25264454737254216, 0.43935115541717973, 0.8801524458237246, 0.7157420601441927], [0.6491654397119059, 0.2939967480963315, 0.1327933133974264, 0.656469831855535]], [[0.5866427294666772, 0.9880610396758992, 0.639221573668538, 0.9936420716620719], [0.24657339290842062, 0.9864885499704058, 0.9842384898460576, 0.9584527897489776], [0.7223195171934346, 0.7360271104282831, 0.8172424566867642, 0.9742737792485618]]], [[[0.5149475113584814, 0.5755072658544693, 0.061070371235920407, 0.6807482619352412], [0.567955961588093, 0.22763600129760542, 0.4220263716065782, 0.47403639363582806], [0.3972157966493314, 0.807608950782692, 0.3377697917348029, 0.8134447927081143]], [[0.24280576136460008, 0.509144414625858, 0.7611489848051081, 0.8431305542231905], [0.845965628862353, 0.797348781829697, 0.7886132272356798, 0.4948918672596838], [0.3597785739547653, 0.21291968486208446, 0.29629482176115907, 0.19803895267638827]]], [[[0.12272833360640356, 0.8874649679955368, 0.3476765803003051, 0.137416140777753], [0.390645195077464, 0.7931292782496729, 0.5202471591096526, 0.6775670727039582], [0.6157725081331094, 0.612859173565169, 0.43294398845783777, 0.9915646842238843]], [[0.8758155969359598, 0.0520449363991935, 0.22003010518777044, 0.9984509836262333], [0.25798758021368084, 0.902752547695231, 0.5725504546076325, 0.39376437029862954], [0.6196532023430871, 0.6749110086247015, 0.41034155465344635, 0.07431812713386221]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 6.706662e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.1806047340420741, 0.6420314431276954, 0.836680379062931] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.3783328333823852, 0.24075046829035174], [0.669637086279931, 0.6208991268058974], [0.4600083215157381, 0.5561111223678822], [0.7685157503036292, 0.35776861625439493]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.5014500470804437, 0.6198223113701952, 0.8584742740844648, 0.6592436940807501], [0.09028758800523495, 0.062208079249841286, 0.14764107098656565, 0.7584667784369181]], [[0.8275249750545722, 0.035913909955379686, 0.66683988891868, 0.1847721165397357], [0.5120354652966768, 0.2826987153028725, 0.6211072393825734, 0.530549500643056]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on FunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.3344140442229344, 0.07384742133122324, 0.6296026720854615, 0.11619842745365061], [0.5476185677003443, 0.6711125124766801, 0.5247680213783656, 0.5502805188489758], [0.9071428516928248, 0.9674898220168037, 0.9263101496882113, 0.17993746162412105]], [[0.9447161653672127, 0.6995035815535086, 0.5052662909904108, 0.19527507806475541], [0.12174564423962597, 0.06485487359517372, 0.4904294936718465, 0.6509864692598115], [0.8439161891968794, 0.603995457376195, 0.3008052035495644, 0.9585900001117356]]], [[[0.22887207908895368, 0.038466847676040694, 0.09118120412099451, 0.1721716174994109], [0.08484270493733814, 0.5642509643855499, 0.7460868534384093, 0.9399995763284531], [0.8340850665283215, 0.3696071934274232, 0.18686895112325297, 0.9721099213848409]], [[0.14269964099802157, 0.8720572407477846, 0.735123088844361, 0.8214269511861152], [0.32134129011912593, 0.9311155344298172, 0.06220378698195428, 0.9052490397628026], [0.05251737078742713, 0.8201187552971062, 0.6947354607130618, 0.9631674355704753]]], [[[0.45677664029207776, 0.6021482735524599, 0.9915947848852379, 0.6781146312052093], [0.48068274926895205, 0.26520835580331537, 0.4445203912730439, 0.3009036927784602], [0.4438203206574812, 0.07431129522156255, 0.13559347212416573, 0.003917748141243127]], [[0.22901747401051775, 0.310120562228426, 0.9240248391500622, 0.5794701587785138], [0.4208972918287953, 0.5735482028795404, 0.5123769366914721, 0.6537985411894108], [0.7125703365202771, 0.45727382548797946, 0.4919675813816522, 0.150540323931747]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_Solution_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 5.521989e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_Solution_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.6812810573536356, 0.10968096570012964, 0.7697488677961024] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_Solution_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.4882196511630268, 0.8951141848094238], [0.5817958311582203, 0.8215955596831104], [0.6326606034201372, 0.08664054663038256], [0.9527137138351295, 0.9180643836127376]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_Solution_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.9160532653262539, 0.7215798276701644, 0.9343287576915817, 0.5888525958626359], [0.5713864326300565, 0.5104428608013275, 0.13703239235171416, 0.8441761900919201]], [[0.601603122595341, 0.9965693098197328, 0.5739489978551423, 0.2515124269928284], [0.9570900963849621, 0.5232861763658169, 0.36738734113776095, 0.35439051239644237]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_Solution_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.2917319241795221, 0.27581659210027243, 0.6224072413462947, 0.3440615591731835], [0.15876097698196856, 0.0432338704923213, 0.32094123494386617, 0.7840201486918309], [0.0643366348469595, 0.5074984620115042, 0.03989554259841266, 0.0889303277634228]], [[0.09247056096964923, 0.13722088193869197, 0.3326011261134003, 0.5874085553839858], [0.8137212854985243, 0.6416431629059236, 0.17734292522020545, 0.6301757745600586], [0.06992286340899201, 0.282872255082392, 0.010675002842713055, 0.5404060890435098]]], [[[0.11555732313796008, 0.6612158122055297, 0.10028478066254376, 0.035860945349660556], [0.3519918291339904, 0.07293643968084573, 0.7192284937633354, 0.41677607134743433], [0.06779693557842237, 0.08098176041306171, 0.13739643960243253, 0.014803250675712754]], [[0.40810151363855607, 0.4774389407164168, 0.3455337744082687, 0.6164389564400223], [0.6176632300584821, 0.0033413570266866977, 0.7924241518266066, 0.03592524925212004], [0.9930233952416735, 0.8945468808194054, 0.628151735344806, 0.8093334260516684]]], [[[0.6155014092846349, 0.3671815628989522, 0.031854783064013126, 0.8860342906119493], [0.7110994202370878, 0.4409441503039453, 0.5704446997065151, 0.2779117523356358], [0.3947901505316598, 0.39890587150375467, 0.3150747614539722, 0.48100397291017716]], [[0.33180588726230986, 0.36249287231499205, 0.9747689583506707, 0.33774236247886713], [0.8957182479580069, 0.09586517187696941, 0.13130799242243718, 0.2182348926190436], [0.6834389607338858, 0.7285815772577423, 0.3224952860074962, 0.8236436774836979]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedSolution_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 3.357124e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedSolution_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.6702778817959021, 0.8806594285147338, 0.07700156829133575] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedSolution_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.753659675939755, 0.35544949244083124], [0.44163501246875325, 0.6025957223985594], [0.4493369875306744, 0.9344798447832641], [0.17983217222500303, 0.932082121795223]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedSolution_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.46301262561535406, 0.7729417147600186, 0.7764425369253872, 0.37534379980580457], [0.16785491050550438, 0.44815184305133815, 0.9660069053013806, 0.7231371643386225]], [[0.2720606303302372, 0.5746822745414204, 0.6473360494743179, 0.8705637758140671], [0.7862581867094424, 0.8099537567396707, 0.5546799595438864, 0.7696019380646827]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedSolution_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.20238328267865435, 0.03520622375939042, 0.36805748711548003, 0.10574644668385247], [0.28959401508768645, 0.7210783162707693, 0.19964346842659864, 0.8645858229044044], [0.8202081750272456, 0.5379525096713614, 0.4310976457662846, 0.8038451189821476]], [[0.7892155016572852, 0.4275124489142469, 0.5703199134538813, 0.11349104106429164], [0.0005286821168187794, 0.8081049635389059, 0.10065360307907756, 0.25061948658724476], [0.6376170651726422, 0.9970040233222386, 0.11781882515610997, 0.9557035301517579]]], [[[0.5137364546411383, 0.33928989908679175, 0.8028010031002306, 0.895671109038515], [0.9530081625008237, 0.5731371912985467, 0.2612323570200339, 0.46337176372077127], [0.4283876768125843, 0.5600146629833133, 0.03421199389379126, 0.07444293025641546]], [[0.3727084813059123, 0.6715400598270179, 0.8633101079991352, 0.7461173722284687], [0.7342460785920198, 0.9407789807521757, 0.28153172346198574, 0.3548189726786344], [0.18458016456081627, 0.2674402334150121, 0.90783551311868, 0.7990383200463504]]], [[[0.9350141614331576, 0.40438690904476693, 0.9516907613416633, 0.8430600217608702], [0.202509418252945, 0.5690132762321523, 0.8852332012111792, 0.029472677070449538], [0.22868726615907498, 0.4203191533385018, 0.9586654207333102, 0.12885180289414944]], [[0.8420573165681737, 0.20708711099012533, 0.07928977487581546, 0.47448148074970287], [0.8701886612417312, 0.5882406239797682, 0.49519561019567016, 0.07498564857239398], [0.748884786209675, 0.13540662596620046, 0.19487299200503583, 0.41098764433186974]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 4.781588e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.9771882822101213, 0.1353351682591536, 0.10301632081693357] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.8756680474082679, 0.7005861034108848], [0.21215541352861134, 0.7090441409652147], [0.9886029551235619, 0.16368365483570235], [0.9091431798460116, 0.41577065335116103]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.4002077234040202, 0.5821004059547503, 0.522439415415772, 0.7271992235096507], [0.8352778826129994, 0.9849692956816269, 0.3294115276888475, 0.5262923755538214]], [[0.36306077740520293, 0.7727890707810325, 0.3116492572347739, 0.3900050747564118], [0.7275622321960616, 0.35156090209041924, 0.3689226617815363, 0.022572746447523606]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.16930984024551765, 0.2052932822028719, 0.41724186552363607, 0.3808238005554927], [0.9894924941306565, 0.37854164892428, 0.6956064788798931, 0.4516403364833551], [0.7760847332712774, 0.1797249473852176, 0.5604470978404155, 0.17976533006790307]], [[0.3142199477437857, 0.1176878337991738, 0.8032154115210819, 0.7716192725126861], [0.5198627172137245, 0.9066649906889969, 0.36069735378155865, 0.36911710372340656], [0.9652439430451557, 0.7891472644920869, 0.7448256407716846, 0.6437909824485715]]], [[[0.557488151838626, 0.12319982342521996, 0.20749100347818927, 0.5704815459655864], [0.6520230678671204, 0.35497694129185464, 0.32131266825388016, 0.012422661115001565], [0.2990741449194453, 0.17013139970970015, 0.21414753551758592, 0.5703581370065627]], [[0.5043310922558678, 0.887379443711487, 0.38486947705980923, 0.5918897015304999], [0.6926995168123738, 0.5515205527517397, 0.37203868623992475, 0.17100420325918364], [0.9478707790534879, 0.031940648555831785, 0.8769829151483305, 0.002056456703795617]]], [[[0.5204023128045921, 0.542070692835853, 0.951759183041068, 0.8571458453609877], [0.9004196704240953, 0.49006922258850505, 0.8763926717446734, 0.6402971654177423], [0.7634170652272132, 0.5924667961042385, 0.38051361546994134, 0.551071791183054]], [[0.8509840522825043, 0.6811633346679122, 0.4959879964026973, 0.18570638186404065], [0.34759874973157856, 0.6343429757349038, 0.004411319981810435, 0.09696888952807059], [0.6054690630923655, 0.5209408454269387, 0.07280547868313891, 0.892340836460781]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 3.280712e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.047239407147456114, 0.5133833955313981, 0.4164357837777749] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.9504145850942762, 0.8711117499942452], [0.9048359700686238, 0.05653398657398545], [0.8655833613941641, 0.6737466524696735], [0.4279670670761089, 0.0827242551835875]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.8440403893235247, 0.9578372464373229, 0.575026757636108, 0.7513123535465035], [0.8167037981203067, 0.2499296955829866, 0.07361612888811064, 0.046535914663583644]], [[0.9924585235768051, 0.41577243069152403, 0.6083636811753431, 0.165968645679769], [0.576066016688442, 0.4694423081683816, 0.364417517482695, 0.7964148143397334]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedContinuousFunction_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.9135722983852855, 0.44050796245124224, 0.5139064494798129, 0.015990160366908368], [0.31180213893722153, 0.4466441812075328, 0.7673975285402893, 0.5560015482612994], [0.6634756287090158, 0.7917716280156422, 0.5979986547342357, 0.3657302483431226]], [[0.7673782062530351, 0.789002145332758, 0.19384399687993792, 0.7244343429536458], [0.6970448295177649, 0.4229071719304215, 0.453594387677712, 0.8329331182940052], [0.9520435979420216, 0.6156160899448542, 0.7107250847205517, 0.6439477697541416]]], [[[0.6659031755196494, 0.5585605231391902, 0.5040719799053046, 0.8958458490834424], [0.8708063802321656, 0.3439622310778292, 0.5130149683962194, 0.10522215221162368], [0.7022203085690099, 0.9934799168250917, 0.2506636246729308, 0.16427791627165245]], [[0.7391922114000964, 0.043097456493444875, 0.17129582922316333, 0.1553678727609511], [0.18491233626309656, 0.4057679465572568, 0.03207740311485385, 0.9001331921701106], [0.4818434758756335, 0.7926873428831903, 0.0028809437087680134, 0.3959817876223608]]], [[[0.2554274880589791, 0.712440006606618, 0.6408219564704969, 0.5877217795198632], [0.4187051444596033, 0.9936209960663148, 0.8309217275627961, 0.7504933014287744], [0.3480921182300606, 0.15455535116754449, 0.6817701257087642, 0.6187081452635974]], [[0.09237135102324212, 0.8929575338370407, 0.5808238735234725, 0.46027662219474885], [0.9814313079171213, 0.6373459715599616, 0.5915812869094025, 0.9692182106977938], [0.35126792127525996, 0.1931203749122251, 0.39080644401612863, 0.12312865309909504]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_FunctionOnBoundary_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 6.565931e-02
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_FunctionOnBoundary_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.2945673869725772, 0.5137746159099191, 0.27584369305444145] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_FunctionOnBoundary_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.338342919548857, 0.8634157519287008], [0.15819195375178852, 0.7158265352723212], [0.11520666423861647, 0.3468523772517941], [0.7708104703827059, 0.9674461665222623]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_FunctionOnBoundary_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.5003009108975279, 0.10443572119239564, 0.015034204807924167, 0.2459254241903711], [0.8527920986454199, 0.6535467158709842, 0.9458451512349537, 0.6568106766792589]], [[0.8957116197888478, 0.003118373399681018, 0.14850290779965825, 0.46254240610133956], [0.9383459723823225, 0.2636257332331948, 0.24170263479269138, 0.5164953041778834]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_FunctionOnBoundary_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.7324624087849464, 0.027263694709999675, 0.594317151844773, 0.4193126648417973], [0.06681867047030021, 0.891471125571623, 0.30825824912062927, 0.7527258194756871], [0.09913990850114562, 0.9626163751381577, 0.08478819802747783, 0.15297182776989715]], [[0.744164030067323, 0.528632217764034, 0.3264545486093663, 0.6698188504745249], [0.7593062875103713, 0.7392857862648838, 0.9936141867997563, 0.17748288059795625], [0.6307145974309565, 0.7117932965662842, 0.8463383032406973, 0.8678730337151794]]], [[[0.8630304551884701, 0.642921837195083, 0.09292137498151853, 0.565449383147089], [0.07643516079250168, 0.4161784806507062, 0.9066806723324461, 0.8084988183054603], [0.6413224184753955, 0.18734131084482852, 0.3606631195387867, 0.49150252335102995]], [[0.37717599906361876, 0.8953044246685854, 0.7318229281399358, 0.9329770828458755], [0.4792510680036539, 0.38509653844411307, 0.5479744052286912, 0.06894496382359827], [0.2581672957474246, 0.5600047176086903, 0.8138671648724446, 0.8220335488717734]]], [[[0.9347093528821342, 0.11237749710571576, 0.5296114857573128, 0.8050866792471423], [0.33259727332532096, 0.5315704177301697, 0.31920155340285683, 0.3914547977862284], [0.878104003272701, 0.5994682105740773, 0.44932355283867553, 0.7582858242405067]], [[0.06571082489839197, 0.19957682868800752, 0.06242393757196962, 0.12627014171277118], [0.6668100403089322, 0.6013735242463771, 0.6177983259873522, 0.24086714257220454], [0.6288987822353965, 0.5649825815016367, 0.9110750140839613, 0.6008866436716015]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedFunctionOnBoundary_rank0(self):
      """
      test integral of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = 2.341673e-01
      arg = Data(val,(),w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,float),"wrong type of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedFunctionOnBoundary_rank1(self):
      """
      test integral of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([0.48581036767138597, 0.8176997318253109, 0.7449931690083985] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedFunctionOnBoundary_rank2(self):
      """
      test integral of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[0.16905875433591022, 0.8927668363281058], [0.7568961164373523, 0.14323884542187504], [0.7465301825060964, 0.8316663752956572], [0.021811991114266016, 0.037607928692068415]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedFunctionOnBoundary_rank3(self):
      """
      test integral of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[0.3155736690899592, 0.8900122916079399, 0.7542239833254646, 0.5613148349617835], [0.22295572922016005, 0.008337060750960745, 0.3016440321827637, 0.3825614629651102]], [[0.9626560512015334, 0.5896673784702126, 0.04273629905669185, 0.4741032067837616], [0.6367552160759774, 0.9922416338595965, 0.31619162768439757, 0.9272014432533398]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")
   def test_integrate_onReducedFunctionOnBoundary_fromConstant_ReducedFunctionOnBoundary_rank4(self):
      """
      test integral of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      DIM = self.domain.getDim()
      w_int = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      m = [ inf(self.domain.getX()[d]) for d in range(DIM) ]
      M = [ sup(self.domain.getX()[d]) for d in range(DIM) ]
      if DIM == 2 :
          V = 2 * ( (M[0]-m[0]) + (M[1]-m[1]) )
      else:
          V = 2 *  ( ((M[0]-m[0]) * (M[1]-m[1]) ) + ((M[0]-m[0]) * (M[2]-m[2]) ) + ((M[2]-m[2]) * (M[1]-m[1]) ) )
      val = numpy.array([[[[0.6267459102572278, 0.5362627848876216, 0.2754801847778746, 0.5692720875367634], [0.6857781455312725, 0.10043525699174694, 0.509381001895652, 0.16531992058339318], [0.4812507128514596, 0.39543688539056876, 0.6307843451501978, 0.3440430388056379]], [[0.22425851186164125, 0.10154973695675518, 0.10791091820151943, 0.5722034134850347], [0.10577946028202467, 0.18143494047290631, 0.4128652803523517, 0.27985783799824704], [0.8374825620081713, 0.8307496709593355, 0.5428544158028531, 0.20142592011195481]]], [[[0.36442035652581095, 0.5026111748268273, 0.8116320467432648, 0.6064444955025564], [0.907760726925826, 0.18743127007128535, 0.6137935426553189, 0.7551347389377417], [0.24604177714984354, 0.3056362889199954, 0.7218875452035804, 0.437928983736083]], [[0.5842918872230398, 0.8165926136052304, 0.2454431302609027, 0.8652815946146846], [0.7484208900149991, 0.02456487153523479, 0.19029555443901214, 0.585218434087785], [0.062075052904296935, 0.2689114296062425, 0.3548690764455854, 0.6559697109057456]]], [[[0.8606025882713398, 0.23353230468147823, 0.3892040509409326, 0.5331494010611918], [0.792346428914537, 0.15015459170206436, 0.01972927464430474, 0.19381225295923243], [0.16021425365471453, 0.5898723088641656, 0.9782822845156297, 0.031019522234899677]], [[0.3976100270937254, 0.4564457102723404, 0.2941671644294068, 0.34084336459208364], [0.8370658319833768, 0.5080538034480958, 0.7256738684163778, 0.45173735750620314], [0.3109078637200383, 0.7828143033077204, 0.6809078470807235, 0.41422407599615607]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      ref =  V* val
      res=integrate(arg,where=w_int)
      self.assertTrue(isinstance(res,numpy.ndarray),"wrong type of result.")
      self.assertEqual(res.shape,(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-ref)<=self.RES_TOL*Lsup(ref),"wrong result")

   def test_interpolate_onFunction_fromConstant_onSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = 6.362630e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([0.7957880244283448, 0.15533461547618588, 0.31925558975590607] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[0.5337097756537618, 0.669144212053789], [0.8778275460207027, 0.9408348033958438], [0.9919800317313592, 0.08512409243777908], [0.6568953491049027, 0.5453496946801801]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[0.5966951769938136, 0.9023800941690662, 0.7543928363648372, 0.02125729862699577], [0.8617006703446602, 0.19596308006086305, 0.6290260410166618, 0.9811949926357305]], [[0.23559148624608328, 0.19976533379462902, 0.03540032306051344, 0.9435607556490764], [0.31608892743117745, 0.3473104124357085, 0.9934462072029586, 0.27540536718946873]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[[0.1691131235693435, 0.6055594050441906, 0.7400865559364034, 0.6806037660239634], [0.5027676741487966, 0.28670377793963076, 0.7507170388527049, 0.22392593560562624], [0.32411717270339746, 0.45246860503476105, 0.36851989235523286, 0.5519653438071802]], [[0.6786519875629724, 0.8114097669629837, 0.123316362301787, 0.6499567602584536], [0.9612184906760717, 0.9345484321213999, 0.7470273148400671, 0.724248435412491], [0.8742842737881692, 0.350520533873211, 0.5624272559965394, 0.3475523964102293]]], [[[0.8352910738090578, 0.23073694505833098, 0.4881185812984198, 0.3234947704964498], [0.3777094349118537, 0.5138306419299246, 0.723324110275431, 0.8000651679901943], [0.7032663932010323, 0.021322898941749546, 0.49434113052521145, 0.4735751745643312]], [[0.4198616798316801, 0.8717761776532108, 0.026630328445474083, 0.37936155400847094], [0.6593584959905494, 0.3518822236130449, 0.5216959256650641, 0.5554302397678491], [0.5171075306880052, 0.2744288880526651, 0.34754996006843186, 0.8834443924378164]]], [[[0.13260986756648463, 0.15248351031704632, 0.2057701331582673, 0.6118614857156207], [0.31567423197178424, 0.9478817292191342, 0.19002330653826938, 0.7108763034183044], [0.5958255488011834, 0.6533230731696942, 0.2890221682034334, 0.5023634410996681]], [[0.005845250046578432, 0.96349348400995, 0.9902938895770659, 0.6292937335964623], [0.3979908531302859, 0.8125169905012566, 0.9279306639788383, 0.9711617514765443], [0.8606759832229987, 0.5740243646557076, 0.5439142860935084, 0.12276296123271802]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = 9.300056e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([0.5720473919254031, 0.7735791099614414, 0.940160014446488] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[0.19386089508579973, 0.35216292702931806], [0.7547503277968484, 0.16640497214044103], [0.9090461540570057, 0.990123429441133], [0.8626431222987505, 0.1639923938973007]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[0.925049279799758, 0.9582667321701897, 0.8032728770748506, 0.0692901783185319], [0.5604273748159787, 0.9632692968062707, 0.43996851475907894, 0.20360878035494612]], [[0.7528262356163642, 0.563113710573778, 0.33819880179467743, 0.916829951049813], [0.8561133876813428, 0.12166096328071829, 0.7810547331591118, 0.11112962027788542]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[[0.534892202899736, 0.14438882912964368, 0.47139030653113767, 0.3878266069644377], [0.8857841695411337, 0.017622845046403657, 0.2348163007553563, 0.3221662422686701], [0.7760234364273709, 0.9725487772727028, 0.22566643102468198, 0.34815164395002396]], [[0.6431128039350013, 0.7101100959861724, 0.8940848637837294, 0.8830418934856877], [0.7762675382409253, 0.7595017385120263, 0.46661058103907727, 0.06425677098026605], [0.6658678564741213, 0.39616390430311144, 0.3148603688302921, 0.18338203783492424]]], [[[0.7283898902348546, 0.7287684361712117, 0.2606646905254971, 0.33114189608936484], [0.8508413094408469, 0.0883348760063053, 0.8817492997055513, 0.7115356634186207], [0.93906623889196, 0.523494600849017, 0.6503223649898034, 0.031463748239435985]], [[0.8474561630688064, 0.48441996359311457, 0.8886662913673627, 0.6007483126822258], [0.8325725892213632, 0.2739157486300309, 0.4117906927916668, 0.5104453001476653], [0.5670056768266356, 0.6013381493147748, 0.023236727481303676, 0.36999598568553327]]], [[[0.5072680465215346, 0.8265191807096115, 0.5861067519323756, 0.22141663546392287], [0.9574288611808698, 0.7794012559449628, 0.5549581839981548, 0.8025243456671689], [0.8987023346117079, 0.631438492390796, 0.6288652592325653, 0.09840467517999973]], [[0.748108251586945, 0.029448736325948133, 0.45885848578614175, 0.6973807249739707], [0.8267709666527009, 0.23464722557533646, 0.779310527109466, 0.8163930558754177], [0.9686035108121182, 0.27361593450101573, 0.5916200035259509, 0.04539880264503737]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = 2.247503e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([0.6445867016253057, 0.313880983304601, 0.2737773769510353] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[0.33707194263488527, 0.9630956282230024], [0.6510363673473158, 0.6970870528898714], [0.6419317185930951, 0.12746828138635213], [0.3351947035845553, 0.5978452458428561]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[0.12213004243179515, 0.2784237245234691, 0.9546494291259062, 0.7371354605024819], [0.34779475823532235, 0.022828832154297762, 0.5022890253853662, 0.4279518504816101]], [[0.8645619763948243, 0.5488021749619477, 0.9622143743956186, 0.8198574046539782], [0.06963968225376582, 0.83577757487137, 0.24277444984478913, 0.09379841127974142]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[[0.2614323799260653, 0.8351042413345324, 0.7692308029368904, 0.2557005537146777], [0.7870363454045636, 0.24385941634421826, 0.31269851568289897, 0.026850874086864707], [0.9356315357918102, 0.2659278978585883, 0.8478134748134306, 0.8002775491090136]], [[0.08991515755148205, 0.000720124857735871, 0.8263194296435186, 0.14083100900967482], [0.7340752889936754, 0.03757062737708783, 0.36986277705122594, 0.9657274690515972], [0.43372215877805453, 0.9944754498401112, 0.638438098217256, 0.17988176344123008]]], [[[0.8719945854479118, 0.5430760793908778, 0.8971206123644734, 0.9347942282193488], [0.6366109234737618, 0.3992474033538371, 0.9119390190972835, 0.8923966249995334], [0.7797945316627413, 0.3819562112596737, 0.2651896652904343, 0.4419252778843903]], [[0.5368798322893592, 0.4681925958952069, 0.0026925769788938014, 0.48910459180515], [0.35189296228277345, 0.8995624667827478, 0.2686609671752668, 0.6502525818801146], [0.693052452676061, 0.6647133652576568, 0.08563980600430954, 0.6118914626679669]]], [[[0.05365567041344044, 0.6320237261138029, 0.5315273195236281, 0.047604032750147796], [0.7954758473285501, 0.984611766955196, 0.07015841071899076, 0.07981171219177563], [0.9952972242217776, 0.0065753173984242075, 0.5850967344116169, 0.7032906564253522]], [[0.35863003513491987, 0.6529020111537239, 0.6610405768959473, 0.8886386003176545], [0.7253125602245284, 0.5847163004507515, 0.7379894904581126, 0.345781356713295], [0.9589929182114783, 0.39050438530567333, 0.7400184948062529, 0.573603761781558]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = 8.951148e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([0.6066732546949051, 0.8016398923432774, 0.4742691549563627] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[0.08556466722946376, 0.4950085940164848], [0.7730714517331375, 0.9837986193454442], [0.22625803764676578, 0.6552668077134152], [0.6810217208693635, 0.8713404367856101]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[0.7058179274094152, 0.22337081992051633, 0.8122760445478099, 0.06728031553337865], [0.9756190604873851, 0.6496267984929114, 0.5902853051221699, 0.3767000223828797]], [[0.266337667800284, 0.4207323904799788, 0.5750824863784477, 0.6987185460777049], [0.043561259147539166, 0.5622387926011073, 0.9320712857652301, 0.9526708785827293]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onReducedContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[[0.9053796836775925, 0.8797433141622261, 0.35691574679425453, 0.9118326265463733], [0.9077954721023275, 0.4064689850148223, 0.5303614090385271, 0.40604231144823466], [0.11607424724026061, 0.5599600398772994, 0.814075551661732, 0.54254591030764]], [[0.9062004798239428, 0.7359957714009656, 0.01798015865607705, 0.24826118958393129], [0.0543279497977297, 0.5433018354934592, 0.6342410596898008, 0.8102700194136568], [0.40220290876693743, 0.7521083582294243, 0.8312614071899106, 0.26542605255033247]]], [[[0.7622540278299255, 0.48609979776990997, 0.5803413362149745, 0.5607529899926853], [0.20494496825065767, 0.3502799847257706, 0.7445213672889649, 0.9689376198705607], [0.5234912090436175, 0.7184819779425649, 0.4615415726839104, 0.7355891398407086]], [[0.2671919789484164, 0.9807156056044939, 0.5884608670312474, 0.35556723211294106], [0.3698459328833036, 0.021662160991340995, 0.5754105446763756, 0.6998817470727063], [0.08539102866608617, 0.737917818260024, 0.13807049592934928, 0.873592672888957]]], [[[0.02309680162950456, 0.6501015293782897, 0.543023930558804, 0.16724059191177], [0.5922782207754151, 0.8174937364601335, 0.6860353552710656, 0.3538636852781476], [0.15891103131880369, 0.6861738109435661, 0.47078753964243025, 0.6884582686306938]], [[0.893188311715661, 0.8865432429837392, 0.9016858446102259, 0.6387469290276518], [0.38674084211945625, 0.6178487648759542, 0.5314659765084864, 0.8207842812941796], [0.27063765117779504, 0.5550043283273147, 0.16119769636828118, 0.5476268785609076]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Function(self.domain)
      val = 9.735540e-02
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([0.2743715749292459, 0.21469786552223857, 0.8593033737580423] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([[0.6846912417947641, 0.06894788069207214], [0.978900026010339, 0.09873703418021096], [0.4919546781747247, 0.39911956020002926], [0.04553381200128093, 0.936072200755137]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([[[0.6123958941856611, 0.48679257122335007, 0.8904002629859167, 0.9572963840936238], [0.6243391148042786, 0.7709419990983722, 0.954913195292853, 0.7623141780447016]], [[0.2742632577925228, 0.6058872464545579, 0.10700255297665273, 0.7789171399591038], [0.9627200051077933, 0.6221603524102973, 0.9542257157007191, 0.22639416924078581]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunction_fromConstant_onFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on Function
      """
      w_ref = Function(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([[[[0.2970124845724982, 0.9664209731219923, 0.7282801354365919, 0.7230929641464919], [0.45006095507091426, 0.012786087660424661, 0.5789425955397912, 0.3067754784298242], [0.8621451350032959, 0.3777843119038977, 0.837991057837885, 0.3266075851449377]], [[0.10559985815885153, 0.2804607320052436, 0.8951175530640482, 0.315420817580812], [0.09613077434959427, 0.39319781168821666, 0.7494837497122231, 0.5621137256709696], [0.2869247718285539, 0.9704518484296983, 0.8624881513697881, 0.6532896599378017]]], [[[0.8067849518701942, 0.23366554103475312, 0.7786214838162852, 0.3688271473922984], [0.261624144653578, 0.655555167205075, 0.360295197528421, 0.5785499957774662], [0.3337932634810461, 0.5122950843101022, 0.4658186434277789, 0.3775442280831859]], [[0.07906213284684738, 0.7274934339748216, 0.30650282320570854, 0.27943797535753134], [0.9804466861381568, 0.13421200844964265, 0.8676702212229163, 0.6928499188226049], [0.35153120975853436, 0.21130282331022066, 0.8118771496566942, 0.8880110735985497]]], [[[0.03512494307952818, 0.7608765341895213, 0.3101236286036795, 0.5745214687846504], [0.6625962485952909, 0.8317739766430151, 0.8232216834384922, 0.45933835690782554], [0.1539455147736789, 0.6455194134516086, 0.6413842437290627, 0.30636459274182914]], [[0.8916782444314838, 0.9829897770361029, 0.8540622107815008, 0.6064961419219922], [0.29805293341076233, 0.6659362982990858, 0.7471676189872066, 0.06246468676823902], [0.8039682465326102, 0.923490451650216, 0.8922927240835877, 0.3130749046132094]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = 3.728258e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([0.012971727303935032, 0.3957173440042402, 0.5409633187570417] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[0.19515767965056197, 0.6973560941989282], [0.1026344066296202, 0.19376496449411318], [0.1489826290982933, 0.32190319858015215], [0.7863186531195617, 0.21050167126598007]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[0.6131998020720266, 0.253807496184978, 0.1691697388015262, 0.2765875384637184], [0.8224880214516279, 0.2640941443307725, 0.8840157747171888, 0.16343699468508355]], [[0.7224141035444551, 0.7465831267725502, 0.38210603950193256, 0.02578542820312768], [0.4562390904340804, 0.12744676539494748, 0.7651594504829711, 0.41466620658569875]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[[0.7575419952110635, 0.9020775334918304, 0.7572501252215946, 0.4154368797961233], [0.15043484189529976, 0.004857332752311039, 0.2349518517295438, 0.7304171671907722], [0.1050715465877049, 0.04447526835184923, 0.7054732646870934, 0.6745068154308503]], [[0.247726729905123, 0.11082510313222316, 0.9124795867158895, 0.17286167920667295], [0.5463160264747804, 0.8122346367461396, 0.7217802017863385, 0.6922759037372286], [0.21305609701662598, 0.30494843853884457, 0.020576671723682938, 0.527881550311816]]], [[[0.5485089326345317, 0.9796056792579599, 0.5336920404274481, 0.7796198687611161], [0.40346400278024164, 0.6161139739077297, 0.6083992706043033, 0.21422040582287794], [0.9815921964958273, 0.5538942350259384, 0.26499117558610563, 0.7790581530847276]], [[0.2340825486805166, 0.898136081400828, 0.5359664808913567, 0.7768956047745614], [0.4972377601123942, 0.8988399867675314, 0.03607509358508598, 0.9473643960842325], [0.008867366706068913, 0.15432797413445842, 0.7519109235674708, 0.16889193138926994]]], [[[0.33463706661310477, 0.6250277603945628, 0.2848531077188371, 0.385805614067793], [0.26761389179709183, 0.20819682811775042, 0.40458605506027934, 0.14173898654864892], [0.8323891029374194, 0.41330347063797723, 0.9945924451945423, 0.08696191317702207]], [[0.7883901317275269, 0.21989816004612694, 0.9907669081920188, 0.8355286046419015], [0.3273140987294253, 0.5570362711124685, 0.6058717832691248, 0.8109337748012123], [0.8948355192793349, 0.26719368475866867, 0.793392371729907, 0.12265148012043736]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = 3.150930e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([0.7490294729809388, 0.20398195461443003, 0.6815304189962305] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[0.18928827137747295, 0.6565724473867753], [0.05795388490535902, 0.5940015021513085], [0.21608441479396467, 0.0577813802223468], [0.7927201986393057, 0.2769882045611559]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[0.2306177773104542, 0.04668454787233478, 0.32151273485515397, 0.6613767848691384], [0.2462933276403272, 0.6220697333417127, 0.5496069482737364, 0.15726399552366555]], [[0.9310155631859295, 0.9123509065656492, 0.44211833373708387, 0.7902703442335067], [0.8274530542083902, 0.5834972153270758, 0.35600865651056834, 0.6786525605923202]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[[0.1078387984939646, 0.6779191083947568, 0.22082250181160779, 0.6637086193441153], [0.6124058648635899, 0.2436252078681913, 0.1794680271178265, 0.9659476891164605], [0.6645947631357397, 0.10388389324490832, 0.9915002528444541, 0.3298558232280008]], [[0.7850034474680475, 0.37249080241888555, 0.7547940194315218, 0.25593589128673333], [0.5967018891553976, 0.7562324747433757, 0.4926152716329022, 0.12403837698443199], [0.09709556618249504, 0.8598669518766141, 0.9917116480055214, 0.5226983011447389]]], [[[0.5366184544364865, 0.07073016370892626, 0.183662818999751, 0.7053252141625185], [0.8604580418347838, 0.5621418179116249, 0.18355158200074295, 0.9347752517042563], [0.4020762395441769, 0.6776406738997084, 0.18700240935855272, 0.6578768497841756]], [[0.8432841079564232, 0.39384184490035934, 0.46419110526588603, 0.9250719815363536], [0.9819988348224011, 0.4891210910740228, 0.4787784641401054, 0.440334617183876], [0.6008111655338242, 0.863159029860924, 0.8193345864340122, 0.0185308922374684]]], [[[0.617824673895684, 0.08902970422841205, 0.9855167614108061, 0.5619965493225246], [0.1701569789929529, 0.7514267586278164, 0.09458771333902516, 0.7071314883631369], [0.14975674152904972, 0.1824365014765862, 0.5206117642595841, 0.391518541184131]], [[0.7155790453366037, 0.19045442285834502, 0.9191791662001249, 0.9205677449372358], [0.7446577830708815, 0.5719615797850744, 0.3174629245033359, 0.901426632886397], [0.7559743653238208, 0.9352772535071997, 0.6145229483534825, 0.1286554625074503]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = 3.260421e-03
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([0.8968728525758245, 0.9812864362074349, 0.10156242799983295] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[0.4118649570395786, 0.4111829501471638], [0.8415008717865939, 0.09924301711503469], [0.24257843038149807, 0.8600844611069439], [0.9630118611169631, 0.17844579151249906]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[0.41652392448901765, 0.1674334113982131, 0.49750788947327484, 0.2897070892908381], [0.3638127879744031, 0.35649423868667085, 0.8479688643778721, 0.3137442583557205]], [[0.5392532317376467, 0.2189721215014382, 0.4144313798930708, 0.15111584874912964], [0.7204946884353611, 0.9754381515852647, 0.8218854665066333, 0.21839242530963188]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[[0.6112767869006167, 0.40119849661820317, 0.48263512412666243, 0.512738683372073], [0.67551549566529, 0.47916295191525216, 0.6457894956307684, 0.5190529996145664], [0.42030364167772405, 0.37963410849906265, 0.6197852711086739, 0.08893320512728975]], [[0.29311909322470664, 0.9078039524622689, 0.5667621427360752, 0.8815268118688634], [0.1976920816718083, 0.22174905824587832, 0.6580450956634241, 0.29479978564545883], [0.7207450235602346, 0.09792934708788881, 0.5032680829616303, 0.05461134674580792]]], [[[0.47450656962533, 0.5408692534965368, 0.16846796750872994, 0.3118660047568367], [0.09352135536459416, 0.22568170567436097, 0.8169337152668241, 0.7335264598933963], [0.32016349396653665, 0.8241224536455416, 0.6121966269084769, 0.46572591656401485]], [[0.9676557388760243, 0.17988071339325162, 0.6179773240396723, 0.8853602707502274], [0.23128659135991025, 0.7067266570445252, 0.5186863016947421, 0.17210683444542862], [0.8714022446592182, 0.27014720911567613, 0.35057438838383803, 0.7229915787046332]]], [[[0.9503623687583985, 0.4642172911766458, 0.567314413657875, 0.7086928932940554], [0.49312244840567676, 0.32885543023329333, 0.46398948700228637, 0.6220319579542395], [0.18184864131884415, 0.870736860550771, 0.32192956322217137, 0.7038770694346873]], [[0.929520906640193, 0.9641199393138358, 0.43063472362502386, 0.08824458887043662], [0.8963625912136505, 0.45808207008556145, 0.45011466083504725, 0.40398102131892455], [0.7850918943306927, 0.90342357596291, 0.9404421740394262, 0.8124667905888024]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = 8.575434e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([0.5725918096409758, 0.6003841830725746, 0.1441542920384522] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[0.6293856675726806, 0.6893340206451204], [0.808534788550273, 0.8619664687725175], [0.9223392946188489, 0.14192622876600436], [0.23801236915392276, 0.6855732439869456]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[0.34249822058084856, 0.9161409852900473, 0.35308880082288896, 0.8481508363708606], [0.04395025574198774, 0.7444232225770522, 0.48328911269573127, 0.4312664086925645]], [[0.11879143352969523, 0.24235522740831483, 0.4684045454422867, 0.03647719487907586], [0.049185045125999416, 0.7148935616360678, 0.9038590951599494, 0.7565073711836524]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[[0.006017857670573057, 0.23364131519784181, 0.4985494960868405, 0.27109100097392147], [0.46777297787696615, 0.6253125640259952, 0.40452662800606065, 0.37763688636458537], [0.42176263329078234, 0.33627005175270563, 0.7284328028931366, 0.7554306088625987]], [[0.445605085403727, 0.24168586191612962, 0.8761435178594109, 0.6336418596811683], [0.12067539387160242, 0.2760591431484126, 0.8302040359677788, 0.5249849665915117], [0.25910328032811913, 0.46022280614122835, 0.09338024314795279, 0.3025158275666153]]], [[[0.9011600112580125, 0.9067598044865398, 0.36675480908395464, 0.24672806654954949], [0.3354970078650197, 0.3328614738307458, 0.9439130347784399, 0.6661090918695912], [0.033963254642413454, 0.3784876992752988, 0.8642207159457712, 0.43850579779608057]], [[0.41036053467979716, 0.20771066763908475, 0.436903095299338, 0.5186597221271169], [0.1528880722368523, 0.7947807451302626, 0.26183907725585254, 0.7347606333872441], [0.2276041986246412, 0.1997750912337225, 0.4079301945121403, 0.3979885600222165]]], [[[0.7111046900053858, 0.26301022503920446, 0.005970573072419216, 0.4946615862812376], [0.15967084945823962, 0.6539354868230908, 0.9475004902589886, 0.6220902877158248], [0.7865229772442921, 0.8417263055532432, 0.1659834833377698, 0.36162369312272435]], [[0.8803451748350127, 0.04879713643877015, 0.9073005937541219, 0.7580478879811012], [0.601170013261482, 0.9801415344221798, 0.4037308171429357, 0.7833100529376933], [0.03621439688551875, 0.07125911770743965, 0.5897957915039718, 0.39439908929824796]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      val = 4.027490e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([0.3250182911594902, 0.37482274213325484, 0.9026028760284558] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([[0.8626369066899092, 0.4585408354680027], [0.9038270065192777, 0.5132186814212445], [0.2754318743152945, 0.7606957160579115], [0.26929872320779824, 0.12523388730012663]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([[[0.1710848061970881, 0.4210605266008811, 0.1799412900566273, 0.5680036275667364], [0.08272971547755459, 0.6561558440334991, 0.43166459070419705, 0.73553987620405]], [[0.08553882488658737, 0.7668629602548288, 0.8239733431217472, 0.6790092094499529], [0.30524026079530386, 0.9218782475268773, 0.39672505422645843, 0.2779629842723299]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = Function(self.domain)
      val = numpy.array([[[[0.3290657098736812, 0.4322039583426357, 0.41679124068929796, 0.8126181140955258], [0.0364108453724622, 0.13153050469390792, 0.42667362159493416, 0.7269177132160766], [0.13139887775061831, 0.14434199337800335, 0.6122122378647069, 0.6415719838392016]], [[0.4950747908561032, 0.05818933183041919, 0.5637593942515523, 0.16802703084948512], [0.5938645385125116, 0.8887302141326991, 0.5046494841030463, 0.8680400546868173], [0.9219797057865186, 0.5312281531664578, 0.7513129868079005, 0.2855893695807905]]], [[[0.49672165560847126, 0.9980575642511861, 0.4858508883100776, 0.7156314913205017], [0.8531148185359293, 0.9307807538610622, 0.4065721774005391, 0.1389553549495719], [0.2570744157865883, 0.5193374245774657, 0.3548465058431266, 0.5325510266778257]], [[0.056082832572043606, 0.8804164784307431, 0.7682033179219527, 0.96189826806205], [0.5767667365187719, 0.2047025018997496, 0.10002088039489676, 0.9525702281176113], [0.39612044468926366, 0.41101913208017515, 0.3294265556585515, 0.6608604529763971]]], [[[0.26594993597807526, 0.9942637994644868, 0.009990794210599674, 0.3452102472130977], [0.0746892268534799, 0.937241251111298, 0.8203467445106463, 0.6066768531501937], [0.43265136987239683, 0.6425094052895142, 0.6979610366846875, 0.5082328192200561]], [[0.6625510907969888, 0.4668463667680315, 0.2517549671029965, 0.2883304737321145], [0.1755580391237066, 0.33008357665124555, 0.8511290836017738, 0.6575328633944767], [0.11094017588286831, 0.6654188284769922, 0.10239978622283252, 0.8684316764300468]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      val = 8.414660e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      val = numpy.array([0.05849068302647553, 0.19447582046764433, 0.8561679113512668] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      val = numpy.array([[0.9506619006477389, 0.45158015390940354], [0.13794385505939888, 0.2650022089772285], [0.137496234037903, 0.6061453738559234], [0.13136958789709163, 0.8903658787553632]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      val = numpy.array([[[0.5049445622092998, 0.9503234641849262, 0.8566292410717335, 0.7211612275027681], [0.9799186194859697, 0.9742963916662968, 0.5180422269851891, 0.1453140389575599]], [[0.6937181277704628, 0.5812458331575757, 0.8393868082169451, 0.5815957353405391], [0.843105079526146, 0.27458482867796785, 0.8573812421346612, 0.015365887109309528]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunction_fromConstant_onReducedFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunction
      """
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedFunction(self.domain)
      val = numpy.array([[[[0.45798366622514064, 0.5193102937785187, 0.20026259482112807, 0.5752762940618534], [0.3256477095285484, 0.837553585825753, 0.6405909508206649, 0.0854938394042647], [0.6344348236055513, 0.7355398523726475, 0.11662713982751316, 0.19784692643398172]], [[0.11040561381615766, 0.7721002746257891, 0.9637314196990963, 0.3175732518686092], [0.2654213837064784, 0.6295950193017563, 0.5510307578728122, 0.34767344894886876], [0.632331011074255, 0.762096748207405, 0.8968910997446898, 0.09367919363975996]]], [[[0.9116317371108273, 0.607571071079898, 0.9601920527563665, 0.40327404883922624], [0.26165673238553044, 0.6831009504228737, 0.8995423716551886, 0.6556910293351392], [0.08066187035403649, 0.8195870021848646, 0.8357368655590256, 0.005099761546050963]], [[0.09386702122280133, 0.060675152908738705, 0.23283975759278575, 0.38877792949780265], [0.7976392026081873, 0.275800847647067, 0.9217423207211316, 0.15034914823275425], [0.7517744507415166, 0.7329445993447944, 0.28057447335288366, 0.02840906617234229]]], [[[0.2528689454628795, 0.6952180766780145, 0.8969218154535067, 0.7123216491520122], [0.11177689822920944, 0.7794409198464872, 0.5449226329759524, 0.42849681173791354], [0.6174140243613375, 0.29872651066827094, 0.12421663225288804, 0.21211874568709999]], [[0.513926505495494, 0.9768158291160023, 0.6114756969812538, 0.1870935495775632], [0.006285103664050662, 0.14487813193385723, 0.9702978777341069, 0.9666390008551409], [0.16475839853805674, 0.47693762634427506, 0.6539215859843611, 0.42602005660059983]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = 9.192167e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([0.41038789718937674, 0.8753158910702785, 0.40646509008814946] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[0.9777665519705623, 0.30515152513007326], [0.6178842521178955, 0.8690952716246513], [0.2260807519749941, 0.44002942371896236], [0.0859769086897264, 0.7760803649079892]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[0.51850652220834, 0.05478610401807382, 0.9500183440368776, 0.008561716687065624], [0.5856735366173363, 0.8411375892234033, 0.4898273919923175, 0.6173720111374574]], [[0.027715270037911188, 0.3023149932544965, 0.14538979731538693, 0.5136462244271097], [0.3977251731942798, 0.9079992135171645, 0.04346413556660944, 0.2888627174370282]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[[0.20622408379238066, 0.005100009559660501, 0.23337931752995833, 0.10601497183723918], [0.8165023555426181, 0.396580852315114, 0.47029308560599625, 0.3874301860339717], [0.2011358318819475, 0.5461596158643076, 0.2239654551721597, 0.8856310800632324]], [[0.4865526283067304, 0.3545397483398661, 0.38998059664381524, 0.565302365497114], [0.2036231799281505, 0.4818603523191022, 0.3928413820152391, 0.4712734087844712], [0.3923227826719422, 0.4920022951224834, 0.2743085894289139, 0.3692164595954831]]], [[[0.11352919547716644, 0.09758472955064057, 0.3769346202405284, 0.8611052931614419], [0.005233705552234791, 0.4884777542806439, 0.8122024660886195, 0.814485975702801], [0.6353940925242333, 0.1512161696283696, 0.8613782896867531, 0.5499228421963279]], [[0.9378121425611091, 0.14334109698436082, 0.646216595524758, 0.7889063463455013], [0.8328778640005114, 0.7875327973910158, 0.8341424415381581, 0.5476176259551765], [0.5919432489217056, 0.3851719939103463, 0.2937238668485481, 0.001323464137289232]]], [[[0.907197659013393, 0.9771085376225925, 0.7681934429785708, 0.1437626548077089], [0.10609723386579106, 0.7970325410650613, 0.13580107033208622, 0.3095278591797175], [0.5426885800883268, 0.058891750718371494, 0.5726727281494745, 0.10413526077153223]], [[0.8808911345392071, 0.43402439305259, 0.5994893869237345, 0.8682314872608294], [0.9534415618238753, 0.8617431942968606, 0.6009208065252089, 0.7934262509150228], [0.4253492297908934, 0.6435973759874094, 0.012744602269634253, 0.12137625323158896]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = 4.506794e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([0.14246416290793984, 0.9973780177886851, 0.08384063678307241] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[0.9999399549724203, 0.15213235003893677], [0.7107206921345622, 0.4116454322858999], [0.9178626025820907, 0.17634563548224114], [0.7389241754895396, 0.17012409710168386]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[0.7966643608859245, 0.975748924838338, 0.7796226785530369, 0.8259173738472025], [0.7829656332967383, 0.6872008388513661, 0.12445630733343327, 0.6434417403109474]], [[0.1757612623897169, 0.8783903523758947, 0.7283411636244587, 0.5310367202876731], [0.5460503107333516, 0.04306599554577206, 0.865427397809365, 0.8318637128939894]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[[0.14582397292102844, 0.6544969671881469, 0.38797238806491974, 0.39520238186905965], [0.036291902683921684, 0.9165814584950539, 0.7954558978953897, 0.16831695835511618], [0.0863580677155169, 0.08941808033906129, 0.61927614235708, 0.3018976003034982]], [[0.7493211909675515, 0.6086166656015553, 0.377440956410247, 0.35024282326748046], [0.5981380424320348, 0.9822820806997827, 0.9288510056981487, 0.5859141321672025], [0.7484786688292904, 0.20141594168627341, 0.8319229243827568, 0.09933228360785518]]], [[[0.26052728871085606, 0.13811987233331924, 0.15981560203754974, 0.1307231063931723], [0.945065291608868, 0.14397226589372414, 0.18137931889974546, 0.3449983036342097], [0.9919646435245192, 0.8917228330694286, 0.3805270332960442, 0.7302765092819249]], [[0.6074474487998708, 0.08460704019903975, 0.833083329714192, 0.10125836914566688], [0.005552871595602471, 0.1465109871455753, 0.9309488847018791, 0.6696103205315634], [0.9993274454875536, 0.40522013004902746, 0.9996446213480157, 0.2706134570525909]]], [[[0.5291260863701185, 0.9376952779289915, 0.8203658473177817, 0.9122289837016366], [0.0029039661476604195, 0.83102237275781, 0.4296819023930326, 0.9857948972993257], [0.4073595529668991, 0.8038699633316648, 0.8316872243584207, 0.5275302089170382]], [[0.7902884489236343, 0.307906619368144, 0.6484099713379581, 0.05690675285413049], [0.28044876022275866, 0.5996827628977335, 0.7902636684719875, 0.9309904897922812], [0.24476368641710256, 0.7018415281316769, 0.5905788192579572, 0.9412065444516884]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = 1.108449e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([0.9497319589257759, 0.9636075516101584, 0.051564768313318354] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[0.5563306374540675, 0.05583183073812581], [0.7245573572829551, 0.7847336321885154], [0.207802490269958, 0.27956915421799033], [0.6108548164634193, 0.41288965529993027]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[0.35892684451205115, 0.2966906157972712, 0.6582960973053003, 0.835667345496716], [0.8046201458736681, 0.6084967085002652, 0.3239358482598148, 0.14452765410991408]], [[0.6504531677609211, 0.8678550210496763, 0.2821918001955528, 0.44618525404522624], [0.2417077771849203, 0.5973709757045786, 0.46672174991455884, 0.6337650537799687]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[[0.2944251353507551, 0.07501081348075567, 0.8309177134544613, 0.5776490765350085], [0.021689489790741034, 0.49889730525325804, 0.5139408231457663, 0.9882435241915973], [0.08137730461123893, 0.35864132100205703, 0.26430599503988317, 0.5658329672272592]], [[0.3163171611555239, 0.7054659851923525, 0.6513635213332974, 0.8495823569979479], [0.35877021541226894, 0.3548827234348809, 0.8524360587914576, 0.782041719369517], [0.9133825563780668, 0.7118339530746202, 0.7815080064303972, 0.6222191137687296]]], [[[0.4248723817383898, 0.8795001935037968, 0.01666287947492029, 0.4084386155392602], [0.13414701303594956, 0.9384171855671386, 0.5957528481119125, 0.5272189225595069], [0.028716420311946322, 0.11191082995378943, 0.9778164086389094, 0.6167966785297105]], [[0.6081155888593136, 0.9650418205445292, 0.8877805345004621, 0.30297960074429253], [0.645074469475765, 0.7313547180092884, 0.19773820979384327, 0.9470583356818679], [0.172359967515351, 0.6916943710926826, 0.33277583453620574, 0.6449351779555812]]], [[[0.6102478623820575, 0.35544894157761286, 0.7518816958884267, 0.40342926789366684], [0.8502768175957146, 0.5591442930799826, 0.39238438198963366, 0.4540480434719212], [0.18945099002280996, 0.7900497565669576, 0.02967265631338767, 0.1051417886148277]], [[0.48721929377589923, 0.4777693001340729, 0.7047844215612442, 0.7918864544058136], [0.4715159800785831, 0.011874498977747727, 0.5805593477162783, 0.9815978294924813], [0.9281386722114529, 0.5609485520373018, 0.4063300030918854, 0.20309197759679953]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = 7.070421e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([0.7607518371224172, 0.8464643338635388, 0.9447448149773134] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[0.9052236375670168, 0.9933326385559299], [0.6872993494967177, 0.5609083293415859], [0.5858670569693469, 0.25637763665160507], [0.5937208221067354, 0.41695511979912836]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[0.5038771000678699, 0.936487893943485, 0.21947511973273015, 0.9308205058304588], [0.76166507153111, 0.9093224836696384, 0.4950264500227347, 0.827000080933229]], [[0.022798683770178574, 0.8664069404036596, 0.5969559321683152, 0.37028573368723894], [0.37908518722638784, 0.09278145963555984, 0.36092463038218936, 0.4072068019394879]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on FunctionOnBoundary
      """
      w_ref = FunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[[0.9954944184525968, 0.2804719312247317, 0.42981219210236377, 0.16898220127865404], [0.5304958374278685, 0.9821892954307407, 0.8795519211742088, 0.5695376679142153], [0.497351879619535, 0.1297996965378112, 0.7286146022068011, 0.233846112327177]], [[0.780008079251037, 0.06840564213822298, 0.12317179942058809, 0.9736047882746599], [0.562603853324446, 0.5311572053208246, 0.10562928629388313, 0.9235918257363729], [0.5718594734897233, 0.3836307796287002, 0.75672637306399, 0.9662234483401362]]], [[[0.9677971491907695, 0.05362933942890469, 0.7369240913481204, 0.26033931289076173], [0.08441528314834867, 0.7686010092575635, 0.8788974306293529, 0.3480234012044001], [0.8827319535640358, 0.34139199598989645, 0.2387225815541889, 0.4271550167444289]], [[0.9433318959482143, 0.6926092314386534, 0.0579857353278046, 0.012760602879228422], [0.8299506048071782, 0.16514401419229907, 0.9622492642442476, 0.9719336749500798], [0.14052933402133672, 0.040167444311985157, 0.8448263470459549, 0.07356494691919657]]], [[[0.9045874679372933, 0.5543948259159743, 0.27114391934801696, 0.6124477204630296], [0.8362921077428459, 0.9246955662318225, 0.34346527024414963, 0.9720097520031376], [0.7885326599660657, 0.3496399645885666, 0.14977077719533283, 0.34371673899701694]], [[0.8963518220111483, 0.30675656220409964, 0.9749762460041155, 0.4950889266330232], [0.8059279482797532, 0.04045052568497254, 0.5951258122513224, 0.5306970651564509], [0.5957654454868341, 0.5844187015725724, 0.17077377682759753, 0.5705581127663562]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = 1.852028e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([0.9524901431348591, 0.968228116597958, 0.4374117029306468] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[0.6283815860104343, 0.6957769522914173], [0.5837443985047255, 0.05763188452653645], [0.3260167444620985, 0.6556166631341867], [0.2292972433996443, 0.8301435680946683]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[0.6201259108475821, 0.5746426093351019, 0.05949913896726189, 0.7450953635359022], [0.43460250008615053, 0.3074373282370092, 0.9385315952680193, 0.7663312029539153]], [[0.43275709490573866, 0.03520868970279034, 0.540679458264847, 0.24781018681665112], [0.17817695089351604, 0.5306920370376325, 0.715462213991392, 0.1975429679639551]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[[0.07958145536780192, 0.586292071403468, 0.24305878981930418, 0.10956835948378785], [0.18790202669214506, 0.9999991093916525, 0.7648710047696148, 0.3995601886990934], [0.8242421198377108, 0.8709616320531619, 0.036941579592055906, 0.8737747050470887]], [[0.5456111046652553, 0.41728019065821387, 0.283359484508531, 0.9954511004461133], [0.703016807474764, 0.9439428840593841, 0.585557925285333, 0.832581927086008], [0.4410255696163289, 0.8864814087471373, 0.9820280215380782, 0.5621575030502604]]], [[[0.20481617776433025, 0.17129844891572055, 0.26939985996301596, 0.21001377356656958], [0.7613064609535483, 0.8740126758736518, 0.9892618069319213, 0.4638242187573097], [0.004712752503956641, 0.37468534675549003, 0.7664122560864787, 0.9508171638156188]], [[0.43720149531198615, 0.5350997812247933, 0.3534193335198461, 0.233148117167021], [0.20078658660747417, 0.869777001672443, 0.5394745541010878, 0.5925790039861761], [0.5966763298689197, 0.23289144391552263, 0.20751974516464333, 0.6159618865654661]]], [[[0.8705034681570242, 0.5805245326514258, 0.610735106012795, 0.17886042204038144], [0.350222261590468, 0.04795881412971248, 0.3139774623374556, 0.438333140713854], [0.8974386154653958, 0.7418304775313741, 0.9064616502896158, 0.18924628026574297]], [[0.461472893466374, 0.7125143823711997, 0.7484095041412295, 0.33678674983473866], [0.6536630858906187, 0.12390985652750697, 0.53332627624869, 0.9957473288269107], [0.5203312787963457, 0.3663705083061257, 0.05755252274556655, 0.4511805743101691]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedSolution_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = 4.360027e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedSolution_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([0.9814583841303897, 0.8576165958910862, 0.07785369478500859] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedSolution_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[0.4113075230582314, 0.5465483395661529], [0.21734053376505624, 0.6973676513882143], [0.44990454080518905, 0.08465419987996214], [0.8038857599681649, 0.6668434230497731]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedSolution_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[0.005521547430829443, 0.6758088987530083, 0.24718267947230133, 0.3426960092590303], [0.0077141969718255154, 0.680022518930657, 0.845368261124892, 0.9866112987747238]], [[0.9296936244367552, 0.8595840995330449, 0.5935178029669631, 0.39357549670582404], [0.6495599674604388, 0.3055336213506945, 0.3821621215123423, 0.14644551577188558]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedSolution_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[[0.679265471025438, 0.5435103016127519, 0.7413003186327275, 0.935860504866077], [0.7828958609795503, 0.5652809263702883, 0.044183618379292966, 0.2513865577643034], [0.9874242525690592, 0.8857390512025201, 0.08439688172768889, 0.610831917980801]], [[0.21416107431777365, 0.7291977503968069, 0.6210228718396265, 0.5215508221490089], [0.2758492536359548, 0.7973590703068396, 0.6879543466908404, 0.15176106531240285], [0.45786106049097175, 0.1586336179213501, 0.6337039744368264, 0.15230888930935305]]], [[[0.5181281106408817, 0.5309557878052082, 0.5040793211461676, 0.2678067100887005], [0.8896641065573563, 0.8399568773078349, 0.3527496428117566, 0.1857666234673817], [0.7287270147890453, 0.9809007761842233, 0.6753454929742083, 0.8602887174170138]], [[0.382703562904562, 0.3385444290235794, 0.08938242227012627, 0.6752310924254504], [0.6288069041764609, 0.45650713766295725, 0.4892316353544307, 0.6403862534439819], [0.934314836322674, 0.42563999164628763, 0.17994155635430764, 0.5939345775678037]]], [[[0.7630638611938967, 0.14076938780328685, 0.029311815064348767, 0.9010601687085374], [0.18924465172377591, 0.01517447974274011, 0.5774557644978928, 0.05195525587356875], [0.6825411046583563, 0.5329477579655328, 0.700373962205211, 0.09629415636243366]], [[0.5955791154023462, 0.22678959171155633, 0.5513514148685765, 0.8560204320046453], [0.02304511301376133, 0.969956630928132, 0.15467884801914122, 0.6812422589047169], [0.6479989921871985, 0.5257271957338543, 0.55313996044493, 0.28178786135956313]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = 2.047214e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([0.02416772253162347, 0.7761099099114784, 0.40972129988347694] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[0.406066297425263, 0.3224942848549607], [0.5506703452768537, 0.36144231520987324], [0.9192445688343615, 0.6703618020979031], [0.7576894134836163, 0.4794818865895989]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[0.5950142355321145, 0.46471555387973196, 0.21003910909735324, 0.6888954330446211], [0.860245466049949, 0.1277561856062588, 0.9191123475151095, 0.713630481895244]], [[0.9524526830805979, 0.6148944845454374, 0.5624943752237618, 0.7795324601114211], [0.6863358797711895, 0.514542697325243, 0.002538939522166772, 0.2835811052149275]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[[0.1855706263439878, 0.8551665320817079, 0.2425488726028966, 0.6648349192350533], [0.527867350742479, 0.550960368040881, 0.18669124610951326, 0.46930288077705784], [0.38377417192152397, 0.8767600932147964, 0.7255646177024861, 0.4072382963136606]], [[0.7039209431076947, 0.6652829637729729, 0.28028768811925564, 0.3768702226572559], [0.30464282801762654, 0.19464044848645956, 0.4930516614499366, 0.5507637280449756], [0.627153438269037, 0.5568723793137564, 0.6147079043668865, 0.4298516309026128]]], [[[0.04797436564854085, 0.973670681618537, 0.19618677129164674, 0.24815278301888444], [0.6519542158984761, 0.9977022169074248, 0.8734428134091466, 0.14655113838432576], [0.30416530921864937, 0.14773289318561966, 0.05086002504233933, 0.33197565393887885]], [[0.8713686176140557, 0.006316687099104268, 0.37064730623416287, 0.2607529108811799], [0.6152686009060456, 0.7540097099842205, 0.8240421523076797, 0.7022497831828466], [0.2587154628437103, 0.49261076678351967, 0.7697739449092331, 0.4753418726160892]]], [[[0.9810468447762527, 0.25112012705787967, 0.13083245527033638, 0.12234701676906157], [0.5081471243611793, 0.8668401414958787, 0.6199507692602556, 0.5819302930458916], [0.6147554572177973, 0.9087270079596469, 0.34264454954047374, 0.577807264554528]], [[0.7205978925977358, 0.10686082894453719, 0.9198491148582825, 0.8095857598809354], [0.7680838830717062, 0.07583810572933103, 0.9432620748938074, 0.658572680079183], [0.8126907736231406, 0.37142498540243196, 0.20011631134474783, 0.43471749614165733]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = 4.196279e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([0.04180672759508275, 0.11505345429909464, 0.45704913150906934] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[0.4889630982370293, 0.4636455033993767], [0.7635907559638243, 0.13608689365316395], [0.5149121340618423, 0.12383229788560945], [0.7649999576978991, 0.4802461876893527]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[0.930100580213905, 0.33926613616565016, 0.7187963853028017, 0.2621606679342313], [0.51963948787346, 0.16154898663612272, 0.10060537215837162, 0.260302427567349]], [[0.1846219243758611, 0.6060990096209946, 0.9332109691746319, 0.7972428773295887], [0.8755406973982766, 0.02761865004478037, 0.7580350288936898, 0.12627145142681884]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedContinuousFunction_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[[0.7525538643717313, 0.8927982585642765, 0.4210826548629708, 0.5555735741904018], [0.17858267313856324, 0.5740398589900949, 0.40870781372307463, 0.909580303943474], [0.6526248633659558, 0.09019485320260956, 0.7337664840405913, 0.7468973066512248]], [[0.8095110813170832, 0.3807384017728831, 0.7392521239843176, 0.15203082514650257], [0.2727674585205959, 0.5812445347504359, 0.36571712468305095, 0.4687344146249067], [0.07893014883999883, 0.9361136247528623, 0.18014148375480055, 0.38083376317998774]]], [[[0.8632974181093962, 0.8016592808822394, 0.4983642968435722, 0.21819560532075755], [0.6398224952588359, 0.3432209152210235, 0.40024729342795706, 0.14286155656344757], [0.1841275903706291, 0.5727160205669964, 0.24924225685732992, 0.9555915720156908]], [[0.23101720296251682, 0.9537586137251547, 0.018561656010396055, 0.906796750480803], [0.14709464565937658, 0.9307564393767921, 0.4498701606097416, 0.22457529341173965], [0.6192439321514038, 0.21363503499184533, 0.6735094873176105, 0.7626716974408687]]], [[[0.0036716278032534744, 0.8434038122422595, 0.6382221799882414, 0.5271284448583157], [0.13326687583289554, 0.8333772952654219, 0.11011908516385194, 0.4007597680302103], [0.6917088620398234, 0.43902229159188766, 0.49699742476142716, 0.35272680611196183]], [[0.46959397470725106, 0.7360781515812774, 0.14415449819117165, 0.8678607148641525], [0.5431062391475556, 0.028681049971717165, 0.8589363499474931, 0.08848063545993701], [0.9405049960837757, 0.310728367912534, 0.769507495982587, 0.7517762842282233]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onFunctionOnBoundary_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      val = 2.334151e-01
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onFunctionOnBoundary_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      val = numpy.array([0.8882196658190877, 0.9449724811444659, 0.290189677020929] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onFunctionOnBoundary_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      val = numpy.array([[0.6062386890051852, 0.7602308024038457], [0.6222717023701925, 0.3989859465331288], [0.5079766923359104, 0.9593221240301741], [0.08051545932215776, 0.1275281797176182]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onFunctionOnBoundary_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      val = numpy.array([[[0.043675826386976535, 0.919615411297421, 0.027808465322756493, 0.27505330587245824], [0.8412813246837996, 0.05917923496352562, 0.8955603500026582, 0.8871120117624177]], [[0.058079346099858986, 0.9055741155888302, 0.3376080785604336, 0.5778695156221998], [0.36661632473039185, 0.800994330349761, 0.09128826289541214, 0.2382078284920388]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onFunctionOnBoundary_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = FunctionOnBoundary(self.domain)
      val = numpy.array([[[[0.2145524013903527, 0.24089792961072476, 0.9932916227285838, 0.012733683478825641], [0.5142011161806974, 0.16405399624392414, 0.8655791108036284, 0.3390126493887843], [0.21827060140896748, 0.27693953098301016, 0.39063049876734735, 0.5364451186486017]], [[0.3019752734996425, 0.3215151797488027, 0.7521368591896964, 0.46229900343725694], [0.19080450491207868, 0.23696866552500706, 0.007366183215039124, 0.7755022403454018], [0.7699223196351326, 0.8595117783157447, 0.8112482083998037, 0.13249220895293223]]], [[[0.7936166212806668, 0.18751400052329203, 0.27540072394534865, 0.36834425307838037], [0.6116108630046057, 0.18349119780248235, 0.14197713359273179, 0.6048285285550714], [0.551824051185101, 0.1473745682559563, 0.5008165740462018, 0.7751839264048938]], [[0.24879062392423779, 0.9609030000048912, 0.6109640073511698, 0.97674777683831], [0.8731693641272898, 0.0940807350810432, 0.41320254452468774, 0.673518865581775], [0.07920917584712195, 0.17757928557861602, 0.0551238374634162, 0.2097289863544426]]], [[[0.5434193211546073, 0.5768343694893899, 0.782221613606624, 0.6600146089466845], [0.9964243690573472, 0.12270068191803041, 0.9190358885047175, 0.7793851260350958], [0.906184505321438, 0.4106607249361641, 0.8981909427253475, 0.6281699603991318]], [[0.9963724531537237, 0.098881184736761, 0.6201804906367727, 0.9355731882315826], [0.5029881365361015, 0.31198238883404406, 0.0151272630521605, 0.06757634116237543], [0.4874585428336724, 0.6207127438960461, 0.4796838683083924, 0.7842448812877467]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedFunctionOnBoundary_rank0(self):
      """
      test interploation of constant rank 0 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      val = 7.288496e-02
      arg = Data(val,(),w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedFunctionOnBoundary_rank1(self):
      """
      test interploation of constant rank 1 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      val = numpy.array([0.25841556630698104, 0.1573540961311457, 0.9584429052895114] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedFunctionOnBoundary_rank2(self):
      """
      test interploation of constant rank 2 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      val = numpy.array([[0.6790821330183561, 0.7564906985837898], [0.5632568165696104, 0.08413153914498972], [0.31679030112704576, 0.5923791304740935], [0.8378495209362984, 0.5831567791999903]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedFunctionOnBoundary_rank3(self):
      """
      test interploation of constant rank 3 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      val = numpy.array([[[0.26631081093084663, 0.5742266726541079, 0.12753861068584116, 0.3584424932381006], [0.22597046302695967, 0.9783592503636329, 0.44750713810073517, 0.9372202753758945]], [[0.33545127011318, 0.07002843394376845, 0.3566614085333282, 0.43701344251938046], [0.9247924175230922, 0.41668241916224347, 0.9929687668000138, 0.407022205915038]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")
   def test_interpolate_onReducedFunctionOnBoundary_fromConstant_onReducedFunctionOnBoundary_rank4(self):
      """
      test interploation of constant rank 4 Data on ReducedFunctionOnBoundary
      """
      w_ref = ReducedFunctionOnBoundary(self.domain)
      w_arg = ReducedFunctionOnBoundary(self.domain)
      val = numpy.array([[[[0.8691434338403439, 0.6331464365176072, 0.8390644610809543, 0.4118416884147633], [0.4407318429840573, 0.877257517977619, 0.09230628707625521, 0.8285490604154563], [0.14981523179372958, 0.44792226923501977, 0.6546835167890781, 0.6348195060584106]], [[0.6591264484692713, 0.4208058158490403, 0.6060911404336305, 0.6805808577849918], [0.026865886829161467, 0.44763734162689484, 0.4285147051423571, 0.38633495777979754], [0.01197160601400138, 0.056387138751563426, 0.6235605595963108, 0.11883773519918384]]], [[[0.7165555921184351, 0.3448408207023229, 0.6403150812646062, 0.5074650351107], [0.08091639767885062, 0.8933089296805964, 0.8229294388271333, 0.6128834637309405], [0.6648810713352777, 0.7124734916410694, 0.7762079386938845, 0.06375120535252199]], [[0.2237681392590113, 0.8405738776179142, 0.14107681496964397, 0.6310989760807912], [0.6385013735546666, 0.464476194499888, 0.33573313838099095, 0.5268345106982509], [0.42682926579930924, 0.908175977790712, 0.5886079168303059, 0.2479777069953224]]], [[[0.4360956389458204, 0.46278050546015437, 0.581426071830637, 0.70790178534635], [0.3710587791530059, 0.5125796775186596, 0.12018187143158043, 0.40115255163024455], [0.13760882029302024, 0.5670063384417476, 0.7741185467231643, 0.1906377624589708]], [[0.5693881769080421, 0.7060457171887292, 0.37712214911798647, 0.42549139782130463], [0.5092197372656926, 0.4859093480715948, 0.5545563851154015, 0.725593183454442], [0.3046905769902506, 0.38681544902199805, 0.6303924508020141, 0.34799911513307624]]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=interpolate(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3, 2, 3, 4),"wrong shape of result.")
      self.assertTrue(Lsup(res-val)<=self.RES_TOL*Lsup(val),"wrong result")

   def test_grad_onFunction_fromConstant_onSolution_rank0(self):
      """
      test gradient of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = 7.936868e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onSolution_rank1(self):
      """
      test gradient of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([0.9809538471304521, 0.7826481973394941, 0.5359055119881645] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onSolution_rank2(self):
      """
      test gradient of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[0.5264876431264686, 0.5254916987976153], [0.017646227806965453, 0.024391719214934215], [0.01492010233387886, 0.8672061088884643], [0.02227531320782039, 0.9391080329087559]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onSolution_rank3(self):
      """
      test gradient of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[0.0638336511598484, 0.21381761049151415, 0.14615683522800482, 0.35430794861767967], [0.04936513833613787, 0.05126121312906007, 0.3651002200997059, 0.703585003125905]], [[0.10297211801323858, 0.42470299613921314, 0.8190282592661424, 0.07165012838752027], [0.6381827419590647, 0.5606822747364155, 0.8513170261447275, 0.4316264240928648]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedSolution_rank0(self):
      """
      test gradient of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = 7.488921e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedSolution_rank1(self):
      """
      test gradient of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([0.059211838949910556, 0.24173955393668212, 0.7686980774358706] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedSolution_rank2(self):
      """
      test gradient of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[0.8734085126677908, 0.0720449950676273], [0.31625739790011276, 0.7426342909522352], [0.09558658744083603, 0.38925303937358624], [0.9920865957485483, 0.17226707437488598]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedSolution_rank3(self):
      """
      test gradient of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[0.28552147486426316, 0.4103411107471152, 0.6522588475155091, 0.5046369427699156], [0.8648254675874936, 0.1618520354287829, 0.32785900988934524, 0.3926523876956626]], [[0.46620988315281997, 0.6524971397690709, 0.5952078893622407, 0.6554834658592923], [0.6136447466933086, 0.735004331274266, 0.27502278543859604, 0.1315703954470645]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onContinuousFunction_rank0(self):
      """
      test gradient of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = 5.269579e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onContinuousFunction_rank1(self):
      """
      test gradient of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([0.845177962924561, 0.14905292163118622, 0.8388568363520449] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onContinuousFunction_rank2(self):
      """
      test gradient of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[0.3116997247125688, 0.7004984855725862], [0.41150058328266437, 0.8970526833085086], [0.25511618503381306, 0.8358449760582636], [0.7722557165865692, 0.4287022961406195]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onContinuousFunction_rank3(self):
      """
      test gradient of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[0.04134745420061481, 0.504387672432746, 0.41088627257751564, 0.10086947371578103], [0.6549900611304361, 0.051277464724197475, 0.23854141764417613, 0.233826361853411]], [[0.39315819045674394, 0.4657074352867523, 0.4592996576631705, 0.24145767676623853], [0.6422061352555105, 0.8160257951107688, 0.4240265851482867, 0.7389537886064693]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedContinuousFunction_rank0(self):
      """
      test gradient of constant rank 0 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = 6.051475e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedContinuousFunction_rank1(self):
      """
      test gradient of constant rank 1 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([0.31935296515041944, 0.15700012210594494, 0.5841980565105753] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedContinuousFunction_rank2(self):
      """
      test gradient of constant rank 2 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[0.4019051515468195, 0.20612280864683086], [0.6803057116746596, 0.7410026278483777], [0.594420036738727, 0.10260115606085685], [0.2893726426244503, 0.27992958526754064]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onFunction_fromConstant_onReducedContinuousFunction_rank3(self):
      """
      test gradient of constant rank 3 Data on Function
      """
      DIM = self.domain.getDim()
      w_ref = Function(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[0.29443478707098425, 0.5152868350547829, 0.37874747649180285, 0.6371180027384423], [0.2695522164301283, 0.36887170751980125, 0.07091133713789688, 0.43394378103093756]], [[0.9785303460136326, 0.3291542849672704, 0.47880457099226736, 0.10385546999506912], [0.3548175157093879, 0.9330065510320062, 0.12260889668231656, 0.03656783986450385]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onSolution_rank0(self):
      """
      test gradient of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = 1.654730e-02
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onSolution_rank1(self):
      """
      test gradient of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([0.7885601234544278, 0.17116843752490885, 0.24084869120135122] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onSolution_rank2(self):
      """
      test gradient of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[0.12754753594192803, 0.2785389487377298], [0.8719229754433291, 0.9537732548894249], [0.5135880050569422, 0.09294527298920152], [0.1499038786783614, 0.16283599571553276]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onSolution_rank3(self):
      """
      test gradient of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = Solution(self.domain)
      val = numpy.array([[[0.4308409382228361, 0.29524685687192065, 0.2084676620865784, 0.5336570468830846], [0.43625031347446874, 0.43270235007431146, 0.9935416027428287, 0.44225659355109315]], [[0.42658497394894146, 0.9315894905872781, 0.8981898753969447, 0.06477218591866962], [0.4442874712674957, 0.6765953968396149, 0.682338503522219, 0.2525981822961223]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedSolution_rank0(self):
      """
      test gradient of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = 7.938059e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedSolution_rank1(self):
      """
      test gradient of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([0.13725522062510154, 0.8540510717261819, 0.4150096653408736] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedSolution_rank2(self):
      """
      test gradient of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[0.06560955350725395, 0.014980515593223664], [0.12379727352446901, 0.27095047904653524], [0.7665511415078269, 0.8981040892437574], [0.0047663000707844905, 0.6695831127844766]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedSolution_rank3(self):
      """
      test gradient of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedSolution(self.domain)
      val = numpy.array([[[0.6656716208966644, 0.6013490931897385, 0.31774720134989376, 0.9093635785050875], [0.40041167635118324, 0.6043035560991207, 0.33386573139887143, 0.6701907760151365]], [[0.1690533393445095, 0.7983908265260318, 0.3330832349818498, 0.12806983544865447], [0.08125317738329119, 0.44355855504296493, 0.875298380313076, 0.03545281427942437]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onContinuousFunction_rank0(self):
      """
      test gradient of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = 1.512591e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onContinuousFunction_rank1(self):
      """
      test gradient of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([0.2950829229504748, 0.39309332317737233, 0.0768718520453745] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onContinuousFunction_rank2(self):
      """
      test gradient of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[0.492956708317943, 0.5994258250266143], [0.17399852479832656, 0.13066756142273983], [0.9580607856166072, 0.7137201136844993], [0.35687704768905193, 0.1322470612721962]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onContinuousFunction_rank3(self):
      """
      test gradient of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ContinuousFunction(self.domain)
      val = numpy.array([[[0.7379911687606209, 0.9526205213896735, 0.017776453214885746, 0.5470556160731032], [0.6809106216633825, 0.31670663321571657, 0.05206427134064118, 0.6894556787047617]], [[0.1738331159027665, 0.5467201584578865, 0.10106010938221999, 0.12434149173432973], [0.6935924904126988, 0.8726902909661227, 0.44602639563271707, 0.9636765938289744]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedContinuousFunction_rank0(self):
      """
      test gradient of constant rank 0 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = 3.359234e-01
      arg = Data(val,(),w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),()+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedContinuousFunction_rank1(self):
      """
      test gradient of constant rank 1 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([0.347219154849003, 0.08922399119247204, 0.3192081287195233] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(3,)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedContinuousFunction_rank2(self):
      """
      test gradient of constant rank 2 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[0.27673002125271384, 0.038407754143686135], [0.9544262319245038, 0.3718893877821472], [0.08889976536735855, 0.5298883140395086], [0.8239855642862909, 0.7993343782624673]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(4, 2)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")
   def test_grad_onReducedFunction_fromConstant_onReducedContinuousFunction_rank3(self):
      """
      test gradient of constant rank 3 Data on ReducedFunction
      """
      DIM = self.domain.getDim()
      w_ref = ReducedFunction(self.domain)
      w_arg = ReducedContinuousFunction(self.domain)
      val = numpy.array([[[0.9485628855052038, 0.04284253201472121, 0.08486235607984038, 0.20068489136978185], [0.17658695575216832, 0.7319768182413037, 0.5240305565465515, 0.20858235741970155]], [[0.17966267066122366, 0.36514220687322674, 0.9248876071965655, 0.6719804014575069], [0.45897378517215015, 0.5801829493513729, 0.592642754829264, 0.5881306364541707]]] ,numpy.float_)
      arg = Data(val,w_arg)
      res=grad(arg,where=w_ref)
      self.assertTrue(isinstance(res,Data),"wrong type of result.")
      self.assertEqual(res.getFunctionSpace(),w_ref,"wrong functionspace of result.")
      self.assertEqual(res.getShape(),(2, 2, 4)+(DIM,),"wrong shape of result.")
      self.assertTrue(Lsup(res)<=self.RES_TOL,"wrong result")

